/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2022, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     RomLFNST.cpp
    \brief    LFNST Tables
*/

#include "Rom.h"


#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iomanip>

// ====================================================================================================================
// LFNST Tables
// ====================================================================================================================

#if EXTENDED_LFNST || JVET_W0119_LFNST_EXTENSION
const uint8_t g_lfnstLut[NUM_LFNST_INTRA_MODES] =
{  // 0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15  16  17  18  19  20  21  22  23  24  25  26  27  28  29  30  31  32  33  34  35  36  37  38  39  40  41  42  43  44  45  46  47  48  49  50  51  52  53  54  55  56  57  58  59  60  61  62  63  64  65  66  67  68  69  70  71  72  73  74  75  76  77  78  79  80  81  82  83  84  85  86  87  88  89  90  91  92  93  94  95  96
    +0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 33, 32, 31, 30, 29, 28, 27, 26, 25, 24, 23, 22, 21, 20, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2
};
#else
const uint8_t g_lfnstLut[ NUM_INTRA_MODE + NUM_EXT_LUMA_MODE - 1 ] =
{//0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63 64 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 86 87 88 89 90 91 92 93 94
   0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
};
#endif

#if JVET_W0119_LFNST_EXTENSION
const int8_t g_lfnst4x4[ 35 ][ 3 ][ 16 ][ 16 ] = {
 {  //0
   {
     {  105, -37, -16,   0, -55,  20,   8,   0, -12,   4,   2,   0,   2,  -1,  -1,   0 },
     {   46,  74, -45,  -4,  46, -43,   1,   2, -48,  -5,  19,   1,   1,   3,  -1,   0 },
     {  -14,  70, -15,  -6, -83, -16,  28,   3,  47, -27,  -7,   2,  10,   2,  -3,  -1 },
     {  -19, -35, -18,  15, -21, -78,  45,   6, -14,  69,  -4, -15,  17,   3, -13,   1 },
     {  -36, -16, -87,  32,  -3,  54,  27, -20, -17, -12,  35,  -5,  18, -11,  -8,   3 },
     {  -21,  11,  48, -12, -37,  13,  -5,   1, -86, -12,  13,   6,  62,  -4, -15,  -1 },
     {   12,  17,  25,  -1,  29,  32,  88, -33, -10,   4, -61,  10,  -5, -17, -14,  11 },
     {    2,  40,  -2, -10,  -8,  38, -32,  -1,   9,  85,  -2, -19,   9, -64,  12,  13 },
     {   11,  17,  28, 110,  -5,  -9, -21, -35,  -1,  -7, -11, -33,   3,   2,   6,   6 },
     {   22,  -5,  -6,  -1,  37,  -2, -10,   2,  53,  -5, -13,   1, 102,   6, -32,  -2 },
     {   11,  12,  41,   3,  11,  15,  35,  13,  23,  12,  84, -31, -19,   8, -59,   1 },
     {    2,  -5,   6, -20,  -7, -23, -15,-102,  12,  10,  34,  51,  -1,  -2,  -7,  27 },
     {    4, -22,   7,   6,   5, -37,   8,  18,   5, -50,  14,   3,  -1,-104,  -2,  25 },
     {    5,   7,  16,  39,   6,   6,  24,  28,  10,  21,  30,  85,  11,  -9,  40, -52 },
     {    5,  -4,  16, -21,  10,  -2,  34, -16,  10,  -8,  34, -48,  30,  11,  96,  16 },
     {    0,  -4,  -1, -19,   0,  -9,  -2, -42,  -1,  -9,  -3, -38,  -3, -27,  -7,-109 },
   },
   {
     {  113, -45, -15,  -1, -30,   7,   4,   2, -17,   7,   4,   0,  -2,   2,   0,   0 },
     {   39,  64, -41,  -3,  80, -19, -25,   3, -32, -21,  13,   4,  -8,   1,   5,  -1 },
     {  -12, -88,  18,  14,  77, -35, -11,   3,  -6,  24,  -4,  -3,  -8,   7,   2,  -1 },
     {  -18,  -1, -33,  13, -41, -92,  24,  20, -41,   7,  36,  -3,  16,  20,  -4,  -7 },
     {   -6, -17, -86,  26,  -7,  -4, -31,  14,  81,  -1,   0,  -5, -10,  -2,   9,  -2 },
     {  -34, -26, -47,  18,  -8,  67, -27,  -5, -71,  18,  35,  -6,   7,  -9,   6,   0 },
     {    8,  26, -20,  10,  22,  13,  61, -12,  13,  91,  -3, -35,  10,  -6, -29,   0 },
     {    4,  19,  15, -21, -21, -21, -87,   8,  -8,  68, -37,  -1,  24,   8,   8,  -2 },
     {   12,  19,  35, 105,  -2,   9, -10,  49,  -5,  -8, -12, -25,   8,   2,   0, -12 },
     {    3, -12, -11,  -2,  16,   7,   0,  -4,   8, -28, -12,   9, 114,   1, -42,   1 },
     {   -8,  -6, -32,  19,  -9, -19,  15, -21, -35, -14, -93,   6, -14, -51,  -5,  29 },
     {    6,   1,  18,  -3,   1, -17,  -8,  18,  13,  11,  48,   7,  14,-109,  12,  22 },
     {   -7,  -4, -13, -32,   8,  20,  24,  99,  -8,  13, -24,  52,  -8,  -3, -11, -22 },
     {    3,   7,  -1,  27,   6,   4,  27, -20,   5,  23,  -3,  67,  32,  17,  89,  14 },
     {   -2,  -6,  -6, -35,   5,   3,  17,  24,  -6, -17, -21, -80,  26, -10,  72, -32 },
     {   -1,   0,   1, -11,   2,   7,   2,  38,   1,  -3,   5, -27,   4,  30,   7, 114 },
   },
   {
     {  115, -28, -29,   2, -17,  28,  -2,  -4, -18,   1,   9,  -1,  -2,  -1,   1,   1 },
     {   15, 108, -39, -18, -21,  -3,  36,  -6,   2, -27,   6,   8,   2,  -2,  -3,   1 },
     {   13,  30, -16,  -4, 111,   4, -33,  -4, -25,  23,  11,  -4, -11,  -6,   7,   2 },
     {  -13,  14,  42, -14,   0, 111,  -9, -20,  18, -14,  23,   2,  -1, -13,  -3,   8 },
     {   39,  23,  99, -32,   9, -24,  14,  32, -21,  -7, -30,  12,   1,   1,  -5,  -5 },
     {   24,  10,  -3,  -6,  13,   0, -14,   3, 108,  24, -47,  -8, -10,  23,  12,  -7 },
     {   -3,  32,  11,  -4, -44,   4, -27,   7, -22,  91,   9, -43,   5, -12,  41,   1 },
     {    2,  -8,   9,  39,  25,  13, 104, -15,   6,  52,   2,   2,   2,  -5,  -9,   3 },
     {   -8, -28, -16,-108,   2, -10,  21, -40,   2,  27,   9,  25,   3,  -1,   2,  10 },
     {    2,   5,  -2,  -2,   8,   7, -10,   0,  -2,  11,   3,  -8, 111,  45, -37, -14 },
     {   16,   1,  27,   4,   3, -40,  -2,   4,  45,  -6,  99, -14,   6,  -4,  -2,  39 },
     {   11,  10,  28,  25,  -5, -24, -16,-110,  -9, -13, -24, -19,   3,   5,   4,  21 },
     {   -4,   0,   9,  -1,   0,   7,  12,  -7, -15,  -8,  29,  -1, -32, 104,  43, -35 },
     {    2,  13,   6,  30, -14,  -3, -29,  -9,   2,  36,   9, 112,   1,   7,  -2,   7 },
     {    2,  -9,   0,   5,  13,  -1,  15,   1,   7, -29,  -4,  18,  52, -30, 105,  -4 },
     {    5,   1,   9,   0,  -1, -13,  -5, -25,  14,   2,  29,   1,  -2, -40, -13,-112 },
   }
 },
 {  //1
   {
     { -124,  14,  16,   4,  15,  -8,  -1,   0,  17,  -1,  -2,   0,   3,   0,   0,   0 },
     {   10,  -8,  -5,   1, 122,   0, -19,  -4, -24,  10,   3,   0, -15,  -1,   2,   1 },
     {   -9,-119,  16,  12,  -6,  -7,  -8,   2,  21,  32,  -6,  -4,  -1,   5,   1,  -1 },
     {  -18, -17, -15,   4, -10, 107,  -1, -15, -44,  -1,  33,   0,   6, -26,  -1,   4 },
     {   18,  18,  31,  -6,  18,  60,  -4,  -7,  98,   2, -29,  -3, -18,  -8,   3,   2 },
     {   -8,   3,-108, -10,  -2,  -7, -18,  -6,  44,  36,  29,  -3,  -5,   2,   8,   1 },
     {   -3, -32, -33,   0,  12,   0,  -2,  -2,  20,-113,  -3,  19,  13,   8, -20,  -2 },
     {    0,  -7, -12, -12,  19,   2, 118,  11,   9,  15,   7,  15,   3,  -2, -30,  -2 },
     {    5,   8,  -8,  14,  13,  13, -10,   1,   3,  21, -35, -13, 112,  21, -28,  -9 },
     {    4,  -3,  34, -81,   3, -10, -17, -15,  16,   6,  67,  44,  36,   4,  -5,  10 },
     {   11,  10,  22,  83,   7, -15,   4,  -5,  28,  -7,  82, -17,  16, -14, -10,  10 },
     {   -1,   4,   5,   9,   0,  26,   5,  25,  -1,  -1,  25,   2, -13, 112,  25, -30 },
     {    0,   1,  -1,  -2,   2,   8, -20, 117,   4,   2,  10,  19,   2, -33, -12, -24 },
     {   -1,   8,   0,  23, -10,   3, -27, -20,  -5,  22,  -8,  57, -33,  18, -96, -11 },
     {    1,   3,  -5,  41,   1,   1,   7,  -3,   0,   8, -19,  98,  17,  -6,  61,  24 },
     {    2,   0,   3,   3,   1,  -7,   4, -32,   3,   0,   5,  13,   5, -29,  20,-118 },
   },
   {
     { -115,  35,  10,   4,  39,   7,  -7,  -2,  12,  -8,  -3,   0,   3,  -2,   0,   0 },
     {   39,  56, -21,  -4,  76, -45, -28,   4, -33, -33,  17,   8,  -7,   8,   9,  -1 },
     {   21,  88,   6, -14, -46,  58, -29, -10,  14, -20, -25,   9,   6, -16,   4,   6 },
     {  -21,  46,  -9,  -4, -68, -68,  14,  15, -24,  29,  41,  -2,  16,  22, -10, -11 },
     {   -2,   6,  56,  -1,   7,  33,  31, -24, -99,  10,  -7,  -9,  -4,  20,  -9,   2 },
     {   18,  39, -21,  -4,  39,  10,  58,  -3,  26,  72, -12, -35,  -3, -20, -37,   3 },
     {   17,  11,  99,   6,   5, -53,   3, -20,  46,  -6, -24,  -6,   2,  -2,  -4,   4 },
     {    0,  -9,   7,   0,   8,  -2, -91, -15, -10,  84, -12,  12,   0,  10,  -5,   4 },
     {  -11,  -4, -36,  16, -15, -44,   5, -16, -32,  -8, -89,   0,  16, -34,   2,  42 },
     {   -7, -12, -10,-110,   2, -12,  10, -55,   6,  -2,   5,  20,   0,   4,  -1,  13 },
     {   -8,   6,   4,  -1, -16, -12,   1,   1,  -7,   9,   9,  -1,-115, -41,  25,  10 },
     {    1,   6, -28,  44,  -8,   1,   1, -90,  18,  -7,  11, -38, -15,  53,  23,   7 },
     {    1,  -5,   7,  22,   1,   1,  -7, -44, -12,   1,  61,   0,  39, -90,   4,  17 },
     {    5,  11,   5,  12,  10,   5,  38,   5,   8,  36,   1,  65,  17,  13,  89,  22 },
     {   -1,  -2,   7, -33,   0,  -1, -14,  26,  -7,   6,  -5, -91,  21, -10,  73,  -3 },
     {   -1,   0,  -8,   2,   0,  -8,   7, -28,  -7,   5, -29,  15,   2, -25,  19,-115 },
   },
   {
     { -121,  38,  12,   2,  11,  -4,   0,   0,   9,  -3,  -1,   0,   4,  -1,   0,   0 },
     {  -31,-112,  49,  10,   2,  12,  -3,  -2,   3,   9,  -4,  -1,   1,   3,  -1,  -1 },
     {    3, -16, -37,  13, 107, -33, -39,   6,  20,   0,  -3,  -2,  -2,   2,   3,  -1 },
     {  -22, -29, -96,  63, -33,  19,  13,  -7,  -5,   6,  12,  -7,   1,   1,   3,  -2 },
     {   -1, -13,   3,  12, -12,-105,  60,  18,  -5, -25,  16,   2,  -1,   2,  -2,   0 },
     {   -6,   5,   6,   1,  18,  -7,   0,  -2, -97,  55,  28, -14, -44,  21,  12,  -4 },
     {  -15, -22, -36, -62, -44, -33, -73,  39, -15,  -8,  -1,  15,  -3,  -1,   7,   0 },
     {   11,  22,  41,  86, -26, -24, -64,  26,  -2,   2, -21, -12,  -1,   1,   0,  -5 },
     {    1,   3,  -8, -17, -17, -28,   7,  -6,  50,  96, -28, -35,  10,  31,  -7,  -9 },
     {   -2,   0, -11,   6,  10,  11,  28,  45, -25,  10, -84,  61,  -9,  16, -37,   5 },
     {   -3,  -4,  -7,  -6,  -9, -29, -19, -98, -28, -22, -57, -14,  -3,  -5, -23,   9 },
     {    1,   0,   0,   0,   9,  -4,  -2,  10, -39,  27,  13,  -7,  97, -50, -40,  17 },
     {   -1,  -2,  -1,   1,  -3,  -6,   5,   5,  18,  28,  -7,   0, -60,-101,  -2,  36 },
     {    2,   4,   6,  14,  -7, -14, -18, -42,  18,  32,  46,  96,   2,  15,  -8,  28 },
     {   -1,   0,  -1,  -3,  -1,   7,  -9,  13,  11, -15,  33, -31, -32,  30,-102,  37 },
     {    0,   0,   0,  -1,  -1,  -2,  -4, -10,   4,   5,  16,  19, -14, -29, -47,-111 },
   }
 },
 {  //2
   {
     {  112, -41,  -6,  -4, -42,  -6,  10,   1,  -8,  12,   4,  -2,  -4,   1,  -2,  -2 },
     {   44,  49, -22,  -2,  76, -52, -23,   2, -33, -29,  19,   6,  -8,   7,   9,  -2 },
     {   31,  38,   0,  -6,   4,  84, -44, -13,  13, -38, -46,  12,  -6, -23,  10,  15 },
     {   -2,  88, -22,  -7, -80, -21, -12,   8,  -8,  15,  16,   8,  17,   5,  -7,  -4 },
     {   15,  38,  -2,  -7,  35,  24,  49, -15, -17,  70, -31, -39,  -4,  -7, -50,   6 },
     {  -15, -13,  20,  -1, -14,  33, -10,  -9,-113,   0,   7,   8, -10,  33,   8,   4 },
     {   13,  26, 111,  -8,  -8, -26,  16, -34,   8, -25,  -6, -11,   5,   3,   0,  -7 },
     {   10,  10,  16,  -5,  16,  54, -11,  18,  23,   8,  82, -13,  -5,  26, -16, -62 },
     {    2, -14,  29,   0,  11, -19, -89,  -1,  11,  72, -17,  28,   5,  15, -17,   2 },
     {    5, -11,   1,   1,  13,   9,  -2,  -4, -20,  -5,  14,   6, 117, -39, -13,  -5 },
     {   13,  21,  21,  33,  16,  20,  40,  32,  12,  38,  13,  53,  14,  20,  65,  39 },
     {    5,  -1, -18,  -1,   3,   2,   6, -10,  21, -20, -39, -14,  42, 107,  -6,  -5 },
     {   -4,  -4, -21, -73,   4,   4,   2, -81,  12,  23,  31,   4,   3,   4,  45,  17 },
     {   -2,  -4,  11, -90,   7,  -1,  22,  54,  -1, -12, -14,  56,  -3,   4, -34,   0 },
     {   -1,   3, -12,  14,   2,  -6,  21, -29,  -8,  12, -41,  51,  -4, -12,  23, -98 },
     {   -1,   2,   9, -38,   0,  -4, -14,  53,  -9,  18, -29, -69,  12,  -9,  70, -28 },
   },
   {
     {  -87,  62,  -9,   3,  60, -31,  -8,   3,  -6,  -9,  11,  -1,   3,   2,  -1,  -1 },
     {   55,  19, -34,   2,  20, -80,  40,   1, -34,  40,  11, -15,   1,   2, -15,   4 },
     {   50,  23,   5, -13,  14, -16, -59,  25,  12, -58,  64,   0, -14,  20,   2, -22 },
     {   -2,  65, -64,  11, -71,   9,  15,  14,  44,  -9,   2,  -7,   6, -18,   4,   0 },
     {   43,  38, -23,  10,  35,  21, -29, -24, -29, -20, -62,  43,   8, -23,  39,  18 },
     {   -4,  32,  29, -34, -36,   6, -57,  36, -44,  63,   2,   9,  30, -20, -16,  -3 },
     {  -11, -18,   4,  25, -13, -29,   6, -19,   4,  13,  24,  49,  28, -27,  53, -80 },
     {    8,  21,  59, -52,  -9, -53,  11,  -1,  62,  -4, -43,  20, -18, -10,  11,   9 },
     {   15, -24, -31,   3,  58,  14, -26,  33,  69,  42,  -8, -10,  17, -41, -21, -15 },
     {   28,  56,  43,  -4,  23,  54,  40, -35,   3,  16,  11, -33,  -1,  -5, -15, -51 },
     {    2,  -9,  20, -14,   3,  -8,  18,   1, -11, -41,  23, -46,  64, -70,  24,  35 },
     {    5,   8,  19,  47, -12, -19, -30, -70,  25,   2,  18,  25,  34,  -3, -62,  31 },
     {   13,  18,  27,  29,  16,  23,  35,  41,  21,  27,  39,  43,  27,  35,  37,  52 },
     {    1,   8,  26,  62,  -9, -14, -36,  -8,   8,  31,   1, -55, -53, -18,  51,  19 },
     {    6,   6,  31,  63,  -6, -17,   5,  58,  -7, -32, -55, -17,  34,  20, -28, -32 },
     {   -1,   2,  15,  23,   1,   7,  26,  32, -20, -16,  19,  46, -61, -73, -40,   6 },
   },
   {
     { -126,  15,  10,   3,   9,   4,   5,   0,  11,  -1,   0,  -1,   7,  -1,   0,   0 },
     {   15,  26,  -9,  -1, 111,  13, -19,  -4,  46,  -7, -15,  -3,   8,  -8,  -7,   0 },
     {   -8, -79,   4,   4,  20, -85,   5,   6,  19, -42,   6,   6,   8,  -9,   4,   3 },
     {  -12,  -3, -73,   5, -22,  -9, -86,   8,  -6,  -8, -48,  11,   2,  -5, -14,   8 },
     {    3, -44,  -6,  -3, -26,  26,   3,   0,  79,  55,  -5,  -3,  55,  24, -11,  -4 },
     {  -11, -46, -20,   3,  44,   2, -10,   5, -49,  68,  19,   2, -35,  52,  15,  -3 },
     {    4,  11,   0,  66,   2,   8,   9,  85,   7,   1,  15,  58,   7,  -2,  14,  25 },
     {    3,  59, -36, -20,  -6, -57,  -6, -12,  13,  11,  58,   9,  35,  36,  39,   4 },
     {    6,  34,  45,  25,   1, -54,   1,  13, -15,  25, -56, -15,  19,  48, -48, -15 },
     {    2, -13,  52,  -2,   2,  26, -38, -22, -13, -35, -24,  12,  32,  50,  63,  37 },
     {    1,  12,  15,   8, -22, -13, -16,   5,  67,  -2,  -9, -10, -93,  29,  29,  -9 },
     {   -1,   5,   6, -59,   0, -11,  20, -15,   7,  18, -20,  73, -23,   2, -33,  66 },
     {   -2,  -5, -52,  -4,   2,  26,  53,   7,   1, -62, -11,  -1,  -3,  74, -25,  -9 },
     {    0,  -6,  37, -24,  -2,  15, -58,  18,   2, -25,  65,  20,  -3,  24, -62, -28 },
     {    1,   0,  -7,  79,  -3,   1,  -5, -73,   6,  -2,  31,  -2,  -9,   6, -36,  48 },
     {    0,   0,   0, -21,   0,   0,  -7,  48,   1,  -2,  17, -81,  -2,   4, -16,  81 },
   }
 },
 {  //3
   {
     {  108, -44,  -5,  -3, -49,   0,  11,   1,  -9,  14,   1,  -2,  -3,   1,  -3,   0 },
     {  -45, -30,  20,   0, -76,  63,  11,  -3,  45,  17, -24,  -1,   8, -14,  -5,   3 },
     {   29, -14,  11,  -2,  38,  62, -49,  -4,  28, -71, -15,  15, -19,  -8,  28,   2 },
     {   20,  97, -39,  -5, -55,   2, -25,  11,  18, -21,   7,   8,  14, -14,   4,  -1 },
     {  -24, -30,  20,   2, -46, -35, -19,  17,  -2, -50,  68,  14,  -3,  44,  27, -24 },
     {   -7,  15,  20,  -4, -17,  52, -29,  -9,-103,  12,  -5,  14,  14,  25,   2,   5 },
     {   19,  36, 104, -39,  -3, -26,   1, -27,  19,  -2, -17,   5,   3,   9, -17,  -1 },
     {   12,  15,   2,  -6,  23,  49, -10,  15,  26,  28,  59, -35, -14,  40, -60, -31 },
     {    1,  24,  -6,   8,  -8,  23,  90, -27, -16, -50,  -4, -26, -43,  31,   8,   6 },
     {   -7,  10,  -2,   0, -16, -13, -34,   8,   2,  40, -20,   6,-109,  15,  22,   4 },
     {   15,  22,  18,  21,  21,  25,  32,   7,  20,  50,  24,  34,  19,  13,  84, -24 },
     {   -2,   3,  29,  -3,  -3,  10,  13,  14, -28,  -4,  52, -19, -30,-100,   6, -15 },
     {    9,  12,  36, 101,  -5,  -9, -11,  42,  -4, -10, -29, -39,   6,   4,  -8, -10 },
     {   -1,   0,   6, -47,   7,   7,  30, 109,  -8, -11, -27,  18,  -1,   8,  -5,   0 },
     {    2,   2,  10, -15,   1,   3, -13,  15,   7,   9,  27, -61,  12,  11,  36,  97 },
     {   -3,  -4, -10, -41,  -3,  -6, -16,  -5,  -7,   1, -33, -81,  14,   4,  44, -66 },
   },
   {
     {   98, -53,   1,  -4, -56,   3,  19,  -1,   1,  20,  -1,  -5,  -4,  -2,  -5,  -2 },
     {   34,  24,  -8,  -4,  35, -87,   3,   9, -14,   1,  64, -12,  -4,  15, -13, -29 },
     {    9, -19,  33,  -9,  -9,  35, -58,  15,  34, -68,  52,   5, -14,  22,   7, -26 },
     {   26,  79, -44,   4, -48,  11, -57,  23, -18,  12,  -4,  29,   0,  -2,  -1,   6 },
     {   47,  -7, -17,   8,  57,  -9, -12, -16, -18, -50, -36,  35,   9, -19,  63,  19 },
     {  -18,   0,   3,  22, -27,  -7,  14, -24,  15,  20,   8,  42,  26, -31,  47, -88 },
     {   -1,  22, -51,  -2, -19,  -7,  45,  25,  69, -41,   7, -43,   1, -35,  23,  12 },
     {   42,  45,  15, -10,  48,  47,   3, -33,  21,   8, -21, -30, -12, -29, -41, -46 },
     {   -5, -21, -65,  22,  18,  53,  20,  23, -58,  -4,  21, -29,  -7,  29,  11, -38 },
     {   -1, -45, -38,  39,  38,  -9, -42,  44,  51,  34, -16,  25,  -2, -15, -35,  -8 },
     {   22,  29,  30,  18,  29,  39,  36,  33,  30,  43,  37,  28,  27,  37,  34,  30 },
     {    1,  12, -24,  36,  -9,  -8,  29, -48,  31, -21, -17,  39, -52,  68, -20,  -4 },
     {    2,  -3, -11,  66,  -6,  12,  -9, -57,  -7, -15,  48,  -6,  52, -28, -32,  36 },
     {    2,  12,  41,  51,  -8,   1,  38,  53, -33, -33,  -1,  29, -42, -46, -27,  -2 },
     {   -3,   0,  12,  39,   0,  -2, -31, -22,   3,  41,  18, -43, -76, -20,  55,  17 },
     {   -9, -13, -32, -59,  12,  22,  16, -22,  -2,  13,  52,  57, -40, -43, -12,  21 },
   },
   {
     {   98,  13, -20,  -6,  70,   7, -22,  -5,  11, -14, -10,   1, -12, -13,   0,   2 },
     {  -51,  82,   6,  -4,  45,  65,  -6,  -8,  15,  10, -16,  -5,  -2,  -8,  -9,   1 },
     {  -57, -36, -26,   4,  64, -52, -30,   5,  50, -22,  -4,   6,  19,  -3,   7,   4 },
     {    5,  -8, -67, -14, -44,  25, -68, -10,  20,  43, -33,   1,  22,  17, -13,   4 },
     {  -21,  26, -33,  -2, -12, -21, -44,   1, -72, -59, -14,  14, -42, -23,  12,  14 },
     {   12,  13,  30,   5, -37, -10,   1,  -2,  38, -56, -59, -10,  37, -48, -42,  -1 },
     {   -7,  -9,   9, -68,   8, -17,   2, -85, -13,  -5,  -5, -56,  -5,  10,  -5, -21 },
     {   16,  60,   2, -12, -16, -30,  -9, -10,   7, -26,  31,  14,  69,  44,  48,  12 },
     {   -1, -11, -25, -32, -28,  34,  11, -17,  49, -21,  52,  15, -19, -64,  44,  15 },
     {   -2, -46,  40,   3,   9,  48, -10, -18,  -8, -19, -49,  12,   7,  27,  55,  56 },
     {    2,  -1,  -2,  55, -11,  16, -22,  10,  14, -19,  -2, -73, -12,   5,  53, -59 },
     {   -6, -34, -30,   3,  23,  46,  13,   4, -54, -24,  25,  -9,  78, -23, -18, -20 },
     {    0,  16, -27,  20,   7, -35,  42, -18, -21,  54, -46,  -7,  23, -55,  49,  13 },
     {   -2,   3, -68,   0,  -4,  11,  75,   0,  18, -41, -32,   1, -19,  51,  -5,  -1 },
     {    2,   6, -14,  57,  -3,  -5,  -1, -33,   4,  -3,  37, -51,  -2,   2, -34,  81 },
     {    0,  -3,   3,  61,  -2,   4,  -4, -81,   3,  -4,   5,  65,  -4,   3,  -9, -40 },
   }
 },
 {  //4
   {
     {  122, -26, -13,  -4, -14, -10,  -1,   1, -18,   1,   4,   1,  -9,   1,   2,   0 },
     {   14,  34, -17,   0, 109, -15, -25,  -2, -12, -36,   0,   3, -20,  -5,  10,   2 },
     {  -21, -86,  19,   7,  29, -72,  27,   6,  -7,   4,  26,  -3,  -4,  22,   8,  -6 },
     {   13, -49,  17,  -2,  37,  43,   4,  -7,  87,  14, -36,  -9,   0, -31, -17,   2 },
     {   10,  21,  66, -13,  31,  30,  54, -24, -45,  47,  -5, -21, -14,  19, -25,  -5 },
     {    7,  31, -54,   9,  20, -33,   6,  11,  29,  81,  14, -10,  38,  21, -31, -10 },
     {   20,  51,  45, -12, -12, -53,  43, -12,  53, -45,  13,  -7,  33, -12,  -1,  -5 },
     {   -3,  -3,  17, -17,  -8, -52, -41, -22, -23,   5, -87,   4,  16,  -6, -44,  23 },
     {   10,  -3,  45,  -8,   7,  20, -59,  -2,  18,   0,  11,  19,  44,  83,  20, -11 },
     {    9,   9,  31,  82,   6,   3,  12,  81, -10,  -4, -24,  28,  14,  -5, -14,   2 },
     {    1,  25,  13,   7, -19, -27, -14,  12,  41,  19, -12,   2,-102,  38,   5,  -1 },
     {   -2,  -6,  17, -10,  -5,   3, -42,  18,  -3, -21,  51, -18, -15, -18, -90, -43 },
     {    1,   7,  38, -30,   4, -15, -28,   9,  -1,  50,  37,  66,  -3, -60,  21,  25 },
     {   -3,  -6, -28, -44,   5,  15,  44,  16,   6, -26,  -2,  76,  -6,  39, -48,  30 },
     {    2,  -1,   0,  74,  -3,   2,  -8, -76,   7,  -4,  33,  21,  -5,   5, -28,  51 },
     {   -1,   0,  -8,  20,  -1,  -4,  10, -43,  -3,   3, -23,  60,  -4,  -9,   6, -98 },
   },
   {
     {   85, -58,   6,  -3, -66,  32,   6,  -1,  11,   7, -10,   1,  -4,  -3,   2,   1 },
     {   55,   8, -21,   0,  24, -73,  31,  -1, -52,  52,   6,  -9,   8,   0, -18,   5 },
     {   42, -33,  25,  -6,  65,  -1, -45,   4, -24, -54,  37,   6, -20,  33,   7, -11 },
     {  -41, -66,  52,  -3,  24,  27,   1, -13, -48,  51, -29,   6,  16,  -3,  -9,   6 },
     {   28,   9, -30,  11,  48,  21,   6, -18,  -6, -17, -68,  24,  10, -49,  57,  13 },
     {  -15, -59,  -4,  13,  33, -36,  60, -10,  67, -10,  -8, -19, -37,   4,   0,   3 },
     {  -12, -15, -69,  47,   4,  53,  36, -20, -42, -19,  25, -23,   8,  30, -17,   6 },
     {  -26, -29,   2,  14, -42, -48,  18,  -4, -30, -30,  21,  33,  33,   9,  72, -24 },
     {    0,  30,  14,  -7, -19,  -1,  20, -28, -22,  -3, -50,  31, -64,  74,   7,   8 },
     {  -25, -41, -55,   3, -21, -35, -61,  29, -15,  -8, -26,  11, -32,  -9, -18,  48 },
     {    1,   6,  32, -35, -12, -12,  29,  -5, -19, -60, -15, -42,  35,  -2, -13,  75 },
     {   -7,   0, -12, -41,   4,  23,  18,  72, -19,  21,   5, -52, -34,   7,  60,  -2 },
     {   -8,   7,  23,  10, -18,   1,  20, -16, -43, -26,  22,  -2, -71, -78, -16, -18 },
     {   -6, -13, -28, -68,   8,   9,  37,  35,  -8, -28, -15,  53,  16,  -4, -46, -42 },
     {    7,  11,  24,  43,  10,  16,  32,  61,   6,   9,  30,  65,  -5,  -2,  -2,  58 },
     {   -5,  -6, -25, -64,   5,  13,   0, -58,  11,  29,  57,  31, -14,  -8,  25,  45 },
   },
   {
     {  103, -46,  -1,  -4, -55,  -8,  14,   1,  -1,  18,   6,  -3,  -4,   0,  -6,  -3 },
     {   60,  37, -17,   0,  66, -46, -28,   1, -26, -47,  19,  10,  -6,   5,  22,  -1 },
     {   31,  18,  -5,  -3,   3,  87, -33,  -6,  21, -34, -56,   8,  -9, -28,  13,  22 },
     {   22,  76, -14,  -5,   9,   5,   6,  -6, -25,  70, -20, -17,   1,  -6, -59,   2 },
     {   18, -35,  27,  -2,  80,   6,  40, -10,  58,  13,  -2, -27,  -2, -27, -27, -17 },
     {   11,  30,  39,  -4,  -9,  49, -22,  11,  24, -13,  69,  -8, -13,  53, -21, -41 },
     {   -1, -50,   9,   0,  40,  42,   8, -12, -83,  10, -14,  -6, -14,  52,  -7,   7 },
     {    6,  39,  75,  15, -20,  -3,  68,  -7, -34, -39, -13, -18,  -4, -21,  17,   0 },
     {  -14, -20,  12, -20, -19, -28, -42, -18, -16, -47, -17, -31, -44, -26, -73, -16 },
     {   -8,   0, -55,   3,   1,  29,  40,  38, -23,  -4,  41,   2, -65, -50,   7, -22 },
     {   -3,  11, -34, -54, -10,   6,  12, -73,   3,  -1,   3, -58,   2,   9,  45, -32 },
     {    0,   6, -15, -22,  -2,  -4,  20,  31,  21, -19,  14, -56, -17,  33,  -9,  93 },
     {   -1,   4,  40,  10,   3, -17, -29, -29,  15,  48,  -1,  12, -88,   1,  40,  24 },
     {    2,   4, -22,  72,  -5, -12,  -4,  32,  16,  -3, -48, -56, -18,  37,  17, -42 },
     {    0,  12, -32,   8,  -6,  -8,  52, -43,  29, -29, -22,  61, -32,  50, -32,  -1 },
     {    1,  -4, -16,  83,  -4,  14,  -9, -64,  -6,  -6,  47, -18,  17, -23, -15,  33 },
   }
 },
 {  //5
   {
     { -118,  32,   8,   3,  32,  12,  -3,  -1,   9,  -6,  -4,  -2,   3,  -1,   1,   0 },
     {  -33, -38,  17,   1, -99,  33,  26,   1,  19,  38,  -5,  -5,   7,  -2, -10,  -1 },
     {   10,  99, -19, -11, -56,  19, -23,  -4, -42,  -7,   1,   6,  -4,   7,   5,   3 },
     {  -25, -13,   4,   5, -20, -82,  24,  12, -57,   7,  54,   2,   0,  36,   9, -12 },
     {    2,  31, -35,   6, -22, -57,  -5,  25,  88,  -9,  17,   8,  35, -18,  -9,  -6 },
     {   10,  31, -12,  -6,  33,  -5,  28,  -7,  -5,  98,  -5, -25,  16,   4, -52,  -5 },
     {  -15, -33, -98,  15,   3,  27, -38,  34, -15,  26,  19,  17, -20,   3,   7,  -2 },
     {   -7, -15,  17,  22, -11, -39, -68,   9, -20,  17, -72,  24,  33,  17, -20,  26 },
     {   10,   3,  24,  -4,  11,  33, -34,  -7,  31,  19,  36,   9,  38,  88,  30, -22 },
     {    0, -12, -18,   0,   6,  26,  20,   3, -39, -29,  12,  -1, 107, -26, -12,  -6 },
     {    0,  -4,  47,  -9,   2,   6, -59,   0,  -6,  20,  65,  32,  -6, -60, -32, -14 },
     {    0,   4,  -1, -10,   6, -10,  12, -28,   1,  46,   4,  35,  18, -35,  86,  53 },
     {  -12, -22, -31, -72, -10, -23, -36, -76,   3,  -6,  -2, -40,   5,  -2,  -3, -13 },
     {   -4,  -4, -10, -62,   5,   5,  31,   0,   5, -14,   0,  85,  -7,  29, -45,  33 },
     {   -1,  -1,  17, -76,   5,  -3,  -4,  80,  -8,  13, -27, -14,   5, -12,  33, -36 },
     {    0,  -3,   8, -19,   0,   7, -16,  35,   4,  -9,  33, -59,   2,  13, -15,  97 },
   },
   {
     {  -86,  49,  -1,   3,  71, -27, -11,   0, -13, -16,  13,   0,   2,   8,   1,  -3 },
     {   36,  23, -21,  -1,  12, -73,  33,   0, -43,  66,   5, -11,   9,  -4, -32,   7 },
     {   65, -27,  10,  -3,  55, -29, -23,   4, -18, -55,  42,   3, -12,  33,  17, -13 },
     {   14,  58, -50,   5, -35, -33,  22,  11,  66, -39,  23, -12, -26,   8,   4,  -5 },
     {  -16,   2,  36, -14, -31,   9, -36,  14,   5,  18,  57, -16, -21,  42, -77,  -9 },
     {   29,  71, -17,  -7, -24,  27, -66,  17, -45,  -2,   5,  25,  33, -15,  14,  -5 },
     {   27,  11,  33, -25,  45, -12, -41,  16,  70,  25, -20,  -1,  -3, -59, -19,   2 },
     {  -29, -21,  42, -21, -47, -61,  -9,  31,  -2,   5,  31,   7,  16, -25,  57, -26 },
     {    4,  24,  29, -26,  -9, -17, -11,   2,  -8,  16, -60,  33, -74,  51,  25,   7 },
     {    3,   3,   3,  36,   2,  -3,   4, -52,  18,  19,  10,  70,   1,  -1, -11, -81 },
     {  -11, -33, -58,  47,  12,   8, -43,  48,   2,  45,   8, -15, -44,   0,  25, -20 },
     {  -17, -34, -26,  18, -24, -47, -40, -13, -15, -43, -22,  42, -14, -33, -47,  40 },
     {   -3, -16, -15,   7,  -1, -25, -29,  11,  39,  11, -36,   5,  77,  73,   0,   6 },
     {   -8, -18, -37, -68,  12,  14,  32,  63,  -9, -15, -13,  40,   6,  -3, -31, -40 },
     {    3,   3,  10,  20,  12,  16,  27,  31,  19,  24,  49,  74,   4,   8,  14,  69 },
     {   10,  18,  48,  74,  -3,  -4,  25,  66, -12, -28, -34,  -6,   6,  -4, -25, -20 },
   },
   {
     { -103,  42,   3,   3,  53,  12, -15,  -2,   1, -23,  -8,   2,   4,   1,  11,   3 },
     {   -3, -37,  18,  -1, -35,  85, -27,  -1,  41, -51, -20,   9,  -7,  -2,  26,   0 },
     {   53,   3,  -9,   2,  64, -29, -34,   6,   0, -70,  25,  17,  -3,   6,  38,  -2 },
     {   27,  17, -33,   6,  27,  15,  17, -15,  -5,  18, -83,  11,  21, -56,  32,  27 },
     {  -37, -41,  -5,  16, -43, -43,  33,   6, -30, -15,   9,  30,  17,  -3,  76,  13 },
     {  -12, -61,  28,  -3,  44, -26,  48,  -4,  66,  -6,   3, -23,  17, -35, -17,  -7 },
     {  -22, -35, -33, -14, -26, -48, -53,  -6, -17, -39, -38, -16, -26, -34, -47,   6 },
     {   -8, -17, -46,  26,   2, -12, -47,  38,  55,  40,   4,  48,  39,  23, -10,  15 },
     {    6,  74,  27,  -8, -53, -37,  -2,  -8,  56, -25,   8,  11,  20, -38,   2,   3 },
     {    4,   0,  24, -15,   2,   0,   5,  32,  -2,  -9,   5, -25,  -2,  17,  -9, 116 },
     {    3, -20,  85, -14,  13, -11, -45, -30, -30,  25, -16,  51,  29,  -7, -10,   4 },
     {    1,  -3, -13, -41,  -6, -17, -21, -55,  18,   6, -21, -56,  48,  59,  31,  -2 },
     {   -2,  -4,   5, -42,   0,  14, -36,  63, -20,  23,  28, -45,  35, -54,  29, -21 },
     {    3,   0, -21,   4,  -6,  24,  28,  -3, -35, -48,  15,  12,  89,  -1, -52,   0 },
     {   -8,  -9, -40, -61,   5,  20,   1, -57,   7,  16,  58,  36, -17, -34,   2,  28 },
     {    1,  -4,   2,  87,  -3,   9, -33, -49,  -6,  13,  39, -43,  10, -32,   8,  21 },
   }
 },
 {  //6
   {
     { -112,  39,   4,   3,  44,   7,  -8,  -1,   2, -12,  -3,   0,   2,   0,   3,   0 },
     {   48,  14, -13,  -1, 102, -33, -19,  -1, -10, -39,   7,   4,  -9,   0,  12,   0 },
     {  -14, -70,  33,   3,  31, -26,  34,  -2,  79,   2,  -3,  -9,  21, -21, -11,   0 },
     {  -25, -31,  14,   4, -15, -79,  29,   5, -57,  -5,  49,  -1,  -2,  36,  13, -11 },
     {    8,  59, -16,   1, -39, -45,  -6,  12,  48, -42,  28,  10,  60, -17,   6,  -5 },
     {   13,  42,  -2,  -4,  29, -12,  19,  -4,   6,  87,  15, -20,  33,  32, -49, -16 },
     {   15,  42,  69, -30,   0,  17,  64, -38, -17, -29,  10, -20, -14, -30,  -8,   1 },
     {    9,  -3,  22,   1,   9,  55,   5,  -3,  25,  -6,  39,  -1,  26,  70,  67, -20 },
     {   -4,   9, -42,   5,   1,  10,  19,  13,  29,  23,  75, -11, -66, -43,  18, -18 },
     {    6,  31,  58,   5, -15, -45, -44,  -7,  40,  26, -20,   8, -62,  24,  24,   8 },
     {    2,  12, -12,  10,   8, -17,  35,   1, -16,  48, -34,   1,  25, -37,  85,  37 },
     {   -1,  16, -54,   1,  -8, -13,  66,  -5,  28, -28, -49, -14, -37,  57,  -3,   3 },
     {   10,  16,  34,  84,   5,  14,  27,  78,  -9,  -9,  -9,   9,  -5,  -3, -13, -10 },
     {    0,  -2,  -2,  22,  -7, -10, -22, -12,  -8,  -7, -27, -92,   7, -16,  20, -71 },
     {    2,  -1,  -8,  80,  -4,   4, -10, -68,   2,  -7,  27, -27,   3,   7, -13,  59 },
     {    0,   0,   6, -36,  -1,   3, -14,  60,   2,  -8,  15, -75,   0,  13,  -3,  72 },
   },
   {
     {   88, -46,   1,  -3, -72,  23,  11,   0,  13,  17, -11,   0,  -1,  -9,  -1,   2 },
     {   47,  17, -15,  -1,  25, -75,  21,   1, -57,  49,  18,  -8,  13,   8, -26,   3 },
     {   26, -76,  41,  -3,  65,  15, -17,  -9, -36, -29,   9,   5, -10,  29,  11,  -7 },
     {   49,  39, -22,  -2,   0, -28, -17,  11,  32, -71,  47,   0, -26,  33,  17, -19 },
     {    9, -25, -17,   7,  46, -35,  58, -13,  64,   0, -41, -14, -41, -23,  11,  15 },
     {   39,  31, -46,  10,  36,  32,  -8,  -6, -30, -23, -43,  15,  43, -42,  44,   8 },
     {   25,  17,  54, -25,  25, -20, -69,  19,  39,  27,  -1,  12,  -1, -58,  -6,   0 },
     {   24,  23, -24,   7,  36,  58, -13, -10,  19,  10,  -4, -15, -13,  23, -89,  17 },
     {    8,  37,  19, -23, -10,   1, -13,   0, -23,  19, -68,  33, -62,  53,  22,   8 },
     {    9,  23,  56, -40,  -6, -14,  40, -31,  14, -43, -26,  13,  63,  18, -29,  15 },
     {   14,  25,  19,  -1,  27,  38,  30,  -1,  36,  59,  27,  -8,  29,  39,  49, -40 },
     {    2,   3,   6,  64,  -6, -12, -10, -78,  10,   9,  11,  70,  -5,  -6, -10, -23 },
     {   -8, -35, -36,  19,   2, -33, -42,  34,  38,  10, -38,  22,  55,  54,  -6,   0 },
     {    8,  16,  33,  61, -11, -11, -29, -30,   1,  -1, -14, -78,   8,  18,  24,  44 },
     {    1,   0,   1,   2,   5,  10,  12,  17,   9,  11,  45,  48,   1,   9,  21, 104 },
     {   -9, -16, -43, -72,  -3,  -7, -36, -78,   9,  17,  14, -20,   3,   8,  19,  15 },
   },
   {
     {   77, -48,   4,  -3, -77,  37,   5,   1,  25,   6, -10,   1,  -6,  -6,   0,   1 },
     {   31,  20, -21,   1, -10, -58,  37,  -2, -37,  79, -18,  -4,  20, -31, -17,  12 },
     {   93,   3,  -9,  -2,  68, -28, -18,   1,  -8, -39,  12,   1,  -7,   3,  12,   0 },
     {    7,  -1,  23,  -7, -17, -21, -23,  10, -15,  15,  64, -12, -20,  56, -79,  -2 },
     {  -10, -32,  10,   7, -33, -51,  34,   2, -33, -19,  47,  11,   6,  33,  73,  -4 },
     {   19,  32,  28,   3,  33,  50,  41,   0,  36,  53,  33,  -1,  33,  42,  29,  -7 },
     {  -10, -12,  -7,   3,  -1, -56,  -3,  19,  98,  -2,  32,  -1,  26, -33,  -8,   0 },
     {  -15,-104,  -1,  10,  52,   1,  18, -11,  -2,  24, -21,   4,  14,  18, -23,  -7 },
     {   -1,   2, -60, -22, -15, -10, -66, -25,   8,  19, -20,  -7,  49,  58,  22,  -1 },
     {   -1, -19,  55, -33,   5,  12, -57,  36, -33,  19,  21,  -9,  52, -47,  19,  -2 },
     {   -6, -11, -11, -30,   7,   5,  16,  31,   5,  -4,  -3, -71, -14,  12,  12,  92 },
     {    0,  -1,   1, -38,  -1,  -8,  21, -12,   2,  -8,  -8, -88,  -7,  -5,   3, -80 },
     {    2,   3, -35,  26,  -8,  27,  33, -12, -25, -50,  32, -17,  80, -16, -37,   8 },
     {   -7, -15, -79,  -8,  11,  35,  -7,  31, -11,  29,  59,   9, -39, -30,  15, -20 },
     {   -3,  -7,  10, -42,   4,   3,  -3,-105,   4,   7,  38,   5, -13, -31,   3,  28 },
     {   -2,   4,  -9, -99,   1,  -1,  39,  27,   1, -22, -13,  54,  14,  13, -20,  -6 },
   }
 },
 {  //7
   {
     {  113, -34,  -5,  -3, -46,  -6,   6,   1,  -7,  12,   3,   0,  -1,  -1,  -2,   0 },
     {  -37,  -3,   6,   0, -95,  47,   7,   1,  46,  27, -16,  -1,  13, -19,  -8,   3 },
     {  -26,  20,   0,   1, -50, -28,  12,   1, -92,  32,  24,  -1,   4,  41,  -6,  -5 },
     {  -26, -88,  38,  -2,  16, -41,  46,  -6,  19,  38,   8, -12,  -2,  10, -17,  -3 },
     {   16,  53, -19,  -1,  24, -35,  12,  -2,  21,  72, -10,  -8,  53, -25, -41,   6 },
     {   -7,  17,  -5,   3, -36, -77,  12,   9,  41, -59,  43,   4,  33,  -1,  18,  -7 },
     {   14,  35,  86, -31,  -5,  -5,  32, -32, -11, -34, -52,   4,  11,   5, -16,  14 },
     {   -6, -42,   4,   5,   9,  25, -13,   1, -42, -21,  -1,   9,  99, -36,  13,   4 },
     {  -13,  -8, -16,   6, -14, -40,  10,   6, -34,  -7, -32,   9, -44, -95,   3,  26 },
     {   -7, -21,  -8,  15, -11, -43, -63,  23,   9,   4, -78,  29,  11,  43,  -2,   9 },
     {   -3,  -6, -70,  -1,   2,  12,  78,  12,  -5, -35, -49, -16,  11,  29, -15,  -2 },
     {    3,  11,   0,  -3,   5,  -6,  25, -13,   6,  37, -22,  -7,   8,   7, 115,  11 },
     {   -8, -12, -33, -87,   1,  -1,   0, -31,   5,   7,  17,  73,   0,   7,  -3,  28 },
     {   -7, -13, -27, -38, -10, -16, -40, -72,  -3,  -8, -14, -77,   1,  -5,  -2, -25 },
     {   -1,   1,  14, -77,   2,   0,  -9,  90,  -4,   3,  -4, -39,   1,  -9,   8, -19 },
     {    0,  -3,  -1,   2,  -1,   3, -10,  10,   3,  -6,  18, -45,   4,  18,  -6, 116 },
   },
   {
     {   87, -51,   5,  -4, -70,  29,   5,   1,  15,   8,  -9,   1,  -2,  -7,   1,   1 },
     {  -46, -37,  37,  -4,   2,  73, -36,   1,  39, -52,  -2,   7, -17,  10,  15,  -5 },
     {   21, -63,  49, -13,  62,  -2, -20,  -1, -66,  26,  -3,   4,   7,  10,  -4,   0 },
     {   49,  13, -11,  -2,  20, -22, -20,  10,  -1, -63,  60,  -9, -26,  60, -15, -12 },
     {   -4,  38,  16, -14, -39,  20, -65,  27, -37,   6,  43,   0,  60, -23,  -8,  -8 },
     {   11,  25,  62, -45,  -4, -52, -34,  28,  47,  24, -19,  10, -40,  -7,  14,  -7 },
     {   50,  18, -24,   6,  59,   9, -20,   2,  20, -34, -20,  12,  24, -63,  48,  -2 },
     {    1,  29,  -4, -10, -28,   5,   3,  -1, -57, -20, -49,  39, -11,  45,  65, -24 },
     {   17,  34, -36,  16,  20,  56, -45,  15,  -8,  51, -22, -11, -59,  13, -25,  10 },
     {   20,  33,  17, -35,  19,  24,  18, -19,  20, -12, -42,  40,  47,  31, -67,   8 },
     {    8,  13,  29,  67, -10, -16, -32, -82,   6,  10,  16,  48, -11,  -5,  -5,  -4 },
     {  -11, -13, -13,  -9, -22, -25, -17,   9, -42, -55, -27,  25, -42, -52, -50,  35 },
     {  -13, -43, -42,  37,  -3, -40, -51,  39,  27,   6, -36,  24,  38,  35,  -9,  -9 },
     {    9,  18,  33,  32,  -8, -10, -16, -10,  -5, -21, -38, -70,  20,  24,  14,  73 },
     {    6,  14,  35,  51,  -1,   1,  17,  26, -11, -19, -33, -38,  -3, -18, -33, -83 },
     {   -7, -13, -33, -52,  -7, -14, -40, -72,   0,  -4, -24, -53,   3,  -1,  -7, -43 },
   },
   {
     {   76, -39,   0,  -3, -81,  31,   8,   2,  29,  14, -14,   1,  -2, -18,   0,   4 },
     {   92,  20,  -9,  -2,  82,   6, -13,  -3,  14, -14,  -6,  -1,   0,  -5,  -2,  -1 },
     {   17,  29, -16,   0,  -4, -54,  25,   0, -33,  83, -11,  -4,  32, -39, -24,  15 },
     {  -31,  43,   6,  -3,  14,  77,   3,  -6,  61,  47, -23,  -6,  24,  -6, -17,  -1 },
     {  -25,  -9, -17,   6,   6, -18, -16,   7,  28, -40, -47,  12, -16, -96,  -7,  19 },
     {   11,  60,  37,   1, -15,  43,  34,   0, -60, -18,   4,  -2, -55, -32,  -6,   2 },
     {   -1, -14,  45,  23,  20, -11,  64,  32,  40,   3,  58,  24,  12, -27,  35,  18 },
     {    1, -13,  30,   1,   1,  -7, -27,  17,  15,  -1,  44,   4, -19,   2,-107,  20 },
     {   10,  85, -17,  -6, -44, -39, -29,   3,  45, -25,  37,   0,   3,   6,  17,   2 },
     {   -4,   0, -26,   8,   7, -32,  53,  -6,  43,  18, -28, -13, -84,  32, -13, -13 },
     {    7,  16,  54,  28, -11, -27,  31,  12,   7, -41, -58, -25,  49,  21, -28, -33 },
     {   -3, -13,  25, -63,   4, -13,  19, -81,  15, -11,  24, -52,   7, -25,  -2,  -8 },
     {   -2,   2, -79,  12,  -4,  25,  60,  -8, -10, -45,  23,  -7,  36,   4, -35,  18 },
     {   -2,  -8, -22,  20,   2,   8, -10,  33,  -2,  11,  35, -44,  -5, -37,  -3, -97 },
     {    2,   3,   2,  27,  -3,  -4,   3, -67,   2,   1,   6,  89,   4,  -6, -16, -53 },
     {    2,  -3,   6,  98,  -1,   2, -23, -51,   0,   8,  12, -46,  -8,  -3,  11,  32 },
   }
 },
 {  //8
   {
     {  100, -41,  -3,  -4, -64,   7,  11,   1,   5,  16,  -2,  -2,  -2,  -5,  -2,   1 },
     {   53,   5, -10,  -1,  67, -65,   2,   0, -59,   0,  26,  -2,  -1,  24,  -4,  -4 },
     {   35, -23,   0,   1,  52,  32, -25,  -1,  43, -79, -10,  10, -27, -10,  34,   0 },
     {   23,  96, -28,  -4, -42, -13, -41,  13,   5, -32,  20,  17,   9,  -3,  14,  -6 },
     {   24,  30, -20,   0,  42, -18,   9,  -2,  47,  43, -39,  -6,  18, -75, -20,  19 },
     {  -16, -18,  12,   0, -17, -78,  26,  11,  73, -13,  48, -10, -14,   2,   6, -15 },
     {  -12, -22,-105,  19,   7,  28,  14,  35,   0,   7,  40, -15,  -9,  -6,  -4, -13 },
     {  -21, -22, -22,  11, -27, -43,  20,   1, -44, -26, -38,  19,  -2, -50,  67,  19 },
     {   -8, -47,   5,  11,   4, -15, -88,  22,   6,  25,  17,  36,  55, -12,   7,  -8 },
     {   -3,   5,   3,   6,  -1,  -8, -45,   9,   0,  54,  -7,   7,-104,   0,  20,   3 },
     {   -3,  -7, -36,  12,  -7, -33, -10,  22,  22, -12, -86,   6,   3,  66, -27,  -5 },
     {  -10, -16,   4,   2, -17, -15, -20,   9, -25, -48,   5,   8, -32, -44, -93,  21 },
     {    4,   4,   6,  98,  -8,  -5, -24, -34,   0,  -7,   2, -68,   7,   1,   4,  15 },
     {  -10, -16, -41, -64,  -9, -15, -36, -83,   5,  -2,   2, -36,   1,   7,   1,  18 },
     {   -1,  -1,  18, -44,  -1,  -1, -19,  74, -11,  -7, -10, -85,   8,  -5,  15,  22 },
     {   -1,  -1,   5,  -6,  -2,  -4,  -9, -10, -14,  -5, -24, -27,  -5, -31,  -5,-116 },
   },
   {
     {  -73,  58, -13,   4,  66, -48,   5,   0, -22,   8,   9,  -4,   6,   2,  -6,   1 },
     {   57,   4, -34,   8,  -4, -51,  52, -11, -34,  60, -30,  -2,  15, -23,   0,  10 },
     {   16, -47,  53, -15,  48,   8, -43,   6, -70,  26,  11,   5,  27,  -4,  -7,  -1 },
     {  -60, -15,   9,  15, -14,  43,  13, -27,  -4,  25, -62,  31,  28, -41,  37,  -3 },
     {   31, -24, -24,  30,  67,   8,  21, -38,  -6, -56, -23,  29, -21,  26,  33, -13 },
     {   43,  52,  -6, -23,  20,  -2, -47,  29,  21, -18, -10,  13,  26, -58,  50, -14 },
     {    2,  18, -54,  41, -22,  42,  -6, -16, -50, -27,  47, -11,  55, -12, -12,   4 },
     {   22,  28,  -5,   1,  51,  63,  -3, -15,  37,  31, -26, -22,   0,  -6, -63,  27 },
     {   10,  58,  -5, -18, -30,  27, -24,   6, -57,  10, -40,  33, -38,  53,  -1,  -9 },
     {  -14, -22, -29, -38,  11,  21,  23,  59, -28, -29, -35, -61,  -5,   2,  24,  42 },
     {   -7, -18, -14,  10, -14, -36, -37,  -2, -16, -47, -38,  27, -25, -51, -63,  33 },
     {   11,  36,  63, -45, -10,   2,  59, -38, -17, -50,   4,   2,  23, -14, -14,  18 },
     {    1, -10, -10, -13,  -4, -26, -12,  14,  27, -16, -42,  16,  88,  60, -24,  -6 },
     {   -7, -14, -25, -35,  12,  21,  42,  44,  -6,  -4,  10,  35,  -4, -25, -42, -79 },
     {    5,   7,  17,  35,   9,  12,  30,  64,   7,   6,  24,  70,   5,   7,   5,  62 },
     {  -10, -20, -51, -79,   4,   4, -12, -40,  10,  18,  34,  42,  -1,   4,  12,  42 },
   },
   {
     {  125, -12,  -9,  -4, -12,  -8,  -4,  -1, -12,  -3,   1,   1,  -8,  -1,   2,   1 },
     {  -17, -58,   8,   5, -97, -35,  21,   9, -20,  17,  20,   1,   7,  19,   5,  -5 },
     {   -1, -70, -21,   8,  48, -62, -16,  11,  58, -14,   0,   6,  30,   0,   2,   0 },
     {  -16,   4, -14,   6,   0, -49, -30,   6, -64, -75, -14,  15, -36, -27,  11,  15 },
     {    8,  -3,  76,  11,  31, -18,  80,   7,  -3, -30,  36,  -7,  -9, -16,  10,  -7 },
     {   -7, -20, -36, -23,  39,   4,   1, -12, -33,  37,  70,  11, -41,  25,  51,   3 },
     {    5,  20,  -9,  67,   0,  14,  -1,  84,   3,  10,  16,  56,   7,   9,  13,  17 },
     {    5,  47,  20,   5,   6, -39, -18, -14, -19, -11,   3, -15,  59,  79,  34,  -8 },
     {   -1, -31,  18,  33,  38,  -9,  -1,  17, -49,  40, -46, -25, -33,  40, -47, -22 },
     {    0, -53,  30,  18,   0,  63, -15, -11,   3, -31, -38,  -8,  -4,  22,  66,  26 },
     {    0,  -9,  28, -44,  11,  -7,  11, -19, -16,  18, -20,  73,  16,  18, -29,  73 },
     {    2,  26,  49,  23, -15, -43, -42, -21,  44,  41,  -4,  -1, -60, -10,  25,  23 },
     {    0,   8, -25,  15,   9, -25,  39,  -6, -28,  50, -46, -20,  35, -53,  52,  24 },
     {   -1,  21, -43, -20, -10, -16,  62,  13,  39, -19, -46,   5, -51,  50,  17,   1 },
     {    1,   1, -35,  76,  -2,   5,  23, -67,   4, -16,  25, -11,  -1,  14, -32,  46 },
     {    0,  -3,   1,  26,   1,   1,   7, -53,  -2,   5, -21,  77,   1,  -8,  15, -78 },
   }
 },
 {  //9
   {
     {  123, -30, -10,  -4,  -4, -11,  -1,   1, -10,  -2,   3,   1,  -5,   0,   1,   0 },
     {    8,  29, -10,   1, 117, -14, -20,  -2,   1, -31,  -3,   2, -10,  -5,   5,   2 },
     {  -23, -33,   4,   5,  -1, -77,  13,   6, -77, -19,  33,   4, -16,  26,  14,  -2 },
     {   12,  77, -22,  -4, -25,  42, -24,  -1, -76,  -5,  -5,   6, -31,   3,   4,   5 },
     {   -9, -20, -16,   6, -33,  -7, -32,  10,  20, -88, -23,  19, -27, -52,  28,  14 },
     {    7,  57, -50,   3, -25, -64, -28,  16,  30,   0,  15,   8,  59,  12,  -3,  -3 },
     {  -18, -54, -68,  10,  10,  13, -66,  13,  -6,  47, -20,  11, -18,  23,   0,   4 },
     {    3,   3,   2,   1,  -3,  33,  -5,   2,  30, -19,  47,   5,  -7,  66,  83, -21 },
     {    2, -15,  38,   1,   0,  15, -30, -10, -26, -39, -61,   8,  66,  55, -10,   3 },
     {    7,  28,  48,  -7, -12, -51, -27, -11,  29,  31, -52,  -5, -62,  27,  18,  10 },
     {   -7,  -9, -24, -80,  -5,  -7, -14, -87,  -1,  -8,   8, -35,   6,  -5,   7,  -1 },
     {    1,  -3,  30,  -1,   9,  -4, -23,  -1, -26,  46,  -3,  14,  46, -61,  77,  11 },
     {    2,  -5,  55, -18,  -5,   7, -79,  20,   4,  -4,  66,  -6,  -3, -10, -43,  -5 },
     {    1,   0,  -3,  47,  -4,  -3, -16,   2,  -8,  -6, -14, -97,   2, -14,  11, -62 },
     {   -4,  -2,  -3, -84,   4,   0,  13,  79,  -5,  -1, -27,  -7,  -1,  -2,   8, -44 },
     {    0,  -1,  -4, -11,   1,   5,   5,  38,   0,  -4,   6, -68,   7,  13,  12,  98 },
   },
   {
     {   89, -59,   6,  -3, -62,  25,  11,  -2,   5,  14, -10,   0,  -1,  -4,   0,   2 },
     {   50,  21, -28,   0,  25, -79,  31,   3, -47,  42,  16, -13,   7,   5, -17,   2 },
     {    2,  72, -57,   9, -71,  -6,  16,  10,  45, -10,   6,  -7,   3, -14,   3,  -1 },
     {   53,  13,  -4,  -7,  25,  -8, -46,  18,   9, -73,  54,   3, -20,  31,  12, -19 },
     {    5, -23,  12,  -4,  30, -46,  40, -11,  84,  -8, -22,  -8, -56,  -5,   7,   7 },
     {   -4,  26,  70, -62, -35, -36, -39,  37,  15,  33,   0,   9,  -1,  -1,  -8,  -8 },
     {   42,  34,  -7, -10,  39,   8, -25,  -3,   1,  -2, -62,  36,  18, -58,  47,   5 },
     {   -1,  22,   9, -20, -22,  -7,  37, -21, -33, -32, -53,  47, -17,  74,  11,  -6 },
     {    4,  16, -28,  14,   6,  30, -41,  22, -15,  51, -20, -10, -91,  21,  -1,   4 },
     {   29,  35,   7, -23,  33,  43,   4, -18,  27,   2, -13,  -7,  19,  17, -90,  29 },
     {  -10, -23, -40, -67,  18,  22,  37,  79,  -9, -17, -16, -25,   6,  -1,   6,  -4 },
     {   -3,  14,  21, -27, -17,   5,  25, -22, -48, -40,  22,  -4, -57, -64, -15,  40 },
     {   15,  37,  31, -21,  18,  48,  47, -28,   5,  35,  39, -29,  -4,  10,  52, -37 },
     {   -9, -17, -32, -26,   7,   8,  10, -11,  10,  23,  36,  88, -15, -23, -29, -52 },
     {   11,  19,  45,  66,   4,  10,  36,  61, -10, -16, -16,   9,  -9, -21, -30, -48 },
     {   -4,  -5, -13, -22,  -8, -13, -24, -43, -13, -19, -39, -55,  -7, -18, -26, -84 },
   },
   {
     {  -94,  53,   0,   4,  60, -11, -19,   1,  -2, -22,   8,   5,   3,   3,   6,  -3 },
     {   40,  26, -17,  -3,  30, -90,  15,  11, -28,  21,  47, -16,  -1,  15, -19, -14 },
     {  -41,   6, -12,  -3, -12,  -4,  57,  -4, -11,  68, -40, -37,   4, -12, -40,  40 },
     {   51,  40, -37,  -1,  44,   8, -21,  -6, -26, -14, -60,  13,  -3, -21,  13,  55 },
     {   -2,  78, -22,  -4, -79,   9, -42,  20,  -5,  23,  12,  18,   8,   0, -13,  -6 },
     {   -9,   1, -18,  30, -13, -30,  35, -32, -11,  24, -25,  52,  31, -31,  63, -43 },
     {    4,   4, -50,   1,   3, -15,  15,  36,  91, -19,   4, -21,   1, -54,   1,  -4 },
     {   31,  46,  22,  -6,  33,  48,  17, -39,  11,  17,  -6, -25,  -6, -23, -30, -71 },
     {   -7, -22, -81,  27,  14,  52,   9,  36, -46,   5,  20, -16,  -7,  19,   5, -31 },
     {   -2, -39,   5,   7,  35,  -2, -61,  24,   7,  63,  -3,  46,   9, -35, -41, -11 },
     {  -20, -29, -36, -26, -26, -37, -42, -45, -25, -35, -37, -31, -17, -23, -30, -36 },
     {   -6,   5,  32,  11, -14,   0,  28,  38, -48, -28,  14,  17, -58, -75, -10,   1 },
     {   -7,  -6, -29, -91,   7,  19,  12, -35,  -1,  22,  49,  29, -21, -20,  22,  19 },
     {   -2,   5, -20,  27,  -1, -10,  12, -22,  35,   9, -16,  46, -92,  44, -19,  -4 },
     {    1,  -3, -19,  57,  -5,  11,   4, -65,   1, -23,  54,  15,  28, -24, -44,  39 },
     {   -1,   0,   9,  37,  -1,  -1, -44, -25,   4,  44,  23, -61, -43, -22,  60,  17 },
   }
 },
 {  //10
   {
     { -114,  33,   5,   4,  46,   2,  -6,  -1,   9,  -9,  -2,   0,   4,  -1,   0,   0 },
     {  -40,  -6,   7,   0,-101,  40,   9,   2,  41,  26, -12,  -1,  14, -12,  -6,   1 },
     {   27,  -2,  -3,  -1,  38,  40, -16,  -1,  95, -36, -23,   0, -14, -35,   8,   3 },
     {   18,  93, -34,   2, -23,  42, -40,   5, -35, -27,  -4,  10,  -4,  -9,  12,   2 },
     {   21,  36, -10,  -1,  32,  -1,   2,  -3,  15,  77, -17,  -6,  71, -18, -40,   2 },
     {   -5, -38,  12,  -1,  27,  81,  -6, -10, -51,  38, -38,  -2, -38, -16,  -9,  11 },
     {   -9, -10, -11,   3, -12, -43,   2,   4, -25, -11, -12,   5,  -6,-113,  -4,  18 },
     {   14,  39,  92, -18,  -4,   2,  48, -28, -11, -34, -31,  -8,   9,  -4, -16,   9 },
     {   -5, -46,  -2,   4,   3,  32, -17,   2, -26, -57,   0,  12,  94,  -2,   4,  -2 },
     {   -3, -10,   7,   0, -13, -44, -65,  -2,   3,  -2, -90,  19,  -2,  31, -10,  21 },
     {   -4,   1, -74,   9,   1,   0,  79,  -9,  -4, -26, -53, -10,  -2,  20, -16,  13 },
     {   -2,  -6,  -6,  -8,  -6,   5, -22,   6,   1, -30,  28, -13, -20,   4,-116,   2 },
     {    6,   6,  25,  96,   2,   2,   7,  68,  -4,  -4, -18, -24,  -2,  -1,  -7, -27 },
     {   -2,  -4, -10,   1,  -6,  -7, -26, -51,  -6,  -5, -11,-103,   4,  -7,  12, -45 },
     {   -3,  -1,   0, -81,   3,   1,   6,  86,  -6,  -3, -22, -28,   2,  -2,   6, -31 },
     {    1,   0,   4,   2,   0,   3,  -7,  25,   3,   0,  18, -59,   8,  10,  14, 107 },
   },
   {
     {   82, -48,   4,  -3, -74,  34,   5,   1,  23,   5, -10,   1,  -6,  -7,   3,   1 },
     {   51,  17, -23,   0,   5, -66,  32,  -1, -48,  66,  -4,  -7,  18, -16, -16,   7 },
     {   17, -69,  40,  -4,  53,  29, -26,  -5, -69,   5,   1,   7,  17,  15,  -4,  -1 },
     {   57,  19, -14,  -2,  20, -29, -15,   9,  -3, -56,  56,  -9, -31,  58, -15,  -9 },
     {   10, -35,  16,  -4,  52, -28,  27,  -9,  47,  17, -43,  -1, -76,   6,  10,  10 },
     {    4, -14, -72,  34,  20,  31,  65, -30, -20, -40, -26,   5,  15,  14,  14,   5 },
     {  -47, -26,   5,   4, -43,   3,  33,  -5, -13,  31,  13, -19, -21,  69, -59,  13 },
     {    6, -25, -22,  13,  39,  19,   0,   2,  46,   6,  43, -38,  22, -48, -66,  23 },
     {  -24, -61,  23,   2, -20, -71,  35,   1,  10, -41,  32,   2,  32, -19,  24, -12 },
     {  -19, -13, -26,  19, -21,   8,  -8,   7, -51,   0,  41, -20, -78, -53,  24,   9 },
     {   -4,  -6, -16, -66,   6,  10,  21,  82, -11, -15, -22, -52,   9,   6,  17,  21 },
     {   -4, -19, -29,  28,  14,  11, -11,  18,  29,  58,  35, -27,  16,  36,  60, -48 },
     {  -12, -36, -57,  29, -11, -37, -68,  37,   2,  -7, -43,  26,   2,   8, -23,   9 },
     {    7,  11,  28,  52,  -8, -13, -23, -21,  -1,  -9, -17, -67,  16,  17,  32,  71 },
     {    6,  13,  33,  57,  -1,   1,  15,  34, -13, -20, -37, -42,  -7, -15, -32, -74 },
     {    7,  12,  27,  53,   9,  15,  34,  72,   5,   9,  23,  60,   0,   4,   6,  43 },
   },
   {
     {   99, -43,  -2,  -4, -62,   3,  15,   1,  -1,  24,  -2,  -4,  -3,  -1,  -8,   1 },
     {  -57, -30,  12,   0, -55,  69,  11,  -4,  42,  20, -38,  -2,   5, -20,  -5,  11 },
     {   40,  -3,   1,   3,  23,  53, -37,  -5,  20, -78, -23,  23, -12, -13,  47,   4 },
     {  -32, -64,  19,   4, -32, -21,   0,   7,   4, -35,  58,  15,   0,  42,  38, -35 },
     {   -5,  77, -11,   2, -78,   9, -45,  10, -26,  -7,  14,  24,  15,  16,   8,  -8 },
     {  -20, -25, -43,   8, -23, -55,  11,   3, -33, -12, -42,  23,  12, -37,  47,  50 },
     {   -7, -16,  21,  -2,   7,  44,  13, -17, -93,  -2, -23, -11, -13,  55,   0,  25 },
     {  -11, -46, -46, -13,   7,   2, -80,  23,  -9,   4,   6,  14, -30,   3, -59,  15 },
     {    0,   1, -89,  12,  16,  47,  35,  35, -13,  20,  21,  -8,   3,   3,  30, -38 },
     {   -4,  16, -19, -16, -24,  -8,  27,  32,  21, -58,  12, -73, -29,  20, -17,  47 },
     {   -9,  -6, -20, -18, -18, -19,  -8, -50, -22, -32, -40, -46,  -9, -24, -16, -79 },
     {   -7,   5,  30,  10, -10,  12,  29,  29, -40, -16,  33,  24, -61, -75, -19, -12 },
     {    3, -16,   2,  -8,   5,  18,   9,   8, -24, -40,  27,  -2,  99, -31, -43,   8 },
     {   -3,   5, -20, -86,  -1,  14,   6, -64,   1,  11,  49,  11, -12, -17,  19,  28 },
     {    0,   8, -35,  44,  -7,  -2,  43, -55,  19, -31,   2,  57, -15,  28, -51,   8 },
     {    0,  -6,  -5,  76,  -3,  10, -33, -51,  -6,  15,  45, -52,  -1, -27,  18,  26 },
   }
 },
 {  //11
   {
     {  123, -28,  -9,  -4,   4, -13,   0,   0, -11,  -3,   3,   1,  -7,   1,   1,   0 },
     {   -4,  16,  -6,   1, 121, -20, -14,  -2,   4, -25,   1,   1, -13,  -3,   5,   1 },
     {   18,  83, -38,   2,  -8,  60, -40,   5, -43,  -1,  -9,   6, -18,  -4,   6,   2 },
     {   19,  10,  -9,  -1,   7,  43, -15,  -1, 101,  10, -29,  -2,  35, -34, -13,   3 },
     {   -7, -43,  -6,   7,  26,  36, -14,   2,  -7,  91, -26,  -7, -42,  33, -15,  -2 },
     {   14,  32,  71, -41,  14,  38,  62, -46,  -8,   9,   3, -19, -12,  -5, -10,   2 },
     {  -10, -34,   1,   2, -12,  18,   0,   2,   6, -29,  -6,   3, -85, -77,  20,  13 },
     {    1,   2,  38, -11,  -7, -24, -26,   2, -20, -21, -95,  34,   2,   3, -40,  28 },
     {    6,  34,   2,  -4, -20, -25,   3,   2,  59, -20,  -4,   1, -70,  72,  10,  -8 },
     {  -11, -18, -38, -85,  -6, -10, -34, -66,   4,   5,  10,  19,   4,   7,  20,  27 },
     {   -4, -49,  21,   9,   6,  72, -15,  -4,  -6, -62,  -6,   7,  18,  50,  34,  -7 },
     {   -7, -15, -60, -12,   4,  23,  51,   2,  -4, -34,   7,  12, -10,  13, -87,   3 },
     {    3,   6,  15,  -9,   8,  13,  25,  44,   8,  21,  31,  97,   2,   4,  18,  45 },
     {   -2,   5, -49,  19,   4,  -6,  70, -21,  -6,   7, -63,   9,  10,   2,  62,   0 },
     {   -3,  -3,  -1, -67,   3,   1,   4,  50,  -4,   1, -19,  17,   1, -11,  11, -92 },
     {    2,   1,   8,  47,  -3,  -4, -12, -70,   3,   5,  11,  67,  -6, -10, -16, -64 },
   },
   {
     {  -85,  45,  -3,   3,  74, -31,  -3,  -1, -22,  -2,   7,  -1,   6,   4,  -3,   0 },
     {  -56, -14,  21,  -1, -12,  62, -27,   3,  60, -57,   3,   3, -27,  10,  11,  -4 },
     {   -5,  74, -46,   6, -50, -39,  30,   3,  58,  -6,   4,  -7, -18,  -1,  -1,   1 },
     {  -54, -14,  12,   1, -33,  16,  18,  -6,  12,  70, -51,   6,  17, -57,  11,   8 },
     {   19, -29,  17,  -7,  49, -36,  17,  -5,  41,  23, -33,   3, -85,   4,  16,   3 },
     {   -4, -27, -61,  51,  20,  31,  63, -51, -10, -26, -22,  11,   4,  10,   3,   5 },
     {   41,  25, -16,   0,  50,  12, -25,   3,  32, -27, -14,   9,  33, -78,  37,  -3 },
     {    4,  45, -19,   3, -19,  39, -24,   4, -67,   5, -14,  11, -68, -12,  44,  -3 },
     {   -4, -21, -36,  29,  13,  21, -17,  10,   9,  34,  67, -56, -30, -40, -31,  21 },
     {   24,  43, -10,  -4,  33,  60, -18,  -4,  31,  58, -23,  -2,  16,  56, -25,  -4 },
     {    7,  16,  39,  64, -12, -20, -42, -76,  10,  15,  24,  38,  -5,  -5,  -5,  -4 },
     {  -12, -32, -37,  28,  -4, -20, -28,  24,  14,  29,  16,  -3,  20,  34,  76, -51 },
     {    8,  23,  52, -15,   9,  28,  70, -19,   3,  15,  58, -19,   5,  -3,  50, -19 },
     {    6,  12,  29,  40,  -5, -10, -11,  -2,  -4, -15, -38, -82,  15,  21,  35,  56 },
     {    7,  13,  32,  61,   3,   6,  20,  45,  -7, -12, -26, -21,  -9, -16, -38, -77 },
     {    4,   4,  15,  49,   6,   7,  21,  70,   6,   8,  19,  63,   3,   9,  10,  61 },
   },
   {
     {   83, -45,   0,  -4, -75,  29,   9,   1,  19,  14, -13,   0,  -3, -11,   0,   3 },
     {   42,  19, -12,  -4,   4, -72,  21,   3, -49,  62,  23, -13,  10,   9, -37,   3 },
     {   77, -12,  -6,  -1,  65, -21, -24,   3,  -4, -62,  17,   9, -14,   7,  24,  -7 },
     {  -27, -40,  24,   1, -49, -22,  -2,   5, -34, -32,  59,   3, -16,  61,  15, -27 },
     {   20,  26,  27, -12,  22,  57, -28,   0,  30,  21,  34, -18,  -2,  55, -62, -16 },
     {  -11, -55,  42, -11,  31, -35,  52, -18,  66,   3,  19, -28,  17, -15, -13,  -2 },
     {  -16, -30,  -2,   1, -14, -28, -50,  21,   0, -34,  -9,   0, -33, -34, -81,  36 },
     {   -6, -69,   9,   3,  44,  41,  16, -15, -65,  19, -41,  -1,  -3,  25, -13,  18 },
     {   -9, -39, -51,  21,   7, -26, -44,  23,  41,  26, -20,  26,  53,  48,   5,  -6 },
     {   -2,  12, -56,  20,  -2,  -8,  54,  -2,  29, -15, -24, -20, -69,  51, -11,  16 },
     {    5,   4,  29,  56,  12,  15,  35,  73,   6,  11,  29,  55,  -5,  -4,   0,  36 },
     {    0,  -2,  47, -11,   7, -27, -40, -16,  21,  54, -23,  33, -72,  10,  34,   1 },
     {  -13, -34, -62, -16,  19,  29,  -9,  10,   2,  39,  66,  -6, -37, -43,  14, -16 },
     {   -1,   3,   1, -37,  -4,   3, -17,   7,   4,   3,  24, -38,  14,  22,  37, 103 },
     {   -2,  -3,  12, -47,   5,  -3,   3,  95,  -5,   2, -36, -46,  -9,   2,  14, -36 },
     {    3,  -3,  15,  90,   0,  -1, -32,  -1,  -4,  12,   0, -79,  -6, -11,  21,  -6 },
   }
 },
 {  //12
   {
     {  110, -28,  -4,  -3, -58,   1,   6,   1,  -5,   9,   0,   0,   0,   1,  -1,   0 },
     {  -45,  -6,   5,   0, -85,  40,   5,   2,  69,  11, -14,  -1,   7, -17,  -2,   1 },
     {  -38,   8,   4,   1, -62, -25,  13,   1, -86,  37,  15,   0,  24,  26,  -8,  -2 },
     {   18, 108, -21,  -1, -25,   7, -27,   3,  -3, -46,   8,   7,  14,  -4,  11,  -1 },
     {  -17, -26,   8,   1, -29,   6,  -3,   2, -25, -68,  15,   4, -88,  17,  31,   0 },
     {   -5,  -8,   8,   1, -15,-102,  16,   4,  48, -34,  29,   0,  23,  20,   1,  -6 },
     {    7,  47,  -4,  -2,   0, -29,   8,  -1,  27,  71,  -9,  -6, -80,  28,  -9,  -1 },
     {   -4,   4,  -7,   3,  -8, -41,  10,   2, -19,   9, -13,   2, -21,-113,  20,  14 },
     {   -8, -19, -97,   3,   2,  10, -12,  23,   8,  19,  70,   3,  -4,  -6,   9, -17 },
     {   -4, -20,  12,   5,  -7, -23,-112,   9,   3,  23, -18,  26,   4,   7,  30,   2 },
     {   -8, -13, -76,   4,  -6, -21,   6,  -1,  -7, -23, -89,   7,   0,  21, -24,  18 },
     {   -2,  -4,   9,  -7,  -6,  -4, -39,   4,  -4, -18,  23,  -8, -20, -21,-113,  10 },
     {   -3,  -1,  -4, -97,   1,   0,   9, -31,   3,   3,  11,  73,   0,   1,  -1,  23 },
     {   -3,  -4, -17, -16,  -5,  -5, -22, -98,   1,   0,  12, -70,   2,   0,  12,  21 },
     {   -2,  -1,   2, -77,   1,  -3,  -9,  68,  -3,  -2, -12, -73,   4,   2,  13,   8 },
     {    2,   2,   1,  26,   2,   5,   5,  22,   5,   4,  22,   2,   4,  13,   8, 120 },
   },
   {
     {   80, -61,  10,  -3, -64,  40,   4,  -2,  14,   3, -12,   3,  -3,  -4,   3,   1 },
     {   52,  13, -36,   5,   5, -65,  50,  -5, -39,  57, -10, -12,  11, -10, -13,  10 },
     {    8, -60,  59, -12,  62,   0, -29,  -3, -60,  24,   2,   6,  11,   7,  -7,   0 },
     {  -57, -16,   6,  12, -16,  29,  31, -26,  -4,  45, -70,  21,  22, -40,  22,  10 },
     {   12, -35, -16,  25,  56,  -9,  48, -41,  32, -38, -29,  15, -46,  32,  14,   0 },
     {   43,  32, -40,  25,  31,  33, -31, -14, -35, -31,  -9,  32,  34, -34,  46, -17 },
     {   30,  25,  37, -50,  24, -35, -21,  29,  53,  13, -43,  17, -21, -36,  33,  -6 },
     {    6, -24, -21,  15,  48,  23,  -1,   9,  50,   6,  18, -54,  31, -58, -30,  34 },
     {   16,  39, -22,  -6,  20,  58, -28,   7,  -9,  39, -40,  -5, -47,  35, -43,  30 },
     {   18,  24,  27,   4,   5,  -5,  -3, -40,  42,   5, -11,  38,  75,  38, -51,  -9 },
     {    4,   9,  23,  64, -13, -21, -45, -54,  18,  39,  34,  13, -46, -30,   1,   0 },
     {   11,  32,  43, -21, -13,  -2,  21, -41, -32, -56,  -7,  -8, -12, -37, -23,  65 },
     {  -14, -38, -37,  21, -17, -50, -52,  29,  -4, -30, -40,  36,   4,  -5, -33,  40 },
     {    8,  13,  27,  44,  -8, -16, -23,   2,   2,  -5, -38, -72,  27,  45,  50,  28 },
     {  -10, -18, -44, -67,  -4,  -8, -33, -61,  11,  19,  22,  -4,  11,  21,  38,  40 },
     {    4,   6,  15,  22,   8,  14,  28,  44,  11,  20,  42,  56,   9,  19,  36,  74 },
   },
   {
     {  108, -46,  -6,  -4, -46,  -2,  11,   1,   1,  16,   1,  -3,  -3,  -1,  -5,  -1 },
     {  -58, -29,  23,   1, -88,  21,  32,  -1,  10,  43,   1, -10,   9,   7, -13,  -4 },
     {  -16, -41,  10,   8,   6, -88,  26,  10, -38, -12,  53,   0,   0,  29,  18, -15 },
     {  -17, -12, -35,   6, -60, -15, -49,  14, -35, -59, -28,  28, -21, -28,  17,  23 },
     {  -15, -78,  21,   9,  30,   1,  17,   3,  67, -38, -18,   0,  -1, -43,  11,  11 },
     {  -14, -31, -62,  -9,  10, -39, -57,  -6,  34,  45, -21,  -6,  35,  23, -30,  -6 },
     {    1,  29,   3,   7,   1, -53,  29,  13,  -5,  24, -10,   2, -14, -74, -68,  24 },
     {  -10, -49,  10, -17,  31,  20,   0, -36, -69,   7, -45, -33, -39,  11, -35,   1 },
     {   -4, -26, -18,  57,  26,  35,   1,  73, -19,  39,  10,  45, -23,  11,  -7,  13 },
     {   12,  12,  47,  48,  -7, -16,  -1,  30,  -6, -34, -60, -10,  54,  38, -31, -14 },
     {   -3,   0, -16,  10,  -8,  16, -17,  25,  11, -30,  29, -33, -30, -17, -41, -95 },
     {   -4,   9, -78,  12,   2,   5,  70,  18,  -7, -14, -27, -53,  15,   1,  26,   5 },
     {   -3, -16, -22, -47,   6,  31,  28,   1, -21, -38,  28,  51,  60,   2, -52,   3 },
     {    0, -15,  11,  20,   9,  16, -29,   0, -46,  25,  17, -29,  69, -67,  34, -15 },
     {    0,   2, -13,  23,   3, -11,  30, -48,  -6,  18, -42,  70,  -1, -15,  24, -67 },
     {   -2,   1,  19, -83,   6, -12,   3,  71,  -6,  20, -41,   7,  -3,  -8,  31, -29 },
   }
 },
 {  //13
   {
     { -119,  25,  10,   3, -33,  15,   3,   0,  11,   7,  -3,  -1,   9,  -1,  -2,   0 },
     {  -25,  32,  -5,   2, 114,  -6, -16,  -2,  28, -22,  -4,   0,  -4,  -9,   3,   1 },
     {  -28, -16,   5,   2,  18, -56,   8,   0, -97, -18,  25,   2, -29,  23,   9,  -1 },
     {  -18, -84,  26,   2,  10, -62,  26,   0,  53,   0,   5,  -5,  27,   5,  -4,  -3 },
     {    3,   8,  11,  -3,  30,  14,  11,  -7, -28,  78,  -1,  -6,  65,  53, -31,  -5 },
     {   -8, -37,  30,  -3,  19,  40,  16,  -8,   7,  52, -17,  -8, -92,  12,  -1,   2 },
     {   14,  34,  72, -14,   4,  16,  79, -17,  -4, -22,  35, -12,  10, -26,   5,  -4 },
     {   -6, -28,  15,  -5,  11,   6,  -7,  -9, -42,   8, -57,   7,  22, -87, -36,  25 },
     {    7,  52, -24,  -1, -10, -74,  21,   7,  18,  65,   5, -11, -27, -40,  -7,   4 },
     {  -11, -36, -50,  10,  15,  38,   9,   6, -10,  28,  77, -10,  14, -49,  30,  -9 },
     {   -4,  -4, -19, -74,  -5,  -7, -19, -88,   2,  -3,   0, -47,   0,   2,   8,  -9 },
     {    3,   4,  46, -12,   0, -10, -38,  -6,  -2,  33,  -7,  40,  17,  -9,  96,  17 },
     {    4,   3,  55, -16,  -6,  -6, -77,  21,   6,   6,  61, -15,  -9, -14, -49,  -1 },
     {   -4,  -3, -20, -63,   2,   3,  19,   0,   9,  -1,  26,  90,  -8,   8, -30,  40 },
     {   -3,  -3,   1, -58,   4,   1,   4,  51,  -7,   2, -22,  12,   1, -12,   6, -98 },
     {   -2,  -3,  -6, -53,   4,   6,  11,  71,  -3,  -5, -11, -60,   7,  11,  20,  64 },
   },
   {
     {  -82,  41,  -2,   3,  77, -31,  -3,  -2, -30,  -1,   8,  -1,   9,   7,  -3,   0 },
     {  -56, -14,  18,   0,  -6,  57, -24,   2,  57, -63,   6,   3, -33,  19,  12,  -5 },
     {  -23,  68, -31,   2, -48, -45,  28,   3,  64,  14,  -8,  -6, -23, -16,   1,   3 },
     {  -56, -30,  16,   1, -20,  34,   9,  -5,   2,  56, -46,   7,  30, -62,  16,   6 },
     {  -29,  30,   1,  -1, -70,  16, -18,   8, -36, -16,  38,  -2,  70,  23, -23,  -6 },
     {   -6, -12, -69,  38,   2,  27,  71, -38, -17, -33, -25,  11,  16,  21,   6,   2 },
     {   37,  35, -14,   1,  35,  12, -24,   2,  33, -43,  -3,  11,  65, -58,  31,  -9 },
     {   -2, -38,  -3,   1,  30,  -6,  16,   1,  67,  21,  19, -25,  51,  22, -67,  13 },
     {    5,   6,  50, -15,  -4, -33,   6, -18,  18,  -5, -63,  47,  35,  60,  20, -20 },
     {  -20, -61,  17,   1, -21, -75,  24,   1,  -5, -54,  25,   0,   9, -37,  17,   2 },
     {   -3,  -4, -19, -66,   5,   8,  31,  78,  -9, -16, -33, -44,  12,  17,  22,  11 },
     {    9,  29,  41, -22,  -2,  10,  20, -27, -18, -41, -33,  -4, -11, -29, -61,  65 },
     {   -8, -23, -54,  23, -10, -27, -69,  33,  -6, -14, -53,  28,  -4,   1, -36,  22 },
     {    4,   4,  19,  57,  -5, -11, -24, -14,   2,   1, -19, -78,  14,  24,  45,  51 },
     {    7,  13,  37,  65,   2,   5,  24,  51,  -9, -15, -27, -22,  -7, -16, -37, -65 },
     {    4,   7,  20,  41,   6,  11,  29,  61,   6,  11,  30,  62,   4,   9,  19,  63 },
   },
   {
     {   79, -83,  48, -22,  12, -12,   6,  -3,   1,   2,  -3,   2,  -1,   2,  -2,   1 },
     {  -23,   5,   4,  -2,  67, -79,  49, -20,  29, -33,  18,  -6,   8,  -5,   1,   1 },
     {   75,  28, -69,  32,  52,  -1, -32,  14,   1,  -7,  -1,   0,  -8,  -5,   4,  -1 },
     {   12,  -4,  -2,  -1, -36,  29, -12,   4,  68, -71,  39, -14,  32, -35,  16,  -4 },
     {   19,  -4, -23,  18, -55, -35,  34,  -5, -68, -30,  47, -12, -34,  -9,  25,  -4 },
     {  -34, -55, -19,  50, -19, -48, -37,  58,   7, -16, -27,  29,   7,  -2,  -8,   5 },
     {   32,   9, -33,  19, -55, -25,  49, -19,  39,  38, -19,  -2,  44,  29, -36,   7 },
     {   22,  40,  11, -39, -21, -16,   5,  -1,  -7, -47, -53,  63, -20, -26, -23,  36 },
     {    3,   5,   1,  -6,   7, -10,   0,   0, -46,  22, -10,   8,  90, -68,  27,  -6 },
     {    1,   9,  11, -10,  11,   4, -22,  10, -44, -43,  33,   1,  56,  65, -54,  14 },
     {   13,  24,  43,  50,  15,  31,  55,  61,   5,  12,  30,  44,   1,   1,  11,  16 },
     {  -18, -42, -21,  49,  21,  49,  25, -58, -17, -21, -11,   8,  -3, -27, -34,  43 },
     {  -11, -22, -38, -37,   1,  -1, -12, -17,  13,  26,  47,  61,   8,  25,  46,  54 },
     {    1,   7,   6, -15,  -4, -20, -19,  29,   7,  36,  41, -40, -18, -49, -54,  64 },
     {   12,  28,  51,  51, -11, -24, -39, -38,   3,  -2, -17, -25,   8,  21,  46,  55 },
     {   -6, -13, -26, -30,   8,  20,  40,  48,  -8, -21, -45, -56,  10,  21,  38,  51 },
   }
 },
 {  //14
   {
     {  119, -41,  -3,  -5,  10, -10,   2,  -1, -17,   3,   2,   0,  -5,   2,   0,   0 },
     {  -11,   0,   2,   0, 121, -32,  -7,  -3,  11, -12,   1,   0, -14,   1,   2,   0 },
     {   28,  86, -64,  12,  11,  35, -33,   7, -28, -20,  13,  -1,  -6,  -7,   7,  -1 },
     {  -22,  -5,   9,  -1,   0, -27,  11,  -2,-114,  21,  13,   1, -26,  25,   0,   0 },
     {   -4, -17,   6,   0,  22,  68, -42,   6,  -3,  77, -50,   8, -19,  10,  -9,   2 },
     {   -3, -13,   4,   0, -16,  12,  -3,   1,  18, -33,   4,   0,-117,  -6,  25,   1 },
     {   18,  53,  70, -50,   7,  25,  31, -30, -14, -26, -37,  18,  -2, -15, -18,  14 },
     {  -14, -46,  18,   0,  11,  46, -31,   6, -36, -43,  17,   3,  28, -71,  37,  -3 },
     {   -3,  -7,  -7,  16, -16, -44, -46,  39, -11, -40, -71,  48,   1,  -4, -28,  17 },
     {    2,  19, -21,   4,  -5, -42,  28,  -5,   0,  54, -26,  -1, -14, -93,  22,   9 },
     {   -8, -11, -33, -83,  -7, -16, -32, -33,   8,  14,  25,  57,   4,  10,  27,  29 },
     {  -13, -33, -55,   5,  10,  34,  50, -25,  -8, -17,   2,  16,  -8, -18, -74,  36 },
     {    7,  13,  23,  30,   9,  18,  38,  62,   8,  18,  40,  76,   1,   5,  18,  31 },
     {    4,  11,  32, -20,  -6, -16, -50,  30,   6,  19,  58, -25, -15, -32, -70,  24 },
     {   -6,  -7, -26, -70,   6,   9,  32,  77,  -3,  -4, -12, -12,  -2,  -4,  -9, -57 },
     {    2,   3,   8,  19,  -3,  -4, -11, -36,   3,   8,  21,  61,  -8, -17, -33, -94 },
   },
   {
     {  -88,  42,  -1,   3,  75, -29,  -4,  -2, -17,  -5,   7,  -1,   0,   7,  -1,   0 },
     {  -43, -68,  37,   1,  19,  77, -29,  -4,  14, -29,  -7,   6, -11,  -1,  12,  -2 },
     {   37, -39,   8,   0,  47,   3,  -4,  -4, -94,  26,   2,   4,  40,  -1,  -6,   0 },
     {  -34, -29, -25,  18,  -5,   3,  57, -21,   9,  61, -61,   3,   1, -43,  13,  12 },
     {    2,  -1, -74,  33,   6,  36,  51, -26, -12, -62,  14,   0,  13,  26, -11,   2 },
     {  -50,  21,   6,  -2, -68,  20, -12,   7, -23,   3,  16,   0,  84, -15, -10,  -3 },
     {   36,  19, -13,   1,  53,  13, -19,   2,  54, -15,  -2,   0,  59, -67,  10,   0 },
     {   12,  36,  21,  -8,  -8,   0,   1, -17, -23, -38, -67,  52,   9,  15,  64, -23 },
     {    5,   5,  25,  78,  -8, -14, -32, -81,   7,  14,  24,  28,  -4,  -5,  -9,   2 },
     {   -2,  30, -27,   8, -16,  30, -23,   9, -48,  -3,  26, -13, -57, -71,  38,   5 },
     {  -15, -62,   3,   7, -15, -77,  14,   9,  -2, -51,  26,   3,   9, -34,  36,  -4 },
     {    8,  13,  33,  36,  -3,  -5,  -1,   1,  -9, -27, -38, -84,  10,   9,  14,  63 },
     {   -2, -12, -40,  36,   3,   1, -33,  21,  15,  39,  11, -32,  17,  46,  71, -39 },
     {   -6, -17, -52,  25,  -8, -20, -69,  36,  -6, -12, -52,  31,  -5,   0, -39,  28 },
     {   -3,  -7, -26, -51,   0,  -1, -17, -35,   9,  15,  26,  25,  10,  22,  42,  87 },
     {    6,  11,  28,  54,   7,  15,  36,  71,   5,  11,  27,  54,   2,   7,  16,  41 },
   },
   {
     { -113,  12,  10,   3, -14,  22,   2,  -1,  49,   7, -10,  -2,   7,  -7,  -3,   1 },
     {   -3,  93, -11,  -5,  34,  20, -24,  -1, -16, -68,   0,   9,  -8, -14,  15,   2 },
     {  -21, -40,   4,   5, 107, -27,  -4,   0,  11,  -3,   1,  -1, -44,   3,   8,   0 },
     {  -48,   7,  -8,   5, -13, -59,   7,  10, -70,  -9,  45,   1,  10,  53,   8, -14 },
     {   -2,  15,  94, -11,   6, -12,  28, -19, -35,  -2, -63,  -2,   0,  13, -20,  10 },
     {  -17,   8, -21,  -6,  19,  71,   0, -12, -71,  64,   4,  -6, -24, -11, -15,   5 },
     {  -18, -50,   4,   7, -44,   5,  -4,   5, -36, -52, -10,  15, -58, -52,  28,  18 },
     {    4,   5,  17, -30,   5,  12,  87, -22,  12, -22,  69, -16, -12, -27, -21, -10 },
     {   -5,  -5,  -9, -94,   0, -11, -10, -17,   5,  12,   1,  80,   2,   7,   5,  19 },
     {  -13, -27, -19,   6,  28,  -9,   6,  11, -30, -22, -12,  13,  85, -59, -39,  11 },
     {   -2,  44, -23,   2, -14, -70,   1,  11,   5,  38, -12,  -6, -45, -50, -49,  15 },
     {    0, -13, -36,  -2,  -5,  24,  23,  37,   4, -38, -33,  19, -28,  49, -75, -24 },
     {   -6, -16, -34, -27,  -8, -10, -24, -85,  -1, -27, -16, -61,   1,  18, -24,  23 },
     {    2, -11,  54,   4,  -5,  10, -76,  -4,   2,  -8,  56,  10,  -8,  -7, -61, -18 },
     {   -4,  -5,  -1, -49,  -2,  -9, -13,  11,  -9,   3, -24, -33,   1, -28,  17,-104 },
     {    1,   5, -11,  56,  -2,  -7,  13, -77,   2,   5,  -8,  62,  -4,  -7,  -2, -53 },
   }
 },
 {  //15
   {
     { -123,  15,   6,   3,  27,   3,  -1,   0,  16,  -3,  -1,   0,   1,  -1,   0,   0 },
     {  -21,  -7,   3,   0,-115,  17,   7,   3,  38,  13,  -4,   0,  24,  -5,  -3,   0 },
     {   23,  -4,  -2,  -1,  36,  17,  -5,   0, 116, -14, -12,  -2,  -2, -18,   1,   0 },
     {    6,   6,  -3,   0,  25,  -5,  -1,  -1,  -4,  26,  -2,   1, 121,  -6, -16,  -2 },
     {   12, 109, -11,  -1,  -7,  53, -22,   0, -11, -18, -15,   3,   0, -16,   1,   3 },
     {    4,  45,   7,   1, -16, -94,   2,   3,  19, -58,  23,   1,  13,  24,  10,  -6 },
     {    0,  27, -62,  -4,   1, -28, -28,   4,  20,  85,  17,   2, -19,  43,   2,  -4 },
     {  -10, -30, -71,  -9,  -9, -19, -69,  -3, -13, -36, -29,  11,   6, -49,   4,  12 },
     {    2,  23,  17,   1,  -1, -50,  34,  -3,  -2,  48, -20,  -8, -19, -93,  -7,  11 },
     {   -2,  -2, -43, -10,   2,  22,  35,  13,   0, -13, 101,   8,   4, -41,  18, -12 },
     {   -3,   3, -67, -13,   3,   1,  85,  17,  -3, -20, -54, -18,   3,  20,  -2,   2 },
     {   -2,   2,   1, -81,  -2,   1,   2, -91,   1,  -1,   7, -37,   1,   4,   6,   2 },
     {   -1,   1,  -8,  -5,  -2,  -2,  -5,   3,   1, -13,  18, -14, -13,   0,-124, -11 },
     {   -1,   1,  14, -76,   2,  -3,   8,  28,   1,   4, -11,  97,  -3,   3, -10,  -2 },
     {    1,   1, -18,  60,  -1,   0,  23, -82,   2,  -3,  -4,  69,  -2,   2, -11, -16 },
     {    1,   1,  -2,   7,   0,   1,   8,  -7,   4,  -2,  18,  11,   2,  11, -10, 125 },
   },
   {
     {   81, -56,   9,  -4, -68,  39,   1,   0,  18,   2, -10,   2,  -5,  -4,   3,   1 },
     {  -51, -15,  29,  -3,   0,  63, -41,   4,  43, -64,  11,   8, -19,  15,  14,  -9 },
     {   -8,  60, -48,   8, -56, -20,  32,   3,  69, -18,  -1,  -8, -24,  -3,   4,   1 },
     {   61,  17,  -7, -10,  27, -29, -24,  18,  -7, -48,  60, -15, -24,  50, -21,  -8 },
     {   17, -36, -23,  27,  61,  -2,  48, -40,  22, -20, -41,  16, -49,  19,  17,   3 },
     {    6,   1,  59, -49,  11, -45, -22,  31,  42,  36, -30,   5, -52,  -6,  12,  -2 },
     {   46,  37, -15,  -4,  36,   8, -31,   6,  12, -25, -18,  25,  36, -59,  57, -20 },
     {   -5,  32,   1,  -5, -41,   4,  -3,  -6, -76, -14, -22,  37, -52,  24,  45, -20 },
     {    4,   9,  33, -20, -13, -28,  20, -18,  16, -27, -36,  45,  70,  59, -13, -18 },
     {    7,  13,  33,  61, -13, -24, -41, -72,  17,  22,  32,  36, -14, -15, -14, -10 },
     {  -23, -56,  12,   6, -21, -61,  30,   7,  -4, -33,  43,  -8,  10, -28,  52, -30 },
     {   -8, -19, -44,  23,   6,   7, -27,  30,  22,  58,   8,  11,   9,  51,  28, -65 },
     {    7,  11,  27,  53,  -7, -14, -25,   0,  -1,  -4, -35, -73,  18,  36,  49,  35 },
     {  -12, -35, -40,  30, -14, -43, -53,  39,  -8, -30, -42,  34,  -3, -12, -35,  30 },
     {   -7, -15, -34, -46,  -3,  -7, -22, -33,  11,  19,  32,  27,  11,  25,  48,  76 },
     {   -8, -14, -33, -50,  -9, -17, -41, -66,  -6, -12, -30, -54,  -2,  -7, -15, -43 },
   },
   {
     {  -78,  66, -24,   9, -49,  49, -19,   6,  -6,   6,  -1,  -1,   2,  -3,   2,  -1 },
     {  -72, -24,  66, -26, -14, -33,  56, -25,  16,  -1,   3,  -3,  10,   4,  -9,   4 },
     {  -39,  32, -24,  13,  82, -33, -13,   7,  49, -47,  11,   0,  10, -13,   5,   0 },
     {   -8,   8, -16,  13,  -7, -37,  24,   2, -74, -38,  63, -17, -40,   4,  29, -12 },
     {  -44, -27,  -3,  28,  41, -22, -26,  29, -41,  57, -35,  12, -30,  42, -14,  -2 },
     {  -14, -55, -21,  52, -59, -34, -15,  52,  32, -26,   2,  12,  20, -12,   7,  -4 },
     {   -9, -22,  25, -12,   0,  28, -18,   4,  25, -20, -19,  16, -96, -39,  38,   4 },
     {    3,  21,  15, -21,  -5, -23, -10,  14, -52, -46, -43,  67,  17, -29, -25,  39 },
     {  -19, -32,  18,   0,  28,  24, -22,   2, -39,  22,   9, -12,  54, -72,  47, -19 },
     {  -19, -33, -57, -62, -19, -34, -51, -57,   0,   0,  -6,  -7,   2,   9,  14,  13 },
     {    4,   7,  38, -31,  -2,   9, -33,  22,   1, -31,  -2,  25,  23,  68,  50, -55 },
     {  -17, -55,   0,  15,  20,  70, -12, -18, -16, -54,  24,  -3,   7,  31, -42,  25 },
     {   -2,  -4, -12, -10,   9,  12,  17,  14,  18,  39,  60,  68,   8,  16,  39,  60 },
     {    3,  11,  40, -41,  -5, -15, -57,  53,   5,  11,  49, -40,  -7,  -9, -40,  28 },
     {   -9, -17, -37, -48,   7,  13,  26,  29,   3,   9,  24,  41, -11, -24, -49, -72 },
     {    6,  12,  35,  46,  -9, -19, -49, -63,   7,  15,  39,  47,  -8, -13, -28, -32 },
   }
 },
 {  //16
   {
     { -117,  46,   2,   5,   9,   1,  -2,   0,  19,  -6,  -1,  -1,   1,  -1,   0,   0 },
     {    2,  -5,   1,   0,-100,  34,   5,   3,  71,  -7,  -8,  -1,   4,  -9,   1,   0 },
     {   20,  -7,   0,   0,  70,  -7,  -7,  -1,  98, -30,  -6,  -2, -18,  -8,   3,   0 },
     {   33,  83, -64,  13,  -9,   1,  -7,   3,  -9, -51,  34,  -6,  -1,  -1,   5,  -2 },
     {    7,   8,  -4,   0,  14,  -2,  -1,  -1,  13,  12,  -5,   1, 123, -21, -14,  -1 },
     {   -6, -13,  12,  -4, -24, -98,  57,  -8,  10, -33,  26,  -5,   9,  22, -10,   0 },
     {   16,  43,  68, -54,  -5,  -1, -12,  -1, -14, -49, -47,  42,   4,   1,   5,   3 },
     {  -21, -58,  15,   7,   2,  33, -19,   3, -22, -83,  48,  -7,  17, -21,  21,  -4 },
     {    0,  -5,   5,   0,   5,  30, -16,   2,   8, -11,  11,  -3,  16, 113, -41,  -6 },
     {  -10, -25, -36,  27, -16, -49, -69,  53,  -2, -15, -43,  34,   3,   8,   3,   0 },
     {   -8, -21, -44,  29,   9,  26,  59, -36, -14, -36, -67,  30,   1,   2, -18,  19 },
     {  -13, -20, -50, -81,   4,   7,  11,  10,   7,  14,  36,  69,  -3,  -4, -10,  -8 },
     {   -1,  -1, -11,   0,   3,   2,  19,  -8,   4,   9, -13,   8,  16,  39, 107, -47 },
     {   -8, -14, -32, -50, -11, -20, -50, -81,  -2,  -7, -23, -49,   1,   3,   3,   7 },
     {   -5,  -6, -21, -51,   6,  10,  31,  74,  -7, -14, -34, -71,   2,   3,   4,  11 },
     {    1,   1,   1,  -1,   0,   0,  -2,   1,   5,   9,  16,  11,   8,  19,  44, 116 },
   },
   {
     {   84, -25,  -4,  -2, -84,  23,   5,   2,  32,  -6,  -3,  -1,  -6,   0,   1,   0 },
     {  -18, -69,  38,  -2,  20,  78, -40,   1,  -7, -39,  14,   1,  -2,  10,   1,  -2 },
     {  -54,  15,   2,   0, -16,   2,   1,   2,  93, -20,  -7,  -3, -62,   8,   6,   1 },
     {   13,  29,  57, -44,  -9, -32, -66,  50,   4,  13,  37, -23,  -2,   1, -10,   3 },
     {   11,  40, -29,   9,  -2,  -2,   3,  -5, -10, -79,  47,  -3,  10,  64, -31,   0 },
     {  -67,  13,   5,   1, -73,  16,   0,   2,   0,   2,   2,   0,  77, -10,  -7,  -1 },
     {   30,  -2,  -3,  -1,  55,  -3,  -4,  -1,  78, -10,  -3,  -3,  77, -16,  -8,   0 },
     {  -11, -29, -47,   0,   4,  15,  20,  23,   5,  30,  57, -52, -13, -31, -55,  31 },
     {   -4,  -5, -22, -65,   8,  15,  40,  71, -10, -24, -46, -30,   8,  19,  25,   2 },
     {  -10, -54,  27,  -5,  -3, -38,  22,  -5,  13,  38, -13,   5,  10,  84, -42,   0 },
     {   11,  59,  -6,  -4,  13,  79,  -8,  -7,   7,  64,  -8,  -5,  -1,  43,  -9,  -7 },
     {   -5,  -8, -25, -51,   2,   5,   9,  13,   5,   9,  32,  80,  -7, -17, -26, -65 },
     {   -6, -25, -47,  28,  -2, -14, -30,  19,   8,  24,  37, -21,  10,  32,  74, -47 },
     {   -3,  -7, -55,  26,  -4,  -8, -73,  36,  -3,  -2, -57,  28,  -3,   4, -40,  19 },
     {   -5,  -9, -30, -56,  -2,  -4, -19, -37,   5,  10,  23,  35,   6,  16,  40,  82 },
     {   -4,  -7, -25, -54,  -5, -10, -34, -72,  -4,  -8, -28, -57,  -3,  -6, -19, -43 },
   },
   {
     {   99, -14, -10,  -3,  71, -15,  -8,  -1, -15,  -5,   1,   1, -25,   2,   3,   0 },
     {   29,  79, -34,  -1, -36,  51, -18,   0, -53, -21,  14,   2,  -6, -18,  10,   1 },
     {   58, -42,   8,  -1, -83, -29,  18,   2, -28,  10,   5,   1,  44,  14,  -8,  -1 },
     {  -36, -26,  17,   2,   0, -26,  14,   1, -91,   1,  13,   1, -67,  13,   8,   0 },
     {    7, -14, -15,   6,  -9,  48, -33,   5,  -3,  95, -37,  -4, -23,  36, -10,  -6 },
     {   17,  23,  84, -28,   3,  35,  49, -24,  -7,   4, -41,   2,  -7, -19, -34,  11 },
     {   19, -31,  -1,   0, -49,  12,  -8,   2,  48, -17,  -6,   1, -80, -53,  24,   5 },
     {   12,  24,   9, -10,  -6,  10,  53, -21,  32,  26,  75, -29, -23,  48,  29, -18 },
     {  -11, -38, -16, -65,  14,  25,  -8, -50, -17,  28,  34,  18,  28, -45,  30,  27 },
     {    1, -25,  22,  60,  20,  40,  23,  45, -17,  21,  29, -22,  26, -58,  16, -27 },
     {    5,   7,  57,  -3,  -1,   3, -38,  11,  -2, -10, -21,  21,  16,  27,  97, -11 },
     {    2,  53,  -8,   2,  -2, -74,  20,   3,   2,  64, -20,   6,  -2, -49,  26,   2 },
     {   -3,  -4,  -1,  -6,  -2, -11, -24, -52, -11, -11, -28, -96,   7, -14,   8, -49 },
     {   -4, -16, -57,   2,   7,  23,  78,   0,  -6, -17, -60,  -1,   7,  15,  46,   5 },
     {    3,   2,  10,  62,  -1,  -1,  -8, -39,   0,   2,   2, -32,   3,   7,  14,  98 },
     {    1,  -2,  -2,  60,  -1,   2,   2, -80,   1,  -1,   0,  63,  -1,  -2,  -5, -48 },
   }
 },
 {  //17
   {
     {  116, -17,  -6,  -3,  16,  -5,   0,  -1, -48,   6,   2,   1,  -6,   2,   0,   0 },
     {   -1,   2,  -1,   0, 115, -11,  -7,  -3,  40,  -9,  -2,  -1, -35,   1,   3,   1 },
     {   49, -12,  -3,  -2, -26,   5,   1,   1, 108,  -5,  -7,  -2,  34,  -8,  -2,  -1 },
     {  -17,-100,  14,   2,   4, -17,   6,   0,   3,  73,  -7,  -2,  -2,  14,  -5,   0 },
     {    9,  -4,   0,   0, -43,   7,   2,   1,  22,  -5,  -1,  -1,-117,   9,   9,   2 },
     {    6,  43,  -1,  -5,  11,  90,  -3,  -6,   5,  76,  -8,  -5,  -2,  16,  -6,  -1 },
     {    7,  12,  97,   9,   1,  -3,  12,   0,  -5, -12, -79,  -7,  -1,   1, -15,   0 },
     {   -6, -43,   1,   3,   6,  61,  -3,  -2, -11, -27,  -6,   1,  -5, -98,   7,   7 },
     {   -3,  -4, -51,   9,  -6,  -7, -87,  13,  -5,   0, -73,  10,  -1,   4, -19,   5 },
     {   -3,  -2,   7, -98,   0,  -2,   1, -20,   1,  -1,  -8,  77,   0,   1,  -3,  17 },
     {    5,  46,  -1,  -3,  -6, -63,   3,   3,   5,  64,  -3,  -3, -10, -77,   0,   7 },
     {   -2,  -1, -38,   2,   3,   1,  61,   3,  -1,  -8, -17,  -1,  -8,  -4,-104,   1 },
     {   -2,  -3,  -9, -45,  -4,  -6, -12, -87,  -3,  -6, -10, -78,   0,  -2,  -4, -16 },
     {   -2,   2, -48,  25,   3,  -2,  64, -34,  -2,   2, -59,  23,   4,   5,  64,   0 },
     {   -3,  -2, -17, -61,   4,   3,  23,  82,  -3,  -3, -25, -47,   1,   0,  25, -40 },
     {   -1,  -1,  -6, -13,   1,   2,   9,  18,  -1,  -2,  -5, -38,   4,  10,   9, 119 },
   },
   {
     {   97, -23,  -5,  -3, -77,  17,   4,   2,   3,   2,  -1,   0,   9,  -3,   0,   0 },
     {   38,  63, -19,  -1,  16, -64,  10,   0, -71,  11,  12,   1,  27,  11,  -8,   0 },
     {   12, -67,  10,   0,  45,  48, -10,  -2, -80,  10,   3,   2,  25, -10,   0,   0 },
     {  -14, -15, -93,   4,   2,  18,  81,  -1,   2,  -5,  -2,  -4,   4,  -3, -17,   2 },
     {   52,  -5, -10,  -2,  54,  -4,   2,  -1,   3, -21,   2,   0, -98,  20,   8,   2 },
     {    4, -27,   2,   0,   8, -39,  12,   3,  14, 105, -22, -10, -21, -42,   3,   6 },
     {   48,  -4,  -3,  -2,  70,  -7,  -3,  -2,  66, -11,  -7,  -1,  67,  -9,  -5,  -2 },
     {    4,   1,   0,  96,  -2,  -1,  -6, -82,   2,   2,   9,  -2,  -1,  -3,  -2,  19 },
     {   -2,  -9, -25,  -4,   0,   0, -37,  12,  14,  17, 107, -13,  -6, -11, -44,  -1 },
     {   -8, -51,  12,   6,  -5, -38,  11,   2,   8,  15,   0,  -5,  12, 104, -23, -10 },
     {   -1,  -4,  -3, -24,  -1,  -4,   2, -39,   5,  10,   8, 109,  -4,  -4, -13, -44 },
     {   -5, -11, -52,  11,  -3,  -8, -38,   8,   3,  11,  15,   4,  11,  18, 102, -26 },
     {    9,  58, -13,  -4,  10,  77, -17,  -6,   6,  60, -15,  -4,   1,  48, -13,  -6 },
     {   -5, -14, -58,   4,  -7, -18, -78,   7,  -5, -13, -60,   5,  -4,  -7, -49,   9 },
     {    4,   6,  10,  55,   2,   3,   7,  41,  -3,  -4, -10, -16,  -6, -15, -21,-102 },
     {   -3,  -5,  -5, -58,  -3,  -7,  -7, -78,  -3,  -6,  -5, -62,  -1,  -6,  -1, -53 },
   },
   {
     {  -74,  83, -51,  23,  18, -20,  10,  -3,   2,  -4,   4,  -3,   1,  -1,   0,   0 },
     {  -14,  16, -13,   8, -71,  80, -48,  20,  24, -26,  13,  -4,  -1,  -2,   4,  -3 },
     {   80,  14, -73,  45, -31,  -6,  30, -17,  -9,   5,   3,  -6,   2,   0,  -1,   2 },
     {  -11,   8,  -2,   1, -19,  23, -18,   8, -74,  81, -45,  18,  15, -14,   7,  -2 },
     {  -37, -23,  12,   7, -62, -11,  60, -38,  50,  20, -44,  18,  -6, -10,   7,   1 },
     {   38,  69,  22, -74, -38, -36,   4,  24,   9,  -7, -12,  18,  15,  -5,   0,  -1 },
     {   -1,   8,   2,  -7, -12,   2,   1,   2, -12,  12,  -9,   5, -76,  85, -48,  17 },
     {   27,  15, -16,   2,  50,  19, -30,  11,  63,  26, -52,  25, -46, -20,  27,  -9 },
     {   -5,   0,   5,  -9, -18, -10,  12,  -4, -45, -25,  20,  -3, -84, -43,  57, -19 },
     {    7,  14,   5,  -8,  24,  43,  16, -56, -31, -57, -28,  67,  16,  18,   5, -10 },
     {   11,  31,  16, -37,  20,  54,  29, -60,  10,  39,  32, -54,  -7, -11,  -4,  10 },
     {   17,  37,  66,  65, -15, -31, -50, -49,   4,   5,   4,  -3,  -3,   0,   4,   5 },
     {    0,  -1,  -3,   0,   3,   2,  -5,   0,  -9, -19, -12,  13, -19, -67, -58,  86 },
     {    8,  19,  32,  27,   9,  20,  35,  37, -16, -38, -69, -68,   3,   4,   2,  -2 },
     {   -8, -21, -40, -40, -11, -29, -56, -54,  -7, -22, -46, -49,   0,  -3, -13, -19 },
     {   -2,  -6, -10, -11,  -1,  -5,  -9,  -5,   0,  -3,  -6,  -9,  14,  40,  80,  87 },
   }
 },
 {  //18
   {
     {   97, -18,  -8,  -3,  76, -14,  -7,  -2,  -4,   1,   0,   0, -24,   5,   2,   1 },
     {   13,  86, -34,  -3,  12,  77, -30,  -3,   1,  -1,   0,  -1,  -2, -24,   9,   1 },
     {  -31,   3,   2,   1,  58,  -9,  -5,  -1,  99, -13, -11,  -2,  41,  -4,  -5,  -1 },
     {   72, -12,  -7,  -2, -70,   8,  -1,   2,  35,  -3,  -4,  -1,  68,  -8,  -5,  -2 },
     {  -15, -29, -82,  24,  -9, -28, -78,  22,  -3,  -3,  -7,   1,  -1,   8,  19,  -6 },
     {   -4, -17,   4,   0,   4,  36, -11,  -4,   9, 101, -29,  -9,   5,  55, -14,  -6 },
     {   14,  -1,   0,   0, -44,   5,   2,   1,  70,  -6,  -5,  -2, -96,   8,   6,   2 },
     {    7,  10,  19,  87,   6,  11,  21,  85,   1,   4,   5,  13,   0,   0,  -4, -17 },
     {    2,   5,  21,  -4,  -3,  -5, -27,   9, -14, -23, -98,  27,  -8, -16, -58,  15 },
     {   10,  73, -15,  -4,  -8, -56,  12,   1,  -1, -13,   1,   1,   7,  82, -20,  -7 },
     {    2,   2,   7,  28,  -1,  -3,  -4, -25,  -7, -15, -22,-102,  -4, -10, -14, -59 },
     {    6,  15,  71, -16,  -4, -11, -54,  14,  -2,  -5, -17,   3,   7,  15,  80, -20 },
     {   -5, -41,   7,   2,   6,  67, -11,  -5,  -6, -69,  11,   7,   5,  70, -13,  -8 },
     {   -2,  -6, -41,  12,   4,   9,  66, -18,  -5, -10, -67,  17,   5,   9,  67, -16 },
     {    4,   7,  17,  74,  -3,  -5, -14, -57,  -1,  -2,  -2, -13,   4,  10,  16,  81 },
     {   -2,  -4, -10, -39,   3,   7,  16,  65,  -3,  -9, -16, -67,   4,  11,  16,  71 },
   },
   {
     {   88, -45,   2,  -3, -73,  33,   1,   2,  14,  -2,  -3,   1,  -1,  -2,   1,   0 },
     {   52,  56, -49,   5, -10, -65,  41,  -3, -31,  29,  -3,  -2,  15,  -4,  -5,   2 },
     {   33, -44,  20,  -3,  54,  12, -19,  -1, -89,  21,   7,   2,  33,  -8,  -3,   0 },
     {   60,   3,  -1,  -6,  71, -13, -13,   3,  29, -30,   9,  -2, -70,  26,   3,   0 },
     {   10,  57,   9, -19, -25,   9, -49,  26, -26, -59,  57,  -9,  26,  21, -15,  -4 },
     {    6,   9,  64, -42, -14, -48, -43,  38,  18,  59,  -6, -10, -12, -20,   8,   0 },
     {   26,   3,  -2,  -1,  44,   2,  -6,   1,  70,  -5,  -4,  -2,  91, -22,  -6,  -1 },
     {   15,  58,   7,  -8,   6,  46, -15, -10, -15, -12, -55,  33, -10, -51,  55,  -9 },
     {   -2,   7, -48,  19,   8,  37, -23,  14,   8,  40,  54, -40, -26, -66,  -7,  17 },
     {    9,  13,  34,  71, -11, -21, -41, -75,   8,  15,  26,  31,  -3,  -4,  -7,  -2 },
     {   10,  46,  45, -18,  10,  57,  42, -25,   3,  37,  -1, -19,  -5,  26, -56,  20 },
     {   -1,  11, -47,  20,   1,  27, -58,  24,   5,  50, -37,  13,   7,  68,  -8, -14 },
     {    5,   7,  22,  52,  -3,  -6,  -7,   8,  -8, -18, -38, -89,   9,  15,  33,  42 },
     {    2,   7,  14,   5,   5,  17,  33,  -1,   8,  30,  52, -16,  10,  35,  79, -57 },
     {    5,   8,  31,  72,   5,   6,  31,  81,  -1,  -5,   0,  32,  -6, -15, -26, -23 },
     {    0,  -1,  -1,  -1,  -2,  -5, -11, -20,  -5, -13, -29, -53,  -8, -22, -40, -99 },
   },
   {
     {   90, -74,  18,  -4,  32, -24,   1,   2, -19,  18,  -7,   2,  -8,   6,   0,  -1 },
     {  -58, -38,  77, -27,   4, -44,  45, -12,  18,  -3, -13,   9,   2,   6, -10,   4 },
     {    8, -18,  34, -20, -88,  58,   5,  -8, -41,  38,  -9,   1,   7,  -5,   0,   0 },
     {  -43,  11,  -7,   9,  45, -11,  -7,   1, -84,  63,  -8,  -4, -25,  18,   2,  -4 },
     {  -37, -53,  -1,  39,  -5,  11, -67,  52,  32,  17, -43,  12,   8,  -3,   3,  -8 },
     {  -13, -33, -34,  55, -57, -55,  30,  22, -30, -20,  41, -15,  -5,   7,   2,  -4 },
     {    0,  -4,   4,  -2, -14,  17,  -3,  -4,  34,   1,   8, -14,-105,  58,   2,  -6 },
     {  -26, -42,  19,   1,  33,  47, -18, -15,  -9, -24,  71, -42,  -2, -31,  36,  -9 },
     {    3,  13,  24, -29,  -7, -13, -35,  37, -51, -69,  -4,  51, -36, -20,  16,   5 },
     {   19,  33,  52,  60,  22,  35,  48,  62,   5,  10,  17,  18,  -2,  -3,   0,  -4 },
     {  -13, -31, -18,  18,  19,  46,  10, -18, -21, -38,   2,  29,   9,  32, -77,  45 },
     {    8,  18,  30, -25,  -4, -12, -40,  32,  -8,  -9,  34, -32,  47,  77, -24, -26 },
     {  -10, -17, -34, -43,   2,   4,  13,  26,  22,  40,  58,  74,  11,  18,  24,  13 },
     {   -6, -16, -20,   6,  10,  28,  36, -16, -14, -35, -38,  23,  21,  45,  50, -70 },
     {    9,  16,  36,  57,  -9, -17, -41, -69,   4,   7,  20,  47,  11,  24,  31,  18 },
     {    0,   1,   3,   5,  -2,  -6, -13, -20,   5,  12,  27,  33, -16, -36, -64, -90 },
   }
 },
 {  //19
   {
     {  114, -28,  -7,  -4,  33,  -4,  -3,  -1, -36,   9,   2,   1,  -8,   1,   1,   0 },
     {   -3,   0,   0,   0,  96, -18,  -6,  -2,  82,  -8,  -6,  -1,  -6,   5,   0,   0 },
     {  -42,  18,   0,   2,  58,  -6,  -4,  -1, -75,  14,   5,   2, -70,   5,   6,   1 },
     {  -23, -90,  42,  -5,  -4, -28,   9,  -1,   1,  61, -29,   3,   0,  21,  -6,   0 },
     {   21,  -3,  -2,   0, -47,   6,   3,   1,  49,  -3,  -4,  -1,-104,  18,   9,   2 },
     {   -5, -32,  15,  -1,  -7, -78,  30,  -2, -12, -81,  25,   1,  -9, -23,   2,   2 },
     {   14,  37,  84, -28,   5,  19,  23,  -7,  -6, -24, -65,  21,  -4, -15, -22,   6 },
     {   10,  48, -15,  -1,  -9, -58,  25,  -1,  -5,  12, -13,   2,  13,  91, -31,  -2 },
     {    5,  12,  40, -10,   8,  25,  78, -19,   9,  26,  76, -17,   4,   9,  20,  -2 },
     {    7,   5,  26,  95,   3,   4,  14,  42,  -3,  -4, -16, -59,  -2,  -4, -12, -29 },
     {   11,  44,  -8,   0, -12, -63,  12,   1,  11,  62, -15,  -4,  -3, -71,  27,   3 },
     {    6,  15,  49, -13,  -6, -17, -56,  20,  -2,  -8,  12,  -7,  11,  34,  86, -28 },
     {   -2,  -1,  -4, -30,  -3,  -4, -16, -77,  -4,  -6, -18, -90,  -2,  -2,  -8, -26 },
     {   -5, -13, -43,  11,   7,  17,  61, -17,  -6, -17, -59,  15,   7,  11,  71, -25 },
     {    5,   5,  20,  72,  -4,  -5, -22, -82,   1,   1,   5,  34,   3,   7,  15,  45 },
     {    1,   2,   4,  12,  -2,  -4,  -8, -29,   4,   7,  17,  52,  -5, -12, -27,-107 },
   },
   {
     {   98,  -7,  -6,  -3, -80,   6,   5,   2,   1,  -1,   0,   0,  11,   0,  -1,   0 },
     {   31,  -7,  -3,  -1,  43,   3,  -2,  -1,-108,   5,   8,   3,  41,  -4,  -3,  -1 },
     {   -9, -95,  13,   6,   2,  81, -11,  -6,   7,   0,   0,   0,   0, -17,   2,   1 },
     {  -12, -13, -93, -17,   1,  11,  80,  15,   1,   1,   0,   0,   4,  -3, -18,  -3 },
     {   55, -10,  -9,  -3,  55,  -4,   2,   0,   0,  10,   0,   1,-100,   4,   6,   2 },
     {    5,  21,  -2,   0,   7,  36,   0,  -5,  -4,-108,   4,  11,  -5,  52,  -3,  -5 },
     {   49,  -8,  -3,  -2,  70, -12,  -2,  -1,  67,  -6,  -3,  -2,  66,  -3,  -5,  -1 },
     {    2,   3, -18,  94,  -1,  -4,  17, -81,   0,   2,  -3,   0,   0,   0,  -2,  18 },
     {    2,   1,  21,   2,   3,   3,  35,   9,  -8,  -6,-106, -22,   4,   1,  51,  10 },
     {    6,  54,   1,  -5,   6,  42,   1,  -3,   3, -24,   2,   2,  -6,-104,  -2,  10 },
     {    1,   2,  -4,  22,   1,   3,  -7,  35,  -2, -10,  22,-106,  -1,   6, -13,  51 },
     {   -3,   1, -57,  -4,  -2,   1, -44,  -3,   2,  -2,  24,   1,   8,  -1, 102,  10 },
     {   -6, -59,   0,   7,  -8, -79,   0,  10,  -7, -62,   0,   7,  -7, -49,   0,   7 },
     {    2,  -3,  55, -22,   2,  -4,  75, -28,   2,  -3,  59, -20,   2,  -1,  49, -11 },
     {    2,   4,  -3,  55,   2,   3,   0,  41,   0,  -2,   5, -25,  -2, -10,  11,-104 },
     {    3,   6,  20,  56,   4,   8,  26,  75,   3,   7,  20,  58,   3,   7,  16,  47 },
   },
   {
     {   71, -76,  40, -16, -38,  41, -20,   6,  -7,   9,  -6,   3,   1,  -1,   1,   0 },
     {  -30,  17,   0,  -1, -42,  55, -36,  14,  61, -64,  31,  -8,  -7,   4,   1,  -3 },
     {   65,  19, -76,  46, -38,   5,  32, -23,  -8, -16,  21, -11,   2,  -2,  -1,   2 },
     {   36, -24,  12,  -8,  49, -42,  20,  -8,  47, -42,  20,  -7, -44,  51, -26,   7 },
     {  -29, -25,   7,  15, -40,  -2,  59, -44,  56,  10, -58,  31,   3,   4,  -1,   3 },
     {   40,  65,  20, -69, -41, -30,  14,  21,  -3, -22, -24,  33,   8,   4,  -2,  -4 },
     {   26,  -9,   2,  -3,  43, -14,   1,  -3,  45, -24,   3,  -1,  87, -54,  20,  -6 },
     {   31,  23, -23,   4,  38,  25, -27,   8,  35,  22, -42,  23, -62, -28,  50, -20 },
     {    4,  -3, -17,   3, -34, -43, -18,  54,  55,  65,   5, -45,  11,  18,  -1,  -7 },
     {   15,  25, -18,   9,  31,  46, -29,   0,  10,  21, -36,  23,  44,  59, -56,  16 },
     {    9,  24,  15, -22,   6,  24,  17, -17,  20,  28,  10, -35, -25, -59, -50,  69 },
     {  -20, -39, -66, -68,  11,  23,  31,  24,   5,  16,  34,  39,  -2,  -7, -13, -10 },
     {   -8, -29, -25,  31, -12, -40, -35,  46,  -9, -36, -34,  41, -13, -37, -38,  40 },
     {    6,  11,  15,  13, -14, -28, -44, -50,  17,  36,  60,  67,  -4, -12, -18, -14 },
     {    8,  18,  43,  51,  11,  26,  58,  68,   4,  12,  30,  36,  -2,  -8, -12, -15 },
     {    0,   0,   1,   1,   0,   0,   3,   8,   3,   8,  22,  27,  12,  34,  71,  93 },
   }
 },
 {  //20
   {
     {  122, -14,  -6,  -3, -11,   4,   0,   1, -32,   2,   1,   0,   0,  -1,   0,   0 },
     {    6,  -5,   1,   0, 120,  -7,  -7,  -2, -17,  12,   0,   1, -36,   2,   3,   0 },
     {   33,  -6,  -3,  -1,  14, -10,   0,  -1, 121,  -8, -10,  -2,  -4,  11,   0,   0 },
     {   -2,   7,   0,   0, -37,   1,   2,   1,   1,  13,  -1,   0,-121,  10,  11,   2 },
     {   13, 107,  -3,  -1,  10,   6,   8,  -1,  -3, -67,   4,   2,  -5,  -5,  -4,   0 },
     {    0,  -5,   4,   0,  -6,-114,  14,   4, -15, -21, -17,   4,   1,  44,  -9,  -4 },
     {    5,   4,  98,  12,   1,  14,  17,   6,  -3,   3, -75,  -7,  -1, -10, -15,  -4 },
     {   -8, -68,   3,   3,   4,  17,   8,   0,  -5,-103,   6,   5, -18, -16, -11,   1 },
     {    3,   3,   9,   0,  -3, -50,  -1,   2,   8,   8,  14,  -3,  -7,-114,  12,  12 },
     {   -2,  -1, -38, -16,   6,   9, 102,  20,   1,  18, -12,  11,  -4, -15, -54,  -9 },
     {   -2,  -1,  13, -96,  -1,   1, -13, -10,   1,   0, -12,  80,   1,  -1,  10,  10 },
     {    3,  -2,  62, -14,   3,  -1,  58, -18,   5,   3,  81, -12,   4,  19,  39,   0 },
     {   -3,  -4, -32,  -7,   3,   4,  36,  10,  -5, -10, -52, -17,   8,  -2, 103,  -6 },
     {   -1,   0,  11, -46,   1,   0, -21, 101,   1,  -4,  16, -40,  -1,   2,  -3, -40 },
     {   -2,  -1,  -9, -61,  -2,  -3,  -7, -69,  -3,  -3, -15, -78,  -3,  -6, -20, -31 },
     {   -1,   0,  -1, -25,   1,   2,   3,  19,  -2,  -2,  -6, -41,   2,  11,  -8, 116 },
   },
   {
     {   86, -43,   0,  -3, -73,  38,   1,   2,  14,  -7,  -2,   0,   2,  -2,   1,   0 },
     {   38,  61, -58,   8, -25, -61,  54,  -6,  -3,  16, -10,  -2,   8,   2,  -5,   3 },
     {   41, -21,   5,  -1,  38,  -5,  -8,  -1, -96,  34,   7,   1,  43, -20,  -4,   1 },
     {   12,  32, -19,   3,   4,  18, -16,   3, -27, -84,  59,  -6,  16,  42, -29,  -1 },
     {  -25, -38, -48,  58,  13,  43,  49, -58,  -8, -14, -14,  18,   7,  -2,  -3,   3 },
     {   55, -14,  -9,   3,  67,  -9,   1,  -6,   7,  -4,   0,   1, -89,  23,   6,   1 },
     {  -34,  19,   0,  -1, -58,  18,   2,   2, -73,   9,   2,   6, -74,   2,  15,  -2 },
     {  -13, -22, -32,  19,  -6,   2,  -5,  15,  15,  43,  68, -61, -17, -27, -39,  32 },
     {   15,  46, -28,   4,  17,  39, -24,   4,  10, -20,   7,  -1,  -4, -87,  51,  -5 },
     {  -15, -23, -39, -60,  18,  30,  49,  66, -11, -18, -31, -31,   4,   6,   6,   3 },
     {  -10, -19, -39, -47,   0,  -1,  -1,   2,  13,  26,  51,  80,  -3,  -7, -13, -46 },
     {    6,  18,  38, -32,   5,  11,  33, -19,  -1, -19, -13,  25, -17, -52, -79,  40 },
     {  -10, -55,  13,   6, -13, -73,  22,   6, -11, -60,  23,  -1,  -9, -45,  27,  -4 },
     {   -5, -16, -47,  34,  -7, -20, -64,  45,  -6, -15, -51,  33,  -6, -10, -40,  22 },
     {    5,   9,  23,  45,   4,   9,  21,  36,   0,   0,  -9, -19,  -8, -19, -43, -96 },
     {   -6, -11, -31, -50,  -7, -15, -40, -67,  -6, -12, -31, -52,  -5, -10, -23, -42 },
   },
   {
     {  -75,  80, -47,  22, -22,  28, -16,   6,   1,  -2,   3,  -2,   2,  -4,   3,  -2 },
     {  -26,  18,  -9,   6,  62, -71,  42, -18,  38, -44,  25,  -9,   8,  -8,   1,   1 },
     {   66,  24, -58,  25,  64,  16, -45,  16,  23,  11, -15,   2,   1,   6,  -1,  -2 },
     {   25, -10,   1,  -2, -40,  33, -18,   8,  56, -63,  36, -15,  44, -40,  19,  -6 },
     {   38,  16, -27,  12, -13,  -7,  18,  -8, -69, -26,  52, -19, -53, -27,  33,  -7 },
     {   18,  44,  26, -53,  25,  51,  30, -59,  16,  32,  21, -31,   6,  10,  12,  -8 },
     {  -10,  -1,   4,  -1,  30,  -8,  -3,   2, -50,  29, -13,   7,  76, -69,  33, -11 },
     {  -46, -26,  37, -15,  55,  30, -50,  22,   6, -13,  15,  -6, -54, -21,  33, -10 },
     {  -16, -37, -17,  38,  -9, -13,  -1,   6,  15,  47,  34, -58,  18,  41,  38, -48 },
     {  -11, -10,   5,  -2,  29,  23, -21,   5, -54, -37,  44, -18,  46,  49, -54,  14 },
     {   11,  21,  41,  47,  14,  29,  52,  60,   9,  19,  38,  44,   6,  11,  15,  19 },
     {   19,  53,  30, -55, -16, -51, -33,  56,  -6,   0,   7,  -3,  10,  28,  19, -28 },
     {  -11, -24, -43, -41,  -5, -11, -18, -19,   7,  19,  46,  59,  11,  25,  45,  50 },
     {   -1,  -5,  -6,   3,   9,  24,  21, -19, -17, -52, -39,  46,  11,  46,  46, -62 },
     {   11,  27,  56,  60,  -9, -24, -54, -60,  -3,  -3,   2,  10,   6,  14,  24,  23 },
     {   -1,   0,  -1,  -3,   4,   8,  17,  23, -10, -24, -49, -57,   8,  24,  54,  74 },
   }
 },
 {  //21
   {
     { -104,  30,   4,   3, -65,  12,   5,   1,   4,  -9,   2,  -1,   7,  -4,   0,   0 },
     {   42,   7,  -6,  -1, -67,  32,   1,   2, -90,  13,  11,   1, -22, -11,   5,   0 },
     {  -10,  80, -29,   1,  55,  64, -27,  -4,  -6,  26,   2,  -4, -20,  -1,   9,  -1 },
     {  -51, -32,  17,   2,  45, -36,   4,   1, -54,  29,  -1,   3, -71,  16,   6,   1 },
     {    7, -24,   1,   0, -33,  24, -17,   5,  34,  80, -39,  -1,  -7,  67,  -9,  -7 },
     {  -23, -17,   6,   1,  32,   3, -11,   1, -56,  34, -10,   2,  98,  -2, -17,  -3 },
     {    6,  16,  67, -27,   9,  25,  78, -29,   8,  26,  44, -14,   9,  15,  12,  -1 },
     {  -11, -55,  18,  -1,   0,  43, -24,   3,  25,  32,   6,  -6,  -7, -83,  43,   1 },
     {    6,  18,  63, -20,   5,  18,   3,   3,  -5, -19, -83,  32, -14, -27, -42,  12 },
     {  -14, -54,   6,   4,  16,  75, -13,  -7,  -8, -65,  17,   4,  -5,  48, -14,  -4 },
     {    4,   7,  18,  68,   6,  11,  26,  84,   5,   8,  20,  52,   4,   5,   7,  13 },
     {    6,  20,  49, -14,  -2, -15, -42,  21, -10, -24, -17,  -7,  14,  31,  85, -40 },
     {    4,   4,  19,  69,   3,   4,  12,   7,  -3,  -6, -22, -91,  -6, -11, -23, -40 },
     {   -5, -12, -49,  21,   7,  14,  68, -29,  -6,  -9, -59,  25,   6,  -1,  52, -26 },
     {    4,   6,  20,  55,  -4,  -5, -18, -52,  -3,  -4, -11,  -9,   7,  14,  31,  92 },
     {    5,   7,  22,  48,  -6, -10, -29, -66,   5,   9,  25,  58,  -3,  -8, -19, -61 },
   },
   {
     {   90, -14,  -8,  -2, -84,  15,   7,   2,  25,  -8,  -2,  -1,  -1,   2,   0,   0 },
     {  -10,  72, -10,  -2, -24, -79,  15,   3,  43,  36, -14,  -2, -24,  -2,   7,   1 },
     {   52,  32, -10,  -3,  16, -36,   6,   1, -89,  19,   5,   2,  50, -10,  -5,  -2 },
     {  -10, -11, -76,  -4,   9,  16,  87,   4,  -5, -10, -47,   1,   2,   3,  11,  -3 },
     {   51, -37,  -4,  -1,  57,  -3,  -2,  -1,  10,  61,  -5,  -2, -61, -37,  13,   5 },
     {  -46, -31,   0,   3, -54,  10,   4,   2, -10,  68,   2,  -5,  35, -65,  -1,   5 },
     {   23, -13,  -3,   1,  45, -15,  -1,  -3,  74,  -3,  -4,  -2,  87, -16, -12,   0 },
     {    2,   1,   1,  75,  -4,  -2,   3, -89,  -1,   2,  -8,  51,   0,  -2,   7, -14 },
     {   -3,   4, -54,   6,   2,  -1,  11,  -8,   8,   2,  89,   3, -15,  -4, -70,   5 },
     {    8,  60,   0,  -2,   9,  38,  -1,  -2,   4, -48,   0,   1, -12, -92,   9,  11 },
     {    7,  61, -12,  -3,  10,  80, -16,  -5,  12,  61,  -8,  -5,  14,  40,  -3,  -5 },
     {   -2,  -1,   0, -57,  -1,   1,  -3,  13,   4,   4,   6,  89,  -1,  -8,   0, -69 },
     {    5,   5,  60,   5,   3,   6,  40,   5,  -3,   5, -44,   2, -11,  -5, -94,  -5 },
     {    3,   6,  57, -26,   5,   9,  75, -32,   5,   7,  57, -23,   5,   6,  39, -14 },
     {    2,   3,   0,  66,   3,   3,   3,  48,   1,  -2,  11, -36,  -2,  -8,   7, -90 },
     {    3,   6,  21,  49,   4,   8,  29,  69,   4,   8,  23,  63,   5,   9,  19,  53 },
   },
   {
     {  105, -54,   0,  -3,   4,  24, -15,   2, -34,  20,   6,  -3,  -1,  -6,   4,   0 },
     {  -16, -33,  19,  -1,  81, -47, -14,   5,  14,  57, -33,  -2, -20,  21,  15,  -8 },
     {  -30, -70,  62, -10, -62,  19, -10,  10,  32,  13, -27,   2,  18,   0,   0,  -4 },
     {   47, -10, -12,   2,  -7, -47,  29,  -4,  66, -49, -24,  12,   3,  49, -31,  -3 },
     {  -11, -17,   6,   7, -36, -62,  55,  -6, -81,  16,  -7,  10,  -4,  15, -22,   0 },
     {   18,  39,  66, -70,  24,  14,  18,  -3, -17, -22, -30,  46,  -7, -14,  -7,   3 },
     {   20,  51, -19,  -8, -49,  17,   3,   2,   9,  58, -45,   2, -25,  43,  40, -26 },
     {  -11, -15,  17,  -5,  -9,   6, -25,  17, -13, -38,  31,  -7,-102,  40,   4,  -2 },
     {    1,   6, -14,   8,  -8, -32, -62,  60, -31, -43, -46,  10,  19, -18,  12, -42 },
     {  -13,   8,   8,  -8,  18,  36, -28,   8, -29,   5,  10,  -8,  48,  87, -51,  -7 },
     {  -13, -24, -48, -76, -18, -30, -37, -28,   1,   4,  29,  40,  16,  14,  28,  22 },
     {  -23, -32, -47,  26,  14,  49,   6, -27, -10,  -7, -50,  67, -25,  -5, -27,  -3 },
     {    9,  23,  -4,   0, -21, -18, -39,  30,  12,  38, -13,   5, -23, -23, -69,  70 },
     {    3,   2,  10,  32,   5,   4,  24,  71,  11,  16,  46,  76,  13,  16,  28,  15 },
     {   -9, -16, -38, -56,   5,  15,  40,  55,  15,  21,  16, -18, -13, -22, -42, -54 },
     {   10,  17,  25,  21, -13, -22, -35, -43,  14,  26,  43,  40,  -7, -18, -37, -72 },
   }
 },
 {  //22
   {
     { -103,  48,  -4,   5, -52,  17,   2,   1,   9, -14,   4,  -1,   7,  -5,  -1,   0 },
     {   42,  61, -50,   7, -10,  73, -36,   3, -29,  19,   9,  -5,  -8,  -7,  12,  -3 },
     {  -17,  42, -25,   6,  89,  -6, -18,  -1,  71, -14,  -3,  -3,   8,   7,  -1,   0 },
     {  -47, -10,  14,  -1,  44,   6, -15,   1, -41,  77, -25,   2, -50,  31,   5,  -4 },
     {   -8,  28, -13,   2,  33, -37,  25,  -6, -61, -52,  49,  -2, -53, -22,  12,   5 },
     {   11,  35,  61, -46,  17,  42,  62, -45,   8,  19,  24, -10,  11,  -1,   0,   5 },
     {   25,  19,  -7,  -1, -42, -11,  20,  -4,  52,  -4,  -7,   1, -86,  54,  -4,   2 },
     {   12,  46, -14,  -3,   2, -32,  21,  -1, -41,  -5, -35,  19,  40,  56, -66,   8 },
     {   -5, -12, -37,  27,  -7, -21,  23, -11,   3,  36,  67, -53,  33,  49,  19, -21 },
     {   16,  48,  -9,  -9, -19, -69,  14,   3,  10,  57, -26,  -5,   6, -55,  35,   0 },
     {    9,  17,  34,  66,   8,  15,  41,  78,   7,  15,  20,  40,   5,   1,   8,   5 },
     {   -7, -24, -41,  30,   2,  13,  32, -25,  14,  31,  11,  -2, -17, -47, -66,  56 },
     {   -7, -11, -29, -47,  -4,  -8,  -9,  -2,   6,  16,  41,  83,  14,  22,  31,  49 },
     {    7,  19,  46, -23,  -9, -23, -64,  35,   7,  16,  54, -30,  -6,  -3, -49,  30 },
     {    6,  12,  32,  56,  -3,  -7, -24, -45,  -8, -13, -18, -16,  10,  22,  42,  78 },
     {    6,  10,  24,  46,  -7, -14, -33, -64,   6,  13,  29,  57,  -3, -11, -22, -57 },
   },
   {
     {   98,  -5, -10,  -3, -80,   7,   8,   2,  10,  -7,  -1,   0,   7,   2,  -1,   0 },
     {  -11,  87, -11,  -3, -21, -82,  16,   4,  27,  14, -12,  -2,  -4,  11,   3,  -1 },
     {   42,  25,  -9,  -2,  39, -23,   2,   0, -99,   7,  10,   2,  43,  -5,  -7,  -1 },
     {  -19,  -7, -83,   7,   1,  19,  84, -11,  -5, -24, -22,  11,   8,  11,  -9,  -5 },
     {   44, -26, -24,   0,  49, -20,  18,  -2,  19,  73, -12,  -2, -56, -32,  14,   4 },
     {  -42, -29,  -5,   2, -52, -10,   8,   2, -10,  72,   1,  -5,  59, -45,  -2,   2 },
     {   25, -25,  16,   3,  44, -19,   2,   0,  50,  -3, -50,  -5,  79,  32,  14,  -5 },
     {   10,  -4, -36,  -7,  23, -12, -15,   0,  43,   2,  89,   3,  26,   3, -60,   3 },
     {    4,   4,   7,  82,  -3,  -7, -10, -90,   2,   4,   5,  37,   1,  -1,   0,   0 },
     {   12,  46,   6,  -1,  20,  33,   6,  -3,  27, -32, -13,  -2,  21,-100,  -1,  14 },
     {   -4, -61,  18,   4,  -6, -80,  20,   6,  -9, -59,   6,   6, -13, -40,  -6,   8 },
     {   -1,  -1,   5, -54,   0,  -1,  -5,  -6,   4,   4,  -1, 101,   2,  -9,   9, -54 },
     {   -4,  -5, -55,  -5,  -1,  -9, -38,  -4,   4, -18,  32,  -4,  13,  -6, 100,  13 },
     {    5,   8,  59, -11,   7,  11,  79, -15,   7,   9,  60, -11,   8,  10,  45,  -5 },
     {    2,   2,  -5,  57,   2,   0,  -2,  43,   2,  -4,  11, -24,  -1, -13,   9,-101 },
     {    2,   5,   8,  58,   3,   8,  11,  78,   3,   7,   9,  62,   4,   8,  10,  51 },
   },
   {
     {  -73,  70, -34,  14,  41, -48,  25,  -9,   6,  -4,   1,  -2,  -3,   4,  -2,   1 },
     {   82,  19, -55,  22,  23, -21,  30, -16, -58,  11,  13,  -3,   1,  -5,  -5,   4 },
     {  -10, -35,  39, -16,  53, -36,   1,   2, -32,  75, -47,  11,   2,  -7,  12,  -3 },
     {  -46,  16,   1,   2, -40,  34, -12,   5, -87,  13,  17,  -4,  27, -47,   9,   1 },
     {    5,  11,   9, -13,  78,  32, -58,  23, -17, -22,  41, -14, -44, -10,  14,  -2 },
     {   28,  60,  40, -68, -10, -15,   7,  19, -22, -39, -31,  44,  -1,   2,  -7,  -9 },
     {    2, -18,  12,  -6,  43, -10,  -8,   3,   3, -34,  19,  -7, 106,  -2, -33,   6 },
     {   23,  36, -23,   4,  11,  28, -15,   1,  51,  23, -31,  10,  34, -69,  52, -13 },
     {   15,  21,  -9, -16, -34, -52, -41,  64,   7,  36,  26, -48,  18,  26,  18, -13 },
     {    1, -34,  23,  -8,  -8, -62,  31,   2,   7, -38,  36, -14, -21, -74,  29,   5 },
     {   15,  29,  57,  67,   3,   6,  11,  17,  -8, -21, -45, -64,  -6,  -9, -17, -12 },
     {  -10, -30, -21,  31,   2, -16, -16,   9, -26, -49, -27,  26,  11,  32,  62, -63 },
     {    5,  19,  41, -19,   9,  29,  55, -34,   3,  19,  47, -25,  17,  36,  55, -30 },
     {   12,  23,  34,  28, -16, -34, -63, -72,  -4,  -2,   9,  16,   7,  14,  28,  44 },
     {   -7, -13, -26, -38,   5,  12,  12,   3, -13, -28, -46, -52,   6,  19,  47,  74 },
     {    3,   5,  21,  48,   5,   9,  28,  66,   3,   7,  23,  57,   8,  15,  29,  58 },
   }
 },
 {  //23
   {
     {  119, -28,  -6,  -3,  25,   5,  -4,   0, -22,  12,   0,   0,  -6,   1,   1,   0 },
     {   22,  10,  -3,   0,-114,  24,  10,   2, -35, -22,   8,   0,  16, -12,  -2,   0 },
     {  -28,  11,   2,   1,  25,  26,  -7,   0,-105,  32,  15,   1, -33, -25,  11,   0 },
     {  -20, -95,  24,   1, -20, -68,   6,   5, -30,   5, -14,   4,   2,   7,  -8,  -1 },
     {   -3,  -3,   1,   0,  36,   2,  -1,  -1, -28, -45,  14,   1, 102, -34, -22,   1 },
     {  -11, -40,   6,   0, -13,  58, -29,   0,  23,  83,  -9, -12,  48,  -4,  18,  -4 },
     {   -7, -12, -93,   9,  -7, -21, -79,   2,  -5, -17,  -2, -10,  -2,   0,  17,  -5 },
     {   -2,  26, -10,   0,   1,  -3,   1,  -2, -34,  17, -20,   1,  31, 108, -30, -13 },
     {   15,  60,   0,  -3, -13, -79,   3,   5,  -4,  60, -17,  -8,  29, -33,  16,   3 },
     {    2,  11,  22,   6,   0,  17, -36,  22,  -5, -12,-101,  24, -12, -27, -43,  -3 },
     {   -7, -15, -77,   9,   2,   9,  70, -10,   5,  30, -13,  25,  -3, -22, -53,  11 },
     {   -3,  -2,  -2, -96,  -3,  -1, -12, -78,  -2,  -2, -15,  10,  -1,  -2,  -2,  24 },
     {   -3,  -7, -21,  -9,   9,  11,  50,   4, -11, -23, -59,  -4,  14,   6,  89, -27 },
     {   -1,   2,  -2, -17,  -1,  -1, -10,  43,   3,   5,  18, 106,  10,  13,  29,  38 },
     {    4,   3,  13,  80,  -3,  -3,  -9, -87,  -2,  -1,  -9,  39,   4,   2,  20,  13 },
     {    1,   2,   2, -14,  -2,  -5,  -6, -14,   6,   7,  20,  44,  -3, -12, -11,-115 },
   },
   {
     {   89, -26,  -6,  -3, -79,  29,   5,   2,  20, -15,  -1,   0,  -1,   3,   0,   0 },
     {  -31, -78,  41,   1,  17,  71, -42,   0,  13, -11,  14,  -3, -11,  -6,   2,   1 },
     {   52, -17,  -2,  -1,  41,  12,  -8,  -1, -96,   8,  16,   2,  42, -15,  -8,   0 },
     {   25, -13,  33, -19,   1, -38, -32,  25,  16,  88,  -6, -20, -18, -43,  22,   8 },
     {   -5, -46, -50,  41,  13,  16,  58, -43,   9,  44, -35,  13,  -5, -29,  11,   3 },
     {  -57,  -2,  10,   1, -74,   3,   3,   4, -22,  19,   1,  -5,  76, -30,  -5,   2 },
     {  -19,  33,  -4,   1, -35,  29,   4,  -1, -55,  -9,  12,   5, -80, -51,  23,   3 },
     {    1,  -1,  41, -22,  -1,   1,   6,   5, -21, -12, -86,  39,   6,  19,  55, -36 },
     {   12,  21,  41,  66, -12, -21, -44, -75,   3,   5,  15,  37,   2,   2,   2,  -6 },
     {   19,  38,  -2, -10,  33,  25,  -6,  -2,  46, -37,  -7,  11,  37, -86,  16,   4 },
     {   -3, -60,  20,   2,  -5, -79,  22,   5,  -8, -59,   5,   9, -14, -41, -10,   8 },
     {   -6, -11, -23, -47,  -3,  -3,  -2,   1,  10,  23,  43,  86,  -4,  -7, -19, -56 },
     {   -5, -17, -55,  14,  -3, -19, -35,   7,   1, -15,  34, -15,  15,  11,  94, -20 },
     {   -5,  -5, -53,  31,  -7,  -7, -71,  41,  -7,  -6, -54,  29,  -8,  -9, -39,  14 },
     {    4,   5,  13,  54,   4,   4,  12,  36,   3,  -1,   1, -38,  -4, -15, -22, -98 },
     {   -4,  -7, -26, -56,  -6, -11, -35, -72,  -6,  -9, -29, -53,  -7, -11, -24, -38 },
   },
   {
     {   87, -66,  15,  -5,  52, -32,   0,   1,  -2,  13, -11,   3,  -5,   8,  -3,  -1 },
     {   50,  33, -52,  10,   4,  69, -53,   7, -18,  40,  -6, -10,  -4,   4,  13,  -8 },
     {   13, -43,  40, -12, -80,  26,  17,   0, -65,  29,   4,   2,  -7,  -9,   9,  -1 },
     {  -50, -20,   1,  19,   8, -21, -34,  32, -14,  51, -67,  25, -18,  41, -23,  -4 },
     {   -3,  24,  51, -59,  23,  41,  22, -48,  31,  45, -27,  -2,  20,  20, -12,   3 },
     {  -19,  29,  -4,  -3,  50, -20,  32, -23, -64,  29,  36, -17, -59,  13,  21,  -2 },
     {   40,  29, -23,   0, -43, -10,  32, -11,  10, -21,  -6,  13, -41,  68, -53,  21 },
     {   -9, -26,  -7,  20, -25, -11,   7,  -4,  49,  33,  30, -58,   3,  55,  40, -44 },
     {    5,  16, -28,  19,   3, -30,  28, -11, -43,  26,   4, -14,  96,  13, -33,  -1 },
     {   27,  42, -16,  -4, -31, -53,  31,   4,  24,  42, -45,  14, -11, -49,  43,  -7 },
     {  -10, -18, -37, -49, -17, -30, -43, -63, -19, -25, -34, -46, -11, -14, -14, -21 },
     {   15,  35,  25, -32,   4,  -9,  -8,  31, -31, -47, -18,  21,  19,  41,  37, -69 },
     {  -12, -35, -38,  37,  13,  42,  53, -47,  -3, -22, -39,  39,  -1,   2,  26, -30 },
     {   11,  19,  36,  46,   9,  17,  22,  17,  -3, -12, -33, -62, -21, -34, -50, -47 },
     {   11,  21,  45,  57,  -4, -10, -29, -42, -13, -23, -32, -21,  10,  23,  46,  55 },
     {    8,  14,  27,  36, -11, -21, -41, -55,   9,  19,  40,  56,  -5, -12, -28, -51 },
   }
 },
 {  //24
   {
     {  119, -32,  -9,  -3,  10,  17,  -5,   0, -22,  10,   2,   0,  -3,   0,   1,   0 },
     {   -8, -15,   3,   0, 108, -33, -17,  -1,  24,  39, -17,  -2, -18,  19,   6,  -2 },
     {  -25,  18,   3,   0,  29,  49, -14,  -3, -86,  37,  30,  -4, -11, -42,  16,   5 },
     {   27, 102, -24,  -4,  37,  17,  22,  -8,  21, -39,  17,   5,  -8,  -1,  -7,   5 },
     {   -3, -36,  15,   1,  39, -38,  16,  -1, -24, -75,  26,  10,  34, -52, -26,  12 },
     {  -12, -37,   2,   4,  17,  87, -26, -10,  65, -16,  37,  -7,  14,  -9,  -1,   9 },
     {   14,   5,  18,  -8, -16, -33,  22,   0,  45,  30,  33, -13, -29, -74,  57,  16 },
     {    7,   9, 101, -15,   9,  32,  52,   9,   4,  11, -31,  21,   2,   1, -18,  -2 },
     {    8,  26,  -6,  -3,   2,  -4, -18,   7,  12,  39, -26,  -8, 109, -29,   4,  -5 },
     {   -6, -15, -26,   5,   1,  -7,  76, -26,   0,  48,  60,   6,  25,  32, -31,  20 },
     {  -12, -25, -51,  10,  14,  30,  62, -16,  -1, -14, -59,  26,   0, -23,  46, -27 },
     {   -1,   0, -28,  12,  -8,   6,  -6,  -1,  16,  28, -40,  13, -32, -56, -85,  34 },
     {   -3,  -4,   1, -88,  -3,  -3,  -8, -87,  -5,  -5, -17, -24,  -1,  -2,  -4,   8 },
     {   -4,  -2, -15, -58,  -4,  -3, -20,  28,   5,   7,  17, 104,   3,   4,  11,  13 },
     {   -7, -10, -25, -69,   7,  11,  27,  81,   0,  -3,   0, -55,  -4,  -2, -14,  -8 },
     {   -2,  -2,  -1,   6,   3,   7,   6,  12, -10, -12, -28, -12,  10,  22,  34, 115 },
   },
   {
     {  -92,  -6,  14,   3,  82,   0, -15,  -3, -25,   9,   6,   0,   3,  -6,  -1,   1 },
     {   25, -73,  -2,   6,  14,  83,  -6,  -9, -32, -39,  16,   6,  15,   5, -12,  -1 },
     {   44,  42,  -5,  -6,  24, -28,   2,   4, -92,  -2,   6,   2,  52,  -2,  -7,  -3 },
     {   33, -10,  58,  12,   3,  -9, -77, -11,   4,  44,  47,  -3, -14, -31,  -9,  10 },
     {  -40,  28,  47,  12, -50,  -1, -37,  -9,   1, -63,  11,   4,  40,  43, -10,  -4 },
     {   49,  12,  19,   2,  70, -13, -19,  -5,  31, -32, -27,   4, -26,  66,  14,  -7 },
     {    2,  37, -32,  19,  13,   0,  16, -38,  12, -26,  56,  47, -36, -12, -63, -15 },
     {   -3,  41,  27,   5, -12,  32,  -1,   1, -43, -37, -25,  -7, -76, -39,  37,   9 },
     {    1, -11,   4,  74,  -7,  -7,  10, -79,  -7,  22, -39,  31,  18,  -4,  33,  -5 },
     {  -14,   8, -17,  -5, -27,  26,  -9,  -2, -41,  60,  22,  10, -38,  81,  11, -18 },
     {   -5, -70,  14,   2, -17, -80,  16,   5, -37, -28,   8,  11, -42,  10,  -8,  -5 },
     {   -1,   3, -48,  41,   3, -13,  -9, -16,  -1, -19,  41, -72,  -6,  15,  22,  62 },
     {    2,   4,  40,  31,  -1,  12,  33,   1, -10,  23, -40, -48, -11,  16, -85,  21 },
     {    7,   7,  63,  -4,  10,   8,  81,  -5,  11,   7,  60,  -1,  11,  10,  37,  11 },
     {    3,   2,  -3,  61,   4,   1,   3,  41,   4,  -5,  18, -40,   0, -10,  11, -94 },
     {    1,   3,  -2,  62,   2,   5,  -1,  80,   2,   4,   0,  61,   3,   6,   3,  48 },
   },
   {
     {   85, -70,  23,  -9,  46, -37,   7,  -1,   3,   2,  -7,   3,  -2,   5,  -4,   1 },
     {  -54, -17,  51, -18,  -3, -56,  63, -21,  12, -40,  26,  -1,   3, -12,   0,   6 },
     {  -14,  38, -47,  22,  82, -37, -14,   4,  54, -29,   2,  -4,   6,   4,  -4,   0 },
     {   47,   4, -12,  -1, -34,  29,   0,  -2,  38, -58,  43, -16,  44, -51,  25,  -4 },
     {  -28, -42,   7,  16, -31, -16, -40,  42,  31, -23, -57,  37,  39, -15, -32,  10 },
     {  -22,  -1,  52, -45,  46,  52, -14, -24,  39,  37, -15,  -8,  45, -12,  10,  -7 },
     {  -35, -35,   0,  22,  35,  -6, -35,  28, -53,   5,  25,  -6,  -2, -60,  54, -14 },
     {   20,  39, -22,   0,   4, -33,  40, -20, -40,  32, -38,  22,  60, -50,  -4,   9 },
     {    0,  -3,  -7,  12, -37, -43,   9,  12,  48,  49, -16, -25,   8,  18,  66, -49 },
     {  -17, -34, -19,  34,   3,   4,   7,  -3, -23,  14,  41, -42,  74,  48, -36, -12 },
     {  -11, -19, -35, -49, -15, -26, -45, -64, -16, -23, -31, -51,  -9, -13, -13, -22 },
     {   14,  43,  30, -43,   5, -22, -29,  34, -41, -41,   1,  19,  34,  47,  20, -40 },
     {   12,  33,  30, -18, -16, -46, -43,  30,  12,  43,  44, -35,  -4, -28, -43,  31 },
     {  -10, -21, -41, -43, -10, -16, -21, -21,   7,  20,  40,  55,  18,  27,  38,  59 },
     {  -10, -19, -49, -70,   9,  15,  46,  72,   3,   6,  -3, -21,  -6, -11, -15,  -9 },
     {   -2,  -7, -13,  -7,  -3,  -2,  -5, -19,  13,  21,  42,  64, -11, -22, -44, -82 },
   }
 },
 {  //25
   {
     { -113,  45,   4,   3, -21,  -9,   9,   0,  24, -15,   0,   0,   5,  -1,  -1,   0 },
     {   17,  11,  -6,   0,-105,  47,  10,   0, -34, -21,  19,  -1,  15, -19,  -1,   2 },
     {   17,  90, -53,   5,  45,  34,  -6,  -8, -27,  -7,  28,  -4, -13, -11,   5,   2 },
     {  -42, -22,  21,  -1,  13,  15, -14,   3, -91,  61,   5,  -2, -22, -16,  18,  -2 },
     {   -7,  -7, -10,   3,  -7,  74, -57,   4,  58,  52, -14, -17,   7,  12,  24,  -8 },
     {   18,  40, -15,  -2, -43, -38,  20,   3,   1,  52, -37,   3, -53,  58,  13, -10 },
     {   -7,  -9, -70,  29, -11, -44, -46,  18, -28,   7, -28,  -6,  64,   6, -13, -12 },
     {   13,  42,  44, -24,  13,  18,  32,  -5,  -8,  37, -46,  21,  74,  -2, -30,  -3 },
     {  -17, -22,  -7,   5,  15,  48,   2,  -9, -28, -36,  -9,  17,  -6,  90, -46, -10 },
     {   -7, -23, -32,  16,   0,  -3,  57, -46,  14,  50,  62, -25,  32,  21,  -7,  20 },
     {   -9, -23, -45,  33,   3,  26,  37, -12,  12,  12, -54,  34, -36, -54, -40,  15 },
     {    8,  10,  31,  90,   8,  10,  30,  64,   2,   3,   9, -32,   1,   2,  -7, -30 },
     {   -6, -17, -23,  -5,  13,  17,  49, -11,  -9, -27, -35,   0,  20,   5,  90, -44 },
     {   -2,  -4, -10,   5,   1,   2,   6,  56,   8,  13,  35,  96,  13,  15,  30,  32 },
     {   -9, -14, -35, -74,   5,   9,  31,  74,   5,   9,  11, -31,  -7,  -8, -29, -27 },
     {    1,   0,   2,   1,  -5,  -9, -12, -28,  11,  14,  35,  47,  -6, -16, -23,-103 },
   },
   {
     {   94,  -5, -14,  -3, -80,  13,  14,   2,  18, -16,  -4,   1,   0,   6,  -1,  -1 },
     {   11,  87, -25,  -4,  -1, -76,  31,   2, -26,   9, -15,   3,  18,   6,  -1,  -3 },
     {   52, -14,   0,  -1,  41,  23, -11,  -1, -95,  -4,  20,   2,  38, -13, -10,   0 },
     {   28,   7,  54, -13,  -8, -38, -66,  20,  10,  63,  24, -22,  -8, -27,  11,  11 },
     {   14, -41, -50,  14,  17,  -5,  47, -13,   5,  81, -23,  -1, -14, -41,  14,   4 },
     {   54,  16,  -1,  -4,  73,   0,  -6,  -1,  22, -19, -11,   4, -77,  28,  11,  -2 },
     {    1,  -4,  50,  -2,  -4,   9,   4,   8, -20,  -5, -96,  -2,  13,  -1,  61, -13 },
     {  -15,  39,   5,  -2, -23,  28,   9,   1, -26, -26,   7,   5, -59, -91,   5,   9 },
     {    7,   9,  24,  75,  -7, -10, -20, -87,   1,   5,   3,  43,  -1,  -1,   2,  -7 },
     {   23,  -7,   5,  -1,  41, -21,  -1,   0,  58, -43, -17,   3,  56, -66, -22,   7 },
     {   -6, -69,   9,   6, -16, -82,   6,  10, -32, -39,  -5,   6, -38,  -6, -16,  -2 },
     {   -1,  -5,   2, -54,  -1,  -1,  -6,   7,   5,  16,   7,  93,   1,  -8,  -2, -66 },
     {   -4,  -9, -50, -14,  -1, -17, -32, -13,   2, -27,  31,  11,  13,  -5,  98,  23 },
     {   -6,   6, -63,  27,  -9,   6, -80,  28,  -8,   2, -54,  14,  -7,  -3, -33,   1 },
     {   -2,  -3,   8, -55,  -4,   1,  -1, -34,  -6,   8, -26,  39,  -2,  11, -24,  96 },
     {   -3,  -3, -18, -59,  -5,  -5, -23, -77,  -5,  -5, -17, -59,  -6,  -7, -15, -44 },
   },
   {
     {  115, -26, -11,  -4,  19,  24, -13,   0, -21,  27,   0,  -4,  -6,   4,  10,  -4 },
     {  -32, -21,   3,   0,  81, -33, -31,   3,   6,  62, -32,  -9, -14,  23,  21, -14 },
     {   32, -36,   6,   2,   9, -72,  21,   5,  47, -50, -33,  16,   3,  26, -40,   1 },
     {  -10, -50,  32,   1,  58,   3,   2,  -3, -20, -27,  59,   4,  11, -56, -20,  32 },
     {  -18, -15, -24,  11,   7,  48, -46,   1, -30, -12, -26,  13, -12,  27, -89,  -5 },
     {  -13, -66,  21,  11, -64, -36, -10,  26, -60,  23, -20,   7, -10, -15,  12,  -3 },
     {  -20, -54, -14,   1,  17,  47,  39, -12, -14, -42, -14,   3,  32,  42,  48, -39 },
     {   -3, -36, -16,  -7, -28,   3, -68, -14,  53,  -8,  10, -62,  39, -20,   6, -16 },
     {    3,  12, -49,  56,  14,   0, -28,  67,   9, -21, -16,  41,  25, -31,  31,  15 },
     {    7,  37, -16, -14,  22, -34,  -1, -10, -55, -36, -26, -33,  12, -48,  -8, -65 },
     {   -7, -10, -84,  11,  -2, -39,  -6, -50, -26, -11,  40,  -4, -32,  21,  15,  30 },
     {    4,   1,  19,  45,  -9, -11, -22, -31,  16,  12,  50,  59,  -5,   3,  -4, -77 },
     {    9,  32,  43,  10,   4, -24, -45,   0, -42, -23,  19,  -7,  56,  63,  14,  27 },
     {    1,   3,  27,   9,   5,  10, -26,  39,   5, -52,  16, -32, -89,  18,  27, -15 },
     {    0,  -1, -12,  60,   2,  -7,  49,  42,  -3,  28,  36, -64,  16,  22, -34, -17 },
     {   -6,  -4, -30, -82,  -1, -13,  -4,  63,  -2,   8,  49,  23,  12,  21, -10, -27 },
   }
 },
 {  //26
   {
     {  106, -49,   1,  -4,  41,  -2,  -9,   1, -17,  22,  -5,   1,  -7,   5,   1,   0 },
     {   35,  59, -41,   3, -32,  80, -22,  -3, -30,   8,  23,  -8,   2, -18,  14,  -1 },
     {   -5,  57, -34,   4,  93,  -2, -19,  -4,  51,   2,  -4,  -3,  -8,  16,  -1,   0 },
     {   49,  14, -12,  -1, -29, -14,  30,  -6,  53, -84,  18,   6,  23,  -4, -27,  10 },
     {   -1,  30,   0,  -3,  32, -36,  42, -10, -72, -33,  41,   8, -51, -23,   2,  13 },
     {   -7, -28, -74,  49, -19, -42, -53,  34, -12, -25,   6, -10, -21,   5,  12, -12 },
     {  -16, -29,  10,   2,  38,  10, -25,   4,   0, -22,  26,  -9,  55, -90,  24,   3 },
     {   -1, -21,   2,   2,  -7,   1,  26, -21,  38,  17,  74, -47, -16,  25,  65, -17 },
     {    1,  13, -38,  19,   4, -35,  35, -15, -31,  40,  20, -18,  85,  21, -28,   0 },
     {   28,  44,  -3,  -8, -32, -61,  10,   8,  26,  45, -31,  -1, -10, -58,  39,  -1 },
     {   -5,  -8, -26, -66,  -7, -18, -33, -78, -14, -19, -25, -47,  -4,  -3,  -8, -16 },
     {   12,  34,  39, -25,   5, -20, -25,  37, -30, -35,  -5,   0,  39,  41,  47, -54 },
     {    8,  21,  41, -18, -13, -27, -60,  24,   7,  21,  58, -30,  -7,   2, -54,  32 },
     {    7,  12,  31,  54,   7,  12,  19,   7,  -4, -10, -30, -81, -16, -20, -34, -48 },
     {   -8, -15, -34, -57,   3,   8,  26,  47,  12,  17,  24,  13, -14, -22, -40, -73 },
     {   -6,  -8, -22, -43,   7,  12,  29,  63,  -7, -11, -28, -60,   4,   9,  17,  63 },
   },
   {
     { -100,   9,  14,   2,  73, -18, -12,  -1,  -6,  14,   0,  -1,  -3,  -3,   2,   0 },
     {   28, -58,   5,   2,  51,  62, -22,  -5, -63,  -2,  27,   2,  13, -17,  -9,   2 },
     {   42,  69, -14,  -6,  34, -52,  12,   4, -69,  12,  -3,   2,  24,  -4,  -1,  -3 },
     {   46, -19,  16,  -4,  35, -27, -30,   5,  35,  77,  -6, -11, -49, -15,  27,   7 },
     {   -9,  17,  71,   0, -47, -16, -72,   7, -16,   6,  37, -15,  28, -19,  -7,  10 },
     {   32,  -1,  39,  -3,  45, -13, -32,   4,  26, -56, -20,   1,   6,  79,  -6,  -9 },
     {    9,  68,   2,   3,  17,  42, -10, -11,  16, -38,  40,   6, -72, -26, -10,  12 },
     {   -9,  12,  42, -13, -16,  34,   7,  25, -41, -11, -73, -14, -30,  -8,  61, -10 },
     {  -13,  30, -24,   7, -19,  45, -16, -17, -17,  55,  27,  -4,  -1,  83,  28,  -8 },
     {    5,  -1,  19,  81,  -4, -11,  -3, -84,  -8,  -3, -24,  34,  -1,  -9,  17,  -3 },
     {  -14, -42, -12,   0, -27, -55,  -4,   4, -53, -23,  28,   8, -74,  29,   1,   3 },
     {    1,  -5, -44,  14,   9, -14, -21,  -6,  10, -39,  20, -54,  23, -11,  78,  43 },
     {   -6,   7, -64,  19,  -8,   6, -81,  18,  -6,  -2, -57,   3,  -9, -10, -37,  -5 },
     {   -1,  -2, -17, -52,  -1,  -6, -30,  -8,  10, -12,  15,  85,  14, -16,  52, -38 },
     {   -3,   0,   4, -51,  -4,   4,  -3, -37,  -8,   7, -28,  28,   0,  12, -14, 102 },
     {   -2,  -3,  -6, -61,  -3,  -5, -10, -79,  -4,  -5, -11, -59,  -7,  -8, -16, -47 },
   },
   {
     {  116,  14,  -6,   1, -42,  23,   6,  -1, -14, -10,   7,   1,  -1,  -3,  -2,   1 },
     {  -26,  52,   4,   4, -79, -44,  30,   4,  30, -46, -22,   7,   4,  13, -13,  -4 },
     {   33,   2,   1,  -4,  14, -70, -21,   9,  28,  56, -54, -17, -10,  31,  33, -20 },
     {   17, -64, -15,   2,  31, -10,  26,   3,   1, -58, -46,  25,   4,  39, -45, -23 },
     {  -17,  33, -46, -10,   1,  60,  26,   3, -28,   0, -44, -19,  13,  33,  50, -41 },
     {  -17, -81,  -9,   1, -78,  17, -41,   1,   5,  10,  -5, -27,  15,   3,  16, -10 },
     {   -7,  14,  23,   9,  -9, -21, -29,  10, -76,   6, -24,  10,  -1, -44, -33, -67 },
     {    0, -31,  23, -15,  -6, -31,  60,  36, -53,  -3, -14,   5,  17, -25,  52,  46 },
     {  -11,  12, -26,  22, -19,  21, -23,  44, -25,  38, -35,  43, -28,  28, -36,  60 },
     {   -4, -20, -29,  14, -19,  -3,  51, -37,  16,  50,  16,  73, -21, -22,  10, -34 },
     {   -1,   2, -90, -10,   2, -51, -19, -43, -40, -19,  17, -18,  -6,  -9,   0,  26 },
     {    0,   5,  37,  57,   0,  -7, -27, -55, -32, -14,   6,  36,  29,  53,  38,  12 },
     {   -7,  -5,  37, -54, -17,   1,  30, -44, -36,  27,  12, -29, -45,  53, -32,   3 },
     {   -2,  -8,  13,  35,   2,  11,  -5, -14,   4, -37, -27, -15,-101, -27,  35,   5 },
     {    1,  -3,  -4,  79,   2,   2,  47, -20,  -1,  31, -16, -64,  21, -12, -38,  11 },
     {   -2,  -5, -15,  38,   0, -22,  12,  64, -15,  -4,  72, -14, -28,  47,  10, -35 },
   }
 },
 {  //27
   {
     { -107,  48,   2,   3, -43,   2,  12,  -1,  11, -21,   7,   0,   7,  -8,   0,   1 },
     {  -40, -20,  15,   0,  79, -71,   4,   1,  41,  -3, -26,   6,  -4,  21, -12,  -1 },
     {    2, -73,  49,  -6, -63, -45,  38,   3, -13, -27,  -1,   8,  10,  -9, -10,   3 },
     {   44,  29, -24,   1, -11,  -8,  17,  -6,  74, -75,  15,   0,  27,  -3, -19,   7 },
     {    0, -13, -30,  15, -51,   0, -46,  21,  27,  20, -71,   8,   4,  50, -31, -12 },
     {   27,  58,  11, -14, -20, -31,  64, -10,  -8,  23, -10,   9, -54,  53,   3,  -2 },
     {    2,   4, -63,  20,  -7, -58, -29,  13, -51, -46,   3,   3, -48, -23,   8,   0 },
     {   -2,  14,  13,  -4,  25,  28,  11,  -2, -46, -27, -23,  39,   0,  -6, -93,  28 },
     {   20,  43,  14, -20,  -4, -30,   3,  13,  -5,  19, -65,  22,  39, -72,  24,  -6 },
     {    0,   7, -47,  22,  -1, -38,  28, -12, -33,  38,  27, -20,  81,  21, -25,   1 },
     {    5,   7,  14,  75,   6,   6,  21,  87,   9,  10,  21,  40,   2,   0,  11,   8 },
     {    2,  -3,  -8,   9, -20, -20,  -5,  -6,  43,  59,  25, -20, -41, -56, -50,  42 },
     {   15,  40,  59, -25,  -9, -35, -69,  32, -14,  -5,  39, -24,  15,  24, -17,   6 },
     {   -6, -10, -24, -46,  -8,  -9, -18,  -7,   8,  14,  26,  83,  13,  19,  28,  61 },
     {    9,  14,  36,  79,  -6, -10, -28, -78,  -3,  -4, -11,  18,   5,   6,  19,  28 },
     {    2,   2,   2,   1,  -5,  -9, -15, -25,  11,  16,  37,  61,  -6, -15, -26, -95 },
   },
   {
     {   96, -18, -12,  -3, -74,  27,  10,   1,   8, -18,   1,   2,   3,   4,  -2,  -1 },
     {   27,  77, -44,   0,   2, -60,  42,  -3, -44,   3,  -6,   6,  20,   0,  -6,  -2 },
     {   37, -28,   5,   0,  52,  34, -23,  -3, -88,   7,  30,   1,  24, -23,  -9,   3 },
     {   42,  -5,  32, -21,   8, -49, -34,  28,  14,  81, -16, -22, -12, -16,  29,   4 },
     {    0, -47, -60,  38,  16,  11,  52, -36,  15,  56, -16,   3, -13, -28,   7,   8 },
     {   44,  28, -11,  -4,  64,  15,  -3,  -6,  26, -29,   5,   7, -86,  13,   8,   2 },
     {  -17,  36,  -5,  -1, -33,  12, -10,   3,   7,  13,  64, -17, -32, -78, -28,  31 },
     {   10, -17, -37,   8,  14, -28, -25,  -2,  32,  16,  73, -17,  24,  62, -38,   9 },
     {  -13, -22, -40, -79,  11,  17,  38,  75,   0,   1,   1, -10,  -4,  -3, -12, -12 },
     {   18, -34,  23,   5,  22, -49,   6,   4,  19, -33, -33,  10,   4, -39, -85,  12 },
     {  -18, -52, -26,   6, -32, -61, -20,   4, -47, -31,  14,   6, -57,   2,  28,  -3 },
     {    1,   0,  -5,  34,   6,   6,   8,  24, -10, -26, -25,-103,   3,   9,  10,  48 },
     {  -11,   5,  37,   3, -25,  17,  30,  -3, -42,  46,  -9,  -6, -52,  53, -59,   4 },
     {    9, -20,  62, -13,  12, -24,  79, -18,   7, -15,  55, -12,   6,  -1,  33,   4 },
     {   -4,  -5,  -7, -59,  -3,  -1,  -8, -41,  -4,  -1, -16,  24,   7,  15,   9,  99 },
     {   -2,  -1, -10, -57,  -3,  -4, -14, -77,  -2,  -4, -10, -61,  -4, -10, -14, -52 },
   },
   {
     {  108,  31,  -6,   1, -57,   8,  12,   0,  -2, -15,   1,   2,  -2,   1,  -2,   0 },
     {   37, -58, -21,  -1,  53,  78,  -3,  -8, -36,  -1,  29,   5,   2,  -9,  -5,   4 },
     {   40, -12,  20,  -1,  27, -37, -61,   3,   7,  80,   3, -28,  -9,  -8,  32,   8 },
     {  -26,  43,  21,   7, -30,  36, -21, -21,  13,  12,  83,  -5,  -3, -35,  -1,  42 },
     {   12,  86,  14,  -4,  74,  -7,  29,   9, -40,   5,  -3,  15,  -7, -16,  -2,  -3 },
     {   22, -31,  36,  21,  24, -49,   0, -14,  25, -15,  26,  54, -21,  -8, -69,  18 },
     {   -4,  -6,  55,  11, -15,  -7, -54,  24, -69, -46,  -9, -32,  24, -16, -24, -10 },
     {    9, -17,  18, -25,  16, -15,  33,  44,   8, -28,   5, -28,  16,  21,  26,  92 },
     {   -6, -13, -70, -10, -19, -60,   9, -19, -66,  11,  27,   8,  18, -27,  -2,  22 },
     {    9,  28, -60,  15,  37, -11, -53, -18,  35, -51,   9, -45,   3,  25, -22,   6 },
     {   12, -19,  29, -29,  18, -28,  35, -47,  21, -35,  35, -39,  19, -34,  38, -48 },
     {    1,   5,  23,  41,   3,  -4, -11, -62, -23,  -8,   3,  38,  36,  67,  47,  15 },
     {    0,   9,   0, -88,   0,  -3, -34,  27,   2,  -4,  39,  43,  22,  44,  -3, -27 },
     {   -4,  -5,  -2, -21,  -4,  -5, -28, -12, -17, -43, -10,  25, -96, -12,  51,  12 },
     {   -5,   0,  17, -28, -10,   1,  27, -40, -31,  26,   9, -55, -50,  61, -46,   2 },
     {    1,   7,   8, -57,   0,  17, -20, -63,   9,   3, -64,  10,  28, -35, -23,  40 },
   }
 },
 {  //28
   {
     {  118, -29, -10,  -3,   2,  24,  -7,   0, -25,  14,   2,   0,  -4,   1,   1,   0 },
     {   -6, -16,   4,  -1, 106, -25, -22,   0,   7,  51, -16,  -3, -21,  17,   9,  -3 },
     {  -27,  17,   6,   0,  20,  51, -12,  -3, -83,  22,  39,  -3,   4, -55,  15,   7 },
     {   27, 100, -24,  -6,  41,  13,  32,  -7,  15, -32,  17,   9, -18,   8, -11,   4 },
     {    6, -36,  14,   2,  43, -43,  18,   1, -29, -76,   8,  16,  27, -31, -45,  11 },
     {   15,  42,   2,  -8, -24, -83,  13,  17, -52,  24, -44,  -2,  -1, -19,  27,  -7 },
     {   16, -11,  28,  -4,   3, -29,  33,   2,  50,   8,  64,  -7,  -1, -48,  63,  16 },
     {   -8, -13,-101,   3,  -6, -37, -39, -18,  -9, -15,  42, -13, -11,   5,  16,   7 },
     {  -11, -24,  15,   6, -12, -12,  58,  -4, -39,   6,  40,  10, -74,  53, -12,  -3 },
     {    4,   6,  -4,  -3, -16, -14, -10,   2,  26,  39,   4, -24, -47, -58, -82,  25 },
     {    1,  -2,  24,  -8,  -2,  11, -51,   7,   2, -58, -31,   5, -82, -23,  37,   0 },
     {   12,  30,  54,  -9, -11, -29, -72,  14,  -7,  -1,  59,  -3,  16,  40, -21,  -8 },
     {    2,   4, -12,  93,   5,   4,   0,  84,   4,  -3,   8,  -5,  -4,  -5,   0, -19 },
     {   -3,  -1, -13, -10,  -7,  -1, -12,  25,   7,  21,   4, 116,  -1,  -4,  -1,  37 },
     {    8,  11,  21,  85,  -6, -12, -13, -84,  -2,   6,  -9,  18,   0,  -2,   4,  23 },
     {   -1,   2,   4,   2,   2,   5,   2,  26, -11,  -7, -13, -37,   7,  30,  12, 113 },
   },
   {
     {   97,  16, -14,  -4, -76,  -3,  15,   3,  18, -10,  -5,   0,  -4,   5,   0,  -1 },
     {  -24,  76,  16,  -8, -13, -87,  -6,  12,  19,  33, -14,  -9,  -2,  -3,  11,   2 },
     {   40,  46,   4,  -6,  37,  -9,  -6,   3, -93, -21,  15,   6,  41,   0, -13,  -4 },
     {   24, -15,  69,  23,  -9,  -2, -85, -20,   7,  21,  44,  -3,  -3,  -7,  -6,   9 },
     {   43, -21, -32, -14,  46,  -8,  10,  12,  -6,  83,  11,  -9, -31, -47,   7,   9 },
     {  -25, -42, -41,  -9, -46, -31,  -2,   3, -15,  36,  46,   2,  70,  14, -20,  -2 },
     {   26, -42,   9,   2,  23, -24, -17,   5,  -6,  13, -57, -17,  28,  64,  62,  -1 },
     {    2,   4, -16,  77,   0, -16,  21, -85,  -8,  16, -17,  37,   5, -11,   8,  -3 },
     {  -20, -13,  30,  14, -47,  17,  11,  11, -43,   7, -42, -44,  19, -68,  28,  25 },
     {   15, -46,  18,  13,  17, -56,   9,   9,   4, -26, -40, -15,  -4, -15, -85,   2 },
     {    9,  34,  23,  37,  20,  45,  33,  10,  35,  41,  -1, -48,  49,  26, -36,   0 },
     {    0,  -7,  50, -28,  -9,  15,  17,  14,  -1,  35, -29,  71,  19, -14,  -8, -67 },
     {   14, -21,  44,  -2,  23, -32,  63,  -2,  27, -34,  57,  -2,  24, -23,  44,   4 },
     {  -12, -13,  31,  18, -26,  -8,  43,   7, -53,  24,  27, -22, -65,  53,  -6, -21 },
     {   -2,   1,  30, -63,  -4,   9,  24, -50,  -6,  20, -13,  20,   2,  23, -20,  81 },
     {   -1,  -1,   3,  52,  -1,   0,   5,  73,  -3,   4,   2,  64,  -2,  10,   3,  63 },
   },
   {
     {  -90,  81, -34,  13,   0,  -8,  10,  -5,   9, -11,   5,  -2,   2,  -2,   0,   0 },
     {   21,   3, -19,  10,  85, -81,  35, -11,  -6,   9,  -6,   2,  -9,   9,  -5,   2 },
     {   64,  41, -82,  29,   7,  36, -29,   1, -15,  12,  14, -12,  -3,  -1,   8,  -1 },
     {  -20,   6,   8,  -5,  -6,   5,   1,  -3, -83,  86, -38,   8,  -1,  -6,   8,  -4 },
     {  -33, -30, -10,  28,  40,  -1, -82,  50,  27,  11, -41,   6,   0,   9,  -3, -11 },
     {  -24, -12,  22, -18,  65,  40, -22, -29,  -7,  16,  65, -42,   4, -21,  31,   8 },
     {   24,  49,  23, -61,  37,  39,  -4, -22,  42,  14, -48,  33,   7,   8, -14,   7 },
     {   -9,  -3,   4,   0,  -2,  16,  -1, -11,  -2,  12,  16, -18, -79,  88, -36,   6 },
     {   13,  18,   6, -21, -31, -38,  -7,   2,  48,  42,  -9, -60,  -4,  17,  61, -37 },
     {  -17, -43, -26,  52,  -1,  32,  40, -59,  49,  36, -27,  -1,  17,   1,  -2,   4 },
     {  -13, -25, -41, -50, -16, -28, -45, -67, -19, -27, -29, -31,   1, -14, -28,   4 },
     {    9,  14,  11,   5, -19, -26, -17,  16,  26,  47,  23, -26,  22, -16, -53,  82 },
     {   -1,  -3,  -5,   1,  -7,  -5,  -2,  -2,  19,   5, -11,  22, -94, -68,  30,  32 },
     {   16,  30,  50,  50,  12,  14,   3,  -9, -16, -30, -48, -69, -13, -26, -25, -10 },
     {  -13, -26, -42, -45,  14,  30,  54,  62,  -2,  -8, -28, -50,  -2,  -6,  -4,  20 },
     {    0,   2,   7,  14,   0,  -1,  -8, -12, -19, -31, -30,  -1,  16,  45,  70,  81 },
   }
 },
 {  //29
   {
     { -116,  38,   7,   3,  -6, -21,   8,   0,  24, -17,  -2,   1,   6,  -2,  -2,   1 },
     {   13,  18,  -9,   1,-104,  37,  19,  -1,  -3, -46,  19,   3,  19, -17,  -8,   5 },
     {    4,  88, -46,   2,  32,  57, -14, -10, -26,  18,  24,  -8, -12,  -6,  12,  -1 },
     {  -43, -50,  25,   2, -13,  46, -38,   6, -63,  41,  14, -10,  13, -36,  15,   2 },
     {   -6,  -4, -19,   5, -42,  30, -43,   6,  53,  50, -46, -16,  -8,  54,  15, -23 },
     {    6,  39,   3,  -6, -37, -56,  16,  19, -73,  34, -50,   7, -18,   8,   3, -15 },
     {    9,  29,  85, -39,  13,  51,  35,  -3,  14,  18, -33,  23,  15,   3, -18,  -2 },
     {   15,  19,  26, -11, -20, -32,  14,   0,  39,  46,  40, -20,  -9, -40,  75,  21 },
     {  -18, -24,   3,   6, -12,  18,  56, -26, -14,  20,  43,   9, -74,  54,  -5,   9 },
     {    3,   8,  36, -20,   0,   7, -54,  31,  -4, -61,  -6,   9, -75,  -1,  33,  -9 },
     {   -3, -10, -27,  11,  -4,  12,  14,   1,  28,  21, -41,   9, -59, -81, -43,  14 },
     {   11,  26,  34,   0, -17, -29, -55,  -3,   8,  27,  49, -14, -10,   7, -80,  18 },
     {    6,   9,  23,  88,   8,  13,  23,  81,   7,   5,  14,  -3,   2,   3,  -4, -18 },
     {    1,  -1,  15,  -7,   5,   2,  21, -19,  -4, -16,  -4,-103, -11, -12, -13, -62 },
     {   -8, -13, -35, -79,   2,   2,  18,  79,   8,  14,  31,   2,   4,   1, -18, -24 },
     {    2,   0,   3,   9,  -5, -10, -10, -31,  10,  11,  28,  62,  -1, -22,   2, -99 },
   },
   {
     {  -99,   6,  14,   3,  72, -22, -13,  -1,  -9,  20,   0,  -2,   1,  -6,   2,   1 },
     {   10,  89, -24,  -6,  -9, -73,  36,   6, -18,   8, -22,   1,  11,   1,   3,  -3 },
     {   47,   4,  -6,  -1,  57,  22,  -7,  -3, -93,   1,  23,   4,  25, -21,  -9,   2 },
     {  -49,  19, -16,   7, -27,  35,  32, -11, -22, -85,  12,  19,  28,  12, -32,  -6 },
     {    0,  27,  82, -21, -28, -16, -66,  31, -17, -27,   5, -15,  24,  13,   2,   0 },
     {   27,  50,  11,  -9,  57,  26,  -8,   0,  33, -44,   5,  11, -73,  21,   2,  -4 },
     {  -16,  40,   1,   3, -33,  19,  -9,  -7,   7,  26,  69,  -3, -28, -76, -18,  24 },
     {   12,  -6, -31,  11,  11, -43, -36, -23,  24,   3,  72,   9,  21,  56, -43,   0 },
     {   17, -45,  39,  -5,  18, -60,  34,   9,  11, -37,  -2,  23, -14, -50, -51,   5 },
     {   -9, -11, -34, -88,   6,  10,  15,  78,   6,   6,  23, -10,   3,   6, -11, -10 },
     {  -19, -30, -21,  -6, -35, -36, -18,  -3, -63, -16,  13,  10, -79,  18,  28,  -1 },
     {    6, -12, -14,  -3,  17, -21,  10,  -3,  18, -48,  31, -30,  27, -15,  77,  62 },
     {    0,   1,  31, -27,  -3,   2,  22, -17,   6,  17,  32,  94,  18,   6,  51, -26 },
     {    2,   1, -55, -10,   4, -10, -73,  -4,  18, -29, -33,  34,  16, -54,   9, -32 },
     {   -3,   2,   2, -77,  -5,   5, -11, -71,  -4,   8, -31,   5,  -1,   8, -25,  58 },
     {    2,  -5,  12, -33,   2,  -9,  16, -56,   3, -11,  22, -62,   1, -15,  15, -82 },
   },
   {
     {   81, -68,  19,  -5,  51, -41,   4,   2,  -1,  11, -14,   5,  -7,  11,  -6,   0 },
     {   46,  34, -62,  14,  32,  49, -63,   8,  11,  27, -15, -13,   2,  10,   8, -15 },
     {   25, -22,  19,  -7, -64,  40,  -1,   3, -76,  53, -10,   2, -20,   6,   5,  -4 },
     {   69,  18,   1, -18, -27,  30,  34, -26,  14, -40,  59, -20,  19, -30,  22,   5 },
     {   12, -36, -28,  40, -49, -17, -38,  51,  -3, -48, -13,  25,  25, -44,  -5,   9 },
     {   -1, -12,  42, -40,  -8,  38, -13, -16,  43,  20, -53,  15,  68, -27, -14,   4 },
     {  -36, -39,  13,   5,  42,  12, -22,   5, -23,  19,  39, -25,  21, -52,  65, -22 },
     {   -3, -31,  -7,   7, -54, -24,  -6,  -5,  53,  13,  -2, -50,   9,  48,  33, -53 },
     {   21,  46,  24,  22,  -4, -19,  31,  52,  25,  38,  -9,  49,   2,  -3,  65,   1 },
     {   -9, -36,  30,  42,  12,  68,   9,  40,  33,  -2,  45,  13,  -4,  42, -26,  15 },
     {   -3,  -9,  25, -56,  -6,  17, -49,   2,  21, -43,  -5,  31, -70,   7,  40,  -1 },
     {   -5,  -9, -19,  -4,  23,  34,  44,   8, -39, -58, -44,  15,  31,  41,  34, -39 },
     {  -14, -39, -61,  -6, -10,   2,  12, -45,  13,  23,   9,  42,  11,  20,  38,  65 },
     {   -4, -21, -19,  31,   2,  31,  50, -12,  32,   5, -48, -20, -62, -57,   0,  -8 },
     {    0,   4,   6, -20,   3,   2,   5,  47,  -6, -11, -29, -76,   3,  14,  23,  79 },
     {    8,  15,  50,  76,   0,   0, -31, -64, -10, -24, -24,  -7,   1,  14,  27,  25 },
   }
 },
 {  //30
   {
     {  117, -29, -12,  -3,  -6,  29,  -7,   0, -24,  14,   4,  -2,  -4,   0,   5,  -2 },
     {    6,  27,  -7,   0,-100,  22,  29,  -1,   7, -58,  14,   6,  16,  -9, -15,   6 },
     {  -32,  19,   0,   0,   9,  78, -20,  -8, -58,  19,  51,  -9,   6, -42,  13,   8 },
     {   29, 100, -33,  -7,  50,   2,  32, -11,  10, -16,   7,  10, -13,   7,  -3,   0 },
     {    5, -42,  13,   4,  54,  -5,  14,  -5, -10, -75,  33,  22,  17, -22, -53,  20 },
     {    0, -13,  -4,   2,  14,  62, -30, -15,  97,  -7,  22, -16,  10,  25,   4,   4 },
     {   19,   5,  32,  -6,  -4, -44,  18,   5,  39,  19,  48,  -6,   5, -65,  48,  42 },
     {   -4,  -7, -93,  11,  -7, -43, -54, -18,  -7, -20,  23, -31,  10, -11,  10,  10 },
     {  -18, -43, -61,  12,  -2,  18,  72, -23,  15,  31,   3,  33, -40, -20,  -1,   0 },
     {   -6, -18,  -9,  -3,  18,  12,  34,  -4, -19, -38, -17,   4,  54,  31,  91,   1 },
     {   -3,  -6,  14,  -7,  -1,   9, -27,   4,  -4, -55, -11,   6,-100,  -6,  44,   5 },
     {    3,   5, -15, -12,  11,  20, -18,  14,  21, -13, -75,  10,  25, -85,   1, -26 },
     {   -2,  -6,  13, -81,  -5,  -4,   4, -91,  -7,   2, -19, -22,   0,  -2,  -9,  18 },
     {   -2,   8, -10,   6,  -5,  12, -26,   7,  -4,  13, -35,  54,   8,  15,  -5, 103 },
     {    5,   8,  15,  24, -10, -16, -37, -56,   3,   4,  20,  88,  12,  -2,  17, -44 },
     {   -7,  -8, -24, -93,  -3,  -3, -12,  58,   5,   5,  33,  44,   4,  12,   2, -18 },
   },
   {
     { -109,  -9,  16,   3,  60, -14, -14,  -1,   1,  13,   0,  -1,   0,  -3,   1,   1 },
     {  -32,  56,   2,  -3, -59, -67,  21,   8,  51,  -7, -29,  -1,  -4,  14,   4,  -3 },
     {  -28, -82,   6,   7, -56,  40,  -9,  -4,  61, -12,   2,  -3,  -6,  11,  -2,   2 },
     {  -32,  18, -13,   3, -13,  34,  41,  -4, -35, -90,   2,  20,  27,  10, -35,  -7 },
     {    6, -22, -72, -14,  43,   5,  63,   2,  25,  11, -43,  12, -25,  27,  14, -17 },
     {   11, -56,  25,  -3,  -3, -51, -12,  22, -44, -10, -57,  -4,  45,  38,   1, -22 },
     {  -24, -25, -62, -13, -42, -28,  21,  -2, -37,  44,  39,   4,  42, -37,  -5,  13 },
     {  -22,  28,   0,  -9, -25,  58,   4,  15, -25,  16, -32, -38,  25,   9,  82,   7 },
     {    3,  16, -38,   6,  10,  -5, -45, -36,  18,  -3,  43,   5,  41,  84,  14,  -4 },
     {    0,   3,  -7, -85,   8,   5, -15,  79,  13, -13,  29, -26,  -2,  15, -21,   2 },
     {    8,   9,  55,   1,  11,  23,  57,  18,  28,  47,  21,  41,  58,  17, -11, -17 },
     {  -16,  21, -14,  -4, -28,  32, -34,   1, -22,  42, -22,  17, -25,  10, -44, -83 },
     {   -8, -13,  20, -31, -18, -15,   4,   1, -31, -10,  31,  77, -50,  18,  58,  -7 },
     {   -6,  -7,   9,  58, -10,  -9,  39,  35, -30,  12,  44, -51, -44,  48,  -7,  -5 },
     {   -7,   7,   8, -21, -15,  17,   1, -17, -24,  35, -33,  17, -20,  44, -45,  86 },
     {    0,   8, -32,  62,   3,   8, -34,  79,   8,  -2, -14,  53,   9,  -7,   4,  28 },
   },
   {
     {   93, -73,  17,  -5,  31, -10, -11,   5, -15,  23, -10,   0,  -5,   5,   3,  -3 },
     {  -41, -55,  75, -16, -50, -31,  38,   6, -13,  -8, -13,  17,   6,  -5, -13,   7 },
     {  -18,   4,  -1,   4,  80, -80,  27,  -3,  36, -21, -10,  12,  -8,  14, -13,   5 },
     {  -48,  -1,  -4,  11,  21, -12, -31,  17, -51,  76, -48,   2, -22,  20,   9, -20 },
     {  -21, -36,  -2,  12,  -6,   9, -72,  36,  68, -23, -39,  -2,  30,  -4,  -4, -14 },
     {    1, -28, -57,  66, -37, -46,  -4,  34, -38, -32,  23,  -3, -16,   0, -10,   4 },
     {  -37, -33,  13,   3,  36,  -3, -15,  -8, -12,   3,  62, -48,  19, -47,  57, -10 },
     {   13,  20, -14,   5,   9, -15,  10,   7, -29,   8, -36,  34,  76, -76,  -4,  15 },
     {    9,  -6, -32,  -1, -42, -40,  25, -51,  30,  36, -22, -57,  39,  19,  12, -29 },
     {  -30, -42, -24,   6,  25,  33,  -8, -55, -17,  -1,   3, -16,  18,   6, -76,  40 },
     {    0,   0,  21,  72,  13,  33,  47,  38,  33,  48,  26,   5,  34,  24,  -8,   6 },
     {    4,   5,  -3,  -5, -24, -29, -26,  -2,  36,  49,  15,   2, -27, -22,  12,  95 },
     {   19,  45,  46,  -3,   0, -26, -28,  36, -21,  -3,   8, -73,  16,   2, -54,   5 },
     {    6,   2,  17,  50,   4,  21,  24, -20,  17,  -9, -51, -37, -62, -63,  -5,  -7 },
     {   -1,   1,   6,  13,  12,  17,  13,   5, -27, -42, -48, -33,  23,  45,  58,  62 },
     {  -11, -23, -49, -60,   7,  19,  50,  69,   9,   9, -10, -35, -12, -17, -12,   8 },
   }
 },
 {  //31
   {
     {  116, -32, -12,  -2,   1,  31, -11,   0, -23,  14,   5,  -2,  -3,  -1,   3,  -1 },
     {   -8, -34,   8,   0,  97, -34, -30,   4,   4,  52, -23,  -4, -11,  12,  11,  -5 },
     {  -31,  24,   0,  -2,  14,  70, -24,  -9, -37,  42,  50, -19,   4, -37,  40,   5 },
     {  -31, -97,  31,   9, -55,   3, -34,  11, -14,  11,  -6, -13,  12,  -6,   0,  -1 },
     {    9,  31, -13,  -8, -51, -22,  -2,   6,  16,  59, -44, -24,   3,  24,  67, -24 },
     {    0,  -6, -16,   3,   8,  52, -44, -12,  97, -17,   5, -25,  14,  29,  -3,  -3 },
     {  -19,  20, -39,  12,  -9,  40, -40,   8, -44,  13, -53,  15, -23,  36, -57, -20 },
     {   -1,  -2, -79,   3,  -6, -57, -52, -16, -19, -28,  29, -37,   5, -19,   6,  10 },
     {   20,  47,  69, -15,   1, -15, -59,  42, -11, -26, -14, -41,  13, -15, -15,   3 },
     {   -9, -16,  -7, -13,  30,  23,  22,  -1, -39, -60, -28, -27,  53,  31,  43, -39 },
     {   -4,  -7,  18,   2,  -2,   2,  -4,   8, -11, -34,  29, -17, -99,  49,  36,  -6 },
     {   -1,   0,  21, -62,  -7,  -7,  14, -76, -14,  23,  10, -48,   3,  35, -40,  11 },
     {   -3,   2,  -6,   0,   0,   7,  -5,  18, -14,  -1, -11,  12,  23,  48,  22, 110 },
     {   -2,  -6,   1, -57,  -1,  11, -28, -42,   7, -30, -54,  42, -31, -50,  33,  17 },
     {   -8, -17, -32, -32,   8,  20,  45,  54,  12,   4, -29, -66, -29, -39, -18,  23 },
     {    6,   6,  18,  87,   4,   6,  13, -56,  -4,  -9, -42, -43, -14, -24,   4,  23 },
   },
   {
     {   98,  19, -15,  -4, -73,   1,  19,   3,  11, -17,  -7,   1,  -2,   7,  -1,  -1 },
     {  -36,  61,  14,  -5, -32, -81,   1,  12,  41,  25, -26, -10,  -7,   5,  16,   0 },
     {   29,  70,   2, -10,  47, -26,   5,   8, -82, -10,   3,   3,  26,  -7,  -7,  -3 },
     {  -51,  15, -26,   0, -21,  17,  55,  -1,  -6, -78, -32,  15,  21,  32, -13, -15 },
     {    5,   3, -72, -15,  32,   1,  70,   5,  14,  44, -10,  13, -34, -37,   0,  -6 },
     {   30, -47,  -4,  -8,  39, -27,   1,  16,  -3,   8, -71, -16,  16,  57,  40, -13 },
     {   18,  43,  48,  12,  42,  42,   1,  -4,  34, -32, -36,   0, -66,  -8,  16,  -2 },
     {  -18,  15,  -4, -38, -22,  49,   4,  50, -16,   7,  -4, -63,  14, -18,  54,  36 },
     {    3, -34,  26, -37,  -1, -25,   2,  63,  -4, -23, -21, -20, -20, -37, -73, -12 },
     {    6,  20, -21, -69,  25,   8, -26,  33,  39, -11,  52,  27,  -5,  54,  -2, -17 },
     {   14, -21,  24,  -4,  26, -38,  48,  -5,  20, -34,  35,   3,   7,   3,  11,  88 },
     {   -4, -14,  67, -34, -19,  12,  42,  10, -16,  29,   0,  70,  17, -13,  33, -26 },
     {   11,  23,  28,  28,  21,  33,  36,  14,  43,  40,   4, -29,  63,  22, -48,  -8 },
     {   13,  -2, -15,  -2,  23, -11, -22,  -1,  47, -41,  -5,   8,  62, -72,  32, -29 },
     {   -2,  14, -19, -10,  -3,  23, -39,  10,   7,  18, -57,  61,  22,  -6, -29,  70 },
     {    1,   1,  12, -78,   0,   4,   3, -89,   4,   6, -21, -33,  10,  -6, -22,   2 },
   },
   {
     {  -87,  82, -37,  15,  -6,   1,   3,  -2,  12, -14,   8,  -3,   3,  -3,   1,   0 },
     {   18,   5, -16,   5, -72,  88, -47,  12,  -3,  -2,   9,  -8,   9, -11,   8,  -3 },
     {  -61, -30,  82, -37, -47,   5,  19,  -3,  20, -19,  -7,  10,   7,  -3,  -6,   3 },
     {  -37,  -2,  14,  -3,   8,   9, -19,   9, -72,  85, -40,   7,  -4,  -1,   9,  -8 },
     {   31,  37,   3, -24, -36,   6,  78, -49, -48,  13,  32,  -5,  -2,  -8,   5,   8 },
     {  -23, -20,  18, -11,  56,  27, -18, -21, -11,  -1,  69, -43,  17, -39,  42,  -6 },
     {  -27, -54, -22,  68, -47, -39,  16,  19, -35,  -8,  35, -28,  -8,   2,   0,  -5 },
     {  -13,  -5,   7,  -6,  15,  31,  -8, -16,   0,   5,  29, -19, -79,  80, -25,  11 },
     {    9,   9,   0, -10, -33, -32,  -2, -11,  55,  44, -13, -47, -18,  16,  72, -28 },
     {  -21, -44, -31,  38,  10,  39,  30, -73,  39,  26, -28,  20,  13,  -6,  -4,  13 },
     {   -4,  -9, -18, -28,  -2,  -6, -13, -38, -19, -25, -33, -60,  23,   8, -50, -69 },
     {    4,   8,   3,  -9, -14, -26, -28, -11,  18,  40,  13, -49,  30,  -9, -52,  80 },
     {  -17, -31, -55, -66, -21, -34, -40, -33, -17, -14,  14,  42, -14,  -1,  22,  20 },
     {  -11, -19, -29, -27,  14,  26,  39,  37, -18, -30, -43, -43,  24,  33,  39,  49 },
     {   11,  22,  40,  43,  -3, -16, -42, -56, -33, -40, -23,   2,  22,  36,  38,  25 },
     {   -2,  -3,  -6,  -8,  -1,   0,   4,  12,   8,  28,  44,  29,  81,  74,  -1, -25 },
   }
 },
 {  //32
   {
     {  112, -51,  -4,  -2,   6,  20, -11,   1, -20,  14,   4,  -1,  -3,  -1,   3,   1 },
     {   18,  62, -37,   1, -61,  69,   8, -11, -11, -23,  35,  -5,   6, -15,   5,   4 },
     {   12,  61, -41,   1,  94,   3, -12,  -9, -15,  31,  14,  -5, -12,  -4,  18,   2 },
     {  -43, -33,  16,   7,  -6,  54, -67,   9, -26,  56,  11, -26,   5, -17,  26,  -3 },
     {   -6,   5,  52, -17,   1, -12,  59,  -5, -79,  20,  31,  16,   2, -38,   8,  15 },
     {  -12, -37,  20,   0,  47,  30,   1, -21,  39, -53,  70,  -2,  13, -24, -13,  27 },
     {  -21, -37, -74,  41,  -6, -37, -13,  -8, -61, -27,  19,  -7,   6, -11, -30,  -2 },
     {   -1,  18,  28, -18,  28,  36, -36,  34, -43, -52, -38,  29,   7,   2, -54, -24 },
     {  -25, -36, -26,   9,  17,  62,  56, -35,  -4,  22, -25,  39, -22,  41, -14, -11 },
     {    2,  -4, -21,   5,   9,   7,  14,  -6,  19,  17, -43,  24,  93, -61,   3,   2 },
     {   -5, -14,  -3, -35,  14,   5,  21, -18, -20, -43, -10, -56,  30,  20,  57, -63 },
     {   -1,   2,  -6, -76, -12, -14, -29, -82,  -6,  14, -11, -14,   2,   0, -42,  21 },
     {    8,  16,  22,  29,   5,   4,  21,   7,  -1,  34,  23, -56,  53,  48, -64,  -9 },
     {   -8, -11, -30, -62,  -6, -11,  -7,  47,   1,  17,  51,  51,  38,  45,   9,  -5 },
     {   -7, -15, -30, -45,   3,   9,  33,  50,  22,  19,   1, -41, -38, -55, -41, -29 },
     {   -7,  -8, -17, -20,   8,  16,  17,  37, -16, -18, -36, -40,   8,  23,  15,  97 },
   },
   {
     {   99,  -9, -16,  -3, -68,  28,  14,   0,   7, -23,   1,   2,  -2,   5,  -3,   0 },
     {   38,  79, -26,  -8,  30, -41,  28,   4, -64,   5,   1,   5,  17, -13,  -5,  -1 },
     {   29, -46,   6,   2,  57,  60, -30,  -7, -59,   4,  38,   1,   9, -21,  -8,   5 },
     {   47, -13,  11,  -6,  16, -38, -32,  10,  25,  86,  -7, -22, -22, -13,  40,   9 },
     {   -9, -27, -81,  33,  17,  13,  61, -43,  16,  34,  -2,  14, -12, -10,   7,   5 },
     {  -19, -48,  -7,  10, -55, -40,  -2,   2, -45,  31,   3, -11,  74, -11, -14,   1 },
     {  -26,  51,  -1,   5, -45,  30, -14, -11,  12,  27,  72,  -9,  -8, -54,   0,  27 },
     {   10,  -4, -45,   6,  15, -41, -42, -12,  21, -20,  65, -21,   2,  62, -28,  15 },
     {  -16,  28, -33,  -9,  -9,  51, -20,   6, -15,  10, -16, -35,  36,  51,  72,  -7 },
     {  -13, -20, -38, -83,   5,   9,  28,  74,   9,  10,  18, -10,  -7, -10, -18,  -5 },
     {  -18,  -7, -48,   0, -31, -12, -55,  -4, -47, -13, -31,  -9, -71, -22,  -4, -27 },
     {   -5, -28,  10,  -4,  -2, -32,  32,   1, -23, -51,  20, -28, -24, -15,  67,  61 },
     {   -4,  -8,  -1, -45, -11, -17, -16, -27,  -5,   2,  29,  97,   3,  10,  46, -19 },
     {  -11,  -2,  36,   5, -25,  10,  39,   3, -47,  45,  21,   3, -53,  69, -17,   6 },
     {    2,  -7,   8,  45,   2, -13,  22,  34,   3, -12,  48, -15,  -6, -10,  33, -92 },
     {    1,  -4,  19, -63,   1,  -6,  18, -82,   1,  -2,   6, -55,   0,  -7,  -6, -42 },
   },
   {
     {  112, -44,  -2,  -3,  40,  -2,  -8,   1,  -1,  13,  -1,   0,   1,   1,   5,   1 },
     {   29,  38, -22,   0, -54,  80, -21,   0, -40,  35,  17, -11,  -5,  -8,  22,  -2 },
     {   47,  22,  -3,   1, -75, -18,  42,   0,  19, -68,  -1,  14,   3,   2, -29,   2 },
     {    6,  16, -32,  10, -13, -17, -30,  15,  -5,  28, -82,  15, -25,  54, -35, -29 },
     {   17,  44,  17, -14, -26, -65,  -2,   7,  34,  47,  -7, -36,   3,  -6,  62,   9 },
     {   -7, -55,  53,  -7, -44, -26,  33,   5, -70,  30,  -6,   9, -28,   4,   3,   0 },
     {    2,   6,  11,  17,  -9,  -7, -33,  30,  12,  22,  18,  27, -18,  -3, -41, 100 },
     {  -13, -73, -10,   5, -53,   1, -66,   7,  32, -14,  -6, -34,  29, -10,   7,  -6 },
     {   -9, -17, -29, -17,   9,  24,  39,  40, -15, -28, -42, -27,  21,  40,  47,  58 },
     {   -2, -12,  15, -20,  -6,  21,  14, -44,  39,  13,  37, -40, -38,  81, -17,   7 },
     {    2,  -1,  67,  -7,  -6,  48,  22,  16,  47,  32, -33,  33,  56,   5, -10, -10 },
     {  -10, -37, -64,  12, -17,   6,  49, -16,  45,  43,   0,  44, -27, -27,  20,   3 },
     {    1,  -2,  10,  72,   2,   1,   3,  72,  12,  -8,  44,  13, -17,  37,  30, -33 },
     {   -1,  -1, -38,   8,  -8, -33,  12, -11, -28,  34,  42,  11,  87,  39, -23,   0 },
     {    2,   3,  19,  47,  -4,  -2, -27, -73,  -9, -24, -18,  48,   8,  28,  55,  24 },
     {    3,   1,   6,  86,   4,   6,  33, -24,  -2,  16, -28, -67,   8, -25, -29,  12 },
   }
 },
 {  //33
   {
     { -115,  27,  14,   2,  13, -39,   5,   1,  19,  -4, -13,   1,   1,   3,  -2,  -4 },
     {    0, -55,  15,   2,  87, -16, -42,   4, -11,  55,  -8,  -9,  -7,   1,  13,   0 },
     {   39, -13,   3,  -4, -11, -85,  24,  15,  24,  -7, -66,   9,  -6,  31,  -5, -21 },
     {   29,  95, -23, -12,  67,  -8,  28,  -4,  -5,  12,   4,   9, -13,   1,   9,   5 },
     {    1, -38,  15,   7,  51,  -2,  44, -14,   1, -65,  27,  40,  -3,  -2, -56,  15 },
     {  -19,  -5, -43,  10,  12,  50,   0,  -4, -47,  -8, -60,  24,  -3,  23, -18, -64 },
     {   -1,   7,  44,  -9, -16, -30,  22,  27, -98,  11,  -1,  12,   8, -45,  -3,   1 },
     {   -8, -16, -93,  17,  -8, -57, -11, -27, -35,  -4,  37, -12,   4,  -6,  -3,  11 },
     {    2,  23,  -1,  22, -20,  -1, -48,  20,  -4,  33,  -8,  39, -36,  21, -75,  47 },
     {  -14, -27, -11,   2, -19,  13,  77, -23,   9,  73,  10,  39, -29,  10,  16,  12 },
     {  -12, -14, -23, -11,  10,  21,  24,  51, -22, -19, -30, -22,  18,  45,  26,  80 },
     {    4,   4, -14,   2,   5,   4,   9,  -5,  22,  35, -22,  12, 106, -30, -36,  11 },
     {    4,  13,  38,   7,  -6, -10, -10, -49, -32,   6,  29,   7,  42,  91,   7,  -5 },
     {   -6,  -8, -11,-113,  -6,  -2, -28, -28,  -2,  -7,  -8,  39,  -2,  -3,   3,  13 },
     {    4,   5,  15,  34,  -1,   3, -10, -82, -10, -17, -60,   0, -13, -32,  23,  49 },
     {    2,   1,  -7,  34,   0,  -6, -29,  29,   9, -18,  10,  92,  18,  -7,  64,   1 },
   },
   {
     { -102, -22,  16,   4,  67,  -2, -19,  -3,  -8,  15,   6,  -1,   2,  -5,   0,   1 },
     {  -41,  61,  12,  -5, -39, -79,  10,  12,  38,  10, -32,  -7,  -4,   9,  10,  -4 },
     {   27,  81,  10, -11,  55, -18,  -2,   9, -70,  -7,  13,   3,  15, -12,  -6,   1 },
     {  -38,  11, -28,   1, -10,  28,  58,  -3, -19, -79, -31,  21,  17,  26, -18, -22 },
     {    0,  12, -78, -15,  31,   3,  64,  -2,  21,  50,   6,  12, -27, -28,   9,   8 },
     {  -35,  51,   6,   5, -45,  48,  -9, -15,  18,  -8,  71,   5,  -9, -38, -21,  20 },
     {   16,  28,  52,  18,  37,  43,  11,  -2,  41, -19, -46,  -2, -65, -16,  18,  -1 },
     {   12, -17,  11,  21,   7, -42,   0, -23,   5,  -3,   8,  53, -31, -29, -79, -45 },
     {   -7, -18,  19, -69, -12,  11,  10,  86,  -7,  -2,  -6, -20, -12, -37, -32, -15 },
     {   12,  12, -34, -35,  38, -10, -40,  17,  53, -40,  38,  -2, -23,  60, -16,   0 },
     {   11, -26,  42,   2,  17, -35,  58,   8,  16, -27,  32,  13,  14,   0,  -6,  80 },
     {   -9, -19, -37,  -6, -15, -34, -30, -18, -37, -52, -10, -20, -60, -44,  23,  35 },
     {   -5,  -5,  21, -60, -13,   3,  -8, -14, -11,   4,   7,  91, -11,   7,  56,  -7 },
     {    9, -11,  13, -16,  17, -21,  23, -28,  27, -34,  38, -37,  29, -42,  45, -65 },
     {    8,   8, -23, -13,  18,   6, -40,  -5,  43, -18, -47,  27,  66, -56, -13,  29 },
     {    0,   2,  18, -72,  -2,   7,  11, -79,  -4,  16, -23, -38,  -3,  16, -41,  16 },
   },
   {
     {   93, -79,  26,  -9,  19,  -9,  -4,   3, -11,  13,  -7,   2,  -3,   3,  -1,   0 },
     {   35,  12, -28,   6, -65,  85, -40,   7, -21,  13,  11, -10,   5, -10,  10,  -3 },
     {   45,  41, -83,  31,  63, -16, -19,   1,  -1,   4,   6,  -8,  -9,   7,   3,  -3 },
     {  -27,   5,   6,  -1,  15,  -3,  -5,   3, -76,  86, -37,   7, -20,  14,   6,  -8 },
     {  -37, -38,   0,  21,  22,   6, -84,  46,  39,   0, -40,   2,  13,  -1,   0, -14 },
     {   24,  13, -23,  17, -63, -36,  18,  24,  13,  -9, -66,  33, -17,  38, -34,  -8 },
     {   25,  44,  14, -51,  26,  24,   3, -21,  30,  18, -57,  34,  49, -35,  -1,   5 },
     {   -1,   9,  15, -34,  23,  44, -14, -29,  38,   7,   1,  -9, -58,  75, -28,   9 },
     {    7,   5,  -3, -24, -23, -29,  -2, -25,  28,  16, -17, -64,   0,  10,  58, -70 },
     {   23,  52,  40, -48,   4, -24, -34,  67, -32, -32,  19,  -3, -19,   4,   5, -12 },
     {  -12, -18, -11,  -5,  27,  36,  16, -16, -44, -60, -21,  21,   1,   3, -19, -81 },
     {   -2,  -1,  -7, -20,  -9, -24, -28, -25, -26,  -2,  -1, -61,  52,   9, -81,   8 },
     {   16,  28,  49,  63,  20,  31,  42,  41,  19,  22,   1, -36,  24,  12, -25, -29 },
     {  -12, -21, -32, -32,  14,  26,  43,  47, -15, -25, -38, -45,  22,  30,  33,  46 },
     {  -11, -23, -43, -51,  -3,   3,  29,  49,  30,  46,  43,  17,   2, -13, -36, -37 },
     {    0,  -1,  -2,  -2,   4,   8,  11,   5,   6,  -7, -36, -45, -74, -79, -30,   9 },
   }
 },
 {  //34
   {
     { -110,  48,   6,   2,  -9, -29,  19,  -1,  18, -17,  -8,   5,   1,   2,  -5,  -2 },
     {  -13, -56,  26,   0,  58, -70, -16,  15,   2,  30, -48,   3,  -5,  13,   1, -17 },
     {   43,  -2,  10,  -4, -16, -47,  70,  -3,  23, -60, -12,  35,  -4,  14, -36,   7 },
     {  -10, -36,   9,   8, -98,  -3, -11,  23, -37,   0, -45,   0,   2,  -2,  -2, -34 },
     {  -31, -72,  53,   1, -12,   5,  -4,  -4,   7, -24,  61, -10,   8, -16, -12,  48 },
     {   -1,  18,  50, -26,   4,   4,  61,  -8, -69,  54,   8,  25, -16, -10,  16,  13 },
     {  -27, -35, -20,  18,  38,  52,   9, -28, -32, -25,  -2,  48,  -8,  -7, -43, -52 },
     {    4, -10,   8, -28,   5, -18,  27, -39,   3, -26,  18, -48,  28, -40,  48, -69 },
     {    5,  10,  73, -37,   3,  59,  -6,   9,  60,   2, -37,  14,  12,  13,  -2, -22 },
     {   12,  40,  36, -13,  27, -11, -39,  51, -55, -53,   2, -19,  12, -27, -33,  -6 },
     {   17,  27,  29,  21, -23, -37, -41, -38,  25,  36,  37,  30, -27, -37, -39, -35 },
     {   -1,   3,  15, -13,  -3,   2,  -2, -24, -15,  -7,  18, -60, -62,  79, -26, -22 },
     {    3,   2,  17,  66,   7,   7,  25,  70,  16,  -2,  41,  10, -25,  13,  41, -41 },
     {   -4,  -7,  -8, -43,  -2,  -4, -33,  -4,  -1, -42,  -4,  44, -76, -15,  59,  10 },
     {    6,   8,  28,  69,   9,  17,  11, -37,   0, -18, -55, -38, -38, -41,   9,  30 },
     {   -7, -15, -30, -37,   1,  11,  33,  48,  27,  25,  -3, -36, -49, -58, -44,  -4 },
   },
   {
     { -105, -21,  21,   5,  56, -15, -25,  -1,   0,  19,   1,  -5,   2,  -3,   2,  -1 },
     {  -23,  53,  10,  -8, -43, -80,  15,  20,  30,   6, -49, -10,  -1,  13,  14, -12 },
     {   13,  88,   8, -15,  62,  -9,  12,  10, -58, -11,  15,  10,   6, -13,  -8,   5 },
     {  -41,  -4, -31,   5,  -5,  27,  64,  -8, -12, -67, -33,  33,   7,  17, -24, -34 },
     {  -11,  22, -69,  -3,  21,   7,  50, -23,  47,  51,  21,  13, -29, -15,  21,  24 },
     {  -39,  43,  29,  11, -54,  21,  -7, -14,  28, -31,  66,   6,   2, -11, -27,  39 },
     {  -13,   1, -11, -24,   0,  32,  16,  29,  -9, -22, -11, -55,  34,  38,  58,  61 },
     {   12,  17,  71,  24,  26,  54,  26,  17,  46,   2, -42,  -4, -40, -14,  11,  -2 },
     {   -9, -23,   1, -67, -11,   5,  25,  83,   1,  13,  16,  -3, -12, -41, -38,  -4 },
     {   23, -35,  22,   7,  35, -52,  29,   6,  17, -22,   7,  32, -10,  38, -34,  68 },
     {   -7, -17,  29,  42, -30,  -6,  55,  -6, -49,  46, -10,  12,  45, -44,   8,  25 },
     {   -5, -10, -25,  17,  -6, -20, -10, -16, -28, -42, -27, -45, -63, -62,  -5,  37 },
     {   -7, -10,  23, -56, -18,   0, -13, -22, -19, -12,  -8,  75, -30, -14,  65,  16 },
     {  -14,  20, -29,  26, -23,  32, -40,  37, -29,  37, -41,  38, -27,  36, -32,  37 },
     {    7,   9,  -9, -38,  14,  13, -28, -33,  36,  -3, -55,  11,  63, -42, -37,  31 },
     {   -6,   1,  27, -54, -15,  11,  25, -62, -28,  36, -11, -39, -30,  40, -40,   1 },
   },
   {
     { -104,  48,   2,   3,  32, -43,  10,  -1,  13,   0, -10,   1,   2,   2,  -1,  -2 },
     {   37,  17, -25,   1,  88, -24,  -1,  -5, -57,  49,   5,  -1,  -5,  -9,  10,   3 },
     {   37,  79, -58,   3, -36, -19,  52,  -8,  -2, -30,   1,   9,   4,   3,  -9,  -1 },
     {   30, -27,  16,  -6, -10, -78,  21,  13,  33,  26, -65,   7, -20,  32,   7, -14 },
     {  -15, -10,  20,  -4, -51, -24,  18,  12, -70,  26, -20,   2,  63, -45,  -3,  -7 },
     {    7,  49,  -4, -10, -36,   3, -50,  17,  30,  60,   3, -44,  -4, -17,  54,  10 },
     {  -22, -49, -66,  63, -18,  -7,  17, -46,  14,  36,  11, -23,   4,  -1,   5,  13 },
     {   -4,  -8, -49,  16,  -1, -17, -67,  32, -27, -45, -24, -28,  13,  22,  -2, -56 },
     {  -12,   7, -16,   1,   8,  63,  14, -21,   2,  12, -70,  44,  -8, -17,  44, -46 },
     {  -21,  -5,   3,  -1, -37, -11,   1,  -3, -61,  -5,  15,   6,-100,   2,  21,   0 },
     {   -9,  -8,   2, -17,   5,  18,  30,   2, -29, -21, -11, -34,  30,  73,  65,  40 },
     {   10, -23,   3,  -3,  20, -34,  19,  -4,  21, -55,  20, -17,   1, -70,  68, -12 },
     {  -20, -30, -53, -58,   5,  20,  40,  73,  12,  20,   8, -15, -14, -15, -16,  -3 },
     {    7,  15,  32,  88,  10,  22,  34,  68,  -1,   0,  -9, -24, -12,  -5,  -1,  -3 },
     {   -2,  -6, -29,  16,   0,  -9, -36,  34,   0, -19, -29,  60,   1, -20,  11,  86 },
     {   -3,  -5,  -3,  14,  -9, -14,   0,  30,   8,  22,  67,  72,  22,  34,  40, -39 },
   }
 }
};

const int8_t g_lfnst8x8[ 35 ][ 3 ][ 32 ][ 64 ] = {
 {  //0
   {
     {  115, -23, -18,  -2,  -4,  -1,  -2,   0, -35,  21,   3,   0,   1,   0,   0,   0, -24,  -1,   6,   0,   1,   0,   1,   0,   1,  -2,  -1,   1,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   18, -33,   2,   4,   1,   1,   0,   1, 105,  -1, -28,  -2,  -5,  -1,  -2,   0, -41,  31,   7,  -3,   1,  -1,   0,   0, -16,  -6,   7,   1,   1,   0,   0,   0,  -5,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {   19, 106, -34, -10,  -8,  -4,  -3,  -1,  29, -17,  13,  -3,   0,  -1,   0,   0, -33, -27,  10,   6,   3,   2,   1,   1,   2,  -2,  -2,   1,   0,   0,   0,   0,  -2,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -39, -14,   1,   5,   3,   2,   1,   1, -24,  42,   1,  -8,  -1,  -2,   0,  -1, -88, -14,  35,   2,   4,   1,   2,   0,  44, -27, -13,   5,  -1,   1,  -1,   0,   9,   7,  -5,  -2,   0,   0,   0,   0,   6,  -3,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0 },
     {    1, -19,   0,   5,   0,   1,   0,   0, -19,-104,  16,  21,   5,   5,   2,   2, -43,  24, -13,  -1,   2,   0,   1,   0,  30,  21,  -5,  -8,  -2,  -2,  -1,  -1,   0,   5,   1,  -1,  -1,   0,   0,   0,   3,   5,  -1,  -1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   1,   2,   0,  -1,   0,   0,   0,   0 },
     {   17,  17, 107, -35,  -8,  -7,  -1,  -3,   3,   4,  -7,  17,  -4,   1,  -1,   0, -20, -27, -33,  14,   7,   3,   1,   1,   0,   3,  -2,  -4,   1,   0,   0,   0,   2,   1,  -4,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -5,  31,  14,  -8,  -2,  -3,  -1,  -1, -35,   0, -17,   6,   3,   2,   1,   0, -20,  75,   9, -24,  -1,  -4,   0,  -2, -45, -37,  41,   8,   0,   2,   1,   0,  24, -27, -16,  10,   0,   1,   0,   0,   4,   5,  -2,  -2,   0,   0,   0,   0,   5,  -3,  -2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },
     {  -11, -26, -12,   4,   3,   2,   1,   1, -10, -22,   1,   3,   1,   1,   0,   0, -26, -63,  14,  18,   3,   4,   1,   1, -87,   8,  11,   1,   5,   1,   2,   0,  40,   4,  -9,  -4,  -2,  -1,  -1,   0,   3,   8,  -1,  -2,  -1,   0,   0,   0,   7,   2,  -1,  -1,   0,   0,   0,   0,   1,   2,  -1,   0,   0,   0,   0,   0 },
     {    0,  -8,  11,  -2,  -1,   0,   0,   0,  15,   9, 109, -18, -17,  -4,  -3,  -2,  -3,  27, -10,  20,  -2,  -2,  -1,   0, -22, -22, -29,   7,  10,   2,   2,   1,   4,  -7,  -6,  -2,   0,   1,   0,   0,   0,   2,  -7,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0 },
     {    7,  13,  22, 107, -16,  -9,  -9,  -2,   1,   6,   7, -11,  15,  -2,   1,  -1,  -9, -13, -33, -43,   6,   8,   4,   2,  -5,  -4,   2,   1,  -5,   1,   0,   0,   4,   2,   3,  -2,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   1,   1,   0,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {  -16,  -3, -17,  -8,   4,   1,   2,   1, -10,  27,  10,  -1,  -4,  -2,  -1,  -1, -34,   5, -41,   4,  12,   1,   2,   0, -20,  47,  31, -31,  -3,  -2,   0,  -1, -46, -36,  47,  17,  -4,   2,   0,   0,  24, -14, -21,   7,   2,   0,   0,   0,   6,   3,  -1,  -2,   0,   0,   0,   0,   4,  -2,  -2,   0,   0,   0,   0,   0 },
     {    1,   6,  -4,  -2,   1,   0,   1,   0,   7,  26,   8,  -2,  -3,  -2,  -1,   0,  -4,  20,  -7,  -5,   2,  -1,   0,   0,  20,  69,   0, -26,  -4,  -4,  -1,  -1,  86, -10, -10,   2,  -7,   0,  -2,   0, -35,  -5,   5,   4,   3,   1,   1,   0,  -7,  -8,   3,   2,   1,   0,   0,   0,  -3,  -3,   0,   1,   0,   0,   0,   0 },
     {   -4,  -5, -36, -16,   8,   3,   2,   1,   1,   5, -12,   9,  -3,   0,   0,   0,  -5,  -4, -87,  12,  21,   0,   3,   1,   9, -59,   6, -13,   2,   6,   0,   1,  32,  28,  12,  -5, -12,  -2,  -2,  -1, -11,  11,   9,  -1,   0,  -2,   0,   0,  -1,  -1,   5,   1,  -1,   0,   0,   0,  -2,   2,   2,   0,   0,   0,   0,   0 },
     {    1,  -3,   5, -18,  -3,   3,   1,   0,  -6, -17, -12,-106,   6,  14,   6,   3,  -8,  -3, -23,   0, -23,   3,   1,   1,   6,   9,  16,  44,  -3, -11,  -3,  -1,   7,   3,   8,   6,   6,  -2,  -2,   0,  -2,   0,   0,   6,   0,   0,   0,   0,  -1,   0,   1,   1,   0,   0,   0,   0,  -1,   0,   1,   3,   0,   0,   0,   0 },
     {    4, -16,  -4,  -6,   2,   2,   0,   1,  17,  -7,  24,  10,  -4,  -2,  -1,  -1,  11, -36,   2, -15,   3,   4,   1,   1,  34, -20,  56,   8, -26,   2,  -3,   0,   8, -41, -32,  52,   5,  -2,   0,   0,  -8,  23, -33, -15,  13,  -1,   1,   0,  -5,  10,   7,  -8,  -2,   1,   0,   0,  -2,   2,  -1,   0,   0,   0,   0,   0 },
     {   -6,  -8,  -4,  -7,-104,  35,   2,   4,   0,  -1,  -1,  -9,  20, -17,   3,  -1,   5,   5,   7,  14,  52, -13,  -6,  -1,   0,   1,   2,   2, -10,   9,  -1,   0,   0,  -1,  -1,  -3,  -1,  -1,   2,   0,  -1,   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   2,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {    0,  -2,  -6,   5,   0,   0,   0,   0,  -2,  -4, -29,  -6,   5,   2,   1,   1,  -2,  -3, -16,   9,   1,   0,   0,   0,   0,  -6, -68,   3,  18,   1,   2,   0,   4, -88,  -4,   4,   3,   7,   1,   2,   1,  38,  10,  -8,  -6,  -3,  -1,  -1,  -1,  17,  14,  -4,  -2,  -2,   0,  -1,   0,   0,   5,   1,  -1,   0,   0,   0 },
     {    4,   4,   0,  -1,   0,   1,   0,   0,   7,   7,  -2,  -2,   0,   0,   0,   0,   8,   8,   0,   0,  -2,  -1,   0,   0,   8,  12,  -2,  -2,   0,  -1,   0,   0,  29,  20,  -3, -10,  -4,  -2,  -1,   0, 103,  39, -28, -12,  -7,  -3,  -2,  -1, -28,   5,  15,   0,   1,   0,   0,   0, -15, -13,   4,   5,   1,   1,   0,   0 },
     {   -3,  -6,  -4, -41,  -7,   6,   3,   1,   2,   0,  10,   7,  -4,  -1,   0,   0,  -4, -11,  -1, -96,   9,  14,   3,   3, -16,   2, -42,  10, -16,   2,   2,   1,   3,  15,  24,  33,   1, -11,  -2,  -2,   5,   1,  12,  -1,   2,   0,  -1,   0,   0,  -2,  -2,   8,   0,  -1,   0,   0,  -1,  -1,   1,   0,   1,   0,   0,   0 },
     {   -2,   1,  -4,  13, -32,   4,   3,   0,  -7,  -4, -12,   4,-106,  16,  10,   2,  -1,  -5,  -3,  -9, -10, -20,   6,   1,   4,   2,   6,   2,  44, -12,  -6,   0,   0,   4,  -1,  10,  11,   4,  -3,  -1,   1,   2,   0,   0,  10,   0,  -1,   0,   1,  -1,   2,   0,   2,   1,  -1,   0,  -3,  -1,   0,   0,   3,   0,   0,   0 },
     {   -8,   0, -14,  -6,   0,   2,   1,   0,  -3,  16,   5,  11,   3,  -4,  -1,  -1, -14,   4, -30,  -6,   3,   3,   1,   1,  -7,  32,   7,  25,  -4,  -8,   0,  -2, -30,  13, -62, -16,  40,   1,   2,   0,  -1,  12,  33, -50, -14,   9,   0,   1,   8, -13,  29,  19, -15,  -3,   0,   0,   3,  -5,  -7,   5,   4,  -1,   0,   0 },
     {   -1,   0,  -5,   5,  -6,  -2,   1,   0,   2,   7,  -1,  44,  10, -10,  -3,  -1,  -1,   5,  -9,  17, -10,  -1,  -1,   0,   4,   8,  -3,  84,   6, -22,  -2,  -3,  23,  -5,  49,   1,  25,   1,  -6,  -1,  -3, -16, -25, -24,   2,  14,   2,   1,  -4,  -2, -18, -11,  -1,   1,   2,   1,   1,   2,   3,  -8,  -3,   1,   0,   0 },
     {    1,  -2,  -9,  -3,   1,   2,   0,   0,   1,  -5,  -8,   1,   1,   1,   0,   0,   1,  -4, -18,  -4,   4,   2,   0,   1,  -2,  -6, -17,   2,   4,   0,   1,   0,   9, -12, -49,   7,  15,   0,   1,   0,  29, -94, -12,  22,   0,   7,   0,   2, -27,  33,   4, -10,   0,  -2,   0,  -1,   3,  15,  10,  -7,  -2,  -1,   0,   0 },
     {    4,   7,   7,   6,  18,  99, -11,  -5,   2,   3,   3,   8,  14,   3,  12,  -2,  -5,  -6,  -7,  -5, -40, -56,   8,   6,  -3,  -4,  -4, -11, -10,  -4,  -7,   2,   1,   1,   2,   3,  12,   5,  -1,  -2,   0,   0,   0,   3,   0,   0,   1,   0,   3,  -1,  -1,   0,  -1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {   -3,  -3,   2,   2,  -1,   0,  -1,   0,  -4,  -2,   1,   2,   0,  -1,   0,   0,  -3,  -4,   3,   2,  -5,   0,   1,   0,  -8,  -7,   3,  -1,   1,  -1,   1,   0,  -8,  -8,  11,   3,   0,   0,  -1,   0, -12,   2,   6,   6,   2,   0,   0,   0,-101, -48,  26,  15,   8,   4,   2,   1,  42,   4, -20,  -3,  -1,   0,  -1,   0 },
     {    4,   6,   7,   8,  50,  20,  -9,  -2,  -2,  -3,  -1,  -9, -21,  14,   3,   0,   4,   2,   8,  -3,  84, -30,  -7,   0,   0,  14,  -1,  40,  -5,  14,  -6,  -2,  -4,  -3,  -3,  -7, -43,   2,  11,   0,   0,  -6,   1, -12,   6,  -6,   0,   2,  -5,   0,   1,   1,  -5,   2,   0,   0,   4,   2,  -1,  -2,   0,  -1,   0,   0 },
     {   -1,   8,   1,  11,   4,  -1,  -1,  -1,  -8,   2, -16,  -8,   1,   4,   0,   1,  -3,  17,   0,  27,   4,  -1,  -2,  -1, -16,   5, -32, -13, -13,   1,   6,   0,  -8,  32,  -7,  51,  10, -33,  -3,  -2, -10,   4, -21, -23,  55,   9,  -8,   0,   0,  -8,  20, -45, -20,  21,   3,   0,   5,  -5,  10,  14, -13,  -5,   2,   0 },
     {    1,   1,   2,   0,  13,  -8,  -4,   3,  -4,  -8,  -6, -10, -16, -93,  14,   6,  -3,  -2,  -4,  -4, -13, -15, -28,   5,   1,   5,   2,   5,   6,  68,  -8, -12,   3,   1,   5,   0,  17,   5,  16,  -2,   0,   0,   0,   1,   3,   2,  -1,   3,  -2,   1,  -1,  -3,  -5,   1,   1,   0,  -2,  -2,   2,   1,   0,   1,   0,   0 },
     {    2,   1,   2,  13,   4,  -2,  -1,   0,   3,   1,   6,   4,   1,  -2,   0,   1,   2,   2,   2,  27,   3,  -4,  -1,   0,   4,   3,  10,   6,  -5,  -2,  -1,  -1,   7,  -1,   6,  59,   3, -18,  -3,  -1,  31,   0,  89,   6, -13,   0,  -6,  -1,  -7,  10, -36,  -8,   9,   4,   3,   0,   0,  -5, -19, -10,   5,   2,   2,   1 },
     {    0,   1,  -1,   5,  -3,   3,   0,   2,  -2,   0,  -4,   7, -39,  -2,   6,   2,  -1,   2,  -4,  11, -13,  13,   0,   0,  -4,  -2,  -4,  12, -86,   0,  20,   0,   1, -18,   4, -42,   4, -38,   2,   7,   5,   2,  12,  20,  31,  -3, -19,  -1,   1,  11,   3,  19,   9,   4,  -2,  -4,  -2,  -1,  -1,  -6,   8,   2,   0,   0 },
     {   -1,   3,   5,  -3,  -2,   2,   1,   1,  -2,   1,   8,   4,  -1,  -1,   0,   0,   0,   3,   8,  -6,  -5,   0,   1,  -1,  -3,   1,  15,   2,   1,  -4,  -1,   0,   0,  12,  22, -12,  -5,   4,  -2,   0, -20,   5,  16,  -6, -12,  -1,   1,  -1, -22,  89,  41, -27,  -9,  -8,   0,  -3,  28, -42,  -8,  25,   1,   1,  -1,   0 },
     {   -2,  -3,  -1,  -7,   1,   1, -98,  26,  -1,  -2,  -1,  -3,  -4, -12,  -1,  -9,   3,   2,   4,   3,   7,  15,  65, -20,   1,   1,   2,   0,   8,  12,  -3,   8,   0,  -2,   2,  -3,   8, -13, -14,   4,  -2,   0,  -1,  -1,  -3,   3,   2,   0,  -2,   0,  -1,   0,  -2,   3,   5,  -1,  -7,  -2,   1,   1,   2,   0,   0,  -1 },
   },
   {
     {  -93,  56,   4,   0,   0,   0,   0,   0,  62, -19, -14,   0,  -1,   0,  -1,   0,   5, -15,   4,   2,   0,   0,   0,   0,  -2,   1,   3,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   61,  37, -43,  -1,  -5,   0,  -2,   0,  46, -52,  -3,  10,   0,   2,   0,   1, -57,  -3,  29,  -1,   2,   0,   1,   0,  -3,  15,  -2,  -5,   1,  -1,   0,   0,  -3,   0,  -1,   1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {  -22,  53, -38,   0,  -1,   0,  -1,   0, -76, -39,  39,  13,   4,   2,   2,   1,  15,  20,  11, -11,  -3,  -1,  -1,  -1,  35,  -3, -18,  -2,   0,   0,  -1,   0,   1,  -6,   3,   3,   0,   0,   0,   0,   4,   0,  -1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {   37,  67, -20, -21,  -3,  -4,  -1,  -1, -16,  43, -32,   1,   2,   0,   0,   0,  41, -56, -11,  20,  -2,   2,  -1,   1, -17, -16,  23,   1,  -2,   1,   0,   0,  -2,   2,   4,  -4,   0,   0,   0,   0,  -3,  -1,   1,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0 },
     {    9, -21, -51,  30,   0,   5,  -1,   2,  32, -26,  27,   6,  -7,   0,  -1,   0,  75,  16, -13, -22,  -4,  -1,  -1,  -1, -50, -15,   6,   5,   7,   1,   1,   0, -14,  12,  10,  -1,   0,  -1,   0,   0,  -2,   2,  -3,  -1,   0,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 },
     {   21,  38,  13, -11,  -7,  -2,  -2,  -1,  31,  26,  27, -32,  -1,  -3,   0,  -1, -12,  51, -56, -11,  11,  -2,   1,   0,  19, -44, -24,  30,   1,   1,   0,   1, -13, -13,  20,   5,  -5,   1,   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,  -3,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 },
     {   24,  15,  76, -46,  -8,  -5,   0,  -2,  -3, -65,   0,   9,  12,   4,   1,   1,  37,   7, -18,   7,  -1,  -2,  -1,   0, -22,  23,  -5,  -5,   0,  -1,   1,   0, -12,   5,  -4,   4,   1,   0,   0,   0,   1,   3,  -2,   0,  -1,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {    5, -17, -15, -17,  12,  -1,   3,   0,  14,  -4, -71,  34,  10,   0,   1,   0,  30,  40,  17,  10, -20,  -3,  -3,  -1,  51, -22,  -8, -10,  -6,   4,  -1,   1, -35, -22,  19,   6,   4,   1,   0,   1,  -2,  10,   7,  -6,  -1,  -1,   0,   0,  -4,   0,   0,   0,   1,   0,   0,   0,  -1,   1,   2,  -1,   0,   0,   0,   0 },
     {  -10, -19, -12,  12,  -2,   3,   0,   1, -15, -46, -10, -10,  14,   3,   2,   1, -22, -16, -44,  43,  12,  -3,   1,   0,  17, -30,  52,  23, -24,   1,  -2,   0, -13,  30,  26, -33,  -4,   3,   0,   0,   6,   8, -13,  -3,   6,  -1,   0,   0,  -1,   2,   0,   1,   0,  -1,   0,   0,   2,   1,  -1,   0,   0,   0,   0,   0 },
     {   18,  -5,  19,  43, -19,   0,  -4,   0,  21,  -4,  28, -16, -16,   4,  -1,   0,  28, -55,  -1,  -3,  11,   5,   0,   1,  70,  21, -16, -10,  -1,  -5,  -2,  -1, -43,  -4,  -1,  -1,   5,   0,   2,   0, -10,   8,   3,   2,   1,  -1,   0,   0,  -4,   2,  -2,  -2,   0,   0,   0,   0,  -2,   1,   0,   0,   1,   0,   0,   0 },
     {    9,  27,  32,  81, -44,  -7,  -7,  -2, -20,  -6, -48,   3,  -3,  10,   2,   2,  -3,  29,   0, -30,  12,   1,   0,   0, -14,  -5,  23,  -5,   1,  -2,   0,   0,  11, -11,  -4,  -3,   3,   1,   0,   0,   4,  -1,   2,  -1,   0,  -1,   0,   0,   0,  -1,  -1,  -2,   1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0 },
     {    1,  -3, -33,   2,  -1,   3,   0,   1,   4,  -5, -20, -47,  23,   4,   3,   1,  22,  23, -31,  30,  20, -13,   0,  -2,  10,  64,   0,   6, -23,  -8,   0,  -2,  27, -35, -25,   4,  -3,   7,   0,   1, -15, -18,  20,  11,  -1,   1,  -1,   0,  -1,   2,   5,  -4,  -1,   0,   0,   0,  -2,  -1,   1,   1,   0,   0,   0,   0 },
     {    6,  16,   8,   1,  -5,   0,  -1,   0,  15,  22,  32,  -6,   4,  -6,  -1,  -1,   8,  41,  18,  26, -27,  -7,  -1,  -2,  12,  -5,  39, -50, -22,  13,  -1,   1, -11,  21, -41, -34,  32,   3,   0,   0,   5, -14, -20,  25,   7,  -6,   1,   0,  -2,  -3,   4,   3,  -4,   0,   1,   0,   1,  -2,  -2,   0,   0,   1,   0,   0 },
     {   12,   0,   7,   8,  15,  -5,   0,  -1,  18,  -6,   7,  60, -18,  -9,  -2,  -1,   8,  -5, -40,  -6, -17,  10,   2,   1,  34,  -4,   4,  -3,   2,   6,  -3,   0,  75,   7,  -8, -13,  -7,  -2,  -3,   0, -36, -19,   5,  -2,   6,   2,   1,   1, -11,   6,   8,  -1,   0,  -1,   0,   0,  -3,   1,  -1,  -2,   1,   0,   0,   0 },
     {   -6,  11,  -8,  10,  31, -10,   1,  -2,  -7,  19,   1,  42, -19, -13,   1,  -1, -21,   8, -54, -10,  -6,  12,   4,   1, -12,  61,  15, -18,   3,  -1,  -3,  -1, -49, -10,  20,  -2,  -1,   0,   1,   1,  29,  -8,  -9,   0,   1,   2,   0,   0,  11,  -9,   0,   2,  -2,   0,  -1,   0,   1,  -3,   2,  -1,   0,   1,   0,   0 },
     {   -1,   0,  10,  25,  -5,  -2,  -2,   0,  -1,  -2,  31,  19,   7, -10,  -2,  -2, -10, -11,  -2,  50, -26, -11,   2,  -1, -29, -17,  -9, -18, -26,  19,   4,   2,   3, -66,  20,   8,  13,  12,  -4,   2,  -9,  28,  31, -15,  -2,  -5,  -1,   0,   8,  15, -14, -13,   4,   0,   1,   0,   0,   0,  -3,   2,   2,  -1,   0,   0 },
     {   11,  14,  19,  32,  92, -38,  -9,  -5,  -3, -13,  -6, -39,   4,  -7,   8,   2,   6,   0,  19,  -8, -35,  12,   5,   0,   0,  -3,   4,  26,  -8,   1,  -3,  -1,   2,  -1,  -6,  -5,  -7,   3,   0,   1,  -3,   6,  -1,  -1,   2,   0,   0,   0,   0,   0,  -3,  -1,  -1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 },
     {   -4,  -7, -11,   7,   8,  -1,  -2,   0,  -8, -21, -17, -16,  11,  -3,   3,   0, -15, -17, -41,  -1, -17,  12,   4,   2,  -2, -31, -12, -31,  33,  17,  -2,   1, -10,  12, -33,  45,  35, -20,  -2,  -1,   9, -14,  28,  33, -30,  -8,   2,   0,  -4,   7,  16, -15,  -7,   6,   0,   0,   2,   2,   0,  -1,   1,   0,  -1,   0 },
     {   -4,  -7,  -3, -34,   6,   3,   2,   0,  -4, -12,   6, -14, -45,  14,   2,   2,  -4,   0,   1, -35,  14,  21,  -3,   2,   5, -12,  66,   0,  12,  -9,  -8,   0,  -8, -56, -18,   2,  11,  -1,   4,   2, -26,  18,   6,  11,  11,  -5,   2,  -1,   8,  23,  -5,  -3,  -5,  -4,   0,  -1,  -1,  -1,  -6,   2,   2,   0,   0,   0 },
     {    6,  -7,  -1,  -2,   0,   3,   0,   0,  10,  -6,   5,  -1, -11,   3,   0,   0,  13, -10,  -2,   5,  -1,   5,  -2,   1,  14,  -6,  10,   4,  -2,   0,  -2,   0,  34, -35, -13,  15,   1,   1,  -1,   1,  92,  15, -32,  -8,  -6,  -2,  -2,   0, -43, -18,  10,   4,   4,   1,   1,   0, -14,   9,  10,  -1,   1,  -1,   0,   0 },
     {    1,  -1,  -2,  -8, -10,  16,  -1,   3,   6,   5,   7, -14,  72, -25, -10,  -1,  -1, -11,  -7, -53, -27,  -3,  15,   3,  14,  16,  34, -14,  -9,  22,   0,  -3,  13,  -2,  25,  24,   5, -12,  -9,  -1, -11,  19, -19, -18,  -5,  -1,   5,   1,   0,  -1, -19,   9,   8,   0,   1,  -1,  -2,  -3,   1,   6,  -4,   0,   0,   0 },
     {    1,  -1,   8,   5,  13,   1,  -1,   0,   5,   5,  14,  43,  46, -19, -11,  -2,  -2, -12,   7, -22,  31, -23,   1,   1,   0, -20,  14,  29, -33, -19,  12,  -1, -25, -14, -45,  11, -21,  16,   9,  -2,  11, -21,  13,  18,  -4,  10,  -5,   0,   1,  17,  23, -14, -10,  -1,  -2,   1,   0,   4,  -7,  -8,   2,   2,   1,   0 },
     {   -2,  -6,  -6,  -4,   6,  -1,   0,   0,  -5,  -8, -18,  -2,   7,   5,  -3,   0,  -6, -21, -15, -31,   8,  -8,   4,   2, -12,  -9, -42,  -7, -22,  18,  12,   0,   4, -20,   2, -32,  37,  30, -10,   0,  -5,  11, -26,  35,  44, -26,  -7,   1,   3,  -7,  14,  34, -25, -13,   5,   0,  -1,   2,  11,  -8,  -9,   5,   1,   0 },
     {    3,   6,   0,   6,  39,  -5,  -6,  -2,   4,   6,   2,  11,   9,  14,  -7,   0,   2,   8, -10,   3,  61, -22, -15,   2,   3,  -8,   0, -46,   0, -24,  11,   5,  17,  18,  -5,  12,   1,  -4,  10,  -3,  -8,  61,  -3,   3, -18,  -9,   1,  -4,   5, -19, -29,   2,  -6,   8,   2,   1,  -2, -11,   7,  11,  -1,   1,  -1,   0 },
     {    5,   9,  10,   9,  26,  98, -41,  -5,  -5,  -2,  -8,  -2, -20,   2, -17,   8,  -1,   1,  -2,   3, -16, -44,  20,   6,   1,   4,   1,  12,  18,  -4,   7,  -5,  -2,  -1,  -5,  -7,   3,  -2,  -1,   0,   1,   2,   3,  -1,   1,  -2,   1,   0,   2,  -1,   1,   0,  -3,  -3,   2,   0,   1,  -1,   1,  -1,   0,   1,   0,   0 },
     {   -3,   0, -10,  -1, -22,  -2,   8,   0,  -1,   5,  -5,  15,  -2, -12,   3,  -2,  -3,   4, -14,   1, -26,  14,   8,  -1,   1,  11,  -8,  50,   0,   3,  -7,  -3,  -9,  15, -65, -10,  18,  -2,   0,   1,  -3,  64,  14, -23,  -1,  -4,  -1,  -1,  17, -28,  -6,   6,   0,   1,  -2,   1,  -1, -16,   7,   4,  -1,   2,   0,   1 },
     {    0,   0,  -3, -15,   1,  -5,   3,   0,  -1,   0,  -5,  -6, -35,  18,   3,   0,   0,  -1,  -3, -32,  -2, -20,   7,   3,   5,   5,   6,  -2, -52,  22,  25,  -6,   3,  34,   5,  15,  17,  40, -23,  -8,  17,  -7,  40, -17,  -2, -14, -18,   6, -11,  -2, -17, -34,  15,   3,   6,   3,  -2,  -2, -11,  15,  15,  -5,   0,  -1 },
     {   -5,   3,   0,   0,   3,   0,  -2,  -1,  -7,   6,   1,  -2,   3,  -3,   0,   0,  -7,   4,  -2,  -3,   3,   0,   0,   0, -13,   9,   2,  -8,   1,  -4,   0,   0, -16,   3,   0,  -7,   2,   1,   3,   0, -29,  25,  11, -14,   2,   1,   1,  -1, -94,  -6,  38,   6,   6,   1,   1,   0,  50,  19, -20,  -7,  -2,  -2,  -1,   0 },
     {    0,   1,   1,   4,   4,  -1,  16,  -2,   4,   9,  10,  11,  25,  92, -29, -11,  -6,  -3,  -7,  -3, -38,  -4, -35,  13,   0,   7,   3,  18,  -2, -33,   9,  10,  -1,  -3,   6,   2,  24, -10,   4,  -3,  -2,  -7,  -4,  -5,  -5, -10,   6,   2,   1,   2,   6,  -1,  -2,   1,   2,  -1,   2,   0,  -2,  -2,  -1,  -1,   0,   0 },
     {    1,   3,   6,   0,  -2,  -4,   1,   1,   2,   7,   5,  13,  -7, -10,   2,   1,   7,   6,  20,   6,  17,  -5,   7,  -4,   3,  19,  10,  39,  -8,  18, -11,  -5,   9,   2,  33,   2,  30, -25, -19,   4,  -6,  11,  -8,  32, -39, -38,  19,   2,   0,  -6,  18, -26, -43,  31,   9,  -3,   0,   3,  -8, -20,  18,  11,  -7,  -1 },
     {    2,   3,   0,  -4,   2,  11,  -1,   1,   4,   5,  -2,  -1,  28,  -6,  -3,  -1,   4,   2,   1, -11,   5,  27,  -7,  -1,   3,   9, -13,  -4,  23,  -9, -16,   5,  10,   8,  10, -57,   5,  -2,   5,   6,  40,  17,  42,  10,   8, -13,  -3,  -3, -11,  52, -17, -23,  -9, -14,   5,  -2,  -5, -16, -33,  12,   7,   4,   3,  -1 },
     {    1,   2,   4,   3,   4,  45, -18,   0,   1,   1,   3,   1,   5,  -5,  21,   0,   0,   3,   4,  -1,  -1,  68, -25, -12,  -6,  -4,  -4,  -5, -51, -24, -27,  16,   4,  10,  12,  27,   4,  -4,  18,   4,  -3,  -4,  13,   9,  19,  -4, -13,  -5,   2, -30,   1,  -8, -11,  -7,   7,   5,   2,   5,  12,  -6,   5,   4,   0,  -1 },
   },
   {
     {  108, -53, -22,   6,  -1,   0,  -1,   0, -28,  15,   1,  -1,   0,   0,   0,   0, -14,   6,   4,  -1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   40,  93, -66, -15,  -2,  -2,  -2,  -1,  24, -18,   1,   1,   0,   0,   0,   0, -16, -12,  13,   3,   0,   1,   0,   0,  -1,  -2,   2,   0,   0,   0,   0,   0,  -4,  -3,   3,   1,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    1, -41, -14,  26,   1,   2,   0,   1, 101, -34, -34,   8,  -2,   0,  -2,   0, -25,  18,   6,  -6,   1,   0,   1,   0, -11,   5,   5,  -2,   0,   0,   0,   0,  -7,   4,   2,  -2,   0,   0,   0,   0,  -3,   2,   2,  -1,   0,   0,   0,   0,  -2,   1,   1,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {  -41, -26, -83,  63,  12,   1,   0,   2, -37, -18,  25,  -2,   1,   1,   1,   0,   2,   7,  19, -14,  -3,   0,   0,  -1,   8,   2,  -3,  -2,   0,   0,   0,   0,   4,   2,   3,  -3,  -1,   0,   0,   0,   2,   1,   0,  -1,   0,   0,   0,   0,   1,   1,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   14,   1,  32,  -5, -12,   1,  -1,   0, -15, -97,  59,  26,  -8,   4,   0,   1, -18,  13,  -8,   2,   2,  -1,   0,   0,   6,  15, -13,  -4,   2,  -1,   0,   0,   0,   5,  -4,  -1,   1,   0,   0,   0,   1,   4,  -3,  -1,   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1,  -1,  -1,   0,   0,   0,   0 },
     {  -14,   6,  17,  -2,  -2,   0,   0,   0,  -1,  32,  24, -24,  -4,  -1,   0,  -1,-100,  27,  39, -10,   2,   0,   2,  -1,  26, -15, -11,   7,   0,   0,   0,   0,   8,  -4,  -5,   2,   0,   0,   0,   0,   5,  -3,  -3,   2,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   2,  -1,  -1,   1,   0,   0,   0,   0 },
     {   19,  28,  26,  76, -58,  -1,  -4,  -1,  24,  27,  40, -23, -11,  -1,   0,  -1,  23,   6, -21, -20,  15,   0,   0,   0,  -7,  -6,  -5,   3,   3,   0,   0,   0,  -3,  -3,  -2,  -3,   3,   0,   0,   0,  -2,  -2,  -2,   0,   1,   0,   0,   0,  -1,  -1,  -1,  -1,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 },
     {   -2, -23, -21, -42,  31,   4,   3,   1,  40,  14,  74, -54, -24,   7,  -1,  -1,  28,   1,  -5,  18, -12,  -1,  -2,   0, -10,  -2, -15,  13,   5,  -2,   0,   0,  -4,   0,  -2,   4,   0,  -1,   0,   0,  -4,   0,  -3,   3,   1,   0,   0,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,   0,   0 },
     {    0, -12,  12,  21,  -8,  -1,  -1,   0,  10,   5,  12,  13, -13,  -1,  -1,   0, -15,-103,  43,  38,  -8,   5,  -1,   2,  -2,  15, -11,  -5,   4,   0,   0,   0,   1,  13,  -6,  -7,   2,   0,   0,   0,  -1,   3,  -2,  -2,   1,   0,   0,   0,   0,   2,  -1,  -2,   1,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0 },
     {   14,  20,  26,  32,  79, -56,  -8,  -3,  10,  18,  19,  39,  -7, -11,  -3,  -1,   8,  13,  11, -16, -30,  15,   2,   0,  -9,   2,   0, -11,   0,   4,   1,   0,   0,  -4,  -3,  -2,  -3,   3,   1,   0,  -1,  -1,  -2,  -2,   0,   1,   0,   0,   1,  -1,  -1,  -1,  -1,   1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0 },
     {   -6,   2,   8,  -6, -11,   4,   1,   0, -15,  -6,   8, -20,   4,   4,   0,   1,  -9,   6,  37, -17,  -9,   3,   0,   0, -96,  23,  58,  -9,  -8,   1,   2,  -1,  -2,  -3,   1,   3,   0,  -1,   0,   0,   8,  -1,  -7,   3,   1,  -1,   0,   0,   8,  -2,  -5,   2,   1,   0,   0,   0,   2,   0,  -2,   1,   0,   0,   0,   0 },
     {   10,   6,  18,  22,  31, -22,  -6,  -2,  -7, -36, -18, -78,  62,   9,  -2,   3,  -7, -27, -16,  -7,  -4,  13,   1,   2,  11,   1,   1,  23, -18,  -1,   1,  -1,  -1,   5,   0,   2,  -4,   0,   1,   0,   1,   3,   1,   4,  -3,  -1,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   1,   1,   1,  -1,   0,   0,   0 },
     {   -9,   0, -13,  14,  20, -11,  -2,  -1,  -5,  13,  10,   6,  -3,  -6,   0,  -1, -48,  -9, -81,  52,  34, -14,   3,   1, -31,   6,  19, -13,   4,   2,   0,   0,   5,  -2,  15, -10,  -7,   4,   0,   0,   4,  -1,   0,  -2,  -1,   1,   0,   0,   4,  -1,   1,  -1,  -2,   1,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0 },
     {   -2,   0,  -2,  -3,   0, -36,  10,   2, -14, -14, -32, -28, -83,  39,  19,   1, -11, -15, -19, -34,   2,  25,   2,   0,  -1,  13,  -6,   2,  31, -10,  -6,   0,   2,  -1,   4,   7,   3,  -2,  -2,   0,   0,   1,   3,   3,   4,  -2,  -1,   0,   0,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   1,   1,   1,  -1,   0,   0 },
     {   -2,   4,  -4,  -9,   3,   6,   0,   0,   3,  17,  -2,  -5,  27, -10,  -3,   0,  -3,   4,  -2, -26,  21,  -1,  -1,   0, -11,  89, -61, -34,  20,  -5,   3,  -2,   1,   8,  -2,  -3,  -1,  -1,   0,   0,   3, -12,   7,   5,  -6,   2,   0,   0,   1,  -5,   3,   3,  -2,   0,   0,   0,   0,  -2,   1,   1,  -1,   0,   0,   0 },
     {    8,  16,  16,  24,  43,  86, -36, -20,  -4,   0,  -9,  -2, -18,  42,   6,  -8,  -3,  -1,  -6,  -5, -20, -30,  17,   7,  -1,   5,   0,   0,   7, -15,  -1,   3,   0,   1,   0,  -2,  -2,  -6,   1,   1,   1,   0,   1,   0,   1,  -3,   0,   1,   0,   0,   0,  -1,  -1,  -2,   1,   0,   1,   0,   0,   0,   0,  -1,   0,   0 },
     {   -4,   2,   4,  -2,   0,  -2,  -1,   0,  -5,   3,   2,  -2,  -3,   0,   1,   0,  -2,   3,   4,  -1,  -2,   0,   0,   0,  10,  -3,   3,  -5,   0,   1,   0,   0,-100,  39,  50, -22,  -2,   1,   3,  -1,  33, -16, -13,   9,  -1,  -1,  -1,   0,   0,   1,  -3,   1,   1,   0,   0,   0,   3,  -2,  -1,   1,   0,   0,   0,   0 },
     {   -3, -11,  -8, -20,  11,  25,  -1,  -3,  11,   9,  25,  25,  10,  -6,  -7,  -3, -13, -45, -29, -78,  28,  31,   2,   4,   2, -21,  15,  -7, -26,   8,   4,   2,  -1,   6,   6,  24,  -8, -11,   0,   0,   0,   3,  -4,  -1,   1,  -1,   0,   0,   0,   3,   0,   3,  -1,  -2,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 },
     {   -3,  -1,  -1,   1,   4,   3,  -2,  -2,  -9,  -7, -10,  13,  14,  -6,  -1,  -1,  -6,  -5,   0,  -7,  10,   0,   1,   0, -62, -54, -69,  42,  36,  -3,   3,   3, -17,  -9,  -8,   1,   7,   1,   0,   1,   9,   6,   7,  -6,  -6,   1,   0,   0,   5,   4,   4,  -3,  -3,   0,   0,   0,   2,   1,   2,  -1,  -2,   0,   0,   0 },
     {    5,   2,  11,   2,  27,  -2, -13,  -1,  -1, -12,  -4, -16, -17, -28,  19,   7,  15,  -4,  32, -10,  74, -48, -24,  12,   7, -10,  12,  -7,  19,  30, -23,  -2,  -1,  -1,  -3,   6, -30,  17,   8,  -5,  -4,   4,  -2,   2,  -1,  -1,   2,   0,  -1,   0,  -2,   0,  -3,   2,   2,  -1,   0,   1,   0,   1,  -1,   0,   0,   0 },
     {    1,   1,   2,   2,   7, -14,  28,  -7,   4,   9,   8,  19,  18,  83, -36, -17,   8,   8,  20,  11,  50,   3, -32,  -1,   1,  -1,   4,   7,  -6, -35,  12,   7,  -1,   0,  -1,  -3, -14,  -1,   4,   2,   1,  -1,  -3,  -3,  -2,  -4,   2,   2,   1,  -1,  -1,   0,  -2,   0,   0,   0,  -1,   0,  -1,   0,  -1,  -2,   1,   0 },
     {    4,   4,   3,   9,   9,  15,  91, -56,   2,   0,   2,   0,   8, -11,  50, -19,  -3,   0,  -5,  -2, -14,  -6, -19,  18,  -1,   0,  -1,   0,  -3,   4, -16,   7,   1,   7,  -3,  -4,   3,  -1,  -8,   3,   2,  -1,   0,   0,  -1,   1,  -4,   1,   1,   0,   0,  -1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -1,   1 },
     {    3,   4,   0,  -6,  -2,  -1,  -4,   3,   1,   0,   0,   1,  -3,   0,  -1,   2,   3,   6,   1, -11,  -1,   2,   3,   0, -11, -19,  -5,  10,   1,   1,   2,   0,  55,  86,  16, -55, -21,   2,  -1,  -2, -10, -27,   9,  19,  -3,  -1,   0,   0,   0,  -4,  -6,   4,   4,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    1,  -2,   1,  -2,  -6,   2,   5,   0,   0,  -6,  -2, -17,  11,  30,  -2,  -7,   1,   0,   1,   6, -15,  13,   4,  -3,  -6, -42, -20, -92,   8,  45,   5,   4,   0, -15,  -2, -14, -18,   8,   7,   1,  -5,   9,   5,  17,  -6, -15,   2,   1,   1,   3,   0,   4,   2,  -4,  -3,   0,   1,   2,   0,   3,   0,  -2,   0,   0 },
     {    0,   1,  -6,   0,   6,  -3,  -6,   2,   0,   1,  -2,  -3,   2,   2,  -2,  -1,   0,   4,  -9,   5,  11,   1,  -4,   0,   4, -10,  17, -19,  15,   2,  -5,   2, -15,  46, -92,  11,  45, -18,   2,   0,  31, -19,  -7,   5,  -3,   2,  -1,   1,  -5,   0,  10,  -3,  -6,   2,   1,   0,   1,  -1,   1,   1,  -1,   0,   0,   0 },
     {   -5,   4,   1,  -2,   1,  -1,   0,  -1,  -8,   5,   3,  -3,   1,   2,  -1,  -2,  -8,   5,   1,  -3,   4,  -1,   1,   0,  -5,   3,   6,  -5,  -2,   1,   1,   1, -36,  31, -11, -10,  16,  -4,   1,  -1, -86,  45,  46, -37,   3,   3,   2,  -1,  -4,  -4,  11,  -3,  -4,   2,   0,   0,   3,  -5,  -1,   3,  -1,   0,   0,   0 },
     {    0,   6,   0,   8,   2,   4,  -2,  -3,  -6,   0, -13,  -1, -27,   1,  12,   0,   0,  13,   1,  21,   9,  12, -12,  -5, -14,  11, -33,  15, -69,  39,  31, -14,  -6,  16, -16,  22, -27, -31,  30,   0,   4,   0,  -3,  -1,  29, -18,  -9,   7,   3,  -2,   1,  -6,   7,   2,  -5,   1,  -1,   0,   2,  -2,   3,  -1,  -1,   0 },
     {   -2,  -3,  -4,  -6,  -7, -30,  18,   1,   0,  -1,  -2,  -1,  -8,   9, -17,   7,  -3, -14,  -7, -24, -13, -83,  41,  22,  -5,  -8, -15, -11, -38, -14,  28,   6,  -4,   5, -11,   7,   4,  28, -14,  -6,   2,  -2,  -3,   2,  12,   3,  -3,  -3,   1,   1,   2,   0,   3,   4,  -3,  -2,   0,   1,   0,   0,   2,   1,  -1,   0 },
     {   -1,  -3,  -2,  -5,  -5, -15, -19,  31,   5,   5,   7,  10,  18,  21,  75, -54,   4,   1,   7,   4,  15,  -2,  45, -29,  -3,   1,  -4,   3, -17, -11, -27,  25,  -2,   3,  -4,   4,  -9,   0, -10,   6,   1,   1,  -1,  -1,   1,  -3,  -4,   2,   0,   2,   0,  -1,   1,   0,  -1,   0,   0,   0,   0,  -1,  -1,   0,  -1,   1 },
     {    3,   3,  -2,  -4,  -1,   0,   0,   1,   4,   6,  -3,  -6,   0,   0,  -2,   1,   4,   5,  -1,  -5,  -2,   1,   3,   0,   4,   5,  -3,  -7,   4,   3,   1,   0,  18,  23,   2, -21,  -6,   4,  -1,  -2,  50,  82, -37, -60,   3,   4,   1,  -1,  14,  11,  -4, -10,  -3,   1,   0,   0,   4,  -1,   1,   2,  -2,   1,   1,   0 },
     {    0,   2,   2,   6,  -8,  -9,   7,   3,   0,   1,   1,  -1,  11,  -6,  -9,   2,  -1,   4,   0,  11, -19, -17,   5,   2,   0,  -8,   3, -16,  36, -23,  -7,   9,   7,  39,  22,  83, -39, -42,   8,  -1,  -2,  11,   5,   5,  -7,  -3,   2,  -1,  -1,  -1,  -5, -15,   9,   8,  -3,   0,  -1,   0,   1,   0,  -1,   2,   1,  -1 },
     {   -5,  -8,  -8, -10, -16, -25, -47, -89,  -1,  -3,  -2,  -4,  -3, -11, -13, -38,   4,   5,   7,   7,  12,  15,  24,  37,   3,   1,   1,  -1,   2,  -2,  10,  25,   2,  -2,   1,  -3,  -1,  -8,   2,   7,  -2,   0,  -4,   2,   0,   1,   0,   3,   4,  -2,  -4,   1,  -1,   3,   1,   0,  -1,   0,   0,   0,   0,   1,   1,   0 },
   }
 },
 {  //1
   {
     {  124, -16, -14,  -4,  -4,  -2,  -2,  -1, -13,   8,   0,   0,   0,   0,   0,   0, -15,   1,   3,   0,   1,   0,   0,   0,  -5,   0,   1,   0,   0,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    1, -68,   8,   4,   4,   2,   1,   1, 102,  -6, -22,  -2,  -5,  -1,  -2,   0, -12,  20,  -1,  -1,  -1,  -1,   0,   0, -10,   2,   2,   0,   0,   0,   0,   0,  -6,   3,   1,   0,   0,   0,   0,   0,  -4,   1,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 },
     {    9,  94, -23,  -8,  -6,  -4,  -2,  -1,  64,  35,  -5,  -8,  -5,  -3,  -2,  -1, -29, -11,  11,   1,   2,   1,   1,   0,  -9, -11,   2,   2,   1,   0,   0,   0,  -4,  -6,   1,   1,   1,   0,   0,   0,  -3,  -4,   0,   1,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 },
     {  -20, -39, -20,  10,   3,   3,   1,   1, -26, 101, -14, -16,  -2,  -5,   0,  -2, -34,   7,  26,  -4,   1,  -1,   1,   0,   9, -21,   3,   3,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,   3,  -5,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 },
     {    5,  15, -33,   4,  -1,   1,   0,   0,  -1,  20, -39,  -5,   0,  -1,   1,   0,  84,  66,  -8, -20,  -7,  -6,  -3,  -2, -21,   3,  21,   0,   1,   0,   0,   0,  -9, -11,   4,   2,   1,   0,   0,   0,  -2,  -2,   3,   1,   0,   0,   0,   0,  -2,  -3,   2,   0,   0,   0,   0,   0,  -2,  -2,   2,   1,   0,   0,   0,   0 },
     {   13,  -4,  62, -21,  -6,  -3,   0,  -1,  20,  52,  50, -13, -12,  -5,  -3,  -2,  65, -30, -14,  13,  -5,   2,  -1,   1,   1,  -7, -17,   4,   1,   1,   0,   0,  -6,  -1,  -5,   2,   1,   0,   0,   0,  -3,  -2,  -5,   1,   1,   0,   0,   0,  -3,   0,  -2,   1,   1,   0,   0,   0,  -2,  -1,  -1,   0,   1,   0,   0,   0 },
     {    4,  19,  90, -13,  -9,  -4,  -2,  -2, -15,   6, -20,  13,   0,   1,   0,   0, -40,  60,  -9, -12,   3,  -3,   1,  -1, -24, -12,  20,   0,   1,   1,   1,   0,   2, -11,  -5,   3,   1,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   3,  -2,  -4,   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {   -3,  -6, -12, -15,   5,   1,   2,   0,   0, -28,  56, -27,  -8,   2,  -1,   0,   1,  32,  74,   5, -19,  -3,  -4,  -1, -31, -38,   5,  34,   3,   2,   1,   1,  -1, -14, -21,   6,   3,   1,   1,   0,   3,   3,  -8,   0,   1,  -1,   0,   0,   3,   1,  -4,   0,   1,   0,   0,   0,   1,   1,  -3,   0,   0,   0,   0,   0 },
     {   -7, -11,  -7,  12,   3,   0,   0,   0, -10,   5,  -9,  14,   5,  -1,   0,   0,  -1, -47, -18,   8,   8,   4,   1,   1,-106, -13,  22,  -1,   8,   3,   3,   1, -22, -10,   8,   0,   1,   1,   1,   0,  14,   2,  -4,  -3,  -2,   0,   0,   0,   8,   5,   0,  -2,  -1,  -1,   0,   0,   3,   2,   0,  -1,  -1,   0,   0,   0 },
     {   -9,  -4, -31, -21,   0,   2,   2,   1,   0,  15,  70,  10, -15,  -3,  -4,  -1, -25,  49, -41,  12,   6,  -3,   1,  -1, -22,  51, -11, -24,   5,  -5,   1,  -1,  -6,   1,  21,  -3,   0,   0,  -1,   0,   1,  -8,  -2,   0,   0,   1,   0,   0,   1,  -6,   2,   2,   0,   0,   0,   0,   2,  -3,  -1,   0,   0,   0,   0,   0 },
     {    7,  10,  12, 106,  16, -10,  -8,  -3,  10,   6,  40,  28,  15,  -4,  -4,  -2,   3,  10,   5, -34,  -3,   4,   1,   0,   2,  -3,  -6,  -6,  -4,   1,   1,   1,  -2,  -4,  -4,  -4,  -1,   1,   1,   0,  -1,  -1,  -6,  -4,  -1,   1,   1,   0,  -1,  -1,  -1,  -3,  -1,   0,   0,   0,   0,   0,  -1,  -1,  -1,   0,   0,   0 },
     {    0,   1, -13,   2,  -6,   0,   0,   1,   0,   0,  -8,  47,  -3,  -6,  -3,  -1,   5,  23, -16,  49,  23,  -8,  -2,  -2,   1, -56, -57,  -2,  32,   9,   3,   2,   7,  -4, -26, -32,   2,   7,   2,   1,   2,  12,   9, -13,  -6,   0,   0,   0,   0,   3,   5,  -3,  -3,   0,   0,   0,   0,   2,   3,  -2,  -1,   0,   0,   0 },
     {    3,  10,  20,  35,  13,  -5,  -3,  -2,  -9,  -1, -27, -56, -11,  10,   4,   2,  -3,   7,  20,  43, -23,  -9,  -3,  -1, -20,  51, -42,  10,  30,  -6,   1,  -2,  -9,  14,   5, -32,   2,   3,   1,   0,   0,  -8,  10,   5,  -3,   0,   0,   0,   1,  -3,   1,  -3,   0,   1,   0,   0,   1,  -3,   2,   2,   0,   0,   0,   0 },
     {   -2,  -2,   2,   0,  -1,   0,  -1,   0,  -1,  -1,  -6,   7,   2,   0,   0,   0,  -4,  -2,   7,  12,   3,  -2,  -1,  -1,  30,  21,  -2,   2,   5,  -1,  -1,  -1, -83, -74, -10,  12,  14,   9,   5,   3,  32,  12, -15, -13,  -6,  -1,   0,   0,   4,  11,   7,  -1,  -3,  -1,  -1,   0,  -2,  -3,  -1,   0,   0,   0,   0,   0 },
     {    0,   1,   9, -10,  -8,  -3,   0,   1,   4,  13,  -5,  66,  17, -16,  -6,  -2,  15,  -4,  68,  37,  11,  -5,  -8,  -3,  -3,  50,  30, -15,  -6,   0,  -1,  -1,   9,  17,  -8, -12,  -2,   1,   0,   0,  -3,  -9,  -2,  -5,  -2,   2,   1,   0,  -2,  -4,  -6,  -2,   0,   1,   1,   0,  -1,  -3,  -3,  -3,  -1,   1,   0,   0 },
     {    2,   2,   2, -26,  81, -35,   1,  -1,   2,   3,   1, -17,  76, -20,  -6,   1,   0,   8,  -6,   4,  14,  11,  -8,   0,  -2,   0, -11,  -2, -11,  15,  -1,   0,  -2,   0,  -4,  -2, -12,   8,   1,   0,  -1,   1,  -1,  -4,  -8,   3,   2,   0,   0,  -1,   3,  -2,  -6,   1,   1,   0,   0,   0,   1,   1,  -4,   0,   0,   0 },
     {    3,  -3,  -3,   7,   4,  -1,   0,  -1,   4,   3,  -5,  -8,  -2,   1,   1,   0,   0,  -7, -11,  14,  -3,   0,   0,   0,  -9,  27,   3,   1,   5,  -4,   1,  -1,  85, -72, -30,  18,   1,   4,  -1,   2,   5,  11, -22,  -4,   3,   1,   0,   0, -17,  13,   8,  -6,   1,  -1,   1,   0,   1,  -3,   1,   0,   0,   0,   0,   0 },
     {   -3,   2,  -7,  -3,   1,   1,   0,   0,  -3,  11,  -6,  12,   2,   0,  -1,   0,  -7,   3, -19,  -8, -13,   6,   1,   2, -11,  26, -14,  50, -20, -10,   2,  -3, -15,  38, -68,  26,  33, -16,   3,  -2, -17,  21,  -8, -34,  23,  -2,  -1,   0,  -1,  -5,  14, -13,   2,   3,  -2,   1,   1,  -2,   4,  -6,  -1,   2,  -1,   0 },
     {   -3,  -3,   3, -22, -29,   6,   3,   1,  -1,   0,  -6,  -3,   8,  -2,   1,   0,   2,  -7,  32, -71,  21,  15,  -1,   2, -24,  23, -65, -34,  -1,  12,   5,   2,   1, -11, -23, -10,  -9,   1,   2,   1,   3,   3,  17,   0, -11,  -2,  -1,   0,  -2,   4,   8,  10,  -2,  -3,   0,  -1,   2,  -3,   4,   5,  -1,   0,   0,   0 },
     {   -2,  -3,  -3,  18, -78,  16,   5,   0,   2,  -4,  10, -38,  70, -22,  -3,   1,  -2,  10, -13,  29,  14,   7,  -8,  -1,  -1,  -3,  10,   3, -28,   8,   4,  -1,  -1,   1,  -5, -13,   6,  -3,   0,   1,   2,   1,  -3,   5,  -4,   1,  -1,  -1,  -4,  -3,   5,  -3,   2,   0,   0,   0,  -1,  -3,   0,   1,  -2,   0,   1,   0 },
     {    5,   6,   2,  -2,   1,   1,   0,   0,   7,   6,   4,  -7,  -4,  -1,   0,   0,   4,   6,   0,  -7,  -6,  -1,   0,   0,   5,   6,  10,  -3,  -3,  -1,  -1,   0,  19,  34,   4, -16,  -7,  -1,  -1,  -1,  89,  60, -10, -16, -14,  -6,  -3,  -2,  30,  28,   5,  -3,  -5,  -3,  -2,  -1,   5,   4,   3,   2,  -1,   0,   0,   0 },
     {    1,   1,   0,  11,  11,   2,  -2,  -1,  -1,  -4,   8, -36, -30,  10,   4,   2,   1,  -2,  -4,   7,  48, -11,  -4,  -1,   6,   0,  46, -44,   5,  25,  -4,   0, -11,  11, -62, -17, -18,   9,   9,   1, -19,   7,   2, -16,  -7,  -5,   3,   0,   6, -11,  18,  11, -12,  -3,  -1,   1,   0,   2,  -6,   6,   2,  -2,   0,   0 },
     {    1,   3,   0,  -5,  20,  46,  -5,  -3,   1,   5,   0,  11,  16,  42,   1,  -2,  -1,   2,  -2,  18, -58,   8,  20,  -1,  -4, -12,  -3, -40, -45, -17,  19,   4,  -7,  -9, -23, -24, -18, -21,   3,   5,  -5,   0,   3,  10,  -4, -17,  -4,   0,   2,   4,   9,   8,   9,  -6,  -5,  -2,   1,  -1,   0,   5,   6,  -1,  -2,  -1 },
     {    3,  -1,  -5,   3,   3,   1,   0,   0,   4,  -4,  -6,   0,  -2,   2,   0,  -1,   4,   0,  -9,   2,   9,   3,   0,   1,   3,   0, -11,   5,  -9,   2,   1,   0,   9,   1, -34,  11,   7,  -8,   3,   0,  54, -76, -10,  17,   4,   4,  -1,   2,  49, -33, -30,  14,   0,   1,  -1,   0,  16,   6, -18,  -3,   3,   1,   0,  -1 },
     {   -1,  -4,  -5, -13,  -6,   2,   0,   0,   4,   5,  11,  10,  21,   6,  -3,   0,  -5, -11, -17, -30, -30,  -5,   5,   1,   8,  10,  30,  36,  39, -14, -12,  -3,   5,  -8, -22, -59,  14,  37,   4,   2,  -6, -11,  22,  -2, -44,  -5,   4,   2,   3,  -4,  -6,  20,   7,  -4,  -2,  -1,   2,   4,  -1,  -5,  -4,   1,   2,   0 },
     {    3,   4,   6,   3,  15,  68, -10,  -1,   4,   6,   7,   1,  28,  57,   0,  -6,   5,   3,   6,  -1,  58,  -1,  -2,  -2,   1,   9,  -2,  31,  28, -18,  -8,  -1,   3,   4,  13,  25,  -7, -10,  -2,  -1,   0,   6,   6,   5, -12,  -9,   1,   0,   1,  -3,  -1,  -3,  -8,  -6,   1,   1,   6,  -2,  -2,  -2,  -5,  -2,   0,   0 },
     {   -4,  -6,  -6,  -3, -27, -79, -18,   1,   2,   4,   2,  -1,  15,  67,  -6,  -5,   4,   4,   5,   8,  11,  24,  30,   0,   0,  -4,   2, -12,   7, -41,  -6,   8,  -2,   3,  -7,   7,  -8,   7,  -9,  -3,   1,   5,  -2,   4,   6,  -2,   4,  -1,   4,  -6,   1,  -6,   2,   5,  -1,   1,   1,   0,   1,  -1,  -2,   0,  -1,   0 },
     {    0,  -1,   1,   7,   3, -13,   1,   0,   0,   1,   1,   6,  -8,  -9,   1,  -1,  -1,  -2,  -1,  11,   4,  -1,  -2,  -1,  -3,   4,   0,  18, -15,   1,   2,  -1,   1, -11,   1,  16, -13, -11,   2,   0,  21,  36,  63,  15,  -3, -13,  -8,  -2, -24, -56,  10,  49,  21,   5,  -1,   1,   0, -18, -33,   0,  13,   6,   2,   1 },
     {   -1,   0,   0,  -4,   6,   5,   2,   0,  -1,  -3,  -1,   0,  -1,   9,   1,  -1,  -2,   0,  -1,  -8,   4,  -1,   2,  -1,  -2,  -8,   3,  -3,   8,  -2,   0,   1,  -7,  21,  -8, -14,   7,   1,  -1,   0,  46, -27, -43,   6,   0,   4,  -1,   1, -94, -18,  18,   1,   6,   5,   3,   1, -21, -16,   8,   3,  -1,   0,   0,   1 },
     {   -1,   0,  -4,   0,  -6,  -3,   2,   0,  -1,   2,  -1,   5,   3,  -4,  -1,   2,  -3,   1,  -7,  -3,  -3,  -8,  -6,   1,   1,   8,  -2,   9,  18,  24,  -6,  -6, -11,   5, -18,   7, -52,  10,  22,   0,  -3,  12, -24,  48,  14, -52,  -1,   3,   2,   6, -21, -23,  55,   9, -12,  -1,   2,   1,  10, -22, -11,  15,   3,  -1 },
     {    0,   0,   0,   0,   4, -11,   3,   0,  -1,   0,  -2,   3, -27,   7,  10,   2,   2,   2,  -2,  -3,  41,  15,  -4,   0,  -2,   8,  -4,  54, -59, -11,  16,  -4,   0, -16,  16, -57, -37,  -2,   6,   2,  -7, -10, -17, -15,   6,  -5,  -1,   2,  -2,  16,  -5,   8,   2,  -9,  -2,   1,   1,   0,   2,   2,  11,   1,  -5,  -1 },
     {   -1,   0,  -2,   2,  -8,  -2,  -9,  -2,  -2,   1,  -5,   2, -12,  15, -35,  -3,  -1,   5,  -2,  10, -17,  69, -18, -15,   3,   4,   2,   7,   9,  47,  30, -11,   3,   7,   9,  19,  -5,   8,  27,   1,  -2,  -2,  -6, -18, -33, -14,  26,   9,  -8, -11,  -8,  -1, -12, -32,   1,   9,   5,   1,  -1,   4,   9, -12,  -6,  -1 },
   },
   {
     {  116, -34, -10,  -4,  -3,  -1,  -1,   0, -37,  -5,   5,   3,   2,   1,   1,   0, -11,   5,   3,   0,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -40, -63,  21,   7,   6,   2,   2,   1, -81,  27,  30,   0,   2,  -1,   1,   0,  27,  32,  -7,  -9,  -3,  -2,  -1,  -1,  11,  -2,  -9,  -1,   0,   0,   0,   0,   6,   0,  -2,   1,   0,   0,   0,   0,   3,  -1,  -2,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   0 },
     {  -23, -25, -16,  10,   3,   3,   1,   1, -14, -97,   9,  23,   6,   5,   2,   2, -36,   6,  41,   4,  -1,  -1,   0,  -1,  12,  32,   4, -11,  -4,  -2,  -1,  -1,   4,   7,  -4,  -4,   0,   0,   0,   0,   3,   6,  -2,  -2,   0,   0,   0,   0,   1,   2,  -1,  -1,   0,   0,   0,   0,   1,   2,  -1,  -1,   0,   0,   0,   0 },
     {    0, -91,  -3,  15,   6,   5,   2,   2,  81,   2,   9,   0,  -7,  -2,  -3,  -1,  18,   5,  -2,  -7,  -3,   0,   0,   0, -22,  -5,   2,   0,   2,   1,   1,   0,  -7,   7,   3,   0,   0,   0,   0,   0,  -5,   2,  -1,  -1,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 },
     {    5,  13, -60,   0,   4,   1,   1,   1,  -4, -26, -28,  23,   7,   2,   1,   1,  92,  29,   4,   3, -13,  -5,  -4,  -1,  14, -16,  -4,  -5,  -5,   1,   0,   1, -14, -13,   8,   2,   2,   2,   0,   0,  -5,   1,   6,  -1,   0,   0,   0,   0,  -3,   0,   3,  -1,   0,   0,   0,   0,  -2,   0,   1,  -1,   0,   0,   0,   0 },
     {   14,  34,  17,  -3, -10,  -2,  -3,  -1,  28,  -5,  61,  -5, -11,  -3,  -2,  -1,   0,  69,  18, -41,  -8,  -4,   0,  -2,   3,   6, -45, -18,   8,   3,   2,   1,  -8, -15, -15,  10,   8,   1,   1,   0,  -1,  -4,  -4,   5,   1,  -1,   0,   0,  -1,  -3,  -1,   2,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0 },
     {  -14,  -5, -86,  -2,  12,   4,   4,   2,  -3,  57,  -2,  15,   2,  -7,  -2,  -2, -52,  27,  32,   1,  -8,  -4,   0,  -1, -15,   2,   2,  -9,  -3,   2,   1,   0,   5,  -3,   0,  -1,   1,   1,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0,   4,  -2,   2,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 },
     {   -6,  15, -20,   9,  -2,   1,  -1,   0, -18, -23,  53,  16,  -7,   0,  -1,  -1,   2, -29,  -7, -16,  -3,   4,   1,   1, -67, -59,   0,  18,  16,   7,   3,   2,  -6,  18,  29,   7,  -2,  -4,  -2,  -1,  11,  15,  -3, -10,  -4,  -1,   0,   0,   4,   4,  -2,  -2,   0,   0,   0,   0,   3,   3,  -2,  -1,   0,   0,   0,   0 },
     {   -7, -10, -24, -11,   8,   4,   1,   1,   1, -27, -26, -30,  12,   5,   3,   2, -32,  35, -61, -17,  27,   2,   3,   0,  14, -10, -32,  46,  18,  -6,  -2,  -1,   4,  -8,  22,  31, -12,  -8,  -1,  -1,   3,   1,  13,   0, -11,   0,   0,   0,   0,   0,   6,  -2,  -3,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0 },
     {    4,  -3, -32,  21,   1,   2,   0,   0,  10,   8,  49,  19, -13,  -4,  -4,  -1,  -1, -55, -19,  -9,  -2,   8,   3,   2,  80,  -9, -25,   1,   1,   2,  -2,   0,  27,  -5,  -1,   2,  -2,  -1,  -1,   0, -10,  -6,   3,   0,   1,   1,   1,   0,  -8,   3,   5,  -1,   0,   0,   0,   0,  -4,   1,   0,  -1,   0,   0,   0,   0 },
     {    6,  12,  15,  98,  23, -15,  -9,  -4,  -6,   4, -32,  33,  -7, -13,   0,   0, -15,  12, -26, -32, -19,   5,   7,   2,  -7,   7,   4, -11,   0,   7,   2,   0,  -9,  -6,   7,  -2,   2,   3,   0,   0,   3,   0,   6,  -2,   0,   0,  -1,   0,   2,  -1,   1,  -3,  -1,   0,   0,   0,   1,  -1,   1,  -1,   0,   1,   0,   0 },
     {    2,  14, -26,   2,   5,   1,   0,   0,  -1,  -1,  13, -25,  -7,   3,   1,   1,  25,   5, -26, -37,   6,   6,   2,   1, -14,  59,  29,   3,   9,  -6,  -3,  -2,  36,  62,  -1, -20, -10,  -9,  -3,  -2,  -2, -15, -28,  -3,   4,   3,   2,   1,  -8, -14,  -3,   9,   4,   1,   0,   0,  -1,  -4,   0,   3,   0,   0,   0,   0 },
     {    7,  12,  12,  18,   1,  -6,  -3,  -1,  14,   6,  32,  24,   9,  -9,  -3,  -2,  -2,  40,  -1,  47,  15, -16,  -4,  -3,  16,  -5,  29,  47, -31, -20,  -3,  -1,   4,   3,  26, -24, -43,   1,   5,   2,   0,  -2, -10, -29,  -5,  13,   4,   0,  -3,  -3,  -6,  -8,   6,   5,  -1,   0,   0,  -1,  -3,  -3,   2,   1,   0,   0 },
     {    2,  -1,  17, -20,  -4,  -3,   1,   0,   8,  -1, -20,  56,  28,  -7,  -3,  -2,  -4,  11, -21,  23,  -8, -12,  -1,   0, -16, -16, -49, -24,  -5,   9,   8,   3,  62,  41,  -8,  -1,   0,   1,  -3,  -2,   7,  -8,  -3,   0,  -4,  -1,  -2,   0, -11, -13,   6,   4,   1,   2,   1,   0,  -1,   1,   5,  -1,  -2,   0,   0,   0 },
     {   -5,  -1,  -6, -34, -10,   0,   3,   1,   1,   6,  24,  52,  25,  -7,  -6,  -2,  -4, -13, -56,  -1,   1,  -5,   2,   2, -12,  54,   3, -18, -11,  -4,   3,  -1, -60, -11,  11,   8,   6,   4,   2,   0,  -3,   2,   3,  -2,  -2,   1,   0,   0,   9,   4,   1,   0,  -1,   0,   0,   0,   2,  -3,  -1,  -1,   0,   0,   0,   0 },
     {   -5,  -5,  -6,   6, -86,   5,  14,   1,  -4,   2, -14,  22, -71,  18,  14,  -2,  -6,  16, -14,  21,  -8,  19,   1,  -3,  -1,   8,  -1,  12,  15,   7,  -7,  -2,   0,   3,   4,  -2,  12,   0,  -6,   0,   3,  -4,  -1,  -9,   7,  -2,  -2,   2,   1,  -3,   2,  -2,   6,  -2,  -1,   1,   0,  -1,   1,  -2,   3,  -1,   0,   0 },
     {    0,   4, -20,  42,  -5,  -4,  -3,   0,  -5,  -5,  21, -19,   4,   2,   0,   1,  15, -12, -14,  51,  36, -10,  -5,  -2, -41,  24, -29,   8,  -4, -14,   1,   0,   9, -25, -55,   1,   2,   4,   6,   2,   5, -19,  10,  27,  -1,   2,   0,  -1,   6,   3,  14,  -2, -10,  -1,  -1,   0,  -1,  -1,   5,  -1,  -3,   1,   0,   0 },
     {    2,  -1,  13,  -5,   6,  -2,   1,  -1,  -3,  16, -16,  32, -11,  -1,  -2,   0,  19, -26,  49, -25,  13,   4,  -4,   0, -18,  43, -33,  36,  26, -21,   0,  -3,  14, -18,  14,  37, -38,  -8,   2,   0,  -4,  -4,  11, -24, -12,   8,   3,   1,   2,   1,  -5,  -3,   0,   3,   0,  -1,   0,  -2,   1,  -5,   0,   1,   0,   0 },
     {    1,  -2,   4, -24,   3,   5,   3,   0,   3, -11,   6,  26,  -1,  -4,  -1,  -1,  -1,   8, -20, -28, -15,   5,   3,   1, -15,  -8,  60,  14,  -2,   3,  -5,  -2,  56, -64, -34,   9,   4,   3,  -2,   2,  -1, -17,   5,   1,   3,   3,   1,   1,  -3,  14,  12,  -3,   0,  -2,  -1,  -1,   0,   2,  -4,  -3,   0,   0,   0,   0 },
     {    5,  -1,  -6,  -1,   2,  -3,   0,   0,   8,   5,  -2,  -2,   0,  -1,  -1,   0,   8,  -2,  -7,  -1,   1,   1,   0,   1,  11,  10,  -2,  -2,   0,   0,  -2,  -1,  13,  -8, -20,  -1,   3,   1,   0,   1,  92,  59, -13, -17,  -9,  -6,  -3,  -2,  32,   1, -28, -10,   0,   1,  -1,   0,  -7, -19, -10,   5,   6,   3,   1,   0 },
     {   -2,  -4,  -4,  27, -78,   4,   0,   1,   2,   9,  -7,  -3,  65,  12, -12,   1,   1,  -6,  16, -43,  38, -15,  -5,   2,   6, -13,  12,  -6, -16,  -9,   7,   1,   5,   0,   1,   5,   2,   1,   1,  -1,   0,   2,  -4,  -4,  -2,  -1,   0,   0,  -1,   5,  -2,   1,   1,  -1,   1,  -1,   2,   0,   0,   2,  -4,   0,   1,   0 },
     {    0,  -3,   5,  -7,   8,   1,  -1,   1,   1,   0, -13,  34,  -2,  -6,  -2,   0,  -4,   0,   8, -16,  12,   6,  -3,   1,  17, -27,   3,  34,  21, -16,  -5,  -1, -40,  19, -45,  -8,  -5, -10,   8,   2,  27, -28, -53,  28,   6,   5,   6,   0,  10, -13,  24,  26, -15,  -2,  -1,  -1,   3,   4,   8,  -9,  -8,   2,   0,   0 },
     {   -3,  -6,  -7, -12,  -3,  -3,   2,   0,  -6,  -9, -14, -13, -19,   3,   4,   3,  -6, -14, -10, -29, -11,  -6,   2,   2, -13,  -8, -34,  -4, -44, -13,  14,   3,  -3, -18,  -5, -41, -38,  41,  20,   2,   3,  -3,  -7, -14,  42,  38,  -3,  -4,   7,   4,   2,  17,  24,  -3, -13,  -4,   3,  -1,   3,   5,   3,  -6,  -4,   1 },
     {   -1,   0,   0, -10,  13, -13,  -4,   1,  -1,   1,  -5,  22, -26,  -5,   1,   0,   0,  11,  -5,  -8,  80,  22, -15,   0,   0, -16,   8, -47,  26, -20, -12,   6,   7,  -9,   4, -35, -15,  -8,   5,   4, -19,  28,  12,   8,  10,  -3,   5,  -3,   0,   3, -11,  -3, -11,  -1,   4,   0,  -3,   4,  -5,   4,   1,   3,   1,   0 },
     {    5,   8,   8,  12,  20,  87,  17,  -6,   2,   6,   1,  10,   8,  72,   0, -12,  -3,   1,  -6,   0,  -7,   9, -25,  -8,  -4,  -1,  -8,  -9,  -6, -18, -16,   2,   0,  -5,   4, -11,  -6, -15,  -3,   7,   4,   0,   1,  -1,   6,  -8,   0,   3,   2,   3,  -1,   1,   1,  -6,   1,   1,   3,   1,  -1,  -1,  -2,  -3,   0,   2 },
     {   -2,   5,  -4,   1,   1,   4,   2,  -3,  -4,   6,  -6,   8,   5,  -2,   1,  -1,  -4,   5,  -4,  -5, -13,  -5,   1,   1,  -2,   1,  -7,  33,   0,  -6,   1,  -3,  -8,  25, -58,  -2,   8,  -4,   3,   0, -50,  71,  -3, -16,  10,  -2,   2,   0, -25,  40,  -1, -12,   3,  -2,  -2,  -2,  -7,   1,  -6,  -2,   3,  -1,   1,   0 },
     {   -1,  -3,  -4,   8,   5,  -2,  -6,   0,  -2,  -3,  -2,  -3,  17,  -1,  -6,   1,  -2,  -3,  -6,  20,   7,   2,   0,   0,  -5,  -1,  -9,  -4,  31,  14,  -9,  -3,   0, -36,  -2,   2,  19, -12, -13,   1, -19,  -4, -57, -41,  17,   0,   8,   6, -13, -19, -42,  30,  39,   0,   2,   1,   2,   1,  17,  33,  -6, -13,  -2,   0 },
     {   -2,  -4,  -3,  -3,   3, -69,  -9,   3,   5,   5,   7,   2,  25,  55,  34,  -4,  -1,   1,   1,   1, -17,  65,  -4, -10,  -4,  -3,  -6,   1, -23, -21, -23,  -1,   0,  -1,  -2,   5,   0, -11,  -3,   6,  -2,   1,  -4,   1,  -3,  -1,   2,   2,  -2,   0,   6,   0,   1,   2,   0,   2,  -1,  -5,   4,   1,   2,  -1,   0,   1 },
     {   -2,  -2,   0,   0,  -1,   4,   5,   2,  -3,   2,   5,   0,  -2,   0,  -2,   0,  -3,  -4,   2,  -6,   1,   1,   0,   0,  -4,   3,   8,   4,  -3,  -5,   0,  -1, -12, -11,  -5,  -5,  -2,   7,   7,   1,  28,  26,  25,   2,  -8,  -3,  -5,  -2, -90, -61,  21,  13,   5,   7,   2,   1, -19,   2,  25,   4,   3,  -1,   0,  -1 },
     {   -1,  -1,  -3,   1,  -9,   6,   1,   3,  -2,  -4,  -2, -10,  -4,  -4,   0,   0,  -4,  -1,  -9,   2, -14,  -1,  -3,  -2,   3, -17,  10, -38,  14, -12,  -4,   5, -15,  12, -35,  18, -33, -34,  21,   3,   0, -16,  12, -21, -48,  44,  23,  -2,   3,  -3,   0, -22,  34,  31, -11,  -4,  -4,   3,  -5,  11,  15,  -7,  -8,   1 },
     {    0,   0,   1,  -1,  -9,  -2,  -4,   2,   2,   1,   3,  -7,  36,  -2,  -7,   2,   1,   1,   4,   6, -17,  17,  -1,   0,   2,   9,  -9,   6,  62,  34,  -6,  -5,  -6,  -9,   1, -66, -17,  -1,  -6,  -1,  13,  -8,  27, -24,   0,   4,  -1,   6,   1,  23,  39,   2,   2,  -9,  -4,   1,  -6, -15,  -3, -12,  -5,  -5,   0,   3 },
     {    1,   2,   1,   1,  -7,  37,   3, -17,  -1,  -1,   0,  -2,  11, -50,   2,  -4,   6,   4,   7,   3,  12,  63,  46,  -6,  -5,   0,  -7,  -2, -35,  21, -26, -19,  -3,   3,  -5,   8, -20, -22, -22,  10,   3,   5,   9,  13,   5,   9,  -8,  10,   2,   4,  -8,  14,  14,  -7,   0,  -1,   5,   4,  10,  -8,  -4,   5,   5,   0 },
   },
   {
     {  123, -23, -16,  -6,  -4,  -2,  -2,  -1, -15,   1,   2,   1,   0,   0,   0,   0, -12,   2,   2,   1,   0,   0,   0,   0,  -5,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   19, 119, -21, -26,  -9,  -5,  -2,  -2,   5, -14,  -3,   2,   1,   0,   0,   0,  -3, -13,   2,   3,   1,   0,   0,   0,   0,  -4,   1,   1,   0,   0,   0,   0,  -1,  -5,   1,   1,   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,  -1,  -2,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   -2, -15, -76,  10,  12,   4,   2,   2,  94,   0, -27,  -2,  -3,  -1,  -2,   0, -10,   1,  13,   0,  -1,   0,   0,   0,  -9,   0,   6,   1,   0,   0,   0,   0,  -6,   1,   5,   0,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0,   0,  -2,   1,   2,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {  -21, -10, -88,   2,  17,   5,   3,   3, -69, -39,  21,  21,   6,   2,   1,   1,   8,  10,  15,  -1,  -3,  -1,  -1,  -1,   8,   5,   0,  -3,  -1,   0,   0,   0,   5,   3,   4,  -1,  -1,   0,   0,   0,   4,   1,   0,  -1,   0,   0,   0,   0,   2,   1,   2,   0,   0,   0,   0,   0,   1,   1,   0,  -1,   0,   0,   0,   0 },
     {    8,  -4,  34,  25,  -4,  -3,  -3,  -1,  20,-110, -23,  30,   9,   6,   1,   3,  -6,   3,  -4,  -4,   0,   1,   0,   0,   0,  15,   2,  -5,  -1,   0,   0,   0,  -2,   6,  -1,  -3,   0,   0,   0,   0,  -1,   6,   1,  -2,   0,   0,   0,   0,  -1,   2,  -1,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0 },
     {    6,  20,   2,  81,   6, -12,  -8,  -3,  17,   8,  82,  12, -20,  -6,  -5,  -2, -23,   5,   7, -21,  -3,   2,   2,   1,   1,  -2, -14,  -7,   1,   2,   1,   0,  -2,  -2,  -4,  -6,   0,   1,   1,   0,   0,  -1,  -5,  -2,   1,   0,   0,   0,   0,   0,  -1,  -2,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 },
     {   -7, -20,   5, -81, -21,  14,   9,   3,  30, -23,  75,  15, -13,  -2,  -4,  -1, -23,  -2,   7,  16,   6,  -1,  -1,   0,   0,   4, -12,  -2,   0,   0,   0,   0,  -1,   1,  -3,   4,   1,  -1,   0,   0,  -1,   1,  -5,   0,   1,   0,   0,   0,   0,   1,  -1,   2,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0 },
     {  -12,  -3,   3,   6,   5,  -1,  -1,   0, -20,   6, -23,  -3,   7,   1,   2,   0,-108, -44,  30,  19,   6,   4,   3,   1,   5,   3,   7,   1,  -1,  -1,   0,   0,  10,   5,   0,  -2,  -1,   0,   0,   0,   5,   1,   0,  -1,  -1,   0,   0,   0,   3,   1,  -1,  -1,   0,   0,   0,   0,   3,   1,   0,   0,   0,   0,   0,   0 },
     {   -2,   5,   3,  -2, -16,   3,   1,   1,  -5, -22,  -1, -83, -22,  18,   8,   4, -20,  69,  40, -20,  -9,  -6,  -1,  -2,   1,   5,   4,  19,   6,  -3,  -1,  -1,   0,  -5,  -3,   4,   2,   0,   0,   0,   2,  -1,  -1,   5,   1,  -1,  -1,   0,   0,  -2,  -1,   1,   1,   0,   0,   0,   1,  -1,  -1,   2,   1,   0,   0,   0 },
     {    0,  -9, -16,  14, -12,  -1,   1,   1,  -2, -26,  20, -69, -23,  16,   6,   3,  18, -73, -44,   9,  23,  10,   2,   3,   2,   3,   8,  19,   5,  -2,  -1,   0,   0,  10,   8,   2,   0,  -1,  -1,  -1,   0,   4,   0,   4,   1,  -1,  -1,   0,   0,   3,   1,   0,   0,  -1,   0,   0,   0,   2,   0,   1,   0,  -1,   0,   0 },
     {    6,   4,  17, -22, 115,  -8, -13,  -3,   7, -11,  15, -24,   8,  -2,  -1,   1,   1,  -6,  -1,   0, -28,   2,   2,   1,  -1,   2,  -2,   5,  -4,   1,   0,   0,   0,   0,  -2,   3,  -7,   0,   1,   0,   0,   1,  -1,   3,  -2,   0,   0,   0,   0,   0,  -1,   1,  -3,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 },
     {    4,  -3,  14,   9,  -9,  -1,   0,   0,   1,  -3,   7,   2, -11,   1,   0,   1,  45, -43,  94,  29, -30,  -2,  -7,  -1, -13, -12,  28,   4,  -2,   1,  -1,   0,  -5,   3,  -8,  -4,   2,   0,   0,   0,  -2,   3,  -4,  -2,   1,   0,   0,   0,   0,   2,  -5,  -2,   1,   0,   0,   0,  -1,   1,  -2,   0,   1,   0,   0,   0 },
     {   -6,  -4,  -1,   0,   2,   2,   1,   0, -10, -12,   0,   4,   6,   2,  -1,   0, -11,   1,  -9,  -7,   5,   2,   1,   0, -85, -76, -11,  24,  20,   8,   4,   2, -26, -25,   0,  10,   7,   2,   0,   0,   6,   8,   4,  -1,  -2,  -1,  -1,   0,   7,   6,   1,  -1,  -2,  -1,   0,   0,   3,   2,   0,  -1,   0,   0,   0,   0 },
     {   -3,  -5,  -2, -18,   0, -18,  -2,   2,  -9,  -3, -23,  17, -89, -28,  13,   5,  -6, -15,  -6, -46, -39,  19,  14,   4,   1,   0,  -1, -17,  20,  17,   1,  -1,   3,   1,   5,   8,  11,   3,  -2,  -1,   1,   3,   1,   0,   8,   2,  -1,  -1,   0,   0,   1,   2,   2,   0,  -1,   0,   0,   0,   1,   0,   3,   1,  -1,   0 },
     {    2,   6,   1,  15,  18,  11,  -2,  -2,  -4,  -2, -11,   9, -58, -16,   7,   3,   1,  25, -11,  87,  23, -33, -10,  -3,   1,   3, -19,  13,  30,   1,  -3,  -2,   2,  -2,  -7, -14,   0,   3,   2,   0,   0,   1,   2,  -5,   4,   2,   0,   0,   1,  -1,   2,  -4,  -1,   1,   1,   0,   0,  -1,   1,  -2,   1,   0,   0,   0 },
     {   -5,  -8,  -7, -10, -11,-106, -25,   7,   1,  -5,   3, -10,  17, -38, -12,   4,   3,  12,   6,  22,  19,  22,   3,  -3,   0,   3,   0,  10,   3,  14,   4,  -1,   1,  -1,  -3,  -2,  -1,  10,   1,  -1,   0,   1,  -1,  -1,  -2,   4,   1,  -1,  -1,   0,  -1,  -1,   0,   3,   1,   0,   0,   1,   0,   0,  -1,   1,   1,   0 },
     {    5,  -7,  -2,   1,  -2,   1,   0,  -1,   8,  -9, -10,  -2,   8,   3,  -2,   0,   7, -13,  14,  -8,  -2,   2,  -1,   0,  71, -62, -62,  18,  11,   4,  -1,   2,  29, -24, -25,  14,   4,   2,  -1,   1,   0,  -2,   1,   3,   0,  -1,  -1,   1,  -5,   4,   4,   0,  -1,   0,   0,   0,  -3,   4,   3,   0,  -1,   0,   0,   0 },
     {   -1,   1,  -2,   0,  -2,   1,   3,   0,  -3,   5,  -9,  -5,  11,   1,  -1,   0,  10, -20,  21, -14,   4,   0,   0,   0, -46,  68, -81,   2,  25, -12,   6,  -2,  13, -12, -18,   9,   1,   2,   2,   1,  -4,  -2,   8,   2,  -2,   0,  -1,   0,   4,  -3,   3,   3,  -3,   0,  -1,   0,   0,  -1,   3,   0,   0,   1,   0,   0 },
     {   -2,  -3,  -2,   1,   1,  -2,   1,   1,  -2,   2,  -1,  -1,  -3,   6,   1,  -2,  -3,  -7,  -3,   5,   0,   1,   1,   1,  26,  24,   8, -11,  -8,  -7,  -2,   1, -72, -77, -22,   9,  17,  10,   5,   3,  32,  34,  13,  -3,  -6,  -4,  -2,  -2,  -9,  -7,  -1,   3,   3,   0,   0,   0,   4,   3,   0,   0,   1,   0,   0,   0 },
     {   -1,   0,  -4,   6, -19,  45, -39,   0,  -1,  -9,   4, -26,  29, -79, -12,  14,  -2,  -4, -10,  10, -35, -17,  35,  -2,   0,   2,  -5,  -5,  -2,  29,   4,  -6,  -7,   3,  -8,   0,   8,  -2,  -2,  -1,   3,   0,   3,   1,  -5,   5,   2,  -1,   1,   0,   0,  -1,   2,   1,   1,  -1,   1,   0,   1,   1,  -1,   2,   1,  -1 },
     {    3,   1,   7,  -5,  21,  23,  -3,  -3,  -2,  -3,  -3,  -4,  -7, -45,  -2,   6,   9,  -5,  31, -40,  92,  -4, -17,   4,   6,  -3,  16, -16,  13,  16,  -3,  -1,  -4,  -6,   3,   6, -23,   1,   1,  -2,  -5,   7,   0,   0,  -1,   3,   2,  -1,   2,   0,  -4,   3,  -5,  -2,   1,  -1,   0,   0,  -1,   1,   0,   1,   1,   0 },
     {    3,   3,   3,   8,   5, -14,  99,  18,   2,  -4,   5, -10,  18, -40,  50,  11,  -2,  -2,  -4,   2, -11, -16,  -6,  -3,  -1,  -1,  -3,  -1, -11,  14, -11,  -5,   1,  -2,  -1,  -3,   5,  11, -11,  -4,  -2,   0,  -2,  -2,   0,   5,  -5,  -1,   0,   0,  -2,   0,   0,   0,  -4,  -1,   0,  -1,  -2,   0,   1,   1,  -1,   0 },
     {    1,   0,   0,  -1,  -1,   4,   1,  -1,   2,   4,   5,  24,   3, -18,  -4,  -2,  -2,  -3,  -6, -24,  -7, -12,   3,   5,   8,  23,  38,  98,  28, -25, -14,  -8,  15, -15, -16,   6,  -5,  -1,   1,   0,   2,  -3,  -4, -15,  -7,   4,   2,   2,   0,   0,   1,  -6,  -1,   3,   1,  -2,  -1,   0,  -3,  -5,  -1,   2,   0,   0 },
     {   -2,   2,   4,  -1,  -1,   0,   2,  -1,  -2,   4,  -5,   5,   5,  -5,   7,   1,   0,  -1,  12,  -5,  -4,   5,   1,  -1,  22,  -2, -38,  30,  18,  -7,  -1,  -1, -76,  52,  53,  -4, -17,  -7,   3,  -4,  17,  -2, -25,   0,  -4,  -1,   2,  -1,  -8,  -3,  -2,   2,   1,   0,   1,   0,   7,  -1,   1,  -1,  -1,   1,   0,   0 },
     {   -2,  -4,  -1,  -5,  -6, -18,  -3,  -3,  -3,  -5,  -6,  -4, -15, -11, -28,  -3,  -3, -13,  -3, -18, -10, -79, -21,  21,  -4,  -6, -12,   6, -49, -24,  31,  16,   2,  -9,  22,  -3, -19,  30,  20,  -1,  -4,   3,  -3,  -2,   9,  10,   1,  -2,   1,   4,   2,   3,   1,   5,   0,  -2,  -1,   2,   0,  -2,   3,   2,  -1,  -1 },
     {    1,  -2,   6,   4,  -4,   6,  -3,   6,   1,  -2,   4,   2,   2,   2,   6,  -3,   2,   1,   6,  13,  -8,  16,  22,  -5,   2,   7,  -3,   4,   3,  11,  -9,  -3,  25, -64,  89,  -1, -39,   7, -15,   0, -12, -11,  11,  -3,   2,  -1,  -5,   3,  -4,   1,  -3,  -1,   4,  -4,   0,   1,  -2,   1,  -3,   0,   2,   0,  -1,   0 },
     {   -4,  -3,  -1,   2,   3,   3,  -6,  -1,  -7,  -5,  -3,   1,  -1,  -1,   5,   3,  -5,  -4,  -2,   3,   4,   6,  -3,  -1,   1,   2,   4,   1,  -6,  -3,   0,   2, -33, -23, -15,   9,  19,   7,   3,   3, -67, -78, -35,  12,  24,  15,   4,   3, -20, -12,  -1,  12,   6,   2,   1,   0,  -6,  -5,   4,   4,   1,   4,   3,   0 },
     {   -2,  -3,  -3,  -6,  -5,  -9, -49, -19,   4,   5,   5,   9,   6,  14,  99,  11,   2,  -2,   3,  -6,  10, -30,  26,  -1,  -3,  -3,  -4,  -2, -14,  -9, -13,  -4,   1,   5,  -8,   9,  -9,  10,   3,   4,  -7,   9,  -5, -11,  -5,   1, -10,   1,   1,   8,  -2,  -1,  -2,   4,   1,   0,  -1,   2,   0,  -1,   1,  -1,  -2,  -1 },
     {   -1,  -3,   0,  -5,  -3, -14,  -8,   5,   2,   0,   6,  -6,  23,  15,  -6,  -3,  -4,  -9,  -4, -13, -15, -49, -23,   2,  10,  -2,  15, -33,  80,   8, -38,  -1,  -3,  -1,  16, -18,  23,  26,   5,   2,  -8,  -7,   8,  10,  -9,  -3,   3,   4,   6,   5,  -3,   3,  -5,   2,   0,   1,  -3,  -1,  -2,   1,  -3,  -1,   1,   1 },
     {    1,   2,   3,  10,   3,  -4,  -1,   0,   0,   3,   4,   6,   2,  -2,  -5,   6,   0,   4,   3,  13,   3, -16, -10,   0,   3,   7,   8,   2,   1,   0,   7,   0,   7,  13,  28, 107,  41,  -6, -16,  -9,  -7,  14,   6,  16,   0, -11,  -8,  -3,  -3,   0,   5,   8,  -3,  -2,  -2,   3,  -2,   0,   5,  -6,  -5,   0,   1,   1 },
     {    3,  -2,  -2,   2,  -1,  -4,  -1,   0,   5,  -3,  -3,   1,  -1,   1,   2,  -5,   4,  -3,  -3,   1,   2, -14,   4,  -7,  -9,   2,   7, -12,   4,   4,  -3,  -5,  21, -16,  -5,  15, -14, -10,   4,   1,  78, -37, -72,  34,  -5,  -3,   2,   0,   4, -18,  -7,   6,  -3,   1,   0,  -2,   6,  -2,  -6,   2,   7,   2,  -3,   0 },
     {    1,   2,   3,   1,   2,   6, -18,  98,   1,   3,   1,   5,  -3,   8,   2,  61,   0,  -2,  -1,  -3,   1,   1,  -9, -27,   0,  -5,   0,  -4,   0, -26,  11,  -8,   3,   0,   0,  -1,  -3,   4,   4, -12,   3,  -2,   3,  -2, -11,   8,   4, -15,  -7,  -2,   3,   0,  -2,   2,   5,  -8,   3,  -4,  -2,   7,   1,  -1,  -3,  -5 },
   }
 },
 {  //2
   {
     {  121, -29, -13,  -5,  -4,  -2,  -1,  -1, -19,  -5,   0,   1,   1,   1,   0,   0, -11,   4,   2,   1,   1,   0,   0,   0,  -6,   1,   0,   0,   0,   0,   0,   0,  -5,   2,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   27,  49, -16,  -7,  -5,  -2,  -2,  -1, 104,   3, -28,  -8,  -6,  -2,  -2,   0,   8, -24, -10,   3,   1,   1,   0,   0,  -9, -11,   4,   5,   2,   1,   0,   0,  -9,  -4,   4,   1,   0,   0,   0,   0,  -7,  -1,   3,   0,   0,   0,   0,   0,  -4,  -1,   1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0 },
     {  -13, -86,  -1,  16,   8,   5,   2,   1,  47, -67,   3,  15,   3,   3,   0,   0,  29,  -9,  11,   4,  -3,  -1,  -1,   0,  10,   5,   5,  -3,  -4,  -1,   0,   0,  -1,   6,   2,  -3,  -1,   0,   0,   0,  -3,   6,   1,  -2,   0,   0,   0,   0,  -2,   4,   0,  -1,   0,   0,   0,   0,  -2,   3,   0,  -1,   0,   0,   0,   0 },
     {    6, -30,   7,  -3,   3,   0,   1,   0,  15,  65,   4, -12,  -4,  -4,  -2,  -1,  77,  43, -27, -19,  -8,  -4,  -2,  -1,  22, -10, -27,  -5,   1,   1,   1,   1,   2, -14,  -9,   6,   4,   1,   1,   0,  -5,  -9,   0,   5,   1,   0,   0,   0,  -3,  -3,   1,   2,   0,   0,   0,   0,  -3,  -2,   1,   1,   0,   0,   0,   0 },
     {  -10,  -7, -66,  -3,  15,   4,   2,   1, -34,  -6, -78,   6,  20,   6,   2,   1,  22, -18, -37,  20,  13,   2,   0,   0,  14, -11,   1,  20,   2,  -2,  -1,   0,   9,  -4,  11,   9,  -4,  -3,  -1,   0,   3,  -1,  10,   2,  -4,  -1,   0,   0,   1,   0,   6,   0,  -2,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0 },
     {    7,  44, -40,  -1,  -2,   0,  -2,   0,   2, -38,   0,   9,   5,   2,   0,   0,  17,  56,  30,  -3,  -5,  -6,  -2,  -1,  50,  50,  -7, -22, -14,  -6,  -2,  -1,  15,   2, -21, -15,  -3,   2,   1,   0,   2,  -8, -11,   0,   4,   3,   1,   0,  -3,  -7,  -2,   3,   2,   0,   0,   0,  -1,  -3,   0,   1,   0,   0,   0,   0 },
     {  -17, -33, -40,   7,  13,   4,   1,   1,  24,  38, -21,   7,   8,  -3,  -1,  -1, -47,  60,  11,  -1,  -1,  -7,  -1,  -1, -53,  25,   7,  -7,  -3,  -3,   1,   0, -28,  14,   6,  -4,   0,   0,   1,   0,  -5,   5,   2,  -2,   0,   0,   0,   0,   3,   0,  -1,  -1,   0,   0,   0,   0,   2,  -2,  -1,   0,   0,   0,   0,   0 },
     {    6,   3,  10,  17,   4,  -4,  -1,  -1,  -4,  43,  -3,  31,   1,  -5,  -3,  -1,  36, -18,  40,  44,  -8,  -7,  -4,  -1,   5,  13,  64,  10, -26, -12,  -4,  -1,   2,  10,  20, -26, -24,  -4,   1,   1,  -7,   0,  -2, -22,  -6,   5,   3,   1,  -3,  -1,  -5,  -9,   2,   4,   1,   0,  -2,  -1,  -2,  -3,   2,   2,   0,   0 },
     {   -2,  25, -41,  11,  -2,   1,  -1,   1, -17, -24,  35,  27,   0,  -2,  -1,   0,  38,  13,  29,  -2, -16,  -8,  -2,  -1, -40, -55,  -8,  -6,  -3,   5,   4,   2, -39, -29,  20,  12,   9,   7,   3,   1,  -6,   9,  18,   3,   0,  -1,  -2,  -1,   0,   7,   5,  -5,  -3,  -2,  -1,   0,   1,   3,  -1,  -3,   0,   0,   0,   0 },
     {    9,  10,  11,  72,  27, -13,  -9,  -3,   9,   9,   5,  68,  20, -18, -10,  -3, -23,  -5, -30,  11,  -3, -12,  -1,   1,  10, -10, -39,  -8,  -6,   0,   5,   3,  16,  -1, -17,  -1,  -1,   4,   4,   1,   9,  -3,  -2,   2,  -2,   1,   1,   0,   2,  -3,   3,   0,  -3,   0,   0,   0,   0,  -1,   3,  -1,  -2,   0,   0,   0 },
     {    7,  23,  48,  37,  10,  -6,  -5,  -2, -17, -31, -33,  -5,   0,  -1,   1,   1,  26,  19, -37, -32,  -6,   4,   5,   1, -21,  41,  24,   1,   6,   1,   0,  -1, -47,   1,  20,   3,   0,  -3,  -2,  -1, -23,   3,  11,  -3,  -2,  -1,   1,   1,  -4,   4,   2,  -3,   0,   1,   1,   0,   0,   2,   0,  -1,   1,   0,   0,   0 },
     {   -2,  -9,   7,  -1,  -5,  -1,   2,   0, -11,   9, -44,   0, -16,   2,   2,   1,   0, -33,  16,  -5, -28,   3,   3,   1,  -9,  -1,  -3, -60, -27,  13,   7,   3, -15, -24, -37, -35,  24,  31,   9,   2,  -3,  -2,   0,  16,  35,  11,  -4,  -3,   4,   4,   5,  11,  10,  -7,  -7,  -2,   1,   1,   2,   3,   1,  -4,  -1,   1 },
     {    0,  -3,  27, -14,  15,   1,  -1,  -1,   3,  -4,  -5,  -1,  31,   5,  -2,  -1,  -6,  11,  19,  47,  40,  -3,  -8,  -3,   6,   3,  -5,  24,   6, -20, -11,  -2, -40, -55, -39,   8,  -3,  -9,   2,   4, -22, -25,   0,  24,   5,   2,   6,   2,   1,   5,  14,  13,  -4,  -4,   0,  -1,   1,   3,   4,  -1,  -7,  -3,   0,   0 },
     {   -1,  -8,   9,  11,  -1,   1,  -1,   0,   7,   2, -15,  -9, -13,  -1,   0,   0, -47,  30,   1,  -9, -20,   3,   2,   1,  73, -44,  23,   4, -16,   9,   0,   1, -19, -33,  40,  -3,  -3,   7,   0,   0,  -5,   5,  21,  -9,   3,   3,  -1,   0,  -6,   6,   4,  -9,   3,  -1,  -2,  -1,  -1,   3,  -2,  -3,   2,   0,  -1,   0 },
     {    4,   8,  53, -43,  -5,  -4,   1,  -1,   2, -21, -41,  14,  20,   2,   0,   0,   7,  43,   4,  32,  -4, -13,  -7,  -2, -22, -31, -19, -11, -21,  -3,   4,   3,  22,  19,   9,   2,  -2,  10,   5,   1,  31,  27,   8,  -8,  -8,  -2,  -3,  -3,   3,  -7, -13,  -9,  -3,   1,   0,   0,   1,  -3,  -2,   0,   2,   2,   1,   0 },
     {    4,   3,  10,  13,   9,   8,   0,  -2,   4,  17, -13,  12,   9,  14,  -3,  -3,   5,  -4,  39,   4,   6,  16,  -3,  -2,  17,  14,  14, -13,  34,  19,  -6,  -3,   9,  12,  11,  37,  55,  -2, -19,  -6,  11,  18,  23,  31,   3, -40, -22,  -3,   0,  -1,  -2,  -2, -23, -27,  -1,   5,  -2,  -2,  -3,  -4, -10,  -3,   7,   4 },
     {    3,   4,  -2, -20,  68, -13,  -6,   0,   2,   4,  15, -27,  68, -16,  -8,   0,   1, -18,   9, -36,  21, -10,  -2,   3,   7, -10,   9, -37,  -6,  -3,   3,   3,   1,   8,  16, -19,  -9,   4,   4,   1,  -7,  10,  10,  -1,  -5,   3,   2,  -1,  -8,   5,  -3,   0,  -3,   1,   0,  -1,  -2,   0,  -6,   6,  -2,   1,   1,   0 },
     {    4,  -4,   5,  11,  12,  -4,   0,  -1,   5,  -8,  -8,  -9,   4,  -2,   0,   0,   1,   5,  11, -24, -13,  -3,   1,   2, -23,   1,  30,   2,  -6,   1,   1,   1,  33, -43,  13,  17,  -7,   2,  -3,   0,  45, -60,   1,  12, -10,   5,  -1,   1,  40, -40,   1,   7,  -5,   4,   0,   1,  22, -16,   6,   2,   0,   3,   0,   1 },
     {    1,   3,   1,  11,   4,  -1,  -3,   1,  -2,  -6,  -8,   5,   4,   1,  -4,   0,   4,  16,  -1,  -8,  10,   1,  -4,   0,   6, -44,  29,  10,  23,   2,  -4,  -1, -15,  24,  -5, -17,  19, -10,  -2,  -2,   3,   2, -68,  -2,  23,  -4,   1,   1,   2, -18, -31,  51,  13,  -9,   0,   0,  11,  -2,  10,  28, -15,  -8,  -1,   1 },
     {    4,   3,  -4,  12,   9,   1,   2,   0,   3,   5,  -1, -16,  -1,   1,  -2,   0,   7,  -4,   4,  -4,   3,  -4,   0,   1,   9,  11,  -4,  14,   7,  -2,  -1,   0, -19, -20, -17,  -5,  -3,  -2,  -1,   1,  47,  43, -18, -20,  -2,   0,   2,   0,  70,  54, -21, -18,   1,  -3,  -2,  -1,  18,   4, -21,  -4,   0,  -3,  -1,   0 },
     {    1,   4, -12,  24, -12,   8,  -3,   0,   3,  -9,  35, -40,  14,   6,   3,  -1,   2,  19, -53,  49,   5,   3,   0,  -1,  -4, -15,  43, -21, -19,   8,   1,   1,   0,  20, -21, -19,  16,   9,   1,   0,  15,  -5,  17,  28,   3,  -5,  -9,  -2,   6,   0,   7, -14, -12,  -6,  -5,   0,   4,  -8, -12,  -3,  -5,   1,   0,   1 },
     {    3,   3,   2,  65,  14,  -7,  -2,  -4,  -5,  -3, -23, -67, -17,   3,   6,   3,   5,   8,  46,  10,  18,   4,   0,  -3, -14, -22, -29,  11,  -2,  -4,  -2,  -3,  25,   8, -11,  -6, -10,  -3,   0,   2,  -9,   5,   4,  -4,  -2,   3,   1,  -2, -17,  -9,  -9, -15,   5,   7,   1,   2,  -8,  12,   6,   3,   7,   1,   1,   3 },
     {    0,  -2,   3,   4, -29,   3,   2,   4,   1,   0,  -3,  13,  -6,   0,  -5,   1,   3,   9,  -1,  -5,  52,   1,  -9,  -2,  -3,  -7,  -1, -27,  54,  -3,  -9,   0,  10, -12,  12, -60,  11, -11,  -4,   4,   8,  -2,  31, -38,   1,   3,   4,   3,  15,  -8,  34,  -5,   2,   5,   0,   1,   7,  -4,  16,  -2,  -7,   0,  -1,   0 },
     {    5,   8,   6,   8,  17,  69,   7,  -2,   4,   6,   7,  11,  12,  78,   6,  -3,  -1,   1,  -6, -11,   0,  39,  -9,  -5,  -7,  -4,  -6, -10, -23,   1,  -9,  -1,   6,  -3,   0,  -3, -24,  -7,  -2,   0,   5,  -1,  -2, -14,  -2,  -1,   3,   2,   0,   7,   4,   7,   7, -11,   0,  -1,  -5,  15,  18,   9,   2,  -6,   0,   1 },
     {    0,   2,   1,  -5,   5, -12,  -2,  -1,   1,  -7,  -5,  11,   4,  -7,  -3,   3,  -2,  23,  -8, -19,  -8,  -2,  -1,   0, -23, -13,  42,  13,   8,   8,  -6,   0,  62, -26, -26,   2,  10,  -4,  -9,  -1, -21,   0,  -9,  -2,  12,  -4,   1,   4, -11,  58,  17,  -6,  12,  -4,   2,   2, -32,  29,  -1,  -6,   1,  -9,   1,  -4 },
     {   -1,  -2,  -2,  -2,   1,   7,   2,  -1,   0,  -1,   3,   8,  19,   9,   4,   1,  -2,  -3,  -8,   8,   6,  11,   3,  -1,  -4,   4,   0,  -4,  20,  14,  -3,  -2,  -3, -29,  -3,  -1,  20,  10, -10,   0,  -4, -15, -18, -29,  11,  -5,  -8,   3, -21, -26, -48, -52,  33,   6,  -2,   6, -20, -22, -44,   3,  41,   1,   0,   1 },
     {    1,   2,   2,   2,   3,  -1,   6,   1,   2,   1,  13,  -9,  15,  -4,  -7,  -2,   6,   1,   0,   9,   7,  -8,   2,   6,  -3,  21, -15,  39,  -3,   1,   3,  -3,   0,  10,  36,  -4,   7,  42,   5,  -4,  31, -13,  13,  -6,  59,  16, -13,  -8,  -7,   5,  18,  10,  37, -32, -17,   0,  -9,  16,   9,  -5,   0, -33,   1,   3 },
     {    2,   5,   1, -24,  72,  -2,  -1,  -7,  -1,  -7,   5,  -2, -33,  -7,  -3,  -1,   1,   6,  -4,  26, -45,  21, -11,   3,   0,  -3,  -4,  15,  31,  26, -21,  -1,  -9,  13, -17, -23,   4,  -4, -11,   1,   7,  -6,   4, -11,  -2,  -2,  -3,  10,  -2, -15,  26,  -9,   8,   9,  -1,   3,  20,  10,   9, -16,   7,   2,  -5,   0 },
     {   -1,  -2,  -4,   2,   5,  -2,   0,   1,  -1,   1,   3,  -3,  -6,  -6,  -2,  -1,   0,  -3,  -2, -13,  -3,   2,   1,   0,  -2,   6,  16,  15,   0,  -5,   2,   1, -18, -21, -26,  -6,  -8,  -3,   6,   1,  58,  39,   4,  -2, -12,  -4,  -1,   1, -34, -21,  16,  12,  -5,  -1,   1,   1, -69, -34,  25,  12,   2,   9,   3,   0 },
     {    0,   1,   4,   4, -19,   9,  13,   4,  -1,   4, -15,   8,  20,  -9,   4,   2,   2,  -2,  19, -23,   6,  -3,  -2,   3,  15, -22,  15,  -8,  11,   2,  10,   4, -33,  47, -35,  28,  -4,  -4,   9,   0,  33, -38,  -8,  -3, -16,  12,  -2,   2, -10,   9,  35, -33,  19,  14,  -5,  -1,   1,  29,  -2,  -8,  17,  -6,  -7,  -2 },
     {    3,   4,   5,   2,  12,  70,  11,  -1,  -3,  -3,  -3, -13,  -5,  -9,  -4,   0,  -4,  -8,  -2,   2, -23, -72, -15,   3,   1,   3,   1,   9,  14, -24,  21,   2,   7,  -5,  -2,  -4,  37,   1,   9, -13,  -7,  -5,  -3, -16,  -6,  11,   9,  -5, -12,   1,   1, -11, -16,   2,  13,   6,   2,  -6,   0,   9, -14,  -3,   6,   4 },
     {   -1,  -2,  -3,   3,  -3, -17,  -7,  -1,   2,   4,   8,  -8,   6,  11,  -1,  -5,  -1,  -2,  -6,  27,  -5, -15,  11,  10,   2,  16,   8, -26,  12,  20, -16,   4,   2, -15,   4,  37,  17,   3,  -8,   2,   6,  12,   9, -24, -15,  35,  11,  -2, -12,   5, -13,   9,   0,  24, -14, -10,  11,  25,  42,  51,  14, -18, -33, -15 },
   },
   {
     {   87, -49,  -1,  -2,   0,  -1,   0,   0, -71,  28,  10,   0,   1,   0,   1,   0,   9,  12, -11,   0,  -1,   0,  -1,   0,   1,  -9,   1,   3,   0,   0,   0,   0,   1,   2,   2,  -1,   0,   0,   0,   0,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -53,  -9,  19,   3,   4,   0,   1,   0, -32,  64, -17,  -5,  -3,  -1,  -1,   0,  65, -44, -16,  10,  -2,   2,  -1,   0, -14, -10,  25,  -3,   0,   0,   0,   0,   3,  10,  -6,  -6,   2,  -1,   0,   0,  -1,  -3,  -1,   3,   1,   0,   0,   0,   3,   0,   0,   0,  -1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 },
     {  -45,  40, -16,   0,   0,   0,   0,   0, -65, -10,  28,   7,   4,   1,   1,   0,   8,  53, -16, -12,  -2,  -2,  -1,   0,  39, -36, -20,  12,   0,   2,  -1,   1,   0, -10,  21,   0,  -2,   0,   0,   0,   4,   4,  -5,  -5,   1,   0,   0,   0,   0,  -1,   1,   2,   0,   0,   0,   0,   3,   0,  -1,  -1,  -1,   0,   0,   0 },
     {  -34, -65,  38,  10,   5,   2,   2,   0,  25,  41,  -3, -20,  -2,  -3,   0,  -1, -53,  37, -29,   8,   5,   0,   1,   0,  29, -11,  -4,  13,  -7,   1,  -1,   0,   6, -14,   9,  -6,  -2,   2,   0,   0,  -1,   3,   2,  -2,   3,   0,   0,   0,   0,   1,  -3,   0,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 },
     {   21,  -1, -14,   3,  -1,   1,  -1,   0,  46, -16,  17,  -8,  -3,  -2,   0,  -1,  27,   9, -46,   4,   1,   1,  -1,   0,  -3, -61,  37,  19,  -4,   3,   0,   1, -27,  38,  22, -30,   2,  -2,   1,  -1,  -1,   5, -27,   5,   7,  -1,   1,   0,  -4,  -2,   6,   7,  -3,   0,   0,   0,   0,   0,  -1,  -2,  -1,   1,   0,   0 },
     {   23,  40, -27,   4,  -4,  -1,  -1,   0,   1,  41, -37,   0,  -3,   0,  -1,   0, -37, -38,  18,  20,  -1,   4,   0,   1,  68, -27,  19, -13,  -6,   1,  -2,   0, -20, -12,   7, -10,   9,   0,   1,   0,  -4,  12,   0,   2,   2,  -3,   1,  -1,  -3,   0,  -4,   0,  -2,   0,   0,   0,   0,   0,   1,   1,   0,   1,   0,   0 },
     {  -23, -11, -32,  22,   4,   4,   0,   1, -27,  -6,  49, -14,  -5,  -1,   1,   0, -41, -11,   2,   5,   9,   2,   1,   0,   6,  24,  37, -13,  -7,  -4,  -1,  -1,  -7,  39, -46, -21,  11,  -1,   2,   0,  14, -22, -12,  34,  -1,   0,  -1,   0,   5,  -2,  16,  -8,  -9,   2,  -1,   0,   2,  -1,  -4,  -3,   4,   1,   0,   0 },
     {    1,  -8, -66,  30,   9,   4,   0,   1,  25,  38,  45, -11, -21,  -4,  -3,  -1,   9, -27,  -6, -34,  16,   4,   2,   0,  -3,  11, -29,  31,   9,  -8,   0,   0,   8, -23,  17,  10, -20,   3,   0,   0,  -6,   2,  11, -15,   5,   4,  -1,   0,  -2,   4,  -6,  -1,   4,  -3,   0,   0,  -2,  -1,   0,   3,   0,   0,   1,   0 },
     {   14,  50, -11, -12,  -2,  -2,   0,  -1,  -3,  22, -12,  -4,   1,  -2,   0,   0, -28,  17, -48,  21,   6,   0,   1,   0, -44,   6,  28,  19, -14,   2,  -1,   0,  63, -24,  -6, -15,  -3,   6,  -2,   1,  -9, -20,  11,   4,   8,  -1,   0,   0,  -1,   8,   4,  -4,  -2,  -2,   1,   0,   0,  -3,  -3,   1,   1,   0,   0,   0 },
     {   16,  12,  -4,   2,  -3,  -1,  -1,   0,  30,  15,   5, -16,   1,  -2,   1,   0,  41,  31, -28,   8,  -6,  -3,  -2,   0,  32,  -2, -15, -24,  -2,   1,   0,   0, -14, -12, -47,  31,  15,  -1,   1,   0,  14, -36,  39,  21, -24,   2,  -1,   0, -10,   3,  11, -30,   4,   5,   0,   0,  -4,   1,  -8,   6,   8,  -3,   0,   0 },
     {    6,  17,  19, -24,   7,  -3,   1,  -1,  -9,  -8,  25, -22,   4,  -2,   2,   0,   0, -26, -45,  42,   9,  -1,   0,   0,   7,  54, -20,  14, -24,  -1,  -1,   0, -52, -16,  24,  -9,  -1,  11,  -1,   1,  29,   8,   1,  -9,   9,  -3,  -3,   0,   6, -13,  -7,   4,   1,  -3,   1,   0,  -2,   1,   6,   2,  -2,   0,   1,   0 },
     {   -5, -28,  -1,  14,   0,   2,   0,   1,  -3, -51,  23,   0,  -3,   4,   1,   1,  27, -21,  -6,  20,   3,   0,  -1,   0,  32,   2,  34, -15, -19,   1,  -2,   0,  31, -48,   1, -18,  14,   5,   0,   1, -38,  24,  25,  -2,   7,  -8,   2,  -1,  -8,  26, -19, -10,  -1,  -1,   2,  -1,  -3,   0,  -7,   8,   1,   1,   0,   0 },
     {  -11,  -9, -10,  15,  -2,   0,   0,   0, -18, -19,   0,  -6,   9,   1,   2,   0, -32, -25,  -9,  27,  -5,   1,   0,   1, -32, -37,  16,   3,  10,   7,   1,   1, -13,  14,  23,  43,  -6,  -8,  -2,  -1,  23,   0,  46, -28, -34,   6,  -1,   1,  -5,  12, -10, -33,  23,   5,   0,   0,   1,   7,  -8,  12,   4,  -8,   0,  -1 },
     {   -8, -14, -15, -70,  28,  12,   5,   1,  11,  12,  27,  64,  -8, -21,  -5,  -2, -12,  -9, -21,  -2, -40,  17,   5,   2,   9,   4,  12, -19,  26,   8,  -7,   0,   0,  -6,   2,   7,   1, -15,   1,   1,  -1,  -2,   1,  -3,  -3,   4,   4,  -1,   1,   4,  -1,  -1,   1,   0,  -2,  -1,  -2,  -1,  -2,   0,   1,   1,   0,   1 },
     {    1,  -7,  13, -11,  -1,   2,   1,   0,  -4, -20, -21,  25,  -4,   0,   0,   0,   6, -12,   4, -14,  12,   6,   1,   0,  36,   5,  14,  45,  -7, -20,  -2,  -3,   1,  17, -14, -17, -52,  17,   4,   1,  16, -47,  23, -20,  27,  23,  -9,   1, -10,  17,  13, -12,  14, -18,  -1,   1,  -4,   4, -13,   1,   1,  -2,   6,  -1 },
     {   15,  31,  68, -18, -16,  -4,  -2,  -1,   0,   3,  49, -28, -12,   4,  -2,   0, -15, -38,  -5, -36,  18,  10,   1,   1,  15, -20,   0,   3,  28,  -7,  -2,   0,  20,   4,  -8,  17,  -2, -15,   1,   0, -13,   2,  -9,   6, -11,   3,   4,   0,  -7,  11,   0,  -4,   0,   4,  -2,  -1,  -1,   3,  -6,  -1,   3,  -1,  -1,   0 },
     {   -4, -10,   2,  -2,   2,   0,   1,   0,  -6, -13, -14,  15,  -2,   0,   0,   0,  -6, -19,  -4,  -4,   6,   2,   1,   0, -11, -16, -14,  30,  -2,   1,   0,   1, -17, -30,  12,   6,   3,   3,   1,   0, -12,   9,  14,  48,  -7,  -5,  -2,   0,  24,  -1,  61, -27, -40,   5,  -2,   0,  -3,   4, -12, -47,  20,   7,   0,   1 },
     {    0,   0,  -5, -13,  18,  -6,   0,   0,   2,  -1,   9, -25,  10,   0,   2,  -1,   0,  -9,  13,  41, -23,  -2,   0,  -1,   9,  -7, -47,   5,   0,   9,  -1,   1,  15,  44,  16, -12,  -1,  -3,  -3,  -1, -57, -37,  10,   5,   3,   1,   1,   2,  43,  24,  -5,  -1,   0,  -3,   0,   0, -12, -12,  -3,   1,  -4,   0,   0,   0 },
     {    1,  -8,   2,  12,   2,  -2,  -1,   0,   5,  -3,   9, -11,  -5,   4,   0,   0,   7, -12,  18,  22, -20,   3,  -2,  -1,   4,  -4, -16, -23,   6,   9,  -2,   1,  34, -17,  45, -16,  10,  -8,  -2,   0,  50, -43, -32,   1,   3,  -4,   4,   0, -42,  14,  31,   3,  -1,   1,   2,   0, -12,  29, -14, -11,  -1,  -2,   1,   0 },
     {    2,  17,  18,  -7,  -1,  -1,  -1,   0,   5,  32,  33,  -6, -11,   0,  -4,  -1,   1,  36,  52,  -7, -10,  -6,  -1,   0, -29,   5,  35, -11, -12,  -1,   1,  -2, -46, -19,  23, -25,  -1,   8,   2,  -1, -17,  -6,  22, -10,  13,   4,  -2,  -1,   6,  23,   4, -23,   7,  -7,  -1,   0, -11,  11, -27,   0,   7,  -4,   4,  -1 },
     {   -3,   1,  -4,  -6,   7,  -1,   1,   0,  -4,   4,  -8,  -3,   6,  -5,   2,  -1,  -3,   2,   0,   6,   1,  -1,   0,  -1, -10,   6, -14,   7,   3,  -4,   1,   0, -20,   5,   0,   2,  -4,   0,   1,  -1,  -8,   8,  -9,  21,   0,  -2,   1,  -1, -59,  68,  -3,  -1, -12,   2,  -1,  -1,  65, -44, -10,   4,  -4,   8,  -5,   2 },
     {    2,   4,   3,  -3,  70, -31,  -3,  -2,  -4,  -8, -11,  -3, -75,  26,  11,   3,   1,  13,  -2,  20,  31,   1, -12,  -1,  -2,  -5,   4, -23,  10, -19,   8,   2,   2,   5,   1,  19, -16,  10,   3,  -3,  -4,   9,  -9,  -2,   3,   0,  -5,   1,   2,  -3,   8,  -2,   0,  -1,   2,   0,   2,   5,  -4,   2,   1,  -4,   0,  -1 },
     {    0,   4,   6,  40, -29,  -3,   1,   1,  -7,  -5, -23,  33, -18,   4,  -2,   0,  -6,   3, -42, -25,  20,  12,  -1,   1,   4,  33,   4, -33,  23,  -3,  -2,   1, -21,   7,  41,  -4,   8, -15,   1,  -1, -19, -35,  -5,   7,  -9,   0,   8,  -1,  21,  27, -14,   1,  -3,   6,   1,  -1, -13,  -6,  -7,   4,   3,   2,  -3,   1 },
     {   -8,  -8,  -1,  -3,   0,   0,   3,   0, -10, -11,   1,   2,  -3,   7,   0,   0, -14, -14,  -3,   0,   6,  -2,   2,   1, -22, -27,  -6,   0,   1,   7,   2,  -1, -38, -39, -16,   7,   9,   5,   2,   1, -44, -50,  -2,  12,  16,   7,   1,   1, -30, -17,  15,  37,  24,  -2,  -1,   0,  -2,  14,  37,  23, -19,  -9,  -1,  -1 },
     {   -2,  -3,   5, -21,   9,  -4,   0,  -2,   1,   4,   1,  28, -24,  10,  -1,   1,  -5, -10,  10, -25,  30, -12,   1,  -1,  -6, -18, -22,  19, -41,  17,   5,   1,   4,  19,   1, -18,  51,   9, -13,   1,  23, -16,  27,   6,  -2, -39,   4,   1, -10,  10, -36,  -1, -16,  14,  11,  -2, -11,   9,   4,  16,   8,   3,  -8,  -1 },
     {  -10, -16, -34, -58,   2,  13,   7,   1,  -9, -14, -37, -68,  10,  22,   4,   2,   1,   3,  -7, -39,  11,   9,   1,   3,   2,  10,  14,  -6,  23,   2,  -2,  -1,  -4,  -5,  13,  -2,  13,  -7,   0,   0,   3,  -1,   2,   6,  12,  -4,   2,   0,   4,   9,   1,  -4,   2,  -3,   1,  -2, -13,  11, -19,  16,   4,   0,   2,  -2 },
     {   -4,  -8, -10, -27, -28,  17,   3,  -1,   7,  10,  13,  16,  15,   8,  -3,   2,   1,  -4,  -1,   5,  47, -41,  -2,   0,   6,  -2,   4, -31, -40,  16,  16,  -2,  10,  10,  13,  46, -15,  21, -11,  -3, -10,   1, -33,  -8,  10,  -5,  -3,   7,   9,  15,  30, -12,  17,  -6,   2,  -1,  -3,  -8,  -4,  -3,  -5,  -7,   3,  -4 },
     {    2,   2,   3,   7,   1,   0,  -4,   0,   3,  -1,   9, -11,   5,  -2,   0,  -1,  -2,   1,  -4,  -4,  -6,   0,  -3,   1,  -2, -13,  14, -21,  10,  -3,   8,  -1,  -4,   0,  -5,  28,   2,   6,  -7,   2,  25,  -6,  38, -17,  37, -19,  -5,   3,  21,  28,   2,  47, -40, -32,  13,  -3,  -2,  -5,  17, -35, -37,  27,   0,  -2 },
     {    0,  -1,  -2, -12,  10,  -5,   3,   2,   5,   7,   7,   1,  16,  15,  -3,  -4,  12,  17,  24,  12,  17, -13,  -3,  -2,  20,  28,  43,  35,   7, -17,  -4,  -2,   7,   8,  37,  22,  -6, -24,  -7,  -1, -10, -17,  11,  21, -45, -11,   6,   0, -21, -12,  -4,   8, -16,   8,   0,   1, -11,  -4,  43, -10, -21,   1,  -1,   1 },
     {   -1,   4,  -3,   3,  -2,  -5,   3,  -1,  -1,   7,  -2,   5,   5,   4,  -1,   2,  -2,   3, -16,  -2,   5,  -7,   1,   2,   9,  24,  -5,   0,   0,   8,   3,  -3,   1,  34,  -1,  -5,  19,  -9,  -8,  -2, -38,  15,  29, -34,  -4,  -5,   7,   2, -32, -16,  29,  21,   0,   2,   3,  -3,  35,  64, -33, -25,  -7,  -5,  -4,  -2 },
     {    6,   8,  12,  25,  58, -37,  -3,  -2,  -2,  -4,  -6,  -4,  29,  -4,  -7,  -2,  -4, -10, -12, -47, -36,  -7,  21,   1,   9,  12,  19,   9, -36,  24,  11,  -2,  -1,  -3,   7,  23,  13,  -1, -17,  -3,  -4,  -5, -17,  -4,  -5,   5,  -5,   7,   6,   4,  17,  -2,   5,  -6,   3,   1,  -5, -12, -13,  20, -10,   1,   7,  -3 },
     {   -2,   4,  -3,   8,   8, -12,   1,   2,  -1,   8,  -2,  17,  31,  -1,  -2,  -3,  -4,   4,  -8,  15,   8,  -7,   1,   3,  -7,   1, -15,  11,  22, -17,   4,  -2,  -9,   7, -21,   5, -11,   7,  -1,   0,  -5,  35, -15,  28,   4,  -5,  -3,   6, -23,  39, -12,  -2,  -5, -15,   6,  -2, -69,  45, -13,  20, -13,   7,   8,  -3 },
   },
   {
     {   87, -70,  17,  -7,   3,  -2,   1,   0, -49,  33,  -3,   0,   1,   0,   0,   0,   5,   0,  -5,   3,  -1,   1,   0,   0,  -5,   1,   2,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   67,  10, -41,   9,  -7,   2,  -3,   0,  34, -72,  35,  -2,   2,   1,   1,   0, -36,  29,   5,  -9,   4,  -2,   1,  -1,   1,   1,  -7,   3,   0,   0,   0,   0,  -7,   4,   2,  -1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {  -33,   3, -20,  21,  -4,   4,  -1,   1, -74,   2,  61, -21,   7,  -3,   3,  -1,  -4,  52, -33,  -4,   3,  -3,   0,  -1,  19, -18,  -4,   9,  -4,   2,  -1,   0,   5,   3,   1,  -4,   1,   0,   0,   0,   6,  -4,  -3,   2,   0,   0,   0,   0,   2,   1,   0,  -1,   0,   0,   0,   0,   2,  -1,  -1,   1,   0,   0,   0,   0 },
     {   36,  72, -63,  10, -11,   1,  -3,   0, -48,  19, -20,  16,  -3,   3,  -1,   1,  29, -35,  16,   2,  -1,   1,   0,   0,  -1,  -5,   9,  -5,   1,  -1,   0,   0,   2,  -4,   2,  -1,   0,   0,   0,   0,  -1,  -2,   2,  -1,   0,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 },
     {   21,  13, -11,  13,  -8,   1,  -2,   0,  46,  37, -10, -40,   8,  -5,   1,  -1,   0,   1, -68,  41,   0,   0,   0,   0,  11, -40,  36,  10, -15,   6,  -2,   1,  -8,   4,   3, -13,   6,   0,   0,   0,  -2,  -5,   3,   5,   0,  -1,   0,   0,  -3,   1,   1,  -2,   0,   0,   1,   0,  -1,  -1,   1,   1,   0,   0,   0,   0 },
     {   10,  39,  25, -43,  12,  -8,   2,  -2, -20,  14, -45,  25,  -1,   0,   1,   0, -72,  46,   0,   9,  -7,   3,  -1,   0,  31, -15,  -4,  -3,   0,   1,  -1,   0,  -1,  -7,   5,   1,   0,   0,   0,   0,   6,  -2,  -1,  -2,   0,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {  -20, -28,  -6,  27,  -8,   6,  -1,   1, -30, -40, -15,   9,  12,   2,   2,   0, -39, -30,  21,  55,  -8,   1,   0,   0,   9,   8,  59, -15, -23,   3,  -3,   0,   1,  15,   0, -26,   5,   1,   0,   0,   4,   5,   0,  -3,   2,   0,   0,   0,   1,   4,  -2,  -4,   2,   0,   0,   0,   2,   1,   0,  -1,   0,   0,   0,   0 },
     {  -13, -30, -41,  66, -28,  16,  -6,   3,  20,  51, -27,  10, -14,   3,  -4,   0, -44,  19,  21, -29,  10,  -4,   2,   0,   4,   5, -12,  -4,   9,  -3,   2,  -1,   0,  -7,   2,   4,   0,  -1,   0,   0,   2,  -2,   3,  -3,   1,   0,   0,   0,   1,  -2,   1,   0,   1,   0,   0,   0,   1,  -1,   1,  -1,   0,   0,   0,   0 },
     {    9,  34,  24,  -2, -12,   0,  -2,   0,  -8,  29,  28, -58,  17,  -6,   1,   0, -52, -25,  11,   2,  17,  -6,   6,  -1, -47,  51,  -1,   3,  -4,  -1,   1,  -1,   6,   4, -15,   0,  -1,   2,  -1,   1,   1,  -1,  -2,   2,   0,   0,   0,   0,   4,  -2,  -6,   1,   0,   0,   0,   0,   1,  -1,  -1,   2,   0,   0,   0,   0 },
     {    3,   4,   8,   4, -21,  10,  -4,   1,   3,   3,   5, -15,  28, -12,   3,  -1,   8,  14,  14,   8, -42,  12,  -2,   0,   7, -12,   6, -67,  44,   6,  -5,   2, -18,  20, -38,  39,  18, -25,   8,  -3,   5,  -5,   6,   6, -20,   8,   0,  -1,  -2,   1,  -2,   1,   6,  -2,  -1,   1,   1,  -1,   1,   1,  -3,   0,   1,   0 },
     {    7,   3,  26,   3, -16,   5,  -2,   0,  -1,   0,  12, -36,  16,  -1,   0,   1,  -6, -42,  44, -26,  23, -10,   4,   0,  75, -45, -12,  15,  -9,   1,   1,   0, -17,   2,   6,   1,  -2,   3,  -1,   0,  -1,   8,  -7,   2,   0,  -1,   0,   0,  -6,   2,  -1,   1,   0,   0,   1,   0,   0,   3,  -2,   1,   0,   0,   0,   0 },
     {    2,  -4,  -3,  12, -23,  10,  -3,   1,  -5, -20, -39, -23,  51, -10,   3,   0,  21,  15,  -5,  38,  -8, -13,   4,  -3,  21,  34, -51,  12, -21,   9,  -3,   0,  24, -43,   3,  18,  -5,   5,  -2,   2,  -1,  -2,  15,  -6,  -3,   2,   0,   0,  -1,  -4,   1,   2,   2,  -1,   0,   0,  -1,   1,   3,  -1,  -2,   0,   0,   0 },
     {  -15, -18, -21,  -9,   7,   3,   1,   1, -22, -28, -27, -10,  14,   2,   2,   1, -34, -43, -36,  -9,  16,  10,   3,   1, -31, -35, -17,  27,  40,   1,   0,   1,  -8,   2,  16,  46,   3, -13,   0,  -2,   3,   7,  17,  10, -15,  -3,  -1,   0,   4,   4,   8,   1,  -5,  -2,   0,   0,   2,   3,   3,  -1,  -4,   0,   0,   0 },
     {    9,  21,  55,  28, -64,  22, -10,   2, -15, -26, -30,  12,   1,  13,  -6,   4,  20,  24, -15, -20,  28, -19,   5,  -1, -33, -13,  14,   5,   1,  -6,   6,  -3, -19,  20,  10,  -6,   1,   1,  -1,  -1,   1,   4,  -6,  -4,   2,  -1,   0,   0,   1,   0,   1,   1,   1,   0,   0,   0,   2,   1,   0,  -1,  -2,   0,   0,   0 },
     {   -7, -10, -31, -19,  23,  -1,   1,   1,  -3,  -3, -36, -14,  20,   3,  -1,   0,  14,  16,  -8, -10,  42, -13,   0,   0,  23,  38,   3,  11,  -3, -18,   4,  -3, -33,  61, -41,  -3,  -7,   0,   1,  -1,  14, -12, -16,  18,  -4,   2,   0,   1,  -1,  -4,   4,   2,  -3,   2,   0,   0,   0,  -3,  -1,   3,   0,   0,   0,   0 },
     {    9,  10,  22,  22, -18,   2,  -2,   0,  -4, -13,   5,  14, -47,  14,  -1,   0,  -9, -32, -50,   0,   7,  27,  -4,   3,  52,  41, -12,  -2,  27,  -7,  -5,   0,  35,   4, -28,   2,   6, -12,  -1,   0,   3,  -9,  -2,  10,  -1,  -3,   1,   0,  -2,  -6,   3,   0,  -2,   0,  -1,   1,  -3,  -2,   3,   0,   1,  -1,   0,   1 },
     {    1,   2,   4,   0,  -7,  13,  -6,   0,   3,   4,  11,  10,   8, -20,   7,  -2,  -2,  -5,  -2,   6, -12,  16,  -5,   0,   8,   3,  14,  19,  16, -30,   4,   0,  -6,  16,   5,  22, -75,  19,   6,  -3,  16, -29,  21, -52,  20,  22, -13,   4,  -6,  12, -12,   8,  14, -15,   4,   0,   4,  -6,   3,  -6,  -3,   6,  -1,   0 },
     {    2,  -6,  -8, -16,   6,   3,  -1,  -1,   5,  -9, -16, -24,  20,   2,   0,   1,   8,  14,   8, -31,  30,  -5,   0,   0,   2, -13,  48, -19,  23,  -8,  -1,  -2,  88, -12,   6,  -3,  -9,  -9,  -2,   2,   5, -11, -11,  -6,  -1,   2,   0,   2,  12, -12, -11,   8,   4,  -1,   1,  -1,  -6,   2,   2,   1,  -3,  -1,   1,  -1 },
     {   -2,  -2,  -7, -10, -14,  22,  -9,   1,   2,   6,  -3,   4,  26, -34,  10,  -2,  -2,  -7, -11, -12, -36,  24,   2,  -2,   4,  17, -11,  14,  21,   5, -13,   4,   2,  26,  41, -34,  -2, -18,   4,   0, -11,  32, -63,   0,  11,   0,   2,  -1,  -2, -17,   1,  22,  -6,  -3,   3,   0,   3,   0,  -1,  -2,   0,   5,   1,  -1 },
     {    0,   2,  -2,  15,  16, -30,  12,  -3,  -3,  -1,  -2, -21, -15,  36, -15,   1,   4,   8,  -5,  27,  17, -15,  -6,   4,  10,  12,   1, -28,  27, -28,  17,  -4, -26, -18,  25,  16, -17,   1,   5,  -4, -13,  54, -27, -22,  -1,  10,  -3,   0,  10, -15, -18,  26,  10,  -6,  -4,   1,  -3,   6,   7,  -5, -10,  -1,   2,  -1 },
     {    6,   9,  10,  10,   3,  -5,   5,   0,  12,  18,  22,  23,  13,   1,  -6,  -3,   7,  12,  17,  25,   9, -11,   0,   1,  11,  16,  13,  24,   9,  -5,  -1,  -1,  20,  36,  41,  28,  -5, -24,  -7,  -1,  13,  18,  20,   8, -46, -20,   9,  -3,  -1,   3,   6, -23, -41,  15,   1,   0,  -9,  -2,   3, -27,  -2,  12,  -1,   1 },
     {    0,  -2,   1,  -1,   0,  -6,   5,   1,   0,  -2,   0, -16,  -5,  15,  -8,   0,   4,  -5,   2,   5,  -4,  -1,  -3,   5,  -1,   9,  -9, -15,  16,  -4,   3,  -2, -17, -12,  49, -19,  13,  -4,  -1,   0,  81, -52, -13,   1,  -6,  -1,   0,   1, -25, -11,  16,  -1,   8,   3,   0,   0,  -5,  24, -12,   4,   3,  -6,   1,  -1 },
     {    6,  10,  16,  47,  19, -60,  23,  -7,  -7, -13, -29, -34,   5,   8,  12,  -3,   6,  14,  12, -26, -36,  42, -21,   6,  -8,  -4,   6,  25,  -2,  -1,  -9,   6,   7,  12,  -8,  -6,  -8,   7,   5,   0,   3,   4,   1,   5,   1,  -1,  -1,   3,  -8,   8,   9,  -7,  -8,   2,   1,   1,   4,  -2,  -5, -14,  14,  -1,   2,   0 },
     {   -1,   3,   3,  -2,  13,  11,  -5,  -2,  -5,  -2,  -2,  -8,  14, -12,   8,   2,   4,   7, -10, -30,  14,  20,   0,  -3,  19,  40,  41, -12,  15,  -3,  -8,  -4, -41, -39,  32,  -8,   7, -13,   1,   2, -47, -15,  38,   3,   2,  -3,   2,  -1,  -6,  13,  21,  -7,  -9,   1,   0,  -2,  11,  20,  -4,  -9,  -1,   5,  -2,  -1 },
     {    0,  -1,  -3,  -3,   1,  -2,  -1,   0,  -1,  -2,  -5, -11,  -1,   7,   1,   1,   2,   1,  -1,  -3,  -4,  -2,  -3,   3,   4,   4,  -6, -13,   8,   8,   5,   3,   1,  14,   7,  -3,  28,  14, -20,   3,  14,  -3,  22, -15,  43, -54,  10,   2,  -8,  32, -29,  44, -36, -13,  22,  -9,   9, -21,  26,  -9, -25,  19,  -8,   0 },
     {    2,   2,   9,  15,   5,  -1,  -1,  -4,   1,   2,  11,  18,  12,  -5,  -5,  -2,  -3,   0,  11,  17,   2,  -2,  -1,   0,  -9, -14,   6,   9,   1, -21,  10,   2,  -2, -25, -20,  10, -18,   2,   7,  -4, -11, -33, -31,  52,  13,   4,   1,  -1,   2,  -7,  41,  48, -31,  -7,   2,   1, -10,  16,  29, -25, -23,   9,   1,   0 },
     {  -10, -13, -22, -48, -50,  21,  12,   3,  -1,   2,  -2, -30, -30,  45,  -3,  -2,   5,  14,  18,  11,  -6,  32, -12,   1,  -2,   4,  15,  15,  -4,  12, -11,  -4,  -4,  -8,  -7,  17,  -9,   7,  -7,   3,  -7,   6, -13,  -5,   4,  -1,  -5,   0, -20,  -5,  20,  -6, -17,  11,   0,   4,   3,  -3,  15, -27,  21,   5,  -4,  -3 },
     {    4,   2,   4,   7,   9,   7,  -7,   2,   2,   0,   1,  10,  -8, -30,  10,   5,   5,   1,  -7, -19,   5,   5,  -3,   2,   5,  13,  37,  11, -24,  27, -11,  -1, -13, -34, -19,  44,  -1,   4, -14,   6,  58,  40, -22,   2,  -7,  -8, -10,  -2,  -1,  -8,  -5,   6,   1,  -5,   4,  -1,  37,   8,   4,   1, -13,   0,   4,  -4 },
     {    2,  -2,   2,  -2,  -4,  -3,   4,  -1,   2,  -3,   2,  -6,  -1,   5,   3,  -2,   2,  -4,   0,  -6,  -9,   4,   1,   2,   6,   0,  -1,  -6,  -4,   4,  -1,   1, -13,   6,  13,  -7,   1,   8,   0,   1,  10, -14,  24,  -4,   2,   2,  -6,   5,  84, -65,  21,   0, -14,   3,   0,   0,  27, -36,  21,   3,   4,   5,  -2,   1 },
     {   -1,   0,   4,  11,  13,  -8,  -5,   2,  -5,  -7,  -4,  -1, -10, -15,  -6,   9,  -2,  -2,   1,   3,   8,  -3,  -7,   4,  -8,  -8,  -2,   5, -16,   1, -12,   9, -10,  -3,  -8,  14,  -2, -39,   5,   3, -24, -34, -17, -35,  -9, -23,  22,   3, -32, -51, -30,   1,  -1,  33,  20,  -3,   3,   8,  32,  14,  19,  27,  -6,  -4 },
     {    1,   2,   5,   8,  21,   9, -30,  23,  -5,  -6,  -7,  -7,  -9,   2,  21, -21,  -4,  -7, -13, -15,  -8, -21,  19,  -7,   1,   4,   2, -26,   1,  27,  17,  13,   3,   1,  13,  -2, -19,  44,   7, -13,  -2,   3,   0,  13, -29,  19,  -9,   8, -36,  -5,   3,  18, -14,  33,  -2,   3,  -4, -38,  10, -17,  27,   7,  -9,   3 },
     {   -1,  -1,  -3,  -4, -18, -23,  24, -10,   5,   8,  10,  20,  29,  33, -25,  -1,  -4,  -6, -10, -20, -22,  -7,  19,   2,   9,  14,  17,  15,  -6,  35,  10,  -7,   2,  11,   5,  11,  -8,   0, -15,   0, -10, -12,   3,  -5, -13,   8,  14,   3,   1,  14,   2,  42,  21,  32, -14,  -7,  12,  15,  23,  36,  -5, -27,  -9,   6 },
   }
 },
 {  //3
   {
     {  119, -40, -10,  -4,  -3,  -1,  -1,   0, -10,  -9,   2,   1,   1,   0,   0,   0, -13,   3,   2,   1,   1,   0,   0,   0,  -8,   1,   1,   0,   0,   0,   0,   0,  -6,   3,   1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   -9, -15,  10,   0,   2,   0,   1,   0,-115,  34,  16,   2,   4,   1,   1,   0,  13,  25,  -5,  -3,  -2,  -1,  -1,   0,  18,  -1,  -7,  -1,  -1,   0,   0,   0,  10,  -3,  -3,   0,   0,   0,   0,   0,   5,  -3,  -1,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {   -9,  25,  -5,   0,  -2,   0,   0,   0, -22, -35,  11,   3,   3,   1,   2,   0,-103,  10,  31,   3,   5,   1,   2,   0,  -2,  38,   5,  -6,  -2,  -2,  -1,  -1,  15,   9, -10,  -5,  -1,  -1,   0,   0,  10,  -1,  -5,   0,   0,   0,   0,   0,   5,  -2,  -2,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0 },
     {   30,  88, -35, -10,  -7,  -3,  -2,  -1, -13,  51, -38,  -2,  -2,   0,   0,   0, -14, -23, -15,  12,   3,   3,   1,   1,  -1, -24,   7,  10,   1,   1,   0,   0,   5, -11,   9,   2,  -1,   0,   0,   0,   2,  -6,   5,   0,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0 },
     {   12,  16,  -8,  -3,  -2,  -1,  -1,   0,  25,  -5,   2,  -5,   0,  -1,   0,   0,  -1,  62, -11, -11,  -2,  -3,  -1,  -1,  69,   9, -48,  -5,  -3,   0,  -2,   0,  10, -45, -20,  13,   4,   3,   1,   1,  -8, -22,  13,  12,   1,   1,   0,   0,  -8,  -3,  11,   1,  -1,  -1,   0,   0,  -3,   1,   3,  -1,  -1,   0,   0,   0 },
     {   -7, -22,  20,  -7,   4,   0,   1,   0,  27,  73,  -1, -23,  -3,  -5,  -1,  -1, -37,  51, -24, -15,   2,  -2,   1,   0, -56,   6, -14,   6,   7,   2,   1,   0, -11,   6,  -4,   8,   1,   0,   0,   0,   8,  -3,  -4,   4,  -2,   0,  -1,   0,   9,  -6,  -3,   3,  -1,   0,   0,   0,   3,  -5,   0,   2,   0,   0,   0,   0 },
     {  -12,  -7, -11,   8,   1,   2,   0,   1,  -8, -28,   9,   7,   3,   1,   1,   0, -31,  24, -15,   3,   0,   0,   0,   0,  -1, -64,  -8,  25,   2,   5,   1,   1, -37, -24,  55,  17,  -3,   1,   0,   0,   1,  35,  29, -18,  -8,  -3,  -1,  -1,   4,  17,  -7, -16,   0,   0,   0,   0,   3,   3,  -7,  -2,   2,   0,   0,   0 },
     {  -14, -27, -75,  40,   7,   4,   0,   1,   0,   9, -49,  43,  -1,   0,  -1,   0,   8,  38,  20,   9, -15,  -4,  -1,  -1, -12,  17,  18, -17, -10,   1,   0,   0,  -6,  -1,   5, -13,   1,   3,   0,   0,   1,  -5,   3,  -3,   3,   1,   0,   0,   3,  -4,   3,  -1,   1,   0,   0,   0,   1,  -3,   2,  -1,   0,   0,   0,   0 },
     {   10,  55, -11,  -9,  -5,  -1,  -2,   0, -14, -30,  26,  -5,   1,   1,   1,   0,  39,  34,  21, -25,  -3,  -4,  -1,  -1, -47,  19, -14, -22,   7,  -1,   2,   0, -57,   4, -12,   4,  13,   2,   2,   0,   0,  15,  -6,   9,   2,  -2,  -1,   0,   8,   1,  -6,   4,  -2,  -1,   0,   0,   4,  -3,  -3,   2,  -1,   0,   0,   0 },
     {  -10, -19,  -1,   3,   5,   0,   1,   0, -17, -11, -20,   9,   5,   2,   1,   1, -10, -42,   1,   7,   6,   3,   2,   0, -40,  -1, -28,   1,  10,   2,   2,   0, -14, -44, -32,  35,  14,   4,   1,   1, -10, -27,  36,  41,  -1,  -2,  -1,   0,   7,  11,  38,   2, -15,  -5,  -2,  -1,   3,  10,   9, -12,  -7,  -1,   0,   0 },
     {    2,  10, -14,   8,   0,  -1,  -1,   0,  -6, -44,  -2,  10,   7,   0,   1,   0,  19,  16, -51,   8,   7,  -2,   1,   0, -33,   7, -26,  37,   4,  -1,   0,   0,  46,  26,   8,  20, -17,  -4,  -3,  -1,  46, -19,  -7,  -5, -15,   3,  -1,   1,   2, -27,   8,   0,   0,   4,   0,   1,  -1,  -4,  11,  -2,   0,   0,   0,   0 },
     {    1,   6,  39,  -2,  -1,  -5,   0,  -1, -11, -15, -51,   4,   9,   1,   2,   1,  -1,   7, -54,  30,  10,   1,   1,   0,  14,  57,  11,  13, -12,  -7,  -2,  -1, -51,   8,  10, -10,  -9,   1,   1,   0, -28,   5,   9,  -6,   2,   4,   1,   0,   4,   7,   0,  -2,   2,   0,  -1,   0,   4,  -2,   0,   0,   0,   0,   0,   0 },
     {   -6,  -4, -15,   4,  -1,   3,  -1,   1,  -2, -14,   9,  -6,   3,   0,   1,   0, -17,  11, -34,   8,   5,   3,   0,   0,  20, -52,  23,  -2,   2,   2,   0,   0, -36,  36, -27, -24,  16,  -2,   2,   0,   2, -22, -41,  40,  16,   0,   1,   0,   1,  -6,  30,  29, -10,  -5,  -2,  -1,   8,   4,  13,  -6, -11,  -1,  -1,   0 },
     {  -13, -21, -45, -64,  23,  11,   4,   2,  -6, -10, -28, -60,  32,  10,   2,   1,   8,   7,  17,  -8,  28,  -3,  -3,  -1,  10,  10,  16,  20,  13, -11,  -3,  -1,   0,   6,   5,  18,  -4, -11,  -1,   0,  -3,   7,  -2,   8, -10,  -6,   1,   0,  -2,   3,  -6,   3,  -6,   0,   2,   0,  -1,   1,  -3,   4,  -2,   1,   1,   0 },
     {   -4,  -2, -30,  23,  -3,   2,  -1,   1,  -2,  12,   4,   8,  -6,  -1,  -1,   0, -11, -23, -21, -29,   8,   4,   3,   1,  17,  13, -50,  -8,  30,   2,   1,   0,   1,  63,  -3,  19,  11, -11,  -3,  -2, -43,  36,   8,   3, -11,  -9,   0,  -1, -23,  11,  -1,  -4,  -7,   1,   2,   1,   0,   0,  -3,   0,   0,   3,   0,   0 },
     {    3,  -7,  24,  -1,   1,  -4,   1,  -1,   0,  -4, -52,   6,  -2,   3,   1,   1,  -2,  -9,   5,  -3,  -2,   4,  -1,   1,  23, -13, -19, -29,  13,   5,   1,   1,  -7,   4, -19,  13,  28,   0,   0,   0,  68,  45,  -7,  11,  -3, -12,  -4,  -2,  31, -11, -27,   3,  -9,  -3,   0,   1,  -1, -17,   3,   9,  -2,   1,   1,   0 },
     {   -5,  -7,   2,   1,   2,   0,   0,   0,  -9,  -8, -16,   3,   6,   2,   0,   0,  -9, -12,   5,  -1,   3,   1,   1,   1, -14,  -8, -29,   0,   3,   2,   2,   0,  -8, -30,  11,   2,   2,   3,   1,   0, -26,   0, -33, -18,  19,   1,   1,   0, -42, -45, -32,  50,  24,   1,   1,   1,  -3,  -1,  39,  36, -12,  -6,  -2,  -1 },
     {    4,   3,   8,  50, -11,  -1,  -4,   0,  -5,  -9, -12, -18,  -8,   8,   0,   1,  -2,  -4, -17, -76,  20,  13,   1,   1,   6,  13,  47,  -4,  28,  -7,  -3,  -1,   3, -28,  26,  17,   5, -13,  -1,   0,   4, -22, -14,   8,  -4,  -3,   3,   0,   9,   6,  -6,   5,  -3,   0,   1,  -1,  -8,   5,   0,   3,  -1,  -1,  -1,   0 },
     {   -4, -13, -41,   3,   3,   5,   1,   1,   6,  22,  49,  -6,  -4,  -5,  -1,  -1,  -5, -42, -16,  10,   1,   2,   2,   0,  24,  35, -18,   3,  -5,  -1,  -1,   0, -50, -12,  17,   4,  -5,   3,   0,   0,  39, -24,  -5, -11,  -3,   6,   0,   0,  33, -13, -10,   0,   6,   2,  -1,   0, -10, -11,   9,  10,   3,  -1,  -1,   0 },
     {   -6, -10, -21, -20, -10,   5,   1,   2,  -8, -19, -15, -29,  -6,   2,   3,   1,  -6, -14, -43, -27,  -6,   3,   2,   2, -18,  -5, -25, -36, -10,  18,   3,   1,  17, -17, -10, -54,   9,  19,   3,   1,   2,   0,  -2, -27,  26,  13,  -1,   0,  11,  24,   3, -14,  25,   0,  -2,  -2,   0,  14, -21,   2,  19,  -4,   0,  -1 },
     {   -1,  -2,  15, -24,  -6,   7,  -1,   1,  -1,  -2, -12,  -5, -11,  11,  -2,   1,  -1,   0,  14, -13, -11,   7,  -1,   1,   7,  -1,  -6, -36,   5,   8,   0,   1, -10,  30,  54, -12,  16,  -2,  -3,  -1,   6, -36,  43,   1,  -1,  -8,  -4,   0, -11, -61,  26,  -9,  -3,   1,   3,   1,  18,  14,   7, -13,   5,  -1,   2,  -1 },
     {   -2,   0,   0,  10,   1,  -2,  -1,   0,  -3,  -5,  -1, -10,   6,   0,   2,  -2,  -4,   3,  -9,   9,  -2,   1,  -1,   1,   0, -15,  21, -17,   4,   1,   1,  -1, -19,  18, -39,  22,   2,  -1,  -1,   1,  24, -42,  38, -11, -11,   3,  -2,   1, -59,  13, -33, -24,  28,   0,   3,  -1,  -2, -12, -18,  40,   5,  -4,   1,  -1 },
     {   -6,  -8,  -7, -13, -63,  36,  -3,   2,  -4,  -3,  -4,  -6, -66,  44,  -6,   0,   2,   8,   5,  26, -11,  18,  -7,   1,  -2,   6,   1,  31,  19,  -7,  -5,   0,  -3,  -4, -11,  11,  19, -18,   0,  -1,  -4,  -2, -12,  -3,  13, -13,   2,   1,   0,   1,  -7,  -9,   7,  -4,   3,   2,  -2,   4,  -6,  -1,   6,  -4,   3,   0 },
     {    0,  -3,   9,  11, -13,  -8,   3,   1,  -4,   0, -31,  -6,  -7,  -3,   5,   0,  -1,  -7,  37, -30,   5,  -7,   2,   0,  11, -20, -24,  29,  -3,  -7,   3,  -2, -30,  16, -29,   4, -45,   8,   8,   0,   0, -16, -24, -29, -19,  32,   3,   3,  13, -12,  15, -21,  17,  21,  -7,   0,  26,  13,   1, -15,  20,  -4,  -4,  -1 },
     {   -1,   1,   4,  -4,   1,  -1,   3,  -1,  -1,  10,  -3,   2,  -3,   1,   2,   0,  -1,   1,   5,  -6,   2,   1,   1,  -1,   1,  14,   1,   4,  -1,  -3,  -1,  -1, -18,  -6,   4,  -7,   0,   3,   1,   0,  44,  26, -15,  -9,  -1,   0,  -1,  -1, -63,  15,  43,  -9,  -1,   0,   0,   1, -62,  38,  23, -24,   5,  -1,   2,   0 },
     {   -2,  -1,   6, -63,  28,  -7,   4,   0,   2,  -1,  -4,  76, -13,  -1,  -1,  -1,  -5,   2,  -7, -24, -10,   9,  -2,   2,   3,  -4,   3,  -2,  35,   2,  -3,   0,  -5,   3,   6,  11,  -7, -14,  -2,   0,  -5, -29, -22, -10,   2,   5,   4,   2,  11,  23,   0,   3,   9,  -1,   1,  -3, -18,  -1, -16,  11,   3,  -2,  -2,   0 },
     {   -1,  -2,   1,   4, -23, -11,   7,   4,   2,   3,   6,  10,  -5, -12,  -1,   2,   6,   8,  11,  20,  74,  -4, -14,  -2,  -3,  -5,  -6,  -9,  54, -12,  -9,   2,   0, -10,  -5, -51, -15, -13,   6,   7,   7,   5,  27, -11, -16,   3,   9,   1,   4,  -2,  18,  16,   1,   1,   1,  -4,  -5, -17,  -2,  17,  -5,   1,  -2,   0 },
     {   -4,  -4,  -3,  -8,  -2,   1,   1,   0,  -5,  -8,   2,   4,  -3,  -2,   0,  -1,  -7,  -4, -12,  -9,   1,   0,   3,  -3,  -6, -11,  10,  -2, -12,  -3,   4,   5,  -4, -15, -16,  -9,   1,   4,   3,  -3, -28,   7,  -9,  16, -20,  -3,   4,  -2,  -4, -51,  -2, -36, -19,  22,   5,   2, -53, -57, -10,  -8,  40,  22,  -1,   1 },
     {    4,   6,   1,  20,  69, -23,  -6,  -2,  -3,  -6,  -1, -24, -29,   1,   1,   1,   2,   0,   8,  -1, -37,  27,   5,  -1,   2,   6,  -1,  45,  28,   3, -15,   0,  -4,  -5, -16, -32,  20, -15,  -6,   5,   2,   9,  18, -13,   6,  -6,   5,   2,   2, -16,  19,   6,   8,  -5,   2,  -3,   3, -10, -11,  11,   5,  -1,   0,  -2 },
     {    1,   0,   7,  11,   4,  12,  -1,  -6,  -2,   1, -14,   2,   5,  -9,  -3,   0,   1,  -7,  17,  -1,  11,   1,   3,  -5,  -5,   0, -38,  20, -13, -15,   1,   2,  -9,  -2,  23, -31,   5,   3, -10,  -1,  15, -35, -27,  11,   2, -34,  14,  -4, -27,  20, -22,  -1, -42,  -5,  24,  -3,  12,   0, -48,   0,   2,  31,   2,  -1 },
     {    0,   0,   2,   0,   2,  -1,  -9,   1,  -2,   0,  -7,   2, -18,  13,  -6,  -2,   1,   0,  14, -26,  22,   4,   4,   0,   3,  -5,  -8,  18, -23, -16,   3,  -2,  -1,  22,   6,   4, -30,  36,   2,  -3,   5,  -2,  25,  11,  58,   1, -25,  -1,   4,   3,   6,   8,   7, -45, -13,   3, -30, -41,  -7,   4, -18,  -7,  10,   5 },
     {   -6,  -8,  -9, -15, -36, -48,  25,   1,  -6,  -8,  -9, -23, -38, -42,  22,   5,  -1,   0,   1,   0, -40, -13,  10,   3,   2,   5,   6,  -1,   3,   9,  14,  -2,   1,   4,  13,  10, -14,  11,   7,  -9,   3,  -2,   8,   4, -14,   6,   4,  -9,  -7,  16,   9,  43, -13,   9,   7,  -4,  -8, -20,  -8,   8, -14,  24,  -2,  -1 },
   },
   {
     {   89, -52,   3,  -4,   0,  -1,   0,   0, -67,  29,   6,   0,   1,   0,   1,   0,  15,   6,  -9,   1,  -1,   0,  -1,   0,  -4,  -6,   2,   2,   0,   0,   0,   0,   1,   3,   1,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   61,  -5, -16,   0,  -3,   0,  -1,   0,  45, -55,  15,  -1,   2,   0,   1,   0, -70,  37,   9,  -4,   2,  -1,   1,   0,  16,   6, -13,   2,  -1,   0,  -1,   0,  -9,  -5,   4,   2,   0,   0,   0,   0,   2,   4,   0,  -1,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 },
     {  -30,  36, -18,   5,  -3,   1,  -1,   1, -65, -15,  35,  -2,   5,   0,   2,   0,  -6,  61, -21,  -6,  -2,  -2,  -1,   0,  44, -36, -13,   9,  -2,   2,  -1,   0,  -7,  -2,  13,  -2,   0,   0,   0,   0,   8,   0,  -5,  -1,   0,   0,   0,   0,  -1,  -1,   2,   1,   0,   0,   0,   0,   4,  -1,  -1,   0,   0,   0,   0,   0 },
     {   40,  70, -57,   5, -10,   0,  -3,   0, -11, -39,  15,   8,   0,   3,   0,   1,  47, -31,  14,  -9,   1,  -1,   0,   0, -32,   5,   5,  -2,   3,  -1,   1,   0,   2,   8,  -3,   0,  -1,   0,   0,   0,  -3,  -3,  -1,   1,   0,   0,   0,   0,  -1,   1,   3,   0,   0,   0,   0,   0,  -1,   0,  -1,  -1,   0,   0,   0,   0 },
     {   12, -13,  -5,   6,  -1,   1,  -1,   0,  50,  -1,   9, -14,   0,  -3,   0,  -1,  32,   8, -51,  14,  -4,   2,  -2,   1,  -2, -67,  40,   9,  -1,   3,   0,   1, -29,  39,  10, -17,   3,  -3,   1,  -1,   4,  -5, -13,   5,   0,   1,   0,   0,  -8,   3,   7,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0 },
     {   27,  42, -16,  -6,  -4,  -2,  -1,  -1,  10,  49, -56,  14,  -6,   2,  -2,   0, -14, -26,  15,   7,  -3,   3,  -1,   1,  62, -35,  -2,  -4,   0,   0,   0,   0, -20, -22,  17,   2,   2,   0,   1,   0,  -4,  11,   5,  -6,   1,  -1,   0,   0,  -4,  -4,  -3,   2,   1,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0 },
     {   -8,  10,  -9,   9,  -3,   2,  -1,   0, -33,   3,   4,  -5,   4,  -1,   1,   0, -57, -30,  12,  26,  -1,   4,   0,   1,   9,   8,  57, -28,  -1,  -3,   0,  -1, -13,  49, -40, -10,   8,  -3,   1,  -1,  20, -21,  -2,  17,  -5,   2,  -1,   0,  -1,   7,   6,  -8,  -1,  -1,   0,   0,   5,  -2,  -5,   1,   1,   0,   0,   0 },
     {   12,  27, -28,  15,  -7,   1,  -2,   0,  22,  54,  15, -36,   1,  -6,   0,  -1, -11,  24, -48,  14,   1,   0,   0,   0, -10,  18,  -2,  -2,  -6,   1,  -1,   0,  40, -30, -34,   9,   1,   2,   0,   0,  -9, -32,  25,  11,  -2,   2,   0,   0,   1,   6,   1,  -8,   1,  -1,   0,   0,  -2,  -8,   1,   4,   1,   1,   0,   0 },
     {  -17, -44, -55,  57,  -9,  10,  -2,   2,   9,  19,  44, -26,  -6,  -1,  -1,   0,   3, -37,  21, -21,  16,  -3,   3,   0,  31,   8, -15,  10,  -2,  -3,   0,  -1, -20,  -3,  10,  -2,  -3,   2,   1,   0,   0,  11,  -1,  -3,   2,   0,   0,   0,  -2,   2,  -4,   0,   1,  -1,   0,   0,   0,   1,   1,   2,   0,   0,   0,   0 },
     {    6, -10,  15,  -5,   0,   0,   0,   0,  14, -15,  -1,  -1,   3,   0,   0,   0,  51,  19,  15,  -7, -10,  -1,  -2,   0,  45,  12, -21, -37,   9,  -4,   2,  -1, -38,   3, -54,  32,   8,   0,   1,   0,  17, -31,  24,  15, -12,   4,  -2,   1, -13,   4,   5, -11,   4,   0,   1,   0,   0,  -1,  -1,   4,   0,   0,   0,   0 },
     {   -6, -18,   3,   9,  -1,   1,   0,   0, -13, -49,   1,   0,  11,   1,   2,   0,   6, -29, -20,  42,  -6,   4,  -2,   1,  19, -13,  30,  -6,  -7,   3,  -2,   0,  15, -44,  10,  14,   4,   1,   0,   0, -22,  17,  57, -14,  -7,  -2,   0,  -1,  -2,  32, -17, -22,   5,  -2,   1,  -1,  -2,  -1,  -5,   6,   1,   0,   0,   0 },
     {   -5,   5,  -5,   5,  -1,   1,   0,   0,  -7,  15,  -7,  -1,   4,  -1,   1,   0, -24,  -4,  -3,  -4,   1,   2,   1,   0, -49, -21, -24,  17,  19,  -1,   3,   0, -34,  15,  20,  52, -17,  -1,  -1,   0,  49, -11,  42, -27, -20,   5,  -4,   1, -10,  12,  -8, -19,  13,  -1,   1,   0,   7,   2,  -3,   5,  -1,  -2,   0,   0 },
     {    9,  21,  22, -11,   0,  -4,  -1,  -1,  -5,  -1,  17, -33,  11,  -3,   2,   0,   5, -19, -37,  39,   1,  -1,   1,   0,   4,  51,  -9,   0, -20,   2,  -3,   0, -54, -22,  19,  -9,  11,   3,   1,   1,  38,  27, -17,   1,   0,  -5,   0,  -1,  -4, -24,  -6,   8,  -3,   2,   0,   0,   1,   4,   9,   0,  -1,   0,   0,   0 },
     {    2,   9,   3,   4,  -8,   2,  -1,   0,   5,  16,  14, -16,   8,  -4,   1,  -1,   3,  25,  16,   0, -13,   0,  -2,   0, -12,  -3,  17, -44,  15,  -1,   2,   0,  -7,  15, -13,  14,   1,  -5,   1,  -1, -25,  45,  -7, -37,   1,   1,   0,  -1,   5,  -3, -66,  25,  18,  -1,   1,   0,   7, -21,  20,  23, -12,   1,  -1,   0 },
     {   13,  25,  50, -38,   4,  -7,   2,  -1,   3,   3,  52, -36,   4,  -5,   1,  -1, -11, -48,  -2, -24,  16,   2,   3,   0,  24, -20, -20,  21,  14,  -4,   1,  -1,  16,  20, -10,  14, -15,  -3,   0,   0, -25,  -2,   1,  -2,  -1,   5,   1,   0,  -3,  16,  -1,   0,   3,  -2,  -1,   0,   1,   0,  -6,   2,  -1,  -1,   0,   0 },
     {   -3,   5,   0,  -5,   3,   0,   1,   0,  -2,  12,  10,   1,  -5,   0,  -1,   0, -10,   8,   4, -17,  10,  -3,   1,   0, -26, -17,   2,  -6,   4,   1,   2,   1, -35, -15, -16,  14,  27,  -5,   4,   0, -29,  54,  13,  51, -23,  -9,   0,  -2,  42,  -3,  15, -31, -27,  10,  -4,   1,  -5,  -5, -14, -13,  16,   3,   1,   0 },
     {    7,  11,  34,  52, -50,   8,  -7,   2,  -9, -14, -39, -54,  45,  -2,   4,   1,   5,   8,  28,   9,   4, -14,   2,  -3,  -9, -15, -17,  15, -15,   8,  -1,   1,   6,  11,   0,  -6,   7,  -1,  -1,   0, -10,  -6,  -1,   9,  -3,   1,   0,   0,  10,  -1,   5,   1,  -3,   0,   0,   0,  -2,   2,   1,  -6,  -1,   0,   0,   0 },
     {    7,  16,  26,  -1, -11,  -2,  -2,   0,   9,  24,  29,  -7, -14,   1,  -1,   0,   8,  24,  44, -17,  -8,  -3,   0,  -2, -11,   6,  32, -12, -17,   2,  -1,   0, -15, -24,  25, -41,   7,   5,   0,   1,  16, -12,  11, -16,  13,  -3,   2,  -1, -18,  35,  -4, -31,   8,  -4,   2,  -1, -13,  13, -42,   1,  13,  -1,   2,   0 },
     {    8,   3,  -5,  -1,  -1,  -3,   0,  -1,  13,   4, -10,  -4,   0,   0,   0,   0,  19,  10,  -7,  -2,  -1,  -2,  -1,   0,  34,  27,   0,   4,  -3,  -3,  -2,   0,  50,  47,  10,   0,  -8,  -6,  -1,  -1,  49,  47,   8, -10, -20,  -2,  -2,  -1,  38,  13,  -5, -30,  -6,   2,  -2,   0,   8,  -5, -25,  -9,  10,  -1,   2,   1 },
     {    0,  -7,  -1,  -2,  -3,   4,   2,   1,   5,   2,  21, -18,  10,  -6,   1,   0,  -2,  -8,  19,  12, -31,  10,  -1,   1,  -4, -29,   8, -39,  27,   6,  -1,   1,  39, -29,  11,  25,   8, -15,   3,  -2,  29,  11, -29,   0, -24,   9,   0,   0, -12, -37,  31,  -1,   9,   4,  -2,   1, -21,  39,   1, -12,   3,  -3,  -1,  -1 },
     {   -6,  -5, -24, -23,  19,  -2,   3,  -2,   3,  12,  11,  -6,  18, -12,   3,  -1,   0,   1,  20,  54, -49,   6,  -3,   0,  -4,   5, -50, -10,   7,  13,  -4,   2,  -7,  32,  31, -18,  15,  -9,  -3,   0, -33, -20,   4,   9,  -4,  -3,   3,   0,  15,  26,  -1,  -4,  -3,   0,   1,   0,  -3,  -2, -11,  -5,   0,   1,   0,   0 },
     {  -10, -16, -31, -41,  -2,   7,   4,   1, -12, -19, -40, -54,  -2,  12,   5,   2,  -9, -12, -25, -45,  -1,  11,   4,   1,  -8, -11, -18, -32,   4,   8,   4,   1,  -2,  -9,  -7, -26,  18,   7,   3,   0,  15,   2,  -6,  -2,  18,   1,   1,  -1,   9,  27,  -9,   1,  12,  -3,   0,  -1,  -4,  13,  -9,  13,   9,  -4,   2,  -1 },
     {    1,   1,   5,  -4,  -2,   1,  -1,   0,   1,  -3,   9,   3, -12,   6,  -3,   1,  -1,  -3,   1,  -4,  -1,   3,   4,  -1,   7,  -4,   6,   5,  -6,   2,  -1,  -1,  -9,   8,   1, -15,  -6,   4,   0,   1,   3, -30,  33, -18,   5,   3,   1,   1,  73, -40, -18,   2,  11,  -1,  -4,   0, -67,  23,  24,  -1,   4,  -5,   2,   0 },
     {    5,  10,  28,  40, -44,  13,  -6,   3,  -3,  -4,   1,  30, -30,   8,  -9,   2,  -9, -14, -48, -18, -13,  12,   0,   2,   4,  18, -20, -38,  42,  -6,  -1,   1,  -5,   8,  27, -10,  10, -19,   5,  -2, -18, -13,  -6,   2,  -9,   3,   3,  -1,   7,  10,   0,  -4,   3,   4,  -1,   1,   8,   9, -11,  -3,   3,  -1,   0,   0 },
     {    5,   1,  15,  19,  -5,  -4,  -1,  -1,   4,  -2,  20,  30, -22,   1,  -5,   0,  -5, -21,   3,  21, -11,  -4,  -1,   0, -13, -46, -48,  -6, -23,  11,   2,   2,  18, -15, -44, -27,  21,  15,  -4,   3,  44,   8,  -8,  -1,  17, -13,   0,  -1,  13,   5, -24,   2,  -6,   1,   3,  -1,  15, -12,  -1,   6,   2,   6,  -1,   0 },
     {   -1,  -1,  -3,   2,   7,  -5,   3,  -1,  -4,  -2,  -7,   8,  -5,   4,  -1,   0,  -1,  -3,  -5,   4,  16, -10,   2,   0,   1,   2, -15,  12, -19,  15,  -3,   0,   6,  15,  -5,  -9,  21,  -8,  -2,   1, -22,   0, -11,   1, -33,   6,   2,  -1, -28, -41, -14, -43,  46,  17,  -2,   4,  -6,   9, -19,  61,  24, -19,   3,  -1 },
     {    4,   7,   8,  29,  45, -42,   2,  -3,  -6,  -9, -12, -37, -50,  43,  -1,   1,   3,   9,   9,  32,  27, -11, -15,   4,  -4,  -8,  -9, -22,  20, -21,  17,  -3,   1,   2,  12,   8, -22,  12,  -3,  -1,  -7,  -4, -14,   5,  10,  -3,  -4,   1,   9,   4,   6,  -7,  -1,   1,   4,   0,  -6,  -2,  -7,   4,   0,  -4,   0,   0 },
     {    2,   0,   4,  -5,  12,  -2,  -1,   1,   2,  -1,   5,  -4,   0,  10,  -5,   1,   3,  -1,  11, -11,  31, -13,   5,  -3,   3,  -3,   7,  -7,   7,   9,   0,   0,  21,   2,  38,  -5,  45, -28,  -1,  -2,  26, -24,  16,  38, -28, -19,   9,  -4, -19,  12, -35,  17, -26,  17,  -1,   0, -15, -23,  40,   7,  -7,   2,  -4,   2 },
     {   -2,   1,  -3, -10, -11, -10,  -2,   4,   1,   7,   6,  -5, -21,  15,  -3,   1,  -1,   6,  -1,   0,  12, -13,   2,  -4,  -3,   0,  -8, -22, -31,   9,   7,   0,   7,  32,   2,  -3,  21,  10,  -8,   0, -18,  15,  47, -19,   7, -18,   2,  -5, -24, -19,  17,  10,  -9,   9,   5,  -2,  37,  62,  23, -27,  -4,  -4,  -4,  -1 },
     {   -5,  -8,  -9, -34, -41,  42,  -7,   1,   4,   5,   9,  15,   0,   2,  -6,   0,   2,   1,   0,  30,  53, -48,   8,  -3,  -1,  -6,  -8, -37, -14,  13,   7,   1,   2,   1,  15,  23, -25,  16,  -4,  -1,  -2,  -8, -25,   4,   9,   0,  -4,   1,  11,  13,   5, -14,   3,  -5,   1,   3, -13,  -8,  -3,   1,  -1,   0,   3,  -1 },
     {    3,   6,   2,  13,  13,  -2,   2,  -1,   4,   9,   2,  21,  20,   7,  -3,   3,   2,   7,  -3,  10,  26,  -5,   2,   1,   0,   6,  -3,  13,  16,   3,  -3,   5,   0,   4, -10,   5,  27, -13,   4,   0,   8,  23,   4,  19,  17,   3,  -3,  -2,   9,  35,  12,  32,  27, -35,  -5,  -2, -21,  49, -18,  30, -48, -20,  11,  -2 },
     {   -3,   1,  -4,  -1,  -6,  23, -12,   0,  -6,  -2, -11,  -4, -19, -16,  12,   2,  -1,   7,  -3,  18,  14,  11,  -4,   0,  -3,   1, -19,   6,  28,   0, -15,  -3,   2,  17, -23, -23, -21, -13,   9,   8,  -2,  31,  19,  -1,  15,  -9,   2,  -1, -55, -10,  -9,  -5, -31,  -1,   6,  -2, -63,   2, -12,  -9,   0,  17,  -6,   7 },
   },
   {
     {  -92,  55,   0,   3,   0,   1,   0,   0,  59, -11, -21,   2,  -3,   1,  -1,   0,  -1, -23,   9,   6,  -1,   1,   0,   0,   3,   3,   8,  -5,  -1,   0,   0,   0,  -1,  -2,  -2,  -2,   2,   0,   0,   0,   1,   1,   1,   1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -29, -26,  25,   2,   4,  -1,   2,   0, -29,  84, -29, -11,  -2,  -2,  -1,   0,  31, -33, -34,  27,  -2,   3,  -1,   1,   2, -13,  30,   4,  -9,   2,  -1,   0,   4,   0,  -2, -12,   2,   1,   0,   0,  -1,  -2,   2,   3,   2,   0,   1,   0,   2,   0,  -1,  -1,  -1,   1,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   1 },
     {  -44,   2, -12,   6,   4,   2,   1,   1, -10,  10,  48, -14,   0,  -5,   1,  -1, -12,  54, -59, -18,  12,  -2,   3,  -1,  11, -20, -21,  50,  -3,  -1,  -1,   0,   6,  -3,  18,  -3, -22,   3,   0,   0,   3,  -6,  -2,  -6,   3,   8,  -1,   0,   2,   0,   3,   0,   1,  -1,  -2,   0,   1,  -1,  -1,  -1,   0,   0,   0,   0 },
     {   46,  22, -30,  13,  -3,   0,  -2,   0,  38, -12,  -4, -18,  -5,   0,   1,   0, -25,  -8, -37,  39,   9,   2,   0,   0,  13, -23,  47,  15, -36,   4,  -3,   1,  -5,  -1,  11, -43,  10,  12,  -1,   1,  -2,   1,  -2,   6,  16,  -7,  -2,   0,  -3,   2,   1,  -2,  -3,  -3,   2,   0,   0,   0,   0,   1,   1,   0,   1,   0 },
     {  -33, -17,   6,  18,  -1,   0,  -1,   1, -27, -23,  38, -21,   5,   5,   2,   1,   4,  30,   5,  16,  -7,  -6,  -4,   0,  19, -18,  20, -50, -11,  17,  -1,   2,  -4,   8, -15, -13,  60, -11,  -4,  -1,   0,   9,  -8,  26, -12, -27,   8,  -1,  -1,   2,  -5,  -2,  -9,   9,   5,  -1,   1,   0,   0,   3,   1,   0,  -1,   0 },
     {   -5, -80,  35,   9,   5,   3,   1,   1,  72,  -1,  22, -26,  -5,  -3,  -1,  -1, -22, -11,  -5, -12,  16,   1,   2,   0, -13,  22,   5,   4,   7,  -8,   1,  -1,  -5,  10, -15,  -1,  -3,  -3,   3,   0,  -4,   3,  -3,   7,   0,   2,   1,  -1,  -1,   2,  -2,   0,  -2,   0,  -1,   0,  -1,   1,  -1,   1,   0,   0,   0,   0 },
     {  -21, -22,   9,  -1,   6,   8,   1,   1, -22, -15,  10,  19, -14,  -3,  -2,   1,   6,  -3,  29, -11,   4,  11,  -1,   0,   3,  17,   2,   3, -26, -10,  -2,   0,   8, -16,   7, -42,  -5,  48,  -3,   2,   7,   0,   8, -10,  63, -19, -26,   3,  -1,   1,  -5,   4,  -7, -36,  18,   5,  -1,   2,  -2,   2,   0,   7,   9,  -6 },
     {   -1,   4,  51, -14,  -1,  -6,   0,  -1,  -7,   2, -46, -18,  19,   0,   5,   0, -58,  42,  -1,  41,   1, -11,   1,  -2,  14,  22, -28,  -2, -30,   5,   2,   0,   4, -22,  -4,  15,   3,  16,  -4,   1,   8,  -2,  13,  -3,  -4,  -2,  -6,   1,   1,  -6,  -1,  -3,   2,   1,   1,   2,   2,   0,   3,   0,   0,   0,   0,   0 },
     {   29,  36,  15, -15,  -3,   0,  -2,  -1,  39,  47,   7, -23, -12,  -3,  -3,   0,  21,  15, -14, -38,  -3,   5,  -2,   0,  -9, -19, -39, -24,   8,   9,   2,   2,  -5, -18,  -8,   6,  36,  17,  -5,   1,   0,  -2,   5,  15,  22, -27, -11,   2,  -4,  -3,  -4,   5,  -9, -15,  14,   3,  -1,   0,   0,   4,  -1,   4,   5,  -4 },
     {    0,  -2,   4, -11,  20, -13,   1,  -1,  -2,   3, -12,  20, -24,  14,   1,   1,   2, -10,  13, -20,  34, -18,   0,   0, -11,  17, -20,  36, -29,  10,   4,  -1,  19, -20,  33, -27,  28, -12, -13,   3, -15,  19, -24,  16, -19, -29,  29,  -3,   2,  -2,   2,   2, -17,  38,  -6,  -6,   0,   1,   1,  -2,   7,  -5,  -9,   4 },
     {    2,  28, -13,  17,   4,  -5,  -2,  -1, -24,  -5,  21, -44,  -2,   4,   4,   1,  19, -20,  -6,  10,  45, -14,   0,  -2, -34,  46,  -5,   8, -27, -23,  12,  -2,  -3,  12, -43,  16,  -6,  25,   3,  -3,   7,  -8,  11,  23, -11,   1, -12,   1,   4,  -4,   3, -10,  -6,   5,   1,   3,   2,  -2,   2,   2,   2,   1,  -2,   0 },
     {   -8,  -6, -63,  21,   7,   5,   2,   0,  -4,  51,   3,  24, -19,  -6,  -3,  -1, -67,   1,  10, -11, -15,  12,   3,   2,   5,  33,   0, -12,   8,   8,  -6,   0,  10,  -8, -27,   0,  10,  -3,  -4,   2,   6, -10,   7,  18,  -1,  -3,   1,   1,   4,  -3,   5,  -4,  -9,   1,   0,   0,   2,  -3,   1,  -1,   2,   3,   0,   0 },
     {   -8, -16, -20,  -6,   9,   9,  -1,   1, -13, -24, -27, -10,  18,  -2,   1,   2, -16, -25, -21,   4,  24,  15,  -2,   2,  -7, -13,   1,  22,  38,  -1,  -3,   1,  10,  12,  25,  41,  23,  -4, -15,  -2,  11,   3,  23,   5,  -2, -52, -15,   5,  -3,   0,  -5,  -3, -26, -20,  29,   4,   1,   2,   1,   2,  -6,  11,   7,  -7 },
     {    6,  39,  44, -28,  -5,   3,  -1,  -1, -30,   3,  18, -25,   0,  -1,  -2,   1, -45, -23,   4, -26,  21,  17,   1,   2,  23,  10,  28,   5,  35, -17, -12,  -1,   3,   8, -11, -25,  -5, -28,   9,   2,  -5,   4, -23,   6,  14,   1,  11,  -2,   0,   4,  -5,  14,  -6,  -5,   1,  -3,  -1,   0,  -4,   2,  -4,   3,   0,   0 },
     {   -3,   4,   6, -62,  14,   2,   6,  -1,   3, -13,   9,  33,  25, -20,  -2,  -2,   4,   8, -34,  11, -33,  -1,  12,   0, -52,  25,  11,   6,  -3,  26,  -4,  -3,   4,  28, -22, -15,   7,  -2, -11,   3,   3, -15, -14,  22,   8,  -2,   1,   3,   5,  -2,  12,   3, -10,  -4,  -1,   0,   1,  -3,  -1,  -2,   0,   3,   1,   1 },
     {    0,  -4,   5, -22,  34,  -3,  -4,  -1,   4,   0, -25,  34, -30,  -9,  10,   0,   2,  30,   2,  -7,  18,  18, -14,   0,  23, -30,  12,   0,  -2, -46,  11,   3, -35,  30, -21,   5,  -9,  20,  20,  -9,   5,   6, -10,  33, -14,   0, -16,   0,   0,  -7,   8, -12, -20,  13,   3,   4,   2,  -1,   1,  -2,   9,   5,  -5,   0 },
     {   -2,  -5,  -3,  -3,   5,  -9,   4,  -2,  -5,  -9, -14,   0,  -4,  17, -10,   2,   4,   0,  -9, -10,  16,  -1,   8,   0,   3,   5,  -1,   4,   4,  15, -19,   0,  -3, -14,   6,  -4,  20, -17,  16,   0,  -6,  25,   0,  33,  -5,  28, -23,  -9,  18, -26,  29, -31,  25, -46, -24,  23, -14,  12, -15,  -2,  -6, -27,  45,  -4 },
     {    6, -10,  -1,  -2,  -2,  -1,   6,   1,   7, -24,  -3,  10,   7,  -4,  -1,  -1,  33, -12, -22,   8,  -2,   7,   1,   2,  64,  18, -13,  10,   9,  -1, -10,  -4,  15, -45, -34,   3,  -6, -11,  -1,   5, -18, -39,  15,  30,   8,  -3,  14,  -1,  -6,  -1,  26,  14, -20,   3,   0,  -7,  -5,   0,  17, -12,  -2,   5,  -3,   1 },
     {   11,  17,  47,  40,  -8, -18,   3,  -4,  -2,  -6,  30,  29, -23,  -3,   5,  -3, -13, -36, -16,  -6, -31,   4,   4,   6,  18,   0,  -4,   5,  -7,  20,  13,  -6,  16,  25,  14,  15,  -4,  19, -16, -10,  20,   1,   2,   5, -18,  -7, -20,   5,  10, -24,  18, -22, -13,   6,   2,   5,  -1,   2,   4, -22,   8,   3,   3,   0 },
     {   -2,  -7, -18, -54,  24,  -2,  11,   3,   6,  13,  35, -16,  34,   1,  -8,  -3, -14, -34,  32,   8,  -8, -26,  -5,   8,  31, -34,  -9,   6, -11,   1,  25,   1,  23,  -1,  -5,   6,  -7,  18,   1,  -7,  -7,   2,  12,   8,  -3,   0,  -6,   4,   0,  -9,  -1, -26,  14,   9,   0,  -1,  -7,   8, -14,   5,  17, -11,   0,  -1 },
     {   11,  12,  -1,  -1,   5,   4,  -2,   3,  23,  28,  12,   4,  20,  -6,  -3,  -1,  32,  37,  24,  13,   8, -11,  -3,  -2,  32,  40,  20,   5,  -2, -13, -17,  -4,  30,  29,  16,  -3, -10, -22,  -9,   4,  12,   7,   1, -14, -20, -20,  -7,  -3,  -9,   3,  -9, -20, -26, -23,   3,  11,  -3, -16, -20, -17, -10,   8,  17,   6 },
     {   -3,  -5, -14,  24, -12, -24,  10,   0,   1,   9, -16,  -8,  32,  23,  -1,  -5,  -4,   0,  14,   5,  -8, -28,  -9,   6,   7, -11, -24,  11,   6,   5,  16,  -1,  -7,  35,  -8,  -9, -18,   7,  -5,  -7, -20,   5, -51,  25,  16,  -6,  -9,   4,  11, -20,   2,  44, -15, -13,   4,   0,  -7,   5,  11, -14, -34,   9,   8,  -2 },
     {   -2,   4,  10,   2,  -4,  -6,   0,   0,  -2,   8,  25,   7,  10,  -1,  -3,   1, -12,  -9,  20,  16, -14,  -9,  -1,  -2, -29, -23,  -3,  10,  -7, -10,  -9,   4, -30, -33,  13,   2,  -8, -10,  25,   5,  -1,  18,  33,  26,  -4,   9,   6, -16,   9,  35,   9,  -6, -51, -12, -12,  -3,   6,   0,  -5, -34, -27,  28,   1,   7 },
     {    1,   1, -12, -17,  47,  -3,  -6,   1,   0,   4,  25, -11, -12, -17,   6,  -1,   7, -21,  -5,  29, -10,  12,  17,  -4,  26,  28, -21,  -1,  -4,   9, -15,   2, -55,  -5,  18,  12,   5,  -6,   1,   7,  24,  30, -13, -18,   3,   2,   2,   1,  -2, -25, -21,  26,   2,   1,  -3,   5,   1,   5,  27,  -2, -12,  -3,   6,  -2 },
     {    1,   0,   0,  -5,   0,  23,  -7,   3,   1,   3,  -3,  -7,   2,   4,  -1,   1,   1,   2,  13,  -2,  12,  18,  -7,   1,  10,  -2,   1,   8,   4,  11,   4, -14,  10,  31,  14,  12,  20,  24, -15,   3,  19,  -5,  18,  16,  20,   8,  -3, -20,   0,  27,   9,  30,   1,   2, -62,  -4,  11, -15,  13, -26,  -6, -47, -16,  27 },
     {   -3,  -6,  -6, -21, -61,  32,  -2,   0,   2,  10,  10,  34,  23,   2, -12,   3,  -4, -12, -23, -13,  26, -33,  -2,   5,  19,  21,  12, -16, -11, -12,  25,   0, -30, -17,  18,  15,  14,   9,   4, -13,  24,  13, -23,  -2,  -6,   4,  -8,  -2, -10,   1,  16,   3,  -1,  -1,   4,   3,  16,  -4,   7,   7,  -1,   5,  -2,   0 },
     {    0,   0,  -7,   0, -15,  30,  -1,   2,   1,  -2,   1, -20,  15, -17,   6,   4,   6,   8,   0,  10, -19,  31, -17,  -7, -15,  10,   3,  -5,  15, -13,  -3,  14,  32, -21,  15,  -3,  -7,  14,  17, -11, -31,  47, -16,  12,  -2,   6, -18,  -1,  14, -34,   2,   2, -16,  21,  -3,  12,   3, -21,  27, -21,  37,   0,  -9,  -4 },
     {    0,  -5,  -2,  12, -20,  41,   8,  -9,  -3,   1, -20, -21,  21, -29, -17,   8,   6,  -6,  27, -14,   3,  15,  31,  -8,   9, -17, -27,  25, -16,  17, -29,  -2, -24,  23,  -8, -24,   6,  -6,  -1,  13,  36, -12,  -7,  10,  -5,   9,   7,  -1,  11, -14,  15, -16,  -4,  11,  10, -10,  -4,   2, -12,  -7,  25,  21,  -8,  -7 },
     {   -4,  -4,  -3, -24, -40,  11,   3,   0,   2,   5,  19,  20, -32,  23,   4,   1,  -6, -13,  -7,  47,  24,  14, -16,  -8,   1,  -7, -53, -14,   3,  -8, -29,   0,   9,  40,   1, -19,   4,  -2,   8,  17, -30,   0,  16,  -5,  -9,   2,   2,   8,  -4,  -7,  -1,  -5,   8,  -9,   4,  -5,   8,  -3,   3,   2,  -8,   8,  -4,   1 },
     {    5,   0,   3,  23,  16,  -5,  -5,   8,   3,  -9,  -5,  27,  23, -20,   1,   1,   0, -17, -28,   2,   6, -11,   6,  -4,   3, -12, -30, -16,  -2, -22,  -6,   8,  47,   0, -14,  -6,   5,   0,  30,  12,  42,  29,  -9,  14,   9,  12,  13, -17,   7,   8, -40,  10,  -2,  -5, -14,  -5,   1, -15, -18,  29,  -1,   4,  -1,   4 },
     {   -6, -10, -20, -28,  -8, -13,   9,   1,  -7, -14, -26, -39, -30,  12,  10,  -4,   2,  -3, -19, -28, -20,  -6,  19,   1,  16,  18,  -3, -24,  -5,  17,  20,  -3,  16,  23,  20, -10,  -7,  13,  17,   1,   6,  11,  12,  -1,  -3,  24,   0, -22, -28,   8,   0, -10, -24,   6, -11,  -3, -20,  -4,   0,  -4, -10,  45,  -1,   4 },
     {    1,   6,   3,  14,  -5,  11,  -5,  -2,  -5,  -2,  -7,  11,  11, -19,  -9,   0,   3,  -3, -13,  -8,  -9,   6,  14,   1,  23,  18, -10,  14,   0,  18,  10, -18, -28,  13, -18,  -1,  -4,   9, -24,  -9, -59,  51,  18,  10,  -2,  -3, -10,  10,   4,  41, -27, -10,   4,  -2,  -2,   0, -11,  -2, -28,  19,   3,  -2, -14,   3 },
   }
 },
 {  //4
   {
     {   89, -52,   1,  -4,   1,  -1,   0,   0, -66,  17,  17,  -1,   3,   0,   1,   0,   5,  22, -11,  -4,   0,  -1,   0,   0,  -2,  -6,  -6,   5,   1,   1,   0,   0,   2,   2,   3,   1,  -1,   0,   0,   0,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   33,  23, -17,  -4,  -4,   0,  -1,   0,  23, -79,  24,   7,   5,   1,   1,   0, -37,  42,  30, -23,   2,  -4,   1,  -1,   0,  16, -40,   0,   8,  -1,   1,   0,  -3,  -4,   5,  18,  -5,  -1,   0,   0,   2,   1,   0,  -6,  -5,   2,  -1,   0,  -2,   0,   1,   1,   3,   1,   0,   1,   0,   1,  -1,  -1,  -1,  -1,  -1,   1 },
     {  -44,  11, -13,   1,   1,   1,   1,   0, -19,  14,  30,   0,   1,  -3,   1,  -1,  -2,  52, -53, -15,   2,  -1,   1,   0,  11, -31, -24,  52,  -3,   3,  -2,   1,  10,  -8,  34,  -3, -27,   4,  -2,   1,   3,  -7,  -2, -18,   7,   9,  -1,   1,   1,   1,   3,   2,   6,  -3,  -2,   0,   1,  -1,  -1,  -1,   0,  -1,   0,   0 },
     {   33,   6, -24,   6,   1,   3,  -1,   0,  33, -24,  17,  -6,  -7,  -2,  -1,   0, -22,  -5, -26,  14,   9,   5,   1,   1,  17, -22,  50,  15, -26,   1,  -4,   0,  -4,   0,  13, -64,  14,  10,   1,   1,  -5,   6, -14,  13,  36, -15,   0,  -1,  -4,   5,  -1,   6, -15, -12,   5,  -2,  -1,   0,  -2,   1,  -1,   6,   3,  -1 },
     {   42,  15,  -9,  -8,  -7,  -1,   0,   0,  29,  13, -42,  19,  -1,   1,  -3,   0,   0, -39,   2,   3,  -8,   4,   1,   1, -16,   6,  -1,  34,   6,  -1,  -1,  -1,   8,  -8,  28,   5, -52,   6,  -2,   2,   0, -11,  11, -45,  16,  29,  -4,   3,   0,   0,   3,   3,  25, -15,  -7,   0,  -2,   1,  -1,  -1,  -4,  -5,   6,   0 },
     {  -26, -27,  22, -11,   0,   3,   2,   1, -16,  -4,  15,  10,   5,  -1,  -3,   0,   9,  -3,  32, -30,  -4,   4,   1,   1,   1,  23, -13,  -9,  13,  -6,  -3,  -1,   7, -18,  -3, -23,  -7,  15,   2,   4,   4,   0,   6,  -9,  66,  -7, -11,  -1,  -2,   7,  -8,  18,  -8, -55,  11,  -1,  -3,   3,  -6,   2, -17,  14,  23,  -5 },
     {   31,   7,  11, -18,   1,  -1,   0,   0,  66,  45,   4, -27,  -2,  -6,  -2,  -1,  16,   8, -30, -35,   8,   0,   1,   0, -13, -15, -37,  -1,  23,  -1,   2,   1, -10, -12, -11,  20,  20,  -2,   2,   0,  -4,   2,   4,  23,   8, -17,  -1,  -2,  -4,  -1,  -4,   6, -20, -11,   9,  -1,  -2,   0,  -1,   2,  -5,  11,   6,  -2 },
     {  -17, -86,  29,   5,   7,   2,   2,   1,  49, -19,  36, -15,   2,  -3,   1,  -1,  -6, -13,  -7,  -7,  10,  -3,   2,   0, -12,   6,  14,  16,   7,  -5,   2,  -1,   0,  22,   4,   1, -19, -10,   1,  -2,  -6,   1,  -6, -13, -16,  16,   4,   0,   2,   3,   2,  -4,  15,  12,  -8,   0,   0,   0,   0,  -1,   4, -11,  -5,   3 },
     {    4,   9, -11,  13, -13,  -2,   0,   1,   4,  -3,  18, -25,  18,   5,  -2,  -1,  -4,  15, -20,  20, -20,  -3,   0,   1,  17, -23,  25, -35,  19,  13,  -5,   0, -17,  16, -28,  19,   0, -17,   6,   0,  12, -16,  19,  -7,   1,  36, -12,   1,   2,  -2,   6, -10,  37, -49,  -8,   2,  -3,   2,  -7,   8, -20,  -6,  30,  -4 },
     {   -3,  -5,  -1,   2,   4,   1,   2,   0,  -1,  28, -11,  -9,  -1,  -5,   0,  -1, -66,  13, -19,  16,  14,   1,   4,   0,  -1,  59,  -6,  21, -19, -11,  -2,  -2,  -2,  -4, -48,   0, -23,  15,   5,   2,   7, -11,   7,  31,   5,  19,  -8,  -1,   8,  -6,  11,  -5, -14,  -6, -10,   2,   2,  -5,   2,  -5,   1,   6,   4,   2 },
     {   -7, -15, -17,   7,   7,  -1,   3,   0, -11, -21, -17,   7,  24,  -7,  -1,  -1, -15, -18, -12,  23,  11,  -4,   4,   1,  -5,  -8,  11,  27,  23, -15,  -3,  -1,   2,  14,  18,  36,  -8, -24,   3,  -1,  10,   8,  20,   4, -19, -33,  -4,  -2,  -8,   0, -14, -10, -43, -23,  33,   2,  -2,   6,  -3,   4, -12,  43,  22,  -9 },
     {    0, -22,  46,   3,  -4,  -4,  -2,  -1,  15,   6, -38, -21,   8,   6,   2,   1, -37,  30,   4,  39,  -5,  -6,  -4,   0,  26, -10, -14, -15, -25,  16,  -1,   3,  13, -38,  31,   5,  25,   6, -10,   1,   4,   6,  21, -27,   1, -20,   2,   2,  -6,   2, -14,  -1,  13,  -2,   7,  -2,  -1,   4,   0,   6,  -5,  -4,   0,  -1 },
     {   11,  43,  66, -26,   2,  -5,  -1,  -1, -26, -27,  16, -51,   9,  -1,   3,   1,  10, -31, -11,   1,  39,  -1,   2,   0,  25,   3,   5,  20,   4, -25,   0,  -2,  -4,  -8,  -5,  -1, -15,  -7,  10,  -1,  -4,   5,   2,  -2,  -2,   7,   4,  -5,  -1,   4,  -5,  -1,   4,   2,  -2,  -1,  -1,   2,  -2,   2,   0,  -3,   0,   4 },
     {  -13, -21, -24, -15,   8,   8,   0,   2, -21, -31, -34, -14,   4,  17,   3,   3, -19, -33, -23, -10,  25,  14,  -1,   1, -17, -19, -16,  11,  20,  19,   1,   1,  -1,  -8,  10,  17,  38,  13,  -7,  -1,   2,   9,  12,  30,  22,   5,  -8,  -3,  10,   5,  18,  13,  17,  -9, -29,  -3,   1,  -3,   3,  -3,  -2, -36,  -6,   7 },
     {    3, -17,  -5,  14,   2,   3,   0,  -1,  11, -28,   4,  11, -13,  -1,  -1,   0,  56,  -2,  -6,  15,   1,   3,   1,  -1,  17,  -6, -14,   7, -36,  -8,   0,  -1, -19, -37, -10,  10,  -9,  37,   9,   1,   9,   1,  49,  16,  -1,   9, -21,  -5,   0,  -6,  11, -41, -11,  -4,  -6,   7,  -2,  -1,   2,  -7,  20,   6,   4,   2 },
     {    0,  -4, -27,   1,   2,   7,   1,   2,  -2,  28,  36,  11, -19,  -4,  -5,   0, -38, -40,  15, -29,   3,  12,   2,   3,  60,  -2,   3,  -7,  25,  -2,  -5,  -1,  13, -41,   8,   3,   6, -19,   2,   2,  -1,   1,  27,  -5, -13,  -2,  12,  -2,  -2,   3,  -2, -25,   6,  10,   0,  -4,  -3,   3,  -3,   3,  15,  -5,   1,  -2 },
     {    2,   4,   4,   4,  13, -16,   3,  -3,   0,   7,   3,   9, -10,  13, -11,  -3,   9,  -3,  11,  -5,  18, -27,   5,   2, -13,  19, -12,  19, -26,  19, -11,   0,   9,  -9,  22, -18,  19, -34,   8,   3, -13,  29, -18,  20, -24,   7,   0,  -3,  18, -45,  14, -22,  15, -28,  -5,   8, -10,  18, -12,  12, -14,  -9,  42,  -7 },
     {   -1, -11,  28, -25,  -7,   1,   7,  -1,   5,  -5, -31,  32,  12, -13,  -3,  -3,  -7,  17,   4,  -1, -26,  14,  10,   0,  37, -27,   5,  -3,  23,  -4, -17,  -4, -27,  -8,   8,  -6, -16, -21,  22,   4,  -7,  24, -24,  26,   1,  27,   3, -10,   3, -13,  23,   8, -23,   3, -20,   9,   4,  -4,  12, -32,  11,   5,   8,   4 },
     {    1, -19, -35, -19,  16,   1,   3,   2,  14, -11, -33,  -4,  19,   7,   3,   2,  45,  22,  -6,   8,  20,   0,  -6,   1,  59,  31,  -1,   9,   6, -10,  -6,  -4,  29,  -3, -18,  -1,  -4, -22,  -4,   2, -12, -25, -17,   7,   1,  -5,  16,  -1,  -9,   1,  -3,  18,  13,   3,   1,  -7,  -5,   1,   3,  25, -10,  -6,  -3,   0 },
     {    6,   9,   9,   7,   5,  24, -16,   0,  10,  22,  14,  10,  30, -15,   9,  -7,  10,  17,  29,  12,  26,  22, -18,   0,   2,  18,  17,  21,  15,   7, -17,   2, -20,  15,  28,   3,  19,   2,   1,   0,  66,   0,   9,   9,   8,   2,  -5,  -7,   2, -10,  21,   4,   8,   2,  -1,  -9, -17,  20,   1,  12,  10,  -8, -20,  -3 },
     {   -1, -10, -19,  47,  -1, -18,  -1,  -6,   0,  17, -12, -33, -12,  18,   7,   5,  -5,  -9,  37,  -6,  32,  -6, -21,  -1,  35, -28, -26,  -2, -14, -20,  17,   4, -26,  21,   6,   9, -20,  12,   1,  -9,   4,   7, -29,  -3,   7,   2,  -6,   6,  -1,  -8,  19,  12,  -4,  -9,   5,  -2,   5,   8,  -9, -27, -13,   7,  -4,   0 },
     {   -8,  -5, -36, -70,  15,   2,  -3,   3,   3,  20,  13, -24,  45, -10,   1,   2,  -8,   6,  42,  16,   6, -15,  -2,  -2, -14, -27,   6,  -3, -19, -12,   8,   0,   4,  -2,   1,   0,  -8,  17,  10,  -8, -21,  18,   4,  -2,  -1,  13, -20,   1,  12,   0,  -7,  -9,   1,   3,   4,  10,   2,   2,  -1,   6,  17,   1,   1,  -3 },
     {   -1,  -2,  21, -43,  27, -12,  -1,   2,   1,   1, -20,  41, -32,  19,  -4,   2, -14,  15,   0, -24,  25, -13,   0,   4,  16, -21,  16,   3,  -1,  -2,  11,  -2, -30,  35, -16,  -1,   1,  10, -14,  -3,  15, -28,   8, -11,  -3, -23,  -9,   4,   1,   0,   3, -30,   9,  -5,  14,   2,   7,  -7, -23,  10,   1, -15,   7,   0 },
     {    4,   6,   3, -27,  34,  -9,  -2,  -3,   4,  -1,  24, -12, -10,  -6,   6,  -6,   6, -19,   1,  29, -42,  -9,  12,  -6,  24,  18, -18,   8,   2,  23,  -2,  -5,  20,  28,  16,   8,   9,  13, -34,  -4,   3, -13,  -4,   8,   0, -12,  -2,  15,   2,   8,  34,  -5,  -4,  -9,  -3,   4, -21,  -4,  -5, -50, -14,   9,  -5,  -2 },
     {   -2,   0, -14, -10,  12,   8,  -2,   3,  -3,   7,  12,  12,  -3, -11,  -1,  -1,   2, -14, -21,  15, -14,  -2,   6,  -1,  23,  33, -20,  -3, -12,   3,   4,  -8, -72,  17,  15,  12,  -2,   7,  -2,   1, -10,  50,  -5,  -8,   6,  -7,   5,   1,  -6,   8, -41,  13,  17,  -1,  -6,  -1,  10,  -4,   9,  23, -19, -10,  -2,   1 },
     {    4,   5,   7,  16,  15,  27, -39,   2,   4,   6,  13,  25,   4, -29,  -3,   5,   4,   1,  14,  16,  15,  -1,  10,   2,   2, -12,   2,  19, -13,   7,  24,  -3,   6, -16, -22,  11,  11, -24,  16, -16, -17,  -8,   3, -19,  -8, -10, -20,  -4,   7,  19,  11,  24, -25, -12,  -8,   0,  17, -19,   0, -15, -34, -45,   7,  15 },
     {   -2,   0,  -9,   0, -25, -15,   7,   2,   1,   4,  -5,  -4,  -9,  -9,  35,  -5,  11,  17,   2,   2,  30,   5, -12,   0,  -2,  21,  15,   6,  13,  24, -42,  11,  -8,  -1,  -6,   6,   8, -12, -16, -13, -26,  28,   5, -25,   7,   3, -14,  11,   6,  18,  -7, -11,  -9,   0,  -9,  -7,  -6, -37, -50, -27,  20,   4,   1,   0 },
     {    6,   3,  14,  11,   8,  30, -10,  -6,   3,  -7,  -7,  13,   7, -10,  23,  -4,  -5, -12, -16, -23,  -2,   3,  -8,  -6,  25, -12,   0,  -4,  -6,  22,   8,   3,  46,  16, -23,  16,  -3,  27,  -8, -17,  17,  35, -33,  -4,  -3,  14, -35,  14,   1, -26, -32,  -6,  -2,  -6,  10,  13, -19,  -6,  11,  11,  15,   0,   3, -11 },
     {    4,   9,   6,  45,  76, -11,  -5,  -3,  -5,  -2, -25, -19,  36, -29,  -7,  -3,  -3,   9,   6, -37, -19,   0,  -1,  -2,  -4,  -1,  12,   1,   7,  11,  -7,  -5, -13,  -2,   3, -12, -13,   2,  -4,   0, -23,  -2,  12,   8,   3,  -8,   4,  -3,  12,  13,  -9, -19,  10,   0,  -9,  -5,  -5,  -2,  -9,   3,  13,   2,  -4,  11 },
     {    0,   0,  -8,  -6,  33, -22,   0, -16,   1,  -3,   3, -13, -44, -12,  20,  -3,   2,  14,  -3,  13, -10,  42,  -8,  -9, -10,   9,  16, -13,  14, -50,  -8,  22,   6, -13,  21,  11,   5,   4,  16, -10,   9,  -6,  -8,  -6,  -5, -10, -12,   3,  38, -27, -12,  12,   8, -10,   0,  15,   9, -23,  18,   0,  -1, -13,  10,  -2 },
     {   -4,  -6,   3,  -5,   2,  20,   4,   3,  -6,  -8,  -3, -10, -36, -15,  -2,  -6,  -5,  -2,  20,   8, -17,   4,  -2,   5,  -4, -16,  10,  37,   1,  24, -13,  -2, -25, -30, -18,  21,  11,   7,  14,  -4, -20, -30, -30,   8,  -4,  12,  -8, -13,   7,  -6, -15,   7,  17,  -4,  16,  -3, -19,  20, -25,  14, -18,  30, -35, -21 },
     {   -5,  -5,  -6,   0,  -4, -17,  -2, -29,  -6,  -2,   3,  -1,   6, -49,  13,  -5,  -3, -10,   7,  11,  23,  -1,  32,  14,  -9, -13, -11, -12,   1,  10,   6,  -8, -16, -10,   8,  -5,  -5,  10, -17,  -1,   7, -37,  -9,   5,   2,   6,  21,  14, -46, -42, -16,   7,  -6,   7, -14,   4, -25, -26,  -9,   9,   6,  -8,  23,   5 },
   },
   {
     {   85, -51,   3,  -4,   1,  -1,   0,   0, -70,  33,   5,   1,   1,   0,   1,   0,  16,   4, -10,   2,  -2,   0,  -1,   0,  -3,  -6,   3,   1,   0,   0,   0,   0,   2,   2,   0,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -64,  -3,  23,   0,   4,   0,   2,   0, -33,  64, -23,   1,  -3,   0,  -1,   0,  58, -44,  -4,   6,  -2,   2,  -1,   0, -15,   2,  15,  -5,   1,  -1,   1,   0,   9,   2,  -7,  -1,   0,   0,   0,   0,  -2,  -2,   2,   1,   0,   0,   0,   0,   4,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 },
     {  -42,  24, -14,   7,  -1,   2,   0,   0, -54,  -2,  40,  -7,   5,  -2,   2,   0,   6,  59, -41,  -2,  -2,  -1,  -1,   0,  34, -46,  -3,  15,  -4,   3,  -1,   1,  -4,   4,  16,  -8,   0,  -1,   0,   0,   6,  -2,  -8,   0,   1,   0,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0 },
     {  -32, -17,  33,  -9,   5,  -2,   2,   0, -46,   1,  -2,  11,   1,   2,   1,   0, -42,  29,  42, -16,   1,  -3,   1,  -1,  13,  52, -48,  -9,   2,  -2,   0,   0,  19, -29,  -8,  18,  -3,   3,  -1,   1,   3,   3,   7,  -5,  -1,   0,   0,   0,   4,  -3,  -4,   2,   1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0,   0 },
     {   15,  63, -43,  -4,  -4,  -1,  -1,   0, -43, -40,  16,  23,  -1,   5,   0,   1,  46, -21,  32, -23,   1,  -3,   1,  -1, -31,  19, -15,   0,   7,  -2,   2,   0,  10,   1,   1,   7,  -5,   1,  -1,   0,  -2,  -1,  -1,  -4,   1,   0,   0,   0,   2,   0,   2,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0 },
     {  -18, -15,  -6,  20,  -4,   4,  -1,   1, -27, -26,  40, -20,   9,  -1,   2,   0, -27, -10,   4,  22,  -1,   1,   0,   0, -18,  27,  49, -32,  -2,  -2,   0,  -1,   0,  46, -53,  -5,  10,  -4,   1,  -1,  13, -19,   0,  18,  -8,   3,  -2,   1,   2,   4,   2,  -8,   2,   0,   0,   0,   2,  -2,  -2,   3,   0,   0,   0,   0 },
     {   20,  66, -17,  -5,  -7,  -3,  -2,   0, -26,  35, -45,  -2,   5,   1,   0,   0, -48, -19,  13,  29,  -6,   4,  -1,   1,  45, -12,  17, -22,  -2,   1,  -1,   0, -11,  -6, -11,   6,   9,  -3,   2,   0,   8,  -3,   9,   4,  -7,   1,  -1,   0,  -1,  -2,  -1,  -3,   1,   0,   0,   0,   2,  -1,   1,   1,   0,   1,   0,   0 },
     {   13,  18, -22,  15,  -7,   1,  -2,   0,  25,  32,  18, -40,   5,  -5,   1,  -1,  22,  23, -36,   5,  -1,  -3,   0,  -1, -12,  22, -32, -11,   5,   1,   1,   0,  14, -34, -38,  42,   3,   1,   0,   1,   0, -26,  39,   9, -17,   5,  -2,   1,  -2,   4,   0, -14,   5,   1,   0,   0,  -2,  -3,   1,   5,   0,  -1,   0,   0 },
     {    2,  22, -19,  -4,   9,  -2,   1,   0,  -3,  37,  12,  -8, -14,  -2,  -1,   0, -49,  -9, -28,  -9,  26,   0,   3,   0, -40,  30,   9,  50, -18,  -4,   0,   0,  40,  -4,  11, -36, -13,   9,  -3,   1,  -5,  -6, -16,   7,  18,  -4,   1,   0,   7,  -2,   5,   4, -10,  -1,   0,   0,   0,  -3,  -2,  -2,   2,   2,   0,  -1 },
     {  -14, -25, -29,  39,  -9,   6,  -1,   1, -18, -24,   6, -17,  15,   1,   2,   0, -16, -38,  -2,  38,   1,   0,   1,   0,   0,  -3,  12,  13,  -6,   4,  -1,   0,  19, -34,  44,  17, -13,  -1,  -2,   0,  -9,  22,  36, -39,  -7,   2,   1,   0,  -2,  13, -19,  -8,  13,  -1,   1,   0,   0,   2,   3,   3,  -2,  -1,   0,   0 },
     {   -2, -18, -39,  35,   2,   3,  -2,   1,   8,  19,  32, -18, -21,   4,  -3,   1,  14, -28,  27, -24,  22,  -4,   1,  -1,  57,  17, -13,  -5, -18,  -5,   0,  -1, -28, -20,  -8, -28,  20,  11,   1,   1,  -8,  16, -10,  23,  18, -14,   0,  -1,  -1,  -3,   0,   3, -14,  -1,   2,   0,  -2,   1,   0,   0,   2,   2,   1,  -1 },
     {    8,  11,  23, -16,   0,  -4,   0,  -1,  -2, -30,  -9,   2,   4,   3,   1,   0,  25,  20, -13,  36, -11,  -4,  -2,  -1,   2,  15,  23, -26, -34,  13,  -4,   1,  35, -50,  -4, -39,  39,   7,  -2,   2, -24,  13,   2,  22,  16, -20,   4,  -2,  -1,   6,  -4,   2, -13,   1,   2,   0,  -2,  -1,   0,   1,   1,   2,   0,   0 },
     {   15,  29,  33,  13, -30,   2,  -2,   0,  11,  27,  35, -46,  12,  -3,  -1,  -1,   3,   8,  21,  -5, -21,  -2,   1,   0,  -9,   8,   0, -30,   8,   8,  -1,   0,   4,   7,  16, -15,  -5,  -5,   0,  -1, -10,  19, -36, -44,  14,   8,   3,   0,  -3,  -7, -29,  28,  20,  -3,   0,   0,   2,  -1,   6,   5,  -8,  -1,   0,   0 },
     {  -10, -22, -56,  13,   6,   7,   2,   1,  17,  32,   1,  47, -11,  -8,  -4,   0,  -9,  30,  27, -16, -41,   9,  -2,   1,  -9, -19,  21, -33,  17,  11,  -3,   2,  44,  -8,   0,  -9,   6, -10,   1,   0,  -9, -11,  -3,  -8,   3,   3,   2,   0,  -2,   3,  -2,   1,   6,  -3,   0,   0,  -1,  -1,  -2,   1,  -1,   0,   1,   0 },
     {  -10, -15, -41,  28, -10,   2,   0,   0,  -7,  -3, -52, -16,  34,   0,   2,   0,   3,  25,  -7,  24,  -1, -10,   1,  -1, -27,  17, -24,   0,   0,   6,   0,   0, -16,  13,   4,   8,  13,  -6,  -1,   0,   1,  20, -40,  19,  -1,  -6,   2,   0,  11, -26,   6,  38, -16,   4,  -1,   1,   5,  -5,  21,  -9, -11,   3,  -1,   1 },
     {    6,  -7, -11,   3,   3,   2,  -1,   1,   8, -15, -24,   5,   9,  -1,   1,   0,  30,   6, -12,  15,   6,   0,  -1,   0,  62,  38,   5,  16,  -6, -14,  -2,  -2,  39,  20, -18, -16, -42,  -2,   0,  -1,   1, -35, -18, -29,   6,  21,   0,   3, -10, -10,  -2,  11,  21,  -2,  -1,   0,  -3,  -3,   4,   6,  -5,  -3,   0,   1 },
     {   -3,  -8,   3, -18,   8,   3,   1,   1,  -7, -18,  13,  -4,   0,   0,   0,  -1, -19, -42, -17, -10,   8,   8,   1,   2,   6, -39, -15,  -3,  26,   0,   1,   0,  27, -21, -27,  19,  10,  -5,   3,   0, -28, -17, -15,  18,  13,   3,   3,   0, -10,  -7,  -3,  52,   9,  -6,   1,   0,   6,   5,  42,  22, -24,  -3,  -2,  -1 },
     {    4,   1,  19,  20, -22,   4,  -2,   1,   2,  -1,  14, -11,  10,  -1,   0,   0,   0, -13,  22,  12, -14,  -1,   1,   0,   3, -15,  -6,   2,  -3,   8,  -2,   2,  20, -14,  18,   0,  -6,  -2,   0,   0,   6, -37,  -8,   9,   0,   8,   0,   2,  -1, -18,  76,  -1, -12,   0,  -3,   1,  -6,  35,   2, -56,   3,  -2,   1,   0 },
     {    1,  -1,   0,  -2,   5,  -8,   4,  -1,   1,   1,   5,  -2,  -6,  14,  -5,   0,  -1,  -9,   0,   0,  14, -15,   5,   1,   5,  -4, -11,   1, -10,  20,  -2,   0,  24,   7,  11,  -1,  27, -25,   1,  -1, -10,  14, -29,  14, -55,  14,   3,   0,  10, -47,   5, -36,  39,  16,  -8,   3, -17,  20, -13,  37,   9, -18,   5,  -3 },
     {    0,   3,  -6, -26,   6,   2,   1,   1,   4,   4,  25,  19,   5, -16,   1,  -3,  -1,  -4, -25,   1, -15,  11,   0,   1,  16,  24,  26,  -7,  14, -14,  -3,  -2, -16, -12,  -8,  19, -25,  -4,   7,  -3, -38,  45,  -2,  -2,  -9,  18,  -4,   0,  59, -26,  11,   3, -10,   0,  -6,   1, -18,   2,  13, -30,   7,  -1,  -2,   1 },
     {   -1,  -2, -11, -24,  22,  -1,   3,  -1,   3,   4,  24,  20, -19,  -1,  -1,  -1,   1, -11, -30,  20,  -8,   9,  -2,   1,  -3,  30,   6, -22,   4,   0,  -3,   1, -20, -27,  21,   8,   5,  -1,   2,   0,  69,   7, -20, -12,  -3,  -1,  -2,   0, -48, -27,   6,  13,   5,  -1,   2,   1,  13,  15,   9,  -2,  -7,  -1,   2,   0 },
     {   -7, -12, -28, -65,  25,  -4,  -2,   1,   9,  14,  25, -20,  28, -10,   6,   1,   4,  -1,  44,  45,  -8, -18,   1,  -6,  -5, -26, -21,  15, -31,   4,  10,  -1,   1,   9, -19,  -2, -13,  15,  -3,   1,   5,   2,   4,  -9,   8,  -4,  -2,   3,   2, -10,  -7,   1,  -1,  -1,   4,  -2,  -8,   6,  -6,  -1,   2,   2,   0,   0 },
     {   -6,  -6,  -5,  11, -14,  -5,   0,   0, -10, -10, -16,  11,  -9,  -8,   9,  -5, -12,  -9, -19, -12, -28,  -7,   8,  -2, -22, -10, -16, -11, -35,  14,   7,  -6, -40, -22, -24, -24, -15,  35,  10,  -1,  -8, -17,  -1, -15,  22,  25,   0,   4,   4, -13,   7,  -3,  44,  10,  -5,  -1,   1,   1,  11,   8,  40, -19, -10,   0 },
     {   -9, -16, -26, -52, -14,  14,   7,   2,  -8, -17, -27, -61,   8,   7,  -4,   0,  -2, -11, -12, -48,  -9,  18,   2,   6,   9,   3,   8, -16,  19,  11, -11,   2,  18,  -3,  11, -10,  13,  -5,   1,   0,  22,   0,  -2, -10,   3,  -1,  -2,  -1,   5,   5,   1, -14,  -7,   4,   2,   3,  -7,  -6,  -4, -16,  35,   1,  -2,   1 },
     {   -2,  -2,  -5, -16,   0,   1,   4,   0,   3,   6,  19,  19,  24, -14,  -1,  -2,  -8, -30, -25,  15, -36,  -6,   7,   3,  15,  29, -39,  -8,   3,  27,  -4,   0, -15,  20,  34,  -6,  17, -13,  -8,   1, -22, -28, -16,  11,  -5,   3,   8,   1,   7,  48,  11,  -3,   0,   0,  -3,   0,  10, -28, -20,   8,  -4,   2,  -3,   1 },
     {   -9,  -3, -13, -26,  -2,   3,   5,  -2,  -6,   9,  -5,  -9,  19,   2,  -1,   3,  -1,  22,  13, -13,  19,  -9,   1,   0, -10,  23,  37,   4,   3, -10,  -1,   1, -44, -14,  19,  -5,   1, -16,   2,  -2, -59, -28,  12,  -3, -14,   7,   3,   2, -49,   6,  13,   8,  10,   2,   2,   1,  -8,  30,   6,  11, -10,  -2,   2,   0 },
     {    0,   5,  -1,  14,  65, -35,  -4,   4,  -7,  -8,  -7, -34, -45,  22,  10, -13,   0,   7,   6,  27, -15,   2,  -5,  12,  -7,   1, -13, -18,  34,  -2, -11,  -3,  -7,   8,  18, -12, -14,   3,   8,   0, -21, -13,  -1,  -3,   9,   3,  -4,  -1,   0,   0,  10,  -4, -13,  -1,  -8,   7, -32,  -8,   9,   7,  13,   6,  -8,  -2 },
     {    1,   1,   7,   6,  -9,  13,  -8,   2,   0,   0,  10,   5,   2, -23,   3,  -1,  -2,  -7,   7,   7, -12,   9,  -6,  -3,   1,  -7,  -5,   5,  -1, -28,  11,   3,   9,  -2,   4,   0, -11,   6,   9,  -5, -17,  -8, -18,  10, -37, -10,   4,   1, -42, -15, -23, -13, -32,  52,   5,   3,   6, -35,  27,  -8,  47,  17, -15,   3 },
     {   -1,  -2,   1,   6, -16,   9, -19,   4,  -3,  -4,  -3,   1,  30, -26,  18,   3,  -1,  -4,  -1,  14,  -6,  39,  -8,  -6,  -7,  -4, -14,   3,  41, -33,  -2,  -1,  -6, -23,  -4, -22, -17,   1,  14,  -4,   7,   5, -17,  26,  19,  -3,  -8,   7,  -4,  -5,  -2, -18,   2, -13,   7,  -1, -28,  23, -48,  37,   0,   2,  10,  -1 },
     {   -2,   1,  -6,   1,   2,  -3,  -8,  -1,  -5,  -2, -10,  -2, -22,  -4,  11,   1,  -6,  -5, -19,   2,   1,  -4,   4,  -9,   1,  -4, -22,  -8,  -3,  12,  -1,  -1,  17,  21, -35,   5,  -9,  -5,   2,   0, -28,  56,  -9, -13,  -6,  12,   5,  -2, -51,  30,  14,   6,  -7,   0,  -4,  -4,  14,  38, -38, -26,  11,   5,  -6,   2 },
     {   -2,   1,   1,  -1, -26,   8,   0,  -1,  -2,   2,   1,  15,  -5, -17,   5,   4,  -6,  -6, -17,  -7,  -7,   2,   5,   0,   5,   8, -16,  -9, -14,  12,  -6,   1,   7,  41,   9,   3,  17,  13, -16,   2,  -3,   5,  38,  -9,  17, -18,   0,   0, -24, -31,   0,   6, -14,  -9,   8,  -3, -80,   7,  17,   9,  -8,  10, -12,   4 },
     {   -7, -10, -12, -28, -73,   3,   9,  -5,   0,   1,   5,   0, -69,  22,  10,   0,   4,  11,  11,  44,   3,  16,  -3,  -7,   0,   7,  -4,  17,  16,   2,   4,   1,  -5,   6,   6,  10,   1,  -3,   1,  -5,  -8,  -8,  -4,  13,  10,   7,  -3,  -5,  10,   5,  -7,   6,   2,   6,  -9,  -1,   7,   1,   1,   6,   5, -13,   1,   3 },
   },
   {
     { -118,  44,   9,   5,   2,   1,   1,   0,  10,   2,   0,  -1,   0,   0,   0,   0,  14,  -3,  -2,  -1,   0,   0,   0,   0,   7,  -2,  -1,   0,   0,   0,   0,   0,   5,  -2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {  -25, -52,  28,   6,   6,   2,   2,   0, -97,   5,  35,   5,   5,   1,   2,   0,  -1,  28,   8,  -6,  -2,  -2,  -1,  -1,  11,  12,  -6,  -7,  -1,  -1,   0,   0,   8,   5,  -6,  -2,   0,   0,   0,   0,   5,   1,  -4,   0,   0,   0,   0,   0,   3,   1,  -2,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
     {  -12,  -9, -16,  14,   0,   3,   0,   1,  16, -92,  19,  17,   5,   3,   1,   1, -56,   0,  40,   6,   2,  -1,   1,   0,   3,  28,  14,  -8,  -5,  -3,  -1,  -1,   1,  14,   0,  -9,  -2,  -1,   0,   0,   3,   7,  -4,  -5,   0,   0,   0,   0,  -1,   2,  -2,  -2,   1,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0 },
     {   21,  89, -51,   4, -11,   0,  -3,   0, -63,  10,   6,   2,   2,   1,   1,   0, -29,   2,   9,   1,   3,   0,   1,   0,   3,  -1,   1,   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0,   5,  -3,  -1,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {  -12, -24, -19,  12,   8,   2,   1,   0, -34,  -2, -43,  25,   8,   3,   0,   1,   8, -68,  -5,  39,   6,   2,   0,   1, -17, -15,  42,  28,  -4,  -4,  -1,   0,   7,  14,  32,  -1, -15,  -5,  -1,   0,   1,   9,  10, -13, -10,   0,   1,   0,   2,   3,   1,  -8,  -2,   2,   0,   0,   0,   1,   0,  -3,   1,   1,   0,   0 },
     {   -1,  14,  67, -23,  -3,  -6,   1,  -1,   2,  34,  14, -22,  -7,  -2,   0,   0, -74, -31,   4,  -3,   7,   6,   3,   1, -44,   0,  15,   3,   8,   3,   1,   0,  -2,  16,   5,  -3,   1,  -1,  -1,  -1,   5,   9,  -5,  -5,   0,  -1,  -1,   0,   4,   3,  -6,  -1,   1,   0,   0,   0,   3,   2,  -2,   1,   1,   0,   0,   0 },
     {  -15, -20, -24,  -3,   6,   5,   1,   1,  -6, -24, -10, -22,  16,   3,   3,   1, -45,  19, -55,  -5,  20,   4,   2,   0,  -8, -39, -32,  31,  23,   5,   0,   1, -10, -10,  12,  40,  13,  -5,  -2,   0,   8,   6,  20,  22,  -7,  -9,  -3,  -1,   4,   5,  11,   2, -11,  -6,  -1,   0,   2,   1,   3,  -2,  -6,  -1,   1,   0 },
     {   12,  18,  18,  21, -14,  -1,  -3,   0,  -6, -36,  72,   2, -11,  -2,  -1,   0,  41, -34,  -4, -15, -10,   3,   1,   0, -26, -54,   0,  -3,   6,  10,   3,   1, -19,  -8,  19,   4,  10,   4,   0,   0,  -2,  14,  13,  -2,   1,  -2,  -2,  -1,   0,  10,   2,  -7,  -1,  -2,   0,   0,   0,   5,  -2,  -3,   0,  -1,   1,   0 },
     {   11,  17,  32, -49,   1,  -3,   2,  -1, -25, -48, -29, -19,   7,   9,   3,   1,  26, -52, -11, -13,  14,  12,   2,   1,  26,  25, -20, -11,  18,   5,  -1,  -1,  18,  -1, -23,   9,  18,   0,  -3,  -1,  -4,  -4,  -4,  18,  11,  -6,  -2,   0,  -4,  -2,   3,  13,  -1,  -6,  -1,   0,  -2,   2,   4,   4,  -3,  -3,   0,   0 },
     {   14,  20,  30,  -1,  -3,  -6,  -1,  -1,   6, -17,  26, -16,   6,  -6,   1,  -1,  -2,  19, -54,  27,   6,  -1,  -1,   0,  58,  -6,  22,  34, -13,  -7,  -5,   0,  23,  23,  17, -15, -29, -11,  -1,   0,   0, -10, -14, -24, -15,   6,   4,   2,  -9,  -7,  -6,  -8,   7,  10,   2,   1,  -2,  -2,  -1,   1,   5,   3,  -1,   0 },
     {   11,  21,  52,  13,  16, -16,  -1,  -2,  -2,  -7, -16,  66,   9, -13,  -3,  -1, -14,   4,   4,  38, -13, -14,  -2,   0,   8, -11, -35,   5, -23,  -7,   3,   2,  -9, -41, -15,  16, -10,   7,   6,   3,   1,  -5,  19,  13,  -7,   5,   1,   0,   3,   2,  15,  -1, -11,   2,  -1,  -1,   1,   2,   5,  -7,  -6,   2,   0,   0 },
     {    0,  15,   6, -69,  25,  -2,   3,  -1,  -9, -22,  -6, -15,  34,   1,   1,   0,  29,  47,  22,  19,   6, -12,  -5,  -1, -39,  -5,  25,  11, -10, -12,  -1,   0, -37,  -5,  15,   2, -10,  -3,   3,   1,  -4,  13,   3,  -8,  -8,   2,   3,   0,   3,   6,  -5,  -5,  -2,   2,   1,  -1,   2,   1,  -5,  -1,   0,   0,   0,   0 },
     {    6,   9,   6,   7,   4,  -4,  -1,   0,   6,   5,   8,  15,  -3,   0,  -3,   1,  18,   0,  28, -30,  18,  -4,  -1,   0, -13,  37, -35,  30,  31, -13,  -1,  -2,  24,  14,  26,  47, -24, -25,  -6,  -1,   4,   6,   5, -21, -43, -14,   5,   2,  -5, -11, -12, -18, -12,  12,  10,   2,  -3,  -1,  -2,  -3,   3,  10,   3,  -1 },
     {   -2,  -6,   8, -32, -16,   3,   3,   0,  -2,   5, -25,  -1, -26,  13,   3,   1, -17,   7,  43, -19, -20,  13,   0,   0,  63, -41,  -1, -10,   3,  12,  -3,   0,  -6, -18,  45,   3,   6,   4,  -5,  -1,   8,  21,  22, -21,  -1,  -6,  -4,  -1,  -7,   2,  -9, -18,   2,  -1,   0,   1,  -2,   1,  -7,  -5,   6,   2,   2,   1 },
     {   -3, -14, -25, -16,  36,  -5,   1,   0,  13,  26,  47, -34,  33, -14,  -3,   0,  -8, -43,  42,  14,  19, -16,  -1,   0,  37, -13,  -8,  14, -15, -14,   1,   1,  10, -22, -16,   8, -13,  -1,   7,   2,  -4,  -7,   6,  15,  -4,   3,   4,  -1,  -1,   7,   8,   4,  -8,  -2,   0,  -2,  -3,   2,   0,  -2,  -6,  -1,   1,   0 },
     {   -6,  -8, -11, -18,  -1,  -1,   3,   0,  -6, -11,  -3,  -2, -14,   1,   3,   1,  -6, -10,  -4, -29,  -7,   1,   6,   1, -20,   0, -39,  29, -12,   0,  10,  -1,   7, -24,  10, -14, -48,  25,   9,   0,  -7, -19,  -5, -34,  22,  48,   4,  -1,   5,   2,  11,  13,  42,  14, -15,  -5,   4,   5,   7,   6,   9, -13, -13,  -1 },
     {    4,  -2,  -6, -11,   3,  -1,   0,   0,   7,  -4,  -2,   7,   0,  -4,  -1,  -1,   9,  -1,   2,  11,  -5,  -5,  -3,   1,  -4, -25, -20,   1,  -4,  -3,   2,   0,  31,   9, -24,  -3,   2,   2,   2,   0,  76,  31, -33,  -8,   1,   1,  -1,  -1,  55,   9, -33,  -5,   3,   0,  -3,  -1,  21, -10, -18,   6,   2,  -2,   0,   1 },
     {   -3,  -8,   1, -17, -51,  11,   7,   0,   1,  -7,  11, -15, -46,  16,   8,   0,   7,  11,  23,  36, -12,   5,   0,  -2, -15,   1,  26,  38,  -2,   2,  -6,   1,  28, -29, -16,  27,   2,  -3,  -3,   2,   6, -40,   2,  22,  -3,  -2,   1,   0,  18, -15,  12,   5, -11,   0,  -1,   0,  -2,  -4,   8,  -9,  -5,   2,   1,   2 },
     {   -2,  -3,   5,   4,   4,   3,   0,   3,  -3,  -3,  -7,  19,   4,  -1,  -5,  -2,  -5,   2,  22,   5,  21,  -6,  -7,  -1,  -2, -47,  -4,  11,  22,  -6,  -7,   2, -11,  -5,  -9,   3,  14, -14,  -5,   2, -12, -18, -54,  -7,   5,  -6,   4,   2, -30, -45, -35,  32,  20,   3,   7,   1, -20,  -7,  17,  36,   1,  -2,   0,   1 },
     {   -4,  -3, -25, -51,  50, -12,  10,  -2,   7,   8,  34,  46, -25,   3,   7,   0,  -7,  -2, -23,   9, -41,  20,   6,   2,  -9,   3,   1,  -7,  15,  14,   2,  -1,  28,   9,  11,  -2,   9,  -4,  -5,  -2,   0,   0, -11,  12,   0,  -5,  -7,   0, -23, -16,  16,   1,  -4,  -1,  -2,   2, -13, -10,  -9,  -3,   8,  -2,   3,   2 },
     {    7,   5,  13,  14,  30,   0,  -8,   0,   2, -10, -12, -11,  26,  -5,  -5,   2,   0,  12,   2, -36,   7,  -5,   2,   3, -17, -17,  33, -31,  -8,   9,   8,   1,  56, -25,  28,  -8,  -3,   4,   0,   1,  30, -39,  25,   2,  -5,  -1,  -1,   0,   9, -30,   7,   9,  -6,  -2,   3,   2,  -5,  -8,  17,   2,  -7,   5,   2,   0 },
     {   -6,  -9,  -9,  -7,   3,  -8,  -7,  -3, -10, -14, -13, -14, -24,  -8,  -5,   5, -11, -14, -19, -22, -10, -12,   1,   0,  -8, -18,   7, -29, -17, -14,  -1,   2, -11, -27, -32,   0,  -8, -27,  15,   1, -17, -16, -25, -10, -37,   6,  16,   0, -11,  -4,  -4, -33, -19,  34,  14,  -1,  15, -20, -28, -14,  16,  29,   4,  -3 },
     {   -1,   0,   0,  -3,  -3,  12,  -6,   5,   0,  -1,  -1, -10,  14,  10,  -2, -11,   0,  -3,   8,  21, -26,  11,  -6,   3,  -1, -20, -10, -27,  16,  -4,   6,   1, -15,  55, -34,  29, -19,  16,   5,  -2,  26, -42,  39,  -9, -16,  22,  -4,   7, -26,  -8,  -3, -23,  18,  -2,  -4,   2,  -2,  -9,   5,  11,   7,  -9,  -4,   3 },
     {   -2,  -3,  -8,   2, -13,   7,   6,   2,  -1,  -1,   8, -22,  11,   1,  -3,  -6,  -4,  -5, -19,  35, -17,  -3,   4,   0, -16,  21, -15, -23, -20,  11,  -2,  -2,  27, -39,   1,  16,   2,   9,  -4,  -2, -20,  36,  25, -25,   4,  -6,  -2,  -2, -16, -11, -56, -11,  10,  -5,   5,   0, -33, -20,   4,  20,  -4,   3,  -2,   0 },
     {   -1,  -3,  -2,   0,   3,   6,   0,   0,   1,   5,  16,   4,   3,   1,   4,   0,  -1, -21, -18,  -6, -13,   7,   2,   2,  19,  41,  15,  -3,  -6,   5,   3,  -2, -63, -24,  31,  22,  -7,   5,   3,  -3,  33, -19,  -2,  12,  -7,  -2,  -7,   1,  20, -34, -23,  23,   3,   2,   0,   3,  32, -27,  -5,  20,   4,  -2,   0,   1 },
     {   -2,  -3,  -3,  -3,  -9,  11, -18,  -7,  -3,  -4,  -6, -22,   7, -30,  -9,  -6,   3,  -6,   9,  15, -21, -20,   8,   1, -10,   1, -36,   3, -38,  28,  -6,  -4,  -1,  26,  21,   6,  16,   4, -10,   5,  11,  -1, -24, -21,  20, -14, -10,  -1, -28, -31,  20,   8, -27,   5,   6,   7,  30,  -1,  25, -35,   2,  13,  -4,   0 },
     {    2,   4,   4,  -2,  10,  39,  -5,  -8,   2,   3,   3,   4,   5,  68, -19, -12,  -4,  -3,   0, -14,  -9,  17, -25,  -2,  -1, -12, -18,   2, -44,  -6,  -3,   7,   4,  23,   2,  -9, -16, -12,  12,   3, -28,  18,  -2,  25, -10,  -5,  10,   4,  26, -16,   0,   5, -14,   5,  -3,  -4,   2, -23,  28,   3,   0,  -5,   2,   0 },
     {    4,   8,   4,  19,  61,  10, -18,   5,  -8, -10, -13, -24, -50,  -9,  15,  -6,   3,   3,   7,  -6, -11,  -1,  19,   3,   1,   2,   3,  26, -17,  27,  -4, -13, -17,   4, -21,  -2, -10,  -9, -14,  14,   0,  -4,  11,   6,   2, -16, -16,  14,  12,   4, -28, -13,  -6,   3,   5,  12, -17,  31,  11,  15,   1, -14,  -9,  14 },
     {   -2,  -4,  -4,  -8,  -4,  -1,  -3,   9,   1,   3,   9,   5, -25,  39,   0,   1,   2,  -3, -11,  10,  58,  -5, -17,  -2, -10,  10, -16,  -9, -25, -14,   0,   0, -10, -18,   8, -30,   2,  16,   9,  -5,  47,  -7,   1,   3, -10, -15, -15,   8, -45,  -8,   1, -25,  -2, -10,  -1,   6,  -9,  28,   0,  -7,   9,  -6,   7,   6 },
     {   -1,  -1,  -4,   5,   1, -35,  -1,  -5,  -3,  -2,  -2, -15, -23,  -3,  -5,  -9,   1,   1,  -5,  -3,  -8,   4,  -5,  15,   5,   2,  19, -15,  12, -31,  -3,   1, -14,  -8, -18,  15, -24, -22,  16, -10,  27,  49,   1, -18,   4,  17,  -8,  -4,  -6, -15,  30,  25, -17,  -4,   0,   5, -24,  -4,  58,  -5,  -3, -11,   3,  -2 },
     {    2,   3,   6,  16,  16,   5, -20, -20,  -1,  -3,  -3, -15,   2,  15, -16,  14,  -1,   1,   4,  21, -11,  -3, -16,   7,  -1,   6, -11,  -2,  35, -13,  18,  -3,  -3, -14,  20,   1,  -9, -26,  34,  -3,  -4, -27, -19, -10,  55, -19, -21,   2,  13,  15,   0, -29,  -6, -18,  -2,   2, -14, -22, -10, -15,  28, -13,  12,  -4 },
     {    3,   2,   1,   4,  16,  31,  40,  15,   1,   0,   4,  -2,  10,  -6, -27, -23,  -3,  -4,  -5,   4, -16,  16, -19,   5,   0,   8,  -1,   4,  22, -15, -38,  19,  -2, -28, -10, -21,  29, -21, -18,  -1,   3, -12,   7, -31, -17,  -8,  -3,   0,   9,   3,   2,  -7,  16,  -1,  -1,  -6,  24,  21,  23, -23,  20, -16, -25,  -4 },
   }
 },
 {  //5
   {
     {   82, -50,   2,  -3,   1,  -1,   0,   0, -71,  33,   9,  -1,   1,   0,   1,   0,  17,  10, -17,   2,  -2,   0,  -1,   0,  -2, -13,   5,   4,  -1,   1,   0,   0,   2,   4,   3,  -3,   0,   0,   0,   0,  -2,  -2,  -2,   0,   1,   0,   0,   0,   1,   1,   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0 },
     {   37,  18, -21,  -3,  -3,   0,  -1,   0,   3, -64,  32,   4,   5,   0,   2,   0, -40,  65,  -1, -18,   2,  -3,   1,  -1,  14, -10, -36,  19,   0,   2,  -1,   0,   0,  -8,  24,   4,  -8,   1,  -1,   0,   2,   1,  -4, -12,   4,   1,   0,   0,  -2,   1,   1,   4,   3,  -2,   1,  -1,   1,   0,  -1,  -1,  -2,  -1,   1,   0 },
     {   70,  -4,   3,  -5,  -3,  -2,  -1,   0,  33, -27, -28,   6,  -1,   2,  -1,   1, -19, -38,  52,   1,   0,   0,   1,   0, -21,  45,  -5, -27,   7,  -4,   1,  -1,  -5,  -2, -23,  17,   4,   0,   0,   0,  -2,   3,   9,   3,  -5,  -1,  -1,   0,  -2,  -1,  -3,  -1,  -1,   1,   1,   0,  -1,   1,   1,   0,   0,   1,  -1,   0 },
     {  -33,  -9,  26, -10,   0,  -1,   1,   0, -40,  15, -17,  16,   5,   2,   0,   1,  15,   8,  35, -21,  -6,  -2,   0,  -1, -13,  32, -59,   5,  15,  -2,   1,  -1,  19, -25,  15,  36, -25,   4,  -3,   1,   4,  -3,  15, -35,   5,   6,  -1,   0,   1,   0,  -7,   9,  10,  -8,   2,  -1,   0,   1,   1,  -1,  -6,   1,   2,   0 },
     {   31,  17,  -9,  -9,  -4,   0,  -1,   0,  44,  22, -31,  10,  -2,   0,  -2,   0,  12, -31, -19,  -3,  -4,   3,   0,   1, -14, -26,   4,  39,   3,   3,  -1,   1,   9, -11,  51, -17, -29,   5,  -3,   1,  -9,  -1,  -3, -40,  30,   6,   0,   1,  -5,   7,  -4,  18,  12, -15,   1,  -1,  -3,   2,  -1,  -1,  -9,  -1,   3,   0 },
     {   10,  15,  -5,  10,  -2,  -3,  -2,   0,  11,   8,  -9, -18,  -1,   4,   1,   0,   5,   7, -19,  23,  -4,  -3,  -3,   0,  -1, -21,  -3, -12,   0,   9,   1,   1,  -6,   9,  -2,  46,   0,  -8,  -3,  -1,  10, -19,  26, -28, -50,  22,  -1,   1,   3,  -9,   8, -31,  58,  13,  -9,   0,  -1,   3,   0,  10,   6, -37,   5,   2 },
     {  -17, -27,  11,   8,  -2,   2,  -1,   0, -21, -39,  22,   5,  -1,   6,   1,   1,  -6, -13,  34,   9,  -8,   1,  -2,   0,  11,  10,  43, -18,  -9,   4,  -1,   0,  10,  16,   7, -39, -10,  -1,  -2,  -1,   0, -10, -18, -41,  22,  18,   2,   1,   3,   5,  -2,  10,  50, -13,  -6,   0,  -1,   4,  -6,   9,  -8, -27,   6,   2 },
     {   11,  88, -19,  -9,  -7,  -2,  -2,   0, -58, -17, -28,  17,   4,   5,   1,   1,  28, -13,  23,  14, -10,   0,  -2,   0,  15, -20,   9, -20,  -2,   4,   0,   0,  -3,   0,   6,  -5,  13,  -2,   0,   0,   0,   6,  -7,   2,   1,  -6,   1,  -1,   0,  -1,  -1,   6,  -4,   1,   1,   0,   1,   0,   1,  -2,  -2,   2,  -1,   0 },
     {   11,  20,  11, -17,   0,  -2,  -1,   0,  20,  28,   9, -31,   4,   3,  -1,   0,  19,  22,  -6, -30,   3,  -2,  -2,   0,   4,  -3, -24, -36,  13,   7,   1,   1, -10, -14, -33, -10,  29,   0,   0,   1,  -7, -14, -11,  17,  38,   6,   1,   2,   3,   6,  10,  32,  28, -32,  -7,  -1,  -5,   1,  -5,   9, -27, -29,  12,   2 },
     {   -1,   1, -10,  15, -17,   4,   0,   1,  -1,  -5,  12, -16,  26,  -8,   1,  -1,  -6,   4, -14,  25, -25,   6,   0,   1,   8, -12,  19, -23,  29,  -8,   0,  -1, -12,  18, -25,  33, -29,   0,   4,   0,  16, -18,  31, -25,  23,   5,  -9,   2, -14,  20, -24,  20, -19, -30,  15,  -3,   2,  -4,   2,  -2, -20,  41,  -4,  -2 },
     {    3,  27,   2,  -1,   0,   0,   1,   0, -22,  31, -20, -11,  -3,   0,   0,   0, -68,   2, -18,  30,  10,   4,   2,   1,  23,  40,   7,  18, -30,  -2,  -5,   0,  16, -31, -20, -19, -13,  18,  -1,   3,   8, -20,  21,   3,  18,   9,  -5,   1,   5,  -2,   9, -10,   7, -12,  -3,   0,   0,  -1,  -1,   0,   1,  -4,   5,   0 },
     {    1,  25, -50,   6,   3,   3,   0,   1, -12,  -9,  48,   3, -12,  -2,   0,  -1,  34, -24, -13, -29,  16,   2,   2,   0, -45,  35,   1,  27,   0,  -9,   2,  -2,  10,  15, -27,  -2, -26,  11,  -1,   1,   6, -26,  10,   6,   6,  12,  -5,   2,   4,   2,  12, -13,   5,  -4,  -2,   0,  -1,  -1,  -4,   2,   6,  -3,   1,   0 },
     {  -19, -28, -26,  -5,  15,   2,   1,   1, -28, -41, -33,  -4,  18,   8,   1,   2, -26, -38, -25,   4,  27,   3,   1,   1, -18, -19,  -7,  17,  22,   5,   1,   1,  -4,  -1,   8,  23,  24,  -6,  -2,   0,   5,   9,  15,  26,  10,  -6,  -2,   0,   9,  11,  19,  15,   6, -26,  -8,  -1,  -2,   4,  -2,   1, -24, -24,  11,   1 },
     {   -1,  33,  30, -23,  -1,  -3,  -1,   0, -18,  27,  21, -29,   7,  -2,   1,  -1, -57, -12,   1, -20,  12,   1,   2,   0, -41,   0,  12,   5,  28,  -2,   1,   0,  -2,  46,  12,   8,  -3, -19,  -1,  -3,   0,  14, -32, -11, -16,  -3,   6,   0,   1,  -2, -14,  14,   2,  11,   4,   0,   4,  -2,   1,   4,  -1,   0,  -4,   0 },
     {   10,  10,  53, -26,   6,   0,  -1,  -1,  -2, -23,  44, -36,   5,  -1,  -1,   0,  13, -46,   6, -16,  21,   4,   4,   2,  34, -24,   9,   9,   6,  -7,  -3,  -1,  13, -38,  11,   7,   0,   0,  -1,   2,   9,  -7,  34,   3,   0,  -6,  -1,   3,  -2,  -1,   3, -23,  -7,  -3,   3,   1,   0,   1,  -1,  -3,  10,   7,  -2,  -2 },
     {  -10, -15, -16, -23, -19,   7,   4,   0, -14, -21, -25, -30, -19,   3,   4,   2, -15, -21, -24, -29, -24,  14,   6,   3, -11, -21, -22, -24,  -5,  11,   6,   2, -15, -18, -19, -16,  -6,  24,   9,   3,  -5, -10, -10, -11,  13,  20,   5,   0,  -7,  -5,  -6,   5,  14,  39,   6,   2,   7,   3,  15,  10,  38,  23, -23,  -2 },
     {   -1,   4, -19,  31,  -4,  -7,  -5,   2,  -4,   0,  21, -36,  -1,   9,   3,  -1,   7,  -4,  -8,  19,  15, -11,  -3,   0, -27,  18,  -5, -11, -26,  14,   4,   0,  17, -31,  17,  -3,  33,   9,  -9,   0, -12,  33,  16, -18,  -4, -28,   8,  -1, -14,  20, -53,  12,   3,  11,   2,  -3,  -2,  -5,   5,  29, -11,   2,  -6,   4 },
     {   -3,  -6, -56,  13,   7,   2,   1,   2,  -1,  38,  19,  18, -19,   2,  -2,   0, -27, -24,  24, -35,  10,   9,  -5,   4,  56,  -8,   1,  -9,  27, -17,   0,   1, -19, -14,   5,  11,  -6, -12,   7,   0, -12,  24,   6,  -2,  -5,  11,   3,  -2,   6,   7, -22,  -4,   8,   1,  -8,   2,   2,  -9,   1,  12,   1,  -8,   0,   3 },
     {   -1,   2,   3,  -8,   3, -12,  16,  -4,  -5,  -6,  -2,  -1, -12,  -2, -13,   4,   4,  -5, -13,  -3,  -1, -10,  21,  -2,   0,  10,  -7,  -1, -11,  17, -20,   1, -13,  -5,   1,  -8,  -4, -22,  24,   3,   3,  32, -13,   6, -20,  39, -12,  -3, -19,  -3,   0, -20,  10, -28,  11,   8, -39,  33, -32,  28, -22,  43,  15, -13 },
     {    6,   0,  -1,  16,   8,   2,   4,   1,   1, -16, -10, -15,  10,  -5,   3,   2,  13, -12, -24,   2,  11,  -4,   6,   4,  35,  19,  -2,   6,  10,  -4,  -4,   3,  45,  -7, -16,   9, -20, -31,   8,   0, -20, -13, -37,   2, -23,  12,  14,  -5, -18, -35,  -1,  36,  -3,  13,  -4,   0,  -9, -14,  43,   5, -21,   2,  -3,   1 },
     {   -7,  -6, -20, -49,  -7,  16,   0,   2,  -2,  -1, -20, -35,  23,  11,   0,   2,  17,  29,  13,  -4,  13, -12,   3,  -1,   6,  29,  36,  17,   0, -10,   1,  -1, -27,  -7,   6,   2,  -9, -18,   6,   0, -49,  -2,  14,  -7,  -9,   5,  13,  -1, -23,  36,  10,  -5,  -2,  15,  -8,  -2,   4,   7,  -3, -17,  -1,  -2,   1,  -1 },
     {    3,   4,  20,  -3,   6,  -2,  -2,   2,  -3, -13,   1, -15,  -5,   0,   6,   2,  21, -10, -13,  22,  22,  -2,  -2,  -2,  22,  19, -21,  20, -22, -19,  10,  -5, -70,  24,  -4,  10, -20,  31,   2,  -1,  -5,  34, -15,  -7,   9,   9, -14,   3,  24, -19, -17,  10,   4,  -6,  -4,   1,  15, -19,  13,   4,  -7,  -8,   0,   4 },
     {    6,  -4, -11, -23,   0,  -7,   6,   1,   9,  -4, -21, -15,   5,  16,   5,  -3,  14,  -3, -15, -18,  -2, -10,   2,   2,  38,  16, -10,   5,  -6,  11,   4,  -6,  38,  32, -14,  -7,  -1,  -1, -18,   5,  55,  19, -11,  -8, -11,   0,  17,   2,  27,  41, -19,  -6,  -2,   6,  -5,  -8,  30,   6, -24,  -7,  -6,   5,   1,   2 },
     {   -1,  -6, -20,  48, -22,   1,   1,   4,  -3,  17,  -5, -50,  26,   2,   1,   5,  -2, -13,  46, -11,  12, -23,  15,  -3,  13, -14, -17,  18, -28,  19,   1,   1, -15,  17,  12,   6,   2,  -1,  -5,   1,  16, -23, -14,  -1,  10,  -4,   2,  -2,  -5, -20,  24,   5,  -9,   0,  -3,   6,  -7,  24, -12, -17,   5,  10,  -5,   6 },
     {    1,   2,  -4,  -9,  24, -49,   0,  13,   5,   4,  12,   3,  16, -11,  11, -12,   6,   1,  -9,  14,  -7,  10,  12,   7,  13,  22,   6,  -4,  11,  -6,   4,   4,  15,  14,  19,  20,  19,  14,  13, -10,  -1,  13,   9,   2,  21,  14, -20,  -9, -11,  -1,  19,  31,   8,  19, -31,  17, -11,  23, -17,   2,  36,   9, -25,  36 },
     {   -4,  -7, -24, -55,  11,  23,   1,  -1,   6,  13,  35,  20,  32,  -5,  -2,   4, -10, -30,  -2,  21, -35, -18,   2,   1,  16,   6, -27,  -2, -24,  35,   4,  -3,  -1,  16,   1,   8,  16,  10, -23,   1, -10,  -8, -11,  -7,  -3, -15,  -5,   7, -23, -11,   2,  -8,   0,  -7,   7,  -2, -17, -12,  13,   0,  -5,  -4,   2,   0 },
     {    0,   0, -10,   5, -12, -15,  11,  -1,   1,   1,  18, -21,   6,   6,  -7,   3,   0,  -5,  -2,  14, -33,  -3,  17,   4, -14,  13,  -1, -15,  14,  36,  -9,   5,   5, -21,  22,  -5,  -2, -16,  10,   0, -21,  16,  -1,   3,  -2,  25, -10,  -4,  60,  -7,  15, -11, -12, -10, -16,  13,  34, -29,  12, -34,  -3,   8,  15,   0 },
     {   -3,  -7, -10, -22, -24,  -8,  12,  -1,   0,  -1,  -3, -20, -69,  24,   7,  -6,   5,  12,  11,   9,  -1,  26,  -2,   2,  10,  10,  15,  10,   8,  10,  -6,   6,  11,  12,  15,  31,  12,   2, -26,   3,  -1, -12, -15,   3,  12, -22, -23,  11,  -8, -16,   4, -12,  -5, -22,  11,  15,  12,  22,  28,  14, -11,   7,   4,  16 },
     {    4,   8,  19,  33, -31,  15,  -3,   1,  -3,  -4,   0,  19,  26,   5,  15,  10,   6,  -9, -31, -13,  -9,  33,  10,   2,  19,  40,  -4,  13,  34,  25, -21,  -2, -27,  -1,   7, -12,  15, -16, -36,   8, -11,  -3,   4,   2,  -3, -11,   0,  19,  -9,   2,  -1,   1,   9,   7,  17,  16,  10,  16, -16,   5,  21,  -9,  -3, -11 },
     {   -1,  -4, -15, -21,   5,  10,   3,  -1,   3,   6,   4,  14,  20,   1,  -5,  -1,   5,   9,  -8, -10,  16,  -9,   1,   2, -14,  18,  33, -17,  -7,   0, -10,  14, -31, -42,  23,   6,   6,  -9,  12,   0,  73, -14, -15,  -4,  -7,  -1,  -1,   5,  -6, -31, -13,  18,  -1,  -2,  -6,  10,  26,  10,  17,  -1,   4,  -1,   1,  -1 },
     {   -1,   0,   2,  -9, -24,   4,   7,   9,   0,   1,   4,  -1,  -5,   1,  -7, -17,   3,   8,  14,   1, -32,   0,   5,  -5,  -6,  -1,  11,  37,   3,   4,  -2,  -8,  -7,   4, -14,  -4,   8, -28, -17,   7,  21,  33,  42,  14,  20,  12,  45,  -3,   6, -37,   4,  14,  13,   5, -23, -38, -13,  -8,  16,  23,   0,   6, -10,   8 },
     {   -1,  -3,   1,  -1, -35,  -8, -18, -11,   2,   3,  16,   4,  -5,  25, -23, -18,   1,  -5,   8,  36,  19,  -3, -37, -16,  -2,  -4,  -6,   9,  25,  13,  -9, -18,  -3, -23,  -5,   9,  -5,  11, -14,   7,  22,   5, -34,  15,   4,  33,  -5,  16, -17,  30,   9,   9,   1,  28, -14,   1, -28, -18,  -3, -17,   1,   1,  -1,  -8 },
   },
   {
     {   97, -44,  -3,  -4,  -1,  -1,   0,   0, -66,  17,   9,   2,   2,   1,   1,   0,   6,  11,  -5,  -1,  -1,   0,   0,   0,  -1,  -5,   0,   1,   0,   0,   0,   0,  -1,   3,   1,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   51, -13,  -5,  -1,  -2,   0,  -1,   0,  64, -47,   3,  -2,   0,   0,   0,   0, -77,  15,  18,   0,   4,   0,   1,   0,   4,  25,  -8,  -3,  -1,  -1,   0,   0,  -3,  -4,  -5,   2,   0,   1,   0,   0,  -1,   2,   2,   1,   0,   0,   0,   0,  -2,   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {   24, -28,   8,   0,   1,   0,   1,   0,  50,  15, -18,  -2,  -4,  -1,  -2,   0,  40, -59,   3,   4,   1,   1,   0,   0, -70,  20,  25,  -3,   3,  -1,   1,   0,   7,  24, -13,  -4,  -1,  -1,   0,   0,  -6,  -7,  -2,   3,   1,   0,   0,   0,   0,   3,   1,   0,  -1,   0,   0,   0,  -4,   0,   0,   0,   0,   0,   0,   0 },
     {   33,  85, -50,  -2,  -9,  -1,  -3,   0, -22, -43,   4,  13,   2,   4,   1,   1,  24, -33,  30,  -4,   0,  -1,   0,   0,  -7,  13,  -1, -10,   2,  -1,   1,   0,  -7,   2,  -4,   3,   2,   0,   0,   0,   1,   0,   0,   1,  -1,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   22,   1,  -9,   1,  -1,   0,  -1,   0,  43, -22,   5,  -2,   0,  -1,   0,   0,  36,  14, -25,  -2,  -3,   0,  -1,   0,  14, -70,  11,  13,   1,   3,   1,   1, -52,  21,  39,  -9,   1,  -2,   1,   0,   1,  27, -18,  -9,   1,  -1,   0,   0,  -6,  -5,  -5,   6,   1,   1,   0,   0,  -1,   2,   2,   1,  -1,   0,   0,   0 },
     {   21,  17, -11,   1,  -2,   0,  -1,   0,  18,  66, -47,   0,  -7,   0,  -2,   0, -23, -49,   8,  16,   0,   4,   1,   1,  59, -15,  12,  -9,  -3,  -1,  -1,   0, -11, -25,   2,   0,   4,   0,   1,   0, -16,   8,  12,   0,   1,  -1,   1,   0,  -1,   4,  -5,  -2,  -1,   0,   0,   0,  -2,  -3,   2,   1,   0,   1,   0,   0 },
     {   10, -17,  10,  -3,   1,   0,   0,   0,  23, -17,  -1,   4,   0,   1,   0,   0,  58,   8,   6, -13,  -3,  -3,   0,  -1,  13,  18, -47,  -1,   1,   1,   0,   0,   5, -67,   7,  27,   0,   4,   0,   1, -28,  15,  40, -10,  -3,  -2,   0,   0,  -4,  16, -12, -10,   3,   0,   0,   0,  -5,  -1,  -2,   4,   2,   0,   0,   0 },
     {  -20, -36,  38,  -5,   3,   0,   1,   0, -25, -33,   6,   5,   4,   2,   1,   1,   4, -41,  43,   3,   0,   0,   0,   0,  33,  28,   6, -22,   0,  -3,   0,  -1, -58,  20,  -7,   0,   7,  -1,   2,   0,   7,  35,  -8,   0,  -2,  -2,   0,   0,   7,  -3, -15,   2,  -2,   1,   0,   0,   1,  -1,   1,   2,   0,   1,   0,   0 },
     {  -10,   5,  -3,   3,   0,   0,   0,   0, -21,   7,  -5,   6,  -1,   1,   0,   0, -36, -11,  11,   1,   3,   0,   2,   0, -52,  -5,   4,  15,   3,   2,   1,   1,  13, -16,  54,   0,  -6,  -1,  -1,   0, -14,  62,   3, -36,   2,  -3,   1,  -1,  16,   5, -42,   3,   6,   0,   0,   0,   4, -11,   2,  10,  -2,   0,   0,   0 },
     {   17,  37,  75, -39,  -3,  -8,   0,  -2,  -9, -20, -67,   7,  13,   5,   2,   1,   3,  26, -11,  27,  -9,  -1,  -2,   0, -15,  -3,   8,  -6,  -6,   3,  -1,   0,   0,  -6,  -1,   0,   4,   2,   0,   0,  10,  -3,   2,  -1,  -2,  -1,  -1,   0,   1,  -7,   0,   1,   1,   1,   0,   0,  -2,   2,   3,  -1,   0,   0,   0,   0 },
     {    1,   8, -17,   2,   1,   1,   1,   0,   8,  46, -18,   1,  -3,  -2,  -1,  -1,  -3,  25,  10, -16,  -2,  -3,   0,  -1, -29,  21, -46,  -5,  12,   0,   2,   0, -56,   5,  -1,  34,   0,   0,   0,   1,  54,   1,   0,  -1, -15,   1,  -2,   0,   9, -33,  -3,  -1,   1,   4,   0,   1,  -3,  -5,  13,   1,   2,   0,   0,   0 },
     {   11,  29,  25, -17,   1,  -4,   0,  -1,   0,   7,  45, -43,   5,  -4,   2,  -1,  -8, -33, -67,  19,  19,   2,   3,   0,   7,  40, -17,  35, -12,  -3,  -2,  -1, -26,  -2,   5,  -2, -12,   4,   0,   1,   1,   0,   4,  -3,   3,   3,  -1,   0,   5,  -1,  -6,   0,   1,   0,  -1,   1,   1,  -5,   0,   4,   0,   0,   0,   0 },
     {   -7,   0,  -3,   4,  -3,   2,   0,   0, -10,  -6,   5,   0,   2,   0,   0,   0, -18,   0,  -8,   9,  -1,   1,   0,   0, -20, -25,  24,  -4,   3,   1,   1,   0, -34,   7, -12,  21,   4,   1,   0,   1,  15, -13,  64,   2, -17,   0,  -3,   0,  -8,  53,  -1, -51,   4,  -3,   1,  -1,   6,  -3, -39,   8,  11,   0,   1,   0 },
     {   -3, -18,   6,   3,  -2,   1,   1,   0,  -5, -19, -11,  17,   0,   3,   0,   1,  -3, -39,  13,   9,   2,   2,   0,   0,  15,  -7, -20,  31,  -3,   1,  -1,   0,  20,   4,  47,   7, -26,   3,  -4,   1,  50, -41,  11, -39,  -1,   7,  -1,   1, -27, -10,  18,  -5,  19,  -3,   2,  -1,  -7,  29,   1,  -6,   1,  -6,   1,  -1 },
     {   -7, -17, -36,  33,  -3,   5,  -1,   0,  -4, -15, -20,  20,  -7,   4,   0,   1,  12,  19, -32,  38,  -7,   2,  -3,   1,  -1,  44,  41, -10, -26,   1,  -3,   0, -22, -40,  14, -35,   8,  10,   1,   1,  29,  -1,   5,  -4,  19,  -5,  -1,   0,   4,  -7,  -1,   2,  -3,  -6,   2,  -1,  -3,  -2,   4,  -1,  -2,   1,   2,   0 },
     {    6,   5,  -3,  -3,  -1,   0,  -1,   0,  11,   8,   5, -12,   0,  -1,   1,   0,  15,  18,   0,   2,  -9,   0,  -3,   1,  22,  13,   8, -23,  -1,  -1,   1,   0,  19,  27, -21,  -5,  -5,   1,  -2,  -1,  18, -13,  -6, -42,   1,   3,  -1,   0, -33,   5, -68,   2,  37,  -1,   4,   0,   3, -23,  -3,  50,  -2,  -2,  -2,   0 },
     {    7,  10,  26,  70, -42,   8,  -8,   1,  -6,  -7, -27, -34,   1,   9,   0,   1,   5,   8,  11, -44,  42, -11,   4,  -1,   4,  -1,  19,  30,  -8, -10,   2,  -2,  -3,  -3, -20,   2, -19,   9,   0,   1,   4,   7,   5,  -5,   8,   2,  -2,   1,  -3,  -4,  -1,   2,   1,  -4,   1,   0,   2,  -3,   3,  -1,  -2,   2,   1,   0 },
     {    2,   4,   8,  -8,   2,  -1,   0,   0,   0,   5,  23, -16,   2,  -4,   1,   0,  -5, -10,  -5,  -5,   1,   2,   1,   1, -11, -11,  23, -19,   6,  -2,   1,   0,  13, -43, -18,  13,  15,  -4,   3,  -1,  42,  46, -17,   8, -10,  -4,   0,  -2, -77,  -4,  15,   5,  -5,   4,   2,   1,   9,  23,   5,  -7,   0,   1,  -1,   0 },
     {   -8,   5,  -9,  -3,   1,  -2,   0,  -1,  -8,  15, -20,   5,   4,   1,  -1,   0,  -4,  22,   1,  -7,   7,  -3,  -1,   0, -10,  35, -11,  -1,   3,  -3,   1,   0, -36,  25,  19, -11,  -6,  -1,  -1,   0, -55,  -4,  -5, -12,  15,   5,   2,   0, -64,  32,  17,  17,  11,  -5,   0,  -1,  -1,  39,  -5,  -5,  -6,  -5,  -1,   0 },
     {   16,  26,  30,   0,  -9,  -2,   0,   0,  22,  38,  44,   2, -15,  -4,  -4,   0,  19,  32,  40,   0, -18,  -1,  -2,   1,  15,  29,  31,  -5,  -7,  -4,  -3,  -1,  12,  14,  30, -17,  -9,  -2,  -2,   0,  16,  16,  14, -16,  -3,   2,   0,   0,  19,  10,  16, -17,  -6,  -4,  -3,   0,  -6,  17, -13, -29,   1,  -2,   2,  -1 },
     {    1,   3,   4, -37,  29,  -3,   3,  -1,   2,   0,   1,  61, -42,   6,  -7,   2,  -3,   0, -19, -55,  24,   4,   3,   0,   7,   7,  31,  23,  10, -14,  -1,  -1, -10, -11, -20,   4, -27,   7,   0,   1,   7,   0,  -1, -14,  15,   4,   0,   2,  -1,   4,   0,   6,  -6,  -5,   0,  -1,  10, -20,  -3,  -3,   0,   4,   0,   1 },
     {    0,   9,  30,  12, -11,  -5,  -2,  -2,  -4,   8,  40,   8, -19,   3,   0,   0, -13,  -4,  29,   4, -18,   1,  -1,  -1, -25, -21,  16,  -2, -10,   2,   2,   1, -35, -39,  10,  -7,   5,   6,   1,   1, -31, -52,   2,   2,   4,   1,   1,   1, -22, -35,   0,   7,  10,   3,   1,   1, -12, -25,  18,  23,   5,   1,   0,   0 },
     {   -4,  -8, -21, -39,  45, -16,   7,  -1,  -2,  -1, -11, -57,  43, -15,   9,  -2,   5,   7,  29, -31,  19,  -8,   2,  -2,  -2,   0,  37,   2, -10,  -1,  -1,  -1, -11, -25,  20,  -9,  -6,   3,  -1,   0,   8, -27,   8,   3,   4,   1,   0,   0,  15,  -4,  -8,  12,   1,   0,   0,   0,  -1,  10,  -7,   1,  -6,  -2,   2,   0 },
     {    3,   1,   9,   4,  -7,   3,  -3,   2,   3,   4,  -8,  -5,   4,  -4,   1,   0,   5,  -4,  15,  -7, -13,   2,   2,   0,  -1,   8, -20,   7,  -7,   1,  -5,   1,  -2, -11,  31, -11,   2,  -2,  -3,   2,  19, -18, -41,  18,   6,   3,   3,   0, -13,  47,   7,   1, -27,   2,  -2,   1,  67, -54, -21,   4,   1,   5,  -3,   1 },
     {   -3,  -3, -14, -14,  17,   0,   7,  -3,   2,   6,   0,   1,  11,  -6,   2,   1,   5,   8,  33,  32, -19,  -9,   1,   1,  -5, -11, -12,  60, -49,   9,   4,  -1,  -5,   9, -49, -13,  -9,  19,  -2,   1,  -5,  28,  15, -17,  16,  -7,  -3,   0, -15, -14,  16,  -5,   6,  -4,   4,   0,  24, -16,   3,  -7,   2,   1,   2,   0 },
     {    1,   0,   2,  -7,   2,  -2,   4,  -2,   0,   5,  -8,   6, -13,   8,  -2,   1,   0,  -4,   3, -20,  15,  -7,   1,   1,  -3,   2, -21,   9, -22,   8,  -1,   2,   2, -11,   2, -41,  21,   3,  -3,  -1,   2,  11, -35,  18,  -3,   1,  -3,   2,   0, -19,   5, -47,  17,   2,   3,   0, -30,  17, -64,  36,  25,  -4,   1,   0 },
     {    6,   9,   8,  37,  70, -28,  -6,  -7,  -7, -10, -12, -36, -64,  13,   9,   4,   3,   3,   9,  18, -17,  15,   1,  -2,  -2,  -4, -12,   7,  31,  -9,  -8,   1,   1,   2,  -4, -10,  -4,   0,   0,   1,   1,   2,   4,   5,   3,   4,   2,   1,   0,   1,  -4,   0,  -6,  -7,   0,   2,  12,  16,  -4,   7,  -3,  -1,   2,  -2 },
     {    3,   7,  12,  -5,  -5,   4,  -3,   3,   3,   9,  12,  10,  -6,   5,  -7,   0,  -6,  -9,   6, -15,  12,  -9,   9,   0, -11, -22, -34,  -6,  -7,  15,  -4,  -1,  -5,  -4, -15, -49,  33,  15,  -4,   4,  31, -11,  22,   8,  38, -21,   4,   0,  10,  38, -30,  33,   0, -12,   2,  -6,  -8,  29,  16, -13, -14,   8,  -1,   2 },
     {   -2,  -3,  -5,  -6,   8, -11,  -2,  -1,  -1,  -1,  -7,   3,  -3,  20,  -2,   2,   5,   6,   5,  -3,  24,  -1,  -1,  -2,   8,  12,  -1,  21,  -5,  -6,  -3,   0,  12,  25,  23, -25,  19, -11,  -2,   0,   1,   0,  35,  29, -33, -17,   9,  -1, -39, -13, -27,  -9, -38,  32,   0,   1, -19, -29,  27, -15,  29,  14, -12,   2 },
     {    0,   4,   1,   9,  19,   8,   4,   0,  -2,   2,   0,   0,   9,   3,   6,  -1,  -1,   4,  12,  16,  -9,  -9,  -3,   2,   0,  -1,   3,  24,   4,   4,  -1,   1,  -5, -11, -11,  17, -16,   0,  -3,   6,  11,  -1, -43,   0,   1,   6,   4,   1,  -3,  56,  -2,   0, -12,  -1,  -1,   0, -86, -25,  13,  -6,  -1,  -1,   2,   2 },
     {   -5,  -8, -12, -29, -19,  -5,  -8,  -4,   0,   1,   4, -16, -44,  25,   2,   8,   5,   9,  14,  37,  51, -27,   3,  -6,  -5, -12,  -2, -30, -30,  -5,   5,   0,   4,   5,  12,  32, -26,  33, -18,   5,  -1,  -2, -13,  15,  21,  10,   2,  -5,   0,   4,  -9, -10, -12,  -8,  -3,   2,   5,   6,   7,  -2,   4,   3,  -4,  -7 },
     {   -5,  -7, -11, -32, -23,  37,   0,   1,  -5,  -6,  -9, -41, -32,  41,  -6,  -1,   0,   3,   3, -10,  -9,  29,  -8,   0,   0,   5,   9,  12,  12,  13,  -6,   0,  -6,  -6,  -7,   2,  34, -17, -14,   7,  -4,  -8, -11, -41,   5, -30,  27,   0,  10,   6,  19, -13,   7,   9,  12,  -4,   1,  -4,  14, -11,  11, -12,  -7,   0 },
   },
   {
     {  119, -32, -11,  -5,  -3,  -2,  -1,  -1, -28,  -5,   2,   1,   1,   1,   1,   0, -12,   5,   3,   1,   1,   0,   0,   0,  -7,   1,   1,   0,   0,   0,   0,   0,  -5,   1,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   34,  42, -19,  -6,  -5,  -2,  -2,   0, 101,   5, -30,  -6,  -5,  -2,  -2,   0,   3, -25, -11,   5,   2,   1,   0,   0, -17, -14,   5,   8,   2,   1,   0,   0, -14,  -5,   7,   3,   0,   0,   0,   0, -10,   0,   5,   0,  -1,   0,   0,   0,  -6,   1,   2,   0,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0 },
     {  -12, -67,  -6,  20,   6,   4,   1,   1,  35, -85,   0,  22,   4,   3,   0,   1,   4, -29,  20,  12,  -1,   0,  -1,   0,  12,   7,  17,  -3,  -5,  -1,  -1,   0,   4,  10,   6,  -8,  -2,   0,   0,   0,  -2,   9,   1,  -5,   1,   0,   0,   0,  -2,   6,   0,  -2,   1,  -1,   0,   0,  -1,   4,   0,  -1,   0,   0,   0,   0 },
     {   -5,  42,   5,  -6,  -3,  -1,   0,   0, -19, -21,   5,  -2,   2,   2,   1,   0, -91, -33,  24,  10,   8,   4,   2,   1, -36,  15,  29,   5,   1,  -1,   0,   0,   1,  23,   8,  -8,  -4,  -3,  -1,   0,   9,  10,  -8,  -7,  -1,   0,   0,   0,   6,   1,  -7,  -1,   2,   0,   0,   0,   5,   1,  -3,   1,   1,   0,   0,   0 },
     {   13,  10,  44,  -9, -11,  -2,  -1,  -1,  41,  -5,  59, -17, -14,  -2,  -1,  -1,  -4,  33,  43, -28, -10,  -1,   0,  -1,  13,  32,  -3, -35,  -3,   1,   0,   0,   0,   2, -27, -19,   8,   3,   1,   0,  -6,  -6, -20,   1,  12,   2,   0,   0,  -4,  -4,  -9,   6,   6,  -1,   0,   0,  -2,  -1,  -4,   4,   2,  -1,   0,   0 },
     {    3,  27, -53,   7,   5,   0,  -1,   0,   0, -19, -46,  15,  10,   0,   1,   0, -15,  40,  -5,  11,   3,  -3,   0,  -1,  42,  60, -13,  -9, -10,  -6,  -2,  -1,  22,  11, -28, -10,  -5,   0,   0,   0,   2, -13, -15,   5,   3,   3,   1,   0,  -2, -11,   1,  10,   2,   1,  -1,   0,  -1,  -5,   4,   4,   0,   0,  -1,   0 },
     {   12,  14,  15,   1,  -6,  -2,  -1,  -1, -11,  29,  -3,   7,  -9,   0,  -1,   0,  39, -41,   5,  18, -10,   2,  -1,   1,  34,   6,  51,   6, -20,  -3,  -2,   0,  35,  21,  15, -41, -25,   0,  -1,   0,  -1, -12, -24, -35,   6,  13,   3,   1,  -4,  -7,  -9,   0,  20,   5,  -1,  -1,  -2,   0,  -1,   4,   7,  -4,  -2,   0 },
     {   16,  41,  31,  -6,  -9,  -4,  -3,  -1, -13, -46,  15,  -9,   1,   2,   0,   0,   2, -46, -25,  -6,  14,   6,   2,   1,  50, -21, -36,   7,  14,   2,  -1,   0,  38, -16, -17,  20,   7,  -4,  -2,   0,  14, -12,   2,  17,  -4,  -5,  -1,   0,  -2,  -5,   8,   5,  -7,  -1,   1,   1,  -2,   1,   5,   0,  -3,   1,   1,   0 },
     {    7,  35,   1,  56, -20, -10,  -3,  -1,  -7,  -5,  23,  70, -24,  -9,  -1,  -1,  -1,  13,  19,  32, -29,  -6,   1,  -1,   0, -16,  -9,   2, -24,   4,   3,   0, -16, -29,   1,   2,  -8,  10,   3,   1,  -3,  -5,  16,  -1,  -4,   5,  -1,   0,   1,   3,  10, -10,  -5,   3,  -1,   0,   1,   1,   2,  -9,  -1,   2,  -1,   0 },
     {   -1,  44, -46, -37,   0,   8,   1,   1, -29, -42,   8, -22,  11,  10,   2,   1,  47,   6,  47,  -6,   2,  -1,  -3,  -2,  -1, -14,  27, -13,  -3,  -4,  -2,   0, -27, -21,  15,  -4,   6,   1,   1,   1,  -8,   4,  13,   2,   6,  -2,   0,   0,   1,   7,   3,  -2,  -1,  -4,  -1,   0,   1,   3,  -3,  -3,  -1,  -2,   1,   0 },
     {   -7, -11,  -6,   5,   2,   0,   0,   0, -14,   5, -19,   5,   0,   2,   0,   1,  -9, -35,  -8,   0,  -2,   6,   3,   2, -18, -18,  -4, -27,  -3,   9,   2,   1, -25, -32, -43, -35,  17,  16,   4,   1, -21, -28, -23,  17,  46,  13,   0,  -1,   2,   4,  14,  38,  25, -10,  -7,  -3,   3,   6,  10,  14,  -4, -15,  -2,   0 },
     {    6,  28,  26,  27,   4, -10,  -4,  -1, -18, -17,  -5,  19,   9,  -6,  -3,   0,  43, -16, -30, -12,  12,   1,   1,   1, -27,  41, -12, -13,  18,  -3,   2,   0, -39,  49,  -4,  -2,  11,  -8,   0,  -1, -31,  29,  -8,   0,  -1,  -8,   0,   0, -12,  13,  -9,   2,  -2,  -2,   2,   1,  -1,   5,  -3,   3,  -1,   0,   1,   0 },
     {    4,  -1,  27,  -6,  45, -12,  -7,  -1,   8,  20,   8,  15,  57, -16, -10,  -2,   5,  -8,  32,  36,  35, -22,  -8,  -1,   4,  15,  17,  16,   1, -26,  -2,   0,  -7, -21, -19,  12, -12, -15,   6,   2, -12, -18,   3,  23, -11,  -5,   7,   1,  -1,   1,  12,  10, -16,  -3,   3,  -1,   1,   2,   4,  -1, -13,  -1,   1,   0 },
     {   -5,  -4,  25,  -3, -12,   3,   2,  -1,  -6,  -9,  -8, -19, -15,   9,   4,   0, -25,  10, -23,  -9,  -2,  12,   4,   1,  33,  29,  24,  18,  -3,   2,  -4,  -2, -34, -20,  34,   6, -12,  -3,  -3,   0, -60, -35,  31,   0,  -1,   6,   2,   2, -23,   1,  26,  -2,   5,   4,   0,   0,  -2,   9,   8,  -7,   1,   0,  -1,   0 },
     {   -1, -12,  13, -11, -27,   7,   6,   0,  -6,  26, -37,   0, -36,  13,   5,   0,  14, -40,  38,   8, -16,  14,   2,  -1,   0,  43,  10, -18,  -1,   7,  -2,  -2,   5,  -6, -24,  22,  18,   3,  -5,  -1,  11,  10,  27,  35,  -8, -12,  -7,   0,   8,   3,   8, -11, -30,  -4,   0,   4,  -2,  -5,  -3, -13,  -8,  10,   2,   2 },
     {    6,   7,  -4,  32,  46,  -7, -12,  -3,   0,  14,  -9,   3,  31, -15, -11,  -1, -15, -17,  -1, -54, -13,  -5,   5,   5,  19,  -3,  12, -57, -10,  15,  10,   3,  10,  -6,  32, -11,  14,  13,   2,  -3,   8,   1,  35,   3,  -2,  -5,  -6,  -4,   2,  -8,   3, -12, -10,  -3,   0,   0,   0,  -7,  -6,  -5,   1,   4,   4,   1 },
     {   -6,  -5, -48,  25,  -5,   3,  -2,   3,  -3,  17,  35,  -3, -10,  -2,  -3,   2,   1, -21,  23, -13,   0,  -1,   0,   0, -11, -10,  -2,   2,  18,  -1,  -2,   0,  21,  38,  -3,  19,   8, -10,  -1,   0, -29, -28,  -9,  17,  -9,  -7,   2,   3, -52, -36,  19,   8,  -8,   3,   5,   3, -11,   6,  23,  -4,  -3,   4,   1,   0 },
     {    4,   6,  26,   7,  -4,  -6,  -1,   0,   0, -28, -13,   3,   2,   6,   2,   1,  26,  21, -15, -12,  -6,   3,   1,   2, -58,  18,  24,   0,   0,   4,   0,  -2,  18, -10,  19,   3,  -5,  -5,  -2,  -1,  40, -50,  -3,  11,  -4,  -1,  -2,   0,  18, -47,  21,  12,  -2,   2,   0,   0,   2,  -6,  23,   4,   1,  -2,  -2,  -3 },
     {    5,   6,   4,  10,   1,   9,  -4,  -1,   9,  13,   9,  11,  -4,  17,  -6,   2,  -2,  21,   4,   7,   2,  13,  -2,   0,  28,  -6,  23,   9,  13,   2,  -4,  -1,   4,  28,  33,   7,  22,  -5,  -6,  -3,  18,  33,   4,  26,  28, -15,  -7,  -2,  20,  21,  34,  48,   9, -23,  -5,  -1,  17,  14,  34,   8, -21, -17,  -1,   0 },
     {    1,   0, -34,  19,   6,   8, -10,  -1,   5,   5,  45, -26,  -4,   7,  -1,  -3,   7, -17, -30,  -6,  -4,   3,  11,   4, -15,  28,   3,   2, -10, -16,  -1,   1,  11,   0, -20, -11, -20,  -7,   3,   3,   5, -15,   9,   3, -12,  13,   5,   1,  45,  46,  39, -11, -10,   7,  -1,  -7,  20,  21, -11, -17,   2,   0,  -1,  -4 },
     {   -1,   4,  18, -42,   5,   9,   4,   1,   0,  -7,  -3,   5,  14,  -6,  -2,  -1,   1,  12,  -6,  25,  -9,  -5,  -9,  -2, -23,  -4, -24, -14, -26,  16,   8,   4,  29,  23,  -5, -33,  -5,  23,   4,   0,  -3,  20,  33,  -1,   1,   9,  -3,  -3, -36,   6,  52,  -4,   2,   2,  -1,  -2, -24,  24,  13, -22,   5,   0,   2,  -1 },
     {    0,   0,  -7,   4,   0,  -1,  -1,   1,   0,   7,   3,  10,   2,   4,  -3,  -1,   7, -13,  22, -15,   9,   8,   1,  -1, -30,  39, -36,  10,   8,  -1,  -8,  -1,  52, -49,  30,  11,  -8,  -7,  -6,   2, -25,  19,  17, -26,  17,   3,   4,   1,  -5,  19, -19,  14,  19,  -4,  -4,  -2,  -3,  10,  19,  25,   3, -13,  -5,   0 },
     {   -5,  -7, -11, -11, -18, -65,  -4,   9,  -3,  -5,   0, -14, -22, -74,   1,  16,   1,   1,  -5,   8,  -3, -29,  12,  11,   2,  12,  13,   9,  11,  -4,  19,   4,   9,  -2,  13,   3,  23,   0,   7,  -5,  11,  -4,   4,   8,  13,  -5,  -4,  -7,  -5,  13,   8,   5,   6, -10,   0,  -5,  -8,  13, -22,   3,   4, -17,  -2,  -4 },
     {   -2,  -4,  -7,   6, -22,  -1,  -4,  -3,   3,   9,   6,  11, -11,  10,  -9,   1,  -3,   7,  -1,  22,  57,   8,  -6,  -4,   6,  -1,  -1, -25,  57,   3,  -5,  -5,  -3, -13,   0, -51,  16, -19,  -4,  -1,   8,  -5,  25, -29,  -9, -19,  -2,   0,   9, -19,  22, -25,  -5,  -7,   4,   4,  -6,  -1,  -9,  11,   4,  -4,   4,   0 },
     {    3,   5,  18,  44,   1, -31, -15,   2,  -5,  -8, -39, -36,   7,  20,  14,   0,   1,  19,  30, -20, -28,  -3,  10,   8,  -5, -25,   6,  29,  24, -10, -20,  -9,  13,  -3, -32, -29, -15,  -7,   5,   7,  -1,  14,   7,  -5, -13, -15,   1,   2, -18,  17,  13,  -2,  -1,   2,   7,   0,   0,  14,   0,  -1,  -6,  -5,   0,   4 },
     {   -1,   1,   0, -30, -13,   8,  -2,   3,   3,  -1,  -5,  49,   3,   4,   2,   3,  -2,   5,   4, -39, -12,  -3,  -5,  -1,  -6, -11,  -8,  24,  24, -14,   6,  -3,  20,  21,  10, -30, -19,  -2,   7,   1, -22, -32,  12,  25,  18,  -5,   1,  -2,  16,  18, -14,   1, -15, -26,   1,   2,  -2, -17, -37, -18, -28,   3,   9,   3 },
     {   -4,  -8,   0, -28, -27, -29,  -7,  -1,   0,   3, -16,  22,   4, -15,  -5,   1,  -4, -10,  16, -31,   3, -13,  -4,  -4,  11,  -9, -20, -15,   5,  -3, -14,  -5, -12,   4,  12,  -7, -25,  -4,   4,   4, -15,  18, -23,  -1, -22,  18,  15,   2,  28, -20,  22,   6, -19,  22,  -5,   0,  47,  25,  33,  -3,  18,  10, -13,  -1 },
     {    1,   3,   8, -12,  11,   9,   0, -13,   1,  -1, -14,  17,   3,   8,  -2,   0,  -3,   8,  -2,   0, -14, -16,  -9,   0,  -4,  -7,  10, -11,   3,  -3,  -6,  -5,  16,  -1,  -8,  28,  45, -11, -11,  -8, -10, -30, -28, -36,  12, -26,  -4,   8,  -4,  22,  -2, -44,   1, -18,  -3,   7,  27,  48,  11, -26,   3, -12,  -4,  -4 },
     {   -4,  -5,  -4, -21, -20, -11, -35,  -1,   0,   3,   6,  20,   4,  30, -26,   7,  -9,  -4, -21, -11, -23,  -2,  -5,   7,   1,  -3,  10, -15, -28, -54, -12,  -5,  -7,  -6,  -4,   2,  13, -37,  13,  -3,  13,   7,   8,   5,  -7, -17,  18,  -9, -16,   6,  -2,   9,  -2,   3,  26,   0, -32,  -3,   3,  24,  30, -13,   4,   0 },
     {    1,   3,   8,  18, -33,  19,  32,  11,  -1,  -4, -11,   1,   7,  -9,  21, -14,  -2,  15, -12,  -6,  23, -19, -15,  -5,  -8, -18,  35, -36,   5, -24,  26,  15,  32, -11,  -7,  18, -15,  -5,  -3,  -1, -17,  16,  -7,  13,   9, -11,  15,  12, -16,  27,  15,   4,   0,  14,  -1,  -2,  16, -31,  -6,   1,  19,  10, -13,  -6 },
     {   -1,  -1,   3,  -8,  -3, -18, -17,  -4,   1,  -4,  -3,  -8,  -1,  14,  -4,  -1,   1,   7,   1,  33,  -7,   1,   0,   5,  -8, -15, -11, -36, -16,  22,   0,  -6,  14,  23,   8,   6,  -9, -14, -26,   3, -18, -28,  19,  10, -28,  -7, -13,  -3,  -6,  16, -14,  15,   6,  11,  -1,   1,  44,   3, -30,  54, -24,  -2,  -6,   4 },
     {    0,   1,   0,  -7,  21, -12, -10, -13,   1,   3,   3,  13,   2,  23,  -5,  -6,   1,   3,   2,  14, -13,  19,  13,  -4,   0,   8,  -1,  -6,  42,   3, -12, -15,   3, -16,  31,   4,   0,  27, -11,  -7,  12,   0, -44,  24,   8,  11,  -7,   3, -26,  18,  20,   5, -11,  35,  -5,   8,  -6,   2, -44,  -6,  39,  10,  -1,   1 },
   }
 },
 {  //6
   {
     {  102, -42,  -4,  -4,  -1,  -1,  -1,   0, -61,   5,  11,   2,   3,   1,   1,   0,  -1,  17,  -1,  -2,  -1,  -1,   0,   0,  -3,  -1,  -5,   0,   0,   0,   0,   0,   0,   2,   1,   1,   0,   0,   0,   0,  -1,   0,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {   54,  10, -11,  -2,  -4,  -1,  -1,   0,  72, -53,  -7,   1,  -1,   1,   0,   0, -49, -21,  29,   2,   4,   1,   1,   0, -16,  28,   9, -10,   0,  -2,   0,   0,  -8,   7, -11,  -3,   3,   0,   0,   0,  -3,   2,  -2,   5,   1,   0,   0,   0,  -3,   0,   1,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {  -26,  12,   3,   0,   1,   0,   0,   0, -33, -38,  22,   2,   5,   1,   2,   0, -40,  59,  25, -12,   1,  -4,   1,  -1,  25,  38, -45, -14,   0,  -2,   0,   0,   6, -21, -25,  22,   5,   2,   0,   1,  10,  -6,  11,  13,  -8,  -1,  -1,   0,   4,  -4,   4,  -5,  -5,   2,   0,   0,   2,  -1,   1,  -2,   1,   1,  -1,   0 },
     {  -29, -42,  21,   3,   5,   2,   2,   0, -31,   1,  12,  -1,   1,   0,   1,   0,  -2, -31,  30,   3,   2,   1,   1,   0,  -7,  45,  41, -27,  -2,  -5,   0,  -1,   6,  33, -41, -33,   9,  -2,   2,   0,  -3,  -3, -28,  19,  17,  -1,   1,   0,   5,   3,   0,  16,  -6,  -6,  -1,  -1,   1,  -1,  -1,   1,  -7,   1,   2,   0 },
     {   13,  78, -27,  -4,  -6,  -2,  -2,   0, -46,  20, -36,  10,   0,   2,   0,   1, -34, -26,   6,  26,   1,   4,   0,   1,  25,   7,  26,  -2, -13,  -1,  -2,   0,  18, -10,  -5, -19,  -3,   4,   0,   0,   4, -13,  -1,   1,  10,   2,   0,   0,   2,  -3,   5,   2,   1,  -3,  -1,   0,   1,  -1,   1,  -2,  -1,   0,   1,   0 },
     {  -18,  -6,  -5,   5,  -2,   2,   0,   1, -23, -62,  24,  12,   3,   4,   1,   1, -17,   9,  15,   2,  -2,   1,   0,   0,  15, -23,  26,  17,  -5,   2,  -1,   0,  10,  16,  56, -28, -16,  -2,  -2,   0,   0,  16, -16, -48,  10,   5,   1,   1,  -5,   5, -15,   4,  23,  -2,   0,   0,   1,   5,  -3,   7,   0,  -7,   0,   0 },
     {   11,  18,  -4,  -1,  -3,   0,  -1,   0,  -7, -45,   2,   9,   0,   4,   0,   1,  75, -24,  35,  -3,  -6,  -1,  -3,   0,  38, -18,   1, -34,  -1,   0,   0,   0, -21, -30,  -8,  -2,  24,   3,   3,   0, -15,  11,  17,  14,   7,  -9,  -1,  -1,  -7,  12,   2,  -4,  -8,  -5,   2,   0,  -4,   4,  -4,  -3,  -1,   3,   2,   0 },
     {   11,  26, -10, -10,   0,  -2,   0,   0,  19,   4,  16,  -8,   5,  -3,   0,  -1,   4,  41, -24, -22,   2,  -3,   1,  -1,  15, -18, -12,   3,   5,   0,   0,   0,  -9,   5, -10, -34,   9,  -2,   2,   0, -16,  -3, -51,  20,  41,  -2,   4,   0,  -2,   4,   4,  56,  -7, -19,  -1,  -2,   0,  -2,   2,   2, -32,   0,   6,   0 },
     {    1,  51, -13,  -4,  -2,  -3,  -1,   0, -25, -25,   2,   6,   5,   2,   2,   0,  39,   8,   6,  -9,  -2,  -3,  -1,  -1, -75,  10,  -9,  -2,  10,   1,   3,   0, -10,  49,  -3,  18,  -1,  -5,  -1,  -1,  20,   3, -22,  -1, -18,  -1,  -1,   0,   4, -17,  -4,   3,   1,   9,   1,   1,   5,  -2,   5,   3,   1,   0,  -2,   0 },
     {   12,  24,  79, -24,  -3,  -7,  -1,  -1, -11, -29,   1, -38,  10,   0,   2,   1,   9, -19, -39,  14,  27,   1,   2,   0,  18,  24,  -3,  32,  -6, -13,  -2,  -1, -19,   1,  -7,   2, -22,   0,   4,   1,  -7,   9,   0,   0,  -3,  10,   1,  -1,   2,   3,  -4,   0,   3,   2,  -4,   0,   1,  -2,   0,   2,   1,  -2,  -2,   1 },
     {  -14, -16, -15,   8,   2,   0,   2,   0, -18, -28,  -9,  22,   6,  -1,   2,   0, -19, -14, -14,  20,   6,  -1,   2,  -1, -10, -16,   7,  38,   3,  -5,   1,  -1, -12,   6,  17,  20,  -2,  -7,   3,  -2,  -3,  23,  12,  44,  -6, -13,   2,  -2,   7,  15,  22,  13, -59, -11,   0,  -2,  -1,  -5,  -2, -29, -32,  25,   9,   1 },
     {    1, -18, -13,   9,   0,   1,   0,   1,  10, -35, -12,  15,  -1,   3,  -1,   1,  38,  10, -21,  14,  -4,   0,  -1,   0,  -4,  17,   4,  24, -16,   0,  -2,  -1,  39, -11, -14, -11, -26,   9,  -2,   1,  10, -66,   3,   1,   9,  18,  -1,   2,  -2, -14,  47,   3,  10,  -2,  -4,   0,  -3,   7,  12, -17,  -2,  -7,   0,   1 },
     {    2,   7, -15,  11,  -1,   3,  -1,   1,   5, -11,  35, -16,  -9,  -1,   0,   0,  -8,  24, -34,   8,   4,   3,   0,   0,  23, -30,  39,  -5, -18,   1,  -2,   1, -20,  29, -19, -25,  19,   0,   2,   0,  13,  -6,  -4,  33, -25,  -4,  -2,   0,   1, -16,  12, -44, -10,  22,   1,   3,   1,  -1,  -1,  -3,  37,   3,  -8,  -1 },
     {   -5, -13,  -6,  16,   0,   0,   1,   0,   2,   0, -48,  20,  -1,  -1,   0,   0,  11,  46, -10,  37, -10,  -2,   1,  -1, -12,  30,   7,  -4, -36,  -1,  -1,   0, -62, -10,   0, -22,  -4,  23,   3,   2,   3,  28,   6,  -7,  18,   8,  -7,  -1,  10,  -4, -15,  -3,   5,  -8,  -5,   1,   3,  -9,   0,   6,   5,  -3,   0,   2 },
     {    1,  11,  43,  -8,  -6,  -1,  -2,   1,  -3,  -3,  13, -19,  -1,   4,  -2,   1, -12,  15,   1,   3,  -4,   4,  -2,   1, -58, -19,  15, -18, -11,  12,  -1,   3,  21, -34,  16, -26,  29,  15,  -3,   2,  33,  -8,  29,   5,  34, -21,  -7,   0,  -4,   9,   7, -12, -12, -24,   5,   1,  -5,   6, -10, -11,   0,   5,   8,   0 },
     {   -1,  11, -49,  21,   2,   5,  -1,   1,  -6,   9,  58, -14, -20,   0,  -3,  -1,   8, -25, -12, -28,  24,   5,   3,   2, -17,  36,  -3,  27,   1, -13,  -1,  -1, -19, -24,  -3, -15, -24,  14,   4,   1,   0,  10,  25,  -2,  25,  12,  -4,  -2,   9,  16,  -8, -13,   1, -18,  -5,   2,  -1,  -8, -10,   2,   5,  -3,   3,   2 },
     {   18,  24,  22,   4,  -2,  -7,  -3,   0,  27,  32,  27,  22,  -5,  -5,  -3,   1,  28,  38,  25,   7, -12,  -7,  -2,  -1,  25,  40,  29,  11,  -4,  -4,  -3,  -2,  31,  30,  23,   1, -10,  -6,  -1,  -2,   8,  21,  17,  -5, -11,  -2,   1,  -1,  17,  23,   7,  -6, -19,  -3,   0,  -3,   4,   3,  -7, -19, -15,   8,   1,   0 },
     {    7,   6,  28,  79, -18, -14,  -7,  -5,  -5,  -7, -29,  16, -21,   5,   2,   1,  -3,   3, -23, -50,  14,  20,   6,   1,   7,  12,  18, -13,  35,  -1,  -7,   0,  -7, -12,  11,  -2,   9, -18,  -1,   1,  12,  -7,  -6,  -6,  -7,  -3,  10,   0,   2,  -9,   1,   4,  -6,   2,   3,  -3,  -8,  -6,   1,  -1,  -3,   2,   1,   0 },
     {    6, -11, -10,   0,   1,   2,   1,   0,  11, -10,   2,  -2,  -1,   0,  -4,  -2,  15, -14, -11,  14,   6,  -2,  -2,   1,  20,   0, -10,   1,  -9,  -2,  -1,   2,  40, -13,  -5,   2,  -2,   2,  -1,   1,  70,  26, -16,   8,   5,  -8,  -4,   0,  18, -39, -44,   8, -13,  -3,   0,   0, -12, -44,   9,  17,  -6,  10,   2,   3 },
     {   -1,   0, -10,  10,  -2, -11,   1,  -3,   0, -14, -14,  18,   7,   2,   2,   0,   7,   7, -35,   9,   1,   3,  -4,  -3,  -8,  13, -11,   8,  -7,   1,  -2,  -1,  45,  10, -22,   0,  16,  -7,   0,  -3, -24,   9,  -6,   6,   6, -12,   2,  -1, -46,  43, -34, -20,   1,   1,   7,   0,  -7,   1, -56,  26,   9,   9,   0,  -2 },
     {    0,  -2,   3,  -6,   7,  -2,   0,   0,  -1,   7, -14,   2,  -9,  -3,  -2,   0,   6,  -5,  18, -18,  10,  -2,   5,   0, -10,  14, -21,  20, -17,   2,  -9,  -1,  23, -28,  19, -24,   3,  -3,   1,   3, -20,  49, -48,  16,  -2,   0,   4,   1,   6, -37,  33, -19, -10,  15,  -2,   1, -16,  29, -17, -17,  34,  -5,  -2,  -2 },
     {    3,  10,  28,  25, -32, -10,   5,   5,  -1,   4,  46,  52, -32, -14,  -2,   2, -11, -27, -14,  28, -40,   0,   4,   4,  -3,  -5, -27,   7, -19,  13,   8,  -3,  -2,   3, -13,  11,  19,  14,   1,  -5, -16,   8,  -3,  -4,  10,   7,  -5,  -5, -12, -21,   4,  16,   7,  -1,  -4,  -2,   3,   2,  19,   3,   8,  -2,  -5,   1 },
     {    4,   2,   9,  49,  -9,  -6,  -3,  -2,   1,   3, -18, -33, -17,  19,   6,  -5,   2,  12,  47,   9,  17,  -3,   0,   1,  -3, -44, -18,  26,  -8, -24,   2,  -4,  21,  25, -34,  -7, -19,  16,  16,  -1,  -7,  17,  14,   9,   8,  10,  -6,  -1,  -9,  -4,  -9,  -1,   5, -17,  -9,   1,  25,   9,   9,  -4,   0,   3,   2,   0 },
     {    0,  -3,   2,  -7, -20,  -5,  -2,  -2,   2,   0,   3,  -7, -16, -14,  12,   3,   3,  -1,   8,  -2,  -5,  -4,  18,   2,  -8,  -4,   4,  -8, -16, -12,  15,  -7,  -5, -33,   1,   1, -10,   6,   9,  -4,  -1, -24, -42, -25, -10,   7,  10,   1,  -3,  13, -15, -12, -29,  24,   3,   0,  67, -25, -19, -27,  -5,  33,  -8,  -1 },
     {    1,  -1,  13, -11, -15,  -6,  -1,  -2,   1,  -4, -10,  29,  -1,  10,   5,  -6,  -8, -13,  -1, -10,  -6,  -9,  -1,   5,   5, -25, -33, -11,  11,  -3,   5,   4,  -6,  24,  -8, -17, -11,   4,   2,  -1,  11,  -9,  -3,   4,  12,  19,  -3,   2,  71,  23,  15, -22,  12,  -7,  -9,   0,  -8, -25, -53,  -7,   7, -11,   8,   3 },
     {    5,   8,   9,  21,  67,  -6,   4,  -5,  -4,  -7,   0, -19, -25, -30,   7, -11,  -4,  -7,  -3,  -2, -43,  15,  11,  -3,   2,  -7,   0, -19, -10,  34,  -1,   2,  -2,  -3,   3,  12, -43,   3, -10,   3,  16,   6, -11,  29,  -5,  17,  -4,   3,  -8,  13,  -7,   3,  15,  -2,  -7,   7,  -2,  13, -21, -10, -13,   0,   2,   3 },
     {    0,   0,   0,   3,  10,  -7, -10,   7,   2,   2,  10,   8,   3,  17,   6,  -8,  -3,   3,   5,   3,  -2, -19,  -4,   9,  -8, -15,  29,  27,  -6,  -3,   3,  -9, -11, -35, -46,  12,  17,  -2,  17, -14,  37,  20, -12, -27, -33,  10,  23,   8,   8,  23,  19,  30,  18,   3,  10,  -6, -21,  21, -23,   3,  -8,  -6, -19,  -2 },
     {    0,   5,  -1,   0,   6,   8,   3,  -1,  -2, -16,   1,  -2,   3,  -1,  -5,  -1,  -1,  14, -23,  -7, -16,   2,   6,  -3, -22, -10,  31, -15,   4,   0,   8,   1,  38, -24, -23,   3, -12,   1,   2,   0, -62,  23,  19,   1, -10,   5,   0,   1,  60,  -8, -21,  10,   8,   9,  10,  -2,   1, -10,  24,   0,   1,   9,  -4, -13 },
     {    6,   6,  13,  28,  60, -10, -11,  -1,   3,   3,  10,  25,  37, -20, -17,   3,  -1, -11,   9,   8,   7, -24, -15,   2,  -2,  -2, -18,  22,   4, -14,   0,  -6,  -9,  -8,   5, -30,  15,   0,   2,  -8,  -7, -35,   2,   0, -16, -28,  -3,  -3,  23, -14, -25,   0,  -8,  -1,  13,   0,  26,  16,  -3,  14,   6,  -1,   9,  -4 },
     {   -1,   0,  -3,  17,   4,   4, -22,   7,  -1,   0,   5, -13, -36,   8,   4,  11,  -1,  -4, -10,  23,  40, -19, -14,   7,  -7,   3, -23, -41, -41, -22,  21, -11,   8,   6,  32,   0,   6,  -1,  -2, -11, -10, -10,  -8,  11, -18,   0,   9,   0,  11,  19,  -4,   8,  11,  14,  26, -15, -29,   4,   8, -11, -27, -12,  -6,   6 },
     {   -2,  -2,   4,  -4, -24, -20, -31, -10,   0,   4,   4,  22,  27, -22, -39, -35,  -4,  -2,  -1, -22,  20, -14,   2, -14,  -1, -12,  12, -18, -21, -10,   4,   7,   1,   0,   1,  10, -23,  38, -17,  42,   3,   3,  -1,  15, -17,  25,   8,  29, -17,  -4,  -4,   5,  -5, -18,  12,  -1,   1,  10,  -3,   8,   1,  -4,   7, -12 },
     {   -1,  -4, -17,  19, -25, -24,   8,  -1,   0,  -3,   4, -38,   4,   0,   1,  -7,   7,  -7,  -4,  35,  -3,   4,  15, -12,   5,  11, -24,   3,  11,  43, -17,  -1,  -2,   4,  -8, -18,   9,  -1, -43,  16,  -1,   3,  -7, -21, -21, -28,  18,  34,  24,  15,   3,   4,  -2, -13,  10,  -4,   7,  10,   4, -16,   0,  -6,  11, -23 },
   },
   {
     {   76, -47,   1,  -3,   1,  -1,   0,   0, -75,  36,   9,   0,   1,   0,   1,   0,  24,   8, -19,   2,  -2,   0,  -1,   0,  -2, -16,   8,   4,  -1,   1,   0,   0,   1,   7,   2,  -5,   1,  -1,   0,   0,  -2,  -2,  -3,   2,   1,   0,   0,   0,   1,   1,   1,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   34,  18, -19,  -3,  -3,   0,  -1,   0,  -3, -58,  30,   3,   5,   0,   2,   0, -39,  69,  -7, -14,   0,  -3,   0,  -1,  22, -22, -30,  20,  -2,   3,  -1,   1,  -2,  -7,  31,  -2,  -5,   1,  -1,   0,   1,   4, -10, -11,   5,   0,   0,   0,  -3,   1,   1,   7,   2,  -1,   0,  -1,   1,   1,  -2,  -1,  -3,   0,   1,   0 },
     {  -49,   9, -13,   6,   2,   2,   0,   1,  -3,  16,  26,  -7,  -2,  -3,   0,  -1,  14,  15, -53,   6,   2,   2,  -1,   1,  21, -52,  42,  16,  -9,   2,  -1,   0, -10,  24,   5, -39,  10,  -2,   2,  -1,  -2,   1, -23,  26,   7,  -4,   1,  -1,   3,  -1,  10,  -2, -15,   2,   0,   1,   1,  -3,  -1,  -4,   7,   3,  -2,   0 },
     {   52,  -2, -19,   7,   0,   0,  -1,   0,  32, -31,  15, -12,  -3,  -1,   0,   0, -31, -10,  -1,  15,   6,   1,   1,   0,   0,  10,  38, -26,  -7,  -2,   0,  -1, -23,  26, -49,  -4,  23,  -3,   3,  -1,   4,  -7,   5,  40, -22,  -1,  -1,   0,   0,  -4,  10, -25,  -4,   9,  -1,   1,   0,   0,  -2,   4,  12,  -5,  -2,   0 },
     {   39,  -3,  -2, -15,  -1,   1,   0,   0,  37,  -5, -18,  20,  -3,  -2,  -2,   0,  -2, -35,  13, -18,   2,   4,   2,   0, -16,   6,  10,  25,   1,  -4,  -1,   0,   4,  -4,  26, -46,  -8,   6,   0,   0, -19,  20, -35,  13,  39, -12,   3,  -2,   1,   0,   4,  26, -40,  -1,   2,   0,   1,  -5,   5, -19,  10,  16,  -5,   1 },
     {   -4, -65,  34,   7,   3,   1,   2,   0,  37,  38, -13, -22,   1,  -3,   0,  -1, -64,  18, -20,  20,   3,   1,   1,   0,  31,   4,  -1,   4, -13,   3,  -3,   1,   2, -25,   8,  -6,   5,   3,   0,   1,  -2,   9,   8,  -5,   5,  -5,   1,  -1,  -3,   2, -10,   2,   1,  -1,   2,   0,   0,   0,   2,   4,  -3,   1,  -1,   0 },
     {  -31, -11,  14,  -8,   5,   2,   2,   0, -40,   0,  12,  12,  -2,  -3,   0,   0, -11,  24,  17, -20,   6,   1,   2,   0,   8,  38, -30,   3,  -3,  -7,  -1,  -1,   5, -16, -24,   0,   9,   7,   2,   0,   1,  12,  -4,  49,  -6, -15,   0,  -2,  10, -14,  13, -17, -47,  21,   0,   3,   6,  -7,   9, -19,  37,   7, -12,   2 },
     {  -22, -15,   9,   5,   0,  -2,   0,   0, -31, -18,  20,  -8,  10,   1,   2,   0, -18,  10,  23,   3,  -2,  -4,   0,   0,   5,  31,  16, -29,   6,  -1,   1,  -1,   1,  22, -33, -20,   6,  -6,   1,  -1,  -4,  -5, -29,   2,  35,   3,   4,   0,  -1,  19, -10,  50,   3, -31,   1,  -3,  -2,   0,  -7,   5, -47,  11,  10,  -1 },
     {   -4, -13, -31,   8,   8,   3,   1,   0,  12,   1,  52,  -7, -13,  -4,  -1,  -1,   2,  -5, -36, -17,  20,   2,   2,   0, -40,  40,   1,  37, -12,  -7,  -1,   0,  31, -18, -15, -10, -18,  16,  -3,   1,   3, -26,  32,  -5,  21,  -2,  -3,   1,  -3,  16,  -7, -11,   3, -12,   5,  -1,  -3,   0,  -7,  13,  -3,   4,   1,  -2 },
     {  -22, -37,   9,   7,   0,   4,   0,   1, -29, -45,  17,  11,  -4,   3,   0,   1, -16, -25,  34,   9,   1,   5,  -1,   1,   3,   6,  43,   5, -14,  -2,  -1,   1,  12,  19,  35, -24, -16,   4,  -2,   1,   7,  12,   2, -38, -10,   1,  -1,   1,   0,  -5,  -8, -30,  15,  18,   0,   1,   3,   5,   1,   1,  30,  -6,  -5,   0 },
     {   10,  47, -17,   4,   1,  -1,  -1,   0, -30,  -2, -16,  -1,  -4,   2,   1,   0,  -2, -12,   7,  29,   2,   2,  -1,   0,  38,   2,  19, -16, -26,   2,  -4,   1,  13, -57,   3, -25,  29,  13,   0,   1,   0,   8,  35,   1,  23, -23,   2,  -1, -10,  14, -21,  -7,  -9,  -8,   8,  -1,  -1,   3,  -4,  14,   1,   8,  -1,  -3 },
     {    1,   0,  -4,  10, -19,   8,  -1,   1,   2,  -4,   8, -12,  24, -11,   2,  -1,  -2,   1,  -9,  17, -28,  12,  -2,   2,   7, -13,  14, -14,  29, -10,   1,  -1,  -9,  12, -12,  31, -37,  10,   0,   1,  15, -13,  30, -24,  22,  -5,  -4,   1, -12,  15, -20,  13, -43,   0,   8,   0,  12, -17,  21, -30,  26,  33, -20,   3 },
     {   -2,  -4, -58,  20,   4,   6,   0,   1,  -5,  34,  31,   2, -17,  -2,  -3,  -1, -19, -41,  21, -33,  12,   5,   3,   1,  54,   1, -14,   3,  17, -12,   0,  -3, -38,  -4,   2,  12, -14,  -3,   5,   0,   4,  20,  -6,  -5,   0,  11,  -4,  -1,   2, -12,  -5,   4,   7,  -6,  -1,   2,   0,   1,   6,   3,  -4,  -2,   2,   0 },
     {   -1,  39, -29,   7,  -6,  -2,   0,   1, -30,  32, -39,  21,  -6,   4,   0,  -1, -61,   7, -15,  30,  -8,   2,   1,   0, -22,  29,  12,  26, -13,   3,  -1,   0,   8,  27,   2,   2, -27,  -3,   0,   0,   2,  -6, -15,  -6, -17,  16,  -2,   1,   7,  -7,   7,   1,  -2,   6,  -6,   0,   1,  -2,  -2,  -6,  -7,  13,   3,   0 },
     {   15,  30,  11, -16,  -4,   3,  -1,   0,  17,  38,  12, -14,  -5,  -4,   0,   0,  12,  31,   7, -22, -14,   5,   0,   0,  17,  25,   0, -30,  -6,   0,  -2,   0,  10,   9, -22, -40, -12,   9,   0,   1,  -3, -18, -32, -36,  11,  10,  -1,   2,  -7, -16, -23, -12,  24,  22,   1,   2,   3,  -5,   1,   6,  39,  -7,  -8,   0 },
     {    0,  -3,  24, -39,  12,   2,   1,   0,  -1,   0, -27,  46, -12,  -1,   0,   0, -10,   4,  11, -26,   5,  10,  -2,   0,  28, -16,   6,  20,   0, -11,   0,   0, -25,  20, -17,  -9, -14,  19,  -1,   0,  37, -44,  27,   7,  24,   2,  -6,   2,  -9,  11,  15, -17,   5, -21,   5,  -1,  -3,   9, -18,  11,  -4,   0,   5,  -2 },
     {    2,   3,   1,  29,  -3, -11,  -1,  -1,  -1,   5,   3, -45,   7,   4,   3,   0,  13, -11,  11,  25,   1,  -7,  -2,  -2,  -5,  21, -35,   9, -23,   6,   1,   1, -35,  15,  28, -16,   9,   9,  -8,   1,  51, -26, -15,  -3,   7, -16,   0,   1, -31,  -1,  30,   3, -10,   2,   5,  -2,   2,  26, -19, -12,   8,   2,  -3,   0 },
     {    1,   5,  12,  -7, -14,  -2,  -2,   1,   5,   7,  11,   2,   0,   2,   2,  -6,   9,   5, -11,  -3, -17,  -1,   2,   5,   2,  23,   0,  -9,   2,   7,  -5,  -1, -27,   2,  13,   2, -11, -11,   1,   0, -13,  40,   0,  -2,  -7,  12,  -3,  -1,  13,  52,  -7, -35, -12, -17,  -5,   0,  -8,  -7, -78, -12,   5,  14,  11,   2 },
     {    8,  42,  57, -27, -17,  -4,   1,  -1, -20,   9,  40, -38,  -1,  -4,  -2,   1, -45, -47,  -1, -13,  15,  -8,   2,   1, -14, -31,  -7,   8,   6,   5,   2,  -1,  16,   1,  -3,   8,  -1,   3,  -8,   0,   9,   6,  -2,   4,   3,   6,   3,  -1,   6,  -3,   0,  -5,   0,  -7,  -3,   2,  -2,  -1,   0,   0,   0,  -1,   1,   6 },
     {   -4,   2,  -5,  -2,  14,  -5,   1,   0,  -2,  11, -11,  -4,  -5,   3,   3,  -1,  -6,  14,  12, -14,  18,  -1,   2,  -3, -35, -11, -10,  12,  -3,  -3,   8,  -2, -26,  34,   2,  -4,  21,   9,  -5,   1,  26,  32,   7,   2,  -7, -27,  -2,  -3, -16,  20, -67,   5,  -7,  15,  10,  -2,  15, -31,   5,  40,  -1,  -2,  -6,   1 },
     {   -6, -10, -29,   9,   5,  -1,  -4,  -1,   2,  22,  -2,  -3,  -4,   2,   3,   2, -23,   4,  37, -17, -11,   1,  -4,  -4, -12, -33,  -3, -18,  26,   1,   2,   4,  54,   2,   9,  -5,  14, -22,   3,  -5,  -6, -26,  -3,  14,  -6,  -1,   7,  -1, -47,  44,  17, -10,  -7,   8,  -8,  -2,  13,   2, -22,  -4,  10,  -4,   0,   2 },
     {   14,  21,  21,  -2,  -5,   9,   8,   0,  20,  30,  32,  13,  -8, -15,  -6,  -3,  20,  30,  32,  28,  25, -14,  -3,   6,  21,  26,  26,  15,  12, -13,  -5,   1,  24,  32,  28,  11,  -7, -16,   5,  -3,  10,  25,  29,  16,  -4, -11,  -6,  -2,  -9,   4,  14,  13,  -5,  -4,   1,   1,  -8,  13,  16,   2, -10, -10,   1,  -6 },
     {    6, -10, -14, -13,  13,  -6,  -5,   0,   8, -19, -18, -15,  -2,  -3,   2,  -2,  16, -19, -18,   6,   7,   1,  -6,  -1,  45,   5, -19,   2,  -2,  -5,   1,  -2,  67,  26,  -7,   7,  -4,  -2,  -8,   2,  46,   8, -23,  -1, -16,  -3,  -4,   1,  14, -11, -13,  -5, -14, -12,   1,  -3, -15, -33, -21,   2,  -8,  10,   1,   1 },
     {   -4,  -6, -12, -50,  -1,  15,   1,   6,   4,   3,  26,  27,  18, -14, -16,  -8,  -2, -13, -12,   7, -41,   8,   9,   4,  -3,  15,   0,   1,   0,  33, -17,  -7,  -9, -19,   7,  -8,  16, -36,  -9,   7,  41,  -3,  -2,   4, -19,   4,  25,  -5, -19,   2, -18,   7,  -3,  16, -12,  -5, -10,  13,  17,  -1, -16,  -3,  -5,  19 },
     {   -3,  -5,  -5,   2, -30,  30,  11,  -3,  -4,  -4,  -9, -19,  26, -19,  -5,   0,   3,   3,  18,  24,  -4,  23,  -6,   2, -14,  -5, -19,  14,  30,  -9, -22,   5,  -4, -11,   8, -34,  -6, -10,  36,  -4,   6,  -6,   0,  31,   2,  28, -26,  -3,   2, -17, -13, -24,  12, -31,   2,   6, -11, -19,   7,  27,   7,   6,  11, -10 },
     {    3,   8,  15,  14,  11,   1,  -9,  -4,   2,   7,  16,  17,  -3, -28,  22,   1, -10,  -6, -12, -22, -27,  51, -18,   0,  -7,   9,  -2, -20,   2, -12,   6,  16,   4,   9,  26,  -1,  27,   1,  -5, -22,  -7, -10,  -3,  -7, -26, -29, -19,  19,  11, -12,   5,  -8,  -4,  -7,  39, -13,  -5,  10,   5,   6, -21,  30,   3, -15 },
     {    1,  -6,  10,  26, -29,  21, -22,  11,   7,   3,  22,  44, -42,   3,   1,   2,   0,  -7,   5,  35,  -6, -12,  -5,  -7,  -2, -14, -26, -12, -14,  24,  24,  -6,   3,   3,   1,  12,   5,  -1,  -9,  -1,   0, -21, -19,   5,  16,  19,   2,  -3,  17,   9,  -9,   5, -19,  -6,  21,   4, -22, -10,   0,  34,  19, -11, -15,   4 },
     {    1,  -1,   8,  16, -16,   1,   4,   2,   0,  -3,  -6,  17, -25,  -5,  -5,   2,   0,  15, -14, -19,  21, -10,  10,  -1, -24,  -8,  33, -26,   4, -18,   3,   0,   4, -42,  -4,  16,  -7,   2,  19,  -4,  55,  23, -32,  -3,  -8,  15, -17,  -5, -31, -20,  -8,   4,  -7, -14,   8,  10,   2,  25, -11,  -8,   1,  12,   7,  -6 },
     {   -3,  -6, -22,  -7,  22, -26,  19,  -7,   7,  13,   6,  -7,   2,  21, -17,   2,  -9,  14,  16,   4, -32, -10, -11,   2, -23, -18,  -5, -22,   9,   4,   6,  -8,  11,  14,   2,  -2,  -3,   5,  -8,  -3,  13,  21,  17,  -1,  34,  -6,   9,  -9,  39, -35,  -4, -22, -16, -15,  -6,  10, -22,  47,  -4,  18,   3,  15,   5,  -1 },
     {    1,  -1,   4,  33, -50, -26,  17,  -4,   2,   4,  -3,  24,  54, -15,  -7,   5,  -2,  -8,  -6, -31, -15, -38,  20,   0,  11,   1,   2,   3,  -6,   1,   8,  -3,   7,   4,  -7,  -9, -10,  11,  -7,   1,  -2,  -1,   2,  -3, -12, -28,  11,  -1, -12,  -2,  11, -10,  -8,   7,   1,   4, -29, -10,   4,  32,  -5,  -5,  24,  -6 },
     {    6,   8,  18,  39,   0,  -4,  36, -21,   2,   0,   8,  30,  18,  16, -14,   9,  -2,  -7, -11,   1,  23,  30, -31,  10,   8,   6,   0,  14,  23,  13, -18,   3,   6,   4,  -2,  10,  29,   1,  19, -11,   3, -20,  -3,  -6,   9,  -1,  22, -17,   8, -16, -27,  12,  -2,  14, -33,  -5,  10,   7, -34,  -6,   8,  -7, -10,   6 },
     {    1,   1,   0,  16, -36,  -6,  11,  -5,  -1,  -1, -12, -22, -39,  -2,  11,  11,   2,  13,   9, -19, -40,  24,   0,   6,  -1,   9,  23,  25,  13, -13,  13, -12,   0, -12,   2,   3,  20,  -4, -19,  24,  26,  -4,  -6,   2,   2, -22,  -3,  -4,  36,  20,   6,   1,  11, -16, -31,  10,  -4, -12,  15,  -9,  -1, -30,  -8,  13 },
   },
   {
     {  -99,  58,  -5,   5,  -1,   1,   0,   0,  48, -26,   0,  -1,   0,   0,   0,   0,  -6,   1,   3,  -1,   1,   0,   0,   0,   8,   0,  -2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {  -56, -12,  29,  -4,   6,  -1,   2,   0, -54,  65, -22,   6,  -3,   1,  -1,   0,  54, -33,   1,   1,  -2,   1,  -1,   0, -10,   8,   2,  -3,   1,  -1,   1,   0,  15,  -5,  -4,   2,  -1,   0,   0,   0,  -1,  -1,   2,  -1,   0,   0,   0,   0,   5,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
     {   26,  -7,  10,  -9,   2,  -2,   0,  -1,  81,  15, -49,   7,  -8,   1,  -3,   0,  14, -59,  19,   3,   2,   2,   1,   0, -37,  18,   6,  -1,   3,  -1,   1,   0,  -3,  -6,   1,   4,  -1,   1,   0,   0, -12,   5,   5,  -2,   0,   0,   0,   0,  -2,  -1,  -1,   0,   0,   0,   0,   0,  -4,   1,   1,   0,   1,   0,   0,   0 },
     {  -27, -54,  50, -11,  10,  -2,   3,   0,  -6, -22,  -2,  12,  -1,   2,   0,   0, -59,  10,  42,  -9,   3,  -1,   1,   0,  -3,  54, -19,  -8,  -1,  -3,   0,  -1,  15,  -4,  -8,   2,  -3,   1,  -1,   0,   3,   8,  -4,  -4,   0,  -1,   1,   0,   5,  -1,  -4,   1,  -1,   1,   0,   0,   1,   2,   0,  -1,   1,  -1,   0,   0 },
     {   21,  62, -47,   6,  -9,   1,  -2,   0, -49, -18,  -5,  15,   0,   5,   0,   1,  -9, -36,  49, -10,   6,  -2,   2,   0, -17,  44, -11,  -7,   1,  -3,   1,  -1,  18, -11,  -6,   4,  -2,   1,  -1,   0,   0,   1,   2,  -2,   1,   0,   0,   0,   5,  -3,  -1,   1,   0,   0,   0,   0,   1,   0,   1,  -1,   0,  -1,   0,   0 },
     {  -20, -30,  -8,  18,  -1,   5,   0,   1, -22, -44,  19,   1,   6,   2,   2,   0, -12, -53,   3,  23,   0,   4,   0,   1, -32, -22,  63,  -2,   0,   0,   1,   0, -12,  41,  13, -15,   1,  -3,   1,  -1,   4,  14,   0,  -5,  -1,  -1,   0,   0,   1,  12,  -3,  -5,   0,  -1,   0,   0,   2,   3,  -3,  -1,  -1,   1,   0,   0 },
     {    2,  33,   3, -12,   2,  -3,   0,   0, -12,  59, -41,   0,  -1,  -2,   0,   0, -86,   1, -11,  19,   1,   5,   1,   1,   2, -19,  28,   0,   1,   0,   1,   0, -14,  18,  -2,  -2,   0,  -1,   1,   0,  13,  -5,  -2,   2,  -2,   1,  -1,   1,   5,  -4,   1,  -1,   1,   0,   0,   0,   4,  -1,  -1,   1,   0,   0,   0,   0 },
     {   -4,   1,   5, -22,  11,  -4,   2,   0, -18, -20, -29,  37,  -8,   6,  -1,   1,  -7,  -5,   5, -19,   8,  -1,   1,   0,  -3, -43, -26,  46,  -1,   4,   0,   1,  -7, -25,  72,  -7,  -9,   4,  -2,   0, -11,  29,  -6, -14,   9,  -6,   2,  -1,   1,   1,   5,   7,  -7,   2,  -1,   0,   0,   5,   0,  -6,   2,   1,   0,   0 },
     {    2, -16,  -7,   2,   4,   1,   0,   1,  -5, -25, -46,  28,   0,   4,   0,   0,  11, -23,  26,  12,  -6,   4,  -3,   1,  94, -20,  11, -23,   2,  -4,  -1,  -1, -10,  -1, -20,   7,   2,  -2,   2,   0,   7, -14,  10,   5,  -3,   3,   0,   1,  -9,  -1,   6,  -1,   1,  -1,   0,   0,  -2,   0,   3,   0,   0,  -1,   0,   0 },
     {   18,  32,  35, -28,   5,  -7,   0,  -1,   9,   8,  16,  18, -14,   0,  -2,   0,  25,  17,  31, -18,  -7,  -1,  -1,   0,  19,   7,  29, -10,  -8,  -3,   0,  -1,   3,  47,   6, -48,  -7,  -1,   0,  -1,   2,  12, -52, -18,  14,   0,   3,   1,  -3,  -6, -15,  11,   3,   1,   1,   0,   0,  -3,  -7,   2,   0,   1,   1,   0 },
     {  -11, -34, -54,  41,  -7,   9,  -2,   2,  15,  24, -11,  -9,  -3,   1,  -1,   0, -11,  -5, -11,   1,  15,  -6,   2,  -1,  28,  27,   6,  30, -25,   1,  -4,   0,  41,  -3,  17, -46,  -7,   2,  -3,   1,  -4,  -3, -31,  -4,  17,  -2,   3,   0,   1, -10,  -1,  11,  -3,   1,   0,   0,  -1,  -4,   2,   0,   1,   1,   1,   0 },
     {   -3,  -9,  -5,  44, -19,   7,  -3,   1,  15,  47,  43, -14, -12,  -2,  -2,   0, -11, -10,  59, -34,   1,  -4,   1,   0,  18, -26,  -6, -16,   6,   0,   2,   0, -32, -11,  20,  10,   1,   2,   1,   0,   1,  24,  15, -15,  -6,  -1,   0,   0,   3,   8,  -9, -12,   1,   0,   0,   0,   2,   2,  -4,  -4,   2,   0,  -1,   0 },
     {  -17, -24, -48,  17,   7,   2,   2,   1,  -2,   4, -40,  38,  -4,   3,   0,   0,  16,  55,  17, -14,  -9,  -4,  -3,   0, -39,   6,  -4, -17,   6,  -3,   2,   0, -50,  30, -11,   1,   6,  -1,   2,   0,   2,   5, -10,   9,   1,   0,  -1,   0,   2,  -3,   3,   6,  -2,   1,   0,   0,   3,  -2,   0,   2,  -1,   0,   0,  -1 },
     {   -5,  -4,   5,  -9,  -4,   4,   0,   1,  -8,  -5,   8, -15,  -3,   2,   0,   1, -14, -28,  -3, -28,   9,   7,   1,   0,  -7, -23, -15, -14,   5,   8,   0,   1, -31, -26, -32, -31,  32,  10,   2,   2, -24, -16, -30,  40,  46,  -4,   3,  -1,   3,   5,  15,  45,  -4, -11,  -3,  -2,   4,   4,   9,   6,  -9,  -5,   0,  -1 },
     {    1,   7,   9,   6,  -5,   1,  -2,   0,  -1,   3,   3,  18,   1,  -5,   0,   0,   4,  30,  18,  51, -30,   2,  -2,  -1, -11,   6,  34, -23, -37,  11,  -5,   2,  -7, -64,  12, -35,  11,  11,   1,   2, -24,  -1,  24,  -6,  17,   1,   2,  -1,   3,  11,   7,  -8,   3,  -5,  -1,   1,   0,   1,  -1,  -7,   1,  -2,   0,   1 },
     {   -2,   9,   1,  -6,  -4,  -3,   2,  -1, -12,   0, -12, -32,  16,  -1,   3,   1,   3,  -7, -27,  12,  14,  -3,   0,   0,  29,  54,  -8,   0,   1,  -9,  -2,  -2, -59,  14,  18,  -5,   2,  -7,   1,   0, -49,  41,  12,  -7,   6,  -5,   1,  -1, -16,  21, -16,  -2,   1,   0,   0,   0,  -2,  10,  -9,   4,  -4,   3,   1,   0 },
     {    1,   4,   4,  -2,  20, -12,   4,  -2,   1,   4,  15,  16, -29,  16,  -9,   3,   2,  -6,  -2, -13,  34, -16,   5,  -1,   5,  12,   5,  21, -30,   9,  -4,   0, -18,   6,  -8, -20,  21,  -6,  -1,   1,   2,  -8,  10,  39, -25,   2,   1,  -1,   7,  -5,  49, -50,   0,   7,  -3,   1, -15,  30, -29,  -5,  11,  -8,   5,  -2 },
     {   -4,  -4, -22, -16,  28,  -6,   4,   0,   4,   7,   8,  24, -29,   8,  -3,   0, -11,  10, -13, -29,  14,   0,   0,   2,   4, -14,  26, -18,  20,  -7,   5,  -2,  44,  -1, -19,   7,  -3,   1,  -3,   0, -66,  14,  22,   2,   1,   0,   3,   0,   8,  42, -20,  -9,  -1,   2,   0,  -2,   1, -14,  -8,  13,   0,   1,  -1,   0 },
     {    0,   1,   4,  -5,  -2,   2,   0,   0,   1,   3,   4,  11,  -9,   2,  -1,   0,   3,  -1,   3, -14,  17,  -4,  -1,   0,  -2,  14,   2,  31, -17,   0,   0,  -1, -21,   5,   0, -10, -15,  13,  -4,   2,  34, -55,  31, -12,  19,  -5,   3,  -1, -36,  66, -16,   4,  -1,  -3,   2,  -3,  32, -20,   0,   4,  -3,   2,  -2,   1 },
     {   -7, -12, -24, -52,   3,   5,   6,   2,  -4,  -7, -28, -71,  22,   8,   5,   2,   9,  20,  28, -35,  12,  -1,   0,   0,  -4,  -4,  37, -15,   1,  -4,   6,  -2,  11, -17,  24,  -6,   7,  -1,   1,  -1,  15, -12,   7,   3,  -6,   2,  -2,  -1,  -4,   0,   4, -11,  -1,  -2,   1,  -1,   0,  10, -10,  -6,   5,  -3,   2,   1 },
     {   -7, -13, -45, -46,  68, -19,   6,  -3,   9,  17,  38,  -7, -12,  -7,   3,  -2,  -2, -19,  13,  42, -24,  11,  -5,   0,   8,   4, -29,  -1,   2,   2,  -1,   2, -14,   1,   7,  -4,   5,  -3,   3,  -1,  19,   0,  -7,  -2,  -1,   0,  -1,  -2,  11,   6, -10,   7,  -2,   0,  -1,   0,   1,  -4,   4,  -2,   3,   0,  -1,   0 },
     {    5,   3,   1,  -1, -11,   6,   6,  -1,   9,   8,   1,   3,   1,   3,   4,  -1,   6,   7,  -2,   7, -12,   0,   0,   1,   9,  11,   3,  -5,  13,  -8,  -1,   0,  23,  15,   9,  10,   0,  -9,   1,   0,  15,  29,   3,  18,  11,   1,   0,  -2,  27,  46,  46,  22, -18,  -4,  -3,  -2,  51,  47,  15, -42, -14,   3,  -5,  -2 },
     {   -5,  -1,  -2,  -9,   8,   8,  -3,   2,   1,  11,  24,   5,  -2,  -1,   2,   0,  -3,   6,  20,   5,  -3,  -5,   3,   0,   1,   8,  23,  19,   7, -24,   7,  -3,  -8,  14,  18,  23, -30,   3,   1,  -1, -48, -41,  11,   5,   0,   9,   2,   0, -35, -32,  40,  27,  -4,   6,  -1,   0, -18,  15,  42,  -6,  -8,   4,  -3,   1 },
     {    1,   1,   3,   7,  11,   0,  -6,  -4,   3,   7,   6,   9,  33,  11,  -6,   0,   4,   7,  22,  33,  40, -20,  -7, -12,  -5, -11,   4,   2,   9,  -8,  -5,   3,   6, -23,   6,  17,   6, -21,   8,   4,  -8,   1, -48,  25, -24,   4,  -1,   4, -30,  11,  -9,  -3,  14,   3,  11,   5,  46,   9,  -6,  40, -12,  13,   1,  -4 },
     {   -2,   2,   2,   1,  -5,   5,  -3,   1,  -2,   4, -14, -20,  32, -12,   3,  -1,  -1,   7,  40,  20, -21,   0,   1,  -2,  -9, -34, -42,  27, -18,  10,  -5,   1,  29,  44, -26, -19,  20,  -1,  -7,  -3, -35,   2,  16,   1,  -1, -10,   1,  -1, -29,  22,   6,  -9,  -5,   1,  -2,   2, -11,   9, -16,  -9,   8,  -2,   3,   1 },
     {    2,   2,   4,   9,   8,  -8,  -2,   4,  -3,  -8,  -8, -24, -17,  12,   2,   4,   8,  11,  19,  20,  15,  12,  -9,   3,  -2, -15,   9,  39, -11, -36,  12,  -5, -14, -12, -44,   9, -43,  27,   9,   2,  -5,   6,  -5, -10,  14, -11,   6,  -1,  40,  -5, -21,   6,  -4,  13,   0,   2,   9,  38, -11,   9,  23,  -9,  -7,  -1 },
     {    1,   0,   2, -10,   2,  30,  -3,   2,   2,   1,   6,   3,  29,  -9,   2,  -1,   1,  -7,   6, -13, -25,  -3,   6,   3,   7,   5,  11,  31, -10, -17,  -6,   2, -17,  -7, -12,   4, -17,  -3,   3,   3, -18,   5, -11,  31,  -8,  18,  -4,   4,  30,  -8,   2, -23, -36,  -2,   3,   0,  28, -65, -13, -26,   7,  19,  11,   2 },
     {    4,   3,   1,  14,  38, -27,  11,  -2,   5,   4,   4,  17,  69, -35,   5,  -7,   0,  -5,  -6, -27,  25,   3, -11,  -5,   2,  -2,   3, -11,  -9,  13, -10,   9,  -3,   6,  -4,  -2,  -5,  11,  -6,  13,  -9, -16,  13, -35,  23,  -7,   8,   4,  28, -18,  13, -14,   6,  -3,   9,   2,  19,  16,   4,  -4,  -9,   3,  -8,   0 },
     {   -7, -11, -17, -46, -53,  18,  16,  -2,   7,  11,  18,  20,  24,  31, -11,  12,  -1,  -4,   0,  21,  28, -20, -15,  -3,   3,  -4,  -9,  -6,   3,  18,  -7,   0,  -6,   5, -14,   1,  -2,  22,   6,  -1,  -1,   2,  -4, -35,   0,   8,  16,   6,  22,   7,  17,  -4, -21,  28, -11,   1, -14,  -9,   8,  13,  -5, -17,   8,  -1 },
     {    3,   5,   6,   0,  12, -14,   9,  -5,  -1,  -3,  -3, -11,  -7,  16,   2,   3,   6,   8,  11,   1,  30, -13,   3,  -2,   2,  -5,  12,  15, -39,  16,   9,  -4,   1,   8, -14,   5,  35,  -2, -12,   4,   5,  32,  25,   9, -15, -18,   6,  -2,   5, -18, -20,  20,   3,  11,  -5,   3,  16, -30,  72, -14,  -3,   4,  -4,   0 },
     {    1,  -1,   3,  -1, -10,  18,  18,  -8,  -1,  -5,  -4,  -6,  -7,   4, -23,   9,   2,   1,   7,   8,  -9, -10,  26,  28,   3,   0, -11,   8,  26,  13, -17,  -6,  -7,  12,  16, -21,   6, -18, -19, -25, -23, -40,  13,   3, -13,   9,  20,  -1,  46, -13, -22,  -2,  48,   1,  -3,  -8,  34,  11,   7,   4,   6,   2,  -4,  -9 },
     {   -4,  -8,  -7, -21, -26, -23, -28,  -2,   2,   1,   6,   5, -12,   2,   7, -22,  -1,  -5,  -1,   8,   0, -48,  -8,   9,   1,  -5,  -6, -16, -22,  -8,  -7,  -9,   5,  16,  13,  11,  -3,  17, -17,   4,  -7,  -6,  21,  16,   5, -27,  -8,  -8,   7, -35, -16,  13, -19, -35,  -4,   4,  28,   4, -36,   9, -14,   3, -21,  25 },
   }
 },
 {  //7
   {
     { -115,  37,   6,   4,   2,   2,   1,   0,  38,   1,  -5,  -1,  -1,  -1,  -1,   0,   9,  -7,  -1,   0,   0,   0,   0,   0,   5,  -1,   1,   0,   0,   0,   0,   0,   4,  -1,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   37,  11, -10,   0,  -3,   0,  -1,   0, 103, -28, -16,  -2,  -4,  -1,  -1,   0, -17, -36,   7,   3,   2,   1,   1,   0, -27,  -1,  16,   0,   1,   0,   1,   0, -14,  10,   5,  -4,   1,  -1,   0,   0,  -8,   6,  -1,  -1,   1,   0,   0,   0,  -4,   2,  -1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 },
     {   13,  -1,  -3,  -2,   0,  -1,   0,   0,  16,  51, -16,  -3,  -4,  -1,  -2,   0,  87, -10, -35,   0,  -5,   0,  -2,   0,  -4, -53,  -4,  11,   2,   3,   1,   1, -18, -14,  20,   6,   0,   1,   0,   0,  -9,   2,  10,  -4,  -1,   0,   0,   0,  -4,   2,   1,  -3,   1,   0,   0,   0,  -3,   0,   0,   0,   1,   0,   0,   0 },
     {   27,  75, -19,  -9,  -5,  -3,  -2,  -1,  18,  36, -19,  -7,  -2,  -2,  -1,  -1, -27,  39, -16,  -3,   1,  -1,   0,   0,  25,  12, -37,   2,  -1,   0,  -1,   0,   8, -34, -20,  14,   1,   2,   0,   1,  -1, -24,   7,  12,  -1,   1,   0,   0,  -1,  -8,   9,   3,  -2,   0,   0,   0,   0,  -2,   4,   0,  -1,   0,   0,   0 },
     {    5,  50, -11,  -3,  -3,  -1,  -1,   0, -40,  39, -26,   2,  -2,   1,   0,   0, -30, -58,  -8,  20,   3,   5,   1,   1, -27, -13,  43,  13,  -1,   1,   0,   0,  20,  29,  15, -16,  -7,  -3,  -1,  -1,  12,   0, -13, -10,   1,   1,   0,   0,   4,  -6,  -4,   2,   3,   0,   0,   0,   2,  -2,   1,   1,   0,   0,   0,   0 },
     {   -3,  16, -12,   4,  -2,   1,  -1,   0, -14, -45,   0,   9,   2,   3,   1,   1, -32,   3,  -6,   8,   0,   1,   0,   1,  13, -45, -13,  24,  -1,   4,  -1,   1,  -4, -31,  58,  23,  -8,   2,  -1,   0,   8,  25,  43, -28, -14,  -1,  -2,  -1,  -2,  11,  -7, -28,   5,   3,   1,   0,  -1,   5,  -7,  -1,   9,  -1,   0,   0 },
     {   11,   2, -10,   7,  -3,   1,  -1,   1,   1, -39,  -9,  13,  -1,   3,  -1,   1,  32, -43,   9,  14,  -2,   3,  -1,   0,  74,  -5,  13,  -2,  -8,  -1,  -2,   0,  54, -18,  -9, -13,  -4,   0,  -1,   0,   3, -34,  -3,   1,   6,   3,   1,   1, -11, -11,  16,   6,   4,   0,   0,   0,  -6,   3,  10,  -2,  -1,  -1,   0,   0 },
     {  -14, -27, -14,   6,   7,   1,   2,   0, -16, -15, -29,   8,   8,   1,   2,   1, -19, -35, -24,  14,  10,   1,   2,   0, -28, -14, -18,  16,  11,   0,   2,   0, -14, -21, -25,  33,  12,   0,   1,   0,  -5, -17,  19,  53,   1,  -2,  -1,  -1,   9,   7,  41,  22, -22,  -6,  -3,  -1,   5,   3,  16, -10, -18,   0,   0,   0 },
     {   10,  68, -19,  -3,  -6,  -2,  -2,   0, -32, -57,  16,   7,   4,   3,   2,   1,  51,  -7,  19,  -9,   0,  -3,   0,  -1, -19,   9,  -3, -15,   5,  -3,   1,   0, -43,  12, -12,   3,   9,  -1,   2,   0,  -5,  14,  -7,  15,   1,  -3,   0,   0,   3,   4,   1,   6,  -7,  -2,  -1,   0,   3,   0,  -1,  -2,  -4,   2,   0,   0 },
     {    9,  17,  64, -32,  -2,  -2,  -1,   0,   2,   4,  59, -44,   2,  -1,  -1,   1, -12, -37,   7, -29,  12,   3,   1,   1,   2, -35, -12,  -6,  18,   2,   2,   1,  16, -12,  -6,  12,  12,  -3,   0,   0,   6,  -4,   2,  13,   1,  -5,   0,   0,   1,   3,   3,   4,  -5,  -4,   0,  -1,  -1,   3,   1,  -1,  -3,   0,   0,   0 },
     {    5,   3,   9,   5,  -6,   0,  -1,   0,   6,  16,  14,   5, -10,  -2,  -2,  -1,  15,  -6,  15,   8,  -8,  -1,  -2,  -1,  -3,  26,  16,   0, -13,  -2,  -1,  -1,  10,  12, -11,  13, -13,   3,  -1,   0,  20,   0,  39,  16, -34,   3,  -2,   0,  14,   2,  24, -58, -39,  14,  -2,   1,  -5, -13, -15, -50,  13,  28,   2,   2 },
     {   -4,   0,  -2,   2,  -3,   2,  -1,   1,  -7,  26,  -2,   2,  -8,   2,  -2,   0, -18,  -7,  40,   4,  -9,   3,  -1,   1,  16,  -8,  38, -35, -12,   3,  -2,   1, -37, -48,  -1, -35,  17,   9,   2,   1, -38,   3,  25,   7,  29,  -1,   1,  -1,  -1,  37,  10,  -3,  -1, -12,  -1,  -2,   0,   9, -13, -10,  -4,  -1,   2,   0 },
     {   -4,  -8, -55,  15,  -5,   4,  -1,   2,   6,  14,  19,  18, -10,   2,  -3,   0,  -2,  14,  47, -22, -16,  -1,  -2,   0, -32, -46,  -8, -33,  14,  12,   2,   2,  38,  -1,   0,   5,  26,   1,  -2,   0,  35,  -5,  -1,   8,   2, -13,  -2,   0,  -6, -19,   2,   3,  -5,  -4,   3,   1,  -3,   0,   8,   1,   0,   2,   1,  -1 },
     {    0,  -6,  12,  -4,   2,  -1,   0,   0,  -9,  12, -33,  -1,   1,   3,  -2,   1,  19, -42,  37,  -2,   1,   1,  -1,   0, -27,  47, -32, -20,   9,  -2,   1,   0,  12, -35,  -9,  32,  -1,   3,  -1,   2,   5,   2,  31, -30,  -6,   1,  -1,   0,  -5,  -3, -28,  -9,  26,   2,   3,   0,  -3,   3,   2,  25,  10, -12,  -1,  -1 },
     {    4,  -1,  -3,   8,   5,   1,   1,   0,   3,  12, -25,   8,   5,  -3,   1,   0,  -5, -15,  -6,  -4,   7,  -1,   1,  -1,  47,   0, -12, -16,  16,  -1,   0,  -1,  12,  25,  -8,  18,  32,  -9,   0,  -2,   2,  65,  -5,  28,  -2, -24,   2,  -3,  -2,  24, -41,   2, -28, -11,   4,  -1,  -5,  -9, -26,   9,  -8,  10,   5,   0 },
     {   11,  17,  49,  65,   5, -11,  -3,  -3,   7,  12,  33,  69,  -4, -15,  -3,  -3,  -5,  -7,   0,  31, -15, -10,  -1,   0,  -8,  -9, -22,   2, -14,  -1,   2,   1,  -2,  -2, -10,   2,  -6,   4,   3,   1,   2,   6,  -3,  -1,   3,   2,   1,   0,   1,   3,  -1,  15,   4,  -5,   0,  -1,   2,   2,   6,   9,  -7,  -6,  -1,  -1 },
     {    5,  -4,  15,   3,   1,   0,  -1,   0,   5,  -8, -15,  -3,  -1,   1,   1,   0,   2,   4,  -2,  -2,  -1,   4,   1,   0,   9,  -3,  25,  -5,  -8,   1,  -1,   0, -23, -37,   3,  -5,   4,   4,   1,   1,  59,  -1,  -6,   7,   7,   3,  -1,   0,  74, -24, -25,  12,  -3,  -1,  -3,   0,  10, -47,   5,   9,  -5,   1,   1,   1 },
     {   -4,  -4,   2,  16,   3,  10,  -5,   1, -10, -16,  -8,  -3,  -1,   8,  -5,  -1, -15, -21, -43, -10,  -8,  12,  -3,   2, -10, -14, -35, -27,  -3,  18,   0,   3, -15,  -3, -29, -38,  18,  25,   1,  -1,  -5,  -8,  -7, -17,  15,   8,  -2,   2, -23, -24, -14, -34,   2,   8,   3,   3, -28, -37, -15, -17,  19,  14,   4,   4 },
     {    4,  11,  27, -18,  -2,  -2,   2,   0,  -1, -29, -22,  11,   5,   1,   1,   0,  16,  35, -24,   3,  -9,  -2,   0,   1, -48,  -5,  22,  -2,  -4,   5,  -1,   2,  68, -23, -13, -12,   7,   2,  -3,  -1, -13,   2,   5,   3,  14,  -6,   2,   0,   5,  41,  -7,   0,  -6,  -7,   3,  -1,   4, -11, -25,   2,  -5,   5,   4,   1 },
     {    2,   3,  54, -22,   1,  -6,  -2,   0,  -3,  -9, -66,  18,   9,  -5,   3,   1,  -3,  23,  29, -10, -10,   1,   4,   0,   9, -29,   4, -17,  -1,   9,   0,   1, -22,  23,  -4,  10,   3,   2,  -1,   1,  33,  -3, -11,  -4,   2,   2,  -4,  -1, -27, -15,  26,  -3,   1,  -1,   0,   2,  -9,  27,   4, -14,   6,   3,   5,  -1 },
     {   -3,  -4, -16, -40,  15,   6,   3,   0,   4,   5,  12,  -7,  14,  -6,   0,  -1,   2,   3,  32,  60,   7, -14,  -5,  -2,  -3, -24, -12,  34, -15,  -6,   0,   3,  -5, -14, -48,  -6, -16,  10,   5,   2,  18,  29, -21,  -7,  -2,   7,   1,  -1,  -6,  13,  -4,  -3,   2,   1,  -1,   0, -41, -11,  -1,   3,   4,  -2,  -1,   1 },
     {    0,  -1,  -8,   5,   2,   0,   3,  -2,   2,  -2,   8,  -9,  -1,   0,   1,  -1,  -3,   4,  -8,  15,  -2,   4,  -1,  -1,  12, -19,  14, -10,   0,   2,  -1,  -1, -21,  37, -32,  21,  -1,   2,   0,   2,  33, -56,  30,  -2,  -3,   3,   1,   1, -19,  47, -20,  -9,  21,  -5,   3,   0,  29, -15, -27,  32,   2, -10,  -1,  -1 },
     {    3,   4,  -8,   9,  68, -16, -12,   5,   3,   5,   0,   8,  81, -19, -12,   6,   1,   0,   7, -18,  35, -13,   2,   5,  -2,  -5,   4, -17,   0,  -6,  10,   0,   4,  -4,   9, -16, -10,  -6,  10,  -2,  -6, -13,   5,  -1, -15,   2,   3,  -7,   2,  -6,   4,  -4,  -6,   9,  -6,  -3,   8,  -5, -17,   1,   4,   2,  -4,   1 },
     {   -3,   0,  14,   0, -13,   5,  -5,  -2,  -4,  -2,  -1,   6,   4,  12,  -4,  -5,  -5,  12,  22,  16,  -2,   5,   0,  -4,  -4, -13,  35,  12,   3,  -5,   0,  -3,  -8, -18,   0,  32,   6, -12,   2,   0, -47, -19, -11,  19, -29,  -7,   2,   1, -40, -46, -33,   4, -15,  16,   9,   1,   6, -42,  -7,  15,   2,  15,   0,  -1 },
     {   -5,  -6,  -9, -77, -14,   8,   6,   3,   7,   8,  24,  58,  24,  -4,  -3,  -6,  -6, -13, -26,  15, -21,  -8,  -4,   2,   8,  11,  -5, -35, -13,   4,   7,   0,  -5,   8,  16,  19,  14,   4,  -1,  -6,  -7,  -2,   2, -11,   7,  -7,  -2,   2,  10,  -7,  11,   4,   8,   5,   3,  -1,  16, -14,   3,  -1,   1,   6,   0,   1 },
     {   -1,   4,   6,  -3,  -5, -14,   3,  -4,  -1,  -2,   5,  -8,  -3, -16,  -7,   0,   3,  12,  -8,   7,  19, -10,   2,  -1,  -3,  12,  11, -17, -11,   1,   3,   4,   4,  21,  25,  -7, -10,  -8,  -3,   4,  10,   8,  58,  36,  26,  -4,  -4,  -1, -22, -17,   8,  21,  17, -13,  -4,  -1, -63, -31,  16,  19,   7,  -7,  -5,  -4 },
     {   -2,  -4,  -4,   0, -35,  12,  10,   5,  -2,   5, -11,   6,   2,  -4,   7, -11,  -3, -13,  19, -45,   5, -13,   5,   2,   0, -10, -24,   8, -33, -13,  10,  -3,   2,   2,   2,  -7, -45,   2,  11,  -9,   3,  10,   3,  -8,   6,  27,  -4,  -5, -22,  19,   8,  33, -16,   7, -26,  -5,  21, -41, -15,   3, -16,  21,   7,  13 },
     {    1,   2,   8,  -9,  12,  12,  -5,  -3,  -2,  -4, -16,  -6, -17,  -4,   3,  -1,   0,   5,  31,  20,   5,   2,   5,  -3,   3, -26, -36,  -8,  -1,  21, -13,  -5,  10,  44,   3,  -7, -10,  -1, -10,   1, -54, -19,  21,  -8,   4,   9,  -2,   1,  61, -10,  -2,   0, -20,  -8,  -7,   4,   2, -10,   1,  19,   8,   3,   4,   2 },
     {    1,   4,  -5,   2,  48,  -4,  -4, -10,  -3,  -2,  -3, -16,  -2,  -9,  -4,  -4,  -4,  -8,   4, -31, -69,  -2,   9,   6,   4,  -2,   8,  11, -43,   5,   9,   1,   5,   7,   2,  38,  -7,  12,  -4,  -4, -17,   5, -18,  22,  15,   1,   0,   3,   4,  14,  -7, -17,  21,  -3,  -9,  -4,  -9, -14,  26,  -2,   7,  -6, -11,  -2 },
     {   -2,  -2,  -3, -15,   8,   3, -11,   5,   0,   4,   5,  35, -20,   2,  -7,  10,   1,  -8,  12, -22,   6,  13,  -7,   8,  -2,  11,  -8,  33,   8,  12,  -5,   5,  -7,  -4,  24,  10,  -6,   2,   0,  -2,   5, -34, -20,  29,  16,  15,   1,  -7,  12,   4, -17,   7,   8,   1,   2,  -7, -43,  18, -71,  -5,   7,   2,  -1,   2 },
     {    2,   0,   6,   4,   8,  15,   1,   7,   0,  -3,  -2, -32,  -8,  20,  -7,  -3,   6,   0,   3,  40,   8, -23,   8,   1,  -9,   5, -22, -13, -55, -16,  19,  17,   7,   3,  33,  -4,  27,  25,  17,  -2,  12,  -6,  -6,  26,   8,  -2, -18,   2, -10, -21,  -7,  -1,   3,  -2,  -5,  -1,  28,  26, -27,   2, -12,  15,  -3,   6 },
     {   -2,  -3,  -7,  -3,   3, -51,  11,  15,  -3,  -4,  -2,  -6, -28, -44,  -5, -13,   0,  -1,   1,   8,   9,  -8,  40,  21,   3,  -1,   1,  -4,  13,  16,   9,  23,   1,  -3,   1,   6,  23,   0,  19,  22,  -5,  -1,  -7, -12, -10,  23,   3,  38,   3,  11,  -1,  15,   9,  26,   3, -20,   2,  -5,   1,  -7,  10,  37, -13,   7 },
   },
   {
     {  -73,  43,  -1,   3,   0,   1,   0,   0,  77, -36,  -7,  -1,  -1,   0,  -1,   0, -31,  -4,  17,  -1,   2,   0,   1,   0,   5,  17,  -9,  -3,   0,  -1,   0,   0,  -1,  -9,  -2,   5,   0,   1,   0,   0,   2,   3,   5,  -1,  -1,   0,   0,   0,  -2,  -1,  -2,  -2,   1,   0,   0,   0,   0,   1,   0,   1,   1,  -1,   0,   0 },
     {   26,  20, -17,  -2,  -3,   0,  -1,   0,  -7, -52,  28,   2,   5,   0,   2,   0, -33,  70, -14,  -9,  -1,  -2,   0,  -1,  30, -39, -19,  18,  -2,   4,  -1,   1,  -8,   1,  35, -10,  -2,  -1,   0,   0,   0,   8, -19,  -8,   6,  -1,   1,   0,  -3,   0,   3,  11,  -1,  -1,  -1,   0,   2,   0,   0,  -4,  -4,   3,   0,   0 },
     {   74, -13,   3,  -3,  -2,  -2,  -1,  -1,  17, -29, -15,   2,   0,   1,   0,   0, -38,  -6,  40,  -3,   3,  -1,   1,   0, -13,  49, -31, -11,   2,  -2,   0,   0,   6, -27,  -6,  26,  -4,   3,  -1,   1,   0,   3,  23, -15,  -5,   0,  -1,   0,  -4,   1, -10,  -4,   8,   0,   1,   0,  -1,   2,   2,   4,  -1,  -2,   0,   0 },
     {   38,  -1, -23,   5,  -1,   1,  -1,   0,  34, -21,  23,  -8,  -3,  -2,   0,   0, -22,  -9, -21,  10,   4,   3,   0,   1,   0,  -7,  49,  -8,  -5,  -2,   0,   0, -21,  36, -43, -15,  13,  -2,   2,  -1,   8, -16,   5,  42, -15,   2,  -2,   0,   1,  -2,  18, -35,  -6,   5,   0,   1,  -3,   1,  -9,   5,  19,  -7,   0,  -1 },
     {  -34,  -1,   2,   4,   4,   1,   1,   0, -44,   8,  18,  -5,  -1,  -1,   1,   0,  -7,  42,  -9,   1,   3,   0,   0,   0,  24,  13, -27,  -6,  -6,  -1,  -1,   0,   2, -26, -28,  30,   6,   4,   0,   0,  19, -14,  46,  14, -27,   3,  -3,   0,   2,  -5,   4, -48,   8,   6,   2,   1,  -2,   4,  -7,  13,  22, -12,   0,  -1 },
     {   16,   6,  -8,   3,  -1,   1,  -1,   0,  31,   7, -11,  -1,  -6,   1,  -2,   0,  22, -22, -18,  11,  -2,   3,  -1,   0,  -2, -47,  12,  13,  -5,   4,  -1,   1,  -4, -12,  48,  10,  -3,   2,   0,   0,   7,  11,  36, -44, -16,  -1,   0,   1, -12,  -2, -21, -41,  32,   2,   2,   1,  -4,   9,  -7,  23,  22, -13,  -1,  -2 },
     {   -5,   4,  -7,  19,  -6,   0,  -1,   1,  -8,  -8,  15, -29,  10,  -1,   3,  -1,  -4,   9, -10,  29,  -9,   0,  -2,   0,   5,  -7,  22, -36,   9,   0,   2,   0,  -7,  25, -26,  31,  -4,  -5,   0,   0,  11, -24,  20, -38,  -1,   7,  -1,   2, -14,  24, -30,  34,  29, -14,   1,  -3,   6,  -6,   8,  10, -47,  13,   1,   1 },
     {   31,  39, -17,  -4,  -3,  -3,  -1,   0,  42,  43, -24,  -9,  -1,  -4,  -1,  -1,  30,  18, -37,  -6,  -1,  -2,  -1,  -1,   2, -18, -41,   2,   5,   0,   1,   0, -12, -29, -25,  24,   5,   2,   1,   0,  -7, -20,   5,  27,   4,   2,  -1,  -1,  -1,   0,  14,  22,  -9,  -2,  -1,  -1,   0,  -4,   8,  -3, -16,   3,   0,   0 },
     {    7,  69, -24,  -7,  -5,  -1,  -2,   0, -34, -45,   1,  19,   1,   5,   1,   1,  55, -14,  28, -16,  -4,  -2,   0,  -1, -34,   2,  -8,  -7,  11,   0,   2,   0,   2,  24,  -5,  14,  -6,  -3,  -1,   0,   8, -16,  -3,  -3, -10,   4,   0,   1,   1,  -3,  13, -10,   6,   2,  -2,   0,   0,   2,  -5,  -3,   7,  -1,   0,   0 },
     {    5,  10,   7, -14,   4,  -1,   0,   0,   9,  17,  12, -16,   0,  -3,   0,  -1,  12,  19,   5, -27,   6,  -1,   1,  -1,   9,  18,  -9, -22,   3,  -3,   0,   0,   3,   3, -25, -35,  14,   1,   2,   1,  -9, -12, -42, -17,  21,  -2,   1,   1, -11, -17, -28,   7,  50,   4,  -2,   0,   1,  11,   1,  49,  38, -19,  -5,  -3 },
     {  -11, -60,  17,  -4,   3,   1,   2,   0,  38, -21,  49,  -7,   2,  -3,   1,  -1,  37,  18, -14, -32,   4,  -5,   0,   0, -48,   0, -28,  18,  15,   0,   3,   0,   0,  20,  -1,  20, -15,  -1,  -1,   0,   2, -13,   3,  -4,  -3,   6,  -3,   1,   0,   4,   8,  -1,   3,  -1,  -1,   0,  -2,   1,  -2,  -3,  -3,   0,   1,   0 },
     {    0, -18,  32,   0,   2,  -2,   0,   0,   7, -15, -17,  -7,   2,   2,   1,   1,  26,   7,  29,   8,  -3,  -4,  -1,  -1,  20, -28, -12, -43,   7,   1,   1,   0, -46,  15,  -1,  28,  35, -11,   5,  -2,  -2,  44, -18,  19, -35, -11,   1,   0,   1, -20, -10,   3,  -7,  18,  -3,   2,   5,  -1,  12,  -1,   3,   1,  -5,   1 },
     {   -2,  -1,   1,  -4,  15,  -3,   2,  -2,  -4,  -3,  -3,  14, -19,   2,  -1,   0,  -3,  -4,   2,  -7,  21,  -4,   2,   0,  -5,   2,  -2,  17, -22,   2,  -1,   1,   4,  -1,   7,  -8,  23,  -4,   0,  -1,  -6,   3,  -9,  21, -29,  -1,  -3,   0,   7,   2,  15,  -2,  35,  11,  -3,  -2, -13,  23, -24,  39, -82, -24,  20,  -3 },
     {    6, -14,  13,  -1,   1,   1,   0,   1,  10, -35,  25,  -2,   0,   1,  -1,   1,  47, -19,  22,  -8,   2,   1,  -3,   0,  55, -13,  13, -21,  -3,   0,  -3,   1,  16, -50,  -6, -27,  14,   6,  -1,   2, -16, -19,  25,   8,  27,  -2,  -2,   0,  -4,  24,  14,  -1,  -2, -10,  -1,  -1,  -6,   1,  -5, -11, -11,   2,   2,   0 },
     {  -17, -24, -21,  -1,   8,   3,   0,   1, -26, -39, -26,   0,  17,   3,   3,   1, -22, -35, -23,   2,  14,   6,   2,   1, -19, -27, -12,  10,  24,   2,   2,   1, -18, -19,  -4,  18,  19,   4,   1,   1,  -6,  -5,   9,  27,  31,  -3,  -3,  -2,   1,  10,  13,  33,  22,  -4,  -2,  -3,   7,  13,  17,  30,  27, -15,  -6,  -2 },
     {    0,  13, -31,  26,   6,  -3,   2,  -1,  -6, -12,  38, -29, -11,   5,   1,   0,  16,  -1, -23,  17,  19,  -7,   3,  -2, -34,  22,   4,  -2, -24,   7,  -1,   1,  25, -29,   4,  -5,  32,  -1,  -2,   0, -28,  41,   2,  10, -12, -18,   7,  -2,   2,   2, -34,   9, -15,  16,  -4,   2,   8, -12,  22,   0,   5,   4,  -8,   2 },
     {   -1,  -6, -53,   2,   5,   3,   2,   1,  -2,  15,  44,   9,  -9,  -2,  -2,   1,   4, -39,   2, -24,   8,   4,   1,   1,  37,  20, -11,   8,   4,  -8,  -1,  -1, -49,  -7,   0,   9, -13,  10,   0,  -1,  37,  25,  -5, -13,   5,   8,  -9,   2,  15, -30,  -3,   9,   3,  -9,   2,   1,   1,   3,  15,  -5, -13,   1,   3,  -2 },
     {   -7, -34, -33,  34,  -3,  -2,  -4,  -4,   6, -26, -41,  31, -10,   1,   6,  -3,  31,  21, -11,  25, -14,   1,   0,  -2,  19,  30,  -1,  16, -14,   5,   1,   3,  -1,  10,   1, -13,   1,   1,  -4,  -1,  -7, -31,  -2,  -3, -10,   3,   7,   3, -32, -42,  -8,   7,  -1,   2,   4,   0, -20,  -4,  22,   1,   1,  -2,   0,   4 },
     {   -3,   1, -10, -39, -25,  13,   4,  -8,   0,   6,  -2,  -9,   8,  23,   7,  -6,   3,   7, -14, -16,  -3,   9,   5,  -3,   4,  25,  14,   4,  11,  -7,   0,   3,   3,   6,   8,   4,  29,  -5,  -9,   1,  13,  15,   8,  -2, -13, -25,   2,   0, -53,  17,   6,   9,  -6, -11,  10,   1,  -6,  69,   1, -26,   0,  -1,  -9,   5 },
     {   -2,  -1,  -5,  16, -24,  10,   1,  -9,  -2,   7,  -7,  30, -11,   1,   4,  -9,   0,  -1,   4, -27,   5,  -4,   5,  -4,   3,   5, -19,  15,  16,   4,  -6,   4, -27,  18,  -1, -21,  28, -15, -14,   7,   7,   5,  22,  12,   6, -14,   4,   1, -14,  46, -42,  -2, -16, -14,   4,   2,  -1, -45, -35,  21,   2,   9,   7,  16 },
     {   -3, -13, -39,  13,  -5,   2,  -4,  -1,   9,  29,  20,  18,  -8,   8,  -6,  -7, -22, -13,  38,  -9,  -1,  -3,  -2,   7,  12, -37, -28, -17,  10,  10,  10,   2,  51,  32,   5,   5,  15, -17, -16,  -2, -20, -23,  12,   6, -10, -18,   2,   4,   0,  -8,   4,   1,   0,   7,  -5,  -4,  21,  12,  25,  -4,   1,  -7,  -5,   5 },
     {   -9,   1, -15,  12,   2,  -8,  -3,   2,  -5,  22,  16,   9,  -9,  -6, -10,  -2,  -9,  22,  30,   3, -20,  -3,  -1,  -6, -38,  -2,   5, -20, -21,   3,   5,   5, -62,  -8,   9, -13,  -2,  10,   7,   2, -55,   1,  27,  -3,  24,   5,  -1,  -1, -22,  -3,  23,  -3,  10,  -7,   9,  -7,   3,  17, -11, -10,   0,  -1,   2,   2 },
     {    6,  18,  43,   1, -21,   9, -14,   0,  -2,  -3,  16,  27,  -2,   1, -12, -11,  -6,  -4, -17,   6,  -1, -11,  16,   1, -12,  -1,  -9,  -9,   8,  -5,   2,   3, -15,   4,   8, -39,  -5,   6,  -5,  -4,   7, -16,  38,  11, -11,  -4, -12,  -1,  38, -13, -29,  20,  -6,   1,   0,  -2, -27,  32,  47,   1,  -2,  -6,   0,  -2 },
     {   -5,  -7, -17, -19,  -9,  -5,  12, -12,  -2,  -5,  -6,  35,  42,   1,  -7,   7,  -8,  -4, -24, -32, -19, -19,  14,  17,   8,  -1,  16,  -6,  22,   8,  -5, -10,  -1, -17, -26,   0, -10,  -4,   6, -14, -50,  12,  -2, -19, -27,  24,  28,  -7,  11,   3, -18, -14,   0,  20, -21,   6,   1,   2,   0,  -4, -12,   6,  -4,   3 },
     {   -6, -18, -29, -24,   0,  -8,   3,  -3,  11,   7,  -5,  29,  17,  14,  13,   5,   0,  23,  12,  10, -20, -10,   0,   5, -36, -17,   9, -34, -10,  16,  -6,  -6,   8, -26,  -1, -16,  27,  50,  -4,  -5,  54,  -4, -12,  -7,  -2, -12,  -4,   3,  19,   6,  -1,   0,  -6, -10,  -1,   1,  -3, -11,  -2,   6,  -4,   7,   5,  -6 },
     {   -4,  -6, -25,  -1,  16,   5,  -9,   6,   0,   5, -14, -22,  34,  -7, -14,   3,   2,   9,   1,  19,  36, -16, -24,   2, -10,  11, -13, -20,  47,   3, -21,   5,  -7,  -4,  24, -28,   4, -36,  -4,  11,   1,  -8,  -6,  10,   8,   5, -14,   9,  12,  -4, -10, -38,  15, -14,  -3,   7, -18,   5,  -2, -27, -16,  13,   2,  -4 },
     {    3,   3,  -4,  36,  18, -37,  12,  -4,  -1,  -1,  -9, -31,  36,  -4,   1,  -5,   3,  -8,  14,  -5, -47, -10,   8,  -1,   7,  -9, -25,  33,  -2,  -9,   2,  -5,  11,  24, -23, -13,  10,  24,   3,   2,   2,  27,   5,   4,  12, -12,  -1,   1,  16,  11,  -9, -13,   8,  -9,  16,  -8, -35,  25,  -3,  -6,  -4,   4,   1,   7 },
     {    0,   3,   4,  43,  -3, -20, -10,   4,  -7,   4,  -8, -25, -25,  -7,  37,  -5, -14,  -1,  -6, -48, -12,   2,  20,  -9,  -7,   3,  21, -18,  46, -13,   9,   2,  -1, -20,  28,  -8,   8,  18, -11,   0,  10, -18,  -3,   4, -17,   6,   2,   6, -10,  -6,  15,   0,   7,  33,  -5,   0,  -7,  -9,   1,  -8,  -7,  11, -14, -13 },
     {    2,  -3, -10,  29,   7,  35,   1,  -9,   7,  -1, -16,  24,  -8, -42,   5,  13,  12,  28,   1,  -3,   9,  32, -10,   7,  -6,  -2,  14,  -7,  12, -20,  -5, -18,   6,  -4, -25,  -4,  -9, -12, -16,  18,   9,  28,   9, -23,  17, -25,   0,   8,  33,  11,  10,   7,  -1,  30,   8,  -9,   8,  26, -10,  -5,  -2, -15, -11,   9 },
     {   -1,  -4, -13,  -3, -20,  53, -31,   9,   2,   1, -11, -22,  19,   3,  26, -14,   6,   3,  -9, -15, -21,   1,   8,  -4,  13,  23,   1, -13, -18,   9,   3,  18,   3,  19,  21,  13,  16,   8,   2,  -5, -30,   6,   0, -12,  -3,   8, -21,  16,  45,  10,  28,  -7,  -6, -22,   7,   2, -12, -12,   5,  23,   1,  -6,  -4,  -2 },
     {   -2,  -4,  -6,   2,   7,  12, -11, -12,  -3,  -3, -11, -38, -43,  34, -10,  28,   2,   8,  19,   9,  -7,  15,  25,  12,  -5, -10,  -5,  18,  23,  19, -23,  10, -13,  -6, -20, -18,  -1,   3, -27,   1,  -3,  -2,  -4, -20, -15,  20,  33,  -8,  20,   7,   6,  -1, -16, -23, -14,  18,   7,  15,   9,   2,   4, -12,  24,   9 },
     {    1,   3,   7,  -3, -22, -33,  21,  29,   3,   6,  -2,  -5,  22,  -9,   1,  12,   3,  10,   4,  -7,  12, -29,   8,  25,   1,  13,  19,  16, -10, -20,  -6, -10,  -8,   5,   7,   1,  31, -25, -25,   5,   6,  -3,  20, -20,   0, -20,  13,   1,  14,  -8,  36,  15,  19,  -3,   1,  22,   9, -28,  22, -10,  10, -21,  30,   8 },
   },
   {
     {  -95,  49,   3,   3,   1,   1,   0,   0,  64, -19, -12,  -1,  -2,   0,  -1,   0,   1, -16,   7,   1,   1,   0,   0,   0,  -4,   6,   3,  -2,   0,   0,   0,   0,   1,  -1,  -2,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -39,  -9,  15,   1,   3,   0,   1,   0, -44,  56, -11,  -1,  -2,   0,  -1,   0,  80, -33, -20,   5,  -3,   1,  -1,   0, -21, -19,  20,   2,   1,   1,   1,   0,   3,  12,   2,  -6,   0,  -1,   0,   0,   0,  -2,  -4,   0,   1,   0,   0,   0,   3,   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   -6,  77, -50,   5,  -9,   1,  -3,   1, -62, -20,  19,  10,   3,   3,   1,   1,  17,   8,  13, -10,   0,  -2,   0,  -1,  40, -20, -11,  -1,   0,   0,   0,   0,  -2, -12,   8,   2,   1,   0,   0,   0,   1,   3,   1,  -2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
     {   49,  30, -33,   1,  -6,   0,  -2,   0,  16, -19,  -8,   7,  -1,   2,  -1,   1,  24, -67,  33,   0,   2,   0,   1,   0, -51,  36,  17, -13,   3,  -3,   1,  -1,   5,  16, -19,  -1,   1,  -1,   0,   0,  -4,  -5,  -2,   5,   0,   0,   0,   0,   0,   1,   2,   1,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0 },
     {   -9,   3,   9,  -2,   1,  -1,   1,   0, -24,  55, -45,   7,  -3,   1,  -2,   0, -48, -37,  38,  12,   1,   2,   1,   1,  44,  28,   1, -19,  -3,  -3,  -1,  -1,  22, -29, -20,   5,   3,   2,   0,   0,  -7,  -5,  16,   4,   0,   0,   0,   0,   1,   3,  -1,  -4,  -1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0 },
     {  -31, -11,  13,  -2,   4,   0,   1,   0, -40, -18,  32,  -1,   3,   0,   1,   0,   6,  25,   7, -15,   2,  -3,   1,  -1, -42,  65, -36,   2,   2,  -1,   0,   0,  45, -21, -21,  18,  -6,   2,  -2,   1,   4, -17,  14,   0,  -3,   1,   0,   0,   5,   1,   2,  -4,   1,   1,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0 },
     {  -17, -15, -58,  55, -11,   8,  -3,   2,   3,  35,  39, -30,  -4,  -2,  -1,  -1, -38,  -2,  -9, -10,  14,  -1,   3,   0, -10, -10,  25,  11,  -2,  -1,   0,   0,  27,  25, -15, -16,  -3,  -2,   0,   0,   6, -24, -11,  11,   3,   2,   0,   1,   0,   0,  11,   0,  -1,  -1,   0,   0,   2,   0,  -1,  -1,   0,   0,   0,   0 },
     {    5, -29, -31,  36,  -7,   6,  -2,   1,  35,  22,   7,  -9,  -7,  -1,  -2,   0,  36, -14,  19, -34,   9,  -4,   1,  -1,  11,  -3, -45,   9,   9,  -1,   1,   0, -16, -45,  18,  31,  -6,   2,   0,   1, -17,  22,  29, -14,  -7,   0,  -1,   0,  -2,   7, -13, -10,   4,   2,   1,   0,  -4,  -1,  -1,   4,   3,  -2,   0,   0 },
     {   22,  28,  -9,  -9,  -1,  -2,  -1,   0,  20,  10, -24,   6,  -2,   1,  -1,   1,   2,  20, -36,  12,  -3,   1,  -1,   0, -27, -27,  15,  13,  -5,   4,  -1,   1,  74, -49,  16,  -8,  -3,   2,  -1,   0, -37,  -4,  14,  -7,   8,  -1,   2,   0,  -6,  15,  -1,   1,   0,  -2,   0,   0,  -1,  -2,  -2,  -1,   0,   0,   0,   0 },
     {   10,   6,   4,  -5,   3,  -2,   0,   0,  11,  -9,  39, -45,  13,  -6,   3,  -1,  19, -22, -47,  47,  -3,   1,   0,   0,  37,  26,   6,  -6, -22,   3,  -4,   1, -18, -22, -30,  -3,  13,   3,   1,   0, -13, -12,  30,  20,  -4,  -1,  -1,   0,  -3,  17,   4, -18,  -3,   0,   0,   0,  -2,  -2,  -8,   3,   4,  -1,   1,   0 },
     {    9,   7,  -7,  -2,  -1,  -2,   0,   0,  17,  12, -26,  18,  -4,   0,   0,   0,  19,  37,   2, -14, -11,   2,  -3,   0,   0, -28,  -4, -22,  17,   0,   1,   0,  -6,  12, -49,  35,   7,  -5,   1,   0,  22, -54,  29,  27, -21,   6,  -2,   0, -12,   4,  30, -24,   0,   3,   0,   0,   0,   6,  -9,  -5,   7,  -2,   0,   0 },
     {   14,  39, -11,  -2,  -1,  -2,  -1,   0,   4,  42, -20, -12,   0,  -2,   0,   0, -17,   3, -39,  18,   2,   0,   0,   1, -27,  27, -34,  28,  -7,   2,  -1,   0, -32,  -9,  26,   8, -11,   5,  -2,   1,  55, -16,   1, -13,  -3,   3,  -2,   0,   8, -31,   3,   1,   7,   0,   1,   0,  -2,   1,  10,   1,   1,  -4,   1,   0 },
     {   -7, -10, -36, -59,  56,  -9,   7,  -1,  15,  23,  41,  38, -29,  -6,  -2,  -2,  -8, -20, -10,   2, -23,  15,  -3,   3,  11,   1, -24,   3,  19,   1,  -2,   0,   5,  13,  10,  -2,  -7,  -7,   1,   0, -15, -13,  -5,   2,   3,   3,   1,   0,   3,   4,   7,   4,  -3,   1,  -2,   0,   0,   1,  -2,  -5,   0,   1,   0,   0 },
     {   13,  12,  33,  -9, -10,  -3,   0,   0,   9,   1,  -3, -24,  16,  -2,   0,  -1,  12, -19,  20, -19,  17,  -7,   3,  -2,  27,  -8, -42,  52, -13,  -2,   0,  -1,  17,  29,   1, -14, -26,   9,  -3,   1, -22, -45, -11,  -1,  19,   7,   0,   2, -12,  10,  29,  13,   0,  -7,  -2,   0,   2,  10,  -6, -12,  -5,   2,   1,   0 },
     {    7,   9,  -8,   2,  -4,   0,  -1,   0,  16,  24,  -9,   0,  -5,  -1,   0,   0,  21,  32,  -7, -17,   6,  -4,   0,   0,   7,  26, -16, -15,  -2,  -2,  -1,   0, -13, -21, -26, -21,  20,  -1,   3,   0, -24,  14, -49,  46,  16,  -8,   4,  -1,  20, -18,  13,  39, -32,   0,  -3,   0,  -1, -11,  16, -17, -10,  13,  -2,   0 },
     {   -3, -20,  -7,   1,   2,   4,   0,   1,   5, -10,   3,  11,  -4,   1,   1,   0,  14,  -5,  11,   1,  -1,   0,  -1,   0,   6,  10,  13, -13,  -6,   0,   0,   1,  -1, -46,  28, -24,   7,   5,   0,   1,  57, -22, -26,   5,  22,  -5,   0,  -2, -64,  30,  21,  10,  -7,  -9,   4,   0,   4,  18, -13,  -7,  -5,   3,   1,  -1 },
     {  -12, -11, -10,  16,   0,   4,  -2,   0, -21, -24, -30,  40, -25,   8,   0,   2, -15,  -9, -34, -23,  27,   5,   1,   1,  -8,  18,  16,  28,  -7, -11,   0,  -1, -45, -11,   4, -15,  -8,   8,   3,   0, -28, -24,  19,   8,  13,   5,  -1,   0,  28,  35,  15, -10,  -7,  -7,   0,  -1,   5,  -5, -17,  -6,   2,   2,   1,   0 },
     {  -19, -32, -45,   4,   3,   4,   1,   1, -12, -18, -35,  -8,  28,   4,   4,   0,   1,   8,  28,  60, -37,  -7,  -2,  -1, -21, -15, -34,   8, -22,  16,  -2,   4, -13,   2,   5, -10,  15,   4,   0,  -2, -22, -15,   3,   3,  11,  -6,  -2,   0,   9,  -3,  11,   4,  -1,  -2,   3,   0,   5,   8,   8,  -5,  -3,   1,  -1,  -1 },
     {   -3,  29,  15, -17,   1,  -6,   3,  -1,  -9,  37,  20, -26,  12, -10,   0,  -2, -27,  15,  24,  -6,  -1,  -1,   0,  -2, -48, -10,  -2,  -2,   5,   3,   1,   0, -41,  -2,   6,   6,  10,   1,  -1,   0, -28,  22,   2,  13,  -2,  -5,   2,   0, -26,  49,   6,  -9,  -8,   2,  -1,   0,  27,  10, -26,  -7,   2,   1,  -2,   1 },
     {    5,   7, -17,   6,  -3,  -2,   2,   1,  17,  26, -12,  12,   1,  -3,  -2,  -1,  27,  51,  10,   2, -15,  -4,  -2,  -1,  24,  57,  17,  -5, -13,   1,  -2,  -1,   2,  45,  11, -29,  -7,   2,  -3,   0, -12,   5,  20, -34,  11,  -2,   2,  -1, -11,  14, -21,  -8,   9,  -5,   2,  -1,   9,  11, -24,   6,   2,   0,  -1,   0 },
     {    1,   4,  12,  28, -26,   0,   0,   0,  -4,  -6,  -8,  -1,   1,   8,   1,  -1,  -7, -15, -33,   9,  -4,   2,   0,   2,   6,  11, -39, -55,  42,   4,   1,   3,  11,  23,  50,   2,  12, -23,   3,  -5, -21, -29, -18,  11, -17,   1,   4,   0,  -1,  19,  -8,   4,  -1,   8,  -2,   2,  11,   5, -10,   6,   0,  -2,  -4,   0 },
     {   -5,  -7, -20,   2,  10,   1,   1,   2,  -9, -12, -34,   4,  -4,  -2,  11,   3,  -5, -11, -38,  -2,  15,   8,   2,   0,   5,   5, -29,  15,  -5,  -3,  -1,   2,  11,  24, -37,  12,  -2,   3,   2,   0,  -4,  43, -15,   2,  -9,  -1,  -3,   0, -69,  -2,  -3,  -1,   1,   3,   6,   1,  38, -21,  -1,  15,   4,   2,  -3,  -2 },
     {    5,  -1,   8,  10,   2,  -7,   5,  -1,   5,  -2,   3,  10, -10,  -1,  -5,   2,   7,  -4,  11,   1,   1,   6,   0,  -2,   6,   6,  -4,  14, -27,   9,  -2,   1,  22,   3,  30, -16,  23,  -9,  -2,   1,   9,  28,  -8,  22, -27,  -1,   1,  -2,  19, -26,  46, -53, -11,  18,  -4,   3,  27,   0, -29, -25,  36, -12,  -2,  -2 },
     {   -6,  -6,  -6, -38, -44,  50,   8,   2,   5,  12,  17,  34,   4, -14, -17,  -5,  -4,  -3,  -5,   4,  41, -37,   3,  -1,  -4,  -5,  -8, -21, -34,  21,  21,   3,  -2,   8,   8,   7,   5,  18, -18,  -2, -10, -10,   1,   7,   4, -19,  -3,   1, -11, -13, -10, -18,   7,   3,  10,   2,  -1,  -5,  10,   1,   0,   3,   1,  -3 },
     {    4,   0,  -4, -13,  -1,   7,   2,   2,   6,   2,  -8,  -8,   5,   2,  -1,   0,   4,  -1,   9,  -4,  -7,  -4,   5,  -2,   2, -12, -22,   7,  -7,   4,   1,   2,  16,   8,  -6, -13,   8,  -2,  -2,  -1,  44,   0,  -9,  13,   9,  -3,   0,  -2,  36,  52, -35, -13, -14,  -5,   0,   0,   1, -73, -25,  17,   0,   8,   4,   0 },
     {   -5,  -8,  -8, -17, -40,  36,  -6,   1,   2,   3,   1,  15,  53, -45,   5,  -2,  -3,  -7, -11, -20, -38,  29,   8,  -3,   6,  14,  15,  18,  18,   6, -17,   2,  -4, -13,  -2,   1,  -3, -25,  16,  -3,  -4,   6, -12,  -7, -11,  32,  -3,  -1,   2,  -1,  18,  -7,  12, -13,  -6,   4,   4, -12,   1,  -4,  -8,   5,   8,  -3 },
     {   -4,  -2,  -4, -17,   4, -13,   2,   0,  -6,  -2,  -5, -23,  -3,  15,  -7,   2,  -6,  -1,   0, -28,  -7, -19,   9,   5,  -9,  -4,  -2, -13,   2,   4,  -9,   3, -17, -20,  -9, -30,  21,  -3,  12,   4, -25, -27, -20, -24,   9,  11,   0,   2, -20, -45,  -3, -18,  21,   1,  -2,   0, -23, -26, -39,  37,  30,  -6,  15,   1 },
     {   -5, -11, -33, -41,  32,   8,   3,   2,  -3, -10, -33, -49,  36,   1,  15,   3,   4,   3, -11, -28,  37, -11,   2,   0,   7,   6,  12, -24,  11,  -2,   5,   3,  10,   3,  15,   7, -18,   6,   7,  -5,  10,  10,  -8,   8,  -3,   2,  -8,  -4,  27,   7,   2, -16,   2,   1,   5,   3,  -2,  36, -10, -10,  -2,  -3,  -3,  -2 },
     {   -1,   4,   5,   3,  -2,  19,   4,   2,  -7,  -3,  -8,   4,   2, -13,   0,   0,  -9,  -8, -10, -16,  -7,  15,   9,  -1,  -8,  -8, -28,  11,  15,   9,  -7,  -1,   5,  14, -10, -48,  19,  -5,  -3,   4,  12,  22,  37,  23,   1, -16,   3,   0, -17,  -2, -22,  -6, -22,   4,   2,  -4, -66,  38,  -8, -10,  10,  12,  -2,   1 },
     {    3,   3,   3,   1,  10,  28,  36,  -2,  -7, -11, -10, -35, -35, -24, -30,  -3,   3,   5,   7,  -1,  10,  32,  -8,  -9,  -5,  -6,  -7, -26,   9,  13,  15,   4,  -5,  -3,  -1, -31, -14, -19,  -3,   9,   0,  -1,  29, -12,  18,  12,  -1, -12,  -4, -13,  12,   2,  15,   4,   7,  -1,  21, -30,   9, -33, -27, -10,   9,   6 },
     {    3,   7,   8,   8,  25,  14,   0,   5,  -1,   1,  -3,  -6,  -6, -20,  30,  15,   5,  13,  20,  -3,   3,  25,  -7,   6,  -1,   7,   7,  11, -11, -22,  10,   6,  -5,  -1,  23, -11,   2,   3,  20,  -4, -16, -27, -16,  30,  -9,   3, -12,  -6, -18,   4, -29, -44,   8,  17,   7,   7,  -8, -11,  60,  22,   7,  -1,  -6,  -6 },
     {    0,  -3,  -1,   3,   9, -34,  -4,   3,   2,  -3,   0,   9,  27, -45,  15,   6,  -2,  -9,  -4, -27,  17,  29, -15,  -3,  -1,  -7,   0,  -8, -28,  17,  -8,  -1,   1,  -1,   4,   3,  31, -11,  -1,  13,   7, -21,  20,  -2,  19, -41,  32,  -7,  10, -13, -16,  27,  16,  17, -13,   1,  29,   4,  -6,  13,   0,   2, -13,  -1 },
   }
 },
 {  //8
   {
     {  118, -38, -13,  -5,  -3,  -2,  -1,  -1, -19,  -5,   2,   2,   1,   1,   1,   0, -14,   3,   4,   1,   0,   0,   0,   0,  -7,   2,   1,   0,   0,   0,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   28,  54, -23,  -7,  -6,  -2,  -2,  -1,  95, -10, -33,  -6,  -5,  -1,  -2,   0, -10, -32,  -4,   6,   3,   2,   1,   1, -15,  -9,   8,   5,   2,   1,   0,   0,  -9,  -2,   5,   1,   0,   0,   0,   0,  -5,   0,   3,   0,   0,   0,   0,   0,  -3,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {   18,  41,   6, -18,  -3,  -4,  -1,  -2, -27,  91, -15, -23,  -3,  -4,  -1,  -2,  23,   4, -45,  -7,   3,   1,   1,   1,  -3, -27, -15,  14,   7,   2,   1,   1,   0, -12,   3,  11,   1,  -1,   0,   0,   0,  -7,   4,   4,  -2,   0,   0,   0,   1,  -3,   1,   1,  -1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 },
     {    6,  82, -13, -13,  -8,  -4,  -3,  -2, -59, -20,   1,   1,   7,   4,   3,   1, -61,  -8,  22,   9,   8,   2,   1,   0,   4,  21,   8,  -3,  -2,  -3,  -1,  -1,   9,   5,  -5,  -4,  -1,  -1,   0,   0,   7,   0,  -4,   0,   0,   0,   0,   0,   3,  -2,  -1,   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {  -13, -20, -54,  14,  13,   5,   3,   2, -34,  -6, -64,  28,  16,   5,   2,   2,  18, -34, -18,  41,   8,   0,  -1,   1,   5, -11,  26,  27,  -8,  -6,  -3,  -1,   6,   2,  22,   1, -14,  -3,  -1,   0,   1,   2,  11,  -6,  -7,   1,   0,   0,   1,   1,   4,  -4,  -2,   1,   0,   0,   0,   0,   2,  -2,  -1,   0,   0,   0 },
     {    7,  31, -69,   8,  -1,   3,  -1,   1,  -1, -20,   2,  16,   3,   1,   0,   0,  50,  61,  12, -14, -13,  -8,  -4,  -3,  27,   0, -31, -19,  -6,   1,   2,   1,  -7, -22, -10,   4,   8,   5,   2,   1,  -3,  -7,   4,   7,   3,   0,  -1,  -1,  -3,  -3,   4,   2,   0,  -1,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0 },
     {  -17, -23, -32,  -7,  18,   7,   3,   2,  10, -13,  -7, -23,  20,   1,   2,   0, -59,  41, -37, -20,  26,   1,   3,   0, -20,   2, -32,  29,  29,  -1,   0,   0,  -4,   6,   7,  35,   3, -13,  -5,  -2,   6,   3,   6,  12, -13,  -9,   0,   1,   2,   0,   4,   2,  -8,  -1,   2,   1,   2,   0,   1,   0,  -3,   0,   0,   0 },
     {  -11, -17, -30,  23,   4,   2,   2,   1,  18,  70, -18,  23,  -9,  -8,  -3,  -2, -44,  35,  32,   9, -17,  -7,  -1,  -1, -28,  25,  19, -22, -18,   0,   2,   0,  -5,   7,  -9, -22,  -2,   7,   3,   1,   2,  -3,  -8,  -5,   7,   6,   0,   0,   4,  -3,  -3,   0,   4,   1,  -1,   0,   1,  -2,  -1,   0,   1,   0,   0,   0 },
     {    0,  13, -22,  62, -14,  -1,  -3,   0,  -9,   0,  55,  35, -26,  -7,  -3,  -1, -13, -21, -10, -17, -17,   5,   4,   2, -33, -48,  -5,  10,  10,  15,   5,   2,  -8,   2,  28,   9,   2,   1,  -3,  -1,   3,   9,  10,  -8,  -6,  -2,  -2,   0,   2,   5,   2,  -8,  -1,   1,   0,   0,   1,   2,  -1,  -3,   1,   1,   0,   0 },
     {   14,  10,  39,  58,   1, -14,  -9,  -5,  13,  -4, -20,  39,  -5, -11,  -2,  -1, -21,  13, -51,  10,  -2,   0,   5,   1,  49,  29, -26,   5, -10,  -1,   0,  -1,  23,   0, -14,  -1, -12,  -1,  -1,   0,   0, -14,  -1,   0,  -5,   4,   1,   1,  -3,  -6,   4,  -1,  -1,   4,   0,   0,  -1,  -2,   3,  -1,   0,   1,  -1,   0 },
     {   -4, -10,   8,   8,  -5,   0,   0,   0, -15,   2, -52,  20, -10,   7,   0,   3,  -3, -42,  19, -29, -13,  11,   2,   3, -14,   8, -37, -46,  30,  17,   5,   2,  -2, -10, -23,  30,  42,   1,  -3,  -1,   1,   4,  13,  24,   3, -14,  -6,  -2,   1,  -1,   4,   5,  -7,  -7,   1,   2,   0,   1,   3,   1,  -3,  -1,   2,   0 },
     {    4,  16,  54,  -9,   9,  -9,  -2,  -2,  -5, -18, -24,  35,  21,  -3,  -1,   0,   5,  44,  12,  34,  -8, -17,  -6,  -3, -49, -20,  -7,   5, -14,  -6,   4,   2, -39, -22,  15,  18,  -2,   6,   5,   2,   0,  13,  22,   1,  -9,  -1,  -2,  -1,   4,   4,   3,  -8,  -6,   1,   0,   1,   2,   1,   0,  -4,  -1,   1,   1,   0 },
     {   -5, -20,   0, -19, -14,   5,   4,   2,  11,  12,   3,  -5, -15,   3,   2,   1, -43,  -4,  36,  23,  -8,   1,  -2,   0,  59, -49, -14,   9, -14,   3,  -2,   2,  12, -55,   8,  17,  -2,  10,   2,   2,   5,  -8,  27,   3,  -3,   4,  -3,  -1,  -1,   2,  10,  -9,  -3,   1,  -2,   1,   0,   2,   1,  -4,   1,   1,   0,   0 },
     {   -7,  -7,   5,  -2, -55,   5,   6,   4, -14, -28, -39, -28, -54,  16,  16,   7,  -8,  32, -32, -12, -23,  23,  11,   2, -10,  -2,  19,   2, -10,  19,   0,  -2, -15,  -6,  14, -19,   3,  19,  -2,  -1,  -1,   0,   1, -10,  19,  12,  -4,  -1,   1,   1,   0,  -1,  14,   1,  -6,   0,   1,   1,   0,   2,   6,  -2,  -3,   0 },
     {   -5,  -5, -10,   3,  33,   0,  -2,   0,  -7, -13,   8, -17,  25, -11,  -4,   0, -20, -23, -39,   1,  -5, -17,   3,   4, -12, -27, -13, -21, -39,  -1,  15,   9, -17, -18, -28, -33,   8,  37,  21,   6,  -3,  -1,  -2,  23,  40,  19,  -1,  -6,   3,   8,   9,  17,   6,  -8,  -9,  -5,   1,   1,   2,   4,  -3,  -4,  -2,   0 },
     {   -8, -10, -15, -53, -55,   2,  15,   7,   6,   2,  23,  46, -12,   2,   3,  -2,   0, -10, -26,  46,   5,  -6,  -6,  -6, -19,   1, -44,  -5,  -4,  -4,   3,   2,  16,  26, -21,  -3,  -1,   0,   4,   1,   4,   0, -12,   3,   3,   3,   1,   1,  -1,  -4,   3,   6,   1,   2,   0,   0,  -1,  -2,   2,   1,  -1,   1,  -1,  -1 },
     {    2,   2,   2,  31, -14,  10,  -6,   0,   4,   7,  15,  -7,   1,  26,  -1,  -2,   5,  14,  -4,  42,  48,  28, -14,  -9,   1,  -2,  19,   3,   5,  -2, -26,  -6,  -8, -10, -16, -14,  38,   5, -18,   2,   5,   6,   5,  28,  25, -37, -25,   2,   2,   1,  -1,   1, -18, -31,   3,  12,   0,  -1,  -1,  -2,  -8,  -3,  11,   6 },
     {   -3,  -3,  -2,  29, -22,  -7,  -2,   5,  -6,   1,  -3, -11, -20,   2,   7,   7,   6, -10,  18,   0,  39,   8,  -4,  -2, -17,   5, -12,  12,  20, -42, -14,  -4, -30, -31, -37,  -2, -34, -25,  27,  15,  -6,  -5,   1,  10,   0,  34,  33,   3,   6,   8,  10,   9,   8,  15,  -5, -15,   0,  -1,   1,   1,  -2,  -2,  -8,  -6 },
     {   -5,  -1,  -6, -14,  -7,   6,   3,   7,  -2,  15,  17,  14,   6,   3,   0,   1, -11, -17, -19,  -5,   7,   5,   1,  -1,  36,  38,  -2, -19,  -5,  -4,  -2,  -3, -78,   5,  33,   1,   2,  -3,   0,  -2, -50,  14,  26,  -5,  -1,  -2,   1,  -1, -11,  17,   6,  -9,  -1,  -2,   0,   2,   0,   6,  -4,  -3,   2,   1,   2,   1 },
     {   -4,  -5,  -9,   1,  -9, -76,  15,  10,  -2,  -1,   0, -11,  -4, -68,  35,  13,   6,   1,  13,  12,  18,  -5,  38,  -5,   6,   1,   4,   9,   6,  14,  12, -16,  -9,   5,  -3,   0,  11,  10,  -5, -10,  -2,   7,   1,  10,   1,  -7,  -9,  -5,  -3,  -2,  -2,   1, -11,  -5,   0,   1,   0,   1,  -1,   0,  -3,   5,   4,   1 },
     {    1,   1,  12,  -4, -11,  -7,   3,   4,   2,   2, -24,  30,  25,   0,   4,   6, -13,  18,   7, -27,  27,   3,  -6,  -4,  31, -61,   3, -20,  28,  -9,  -5,   2,  -7,  28,  16, -35,   6, -17,   5,   3,  -1,  21, -32, -12,  20,  -1,  15,   1,  -1,   4,  -6,  21,  12,   0,   7,  -3,   1,   4,   1,  11,  -3,  -4,  -1,  -3 },
     {    0,  -3,   3,  36, -70,  28, -10,   4,   0,   4, -10, -35,  47,  -7,  -7,   0,   3,  -6,  11,  -3,  18, -48,   5,   0,   1,   2,  -9,  -9, -21,  -5,  30,   6,   4,  12,   9,  12,   3,   9,   9,  -8,   2,   7,   4,  -6, -14,  -4,  -7,  -6,  -4,  -4,  -8,  -6,  -3,   3,   3,   1,  -1,   1,   0,   1,   1,   5,   2,   1 },
     {    1,  -1,   3,  -9,   7,  -5,   5,  -3,  -1,   4,  -9,  23, -24,  14,  -6,   0,   6, -11,  21, -44,  45, -16,  -1,  -5, -12,  24, -36,  42, -28,  -8,  12,  -2,  16, -21,  35, -25,  -8,  32,  -8,   4,  -1,   9,  -4,  -7,  27, -17,  -8,   1,   1,  -3,  -1,   4,  -7,  -7,   3,   1,   0,   2,  -2,  -1,   2,   0,   2,   2 },
     {   -1,  -3,  -1, -18, -20,  -7,  16,   6,   3,  -5,  10,  29,  16, -17,  -9,  -5,  -2,   7, -28, -30,  28,  -7,   3,   6, -16,   9,  56, -35,  10,  -8,  -1,  14,  34, -46,  15,  -1,   6,  -5,   4,  11,  -4, -35,  23,  11,  -5,   0,  10,  -5,  -1,  -2,  11,  -8,  -5,   0,   1,  -5,  -1,   0,   0,  -5,  -1,   2,   1,  -1 },
     {    4,   5,   6,   2,   5,  16,  10,   3,   8,  10,   9,  19,  26,  30,  13,  -2,   7,   7,  10,   6,  22,  30,  31,   2,   4,   8,   6,   8,  26,  28,  32, -16,   6,   0,   1,  12,  -4,  27,  14, -26,   3,   2,   4,  -2,   7,  38, -14, -34,   4,   1,  -1,   2,  14,   0, -43, -24,  -2,  -3,  -2,  -2,  -2, -15, -22,   0 },
     {    2,   2,  -4,  -1,   7, -27,  10,  -7,   4,   3,  -2,  -8, -23,  15,   5,   0,   2,   6,  -9, -10,   8,   5, -22,   8,  11,  -2,   5, -27, -14, -38,  -3,  14,  19,  33,   5,  12,  -7,  12,  38,   4,  29,  53,  33,  13, -10,   8,  -6, -24,  16,  16,  -9, -27, -18,  -7, -13,  -2,   0,  -3, -10,  -7,   1,   2,   4,   5 },
     {    1,   0,  -4,  10,  10, -21,  21,  12,   1,  -1,   1, -20,  19,   7,  -7,   4,   1,  -3,  -5,  18,  24,  11, -19,   0, -11,   4, -31, -31,  18,  30, -12,   5,   9, -16,  -3, -31,  -5,  15, -22,  14,  11,  12,  25, -53, -14,   6,  -2,  15,  -3,   5,  -7, -29,  30,  20,   7,   7,   0,   4,  -1,   4,  26,   3,  -4,  -2 },
     {    1,  -2,  -3,  -7, -10,   2,  25,  12,   3,   2,   3,  18,   9,  -7, -18,  -3,  -1, -11, -15, -15, -11, -12, -10,   8,  18,  15,  21,  23,   6,   4,   5,  -6, -40, -35, -27,  14,  11,   3,   1,   0,  62,  24, -38, -18,   0,  -1,  -7,   8,  26,   6, -23, -12,   0,  -5,  -2,   3,  -6,  -8,  -9,   4,   6,   9,   6,  -3 },
     {    4,   6,   5,  15,   4,  14,  73, -14,   3,   2,   2,  -1,  14,  20,  71, -31,  -1,  -7,  -1, -12, -10, -13,   6, -37,  -2,  -3,  -2,  -5, -21, -10, -24,  -7,   0,   1,   0,   7,   1,  -6, -15,   7,   0,  -5,  -6,   7,  -6,   0,   1,   7,   1,   0,   1,   3,  -1,   5,   1,  -1,   2,   1,   1,   1,   2,   3,  -3,   1 },
     {    4,   6,   4,   5,  22,  39,  28,  -8,  -4,  -5,  -1,  -7, -43, -25,  11, -10,  -1,   4,  -3,  28,   4, -50, -16,  12,   7,   6,   2,   2,  56, -12,  23,  10,   0,  -5,   6, -25,  22,   7,  20, -13,  -5,   2,   8, -13,   1, -11,   1, -15, -11,  -3,  -3,   2,  -2,  -3,   4,  -8,  -2,  -1,   5,   4,  -8,  -5,  -8,  -7 },
     {    0,   0,   2,  -3, -13,  20,   8, -18,  -2,  -3,  -6,   1,  16, -39, -10, -15,   1,  -3,   7, -11, -18,   5,  -3,  -6,  -3,   2, -10,   3,  34, -10, -26,  -8,   2,   4,  -1,  -9, -46,   6,  -9,  13,   0,   2,  22,  -5,  26,  28, -36,   3,  -4,  -8,  -3,  -1,  -7, -44, -21,  32,   4,  -3,  -3,  -9,  -5,  -3,  23,  24 },
     {    1,   2,  -1,   4, -14,  -8,  18,  -9,   2,   3,   2,  -4,  22,   5, -11,  -6,   1,   5,   4,   7, -30,  -2,  13,  22,  -1,  14,  -3,  34,   7,  24,  -5,   5,  29, -12,  -3, -14,  23, -20,   8,  44, -38,  43,  12,   8,  16,  -2,  42,  -2,  12,  22, -17,   1,  -4, -14,  -8, -18,  -1,  -3,  -9,  -3,  -6,  -6,   5,   0 },
   },
   {
     {   74, -41,   1,  -3,   0,  -1,   0,   0, -78,  36,   5,   1,   1,   0,   0,   0,  34,   1, -13,   1,  -2,   0,  -1,   0,  -7, -15,   8,   1,   1,   0,   0,   0,   1,   9,   1,  -3,   0,  -1,   0,   0,  -2,  -3,  -3,   1,   0,   0,   0,   0,   1,   1,   1,   1,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 },
     {   45,  17, -17,  -1,  -4,   0,  -1,   0,  -6, -55,  24,   0,   5,   0,   2,   0, -43,  67,  -8,  -6,  -1,  -1,   0,   0,  35, -31, -19,  10,  -2,   2,  -1,   1, -13,  -2,  24,  -4,   0,   0,   0,   0,   3,   7, -11,  -4,   1,   0,   0,   0,  -4,  -1,   2,   4,   0,   0,   0,   0,   2,   1,   0,  -2,   0,   0,   0,   0 },
     {  -78,  16,  -2,   3,   1,   1,   1,   0, -38,  20,  16,  -1,   2,   0,   1,   0,  41,  31, -33,   0,  -4,   0,  -2,   0,  24, -52,   6,  10,  -1,   2,   0,   1,  -6,  11,  19,  -9,   1,  -2,   0,  -1,   5,   1, -12,  -1,   1,   0,   0,   0,   1,  -1,   3,   2,   0,   0,   0,   0,   2,  -1,  -1,  -1,   0,   0,   0,   0 },
     {   29,  11, -18,   1,  -2,   0,  -1,   0,  54,  -5,   7,  -4,  -2,  -2,  -1,   0,  14, -24, -29,   5,  -1,   2,  -1,   1, -19, -42,  50,   5,   1,   1,   1,   0, -21,  52,  -5, -27,   4,  -4,   1,  -1,  -2, -11, -28,  20,   3,   2,   1,   1,   0,  -1,  17,   3,  -7,  -1,  -1,   0,   0,  -2,  -2,  -7,   3,   2,   0,   0 },
     {  -22, -12,  -7,   7,   2,   2,   1,   0, -36, -21,  34,  -7,   2,  -1,   2,   0, -34,  16,  -3,   9,   3,   1,   1,   0,   9,  30,  32, -18,  -3,  -3,   0,  -1,   0,  33, -62,  -3,   7,  -2,   1,   0,  18, -33,  13,  32, -11,   4,  -3,   1,   3,   7,  14, -25,  -2,   1,   0,   0,   0,  -1, -10,   4,   4,  -1,   1,   0 },
     {  -19, -14,  18,  -8,   2,  -1,   1,   0, -27, -15,  -6,  16,   2,   2,   1,   0, -18,  -3,  32, -13,   1,  -1,   1,   0, -11,  26,  -3,   9,   2,   0,   0,   0,  15,  23,  15, -30,  -5,  -2,   0,  -1, -13,  13, -63,  15,  18,  -2,   3,   0,  17, -16,  28,  37, -25,   0,  -5,   1,   3,  -4,   9, -38,   3,  10,  -1,   1 },
     {  -12, -19,   6,   4,   2,   1,   1,   1, -15, -24,  16,  -6,   2,   1,   1,   0, -16, -16,  18,   7,   2,   1,   1,   0, -10,   3,  35, -11,   0,  -1,   0,  -1,   1,  36,  22, -15,  -1,  -2,  -1,  -1,   2,  27, -16, -49,   3,  -2,   3,  -1, -16,   7, -56,  10,  32,  -5,   5,   0,   3,   6,   0,  45,  -8, -11,  -1,  -2 },
     {   -9, -70,  24,   4,   5,   1,   2,   0,  30,  52,  -9, -12,  -3,  -3,  -2,  -1, -55,   8, -20,  15,   1,   3,   1,   1,  46, -12,  11,   2,  -6,   1,  -2,   0, -14, -10,   6, -11,   5,  -1,   1,   0,  -2,  13,  -5,   6,   1,  -1,   0,   0,   2,  -5,  -2,   2,  -5,   1,   0,   0,   0,  -1,   4,  -2,   0,   1,   0,   0 },
     {    8,  24,  -5,  -5,   2,  -2,   0,  -1,  11,  34,  -6, -14,   2,  -4,   0,  -1,   7,  32, -16, -10,   4,  -2,   1,  -1,   5,  22, -27, -14,   4,  -3,   1,  -1,  -3,   3, -45,  -4,  10,  -2,   3,   0, -11, -20, -44,  12,  13,  -2,   4,   1, -11, -12, -19,  49,  19,  -7,   1,  -1,   6,   4,  18,  44, -17, -11,  -2,  -1 },
     {  -13, -73,   9,   2,   3,   3,   2,   0,  35, -26,  32,  -1,   2,  -1,   0,   0,  42,  40,   1, -21,  -2,  -5,   0,  -1, -44,   3, -31,  -3,   9,   0,   2,  -1, -15,  13,  -4,  20,   2,  -1,   0,  -1,   3,   4,   1,   4,  -9,  -1,   0,   1,   0,  -3,   0,  -2,  -1,   2,   0,   0,  -1,   0,   0,   2,   1,   0,   0,   0 },
     {   -1,  -1,  -1,  19,  -6,   0,  -1,   0,   0,  -5,   3, -22,   6,   0,   1,   0,  -2,   2,  -4,  28, -10,  -1,  -2,   0,   0,  -7,   9, -26,   6,   1,   1,   0,  -3,  10,  -9,  36,  -9,  -1,  -1,  -1,   5, -16,  16, -31,   0,   4,   2,   1,  -2,  19,  -6,  49,   1, -13,  -3,  -2,  15, -26,  31, -56, -40,  32,  -5,   4 },
     {    4,   8,  10,   4,  -1,   0,  -1,   0,  -8, -17, -16,  -3,   0,   2,   1,   0,  30, -12,  36,   7,  -1,  -1,  -1,  -1,  38,  -5,   2, -42,  -1,  -5,   0,  -1, -66,   1, -18,  11,  27,   0,   5,  -1,   0,  48,  -6,  27, -14,  -6,  -1,  -2,   5, -15, -13,  -2, -17,   4,   0,   2,   3,  -6,  11,   0,   7,   3,  -1,   1 },
     {  -18, -28, -33,  -2,   8,   2,   1,   1, -27, -40, -38,   3,   8,   5,   2,   0, -26, -38, -29,   7,  14,   6,   2,   1, -23, -29, -14,  20,  13,   6,   2,   1, -16, -17,  -3,  24,  17,   2,   3,   0,  -6,  -1,   7,  30,   7,  -2,   0,  -1,   0,   3,  12,  28,   5,  -5,  -1,  -1,   5,  11,  16,  27, -14,  -9,  -2,  -2 },
     {    3, -33,  -3,   7,   0,   4,   0,   1,  13, -42,  -2,  12,   0,   4,  -1,   1,  44, -17,  16,   5,  -4,   2,  -1,   1,  63,  -5,  13, -12,  -9,  -2,  -4,  -1,  32, -28,  -8, -24,  -3,   1,   0,   1, -15, -42,  -5,  -7,  19,   5,   3,   2, -14,   3,  17,  12,  16,  -3,   1,  -1,  -5,  10,   3,   7,  -3,  -5,   1,  -1 },
     {   -1,   0, -26,   6,   7,   2,   1,   1,   2,  -5,  43,  -7,  -5,  -7,   0,  -2,  11,   0, -22,  -4,  11,   1,   1,   0,  -8,  22,  20,  14, -19,  -6,  -2,  -1,   7, -41,  -7, -33,   9,  10,   4,   2, -27,  36,  17,  22,  28, -16,  -1,  -3,   3,  28, -34,  12, -39,  -6,   4,   0,   5, -20,   8,  -2,  -9,  13,  -2,   2 },
     {    5,   7,  69,   7,   3,  -7,  -4,   1,  -2, -30, -51, -25,   6,   3,   3,   2,  13,  37, -22,  35,   6,  -5,  -4,  -1, -28,   2,  21,   5, -27,   0,  -2,   1,  13, -10,  -3, -22,  15,  13,   1,   3,  13,   3,  10,   8,   9, -13,  -4,  -2,  -5,  -5,   1,  -2,  -3,  -4,   6,   0,   3,   6,  -3,   3,  -3,   1,   1,  -3 },
     {    8,  16,  76,  -8,   7,  -8,  -3,   8,  -3, -13,  48, -27,  -2,  -4,  -4,   5,  -8, -42, -18, -35,   4,   2,  14,   1,  14, -21, -14,   1,  25,   3,  12,   0,  14,  -3,  -6,  22,  15,  -5,   4,  -1, -13,  -1,  -3,  11,   0,  -8,  -2,  -1,  -3,  10,   2,   5,   0,  -5,   0,   1,  -1,   4,  -4,   1,  -4,  -1,   0,   1 },
     {   -4,  -6, -16, -26, -38, -19, -11,  -6,  -1,   5,  -8, -38, -19, -28,  -5, -11,  -3,   2,  18, -14, -13,  -4,  -4,   0,  -6, -13, -12, -35,   2,  13,  -7,   7,  13,   5,   9, -17,  22,  20,  18,   9,  21,  15,   9,  14,  31,  32,   7,  -3,   1,  29,  13,   1,   2, -14,  -9, -17,  -6,  20, -27,  -2, -21,  -2,  -2,  -1 },
     {    2,   2,   1,  12,   8,   2,   5,   5,  -4,   0,  26,  -8, -13,   0,   0,  -3,   7, -26, -13,   5,   4,  -3,   6,  -4,   4,  30, -13,  23,  -8,   2,   0,   4, -48, -18,  13, -16, -10,  14,  -2,   2,  80, -12, -10, -12,  20,  -1,  -4,  -5, -25, -21,  12,  11,   0, -11,   6,  -4,  -3,  17,  -3,  -5,  -7,   8,   3,   3 },
     {    2,   2, -21,  40, -11,  -5,   2,  -2,   0,  -1,  29, -43,   6,   3,   3,  -4,   4,  -9,  -8,  34,   3, -11,   5,   1, -12,   8,   0, -15,  -9,  20,   1,  -4,  29, -29,  14,   3,  17, -25,  -2,   1,  -3,  21, -28,   7, -32,  12,   5,  -3,  22, -36,  16,  -7,  23,   5, -17,   4,  -7,  11,   8,  -6,   0, -18,   9,   3 },
     {    4,  -7, -22, -13,   2,  -2,  -2,  -3,   6, -12, -27, -32,  13, -15,   3,  -5,  22,  -5, -30,  -3,  20,  -4,  11,  -5,  43,  36, -11,  20,  16,  -2,  -5,  -3,  39,  47,  15,  23,   0,  -2,  -4,  -3,  20,  20,  -1,   3, -18, -14,  -3,  -6,  -5,  -2, -11,  -4, -27,  -2,   5,  -1, -13,  -1,  -2,   4,   3,  17,  -2,   3 },
     {   -7,  -7, -12, -22, -11,   4,  15,   3, -12, -16, -16, -45, -13,   8,  12,  16,  -9, -13, -31, -28, -28,  -4,   4,   9,  -9,   3, -14, -25, -11,   0,   5,  10, -14,  -9,  -6, -29, -14,   6,   4,   6, -21, -13,  -1, -20, -13,   0,  16,  -2, -16, -19,  -9,  -7,   9,  20,  19,   9,  -7, -21,   7, -12,  43,  28,   6,  -2 },
     {   -1,  -2,  -6,  48,  -2,  -2, -19,   2,  -6,   2, -10, -42, -42,   0,   1,  -2,   0,   0,  27, -19,  63,   0,   1,  -2,   2, -18, -11,  17, -23, -19,  10, -14,  -9,  14,  -8,   4, -26,  27,  -8,  -6, -20,  -9,  -4,  -7,   8,  -3,  -5,  12,  11,   4,   2, -10,   5,   8,   8,  -3,   4,   4,  -1,  -4,   2,  -2,  -8,  -1 },
     {   -1,   4,   8,   1,  16,  12,   0,   2,  -5,  -3,  -2, -15,  23,  -6,   0,  -9,   6,  -6, -19,  24,  -1, -12,  -6,  -6,   7,  36, -24,  -2,  -5,   6,  -4, -12, -47,  13,  20, -15, -12, -19, -14,  -8, -45,  -5,  -1,  -6,  -4,   8,  13,   6,  -3,  54,  23,  -1,  10,   2, -11,  -2,   1,   7, -37,  -4,  -9, -18,   0,   4 },
     {    0,  -2,  -1, -18,  17, -31,  16,   9,   1,   0,  -5,  17, -29,  34, -20,   6,   1,  -3,   1, -17,  28, -36,  11,  -6,   6,   3,  -7,  13, -46,  33, -14,   8,  -1,  12,   5,  -5,  37, -18,   7,   4,   5, -12,   1,   5, -25,  15,  10,  -6,  -3,  13, -15,  -3,  16,  -1, -11,   2,   2, -24,   3, -12, -23,   8,   4,  -7 },
     {    2,   0,  12,  -1, -20,  -2,   3,   2,   1, -10, -10,   1,  15,  19,   8,   2,   9,   3, -22, -26,   6,  15,   4,   5,  13,  15,  15,  32,  10,  -9,  -9,  -2, -11,   1, -24,  -5, -30,   3,  15,  11,   2,  30,  14,  -5,  10,  59,   1, -12,  29, -10,  -5,   2,  32, -12, -39,  -8,   4, -15,  10, -11,   3, -20,  -1,   8 },
     {    0,  -1,  13,  13, -18, -23,  -7,   0,  -2,  -6,   2,  21, -14, -45,   4,  -2,   0,  -4,   0,  20, -18, -29,  20,  -2,   0,  -5,   2,  23, -12, -17,   9,   2,   0,  -7, -11,  10, -34, -13,  27,  12,   1,   5, -11,  11, -26,  29,  41,   0, -13,   9,  -7,  10, -16,  41,  10, -15,  -3,   5,   0,  17,  -6,  15, -15,  -8 },
     {   -2,  -1,  -6, -58,  13,  21,  -4,  10,   0,   5,  13, -45,  21,  15, -16,   6,   3,   2,  34,  19,  23,   4, -22,   3,  -6, -13,  10,  22,  -3, -46, -17,  19,  -1, -16,   2,   3, -12, -20,  27,   9,   4,  -9,  -5,   9, -17,   3,   5,   3, -20,  -6,   1,   2,  -3,   4,  -5,   1,   3,  -3,  -2,  -5, -11,  -9,  -4,  -1 },
     {    0,   0,   3, -39,   4,   3,   5,  -1,   3,  -4,  23,  11,  -9,   6,   3,   7,   1, -16,  -7,  61,   0,   7,  19,  -8,   3, -11, -42, -12, -37,  -2,  23, -12,  12,  23,  -3,   2, -19,  41,   2, -11,  -5,   3,   3,   8,  11,  20, -26,   1,  15, -11, -12,   3,  -2,  -2, -18,  12,   0,  -8,   3,   6,  14,  -3,  -3,  -2 },
     {    2,   2,   1, -18,   4,  14,  -8,  -1,   2,   1,   6,  -3,  -3,   6,  -2, -15,   4,   4,  -5,   0,  11,  27,   3,  -5,   1,  11,  10,  -1,   5,  27,  -1,   3,   1,  -4,   3,  -1,  22,  15,  -6, -12,  12,   2,   2, -10,   1,  -8,   3,   5,   2,  13,   9,   8,  14,  74, -12, -31,  12,   5,  24,  -8,  15, -10, -61, -15 },
     {    3,   4,  -3,   8,  60,  31,  -2, -13,   0,  -5,  -6,   0, -38,  40, -35,  -3,  -2,   6, -16,  -5, -22, -15, -15,  -7,   0,  -1,  11, -22,   8,  -6,   9,  -2,  14,  -2,  -5,  15, -20,  10,   8,   9,   1,  21,  -3,   1,   0,  14,  13,  -5,  12,  -8,   2,  14,  -7,   0,   1, -27, -22,  17, -30,   4,  -6,   9,  -6,   2 },
     {   -1,   1,   6,  13, -60,  16,   8,  21,  -1,   3,   8,   3, -36,  49,  14,  13,   1,   3,   0,  24, -10,  17,  -4,   0,   3,  14,  -3,  15,  15,  -3, -14,  -7,  -2,  13,  21,  18,  13,  -6,  -5,  11, -19, -12,   7,  27,  15,  -5,  22,  -4, -23, -17, -15,   7,  -6,  -4,   5, -11,  -5,  -6, -27,   1,  -1,  -3, -17, -12 },
   },
   {
     {  -89,  61,  -3,   2,  -1,   1,   0,   0,  57, -21, -19,   5,  -2,   1,  -1,   0,  -1, -19,  16,   2,  -2,   1,   0,   0,   2,   4,   2,  -7,   1,   0,   0,   0,  -1,  -2,  -1,   1,   2,  -1,   0,   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   37,  21, -33,   0,  -4,   1,  -2,   0,  25, -79,  40,   8,   1,   1,   0,   1, -32,  37,  21, -30,   5,  -4,   2,  -2,  -1,   9, -28,   6,   9,  -3,   2,  -1,  -4,   1,   4,   8,  -5,  -1,   0,   0,   1,   1,  -3,  -2,  -1,   1,   0,   0,  -2,   1,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {   49,   0,   7, -14,  -5,  -3,  -1,  -1,  12, -11, -49,  26,   1,   5,  -1,   2,   9, -49,  63,   4, -17,   4,  -4,   2, -13,  22,   6, -43,  18,   0,   2,  -1,  -6,   3, -14,  15,  11,  -9,   2,  -1,  -3,   5,   2,   0,  -8,   0,   2,  -1,  -1,  -1,  -2,   1,   1,   2,  -1,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {   46,  24, -29,  14,  -8,  -1,  -4,   0,  38, -10, -17, -20,   1,   0,   2,   0, -24, -13, -30,  51,   2,   1,   0,   0,  11, -22,  48,   0, -35,   9,  -5,   3,  -6,   1,   5, -33,  17,   9,  -3,   1,  -2,   0,   1,   6,   8,  -9,  -1,   0,  -3,   1,   0,  -3,  -1,  -1,   3,   0,   0,   0,   0,   1,   1,   0,   0,   0 },
     {    4,  72, -49,  -3,  -5,   0,  -1,   0, -76,   1,   3,  26,   1,   4,   0,   1,  34,   0,   3,   1, -15,   2,  -3,   1,  11, -24,   6,  -7,   4,   6,  -2,   1,   3,   0,   9,  -2,   5,  -5,  -1,   0,   2,  -2,  -1,  -2,  -1,  -1,   2,   0,   1,   0,   2,   0,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   28,  26, -12, -20,  10,  -2,   1,  -2,  19,  28, -32,  21, -19,  -3,  -3,   1,   2, -26,  -9, -23,  23,   5,   3,   0, -18,  12, -23,  60,  -6, -18,   3,  -4,   9, -16,  24,  -4, -44,  23,  -3,   3,  -2,  -4,   5, -16,  21,   7, -10,   3,   1,  -3,   4,   1,   0,  -8,   2,   2,  -1,   0,   0,  -2,   2,   1,   2,  -2 },
     {   21,  24,  -9,  -4, -11,  -1,  -1,   0,  30,  29,  -1, -26,  19,  -9,  -1,  -3,   1,  -1, -39,  -4,  -8,   2,   4,   2,  -4, -24,  -3,   3,  49,  -3,  -1,  -1, -11,  18, -10,  47, -17, -42,  11,  -5,  -2,  -7,   2,   0, -39,  24,  12,  -1,  -1,   0,   3,  -1,   7,  11,  -8,  -3,   0,  -2,   1,   0,  -3,  -1,  -3,   2 },
     {   -4, -24, -41,  31,   4,   6,  -1,   2,  16,  -4,  47, -10, -24,   2,  -5,   1,  52, -52,   2, -31,  22,   7,  -1,   2, -20,  -1,  30,   1,  13, -17,   2,  -1,  -7,  22, -12, -16,  -1,  -2,   7,  -2,  -6,   2,  -8,  10,   7,   0,   0,  -3,  -1,   4,  -1,   0,  -4,  -2,   0,   0,  -1,   0,  -1,   1,   0,   1,   0,   0 },
     {   12,  15,   9, -12,  11, -15,   0,  -2,  15,  20,   2,  -7, -19,  17,  -2,   1,  13,   0,   0, -28,  24, -23,   0,   0,  -9,  -4, -26,   6, -12,  25,   5,   1,  10, -18,  19, -13,  51, -25, -18,   6, -12,  14, -18,  21, -20, -35,  40,  -6,  -1,  -1,  -1,   5, -14,  33,  -3, -13,  -1,   0,  -1,   0,   2,  -3,  -9,   4 },
     {    5,  -3,  50, -51,   3,  -1,   3,  -1,  10, -52,   1,  17,  22,  -8,   1,  -1,  55,   1, -23,   4, -14,   0,   3,   1, -14, -33,  18,  16,   3,   5,  -2,   0,  -8,  19,  23, -17,  -7,  -6,   1,   0,  -7,   7, -14,  -7,  10,   2,   3,  -1,  -2,   5,  -1,   6,  -1,  -3,   0,  -2,  -1,   1,  -2,  -1,  -1,   1,   0,   0 },
     {   -3,  17,  16,   5,  -2, -11,  -1,  -1, -38, -36,   0, -41,  26,  11,   6,   2,   4, -29,  -4,  33,  41, -27,   4,  -3, -30,  40,   0,  23, -23, -10,  12,  -6,   5,  12, -13,  23, -12,   2,  -7,  -2,   3,   0,   6,   2, -27,   3,   4,   4,   3,  -2,   0,  -5,   1,  14,  -3,  -1,   1,   0,   2,   0,   0,  -2,  -4,   1 },
     {   25,  46,  36,  -7, -16,  -2,  -6,  -3,  15,  23,  28, -41,  -5,  -6,   1,   0,  35,  15,  -2, -12,  -5,  -1,  -4,  -3, -11,   6, -24, -34, -14,  -3,   9,   2, -10,  -7, -33,   3,  12,  39,   2,  -3,   3,  -9,   7,   6,  25,  -4, -27,   5,  -2,  -3,  -4,   0,   3, -19,   8,   7,  -1,  -1,   0,   2,   0,   1,   4,  -3 },
     {    3,  32,  40, -31,   4,   1,  -1,  -2, -25,   9,  26,  -9, -10,  -8,  -2,   1, -60, -24,   6, -21,  18,  20,   4,   3,  16,   9,  33,   2,  24, -20, -12,  -1,   1,  15, -12, -31,  -3, -18,  14,   1,  -1,   5, -23,  14,  14,   1,   8,  -6,   1,   3,  -3,   9, -10,  -4,   0,  -1,   1,   0,  -3,   0,  -1,   4,   2,  -1 },
     {   -2,  -3,  -2,   0,   5,   8,  -7,   0,  -1,  -2,  -3,  -5,  11, -16,   7,   6,  -5,  -3,  -3,   2,   6,  12, -10,   0,   0,  -9,   4,   0,  22, -20,   8,   6,   4,   9,   7,  21,   4,   3, -20,  -7,   7, -15,  13, -21,   5, -64,   6,  30,  -4,   6,  -6,   6, -13,   2,  71, -23,  -1,   4,  -2,   6,  -7,  20, -15, -26 },
     {    7,   1,  28,  33, -13,  -4, -11,   0,  -6, -15,  -8, -33, -13,  18,   2,   4,  11, -15,  24,  -4,  28,   3, -13,   1,  57, -40, -11,  -2,   6, -23,   3,   6, -13, -26,  39,   8,  -8,  -3,  16,  -3,  -4,  22,   8, -28,  -2,   6,   0,  -7,  -9,   5, -12,   1,  13,  -2,  -4,   2,  -1,   4,  -1,   4,  -3,  -5,   2,   1 },
     {   -2,  -3,  -3, -19,  33,  -9,  -4,  -3,   0,  -5, -24,  21, -29,   2,  17,   0,   3,  25,  -4,  -1,  27,   4, -19,  -2,  18, -25,   6,  -3, -10, -39,  26,   4, -24,  25, -25,  13,   4,  26,  17, -23,   2,   5,  -7,  37, -27,  -1, -20,   2,   3,  -9,  12, -14, -15,  23,  -3,   6,   1,  -1,   1,  -3,  10,  -1,  -9,   1 },
     {  -10, -12, -15, -19,   1,   9,   4,   2, -17, -23, -25, -26,   1,  18,   4,   0, -19, -27, -27, -16,  13,  27,   8,   4, -13, -19, -16,   3,  32,  30,  -1,  -1,  -2,  -1,  10,  25,  41,  18, -11,  -4,   8,  13,  16,  25,  19, -12, -33,  -9,   2,  -1,   4,   2,  -5, -29, -14,  15,   2,   1,   2,   0,  -2,  -7,  12,   6 },
     {   -4, -10, -31, -55,  50,  -1,   3,   3,   9,  16,  39, -20,  28, -24,   3,  -3,  -2, -36,  25,  19, -13, -17,   9,  -3,  42,  -7, -21,   0, -17,   9,   9,  -5, -14,  -4,   4,   9,  -2,   9,  -4,  -1,  -3,   9,   3,  -6,  -3,   2,  -3,   0,  -3,   0,  -5,   1,   2,   0,  -2,   2,  -1,   2,   1,   2,  -1,  -1,   2,   1 },
     {    4,   6,   8,  -3,  47, -11,  -7,  -7,  -2, -20,  15,  -3, -50, -15,  20,   3,  13,  -4, -25,  27, -13,  52,  -7,  -9,  -7,  28,  -6,  -9,   3,  -1, -35,  14,  14, -32,   1,  10,   5, -14,  13,  10,  -1,  -2,  24, -11, -12,   2,   6,  -8,  -1,   0,  -5, -12,   9,   9,   0,   0,  -1,   2,   2,   3,   4,  -4,  -5,   1 },
     {   -3,  -9, -35, -31,   4,  19,   4,   3,   3,  -4, -37, -35,  29,  12,  -3,   3,  31,  37,   4,   2,  28,   6,  -9,  -3,  15,  26,   1,  -6,  10, -31, -25,   4,   5, -21, -18, -18,  -1, -21,  23,  15,  -5,  -5,  -6,  14,  24,   0,  18, -10,   0,   3,  -2,  18,  -8, -11,   2,  -4,  -2,  -1,   0,   2,  -7,   5,   3,   0 },
     {   -1,   1,  -1,   3,  -7,  15, -14,   4,  -2,  -2,   2,  -1,   6,  -7,  26, -14,  -3,  -2,  -7,   2, -11,  20, -22,   8,   1,   1,   1,  -5,   9,  -9,  26, -12,  -5,  -1,  -9,   6, -14,  26, -28,   4,   4,  -3,  13,  -8,  27,  -9,  31, -11, -10,  20, -18,  23, -21,  27, -46, -32,   1,  -7,   3,  -7,   2, -10, -24,  50 },
     {    8,  -3,  11,  10, -23,  -6,   8,   2,   3, -21,  -4,  25,   0,  -4,  -6,  -1,  19, -26, -36, -11,  -3,  -1,  15,   5,  80,  26, -12,  13,   4,  13,   1, -13,  18, -12, -40,   2,  -1,   0, -17,   0, -13, -18,   6,  24,  -1,  -4,   2,  11,   2,   1,  11,  -4,  -8,   1,   6,  -2,  -2,  -2,   1,  -3,   2,   4,   2,  -2 },
     {    3,   4,  11,  32,   9, -14,   1,   0,   0,  -1,  11,  33,  27, -33, -13,   0,  -6, -15,  -8,  21,   2, -22,   1,   5, -18, -20, -24,   3,   1, -27,  -1,  11,  -5, -25,  -3,  10,   9,  -3,  40,  16,   3,  13,  21,  36,  13,  -9,   9, -29,   1,  10,   1,  11, -38, -21,  -1,  -6,   2,   0,   1,  -4, -14,  16,  10,   2 },
     {   -1,  -3,   0, -29,  -7,  23,  -3,   6,   4,  -2,  12,  11,   6,  -1,  -9,  -1,  -7,   2,   0, -18,  19, -19,  -1,  11, -13, -24,  40, -13,  -5,  -9,  20,   0,  63, -44, -14,  10,   4,   5,   7, -13, -20,  -3,  41,  -5, -13,   7,  -3,  -2,  -4,  20,  -8, -24,  14,   3,  -2,   0,  -3,   0,  -5,   7,   5,  -9,   2,   1 },
     {  -10, -12, -12, -12, -14, -11,   6,   6, -15, -19, -19, -20, -20,  -6,   9,   8, -17, -22, -20, -18, -21,  -8,  13,   9, -14, -19, -22, -18, -13,   4,  22,  14, -13, -11, -14, -12,  -6,  19,  34,  14,  -5,  -8,  -4,   1,  16,  32,  35,   5,   6,   7,  14,  13,  24,  28,  23, -23,   1,   4,   3,   5,   5,   8, -12, -29 },
     {    2,   2,   1,  23,  24, -63,  22,  -4,   1,  -3, -15, -29,  23,  -9,  31,  -9,   2,   9,  36, -23, -17,  19,  -6,  -8,  -2, -31,  12,  26,  -2,  15, -11,   0,  42,   8, -23,  -3, -10,   5,  -6,   5, -16, -10,   0,  11,   5,   2,  -2,   1,  -4,  10,   7,   1,  -6,  -3,   0,   1,  -1,   0,  -2,  -2,   0,   3,   1,   1 },
     {    2,   1,   6,  12, -19,  58, -17,  -2,  -3,   1,  -8,  -7,  34, -47,  18,  -5,   3, -12,  16, -24,  -2,  29,  -3,  -9,  -1,  -5, -17,  19, -25,   1, -26,  22,  -2,   3,   5, -18,  16,   0,   8,   0,   7,   0,   4,  14, -21,   5,  -6,   0,   4,  -7,  16, -25,  -2,  25,   5,   6,   0,   0,   0,   0,  25,  -7, -13,   5 },
     {    1,   2,  10,  -3,  20,  -2,  -3,   2,  -1,   0,  -5,  10,  -6,   8, -17,  -5,  -1,   3,  11, -12,  24,  -5,   3,   8,   2, -21,   1,  10, -14,   9, -18,  -4, -37,  21, -24,  -4,   7, -14, -14,  36,  41, -38,  44,  -2,   5,   3,  21,   2, -10,  19,   4, -27,  27, -20,  -4, -22,   0,   4, -13,  13,  -7, -12,   4,   8 },
     {   -4,  -5,  -4,  -8,  -5,  -7, -18,  19,  -4,  -8,  -4,  -9,  -8,  -8,  -5, -10,  -7,  -7,  -8,  -8, -13,  -6, -33,  30,  -7, -12,  -4,  -8,  -9, -16,  -2,   4,  -5,  -6, -10,   0, -16,  -8, -21,  40,  -5,  -2,  -6,  -6,  -8, -19,  17,  18,  -3,  -5,  -4,  -4, -14,   6,   3,  61,   9,  -7,  16,  -9,  24,  -4,  57,   4 },
     {    5,  10,   4,  33,  57,  -2, -21,   5,  -7,  -8, -30,  -5,  23, -39,  -8,   3,   2,  11, -26, -46,   2, -11,  33,   8,   3,  24,  24, -28,  -1,   3,  20, -16, -15,  -7,  13,  -6,   5,   2,  -7,  -6,  -3,  11,  -6, -16,   0,   4,  -2,   0,   0,  -1, -13,   2,   9,  -2,  -3,   6,  -1,   1,   3,   6,  -3,  -7,   6,   2 },
     {   -4,   0,   0,  -7, -21,  -1,  35,  -4,  -5,   3,   5,  -6,  -7,  -1,   7,  -7,   3,  15,  16,  -9,  -4,  16,  36,  -2,  -9,   3,  36,  10,  -7,   8,  -9, -30, -62, -44,   3,  17, -19,   2,  -6,  10, -21,  20,  23,  15, -10, -12,   0,   4,  -3,   6,   5,  -6, -20,  11,   8,   4,  -3,   1,   2,  -9,   9,  14,  -4,  -3 },
     {   -2,   1,   2,   6,  13,  -1, -36,  -5,  -1,   4,   8,  19,  18,  16,  13,  10,  -9,  -6,  -7,   0,  17,  -2, -31,  -7, -11,  -8,  -5,  -7,  -6,  15, -23,  14, -24, -19, -18,   1, -13, -15, -21, -18, -44,  27,   1,  14,  17,  17,   2,  50,  14,  12,  21,  11,  18, -13,   9,  -1,  -5,   3,   2,  -9, -21, -12, -15,   9 },
   }
 },
 {  //9
   {
     { -121,  31,  12,   5,   3,   2,   1,   1,  -3,   9,   3,   0,   0,   0,   0,   0,  14,   3,  -3,  -2,  -1,   0,   0,   0,  13,   0,  -4,   0,   0,   0,   0,   0,   9,  -2,  -2,   0,   0,   0,   0,   0,   5,  -2,  -1,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {    0, -33,  12,   1,   3,   1,   1,   0,-113,  10,  22,   4,   5,   2,   2,   0,  -9,  26,   5,  -2,   0,  -1,   0,   0,  19,  11,  -5,  -3,  -1,  -1,   0,   0,  14,   1,  -5,  -1,  -1,   0,   0,   0,   9,  -2,  -3,   0,   0,   0,   0,   0,   4,  -1,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },
     {   24,  28,  -8,  -7,  -3,  -3,  -1,  -1, -10,  69, -14,  -8,  -4,  -3,  -1,  -1,  74,  18, -32,  -5,  -5,  -1,  -2,   0,  23, -31, -19,   4,   1,   2,   0,   1,  -5, -26,  -1,   7,   2,   2,   0,   1,  -9, -12,   6,   4,   1,   1,   0,   0,  -5,  -3,   4,   1,   0,   0,   0,   0,  -3,  -1,   2,   0,   0,   0,   0,   0 },
     {   13,  83, -19,  -5,  -6,  -3,  -2,  -1, -25,  41, -18,  -1,  -1,  -1,   0,   0, -71,  -8,  -2,   8,   5,   3,   2,   1, -23,  -2,   8,   6,   2,   1,   0,   0,  10,   1,   4,   0,  -2,   0,  -1,   0,  13,  -4,  -1,  -1,  -1,   0,   0,   0,   7,  -6,  -1,   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 },
     {  -10, -11,  -6,   7,   2,   1,   1,   1, -35,  -2, -11,  12,   1,   3,   1,   1,  26, -63,  -5,  16,   0,   4,   0,   1, -38, -53,  38,  15,   2,   3,   1,   1, -20,  11,  42,  -2,  -3,  -1,   0,   0,   1,  27,  14, -14,  -4,  -2,  -1,  -1,   5,  14,  -3,  -9,   0,  -1,   0,   0,   2,   4,  -3,  -3,   1,   0,   0,   0 },
     {   -5, -13,  53, -12,   1,  -2,   1,  -1,  11,  44,  44, -23,  -3,  -4,  -1,  -2,  -8,  25,  -1, -22,   1,  -2,   0,  -1, -71,  -3, -10,  -4,  10,   2,   3,   1, -39,  13,  -4,   7,   8,   0,   1,   0,   0,  15,  -7,   4,   0,  -2,   0,   0,  10,   5,  -8,   1,  -2,  -1,   0,   0,   4,  -1,  -5,   1,  -1,   0,   0,   0 },
     {   -8, -11,   0,   2,   2,   1,   0,   0,   2, -22,   2,   3,   3,   1,   0,   0, -41,  23, -11,   3,   3,   0,   1,   0,   7, -36, -35,  17,   3,   4,   1,   1, -23, -63,  18,  34,   3,   4,   1,   1,  -7,   1,  51,  11,  -8,  -2,  -2,  -1,   5,  21,  23, -15, -10,  -2,  -1,   0,   2,   9,  -1, -13,  -2,   1,   0,   0 },
     {  -20, -40, -62,  17,  10,   4,   3,   2,   4,  26, -53,  20,   5,  -1,   1,   0,  -2,  55,  -5,  12,  -4,  -5,  -1,  -2, -38,  23,  12,   2,  -6,  -2,   0,  -1, -22,   7,   7,  -3,  -4,   0,   1,   0,   4,   0,   1,  -2,  -1,   1,   0,   0,   9,  -4,  -1,   0,   0,   1,   0,   0,   3,  -5,   0,   1,   0,   1,   0,   0 },
     {   -7, -56,  30,  -6,   7,   0,   3,   1,  19,  42, -28, -13,   0,  -2,   0,  -1, -35, -30, -53,   7,   8,   4,   3,   2,  13,  -6,  -8,  23,   5,   2,   0,   1,  49,  14,   4,   8,  -8,  -3,  -3,  -1,  11, -10,  -1,  -3,  -6,   0,   0,   0,  -4, -11,   3,   1,   1,   2,   0,   1,  -3,  -2,   5,   2,   1,   1,   0,   0 },
     {   -4,  11, -23, -28,   4,   2,   2,   0, -24, -48, -31, -34,  14,   7,   5,   2,  13, -14, -47, -23,  18,   5,   5,   1, -23,   6, -42,   4,  24,   2,   3,   0, -13,  22, -22,  23,  17,  -2,   0,  -1,   8,  12, -13,  23,   3,  -5,  -2,  -1,   5,  -4,  -3,  16,  -3,  -2,  -1,   0,   2,  -4,   3,   8,  -3,  -1,  -1,   0 },
     {   -3,  -8,   1,   3,  -1,   0,   0,   0,  -8,  16,  -2,   8,  -1,   0,  -1,   0,  15, -44,  20,  12,   3,   1,  -1,   0, -32,  34,  13,  -5,   1,  -3,   0,  -1,   3, -14, -46,  13,   2,   1,   1,   1,  -9, -46,  23,  43,  -2,   4,  -1,   1,   7,   6,  50,   0, -19,  -3,  -3,  -1,   3,   9,   9, -22,  -8,   0,   0,   0 },
     {   10,  17,  22,  73,   4, -12,  -6,  -3,  -1, -10,   8,  69,   1, -13,  -4,  -1,   4,  -6, -41,   9,  -9,  -5,   2,   1, -14,   2, -43,  -9,  -4,   3,   5,   1,  -2,  18, -14,  -1,   0,   3,   1,   0,   9,   8,  -3,   0,  -3,   1,  -2,  -1,   5,  -5,   1,   1,  -3,   1,  -1,   0,   2,  -4,   4,   0,  -1,   1,   0,   0 },
     {    2, -11, -19,  -2,  -2,   2,   1,   1,  11,  -8,  17,   4,  -1,  -1,  -1,  -1,  13,   7,  28,   6,  -4,  -4,  -4,  -1, -25, -47,  -7,   5,   2,   2,   0,   1,  44, -27, -20,   8,   3,   2,   1,   1,  77,  -8, -23,   6,  -3,  -2,  -3,   0,  33, -19,  -9,   6,  -4,  -1,  -1,   1,   1, -11,   3,   4,  -2,   1,   1,   1 },
     {    7,  21,  19, -19,   7,  -6,   1,   0,   1, -19,  14,   1,  20,  -4,   1,   0,  25,  24,  -7,  45,  19, -12,  -3,  -2,  -9,  28,   8,  62,  -6, -20,  -6,  -3,   0,  13,  13,  28, -39, -16,  -3,  -1,  10,  -9,   1, -12, -34,   4,   4,   3,   3, -11,   2, -13,  -8,  11,   3,   2,   0,  -2,   3,  -6,   0,   5,   0,   0 },
     {    7,   7,  -4,  -2,  -8,  -1,  -2,  -1,   7,  10,  -6, -12, -11,  -2,  -1,  -1,  16,   3,   2, -22,  -8,   1,   1,   1,  13,  32,  -5, -20,  -4,   2,   2,  -1,  28,  30,   9,   6,  -1,   0,  -1,  -2,  39,  48,  46,   0, -16,  -7,  -6,  -2,  30,  36,   9, -37, -20,  -5,  -3,  -1,   5,   3, -18, -24,   0,   4,   1,   1 },
     {   -6, -10, -34, -25,   9,  -1,   6,   2,   6,  28,  22,  12,  23, -14,   0,  -2, -10, -14,  27,  43,  21, -14,  -6,  -3,  15, -23, -43,   5,  -6, -10,   3,   1,  -4,  27, -40,  -6,  -6,   0,  11,   3, -18,  45,  -2,   7,   1,   1,   3,  -3, -17,  20,   0,   2,  -2,   0,  -1,  -4,  -4,   4,  -2,  -2,  -1,   0,   0,  -1 },
     {    4,   5,  66, -35,   0,  -8,  -1,  -2,  -4,  -9, -69,  17,   7,   2,   3,   1,   3,  12,  14,  43,  -4,   1,  -1,   0,   4,  -4,   3, -24, -17,  -3,  -2,   0, -16, -16, -17, -15,  13,   8,   7,   3,  17,  19,  -3,   8,  11,  -1,  -3,  -3,  10,   3,  -8,   3,  -4,  -7,  -2,  -1,   3,  -2,  -2,   1,  -3,   0,   1,   0 },
     {    4,   6,   3,   2,  74, -16,  -8,  -1,   3,   1,  -2,  -7,  75, -24,  -7,  -3,  -1,  -5,   0, -33,  10, -20,   3,   3,   0,  -5,  13, -31, -16, -10,   8,   2,   7, -11,  15,  -6, -13,   0,   7,   1,   3,  -8,   4,   4,  -6,   3,   4,   0,   4,  -1,   1,   3,  -4,   0,  -1,  -1,   1,   0,  -3,   1,  -2,   1,   1,   1 },
     {    5,   5,   7,  22,  18,   7,  -4,   2,   8,  12,   6,  14,  18,  12,  -7,   1,   3,  13,  13,  13,  29,   8,  -5,  -1,  19,   7,  25,  20,  43,   1, -11,  -5,   6,   8,  17,  24,  48, -13, -18,  -8,   2,  14,  -2,  29,  20, -37, -19,  -8,  -6,   1,  -7,  15, -15, -36,  -7,   2,  -1,  -1,  -3,   4, -18, -17,   5,   6 },
     {    0,  -4,   5,  -2,   1,   5,   3,  -1,  -5,  13, -21,  16,  -4,   5,  -7,  -1,  11, -31,  43, -22,  14,   0,  -4,  -3, -22,  41, -50,   8,  12, -11,  -1,  -2,  22, -44,  26,  10, -16,  -1,  -9,  -1, -20,  18,   1, -28,   0, -16,   1,   2,   3,  -6, -27,  15,  -1,   1,  12,   3,   0,  -9,   5,  14,   0,   9,   3,   2 },
     {    1,  10,   8,  -1,   6,   4,  -2,  -3,  -4, -17,  -9,   2,   1,  10,  -4,   1,  11,  33,  -6,   1,   2,   2,   1,  -5, -48, -15,   2,   3,  10,   3,   4,   0,  67,  -2, -10, -17,   6,  -2,   5,   0, -22,  -4,   5, -11,   9,   0,   3,  -5, -51,  41,   4,  -9,   6,  -1,   1,  -4, -17,  31, -17,  -3,   2,  -5,   0,  -4 },
     {    3,   0,  -3,  20,  12,   3,   4,   4,   2,   3,   5, -11,  -4,   2,  -2,   1,  -1,  -9, -24,   4,   7,  -3,   3,   2,  11,  18,  18,  30,   3, -14,   1,  -2,  -6, -37, -26, -14,   0,  -3,  11,   0,  -4,   7, -37,  -8,  15,   9,  13,  -3,  49,  63, -19,   4,   4,   3,   2,  -5,  28,  21, -20,   3,  -6,   0,  -7,  -4 },
     {    2,  -1,  19,  52, -18, -13,  -8,  -2,  -3,   4, -43, -23,  11,  -1,   6,   1,  -3,   8,  46, -39,  17,  11,   0,   0,  13, -27,   4,  25,   5,  -2,  -9,  -9, -13,  19, -30,  26, -27,  -5,  -1,  -2,   4,   9, -12,   0, -16,  16,   6,   7, -15,  11,  10,   0,   6,   9,  -6,  -1,  -6,   3,   2,   0,   0,  -4,  -5,  -3 },
     {    4,   4,   5,   2,   9,  59,  11,  -2,   6,   6,   6,  12,  29,  63,   9,  -6,   0,   3,  -6,  -1,  27,  40, -11,  -7,   2,  -6, -11, -18,   9,  -1, -19,  -4, -13,   3,  -9, -33, -17, -18, -14,   6,  14,  -7,  10, -19,  -9,  -1,   9,   8,  12,  -1,  20,  -6,   8,   7,   3,  -3,   3,  -1,   3,   0,   8,  -3,  -8,  -8 },
     {    1,   4,   5,  -8, -14,   6,  -6,   2,   3,  -4,   2,   4,  -8,  -1,   7,   3,   1,  17,  -1,  -9,  -9,  -9,   1,   4,  -1, -12,   1,   1, -11,  -6,   7,   1,  20,  14,  24, -15, -32,  -6,  -1,  -2, -12,  19,   7,  14,   2,   0,   5,  -5,  20,  14,  39,  69,  -8,  -7,   0,  -7,  27,   6,  39,  15, -44, -11, -11,  -1 },
     {   -4,  -7,  -4,  -3, -29, -47, -27,   1,   0,  -4,   1,  -6,   5, -44, -14,   5,   4,   7,  -3,  11,  50,  14,   6,   2,   0,  -2,  -4, -14,  44,   7,  -2,  -4,  -4,  -5,   1, -46,   2, -22, -10,   0,   4, -14,  15, -26,   0,  -7,   7,  14,  15,  -3,  15,  -3,   9,  -1,  12,   4,  10,   1,   4,  -3,   2,  -3,   7,  -4 },
     {    0,  -1,  -4,  48, -22,  -2, -11,   9,  -2,   0,   9, -67,  22,  16,  -4,  12,   3,  -4,  -9,  48,   4,  10,  -1, -14, -11,   9,  -6, -29, -27,   7,  -9,  -7,   6, -18,   4,  -5,  -6,  13,   5,  -8,  -3,   9,   4,  -2, -11,  -6, -16, -15, -11,  -2, -18,  14, -20,  -4,   5,   7,  -1, -10,   6,   2, -12,  12,   6,  14 },
     {    0,  -1,   1,   4,  -1,   5,   8,  -1,   0,   0,  -5,   6, -13,  -1,   8,  -3,  -5,   6,  -1,  -3,   3,  -9,   2,  -1,   6, -16,  16, -21,   5, -14,   5,   2,   4,   3,  -7,   8,  21, -28,   9,   0, -15,   1,  -8,  16, -44, -45,  23,  -4,  -2,  -2, -10,  -2, -35,  39,  49,  -1,  15,  12,  10,  14,   6,  41,  -2, -24 },
     {    2,   2,   6,   7,  -1,   3,  -5,  20,   0,  -6,  -9,  -9,  10,  17,   8,  22,   1,   8,  14,   2, -10, -25, -12,  11,  -6, -23, -13,   6,   0,   1, -10,  -9,  26,  22,  -5,  -2,  17,  -7,   6,  -3, -56, -10,   7,  -7,  13,  -2,   6,  -2,  50, -33,  -6, -31,   5,   4, -12,  11,  38, -10,  -7, -22,  -2,  -6,  -1,  15 },
     {   -1,  -3,  -3,  -9,  -7, -11,  -7,  -2,   3,   7,  -1,   5,  20,  12, -12,   7,   0,  -6,  10,  25, -37,  -3,   5, -18,   6,   7, -13, -34,  37,  22,   2, -10,  -1,   4,  22,  34,  12, -17,  -3, -13,   3, -13, -24,  -8, -38,  18,   8,  -2,   3,  21,  -2,  13,  21,  25, -16,  -3,   8,  25,   0,  -9,   1, -29, -28,   1 },
     {    0,  -1,   3,   9, -28,   3,   9,   9,   1,   3,   4,   6,  -2,  12,   8,   2,   2,   0,  -7,  -2,  43, -22,  -1,   7,  -1,   7,  12, -18,  15, -35,  13,   7,   6, -11,  17,   8,   9,  -9,  53,   3,   6,  10,  11,  14,   9,  37,  27, -31,  -4, -17,  -4,  -4,  10,  14, -17, -35, -17, -19,  -9,   1,  -7,  -6, -18,  -5 },
     {    1,   1,  -1,  -8,   4, -14,   3,  16,   4,   3,   7,  14,  16,  -5,  50,  13,   5,   5,   5,  -2,   7,  32,  33,  -3,   0,   6,   1,  16,   4,  40,  -2, -37,   9,  -3,  12, -12,   0,  38,  -6, -32,  -9,   9,  -6,  19,  10,  15,  -2, -21,  13,  -4,  -1,  -5, -15,  21,   3, -22, -10,   1,  10,  -2,  16,  20, -12, -24 },
   },
   {
     {   88, -42,  -2,  -4,  -1,  -1,   0,   0, -75,  25,   7,   2,   2,   1,   1,   0,  18,   7,  -7,   0,  -1,   0,  -1,   0,  -3,  -8,   2,   1,   0,   0,   0,   0,   1,   4,   1,  -1,   0,   0,   0,   0,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   61,  -3, -11,  -2,  -3,  -1,  -1,   0,  37, -49,   8,  -1,   2,   0,   1,   0, -79,  33,  13,  -2,   3,  -1,   1,   0,  20,  15, -17,   0,  -2,   0,  -1,   0,  -5, -13,   3,   4,   0,   1,   0,   0,   0,   4,   3,  -2,   0,   0,   0,   0,  -4,   0,  -1,   0,   1,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0 },
     {  -41,  21,  -3,   2,   0,   1,   0,   0, -56, -11,  23,   0,   5,   0,   2,   0,  -8,  64, -16,  -4,  -3,  -2,  -1,   0,  59, -35, -21,   7,  -2,   2,  -1,   1,  -9, -14,  18,   1,   1,   0,   0,   0,   7,   6,  -1,  -4,   0,  -1,   0,   0,  -1,  -1,  -1,   0,   1,   0,   0,   0,   3,   0,  -1,   1,   0,   0,   0,   0 },
     {  -29, -23,  23,  -2,   4,   0,   2,   0, -35,  27, -12,   4,  -1,   1,   0,   0, -30,  -8,  30,  -3,   3,  -1,   1,   0,   6,  66, -31,  -8,  -2,  -2,  -1,  -1,  39, -48, -15,  16,  -3,   4,  -1,   1,  -8,   1,  24,  -6,   0,  -1,   0,   0,   6,   3,  -9,  -3,   2,   0,   0,   0,  -1,  -2,   1,   2,   0,   0,   0,   0 },
     {   18,  84, -38,  -1,  -9,  -1,  -3,   0, -41, -44,   3,  13,   3,   4,   1,   1,  24, -26,  35,  -8,   2,  -2,   0,   0,  -9,  25,  -6, -12,   2,  -2,   1,  -1,   4,  -3, -10,   7,   1,   0,   0,   0,   1,  -1,   2,   2,  -2,   0,   0,   0,   2,  -1,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0 },
     {   25,  11,  -3,  -4,  -2,  -1,  -1,   0,  31,  13, -27,   4,  -4,   1,  -2,   0,  40, -16,   1,  -3,  -2,  -1,   0,   0,  20, -26, -28,   7,   0,   2,   0,   1, -16, -60,  42,  15,   0,   3,   1,   1, -25,  38,  28, -24,   1,  -4,   1,  -1,  -3,   9, -23,  -3,   5,   0,   1,   0,  -3,  -2,   1,   7,  -1,   0,   0,   0 },
     {    5,  20,  -4,   1,  -2,   0,  -1,   0, -11,  43, -34,   4,  -4,   1,  -1,   0, -44, -49,  17,  19,   1,   5,   0,   1,  67, -15,  36, -18,  -2,  -3,  -1,  -1, -26,   9, -10, -14,   8,  -2,   2,   0,   1,   8,  -9,  11,   1,  -1,   0,   0,   1,  -3,   1,   1,  -4,   0,  -1,   0,   1,  -2,   1,  -1,   0,   1,   0,   0 },
     {   13,  11, -22,   4,  -1,   0,  -1,   0,  21,  17,   7,  -8,  -3,  -2,   0,  -1,  17,   9, -32,   4,  -1,   1,  -1,   0,  25,  -4,   1,  -2,  -3,   0,  -1,   0,   9, -29, -45,  10,   4,   3,   1,   1,  -7, -55,  44,  34,  -6,   4,  -1,   1,  -4,  24,  36, -30,  -7,  -2,  -1,   0,  -5,   3, -17,  -9,  10,   0,   1,   0 },
     {   12,  55, -18,  -4,  -4,  -1,  -1,   0,  -1,  53, -20,  -6,  -2,  -2,  -1,  -1, -39,  15, -35,   9,   1,   1,   0,   0, -38, -14,  -2,  25,  -3,   4,   0,   1,  52,  -5,  22,  -4, -10,   1,  -3,   0,   7, -14,  -9, -17,   4,   1,   1,   1,  -2, -11,  -4,  10,   6,   0,   1,   0,   2,  -2,   6,   3,  -3,  -1,   0,   0 },
     {   12,  15,  32, -20,   0,  -4,   0,  -2,  12,  25, -50,  12,   2,   1,   0,   0,  16,  33,  23, -13,  -9,  -2,  -2,  -1,   0,   2, -47, -12,  12,  -1,   2,   0, -34,  17, -16,  29,  -2,   0,   0,   0,  37, -24, -15,   6, -11,   5,  -2,   1,  -5, -30,  19,   8,   1,   3,   0,   1,   0,   6,   8,  -7,   0,  -2,   0,   0 },
     {   -9, -20, -69,  31,   0,   7,   1,   2,  11,  40,  43,  -3, -12,  -3,  -2,  -1,  -7, -21,  24, -40,  10,  -2,   2,   0,   7,   2, -37,  19,  13,  -3,   2,   0, -18,  14,   8,  16, -14,  -2,  -1,   0,  13,  -6,  -7,  -8,  -4,   4,   0,   1,  -1, -12,   5,   3,   4,   1,  -1,   1,   0,   2,   3,  -1,  -1,  -1,   0,   0 },
     {    1,  -7,   3, -12,   4,   0,   2,   0,   4,  -7,  -9,  21,  -7,   3,  -2,   0,   7,  -7,  11, -22,   6,  -2,   1,   0,  11,  -1, -15,  27,  -8,   2,  -2,   1,  23, -22,  23, -39,   1,   1,   0,   1, -34,   1, -44,  34,  18,  -2,   3,   0,  16, -14,  55,  15, -28,   1,  -3,   0,  -7,  14,  -4, -34,   9,   2,   1,   0 },
     {  -13, -15,   8,   0,   0,   0,   1,   0, -22, -24, -22,  26,   0,   4,  -1,   1, -28, -28,  17,   6,  -2,   3,   0,   1, -22, -33,   5,  28,   6,   2,   1,   1,  -4,   8,  42,  21, -15,   0,  -3,   0,  31, -13,  40, -10, -15,   4,  -2,   1,  -2,  16,  34, -36,   0,  -1,  -1,   0,  -3,  17, -19, -18,  11,  -2,   1,   0 },
     {   -4, -10, -12,  18,  -4,   2,  -1,   0,   5,   4, -23,  11,  -3,   3,   0,   2,  -2,  30,  56, -22, -11,  -1,  -2,  -1,  -7, -54,   8, -31,  21,   1,   3,   0,  49,  19, -20,   9,   9, -11,   1,  -1, -42,   3,   1,   3,  -8,   0,   2,   0,  -8,  21,  -9,  -2,   2,   2,   0,   0,   6,  -5,  -7,   5,   1,   0,  -1,   0 },
     {   -6,  18,  -3,  -7,   0,   1,   0,  -1, -10,  31,   5,  -4,  -5,   0,   0,  -1, -24,  28,   2, -11,   6,  -3,   1,   0, -51,   6,   3,  -4,  -1,   2,   1,   0, -56,  -3,  -5,  -3,  15,   0,   3,   0, -24,  45,  12,  31,  -4,  -5,  -1,  -1,  35,  17,   2, -25, -19,   3,  -3,   0,   1,  -7, -25,  -8,  15,   3,   0,   1 },
     {  -15, -31, -73,  16,   9,   2,   2,   2,  -5, -13, -63,  25,   7,   2,   3,   1,  10,  33,  -8,  36,  -2,  -3,  -2,  -1,  -7,  23,  17,   0, -15,  -2,   0,   0, -14,  -9,   1,  -7,  -7,   6,   1,   1,   3,   2,   1,   1,   4,   2,   0,   1,   5, -11,  -1,   9,  -2,   1,  -1,   0,   1,  -5,  11,   0,  -3,   1,  -1,   0 },
     {    5,   4,  16,  72, -16,  -5,  -7,  -1,  -7,  -5, -20, -76,   3,   8,   6,   1,   6,   2,  27,  25,  19, -10,  -1,  -4,  -6,  -5, -19,  10, -27,   1,  -1,   1,  -2,   1,  11, -14,  15,   8,   0,   1,   5,  -6,  -2,  11,   2,  -8,  -2,   0,  -3,  -3,   6,  -4,  -5,   3,   1,   1,  -2,   6,   0,  -3,   6,  -3,   0,  -1 },
     {   14,  14,  11,   8,  -5,  -4,  -2,  -1,  23,  26,  15,  17, -10,  -6,  -6,  -1,  29,  38,  34,   4,  -5,  -5,  -4,  -2,  29,  38,  36,  10, -22, -11,  -3,  -2,  20,  26,  27, -16, -26,   6,  -5,   1,  19,   9,   6, -22, -11,  -4,  -3,   2,  12,   5,  -4, -33, -11,   3,   4,   0,  -5,  -4, -21, -14,  12,   4,   1,   1 },
     {    0,  -8,  -7,   6,  -5,   4,   1,   1,   7,   5,  22,   8,  -2,  -1,  -3,  -2,   1,   3,  15,  -8, -12,   0,  -5,   2, -18, -24,  27, -33,  -4,   7,  -1,   1,  18, -55,  -6,  -9,  32,   5,   2,   2,  71,  23,  -6,  15,   0, -15,  -1,  -3,  -6, -33,   5,  -7,  -7,  -1,   2,   0, -23,   6,   7,  -1,   3,   0,   4,   1 },
     {   -3,   5,   5,  -4,   2,   1,  -3,   0,  -2,  14,  13,   5,  -6,   7,  -2,   3,  -3,  14,  32,  -3,  -9,  -1,  -2,   1, -19,   0,  15, -20, -10,   9,  -8,   0, -37, -26,  11, -37,  -7,  15,  -3,   6,  -5, -41,  -1, -26,  41,  13,   3,   2, -17,  32,  15,  20,  39, -22,  -3,  -4,   9,  18,  -8,   7, -17, -16,   2,   0 },
     {    3,   5,  20,  42, -18,  -2,  -3,  -5,  -7, -10, -26,   3, -10,   4,   0,   0,  -1,   9, -23, -62,  33,   4,   6,   2,   1,  11,  46,  18,  23, -17,  -5,   1, -11, -27, -18,  10, -32,  -6,   7,   1,   5,  12, -14,  -5,  -8,  20,   4,  -2, -15,   5,   5,   3,  12,   0,  -5,   1,  -3,   0,   6,  -1,  -2,  -1,   0,   2 },
     {   -5,  -9, -15, -26, -15,   4,  11,   0,  -3,  -7, -27, -13,  10,   2,  -1,   1,   1,  -6, -12, -29,  11,  11,  -6,   3,   6,   4, -21,  -2,  -3,  -7,   2,   3,   7,  11,  -5, -36,  13,  13,   6,   4,   7,   1, -19, -10,  14,   3,  -7,   1, -16,  -9, -13, -43,  22,   1,   6,   3, -23,   6, -53,  45,  34,  -7,   6,  -2 },
     {   -1,  -3,  -4,   8, -49,  29,   5,   1,   1,   1,   1, -23,  46, -28,   5,  -6,   0,  -2,  -9,  12, -32,  28, -14,   8,   1,   8,  -1, -31,  39,  -9,   1,   3,   1,  -5,  13,  15, -28,  10,   2,   3,  -3,   6, -14, -13,  23,   9, -13,  -1,  11,   5,  16,   2, -21, -16,   4,  -1,   0,   2, -14, -22,   2,  15,   7,   3 },
     {    4,   6,  19,  62, -24,  -9,   0,  -2,   1,   1,   3,  67, -15,  -8,  -1,  -2,  -8, -10, -36,  -9, -18,   1,  -2,  -1,  -3,   5, -29, -26, -21,  15,  11,   2,  -2,  19,   0, -12,  18,  26,  -3,  -7,  -9,   3,  11,  -4,  10,  -7, -14,   3,  -8,  -4,  -3,   3,  -6,  -6,   3,   9,   6,   2,  -5,   3,   6,   0,   3,  -3 },
     {    2,  -4,   6,   8,  -5,   1,   5,  -4,   2,  -7,  -8,  12,   9,   7,   0,   5,   0,  -9,   8, -14,  -7,   4,   3,   4,   1, -13, -12,  14,  -1,   3,  -7,   1,  -2, -12,   6, -15,  -3,  -2,   4,   4,  31, -37, -17,  14,  10,   4,   0,   0,  51,  38, -32,  -3, -15,  -7,  -3,   3,   0, -76,  10,  12,   2,   9,   2,   1 },
     {   -2,   0,  -6, -10,  -5,   5,  10,   1,  -1,   7,  -2,   2,   2,  -3,  -7,  -5,   1,   0,   2,   1,   0,   2,  -6,   3,  -4,  14,  -7,   9,  -6,  -6,   6,  -4,  -2,  -3,  12,   0,   9,  -1,   3,  -3,  19,  12, -19,  24,  -5, -14,   0,   0, -86,  49,  -1,   3, -29,  10,   5,  -2,  44, -18,  -4,  -3,  -1,  17,  -5,   1 },
     {    1,   2,  -1,   9,  40, -36,  14,  -6,  -4,  -8,   2, -18, -70,  23,   2,  -8,  -1,  -1, -10,  19, -17,  27, -17,   1,  -2,   3,  -9, -34,  36, -27,   8,   1,   4,  -4,  17,   0, -19,  16,  12,   3,   2,   2, -15,   3,   9,  13,  -1,   3,   0,  -1,   8, -16,   0,   7,  -1,  -2,   1, -15,  -2,  -2,   8,   6,   1,   1 },
     {   -1,  -3, -10,  -7,   2,  -1, -22,   6,  -3,  -5, -23, -22, -27, -12,   6,   4,   2,   5, -16, -38,   2,  20,  16,  -1,  10,  23,  13, -17,  22,  35, -12,  -7,   2,  18,  30,   9,  36,  -1, -30,   9,   8, -15,  11,  -7,  12, -50,  11,   3,  11,  18,  -2,   8, -10,   1,  17, -11,  -9,   7,   2, -13, -10,  -1,  -1,  -6 },
     {   -1,  -2,  -2,  14,  -8, -10,   6,   0,   1,   4,   7,   3,  21,   5,  -9,   3,  -2,  -4, -13,   7, -20, -12,  -3,  -1,   0,  20,   6, -28,   3,  23,  -4,   2,  -7, -17,  34,   8,  20, -35,  -1,   2, -17, -19, -42,  -5, -59,  10,  23,   4,   2,   7,  15, -17,  27,  32, -10,   7,  -2,  -7,  -4,   4,  12, -19,  -6,   2 },
     {   -5,  -7,  -3, -24, -54,   3,  17,  -2,   1,   1,   6,   6, -21,  25,  -5, -19,   0,   4,   1,   9,  60, -24, -33,  -2,   0,  -3,  10, -19,   4,  -7,  27,  17,  -1,   0,  20,  17,   6,  29, -13,  -5,  -7, -26,  -1,   5,  -4,  -1,  -6, -11,   9,   0, -15,  10, -17,  15, -10,   5,  -8,   3,   4,  -2, -10, -11,   4,   6 },
     {    3,   5,   2,  15,  21, -28,   0,   0,   4,   7,   2,  14,  27,  -3,  -7,  -1,   3,   5,   0,  10,  20,  -3, -16,   0,   3,  13,  -1,   6,  33, -30,  -7,   7,   4,   6,  23,   5,  23, -29,   9, -14,   2,   0,   0,  27,   0, -15, -17,  -6,  -4,   6,  -1,   4,   1, -58,   4,  11, -18,   5, -17,  17, -31, -29,  30,   6 },
     {    2,  -2,  -9,  -8, -15,  -5,   0, -12,   8,   5,  -5,  -8,   8,  52,   5,  -6,   7,   1, -11,  -6,  -4,   8,   8,  -4,  13,  11,  -8,  -2,  -2,  12,   7,   3,  22,  18,  -2,   4,  21, -10,  -2,   0,  28,  40,   7,   2,  18,  29, -22,   1,  19,  33,  21,   4,  23, -12, -42,   3,  25,   9,  15, -13,   5, -27, -13, -11 },
   },
   {
     {  -88,  53,   1,   4,   0,   1,   0,   0,  63, -16, -21,   1,  -3,   1,  -1,   0,  -3, -26,  15,   6,   0,   1,   0,   1,   2,   6,   9,  -9,  -1,  -1,   0,   0,  -2,  -2,  -4,  -2,   5,   0,   1,   0,   0,   2,   1,   2,   0,  -2,   0,   0,   0,  -1,  -1,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0 },
     {   20,  26, -21,  -4,  -4,   0,  -1,   0,  23, -74,  30,   8,   5,   1,   1,   0, -34,  45,  22, -30,   3,  -5,   2,  -2,  -1,  12, -44,  10,  12,  -2,   2,  -1,  -2,  -2,   8,  21, -14,  -1,  -1,   0,   3,  -1,  -1, -10,  -5,   7,  -1,   1,  -1,  -1,   3,   1,   5,   0,  -2,   0,   0,   0,  -1,  -1,  -1,  -1,   0,   1 },
     {   40,  -7,  16,  -5,  -4,  -4,  -1,  -1,   6, -10, -35,   7,   3,   5,   1,   2,   7, -39,  60,   4,  -9,  -2,  -3,   0, -16,  32,  -3, -51,  18,  -2,   5,  -2,  -8,   7, -30,  33,  16, -12,   3,  -3,  -1,   4,   5,   7, -26,   2,   3,   0,   0,  -3,  -2,  -3,   4,  11,  -3,   0,   0,   1,   2,   0,   1,  -3,  -3,   2 },
     {   41,   7, -20,  11,   0,   2,  -2,   0,  26, -38,  14, -14,  -6,  -1,   0,   1, -22,  -2,  -3,  22,   9,   2,  -1,   0,  15, -13,  44, -20, -26,   3,  -4,   2,  -8,   4,  -6, -44,  45,   4,   1,   0,  -4,  10, -11,  29,  11, -32,   5,  -4,  -4,   3,  -4,   1, -21,   6,  12,  -2,   0,   1,  -1,   3,   3,   6,  -4,  -4 },
     {   38,   7,  -5, -10,  -4,   3,   2,   1,  23,   1, -26,  20,  -4,  -2,  -6,   0,  -4, -36,  17,  -5,  -4,  10,   3,   3,  -8,  15,   8,  19,   0,  -9,  -6,  -2,   5, -15,  14, -24, -33,  24,   2,   6,   4,  -6,  10, -26,  57,   4, -19,   2,  -2,   2,  -2,   8,   3, -47,  18,   3,  -3,   3,  -3,   2,  -9,  13,  17, -14 },
     {   41,  35, -24,   6,  -8,  -3,  -4,  -2,  27,   9, -28,   0,  -3,   4,   3,  -1, -14, -17, -30,  29,  -4,  -1,  -1,   0,  -2, -17,  19,  25, -10,   9,   1,   1,  -3,   8,  12,   6, -25,  -9,  -4,  -3,  -4,  -5,  -3, -12, -29,  38,   8,   3,   3,  -2,   9,  -7,  25,  26, -36,  -1,   1,  -4,   3,  -6,   6, -23, -11,  20 },
     {    5,   1,   7, -17,  16,  -6,  -1,  -2,   0,  11, -21,  27, -25,   7,   2,   1,   6, -14,  20, -28,  26,  -6,  -3,  -1, -15,  19, -30,  35, -24,   1,   7,  -1,  17, -24,  31, -24,  14,   9, -15,   4, -12,  17, -18,  12,  -2, -28,  23,  -6,   1,  -2,   0,   6, -21,  40, -14,  -4,   2,  -3,   3,  -7,  13, -11, -14,  11 },
     {    6, -66,  29,  -3,   9,   1,   3,   0,  79,  10,  23, -29,   0,  -9,  -1,  -4, -13, -13, -23, -19,  17,  -2,   6,   0, -17,  10,  -2,  18,  13,  -8,   2,  -3,  -5,   9,  -9,   6, -10,  -7,   4,  -1,  -5,   1,  -1,   7,  -6,   5,   3,  -1,  -1,   1,  -1,   0,  -4,   3,  -2,  -1,  -1,   0,   1,   1,   1,   1,   0,   3 },
     {  -32, -53,   5,  12,   2,   4,   1,   1, -24, -40,  15,  18,  -4,   8,   3,   1, -16,  -7,  32,  17,  -5,   1,  -1,  -1,   7,  27,  32,   1, -23,   1,  -2,  -1,   4,   9,  -3, -33, -22,   8,  -1,   1,  -2,   0, -11, -13,  10,  32,   3,   2,   5,   3,   4,   1,  22,   7, -31,  -2,   1,  -3,  -1,  -5,   3, -21,  -8,  18 },
     {   -1, -10,  35,   2,   4,  -4,  -1,  -2,   3,   8, -38, -22,   2,  -1,   4,   0, -52,  34,  -6,  46,   5,  -3,  -1,  -3,   9,  26, -24,  -1, -42,   7,  -3,   3,   7, -26,  -3,  13,   9,  26,  -9,   4,  12,  -6,  24,  -9,  -6, -13, -14,   5,   2,  -8,   0, -14,  11,  -1,  12,   6,   1,   1,   3,   2,   4,  -5,   1,  -9 },
     {    2,   3,   2,  -6,   4, -11,  -1,   6,   2,   6,   4,  -1,  -4,  17,  -3,  -4,  -2,   2,   2, -11,   8, -16,   0,   7,   2,   1,  -5,  -2,  -4,  21,  -3,  -5,   2, -11,   1, -14,  26, -22,  -2,  12,  -3,  10,  -7,  22,  -4,  18,   3, -15,   8, -14,  13, -17,  15, -35, -34,  33,   4,  -3,   4,  -4,   6, -27,  79, -21 },
     {   -2,  24, -27,   7,   5,   4,   4,   0, -22,  19,  11,  -1, -11,  -8,  -2,  -3, -21,  -7, -10,  -6,  21,   5,  10,   0,  -7,  49,   1,  19,  -9, -25,  -2,  -9, -12,   8, -55,   1, -22,  18,  18,   3,   3,   1,   2,  46,  -2,  19, -19,  -8,   4,  -8,   5, -10, -29,   3,  -9,  13,   3,  -4,   3,  -2,   9,  15,  -2,   0 },
     {  -13, -23, -21,   5,  15,  -3,   3,   3, -22, -35, -22,  13,  20,  -3,   0,   1, -16, -28,  -6,  26,  23,  -7,   3,   3,  -6,  -1,  19,  39,  15, -11,  -4,  -2,  10,  17,  29,  24,  -9, -28,  -3,   0,   4,  10,   9,  -4, -34, -25,   1,  -2,  -1,  -4,  -5, -16, -25,  -5,  23,  13,   1,   2,   2,  -1,  -1,  20,  25, -12 },
     {    2,   0, -25,  27,   7,   5,  -6,  -1,   2, -10,  31, -14, -24,   3,  -1,   5,  44, -21,   2,  -7,  25,   2,  -9,  -2, -25,   3,  -2,   3, -23,  -6,   5,   5,   3,   9,  -2,  12,   9,  33, -19,   1,  10, -21,  25, -16, -13, -23, -21,  18,   6, -10,  11, -31,  33,   1,  20,   6,   0,   0,   1,   1,  13, -33,   3,  -8 },
     {    0,  -1,  -2,  -2,   4,   4, -10,  -3,  -1,  -3,  -6,  -2,  13,  -1,  -4,   9,  -2,   0,  -6,   1,   6,   7, -16,  -3,   1,  -3,  -2,   0,  14,  -4,  -3,  11,  -2,  -3,  -3,  12,   9,   6, -20,  -4,   9,   3,  18,  11,  16, -21, -14,  18,  -4,   1,  -4,  -5, -24, -31, -27,  -6,  -2,  -6,  -2,  -8,  -7, -16,  17,  98 },
     {  -10, -30, -62,  28,  -7,   8,  -2,   3,  21,  42,  -7,  48, -13,   5,  -6,   2, -38,  21,   3,  -9, -31,   9,  -8,   3, -15,   9,  -2, -14,  12,  22,  -4,   8,  11,   9,  -2,   2,  13,  -8, -17,   0,   4, -14,  -2,   2,  -1, -14,   4,  13,   2,  -2,  10,   0,   0,   1,  10,  -3,   2,  -1,   1,  -4,   0,   1,   1,  -6 },
     {    1,  14, -16,  28, -16, -11,  -3,  -1, -10,   7,  27, -34,  12,  13,   3,  -4,   0, -18,  -7,  12,   9, -28,   5,   1, -24,  45,  -4,   3,  -9,  24,   5,  -7,  21, -19,  -6,  -1,   5, -26, -16,  11,  -1, -10,  16, -20,   8,  -3,  44,  -9,  -9,  17, -19,  20,   2,  -8,  -6, -25,  -4,   4,  -6,  10, -26,  18,   4,   8 },
     {  -15, -22, -32, -22,  -5,   9,   8,   4, -22, -32, -35, -33,  -5,  15,  13,   4, -20, -32, -32, -19,   3,  19,  14,   5, -20, -17, -20,  -5,  16,  26,  10,   4,  -1,  -6,  -1,  11,  27,  23,   9,   2,   5,   5,  11,  18,  24,  15,  -1,  -7,   7,   9,  11,  16,  15,   2, -10,  -9,   1,   3,   2,   4,  -2,  -8, -17, -11 },
     {    0, -33, -42,  10,   3,   1,   6,   5,  27, -17, -30,   9,  10,   5,   3,  -4,  69,  30,  -1,  23,  -5, -15,   2,  -1,  16,   0, -19,  -2, -25,  -5,  -1,  -5, -11, -27,  -8,   5,  -9,   7,  15,   7,  -5,   6,  20,  15,   1,  16,   6, -14,  -7,   2,  -1,   1, -17,  -1, -12,  -3,  -1,   1,   3,   1,  -4,  18,   3,   1 },
     {   -1,   6,  20,  -9, -12,  -8,  -1,   1,   2, -16, -18,   9,  20,   5,   5,  -1,  20,  28, -14,  16,  -8, -21,  -4,  -5, -75,   4,   4,   4, -17,  13,  11,   3,   1,  52,  -8,  -6,  -1,  14, -13,  -6,   4, -12, -34,  11,   9,  -4,  -9,   5,   5,  -2,  14,  21, -11,  -7,   0,   6,   1,  -2,   1, -11,  -8,   7,   7,  -4 },
     {   -1,   0,  -1,   3,  -9,  10, -15,  10,  -1,  -2,   3,  -7,   9, -16,  22, -17,  -4,   1,  -4,   7, -13,  16, -20,  16,   4,  -6,   4, -11,  15, -21,  22, -18,  -6,   7,  -9,  12, -19,  24, -22,  18,   7, -11,  13, -18,  22, -25,  27, -19,  -9,  11, -18,  19, -24,  29, -26,  14,  13, -15,  20, -18,  28, -22,  30, -23 },
     {   -1,   1, -10, -20,  21,  -5,  -6,  -3,  -2,   3, -25,  -1,  14,  -5,  17,   7,  -2,  18,   1,  -4,  29,  -7, -15,  -6,  17,   6,  10,   6,  10, -35,  18,   8,  -9,   2, -24, -12,  -8, -16,   0, -11,   4, -19,  -4,  13,  10,  -4,  40,  28,  12,  -1,  26,   3,  21,  -6,  20, -54,  -6,  -1,  -9,  -2, -13, -31,  16, -15 },
     {    2,   5,   7, -64,  14,  13,   7,  -1,   4, -15,  26,  31,  19, -23,  -5,  -3,  15, -11, -26,  14, -37,  10,  19,   0,  -4,  24,   6,   6,  -1,  29, -24,  -2,   6, -15, -16,  -5,   5, -11, -13,  19,  10, -21,  24,   8,  -1,  -8,  16,   6,   3,  -2,  10, -24,   0,   8,   6,  -7,   0,  -2,   1,   1,  18,  -6,  -6,  -1 },
     {   -4,  -6,  -7, -10, -16, -11,   1,   8,  -7, -10, -14, -18, -25, -16,   6,  14,  -5,  -7, -10, -17, -21,  -9,   6,   9,  -7,  -7, -10, -16, -16,  -7,  12,  16,  -5,  -5, -11, -20, -16,   0,  17,  12,  -3, -10, -16, -15,  -9,   8,  26,  26,  -1,  -9,  -6,  -7,   9,  24,  39,  17,   6,  10,   9,  13,  26,  33,  44,  41 },
     {   -7,  -1, -48, -54,  32,   3,   7,   1,   5,  34,  15, -33,  44,  -9,  -3,   2, -12,  10,  38,  10,   8, -19,  -1,  -1, -21,  -3,  10, -14, -10, -12,   6,   3,   2,   3,  22,  -4,   1,   8,  14,   3,  -2,  10,  -5, -15,   0,   4, -13,  -8,  -3,   3, -14,   4,  10,   6,   0,  18,   2,  -2,   3,   8,   5,  -7,  -7,   8 },
     {    2,  -2, -14, -20, -16,  10,   0,   1,   4,  -4, -28, -34,  -5,  23,  -7,  -2,  26,  23,  -6,  -8,  18,  28,  -4,   7,  40,  45,  10,   9,  23,  14, -23, -11,  22,  24,  -7, -10,   2, -13, -23,   3, -12, -18, -29, -16,  -4, -11, -14,   6,  -8, -11,  -8,  12,   9,   5,   6,  25,   2,  -4,   4,   4,   6,   0,  -7,   4 },
     {   -1,  -6,  -6,  -5, -24,  -1,   6,   1,   0,  13,  13,  -5,  27,  15,   3,  -6,  -3, -30,  16,   1, -10, -28,   8,   6,  45,   1, -23,  18, -15,  28,  18, -16, -60,  25,   9,   5,  -2,  15, -33,   8,  10,  10, -13,   8,   3,  -2,  -1,  26,   3, -15,  16,   2,  -7,  -2,  11,  -8,   1,   2,   4,  -6,   2,   7,  -3,  -1 },
     {    2,   1,  -2,  34, -33,  25,  -2,  -5,  -1,   0,   5, -24,  43, -25,   0,  15,   9, -15,   5,  13, -18,   6,   9, -15,  -4,  15, -23,   6,  10,  -6, -20,  26,  22, -24,  15, -12,  -1, -10,  23, -15, -13,  11, -21,  14,   7,   0,  -7,  18,  14,  -7,  21,   8, -11,   9,  13,   3,   5, -13,  11, -26,  17, -30,   4,  -8 },
     {   -1,  -4,   9,  -8, -41,   1, -10,   0,   2,   3,  -5, -16,  17,  34,  -5,   3,  -6,  10,  32, -19,  10, -17,  -8,  22,  -6, -34,  22,   9,   4,   6,   7, -14,  39, -11,  -5, -12, -12,  -1,   1,   9,   5, -32,  32,  17,  -3,   9,  -1,   1,   9,  -5,  22, -31, -22,   3,   0,  -8,  -4, -10,  -4, -11,  25,  20, -16,   0 },
     {    2,  -3,  -6,  -7,  18, -33,  -8, -14,   4,  -9,   7,  14,  12,  -6,  10,  -9,  -2, -24, -23,  -2, -21, -29, -11,   8,  35,   2, -14,  -9,  -4,   2,  41,  -4,  61,   7, -17,   1,   3,  19,  -1, -23, -36,  -7,  -5,   3,  -6,   4, -25,  13,  10,   3, -14,   6,  -7,  -2,   2,   6,  -6, -15,  -7,   5,  -8,  -5,   0,  -4 },
     {    3,   5,  -2,  30,  67, -18,  -5,  -4,  -4,  -2, -14, -42,  -9, -17,  -1,  -3,   2,   5,  29,  -7, -37,  12,  36, -22,  -1, -18,  -3,  13,   8,  18,  -9, -22,  12,  13,  -8,  -7, -11,   0, -25,   8,  -3,  -6,  -5,   2,   6,   2,   4,  -5,   7, -10,  14,  -8, -13,  -3,  -3, -15,   0,   5,  -1, -10,   7,   6,   3,   2 },
     {   -2,  -3,   2,  13,  23, -12, -17,   3,  -6, -10,  -8,   3,  14,   8, -34,  -5,  -7, -11,  -9,  -6,   2, -21, -22,  18,  -7, -22, -15,  -4,  -8, -18, -30,  -2, -13, -17, -26,  -5, -10, -26, -15,  41,  -8,  -7,   0,   5,   9,  -2,   5,  27,   3,  15,   7,  25,   2,  -2,  10,  49,   4,   0,   3,   3,  -8, -19, -35,  -4 },
   }
 },
 {  //10
   {
     {  102, -48,  -2,  -4,  -1,  -2,   0,  -1, -56,   3,  13,   1,   3,   0,   1,   0,  -2,  17,  -2,  -3,   0,  -1,   0,   0,  -3,  -1,  -5,   1,   1,   0,   0,   0,  -1,   2,   1,   1,   0,   0,   0,   0,  -1,   0,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   59,  20, -16,  -1,  -5,  -1,  -2,   0,  75, -51, -16,   2,  -2,   1,  -1,   1, -37, -31,  25,   5,   3,   1,   1,   1, -13,  16,  13,  -7,   0,  -2,   0,  -1,  -8,   4,  -3,  -3,   2,   0,   1,   0,  -4,   4,   0,   1,   0,  -1,   0,   0,  -3,   0,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {  -27,  -8,   8,   1,   1,   1,   0,   1, -22, -63,  33,   3,   6,   2,   2,   1, -29,  48,  44, -18,   1,  -5,   0,  -2,  13,  40, -33, -25,   3,  -3,   1,  -1,  -1,  -4, -24,  13,   9,   0,   2,   0,   5,   2,   0,  10,  -5,  -3,  -1,   0,   1,  -1,  -1,   0,  -4,   2,   1,   1,   1,   0,   0,   0,   0,   1,  -1,  -1 },
     {   22,  56, -24,  -6,  -5,  -2,  -1,  -2,  18,   4,   0,  -6,   2,  -3,   1,  -1,  -8,  50, -35, -13,   1,  -4,   1,  -2,   5, -29, -55,  24,   8,   4,   3,   2,  -4, -21,  15,  40,  -4,  -1,  -1,   0,   6,  -3,  15,   1, -20,  -1,  -2,   0,  -1,  -5,   4,  -8,  -4,   7,   1,   1,   1,  -1,   1,  -1,   2,   3,  -2,   0 },
     {   -6,  45, -36,  10,  -5,   2,  -1,   1, -50, -42,   7,  12,   3,   4,   2,   2, -33,   7, -10,  24,   0,   4,   0,   1,  25, -21,  39,  21, -18,   3,  -5,   1,   7,  10,  31, -35, -11,   2,  -2,   0,   4,   2, -10, -17,  13,   4,   0,   1,   0,   2,   0,   2,   8,  -5,  -1,  -1,   1,   1,  -2,   0,  -1,  -3,   2,   1 },
     {   16,  67,  -3,  -8,  -5,  -4,  -2,  -2, -49,  31, -46,  12,  -1,   3,   0,   1,   5, -27,  34,  12,  -4,   1,  -2,   0,  -1,  35,  -7, -28,  -3,  -3,   1,  -2,   5, -13, -30,  13,   9,   4,   1,   1,   4,  -5,   8,  12,  -5,  -3,  -2,  -1,   3,  -5,   0,  -1,  -5,   3,   0,   1,   1,  -1,   2,   0,   1,   2,  -2,   0 },
     {  -21, -35,  -6,   7,   7,   0,   3,   2, -26, -30, -37,  18,   8,   2,   3,   3, -19, -37,  -1,  19,   9,   0,   3,   1, -24,  -1,  -7,  31,  14,  -3,   3,  -1,   6,   6,  27,  51, -11,  -9,  -3,  -2,  10,  13,  29,   0, -38,  -9,  -3,  -1,   2,  -1,   2, -18, -18,  10,   4,   2,   1,   0,   2,  -6,   2,  10,  -1,  -1 },
     {    6,   4, -15,   6,  -5,   2,  -2,   1,  -2, -46,   0,  18,  -2,   6,  -1,   3,  90, -10,  29,  -5, -13,  -1,  -5,   0,  12, -38,  -3, -24,   2,   4,   1,   2, -16, -19,  20,   5,  16,   2,   0,   0,  -6,  13,  14,  -9,   0,  -8,  -1,  -1,  -6,   4,  -4,  -6,   3,   0,   3,   1,  -2,   3,  -1,   0,   3,   0,  -1,  -1 },
     {    0,  -1,  16, -10,  -3,  -1,  -1,   0,   2,   3, -12,  16,   1,   5,  -2,   1,   2,  -4,  28, -18,  -9,  -2,  -2,   0, -13,  23, -27,  15,   8,   3,   1,   0,  20, -14,  29,   6, -43,   5,  -8,   2,   1, -13,   4, -62,  20,  30,  -2,   6,   1,   1,  -2,   9,  49, -19, -11,  -2,  -2,   2,  -2,   3,  -8, -25,  11,   4 },
     {   10,  42,  68, -21,   1,  -4,  -2,  -2, -20, -22,  38, -41,   5,  -1,   1,   1,   6, -40,  -9, -29,  22,   2,   3,   1,  -5,  -8,   7,  10,  26,  -7,   1,  -1,   1,  18,  11,   8,   0, -17,   1,  -2,  -1,  10,  -9,  -7, -11,  -6,   6,  -1,  -3,   2,  -8,  -1,   0,   6,   5,  -1,   0,   3,  -2,   3,   2,   1,  -2,  -2 },
     {    7, -12,  39,  -1,   2,  -3,   0,  -1,   6, -22, -21, -17,   1,   0,   2,   1,   7,  -9, -25,  23,  14,   4,   2,   1,  72,  18,   0,  21, -18, -10,  -7,  -2,  -4, -54, -14,  -5, -17,   7,   4,   5,  -9, -13,  29,  11,   8,  13,   1,   0,   1,   7,  15, -10,  -4,  -5,  -7,  -1,  -3,  -1,  -1,  -8,   2,   1,   2,   2 },
     {    4,   2,  12,  10,  -7,   2,  -1,  -1,   8,   8,  12,   1, -17,   1,   0,  -1,   9,  13,  11,  14, -13,   1,   0,  -1,   8,  11,  15,  -6, -28,   0,   0,  -1,   1,  12,  -1,  -4, -24,   4,   2,  -2,  -1,  -6,  -6, -27, -50,  21,   7,   2,  -4, -15,  -7, -34,  16,  64,  -2,   4,   4,   1,   7,   4,  40,   9, -37,  -7 },
     {   -1, -11,  16,   3,   0,   4,   0,   0,   5,  35,   6,   5,  -6,   2,  -3,   1, -55, -13,  29,  -9, -14,   6,  -4,   2,  44, -37,   0, -33,  10,  15,  -2,   5,  10, -29,  32,   3,  33,  -6,  -6,   0,  -1,  13,  20, -18,  -4, -25,   1,  -1,  -8,   5, -12, -16,   6,   0,  12,   1,  -2,   6,  -4,   5,   8,  -2,   0,  -3 },
     {    0,   0,  -7,  20,  -5,  -3,  -3,   2,  -1,  -4,   9, -26,  18,   4,   1,   0,  -3,   3, -17,  31,  -7,  -4,  -3,   1,   6, -16,  25, -35,  15,   9,  -1,   1, -14,  23, -32,  36,   9, -19,   0,  -4,  13, -19,  25,  -6, -24,  26,  -7,   7,   4,   4,   8, -13,  48, -23, -14,   0,  -2,   1,  -5,   9, -16, -31,  19,   4 },
     {   -1,   0, -62,  19,   2,   7,   0,   2,  -3,  15,  57,   6, -15,  -4,  -7,  -1,   3, -48,   4, -41,   5,   7,   6,   3,  21,  17,   2,  13,  21,  -6,  -2,  -3,  12, -15, -15,   7, -20, -13,   1,   1, -11, -19,  14,  10,  -5,  13,   7,   3,   4,   9,  15, -11,  -5,   3,  -6,  -2,  -3,  -1,  -3,  -8,   6,   2,   0,   1 },
     {    8,  15,  49,  72,  -6,  -8,  -5,  -3,  -1,  -5,  27,  57, -19,  -8,  -3,  -3, -10, -15, -18,   6, -28,   1,   5,   0, -22, -10, -16,   3, -11,  16,   5,   0,  -8,  -7,  -8,   2,   2,  13,  -2,  -2,   1,  -3,   2,  15,   8,   7,  -6,   0,   6,   1,  11,   6,  -5,  -4,  -8,  -1,   1,  -2,   2,  -5,  -6,  -2,   1,   3 },
     {   12,  14,  14,  13,   6,  -9,  -4,  -1,  20,  25,  26,  25,  14,  -9,  -4,  -3,  21,  29,  31,  27,   6, -16,  -7,  -4,  21,  25,  26,  20,  -1, -23,  -9,  -5,  12,  18,  17,   9, -18, -28,  -6,  -2,   9,   8,   7,  -7, -20, -25,  -2,   0,   0,   2,  -9, -11, -28, -27,  12,   4,  -4,  -5, -10,  -9, -22,   8,  29,   6 },
     {   -6,   3,   3, -32,   6,  -1,   2,  -3,  -2,  12,  40,  -1,  22,   0,   0,  -2, -14,  -5,  40,  37,  18, -12,  -8,  -3, -44, -43,   1,   7, -21, -22,   0,   4, -24, -40, -11,  -3, -18,  12,  19,   9,   1,   1,  19,  11,   5,  19,   5,  -2,   6,   9,  10,  -3,  -2,   1,  -8,  -4,  -1,   0,  -1,  -5,   2,   0,  -5,   0 },
     {    4,  -7,   0, -17,  -8,  -1,  -1,  -2,  10, -11,  11,  11,   9,   6,  -2,   0,   7, -23,   3,  28,   1,  -7,  -5,  -2,  22, -35, -35,  -3, -10,  -4,   4,   5,  84,  16, -29,   1,  -3,   2,   3,   0,  30, -18, -26,   5,  -3,  -3,  -1,   0,  -3, -19,   2,  12,  -2,   2,   3,   1,  -1,  -2,   9,   2,  -4,   1,   1,  -1 },
     {    2,   5,  -1,  -2,  11,   8,  -4,   1,   0,  -4,  12, -23,  -2,   1,  -6,   1,   5,  -2, -11, -18,  -2,   1,  -4,   0, -22,  14,   5,  -8, -29,   6,   3,   3,  25,  10,   8,  -9,   8,  46,   0,   1,  30,   4,  39,  -3,  16,  -9, -41,  -7,   2, -17,   0, -55,  -8, -25,  -7,  13,  -1,  -3,  -2,  -9,  30,   6,  23,  10 },
     {    0,  -1,   1,  -8,  18, -14,   1,  -3,   0,   3,  -4,  12, -17,  17,  -3,   5,   4,   0,  10,  -9,  32, -18,   1,  -2,  -4,   7,  -7,  23, -23,  18,  -4,   3,  11, -10,  17, -20,  40, -21,  -5,   0, -11,  18, -21,  33, -31,   7,   9,  -1,  16, -14,  26, -22,  27,   4, -27,   8,  -4,   4,  -6,   3,  -2, -45,  22,   6 },
     {   -5,  -6,   2, -30, -75, -24,  -1,   2,  -1,   0,  11, -14, -64,  -2,  10,   1,   3,   8,  21,  21,  -7,  28,  14,   0,  -4,  -1,  10,  23,  11,  27,  -5, -11,  -1,   0,  -2,  10,   3,   6, -16,  -4,   8,  -1,   5,   9,   1,   0, -12,   1,   5,   2,   6,   0,  -6,  -9,  -4,   3,  -1,  -3,  -2,  -5,  -8,   2,   5,   1 },
     {    3,  -1,   1,  73,   1,  -4, -14,  -1,  -3,   4, -30, -47, -21,   4,   1,   6,   1,  18,  41, -16,  36,  16,  -3,  -3,  -7, -29,   2,  14,  11, -25,  -9,   1,  15,  -5,  -8,  -7, -12,  -2,  14,   8,  18,   4,  -3,   5,   5,   3,   0,  -4,  -7, -10,   0,   5,   0,   0,  -3,  -1,   5,   2,   6,   1,   0,  -3,  -1,   1 },
     {    1,   3,  15, -17,  11,  -3,  -4,  -4,   1, -16, -15,  21,  -3,   5,  -4,  -1,   1,  30, -12, -21, -14,  16,  -7,   6, -35,  -9,  44, -12,  29,   4, -17,   3,  39, -33,  -8, -15,   0, -32,  -8,  13,  11, -39,  19,   5,  -2,  -5,  23,   9,   7,  -1,  23, -11,  -5,   8,  11,  -8,  -1,  -1,  -3, -11,   6,   7,  -1,  -5 },
     {    5,   7,   6,   6,  17,  34,  14,   1,   9,  10,  14,  18,  34,  44,  17,  -6,   9,  12,  11,  20,  33,  49,   9, -10,   4,   8,  10,  19,  32,  30,  -8, -14,   6,   1,  -3,   8,  15,  17, -23, -15,   0,   3,   1,   3,  12,   1, -23,  -8,  -1,   1,   6,   7,   5,   6, -15,  -3,   0,   1,   8,   3,  13,   1, -24,  -5 },
     {    0,  -1,  15, -30,  10,   5,  -2,  -1,   0,   3, -23,  38,  -5,  -3,  -9,  -5, -10,   5,  18, -34,  -7,  15, -10,   4,  34, -36,   5,   3,  21, -16,   2,   0, -35,  44, -10,   0, -25,   8,  13,  -4,  12,  -3, -11,  19,  -1,  26, -17,  -2,   2, -15,  26, -16,  -9,  -7, -19,   5,   0,  -1,   0, -15,  11,  -1,  11,   9 },
     {    3,   3,  -2,  -3,  67,  -8,  16,   0,  -2,  -5,  -3, -32,  -7, -40,   7,  -5,   3,   1,  30,   5, -50,  -9,  24,  -3,   5, -12,  12,  32,   4,  36,  12, -12,   3,  -9, -15,   8,  -3,  10, -23, -11,   4,  -5, -13,   5,  -3,   0,  -6,   8,   6,  -2,  -2,   7,  -3,   2,   1,   3,  -1,  -1,  -1,   3,  -5,   0,   1,  -1 },
     {    2,   0,  -6,  -7,  16,  -4,  10,   3,   2,  -2,   3,  19, -18, -10,  -5,   5,   2,  -3, -18,  10,   1,  19,   4,   6,   3,   5,  -9,   4,  22,  -5, -15,  -6,  -8, -16, -25, -14,  -3, -18,  12,  13,  53,  56,  -1,  -4,   9,  13,  14,   3,  28,  -6, -51, -17,  13,  -6,  -4, -11,  -3, -13, -14,  23,   9,   3,  -1,  -8 },
     {   -2,  -2,  -6,   1,  -3, -75, -19,   7,   2,   2,  -2,  18,  35, -50,   1,  15,   2,   5,  -3,  -3,  43,   8,  34,   5,   2,  -3,   1, -21,  14,  17,  17, -16,   8,  -1,   1, -13,  -3,  11,   2, -17,  -3,   2,   5, -10,   3,   4,  -5,  -8,  -8,   2,   6,  -8,   1,   6,  -5,   0,  -3,   5,   3,  -1,   8,   4, -10,   1 },
     {    0,   7,   3,  -7,  20,   0,   0,  -2,  -1,  -3,  -5,   6, -27,  -5,   2,   4,   0,  14,   0,  19,  -8,  10,  -4,  -1,  -7,  -8, -16,   5,  14, -11,  -1,   3,  37,  33, -18,  -7,  -2,   1,  15,   6, -66,  25,  28,   9,  11,  10,  -3,  -6, -26,  49, -10, -20,   6,   1,  -2, -12,  -4,  13, -21,   4,   5,  -6,  -2,   3 },
     {    3,   1,   2,   1, -12,  -1,  10,   1,   5,   4,  -3, -13,  17,  12,   8,   2,   1,   6,   2, -18, -25, -11,  11,  -4,   5,   6,  21, -17,  -9,  15,   9,   4,  30,  -3,  27,  13,  -1,  -2,  29,   3,  12,  29,   5,  15,  -6,  41,   8, -33,   7,  11,  -3,  15, -23,   2, -53, -23,   1, -10, -10, -19, -27,  14, -19,   6 },
     {    1,   0,  -1,  -5,   2,  -2,  -8,   1,   1,  -1,   3,   0,  -3, -20, -16,   1,   1,   2,  -4,   2,  -3,  -2, -18,   4,   3,   4,   7,   1,   6, -18,  -9,   9,   5,   3,   4,   8,   2,   5,  -9,   8,  13,   6,   9,   3,  10, -24, -14,  14,  12,  11,  -2,  11,  -6,   5, -11,  31,  -1, -18,  -3, -22,  -7, -61, -73,  39 },
   },
   {
     {   74, -55,   8,  -3,   2,  -1,   1,   0, -71,  43,   3,  -1,   1,   0,   0,   0,  23,  -2, -13,   4,  -2,   1,  -1,   0,  -5,  -7,   8,   0,   0,   0,   0,   0,   2,   4,  -2,  -2,   0,   0,   0,   0,  -1,  -2,   0,   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   45,   8, -29,   4,  -4,   1,  -2,   0,  -3, -54,  46,  -7,   5,  -2,   2,   0, -39,  65, -24,  -6,   2,  -2,   1,  -1,  20, -25,  -9,  16,  -5,   3,  -1,   1,  -7,   4,  14,  -8,  -1,   0,   0,   0,   3,  -1,  -7,   0,   3,   0,   0,   0,  -3,   2,   2,   1,   0,  -1,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {  -58,   5,  -6,  17,   0,   3,   0,   1, -19,  31,  24, -23,   2,  -3,   1,  -2,  19,  13, -55,  25,  -3,   2,  -1,   1,  22, -45,  44,  -3,  -9,   4,  -2,   1,  -8,  21,  -9, -17,  13,  -5,   2,  -2,   4,  -7,  -4,  14,  -6,   0,   0,   0,   1,   2,   2,  -6,   0,   2,  -1,   0,   1,  -2,  -1,   2,   1,  -1,   0,   0 },
     {  -55,  -1,  15, -12,   7,   0,   3,   0, -53,  10,   9,  21,  -2,   1,  -1,   1,  19,  34,   2, -32,   4,  -3,   1,  -1,   4,   2, -47,  35,  -1,   0,   0,   0,  20, -23,  36,  -5, -18,   8,  -4,   2,  -1,   6,  -8, -16,  17,  -3,   1,  -1,   3,  -1,   0,  10,  -4,  -3,   2,  -1,   0,   1,  -1,  -3,  -1,   2,   0,   0 },
     {   13,  61, -52,   6,  -4,  -1,  -1,   0, -34, -27,  28,  10,  -5,   4,  -1,   2,  63, -29,   7, -16,   5,  -2,   0,   0, -35,   4,   7,   1,   5,  -3,   3,  -1,   5,  15,  -9,  -1,  -2,   0,   0,   0,   0,  -9,  -1,   5,  -1,   1,  -1,   0,   1,   2,   4,  -3,  -1,   0,   0,   0,   0,  -1,  -2,   0,   1,   0,   0,   0 },
     {   21,   5,  -8, -12,  13,  -3,   1,  -1,  33,  17, -14,  16, -24,   3,  -4,   1,   2, -23, -15, -14,  23,  -2,   2,   0,  -6, -14,  12,  40, -28,   4,  -3,   2,   7, -12,  33, -54,  20,   3,  -2,   1, -17,  22, -31,  29,   6, -17,   8,  -4,   5,  -9,  12,  -3, -20,  17,  -5,   2,  -2,   1,  -1,  -5,  12,  -6,  -1,   1 },
     {   21,  19, -12,   0, -11,   4,  -1,   0,  36,  28, -17,  -4,  10,  -6,  -1,  -2,   8, -15, -35,   8, -14,   8,   0,   2, -10, -30,   5,  29,  16,  -3,   0,   0,   4,  -4,  30,   7, -44,   8,  -3,   2,   4, -18,  16, -48,  33,  11,  -6,   3,  -8,  12, -10,  18,   9, -25,   9,  -3,  -1,  -1,   0,   1, -13,  11,   1,  -2 },
     {  -11, -36, -17,  32,   3,  -1,   1,   1,  20,   5,  48, -42,  -6,   1,   0,  -1,  -2,  -8, -26,   8,  21,  -8,   4,  -2, -39,  51, -10,  16, -23,   4,   0,  -1,  30, -29,  -8,  -1,   3,  10,  -6,   3,  -4,  -4,  23,  -5,   4,  -7,   2,  -1,  -1,   7, -11,  -4,   3,  -1,   2,  -2,  -1,  -1,   1,   6,  -3,   1,  -1,   0 },
     {    6,  13,  -3,   0,   6,  -8,  -1,  -2,   9,  17, -13,  -2, -10,   9,   1,   1,   5,  10, -17,   2,   9, -11,  -1,  -2, -13, -16, -26,  17,  -7,  13,   4,   2,   3,  -3,  16,  36,   0, -14,  -5,  -3,   3,  -6,  10,  -9, -59,  25,   3,   3,   9, -17,  18, -35,  45,  22, -22,   6,  -1,   2,  -1,   3,  11, -40,  14,   4 },
     {  -10, -57,  25, -17,   7,   3,   2,   1,  42, -30,  14,  16,   2,  -5,   0,  -1,  45,  19,  -9, -30,  -5,   2,   0,   0, -41, -16,   4,  10,  17,  -6,   3,   0,  -9,  44, -11,   0, -13,  -2,   3,  -2,   1, -15, -12,   9,   1,   8,  -4,   2,   0,   1,  15,  -4,  -1,  -4,   0,   0,  -1,   1,  -6,  -1,   3,  -1,   3,   0 },
     {  -21, -38, -11,  14,   2,   1,   2,   2, -31, -55,  -9,  19,   2,   7,   3,   3, -25, -40,  16,  29,   0,   1,   1,   1,  -7,  -7,  42,  20,  -8,   4,  -2,   0,  12,  16,  40, -10, -17,  -8,  -4,  -1,   1,  10,   4, -28, -11,   9,   3,   0,   1,   3,  -1,  -9,  15,   2,  -4,   1,   0,   3,   0,  -1,   0,  -6,   6,   1 },
     {   -9, -28, -49,  22,   7,   7,   0,   2,   9,  31,  30,   6, -24,  -4,  -5,   0, -16, -31,  21, -41,  19,   7,   2,   2,  45,  -3, -10,  10,  18, -21,   4,  -4, -38,  13,  -6,  11, -24,   7,   6,  -1,   6,   8,  -4,   1,  -2,  14, -10,   2,   4, -10,   4,  -6,   4,  -4,  -1,   3,   0,   1,   1,  -1,   4,  -2,   4,  -2 },
     {   -6,  23,   3, -17,  -1,   5,   0,   0, -22,  27,  -1,  14,   2,  -7,   0,  -3, -68,   3, -10, -11,   0,   8,   3,   2, -39,  26,  16,  24,   7,  -6,  -1,  -1,  22,  36, -29, -10, -22,   4,   1,  -1,  16, -32, -11,  16,   6,  11,  -6,   4,   9,  -5,  22,  -9,  -3,  -4,  -2,   1,   1,   0,  -5,  -4,   8,  -3,   3,   0 },
     {   -1,  -2,  -1,   4,  -2,  17,  -7,   1,  -2,  -6,  -6,  -1,   4, -21,  12,  -1,   1,  -1,  -7,   7,  -3,  23, -11,   2,  -1,  -4,  -3,   6,   4, -26,  12,   0,  -5,   6,   0,  20, -15,  32, -12,  -1,   6,  -7,  16,  -9,  -2, -37,   9,   2,  -7,   8,  -8,   3, -29,  61,  -3, -10,  12, -15,  19, -21,  41, -29, -29,  24 },
     {   -2,  -2, -12, -43,  33,   5,  -1,   1,   4,  -4,  15,  45, -29, -10,   3,  -2,  -5,   8, -31,  -5,  -3,  21,  -4,   3,   4,   1,  34, -21,  19, -15,  -3,   2,  31, -38, -20,  18,  -5,  -3,  12,  -3, -31,  26,  23,  -5,  -6,   8,  -6,  -1,  10,   2, -19,  -5,   6,  -2,   0,   3,  -2,  -3,   6,   6,  -1,  -2,   2,  -2 },
     {  -13, -17, -23, -14,   4,   3,   3,   1, -20, -28, -35, -19,  14,  10,   1,   2, -21, -32, -36, -12,  16,  11,   6,  -1, -20, -31, -30,   5,  27,  14,   1,   3, -15, -19,  -9,  25,  31,  11,   0,   2,   2,   2,  15,  37,  21,  -1,  -6,  -3,  10,   6,  15,  19,   1, -12,  -3,  -4,   1,   5,   6,   5,  -4,  -3,  -5,  -2 },
     {   -1,  -3,  -9,  20, -29,  22,  -6,   2,   4,   7,  11, -19,  44, -28,   7,  -4,   0,  -3,   1,   4, -34,  19,   1,  -2, -14,   6,  -5,  -8,  33,  -5,  -7,   6,  21, -14,  17,  -1,  -9, -17,  15,  -9, -31,  39, -39,  12, -10,  16,  -7,   2,  23, -27,  15,   7,  -6,   3,  -7,   5,  -3,   2,   5, -13,   9,  -8,   6,   0 },
     {    9,  26,  73, -29,   1,  -9,   0,  -3,  -9, -13,  53, -39,  -2,  -3,  -1,  -1, -12, -49, -10, -16,  21,   3,   5,  -1,   9, -10,   0,  10,  15,  -6,   1,  -1,  -9,   8,   9,   9,  -5,  -6,   1,  -1,  -5,  18,  -2,  -8,  -7,   1,   1,  -1,  -2,   1,  -8,  -2,   2,   4,   1,  -4,   0,   1,   0,   1,   2,   0,   0,   0 },
     {   -1,   2,  -5,   8,   0,  -7,   7,  -2,  -1,   3,  -9,  10,   0,  -1,  -1,  -1,  -3,   1, -12,  14,   3, -10,   9,  -3,  -9,  -2,  -8,  27,   4,  -4,   1,  -3,  -2,   8,   8,  36, -14, -21,   9,  -5,   2,  14,  11,  18, -45, -10,   0,  -5,  -7,  -5, -10, -18, -54,   8,  27,   4,  -1,   2,  -1,   1,  -4,  70,  -2, -17 },
     {    6,   4,  15,  -7,  -3,  -1,   4,  -2,  -5, -30, -21,  -5,  10, -12,   1,   1,  29,  -3, -22,  30,   6,   6,   6,  -1,  39,  31,  -5,  28, -26, -19,  -6,  -4, -17, -14, -23, -10, -35,  26,  12,   4,  -1,   3,  14,  20,  11,  25, -24,   2,  12, -19,  21, -27,  -3, -20,  -4,   5,  -4,  -2,   1,  -9,  14,  -3,  15,  -2 },
     {    2,   3,  13,  34, -32, -10,   4,  -1,  -3,  -2, -22, -12,   1,  12,  -4,  10,  -5,  15,  22, -29,  31, -19,   1,  -2,  -1, -41,  15,  11,  -8, -17,  18,  -6,  38,  11, -33,   3, -12,  12,  -5,  -3, -51,  14,  25,   2,   4,   2,  -3,   5,  20,  -1, -13,   1,  -3,  -6,  -4,  -1,  -3,  -9,   4,  -6,  -2,  -2,   5,   1 },
     {    0,  -3, -16, -17,  10,  -3,  -3,   4,  -1,  -9, -29, -17,   5,   8,   7,   0,  11,   5, -29,   6,  35, -11,  -2,   2,  29,  40,  -1,  14,  13, -10,   4,  -7,  12,  33, -17,   6, -11, -26,  -8,  -3, -16,   4, -44, -13, -10,   6,  23,   3,  -4,   0, -18,  46,  12,  14,  -8,  -2,   3,  -5,   8,   7, -18,  -1, -14,   2 },
     {   -5,  10,   3,   1,   1,   3,   1,   1, -10,  10,  -9,  -5, -10,   0,  -1,   4,  -8,  28,   0,   7,  15,  -1,   1,  -2, -56,  10,  14,  -2, -14, -14,  -1,  -7, -75,   3,   9,  -2,  -7,   9,  14,   4,  -5,  48,  -1,  -8,   3,  17,  -7,  -8,  13,   2, -30,  12,   5,  -4, -13,   8,   6, -15,   8,   1,  -7,  -4,  -3,   5 },
     {   -1,  -2,  -2, -15, -30,  48, -21,   7,  -2,   1, -11,   6,  31, -48,  15,  -1,   5,  -2,   7, -32,  17,  14,   5,  -4,   8,   4,   6,  27, -17,   0, -18,   7,  -7,  -4,  -7, -18,  24, -25,  12,   2,   4, -11,  29,  -1, -22,   1,   7,  -9, -14,  22, -25, -12,  12,  -1,  -1,   3,   2,  -3,  -4,  21, -14,  -1,   3,  -3 },
     {    5,   7,   7,  15,  20,  -2,  -5,  -3,   6,   9,   9,  15,  16,  -3, -13,   1,   6,  11,  10,  20,  21,   0,  -9,  -2,   6,  12,  14,  24,  24, -17, -13,  -1,   6,  11,  18,  27,  17, -20, -16,  -6,  12,  10,  14,  19,   1, -41, -22,  -5,   1,   6,   7,   2, -18, -47, -14,  -3,   0, -11,  -6, -14, -29, -44,   6,  20 },
     {   -1,  -4,  -4, -10, -21,  -6,  17,   2,   6,   6,  14,  16,  15,  21, -32,   2,   1,  -4,  -2,  -3, -15, -21,  26,   0,  11,   6,   4,   6,  11,  32, -19, -15,  -2,  -6,  -5,  -3, -11, -26,  -3,  28,   1,   1,  23,  11,   9,  32,  24, -25,   1,  21,  -4,   3, -21,   8, -44,   4,   8, -22,  12, -23,  17,  -1, -19,  24 },
     {   -5,  -5, -31, -56,  52,   3,   1,  -2,   4,  14,   6, -47,  41,   2,   2,  -3,   2,   9,  50,  16,   5, -17,   0,  -4,  -6, -25,   7,   9, -15,  -9,   2,  -2,  12,   0,  -1,   2, -14,  -1,  16,   8,  -2,   1,   4,   5,   6,   4,  -1,   3,  -3,   0,  -2,  -9,  -3,   3,   3,  -2,   1,  -1,   1,  -2,  10,   3,   1,   3 },
     {   -2,  -4,  -5, -27, -29,  43,  10,  -9,   2,   2,   8,  13,  13,  -9, -27,   6,   2,   0,  -5,  -2,  19, -29,  13,   3,  -1,  -2,  24,  -5, -18,  28,  15, -20,  -8, -17, -17,  29,  -1,   9, -39,   4,  19,  21, -24, -24,   3, -21,  -6,  24,  11, -17,  20,  -3,  -2,  -1,  14, -10, -16,  12,   1,  -6,   0,   1,   2,  -4 },
     {   -2,  -1,  -2,  -7,   0,  -1, -14,   5,  -1,   1,   7,   3,   3,  -4,  25, -10,  -3,  -5,  -2,   3,  -9,  -4, -21,   7,  -2,  -2,  -1,  -2,  -1,  -1,  30,  -5, -11,  -7,  -5,   3,  -8,   3, -26,   1,   1,  -3,   4,   6,   3,  -6,  38,  -2,  -7,   6,  -5,  11, -25,   3, -47,   7,  -1,  -1,   0,  -8,  -7,   1,  84, -30 },
     {   -2,  -4, -17, -26, -23,  -4,   5,   7,   0,  -1, -21, -32, -18,   8,  -9,  24,   4,  10, -12, -13,  -2,  14,   9,  -6,   4,  16,  -8, -24,  30,   1,  -7, -18,  20,  32,  25, -23,   2,   6, -11,  13,  26,  27,  16,   4,   0, -17,   1,  13,   1,  -2, -30, -38,   4,  -9,  16,   1, -14, -17,  -3, -11,  14,  -3,  28,   1 },
     {   -4,  -2, -11, -18,  -1,  -5,  -1,   9,  -1,   6,  14,  18,  -4, -32,   6,   3,  -9, -17,  -8,  19, -23, -13,   5,  15,   1,  -2, -41,  -2, -28,  25,  13,  -2,   0,  50,  -8,  -2,  15,   6,  -4,  -2, -41,  22,  24,  -8,   9,  -3,  -5,   9, -25, -31,  -2,   3,   0, -13,   6,   6,  23, -11,  -6, -15, -18,  -9,  -9,   7 },
     {   -1,   5,  -1,   1,   7,  18,  -4, -20,  -1,   8,   0,   0,  19,  26, -26,   9,   0,  12,   0,   7,  21,  35, -12, -10,   1,  16,   0,  -7,  17,   4, -20,  -1, -18,  14,   8,  -5,   0,   4, -22,   4, -54,   3,  19, -13,  11,  -8,  14,  -2, -16,  10,  33, -22,   6,  14,  -6,  -5,  24,  29, -29,   5,   8,  14,  -2, -28 },
   },
   {
     {   93, -57,  -1,  -2,   0,  -1,   0,   0, -59,  23,  13,  -2,   2,   0,   1,   0,  -2,  14,  -9,  -1,   0,   0,   0,   0,   1,  -3,  -2,   3,   0,   0,   0,   0,   0,   1,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   40,  35, -46,   5,  -7,   1,  -2,   0,  15, -73,  35,   7,   2,   2,   0,   1, -44,  35,  20, -17,   3,  -3,   1,  -1,   8,  10, -21,   1,   3,   0,   0,   0,  -4,  -3,   2,   6,  -1,   0,   0,   0,   2,   1,  -1,  -1,  -1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   19, -58,  51, -13,   8,  -3,   2,  -1,  75, -15, -21,  -7,  -1,  -1,   0,   0, -47,   5,   3,   9,   2,   1,   1,   0, -19,  26,  -1,  -3,  -2,  -1,   0,  -1,  -1,   2,  -8,   2,   0,   1,   0,   0,  -1,   1,   1,   2,   0,   0,   0,   0,  -1,   0,  -1,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   49,  34, -12, -17,  -3,  -4,  -1,  -2,   5,  -4, -48,  27,  -2,   5,  -1,   2,  15, -65,  48,   9,  -9,   4,  -2,   2, -17,  19,  19, -29,   6,  -1,   1,  -1,  -5,   6, -13,   2,   8,  -3,   1,   0,  -4,   3,   1,   3,  -2,  -2,   1,   0,   0,  -1,   0,   1,  -1,   1,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {   13, -20, -45,  51, -13,  10,  -4,   3,  41,   2,  40, -41,  -3,  -4,  -1,  -1,  28, -41, -16,  -2,  16,  -2,   3,   0, -36,   4,  25,  11,  -5,  -2,   0,  -1,  -7,  22,  -7, -13,  -1,   1,   1,  -1,  -3,  -2,  -5,   5,   3,   0,   0,   0,   0,   1,   1,  -1,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0 },
     {   -3,  -6,  33, -40,  13,  -5,   3,  -2, -12, -51,  31,  14,   6,  -1,   1,   1,  70,   6, -12, -12,  -8,  -2,  -1,  -1, -54,  24, -19,   4,   8,   0,   3,  -1,  -4,  14,  -3,  10,  -4,  -1,  -1,   0,   2,   0,  -3,  -2,  -3,   1,   0,   0,   2,  -1,   0,   1,   0,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {  -34, -37, -11,  28,   3,   5,   0,   2, -21,  19,  15,   7, -14,   2,  -2,   1,  -8, -12,  59, -45,   9,   0,   1,   0,  -3,  46, -37, -14,  20,  -9,   3,  -2,  12,  -8, -23,  25,  -4,  -3,   2,   0,   1,  -2,   8,   4,  -8,   2,   0,   0,   3,  -1,   1,  -3,  -1,   3,  -1,   0,   0,   0,   0,  -1,   1,   0,  -1,   0 },
     {  -16,   5,  19,  12, -17,   3,  -3,   1, -39, -23,  17, -40,  33,  -2,   6,  -1, -18,   2,   1,  52, -11,  -7,   0,  -1,   5,  38,  21, -40, -19,  10,  -2,   0,   2,   9, -40,  -3,  26,  -3,   0,   0,   8,  -8,   3,  19,  -9,  -5,   0,   0,   2,  -1,   1,  -4,  -3,   3,   1,   0,   1,  -1,   0,   1,   0,   1,  -2,  -1 },
     {  -16, -44,  -2,   9,   2,   5,   1,   2,   5, -29,   7,  26,  -2,   1,  -1,   0,  16,  11,  27, -12, -22,   5,  -3,   1,  32, -51,  32, -39,  19,   5,  -2,   2, -36,  37,  -1,  -5,  21, -15,   4,  -2,  -2,  17, -17,   7,  -7,  -4,   4,  -1,  -3,   3,  -3,   3,  -2,   3,   1,   0,   1,   0,   0,   1,  -1,   1,  -2,   0 },
     {  -12,  -4, -32, -39,  66, -17,   9,  -2,  -2,  13,  42,  23, -40,  -5,  -1,  -2, -29, -17, -22,  17, -15,  25,  -3,   3,   5,  20,  25,  -6,   7,  -4,  -6,   0,   9,   7, -15, -17,   1,  -3,   2,   1,   2, -10,  -3,  11,   6,   0,   1,  -1,   2,   2,   4,  -1,  -5,  -1,   0,   0,   0,  -1,  -1,   0,   1,   1,   0,   0 },
     {   -8,  -2, -14,  20, -19,  11,  -3,   3,  -9,   0, -27,  37,  -9,   6,  -4,   2, -12,  46, -11, -14,  -2,   0,   0,  -1, -61,  -8,  46, -14,  11,   0,   0,   1,  56,   4, -14, -10,   3,  -5,   0,   0,   4, -21,   0,   5,   1,   1,   1,   0,   1,   3,   6,  -1,   0,  -1,   0,  -1,   0,   0,  -1,  -1,   0,   0,   1,   0 },
     {  -15,  -8,  26,   6,  -3,   0,  -1,  -1, -38, -34,  -2,  10,  -9,  12,   0,   4, -23, -27,  23, -12,  37,  -7,   2,  -1,   4,  12,  30,  36, -25, -13,   3,  -3,   4,  21,  21, -48, -20,  18,  -3,   1,  -2,  12, -25, -11,  28,   0,  -3,   0,   3,   0,  -2,  11,   0,  -7,   0,   0,   1,   1,  -1,  -1,  -1,  -1,   2,   0 },
     {  -12, -34, -26,  18,  10,   3,   0,   3,   0, -20,   0,  13,   7,  -2,   0,   1,  13,  -7,  16,  39, -41,   2,   0,  -2,  -8,  11, -10,  -5, -20,  24,  -6,   3,  26, -49,  52, -17,  12,   3, -12,   5, -12,  16,  14, -29,  12, -12,   5,   1,  -3,   7, -11,   0,   7,  -4,   4,  -2,  -2,   3,  -1,   3,   0,  -3,   1,  -1 },
     {   -7,  -1,   1,  -8, -19,  18,  -3,   2, -10,   0,  10,  -1,  37, -26,   2,  -2, -29, -26,  -3,   8, -34,  18,   8,   3,  -6,  -1,   9,  13,  40,   1, -11,  -1,   6,  30,  17,  20, -45, -26,   5,  -1,   6, -10,  -2, -39,  -8,  35,   3,   0,   2,   2,  -4,   4,  25,  -8,  -9,  -2,   0,   0,   0,   1,  -5,  -7,   5,   3 },
     {    6,   9,  17,  49,  31, -51,  -4,  -4, -13, -25, -19, -32, -36,  35,  17,   2,   5,   7,   1,  21,   1,  28, -23,  -1,  -7,  -9,   2,  -5,  24, -23,  -2,   8,   3,   4,  11,  18, -15,  -7,  12,  -4,   2,  -2,  11, -16, -13,  12,  -2,  -2,  -2,   2,  -3,  -6,  13,   2,  -4,   0,   1,   2,  -1,   3,   0,  -5,   0,   1 },
     {   -1,   6, -11, -16,  44, -21,   0,  -4,   1,  16,  -5, -42,  24, -18,  17,  -1,  -4,  18,  55,  -3, -16, -12,   2,  -8, -28, -37, -14,  17, -23,  21,   1,   2,  27,  36,  -9,  -4,   0,   8,  -9,   1, -10,  -4, -19,   4,   3,  -3,  -1,   3,   1,  -2,   1,   7,   0,   1,   1,  -1,   0,   2,   0,   1,  -4,  -1,   1,   0 },
     {   -9,   1, -22,  17,   1,   2,   2,  -1,  -8,  19, -27,  14,  -9,   3,   0,   1, -12,  23,  -5,  20, -17,   3,  -3,   0, -31,  30, -27,   5,  -7,   8,   1,  -2, -64,  35,  27, -11,   4,   6,  -7,   0,  48,  -7, -21,   0,   3,  -5,  -3,   3,   8, -23,   9,   6,  -5,   2,   1,   1,   4,  -1,   5,  -5,   1,   2,  -1,   0 },
     {   19,  32,  39,  13, -19,  -4,  -6,  -2,  21,  37,  40,  -4, -21,  -1,  -4,  -2,  16,  28,  23, -14, -31,  13,  -8,  -2,   8,  20,   3, -30,  -2,   2,  -6,   1,   1,  -4,   0, -43,  -3,   0,   3,   3,  -5,  -6, -24, -24,  26,  11,   3,   0,  -2,   0,  -9,  10,  17, -10,  -3,  -1,  -1,  -1,  -3,   3,  -2,  -6,   5,   2 },
     {    1, -22, -30,   3,   6,  -3,   2,   0,  14, -16, -35,   8,  12,  -5,   2,   1,  43,  24,   1,  14,  -5,   0,   0,  -4,  49,  27,  -1,  14, -19, -13,  -5,  -3,   9,   6, -35, -15, -35,   6,   8,  -1,  -4, -40,  -9,   1,   7,  33,  -4,   2,  -3,  -2,  15,   7,  13,  -4, -12,   3,  -2,   0,  -1,  -4,  -4,  -5,   4,   3 },
     {   -5,  -6, -14, -21,   2,  -2,  12,  -1,   6,  13,  18,   2,  -3,  40, -25,   0,  -1,   6,   8,  29,  24, -57,  -5,  10,  -4,  -8,   8, -38,  -2,  -8,  45, -10,  -1,   7,  11,  23, -21,  21, -15, -12,   7,  -7,  -4, -27,  -2,  13, -10,   8,  -2,   6,  -9,   1,  24,  -7,  -2,  -1,  -2,   0,  -1,   6,  -7,  -8,   5,   1 },
     {    4,   4,   6,  -4,  24,  60, -60,   6,  -4,  -8,  -7, -12, -47, -30,  26,  15,   7,   6,   5,  22,  25, -18,  28, -23,   0,  -5,  -7, -24,   8,   1, -16,  -2,   0,   1,  11,   4, -11,   0,   2,   6,   2,  -4,  -2,  -2,  -2,   6,  -2,  -4,  -2,  -1,  -3,  -1,   3,  -1,   1,   1,   1,   1,   1,   4,   1,  -2,   1,   0 },
     {    5,   5,  10,  13,   8, -17,   7,  -5,   0,  -2,  -1,   5, -18,  27, -20,   9,   0, -10, -11, -11,   4, -32,  37,  -4,  13,  10,  -2,  12,  -7,  50, -21, -14,   8,  11,   4,  -3,  14, -44, -24,  21,  -2, -11, -13,  -1, -32,  11,  36,  -3,   0,  -3,  10,  -6,  18,  20, -17, -10,  -2,   2,  -3,   3,  -1, -15,   2,   6 },
     {   -2,   1, -18,  -7,   6,  -3,   5,   0,   0,   6, -16, -18,  16,  -2,  -2,   2,   3,  15, -15,   7,  13,  -5,  -2,  -1,   0,  27,  -2, -16,  30, -20,   8,  -5,  13,  -9,  17,  -5,  -4, -23,   7,  -6, -38,  47, -57,  12,  -7,   5,  15,  -8,  15,  -8, -15,  40, -17,  19,  -8,   0,   5,  -9,  11,   0,  -8,   4, -11,   1 },
     {   -1,   7,  -7,   9,  20,  -6,   3,  -6,  -4,   7,  -8,   9,   6, -19,   7,   4,  -1,  22,  -7, -22,   4,   2,  12, -11, -10,  19,  38,   2, -32,  10,  -9,  -7, -36,  -3, -26,  20,  -3,  13,  -5,   7, -27,  46,  28, -35,  -8,  14,  -8,   1,  23, -11, -33,  -8,  22,  -2,  -4,   6,   3, -12,   5,  15,   2,  -6,   0,  -1 },
     {   -1,   3,   2,   9, -20,  12,  -8,   3,  -1,   2,   2,  23, -22,  16,  -6,  -1,  -8,  -2,  -9,  29, -14,  -4, -13,   0,   1,  -4, -42,  13,  -1,   1,  10,   5,   8,  54, -16,  -9,   5,  -3,   1,  -7, -70,   0,  27,  -2,   7,  -8,   1,   1,  38,   5,  -7,  -5,  -3,   0,   5,   0,   1, -11,   2,   0,   0,   3,   1,  -1 },
     {   -5,  -6,  -8,  -1,   6,  -4,   1,   0, -11, -16, -22, -24,  -1,  14,  -8,  11, -11, -16, -29, -26,  -7,   5,  19,   0,  -7, -11, -21, -24,  24,  24,  -5,  -5, -11, -10, -18,  -1,  18,  13, -10,   4,  -8,  -2,  -1,   6,  55,   8, -15,   6,   9,   4,  13,  33,  23, -45,  -4,   1,  -1,  -1,   1,   3, -23, -18,  19,   1 },
     {   -5,   3,   5,   2,   8,   4,   1,  14,  -8,   4,   7,   1,  11,   1,   4, -15, -11,   6,  17,   5,  -3,  -8,   3,   8, -30, -13,   8,  -3,   0, -16,   2,  -2, -49, -35,  13, -20, -15,   2,  -4,  11, -42, -20,  17,  35,  -4,  37,  15,  -3,   8,  27,  30,  -2,  -7,  11, -34,  -5,   5,   8,  -9,  -7,   5, -14,  -2,  15 },
     {   -9, -13, -20, -42, -22,  17,  -6,   2,  -6,  -6, -11, -31, -14,  20,  -4,  16,  -6,  -6,   0, -17,  -1,  25, -17,   0,  -9,  -9, -12,  -1, -12,  -8,   6,  10,   1,  -5,  -3, -30,  33,   7,  18, -10,   4,  -5,  -3, -15, -20,  20,   1, -10,   3, -13,  -9, -12,  43,  35, -15,   9,   3,   5,   6,  14,  26, -22, -16,  -1 },
     {   -1,   0, -12, -20,  14,   5,   3,  14,   1,   5, -14, -21,  20,   1,  -4, -15,  10,  20,  11,  -8,  33,  -5, -11,  18,   5,  19,  10,   0,  27, -14, -17,   5, -13,  14,  15, -28,  16, -33,  -1,   8,   6, -27,  39, -16,  11, -32,  25,  -3,  11,  22, -31,  -5,   3, -15,  13, -21,  -8,  -6,  -8,  15, -11,  -4,  10,   3 },
     {   -4,  -5,  -7,  -8,  -8, -28, -51,  47,   5,   8,   5,  18,  15,  28,  61, -36,  -6, -10, -17, -17,  -4, -24, -16, -15,   4,  11,   3,  -1,  -6,  15,  -3,  32,  -1,   1,  -2,   6,   3,   2,   4, -13,   1,   4, -11,  -6,   3,  -3,   7,   0,  -6,   0,   4,   4,   6,   2,  -4,  -9,  -4,   1,  -2,  -1,  -3,  -7,   3,   1 },
     {   -1,  -1,   0,   1,  -5,   9,  25, -32,  -1,  -2,  -1,   8,  -6, -32,  29,   0,   0,  -2,  -6,  -2,  44,  17, -60,  18,   0,   3,  -4, -26, -26,  38,  -1,  14,   2,   4,  17,  24,  -2, -20,  -5,  -1, -14, -11, -19,  -2,  12,   3,   6,   1,   6,   9,  23,  -5,  12,  -5,  -5,  -3,  -1,  -2,  -8,  -2,   3, -15,   1,   6 },
     {    3,   3,   5,   8,   7,   5, -15,  -3,  -2,  -7,  -8, -10,   1, -13,  17,   2,   0,  -3, -13, -20, -15,  15,  -2,  19,  13,   9,  11,  -3,  21,   4,  33, -24,  -3,   0,   0,  -4,  -6,  30, -62,  -8, -14, -25,  -2, -12,  -7, -22,  12,  45,   1,   6,   3,  11,   3,  36,  13, -17,   8,   4,   2,   4,  13,   3, -26,  -9 },
   }
 },
 {  //11
   {
     {  120, -36,  -7,  -5,  -2,  -2,  -1,  -1, -24,  -4,   3,   1,   1,   0,   0,   0,  -7,   3,   1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -22,  -6,   6,   0,   2,   0,   1,   0,-114,  33,  10,   3,   3,   1,   1,   0,  24,  21,  -7,  -1,  -2,  -1,  -1,   0,  18,  -7,  -5,   0,  -1,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {   11,  -3,  -2,  -1,   0,   0,   0,   0,  23,  30, -12,  -1,  -3,  -1,  -1,   0, 104, -28, -20,  -1,  -4,   0,  -2,   0, -13, -42,   9,   4,   2,   1,   1,   1, -19,   1,  13,   0,   1,   0,   0,   0,  -5,   6,   2,  -2,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   14,  23,  -9,  -2,  -2,  -1,  -1,   0,  25,   2,  -6,  -2,  -1,  -1,   0,   0,  12,  44, -14,  -2,  -3,  -1,  -1,   0,  90, -13, -34,   1,  -5,   0,  -2,   0,   6, -52,  -1,   7,   1,   2,   0,   1, -11, -10,  13,   3,   1,   0,   0,   0,  -3,   1,   4,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   25,  97, -27,  -6,  -7,  -3,  -3,  -1, -16,  39, -29,   2,  -3,   0,  -1,   0, -31, -34,  -7,  11,   2,   3,   1,   1,  -8, -17,  13,   5,   0,   1,   0,   0,   3,  -1,   8,  -3,   0,   0,   0,   0,  -2,  -2,   1,  -1,   1,   0,   0,   0,  -1,  -1,   1,   1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0 },
     {   -8,  -2,   1,   2,   0,   0,   0,   0, -13, -37,  10,   3,   3,   1,   1,   0, -28,   0,   4,   3,   1,   1,   0,   0,   9, -53,   8,   7,   1,   2,   0,   1, -64, -10,  48,   1,   4,   0,   2,   0, -15,  54,  13, -14,  -1,  -3,   0,  -1,   4,  19, -16,  -8,   0,  -1,   0,   0,   1,   1,  -9,   1,   2,   0,   0,   0 },
     {    5,  16,  -5,  -1,  -1,   0,   0,   0,   7,  -8,   2,  -2,   1,  -1,   0,   0,   2,  39, -11,  -4,  -3,  -2,  -1,  -1,  19, -20,  -2,   0,   1,   0,   1,   0, -21,  55,   5, -11,   1,  -4,   0,  -1,  42,  11, -61,   1,  -3,   0,  -1,   0,   4, -55,  -3,  25,   0,   5,   0,   1,   0,  -3,  23,   2,  -3,   0,   0,   0 },
     {   -2,  -8,   1,  -2,   3,   0,   0,   0,  16,  54,  -8,  -9,  -1,  -3,  -1,  -1, -23,  62, -33,  -7,  -1,  -2,   0,   0, -61,  -5, -35,  15,   4,   4,   1,   1, -27, -18,   5,  23,   0,   3,   0,   1,  10,   0,  17,   3,  -7,   0,  -1,   0,   9,  -1,   3,  -6,  -2,   1,   0,   0,   0,  -4,   0,  -1,   2,   0,   0,   0 },
     {  -10, -56,  17,   2,   5,   1,   2,   0,  18,  72, -17,  -5,  -5,  -3,  -2,  -1, -44, -28, -11,  12,   1,   4,   1,   1,  41, -29,  24,   2,  -1,   1,  -1,   0,  13,  11,   9, -15,   0,  -2,   0,  -1, -10,   3, -12,  -3,   5,   0,   1,   0,  -1,  -3,  -2,   7,   1,   0,   0,   0,   0,   0,   4,   0,  -2,   0,   0,   0 },
     {   10,  21,  13,  -6,  -2,  -2,  -1,  -1,  16,  36,  18, -10,  -4,  -3,  -1,  -1,  15,  31,  22, -11,  -4,  -4,  -1,  -1,  14,  43,  14, -15,  -3,  -3,  -1,  -2,  14,  32,  10, -16,  -3,  -3,  -1,  -1,   5,  50,   2, -25,  -2,  -4,   0,  -1,  14,  33, -39, -23,   2,  -2,   0,   0,   3,  -8, -33,   2,   5,   1,   1,   0 },
     {   11,  24,  78, -21,  -7,  -6,  -1,  -2,   2,   3,  58, -29,  -4,  -3,   0,  -1,  -8, -33, -15, -23,   8,   2,   3,   1,   3, -20, -35,   4,  15,   2,   2,   1,   7,   5, -13,  21,   6,  -2,   0,   0,  -1,   4,  -2,  10,  -6,  -2,  -1,  -1,  -5,  -3,   1,   2,  -3,   2,   0,   0,  -2,   1,   3,   0,  -1,   1,   0,   0 },
     {   -1,  -7,  -3,   4,  -1,   0,   0,   0,   0, -18,   4,   5,  -2,   1,   0,   0,  -2,  -4, -12,   5,  -3,   2,   0,   0,  -7, -31,  -1,  22,  -3,   5,   0,   1,  71, -12,  41,  10, -12,   1,  -4,   0,  66,   1,  21, -30, -11,  -2,  -3,   0,   6, -14, -15, -23,   8,   1,   2,   1,  -8,  -6,  -6,   2,  10,   0,   1,   0 },
     {    1,   7,   3,  -3,  -2,  -1,   0,   0,   5,  -2,  20,   1,   0,  -2,   0,  -1,  -3,  26,  -2,  -2,  -3,  -2,  -2,  -1,   6, -26,  26,   4,  -1,   1,   0,   1,  -5,  37,   8, -15,  -3,  -2,  -1,  -1,  11, -39,   8,   7,  -2,   4,   0,   0, -13,  32,  62, -26,  -5,  -4,  -1,  -1,  -1,  40, -37, -37,   8,  -3,   1,  -1 },
     {    4,   7,  52, -19,   4,  -4,   0,   0,  -5, -21, -27, -13,   8,   1,   1,   1,   5,   2, -59,  36,   3,   4,   0,   1,   5,  44,  18,  30, -20,  -3,  -3,  -1, -19,  -3,  25, -27, -18,   4,  -2,   1,  -3, -19,   2, -21,  13,   7,   1,   2,  10,   0,   0,   5,   9,  -3,  -1,   0,  -1,   0,   0,   3,  -3,  -4,   0,   0 },
     {    0,   7,  -5,   1,  -3,   0,  -1,  -1,   2, -12,  16,  -1,   1,   1,   1,   1,   0,  36, -15,  -8,   0,   0,  -1,   0, -28, -28,   5,  12,   2,   2,   1,   1,  54,  -1,  17, -24,  -2,  -3,   0,  -1, -62,   1, -29, -12,  22,  -2,   4,   0, -34,  19, -12,  33,   6,  -4,   2,   0,  11,   6,   8,  13, -16,  -2,  -1,   0 },
     {    6,  10,  35,  11, -12,   0,  -2,   0,   5,  15,  28,   4, -20,  -1,  -1,   1,   1,  15,  49,   1, -19,   2,  -2,   0, -17, -12,  37, -25, -16,   5,  -2,   1, -16, -56,   8, -33,   4,   9,   0,   4,  19, -30,  -5, -12,  19,   5,   0,   2,  14, -21,   1,  10,  14,  -2,  -1,  -1,   2,  -2,  13,   7,  -1,  -5,   0,   0 },
     {   -3,  -3, -15, -49,  44, -12,   5,   0,  -1,   0,  -8, -57,  56, -14,   5,   0,   2,   4,  20, -31,  33, -15,   2,  -2,  -1,  -4,  29,  -6,  13, -11,   1,  -1,   0, -19,  15,   2,   2,  -6,   3,   1,   8, -16,  -2,   2,  -2,  -2,   2,   1,   3,  -5,  -3,   5,  -2,   0,   0,   0,   4,  -1,   0,   4,  -1,   1,   0,   0 },
     {    4,   5,  61, -18,  -5,  -2,   1,  -2,  -4, -15, -80,  20,   6,   5,   4,   2,   2,  27,  28,  -1,  -6,  -1,  -2,  -1,  -6, -30,   9, -23,   8,  -1,   3,   0,  17,  12, -23,  21,  10,   0,  -1,  -2,  -2,  10,  12,   4,  -8,  -2,  -3,   0,  -7,   4,  -5,  -1,   0,   1,   0,   1,   4,   5,  -1,   0,  -3,  -2,   0,   0 },
     {    0,   3, -12,  14,  -9,   2,   0,   3,   1, -11,  22, -29,   5,  -5,   5,  -3,  -6,  15, -38,  28,  -5,   6,  -1,   2,   4, -24,  43, -32,  -2,  -1,   4,  -1,  -8,  19, -48,  15,   8,   3,   3,   0,  11, -24,  34,  -1,  -4,  -1,  -5,   1, -14,  16, -35,  -6,   9,  -3,   1,  -2,  10, -28,  10,  26,  -2,   2,  -2,   0 },
     {    0,   1,   2,  -7,  -1,   1,   2,  -2,   1,  11,   5,   0,   1,   2,  -1,  -1,   3,   4,  11,   1,   4,  -3,  -4,  -1,   1,  19,  -1, -18, -10,   1,  -1,   1, -10,  16,  27,   7,  -1,   5,  -3,   0, -24,  -2,  31, -20, -14,  -2,  -4,   2, -81, -49,   3, -23,  16,   7,   6,   5, -36,   6,  17,   7,  18,   1,   1,   0 },
     {   -4,  -5, -12, -61, -12,  15,   3,   5,   3,   5,  15, -11,   6,   8,   1,   2,   4,   3,  33,  77,   6, -15, -11,  -3,  -8, -18, -30,  17, -20,  -8,   1,   1,   3,   2, -28, -17, -11,  14,   9,   4,   6,  20,   1,  -2,   7,   6,   0,   0,  -5,   3,   7,   1,   3,  -4,  -2,  -1,   0,   0,   1,  -3,   0,  -2,   0,   0 },
     {   -6,  -9, -16, -26, -12,   5,   2,   3,  -8, -12, -26, -27, -21,   8,   3,   5,  -8, -15, -23, -42, -15,   8,   5,   3,  -6, -12, -37, -38,  -9,  16,   6,   3,  -5,   0, -18, -43,  10,  17,   5,   3,   6,  -4, -12, -24,  23,  15,   2,   2,   1,   3,  -4, -18,  33,   5,   2,  -2,  -8,   5, -23,   6,  24,  -2,   1,  -2 },
     {    0,  -1,   2,  -1,   6,   0,   0,   3,   0,   1,  -3,   4,   0,   3,  -2,  -3,   0,  -2, -23,  -6,  -1,   9,   1,  -1,  -3,  14,  27,   6,   1,  -1,  -4,  -1,  -3, -44, -29,  -7,  13,  -6,   2,   0,  50,  50, -28,   9,   2,  -8,   7,  -1, -61,  29,  17,   9,  -8,  -1,   0,  -4, -23,  12,   4,  -5,  -3,  -2,   1,  -1 },
     {    7,  10,  16,  40,  55,   7,  -5,  -1,   4,   5,  11,  41,  56,   7, -15,  -9,  -4,  -6, -10,   6,  18, -21,  -7,  -2,  -6, -16, -21, -15, -21, -10,   2,   6,  -2,  -5, -27, -20, -12,  -1,   8,   5,  -3,   0, -14, -20,  -8,   8,   8,   1,   1,  -2,  -6, -15,  17,   6,   1,  -6,   4,   6, -12,   3,  20,  -2,   2,  -2 },
     {   -4,  -3,  -9, -73,   4,  -1,   9,   5,   6,   4,  25,  76,  11, -12,  -5,  -6,  -5,  -3, -30,  -5, -26,  -1,   0,   2,   1,   8,  12, -34,   9,   6,   2,   4,  -1,  -4,  13,  11,  10,  -5,  -4,  -3,  -1, -10,  -1,   2,  -5,  -4,  -3,  -1,  10,   5,  -7,   5,   0,   6,   2,   3,   5,   0,   6,   1,  -2,   1,  -3,   1 },
     {   -1,  -3,  -1,   6,  -7,   1,  -2,  -2,   2,  12,  -5,  11,   9,   2,   3,  -2,   0, -12,  22,  -2,   4,   2,  -1,  -1,  10,  10, -31,   4,  22,  -4,   5,  -1, -24,  15,  16,  22, -11,  -1,  -4,  -1,  37, -30,   3, -33,  18,   2,   0,  -4, -34,  37, -14,  39,  32, -16,   1,  -4,  31,  19,  -2,  21, -22, -20,  -3,  -2 },
     {    1,   1,  -3,   1,   0,   2, -16,  13,   1,   1,   2, -12,   7,   6, -15,  -1,   1,   0,  -8, -17,  -2,  -1,   0,   2,   2,   5,   4, -26, -22,  14,   9,   1,  12,  11,  -2,   3, -19,   4,   5,   5,  -1,  37,  34, -30,  -8,   1,   2,   3,  22,  10,  55,  18,  13,   3,  -6,  -9,  22,   5,  63, -11, -13, -14, -14,  -4 },
     {   -2,  -5,   2,   4, -68,  10,   0,   2,   1,   1,  -2,  14,   5,  14,  -4,   2,   1,   3, -17,  -1,  83, -15, -11,  -1,  -3,   2,   3, -35,  12, -27,   1,   3,   2, -10,  17, -11,  -8, -11,  18,   6,   7,  -1,  -2,  13,   1,  -2,   8,  -3,   5,   0,  -1,   1,   0,   0,  -2,  -3,   4,  -1,   1,  -6,  10,   7,  -1,   0 },
     {    2,   2,   5,  14,  11,  -7,   6,  -7,   1,   0,  -3,   2,  22,  -9,  -1,  -4,   0,  -5,  -5,   3,  -2, -11,  -5,   0,  -2,   1, -24, -29, -31,  10,   7,   5,   8,  16,  26, -19,  29,  26,  -2,   2,  17,   5,  27,  34,  47, -25, -20,  -2,  -6,   1,   3,  20, -29, -41,  -7,   0, -15, -17,  -6,  -6, -21,   9,  12,   5 },
     {    1,   4,   1,  -9,  21,   6,  -2, -12,   1,   0,   2,  23, -17,  -5,  -5,   7,  -1,   7,   7, -23,  12,   9,  12,   7,   3, -12,   1,  17,  20, -22, -21,  -3,  -4,  17, -12, -19, -39, -16,  10,   9,   6,   0,  28,  16,  28,  35,  14,   5,   8,  12,   6,  -1,   8, -20,  -2,   9, -56, -27,  17,  26, -13,  -2,  -8,   7 },
     {    1,   2,  -2,   1,   4,  18,  -2,  -6,   3,   4,   9,   6,  23,  33,  16,  -9,   4,   7,  -1,  15,  27,  44,   4,  -4,   2,   4,   8,   7,  42,  30, -15,  -9,   5,   0, -11,   1,  37,   4, -27,  -6,  -9,  13,   4,  -9,  23, -17, -18,  12,  23, -17,  14,  -9,  29, -24,   2,  12, -11,  10,   0,   4,  11, -30,   1,   3 },
     {    1,   1,   2,  10,   3, -42,  21,  16,   0,  -1,  -1,  16, -14, -64,  16,  23,   2,   3,  -1,  13,  26, -29,  30,  -7,   1,   1,  -4,   5,  38,  21,  17, -23,   0,  -2, -11, -20,   2,  32,  10,  -5,   0,   6,  -1, -17,   0,  -6, -16,   9,  -3,   3,  11, -13,   1,   1,   0,   8,   7,  -4,  14,   4,   9,  -4,  -1,  -3 },
   },
   {
     {   80, -46,   1,  -4,   0,  -1,   0,   0, -75,  31,   9,   1,   2,   0,   1,   0,  22,  10, -16,   0,  -2,   0,  -1,   0,  -3, -15,   4,   5,   0,   1,   0,   0,   2,   5,   4,  -4,  -1,   0,   0,   0,  -2,  -1,  -3,   0,   1,   0,   0,   0,   1,   1,   1,   1,  -1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 },
     {   29,  20, -20,  -1,  -4,   0,  -1,   0,   3, -59,  32,   1,   5,   0,   2,   0, -42,  70, -11, -11,   0,  -3,   1,  -1,  26, -25, -24,  18,  -3,   4,  -1,   1,  -6,  -5,  27,  -5,  -3,   0,  -1,   0,   2,   5, -10,  -6,   4,   0,   1,   0,  -3,   0,   2,   5,   0,  -1,   0,   0,   1,   0,  -1,  -1,  -1,   1,   0,   0 },
     {   63, -11,   6,  -4,  -2,  -2,  -1,  -1,  20, -29, -16,   2,   1,   2,   0,   1, -32, -15,  48,  -4,   3,  -2,   1,  -1, -19,  58, -28, -19,   5,  -4,   2,  -2,   3, -16, -22,  29,  -2,   3,  -1,   1,   2,  -1,  20,  -4, -10,   2,  -2,   0,  -1,  -1,  -4,  -6,   6,   2,   0,   1,  -1,   1,   1,   2,   1,  -3,   0,   0 },
     {   35,  -2, -19,   6,   0,   1,  -1,   0,  34, -19,  18, -10,  -3,  -3,  -1,  -1, -19, -17, -17,  14,   4,   4,   1,   1,   1,  -8,  58, -14,  -6,  -3,  -2,  -1, -22,  40, -40, -28,  20,  -4,   5,  -2,   0,  -6, -14,  45,  -6,  -2,  -1,  -1,   2,  -2,  15, -13, -15,   6,  -1,   2,   0,  -2,  -2,  -3,  10,   1,  -2,   1 },
     {   34,   2,   0, -12,  -3,  -1,   0,   0,  39,   0, -24,  17,  -3,   1,  -2,   1,   4, -36,  12, -13,   0,   2,   1,   1, -18,  -7,   7,  26,   2,   1,  -1,   1,   8,  -4,  45, -40, -12,   2,  -2,   0, -20,  20, -36, -13,  39,  -7,   6,  -2,  -1,   4,  -3,  38, -19,  -8,   0,  -2,   0,  -2,   3, -12,  -8,  12,   0,   1 },
     {   32,  15, -14,  11,  -6,  -2,  -3,  -1,  38,   5, -12, -18,   0,   2,   0,   0,  11, -16, -24,  25,  -6,   1,  -2,   1,  -3, -41,  19,  -8,   3,   6,   1,   2,  -9,   6,  17,  25,  -6,  -5,  -2,  -1,   7, -12,  29, -47, -13,  12,  -1,   5, -12,  10, -18,  -4,  48, -12,   0,  -4,  -4,   7,  -5,  23, -24,  -9,   8,  -3 },
     {   20,  16,  -9,  -6,   0,   2,   0,   0,  27,  23, -22,  12, -10,  -3,  -3,  -1,  10,  -6, -22,  -3,   2,   4,   1,   1,  -3, -21, -20,  31,  -7,   1,  -1,   1,  -2, -27,  23,  13,  -2,   8,   0,   2,   1,   9,  23,  19, -26,  -7,  -4,  -2,   5, -18,  17, -49, -22,  29,   1,   8,   1,  -3,   5, -15,  54, -11, -14,   2 },
     {   11,  78, -28,  -4,  -7,  -1,  -3,  -1, -41, -35,  -7,  19,   1,   5,   1,   2,  54, -18,  30, -11,  -5,  -2,  -2,  -1, -16,  -9,  -1, -13,  10,  -2,   3,  -1,  -7,  22,  -3,   9,   1,  -3,   1,  -1,   3,  -3, -10,   4,  -7,   2,  -1,   1,   2,  -4,   9,  -1,  -1,   2,  -1,   1,   0,   0,  -1,  -5,   3,  -1,   0,   0 },
     {   14,  28, -11,  -4,  -2,  -4,   0,  -2,  19,  37, -16, -12,   5,  -3,   0,  -2,   6,  19, -32,  -3,   1,  -4,   1,  -1,  -6,  -5, -37,   0,  12,   1,   3,   0, -13, -14, -32,  32,  10,  -3,   3,  -1,  -5, -10,   3,  41,   7,  -2,   2,  -1,   1,  10,  12,  37, -29, -19,  -2,  -4,  -2,  -6,   1,  -9, -46,  23,  10,   1 },
     {   -9, -33, -11,  -3,   2,   1,   2,   1,  28, -12,  46,   5,   2,  -3,  -1,  -1,  30,   3, -17, -38,   4,  -5,   2,  -2, -57,   9, -17,  32,  18,  -1,   5,   0,   9,  34,  -6,  13, -31,  -1,  -5,   0,   7, -30,  -1,  -5,   0,  16,  -2,   2,   5,  -1,  19,  -7,   8,  -6,  -2,   0,  -2,   0,  -5,  -5,   2,  -3,   3,   0 },
     {   -1,  -1,  -2,  11, -18,   9,  -1,   2,  -1,  -5,   7, -11,  25, -12,   3,  -3,  -2,   0,  -5,  18, -27,  13,  -3,   4,   7,  -6,  14, -16,  26, -13,   2,  -3,  -7,   9, -11,  22, -36,  13,  -1,   2,  12, -11,  23, -27,  29,  -8,  -4,  -2, -11,  13, -21,  20, -40,   4,  14,   0,  14, -21,  24, -34,  38,  13, -25,   8 },
     {    5,  34, -30,  -1,   0,   2,  -1,   0, -20,   6,  16,   8,  -6,  -2,  -2,   0, -15, -17, -20,  -3,   6,   5,   2,   0,   4,  33,  16,  28, -15,  -3,  -5,   1,  25, -32, -15, -32, -17,  15,  -4,   1,   9, -28,  39,  -1,  39,  -3,   4,   0,  -4,  15,  -3, -23,   0, -30,   5,   3,  -8,   6, -18,  18,  -1,  15,   3,  -6 },
     {  -21, -30, -21,  12,   9,   4,   3,   1, -33, -44, -26,  19,  12,   6,   3,   2, -32, -35, -11,  24,  14,   3,   3,   1, -21, -16,  10,  32,  11,   3,   1,   0,  -5,   4,  24,  32,   5,  -5,  -1,  -2,   7,  16,  27,  22,  -8,  -7,  -3,  -2,   7,  15,  17,   6, -17, -13,  -1,  -3,   0,   4,   0,  -6, -23,   0,   5,   0 },
     {    0, -30,   8,  -6,   6,   1,   0,   1,  19, -31,  29,  -5,   3,  -5,   2,   0,  53,  -7,  16, -23,   6,  -2,   1,  -1,  12, -26,  -4, -24,   7,  -3,   1,   0, -12, -32,   5,  -1,  33,   2,   5,   0,   2,  20,  33,  15,  16, -27,   2,  -4, -11,  14, -20, -16, -20, -19,   7,   0,  -5,   8, -12,  20,  -1,  23,   0,  -7 },
     {    0,  -1,   3,   6,  -3,  -4,   0,   0,   0,   2,   4,   6, -11,   9,   0,   2,  -3,   2,   6,   8,  -6,  -5,  -1,  -1,   0,   5,   8,   4, -15,   9,  -1,   1,   6,  13,   8,   4, -15,  -9,  -1,  -1,   1,   3,  -4, -15, -36,  15,   1,  11,  -4,  -5, -11, -19, -20,  16, -23,   2,  -3,   1,  -5,   1, -12, 104,  -2, -10 },
     {    0,   8, -34,  22,   6,  -2,   3,   0,  -4,  -8,  38, -32,  -9,   4,  -4,  -1,  14,  -8, -22,  17,  17,  -8,   6,  -3, -32,  25,   2,  -2, -23,   7,  -5,   1,  35, -26,   6,   3,  25,  -6,   2,  -1, -23,  21,  -4,  -7, -22, -13,   2,  -2,  -2,   6, -31,  24,  -4,  30,   0,  -5,  11, -18,  26, -10,  -3, -10, -10,   8 },
     {   -6, -13, -69,  13,  -3,   5,   4,   3,   1,  37,  31,  19, -11,  -1,  -6,  -1, -22, -31,  25, -37,  -1,   3,   1,   4,  46,  -5,  -7,  -5,  27,  -7,   2,  -2, -18,  -5,   0,  14, -11, -13,   6,  -2, -13,  17,  -7,  -3, -11,  12,   2,   0,   1,  -1,  -6,   3,   4,   2,  -7,   0,   2,  -5,   3,   3,  -4,  -5,   2,   4 },
     {   -7, -39, -55,  25,  -4,   9,   1,   1,  17, -14, -41,  33,  -8,  10,  -5,   3,  43,  42,   9,  30, -19,   1,  -8,  -2,   6,  16,   0,  -6, -29,   3,  -8,  -1, -10, -12, -13, -14,  -7,  10,   1,   2,  -8,  -9,   0,   1,  11,  13,   0,   2,  -1,   1,   8,   3,   6,  -1,  -4,   1,  -2,  -3,   3,  -7,  -3,  -4,   0,   1 },
     {    6,   7,  12,  11,  -4,  -7,   0,  -1,   9,  12,  18,  14,  -8, -11,   2,  -1,   9,  14,  17,  12, -10, -12,   0,  -1,   9,  15,  15,   6, -16, -15,   3,  -2,   6,  15,  14,   0, -22, -17,   2,  -4,   6,  14,   8,  -9, -29, -22,  12,  -3,  -2,   0,  -9, -28, -41, -25,  30,   9, -10, -14, -23, -33, -45, -34,  19,  13 },
     {   -7,   8, -16,   9,  -7,   6,   0,   2,  -5,  25, -11,  17, -11,   0,  -2,  -1, -22,  28,  -4,  -2,  -9,   4,  -5,  -2, -56,   7,  -8,   0,   5,  -4,   3,  -1, -29,  33,  -1,   1,  13,   2,   4,  -2,   3,  37,  -6,   0,  10, -22,   8,   2, -17,   5, -47, -11,  -8, -15,  10,   9,  -3,  13,  -8,  44,  18,  -2, -15,  -7 },
     {    2,  -1,  -8,  45,  -4,  -4,  -4,   1,  -3,  10,   9, -40,  -4,   5,   5,  -1,   5, -21,  15,   9,  26, -11,  -3,  -6,  13,  13, -32,  14, -33,   3,   9,   0, -46,  18,  20,  -3,   9,  19, -19,   4,  45, -14, -11,  -8,  15, -24,   2,   1,  -7, -16,  20,   7,  -2,  -2,  10,  -5,  -1,  15,  -2,  -6,   3,   7,  -7,  -1 },
     {    0,  -1,   0,  -4,   0,  17,  -8,   1,   1,  -2,  -1,   7,  -3, -25,   7,   2,   1,   1,   1,  -3,  -1,  26,  -8,   1,   2,  -2,   1,   6,  -2, -27,   9,   3,  -2,   6,  -3,   1,  -9,  33,  -9,   0,   3, -13,   4,  -5,  -1, -37,  10,  17,  -6,  16,  -7,   8, -16,  65, -44, -13,  -5,   7, -12,  18, -17, -14,  48,  -7 },
     {    1,   5,  15, -38,  10,   2,   5,  -3,  -4, -18,  15,   5,   9,  -5,  -1,  -5,  23, -16, -36,  18,   1,   2,   7,   1,  21,  39,  -4,  24, -11,  -2,  -2,  -6, -67,  -4,  -8,   3, -27,   5,   1,  -1,  -8,  39,  -1,  -2,  -1,  25,  -2,   0,   7,  -7, -20,   3,   7,   0, -10,   3,   3,  -8,   9,   9,  -5,  -5,   3,   3 },
     {    4, -10, -11,  -4,   8,  -5,   3,   0,   6, -16, -27,  -2,  17, -17,   7,  -2,  18,  -7, -15,   4,  14,  -7,   4,   1,  41,  13,  -6,  27,  12, -20,   5,  -1,  40,  15, -12,  12,  -9, -25,   5,   0,  11, -20, -30,   4, -19, -24,   5,   2, -17, -29, -23,  10, -13,  -9,  18,   1,   2,  10,   8,  46,   8,  -2, -16, -11 },
     {    6,   7,  30,  72,  -3,   1,  -2,  -2,  -7, -16, -27, -29, -28,   8,  -3,   5,  11,  10,  -4, -32,  27,  13,  -1,   2,   7,   1,  10,  22,  22, -22,  -7,  -5,  -9,  -5, -14,  -6, -24, -14,  15,  -3, -24,  13,   6,   7,   2,  21,  11,  -9,  -1,  17,  -9,  -4,   1,   0, -13,   5,   1,  -6,  -7,   1,   6,  -3,   2,   4 },
     {    0,   1,  -8,  26, -33,  17, -10,   4,   0,   0,  13, -30,  44, -24,  15,  -4,   3,  -6,  -5,  26, -31,  14,  -7,   2, -12,  10,  -9, -13,  18,  -6,   4,   0,  16, -20,  12,  -1, -11,  -2,   2,  -1, -16,  31, -21,  11,   6,   8,   1,   1,  22, -33,  24, -15,   2, -13, -11,   8, -17,  23, -22,  11,   4,   3,  17, -12 },
     {   -6,  -2, -37, -45,  23,   2,   4,  -2,  -1,  13, -28, -58,  36,   5,   6,   1,   4,  29,  23,  -6,  39,  -8,  -2,  -2,  -5,  14,  22,   5,  12, -14,  -2,  -1, -11,  13,  16,  -4,  -1, -10,  -2,  -6, -18,   5,  12, -12,   1,  -2,   6,  -5,   0,  12,   6, -15,   0,   0,   1,  -5,  -5,  -4, -12, -15,  12,   9,  16,   0 },
     {   13,  17,  18,  23,  18,  10,   2,  -1,  20,  27,  28,  34,  25,  17,  -4,   0,  18,  25,  24,  28,  22,   9,  -2,  -4,  19,  24,  22,  21,  22,   8,  -7,  -4,  18,  21,  20,  18,  13,   3,  -6,  -4,   9,  17,  18,  17,   9,   1, -18,  -4,  15,  18,  14,  16,   9,   1, -16, -11,   7,  11,  12,  16,   9,   1,  -7,  -7 },
     {    0,  -1,   4,   4,  -7,  -8,  -4,   3,   1,  -3,  -2,   5,  20,  13,   3,  -4,   3,  -8, -14,   0, -32, -14,  -6,   5,  11,  23, -12,   2,  18,  38,   5,   1, -12,   2,  10, -12,  15, -46, -24,   0,  -2, -18,   4,   3,  -4,  -4,  63, -11,  -8,  38,   2,  -1,  -1,  23, -12,  -9, -11,   3, -24,   1,  14,  -7, -12,  16 },
     {    2,   3,  -7,   5,  74, -34,  16,  -5,   0,  -2,   7,  -6, -31,   5,   3,   7,   0,   1, -10,  27, -22,  24, -11,  14, -10,  12, -12, -24,  38, -21,   0,   0,  -7, -14,  24, -15, -12, -12,  17,  -9,  24,  -9, -11,   8,   2,   3,  -8,   8,  -8,  -3,   3,   0,  -3,   3,   3,  -6,  -8,  12,  -7,  -7,   5,   2,   5,  -4 },
     {    2,  -3,  -7, -12, -22,  28,  26,   9,   3,  -7, -17, -18, -20,  41,  14,  25,   8,   3, -25, -23,  -9,  35,  13,  18,  19,  24,   1,  -9,  20,  26,  18, -10,  22,  20,   9,  -4,  12,  21, -15,  -6,  28,  18,   1,   3,  -3,  -8, -21,   1,   6, -11, -16,  -7,   1,  -4,  -9,  -4, -12,  -4, -13, -18, -16,  -5, -10,  10 },
     {    0,   0,   0,   5,  -8,   5,   6,   1,  -1,   0,  -1,   2,   5,  10,  -5,  -9,   1,  -2,  -2,  -3,  -7,   4,   5,   1,   1,   1,  -4,   3,   4,   8,  -2, -10,  -1,   0,  -2,  -5,  -9,  10,  -3,   5,  -3,   5,  -1,   7,  -6,  17, -22, -40,   0,  15,   0,   7,   1,  44,  88, -29,  -6,  13, -13,  17,   1,  15,  35,   0 },
   },
   {
     { -104,  57,   0,   5,   0,   1,   0,   0,  43, -11,  -8,   0,  -2,   0,  -1,   0,   7, -11,   4,   0,   0,   0,   0,   0,   2,   1,   0,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -42, -18,  22,  -1,   5,   0,   2,   0, -71,  64,  -2,  -3,   0,  -1,   0,   0,  59,  -3, -25,   2,  -3,   0,  -1,   0,   4, -19,   4,   4,   0,   1,   0,   0,   7,   0,   2,  -1,  -1,   0,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {  -16,  31, -21,   7,  -4,   2,  -2,   1, -58, -28,  40,  -1,   6,   0,   2,   0, -35,  67,  -7,  -9,   0,  -3,   0,  -1,  45,  -9, -25,   4,  -2,   1,  -1,   0,   1, -10,   4,   2,   0,   0,   0,   0,   8,  -2,  -2,   1,  -1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
     {   36,  77, -64,  10, -14,   2,  -4,   0, -34,  10, -17,  14,  -3,   3,  -1,   1,  24, -42,  15,   3,   0,   1,   0,   1,  -3,  -9,  15,  -5,   1,   0,   0,   0,   0,  -1,   3,  -3,   1,   0,   0,   0,  -1,  -1,   2,  -1,   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -18,  -5,  19,  -6,   3,  -1,   1,   0, -46,   7, -14,  14,  -1,   4,   0,   1, -38, -26,  57,  -4,   1,   0,   0,   0,  -7,  70, -10, -26,   2,  -5,   1,  -2,  24,  -7, -33,   8,   2,   1,   0,   1,   2,  -5,   4,   5,  -1,   0,   0,   0,   5,  -1,  -2,   0,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0 },
     {   -8, -15,  -5,   8,  -1,   2,   0,   1, -19, -69,  60,  -7,   7,   0,   2,   0,  62, -13,  14, -15,   1,  -3,   0,  -1, -34,  17,   7,  -9,   4,  -2,   1,  -1,  -8,  28, -13,  -4,   3,  -2,   0,  -1,   1,   2, -10,   4,   0,  -1,   0,   0,   0,   4,  -1,   1,  -1,   0,   0,   0,   0,   1,  -2,   0,  -1,   0,   0,   0 },
     {   18,  16,   9, -21,   5,  -7,   1,  -2,  15,  -4, -11,   9,  -2,   1,   0,   0,  51,  16,   6, -25,  -1,  -5,   0,  -1,  -8,  20, -72,  13,   7,   0,   3,   0,   4, -51,   8,  39,  -9,   5,  -2,   2,  -4,   0,  27,  -9,  -8,   2,  -1,   0,  -2,  -2,   0,  -5,   2,   2,   1,   0,  -2,   0,   1,  -1,   2,  -1,  -1,   0 },
     {  -17, -40, -61,  66, -17,  17,  -5,   4,  19,  40,  19,  -6, -13,   0,  -4,   0,  -8,  11,  17, -36,  11,  -5,   3,  -2, -21,   2, -14,   4,  10,  -2,   2,  -1,  15,  -6,   0,   9,  -4,   0,  -1,   0,  -1,  -4,   7,  -2,  -2,   0,   0,   0,   2,   1,  -1,  -3,   1,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 },
     {    6,  32,  -5, -12,   2,  -4,   1,  -1,  -2,  18,  17, -27,  11,  -7,   3,  -2, -20,  20, -55,  20,   3,   1,   1,   1, -74,  34,   1,  22,  -8,   3,  -1,   0,  28,  10, -19,   3,  -8,   2,  -2,   1,   9, -25,   5,   7,  -3,   3,  -1,   1,   2,   0,   4,  -5,   0,   1,   0,   0,   1,  -3,   0,   1,   1,   1,   0,   0 },
     {  -17,  -9,   5,  -7,   5,  -3,   2,   0, -32, -22,  -7,  22,  -4,   7,  -1,   2, -34, -21,  10,  -7,   8,   1,   3,   1, -53, -22,  -5,  41,   1,   5,   1,   1,   4,  -4,  66,   0, -15,   2,  -5,   0,   2,  34,   2, -30,   2,  -3,   1,  -1,   4,   3,  -9,  -1,   3,   1,   0,   0,   0,   3,  -1,  -2,   2,   0,   0,   0 },
     {    1,   7,  -2,  -8,   2,  -2,   0,   1,   3,   5, -40,  50, -19,  10,  -5,   2,  15,  70,  13, -18, -22,   2,  -6,   0, -41,  -2,   7, -30,  13,   2,   1,   1,  -8,  27,  -2, -14,  13,  -6,   1,  -1,  12,   3, -31,   0,   4,  -3,   2,   0,   0, -11,  -3,  12,  -1,   2,   0,   0,   1,  -1,   1,   2,  -2,   0,   0,   0 },
     {  -15, -34, -27,  26,  -5,   8,   0,   2, -12, -45, -38,  37,   1,   6,   1,   2,   9,   6, -29,  54, -22,   9,  -4,   2,   0,   5,  15,  -4, -10,   3,  -3,   2,  12, -26,  -7,  15,   1,   2,   1,   1,   4, -18,  39,   5,  -9,   3,  -2,   0,   0,   8,  12, -15,  -1,   0,   0,   0,   1,   0,   1,  -4,   0,   0,   0,   0 },
     {    5,  -5,  -6,  -1,   4,  -3,   2,  -2,  10,  -8,   5,  12,  -9,   4,  -3,   1,  18,   3,  -4,  -3,   0,  -3,  -1,  -1,  15,  21,   5,  18, -27,   6,  -5,   1,  70, -37,  16, -58,   5,   7,  -3,   3, -26,  -3, -32,   5,  34, -10,   7,  -1,   2,  -5,   3,  21, -10,  -4,   0,  -1,  -3,  -1,   1,  -3,  -4,   4,   1,   0 },
     {  -10, -14, -49, -40,  75, -32,  20,  -7,   9,  17,  34,   8, -10,  -8,   4,  -3,  -6,  -7,  19,  19, -38,  18,  -8,   4,   1, -13,  -8,  -5,   4,   7,  -5,   3,   4,   4,  -2,   6,   4,  -2,  -1,   0,  -4,   1,   4,   3,  -3,  -1,   1,   0,   0,   2,   3,  -2,  -1,   1,   0,   0,   0,   1,   1,  -1,   1,  -1,   0,   0 },
     {   -7, -12, -23,  -2,  19,  -5,   5,  -3,  -4,  -2, -21,  27, -22,   7,  -3,   1,   1, -14, -26,  -9,  31,  -4,   4,  -2,  21,  37, -28,  50, -19, -11,   3,  -3, -45,  37, -11, -10, -20,  12,   0,   2,   9,  -7, -21,  21,   6,   3,  -2,   0,   6, -13,  18,   5, -12,   4,  -2,   0,   2,   1,   3,  -5,  -1,   1,   0,   0 },
     {   -3,   0, -13,  19,  -5,   4,  -2,  -1,  -1,  15,  -1, -26,  19, -10,   5,  -2,   6,  12,   0,  39, -24,   1,  -3,  -2,  -8,  45, -14, -12, -28,  13,  -6,   3, -57, -18,  19, -17,  19,   4,   1,   3, -17,  50,  -6, -11,  13,  -9,   3,  -3,  11,  -5, -26,   7,  -3,  -3,   0,  -1,  -2,  -6,   1,   2,   0,   4,   0,   1 },
     {   -7,  -5,  -4, -11,   4,   4,  -2,   4, -12, -10, -10, -12,  -2,   7,   1,   1, -18, -15, -16, -20,  12,   0,   2,   3, -28, -31, -14, -13,  15,   9,   5,   4, -22, -36, -15,   2,  50,   7,   4,   1, -11,   6,   4,  61,  24, -19,  -2,  -2,   8,   7,  23,  24, -27, -14,  -4,  -3,   1,   0,   5,  -5, -12,   3,   2,   1 },
     {  -11, -21, -41, -33,   8,  11,   0,   2,  -7, -18, -56, -51,  57, -14,  12,   2,  11,  15,  18,  -2,  20, -14,   7,  -6,   4,  -6,   0,  10,   8,  -5,  -1,  -1,  20,  13,  -3,   9,  -8,  -8,   3,  -2,  -1,   2, -19,  -3,  -2,   7,   0,  -1,  -1,  -3,  -7,  17,   6,   0,  -2,  -1,   1,   0,   2,   5,  -5,   0,  -1,  -1 },
     {    9,  14,  22,  57,  41, -44, -10, -13, -11, -16, -30, -50, -30,  13,  17,  10,   6,  14,  20,  -1,  -5,  27,  -6,   0,  -6,  -9,   5,  23,   8, -15, -12,   0,   3,  -3, -15, -11,  -8,   1,   9,   5,   2,   5,   8,   4,   3,   1,   0,  -3,   1,  -5,   1,  -2,  -3,   1,  -1,  -1,   1,   0,   3,  -3,   0,   0,   0,   0 },
     {    4,  -5,   6,  14,  -7,  -1,   2,   1,   5,   0,   4,  -9,  10,  -5,   3,  -1,   3, -20,  11,  14, -18,   4,  -1,  -1,   1,  -9, -25,  -4, -16,  18,  -6,   4,   8,  -9,  16,   3,  14,   3,  -2,   3,  97, -18, -28,   2,   0,  -8,  -6,   1,   2, -40,  12,  12,  -3,  -3,  -1,   3,   4,   1,  12,   0,  -5,  -2,   1,   0 },
     {   -3,   0,   3, -20,   0,  13,  -8,   0,   3,  12,   9,  -1,   8,   0,  -1,   0,   7,  26,  56,   7,   9, -24,  -1,  -9, -19, -19,  29,  22, -47,  -9,   0,  -4, -31, -27,  -9, -22, -15,  28,   2,   5,  13, -26,  21,   9,  10,   9,  -6,   2,   4,  12,  24, -23,  -6,  -1,   0,   1,  -3,   0,  -7, -11,   7,   1,   2,   2 },
     {    6,   7,   8,   9,  -4,  -4,   3,  -1,   8,  10,   6,  12,  18,  -8,   2,  -3,   8,  12,  16,  14,   2, -10,  -2,   1,   7,   6,  11,  18,   3,  -9,   1,  -1,  22,  17,  14,  19,  -2, -23,   3,  -5,   3,  34,   3,  30, -36, -23,   2,  -7,   8,   0,   9, -10, -71,   4,   2,   0,  -2,  -8,  -6, -34, -14,  35,  10,   4 },
     {   -3,  -6, -14, -21,   8,  -7,   1,  -8,   2,  -1,   9,  -2, -16,   6,   1,   1,   2,   4, -22, -21,  17,  10,   7,  -4,   5,  25,  61,  -6,  22, -32,   2,  -4, -16, -59,   4,  14, -13, -10,   8,   2,  39,  18, -15,  -1,  -9,  11,   8,  -1,  15,  -7, -20,   4,  -5,   6,   0,  -2,  -9, -10,   3,   8,   8,   9,  -3,  -3 },
     {    4,   6,   8,  30,  29, -38,   9,   6,  -1,  -2, -12,  20,  53, -37,  11, -12,  -5,  -5, -21, -59, -14, -12,  16,   1,   4,  13,  23, -18, -20,  19,   3,   1,  -7,  -1,  16,   2,  -2,  14, -15,  -6,   0, -10,  10,   0,   4,   3, -11,   5,   5,   5,   9, -12,   0,  -7,  -4,   4,  -2,   3,  -3,  -4,   1,  -2,   5,   1 },
     {   -1,  -3,  -5,  -8,   2,  -3,   3,  -2,  -2,  -4,  -8,  -9,  -9,   8,  -1,   1,   1,  -1,  -3, -13,  17,  -6,   3,  -1,  -2,  -2,  -2,  -3, -26,  18,  -4,   4,  11,  -2,   1, -15,  41,  -7,  -2,  -2,  -6,  18,  -8,  22, -50, -11,   6,  -6,   7, -23,   1, -54,  14,  47,   1,   4,  -3,   4,  -4,  14,  46, -22, -11,  -1 },
     {   -2,  -3,  -7, -20,  -5,   6, -14,   9,   1,   4,  -5, -23,   0,  17,   1,   5,   5,   5,   2, -14,   0,  24, -12,  11,  12,  37,   7,  -2,  37,   4, -19,  10, -13,  13,  51, -30,  20, -25,  -9,   0,  11, -36,  32,  -8,   7, -20,   6,   0,   7,  11,  29, -20,  -1, -10,   5,  -4,  -1,   7, -13, -20,  -1,  -7,  10,  -2 },
     {    1,   4,  11,  17,  13,   7, -13,  -2,  -2,   5,  12,   9,  13, -11,  -4,   5,  -3,   0,   9,  18,   6,  -2,   0,  -7, -11,   5,   6,  -6,   8,  -9, -16,   6, -18, -21,  30,  14,  -9, -32,  15,   1, -41, -29, -34,  12, -26,  27,   5,  -3,  -6, -28,  43,  17,  11,  13, -20,   7,   3,  13,  34,  -4, -11, -16,  -7,   0 },
     {   -3,  -4,  -6,  -2, -33, -49,  63,  15,   2,   4,   3,   4,   4,   6,  15,  -8,   4,   8,   5,  10,  55,  23, -43,  -3,  -5,  -3,  -1, -21, -13,   4, -24,  10,  -2,  -4,   1,   3, -11,   2,   5,   6,  -1,   6,  -2,  -2,  -2,  -5, -12,   0,  -2,   6,   6,   3,   5, -11, -14,   5,   4,  12,  -1,  -3,  -8, -10,   7,   4 },
     {    0,   0,  -5,   1,  28,  -5, -25,  10,   0,  -1,   0, -13, -29,   4,  19,  18,   2,   2, -12,  21,  37, -37,   5, -15,  -1,   5,  -6, -55, -12, -28,  26,  -1,  11,  11,  37,  10,   5,  37,  -7,  -6,   1,  -7,   0, -11,  -4, -15, -22,   4, -14,   4,   4,   0,  -7,   0,   4,   7,  -4,   8,   5,  -9,  -2,  12,  -1,   0 },
     {    3,   0,   0,  -1,  -5, -10,  -8,   8,   5,  -1,   4,  -1,   4,  18,   0,  -1,   1,  -1,  -1,   9,  -6, -13,  -2,   9,   2,  -9, -11, -13,  12,  -7,  13,  -2,  10,   7,   6,  -9, -17,   9,   5,  -2,  -6,   5,   2,  12,  -2,  24, -20,   1, 102, -16,   1,  -4,  15,  -7, -12,  -5,  -2, -25,  -8, -12,  -5, -21,  16,   6 },
     {    3,   4,   4,  18,  24, -13,   6,   2,   4,   6,   5,  22,  28,  -1, -12,  -7,   1,   0,  -1,  21,  22, -45,  -3,  15,  -3,  -3, -12,   7,  17, -40,  22,  -6,   0,  -6,  -8, -29,  18, -35,  27,   0,  19,   8,  16,  -1,   5, -16,  22,  -1, -11,  -1,  -6,  -6,  14, -26,   4, -15,   0,  13, -14,  20,  -2, -34,   2,  -3 },
     {    1,   2,   0,   6,   4, -18,  18, -21,   1,   2,   1, -10,  -7,  49, -26,   0,   4,   8,  10,   6,   2, -34,  22,  -5,   6,  10,  24,  18,  -6,  40,   4, -16,  -6,  -2,   7,  28,  13,  -1, -30,   5,  -6, -19,  -7,   5, -23, -28,   1,   4,   7,  14, -14,  18,  18,  -7,  11,   0,   0,   0,   4,  10, -42, -16,  10,  -6 },
   }
 },
 {  //12
   {
     { -120,  32,  11,   5,   3,   2,   1,   1,  -6,  10,   1,   0,   0,   0,   0,   0,  18,   4,  -5,  -1,  -1,   0,   0,   0,  13,  -2,  -3,   0,   0,   0,   0,   0,   6,  -3,  -1,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    4, -33,  11,  -1,   3,   0,   1,   0,-113,  18,  19,   2,   5,   1,   2,   0,   4,  29,  -2,  -2,  -1,  -1,  -1,  -1,  21,   4,  -9,  -1,  -1,  -1,  -1,   0,  10,  -2,  -3,   1,   0,   0,   0,   0,   5,  -2,  -1,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },
     {  -25, -24,  12,   4,   3,   2,   1,   1,  -2, -63,  23,   2,   5,   2,   2,   1, -77,   0,  36,  -2,   5,   0,   2,   0, -13,  44,   9,  -9,   0,  -3,   0,  -1,  10,  22, -11,  -6,   0,  -1,   0,  -1,   7,   5,  -9,   0,   1,   0,   0,   0,   3,   1,  -3,   1,   0,   0,   0,   0,   2,   1,  -1,   0,   0,   0,   0,   0 },
     {   12,  84, -31,  -3,  -8,  -2,  -3,  -1, -30,  39, -24,   4,  -2,   0,   0,   0, -67,   0,  -2,  11,   3,   3,   1,   1, -17,   4,   4,   5,   0,   1,   0,   0,   9,   1,   1,   0,  -2,   0,  -1,   0,   9,  -3,   0,  -1,  -1,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 },
     {   12,  18,  -3,  -5,  -3,  -1,  -1,   0,  32,  -4,  10,  -9,  -1,  -2,   0,  -1, -12,  72,  -2, -14,  -2,  -4,  -1,  -2,  53,  28, -49,  -7,  -4,  -1,  -1,  -1,  14, -36, -26,  12,   1,   3,   1,   1,  -2, -21,   5,  10,   0,   2,   0,   1,  -3,  -6,   6,   2,  -1,   0,   0,   0,  -1,  -2,   2,   0,  -1,   0,   0,   0 },
     {   -4,  -7, -68,  33,  -3,   6,   0,   2, -13, -48, -43,  44,  -1,   7,   0,   2,  11,  -8,  11,  24,  -9,   2,  -2,   1,  44,  -4,  12,  -1, -12,   0,  -2,   0,  13, -17,  12,  -7,  -3,   2,   0,   1,  -2,  -6,  13,  -5,   1,   1,   0,   0,  -4,   1,   7,  -3,   1,   0,   0,   0,  -1,   2,   3,  -2,   0,   0,   0,   0 },
     {  -17, -50, -31,  25,   4,   6,   2,   1,  17,  51, -39,  18,  -7,   0,  -2,  -1,  -2,  49,  -8,   4,  -8,  -3,  -1,  -1, -44,  37,   4,  -6,  -2,  -1,   1,  -1, -13,  23, -13,  -7,   1,   0,   0,  -1,   3,  -1, -10,   3,   2,   1,   0,   0,   5,  -5,  -2,   3,  -1,   1,   0,   0,   1,  -5,   0,   1,   0,   1,   0,   0 },
     {  -15, -24,  -2,   7,   3,   2,   1,   1,   2, -15,   1,   3,   3,   0,   1,   0, -42,  29, -24,   4,   1,   0,   1,   0, -26, -49, -28,  30,   4,   6,   2,   3, -29, -33,  42,  32,  -2,   3,   0,   1,   3,  21,  38,  -9, -12,  -3,  -2,  -1,   3,  14,   2, -15,  -1,   0,   0,   0,   1,   3,  -3,  -3,   1,   0,   0,   0 },
     {    6,  46, -27,  -1,  -7,   1,  -2,   1, -18, -40,  23,   4,   2,   2,   1,   1,  41,  42,  23, -16,  -5,  -6,  -2,  -2, -34,   8,  -2, -18,   2,  -1,   2,   0, -61,  14,   4,  -5,  10,   2,   3,   0, -10,  26,  -7,  -2,   3,  -2,   0,  -1,   3,   8, -10,   0,   0,  -1,   0,   0,   3,   0,  -6,   2,   0,   0,   0,   0 },
     {    5,   3,   6,   2,  -3,   1,  -1,   1,  17,  13,  11,  -4,  -5,   0,  -1,   0,  -7,  35,   2,  -5,  -8,   0,  -2,  -1,  36, -11,  22,   0,  -9,   2,  -2,   0,  23,  27,  50, -30, -13,  -3,  -3,  -2,  13,  31, -18, -56,   1,  -1,   0,  -1,  -7,  -5, -42, -11,  22,   4,   4,   1,  -3,  -5, -11,  14,  10,  -2,   1,   0 },
     {    7,   9,  36,  61, -21,  -7,  -7,  -2,   7,  23,  43,  57, -32,  -7,  -7,  -2,  -7,  -5,  24,   6, -30,   2,  -1,   1, -10, -18,  -2, -22, -14,   9,   3,   1,  -3, -22,  -9, -15,   4,   9,   2,   1,   1,  -8,   6,  -2,   7,   3,  -1,   0,   2,   4,   7,  -4,   1,  -1,  -1,  -1,   0,   2,   1,  -4,   1,   0,   0,   0 },
     {  -10, -19, -31, -35,  14,   4,   4,   2,   7,  29,  11, -26,   9,  -1,  -1,  -1,  -7,  10,  78,  -4,  -4,  -6,  -4,  -2,  -4, -35,  37, -15,  -7,  -1,  -1,   1,   1, -45,  -3,  -9,   7,   4,   2,   2,   6, -10,   4,   6,   8,   0,   0,   0,   5,   7,   4,   2,   0,  -3,  -1,  -1,  -1,   3,  -2,  -2,  -1,  -1,   0,   0 },
     {   -3,  -6,  19,  -3,   0,   0,   0,  -1,  -1,  25, -20,  -5,  -2,   2,  -1,   0, -24, -17,   0,   8,   5,   6,   0,   1,  47,  30,  13, -13,  -2,  -3,  -3,  -2, -62, -12,   4, -14,   8,  -1,   3,   1, -62,  23,  17,   0,  12,  -2,   3,   0,  -6,  32,  -5,  -3,   1,  -3,   0,  -2,   2,   6, -11,  -1,   0,   0,   0,   0 },
     {   -3,  -3,  -8,   3,  -3,   1,  -1,   1,   6, -11,  20,  -7,   2,  -2,   0,   0, -24,  30, -44,  14,   0,   3,  -1,   1,  26, -54,  40,   1,  -3,   3,  -2,   2, -26,  33, -19, -31,  12,  -3,   3,  -1,   1, -23, -28,  34,   9,   3,   1,   1,  -2,  -1,  27,  12, -10,  -4,  -2,   0,   2,   1,   3,  -7,  -4,   2,   0,   0 },
     {   -2,  -9, -18,  19,  -5,   8,  -1,   2,   0,   6, -13,  -4,  -8,   7,  -3,   1, -13, -17,  -1, -50,   5,  11,   1,   3,  15, -27, -26, -47,  42,  10,   2,   3,  15,  13,  -6,   6,  51,  -8,  -3,  -2,  13,  34,  -2,  18,  11, -26,  -4,  -2,   0,   5, -16,   8,  -9, -11,   5,   0,  -3,  -4,  -7,   6,  -6,   1,   4,   0 },
     {    5,   8,   0,  22,  76, -25, -12,  -3,   4,   7,   3,  17,  69, -35,  -8,  -2,  -4,  -6,  -6, -11,   4, -26,   4,   3,   3,  -7, -10, -25, -19,  -6,  13,   4,   0,   9,   0,  -6,  -8,   8,   9,  -1,   2,   7,   6,   3,  -4,   5,  -1,  -2,  -2,   4,  -2,   0,  -4,   1,  -1,   0,  -1,   0,  -1,   0,  -1,   1,   0,   0 },
     {    6,  10,  69, -21,  -1, -13,  -2,  -4,  -6, -24, -77,  17,   9,   7,   4,   3,   4,  32,  13,  15,  -9,  -1,  -3,  -2, -10, -18,  15, -21,  -5,   3,  -2,   2,   5, -11,  -4,  -5,  16,   8,   3,   2,  18,   8,  -4,   4,   8,  -5,  -2,  -3,   6,  -3,  -1,   7,  -4,  -5,  -1,   0,   0,  -1,   1,  -1,  -4,   0,   1,   1 },
     {    4,   7,  11,  39,   9,  -7,  -9,   0,  -6, -10, -11, -24,  -1,  -1,   1,   4,   1,   7, -17, -66,  -3,  15,   8,   2,  -7,  23,  55,  -4,   9,   7,  -6,  -3,  -2, -29,  38,   9,  -1,  -7,  -6,   1,  -3, -43,   6,   1,  -1,  -1,   3,   4,  10,  -7,   9,  -3,   1,   3,   2,  -1,   5,   3,   5,  -3,   2,   1,  -1,  -1 },
     {    5,   7,   4,  -5,  -4,  -3,   2,   2,   9,  12,  10,   3,  -6,  -4,   2,   1,  11,  12,  13,   4,  -8,  -2,   0,  -2,  20,  20,  24,   4,  -8,  -2,  -3,  -2,  15,  34,  21,  18,  -4,  -6,  -5,  -4,  23,  31,  37,  26, -25, -12,  -9,  -4,  20,  19,  27, -25, -51,  -7,  -2,  -1,  -1,  -4,  -7, -41, -22,  10,   5,   3 },
     {   -5,  -6, -13, -22,  -7,  -2,   7,   0,  -6,  -9, -10, -31, -20,  -4,  10,   0,  -5, -14, -23, -24, -48,   6,  18,   1,  -8,   1, -18, -31, -47,  31,  19,   3,   5,   9,  -2, -24, -11,  44,   5,  -2,   5,   4,  20,  -4,  14,  25, -11,  -5,   1,   7,  11, -11,   7,   2, -11,  -1,   0,  -3,  -3, -11,   4,   0,  -3,   2 },
     {    2,  -4,  -3,   4,   7,   2,   0,  -1,   3,   1,   3,  -5,  -7,  -1,   0,   2,  -1, -21, -15,   9,  -2,   3,   4,   1,  21,  16,   3,   9, -10,  -3,   1,  -1, -42, -34, -13,  -2,  -2,   5,   4,   3,  68,  16, -22,   5,  -1,   4,   1,   2,  65,  -9, -31,   5,  -8,   2,   0,   1,  10, -23,  -3,  10,   0,   4,  -1,   0 },
     {   -3,  -6,   5,  -6,  -3,   1,  -2,  -1,  -1,   9, -12,   2,  -6,  -2,  -1,   2, -11, -10,  32,  -7, -21,   0,   4,  -2,  22, -11, -23, -14,  -5,   8,   9,  -2, -37,  26,  35,  29, -12,   0,   1,  -5,  -7, -12, -23,   0, -31,   7,   3,   1,  -6, -59,   9,  23, -14,  21,   3,   5,   7,  -2,  40,  -4, -13,   6,  -5,  -1 },
     {    2,   2,   6,  -5,  16,  53, -61,   6,   0,   1,   2,  -4,   8,  51, -61,  13,   0,   0,  -4,  -1, -16,  15, -21,  21,  -3,   2,  -2,   0, -19,  -5,  12,   6,   1,   3,  -2,   3,  -8,  -2,  19,  -6,   4,  -1,   6,   2,  -1,  -8,  11,  -7,  -5,   2,   2,   0,  -2,  -8,   3,  -2,   0,   2,   0,   1,   1,  -4,   2,   0 },
     {    1,   3,  -1,  19,  13, -12,  -6,  -5,  -2,  -2,  -9,  -9,  -7,  -8,   8,  -7,  -1,   3,  13, -14, -40,  20,   2,  -6,   5, -12,  14,  38, -29,  15,  -7,  -7,  10,   2, -30,  51, -16,  -2,  -4,  -8, -19,  28, -41,  28, -12,  -4,  -6,  -4, -15,  26, -27,   6,   3,  -5,  -3,   1,  -8,   9,  -7,  12,   5,  -5,   0,   1 },
     {    0,   0,   2,   0,   1,  -4,  -1,  -1,  -2,   0,  -8,  -2,   1,   5,  -2,   0,  -2,  -3,   5,   1,   0,   8,   1,  -1,   6, -10,   7, -16,  12,   0,  -2,  -3, -18,  11, -34,   8,   5,  -5,  -2,   1,   9, -15,  11,  -2, -39,   4,   1,   5, -12, -19,  -7, -73,   2,  25,   9,   8, -12,  -3, -20,   0,  67,  16,   0,   0 },
     {   -2,   0,  -8, -64,  51, -11,   7,  -2,   3,  -1,  17,  61, -29,  -1,  -6,  -2,  -3,  -1, -19,  -9, -33,  17,  -5,   4,   0,  13,   8,   1,  21,  16, -10,   0,  -1,  -1,  13,  10,  17,  -4, -16,   1,  -5, -17,  -5,  -2,   9, -11,   0,   6,   8,   1,  -1,  -2,   3,  -5,   4,   0,   1,   4,  -4,   5,   7,  -1,   3,  -2 },
     {    1,   2,  -1,   5,  -5,   5,  -5,   2,   2,   2,  18, -16,  12,  -6,   5,   0,   5,   3, -10,  46,  -1,   2,   6,  -4,   1,  15,  25, -21, -17,  24,   6,  -4,  -3,   8,   4,  34,  38,  23, -16, -10,   9, -19,  37,   7,  -4, -27, -24,   2, -17, -14, -28,   3,  17, -14,  12,   6,   6,   1,  13,  40, -16,  -5,   2,  -3 },
     {    5,   6,   7,  26,  41,  10, -17,  -2,  -4,  -4, -12, -39, -20, -19,  -3,   9,   4,   1,  24,  20, -35, -15,   9,  12,   7,   2, -15,  41,  37,  14,  -8,  -6, -10,  13, -14, -20,  28,  -2, -19,  -4,   8,  -1,   7, -30,  11,  -2, -16,  11,   2,  -2,  28,  -4,  11,  -3,  -1,   6,   3,   0,   3, -11,  -6,  -4,   4,   0 },
     {    6,   9,  10,  12,  16,  57,  53,   5,  -1,   0,   0,  -2,   1,  20,  18,  -9,  -6,  -8,  -8,  -8, -32, -55, -33, -21,   1,   0,   4,   4,  -7, -30, -14,   2,   3,   3,   8,  14,  22,  14,  17,   6,  -7, -11,  -5,   5,   5,   5,   3,  -6,   0,  -4,  -9,  -8,  -6,   2,  -4,  -7,   0,  -1,  -7,   1,   4,   1,   0,  -2 },
     {   -1,  -1,  -2,  -1,  -6,   7, -10,   2,  -1,  -2,   7,  -2,  15,  -4,  -4,  -1,   2,  -1,  -6,  11,   1, -10,  14,  -5, -12,  15,  -6,  23, -15,  22,  14,  -3,  14, -31,  25, -29,  27,  18, -13,  -2, -24,  17, -33,  21,  16, -34,  -5,   2,  -8, -28,  -7,  -5, -43,   6,  16,   9, -11,  -7,   4, -20,  36,  27,   6,  -1 },
     {    4,   6,   5,  14,  24,  25,  22, -13,   5,   7,   8,   2,  36,  55,  39, -14,   3,   1,   7,   6,   7,  55,   9, -15,  -4,  -4, -17,  -8, -12,  24, -31, -18,  -4,   3,  -1, -14, -12,  -7, -29,   6,   3,   3,   1,  -6,   1,  -4,  -6,  17,   7,  -1,  14,   4,   4,  -3,  -1,   9,  -2,  -4,   4, -11,  -3,   0,   3,   2 },
     {   -1,  -2,   0,   2,  -2, -17,  -3,  -6,   1,   6,  -2,   8,  -4,   3, -17,  -8,  -1,  -4,  10, -17,  33,   7, -35,  -3,  10,   5,  -3,  16, -10, -17, -14,   2, -16,  12,  13,  17,   4,  40,  -9,  -1,  19,   4,  -9,  10,  38,  14, -47,  -3,  -9, -17,  16, -14,  18, -34, -22,  19, -13,  11,  10, -13,  23, -14,  15,  15 },
   },
   {
     {   75, -42,   2,  -4,   0,  -1,   0,   0, -77,  36,   5,   2,   1,   1,   1,   0,  33,  -1, -12,   1,  -2,   0,  -1,   0,  -9, -12,   8,   1,   1,   0,   0,   0,   2,   9,   0,  -3,   0,  -1,   0,   0,  -2,  -3,  -4,   2,   1,   0,   0,   0,   2,   1,   2,   1,  -1,   0,   0,   0,  -1,  -1,   0,  -1,   1,   0,   0,   0 },
     {   35,  17, -17,  -2,  -3,   0,  -1,   0,  -6, -51,  26,   2,   5,   0,   2,   0, -35,  66, -15,  -7,  -1,  -1,   0,  -1,  33, -42, -10,  15,  -2,   3,  -1,   2, -16,  10,  27, -14,   1,  -2,   0,  -1,   2,   7, -24,   3,   3,   0,   1,   0,  -1,  -4,  11,   6,  -5,   0,  -1,   0,   1,   1,  -1,  -7,   2,   1,   0,   0 },
     {   69, -14,   2,  -2,  -1,  -1,  -1,  -1,  18, -26,  -9,  -3,   0,   0,   0,   0, -47,  -1,  32,   0,   4,  -1,   1,   0,  -5,  48, -29, -10,  -1,  -2,   0,  -1,   8, -35,  -7,  25,  -2,   4,  -1,   2,   0,   4,  30, -17,  -4,  -1,  -2,   0,  -6,   5, -17,  -5,  10,   0,   2,   0,  -1,   1,   1,  11,  -6,  -2,   0,   0 },
     {  -48,   5,   9,   2,   2,   0,   1,   0, -55,  19,   3,  -2,   2,   0,   1,   0,   9,  43,  -3,   0,  -2,  -1,  -1,  -1,  32,   0, -42,   2,  -3,   2,   0,   1,  13, -45,  16,  24,  -4,   4,  -1,   2,   5,   6,  29, -27,  -3,  -2,  -1,  -1,  -5,   5, -22,  -3,  15,  -1,   3,  -1,   1,   1,   1,  15,  -8,  -3,   0,  -1 },
     {   20,   5,  -3,   1,  -2,   0,  -1,   0,  35,   5, -15,  -1,  -2,   1,  -1,   0,  28, -27,  -6,   4,  -3,   1,  -1,   1,  -6, -45,  10,   7,   2,   3,   0,   2,  -9,  -7,  57,  -4,  -3,  -1,  -2,   0,  -8,  32,   1, -50,   8,  -4,   2,  -2, -14,   7, -41,  15,  22,  -1,   5,  -1,   0,   3,   3,  24, -17,  -5,   0,  -1 },
     {    4,   2, -20,  16,  -3,   2,  -1,   1,   6,  -6,  27, -23,   4,  -4,   1,  -1,  -5,   9, -33,  25,  -4,   3,  -1,   1,  12, -20,  40, -24,   2,  -2,   0,   0, -18,  25, -42,  22,   3,   0,   2,   0,  20, -30,  45, -10, -13,   4,  -5,   2, -16,  23, -24, -19,  22,  -4,   5,  -2,   3,  -4,  -2,  24, -11,  -2,   0,   0 },
     {  -13,  -5,   8,  -1,   4,  -1,   1,  -1, -19,  -2,  12,  -8,   4,  -2,   2,  -1, -15,  15,  16,  -6,   5,  -3,   2,  -1,  -5,  33,   4, -19,   3,  -4,   1,  -2,   1,  33, -25, -17,   7,  -4,   3,  -2,  -6,   8, -55,  -8,  13,  -1,   6,  -1, -10,  -8, -39,  51,  17,  -6,   6,  -1,   7,   3,  15,  44, -28, -12,  -3,  -2 },
     {   28,  40, -17,  -4,  -1,  -2,   0,  -2,  34,  46, -28,  -6,  -2,  -3,   0,  -2,  19,  19, -37,  -3,   0,  -2,   1,  -1,   2,  -9, -39,   4,   2,   0,   2,   0, -11, -24, -31,  18,   7,   1,   3,   1, -12, -23, -11,  29,   3,   0,   2,   1,  -4, -10,  10,  33,  -1,  -4,  -1,  -1,   3,   1,  17,  15, -15,  -5,  -3,  -1 },
     {    1,  -2,  -2, -14,   9,  -1,   2,  -1,   5,   9,   2,  16, -12,  -1,  -4,   0,  -4,   0,  -6, -21,  13,  -1,   3,  -1,   2,   3,   1,  22, -13,   0,  -3,   0,   0,  -8,   1, -32,  18,   0,   4,   1,  -1,   6,  -2,  30, -10,  -5,  -5,   0,   0, -10,   5, -54,  24,  16,   2,   1, -16,  26, -37,  68,  -4, -29,  10,  -7 },
     {    9,  64, -12,  -3,  -4,  -1,  -2,   0, -30, -47,  -8,  11,   1,   5,   1,   3,  52,  -8,  37, -11,   0,  -3,   0,  -1, -29,   1, -20, -13,   7,  -3,   2,   0, -10,  25,  -3,  26,  -2,   2,  -1,  -1,  16, -11,  -2,  -7, -18,   2,  -1,   2,   0, -10,  12, -10,  19,   1,  -1,   1,  -1,   9,  -6,   6,  -5,  -6,   2,  -2 },
     {    9,  45, -11,   2,  -3,   0,  -1,   0, -30,  -5, -19,   4,  -2,   2,   0,   0,  -1, -29,  18,  15,   1,   4,   1,   1,  51,   2,  28, -24,  -4,  -5,   0,  -1, -26, -30, -15, -22,  19,  -4,   2,   2, -13,  38,   9,  29,  14,  -1,  -3,  -6,  -1,  12, -22,   1, -34,   3,   4,   0,   1,  -8,   0,   2,   2,   5,  -1,   2 },
     {   -3,  -4,  -1,   2,   2,   1,   1,   0,  -6,  -7,  -1,  10,  -7,   1,  -2,   1,  -5,  -7,   2,   4,   3,   1,   1,   0,   2,   1,   8,  11,  -9,   0,  -2,  -1,   4,  -5,   5,  -7,   2,  -3,   5,   2,   0,  -7,  10,   9,   4, -13, -12,   3,   9,  17,  27,  13,  13,  24, -14,  -4,  -6,  -1, -12, -18,-107,  27,  13,   4 },
     {    6,  58, -15,   4,  -5,   1,  -2,  -1, -30,  35, -26,   8,  -4,   2,  -1,   0, -63, -15, -17,  20,  -3,   6,   0,   3, -11,  11,  21,  24,  -8,   4,  -2,   1,  30,  18,  17,  -8, -19,   1,  -4,   0,   8, -16,  -6, -26,  -3,   1,   0,   2,   2,  -5,   5,  -4,  18,  -2,  -2,   0,   1,   2,   1,  -3,   9,  -1,  -1,  -1 },
     {   25,  32,  26,   1,  -4,  -2,  -3,  -2,  34,  40,  35,  -1,  -5,  -3,  -4,  -2,  36,  34,  30,  -5,  -8,  -4,  -5,  -3,  37,  29,  18, -13, -11,  -3,  -5,  -2,  25,   9,   1, -25, -10,  -1,  -2,  -2,   6, -10,  -8, -25,  -2,   3,   2,   1,   1,  -3,  -6, -22,   0,   3,   1,   0,  -4,  -9, -15, -21,  -6,   7,   4,   3 },
     {    0,  17, -30,  -8,  -3,  -3,   2,   1,  -3, -20,  32,  14,   3,   3,  -2,  -1,  19,   3, -19, -25,   1,  -6,   3,   0, -37,  18,   0,  31,   3,   5,  -2,  -2,  33, -16,  -4, -27, -15,  -4,   1,   3, -12, -19,  20,   2,  40,   1,  -2,  -2, -13,  44, -24,  10, -32,  -7,   9,  -5,   1, -14,  -9,   9,   6,   4,  -3,   3 },
     {   -2,  11,  -2,  -1,  -2,  -2,  -1,  -1,   3,  31,   1,   3,  -2,  -2,  -1,  -2, -14,  30,  -8,  -5,  -2,  -4,   2,  -2, -47,   8, -16,   3,   5,  -1,   3,   0, -28,  33,   0,  21,   5,  -5,  -1,   0,  14,  58,   4,  13, -13,  -1, -15,  -4,   3,   4, -37, -21, -23,  25,  -1,   1,  -5, -22, -20, -11,  -6,  29,   5,   7 },
     {    0,   7, -39,  38, -12,  10,  -6,   2,  -2,  -9,  44, -38,   8,  -8,   3,   1,  14,  -5, -23,  25,   0,   2,  -2,  -2, -23,  20,   5,  -9, -13,   0,   0,   2,  29, -30,   8,  -1,  14,  -4,  -1,   1, -28,  34, -15,   7, -26,   4,   2,  -1,  21, -28,   6,  -7,  12,   5,  -7,   4,  -4,   6,  11,  -7,  -4,  -5,   1,  -3 },
     {   -1,   2,  -8, -10,   4,   1,   0,   1,  -3,   0,  -7, -17,  10,  -3,   0,   1,  -1,   1,  -6, -12,   6,   0,   1,   1,  -6,  -3,  -8, -23,  14,  -3,   1,   1, -11, -14, -17, -22,  21,  -6,   3,   5,  -7,  -2,  -7, -13,  48, -11, -16,   3,   7,   7,   4,   1,  71,  28, -19,  -3, -10,  -9, -23, -21,  35,  36,   4,   3 },
     {   -8, -31, -59,  24, -11,   5,   0,   1,   8,  -4, -45,  38, -14,   9,   0,   1,  24,  42,  14,  33, -12,   5,   0,  -1,   1,  28,  12,   6, -24,  -1,  -1,   3, -23,   2,  -4, -18, -21,  -3,  -2,   5, -10,  -4,  -4, -13,  -2,   2,  -6,   3,  -8,   3,   2,  -2,   7,   8,  -3,  -1,  -1,   4,   5,   2,  10,   5,  -3,   0 },
     {    5,   8,  12,  16,   6,  -4,  -3,   0,   9,  15,  22,  23,  11,  -5,  -4,  -2,   7,  14,  18,  20,   7,  -6,  -3,  -2,   3,  14,  18,  20,  12,  -9,  -3,  -2,   5,  16,  21,  23,  13, -15,  -9,  -2,  14,  30,  29,  30,  21, -10, -18,  -8,  19,  33,  31,  32,  26,  -7, -16,  -9,  14,  22,  22,  21,  21,  -6, -11,  -6 },
     {   -4,  -7, -52,   7,  -2,   3,   4,   2,  -2,  26,  47,   1,  -3,  -1,  -7,  -3,  -7, -43,   6, -32,  11,  -4,   4,   1,  33,  23, -24,  20,   9,  -1,   0,  -4, -54,   6,  10,   7, -20,  -1,  -2,   2,  24,  -1,  -5, -13,   7,   8,   0,   3,  -3, -11,   7,   4,   5, -10,   0,  -1,  -1,   4,   1,  -7,  -2,   1,   4,   1 },
     {    0,   1,  -3, -10,  22, -12,   1,  -3,  -1,   0,   0,  11, -31,  18,  -2,   2,   1,   0,   1, -12,  32, -16,  -1,  -2,  -1,  -1,   0,   8, -31,  18,  -1,   4,   2,  -1,   1, -11,  39, -15, -11,  -2,  -4,   2,   3,  10, -36,  38,   7,  -4,  -4,  13,  -9,  -1,  32, -52,  12,  -4,  10, -18,  16, -14,   1,  36, -16,   6 },
     {   -5, -11, -54,  10,  -4,  -1,   1,   2,   5,  28,   6,  19,  -7,   1,  -3,   1, -20,   3,  49, -26, -11,  -2,  -6,   3,  -6, -45,  -2, -39,  21,   2,   1,   2,  46,  -2,   2,   6,  28, -11,   4,  -6,   8,  -9,  -2,   8, -10, -22,   3,  -5, -12,   3,  -1,   1, -10,   3,   2,   0,   4,   7,   0,   1,  -1,   4,  -4,  -2 },
     {    2,   0,   1, -17,   4,   2,   0,   1,   1,  -9,   9,  16,  -5,  -5,  -5,   1,   7,  -5,  -4, -13,   7,   4,   1,   5,   4,  -1,  10,  15,  -5,  -2, -11,  -1,   9,  -4,  -2, -11,   3,  16,   1,   3,  22,  -8,  12,   4,  -1, -14, -11,  -1,  -3, -40, -17, -15,   4,  55,  -9,  11,  29, -22,  75,   8,   3,  19, -20,   5 },
     {    6, -13, -23,   0,  15,   4,   5,   3,   9, -20, -32,  -2,  20,  10,   1,   0,  22, -13, -28,   6,  18,   2,   1,  -1,  50,  19, -11,  20,  20,   3,  -3,  -5,  55,  31,  -2,  22,  10,  -6,   3,  -8,  31,  15, -12,   9,   0, -13,  -2,  -1,   6,  -8, -20,   0,   0,  -5,  -2,   0,  -8, -15, -23,  -2,   2,  11,   6,   2 },
     {   -1,  -4,  -5,   6,   1,   0,  -7,   2,   1,   3,  -6,  -3,   8,  11,  -6,   6,  -2,   3,   6,   7,   6,   0, -11,   5,  -2,  -5,   0,  -8,  10,   9, -11,  10,  10,   0,   8,   3,  16,  -1, -42,   4,  17,  -6,   4,   4,  33,  88, -25,  -9,  -3, -17,   3,   5,  -6,  34,  30,  -4,  -5,   3,  -4,  -3, -11, -21,   9,  -2 },
     {    2,   1,   4,  57,   9, -10,  -5,   3,  -3,   4, -12, -51, -22,  16,   3,   1,   2,   2,  24,   2,  44,  -6,  -4,   0,  -4,  -8, -18,  20, -30,  -6,  11,  -1,  -5,  16,   8, -15,  17,  33, -12,  -1,  13, -18,   7,   5,  15, -31,   7,   2,  -3,  12,  -6,   4, -13,  18,   5,  -5,   2,  -4,  -3,  -5,   7, -13,  -3,   3 },
     {    5,  -1,  30,  51, -32,   1,   0,  -1,   2, -11,  21,  46, -43,   8,  -1,  -4,  -3, -26, -18,   7, -35,   8,   6,   1,   7, -16, -32,  -6, -20,  20,   4,   3,  15,   2, -24,  -1,  -4,  13,   1,   1,  20,   9, -12,   6,  12,   6,  -9,  -1,   2,  -2,  -9,   9,  10,   0,  -7,   0,  -6,  -3, -11,   7,  10,  11,  -1,   1 },
     {    0,  -1,   3,  23, -22,  10,  -7,  -3,  -1,   2,  -8,   2,   7,  -7,   0,   2,  -6,  -2,  23, -33,  15,  -1,   6,  -1,  21, -23,   0,  29,   9,  -8,  -1,  -6,   9,  27, -22,  18, -24,  -4,  -1,  -5, -75,  -1,  14,  -1,  -6,  22, -10,   6,  37,   2, -12,  -6,  11,   8,  -4,   2,  -9, -19,   8,  10,   1,   3,   1,   5 },
     {   -1,  -1,  -8,  -5,   0,  19,  -6,   3,   1,   2,   0,   6,   6, -22,   1,   2,   1,   4,   2,  -1,  -6,  33, -11,   5,  -3,  -1,  -1,  -3,   9, -34,  -9,  12,   3,   1,   3,  -2,  -4,  67, -48, -13,   2,   7,  -2,  13,  10,   9,  51, -17,   5,   7,   3, -10,   7, -22, -10,   2,   1,  -3,  -4,  16,  -7,  34,  -2,   2 },
     {    3,   6,  15,  37,  48,  13,   2,  -3,   1,   5,  20,  47,  47,  23,  -1,  -8,  -5,  -3,   9,  33,  30,   5,  -9, -11, -14, -14,  -3,  11,  22,  -3,  -9,  -5, -16, -24, -12,  -1,   1,  -8,  -6,  -7, -11, -21, -18,  -6,  -3, -10,   1,  -1, -12, -17, -15, -10,  -4,  -5,  -1,   7,  -5,  -9,  -9,  -3,  -2,   8,   5,   2 },
     {    0,  -4,  10,  15, -43,  20, -20,   5,  -1,   3, -12, -16,  51, -32,  30, -11,   1,  -9,  17,   5, -25,   5, -19,   6,   1,   5, -22,  25,  -2,   8,   6,  -3,  -8,   3,  17, -27,  22, -24,  13,   4,  17, -23,  -3,  18, -16,  10,  -8,  -3, -19,  13,  -6,  -9,   5,  -8,   7,   9,   8, -10,   3,  11,  -3,  21, -11,  -3 },
   },
   {
     {   83, -58,   7,  -3,   1,  -1,   0,   0, -65,  37,   5,  -2,   1,   0,   1,   0,  13,   5, -13,   3,  -1,   0,  -1,   0,  -2,  -8,   5,   2,  -1,   1,   0,   0,   1,   3,   0,  -2,   1,   0,   0,   0,  -1,  -1,   0,   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   52,  17, -37,   4,  -5,   1,  -2,   0,   7, -65,  42,   0,   2,   0,   1,   0, -45,  52,  -1, -15,   5,  -3,   1,  -1,  15,  -5, -20,  12,  -1,   0,   0,   0,  -5,  -3,  10,   0,  -3,   1,  -1,   0,   2,   2,  -3,  -2,   1,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {  -34,  49, -42,  13,  -4,   3,  -1,   1, -67,   3,  37,  -1,   1,   1,   1,   0,  53,   7, -23,  -7,   1,  -1,  -1,   0,   9, -31,   7,  10,  -1,   1,   0,   1,  -3,  10,   7,  -9,   0,   0,   0,   0,   3,  -3,  -6,   2,   2,   0,   0,   0,   1,   0,   2,   0,  -1,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   0 },
     {   46,  42, -30,  -7,  -5,  -1,  -2,  -1, -12, -20, -18,  26,  -5,   5,  -2,   2,  29, -56,  55, -18,   0,   1,   0,   0, -25,  32,  -9, -15,  13,  -6,   3,  -2,  -1,   0, -11,  15,  -4,   0,   1,   0,  -2,   1,   4,  -2,  -3,   2,  -1,   0,   1,  -2,   0,   0,   1,   0,   0,   0,  -1,   1,   0,  -1,   0,   0,   0,   0 },
     {  -26,  12,  31, -24,   4,  -3,   2,  -1, -57,   8, -32,  32,   1,   3,   0,   1, -27,  31,  35, -17,  -6,   0,  -1,   0,  42,   7, -43,   0,   7,  -1,   0,   0,   5, -29,  16,  15,  -8,   1,  -1,   1,   2,   8,   4, -14,   2,   1,   0,   0,   1,  -3,  -3,   4,   2,  -1,   0,   0,   1,   1,   1,  -1,  -1,   0,   0,   0 },
     {   25,  38, -15,   4, -11,   0,  -2,   0,   4,  20, -39,   1,   5,   0,   0,   0, -36, -16,   5,  30, -12,   5,  -2,   2,  59, -43,  39, -25,  -2,   4,  -3,   1, -30,  14,  -5,  -9,  17,  -7,   3,  -1,   4,   4,  -6,  10,  -5,  -1,   1,  -1,  -4,  -1,   2,  -1,  -1,   2,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {  -16, -21, -43,  62, -16,  10,  -5,   3,  -4,  43,  20, -35,  -6,   1,  -1,   0, -34, -27,  28, -17,  25,  -7,   4,  -1,  27,  25, -19,   6,  -5,  -4,   2,  -2,   1, -17,  -5,   7,  -4,   5,  -1,   1,  -2,   3,   8,  -2,   1,   0,  -1,   1,   0,   0,  -4,  -2,   1,  -1,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0 },
     {  -14,   8,  14,   5, -11,   1,   0,   0, -43, -26,  17, -27,  28,  -4,   5,  -1, -25,   9,  -2,  43, -19,   1,  -1,   0, -13,  44,  16, -42,  -1,   6,  -3,   0,  14,   2, -47,  19,  17,  -8,   2,  -1,   8, -16,  19,  10, -18,   4,   0,   0,   1,   4,  -1, -10,   6,   1,  -1,   0,   1,  -1,   0,   3,  -1,  -1,   0,   0 },
     {   20,  47, -17,  -9,   3,  -5,   0,  -2,   3,  30, -17,  -4,  -5,   1,   0,   0, -18,   2, -35,  25,   2,   1,   0,   0, -22,  16,   4,  15, -21,   7,  -3,   1,  60, -56,  16, -19,   4,   7,  -4,   3, -22,   1,  13,  -5,  12,  -9,   3,  -1,   2,   5,  -7,   2,  -3,   1,   2,  -1,  -2,  -2,   1,   2,   0,   1,  -1,   0 },
     {   17,   5,  14,  22, -36,  11,  -6,   2,  20,  -9, -14, -39,  43, -10,   3,  -1,  49,   5,  -3,  24, -21,  -7,   2,  -3,  10, -12, -36,  -4,   1,  12,  -4,   2, -19, -32,  29,  25,  -1,  -4,   1,   1,   2,  17,  14, -26,  -5,   3,  -1,   0,  -5,  -2, -13,   5,  11,  -2,   0,   0,  -1,   1,   4,   3,  -5,  -1,   1,   0 },
     {   -1,  -7, -20, -38,  52, -18,   9,  -3,  11,   8,  43,   2, -25,  -2,   0,  -2,   7, -28, -25,  31, -18,  18,  -6,   3,  35,  29, -20, -15,   7,  -3,  -3,   1, -36, -26,  11,  12,   2,  -3,   5,   0,  13,  11,   7,  -7,  -3,   0,  -1,  -1,  -2,  -4,  -6,  -1,   4,   1,   0,   0,  -1,   2,   1,   3,  -2,   0,   0,   0 },
     {    0,  -2, -35,   2,   1,   7,   1,   1,  23,  36,  -3,  22,  -6,  -9,  -2,  -2,   0,  33, -13, -17, -26,  16,  -3,   3, -12, -26,   0, -22,  48,  -9,  -1,   0,  22,  -1, -21,  45, -22, -14,   9,  -3,  -1, -26,  26, -13, -15,  21,  -6,   1,  -3,   7,   1,  -9,  15,  -6,  -3,   1,   0,  -2,  -1,   3,  -2,  -2,   3,   0 },
     {   16,  45,  28, -21,  -4,  -2,  -3,  -1,   0,  32,   1, -26,   9,  -7,   1,  -2, -19, -10, -23,  -5,  19,  -6,   6,  -1,  -9,  11, -26,  41,  -8,  -6,   4,  -4, -29,  34, -13,  17, -29,  15,  -5,   1,  36, -37,   9,  -7,  -1,  11,  -8,   5,  -5,  -6,  12,  -8,   9,  -7,   1,   1,   1,   2,  -3,   1,   0,  -2,   3,  -2 },
     {  -14,  -4,  -8,  -4,  -1,   3,   2,   0, -22,  -2,  12,  -1,  16, -10,   2,  -1, -50, -33,   5,  16, -21,  12,   2,   1, -46, -15,  11,   9,  22,   6,  -4,   3,   1,  28,  51,   3, -16, -14,   2,  -1,   5,  28, -15, -39,   2,   7,   1,  -2,  -2,   1, -20,  16,  13,  -4,  -1,   0,   2,   2,   3,   3,  -9,  -1,   1,   0 },
     {    2,   5,  15,  27,  24, -32,   0,  -3, -11, -24, -17, -17, -43,  44,   1,   3,   1,   4,  -6,  11,  39, -16, -11,   4,  -9, -25,  20, -14,  -5, -13,  20,  -3,  -3,  -3,   3,  33, -10,  18,  -8,  -4,   2,  13,  26, -29,  -7,  -3,  -7,   6,  -7,   6, -19,  -8,  22,  -2,   4,  -1,   1,   3,   0,  12,  -8,  -5,   1,  -1 },
     {   -1,   3, -27,  28,  15, -22,  -1,  -2,   1,  20, -30,  -8, -18,  20,   6,   0,   4,  43,  12,  25, -17,   0, -13,  -1, -37,  18, -25, -20,  -1,   5,   2,   4, -16,   9,  20, -10,  13,   1,  -2,  -2,  42, -14, -39,   9,   2,  -6,   3,   1,  -5, -24,  20,  16,  -9,   3,   1,   0,   2,   5,   6, -11,   2,   2,  -2,   0 },
     {   14,  36,  37,  -4,   3, -11,  -4,  -1,   9,  31,  34, -33,  -9,   1,   4,  -1,   0,  17,   6, -19,   9,  -4,  -3,  -2,  -5,   0,  -4, -28,  24, -13,   6,  -1,   3,   6, -15,  10,   1, -14,   9,  -6, -38,  44, -41,   9,  -6,   7,   2,  -3,  18, -18, -11,  23, -11,   6,  -6,   2,  -1,  -6,   9,  -9,  -1,   3,  -2,   2 },
     {    2,  -2,  -1,   1,  38, -34,   8,  -4,   2,  -6,   2, -38,   6,  11,   6,  -1,   5, -10,  39,  23, -32,  10, -10,   1,  13, -32, -38,  24,   2,   4,  -1,   4,  34,  33,  -7, -15,  -4,  -2,  -2,  -2, -39, -26,  13,   7,   5,   2,   1,   2,   8,  14,   7,  -6,  -1,  -2,   0,   0,  -2,  -4,  -5,  -1,   4,   0,   0,   0 },
     {   -1,   1,  -4, -12,   2,  -9,  10,  -2,   3,  12,   5,   6,  -6,  16, -16,   5,  -5,  -5,   0, -12,  11, -33,  21,  -5, -13, -12, -18,   8, -16,  44, -11,  -2,   0,   4,  12,  12,  40, -43,  -4,  10,  -1,   4,   4,  12, -55,   7,  23, -11,   2,  -5,   3, -25,  22,  23, -21,   4,  -3,   5,  -5,  10,   2, -17,   7,   4 },
     {   -4,  -9,  -9,  18,   7,  -6,  -3,  -4, -17, -32, -42,  -8,   5,   1,  12,   3, -10, -23, -45,   2,  15,  23,  -2,   1,   0,   2, -12,  30,  28, -16,  -4,   3,  -5, -13,   3,  26, -11, -16,  16,  -4, -17,  12,  -7,  39, -22,  15,  -1,  -6,  19, -16,  27,   0, -22,   9, -12,   5,   1,  -3,  10, -16,   9,   0,  -1,   1 },
     {   -7,   5, -14,  -2,  10,   0,   4,   0,  -7,  16, -12,  -1,   5, -14,   3,  -3, -10,  27,  11,  -5, -10,   6,   0,   1, -44,   0,  -2,  13,  -9,  -4,   2,   1, -71,  -6, -10,  -6,  17,  13,   3,   3, -40,  26,  41,  16,   6,  -9,  -5,  -1,  15,  27, -15, -18, -12,   0,   6,  -1,   3,  -4,  -9,   6,   4,   5,  -2,  -2 },
     {  -10, -18, -44, -38,   0,   8,   7,   5,   3,   3, -14,   9,  31,  10,  -9,  -5,   0,   5,  -8,  28,  17, -51,   9,   1,   5,   2, -10,  -9, -17,   8,  29, -10,   2,  21, -13,  27, -19,  22, -22,  -4, -32,  10, -16,  -2,   3,   6, -10,  12,  16, -12,   3,  22,  -3,  -6,   4,  -1,   2,  -4,  12,  -6,  -4,   1,  -1,   0 },
     {    0,   0,   1,   6, -32, -41,  69, -23,   2,   4,   5,   3,  28,  33, -58,   8,  -3,  -2, -11, -12, -17,   8,  -5,  22,   3,  10,   4,   6,  13, -23,  18, -14,  -5,  -6,   0, -14,   0,   5,   1,  -3,   5,   3,  -1,   3,  -2,   1,  -1,   1,  -4,   2,  -4,  -2,  -3,  -1,  -1,   1,   1,   0,  -2,   1,   1,   1,   0,  -1 },
     {    5,   5,  -9,  17,   7, -10,   1,   0,   7,   7, -18,  12,   3, -13,   4,  -2,  21,  32,   2,  13,  -2,   9,   2,  -2,  29,  55,  24,  31, -10, -12, -13,   4,  12,  47,   7,  -3, -22, -16,   2,  -3,  -7,  26,  -3, -25, -30,  -3,  -1,  -3,   5,  -6, -31, -25,   0,  13,   0,   3,  -2, -10,  -7,   9,   9,   0,  -4,   2 },
     {   -2,  -5, -18, -19,  36, -29,  20,  -7,   1,   1, -10, -30,  41, -32,  15,  -8,   3,   5,  21, -34,   8,  12,  -4,  -1,  -5, -18,  30,  11, -21,  13,  -3,   5,  16, -23, -18,  25,  -1,  11,   2,  -1,  40,  28, -11,   6,  -4,  -9,  -5,  -5, -24,  -8,   0,   0,  -7,   4,   1,   2,  -3,   3,   0,   3,  -2,   3,  -1,   0 },
     {    2,   4,  14,   1, -10,   7,  -4,   5,   6,  11,  32,  14,  -1,   3,  -2,  -2,  -2,  -1,  24,  22, -10, -11, -11,   3,  -6, -14,  14,  -1,  -7,  -7,  17,  -1,  13,  -5,  32,  19, -14,  16,  -8,  -8,  16,  20,  19,  11, -17,  -6, -13,   5,  15, -13,  34, -50, -35,   2,   1,   2,  -6,   2,  -8, -41,  36,  10,  -1,   0 },
     {    2,   2,   2,   5,  -8,  -1,  -2,  -2,   3,   3,   1,  12,   0,  -5,   5,  -1,  -5, -11, -22,  -8,   1,  12,  -7,   3,   4,  -7, -26,  -1,  17,  10,  10,  -4,  14,  28,  -4,  27,  31,   0, -29,  -1,  10,   2,  14,   3,  -7, -69,  17,  11,  -4,   0, -22,   9, -35,  19,  33, -16,  -2,  -1,   6,   6,  -9,  24, -22,  -4 },
     {   -5,  -7, -18, -42,  -4,  20, -18,  10,  -2,  -2, -13, -36,   1,  17, -12,   5,   3,   5,   8,   9,  21, -15,  21, -12,  -3,  -8, -13, -10,  36, -40,   3,  -4,  14,   2,   9, -31,   1,   5,  27,  -1,  27,   5,   8,  19, -14,  -5,  -5, -13, -22,   7, -34, -16, -14,  11,   3,   8,   3,  -9, -12,  15,  17,   9,  -9,   1 },
     {    0,  -4,  -2,   4,  -3,   7,  -1,   2,  -1,  -7,  -1,   5, -13,   5, -13,  -1,  -4,  -4,  -1,   0, -21,   8,  13,   1,  -6, -22, -12,  10,  -9,   2, -18,  -3,   5,  -1, -36, -15,   7,  20,  10,   7,  21,  15,  19, -12,   6,   3,  -9,  -4,  50, -73, -28,   0,   6,  -2,   1,   9, -31,   4,  22,  11,  11,   1,   7,  -5 },
     {   -4,  -6,  -7, -20, -28,   4,   0,   0,   2,   4,   3,   5,   7,  -6,  35, -14,  -3,  -5,  -1,  -5,  11,  18, -66,  17,  -4,  -4, -16,  -1, -19, -23,  30,  15,   4,   1,   5, -13,  24,   7,  16, -35,  -6,  -5, -10,   0, -25,  -2, -13,  15,   7, -11,  -4, -14,  32,  13,  -2,   4,   0,   6,   5,  20,  -4, -24,   3,  -2 },
     {   -3,  -5,  -5,  -9, -14, -48, -42,  52,   3,   4,   3,   2,  24,  48,  26, -24,  -5,  -7, -13, -31, -34,  -9,  11, -19,   5,  11,  17,   8,   9,  -6,   3,  13,  -9,  -9,  -4,  -9,   2,   3, -16,  -6,   0,  -2,   8, -10,   2,   1,  13,   4,   0,  -1,  -1,   2,  -2,   5,  -2,  -4,  -2,  -2,   4, -10,   7, -12,  -2,   4 },
     {   -1,  -2,  -9, -16,  -5,   9,   7,  11,  -6,  -9, -24, -37, -13,   2,  -7,  -9,  -1,  -2, -23, -35,   3,  16,  -7,  22,   7,  15,  -6, -33,  13,  43,  -2,   6,   3,  15,  12,  -6,   4,   3, -19,   4, -11,   6, -13, -17,  20,  -9, -15,  -3,  10,  -5,  10, -43,   5, -30,  10,  10,  -8, -15, -28,  -6,  21,   1,  21, -10 },
   }
 },
 {  //13
   {
     {  118, -38,  -6,  -5,  -2,  -2,  -1,  -1, -24,   1,   3,   1,   1,   0,   0,   0, -15,   6,   1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   20,   2,  -3,   0,  -1,   0,   0,   0, 113, -29,  -9,  -4,  -3,  -2,  -1,   0, -33, -12,   6,   1,   2,   1,   1,   0, -28,   9,   4,   0,   1,   0,   0,   0,  -3,   4,  -1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   18,  -7,  -1,  -1,   0,   0,   0,   0,  23,  14,  -7,   0,  -2,   0,  -1,   0, 109, -29, -12,  -2,  -3,  -1,  -1,   0, -28, -25,   8,   1,   2,   1,   1,   0, -26,   9,   6,   0,   1,   0,   1,   0,  -3,   5,  -1,  -1,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 },
     {   11,   9,  -5,   0,  -1,   0,  -1,   0,  34,  -7,  -4,  -1,  -1,   0,   0,   0,  20,  25,  -9,  -1,  -3,  -1,  -1,   0, 105, -26, -18,  -1,  -4,   0,  -2,   0, -12, -34,   6,   1,   2,   1,   1,   0, -17,   2,   7,   0,   1,   0,   0,   0,  -5,   2,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   30, 104, -39,  -1, -12,  -2,  -4,  -1, -14,   7, -11,   4,  -1,   2,   0,   0,  -8, -45,  12,   2,   4,   1,   1,   1,   5,  -8,  12,  -3,   1,  -1,   0,   0,   7,   2,   0,  -3,   0,   0,   0,   0,  -4,   0,  -2,   0,   0,   0,   0,   0,  -2,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
     {   -5,   8,  -2,   1,  -1,   0,   0,   0, -13, -23,   9,   1,   2,   1,   1,   0, -27,  12,   1,   1,   1,   0,   0,   0,  -3, -38,   8,   1,   2,   1,   1,   0,-100,  27,  24,   0,   4,   0,   1,   0,  20,  35, -10,  -2,  -2,  -1,  -1,   0,   7,  -4,  -5,   1,  -1,   0,   0,   0,   2,   1,   0,   1,   0,   0,   0,   0 },
     {   -7,  -9,   4,   0,   1,   0,   1,   0, -11,  -3,   2,   1,   1,   0,   0,   0, -13, -23,   8,   1,   2,   1,   1,   0, -21,   0,   4,   1,   0,   0,   0,   0, -21, -43,  10,   3,   3,   1,   1,   1, -89,  21,  32,   0,   5,  -1,   2,   0,   1,  52,  -3,  -6,  -1,  -2,   0,  -1,  12,   5, -11,  -1,  -1,   0,   0,   0 },
     {   -2,  -9,  -2,   3,   0,   0,   0,   0, -18,-103,  28,   7,   8,   3,   3,   1,  26, -33,  26,  -4,   3,  -1,   1,   0,  21,  30,   4,  -8,  -2,  -2,  -1,  -1,  10,  17,  -8,  -3,  -1,  -1,   0,   0,  -2,   5,  -4,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {   11,  44, -14,  -3,  -4,  -1,  -1,   0,  -2, -28,  21,  -2,   3,   0,   1,   0,  25,  76, -16, -11,  -3,  -4,  -1,  -1, -40,  33, -37,   6,   0,   1,   0,   0, -14, -33,  -5,  13,   0,   3,   0,   1,   2, -10,  13,   3,  -1,   1,   0,   0,  11,   3,   3,  -4,  -1,   0,   0,   0,   1,  -2,  -2,  -1,   1,   0,   0,   0 },
     {   -1,   2,  -1,  -1,   0,   0,   0,   0,  -3, -13,   7,   1,   1,   1,   0,   0,  -2,   9,  -4,   1,  -1,   0,   0,   0,  -6, -28,   9,   3,   2,   1,   1,   0,  -9,  16,  -3,  -1,  -1,  -1,  -1,   0,   3, -62,   6,   6,   1,   2,   0,   1, -75,  29,  41,  -5,   5,  -1,   2,  -1,  12,  48, -12,  -9,  -1,  -3,  -1,  -1 },
     {  -10, -26, -74,  38,  -2,   9,  -1,   3,   4,  14,   0,  10,  -7,   1,  -2,   0,   5,  19,  70, -34,  -1,  -5,   0,  -2,  -6, -19,   2, -19,   9,  -1,   2,   0,   1,  -8, -23,  11,   4,   0,   1,  -1,   5,   8,  -1,   8,  -3,  -1,  -1,   0,  -2,   2,   5,  -3,  -1,   0,   0,   0,  -2,   0,  -1,  -2,   1,   1,   0,   0 },
     {    0,  -2,  14,  -6,   1,  -1,   1,  -1,   2,  31, -14,   0,  -2,  -1,  -1,   0,   2, -29,  20,  -3,   2,   0,   1,   0,  22,  71, -19, -17,  -1,  -5,   0,  -2, -62,   4, -38,  17,   1,   3,   1,   1,   9, -32,  13,  16,  -4,   3,  -1,   1,   7,   0,  14,  -7,  -2,   0,  -1,   0,  -2,   6,  -2,  -3,   2,   0,   0,   0 },
     {    7,  15,  10,  -2,  -1,  -2,   0,  -1,  12,  25,  16,  -5,  -1,  -1,  -1,  -1,  13,  34,  15,  -6,  -2,  -3,   0,  -1,  18,  34,  20, -10,  -3,  -4,  -1,  -1,  14,  58,   6, -15,  -3,  -4,  -1,  -2,   4,  38,  -2, -15,   0,  -4,   0,  -2,  -2,  59, -17, -13,   0,  -4,   0,  -1,  25,  14, -31,  -3,  -1,  -1,  -1,   0 },
     {    5,  14,  40, -37,  16, -10,   3,  -2,   7,  16,  66, -49,  14, -10,   3,  -2,  -2, -12,  41, -31,   8,  -4,   2,  -1,  -4, -32, -10,  -1,   6,   2,   2,   0,   5, -16, -22,  14,  -1,   3,   0,   1,   6,   4, -11,  11,  -4,   1,  -1,   0,  -2,  -2,  -4,   4,  -3,   1,  -1,   0,  -5,  -2,   2,   1,  -1,   1,   0,   0 },
     {    0,  -6,   0,  -1,   3,   0,   0,   1,  -2,  -7,  -5,   1,  -1,   1,   0,   0,  -2, -28,   5,   5,  -2,   3,   0,   1,  -1,  -9,  -7,  10,  -1,   1,  -1,   1,   7, -41,  26,  19,  -3,   3,  -2,   1,  77,   2,  47, -21,  -8,  -3,  -2,  -2,  12,  41,  -8, -32,   2,  -4,   1,  -2,  -6,   5, -28,   0,   6,   0,   0,   0 },
     {    7,  16,  58, -20,  -2,  -4,  -2,  -2,  -8, -22, -75,  21,   5,   7,   2,   3,   6,  29,  41,  -6, -10,  -2,  -3,  -2,  -8, -25,  15, -22,   7,   0,   0,   0,   9,  -3, -28,  10,   3,  -1,   1,   0,  -1,  14,  12,   6,  -5,  -1,  -1,  -1,   1,  -5,   1,  -7,  -1,   0,   0,   1,  -1,   3,  -1,  -4,   1,   0,   1,   0 },
     {    8,  10,  35,  74,  -3, -13,  -9,  -2,  -3,  -2,  -5,  -9,  -6,   0,   0,   1,  -5,  -7, -28, -76,   4,  15,   8,   3,   4,   8,  19,   5,  15,  -3,  -1,  -2,   0,  -3,  12,  28,  -2,  -8,  -4,  -2,  -3,  -3,  -7,  -5,  -8,   0,   2,   0,  -4,  -2,  -2,  -6,   2,   3,   2,   1,  -4,   1,   2,   0,   4,   1,  -1,   0 },
     {    0,   1,  -9,  12,  -4,   3,  -1,   2,   1,  -8,  22, -15,   1,  -5,   2,   0,  -5,  10, -40,  23,  -6,   6,  -1,   1,   4, -24,  54, -21,  -1,  -2,  -3,   0,   0,  13, -61,  10,   5,  -1,   2,   1,  -2, -17,  32,  25,  -8,   2,  -3,   0,  31,  14,  -1, -26,  -4,  -1,  -2,  -1,  -9, -28,  -5,  10,   3,   3,   0,   1 },
     {   -1,   0,  -3,   3,  -5,   0,  -4,  -1,   0,   1,   2,   6,  -2,   1,  -2,   0,  -5, -11,  -7,  -1,   0,   0,   0,   1,   1, -24, -69,  22,  13,   2,   3,   1,  14,  60, -14,  36, -10,  -2,  -3,  -2, -22,  15,  46,  -8, -15,   2,  -5,   1,   1, -21,   7, -25,   5,   3,  -1,   1,  13,  -4,  -6,  -3,   6,  -1,   1,   0 },
     {   -8, -13, -40, -48,   0,   7,   6,   2, -10, -16, -48, -76,  -1,  13,   9,   3,  -3,  -3, -28, -42,  -4,   9,   8,   2,   3,   6,  -2,  -5,   7,   6,   1,   1,   0,   9,  -4,   3,   8,   3,  -2,   0,  -2,   4,  -7,  -7,   3,  -1,   0,   0,  -1,   8,  -6,  -5,   0,   1,  -2,   0,   0,   5,  -4,  -3,   0,   2,  -1,   2 },
     {    1,  -2,   2,   0,  -4,   1,  -1,   0,   1,  -5,  -2,  -2,  -1,   0,  -2,   0,  -1,  -2,   8,  -7,   8,  -1,   2,  -1,   2, -18,   1,   1,   7,  -2,   0,  -1,   4,   2,  16,  -6,  -3,  -3,  -2,  -1,   0, -41,  -9,   6,   4,   5,  -1,   2,  79,  -4,  13,   8,  -5,   0,  -2,   0,  76,  29,  -8, -15,  -9,  -3,   0,  -2 },
     {   -1,  -2,   1,  -5,  -4,   1,  -3,   3,   0,  -2,  -6,  10,   1,   2,  -1,   1,  -2,  -5,  -2,  -8,   2,   2,  -2,  -1,  -1,  -2,  -9,  -5,   6,   0,   4,  -2,  -6,  -6,  -4,   1,   5,  -2,   1,  -1,  13, -16,  -3,   7,   2,  -5,   2,   1, -41,  12, -27,  17,   0,   2,   3,  -1,  71, -81,  -7,  27,  -3,   6,  -4,   2 },
     {    0,  -3,  14,  21, -76,  53, -28,  14,   0,   2,  -8, -27,  43, -17,  12,  -5,   2,   0,   7,   7,  26, -30,  16,  -6,  -1,  -1,  -6,   8, -19,  12,  -6,   4,   0,  -1,   0,  -7,  -5,   8,  -6,   5,   4,   1,   3,   5,   6,  -1,  -2,  -1,  -4,   3,  -3,   1,   2,  -1,  -1,   1,  -2,   2,   4,   3,  -3,   1,   0,  -1 },
     {   -5,  -7, -12, -51, -34,   6,   3,   4,   7,   8,  18,  60,  31,  -2,  -4,  -5,  -5,  -4, -24, -29,   1,   0,   5,  -1,   2,   2,  15, -33,  -1,   8,   4,   0,   5,  -2,  12,  52,  12, -10, -10,  -1,  -4,   1, -22,  -9, -17,  -2,   4,   0,   4,   4,   5,  -6,   2,  10,   0,   0,  -4,   7,   1,   0,   3,  -4,   0,  -3 },
     {    0,   2,   1, -13,   2,   6,  -3,   3,   2,   6,  13, -14,   1,  -3,   0,   2,   4,  14,  24, -10,   0,   2,   2,   1,   2,  13,  34, -35,   4,   0,  -4,  -2,  -3,  12,  57, -15,  -4,   1,  -9,   0, -14, -27,  64, -21,   0,  -1,  -7,  -2,  -8, -47,  -2, -16,   6,   1,  -2,   2,  -3, -17,  14,   4,   0,   3,  -1,   0 },
     {    0,   2,   4,  -8,  -2,   1,   1,   3,   0,  -4,  -9,  12,  -6,   0,  -1,  -1,  -2,   5,  16, -16,  -2,  -7,  -1,  -2,   3, -14, -22,  11,  10,   1,   3,   1,   0,  33,  30,  -6,  -2,   1,  -3,  -3, -11, -54, -11,  17,  12,  -2,   2,   1,  30,  50, -12,   9,  -8, -11,   2,  -3, -59, -37,   4,  11,  -1,   4,   2,   1 },
     {    1,   1,   4,  12,  12,  -4, -17,  -7,   0,   0,   4,  14,  16,  -7, -20,  -1,  -6, -12, -24, -20,  -3,   0,  -8,   5,  -7, -13, -41, -70,   4,  13,   0,   3,   2,  -1, -13, -48,  10,  27,  -2,  14,   7,   4,   1, -10,  28,  12, -10,   5,   2,   7, -11, -16,   8, -18,  -4,   6,   0,  17,  10,  20,  -8,  -5,   6,   2 },
     {    0,   0,  17,  -7, -56,  29, -12,  -3,   0,   1,  23,  -4, -72,  52, -13,   5,  -1,  -4,   6,  -1, -35,  25, -16,   6,  -1,  -2,  -5,  -7,   7,   4,   0,   5,   1,  -6,  -4,  -4,  20,  -1,   4,   3,   2,   2,  -4,  -4,  10, -10,  -1,   2,   0,  -3,   2,  -1,   7,  -6,   6,   4,   0,   2,   1,  -5,   5,  -4,   3,   0 },
     {    1,   2,   2,   2,   9,   7,  -8,   3,   0,   0,  -1,  -8,   4,  -6, -18,   1,   0,  -5,  -9,  -4, -10,  -1, -13,  -1,   1,   5,  -5, -17, -11,  -6,  -9,   1,  11,   6,  25,  12,  16,   0,   1,   1,  14,  30,  13,  45,  14,  -3,   1, -10,  -4,  16,  60,   7, -38, -16,  -7,  -6,   8, -26,  33, -46, -17,   6,  -2,   2 },
     {    0,   0,  -1,  16,  -8, -30,   7, -19,   1,   3,  -1, -11,  36,  -8,  29, -20,   0,  -3,  15,   6, -62,  26, -15,   0,   0,   2,  -4,  10, -29,  57, -14,  19,  -1,   4,   7,  -2,  27,   0,  -4,   5,  -5,  -8, -10,  -8,  -4, -12,   3,   5,   6,  -1,  14, -12,  12,   6,   7,   3,   5, -13,  -3,  -1,   3, -10,   4,   1 },
     {    1,   2,  -1,  10,   9,   9,   0,  -3,  -1,  -3,  -2, -14,  -7, -10,  -9,  -9,   1,   2,   4,  24, -25,  -7,   3,   0,   2,   1,   3, -10,  15,   7,  -2,   6,  -1,   5,  -6,  41, -12,   5, -10,  -6,   4,  -7,   5, -35, -12,   5,  -6,   5,  -1,  27, -28,  30,  12,  -2,   5,   4,   8,  17,  73,   1, -42,   6,  -8,  -5 },
     {    1,   1,   2,  13,   0,   0,  -4,   8,  -2,  -2,  -1, -15,  -3, -16,   8,  -4,   0,  -3,  -3,  24,  -8, -16,   1,  -2,  -2,   3, -18, -32,  46,   3,  -2,   3,   6,  -1,  19,  -3,  46, -29,  11, -11,  -1,  -2,  -4,  28, -38, -35,  15,  -9,  -4,  -4, -21,   4, -30,  22,   9,  -1,  -3,  24, -12,  16,  13,  21, -14,  -2 },
   },
   {
     {   73, -38,   1,  -3,   0,  -1,   0,   0, -78,  35,   3,   2,   1,   1,   0,   0,  40,  -7,  -8,   0,  -2,   0,  -1,   0, -16,  -7,   7,   0,   1,   0,   1,   0,   5,   7,  -2,  -1,  -1,   0,   0,   0,  -2,  -4,   0,   0,   0,   0,   0,   0,   2,   2,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 },
     {   65,  -1, -10,  -2,  -3,  -1,  -1,   0,  -4, -41,  14,  -1,   3,   0,   1,   0, -59,  56,  -2,  -1,   0,   0,   0,   0,  43, -19, -16,   4,  -3,   1,  -1,   1, -18,  -9,  17,  -1,   2,   0,   1,   0,   4,  11,  -5,  -3,   0,  -1,   0,   0,  -4,  -3,  -1,   2,   1,   0,   0,   0,   1,   1,   1,   0,  -1,   0,   0,   0 },
     {   53, -24,   6,  -2,   0,  -1,   0,  -1,  42,   0, -17,  -1,  -3,   0,  -1,   0, -36, -38,  27,   0,   5,   0,   2,   0, -32,  61, -10,  -5,  -1,  -2,   0,  -1,  25, -29, -13,   9,  -2,   2,  -1,   1, -14,   4,  18,  -4,   1,  -1,   1,   0,   2,   4, -10,   0,   1,   0,   0,   0,  -2,  -2,   3,   1,   0,   0,   0,   0 },
     {  -42,  -8,  17,  -1,   3,   0,   1,   0, -50,  26, -11,   2,   0,   1,   0,   1,  -8,  17,  13,  -1,   1,  -1,   0,  -1,  36,  25, -38,   2,  -4,   0,  -2,   0,  20, -63,  19,  10,   0,   3,   0,   2, -15,  30,  14, -16,   2,  -3,   1,  -1,   7,  -3, -20,   7,   1,   1,   0,   0,   0,  -3,   9,   0,  -1,   0,   0,   0 },
     {   21,  -1,   3,  -1,  -1,   0,  -1,   0,  40,   6, -18,   1,  -2,   0,  -1,   0,  37, -26,   3,  -1,  -1,   0,  -1,   0,  -5, -39,  -3,   4,   2,   1,   1,   1, -32, -11,  50,  -2,   3,  -1,   1,   0, -14,  60, -16, -23,   2,  -5,   1,  -2,   3, -15, -30,  20,   2,   3,   0,   1,   2,  -5,  19,   0,  -5,   1,  -1,   0 },
     {   16,  13, -15,   3,  -2,   0,  -1,   0,  27,   8,   4,  -3,  -2,  -1,  -1,  -1,  34,   0, -24,   5,  -3,   1,  -1,   1,  25, -28,   0,   0,  -2,   0,   0,   0,  -5, -44,  -8,  14,   1,   4,   0,   2, -13, -20,  65,  -1,  -3,  -1,  -1,   0, -12,  46,  -8, -35,   6,  -6,   2,  -3,  -4,  -5, -27,  15,   6,   0,   2,   0 },
     {  -15, -11,  -9,  10,   0,   2,   0,   1, -18,  -9,  27, -15,   3,  -2,   1,  -1, -24,  -2, -12,  16,   0,   3,   1,   1,  -8,   6,  36, -18,   2,  -3,   1,  -1, -11,  26, -22,  11,   0,   1,   1,   0,  12,  21,  26, -23,  -3,  -5,  -1,  -2,  -9,  19, -72,   6,  15,  -1,   4,  -1,  14, -30,  26,  23, -15,   5,  -4,   2 },
     {   13,  74, -20,  -2,  -6,  -1,  -3,   0, -31, -63,   9,   8,   4,   3,   2,   2,  39,   2,  26, -11,   1,  -4,   0,  -2, -35,  21, -22,  -2,   2,  -1,   1,   0,  11,  -7,   3,  13,  -4,   3,  -2,   1,   1,   4,   6, -12,   1,  -2,   0,   0,  -1,   2, -10,   3,   2,   0,   1,   0,   2,  -4,   3,   1,  -2,   1,   0,   0 },
     {  -14, -14,  17,  -8,   3,  -1,   1,   0, -21, -17,   5,   9,   1,   1,   0,   1, -22, -14,  26, -11,   3,  -1,   2,  -1, -19,   0,  15,   8,   2,   1,   0,   1,  -9,  19,  35, -20,   2,  -3,   1,  -2,  -5,  36,   1,  -6,   1,  -3,   0,  -1,   5,  35,  -4, -39,   2,  -5,   0,  -2, -14,  10, -73,  13,  23,  -2,   6,  -1 },
     {    9,  23,  13,   0,  -2,  -2,   0,   0, -18,   5, -37,   0,   0,   2,   0,   1,  -1, -39,  41,  12,  -1,   1,   0,   1,  49,  12,   2, -32,   4,  -6,   1,  -3, -63,  11, -25,  20,   4,   1,   2,   1,  29,   3,   9,   8, -12,   2,  -4,   1,  -4, -14,   7, -12,   5,   0,   1,   0,  -2,  10,  -3,   2,   1,  -2,   1,  -1 },
     {  -16, -76,  20,   1,   4,   0,   3,   1,  15, -51,  23,  -4,   4,  -1,   1,   0,  50,  27,  23, -11,   1,  -4,  -1,  -2,   0,  15, -23, -13,   2,  -2,   1,  -2, -31,  -2, -17,  13,   1,   2,   1,   1,   1,  -1,   6,   8,  -3,   1,  -1,   0,  -2,   3,   7,   3,  -1,   0,  -1,   0,   2,   2,   4,  -1,  -3,   0,   0,   0 },
     {  -14, -21,  -5,  10,   0,   3,   0,   2, -26, -43,   4,   9,   0,   4,   1,   2, -25, -49,   2,  16,   1,   5,   1,   3,  -9, -34,  27,  12,  -1,   3,   0,   2,   2, -21,  28,  10,  -2,   2,  -1,   1,  -4,  -7,  37,   0,  -5,   1,  -1,   1,   3,   5,  40,   2,  -8,   0,  -2,   0,  10,  19,  41, -26, -14,  -3,  -4,  -2 },
     {    4,   8,  -9,  -1,   2,   1,   0,   0,  -6, -20,  12,   3,  -1,   0,   1,   1,  12, -21,   3,  -2,   5,   2,   0,   0,  44,  15,  36,  -8,  -2,  -4,  -3,  -2,   2, -13, -27, -42,  12,  -3,   5,  -2, -60,  17, -23,  37,   8,   0,   5,   0,  19,  26,   0,  26, -20,  -2,  -5,   0,   4, -21,  -2,  -7,  -3,   6,  -1,   2 },
     {    1,  -4,  14,   7,  -3,  -2,   0,  -1,   1,  -3, -19,  -9,   6,  -1,   2,   1,   2,  -6,  23,  10,  -6,  -1,  -2,   1,  20,  -6,  -5, -15,   9,   1,   2,   0,  16,  17,  19,   5,  -5,  -6,  -4,  -2,  -9, -43, -18, -51,  24,   4,   8,   0, -33,  48,   0,  55,  -3,  -7,   4,  -2,  10,  15, -19,   8, -20,   2,  -4,   1 },
     {   -1,  -5,  -1,  -2,   1,   0,   0,   1,  -3, -10,  -5,  12,  -1,   3,   0,   0,   0, -11,   2,  -7,   1,   0,   1,   0,   6,  -3,  -2,  13,  -2,   2,  -1,   1,   6,  -6,   4, -12,   3,  -3,   0,   0, -10, -11, -14,  11,  10,   3,   1,   1,  -6, -18,   5, -23,  -6,   9,   4,   3,   6,  20,  18, 108, -33, -11,  -5,  -2 },
     {    8, -11,   8,  -6,   2,   0,  -3,  -1,  11, -25,  11,  -2,  -3,   2,  -2,  -1,  27, -19,   9,   1,   1,   2,  -2,   1,  56,   6,  28,  -4,  -8,   1,  -3,  -2,  64,   9,  13, -16,  -4,  -5,  -5,  -1,  31,  -8,  -2, -27,  -4,  -4,  -3,   0,  -3, -38, -16, -26,  18,   3,   2,   1, -12,  -4,   2, -10,  21,  -3,   4,  -1 },
     {    4,   9,  68, -19,  -1,  -6,  -1,  -3,  -1, -22, -56,   8,   7,   5,   1,   1,   0,  35,  -3,  24, -12,  -1,  -4,  -1, -19, -31,  22, -24,   2,   3,   1,   1,  29,   6, -19,  12,  11,  -5,   3,  -2, -26,   7,  10,   9, -15,   1,  -2,  -1,  13,  -2,  -5,  -7,   2,   4,  -1,   1,   0,  -8,   7,   3,   1,  -2,   0,   0 },
     {   -1,  -1,  11, -46,  31, -10,   6,  -2,   1,   2, -16,  61, -37,  12,  -5,   1,  -5,  -1,  10, -49,  24,  -4,   3,   1,   8, -10,  -1,  32,  -6,  -5,   4,  -4,  -8,  13,  -9,  -7, -11,   9,  -5,   3,  12, -16,  15,  -1,   7,  -4,   2,  -2, -11,  17, -17,   7,  -3,  -1,   0,  -1,   9, -22,  19, -10,   0,   5,  -2,   2 },
     {   10,  24,  77,  -9,  -1,   1,  -2,  -2,   0,   3,  64, -27,   4,  -1,   4,   0, -11, -40, -19, -27,   2,   1,   5,   0,   7, -17, -30,   5,  12,  -2,   4,   1,   2,   1,  -9,  17,  -4,  -3,   0,  -2,  -2,  -1,  -5,   6,   0,   3,  -2,   0,   1,  -1,   1,   5,  -1,  -3,  -2,   1,   5,  -5,  -1,   2,  -4,   0,   1,   0 },
     {  -10, -23, -28, -12,   8,   1,   0,  -1, -15, -39, -45, -15,  12,   1,   5,   0, -10, -38, -49,  -5,  10,   3,   5,   1,   3, -19, -37,  11,  15,   4,   6,   1,   9,  -4, -23,  19,   8,   0,   2,   2,   3,  -3, -20,  13,  10,  -1,   1,   0,  -5, -14, -22,  13,  16,   1,   3,   1, -10, -17, -32,   0,  13,   4,   4,   2 },
     {   -5, -10, -41,   4,  -5,   3,  -1,   2,   4,  19,  18,  13,  -4,   5,  -4,  -1, -12,  -4,  51, -29,  -4,  -6,  -2,  -1,   5, -46, -18, -35,  24,  -3,   8,   1,  34,  21, -13,  36,   5, -10,   2,  -5, -34,   8,  -1,   7, -25,   9,  -3,   2,   5,  -5,   0,  -6,   7,   7,  -4,   3,   5,  -4,   0,   2,   6,  -6,  -1,   0 },
     {   -6,   1,   3,  -7,   4,   0,   5,   3, -10,  -1,   9,  -8,  -5,   4,  -3,   2, -13,   0,   8, -18,   2,   2,   0,   3, -25, -17,  11, -29,  -3,   0,  -1,   1, -31, -39, -16, -41,  17,  10,   4,   2, -33, -38, -17, -34,  11,  11,   6,   4, -27, -27, -15, -17,  31,   9,   9,   4, -13,  12,   9, -11,  26,  -2,   6,  -2 },
     {    0,   0,   1,  -1,  -1,   2,   1,   0,  -1,  -2,  10,   6, -12,   5,  -5,   1,  -3,  -1,  -1,   2,   1,   4,  -3,   3,  -7,  -7,  12,  -2, -13,   3,  -2,   0,  -2, -17,  -5,  -3,  -4,  11,  -1,   2,  18,  -3,  12,  -6, -55,   7,  -2,   4,  12,  -8,   4,  67, -16, -21,  -7,  -4, -17,   6, -10,  41,  61, -13,   6,  -5 },
     {    1,  -1,   4,  17,   1,  15,  -2,  -5,  -1,   4,  -6, -30,  -4,   9,   6,   6,  -2,   3,  37,  -6,   4,   5,   1,  -1, -15, -34,   0,  -7, -18,  -9,   3,   4,  21, -22, -11, -21,  21,  13,   2,   4,  45,  10,   1,  40,  30, -27,  -2, -10, -38,   1, -12,   5, -37,  -4,   3,   1,  20,  -7,  -4,   0,   8,  14,  -6,  -3 },
     {    3,   5,  -1,  22,  56,  -6,  -8,  -7,  -3,  -6,   4, -25, -73,  -9,   4,   1,   3,   2,  -7,  28,  49,   4,  -4,  -7,  -2,  -1, -11, -15, -27, -13,   3,  -1,   4,  11,   5,  18,  14,   5,   0,  -3, -17,  -1,  -9, -10,  -8,  -2,  -4,  -1,  10,  -2,   3,  -9,  -9,   6,   1,   1,  -5,   9,  -8,   0,  -7,  -2,  -4,   0 },
     {    3,  -2,  18,  78, -44,   6,  -2,  -3,  -1,  -8,  -9,  52, -36,   6,   6,  -2,  -3,  -1, -32, -31,   3,  -2,   2,   3,   0,   7,   2, -27,  15,  -2,  -1,   4,   2,  -4,   2,  -1,  16,   3,   1,   0,  11,   2,  -3,  -4,   7,  -5,  -4,   0,   1,  -4,   0,   1,   1,  -3,   2,  -1,  -3,  -7,  -9,  -2,  -5,   4,  -1,   1 },
     {    2,   2,   6,   9,   3,  -1,   5,  -1,   0,   2,  -2,  -1,  -1,  -7,  -3,   4,   3,  -1,  10,   0,   7,  -5,   5,   0,  -2,   2,  -6,  14, -14,   3,  -6,   3,   8,  -4,  13, -19,  26,  -6,   8,  -3,   8,  -1,  -6,  38, -54,  -3,  -2,   1, -23,  15,  -3,  15,  82, -23,   8,  -6,   6,   8,  -1,  -5, -33,   9,  -3,   2 },
     {    2,   5,  -3,  -9,   9,  -3,   0,  -5,   3,   1,   7,  13,   4,  -1,   2,  -1,   4,   7, -20,  -9,   9,  -2,  -5,   1,   6,   5,  12,   0,  13, -10,   6,  -4,  12,   9, -33,   6,  -2,   0,   3,  -2, -10,  43,  19,   2,  22, -17,  -1,  -7, -32,  -9,  -5,  18,  10,  -7,   2,  -6,  -5,  91,  -1,  -9,   3, -13,  -2,  -8 },
     {    1,   1,   7,  31, -35,  17,  -7,   9,  -2,   2, -21, -32,  19,   1,   0,  -5,   1,   7,  19, -21,  52, -28,  19,  -8,   7, -10,   5,  42, -29,   8,   0,  -4,  -2,  23, -18,   1, -19,  27, -11,   8, -27,  -1,  16, -14,   7,   1,  -4,   3,  17,  -7,   1,   0,  11, -10,   8,  -2,  -3,   1,   7,   2,   8,  -1,   1,  -4 },
     {   -1,  -3, -11, -26, -18,  11,   8,   1,  -2,  -7, -23, -42, -28,  -1,   6,  12,  -1,  -2, -20, -46, -21,   8,   9,   4,   1,   0, -11, -34,  -9,  -1,   5,   6,   8,   5,  -6, -34, -29,  15,  11,   8,  12,  14,   4, -19, -35,   0,   2,   4,  19,  18,   8, -11, -16,  -9,   3,   0,  11,  19,   7,  -5, -30,   2,   0,  -2 },
     {    4,   4,   8,  55,  43,  -1,  16,   0,  -3,  -5, -19, -22, -10,   1,   3,   9,   1,   4,  11, -11, -35,  21, -25,  12,   3,   1, -10,  35,  37,   7,   3,   1,  -4,  11,  -6, -23, -38, -17,  -1,  -1, -13,  -4,  12,   8,  -6,   3,   9,  -8,  -2,  -1, -12,  -6,   8,   2, -13,   1,  15,   1,  13,   0,  24, -10,   8,   3 },
     {    0,   1,  -4,   8,   4, -51,  43, -19,  -1,  -1,   5, -15, -13,  54, -49,  11,  -1,  -1,  -2,   6,  -7, -44,  24,  -4,  -2,  -3,   0,  -8,  -2,  29, -17,  -7,   1,  -2,  -6,  -1,  -3, -22,  -1,  15,   4,   9,   5,  -1,  14,   3,   4,  -1,   2,  -6,   6,   3,  -2, -16,  -3,  -3,  -7,  -3,  -5,   0, -10,   4,   6,  -2 },
   },
   {
     {  116, -19,  -9,  -3,  -3,  -2,  -2,  -1,  23, -19,  -4,  -1,  -1,   0,   0,   0, -24, -16,   5,   1,   2,   1,   1,   0, -22,   1,   8,   0,   1,   0,   0,   0, -10,   7,   3,  -1,   0,   0,   0,   0,  -4,   4,   0,  -1,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {  -25,  30,  -6,   3,  -1,   1,   0,   0, 104,  -1, -17,  -1,  -5,  -1,  -2,  -1,  20, -37,  -7,   1,  -1,   1,   0,   0, -25, -24,   9,   3,   2,   1,   1,   1, -16,   0,  12,   1,   1,   0,   0,   0,  -7,   6,   5,  -2,   0,  -1,   0,   0,  -3,   3,   0,  -2,   0,   0,   0,   0,  -1,   1,   0,  -1,   0,   0,   0,   0 },
     {   35,   5,  -9,  -1,  -2,  -1,  -1,  -1, -17,  46, -14,   2,  -2,   0,  -1,   0,  70,  12, -31,   2,  -6,   0,  -2,   0,  25, -51, -18,   7,  -2,   2,   0,   1, -15, -39,  14,   9,   1,   2,   0,   1, -12,  -5,  20,   0,   0,   0,   0,   0,  -5,   5,   7,  -4,  -1,   0,   0,   0,  -2,   2,   1,  -2,   0,   0,   0,   0 },
     {   16,  46, -10,  -2,  -2,  -2,  -1,  -1,  39,  17, -12,  -6,  -1,  -2,  -1,  -1, -30,  44,  -8,  -4,   1,  -2,   0,  -1,  28,  36, -39,  -3,  -4,  -1,  -1,  -1,  19, -30, -40,   9,  -2,   2,   0,   1,  -4, -39,  -3,  17,   1,   4,   0,   2,  -4, -15,  13,   8,  -1,   2,   0,   1,  -1,  -2,   8,   0,  -1,   0,   0,   0 },
     {   10,  15,  -5,  -2,   0,  -2,   0,  -1, -14,  60, -15,  -1,  -1,  -2,  -1,  -1,  30, -17, -16,  -1,  -1,   0,   0,   0, -40,  14,  16,  -2,   3,  -2,   1,  -1,  11,  54, -27, -12,  -2,  -4,  -1,  -2,  12, -18, -49,   9,   2,   2,   1,   1,   1, -30,  -4,  23,   1,   3,   0,   2,   1,  -8,  11,   8,  -3,   0,  -1,   0 },
     {   -1, -72,  17,  -2,   4,   1,   1,   1,  38, -21,  28,  -6,   1,  -1,  -1,  -1,  61,  26,  15, -17,  -2,  -4,  -2,  -2,  16,  22, -10, -20,   0,  -4,   0,  -1,  -9,   6, -29,  -8,   4,  -1,   2,   0, -12, -11, -19,  10,   5,   1,   2,   1,  -6,  -5,   1,  12,   0,   0,   0,   0,  -2,   0,   4,   4,  -2,   0,   0,   0 },
     {   -9, -29,  -1,   6,   1,   2,   0,   1, -10,  -3,  -9,  10,  -2,   3,  -1,   1,   4, -44,  -1,  10,   0,   3,   0,   1, -23,   7,  13,   7,   0,   1,   0,   0,  21, -14, -20,  13,  -5,   3,  -1,   1,   1, -52,  34,  27,  -5,   5,  -1,   3,   4,   7,  65, -10, -12,  -1,  -4,   0,   2,  19,  15, -34,  -5,  -2,  -1,   0 },
     {  -11, -40,  22,  -4,   5,   0,   3,   1,  13,  44,  15,  -9,   1,  -4,   1,  -2, -37,  39, -27, -11,   1,  -2,   2,  -1, -59, -12, -34,  11,   7,   5,   3,   2, -26,   2,   2,  27,   4,   4,   0,   1,  13,  16,  16,  14,  -8,  -2,  -3,  -1,  14,   6,   2,  -7,  -8,  -1,  -2,   0,   3,  -4,  -4,  -5,  -1,   2,   0,   1 },
     {   15,  42,  72,   2,  -4,  -1,  -3,  -1,  -4,  24,  64, -11,  -6,   0,  -2,  -1,   2, -32,  25, -27,  -5,   1,   0,  -1,  -4,  -7,  -3, -27,   3,   0,   1,  -1,   5, -14, -16,  -9,   8,   0,   1,   1,  -5,   0,   7,   1,   7,  -2,   1,  -1,   1,  15,   2,  -7,   1,  -3,   0,  -1,   1,   7,  -6,  -5,   1,  -1,   1,   0 },
     {    0,  42, -14,   4,  -3,   2,  -2,   0, -16, -56,  29,   0,   7,   1,   1,   1,   7,  30, -13,  -7,   0,  -1,   1,   0, -12, -42,  -5,  -1,   6,   2,   2,   1, -49,  28, -11,   2,   8,   0,   3,  -1,   9, -11, -21,  29,  -3,   2,  -1,   0,   5,  -6,  34,  15, -12,  -1,  -2,   0,   7,   8,  13, -14,  -8,   0,  -1,   0 },
     {   -1, -29,  25,   3,   1,   2,   0,   2,  21,   9,  21,  -3,  -1,  -1,   0,   1, -40,   1, -34,   2,  -3,   5,  -2,   1,  44, -58,  15,  13,  -2,   5,  -3,   2,  35,  14,  20, -21,  -5,  -4,  -2,  -1,  12, -26, -27, -12,   2,   3,   2,   2, -14, -22,  12,  11,   7,   2,   2,   1,   0,   7,  14,  -1,  -1,  -3,  -1,   0 },
     {  -10, -15,   3,  -4,   7,   2,   3,   2, -20, -25, -23, -11,   8,   3,   3,   3, -18, -54, -25, -16,  12,   4,   4,   3, -10, -22, -43,  -9,  17,   3,   4,   1,  -7, -36, -30,   7,  18,   3,   3,   2, -21,  -3, -19,  13,  17,   0,   4,   0, -11,  -9, -23,  33,  12,   0,   3,   0,  -4,  -9,   5,  32,   2,  -1,   0,   0 },
     {  -11,  -8, -10,  -3,  -2,   1,   0,   0,  -5,  32,  -6,  -1,   0,   1,   0,  -1, -30,  19,  17,  13,   1,   2,   1,   0,  -7,  -3,  39,  -8,  -4,  -2,  -2,  -1, -53, -16,  -6, -38,   4,   1,   2,   0, -62, -20, -12,  -6,  22,   9,   6,   3, -21,  27,  10,  12,  14,  -2,   1,  -1,   3,  21,  -4,   1,  -1,  -6,  -2,  -1 },
     {   -4,  -4, -19, -51, -29,   2,   5,   3,  -5,  -6, -23, -63, -36,   3,   8,   3,  -3,   0, -20, -50, -23,   6,   5,   4,   3,   0,   2, -33, -13,   7,   6,   3,   6,  12,   3, -21,  -2,   8,   2,   1,   8,   3,   9,  -6,   5,   4,  -1,   0,   4,  11,  11, -14,   1,  -1,  -2,  -1,   3,   7,  -1, -14,  -1,  -2,  -2,   0 },
     {   -5, -10, -75,  -1,  -5,   2,   3,   4,   5,  26,  34,  24,   2,   2,   0,   0, -12,  -4,  50, -18,  -7,  -7,  -5,  -2,  -5, -35, -17, -32,   8,   2,   5,   1,  31,   3,   0,   7,  22,   2,   0,   0,   5,   4,   2,  10,   3, -12,  -2,  -2, -17,   1,  -2,   1, -10,  -5,   3,   1,  -4,   7,   1,   6,  -2,   1,   1,   1 },
     {   -2,   9,  -4,  -2,  -5,   1,  -2,   1,  -8, -21,  19,  -2,  -1,   2,   1,   2,  15,  16,   6, -11,   1,   1,   3,   1, -53,   4, -24,  13,  -1,   6,   3,   1,  12, -10,  24,  -1, -22,   4,  -5,   1,  -9, -42,   3, -54,   1,   9,   1,   3, -34, -26,  -4, -20,  33,  11,   7,   2,  -9,  19,  14,  14,  19,  -8,  -1,  -3 },
     {    0,   2, -17,   8,   6,   5,  -1,   2,   1,  -7,  19,  -6,  13,   0,   0,  -1,   0,   2, -39,   6,   3,   1,   1,   2,   4, -11,  26,  -4,   4,  -4,  -1,  -1,   2,  25, -46,   3,  -1,  -6,   1,  -2,   9, -34,  28,   7,  -6,   1,   0,   2, -16,  34, -27, -42,  12,  -3,   7,  -2, -11, -13, -46,  33,  20,   4,   5,   0 },
     {    2,  -1,   5,  60,  17,   0,   1,  -2,  -2,  -6, -18,  49,  -6,  -8,  -1,   2,  -6,   1, -33, -20, -40,  -8,   2,   3,  -6,   7, -21, -48, -33,   9,   6,   5,  -1,  12,   0, -34,  -4,  18,   5,   5,  -3,  12,  11,  -9,  10,   9,  -3,   0,   1,   2,   9,  -6,   3,  -2,  -5,  -1,   1,   1,   2,   0,   2,  -4,  -1,   1 },
     {    0,   6,   3,   2,   3,  -2,   2,  -1,  -3,   5,   3,   3,   1,  -4,   0,  -3,  -2, -13, -24, -22,   6,  -3,   3,  -3,  23,  26, -24,  15,  22,   2,   3,  -2,   5,  50,  20,  24,   9, -13,  -1,  -4, -62,  16,   2,   7, -22, -13,  -1,   0, -47,  11,  12,  -9, -18,   3,   6,   1,  -6,  26,   2,  -7,  -2,   9,   3,   0 },
     {    2,   4,  -4, -14,  68, -18,  -1,  -2,   1,   3, -11, -25,  69, -30,   0,  -3,   0,   1,   1, -31,  26, -30,   4,   1,   0,  -2,  12, -21,  -3, -20,   6,   0,   3,  -4,  13, -13,  -5,  -8,   6,   2,   8,  -2,   9,  -6,  -3,   5,   5,   0,   2,  -3,   6,  -6,  -3,   4,   1,   1,   2,   7,  11,  -6,  -8,  -2,  -3,  -1 },
     {    0,   2,   2,   0,  -5,   6,  -1,   4,   1,   1,  10,  -7,  -6,   1,   5,   1,   0,   5,  -8,  12,  -7,   3,   0,   2,   1,   5,  20, -10,  -2,  -3,  -3,   1,   0,   6,  -6,  14,   2,   0,   2,   0,   2,   5,  38,  -7,  -6,   4,  -1,   0,  -2,   8,  -7,   6, -13,   2,   1,   0,  29,  13,  69,  69, -49,  -9, -11,  -4 },
     {   -3, -10, -13,  23,   0,  -9,  -3,   2,  -1,  22,   6, -17,  -8,  -6,  -5,   0, -13, -34,  17,  -2,   4,   5,  -1,   2,  42,  14, -10,  -9,   8,   6,   5,  -1, -72,  13,  -3,  17,   0,  11,   0,   2,  31, -12,   8, -16, -10,   8,  -5,   2,  -3, -38,   2, -27,  12,  11,   2,   3,  -7,  -2,  19,  -2,  19,  -1,  -1,   0 },
     {   -1,  -1, -46,  33,  -4,   9,  -6,   3,   1,   5,  58, -40,  15,   2,  -6,   5,   1, -21, -37,   7,  -6,   3,   0,  -3,  -3,  24,  -2,  15, -23,  -1,  -6,   1,   3, -19,  -1,  -5, -24, -12,  -5,   3, -19,  16,  -5, -18,   1,  -5,   0,   2,  20,   8,  -4,  23,   1,  -3,   1,   2,  10, -18,   9, -27,  -7,   2,   3,   3 },
     {    2,  -3,   0,  65, -25,   4,  -2,   5,  -1,   2, -21, -16, -15,  12,  -7,  16,  -1,  13,  11, -50,  27,  -3,  -6,   2,  -4,  -4,  17,   0,  31, -25,  -4,  -3,   5, -16,  10,   2, -17, -20,  12,   5,  13,  -8, -13,  -3, -24,   2,  21,   1,   9,  21,   7,  13, -17,  11,   1,  -6,  -1,   9,  -7,  19,  10,   7,  -5,  -4 },
     {    5,   4,   2,   3,  23,  55,  35,  19,   5,   8,  -3,   1,  29,  64,  43,  16,   4,   5,  -1,  -2,  20,  44,  20,   6,   4,   6,  -1, -15,  11,  16,   6,  -4,  -1,   3,   3,  -7,   5,   6,  -6,  -6,   4,   4,   1,  -8,  -2,  -2,  -7,  -4,   4,  -1,   5,  -5,   1,  -1,  -4,  -4,   0,   3,   4,  -9,   2,  -1,  -2,  -2 },
     {   -1,   2,   6,  -2,   4,  12,   5,  -3,  -3,   1,   8, -10,   2,  11,   9,  -5,  -4,   2,  11, -15, -21,  -1,   9,  -2,  -3,   3,  27,  -5, -37,  -1,   5,   6,  -6,  -3,  34,  16, -24,   1,   6,   6, -14,  -7,  29,  51, -13,   0,   2,   4, -23, -35,   5,  37,  -1,   1,  -2,   6, -26, -38, -11,  26,  24,   8,   5,   3 },
     {    1,   2,  -1,  21, -12,  -5, -19,  -1,   0,  -3,   2, -11,   5,  -8, -17,   3,   5,  16, -20,  -4,  12,  10,  -4,   9,  -6,  22,  27, -15,  42,  21,   0,  14,   7,  -6,  29,   0,  54,  -2,   2,   0,   4,  -8,  21,   5,  46, -25, -10,  -8,  -2, -26,   5,   2,  11, -33,   6,  -3,  -7, -15,   1,  -6,  -3,  -9,  16,   3 },
     {    3,   3,  -9,  -4,  14,  -7,   4,  -4,   5,   8,   8,   0,   4,  -2,  -2,  -3,   6,  -7,   4, -11,  -1,   3,  -4,  -4,  15,  15, -16,  29,   7,  17,   7,   0,   2,  21,  17,   7,  -9,  17,  -1,   1,  22,   6,   5,   7,  31,  17,   0,   4,  21,  38,  35,  39,  42,  -9, -10,  -4,   0,  26, -20,  37,  18, -31, -12,  -5 },
     {    1,   2,   6,  37,  23,  -6,  -3,  -7,  -2,  -3, -17, -54,  -8, -10,  -5,  11,   4,  15,  21,  31,  -2,  21,  -5,  -7,  -9, -20,   4,   0,  -5,  16,  -7,  -5,  21,   5, -32,  10,  -5,  25,  10,   1,   7,  39,   6,   4,   2,  10,  -2,  -9, -52,  -2,   5,   8,   8,  -4,  -5,  -6, -12,  16,   2, -16,   1,  -3,  -4,  -1 },
     {   -3,  -2,  -2, -15, -31, -39, -16,   7,   2,   3,   7,  16,   9, -10,  -2,   6,   2,   1,  -7,  17,  57,  36,  15,   8,  -1,  -2, -20, -26,   6,  12,  -6, -10,   1,   9,  -5, -44, -31, -13,  -7,  -4,   3,  20,  32,   4, -11,   3,  10,   7,  -7,  -5,  20,  16,  -3,   8,   4,   2, -12, -14,   0,   6,   7,   2,  -1,   0 },
     {    1,   1,  -3,  12, -11, -37,  34,  18,   1,   0,  13,   0, -13, -41,  58,  24,   3,   1,  -2,   5,  -2, -15,  50,  12,   0,   4,  12,  17,   9,   6,  39,  -6,   5, -10,   2,   0,   8,  19,  20, -18,  -5,  -5,  -6,   0,   1,   4,  -5, -20,   4,   1,  -4,   1,  -7,   2,  -8,  -8,   1,  -6,  -1,   0,   3,   7,  -4,  -1 },
     {   -1,  -4,   7,  -5, -37,  23, -51,   2,   2,   4,  -2,  -9,  35,  15,  -8,  38,   0,  -7,   0,  24,   2, -25,  16,  -1,   3,   2, -18, -14,  -9, -21,  17,  -5,  -1,  13,  17,   8,  -1,  22,  21,  14,   2, -10,  -1,  -2,  12,  21,  -3, -25, -10,  -9,   1,   2, -14,  -1, -19, -17,  20,  11, -31,   4, -13,   6, -15,   3 },
   }
 },
 {  //14
   {
     {  123, -27,  -6,  -4,  -3,  -2,  -1,  -1, -14,  -4,   1,   0,   0,   0,   0,   0, -10,   0,   1,   0,   0,   0,   0,   0,  -5,   0,   1,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   10,   6,  -3,   0,  -1,   0,   0,   0, 115, -24,  -8,  -3,  -3,  -2,  -1,  -1, -35, -16,   5,   0,   2,   0,   1,   0, -25,   6,   4,   1,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {  -12,   5,   2,   1,   0,   0,   0,   0, -34, -21,   7,   0,   3,   1,   1,   0,-113,  20,  13,   2,   4,   1,   2,   0,  11,  27,  -4,   0,  -1,   0,  -1,   0,  15,  -1,  -4,   0,  -1,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
     {  -11,  -9,   4,   0,   2,   0,   1,   0, -31,   4,   3,   2,   1,   1,   0,   0,  -6, -37,   6,   0,   2,   1,   1,   1,-107,   9,  21,   2,   5,   1,   2,   0,  -7,  39,  -1,  -1,  -1,  -1,   0,  -1,  11,   3,  -6,   0,  -1,   0,   0,   0,   2,   2,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
     {   -5,   4,   0,   1,   0,   0,   0,   0,  -7, -20,   4,   0,   2,   1,   1,   0, -16,  10,   1,   1,   0,   0,   0,   0,   7, -43,   4,   1,   1,   1,   1,   1, -96,   7,  31,  -1,   6,   0,   2,   0,  -1,  53,  -4,  -4,  -1,  -2,  -1,  -1,  17,  -2, -14,   1,  -2,   0,   0,   0,   0,  -4,   1,   2,   0,   1,   0,   0 },
     {   24, 121, -13,  -5,  -9,  -4,  -4,  -2, -12,   3,  -8,   0,  -1,   0,   0,   0,   0, -23,  -2,   1,   1,   1,   0,   1,  -4,  -6,   1,   1,   1,   0,   0,   0,   2,  -3,   1,   0,   0,   0,   0,   0,   3,  -3,   0,  -1,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -5,  -5,   0,   1,   0,   0,   0,   0,  -9,  -8,   2,   1,   0,   1,   0,   0,  -9, -23,   1,   3,   1,   1,   0,   0, -17, -20,   3,   3,   1,   1,   0,   0, -13, -53,   7,   5,   2,   2,   1,   1, -65, -18,  44,   2,   7,   1,   2,   0, -23,  57,  20, -11,   1,  -3,   0,  -1,   9,  23, -14,  -6,  -1,  -1,   0,  -1 },
     {   -4,  -7,  -7,   3,   0,   2,   0,   1, -16,-100,   3,   8,   6,   5,   2,   2,  22, -54,  19,   5,   3,   2,   1,   1,  28,   4,  19,  -1,   0,  -2,  -1,  -1,  23,  13,   7,  -5,  -2,  -2,  -1,  -1,   5,   9,  -3,  -4,  -1,  -1,   0,  -1,  -4,  -1,  -3,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 },
     {   -1,   3,  -8,   0,   0,   0,   0,   0,   2, -22,  13,   0,   2,  -1,   1,   0,  -4,  34, -14,  -1,  -3,   0,  -1,   0,   3, -52,  10,   4,   3,   1,   1,   1, -13,  44,  -5,  -3,  -2,  -1,   0,  -1,  31, -58, -12,  10,  -2,   3,  -1,   2, -42,   7,  42,  -6,   4,  -1,   1,  -1,   6,  28, -11,  -8,   0,  -2,   0,  -1 },
     {    2,  15,  -9,   1,  -1,   1,  -1,  -1,  -2, -50,  10,   3,   4,   1,   2,   0,  32,  62,  15, -10,  -2,  -5,  -1,  -2, -31,  60, -25,  -9,  -3,  -2,  -1,  -1, -35, -21, -26,   9,   1,   4,   1,   2,   0, -12,   6,  10,   1,   2,   0,   1,  11,   4,   8,  -1,  -2,  -1,  -1,   0,   5,  -1,   0,  -3,  -1,   0,   0,   0 },
     {   -6, -10, -83,  17,   5,   3,   2,   2,  -5,   1, -79,  22,   5,   3,   2,   1,  -3,  14, -20,  20,  -1,   2,   0,  -1,  -2,  14,  17,  10,  -3,   0,  -2,   0,  -2,   5,  20,  -4,  -4,  -1,  -2,   0,  -1,  -3,  10,  -7,  -2,   0,   0,   0,   1,  -1,   2,  -3,   1,   0,   0,   0,   1,  -1,  -1,  -1,   1,   0,   0,   0 },
     {   -1,  -4,   5,   0,   0,  -1,   1,   0,  -1,  26, -13,  -1,  -1,   0,  -1,  -1,  -5, -46,  13,   6,   3,   1,   0,   0,  39,  44,  25, -16,  -3,  -6,  -1,  -2, -56,  27, -43,  -7,   1,   1,   2,   0, -16, -30, -12,  23,   2,   5,   0,   1,  17,   1,  27,  -1,  -3,  -1,  -1,   0,   1,   9,  -2,  -8,   0,  -1,   0,   0 },
     {   -4,  -9,   1,   3,   0,   0,   2,   0,  -8, -11, -12,   5,  -1,   2,   0,   0,  -9, -25,  -6,   4,   1,   2,   0,   1, -15, -22, -23,   9,   1,   3,   0,   2, -17, -50,  -5,  16,   2,   4,   1,   2,  -2, -34,  -6,  14,  -1,   3,   0,   1,  -9, -72,  19,  19,   1,   5,   0,   2, -24,  -8,  51,   1,   1,   0,   1,   0 },
     {    4,   5,  61,   7,  -2,  -6,  -1,  -2,  -2, -18, -19, -14,  -1,   1,   2,   0,   1,   4, -81,   2,   5,   7,   4,   2,   2,  34,  23,  32,  -1,   0,  -3,  -2, -10,  -2,  33,  -4, -11,  -3,  -3,   0,  -3, -20,  -3, -16,   0,   2,   1,   2,   9,   0,  -3,  -2,   5,   2,   2,   0,   2,   2,  -1,   2,   2,  -1,  -1,   0 },
     {    0,   6,  -1,  -5,   1,   0,   0,   1,   0,   4,   6,  -6,   2,  -1,   0,   0,   6,  17,   3,  -5,   1,  -3,  -1,  -2,   2,  25,   3,  -4,  -1,  -2,  -1,   0,  -3,  39,  12,  -8,  -1,  -5,   0,  -1, -66,  20,  -9,  -6,   1,   0,   0,  -1, -79, -24,  -9,  11,   5,   4,   1,   1, -35,  -6,  14,   9,   2,   2,   0,   1 },
     {    1,   0,   6,   0,   0,  -1,  -1,   1,   0,   9,  -7,   0,  -1,   1,  -1,   0,   5,  -3,  12,  -4,   1,   0,   1,  -1,  -1,  19, -10,  -2,  -2,   1,  -1,  -1,   6, -11,  26,  -2,   0,  -1,   0,   0,  10,  34, -17,  -8,  -1,  -2,   0,  -1,  -1, -19,  24,   2,   3,   1,   0,   1,  -1,  99,   5, -46,   2,  -9,   0,  -4 },
     {    5,   6,  18,  88,  23,  -7, -13,  -8,   2,   2,   7,  59,   8,  -8,  -9,  -6,  -3,  -6, -15, -16, -18,   0,   2,   3,  -1,  -3, -19, -40, -13,   8,   7,   4,   1,   7,  -1, -19,   2,   6,   4,   0,  -1,   2,   5,  -3,   6,   2,  -1,  -2,  -4,   0,  -1,   3,   3,  -2,   0,   0,   2,   0,   1,   1,   0,  -1,   0,   0 },
     {   -4,  -4, -65,  -6,   4,   7,   0,   2,   4,   8,  84,   9,  -3,  -8,  -5,  -1,  -4, -19, -51, -16,   6,   5,   4,   4,   4,  22,   8,   9,   0,  -1,  -2,  -1,  -5, -12,  14,  -6,  -1,  -1,   0,   2,  -1,   0, -11,  -2,   5,   0,  -1,   1,  10,  -5,   7,   2,  -2,   1,   0,   1,  -4,   1,  -1,  -1,   3,   1,   0,   0 },
     {    2,   4,   6,   9, -10,  -1,   0,   1,   3,  11,  19,   9, -19,   0,   0,   0,   4,  22,  51,   3, -18,   3,  -3,  -1,  -2,   7,  75, -11, -16,   2,  -6,  -1,  -3, -25,  43, -27,  -9,   1,  -4,   1,   5, -30,   8, -24,   4,   4,   0,   2,   8, -21,   3,  -7,   7,   1,   1,   2,  -2,  -9,   7,   3,   5,   0,   0,   0 },
     {    1,   5,  -1,   2,   0,  -3,   2,   1,  -1, -14,  -2,   0,   0,   2,  -2,  -3,   0,  24,  -3, -10,  -5,  -6,   4,   1,  -8, -29,   5,   8,  -1,   6,   0,  -1,  29,  24,   0, -16,  -3,  -5,  -1,  -6, -69,  -1, -28,  25,  -1,   6,   3,   2,  71, -12,  27,  -5,  -5,   2,  -4,   0,  -9,  -1,   4, -11,   3,  -3,   2,  -1 },
     {    3,   6,  11,  14,  23,   0,  -8,  -3,   6,   9,  24,  42,  39,  -4, -11,  -5,   5,   8,  19,  67,  27, -13, -16,  -6,   1,   2,   6,  53,   5, -18, -10,  -4,  -1,  -2,  -3,  19, -14, -15,  -2,   1,  -5,   1, -13,   4, -20,  -6,   3,   4,  -1,   3,  -2,  -8, -11,   2,   4,   1,  -3,   0,   0,  -7,  -6,   3,   3,   1 },
     {    1,   1,   7,  -1,  -1,  -1,   1,  -1,   2,   5,   5,  -3,   4,   0,  -3,   0,   2,  -2,   0,  -6,  -1,  -3,   0,  -1,   6,  11,  -9,   5,  -2,  -4,   2,   2,   4,  24,  15,   8,  -6,   0,   0,   0,  18,  28,  66,  -8, -17,  -4,  -6,  -1,   7,   0,  73, -33, -13,  -5,  -5,  -2, -29, -27,  28,  -8,   0,   2,   2,   3 },
     {   -1,  -2,   4,  13, -63,  39, -14,   6,   0,   0,   6,  31, -64,  41, -11,   5,  -1,  -1,  -4,  22, -33,  16,  -6,   1,   0,   2, -15,  27,  -1,   3,  -5,  -2,  -1,   8, -17,   9,   1,  -7,  -2,   2,  -4,  10,  -5,  -2,   0,  -4,  -1,   2,  -4,   5,   0,  -6,  -1,   4,   1,   1,  -3,   1,  -2,  -4,   2,   0,   0,   1 },
     {    0,   2,  13, -61,  46, -20,   7,   2,   1,  -4, -12,  73, -49,   1,   3,  -5,  -2,   5,  -7, -19,  12,  10, -10,   2,   0,  -1,   7, -11,   9, -17,   5,   3,   1,  -1,  -2,   6,  -8,   3,   4,   1,   0,   5,  -5,  -2,   8,   0,  -2,   5,  -3,   7,   4,   1,  -6,  -1,   4,  -4,  -2,  -5,   5,  -5,   3,   0,  -2,  -2 },
     {    0,   2,   7,   2, -16,   3,  12,   5,   1,  -5,  -3,   1,  10,   3,   2,  -3,   0,  16, -29,  24,  12,  -4,  -1,  -9,  -9,  -9,  48, -17,   5, -12,  -2,  -3,  11, -31, -47, -43,  32,   2,  18,   2,  13,  42,  -9,  10,  14, -13,  -3,  -1, -13,   9,  29,  22,  -3,  -1,  -1,  -4, -15,  -4,   8,   0,  -5,  -2,   2,   3 },
     {    0,  -2,   6,  -3,   2,  -2,  -1,   1,  -2,   1,  -7,   5,   0,   3,   4,   0,  -1,  -7,   5,   0,   2,  -2,  -1,   0,  -5,   3, -13,   0,   0,   2,  -3,   2,  -3, -18,  10,  -2,   3,  -5,   1,  -1,   4,   4, -11,   1,  -1,   0,  -3,  -2,  -2, -32,  30,  -6,   0,  -3,   0,   1, -40,  -9,-104,  24,  15,   6,   7,   2 },
     {   -1,  -1,   4, -40,  -5,   0,  11,   1,   1,  -2,   9,   9,   7,   3,   2,  -5,   1,  -5, -11,  71,  -5,  -9,   1,   7,  -5,   3, -41, -33, -35,  20,  -5,   2,  -2,  13,  18, -47,   3,  12, -11,  -2,   0, -12,  14,  -3,  16,  15,   9,  -1,  10,  -1,  -3,  -3,  18,  -6,  -6,   0,   0,   6,   6,  12,   3,  -2,  -3,  -6 },
     {    0,   0,   0,  -4,   8, -13,  -3, -10,   1,   1,   6,  -3,  -3,  19,  17,  -7,   1,  -2,  -3,   7, -18,  -3,  -1,   4,   0,   4,   1,   1,  -8,  -9, -11,  10,   4,   4,   3,   5,   3, -18,   9,  -5, -15,  17,  13,  -5, -11,   3,   1,  -1,  -9, -36,  33,  31,  -9,   7,  -1,   1,  89,  -7, -15,  19, -30,   0,  -8,   4 },
     {   -1,  -2,   0,  15, -33, -12, -11, -10,   1,   2,   6,   3, -10,  25,  11,   4,   2,   6,  -4,   7,  70,  12, -13,  -2,   0,   0,  -3, -43,  53, -17, -16,   8,   3,   7,  23,   0,   7, -26, -12,  -7,  -2, -11,  20,  21,   5,  10,   9,  -2,   7, -12, -11,  -1,  -6,  -7,   2,   1,   3,   3,  -1,  -4,   0,   7,  -4,  -9 },
     {    0,   2,  -3,  -4,  11, -17,   4,   9,   1,  -6,  11,  -1,   2,  12,  14,  -6,  -2,   8, -23,   9, -18,   3,  -4,   0,  -1, -13,  32, -14, -24,   6,  -5,  -7,   2,   7, -45,  32,  -2, -22,   4,  -7,  -7,   5,  51,   1, -18,  10,  -3,  -5,   9, -36, -41, -23,  10, -15,   5,   2, -17,  31, -12,  -5,   6,  -2,   0,  -7 },
     {   -1,  -2,  -2,  10, -34, -30,  16,  13,  -1,  -2,   4,   8, -19, -57,  24,  17,   0,   1,  -1,  19,  -7, -32,  39,  -2,   1,  -5,  -5,   0,  21, -37,  -9,  12,  -1,   7, -11,  10, -35,  17,   2,  -7,  -4,   2,  10, -18,  12,   9, -16, -11,   6,  -3,   1,  15,   1,  22,  -1,  -4,   0,   7,  -4,   4,  17,  -3,  -5,   0 },
     {   -3,  -4,   3, -31, -33,   0, -27,  -1,   2,   1,  -2,  29,  44,  -5,  -2,  -3,  -2,   0,  -6,  -7, -16,  10,   0,   0,   2,   2,   2, -20,  41,  23,  -6,  10,  -1,   2,  -1,   1,   8,  20,  39,  -6,  -9, -13,  -1, -28, -31, -31,  -5,  -4,   9,  -7,  -4, -15,  19,  26,   7,  11,   0,   2,  -2,  -2, -21,  -5,   5,  21 },
   },
   {
     {   73, -49,   6,  -4,   1,  -1,   0,   0, -72,  43,  -1,   1,   0,   0,   0,   0,  33, -11,  -7,   2,  -2,   0,  -1,   0, -12,  -2,   7,  -1,   1,   0,   0,   0,   5,   3,  -3,   0,   0,   0,   0,   0,  -3,  -2,   1,   1,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   57,   5, -24,   3,  -4,   0,  -2,   0,  -7, -42,  33,  -6,   5,  -1,   2,   0, -41,  59, -22,   1,  -1,  -1,   0,   0,  38, -38,   1,   7,  -3,   2,  -1,   1, -21,  12,  10,  -7,   2,  -1,   1,  -1,   9,  -1,  -9,   3,   0,   0,   0,   0,  -5,   0,   4,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0 },
     {  -48,  48, -29,   9,  -4,   2,  -1,   1, -60, -13,  37,  -3,   4,   0,   1,   0,  58,  12, -25,  -2,  -2,  -1,  -1,   0, -15, -23,  13,   4,   0,   1,   0,   1,   0,  20,  -3,  -5,   0,  -1,   0,  -1,   5, -11,  -4,   4,   0,   1,   0,   0,  -1,   4,   4,  -2,   0,   0,   0,   0,   1,  -1,  -2,   0,   0,   0,   0,   0 },
     {   58,  35, -35,  -2,  -4,  -1,  -2,   0,  21, -50,  17,   6,  -1,   2,   0,   1,  25, -17,  16, -15,   4,  -3,   1,  -1, -52,  36, -16,   7,   1,   0,   1,   0,  29, -18,   3,   3,  -4,   2,  -2,   0, -14,   4,   4,  -4,   3,  -1,   1,   0,   4,   1,  -3,   2,  -1,   0,   0,   0,  -2,  -1,   1,  -1,   0,   0,   0,   0 },
     {   21, -13, -12,  12,  -4,   2,  -1,   1,  59,   4,   9, -20,   2,  -4,   0,  -1,  22, -27, -27,  21,  -4,   3,  -1,   1, -18, -19,  49, -18,   3,  -1,   1,   0, -17,  44, -38,   3,   4,  -3,   1,  -1,  14, -30,  13,  10,  -8,   4,  -2,   1,  -8,  12,   2, -11,   4,  -2,   1,   0,   2,  -3,  -4,   5,  -1,   0,   0,   0 },
     {  -19, -20, -24,  31,  -5,   5,   0,   1, -15,  13,  45, -39,   7,  -4,   1,  -1, -48,   4, -29,  26,   2,   1,   1,   1,  -9,  36,  15,  -9,  -7,   1,  -2,   0,  40, -31, -19,   3,   5,   0,   1,   0, -24,   6,  21,   1,  -3,   0,   0,   0,   9,   7, -14,  -3,   2,   0,   0,   0,  -1,  -5,   5,   4,  -1,   0,   0,   0 },
     {  -13, -39,  39, -23,  13,  -4,   3,  -1,  25, -19,  -2,  13,  -4,   1,   0,   0,  20,  47, -37,  -2,  -4,   0,  -1,  -1, -41, -29,  15,  14,  -1,   4,  -1,   2,  46,  -2,  12, -18,   1,  -2,   0,  -1, -34,   9, -12,   6,   4,  -2,   2,  -1,  15,  -2,   3,   3,  -6,   2,  -1,   0,  -4,  -1,   0,  -3,   2,   0,   0,   0 },
     {    5,  -9, -23,  26,  -9,   4,  -1,   1,  40,  28,  15, -30,   4,  -3,  -1,  -1,  55, -11, -21,   7,   1,  -2,   1,  -1,  38, -11, -18,  11, -10,   4,  -2,   1, -13, -36,  37, -14,   7,   0,   1,   0, -10,  37, -22,   1,   2,  -4,   1,  -1,   9, -20,   0,   7,  -5,   3,  -1,   1,  -4,   4,   4,  -4,   2,  -1,   0,   0 },
     {   12,  20,  23, -10,  -5,  -1,  -1,  -1,   1, -17, -23,  -1,  12,   0,   2,   0,  41,  30,  -7,  22, -20,   2,  -3,   0,  46,  -3,  -2, -35,  10,  -2,  -1,   0,  18, -36, -26,  35,  -3,  -1,   1,   0, -10,  -5,  42, -20,  -6,   4,  -2,   0,  -3,  14, -25,   0,  10,  -4,   1,   0,   3, -10,   7,   8,  -7,   2,   0,   0 },
     {  -17, -56,  25,  30, -23,   8,   0,   1,  -8, -52,  11, -31,  35,  -5,   4,   1,  18,   8,  23,  14, -18,  -1,   0,  -2, -23,  17, -28, -10,   9,   3,   0,   1, -23,   8,  17,  15,  -4,  -2,   0,   0,  28,  -4,  -7, -12,  -2,   3,  -1,   0, -16,  -1,   4,   6,   6,  -3,   1,   0,   4,   5,  -2,  -2,  -5,   1,   0,   0 },
     {  -16, -50, -17, -10,  30,  -5,   5,   1,  12,  -9,  40,  25, -34,   5,  -3,   0,  23,   4,  -5, -39,  26,  -2,   0,   1,  31,  20, -19,  23,  -6, -10,   3,  -3,  -4,  -7, -15,   4, -13,  12,  -3,   3,  17, -29,  27,  -6,   9,  -4,  -1,   1, -14,  20,  -7,  -5,   0,  -2,   3,  -2,   3,  -7,  -3,   6,   0,   2,  -1,   0 },
     {   -8, -16,   2, -14,  22,  -7,   2,   1, -16, -43,  14,  17, -19,   9,  -1,   1,  -5, -36,  -1,   9,  10,   1,  -1,   1,  -6, -13,  36, -19,   2,  -4,   2,   0, -52,  -9,   1,  12,   6,   3,   1,   0, -23,  60,  15,  -9,  -4,  -3,  -1,  -1,  20, -11, -27,   2,   2,   2,   0,   0,  -7,  -1,  11,   4,  -1,   0,   0,   0 },
     {   -1,   5, -18,   6,  -2,   2,  -2,   0,  11,  34,   0,  -3,   1,  -4,  -2,  -2, -15,   6,  -5,  -9,  -5,   3,   0,   0, -39, -40, -11,  14,  15,   0,   2,   0,  -8,   6,  37,  18, -22,   3,  -1,   0,   3,   5,  23, -51,  15,   0,   0,  -1, -18,  30, -45,  24,   9,  -9,   4,  -2,  12, -17,  12,   6, -15,   7,  -2,   1 },
     {    4,  16,   8, -36,  24,  -7,   4,  -1,   6,  19,  13,  10, -13,  -2,  -1,  -3,  -2,  21, -49,  23, -10,   6,  -2,   2, -28,  42,   1, -20,   8,  -3,  -2,  -1, -23, -27,  10,  12,   0,  -1,   2,   0,  57,  11, -18,  -6,  -3,   1,  -2,   0, -28, -17,  12,   5,   2,   1,   0,   1,   9,  10,  -2,  -3,   0,  -1,   0,  -1 },
     {    5,  10,  26,  -2,  -6,   1,  -3,   0,  -9, -21, -19,  -6,   6,   1,   2,   2,   1, -20, -14,  16,  13,  -5,   3,  -2,   0,   6,  12,  26, -41,  12,  -4,   2,  -7, -21,  20, -48,  34,   1,  -2,   2,  19,  -6,  15,  17,   0, -17,   7,  -4, -36,  43, -28,  -4, -10,  10,  -1,  -1,  19, -23,   3,   8,   4,   0,  -4,   3 },
     {   -4,  -4,  -9,  -1,  -2,   1,   0,   1,  -5,  -5, -11,   2,   2,   1,  -1,   1, -11, -17, -16,   6,  -1,   2,   0,   0, -22, -36, -15,   7,   7,   3,   2,   1, -35, -60,   5,  17,  -4,   3,   0,   1, -36, -39,  24,  17,   5,   4,   0,   1, -20,  12,  54,  -9,  -1,  -3,   1,  -1,  -6,  24,  -3, -10,   4,  -3,   1,  -2 },
     {   -6, -14, -16, -34, -15,  18,  10,   1,  -2, -15,  20,  38,  33, -24,  -9,  -3,  -8, -48, -10,   4, -42,  21,   8,   4,  32,  -2,  20,  -1,  25,  -1, -11,   0,  32,  -4,  19, -13, -14, -12,   6,   1,   1,  -5, -16,  -6,  10,   9,   1,  -3, -16,   7,   7,  13,  -3,  -6,  -2,   1,   9,  -4,   0,  -4,  -3,   5,   0,   0 },
     {    1,  -1,  -5,  12,  48, -20,  -9,  -1,  -7, -19,  -7, -23, -42,  27,  18,   1,  -1, -25,  10,  42,  12,  -7, -17,   0,  17, -17,  -2, -25,   5,   5,   8,   4,  41,   9,  30,   5,  -9,   0,  -5,  -7,   0, -16, -26, -16,   8,   0,   1,   2, -20,  10,  18,  12,  -1,  -1,   0,  -2,  11,   1,  -3,  -3,  -4,   3,  -1,   0 },
     {   12,  31,  77, -26,  -4,  -8,   2,  -5,   5,  12,  71, -36,  -2,  -7,   0,  -3,  -4, -26,   8,  -8,   4,   0,   1,   2,   6, -10, -17,   7,   7,  -1,   2,   0,  -5,   4,  -4,   3, -10,   8,  -1,  -2, -15,  -7,  -1,  -1,   7,   0,  -2,   2,   3,   5,   2,   2,  -2,  -2,   2,  -1,  -2,   0,   1,  -3,  -2,   3,   0,   0 },
     {   -2,  -4, -13,   0,  18, -13,   4,   3,   7,  17,  17,   7,  -7,  -3,   5,  -3,   6,  23,  28,  -5, -24,   5,  -8,   0,  -9,   8, -18, -40,  28,   7,   1,  -1, -24,   0,  -6, -12,  19, -26,   8,   3, -29,   4, -15,  43,  -3,   1,   9,  -8,  -8,  41,  -7,  13, -28,   8, -10,   4,  25, -31,   9, -15,   5,   9,   0,   0 },
     {    0,   0,  -1,   6, -12, -47,  53, -15,   1,   1,   2,   3,   8,  52, -63,  15,  -1,  -1,  -4,   5,  -7, -36,  40,  -5,   3,   1,   2,  -3,   4,  19, -13,  -5,   2,   3,   1,   2,  -4,  -5,  -3,   7,   1,  -4,   0,  -4,   4,  -2,   7,  -7,  -1,   1,   1,   3,  -1,   2,  -4,   3,   2,  -1,   1,  -3,   0,  -1,   2,  -1 },
     {   -2,  -3,  -2,  -9,  10,   0,   1,   1,   4,   8,  19,  13,  -4,  -3,  -3,  -2,  -2,  -6,   9,  -3,  -1,  -2,   0,  -2,  -4, -14, -22,  -4, -11,  20,  -8,   3,  13,   5,  23,   7,  34, -38,   5,   0,  -2,   3,   7,   1, -63,  28,   2,  -5,  -9,   0,  -2, -38,  52,  -6, -11,   5,  -2,   6,  -9,  28, -14, -14,  13,  -6 },
     {    1,  -1,   7,  18, -15, -16,  -4,   6,   3,   6,  10,   7,  14,   8,  -2,  -3,   1,   4,  15, -29,  19,   6, -10,   3,   0, -18,  24,  -5,  -8, -27,  12, -13,  23,  -8, -20,  16,  12,  -2,   3,  -2,   0,  44,   3,  -4,  -7,  -4,   9,  -3, -63,  -3,  12,   8,  -2,   3,   0,   1,  33,  23, -37, -23,   8,  -6,  -2,   1 },
     {   -1,   0,  -4, -12,  -3,   2,  -4,   2,   1,   4,  -4, -11,   3,  -2,   0,   0,   3,   7,   5,   5,  -1,  -8,  10,  -3,   1,   9,  10,  -6,  12,  -3, -10,   5,  -7,  -3,  15,   1, -18,  13,  -2,  -4,   0,  26,  -5, -10,  16,  -8,  -3,   1,  24,  50,  20, -42,  -3,  -6,   6,  -5,  -4, -14, -84,  27,   4,   0,   3,   2 },
     {    0,  -5,   1,  12,  -3,  -2,   0,   5,   0,  -1,  11,   3,  -6,   2,   2,  -6,  -7, -11,  24, -11,  -3,  -2,  -3,   2, -15, -53,   4,  -9,  13,   9,  -5,   0,  24, -50, -17,   2,   0,   5,  -2,   4,  70,   0,  -3,  19,  -2, -10,   2,  -6,  38, -12,  -4, -14, -14,   4,  -6,   2,  12, -22,   4,   8,   1,   6,  -1,   0 },
     {    7,  10,  34,  68, -38,   5, -17,   6,   1,  -1,  10,  55, -46,   6,  -9,   9,  -3,  -3, -23,   7,  -7,  12,   2,  -5,   1,   9,  -5, -11,  17, -22,  -8,   4,  -2,   0,  12,  -5,  -8,   4,   5, -10,  -5,  -3, -10,   3,   4,   3,   4,  -3,  -2,   6,   0, -17,   8,  -5,   6,  -3,  -4, -10,   6,   3,  -3,  -2,   1,  -1 },
     {    1,   2,   1,   4,  -4, -11,   6,  -2,   9,  19,  30,  23,  -5, -11,  11,  10,  10,  28,  58,  25, -20,  -4,  -4,   4,  -2,   9,  52,   1, -33,   5,   2,   3,  -6,  -9,  34, -11,  -3,  -2,  -8,  -5,   3, -15,  30, -23,  -1,  -6,  -3,   1,   4,  -4,  23,  -8, -10,   8,   1,  -5, -14,  11,  13, -13,   4,   1,  -4,   2 },
     {    0,   1,  -7,  -1,   8, -24,  31, -18,  -2,  -1,  -7, -19,   3,  15, -16,   3,   0,   2,   7,   1, -33,  45, -26,  18,  -1,  -1,  -3,  11,  17, -50,  37, -14,   3,  -1,  16, -23,   2,  36,  -6,  -5,   2,   1,   8,  24, -17,  -3, -15,  16,   0,  -1, -10, -23,  17,   2,  10,  -5,  -4,   2,   6,  14,  -6,  -4,  -3,   4 },
     {   -3,  -4,  -3, -12, -27, -38, -32,  41,   3,   4,   6,  17,  20,  42,  26, -45,  -4,  -8, -14,   6, -18, -34, -29,  33,   3,   8,  -5,  14,  -1,   6,  24, -21,  -5,   3,   2,   3,  -6,  -4, -14,  -5,  -3,  -4,   0,  -2,   3,   6,  11,   5,   8,  -4,  -3, -10,   0,   4,  -5,  -2,  -5, -14,   2,   2,   4,  -6,   0,   2 },
     {    1,   1,   7,   0,  -5,  18, -18,   8,  -4,  -6, -11, -37, -25,  23,  -8, -11,   1,   8,   5, -55, -16,  -2,  11,   5,   4,  15,  52, -11,  -1,  11,   5,  -9,  -7, -24,  22,  -5, -28, -14,  29,  13,  -1, -19, -16,  12,   1,  24,  -5,  19,  -2,   4, -14,   6,  25,  -6, -13,   4,   3,   6,   2,  -4,   0,  -6,   1,  -4 },
     {   -3,  -6,  -7, -28, -46,  46,  16, -17,   2,   3,   7,   3, -15,   8, -15,  -1,   1,   6,   6,   6,  36,  -9, -10,  12,  -1,  -1,   6, -31,  -6, -12,  37, -14,  -2,  -6,   9,  10,  24,   9, -41,  15,  -6, -14, -25,  -6,  -4,  -5,  24,   0,   8,   7,   9,  17,  11,  -5,  -1,  -7,   6, -14,   3,  11, -12,   4,  -9,   5 },
     {    1,   1,   5,   1, -16,  -4,  -1, -13,  -1,  -5,  -4,  -4, -14,  10,  -7,   1,  -2,  -8, -20,  -7,   7,  15, -18,  10,   5,   6,  -9,  -1,  13,   9,  24,   5,   7,  10,  18, -10,  16, -34,   6,   8,  25,   6,  44,  -3, -21,  18, -15, -14,  39,  24,  10,  34, -19, -19,  15,  -2, -11,  23, -22, -37,  23,   9,  -3,   4 },
   },
   {
     {  103, -46,  -5,  -4,  -1,  -1,  -1,   0, -48,   8,  11,   0,   2,   0,   1,   0, -24,  19,  -1,   0,   0,   0,   0,   0,  11,   0,  -6,   1,  -1,   0,   0,   0,   2,  -3,   0,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   26,  30, -23,   1,  -5,   0,  -2,   0,  63, -50,   2,   1,   0,   1,   0,   0, -77,   2,  28,  -3,   5,   0,   2,   0,   4,  27,  -9,  -4,  -1,  -1,   0,  -1,  10,  -5,  -9,   3,  -1,   0,   0,   0,  -2,  -3,   2,   2,   0,   0,   0,   0,  -2,   0,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {    8,  72, -55,  11, -11,   1,  -3,   0, -62,  -1,   9,   8,   2,   3,   1,   1,  14, -39,  34,  -9,   4,  -1,   1,   0,  28,   3,  -5,  -6,   0,  -2,   0,   0,  -5,   1,  -5,   2,   1,   0,   0,   0,  -4,   0,   2,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   50,  -1,  -5,  -5,  -1,  -1,  -1,  -1,  38,  11, -21,   2,  -4,   0,  -1,   0,  42, -68,  11,   6,   1,   2,   0,   1, -57,  14,  26,  -7,   4,  -2,   1,   0,   0,  28, -12,  -5,   0,  -2,   0,  -1,   4,  -8,  -6,   5,   0,   1,   0,   0,   0,  -2,   2,   1,  -1,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   0,   0 },
     {   21,   2, -19,   8,  -4,   2,  -1,   0,  23, -64,  41,  -7,   5,  -1,   1,   0,  66,  28, -13, -17,   0,  -5,   0,  -2,  -3,  -2, -27,  11,   0,   1,   1,   0, -45,  -3,  16,   7,   0,   1,   0,   0,  10,   9,   0,  -5,  -2,   0,  -1,   0,   3,  -2,  -2,  -1,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0 },
     {  -26, -19,   8,   6,   1,   2,   0,   0, -42, -19,  30,  -4,   4,   0,   1,   0,   0,   4,  14, -15,   5,  -4,   2,  -1, -58,  71, -26,   2,   2,  -2,   1,  -1,  28,   1, -26,  13,  -6,   2,  -2,   0,  11, -24,  11,   3,  -3,   2,   0,   1,   0,   1,   5,  -4,   0,   0,   0,   0,   0,   1,  -1,  -1,   1,   0,   0,   0 },
     {   11,  36,  55, -61,  13, -13,   2,  -3, -23, -40,  -6,  11,   7,   2,   3,   1,   2, -14, -31,  46, -14,   7,  -2,   1,  15,  32,   4, -12,  -8,  -1,  -2,  -1, -15,  -2,   2, -10,   6,   1,   1,   0,   2,  -3,  -1,   4,   2,   0,   0,   0,   3,  -2,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -1,  34, -20,   2,  -6,   0,  -1,   0, -24,  -6,  16,  -4,   4,   0,   1,   0, -34,  28, -32,  13,   0,   1,   0,   0, -65, -31,  30,  13,  -1,   4,   0,   1,  -1,  31,  27, -20,  -1,  -2,  -1,  -1,  35,   6, -29,  -4,   3,  -1,   0,   0,  -4, -12,  -1,   8,   1,   1,   0,   0,  -1,   1,   4,   0,  -2,   0,   0,   0 },
     {  -21, -35,  14,   3,   6,   0,   1,   1, -13, -12,  11,  -1,   0,   1,   0,   0, -28, -19,  32,  -7,   2,   1,   1,   0,  20,  -2,   6, -13,   6,  -3,   1,  -1, -69,  68, -18,   2,   5,  -4,   2,  -2,  16,   5, -20,  11,  -5,   1,  -1,   0,  11, -19,   7,   4,  -4,   3,  -1,   1,  -1,   2,   3,  -3,   0,   0,  -1,   0 },
     {    9,  14,  45,  44, -60,  13,  -7,  -1,  -9, -13, -23, -17,  13,   5,   1,   3,  -4, -10, -29, -28,  53, -15,   4,   1,  12,  17,  27,  14, -16,  -7,   1,  -3, -11,  -4,  -1,   2, -15,   9,   0,   0,  -1,  -2,  -6,  -2,   6,   2,  -1,   1,   4,   3,   0,  -1,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -3,  -6,  -9,   3,   5,   1,   0,   1,  14,  12,  58, -46,   6,  -8,   2,  -2,  -7, -40, -36,  23,   7,   2,   2,   0,  28,   2, -23,  38, -13,   3,  -2,   0,  25,  36,   6, -22, -10,   0,  -2,   0, -25, -22,  -2,  -3,   8,   2,   1,   1,  -5,  -2,   5,   6,   1,  -1,   0,   0,   5,   2,  -2,  -1,  -1,   0,   0,   0 },
     {    1,  -3,  -2,   3,  -3,   1,  -1,   1,   0, -36,   2,   7,   4,   1,   1,   1,  27,  28,  -2,  -4,  -3,  -2,  -1,   0,  28, -23,  38, -23,   4,  -1,  -1,   0,  53,  27, -47,  -6,   7,  -4,   1,  -1,   1, -32, -22,  26,  -4,   2,   1,   0, -28,  -6,  31,   0,  -4,   3,  -1,   0,   5,  11,  -2,  -9,   1,   0,   0,   0 },
     {  -13, -29,   6,   5,   2,   3,   0,   1, -19, -50,  17,   1,   7,   2,   2,   1,  -4, -26,  23,   2,   1,  -1,   0,  -1, -17, -17,  33,  -6,  -1,   2,  -1,   1,  22, -11,  15, -20,   7,  -1,   1,   0, -59,  54,  -9,  -5,  11,  -6,   3,  -2,   2,  21, -23,   9,  -3,  -1,   0,  -1,   7,  -9,   2,   3,  -4,   2,  -1,   0 },
     {   -9, -18, -41,  33,  -5,   6,  -2,   1,   4,  10,   9, -17,   3,  -2,   1,   0,  -1,   6, -25,  38, -14,   3,  -1,   0,   7,  41,  38, -35, -14,   2,  -3,   0, -29, -33,  -6, -26,  30,  -1,   2,   2,   3, -11,  -2,  30,   1,  -6,   1,  -1,  22,  16,   0,  -3, -13,   0,   0,  -1,  -5,  -6,  -1,  -3,   2,   3,  -1,   0 },
     {  -10, -16, -21, -50, -56,  36,  10,   4,   8,  11,  11,  24,  23,  -2,  -8,  -2,   5,   7,  14,  34,  38, -36,  -8,  -2,  -7, -11, -20, -30, -25,   7,  12,   3,   3,   8,   4,  -1,  -5,  14,   0,  -1,   0,   0,   5,   6,   5,  -3,  -5,  -1,   0,  -3,  -3,   0,   2,  -3,   0,   1,   2,   0,   0,   0,  -1,   0,   1,   0 },
     {   -4, -10, -12,  -5,  -5,   2,   4,   3,   4,   1,  25,  51, -31,   5,  -8,  -1,  -8, -20, -45, -27,  15,  11,   1,   4,  13,  17,   7, -40,  43, -12,   3,  -1,  10,   6,  43,  20, -16, -15,   1,  -5,  -7, -14, -20,   3, -13,   7,   3,   0,   2,   7, -10,  -2,   6,   4,  -3,   0,  -1,   1,   2,   1,   1,  -1,  -1,   0 },
     {   -4, -23,  -3,   8,   1,  -1,   1,   1,   2,  -8, -10,  -1,   2,   1,   1,   1,  13, -16,  22,   5,  -1,   2,  -1,  -1,  30,  25,  -6,   4, -11,   1,  -3,   0,  33, -19,  31, -23,  -1,   1,  -2,   1,  63,  14, -44, -17,  11,  -5,   2,  -1, -10, -37, -16,  26,   3,   0,   2,   0, -17,   2,  27,   1,  -6,   1,  -1,   0 },
     {   -9, -25,  -1,   5,   4,   0,   1,   0, -10, -29,   6,   3,   0,   2,   0,  -1, -14, -40,   7,  14,   4,   3,   0,   2,  -2, -35,  15,   6,  -2,   2,  -1,   2, -14, -30,  20,   5,  -6,   8,  -1,   2,  43, -35,  37,  -8,   2,   2,  -4,   1, -41,  33,  19, -23,  11,  -7,   2,  -2, -12,  25, -19,  -1,   5,  -4,   3,  -1 },
     {    8,  19,  32, -12,  -8,  -2,  -3,   0,  10,  24,  61, -39,   1,  -4,  -1,  -2,  -5, -15,   5,  -8,   5,  -2,   0,   0,  -2, -19,  12, -31,  14,   0,   3,   1,   8, -32, -27,  32,  14,  -5,   2,  -2,  31,  37,  -8,  14, -21,  -5,  -1,  -3,  -7,  -7,  -4, -13,  -4,   4,   2,   1, -10,  -9,   5,   7,   3,   3,  -1,   0 },
     {   -3,  -5,  -7,   0, -24, -63,  48, -15,   2,   3,   3,  -4,  17,  28,  -6,  -1,   1,   3,   3,   0,  14,  44, -55,  16,  -4,  -3,  -5,  -1, -15, -35,  15,   5,   2,   0,   0,  -1,   6,  -2,  23, -11,  -3,   1,   1,  -2,   0,   8,  -8,  -3,   1,   1,   3,   0,  -4,   0,  -4,   5,   0,   2,   0,  -1,   2,   0,   1,   1 },
     {    1,  -3,   6,  38, -28,   8,  -6,  -7,  -6, -18, -27, -17,  27,  -6,   3,   4,  -1,  -6,  -8,  33, -27,   7,  -2,   5,   0,  -9, -29, -25,  23,   8,   3,   0,  19,  27,  18,  16,  20, -26,   5,  -3,   1,   1,  29,  -2, -43,   4,   0,  -2,  -5, -18, -32, -11,  11,  15,  -2,   3,   5,  -1,   3,  12,  10,  -3,  -2,   0 },
     {   -6, -12, -33, -11,  14,  -2,   5,   2,   0,  -9,  -4,   2,   5,  -3,   1,   0,   4,   4, -21,  15,   6,   3,   2,  -4,  18,  12,  41,  12, -11,  -7,  -2,  -3,  12,  19, -35,  34, -34,   7,  -5,   1,  22,  27,  20, -41,  -4,  13,  -7,   3,  22,   0, -24, -24,  24,  -7,   0,   0,  -6, -24,  -7,  27,   2,  -6,   4,  -1 },
     {   -2,  -4,  -3, -11, -16,  -9,  12,   2,   5,   7,   8,  29,  51, -29,   5,  -8,  -7, -14, -15, -42, -27,  22,   7,   2,   2,   5,   9,   7, -42,  46, -18,   1,  -1,   2,   3,  27,  35, -19, -14,   3,   0,  -6,  -8, -16,   4, -18,  10,   5,  -1,   4,   8,  -1,  -3,   4,   5,  -4,   2,   1,  -1,   2,  -2,   3,  -1,  -2 },
     {   -7, -16, -20,  -4,   6,   2,  -5,   0,  -9, -22, -15,   6,  -1,   1,   0,   4,  -9, -26, -26,  11,  12,   2,   3,   1, -12, -24,  -8,  23,  -4,   1,   2,  -1,   0, -35, -13,  32,  -6,  -5,   3,  -2, -15,   3,   2,  23, -14,   5,   0,   1,  33, -56,  41,   3, -12,  13,  -7,   5, -11,  -3,  39, -20,   3,  -2,  -1,   1 },
     {   -4,  -5, -21,   2,   5,   0,   0,   3,  -2,  10, -17,  -6,   5,  -2,  -1,  -2,  -3,   1, -21,  17,   5,   0,   4,  -3,  -4,  27,  -9,   1,   1,  -5,   0,   0, -24,  -5, -13,  26,  -7,  -2,   1,   2, -19,  40, -27,  -4,  -5,   2,   0,   0, -91,   0,  16,   7,   6,   4,   0,   1,  24,  15,   6,   1,  -3,   0,  -1,   1 },
     {   -1,   0, -13, -23,  38, -11,  15,   6,  -3,  -1, -23, -45,  55, -14,  10,  -1,   1,   3,   8, -20,  17, -19,   3, -14,   1,   1,  11, -11,   5,  -5,   8,  -1,   1,  -1,  33,  21, -20,  14,  -5,   5,  -7, -22, -24,  29, -24,  12,  -8,   4,  10,  21, -20,  -3,   0,   3,  -2,   1,  -2,  12,   8, -13,   2,  -1,   1,  -1 },
     {    0,   4,   4,  -4,   7,   3, -15, -11,   1,   6,  16, -12,  12,   6,  -4,   1,   0,  10,  13,  -3,   1,  -2,  -7,  18,  -7,  -2,  25, -20,   7,   8,   1,  -4, -21, -18,  -5,  -2,   1,  -3,  10, -13, -35, -46,   5, -37,  13,   1,  -3,   9, -12, -34, -10,  23,  47,  -3,   4,   2, -15,  13,  37,  32,  -3, -20,   1,  -2 },
     {   -4,  -5,  -7, -30, -33,  36,   3,   4,   0,   1,  -3, -13,  24, -22,  -6,   6,   0,   0,   7,  -8, -20,  29,   2,   0,   2,   1,   7,  24,  41, -24, -25,  -3,  -8, -18, -15, -18, -16, -36,  24,   8,   8,   0,  -3,  -3,  -2,  40,  13,  -7,   7,  14,   7,  27,   0,  -8, -17,  -6,   3,   0,  -9,  -8, -10,  -6,   4,   6 },
     {    2,   3,   8,   6,  -5,   0,   3,  -3,   4,  11,  11,  -6,   5,   4,  -9,   6,   7,  11,  24,  14, -12, -10,  -2,  -1,   3,  22,  20,  -3,   0,   7,   1,  -5,   7,   0,  35,   2, -13,  -9,  -6,  -1, -14,  23, -14, -26, -13,   8,   5,   1,  23, -20,  60, -33,  19,   9,  -6,   0,   0,  39, -38, -14,  25, -17,  -1,  -2 },
     {    2,   4,   2,   4,  13,   4,  29, -20,  -4,  -5,  -8, -13, -29, -46,  19, -13,   5,   8,   9,   9,  38,  20, -29,  24,  -1,   1,  -2, -13,  -1,  53, -23,  12,  -3,  -1,   0,  -6, -36, -32,  -1,   2,   2,   5,   5,  21,  13,  -8,  21,  -4,  -1,  -3,  -2, -10,   9,   5,   1,  -7,   4,  -2,   0,   1,  -1,   0,  -8,   4 },
     {    2,   2,   0,   1,  17,  37,  29, -41,  -1,  -2,  -3,   9,   4, -19, -27,  13,  -5,  -9, -11, -16, -21, -34, -27,  48,   6,   6,   4,  27,  11,  10,  33, -12,   0,  -6,  -9,  -9,  12,  10,  10, -34,   6,   7, -11,   8,  -8,  -2, -23,  -1,  -4,   3,  -8,  -8,  -3,  -3,  -5,  21,   2,  -2,  -8,   0,  10,   0,  10,  -3 },
     {   -3,  -5, -18, -34,  11,  16,   5,  12,  -3,  -2, -14, -26,  -4,  24,   7,   3,  -1,   1, -14, -34,  27,  -9, -17,   6,   6,  16,  19,   6,  22,  -6, -18,  14,   4,  10,  12, -15,  42,  -7,   0,  10,   1,  12,  45,  -4,  -2, -26,   7,  -5, -13, -38,  -1,  -3, -24,   1,  11, -10,  -5,  19, -20,  10,   4,   7,  -3,   3 },
   }
 },
 {  //15
   {
     {  120, -36,  -7,  -5,  -2,  -2,  -1,  -1,  12,  -7,   0,  -1,   0,   0,   0,   0, -17,   0,   2,   0,   1,   0,   0,   0,  -8,   2,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -8,  12,  -2,   1,  -1,   0,   0,   0, 116, -19, -10,  -4,  -4,  -2,  -2,  -1,  26, -16,  -1,  -1,   0,   0,   0,   0, -30,  -4,   4,   0,   1,   0,   1,   0, -16,   3,   2,   0,   1,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {  -19,   9,   2,   1,   0,   0,   0,   0,  15, -20,   3,  -1,   1,   0,   0,   0,-114,   6,  14,   3,   4,   1,   2,   0, -34,  24,   3,   1,   1,   0,   0,   0,  13,  10,  -3,   0,  -1,   0,   0,   0,   8,   2,  -2,   0,  -1,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    4,  11,  -3,   0,  -1,   0,  -1,   0,  38,  -8,  -4,  -1,  -1,  -1,   0,   0, -22,  17,  -1,   1,   0,   0,   0,   0, 101,   2, -15,  -2,  -4,  -1,  -2,   0,  46, -30,  -5,  -1,  -1,   0,  -1,   0, -15, -12,   5,   0,   1,   0,   0,   0,  -7,  -1,   2,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0 },
     {   -6,  -6,   2,   1,   1,   0,   0,   0,  -3, -15,   4,   0,   1,   0,   0,   0, -26,   7,   3,   1,   1,   1,   0,   0,  27, -27,   0,   0,   0,   0,   0,   0, -88, -21,  20,   1,   5,   1,   2,   0, -55,  35,  12,   0,   2,   0,   1,   0,  12,  24,  -6,  -1,  -1,  -1,   0,   0,   8,   0,  -4,   0,  -1,   0,   0,   0 },
     {  -29, -99,  28,   5,  10,   3,   4,   2,   6, -49,  19,   0,   4,   1,   1,   0,  14,  24,   3,  -4,  -2,  -2,  -1,  -1,   6,  29,  -7,  -3,  -2,  -1,   0,  -1,   4,   9,  -7,   0,  -1,   0,   0,   0,   2,   1,  -3,   1,   0,   0,   0,   0,   1,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -5, -10,   0,   0,   1,   0,   0,   0,  -9,  -4,   2,   2,   0,   1,   0,   0,   0, -15,   3,   1,   1,   0,   1,   0, -30,  -6,   3,   2,   1,   1,   0,   1,  16, -32,  -3,   3,   0,   1,   0,   1, -59, -48,  22,   4,   5,   2,   2,   1, -69,  21,  27,  -1,   5,   0,   2,   0,  -6,  37,   2,  -4,   0,  -1,   0,  -1 },
     {    2,   4,   4,  -4,   0,  -1,   0,  -1,  16,  63,  -2,  -9,  -3,  -3,  -2,  -2,   3,  88, -10,  -9,  -5,  -4,  -2,  -2, -16,  42, -22,  -4,  -2,  -1,  -1,   0, -24,  -7, -16,   2,   1,   2,   1,   1, -16, -16,  -1,   4,   1,   2,   0,   1,  -2,  -3,   2,   1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   -1,   4,  -2,   1,   0,   0,   0,   0,   2, -14,   3,   0,   1,   0,   1,   0, -11,  22,  -4,   1,  -2,   0,  -1,   0,  25, -33,   1,   0,   0,   0,   0,   0, -51,  27,   3,   1,   1,   0,   1,   0,  64, -30, -11,   2,  -3,   1,  -1,   1, -57, -25,  23,   1,   3,   1,   2,   0,  -9,  28,   3,  -2,   0,  -1,   0,   0 },
     {  -16, -61,  11,   6,   7,   1,   3,   1,  19,  80, -25,  -4,  -6,  -3,  -2,  -1, -26, -20,  -7,   7,   0,   2,   0,   2,   3, -52,  13,   7,   3,   3,   1,   1,  13,  -5,  14,  -1,  -1,   0,   0,   0,  12,   6,   3,  -4,   0,  -1,  -1,   0,   2,   2,  -1,  -2,   0,   0,   0,   0,  -3,   2,  -1,   0,   1,   0,   0,   0 },
     {   10,  20,  92, -22,  -1,  -8,  -2,  -3,   4,   9,  52, -22,  -1,  -6,   0,  -3,  -6, -23, -34,  -9,   6,   3,   3,   0,   0, -11, -35,   8,   6,   3,   1,   0,   1,   8,  -7,  10,   0,   1,   0,   0,  -4,   3,   0,   2,  -2,  -1,   0,   0,  -4,  -3,   0,   0,  -1,   0,   0,   0,  -2,  -1,   0,   1,   0,   0,   0,   0 },
     {    1,   4,  -1,  -1,   0,   0,   0,  -1,  -3, -30,   5,   1,   1,   1,   1,   1,   2,  41, -20,   1,  -4,  -1,  -1,  -1, -32, -43, -18,  16,   0,   5,   1,   2,  16, -74,  19,  14,   1,   5,   0,   2,  43,  -1,  27,  -5,  -2,  -2,  -1,  -1,  22,  14,   2, -10,  -1,  -2,  -1,  -1,   0,  -3,  -5,  -2,   1,   0,   1,   0 },
     {   -3,  -7,  -3,   1,  -1,   1,   1,   1,  -5,  -6,  -5,   4,   0,   1,   0,   1,  -6, -14,  -3,   2,   0,   1,   1,   0,  -7,  -7,  -6,   3,   0,   1,   0,   0, -17, -30,  -5,   5,   2,   2,   1,   1, -22, -29,  -5,   9,   1,   2,   0,   1,   8, -82,   4,  15,   1,   5,   0,   2, -40, -54,  36,  11,   4,   3,   1,   1 },
     {    0,   1,   2,  -8,  -5,   1,   0,   0,   4,  11,  45, -12,  -3,   0,  -1,  -1,   4,   5,  82, -27,   3,  -5,  -3,  -2,  -4, -32,  13, -29,   7,  -3,   0,   0,   1, -28, -44,  -8,   7,   2,   4,   1,   9,   1, -30,  15,   5,   3,   2,   0,   6,   7,  -3,  11,   0,   0,   0,   0,  -1,   1,  -1,   2,  -2,  -1,   0,   0 },
     {    1,   7,  -2,  -4,   1,   1,   0,   0,   0, -21,   7,   4,   2,   1,   1,   1,   0,  51, -19,  -5,  -4,   0,  -1,   0, -10, -67,  24,   6,   4,   2,   1,   1,  32,  56,  -2, -14,  -2,  -5,  -1,  -2, -46, -12, -14,   6,  -1,   2,   1,   1,  18,  -9,   0,   6,   0,   1,   0,   1,   7,  -5,   0,   0,  -1,   0,   0,   0 },
     {    1,   5,  -1,  -4,   1,   0,   0,   0,  -1,   0,   0,   2,   5,  -1,   2,   0,   6,  22,   5,  -3,   3,  -1,   0,  -3,  -8,  -5,  -4,  -3,   2,  -1,  -2,  -1,  22,   1,  14, -10,   4,   2,   0,  -1,  -3,  78,  -6, -18,   1,  -5,   0,  -3, -67,   2, -17,   3,   3,   0,   3,   0, -52, -23,  11,  13,   2,   2,   1,   1 },
     {   -7, -12, -74,  21,   4,   7,   1,   3,   8,  16,  83, -17,  -2,  -8,  -1,  -2,  -4, -15, -25, -10,   4,   0,   2,   0,   0,   0, -33,  15,   2,   5,   2,   2,   2,  11,  16,   5,  -2,  -1,  -2,   1,  -3,  -1,  -3,  -2,  -2,   0,   1,   0,  -1,   0,  -2,   0,   1,   1,   0,   0,  -1,  -1,   0,   1,   2,  -1,  -1,   0 },
     {    8,   9,  25,  76,   9,  -8, -12,  -4,   6,   9,  27,  78,   4, -16, -14,  -4,   1,   3,   9,  33, -12, -11,  -7,   0,  -2,  -3,  -4,  -3, -11,  -3,  -1,   4,  -1,  -2,  -5, -12,  -6,   0,   5,   2,  -1,   0,  -2,  -9,  -1,   2,   3,   1,   1,  -2,   1,  -7,  -2,   1,   1,   0,   0,  -1,   0,  -3,   0,   1,   1,   0 },
     {    1,   1,   0,  -2,  -1,   1,   1,   0,  -1,   1,   0,   1,   0,   1,   0,   1,   3,   3,   2,  -1,   1,  -2,   0,   0,  -4,   5,  -4,   2,   1,  -2,  -1,   0,   9, -14,  10,  -1,   2,  -2,   0,   0, -13,  38, -10,  -5,  -2,   0,  -1,  -1,  26, -58,   3,  11,   2,   2,   0,   2, -12,  95,  12, -24,   0,  -7,   1,  -3 },
     {    3,   7,  17, -15,  -4,  -1,   4,  -1,   6,  11,  39,  -9,  -9,  -1,   0,   1,   7,  14,  40,   5,  -9,   2,  -4,  -3,   8,  19,  58,  12, -12,   3,  -6,  -3,   4,   7,  69,   1, -12,  -1,  -7,  -3,   2, -13,  41, -15, -13,  -2,  -3,  -1,   5, -13,  10, -15,   1,  -3,   0,   1,  -7,  -9,   6,  -5,   2,  -1,   1,   0 },
     {    0,  -1,   1,   2,  10,  -4,   0,  -1,  -1,   0,   0,  -7,   8,  -4,  -2,  -1,   2,   1,   6,   1,   6,   0,   0,  -2,  -6,  -2,  -3,  -2,  -2,   1,   1,  -2,   7, -16,  14, -10,   1,   5,   1,   0,  -5,  18,   1,  -5,   0,  -1,  -3,   0, -36, -54,  -6,  18,   0,   4,  -1,   3,  87, -19, -48,   7,  -4,   3,   0,   2 },
     {    5,   4,  13,  71,  -5,  -6, -13,  -4,  -4,  -4, -16, -21,  -7,   5,   3,   4,  -1,   1,  -9, -83,  11,   7,  18,   3,   3,   5,  25,   8,  17,  -1,   2,  -8,   1,  -7,  12,  28,  -4,  -5,  -8,  -4,  -2,  -1, -15,  -7,  -8,  -2,  -1,   2,  -4,   0,  -3,  -8,   2,   2,   1,   0,   3,   0,   5,   4,   3,   2,   0,   0 },
     {    0,  -1,  -6,   4,   1,   6,  -6,   0,  -1,   1, -17,   7,  -1,   6,  -2,  -2,   2,   1,  41, -16,  14, -10,   3,  -9,   0, -17, -50,  -4,   5,   8,  11,  -4,   4,  37, -20,  38, -11,  13,   0,  -3,  -6,  15,  66,   0, -23,   2, -11,   0,   5, -16,  25, -26,   4,   1,  -1,   2,   0,  -5,  -4,  -6,   7,  -2,   1,   0 },
     {   -1,  -5,   7,  11, -76,  48,  -2,   0,   0,   0,   5,   8, -57,  43,  -5,  -4,   0,   2,  -4,   8,   7,   3,  -4,   2,  -1,   0, -11,  -7,  22, -13,  -5,   5,   1,   2,   4, -21,  24, -10,   2,   5,   0,   1,   6, -14,  12,  -7,  -1,   1,  -2,  -5,   6,  -3,   0,  -3,  -1,  -2,   8,  -4,  -4,   1,  -5,   1,  -1,  -1 },
     {    0,   1,  -3,  -9,  21,  -9,  17,  -9,   1,   1,   2,  28, -17,  -4,   7,  -7,  -5,  -4, -11, -29, -53,  27, -12,   5,  -1,   1,  -6, -62,  12,  21, -10,   7,   6,   7,  21,  19,  52,  -5,  -4,   2,   1,  -2,  16,  22,   9, -14,  -3,  -4,  -2,  -5, -11,   0,  -9,  -6,  -3,  -2,   0,   2,  -2,  -1,  -3,  -2,   2,   1 },
     {    0,   1,  -5,  10,   4,  18,  -9,   6,   1,   1,  20, -13,   4,  -6,  -1,  -5,  -3,   1, -43,  12,  18,   6,  10,  -8,   4,   3,  62, -29, -14,  -8,  -4,  -3,  -4,  -7, -58,   6,  13,  18,  -1,   2,   6,  17,  47,   7,   3,   2,  -6,   0,  -1, -16,  -4, -15,   8,  -5,  -1,  -3, -10,   2, -18,   0,   3,  -2,   1,   0 },
     {   -1,  -1,  -2,   3, -17,  -4,   4,   4,  -1,  -3,   2,  -4, -43,   1,   3,   4,   1,   2,  10,  22, -56,  23,   7,  10,   1,  -1,  11,  53, -29,  15,   0,   0,   1,   1, -28,  56, -14,   1,   0,   1,  -5,  16, -14,  23,   0,  -1,  -1,   1, -13,  -4, -16,   1,  -4,   0,  -5,   0,   5,   1,  -7,   1,   0,   5,  -3,   1 },
     {   -1,  -1,   1,   0, -13,  -1,  -8,   5,  -4,  -6,  -6, -17, -23, -13,   0,   4,  -4,  -8,  -8, -25, -46, -20,  22,   4,  -4,  -5, -13, -32, -60,  16,  21,   5,  -1,  -1,   1, -44, -22,  33,   9,   0,  -1,  -2,   0, -30,   1,  20,   0,   0,   4,  -1,   6,  -9,  13,  11,  -1,   0,   0,   1,   3,   3,  15,   1,  -4,  -2 },
     {   -1,  -1,  -1, -15,  -4,   4,  -5,   7,   0,  -1,   0,  15,   8,   8,   6,   1,  -1,  -2,   7,  -4,   2,   4,  -4,  -3,  -1,  -5,  -8,  -6,   4, -14,  -6,   5,  -6,   0,   2,  16, -10, -12,   2,  -3,  -6, -33,  -7, -33,  -8,   2,   0,   2,  -2, -12, -79, -33,  30,  10,   7,   3, -22,  13, -48,   8,  26,  -1,   2,   0 },
     {    5,   7,   5,   1,  55,  59,  12, -13,   0,   1,  -2, -12,   7,  27,  -3, -23,  -2,  -4,   4,  -7, -23, -35,   4,  -2,  -1,  -3,  -1,  16, -22, -48,   1,  13,   1,   1,   1,  10,  20, -13,   6,  11,  -2,   0, -11,  -8,  15,  -1,  -8,  -1,  -4,  -2,   1, -25, -10,   9, -14,  -8,   7,  -5,  12, -17,  -4,   6,   0,   0 },
     {   -1,   0,  -7, -27,  43, -25,  11, -17,   1,   0,   6,  36, -53,  10,  -6,  13,   0,  -1,  -4, -19,  18,  12,  32,  13,   2,   3,   5,  22,  10,  -2, -13,  -8,  -5,  -5, -19, -18,  -1, -16, -21,   0,   2,   7,   0, -28,  13,  -1,  18,   4,   6,  -3,  32,  -8,  24,  10, -14, -11,  -4,  -5, -17,   3,   8,  -5,  -7,   2 },
     {    3,   5,   5,  30,  40,  -2, -51,  20,  -4,  -5,  -8, -45, -23,  14,  25,  -4,   3,   2,  15,  28, -19,   5,  15,  -4,  -2,  -1, -15,   2,  24,  13, -28,  -5,  -1,   5,   3, -34,   3,  -6,   7,   8,   1,  -7,  17,  15,   0, -14,  17,  11,   4,  -1,  -9,  15,  -1,  -1,   4,  -4, -18,   0, -19, -15,   6,  -3,   4,   4 },
   },
   {
     {   88, -35,  -2,  -4,  -1,  -1,  -1,  -1, -78,  23,   5,   3,   2,   1,   1,   0,  25,   2,  -5,   0,  -1,   0,  -1,   0,  -4,  -6,   2,   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   49,  -2,  -5,  -2,  -2,  -1,  -1,   0,  18, -29,   5,  -1,   2,   0,   1,   0, -85,  37,   4,   2,   1,   1,   1,   0,  58,  -6, -11,   0,  -3,   0,  -1,   0, -20, -10,   7,   0,   1,   0,   1,   0,   4,   7,  -1,  -1,   0,   0,   0,   0,  -3,  -2,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
     {   54, -24,   1,  -2,   0,  -1,   0,   0,  63,   2, -10,  -2,  -3,  -1,  -2,   0, -23, -33,  10,   0,   3,   1,   1,   0, -56,  38,   3,   0,   1,  -1,   0,   0,  42,  -7, -12,   2,  -2,   0,  -1,   0, -18,  -8,   8,   0,   1,   0,   0,   0,   2,   5,  -2,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 },
     {   33,  83, -29,  -4, -10,  -2,  -3,  -1, -11, -72,  14,   6,   6,   3,   2,   2,  27,  -6,  11,  -6,   0,  -2,   0,  -1, -21,  19,  -8,   0,   0,   0,   0,   0,  -4,  -2,   1,   2,   0,   0,   0,   0,   7,   0,  -1,  -1,   0,   0,   0,   0,  -3,  -2,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {   29, -16,  -1,  -1,   0,   0,   0,   0,  54,  -7,  -1,  -3,  -1,  -1,   0,  -1,  31,   4, -13,   0,  -3,   0,  -1,   0, -16, -48,  17,   1,   4,   1,   2,   0, -53,  53,   5,  -3,   1,  -2,   0,  -1,  42, -12, -17,   3,  -3,   1,  -1,   0, -16,  -7,  11,   0,   1,   0,   0,   0,   3,   6,  -3,  -1,   0,   0,   0,   0 },
     {  -20,  -6,   0,   2,   1,   2,   1,   1, -31, -20,  19,  -1,   3,  -1,   1,   0, -41,  48, -14,  -1,  -2,   0,   0,   0, -47, -14,  11,   6,   3,   2,   1,   1,  50,  36, -10,  -5,  -4,  -2,  -1,  -1,  21, -39,  -6,   3,   0,   1,   0,   1, -30,  12,  13,  -1,   2,  -1,   1,  -1,  12,   5,  -8,   0,  -1,   0,   0,   0 },
     {    4, -23,   0,   1,   2,   1,   1,   1,  26, -24,  14,  -1,   1,  -1,   0,   0,  38,  66, -20, -10,  -5,  -4,  -1,  -2, -20, -34, -21,  14,   0,   4,   0,   2,  22, -37,  31,  -2,   0,   0,   0,   0, -37,  31,   1,  -8,   3,  -2,   1,  -1,  16,   2, -12,   6,  -1,   1,   0,   0,  -1,  -6,   4,   0,  -1,   0,   0,   0 },
     {   16,  16,  -9,  -1,  -2,   0,  -1,   0,  26,  11,  -3,  -2,  -2,  -1,   0,   0,  40,   1, -12,   2,  -2,   1,  -1,   0,  65, -16,   1,  -2,  -1,  -1,  -1,  -1,  47, -11, -18,  -3,  -2,   0,  -1,   0, -15, -54,  13,   5,   3,   2,   1,   1, -37,  26,  15,  -4,   2,  -1,   1,  -1,  12,   3, -10,   1,  -1,   0,   0,   0 },
     {   -7, -14, -78,  29,   3,   7,   0,   3,   7,  22,  77, -22,  -6,  -6,  -2,  -2,  -6, -22, -22,  -6,   8,   1,   3,   1,   4,  13,  -9,  16,  -4,   0,  -1,   0,  -5,  -3,  14,  -9,  -2,   0,   0,   0,   4,  -1,  -9,   2,   3,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0,   0 },
     {  -11, -47,  12,   3,   4,   1,   1,   1,   1, -26,   9,  -1,   2,   0,   1,   1,  25,  18,  15,  -6,  -1,  -2,  -1,  -2,   0,  53, -35,  -7,  -1,  -1,   0,  -1, -24, -36,  -7,  22,   0,   4,  -1,   2,  56, -16,  23,  -8,  -4,   0,  -2,  -1, -33,  12,   0,  -9,   5,  -1,   1,   0,   6,   5,  -7,   5,  -1,   0,   0,   0 },
     {    3,   4,   2,  -3,   1,   1,   0,   0,   3,  14, -12,   2,  -3,   1,  -2,   1,   6,  -1,  10,  -5,   1,  -2,   0,   0,  11,   6, -20,   7,  -3,   2,  -1,   1,  42, -10,  22,  -5,   0,  -2,  -1,  -1,  31,  15, -43,  -6,  -1,  -1,   0,  -1, -28, -75,  23,  19,   1,   5,   0,   2, -11,  40,  18, -15,   1,  -3,   0,  -1 },
     {   -6, -14, -42,  10,   1,   5,   2,   2,  -1,   6,  -9,  17,  -4,   1,  -1,   1,   1,   4,  84, -31,  -2,  -5,  -2,  -2,   0, -34, -39, -10,  14,   1,   5,   1,   7,  31, -15,  28,  -6,  -1,  -1,   0, -16,   0,  15,  -7,  -6,   2,  -1,   0,   4,  -4,  -1,  -5,   4,   0,   0,   0,   4,   2,  -2,   1,   0,  -1,   0,   0 },
     {   12,  54, -17,  -1,  -3,  -1,  -2,  -1,   7,  73, -22,  -2,  -4,  -3,  -2,  -2, -16,  28,  -8,   1,  -3,  -1,   0,   0, -35, -16, -17,   3,   1,   3,   1,   0, -16, -40,   8,  17,   2,   3,   0,   1,  30,  -3,  23,  -8,  -2,  -2,  -1,   0, -13,  14,  -1,  -7,   3,  -2,   1,   0,   5,   0,  -7,   3,   0,   0,   0,   0 },
     {    8,  12,  -4,   0,   0,  -2,  -1,  -2,  14,  21,   1,  -5,  -1,  -2,   1,  -1,  18,  39,  -4,  -4,  -3,  -2,  -2,  -1,  23,  35,  -5,  -7,  -2,  -3,  -2,  -1,  35,  37, -26,  -1,  -4,   0,   0,   0,  48,  12, -13,   1,  -5,  -1,  -3,   0,  59,  -6, -29,  -5,  -5,   1,  -1,   0,   1, -37,  -3,   7,   1,   1,   0,   1 },
     {    6,   5,  11,  64,  12,  -9, -11,  -3,  -7, -10, -21, -78, -17,   9,  12,   5,   5,   4,  24,  48,  16,  -8,  -6,  -4,  -5, -10, -16,  -9, -16,   3,   2,   3,   4,   4,  11,  -7,   6,   2,  -1,  -1,  -1,  -6,   1,  12,   2,  -3,  -1,   0,   3,   2,  -3,  -5,  -3,   1,   1,   1,  -4,   1,   3,   3,   0,   0,  -1,  -1 },
     {   -1,  -5, -19,  11,  -4,   4,   0,   1,   2,  -7,  13,  -8,   5,   0,  -1,  -2,   3,  13, -10,  12,   0,   0,  -1,  -1,  -2, -16,  50, -35,  -5,   1,  -1,   0,   6, -21, -66,  22,  19,   2,   4,   1,   0,  38,  31,  12, -23,  -1,  -4,   0, -16, -39,  -1, -14,   9,   4,   1,   1,  -2,  14,   7,   5,   2,  -2,   1,   0 },
     {   -1,  -1,  -1,   0,   1,  -2,   0,   0,   2,   5,  -2,   2,   1,   1,  -1,   0,   4,  27,  -4,  -7,   0,  -3,   1,  -1,  -8,  15, -13,   4,  -3,   1,  -1,  -1, -24, -11, -27,  -2,   4,   2,  -1,   0, -20, -35,  -6,  41,  -4,   6,   0,   4,  51, -12,  66, -15, -13,  -1,  -3,  -1, -25,  25, -16, -31,  14,  -4,   2,  -2 },
     {    0,   2,   5,  -3,   2,  -3,   1,  -2,  -1,  -6,   6,  -2,   0,  -1,   0,   1,  -4,  -3, -25,   5,  -2,   6,  -1,   1,   0,  -9, -13,   9,   5,   0,   3,   1,  -4,  10,   0,  19, -12,   3,  -2,  -2,  -3, -54,  45, -25,   0,   2,   0,   1,  26, -29, -40,  11,  13,   0,   3,   1, -46,  27,  51,   5,  -7,  -2,   0,  -2 },
     {   -8, -24, -70,  25,   3,   9,   3,   1,  -7, -20, -77,  33,   1,  10,   1,   0,   1,   8, -20,  16,   0,   2,  -3,   1,   1,  12,  27,   3,  -1,  -4,  -4,   0,  -1, -11,  14, -12,  -1,  -1,  -2,   1,  10, -11,  -5,  -4,   5,  -2,   1,   0,   9,   4,  -5,   6,  -2,  -2,   0,  -1,  -3,  -2,   4,  -1,  -3,   1,  -1,   1 },
     {    5,   7,  24,  63, -21,   0,  -7,   0,  -2,  -1, -14,  -7,  -3,   1,   2,   3,  -1,   1, -10, -75,  27,   6,   5,  -3,   5,   3,  24,  45,  10, -17,  -7,  -3,  -4,  -3, -17,   6, -25,   7,   3,   2,   2,   9,   3, -11,   6,   4,  -1,   1,  -5,  -1,   2,   0,   4,  -4,   1,   0,   1,   1,   0,  -3,  -2,   1,   0,  -1 },
     {   -1,  -4,   2,  13, -62,  47,  -8,   7,   1,   2,  -8, -18,  71, -42,   9,  -2,   0,  -4,   8,   9, -34,  11,   8,  -5,  -1,   0,  -9,   6,   1,  12, -12,   5,   1,   0,   3, -11,  13, -13,   8,  -4,   1,  -5,   0,   5,  -8,   6,  -4,   0,   4,   0,  -4,   3,   4,  -2,   0,   2,   1,  -1,   3,  -5,  -4,   0,   0,   0 },
     {   -1,  -9,   8,  -1,   3,   0,   2,   1,  -4, -23,   8,   5,   2,   6,   1,  -1,  -9, -45,   2,   1,   2,   1,   0,   1,   4, -55,   5,  13,   2,   5,   3,   5,  33, -38,  10,  12,   3,   2,  -4,  -2,  58,  -1,  15,   4,  -5,  -2,  -3,   2,  51,  11,  17, -11,  -6,  -2,  -1,  -2,  12,  -3,  -9, -11,   3,  -2,   0,  -2 },
     {    1,   2,  -2,  -4,  -1,   2,  -4,  -5,   3,   4,   1,  -3,  -2,   0,   6,   0,   5,   7,   1,   0,   4,   3,  -1,   2,   7,  14,  10,  -4,   3,   0,  -2,  -1,   8,  22,  12,  -4,   3,  -2,   5,  -1,   5,  28,  12,  -7,   1,  -5,  -4,   0,  16,  44, -23,  -8,   3,  -3,  -1,   0,  14, 100, -18, -23,  -3,  -8,   1,  -6 },
     {    3,   4,  11,  34,   0,   2,  -9,   1,   0,  -1,   2,  14, -10,   2,   2,  -2,  -3,  -4, -28,  -7,  -7,   2,  11,   1,   4,  10,   0, -46,  37,  15,  -1,  -2,   5,   5,  34,  61,  11, -35,  -6,  -1, -10,  -7, -28,   1, -37,  12,  11,   5,   7,  11,  10,  -9,  12,  11,  -7,  -2,   6,  -2,   7,   5,   3,  -7,  -2,   3 },
     {    1,   2,   4,   6,  -1,   5,  -7,  -2,   3,   5,  13,   9,  -6,   0,   1,   1,   2,   8,  32, -13,  -9,   0,   4,  -3,   1,  -6,  31, -23,  12,  -3,  -3,   0,   9, -15,  15, -25,   4,  -4,   5,   2,  19,   5, -13,  12,  10,   3,   2,  -1, -15,  36,  -5, -15,  -6,   0,   0,  -1, -93,  -4,  13,   6,  14,   3,   3,  -2 },
     {    4,   5,   3,  -7,  43,  56,   9,   0,  -6, -11, -10,  -8, -43, -72, -16,  11,   2,   0,  -3,   6,   7,  26,  26,   0,   0,  -3,  -3,   4,  16,   5, -16,  -8,   0,  -2,  -7,   2, -19, -12,   1,   9,   1,   6,   3,   5,  15,   5,   1,  -4,   0,  -6,   0, -10,  -6, -10,   0,   1,   2,   1,  -5,   4,   4,   4,   2,  -1 },
     {   -1,  -2,  -2,  -5, -12,  -4,   5,  -1,  -2,  -5, -10,  -1,   5,   5,  -3,   1,  -1,  -8, -18,   7,   7,   1,  -4,   0,   5,   3, -12,  17,   1,   7,  -5,   7,  10,  21,  24,   2,  -4,  -1,   2,   0,  -6,  10,  27, -16, -11,  -3,  -2,  -1,  -6, -29,   6, -44,  19,  -4,   2,  -3, -35,  -2, -75,  37,  30,  -8,   7,   0 },
     {   -1,  -2,   2, -12, -27,  44, -15,   6,   1,   1,   1,   5,  -2,   0,  -5,  -1,   1,   2,  -2,   1,  51, -64,  26,  -4,   1,  -1,   8, -16, -32,  30,  17,  -9,   0,   1,   3,  18, -12,  27, -28,   9,  -1,  -5, -12,  -9,  12, -14,   4,   9,   5,   5,   2,  -3,   5,  -5,   8,  -9,  -6,   2,   1,   7,  -8,   2,  -1,  -1 },
     {   -5,  -7, -19, -50, -19,   8,  21,   0,  -6,  -8, -22, -69,  -3,  16, -14, -12,  -2,  -3, -10, -37,  27,   6, -14,  12,  -1,  -5,  -2, -22,  42, -11,  10,  -2,  -1, -13,  -5,  -4,  13, -12,  -1, -13,   2,  -8, -16,   3,   5,   7,   8,   2,   3,   0, -11,   5,   8,  -4,  -1,  -4,  -3,  -1,  -6,   0,  -1,   3,  -2,   2 },
     {    0,   0,  -8,  -4,  51, -33,   5, -18,  -1,  -1,  -9, -25,  43, -14,  10,  -6,  -1,  -1,   5, -27, -22,   7,  28,  11,   1,   0,  14,   8, -27,  36,  -7, -20,  -1,  -4,  -9,  15,  22,  -6, -21,  13,  -3,  -5, -20, -28,  12, -16,  14,   7,   5,   3,   4,  -8,  13, -11,   5, -11, -10,   3,  -8,   7, -17,   7,  -2,  -1 },
     {    1,   1,   3,  -7, -13,  10,   4,  -2,   2,   7,  16, -11, -25,   0, -10,  -7,   6,  21,  47,  -3, -10,  12, -17,   3,   1,  14,  64,   4, -19,  10,  -2,   2,  -8, -15,  35, -11, -12, -25, -13,   2,  -3, -30,  15, -23, -15,  -1,  10,   2,  19, -22,   7,   7,  10,  12,  -8,   3,  25,   0, -10,  10,  -1, -12,  -6,   6 },
     {    1,   2,   6,   8,   4,   5,  17,  -9,  -1,  -1,  -5,  19,  -8,  -4, -28,  -2,  -1,  -3,   2, -22,  19, -11,  -3,  25,  -3,  -5, -12,  22, -44,  14,  12,  -1,  -2,  -5,  -1, -14,  46, -31,  16, -13,   0, -15, -12,  43, -26,  14,  -5,  -3,  -3,   0, -36, -19,   2,   0,  -2,  15,   8,  11,  11,  29,   5,   0,  10,  -7 },
   },
   {
     {  -90,  77, -37,  25, -13,   7,  -4,   2,   5,  -4,   2,  -1,   0,   0,   0,   0,   1,  -3,   3,  -3,   2,  -1,   1,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   80,  36, -73,  40, -32,  15, -11,   4,   6, -11,   7,  -4,   2,  -1,   0,   0,  -8,  -2,   6,  -4,   4,  -3,   2,  -1,  -2,  -1,   1,   0,   0,   0,   0,   0,  -4,   0,   1,  -1,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,  -1,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    4,   7, -11,   6,  -5,   2,  -1,   0, -93,  74, -35,  23, -10,   5,  -2,   1,  12,  -2,   0,  -1,   0,   0,   0,   0,   3,  -4,   3,  -2,   2,  -1,   0,   0,   2,  -3,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,  -1,   1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   -9, -10,   3,   6,  -3,   4,  -2,   2, -77, -43,  66, -33,  26, -11,   8,  -2, -35,  19,   0,   0,   2,  -2,   2,  -1,  11,   6,  -6,   4,  -3,   2,  -1,   0,   5,   5,  -3,   1,  -1,   0,   0,   0,   4,   1,  -2,   1,  -1,   1,  -1,   0,   1,   1,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
     {    0,   8,   0,  -5,   3,  -3,   1,  -1,  14,  18, -25,  15, -11,   5,  -3,   1,-104,  55, -20,  18,  -7,   4,  -1,   1,  -8,   2,   4,  -2,   2,  -1,   1,   0,  -5,  -3,   4,  -2,   2,  -1,   0,   0,   1,   1,   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    4,   4,  -3,   1,  -2,   1,  -1,   0,  10,   4,   1,  -6,   3,  -4,   2,  -1,  31,  48, -42,  18, -13,   3,  -3,   0,  96, -31,  -7,  -1,  -4,   1,  -3,   1,   3, -15,   3,  -2,   1,   0,   1,   0,   2,  -5,   1,   1,   0,   1,   0,   0,  -3,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   11,  23,   5, -25,  21, -19,  11,  -6,  -1,  -5,  -8,  10,  -6,   5,  -2,   2, -41, -45,  61, -31,  21,  -6,   3,   0,  70,  -5,  -8,  -2,  -1,  -2,   0,  -1,   1,  -1,  -2,   4,  -4,   3,  -2,   1,   4,   1,  -2,   0,   0,   0,   0,   0,  -1,  -2,   0,   0,  -1,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0 },
     {  -29, -64,   7,  52, -49,  48, -28,  16,  13,   8, -14,   7,  -4,   1,  -1,   0, -18,  -8,  21, -18,  17, -13,   9,  -5,  26,   2,  -5,  -1,   0,  -1,   0,  -1,   1,   0,  -2,   0,   1,  -1,   1,   0,   0,   2,   2,  -1,   1,   0,   0,   0,   1,   1,  -1,   0,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {    3,   4,  -1,   1,   1,   0,   1,   0,   6,   4,  -5,   0,   2,  -3,   2,  -1,  -2,   2,  -2,   2,   0,   0,   1,   0,   1,  38, -17,   5,  -4,   0,   0,  -1, 115, -14, -17,   3,  -6,   1,  -2,   0,  -4, -25,   7,   2,   1,   3,  -1,   0,  -4,   1,   2,  -2,   0,  -1,   0,   1,  -1,  -2,  -4,   1,   0,   1,  -1,   0 },
     {    3,   8,   2,  -3,   0,   0,  -2,   0,   2,   8,  -6,   3,  -1,   0,   0,  -1,  10,  16,  -6,  -2,   2,  -3,   1,  -2,  12,  74, -68,  32, -19,   5,  -2,  -1, -41,  33, -25,  17,  -9,   4,  -1,   1, -14,  -2,   2,   1,   1,   1,   0,   0,  -8,  -3,   5,  -2,   1,   0,   0,   0,  -1,   1,   2,  -1,   1,   0,   0,   0 },
     {   -4, -10,  -9,   8,   2,  -2,   3,  -1, -26, -72, -15,  64, -47,  41, -21,  11,   8,  -4, -12,  17, -14,  11,  -7,   3,   8,  12,  -2,  -7,   9,  -9,   6,  -4,   4,   7,   1,  -5,   5,  -4,   2,  -1,   4,   3,   0,  -3,   3,  -2,   1,   0,   0,   3,   1,   0,   1,  -1,   0,   0,   1,   1,   0,  -1,   1,   0,   0,   0 },
     {   -3,  -3,  -1,   1,   2,  -2,   2,   0,  -6,  -9,  -2,   2,   2,  -1,   2,   0,  -5,  -9,   0,   2,   1,   1,   0,   0,  -3,  -9,  -2,   2,  -1,   1,   0,   0, -11, -37,  14,  -1,   2,   2,  -1,   1,-101, -19,  32,  -2,   7,   1,   2,   0, -45,  16,  15,  -5,   3,  -1,   1,   0,  -8,   8,   0,   0,   1,   0,   0,   0 },
     {   -1,   1,   0,   1,  -1,   2,  -1,   1,   0,  -1,  -5,  -1,   5,  -4,   3,  -2,  -5,  -7,  -7,   9,  -4,   3,  -1,   1,   6, -26,  32, -22,  14,  -7,   4,  -2,   2,  80, -61,  29, -15,   3,  -1,  -1, -14, -38,  17,   0,   1,   5,  -3,   2, -15,   9,   8,  -7,   3,  -3,   1,   0,   0,  -1,  -6,   2,  -1,   1,  -1,   1 },
     {   22,  55,  67, -32, -37,  42, -42,  22,  -9, -19, -13,   9,   5,  -5,   5,   0,   0, -11, -20,   5,  12, -13,  14,  -7,  -2,  -1,   3,  -3,   3,  -1,   1,  -1,   1,  -2,   6,  -1,   3,  -3,   1,  -2,  -2,   2,   2,  -1,   0,   0,   0,   0,   3,   0,  -3,   0,  -1,  -1,   0,   0,   3,  -2,   0,  -1,  -1,   0,   0,   0 },
     {    2,   5,   1, -10,   6,  -4,   1,   0,   0,  -7,  11,  16, -22,  15,  -8,   4,  21,  71,  18, -63,  46, -36,  14,  -7, -19, -22,   3,   5,   2,  -2,   4,  -2,   5,   0,  -6,  10,  -9,   9,  -7,   3,  -4, -11,   5,   0,   2,  -2,   1,   0,   1,   0,   2,  -1,  -2,   2,   0,   1,   3,   0,  -4,   1,   0,   0,   0,   0 },
     {    0,   0,   4,   0,  -2,   2,  -2,   0,  -1,   0,  -1,   2,  -2,   0,   2,  -1,   0,  -3,   4,   1,   0,   0,   0,   0,  -7,  -9,  -9,  11, -11,   7,  -2,   1,  -5,   0,  21, -11,   4,  -1,   0,   1,  56, -34,  -9,   7,  -6,   2,  -2,   0, -98,   8,  23,  -3,   4,  -1,   1,   1,  17,   8,  -9,   0,   1,  -1,   0,   0 },
     {    5,  12,  21,  -2, -19,   8, -12,   2,  13,  41,  70,  -9, -52,  40, -32,  13,  -3,  -3,  21,   9, -29,  20, -14,   9,  -1, -12, -11,   5,  -1,   0,   4,  -2,   1,  -8, -16,   5,   4,  -3,   3,   0,  -5,  -9,  -3,  -2,   4,  -3,   2,  -1,   1,   5,   1,   1,   0,  -1,   0,  -1,  -2,   0,  -1,   0,   1,  -1,   0,  -1 },
     {    2,   4,   2,   1,  -3,   0,   1,  -1,   1,   3,   0,   5,  -4,  -1,   2,  -2,   2,   7,   2,  -2,   1,   0,  -3,   3,   1, -13,   4,  10,  -9,   6,  -4,   2,  25,  30, -11,  -1,  -3,   0,   0,   0, -27,  84, -54,  19, -10,   0,   1,  -1, -41,  26, -16,   9,  -1,   0,   0,  -1, -12,  14,  -2,   0,   1,  -1,   0,   0 },
     {   -2,  -4,  -2,   8,   0,  -3,   5,  -3,  -5, -11, -11,   7,   1,   2,   0,   0, -15, -30, -29,   5,  10,  -1,   1,   0, -18, -66, -44,  56, -18,  16,  -4,   3,   4, -22, -27,  36, -17,  11,  -7,   3,  10,  -1,   0,   8,   0,  -1,   0,   0,  19,   5,  -1,   1,   0,   0,  -1,  -1,   9,   0,   1,   0,  -3,  -1,  -1,  -1 },
     {   -1,  -1,   1,   0,   0,   2,  -5,   3,  -3,  -4,  -1,   2,  -3,  -3,   0,   0,  -2,   0,   2,   4,   0,   2,   0,   1,  -4, -10,   3,   3,   0,  -1,   3,  -2,   3,  -3, -10,   5,  -5,   5,  -1,  -1,  -3,   1, -15,   4,  -4,  -1,  -1,   2, -18, -92,  26,   5,  -1,   9,  -1,   3, -68,  -4,  36,  -1,   8,   0,   2,   0 },
     {   -3,  -7, -14,  -2,  12,  -8,   2,   3,   3,  16,  20, -24, -22,  26, -21,   8, -11, -38, -66,   4,  49, -37,  28, -13,   9,  19,  10, -10,  12, -17,  14,  -7,  -2,   1,   9,  -5, -12,  10,  -7,   4,   2,   3,  -6,   3,  -5,   3,  -1,  -1, -12,   5,  -8,   4,   0,   4,  -2,   1, -10, -11,   4,   6,  -3,   0,  -1,   1 },
     {   12,  21,  57,  72, -25, -46,  36, -32,  -1,  -2,  -4,  -7,  -8,   1,  11,   0,  -3,  -3, -20, -26,   8,  12,  -8,   6,   0,   3,  -3, -12,   7,   6,  -2,   0,  -1,   5,   4,  -4,   6,  -2,  -2,   0,  -1,   2,   7,  -4,  -1,  -2,  -1,   0,  -3,  -9,  -4,  -2,  -3,   1,   1,   1,   1,  -4,  -1,  -1,   1,   1,   0,   0 },
     {    0,  -1,  -3,  -1,   6,   2,   1,   4,   2,   5,   6,   1,  -5,   5,   0,  -1,  -5,  -8,  -9,  -4,  10,  -1,   0,  -1,   2, -20,  -4,  42, -28,  14,   1,  -1,  15,  62,  40, -60,  26, -18,   3,  -6, -23, -24,  -5,  -6,  11,   0,   3,  -2,  23, -16,   5,   2,   0,  -1,  -3,   2,   6,   7,  -4, -11,   4,  -4,   3,   0 },
     {    1,   2,   1,  -1,  -1,   4,   3,  -1,  -1,  -3,  -2,   3,   1,  -8,   1,  -1,   0,   2,   0,  -9,  -1,   8,  -6,   6,  -1,  -2,  -2,  12,  -6,   3,  -1,  -1,  -3,   4,   7,  -8,  -1,   3,  -3,   3,  22, -20,  19, -20,  16,  -9,   4,   0,  -3,  55, -43,  13,  -8,  -1,   1,  -2, -91,   9,  12,   4,   2,   2,   0,   0 },
     {    2,   3,  -1,   1,  27,  41,   9, -17,   5,   9,  10,  15,  27, -10,  -8,   6,  -4,  -4, -16, -33,  -3,  29,  -4,  -1,  -4, -15, -42,  -9,  45, -45,  37, -18,   4,   7,   6,  -1,  18,  -8,  11, -12,   2,   9,   3,  14,  11, -12,   9,  -4,  -6,  -5,  -1,  -9,   5,  -8,  13,   0,  -1,   4,  -4, -12,  -5,   0,   8,   3 },
     {   -1,   0,  -1, -11, -15,  -3,   5,   5,  -9, -16, -51, -70,  21,  46, -42,  40,   3,  10,   9,  -7,   2,   5, -19,  13,  -2,  -4,  -7,   4,  11,  -1,  -2,  -5,   2,   2,   1,   3,  12, -14,   4, -11,  -1,  -4,  -7,  10,  -8,  -4,   4,  -5,   3,   7,  10,  -1,   5,  -2,   4,  -1,  -6,   4,   5,   4,   0,  -4,   2,  -2 },
     {   -1,  -1,   1,   2,  -5,  -4,   4,   4,  -2,  -4,  -4,   6,   0,  -4,   7,  -5,   0,   2,   5,   2,  -1,   3,   3,  -2,  -3,  -1, -12, -17,   3,  -2,   8,  -6, -12, -12,  -7, -10,   8,   3,   4,   1, -18, -56, -59,  62, -17,  23, -10,   6,   9,  -1, -45,  30,  -8,   2,  -5,   4,  -1,  -5,  -8,  10,  -3,   1,   1,  -1 },
     {   -1,  -1,  -1,  -1,   1,   0,  -6,  -5,   0,  -2,  -3,   0,   1,  -1,   1,  -2,   0,   4,  10,   0,  -5,   5,  -4,  -6,  -2,  -8,  -9,   9,   1,  -4,   1,   1,   3,  10,   9,  -5,   1,  -5,   7,  -1,  -6,   7,   4,  -4,  -6,  -5,   2,  -3, -11,   9,  -2,   0, -10,   3,   0,   4,   1,-117,   9,  34,  -6,  10,  -3,   5 },
     {    6,   9,  16,  35,  44,  16,  -8,   4,   2,   3,   3,  14,  27,  10, -27,  39,  -1,  -1,   1,  -7, -30, -14,   9, -29,   0,   3,  16, -12, -48,   8, -14,  -4,  -1,   3,  15,  11, -20,  25,   3, -12,  -1,  -3,  -1,  17,   9,   9,  26,  -8,   1,  -4, -14, -17,  20,  -6,  16,   0,  -4,  -6,  -1,   1,   0, -11,  -4,  -7 },
     {   -2,  -4, -10, -21,  -5,  34,   7, -13,   1,   2,   9,  11,   0,  -3, -13,   3,  -5,  -9, -32, -43,  18,  48, -31,  36,   4,  13,  16, -12, -12,  33, -35,  15,  -1,   3,  17,  13, -24,   3, -11,   2,  -4,  -2,   2,   3, -14,  16,  -2,  -5,   0, -12,   1,  16,   1,  -3,   1,   1,   4,  -4, -13,  13, -14,   7,   3,   2 },
     {   -1,  -1,   1,   0, -12,  -9, -10,   7,   0,   1,  -2, -11,  -8,  10,  22,   4,   3,   6,  16,  18,   2,  10, -27,  -3,   1,  -3,  -4,  11,  11, -25,  20, -21,   8,  23,  62,  31, -48,  26, -22,  11,  -2,   5,  19,  14,   0,  16, -14,  15,  -1,  -9, -19,  12,  -8, -10,   5,   4,   9,   1,   9, -21,  -1, -11, -11,  -7 },
     {   -1,   0,   2,  -3,  -9,  -7,  -5, -13,   2,   3,   8,  13,  -2,  -5,   8,   9,   1,   0,   1,   3,   7,   5,  -2,  -5,   2,   5,   6,  11,  10,   4,  -2,   3,  -3,  -1,   6, -15,  -7,   2,   0,   9,   5,   1,   2,  52, -40,  16, -10,  12,  23,  31,  52, -64,   3, -13,   5,   7, -33,  -5,  -6,  -8,   0,  -9,   5,  -5 },
   }
 },
 {  //16
   {
     {  124, -28,  -8,  -5,  -3,  -2,  -1,  -1,  -8,   0,   0,   0,   0,   0,   0,   0,  -7,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    7,   3,  -1,   0,  -1,   0,   0,   0, 121, -17,  -8,  -4,  -3,  -2,  -2,  -1, -28,  -4,   2,   1,   1,   0,   1,   0, -19,   2,   1,   0,   1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    7,  -5,   0,  -1,   0,   0,   0,   0,  25,   9,  -3,   0,  -1,   0,  -1,   0, 121, -15,  -8,  -4,  -3,  -2,  -2,  -1, -18,  -9,   2,   0,   1,   0,   0,   0,  -8,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 },
     {   -5,  -5,   2,   0,   1,   0,   0,   0, -22,   5,   2,   1,   0,   0,   0,   0, -15, -15,   3,   0,   1,   0,   1,   0,-121,  13,  10,   3,   4,   1,   2,   1,   3,  16,  -1,   0,  -1,   0,   0,   0,   9,   1,  -1,   0,   0,   0,   0,   0,   6,   2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {   -2,   1,   0,   0,   0,   0,   0,   0,  -4,  -9,   2,   0,   0,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0,   2, -20,   2,  -1,   1,   0,   0,   0,-115,  17,  11,   3,   3,   1,   2,   0,  34,  20,  -6,   0,  -2,   0,  -1,   0,  22,  -9,  -2,   0,   0,   0,   0,   0,  -5,   0,   1,   0,   0,   0,   0,   0 },
     {   -4,  -7,   1,   0,   0,   0,   0,   0,  -6,  -4,   1,   0,   0,   0,   0,   0,  -4, -10,   2,   0,   1,   0,   0,   0, -11,  -9,   2,   0,   0,   0,   0,   0, -42, -17,   6,   0,   2,   1,   1,   0,-104,   3,  13,   2,   4,   1,   2,   0, -48,  20,   6,   1,   1,   0,   1,   0,  -2,   9,   0,   0,   0,   0,   0,   0 },
     {    0,   2,  -1,   0,   0,   0,   0,   0,   0,  -9,   3,   0,   1,  -1,   0,   0,   1,   8,  -1,   0,   0,   0,   0,   0,  -2, -15,   2,  -1,   1,   0,   0,   0,  -6,  19,  -1,   1,  -1,   0,   0,   0,  49, -31,  -3,  -2,  -1,   0,   0,   0,-102,   8,  16,   1,   4,   0,   2,   0,  35,  15,  -7,   0,  -2,   0,  -1,   0 },
     {   26, 123, -12,  -3, -10,  -4,  -5,  -2,  -6,  -1,  -1,  -1,  -1,   0,   0,   0,   3, -10,   0,  -1,   1,   0,   0,   0,  -5,  -1,   0,   0,   0,   0,   0,   0,  -2,  -2,   0,   0,   0,   0,   0,   0,  -7,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {    0,  -1,   0,   0,   1,  -1,   0,   0,   2,   2,   1,  -1,  -1,   0,   0,   0,   1,   9,   0,  -1,   0,   0,   0,   0,   2,  -2,   0,  -1,   0,   0,   1,   0,  -3,   6,  -1,   0,   0,   0,   0,   0, -13,   6,   3,   1,   1,   0,   0,   0,  35,  34,  -7,   0,  -2,  -1,  -1,  -1, 114,  -9, -22,   0,  -5,   0,  -2,   0 },
     {    1,   6,  10,  -3,  -1,  -1,   0,  -1,  14, 109,  -7,  -6,  -9,  -4,  -3,  -2,  -7,  54, -11,  -2,  -5,  -1,  -1,  -1,  -8,  -1, -10,  -1,   0,   1,   0,   0, -12, -15,  -4,   2,   1,   1,   0,   1,  -1, -14,   2,   1,   1,   1,   1,   0,  -5,  -3,   3,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0 },
     {   -1,  -6,   2,  -1,   1,   0,   0,   0,   3,  56, -13,  -1,  -4,  -1,  -1,  -1, -16,-106,   6,   6,   6,   5,   3,   1,  21,  -4,  12,  -4,   2,  -2,   1,  -1,  -5,  23,  -4,   1,  -1,  -1,  -1,   0,   6,  -2,  -1,   0,   0,   0,  -1,   0,  -3,   1,   0,   0,  -1,   0,   0,   0,   7,  -2,   1,   0,   0,   0,   0,   0 },
     {   -1,  -3,   1,   2,   0,  -1,   0,   0,  -3,  -9,  -6,   3,  -1,   1,  -1,   0,  -9, -21, -11,   4,  -1,   2,  -1,   1, -14, -79,  -4,   8,   3,   5,   1,   2,   4, -80,  12,   7,   4,   3,   1,   2,  16, -36,  15,   3,   3,   1,   1,   1,  17,  -9,   9,   1,   1,   0,   0,   0,   7,   0,   2,  -1,   0,   0,   0,   0 },
     {   -5,  -5, -73,  34,  -9,  11,  -3,   3,  -5,  -1, -77,  34,  -8,   9,  -2,   3,  -2,   6, -35,  21,  -6,   5,  -1,   2,  -1,   9,  -5,   9,  -2,   0,  -1,   1,  -2,   8,  10,  -2,  -1,  -1,   1,  -1,  -1,   1,  12,  -4,   1,  -2,   0,  -1,   0,  -2,   8,  -3,   2,  -1,   1,   0,   2,  -1,   3,  -1,   2,   0,   1,   0 },
     {    1,   2,   2,   1,   1,   0,   0,   0,   1,   8,  -1,   3,   3,   1,   1,  -1,  -5,  16,  -9,   2,  -3,   0,  -1,   0,  -3, -83,  13,   3,   4,   1,   3,   2,  30,  50,  15,  -7,  -2,  -3,   0,  -2, -18,  58, -20,  -3,  -5,  -1,  -1,  -1,  -8, -31,  -6,   5,   1,   2,   0,   1,   3,   0,   4,   0,   0,  -1,   0,   0 },
     {   -6,  -6, -86,   4,  -1,   6,   3,   3,   7,  14,  86,  -5,   6,  -7,  -1,  -4,  -2,  -7, -16,  -7,   7,   0,   1,   0,  -1,  -2, -14,   5,   0,   1,   0,   0,   1,   2,  16,   0,   0,  -2,   0,  -1,  -1,  -6,  -9,   1,  -1,   0,  -1,   1,   3,   3,   1,   2,   0,   0,   0,   0,  -1,  -2,   1,   0,  -2,   0,   0,   0 },
     {    0,  -1,  -3,   2,   1,   3,   2,   0,   0,   7,  -1,   2,   0,   1,  -1,  -2,   3, -13,  14,  -5,   5,  -2,   2,   0,  -5,  34, -25,  -3,  -6,  -3,  -1,  -2,   1, -57,  19,   4,   4,  -1,   1,   2,  12,  71,  -7,  -7,  -7,  -3,  -2,   1, -24, -53,   1,   6,   1,   4,   1,   1,  27,  28,  -2,  -2,  -2,  -3,  -1,   0 },
     {   -1,  -3,   0,  -1,  -2,  -2,   2,   1,  -2,  -9,  -1,  -5,  -1,   1,   1,   1,   0,  -1,  -3,  -4,   0,   1,   0,   1,  -1,  11,  -9,  -4,  -1,   0,  -1,   0, -10,  11, -20,   1,  -2,   0,   0,   0, -23, -42, -17,  12,  -1,   3,   0,   2,  -3, -87,   6,  18,   1,   4,   0,   2,  24, -62,  14,  10,   2,   2,   1,   1 },
     {    3,   3,  46,   3,  -1,  -3,  -3,  -3,   0,  -7,  17,   3,  -4,   3,  -1,   1,  -5, -12, -93,  -1,   0,   5,   8,   3,   1,  16, -37,  12,   2,   2,   1,  -2,   1,  16,  45,   8,  -1,  -4,  -4,  -2,  -1,  -5,  24,  -5,  -1,  -4,  -1,  -2,   0,  -7,  -8,  -5,  -1,   0,   2,   0,   1,   3,  -4,  -4,   1,   1,   1,   0 },
     {    6,   5,  23,  72,  23,  -7, -15,  -6,   5,   5,  29,  73,  14, -11, -11,  -6,  -1,  -3,   8,   7, -15,  -5,  -2,   0,  -2,  -3,  -6, -33, -19,   3,   5,   2,  -3,  -3, -15, -26,  -6,   6,   3,   4,   0,  -3,  -2, -14,   0,   2,   0,   1,   1,   0,   0,   1,   1,   0,  -1,   1,   0,  -3,   2,   3,   3,   0,  -2,   0 },
     {    1,   1,   1,   1,   3,   2,   2,  -1,   0,  -2,   2,   1,   0,   5,   2,   2,   0,   2,   1,   0,   1,   0,   2,   2,   2,   1,   2,   0,  -1,   3,   0,  -1,  -1,  23, -12,  -3,  -1,  -1,  -1,   0, -16, -35,  -9,  16,   0,   4,   0,   2,  29, -37,  38,   1,   5,   0,   0,  -1,  10,  97,  -3, -22,  -3,  -6,   0,  -3 },
     {    0,   0,   0,   5,   3,  -3,  -3,  -2,  -2,  -4, -11,  -1,  11,  -4,   0,  -2,  -2,  -7, -28,   3,  14,  -2,   3,  -2,  -5, -11, -56,  11,   8,  -1,   4,   0,  -4,  -7, -67,  18,   8,   1,   2,   3,  -2,   9, -61,  23,   2,   2,   0,   1,  -4,  21, -32,  17,  -2,  -2,   1,  -1,  -6,  13,  -8,   6,   0,  -1,   0,   1 },
     {    3,   0,   2,  74,  13, -19, -11,  -6,  -3,  -2,  -1, -53, -44,   4,   8,   5,   0,  -1,   0, -44,   7,  10,   6,   3,   3,   1,  11,  42,  15,   1,  -3,   5,  -1,  -4,  -6, -21,  -9,   9,   3,  -4,   2,   5,  -2,  -5,  -3,   6,   4,  -4,   0,   2,   5,   8,   9,   1,  -3,  -7,   0,   1,   0,   2,   3,   4,   2,   0 },
     {    1,   1,  -4,   6,  11,   9,  -2,  -3,   0,  -3,  14, -17,  -6, -14,  12,  -3,  -2,   2, -50,  22,  -9, -17,   6,  -1,   3,   7,  65, -22,   0,   9, -12,  -6,  -2, -15, -53,  25,  16,  -4,  17,   2,   3,  21,  21,   1,  -7,  14, -16,   7,  -5,  -9,  -6, -12,   9,   8,  -3,   1,   7,   5,  24,  -5,  -2,   0,  -6,  -1 },
     {    0,  -1,  13,   3, -47,  29,   7,   5,   1,  -2,  27,  17, -64,  42, -11,   2,   1,  -3,  13,  37, -39,  12,  -4,   1,   0,  -2,  -8,  33, -16,   8,  -4,  -2,  -1,   0, -20,  11,  -1,  -5,  -5,   2,  -1,   6,  -9,  -2,   9,  -8,  -1,   4,  -2,   0,   2, -16,  11,  -4,  -1,   1,   1,  -5,   2,  -8,   2,   2,  -1,   1 },
     {   -3,  -3, -11, -29, -11,   0,   1,   5,  -4,  -7, -11, -22, -32,  -2,  -3,   7,  -4,  -7, -20, -34, -33,   4,   7,   6,  -4,  -6, -21, -63, -29,  17,   2,  11,  -1,  -2, -11, -54,   0,  10,   6,   4,   2,   2,  -8, -26,  20,  15,  10,   2,   2,   5,  -1,  -2,  14,   1,   5,   0,   0,   6,   4,   6,   7,   1,   0,  -2 },
     {    0,   0,  -3,  -5,   8,   0,  -5, -10,   0,  -1,   3,  -7,   2,   5,  -1,  -6,   0,  -2,  12,  -8,  11,  -6,   0,  -6,  -2, -13, -54,   3,  -5,  10,  11,   8,   2,  19, -39,  12,  -7,   9,   2,  12,   5,  24,  68,   5,  -6,   0,  -3,   4,   2,   7,  52,   1, -17,  -5,  -6,  -5,   2, -12,  36,  -6,  -7,  -4,  -3,  -2 },
     {    0,  -3,   9,  30, -62,  35, -14,  -1,  -1,   1,  -1, -41,  52, -32,  -1,  -2,   0,  -2,   0,  19, -35,  29,  -6,   3,   0,   3,  -2,  -7,  11, -10,  16,  13,   1,  -1,   5,  -4,  -6,  12,  15,  13,   0,   1,  -4,  -4,   0,  -3,  -5,   8,   2,   1,  13,  12, -14,  -3,  -3,  -6,   0,   2,   1,   6,  -5,  -3,   3,  -5 },
     {   -2,  -2,   0, -29, -17,  -7, -16,  -2,   3,   4,   4,  39,  27,  12,  11,   1,  -3,  -3, -15, -39, -27,   1,   7,  -8,   5,   7,  25,  44,  28,  34,  -6,  -8,  -4,  -9, -20, -24,   4,  -2, -11,   4,   2,   5,  -2, -28, -16, -15,  17,  10,   0,   1,   8,  31, -10, -18,   3, -15,   5,   5,  15,  15,   4,  -4,   4,  -2 },
     {   -3,  -3,   2,  -6, -45, -59,  25,  18,  -1,  -3,   3,  23, -29, -68,  -7,  22,   1,   0,   1,  16,  13, -11,   4,  12,   1,   2,  -3,  10,  24,  10,  -1,  -7,   2,   1,  -1,  -4,  11,  23,  16,  -6,   0,   7,  -2,   0,   3,  14,  -1,  -2,   1,   0,  18,  -4,  -4,   4,  -2,   2,   0,  -2,  -1,  -5,  -9,   0,   4,   2 },
     {    1,   0,   8,   3, -21,  50, -33,  15,   0,   0,   2,  18, -21,   8, -32,  12,   1,   4,  -4, -28,  48,  -3,  21,   1,   2,   2,  13, -30,  44,   8,  27,   1,   2,  -5,   7,  -8,  20,  -6,   2,  -5,   1,   4, -13,  32,  -2,  -6,  -2,  -9,  -1,   6,  14,  -9,  -4,   1,   1,  -8,   1,  -7,  14,  -7,  -3,   5,  -4,   4 },
     {    0,   1,   4,  -6,  16,  12, -29,   0,  -1,   0,  -2,  -8,   1, -13,  -1,  27,   0,  -4,  -4,  14, -11, -21, -10,  -1,   0,   3,  -5,   0, -15,  28, -25,   8,   1,  -2,  10,  10,   2,  -3,  -5,  16,  -1,   7, -22,   8, -52,  15,   5,  -7,   4,  10,  52,   5,  39,  -1,  -1, -21,  -9, -19, -43,   2,   4,   8,   8,  10 },
     {    0,  -1,   7,  -1, -25,  -5,   4,   4,   2,   4,   5,  -3,  12,  28,  21,   4,   0,  -2, -11,  28,  38, -35, -33,   1,   0,   4,   9,  16, -29,   1,   6,  45,  -2,  -6,   0, -56,   1,   1,   8, -18,   4,   2,   4,  23,  -9,   3,  13,   1,  -2,   4,  -5,  15,   0,  32,  -6, -12,   2,   2,  13,  10, -13,  -3,  -5,  -3 },
   },
   {
     {   74, -54,  11,  -5,   2,  -2,   1,   0, -68,  46,  -5,   2,   0,   0,   0,   0,  27, -11,  -6,   3,  -2,   1,  -1,   0,  -8,  -2,   6,  -2,   1,   0,   0,   0,   4,   2,  -2,   0,   0,   0,   0,   0,  -3,  -1,   1,   0,   0,   0,   0,   0,   2,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   67,  14, -39,   8,  -7,   1,  -2,   0,  -5, -49,  43, -10,   5,  -1,   2,   0, -32,  48, -22,   0,   1,  -1,   0,  -1,  26, -26,   1,   7,  -4,   2,  -1,   1, -15,   8,   7,  -6,   2,  -1,   0,   0,   6,  -1,  -6,   3,   0,   0,   0,   0,  -4,   1,   3,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
     {   15, -54,  46, -16,   7,  -2,   2,  -1,  50,  18, -38,   6,  -3,  -2,   0,   0, -64,  20,   5,   8,  -3,   3,  -1,   1,  38, -14,   1,  -5,   0,   0,  -1,   0, -14,   1,   2,   1,   1,  -1,   1,   0,   3,   3,  -1,   0,  -1,   0,   0,   0,  -3,  -1,   0,   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
     {  -65,  -6,  13,   5,   2,   2,   1,   1, -51,  25,  12,  -8,   4,  -2,   2,  -1,   5,  36, -35,  12,  -6,   2,  -2,   0,  33, -50,  23,  -2,  -1,   2,  -1,   1, -22,  26,  -2,  -6,   4,  -3,   1,  -1,  12,  -8,  -5,   5,  -2,   1,  -1,   0,  -4,   3,   3,  -2,   1,   0,   0,   0,   3,  -1,  -2,   1,   0,   0,   0,   0 },
     {    6, -34,   9,   4,   2,   0,   1,   0,  58,  -9,  16, -19,   4,  -4,   1,  -1,  39,  17, -47,  19,  -7,   2,  -2,   0, -51, -22,  38,  -5,   1,   3,   0,   1,  29,  19, -18,  -5,   3,  -3,   1,  -1, -13, -12,   5,   6,  -3,   2,   0,   0,   4,   6,   0,  -4,   1,  -1,   0,   0,  -2,  -3,  -1,   2,   0,   0,   0,   0 },
     {  -21, -10,  22, -14,   7,  -3,   2,  -1, -40, -17,  10,   8,   1,   1,   1,   0, -44,  45, -12,   0,  -1,   0,   0,   0, -31,  17,  -8,   8,  -2,   2,  -1,   1,  65, -35,   8,  -4,  -2,   1,  -1,   1, -38,  17,   1,  -3,   4,  -2,   1,  -1,  15,  -1,  -5,   4,  -3,   1,  -1,   0,  -3,  -1,   2,  -2,   1,   0,   0,   0 },
     {   13,  24,  32, -55,  19,  -9,   3,  -2,  -4, -38, -31,  58, -16,   6,  -1,   1,  42,  29,   8, -27,  -2,   2,  -3,   0,  -7, -30,  -2,   8,   8,  -4,   3,   0,  -6,  17,   4,  -1,  -6,   2,  -1,   0,   4,  -6,  -5,  -1,   4,   0,   0,   1,  -1,   0,   4,   1,  -2,   0,   0,   0,  -1,   1,  -2,  -2,   1,   0,   0,   0 },
     {    9,  31,  -4,  -7,  -2,  -1,  -1,   0, -23,  -1, -15,  10,   0,   1,   1,  -1, -58, -14,  16,   9,  -4,   5,  -1,   1, -34,   0,  36, -24,  12,  -5,   3,  -1,  11,  45, -48,  16,  -3,  -2,   1,  -1,  10, -38,  22,   3,  -6,   6,  -3,   2,  -8,  17,  -1,  -8,   6,  -3,   1,   0,   5,  -4,  -3,   4,  -2,   1,   0,   0 },
     {  -16, -39,  18,  -1,   5,  -1,   2,   0, -11, -23,  22,  -6,   3,   1,   1,   0, -17,   8,   3,  -9,   6,  -4,   2,   0, -64,  29, -23,  19,  -5,   2,   0,   0, -47,  16,  19,  -4,  -1,   3,  -2,   1,  56,  -7, -12,  -1,   1,  -2,   0,   0, -21,  -1,   6,   2,  -1,   1,   1,   0,   7,   4,  -3,  -1,   0,   0,   0,   0 },
     {   -8, -15, -37,   9,  22,  -5,   2,   1,  14,  37,  16,  19, -46,  14,  -7,   0, -18,  -5,   8, -47,  44, -13,   4,   1,  -6, -23,  -2,  40, -17,   0,   4,  -2,  22,  24,   2, -22,  -2,   7,  -6,   2, -10, -17,  -4,   8,   8,  -5,   3,   0,   0,   7,   6,  -2,  -5,   2,  -1,   0,   0,  -2,  -3,   0,   3,  -1,   0,   0 },
     {   -3,  -1, -12, -36,  56, -26,   9,  -2,   1,   0,  35,  16, -47,  18,  -3,  -1,  -7, -21, -37,  25,   8,   4,  -4,   4,  12,  30,  24, -29,   4,  -7,   2,  -2, -23, -23, -11,  13,   0,   4,   1,   0,   7,  18,   2,  -1,  -2,  -2,   0,   0,   3,  -9,  -3,  -2,   1,   2,  -1,   0,  -2,   1,   2,   2,   0,  -2,   1,   0 },
     {   11,  38, -13,  -1,  -6,   1,  -2,   0,   4,  34, -25,   4,   0,  -2,   0,  -1, -25,  -4,  -9,  12,  -6,   4,  -1,   0, -54, -32,  26,  -1,   2,   3,  -1,   1, -32,  -6,  34, -18,  10,  -5,   3,  -1, -10,  48, -31,   6,  -1,  -2,   2,  -2,  20, -24,   3,   7,  -6,   4,  -2,   1,  -4,   4,   4,  -5,   3,  -1,   0,   0 },
     {  -10, -26,   5,   6,  -2,   1,   1,   1,  -8, -29,  14,  -2,   3,   1,   0,   1,  -2, -10,  18, -14,   6,  -3,   1,  -1,  -6,  -5,  -1,   3,   1,  -2,   2,   0, -55,  24, -17,  12,  -5,   4,  -1,   0, -60,  11,  25,  -4,   3,   2,   0,   1,  67,   5, -16,  -2,   0,  -2,   0,  -1, -19,  -7,   6,   1,   0,   1,   0,   0 },
     {   -5, -18,   5,   0,   2,  -2,   2,  -2,  -8, -27,  11,  -4,   5,   2,   0,   2,  -2, -23,  20,   1,  -3,   0,   0,   1,  16,   3,   6,  -9,   7,  -3,   2,  -2,  28,  37, -17,  -6,   3,  -6,   3,  -3,  -2,  15, -47,  27, -11,   6,  -2,   1,   7, -63,  46,  -5,  -2,   4,  -3,   2, -16,  32,  -5, -11,   7,  -5,   3,  -2 },
     {   -9, -20, -34,   2,   4,  13,  -2,   2,  10,  21,  19,  15,  -1, -18,   5,  -2,  -2,  15,   9, -18, -27,  26,  -8,   1,  -8, -16, -15, -16,  58, -28,   4,   1,   7,   2,  10,  38, -49,   7,   9,  -6,   7,   7,  -7, -26,  13,  14, -14,   4,   2,  -9,   2,   7,   4, -14,   6,  -1,  -3,   4,   2,  -2,  -4,   6,  -1,  -1 },
     {    0,   6,   0, -39,  12,   0,   3,  -1,   4,   8,  25,   6,  20, -23,   5,  -2, -12, -37, -13,  22, -36,  26,  -2,   2,   9,   9, -24,  17,  11,   0,  -8,   3,   2,  22,  36, -39,   0,  -9,   4,  -1, -24, -37, -17,  21,   9,   1,   2,  -1,  13,  32,   5,  -3, -10,   0,   0,   0,   4, -10,  -5,  -2,   4,   2,  -1,   1 },
     {   -6, -14,   0,   4,  -1,  -2,   3,   0,  -6, -17,   6,   3,   0,   1,  -1,   3,  -3, -18,  12,  -2,   0,   0,   1,  -1,  -9, -22,  11,  -8,   5,  -2,   1,   0, -24, -18,   7,   4,  -1,   1,   0,  -1, -67,   5,   2,   1,   1,   1,   2,   0, -86,  15,  22,  -2,   5,   2,   0,   0,  18,  18,  -5,   0,  -2,   0,   0,   0 },
     {    6,  10,  10,  26,  34, -37,  -3,  -1,  -4,  -5, -16, -44, -30,  37,   8,   3,   7,  13,  26,  39,   2, -17, -11,   1,  -8, -18, -37, -26,  11,   5,  10,   2,  10,  22,  32,  12, -11,   1,  -5,  -3, -11, -16, -17,  -5,   5,   0,   2,   1,   7,  15,   2,   1,  -1,  -1,   1,  -1,   2,  -5,  -2,   1,  -1,   0,   0,   0 },
     {    2,  10, -21,   9,  -1,   3,  -2,   0,   6,  30, -28,   8,  -7,   2,   0,  -3,  11,  61, -16,   1,  -6,   1,   0,   1,  -3,  62, -16,  -4,   3,  -7,  -5,  -4, -31,  11, -17,  -6,   4,  -4,   6,   0, -33, -22,  -3,  23,  -3,   7,  -5,   3,  -4,  -6,  32, -15,   2,  -4,   0,   0, -10,  12, -12,  -3,   5,  -2,   1,  -2 },
     {  -15, -43, -47,  32,  -4,  15,   0,   0, -12, -40, -41,  41, -14,  17,  -3,   5,   2,  -5,   7,  16, -14,  10,  -4,   4,   2,   1,  24, -19,   7,  -3,  -3,   2,   5, -18,  22, -21,  16, -10,   6,  -1,  15, -12, -10,  17,  -6,  -1,   2,   0,  13,  23,  -4,  -2,  -5,   4,  -1,   3,  -2, -15,   1,   2,   2,  -3,   0,   1 },
     {    0,  -4,  -3,   0,  -1,   0,   0,   1,   1,  -1,  -2,   1,   0,  -2,  -1,  -1,   4,   4,   3,  -3,  -1,  -1,  -3,   2,   4,   9,   4,  -1,  -2,  -2,   1,   0,   1,   7, -11,   5,  -2,   3,  -1,   1,  -5,  16, -12,   2,   2,  -1,   2,   1,  28,   0,  10,   2,  -1,   1,   0,   1, 118,  -7, -25,  -3,  -6,   2,  -3,   0 },
     {    3,   5,  11,  14, -10,  -2,  -6,   1,  -6, -16, -23, -14,   0,   4,   7,  -2,  -2, -10, -23,   6,  10,   9,  -7,   8,   8,  13,  22,  17,   3, -39,  18,  -9,  -5,  -2, -14, -17, -33,  46, -12,   3,  -1,  -4, -15,  -1,  54, -32,  -4,   7,  -4,   3,   9,  21, -35,  -1,  17,  -9,  -5,  -4,   7, -11,   9,  12, -13,   4 },
     {    2,   1,   9,   4,  16,  28, -57,  32,  -2,  -6,   0,   2, -18, -32,  67, -32,   1,  -2,   3,   2,  21,  16, -43,  14,  -1,   0,  -1,  -5, -18,  -2,  15,   2,   0,  -2,   5,   1,  15,  -5,   4,  -7,  -4,  -2,   1,   1, -11,   7,  -8,   7,  -1,  -3,   0,   1,   6,   0,   5,  -4,   0,   1,   0,   2,   0,  -2,  -3,   2 },
     {   -6, -11, -31, -12,  -3,  10,  -5,  -1,  -1,  -3, -15,  24,   1,  10,   6,  -4,  -2,  -6, -16,  40, -12, -10,   0,  -8,  -2,  -3, -23,   9, -32,  29, -19,  14,   4,  27,   1,   8,   1,  -2,  -2,   2,  -8,  -9,  32, -47,  21,  -7,   1,   1,  -6, -34,   3,  37, -10,   1,   3,  -1,   8,  29,  12, -19,   2,   5,  -5,   1 },
     {   -1,  -3,  -4,  -3,   1,   4,   0,   2,   2,   4,   6,   2,  -1,   0,   3,  -6,   6,  14,  11,  -7,  -3,  -8,   1,  -1,  10,  33,   8, -14,   1,   3,  -8,   0,   8,  55,  -9, -11,   8, -17,   7,  -1, -11,  48, -33,  -7,   3,   0,  -1,  -1, -34,  19, -50,  31, -10,   1,  -1,   3,  -4, -34,  24,  -5,  -2,   7,  -4,   3 },
     {    0,  -1, -10,  17,  29, -38,  29, -14,  -2,  -2, -10, -17,   1,  -6,   5,   2,   3,   8,  20,  -2, -31,  50, -32,   8,  -2,  -7,  -6,  39,  -1, -26,  10,  -5,   1,   0, -22, -21,  31,  -8,  12,  -1,   5,  16,  35,   0, -10,  -4,  -8,  -5, -10, -15, -15,   4,  -2,   6,   0,  -1,   6,  12,  11,  -4,   2,  -3,  -3,   1 },
     {    8,  14,  39,  40, -35,   3,  -3,  -5,  -1,  -6,   5,  23, -33,  25,  -5,   8,  -8, -21, -49,  -8,  14,  24,  -2,  -7,   3,   6, -23,   2,  27, -15,  -4,   8,   1,  15,  14,  -7,  -5, -12,  14, -10,  -8,  -8,  12,   8, -20,  10,  -4,  -5,  -4, -18, -20,   7,  18,   2,  -9,  -3,  15,   8,  22,  -9,   4,   1,   0,  -2 },
     {    7,  10,  33,  42,   0,   0,  -8,  -4,  10,  19,  53,  52,  -7, -16, -10,   3,   6,  18,  40,  22, -17,  -4,   4,   5,  -2,   1,   9, -25,  -3,  -2,   5,  -3,  -1,  -2,   1, -17,  17,  10,  -8,   9,   6,   1,  14,  -4,  18, -25,  13,  -4,   7,   8,  15,   2, -14,   9,  -4,  -5,  -1,  11,   3, -18,   2,   2,  -2,   2 },
     {    5,   7,  16,  30,  34,  -7, -17,   9,  -4,  -8,  -7,  -7, -11, -10,   9, -12,   0,  -4,   1, -10,  -8,   8,  29,  -8,   1,  -2,  12,  21,  15,  27, -65,  23,  -4, -14, -21,  -1, -20, -19,  43,  -8,   5,   1,  -4,  -1,  30,  -2,   1, -12,   0,   0,   3,   0, -16,   3, -13,   8,  -3,  -5, -21,   2,   3,  -4,  10,  -3 },
     {    1,   2,   3,  -4, -10,  14,  -9,  -3,  -1,  -3,  -5,  -8, -10,  19,   5,   9,   1,  -3,  -6,   5,   1,  15,  -6,  -1,   9,  17,  17,  18,   4,  -5, -10, -11,  13,  34,  28,   4, -10,  -9, -12,   7,  11,  53,  31, -16,   0,  -8,   4,   2,   3,  46,  21, -32,   0,  -9,  -6,   8, -11,  41, -36,   1,   1,  -9,   6,   3 },
     {    2,   4,   5,   1,  20,  34,   1, -44,  -6,  -9, -10, -24, -28, -47, -16,  49,   1,   1,  -3, -14,  17,  25,  12,   1,  -2,  -1,  -4, -34, -21,  -2,  -6,  -8,   0,   4,   7,  -4,  14, -13,  -5,  19,  -3,  -9,  -5,  -6,  21,   9, -15, -17,   4,   0,  -3,  15,  -2, -19,  13, -11,   2,  16,   2,   6, -10,  -9,   3,   2 },
     {    1,   0,  -6,  19,  22,   8, -16,   5,  -2,  -4, -14,  20,   0, -18,  -4,  24,  -3,  -7, -19,  18, -22,  10,  16,  15,  -3,  -8, -32,  23, -14,  26,  12,  -4,   2,  10, -25,  28, -20,  28, -19,  17,   1,  28, -14,  27, -13, -19,   7, -13,  -7,  13,   0, -23,  21,  12, -14, -28,  -1, -11,  23,  13,  -1,   1,   9, -11 },
   },
   {
     {  102, -24, -13,  -3,  -4,  -1,  -2,   0,   0,  -7,   2,   0,   0,   0,   0,   0, -67,  12,  12,   1,   3,   1,   1,   0,   1,   8,  -2,  -1,   0,   0,   0,   0,  16,  -2,  -4,   0,  -1,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -6,  82, -49,  11, -12,   2,  -3,   0, -11,   3,  -6,   5,  -2,   2,  -1,   0, -29, -57,  41,  -6,   9,   0,   2,   1,  10,   1,   6,  -6,   1,  -2,   0,   0,  13,  16, -12,   0,  -1,  -1,   0,  -1,  -2,  -3,  -1,   1,   0,   0,   0,   0,  -3,  -3,   3,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {   -2,  -4,   3,  -1,   1,   0,   0,   0, -99,  26,  10,   3,   3,   1,   1,   0,  -1,  15,  -5,   1,  -2,   0,   0,   0,  68, -13, -12,   0,  -3,   0,  -1,   0,   5, -14,   3,   0,   0,   0,   0,   0, -18,   2,   6,   0,   1,   0,   0,   0,  -1,   4,  -1,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0 },
     {   67,   1, -14,  -3,  -4,  -1,  -1,   0, -14,  10,  -4,   1,  -1,   0,   0,   0,  82, -39,   0,   0,  -1,   0,  -1,   0,  -5, -14,   8,  -1,   1,   0,   0,   0, -47,  17,   5,  -2,   2,  -1,   1,   0,   1,   7,  -3,  -1,   0,   0,   0,   0,   9,  -2,  -2,   1,  -1,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0 },
     {    9,  31,  58, -55,  10, -10,   2,  -3, -11,  -7,   2,  -5,   5,  -2,   2,   0, -19, -42, -47,  50,  -5,   7,   0,   2,  -2,  14,   4,   5,  -6,   1,  -2,   0,   9,  25,  12, -18,  -1,  -1,   0,  -1,   2,  -7,  -5,   0,   1,   0,   0,   0,  -2,  -7,  -2,   4,   0,   0,   0,   0,  -1,   1,   1,   0,   0,   0,   0,   0 },
     {   10,  15,   3,  -6,   0,  -1,  -1,  -1,  66, -25,  -4,  -3,   0,   0,   0,   0,   9,  -3,  -7,   4,   0,   1,   0,   0,  86,  -6, -14,   0,  -5,  -1,  -2,   0, -11, -20,   6,   1,   0,   1,   1,   0, -48,   4,  11,   0,   3,   0,   1,   0,   3,   8,  -2,  -1,   0,  -1,   0,   0,   6,   1,  -3,   0,  -1,   0,   0,   0 },
     {   -4, -15,   7,   3,   1,   0,   0,   0, -18, -57,  40,  -8,   9,  -1,   3,   0,  16, -12,  19, -14,   7,  -3,   2,  -1,   6,  72, -42,   3,  -5,  -2,   0,  -1,  -9,  27, -28,  14,  -5,   1,  -1,   0,  -2, -26,  10,   5,  -2,   3,  -1,   1,   3,  -9,   9,  -3,   0,   1,   0,   0,   1,   4,  -2,  -2,   0,  -1,   0,   0 },
     {   12, -18,   4,  -2,   4,  -1,   1,   0,  11,  16,  -5,  -2,   0,  -1,  -1,   0,  45, -19,   4,  -3,  -2,  -1,   0,   0,   6,  -8,   3,   1,   0,   0,   1,   0,  98,   4, -26,   2,  -5,   0,  -3,   0,  -8, -19,   1,   3,   0,   1,   0,   0, -44,  -2,  14,  -1,   3,   0,   1,   0,   3,   6,  -1,  -2,   0,   0,   0,   0 },
     {  -13, -19, -40, -57,  38,  -2,  10,   1,   6,   7,  12,   0,   4,  -5,   0,  -2,   3,  10,  51,  54, -37,   1,  -9,  -2,  -2, -13, -17,  -4,  -4,   6,   0,   2, -10,  -1, -26, -18,  14,   1,   3,   1,   0,   8,   7,   5,   1,  -2,   0,  -1,   2,   0,   6,   5,  -4,   0,  -1,   0,  -1,  -2,  -1,  -2,   0,   0,   0,   0 },
     {  -18, -52,  26,  10,   0,   1,   2,   0,   7,   4,   2,  -4,   0,  -1,  -1,   0, -25, -34,  22, -14,  10,   1,   1,   1,  27, -34,  22,  -6,   4,  -3,   2,   0, -11,  66, -23,   1,  -4,  -4,   1,  -2,  -8,  31, -25,   6,  -2,   1,  -1,  -1,   6, -24,   3,   5,  -1,   2,  -1,   1,   0,  -6,   7,   0,  -1,   0,   0,   0 },
     {    1,   8,   0,  -1,   0,  -1,   0,  -1,   6,  18,  43, -42,   9,  -8,   2,  -2,  -9, -12,  -4, -11,  11,  -4,   3,   0, -23, -40, -50,  62,  -8,   8,  -1,   2,   1,   6,   9,  16, -14,   4,  -3,   1, -13,  26,  25, -28,   1,  -2,  -1,   0, -10,   4,  -5,  -7,   6,  -1,   1,   0,   6,  -4,  -7,   5,   0,  -1,   0,   0 },
     {    8,  17,  -7,  -3,   1,   0,   0,   0,  14,  -6,  11, -10,   3,  -2,   1,  -1,   9,  21,  -7,  -2,  -2,  -1,  -1,   0,  44, -13, -17,  10,  -2,   1,   0,   0,  11,   9,   4,   4,  -3,   0,  -1,   0,  97,   3, -18,  -4,  -6,  -1,  -3,   0,   4, -31,  -3,   1,   0,   1,   1,   1, -35,  -4,  11,   0,   2,   1,   1,   0 },
     {    5,  10,   6,  25,  54, -59,  19, -10,  -2,  -3,  -4,  -5,   1,  -6,   5,  -1,  -4,  -6,  -5, -32, -49,  55, -16,   8,   0,   3,   6,   9,   0,   7,  -8,   2,   2,   1,   1,  20,  19, -23,   5,  -1,  -2,  -2,  -3,  -4,  -3,  -1,   4,  -1,   3,   0,  -3,  -6,  -4,   6,  -1,   0,   0,   2,   0,   1,   2,  -1,   0,   0 },
     {    3,   7,   5,   1,  -1,   0,   1,  -3,  21,  88, -26,  -7,  -7,  -2,  -2,  -2,  -7,  15, -12,  -1,   1,  -2,  -1,   1,   8,  33, -30,   3,   3,  -4,   1,  -1, -19,  27, -28,  11,  -3,   0,   1,   1,  -6, -35,  14,   8,  -4,   6,  -2,   2,  20, -21,  16,  -5,  -2,   2,  -2,   0,   4,   6,  -1,  -5,   1,  -1,   0,   0 },
     {   -2, -18,   6,  -2,   0,  -1,   1,   1,   4,  18,   1,   0,  -4,   0,  -3,   0,   6, -32,  13,   0,   5,   0,   1,  -2,   1,  26,  -7,  -4,  -1,   1,   0,   0,  40, -26,   6,   2,  -2,  -1,  -2,   2,   5,  24,   6,  -9,   1,  -1,   1,  -1,  87,  15, -36,   0,  -3,   0,  -2,  -2, -14, -35,   0,   9,   1,   2,   0,   1 },
     {    3,  10,  23, -23,   5,   0,  -4,  -2,  -4,   3, -19, -37,  32,  -8,   7,  -2,   1,   3,  23, -10,  10, -10,   7,  -1,   5,  11,  48,  33, -37,   2,  -5,   1, -16, -35, -43,  29, -11,  11,  -1,   2,   7,  -1, -30,  -1,   9,   2,   0,   0,   3,  22,  17, -10,   0,  -2,  -1,  -1,  -3,  -2,   4,  -3,   0,   1,   0,   0 },
     {   -2,   2, -43,   9,   5,  -1,  -1,  -2,  -3,  -7, -17, -14,  20,  -3,   5,  -2,  12,  36, -13,  10,  -4,   0,   4,   1,   4,  30,  25,  33, -32,   3,  -6,   0,  19,  43,  24, -26, -11,   2,  -3,  -1, -24,   8, -21, -23,  17,   1,   1,   1,  12, -29, -16,  17,   7,  -2,   0,   1,  22, -15,   7,   7,  -8,  -1,   0,   1 },
     {  -15, -47, -21,  17,  12,   1,   0,   0,   3,  10, -12,  -9,   3,  -5,   3,  -1, -15, -49,   4,  12,   4,   2,   5,  -2,  16,   4,  11,  23, -16,   5,  -5,  -1, -15, -13,  39, -17,  -9,   3,  -4,   2,  22, -57,  13, -14,  10,   2,   1,   3, -10,  18, -10,   4,   7,  -3,   1,  -2, -16,  35,  -5,   0,  -2,  -3,   0,  -1 },
     {   -6,  -8, -11, -17, -47, -50,  38,   5,   1,   2,   2,   7,   5,  -4,   4,  -3,   6,  11,   9,  16,  59,  50, -34,  -8,  -1,  -2,  -3,  -5,  -9,   1,  -8,   5,  -2,  -3,  -2,  -4, -28, -21,  13,   4,   3,   0,   1,   4,   5,   2,   5,  -2,  -2,   3,   2,  -2,   8,   5,  -5,   0,  -1,   3,   1,  -1,  -4,  -4,  -1,   2 },
     {    4,  12,   2, -11,   0,  -4,  -5,   8,  12,  25,  76, -24,  -3,  -6,  -9,  -1,   2,  10,   6, -20,   3,   3,   5,  -4,  -2,  -9,  15, -23,   0,   2,   7,   1,   3,  -4,  15, -13,   9,  -2,  -1,  -1, -11, -37, -47,  26,  13,   0,   1,   1,  16,   6, -19,  24,  -6,  -1,   0,  -2,  17,  20,  10,  -5,  -9,   0,   1,   0 },
     {    0,  -1,  -2,  -4,   1,  -4,   3,   5,  -1, -18,  -6,   3,   5,   1,   0,   1,   0,   3,  -5,   3,   1,  -4,   3,  -4,   8, -26,  10,   2,   6,  -4,   3,   3,   8,  13,  -6,  12,  -2,   4,   0,   4,  35, -21,  19,   1,  -9,   6,   0,   0,  36,  12,   6, -12,  -4,   1,   0,  -3,  94,   9, -41,   0,   0,  -1,  -1,   0 },
     {   -5,  -8, -26, -47,  29,   4,   2,  -2,   1,   6,  -3,  13,   2,   3,  -3,  -2,  -1,  -1,  -3, -15,  10,  -4,   8,   5,   2,  12,   7, -13,  16,  -8,   7,  -9,   0,  10,  41,  63, -44,  -1, -12,  -4,  -7,   2, -13,  10, -31,   8,  -6,   7, -16,   0, -22, -24,  17,   5,   1,   3,   4,  -4,   9,  -1,   8,  -2,   0,  -1 },
     {    0,   5, -14,  -4,   5,   8,   1,  -3,  -9, -31,  -4,   4,   9,   3,   3,  -5,   0,  11, -20,   1,   3,  -3,  -1,   3, -14, -47,   7,   8,  -4,   6,  -1,   6,  14,   9, -12,  13, -12,  -1,  -3,  -1, -30, -38,  15,  17,  -5,   6,   3,  -1,  59, -22,  22, -13,   0,   0,   0,  -1, -46,  26,  16, -18,   4,  -7,   0,   4 },
     {    1,  -1,   2,   9,   0, -14,  12,  21,   2,   6,  -6,  23,  43, -23,  -3,  -6,  -5, -11, -10,  -8, -10, -11,  -2, -15,  -7,  -8, -23, -42, -57,  39,  12,   7,   0,   1,   7,  19,  -1,  40,  -8,   1,   3,  10,   9,  22,  23, -12, -10,   0,  -6, -13, -10,   1,  -3, -19,   6,   5,   4,   2,  -3,   0,  -4,   7,   4,   0 },
     {   -2,  -2,  -2,   0,   1, -27, -61,  40,   0,   0,  -2,   4,   3,  -3,   1,   2,   2,   4,   2,  12,   8,  34,  63, -47,  -2,  -4,  -3,   2,  -2,  -3,   2, -12,  -2,  -2,  -5,  -2,   1, -22, -29,  18,  -2,   2,   6,   7,  -3,   1,  -2,  14,  -1,  -8,   5,  -5,  -4,   9,  10,  -2,  -7,  -8,  -6,   0,   1,  -2,   0,  -4 },
     {    5,  16,  43, -31,   6,  -7,  -8, -16,  -3,  -8, -32,  29, -12,   0,   0,   3,   8,  23,  56, -35,   9,  -3,  11,   8,  -4, -14, -16,   9,  -4,   7,   1,  -2,   1,  14,  11, -14,  -8,  -1,  -7,   4,  -2, -18,  13, -23,  18, -10,  -2,   4,   3,  -5, -37,  20,   3,  -2,   7,   0,  -2,  25,   1,   1, -14,  -3,   4,   0 },
     {   -1,  -7,   5,   7,   9,  -2,  -2, -10,  -4,  -7, -10,   4,  -2,  10,   7,  -3,  -3, -20,   6,  10,  11,   7,   0,   8,  -8, -15, -17,  11,   9,  -6,  -8,  -1,  -9, -48,  -3,  10,  -5,  12,  -1,   0,  -4, -27, -11,  11,  -2,   9,   9,   1,  -6, -67,   0,  31,  -2,  -2,   4,   4,  35, -40,  38,  20,  -8,   0,   0,   2 },
     {   -3,  -7,  -8,  -9, -24,  -6,  -2, -16,   6,   8,  34,  66, -28,   2,  -8,   7,  -3,  -7, -11,   5, -14,   2,   6,   3,   4,   4,  11,  24, -14,  22,   8, -18,  -2,  -6, -13,  18,  15,   7,  -8,   1,   0,  -4, -23, -45,  21,  -9, -10,   8,   7,  -5,  24, -31,   9,   2,   6,   3,   8,   5,  16,  11,  -6,  -5,   6,   4 },
     {    3,   6,  -1,  19,  35, -36,   2,   3,  -1,  -2,   4,  15, -25,   7,  -5,  -6,   1,   4,  -7,   8,  14,  -3,   6,   4,  -1,  -3, -10,   9,   3, -43,  21, -12,  -1,   0, -16, -21, -37,  52,  -5,  -7,   2,  12, -14,   8,  14,  35, -24,   0,   7,  24,   9,  12,  11, -32,  -3,  10,  -5,   7,  -1,  -2,   1,  -6,   9,  -2 },
     {    4,   9,  22,   7,   0,   3,   8,  -4,   5,   9,  35,  -4,  -2,   7,   2,   1,   2,  11,  25,   1, -12,   7,   1,   5,   2,  -4,  43, -20, -25,  -9,  17,  -8,  -6,   3,  15, -10, -19,   3,  -2, -10,   0, -22,  51, -11,  -2,  14,  -1,   6,  -7, -22,  12, -13,  24,  -3,  -3,   5, -15, -48, -27,  16,  20, -12,  -1,   4 },
     {    3,   3,  12,  12,  18,   7,  -4,   9,   2,  10,  14,   3, -10,  16,   9,   2,   5,  -4,  16,  10,   8,   0,  -3,   8,  -2,  20,  10,  -1, -15,  -6, -15, -10,   6, -27,  15,  -1, -13,  -4,   3,  -3,   1,  47,   6,   8,   0,  -9,  18,   1,  10, -53,  15,  -8,  11,   2,  -4,  -2,  12,  70, -18, -14,  11,   1, -18,   0 },
     {   -1,  -5,  -1,  -9, -12,   7,  -5,   9,  -1,  -2, -12,   3,  26,  28, -33,  14,  -3,  -9,   0,  -8, -17,  -2, -11,  -8,  -5,  -5, -19,   1, -31, -54,  51,  -2,   3,  -7,  10,  11,  16, -34,  33,  -2,   0,   2,  -7,  -8,  19,  34,  -9,   3,   0,  -8,   1,   3,  -3,  21, -11,  -6,   0,  10,   1,   1, -10, -12,   0,  -3 },
   }
 },
 {  //17
   {
     { -118,  14,   5,   4,   3,   2,   2,   1,  47,  -2,  -2,  -1,  -1,  -1,  -1,   0,  10,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   29,  -3,  -1,  -1,  -1,   0,   0,   0,  91, -10,  -4,  -3,  -2,  -2,  -1,  -1, -84,   3,   4,   2,   2,   1,   1,   1,   2,   3,   0,   0,   0,   0,   0,   0,   9,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -33,   8,   2,   1,   1,   1,   0,   0, -67,   3,   4,   2,   2,   1,   1,   0, -81,   9,   4,   2,   2,   1,   1,   0,  61,   1,  -3,  -1,  -2,  -1,  -1,   0,  15,  -3,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   12,   0,  -1,  -1,  -1,   0,   0,   0,  25,  -3,  -1,  -1,   0,   0,   0,   0,  24,   3,  -1,   0,  -1,   0,   0,   0,  87, -12,  -5,  -2,  -2,  -1,  -1,   0, -83,  -1,   5,   1,   2,   1,   1,   0,  -6,   7,   0,   0,   0,   0,   0,   0,  17,  -1,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 },
     {  -12,   5,   1,   1,   0,   0,   0,   0, -25,  -1,   2,   1,   1,   0,   0,   0, -44,   1,   2,   1,   1,   0,   0,   0, -69,   0,   4,   1,   2,   1,   1,   0, -88,  10,   5,   2,   2,   1,   1,   0,   7,   9,  -1,   1,   0,   0,   0,   0,  28,   0,  -2,   0,  -1,   0,   0,   0,   6,  -1,   0,   0,   0,   0,   0,   0 },
     {    1,   4,   0,   0,   0,   0,   0,   0,   3,  -3,   0,  -1,   0,   0,   0,   0,   1,   4,  -1,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0, -20,  11,   0,   1,   0,   0,   0,   0,  96, -14,  -6,  -2,  -2,  -1,  -1,   0, -77,  -1,   6,   1,   2,   1,   1,   0,  18,   5,  -2,   0,  -1,   0,   0,   0 },
     {   -3,  -1,   0,   1,   1,   0,   0,   0,  -5,  -3,   1,   0,   1,   0,   0,   0,  -4,  -4,   1,   0,   0,   0,   0,   0, -10,  -6,   1,   0,   0,   0,   0,   0, -27,  -7,   2,   0,   1,   0,   0,   0, -76,  -3,   6,   1,   2,   1,   1,   0, -94,  10,   7,   2,   2,   1,   1,   0, -24,   9,   2,   1,   1,   0,   0,   0 },
     {  -15,-125,   1,  -3,   7,   3,   4,   1,  -1,  13,   5,   2,   0,  -1,   0,   0,  -3,  15,   3,   1,   0,   0,   0,   0,   1,  -1,  -1,  -1,   0,   0,   0,   0,  -3,   7,   0,   1,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {    0,  -1,   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,  -1,  -4,   0,   0,   0,   0,   0,   0,  -3,  -2,   1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  29,  -1,  -3,  -1,  -1,   0,   0,   0,   8, -16,   0,  -1,   0,   0,   0,   0,-123,   6,  10,   1,   3,   1,   2,   0 },
     {   -2,  -6,   0,  -1,   0,  -1,   1,   0,  -7,-112,   5,   2,   8,   4,   4,   2,   8,  53,   5,  -2,  -2,  -2,  -2,  -1,  -3,  13,  -2,   1,  -1,  -1,   0,   0,   2, -15,   0,  -1,   2,   0,   0,   0,   1,  14,   2,   0,  -1,  -1,  -1,   0,   1,  -1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    2,  13,  -1,  -1,   0,  -1,   0,  -1,   5,  48,   1,  -3,  -3,  -1,  -1,   0,   2,  92,  -4,  -2,  -6,  -3,  -3,  -1, -11, -34, -10,   2,   0,   1,   0,   1,   0, -60,   2,   1,   4,   2,   1,   1,   5,   6,   6,   1,   0,   0,   0,   0,   0,  10,  -2,  -1,  -1,   0,   0,   0,  -4,   0,  -2,   0,   0,   0,   0,   0 },
     {   -4,   2, -95,  10,   2,   7,   1,   3,   3,   1,  76,  -6,  -4,  -7,  -1,  -3,   0,  -2,  -7,  -1,   2,   2,   1,   0,  -1,   4, -21,   2,  -1,   0,   1,   0,   1,   3,  23,  -1,  -2,  -1,  -1,  -1,  -1,  -2, -13,   1,   0,   2,   1,   0,   1,   0,   5,   0,  -1,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0 },
     {   -1,  -8,   0,   2,   0,   0,   0,   0,   0,  11,  -4,  -1,   0,   1,  -1,   1,   1, -32,  10,   1,   1,   1,   1,   0,   1,  75, -14,  -3,  -4,  -3,  -2,  -1,  -7, -81,   8,   4,   4,   3,   1,   2,  12,  45,   0,  -4,  -1,  -2,  -2,   0,  -9, -15,  -3,   2,   0,   1,   0,   1,   3,   2,   3,  -2,   0,   0,   0,   0 },
     {    2,  12,   4,  -5,  -4,  -1,  -2,  -1,   4,  28,   2,   1,   0,  -2,   0,  -2,   6,  57,  -3,  -3,  -3,  -2,   1,   0,   5,  88,   0,  -4,  -5,  -2,  -2,  -1,  -4,  60, -11,  -4,  -4,  -2,  -2,  -1, -11,   0,  -7,   0,   0,   1,   0,   0,  -6,  -7,  -3,   0,   1,   0,   1,   0,  -6, -10,   0,  -1,   0,   0,   0,   0 },
     {   -1,   1, -48,  15,  -3,   7,   0,   1,  -1,   3, -31,   2,  -1,   3,  -1,   1,   3,   3,  90, -10,   1,  -9,  -1,  -3,   1,  -5,  20,  -8,   4,   0,   1,   0,  -3,  -1, -61,   1,  -6,   5,   1,   1,   0,   3,   0,   5,  -1,   0,   0,   0,   0,   1,  12,   0,   1,  -1,  -1,  -1,   0,  -7,   3,  -3,   0,  -1,   1,   0 },
     {   -1,  -2,   1,  -3,  -1,  -1,  -1,   2,   0,  -6,   5,   2,   3,   1,   0,   1,   0,  -2,  -2,   4,  -3,   2,   0,   0,   0,  33,   0,  -4,  -4,   2,  -1,  -4, -11, -33, -16,  -3,   0,   2,   1,   2,  -1,-101,   9,   6,   6,   2,   2,   3,  17,  24,  18,  -1,   2,  -3,  -1,  -1,   0,  46,  -4,  -2,  -1,  -2,  -1,  -1 },
     {   -6,  -1, -14, -71, -56, -39, -20, -10,   3,   4,   3,  43,  38,  24,  11,   7,   0,  -2,  10,  17,  12,   8,   4,   2,  -3,  -7, -13, -21, -16,  -8,  -7,  -1,   1,   2,   2,   7,   6,   4,   3,   2,   0,   3,   3,  -3,  -3,  -2,  -1,  -1,  -1,   0,  -1,   3,   3,   1,   0,   1,   0,   1,   2,   0,   0,   0,   0,   0 },
     {   -3,  -4, -60,  22,   7,  -1,   1,   2,  -4,  -3, -87,  26,  12,  -4,   1,   4,  -3,   3, -55,  12,  -2,   2,   5,  -1,  -1,   5, -15,   2,  -6,   3,   1,  -1,   0,   3,  10,   5,  -1,   1,   0,   0,   0,  -4,  11,   1,  -3,   1,   1,   1,   0,   1,   5,  -1,   0,   0,   0,  -1,   0,  -5,   4,  -2,   1,   0,   0,  -1 },
     {   -2,  -2, -10, -10,  -6,  -5,  -2,   7,   3,   3,  21,  29,  29,  12,   2,   0,  -5,  -6, -42, -47, -35, -16,  -7,   2,   5,   4,  60,  37,  25,   8,   3,  -2,  -3, -12, -40, -15,  -7,  -6,  -2,  -3,   3,  13,  17,   7,  -2,   3,   1,  -3,  -1,   1,   0,  -2,  -2,   0,   0,   1,   0,  -2,  -3,   1,  -2,   0,   0,   1 },
     {    0,   0,   3,  -2,  -1,   0,   0,   3,   0,  -4,   5,  -6,   1,  -5,  -2,   0,   0, -10,   0,   6,   2,   4,   4,  -1,   1,  11,  -4,  -3,  -1,   1,  -3,   0,  -2,  -7,  -3,   2,  -4,   3,   2,   1,   9,  -8,  23,   0,   6,   1,   3,  -3,  -1,  91, -20,  -3,  -1,  -4,  -3,  -2, -14, -78,   1,   5,   2,   4,   3,   2 },
     {    4,   1,  24,  67,  18, -11, -11,  -8,   4,   3,  26,  68,  30,  -2,  -9,  -2,  -1,  -2,   3, -12,  -7,  -5,   6,   0,  -3,  -6, -28, -43, -26,  -1,   3,   5,   0,   3, -15, -15,  -4,   4,   5,  -1,   1,   5,  -1,  -4,  -5,   3,   3,  -1,   0,   6,   0,   1,   1,   2,  -2,  -1,   0,   3,   5,   0,   2,   1,   0,  -1 },
     {    1,   5,  -2, -48,  66, -10,   4,   4,   0,  -3,  -1,  55, -61,  15,  -9,   0,   0,   0,   8, -26,  17, -18,   1,   3,   0,   2, -14,  12,  -3,   4,   4,  -2,  -1,   1,   6,   8, -17,   4,  -5,   0,   0,  -2,   3,  -9,   5,  -5,   4,   0,   0,   3,   3,   4,  -2,   4,   0,  -2,   0,  -2,   1,  -7,   5,   0,  -1,  -1 },
     {    0,   1,  -7,   3,  19,  -1, -11,  -5,   0,  -1,  18,  -6, -29,   7,  15,   5,   1,   2, -36,  40,  44,  14,  -5,   3,   1,  -1,  50, -35, -35,  -7,   1,   7,  -1,  -6, -47,  21,  27,  -1,  -1,  -3,   1,  13,  16,   1, -12,   0,   0,  -2,  -1,  -4,   2,  -1,  -2,   1,  -1,   1,   0,   3,  -2,   4,   3,   0,  -1,  -1 },
     {   -1,  -3,   5,   2, -17,  14,   7, -17,  -1,  -4,   4,  13, -15, -11,  24,  -9,  -2,   0,  -6, -23,   1,  13, -18,   7,  -2,  -3,  -4, -16,   9, -13,  -3,   7,  -5, -17,  -2,   9,  -3,   1,  -3,   1,  -6, -44,   7,   5,   4,  12,  -6,  -2,  -4, -65,  -1,  12,   1,   7,   0,   8,   5, -71,  18,   5,   4,   0,  -2,   0 },
     {    0,  -1,   8,  -8, -11,  10,   9,  -8,  -1,  -3,   8, -13,  -5,   8, -14,   4,  -1,  -7, -25,   0,   9,   6,   4,  -3,  -3,  -5, -80,  30,  14,  -1,  -4,   1,  -3,   6, -75,  14,  11, -11,   1,   1,   0,  11,   2,  15,  -1,   0,   3,  -5,   0,   0,   9,   7,  -7,  -5,  -1,   3,   1,   4,  15,   2,  -2,   1,  -1,  -1 },
     {    0,  -3,   5,  29, -59,  49, -26,  13,   0,  -1,  -1,  33, -52,  54,  -7,   1,   0,   1,  -2,   8,  -3,  -6,  -8,   6,   1,   2,   5,  12,   6,   1,   0,   2,   1,   0,  12,   1,  14,   0,  -7,  -5,   0,   1, -15,  -5,   8, -18,   8,  -3,   0,  10,   6, -12,   2,  -9,   5,  -3,  -2,   0,  -5,  -5,   4,   2,   1,  -2 },
     {    1,   4,   3, -29, -13,  63,  45,  10,   0,  -1,   3,  38,  11, -56, -32, -10,   1,   1,   2,  16,  -1,   3,   6,   0,   1,   0,  13, -20,   4,   9,   4,   7,   1,  -1,   2,  14,  10, -23,  -8,  -5,   1,  10,  -2,  -3,   3,  21,   1,  -4,   1,   7,  19,   0,  -7, -12,  -3,  -1,   0,   8,   1,   3,  -7,   3,   9,   0 },
     {    0,  -1,   5,  18,   0, -29, -17,  -6,   1,   0,   8,  31,   0, -40, -13,  10,   2,   2,  10,  64,  14, -28,  -6,   7,   1,   1,  11,  60,  20, -30, -12,  -5,  -2,  -6,  -4,   4,  -2, -11,  -4,  10,  -2, -11, -19,   1,  -5,   7,  12,  -2,  -2, -11, -13,  -3,   6,   4,   0,  -1,   1, -13,  -4,  -2,   1,   4,  -1,   2 },
     {    0,   0,   4,  -5,   8,  10, -16,  20,   1,   0,  -6,   8,   8, -10,  52, -22,   0,   1, -13,  -2,  36,  15, -14,   3,   0,  -1,   2,  -9,  27,  -9, -11,   4,  -1,  -7, -10, -23, -33,  -1,  22,   7,  -1,  10, -64,  12,  30,  -3, -17,   1,   1,  21,  19,  -2,  -3,  -6,  -8,  -1,   0,   5,  15,   5,  -1,   7,   5,   7 },
     {    1,   0,   4,   1, -12,  30,  11, -11,   1,   2,  -1,  -8,  29,   8,   2, -18,   0,   0,   0, -18,  50, -29, -17,   7,   0,   0,  -1,  43, -56, -22,  19,  32,   0,   1,   7,  -3, -29,  28,   2,   8,   0,  -1,  12,  -9, -12,  12,   6,   5,   1,   2,  11,   5,  -7,   2,  -5,  -9,  -1,   1, -11,   5,   4,  -7,   4,  -6 },
     {   -1,  -2,   6,   6, -13, -33,   3,  -5,  -1,  -3,   1,  -2, -15, -14, -19,  -1,   0,   0,  -7, -15,  -8,  16,  33,  -7,   0,   0,   9,  12, -19,  26, -43,  33,   0,  -4,  -3,  26,  -7,  38,   1,  -6,  -2,   2, -29,   9,   1,  16, -10,  -3,   1,  -2,  45, -23,  14, -21,   3,  -2,  -2, -15, -25, -19,   5,   1,  19,  -5 },
     {    1,   1,   6,   5,   0,   2,   3,  30,   1,   3,   3,   2,  14,  -4,  15,  44,  -1,  -1,   0, -18,  11, -39, -32, -28,   0,   2,  -5, -11,  11,  -6,  -2, -15,   2,   4,  13,  55,  12,   7,  14, -11,  -2,  -4,  -9,  32, -41, -10, -23,   2,   0,   6,  10, -17, -26,   3,   7,  -3,  -1,  -3,   0,  -7,   6,  10,   5,  -2 },
   },
   {
     {   98, -50,   8,  -9,   1,  -2,   0,  -1, -56,  29,  -5,   5,   0,   1,   0,   0,  -1,   1,   0,   1,   0,   0,   0,   0,   7,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -42, -63,  63, -21,  15,  -3,   4,   0,  24,  50, -44,  14, -10,   1,  -2,   0,  13,  -1,  -5,   2,  -2,   1,  -1,   0, -14,  -8,   9,  -3,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -29,  19,  -8,   4,  -2,   1,   0,   0, -68,  17,   6,   2,   2,   1,   1,   0,  88, -27,  -1,  -4,  -2,  -1,  -1,   0, -39,   7,   2,   1,   2,   0,   1,   0,   5,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   50, -11,  -7,   1,  -1,  -1,   0,   0,  74, -22,   2,  -6,  -1,  -1,  -1,   0,  41, -18,   1,  -2,   0,   0,   0,   0, -68,  19,   1,   3,   2,   0,   1,   0,  24,  -1,  -3,   0,  -1,   0,  -1,   0,  -6,  -2,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 },
     {    3,  21,  -9,   0,  -1,  -1,  -1,   0,   3,  32, -34,  16,  -9,   3,  -2,   0, -28, -74,  62, -22,  13,  -1,   2,   1,  11,  39, -25,   6,  -3,  -2,   0,  -1,  11,  -2,  -6,   4,  -3,   1,  -1,   1,  -9,  -3,   5,  -2,   2,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {  -17,   0,   2,   0,   1,   0,   0,   0, -31,   3,   7,  -1,   1,   0,   0,   0, -49,  21,  -8,   6,   0,   0,   1,   0, -43,   8,   7,   0,   2,   1,   0,   0,  92, -14,  -8,  -1,  -4,   0,  -2,   0, -37,  -2,   4,   0,   2,   0,   1,   0,   2,   5,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   0 },
     {   13,  39,  21, -59,  52, -42,  21, -10,  -9, -34,  -9,  38, -34,  27, -12,   5,   4,   2,  -7,  11, -11,   8,  -5,   3,   3,   9,   3, -12,  11,  -9,   4,  -2,   1,  -4,   0,   1,  -1,   1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   1,  -1,   1,  -1,   0,   1,   1,  -1,   0,   0,   0,   0,   0 },
     {    3,  -7,   1,   0,   0,   1,   0,   0,  22,  -3,   4,  -3,   1,  -1,   0,  -1,  54,  -5,  -5,  -2,  -2,  -1,   0,   0,  82,  -9,  -8,   0,  -3,   0,  -1,   0,  52, -17,  -3,   0,  -2,   1,  -1,   0, -52,  -1,   6,   0,   3,   0,   1,   0, -12,   6,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
     {    9,  21, -18,   5,  -3,   0,   0,   0,   9,  18, -16,   8,  -8,   3,  -2,   0,  -2, -12,   6,  -2,   1,   1,  -1,   1,  -1, -62,  50, -22,  13,  -3,   2,   0,  11,  64, -47,  18,  -9,   0,   0,  -2,  -7, -16,   7,  -1,   0,   2,  -1,   1,  -7,  -5,   8,  -3,   2,  -1,   1,   0,   2,   1,  -2,   1,   0,   0,   0,   0 },
     {   -1,  -2,   2,  -2,   1,   0,   0,   1,   0,   0,   0,   0,   1,  -2,   1,  -1,   4,   0,   3,  -1,  -1,   1,  -1,   0,   6,   4,  -5,   2,  -2,   1,   0,   0,  36,  -2,   1,  -2,   1,  -1,   0,   0,  73, -15, -11,   1,  -4,   0,  -1,   0, -93,   0,  13,  -2,   4,   0,   2,   0,  18,   8,  -3,   0,  -1,   0,   0,   0 },
     {    4,  14,  12, -10,  -2,   2,  -3,   2,   2,  12,   0, -29,  35, -29,  16,  -7,  -9, -37, -16,  60, -55,  41, -18,   7,   8,  22,  12, -25,  18, -10,   1,   1,  -1,  -1,  -5,  -5,   9,  -9,   7,  -3,   0,  -7,   2,   6,  -7,   6,  -3,   1,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,  -1,   0 },
     {   21,  51,  52,  -4, -41,  32, -27,   9,  -6, -23, -59,  16,  18, -14,  16,  -4,   7,  14,  -7,  -7,  14, -12,   7,  -3,   0,  -2,  22,  -4,  -8,   7,  -7,   2,   3, -22,   2,   0,   2,   1,   1,   0,   2,  13,  -3,  -1,   1,  -1,   1,  -1,   0,   2,  -3,   1,   1,  -1,   1,   0,   1,  -3,   1,   0,  -1,   0,   0,   0 },
     {   11,  33, -49,  18,  -5,  -3,   2,  -1,  20,  67, -27,  20, -24,  14, -12,   3,   4,  26, -21,  14, -12,   6,  -5,   2,  -4, -13,   5,  -7,   9,  -6,   4,  -1,   2, -41,  35, -18,  11,  -4,   3,   0,   6,   3,  -6,   3,  -2,   0,   0,  -1,   1,   8,  -6,   2,  -2,   1,   0,   0,   1,  -2,   1,  -1,   1,   0,   0,   0 },
     {   -2,   0,  -1,   0,  -1,   1,  -2,   2,  -5,  -2,  -2,   2,   1,  -1,   1,  -1,  -9,  -2,   2,   1,  -1,   2,  -1,   0, -26,  -5,   1,   0,   1,   1,   0,   0, -49,  -7,  10,  -2,   3,  -1,   1,   0, -78,   4,   5,   1,   2,   0,   1,   0, -80,  12,  10,  -1,   3,   0,   1,   0,   8,   9,  -4,   0,  -1,   0,   0,   0 },
     {    2,   7,   1,  -1,  -4,   4,  -3,   2,   0,   7,  -9,   5,  -2,  -1,   1,  -1,   4,  27, -14,   1,   0,  -1,   1,  -1,   2,  30, -17,   8,  -6,  -1,  -1,   0, -11,  10, -18,   6,  -3,   2,  -1,   1,   1, -85,  57, -19,  10,   1,   0,   2,   4,  44, -17,   2,   0,  -4,   2,  -2,   8,   2,  -8,   4,  -2,   0,   0,   0 },
     {   -7, -11, -33, -65,   3,  43, -31,  25,   7,   7,  29,  56,   0, -36,  25, -19,  -1,  -2,  -1,  -7,  -3,  -1,   3,  -1,  -3,  -8, -11,  -9,   1,  11,  -9,   8,   3,   7,  13,   2,   4, -10,   6,  -5,   0,  -2,  -5,  -3,  -3,   3,  -1,   1,   4,   1,  -3,   4,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   1 },
     {   -9, -28, -25,  14,  -1,   5,   1,   1,  -6, -22, -17,  16,   2,  -4,   6,   0,  11,  37,  46, -14, -18,  12, -13,   3,   1,  21,  14, -46,  31, -19,   7,  -3,  -9, -41, -35,  35,  -9,   4,   5,  -2,   6,  12,   5,   6,  -7,   6,  -5,   2,   1,   9,   6,  -9,   4,  -4,   0,   0,   2,  -4,  -2,   0,  -2,   1,  -1,   0 },
     {   -1,  -1,  -1,   0,   0,   3,   1,   1,  -1,  -1,   0,   2,   0,   1,  -2,  -1,  -1,   0,  -3,  -1,   0,   0,   2,   0,   2,   2,   6,  -2,   1,   0,  -1,   1,   1,  -6,  -4,   2,   1,   2,   0,   0,   4, -14,   3,   0,   2,  -1,   1,  -1, -21,  -2,  -6,   3,   0,   0,   0,   0,-123,   1,  22,  -2,   3,   0,   2,   0 },
     {   -3,  -9,  -6,   0,  -4,   7,   0,   1,   3,  10,  29,   9, -17,   5,  -9,   1, -10, -36, -54,  -6,  36, -13,  12,   0,  12,  45,  65, -17, -11,  -3,  -5,  -3, -11, -28, -24,  10,  -4,  10,   0,   2,   5,   5,   0,   2,   3,  -2,   0,   0,  -3,   6,   3,  -2,   2,   0,   0,   0,   8,  -5,  -5,   2,  -1,  -1,   1,   0 },
     {    4,  18,  -3,  -5,   0,   4,   2,  -2,   7,  33,  -7,  -1,  -1,  -8,  -2,  -3,   9,  49, -18,  -2,  -4,  -2,   1,  -3,   7,  64, -23,   4,  -7,  -2,  -3,  -1,  -2,  55, -34,   7,  -9,   4,  -2,   1, -13,  31, -29,  11,  -3,  -1,   0,   2, -10, -21,   8,  -1,  -1,  -2,   1,  -1,  -3,  -8,   6,  -2,   1,   0,  -1,   0 },
     {   11,  37,  36, -32,   9,  -8,   6,  -4,  14,  50,  55, -40,  13, -13,   3,  -2,   3,  20,  34, -30,  10,  -8,   1,  -1,  -6, -16,   8,  -3,   6,   0,   2,   0,  -6, -32, -16,  10,   0,   0,   2,   0,   3,  -2,   2,   7,  -6,   2,  -1,  -3,   0,  11,  -3,  -1,  -1,   1,   0,   1,  -1,  -7,  -2,  -1,  -1,   2,   0,   0 },
     {   -3,  -7, -13,   0,  15, -12,   7,  -3,  -2,  -8, -13,   1,   6,  -4,  -2,   1,   4,  11,  22,   0,  -3,  -3,   4,  -4,   2,   8,  26,   9, -39,  35, -26,  12,  -2,  -3, -28, -40,  55, -39,  19,  -3,  -4, -14,  -8,  37, -24,   7,   6,  -8,   6,  12,  16,  -9,  -2,   9,  -9,   6,  -1,  -4,  -4,  -1,   2,  -3,   2,   0 },
     {   -5,  -7,  -4, -23, -62, -34,  27,  -8,   5,   4,  11,  37,  62,  34, -24,  -1,  -4,  -4,  -7, -23, -33, -16,   9,   6,   1,   3,   4,   7,   7,  -1,   6,  -3,   2,   6,   2,   5,  13,  -1,  -6,   1,  -1,   2,  -1,  -7, -14,  -1,   1,   2,  -1,   4,  -6,   2,   5,   5,  -3,  -1,   1,  -5,  -3,   1,   0,   0,  -2,   1 },
     {    1,   0,  16,  10, -10,   1,  -4,  -1,  -1,  -4,  10,   6,  -4,   5,   1,   1,   0,   0,   9,  -6,  11,  -6,   3,   0,   4,  13,   9, -24,  26,  -9,   4,  -3,   3,  24,  46, -16, -27,   2,  -2,  -8,  -9, -40, -53,  45, -13,  11,  -5,  -1,  10,  38,  17, -32,  15, -11,  -1,   3,  -4,  -4,  -5,   5,   0,   3,  -1,   2 },
     {    1,   0,   3,   8,   3,   1,   7,  -4,  -1,  -6,   0,  15,  -1, -20,  17, -10,  -2,   1, -13, -55,  -1,  49, -44,  21,   3,   1,  24,  56,   7, -33,  23, -13,   0,   0,  -8, -31,   0,   1,   6,  -5,  -2,  -5,  -8,   1,   3,  10,  -9,   5,   2,   1,   4,  10,  -7,  -2,   5,  -3,   0,   3,   0,  -6,   2,   0,  -3,   1 },
     {   -1,  -4,  -8,  -2,   8,  -7,   3,  -2,  -2,  -6,  -7,  -8,  -2,  -6,   7,  -1,  -1,  -8,   3,   3,   3,   1,  -1,   0,   0,  -5,   4,   3,  -4,   0,   2,   0,   4,  17,   0,   2,   3,  -2,  -3,  -1,   7,  38, -11, -10,   2,   0,  -1,   3, -12,  77, -52,   9,  -9,   2,   1,  -3,   2, -65,  23,   1,   1,   3,  -3,   3 },
     {    2,   0,  32,  17, -34,  17, -20,   8,   0,  -3,  32,  13, -47,  21, -20, -17,   0,  -4,   5,   8,  -8,   4,  18,   4,  -1,  -3, -26, -16,  23, -36,   8,   6,   1,   7, -26, -26,  44, -18,  18,  -1,   1,   9,   5,   3,  10,   1, -13,  13,  -2,   7,   2,   5, -14,  18,  -6,  -6,   1,   0,  11,   1,   0,   0,  -1,  -3 },
     {    1,   4,  -2,  -2,  -2,  -3,  -3,   2,   0,   0, -11,   2,  -6,   4,   7,  -6,  -7, -19, -40,   2,  -3,   5,   5,   7,  -7, -29, -48,  30,  -3,   4,  -1,   3,  -5, -26, -39,  38,  -8,   2,  -3,   3,   3,   4,  -6,  28, -32,  22, -12,  15,   5,  21,  29, -40,  13,  -6,   1,  -4,  -5, -12,  -3,  -2,   8,  -1,   5,  -4 },
     {   -6,  -9, -20, -44, -24,   8,   1,   1,  -1,  -2,  -9, -14,   5,   9,   2,  13,   6,   8,  24,  50,  30, -19, -11,  -8,   0,  -2,   7,  17, -14, -28,  41, -18,  -3,  -5, -20, -42,  -7,  25, -16,  17,   2,   5,   7,  13,   3,  10, -23,   6,   1,   4,   7,  -8,   6,  -4,   8,  -5,  -3,   4,   0,  -3,   5,  -5,   2,  -2 },
     {    0,   2,   6,  -1, -20,  15,  39, -27,   1,   0,   5,  11,   2, -10, -11,  -1,   1,   1,  -5,  11,  45,   8, -54,  50,  -5,  -9, -26, -41, -26,  14,   1, -15,   2,   3,   6,  10,   8,  -9,  20, -15,  -1,  -5,   7,   7,   4,  -6,   0,   5,  -2, -11,   0,   0,   6,   3,  -8,   2,  -1, -20,   0,  -1,  -5,  -2,   0,  -1 },
     {    1,   1,   0,  -7,  10,  53,  10, -32,  -3,  -7,  -4,  -6, -13, -26, -36,  42,  -3,  -6,  -6, -14, -27, -14,   9,   3,   2,   2,   3,   8,  18,  17,  41, -32,   0,   0,   2,  14,  24, -16, -35,  19,  -2,  -9,  -6,  12, -13,  -7,  15,   0,   0,  -2, -12,  -4,  -4,  11,   3,  -4,   4,   3,   8,  -1,   1,   3,   0,  -2 },
     {    1,   2,   0,   0,   9,  -7,   1,   4,   0,   1,  -5,  -8,   9,  -9,   2,  19,  -1,  -2, -17, -24,  25,  -3,  -7, -26,   2,   8,  -7, -20,  25, -13,   2,  -3,   7,  21,  19, -15,  19, -32,  11,  -1,   5,  36,  55,  -3, -28,  23, -15,  -2,  -3,   1, -14, -29,  33, -21,  18,   1,  -7,  15, -17,  -4,   8,  -5,   8,  -3 },
   },
   {
     {  102, -17, -18,  -5,  -4,  -2,  -2,  -1,   6,  -5,   0,   0,   0,   0,   0,   0, -70,   9,  14,   3,   3,   1,   1,   0,  -4,   4,   0,   0,   0,   0,   0,   0,  11,  -1,  -3,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -2,  90, -29, -11,  -7,  -3,  -2,  -1,  -4,   4,  -4,   0,   0,   0,   0,   0, -18, -74,  26,  10,   6,   3,   2,   1,   6,  -3,   4,   0,   0,   0,   0,   0,  11,  18,  -7,  -2,  -1,  -1,   0,  -1,   0,  -1,  -1,   0,   0,   0,   0,   0,  -2,  -2,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {    8,   4,   1,  -1,  -1,   0,   0,   0,  96, -17, -13,  -4,  -3,  -1,  -2,   0,  19, -14,  -4,   1,   0,   0,   0,   0, -72,  11,  13,   2,   3,   1,   2,   0, -20,   9,   4,   0,   0,   0,   0,   0,  11,  -1,  -3,   0,  -1,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -58, -12, -51,  27,  13,   6,   2,   2,  15,   0,  -4,   1,  -1,   0,  -1,   0, -49,  30,  57, -20,  -9,  -4,  -1,  -1, -20,   4,   4,  -1,   1,   0,   0,   0,  31, -11, -21,   5,   2,   1,   0,   0,   5,  -2,   0,   1,   0,   0,   0,   0,  -3,   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -37,  15,  65, -16,  -9,  -3,  -1,  -1,   9,  -5,   0,  -2,  -1,   0,   0,   0, -73,   5, -47,  18,  13,   5,   3,   1, -13,  10,   0,   2,   1,   0,  -1,   0,  38,  -2,   9,  -6,  -5,  -1,  -1,   0,   6,  -3,  -2,   0,   0,   0,   0,   0,  -2,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    0,   2,   0,   1,   1,   0,  -1,   0,  75, -17,  -5,  -1,  -2,  -1,  -1,   0,   7,   3,  -1,  -1,  -1,   0,   0,   0,  87,  -4, -11,  -4,  -3,  -1,  -2,   0,  29, -14,  -3,  -1,  -1,   0,   0,   0, -38,   0,   6,   1,   2,   0,   1,   0, -10,   2,   2,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {  -13, -17, -21, -82,   6,  17,   9,   3,   1,  -3,   4,   2,   3,   0,  -1,   0,   7,  13,  22,  79,  -5, -16,  -8,  -3,  -3,   5,  -5,  -3,  -4,   0,   1,   0,   1,  -4,  -9, -26,   1,   6,   2,   1,  -2,  -1,   3,   3,   2,   0,  -1,   0,   1,   1,   1,   4,   0,  -1,   0,   0,   1,  -1,   0,  -1,  -1,   0,   0,   0 },
     {   -1,  -8,   3,   9,  -1,   0,  -1,   0, -15, -69,  20,  19,   4,   3,   1,   1,   8, -13,   6,  -5,   1,   0,   1,   1,   4,  87, -22, -25,  -4,  -4,  -1,  -2,   4,  22, -12,  -2,   0,  -1,  -1,  -1,   2, -25,   4,   8,   1,   1,   0,   1,  -2,  -7,   4,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {  -21,  12,  -2,  -1,   1,   1,   0,  -1,   9, -14,   3,   2,   0,   0,   1,   0, -47,  12,   4,   1,   0,   0,   1,   1,  25,   3,  -5,  -5,  -1,   0,   0,   0,-103,   0,  15,   4,   5,   2,   1,   0, -13,  10,   2,   2,   0,   0,   0,   0,  33,   0,  -7,  -1,  -2,  -1,  -1,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 },
     {    6,   6,  14,  -6,  87, -18,  -9,  -3,   0,  -1,   2,  -3,   0,  -3,   1,   0,  -6, -10,  -8,   3, -82,  16,   9,   2,   1,   1,  -1,   5,   1,   3,  -2,   0,   3,   4,  -1,   0,  27,  -6,  -2,   0,   1,   0,  -1,  -2,  -3,  -1,   1,   0,   0,  -1,   1,   0,  -4,   1,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   -1,  -5, -11,   3,   7,   0,  -1,   2, -11, -12, -66,  15,  19,   4,   2,   2,   0,   3, -10,   6,  -1,   0,   1,  -1,  17,  17,  87, -18, -27,  -4,  -2,  -2,   2,   2,  21, -11,  -6,   1,   0,   0,   2,  -6, -27,   5,  10,   1,   1,   0,   1,  -2,  -6,   3,   2,   0,   0,   0,  -2,   1,   1,   1,   0,  -1,   0,   0 },
     {   -9, -32,   7,  10,   3,  -1,   0,   1, -11,   5,  -4,  -1,   1,   0,   0,   0, -10, -30,   5,   4,   3,   2,   1,   0, -26,  -5,   9,   4,   0,   0,   0,   0,  -1,  39,  -4, -12,  -4,  -1,  -1,  -1,-102,  15,  10,   0,   4,   1,   2,   0, -14,   0,   1,   4,   1,   0,   0,   1,  19,  -1,  -3,   0,  -1,   0,  -1,   0 },
     {  -10, -47,   8,  11,   0,   2,   1,   1,  13,   4,   2,  -3,  -2,  -1,  -1,  -1, -12, -30,  10,   7,   5,   2,   1,   0,  29, -26,   1,   5,   2,   0,   0,   1,  -2,  78, -14, -17,  -5,  -4,  -2,  -1,  51,  13, -11,  -5,  -3,   0,  -1,  -1,   5, -34,   4,   7,   1,   1,   1,   1, -15,  -2,   3,   1,   0,   0,   0,   0 },
     {    6,   9,   8,  12,  11,  81, -24, -10,  -1,   1,   2,   4,  -4,  -2,  -3,   1,  -3,  -5, -11,  -9, -12, -80,  22,  12,   0,  -3,   2,  -6,   4,   4,   4,  -1,   3,   4,   3,   2,   5,  30,  -6,  -5,  -1,   3,  -2,   1,  -2,  -3,   0,   0,  -1,  -3,   0,   1,  -2,  -6,   1,   1,  -2,  -1,   0,   0,   0,   1,   1,   0 },
     {    3,  -5,  -4,   2,   1,   0,   0,  -1,   1,   7,   2,  -1,  -2,   1,  -2,   0,  10,  -6,  -1,  -1,   0,   0,   0,   1,  -3,   9,  -3,   2,   1,  -2,   1,   1,  35,   0,   9,  -3,  -3,  -2,   1,   0, -12,   8,   0,   0,  -2,   0,   0,  -1, 118,  -7, -19,  -2,  -2,  -1,  -3,  -1,   2, -10,   1,   1,   0,  -1,   0,   0 },
     {   -1,  -1,   0, -10,  -4,  11,  -2,  -2,  -8, -20, -10, -65,  -6,  21,  10,   4,   2,   3,   6,  -3,   2,  -3,   5,   1,  10,  26,  18,  84,  13, -28, -12,  -5,  -3,   3,  -8,  16,   3,  -5,  -3,   0,  -1,  -3,  -6, -28,  -6,   8,   3,   2,   0,   1,   2,  -6,  -3,   1,   0,   0,   4,  -3,  -1,   1,   1,   1,   0,  -1 },
     {    6,   1,  57, -13, -16,  -3,   0,  -1,  -2, -11,   1,   9,   0,   3,  -2,   1,   5,  -6,  39, -10,  -5,  -2,  -3,   0,   0, -13,  27, -11,  -9,   2,   2,  -1,  -8, -19, -80,  16,  20,   4,   4,   0,  -4,  15, -25,   5,   9,  -5,  -1,  -1,  16,  11,  24,  -7,  -8,  -1,  -2,   0,  -2,  -3,   3,   0,  -1,   2,   0,   0 },
     {    7,  24,   7,  -3,   0,   0,  -1,   0,  10,  76, -17, -19,  -7,  -4,  -7,   1,   4,  38,  -2,  -5,   0,   0,  -4,   0,   6,  46,  -5, -10,  -3,  -1,   1,  -4, -10,  34, -30,  -4,   0,  -1,   1,  -2, -12, -38,   2,  20,   3,   2,  -1,  -1,   2, -28,  14,   4,  -2,   1,   1,   0,  -3,   1,   1,  -4,   0,   0,   1,   1 },
     {   -5,  -6,  -7, -13, -13, -19, -86,   8,   1,   4,   2,   3,   3,   7,   3,   4,   5,   8,   6,   6,   9,  18,  81,  -7,  -2,  -3,  -2,  -7,  -2,  -5,  -1,  -3,  -1,  -1,  -3,   1,  -3,  -4, -25,   0,   0,   1,   0,   1,  -1,   2,   5,   1,   1,   1,   4,   1,  -1,   1,   5,   0,   1,  -1,   1,   0,  -1,   0,  -1,  -1 },
     {    8,  50,  -8,  -5,  -1,  -3,   1,   1,  -8, -50,   4,   4,  -1,  -1,  -1,   1,  12,  68, -10,  -6,  -3,   2,  -4,  -4, -10, -31,   4,   1,   8,   3,  -1,  -1,  11,  42, -11,  -9,  -4,  -2,  -2,  -2, -14,  33,  -3,   1,   0,  -2,  -2,  -2,  -1, -36,  -2,   8,   2,   1,   0,   2,   1,  -9,   4,   0,   1,   0,  -3,   2 },
     {   -1,  -4,  -3,   3, -13,   0,   5,  -7,  -4,   1, -20,  16, -61,   4,  23,   2,   1,  -1,  -2,   8,  -6,   4,   1,   6,   4,   5,  14, -24,  85,  -9, -32,  -5,   1,  -5,  -1, -10,  22,  -9,  -9,  -1,  -2,   3,  -3,   6, -26,   2,  10,   0,  -1,   5,   1,   3,  -7,   2,   2,   0, -10,   1,  -1,   0,   2,  -2,  -1,   1 },
     {   -1,  -1,  -3,   2,  -1,  -3,   1,   4,  -3,   0,   0,   0,   6,  -1,  -1,   1,   0,  -4,  -3,   2,  -2,   1,   1,   0,  -9,   2,  -5,   9,  -7,  -4,  -2,   1,   1, -11,   3,  -1,   0,   5,  -2,  -3, -22,   5,  -3,  -2,   1,   2,   0,   1,  -2, -12,   2,  -2,  -1,  -1,   1,   0,-121,  -3,  21,   0,   3,   3,   2,   1 },
     {   -5, -10,  -4, -59,  10,  16,   9,   0,   2,   8,  -3,  25, -10,  -3,  -7,   1,  -2,  -2,   2, -35,   6,   8,   0,   2,   2,   5,   1,  -8,   3,   2,  -1,  -6,   7,  24,  19,  83, -11, -21, -11,  -9,  -4,  10,  -6,   0,   1,  -1,  -1,  -3,  -5,  -3,  -9, -27,   4,   7,   4,   3,  -4,  -6,   5,   5,   1,  -1,   0,   4 },
     {    3,   4,   8,   6,   2,  15,   4,  83,   0,   2,  -3,  -3,  -4,   1,   1,   1,  -2,  -4,   1,   1,  -7,  -7,  -3, -86,   1,   2,  -2,  -5,   7,  -5,  -3,   4,   0,   1,   3,   5,   3,   0,  -2,  33,   1,   3,  -1,   1,  -3,   3,   0,  -3,   0,   0,  -6,  -5,  -1,  -2,  -1,  -6,   1,   5,  -1,   1,   2,  -3,  -2,   0 },
     {   -1,  -4,  -7,   4,  -4,   0,   1,   5,  -8, -14, -72,  17,  18,   2,  -2,   6,  -2,  -4, -20,   9,   0,  -1,   1,   4,  -4, -13, -27,   7,  -3,  -3,  10,   2,  -2,   2, -27,  18,  14,   3,   3,  -3,   8,  -4,  71,  -8, -21,   2,  -4,  -1,  10,  -3,  28, -16, -10,   0,  -4,   3,  -1,  -1, -16,  -1,   8,   1,   1,   0 },
     {   -1,  -1,  -2,  -8,  -1,   3,   1,   2,  -2, -25,  12,  -4,   2,  -5,   3,   8,  -2,   6,  -4,  -5,  -3,  -3,   2,   0,  -3, -48,  11,   4,   6,   7,  -3,  -8,  -1,  21,  -5,   8,  -4,  -9,   2,   3, -10, -93,  -3,  18,  -2,   8,   8,   6,  17,  25,  -2,  -2,   7,  -2,  -3,  -3,  -9,  40,   3, -11,   1,  -5,  -2,  -2 },
     {    1,   5,   4,  -2,   5,  12,  -7, -18,   6,  11,  10,  24,  25,  56,   8, -18,  -1,   4,   1,  -7,  -2,  -1, -14,  -4,  -5, -12,  -6, -14,  -8, -82, -19,  32,   0,   6,   1,  -4,  -2, -17,  12,  13,   0,  -3,   3,   2,  -7,  26,  14,  -7,  -1,   8,   2,   6,   2,   8,   1,  -3,   1,  -4,   5,   0,   1,   3,   1,  -3 },
     {   -6,  -6, -52,   9, -10,  10,  -1,  10,   3,   3,  22,   6,   5,   6,  -3,  -1,  -7,  -6, -65,  15,   0,   7,  -3,   0,   0,   2,  -1,  -2,  -4,   1,   7,  -4,   0,   0, -25,  14,  35, -16,  11,  -9,  -7,   5, -48, -13,  12,   1,   1,   3,   7,   3,  46,  -8, -16,   8,  -6,   6,   5,   3,  -4,   0,  -1,  -1,  -4,   1 },
     {    0,  -4, -10, -12,  49, -12, -20,  16,  -3,  -7,  -4,   2, -34,   8,   0,  17,  -2,  -9, -18, -14,  31,  -2,  -9,   6,   0,  -2,  -1,   6,  13, -12,  -1,   2,  -5, -11, -22,  -3, -64,  28,  45,  -2,   4,   8,  -2,  25,  18,   0,  -2,  -5,   4,   3,  21,   5,   1,  -8,  -9,  -1,  -2,   2,  -9,  -6,  -4,   5,  -4,   2 },
     {   -2,  -7,  -7,  -8,  -4,  11,  -4,   7,  -4, -12,   4, -61,   1,   3,  -7,  -1,  -1,  -6,  -4,  -8,   6,  15,  -3,  12,  -2, -10,   5, -29, -13,  -2, -15,  18,   5,   4,  11,   6,  31,   2,   5,   1,   5,   6,  14,  61,   4, -10,  10, -11,  -6,   5, -12, -10, -47, -19, -10,  -2,  -5, -28, -14, -14,   8,   8,  -3,   6 },
     {   -2,  -9,   2,  -7,   2,  16,   6,  -9,  -5, -12,   0,  -9,  -6,   5, -22, -13,  -3, -19,   3,  -7,   2,  11,  10,  -8,  -3, -14,   6,  -5,  17,  -3,  40,   3,  -6, -43,   3,   4,  -1, -16,   5,   8,  -9, -31,  -2,   2,  -3,  10,   5, -11,  -3, -82,   6,  28,   3,  13,  -2,   2,   8, -22,   0,   1,   0,  -7,  -8,  -4 },
     {    2,   4,   0,   7,  25,  16,  19, -14,   0,  -2,   4,   4,  -1, -21,  14, -29,   2,   3,  -1,   6,  23,   4,  19,  -5,   0,   2,   5,   9, -12,  34, -44,  49,  -5,  -4,  -9,  -4, -25, -22, -30,  27,  -1,  -7,   2, -11,  -3, -13,  23, -27,   6,   4,  32,  -5,  -1,  13,   3, -12,  -3, -16, -16,  -4,  -2,   0, -14,   3 },
   }
 },
 {  //18
   {
     {  122, -19,  -9,  -5,  -4,  -2,  -2,  -1,   4,  -1,  -1,   0,   0,   0,   0,   0, -28,   4,   2,   1,   1,   0,   0,   0,  -8,   2,   1,   0,   0,   0,   0,   0,  -5,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    5,  -1,   0,   0,   0,   0,   0,   0,-121,  14,   9,   4,   4,   2,   2,   1,  -8,   0,   1,   0,   0,   0,   0,   0,  35,  -3,  -3,  -1,  -1,   0,  -1,   0,  11,  -1,  -1,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },
     {   27,  -6,  -1,  -1,  -1,  -1,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0, 121, -10,  -9,  -4,  -4,  -2,  -2,  -1,  -4,   0,   0,   0,   0,   0,   0,   0, -21,   2,   1,   0,   1,   0,   0,   0,  -6,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 },
     {   -8,   2,   1,   0,   0,   0,   0,   0, -35,   3,   2,   1,   1,   1,   1,   0,  -6,   0,   1,   0,   0,   0,   0,   0,-118,  10,  10,   3,   4,   1,   2,   0, -15,   1,   2,   0,   1,   0,   0,   0,  24,  -2,  -2,  -1,  -1,   0,   0,   0,  12,  -1,  -1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0 },
     {   17, 111, -15,  -2, -10,  -4,  -4,  -2,   4,  39,  -6,  -1,  -3,  -2,  -1,  -1,  -2, -38,   4,   0,   3,   1,   1,   1,  -1, -16,   2,   0,   1,   0,   0,   0,  -3,  -2,   0,   0,   0,   0,   0,   0,  -1,  -4,   0,   0,   0,   0,   0,   0,  -1,  -3,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   -7,  -1,   0,   0,   0,   0,   0,   0,  -3,  -1,   1,   0,   0,   0,   0,   0, -17,   3,   1,   1,   0,   0,   0,   0,  24,  -1,  -1,  -1,  -1,   0,   0,   0,-120,   9,  10,   2,   4,   1,   2,   0,  23,  -1,  -1,   0,  -1,   0,   0,   0,  15,  -1,  -1,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },
     {    6,   0,   0,   0,   0,   0,   0,   0,  10,  -1,  -1,  -1,   0,   0,   0,   0,  10,  -1,  -1,   0,   0,   0,   0,   0,  17,   0,   0,  -1,   0,   0,   0,   0,  23,  -1,  -3,  -1,  -1,   0,  -1,   0, 121, -10, -11,  -2,  -4,  -1,  -2,   0, -10,   0,   0,   0,   0,   0,   0,   0, -19,   1,   1,   0,   0,   0,   0,   0 },
     {   -1,  -5,   1,   0,   1,   0,   0,   0,   8,  86,  -8,  -6,  -6,  -3,  -2,  -2,   6,  83,  -6,  -6,  -5,  -3,  -2,  -2,  -2, -15,   1,  -1,   1,   0,   1,   0,  -2, -35,   2,   2,   2,   1,   1,   1,   0, -10,   1,   1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0 },
     {    8,  12, 102, -27,   6, -10,   0,  -3,   4,   7,  55, -15,   3,  -5,  -1,  -1,  -3,  -5, -29,   6,  -2,   2,   0,   1,  -2,  -3, -22,   5,  -2,   2,   0,   1,  -1,   0,  -5,   1,  -1,   1,   0,   0,   0,   0,  -5,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {   -7, -54,   4,   3,   5,   2,   2,   1,   8,  79,  -9,  -4,  -5,  -3,  -2,  -1,  -7, -74,   7,   4,   5,   3,   1,   1,  -2, -14,   1,   0,   1,   1,   0,   0,   1,  34,  -1,  -1,  -2,  -1,   0,  -1,   0,  -4,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   -5,   0,   1,   1,   0,   0,   0,   0,  -6,   1,   0,   0,   0,   0,   0,   0,  -7,   1,   0,   0,   0,   0,   0,   0, -10,   1,   1,   1,   0,   0,   0,   0, -17,   1,   1,   0,   0,   0,   0,   0,  -7,   0,   2,   0,   0,   0,   0,   0,-122,   9,  11,   2,   4,   1,   2,   0, -25,   2,   3,   0,   1,   0,   0,   0 },
     {   -1, -14,   1,   2,   0,   1,   0,   1,  -1, -20,   2,   4,   0,   1,   0,   0,  -1, -20,   3,   5,   1,   2,   0,   0,  -7,-101,  11,  11,   5,   5,   1,   2,  -4, -59,   5,   7,   3,   3,   1,   1,   3,  25,  -2,  -2,  -1,  -1,   0,  -1,   1,  24,  -2,  -2,  -1,  -1,   0,   0,   0,   7,  -1,  -1,   0,   0,   0,   0 },
     {    1,   1,  11,  -1,  -2,   0,   0,  -1,  -4,  -4, -62,  20,  -3,   6,   1,   1,  -7,  -4, -95,  27,   0,   6,   2,   2,   0,   0,   2,   1,   3,  -1,   1,   0,   4,   1,  42, -11,   2,  -3,  -1,   0,   1,   0,  13,  -4,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   3,  -1,   0,   0,   0,   0 },
     {   -1,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0, -16,   0,   1,   0,   1,   0,   0,   0,  27,  -1,  -2,  -1,  -1,   0,   0,   0,-123,   8,  10,   2,   3,   1,   1,   0 },
     {    5,   2,  23,  87, -20,  -1,  -7,  -1,   4,   4,  22,  73, -19,  -3,  -5,  -1,   0,   1,   2,   1,  -5,  -3,   0,   0,  -1,  -1,  -9, -28,   5,  -1,   2,   0,  -1,  -1,  -8, -20,   6,   0,   1,   0,   0,   0,  -3,  -7,   2,   1,   0,   0,   1,   0,   0,  -2,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0 },
     {    4,   5,  59, -24,   4,  -7,   1,  -2,  -5,  -6, -81,  32,  -4,   8,  -1,   1,   3,   2,  48, -17,   3,  -3,   0,   0,   2,   1,  17, -10,   2,  -2,   0,   0,  -3,  -3, -33,  11,  -1,   3,   0,   1,   0,  -1,   9,  -2,   0,  -1,   0,   0,   0,  -1,  -2,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
     {   -1,  -8,  -1,   0,   0,  -1,   1,   0,   1,   9,  -1,   0,  -1,   0,   0,   0,  -3, -38,   3,   2,   2,   2,   0,   1,   5,  65, -11,  -5,  -4,  -2,  -2,  -1,  -6, -92,   8,   9,   5,   5,   1,   2,  -1, -16,   3,   2,   1,   1,   0,   0,   2,  37,  -2,  -4,  -1,  -2,  -1,  -1,   0,  -3,   0,   0,   0,   0,   0,   0 },
     {    4,   3,  10,  47,  24,  -5,  -7,  -3,  -2,  -4, -12, -32,   6,   0,   2,   1,  -6,  -7, -23, -95, -19,   7,  11,   5,   0,  -1,   3,   3,  -1,   3,   1,   0,   2,   3,  12,  41,   7,  -2,  -4,  -2,   0,   2,   2,   4,  -1,  -1,   0,  -1,   0,   2,  -2,  -5,  -2,   0,   1,   0,   0,  -1,   0,   2,   0,   0,   0,   0 },
     {    1,   1,  20,  -1,  -5,  -1,  -1,   0,   1,   2,  17,  -3,  -4,  -1,  -1,   0,   2,   0,  21,   0,  -3,  -2,  -2,   0,   9,   9,  96, -13,  -5,  -5,  -4,  -2,   6,   0,  66, -13,  -4,  -3,  -2,  -1,  -2,  -2, -17,   1,   0,   1,   1,   0,  -3,   0, -25,   4,   1,   1,   1,   0,   0,  -1,  -8,   1,   0,   0,   0,   0 },
     {   -6,  -9,  -2, -16, -81, -31,  -4,   2,  -5,  -7,  -4, -22, -73, -26,   1,   3,  -1,  -1,  -3, -15, -11,  -1,   5,   2,   1,   1,  -3,   3,  25,  10,   3,   0,   2,   4,  -1,   9,  22,   9,  -1,  -3,   1,   0,   0,   4,  11,   3,  -1,  -1,   0,   1,   0,   1,   2,   1,   0,   0,   1,   2,   0,  -1,   1,   0,   0,   0 },
     {    0,  -6,   0,   1,   1,   2,   0,   0,  -1, -15,   4,   3,   1,   1,  -1,  -1,   0,  -9,   0,   1,   0,   1,   0,   0,  -2, -23,   1,   1,   1,   2,   1,   0,  -1, -14,   2,   8,   1,   1,   0,   0,  -8,-109,   7,  16,   5,   7,   1,   3,  -3, -50,   3,   9,   2,   3,   0,   1,   2,  15,   0,   0,  -1,  -1,   0,   0 },
     {    6,   6,  21,  61,  25,   3,  -6,  -3,  -6,  -8, -27, -73, -32,  -1,   8,   6,   3,   4,  16,  43,  13,   0,  -4,  -3,   1,   3,   1,  21,   9,   2,  -3,  -2,  -2,  -4, -10, -31, -11,   1,   4,   2,   0,  -2,   5,   9,   4,   1,  -1,  -1,   1,   0,  -1,  -1,   0,   0,   0,   0,   0,   2,   1,   1,   1,   1,   0,   0 },
     {    0,  -2,  11,   8, -50,  51, -14,   6,   0,   1,  -2,  -9,  20,   5,   1,  -1,   1,   2,  -7, -18,  78, -44,  15,  -7,   0,   0,  -1,  10,  -4,  -3,   0,   0,  -1,  -1,   3,  10, -33,  14,  -5,   3,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0,  -2,  -1,   4,  -4,   1,   0,   0,  -1,   0,   1,  -3,   1,   0,   0 },
     {   -1,  -1,  -6,   4,  -1,   3,   1,   2,   0,   1,   8,  -8,   1,   2,   2,  -1,  -3,  -3, -45,   9,   5,   4,   1,   1,   5,   5,  66, -20,  -3,  -2,  -1,  -1,  -6,  -4, -78,  12,   6,   4,   4,  -1,  -4,  -3, -32,  13,   5,   0,   2,   1,   4,   9,  37,  -7,  -1,  -2,  -1,   0,   0,   1,   0,  -1,  -2,   0,   0,   0 },
     {   -1,  -1,  -5,   8,  15, -65,  33, -17,   0,  -1,  -4,   7,  29, -70,  38, -15,   1,   1,  -3,   1,  31, -31,  16,  -5,   1,   0,   4,  -4,   9,   6,  -3,   2,   0,  -1,   1,  -4,  -6,  15,  -7,   4,   0,   0,   2,   1,  -7,  12,  -7,   2,   0,  -1,   2,   1,  -5,   7,  -4,   1,   0,   0,   0,  -1,  -3,   3,  -2,   0 },
     {    1,   2,   3,  17,  -4,  -6,   0,  -3,   1,   4,   4,  25,  -5,  -6,   0,  -1,   1,   3,   4,  30, -13,  -7,  -3,  -1,   4,  10,  16,  88, -15, -12,  -7,  -4,   2,   8,  -1,  64,  -1,  -4,  -7,  -3,  -1,  -3,  -2, -18,   0,   2,   0,   1,   0,   1,   0, -27,   2,   3,   1,   2,   0,   0,  -2,  -9,   1,   2,   1,   0 },
     {    1,   6,   0,  -2,   0,   2,   3,   0,   0,  -3,  -1,  -1,   3,   2,   4,   1,   1,  13,   1,  -1,  -2,  -2,  -2,   0,  -1, -11,  -3,  -6,   8,  -2,   2,   0,   3,  36,   0,  -5,  -1,   2,   0,   0,  -3, -51,   6,  -6,  -1,   5,  -3,   2,   5, 103, -17, -17,  -5,  -6,  -1,  -3,   1,  20,  -8,  -7,  -3,  -1,   0,   0 },
     {    0,   2, -12, -18,  58, -12,   4,  -9,   0,  -1,  18,  24, -72,  24,  -8,   6,   1,   1,  -8, -14,  48,  -7,   2,  -3,   1,   2,   1,   0,  38,  -8,   3,  -3,   0,   0,   7,  14, -32,   2,  -4,   3,   0,   1,  -6,  -7,   7,  -4,   1,   1,   0,   0,   5,  -2,   0,   1,  -1,   0,   0,   0,   0,  -3,  -1,  -1,   0,   1 },
     {   -3,  -5,  -2,   7, -24, -58, -28,   1,   1,   2,   0,  -3,  19,  25,  -1,   5,   4,   7,   4,  -1,  43,  75,  20,   4,  -1,  -2,  -1,   4, -19, -11,  -9,  -1,  -2,  -1,  -1,  -2, -25, -33,  -8,   2,   0,  -3,  -2,  -4,   3,   2,   4,   1,   1,   4,   2,   0,   5,   6,   2,  -1,   0,   4,   2,   1,  -1,  -3,   0,   0 },
     {   -1,   0,  -5,  -1,   4,  -3,   4,  -4,  -1,  -1, -16,   4,   1,   3,   4,   0,  -1,  -1,  -6,  -4,   2,  -2,   3,  -2,  -1,   0, -20,  13,  -7,   2,   6,   1,  -2,  -2,  -9,  -8,   9,   5,   0,   2,  -8,  -4, -97,  -2,  23,   4,   2,   0,  -6,  -3, -64,  11,   9,  -1,   4,   1,   2,   6,  21,   4,  -3,   0,   1,  -2 },
     {   -2,  -2,  -5,  -4,  16, -24, -78,  -4,  -2,  -2,  -4,   2,   7, -34, -66,   6,   0,   1,  -2,   9,  -1, -20,  -2,  11,   1,   1,   3,  -2, -10,  22,  30,   3,   1,   2,   1,   4, -12,  23,  26,  -6,   0,  -1,   0,   8,  -5,   4,   8,  -4,   0,   5,  -4,   1,   0,  -4,   1,   1,   0,  -1,   1,   1,   1,  -1,  -1,   0 },
     {    0,   1,  -1,   8,   2,  -6,   0,  -1,   0,   0,   0,  -8,  -4,   9,   9,   3,   2,   4,   4,  35,  14,  -1,  -6,  -8,  -4,  -5, -12, -56, -15,  -7,  -3,   2,   3,   8,  10,  65,  14,   6,  -4,  -2,   2,  10,   8,  56,   6, -12,  -8,  -6,  -4,  -6, -33, -35,  -8,  -7,   2,   3,   0,  -2,   9,  -4,  -1,   1,   5,   3 },
   },
   {
     {   80, -58,  11,  -5,   2,  -1,   1,   0, -66,  43,  -4,   1,   0,   0,   0,   0,  14,  -5,  -4,   2,  -1,   0,   0,   0,  -2,  -1,   2,   0,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   68,  25, -50,  13,  -9,   2,  -3,   0,   2, -53,  45, -10,   5,  -1,   2,   0, -38,  37, -11,  -2,   3,  -2,   1,  -1,  15, -10,  -3,   5,  -3,   1,  -1,   0,  -8,   3,   4,  -2,   1,   0,   0,   0,   2,   0,  -2,   1,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {   16, -50,  44, -16,   7,  -2,   2,   0,  59,   5, -32,   6,  -4,   0,  -1,   0, -70,  31,   3,   2,   0,   1,   0,   0,  26, -12,  -1,   0,  -1,   0,   0,   0,  -9,   3,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   58,  10,  -7, -10,  -1,  -2,   0,  -1,  59, -14, -24,  12,  -5,   2,  -2,   1,  17, -45,  31,  -7,   2,   0,   0,   0, -53,  41,  -7,  -3,   4,  -2,   1,  -1,  13, -10,  -3,   5,  -2,   1,  -1,   0, -10,   3,   3,  -2,   1,   0,   0,   0,   1,  -1,  -1,   1,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0 },
     {    2,  32, -36,  22, -12,   5,  -3,   1, -14,  45, -26,   3,  -3,   0,  -1,   0, -46, -46,  53, -13,   8,  -1,   2,   0,  49,   2, -17,   0,  -1,  -2,   0,  -1, -19,  -1,   5,   1,   0,   0,   0,   0,   8,  -1,   0,  -1,   0,   0,   0,   0,  -3,  -1,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },
     {  -11, -49, -31,  59, -12,   8,  -3,   3,  43,  37,  34, -59,   7,  -5,   1,  -2,   1, -18, -18,  17,   4,  -2,   1,   0, -12,   8,   5,   1,  -5,   2,  -1,   0,   1,   0,  -1,  -4,   2,   0,   0,   0,  -2,  -1,   0,   3,  -1,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0 },
     {  -23,  -9,  -1,   5,   1,   1,   1,   0, -37,  -2,  12,  -1,   1,   1,   0,   0, -63,  12,  13,  -5,   4,  -1,   2,   0, -37,  36, -15,   4,   0,   0,   0,   0,  70, -35,   0,   3,  -4,   2,  -2,   1, -23,  11,   4,  -4,   2,  -1,   1,   0,  13,  -3,  -3,   1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {   16,  50, -10,  -6,  -6,  -1,  -2,   0,   7,  44, -26,  -2,   0,  -2,   0,  -1, -25,   3, -35,  30,  -7,   4,  -1,   1, -27, -39,  59, -20,   3,   2,   0,   1,  29,  11, -21,  -1,   5,  -4,   1,  -1, -12,  -8,   7,   4,  -2,   1,   0,   0,   4,   1,  -1,  -2,   1,   0,   0,   0,  -1,  -2,   1,   1,   0,   0,   0,   0 },
     {    6,  18,  30, -33,  18, -12,   4,  -2,  -8,  -8,  30, -23,   3,  -1,   2,  -1, -24, -49, -35,  55, -11,   8,  -2,   2,  28,  50,   0, -22,  -2,  -1,  -1,  -1, -18, -15,   3,   4,   3,  -1,   1,   0,  10,   3,  -3,   2,  -2,   0,   0,   0,  -4,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0,   0 },
     {   -8, -13, -40, -20,  64, -26,   8,  -1,  13,  23,  30,  29, -69,  25,  -5,   0,  -4,  -6, -14, -15,  25,  -3,  -2,   2,   1,  -4,   8,   1,   2,  -6,   3,  -1,   0,   0,   2,   1,  -7,   5,  -1,   0,  -1,   0,  -3,  -1,   5,  -2,   0,   0,   1,   0,   2,   0,  -3,   1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0 },
     {   13,   5,   2,  -1,  -1,  -1,   0,   0,  21,   2,   0,  -3,   0,   0,  -1,   0,  37,  -3,  -8,   1,  -1,  -1,   0,   0,  74, -10,  -9,  -1,  -3,   1,  -2,   0,  53, -26,  -3,   3,  -2,   1,  -1,   1, -67,  18,   9,  -3,   3,  -1,   1,   0,  11,  -1,  -5,   1,   0,   0,   0,   0,  -6,   0,   2,   0,   0,   0,   0,   0 },
     {    8,  29,  -8,  -2,  -2,  -2,   0,  -1,  12,  46, -20,  -3,  -1,  -2,  -1,   0,  15,  49, -27,   4,  -6,  -1,  -1,  -1,  -9,  19, -27,  11,  -1,   0,   0,   0, -21, -65,  49,  -5,   0,   4,  -1,   2,  18,  17, -12,  -6,   4,  -3,   1,  -1,  -7,  -6,   2,   4,  -1,   0,   0,   0,   1,  -1,   1,  -1,   0,   0,   0,   0 },
     {  -12, -31, -47,  10,  12,   3,   1,   2,  -5, -19, -23,  21,   4,  -2,   0,   0,   9,  22,  22,  27, -42,  13,  -5,   0,  11,  15,  25, -62,  29,  -6,  -1,   1, -10, -26, -13,  26,   2,  -7,   4,  -1,   6,  13,   5,  -6,  -7,   5,  -1,   0,  -2,   0,  -1,  -1,   3,  -1,   0,   0,   1,   0,   1,   0,  -1,   0,   0,   0 },
     {    0,  -6,  16,  31, -44,  28, -14,   5,  -6, -18, -11,  33, -25,  13,  -4,   2,  -3, -11, -38, -20,  55, -26,  10,  -2,  12,  25,  40, -23,  -9,  -1,   1,  -2, -10, -23,  -4,  13,  -4,   5,  -2,   1,   8,  11,  -3,  -5,   3,  -3,   0,   0,  -4,   0,   1,  -1,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0 },
     {   -7,   2,   1,   0,   0,   0,   0,   0, -10,   2,   3,   0,   2,   0,   0,   0, -14,   2,   0,  -1,   1,   0,   1,   0, -26,   2,   4,   0,  -1,   1,   0,   0, -74,   0,   4,   3,   3,   0,   2,   0, -88,  14,  15,  -3,   3,   0,   1,   0,  39,  -2,  -8,   2,  -1,   0,  -1,   0,   8,  -1,   1,   0,   0,   0,   0,   0 },
     {  -10, -17, -28, -52, -28,  40,   9,   1,  16,  26,  38,  48,  33, -44, -10,  -1,  -7, -15, -20, -25, -14,  14,  10,   0,   2,   3,   4,   9,  -1,   4,  -6,   0,   1,  -1,   0,   5,   3,  -6,   1,   1,   0,  -2,  -3,  -6,  -2,   5,   1,  -1,  -1,   2,   2,   3,   1,  -2,  -1,   0,   0,  -1,  -1,  -1,  -1,   1,   0,   0 },
     {    4,  19,  10,  -6,   0,  -2,  -1,  -1,   6,  30,  14, -15,   0,  -2,   0,  -1,   9,  40,   2, -20,  10, -10,   1,  -1,   8,  39, -18,  -8,   2,  -4,   2,  -2,  -9,   9, -58,  46, -14,   5,  -1,   0,   5, -41,  45, -14,  -5,   7,  -3,   3,  -2,   7,  -5,  -6,   8,  -4,   1,   0,   0,  -4,  -1,   5,  -2,   0,   0,   0 },
     {    3,   6,  23, -20,   9,  -6,   1,  -1,   4,   4,  35, -33,  12,  -7,   3,  -2,   1,  -9,  39, -23,  15, -11,   4,  -2, -10, -47,  28, -10,  -1,   5,  -1,   2, -10, -48, -18,  35,  -8,   6,  -1,   1,  18,  47, -12, -11,  -2,   0,  -1,   0,  -6,  -6,  -5,   5,   1,   0,   0,   0,   0,   3,  -2,   2,  -2,   1,   0,   0 },
     {   -6, -10, -29, -26,  20,   0,   3,   0,  -4,  -6, -24, -10,  18,  -3,   4,   1,   7,  14,  16,  13,  20, -26,   7,  -4,   7,  23,  25,  19, -70,  34,  -7,  -1,  -5, -11, -25, -22,  41,  -4,  -8,   6,   4,   3,  11,  11,  -7, -12,   8,  -3,  -2,  -1,   1,  -3,  -4,   7,  -2,  -1,  -2,   0,  -1,   1,   3,  -3,   0,   1 },
     {   -2,  -1,   0,  -1,   2,   0,   1,   0,  -3,   0,  -1,  -1,   2,   1,   0,   1,  -2,  -1,   0,  -1,   1,  -1,   0,  -1,  -7,  -1,  -3,  -3,   0,   2,   1,   0,   0,  -1,  -1,   1,   2,   0,  -1,  -1, -33,   0,  -1,   3,   0,  -1,   0,   0, -91,  15,  18,  -4,   4,   1,   1,  -1,  79,  -7, -12,   2,  -3,  -1,  -1,   0 },
     {    2,   3,   4,  -1,  25,  33, -64,  29,  -1,  -2,  -3,  -3, -22, -43,  70, -28,   1,   1,   3,   2,  10,  22, -27,   5,  -1,   1,  -1,   4,  -5,   0,  -4,   7,   1,  -2,   0,  -2,  -2,  -4,  11,  -7,  -1,  -1,   2,   0,   4,   2,  -8,   4,  -1,   0,  -1,   1,  -3,  -1,   4,  -1,   0,   0,   0,   0,   1,   1,  -2,   1 },
     {   -6, -15, -37, -19,  18,  -1,   6,  -1, -10, -26, -54, -28,  27,  -2,   6,   1,  -9, -27, -49,  -9,  24,   0,   3,   2,  -4, -20, -23,  21,  10,  -2,   2,   0,   0,  -2,   2,  40, -27,   6,  -1,   0,   6,   9,  21, -15,   0,   5,  -5,   1,   0,   1,   4,   0,   3,  -4,   0,   0,   0,   3,   4,   0,  -3,   0,   1,  -1 },
     {   -4,  -5, -11, -35, -18,  31,  -9,   5,   2,   5,   7,  -5, -10,   4,  -2,   2,   6,  10,  22,  52,  34, -50,   8,  -6, -11, -21, -39, -43,   2,  20,   5,   0,   7,  20,  25,  11, -19,   4,  -7,   1,  -6, -10, -10,   3,   8,  -3,   1,   1,   4,   1,   1,  -2,  -1,   0,   0,   0,  -2,   1,   1,   1,   0,  -1,   0,  -1 },
     {    0,   9,  -5,  -3,   3,   1,   0,   1,   1,  15,  -8,  -8,   5,  -2,   0,  -1,   1,  20,  -9,  -8,   3,  -2,  -2,   1,   4,  33,  -7,  -5,   3,  -5,   3,  -2,   5,  58, -24,   9,  -5,  -4,   1,  -2,  -3,  70, -42,  -1,   2,  -2,   0,  -1,  -3, -51,  19,  12,  -7,   4,  -1,   2,   2,   3,   4,  -6,   1,   1,  -1,   0 },
     {    0,  -2,   3,  15, -20,  14, -11,   5,   1,  -2,   7,  32, -32,  19, -11,   3,  -2, -10,  -6,  48, -27,   9,  -4,  -1, -10, -26, -36,  41, -21,  12,  -1,   1,  -6, -11, -45,  18,  11,   0,   0,   1,  11,  32,  20, -28,   7,  -8,   3,  -2,  -6,  -6,  -1,  -1,   4,  -2,   1,   0,   2,   1,   2,  -3,   2,  -2,   1,  -1 },
     {   -5,  -1,   0,   0,  -2,  -2,   1,   0,  -6,   0,   1,   1,   0,   1,   1,   1,  -6,   1,   2,   1,   2,   1,  -1,   1,  -8,   0,   2,   1,   2,  -2,   1,   1, -11,   1,   2,   1,  -1,  -2,   2,   0, -28,  -1,   2,   0,  -2,   2,   0,   0, -75,  -1,   3,   2,   4,   1,   0,   1, -97,   6,  12,  -1,   2,   0,   2,   1 },
     {   -5,  -6, -12, -50, -35,  40,  -9,  12,  -4,  -5, -10, -54, -41,  44, -12,  10,   1,   5,   7, -16, -13,  21,  -6,   3,   4,   7,  13,  15,  20, -21,   6,  -4,   0,  -5,  -2,  -6,  23, -18,   9,  -3,   0,  -2,   3,  14,   1,  -3,   0,  -1,   0,   2,  -1,   6,  -4,   0,  -1,   0,   1,  -2,   0,   0,   2,   0,   0,   0 },
     {   -3,  -5,  -6,  -8, -35, -19,  45, -19,   1,   0,   2,   5,  -2,  -6,  15, -10,   3,   6,   8,  13,  36,  31, -75,  35,  -2,  -4,  -4, -10, -26, -13,  32,  -6,   1,  -2,   2,   6,   3,  -1,   8, -10,   1,   2,   1,   0,  -2,   5, -10,   5,   2,   0,  -1,  -2,   1,   1,   1,  -1,   3,   0,  -1,   2,   2,  -1,  -1,   1 },
     {   -2,  -4,  -7,  -8,  -1,   8,  -5,   0,  -1,  -2,  -2,  -1,  11,  -3,  -5,   7,   3,   7,  18,  25,  -5,   0,   1,  -6,   7,  12,  34,  28,  -9, -40,  17,  -5,  -4,  -3,  -9, -17, -49,  56, -11,   1,  -3, -11,  -8, -16,  52, -14, -12,   9,  -2,  -1,   4,  12, -15, -13,  14,  -3,   2,  -3,  -1,  -6,  -2,  11,  -5,   0 },
     {   -5,  -8, -10, -17, -28, -47, -21,  56,   6,   8,   9,  16,  34,  52,  21, -51,  -2,  -4,  -5,  -5, -11, -27,  -3,  11,   1,   0,  -1,   2,   8,  -9,   1,   9,   0,  -1,  -2,  -4,   3,  16,   1, -11,   1,   3,   2,   6, -10,  -6,  -2,   6,   1,   1,   1,  -3,   3,   3,   2,  -3,   0,   1,   0,   0,  -2,  -1,  -1,   1 },
     {    0,   2,  -1,  -1,   0,  -5,   2,   2,   0,   2,   1,  -7,   0,   5,   1,  -7,   0,   2,  -1, -11,  -3,  -1,  -4,   5,   1,   9,   7, -15,  -5,  20,  -4,   1,   6,  28,  24,  -9,   9, -12,  -1,   2,   9,  22,  22, -54,  24, -10,   3,  -1, -12,  32, -64,  49,  -3,  -8,   7,  -6,   5, -21,  18,  -1, -12,  12,  -5,   3 },
     {    0,   4,  -3,   1,   7,  10,  -1,   0,   0,   4,  -6,   5,   4,  -4,  -1,   4,   0,   4,  -7,   4,   5,  -3,  -2,  -2,  -1,   5, -16,   6,   8, -16,  14, -12,  -4,   1, -31,  -8, -10,  14,  -5,  11,   1,  32, -35,  42, -24,   4,   2,   0,   5,  82, -29, -13,   4,  -2,  -2,  -4,  -7, -28,  15,   5,   3,   2,  -1,   0 },
   },
   {
     {  -91,  35,  14,   0,   1,   1,   1,   0, -72,  27,  13,  -1,   1,   0,   1,   0,  -7,   2,   3,  -1,   0,   0,   0,   0,  17,  -7,  -2,   0,   0,   0,   0,   0,  13,  -5,  -2,   0,   0,   0,   0,   0,   5,  -2,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {   14,  72, -58,   4,  -6,   0,  -2,   0,  19,  61, -50,   2,  -5,  -1,  -1,  -1,   1,   5,  -2,  -1,   0,   0,   0,   0,  -7, -17,  16,  -1,   1,   0,   0,   0,  -5, -12,  11,   0,   1,   0,   0,   0,  -2,  -5,   4,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0 },
     {   73, -13,   0,  -8,  -2,  -2,  -1,  -1, -78,  32,  13,  -4,   2,   0,   1,   0, -53,  21,   9,  -2,   1,   0,   1,   0,  -8,   2,   2,   0,   0,   0,   0,   0,   4,  -3,  -1,   1,   0,   0,   0,   0,   5,  -3,  -1,   1,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {   15,  40,  54, -58,  -3,  -6,   1,  -2,  33,  33,  47, -53,  -4,  -5,   0,  -2,   9,   4,   5,  -6,  -1,   0,   0,   0,  -6,  -9, -13,  15,   1,   1,   0,   0,  -5,  -7, -10,  11,   1,   1,   0,   0,  -3,  -3,  -4,   4,   0,   0,   0,   0,  -1,  -1,  -2,   2,   0,   0,   0,   0,  -1,  -1,  -1,   1,   0,   0,   0,   0 },
     {  -36,   6,   2,   3,   1,   1,   0,   0,  35, -10,   2,  -2,  -1,  -1,   0,   0, -82,  30,  17,  -3,   2,   0,   1,   0, -67,  26,  14,  -3,   2,   0,   1,   0, -13,   5,   3,  -1,   1,   0,   0,   0,   9,  -4,  -2,   1,   0,   0,   0,   0,   7,  -3,  -2,   1,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0 },
     {   13,  25,  24,  45, -61,   9,  -6,  -1,   6,   3,  39,  43, -58,   9,  -5,  -1,  -2, -23,  26,  10, -12,   2,  -1,   0,  -6, -15,   3,  -8,  12,  -1,   0,   0,  -4,  -2,  -7,  -8,  12,  -2,   1,   0,  -1,   3,  -6,  -4,   6,  -1,   0,   0,   0,   1,  -3,  -2,   3,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0 },
     {   -4, -15,  24,  16, -20,   5,  -2,   0,  12,  36, -13,  15, -21,   3,  -2,  -1,  16,  76, -54,   0,  -2,  -3,   0,  -2,  12,  41, -36,  -8,   8,  -4,   1,  -1,   3,   1,  -4,  -6,   5,  -1,   1,   0,  -2, -11,   8,  -2,   1,   0,   0,   0,  -2,  -6,   5,   0,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0 },
     {    0,   4,  -2,   0,  -1,   0,   0,   0,  36, -10,  -3,   1,  -1,   0,   0,   0, -67,   6,   2,   6,   2,   1,   1,   1,  80, -24, -14,   3,  -2,   0,  -1,   0,  52, -14, -10,   1,  -1,   0,  -1,   0,   3,   0,  -1,   0,   0,   0,   0,   0,  -9,   3,   2,  -1,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0 },
     {   11,  21,  30,  37,  44, -47, -11,  -5,  11,  17,  20,  47,  45, -46, -12,  -4,  -1,  -2,  -6,  26,  15, -12,  -4,  -1,  -8,  -9, -11,   1,  -5,   8,   2,   0,  -5,  -5,  -5,  -9,  -9,  10,   3,   0,  -2,  -1,   0,  -8,  -5,   6,   1,   0,   0,   0,   0,  -3,  -2,   2,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0 },
     {   -1, -11,  -9,  23,  13, -11,  -3,   0,  12,  18,  25,  -9,   5, -11,  -3,  -1,  25,  39,  63, -53, -12,  -2,  -2,  -1,  28,  20,  40, -39, -15,   4,  -1,   0,   8,   1,   4,  -6,  -6,   3,   0,   0,  -1,  -5, -10,   8,   0,   0,   1,   0,  -3,  -3,  -6,   5,   1,   0,   0,   0,  -1,  -1,  -2,   1,   1,   0,   0,   0 },
     {   18,  70, -33,  -1,  -3,  -3,  -3,  -1, -15, -60,  28,  -2,  10,   0,   1,   1,   4,   8,  -5,  -8,   3,  -1,   1,   0,  11,  57, -32,  -7,  -1,  -2,   0,  -1,   3,  22, -13,  -3,  -1,  -1,   1,  -1,   0,  -1,   2,   1,  -1,   0,   0,   0,  -3,  -6,   4,   2,  -1,   0,   0,   0,  -1,  -1,   1,   1,   0,   0,   0,   0 },
     {   -6,   1,   4,  -3,   0,   0,   1,   0,  -2,   1,   1,  -1,   0,   0,   0,   0, -27,   3,   0,   1,   0,   1,   0,   0,  44,  -3,   2,  -4,  -2,  -1,  -1,   0, -90,  22,  20,  -4,   2,   0,   1,   0, -64,  15,  13,  -2,   2,   0,   1,   0,  -7,   1,   2,   0,   1,   0,   0,   0,   9,  -2,  -2,   1,   0,   0,   0,   0 },
     {    5,   8,  10,  10,  26,  61, -52,  -7,   5,   8,  10,   8,  31,  59, -53,  -8,   1,   2,   3,   1,  13,  15, -14,  -3,   0,  -2,  -1,  -1,  -2, -12,  10,   1,  -2,  -2,  -2,  -1,  -6, -13,  11,   2,   0,  -1,  -1,   0,  -5,  -7,   6,   1,   0,   0,   0,   0,  -2,  -3,   2,   0,   0,   0,   0,   0,  -1,  -1,   1,   0 },
     {    1,  -1,  10,  19, -14,  -4,   1,   2,   0,   2,  -8, -17,  14,  -4,   2,   2, -10, -26, -22, -65,  59,   4,   0,   3,  -9, -20, -23, -48,  46,   7,  -2,   3,  -3,  -1, -10,  -8,   9,   3,  -1,   0,   0,   4,   1,  10,  -8,   0,   0,  -1,   1,   2,   3,   7,  -7,  -1,   0,   0,   1,   0,   2,   3,  -3,   0,   0,   0 },
     {  -14, -33, -57,  27,  10,   5,   1,   3,  11,  41,  35, -22,  -3,  -8,  -1,  -1,  -2, -17,  23,   5,   0,  -4,   0,  -1,  -6,   6, -56,  31,   9,  -2,   0,  -1,  -8,  18, -38,  13,   8,  -3,   1,  -1,  -6,   2,  -1,  -2,   2,   0,   0,   0,  -2,  -1,   9,  -5,  -1,   1,   0,   0,   1,  -2,   3,  -1,  -1,   1,   0,   0 },
     {   -1,   0,   1,   1,   0,   1,   0,   0,  -4,   1,   3,  -1,   0,   0,   0,  -1,   3,   0,   0,  -1,   0,  -1,   0,   0, -33,   5,   6,   1,   0,   0,   0,   0,  60,  -6,   2,  -2,  -3,  -1,  -1,   0, -99,  13,  16,   0,   2,   0,   2,   0, -33,   5,   6,   1,   1,   0,   1,   0,   7,  -1,  -1,   1,   0,   0,   0,   0 },
     {    2,  -2,  27, -19,   2,  -3,  -1,  -1,   6,  38, -34,   7,  -3,   0,  -2,   0, -10, -52,  13,  10,  -3,   4,   0,   1,  10,  53,   4, -22,  -6,  -3,   1,  -2,  20,  67, -12, -25,  -3,  -4,   1,  -2,   7,  15,  -3,  -8,   1,  -1,   0,   0,   0,  -6,   0,   2,   1,   0,   0,   0,  -1,  -5,   1,   2,   0,   0,   0,   0 },
     {   -4,  -6,  -7, -13, -16, -25, -53,  60,  -5,  -7,  -8, -11, -16, -27, -51,  60,  -1,   0,  -3,  -2,  -5, -11, -12,  16,   1,   0,   2,   3,   2,   2,  10, -11,   1,   1,   1,   4,   2,   5,  12, -13,   0,   1,   0,   2,   2,   4,   6,  -7,   0,   0,   0,   1,   1,   2,   2,  -3,   1,   0,   0,   0,   1,   1,   1,  -1 },
     {   -2,  -4,  -5, -19, -18,  17,   7,   0,   1,   0,   3,   9,  11, -12,   1,  -1,  11,  16,  27,  31,  58, -59, -11,   0,  12,  13,  26,  16,  46, -44, -13,   1,   4,   4,   8,   2,   9,  -6,  -5,   1,   0,   0,  -4,  -3,  -8,   9,   0,   0,  -1,  -1,  -4,  -1,  -5,   6,   1,  -1,  -1,  -2,  -1,  -1,  -2,   2,   1,   0 },
     {   -2, -11,  -6,   3,  -1,   1,  -1,   1,   5,  15,  19,  -4,  -6,  -2,  -2,   0,  -9, -47,  -7,   9,   7,   2,   1,   4,  10,  46,  -3, -11,  -5,  -3,  -2,  -1,  -2, -47,  48,   6, -12,   5,  -4,   2,  -3, -65,  31,  17,  -7,   7,  -2,   3,   2, -14,   1,   7,   0,   1,  -1,   0,   3,   3,  -5,   0,   2,  -1,   0,  -1 },
     {   -3,  -6, -32,   6,   4,   2,   0,   1,   8,  11,  59, -25,  -8,  -2,  -2,  -2,  -3,   2, -54,  22,   3,   1,  -2,  -1,  -1, -18,  31, -20,   0,   3,  -4,   1,  16,  33,  40, -41,  -7,   0,  -2,   0,  13,  23,  -1, -11,  -4,   0,   0,  -1,   2,   2,  -7,   2,   0,   1,   0,   0,   0,  -3,  -3,   4,   1,   0,   0,   0 },
     {    3,   4,  19,  50, -57,  16,  -5,  -1,  -1,   1, -11, -40,  39, -14,   7,  -2,  -1,  -5,  -7,  -9,  19, -10,   6,  -3,   3,   8,  15,  42, -47,   8,   3,  -4,   3,  15,  10,  13, -26,   4,   3,  -3,   4,   6,   4,  -5,  -2,   1,  -1,   0,   0,   0,  -1,  -5,   6,   1,  -1,   1,  -1,   0,   0,  -1,   1,   0,  -1,   0 },
     {    3,   1,  -1,   0,  -1,   0,   1,   0,   4,  -2,  -1,   0,   1,   1,   0,  -1,   2,   1,   1,   0,   1,  -1,   0,  -2,   4,  -1,  -4,   2,  -1,   0,   0,  -2,  14,  -2,  -4,  -3,   0,   1,   0,  -1, -19,   2,  -6,   0,   2,   1,   1,   0,  91, -10, -15,   0,  -1,   0,  -1,   0,  83, -10,  -9,   0,  -2,   0,  -1,   0 },
     {    7,  10,  11,  14,  18,  29,  43,  62,   8,  11,  13,  15,  24,  29,  43,  62,   2,   3,   3,   7,   4,  11,  13,  16,  -1,   0,  -1,  -3,  -4,  -5,  -5, -11,  -1,  -2,  -1,  -4,  -3,  -7, -10, -13,   0,  -1,  -3,  -1,  -2,  -6,  -6,  -7,  -1,  -1,   0,  -1,   0,  -3,  -2,  -3,   0,  -2,   0,  -1,   0,  -1,  -1,  -1 },
     {   -2,  -3,  -4,  -4, -17, -19,  11,   6,   0,   1,   0,  -1,  12,  10, -10,  -3,   7,  12,  14,  25,  30,  65, -43, -21,   7,  15,  14,  18,  24,  56, -35, -20,   2,   1,   2,   2,  15,  13,  -7,  -7,  -1,  -2,  -2,  -1,   1,  -8,   6,   2,  -2,  -2,   0,  -2,  -4,  -6,   5,   3,   0,  -1,   0,  -2,  -3,  -3,   2,   1 },
     {    1,   0,   6,  -9,  -2,  10,  -1,   0,  -1,  -6,  -9,  -5,   0,   4,   0,   3,   6,  11,  39,  -8,  -5,  -8,  -1,  -3,  -6, -13, -35,   4,   4,  14,  -1,  -1,   3,  -5,  27, -47, -10,  24,  -1,  -1,  15,  10,  64, -55, -14,  15,  -3,  -1,   5,   2,  25, -14,  -8,   3,  -1,   0,   3,  -3,   1,   3,  -2,  -1,   0,   0 },
     {    8,  12,  24,  57,  40, -23, -19,  -3,  -7, -12, -24, -64, -25,  21,  16,   5,   0,  -1,   6,  23, -13,  -5,  -1,   0,   5,   9,  10,  12,  33, -18, -12,  -2,  -1,  -7,   1, -19,  21,  -5,  -4,   1,  -2,  -8,  -6,  -3,   4,   1,   0,   2,  -1,  -2,  -3,   0,  -4,   1,   2,   0,  -1,  -1,   0,   1,  -2,   1,   1,   0 },
     {   -3,  -5,  -7,  -8, -21,  24,   1,   2,  -2,  -4,  -7, -21,  25,  -8,   2,  -1,   1,   0,   3,  38,  -3, -17,  -4,  -3,  -3,  -2,  -3, -38, -21,  18,   7,   3, -12, -29, -42, -55,  18,  30,   8,   7, -13, -24, -31,  -4,  24,  13,   1,   3,  -3,  -5,  -2,  11,   8,  -3,  -2,   1,  -5,   4,   4,   6,  -1,  -3,  -1,  -1 },
     {    1,   1,   3,   2,  -5,  -2,  -1,  -1,  -1,  -9,   0,   0,   8,   0,   2,   2,   2,  13,   2,  10,  -9,   1,  -1,  -6,  -4, -42,  14,  -4,   3,  -1,   5,  -3,   6,  53, -20,   6,   6,  -6,   1,  -2,  -3, -56,  37,  25,  -9,   3,  -4,   1,  -1, -61,  28,  20,  -7,   5,  -3,   1,   2,  -9,   5,   6,  -2,   1,   0,   0 },
     {    0,   0,   1,   0,  -1,   1,   1,  -2,   0,  -1,  -2,  -1,   2,  -3,  -2,   1,   1,   0,   0,   1,   0,   1,  -4,   1,  -2,  -2,  -2,  -3,  -3,  -3,   1,   1,  -6,   1,   2,  -1,   5,  -4,  -4,   1,  29,  -6,  -5,  -2,   7,  -1,  -4,   0, -79,   6,   2,  -2,   5,   2,   1,   1,  94,  -5,  -5,   0,  -2,   0,  -1,  -1 },
     {   -1,  -2,  -2,  -6,   5,   0, -23,  15,   1,   1,   3,   3,  -6,   4,  15, -15,   4,   3,   6,   9,  19,   5,  70, -51,   5,   5,   7,   3,  19,  19,  55, -47,   1,  -1,   6,   5,   4,  14,  11, -14,   0,   2,  -5,   0,   0,   2, -11,   5,  -5,   0,  -4,  -1,  -2,  -2,  -9,   8,   2,  -2,  -1,  -1,  -2,   0,  -3,   4 },
     {   -1,  -2,  -1,  -4, -12,  -8,   4,   0,   2,   2,   7,   6,  36, -12, -13,   1,  -2,  -6,  -6, -23, -14,  22,  10,  -3,   3,   3,  10,  18,   5, -23,  14,   2,   5,  -3,  14, -13,  69, -40, -13,   7,  -1, -13,   8, -35,  50,  -1, -18,   6,   7,   0,   4, -10,  11,   8, -10,   2,  -7,   4,   1,   2,  -4,   2,   0,  -3 },
   }
 },
 {  //19
   {
     { -121,  15,   6,   4,   3,   2,   2,   1,  32,  -4,  -1,  -1,  -1,  -1,   0,   0,  19,  -1,  -1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   17,  -5,   0,   0,   0,   0,   0,   0, 105,  -4,  -6,  -3,  -3,  -2,  -2,  -1, -68,   7,   3,   2,   2,   1,   1,   1, -16,  -2,   1,   0,   0,   0,   0,   0,   9,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -30,   5,   2,   1,   1,   1,   0,   0, -55,   8,   2,   1,   1,   1,   1,   0, -98,   3,   5,   2,   3,   1,   1,   1,  42,  -8,  -2,  -1,  -1,  -1,   0,   0,  29,   0,  -1,   0,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    5,  -3,   0,   0,   0,   0,   0,   0,  26,  -1,  -2,  -1,  -1,  -1,   0,   0,   7,  -7,   0,   0,   0,   0,   0,   0, 103,   0,  -7,  -2,  -3,  -2,  -2,  -1, -69,  10,   4,   2,   2,   1,   1,   0,   4,  -4,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 },
     {  -12,   5,   1,   1,   1,   0,   0,   0, -22,   7,   1,   1,   0,   0,   0,   0, -38,   6,   2,   1,   1,   1,   0,   0, -59,  10,   3,   2,   1,   1,   1,   0, -99,   1,   7,   2,   3,   1,   1,   0,  18, -11,  -1,  -1,   0,   0,   0,   0,  17,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    2,  -6,   0,   0,   0,   0,   0,   0,   6,  -2,   0,  -1,   0,   0,   0,   0,   4,  -6,   0,   0,   0,   0,   0,   0,   6,  -2,  -1,   0,   0,   0,   0,   0,  18, -15,   0,  -1,   0,   0,   0,   0, 123,  -9,  -9,  -2,  -3,  -1,  -2,  -1,  12,  11,  -1,   0,  -1,   0,   0,   0,  -6,   2,   1,   0,   0,   0,   0,   0 },
     {    2,   0,   0,  -1,   0,   0,   0,   0,   3,  -4,   1,   0,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0,  12,   4,  -1,   0,  -1,   0,   0,   0, -16, -12,   2,   0,   1,   0,   0,   0, 118,  -6, -10,  -2,  -3,  -1,  -2,   0, -38,  12,   3,   1,   1,   0,   1,   0 },
     {   16, 121,  -4,   4,  -4,  -2,  -4,  -1,   3, -22,   4,  -1,   0,   0,   1,   0,  -2, -28,  -2,  -1,   1,   0,   1,   0,   1,   9,  -2,   0,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 },
     {   -1,   0,   1,   0,   1,   0,   0,   0,  -2,   2,   0,   0,  -1,   0,   0,   0,  -1,   6,   0,   0,   0,   0,   0,   0,  -4,   3,   1,   1,   0,   0,   0,   0,  -2,   4,   0,   0,   0,   0,   0,   0,  -2,   5,   0,   0,   0,   0,   0,   0, -38,  16,   2,   1,   1,   0,   0,   0,-120,   5,  11,   1,   3,   1,   1,   0 },
     {   -1,  -6,   4,   2,   0,   1,   1,   0,   7,  87,  -4,  -3,  -2,  -4,  -3,  -1,  -1, -86,   5,   3,   3,   3,   2,   2,  -4,  31,  -8,  -3,  -1,  -1,  -1,  -1,   6,  -2,   6,   1,   0,   0,   0,   0,  -5,  -6,  -2,   0,   0,   0,   0,   0,   3,   4,   0,   0,   0,   0,   0,   0,  -3,  -3,   0,   0,   0,   0,   0,   0 },
     {   -4, -35,   1,   1,   2,   1,   1,   1,  -6, -84,   8,   0,   4,   4,   3,   2, -11, -72,   1,   1,   2,   3,   2,   2,  -7,  33,  -8,   1,  -2,   0,  -1,   0,   1,  36,   0,  -1,  -2,  -1,  -1,  -1,   0,   7,   2,   0,   0,  -1,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0,   0 },
     {   -7,  -2, -96, -38, -20,  -4,   0,   3,   4,   6,  55,  17,   8,   0,  -1,  -2,   1,   3,  26,  15,   7,   1,   0,  -1,  -2,  -2, -26, -10,  -4,   0,   1,   1,   0,   3,   4,   1,   0,   0,   0,  -1,   0,  -2,   0,   0,   1,   0,   0,   0,  -1,   0,   2,   1,   0,   1,   0,   0,  -1,   0,  -1,   1,   0,   0,   0,   0 },
     {    1,   5,   3,   1,   2,   2,  -1,   0,   1,  19,   1,   1,  -3,  -1,  -1,   0,   2, -19,   1,  -1,  -1,   0,   0,   0, -10, -82,  -1,   4,   5,   5,   4,   2,  -5,  76, -15,  -3,  -4,  -3,  -3,  -2,  15,  33,  11,  -2,   0,  -2,  -1,  -1,  -2, -35,   6,   2,   2,   2,   1,   1,  -3,   0,  -4,   0,   0,   1,   0,   0 },
     {    1,   8,  -2,  -2,  -2,   0,  -1,  -1,   0,  21,  -8,  -2,   1,   0,   0,  -1,   1,  47, -10,  -3,  -2,   1,  -2,   0,   3,  78, -12,  -2,  -3,  -2,  -1,  -1,  10,  75,  -3,  -6,  -4,  -2,  -2,  -2,  13,  33,   5,  -3,   0,  -3,  -1,  -1,   9,   6,   3,  -1,   0,  -1,   0,   0,   6,   1,   0,   0,   0,   0,   0,   0 },
     {    1,   3,  13,   7,  -1,  -2,  -2,  -2,  -4,  -7, -48, -26,  -7,  -2,   3,   4,   6,   8,  81,  41,  11,   1,  -4,  -6,  -4,  -7, -58, -21,  -5,   2,   4,   4,   2,   6,  20,   2,  -2,  -4,  -3,  -2,   0,  -7,  -4,   3,   3,   3,   2,   1,  -1,   5,  -2,  -3,  -1,  -1,  -1,   0,   1,   1,   2,   2,   0,   0,   0,   0 },
     {    2,  -1, -30,  52,  55,  10,  -5,  -5,  -2,   1,  32, -56, -58,  -9,   6,   6,   1,   4, -18,  20,  23,   2,  -3,  -2,   1,   0,   5,   1,   2,   1,   0,  -2,   1,   4,   5, -10, -13,  -3,   1,   1,   0,  -7,  -8,   7,  11,   4,  -1,  -1,   0,   2,   4,  -3,  -4,  -1,  -1,  -1,   1,  -3,  -1,   2,   3,   0,   0,   0 },
     {    4,   0,  67,  15,  -7,  -6,  -7,  -4,   5,   5,  93,  17,  -2,  -7,  -8,  -4,   2,   9,  31,   8,  -2,  -4,  -2,  -3,   0,   7, -22,  -5,  -2,   2,   2,   1,   0,   3, -19,  -1,   3,   1,   2,   1,   0,  -4, -13,  -3,   1,   1,   2,   0,  -2,  -4,  -7,  -2,   0,   1,   0,   0,   0,  -1,  -3,  -1,   0,   1,   0,   0 },
     {    0,  -4,  -3,   5,   4,  -1,   1,   1,   0,  -2,   0,  -9,  -1,  -2,   1,  -1,  -1,  -3,  11,  11,   1,   1,  -1,  -1,   4,  18,  -2,  -4,   0,   0,  -1,   1,  -9, -52, -10,   3,   1,   2,   1,   1,   5,  92,  -2,  -6,  -3,  -3,   0,  -2,   8, -60,  11,   4,   3,   2,   1,   0,  -8,  16,  -7,  -1,   0,   0,   0,   0 },
     {    0,  -1,   1,  -2,   0,  -2,   0,   1,   0,   0,  -3,  -1,  -1,  -1,  -1,  -1,   1,  10,   2,   0,  -1,   3,  -1,   0,   3,  22,   4,   0,  -1,  -1,  -1,   0,   8,   3,  15,   2,   1,  -1,   0,  -2,   7, -57,  27,  -3,   2,   0,   2,   0,  -9, -95,  11,   5,   3,   3,   3,   2, -14, -42,  -9,   2,   0,   2,   1,   1 },
     {    0,   4,  22, -70,  42, -15,  17,  -2,   0,   0,   7, -10,  -3,  -3,   2,   3,  -1,  -2, -29,  67, -37,  19, -15,   3,   0,  -2,  -3,   7,   6,   1,   2,  -1,   1,   0,  17, -21,  13,  -5,   6,   0,  -1,   2,   1,  -1,  -2,   0,   0,   0,  -1,   2,  -1,   6,  -2,   1,  -1,   0,   1,   3,  -1,   2,   0,  -2,   0,  -1 },
     {    1,  -2,   8,  18,   1,  -4,  -2,  -5,   1,   0,   6,  11,  12,  -2,   0,  -3,  -2,   2, -64, -15,   8,  12,   6,   9,  -5, -13, -81, -31, -14,   5,   6,   4,   0, -13,  38,   9,  -1,  -7,  -6,  -5,   1,   7,  32,  11,   2,  -2,  -3,   0,   0,   2,  -1,   2,   1,   2,  -2,   1,  -1,   2,  -6,  -1,   0,   0,   0,   1 },
     {    0,   3,   5, -32,  52, -28,  15, -11,   0,  -1, -11,  54, -58,  32, -18,  10,   0,  -1,  16, -38,  29, -19,   9,  -4,  -1,   3, -15,   4,  -6,   1,  -1,   1,   0,  -4,  -6,   1,  -8,   2,  -1,  -3,   0,   3,   3,  -5,   9,  -3,   3,  -1,   1,  -1,   3,   2,  -2,   4,   0,   1,   0,   0,   0,  -1,   2,  -1,   1,   1 },
     {    0,  -1,  -2,   3,   2,   0,   2,   1,   0,   3,  -2,   2,   0,  -3,   1,  -2,   0,   1,   0,  -1,   2,   2,  -1,  -2,   1,   8,   3,   1,   0,  -1,  -3,   1,   2,   7,  -2,   0,  -3,   1,   0,   1,   2, -31,  11,   1,  -1,   3,   0,  -1, -17, -26, -24,   6,  -3,   1,   0,   3,   3, 114, -20,  -7,  -2,  -5,  -1,  -1 },
     {    0,  -2, -10,  38, -35,  25, -22,  12,  -1,  -1, -15,  54, -47,  28, -26,  14,   0,  -1, -16,  50, -29,  16,  -7,   4,   0,   1,  -6,   9,  -5,   0,  -1,  -1,   0,  -3, -15, -17,  14,  -2,  10,  -3,   0,  -2,   3, -12,   7,  -7,   7,  -5,   0,   0,  -1,  -7,   4,  -4,   3,  -1,   0,   0,   1,  -6,   2,  -2,   2,  -1 },
     {    0,   0,   0,   4,  -3,  -3,  -5,  -6,   0,   0,   2,  13,  -5,  11,  -5,   5,  -1,   1, -11,  -6,  -5,  -4, -11,  -5,   0,  -4,  11, -16,   4,  -1,   2,   3,   5,  15,  67,  32,  17,   3,  -1,  -6,  -3,   4, -86, -10,  -4,   7,   0,   6,  -2, -22,  -7, -19,  -5,  -3,   3,  -3,   0,   7,  17,   2,   1,  -2,  -3,  -1 },
     {    3,   2, -11,  17,  40,  57,  39,  23,   1,   0,   2,  11,  14,  -7,  -5,  -6,  -2,  -1,  14, -22, -52, -42, -34, -20,   0,   3, -21,   6,  13,  19,  12,   9,   1,  -2,   2,   5,  15,   6,   9,   0,  -1,  -2,   6,  -8,  -8,  -6,  -5,   1,   0,   2,   3,   0,  -1,  -2,  -4,   1,   0,  -1,  -4,   3,   1,   1,   0,   0 },
     {    2,   3,   5, -14,   0,  37,  38,  31,  -3,  -5,  -3,   7, -17, -56, -49, -16,   2,   3,   3,  -5,  11,  46,  28,  11,   0,   0,   2,   6, -11,  -6,  -3,  -2,  -1,   0,  -1,  18, -17, -18, -27, -12,   0,   2, -10,  -3,  -2,  19,  10,   7,   0,  -1,  -2,  -2,   2,  -9,  -4,  -4,   1,  -1,   8,  -8,   2,   1,   1,   1 },
     {    1,   1,  -2,   1,  18,   4,   9,  -7,   0,  -1,  -6,  11,   6,   6,  -3,  -5,   0,  -2,  -8,  24,  -1,  -5,  -9,   5,  -1,   3,  26, -95,   3, -14,  17,  -8,  -2,   3, -40,  40, -10,  10,  -4,  -4,   1,  -3,   9,  -5,   2,   5,   2,   8,  -2,   2, -23, -10,  -3,   7,   1,  -2,  -1,  -8,  -7,   8,   0,  -6,   2,   0 },
     {   -1,  -1,  -7,  -6,  12, -10, -15,   3,  -1,   0, -18,   2,  15, -18,   6,  -9,  -1,   1, -25,  12,   2,  -1,  19, -14,  -1,   4, -39,  22,   9,  36,   5,  15,  -1,   0, -56,  23, -16,  33,   5,  20,  -3, -16, -51,   6,   6,  -7,  28, -14,  -2, -12,   0,  -3,   0,  -4,   9,  -3,  -2,  -5,  -6,  -6,   1,  -3,   0,   1 },
     {    3,   3,  -2,  24,  42,  17,   7,   3,   4,   4,   1,  33,  60,  39,  14, -12,   3,   3,   6,  24,  34,  31,   0,   1,   1,  -1,   6,  16,  -8,  -4, -11,  -6,  -1,   0,   6, -15, -37, -23, -11,   2,  -1,  -3,  -9, -26,  -9, -18,  -5,   3,  -1,  -1,   5,  -8, -11,  -5,   5,  -3,   0,  -1,   3,   0,  -4,  -5,   0,   0 },
     {    0,  -1, -14,  31,  28, -58, -20,  10,   0,   0,   0,  17,   1, -19,   6,  -4,   0,   0,  19, -10, -50,  38,  20, -22,   1,   0,   6,  -1,  23,   6, -20,   2,   1,   2,   9,  18,   6,  -2, -13,   2,   0,   4,  20, -11, -36,  15,  -3, -14,   1,   9,  -5, -17,   5,   5,  -5,  -3,   1,  -4,  -6,  -2,   6,  -6,   0,  11 },
     {    0,   0,   8, -23,  -3,  39, -33, -11,   0,  -1,  -1,  21, -22, -24,  52,   1,   0,   0,  -1,  -1,  -1,  -7,  -1,   8,   0,   2, -10,  -3,  51,  -5, -36,  -9,  -1,  -3,   2,  21, -35, -26,  19,   8,   1,   0,   6,  -6,   5,  19, -18,   4,   0,  -1,   9,  -5,  -6,  -8, -11,  -5,   0,   1,  12,  -9, -22,   2,  -2,  -1 },
   },
   {
     {   69, -65,  28, -11,   5,  -3,   2,  -1, -54,  52, -18,   4,  -1,   0,   0,   0,  17, -12,  -2,   5,  -3,   1,  -1,   1,  -5,   2,   4,  -3,   1,   0,   0,   0,   2,   0,  -2,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -66,   7,  34, -24,  14,  -5,   3,  -1,   0,  33, -50,  34, -16,   5,  -3,   1,  28, -41,  36, -17,   2,   2,  -2,   1, -14,  20, -12,  -2,   7,  -5,   3,  -1,   7,  -7,   1,   5,  -5,   2,  -1,   0,  -3,   2,   1,  -3,   2,  -1,   0,   0,   2,  -2,   0,   1,  -1,   0,   0,   0,  -1,   1,   0,  -1,   1,   0,   0,   0 },
     {    1,  34, -53,  38, -20,   9,  -5,   2, -48,  10,  28, -20,   8,  -1,   1,   0,  58, -43,   8,  -4,   5,  -4,   2,  -1, -29,  21,  -2,   0,  -1,   0,   1,   0,  12,  -6,  -3,   2,   0,   0,   0,   0,  -4,   1,   2,   0,  -1,   0,   0,   0,   2,  -1,   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   72,  15, -15, -17,  10,  -8,   2,  -2,  38, -36,  -7,  27, -19,  10,  -4,   2,  -2, -17,  38, -33,  17,  -5,   1,   0, -21,  32, -30,  17,  -2,  -4,   3,  -2,   7, -14,  11,  -1,  -6,   6,  -3,   1,  -3,   5,  -3,  -3,   5,  -3,   1,   0,   0,  -2,   1,   1,  -2,   1,   0,   0,  -1,   1,   0,  -1,   1,   0,   0,   0 },
     {    6,  43, -22,  -6,   4,  -3,  -1,   0, -66,   5, -11,  32, -19,  11,  -4,   2, -32,   5,  23, -26,  15,  -4,   1,   0,  58, -30,  -2,   5,   0,  -4,   3,  -2, -26,  16,   3,  -3,  -1,   2,   0,   0,  11,  -4,  -4,   2,   1,  -1,   0,   0,  -3,   0,   3,  -1,  -1,   1,   0,   0,   2,   0,  -1,  -1,   1,   0,   0,   0 },
     {   11, -14, -14,  44, -42,  26, -12,   5,  47,  38, -30, -21,  24, -16,   4,  -1, -25, -36,  39,  -7,   1,  -1,   4,  -2,  31,  -6,  -6,  -3,   1,   2,  -2,   1, -24,  14,  -3,   4,   1,  -3,   2,  -1,   7,  -5,   1,   0,  -2,   3,  -1,   0,  -4,   1,   0,   0,   2,  -1,   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0 },
     {  -27, -20,   6,  16, -11,   8,  -3,   1, -32,  14,  12, -16,  10,  -4,   1,   0, -61,  15,   8, -13,  18, -11,   7,  -3, -16,  30, -39,  36, -24,  10,  -3,   0,  27, -34,  30, -19,   4,   5,  -5,   4,  -9,  12,  -9,   1,   6,  -7,   4,  -2,   2,  -2,   1,   2,  -4,   2,  -1,   0,   1,   0,   0,  -1,   1,   0,   0,   0 },
     {  -14, -36,   8,   5,   6,  -3,   3,   0,  16, -12,  22,  -8,  -7,   6,  -3,   1,  68,   4,  -9, -21,  22, -15,   5,  -3,  59,  -3, -28,  33, -26,   8,  -2,  -1, -23,   2,  21, -18,   6,   5,  -5,   3,   5,   3, -10,   5,   3,  -6,   3,  -1,   0,   1,   1,   0,  -3,   3,  -1,   0,   0,   1,  -1,   0,   2,  -1,   0,   0 },
     {   -7,  -3,   7,  -4,   1,   0,   0,   0, -12,  -4,   5,  -4,   5,  -2,   2,   0,  -5,   8,   0,   0,  -1,   0,   0,  -1, -54,  21,  -8,   8,  -2,   1,   0,   1,-101,  10,  12,  -2,   3,   0,   1,   0,  42, -15,  -4,   1,  -2,   1,  -1,   1,  -5,   2,   3,  -1,   0,   0,   0,   0,   3,  -1,  -1,   1,   0,   0,   0,   0 },
     {  -10, -20, -35,  22,  20, -17,   9,  -2,  27,  43,   8,   5, -48,  32, -15,   3, -13,   4, -10, -19,  40, -22,   5,   1, -38, -33,  30,   8, -14,   3,   7,  -4,   2,  17, -12,  -7,   4,   5,  -6,   2,  -4,  -5,   2,   5,   0,  -5,   4,  -1,   4,   1,   0,  -4,   1,   3,  -2,   0,   0,   0,  -1,   1,   1,  -2,   1,   0 },
     {    0,  -1,  21,   3, -36,  29, -13,   4, -12, -24, -24,  15,  24, -18,   8,   0,  26,  41,  19, -39,   8,  -8,   5,  -5, -41, -45,   2,  27,  -8,   7,  -2,   2,  22,  35, -14,  -8,  -1,  -1,  -1,   0,  -3, -12,   7,   2,   1,   1,   0,   0,   0,   2,  -1,  -1,  -1,  -2,   2,  -1,   2,   2,  -1,   0,   0,   0,   0,   0 },
     {   14,  30, -15,   5, -17,  12,  -8,   3,  15,  31, -37,  19,  -4,  -3,   0,   0,  24,  34, -22,   9, -13,   7,  -4,   1,  -8, -22,  12,  -2,   3,  -1,  -1,   2, -16, -54,  49, -26,  13,  -4,   4,  -1,  -8,  33, -24,   6,   3,  -4,   2,  -2,   6,  -5,   0,   5,  -6,   4,  -1,   1,  -2,  -1,   2,  -2,   1,   0,   0,   0 },
     {    0,  -3,  -2,   1,   0,   1,  -3,   1,   4,   1,   0,   2,  -3,   0,   1,   0,   2,  -4,  -1,  -1,   0,   2,  -1,   0,   2, -10,   3,  -2,   1,  -2,   0,   1,  44, -21,   2,  -2,  -1,   0,   0,   0, 116,  -4, -15,   3,  -5,   0,  -2,   0,   0,  11,   2,  -2,   1,   1,   0,   0,  -3,   2,   0,   1,   0,   0,   0,   0 },
     {  -11, -26, -22,   7,   2,  14,  -7,   3,   2,  12,  22,   9,  -8, -15,   8,  -3,   4,  26,  -3, -27,  -3,  24, -12,   3,   1,  -7, -16,   2,  42, -43,  16,  -3,  -6, -14,   7,  32, -56,  29,   2,  -7,  -1,   5,   6, -27,  24,   4, -17,  10,   2,   1,  -4,   7,   1, -12,  11,  -3,  -2,   0,   1,  -1,  -2,   5,  -2,  -1 },
     {   -1,  -1,  -2,   0,  -1,   2,  -1,   0,   1,   3,  -2,   2,  -3,   1,   0,  -1,   3,   2,  -5,   0,   2,   0,   0,  -1,  -2,   5,   1,  -1,   0,  -2,   0,   1,   7,   4,  -1,   1,  -1,   3,  -1,   0,   0,  17,  -9,   1,   1,  -2,   1,   0,-115, -11,  23,  -7,   6,  -1,   2,  -1,  42,  -2, -12,   3,  -2,   0,  -1,   0 },
     {    0,   3, -17, -35,  17,  28, -19,   5,  11,  22,  28,  23,  -6, -47,  26,  -7, -10, -32, -23,  -8,  -5,  40, -13,  -1,  -1,   4,  10,  31, -12, -12,  -7,  10,   6,  23,   1, -37,  12,  -1,  11,  -7,  -5, -17, -11,  22,   1,  -3,  -4,   1,   1,   4,   7,  -8,  -8,   5,   0,   0,  -3,  -3,  -2,   0,   5,  -1,  -2,   0 },
     {   10,  32, -12,  -1,  -3,  -1,   0,  -1,  12,  42, -23,   6,  -3,   2,  -4,   0,  11,  39, -32,  15,  -5,  -2,   1,  -1,  13,  39, -32,  20, -19,  11,  -5,  -1,  10,  11, -14,   9,  -3,   3,  -5,   4,   9, -52,  45, -22,   9,  -3,   3,   0,  -5,   1,  -3,  -4,   6,  -3,   3,  -2,   0,   0,  -2,   3,  -3,   1,   0,   0 },
     {    0,  -2,  14,  -9, -13,  47, -44,  23,  -2, -10,   5,  24, -18,  -9,  17,  -8,  -3,  -7, -28,  -6,  42, -36,  26, -13,   9,  26,  31, -26, -11,  14, -10,   3, -11, -28, -15,  28,  -7,   4,  -3,   3,   0,   6,   7,  -8,   0,  -2,   0,   0,   2,   2,  -2,  -2,   3,  -1,   1,   0,  -5,  -1,   0,   2,  -1,  -2,   1,  -1 },
     {   16,  51,  46, -38,  -2, -11,   4,  -6,  12,  39,  36, -47,   8,  -8,   3,  -3,   0, -16,   2, -12,  12, -12,  13,  -5,  -6, -38,  -6,  17,  -9,   5,   0,   3,  -5, -12,  -1,  18, -19,  12,  -3,   2,   1,  16,   1, -11,   8,  -3,  -2,   1,  -3,   1,  -7,   4,   3,  -5,   2,   1,   1,  -5,   1,   4,  -2,   2,  -1,   1 },
     {   -1,   2,   1,   2,  -3,  -3,   3,  -2,  -1,   2,  -2,  -1,  -2,   0,  -1,   2,   0,   6,   1,  -2,  -3,   0,  -1,   1,  -1,   6,   2,   4,  -3,   2,   1,  -2,  -1,   7,  -4,   1,  -2,   1,   1,   0,  -3,  12,  -6,   2,  -1,   0,  -1,   0, -42,  13,   5,  -4,   6,  -1,   1,  -1,-116,  -4,  21,  -4,   5,  -1,   3,  -2 },
     {   -1,  -4,   4,   0,  -3,   2,   2,  -2,  -2,  -8,   4,  -1,  -5,   3,  -3,   3,  -1, -10,   6,  -5,   2,   0,   1,   2,  -1,  -9,  10,  -7,   5,  -5,   1,   0,   1, -20,  23, -15,  11,  -6,   3,  -3,  -9, -35,  22,  -3,   3,   0,   1,   2, -26,  81, -54,  22, -10,  -2,   2,  -2,  12, -34,  21,  -2,  -4,   5,  -4,   2 },
     {    5,  11,  18,   9, -13,  -2,   0,   2,  -4, -17, -24, -12,   7,   3,   6,  -6,  -4, -18, -17,  19,  18,  -3, -24,  17,   2,   6,  21,   3, -16, -27,  45, -21,   0,   8,   2, -24,  -5,  51, -42,   7,   2,  -5,   2,   5,  23, -31,  -1,  22,   2,  -2,   3,   2, -16,   7,  18, -21,  -1,   5,  -1,   0,   3,   3, -10,   4 },
     {   -3,  -5,  -7, -20, -32,  13,  41, -23,   4,   2,  11,  33,  29,  -2, -57,  28,  -7, -21, -29,  -4, -10, -16,  46, -21,   1,   4,  10,   2,  16,  -8,  -4,  -7,   2,   7,  12,  -8, -18,  13,  -8,  14,   0, -12, -11,   4,   8,  -1,   2,  -6,   2,  -2,  10,   0,  -7,  -1,   1,   1,  -1,  -3,  -2,  -1,   0,   2,   0,  -3 },
     {   -2,  -6,  -3,   5,  -9, -12,  25, -10,  -1,  -5,  -1,   7,  11,  -5,  -9,  -3,   0,   4,   9, -27, -14,  34,   4,  -5,   9,  13,  47,  30, -39, -18,  -4,  12, -13, -41, -37,   8,  10,  14,   7,  -7,   5,  20,  33,   5, -17, -11,  15,   2,  -2, -15, -20,   1,  20, -16,   4,   5,   5,   6,  11,  -5,  -7,   5,  -4,   3 },
     {   -5, -17, -29,  -4,  23,  -4,  -9,   7,  -3, -19, -13,  16,  10,  -7,   6,  -5,  -5, -32,   1,  29, -20,   4,   5,  -4,  -9, -41,  -4,  16, -20,  45, -25,  10,  -6,  -6,  13,   8, -15,  12, -13,  16,   3,  10,  18, -45,  33, -22,  19,  -8,  -5,   1,   0,   4,   1,  -9,   9,  -7,  -6,   9,  -5,  -2,  -7,   5,  -3,   2 },
     {   -4,  -5, -41, -29,  41, -22,   7,  14,   1,  11, -20, -15,  52, -34,  10,  -2,   6,  25,  20, -13,   8, -32,  13,   0,   4,  17,  24, -22, -20,  10,   2,  -8,   4,   5,  11,   4, -19,  14,  -6, -14,  -2,  -5, -21,  16,  -1,   4,  -9,  12,   0,   7,  -7,   2,   9,   4,   1,  -3,   5,  -4,   7,  -6,   3,   1,   1,   1 },
     {    0,  -4,   8,   1,  11,  -2,  -9,   6,   0,  -5,  12, -10,  -1,  -3,  -2,   7,   3,  -1,  26,  -5,  -4,   1,   5, -13,   1, -19,  17, -13,   7,  -4,   5,   1,  -1, -45,  25, -15,   1,   4,  -8,   3, -11, -70,  23,  11,  -8,   4,  -6,  -1, -12, -23,  17, -20,   8,   0,   1,   5, -18,  41, -37,  16,  -3,  -4,   3,  -2 },
     {    8,  14,  22,  45,  30, -15,   2,  -7,  -7, -19, -40, -21,  -8, -13,   5,  13,  -3, -16, -31,  -5,  20,   8,  16,  -9,   2,   6,  15,  26,  18,   9, -29,   7,  -1,  -2,  -1, -21, -44,   9,  27, -16,  -1,  -7,  -3,  14,  11,  17, -18,  -8,   1,   4,   8,  -4,  -2, -16,  -2,  14,   3,  -3,   5,  -7,  -4,  -1,  11,  -1 },
     {    2,   5,  11,   9,   7,  13,  -4, -15,   1,   9,  14, -17, -24,   9,   1,  10,   6,  26,  32, -12, -24,  15,  24, -17,   5,  25,  21, -25,   5,  12, -25,  26,   3,  17,  -4, -15,  -5,   0,  -2,  -5,   6,   2,  -8,   7,  46, -52,   9,   9,   1, -10,  -8,  15, -30,  14,  14,  -4,   0,   0,  -4,   4,  11,  -3,  -4,   5 },
     {   -2,  -2, -12, -15,  10,  -1, -42,  31,  -1,   0,  -8, -11,  -2,  18, -12,  -3,   1,   2,   9,  30, -11,  -7,  40, -34,  -2,  -4, -16,  17,  18, -32,  11,  11,   0,  14,  -7, -30, -10,  12, -10,   8,   6,  27,  28,  24, -18,   4,  -3, -10,  -5,  -7, -40,  -2,   9,   4,   1,   8,   0,  12,   3,  -7,  11,   0,   2, -10 },
     {   -1,  -1,  -1, -11,   9,  -5, -28,   4,   0,  -3,  -4,  -9,  34,  13, -29,  17,  -3,  -8, -14, -16,  28,  10,  -8,   3,   0,   3,  10,  -8, -10, -16,  20, -18,   3,  10,  14,  -8,  17, -25,  50, -28,   8,   6,  21, -40,  23,  -6,   0, -13,  -6, -23,  -8,  27, -20,   8,  -9,  12,  -8,  14, -12,  -6,   9,  -2,  -4,  10 },
     {    1,   1,  -5,  -2,  16,  31,  34, -29,  -3,  -3,  -6, -15, -26, -39,   7,  13,   3,   7,  13,  17,  19,  -4,   3,   3,   1,   3,   9,   1,  -9,  -9,   9,   7,   1,   6,   8,   2,  12, -19, -24,  32,   6,  12,  20, -10,   1,  35, -23, -16,  -5, -12,  -5,  44, -16,  -3,   8,   5,  -1,  28,   3, -19,  12,  13,   4, -15 },
   },
   {
     {  102, -40,  -3,  -6,  -1,  -2,  -1,   0,   5,   1,  -2,   0,   0,   0,   0,   0, -60,  23,   4,   3,   1,   1,   1,   0,  -3,  -1,   1,   0,   0,   0,   0,   0,  10,  -3,  -2,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   15,  73, -66,  22, -16,   4,  -5,   1,  -8,  10,  -4,   2,  -1,   0,   0,   0, -34, -41,  45, -13,   9,  -1,   3,   0,   4,  -7,   2,   0,   0,   0,   0,   0,  11,   4,  -7,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   14,   9, -12,   3,  -2,   0,  -1,   0,  96, -25,  -7,  -3,  -3,  -1,  -1,   0,  23,  -9,   4,  -3,   0,   0,   0,   0, -67,  18,   7,   1,   2,   0,   1,   0, -20,   2,   2,   0,   1,   0,   1,   0,  10,  -3,  -2,   0,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 },
     {   62,  -4, -11,  -2,  -2,  -2,   0,  -1, -16,  -3,   5,   0,   1,   0,   1,   0,  81, -39,   1,  -2,  -1,   0,  -1,   0,  37,   3,  -9,   0,  -2,   0,  -1,   0, -44,  19,   3,   0,   2,   0,   1,   0, -16,  -2,   4,   0,   1,   0,   0,   0,   5,  -3,  -1,   1,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0 },
     {    6,   1, -11,   8,  -4,   2,  -1,   1,  -1,  70, -50,  16, -12,   2,  -3,   0,  19,  23,  -9,  -2,  -2,  -1,   0,  -1, -17, -56,  47, -12,   9,   0,   2,   1, -25, -21,  13,  -1,   3,   1,   1,   0,   7,   9, -10,   1,   0,  -1,   0,   0,   7,   4,  -2,   0,  -1,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0 },
     {   23,  18,  37, -39,   8,  -8,   3,  -2, -16,  16,  -4,   0,  -1,   0,   0,   0,  19, -55, -13,  24,  -2,   4,  -1,   1, -36, -14,  12,   0,   3,   1,   1,   1,  65,  13, -11,  -5,  -4,  -1,  -1,   0,  30,   7,  -8,   1,  -2,  -1,  -1,   0, -22,  -1,   4,   1,   1,   0,   0,   0,  -5,  -1,   1,   0,   0,   0,   0,   0 },
     {   -4,  13,  14, -14,   3,  -4,   1,  -1,  72,   2, -15,  -1,  -4,  -1,  -1,  -1,  -9,  -9, -11,  13,  -3,   4,  -1,   1,  72, -37,   2,  -2,  -1,   1,  -1,   0,  29,   9,  -3,  -4,  -1,  -1,  -1,   0, -45,  16,   3,  -1,   2,  -1,   1,   0, -12,  -3,   2,   0,   0,   0,   0,   0,   2,   0,  -1,   1,   0,   0,   0,   0 },
     {   -2, -49, -26,  40,  -7,   9,  -2,   2,   7,  -4,  -3,   1,   1,   0,   0,   0,  46,  15,  34, -40,   7,  -9,   1,  -2,   6,  -7,   3,   0,   0,   1,  -1,   0,  67, -12, -26,  13,  -5,   2,  -1,   0,   6,   8,  -1,  -2,   0,   0,   0,   0, -23,   2,   9,  -3,   1,   0,   0,   0,  -3,  -1,   0,   1,   0,   0,   0,   0 },
     {  -22, -45,  25,  -2,   8,  -1,   2,  -1,   3,   6,  -5,  -1,   1,   0,   0,   0, -29, -41,  47, -16,   7,   0,   2,   0,  -7, -26,  15,   0,   1,   2,   0,   0, -39,  58, -40,  14,  -4,   0,   0,  -1,   3,  13,  -7,  -1,   0,  -1,   0,  -1,   8, -11,   8,  -2,   0,   1,  -1,   0,   0,  -1,   0,   1,   0,   0,   0,   0 },
     {   -6,  -7, -39, -28,  67, -36,  16,  -4,  -3,  -2,  -2,   4,  -2,   1,   0,   0,   8,  15,  36,  22, -60,  31, -14,   1,  -2,   3,   1,  -5,   5,  -3,   0,   0,   2,  -9, -16,  -5,  19,  -9,   3,   0,  -5,  -1,   2,   2,  -3,   2,   0,   0,  -1,   2,   5,   1,  -4,   2,   0,   0,   3,   1,  -2,  -1,   1,  -1,   0,   0 },
     {    0,   1,   2,   3,  -2,   1,   0,   0, -19,  17,  -8,   5,  -2,   0,   0,   0,   5,   5,  -4,  -2,   3,  -2,   0,   0, -50,  14,  -3,   0,   2,   0,   0,   0,  15,  19,  -3,   0,  -3,   1,  -1,   0,-101,   0,  24,  -1,   6,   1,   2,   0,  -7, -19,   4,   0,   2,   0,   1,   1,  28,   3,  -8,   0,  -2,   0,   0,   0 },
     {    6,  15,   9, -12,   7,  -5,   0,  -1,   8,  34,  39, -50,  20, -13,   3,  -3,   7,  14,   9, -11,   0,  -2,   0,   0, -16, -40, -39,  58, -19,  12,  -2,   1, -12, -26, -16,  22,  -4,   5,   0,   1,  -3,  14,   9, -15,   3,  -2,   0,   0,   0,   6,   5,  -4,   1,  -1,   0,   0,   0,  -3,  -1,   2,   0,   0,   0,   0 },
     {   -1,   9,  -4,   1,  -2,   1,  -1,   0,  13,  57, -30,   7,  -8,   0,  -2,  -1,   0,  25, -12,   4,  -3,  -1,   0,   0,  13,  44, -36,  10,  -4,  -1,   0,   0,  -7,  42, -27,   4,  -2,  -1,   0,  -1,  25, -28,  22,  -8,   1,   0,  -1,   0, -40, -16,  21,  -3,   2,   1,   0,   1, -24,  -2,   3,   0,   2,  -1,   1,   0 },
     {    1,  -4,   3,  -4,   2,   0,   1,   1,  12,  22, -13,   0,  -3,  -1,  -1,   0,   7,  -3,   2,  -2,   2,  -1,   0,  -2,   9,   8, -16,   5,  -3,   1,   0,   0,  32,   5, -10,   4,  -4,   1,   1,  -1,  19, -34,  15,  -4,   2,   0,   0,   0,  98, -21, -11,  -1,  -3,   0,  -2,   1,  40,  13,  -9,   0,  -3,   0,  -1,   0 },
     {   12,  19,  33,  57,  32, -33,  -6,  -9,   1,   3,   1,   1,  -1,   0,  -2,  -1, -12, -16, -34, -62, -32,  31,   3,   7,   1,   4,   2,  -3,   2,  -1,   0,   2,   4,   7,  17,  31,  12, -12,  -1,  -3,   2,  -2,  -3,   0,   0,   0,   2,  -1,   0,  -4,  -5,  -9,  -2,   4,   1,   1,  -1,   0,   2,   0,   0,   0,   0,   0 },
     {   11,  37,  18, -37,  13,  -6,   1,  -2,  -5, -18,  -6,   7,   2,  -2,   2,  -1,   8,  29,   5, -20,  17, -17,   4,  -1,  10,  17,  34, -18,  -2,  -7,   1,  -2,  -8, -17, -43,  58, -29,  19,  -5,   1,  -9, -14, -21,  13,   3,   3,   0,   1,   3,   1,  20, -19,   7,  -3,   0,   1,   0,   8,   0,   1,  -2,   0,  -1,  -1 },
     {    7,  23, -24,  15,  -8,   2,  -1,   2, -10, -29,  17,  -5,   1,   0,   0,   1,  10,  46, -38,  14, -12,   2,  -1,  -1,  -7, -28,  10,   4,   1,   2,   0,   1,   8,  60, -26,  -7,   3,  -3,  -2,   1,  12,  42, -19,  -3,  -2,  -2,   0,  -2,  21, -36,  16,   2,  -1,   2,  -1,   1,  -1,  -3,   3,   1,   0,   0,   0,   1 },
     {   -6, -12, -24,  -8,  16,  -2,   3,   2,  -7, -12, -44, -31,  41, -14,  10,   1,  -3,  -7, -10,   0,   1,  -3,   1,   0,   7,  20,  45,  40, -52,  20, -11,   0,   8,  18,  35,   7, -15,   4,  -3,  -2,   1,  -4,  -6, -11,  16,  -5,   2,   1,  -5, -10, -11,  -2,   6,  -1,   1,   0,  -3,   1,  -2,   0,   1,  -1,   1,  -1 },
     {    0,   2,   0,   3,   0,   4,   0,   2,   2,  -6,   2,  -2,   0,  -1,   1,   1,  -2,   1,  -4,   1,  -3,  -1,   1,  -2,   8,  -4,   5,  -2,  -1,   0,   0,  -1, -14,  -2,  -1,  -1,   1,   1,  -1,   0,  30,  -2,   4,  -3,   2,  -1,   0,  -1, -47, -10,  11,  -1,   2,   1,  -1,   0, 109,  -6, -24,   2,  -5,  -1,  -2,   0 },
     {    2,   2,   3,  -1,  31,  36, -67,  27,   1,   2,   3,   9,   3,  -4,   0,   1,  -1,  -4,   2,   5, -36, -34,  64, -25,  -1,   0,  -2,  -4,  -4,   8,  -5,   0,   1,   0,  -2,   1,  19,  11, -25,   9,  -1,   3,   3,   3,   0,  -6,   5,  -1,   2,   3,  -3,  -2,  -3,  -3,   6,  -3,  -2,   1,  -1,  -2,   0,   1,  -1,   1 },
     {   -4, -13, -21,   9,   6,   0,   1,  -1, -12, -44, -42,  31,  -4,  12,  -3,   7,  -9, -25, -38,  18,   4,   2,   3,   6, -10, -38, -25,  26,   3,  -5,   4,  -3,  -9, -23, -21,  28,  -7,   8,  -1,   2,   3, -13,  42, -16,  -3,   0,   1,   0,   0,   3,  26, -12,  -2,  -2,   1,   0,  -8,  11,  -5,   0,   3,  -1,   0,  -1 },
     {    1,   3,   4,   6,  -1,   3,   0,  -2,   1, -11,  34, -15,   1,  -2,   0,  -3,   3,  12,   8,   1,  -3,   0,   2,  -1,  -2, -43,  31,  -4,   5,   1,   0,   1,   1,  13,  10,  -9,   3,  -3,  -1,   1,   1, -84,  23,  10,  -2,   6,   1,   3, -21, -33,  -6,   9,   1,   1,   1,   1, -19,  38, -12,  -1,   0,  -2,  -1,  -1 },
     {   -6, -11, -31, -25,  36,  -4,   6,   3,   4,   9,  22,   9, -20,   4,  -5,   5,  -1,   1,  -7, -15,  19, -13,   5,  -4,  -4, -13, -24, -17,  23,  -8,   8,  -3,   5,  27,  53,  27, -53,  16,  -8,   3,  11,  16,  13,  -4,   0,   0,   1,   0,  -7, -11, -27,  -9,  23,  -1,   1,   2,  -1,  -1,  -2,   0,  -4,   1,   1,   0 },
     {   -9, -29, -39,  19,   8,  13,  -4,  -7,  11,  36,  42, -18,  12,  -3, -12,  -1, -10, -34, -45,  34,   2,   4,  -1,  11,   2,  17,  17, -16,   3,   2,   3,   3,  -1, -13, -15,  25,  -7,  -3,   5,  -1, -11, -12, -33,  14,  -8,  -2,   4,   1,   4,   2,  27, -16,   0,  -1,   0,   3,   1,   8,  -2,   4,  -2,   5,  -5,   0 },
     {   -4,  -5,  -9, -26, -20,  19,  -4,   4,  -4,  -4, -11, -41, -32,  35,  -3,   6,  -1,  -1,  -1,  -8, -11,   0,   3,   1,   5,   9,  16,  42,  40, -53,   4,  -2,   6,  12,  16,  32,  36, -21,  -4,  -5,  -4,  -3, -11,  -6,   2,  14,  -1,  -3,  -1,   7,  -3,  -9,  -9,   7,   0,  -1,   3,   1,   2,  -1,  -2,   2,  -1,  -1 },
     {    1,   4,  -3,  10,  -6,  -4,   3,   4,   0,   2,   2,   1,   1,  -4,   1,  -1,   2,  12,  -8,   8,  -5,   4,  -2,  -3,  -3,  -3,   1,   1,  -3,   6,  -2,   2,   4,  39, -15,   2,  -5,   3,  -1,  -1,  -1,  -1,   4,  -4,   3,  -1,   1,   4,  -1,  93, -30, -12,   5,  -6,   1,  -2,  14,  62,  -5,  -7,  -2,  -3,   2,  -3 },
     {    1,   2,  10,   3,  -9,  23, -14,  -3,   2,   1,  25,  50, -32,  19, -24,   4,   2,   3,  16,  17, -19,  16, -12,  11,   0,   1,   8,  24, -29,  15,   1,  -6,   1,  -3,   3,  22, -10, -15,  22, -15,  -3,  -7, -23, -39,  48, -17,   4,   2,  -3,  -8, -18,  -4,  12,  -5,   3,   2,  -1,  -1,   1,  13,  -6,   5,  -2,   1 },
     {    5,   8,   8,   9,  24,  49,  40, -30,   1,   1,   1,   3,   0,  -4,   5,  -9,  -5,  -6,  -3, -10, -34, -67, -35,  24,   0,   0,  -1,  -3,   0,  -2,   6,   8,   2,   6,   8,  -2,  23,  38,  16,  -7,   1,   4,   4,  -8,   0,   6,  -3,  -2,   2,   4,   7,  -2,  -6,  -9,  -6,   3,  -2,   3,  -2,   1,   1,  -2,   3,  -2 },
     {    3,   3,  15,  40,  12, -33,  10,  -7,   1,   5,   5,  -4, -14,  25,  -4,  -2,   1,  -2,  11,  26,   4,  -9,  14, -13,   5,  13,  18,  33,  10, -27,  17, -17,  -2,  -6,  -5, -31, -20,  48, -26,  24,  -4,  -2, -17, -29, -10,   8,  -8,  19,   2,  -5,   0, -10,  16, -21,   6,  -5,  -1,  -2,  -9,  14,  -1,   0,  -5,  -5 },
     {    0,  -1,   4,  14,  -6,   7,  12,   1,  -1,  -2, -20,  17,  -3,   6,   8,  -5,   0,  -1,   8,  17, -14,  -2,  -5,  -4,  -3,  -8, -34,  32,  -8,   4,   0,   9,  -1,  -3,   1,  17,  -4,   3,  -3,  11,  -5,  -5, -50,  67, -23,   9,   1,  -7, -11, -25, -34,   1,  20,  -4,   5,  -7,   8,  19,  12, -29,   8,  -4,  -1,   4 },
     {    0,   1,   0,   1,   6,   5, -16,  -7,  -1,  -1, -10, -21,  21,  33, -38,  -5,   0,   0,  -4,  -7,  -8,  -3,   9,  14,  -5,  -9, -12, -26, -45, -48,  44,   5,  -3,  -4, -13, -19, -19,  -9,  10, -14,   2,   5,   5,  15,  26,  29, -28,  -2,   0,   1,  -5,  15,  19,  11,  -2,   3,   1,   7,   2,  -3,   6,   1,   9,   5 },
     {    0,  -1,   7,  -3,  -7,  -1,   2,  10,   1,   9,   4,   2, -15,   6,  -4,   2,  -1,  -5,   9,  -2,   0,   7,   0, -10,   5,  23,  10,   5,  -6,  13,  11,  -6,  -3, -17,  16,  10,   7,   8,  12,   6,   9,  51,  18,  10,  -1,   8,  -1,  -4,   0, -29,  25,  19, -11,   8,   2,  -6, -12,  78, -43,   8,  12,  -7,   1,   0 },
   }
 },
 {  //20
   {
     {  124, -25,  -8,  -5,  -3,  -2,  -1,  -1,  -5,   3,   0,   0,   0,   0,   0,   0, -16,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    7,  -5,   0,   0,   0,   0,   0,   0, 120, -16,  -9,  -4,  -3,  -2,  -2,  -1,  21,   8,  -3,   0,  -1,   0,   0,   0, -29,   7,   1,   1,   1,   0,   0,   0, -13,   0,   1,   0,   0,   0,   0,   0,  -4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {  -14,   5,   1,   1,   0,   0,   0,   0,  22,   7,  -3,  -1,  -1,   0,  -1,   0,-118,  14,   9,   3,   3,   2,   2,   1, -16, -14,   2,   0,   1,   0,   1,   0,  29,  -4,  -2,  -1,  -1,   0,   0,   0,   4,   2,   0,   0,   0,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 },
     {   -1,   4,  -1,   0,   0,   0,   0,   0, -30,   7,   2,   1,   1,   1,   0,   0,  -6,  13,   0,   1,   0,   0,   0,   0,-101,  17,   9,   3,   3,   1,   1,   0, -58, -13,   8,   1,   2,   0,   1,   0,  13, -17,   0,   0,   0,   0,   0,   0,  22,  -5,  -3,   0,  -1,   0,   0,   0,   8,   1,  -1,   0,   0,   0,   0,   0 },
     {    5,   1,  -1,  -1,   0,   0,   0,   0, -11,  -8,   2,   1,   1,   0,   0,   0,  33,   5,  -4,  -1,  -2,   0,  -1,   0, -62,  -6,   7,   1,   2,   1,   1,   0,  95, -20, -10,  -2,  -3,  -1,  -1,   0,  11,  23,  -3,   0,  -1,   0,   0,   0, -29,   2,   4,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {    1, -10,   2,   0,   1,   0,   0,   0,   8,  -9,   0,   0,   0,   0,   0,   0,   4, -10,   0,   0,   0,   0,   0,   0,  16, -11,  -1,   0,   0,   0,   0,   0,  22, -21,  -1,  -1,   0,   0,   0,   0,  90, -21, -11,  -2,  -3,   0,  -1,   0,  74,   7, -13,   0,  -3,   0,  -1,   0,  16,  13,  -4,   0,  -1,   0,   0,   0 },
     {    2,   4,  -1,  -1,   0,   0,   0,   0,  -3, -10,   3,   0,   1,   0,   0,   0,   4,  10,  -1,   1,  -1,   0,   0,   0, -10, -15,   3,   0,   1,   0,   0,   0,  31,  16,  -5,   0,  -2,   0,  -1,   0, -76, -11,  12,   2,   3,   1,   2,   1,  86, -18, -14,  -1,  -3,   0,  -2,   0, -11,  19,   1,   0,   0,   0,   0,   0 },
     {   24, 120, -18,   0, -10,  -3,  -5,  -2,   4, -17,   5,  -1,   1,   0,   0,   0,  -1, -23,   1,  -2,   0,   0,   1,   0,   2,  -2,   0,   1,   0,   0,   0,   0,  -2,  -3,   0,   0,   0,   0,   0,   0,   7,   1,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
     {    0,   2,  -1,   0,   0,   0,   0,   0,  -2,   2,  -1,   0,   0,   0,   0,   0,   0,  10,   0,   0,  -1,   0,   0,   0,  -4,  -3,   1,   1,   0,   0,   0,   0,  -4,   5,   0,   0,   0,   0,   0,   0,  17,  -2,  -3,   0,  -1,   0,  -1,   0,   4,  37,  -3,   0,  -1,  -1,   0,  -1,-117,  10,  24,   0,   5,   0,   2,   0 },
     {   -3, -24,   7,   7,   3,   0,   1,   0,  -9, -96,  15,   7,   6,   3,   3,   2, -14, -67,   5,   5,   4,   2,   2,   2, -13, -23,  -3,   5,   1,   1,   0,   1, -14,   1,  -6,   3,   0,   0,   0,   0,  -7,   9,  -2,   0,   0,   0,   0,   0,  -7,   6,   1,  -1,   0,   0,   0,   0,  -6,  -1,   2,   0,   0,   0,   0,   0 },
     {    0,   4,  -2,  -1,  -1,   0,  -1,   0, -10, -69,   4,   3,   4,   3,   2,   1,  -7,  67, -15,  -1,  -5,  -1,  -2,  -1,  20,  62,   3,  -4,  -3,  -3,  -1,  -2,  11, -22,  18,  -1,   1,   0,   1,   0,  -2, -33,   1,   4,   0,   2,   0,   1, -11,  -3,  -7,   1,   0,   0,   0,   0,   1,   7,  -3,  -2,   0,  -1,   0,   0 },
     {   -1,   1,  -4,   0,   0,  -1,   0,   0,  -4, -10,  -3,  -1,   0,   0,   0,   1,  -8,   7, -12,   1,  -2,   1,   0,   0,  -9,  53, -25,   1,  -3,  -1,  -2,   0,   3,  81, -21,  -5,  -4,  -2,  -2,  -1,  19,  62,  -4,  -8,  -1,  -2,  -1,  -1,  21,  21,   5,  -4,   1,  -2,   0,   0,  13,   8,   1,   0,   0,   0,   0,   0 },
     {  -11, -15,-111,  -9,  -6,   7,   4,   6,   2,  -3,  32,  -4,  -1,  -4,  -2,   0,   4,   9,  45,  10,   3,   0,  -2,  -1,   0,  -3,  -9,   1,   0,   1,   1,   1,  -1,  -2,  -6,  -1,  -1,   0,   0,   0,   1,  -1,   1,  -1,   0,   0,   0,   0,  -1,   0,   4,   0,   0,  -1,  -1,   0,   1,   0,  -2,   2,   0,   0,   0,   0 },
     {   -3,  -9,  -7,   4,  -1,   1,  -1,   1,   7,  30,   8,  -9,   0,  -2,   1,   0, -11, -72,  -4,   7,   3,   3,   2,   2,   0,  84,  -6,  -6,  -5,  -4,  -2,  -2,  20, -32,  20,   4,   4,  -1,   1,   0, -17, -12,  -8,   2,   1,   1,   0,   0,   8,   1,  -1,   0,  -1,   0,  -1,   0, -11,   6,   0,   0,   0,   0,   0,   0 },
     {    1,  -1,   2,  -1,  -1,   0,  -2,   0,   1,  -6,   3,   0,   3,   0,   0,  -1,   0,  -5,   2,  -1,   1,  -4,   1,  -1,   0,  12,   0,  -3,  -1,   0,   0,   0,  10,  25,   4,  -5,  -2,  -3,   0,   0,  21, -14,  24,  -3,   4,  -1,   0,  -1,  12, -73,  33,   7,   5,   2,   1,   1, -25, -80,  13,  13,   1,   4,   1,   2 },
     {    3,   3,  43, -24,  -8,  -1,  -1,   0,   5,   1,  76, -26, -15,  -2,  -2,   1,   3,   6,  61, -19, -13,  -2,  -1,   1,   2,   9,  40, -11,  -9,   2,   0,   1,   1,  12,  19,  -5,  -8,   0,  -2,   1,   4,  12,   8,  -1,  -6,  -1,  -1,   1,   1,   5,  -2,   0,  -2,  -1,   0,   1,   2,   6,  -4,   1,  -1,  -1,   1,   0 },
     {    0,   0,   1,   2,  -2,   0,   1,   1,   0,   5,  -2,   3,  -2,   1,  -1,   0,   2, -18,   9,   2,   1,   1,   0,   0, -13,  -4, -18,   3,  -2,   1,   0,   0,  18,  66,   9,  -8,  -2,  -4,  -1,  -2,  15, -76,  27,   5,   5,   2,   1,   2, -33, -11, -27,   5,  -1,   3,  -1,   1,   2,  40,  -5, -11,  -2,  -2,   0,  -1 },
     {    1,   1,  10,   1,  -6,  -4,  -3,   0,  -5,  -9, -73,  -3,   7,   9,   4,   0,   3,  -4,  78, -12,  -2,  -1,   3,  -1,   3,  24,  29,  17,  -4,  -2,  -3,   0,   0,  -7, -49,   4,   2,   2,  -1,   1,  -1,  -6,  10,  -4,   2,  -3,   0,  -2,   1,   0,   7,  -3,   0,   0,   0,  -1,  -1,   6,   1,   1,   0,  -1,  -1,   0 },
     {    3,   0, -11,  72,  41,   8,  -9, -11,   1,   2,  12,  -2,   2,   2,  -2,   0,  -2,   1,   6, -73, -43,  -8,   5,   8,  -1,  -1, -16,   3,  -4,  -3,  -1,  -2,   2,  -1,   4,  29,  16,   1,  -3,  -5,   1,   2,   8,  -2,   1,   3,   0,   1,   0,  -1,   0,  -4,  -6,   0,   2,   3,  -1,   1,  -1,   2,   3,   1,   0,   0 },
     {   -1,  -2,  -3,  -1,  -3,  -1,   3,   0,   1,   0,   4,   0,   4,   3,   3,  -1,  -4,   0, -12,   1,   0,  -3,  -1,   2,   6,   1,  10,  -3,   0,  -3,   2,   0, -12, -16, -10,   8,  -2,   1,   1,   1,  17,  40,  11,  -8,   2,  -4,   0,  -3,  -9, -80,   7,  19,   3,   5,   0,   3, -21,  74, -10, -15,  -5,  -4,   0,  -1 },
     {    3,   4,  28,  -1,   1, -10,   0,   2,   4,   7,  44,   7,   2,  -3,  -3,   0,   1,  10,  10,   8,   2,   5,  -3,   0,  -2,   6, -70,   1,  -1,   4,   2,   2,  -4, -25, -77, -11,   6,   5,   5,   4,  -5, -22,   0,  -8,   3,  -1,   3,  -1,  -2,  -1,  23,   1,   0,   0,   0,   0,  -1,   5,   9,   2,   1,  -1,  -2,   1 },
     {   -4,  -5, -11, -55, -27,  -7,  15,  13,  -6,  -9, -22, -63, -39,   0,  17,  11,  -5,  -8, -26, -52, -24,  -4,   7,   4,  -3,  -7, -23, -13,  -5,   6,   5,   1,  -2,  -6,  -9,   8,  12,   6,  -1,   0,  -1,  -5,   9,   3,   5,   3,  -1,   1,  -1,   2,   2,   4,   0,   0,   0,  -1,  -1,  -2,   0,   3,   2,   1,  -1,   1 },
     {    0,  -5,   4,  35, -70,  23,  -1,   7,   0,   1,   8,  -8,   8, -12,  -2,   1,   1,   6,  -2, -37,  66, -10,   6,  -4,   0,   1,  -8,  19,  -2,  16,   1,  -2,   0,  -2,   2,  24, -34,   6,   0,  -2,   0,   0,  -1, -14,   4, -11,   2,   1,   0,  -1, -11,  -9,  14,   0,   0,  -1,   2,  -1,  10,   1,   1,   5,  -1,  -1 },
     {    1,   2,  16,   0, -16, -14,  -1,   3,  -2,  -4, -23,  -2,  13,  20,  11,  12,   5,   9,  48,  12,  -8, -19, -12,  -3,  -6, -10, -71, -12,  -1,  -2,   1,   4,   3,  -4,  64,   6,   3,  -5,   1,   1,  -3,  15, -31,   2,  -1,   3,   4,   3,   3, -13,   9,   0,  -1,  -1,   1,  -1,  -3,   8,   3,   4,  -6,   1,   1,   2 },
     {    5,   5,   7,  39,  35,  24,  15,   4,  -7, -10, -15, -59, -57, -36, -26,  -8,   4,   7,  14,  37,  23,  17,   6,   6,   1,  -2, -11,  -5,   5,   9,   5,   3,  -1,  -4,   7, -12,  -8,  -5,  -4,  -4,   2,   5,  -1,   2,   9,   4,   7,   1,   0,  -8,   4,  -4,  -1,  -4,  -2,  -3,   0,   5,   8,   1,   1,   1,  -2,  -2 },
     {    0,   1,   3,   2,   2,   6,  -4,   7,   0,   2,   5,  -6,  -4,   0,   1,   7,   0,   2,   7,   1,  -9,  -3,   4,   3,   0,   5,  -1,   1,   6,   1,   7,   1,  -2,   8, -29,   3,  -3,   4,   0,   2,  -1,   4, -68,  23,   5,   6,   1,   3,  -7, -28, -79,  22,   7,   5,   1,   2, -14, -27, -33,   7,   5,   3,   3,   0 },
     {    1,   3,   0, -27,  55, -24,  17,  -6,   0,  -1,  -1, -32,  68, -38,  11, -14,   1,   2,   4, -26,  51, -11,   8,  -4,  -1,   0,  -6, -12,  10,  -4,   5,   3,   0,   1,   2,   2,  -6,  -3,  -4,   1,   0,   0,  -1,   7,  -3,   6,  -2,   1,  -2,   1,  -7,   4,  -9,   4,  -2,   2,   0,   0,   0,   1,  -8,  -1,  -2,   1 },
     {    0,  -3,  -7,  36, -17, -20,  -6,   4,   0,  -1, -10,   3,   6,  -3,   7,  12,   0,  -3,  -2,  -9,   1,  -2,   8,   0,  -1,  -2,  11, -80,  -9,   9,  20,  11,  -3,  -1,  -7, -56, -11,  10,   3,   1,  -3,  -3,   2,  -8, -16, -14, -13,  -2,  -1,   6,   1,  25,   3,   1,   7,  -5,   2,   7,  -4,  40,   6,  -5,  -7,  -4 },
     {   -4,  -6,  -6,   1, -18, -72, -63, -46,  -2,  -2,   1,  -9, -13, -19, -21,  -7,   1,   2,  -1,  -1,   6,  29,  19,  19,   1,  -1,   4,  12,  19,  10,  13,   6,   1,   6,   5,  11,  14,   6,   1,  -2,   0,  -2, -17,   1,   0,   0,   1,   3,  -2,  -8,   4, -11,  -6,  -6,  -1,   3,   0,   3,   7,  -2,  -2,   0,  -3,  -1 },
     {    1,   2,  -2,  -9,   6,  15,  16,   7,  -1,  -2,   0,  10,  -3,  -5,  -7,  -6,   1,   0,  -5, -16,   5,   7,  18,   9,  -3,  -4,  22,  -7, -32,  -5,  -6,  -1,   5,  23,  -2,  10,   3,   8,  10,   1,  -6, -28, -82, -20,  -1,   3,   9,   9,  -2, -10,  57,  -7, -13,  -7,  -6,  -1,   1,  18,   3,  11,   0,   8,  -2,   1 },
     {    0,  -1,   4,   7, -25,  40, -40,   4,  -1,   0,  -1, -43,  51, -22,  16, -16,   1,   3,  -5,  28, -35,  29, -19,  23,  -2,  -7,   0, -17,   0, -34,   2, -11,  -1,   5,  -8,  23, -12,   9,   4,  -1,   1,  -7,   4,   7,  -2,  -3,   2,   0,  -1,   5,  13,  -4,   5,   4,   0,  -2,   0,   1, -12,   6,   2,   8,  -1,  -1 },
     {    0,  -1,   1,   6, -21,   3,  44, -11,   0,   0,  -3,  -1,  20, -22,  -2, -10,  -1,  -2,   1, -22, -30,  42,  -2,   8,   2,   2,  -6,  52,  26,   1,  13,   8,  -1,  -4,  15, -63, -11,  -1,  -3,  -4,  -3,   2, -11,  -1, -23, -12,   5,   2,   2,  -3,   7,  13,  -7,   0,   8,   6,  -2,  -1,  -6, -14,   8,   1,   7,  -5 },
   },
   {
     {   97, -72,  30, -21,  10,  -6,   3,  -2, -12,  11,  -6,   3,  -2,   1,   0,   0,  -2,   2,  -2,   2,  -1,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   73,  42, -72,  38, -29,  13,  -9,   3,  15, -26,  20, -12,   7,  -3,   1,   0,  -9,   0,   5,  -4,   4,  -2,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   2,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -6, -21,  21, -12,   9,  -4,   3,  -1, 104, -55,  18, -15,   4,  -3,   0,  -1, -15,  18, -10,   6,  -3,   2,  -1,   0,  -3,   0,   0,   1,  -1,   1,   0,   0,  -2,   3,  -1,   1,   0,   0,   0,   0,  -1,   1,  -1,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {    8,   2,  -5,   3,  -2,   0,   0,   0,  41,  15, -27,  13, -11,   4,  -2,   1,  92, -48,  21, -17,   7,  -4,   0,   0, -35,  20,  -5,   3,   0,   0,   0,   0,   6,  -4,   1,   0,  -1,   1,   0,   0,  -4,   1,   1,  -1,   1,  -1,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   -3,  -4,   8,  -8,   6,  -5,   3,  -1, -46, -55,  58, -28,  22,  -8,   6,  -1,  68,  13, -31,  14, -12,   4,  -4,   0,  -9,   4,   2,   0,  -1,   1,   0,   0,  -4,  -2,  -1,   1,   0,   0,   0,   0,   4,   4,  -3,   2,  -2,   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0 },
     {    1,  -3,  -3,   3,  -1,   1,   0,   0,  14,   4,  -9,   4,  -4,   1,  -1,   0,  37, -11,  -6,   2,  -2,   1,   0,   1, 115, -24,  -4,  -5,  -2,  -1,  -1,   0, -21,   9,   0,  -1,   1,   0,   0,   0,  12,  -3,  -1,   1,  -1,   1,   0,   0,   0,   5,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   -8, -14,  12,  -3,   4,  -2,   2,  -1, -15, -28,  35, -23,  18, -10,   6,  -2, -27, -67,  61, -32,  22,  -8,   5,  -1,  20,  15, -25,  14, -10,   5,  -3,   0,  18,   8, -10,   5,  -4,   2,  -2,   0,  -5,   1,   0,   1,  -1,   0,   0,   0,   3,   0,  -2,   1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {    3,   3,  -3,   1,   0,   1,   0,   0,   2,  -3,  -4,   4,  -3,   3,  -2,   1,   9,  10, -11,   6,  -4,   2,  -1,   0,  16, -25,   6,  -2,   2,   0,   0,   0, 118,  -5, -16,   1,  -6,   0,  -2,   0, -24,   9,   4,  -3,   3,  -2,   1,   0,   6,  -7,  -1,   0,   0,   0,  -1,   0,  -1,   3,   1,   0,   0,   0,   0,   0 },
     {   27,  67,   1, -55,  53, -53,  32, -17,  -1, -11,  -2,  12, -14,  12,  -7,   3,  -7, -12,   4,   4,  -7,   9,  -6,   4,   5,   2,  -2,   1,   0,  -1,   1,  -1,  -2,  -3,   0,   2,  -2,   2,  -1,   1,   0,   0,   1,   0,   0,  -1,   0,   0,  -1,  -2,   0,   1,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    0,   2,   1,   0,  -1,   0,   0,   1,  -1,   3,  -1,  -1,   2,  -1,   1,   0,   4,  10,  -4,   2,  -2,   1,  -1,   0,  11,  21, -15,   6,  -5,   2,  -1,   0, -20,  21,   2,  -2,   1,  -1,   0,   0,-112,   5,  26,  -4,   7,  -1,   3,   0, -31,  -7,   4,   1,   1,   0,   1,   0, -13,  -2,   2,   0,   1,   0,   0,   0 },
     {    0,   1,  -1,   2,  -1,   1,  -1,   0,   3,  11,  -3,  -4,   3,  -4,   1,  -1,  -4,  13, -16,  12,  -9,   6,  -4,   2,  19,  69, -58,  28, -19,   6,  -2,  -1,  20, -54,  37, -21,  12,  -5,   3,   0,  16, -10,   3,   0,  -1,   2,  -1,   0,  -3,   9,  -4,   2,  -1,  -1,   1,  -1,  -4,  -2,   1,  -2,   1,   0,   0,   0 },
     {   -1,  -3,  -3,   1,   1,  -1,   1,   0,  -3,  -7,  -3,   1,   0,   0,   0,   0,   1,   4,   1,  -1,   0,   0,  -1,   1,   1,  -2,  -3,  -1,   3,  -3,   2,  -1, -12, -11,   6,  -5,   3,  -1,   1,  -1, -33, -25,  12,   3,  -1,   5,  -1,   1, 112,  -2, -25,   5,  -9,   1,  -2,   0,  21,  15,  -6,  -1,  -1,   0,  -1,   0 },
     {    9,  23,  13, -14,  -2,   5,  -7,   5,  18,  52,   0, -47,  51, -50,  31, -15,  -2, -24,  -5,  29, -30,  26, -15,   6,  -5, -11,   4,  -3,   1,   2,  -2,   2,   2,   9,  -9,   2,   1,  -3,   2,  -3,   0,  -5,   3,   1,  -4,   5,  -4,   2,   4,   1,  -2,   0,   1,  -2,   1,  -1,   3,   0,  -1,   1,  -1,   1,  -1,   1 },
     {    0,  -2,   7,  -4,   1,  -1,   0,  -1,   0,  -6,   6,  -7,   8,  -7,   5,  -2,  -2, -15,  16, -11,   9,  -5,   2,  -1,   0, -40,  38, -26,  18, -10,   6,  -2,  -3, -76,  54, -26,  16,  -5,   2,   1, -22, -12,   8,   4,  -2,   5,  -2,   2, -22,   6,  -5,   6,  -3,   2,   0,   0, -11,  -2,  -1,   1,  -1,   1,   0,   0 },
     {   21,  54,  65, -30, -30,  36, -41,  20,  -4, -15, -13,  11,   4,  -6,   9,  -5,   4,  15,  -6, -17,  27, -28,  22, -12,   5,   5,  -2,   2,  -4,   3,  -2,   1,  -1,  -5,  -6,   4,  -3,   3,  -1,   0,   3,   5,   1,   0,  -1,   0,  -1,   0,   4,   0,  -2,   2,  -1,   0,   0,   0,   2,   0,   0,   0,  -1,   1,   0,   0 },
     {    0,   0, -13,  -3,  21, -18,  17, -10,  12,  45,  20, -36,  16, -11,   4,   0,  18,  63,  19, -50,  36, -30,  13,  -7,   7,   9,  11,   0,  -9,   9,  -8,   3,   4,   2,   4,   8, -10,   9,  -5,   4,   3,  -4,  -9,   2,   1,  -2,   0,   0,   4, -10,   0,   2,  -1,   4,  -3,   3,   2,  -7,   4,   4,   0,   0,   0,  -1 },
     {    1,   2,   1,   0,   1,   0,   0,   1,   0,   2,   2,  -1,  -1,   0,   0,   0,  -1,   1,   2,   1,  -3,   1,   1,  -1,  -2,  -3,   0,   1,  -1,   2,  -1,   1,  -2,   4,   1,  -1,  -1,   1,  -1,   1,   2,  21, -17,   5,  -4,   1,  -1,   0,  32,  17,  -2,  -4,   2,  -2,   1,  -1,-115,  -5,  31,  -4,   8,   0,   3,   0 },
     {    0,   0,   1,   2,  -1,   1,   0,  -2,   2,   5,   0,   0,   2,  -1,   1,  -1,  -3,  -7,  -2,   3,  -2,   3,  -1,   1,   6,   7,   0,  -3,   3,  -2,   0,   0,  -8, -19,  13,  -4,   1,   2,   0,   1, -15,  83, -56,  22, -12,   2,  -1,  -1,  19, -49,  31, -14,   8,  -3,   2,   0,  18,   7, -11,   2,  -2,   1,   0,   0 },
     {   -4, -10, -19, -11,   8,   1,  -3,   9, -14, -45, -59,  19,  30, -25,  31, -16,   7,  39,  50, -13, -17,  21, -25,   8,   6,  16,  13,  -7,  -4,   2,  -3,   1,  -1, -10, -10,  10,  -5,   0,   4,  -3,   4,   8,   4,  -4,  -1,   1,  -2,   1,  -4,  -1,   4,   1,  -3,  -1,   0,   0,  -2,   2,   2,  -2,   1,   0,   0,   0 },
     {    0,   1,   6,   8,  -1,   1,   0,  -2,   1,   3,   7,  -7,  -6,   2,  -5,   3,  -8, -11,  -9,   7,  -1,   0,   0,   2,  18,  74,  40, -61,  33, -28,  11,  -8,   7, -11, -20,  24, -24,  16, -14,   5,  -3, -14,  -5,   7,  -5,   2,   1,  -2,   1,   5,  -1,   0,   3,  -2,   0,  -1,  -5,   4,  -2,   0,   0,  -2,  -1,  -1 },
     {   -1,   0,  -5,  -2,   3,  -1,   1,  -1,   0,   5,  -2,  -3,   1,   0,  -3,   0,   0,   5,  -1,  -4,   2,   0,   1,   1,  -1,  -1,  -1,   1,  -2,   2,  -2,   1,  -1,   1,  -6,   1,   1,   3,  -3,   0, -14,  44, -32,  12,  -8,   3,  -3,   1, -12,  86, -55,  19, -13,   4,  -1,  -1,  18,  35, -13,  -2,   0,  -2,   1,  -1 },
     {   11,  19,  58,  73, -26, -36,  32, -39,  -5, -11, -27, -19,  12,   5, -11,  13,   1,   9,   5,  -8,  -3,  12,  -9,   9,   0,  -6,  -3,   5,   0,   2,  -2,  -1,  -3,  -6,  -9,  -2,  -1,   6,  -3,   2,   3,   6,   9,  -1,  -3,   1,  -3,   1,  -1,   1,  -4,  -4,   0,   1,   1,   0,   0,  -1,   1,   2,  -1,  -1,   1,   0 },
     {    4,  11,  19,  -8, -14,  22, -19,   8,   4,  15,  34,  -7, -25,  33, -31,  18,   7,  21,  52,   1, -38,  39, -30,  16,   4,   5,  26,   6, -23,  20, -13,   5,   2,   2,  12,   5,  -4,   1,   5,  -5,  -3, -11, -10,  15,  -7,   4,   1,  -1,  -2,  -5,  -6,   6,  -3,  -1,   5,  -2,   3,  -1,   1,  -1,   0,  -4,   3,  -2 },
     {   -2,  -6, -10,  -2,  -2,  -3,  -3,   4,   1,   1,  -5,  -5, -10,   0,   2,   0,  -1,  -6, -12,  10,   8,   1,  -3,  -1,  -3, -24, -12,  20, -24,  21, -10,   5, -16, -66, -52,  47, -30,  28, -14,   8, -16, -24, -18,  14,  -1,   5,   0,  -2,  -8,  -2,   2, -12,   8, -10,   5,  -3,  -2,   2,   6,  -3,   5,  -6,   3,  -2 },
     {    0,   0,   2,   0,  -1,   1,  -1,  -5,   1,   4,   5,   3,   2,  -2,   0,  -1,   0,   2,   0,  -6,  -1,   0,   0,   1,   0,  -2,  -3,   3,  -2,   4,  -3,  -1,  -1,   1,   3,   1,  -1,   4,  -5,   0,   9, -13,  11,  -7,   6,  -5,   3,   0, -11, -28,   9,   3,   0,   5,   0,   0, -30, 107, -47,   6,  -6,   0,   0,  -2 },
     {   -2,  -3, -11, -17,   3,  10,  -2,  -2,  -7, -10, -37, -67, -12,  17, -22,  24,   6,   9,  35,  54,   3, -27,  20, -32,   0,   0,  -2,   8,  14,  -8,   5,  -5,  -1,   0,  -9, -10,  12,  -3,  -2,  12,   4,   9,  16,   4, -10,  12,  -6,   0,  -2,  -6,  -5,  -9,   7,  -5,   8,   2,   0,  12,   8,   4,   2,   0,   1,  -1 },
     {    0,   2,   4, -11, -16,   5,   1,  -2,   3,   9,  15,   9,  13,   9, -12,  -7,   6,  21,  28,  -1, -17,   7,  -1,   6,  -5, -26, -62,  -8,  50, -39,  37, -16,  -2,  -3,  15,  25, -25,  19, -13,   3,   0,  -1, -12, -17,  16,  -8,   1,   3,  -3,   3,  -5, -11,   6,   0,   5,  -3,   3,   1,   8,  -1,  -4,   6,  -4,   2 },
     {    1,   2,   3,   2,  -2,  -1,   8,  -9,  -4, -11, -13,   9,   4,  -7,   6, -10,   1,   3,   0,   0,  -2, -12,   4,  -4,  -4,  -3,  -9,  -4,   4,   6,   0,  -5,  12,  32,  23, -17,  -2,  -2,   0,   1, -12, -56, -50,  60, -29,  24, -10,   5,  -8,  11,  15, -31,  16, -16,   8,  -3,   5,  13,  11,  -2,  -6,   0,  -3,   2 },
     {   -5,  -6, -12, -39, -62, -11,  42,  -3,   3,   5,   1,  17,  50,  26, -25,  -2,   0,   0,  -1,  12,  17,  -8,   2,  13,   1,   7,   7,  -7, -15,  -7,  -1,   5,   1,  -4, -16,  15,  37, -12,  17, -13,  -1,  -6,  -8,   0, -10,   1,   0,   3,  -2, -11, -19,  -4,   8,   4,  -1,  -3,  -8,  -7, -14,   1,  -5,  -4,  -3,   0 },
     {   -5,  -8, -16, -25, -18, -25, -39,  -2,  -4,  -6, -21, -33,  -2,  15, -15,  -1,  -5,  -9, -32, -45,   1,  18, -14,  22,  -2,  -3,   0,   6,   3, -11, -15,  -5,   3,   9,  19,   5, -12,  24, -11,   3,   4,  10,  28,  -3, -34,  20, -23,  19,   2,   3,  -1, -24,   7,   3,  14, -10,  -3,  -8, -17,   0,   7,   7, -10,   5 },
     {   -1,  -2,  -1,  -3, -16,  -3,  18,  -6,  -1,  -6,  -3,  13,   5, -11,   4,  -8,   0,  -7,   5,  27,   0,   5,  23, -24,  -1,   1,  11, -12, -27,  19, -21,  25,   5,  19,  57,  19, -16,  47, -40,  21,   4,   8,   8, -15,  10, -13,  12,  -7,   4,  -1, -17,  -1,  14, -20,  21,  -3,   4,  -5,  -4,  -8,   5, -10,   6,   1 },
     {    0,  -1,  -1,  11,  15,   2, -10,   0,  -2,  -6, -12,  -2,   4,   8,   8,  -9,  -1,  -6,  -9,   4,   0, -12,  -8,   7,  -1,  -2, -10,  -1,   2,   5,  14,  -8,   1,  10,   8,   5,  -7,   7,  -5,   3,   3,  -8, -18,  15,  -9,  18, -18,   6, -17, -61, -69,  50, -22,  20,   3,   0, -11, -12,  -1, -16,  10,  -9,   7,  -6 },
   },
   {
     {  100, -23, -14,  -4,  -3,  -1,  -1,   0,   3,   5,  -3,   0,  -1,   0,   0,   0, -71,  13,  12,   2,   3,   1,   1,   0,   1,  -6,   2,   0,   0,   0,   0,   0,  14,  -1,  -4,   0,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -5,  86, -28,  -8,  -6,  -3,  -2,  -1, -10,   1,   8,  -2,   0,  -1,   0,   0, -33, -68,  29,  10,   6,   3,   1,   1,   4,   0, -10,   3,   0,   1,   0,   0,  20,  21, -10,  -5,  -2,  -1,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,  -4,  -4,   2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {   11,  17,  -9,  -3,  -2,   0,   0,  -1,  91, -25,  -9,  -3,  -3,  -1,  -1,   0,   8,  -5,   1,   2,   0,   0,  -1,   0, -76,  21,  10,   1,   3,   0,   1,   0, -10,  -6,   4,   0,   1,   0,   1,   0,  18,  -6,  -4,   0,  -1,   0,   0,   0,   1,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0 },
     {  -68,  -5,   6,   9,   2,   2,   1,   1,  11,  -2,  -2,   0,   0,   0,   0,   0, -73,  41,   9,  -4,   1,   0,   1,   0, -16, -13,   7,   1,   1,   0,   1,   0,  55, -23, -10,   2,  -1,   0,  -1,   0,   3,   8,  -4,  -1,   0,   0,   0,   0,  -8,   4,   3,  -1,   0,   0,   0,   0,   1,  -2,   1,   1,   0,   0,   0,   0 },
     {   -7, -20, -78,  33,  11,   5,   2,   2,   0,  -6,   7,  -5,   1,   0,   0,   0,  22,  18,  72, -31, -11,  -4,  -2,  -2,   5,   9, -12,   9,  -2,   0,   0,   0,  -9,  -7, -26,  11,   5,   1,   1,   0,  -5,  -2,   6,  -4,   1,   0,   0,   0,  -1,   2,   5,  -2,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {    6,  12,  -7,  -1,  -1,   0,  -1,   0, -70,  36,   1,   0,   0,   0,   0,   0,   4,   6,   4,  -1,  -1,  -1,   0,   0, -73,  -9,  18,   5,   4,   2,   2,   1, -15, -28,   2,   3,   3,   1,   1,   0,  51,  -4, -14,  -2,  -2,   0,  -1,   0,   6,  11,  -1,  -2,  -1,  -1,   0,   0,  -5,   0,   3,   0,   0,   0,   0,   0 },
     {    0,  -6,   3,   7,   0,  -1,   0,   0, -37, -56,  23,  11,   5,   4,   1,   1, -14,   3, -10,  -3,   0,   4,   0,   0,  -9,  88, -22, -17,  -4,  -4,   0,  -2,  14,  -2,  14,  -2,   0,  -3,  -1,   0,   8, -37,   8,  10,   1,   3,   0,   1,  -6,   0,  -6,   2,   0,   1,   1,   0,   1,   5,  -1,  -2,   1,  -1,   0,   0 },
     {   26,  -9,   2,  -1,  -1,   1,  -1,   0,  -5, -12,   5,   1,   1,  -1,   2,   0,  56, -16,  -6,   0,   0,  -1,   0,   0,  -9,  -7,  -2,   0,   1,   1,   0,   0,  94, -16, -17,  -3,  -3,  -1,  -1,   0,  10,  26,  -4,  -2,  -2,  -1,   0,   0, -37,   8,  10,   0,   1,   0,   0,   0,   1,  -8,   0,   2,   0,   0,   0,   0 },
     {   10,  15,  24,  76, -20, -14,  -6,  -3,   3,   6,   6,  -9,   5,   0,   0,  -1,  -6, -13, -25, -75,  19,  15,   5,   3,  -5,  -8, -10,  15,  -8,   0,   1,   1,   1,   5,  13,  29,  -6,  -8,  -2,  -2,   0,   4,   4,  -7,   4,   1,  -1,   0,  -1,  -2,  -4,  -7,   1,   1,   1,   1,   0,  -2,   0,   2,  -1,   0,   0,   0 },
     {  -15, -54,  14,  11,   6,   1,   3,   0, -13,   5, -15,   0,   3,   1,   0,   0,  -8, -32,  14,   2,   1,   3,  -1,   0, -38,   2,  16,   6,  -2,   1,   0,   0,   8,  82, -23, -13,  -4,  -4,   0,  -2,   0,  -3,  10,  -4,   1,  -1,   1,   0,  -5, -40,   9,   9,   1,   2,   0,   1,  -1,   2,  -6,   0,   1,   0,   0,   0 },
     {   -1,   5, -12,   4,   0,   1,  -1,   0,  -2, -11, -57,  17,  14,   1,   2,   2,   1, -11,   4, -15,   3,   1,   3,   1,  30,  17,  79, -27, -22,  -1,  -4,  -1,   6,   2,   3,  21,  -4,  -2,  -2,   0,  16, -11, -42,  13,  11,   0,   1,   0,  -2,   5,  -3, -11,   2,   1,   0,   0,  -6,   1,   9,  -2,  -3,   0,  -1,   0 },
     {    7,  29, -11,  -4,   0,   0,  -2,   0, -21,   9, -11,   4,   3,   1,   1,   0,  10,  30,  -9,  -2,  -3,  -1,   0,  -1, -38,  14,  26,  -4,  -3,   0,  -1,   0,   3,   0,   8,   5,  -2,   0,   0,   0, -98,  13,   7,   4,   7,   0,   2,   0, -14, -22,   1,  -1,   3,   1,   0,   0,  23,  -7,  -6,   0,  -2,   1,   0,   0 },
     {   -6, -11, -12,  -7, -75,  28,  18,  -3,  -2,  -1,  -1,  -6,  10,  -6,   0,   1,   5,   7,   9,  11,  76, -26, -20,   1,   0,   4,  -1,   8, -14,   9,   1,  -2,  -3,  -3,   3, -10, -33,   9,  11,   1,   0,   1,   0,  -5,   6,  -4,   0,   2,  -1,  -1,  -1,   3,  10,  -3,  -3,   0,   2,  -2,   0,   2,  -3,   1,  -1,   0 },
     {    2,  23, -11,  -3,  -1,  -4,   3,  -1,  19,  79, -22, -15,  -3,  -2,  -4,  -1,  10,  40, -12,  -5,   0,   1,  -1,  -2,  17,  40, -16,  -1,  -1,  -1,   3,  -1,  20,  26,   2, -12,   1,  -4,  -2,   0,  24, -34,  11,   6,  -1,   1,  -1,   0, -30, -12,   0,   9,   1,   2,   0,  -1,  -7,   5,   0,  -2,   0,   0,   0,   0 },
     {    5,  11,  52, -15, -20,   0,   0,  -1,   2,   5,   4,  11,  -1,   0,   0,   1,  -2,   4,  22,  -9,   3,  -2,  -2,   0,   0,  16,  -2, -15,  -1,   4,  -2,   0, -25, -12, -80,  28,  18,   0,   2,   0,  -8, -10,   5,  -1,   0,   0,   2,   1, -26,  15,  44, -14, -10,   0,  -1,   1,  -5,   0,   0,   3,   1,  -1,   0,  -1 },
     {    0,   1,   9, -10,  -6,   4,   0,   4,  -6, -14,  -9, -50,   2,  14,   5,   1,  -4,  -1,   0,   7, -11,  -5,   1,  -1,  11,  22,  22,  86, -10, -26,  -7,   0,  -2,  -4,   1,  -1,  26,   2,  -4,  -2,  -2,  -3, -15, -42,   7,  18,   2,   1, -13,   5,   1,   1, -10,  -2,   2,   0,   2,  -4,   2,  10,  -1,  -4,  -1,  -1 },
     {    0, -20,  18,  -6,  -2,   5,  -2,   1,   7,  30,  -9,  -9,   0,  -1,  -3,   1,   8, -34,  17,  -2,   2,  -6,   2,  -1,  -6,   0,  -2,   6,  -2,   0,  -2,  -2,  34, -33,  -5,  14,   6,   3,  -1,  -1, -26, -60,  10,   4,   4,   5,   4,   1,  57,   5, -22,  -8,  -5,   0,  -1,   0,  23,  36,  -9,  -6,  -4,  -2,  -1,   0 },
     {    6,  11,  11,  18,  25,  74,  -4, -31,   0,   2,  -1,   2,   4,  -4,   4,   4,  -6,  -4,  -9, -13, -23, -72,   0,  30,  -1,   2,  -4,  -4,  -6,   6,  -6,  -3,   1,   8,   2,   1,   8,  33,   2, -14,   4,   5,   3,   1,   0,  -3,   4,   1,  -3,   2,  -2,   3,  -2,  -9,  -1,   4,  -1,  -1,  -3,   0,   1,   0,  -2,   0 },
     {  -13, -39,   3,   6,   9,   3,   1,   3,   9,  44,  -1,  -3,  -3,   0,  -6,  -2, -17, -54,  14,   9,   0,   2,   4,  -3,  -5,  33, -12,  -7,  -3,   2,   2,   1, -17, -25,  32,   7,  -4,   2,   0,   1, -21,  23, -10,   7,   3,  -3,  -1,   1, -47,  35,  -5,  -8,   7,   0,   0,   0,   5, -30,  11,   1,   1,   1,   2,  -1 },
     {    1,   0,   2,   5,  -7,   4,  13,  -9,   4,   3,  18, -13,  59, -19, -15,   1,  -1,  -1,   0,   8, -21,  12, -10,   1,  -4, -11,  -6,  -4, -75,  21,  21,  -3,  -2, -10,   1, -25,  32, -21,  -1,   8,  -1,   1,  -6,  16,  23,  -6, -12,   0,  -1,  -7,   5,  12, -17,  10,   0,  -3,  -3,   4,   1,  -7,   0,   2,   2,   1 },
     {   -1,  -1,  -1,   3,   0,  -4,  -3, -10,   2, -11,  -5,   0,   4,   2,   2,   1,  -3,  -5,   3,   2,   3,   2,   6,   8,   8, -19,   7,  -3,   5,  -5,   1,   3, -16, -17,   8,  -4,   2,  -1,  -2,   1,  28, -15,  11,   6,  -6,   4,   5,  -2, -46, -25,   7,   7,   0,   1,   3,   1,  95,   7, -38,  -9,  -6,  -3,  -3,   1 },
     {    5,   7,  32,   0,  -2,  -6,   1,   0,   9,   7,  77, -17, -20,  -3,  -5,   1,   8,  13,  44,  -3,  -3, -10,   0,  -2,   5,  -5,  34,  -6,   8,  -3,   1,  -1,   3,  12,  18,   0,  -9,   8,  -3,   1,  -2, -16, -44,  18,   7,   0,  -2,  -1, -18, -23, -25,   0,   8,   2,   4,   2,  -6,   2,   8,  -8,  -1,  -1,   3,   0 },
     {   -2,  -2,   0, -35,  13, -19,  39, -40,  -2,  -3,   1,   0,  -7,   3,  -5,   0,   1,   5,   7,  -8,  -4,  13, -41,  48,  -2,  -2,  -4,   3,  -6,   5,  17,   2,   6,  13,  14,  46,  -6, -10,  18, -27,   7,   5,  17, -14,  20,   1, -14,  10,  -5,   3, -17, -29,   6,   4,  -5,   5,  -1,  -1,  -8,  10, -15,  -3,   4,  -3 },
     {    2,   2,   7,  -6,  -2,   2,   5,   1,   6,  20,   8,   4,  -9,  -1, -15,   6,   1,   0,   2,  -5,  -3,  -1,   1,  -2,  10,  42,  -5,  -6,  -3,   0,   9,  -5,   9,  -8, -11,  11,   1,  -3,  -6,   6,  18,  48, -23, -15,   4,  -8,  -5,  -2,  60, -43,  12,   5,  -5,  -1,   1,  -1,  41, -42,   0,  14,   1,   0,   0,  -1 },
     {    5,   8,   8,  27,  19,   2,  69,  16,   2,   4,  -2,   2,   3,  17,   5,   0,  -4,  -8,  -7,  -2,  -9,  -2, -71, -24,   2,   1,   6,  -6,  11, -27, -13,  -7,  -2, -11,  -9, -19,  -3,   4,  27,  15,  -5,   0,  -9,   6, -16,  12,   2,  -2,   2,   2,   5,  19,   2,  -1,  -8,  -6,   5,   5,  -2,  -5,   7,  -3,  -1,   1 },
     {    6,  12,  46, -11, -20,   2,   1,  -1,  -5,  -9, -41,  24,  26,  -5,  -7,  -4,   7,  11,  59, -20, -20,   5,   3,  -2,   1,   1,  -9,  11,   8, -10,   4,   0,   2,  -6,  23, -14,   2,   1,  -5,   7,   5,  14,  30,   3, -26,   4,  -3,  -4,   0,  -4, -39,  19,  16,   1,   4,  -1, -10, -21,  -3,  -6,   7,   4,   0,   2 },
     {    1,   5,   1,  -2, -11,  13,  -3,   1,  -3,   0,   3, -10,   5,  -7,  -1,  -7,   2,  18,  -3,   2,  -6,  14,  -2,   0,  -7,   9, -14,   6,  -5,  -7,  -4,   1,   3,  49, -19,   6,   7,  -6,   1,   4,   1,  28, -29,  24, -17,   3,   5,  -9,  14,  66, -31, -16,  14,   0,   4,   1,  48,  21,   8, -22,   0,   1,  -4,   2 },
     {   -2,  -5, -11,  20, -51,  -3,  16,  -1,   1,   3,  -4,  35, -19,  -6,   0,  -2,   1,  -1,  -1,  13,  -9,  -4,  -9,  12,  -2,  -9,  -8, -17,  27,  -5,  -3,   2,   7,  11,  20,  -8,  77,   3, -19, -10,  -5,  -8, -18, -13,   5,  17,   6,  -1,  -3,  -3,  -6, -16, -31,   5,   6,   6,   1,  -8,   6,   8, -11,  -2,  -2,   4 },
     {    3,   5,   7,  26,  16, -23, -22, -13,   3,   6,   2,  46,   0, -12, -23,   7,   3,   6,   4,  13,   4,   8,  -1,  18,   1,   3,   2,   6, -13,  -2,  26,   0,  -2,   0, -19, -43, -25,  27,  26,  -7,  -8, -14, -18, -48,  10,  24,  -5,  -4,  -1,  20, -12,   8,   4, -15,  -3,  13,  13,   3,  13,  19,  -6,  -8,   2,  -5 },
     {   -1,  -1,  -4,  -8,  -3,   8,  -4,  10,   3,   5,   6,   7,   5,  48, -14, -16,   1,   1,  -1,   0,   3, -15,  12,  -6,  -7, -11, -12, -23, -30, -77,  19,  28,   0,   3,   0,  10,  -7,   4, -21,   1,   4,   5,   9,   2,  23,  37, -13, -23,   1,  -4,   3, -10,  10,   0,   6,  -4,  -7,   2,  -3,  -3,  -9,  -4,   2,   6 },
     {   -1,  -2,   3,   7, -13,  -2,  -5, -20,   4,  13,   6,   8,  21,   2,  11,  -7,  -2,  -3,   8,   5, -17,   4,  -2,  12,   3,  19,   0,  -6,   6, -15, -25,   1,  -4, -15,  11, -20, -12, -16, -11,  -4,   0,  34, -12, -24, -15, -16,  12,   1,  -1, -19,  17, -30,  21,   8,   4,  10, -12,  76, -13,  16,  -5,  26,  -2,  -6 },
     {    0,   0,   1,  12, -13, -18,   3,   2,   2,   8,  21, -19, -22,  -4,  -1,  -5,   1,   0,   7,   4,  -2,   4,  -3,  18,   7,  18,  53, -17, -15,   3,   3,  17,   5,   5,  12, -17,   8,  24, -10,  -4,   6,  24,  58,   1, -14,  -1,  -6,  -5,  17,  36,  16,  12, -19, -16,   7,  -6,  -1,  10, -29,   5,  22, -16,  10,   3 },
   }
 },
 {  //21
   {
     {  122, -33,  -7,  -4,  -2,  -2,  -1,  -1,  -8,   7,  -1,   1,   0,   0,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0,   1,   2,  -1,   0,   0,   0,   0,   0,  -4,   3,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    7, -10,   1,   0,   0,   0,   0,   0, 120, -19, -10,  -3,  -3,  -2,  -2,  -1,  11,  18,  -4,   1,  -1,   0,  -1,   0, -25,  10,   1,   1,   0,   0,   0,   0,  -7,   2,   1,   0,   0,   0,   0,   0,  -7,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    8,  -2,  -2,  -1,   0,   0,   0,   0, -16, -18,   4,   0,   2,   1,   1,   0, 119, -14, -12,  -2,  -4,  -1,  -2,   0,  -2,  29,  -4,   1,  -1,   0,   0,   0, -20,   2,   3,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {    2,  15,  -2,   0,  -1,   0,   0,   0, -26,   9,   3,   1,   1,   0,   0,   0,  -8,  33,  -2,   1,  -1,   0,   0,   0,-108,  19,  17,   1,   4,   1,   2,   0, -23, -31,  10,  -1,   2,   0,   1,   0,   9, -13,   0,   0,   0,   0,   0,   0,   5,  -6,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {   -9,  -4,   2,   0,   1,   0,   0,   0,   3,  25,  -4,   0,  -1,   0,   0,   0, -23,  -5,   3,   1,   1,   1,   1,   0,  28,  40,  -9,  -1,  -3,  -1,  -1,  -1,-101,  21,  23,   1,   5,   0,   2,   0,  -5, -36,   6,   0,   2,   0,   1,   0,   3,  -3,  -1,   0,   0,   0,   0,   0,   2,  -5,   0,   0,   0,   0,   0,   0 },
     {    1, -14,   2,   0,   1,   0,   0,   0,  10,  -7,  -1,   0,   0,   0,   0,   0,   0, -22,   3,   1,   1,   1,   0,   0,  10,  -9,  -2,   0,   0,   0,   0,   0,  -6, -45,   5,   2,   2,   1,   0,   1,  87, -29, -27,   1,  -5,   0,  -2,   0,  29,  48, -19,  -3,  -3,  -1,  -1,  -1, -10,  21,   3,  -2,   1,  -1,   0,  -1 },
     {   -4,  -7,   0,   1,   0,   0,   0,   0,   6,  22,  -1,  -2,  -1,  -1,   0,   0,  -5, -20,   2,   2,   1,   1,   0,   0,   2,  28,  -2,  -1,  -1,  -1,   0,   0,  -7, -32,   3,   2,   1,   1,   0,   0,  36,  53, -18,  -4,  -4,  -1,  -1,  -1, -72,   2,  34,   1,   6,   0,   2,   0,  10, -49,  -2,   6,   0,   2,   0,   1 },
     {   32, 118,  -9,  -2,  -9,  -3,  -4,  -2,  11, -18,   3,   0,   0,   0,   0,   0,  -5, -15,  -3,   0,   0,   0,   0,   0,  14,   4,   0,   0,   0,   0,   0,   0,  -6,  -7,   0,   1,   1,   1,   0,   1,   8,  -1,  -2,   0,   0,   0,   0,   0,  -3,   0,   0,   1,   0,   0,   0,   0,  -1,  -2,  -1,   0,   0,   0,   0,   0 },
     {    2, -11,   5,   1,   0,   0,   0,   0,   6, -23,   7,   1,   1,   0,   0,   0,   3, -32,   8,   3,   1,   1,   0,   1,   5, -28,   8,   3,   2,   1,   1,   1,   0, -41,   7,   5,   2,   2,   0,   1,   2, -43,   5,   7,   1,   2,   0,   1,  12, -74,   1,  11,   0,   3,   0,   2,  39, -42, -26,   9,  -4,   2,  -2,   1 },
     {   -3, -18,  11,  -4,   1,   0,   1,  -1, -17,-108,  15,   6,   8,   4,   3,   2, -33, -20, -15,   5,  -1,   3,   0,   1, -10,  29, -10,   2,  -2,   0,   0,   0, -15,  12,  -5,   0,   0,   0,   0,   0,  -1,  17,  -1,  -2,   0,  -1,   0,  -1,  -7,   7,   4,  -2,   1,  -1,   0,   0,  -2,   4,   2,  -1,   0,   0,   0,   0 },
     {    2,   2,   4,  -2,   1,  -1,   0,   0,  -5, -18,  -2,   3,   1,   2,   0,   1,   7,  20,   8,  -3,  -1,  -1,   0,  -1,  -5, -30,   0,   4,   1,   2,   0,   1,   2,  22,   6,  -3,  -1,  -1,   0,  -1,  -5, -38,   2,   7,   0,   2,   0,   1,  -5,  47,   1,  -7,   0,  -2,   0,  -1, -28, -91,  27,  19,   4,   5,   2,   3 },
     {   -1,  -1,  -2,   0,   0,   0,   0,   0,  -9, -22,  -5,   3,   1,   1,   1,   0,   0,  88, -18,  -7,  -5,  -3,  -1,  -2,  51,  27,  29, -11,   0,  -4,  -1,  -2,   6, -51,  16,   7,   2,   2,   0,   1,  -9,  -5, -13,   4,   0,   0,   0,   1,  -3,  -3,  -1,  -1,   1,   1,   1,   0,   7,  -2,  -2,  -1,   0,   0,   0,   0 },
     {   -1,  -2,  -7,   2,  -2,   0,   2,   0,   5,  17,  10,  -5,   2,  -2,   0,  -2, -11, -43,  -8,   8,  -1,   2,   0,   3,  -4,  68, -11, -12,  -3,  -4,  -1,  -2,  52, -22,  36,   4,   0,   0,   0,   0, -44, -30, -17,  13,   2,   3,   0,   2,  12,  19, -10, -12,  -1,  -2,   0,  -1,  -5, -10,  11,   4,   0,   0,   0,   1 },
     {    0,  -3,  -1,  -1,   0,   1,   0,   1,   2,  -4,  10,   1,   2,   1,   0,   0,   4, -24,  12,   3,   2,   1,   0,   0,  -2, -49,  20,   7,   2,   2,   0,   1, -39, -50,  -8,  24,   0,   5,   0,   3, -65,   9, -35,   8,   2,   2,   1,   1, -24,  29,  -8, -12,   2,  -2,   1,  -1,  -7,  19,  13, -10,   0,  -2,   0,  -1 },
     {   -6,  -5, -79,  22,   2,   6,   2,   1,  -5, -11, -79,  18,   7,   7,   3,   1,  -4, -16, -38,  -2,   7,   4,   3,   1,  -4, -14,  -1, -11,   4,   1,   1,   0,  -5,  -7,  14,  -6,   0,   0,   0,   1,  -5,   1,  13,  -5,  -1,   0,  -1,   1,   0,   3,   8,   0,  -2,   0,  -1,   0,   2,   0,   4,   2,  -1,   0,   0,   0 },
     {    5,   5,  72, -10,   7,  -9,  -1,  -3,  -5,  -1, -85,  19,   3,   9,   3,   2,   0, -14,  19, -13,   0,  -2,   0,  -1,  -1,  22,  19,  13,  -6,   1,  -4,  -1,   6, -11, -26,   4,   6,   4,   1,   1,  -9,  -2,  11,  -5,   0,  -1,  -1,  -1,   2,   5, -12,   3,   0,   1,   0,   0,   0,  -3,   0,  -3,   0,  -1,  -1,   2 },
     {    0,  -2,   1,   1,  -2,   2,   0,   1,  -1,  -8,  -4,   6,  -3,   1,  -3,   1,   2,   0,   6,  -5,   0,  -2,  -1,   0,  -2,  -4,   5,   0,   1,   0,   1,  -1,  17,   9,  11,  -4,   0,  -1,   0,  -1,  14, -60,  12,  21,   1,   5,  -2,   2, -85, -22,  -2,  15,   4,   3,   1,   1, -36,  38,  16, -18,   2,  -4,   2,  -3 },
     {    0,   4,  -1,  -1,  -1,   1,   1,  -1,  -2,  -2,  -1,   3,  -3,   2,  -2,   1,   2,   8,  -1,   0,   2,  -2,   0,  -1,  -5,  -4,  -6,   7,  -3,   1,   1,   2,   8,  19,  -3,  -8,   2,  -2,   1,  -1,  -4, -28,   4,   9,  -1,   1,  -1,   1, -33,  52,  -4, -23,   1,  -5,   1,  -3,  87,   9, -51,  -5,  -7,  -1,  -1,   0 },
     {    3,   3,  56, -18,  -7,  -1,   2,   0,   1,  13,   2,  10,  -8,   2,  -2,   1,  -1,  -4, -90,  21,  12,   2,   1,  -2,  -3, -31, -18, -16,  10,  -1,   2,   2,  -5,  -4,  40, -13,  -4,   0,   1,   3,  -1,   9,   8,   9,  -4,   0,  -1,   1,  -4,   2,  -4,   6,   2,   0,   0,  -1,  -1,   0,   0,  -2,   2,   1,   0,   0 },
     {   -4,  -1, -21, -91, -18,  13,  20,   7,   0,  -2,  -8,  29,  -3, -11,  -3,  -1,   4,   3,  25,  53,  33,  -7,  -7,  -4,   1,   4,   8, -19,  -9,   8,   5,   4,   2,   1,   1,   5,  -2,  -6,   0,  -1,   3,   3,   4,  -4,   7,  -3,   4,   0,   0,  -1,  -8,   0,  -6,   2,   0,   1,   1,  -2,   0,  -4,   3,  -3,  -3,   1 },
     {    2,  -2,  10,  -5,   1,   0,   0,   1,   3,  -9,   9, -19,  -2,   1,  -5,   1,   3, -17,  27, -21,  -1,   9,  -3,   1,   4, -11,  53, -32,  -1,  -2,   0,  -1,   1,  -7,  55, -20,   6,  -1,  -4,  -4,   0,   9,  59, -24,  -6,   4,  -6,  -2,   6,  18,  39, -16,   0,  -4,   0,   0,   4,  12,   7,  -5,   1,   0,   1,   1 },
     {   -3,  -6, -19, -12,  -2,  -5,   4,  -1,   7,  15,  35,  21,   8,   5,  -7,  -3,  -7, -20, -45, -22,   0, -11,   5,   3,   5,  17,  61,  16,  -7,  -2,  -4,  -3,  -2,  -4, -51, -10,   1,   1,   8,   4,   3,  -3,  41,   1,  -4,  -3,  -8,  -1,  -3,   9, -35,   6,   5,   6,   5,   1,  -2, -14,   5,  -1,  -2,  -4,   1,   0 },
     {    3,   1,   9,  53,  13, -10, -10,  -8,   4,   3,  15,  73,  29,  -8, -14,  -3,   4,   3,  13,  51,  24, -12, -10,  -3,   4,   2,  19,  24,  17,  -7,  -8,   1,   4,   7,  19,  13,  13,  -1,  -5,  -1,   6,   6,   9,   3,   9,  -2,   4,  -4,   5,   3,  13,  -4,  -5,   2,  -1,  -5,   3,   5,   5,   1,  -2,   4,  -1,  -2 },
     {    0,  -1,  -5,   2,   6,   2,  -1,  -6,   0,   0,  10,   3,  -8,   7,  -6,  -1,   1,  10,  14,  15,   0,   0,  -6,   0,   5,   4, -66,  14,  11,   1,   1,   3,  -3, -53, -22, -21,   9,   2,  -2,   1, -17,  -3,  69, -12, -15,  -5,  -2,   0,  -3,  20,   3,  19,   0,   3,   2,   0, -18,   1, -17,   3,   8,   0,   2,  -1 },
     {   -1,   3, -10, -22,  63, -48,   3,  -7,   0,   2,  -1,   1,  -6,  22, -13,   2,   0,  -1,  -3,  38, -57,  42, -19,   8,   1,  -1,  13, -10,  13,  -3,   3,  -1,   0,   6,  -7, -17,  17,   3,  -1,  -7,   0,  -4, -14,   3, -18,  10,   0,  -1,  -1,  -2,   3,   4,  -2,   2,   3,  -1,   0,  -1,  -2,  -4,   0,  -1,   4,   0 },
     {   -2,  -1,  -5, -37, -21,  -6,   6,   3,   4,   3,   3,  43,  45,  19,  -9, -11,  -1,   1,  12, -44, -46, -15,  14,  11,   1,  -6, -20,  33,  28,  -1,  -5,  -5,   0,  -3,  33, -29, -12,   0,   8,   3,  -2,  10, -12,   5,   7,   0,  -5,  -3,   3,  -5,   3,  -8,   3,   4,   4,   1,  -1,   5,  -3,   3,  -8,  -3,  -5,  -1 },
     {    0,   0,   7, -13,   0,   4,  -3,   3,   2,   5,   6,  13,   2,  -3,   0,   1,   1,   2, -21,  -8,  16,  13,  -2,   6,   3,   3, -15,   5,  11,   0,   6,   1,   3,  -9, -43, -13,   7,  -1,  -2,   2,  -1, -33, -13, -40,   0,   0,   0,   3,   4, -12,  61, -47,  -2,  -7,  -9,   0,  18,  11,  51,  -2, -13,   3,   0,  -2 },
     {    1,   2,  -4, -19,  52, -40,   8,  -4,   0,  -1,  -2, -33,  53, -38,  27,  -6,   0,   0,  -1, -12,  44, -29,  18, -12,   0,  -1,  -6,  -1,  27, -24,   5,  -9,  -1,  -2,   0,  -5,   1,  -5,   4,   2,  -2,  -1,   1,   6,   2,   5,  -6,   6,  -3,   4,   1,  16,   2,   3,  -3,   2,   4,   4,   3,   6,  -2,   8,  -1,   2 },
     {   -1,  -1,   1, -12,   1,  -2,  -5,   7,   0,   3,   1,   7,   3,  -4,  -8,   4,  -2,  -6, -17,  -4,  -7, -11,   7,  -5,  -1,   3,  16,  -2,   0,   8,  -2,  -7,   2,   4, -18,  26, -18,   9,  -1,  -2,  -6, -20, -11,   5,  11,  -2,   1,  -5,  11,  15,  70,  16,  -6,   0,   3,  -2, -40,   8, -70,  19,  15,   4,   7,   2 },
     {   -2,  -2,  -7, -20,  -4,   5, -51,  18,  -3,  -2, -12, -26,  -5, -41, -12,  18,  -2,   0, -12,  27, -30, -13,   2,   1,   2,  -3,   4,  64, -27, -13,  -6, -12,   3,  -2,  15,  10,   8,   9,   0,  -8,   3,   2,  14,  -4,  13,  21, -11,   4,   2,   0,   6,  -2,  21,  -4,   3,   6,   9,   6,  17,   7,  11,  12,   1,  -1 },
     {   -4,  -6,  -6,   5, -68, -71,  19,   0,  -2,  -3,  -5,  -6, -27, -16,   2,  -2,   0,  -1,   3,  -3,  13,  25, -21, -10,   1,   3,   4,  10,  33,  -8, -10, -13,  -1,   0,  -6,  -6,  20,   5, -17,   5,   0,   2,   0,  16,  21,  11, -11,   1,   5,   4,   7,   9,  16,  -6,   1,   5,  -1,   1,  -1,   2,  -1,   4,  -1,   3 },
     {    2,   1,   6,  14, -20,  16,  62,  12,   0,   2,  -5, -31,  48, -13,  -5,   9,  -1,   0, -10,  17, -14,  -2, -32,  10,   3,   0,  11,  31, -12,  32,  27,  22,   2,   2,   1, -10,  18,  -9, -14,   3,  -4,  -5,  -7,  -9, -20,  -3,   0, -13,   1,   6,  12,  28, -11,   2,  -2,   5,   3,   0,   3, -12,  -1,  -1,   3,   5 },
   },
   {
     {   77, -58,  15,  -7,   3,  -2,   1,   0, -61,  47, -10,   3,  -1,   1,   0,   0,  22, -14,  -2,   3,  -2,   1,  -1,   0,  -9,   3,   4,  -2,   1,   0,   0,   0,   3,   0,  -2,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -72, -15,  47, -18,  12,  -3,   4,  -1,   1,  43, -51,  22,  -9,   3,  -2,   1,  27, -37,  25,  -7,  -1,   2,  -1,   1, -11,  15,  -5,  -3,   5,  -3,   2,  -1,   7,  -6,  -1,   4,  -3,   1,  -1,   0,  -2,   1,   2,  -3,   1,   0,   0,   0,   3,  -1,  -1,   1,  -1,   0,   0,   0,  -1,   1,   1,  -1,   0,   0,   0,   0 },
     {  -25,  43, -40,  21,  -9,   4,  -2,   1, -59,  -5,  34, -12,   6,   0,   1,   0,  67, -29,  -5,  -2,   1,  -2,   0,   0, -30,  19,   0,   1,  -1,   0,   0,   0,  13,  -8,  -1,   0,   0,   0,   0,   0,  -3,   2,   1,   0,   0,   0,   0,   0,   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -54, -27,  17,  11,  -1,   3,   1,   1, -39,  19,  18, -22,  11,  -5,   2,  -1,   3,  37, -47,  26, -11,   2,  -1,   0,  31, -44,  32, -10,  -1,   3,  -2,   1, -16,  20, -10,  -3,   6,  -5,   2,  -1,   9,  -8,   1,   5,  -5,   2,  -1,   0,  -1,   3,   0,  -3,   2,   0,   0,   0,   2,  -1,   0,   1,  -1,   0,   0,   0 },
     {    6, -38,  11,   5,   2,   0,   2,   0,  65, -12,   6, -16,   3,  -2,  -1,  -1,  55,  22, -34,  13,  -6,   0,  -1,   0, -63,   4,  24,  -4,   0,   2,   0,   0,  19,  -3, -10,   0,   2,  -1,   0,   0, -10,   2,   4,   1,  -1,   1,   0,   0,   3,   1,  -2,  -1,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   12,  28,  29, -53,  23, -13,   4,  -3, -30, -51, -13,  52, -17,  10,  -1,   2,  35,  42, -13, -19,   0,   1,  -3,   0,  -4, -20,  10,   5,   3,  -3,   2,   0,  -7,   5,  -2,  -1,  -2,   2,   0,   0,   4,   1,  -1,  -2,   2,  -1,   0,   0,  -1,  -1,   1,   2,  -1,   1,   0,   0,   1,   1,  -1,  -1,   1,   0,   0,   0 },
     {  -15,   4,   3,  -4,   4,  -2,   2,   0, -33,   6,   4,   4,  -3,   3,  -1,   0, -61,  26,  -1,   0,   3,  -2,   2,   0, -49, -10,   8,   3,   0,   1,   1,   0,  78, -20,  -8,  -2,  -3,   0,  -1,   0, -31,  14,   7,  -2,   3,  -1,   1,  -1,  10,  -7,  -3,   1,  -1,   0,   0,   0,  -2,   1,   1,  -1,   0,   0,   0,   0 },
     {   22,  49, -10,  -5,  -7,  -1,  -3,   0,  10,  17, -27,   6,   1,  -1,   0,   0, -11, -31,   6,  21, -12,   8,  -3,   2, -18, -28,  52, -39,  16,  -4,   2,   0,   1,  36, -47,  21,  -1,  -5,   3,  -2,   4, -19,  20,  -1,  -8,   6,  -3,   1,  -2,   2,  -4,  -4,   5,  -2,   0,   0,   1,   0,   0,   2,  -1,  -1,   1,   0 },
     {   -8,   8,  -3,   1,   0,   0,   1,   0, -15,  19,  -3,   1,  -2,   1,  -1,   0, -34,  25,   0,   1,   0,  -1,   1,  -1, -73,  21,   9,  -1,   3,  -1,   2,  -1, -69, -22,  19,   1,   2,   1,   1,   1,  44, -20, -10,   0,  -1,   1,  -1,   0,  -3,   4,   4,  -1,   0,  -1,   0,   0,   3,  -3,  -1,   1,   0,   0,   0,   0 },
     {    7,  15,  44, -26,  -1,  -4,   2,  -4, -11, -32,   5, -17,  30, -10,   5,   0, -15, -22, -28,  53, -32,  15,  -3,   3,  13,  55,  -2, -29,   6,  -1,  -3,   1,   3, -32,   7,   9,   6,  -5,   3,  -2,  -1,   8,  -2,  -2,  -5,   3,  -1,   0,  -1,  -2,  -1,   0,   2,   0,   0,   0,   1,   0,   0,   1,  -1,   0,   0,   0 },
     {   -7, -10, -31, -18,  59, -36,  14,  -4,  11,  17,  37,   7, -59,  31, -11,   1, -11, -23, -28,  16,  15,  -3,  -2,   4,  13,  17,  14, -20,   7,  -9,   5,  -3,  -3,  -9,   0,   9,  -7,   7,  -2,   0,  -1,   3,  -6,   0,   3,  -4,   0,   0,   3,   1,   3,  -2,  -1,   2,   0,   0,  -3,  -2,  -1,   2,   1,  -1,   1,   0 },
     {  -11, -30,  21,  -8,  13,  -8,   6,  -3, -13, -38,  32,  -9,   2,   3,   0,   1, -17, -24,  11,  -4,   7,  -2,   1,   1, -21,  19, -28,  18, -10,   3,   1,   0,   7,  60, -27,   0,  -3,  -1,  -3,   1,  16, -54,  31,  -2,  -1,   3,   0,   1,  -7,  14,  -9,  -3,   3,  -1,   0,  -1,   2,   0,   0,   3,  -2,   0,   1,   0 },
     {    1,  -6,   1,   2,   2,  -2,   1,  -2,   2,  -6,   0,  -2,   2,   0,  -1,   0,   2, -12,   1,  -4,   1,   0,   1,   0,   3, -15,   2,   2,   1,   0,   0,   0,  44, -16, -14,   3,  -4,   1,  -1,   0,  86,  25, -22,  -2,  -2,  -1,  -2,   0, -63,  20,  22,  -2,   3,  -1,   1,   0,   6, -11,  -6,   2,  -1,   1,   0,   0 },
     {   10,  27,  20, -12,   3,  -6,  -1,  -1,   6,  11,  -6,  -7,  -6,   6,   0,  -2,  -6, -34, -23,  12,  20, -13,   5,   0,  -9, -11,  13,  30, -48,  27,  -8,   2,   1,  11,  13, -55,  47, -13,  -2,   2,  -2,   0, -15,  33, -14,  -9,  11,  -6,  -2,  -4,   6,  -7,  -5,  12,  -7,   2,   1,  -2,  -2,   0,   5,  -5,   1,   0 },
     {    3,  -7,  -3,   0,  -1,   2,   0,   0,   4, -12,  -3,   1,   1,   0,   0,   0,   6, -14,   4,   0,   3,  -1,   1,   0,  12, -25,   7,  -3,   1,   1,   0,  -1,  23, -43,   6,  -3,   3,   0,   0,   3,  48, -45, -10,   5,  -7,   3,   0,  -1,  77,  15, -34,   2,  -4,   0,  -3,   0,   3,  26,   0,  -2,  -1,  -1,   0,   0 },
     {    2,   8, -16, -16,  17,  -7,   3,   1,   7,  15,   6,   2,   4, -10,   3,  -1,   6,  -1,   9,  25, -49,  30, -11,   3, -16, -30, -34,   1,  40, -17,   1,   3,  17,  34,  52, -33,  -9,  -5,   5,  -5,   1, -15, -23,  18,   6,   3,  -2,   2,  -4,   4,   4,   0,  -7,  -1,   2,   0,   0,  -1,   3,  -2,   1,   2,  -2,   0 },
     {    8,  12,  24,  40,  13, -45,   1,  -2, -12, -20, -30, -52, -15,  48,  -3,   1,  12,  19,  38,  31,   4, -25,  -6,   0,  -6, -20, -21, -18,   1,   7,   7,   0,  -3,   0,   6,   6,   0,   1,  -6,   1,  -2,   6,  -3,   1,   1,  -3,   3,  -1,  -1,  -3,  -1,  -3,   0,   2,   0,   0,   1,   5,  -1,   1,   0,  -2,   0,   0 },
     {   -7, -29,  16,  -6,   0,   2,   0,  -1,  -6, -40,  32,  -4,   7,  -9,   0,   1, -12, -53,  29,  -7,   4,   7,   1,  -1, -23, -53,  22,  -2,  13,  -9,  -1,   3, -33, -16,  -2,   4,  -7,   9,   1,  -1, -22,  34, -29,   7,   0,   0,  -2,   0,   4, -16,   7,   3,  -2,   1,   1,   1,   3,  -3,  -6,  -1,   2,   3,  -2,   1 },
     {  -14, -43, -51,  11,  13,   8,   3,   3, -14, -38, -32,  30,   1,   1,   0,   4,  -2,  11,  32,  18, -18,  -2,   2,  -1,   3,  26,  20, -27,  -3,  14,  -8,   3,   2,   6,  -8, -21,  31, -19,   7,  -4,   0,   5,  -4,  26, -23,   1,   6,  -3,   2,  -5,  12,  -9,   0,  12,  -8,   3,   1,  -6,   2,   1,   4,  -7,   3,   1 },
     {    0,  -1,  -1,  -4,   2,   3,   3,   1,   3,   2,  11,   6,  -4,   0,  -5,  -1,   7,   1,  19,   6,   0,  -4,   0,   3,   6, -18,  10,  -2,   2,   6,   0,  -3,  -2, -49,  13,  -6,  14,  -6,   4,   3, -40, -44,  21,   0, -10,  10,  -1,  -1, -50,  50, -21,  -3,  10,  -8,   1,  -4,  44, -20,  -4,   3,  -7,   6,   0,  -1 },
     {    3,  -3,  21,  49, -17,  10,  -6,  -1,  -8, -25, -22,  16, -10,  15,   0,  -4, -11, -19, -38, -30,  23,  -5,   5,   7,   2,  17,  28, -27,  21, -17,   3,  -1,  14,  21,  47,  -1, -19,   8,   2,  -5,  -9, -24, -24,  18,  12,   0,  -3,   2, -11,   8,   2,  -6,  -6,  -2,   2,   2,   9,  -6,  -1,  -3,   1,  -1,   0,   0 },
     {   -1,  -3,  -3,  -2,   6,   3,  -6,  -4,   2,   3,   3,  -2,  -2,  -6,   4,  -7,   1,   2,   1,   0,   6,   0,  -3,   3,   3,   2,   2,  -8,   4,  -2,  -1,  -9,   5,  -5,   5,  -6,   5,   5,  -1,   5,   8, -18,  15,   6,   0,   2,  -3,   2, -30, -78,  -6,  22,  -1,   4,   4,   5,  51,  58, -34, -19,   2,  -5,   0,  -2 },
     {   -1,  -1,  -8,  -8, -21, -15,  65, -34,   3,   5,   1,   5,  37,  22, -62,  29,   1,  -2,  -2,  -6, -12, -13,  41,  -7,   3,  -4,   7,   3,   8,   5, -16,  -3,   1,  -1,   9,   5,  -7,   5,  -6,  10,  -1,  -8,   5,   3,  10,  -5,  10,  -6,  -2,  -9,  10,  -2,  -7,   4,  -5,   7,  -1,   1,  -3,  -4,   2,  -6,   3,  -2 },
     {   -5, -10, -29,   9, -10, -13,   1, -13,  -8, -18, -48,   5, -12,   1,  -1,  -8, -11, -26, -59,  -1, -19,   0,  -2,  -9, -12, -24, -50,   7,  -4,  -1,  -8,  14, -17, -16, -27, -10,  -8,  -4,  11,  -2,  -7,   8,  15,  -4,  14,   5,  -3,  -1,  -3,  10, -15,   5,   8,  -3,  11,   7,   5,  -3, -29,  -6,   5,   4,  -5,  -7 },
     {    2,   2,   9,  34,  -3, -40,  26, -11,  -2,  -3,  -4,  -5,  -2,   2,  -6,   3,  -4,   1,   1, -28, -22,  51, -36,  10,   8,  15,  26,  40,   3, -41,  25, -11,  -3, -11, -17, -21,  22,   0,   6,   1,   0,   4,  11,   1, -21,   8,  -9,   2,   2,  -3,   0,   3,   7,  -3,   0,  -3,  -2,   4,   5,   3,   2,   1,   1,   1 },
     {    2,   6,  17,   3,  -9,  -4,   0,  -4,   4,  12,  23,   4, -10,   7,  -5,  -6,  -1,   2, -11, -35,  -4,  -8,   2,   3,  -5,  -4, -32, -20,  24,   9,  -9,   1,   0,   2,  -3,  25,  13, -50,  19,  -4,   0,   9, -10,  16, -54,  40,   9, -14,   2,  -4,  -4,  -9,  24,  10, -32,  16,  -1,   5,  -2,  -3,  -1, -13,  12,   2 },
     {    0,   2,   3,  -6, -14,   2,  10,  13,   0,   4,   6,   0,  -3,  -5, -17,   0,  -2,   4,   1,  -4,  -7, -12,  -9,  -8,  -3,  13,  -4, -18,  -9, -11,  31,  -3,   0,  28,  -5, -15,  -4,  -1, -18,  22,   6,  36, -12,  -5,  -4,  -1,  27,  -7,  31,  25,  -5,  14,   0, -10,  -5, -10,  73, -10, -28,  18, -13,   9, -16,   8 },
     {    1,   5,   8,  -5,  -9,  -5,  -6,  -8,   3,   9,  11,  -3,  -4,  -6,   5,  -2,   3,  12,   9, -10, -10,  -3, -11, -15,   5,   6,  -9,  -9, -10, -15,   2,  16,   4, -15, -28,   8,   2,   9,   4, -12, -16, -34, -24,  29,  13, -11,  -1,   5,  12,  14,  67, -36,  -7,   7,   6,   0,   0,  12, -54,  11,  17,  -3,   1,  -5 },
     {    1,   1,  -2,  16,  33, -20,   5, -16,   2,   3,   2,   2,  21, -11,  19,   7,  -3,  -7,  -5, -40, -32,   8, -16,  -9,  -4,  -9,   3, -18, -27,  36,  -6,   7,   4,   9,  24,  44,  37, -10,  -4, -14,  -1,  -2,  -7, -31,  14, -27,  12,  17,   3,   3,  -3,   5, -10,   8,   5,  -7,  14,  -3,  -1, -13,   4,   5, -14,   2 },
     {    0,   5,   7,   9,  11,   6,   1,  -2,  -3,   3,   7,   2,  -2,  -8, -12,  -3,  -4,   4,   6,  -3, -10, -15,  12,  -3,  -2,  13,   6,  -1,  10,  -6, -14,   8,  -7,  12, -14, -14,   6,  17,   0, -26,  -8,  26, -11,   5,  -2,  -6,   3,   0, -12,  58, -18,   9,  -2,  -1,   1,  -2,  -8,  79,   8, -40,   9,   9,   6, -10 },
     {    2,  -1,  24,  31, -44,  25,  -2,  -3,   2,  -1,  29,  44, -56,  20, -18,  11,   0,  -3,   8,  29, -34,  10,  -1,   1,  -3,  -7,  -9,  -6, -19,   8,  -3,  -2,   1,  -2,   1,  -4,  15, -14,   4,   1,   5,  -5,   2, -33,   7, -16,   0,  -4,  -1, -11,   9,  10,  -1,  -3,  11, -11, -13,   8,   1,   0,  -1,  11,  -3,   4 },
     {   -2,  -5,  -3, -18, -13,   1,  -8,  18,   4,   5,  11,  -4,  18,  61,  12, -31,  -1,  -1,   7, -15, -19,   2,  -1,  28,   2,   4,  14,   0, -11, -13, -11,  44,  -2,  -1,  -1, -13,  -2, -18,  10, -21,   6,   9,  11, -14,  20,   4,  -5, -11,   3,  -2,  -6,  -5, -41, -25, -12,  16,  15,  -7,  -2,  -9,   6, -17,  16, -23 },
   },
   {
     {  104, -30, -11,  -4,  -3,  -1,  -1,  -1,  12,  10,  -7,   0,  -2,   0,  -1,   0, -59,  19,   7,   1,   2,   0,   1,   0, -10,  -7,   5,   0,   1,   0,   0,   0,   6,  -4,  -1,   1,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    7,  84, -53,   9, -11,   1,  -4,   0,  19,   8,   2,  -7,   1,  -2,   0,  -1, -24, -52,  41,  -6,   7,   0,   2,   1, -20, -10,  -1,   7,   0,   2,   0,   1,   4,   6,  -9,   1,   0,   0,   0,   0,   4,   1,   0,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -1,  -5,   4,   0,   1,   0,   0,   0,  99, -26, -12,  -1,  -4,   0,  -1,   0,  27,  14, -11,   0,  -2,  -1,  -1,   0, -59,  21,   9,  -1,   3,   0,   1,   0, -22,  -6,   7,   1,   1,   1,   0,   0,   5,  -5,  -2,   1,  -1,   0,   0,   0,   3,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   62,  -2,  -9,  -6,  -1,  -2,   0,  -1, -15,  -1,   3,  -1,   0,   0,   0,   0,  81, -45,  -3,   3,  -3,   1,  -1,   0,  18,  18, -12,   2,  -3,   0,  -1,   0, -43,  27,   5,  -3,   2,  -1,   1,   0,  -8,  -5,   5,   0,   1,   0,   0,   0,   3,  -3,  -1,   1,  -1,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 },
     {   10,  40,  62, -53,  11, -14,   2,  -3,   5,  27,   9,  -4,  -7,  -1,  -2,  -1, -15, -28, -46,  45,  -9,  10,  -1,   2, -17, -28,  -8,   3,   9,   0,   2,   0,  -3,   1,  10, -12,   4,  -1,   0,   0,   5,   5,   0,   0,  -2,   0,   0,   0,   2,   0,  -2,   2,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    0,  -6,   9,  -5,   2,   0,   0,  -1,  33, -65,  31,  -6,   5,   0,   0,   1, -30, -22,  -5,  14,  -2,   4,   0,   2,  45,  31, -46,   7,  -6,  -1,   0,  -1,  31,  35,  -7, -11,   0,  -4,   0,  -1, -17,   6,  15,  -4,   0,   0,   0,   0, -10,  -6,   4,   3,   0,   1,   0,   1,   0,  -2,  -1,   1,   0,   0,   0,   0 },
     {    3,   4,  -1,   2,  -1,   0,  -1,   0, -58, -33,  33,  -2,   5,   1,   1,   1, -15,  -3,  -6,   6,  -1,   2,   0,   1, -68,  59, -15,   2,   0,  -2,   2,  -1,  -5, -15,  16,  -8,   3,  -1,   1,  -1,  30, -21,   0,   2,  -2,   2,  -2,   1,   4,   1,  -4,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0 },
     {   -8,  34, -11,   4,  -6,   1,  -2,   0,   4,  15,  -2,  -1,   0,   0,  -1,   0, -36,  39, -17,   1,   2,   0,   1,  -1,  34,  29,  -5,  -9,   0,  -2,   0,  -1, -70,   6,  39,  -6,   3,   1,   1,   1, -34, -34,  15,   7,   1,   2,   1,   1,  12, -12,  -9,   4,   0,   0,   0,   0,   3,   1,  -2,  -2,   1,   0,   0,   0 },
     {   15,  18,  41,  50, -56,  13, -11,  -1,   4,   6,  22,   0,   3,  -6,   0,  -2,   6,  -4, -39, -42,  50, -12,   7,   1,  -3, -14, -23,  -2,  -2,   8,   0,   3,   8, -10,   7,  12, -15,   3,   0,   0,   0,   6,   1,   1,   1,  -3,   0,  -1,  -2,   2,  -1,  -3,   3,  -1,   0,   0,   0,   0,   1,   0,  -1,   1,   0,   0 },
     {  -24, -46,  32,  11,  -4,   4,   1,   0,   1,  10,  -7,   6,  -1,   0,  -1,   0, -39, -45,  29, -12,  14,  -2,   2,   0, -19, -14,   0,   1,  -2,   3,   0,   1, -55,  52, -18,   3,  -1,  -2,   2,  -1,  19,  -8,   6,  -6,   1,  -1,   0,   0,  19, -10,   2,   1,  -2,   1,  -2,   1,   0,   2,  -2,   1,   0,  -1,   0,   0 },
     {   -2,  -7,   5,  -4,   3,  -1,   1,   0,   8,  22,  57, -35,   2,  -6,  -1,  -2,  11,  29,  33,   1, -12,   1,  -2,  -1,  -7, -20, -47,  49,  -2,   4,   1,   1, -27, -32, -33,   5,  16,   1,   3,   2, -14,  -5,   7, -18,   3,   1,   0,   1,   5,   3,   7,  -3,  -6,   1,   0,  -1,   5,   0,  -1,   3,  -1,  -1,   0,   0 },
     {    2,  -6,  -4,  -1,   2,  -2,   1,   0,  12, -40,  19,   1,   1,   2,   1,   0,  -9, -23,   3,   4,  -1,   3,   1,   0,  42, -28,   9,  -6,   1,  -1,   0,   0, -26, -50,  17,   9,  -1,   3,   0,   3,  48, -15, -39,  13,  -4,   3,  -2,   0,  34,  35, -22,  -6,   0,  -2,  -1,  -1,  -4,  11,   8,  -5,   0,   0,   0,   0 },
     {   -9, -13, -17, -51, -55,  42,   6,   6,  -4,  -9, -12, -23, -11,   3,   8,   1,   4,   4,   8,  40,  45, -41,  -5,  -4,   3,   4,  14,  31,  14,  -5, -11,  -1,   0,   2,   7,  -3, -10,  13,  -1,  -1,   1,   0,  -2,  -7,  -1,   1,   3,   0,   0,  -2,  -3,   0,   2,  -3,   0,   1,  -1,   0,  -1,   0,   0,   0,   0,   0 },
     {    6,  17,  -6,   1,   3,  -1,   1,   0, -25, -66,  18,   9,   5,   4,   2,   0,  -2,  20, -14,   2,  -2,  -2,   1,   0, -21, -46,  33,  -2,   2,   1,   0,   2, -32,  14, -15,   6,   0,   2,   0,   0, -43,  46,  -9,  -7,   6,  -4,   2,  -1,  17, -16,  13,  -6,   0,   0,   0,   1,   7,  -5,   3,   3,  -3,   0,   0,   1 },
     {   -1,  -2,   1,  -6,   7,  -6,   1,   2,  -3,  -5, -24, -48,  42,  -5,   6,   2,  -7, -10, -20, -21,  -4,  17,  -1,   2,   0,   3,  22,  57, -58,   8,  -4,  -2,   6,   8,  30,  26,   0, -19,   1,  -3,   0,   2,  -7,  -9,  23,  -5,  -2,   1,   1,  -2, -12,  -8,   0,   6,   0,   0,   1,  -1,   0,  -2,  -4,   1,   1,   0 },
     {  -11, -26, -54,  31,  -3,   4,  -2,   4,   1,  10,  -8,  -3,   5,  -5,  -2,  -1,  -1,  -3, -39,  23,  -3,   9,  -1,   0, -13, -35, -37,  12,   4,   8,   5,   1,  -2,  11,  34, -51,  12,  -8,   2,   0,  16,  30,  16,  -9, -11,   0,  -2,   0,  17,   8,  -7,  15,  -6,   1,  -1,   0,  -3,   1,   2,  -1,   4,  -1,   0,   1 },
     {   -2, -24,   3,   5,   0,   2,  -1,   3,   1,  -5,  11,   4,   1,  -3,   1,  -1,   5, -33,  15,   5,   2,   1,   1,   1, -18, -25,   9,  -5,   3,   4,   0,   1,  30, -26,  27,  -3,   6,   1,  -1,  -1, -61, -42,  17,  22,   0,   4,   0,   3,  21, -30, -40,  11,  -1,   1,   2,   1,  25,  18, -21,  -8,   0,  -2,  -1,  -1 },
     {   -4,  -5,  -7,  -5, -34, -60,  53,  -8,  -3,  -6,  -5,  -7, -18,  -2,  -2,   7,   2,   4,   4,  -5,  26,  53, -52,   5,   1,   1,   1,   7,  28,   2,   4, -11,   1,   1,   1,   8,  -5, -17,  19,  -1,  -3,  -2,   2,   2,  -9,  -2,   0,   4,   1,  -4,   3,  -3,   0,   6,  -5,   0,   0,   2,   0,  -1,   1,   0,   1,   0 },
     {  -10, -24,   2,   5,   0,   1,   2,   0,   2,  10,   8,  -1,  -1,   1,   1,   1, -15, -41,  11,   2,  -1,   3,   0,   2,  -4,  16,   6,  -5,   5,  -5,   1,  -1, -26, -39,  37,   2,  -1,   2,  -3,   3, -39,  44,  -7,   4,  -6,   0,   2,   0, -60,  32,  18, -11,   4,  -5,   4,  -3,   9, -24,   9,  -1,   0,   1,   0,   1 },
     {   -2,  -1, -11, -17,  10,  -1,  -5,  -7, -10, -18, -65,  21,  -4,  13,  -2,   5,  -9, -18, -25,  -8,  10,   1,  10,   6,  -4,   5, -47,   9,   2,  -1,   4,  -4, -10, -43, -32,  31,  -2,   9,  -1,  -2, -14,  -4,  28, -18,   3,  -1,  -3,   2,  20,  11,  14,  -7,  -2,  -4,  -2,  -1,  19,  -1,  -8,   6,  -4,  -4,  -4,  -1 },
     {    7,  11,  22,   9,  -6,  -1,  -1,  -8,  -2, -29, -14,  12,   9,  10,   4,   6,   4,   2,  16,  -7,   2,   3,   2,   5,  12, -23,  24,  20,  -6,  -9,   3,  -3, -12, -25,   8, -32,  19,  -3,   3,  -3,  28, -20,  41, -23, -18,   8, -10,   2, -45, -19,  13,  36, -11,   3,  -1,   1, -10, -14, -32,  19,  11,  -4,   4,  -2 },
     {   -3,  -5, -19, -43,  23,   8,   3,   0,   1,  -4,  22, -29,  -5,   6,  -4,   2,   3,   7,  -1, -40,   7,   1,   5,  -1,  -5, -24, -14, -25,   6,   1,   1,   2,  -4,   9,  25,  30, -36,  -2,  -9,  -1,  24,  10,  49,  38, -14, -11,  -2,  -5,  -7,  -2,  11,  13,  13,  -3,   2,   1,   0,  -6, -17,  -4,   8,   3,   0,   3 },
     {    2,   3,   6,  15,  -5,   6,  -4,   8,  -4,  -5,  -6, -18, -43,  25,   8,   1,  -3,  -3,  -9, -22, -32,  -2,  15,  -3,   1,  -2,   4,  30,  52, -36, -13,   2,  10,  19,  21,  35,  43,  -7, -20,  -3,   2,   1,   4,  -4, -13,  23,   3,  -4,   3,   4,  -9, -11, -16,   4,  10,   0,  -8,   1,   4,  -4,   0,  -6,   2,   1 },
     {    3,   4,   0,  14,  18,  36,  54, -46,   3,   8,   7,   7,   5,  27,   7,   1,  -3,  -3,  -9,  -6, -15, -33, -46,  53,  -4,  -5,  -8,  -1, -16, -27,  -4,   2,   1,   2,   2,  -1,  -7,   8,  11, -23,  -1,  -1,  -3,   0,   5,  12,   4,  -2,   2,   2,  -7,  -3,   6,  -1,   1,   6,  -2,   0,  -2,   0,   0,  -1,   0,   1 },
     {   -6, -16, -30,   5,  -5,   0,   0,   9,  13,  24,  35,   4,   1,   8,  -1,  -4,  -6, -20, -38,  16,   0,   6,   4,   1,  13,  16,  20, -11,  -1, -17,  -4,   3,  -1, -20, -13,  30,   6,  -2,  -1,  -1,  20,   6, -13, -15,   7,   3,  -5,  -6,   8, -65,  26,   1,  -5,  10,  -1,   2,  15, -20, -31,  19,  -8,   2,   2,   5 },
     {    5,  10,  28,  -5,  -2,  -5,  -7,  -1,   3,   9,  -7,   3,  -4,  20,  -5,   6,   5,   9,  41, -10,   4,  -4,   0,  -2,   8,  28,  -4,   3,  -7, -20,  14,  -4,   8,  -9,  26, -17,  -7, -11,   5,   6,   5,  61,   0,  -8,  -9,  -4, -12,  -3,  58,  -8, -18,  11,   0,   3,  -4,  -2,  30, -20, -23,  10,   2,   2,   1,  -4 },
     {    0,  -1,   2,  -1,  -6,  -1,  15, -10,  -2,   2, -12, -56,  23,  -4,   4,   9,  -6, -10, -20, -22,  -4,   4,  13,  10,   3,  18,  15, -10,  10,   1,  -2,   7,  -3,  -6, -40, -30,  16,  25, -10,  11, -12,   5, -21,  28, -42,  14,   5,   0,  14,   1,  13,  35,  -6, -21,   0,  -6,  15,   2,  -3,  -8,  20, -13,  -5,  -1 },
     {   -1,  -2,  -3,   0,  10,  -2,  -2,   8,   4,  13,  22,   0,  12,  17,  12,   6,  -4,  -8,  -8,   1,   8,  -1,   0,  -1,   2,  19,  29,  -7,   3,   8,  -2,   5,   2,  -3,  -1,  10,  -6,   4,  20,  -6, -12,  18,  30, -37, -21, -14,  -1, -12,  12,  30,   0,  -3, -21,  -8, -10,   5, -31,  75, -20,  13,   1, -16,   3,  -7 },
     {    1,   1,  -2,  25,  25, -43,  -5,   6,   0,   0,  -1,  -2,  12,  26, -14,   0,   1,   3,  -2,  14,  11,  -6,   8, -12,  -1,  -4,  -1,  37,  20, -29,  -8,  15,  -2,  11,   4,  -6, -57,  52,  -5,   0,   5,  -9, -11,  -8,  -4,   1,  12, -12, -16,  -3, -14,  -3,  22, -19,   6,   3,  17,   4, -12,  -2,  -4,   0,  -7,   7 },
     {    0,  -5,   3,  -9,  14,   6,  -8,   3,   0,   0,  -5, -10,  -3,   5,   6, -14,  -1, -12,   3, -21,  -3,   4,   0,   1,   0,   1,  -4,   2,  15,   4,   6,  11,   7, -18,  13,  -9, -25,   8,  21,  -6, -18,  -9, -11, -26, -21, -11,  15,  10,  23, -35,   2,   7,   3,  -6,  -9,   8, -68, -34,  47,  23,   4,   7,  -5,  -7 },
     {    3,   5,   7,   9,  10, -11,   5,  14,  -1,  -3,  -4,   6, -19, -38,  21,   4,  -2,   1,   4,  -1, -14, -17, -12,  12,   4,   7,  -1,   8,   6,  57, -51,  -1,   1,   0,  13,  12,  10,  30,  10, -37,  13,  13,   2,   0,  -8, -16,  34,   0,   6,  -2,  -4,   5, -12, -12,  -3,  18,  19, -19, -21,   8,   3,   0, -11,   2 },
     {   -2,  -1,   2,   8, -35,  10, -17,   7,  -2,   2,  -8,  16,   1,  22,   1,  34,  -2,   4,   5,  23, -34,  26,  13,  34,  -2,   5, -11,  16,  -5,   6,  -9,  10, -11,  -6, -18,  19, -12,  -7,   8,  -3,  -1,  18,  -7,  38,  -9, -30,   3,  -8, -10, -15, -24,  22,  29,   3,  -3,  -2, -37,   9, -16, -17,   2,   2,   6,  -3 },
   }
 },
 {  //22
   {
     { -122,  25,   7,   4,   3,   2,   1,   1,   2, -10,   2,  -1,   1,   0,   0,   0,  21,  -8,   0,  -1,   0,   0,   0,   0,   7,  -2,   0,   0,   0,   0,   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    5,  12,  -2,   0,  -1,   0,   0,   0,-117,  11,  11,   3,   4,   1,   2,   1,   4, -30,   5,  -1,   1,   0,   1,   0,  32, -12,  -5,   0,  -1,   0,   0,   0,  11,   3,  -4,   0,  -1,   0,   0,   0,   3,   3,   0,   0,   0,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {   18,  -3,  -3,  -1,  -1,   0,  -1,   0,  -3, -24,   3,  -1,   1,   0,   1,   0, 106, -12, -16,  -1,  -4,  -1,  -2,   0,   1,  48,  -8,   0,  -2,   0,  -1,   0, -32,  10,  11,  -1,   2,   0,   1,   0,  -3,  -8,   1,   1,   0,   1,   0,   0,  -1,   2,  -3,   0,   0,   0,   0,   0,  -2,   2,   1,  -1,   0,   0,   0,   0 },
     {   -6,  17,   0,   1,   0,   0,   0,   0, -35,   7,   6,   1,   1,   0,   1,   0,  -6,  29,   0,   0,  -1,   0,   0,   0, -83,  18,  24,   0,   5,   0,   2,   0, -20, -54,  21,   1,   4,   1,   1,   1,  24, -34,  -8,   5,  -1,   2,   0,   1,  12,  -2, -11,   1,  -1,   0,   0,   0,   3,   4,  -1,  -2,   0,  -1,   0,   0 },
     {   11,  12,  -4,  -1,  -2,  -1,  -1,   0,  -8, -34,   5,   0,   2,   1,   1,   1,  38,  27, -11,  -1,  -3,  -1,  -1,   0, -41, -32,  16,   2,   4,   1,   2,   1,  67, -11, -35,   2,  -5,   1,  -2,   0, -14,  55,   3,  -7,   0,  -2,   0,  -1,  -9, -11,  13,   1,   1,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0 },
     {    5, -16,   0,   1,   1,   1,   0,   0,  12, -16,  -4,   1,   0,   1,   0,   0,  11, -27,  -2,   2,   1,   1,   0,   0,  16, -20,  -7,   2,   0,   1,   0,   0,  22, -42, -12,   5,   0,   2,   0,   1,  45,  -2, -55,   9,  -6,   2,  -2,   1,   4,  64, -28, -11,  -2,  -3,   0,  -2,  -9,  27,  19, -14,   1,  -3,   0,  -1 },
     {    3,   5,  -1,  -1,   0,   0,   0,   0,  -4, -25,  -2,   3,   0,   2,   0,   1,   4,  21,  -1,  -3,   0,  -1,   0,   0,   0, -32,   0,   4,   0,   2,   0,   1,   9,  36,  -9,  -5,  -1,  -1,   0,  -1, -20, -46,  26,   8,   4,   3,   1,   1,  29, -16, -67,  12,  -3,   3,  -1,   1,  -3,  53,  12, -25,  -1,  -6,   0,  -3 },
     {   26, 118,  -6,  -1,  -7,  -4,  -4,  -2,  23,  -5,  11,  -2,   1,  -1,   0,  -1,  -5, -26,   1,  -1,   1,   0,   1,   0,   9,  -4,  -1,   1,   0,   0,   0,   0,  -7,  -4,   0,   1,   0,   1,   0,   0,   1,  -2,  -3,   0,  -1,   0,   0,   0,  -1,   4,  -1,  -1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0 },
     {   -9,  20, -24,   3,  -1,   1,  -1,   0, -17,  35, -33,   3,  -2,   1,  -1,   0, -13,  45, -29,   2,  -2,   0,  -1,  -1, -15,  39, -27,   0,  -1,  -1,   0,  -1, -10,  40, -21,  -5,   0,  -2,   0,  -1,  -7,  34, -16,  -6,   1,  -2,   0,  -1, -11,  43,  -8, -12,   2,  -3,   0,  -2, -12,  23,  19, -15,   2,  -3,   1,  -1 },
     {    1,   5,  -3,  -4,  -1,   0,   0,   0,  13,  98,  -8,  -7,  -7,  -4,  -2,  -2,  50,  19,  32,  -7,   0,  -4,  -1,  -2,  11, -38,  19,   4,   1,   1,   0,   0,   7, -13,   0,   5,  -1,   1,   0,   0, -11,  -8,  -4,   8,   1,   1,   0,   0,  -6,  -1,  -7,   1,   1,   1,   0,   0,  -1,   1,  -1,  -2,   1,   0,   0,   0 },
     {   -4,  -9, -19,   5,   0,   2,   0,   1,  10,  26,  22,  -7,   0,  -2,  -1,  -1, -10, -32, -27,   6,   1,   2,   1,   1,  12,  33,  21,  -8,  -1,  -2,  -1,  -1, -10, -36, -26,  12,   2,   4,   1,   1,  12,  39,  17, -17,  -2,  -4,  -1,  -2,  -6, -44, -23,  22,   3,   5,   1,   2,   9,  50,  -7, -28,  -2,  -6,   0,  -3 },
     {    2,  10,   1,  -3,  -2,  -2,   0,   0,  -8, -25,  -8,   1,  -1,   2,   0,   1,  -5,  70, -19,  -9,  -3,  -3,   0,  -2,  70,  16,  49, -11,   0,  -5,  -1,  -2,  -2, -35,  15,  18,   2,   3,   0,   1, -16,  -4, -16,   6,   1,   1,   0,   1,  -3,  -2,   1,  -1,   1,   0,   0,   0,  -2,  -3,  -1,  -1,  -1,   1,   0,   0 },
     {    0,   0,   3,  -1,   0,  -1,   0,   0,  -5, -14,  -8,   4,  -2,   1,  -1,   0,   7,  20,  10,  -6,   3,  -3,   1,  -1,   5, -43,  12,  14,   1,   4,  -1,   2, -75,   3, -55,   7,   2,   3,   3,   2,  37,  14,  14, -35,  -4,  -6,  -1,  -2,  25,   9,  19,   7,  -6,   1,  -3,   0, -10,   2,  -9,  10,   4,   1,   1,   0 },
     {    0,   3,  -9,   5,   0,   1,   0,   0,  -1,   1, -23,   6,   1,   2,   0,   1,   0,   4, -13,   5,   0,   1,   0,   0,   2,   2, -25,   5,   1,   1,  -1,   1,   5,  -1, -11,   3,   1,   0,   0,   0,   3,  -5, -30,   8,   3,   2,   1,   1,  -1, -10, -35,  16,   2,   5,   0,   1, -13,  -8,-106,  24,   9,   4,   4,   2 },
     {   -5,  -5, -84, -10,  -7,   5,   3,   3,   2, -14,  47, -16,  -2,  -7,  -2,  -2,   0,  28,  44,  32,   1,   0,  -3,  -2,  13,   5, -36,  10,   8,   3,   3,   1,  -2, -18,  -4, -18,   2,   1,   2,   1,  -5,  -4,   2,   2,  -6,   1,   0,   0,   4,   2,   2,   1,   0,  -2,   1,   0,   1,  -5,   1,   1,   0,   1,  -1,   1 },
     {    3,   7,  -1,  -9,   0,  -1,  -1,  -1,   4,   9,  -8,  -9,  -1,  -1,   2,  -1,   5,  19,   1, -11,  -5,  -4,   0,  -1,  15,  20,   2, -14,  -4,  -5,  -1,  -1,  44,  13,  24, -28,  -4,  -7,  -1,  -2,  79,  -3,  53, -14, -12,  -1,  -4,  -1,  13,  13,   7,  27,  -7,   4,  -2,   1, -12,   5,  -9,  15,   3,   1,   1,   1 },
     {    3,   3,  70,  -3,   0,  -5,  -3,  -2,   2,  17,  57,   5,   0,  -2,  -4,  -2,   3,  32, -33,  29,   0,   1,   1,  -1,  11,  -3, -55,   0,  10,   2,   4,   0,   0, -23, -15, -23,   4,   2,   2,   0,  -4,  -8,   4,  -9,  -5,   3,  -1,   1,  -4,   1,   2,   1,  -5,   0,   0,   0,  -1,   2,   0,   0,   0,  -1,   1,   0 },
     {   -2,  -3,   3,  -1,  -5,   0,  -1,   0,  -4, -14,   4,  10,   2,   0,   0,   1,  -3,   1,   8,  -1,   4,  -1,   2,  -3,  -5, -16,   8,   9,   0,   4,  -1,   2, -21,  -6,   9,  -8,   1,  -3,   0,   0,   2,   7,  32,  16,  -9,   1,  -4,  -1, -91,  20, -28,  33,  14,   1,   4,  -1, -34, -21,   1, -26,  16,  -2,   4,   0 },
     {    0,   0,  38, -19,  -1,  -2,  -4,  -2,  -1,  -3,  55, -21,  -5,  -1,  -5,   0,  -6,  11,  56, -13, -10,   0,  -6,  -2,  -7,  35,  27,  16, -17,  -1,  -5,  -2,   3,  40,  -9,  27,  -6,  -1,  -1,  -2,  12,  22, -25,   6,   7,  -3,   3,  -2,   2,  10, -22,  -2,   8,   0,   1,  -1,  -6,   4, -15,  -2,   4,   1,   0,   0 },
     {   -2,   3,  -1, -88,  36,   9,  -2,  -2,   0,   0,  -5,  51, -26,  -3,  -1,  -2,   1,  -8,   3,  39,  11, -11,   0,   0,  -4,   1,  11, -37,  11,   4,  -3,   1,  -2,   5,  -2,   8, -12,   0,  -1,   0,  -2,   2,   1,  -1,   0,   0,   3,   0,   2,   5,  -2,  -1,   0,   0,  -1,   2,   1,   0,  -5,  -2,   2,   1,   0,   0 },
     {   -2,  -6, -31,  20,   5,   1,   0,   0,   3,  11,  56, -17,  -9,  -8,  -2,   0,   2,   0, -60,  11,  15,   8,   4,  -1,  -9, -33,  27, -34,  -7,  -1,   1,   3,  -5,  40,  15,  27, -12,  -3,  -3,  -4,  16, -15, -11,   1,  11,   0,   2,   1,  -6,  16,   5,  -4,  -4,  -2,  -2,  -1,  -6, -13,  -8,   6,   5,   2,   1,   1 },
     {    3,   1,   0,  70,  22, -11, -14,  -9,   3,   2,  11,  77,  19, -15, -14, -13,   4,   4,  28,  24,  15,  -8, -10,  -8,   4,  12,  21, -10,  11,   0,  -2,   2,   6,  11,   1, -16,  -3,   3,   2,  -1,   7,   7,  -4,  -9,  -3,   4,   6,   2,   4,   1,  -7,  -7,  -3,   0,   2,   3,   4,  -2,  -2,  -2,  -2,  -1,   1,   1 },
     {    0,   1,  -5,  -2,   3,   0,   4,   1,   1,   5,  -7,  -2,  -4,   3,  -2,   1,   0,   2, -10,   4,  -1,   0,   0,   2,   3,   8,  -3,   1,  -1,   0,  -1,   0,   4,   1, -22,   2,   2,  -1,   3,  -2,  14,   3, -17, -23,  10,   2,   1,  -1,  -2, -22, -64,  27,   5,   8,   0,   3,  17, -60,  50,  48, -16,   6,  -5,   3 },
     {   -1,  -5,   4,  10, -12,   0,  -5,   2,  -1,   5,   3,  -6,  -1,  -2,   1,  -4,  -4, -18,   5,  16,  -3,   7,  -1,  -2,  -5,  15,  13,  -9,   8,  -2,  -4,   0,   7, -17,   0,  17,  -2,   2,  -4,   2, -51,   3,  15, -18,  -7,   1,   0,  -1,  38,  37,  -1,  44, -21,   1,  -5,  -1, -64,  -2,   8,  29,  22,  -5,   6,   1 },
     {   -4,  -5,   7, -18, -82, -55, -28,  -2,  -1,  -2, -16,  -1,   3, -16,   2,  -4,   3,  -1,   3,  40,  40,  27,  13,  -3,   0,   1,  13,   1,   8,  -2,  -1,   2,   1,   9,   3,  -3,  -3,  -5,   2,   0,   9,   0,  -8,  -2,   5,   0,  -3,   2,  -1,  -8,   1,  -3,   6,   4,  -2,  -3,   7,   3,   1,  -4,  -1,   1,   2,   2 },
     {    6,   8,  22,  34,  32,  19,   3,   1,  -7, -13, -34, -52, -40, -19,  -3,   1,   5,   9,  28,  51,  30,  10,   1,  -1,   0,  -4, -16, -30, -11,  -7,  -1,   3,  -3,  -2,  12,  21,   4,  -2,  -4,  -3,   8,  10,   0,  -6,   2,   3,   5,   2,  -9, -12,  -1,   5,   3,  -4,   0,   1,   4,   9,   4,  -8,   3,   1,   1,  -2 },
     {    0,  -1,   8,   2,  -1,  -5,   5,  -6,  -1,  -1,   1, -10,  13,  -2,   8,   0,   0,   1,  -2,   8, -25,   7,  -7,  -2,   0, -24,  -1, -20,  20, -12,   5,   1, -32,  -4,  42, -30, -32,  -6,  -9,   7,  -2,  62,   8,  43,  -1,  -9,  -2,  -6,  35,   6, -29,  -6,  21,   9,   4,   0,   4,  -2,  -2,  10, -11,  -3,  -3,   0 },
     {    0,  -2,   1,  14, -47,  47, -23,  23,   0,  -2,   8,  21, -64,  34, -23,  28,   1,  -3,   7, -14, -12,  -3,   1,   2,  -1,  -4,   4, -23,  15, -23,  19, -10,  -2,   2,   5, -22,  15, -14,  10,  -1,  -2,   4,  -8, -11,   8,  -7,   4,  -2,   0,  -1,  -3,  -2,   3,  -3,   2,  -2,   2,   1,  -4,  -3,   0,   1,  -2,  -1 },
     {    0,   3,  -3,  11, -13,   6,   5,  -3,   0,   4,  -6,  33, -29,   9,  -9,   3,   1,   5, -15,  42, -46,  14,  -5,  -1,   3,  -3,  -9,  53, -30,   7,  -8,  -6,   5,   2,   9,  46, -12,   5,   2,  -4,  19,   2,   9,  37,  -7,  -7,  -2,  -1,  14, -10,  14,  19,   6,  -6,  -2,  -1,   3,  -6,  11,  -7,   5,  -2,   3,   0 },
     {    0,  -2,   1,   8,   1,   1,  -6,   6,   0,   4,   0, -10, -14,   4,   1,   9,  -2,  -9,  -7,  16,   9,   1,   0,   2,   3,  15,  19,   9,  -2,  -5,   1,   0,   6, -13, -33,  10,  11,   7,   0,   4,  -6,  -9,  60,  22, -23,   0,  -2,   1,   0,  27, -26, -72,  18,  -8,   9,  -2,   4, -11, -13,  28, -10,   9,  -1,   2 },
     {    1,   0,   7,  -3, -21,  78, -35,  21,   1,   2,  -4,  -7,  39, -18,  24, -15,   0,   0,  -8,  21,  12, -41,  24, -16,   2,   1,  10,  15, -11,  24, -22,  10,   1,   3,   3,  -8, -12,  -8,   5,  -9,   2,   2,  -3,   2,   4,  -3,  -5,  -4,   8,  -2,  -2,  -4,   6,  -2,  -1,   1,   2,  -6,   2,  -4,   5,  -1,   5,  -1 },
     {    1,   1,   5,  -3,  14,   2,  16,  -8,  -1,   1,  -8, -31,   2,  21,  -4,   1,  -2, -11, -10,  52, -30,  -2, -15,  -3,   3,   2,  48,  38,  34,   8,  22,  -4,   2,  19,   0, -45,  14,  -4,   1,   1,   1,  -6, -16, -14,   4,   8,  -5,   2, -11,   2,   2,  -4, -33,  -7,   1,   5,   1,  -1,  -7,  -7,  -6,  -9,   7,  -1 },
   },
   {
     {   97, -13, -12,  -4,  -4,  -2,  -2,  -1, -77,  18,   9,   2,   3,   1,   1,   0,   1, -13,   2,   1,   1,   0,   0,   0,  13,   2,  -4,   0,  -1,   0,   0,   0,  -3,   2,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -37,  29,  -1,   0,  -1,   0,   0,   0, -57, -21,  12,   3,   5,   2,   2,   1,  93,  -8, -15,  -3,  -5,  -2,  -2,  -1, -28,  19,   4,   0,   1,   0,   0,   0,  -3, -10,   3,   0,   1,   0,   0,   0,   5,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    9,  90, -24,  -7,  -8,  -4,  -2,  -2,  -3, -62,  23,   3,   6,   2,   1,   1, -38, -11,  -2,   6,   2,   2,   1,   1,  32,  15,  -9,  -4,  -2,  -1,  -1,  -1,  -4,   1,   5,  -1,   1,   0,   0,   0,  -1,  -1,   0,   1,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {  -59,  -5,   7,   6,   3,   2,   1,   1, -42,  44,   0,  -4,  -1,  -2,   0,  -1, -18, -52,  13,   4,   4,   2,   2,   1,  66,   4, -21,  -2,  -4,  -1,  -2,   0, -19,  19,   8,  -2,   1,  -1,   0,  -1,   5, -11,   2,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {  -28,   9,  20,  -4,  -1,  -1,   0,   0, -57, -13, -16,  12,   3,   3,   1,   1, -27,  57,  10, -14,  -2,  -4,   0,  -2,   7, -62,  15,   9,   2,   2,   1,   1,  37,  11, -25,  -2,  -2,   0,  -1,   0, -15,  12,  10,  -3,   1,  -1,   0,   0,   3,  -7,   0,   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
     {   -8,  18, -21,   5,   0,   0,   0,   1,  -5,  30,  14, -12,  -2,  -3,   0,  -1, -21, -52,  12,  13,   3,   3,   0,   1, -54,   7,  -7,   1,   4,   1,   2,   0,  82,   5,  -8,  -7,  -5,  -2,  -2,  -1, -26,   8,   7,   1,   1,   0,   0,   0,  -4,  -9,   1,   1,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0,   0 },
     {  -16, -24, -81,  27,   8,   7,   3,   3,  -2,   5,  68, -25,  -4,  -4,  -1,  -2,   3,  37,   0,   5,  -6,  -3,  -2,  -1,   5, -25, -19,   6,   5,   2,   2,   1,  -8,   0,   8,  -6,   0,   0,   0,   0,   1,   2,  -4,   2,  -1,   0,   0,   0,   2,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {  -19,   4,   1,   3,   0,   0,   0,  -1, -29,  22,   6,  -1,  -1,   1,   0,   0, -49,  13, -14,   3,   3,   1,   2,   0, -26,  24,  29, -12,   0,  -3,   0,  -1, -10, -74,   8,  17,   3,   4,   1,   2,  40,  13, -33,  -5,  -2,  -1,  -1,   0,  -7,  17,  13,  -6,   0,  -1,   0,  -1,   0,  -8,   3,   3,   0,   0,   0,   0 },
     {    4,  41, -11,   0,  -6,  -1,  -2,   0,   0,  26, -21,  -4,   3,  -1,   0,   0,  -3,  -4,  56,  -6,  -3,  -1,  -1,  -2, -49, -34, -20,  29,   6,   5,   2,   2, -47,  12,  -6, -18,   6,  -1,   3,   0,  46, -17,   8,  -3,  -7,   0,  -2,   0,   1,   8,  -4,   3,   0,  -1,   0,   0,  -4,  -2,   4,   0,   1,   0,   0,   0 },
     {  -11, -13, -28, -88,  -5,  13,  11,   4,   3,   7,  14,  68,  -1, -11,  -7,  -2,   4,   4,  22,  14,  12,  -2,  -3,  -2,  -3,  -2, -15, -29,  -4,   6,   3,   1,   0,  -5,  -3,   9,  -4,  -2,  -1,   0,   0,   0,   2,  -2,   1,   0,   0,   0,   2,   3,   1,   2,   1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 },
     {   -4, -23, -16,  13,   2,   3,   1,   1,  -7, -28, -33,  -7,  11,   2,   3,   1,   4,  16,  76, -12, -16,  -4,  -5,  -2,  16,  51, -31,  16,  -1,  -3,   0,  -1,  16, -42, -10,  -5,   9,   3,   2,   1, -20,   6,   8,  -2,  -4,   1,   0,   0,  -2,   4,   3,   1,   0,  -1,   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0 },
     {   10,   6, -13,   1,  -1,   1,   0,   0,  23,   0,   0,  -4,  -1,  -1,   0,   0,  31, -36,  11,  -3,   3,   0,   0,   0,  37, -39,  16,   8,  -4,   2,  -2,   0,  13, -22, -54,  14,   4,   3,   1,   1,  27,  58, -16, -24,  -4,  -3,  -1,  -1, -14,  11,  28,   0,  -3,  -1,  -1,   0,  -3,  -8,  -2,   7,   1,   0,   0,   0 },
     {   -1, -32,  22,  -2,   1,  -1,   1,   1,  -5, -47,   7,   2,   2,   1,   1,  -1,  -3, -13,  -1,   9,  -1,   2,  -2,   1,  -1, -17, -37,   1,   8,   0,   2,   1,  39,  11,  28, -22, -11,  -2,  -2,  -1,  63,  -6, -21,  21,  -7,   0,  -2,   0, -36,  30, -10,  -5,   8,  -1,   1,   0,  -2,  -8,  12,  -3,  -1,   1,   0,   0 },
     {   -4, -13, -13,  10, -14,  -2,   2,   2,   3,  -4,  19,  43,  27,  -5,  -5,  -3, -12, -17, -14, -87, -11,  17,   9,   4,  -8,   9,   7,  35, -23, -12,  -5,  -1,   8,  14,   2,  19,  23,  -3,  -1,  -2,   8, -14,  -2, -18,  -3,   6,   1,   1,  -8,   2,  -3,   2,  -3,   0,   0,   0,   3,   0,   0,  -1,   0,   0,   0,  -1 },
     {    2,   5,  -5,  -4,  74, -40,   4,  -5,  -2,  -3,   2,  17, -69,  41,  -8,   3,  -1,  -4,   5, -29,  11, -15,   9,   0,  -2,   5,  -5,   9,  16,  -5,  -4,   1,   1,   0,   3,   7, -13,   9,  -1,  -2,   1,  -1,   0,  -7,   7,  -6,   1,   0,   0,   3,   0,   1,  -2,   2,  -1,   1,  -2,  -2,   0,  -2,   1,   0,   1,   0 },
     {  -11, -45,   9,   7,   9,   0,   2,   0, -19, -60,  15,   1,   5,   5,   1,   0, -30, -42,   7,  21,  -3,   1,   1,   1, -38,  -8,   4,  -4,  -1,   0,   2,   2, -42,  11, -27,   5,   3,   1,   1,   0,  -7,  15,   7, -13,   2,  -4,   1,  -1,  27, -23,  14,   3,  -6,   2,  -2,   0,   3,   7, -12,   4,   2,  -1,   0,   0 },
     {    3,  -2,  -2,  -1,   1,   2,   0,   4,   3, -10,   4,  -4,   2,   0,   1,  -2,   6, -21,   7,   0,  -1,  -4,   3,  -1,   7, -27,  13,   5,  -1,   2,  -1,   1,  11, -27, -10,   7,   5,   2,  -2,   0, -18, -42, -27,   8,   2,   6,   2,   2,  67,  52, -34, -19,  -8,  -5,  -1,  -2, -22,  19,  31,  -1,  -1,  -3,  -1,   0 },
     {   -6, -10, -32, -10,   8,   1,   0,   3,  -6, -10, -13, -14,  -2,   6,   4,   0,  -2,   4,  11,  19,  -2,  -4,  -2,  -2,   4,  26,  65,  -2,  -3,  -2,  -6,   0,   6,  40, -42,  -2,   3,   2,   5,  -1,  15, -56, -11,   2,   1,   4,   1,   2, -38,  18,  13,  -5,  -2,  -1,  -1,   0,  12,  -3,   1,   3,  -1,  -1,  -1,  -1 },
     {   -2,  -7,   3, -10,   2,   1,  -3,   2,   1,   3,  18,   4, -10,  -2,   2,  -4,  -5,   0, -24,   5,   8,   2,   3,   0,   9,  17,  13,  21,   3,  -3,  -2,  -2,   2, -25, -39, -48,  12,   3,   4,   4,  21,   3,  59,   7, -36,  -3,  -8,  -1,   7,   4, -39,  35,  19,  -4,   2,  -3, -11,   8,   2, -24,   4,   6,   1,   1 },
     {   -5,   4,  -8,   6,  -2,  -2,   4,   3,  -8,  11, -18,   7,   4,   6,  -4,  -3, -12,  22, -35,   7,   6,  -1,  -4,  -1, -14,  37, -46,   6,  -5,   7,   4,   0, -10,  42, -44,  -5,   0,  -3,   2,   0,  -9,  38, -21, -11,   2,  -7,   0,  -1,   6,  52,   1, -13,  -3,  -2,  -2,  -2, -14,  13,  20,   0,   3,  -3,   0,  -1 },
     {    2,   3,  58, -46,   5,  -7,  -4,  -3,   3,   8,  68, -44,   2, -10,   3,   1,   1,  16,  26,  -7,   4,   3,   0,   1,   0,  17,  -5,  32,  -3,   4,  -5,  -3,  -2,   6, -12,   2,   6,  -4,   4,  -3, -11,  -4, -16, -18,   0,   3,   1,   2,  -7,   6,  14,   0,  -8,   0,   0,  -1,   1,   2,   7,   7,   1,  -3,  -2,   0 },
     {   -2,   1,  -5, -26,   2, -14,   2,   4,   3,  10,  -4,   6,  29,  21, -10,  -4,  -2,  -6, -12,  10, -75,   0,  15,   3,   5,  -2,   7,  20,  55, -30, -10,   3,  -4,   3,   7, -22,  -1,  34,  -1,  -3, -11,  12, -20,  10, -17, -10,  11,   1,   5,  -6,  10,  -6,   8,  -5,  -4,   1,   1,   2,   0,   3,  -4,   1,  -2,   0 },
     {   -6,  -8,  -6,  -4, -43, -78,  -9,  -8,   4,   5,   5,  -9,  27,  61,   0,   3,   6,   7,   8,  13,  27,  12,  14,   1,  -1,  -7, -10, -19, -26, -25,  -4,  -1,   2,  -4,  -4,   5,   3,   7,  -3,   1,   2,  -6,   5,  -3,   0,   0,   3,   1,  -1,   1,  -5,   6,   2,   1,   1,  -1,  -4,   5,  -7,  -2,   0,   1,   1,   1 },
     {   -3,  -6, -12, -12, -14,   8,  10,   0,  -2,  -4, -13,   7,  -9,  10,  -7,   5,   0,  -5,  -4,   4,  42, -23,   1,  -1,   2,   1,  21,  52, -35,   3,  -4,  -5,  -2,   0,  29, -61,  10,  10,   1,   0, -13,  17, -37,  22,   8,  -2,   7,   1,  14, -10,  27,   3,  -8,  -2,  -2,   0,   1,  -1,  -5,  -1,   0,   1,  -1,  -3 },
     {   -1,  -2, -11, -12,  10,   4,  -8,  -2,  -2,   0, -10, -16,   1,  -2,   5,  -2,   1,  12,  15,   5, -13,   5,   4,   3,   7,  24,  19,  -6, -14,  -7,  -6,   2,  23,  32,  20,  14,   6,  -3,   4,  -6,  46,  23, -11, -16,   4,   1,   2,   3,  39, -41, -12,  17, -13,   7,  -2,   2, -61,  15,  18,  -8,   8,   0,   3,  -1 },
     {    1,   4,  -1,  -3,   8,   0,   4,   3,   0,   3,  -9,  -2,  -2,   0,   0,  -1,   2,   2,  -7,   7,  -5,   8, -14,   2,   3,  -5, -18,  12,  -2,   9,   9,  -2,   4, -10, -17,  10,   6,  -2,   4,   4,   8, -10, -50,  -3,   2,  -2,  -3,   9, -11, -28, -36,  14,  -3,   4,  -1,  -2,  30,  72, -50, -28,   0,  -8,   1,  -2 },
     {    1,  -1,  -1,  -1,  -3, -15,  -5, -16,   3,   4,  -4,   7,   5,  -9,   7,  18,   1,   7, -19,   0, -15,  -2,  13,  -1,   9,   9, -17,   7,  13,  15,  20,  15,  19,  -6, -12, -17, -18, -32,  -5,   8,  24, -32,   4,  -8,  13,   2,  -9,  -6,  44,   7,  45,  30, -21,  -5,   5,  11,   1,  -7, -45,  33,   6,   3,  -1,   1 },
     {   -6,  -8, -27, -51,  19,   1, -12,   1,  -6, -10, -32, -52,  20,  10,  -3, -28,  -5,  -9, -22, -14,   8,  20,   9,   6,  -1,  -7,   4,  30,  -3,  15,  24,   7,  -4,  -5,  22,  24, -13, -12,  -5,   3,   0,  16,  30,   9,   7,  -4, -17,   5,   0,  15, -11,  -9, -10,  -6,   3,  -7,  16,  -3,  -7,   4,  -4,  -8,   0,   6 },
     {   -1,  -2,   1,   3, -16,  -3, -17,  26,   3,   3,   5,  11,  13,  32,  33, -28,  -3,  -5,  -5,  -3,  -9, -60, -25,  15,   1,   0,   1,   4,  14,  48, -27, -14,   3,  -2,  10,   8,  -5,  -6,  46,   1,   0,  -4,   7, -15,   1, -10, -23,  11,  -1,   1,   5,   9,  -5,  11,   0, -11,  -2,   2,   3,  -2,   4,  -8,   1,   3 },
     {   -3,  -4,  -9,  -5,  10,   4, -82, -25,   0,   0,  -6,  15, -14, -12,  62,   6,   1,   1,  -3,   5,   5,  40,  -1,  12,  -3,  -5,  -7,   6,   1, -20, -27, -20,  -3,  -2,   1, -11,   0,   5,  10,   1,  -8,   1,  -8,   1,   1,  14,  -3,   4,   1,   2,   9, -12,   0,  -6,   0,  -5,   4,   2,   0,  -3,   0,   0,   3,  -2 },
     {    3,   6,   8,  -2,  38,   1, -17,   4,   4,   8,   8,   4,  18,  20,   4,  10,  -1,   0,   0,   5, -37, -30,  -3,   2,  -2,  -5,  -2, -20, -41,   9,  28, -29,   2,   4,   1,  -5,  54, -11, -31,  23,   4,   9,   8,  28,   4,  25,  -3,  -1,   0,   7,  11,  -6,   0, -12,  19, -17,   4,   1,  -5,  -7,  -4,   2,  -5,   4 },
     {   -1,   0,   0,   1,   0, -14, -13,  -6,   2,   3,  11,   0,  10, -13,   4,   9,  -1,  -4,  16, -27, -10,  -7,  24,   4,  -2,   6,  13,  -8, -11,  14,   7,   9, -11,   3, -19,   0, -50, -25,  12,   2,  -4,  14, -11,  63,  12, -19,  -2,   1,   2,  -1,   8,  -5,  38,  33,   4,  -1, -16,   8,  -3, -29, -23,   5,  -6,   2 },
   },
   {
     {  107, -49,  -3,  -5,  -1,  -1,   0,   0,  36,  -3,  -9,   0,  -2,   0,  -1,   0, -19,  21,  -4,   0,   0,   0,   0,   0, -14,   7,   3,  -1,   1,   0,   0,   0,  -6,   1,   2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   35,  25, -19,   0,  -5,   0,  -2,   0, -80,  63,  -4,   0,   0,  -1,   0,   0, -47,   1,  20,  -4,   4,  -1,   2,   0,  10, -27,  11,   1,   1,   1,   0,   0,   8, -10,  -2,   3,  -1,   1,   0,   0,   4,  -2,  -2,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {   -1,  74, -48,   6,  -8,   1,  -3,   0,  69,  32, -30,  -5,  -5,  -2,  -1,  -1, -10,  18,  10,  -9,   2,  -2,   1,  -1, -26,   1,  17,  -3,   2,   0,   1,   0,  -9,  -7,   8,   1,   1,   0,   0,   0,  -4,  -4,   2,   1,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0 },
     {  -47, -31,  31,  -3,   6,   0,   2,   0,   7,   2,  -7,   4,  -2,   1,   0,   0, -74,  66, -12,   2,   0,  -1,   1,  -1, -31,   2,  17,  -7,   4,  -2,   2,  -1,  11, -21,   9,   1,   0,   1,   0,   0,   7,  -7,  -2,   3,  -1,   0,   0,   0,   3,  -1,  -2,   0,   0,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0,   0 },
     {   -1,  -8, -14,   8,   0,   2,   0,   1, -39,  31, -32,  10,  -1,   2,   0,   1,  52,  48, -44,  -7,  -3,  -2,  -1,  -1,  -6,  57,   0, -19,   2,  -4,   0,  -2, -19,   9,  29, -11,   1,  -2,   0,  -1,  -6, -10,  15,   2,  -2,   1,   0,   0,  -2,  -5,   2,   3,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0 },
     {  -13, -31,   1,   6,   2,   2,   1,   1,  31,  34, -43,   7,  -5,   1,  -1,   0,  19,  20,  -8, -10,   1,  -3,   1,  -1,  81, -44,  10,  -7,  -3,   0,  -1,   0,  34,  -2,  -9,   4,  -4,   1,  -1,   0,  -6,  13,  -4,   0,   1,  -1,   0,   0,  -6,   6,   1,  -1,   1,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0 },
     {   12,  28,  66, -57,  12, -13,   3,  -3,  17,  49,  39, -33,  -3,  -6,  -2,  -2,  26,  22,  -7,  13, -14,   2,  -3,   0,   6,  -2,  -7,  14,  -2,   0,   0,   0,  -5,  -9,  -1,   6,   2,   0,   0,   0,  -5,  -7,  -2,   3,   1,   1,   0,   0,  -2,  -3,  -1,   2,   0,   1,   0,   0,  -1,  -2,  -1,   1,   0,   0,   0,   0 },
     {   -3, -27,   0,   4,   3,   2,   1,   1,  -8,   2,   2, -10,   4,  -1,   1,   0,  52,  -4,  37, -29,   3,  -5,   0,  -1, -37, -24,  62,  -5,  -5,   2,  -2,   1,  -1, -51,  16,  24,  -9,   5,  -2,   2,  11, -13, -19,  17,  -2,   1,   0,   0,   3,   2, -11,   1,   3,  -1,   0,   0,   1,   1,  -3,  -1,   1,   0,   0,   0 },
     {  -22, -38,  16,   5,   4,   1,   1,   0,  10,  41, -19,   0,  -3,  -1,  -1,  -1, -10, -21,  37, -10,   4,  -1,   0,   0, -11,  -6,   4,   9,  -5,   3,  -2,   1, -76,  50, -11,   8,   2,  -1,   1,  -1, -28,  -5,  13,  -6,   7,  -2,   1,   0,   4, -13,   5,  -1,   0,   1,   0,   0,   2,  -3,  -2,   1,  -1,   1,   0,   0 },
     {   -4,  -4, -28,  -5,   9,   0,   2,   1,  -1, -17,  32, -53,  24,  -7,   4,   0, -13,  32,  41, -50,   1,   1,  -2,   0,  38,  42,   2,  -1, -25,   5,  -5,   0,  12,  22,   6,  15, -12,  -1,  -1,  -1,  -9,  -1,   8,   8,   1,  -2,   0,   0,  -7,  -5,   3,   3,   1,   0,   0,   0,  -1,  -3,  -1,   3,   0,   0,   0,   0 },
     {   -7, -10, -33, -59,  59, -16,  11,   0, -10, -14, -50, -32,  32,   0,   5,   2,  -9, -22, -23,  17, -19,  17,  -3,   3, -11, -14,   6,  11, -16,   5,   0,   0,  -2,  -1,  11,   1,  -4,  -3,   1,   0,   3,   3,   6,  -1,  -2,  -2,  -1,   0,   3,   3,   2,   1,  -1,   0,  -1,   0,   0,   1,   1,   1,  -1,   0,   0,   0 },
     {   -6, -11, -16,   7,   2,   2,   0,   1,  -9,  15,  -5, -10,   4,  -2,   0,   0,  25,  31,   5,  -5,  -4,  -1,  -2,  -1, -48,  -1, -22,  33,  -5,   4,   0,   1,  29,  12, -68,  13,   6,  -4,   2,  -1,   2,  46, -18, -25,  10,  -4,   1,  -2,  -7,  14,  16, -16,   0,   1,   0,   0,  -3,   1,   7,   0,  -3,   1,   0,   0 },
     {   13,  36,   8, -14,   1,  -4,  -2,  -1, -21, -53,   1,  20,  -4,   5,  -1,   1,   3,  42, -19,  -4,   0,  -6,   0,  -1,  14, -33,  45,  -6,  -2,   2,  -2,   1, -38,  19, -18,  24,  -7,   3,  -1,   0, -39,  25,  -5,  -4,  11,  -5,   3,  -2,   2, -10,   5,  -8,   3,   2,   0,   0,   1,   0,  -1,   0,  -1,   1,   0,   0 },
     {    5,  10,  31, -18,   4,  -4,   1,   0,   1,   4, -10,  -9,  17,  -5,   2,  -1,   0, -18,  -8, -44,  42, -10,   6,  -2, -24,  -7,  10, -59,  21,   7,  -2,   2,  21,  33,  -6, -32, -21,  14,  -5,   1,  14,  32,  11,  -6, -22,   3,   1,  -2,   5,   5,  12,   6,  -9,  -2,   1,   0,   1,   0,   3,   6,  -3,  -1,   0,   0 },
     {    6,   8,   7,  46,  48, -53,   6,  -4,   7,  10,  17,  40,  52, -45,  -2,  -3,   3,   9,   7,  15,  10,   7, -15,   1,   4,   2,  -1,  -1,  -9,  25,  -7,   0,  -3,  -7,  -2,  -7,  -7,  15,   2,   1,  -4,  -4,  -3,  -7,  -4,   4,   4,   1,  -1,  -3,  -4,  -3,  -2,   2,   2,   0,  -1,   0,  -2,  -1,   0,   0,   1,   0 },
     {    1,   3,   3, -11,  21, -14,   4,  -2,   1,   3,   4,  31, -28,   7,  -9,   1,   4,   1,  45,  29, -46,   9,  -5,  -2,   3,  27,  50, -16,   0, -16,   5,  -5,  27,  43,   5, -16,  15, -13,  -2,   0,  30,  20,  -1,  -5,   6,   1,  -4,   2,   3,   4,   0,   5,  -1,   4,  -1,   0,   0,  -1,   0,   2,   1,   0,   1,   0 },
     {   -1, -10, -17,   7,   0,   5,   0,   3,  -2,  -8,  22, -11,   0,   1,   0,   0,  17,  23,  13,   9,  -7,   1,  -3,  -1, -11, -45,   2,  12,  12,  -2,   1,   1,  14,  -6,   8, -41,  15,  -3,   3,  -1, -13, -14,  68, -28,  -9,   4,  -3,   1,   9, -39,  31,  18, -18,   5,  -2,   2,   3, -11,  -5,  16,  -3,   0,   0,   0 },
     {    7,  11,  10,  -7,   1,  -3,   0,  -2,   0, -12, -23,  11,   1,   2,   2,  -1,   6,  13,   1, -17,   7,  -3,   1,  -1,   8,   0,  -8,   9,  -8,   4,  -3,   0,  -1,  15, -39,  20,  -5,  -1,   2,  -3,  66, -54,   5,  -8,   7,  -3,   2,  -1,  58, -31,  -9,   1,  -6,   4,  -4,   2,  10,   8, -11,   4,  -3,   0,  -1,   0 },
     {   -3,  -4, -16,   5,  -3,   2,   0,   3,   1,  -5,  17, -22,   8,  -1,  -2,   1,   2,  24,   2,   6,  -4,   0,  -2,   0,  23, -19,  -5,  -5,  16,   0,   2,  -1, -65, -10,   6, -28,  11,   4,   2,   1,  70,  24, -21, -15, -14,   3,  -2,  -1,  -6,  34,  15,  -3,  -6,  -4,   1,   0,  -2,  -4,  11,   4,  -2,  -1,   2,   1 },
     {   -3,  -3, -18,  -3,   7,   3,  -4,   5,  -4,  -4,   9,  -9,  -1,  -1,   1,   2,   9,  25,   6,   2,  -7,  -3,   0,  -2, -12, -31, -15,  -4,  15,   0,   1,   2,   3,  36, -11, -34,   8,   7,  -1,   0, -17, -25, -16,  23, -12,  15,  -2,   2,   0,  20, -77,  19,  -1,  -2,   5,  -1, -11,  33, -20, -17,   8,  -5,   3,  -1 },
     {   11,  23,  54,  20, -30,  -2,  -6,  -3, -11, -23, -55,  -8,  23,  -5,   3,   0,   2,   8,  25, -18, -18,  14,  -5,  -1,  -4,  -7,  -8,  39, -26,   7,   0,  -6,   4,   4,  27, -20,  11, -10,   2,  -3,  10,   9,  20,   0,  -2,   1,  -4,   2, -21,  21, -13,   3,   5,  -2,   2,   2,   0,  -1,   3,  -4,   1,  -2,   2,   1 },
     {    0,  -5,  -1,  41, -19,   4,  -6,   1,   7,  18,  30, -33,  12, -13,   4,  -3,  -4, -13, -58,   5, -16,  21,  -1,   6, -13, -25,  31,  -2, -35,  -1,   2,   1,  10,  41,  21,  12,  -3, -21,   2,  -3,  19,   9,  -3,   4,  17,  -8,  -5,   1,  16,  -8,  -5,  -4,   7,   4,  -4,   2,   0,   5,  -5,  -1,   0,   3,   0,   0 },
     {    3,   5,   6,  12,  18,  -8,   4,  -8,   1,   1,   5,   6, -15,  34,  -4,  -2,  -2,  -3,  -1, -41, -42,  46,   5,   4,  -3,  -7, -24, -45, -32,   6,  26,   3, -12, -29, -18, -25,  -4, -16,  19,   5, -20,  -7, -14, -18,   4, -15,   1,   8,   9,   2,   1,  -5,  -4,  -4,  -9,   4,   2,   3,  -4,   3,  -2,  -3,  -4,  -2 },
     {   -5,  -9, -11, -11, -35, -62,  11,   9,  -7,  -9, -10, -17, -54, -60,   3,  16,  -4,  -6,  -7, -17, -29, -15,  -5,   7,  -3,  -4, -11, -15,  -6,  16,  -8,  -2,  -2,  -7,  -1,  -2,   7,  21,  -7,  -8,  -3,   2,   0,  -4,  13,  13,   1,  -6,   4,  -2,   1,   3,   3,   7,   1,   0,   8,  -6,   0,   3,   3,   1,   1,   1 },
     {   -2,  -1,  -7,  -2,  -5,   7,   1,   2,   0,   4,   6,  15, -23,  10,   0,  -2,   3,  10,   5,  19, -10,  13,  -7,   0,  -8,  -8, -17,   4, -26,  30,  -7,   0,  -6,   3,  -2,  10, -62,  22,   4,  -6,  18,  24,  22,  36, -33, -25,  13,  -7,  -6,  -5,   7,  38,  21, -26,   2,  -1,   1,  -3, -14,   4,  22,  -6,  -1,   0 },
     {    1,   0,   4,   9,  -7,  -1,   2,  -2,   0,  -1, -15, -16,  24,  -5,   1,  -2,   0,  -1,   3,   6,  -8,   9,   2,  -2,   5,   7,  -6,  -5,  23, -13,  14,  -4,  -5, -10, -15, -13,  25,  -6,   3,   5,  -1,  35, -37,  37,  -3,   5,  -4,  -3,   3, -66,  -9,  29,  15,   1,   2,   0,  26, -38, -31,   6,   4,   3,   0,   3 },
     {    0,   1,  -2,  -4,   4,  -8,   0,   1,  -1,  -4,   2,   7,   1,   1,  -5,   3,   1,   0, -10,  -8,   1,   6,   8,  -2,  -8, -15,   2, -16,   6,   2,   4,   3,   1,  17, -24,   8,   4, -10,  -2,  -1,  26, -47,   4,  14,  21,  -8,  -2,   4, -97,  -8,  12,   8,   9,   5,  -4,   3,  -8, -25,  -4,   3,  -3,   9,   0,  -2 },
     {   -2,  -1,  -1,  -7,   3,   0,  -5,   1,   4,   7,   8,  16,  -3,  -1,  10,  -7, -10, -15, -24, -23,   6,   0,  23,   0,   8,  21,  27,  21, -25,  -5, -13,   4, -11, -21, -21,   6,  26,   2,  -5,   4,   8,  21,  16, -31, -11,  19,  -7,   3, -20,   5, -16,  52, -22,   3,   4,  -1, -13,  21, -43,  13,  18, -18,   7,   0 },
     {    4,   6,   5,  26,  45, -35,  -8,  -6,   0,  -1,   5,   1, -26,  26, -13,   0,  -1,   0, -10, -44,  -8,   5,   7,   6,  -6, -16, -25,  27,  35, -42,  -5,   3,   1,   9,  26,  31,   7,   5, -15,  -2,  11,  15,  12,  23,  -3,  10,  -4, -14,   5,  -1,   9,   9,   4,   0,   4,  -6,  -9,   4,   8,  -6,   7,   1,   1,   5 },
     {   -2,  -2,  -4,  -1, -23, -15,  67, -38,   0,   0,   0,  -4,  -5,  14,  20, -16,   0,   0,  -2,  -3,  19,  40, -54,  26,   2,   4,  10,   3,  30,  -3, -31,   8,   0,   1,  -5,   6,   7, -15,   6,  10,   0,   0,   7,  -5,  -4, -10,   5,   4,  -6,   1,  -9,  -3,   1,  -9,   3,   2,  -2,  10,  -7,  -3,  -1,  -5,   3,   3 },
     {    2,   2,   8,  18,  -5,   9, -28, -10,  -1,   0,  -7, -23,   9,  12,   2,  -3,  -5,  -9, -10,   7, -37,  19,  -8,   2,   5,  13,  26,  -6,   7,   2, -25,  18,   2,   0, -34,  -6,  15,  52, -19,   6,  -8, -26,   1,   5, -28,  27,  15, -22,  -1,  -1,  21, -17,   4, -28,  23,  -8,  -3,  -2,  16, -10,   7, -12,   1,   5 },
     {   -2,  -2,  -2,  -9,  -5,  -6, -33,  41,   0,   2,   1,  -3,   0,  10, -68,  48,  -2,  -4,  -6,  -8,  20,  28, -43,  21,   3,   6,  10,   8,  14,  13,  -2, -14,   3,  -1,   1,   5,   8,  -3,  17, -14,  -1,   1,  -2, -16,  -5,  -5,  14,  16,  -1,  -4, -11,  -1,  -1,  -8,  10,  11,   4,  -4,  -8,   2,  10,  -7,   5,   1 },
   }
 },
 {  //23
   {
     {  121, -28,  -9,  -4,  -3,  -2,  -1,  -1, -16,  15,  -2,   1,   0,   0,   0,   0, -16,   4,   0,   0,   0,   0,   0,   0,  -2,   3,  -1,   0,   0,   0,   0,   0,  -5,   4,   0,  -1,   0,   0,   0,   0,  -2,   1,   2,  -1,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    9, -18,   2,  -1,   1,   0,   0,   0, 110, -15, -14,  -2,  -4,  -1,  -2,  -1,  -3,  42,  -8,   0,  -2,   0,  -1,   0, -34,  14,  10,  -1,   1,   0,   1,   0,  -7,  -6,   5,   1,   0,   0,   0,   0,  -4,  -1,   0,   1,   0,   0,   0,   0,  -2,   1,   0,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {  -13,  -1,   4,   1,   1,   0,   1,   0,  10,  36,  -7,   0,  -2,   0,  -1,  -1, -95,  13,  27,   0,   5,   0,   2,   0,   7, -60,   9,   3,   2,   1,   1,   1,  23,  -9, -16,   3,  -2,   1,  -1,   1,   3,   5,  -6,  -1,   0,  -1,   0,   0,   1,   3,   0,  -1,   0,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0 },
     {   -2, -25,   3,   0,   1,   0,   1,   0,  37,  -2,  -9,  -2,  -2,   0,  -1,   0,  -1, -38,   5,   2,   1,   1,   1,   1,  63,  -8, -43,   4,  -6,   1,  -2,   1,  -6,  72, -16, -12,  -2,  -3,   0,  -2, -16,  13,  24,  -9,   1,  -2,   1,  -1,  -2,  -3,  10,   0,  -1,   0,   0,   0,  -1,  -2,   4,   1,   0,   0,   0,   0 },
     {   -8,   0,   5,   1,   0,   0,   0,   0,   1,  28,  -3,  -1,  -2,  -1,  -1,   0, -25,  -3,  14,   0,   1,   0,   1,   0,   8,  25,  -5,  -3,  -2,  -1,  -1,   0, -34,  13,  52,  -6,   3,  -3,   1,  -1,   6, -74,  24,  28,  -2,   6,  -1,   3,   4, -14, -44,  19,   3,   3,   1,   1,  -3,  10, -16,  -9,   5,  -2,   1,  -1 },
     {   -5, -18,   6,   4,   0,   1,   0,   0,  11,  26, -11,  -6,   0,  -2,   0,  -1, -10, -37,   9,   7,   0,   2,   0,   1,  14,  30, -24,  -8,   0,  -2,   0,  -1, -14, -26,  34,   7,  -1,   2,   0,   1,  13,  -5, -63,   5,   5,   1,   1,   1,  -3,  46,  23, -38,  -3,  -6,  -1,  -3,   4, -16,  24,  18,  -7,   1,  -2,   1 },
     {  -12,  30, -17,  -2,  -1,   0,  -1,   0, -20,  47, -22,  -5,  -2,   0,   0,   0, -21,  45, -17,  -7,   0,  -1,   0,  -1, -20,  43,  -8, -14,   2,  -3,   1,  -1, -18,  34,   7, -20,   4,  -4,   1,  -2, -10,  18,  26, -24,   3,  -4,   2,  -2,   0,  -7,  42, -17,  -3,  -1,  -1,   0,   6, -16,  23,   1,  -6,   2,  -2,   1 },
     {  -30,-111,  11,   3,   7,   3,   3,   2, -32,  11, -17,   1,  -1,   2,   0,   1,  13,  35,  -2,  -3,  -2,  -2,  -1,  -1,  -7,   0,   5,   1,   1,   0,   0,   0,   2,   0,  -3,  -3,   1,   0,   0,   0,   4,   1,   5,  -2,   0,   0,   0,   0,   1,  -2,   0,   2,  -1,   0,   0,   0,   1,   0,  -1,  -1,   0,   0,   0,   0 },
     {    9,  18,  26,  -3,  -3,  -2,  -2,  -1, -17, -36, -33,   4,   5,   3,   2,   2,  19,  39,  30,  -6,  -5,  -4,  -3,  -2, -19, -40, -26,  10,   6,   4,   1,   1,  18,  38,  18, -16,  -8,  -3,  -2,  -2, -13, -31, -10,  20,   7,   4,   2,   2,  11,  29,  -4, -28,  -3,  -4,  -1,  -2,  -1,  -9,  24,  19,  -4,   0,  -1,   1 },
     {   -2,  -7,   8,  -3,   2,  -1,   1,  -1, -16, -77,   6,   9,   3,   3,   1,   2, -61,  13, -48,   4,  -1,   3,   1,   2,  26,  33,   2, -24,   0,  -6,   0,  -2,  22,   5,  24,   3,  -4,  -1,  -2,   0,  -4,   3,  -3,   7,   0,   0,   0,   0,   0,   0,   1,  -4,   2,   0,   0,   0,   1,  -1,   3,   0,  -2,   0,  -1,   0 },
     {   -2,   2, -19,  -3,   2,  -1,   0,   1,  -2,   5,  -6,   3,  -1,   1,  -1,   0,   0,   5, -22,  -4,   5,   0,   1,   0,   1,   0,  -5,   2,  -1,   1,   0,   0,  -8,   7, -29,  -4,   6,  -1,   2,   0,  -4,   0, -16,  -4,   0,   3,   1,   1, -15,  24, -58, -37,  21,  -2,   6,  -1,   5,  21,  39, -78,  -4,  -1,  -2,  -1 },
     {    4,  11,  -2,  -2,  -2,   0,  -1,   0,   6,  27,  -9,  -5,  -3,  -1,  -1,   0,  29,  37,   7, -14,  -3,  -6,  -1,  -2,  68,   0,  46, -14,  -3,  -5,  -2,  -2,  28,  -1,  29,  26, -11,   2,  -5,   1, -29,  18, -10,  36,   0,  -1,   1,   0, -20,  -4,  -2,   0,   9,   0,   3,   0,  -2, -11,   3,  -8,   3,   2,   0,   1 },
     {    0,   4,   0,   0,   0,   0,   1,   0,  -4,  -9,  -3,   3,  -1,   3,   0,   0,   0,  20, -10,  -2,   2,  -2,   1,  -1,  31, -11,  31,   2,  -1,   0,  -2,   0, -57,   4, -27,  37,   2,   6,   3,   2, -23, -35, -26, -43,  21,   4,   7,   1,  32, -14,   3, -23, -10,   2,  -5,   1,   2,  25,   0,  29,  -9,  -8,  -2,  -2 },
     {    0,   5,   5,   0,  -1,   0,  -1,   1,  -5,  -1, -24,  -4,   5,   1,   2,   1,  -2,   5,   2,   2,   0,   1,   0,   1,  -6,   2, -24,  -5,   6,   0,   1,   0,   0,   3,   2,  -1,  -2,   0,   0,   1, -16,   6, -31,  -7,   5,   0,   5,   0,  -1, -21,  10, -39,   2,   9,   2,   3, -25,  -8, -85, -35,  47,   4,  10,   1 },
     {   -3, -11, -17,  12,  -2,   0,  -1,   1,   4,  23,  19, -18,   2,   0,   1,  -1,   3, -25, -13,  24,  -2,   2,  -2,   0, -36,  -1,  -9, -24,   7,  -1,   3,  -1,  53,  18,  18,  20, -13,  -2,  -4,   0, -50,  -5, -10,   3,  19,   4,   6,   0,  17, -36, -14, -22,  -1,   4,  -1,   2,  10,  32,   9,  16, -17,  -8,  -3,  -2 },
     {   -6,  -9, -76,  -1,   5,   2,   3,   2,  -4, -28, -61,  -9,   8,   4,   6,   2,  -7, -38,   9, -30,   3,  -1,   2,   0, -19, -17,  29,  -7,  -7,  -1,   0,   2, -21,   4,  17,   8,  -6,  -1,  -4,   1, -10,   9,   4,  17,   2,  -3,   0,  -2,   3,  -2,   5,   8,   4,  -1,   0,  -1,   7,  -9,   1,   1,   1,   3,   0,   0 },
     {    5,   9, -25,  -2,   3,   0,   0,   0,   8,  17, -39,  -8,   3,   2,   0,   2,  18,   8, -38, -12,   8,   0,   2,   1,  29, -11, -15, -30,  10,   0,   2,  -1,  43, -19,   0, -32,   3,   2,   1,   0,  41, -21,   5, -30,  -1,   6,  -3,   2,  17,  20, -12,   9,  -9,   0,  -5,   0, -10,  35, -15,  23,   3,  -8,   1,  -3 },
     {    5,  -9,  16, -19,   4,  -2,   0,  -2,   8, -15,  25, -29,   5,  -2,   2,  -3,   8, -17,  25, -29,   6,  -2,   3,  -1,   8, -17,  24, -31,   6,  -1,   2,   1,   7, -20,  24, -30,   4,   1,  -1,   2,   9, -21,  25, -33,  -4,   4,  -1,   2,  16, -27,  26, -41,  -7,   7,  -3,   3,  10, -17,  18, -39,  -5,   8,  -3,   2 },
     {    4,   7,  74, -11,   0,  -5,  -2,  -3,   0,  13, -61,  19,   2,   3,   4,   3,  -4, -41, -15, -31,   8,   0,   1,   3, -17,  23,  30,  11, -11,  -3,  -4,  -2,  22,   3, -13,  11,   7,   2,   2,   0, -11,   0,   4, -11,   6,   4,   1,   0,  -4, -10,  -7,   8,  -2,  -1,   1,   1,   6,   4,   3,  -4,  -9,  -1,  -1,   0 },
     {    1,   1,   6,  13,  -6,   0,  -2,  -2,  -3,  -2,  -2, -20,   6,   2,   2,   0,   2,   3,   8,  16,  -4,  -1,  -1,  -2,  -7,  -3,  -1, -22,   6,   1,   0,  -1,   7,  15,  11,  29, -12,   0,  -2,   0, -11, -13, -13, -33,  28,   2,   3,   0,  14,  37,  21,  68, -29, -18,  -6,  -4,  -6,  -6,   5, -55,   7,   6,   5,   2 },
     {    1,  -2,  -5,  71, -42,  18,   5, -10,   2,  -6,  11,  17,   6,  -5,   2,  -1,  -3,   5,  22, -53,  44, -11,  -5,   4,  -1,  21, -14, -15,   1,  12,  -9,   3,  10,  -4, -22,   9, -28,   8,   3,  -2,   3, -14,   3,   2,  -1,  -8,   4,  -3,  -5,  -3,   5,  -4,   9,  -2,  -1,   2,  -5,   3,   1,   1,  -2,   1,  -2,   2 },
     {   -2,  -2, -33,  -6,  -4,   8,  -3,   2,  -2, -16,  -7, -19,   6,   0,   1,   0,  -9,   8,  49,   2, -15,  -3,  -5,  -3,  15,  48,   1,  55,  -3,  -1,  -4,  -4,  44,  -9, -14,  -1,  38,   0,   8,  -1, -12, -20,  -3, -30,   5,  19,   2,   6, -17,  -4,  -6,   0, -16,   4,   2,   0,   0,  -2,  -1,   2,  -7,  -3,   1,  -1 },
     {   -5,  -9, -23, -36, -12,   2,   3,   6,   7,  19,  40,  43,  17,  -7,  -6,  -5,   1, -10, -35, -40,  -3,  11,   9,   7,  -9,  -6,  17,  26,   5, -12,  -7,  -3,  20,  19,  -3, -11,  -2,   9,   8,   5, -22, -33,  -9,   2,   8,  -5,  -2,   0,  13,  31,  12,   1,  -2,  -3,   0,  -1, -15, -30, -10,   2,  11,   5,   4,   2 },
     {    1,  -1,  -9,  31, -44,  25, -11,   7,   2,   1, -16,  48, -50,  30, -11,   4,   6,  -8,  -9,  34, -39,  20,  -3,  -3,   6, -14,  13,  12, -21,   5,   1,  -6,   9,  -8,  17,  -4,  -4,   1,   3,  -2,   9, -11,  13, -13,  -8,   0,   1,  -1,  10, -10,  13, -16,  -8,   1,   0,   0,   4, -10,   7, -17,   2,   5,   0,   2 },
     {    0,   0,  -5, -16,   1, -10,  -7,  -1,   1,   1,  10,  -7, -13,   1,   0,  -7,  -1,   1,  21,  -8, -12,  -5,   6,  -1,  -9,  23,  21,   2, -19, -11,  -1,   4,  20,  46, -10,  42, -12, -12, -10,  -4,  73,   8,  -1,  15,   7,  -8,   0,  -6,  39,   7,   5, -22,   6,  11,  -5,   3, -12,  19,  -5,  -5,   8,   7,   0,  -1 },
     {   -4,  -3,  -8, -59, -31,  -9,  12,   8,  -1,  -5, -17,  24,  -6,  -2,  -5,   2,   2,  -1,  30,  49,  47,   4, -11, -13,   4,  21,   9, -19,  19,   9,   7,  -1,   5,  -4, -22, -14, -36,  11,   7,   7,  -4, -13,  -5,   2,  -5, -16,   1,  -4,   3,  -7,   7,   8,  18,  -5,  -6,  -1,   4,   4,   6,  -1,   4,   3,   0,   4 },
     {    3,   4,  -6,  21,  80,   5,   5,   1,   3,   2,   1,  44,  15,  27,   5,   2,   3,  -1,  29,   6, -28,  10,   9,  -2,   0,  16,   3, -34,  -6, -22,   7,   0,  10,  -3, -27,  -9, -22,   3,  -9,   0,  -5, -21,  -3,  -8,  -4,  -5,   3,  -3,  -7,  -4,   2,   2,  14,  -5,   1,  -2,  -2,  -9,   5,   2,   7,   6,  -1,   2 },
     {   -3,  -4,  12,  -1, -19,  10,  -5,   6,  -4,  -5,  17,  -9, -15,  13,  -5,   0,  -7, -10,  28, -12, -20,   9,   2,  -5, -11,   7,  33, -21,   2,  -5,  -3,  -5, -12,  13,  16, -29,  -1,  -2,   4,  -1, -22,  25,   9, -27,   4,   1,  -1,  -3, -23,  41, -13,  -3,  23,  -9,   5,  -4, -27,  46,  -3,  24,  34, -20,   5,  -7 },
     {   -1,  -1,  -3, -10, -12,   6,   4,  -3,   0,  -1,   3, -13,   4,  12,   0,   7,   4,  -1, -13,  12,   6,   0,   2,  -3, -10, -17,   2, -18, -15,   1,  -1,  -3,   3,  16,   1,  26,  -7,  -5,   0,  -9,  30, -21,   5, -31,  38,  18,   5,   3, -69,  -3, -10,  -5,  38,  11,   6,   3,  -2, -43, -11,  17, -17,  -3,   2,   3 },
     {   -2,   0, -10, -31,  23,  -4,   5,  -2,   0,  -1,  13,  47, -46,   7,  -3,   1,   3,   7,  14, -26,  35, -19,  -1,  -4,  -6,  -5, -25, -20, -33,  32,   0,   4,   3,  -3,  22,  18,  25, -31,  -6,  -4,  -8,  13,  -9, -14,   3,  22,  -3,   3,  -8,  -3,  -3,   7, -24,  -4,   7,  -1,   3,  10,  -2,  -1,   3,  -6,  -9,  -2 },
     {   -4,  -4,  -5, -13, -38, -50, -28, -18,   4,   2,   1,  36,  45,  27,  26,  -5,  -1,   4,  13,  -9, -41, -31,  -8,   6,   2,  -4, -16, -15,  12,  -6,  -4,   0,  -8,  -3,  13,   4,   1,   5,   0,   8,   7,  13,  -8, -14,  -5, -15, -17,  -6, -12, -14,  -8,   4, -10,   2,   2,   7,   4,  13,  -4,  -4, -23,  -4,  -5,  -2 },
     {    2,  -1,  -4,  41,  11, -34,  11,  -8,   2,  -3,  -2,  37,  -7, -31,  16,  -9,   4,  -7,   4,  31,  33, -24,   7,  -3,   2,  -5,  29,   8,  40, -23,   8,  -7,   4,  13,  20, -10,  42,  -6,   4,   6,   9,   9,   9,  -9,  33,  -2,  -2,   4,   2,  10,  -2, -10,   8,   9,  -7,   0,   1,  -6,  -2,   0,   8,   4,   3,   3 },
   },
   {
     {   99, -28,  -9,  -4,  -2,  -1,  -1,   0, -67,  28,   5,   2,   1,   0,   1,   0,  -4, -12,   4,   1,   1,   0,   0,   0,  12,   1,  -4,   0,  -1,   0,   0,   0,  -4,   2,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   38,  -6,  -4,   0,  -2,   0,  -1,   0,  65,  -3,  -9,  -3,  -2,  -1,  -1,   0, -94,  21,  13,   2,   4,   1,   1,   0,  22, -21,  -2,   1,  -1,   1,   0,   0,   4,  10,  -4,  -1,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   14,  81, -47,   5, -10,   0,  -3,   0, -23, -59,  45,  -7,   8,  -1,   3,   0, -14,  -2,  -8,   6,   0,   1,   0,   0,  23,   7,  -6,  -1,  -1,   0,  -1,   0,  -8,   0,   4,  -1,   1,   0,   0,   0,   3,  -1,  -1,   0,  -1,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {   55,  -2, -11,  -2,  -2,  -1,  -1,   0,  45, -33,   5,  -1,   0,   0,   0,   0,  35,  33, -19,  -1,  -4,  -1,  -1,   0, -76,  18,  19,   0,   3,   0,   1,   0,  16, -24,  -3,   3,   0,   1,   0,   0,  -2,  10,  -4,  -1,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {    1,  29, -25,   7,  -4,   1,  -1,   0,  17,  37,  -2, -11,  -1,  -3,  -1,  -1, -13, -76,  42,   4,   5,   2,   1,   1, -45,  24, -27,   8,   1,   1,   1,   0,  41,  -1,   0,  -8,  -1,  -1,  -1,   0,  -5,   3,   4,   1,   0,   0,   0,   0,  -2,  -3,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {  -24,   9,   4,  -1,   1,   0,   0,   0, -54,   4,   4,   5,   1,   1,   1,   0, -38,  32,  -1,  -3,   1,  -1,   1,   0, -32, -43,  25,   1,   4,   1,   1,   0,  70, -23, -23,   3,  -3,   1,  -1,   0,  -7,  25,  -2,  -4,   0,  -1,   0,  -1,  -2,  -5,   5,   0,   0,   0,   0,   0,   3,  -1,   0,   1,   0,   0,   0,   0 },
     {  -15, -40, -52,  54, -14,  14,  -3,   3,   9,  30,  49, -53,  16, -13,   3,  -3,   8,  15, -15,  16,  -9,   3,  -2,   0,  -5, -25,   0,   6,   1,   2,   0,   1,  -5,  12,   0,  -7,   3,  -2,   1,  -1,   3,  -1,  -2,   3,  -2,   1,   0,   0,  -1,   0,   2,  -2,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0 },
     {  -13,  -2,   1,   4,  -1,   1,   0,   0, -10,  17,   3,  -5,   1,  -1,   0,   0, -37,  -2,  -1,   6,   1,   1,   1,   0, -16,  28,  12,  -9,   2,  -2,   1,  -1, -35, -69,  23,  11,   2,   4,   1,   1,  61,  -1, -41,   0,  -3,  -1,  -1,   0, -10,  27,  10,  -7,   0,  -2,   0,  -1,   0,  -8,   4,   2,   0,   1,   0,   0 },
     {    7,  37, -19,  -3,  -3,   1,  -1,   0,   2,  22, -21,   9,  -3,   1,  -1,   0,  19,  11,  39, -25,   5,  -6,   1,  -1, -25, -59,  19,  24,  -4,   6,   0,   2, -47,  19, -33,   3,   7,  -2,   3,   0,  41, -10,   6, -13,  -1,   0,  -1,   0,  -1,   8,   2,   2,  -2,   0,  -1,   0,  -4,  -1,   3,   1,   1,   0,   0,   0 },
     {  -10, -31, -39,  18,   8,   3,   1,   2,  -9, -25, -11,   9,  -2,   4,   1,   1,   7,  36,  67, -46,   0,  -8,  -1,  -2,  19,  32, -27,  26, -11,   1,  -2,  -1,   7, -32,  -5,   4,   8,   1,   1,   1, -17,   4,   3,  -6,   3,   0,   0,   0,   1,   3,   2,   0,  -3,   0,   0,   0,   2,   1,   1,   0,  -1,   0,   0,   0 },
     {  -13,  12,   3,   1,   0,   0,   1,   0, -28,  17,  -1,   0,   1,   0,   0,  -1, -36,  37,   1,  -1,  -3,   1,   0,   0, -51,  30,  -1,   2,   3,  -1,   1,   0, -35,  19,  32, -12,   4,  -2,   1,  -1, -29, -52,  29,   9,   2,   3,   1,   1,  24, -26, -19,   6,   0,   1,   0,   0,   4,   7,  -3,  -4,   0,  -1,   0,   0 },
     {   11,  16,  47,  46, -52,  12,  -9,   0,  -9, -19, -47, -37,  54, -14,   8,   0,  -1,  -4,   8,  -5, -18,   8,  -2,   0,   5,  12,  13,  16,  -7,  -1,  -1,  -1,   1,  -4,  -9,  -5,  11,  -3,   1,   0,  -2,   2,   2,   2,  -5,   2,  -1,   0,  -2,  -2,   1,  -2,   2,  -1,   1,   0,   1,  -1,  -1,   1,  -1,   1,   0,   0 },
     {    6,  42, -27,   7,  -7,   2,  -1,   0,  16,  62, -25,  -2,  -5,  -3,  -1,  -1,  17,  20, -12, -14,   6,  -4,   1,  -2,  33,  12,  40, -24,  -3,  -4,  -1,  -1,   2, -25,  12,  30,  -7,   3,  -1,   1, -47,   8, -10,  -2,  10,  -1,   3,   0,   6, -10,   5,  -8,   0,   1,   0,   0,   3,   1,   0,   1,  -2,   0,   0,   0 },
     {   -5,   1,   1,  -8,   1,  -2,   1,   1,  -2,  12,   3,  -2,  -2,   0,   0,  -2,  -8,  12,   0,  11,  -5,   1,   0,  -1,  -3,  28,  -2,   0,   3,  -1,   0,   0, -26,   6, -12,  -5,   5,  -1,   3,   0,  10,  54,  33, -22,  -5,  -4,   0,  -2, -48, -55,  41,  24,   1,   4,   1,   1,  14, -22, -33,   5,   5,   1,   1,   0 },
     {   -7, -22,  -5,  15,   0,   2,  -2,   0,  -2, -14,  18,   9,  -4,   0,   1,   1, -21, -37, -26, -34,  28,   1,   4,   2,  -6,  29,  49,   4, -38,   3,  -6,  -1,  -2,  15, -40,  36,  18,  -7,   2,  -1,   1, -19,   7, -35,   9,   6,   1,   1,  10,  -5,   3,   9, -12,   1,   0,   0,  -3,   7,  -3,  -2,   3,  -2,   0,   0 },
     {    5,  11,  22,  21, -13,   2,  -3,   0,   5,  11,  15,  16, -11,  -1,   0,  -2,  -1,  -2, -36, -56,  39,  -6,   1,   3,  -7, -22, -31,  41, -26,  10,   0,   0,   0,  -1,  53, -16,   3,  -6,   1,  -1,   0,  28, -18,   5,   3,  -2,  -1,   0,  -4, -14,   2,  -2,   1,   1,  -1,   0,   3,  -2,  -3,  -3,   2,   0,   0,   0 },
     {    0,   4, -15,   4,   1,  -1,  -1,   0,  -3,   6, -18,   3,  -4,   1,   1,  -1,   5,  22,  -9, -20,   2,  -2,   1,  -1,   7,  26,  13,  -8,  -6,  -3,  -1,  -1,  44,  48,  19,  -2,  -6,  -4,  -2,  -2,  40, -39,   0,  16,   2,   4,   0,   1, -70,  17,   4,  -9,   8,  -1,   3,   0,  11, -14,   6,   1,  -2,   1,  -1,   0 },
     {    9,  14,  21,  50,  57, -20, -15,  -9,  -7, -11, -16, -50, -58,  26,  13,   8,   0,  -3,  -1,   8,  14, -18,  -2,   0,   4,   5,  12,  16,  14,   5,  -4,  -1,  -3,  -4, -10, -11, -12,   5,   5,   2,  -1,  -2,   5,   7,   8,  -5,  -3,  -1,   0,   1,   2,  -1,  -2,   3,   0,   0,   0,  -2,  -1,   1,   0,  -1,   0,   0 },
     {   -3,   2,   5,  -5,   1,  -3,   1,  -1,  -9,  -3,  -4,  -8,   2,  -1,  -1,   1,  -8,   9,  -2,   1,  -6,   1,  -2,   1, -20,  11, -23,  -2,   3,   3,   1,   2, -13,  40,  -7,   9,  -6,  -3,   0,  -4, -25,  12, -35, -28,  22,  -6,   6,  -3,  11,  22,  60, -52, -15,   2,  -3,  -1,   5, -18,  31,  31, -12,  -2,  -2,   0 },
     {   -7, -24, -34,   0,  10,   2,  -1,   3, -10, -38, -34,   0,   5,   4,   3,   6,  -9, -34,  14,  -3, -10,   5,   5,   0, -15, -24,  40,   0,   8, -12,  -5,   2, -18,  20,  45,  21, -13,   4,  -5,  -3, -12,  23, -30,  27,   7,   0,   6,  -3,  -9, -32,  -8,  -8,   9,   5,  -2,   4,   6,   0, -13,   5,  -5,  -1,  -1,  -2 },
     {    7,   5,  50, -18,  -4,  -7,  -1,  -2,   9,   4,  63, -21,   0,  -3,  -5,  -5,  11,   4,  51,   0,  -6,  -4,  -1,   2,   4,  -9,  17,  16,  -5,   2,  -5,   0,   4,  -2,  11,  42,   0,   4,  -1,   0, -15, -28, -24,  11,  13,   1,   4,   3, -23, -16,   2,   1,  -2,   1,   0,   1,  13,  -6, -18,  15,   1,   0,   0,  -1 },
     {    0,   5,  -8, -31,  24,  -4,  10,   1,   3,  10,  -2, -13,  -3,   1,  -3,   2,  -1,  -2, -13,   3, -33,  12,  -5,   3,  11,   9,  28,  59,  -8, -30,   5,  -7,   7,   0,   2, -39,  55,   7,  -6,   4, -12,  -5, -28,  -7, -29,  20,   4,   1,   1,  -1,  10,   0,  -3, -16,   1,  -1,   1,   4,   8,   0,   4,   1,  -3,   0 },
     {    4,   8,   9,  29,  34, -45, -16,  -1,   5,   9,  13,  13,  17,   9,  -2,  -5,  -2,  -3,  -9, -41, -54,  51,  20,   1,  -3,  -8, -19, -18,  13, -26,   2,   6,   1,   0,  -1,  26,   7,  -8, -16,  -3,  10,   6,  14,   1,  -1,   8,   1,  -7,   5,  -2,  -3,  -7,  -2,   3,   6,   2,  -2,   2,  -2,  -5,   0,   2,   1,   2 },
     {    1,   4,   3,  -4,  -7, -11, -14,   9,   2,   7,  -1,  -9,  -7,  10,  16, -14,   7,  24,  21,  16,   2,   2,  -3,  -4,  10,  20,   2,  -9, -15,  -4, -11,  -1,  28,  32,  22,   7,   6,   4,   2,  -7,  51,  20, -24, -17,   3,   0,   4,   4,  50, -39,   0,   7,  -9,   9,  -7,  -2, -37,  28,   4,  -8,   3,  -5,   1,  -1 },
     {   -1,  -2,  -4,   7, -24, -49,  57, -19,   0,   1,  -2, -14,  16,  54, -60,  20,   0,   1,  -1,   4,   7, -22,  23, -14,   0,   1,  -1,  -2, -14,  -6,   2,   4,   1,   2,  -2,  -3,   2,  11, -14,  -1,   3,   1,  -1,   4,   1, -12,  12,  -1,   3,  -3,   0,   3,   0,   4,  -7,   4,  -1,   3,   1,  -3,  -1,  -2,   2,  -1 },
     {   -4,   2,  11,   2,  -1,   2,  -7,  -3,  -4,   6,  22,   3,   1,  -3,  -3,   2,  -7,   5,  15,  -7,   3,  -6,  14,  -5,  -8,  19,  21,   5,   1,  -7,  -8,   6, -15,  22,   2,   3, -10,   7,   3,  -5, -22,  39,   8,  19, -23,   3,   0,   0, -17,  37, -14,  13,  19,  -7,  -1,  -4, -78, -23,  28,  -7,   9,   7,   2,  -1 },
     {    3,  -6,   5, -17,  19,   2,  -3,  -7,   6,  -8,  13, -25,  25,  -8, -11,   1,   5, -12,  18, -38,  13, -10,   6,   6,   8, -12,  34, -37,   8,   8,  -5,  -5,   9, -11,  35, -41,  -6,   1,  -1,   4,  10,  -9,  37, -10, -20,   9, -10,   0,  25,  -3,  37,   0, -11,  13,  -1,  -3,   0, -19,  17,  15,  -6,   6,   0,   6 },
     {   -2,   0,  -8, -45,  18, -16,  22,  -3,  -3,  -1, -13, -55,  14, -18,   4,  -1,  -2,   2, -12,  -3,  -6,  15, -20,   5,   0,  -1,  -8,  28, -27,  13,   2,  -1,   1,  -6,  16,  30, -17, -24,   4,  -5,   7,  15,  31, -15,  35, -13,  -9,  -2,  -1,  13, -21,   3,   9,   5,   7,  -1,  -2,  -6,  20, -24,  -3,  -3,   3,   2 },
     {    1,   4,   2,  11,  18,  -3,  -6, -10,  -3,  -4,  -8,  -2,  -2,   7,   2,   3,   4,   5,  15,  26,  11,  -7, -14,   0, -10, -15, -11, -44, -63,  11,  18,   0,  -4,   0,  12,  11,  57, -41,  -9,  -2, -10,  13,   4,  19, -16,  25,  -9,   1,  -9,   9,  -3, -10,  -7,  -8,   5,   1,   1,  -7,  -2,   8,   9,  -1,  -6,   2 },
     {    1,  -1,  -4,   6, -12,   1,   4,   2,  -2,  -5, -12,  -6,  -7,  12,   3,   2,   3,   0,  -4,   6,   7,   0,  -9,  -5,  -1, -11,  -9,   3,  35,  -7, -11,   0,   9,   3,  36,  30,  16,  13, -14,  -4,  -1,  -1,  15, -54,  -9,  12,  -5,  -3,  -4,  20, -26,  26, -32, -35,  -6,   4,  -1, -18,  -4,  47,  28,  -3,  14,  -2 },
     {    5,   7,   5,  11,  24,  45,  30, -13,   1,   2,  -1,   7,  -1, -12,  -6,   5,  -5,  -7, -14, -14, -53, -53,  -5,  11,   0,  -2, -13,  -4,   0,  22,  -7,  -6,   7,   2,   2,  29,  26,  25,   5,  -5,  13,  -2,   9,  12,  -8, -36,   2,   4,   5, -20,  -1, -10,  -8,   6,  -8,  -3, -12,  -7,  13,  -2,  15,   2,   2,   0 },
     {    0,   0,  -9,   7,   1, -20,  20, -27,  -3,  -4, -14,   4,  -4, -27,   2, -15,   1,   4,  -4,   1,   0,  21, -20,  16,  -2,  -2,  -5,   5, -14,  11,  14, -11,   5,  10,   1,  13, -22,  13,  -8,  15,  -4, -14, -16,  29, -13, -11,   5,   8,  24,  20,  46,  53, -12,  -5,   7, -19, -11, -16, -34,   2,  25,  11,  -4,   9 },
   },
   {
     {  111, -51,  -4,  -4,  -1,  -1,  -1,   0,  17,   6,  -9,   1,  -2,   0,  -1,   0, -19,  20,  -3,   0,   0,   0,   0,   0, -11,   6,   3,  -1,   0,   0,   0,   0,  -6,   1,   1,   1,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   20,  14, -11,   0,  -2,   0,  -1,   0, -98,  52,   5,   0,   2,   0,   1,   0, -31, -12,  22,  -4,   4,  -1,   2,   0,  19, -30,   8,   2,   0,   1,   0,   0,  13, -11,  -4,   3,  -1,   1,   0,   0,   6,  -2,  -4,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {  -14,  45, -23,   3,  -5,   1,  -1,   0,  42,  38, -27,  -3,  -5,  -1,  -1,  -1, -61,  52,  11, -10,   3,  -3,   2,  -1, -33,  -9,  31,  -6,   3,  -1,   1,   0,   1, -22,  12,   3,  -1,   1,   0,   0,   3, -11,   1,   3,   0,   0,   0,   0,   2,  -4,   0,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0 },
     {  -41, -69,  58,  -9,  12,  -2,   4,   0, -19, -19,   6,   7,   1,   2,   0,   1, -52,  39, -17,   8,  -1,   2,   0,   0,  -2, -10,   6,  -4,   3,  -1,   1,   0,  20, -14,  -1,   1,  -1,   1,  -1,   0,   8,  -1,  -5,   2,  -1,   0,   0,   0,   2,   2,  -3,   0,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0 },
     {    4,  -3,   0,   1,  -1,   1,   0,   0,  49, -22,  10,  -6,   0,  -1,   0,  -1, -33, -40,  43,  -2,   4,   0,   1,   0,  38, -66,   4,  17,  -4,   4,  -2,   2,  21,  -2, -34,  16,  -3,   2,  -1,   1,  -2,  17, -19,   0,   2,  -2,   1,   0,  -2,   9,  -4,  -3,   1,  -1,   0,   0,  -1,   3,  -1,  -1,   0,   0,   0,   0 },
     {  -15, -28,  -3,   9,   2,   3,   0,   1,  17,  54, -58,   9,  -6,   2,  -2,   0,  38,  27, -18, -13,   1,  -3,   1,  -2,  66, -16,   5, -13,  -1,  -1,  -1,   0,  15,  16,  -4,  -1,  -5,   0,  -1,   0, -10,  13,   3,  -2,   0,  -1,   1,  -1,  -6,   4,   6,  -3,   1,   0,   0,   0,  -2,  -1,   3,  -1,   0,   0,   0,   0 },
     {   -5,   8, -32,  23,  -4,   3,  -1,   1,  -4, -14, -21,  23,  -2,   3,  -1,   1, -63,  -7, -26,  23,   4,   4,   1,   1,  18,  19, -57,   3,   7,  -2,   2,  -1,  -3,  46, -18, -24,   9,  -4,   3,  -2,  -5,  12,  18, -21,   2,  -1,   1,  -1,  -1,  -2,  13,  -5,  -3,   1,   0,   0,   0,  -2,   4,  -1,  -1,   0,   0,   0 },
     {  -11, -33, -61,  50,  -5,  10,  -2,   3, -20, -45, -38,  27,   9,   3,   2,   1,   9, -13,   1, -19,  14,  -4,   2,   0, -25,  -6,  26, -16,   3,   1,   0,   1,  15, -26,   4,   5,  -6,   3,  -2,   1,  13,   4, -16,   6,  -3,   0,   0,   0,   2,  10,  -8,  -2,   1,  -2,   0,   0,   0,   4,  -1,  -1,   0,   0,   0,   0 },
     {   27,  51,  -6, -17,   0,  -4,  -1,  -2,  -8, -41,  16,   8,  -1,   3,   1,   1,  10,  38, -53,  15,  -5,   1,  -1,   0,  29, -10,   4, -12,   6,  -3,   1,  -1,  57, -28,  -5,   0,  -5,   3,  -2,   1,   4,  22, -13,   2,  -2,   0,   0,  -1,  -5,  10,   1,  -3,   1,  -1,   0,   0,  -3,   3,   2,  -1,   0,   0,   0,   0 },
     {   -3,   2,  13,  -1,  -1,  -2,   1,   1,  -6,  24, -16,  15,  -5,  -1,  -1,   0,  26,   9, -15,  16,  -3,   0,  -1,   0, -65, -25, -11,  29,   4,   4,   2,   1,  -2,  -3, -62,  11,  10,   0,   4,   0,  -4,  40, -22, -29,  11,  -3,   2,  -2,   1,   8,  14, -20,  -1,   1,   0,   0,   2,  -1,   8,  -2,  -4,   1,   0,  -1 },
     {    6,   4,  17,  -6,   1,  -5,   0,  -2,  17,  24, -25,  29, -10,   1,  -4,   1,   0, -47, -35,  44,   3,   3,   2,   1,  -5, -27, -22,  -3,  25,  -3,   4,   0, -10, -40,  23, -20,  12,   5,   1,   2,  25, -36, -12,   9, -10,   8,  -2,   2,   6,  12, -28,   7,  -3,   0,   1,  -1,  -1,  10,  -7,  -4,   2,  -2,   1,  -1 },
     {   -8, -22, -23,  14,   0,   4,   0,   0,  16,  33,  19, -38,   8,  -5,   1,  -2,   6,   2,  19,   1, -13,   6,  -3,   2,  -5,  38, -44,  24,  -7,   0,   2,  -1,  64, -28,  -2, -13,  10,  -6,   2,  -1,  39,  -4,  -8,  -4,  -3,   3,  -4,   2,  -6,  18,  -2,  -1,  -2,   1,  -1,  -1,  -3,   3,   3,  -1,  -1,   0,   1,   0 },
     {   -4,   3,  -6,  19, -17,   7,  -4,   2,  -5,   3,   3,  20, -21,   7,  -4,   2,  -6,  17,  -2,  26, -24,   3,  -2,   0,  19,  30,  -5,  39, -24,  -2,  -1,  -2, -20,   2,  -4,  54,  -7,  -7,   3,  -2, -15, -13, -41,  37,  17,  -9,   5,  -1, -10,  11, -35,  -3,  22,  -6,   3,  -1,  -5,   6,  -7, -12,   9,   1,  -1,   1 },
     {   -7,  -6, -37, -53,  67, -23,  14,  -3,  -7, -11, -38, -26,  29,  -2,   2,   1,  -7, -11,  -7,  21, -35,  25, -10,   3,  -2,  -1,  11,  14, -23,   6,  -1,  -1,  -8,   3,   6,   8,  -4,  -7,   5,  -3,   0,   2,  -3,   1,   2,  -5,   2,   1,   1,   2,  -2,  -1,   1,  -1,  -2,   2,  -2,   1,   0,  -2,   2,   0,   0,   1 },
     {    5,   5,  -1,   4,  -2,   1,  -2,   0, -10, -21, -14,   6,   3,   0,   2,   1,  23,  38,  18,  -6, -12,   0,  -3,  -1, -17, -43, -18,  42,  -1,   2,   3,   2,  23,  29,  -9, -24,   9,  -8,   2,  -2,  -6, -39,  42,  18,  -6,   6,  -3,   1,   8, -14, -41,  24,   0,   2,   1,   0,  -4,  12,  -9,  -8,   5,  -2,   1,  -1 },
     {   14,  29,  49, -15,  -9,  -4,  -3,  -1,  -8, -26, -60,  23,  14,   3,   1,   0,  -4,  -4,  31, -38,   6,   1,  -2,   1,  14,  33,  -8,  16, -20,   0,   1,  -2,   1, -13,  -5,   6,  14,  -8,   4,  -1,  36, -11,  -4, -15,   5,   3,  -5,   3,  -2,  21,   8, -10,  -6,   3,  -2,   0,   1,  -1,  12,  -1,  -4,   0,   1,  -1 },
     {   -3,  -7, -25,  13,   5,   4,  -1,   1,  -3, -14,  23, -15,  10,  -4,   1,   0,   9,  43,  -2,  -8,  -6,   2,   0,  -2,  36, -22, -15,  18,  10,   1,   3,   0, -67, -37,  -8, -10,  27,   3,   8,   3,  19, -21, -11, -32,   0,   5,   0,   3,  16,  11,   9, -15, -12,   2,  -3,   1,   5,  -2,   9,   3,  -8,   1,   0,   1 },
     {    1,   7,   5,   1,  16, -13,  -1,   1,   1,   7,  -3, -24,  52, -26,   9,  -8,  -6,   5, -26, -32,  54, -13,   5,   1,  -5,   2, -37, -11,   9,  20, -13,  13,  -9,   5, -25,  17, -21,  21,  -6,   2,  -9, -14, -23,  25,  -6,   5,   6,  -3,  -5,   1, -26,  -1,   8,  -2,   5,  -2,  -4,   4,  -5,  -8,   1,   1,   0,  -1 },
     {    7,  10,  16,  49,  49, -35, -11, -10,   8,  11,  22,  52,  38, -26, -20,  -4,   8,  13,  31,  30,   1,   6, -15,  -3,   8,  12,  17,  -1,  -3,  14,   3,  -6,   7,   4,   7, -13,  -2,   8,   8,  -1,  -5,   5,  -4,  -9,  -4,   7,   4,   3,  -4,  -4,   0,  -2,  -6,   6,   0,   3,   0,  -4,   0,   0,  -2,   0,   1,   1 },
     {   -4,   1,  -4, -17,   0,  -1,   5,  -5,  -6,  -1,   3,  -3, -12,   9,  -5,   1,  -1,  13,  22,  -3,  -4,  -2,  -4,   0,  -5,   6,   8,  -5,  19,  -7,   2,   0, -15,  30,   9, -32,   6,  15, -13,   8,  -2,  32, -29,   9, -45,  28,  -6,  -1, -19,  57, -30,  -2, -16, -15,  15,  -9,  -3,  12,  22, -22,  11, -17,   5,  -2 },
     {    2,   1,  -8,  -9,   5,  -1,   2,   3,   5,   5,  13,   8, -11,  -1,   0,   0,   5,   6,  -3,   7,  -2,  -4,   4,  -4,   3,   5,  23, -19,   7,  -2,  -1,  -1,   0,  53, -28,  10, -18,   5,  -1,  -2,  85, -15, -17,   1,  -2,  -8,   4,  -3,  49,  -9,  -7,  -3,   4,  -3,  -3,   0,   1,   9,  -5,  -2,  -1,   0,  -1,   0 },
     {   -3,  -5, -10, -16,   5,  -3,   9,   0,  -2,  -5,  -7,  -5,   7,  -9,   7,  -6,   2,   0,  13,  21,  19, -28,   0,   1,   9,  20,  23,   8,  44, -39, -15,   8,  13,   9,  -8,  14,  54,   0, -28,   2, -11,  -8,  -8,  -7,  35,  31, -19,  -2,  -3, -17, -12, -10,   1,  29,   1,  -5,  -2,   1, -13,  -2,  -7,   4,  10,  -5 },
     {   -1,   1,  -1,   2,  -2,  -2,  -4,   3,  -1,   0,  -7,   2,   8,  -3,  -1,   4,  -1,   4,  -3,   8,   5,  -8,  -3,   1,   5,   8,   0,  24,  10, -10,  -3,  -3, -14,  -8, -12,  22,   8,   4,  -4,   2,  16,  29,   2,  19, -31,   2,  -5,  -1,   0,  -1,  21,  76, -27,  -7,   0,  -5,  -8,  -5, -30,  57,  14, -12,   4,  -5 },
     {    3,   3,  17,  25, -44,  20,  -3,  10,  -2,  -1, -12, -40,  54, -28,  12,  -3,  -3,  -6, -25,  38,   7,  -7,  -6,   2,   2, -12,  36,  27, -22,  -7,  10,  -2,  -3,  22,  25, -12,   3, -12,   7,  -1,  11,  14,  -3,  -1,  -3,  -4,  -7,   4,   8,   9,  -2,  -7,  -4,  -1,  -5,   2,   3,  -4,  16,  -6,  -1,  -1,  -3,  -2 },
     {    5,   8,  16,  33, -10, -30,  17,  -4,   0,   1,  -6, -20,  10,  11,  -1,   0,  -2,  -2,  -3, -20, -31,  47,  -8,   7,  -2,  -5,  -5, -43,   2, -17,  17,   2,  -7,   2, -24, -18,  35, -35,  -1,   6,  -6,   8, -31,  33,   3,  -1, -17,   3,  12, -20,  -4,  11,   3,   8,  -4,  -5,   3,  -4,  -7,  -4,   5,  -1,   2,  -4 },
     {    2,   2,   6,  -1, -10,  -8,   8,   0,  -8, -16, -27, -18,   0, -10,   0,   5,   9,  16,  42,  34,  -9, -19, -15,  -3,   0,  -5, -29, -11,  42,   2,  -7,  -9,  -3,  -8,  12, -10, -11,  19,  -6,  -7,   1,  11,  -6,  35,   1, -22,  18, -12,  16,  -6,  27, -17,  35, -23,  -2,   0,  13, -27,   5,   8,  11,  13, -13,   5 },
     {    1,   2,   4, -10,  12,  58, -54,   5,   0,   1,  -2,  -9,  29,  52, -38, -14,   2,   2,   5,   8,  14,  27, -14, -19,  -2,   2,   5, -14,  19,  20,  -1,  -8,   3,   7,  -4,   1,  27,  -2,  12,  -7,  -2,  -1,  -8,  -1,   7,  -7,  14,  -1,  -3,   1,  -8,   6,  10,  -4,   4,   2, -10,  -7,  -6,   4,   6,   8,  -5,   0 },
     {   -1,  -4,  -3, -17, -16,   7,  -7,   9,   3,   3,  -6,  11,   5,  -4,   6,   5,  -2,  -4,   5,  -8,  -4,  15,   3,  -3,   9,  21,  -1,  18,   1,  15,   3,   1,  13, -13,   8,  -4,   2,  21,   1,  23, -31,  21, -17, -16, -25,  13,   6,  18,  70, -37, -18,  -2, -10, -13,  10,   4,  40,   1, -21, -17,  -6,  -4,   1,  -2 },
     {    3,   4,   9,  12,  -4,  18,   1,  -3,  -4,  -6, -16, -20,  -6, -10,  18, -29,  -3,  -7,   3,  20, -45, -31,   5,  -9,   3,  15,  16, -32, -17,  10,  -3,   2,   4,  -7, -47, -13,   6,  38,  17, -13, -18, -35,  -8, -14, -15,  16,  18,  -1, -15,  -9,  -1,  10, -18, -14,   8,  13,  -9,   1,  -5,   3,  -3,  -6,  -3,   0 },
     {    2,   4,   8,  15,  13,  11,   5, -20,   0,   0,  -4, -12,   4,   7,  -8,  13,  -4,  -6,  -6,  -3,  -8,  -6,  25,   6,   9,  11,  13,  12,   7,  -5,   8, -14, -19, -23, -21,  -7,  -6,  19,  -9,   4,  29,  36,  36,  -6, -14,   1,  -9,   4, -27, -24, -19,  13,  44,   4,  -1,   1,   2,  -9, -21, -54,  13,  13,   2,   5 },
     {    6,   9,  17,  41,  16, -34,  10, -11,  -3,  -4, -10, -36, -20,  42,   0, -38,  -2,  -5, -12,  -6,   3,  16,  28, -10,   1,   3,  12,  30,  30,   2, -19,  13,   7,  10,  19,  12, -11,  11,   0, -18,  -4,   1,   3, -20,   4,  -2,  19, -13,  15,  17,  -7,  -3,  -7,  -1, -14,  13,  18,   2,  -4,   8,   4, -11,   3,   4 },
     {   -5,  -7,  -9, -11, -43, -35,  14, -13,  -2,  -2,   1,  -4, -15, -24,  -9,  12,  -1,  -1,  -6,  -2,  13,  36, -41,   2,   4,   4,  18,  -2,   4,  47, -38, -13,   0,   4,  -2,  -2,  31,  21,   5, -13,  11,   4,   8, -17,  11, -12,  17,   4, -20,   3, -17,  14,   8,  -6,   6,  16,  -8,   1, -10,   0,  -6,   9,   3,   0 },
   }
 },
 {  //24
   {
     {  107, -15, -13,  -4,  -4,  -2,  -2,  -1, -61,  20,   8,   1,   2,   1,   1,   0,  -2, -13,   2,   1,   1,   1,   0,   0,   4,   4,  -4,  -1,   0,   0,   0,   0,  -3,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -36,  50,  -4,  -1,  -2,  -1,  -1,  -1, -78, -33,  23,   4,   6,   2,   2,   1,  60, -23, -17,   0,  -2,   0,  -1,   0,   2,  16,  -1,  -3,  -1,  -1,   0,   0,   3,  -5,   1,   1,   0,   0,   0,   0,   2,   2,  -1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
     {   42,  22,  -3,  -8,  -3,  -3,  -1,  -1,  29, -60,  -3,  10,   1,   3,   0,   2,  28,  63, -25,  -9,  -5,  -3,  -2,  -1, -48,   4,  33,   1,   3,   0,   1,   0,   7, -18,  -9,   5,   0,   2,   0,   1,  -5,   9,   1,  -3,   0,  -1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   18,  87, -19, -10,  -7,  -4,  -3,  -2,  17, -30,  20,   3,   1,   0,   0,   0, -70, -16,   1,   8,   6,   3,   2,   1,  31,  -8,  -7,  -5,  -2,   0,   0,   0,  -2,   8,   3,   0,  -1,  -1,   0,   0,   0,  -3,   1,   1,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -12,  15,  17,  -5,  -2,  -1,  -1,   0, -43,   6, -25,  10,   4,   2,   2,   1, -15,  37,  39, -20,  -1,  -4,   0,  -2,  -3, -76,   8,  28,   3,   5,   1,   2,  20,  14, -40, -12,   3,   0,   1,   0,  -1,   6,  18,  -5,  -4,  -1,  -1,  -1,   3,  -3,  -5,   4,   2,   0,   0,   0,   0,   1,   1,  -2,   1,   0,   0,   0 },
     {   -8,  31, -52,  10,   1,   1,   0,   1,   7,  65,  35, -24,  -7,  -5,  -2,  -3,   4,  -4,   9,  17,   1,  -1,   0,   0, -60, -16,  15,  -3,  10,   2,   3,   1,  26, -29,  -5,   2,  -3,   2,   0,   1,  -2,   9,  -5,  -3,   0,  -1,   0,   0,   2,  -5,   6,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,   1,   0,   0,   0 },
     {   20,  12, -12,   2,  -5,   0,  -1,   0,  36, -21,  -1, -13,   6,  -2,   1,  -1,  47, -27,  41,  15, -13,   1,  -4,   0, -22, -10, -63,  28,  13,   3,   2,   1,  -5,  34,  -2, -38,   1,  -1,   0,  -1,  -2,  -3,  18,  10,  -9,   0,  -1,   0,  -2,   4,  -7,   1,   3,  -1,   0,   0,  -1,   0,   2,   0,  -1,   0,   0,   0 },
     {   11,  47,  57, -19, -11,  -3,  -4,  -2,   7,  37, -62,  20,   5,  -2,   1,   0,   5, -30,  16, -17,   4,   3,   1,   1, -33,  32,   9,   2,  -5,  -2,   0,  -1,   4, -20,  14,   0,   4,  -1,   0,   0,   2,  -8,  -2,   4,  -2,   2,   0,   1,  -1,   0,  -1,  -1,   1,  -1,   0,   0,   1,  -2,   1,   1,  -1,   1,   0,   0 },
     {    8, -21,  38,  -8,   3,  -3,   0,  -1,  12, -23,  21,  12,  -3,   1,   0,   0,  -9, -63, -30,  13,  12,   6,   2,   3, -23, -20,   3,  -5,  -7,   3,   0,   1,  55,  12, -41,  12,  -4,  -3,  -2,  -1, -17,  33,  -1, -14,   5,  -1,   0,  -1,   0,  -5,   6,   2,  -1,   0,   0,   0,  -1,   3,  -4,  -2,   2,   0,   0,   0 },
     {    5,   3, -13,  -6,  -1,   2,   1,   0,  19,   5,   4,   4,  -1,  -3,  -1,   0,  17, -23,  11, -11,   0,   2,   0,   0,  28,  11,  20,  33,  -9,  -2,  -3,  -1, -50, -35, -57,   2,  25,   8,   6,   3,  25,  33,   5, -38, -18,  -3,  -1,  -1,  -6,   0,  18,  20,  -1,  -4,  -2,  -1,  -1,  -1,  -7,  -3,   4,   2,   0,   0 },
     {   -9, -10, -31, -72, -10,  12,   7,   4,   9,  17,  26,  75,   2, -16,  -8,  -3,  10,   4,  -4, -25,  18,  11,   2,   0,  -8,  -8, -13,  -6, -20,  -1,   3,   2,  -2,  16,  10,   1,   8,  -7,  -4,  -2,   0, -12,   8,   0,  -3,   6,   0,   0,   0,   2,  -6,   4,   1,  -2,   1,   0,   0,   0,   2,  -1,   0,   0,  -1,   0 },
     {    3,  -8,   8, -10,  -4,   3,   2,   0,   4, -22,  16,  10,   7,  -5,  -1,  -1,   7, -12,  46,   7,  -5,   3,  -2,  -1,   3, -28,  -4,  16,   3,   2,   0,   2,  -7, -44,   7,  37,  -4,   7,  -1,   1,   1,  -4, -72,  19,  18,   3,   4,   0,  -3,  25, -11, -37,   7,   0,   1,   0,   2,  -1,   9,  -2,  -3,  -1,  -1,   0 },
     {  -17, -29, -35,  -7,  -2,   9,   4,   3, -26, -39, -21,   9,  15,  -2,   3,   2, -47,  -4,  36,  -4, -13,   4,  -3,   0, -27,  47, -15,   6,   5, -10,   1,  -2,  37, -31,   2, -22,  10,   4,  -1,   2,  11,  14,   3,  -5, -16,   2,   0,   0,  -2,   0,  12,   7,  -1,  -4,  -1,  -1,   2,  -3,  -3,   4,   2,   0,   0,   0 },
     {   -5,  -4,  65, -13,  -1,  -7,   1,  -2, -13,   6,  61,   1,  -4,  -3,  -3,   0, -26,  24, -13,  23,  -6,   4,  -2,   0, -28,   2, -28,  14,  18,  -3,   4,  -1, -27, -25,   6, -28,  11,   5,   2,   2,  27, -11,   5,  -3, -18,   2,  -1,   0,   2,   2,   0,  11,   1,  -3,   0,  -1,   1,  -3,  -2,   0,   2,   1,   0,   1 },
     {    1,   0,  -1,  38,  -4,  -4,  -4,   0,   5,   3,  24,  -7,  24,  -4,   1,  -3, -10, -13, -21, -60, -26,  20,   5,   5, -11,   1,  12,  52, -20, -29,  -4,  -1,  12,  18,  16,   4,  39,   4,  -8,  -4,  -9, -24, -10,  -1,  -7,  18,   7,   2,   3,   3, -13,  -6,  -3,  -6,   2,   2,   0,   1,   5,  -1,  -2,  -1,  -2,   0 },
     {    2,   4,  -5,   2,  73, -28,  -4,   0,  -1,  -3,   3,  18, -70,  34,   2,  -3,  -3,  -5,   4, -35,  12, -26,   6,   2,  -5,   8, -16,  24,  10,   9, -10,   0,   6,  -6,  13,  -3,  -5,   3,   8,  -2,  -1,   2,  -7,  -6,   2,  -2,   0,   3,   2,   1,  -1,  -1,  -3,   1,   0,  -2,   1,   0,   2,   0,  -1,  -2,  -1,   1 },
     {    0,   1,  11, -33,  -1,   1,   0,  -1,   3,   7,  21,  -9, -19,  -6,   0,  -1,   8,  14,  31,  36,  -9, -14, -12,  -4,  29,  42,  32,  29,  11,  -3,  -6,  -7,  31,  12, -19,   6,  21,  16,   4,   1, -50, -31,   7,  -1,   0,   9,   8,   5,   3, -14, -19,   2,  -3,  -2,   1,   2,   0,   6,  -2,  -5,  -1,  -1,  -1,  -1 },
     {    1,   5,   6,  -2,  -4,  -1,   0,   0,  -2,  -2,  -3,  -6,   2,   3,  -2,   0,   4,  13,  14,   5,  -3,  -4,  -2,  -2,  -7,  -3, -26, -13,  -6,   2,   0,   2,   0,  16,   8,  25,  15,  -8,  -1,  -3, -12, -14, -18, -61,  13,  16,   6,   3,  22,  11,  45,   7, -62, -10,  -4,  -1,  -5,   2,  -5,  39,  22,  -9,  -2,  -2 },
     {   -7,   8, -14,  14,  -3,  -1,  -1,   2, -12,  16, -26,  28,   3,   2,  -3,   1, -15,  20, -37,  31,   0,   0,  -1,   0, -15,  22, -37,  26,  14,  -2,   3,  -3, -16,  20, -32,   6,  24,  -3,   4,  -2, -17,  12, -22, -17,  33,  -6,   6,  -2,   0,   0,   4, -42,  27,   0,   1,   1,   4,  -7,  13, -22,   4,   3,  -2,   2 },
     {   -5, -11, -15,   4,   8,   6,   3,   1, -10, -23, -14,  -1,  -3,   6,   4,   3, -16, -32,   1,  25,  23, -11,  -2,   1, -43,  -8,  39,  -4,  -1,  17,  -6,  -4, -35,  21, -16,   2,  12,   0,  10,  -3,  28, -62,   5, -10,  -1,   6,  -1,   8,  13,  16, -28,  -2, -11,   1,  -2,   1,  -1,   9,   9,  -8,   0,  -2,   0,  -2 },
     {    4,   4,  11,  73,   3,  -9, -11,  -4,   7,   3,  30,  56,  20,  -5,  -4,  -5,   5,  11,  39, -10,  21,   5,   0,  -2,   3,  24,  -4, -29,  -9,  20,   5,   3,   4,  -7, -27,  -1, -11,  -5,   8,   5, -12, -20,  10,  -2,   3,  -7,  -3,   1,  -1,  -7,  -4,   3,  -1,   6,  -1,  -1,   1,   1,  -5,  -2,  -4,   0,   2,   1 },
     {   -6,  -1,  10, -22,   7,  -1,   1,   0,  -9,  -3,  22, -36,   1,   6,   2,   0, -15,   8,  30, -21, -14,   9,   2,   0, -31,  30,  -4,   8, -47,  13,  -1,   5, -27,  32, -15,  40, -28, -25,   6,  -8,   3,  13,  10,  13,  27, -22,  -1,  -3,   0, -18,   7,   1,  18,   4,  -4,   1,   3,  -5,  -3,  -5,   4,   2,   0,   0 },
     {   -3,  -5,  -5, -12, -21, -18,   9,   2,   5,   8,   8,  19,  39,  34, -13,  -4,  -5, -13, -10, -18, -50, -33,  16,   5,   1,   2,  10,   7,  46,  12, -27,  -8,  -3,   5,  -3,   2, -33,  18,  29,   3,  -2,  -7,   3,  -2,  10, -30, -14,   4,  10,   4,   3,   1,  -9,  15,   0,  -5,  -2,   1,  -2,   8,  13,  -9,   1,   2 },
     {    5,   8,   9,   9,  42,  81,   4,  -8,   0,   1,   3,  12,  -2, -46,  16,  12,  -4,  -7,  -5, -12, -46, -30, -26,   5,  -3,  -4,  -2, -14,  14,  13,  -3, -15,  -6,   5,   0,  11,   4,   1,   7,   4,   1,   5,   5,   1,   1,   2,  -2,   0,  -5,  -6,   3,  -2,   1,  -3,  -1,  -1,   0,  -1,   0,   1,   0,  -1,   0,   1 },
     {    1,   2,  -2,  13,  -3,  -2,   4,   1,  -1,   0,  -7,  10,  -2,   9,  -4,   1,   2,   3,   1,  14,  -8,  -9,   1,  -4,  -1,  -6, -10,   0,  -4,   2,  -1,   0,   6,   0,  13,  31,  10,  -6,   0,  -2,  -2,   2,  -6,   8,  -7,   9,   0,   2,  -9,  18, -20,  59,  23, -21,   2,  -6, -10,  -9, -21, -34,  82,  11,  -1,   4 },
     {    1,   3,   4,   6,  -9,  -3,  -4,   2,   1,   4,   8,  10,  -9,  -8,  -1,  -1,   6,   8,   9,   9,   5, -10,  -3,  -6,   1,  11,  28,   9,   7,  -1,  -3,  -3,  23,  52,   5, -10,  -1,  10,   0,   2,  57,   2, -36,   0,   1,  -2,   2,   3, -68,   3,  32,   1,   3,   4,  -6,  -3,  10, -18,  -9,  10,   2,  -2,   1,   0 },
     {    5,   8,   0, -17,  14,  12,   4,  -3,   6,  14,  -1, -24,   4,  18,   2,  12,   7,  15, -18, -15, -17,  13,  10,  -2,  17,  15, -28, -20, -19,   5,  -8,   8,  36, -15, -53,  -9,  -9, -15,   8,   8,  24, -36,  -8,   1,  -2,   6,  -5,  -1, -21,  49, -16,  -3, -10,  -6,   1,  -6,  -6,   2,  20, -12,   0, -11,   0,  -2 },
     {    0,   3,   1,  -2,  32,  -7,  -4, -11,   0,   2,  -2,  16,   6,  28, -10,   5,   3,   0,  17,  23, -41,   6,  39,   7,  -8,  -9,   2, -43,  -5, -46, -19,  12, -12,   9, -11,  15,  17,  26, -33, -22,   4,   0,   8,   5, -16,   5,  31,  -5,  -8, -11,   5,  -4,   8,  -7,  -5,  14,   4,  -2,   0,  -2, -15,   3,  -2,  -3 },
     {    0,  -2,   0,   3,  -8,  -5,  -1,  -1,   2,  -1,   1,   6,  -1,  10,  -3,   2,   2,   0,   7,  11, -13, -13,   5,   6,   4,  -2,  -6,   4,   3,  -3,  -7,  -2,   5, -14,  10,  16,  23, -30,  -7,  -4,  28,  14,  28,  -2,  48,  44, -25, -12, -22, -21, -27,   3, -27,  38,  35,   5,   3,   5, -10, -16, -13, -28,   2,  10 },
     {    2,   0,   5, -24,  44, -24,   4,  -2,   4,  -2,   8, -32,  58, -21,  -3,  -3,   5,   1,  11, -28,  39, -12,  -1,   2,   3,   1,   0, -27,  32,   3,  -6,  -5,   3,   3,   3, -10,  23,  10,  -3,   1,   0,   0,  -1, -22,  19,  12,  -2,   1,  -6, -10,   4, -13,  18,   3,  -2,   3,   2,  -7,   2, -29,  19,  11,  -3,   4 },
     {   -4,  -6,  -5, -11,  -6, -14, -76, -13,   3,   4,   5,   5,  18,  27,  65,   0,  -1,   1,  -3,   6,  -2, -26, -27,  25,  -1,  -7,   0,  -8, -22,   4, -13, -16,  -5,  -4,  -1,  -9,  14,  -5,   5,   2,  -8,   2,  -2,  15,  -4,  10, -10,  -2,  -3,   4,  13,  -3,  10, -14,   3,  -3,   2,  -3,   5,  10,   3,  -1,  -2,   0 },
     {    0,  -2,   3,   6, -56,  28, -14,   3,  -1,  -6,   5, -19, -33,  -5,   2, -13,  -1,   3,   2, -34,  34, -28,  14,  15,  -5,   5, -28, -14,  19,  -4, -35,  -3,   5,  -7, -12,  26,   4,  29,   1, -30,   8,  -7,  15,  -2,  -6,   3,  22,   6,  -6,  -4,  -2, -11,   7,   4,   0,  14,   5,   3,   3,  -7,   3,   9,  -6,   6 },
   },
   {
     {  112, -31,  -9,  -4,  -3,  -2,  -1,  -1,   6,  27, -11,   0,  -3,   0,  -1,   0, -32,  23,   5,  -2,   1,  -1,   0,   0,  -8,  -1,   8,  -2,   1,  -1,   0,   0,  -4,  -2,   6,   0,   0,   0,   0,   0,  -1,  -3,   3,   2,  -1,   0,   0,   0,  -2,   0,   0,   2,  -1,   0,   0,   0,  -1,   0,   0,   1,  -1,   0,   0,   0 },
     {  -16, -11,   3,  -1,   1,   0,   1,   0,  86, -24, -16,  -1,  -4,   0,  -2,   0,  13,  57, -23,  -3,  -4,  -1,  -2,  -1, -33,  33,  23, -11,   2,  -3,   1,  -1,  -4, -16,  23,   3,  -1,   0,   0,   1,  -3,  -7,  -1,   8,  -1,   1,   0,   0,  -1,   0,  -3,   2,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   26,  -6,  -6,   0,  -2,   0,   0,   0, -22, -21,   6,  -1,   2,   0,   1,   0,  51, -30, -28,   5,  -5,   2,  -2,   1,   3,  67, -35, -14,  -1,  -3,   0,  -2, -12,  24,  42, -23,  -1,  -4,  -1,  -2,   4, -16,  23,  15,  -8,   1,  -2,   1,  -3,  -1,  -6,  11,   2,  -1,   1,   0,  -1,   0,  -1,   0,   2,   0,   0,   0 },
     {   -6, -24,   4,  -2,   1,   0,   1,   0,  43,  -7, -18,   0,  -3,   1,  -1,   0, -13,  -3,   1,  -2,   1,   0,   1,   0,  14, -15, -38,   9,  -2,   3,  -1,   2,  -8,  57, -34, -32,   5,  -6,   2,  -3,   2,  11,  49, -36,  -9,  -3,  -2,  -1,   7,  -7,  24,  19, -21,   1,  -3,   0,  -3,   3,  -3,  18,  -1,  -4,   1,  -1 },
     {  -31,  11,   6,  -1,  -1,   0,   0,   0,   8,  57, -15,  -6,  -4,  -3,  -1,  -1, -41,   4,  32,  -7,   1,  -1,   1,   0,  10,  -3,  -8,  -6,  -1,  -2,   1,   0,   2,  13,  50, -22,  -6,  -3,  -2,  -1,  10, -37,  28,  39, -14,   3,  -4,   2,  -6,   0, -28,  22,   8,  -2,   3,   0,  -1,  -1,  -4,  -6,   7,   0,   1,   0 },
     {    0, -45,  14,   6,   4,   2,   1,   1,  38, -14,  -8,   9,   4,   0,   0,   0,  -5, -34,  13,  16,   2,   4,   0,   2,  14,  -9, -39,  19,   3,   4,   1,   0,  -4,  19,  -4,   8,   3,   2,  -1,   0, -12,  -1, -36,  29,  19,   0,   2,   0, -12,  12, -30, -38,  35,   1,   5,   0,   5,  -6,  14, -34,   1,   8,  -2,   2 },
     {   11,   3,   3,  -1,   3,  -2,   1,  -1,  -5, -32,  14,  11,  -1,   3,  -1,   1,  14,  11,  -4,  -1,   4,  -3,   0,  -2,   1, -20,  21,  18,  -4,   4,  -2,   1,   1,   5, -27,   9,   8,  -1,   1,  -1,   6,   2,  29,  11, -10,   4,  -6,   0, -12,   2, -37,  58,  27,  -8,   4,  -3,  -7,  -3, -17, -45,  55,   7,   1,   2 },
     {  -29,-105,  16,   2,   7,   3,   3,   2, -37,   3, -12,  -1,  -3,   2,   0,   1,   6,  15,  -6,  -9,  -1,  -3,   0,   0,  21,   2,  22, -12,  -5,  -2,  -2,   0,   7, -13,   8,   9,  -5,   1,  -1,   1,   8,   1,  -2,   0,  -9,   1,  -1,   0,   2,   5,   5,  16,  -9,  -3,  -2,  -1,  -2,   3,  -7,  11,   4,  -4,   1,  -1 },
     {  -13,   5,  -8,   3,   1,   2,  -1,   0, -53,   0, -50,  21,   6,   5,   2,   1,  -3,  61, -19,  -9,   2,  -4,   0,  -2, -28,   3, -15,  10,  13,  -1,   4,   0,  -4,  27, -14, -24,  10,  -4,   2,  -3,  -8,  -2,  11,  -1,  16,  -4,   2,  -1,  -1,  -5, -10, -34,  20,   3,   2,   1,   6,  -6,  19, -24,  -5,   6,  -3,   2 },
     {   -3,  -5,   4,   1,   2,   0,   1,   0, -10, -45,  12,   4,   3,   3,   1,   1, -66, -15, -31,  18,   2,   6,   2,   3, -37,  -9, -29, -21,  12,  -3,   4,   1,  28, -33,   8, -38,  -3,  -2,   2,   3,  18,   6,  -7,   0, -26,   6,  -1,   1,   0,  19,  -4,   6,  -8,  -3,   0,  -3,   4,   3,   8,  -2,  11,  -4,  -2,  -1 },
     {   19,   4,  36, -16,  -4,  -2,  -1,  -1,  -6, -51,  -2,   0,   4,   2,   3,   2,  33,  14,  38, -20, -12,  -2,  -3,  -1,  -7, -45,   5,  14,   2,   4,   1,   3,  27,   7,  12, -21, -20,   2,  -4,   0,  -2, -14,  11,  35,  -9,   1,  -2,   2,   3,   0, -17, -12, -16,   5,   1,   2,   0,  11,   6,  33, -27,  -9,   1,  -3 },
     {   -2, -14,   1,   3,   0,   1,   0,   1,   3,  35,  -4,  -4,  -1,   0,   0,  -1,  44, -10,  34,   1,  -2,   0,   0,   0, -55,   0, -19,  34,   3,   6,  -1,   1, -26, -38, -19, -28,  20,  -4,   5,   2,  19, -18, -13, -19, -24,   7,   1,   3,   1,  26,  -5,   5, -13,  -4,  -1,  -4,   5,   1,  15,   0,  18,  -4,  -6,  -1 },
     {    0,   5,   3,   9,   1,   1,  -1,   0,  -3,   3,  -8,  13,   6,   1,   1,   0,  -2,   6,   8,  11,   0,   1,  -1,  -1,  -9,   6, -10,  14,   9,   2,   2,  -1,   6,   7,   7,   7,   2,   2,   2,  -3,  -5,  22,  -7,  27,  15,  13,  -4,  -6,  13,   5,  35,   6,  52,  -5, -22,  -2,  -1,  11,  -2,  58,  54, -48,  -3,  -6 },
     {   -2,  -5,  -6,  -6,   8,  -2,   1,   0,   3,   5,  10,   8, -12,   2,   0,  -1,   1,  -5,  -1,  -9,  10,  -2,   2,   0,  -2,   5,   9,  14, -13,   2,  -2,   0,   0, -16, -10, -17,  17,  -6,   7,   2,  11,  26,  16,  15, -34,  18,  -2,  -6,   2,  -2,  16,  12,  67, -15, -13,  -1,  -3,  -7,  -6,  -5, -76,   2,  10,   3 },
     {    5,  12, -17, -13,  -6,  -1,   1,   0,  10,  19, -26, -17,  -8,   0,   1,   1,  28,   3,  -9, -23,  -9,  -4,   2,   0,  30, -11,   0, -28,  -7,  -3,   2,   3,  28,   3,  -2, -20, -28,   0,   5,   1,  -6,  48, -15,  13, -37,   4,   4,  -4, -10,  36,  15,  -2,   0,  -7,   2,  -7,  11, -13,  40, -17,  26,  11, -13,   2 },
     {    5,  18,  14, -16,   7,  -3,   0,  -1,  -1, -27, -18,  22,  -7,   4,   0,   1, -23,   1,  -9, -28,  13,  -2,   1,   1,  47,  16,  18,  15, -21,   4,  -1,  -1, -35,   1,   1,   7,  32,  -3,   2,  -1,   8, -39, -17, -19, -20,  18,   3,   5,   1,  38,  -8,  -3,  -1,  -6,  -3,  -5,   9,   4,  33,  15,   3, -21,  -8,  -2 },
     {    7,  12,  96, -17,   3,  -8,  -2,  -4,   2,  39,  18,  27,  -9,   0,  -4,  -1,  14,   9, -41,  21,   9,  -1,   3,  -1,   3, -11,  -8, -20,  14,   2,   2,   1, -15,  -1,  -7,  -7,  -8,   7,  -1,   2,  -3,   4,  -1,  -3,  -3,  -4,   1,  -1,   0,  -1,   2,   0,   4,  -3,  -1,   0,   0,  -3,   1,  -3,   0,   2,  -2,   0 },
     {   -3,   9, -13,  18, -14,  -2,  -2,   0,  -5,  11, -22,  24, -16,  -4,  -1,   0,  -2,  11, -21,  21, -14,  -5,   0,   0,  -2,   7, -21,  19, -14,  -7,   0,  -1,  -4,  10, -21,  21, -16, -13,   2,   1,  -6,  17, -27,  25, -17, -23,   8,  -3, -16,  29, -33,  28, -27, -23,  17,  -6, -15,  23, -25,  24, -28, -15,  15,  -7 },
     {   -6, -10, -18, -20,  -7,   5,   3,   2,  10,  17,  26,  24,   9,  -5,  -5,  -2,  -9, -16, -25, -24, -13,   9,   3,   3,  12,  18,  24,  25,   9,  -9,  -3,  -2, -11, -20, -28, -27, -10,  15,   5,   3,  13,  20,  28,  26,   8, -20,  -8,  -3, -15, -25, -31, -27,  -8,  18,  11,   6,  13,  20,  22,  19,  10, -14, -10,  -6 },
     {   -1,  -2, -12,   1,   6,   0,   1,  -2,   1,  -7, -39,  11,   3,   4,   4,   0,   1, -45, -18, -26,  16,   1,   8,   0, -39, -33,  19, -37,  -8,  10,  -2,   6, -58,   5,  -3,  11, -38,   6,  -1,   2,  -7, -12,   2,  15,  -7, -17,   5,  -6,  18,  -9,   2,   5,  12,  -5,  -9,   1,   2,   6,  -3,   7,   1,   6,  -6,  -1 },
     {   -8,  -5,  24, -21,  -5,   1,   4,  -1, -16,  -8,  37, -23,  -7,   1,   3,   1, -32,  14,  25, -10, -25,   8,   0,   1, -34,  37,  -2,  17, -34,  14,   2,   0, -30,  32,  -9,  21, -30,   5,   7,  -3, -22,  20,  -4,  13, -20,   8,   7,  -6,  -2,  -2,  12,   2,  -9,  -1,   2,  -4,  10, -15,  19,  -9,   5,   6,  -6,   2 },
     {   -4,  -6, -37,  13,  -4,  -5,   2,  -2,  -1,  -5,  53, -21,  -1,   5,  -2,   3,   8,  44,  -8,  33,  -6,  -5,   3,  -5,  28, -33, -12, -13,  21,   4,   0,   5, -54,   6,   0, -16, -21,   5,   6,   1,  13, -19,  -6,  11,  -3,  -3,  -1,  -3,  12,  20,   0,  -4,  11,   0, -14,   0,  -2,   2,  14,   6,  -6,   1,   0,  -3 },
     {    4,   0,   5,  82,  -4,   3,   1,   0,   5,   1,  29,  53,  21,   5,   1,  -5,   1,  10,  28, -28,  29,   1,   8,   0,   2,  20,   2, -25,  -2,  17,  -1,   3,  14,   8, -11,  -8, -20,   7,   7,   5,  -2,  -2,  -3,  -6, -14,   3,   4,   4,  -9,  -1,   0,  -4,  -6,  -5,  -1,   1,   0,  -6,   7,  -1,  -1,   0,  -6,   2 },
     {    3,  11,  19,  -1, -15,  -2,   1,   1,  -1, -19, -35,   7,  28,  10,  -2,   3, -11,   2,  27,  -7, -28,  -5,   6,  -3,  30,  21,  -7,   8,  30,   2,  -5,  -1, -40, -32,  -2, -18, -29,  -6,  12,   7,  31,  24,  -8,   5,  18,  14,   2,  -6,  -8,  -3,  21,   7,  -9,   2,  -3,   6, -11, -15, -35, -19,   0,   7,  11,   1 },
     {    1,   1,   0,   1,   2,  -2,  -2,   3,  -1,  -2,   1,   1, -11,  -3,   2,   1,   1,   7,   0,  -3,   7,  -7,  -1,   3,   9,  -1, -10,   1, -13,  -4,  -8,   1,  -3,  -9,  -1, -11,   2, -34,  -1,  18, -19,  -2, -16,   0, -31, -10,  56,  -5,   2, -17,  -2,   0,  11,  89,  -6,   3,  -5,   5, -22,  -1,  11,  -5,   2,   0 },
     {   -1,  -2,  -5,   0,  -3,  -1,   9,  -1,   1,  12,  -4, -14,  -4,  10,   5,   2,   7,  -6, -11,   0,  -9,   2,   9,   1,  -8,  -4,   2, -24,   4,   8,   2,   7,   5,   9, -15,   6, -10,   5,  30,  -3,  -8,  -3,   3,  -4,   5,  77,  -4,  -6,  -8,  -8, -27,  -7, -16,  14, -17,   4,  -5,  15,  -4, -26, -12, -65,  12,  -2 },
     {   -1,  -2,   6, -20,  30, -33,  18,  -3,  -1,  -1,  16, -27,  43, -39,  21,  -6,  -4,   3,  13, -26,  33, -31,  13,  -5,  -9,  10,   2, -19,  24, -22,   7,   4, -10,  12,  -5,  -7,  13, -25,   5,   6, -10,  18,  -7,  -1,   9, -21,   6,   3, -16,  19, -12,   1,   2, -16,   8,  -2,  -8,  11,  -9,   0,  -3, -21,   8,  -1 },
     {   -2,  -2, -11,  -2,   6,  12,  -7,   1,  -2,  -6,  -5,   6,   7,  -1,   7,  -1,   0,   2,   4,  13,   8,  10,  -2,   0,  14, -10,   3,   9,  26,  -2,   8,  -3, -24, -13,  15, -23,  23,  35, -10,   5, -75,  10,  -8,   3, -35,  16,   7,   1,  -7, -32,   1,   3, -14, -27,  27,  -1,   4, -18, -15,   4,   5, -11,  -3,  11 },
     {    8,   9,  31,  67,  29,   4,   4,   2,  -3,  -4, -22, -64,  -9,  -6,   0,   2,  -8, -12, -13,   8, -22,  -6,  -5,   2,  10,  13,  27,  20,  12,  -4,  -9,  -3,  -1,  -5, -17, -17,  -5,   2,  14,  -5, -14,  -7,   6,   2,  -9, -19,  -8,   1,  12,  -4,  -8,  -1,  12,   1,  -2,  -4,   3,  14,  14,   1,  -3,  -3,  -2,   1 },
     {    1,  -1,   9,  20, -27,   7, -19, -30,   0,   3,   6, -36,  14,  -4,  -4,  -6,   1,   9, -34, -46,  29,  18, -10,   6,   3, -20, -33,  26, -33,   5,  34,   1,  -9, -24,  15,   2,  -4,  14,   4,   4,  -6,   6,   8,   0,  25,  -2,  -2,   6, -15,  12,   2,   8,   1,   7,   3,  12,  -4, -15,  -8,   1,  -2, -11,  12,   4 },
     {   -3,  -5,   6,   0, -82,  -6,   0,  -2,  -4, -13,  -3, -14, -56, -30,   1,  -8,  -5,  -4,  11,  -9,  29, -42,  -3,   2,   1,  13,  -4,  -9,  24,  -2, -16,   5,  -1,  -6, -14,   0,   6,  14,   5,   6,  -5,  -5,   3,   3,   0,   0,   1,  -1,   6, -10,  -3,  -7,   6, -11,  -4,  -4,   0,   8,   0,   0,   6,   4,  -3,  -4 },
     {    2,   5,  -4, -34,  21,  29,  -5,   0,   0,  -1,  -8, -22,  -4,  11,   2,  -6,   2,  13,  13,   2,  21,  16,  -3,  -5,  12,  25, -15, -14,  -2,  33,  15,  -3,  21,  -4, -31,  -6, -24,  37,  11,   5,  19, -23, -15, -11, -16, -31,  20,  13,  -1, -16, -11, -11,  18, -24, -10,  19, -16,  12, -22,   1,   6, -14,  -1,  -1 },
   },
   {
     {  107, -44,  -5,  -5,  -1,  -1,  -1,   0,  28,   5, -12,   0,  -2,   0,  -1,   0, -26,  28,  -5,  -1,  -1,   0,   0,   0, -17,   9,   6,  -2,   1,  -1,   0,   0,  -7,   0,   5,   0,   0,   0,   0,   0,  -3,  -1,   2,   1,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {  -31, -50,  36,  -1,   6,   1,   2,   0,  63, -67,   7,   6,  -1,   2,  -1,   1,  26,   6, -29,  11,  -4,   2,  -2,   1,  -8,  26, -17,  -1,   1,  -1,   0,   0,  -6,  13,  -1,  -5,   1,  -1,   0,   0,  -4,   5,   1,  -2,   0,   0,   0,   0,  -1,   2,   0,  -1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {    9, -64,  43,  -7,   7,  -1,   2,   0, -84, -14,  27,   8,   3,   3,   1,   1, -14, -14,  -8,  11,  -1,   2,   0,   1,  28, -12, -15,   4,   0,   0,   0,   0,  13,   3, -12,   0,   0,   0,   0,   0,   5,   6,  -6,  -2,   0,   0,   0,   0,   1,   3,  -1,  -2,   1,   0,   0,   0,   1,   1,   0,  -1,   0,   0,   0,   0 },
     {  -49, -20,  25,  -2,   5,   0,   2,   0,   4,  23, -21,   5,  -4,   1,  -1,   0, -66,  69, -13,  -4,   2,  -2,   1,  -1, -14,  -9,  26, -14,   5,  -2,   1,  -1,  17, -25,  14,   0,  -3,   2,  -1,   1,   8,  -9,  -1,   5,  -2,   1,   0,   0,   3,  -1,  -2,   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {    1,   6,  27, -16,   2,  -3,   0,  -1,  44, -13,  38, -19,  -1,  -3,  -1,  -1, -42, -24,  52,   6,  -4,   3,  -1,   1,  -5, -51,   7,  31,  -7,   6,  -1,   2,   9, -12, -32,  23,   0,   2,   1,   1,  -1,   9, -22,   1,   6,   0,   1,   0,  -1,   6,  -6,  -5,   3,   0,   0,   0,   0,   1,  -1,  -3,   1,   0,   0,   0 },
     {  -13, -32, -60,  58,  -9,  11,  -3,   3,  -5, -45, -33,  28,   8,   3,   2,   1, -34, -25,  23, -21,  17,  -2,   4,   0,  -5, -13,  10, -10,  -1,   3,   0,   1,   9,   5, -11,   3,  -5,   2,   0,   0,   3,  11,  -8,  -3,   1,  -1,   0,  -1,   1,   6,   1,  -6,   2,  -1,   0,   0,   1,   1,   1,  -3,   1,   0,   0,   0 },
     {    7,  39, -11,   0,  -5,   0,  -2,   0, -26, -44,  42,  -1,   3,   0,   2,   0, -47,  -4,  -7,  22,  -5,   3,   0,   1, -60,  44, -22,   7,   7,  -5,   3,  -1,   0, -14,  13, -15,   9,  -2,   1,   0,  17, -16,   5,  -3,  -3,   3,  -1,   1,   6,  -4,  -3,   2,  -3,   1,   0,   0,   3,   0,  -1,   0,   0,  -1,   0,   0 },
     {   -4,  13,  -3,   4,  -4,   0,   0,   0,  18,  27, -26,  18, -11,   2,  -2,   0, -45,  -6, -24,  37,  -1,   2,   1,   1,  24,  -9, -61,  18,  11,  -3,   2,  -1,  -1,  41, -31, -25,  18,  -5,   1,  -1,  -5,  11,  18, -29,   5,   1,  -1,   0,   0,  -1,  14,  -6,  -4,   2,  -1,   0,  -1,  -1,   4,   0,  -1,   0,   0,   0 },
     {  -20, -30, -18, -20,  31,  -4,   4,   1,   1,  14,   2, -55,  28,  -3,   2,   0, -22,  -3,  40, -36,  -7,  10,  -3,   1,   3,  35, -16,   4, -26,   6,  -2,   0, -11,  41,   6,  -7,  -7,  -8,   2,  -2,   7, -12,  29,  -3,  -3,  -4,  -1,   0,   5, -10,   9,  10,  -7,   1,  -2,   1,   1,  -2,  -2,   7,  -3,   0,   0,  -1 },
     {   -2,  -6,  29,  46, -63,  19,  -8,   2,   7,  26,  43,   5, -30,  -5,   3,  -3,   4,  17,  31, -35,  25, -20,   6,  -3,   5,  27, -18,  -7,  12,  -2,  -4,   1,  -5,  15,  -9,   0,   4,   0,  -2,   0,   2, -10,   6,  -1,   1,   0,  -1,   1,   1,  -5,   2,   4,  -2,   1,  -1,   1,   0,  -2,  -1,   4,  -2,   0,   0,   0 },
     {   17,  32,  -4,  -7,  -5,  -2,  -1,   0,  16, -39,  23, -13,   6,  -4,   2,  -1, -24,   6, -13, -16,  11,  -3,   1,   0,  72,  -9,  -2, -35,   2,   0,  -1,   1,  45,   0,  26, -21, -16,   5,  -4,   1,  -1,   4,  19,  13, -18,   2,  -1,   0,  -8,   1,   1,  16,  -6,  -2,   1,  -1,  -4,   1,  -3,   7,   0,  -1,   1,   0 },
     {    3,  11, -31,   8,  -1,   3,   0,   0, -14, -17,  14, -18,   7,   2,   1,   1,  19,  68,  -9, -12, -20,   3,  -4,   1,  11,  17, -11,  32, -20,   0,  -1,  -1,  38,  -6, -46,  26,   6,  -9,   2,  -1,  -7,  29, -15, -15,  19,  -7,   1,  -1,  -7,   5,  12, -15,   5,   2,  -2,   0,  -2,  -1,   6,  -1,  -2,   2,   0,   0 },
     {   -9, -30, -29,  28,   0,   2,   1,   2,  21,  37,  17, -18,  -2,  -9,   2,  -3,  -7, -25,  -8,  34, -16,   6,  -4,   2,  29,  30, -25,  13,  -8,  -7,  -1,  -2,  20, -43,  39,  12,   6,  -2,  -1,   1,   6, -12, -30,  19,   6,   3,  -2,   2, -15,  16, -13,  -9,  11,  -1,   2,  -1,  -1,   0,   1,  -8,   3,   1,   0,   0 },
     {   -8, -12, -18, -47, -51,  42,  14,   5,  -8, -15, -26, -49, -41,  29,  21,   5,  -6, -16, -22, -12,  -3, -14,  14,   5,  -6, -12,  -5,   8,   8, -21,  -1,   4,  -2,  -7,   6,  13,   3, -10,  -9,   1,   1,   2,   3,   6,   3,  -1,  -7,  -2,   2,   5,   1,   1,   2,   1,  -3,  -2,   1,   1,   1,   1,   0,  -1,  -1,   0 },
     {    0,   7,   0,  -1,  -2,  -2,   2,  -2,  -6, -23,  -6,  34, -34,   9,  -4,   2,  -8,  13,  35,  25, -46,  -5,  -1,  -2,  37,  16,  37,  23, -15, -27,   3,  -4, -13,  28,  23,  14,  22, -23,  -2,  -1,  -3, -22,  16,   9,  22,   1,  -7,   3,   2, -12,  -9,   9,   7,   8,  -3,   0,  -2,  -2,  -6,  -1,   3,   4,   1,   0 },
     {    3,  26,  20,  -7,  -3,  -4,   1,  -1,  -2,  -2, -10,   9,   7,  -1,   2,   1, -34, -15, -31, -16,  19,   6,   3,   1,  17,  35,  -1, -19, -26,  12,  -5,   1, -35,  23,  -5,  55, -28,   0,   3,  -1, -10,   4, -32,  22,  24, -15,   5,  -2,  -2,   2,  -5, -20,  23,  -2,  -1,   1,   2,  -2,   5, -13,   3,   3,  -2,   1 },
     {   -7,  -7, -27,  27,  -6,   3,   0,   2,  -6,  -9,  40, -27,  16, -10,   3,   1,  -8,  25, -29,  13,   3,   1,   0,   0,  13, -43,   6,  13,   4,   6,  -2,  -1, -71,   0,  19,   2,   6,   6,   1,  -2, -44,  -3,  11,   0,   3,   1,   5,  -1,   8, -17,   1,   3,   0,   1,   1,   2,   7,  -5,  -7,   1,   2,   0,   1,   2 },
     {  -10, -13, -38, -24,  28,  -8,   9,   1,   2,  12,  38,   3, -36,  15,  -5,   3,  -3,   5,  13,  21, -11, -15,  11,  -5,   0,  17,  13, -50,  44, -14,   1,   1,  -6,  22, -23,   2, -11,  28, -12,   3,  -9,  22, -26,   5,  -7,   3,   8,  -7,   4,  -3,  13, -11,   3,  -8,   3,   0,  -1,  -2,   4,  -1,   2,   0,  -2,   1 },
     {   -3,  -1,  -4,  13, -16,   9,  -2,  -5,  -3,   9,   2,  11, -23,  13,  -5,   0,  -2,   4,  -6,  23, -30,  16,  -3,   1, -16,  -2,  -6,  14, -37,  22,  -1,   1,   2,  14,  -7,  11, -51,  13,   4,  -3,  -7,  18,   6,  31, -48, -10,  14,  -7,  -3,   9,  -4,  42, -22, -19,   8,  -3,  -7,   8, -14,  24,   0,  -9,   3,   0 },
     {    0,   2,  -3,   0,   4,  -4,   1,  -3,  -6,  -3,  -7,  -4,   4,   2,   0,   0,   8,  21,  11,  -6,   4,   1,   0,   2, -14, -17, -27,  14,  17,  -3,   5,   0,   8,  28,  -1, -21,  -2,   4,  -7,   0,  -2,   0,   5,  55,  -9,   5,  -4,  -3, -10,   5, -45,   8,  60,  -9,   4,  -2,  -1,  -7,   4, -46,  31,  11,  -4,   2 },
     {   -2,  -3,  -6,   4, -16, -62,  71, -18,  -2,  -4,  -1,   2, -33, -24,  29,  11,  -2,  -3,  -4, -19,  -8,  27, -35,  27,  -3,  -5,  -5,  -7,   8,  14, -20,   0,  -1,  -5,  -4,   4,  10,  -2,  -1, -12,   1,  -4,   7,   1,  -1,   3,  -1,  -4,   0,   4,   3,   2,   0,   2,  -1,   0,  -1,   3,  -1,   1,   1,  -2,   1,   1 },
     {   -5,  -6, -12, -32, -19,  15,  10,   8,  -3,  -4,  -8,   8,  18, -32,  13,   4,   2,   9,  15,  40,  43, -40, -25,  10,   8,  21,  22,  34,   5,  27, -40,  -5,  11,  15,  10,  -1,  -6,  30,   3, -17,   1,   3,  -1,  -5,  -3,   1,  17,  -4,   1,  -1,  -2,  -2,   1,  -5,   2,   4,  -2,   0,   0,  -2,  -3,   1,  -1,   1 },
     {    3,  -3, -19,   0,   7,   1,  -2,  -3,   7,  -6,   1,  -2,  -2,   5,  -2,  -6,  10,  12, -16,   4,   6,   2,   2,  -4,  20, -15, -11,   2,  12,  10,   1,  -1,  -6,   2, -34,  23,  -7,  13,   4,   0,  64, -70,  -5,   4,  -2,  -3,  10,   2,  39, -17, -22,   3, -10,  -2,   0,   5,  -4,  17, -14,   0,  -1,  -5,   2,   1 },
     {   -2, -12, -12,   1,   4,   4,  -9,   5,   7,  15,  24,   5, -13,  -9,   5, -10,   6, -18, -22,  15,  11,   6,  -2,   3, -28, -15,  11, -11,  -3,  12,  -1,   5,  51,  14,   4,  21, -23, -21,   4,  -7, -13,  -3,  46,   5,  40, -25,  -6,   2,  10, -43,   7,  -2,  19,  20,  -6,   5,   5,  -5, -12,  -1,   4,   9,   2,   1 },
     {    2,   2,   4,  21, -23,   3,   6,  -4,   4,  11,  14, -29,  13,  -9,   2,   6,   1,  -3, -23,  18, -11,   5,  -7,   0,  -8, -11,  42, -15, -22,  -4,   3,  -7,   4,  49,   9, -31,  12, -17,   0,  -2,  54,  34, -32,  -7,   7,  -5,  -7,   0,  25,  12,  -8,  -1,   5,  -1,  -7,   0,  -7,   7,   5,  -5,   0,   2,   0,  -1 },
     {   -4,  -6,  -8,  -5,   8,   2,   1,  -7,   8,  13,  17,  11,   2,  -2,  -4,  -3, -11, -20, -27, -14,  13,  -8,   4,   4,  13,  25,  33,  13,  -9, -14,  -2,   2, -15, -28, -36,  10,  21,   1,   2,   4,  19,  23,  30, -22, -23,   5, -11,  -1, -10,  -3,  -2,  54,  23,  -1,   1,   0,  -2,  -2, -14, -26,  31,   4,  -2,   3 },
     {    0,  -1,   0,  -7, -10,  28, -15,   4,   1,  -1,   4,  -4, -20,  46, -24,   1,  -3,  -3,   3,  -6,  -2,  59, -41,  -1,   9,   8,  16, -10,  -1,  39, -14, -20,   1,  -7,  -1, -21,  17,  15,  27, -15,  -8,  -9,   0, -17,   2,   5,  21,   5,  -3,  -9,  -1, -14,  13,  -5,   9,   6,   6,  -9,  11, -21,   9,   2,  -2,   3 },
     {    0,  -1, -13,  -7,  32, -10,   4,   7,   1,  -1,  21,  16, -48,  15,  -6,  17,   5,  10,  -1, -14,  26, -32,  14, -10,  -4, -20, -17,   9, -40,  16, -11,  10, -10,  -3,  31, -14, -15, -20,  14,  -1,  17,  -2,   1, -33,   2, -22,  10,  10,  -1,  34,  -4,  -9,   3, -12, -12,   3,  -5,  -3,  12, -16,  11,   5,   1,  -3 },
     {    6,   9,   9,  16,  23,  45,  50, -36,   6,   8,  10,  13,  27,  53,  46, -20,   6,  10,  13,  16,  18,  14,   2,  19,   4,   3,   2,  -1,   0, -11, -12,  26,   0,   1,   5,   1, -10, -12,  -8,  10,  -1,   0,   2,  -9,  -5,  -4,  -7,   3,   4,  -1,  -3,  -7,  -1,  -6,  -2,   0,   2,  -1,  -4,   0,   1,   1,   0,   1 },
     {   -3,  -1, -12, -12,  15,  14,  -9,  14,  -1,   6,  12,  11,  -9,  -2,   2,  12,   0,   5,  -8,  -4,  25,   6, -18,  11, -10,  -5,  -5, -18, -20,   0,  -6,   8,   6,  19,  -7,   9,  21, -19, -12,  -6, -18,  -2, -17,   3,   9,  49, -18,   0, -18,   7, -50,  15, -41,  33,  18, -17,   0,  17, -13,  -2,  -2, -14,  24,  -3 },
     {   -2,  -5,  -6, -19, -19,  10,  -8,  14,   1,  -1,  -2,  12,   2, -29,  36,  -1,   2,   4,  12,  13, -29,  -3,  28,  24,   7,   4,  -1, -34, -38,  32,   8,  29,  -9, -22, -30, -42,   0,  -2,  11,   7, -16, -19, -15,  -4,   0,  -6, -18,   7,   4, -16,  -7,   0,  12,  -4, -10,  -2,  15,  -3,   2, -17,  -6,   4,  -7,   3 },
     {   -5,  -7, -16, -25, -23,   5,  29, -17,  -2,  -2,   3,   9,  12,  -9, -26,  -3,   3,  10,  14,  20,  31,   6, -29, -28,   3,  -1,  -7, -15,   1,  -1,  45, -27, -11,  -5, -11,   0, -11, -33,  28,  25,  -3,  -5, -11,   7, -10, -21, -21,  29, -20,   3,   2,   9,  -1,  11, -18,  -1, -11,  -6,  -7,   6,  -4,  15,   1,  -7 },
   }
 },
 {  //25
   {
     { -115,  22,  13,   3,   4,   1,   2,   1,  29, -33,   1,   0,   0,   0,   0,   0,  22,  -1,  -8,   0,  -1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   3,   2,  -2,  -2,   0,   0,   0,   0,   1,   1,   1,  -1,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },
     {    2, -33,   9,  -1,   2,   0,   1,   0,  86,  16, -32,   0,  -7,  -1,  -3,   0, -34,  58,  16,  -9,   1,  -3,   0,  -2, -13, -24,  25,   8,   0,   1,   0,   1,  -2,  -4, -10,   7,   2,   1,   0,   0,  -3,  -1,   1,  -3,   1,   0,   0,   0,  -1,   1,   0,   1,   0,   0,   0,   0,  -1,  -1,   0,   0,   1,   0,   0,   0 },
     {  -26,  -9,   7,   4,   1,   2,   0,   1, -10,  50,  -4,  -4,   1,  -2,   0,  -1, -32, -42,  53,   5,   4,   2,   1,   1,  20, -48, -39,  28,   1,   6,   1,   3,  -3,  24, -29, -20,   8,  -1,   2,  -1,   5,  -2,  11, -11,  -4,   1,   0,   0,   1,   1,   0,   0,  -2,   1,   0,   0,   1,   0,   0,   0,  -3,   2,   1,   0 },
     {    2,  50,   5,  -9,  -1,  -2,  -1,  -2, -43, -17,   1,  13,   1,   2,   1,   1,   9,  36,  16, -24,   1,  -4,   0,  -2,  -5, -45,  48,  31,  -4,   3,  -2,   2,   1, -18, -50,  16,  12,   4,   3,   2,   0,   9,   1, -20,   0,  -1,   1,   0,   3,  -3,   0,   3,  -2,   1,  -1,   0,   0,   1,   0,  -2,  -1,   1,   0,   0 },
     {  -29,  22, -12,   4,  -1,   0,   0,   0, -17,  60,   9, -18,  -4,  -3,  -2,  -1, -27,  -1,  17,  -3,   2,  -3,   1,  -1,  17,  -4,  41, -30,  -7,  -2,  -2,  -1,  10, -28,  40,  42, -20,   1,  -6,   1,  -4,  -7, -31,  26,  11,   0,   2,   1,  -1,   1,  -3,  -4,   2,  -4,   1,   0,   1,  -1,  -1,   2,   7,  -5,  -1,   0 },
     {   -1,  20,   5,  -9,  -4,  -1,  -2,   0, -12,   7, -18,   3,   0,  -3,   1,  -1,   5,  22,  16, -22,  -7,  -1,  -2,  -1,  -9,  -3,  -1,   4,   3,  -4,   2,  -2,  13,   1,  28, -48, -15,   3,  -4,   1,  11,  -1,  37,  52, -47,  -9,  -7,  -1,  -6,   7, -19,  36,  35, -18,   1,  -4,   1,  -5,   2, -12,  15,   7,  -2,   2 },
     {   28,  94,  -4,  -6,  -6,  -4,  -3,  -2,  57,  -7,  28,   1,  -1,  -1,  -1,  -1, -32, -24,  -2,  18,   4,   3,   1,   1, -12,  -4, -16,  -3,   5,   1,   1,   1,  -2,   2,   3,  -4,   0,  -1,   0,   0,  -2,  -3,   3,   0,   1,   0,   0,   0,  -1,  -3,  -1,   1,   0,   0,   0,   0,  -1,  -1,   0,  -1,   1,   0,   0,   0 },
     {   -4,   1, -11,  -1,   2,  -3,   2,   0,  -2,  11,   2, -11,  -8,   3,  -2,   1,  -7,   1, -13,  -4,   5,  -4,   2,  -1,   3,   6,   7, -21,  -9,   3,  -2,   1,  -6,  -3,   0,  -8,   7,  -5,   2,  -2,  10, -13,  19, -45, -33,  17,  -7,   5,   3,  14,  10,  48, -59, -30,   2,  -3,   1,   6,   1,  19,  42, -38,  -7,  -2 },
     {    2,   3, -15,   9,  -5,   2,  -2,   1,  16,  39,  13, -20,   4,  -7,   0,  -3,  61,  -1,  34,  14, -14,   3,  -6,   0, -54,  19,  -8,  30,  17,  -2,   4,   0, -11, -46,  -2, -13,  16,   9,   3,   2,   2,  -1, -24,  -8, -14,   3,   0,   1,   1,   5,   7,  -7,  -4,  -8,  -1,   0,   1,   0,   2,   6,   0,  -1,  -2,  -1 },
     {   10,  19,  24,  18,  -2,  -5,  -2,  -3, -18, -25, -38, -26,   4,   7,   3,   3,  19,  31,  35,  20,  -6,  -9,  -3,  -3, -20, -28, -32, -17,  13,  11,   5,   3,  16,  26,  27,   3, -22, -15,  -4,  -3, -11, -20, -11,   7,  28,  11,   1,   2,   4,   6,   0, -18, -31,  -4,   2,   0,   1,   2,   5,  16,  13, -12,  -4,  -2 },
     {  -18,  28, -30,   9,   3,   1,   0,   1, -27,  42, -40,  12,   5,  -1,   1,   0, -27,  39, -31,   1,  11,  -2,   3,   0, -22,  30, -16,  -8,  16,  -6,   4,  -2, -10,   8,   5, -27,  24,  -5,   4,  -2,   5,  -9,  19, -29,  17,   0,   1,   0,   7, -13,  17, -21,   4,  15,  -4,   3,   1,  -5,   3,  -4, -16,  15,   1,   0 },
     {   -8, -17,   8,   6,   1,   3,   0,   1, -19, -23,   9,   5,   5,   3,   2,   1, -51,   1, -23,  24,   1,   5,   2,   2, -29, -29, -15, -20,  27,  -2,   7,   0,  28, -49, -10, -29, -14,  16,  -3,   3,   0,  26, -34,  -1, -28,  -5,   1,   0,   3,  10,   9, -14,  -1, -15,   1,  -1,   4,   2,   6,   3,  -4,  -1,  -4,   0 },
     {   -2,   2,  -2, -12, -13,   5,  -1,   1,   2,   5,   2,   5,  10,  -5,  -1,  -1,   0,  -1,   0, -19, -16,   8,   1,   2,   2,   1,   8,   4,   8,  -8,  -2,  -1,   0,  -6,   2, -24, -22,  14,   1,   2,   5,   0,  16,  15,   2, -12,  -5,  -2,  -7, -14, -15, -18, -32,  35,  -3,   6,  -4,   8,  -7,  15, -47, -82,  23,   3 },
     {    6,  16,  43,  -5,  -9,   1,  -1,  -1,  -1, -17, -50,  12,  10,  -1,   4,   2, -21, -17,  15, -29, -12,   4,   0,   0,  36,  36,  -1,  23,  -5, -13,  -2,  -1, -26, -22,  17,  -4,  22,  15,   2,   0,   0, -15, -38,  -6, -21,   2,   4,   3,   3,  16,  10, -19,  -6, -16,   0,   0,   4,   1,   9,  16,  -3,   1,  -5,  -4 },
     {    5,   5,  80,   1,  -4,  -4,  -4,  -1,   2,  44,   4,  33,   0,   0,  -3,  -1,  31,  -3, -39,  -1,  26,   5,   5,  -1, -15, -22,  -4, -27,  -2,  16,   3,   4, -32,  -6,  -6,   0, -17,  -3,   6,   0,   7, -11,  -1,  -2,   1, -10,  -2,   1,   3,   5,  -5,   0,   2,   1,  -4,   0,   0,   1,   2,  -3,  -1,   2,   3,  -1 },
     {    4,  -4,  10, -11,   9,   5,   1,  -2,   4,  -9,  12, -21,   7,  11,   1,   0,   4,  -8,  15, -16,  12,   7,   0,  -2,   5, -12,  19, -22,   8,  15,   1,  -1,  12, -18,  25, -21,   8,  17,  -3,   2,  10, -20,  18, -14, -12,  38,  -5,   4,   2,  -6,   0,   9, -27,  42,  -1,   0,  -7,   8, -14,  17, -40,  61,   2,  -8 },
     {   -5,  -8,   6,   1,  11,  -1,   2,   0,  -9,  -9,  11,  20,  -6,   9,   0,   1, -17,   2,  -1,  12,  16,  -4,   4,  -1, -17,   3,  -3,  19,  -2,   8,   1,   1,   3, -16,  14,   3,  40, -11,   8,  -3,   8, -40,  -8,   9,   9,  45,  -4,  -5, -23,  10, -61,  -3,  -9,   2,   3,  -2,   3, -12,   6, -49,   1, -21,   2,  12 },
     {    3,   6,  21,  36,   6,  -8,  -4,  -2,  -4,  -9, -24, -35,  -4,  16,   7,   4, -10,  -2,  11,  34, -11, -15,  -6,  -5,  26,  25,   8,  -3,  24,  20,   3,  -4, -27, -36, -19,   0, -17,  -7,   9,   5,  26,  24,  17,  -5,  13,  -6, -10,  -7, -13,  -4,  -2,  30,   9,  22,  -4,  -2,  -5, -10, -20, -27,   4,  -4,  18,   9 },
     {   -4,  -3, -37, -29,   4,   1,   1,   0,  -3, -15,  26,  15, -19,  -1,  -3,   0,   7,  40,  18,  22,  19,  -4,   0,  -2,  43, -12, -19,  -9,  13,  16,   4,   4, -63, -12,   5, -15, -17,   4,   8,   2,   8, -26, -15,   3,  -7, -17,  -6,   1,   7,  18,  -2, -10,   4,   0,  -5,  -5,   2,   2,  11,   8,  -4,   4,   1,  -2 },
     {    3,   5,   4,   8,  14,  10,  -4,   1,  -5,  -9, -10, -12, -22, -16,   5,   0,   6,   9,  12,  11,  23,  17,  -5,   1,  -8, -10, -15, -17, -29, -14,  10,   3,   8,  10,  13,  21,  23,  14, -13,  -4,  -9,  -9, -17, -19, -31, -14,  19,   1,  11,  17,  21,  28,  25,  13, -33, -11, -15, -13, -24, -21, -35, -25,  29,   9 },
     {  -10, -14,  44, -17,  -5,  -2,  -2,   0, -21,  -8,  52,  -6, -10,  -7,  -3,   2, -42,  23,  20,  14, -23,   4,  -4,   2, -33,  31,  -2,  24, -10, -12,   1,   1, -10,   4,  10,  11,   2, -13,   4,   0,  10, -25,  19,  -1,  14, -22,   3,   7,  14, -17,  13,  13,   5,   0,   1,   3,  -1,   6, -12,  25,   3,   5,   7,  -6 },
     {    3,   6,  32,  -3,  -8,  -5,  -6,   0,   2,  24,  32,   8,  -9,  -6,  -6,  -2,  16,  43,  -6,  31,  -4,  -5,  -5,  -1,  58,  10,  -9,  13,  24, -11,   3,  -7,  38,   9,   8, -10,  28,  12, -10,   4, -25,  30,  -1,  -1,  -6,  28,  -2,  -3,  -9,  -6,  12,   0,  -5,  -2,  12,   0,  -2, -11,   1,   3,  -2,  -7,   2,   3 },
     {   -1,   1,   7, -75,  36,  -6,  11,  -6,  -4,   6, -34,   4, -27,  11,  -1,   4,   7, -25,   0,  43, -17, -16,   2,   0, -14,   0,  24,  -8,  35, -18, -10,  -2,  -1,  16,  -6,   8,   1,  24,  -7,   0,   0,  -6,   1,  -1,  -3,   1,  14,  -3,   0,  -9,  -4,   3,   2,  -4,   5,   4,  -2,   5,  -6,  -3,   1,  -3,  -4,   1 },
     {    5,  -8,  14, -12,  18,  -9,  -2,  -3,   7, -12,  17, -24,  21, -14,  -3,   0,   7, -11,  16, -21,  20, -11,  -2,   1,  11, -12,  17, -21,  19, -14,  -4,   2,   7, -15,  15, -24,  19, -10,  -9,   1,  10, -14,  19, -22,  20, -12, -15,   2,  15, -23,  26, -29,  25, -14, -30,  12,  14, -21,  20, -25,  18, -24, -16,  20 },
     {    1,   2,   0,   0,   7, -13,   0,   1,   0,   1,   1, -15, -24,  13,   1,   1,   2,  -2,   1,  -3,   1, -23,  -1,   1,   0,   3,   5,  -7, -19,  16,  -2,  -1,  -4,   2,  -4,  -6,  -5, -32,   4,   2,  -8,   3,   2,  -2, -20,  12, -14,  -2,  -2, -25,   0, -27,  -2, -68,  13,  29,  -2,  -2,  -5,  -2, -43,   9,  55,   2 },
     {   -1,   0,   0, -39,  10,  -5,  -7,   3,   1,  -2, -15,  42, -25,  -4,  -1,  -4,   1,  -6,  34, -11,  35,  -8,  -6,   2, -11,  26, -19,  -6, -21,  24,   3,  -2,  33, -25,   1,   3,  -8, -34,   9,  -3,   7,  35,  -3,   0,  16,   4, -32,  -1,  -7,   7,  24,   7,  -2,  22,   5, -11,   1,  -5,   1,  12,   5,  -7,   0,   5 },
     {    3,   6,   1, -50,  18,  10,  -5,   3,   4,  14,  -8, -48,  17,  11,   2,   2,   7,  14, -40,  -5, -28,  22,  -1,  -2,  22, -14, -35,  18, -26,   3,  11,  -2,  28, -29,  -7,   8,   0, -24,  14, -11,  19, -11,  -1,  -2,  11, -15,  -2,   7,  -3,  16,  -9,  -3,   0,  -1,  -9,  -2,  -5,   3,   4,  -7,   2,   6,   9,  -7 },
     {   -4,  -6,  -2, -28, -91, -16,   6,   8,  -4,  -6, -19, -28, -15, -12,  -9,   7,  -1, -12, -12,  27,  31,  18, -10,  -7,  -3,  -3,  22,  19,  -4,  23,  20,   0,   4,   4,   9,  -1,   2,  -1,  -3,  11,   5,  -3,   9,  -5,  -2,   0,  -3,  -1,   8,   2,   3,  -8,   3,  -1,   3,   2,   0,  -8,   1,  -9,  -3,  -1, -12,   4 },
     {   -6,   3,  12, -30,  12,  -1,  -3,  -1,  -9,   6,  17, -43,  23,  11,  -5,   0, -16,  16,   7, -22,   8,  15,  -9,   1, -21,  29,  -4,  -9, -10,  27,  -9,   1, -32,  36, -19,  -2, -16,  21,  -9,   0, -26,  34, -17,   4, -15,  25,  -9, -10, -16,  12,   1, -11,  -6,   8,  -9,  -6,   6, -11,  14, -24,   3,  -4,  -9,   9 },
     {   -6,  -7, -10, -19, -39, -38, -22,  -7,   2,   1,   4,  24,  51,  27,  -7, -13,   0,   0,   4,   0, -23, -31,  14,  22,  -6,  -9,  -4,  -7,  13, -10, -11,   6,   1,   9,  10,  10,   7,  -2,  -2,  -3,   4,  -2, -13,  -5,   2,   1, -20, -10,  -2,  18,  13,  15,  -3,   0, -34,  -9,  -9,  -2, -15, -15,   5,  10,  40,  -7 },
     {    0,   1,   1,  -2,  -1,   2,  -1,   4,  -3,  -4,  -1, -15, -10, -22,  -4,  -2,  -1,  -2, -14,  -1, -17, -11,  -3,   7, -11, -31,  -4,  -4, -12, -37,  -7,  -5, -41,   0,  14,   3,  24, -20, -42,  -8,  35,  29,   0,  13,   8,  28, -30,  -4,   8,  24,  14,   0,   5,  14,  45,   0,  -6,  -1,   5,   6,  -7,  -4,   0,  10 },
     {   -2,  -3,  -6, -17,   2,  11, -24,   5,  -1,  -7, -12,  14,  54, -16,  -7, -16,   1,   4,  15,  14,  20,  36,  -8,   7,   5,   0,  -9, -16,   3,   2,  16,  22, -18, -20,  -8,  15,   8, -10, -20, -12,   3,   6,  17,  15,   4,  19,  39,  -3, -13, -38,   6,   7,  -2, -28,  23,  -2,  -5,   3, -29,   2,   1,  -5,  -1,  -7 },
   },
   {
     {  118, -36,  -8,  -4,  -2,  -1,  -1,   0, -13,  16,  -3,   1,  -1,   0,   0,   0, -21,   7,   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0,  -6,   4,   1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    7, -12,   3,   0,   1,   0,   0,   0, 112, -23, -15,  -2,  -4,  -1,  -2,   0,  -6,  36,  -7,  -1,  -2,   0,  -1,   0, -35,  12,   8,  -1,   1,   0,   1,   0,  -7,  -3,   3,   1,   0,   0,   0,   0,  -5,   0,   0,   0,   0,   0,   0,   0,  -1,   2,  -1,  -1,   0,   0,   0,   0,  -2,   1,   1,  -1,   0,   0,   0,   0 },
     {  -18,   4,   4,   1,   0,   0,   0,   0,   6,  22,  -7,   0,  -2,  -1,  -1,   0,-103,  18,  22,   0,   5,   0,   2,   0,  12, -51,   9,   3,   2,   1,   1,   0,  31,  -7, -13,   2,  -2,   0,  -1,   0,   2,   4,  -2,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   1,   0,   0,   0,   0 },
     {    2, -14,   3,   1,   1,   0,   0,   0,  37, -13,  -4,  -1,  -1,   0,   0,   0,  -1, -37,   6,   2,   1,   1,   1,   0,  81,  -9, -34,   2,  -5,   1,  -2,   0,  -7,  63, -12,  -7,  -2,  -2,   0,  -1, -24,  11,  17,  -4,   1,  -1,   1,   0,  -5,  -4,   5,   2,   0,   0,   0,   0,  -2,  -1,  -1,   0,   1,   0,   0,   0 },
     {   30,  97, -41,  -2, -10,  -1,  -3,  -1,  20,  -5,  18,  -8,   2,  -2,   0,   0, -13, -46,  19,   5,   3,   2,   1,   1, -22,   3, -11,   6,   0,   1,   0,   0,   0,   4,  -3,  -4,   1,  -1,   0,   0,   3,   1,   3,  -2,  -1,   0,   0,   0,  -1,  -1,   2,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   14,  -4,  -1,  -1,   0,   0,   0,   0,  -4, -39,   8,   3,   2,   1,   1,   1,  35,   3,  -8,  -2,  -2,   0,  -1,   0, -14, -40,   8,   7,   1,   3,   0,   0,  64,  -5, -48,   3,  -5,   1,  -1,   0, -11,  62,  -6, -15,   0,  -3,   0,  -1, -11,   3,  18,  -3,   0,   0,   0,   0,  -2,  -1,   2,   2,   0,   0,   0,   0 },
     {   -2, -31,  13,   1,   3,   0,   1,   0,  -5, -73,  31,   3,   5,   2,   2,   1, -42, -44, -12,  16,   1,   5,   1,   1,  -5,  31, -26,  -4,  -1,  -1,   0,  -1,  14, -20,  12,  -2,   0,   0,   0,   0,  33,  -6, -23,   6,  -4,   1,  -1,   0,  -2,  23,  -7,  -6,   0,  -1,   0,   0,   0,   2,   6,  -1,  -1,   0,   0,   0 },
     {   -7, -38,  10,   2,   3,   1,   1,   0,  25,  54, -14,  -8,  -4,  -4,  -1,   0,  22, -67,  38,   3,   3,   1,   0,   1, -29, -26, -10,  16,   0,   3,   0,   1,  -7, -14, -17,   4,   2,   1,   1,   0,  30,   2, -22,  -4,  -1,   0,   0,   0,   0,  24,   1,  -7,  -1,  -1,   0,   0,  -1,   1,   6,   0,  -1,   0,   0,   0 },
     {  -11, -24,  12,   1,   2,   0,   1,   0, -13,   3,   3,   0,   2,   0,   1,   0, -14, -15,   8,   5,   1,   2,   0,   0, -66,  17, -24,   8,   3,   2,   1,   1,  25,  44,  -5, -23,   2,  -5,   0,  -2, -13,  -7,  55,  -4,  -2,   0,   0,   0,   6, -46,   2,  22,  -2,   4,   0,   1,   4,  -2, -20,   4,   2,   1,   0,   0 },
     {  -13,   5, -17,   6,  -1,   2,   0,   0,  -8,  39, -24,   0,  -1,  -1,  -1,   0, -21,  17, -28,   4,   0,   2,   0,   0, -11,  56, -36, -11,   0,  -2,   0,  -1,   7,  22,  -9, -14,   2,  -3,   0,  -1,   9,  45, -14, -21,   2,  -4,   0,  -1, -18,  34,  33, -24,   3,  -3,   1,  -1,  -1, -11,  25,   0,  -2,   0,   0,   0 },
     {   12,  30,  75, -48,   4,  -7,   1,  -2,   6,  19,  -4,  15, -10,   1,  -2,   0,  -1, -19, -59,  38,   1,   2,   0,   1,  -6, -23,   3, -17,  11,  -2,   2,   0,  -4,   5,  10, -10,  -3,   2,   0,   0,   1,   7,  -1,   3,  -3,   0,   0,  -1,   0,   1,  -1,   2,   1,  -1,   0,   0,   0,   1,   2,  -1,   1,   0,   0,   0 },
     {    3,  -3,  -2,   3,  -3,   0,   0,   0,  18,  34,   0,  -9,  -2,  -3,  -1,  -1,  11, -19,   3,   0,   0,   1,   0,   0,  23,  31,   4, -20,   0,  -5,   1,  -1,  80, -31,  45,   1, -10,   0,  -3,   0, -29,  -4,   1,  30,  -3,   2,  -1,   0, -26,  -8, -12,   5,   8,   1,   2,   0,   3,  -4,  -8,  -3,   1,   1,   0,   0 },
     {    5,  11,   4,  -2,  -2,  -1,   0,   0, -11, -12, -23,   5,   1,   1,   1,   1,  15,  22,  11,  -8,  -3,  -2,   0,  -1, -18, -23, -26,  11,   3,   2,   1,   1,  33,  41,   9, -20,  -4,  -4,  -1,  -1,  -9, -41,  -8,  25,   2,   4,   1,   1,  10,  49, -44, -28,   4,  -4,   0,  -1,  -8,  14,  37, -12,  -4,  -1,   0,   0 },
     {    1,   4,   7,  -2,  -3,  -1,   0,   0,  -7, -18, -55,  27,   3,   3,   1,   1, -11, -34,   1, -33,  17,  -2,   3,   0, -10,  21,  60, -29, -18,   3,  -3,   0,   8,  36,  -9,  39, -15,  -3,  -1,  -1,  10,  -6, -10,   4,  18,  -4,   2,  -1,  -4,  -8,   2, -12,   3,   5,   0,   1,   0,   0,   1,  -1,  -6,   1,   0,   0 },
     {   -4, -12, -56,  28,  -5,   6,  -1,   1,  -1, -11, -52,  28,  -1,   3,   0,   1,   2, -20, -39,  16,  12,   0,   3,   1,  -4, -40,  -6, -29,  22,   1,   4,   1,  -9, -17,  24, -34,  -5,   6,  -1,   0,  10,   5,  18,   5, -17,   1,  -1,  -1,   3,   2,  -3,  15,  -2,  -1,   1,  -1,   1,   2,  -9,   2,   3,   1,   1,  -1 },
     {    8,  12,  10,  -2,  -4,  -1,  -1,  -1,   8,   5,   0,  -2,  -1,  -1,   0,  -2,  14,  23,   6, -12,   4,  -3,  -2,  -1,  23,   6,   8,   0,  -5,   1,  -2,   1,  26,  15,  -2, -18,  -1,  -4,   3,   0,  97, -10,  27,  -8, -18,   2,  -5,   1, -22,   7,  -1,  32,  -6,   1,  -1,   0, -23,   0, -11,   3,  11,  -4,   2,   0 },
     {    3,   1,  15,   2,   1,   1,  -2,  -1,  -4, -17,   5,   0,   0,  -1,   2,   0,   5,   5,  26, -10,   2,  -2,  -1,   1, -11, -29,  -2,  -8,   3,   1,   1,   0, -11,  10,  43,  -9,  -6,  -2,  -3,  -1,  -2,  -8,  -1,  11,   2,   1,   1,   0, -39,   3,  46, -11,   0,  -3,   1,   0,   6, -82,  15,  28,   0,   5,  -2,   2 },
     {    9,  13,  34,  68,  -4, -12,  -8,  -4,   6,  11,  15,   2,  15,   0,  -2,  -2,  -3,  -7, -39, -65,  -2,  18,  10,   1,  -6, -19, -28,   0, -22,  -2,   4,   3,  -4, -10,  13,  14,   4, -15,  -7,  -1,   1,   7,  12,  -2,  12,   4,  -6,  -3,   6,   5,  -4,  -1,  -2,  12,   3,   0,   2,   7,  -6,   2,  -6,   0,   4,   0 },
     {  -10, -21, -48,  -9,  18,   5,   1,   0,  10,  23,  69,  -5, -16,  -8,  -4,  -2,   1,  -3, -44,   5,  11,   7,   0,   0, -11, -22,  27, -22, -15,   1,  -1,   2,  10,  41,   6,  17,   0,  -8,  -3,  -2,   8, -19,  -7,   0,  11,   3,   0,   1,   4,  10,   3,  -7,  -5,   2,   0,   1,  -9,  -5,   7,   0,  -2,   0,  -1,   0 },
     {    6,   2,   5,  -3,  -9,   2,  -3,   0,   5,  -1,  -9, -12,  -6,  -3,  -2,  -1,   8,   0,   6,  -4,   2,  -1,  -1,  -1,  12,   5,  -7, -15,   2,  -3,  -1,   1,  28,  -5,  16,  -1,  -5,   0,  -3,  -2,  11,  -8,  -5, -16,   1,  -6,   3,   0, 104,  -7,  37,  -7, -13,  -8,  -5,   0, -10, -13,   9,  28,  -1,   4,  -3,   1 },
     {    4,   5,  -5,  22,  -7,   3,  -1,   0,   4,   2, -18,  23,  -8,   2,   2,   1,   7,   5, -22,  40,   0,   1,   0,   1,  11,  -1, -16,  39,   9,   4,   2,  -3,  25,  -7, -14,  20,  29,  -8,   7,  -3,  14, -49, -14, -26,  40,   9,   4,   0, -16, -29,  11, -26,  -8,  18,   1,   2,   9, -15, -13,   6, -19,   3,   5,  -1 },
     {   -1,   0, -12, -39,  34, -14,   2,   1,   3,   2,  10,  70, -46,  14,  -9,   0,   3,   6,  15, -35,  25,  -8,   6,   2,   1,   3, -40,   3,   1,   0,   0,   1,   2, -23,   7,  19, -15,   4,  -1,  -3,  -2,   5,  16, -18,  14,  -7,  -1,  -2,   3,   2,  -8,   7,   1,   2,  -2,   1,  -2,  -3,   2,  -3,   0,  -3,   0,   0 },
     {   -3,  -4, -21, -53,  15,  21,   5,   1,  -4,  -8, -31, -57,  17,  26,   4,   4,  -4,  -7, -31, -22, -10,   4,   7,  -2,   0, -14, -17,  37, -21, -10,   1,  -1,   1, -14,  17,  27, -13,  -9,   2,  -5,   9,  -6,  28,   5,  11,  -8,  -4,  -1,  -6,   1,  15,  -3,   5,  -1,  -2,   4,  10,   4,   1,  -4,  -3,  -1,  -3,   2 },
     {    0,   0,  -3,   2,   3,   1,   5,  -1,   1,   1,   9,  -3,   6,   1,   1,  -6,   6,  15,  18,  12, -10,  -4,  -3,   2,  -1, -11, -21, -18,  15,  -3,   8,   4,  -8,  13,  18,  -2, -34,  -5,  -6,   2,  38,  33, -17,  36,  35, -25,   4,  -3,   7, -39,  -4, -31,  33,  28,   2,   5,   2,  18, -18, -16, -30,  12,   1,   0 },
     {   -3,  -5,   7,  -4, -71,  34,  -3,   5,   0,   2,  12,  -7,   0, -17,   5,   0,   1,   5, -11,  -4,  76, -30,  -7,   2,  -2,  -1, -10,  16,  -5,  27, -14,   3,  -5,  -4,   4,   5, -23,   8,  14,  -4,  -5,   3,   2,  -1,   2, -11,   5,   3,  -3,   1,  -6,  -1,   8,   0,  -6,   2,  -3,  -1,  -1,  -4,   1,   1,  -2,  -2 },
     {    0,  -4,   2, -24,   8,  -2,   7,   2,   4,  -1,   9, -16,  33,  -7,   1,   2,   1,  -6,   6, -33,  23,  -4,   8,   3,   5,  -1,   8, -29,  33,  -8,   0,   1,   2, -11,   2, -34,  16,  20,  -3,   6,  -2, -11,  10, -45,  -3,  17,  -6,   4, -15, -13,  -4, -38, -28,  27,  -2,   1,   2,  16,  12,  -4, -44,   1,   4,  -1 },
     {    2,  -1,   3,  17,  10,  -7,  -2,  -1,   2,   0,   2,  -9,  -5,   5,   1,   5,   0,  -7,   2,  27,   9,  -4,   3,  -3,   8,   9,  20,  -6,  -5,   1,  -4,   5,  -4, -23, -19,   6,  10, -15,   8,   0,   4,  11,  63,   1,  12,  -8, -15,  -1,  16,  11, -27,  -6,  33,  14,  -3,   1,  -7, -41,  51, -20, -26,  22,   3,   4 },
     {   -2,  -3,  -4,   1,   0, -25,   7,   1,  -3,  -4,  -6,  -1,  -2,   3, -10,   2,  -2,  -3,   5,  10,  17,   8, -11,  -4,   0,  -1,  28,  51, -11,  -9,   4,  -3,   0,  18,  64, -12,  13, -16,   6,   4,   4,  50, -16, -32,  12,  14,  -7,   3,  11,  -8, -37,   0, -19,   4,   3,   2,  -3,  -8,   2,  -4,   5, -10,   1,   2 },
     {    4,   7,  11,  36,  47, -16,  -6,  -7,  -6,  -6,  -7, -53, -50,  26,  -1,   4,  -2,  -3,  -7,   2,  25, -41, -14,  13,  -1,   4,   4,  -2,  29,  10,  -5,   1,  -1,   3,  -7,  -4, -29,   8,  14,   6,  -5,  -7,  -6,  -2,  13,  -7,  -3,   2,  -8,   2,   0,   0, -18,   0,   3,   3,  -8,   7,  -7,   8,   6, -17,  -7,  -1 },
     {   -3,  -2,   4,  -2, -23,   3,  -8,   1,  -6,  -5,  -3,  -5,  -5,   7,   3,   3,   1,   5,  24,  32, -12,  16,  12,  -1,  -7,  -6,  -5, -31, -21,  -3,   3,   0,  -6,  -8,  16,  13, -10, -19,   3,  -8, -17,   2,  21, -22,  39, -12,  -5,  -1, -18,  30,   7,  10, -28,  27, -13,   0, -50,  32,  -7,  32,  -3,  -8,   9, -10 },
     {    4,   6,   6,  14,  47,  15,   0, -24,   3,   4,   1,  13,  47,  11,   0,  -9,   4,   6,   7,  22,  30, -12,  21,  -7,   0,  -2,  -4,  -1, -32,  13,  23,  13,   3,  -2,  10,   3,  14,  11,   7,  11,  -1,  -1,   0,  -3, -16, -44,  21,  -6,  -3, -12,   1, -21, -19, -15, -21,   4, -15,   7,   3,  -9,  28,  14, -13,   0 },
     {    1,   2,  -1,   3,   0,  -9,  13,   4,   2,   4,   2,  -4,   5,  12,   3,   2,   5,   9,   5,  15,  11,  53,  27,  -3,   4,   6,   0,  -7,  -8,  22,   3,   8,  -1,  -3, -17,  -3, -46,  -7,   8, -22,   7,   0,   8,  -3,  -8,   0,  36,   8,   2,   5, -25,  -8, -25, -14,  12,  22,  45, -18,  -6,  19, -21, -34,  -9,   4 },
   },
   {
     {  101, -54,  -1,  -4,   0,  -1,   0,   0,  33,  -1, -14,   2,  -3,   1,  -1,   0, -22,  29, -10,   0,   0,   0,   0,   0, -14,  11,   4,  -4,   1,  -1,   0,   0,  -7,   2,   5,  -1,   0,   0,   0,   0,  -3,   0,   2,   1,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   32,  30, -27,   1,  -5,   0,  -1,   0, -61,  72, -13,  -4,   0,  -2,   0,  -1, -37,   7,  28, -13,   5,  -3,   2,  -1,  11, -31,  25,  -3,  -1,   1,   0,   0,   8, -16,   4,   7,  -4,   1,  -1,   1,   4,  -5,  -3,   5,  -1,   0,   0,   0,   1,  -2,  -1,   1,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {  -32,  41, -29,   7,  -4,   2,  -1,   1,  68,  20, -41,  -1,  -4,  -1,  -1,   0, -11,  53, -10, -15,   2,  -4,   1,  -2, -21,  14,  28, -18,   3,  -3,   1,  -1,  -1, -13,  25,  -3,  -3,   1,  -1,   0,  -1,  -9,   7,   7,  -3,   1,  -1,   0,  -1,  -3,   0,   4,   0,  -1,   0,   0,   0,  -2,   0,   1,   1,  -1,   0,   0 },
     {  -47, -56,  55,  -7,   9,  -1,   3,   0, -24,  -4,   1,   9,  -2,   3,  -1,   1, -53,  47, -14,   2,   2,  -1,   1,  -1,  -1, -16,  21, -11,   5,  -2,   1,   0,  19, -21,   6,   3,  -5,   3,  -2,   1,   8,  -4,  -6,   6,  -3,   1,   0,   0,   3,   1,  -5,   2,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0 },
     {    1,   1,  22, -19,   2,  -2,   0,   0,  58,  -7,  32, -31,   0,  -4,  -1,  -1, -11, -16,  61, -17,  -6,   1,  -1,   0,   9, -43,  31,  20, -17,   7,  -4,   2,   1, -10, -14,  32, -11,   2,  -1,   1,  -5,   4, -17,  15,   2,  -3,   1,   0,  -3,   3,  -7,   3,   4,  -2,   0,   0,  -1,   1,  -2,   0,   2,   0,   0,   0 },
     {   21,  24,  18, -31,   3,  -5,   0,  -1, -29, -18,  47, -27,   5,  -1,   1,   0,  33,  12,  -7, -10,  -9,   1,  -2,   0, -23,  37,  28, -20,  -4,  -3,   0,  -2,   6, -29,  48,   1, -13,   4,  -4,   1,   6, -16,   4,  29, -16,   3,  -3,   1,   0,  -2, -10,  19,   0,  -4,   0,  -1,   0,   0,  -4,   4,   5,  -4,   0,   0 },
     {    3,  42,  -1, -12,  -2,  -3,   0,  -1,   2, -11,  41, -10,  -3,   0,  -1,   0, -71,  12,   5,  34, -14,   6,  -2,   1, -46,  20, -32,  32,   3,  -3,   3,  -1, -10,   1, -12,  -5,  21,  -8,   4,  -1,   7,  -9,   3, -15,  11,   0,  -1,   1,   4,  -6,   3,  -6,   0,   4,   0,   0,   2,  -1,   0,  -1,  -2,   2,   1,   0 },
     {    2,  11,  49, -46,   6,  -7,   2,  -2,  15,  67,   3, -22, -14,  -1,  -3,  -1,  32,  31, -19,  17, -19,   1,  -2,  -1,  34,   9, -23,  16,  -3,  -5,   1,  -2,   0,  20, -11,  -2,  10,  -6,   0,  -1,  -7,   5,   8, -10,   7,   0,  -1,   0,  -4,  -1,   8,  -3,   0,   1,   0,   0,  -2,  -2,   3,   0,   0,  -1,   1,   0 },
     {   11,  20,  -1,  11, -13,   0,  -2,   0, -20, -20,   0,  29, -12,   4,  -3,   2,  34,  39, -23,  12, -12,  -4,  -2,  -1, -16,  -5,  32,  25,  -7,  -4,  -1,  -1,   1, -18, -33,  50,   4,  -6,   1,   0, -14,  19, -36,  -1,  33, -12,   5,  -2,   0,  -1,   0, -25,  20,   0,  -1,   1,   1,  -2,   4, -10,   1,   4,  -1,   1 },
     {  -23, -44,   6,  20,  -4,   7,   0,   3, -13,  39,  -9,  -6,   1,  -3,   0,   0,  21,  -4,  40, -20,  -4,   0,  -2,   1, -55,  40,   5,  23, -20,   3,  -1,   0, -39,  16,   4,  18,   8, -11,   6,  -3,   2, -23,   7,   0,  16,  -5,   1,   1,   6, -11,  -1,  -1,   5,   4,  -2,   1,   2,  -3,  -3,   0,  -1,   3,   0,   0 },
     {   -1,  -2,   9,  31, -29,   5,  -6,   2,  14,  21,   6,  38, -38,   2,  -8,   0, -10, -22,   3,  43, -22,   1,  -2,  -1,  15,  10, -15,  10,  -5, -10,   1,  -2,   4,  -3,  31,   8, -13,  -6,  -3,  -1,   1,  -4, -10,  48, -12,  -9,  -2,  -2,  -9,  12, -22,  21,  25, -21,   3,  -2,  -1,  -1,  -1,  -9,  26, -10,  -3,   2 },
     {    9,   2,  26,  19,  -9,  -8,  -3,  -2,  16,  15,  15,  32,  -7, -15,  -1,  -3,  30,  -9,  17,  22,  17, -19,   1,  -3, -16, -26,  16,   7,  46, -14,   1,  -1,   3, -38,   8, -22,  42,   8,  -7,   3,   7, -26,   5, -22,   4,  29,  -8,   2,   1,  -6,  -7,  -1, -19,  21,   1,  -1,  -3,   2,  -7,   4, -12,   6,   5,  -2 },
     {    7,  24,  32,  22, -28,   1,  -8,  -2,  -6,  -7,  34,  31, -16,  -4,  -6,  -2,  -8,  26,  18, -21,  27, -18,   0,  -5,   9,   6,  15, -28,  12,  11, -10,   2, -32,  54, -14,  -5, -18,  16,  -3,  -1, -20,  13,  23,  -6,  -9,  -4,   8,  -4,   3, -15,  15,  10,  -3,  -7,   3,   1,   1,  -5,   1,   7,   4,  -5,  -1,   1 },
     {    2,  -7,  -5,  48, -33,  15,  -9,   4,   6,   6,  32, -22,  14, -13,   3,  -4,   4,  32,  10, -33,  20, -10,  -1,  -1,  28,  32, -37,  16, -10,   7,  -7,   0,  55, -23,  -7,   0,   5,  -3,   1,  -2,  14,   1,  -8,  -5,   5,   3,  -1,   0, -10,  12,   0,  -6,  -1,   4,  -1,   1,  -4,   0,   3,  -1,  -3,   0,   1,  -1 },
     {   -4,  -3,   3,   3,   6,  -4,   1,  -1,  14,  23,  19,  -7,  10,  -8,  -1,  -3, -23, -36, -19,  14,  28,  -6,   3,   0,  22,  34,   4, -42,  15,   1,  -4,   0, -19,  -7,  26,  33,  -9,  11,  -3,   1, -13,  -3, -39,  -1,  35,   0,   8,   0,   1,  -9,  10, -41,   6,  16,   0,   2,   2,  -4,   0,   0, -20,  14,   1,   0 },
     {    3,   8,   5,  22,  49, -48,  -1,  -4,   2,   6,   0,  22,  57, -44, -10,  -2,   4,  14,  -2,  22,  17,   8, -26,   4,   4,  -1,  -5,  26, -12,  32, -19,  -1,  -1,   2,   7,   7, -14,  16,   5,  -8,  -7,  -5,  10,  10,  -7,   0,   9,  -2,  -5,  -3,  -8,   8,   7,  -6,   5,   2,   0,  -2,  -4,  -4,   8,  -3,   1,   1 },
     {    3,   8,  28,  -5, -13,   7,  -3,   0,  10,  31,  -6,  -1,   0,  -1,   1,  -4,   9, -26, -23,  -7,  29,   6,   4,   1, -69, -14, -23, -16,   5,  27,  -1,   4,  18, -21, -29,  -9, -32,  15,   8,   1,   4,  24, -19,   1, -23, -15,   9,   0,   0,  11,   4,  -4,   1, -18,   1,   2,   1,   0,   7,  -3,   4,  -8,  -2,   2 },
     {   17,  38,  60, -17, -17,   2,  -5,  -3,  -3, -24, -59,  29,  15,   0,   8,   0, -18, -24,   8, -34,  14,   3,   1,   5,  13,  22,  -8,  13, -20,   1,  -1,   0,   3,  -5,   9,  11,  11, -13,   3,  -1,  15, -21,   0,   3,  10,   2,  -8,   3,  -3,   5, -10,  -6,  -2,   8,  -4,  -1,   0,   1,   1,  -2,  -3,   1,  -1,  -1 },
     {   -5,  -7, -11,  -8,   4,   0,  -1,   3,  11,  17,  23,  21, -11,  -7,  -4,  -2, -15, -27, -38, -17,  12,   6,   3,   1,  19,  29,  38,  11, -32,   0,  -7,  -4, -19, -30, -26,  15,  30,   3,   7,   2,  13,   9,   8, -37, -31,   3,  -1,   1,  -2,   7,   9,  36,  -5, -13,  -2,  -2,   0,   0,  -4,  -6,  14,  -6,   1,   2 },
     {   -1,  -4,  -1, -11,   9, -18,   8,   2,   1,  -3,   3, -12,  13, -32,  19,  -3,  -1,  -5,   0, -10,  27, -42,  17,  -5,  -4,  -6,   1,  -8,  21, -36,   6,   1,   3,   5,  -2,  -8,  21, -31,  -2,   5,   7,   2,   8,   2,  16, -27, -10,   3,   6,   1,   6,  -2,  45, -32, -10,   3,   9, -11,  17, -24,  42, -12, -13,   5 },
     {   -3,  -4,  -5,  -3, -34,  10,   1,   7,  -6,  -8,  -1, -28,  -1,  -3,  10,   1,  -1,   7, -40, -12,  29,  -8,   5,   0,   9, -39, -33,  28,   3,  10,   2,  -3, -50, -28,  12,  23,  -2,   8,   9,   2, -35, -27,  23,  16,   1,   4,   4,   1,  -9, -16, -10,  10,   5,  -1,   6,   1,   3,  -1, -13,   4,   2,  -2,   6,   2 },
     {   -5,   0,  -4, -12,  11,  12,  -6,   1,  -9,  -1, -14, -10,  14,   6,  -5,   4,  -4,   9,  12, -10,   8,  -9,  -3,  -1,  -1,  23,  -8,  15,  28, -19, -15,   4, -23,   8, -18,  -1,  23,  31, -37,   5, -19,  17, -45,  23, -33,  44, -10,  -8, -16,  17, -21,  -2,  -2,  -8,  18, -13,   3,  -4,   9, -20,  11, -11,   2,   0 },
     {    2,  -1, -15, -15,  12,   6,   1,   3,   6,   2,  -1,  10,  -2,   0,  -2,   5,   5,   2,  -8,   5,  10, -18,  11,  -3,   3,   7, -16,  -4,  19, -14,  -6,  10,  16,  15, -45,  44, -25,  15, -19,   4,  51, -49,   1,  22,  -4,  -1,   2,  -5,  25, -29,  -6,  24,  -8,   8,  -4,  -3, -10,  13, -12,  13,   2,   1,  -4,  -3 },
     {    4,   5,   7,   9,  30,  41, -48,  -6,   4,   6,   9,   8,  42,  45, -17, -24,   6,   7,   4,  17,  22,   7,  39, -30,   5,  -1,   5,  10,  -2, -10,  40,  -9,  -3,   1,   5,   4,  -6,  -7,  15,  13,  -8,  -4,   6,  -1,  -4,  -1,   0,  18,   1,   2,   1,   2,   3,  -9,  -1,   6,  -2,   0,   0,  -1,   8,  -3,  -3,   3 },
     {    3,   5,   1,   9,   0,   1,   1,  -3,   0,  -4,  10,  -5,  -2,   3,  -3,  -1,  10,  23,   0,  -3,  -3,   7,  -5,   2,  18, -16,  -4,  -8,  -3,  14,   2,   6, -59,  -2,   2, -28,  -7,   2,  16,   0,  63, -31, -37, -14,  -7, -18,  11,  -1,   5,  36, -20, -19,   1, -10,  -4,   3,  -3,   9,  16, -17,   1,   2, -10,   1 },
     {    1,   1,   5,  15,  -4, -13,  -1,  -7,  -4,  -6, -12, -16,   0,  -2,   0,  -4,   1,   2,   5,  15,  -8, -11,  -4,   1,   0,  -2,  -6, -12,  -2,  -3,   1,  -3,  -2,  -2,   9,  29,  -7, -14,   2,  -3,  -9, -11, -20, -28,  10,  12,  -5,   9,   8,  10,  15,   1, -60, -26,   5,   2,   8,   7,   6,  29,  35, -70,  -4,   5 },
     {   -2,  -3,  -2, -19, -16,  47, -25,  15,  -3,  -4, -10,  -9,  13, -21,  11,   2,   0,  -3,  -5,  40,   8, -46,  11,   3,  -5,  -6,  19,  23, -24,   8, -39,  25,   5,  14,  30,  -3,  -7,  11, -20, -25,  13,  14,   5, -17,   1,  -3,  16, -12,   9,  15,   9,   2, -10,   3,  -6,   4,  -2,   5,   5,  -1,  -4,  10,  -9,   2 },
     {    4,   5,  21,  35, -41,  13, -10,   8,   2,   5,  -6, -35,  54, -25,   1,   5,  -4, -11, -27,  14, -32,  33, -18,   5,  -1, -10,  28, -17,   4, -20,   9,  -2,  -8,  17, -23,  -5,  11,  -4, -18,  10,   8,  -1,  -2,   9,  -4,   4, -11,   2,  14,  -9,   5,  11,  10,   1, -10,  -3,  -2,   4,  -8,   6,   4,   5,   0,  -4 },
     {    1,  -2,   6,  -1,  -8,  -2,  -5,  -1,  -2, -10, -21,  -7,   3,  -3,   4,  -9,   0,  -3,  23,   1, -33,  -2,   2,  -5,  17,  22,  -2,  13,  28, -11,   4, -15,  -2, -34,  -5,  -5,   0,  43,  14,  -9,  -3,  -4,   4,  -8, -19, -19,  53, -10,  17, -21,  32, -11,  14, -35,  11,  22,   5,  -9,  -3,  13,  -6,   5, -11,  15 },
     {    0,   1,   1,  11,  -7,  -4,  54, -38,   2,   5,   5,  12,  14,  -9,  67, -38,   0,   5,  -1,   8, -16,   0,  50, -17,   2,   4,   2,  10, -19,  -4,   5,  11,  -8,   1,   0,  -5, -17,  10, -11,  12,   0,   7,  -5,   2, -11,   9,  -6,   8,   1,  -4,  -1,  -4,  -7,   3,   7,  -4,  -2,   3,   2,   0, -11,   0,   8,  -6 },
     {   -2,  -1,   2,   9,  -6, -41, -19,   0,  -2,   0,   0,  -3,  -4, -25, -39,  17,  -4,  -4, -12, -21, -13, -10,  43,  13,  -7,  -5,  -6,  26,   5, -31,  26,   2, -12,   4,  20,  17, -22,  10, -13,   3,  19,  28,   7,   0, -18,  -6, -17,  11, -11,  11,  10, -17, -23,   2,  -3,   4,  -6,   8,  15,  18,  -7,  30,  -2,  -8 },
     {    2,   3,   7,  21,  16,  -7, -29, -31,   0,   0,  -4, -21, -11,  23,  33, -16,  -5, -11, -20,  -3,  -8, -39, -30,  14,  -4,  -2,  14,  15,  19,  24,   4, -13,   6,  20,   7,  15,  13,  -5,   9,  31,  19,  -1,   0,   5, -15, -18, -22, -24, -18,   8,  10,   4,   2,   3,  35,   4,   2, -18,  -1,  17, -12,  11,   0,   2 },
   }
 },
 {  //26
   {
     {  112, -35, -10,  -3,  -3,  -1,  -1,   0,  -5,  34, -11,   0,  -2,   0,  -1,   0, -28,  13,  11,  -4,   2,  -1,   1,  -1,  -2,  -7,   8,   1,  -1,   0,   0,   0,  -4,   0,   0,   3,  -1,   0,   0,   0,  -1,  -1,   0,   1,   1,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   10,  35, -11,   2,  -3,   0,  -1,   0, -81,  28,  32,  -5,   6,  -2,   2,  -1,   4, -64,  21,  14,   1,   3,   0,   2,  13,  -9, -33,  11,   2,   1,   0,   1,   1,   5,  -6, -12,   4,   0,   1,   0,   5,  -1,   1,  -3,  -3,   1,  -1,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,  -1,  -1,   0,   0 },
     {   35, -18,  -5,   2,   0,   0,   0,   0,  -6, -55,  19,   4,   4,   2,   1,   1,  28, -21, -56,  21,  -3,   5,  -1,   2, -10,  52, -31, -31,   9,  -4,   3,  -2,   2,   4,  30, -19, -11,   1,  -2,   0,   0,   5,   3,  11,  -7,  -3,  -1,   0,  -1,   1,   2,   1,   3,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {   -7, -50,  19,   5,   3,   3,   1,   1,  43,  -2,  -9,   2,   1,   0,   0,  -1,  -6, -27,  22,  19,  -1,   5,  -1,   2,  -5, -10, -60,  32,  10,   4,   3,   1, -11,  23, -24, -42,  23,   0,   5,  -1,   5,  -5,  16, -22, -16,  10,  -2,   3,  -2,   6,  -1,   5, -13,  -3,   4,   0,   0,   0,   2,   0,   0,  -6,   2,   1 },
     {  -27,  -8, -14,   7,  -1,   4,   1,   2,   9,  57, -24, -19,   2,  -2,   1,  -1, -31,   0,   7,   3,   1,   1,   2,   0,  -1,   6, -20, -42,  14,  -2,   4,  -1,  14,  -6,  53, -34, -31,   6,  -5,   2,   5,  -5,   6,  38, -24,  -5,  -1,  -1,  -2,   4,  -2,   8,   7,  -9,   3,  -1,   1,  -2,   1,   2,   3,  -3,  -1,   1 },
     {    5,  28, -12,  -9,   2,   1,   1,   0, -28,  -8, -22,  11,   2,   5,   3,   2,  -1,  30, -20, -26,   9,   1,   3,   0, -11,   2,  14,  -3,   4,   3,   3,   0,   0, -16,  -5, -41,  24,  10,   5,   3,   8, -12,  22, -31, -48,  28,  -1,   6,  -3,   8,  -7,  28, -40, -17,  15,  -1,   1,   0,   3,   1,   8, -26,   6,   3 },
     {    7,  10,   8, -10,  -2,   0,  -3,   0,  -4, -12,  -5,  14,  -6,  -7,   2,  -1,  12,   7,   9, -17,  -5,   0,  -5,   1,  -6,  -3,   2,  28,  -9,  -9,   1,  -2,   0,  -3, -12, -23,   3,   4,  -5,   3,   0,  14,   8,  24, -44, -27,   3,  -7,   5,   5,   7,  17,  60, -42, -22,  -3,   3,  -1,   6,  -6,  16,  47, -29,  -6 },
     {   28,  81, -25,  -2,  -7,  -3,  -3,  -2,  70,  10,  27, -16,  -2,  -5,  -2,  -2,   8,  -6,  16,  20,  -8,   2,  -3,   0, -25,  11,  -8,  11,   9,  -2,   1,  -1,  -5, -10,   8,  -6,   9,   2,   0,   0,  -3,  -5,  -1,   1,  -2,   6,   0,   0,  -2,  -1,  -1,   2,  -3,   1,   2,  -1,  -1,  -1,  -1,   1,   1,  -1,   0,   1 },
     {  -10,  18, -23,  15,   1,  -1,  -3,   0, -17,  27, -34,  24,   4,  -3,  -2,  -1, -19,  31, -32,  18,   8,  -5,  -1,  -2, -17,  26, -22,   6,  19, -13,   3,  -4,  -9,  12,  -6, -14,  32, -15,   1,  -3,   4,  -4,  15, -27,  37,  -8,  -6,   0,   8, -13,  18, -24,  21,  20, -19,   2,   1,  -1,   2,  -2, -11,  30,  -9,  -4 },
     {  -15, -24, -25,  -8,   9,   5,   2,   3,  23,  37,  35,  11, -15,  -9,  -4,  -3, -27, -38, -34,  -3,  18,   9,   3,   3,  22,  32,  22,  -7, -25, -11,  -3,  -3, -15, -19,  -5,  22,  31,   8,   1,   1,   3,   1, -10, -29, -23,   1,   4,   2,   1,   3,  11,  19,   7, -11,  -6,  -2,   0,  -1,  -4,  -5,   4,   8,   1,   1 },
     {   -2, -18,  -2,   0,  -3,   1,  -2,   1,  -7,  28, -20, -10,  -1,  -5,   1,  -2,  66,   0,  28, -26, -13,  -3,  -5,   0,   1,  54,   8,  32, -20,  -9,  -5,  -4,  -9,   4,  36,   8,  30, -12,   0,  -5,   3, -21,   7,  15,   8,  20,  -5,   4,  -6,   1, -13,   4,   2,   8,   9,  -1,   0,  -4,  -1,  -4,  -1,  -2,   3,   2 },
     {   -9, -13,  50,  -7,  -3,  -3,  -1,  -1, -22,  -3,  49,   4,  -9,  -2,  -4,  -1, -43,  30,   9,  35, -17,   0,  -4,   0, -34,  15,   6,  25,  13, -14,   4,  -5,  13, -36,  23,  -1,  25,  -3,  -1,   1,   8, -16, -10,  13,   0,  13,  -5,   3,  -2,   2, -15,   5,   3,   3,   4,  -1,   1,  -1,  -3,  -3,   3,   0,   2,  -1 },
     {    0,   1,   3,  13,   9, -11,  -4,   0,  -3,  -2,  -6, -16, -15,  11,   3,   0,   0,   5,   6,  20,  11, -17,  -4,   1,  -4,  -2, -14, -20, -15,  15,   3,  -1,   6,  13,  15,  26,   7, -33,  -7,   0,  -7, -12, -16, -20,  10,  34,   6,  -1,   3,   4,   5,  -4, -25, -59,   0,   5,   6,   5,  12,  11,  39,  22, -47, -13 },
     {  -11, -38, -44,  13,   5,   1,   2,   2, -11,  25,  30, -28,  -4,   0,  -2,  -1,  56,  28,   3,  30,  -3,  -6,  -2,  -2, -38, -15,  11,  -3,  27,  10,   3,   2,   3, -26, -27,  -6, -13,   9,   5,   4,  -2,  16,  -5, -16,  -7, -12,  -1,  -2,   5,  -1,  12,   7,  -3,  -2,  -3,  -1,   0,   3,  -1,   2,   7,   3,   1,   0 },
     {    6,  -8,   9, -11,  18,  -3,  -3,   0,   9, -15,  14, -17,  24,  -2,  -5,  -1,   9, -16,  18, -21,  26,  -5,  -6,  -1,  12, -20,  20, -25,  22,   2, -10,   0,  15, -20,  20, -22,  18,   7, -16,   3,  12, -15,  16, -13,   4,  23, -25,   5,   5,  -4,   5,   1,  -7,  34, -30,   0,  -4,   9,  -9,  15, -21,  49, -27, -11 },
     {    1,  -3,  -6,   8,   7,   6,  -1,   1,   1,  -2, -19,  12,  19,   3,   6,  -1,   6,  -7,  -1,   3,  14,  10,  -1,   2,   9,  -2,   1,   9,  28,   4,   5,  -3,   7, -22,   1,  -2,  23,  38,  -3,   8, -25,  17, -52,  14, -14,  37,  24,   5,  -4, -10,  -3, -51,  12, -16,  34,  -4,   2,  -7,   3, -12, -24,   7, -11,   4 },
     {   -9, -15, -69,   9,   2,   9,   3,   5,  -8, -48, -17, -29,  11,   5,   9,   2, -43, -20,  23, -22, -17,   8,   1,   4, -36,   5,  -2,  19, -23,  -7,   4,   0,  14, -22,   3,   5,   6, -19,   0,   0,   8,   3,  -9,   2,   5,   1, -10,   1,   1,   6,  -2,  -5,   2,   3,  -2,  -2,   3,   0,   3,  -2,  -2,   2,   4,  -2 },
     {    2,   5,   4,   7,   8,   3,  -5,   2,  -3,  -2,  -7,  -2,  -8,  10,  12,  -2,   3,   6,   5,   8,  10,   5,  -5,   3,  -2,  -4,  -7,  -5,  -9,  13,  13,  -2,   4,   6,   5,   8,   4,   3,  -7,   3,  -5,   2,  -5,   0,  -8,  22,  21,  -5,   2,  11,   7,  14,   7,   5,  -1,  17,   1,   6,   3,  10,   7,  49,  93, -41 },
     {    4,   5,  44,  -6,   0,   2,  -1,   4,   6,  25, -28,  24,   4,   2,   4,  -3,  17, -41, -12, -20,  15,   8,   4,   4, -75,  -4,  -3, -15, -20,  10,   5,   2,  12, -38,  -9,   6, -11, -18,   5,   0,   8,  13, -15,  -9,   5,  -8, -10,   0,   2,   5,   8,  -6,  -4,   4,  -1,   0,   2,   0,   1,   4,  -3,  -2,   2,   3 },
     {   -1,  -1, -16,  -7,  23,  -5,   7,  -2,   3,   2,  17,  14, -12,  17,  -8,   2,  -1,   3, -11,   4,  30,  -8,   8,  -5,  -9, -18,  -2,  -5, -11,  25,  -9,   4,   9,  14,  -8,  -3,   7, -24,   8,  -7,   5,  18,  26,  29,  11,  34, -29,   3,  -6,   4, -11,  19,  32,  19,  58, -16,   0, -14,   0, -31,   5,  10,  -6,   7 },
     {    2,  -5, -16,  67, -44,  17,  -5,   0,   5, -15,  -4,  60, -33,   0,   0,  -2,   7, -21,  20,   8,  13, -21,   9,  -6,   0, -15,  22,  -8,  22, -15,   0,   2,  -3,  -9,  11,   0,   5,  -1,  -9,   3,   5, -11,   0,   9,   1,   3,  -6,  -2,  -3,   6, -13,  14,  -3,   3,  -6,   1,  -1,   3,  -4,   1,   8,  -3,   0,   0 },
     {   -8,  -3,   9,  11,  -5,   5,   0,  -2, -10,  -7,  25,  -5,   6,  -3,   1,   2, -21,   0,  32, -27,  26,  -3,  -2,   3, -49,  31,  15, -16,   9,  15, -13,   9, -57,  42,  -2,   1, -13,  25, -11,   2, -15,   2,  14,  -1,  -4,   6,  17,  -8,   7, -24,  16,  -4,   1,  -4,  10,  -3,   3,  -7,  -2,   3,  -6,   2,  -4,   1 },
     {    3,   1,  17,  55, -44,  28, -14,   5,   3,   8,  26, -28,  43, -30,  19,  -9,   0,  21, -26, -17,  11,  19, -12,   9,  14, -12,  -8,   4, -32,  13,  11,  -6,  -1,  -5,  -5, -12,   7, -24,   8,   1,  -8,   9,   0,   5,  -2,   9, -12,   4,   3,  -7,   6,   1,   3,  -1,   7,  -5,   0,  -1,  -1,   1,  -3,   2,  -2,   2 },
     {   -4,   5,  -6,   5,  -9,  21, -16,   5,  -6,  10, -12,  11, -17,  26, -23,  10,  -6,  11, -13,  13, -21,  28, -20,   5,  -6,  12, -14,  18, -20,  26, -19,   2,  -7,   9, -12,  17, -23,  24, -15,  -2,  -7,  12, -13,  15, -21,  22, -20,  -1,  -9,  12, -14,  17, -26,  27, -17, -15, -11,  15, -16,  20, -21,  22, -29,  -9 },
     {   -5, -11, -27, -30,  -2,  12,  -2,   4,   4,  11,  35,  37,  -4, -15,   0,  -3,  11,  10, -11, -16,   8,  20,  -3,   3, -27, -33,  -8,   1, -22, -18,  13,   4,  35,  36,  12,  -4,   7,  -3, -21,  -4, -15, -12,  13,  16,   4,   9,  23,   4,  -1, -16, -21, -25, -12, -11, -23,  12,   0,   9,   7,  10, -13, -11,   0,  -6 },
     {    6,   8,  10,  16,  20,  13,  -1,  -7,  -9, -14, -17, -23, -26, -17,   2,   6,   9,  14,  16,  24,  27,  19,  -4, -13, -10, -15, -15, -21, -24, -13,   3,   9,  10,  11,  14,  18,  21,  11, -14, -18,  -5, -10,  -9, -16, -19,  -7,  13,  15,   6,   8,   9,  15,  19,   9, -21, -33,  -8, -14, -12, -20, -21, -12,   6,  47 },
     {    2,   2,  11,   5,  -5,  -4,  -1,   2,   1,   5,  -2,  -2,  -8,  -6,  -3,  -1,  -2,  -3,  12,  -3, -35,  -1,  -4,   5,   3,  12,  10, -15,   5, -45,   4,  -4,  38,  19, -12,   0,  13,  -5, -40,   9,  -4,  59,  12,  -7,  -1,  24, -10, -24,   3,  -7,  44,   8,  -5,   2,  22,  -9,  -2,  -1,  -3,  21,   0,  -7,   2,   7 },
     {   -3,  -3,   1, -18,  -4,  14,  -5,  -3,  -4,   1, -25, -18,  13,   1,  13,  -3,   5, -25, -28,  33, -31,  21,   6,   3, -15, -41,  31,  -4,   9, -29,  26,  -6, -58,   8,   9,   9,   9, -12, -16,  15,  -9, -15,  15,   9,  -1,  15,  -5,  -3,   4, -11, -13,  13,   4,  -2,  10,  -6,   2,   0,  -7,  -9,   0,   9,  -5,  -5 },
     {   -3,  -2,   1, -41, -60,   9,   1,  11,  -3,  -2,  -4, -36, -58, -18,  20,  11,   0,   0, -20, -17,   9, -40,   6,  16,  -1,  -9, -20,  -1,   9,   1, -21,  13,  -6,  -8,  -6,  -3,   7,   4, -14,   1,  -4,  -4, -13,   0,   2,  10, -12,  -1,   2,   2,  -1, -14,   5,   4,   3,  -7,   1,  -4,   6,  -4, -11,   5,   6,  -5 },
     {   -1,   1,  -2,  -1,  10, -10,  -9,   5,  -1,   4,  -4,   1,  34, -19,  -9,   6,  -3,   1,  -6,  -4,  14,  -7, -17,  15,  -7,   2,  -6,  -2,  19, -10, -23,  12,  -8,  12, -11,   9,   9,   7, -36,  18, -17,  14, -24,  13,   5,  11, -34,  13, -16,  25, -21,  11,   1,   6, -37,  16,  11,  -8,  24, -22,  48, -15,  18,  18 },
     {   -1,  -3,   3, -11, -38,   0,  -1,   5,  -2,  -3, -26, -16,  15, -14, -11,  -6,  -8, -24,  -3,  28,  15,   6,  -6,   3, -12,  15,  27,  12,  16,  15, -15,  -6,  56,  19,  -3,  18,   6,  21,  27,  -6, -34,   0,  32,  -7,   0,  -7,  11,  23,   8, -18, -10,  21,   3,  -2,   4,   4,  -4,   5,  -8,  -9,  16,   3,  -4, -12 },
     {    5,   5,  16,  53,  42,  -2,  -1,   7,   4,  12,   6, -36, -22,  29,   2,  12,  -5, -21, -45, -27, -19, -30,   0,  -7, -15, -13,  13,  26,   5,  -8, -17,   1,  17,  14,  -1,  -6,   6,  19,  -4,  -7, -12, -14,  13,   9,  -1,  -2,   7,  -1,   2, -10, -12, -10,   3,  -4,  -8,   5,   0,   9,   8,   2,  -1,  -2,   6,   7 },
   },
   {
     {  116, -12,  -9,  -4,  -4,  -2,  -2,  -1, -45,  17,   2,   2,   1,   1,   0,   0, -16,  -3,   2,   0,   1,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0,  -3,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   28, -28,   1,  -1,   1,   0,   0,   0, 100,  12, -17,  -3,  -5,  -1,  -2,  -1, -52,  35,   6,   0,   1,   0,   0,   0, -16,  -7,   6,   0,   1,   0,   1,   0,  -4,   0,   1,   0,   0,   0,   0,   0,  -3,  -2,   1,   0,   0,   0,   0,   0,  -2,   1,   1,  -1,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0 },
     {  -28,  -3,   4,   2,   1,   1,   1,   0, -25,  42,  -1,  -1,  -1,  -1,   0,   0, -80, -31,  27,   3,   6,   1,   2,   0,  49, -44, -14,   3,  -2,   1,   0,   0,  12,  15, -10,  -2,  -1,  -1,  -1,   0,   4,   0,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   0 },
     {   -6,  36,  -1,  -2,  -2,  -1,  -1,   0, -29,  -6,   2,   4,   1,   1,   1,   0, -23,  50,   7,  -6,   0,  -2,   0,  -1, -52, -48,  39,   6,   6,   2,   2,   0,  38, -43, -24,   8,  -1,   2,  -1,   1,   4,  16, -11,  -4,   0,  -1,   0,   0,   4,   0,   2,  -1,   0,   0,   0,   0,   1,   1,   0,   1,   0,   0,   0,   0 },
     {  -20, -99,  10,   5,   6,   4,   3,   1, -25,  51, -16,  -1,  -3,  -1,  -1,  -1,  35,  20,   9,  -4,  -1,  -2,  -1,  -1, -21, -12,   8,   4,   1,   1,   1,   0,   7,  -7,  -6,   1,   0,   0,   0,   0,   2,   1,  -2,  -1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0 },
     {  -13,  11,  -9,   3,   0,   1,   0,   0,  -7,  46,  10,  -6,  -2,  -2,  -1,  -1, -28, -12,  -7,   4,   2,   1,   1,   0, -10,  53,  22, -15,  -1,  -4,   0,  -1, -30, -64,  39,  16,   4,   4,   2,   1,  24, -20, -33,   5,   0,   2,   0,   1,  -3,   9,   1,  -4,   0,  -1,   0,   0,   3,  -3,  -2,   1,   0,   0,   0,   0 },
     {   -9,  27,   5,  -3,  -2,  -1,   0,   0, -22,  24, -22,   0,  -1,   0,   0,   0,  -7,  50,  14,  -9,  -1,  -3,   0,  -1, -17,   7,  -2,  -2,   2,   0,   1,   0, -12,  37,  40, -20,   0,  -5,   0,  -1, -10, -58,  41,  23,   0,   4,   0,   1,   8, -17, -31,   9,   3,   2,   0,   0,  -1,   4,  -2,  -8,   2,   0,   0,   0 },
     {   12,  47, -15,   2,  -4,   0,  -2,   0,  33,  71,   4, -11,  -3,  -5,  -1,  -2,  50, -39,  35,  -1,   1,  -1,   0,   0, -27, -31,  -5,  14,   0,   4,   0,   1,  -8,   4, -14,   1,   1,   0,   1,   0,  -4,   1,   1,  -2,   0,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,  -1,   0,  -1,   0,   1,  -1,   0,   1,   0,   0 },
     {  -10,  19, -17,  -1,  -1,   1,   0,   0, -10,  39, -12,  -6,  -2,  -1,  -1,   0,  -9,  31, -25,  -6,   0,  -1,   0,   0,   2,  39,  -2, -15,  -1,  -3,   0,  -1,  -1,  24, -13, -14,   2,  -2,   0,  -1,  -7,  48,  30, -37,   0,  -6,   1,  -2,   0, -19,  61,  -3,  -7,   1,  -1,   0,   5, -18,   7,  13,  -3,   2,  -1,   1 },
     {    5,  12,   6,  -5,   0,  -2,   0,   0,   0,  -1, -22,   6,  -1,   2,   0,   0,  24,  44,  18, -22,  -1,  -5,   0,  -2,  68, -22,  33,  13,  -7,   0,  -3,   0, -67, -12, -21,  20,   8,   4,   2,   1,  11,  -2, -12, -11,   1,  -1,   0,   0,  -2,   7,   5,  -1,  -1,   0,  -1,   0,   1,  -2,   1,   2,   0,   0,   0,   0 },
     {   -2,  -9, -17,  -5,   3,   0,   1,   0,   7,  10,  33,  10,  -3,  -1,  -2,  -1,  -6, -16, -31,  -8,   4,   1,   2,   0,   7,   8,  40,   9,  -6,  -2,  -1,   0,  -9, -14, -33,  -1,  10,   2,   2,   1,   5,  10,  50,  -1, -12,  -2,  -2,   0, -11, -42, -37,  42,  12,   5,   2,   1,   0,  19, -24, -31,   3,  -1,   1,   0 },
     {   -9, -15, -95, -10,  -1,   4,   1,   2,   0, -10,  53, -11,  -3,  -4,   0,  -1,   4,  36,  31,  23,  -2,  -1,  -2,  -2,  14,  -7, -23,  -1,   7,   1,   2,   0,  -5,  -5,   8,  -6,  -1,   1,   0,   0,  -5,  -3,  -4,   5,  -1,   0,   0,   0,   2,   2,   2,  -1,   0,  -1,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0 },
     {    5,   5,  -4,   6,  -1,   0,  -1,   0,   9,  18,  -1,  -8,  -1,  -2,  -1,  -1,  14,   7,  -7,  -2,   0,  -2,   0,   0,  43,  10,  34, -20,  -5,  -3,  -1,  -1,  52,  -1,  17,  36,  -9,   1,  -4,   0, -75,   2, -20,  17,  16,   2,   4,   0,   7, -23, -10, -14,   5,   2,   1,   1,   2,   6,  -5,  -2,  -1,   0,   0,   0 },
     {   -5,   4, -20,  -7,   2,   1,   1,   0,  -3,  13,  -9,   8,  -3,   0,  -1,   0,  -5,   5, -28, -10,   3,   0,   0,   1,  -1,  11,  -4,   5,   0,  -1,   0,   0,  -1,   9, -28, -17,   4,   0,   3,   1,  -6,   6, -12,   0,   0,   0,  -1,   0, -12,  30, -40, -49,  17,  -3,   4,  -1,  15,  12,  67, -48, -24,  -2,  -3,   0 },
     {   -1,  -2,   1, -12,   3,   0,   1,  -1,   4,  15,  48,  19,  -6,   0,  -3,   0,   3,   2, -65,  -6,  13,   3,   4,   1, -10, -48,  15, -33, -11,   3,   0,   1, -17,  40,  19,  25,  -8,  -6,  -3,  -1,  15, -22, -13, -10,  12,   4,   2,   1,  10,  13,  11,  -5, -12,   0,  -2,   0,  -4,  -3,   5,  13,   1,  -1,   1,   0 },
     {   -4,  -6, -69,  18,   4,   5,   0,   2,  -4, -19, -74,   4,   7,   2,   2,   0,  -4, -35, -14, -30,  12,  -1,   1,   1, -13, -22,  24, -23,  -3,   6,  -1,   1, -10,   3,  18,   1, -15,   3,  -2,   1,   0,   4,   4,   8,  -5,  -2,  -1,   0,   2,   5,   3,   9,  -1,  -2,   0,   0,  -1,   1,  -7,   5,   4,   0,   0,   0 },
     {   -8, -11,   7,   7,   5,   0,  -2,   0, -14, -17,   9,   3,   1,   0,   2,   0, -23, -18,  11,  10,   2,   5,  -1,   0, -40,  -4,  -1,   7,   0,   3,   3,   0, -66,   1, -18,  10,   8,   2,   2,   0, -73,  -4,  -6, -15,  11,   1,   4,  -1,  27, -24,   6, -14,  -4,  -1,   1,   0,  23,   3,  -1,   1,  -7,  -1,  -1,   0 },
     {    2,  -2,   3,  69, -34,   6,  -9,  -1,  -2,  -1,   9, -71,  42,  -9,   4,   3,  -2,   6, -25,  18, -26,  11,  -3,   2,   3, -12,  19,   5,   4,  -6,  -1,  -1,  -8,  14,  -8,  -8,   9,   6,   1,   1,   7,  -9,  -1,   5, -13,   2,   1,  -1,  -5,   3,   1,  -5,   0,  -2,  -2,   1,  -1,   3,   2,  -1,  -4,   0,   1,  -1 },
     {    4,   7,  18,  11,   4,  -1,   0,   0,   4,  13,  -3,  13,  -3,   1,   0,  -1,  12,  13, -31,  23,  11,   0,   1,   2,  17, -34, -35, -35,  26,  -4,   8,   1, -21, -45,  14, -45, -25,  11,  -3,   3, -26,  23,   7,  27, -28,  -5,   1,  -2,  -1,   2,   0,  12,  13, -11,   4,  -1,  11,   3,  -3,  -3,   5,  -1,  -3,  -1 },
     {    4, -12,  14, -14,   0,  -1,   1,  -2,   7, -18,  24, -22,  -2,  -4,  -1,   1,   7, -22,  33, -25,  13,   4,   2,   1,   9, -17,  41, -24,   9,  -6,   8,  -5,   4, -20,  36, -37,   3,   2,  -5,   1,   7,  -3,  37, -16,  -1,   6,  -5,   1,   8, -20,  17, -28,  -7,  10,   3,   4,   5, -10,  37,  -1, -32,   6,  -3,   1 },
     {   -2,  -5,  -7, -13, -18,   4,   1,   1,   7,   8,  18,  43,  42,   3,   0,  -1,  -5,  -8, -13, -51, -49,  -6,   0,   2,   3,   3,   5,  41,  27,  -3, -12,  -3,   7,   1,   7, -21,  -2,  25,   9,   6, -14, -12, -10,  15, -23, -14,   2,  -4,  13,   3,  17,  -5,   8,  -1,   2,   0,   5, -13,  -3,  20,  -2,  -1,  -1,   1 },
     {   -4,  -2,  12,  -5,   1,  -3,   2,  -2,  -7,  -5,  21,  -8,  -6,  -3,   5,   0,  -8,  -1,  26,  -3,  -5,  -1,  -8,  -3, -17,   5,  30,  -1, -10,   1,  -4,  -3, -24,  36,  22,  16, -30,   2,  -6,   1,  -2,  52,  -1,  47,  -7, -19,  -2,  -5, -34,  18,   4,  10,  43, -12,   1,  -4,  -3, -19,  11,  -8,  12,  11,  -4,   0 },
     {    2,   0,   4,  71, -10,  17,  -7,   7,   4,   0,  29,  37,  14,   0,   9,   4,  -1,  10,  22, -44,  45, -19,   3,   1,  -2,  15, -15, -13, -20,   7,  -8,  -2,   2,  -7, -16,  -5, -20, -10,  -5,  -1,  -2,  -4,   7,   6,  19,   6, -14,  -1, -10,  12,   6,  -2,  -3,  18,   1,   2,   0,   4,  -3,  -2,  -2,   0,  -2,  -2 },
     {    1,  -1,  -8,  38, -23,  16,  -8,  -3,   1,  -3, -10,  60, -30,  -5,   5,   5,   4,  -5,  -2,  50, -11,  11,  -2,   0,   5, -13,  19,  22,  -1, -24,   7, -12,   4,  -1,  22,  -7,  32, -10,   3,  -6,   6,  -6,   8,  -9,  20,  -8,   3,  -1,  -6, -14,  14, -12,  30,  -4,  -1,   1,  -8, -23,  11,  -6,   5,  13,   3,   0 },
     {    6,   6,   0,  42,  72,  29,  22,   0,  -2,  -4,  10, -17, -56, -12, -10,   7,  -2,  -1, -10, -20, -22, -18, -11,  -6,   4,  -1,   1,  24,  29,  18,  -1,  -2,   3,   5,   6,  -3, -13,   8,  13,   1,   4,  -8,  -8,  -8,  -6,  -3,   8,   3,  -2, -10,   3,  -3,  -1, -10,   2,   2,  -3,  -3,   1,   0,  -4,  -5,   0,   1 },
     {    0,   0,   1,  12,  -5,  -5,   2,  -2,   0,   4,   5,   2, -15,  -1,   6,   0,  -3,  -2,   7,   4,  -1,   2,   5,   2,   3,  14,  12,  -2,   0,  -8,  -3,  -1,  -2,  -2,  -4,   2,  -6,   5,  -4,  -3,  -8,  14,  21,  -8, -17,  -8,   2,   0,  81,  41, -12,  10,   5, -10,  -5,   0, -65,  22,  21,  10,  13,   4,  -3,  -1 },
     {   -1,   0,   0,   1,  -3,   3,   4,  -5,  -2,  -1,   8,   1, -12, -10,  15,   6,  -1,   0,   3, -19,  21,   9, -12,   0,  -2,  -1,   6,  12, -21, -21,  10,  -4,   9,   4,  15, -10,  35,  12,  -7,   5, -34,  -5,  -7, -28, -48,  39,   8,   9, -34,  32,  -1,  26, -22, -40,  -3,  -4,   4,  -7,   6,  -1,  23,  -4,   6,  -6 },
     {    1,   1,   1,   3,  31, -21, -13,  -3,   0,   0,  -1,  -4,  26,   3,   1, -12,   0,   4,  -2,  -7,  10, -15,   0,  -3,  -2,  -6, -12,  -4,  21,  -3,   4, -13,   3,   7,  15,   6,  29,  13,   3,  17,  -6,   2, -16, -16,   7,  12, -14,  14,  10,  11,  15,  27,  34,  26, -29,  -6, -13,  -4, -11, -42, -21,  55,  25,  -4 },
     {    0,   0,   0, -12,  16,   9,  -1,  -4,   0,  -1,  -1,   7,   7,   7, -12,   1,   1, -10,   9,  42,   6, -16,   0,  -5,   3,  16,  42, -24,  17,   8,  -1,   0,   7,   6, -48, -21, -21,  19,   2,  -2, -16, -52,  12,   5,   1,  -7,   0,  -3, -18,  43,  27,   9,  -2,   1,  -6,  -5,   1, -20,  -4,  -1,   3,   1,   0,  -2 },
     {   -1,  -2,   2,  -3, -41,   7,  -6,  -3,   0,  -3,   0,  -4, -18, -22,  -4,   8,   1,   3,   5, -11,  41,   7,  13,  -3,   1,  12,   3,  -3,  31,  14,  15,  10,   4,  16, -11,  -7, -21,   4, -12,   4,  11, -13, -36, -26, -23, -39,  20,  -2,   9, -29, -16,   2,  15, -26, -29,   6,   8,  -9,  -8,   2,  -2,  30,  -6,  -7 },
     {   -1,  -2,   0,   5, -29,   1,   2,  -5,  -2,  -4,   2, -12, -41,  44, -16,   5,  -1,  -2,   6, -26, -24,  -1,  34,   3,  -7,  -5, -20, -23,  19, -51, -11,  -1,  -1,  -5,  -7,  34,  -2,  16,   0, -24,  -4, -11,  21,  -4,  -7,  -7,  18,   8, -18,  16,  11,  -6,   6,   1,  -4,  -1,   6,   9, -16, -17, -14,   6,  -1,   6 },
     {    2,   1,  -1,  -5,  25,   5, -15,  21,   0,  -3,   1,  -9,  -2, -20, -31,  12,   1,   1,  -4, -13,  -7,  29,  -4,  22,   0,  -5, -12,  -6, -62, -12,  -9,  -5,   6, -16, -16,  -5,   7, -19,  15,   1,  -2, -19,   7,  22, -23, -23,  -1,  17,  14,  -3,  16, -15,  16,  -4, -30, -11,   7, -18,  -6,  -1,   7,  27, -31,  -3 },
   },
   {
     {  103, -13, -27,  -4,  -4,  -1,  -2,   0, -54,  32,  16,  -5,   0,  -1,   0,  -1,  -9, -19,   9,   7,   1,   1,   0,   0,   2,   1,  -7,   0,   1,   0,   0,   0,  -2,   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   17, -72,  20,  11,   3,   3,   1,   1,  47,  44, -47,  -8,  -3,  -1,  -1,  -1, -31,  33,  30, -13,  -3,  -3,   0,  -1,  -4, -22,   8,  13,   0,   1,   0,   1,  -3,   4,  -7,   0,   2,   0,   0,   0,  -1,  -2,   2,  -1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   29,  68, -24, -16,  -5,  -4,  -1,  -2,  65, -12,  10,  -5,  -6,  -2,  -2,  -1, -68,   2,  18,  14,   3,   0,   1,   0,   1, -19,  -2,   3,   4,   1,   1,   1,   0,   1,  -4,   0,   0,   1,   0,   0,  -1,  -1,   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   49,  -3,  17, -20,  -3,  -3,  -1,  -2,   6, -59, -14,  33,   1,   4,   0,   2,  10,  57, -41, -26,   5,  -1,   1,  -1, -17,   9,  42,  -8, -10,  -1,  -1,  -1,  -1, -10,  -6,  16,   1,  -1,   0,   0,  -4,   4,   0,  -5,   3,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,  -1,   1,   1,  -1,   0,   0,   0,   0 },
     {    6,   7,  76, -30, -11,  -2,  -2,  -2, -14,   0, -32,  45,   3,  -3,   0,   0, -35, -45,  -8,  -4,  22,   7,   3,   2,  30,   4, -23, -13, -11,   5,   1,   1,  -2,  19,   4,  -7,  -3,  -5,   0,  -1,   4,  -4,   6,   2,  -2,   0,  -1,   0,  -2,   2,  -2,   1,   1,  -1,   0,   0,   1,  -1,   1,  -1,   0,   0,   1,   0 },
     {   10, -53,  -9,  10,  13,   2,   3,   1,  45, -21,  29,  -8, -12,   1,  -2,   0, -16, -18, -45,  35,   9,   2,   1,   2,   3,  43, -26, -41,   6,   1,   2,  -1,  -3,   6,  32,  -3, -15,  -2,  -1,   0,  -2,   0,  -5,   9,   3,  -2,   0,   0,  -2,   1,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0 },
     {  -16, -26, -32,  15,  11,   2,   2,   2, -36, -48,  11,   3,   5,   7,   2,   2, -43,  32,  11, -15,  -5,  -1,   2,  -1,  73, -17, -14,   6,  -7,  -3,  -3,   1, -10,  28,  -5,  -6,   2,  -1,  -1,  -1,   3,  -4,  10,  -1,  -3,   0,   0,   0,  -2,   2,  -2,   3,   0,  -1,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0 },
     {  -20,  11, -23,   6,   3,   2,   1,   1,  -9,  48,  -3,  -1,  -9,  -4,  -1,  -1, -26,  -5, -22, -23,  23,   2,   4,   0,  30,  11,  51, -33, -34,   1,  -2,   0,  -3, -15,  18,  55,  -5, -13,  -3,  -1,   0,  -4, -19,  -2,  23,   3,  -1,  -1,   0,  -3,   4,  -5,  -3,   4,   0,   0,   1,   0,  -2,   1,   0,   1,  -1,   0 },
     {  -17, -20,  -2, -51,  25,   6,   5,   1, -11,  19,  43,  31, -46,  -3,  -3,   0,   0,  18,  -6,  31,  32, -17,  -4,  -3, -10, -47,   3,  -8,  12,  22,   0,   2,   8,   2, -29,   2,  -9,   1,   6,   1,   1,   6,   3, -12,   2,  -3,  -1,   1,   2,   0,   0,   4,  -4,   0,   0,   0,   0,   1,   0,  -1,   1,  -1,   0,   0 },
     {    7,  10,  44,  68,  -5, -20,  -9,  -5,  -3,  15,  42, -19,  13,   7,  -3,  -1, -16,   7, -52,  -9,   4,  12,   9,   3,  -9, -50,  -2,  -1,   1,   0,   4,   4,   1,   5, -17,  -5,  -3,  -1,  -1,   0,   3,   7,   3,  -7,  -2,  -3,   0,  -1,   3,   0,   2,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0 },
     {    3, -24,  -3,  -2,   4,   6,   0,   1,  25, -14,  33,  10,  -1,  -8,   1,  -2,   3, -44,   2, -14,   3,  17,   0,   4,   1,  15,   5,  41, -33, -11,   0,   0, -16,  -1, -40,  19,  51,  -9,  -2,  -1,   1,   2,  -4, -44,  -2,  24,   1,   1,  -1,   3,   4,   6, -20,  -5,   6,   1,   0,   1,  -1,  -1,   3,  -7,   1,   2 },
     {  -13, -17, -33,  11,   1,   8,   2,   3, -18, -20, -24,   9,  15,  -1,   4,   2, -36, -29,  -2, -13,  11,  11,   2,   2, -52,  -7,   5, -16, -13,   0,   6,   0,  73,   6, -19,  -9, -11, -12,  -4,   0,  -5,  26,  13,  -3,  -3,  -3,  -4,  -1,   2,  -4,   7,   5,   0,   0,   0,  -1,  -1,   1,  -1,   1,   1,   0,   1,  -1 },
     {   -2,  -8,  37,  -8,  13, -15,   1,  -2, -22, -24,  -8, -44, -10,  29,   3,   4, -28, -17,   8,  43, -30, -26,   5,   1,   8,  12,  37,  10,  24, -11, -18,  -1,  21, -16, -19,  27,   1,  18,   2,  -2,  -4,  11, -22,  -8,  13,  -2,   6,   1,   1,  -2,   4,  -8,  -6,   3,   0,   2,   0,   1,  -2,   2,   0,  -3,   0,   0 },
     {    7,  13,  -5,  11,  81, -35,  -9,  -3,   6,   5, -10,  21, -41,  43,  -2,  -2,   6, -15,   4, -43, -16,  -8,  23,   3,  -9,   4, -19,   0,  -1, -15,  -8,   7,  -1,  -1,   8,  -1,  -4,   4,  -6,  -5,   1,  -1,   6,   3,  -1,  -3,   2,  -1,   0,   2,  -1,   3,   0,   0,   0,  -1,   0,  -1,   1,  -1,   2,   0,   0,   0 },
     {   14,  -8,  17,   1,  -3,  -6,  -4,   0,  17, -28,  24, -15,  -1,   0,   4,  -1,  33, -17,  41, -12,   0,   1,  -6,   0,   2, -23,  -6,   2, -33,  20,   0,   2,  11,  32,  -3,  41, -35, -26,   3,   0, -13,   5,  -6,  30,  40, -17,  -8,  -3,  -2,  -3,  -6, -12,  17,  13,  -3,  -2,  -1,   1,  -2,   0,  -3,   5,   0,  -1 },
     {   -6,  -6,  20,  18,  -1,  -1,  -6,  -1, -11,  -6,  33,  33,   4, -15,  -1,  -4, -23,  23,  42, -16,   1,  15,  -5,  -1,  -7,  20, -35,  -3,  12,   6,  11,  -1,  10, -72,   9,  14,  -4,   3,   2,   9,  -7,  16, -36,   1,   4,  -1,  -1,   0,   3,   3,   6, -15,  -1,  -1,   1,  -1,   2,  -1,   0,   4,  -3,  -1,  -1,   0 },
     {    8,  14,  -8,  48,   3,  -6,  -8,   0,  -4, -16, -41,   2, -10,  11,  12,   3,  13,   1,  15,  25,  39, -23, -13,  -4,  -2,   5, -15, -23,   5,  46,   3,  -2, -21,  -2, -18,   4,  -1, -16,  22,   4,  -2,  -7,   0, -38,  16,  12,  -8,  10,   1,   2,   3,  -4, -33,   4,   6,   0,   0,   2,   2,   4,   3, -13,   2,   0 },
     {   -2,   5,  10, -14,   7,   3,   1,  -5, -15,   1,  -3, -11,  -5,   1,   2,   4,  -5,  27,  -3,  15, -11,  10,   0,  -1, -22, -17, -27,  -1,  -6, -23,  16,   3,  23,  21,  40,  10,  30, -26, -28,  -4,  -7, -26, -10,  -3,  32,  50,  -1,  -7,   0,   0,  -9, -20, -34,   8,  22,   8,   1,   1,   4,   4,  -4, -17,   0,   4 },
     {   -8, -14, -14, -16,  -3,   4,   3,  10, -12, -26, -20, -13,   1,   8,   6,  -1, -13, -26, -12,  -5,   5,   6,   3,   2, -32, -45,  -2,  14,  -3,  -1,   6,   5, -57, -20,  27,   3, -14,  -8,  -1,   4,  53,  10, -36,   3,   0,  -8,  -6,  -3,   4,  19,   2, -13,   2,  -2,  -1,  -2,   2,  -1,   6,   2,  -3,  -2,  -1,   0 },
     {   -6, -10, -19, -12, -27, -33,  19,  13,   8,   9,   1,  34,  51,  38, -39, -16,  11,   1,   8,  17, -18,   7,  41,  -1,  -1, -12, -13, -28,   5, -12, -10,  20, -11,   2, -16,  18,  -7,  10,  -1, -13,  -7,   0,  -1, -17,  14,  -5,   5,   4,   1,  -1,   2,  -6, -14,   8,  -2,   1,   1,   2,   1,   4,  -2,  -8,   3,   0 },
     {    6,   7,   4,   7,  39,  71, -11, -17,   5,   4,  -1,  22,  34, -25,  17,   8,   1,  -9,  -4,   4, -57, -21,   3,  16,  -2, -18,  -2, -34,   6,   4, -11,  -4, -15,   2, -13,  15,  -2,   0,   2,  -5,   4,   2,   1,  -6,   4,  -1,   3,   0,   1,   0,   1,  -4,  -5,   1,   0,   2,  -1,   2,   0,   2,  -2,  -2,   0,  -1 },
     {    2,   2,  15, -36,  40, -13,   7,  -7,   0,   2,   5, -48,  35,   9,  -4,   2,   3,  11,  11,   0,  -7,  35,  -4, -10,  11,   4,   5, -31, -26,  19,  34,   5,  -3, -23, -12, -14, -14, -21,   6,  20,   8,   0,  16,  -8, -25,   4, -17,   2,  -1,   4,   1,  23, -19,  -5,   3,  -3,  -1,   1,  -1,   2,   4,  -8,   4,  -1 },
     {   11,  -7,  -5,   4,   6,   0,  -8,  -5,  19, -14,  -2,   0,  13,  -3,   9,   4,  29, -26,  13, -13,  32,   2,  -2,  -1,  31, -38,  23, -17,  17,   2,  -6,  13,  22, -12,  16,  -7,  -9,  41, -28,   1, -16,   5, -22,  10, -28,  33,  25,  -7,  -8,   7, -10,  -5, -21, -18,  16,  11,  -2,   4,   1,   6,  -4, -12,  -4,   0 },
     {    6,   5,  -8, -18, -12,  12,  -2,   2,   9,   6, -26, -17, -17,   2,  22,   3,  18,   4, -28,  -6,  -5,  10,   5,  -1,  34, -13, -39,  11,  10, -19,   8,   7,   3, -18,  -5,   9, -16, -11, -16,   4, -30,  59,   6, -31,  15,  -3, -10, -13,  15, -14,  30,  -5,   2,  -2, -10,  -2,  -4,   6,  -5,  14,  -1,   8,  -8,  -5 },
     {    3,   4,  -4,  28,  25,  -9,  -6, -10,  -2,  -3, -19,  30,  -6, -14, -22,  14,   9,  -1,  10,  53,   1,  34,  -7, -23,   8,  -1,  22,  -2,  -4, -26,  42,   7, -13,   8, -23,  -1,   6,  -8, -33,  16,   7,   7, -12,  12,   5,  -4,   5, -23,  -2,  -1,   7,  -8,  30,  -6,  -5,   5,   0,   0,   0,  -5,   0,  21,  -7,  -2 },
     {    0,  -2, -13,  -8,  31,  -8,   3, -11,   5,   7,  -7,  13,  26, -12,  -5,  11,  -1,  -9, -25,  -4,   2,  35,  -7,  -7,  12,   2,  13,  45,  29,  17,   7, -14,   9,   9,   5, -14, -15,  13,  37,  14, -16,  -4, -14,  -3,  27, -11,  -5,  14,   3, -10, -11, -32, -10,  36,   3,   1,  -5,   3,   1,   3, -24, -20,  19,   2 },
     {   -5,  -7,  -3,   4,   2,   7,   3,  -2,  -9, -12,  -8,  -3,  -2,  -1,  -1,  10, -12, -15,  -5,  11,   4,   2,  14,  -2, -13, -17,  -2,  10,  -6,   1,   2,  10, -25, -33,   9,  16,   0,  -1,  -1,   9, -64, -26,  29,   5,  -4,   1,   0,   1,  44, -14, -55,  16,  16,  -8,  -6,   2,  -3,  18,   3, -21,   3,   4,   6,  -6 },
     {    8,   8,  11,  15,  13,  36,  84,  10,   6,   4,  11,   8,  17,  40, -10,  12,   2,  -3,   3,  -1,  22, -36, -27,   0,   5,  -1,   4,  14, -16, -19,  10,   0,   4, -11,  10, -20,  -3,  -6,  -6,   6,  -6,   4,  -6,  -1,   5,  -2,  -8, -10,   0,  -6,   5,  -3,   3,  10,  -4,   2,  -5,  -5,   4,  -2,  -4,   2,   5,  -1 },
     {    2,   1,   3,   2, -12,  -1,  44, -23,  -4, -11,  -6, -20, -32, -41, -51,  39,  -1, -11,   2, -35, -12,   1,   2, -31,  -6,  -2, -13, -21,  27,   2,   9,   2,  -3,   8,  -7,  23,  -1,  15,  -6,   1,  12,  -6,  10, -11,   0, -11,  16,   2,   3,   5,   3,   7, -17,   8,  -8,   8,  -2,   2,   0,   4,   1,  -8,   9,  -1 },
     {    0,  -8,  -2,   3, -15, -10,  -8,   7,   6,  -4,   6,   8,  -7, -11,   0,  -9,   2, -17,   2, -10, -10, -21,   8,  -7,  13,  -2,  20, -10,  27,  -3,  -7,   2,  -5, -31, -12, -22,   3, -22, -27,  15,  19,   8,  32,  18,  23,  20, -48, -26,  -9,   1, -14,  12,  -2,  49,  27,  -9,   1,   6,  -7,  -6, -11,  -9,  13,  10 },
     {    3,   4,  -4,  12, -25,  24,   7,  13,   3,   4,  -7,  16, -41,  17,  -6,  -4,   5,  -4, -11,   7, -37,  11,  15,  -2,  14,  -3,   4,   9, -10,  -1,  24,   7,  12,  -1, -12,   1, -20, -14,  27,  24, -10,   0, -22,  22, -20,  -6,  -4,   8, -17,  10, -10,   7, -12, -26,  36,  -9,   0,  -3,   6,  -7,  19, -47,  15,  10 },
     {    0,  -6, -11,  -3,  -4, -38,  12, -17,   7,   1,  -8,  17,  12,   4, -10,  -4,   6, -12, -13,  -1, -25, -17, -43,  15,  10, -25,  -9,  -5,  13,   6,  12, -60,  13,  -1,  12,  26,  22,  -2,   8,  25,  -5,  -2,  -8,   5,  -9,  10, -26,   9,   8,  13,   4,  14,   4, -17,   0, -19,   1,   9,  -4,  -5,  18,  20, -19,  -1 },
   }
 },
 {  //27
   {
     {  114, -30, -14,  -3,  -4,  -1,  -2,   0, -23,  35,  -3,  -2,  -1,  -1,   0,  -1, -20,   1,  12,  -1,   1,   0,   0,   0,  -1,  -4,   1,   3,  -1,   0,   0,   0,  -4,   0,   0,   1,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -10, -58,  15,   3,   4,   2,   2,   1,  77,  12, -46,   0,  -5,   0,  -2,   0, -23,  53,  12, -16,   0,  -4,   0,  -2,  -6, -12,  21,   5,  -3,   1,  -1,   0,  -3,   2,  -4,   6,   1,   0,   0,   0,  -2,  -2,   2,  -1,   2,   0,   0,   0,  -1,   1,   0,   0,   0,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   1,   0 },
     {  -34,  13,  -5,   6,  -2,   2,   0,   1,  -6,  75,  -1, -22,   0,  -6,   0,  -3, -15, -28,  69,   1,  -5,   0,  -1,   0,  11, -35, -14,  34,   0,   2,   0,   1,  -2,   4, -14,  -3,  10,  -1,   2,   0,   3,  -6,  -1,  -3,   0,   2,   0,   1,   0,   0,  -1,  -1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {    8,  56,  -4, -15,  -4,  -3,  -1,  -1, -24,  -5, -15,   7,  -1,   1,   1,   1,  10,  38,  11, -46,   0,  -7,   1,  -2,   7, -17,  64,  15, -28,   1,  -6,   1,  -1,  -8, -17,  43,   5, -10,   1,  -2,  -2,  -1,  -2,  -6,  19,  -1,  -2,   0,   1,  -4,  -2,   0,   0,   7,  -1,   0,  -1,   0,   0,   0,   0,   0,   1,  -1 },
     {   29,  76, -17, -12,  -5,  -5,  -2,  -2,  80,  -3,  26,  -6,  -8,  -3,  -3,  -1, -32,  -4,   1,  24,   3,   2,   0,   0, -15,  -3,  -7,  -3,   9,   1,   2,   0,  -4,  -5,   3,  -1,   1,   2,   0,   0,  -3,  -2,  -2,   0,   1,   2,   1,   0,  -2,  -1,   0,   0,  -1,   1,   0,   0,  -1,  -1,   0,   0,   0,   0,   1,   0 },
     {  -19,  -8, -32,   7,   6,   1,   3,   1,  -1,  42,   7, -17,  -8,   0,  -2,   0, -24, -12, -21,   2,  16,  -2,   5,  -1,  17,   6,  18, -63, -15,   7,  -2,   2,   8,   1,  34,  38, -42, -12,  -3,  -3,   0,   2,  -9,  19,  27, -16,  -2,  -2,   1,  -3,   5,  -4,   6,  10,  -6,   0,   0,   0,  -2,   2,   0,   2,   4,  -2 },
     {   -7,  35, -24,  10,  -1,   1,  -1,   0, -24,  25, -47,   3,  17,   0,   2,   0, -11,  36,  -8, -15,   9,  -3,   2,  -2, -13,  -5,   5, -27,  37,  -2,   4,  -2,  12, -22,  25, -53,  -7,  25,  -3,   5,   4,  -4,  12,  11, -38,   6,   1,   0,   0,   1,  -1,   7,  -3,  -8,   4,  -2,   1,  -1,   2,   1,  -1,  -2,   1,   1 },
     {   -9,   1,  31, -12,  10,  -2,   0,  -1, -16,   0,   2,  46,  -9,   3,  -1,   0, -72, -14, -26,  -2,  38,   2,   6,   1,  27, -27,  -1, -16,  -9,   7,   2,   1,   3,   8, -15,   0,  21,  -1,  -2,   1,   4,  -5,   2, -26, -10,  26,   3,   1,  -1,   3,  -3,   5, -20, -11,  13,   1,   1,  -1,   1,  -2,   3,  -8,  -5,   4 },
     {   -7, -10, -43,   5,  11,   3,   0,   2,  13,  28,   8,  -8,  -8,  -3,   2,  -3,  28,  -5, -14,  -5,  11,   2,  -3,   2,  -7,  42,  14,   3, -16,  -7,   3,  -5,  -5, -16,  18,  12,  53,  -4,  -7,   0,  -3, -13, -10, -28, -10,  52,   5,   2,  -4,   0,  -3,   3, -29, -18,  23,   3,   1,  -1,   0,  -2,   5, -11,  -9,   7 },
     {   -1,   3,  -3,  15,   0,  -4,  -2,  -3,  -7,   5, -11,  10,  10,  -3,  -9,  -1,  -2,  11,  -7,  17,   4, -11,  -2,  -5,  -4,   7, -11,   3,  23,  -7,  -9,  -2,   2,   8,   5,  -6,  24, -15,  -3,  -8,  11, -10,  21, -18,  48,  17, -35,   3,  -2,   4,  -3,   2, -16,  72,  -6, -20,  -1,   0,  -2,   0, -10,  -1,  45, -17 },
     {   11,  17, -12,  -2,  -4,   4,  -2,   1,  22,  -9, -53,  11,  12,  -1,   4,  -1,  14, -45,   4, -48,  12,   9,   1,   4,  20,  18, -30,  -4, -40,   7,   0,   1, -13,  42,  13, -14,  -5, -23,   3,  -4,   5,  -8,  28,   5,  -3,  -4,  -8,   1,  -2,   3,   0,  10,   2,   0,  -3,  -2,   0,  -1,   3,   1,   2,   1,   0,  -2 },
     {    8,  14,  75, -25,  -5,  -4,  -4,  -1,  11,  55,  -6,  32, -16,  -5,  -3,  -3,  51, -10, -12,   7,  23,  -8,  -1,  -1, -20,  14,   3, -11,   7,  13,  -2,   0, -13, -13,  12,   3,  -8,   3,   6,   0,  -1,  -9,  -6,   6,   1,  -4,   0,   2,  -2,   0,  -3,  -1,   3,   2,  -1,  -1,   0,  -2,   0,  -1,  -1,   1,   1,  -2 },
     {  -11, -16, -24, -22, -14,   3,   5,   2,  14,  26,  30,  31,  18,  -7,  -9,  -3, -15, -23, -31, -31, -14,  10,  10,   5,  14,  25,  28,  24,   6, -18, -14,  -6, -15, -22, -22, -15,   6,  24,  15,   4,   9,  11,  10,  -1, -20, -33, -12,   1,   1,   0,   2,  10,  23,  20, -11,  -8,  -1,   0,  -1,  -4,  -8,   3,  12,  -3 },
     {    0,  -6, -36,   4,  17,  -5,   5,   1,   6,   5,   8,  -4, -13,  12,  -3,   1,  35,   5, -13,   6,  26,  -9,   5,  -2, -17, -32,  -2, -15,   4,  41,  -1,   5, -16,   7, -51, -11, -23,  -6,  28,   1,   6,   3,  17, -44,  -9, -22,  -7,   9,   3,   3,  10,  15, -25,  -4, -12,  -4,   0,   3,   2,   4,   8,  -8,  -3,  -4 },
     {  -10,  15, -18,  20, -11,   4,   0,   2, -17,  25, -28,  32, -24,   1,   4,   0, -19,  27, -29,  27, -15,  -3,   4,  -1, -22,  28, -28,  23,  -7, -12,  13,  -5, -10,  16, -14,   7,  12, -23,  15,  -6,  -3,   2,   2, -12,  25, -30,  17,  -3,   6, -10,  13, -22,  25, -21,  -3,  12,   2,  -3,   4,  -5,   3,   5, -17,   8 },
     {   -1,  -7, -12,  32,   1,   2,  -3,   0,   1, -21, -22,  29,  31, -10,   1,  -3, -12, -39,  20,  -8,  29,  17,  -8,   4, -56,  -2,   6,   3,  -3,  24,   6,  -1,   1, -42,  -5,  13,  -4,   6,  12,   2,  -1,   2, -40,  -1,  12,  -2,   2,   5,   1,   3,  -4, -30,   5,   7,  -1,   0,   1,   2,   1,  -3, -15,   3,   4,   1 },
     {    7,   4, -14,  19,  -8,   1,  -3,  -2,  12,  -1, -18,  17,  17,  -5,  -3,   1,  25, -11,  -2,  11,  10,  -8,   0,  -6,  46,  -9,   8,  -6,  38,  -4, -19,   3, -17,  33, -13,  19,   1,  56, -10, -12, -12,   1,  -6,  -7,  14,   5,  44,  -9,   1, -14,   5, -23,  -1,   5,  11,  20,  -3,   1,  -5,   0, -13,  -3,   3,  12 },
     {    1,   2,   1,   5,  12,   9,  -7,  -5,  -2,  -3,  -4,  -5, -12, -21,   2,   2,   3,   3,   3,   7,  20,  14, -12,  -7,   2,  -3,  -2,  -9, -13, -21,   9,   5,   2,   3,   5,  10,  25,  16, -23,  -8,  -5, -11, -10, -23, -23, -14,  23,   8,   6,  10,  13,  12,  11, -10, -67,  -9,   2,   6,   3,   9,   8,  44,  28, -43 },
     {   -7,  -8, -18, -72,  -7,   6,   8,   4, -10, -14, -50, -20, -31,   5,   5,   9, -11, -42,  -9,  21, -13, -25,   5,   2, -34, -10,  15,   4,  14, -13, -20,   4,  -3,  -1,   0,  10,   3,  11,  -9, -10,   9,   3,  -2,  -3,   3,   3,   5,  -3,   4,   6,   0,  -1,  -5,   2,  -2,   0,   1,   3,   1,   1,  -2,  -3,   2,  -1 },
     {   -9, -14, -41, -56, -10,  10,   2,   4,  -5, -15,  14,  48, -15, -23,  -2,  -3,  11,  32,  45,  17,  37,  -1, -18,  -6,  22,   6, -15,   0,   0,  19,  12,  -1,  -6,  -2,  15,  -7,  -3,  -5,   2,   7,  -1,  -2,  -1,  18,   3,   3,  -6,  -2,  -3,   1,  -3,  -5,  13,   4,   4,   1,  -1,  -4,   0,  -1,  -4,   7,   1,   2 },
     {   -2,  -9, -15,  -7,   6,   5,   0,  -2,   5,  15,  28,  21,   7, -13,  -4,  -1,  18,   1, -24, -21,   4,  14,  -1,   0, -55, -50,  -3,   0, -13, -20,   2,   6,  38,  38,  -4,  -6,   5,   0, -30, -11,  -8,   3,  25,  24,   8,  14,  14, -12,   3,  -8, -12,  -5,  13,   5,   8,  10,  -2,   3,   0,  -5,  -8,   6,   1,   7 },
     {   -3,   4,  -5,   4,  -3,  15,  -5,  -7,  -5,   6,  -6,   8, -16,  22,  -7,  -8,  -4,   9,  -5,  10, -15,  26,  -7,  -8,  -7,   8, -10,  14, -22,  23,   0, -10,  -6,  10,  -9,  16, -22,  22,   5, -20,  -7,  17, -13,  19, -21,  11,  21, -22,  -8,  14,  -9,  14,  -6,  -2,  32, -33,   5,  -5,   9,  -9,  16, -24,  58, -29 },
     {    3,  -4,  -9,  47, -48,  31, -13,   6,   8, -12,   5,  33, -61,  26,   0,   3,  13, -21,  16,  -4,   1, -21,  22,  -5,   4, -20,  21,  -5,  10, -22,   6,   6,   9,  -9,  14,  -6,   5,  -8,  -2,   7,   4,  -1,   3,  10,  -5,  -2,  -8,   8,  -8,   6, -10,  17,  -2,  -1,  -4,  -1,  -1,   2,  -4,  -1,   5,  -7,   0,  -3 },
     {   -3,  -3,  10,   3, -82,  34,   3,   5,  -4,   1,  -4, -30,  25, -50,  12,   4,  -3,   3, -21,  23,  12,  16, -28,   1,  -4, -11,  18,  -3,  -9,  14,  11, -12,  -9,  11,  -4,   2,   4,  -7,  10,   5,   4,   5,  11,  -6,  -4,   5,  -4,   3,   4,  -4,   1,  11,  -8,  -3,   4,   1,  -1,   1,  -5,   2,   4,  -7,  -1,   3 },
     {    8,   4, -10, -13,   7,   3,  -4,   2,  14,  11, -29,   6,  11,   2,   3,   4,  30,  -7, -33,  21, -18,  17,   3,   1,  45, -42, -11,   5,  -7, -12,  28,  -6,  33, -28,  -9,  -5,   6, -21,  -2,  19, -14,  34, -33,  -1,   2,   5, -19,   5, -11,  14,   4, -18,   2,   2,  -1,  -8,  -1,  -2,   8,  -3,  -2,   1,   3,  -2 },
     {   -2,  -1,  -8,   8,  12,  -2,   1,   1,  -4,  -5,   6, -12,  -3,  -1,   2,   1,   3,   7, -32,  -1,   4,   3,   3,   0,   5, -47,  -5,  18,  -3,  -2,   0,   1, -83,  -4,  28,  -4,   7,  -4,   0,   1,  22, -35, -15,  23,  -1,   0,  -5,   2,   1,  17, -30, -16,  14,  -2,   1,  -4,   1,  -1,   5,  -9,  -6,   3,  -1,  -2 },
     {   -3,  -3,  -5,  -3, -10,  -6,  -7,   5,   2,   5,   4,  12,  13,  12,   4, -11,  -3,  -4,  -5,  -6, -13, -12, -13,   6,   2,   3,   3,   5,  13,  11,  -3, -15,  -7,  -4,  -6,  -6, -13, -11, -16,  10,   3,   8,   7,   8,  14,  16,   0, -27,  -5,  -9, -12, -13, -21, -13, -12,  26,  -1,   1,  -2,   2,  -4,  -3, -35, -95 },
     {    4,   4,   4,  20,  39,  11, -13,  -6,   0,  -2,  -8, -17, -39, -13,  16,   0,  -3, -17,  -6,   2,  16,  -7, -41,   3,   6,  12,  27,   5,  10,  15,  18, -18, -12,  -2, -25, -16,   1,  -9, -23,  25,   5,  23,  14,  10,  -3,  10,  -9, -36,   9, -12,   8,   9,  37,  12,  19,   8,   0,   3,  -8,  -8, -13,  19,   8,  10 },
     {   -2,  -2,  -2,  -2,  26,  -5,  -7,  -4,  -2,  -1, -13,  30,  28,   7, -17,   6,   2, -16,  20,  49, -36,  32,   5, -12,  -7,  -2,  45, -17,  -3,  -8,  35,  -7, -19,  15,   5,   3,   0,  -7,   3,  16, -12, -24,  23,   1,  -4,  -5,  -4,   4,   5, -16, -16,  27,  -5,  -5,  -2,   4,  -3,   7, -13,  -2,  13,  -6,  -4,   2 },
     {   -2,   2,   5,  -4, -16,   5,   2,  -6,  -1,   3,   2,   6,  17,  11,  -3,  11,   5,  10,  19,   1,   1,  12,  20, -12, -12,   3, -18, -30, -17, -36,  13,  29, -43, -19, -21,   6,  -9,  -3, -44,   4,  37,  32,   1, -13,  -3,   6,  -2, -29,  -2,  11,  33,   4,  -7,   8,  14,   8,  -6, -15,  -2,   8,  -7,   1,  -2,   7 },
     {   -1,  -1,  -8,  20, -11, -47,   1,  -1,   2,   0,  -8,  16,  22, -13, -22,  -2,   4,  -5,   0,  21,   8, -35, -11,  -1,  -3,  -6,   4,   0,   1,  -1, -38,  -5,  -3,   6,  -4,  17,   5,  -4, -15, -13,  -1,  24,  -1,  13, -10,  13, -29,  -3,  -1,  17,   2,  25,  27, -25,   5, -34,   2,  -4,  11, -10,  38,  22, -19,   3 },
     {    3,   5,   3,   3,  30,  61,   4,  -9,   2,   3,   9,  23,  19, -10,  28,  16,   0,   5,   9,   2, -36, -49, -25,  22,  -5, -10, -13, -30, -20,  -8, -24, -26, -13,  -4,  -6,   4,   7,   5,   8, -11,  -4,   2,  -4,   5,  -9,  -2,  -9,  12, -15, -10,   7,  -5, -11,  -3,  -3, -11,  -7,  -7,  -6,   2,  -2,  -4,   1,   3 },
   },
   {
     {  115, -30, -11,  -4,  -3,  -2,  -1,   0, -16,  25,  -4,   0,  -1,   0,   0,   0, -32,   8,   6,   0,   1,   0,   0,   0,  -3,  -4,   2,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -3,  16,  -6,   0,  -1,   0,   0,   0,-105,  14,  20,   2,   5,   1,   2,   0,  21, -49,   5,   2,   2,   1,   1,   0,  34,  -5, -16,   1,  -2,   0,  -1,   0,   3,   8,  -4,  -2,   0,   0,   0,   0,   5,   1,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {   31,  -9,  -6,  -1,  -1,   0,   0,   0,   1, -34,   9,   2,   2,   1,   1,   0,  80,  -5, -36,   2,  -5,   1,  -2,   0, -20,  67,  -6, -11,  -1,  -3,   0,  -1, -23,   0,  25,  -2,   1,  -1,   1,   0,  -2,  -6,   1,   4,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   -1, -42,  12,   3,   3,   1,   1,   0,  40,  -4,  -8,  -2,  -2,  -1,  -1,   0,  -6, -45,   7,  10,   1,   3,   0,   1,  50,   7, -54,   3,  -3,   1,  -1,   0, -16,  62,  -4, -20,   0,  -4,   0,  -1,  -7,   0,  23,  -4,  -2,  -1,   0,   0,  -3,  -1,   4,   4,  -2,   0,   0,   0,  -1,  -1,   0,   2,   0,   0,   0,   0 },
     {   26,  92, -42,  -2,  -9,  -2,  -3,  -1,  35,  -8,  24, -12,   1,  -3,   0,  -1, -14, -44,  17,  11,   0,   3,   0,   1, -17,   5, -19,   7,   1,   1,   1,   0,  -1,   4,   0,  -7,   2,  -1,   1,   0,   0,   0,   4,  -1,  -1,   0,   0,   0,  -1,  -2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {   21,  -4,   8,  -5,   1,  -2,   0,   0,  -3, -69,  14,  11,   3,   3,   1,   1,  26,   6,  -7,  -1,  -2,   0,  -1,   0, -11, -45,   1,  22,   0,   5,   0,   1,  26,  14, -63,  -1,   3,   0,   0,   0, -11,  41,   6, -22,   0,  -3,   0,  -1,   0,  -1,   8,   1,  -2,  -1,   0,   0,  -2,   4,   1,   0,   0,  -1,   0,   0 },
     {   -1,  10,  -3,  -2,   1,  -1,   0,   0, -18, -64,  23,   9,   2,   4,   1,   1, -68,  22, -44,  11,   3,   3,   1,   0,  37,  27,  -4, -29,   2,  -4,   0,  -1,  16,   1,  25,  -2,  -7,   0,  -2,   0,  -5,   1,   0,   8,   1,  -1,   0,   0,   0,  -3,   0,  -1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {    2, -35,  14,   6,   2,   1,   1,   0,  17, -22,  25,  -2,   2,  -1,   0,   0,  -9, -55,  13,  19,   2,   4,   0,   1,  -1,  -4,  -5,   8,  -1,   1,   0,   0,   0, -43,   1,  30,   0,   5,  -1,   1,   7,  -1, -66,  17,  10,   1,   1,   0,  -9,  28, -20, -24,  11,  -3,   2,  -1,   0,   4,   4, -11,   1,   0,   0,   0 },
     {   -7,  -4, -17,   6,   0,   2,   0,   1,   6,  24,  -5, -10,  -2,  -1,  -1,  -1,  -9,  -8, -20,   4,   2,   2,   1,   1,  10,  25,  -1, -18,  -2,  -2,   0,  -1,  -9, -16, -19,  13,   3,   1,   1,   0,   3,  41, -26, -54,   7,  -3,   2,   0,  10,   7,  69, -22, -26,   1,  -3,   0,   5,  -9,  10,  33, -10,  -2,  -1,   0 },
     {  -14, -29,   8,   7,   0,   2,   1,   0, -15,   7,   6,  -6,   2,  -1,   0,   0, -23, -20,  -4,  20,  -2,   4,  -1,   1, -76,  18, -38,  -3,  13,   0,   4,   0,  50, -11,   7, -39,  -2,  -1,   0,   0,  20,  15,  21,   0, -16,   1,  -3,   0,  -2,   0,   8,  14,  -1,  -3,   0,  -1,   2,  -3,  -3,   3,   4,   1,   0,   0 },
     {   14,  33,  81, -37,   1, -10,  -1,  -2,  10,  27,  -2,  27, -12,   2,  -3,   0,   2, -22, -53,  21,  11,   2,   2,   0,  -9, -24,   2, -24,   9,   2,   2,   1, -12,   1,   1,  -1,  -9,   2,   0,   1,   3,   3,   0,   1,  -2,  -2,   0,   0,   1,   1,   2,   0,   0,   0,  -1,   0,   0,  -1,   1,   1,  -1,   0,   0,   0 },
     {    8,   4, -17,   4,  -3,   1,  -1,   0,  21,  25, -12,  -9,  -1,  -2,   0,  -1,  32,  -4, -10,  -3,   0,   0,   0,   0,  37,  -7,   7, -26,   2,  -2,   0,   0,  74,  -8,  20,  -3, -22,   4,  -5,   1, -23,  45,  -6,  33,  -6,  -5,  -1,  -2, -26,   3,   2,   3,  16,  -4,   2,   0,  -3,  -7,   2,  -6,   2,   5,  -1,   1 },
     {    7,  19,  15,  -3,  -6,  -2,  -1,  -1, -14, -18, -39,  -2,   8,   2,   2,   1,   8,  24,  21,  -5,  -8,  -2,  -1,  -1, -14, -14, -37,   1,  11,   1,   2,   1,  15,  24,  24, -18, -13,  -4,  -2,   0, -11, -18, -29,  14,  16,   0,   2,   0,  11,  26,   9, -58, -12,   2,  -1,   1,   6,  -5,  37,  22, -23,  -2,  -2,   0 },
     {    0,  -2,   4,   0,   2,  -1,   1,   0,   5,  23,  64, -30,  -7,  -3,  -1,  -1,  18,  50,  -7,  50, -19,  -1,  -4,  -1,  22, -19, -34,  20,  27,  -5,   3,  -1, -10, -20,   4, -22,  14,   9,   0,   1, -21,  -7,  -6,  -3,  -8,   6,   3,   1,   0,  -2,   1,  -4,  -5,  -3,   1,   0,   0,   0,   1,   3,   0,  -2,   0,   0 },
     {   -9,   9, -34,  21,  -3,   2,   0,   0,  -9,  24, -38,  34,  -5,   1,  -2,  -1, -10,  15, -45,  25,   2,   0,   0,   0, -10,  14, -29,  21,   8,  -4,   1,  -1, -13,  12, -34,   3,  18,  -3,   2,  -1, -12,  15, -19,   1,  16,  -6,   3,  -1, -15,  10, -28, -15,  30,  -6,   5,  -1,  -1,   3,   4, -38,   9,   1,   1,   0 },
     {    6,  11,   0, -11,  -1,  -1,  -1,   0,   3, -14, -22,  11,   1,   1,   0,   0,   4,  12,   5, -18,  -1,   0,   0,   0,  30,  -4, -12,  13,   5,  -3,   1,   0, -22, -33,   2, -40,  13,   7,   3,   1,  63,  17,   0,  -1, -40,   4,  -5,   0, -28,  47,   7,  19,   6, -13,   2,  -3,   2, -10,  17,  -3,   6,   2,  -1,   0 },
     {    7,  17,  48, -14,  -6,  -4,  -2,  -3,  -1,   1, -16,  20,   0,  -1,   0,  -2,  -7,   3,  36,  -9,  -7,  -2,  -2,  -1,  21,  59, -11,  34,  -6,  -8,  -3,  -3,  31, -33, -16,   2,  32,   3,   3,   1, -41,  -4,   5, -17,   1,  11,   3,   2, -12, -14,  -3,   4,  -4,   2,   3,   0,   6,  -3,  -6,  -3,  -2,   0,   2,   0 },
     {    7,  10,  31,  69, -26,  -3,  -6,  -1,   5,   9,  18, -16,  30,  -8,   1,  -1,  -3,   4, -32, -58,  -1,  23,   4,   4,  -3, -16, -26,  13, -30,  -3,   7,   1,  -4,  -6,  14,   5,  15, -17,  -5,   2,  -1,  10,  -3,   1,   2,   4,  -4,  -1,   2,  -4,  -2,   2,  -1,   1,   2,  -1,   5,   1,  -3,  -3,   2,  -2,  -1,   0 },
     {   -2,  -9, -20,   4,  -2,   5,   0,   2,  -4, -17, -36,  10,  13,   3,   4,   2,  -2, -23,  -8,  -3,  13,   5,  -1,   2,  -2, -31, -10, -15,  16,   9,   3,   0, -33, -49,  13, -28,  -1,  10,   0,   2, -68,  -1,   4,   2, -29,   5,   1,   1,  17, -18,   0,  11, -12, -12,   3,  -1,  17,   8,  -7,   7,   2,  -6,  -5,   0 },
     {   -9,   1,  16,  -1,   0,  -5,   1,   0, -14,  -2,  22,  -2,   0,  -8,   4,  -3, -16,   5,  23,  -1,  -3,  -4,  -4,   2, -25,  10,  16,   9,  -9,   3,  -1,   1, -45,  23,   9,   1, -15,   1,   3,  -1, -43,  39,   0,  27,  -9, -12,   1,  -1, -44,  10,  26,   5,  27,  -8,  -1,  -1,   6, -44,  17,   1,   9,  10,  -1,   0 },
     {   -4,  -5,  -6, -48,  -7,   6,   4,   0,  -8, -10, -34, -60, -11,   4,   8,   4,  -8, -16, -34,   7, -34,   4,  -3,   7,  -3, -15,   7,  46, -19, -24,  -2,  -3,   0,  -2,  28,  14,  24, -17,  -7,  -5,  -2,  12,   8,   2,  15,   9,  -9,  -1,   1,  -3,   3,   7,   4,   5,   1,  -4,   9,  -3,   2,   0,   3,   1,   4,   1 },
     {   -5, -11,  -8, -18,   5,   4,   1,   0,   5,   8,  24, -14,   3,  -2,   2,   2,  -3,  -9, -10, -22,  25,   0,  -3,  -3,   1,   3,  17, -14,  -3,   0,   9,   1,  -4, -13, -17, -32,  13,  -5,   8,   3,  -2,   1,  29,  -8,  10,   4, -13,   2, -25,  -7, -22, -41,   9,  28,   4,  -3,  38,   2,  32, -23, -56,   5,   6,   1 },
     {   -7, -11, -27, -44,  -1,  10,   6,   3,   4,   6,  29,  57, -29, -13,  -5,  -1,   6,  12,   5, -24,  28,   4,   1,  -1,  -8, -19, -37,   3, -33,   4,   9,  -1,   6,  -5,  33,  14,  20, -21, -10,   0,  -5,  10,  -3,  -2,  18,  20,  -7,   1,   4,   2,   5,   4,  -9,  -3,   5,  -4,  -2, -13,  -3,   7,   7,   3,  -2,   2 },
     {    3,  -4,   0,  47, -38,  -1,  -4,   1,   2, -11,  -8,  36, -12,   5,  -9,   0,   6, -12,   7,  45, -10, -12,   2,  -2,   6, -18,  25,   2,  12, -11,   0,  -1,   0, -16,  18,   1,  15, -25, -10,   3,   5, -12,  36,   0,  30,  13, -20,  -2, -21,   6,  21,   2,  -4,  32,   0,  -6,   2, -13,  16,  12,   2,   0,   7,  -5 },
     {    2,   2,   8,  -6,  -9,   2,  -3,   1,  -1,  -9, -12,  10,  10,  -5,  -4,   1,   3,  11,  25,   7,  -7,   2,   3,  -2,   6,  -3, -21,   2,  18,   1,   2,  -2, -22, -11,  16,   1,  -8,  -1,  -4,   2,  40,  44, -15,  16,  17, -20,   1,  -2,   3, -75,   8, -10,  22,  20,  -1,   3,  18,   9, -31,  -6, -17,   8,   1,   3 },
     {   -3,  -5,   2, -12, -80,  30,   5,   1,   0,   1,  -1, -12,   7, -19,   4,  -5,   0,   4, -11,  15,  46, -26, -12,   6,   4,   8,  12,  27, -18,  35,  -5,  -8,   6,   7,   1, -21, -19,  12,  14,  -3,   3,  -6, -20,  -8,  -2,  -9,   2,  10,  -2,  -9,  -6,   6,   0,  -3, -11,  -2,  -8,   6,   2,  11,   3,   2,  -5,  -3 },
     {    2,   2,  -6,   0,  18,   9,   8,  -1,  -1,  -5,  -8,  -5,  10,   7,   7,  -1,   4,   3, -14,  -2,   7,   3,  -3,   5,  -5, -16, -16,   0,   1,   3,   5,  -1,  10,   4,  -5,  14, -12,   7,   2,   0,   7, -49, -15, -13,   8,   5,   6,   7, -87, -21,  21,   7, -18, -10,  -4,  -1,  32, -10, -31,   9,   2,  -1,  -4,   0 },
     {   -7, -12, -18, -28, -34,  -3,  12,   8,   8,  12,  19,  37,  30,   6, -12,  -6,  -6, -13, -17, -31, -51,  27,   7,   0,   1,   0,  12,  20,  28, -20,   2,  -6,   5,   5,  -6, -14, -12,  14,  -1,   6,  -8, -22,  -7,   4,  -9, -25,  -3,   2,   0,   5,  10,  -9,   6,  16, -17,  -2, -20, -13, -16,   9, -13,  -2,   8,  -8 },
     {    1,   3,   2, -11,   1,   8,   3,  -3,  -3,  -2,  -7, -18,  -1,  -3,  -9,   0,   2,  12,   4,   1,  -2,   0,  -2,   1,   4,   2, -39, -15, -17,  14,   2,   7,  -6, -40, -24,  22, -34, -38,  16,   0,  -9, -16,  31,   5,  13, -51, -17,   0,   4,  28,   9,   7,  21,  18, -33,  -9,  -7,   5,  -9,   1,   2,  14,   1, -10 },
     {    1,  -1,  -1,  22,   4, -29,   5,  -3,  -1,  -5,   6,  -4, -78,  15,   7,   3,  -3, -11, -15, -38,  22, -22, -10,  -3,   2,   0,  -2,  37,  29, -12, -10,   4,   2,  -9,   5, -10, -15,  19,  15,  -1,  -1,  -9,   6,  13,  13, -12,  12,   1,   7,  -2,  20,  -8,  16,  -3,  -5,   5,   0,   9, -14,  11,  -6,  12,  -2,   2 },
     {   -3,  -3,   1,  -7, -51,   1,   1, -14,  -4,  -6,  -1, -19, -38,  -4,   2,  18,   0,   1,   8, -20, -24, -23,  32,   7,  -7,  -3, -20, -50,  15,   1,   0,  -7,  -4,   2, -20,  19,  27,  18, -39,  -4,  -2, -11,  -4,   0, -13,  -7,   2,  -4, -11,  -9,  -1,   4,  -5,  -4,  -2,  -8,  -5,  -9,  13, -23,   3,   4,  -8,  22 },
     {    1,   0,   5,   1,   3,  -6, -12,  -5,  -1,  -5,  -4,  -7,   2,  -9,  43,   8,   2,  -2,  11,   0,  14,   5,  -8,  11,   1,  -4,  10,  -5,  18, -19,  50, -40,  -3,   0,  18, -15,  -7, -17, -25,  24,  -5,   2,   0, -44,  19,  -8,  -3,  -6,  -7,   8, -13,  -6,  13, -24, -14,   9, -38,  -9, -20, -11,  -6,  25,  -6, -11 },
   },
   {
     {  106, -60,   5,  -5,   0,  -1,   0,   0,  19,  -3,  -7,   3,  -2,   1,  -1,   0, -16,  19,  -8,   3,  -1,   0,   0,   0, -12,  10,  -1,  -1,   1,   0,   0,   0,  -7,   4,   1,  -1,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {  -41, -30,  25,   0,   5,   0,   2,   0,  80, -66,   2,   5,  -1,   1,  -1,   1,  19,   9, -25,   9,  -3,   1,  -1,   0, -14,  25, -13,  -1,   2,  -1,   0,   0,  -7,  10,   0,  -5,   2,  -1,   0,   0,  -4,   4,   2,  -2,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   13,  53, -58,  17, -10,   3,  -3,   1,  68,  30, -44,   0,  -3,  -2,  -1,  -1,   4,  22,   6, -16,   5,  -4,   1,  -1, -16,   8,  20, -11,   1,  -1,   0,   0, -11,  -2,  16,  -3,  -1,   0,   0,   0,  -4,  -6,   9,   1,  -1,   0,   0,   0,  -1,  -4,   4,   1,  -1,   0,   0,   0,  -1,  -2,   1,   1,  -1,   0,   0,   0 },
     {  -40, -28,  30,  -5,   6,   0,   2,   0,  -1,  33, -26,   8,  -5,   2,  -2,   0, -67,  67, -18,   1,   2,  -2,   1,  -1,  -7,  -9,  20, -14,   7,  -3,   2,  -1,  15, -21,  13,  -2,  -2,   2,  -1,   1,   8,  -9,   0,   3,  -2,   1,   0,   0,   3,  -1,  -2,   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {    7,  17,  32, -40,  10,  -6,   3,  -2,  47,   9,  45, -47,   6,  -6,   1,  -2, -24,  -3,  53, -14,  -7,   4,  -2,   0,  -5, -25,  22,  18, -16,   7,  -3,   1,  -2, -10,  -6,  24, -10,   3,  -1,   0,  -1,  -3, -11,  13,  -2,   0,   1,   0,  -1,   0,  -6,   5,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0 },
     {  -15, -54,   2,  25,  -4,   5,   0,   1,  27,  22, -41,  17,  -5,   1,  -2,   0,  19, -33,  28, -16,  11,  -4,   2,   0,  48, -55,  18,  -3,  -1,   4,  -3,   2,   6,   6, -17,  11,  -6,   4,  -2,   0,  -9,  14, -12,   2,   1,  -2,   1,  -1,  -6,   9,  -2,  -3,   3,  -2,   0,   0,  -2,   2,   0,  -2,   1,   0,   0,   0 },
     {    6,   9,  35, -48,  15,  -8,   4,  -2,   8,  52, -13, -21,  -4,   1,  -1,   0,  66,  23, -40,  11, -15,   3,  -3,   0,  39,   7,  -9,   0,  -4,  -2,   0,  -1,   5,   6,   9,  -4,  -1,  -1,   0,   0,  -9,   3,  11,   0,  -3,  -1,   0,   0,  -6,  -1,   7,   3,  -3,   0,   0,   0,  -2,  -2,   3,   2,  -1,   0,   0,   0 },
     {    5,  -7,  -7,  -7,  13,  -3,   1,   0, -23, -40,   4, -14,  23,  -3,   3,   0,  32,  12,  -9, -45,  20,  -3,   2,  -1,   3,   0,  51, -50,   0,   9,  -3,   2,  12, -22,  31,   0, -30,  18,  -6,   2,  -3,   5,  -4,  23, -24,   3,   2,  -1,  -1,   2,  -4,  11,  -4,  -6,   3,   0,  -1,   1,  -2,   4,   1,  -2,  -1,   1 },
     {  -25, -46,  10,  11,  -1,   3,   1,   1, -14,  45,  -2, -17,   5,  -5,   0,  -1,  27, -11,  36, -29,   5,  -3,   0,   0, -51,  49,   1,   0,  -9,   3,  -1,   0, -35,  22,  11,   2,  -3,  -3,   2,  -1,   6, -19,  14,   2,  -1,  -1,   1,   0,   8, -11,   3,   5,  -3,   1,  -1,   1,   3,  -3,  -1,   4,  -2,   1,   0,   0 },
     {  -10, -15, -35, -17,  38,  -7,   4,  -1,  10,   8, -13, -42,  43,  -9,   3,  -1, -46, -30, -13,  -7,  16,  11,  -2,   3,  25,   9, -49, -15,  -5,  14,  -2,   1,   7,  21,   6, -24, -13,   8,   0,  -1,   1,   5,  11,  -4, -16,   2,   2,  -1,  -2,   4,   9,   2,  -7,  -3,   3,   0,   1,  -1,   3,   2,  -2,  -2,   0,   1 },
     {   -7,  -9, -32, -10,  27,  -6,   4,   0, -22, -32, -30, -14,  23,   1,   3,   1,  11,  49,   7, -25, -21,   9,  -5,   0,   3,  -4,  11,  41, -39,   7,  -1,  -1,   3,  14, -33,  28,   7, -19,   8,  -4,  -2,   7,   6, -18,  26, -11,   0,   1,   3,  -4,   7, -10,   4,   5,  -5,   2,   0,   0,   2,  -1,  -2,   3,   0,  -1 },
     {    1,  10,  10, -50,  46, -20,   9,  -5,  -8, -11, -50,  30,  -2,  12,  -4,   3,  -7, -46,   0,  25, -15,  13,  -2,   1, -40, -19,  30,  -7,   4,  -3,   5,  -1, -29,   3,   8,  -4,   2,  -1,   3,   0,   2,  -6,   0,   0,   1,   0,   0,   1,   6,  -3,  -7,   1,   0,  -1,   0,   1,   2,   2,  -3,  -2,   1,  -1,   0,   0 },
     {    4,  24,  26,  -8, -11,  -2,   0,  -1,  -8, -14,   6,  17, -13,   4,  -1,   1, -24,  18,  12, -23,  13, -10,   5,  -2,  30,   0,  13, -28,   8,  -1,  -2,   1, -44,  73, -15,  -7,  -7,   3,  -3,   0, -26,  19,  24, -12,  -2,  -3,   2,  -1,   6, -18,  19,   3,  -7,   2,  -2,   1,   4,  -7,   2,   9,  -6,   2,  -1,   0 },
     {    6,   2,  11,  -4,  18, -23,   8,  -2,   4,  -4,   3,  -2,  31, -35,  11,  -2,  15,   6,  15, -12,  38, -34,   6,  -2,  -3, -17,   4,   2,  42, -24,   1,   1,   9, -16,  -9, -24,  47,  -8,  -7,   4,  11, -16,  14, -35,  15,  18, -14,   7,   6,  -5,   6,  -9, -13,  21,  -9,   3,   0,   1,  -1,   2, -11,   8,  -1,  -1 },
     {    2, -11,  -6,   0,  -7,   2,   2,   0,   3, -20,   2,   7, -23,  10,  -1,   2,  -3,  -7,  27,   0, -34,  11,  -3,   3,  38,  -4,   8,   6, -24,  -5,   4,  -3,  20,  -4,  43, -27,   7, -19,   7,  -2,  26, -41,  55,  -4,  -8,   3,  -7,   5,  12, -16,   4,  25, -17,  10,  -8,   3,   0,   0,  -8,  15,  -7,   2,   0,  -2 },
     {   10,  20,  36,  15, -11,   3,  -9,   0, -11, -20, -47, -20,  13,   4,   4,   1,  16,  25,  39,   5,  -6,  -6,  -6,  -2, -27, -30, -44,  10,  19,  14,   4,   1,  26,  24,   6, -31, -22,   3,  -3,   1,   1,  -4,   3,  30,  -9,   1,   0,  -2,   1,   8, -13,   0,   1,  -5,   6,  -1,   1,   1,   2,   2,   5,  -4,   0,   1 },
     {  -13, -26, -51,   5,  29,  -9,  12,  -3,  -4,   5,  36, -19, -17,  11, -10,   0,  18,  34,  24,  48, -29,   0,  -5,  -2,  -6, -27,   8, -12,  40,  -8,   0,  -1, -14,  15,   3, -19,   0,  14,  -9,   3, -15,  13,  -4,   6,  -5,   6,   2,  -3,   2,  -6,   6,  -1,  -1,  -1,   4,  -2,  -1,   1,  -3,   4,   1,  -2,  -1,   3 },
     {    3,   7,  14,  42,   9, -58,  16,  -2,   4,  11,  15,  35,  26, -57,  10,   3,   2,   7,  -2,  20, -14,  16, -26,  15,   2,  -2,   9,   9, -24,  35, -22,   3,  -7,   4,  16,  -6, -12,  13,   1,  -6,  -8,   4,   5,   1,  -4,  -2,   4,  -3,  -3,   0,  -2,   3,   1,  -2,   3,  -2,   2,  -5,   0,  -1,   1,   0,   1,   0 },
     {    3,   5,  -2, -12,  17, -26,   5,   1,   6,   6,  -2,  21, -25,  -6, -10,   5,   0,   0,  35, -12, -21,   1,  -4,  -3,   8,  51,   5, -33,   3,  -6,   2,  -2,  58,  25, -21, -10,   2,   0,  -3,   1,  38,  22, -33,  -7,   4,   1,  -5,  -2,   3,  19, -10,  -8,   4,   0,  -2,   1,   1,  -1,   5,  -4,   0,   1,  -2,  -1 },
     {    0,  -4,  -2,   1, -12,   3,   2,  -1,   6,  14,  16,  -7, -16,   5,   2,   0,   6, -14, -32,  -8,  -3,  -2,   2,   1, -55, -37,  13,  -7, -17,  19,  -6,   9,  47,  12, -31, -17, -17,   3,   7,   2,   4,  27,  29, -36, -14, -11,  -3,   5,   6,  -9,  21,  11, -22,   5,  -6,   1,   5,  -4,   2,   9, -10,   1,   2,  -4 },
     {   -6,  -3,   3,  -8, -10,  10,   0,   1,  -8,   1,  -2,  -4,   2,  -3,   9,  -2,  -8,  -3,  20,  -2,  -9,   4,   8,  -7,  -9,  12,  11,   4, -14,  -2,   9,   0,   4, -11,  36, -33,  24, -21,   5,   3, -55,  51,   8, -18,   8,  -2, -11,   8, -51,  42,  -9,   2,  -2,   4,  -6,   0,  -7,   3,   7,  -3,   3,   0,   4,  -7 },
     {    1,  -1,   0,   7,   4,  -6,   1,   0,   1,   0,   4, -13,  -2,  -8,  -5,   0,  -1,  -7,  -5,  24,  11, -19,   5,  -1,  -4, -10,  -8, -36,  -2,   1,  -1,   4,   8,  11,  29,  26, -36, -39,   4,  -5,   1,  -1,   2,  13,  62, -34, -22,   3,   4,   2,   3,  -9,  15,  37, -28,  -4,   2,   2,  -3,   5,  -5,  14,  10, -13 },
     {    4,   5,  15,  37,  -8,   4,  -6,  -7,   3,   7,   3, -30,  33,  -2,   7,   2,  -4, -10, -28, -10, -40,  33,  -5,  -1,   1, -13,  25, -15,  -1, -28,  35, -21, -13,  21,  -4, -17,  19, -28,   8,  16,  14,  10, -30,  11,  -3,  10, -26,  18,  21,  -9,  -9,  -3,   0,   2,  -2,  -4,   3,   0,   0,  -1,  -2,  -1,   2,   0 },
     {    3,   4,  17,  17,  -9, -13,  -8,   2,  -6, -13, -29, -32,  -9,   8,  -4,   4,  -3,  -4,  17,   9, -27,  -1,  -7,  10,  13,  18,  -6, -14,  20,  13, -16,   6, -32, -52, -10,   6,  -8,   5,   8,  -1,  23,  33,  -9, -23, -13, -16,   3,   9, -13, -14,  40,  -5, -19,  15,  -9,  -2,   6, -11,   3,  12, -24,   7,  -2,   1 },
     {    0,  -3,  -1,  -3,  -2,  -8,  18,   0,   2,  -1,   3,   0,  -5, -19,  29,   0,   1,  -1,   5,  -1,  -4, -11,  23,  -3,   1,  -4,  -5,  -6,  -2, -14,  28,   4,  -1, -12,  -1, -18,  -9, -14,  21,   9,  -2,  -3,   1,  -3,  -6, -49,  30,  10,   6, -12,   8, -14,  31, -49, -18,  32,   1,   0,  -1,  -1,   4,  11, -49,  23 },
     {    4,   6,   9,  10,  29,  12, -30,   2,   7,   9,  22,  28,  21,  27, -36,  -4,  11,  14,  16,  -3,  27,  48, -11, -14,   2,  -3,  -6, -13,  13,  32,  27, -19,  -1, -14,  -9, -12,   0,  -7,  45, -11,  -1, -11,   5, -13,   6, -19,  15,  10,  -6,   5,   2,  -2,  -2,   4,  -8,   8,  -2,   3,   1,  -3,  -8,   4,   3,   0 },
     {   -1,   1,  -8,  -4,   4,   3,  -2,  -1,  -1,  -1,  12,   3, -14,  -1,   7,   0,   7,  16, -11,  -4,  13,  -8,  14,   4,  -1, -26,  -1,   4, -17,  14,  -3,  15, -39,  21,   7, -18,  -6, -12,   1,   5,  71, -15, -20,  -5,  -9,   1, -13,   7, -52,  42,   3,  -6,  -8,  -9,  -6,  -6,   2,  -5,   8,  -6,   0,  -2,  -5,  -6 },
     {   -1,  -1,  -3,  -6, -25, -11,  69, -21,  -3,  -5,  -4, -15, -30,  12,  31,  -6,   1,   1,  -4, -24,  12,  53, -26,  11,  -4,  -8, -13, -20,  21,  20,  -4,  -3,  -1,  -4,  -6,   3,  16,  -4,   5,  -7,  -9, -22,  -5,   1,  26,   4,   2,  -4,  -5,  13,  -2, -10,  -7,   0,   6,   2,   0,   0,   2,   7,  -6,  -4,   1,   3 },
     {    3,   6,   5,  24,  36, -41,  10,  -5,   1,   3,   7,  16, -25,  16,  -9,   9,   1,   5,   1, -34,  -1,   6,  17,   0,  -8,  -9, -42,   1,  -9, -37,  33,  -1,  -4, -21,  -3,  26, -22,  -8, -17,  21, -22,  -1,  12,   4, -17,  11, -26,   7, -12,   8,  15,  11, -10,  -7,  -3, -15,  10,  -1,  -1,  -1,   5,  -3,   2,  -6 },
     {   -1,   0,  -4,   4,   1, -18,   3,  -5,  -2,   0,  -5, -16, -11,   4,  14,  -2,  -2,   2,  -4,   6,  11, -10,   0,   9,  11,  15,  11,   3,  19,  12,  -8,   0, -13, -10, -45, -11,   4,  -6,  24,   1,   4,   5,  29,  16, -18, -29, -31,  12,  -1,   6, -41,  49,  26,   6, -17, -13,  -2,  22, -18, -18,  25, -12,  14, -11 },
     {   -5,  -7, -16, -39, -23,  23, -18,  16,   2,   3,   3,  34,  18, -32,  17,   9,  10,  16,  18,   0, -14,  26,  13,  17,  -5, -13, -33, -36, -11,  14,  10,  21, -15, -21, -18,  -2,   9, -14, -26,   5,  -3,   5, -12,   0,  -5,  -2, -16,  -3,  21,  -6,  -1,  22,   3,   4, -10, -19,  12,  -5, -11,  18,  -1,  -4,   6,  -2 },
     {    6,   9,  13,  29,  29,  13,   0, -12,  -2,  -2,  -9, -20, -16,   6,  10, -28,   4,   7,   9,  34,  22,  -3,  12,   6,   3,   5,  -1, -30, -34,  21,  12,  35,   0,   2,   0,  21,  17, -16,  -8,  -4,  -3,   7,  -6, -22,   2,  38,  10,  15,  14,  -7, -22,  14, -22, -14,   1,  15,  -8,  20, -15,   3,   9, -11, -18,  14 },
   }
 },
 {  //28
   {
     {  113, -18, -20,  -2,  -5,  -1,  -2,  -1, -31,  36,   7,  -4,   0,  -1,   0,  -1, -15,  -9,  12,   4,   1,   0,   0,   0,  -2,  -2,  -4,   3,   1,   0,   0,   0,  -3,  -1,   1,  -1,   2,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    8,  70,  -7, -10,  -2,  -3,  -1,  -2, -61, -42,  44,  13,   4,   2,   1,   1,  19, -34, -35,  11,   5,   3,   1,   1,   3,  16,  -7, -17,   0,   0,   0,   0,   4,  -4,   4,   1,  -5,   0,   0,   0,   1,   3,  -1,   0,   0,  -2,   0,   0,   1,  -2,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {   37, -24,  19, -10,   1,  -1,   0,  -1,   3, -65, -31,  28,   1,   6,   1,   3,   8,  49, -46, -33,   4,  -2,   1,  -1,  -4,  14,  37, -12, -15,  -2,  -2,  -1,   0,   0,  -1,  14,   0,  -4,   0,   0,  -2,   3,   4,  -2,   3,   1,  -1,   0,   0,   0,  -1,   1,   0,   1,   1,   0,  -1,   1,   1,  -1,   0,   0,   0,   1 },
     {    1, -54,  -6,   8,   9,   2,   3,   1,  33,  -9,  28,   1,  -9,   2,  -3,   1, -20, -21, -48,  42,  15,   3,   4,   1,  -3,  17, -32, -53,  16,   7,   4,   2,   3,   4,  20, -10, -30,   4,   1,   1,  -1,   5,  -4,   7,  -2, -13,   3,   0,   0,   1,   4,  -1,   1,  -1,  -6,   1,   0,   1,  -1,   1,   0,   0,  -1,  -3 },
     {   30,  73,   6, -18,  -5,  -5,  -2,  -3,  81,  -6,  10,  11,  -8,  -2,  -4,  -1, -48,  10,   1,  10,   9,   1,   2,  -1,  -7, -15,   4,  -3,   3,   3,   1,   1,  -4,  -2,  -3,   3,  -1,   1,   0,   0,  -3,  -2,   0,   0,   2,   1,   0,   0,  -2,  -1,  -1,   0,   0,   1,   1,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 },
     {  -17,  -2, -28, -11,   8,   2,   1,   2,  -1,  40,  11,   6, -17,  -3,  -1,  -3, -20, -20, -17, -26,  26,   1,   3,   1,  21,   7,  29, -27, -54,   2,  -3,   0,   2,   8,  14,  51,   6, -34,  -6,  -4,   2,  -2,  -2,  -5,  28,  17, -12,  -2,  -1,  -1,   1,   3,  -6,   8,  11,  -4,   0,  -1,  -1,  -1,   2,  -2,   2,   6 },
     {   -9,  25, -25,   5,  -2,   3,   1,  -1, -23,  25, -31, -13,   9,   5,   2,   2, -13,  24,   1, -22,   7,  -2,   7,  -3,   6, -26,  28, -43,  11,  15,  -1,   6,   3,  -9,   5,  -1, -59,  21,  10,   1,  -1,   3,  -2,  16, -13, -43,  15,   3,   3,  -3,   5,  -1,  10, -10, -22,   6,   0,   1,   0,   3,   0,   3,  -3, -10 },
     {   -1,  -1,  72,  -2,  -9,   2,  -2,   0, -21,   6, -37,  38,  17,  -3,   2,   0, -40, -43,   0, -21,  20,  13,   4,   2,  20, -17, -31,  -8, -15,   6,   4,   2,   0,  18,  -2, -18,  -5,  -7,   1,   0,   6,  -2,   8,   3,  -7,  -3,  -3,   1,  -1,   3,  -2,   0,   2,  -2,  -2,  -1,   1,   0,   2,   0,  -1,   0,  -1,  -1 },
     {   -7, -10, -15, -14,  -5,  13,   2,   0,   9,  17,  17,  23,   8, -16,   0,   0, -14, -17, -23, -21,   0,  24,   4,   0,  12,  19,  21,  13,  -5, -30,  -1,   0,  -6, -14,  -6,  -1,  36,  46,  -5,  -2,  -5,  -4, -12, -20, -45,  -7,  39,  10,   0,  -2,   2,   3,   4, -31, -26,  11,   0,   1,   0,  -1,  -1,   8,  -6, -13 },
     {   12, -20,  28, -24,   2,   1,   1,  -2,  20, -30,  41, -29,   0,   3,   0,   1,  22, -31,  31, -16, -15,  13,  -3,   2,  11, -15,   8,  15, -33,  19,  -5,   5,  -9,  13, -23,  36, -37,   7,  10,  -3,  -6,  10, -13,  14,   0, -29,  15,  -1,  -1,   0,  -1,   0,  10, -11,  -9,   4,  -1,   1,  -1,   1,   2,   0,  -2,  -4 },
     {    8,  13,  -7,  18, -17,  -1,   1,  -1,  14, -17, -35, -42,  29,   3,   0,   2, -15, -26,   3, -13, -48,   7,   5,   1,  21,  46,  -8,  -4,  -5, -36,  -8,  -3,   2,   8,  45,   9,  -1,   4, -16,  -7,  -2,  -1,  -5,  26,  12,   1,   5,  -4,  -2,  -1,   1,  -8,   8,   8,   1,   2,   0,  -1,  -1,   2,  -3,   0,   3,   0 },
     {  -10, -15, -56, -13,  18,   0,   4,   2, -20, -54,   2,  -4, -13,  10,   2,   3, -62,   7,  21,  -3,  -4,  -9,   3,   0,  42, -23, -15,  14,   1,  -2,  -6,   1,   2,  28, -13, -13,   5,   0,  -1,  -3,   5,   1,  14,  -6,  -7,   4,   0,  -1,   0,   3,  -1,   6,  -2,  -4,   2,   0,   1,   0,   2,  -1,   2,   0,  -2,   1 },
     {   -1,   0,  -4,   2,  -9,  -3,   3,   0,   0,   2,   0,   7,  -6, -11,  -2,   5,  -3,   1,  -8,   8, -15,  -6,   6,   1,  -4,   7,  -8,  15,  -6, -22,  -1,   6,  -7,   4, -14,   3,   8, -29,  14,  -3,   7, -14,  13, -25,  30, -51, -26,  26,   1,   3,   2,   2,   5,  30, -69,  -7,   0,   3,  -1,   3,  -2,  16,   9, -37 },
     {  -12,  -8,  57, -23,   2, -10,   3,  -4, -36,  17,  36, -19, -10,   6,  -6,   3, -47,  40,  -9,  27, -27,  -8,   5,  -3,  11,  -9,  22,   2,  17, -25,  -4,  -2,  15, -12,  11,  22,   5,   9, -10,  -3,  -1,   0, -12,   9,  14,   3,   6,  -1,   3,  -4,  -2,  -6,   3,   5,   0,   2,   0,  -1,  -2,   1,  -1,   0,   1,   1 },
     {   -8, -12, -12, -19, -21, -15,  -1,   3,  11,  19,  23,  28,  29,  21,  -2,  -6, -12, -17, -20, -25, -25, -16,   3,   7,  17,  20,  23,  24,  25,  11, -11, -10, -18, -23, -23, -23, -19,   1,  16,  13,  13,  17,  15,  16,   4, -12, -31, -15,  -4,  -3,  -2,   3,  10,  26,  25,  -3,   0,   0,  -1,  -4,  -6, -11,   0,   9 },
     {   -7,  -7, -12, -70, -40,   5,   5,   0,  -6,  -6, -12,  27,   9, -29, -12,   1,  14,  23,  43,  31,  18,  16,  -6, -10,  24,  30, -12, -14,   0,   5,  14,   5, -17, -14,  19,   2,  -7,  -5,  -4,   3,  -1,  -7,  -7,  11,   5,  -1,   0,  -4,  -1,   1,   0,  -6,   4,   3,  -1,   1,   0,  -2,   0,   2,  -3,   0,   2,   0 },
     {   -5,  -7, -22, -26, -19,   2,   5,   2,   3,   8,  23,  25,  26,  -7,  -9,  -3,  10,  13, -12, -25, -20,  19,  10,   2, -47, -43, -21,   4,  -9, -23,  -2,   4,  44,  38,  17, -11,  -6,  -6, -16, -10, -10,  -2,  11,  27,  11,   4,   7,  -2,   2,  -1,  -6,  -8,   6,   4,   2,   4,   0,   0,   4,   1,  -3,  -1,   0,  -1 },
     {   -7,  11, -10,  16, -19,  15,  -5,   2, -12,  17, -19,  26, -28,  23,  -6,  -1, -15,  21, -23,  29, -29,  18,  -4,   0, -13,  27, -27,  32, -25,  17,   4,  -8, -16,  24, -20,  22, -15,  -1,  20, -13,  -7,   6,  -7,   2,   7, -19,  27, -14,   1,  -4,   4, -10,  13, -23,  20,  -3,   2,  -4,   3,  -6,   7, -10,   0,   7 },
     {    7,   9,   8,  45,  32, -10,  -3,  -3,   8,  13,  43,  33,  19,   7, -10,  -5,  20,  47,  16, -21,  -4,  13,  12,  -4,  55,   7, -27,  -9,  -7,  -4,   5,   8, -24,  11,  14,  -6,  -4,   0,  -1,  -2,  -9, -20,   1,  10,   3,   0,   3,   3,  -4,  -2,  -6,  -3,   4,   2,  -3,   1,  -1,  -1,  -1,  -1,  -1,   3,   2,   0 },
     {    6,  10,   9, -51,  18,  -5,   4,  -2,  10,  25, -24,  -9, -35,  17,   1,   2,  43,  -7, -29,   5, -12, -33,  12,   3,  31, -11, -10,  -6,  11, -13, -28,   5, -24,  39,   2,  -9,   2,  15, -10, -16,   1,  -9,  26,   2,  -4,   3,  10,  -3,  -1,   1,  -2,  10,   1,  -1,  -1,   4,  -1,  -1,   3,  -2,   4,   0,   1,  -1 },
     {    0,   6,   7, -27,  22,   5,  -5,   3,  -1,   9,   9, -31, -22,   7,  17,  -2,  -5,  20,  -9, -17, -14,  15,  -3,   9, -11,  19, -15,   3, -34,  -6,  35,  -4,  -1, -18,  14, -36,  -3, -36,   2,  23,   4,   2,  12,  17, -41,  -7, -28,  -3,   1,   4,   1,  23,  17, -26,  -3, -12,   2,  -2,   1,  -2,  10,   9,  -6,   0 },
     {   -1,  -1,  -1,  -2,  -6, -13, -14,   4,   2,   3,   2,   3,   5,  14,  24,   0,  -1,  -2,  -3,  -3,  -6, -16, -19,   8,   5,   3,   1,   3,   8,  10,  25,  -4,  -7,   0,  -5,  -3,  -5, -27, -20,  16,   1,   0,   9,   2,  23,  20,  22, -17,  -5, -14, -12, -20, -21, -26, -14,  57,  -3,   2,  -3,   2,  -6,  -3, -45, -54 },
     {   -1,  -1,  -3, -18,  61, -16,  -6,   8,  -1,  -3, -15,  45, -22,  35,  -2, -10,  -3, -20,  41,  -8, -15,  -4,  27,   0, -30,  29,  -6,  -7,  10, -16,  -5,  16,  25, -29,   1,  10, -11,   4,  -8,  -5,  -6,   7, -14,   0,   8,  -5,   5,  -2,  -2,  -2,   2,  -7,   2,   2,  -4,   2,   1,   0,  -3,   1,  -1,   1,   0,  -3 },
     {    2,  -1,  -1,  28,  -5,   9, -14,   1,   0,  -5,   3,  24, -43, -35,  12,   1,   5,  -7,   5,   1,  12, -15, -24,  -1,   9,  -8,  11,   7,   3, -27,   4,  -1,  -2,  -2, -12,  -2,  -6,   3, -49,   4,   8,  25,   1,  37,   1,  -9,  -2, -38,   0,  -4,  11,   2,  46,  11,  -3,  -1,   1,   1,  -2,  -2,  -4,  23,  10,   0 },
     {    1,   0,  -4,  -1,   6,  13,  18,  -5,   0,   0,  -9, -10,  30,  15, -10,  12,   4,   0,  -8,  19, -17,  27,  25,  -7,  -3, -26,  -4, -23, -14, -31,  23,  33, -36, -22, -44,   4,  -5, -11, -40,   0,  10,  16,   9, -25,   0,   3,  -1, -28,   4,  -1,  12,   2, -11,   7,  10,   2,  -1,  -1,  -1,   4,   0,  -9,   4,   7 },
     {   -2,   0,  -1,  -1,  -3,   6, -18,  -7,  -2,   2,   1,   2,   1,  17, -10,  -3,  -4,   3,  -1,  -3,  -7,   9, -22,  -8,  -2,   2,  -2,  -1,  -1,  17, -11,  -4,  -3,   1,  -1,   4, -11,  15, -25,  -8,  -8,   4, -12,   8, -19,  27, -13, -17, -11,  10, -17,   9, -28,   8, -11, -49,   5,  -4,   9,  -3,  13, -14,  41, -73 },
     {   -4,  -3,  -6,  -6,   1,   5,  -6,   2,  -7, -10,   1,  -4, -20,  -5,   4,  -5,  -6,   1, -14, -31,   8,   1,  10,   8, -23, -32, -29,  19,   8,  -7,  -9,  -2, -64, -27,  37,   8,  -1,   7,   3,  -1,  40, -12, -43,   8,  12,  -2,  -2,   1,  -1,  17,   3, -23,  -8,   2,   3,   2,   7,   4,   8,   5,  -5,  -7,  -5,  -1 },
     {   -3,  -5,  -4,  10, -39, -91, -27,   4,  -2,  -3,  -4, -17, -24,  15, -34, -18,   4,  -1,  -5,  -3,  23,  27,  26,  -5,   2,  -1,   2,   8,  -1,  -2,   4,  20,   3,   1,  -2,  -3,   2,   0, -10,  -4,   4,   4,   3,  -2,  -8,  -1,   1,  -2,  -3,   8,   0,   8,   5,  -7,  -2,   4,   1,  -1,  -1,  -1,   5,   9,  -1,   1 },
     {   -6,   2,  10,  -5,  16,  -3,   0,  -6, -10,   2,  17, -20,  13, -15,  -7,   4, -22,  14,  15, -23,  21,   4, -14, -15, -47,  46,  -5,  -9,   1,  15, -23,   9, -28,  22,  -8,  -8,   0,  12,  -7, -26,  31,  -4,  42,  -5,   9,   0,  22, -19,   4,  -6,   2,  19,   6,  13,   1,  18,  -1,   0,  -3,  -1,  -1,   6,   7,  -2 },
     {   -1,  -5,   3,   7, -35,  29, -24, -19,   2,  -5,  15,   6, -19,  47, -29,  36,   1, -10,  22, -19,  -8, -18,  -5,  -1,  -5,  -3,  14, -40,  14,  -6,  17,   1,  -9,  -1,  -4, -19,  35, -21,  13, -26,   0,  -9,  -8,  15,   7, -10,  13,   5,   8,   8,  -4,  10,   2, -14,   4,   0,   7,   0,   0,   9,   9,   7,   6,   7 },
     {   -2,  -3,  -3,   3,  16,   6,  13,  -6,  -3,  -4,  -4,   4,  26,  39,  -5,  -1,  -3, -12,  -1,  43,  26,  -2,  11,   5, -14,  -2,  47,  26, -22,   7,   7,   7, -15,  21,  17, -15,   2,   0, -14,   9,   9, -49,  -9,  23, -10,  -4,  -3,  -9,  -6,  15, -19, -13,  27,   6,  -6,   5,   3,  -8,   7,  -8, -14,  17,   4,  -2 },
     {    3,   3,   2,   4,  18,  23,  23,  -2,  -2,  -1,  -6,  -8, -37, -32, -66, -11,  -1, -14, -13, -15, -20,  14,  10, -27,   6,   6,  15,   9,  12,  24,  40,  14,   7,   3,  -6, -14,   0,   7,   3,  18,   2,  -9,  11,  20,  10,   8,   3,   9, -10,   7,  -7, -15, -11,  16,   0,  14,   0,   1,  -2,   3, -10, -17,  -2,  -7 },
   },
   {
     { -115,  14,  11,   3,   3,   1,   1,   0,  40, -26,  -3,  -1,   0,   0,   0,   0,  21,   6,  -5,  -1,  -1,   0,   0,   0,  -1,   1,   1,   0,   0,   0,   0,   0,   3,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   25, -30,   2,  -2,   1,   0,   0,   0,  95,  11, -23,  -2,  -6,  -1,  -2,   0, -31,  56,   3,  -4,  -1,  -2,   0,   0, -24,  -3,  19,  -1,   2,  -1,   1,   0,  -4,  -7,   4,   2,   0,   0,   0,   0,  -3,  -3,   3,   1,  -1,   0,   0,   0,  -1,   0,   0,   3,  -1,   0,   0,   0,  -2,   0,  -1,   2,   0,   0,   0,   0 },
     {  -27,   4,   5,   2,   1,   1,   0,   0,  -2,  50,  -4,  -2,  -1,  -1,   0,   0, -70, -20,  43,   2,   6,   1,   2,   0,  23, -64, -10,  14,   0,   4,   0,   1,   9,   9, -26,   0,   0,   1,   0,   0,   1,   3,  -4,  -5,   2,   0,   0,   0,   1,   2,  -1,  -5,   1,   0,   0,   0,   1,  -1,   2,  -2,   0,   1,   0,   0 },
     {   -2, -28,   2,   1,   0,   1,   0,   0,  27,  17, -11,  -5,  -1,  -2,  -1,   0,  -5, -39,   3,   6,  -1,   2,   0,   1,  25,  32, -57,  -9,  -2,  -1,  -1,   0, -12,  52,  38, -34,  -3,  -6,   0,  -1,   4, -21,  35,  18,  -9,   1,  -1,   0,  -3,  -1,  -7,  12,   1,   0,   1,   0,  -1,  -1,  -1,   3,   0,  -1,   0,   0 },
     {   15,  -2,   0,  -4,   0,  -1,   0,   0,   7, -38,  -2,   5,  -1,   2,   0,   0,  13,   6, -10,  -5,   0,  -1,   0,   0,   2, -26, -15,  13,  -2,   3,  -1,   1,  -2,  38, -58, -21,   7,  -3,   1,  -1,   1,  27,  44, -48, -11,  -3,  -1,  -1,  10,  -8,  30,  26, -21,  -1,  -3,   0,  -4,   5,  -7,  18,   7,  -6,   1,  -1 },
     {   -1, -95,   6,   6,   5,   3,   2,   1,   9, -19,   6,   4,   1,   1,   0,   0,  20, -31,  -4,  13,   1,   3,   0,   1,  11,  -6, -14,  12,  -1,   3,  -1,   0,  -2,  -6, -29,  20,   3,   3,   0,   0,  -7,  20, -40,  -5,  12,   0,   1,  -1,  -4,  10,  -2, -26,  11,   2,   1,   0,   3,  -1,   7, -14,  -1,   4,  -1,   1 },
     {  -17, -47,  12,  -2,   0,   0,   0,   0, -53,   9, -38,   1,   2,   1,   1,   1,  28,  58,  26, -14,  -5,  -4,  -1,  -1, -25, -27,  -1,   8,   8,   2,   2,   1,  20,  15,  10, -26,  -6,  -3,  -1,   0,  -1, -14,  21,  21,  -4,   1,   0,   0,   2,  -2, -14,  -2,   2,   1,   1,   0,   1,   0,   5,   1,  -5,  -1,   0,   0 },
     {   30,  47,  25,   0,  -2,  -2,  -1,  -1,  24, -56, -14,   8,   6,   3,   1,   0,  13,  -5,  22,   4,  -5,   0,  -1,   0,  -8, -33, -29,  22,   5,   4,   1,   1,  10,  25,  -4,  -6,  -4,  -1,  -1,  -1, -15,  -6, -15,  32,  12,   1,   1,   0,  -8,   5, -30, -31,  21,   2,   2,   0,   2,  -3,  13, -19,  -9,   5,  -1,   1 },
     {    6,  17, -13,   1,  -3,   1,  -1,   0,  26,  64,  12, -14,  -3,  -4,  -1,  -1,  74, -21,  38,   6,  -3,   0,  -1,   0, -40,  -6, -11,  27,   2,   3,   0,   1, -16,  -9, -17,  -4,   9,   0,   1,   0,   0,   0,  -2,  -8,  -2,   1,   0,   0,   0,   2,   4,   1,  -1,  -2,   1,   0,   0,   0,   0,   1,   2,   0,   0,   0 },
     {   -1,  -6,  -6,   0,   5,  -1,   1,   0,   0,  -3,  18,  11,  -4,   1,  -1,   0,   0,   0,  -6,   1,   7,  -1,   1,   0,   2,  -6,  17,  14,  -6,   2,  -1,   0,  -3,  -4, -16,   5,  12,   0,   2,   0,  11,   3,  25,  21, -16,   5,  -4,   0, -16,   4, -39,  57,  43,  -7,   2,  -2,  -5, -10, -16, -55,  51,  14,   0,   2 },
     {   10, -16,  30,  -8,  -6,  -1,   1,   0,  14, -25,  34,  -9,  -4,  -3,   0,   0,  18, -23,  36,  -6, -11,   0,  -1,   1,  21, -28,  29,  -1, -12,   1,  -2,   1,  22, -18,  23,  13, -23,   2,  -3,   1,   8, -10,   3,  27, -22,   0,  -2,   0,   3,   6,  -7,  37, -28,  -3,   0,  -1,  -7,  16, -17,  43,  -2, -15,   2,  -2 },
     {    4,   7,   1,  -1,   2,  -1,   0,   0,   5,  20, -12,   4,  -3,   1,  -1,   0,  33,  22,   7, -19,   1,  -5,   0,  -1,  74,   1,  36,   3, -13,   0,  -3,   0, -29,  41,  -5,  42,   1,   0,  -1,  -1, -29,  -9,   6,   1,  26,  -2,   2,   0,  -1, -22,  -3,  -8,   4,   6,   1,   1,  -3,  -2,  -7,  -4,  -7,   4,   2,   1 },
     {   -3,  -5, -42,   4,   2,  -1,   0,   0,   7,  12,  52, -10,  -4,   0,  -1,  -1,   6,   4, -29,  25,   5,   0,   1,   0, -24, -31,  -2, -34,   8,   4,   2,   1,  41,  33,   6,  14, -30,  -4,  -2,  -1, -23,  12,  16,  18,  26,  -5,  -1,  -1,   2, -29,  -8,  -7,   2,  10,   2,   2,  -2,   5,  -6,  -5, -11,  -1,   1,   0 },
     {   -5,  -4, -82,  -3,  -2,   0,   0,   2,  -2, -34,  27, -26,  -4,   0,   0,  -1,  -8,  34,  45,  20, -10,  -4,  -3,  -1,  32,  10, -16,  22,  13,   1,   0,  -1, -17, -11,   3, -15,   7,   5,   2,   1,  -4, -13,  -8,  -2,  -8,   0,   1,   1,   0,   8,  -4,  -3,  -5,  -4,  -1,   0,   1,   2,   3,   2,  -1,  -1,  -1,  -1 },
     {   -3,  -7, -10, -27,  -3,   1,   0,   0,   7,   6,  25,  21,  -2,  -5,  -1,   0,  -5, -11, -14, -34,  -1,   5,   1,   1,   7,   7,  23,  22,  -9,  -5,  -2,  -1,  -3, -15, -19, -32,   7,   7,   1,   2,  12,  12,  28,  18, -27,  -8,  -2,  -1,  -9, -17, -20, -20,  19,   7,   3,   1,   7,   5,  13,  -6, -76,   0,   3,   0 },
     {   -7, -12,  41,   1,   0,   0,   1,  -1, -16, -13,  49,  -1,  -4,   2,  -1,   0, -30,   7,  20,  20,  -8,   0,  -1,   1, -44,  22,   4,  30,  -6,   0,   0,   0, -42,  21,  -1,  22,  16,  -2,   2,  -2, -10, -34,  13,  -9,  19,   1,   1,   0,  15, -31,  -6,   0,  -5,   5,  -3,   2,   2,  -3, -11,   3, -12,   3,   0,   0 },
     {    4,   7,  54, -10,  -1,   1,   0,  -1,   2,  27,  48,   1,  -5,   1,   0,   0,  10,  52, -19,  27, -14,   2,  -3,  -2,  42,   8, -42,   4,  16,  -6,   4,  -1,  22, -20, -12, -18,   4,   3,  -1,   3,  -3,   4, -10,  -1,  -5,  -2,  -3,   3, -21,   3,  11,   2,   4,  -1,   0,  -2,  -3, -11,   2,   0,   0,   1,   4,   0 },
     {    1,  -1,   0,  -6,   9,  -3,   1,  -1,   0,   3,  16,  18,  -5,   3,   1,  -1,  13,   8, -34,   9,  17,  -3,   2,   1,   3, -50,  -2, -31,  16,   9,   2,   3, -68,   0,  -2,  -9, -18,   6,   2,   0,  30, -43,  -7,  -4,  -8,  -3,   0,   3,  11,  31, -16,  -8, -13,  -7,  -1,   2,  -4,  14,  14,  -4,  -3,  -5,  -1,   0 },
     {   -7,   7, -14,  19, -31,   5,   2,   4,  -9,  15, -13,  36, -23,  12,  -8,  -1, -11,  11, -22,  20, -27,  11,  -4,   1, -10,  17,  -9,  34, -22,  12,  -5,   0, -12,  19, -17,  21, -27,   2,   7,  -1, -17,  23, -15,  29, -16,  -3,  10,  -2,  -8,  30, -21,  18, -23,  -8,   6,   1, -17,  13,   8,  21,  -7, -21,   2,   0 },
     {    3,  -1,  -9,  13,   2,  -2,  -1,  -3,   5,   3,  -9,  11, -10,   3,   1,  -2,   5,  -7, -17,   6,   2,   7,   0,  -2,   9,   0,   0,  11,  -1,  -1,  -8,  -2,  43, -12, -10,   1,  49, -14,  -7,   0,   1, -49,   7, -22,  26,  33, -14,   2,  -7,   9, -38,   7, -44,  26,   8,   0, -13,  19,   1, -13, -14, -16,   6,   2 },
     {   -4,  -4, -13, -57, -52,   0,   8,   2,   3,   0,   8,  59,  33, -18,  -8,  -9,   3,   6,  13, -11,  15,   5,  12,   3,  -6,  -7, -22,  -5,  -6, -19,  14,  -3,   5,   4,  19,  19,   9,  -6,  -5,  -5,  -1,  -1, -11, -10,  -4,  18,  12,  10,  -3, -13,  10,   0, -14,  -6, -17,  11,  -7,   4,  -6, -13,  12,   2,  -5,  -9 },
     {    0,   2,  -4, -17,  10,   4,   9,  -4,   0,  -1,  27,   4, -13,  11,   2,   2,  -1,  15,  21, -26,   8, -13,   0,  -1,  -8,  14,   0,  -8, -38,  21,  -2,  -1,  15,  33,   0,  13,  -1, -32,   3,  -2,  58,   5,  -9,  -2,  28,  -1, -11,  -5, -21,  39,  17,  -1,   6,  13,  12, -12,   4, -17,  31,   2,   0,  -7,   5,  10 },
     {    0,   0,  -2, -18,  -8,  32,  -3,   7,   2,   3,   3,  -3,  -4,   8,  -1,  12,   0,  -5, -16,  12,   1,  10,   4,  12,   4,  -2,  23,  13,  -8,  12,  18,   3,   7,   8,   0, -18,  22,  12,  18,  -9,  -1, -14,   2,   3,   2,  -2,   2, -14,  -5, -12,  -5, -28,  31,  10, -34, -12,  18,  15,  17,  34,  47, -54,  -5, -18 },
     {   -2,   0,   0, -34,  -7,  -3,   4,  -1,  -3,   1, -10, -56,  -6,  -1, -11,  -4,  -4,  -5, -26, -18, -46,  17,  -8,   7,  -7, -12,  -5,  15,   8, -36,  -3, -16,  -3,  11,  -1,  27, -12, -17,  -8,   3,  16, -10,  -3, -13,  -5,  -9,  -3,  13, -30,  -5, -10,  -1, -13, -28,  10,  10,  29,  -1,   7, -17,   4, -16,  -9, -15 },
     {    0,  -1,  -2,   9,  26,   7,  -5,   2,   2,   8,  33,   5,   9,   2,   0,   0,  -3,  -1, -10, -42,  -8,   0,   4,  11,  -9,   2,  17,  11,  12, -18, -12,   1,  24,  38,  10,  -9,  17,   5,   5,  -2, -45, -20, -21, -14, -31,  -5,  15,   3,  18,  50,   5,  -6,  -1, -23, -15,   2, -11,  -7,   6,   0,   8,  19,  -4,  -1 },
     {    1,  -2,  -1,  60, -34,   9, -16,   7,   2,  -1,  12,  -2,  49, -36,  -8,   3,   2,   4,  -7,  14, -21,  21,  -7, -14,   3, -17,  28,  -1, -12, -15,  17,  -1, -13,  18,  10, -19,  16,  -8,  -8,  -2,  21,   6,   0,   2,   0,  20,  -5,   1, -26,  10,  11,  -8,   1,   2,  11,  -2,   8, -26,  -5,   5,  -6,   5,  -2,  -5 },
     {    0,   0,   0,  11, -15,  -1, -15,   6,  -1,  -1,  -4, -13,   9, -19,  -7,  -5,  -2,   3,   5,  -9,  -3,  12,   5,   5,  -1,  10, -13, -15, -35,  -5,  -3,  10,   2, -18, -29,   0, -26, -31,   9,  -4, -10, -26,   6,  -3,   3, -47, -22,  12,   2,  14,  -8,  -2,   6,  26, -53,  -4, -12,  10,  -7,  -5,  -1,  23,  13, -39 },
     {    0,  -3, -11,  24,   9, -17,  -6,  -9,   4,   2,  15,  10,  30,  28,  14,   8,  -1,   7, -17, -58, -13,  10,  -3,  -6,   0, -21, -36,  19, -19,   4, -10,   0, -16, -23,  24, -11,   6,   9,  -1,  -2,  -3,  11, -14,   0,  37, -22,  12,   2,  10, -21, -12,  20,  -4,  10,   0,  -4,  14,  -9, -10,   9,  12, -23,  13,   0 },
     {   -3,  -3,  -1, -14, -24,   3,  -9,   7,  -5, -10,  -6,   1, -41,  -7, -18,  12,   1,  -7,  -5,   4,  21,  20,  10,  -6,  -5, -24,  -5,  -5,   2,   5,  -3,   3, -12, -20,  23,   2,  24, -24,   2,   1, -48,  -6,  17, -12,  15, -22,  14,  -2, -28,  29,  14,  20,  20,   1,  16, -20,  21, -30,  -4,  16, -19,   1,   2,  16 },
     {    0,   2,  -3,  -2,  18, -14,  -4,   6,  -1,   0,   1,   3, -16,   6,  -5,   7,   1,   5,  -8,   0,  -3,  -7, -11,  31,  -1, -16,  -7,  -6, -15, -11, -20,   3, -18,   0,   3,  -8,  22, -54,  33, -14, -10,  14, -22,  32, -36,  31, -18, -13,   6, -21,  15,  -9, -20,  29,   1,  -9,  19,  -4, -21,  -4,   7,   5, -19,  -5 },
     {    0,   0, -12,  13, -16, -11,   5,  14,   1,  -1,  -7,  11, -21,  12,  12,  14,   2,  -1, -11, -11, -16,   7,   3,  11,   0,  -6, -12,  17,  34, -17,  -5,   1,  11,   4,  -5,  24, -10,  15,   6, -20,  33, -29,  -4,   7,   9,  31,  -6,  -9,   8,   8,  37,  12,  49,   8, -19,   0,   7,  10, -33,  18, -20,  16,  22,  -4 },
     {    0,   1,  -7, -11,  22,  -1,  10, -12,   1,   4,  -8,  29,  13,   3, -10,  -1,   2,  -1,  -4,  36, -10,  -9, -17,   9,   2,  -8,  12,  11,  -4,  -9, -22, -35,   2,   3,  13, -11,  23, -15, -34, -33,   9,   8,  -7,   2,  20, -39, -35,  15,   9,  -1,  -2,   6,  14, -33,  -1, -22,   2,  19,  -6,  24,  -5,  22,  -5,  -8 },
   },
   {
     {  101, -59,  -1,  -2,   0,  -1,   0,   0,   5,  19, -21,   5,  -2,   1,  -1,   0, -29,  27,  -1,  -4,   2,  -1,   1,  -1,  -7,  -1,   9,  -3,   1,   0,   0,   0,  -4,   0,   2,   1,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -36, -54,  43,  -1,   8,   0,   2,   0,  56, -56, -10,  17,  -4,   4,  -2,   2,   1,  33, -42,   8,   1,   0,   0,   0, -10,  19,  -2, -13,   5,  -1,   1,  -1,  -1,   4,   4,  -5,  -2,   1,   0,   0,  -2,   3,   1,  -1,  -1,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   11,  50, -60,  17,  -8,   3,  -2,   1,  82, -11, -21, -11,   4,  -3,   0,  -1,  -9,  19,  11, -16,   3,  -1,   0,  -1, -28,  12,  17,  -4,  -2,   0,   0,   0,  -6,  -7,  11,   2,  -1,   0,   0,   0,  -3,  -3,   3,   3,   0,   0,   0,   0,  -2,  -2,   1,   1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0 },
     {  -50, -14,   3,  12,   2,   3,   1,   1,  -9,  52, -62,  13,   1,   0,   0,   0, -20,  37,   8, -34,  15,  -6,   3,  -2,  19, -32,  38, -16,  -4,   5,  -2,   1,   8,  -8,   2,   8,  -9,   2,   0,   0,   2,  -3,   1,   2,   0,  -2,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0 },
     {    7,  -9, -51,  48,  -8,   6,  -3,   2, -47, -29, -24,  33,  13,  -1,   3,   1,  31, -14, -38, -12,  23,  -4,   3,  -1,   1,  27,  -6, -38,  10,   3,   0,   0,   2,   4,  16, -17,  -9,   5,  -1,   0,   1,   2,   7,  -1,  -7,  -1,   1,   0,   1,   0,   3,  -1,  -1,  -1,  -1,   0,   0,   1,   1,   0,   0,   0,  -1,   0 },
     {  -13, -58,  -9,  31,  -3,   8,  -2,   3,  27,   6, -19, -12,  10,  -4,   2,  -2,  40, -46,  52, -27,   0,   4,  -3,   1,   0,   5, -19,  29, -20,   7,  -2,   1, -16,  27, -18,   5,   5,  -8,   4,  -2,  -1,   1,   2,  -5,   4,  -1,  -1,   1,  -1,   2,   0,  -1,   0,   1,  -1,   0,  -1,   1,   0,  -1,   0,   0,   0,   0 },
     {  -16, -20, -21,  49, -15,   6,  -3,   2, -18, -44,  43,  -7,  13,  -5,   2,  -1, -75,  18,  27, -10,   5,   3,   0,   0,   3, -17,  -4,  11,  -5,   3,   1,   0,  20,  -8, -17,   5,   2,  -1,   1,   0,   6,   7, -10,  -3,   3,   0,  -1,   1,   2,   2,  -1,  -4,   0,   1,  -1,   0,   1,   1,   0,  -1,   0,   0,   0,  -1 },
     {   14,   3,   4, -32,  14,   2,   2,   0, -16, -41,  13, -28,  17,   8,   4,   1,  22,   4,  -3, -49,   1,  11,   1,   1,   8,   0,  40, -17, -44,  20,  -3,   2,   0,   0,  -2,  42, -38,  -6,   8,  -1,  -4,   7,  -8,  16,   6, -20,   4,   0,   0,  -1,  -2,   3,   6,  -4,  -5,   2,  -1,   1,  -1,   1,   1,   0,  -1,  -2 },
     {   22,  34,  47,  13, -48,   7,  -9,   0,   8, -18,   4,  57, -27,  -7,   2,  -1,   8,  -8,  17, -22,  42, -29,   4,  -2,  -2, -11,  12,   5,   5,  11, -12,   3,  -9,  10, -26,  16,   1,   5,   2,  -4,  -1,   0,  -2,  -8,   8,   1,   1,   0,  -1,   0,  -3,  -2,   0,   3,   1,   0,  -1,   0,  -1,  -1,   0,   0,   1,   1 },
     {    8,   7, -26, -20,  31, -11,   6,  -1,  22, -27, -24,  19,   1,  11,  -5,   3, -24, -38, -12,  23,   5,   7,   2,   0,  52, -56,  -5, -12,   7,   6,  -1,   2, -10,  31, -30,  -1, -10,   7,   0,   0,  -9,  18,   3, -14,   0,  -5,   4,  -2,  -3,   2,   7,  -3,  -3,  -1,  -2,   2,  -1,   1,   2,  -1,  -1,  -1,  -1,  -2 },
     {    6,  15, -11,  -2,  -1,  -2,   1,   0, -25, -33, -30,  12,   2,   6,   1,   4,  23,  41, -10,  -7, -37,   5,  -1,  -1,  13,  -6,  14,  51, -17, -27,   9,  -5,  -3,   8, -22,   4,  50, -27,  -3,   2,  -5,   1,  12, -22,  16,  16, -15,   2,   0,  -2,   1,  -1,  -5,   9,   2,  -4,   0,   0,   2,  -1,   0,   0,   2,   1 },
     {    9,   2,   1,  23, -40,  16,  -6,   2,  22,  23,  28, -43,  10, -16,   6,  -3,  27,  36, -39,  -4,  11,  -7,  -2,   2,  67, -17, -24,   3,  -3,   7,  -6,   1,   0,  12,   1,  -4,  -3,   2,   0,  -1, -17,   8,   9,  -2,   0,  -1,   2,  -1,  -4,  -3,   7,   2,   0,  -1,   1,   0,  -2,  -1,   0,   3,  -1,   0,   0,   0 },
     {   11,  19,  30,   9,  32, -41,   0,  -6,   1,  -5,  -4,   5,  68, -35,  -8,  -3,   4,  11,   7, -15,  24,  23, -28,   3,  -3,  -4,  -7,  11,   1,  30,  -8,  -8,  -2,  -5,   0, -24,  12,  19,   2,  -4,  -3,  -3,   6,  -6, -20,  24,   2,  -2,  -2,   1,  -3,   3, -11,   1,  14,  -4,   1,  -2,   1,   0,  -2,  -3,   3,   3 },
     {   -5,  20,  12, -20, -10,  17,  -2,   1,  -8,   8, -24,  -5,   6,   8,   4,  -1, -50,  -4, -19, -20,  16,  13,   4,   1,  -5,  35, -48,   3, -36,  18,   7,  -1, -16,  34,   1, -12,  -3, -28,  21,  -5,  13, -19,  30, -15,   9, -17,  -3,   9,   3,  -6,   6,   6,  -5,   1, -11,   4,   1,  -2,   1,   3,   1,  -1,  -2,  -4 },
     {   -3,  -7,  -7, -48,  -3,  36,  -3,   3, -10, -26, -14, -39,   5,  18,  13,  -2,  -2,   0,  13, -34,  30, -31,  15,  -2,  -4,  -2, -10,   6,  36, -13, -12,  11,   7,   0,  -7, -22,  14,  27, -22,   3,   4,  -2,  10,  -5, -28,  30,  -1,  -8,   0,   2,  -1,   7, -15,  -1,  12,  -5,   0,   0,   2,   3,  -3,  -3,   3,   2 },
     {    4,   8,   1,   6,  31, -32,   6,  -5,   5,   3,  23,  20, -32,  20, -16,   1,   6,  31,  43, -35, -18,   2,   8,  -6,  41,  41, -19, -28,  11, -14,   6,   3,  21,  24,  -9, -12,  -9,   6,  -6,   4,   0,  -6,  17,   1, -10,   0,   3,  -2,  -6,  -5,   6,  11,  -4,  -2,   3,   0,  -4,  -2,  -1,   7,  -1,  -2,   1,   1 },
     {  -10, -18, -27, -19,  22,   0,   0,   0,  14,  24,  41,  18, -12, -16,  -5,  -7, -12, -20, -25, -11,  36,  -5,  -5,   3,  14,  23,  23,  -9, -27,  -3, -10,  -3,   1,   1,   6,  29,  37, -16, -15,  -4,   5,  -4,  -1, -20,  13,  42, -15,  -5,   0,   0,  -2,   5, -14,  13,  20,  -9,   1,   0,  -1,  -3,   1,  -2,   7,   6 },
     {   12,  11,  34,   9, -27,   7,  -2,  -1,  12, -15, -28,   0,  20,  -1,   3,  -3, -23, -44,   9,  -9, -24,  13,   6,   3,  48,  17,   6, -14, -12, -33,   7,  -1,  16, -21,  46,  -1,  12, -10, -18,   7,   1,   6, -13,  22,  -5,  17, -14,  -1,  -5,   8,  -5,   2,   3,   3,   7,  -7,   1,   1,  -2,  -1,   1,   1,   3,   3 },
     {   -5,   1,  10,  17,  -7,   2,  -5,   4, -11,  -5,  16, -13,  10,  -9,   1,  -1, -11,   4,   8,  13, -19,   1,   4,  -6,  -4,  -6,  44, -22,   8, -16,   0,  -2, -72,  59,  16,  -3,   2,  10, -17,   3,   6, -25,  19,   8,  -9,  10,   4,  -6,  15, -15,  -6,  14,  -5,  -1,   5,   0,   1,   1,  -8,   2,   3,  -2,   2,  -1 },
     {   -7,  -6, -19, -27,   0,   7, -10,   8,  -9, -16, -18,  13,   0, -37,   4,   5,  -8,  16,  28,  33, -13, -42, -21,   8,  34,  34,  -1,  18,  -5,  24, -45,  -7,  -4,   5,  25,   6,  -5,  16,  10, -27,   1,  -4,  -5,  13,  13,  -8,  16,  -5,  -2,   0,  -1,  -6,  12,   0,   1,   4,   0,  -4,   0,   0,   2,   4,  -1,   4 },
     {    1,  -1,   4,  -2,  16, -37,  23,  -2,   2,  -2,   1, -11,  20, -46,  32,  -2,   0,  -3,   2,  -5,   9, -46,  31,   5,  -4,  -4,  -4, -10,  11, -26,  -5,  24,   3,   1,   0,  -3,  -3, -21, -12,  17,   3,  -4,   7,  -6,  23, -20, -31,  19,   1,   0,   5,  -4,   5,  22, -40,   2,  -1,   1,   0,   3,  -5,  14,  -5, -12 },
     {    4,   5,   6,   5,  36,  35, -73,  18,   6,   8,  10,  18,  37,   8, -19, -19,   7,   8,   7,  20,   2, -23,  46, -32,   4,   3,   3,   7, -17,  -1,  19,   4,   6,  -3,  -2,  -5,  -9,  -2,   1,  14,   0,   4,   0,   1,   1,  -5,  -4,   6,  -2,   0,   0,  -2,   1,   4,   0,   0,   0,  -2,  -2,   2,  -3,  -2,  -1,   3 },
     {   -2,  12,   9,  10,   5,   0,   2,  -2,  -6,  17, -18, -13,  25,   0,   7,   1, -15,   3, -14,  -1,  -6,   3,  10,   5,  -8,  43, -15,  -8,  20, -18,  -1,  -3, -11,  18, -45,  46,   2,  23, -24,  -1, -19,  26, -39,  -2,  18,   2,  18, -20,  -4, -10,  15, -28,  11,  -1,  13,   2,   0,  -7,   5,  -5,  -8,   3,  -1,   7 },
     {    1,   4,   2,  15,  35,  -4,  -9,   5,  -4, -10,   0,  -5, -21,  24,  24, -14,  -4,   5,   0, -20,  14,  -3,  -2,   1,  24,  -3, -24,  12,  29,  11,   8, -11, -67, -40,  24,  16,  10,  -3,  22,  11,  -9, -21, -27,   3,  13,   0, -10,  13,  -4,   2, -11, -11,   3,  -2,   0,  -3,   2,   0,   0,  -7,  -1,  -2,  -1,  -1 },
     {   -5,  -8, -11, -16, -10,  10,  15,  -4,  10,  15,  22,  31,  20,  -8,   1,  -8,  -9, -14, -19, -27, -16,  12,  16,   5,  18,  22,  28,  29,   7, -22, -10,  -3, -18, -20, -18, -15,  14,  34,  24,   5,   5,  -1,  -2, -13, -39, -36,  17,  18,   0,   0,   0,   4,  11, -13, -26,   8,   2,   1,  -1,  -1,  -1,   6,  -7,  -4 },
     {    4,   1,  17,  21,   0,  -3,  -9, -11,   0,  -3,  -7, -31,   6,  28, -13, -13,  -1, -17,   3,  12, -10,  -9,  -6,   6,   7,  17,  33, -12,  23,   8, -20,  14,  15, -12,   4,  12,   5,   6,  40, -13,   1,  -9,  16, -55,  17,   1,  10,  10,   3,   0,  28,  -5, -37,  20, -11,  14, -11,   7,  -5,  16, -11,  -2,  10,   1 },
     {   -3,  -5,  -6, -26, -17,  30,  16,  -9,   0,  -1,   2,   6,  21,  -9,  15,   3,   9,  14,  36,  27,  23,   7, -38,  -4,  15,  21,  16, -26,  18,  -1,  54, -30,  -4,  -1, -16,  -6,  -5, -20,  21,  31,  -5,  10,  -8, -12,   0,  -1, -19,  12,  -2,  -5,   8,  -5,  -5,   3,  -4,  -1,   0,  -2,  -1,   0,  -4,   3,  -1,  -1 },
     {   -7,  -2,   3,  -3,   3,  -1,  -3,   4,  -8,   1,   6,   8, -14,  -1,  14, -19, -11,  -4,   9,  -9,   1,   1,  -3,  -1, -18,  -3,   3,   9,  -6,  -3,  -5,  -2,  -7,   2,  35,  -8,  -3,  -1,   2,   2, -86,  56,  29, -12,  -9,  -5,  -4,  -4,   1, -17,  20,  -3,   1,   2,   1,  -3,  15, -19,  -6,  14,  -2,   4,   0,   7 },
     {   -1,  -2,  -4,   0,  -2,  -2,  29, -22,   0,   0,   3,   8,  23,  28, -43,  10,   0,   1,  -2,   4,  31, -11, -18, -23,   4,  -1,  -9,  41,   5, -34,  13, -13,   4,   6,  34,  30, -32,  21, -21,   1,  -7, -17,  24,  -6,  19,  -5,   8,   8,   9,   2, -14,  -9,  10,  10,  -7,   3,   8,   3,  -6,  -3,  -1,   9,   7,  -6 },
     {   -5,  -9,  -6, -19, -27, -36, -24,  57,  -4,  -7,  -8, -17, -15, -29, -39,  33,  -5, -13,  -5, -16,  -3,  10,   4, -47,   2,   3,  -2,  -7,  16,  11,  19, -19,  -6, -16,  -6,   9,  13,  11,   5,   6,  -9,  -4,   7,   1,   7,   4,   5,   0,   3,  -9,  10,  -3,  -2,  11, -19,   0,   1,  -2,  -2,   2,  -6,  11,   4, -10 },
     {    4,   2,   7,   7,  16,  -3, -11,  -4,   0,  -9,  -7,  -9, -18, -23,  27,   3,   2,  -4,   4,   7,  16,  -5,  -3,  -2,   1,  -6,   0,  -3, -19, -19,  29,  -6,  19,   1,  14,  15,  17,   1,  -9,  -2,  -6, -22,   4, -23, -13,  12,  51,   5,   1, -10,   3,  -9,  -2, -59, -12,  42,  -8,   5,  -1,  -2,   3,  10, -32,  -6 },
     {    1,   2,   3,   6,   2, -16,  26, -14,  -1,  -1,  -2, -15,   0,  19, -32,  21,  -1,   1,   0,   7,  28, -21,  20, -15,  10,  13,  18,  16,  -3,  12,   6,  30,  -9,  10,  10, -22,  20, -27,  29,  -3,  15,  38, -29,  22, -11,  16,   9,   1,  31, -29,  17, -11,  15, -15,  -2,   1,   3,  -7,  -1,  -3,  10,  -3, -19,   1 },
   }
 },
 {  //29
   {
     {  101,  -3, -18,  -5,  -5,  -2,  -2,  -1, -69,  17,  16,   1,   2,   0,   1,   0,   2, -19,  -1,   3,   1,   1,   0,   0,   3,   7,  -4,  -2,   0,   0,   0,   0,  -2,  -1,   3,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -36,  54,  -3,  -6,  -1,  -2,   0,  -1, -57, -55,  22,  10,   5,   3,   2,   1,  61,  -4, -28,  -3,  -2,   0,  -1,   0,  -8,  19,   8,  -4,  -1,  -1,   0,   0,   4,  -9,   0,   3,   0,   0,   0,   0,   0,   4,  -1,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {   21,  77, -23, -12,  -4,  -4,  -1,  -2,  30, -42,  24,   4,  -1,   1,  -1,   0, -68,  -3,  -2,   9,   5,   1,   2,   1,  31,  -8,  -7,  -6,   0,   0,   0,   0,  -7,   9,   4,   0,  -2,   0,   0,   0,   2,  -3,   0,   1,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -47,   1,  -5,   9,   2,   2,   1,   1, -25,  40,  19, -13,  -1,  -3,   0,  -1, -12, -79,  10,  21,   4,   4,   1,   1,  37,  18, -42, -10,   0,   0,   0,   0,  -4,  14,  19,  -7,  -4,  -2,   0,   0,   4,  -6,  -4,   6,   1,   0,   0,   0,  -1,   1,   0,  -3,   1,   1,   0,   0,   2,  -1,   0,   0,  -1,   1,   0,   0 },
     {   16, -28, -20,   6,   5,   1,   1,   0,  45, -10,  24,  -9,  -7,  -1,  -1,   0,  10,  -6, -50,  14,   3,   2,   0,   1,  -4,  68,  13, -35,  -4,  -5,  -1,  -1, -14, -23,  39,  20,  -5,   0,  -1,   0,   0,  -3, -19,   6,   6,   1,   0,   0,  -2,   1,   5,  -3,  -1,   0,   0,   0,   0,   0,  -2,   2,   0,  -1,   0,   0 },
     {   -7,  11, -64,  17,   7,   1,   1,   1,  16,  27,  56, -34,  -9,  -1,  -1,  -1,  14,   5,   3,  28,  -3,  -3,  -1,   0, -52, -34,   0,   2,  13,   4,   2,   1,  28,  -2, -16,  -7,  -6,   1,   0,   0,  -7,  10,   5,  -2,   0,  -2,   0,  -1,   3,  -4,   2,   2,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   0 },
     {   18,  43,  42, -20, -11,  -2,  -3,  -1,  27,  20, -30,  14,   2,  -3,   0,  -1,   0, -55,   0,   5,   7,   6,   0,   1, -62,  20,   4,  -5,   0,   0,   2,   0,  42, -14,  -5,   0,  -3,  -1,  -1,   0, -12,  12,   2,   0,  -1,   0,   0,   0,   1,  -5,   2,   1,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0 },
     {   20, -17,  12,   5,  -3,  -1,   0,   0,  15, -48,  -8,  -6,   6,   3,   1,   1,  24,   6,   9,  20, -11,  -1,  -3,   0, -18,  -5, -59,  -6,  18,   3,   3,   1,  16,  47,  33, -34, -22,  -3,  -2,  -1, -11, -15,  12,  35,   5,  -4,  -1,   0,  -1,   2, -14, -12,   7,   4,   1,   1,  -1,   0,   5,   1,  -4,   0,   0,   0 },
     {    6,  42, -16,  25, -15,   0,  -2,   0,   7,  16, -31, -36,  22,  -1,   2,   0,  20,  17,  52,  -4, -30,  -1,  -3,  -1,  -3,  28,  -7,  21,  15, -10,  -1,  -1, -33, -39,  28,  -9,   5,  10,   0,   1,  13,  -6, -17,  13,  -7,  -1,   2,   0,  -3,   5,   4,  -5,   4,  -2,   0,   0,   1,  -2,   0,   4,   0,   0,   0,   0 },
     {   -6,  15,  21,  58,  -2, -11,  -3,  -2, -28,  13, -19, -47,  12,  12,   2,   2, -31,  14, -34,  -8, -12,  -4,   4,   1,   5,  -3,  41, -17,  -1,  -7,  -3,  -1,  38,  20,  19,  10, -12,   0,  -3,  -1, -19, -16,   6,  22,  10,  -1,   1,   0,   3,  -8, -13,  -4,   3,   4,   1,   1,   0,   3,   0,  -1,  -1,  -2,   1,  -1 },
     {  -15,  21, -21, -47,   5,   6,   1,   2, -14,  46,  -6,  37, -18,  -8,  -1,  -2,   2,  37,  11, -13,  20,  -4,  -2,  -2,  -7,  -8,   6, -28,  -5,   8,   3,   0,   6,   2,  49,  11, -21,  -9,  -2,  -1,  -2, -24,  -2,  36,  13,  -1,  -2,   0,  -2,  -1, -18,  -7,   9,   5,   3,   1,   1,  -2,   3,  -3,  -3,  -1,   1,   0 },
     {    0,  -6,  15,   2,   0,   0,   0,  -1,  -7, -16,   3,  -1,   2,   2,   0,   1,  -2,  -4,  22,  21,  -1,  -2,   0,  -1,  -3, -28,  -5,  -4,   7,   7,   2,   1,   4, -12,  -1,  47, -10,  -2,   0,   0, -17,  -1, -66, -12,  47,   3,   5,   1,   7,   6,  12, -55, -21,  13,   1,   1,   3,   2,   7,  15, -14,  -6,  -1,   0 },
     {  -13, -26, -50, -23,  10,   9,   2,   3, -18, -33, -28,  -2,  -3,   1,   4,   0, -25,  -1,  47,   4, -11,  -6,  -4,  -1,   9,  49,  14,   2,   0,  -6,  -2,  -3,  59,  -3, -10,   5,   2,   2,  -1,   1, -27,   8,   6,  -3,   0,   2,   1,   1,   2, -10,   1,   4,  -1,   0,   1,   0,  -1,   2,  -2,   0,   2,   0,  -1,   0 },
     {   16,  18, -24,   1,   3,  -3,   0,   0,  33,   8, -13, -11, -10,   2,   0,   0,  50, -26,  12, -17,   6,  -5,  -1,   0,  22,   6,  25,  21, -26,   5,  -6,   0, -13,  56, -14,  26,  15, -12,   3,  -3, -28, -33,  12,  -8,  21,   9,  -1,   1,   1,  -6, -15,  -5,  -8,   6,   3,   0,  -1,   3,   3,  -4,  -2,  -4,   1,   0 },
     {    6,  13,  15,  10,  30,  -7,  -6,  -1,  -8,  -7, -32, -32, -49,  11,   9,   2,  11,  12,  20,  46,  37, -21, -12,  -3,   7,   3,  -3, -37,  -3,  31,   7,   1, -17, -10, -13,   5, -18, -19,   3,   2,  13,  13,  17, -11,   4,   2,  -7,  -2,  -3,  -1,   7,  20,  -7,  -4,   2,  -1,   0,  -1,  -6,  -2,  11,   0,   0,   1 },
     {    0,  -3, -31,  43,   2,   3,  -6,   2,   1, -15, -15,  11,  23,  -3,   5,  -2,  -5, -12,  11, -54,  18,  24,   1,   3, -20,  -9, -24, -24, -52,   9,  15,   5,   1,  -4,   2,   9, -18, -37,  -2,   1,  21,  23,   3,   1,  12,  -7, -15,  -3,   0,   9,  14,   2,   2,   5,  -1,  -4,  -1,  -3,   2,   5,   2,   1,   1,  -1 },
     {   -2,  -4,  -1, -16, -67,   5,  23,   1,   3,   6,  -3,  -7,  59, -13, -23,   4,  12,  13,   9,  42,   2,  11,   3, -10,  16,  -5,   6, -44, -11,   6,  11,  10,  -4,   1, -20,  15,  -6,  -8,  -8,  -1,  -8,   3,  21,  -4,   0,  -2,   0,  -4,   4,  -6,  -5,  11,   0,   1,  -2,  -1,  -2,   1,  -3,  -2,   3,  -1,   1,  -2 },
     {    0,  -3, -12,  -7,  -3,  -9,   2,   0,   6,   2,  -8,   9,  11,  17,  -5,  -1,   3, -16,  -6, -13, -13, -22,  15,   5,   9,   3,  17,  -3,  28,  26, -24,  -2, -11,  -3, -21, -10, -52,  10,  45,   4,   1,   4,   1,   6,   9, -58, -28,  16,   0,   2,   6,  -2,   7,  25, -18, -22,   1,   1,   0,   0,  -3,  -2,  10,  -1 },
     {    5, -10,  44, -28,  14,  -4,   0,  -1,   7, -14,  58, -39,   5,   2,   1,  -1,   4,  -7,  43, -20, -12,   7,  -5,   4,   6,  -1,  17,   8, -30,  10,  -7,   3,  12, -14,   3,  26, -22,  -5,   0,  -2,  11,  -1,   1,  32,  -8,  -4,  -6,  -5,  -6,   8, -15,  22,  15,  -2,   5,  -4,  -3,  -3,  -9,  -5,   9,   5,   1,   2 },
     {   -9, -12, -16, -14,  -2,   4,   4,  -5, -12, -19, -16,  -8,  14,   4,  -1,   3, -18, -22,   0,  23,   4,  -9,  -3,   0, -39,  -2,  39,   0,  -5,   5,  -3,  -6, -14,  25,  -9,  -8,  12,  -4,   1,   3,  61, -42, -15,  12,  -6,  10,  -4,  -3, -28,  40,  -3,  -9,   0,   0,   6,  -1,  12, -12,   4,   0,   1,  -1,  -2,   3 },
     {    2,   2,  -1, -11,  -7,   0,  -3,  -1,   5,   4,  -7,  -6,   1,  -7,  -2,   5,  12,   3,  -4,  -1, -12,   1,   6,  -1,  10,  -7,   6,  11,  13,  -7,  -9,   0,  16,  34,  26,  14, -11,  -1,  -4,  -6,  44,  45, -14, -16,  17,   4,   4,  -1, -61, -50,  22,  15,  -1,   4,   5,   6,  17,  -1, -21,  -6,   3,  -3,   1,  -3 },
     {   -2,   3,   5, -12,  -3,  -1,   1,   1,  -3,   4,   5, -10,  12,   1,  -2,   6,  -2,  10,  -1, -21, -12,  -8,   8,  -2,  -8,   0, -16,  -6,  14,  15,  -5,   4,  15,  15,  28,   8,  -9,  -8,  13,  -2,  -8, -34, -10, -48, -10,  10,  -7,  15,   9,  26,  27,  37, -49, -25, -11,  -9,  -6,  -9,  -7,  19,  43,  -4,  -6,  -6 },
     {  -11,   5,  17,  -6,   1,  -4,   9,  -4, -19,  13,  21,  -5,  -2,   8, -10,   5, -33,  32,   1,   7, -11,   2,  -1,  -3, -44,  51, -26,  24, -22,   2,   3,   2, -32,  26, -19,  12,   4, -20,  17, -11,  -1,  -9,  19, -10,  29, -16,   5,   6,   4, -30,   6, -10,   1,   8, -11,   2,  10,  -6,  -2,   5,  -3,  -1,   4,  -2 },
     {   -2,  -4,   5,  12, -46,  -8, -15,  10,  -3,  -4,  10,  -5, -16, -38,   4,  -6,  -2,   9,  -1,   6,  49,  28, -25,  -2,   9,  15,  16,  47,   4,  27,  28,  -7,  15,   1,   8, -24,  -9, -16,  17,  24,  -8, -10, -26,   0,  -3, -11, -15,   8,  -6,  -4,   3,   2,  -1,  -1,  -4,  -7,   3,   1,   0,  -5,   2,   4,   0,  -1 },
     {   -5,  -5, -13, -53, -20,   3,   4,  -5,  -7,  -7, -31, -53, -26,  -5,   7,   7,  -5,  -9, -37,   0, -20,   8,   3,   3,  -3, -27, -12,  20, -18, -28,   2,   0, -13, -12,  17,   4,   8, -26,  -5,   6,  -9,  16,   2,   3,   9,  -9, -28,  -4,  21,   9,   9,   5,   7,  13,  -7, -13,  -4,   6,   4,   1,  -1,   5,   7,   1 },
     {    6,   8,   9,  10,  19,  84,  27,  -6,   1,   0,   1,  -2,   0, -60,  10,  28,  -3,  -5,  -8, -10, -26, -16, -31, -12,   2,   1,   3,  -6,  15,  14,   8, -11,  -3,   3,  -9,   4,  -6,  -8,   5,  10,   3,   0,   5,  -1,   7,  -1,  -4,  -2,   3,   0,  -5,  -1,   2,   0,  -4,   1,  -3,   3,   2,  -4,  -4,   1,  -1,  -3 },
     {    2,   2,   4,  34, -17,  -7,   3,   4,   2,   2,   5,  48, -27,   4,   3,  -5,   4,   2,  13,  34, -36,  -6,   0,  -5,   2,  -1,   4,   8,  -2, -32,  -5, -11,  12,   8,   5,  23,   4, -24,  -1,  10,  13,  -7, -12,  -5,  19, -16, -15, -15,  19,  21,   0,  33,  33,   0,  -4, -10, -10,  -3,  -6, -21,  26,  23,   3,   4 },
     {   -8, -13, -10,   2, -10,   4,  -5,   0, -10, -19,  -6,   5,   4,   5,  15,   1, -16, -26,  13,  -1,  11,  13, -16,  -1, -33, -15,  27,   6,  14,   0,  -7,  -5, -50,  15,  29,  15, -33,  23,  -1,   3, -16,  15,   4,  -5,  -2,  -1,  25, -12,  36, -29, -13,  18,  -5, -21,   7,  12, -11,  18, -10, -16,  15,   5,  -8,   2 },
     {    2,   0,  -2,   7,  -5,  12,   9,   3,  -2,  -9,  -9,  -9,  -8,  -2,  -8,  -3,   2,   0,   3,   5,  15, -11,  -9,  -1,  -9, -15, -15, -10, -29,   9, -10,   0,  11,   7,  22,  30,  38,  29,  27, -13,   2,   0, -15,  -9, -38, -35,  25,  36,  -6,  -6,  -9,  -4,  24, -14, -44, -15,   9,   6,   2, -12,  -8,  18,  15,  -7 },
     {   -1,  -5,  -2,  11,   5,   4,  19,   1,  -4, -10,  -3,   2,   6,   3, -17,   1,  -8,  -4,  -2,   2,   7,   3,  15,   4, -10,  -6, -16,  12,   9,  16, -25,   5,   0, -34,  12,   1,  12, -21,  -4,  -8,  -2, -19,  -5, -22,   8,  16, -21,  20, -13, -19, -56,  21, -15,  21,  24, -20,   7,  20,  -2, -48,  -7, -20,  10,  19 },
     {    2,   2,   0,   3,  16,   5,   5,  -7,  -2,  -5,  -7,  -7, -23, -30, -24,  19,   5,   0,   6,  -1,  16,  46,  32, -20,   0,   3,   6,  -1,   6, -38,  -8,  36,  -1,  -1, -20, -14, -35,  10, -26, -31,   5, -25, -15,  -2,   9, -14,  24,  12,  -1,   8,   1,  15,   1,   7, -30,   5,  -1,   4,   0,  -4,   2,   1,  10,  -7 },
     {   -3,  -6,   8,  14, -44,  10, -21,   3,  -4, -11,  18,  -5, -19, -29,   8,   1,  -6,  -2,  21, -23,  34, -42,  23,   5, -14,   6,   0, -14,  22, -21, -32,   2,  -9,   4,   3,  13,  14,  15, -11, -12,  -2,   2,  26,   7,   7,  25, -18,   8,   3,   5,  23,   0,   3,  27,   2, -15,   4,  -2,   1,  13,  -9,   9,  16,  12 },
   },
   {
     {  110, -39,  -7,  -4,  -2,  -1,  -1,   0,  19,  15, -12,   0,  -3,   0,  -1,   0, -28,  27,  -1,  -2,   0,  -1,   0,   0, -12,   4,  10,  -3,   1,  -1,   0,   0,  -5,  -3,   6,   1,  -1,   0,   0,   0,  -2,  -2,   2,   2,  -1,   0,   0,   0,  -2,   0,   0,   2,   0,   0,   0,   0,  -1,   0,   0,   1,   1,  -1,   0,   0 },
     {   25,  20, -11,   0,  -3,   0,  -1,   0, -97,  46,  12,  -1,   3,  -1,   1,   0, -21, -30,  25,  -1,   3,   0,   1,   0,  21, -29,  -1,   6,  -1,   1,   0,   0,   7,  -3, -10,   3,   0,   0,   0,   0,   4,   1,  -5,  -2,   1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0 },
     {  -28,  48, -17,  -1,  -3,  -1,  -1,   0,  25,  54, -24,  -8,  -4,  -3,  -1,  -1, -35,  51,  18, -17,   2,  -4,   1,  -1, -13, -16,  45,  -9,   0,  -1,   1,  -1,   8, -29,  20,  12,  -6,   2,  -1,   0,   2, -11,  -4,  14,  -1,   0,   0,   0,   0,  -2,  -5,   5,   3,  -3,   1,   0,   0,  -1,  -2,   1,   3,  -2,   1,   0 },
     {   27,  52, -36,  -1,  -6,   0,  -2,   0, -13,  13,  -3,  -7,   1,  -2,   1,   0,  58, -10, -17,  -8,  -2,  -1,  -1,   0,  -5,  55,  -6, -19,  -1,  -3,   0,  -1, -12,  10,  42, -17,  -4,  -2,   0,  -1,   2, -15,  25,  14, -11,   1,  -2,   0,  -2,  -5,  -1,  18,   0,  -4,   0,   0,  -2,   0,  -3,   5,   6,  -5,   1,   0 },
     {  -28, -49,  21,   3,   6,   2,   2,   0, -60,  -8, -15,  14,   2,   4,   0,   1, -23,  44, -44,  -2,   3,   0,   1,   0,  -3,  28,   7, -32,   6,  -4,   2,  -1,  15, -10,  33, -16,  -7,   1,  -1,   0,   9,  -7,  11,  10, -11,   1,  -1,   0,   2,   0,  -2,  10,  -1,  -3,   0,   0,   1,   0,  -1,   1,   3,  -2,   0,   0 },
     {   18,  11,  14,  -8,  -1,  -2,   0,  -1, -25, -47,  40,   2,   1,   0,   1,   0,  38,  13,   1,  -3,  -8,  -1,  -2,   0, -24,   0,  50,   8,  -7,   0,  -1,   0,   9, -46,   2,  41,  -8,   3,  -1,   1,  -3,   4, -35,  23,   8,  -4,   1,   0,  -2,   4, -16,  -5,  12,  -2,   1,   0,   0,   2,  -3,  -6,   3,   1,   0,   0 },
     {    5,  47, -13,   1,  -2,  -1,  -1,   0, -15, -26,  18,  11,   4,   1,   1,   0, -47,  14, -19,  25,   2,   3,   0,   1, -63,  21, -22,  12,  15,  -1,   3,   0, -15, -12, -15, -16,  23,  -1,   4,  -1,   5, -10,   0, -31,  12,   5,   1,   0,   9, -10,   8, -21,  -8,  12,  -1,   1,   5,  -2,   4,  -3, -14,   9,  -2,   0 },
     {   -4,  -2,  32, -11,  -2,  -2,   0,  -1,  16,  -5,  52, -17,  -6,  -3,  -1,  -1, -48, -32,  35,  17,  -9,   4,  -1,   1, -12,  14, -15,   7,  -5,  -1,   0,   0,  -3,  -2,  39,  -7,  -7,  -3,  -1,   0,  14, -26,  14,  34, -17,   1,  -3,   1,  -2,   3, -20,  33,   4, -10,   1,  -1,  -2,   3, -12,   4,  17,  -9,   2,  -1 },
     {   18,  49, -15,  -8,  -2,  -3,  -1,  -1,   8, -57,   7,   8,   1,   3,   1,   1, -43,   5, -30,  -3,   6,   0,   2,   1,  59, -21,   3, -28,  -1,   0,  -1,   0,  27,  17,   3,  -2, -18,   2,  -3,   1,  -9,  20,   6,  11,  -7,  -5,   0,  -1,  -6,   3,   2,   7,   3,  -5,   0,  -1,  -3,   0,   1,   1,   4,  -3,   0,   0 },
     {   -9, -23, -59,  40,   1,   6,  -1,   1,  -5, -40, -31,  10,  18,   0,   2,   0, -22, -43,  27, -30,  20,   2,   3,   0, -32,  -5,  20, -18,  -8,  10,  -1,   2, -18,   1,   8,   5, -17,   5,   1,   1,   8,  -7,  -4,  10,  -7,  -4,   3,   0,   8,   0,  -8,   2,   1,  -4,   0,   0,   3,   4,  -4,  -2,   2,   0,  -1,   1 },
     {    3,   5,   6,  -7,   0,   0,   0,   0, -11, -22,   3,   1,  -2,   2,   0,   0,  18,  26,  13, -20,  -3,  -2,  -1,   0, -23, -32,  20,  16,  -6,   4,  -1,   0,   1,   5, -35, -21,   4,  -1,   1,   0,  15,   4,  48, -29, -31,   2,  -4,   1,   7,   4,   9,  54, -29, -13,  -1,  -1,  -3,  11,  -9,  22,  24, -21,   2,  -2 },
     {    6,  -1, -14,   4,   3,   1,  -1,   0,  17,  11,  -7,  -2,   4,  -1,   0,   0,  15, -26, -19,  23,   6,   3,  -1,   1, -16, -28, -23, -13,  27,  -1,   4,  -1,  46, -58,   6, -39,   4,  13,  -3,   2,  15,  12, -26,  -1, -34,  11,  -2,   2, -10,  33,  -9,   3, -11, -12,   5,  -2,  -1,   7,   8,  -3,   5,  -8,   0,   0 },
     {   -6, -10, -14,   6,   2,   0,   1,   0,  10,  17,  26, -15,  -8,   0,  -1,   1, -15, -24, -24,  16,   3,   0,   2,   0,  18,  28,  20, -30,  -9,  -1,  -3,  -1, -15, -21,  -1,  42,  -1,  -1,   1,  -1,   0,   4, -23, -46,   6,   2,   1,   1,   9,   5,  36,  17, -42,  -8,  -1,  -1,  -2,   8, -10,  32,  20, -17,   0,  -2 },
     {   -8, -19, -52,  23,  -2,   5,  -2,   2,  -4, -12,  22, -39,  11,  -5,   5,  -1,  -6,  35,  30, -12, -27,   5,  -3,   1,  19,  33, -21,  33, -18,  -8,   0,  -1,  51,  -7,   1,  -2,  19, -16,  -1,  -2,   2,  16,   1,  -4,   7,   5,  -6,   1, -11,   8,   7,  -2,  -2,   2,   0,  -1,   0,  -4,   4,   4,  -1,  -2,   2,   1 },
     {    4,  -5,  16, -31,  10,   1,   2,  -1,   4, -13,  18, -39,  15,   2,   3,  -1,   7,  -9,  22, -37,   6,   6,   0,   1,   7, -12,  17, -30,   6,   6,   0,   1,  10, -13,  23, -27,  -6,  11,  -1,   1,   7, -17,  13, -21,  -2,  13,   0,   1,  11, -22,  17, -31, -29,  28,  -3,   3,   4,  -4,   6,  13, -50,  16,   0,   0 },
     {   -5,   2,   6, -25,  -3,   4,   3,   1,  -8,  12,   4, -39,  -1,   3,   3,   0,  -7,   5,  -2, -15, -13,  11,   0,   1, -42,  11, -19, -15, -15,  12,   2,   1,  -8,  21, -18,  -3, -47,  10,   2,   1, -19,  52, -18,  16, -26, -23,  14,  -3,   1,   8,  25,  -9,   9, -27,   6,   0,   9,  -9,  21,   7,   0,  -4,  -4,   1 },
     {   -5,  -6, -22, -62,  19,   3,   7,   1,  -8, -14, -47, -43,   1,  18,   4,   4,  -8, -22, -34,  10, -28,  11,   7,   2,  -5, -23,  -3,  24, -23, -14,   7,  -1, -11, -16,  11,  21,  11, -22,  -2,  -1,   3, -15,  12,   4,  13,  -2,  -7,   1,  -1,   4,  -8,   9,   1,   2,  -2,  -2,  -1,   4,   1,  -5,   4,   2,   2,   0 },
     {   -6,  -8,  -8,  10,  15,  -5,  -3,   0,  -7,  -7,   7,   6,  12,  -6,  -1,   1,  -1,   0,   9,  24,  -7,   1,  -4,   0,   2, -21,  16,  18,  15,  -7,   0,  -1, -35, -16,  39, -11,  15,   9,  -6,   4, -75,  26,  18,  -2, -14,  14,   2,   0, -39, -11,  24,   4,  -3,  -4,   6,  -2,  -2, -24,   6,   6,   0,   2,   0,   1 },
     {    1,   7,   4,   1,  -4,   2,   0,   0,  -7, -17,  -6,   0,  -3,   1,   3,   0,  -1,  26,  15, -12, -10,   2,  -5,  -1,  37,   1, -10,  12,  -4,   1,  -2,   0, -70, -14,  11, -14,   9,   4,   4,  -3,  35, -11, -27,  -4, -20,   3,   2,   1,   1,  55,  -1, -13,  -5, -12,   0,   0,  -5,  -8,  38,   7,  -4,  -6,  -3,   0 },
     {    9,  11,  19,  34,  -4,   0,   1,   1,  -9, -21, -32, -26, -14,   2,   6,  -2,  11,  10,  13,  30, -12, -27,   1,   0, -13, -26, -17,  -3,  12,  -1,  -8,  -4,  22,  26,  24,  21,   1,   5,  -9, -12,  -4, -30, -22,  11,   4,   5,  10,  -5,  21, -11,  22,   4, -12,   4,   5,  -1,  24,  -6,  32,  39,  11,  -5,  11,   1 },
     {    6,  17,  38, -18,   6,  -6,  -4,  -2,  -3, -11, -53,  21,   2,  10,  -2,   0, -16, -31,  30, -11, -14,   4,  -4,  -3,   5,  57,  29,  14,  -2, -10,  -9,  -2,  27,  -2,  -8,  -7,  23,  13,  -9,   0,   2,  21, -16, -11, -11,  19,   0,  -4, -11,   3,  12,   3,  -1,  -2,   9,  -5,  10, -17,  -1,   5,   8,  -3,   3,   1 },
     {  -10, -17, -47, -57,  18,   7,   3,   3,  -3,  -5,  11,  11, -23,   0,  -8,   6,   7,  14,  33,  52,   9, -28,  -3,  -3,   5,  10,  27,  -9,  29,  -8,  -3,  -5,   2,  33,  -5, -11,  -1,  20,  -3,  -3,  12,   2, -14,  -1, -11,  -4,   3,  -5,   8,  -2,  -7,  -1,  -1,  -5,  -4,   4,   7,   6,   4,  -4,  -2,   6,  -3,   3 },
     {    2,   3,  -3,  -8,  55, -37,   0,   1,   3,   5,  -2,  11,  51, -25,  -7,   3,   2,  11,  -8,   9,  28,   6, -11,   0,   9,   7, -15,  34, -17,  21,  -3,   5,   9,  -2,  -4,  23, -25,  23,   5,  -5,  16, -27,  -3,   5, -23,  11,  18,   5,  -1, -18,  18,   1,  -4, -12,  -2,   4,   0,  -7,   7,  -4,   1,  -2,  -4,   3 },
     {   -1,   2,   2,  -3,   1,  11,   4,   4,  -3,   5,   7, -12,   8,  11,  -3,  10,  -5,   3,  -9, -14,  17,   2,  -5,   1,  -2,  12,   7,   1, -13,   3,  -5,  -4,  10,  24, -16,  -2,   6,   1,  -3,  -7, -66, -41, -13, -11, -13,  29,   7,  -6,   0,  29, -51,  -7, -26, -12,  11,  -2,  17,  25,  18, -14,  -2, -17,  -4,   8 },
     {    1,   1,   1, -10,  -3,  -3, -10,   0,  -2,  -4,  -4,   7,  -3,  -2, -21,  -8,   7,   7,  23,  21,   8,  18,  -8, -14,  -4,  -1,  -7, -34, -34,  34,  23,  15,  13,  -5,   1,   2,  -2, -13,   6,  22, -20, -19,   2, -29,   1,  -7, -13,  14,   6,  30,   8,  22,  20,  22,  -6, -17,   3, -30,  19, -16,  21,  43,  13,  -9 },
     {   -1,   0,  -1,  -4,  21,  -4, -12,  -3,   3,  11,  22,  22, -14, -21,  -2,  -5,   0, -10, -19, -18,  22,  -5,   9,  -2, -30, -19,  12,   7, -31, -19,   2,   6,  14,  41,  38,  13,  20, -10, -18,  -5,  27,  34,   1,   0,  15,  29, -11,  -2,  -9,  29,   4,   0, -13,   2,   1, -10,  -3,  -4,  33,   0, -11,   4,   4,   3 },
     {    1,   2,  -1, -19,   9,  -2,  18,  -3,   0,  -6,  -4,  30, -43,  27,   9,  19,   2,   4,  19, -29,  36, -33,  17,  -1,  -1,   3, -43,  11, -19,   3,  -7,  -8,   9, -31,  21,   3,  -6, -14,  15, -18, -21,   4,  -9, -10,   9, -21,  12,   0,  -3, -11,   1,  12, -17, -11, -15,  11,  -3,  -8,   0,   5,  -4,   8, -16,   9 },
     {    3,   5,  17,  -9,  22, -21,   7,   0,  -5, -10, -10, -30,  30, -45,  11,  -7,   0,   3,  18, -10,  34, -26,  22,  -6,  -4,   8,  -4, -24,  33, -23,  28,  -2,  -5,   2,  -8,  -1,  20, -32,   7,  -2,  -9,   4,   1,  -4,  19, -20,   4,   5, -16,  24,  -4,  -8,  13, -10, -13,   1,  -9,  15,  -8,   8,  25,  -4,   0,   0 },
     {    5,   7,   0,  14,  66, -14,  20,  -4,   4,   7,   9,  26,   3,   9,  16,   4,   4,   2,   8,  -5, -53,  30, -10,  13,  -3, -11,   6, -22,  -2,   1,  -2,  -2,   0,  14,   5, -27,   9, -32,  10,  -2,  -9,   0, -28,   4,  11,  -8,  -8,   3,  28,   0,  -2,   1,  -8, -14, -17,  10,  10,   3, -15,   1,   5,   2, -10,  -2 },
     {    0,   1,   4,  -6, -14,  17,   3,   2,   1,   1,  -1,  16,  24,  21,   3,   6,   2,  -4,   6,  21,  -7,   4,  -5,   8, -10,  -3,  20,   1,  -6,  24, -18,  -1,  16,  12,   8, -18, -29, -16,  19, -11,  14, -23,   7, -13,  42, -15,   3,   8, -65,  16,  16, -18,   1, -18, -12,   2, -13,  11,  17,  18,   1, -23,  -4,   2 },
     {   -3,   0,  -3, -14, -19, -20,   1, -19,  -4,   0,  -7,  12,  30,   7,   0, -19,   0,  11,  15,  -4, -30, -18, -13,   8, -10,  -4, -32, -38,   0,   1,  -8,  12,  -7,   9,  -9,  28,  18,  15,  -7,  -7,   0,   8, -18,  13,   1,   7,  -5,  15, -32, -11, -19,  27, -43,   2,  -3,   9, -35,   7,  -1, -14, -20,   9,  15, -13 },
     {    0,  -1,  -2,   9,   0, -24,   5,  12,   1,   1,  14,  -9,   4,  52, -35, -18,   1,   2,  -7, -20,  -7,  -4,  51, -15,  -1,  -6,   3,   9,  17, -15,  12,   9,   0,  16,   9,   1,   5,  29,  23, -16,   5, -13, -11,  -9,  -6,   3, -25,  25, -13,   3,  21,  -3,   9, -23,  -7, -10,  22,   7, -13, -21,  12,  17,  13,   1 },
   },
   {
     {  113, -39, -16,  -1,  -3,   0,  -2,   0,  -7,  33, -12,  -3,  -1,  -1,   0,   0, -17,   9,  11,  -4,   0,   0,   0,   0,  -3,  -2,   4,   3,  -1,   0,   0,   0,  -5,   1,   1,   1,   1,   0,   0,   0,  -1,  -1,   1,   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -10, -69,  29,   4,   6,   1,   2,   1,  61, -40, -41,  18,  -3,   4,  -1,   2,  -5,  42, -29, -13,   5,  -2,   1,  -1,  -4,   8,  12, -12,  -2,   0,   0,   0,  -1,   5,   2,   1,  -3,  -1,  -1,   0,  -2,   2,   3,  -1,   1,   0,  -1,   0,  -1,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {  -38,  -2, -13,   8,  -1,   4,   0,   2,  15,  63, -52, -22,   6,  -4,   2,  -2,  -7,  21,  54, -41,  -5,  -1,  -1,  -1,   8, -23,  24,  21, -18,   2,  -1,   0,  -1,   0,  -4,  10,   3,  -6,   1,   0,   1,  -5,   2,   2,   2,   0,  -2,   0,   0,   0,   0,   1,   1,   0,   0,  -1,   0,  -1,   1,   0,   0,   0,   0,   0 },
     {   -4, -31,  19,   4,  -4,   0,   0,   0,  48,   3,  26, -28,  -6,  -5,  -2,  -1,  10, -19,  43,  41, -34,   2,  -6,   1, -10,   4, -39,  51,  19, -15,   3,  -2,  -5,   4,  -9, -17,  26,   3,  -2,   1,  -1,  -2,  -3,  -2,  -2,   7,   1,   0,  -2,   1,  -1,  -2,   1,   1,   1,   1,   0,   0,  -1,   0,  -1,   1,   0,   1 },
     {   26,  70, -48,  -9,  -3,  -3,  -1,  -1,  80, -19,   4, -17,  -5,   0,  -1,   0,   4,  21,  -3,  -1,  -7,  -2,   0,   0, -13,   8,  16,  -3,  -1,  -3,  -1,   0,  -4,  -5,   6,   9,  -2,  -1,  -1,   0,  -4,   0,   1,   2,   5,  -2,   0,   0,  -1,  -2,   1,   1,   1,   1,  -1,   0,  -1,   0,   0,   0,   0,   0,   1,   0 },
     {   15,   4,  20, -22,  -4,   1,   0,   0, -13, -38,  10,   4,  -4,   0,   5,   1,  23, -14,   6, -38, -17,  11,  -3,   5,   3,  12,  11,  38, -64,  -8,   5,  -1,  -4,  12, -14,  32,  24, -44,   3,  -1,   2,  -1,   5,  -7,  23,   4, -16,   4,  -1,   1,  -3,   3,   0,   6,  -2,  -2,   0,   0,   0,  -1,   2,   1,   0,  -1 },
     {    3,  17,   1,   2,  -4,   3,  -8,  -1, -18,  -6, -18,  20,  11, -12,   2,  -3,   4,  19, -10, -13,  -4,   2,  -9,  -2,   0,  14,  14,  20,  15, -39,   0,  -7,   9,  -8,  12, -11,  57,  24, -42,   1,  -1,   2,  -1,  -1, -18,  55,  17, -23,   1,  -2,  -1,  -3,  -2, -11,  29,   8,   0,  -1,   0,  -2,  -2,   1,  -4,   8 },
     {    5,  26,  63, -42,  -2,  -8,   0,  -3,   8,  18,  26,  24, -32,   0,  -6,   0, -48,  32,   9,  20,  10, -17,   2,  -3,   7, -34,  24,   9,  10,   2,  -6,   1,  -2,  -6, -19,  18,   6,   4,   0,  -2,   0,  -4,  -5,  -6,   9,   3,   2,   0,  -1,  -1,  -4,   0,  -1,   3,   1,   2,   1,  -1,  -1,  -1,   1,   0,   0,   1 },
     {   19,   7,  27, -10,   5,  -6,  -1,  -3, -16, -52,  -2, -28,  29,   7,   2,   1,  35,  12,  40, -24, -12,  -2,  -8,  -1,   3, -28,  22,  10,  21,  23, -12,   3,   3,  -5, -15,  -6, -32,  39,   2,   0,  -5,   7,  -6,  18, -21,  -8,  16,  -8,   1,   0,  -3,   0,   2,  -7,   6,   2,  -1,   1,   0,   2,  -2,  -3,   0,   2 },
     {    3,  25,   4,  20, -18,  -4,   1,  -1, -39, -13, -39,   7,  -5,  13,   4,   5,   2,  57,  -6,  20, -38, -14,   5,  -3, -25,  12, -11,  28,  24,  -8,  11,  -3,  -3,  -5,  -7, -25,   2, -21,  22,   4,   4,  -6,  12,  -7,  15, -26, -16,  14,   1,  -2,   0,  -5,   3,  10, -18,  -2,   1,   0,   2,   0,   0,   1,   0,  -5 },
     {   17,  32,  23,  19, -11,  -2,  -7,  -1,  24,   2, -37,  57,   8, -10,   3,  -4,  20, -35,  13, -12,  46,  -3,  -4,  -1,   3,  -7, -32,  13,   6,  30,  -3,  -1, -11,  15, -19, -25,   9,   2,  18,  -3,   2,  -5,  10, -14,  -6,   3,   0,   9,  -2,   2,  -2,  -2,  -4,   0,   0,   0,   0,  -1,   1,  -2,  -1,   0,   0,   1 },
     {   -7,  -7, -41,  40,   4,   4,   0,   0,  -7, -50,  14,   8,  22,   0,   2,   0, -79,  -1,  19,  -4,   8,  12,  -1,   1,  -1, -34,  -3,  17,  -5,   6,   5,   0,   2,   1, -27,  -1,  11,  -2,   4,   1,   1,   3,  -5, -17,   2,   5,   1,   1,   2,   1,   1,  -4,  -6,   1,   1,   0,   1,   1,   0,  -2,  -1,   0,   0,   1 },
     {    1,   0,   0,   0,   9, -15,  -2,   1,  -2,  -7,   0, -10,   0,   3, -12,  -3,   1,   0,   9,   0,   9, -22,  -5,   3,  -4,  -9,  -4, -10,  11,  10, -17,  -3,   7, -10,   6, -16,  -3, -24, -13,  10,   0,   6,   1,  11,   0,  28, -72, -17,   0,   8,  -1,  11,  -4,  29,  32, -61,   2,   0,   3,   0,   5,  -4,  23,  13 },
     {   -6, -23, -42,  22,   0,  -6,   1,   2,  -9,  -5,  13,  16, -32,   0,  -3,   0,  49,   8,   3,  17,  15, -43,   1,  -4,  18,  -8,  30,  14,  23,   8, -31,   1,  -6,  10, -26,  35,  15,  13,   6, -13,  -3,   0,  -3, -22,  26,   4,   7,   1,   1,  -1,   0,  -5,  -8,  12,   2,   3,  -1,   0,   0,  -3,   0,  -2,   3,   1 },
     {   -3,  -4,  -6, -14, -11,   3,  12,  -3,   5,   8,   9,  18,  14,  -9, -22,   1,  -9, -14, -18, -23, -16,  11,  17,   0,  11,  17,  21,  21,   5, -24, -30,   3,  -7, -14,  -6,  -6,  18,  42,  23,  -7,  -5,  -6, -10, -13, -37, -31, -15,  25,  -2,   4,   0,  10,   7,   6, -27, -45,   2,   0,   4,  -1,   3,   3,  22,  -2 },
     {   -4,  -5,  21,  42, -30,   0,  -7,  -1,  -1,   7,  44,   1,  10, -25,   6,  -7,  -5,  20,  23, -17,  29,  -7, -15,   5, -54,  47,  21,  -2,  -5,  22, -11,  -6,  -5, -23,  24,   9,   0,   1,  13,  -4,   0, -13,  -1,   9,   3,   2,   1,   6,  -1,  -3,  -5,   1,   2,   3,   0,   5,   0,  -1,  -3,   1,   0,   1,   0,   1 },
     {  -10,  -7, -22, -55,  59, -14,   2,   0, -12,   0,  -2, -13,   0,  18,  -5,  -6,  -7,  19,   2,   0,  19,   5, -12,   3, -26,  28, -13,  14,  -3,  38, -12, -11, -11,   4,   1, -16,  21,   6,  31, -14,   3, -12,  12, -21, -11,  13,  11,  12,   1,   0,  -1,   7, -16,  -6,  11,   6,   1,  -2,   1,  -1,   0,  -4,  -2,   7 },
     {    4,  -4,   4, -12,  21, -15,   8,   2,   2, -11,   7, -18,  19, -24,  15,  12,   9,  -6,  13, -12,  23, -22,   6,   7,   6, -17,  13, -19,  23, -23,   8,  17,  14, -15,  15, -19,  25, -15, -13,  20,  10, -14,  10, -17,   4,   9, -17,  41,  -3,   2,  -7,   4, -19,  15, -50,  23,   0,   2,  -3,   3,  -6,   8, -11, -28 },
     {    3,   6,   9,  21, -20,  -7,   1,  -2,   2,  13,  36, -36,  19, -16,  -3,  -1,   7,  50, -34, -14,  10,   7, -12,   0,  63, -11, -24,   7, -10,  12,  -6,  -8, -13,  33,   2, -20,   8,  -3,  11,  -5,  -1,  -3,  23,  -9,  -6,   6,   1,   4,   1,  -1,   2,  10,  -6,   2,   4,   0,  -1,   0,  -1,   1,   3,  -3,   2,   3 },
     {   -4,  -2, -17, -18, -17,  14,   7,   8,  -8, -27,   0,  12, -51,  -7,  14,   0, -25,   3,  41, -18,   2, -37,   6,   4,  28,  29, -19,  -7,  -2, -12, -20,  12,  -5,  29,  22, -21, -12,   1,  -9,  -9,   4, -10,  21,  28, -17,  -6,   3,  -1,  -4,   4,  -4,  15,  10, -15,  -5,   2,   0,  -2,   3,  -2,   6,  -1,  -5,  -1 },
     {    7,  16,  31,  28,   7, -27,  -8,  -2,   0,  -9, -36, -34,  11,  23,   8,  -2, -34, -34,  -2,  18, -16, -24,  -5,   8,  38,  37,  12,  -8,   1,   7, -25, -15, -12,   0,  21,  28,  10,  10,  13,  -7,   0,  -7, -11,  -8,  12,  10,   2,  11,  -4,   1,   0,  -3, -12,   1,   1,   4,  -1,  -1,   0,   1,   2,  -2,  -1,  -1 },
     {   -6, -12, -12, -18,  -9, -11,  -4,   9,  12,  14,  21,  20,  24,  17,  18, -12, -11, -21, -19, -25, -19, -27,   2,  11,  20,  17,  27,  23,  25,   9,   8, -13, -13, -21, -17, -26, -13, -13,  15,  30,  13,  10,  17,  10,   7,  -9, -14, -20,   1,  -5,   1,  -1,  11,  10,  23,  37,   0,  -3,  -3,  -3,  -9,  -7, -22,   4 },
     {   -3,  -4,  -1, -22, -53,  21,  23,   8,  -4,  -6, -23, -48, -25, -15,  34,   9,  -9, -21, -39, -11,  14, -17, -18,  22, -28, -24, -10,  11,   6,  12, -17,  -3,  -5, -23, -14,  -1,   4,   1,   1, -13,  -2,   5,  -5, -17,  -8,   6,   2,  -3,   2,  -1,   4,  -1,  -5,  -8,   3,   0,   0,   0,   0,   3,  -3,  -1,  -1,   2 },
     {   -3,   2,   1, -23,  -3, -10,   1,   3,  -3,   3,  -8,  -4,  42, -23, -26,   5, -16,  -6,  -5,  -9,   3,  -6,  -8,   5, -24,  10, -18,   0,  14, -18, -30,  13,  -7,  37, -10,   9, -13,  11, -26, -31,  14,  -4,  21,   5,  46, -12,   9, -27,   8,  -4,   4, -12,  12,  38,  -8,   6,  -1,  -4,   4,   3, -11,   5,  10,  -6 },
     {   -5,  -9, -18, -41, -37,  10,  12,   0,  -5, -14, -19,   9,  20, -32,  -8,  10,  -1,   5,  20,  39,  17,  18, -46, -16,  32,  16,  14,   0,   6,  15,  24, -31,   4,   3,  15,  11,   2,   1,  13,  24,   5,   0, -19,   8,  11,  -1,  -7,   8,   2,   2,  -4, -10,  10,   8,  -2,  -6,  -1,  -1,   1,  -3,  -4,   1,   5,  -4 },
     {   -3,  -2,   0,  -2,  -3,  23, -18,  17,  -4,  -6,  -3,  -3,  18,  -3,   5,   0,   3, -13,  -7,  34,   1,  -6,  23,  -8, -23, -32,  43,   7, -15,  -1,   0,  -5, -72,  35,  31, -17,   3,   0,  -2,   2,   0, -17,  12,   9,  -8,   6,  -7,   5,   9, -10, -11,  16,  -1,  -4,   2,   0,  -1,   3,  -2,   1,   4,  -3,   2,   5 },
     {   -1,  -2,  -7,  11,   6, -71,  55, -21,   0,   3,  -4,  26,  -6, -26,  13,  10,   3,  -4,  -2,  18, -21,  17, -25,  36,  -6, -21,  15,   5, -15,   5, -13,  10,  -8,   1,  21,  -5,  -2,  -7,  -2, -11,  -4,   2,  13,  12, -12,  -7,   4, -11,  -1,   0,  -6,   6,   3,  -9,   4,   3,   1,   4,   1,   0,   3,  -5,  -6,  -2 },
     {    0,   5,   7,  -1,   4,   4,  37,  -7,  -3,   4,  -4, -13,  19,  -1,  37,  -7,   1,  15,  12,  23,  28,   4,  28,   6,   7,  43,   7,   0,  -4,   0,  16,  45, -23,  11, -36,  10,  -5,   4, -23,  27,   5,   3,  -4, -30, -12, -11,  -9, -31,   0,  -4,   1,  -1,  -1, -11,   1, -17,  -1,   1,   4,   1,   4,  -4,  -2,   2 },
     {    3,  -3,   4,  -8,   5,  -5,  25, -20,   1,  -9,   0,  -5,  14, -31,  26, -53,   2,  -8,   6,  -6,  -2, -15,  33, -34,   2, -14,   1, -12,  12, -13,  14, -19,  16,  -6,  20,  -5,  12, -15,  20, -26,  -9, -18,   1, -12,  22, -19,  16,  -1,   7,  -4,  20,  -6,  13,  -9,  23, -12,   0,  -9,   8,  -3,  10,  -3,   9,  31 },
     {    2,   3,  -1,  -4,  29,  33, -29,  23,   4,   7,   0,  16,  29, -28,  50,   4,   2,  16,  -6,  30,  -9, -14,   6,  32,  14,  -2,  -1,  13, -26,  18, -30,   2,  32, -13,   8, -11,   4,   0,   2, -23, -10,  22, -13,  14,   8, -10,  -2,  -9,  -8,   8,  -5,  -8,  11,  -3, -11,  -9,   4,   4,   4,  -4,   3,   0,   0,  -8 },
     {   -1,  -2,   2,   2, -17, -17,  -9,   3,   2,   5,  -6, -15,   7,   6,  -5,  -4,  -5, -14,  -4,  27,   7, -11,   1, -13, -19,  11,  34,  -4, -15,   1,  -1, -17,  65,  35, -37,  -9,   3,  -3,  -3,   4, -15,  10,  47,   0, -22,   1,   4,   8,  -1,  -6,  -6,  32,  16,  -9,  -4,   2,  -1,   0,   2,  -8,  12,  12,   1,   1 },
     {   -5,  -8, -13, -20, -45, -21,   2,  -3,   0,  -3,  -5,   4,  27,  33, -21, -21,   2,   8,  -2,  14,  18,  -2,  42,   5,   7,  -6, -18,  22, -14,  11, -13,  41,  20, -15,  24,   6,  -3,  -3,  -1,  -2, -38, -15, -11,  15,   5,  17,   0,  13,   3,  -9, -24,  -7,  -6,   2,   2,  10,   9,   2,  -8,  -1,  -7,  -4, -15,   3 },
   }
 },
 {  //30
   {
     {  118, -11, -12,  -3,  -4,  -2,  -2,  -1, -31,  27,   1,   0,   0,   0,   0,   0, -18,  -2,   8,   0,   1,   0,   0,   0,  -1,  -3,   2,   1,   0,   0,   0,   0,  -3,  -1,   1,   1,   0,   0,   0,   0,  -1,   0,  -1,   2,   0,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0 },
     {  -15,  35,  -2,   1,  -1,   0,  -1,   0, -92, -11,  28,   1,   6,   1,   2,   0,  24, -59,  -3,   8,   0,   2,   0,   1,  16,   5, -28,   2,   0,   1,   0,   0,   2,   8,  -4, -10,   2,  -1,   0,   0,   5,   1,   3,  -5,  -2,   1,   0,   0,   2,   0,   1,   0,  -3,   1,   0,   0,   1,   1,  -1,   1,  -2,   0,   0,   0 },
     {   27,   4,  -5,  -2,  -1,  -1,  -1,   0,  -1, -66,   2,   6,   2,   3,   1,   1,  48,  22, -56,  -2,  -4,   0,  -2,   0, -16,  56,  13, -24,   0,  -5,   0,  -1,   0,  -9,  27,   3,  -5,   0,  -1,   0,  -3,   2,  -1,   9,   0,  -1,   0,   0,  -1,  -1,   0,   1,   3,  -1,   0,   0,  -1,   0,   0,  -1,   2,   0,   0,   0 },
     {   11, -57,   2,   6,   3,   3,   1,   1,  39, -11,  -6,   0,   2,   0,   0,   0,   1, -44, -11,  20,   1,   5,   0,   1,   4,  21, -69,  -1,   7,   1,   1,   0, -10,  37,   2, -37,   4,  -3,   2,  -1,   3,  -2,  19, -10,  -4,   1,   0,   0,  -1,   1,   5,  -3,  -5,   4,   1,   0,  -1,   1,   0,   2,  -8,   1,   2,   0 },
     {   14,   1,   3,  -2,   3,  -1,   1,   0,   2, -42,   0,  12,   0,   4,   0,   1,  11,  13, -10,  -4,   5,  -1,   1,   0,  -7, -24,   9,  26,  -2,   5,  -1,   1, -15,   3, -69,   4,  27,   0,   4,   0,  -1,   6,   5, -69,   3,   9,   1,   1,   7,  -2,  16,   0, -35,   2,   1,   0,  -1,   7,  -1,  10,  -3,  -7,   1,  -1 },
     {   -1,  10,  -2,  -7,   0,  -2,   0,   0,  -4,   4,  -6,  -1,  -8,   1,  -1,   0,  -4,  13,  -2, -16,   1,  -3,   0,  -1,  -2,  -1,   1,  -5,  -8,   1,  -1,   0,   3,   4,   3, -36,   0,  -3,   0,  -1,  15,  -3,  43,  -3, -55,   3,  -8,   1,  -1,  18,  -2,  67, -11, -34,  -1,  -3,   3,   0,   5,   2,  50, -13, -10,   0 },
     {  -10, -16, -28,  -1,   5,   1,   1,   0,  16,  36,  40,  -4, -10,  -2,  -2,  -1, -24, -39, -36,   7,  12,   3,   3,   1,  23,  36,  25, -25, -17,  -3,  -3,  -1, -14, -23,   3,  36,  13,   1,   1,   0,   2,   0, -20, -34,  -4,   6,   0,   1,   1,   4,  15,  16, -18,  -9,  -1,  -1,   1,   2,  -3,   4,  14,  -4,  -2,   0 },
     {  -22, -94,  -5,   3,   4,   3,   2,   1, -54,  21, -24,  -3,   0,   1,   0,   0,  39,  23,   8, -15,  -4,  -3,  -1,  -1,   5,   8,  18,   8,  -4,  -1,  -1,   0,  -1,  -5,  -2,   6,   4,   0,   1,   0,   1,   0,  -2,  -3,  -1,   1,   0,   0,   1,   2,   1,   2,  -2,  -2,   0,   0,   1,   0,   1,   1,   3,  -1,  -1,   0 },
     {  -16,  28, -28,   7,   2,   1,  -1,   1, -26,  42, -40,   6,   5,   1,   0,   1, -31,  39, -31,  -1,   8,  -1,   2,   0, -23,  31, -16, -17,  15,  -4,   3,   0,  -9,   7,  10, -29,  15,  -1,   2,   0,   5, -13,  20, -30,  11,   5,   1,   1,   6, -13,  15, -19, -12,  14,   0,   1,   0,  -1,   1,   4, -21,   3,   5,   0 },
     {   -8, -33,  18,   3,   6,   2,   2,   0, -29, -45,  -3,  16,   7,   5,   2,   1, -78,   2, -33,   7,  10,   4,   3,   1,  -1, -29,  -2, -26,  11,   1,   3,   0,  24, -26,  -1, -13, -12,   8,  -2,   1,   3,  12, -19,   6, -14,  -1,   1,   0,   0,   9,  -4,  -7,   2,  -6,   1,   0,   2,   0,   3,  -5,  -3,  -1,  -1,   1 },
     {   -2,  -5, -13, -18,   5,   4,   0,   0,   5,   6,  18,  26,  -3,  -5,   1,   0,  -4,  -8, -18, -23,  11,   6,  -1,   1,   4,  10,  19,  28,  -7,  -7,   1,  -1, -11, -14, -27, -22,  26,  12,   0,   2,  10,  16,  24,  10, -26, -12,   4,  -1,  -3, -10,  -6,   8,  61,  12,  -2,   0,  -5,  -9,  -9, -32, -35,  38,  11,   2 },
     {    0,   0,  -1,   5,   1,   2,   0,   0,   6,  11,  13,  -3,   5,  -3,   1,  -1,  17,  -8,  21,  21,  -2,   4,  -2,   1, -60,  13, -16,  25,  22,  -1,   5,   0,   2, -65,   5, -27,  25,  13,   3,   1,   3, -10, -39,  -5, -28,  16,   0,   1,  -4,  15, -10, -13,  -9, -16,   3,   0,   5,  -3,   9,  -6,  -1,  -7,  -4,   2 },
     {   -4,   4,  -7,  25,   0,  -5,  -1,   0,  -8,   4, -14,  26,  -2,  -3,  -2,   0,  -9,  12, -16,  34,   1,  -8,   0,   0, -11,  13, -23,  26,   3,  -6,  -2,   0,  -9,  19, -23,  27,  16, -16,   1,  -2,  -8,  15, -17,   5,  27, -15,  -1,  -3,   0,  11,   2,   1,  43, -26,  -6,  -2,  10, -11,  19, -25,  58,   5, -19,   3 },
     {    2,   7,  31,  -9,   8,   0,   0,   0,  -4, -18, -29,  24,  -6,   1,   1,   0, -20,  -6,  11, -26,  11,   0,   1,   0,  36,  23,   2,  23, -24,  -1,  -3,   0, -23,   7,  26,  10,  42,  -8,   3,   0,   4, -50, -14,   0,  -1,  31,  -1,   0, -13,  14, -35, -11, -11,  -8,   3,   0,   3,  -6,  10, -15,  -5,  -7,  -2,   2 },
     {   -5,   0, -77,   2,   0,   4,   3,   1,   1, -37, -45, -28,   7,   4,   4,   2, -19, -41,  20, -32,  -7,   3,   0,   2, -34,  -4,   7,   5, -24,   1,   1,   0,  -9,  -4,  -1,  13,  -6, -12,   1,   0,   8,  -3,  -1,   6,   6,  -7,  -3,  -1,   2,   6,  -2,   0,   5,  -1,  -2,  -1,   0,   3,   2,  -2,   3,   3,  -2,  -1 },
     {   -5,  -6, -63,  -3,   1,   1,   2,   1,  -5, -29,  44, -25,  -4,  -1,  -1,   0, -14,  47,  22,  35, -12,  -4,  -2,  -1,  45,  -1,  -7,  12,  25,  -3,  -2,  -1, -16,   9,   7, -11,  12,  13,  -1,   1,  -3, -20,   2,   5,  -5,   7,   5,   2,  -1,  -2, -16,  -3,  -1,  -1,  -1,   2,   0,  -1,   1,  -7,  -1,  -1,   0,  -1 },
     {    9,  15, -35,  13,   2,   3,  -2,  -1,  18,  14, -45,  14,   8,  -4,  -2,   2,  41,  -7, -26,  -2,  17,  -5,   1,   0,  61, -34, -13, -15,  15,   3,  -4,   1,  31, -27,  -7, -11,   2,   8,   0,   1,   7,   8, -20,  -3, -10,   4,   0,   2,  -5,  23, -11, -11,   1,  -3,  -1,  -2,  -2,   9,   6, -11,   0,  -2,  -2,  -1 },
     {    5,   6,  12,  23,  29,  24,   4,  -3,  -8,  -8, -16, -41, -48, -22,   1,   6,  11,  10,   8,  30,  29,  19,  -5,   2, -11, -18,  -9, -30, -28,  -4,  10,   9,   2,   8,   7,  12,  24,  -2, -20,  -8,  -2,   2,  -7, -11, -15,  14,  25,   6,  -2,  -2,   6,  12,  16,  -9, -10,   4, -10, -10, -19, -11, -15,  21,   4,  -1 },
     {   -4, -11, -34, -33,   1,   0,  -1,   2,   2,   9,  40,  48,  -7,  -5,   0,  -2,  14,  10, -12, -12,  31,   7,  -2,   1, -36, -37, -14, -13, -26,   1,  11,   0,  25,  22,  12,   1,  11, -26, -17,   2,  -7,  -9,   4,  14,  16,  29,  -4,  -3,   3,   4, -16, -11, -13,  -2,   8,   2,  -1,   1,   6,  -3,   4,  -2,  -8,  -2 },
     {    4,  -7,  11, -16,  -5, -18,   6,  -3,   5, -16,   4, -12,  12, -21,   0,   1,   7, -15,  23,  -4,  19, -20,   5,  -4,   5, -12,  21, -22,  29,  -7,   4,   0,  20, -12,  26, -12,  24,  -5,  -4,   2,  28, -27,  19, -29,  27,  -6,  -4,  -3,  26, -19,  34,  -9,  21,  -4, -13,   4,  18, -16,  27, -15,  21,  20, -13,  -2 },
     {    2,  -5,   0,  50,  12,  -1,  -1,   2,   0,  -2,   4,  11,  -2,   0, -16,  -3,  -2, -11,  -5,  19,  -1, -13,   0,  -2, -16,  -5,  16,  12,  24,  -6,  -4,   0,  22,  -4,   5,  12,   9, -10,  -5,   0,  36, -18,  18,   0,  23,  14, -26,  -4,  -9,  -3, -34,  38,  -4,  45,  -2, -18, -15,  19, -23, -12,   4,  -4,  25, -10 },
     {   -1,  -4,   3,   8,   2,   2,   9,  -6,  -2,   6,  -1,   4,   1,   1,   7,   0,  14, -11, -20,  16,  31,  -5,   1,   4, -15, -31,  17, -21,  10,  35,  -1,  -1, -69,  -4,  -6,  -9, -30,   4,  24,   5, -10, -44,   6,   5,   0, -26,   4,   2,   6,  -9, -30,  -3,   2,  -5,  -6,   2,   9,   6,  -4, -18,   3,  -5,  -6,   5 },
     {   -4,   1,  -7, -66, -11, -14,  10,  -3,  -4,  -4, -32,  19, -24,  13, -12,   3,  -2, -10,  24,  46,  19, -15, -11,  -3,  -5,  28,  11,  -7,  21,  23,   0,  -4,  19,   8, -17,  15, -20,  14,   1,   3,  14,   1,   0,  -5,   0,   9,  18,  -2, -26,   3,   0,  14,  -1,   2,  11,   6,  -1,   2, -16, -14,  -6,  -5,   3,   0 },
     {    1,   2, -10,  46,  33, -14,  -4,  -5,   1, -10,  10,  34,  37,  -9, -24,  -3,  -1,   7,  37, -19,  44,   0,  -7, -13,   1,  17,  -8, -10,   1,  26, -10,  -3, -13,   2,  -1,  11, -16,  32,  -6,  -9,   5,  10,   2,  10,  -7,  18,  14,  -6, -10,  -6,  29,   8,  -1,  -3,  13,  17,  -4, -12,   2,  10,  -4,  -1,  -5,  -1 },
     {    2,  -1,   2,   2,  -8,   1, -16,   4,   1,   1,   0,  18,   2,  39,  -2,  -2,   3,  13,   1,  19, -30,  15,  -5,  -9,   5,  -5, -27,   1, -35,  16,   2,  -4,   3, -28,   1,  21, -23,   4,  -3, -11,  38, -37,   9, -12,   2,   2,  10,  -5,  34,  -4,   6,  10,  12, -17,  11,   2, -18,  26,  17,   8, -11,  44,  -5,   0 },
     {   -1,   1,   0,  -4,  -9, -16,  30,   1,  -3,  -7,   7,  21, -11,   5,  17,  10,   3,  19,  27,  -1,   9,  -3,  13,   2,   8,  -1, -25, -20, -16,  -6,  11,  19, -37, -23,   9,  -2,  -8, -25,   6,  13,  30,  22, -11,  -4,   7, -16, -41,  34, -30,  20,  23,   3,  -1,  12, -22, -15,  -9,   0,   0,  -1,  -4,   4,  20,  -3 },
     {   -1,   0,  -6, -17,   9,   6,  18, -12,  -1,   0,  -6,  33,   6,  -3,  -6,   6,   3,  -5,   0,  -7, -16,   6,  20,  -2, -16, -19, -10,   6,  15, -36, -24,  12,  -4,  19,  29,  26,   7,  54,   4, -11,  -7, -10, -10, -10, -14, -19,  23,  28,  17,  -1,   3,  11,   1,   0, -42,  -8,  -3,  14, -14,   9,   5,  -4,   9, -12 },
     {    3,   3,   6, -28,  67,  -9,   8,  -9,   2,   9,   6, -32,  31,  23,   3,  -2,   3,  11, -19, -13, -11,   4,  -3, -12,  -1,  -7, -21,  13,  -3,   9,  16,   3,   2,   2,   6,  11,  -8,   7,   6,   9,  35,  10,  -6,  -7,  11,   5,  -6,   6,  -4,  -4,  -7,  -6,  -4,  -4,   7,  -3,  35, -10, -39, -31,  13,  -2,  -3,  11 },
     {    2,   1,  -6,  22,  40, -30,   5,  14,   2,   3,   9,   5, -70,  24,  28,   9,  -1,  -9,  -7, -13,  -5, -31,  -1,   2,  -5,   4,  11,  15,   7,   2, -19,   1,  17,   8,  -5, -11, -17,  21,  13,   5,   6,  -5,   8,  12,   9,   1,  -1,  15,  10,  21,   5, -17, -16,  -7, -10,  11,   2, -18,  20,   6, -11,  13, -13,  -3 },
     {    0,   1,  -7, -10,  26,  11, -12,   5,  -1,  -6, -20,  25, -19,  19, -11,   0,   3, -14,   5,  17, -21,  11, -16,  -5,   7, -22,   8,   6,  16,  15,  -1,  -2, -23, -14,  40,  -3,  10,  -2,   6,   4, -49,  25,   6,  -7,   5,  24, -41,   2,  13, -25,  20,  22,   3,   2,  20, -14,  16, -10,   5,   6,  -3,  12,   3,  13 },
     {    0,   1,  -6,   5,  -3, -15, -11,   3,   0,   0,   3,   9,  -7,  -2, -38,   1,   3,   2, -14,  10, -16,  14, -10,   8,   0, -17,  25,  19, -16,  23,   8, -17,  -4,  32,  24, -34,   1, -15,  14,  -5,  35,  -5, -50,  -6,  -4, -19,  20, -19,  -5,  19,  27,  11,   6,  13, -14,  -4,  16, -21,  -1,  17, -11,  -9,  11,  -3 },
     {    0,   3,   3, -15,  -2,  16, -27,  -7,  -2,  -1,   0,  -3, -22, -14,   0, -27,   1,   6,   6, -25,   7, -18,  -2,  -7,  -5,   7, -16, -22,  35, -21, -12,  -6, -26,  21, -21,  13, -13,  -3, -13,  -8,  23,   1, -40,   7, -21,  10, -29, -23,  43,  19,  -2,   7,   1,  15,  14,  -5,  17,   5,   0,   0,  -2,  18,   9,  11 },
   },
   {
     {  107, -42,  -8,  -3,  -2,  -1,  -1,   0, -24,  30,  -6,   0,  -1,   0,   0,   0, -37,  10,   9,   0,   1,   0,   1,   0,   3,  -8,   2,   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -10,  10,  -3,   1,  -1,   0,   0,   0, -98,  29,  16,   0,   4,   0,   2,   0,  35, -52,   8,   2,   1,   1,   0,   0,  32,  -2, -18,   2,  -2,   1,  -1,   0,  -1,  10,  -2,  -3,   0,  -1,   0,   0,   4,  -1,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,  -1,   0,   0,   0 },
     {   11,  82, -52,   3,  -8,   0,  -3,   0,  16,   6,  23, -16,   1,  -3,   0,  -1, -51, -25,  38,   6,   1,   1,   1,   0,   4, -25,  -8,  12,   1,   2,   0,   0,   8,  -1,  -8,  -1,   1,   0,   0,   0,   1,   2,  -1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   51,  27, -31,   0,  -5,  -1,  -1,   0,  11, -37,  25,  -5,   3,   0,   1,   0,  58, -22, -21,   8,  -4,   2,  -2,   1, -39,  57, -11,  -6,   2,  -2,   1,  -1, -16,  -3,  20,  -4,   1,   0,   1,   0,  -1,  -5,   1,   4,  -1,   1,   0,   0,  -1,  -2,   0,   1,   1,   0,   0,   0,  -1,   1,  -1,   0,   1,   0,   0,   0 },
     {    5, -17,  11,  -3,   3,   0,   1,   0,   8, -62,  27,   6,   1,   3,   0,   1, -37, -12, -24,  21,   1,   3,   1,   1,  58,   6, -47,  -7,   2,  -1,  -1,   0,  -6,  46,   1, -22,  -1,  -2,   0,   0,  -6,   1,  18,  -2,  -3,   0,   0,   0,  -2,  -2,   1,   4,  -1,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   0 },
     {    0, -22,   2,   5,   0,   2,   0,   0,  55,  43, -34,  -9,  -2,  -3,  -1,  -1,  28, -45,  41,  -3,  -2,   0,   0,   0,   5,   3, -31,  18,  -2,   2,  -1,   1, -37,  42,  -2, -14,   6,  -2,   1,  -1,   0, -14,  18,  -1,  -3,   2,   0,   0,  -1,  -1,  -2,   4,  -1,   0,   0,   0,  -2,  -1,   0,   0,   1,   0,   0,   0 },
     {    8, -27,  -8,  13,   2,   2,   0,   0,  -5, -47,  12,  -3,   8,   1,   2,   0,  26,  -8,  25, -14,  -5,   2,  -1,   0, -28, -33,  -2,  38,  -5,   5,   0,   1,  19,   7, -67,   3,  13,  -1,   2,   0,  -6,  33,   2, -35,   2,   1,   1,   0,   4,  -1,  15,  -3, -10,   2,   1,   0,  -1,   4,   0,   4,  -4,   0,   1,   0 },
     {  -14, -23, -68,  56,  -5,   8,  -1,   2, -12, -17, -18, -11,  17,  -1,   2,   0,   3,  33,  22, -47,  -1,   0,   0,  -1,  34,  26,   7,  -2, -17,  -2,  -1,  -1,  -2,   8,  18,   6,  -2,  -4,   0,  -1,  -6,  -7,   6,   9,   1,   0,   0,   0,  -3,  -3,  -2,   1,   3,   0,   0,   0,   0,  -2,   0,  -1,   1,   0,  -1,   0 },
     {   26,  46,   1, -24,  -1,  -3,  -1,  -1,   4,  -6, -19,  18,  -7,   2,  -1,   0,  42,  33, -21, -17,   3,  -5,   0,  -1,  46, -37,  38,   0,  -7,   2,  -2,   0, -39,  31, -17,  14,   1,  -1,   0,   0, -16,   0,  14,  -6,   5,  -1,   1,   0,   0, -10,   2,   5,   0,   1,  -1,   0,  -2,   0,  -1,   0,   2,  -1,   0,   0 },
     {    8,  22,  10, -19,   2,  -1,   0,   0, -10, -25, -68,  51,  -2,   5,   0,   1, -20, -36,  13, -34,  21,   0,   3,   0, -12,  30,   3, -20, -12,   3,   0,   0,  40,  10,  -5,   1, -10,  -5,   0,   0,  -3,  13,  15,  -2,  -1,  -3,  -1,   0,  -4,  -1,   3,   4,   1,  -1,   0,   0,   0,  -2,   2,  -1,   1,   0,  -1,   0 },
     {    3, -19,   7,  -2,   4,   1,   0,   0,  20, -16,  16,   1,  -3,  -1,   0,   0, -10, -51,   8,  -6,  10,   2,   1,   0,  29,  10,  31, -15, -23,   4,  -3,   0, -27, -14,  -2,  63,  -6,  -5,   1,  -1,   1,  -1, -52,  -5,  30,  -1,   1,   1,   2,   9,   4, -26,  -4,   8,   0,   0,   1,   3,   0,   2,  -9,   0,   2,   0 },
     {    9,  -4, -10,   4,  -3,   0,   1,   0,  31,  30,   2, -15,  -3,  -2,  -1,   0,  30, -10, -16,  13,  -1,  -1,  -1,   0,  38,  -2,  -9, -18,   5,  -2,   0,   0,  61, -13,   4,   3, -15,   3,  -3,   0, -49,  55,  -6,   6,   1,  -6,   0,  -1, -18,  -6,  22,  -2,   7,  -2,  -1,   0,   0, -10,   1,   4,   2,   0,  -1,   1 },
     {  -13,  -2, -43,  25,   0,   3,   0,   1,  -3,  27, -27,  14,  -6,   4,  -1,   0, -18,  -7, -57,  22,  16,   0,   3,   0, -14,   3,  -7, -26,  14,   1,   2,   0, -27,   7, -20,  -6,  10,   0,   2,   1,   8,  12,  -3, -50,   9,   3,   1,   1,  12,  -4,  34, -14, -24,   6,   0,   0,   1,   2,   2,  12,  -9,  -3,   2,  -1 },
     {   -2, -19, -20,  19,   0,   1,   0,   0,   4, -10,  10,   1,   0,   1,   1,   0,   8, -26, -11,  -1,  21,  -3,   2,  -1, -29, -55,  22, -46,  10,  16,   0,   2,  -4,  28,   3,  -8, -50,   8,   1,   1,  20,   8,   4,  38, -15, -21,   1,  -1,   5,   7, -12,   9,  21,  -6,  -4,  -1,  -2,   5,  -1,  -8,   6,   4,  -1,   0 },
     {   10,   8,  -4,   2,  -5,   1,  -1,   0,   7, -12, -25,   8,   5,   1,   1,   0,  17,   6,  -7,  -4,  -2,  -1,   0,   0,  24, -17, -22,  10,   8,  -1,   1,   0,  -1, -21,   3, -21,   8,   5,   1,   1,  56,  11, -37,   4, -17,   1,  -2,   0, -35,  67,   4, -20,   5,  -9,   0,  -2,   0, -12,  32,  -2,  -1,   1,  -1,   1 },
     {   12,  22,  48,  53, -45,   1,  -6,  -1,   8,  17,  25,  11,  18, -16,   1,  -2,  -2,   5, -21, -44,  41,   7,  -2,   0,  -7,  -6, -33,   3, -23,  21,   3,   1,  -8, -14,  -1,   1,  -6,  -9,   7,   1,  -7,   0,   6,  -6,   5,  -4,  -3,   2,   0,   1,   0,   0,   0,   0,  -2,   0,   1,   1,  -1,  -1,  -1,   2,   0,   0 },
     {    1,  -6,  -8,   1,  -2,  -2,   0,   0,  -5, -25, -30,  17,   4,   2,   0,   1,   1, -17,   4,   4,   1,   0,   2,   0,   5, -34, -20,  10,  15,   2,   0,   0, -35, -59,  29, -20,   6,  10,   0,   2, -55,  11, -10,   5,  -9,   2,   2,   1,  38,  -4, -21,  -3,  -3,  -6,   0,   0,  13,  21,  -1,  -8,   2,  -2,  -3,  -1 },
     {   -8, -10, -24, -26,  34, -14,   4,   1,   4,  13,  40,  45, -54,  12,  -6,   0,  10,  28,  28, -22,  32, -11,   1,  -1,  -1,   4, -43, -11,   6,  11,   0,  -1, -11, -17,  -2,  11, -16,   3,   2,   1,   0,   4,   6,  -9,   6, -10,   2,   1,  -3,   5,   3,   1,   1,   1,  -1,   0,  -5,  -4,   8,  -2,  -1,   2,   0,   1 },
     {    4,  10,   4,   4,  -1,  -1,  -2,   0, -10,  -8, -21,  -7,   5,   7,   0,  -1,  11,  23,   8,  12,  -6,  -9,   0,  -1, -17, -17, -34,  -2,  18,   9,  -1,   2,  21,  29,   7,   5, -20,  -7,   1,  -1, -17, -31, -22,  14,  50,  -4,   0,   0,   2, -13,  -4, -60,   3,  27,   1,   1,   8,   1,  16,   2, -34,   4,   6,   1 },
     {    7,  -7, -19,  13,   0,  -1,   1,  -1,  18,   5,  -2,  18, -14,   4,  -1,  -1,  17, -13, -25,  12,   7,   3,  -1,   0,  21, -14,  -2,   5,  -1,  -1,   1,  -3,  49, -20, -18,   9,  12,  -9,   1,   1,  18, -51,  12,  -7,  10,  15,  -3,   1,   2,   3, -47,  11, -13,  10,   6,   0, -20,  45, -20, -20,   3,  -7,   5,  -1 },
     {   -7,  -8, -21, -47,  12,  12,   6,  -1,  -5,   0, -15, -48,   1,  11,   1,   3,  -9,  -9, -45, -25,  11,   6,   2,   3, -15, -19, -26,  37, -44,   3,   6,   1,   0,  -3,  23,  21, -14, -24,   5,   2,   0,   5,  19,  -2,  15, -15,  -5,   1,  -5,  11,   2,   8,   5,   1,  -7,  -2,  -3,   6,   5,  -5,   2,   3,   2,  -1 },
     {   -8,   4,  -9,  20, -18,  10,   0,   0, -11,  11,  -9,  32, -16,   4,  -3,   2, -13,  11, -13,  36, -18,   1,  -1,   1, -16,  21, -14,  31, -12,  -5,  -3,   1, -17,  24, -28,  36,  -4,   1,  -2,  -4, -25,  28, -14,  24,   5, -11,   5,  -3, -15,  21, -32,   7,  33, -12,   4,  -1,  -2,   8,   2, -29,  20,   1,  -3,   2 },
     {   -5, -10,  -5, -19, -54,  46,  -4,   1,  -3,  -9,  -2, -15, -29,  -6,   9,  -2,   0,  -4,  -1,  -3,  46, -56,   7,   3,   2,   5,   6,  21,  28,  -7, -28,   4,   7,   5,   6,  10,   8,  28, -13,  -6,   3,  -5,   2,   3, -10,  13,  14,  -8,   0,  -5,   6,  -7,   0,  -1,   6,   4,   1,  -3,  -4,   1,   1,   3,   2,   3 },
     {   -4,  -9,  -7,  13,   1,  -5,  -3,   1,   0,  -1,  17,  26, -23,   1,  -5,  -1,  -1, -13,  -4,   7,  14,  -1,  -3,   2, -15, -21,  35,  17, -16,   2,   2,  -5,  10,  23,  49, -24,  32, -21,  -1,  -5, -20,  10,   4, -18,  18,  24, -17,   2, -23,  14, -27,  -3,  -6,  14,   9,  -4,  16, -45,  12,  -5,  -5,  -2,   9,   1 },
     {   -1,   1,  -1,  -8,   6,   1,   1,  -2,   1,  11,  15, -11,   0,  -2,   0,  -2,  -3,   0,  -7, -12,  10,   6,   2,   0,  -3,  14,  15,  -8,  -7,  -5,  -1,   1,  15,  21, -18, -11,   8,   5,  -5,   0, -43, -25,   0,   3, -16,   2,   6,  -1,  56,  58,  -4, -15,  -7, -17,  -2,   0, -30,   1,  53,  10,   4,  -1,  -8,   2 },
     {   -4,  -8, -11, -12,   3,  -1,   1,  -2,   9,  13,  16,  12,   2,  -6,  -3,   1,  -9, -16, -22, -22,   4,   8,   5,   0,  15,  24,  31,  32,  13,  -7,   2,  -4,  -6, -16, -30, -42,   0,   1,  12,   6,   7,   7,  15,  15,  -1, -23,  -8,   2,  -1, -25, -12, -38,  42,  18, -12,   3,  18,   8,  19, -14, -33,  17,   0,  -1 },
     {   -4,  -6,  -5,   0,   5,   0,  -5,   4,  -4,  -7,  -2,  -1,   5, -13,   9,  -1,  -5, -11,  -5,   7,  -5,   3,   0,   1,  -6,  -7,   2, -13,   9,  10,  -7,   1,  -8, -25, -27,  16,  -3,   4,   7,  -4, -41, -56,  39,   3, -13,  -2,   2,   1, -56,  30,  20,  -9,   8,  -5,  -2,   1,  38, -23, -10,  11,   4,   2,   0,   1 },
     {    8,  12,  18,  40,  35, -26,  -8,  -3,   1,   1,   1,  -6, -35,  26,   1,   3, -10, -21, -29, -46, -33, -14,  15,   3,  -5,  -8,  -2,  29,  31, -30, -11,  -1,   9,  13,  17,  16,  -4,  29,  -7,  -5,  -1,  -8,  -6,  -2,  -7,  -6,  16,   0,   1,   0,  10,   2,  -1,  -7,  -1,   1,   9,   0,  -5,   5,   5,  -2,  -3,   2 },
     {    6,  11,  18,  28,  25,  -6,  -6,  -3,  -6,  -8, -16, -29,  -9,  29,   6,   0,   0,   4,  13,  35,  27, -22,  -2,   3,  14,  23,  24,  25,   4,  39, -16,  -4,  -6, -14,   0, -14, -39,   6,  32,  -4,  -2,   5,  16, -24,   7, -26,  -5,   8,  -2,  15,  -8,   8,  -9,  13,  -3,  -4, -18,  -4,   5,  -9,   3,  -9,   3,   4 },
     {   -1,   0,  -5,  -6,  17,  -9,   0,   6,  -2,   0,  -4,  -5,  -4, -10,   1,   3,   0,   4,  -6,  -1,  39,  16, -12,  -9,  -2,   5,   5,  19, -15,  -3,  10,   3,   5,  17, -13, -39,  -7,  25,   5,  11, -19, -28, -64,  12,  13,   6,   4,  -2, -15,  -8,  28,  30, -14, -17,  -1,   6,  17,   6, -18, -34,   5,  11, -10,  -1 },
     {    0,   0,  -6, -10,  27, -34,   1,  -2,   5,   8,   0,  -3,  54, -20,  -9,  -3,   3,   6,  -5,   6,  33,  -1, -13,   7,   2,   7,  -7,   8,  15,   9, -10,  -6,   9,  13,   6,  34,   4,  40, -10,  -6,  15,  17,  17, -15,   0,  16,  14,  -2,  18,   7, -33, -11,  -2, -13,   3,   3,  41, -12,  -4, -18,   0,  10, -12,   4 },
     {    1,   4,   7,  -8,   3,  19, -26,   0,  -2,   0,  -2, -36, -11,   6, -20,  -2,  -2,   2,   0, -26,   0,   3,   1,   5,  -1,   8, -11, -35,  22, -25, -21,  -7,  -9,  -1, -31, -11, -10, -34,   7,  -6,  -5,  10, -18, -12,   0,   6, -11,   0, -10,  -9, -48,   5, -24,   3,   3,  -7,  32, -13,  14, -25,  22, -18, -14, -11 },
   },
   {
     {  110, -32, -22,  -1,  -4,  -1,  -2,   0, -31,  36,   1,  -6,   0,  -1,   0,  -1, -14,  -3,  12,   3,  -1,   0,   0,   0,  -3,  -1,  -1,   3,   2,   0,   0,   0,  -3,  -1,   1,   0,   1,   1,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    8,  85, -26, -13,  -4,  -4,  -2,  -2, -46,  -6,  58,  -6,  -1,  -1,   0,   0,  15, -41,  -3,  23,  -1,   1,   0,   1,   1,   3, -13,  -1,   6,   0,   0,   0,   3,  -5,   0,  -1,  -1,   2,   0,   0,   1,   1,  -3,   0,   0,   0,   1,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {   35,  -7,  32, -15,   0,  -3,  -1,  -1, -20, -75,   8,  46,   2,   6,   0,   2,   8,  14, -62,  -8,  15,   2,   2,   1,  -4,  18,  10, -22,  -8,   2,   0,   0,   1,   2,  -1,   5,  -3,  -3,   0,   0,  -1,   4,   2,  -4,   1,  -1,  -1,   0,   0,   1,  -1,   1,  -1,   0,   0,  -1,   0,   1,   1,  -1,   0,   0,   0,   0 },
     {  -24,  -2,   2,   8,  -3,   4,  -1,   1, -86,  18, -36,   8,  20,   0,   7,   0,  29,  -4,   9, -55,   2,   6,   0,   1,  14,  -6,  23,   6, -31,   0,   0,   0,   5,   0,  -4,  11,   3, -11,  -1,   1,   3,  -1,   3,   0,   1,   1,  -4,   0,   2,   0,  -1,   1,   0,   0,   0,  -2,   1,   0,   1,   0,   0,   0,   0,   0 },
     {   29,  73, -26, -15,  -8,  -2,  -2,  -1,  43,  -7, -35,  10,   6,  -3,   2,  -1, -24,  31,  10, -41,   6,   3,   1,   0,  -3, -14,  37,   6, -21,   0,   0,   0,  -5,  -5,  -3,  15,   5,  -6,   0,   0,  -2,  -3,   4,  -1,   2,   2,  -1,   0,  -2,  -1,  -1,   2,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 },
     {  -15,  -4, -24,   2,  15,   3,  -1,   2,   3,  38,   3,  19, -19,  -3,   2,  -3, -28, -14, -23,   4,  58,   0,   0,   2,   5, -14,  -3, -60, -21,  34,   9,   2,   3,   3,  11,  11, -34, -26,   7,   5,   3,  -1,   1,   1,  11,  -5, -13,  -1,   0,   0,   2,   2,  -3,   3,   1,  -4,   1,   0,   0,   0,   2,  -1,   0,   1 },
     {    4,  22,  88, -28, -19,  -4,  -4,  -3,  -8,  35,  -5,  40,  -6,  -9,  -3,  -3, -33, -11,  17,  10,  16,   2,   0,  -1,  10, -29, -14,  16,   9,   5,   1,   1,   0,   1, -16, -10,   8,   5,   2,   0,   4,  -4,   0,  -4,  -4,   2,   3,   1,   0,   1,  -3,  -1,   1,  -1,   0,   1,   1,  -1,   0,  -1,  -1,   0,   0,   0 },
     {    8, -21,  18, -30,   9,   5,   1,   1,  28, -18,  37, -21,   1,  -4,   3,  -1,   4, -41,  23, -23,  -6,  26,  -6,   6,   6,   4, -10,  23, -64,   8,  11,   0,  -7,  14, -10,  28,   3, -39,   4,   2,   0,   2,  -2,   0,  13,   0, -12,   0,  -1,   2,  -2,   3,   1,   1,   1,  -3,   0,   0,  -1,   0,   2,  -1,  -1,   0 },
     {  -10, -12,   3, -45,  21,   0,   2,   0, -15,  19,  37, -11, -52,   6,   0,   1,  19,  53,  -1,  13,  -2, -31,  -3,  -3,   2, -15,  47,  11,   1,   5, -11,  -3,   2,  -9, -14,  33,   7,   1,   3,  -3,  -2,  -2,  -4,  -7,  12,   5,   1,   2,   1,  -2,  -3,   3,  -3,   2,   1,   2,   0,  -1,  -1,  -1,   2,   0,   0,   0 },
     {   12,  13,  18,  11,  12,  -7,  -4,  -2, -19, -43, -23, -32,  -5,  17,  11,   4,  11,  27,  41,  17,   0, -18, -13,  -1, -13, -34, -22,  -3,  11,  40,  12,   1,   4,   9, -11, -14, -41, -28,  11,   9,   0,   3,   7,  12,  14, -11, -19,  -6,   1,   2,  -2,  -2,  -1,   6,   4,  -3,   0,   0,   1,   2,   1,  -2,  -1,   1 },
     {  -17, -13, -23,   2,  12,  -1,   6,   1, -30, -37,  18,  -6,   2,  13,   0,   4, -93,   9,   7,  -4, -14,   0,   9,   1,  44, -25,  -3,   6,  -2, -16,  -4,   1,   8,   9, -11,   0,   6,   6,  -6,  -3,   6,   2,   1,  -2,  -4,   1,   3,  -2,   1,   2,  -1,   1,   0,  -2,   0,   1,   2,   0,   0,  -1,   0,   0,   0,   0 },
     {    6,  -8,   8, -10,  14,  -7,   2,  -3,  12, -12,  13, -15,  31,   0, -10,  -2,  12, -24,  18, -25,  20,  -2,  -2,   0,  21, -12,  23, -15,  15,  32, -36,   5,  -9,   6, -11,  14, -37,  60,  14, -18,  -4,   3,  -7,   5, -17, -19,  36,   2,   0,   0,   0,   0,   1,  -7,  -4,   9,  -1,   2,  -2,   2,  -3,  -1,   1,   0 },
     {   -3, -14, -45,   3,   4,  10,  -3,   4,   4, -25, -11,  56, -29, -17,  10,  -2,  19, -21,  28,   2,  33, -23, -16,   1,   7,  -9, -17,  32,   8,  18, -12,  -6, -13,  21, -30, -12,  29,   4,   8,  -4,  -1,  -1,  10, -23,  -1,  13,   4,   0,   0,   1,   0,  -1,  -4,  -2,   4,   1,  -1,   0,   2,  -3,  -1,   0,  -1,   1 },
     {   -3,   0,  -9, -83,  15,  14,   8,   5,  -7,  -7, -62, -17, -19,  11,  10,   6,  -6, -40,  -7,  13, -16, -17,   3,   6,   2,  18,  -7, -11,   6, -11, -12,  -2,   0,   8,  19,  -8,  -7,   4,  -5,  -5,   4,   2,   7,   9,  -7,  -2,  -1,  -2,   0,   1,   2,   4,   2,  -2,   0,  -2,   1,   1,   1,   1,   1,   1,  -1,   0 },
     {   -4,  -3,  -7,  -2, -12,  -4,  12,   6,   2,   7,   4,  11,   4,  17,  -4,  -5,  -9,  -7, -10,  -7, -16, -10,  18,   8,   5,  17,  11,  22,  10,  14,  -6, -13,  -8, -16, -11, -12,  -6,  -6,  66,  18,   0,  -7,   0, -13,  -8, -49, -38,  43,  -1,   2,   0,   4,   5,  12, -27, -25,   2,  -1,   1,  -1,   0,  -1,   5,  -2 },
     {   11,  18,  31,  55,  19, -17, -14,   0,   2,  -1, -25, -26, -25,  18,  15,   1, -16, -39, -13,  10,   0, -34,  -5,   5,  20,  28,  23,   2,  -1,   5, -22, -12, -10,   3,   0,  33,  19,  -3,   7,  -3,  -2,   0, -10, -19,  21,  16,  -1,   1,   0,  -2,   4,  -5, -12,   3,   7,   1,   0,   0,  -2,   1,   1,  -1,  -2,   1 },
     {   -2,  -7,  -4,   6,  30,  -4,   3,  -2,   2,   1,  -7,  -2,   9,  22,   1,   2,  -6, -37, -23,   6,  -1,  12,  22,   3, -72, -51,  27,  10,   2, -11,   2,   9,  25,  -8, -42,  -3,   5,   0,  -9,   2,   0,  15,   1, -19,  -5,  -1,   1,  -1,   3,   1,   3,  -1,  -5,  -1,  -4,   0,   0,   4,   1,  -1,  -1,   0,  -1,  -1 },
     {  -12, -10,   4,   0, -29,  15,  -4,   5, -16, -18,  13,   7, -10, -18,  18,   0, -39,  -6,  44, -12,  10, -10, -11,   6, -69,  41,  15,  -8,   4,  11, -13,  -5,  14, -33,  16,  12,  -3,   4,   4,  -9,   6,  -3, -14,  13,   3,   1,   4,  -1,   5,  -5,  -1,   0,   2,  -1,   2,   1,   3,  -1,  -3,   3,  -1,  -2,   1,   2 },
     {   -4,   9,  -6,   9,  -8,  25,  -5,  -7, -11,   9, -18,   8, -18,  20,  -6,   2,  -9,  15, -15,  23, -22,  32,  -1, -13, -16,  12, -21,  14, -25,  18,  10,  -6,  -8,  16, -10,  20, -11,   2,  30, -38,   3,  -5,   6,  -8,  19, -28,  57,   5,   3,  -4,   2,  -7,   5, -16,  -7,  30,   1,   0,   1,   0,   2,   0,  -5,   1 },
     {   -3,   3,   2, -14,  55, -14, -14,  -3, -10,   1,  -4, -22,  17, -12,  -1,   0, -15,  13, -13,   2,  22,  33, -42,  -5,  -7,  16,   2,  20,   2,  29,  25, -34,  -2, -18,  14, -24,  26,   9,  22,  15,   1,  -8,  -4, -11, -26,  21,   6,  10,   1,   1,  -2,   4,  -9, -15,   7,   4,   1,  -1,   1,   0,  -2,   0,  -4,  -1 },
     {    1,   4,   0,   1,  63, -34,  -9,  -4,   1,   3,   2,  52,  -4,  37, -25,  -7,   5,   1,  52,  -2, -20,  12,  24, -15,  -3,  34,  -1, -20,   2, -15,  11,   9,   1,  -1,   7,   4,  -8,   1,  -5,   4,   4,  -8,  -4,   3,   6,   0,  -1,   1,   0,   3,  -4,  -3,   0,   2,   2,   0,  -1,   0,   0,  -3,   1,   1,   2,   1 },
     {   -4, -10, -11, -12, -34, -14,  17,  -6,  -5, -16, -25,   0,  29, -20, -36,  12,   2, -11,  21,  57,  10,  20,  -8, -30,  17,   5,  37, -12,   7,   2,  26,  -1,   7,  -8, -15,  33,   6,  -5,   3,  16,  -2,   4, -13, -19,  19,   4,  -4,   2,   2,  -1,   4,  -4,  -8,  11,  -2,  -1,  -2,   0,  -2,   0,  -1,  -1,   1,  -1 },
     {   -5,   5,  -5, -11,  25, -44,  26,   3,  -5,  17,  -4,   7,  43, -42,   3,   5, -12,  21, -17,  10,  -8, -17, -13,  16, -17,  24, -28,  18,  -9,   9, -37,  13,  11,  16, -18,  14,  -8,  -4,  -5, -20,   5,   4,  11,  -9,  26,  -9, -10, -16,   2,  -4,   8,  -6,   4,  18,   4,  -8,   1,  -1,  -1,  -4,   2,   0,   5,   3 },
     {   -7,  -8, -11, -10, -16,  -5, -11,   4,   8,  16,  10,  20,  19,  35,  21,   8,  -8, -13, -17,  -7, -18, -12, -20, -10,  10,  19,  18,  21,  17,  18,  19, -18, -28, -24, -32,  -8, -27, -10, -11,  16,   0,  16,   8,  19,   6,   1,   0, -50,  -1,  -8,   1,  -5,  11,   1,  33,  18,   1,  -1,  -2,  -3,  -1,  -5,  -9,   8 },
     {    6,   5,   6,   4,  32,  75, -35, -18,   7,   4,   4,  19,  41, -25,  24, -11,   2,   4,   9,  28, -27, -20,  -9,  20,  16,   3,  18, -16,  -7,  -5,   4,   0,  24,  -1, -12,   1,  -5,  -2,   6,   5,  11,   7,  -4,  -4,   2,  -4,  -5,   7,  -9,   3,   3,  -4,   2,   2,  -3,   1,  -2,   2,   0,  -5,   1,   4,  -3,  -2 },
     {    2,   2,   4,   8, -17, -18,  -1,  -4,   5,   7,   7,  -7, -17,   2,  -7,   0,  -2,  -4,   5,   4,   2,  -1, -14,   2,   6,  31,  35,  -6, -13,   2,   7,  -6,  68,  52, -22, -32,  -1,   2,   6,  -4, -10,   0,  42,  16, -19,  -4,   0,  -3,  -3, -14, -13,  18,  10,  -8,   3,   4,  -8,   1,   0,  -5,   6,  -2,   0,   3 },
     {    1,  -3,  -5,   9,  13, -24, -12,  24,   1,  -6,  -7,  21, -13, -33,   5,   0,   3, -15,  -9,  11, -13,   8,  -8,  -3,   2, -32,  17,  17,   0,  -5,  -2, -35,   9,  10,  27,   9,   1,   2,   2,   3,  -6,  16, -28,  65,  18, -21,   1,   9,  -9,  -2, -13, -15,  37,  -1,  -8,   2,   6,  -1,   7, -10,  -2,   8,  -7,  -1 },
     {    5,   0,  -2,  -4,   4,   3,  11,   1,   8,  -1,  -9,   1,  -9,  15,   7,  -2,   8,   1,   0, -11,  11,  10,  -2,   9,  28,   5, -31,   6,   9,   6,   1,  13,  64, -58, -12,  16,  11,  -2,   7, -11, -54,  18,  -5,   8,   5,   0,   3,  -3, -23,  -3,  14, -22,  -1,  -3,  -6,   3,   7,  -6,  -7,   4,   0,  -4, -11,   4 },
     {    5,  10,  12,  16,  30,  20,  38,  -3,  -4,  -1,  -1,  -6, -37, -42,  18,  27,  -5,  -3,   3, -17,   3,  21,  -2, -16,   5,  23,  11,   6,  22,  -7,  23,  20, -18,   2, -16,  -2, -16,   3, -22,  26,  -7,  20,   5, -12,  -5, -34,   2, -15,   0,  -7,  20,   5,  15,   0, -27,   2,  -3,   1,   2,   3,   0,  11,   3,  -8 },
     {    0,   7,  -1,  15,  -1,  22, -10,  17,  -5,   2, -11,   8, -24,   8, -53,  17,  -2,   5,  -3,  18, -16,  25, -30,  19,  -1,   3,  -1,   8, -26,  14, -34,  11,  -8, -11, -11,   5, -15,  19, -11,  -1,  -4,  20, -14,   7, -34,   6, -28, -23,  -1,  15,  -4,  12,   2,  -4,  12, -41,   2,   2,   1,  -2,   4,  -4,  10,  -4 },
     {   -1,   2,   0,  -7,   6,  35, -15,   5,  -3,   2,   1, -13,  -1,  13, -25, -35,  -7,  -1, -11, -25,  19,  28, -13, -11,  -6,  11,  -4,  15,  46,  -5,   1,   2,   5,  29,  -5,  10,  -1,   7,   0,  -9,   4,   1,  -1,   7,  44,  10, -19,  -3, -14, -27, -20,  -8,   0,  38,   0, -13,  -7,   4,   5,   3, -22,   1,  23,   7 },
     {    0,  -2,  -1,  -1,  -3, -14,  -3,   6,   1,   0,  -3,   8, -11, -31,   4, -13,   0,  -8,  -1, -17, -57,  15,  -6,   3,   2, -17, -12, -42,  25,  21,   4, -14, -11, -14, -16,  15,   7,  -5,   7, -18, -18,  15,  15, -12,   5,  21,  -9,  10,   9, -25, -23,  50,  -8, -13,  -4,   3, -18,   9,  -7,   1,   1,  -2,  -2,  11 },
   }
 },
 {  //31
   {
     {  115, -20, -13,  -3,  -4,  -1,  -1,   0, -13,  36,  -6,  -1,  -1,  -1,  -1,   0, -26,   8,  12,  -3,   1,  -1,   1,   0,  -2,  -7,   7,   2,  -1,   0,   0,   0,  -4,  -1,   0,   3,   0,   0,   0,   0,  -1,  -1,   0,   1,   1,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -4, -37,   7,  -1,   3,   0,   1,   0,  84, -10, -36,   2,  -6,   1,  -2,   0, -12,  66, -10, -14,  -1,  -3,   0,  -1, -13,   2,  34,  -5,  -2,  -1,   0,   0,  -2,  -5,   3,  14,  -2,   0,   0,   0,  -5,   0,  -2,   1,   5,  -1,   1,   0,  -1,  -1,   0,  -1,   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   38, -13,  -6,   0,   0,  -1,   0,   0,   0, -64,   9,  10,   2,   3,   1,   1,  31,  -4, -63,  13,  -2,   3,  -1,   1, -13,  55, -20, -27,   6,  -4,   1,  -1,  -1,   4,  24, -15,  -5,   0,   0,   0,  -1,   4,   6,   4,  -5,  -1,   0,   0,  -1,   1,   2,   2,   0,  -2,   0,   0,   0,   1,   1,   0,   1,  -1,   0,   0 },
     {    2,  47, -11,  -7,  -2,  -2,   0,   0, -42,  -7,  11,   0,  -1,   2,   0,   1,  12,  29, -19, -21,   1,  -4,   1,  -1,   8,   5,  68, -23, -13,  -1,  -2,   0,   9, -25,  14,  50, -15,   0,  -2,   0,  -6,   3, -20,  12,  20,  -4,   1,  -1,   1,  -6,  -1,  -7,   3,   5,   0,   0,   0,   0,  -2,   0,  -2,   0,   0,   0 },
     {  -16,   7, -15,  -4,  -2,   2,   0,   1,  -4,  38, -20, -18,  -1,   0,   1,   0, -15,   6,   0, -15,   0,   2,   2,   0,   0,   7,  -2, -43,   4,   0,   3,   0,  17, -14,  54, -39, -31,   9,  -3,   3,   7,  -4,  12,  38, -47,  -2,   0,   1,  -3,   9,  -6,  24,   2, -22,   4,  -1,   2,  -2,   3,   2,  11,  -7,   0,   0 },
     {   21,  92, -22,  -8,  -6,  -4,  -2,  -1,  68,  -2,  28, -11,  -4,  -3,  -1,  -1,  -7, -10,   9,  18,  -3,   2,  -1,   0, -21,   2,  -9,   6,   7,   0,   1,   0,  -5,  -6,   1,  -5,   2,   2,   1,   0,  -3,  -1,   2,   1,  -2,   0,   0,   0,  -1,  -2,   0,   2,   1,   0,   0,  -1,  -1,   0,   0,   0,   1,   1,  -1,   0 },
     {  -11, -26, -10,  15,   1,   0,   1,   0,  23,  30,  16, -22,  -1,   0,  -2,   0, -21, -32,  -3,  34,   2,  -1,   1,   0,  13,  13, -20, -38,   2,   4,  -1,   0,   4,  12,  35,  37, -21, -10,  -1,  -2,  -7,  -6, -24,   9,  49,   3,   0,   1,  -2,  -7,   1, -28, -15,  20,   3,   1,  -1,   1,  -4,   5,  -9, -11,   2,   0 },
     {    2,  -7,  12,   4,  -4,  -1,  -2,   0,   5,  -4,   3,   9,   0,  -7,   0,  -1,   7,  -5,  15,   8,  -9,  -1,  -3,   0,  -1,   2,  -8,  18,   1,  -8,  -1,  -1,   4,   7,   8,   7, -16,  -3,  -4,  -2,  -4,  22,  -7,  56,  10, -34,   0,  -5,   5,  -1,   6,  -9,  79,  -2, -16,  -1,   1,  -4,   3, -10,   2,  51, -11,  -2 },
     {    2,  -6, -10,   3,  -4,   1,  -1,   1,   7,  39, -14, -20,   2,  -6,   1,  -2,  82,  -3,  29, -15, -16,   0,  -5,   0, -15,  48,   5,  30, -11,  -7,   0,  -1, -15,  -5,  24,   6,  26,  -6,   0,   0,  -1, -19,  -2,   4,   6,  14,  -2,   1,  -4,  -1,  -7,  -2,  -4,   3,   4,  -1,  -1,  -2,   0,  -1,  -2,  -2,   0,   1 },
     {   17, -28,  28, -18,  -1,   1,   1,   0,  26, -38,  40, -22,  -4,   4,   0,   0,  25, -37,  31, -13, -11,   7,  -1,   2,  20, -26,  20,   1, -20,  10,  -2,   3,   7, -10,  -1,  18, -27,  10,  -2,   3,  -3,   6, -16,  24, -27,   4,   1,   1,  -8,  11, -16,  20, -17, -12,   8,   0,  -1,   4,  -4,   4,   7, -19,   3,   1 },
     {   14,  24,  29,  16,  -6,  -4,  -1,  -1, -21, -34, -39, -20,   8,   7,   2,   2,  24,  36,  35,  15, -11,  -9,  -3,  -2, -23, -34, -27,  -6,  19,  11,   3,   2,  18,  24,  13, -10, -27,  -9,  -1,  -1,  -7,  -9,   0,  18,  25,   6,   0,   1,   0,  -2,  -9, -19, -20,   6,   6,   1,  -1,   1,   3,   8,  -1, -15,   0,   0 },
     {    6,  16, -25, -10,  -1,   0,   0,   0,  13,  -1, -33,  -6,  -4,   4,   1,   1,  18, -18,  -5, -39,   5,  -2,   3,  -1,  40,  -3,  -6, -17, -38,  10,  -5,   2, -13,  60, -16,   9, -28, -20,   5,  -4,   5,   7,  39,  -8,  13, -23,  -3,  -2,   3,  -3,  13,  15,   0,   6,  -9,   1,  -2,   2,  -1,   6,   4,   2,   1,  -2 },
     {   -8, -16, -83,  15,   2,   4,   2,   0, -11, -46,   7, -38,   9,   2,   2,   1, -31,  16,  30,  -6, -30,   2,  -2,   1,  12,   9,  -3,  25,  -7, -16,   1,  -2,  19,   1,   6,  -2,  14,  -5,  -7,   0,   0,   8,   0,   4,   0,   8,  -2,  -2,   0,   0,   1,   0,   1,  -1,   3,   0,   0,   0,   0,   0,   2,  -1,   0,   1 },
     {   -1,  -3,  -4, -15, -10,   9,   2,   0,   4,   2,   4,  10,  22,  -5,  -1,   0,  -1,  -6,  -3, -24, -15,  14,   3,   0,   5,   1,   9,  10,  21,  -7,  -3,   0,  -5, -12,  -7, -31, -16,  25,   5,   1,  10,  10,  19,  21,   9, -18,  -5,   4,  -5, -13, -10, -19,   5,  62,   7,   1,  -8,  -1, -15,  -5, -55, -31,  32,   4 },
     {    6,  26,  44,  -5, -10,   1,  -2,  -1,   3, -26, -42,  16,  15,  -3,   2,   0, -49, -23,  12, -21,  -8,   4,   0,   2,  42,  34,  -5,  16, -16, -15,  -1,  -3, -12,  -3,  33,   7,  27,  -2,  -4,   0,   0, -22, -15,   8,   1,  17,   0,   1,  -5,   6, -10,  -9,  -1,  -2,   5,   1,   2,  -4,   2,  -1,  -2,  -2,   1,   1 },
     {   -2,  -5, -22,  32,   3,   5,  -1,   3,   1, -19, -30,  11,  27,  -2,   4,   0, -10, -45,  13, -31,  29,  17,  -1,   4, -68,  -5,   6,  -9, -20,  26,   2,   1, -17, -26,  -5,   5, -15,  -7,   9,  -1,  11,  -6, -19,  -5,   1,  -9,   2,   4,   4,   5,  -2, -11,   3,  -4,  -5,   3,   5,   5,  -1,  -3,  -3,   4,   0,  -2 },
     {   -8,  -6,  35, -37,  11,  -4,   9,  -3, -16,   0,  37, -29,  -3,   6,   2,   1, -35,  20,  11,  -8, -18,  15,   1,   0, -47,  28,  -5,   7, -30,   7,   7,   2, -15,   8,   7,   3,  -9, -26,   6,  -1,  17, -19,  28,  -3,   8, -15,  -8,   2,  18, -20,   7,  11,  -3,   7,  -9,   3,   5,   0, -15,  13,   3,   3,   3,  -3 },
     {    7,  -6,   8,  -7,  31, -17,   2,  -2,   7, -18,  14, -20,  14, -17,   0,  -1,  10, -11,  25, -22,  45, -23,  -1,  -1,  10, -16,  15, -27,  25, -13, -12,  10,  11, -12,  26, -18,  22,  -5, -24,   4,  10, -11,  23, -18,  12,  18, -21,   2,   6, -13,  21,   0,   5,  23, -13,  -3,   2,  -9,   4,   5,  -6,  37,  -3,  -4 },
     {    3,   2,   9,  -9, -10,  -4,  10,  -6,   1,   1, -12, -25,  20,  -2,  -1,   2,   8,   6,  14,   1, -21,  -2,   2,   0,  23,  21,  -9,  -9,  15,   2,   1,   4, -13, -31, -12,  -7, -20,  21,  13,   2,   4,  16, -28, -34, -14, -36,  13,   3,  22,  11,  51, -27, -10, -10, -24,   2,   2,  -4,   8,  40,  -1,   5,  -8,  -8 },
     {    5,   8,  17,  68,  14, -28,  -6,   3,   7,  12,  44,   3,  27,  -2,  -6,  -3,   2,  35,  -1, -46,   2,  31,   1,  -7,  26,  12, -33,  -7,  -4,  23,   2,   0,  10,  -2, -10,   4,   5,   4,   5,  -4,   0,  -2,   2,   7,  -4,  -1,   2,  -2,   4,   0,  -5,  -4,  -3,  -6,   6,   1,  -5,   3,  -2,  -7,  -1,   2,  -3,  -1 },
     {   -3,  -3, -28, -21,  28,   2,   3,   3,  -7, -17,  13,  34, -31,   3,  -1,  -4,   3,  27,  34,   9,  22, -10,  -1,   6,  27,   5, -17,  -7,   5,  37,   4,  -1, -59, -15,   9,  -7, -12,   6,  18,   0,  -7, -37,  -7,  11,   4,  -8,   4,   4,   5,  13, -11,  -5,   3,   2,   1,   4,   4,  -6,   1, -11,  -5,  -1,   1,   4 },
     {   -2,  -4, -16, -40,  -6,   6,  14,  -6,   2,   3,  15,  43,  18, -18,  -8,   2,  10,  13,  15, -12,   8,  30,   9,  -4, -16, -15, -19, -17, -24, -30, -12,   9,  23,  28,  14,   8,  12,  20, -11, -20, -13,  -5,  -7,   0,  -8,  12,  42,   3,  11,  -6,  -3, -29,  -8, -14, -11,  16,   2,  -5,   3,  15,  -9,   3,  -4,  -6 },
     {    7,   6,   7,  13,  26,  18,  -4,   8,  -2,  -8, -19, -22, -30, -25, -22,  10,   9,   2,   2,  34,   9,  48, -11, -11,  13, -13,   6, -13, -34,  -4,  16, -16,  19, -14,   4,  -6,  28,   7,  27,  -1,  12,  -1,   7,   2,   0, -17,   7,  40,   4,   0,   6,   5,   0,   4, -21,  -2, -18,  -1, -11, -11, -19,   2,   8,   3 },
     {   -1,   2,   5,  -3,   7,  23,  -4,  23,   0,   0,   7,  29,  13,  13,  10, -17,  -8,   4,  43,  20,  25,   5, -10, -19,  -3,  49,  26,  -2,   7,  23, -24, -15,  36,  25,  -9,  -7,  -4,  -1,  18,  -3,   2,  26,   2,   5,   3,  -4,  -8,   1, -22,   0,  26,  11, -17,  -9,  16, -10,  -5,  -4,  17,   1,  -8,   2,   6,  -6 },
     {    2,   1,  -6, -32,   5,  -4,  -1, -10,   6,   7,  -3,  24, -14,  -4,   4,  -3,   9,   7,  -5, -15,   0,   2, -10,   9,   5,  -8, -38,   1, -28,   9,  23,  -9,  39, -40,  -2,  -3,   3, -51,   0,  15,  12,  28, -16,   2,   6,   8, -37, -13,   5,  19,   8, -13,  -3,  14,  17, -10,   3,  -9,  20,  11,  -5,  -5,   3,   9 },
     {    2,  -5,  -8,  32, -39,  -8,   7,   2,   7,  -5,   6,  47, -20,  24, -10,  -2,   7,  -7,  14,  12, -16,  -5,  15,  -4,   8,  -7,   9,   5,   0, -16,  -6,   2,  29, -11,  18,   6, -11,  -4,  22,  13,  35, -41,  12, -12,  -6, -16, -15,  -5,  25, -23,  -9,  13, -10,  12, -17, -12,   4,   5,  -3,  22,   5,  13,   2, -11 },
     {   -1,   2,  -2,   0, -18,  19,   6,  16,  -5,  -2,  -9, -10,  -9,   1, -21,  14,  -1,   9, -11,   5,   9,  21, -32,   9,   1, -11, -17,  12,  -8,  33, -41,   1, -22,  -1,  17,  23,   0,  18, -20,   2, -10,  17,   2,  -2, -35,  16, -25, -22,   5, -40,   0, -11,  -7,  10,   9, -32, -13, -12,  -5,  17,  -6,  24,   8,  -7 },
     {    0,   3,  -5,  -5,  68, -36, -22,  27,  -1,   3, -20,  24,   2,  16, -11,   6,   1, -14,   4,  12, -38,  34,  30, -14,  -8,  -5,   5,   1,  15, -16,  -8,  11, -15,  -6,   4,  16,  -6,   0, -21,   8, -11,  10,   9,   6,  -3,   1, -11, -12,  11,  -1,  12,   9,  -5,  -1,   8, -10, -12,  14,  11,  14,   1,  -8,  -3,  -3 },
     {   -2,   0,   8,  33, -32,   7,   2,   1,  -4,  -6,  15,  21, -48, -14,   9,  -1,  -2,  -1,  24, -14, -16,  14,   6,  -4, -10,  -4,  14, -24,  -4, -16,  32, -16, -40,   9,  16, -20,   4,  -1, -32,   1, -10,  40,   5,  -8,  14,  10,  -8,  -4,  -2,  -3,  28,   2,  -8, -10,   7,   3,   9, -11,  -6,   5, -13,  -7,  -8,   1 },
     {   -3,  -3,  -7,  -8, -18, -10,  -7,   3,  -1,  -2,   5,  13,  42, -46, -10,  22,   5,  13,   7,  25,  15,  25,  -3,  -6,  14,  -3,  13,   2,  13,  -1,  18,  10, -18,   6,   8,  -4,  -8, -39,  -7,  19,  14,   0,  -1, -10,  -2, -11,  -9, -12, -38,   3,  -9,  -3,   8, -13,  -6, -15,  -4,  15, -38,  12,  27, -13,  16, -14 },
     {    0,   3,  -2,  -4,  -3, -36,  27, -27,  -3,  -2, -10,  -5, -29,  -7, -10,   9,  -2,   1,   5,   0,   7,  20, -21, -19,  -4,  14,   0,   6,  26,   3, -12,   9,  -2,  16,  -5,  20, -14,   5, -14, -10,  70,  14, -12,  -1,  -4,  26,  23,   3,  13,  30,   3,   9,   8,  10,  13,  -4,  10,   3,   5, -18,  -7,  -2,  11, -11 },
     {    0,  -1,   1,  -3,   6, -26, -29,  -2,   0,  -2,  -6,   4,  -7, -31,  21,   2,  -2,  -4,  13,   7, -11, -12, -29,  30,   0,  12,   1,  -9,   9,   3,   3, -28,  13,   2,  -3,  20, -13,  -3,  -3,  17,  10,  -2,  16, -20, -23, -10,  16, -26,  -3, -24, -19,  -8,  -5, -20,  -1,  24, -11,  20,  -4,  -9, -47,   9, -22,  35 },
   },
   {
     {  113, -34, -11,  -4,  -3,  -1,  -1,   0, -39,  24,   2,   1,   0,   0,   0,   0, -15,  -2,   3,   1,   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -4,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -20,  33,  -8,   0,  -2,   0,  -1,   0,-100,   3,  25,   3,   5,   1,   2,   0,  50, -37,  -7,   1,   0,   1,   0,   0,   9,   8,  -5,  -2,  -1,   0,   0,   0,   6,  -4,   1,   0,   0,   0,   0,   0,   3,   1,  -2,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
     {   28,  -5,  -1,  -3,   0,  -1,   0,   0,  22, -54,   7,   6,   1,   2,   0,   0,  74,  25, -37,  -3,  -6,   0,  -2,   0, -50,  39,  14,  -4,   1,  -2,   0,   0,  -3, -12,   5,   3,   1,   1,   0,   0,  -7,   5,  -1,  -1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 },
     {  -33, -93,  40,  11,   7,   3,   2,   1, -20,  43, -28,  -2,   0,   0,   0,   0,  38,  11,   1,  -7,  -3,  -1,  -1,   0, -12,  -4,   4,   4,   0,   0,   0,   0,   5,  -1,  -5,   0,   0,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   1,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -2,  29,  -2,  -5,  -3,  -1,  -1,   0, -34,   4, -14,   8,   2,   2,   1,   0, -14,  69,   8, -21,  -1,  -5,   0,  -1, -37, -50,  48,  10,   3,   1,   1,   0,  30, -25, -23,   7,   0,   2,   0,   0,  -1,   8,   1,  -4,   0,  -1,   0,   0,   5,  -4,  -2,   1,   0,   0,   0,   0,   1,   1,   1,   0,   0,   0,   0,   0 },
     {   -6,  26, -32,  10,  -2,   2,  -2,   1,  16,  78,  -7, -18,  -5,  -5,  -1,  -1,   2,  -4,   4,   3,   2,  -1,   1,   0, -40,  40,  19, -17,   3,  -3,   1,  -1, -11, -40,  38,   8,   1,   1,   0,   1,   9, -15, -12,   6,   0,   2,   0,   1,   1,  -2,   3,   0,  -1,  -1,   0,   0,   1,  -4,  -2,   2,   1,   0,   0,   0 },
     {   14,  18, -18,   6,  -4,   0,  -1,   0,  27,  21,  -5, -18,   2,  -2,   0,  -1,  45, -31,  50,   7,  -8,  -1,  -1,   0, -43, -38, -29,  31,  10,   6,   1,   1,  22,  26, -26, -26,  -1,  -1,   0,   0, -14,   8,  19,   2,  -4,  -1,  -1,   0,   0,  -4,  -3,   2,   1,   0,   1,   0,  -2,   2,   2,   0,   0,  -1,   0,   0 },
     {    8,  29,   1,  -8,  -4,  -2,  -1,   0,  -2,  18, -44,  12,   1,   1,  -1,   0,  35,  41,   6, -29,  -1,  -5,   0,  -1,  40,   5,   3,   4,  -6,  -1,  -2,   0, -44,  44,  20, -14,   3,  -2,   1,  -1,  -3, -36,  37,  10,  -3,   1,   0,   1,  -1,  -8, -13,  12,   2,   1,   0,   0,  -2,  -1,   1,  -2,   2,   0,   0,   0 },
     {   14,   5,  -5,  -3,   0,   1,  -1,   0,  28,  10,   0,  -6,  -2,  -2,  -1,  -1,  42, -23,  18, -20,   1,  -2,   0,   0,  50, -24,  44,  11, -18,   2,  -4,   1, -23, -22, -24,  52,   2,   3,  -1,   1,  -5,  20, -39,  -8,  11,  -1,   2,   0,  -7,  10,   7, -11,   1,  -1,   0,   0,  -1,   0,   1,  -1,   0,   2,  -1,   0 },
     {   19,  39,  94, -18, -14,  -9,  -2,  -2,   5,  11, -34,  20,   7,   1,   0,   0,  -8, -43, -20,  -3,   7,   6,   2,   1, -26,  -2,   5,   1,  -4,   0,   1,   0,   0,  -5,   0,   2,   0,  -1,   0,   0,   8,  -1,  -5,   1,   0,   0,   0,   0,   0,   5,  -2,  -3,   1,   0,   0,   0,   0,  -2,   2,   0,  -1,   0,   0,   0 },
     {    1,   2, -16,   1,   1,   2,   0,   0,  14,  26,  -8,   9,  -7,  -3,  -1,   0,   1, -19, -45, -10,  13,   2,   2,   0,  19,  13,  29, -26, -15,  -2,  -1,   0,  69,  31, -10,  10,  -4,  -5,  -3,  -1, -44,  18,  32,  -5,   3,   1,   1,   0,  -2, -22,   3,  15,   2,   1,   0,   0,  -4,   4,  -6,  -4,   2,   2,   0,   0 },
     {    5,   1,  10,   5,  -5,  -1,  -1,   0,  -2, -15,  -7, -15,   7,   1,   1,   0,   9,   7,  20,   3,  -8,  -2,  -1,   0,   6, -17,  -3, -26,   4,   4,   0,   1,  33,  12,  52,  16, -46,  -4,  -4,   0, -13,   1, -24,  66,  25, -13,  -1,  -2,  -7,   1, -16, -27,  23,  10,   0,   0,   1,  -2,   4,  -1,  -9,   4,   0,   0 },
     {    1,   0,  -9,  28, -10,   1,  -3,   1, -14, -34, -66,  -6,  25,   5,   4,   0, -14, -14,  46, -25, -17,   1,   0,   0,  11,  55,   2,  15,  -2, -10,  -3,  -1,  27, -25,  -4,  -5,  13,   6,   0,   0, -15,  10,  -2,  -6,  -7,   3,   1,   1,  -1,   0,   4,  -2,  -3,  -3,   0,   0,   2,   0,   1,   2,   0,   0,  -1,   0 },
     {  -10, -16,   2, -11,   4,   1,   2,   0, -10, -17,  29, -10,   2,  -1,   1,   0, -24, -25,  34, -32,   4,   4,   0,   2, -45,  19,  41,  13, -36,   3,  -5,   1, -21,  44,  -1,  34,  -1, -12,   0,  -2,  -3, -26,  12,   8,  15,   0,   1,   0,  12, -17, -14,  10,   3,   2,  -1,   0,   3,   0,  -8,  -2,   3,   1,  -1,   1 },
     {   -4, -12, -45,  32,   0,   4,  -2,   2,  -5, -21, -49,  28,   7,   1,   1,   1,  -4, -37, -23, -14,  29,   0,   4,   0, -26, -41,  13, -36,   2,   9,   2,   2, -29,  13,   6,  -9, -16,  -2,   0,   0,  28,   6,  -4,  -5,   0,  -1,  -3,   0,   5,  10,   4,  -7,  -8,   4,   0,   0,   2,   1,   2,   3,  -3,  -5,   2,   0 },
     {    6,   7,  14,  70, -22,  -3,  -2,  -2,  -1,   0,  10, -68,  33,   0,   1,   1,  -6,  -1, -42,   7, -30,   5,   1,   0,   6, -15,  27,  14,   5, -12,  -2,   0,  -9,  19,  -4,  -7,  11,   6,  -2,  -1,   6, -11,   2,   3,  -5,   3,   2,  -1,  -1,   2,  -3,   0,   1,  -1,  -1,   1,   0,   2,   0,   2,   1,   1,   0,  -1 },
     {   -1,   7,  13,   0,  -3,  -2,  -1,   0,  -6,   2,   1, -14,   6,   0,   0,  -1,  -1,  15,  14,   0,  -4,  -1,  -1,   0,  -5,   4,  -5, -15,   1,   3,   1,   1, -22,  22,  12,   4, -15,  -2,   0,  -1,   1,  52,   7, -14,  10, -10,   1,  -2,  -8, -36,  76,   3,  -5,   4,  -1,   1,   9, -40, -21,  36,  -1,   4,  -1,   2 },
     {    8, -11,  12,  -1,   2,   0,  -1,   0,  11, -19,  11,  -5,   3,   1,   1,   1,  15, -19,  21,   1,   4,  -3,   0,   0,  20, -27,  21, -12,   5,   3,   1,  -1,  23, -42,  33, -10,  -2,   3,   1,   1,  32, -29,  30,   2, -18,   5,  -2,   0,   7, -25,   1,  49, -14,   3,  -3,   1,  -1,   5, -48,  18,   8,  -1,   1,   1 },
     {   -1,   0,   7,  10,  -2,   1,   0,  -1,  -4,  -2,  -3,   5,   4,   1,   1,  -1,  -4,  -3,  -1,  13,   0,   1,   0,   1,  -1, -15,  -9,  -7,  15,   1,   1,   0, -37, -23,   6,   3,   3,   6,   1,   1, -72,   4,  -2,  -4,   8,   2,   3,  -1,  70, -43, -10,  -9,  -7,   4,  -1,   1,   2,  30, -13,   2,  -6,  -4,   0,  -1 },
     {    3,   0,   0,  -1,   6,  -5,  -3,  -1,   1,  -4,   4,   3,  -4,   1,   2,   1,   4,   0,   2,  -7,  -4,   4,   1,   1,   0, -11,   0,  -3,  -6,  -3,   2,  -1,  23,  14,  19,   1,   8,   2,  -2,   0, -26, -44, -19, -14,  -9,  11,   5,   0,  48,  53,  32,  -4, -22, -11,  -4,  -1, -34, -43,  17,  37,   8,   1,   2,   1 },
     {    7,   8,  19,  65, -20,   3,  -6,   0,   8,  10,  39,  22,   0,  -9,  -1,  -2,   3,  18,  10, -61,  34,   3,   4,  -1,   0,   5, -45, -13, -17,  12,   0,   1,   6, -17,  -7,  11,  -8,  -6,   2,   0,   1,   4,   9,  -2,   3,  -3,  -2,  -1,   1,   9,  -3,   2,   0,   2,   1,   0,  -8,   4,  -3,  -2,   2,   0,   1,   1 },
     {    5,  12,  11,  22,   9,  -5,   0,   0,  -7,  -6, -20, -18, -27,  16,   1,   2,   5,  12,  15,  40,   3, -24,   1,   1,  -2,  10, -12, -30,   9,  18,  -2,   0,   9,  12, -24,  40,  -7, -17,   5,   0,  10, -28,   4, -31,  42,  14,  -7,   3, -13,   1,  -9,  -6, -39,  10,   5,   0,   0,  19,  -9,  10,  13, -15,  -2,  -2 },
     {   -1,  -3, -13, -14,  -5,  11,  -2,   2,   1,   0, -10,  -1,   8,  -4,   1,  -1,   1,   1, -25, -18,  -7,  19,  -1,  -1,  -4, -24, -33,  41, -30, -18,  11,  -2,  16,  -4,  44,  16,  35, -25, -14,  -5,   8,  -3,  -8, -14,  32,  30,  -1,  -4, -15, -20,   3, -19, -22,  18,  13,   5,  11,  12,  -9,   9, -12, -14,   1,   1 },
     {   -1,   4,  -2, -36,  24,   7,  -2,   2,  -5,   2, -14, -54,  22,   0,   1,  -1,  -4,   6, -12, -22,  -6,  10,   0,   2,  -5,  -1, -25, -18, -28,  12,   0,   1, -13, -10,  -4,   8, -27,  -3,   1,   0,  -2,  40,  11,   7, -12,   7,  -7,  -4,  12,  34, -19,  35, -19,  -3,   4,  -4, -26,  25,   0,  11,   7, -16,   2,  -1 },
     {    3,  -3,  10, -27,  16,  -5,   2,  -2,   7,  -2,  18, -34,  30,  -8,   4,  -4,   6,  -3,  13, -33,  28,  -8,   4,   1,   9,   3,  15, -36,  22,  -1,  -4,   3,   9,  -4,  -3, -33,   8,   2,  -3,  -2,   7, -15,  12, -26,   9,   4,  -3,   2,   3,  -9,   7, -41, -14,  14,  -2,   3,  26,  16,  42,   3, -34,  -2,  -5,   1 },
     {   -5,  -8,  -6, -20, -76,  18,   7,   3,   0,  -1,  -2,   2,  49, -29,  -4,  -1,   7,  14,  15,  44,  23,  20, -12,  -1,   2,   0,   2, -26, -31,   9,  13,   3,  -1,  -4, -15,   1,   8, -19,  -1,   2,  -5, -12,   4,  -3,   3,   4,  -7,   2,   1,   8,   5,   4,   4,  -4,   1,   1,  -3,  -2,   2,  -6,   5,   9,  -2,   0 },
     {   -5,  -5,   6,   2,  -9,   1,  -5,   0,  -7,  -8,  11,   2, -12,  -4,  -3,   3,  -6, -10,  10,  -4,  -5,   1,   7,   2, -11, -17,  11, -19,  14, -18,   2,   3, -25, -16,  15,  -8,  12,   0,  -4,   4, -66,   0,  17,  -1,  -1,   7,   3,  -1, -71,  31, -11,  19, -18,   7,  13,  -2,  15, -17,   3,   6,   2, -20,   5,   3 },
     {    6,   8,  12,  20,  37,  10,  -1,  -3,  -9, -15, -16, -34, -41, -30,  11,  13,   4,   6,   8,  24,  60,  13, -24, -10,   1,  -3,   5,   7, -31,  -3,  25,   5,  -3,  -6,  10,  -9,  16,   1, -22,  -6,  -6,   8,   5,   6,  -9,   4,  19,   0,   0,  -4,  -1,  -2,   6, -18,  -3,   7,   2, -15,   8, -19,  -4,  14,  -5,   3 },
     {   -4,  -6,  -4, -20, -55,  38, -18,   1,  -3,  -5,  -6, -26, -26,  -2,  11,  -3,  -4,  -3, -14, -12,  44, -52,  14,   3,   1,   9,   0,  34,  30,   3, -17,   5,   2,   6,   3,   0,  -9,  20,  -7,  -2,   1,   6, -12,   7,   5,  -5,  11,  -1,   9,   2,  -8,  12, -13,  12,   6,   3,  -8,  -9,  -6,   9,  -6,  -3,  -5,   2 },
     {   -2,  -4, -11,  -2,  12, -11,  12,  -6,  -1,  -2, -11,  -8,   9,   1,  -1,   9,  -5,  -9, -26,   3,   9,   2,  -6,   8,   3,   1,  -8,  41, -17,  36,  -7,   5,  -6, -41, -23,   0, -41,  -1,   2,  -2, -18, -42,  33,  11,  24, -24, -18, -10, -11,   2,  26,  -5,  14,  16,   1,  -8,   9, -21,   5,   7, -16,  -3,  -3,   4 },
     {   -1,   0,   5,   6,   1,  -7,  21,  33,   0,   3,  11,  15,  17, -20,   9,  18,  -3,   0,   5,  11,   7, -26, -25, -18,   0,   4,  17,  22,  -2,  29, -18,  -4,   0,   8,  26,   2, -12,  24,  -4,   4,  -4,  14,  15,  -2, -28,  28, -20, -16, -15,  -1,   6, -41, -23,  -9,   1, -16, -29,  17,  -5,  -9,  10, -24,  22,   7 },
     {   -1,   0,  -1,   1,  17, -32, -11, -28,  -1,   1,  -2,  -2,  29,   4,   3,  11,   0,   3,   3,   6,  17, -40, -15,  -2,  -6,  -4,  -7,  -4,  -3, -16,  -9,  12,  -2,   5,   9,   1,  21,   2,  -2,   4,  -8,   4, -16, -20,  10,  10, -25,   2,  -4,   7,   1,  25,  34,  55, -11,  -8, -37, -11, -18, -31, -15,  16,  10,   1 },
   },
   {
     {  111, -14, -29,  -4,  -5,  -1,  -2,  -1, -37,  33,  12,  -5,   0,  -1,   0,  -1, -11,  -9,  10,   5,   0,   0,   0,   0,  -3,   0,  -2,   2,   2,   0,   0,   0,  -3,  -2,   1,   0,   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    3, -92,  17,  22,   4,   5,   1,   2,  29,  26, -59,  -8,   5,   1,   1,   0,  -9,  30,  18, -22,  -6,  -1,   0,  -1,  -1,  -5,   9,   8,  -4,  -2,   0,   0,  -3,   4,   0,  -1,   1,  -1,  -1,   0,   0,  -1,   2,   1,  -1,   0,   0,   0,  -1,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
     {  -30,  17, -41,  12,   5,   1,   1,   1,  28,  61,  15, -55,  -8,  -3,  -1,  -2,  -4, -19,  55,  24, -20,  -4,  -3,   0,  -1, -13, -16,  21,  14,  -2,   0,   0,  -1,  -3,   1,  -5,   2,   3,   0,   1,   0,  -3,  -3,   3,   0,   0,   0,   0,   0,  -1,   1,  -1,   0,   1,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0 },
     {   37,  40,   7, -24,  -6,  -5,  -1,  -2,  92,  -8, -11,  14, -13,  -2,  -4,   0, -46,  26,   2,   7,  12,  -4,   0,  -2, -10, -14,  10,  -3,   6,   6,   0,   0,  -7,  -2,  -3,   3,  -2,   2,   3,   0,  -3,  -3,   0,  -1,   2,  -1,   1,   1,  -2,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   -6, -54,  11,   3,  16,  -1,   1,   1,  17,   6,  46,   8, -31,   1,  -5,  -1,  -7,  -9, -28,  61,  26, -15,  -2,  -2,  -3,   5, -27, -33,  31,  21,  -3,  -1,  -1,   3,   4,  -9, -19,   8,  10,  -1,   0,   2,  -4,  -2,   1,  -6,   2,   3,   0,   1,   1,   0,  -2,   1,  -1,   0,   0,   1,  -1,  -1,   1,  -1,   0,   0 },
     {   13, -16, -55,  14,  18,  -2,   4,   0,  17, -71,  14, -37,  -1,  19,   1,   5,  21,  39,  -9,  10, -39,  -9,   2,  -1, -13,  19,  11,  10,  14, -15,  -5,  -1,  -1,  -1,   3,   0,   1,   7,  -2,  -2,  -3,   4,   0,   2,   1,  -3,   3,   0,   0,   0,   1,  -1,   0,   1,  -1,   0,  -1,   1,   0,   1,   0,   0,   0,   0 },
     {   17,   3,  80, -16, -17,  -9,  -1,  -3,  -9, -15,   7,  -8,   1,   7,  -1,   1,  21,   7,  27,  28, -37, -11,  -4,  -1,  -8,  -4, -19,  48,  33,  -9,  -7,  -4,  -3,   0, -16, -16,  16,  17,   5,  -2,  -1,  -1,  -2,  -1,  -4,  -1,   5,   2,   0,   0,  -3,  -3,   2,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 },
     {    3, -32,  11, -24,  -4,   9,   4,   1,  31, -14,  39,  -8,   5, -10,  -2,   2, -19, -40,  -8,  -3, -22,  41,   3,   4,   2,  19, -25,  15, -47, -34,  21,   4,  -2,   9,   4,  10,  25, -27, -22,   4,   1,   2,  -2,  -1,   3,  12,  -5,  -9,  -1,   1,   0,   2,   2,  -3,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,   0 },
     {   -8, -12,  -6, -60,   8,  13,   2,   2,   1,  31,  25,   7, -44, -14,   4,   0,  52,  34,  -4,  -7,  10, -12, -12,  -1, -22,  -2,  41,  19, -17,   3,  -1,  -4,  -1, -11, -13,  26,  16,  -9,   0,   0,  -3,  -2,  -2,  -7,   7,   7,  -2,   0,   0,  -2,  -2,   3,  -1,   0,   1,   0,   0,  -1,   0,  -1,   0,   1,   0,   0 },
     {  -18, -17, -24, -46,  20,  14,   2,   2, -35, -34,  -3,  19, -26,   2,   8,   2, -59,  11,  42,   2,   3, -17,  -3,   0,  42, -23, -11,  17,   1,   3,  -6,  -1,   1,  17, -14,  -6,   5,   1,   1,  -2,   6,  -1,   5,  -5,  -3,   2,   0,  -1,   0,   3,  -1,   2,  -1,  -2,   1,   0,   1,   0,   1,  -1,   0,   0,   0,   0 },
     {  -10,   3, -10,   5, -20,  -1,   5,   1,  -7,  25,  -7,  23, -13, -13,  -4,   4, -17,  10, -28,  13, -13, -18,  16,  -3,   9,   4,  13, -11,  33, -63, -16,  13,   8,  -6,  15,  -4,  32,  28, -52,  -6,  -1,   0,  -1,   2,  -7,  25,  11, -21,   1,  -1,   1,   0,   2,   0,   6,   2,   0,   0,   0,  -1,   0,   2,   1,   0 },
     {    5,   2, -10, -60,  -9,  16,   6,   3,   9,  -7, -63, -12, -14,  -1,  11,   4,  28, -42,   2,  12,  -8, -15,  -3,   3,   1,  44, -19, -21,   9,  -1,  -8,  -3,  -5,   6,  33,  -9, -17,   1,   0,  -3,  -1,   0,   4,  17,  -2,  -7,  -2,   1,  -1,   1,   2,   1,   4,  -1,  -3,  -1,   0,   0,   1,   2,   1,   1,   0,  -1 },
     {    1,  11,  40, -12,  32, -15,  -1,  -2, -15,  19,   3, -49, -19,  45,   3,   0, -33,   3, -24,  -4, -31, -17,  23,   6,  20,   0,  27, -22, -17,  -9, -12,   4,   4,  -6,  12,  30, -20, -12,   0,  -4,   1,   1,  -3,   8,  17, -11,  -6,   1,   1,  -1,  -1,   2,   0,   4,  -2,  -2,   0,   0,  -1,   1,   2,  -1,   0,  -1 },
     {   -8, -17, -16, -18, -11,  -5,   7,   4,  16,  14,  22,  21,  29,  10, -17,  -9, -10, -30, -16, -26, -17,   0,  18,   7,  17,  22,  23,  21,   9,   2, -39, -14,  -8,  -4,  -8,   8,   6,  45,  43, -16,   2,  -1,  -2, -10, -14, -25,  17,  27,  -2,  -1,  -2,   2,   3,   1, -10,   1,   0,   1,  -1,  -2,  -3,  -2,   2,   0 },
     {  -14,  -7,  -6, -48, -20,  -1,  21,  -1, -18,  13,   6,  -8,  45,  -3, -32,   3, -10,  45,   2,  19, -14,  45,  13, -19, -11, -12,   3, -22,   9,  -1,  30,  12,  11, -13,  11, -13, -29,   6,   0,  10,  -1,   0,  -2,  15,  -8, -18,   4,   1,   1,   1,   0,  -2,   3,  -1,  -6,   2,   0,   0,   1,   2,  -2,  -1,   0,  -1 },
     {   -3,  -2,   9,  16, -71,  13,  10,   1,  -9, -10,  14, -35,  -6, -47,   5,  12, -24,  20,  -1,   6,   4, -16, -36,   3,  30,  26,  18,  -3,  -6,  14,  -7, -23,   4,  -1,  19,  22,   0,  -6,  13,  -2,  -2,   2,  -7,   9,  11,  -2,   0,   4,  -1,   1,   0,  -2,   5,   3,  -1,   0,   0,  -2,  -1,   1,   0,   1,   0,   0 },
     {   10,  12,   6,   6,  19,  -9, -10,  -1,  15,  18,  22,  19, -10,  -1, -10,  -4,  34,  42,  12, -26,  -8,   8,  -1, -11,  75,  26, -35,  -9,  -2,  -2,   6,   3, -26,  15,  27,  -8,  -7,   3,  -1,   3,  -2, -12,   7,  12,  -3,   0,   4,   0,  -6,  -2,  -1,   0,   5,  -1,   0,   3,   0,  -4,   0,   1,   0,   1,   0,   0 },
     {    8, -11,  16, -19,  23, -17,   5,  -2,  11, -20,  23, -33,  34, -33,   4,   6,  14, -18,  27, -28,  31, -10,  -8,  10,   8, -18,  16, -20,   8,  10, -29,  26,  -2,  -1,  -1,   9, -19,  34, -38,  -1,  -4,   7,  -9,  13, -14,  17,   6, -20,   0,   2,  -1,   2,  -2,   0,   7,  -1,   0,   0,  -1,   1,  -1,   0,   1,  -1 },
     {   -6,   2,  10,  10,  34, -17, -12,  -8, -10,  -2,   8,  29, -12, -11,  -9,  -1, -28,   4,  49, -12,  -6,   9, -13,  -1, -39,  66,   6,  -6,   8,   0,   3,  -9,  21, -42,  24,  13,  -3,   6,   1,   2,  -1,   3, -14,   5,   3,   4,   2,   0,   0,  -2,  -1,  -4,   4,   1,   3,   0,   1,  -1,  -2,   0,  -1,   1,   1,   0 },
     {   -3, -10,  -7,  -7, -13, -15, -11,   3,  10,   5,  12,  17,  14,  10,  23,   3,  -3, -16,  -6, -10, -17, -26,  -7,   9,  22,  10,  27,   7,  26,   4,  26,  -2, -15, -30, -20, -22, -13, -19,  -6,  60,   2,   9,  -3,   6, -11,  -8, -53, -21,  -3,  -1,   0,   0,   7,  11,   6, -15,   1,   1,   0,   1,   0,   0,  -1,  -1 },
     {    1,  -9,  -6,  13, -42, -18,  17,   6,   6, -15,   4,  36,  -9,  17, -30,   1,  24, -15,  50,  13,  -9, -16,  33,  -7,   5,  -9,   7, -20,  -9,  -8, -10,  24,  -7,   3, -11,  40, -20, -20, -11, -11,  -3,   4,  -6,  -1,  34, -12,  -8,  -2,   1,   3,   1,  -4,   1,  14,  -6,  -2,  -2,   0,   0,   0,   1,   0,   4,  -2 },
     {   -2,  -2,  -3,  -2, -13,   5,   1,   5,   2,   8,  27,  20, -10, -12,  21,   3,  -5,  10,  -8, -48, -31, -17,  -4,  27, -40, -29, -32,   0,   5,  -8, -25,  -7,  28,  24,  24, -19, -34, -10,  -1, -10,  -5,   1,  11,  29,  11, -17, -14,  -1,   3,   1,  -4,   0,  10,   7,  -4,  -7,   0,   1,   1,   1,  -1,   3,   1,  -1 },
     {   -7,  -9, -11, -12, -30, -86, -30,  17,  -7,  -8,  -9, -24, -45,   1, -12, -12,  -4,  -6, -13, -24,  23,  30,   4, -14,  -4,  -2,  -7,  20,  15,   0,   3,   3,   3,   5,   4,  -7,  -5,   1,   1,  -4,   6,   4,   7,   2,  -3,   2,  -1,  -1,  -1,  -1,   1,   5,  -2,   1,   2,  -1,   1,   0,  -1,   1,   0,  -2,  -1,   0 },
     {    6,   3,   2,  -1,   4, -10, -12,   2,  10,   8,   1,  -1,   5,  -8,  -5,  -7,   6,  -2,   8,   9, -12,  -7,  -7,  -9,  17,  30,  20, -22, -10,  -1,   5,   2,  81,  44, -50, -21,  -1,   0,   1,   1, -23,   5,  29,  -3,  -1,   0,   4,  -1,   1,  -9,   1,  10,  -3,   1,  -3,   6,  -6,   0,   0,   0,   0,  -2,   0,  -2 },
     {    8,   3,  11,  19,  11,  46,  15, -13,   1, -12,   0,  -9, -52, -28, -32,  15,   0, -32,  -7, -33, -15,  17,  10, -45,   6, -15,  21,   7,  26,   1,  21,  10,   2,  -6,  -7,  -7, -18,  11,   1,  13,  -3,   9,   7,   7,  -3, -18,   6,   2,  -7,  -2,  -2,   5,   7,   1,  -8,   1,   1,   0,   0,   1,  -2,  -1,   2,  -3 },
     {    3,   0,  -6,   3,  35, -52,  45, -12,   0,   3, -18,  24,   2, -30, -14,  25,   7, -16, -13,  24, -34,  15, -46,   9,  15, -29,  19,   2,  -2,   2,  -2, -24,  -3,  12,  13,   4,   6,  -7,   6,  -1,  -3,   6,  -4,   6,  14,   0,  -2,   4,  -1,   7,  -5,  -2,   1,   3,  -1,  -1,  -1,  -1,  -1,   0,   2,   2,   0,  -3 },
     {   -4,  -1,  -7,  -1, -10,  11, -33,  -7,   0,  10,  -2,  13,   1,  25,  -5,  11,  -6,  -3, -12,   4, -24,  12, -46,  -9,  -2,   7,  -1,  10,  -7,  22,  -7,  -7, -20,   0, -28,  10, -37,   9, -26, -25,  14,   8,   1,   3,   7, -10,  37, -57,   2,  -6,  13,  -4,  10,  -6,  19,  11,  -1,   6,   1,   2,  -1,  -3,  -1,   0 },
     {    3, -11,  -2,  -6, -13,   8, -36,   1,   9, -10,   4,  10,   5,  28,  18, -22,  13, -18,   8,  -5, -14,  -1, -17, -15,  17, -45,   4, -17,  12,  -4,  32, -27,  34, -19,  33,  30,   6,  15,  11,   6, -26,   4, -12,  11,  -1,  25,  21,  15,  -4,   0,  -5, -12,  -7,  -7,  12,  14,   0,  -8,  -6,   6,  -7,  -6,  -5,   1 },
     {   -4,  -2,   3,   5,   3,   4, -10,   7,  -4,  -3,  10,   6,  12,  15, -26, -21,  -8, -11,   4,  17,  11, -19,  -3, -18, -18,   6,  44,  25, -10,  11,   7,   8, -17,  48,  40, -14,   1,  -2,  -2,   7,   3, -35,  22,  37,  -1,  15,  -1,  -7,  -5,   3, -31,  -9,  23,   3,   9,   1,  12,  -6,   2,  -6, -10,   4,  -5,  -2 },
     {    2,   2,   3,   4, -11,  -5,  69,  22,   4,   3,   8,  -3, -10,  60, -11,  21,   1,  -1,   5, -18,  34, -17, -29, -15,   0,   5, -14,  10, -20,  -6,   6, -11,  22, -16,   9, -15,   8,  14,  -8,   4, -16,  12,   4,   2, -21,  -8,   3, -16,   0,  -2,   5,  -1,  11,  -8,   3,   1,   3,  -3,   4,  -3,   0,   2,  -1,   2 },
     {   -4,  -1,  -5, -15,   2,  -4,  12, -30,  -2,   7,   1, -10,  27,   5, -24,  17, -10,   0, -15, -30,   1, -27,  -4, -39, -14,  11, -36,  -1,  31,  12,  -1, -14,  -4,  13, -11,  29,  14,  -9,   4,  14,   1, -13,  23, -12,  36,  33,  -9,   0,   0,  -4,  -2,  -1, -26,  25,  21,   7,   1,   5,  -2,   4,  -4,  -8,   2,   3 },
     {    2,   2,  -2,  -4,  24,   8, -17,  50,   0,   0, -10,  -3,  18, -13,  -8,  -8,  -2,   0, -18,   5,  21,  -2, -13, -21,  11,   0, -13,  25,   5, -17,  -8,   5,  10, -15, -15,   4,   2,  -1,  -2,  -6, -33,  -8, -24,  19,  43, -21,  -8,   0,   4,   0,  15, -42,  20,  30, -17, -13,   2,   3,   6,   3, -14,  13,  13,  -1 },
   }
 },
 {  //32
   {
     { -112,  17,  14,   3,   4,   1,   2,   0,  13, -43,   6,   3,   2,   1,   1,   0,  23,  -5, -19,   5,  -1,   1,  -1,   0,  -1,  10,  -8,  -6,   3,  -1,   0,   0,   5,   0,   3,  -7,  -2,   1,   0,   0,   1,   2,   1,   1,  -4,  -1,   0,   0,   2,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1 },
     {    6,  42, -10,   2,  -3,  -1,  -1,   0, -75,   7,  44,  -4,   5,  -1,   2,   0,   9, -67,   4,  22,   0,   4,   0,   1,   5,   4, -36,   0,   5,   0,   1,   0,   3,   1,   3, -14,  -2,   1,   0,   0,   3,   1,  -2,   3,  -6,  -1,   0,   0,   1,   0,   0,   0,   1,  -4,   0,   0,   1,   0,   0,   0,   1,   0,  -2,   1 },
     {  -42,  23,   7,  -2,  -1,  -1,   0,   0, -13,  49,   0,  -7,  -3,  -4,  -1,   0, -10,   0,  63, -15,  -3,  -4,   0,  -1,  13, -41,  25,  42, -15,   2,  -3,   1,  -5,   1, -29,  30,  15,  -7,   1,  -1,   1,  -5,  -4, -10,  18,   2,  -2,   1,   1,  -2,  -2,  -3,   0,   7,  -1,   1,   0,  -1,  -1,  -1,  -1,   1,   1,   0 },
     {    8,  29, -12,  -4,   3,  -4,   0,  -1, -22, -20,  22,   3,  -6,   3,  -1,   1,   6,   8, -35,  -8,   8,  -6,   1,  -1,  13,  15,  48, -30, -23,   2,  -4,   0,   4,   1,  11,  65, -10, -16,   0,  -3,  -3,   1, -10,  -2,  49,   1,  -5,   0,   1,  -6,   1,  -9,  -6,  23,   2,   0,  -2,   1,  -4,   0,  -3,  -3,   7,  -1 },
     {  -13,  14,  -9,  -5,  -7,   1,   0,   1, -14,  25, -11, -16,  -3,  -3,   2,   0, -10,  11,   7, -21,  -7,   0,   2,   1,   5, -12,  24, -34,  -7,  -1,   2,   2,  12, -14,  26,   1, -54,   9,  -2,   4,  -5,  14, -11,  49, -31, -35,   8,  -2,   3,   0,   5,   5,  33, -34,  -9,   2,   2,  -1,   3,   1,   9,   8, -18,   3 },
     {   25,  43,  19, -10,  -7,  -2,  -2,  -1, -44, -54, -22,  18,  10,   3,   2,   1,  33,  41,   3, -29, -12,  -2,  -2,  -1, -13, -10,  24,  32,   6,  -2,   0,   0,  -1, -10, -25, -19,   8,  10,   2,   1,   0,   4,  10,  -1, -20,  -5,   1,   0,   1,   0,  -3,   5,   7,  -6,  -2,  -1,   0,   0,   2,  -1,   0,   2,  -1,   0 },
     {  -18,  30, -25,   0,   6,   5,  -1,  -1, -23,  40, -26,  -3,  10,   6,  -1,  -2, -19,  25, -12, -18,  19,   3,   2,  -3,  -1,   3,  13, -33,  20,   5,   0,  -2,  11, -16,  27, -32,  14,  17,  -3,  -2,   9, -13,  16, -14, -15,  36,  -4,  -1,  -4,   5,  -7,  11, -34,  25,  15,  -7,  -2,   1,  -3,   3, -10, -10,  25,  -6 },
     {  -18, -85,   3,   5,   6,   4,   2,   1, -70,   2, -36,   1,   4,   2,   1,   0,  24,  -8,  -3, -27,  -1,  -1,   1,   0,  18,  11,   2,  -2, -17,  -1,  -2,   0,   1,  14,   3,   5,   1,  -8,  -1,  -1,   5,   0,   7,  -2,   4,   1,  -3,  -1,   1,   3,   1,   2,  -2,   2,   1,  -1,   1,   0,   0,   1,   0,  -1,   1,   1 },
     {    1,   1,  -4,  -8,   4,  -3,   3,   1,   2,  -1,  10,   1, -17,   6,   2,   2,   1,  -7,  -9, -10,  10,  -3,   3,   2,   3,   5,  13,  -2, -27,   8,   2,   2,  -3,  -9,  -4,  14,  17,  -6,   5,   3,  -9,  -7, -20, -21, -35,  33,  11,   6,  -6,   0, -11,   6, -37, -65,  26,   5,   5,  -5,   7,  -7,  22, -32, -44,  13 },
     {   -5, -18,  16,   5,   9,   1,   3,  -1, -21, -18,   7,  23,   4,   8,  -1,   2, -53,   3, -22,  18,  22,   1,   6,  -2,  11, -56,   4, -26,  24,  16,   1,   2,  -3,  -2, -46,   0, -24,  23,   6,   0,  -5,   9, -12, -28,  -4, -15,  14,   1,   6,  -4,   8, -12, -12,  -2,  -5,   4,   0,   4,  -2,   3,  -7,  -4,   0,  -3 },
     {    4,   6,  14,  13,  -9,  -8,   4,   0,  -5, -13, -22, -21,   6,  15,  -1,   2,  10,  15,  26,  17, -17, -14,   5,   0, -13, -21, -25,  -9,  16,  21,  -1,   4,   8,  17,   9,  -5, -43, -21,   9,  -1,   1,   1,   6,  25,  36,  10,  -9,   1,  -8,  -6, -18, -18, -30,   8,  35,   6,   1,  -6,   2,  -4,   0, -46, -18,  12 },
     {    4,  24,  17, -21,   3,  -3,  -1,   1,  11, -17, -22,  26, -12,   3,   0,  -1, -44, -23,  -7, -32,  19,  -3,   2,   1,  22, -14, -22,  -8, -42,   7,  -1,   1,   1,  40,   2, -12,  -3, -37,   0,   0,  14,  -4,  38,   9,  -1,   0, -20,  -3,  -5,  10,  -6,  23,  10,   2,   1,  -8,   2,  -3,   4,  -4,   9,   5,   2,   4 },
     {  -12,  21, -26,  22,  -6,  -6,   1,   1, -20,  31, -37,  24,  -9,  -5,   1,   2, -18,  31, -30,  22,  -3, -14,   2,   1, -20,  23, -23,   8,   6, -13,   5,   0,  -7,  12,  -9,  -2,  15, -27,   8,   0,   0,  -4,   6, -19,  25, -22,   2,   4,   8, -15,  15, -22,  17, -24, -13,   9,   5,  -3,   6,  -3,   6,   8, -36,   6 },
     {   -8, -21,  47,  -4,  -5,  -4,   0,   0, -29,  -7,  34,  15, -14,  -3,  -1,   0, -57,  29,   4,  28, -15,  -6,  -2,   1, -24,  10,  11,  17,   4, -19,   3,  -2,  13, -27,  33,   2,  19, -14,  -7,   2,   8, -11,  -1,  25,   0,  14, -14,  -1,  -6,   5, -15,  12,  10,   1,   6,  -4,   2,  -2,   0,  -5,  10,  -2,   0,   2 },
     {   -5,  -7, -70,  27, -11,   7,  -1,   2,  -5, -43,  19, -46,  14,  -8,   3,  -1, -40,  21,   5,  -1, -33,   8,  -2,   1,  30, -13,   1,  13,  -5, -19,   2,  -3,   4,  18,  -5,  -5,  12,  -5,  -6,   0,   3,   5,  13,  -1,  -7,   4,  -4,  -2,   2,   2,   2,  10,  -1,  -4,   3,  -3,  -1,  -1,  -1,   1,   4,  -5,  -2,   1 },
     {    7,  11,  19,  20,  16,   5,  -8,  -1, -11, -22, -22, -31, -19,  -3,   6,   0,  11,  21,  27,  24,  22,   1, -12,  -2, -10, -26, -25, -26, -13,   3,  11,   0,  17,  22,  27,  20,  15,  -4, -19,  -4, -13, -17, -18, -13,  -3,  22,  16,  -1,   8,   7,  11,   0,  -4, -20, -26,  -5,   0,   7,  -1,  14,  10,  38,  13, -15 },
     {    5,  19,  41,  19, -15,  -7,  -2,  -2,   5, -14, -44, -16,  35,  12,   2,   1, -38, -31,   8, -10, -25,  -2,   4,   2,  39,  38,   3,   1,  12,  -7, -10,  -6, -26, -14,  15,  13,   8,  18,   3,  -3,   6, -10, -23,  -6,   4,  -2,  12,   8,   7,   3,   7, -14, -11,  -6,  -8,   2,   5,  -4,   2,  10,  -2,   2,  -5,  -4 },
     {    4, -11,   1, -11,   0, -12,   9,  -6,   7, -13,  11, -22,  10, -17,  17,  -9,   8, -19,  -2, -16,  21, -13,  13, -12,   2, -29,  23,   1,  21,   6,  12,  -6,   7, -24,  26, -20,   8, -20,   1,  11,  29, -37,  19,  -9,  12,   2, -18,   8,  41, -24,  17, -26,   5,  -7,  -5,  10,  14,  -2,  18, -11, -11,  14, -26,   1 },
     {   -2,  -1, -26,   5,  12,  31,   2,  -8,   0, -20, -26,  -4,  17,   1,  11,  22, -17, -29,   7,   3,  25,  -2,  -1,   4, -60,  -4,   3,  21, -28, -20,   5,   3,  -4, -28,   8,  18, -24,   9,  10,  -9,  14, -20,   7, -10,  -5, -10,   4,   1,  -5,  25, -15, -15,   2,  12,  -9,   4,   6,  14, -21,  -6,  12, -16,   3,   5 },
     {   -5,  -3,   6, -29,  25,  -2,   7,   4,  -8,   0,  30, -21,  -1,  16,  -2,  -7, -18,  15,  13, -30, -22,   5,  -1,  -2, -42,  18, -10,   0, -11,  18,   2,  -4, -34,  23,  21,  -3, -10,   2,  30,   1, -22,  -3,   9, -25,  -5, -26,   2,   7,   8, -17,  32,   0, -22,   1, -12,  -7,  14,  -5,   8,  24, -11,  -5,   9,  -1 },
     {   -5,  -4, -29, -45, -16,   8,  -6,  -3,  -8, -23, -18,   5, -25, -17,  18,  -7,  -8,   6,  39,  41,  28, -17, -13,  -3,  25,  33,  11,  -1,  17,  44,  -1,  -7, -16,  -4,   3,   0,   0,   1,  28,  12,   0,   0,  -8,  -2,  -5, -14, -11,  12,  -4,  12,  -9,   7,   2,   3,  -3,  -9,  -3,  -5,  -1,  -2,  -2,   4,   0,   5 },
     {    0,   3,   5, -58,  -3,  30,   7,  -2,  -1,   5, -27, -31, -32,   6,  12,   9,   1, -19, -31,  17, -26, -42,  10,   9,   3, -22,   2,  12,  16, -38, -12,  -3,   9,   7,   2,   7,   5,  10, -15,  -2, -22,  16,   3,  -7,  -2,   6,  -8, -11,   3,  -4,  11,  -4,   0,   0,  12,  -7,  13,  -3,   4,   7,  -6,  -3,   9,   2 },
     {    4,   0,   1,  12,  25, -15,  -1,   4,   4,   1,  -5,   4,  15, -11,   9,  -5,  17,  -2,  10,  21, -18, -14,  20,  12,   4, -15,  12, -18,  -8, -12, -19,   3,  -2, -34,   3, -18,  -6, -32,   2, -12,  10,  22, -11, -31,   5, -17, -37,   3, -22,  23,  26,  37, -17,   1, -28, -26,   7, -12,   2,   7,  18,  -9,  -7,  -3 },
     {    4,   6,  25,  27, -13,  12,  -3,  10,  -1,  16,   0, -24,  -7,   5,  20, -27,  15,  12, -37,  22,   3,  19,  -3,   2,  24, -13,   2,  21, -16,   1,  30,   1,  -3,  10,   8,  10,  -9,   9,  33,  16,  -1, -38,   3, -13, -27, -22, -22,  12, -26,   6,   2,   0,  11,  12,   7, -17,  15, -10,  -3, -16, -11, -13,  10,   9 },
     {   -1,   0,  -8,  -2,  -7,   9,  21,  -2,  -3,  -3,  10,  24, -14, -19,  -1,  10,  14,  21,   1,  -4,  -4,  -4,  -3,  12,  26, -28, -30,  -9, -12, -20,   1,   8, -86, -25,  27,   2,  -8,   0, -16,   0,   1, -11,  -1,   2,  -3,   7,  -4,  -1,  -3,   1, -12, -16,   0,   4,   8,   6,  -8,  15,   6,   9,  -9,  -1,   2,  -6 },
     {    2,   5,  20,  28,  15,  17,   5,  -1,  -4,   5,  -3, -46, -33,  -4,  -8,   4,  -4,  -1,  -9,  10,  15,   0, -13,  -5,  -3,  25,  12,   3,  -5,  19,   3,  -1, -29, -25, -34,  -3, -13,  -5, -12,   8,  33,  37,  40,   5,   0,   7,  -5,  -8,  -3, -23,  -2,   5,  -4,  -6,   7,   2,  21,  15,  13,  10,  17,  11,   8, -12 },
     {   -1,   0,   7,   3,   2,   0,   8,  -7,   0,   6,  -5,  18,  33, -33,  24,   6,   6, -12,   3,  37,  -9,   3, -15, -12, -10,  -1,  43, -12,  10,  -9,  -7,  14, -28,  41,   0,   4,  -7,   2,  -6,   0, -16, -26,  34,  11,  -6,  11,  20, -17,  -5, -10,  17,  36,  -6,  -5,   7,  27,  13,  -2,   2,   2,  10,  -1,   1,   6 },
     {    2,  -2,  -9,  47, -34,  42,   1,   2,   5,  -7,   8,  34, -31,  26,  -8, -20,   5, -15,  14, -10, -17, -42,   1,   1, -15,  -9,  10,  -9,   4,  21, -27,   1,   4,   8,  17,   8,   8,  23,   3, -10,  21, -11,   7,  -6,  -4,  -2,  -1,   4,   5,  -6,  21,  19,  12,  -3,  10,  -9,   5,  -1,   7,  11,  -8,  -5,  -4, -13 },
     {   -1,   1, -13, -10,  32, -30,  -7,  21,   1,  -3, -10,  15, -11,  41,   6, -16,   3, -11,  16,  15, -21,  11,  -5, -12, -13,  -3,  17, -13,  11, -27,  28, -22, -17,  26,   2,  13,  -5,  22, -15,  27,  48,   8,  -3,  -7,  -9,  10, -24,  -2,  -9,  -4, -13,  -5,   7,  -2,   4,  -6, -12,  12,  -6,   8, -11,  12, -18,  10 },
     {    3,   3,   2,  14,  48, -14, -28,  -8,   0,   2,  11,  -3, -37,  11, -14,   4,   6,   6,  -2,   5,   5, -22,  28,  -4,  34,   9,  12,  25,  10, -14, -18,  10,   0,   9,  13, -16, -18,  13,  13,  -8,  -9, -23,  28,   5,   3,  -5,  28,  11,  19,  18, -16,   5,  -7,   5, -25,  10, -10,  17, -24,   7,   2, -13, -18,  -1 },
     {    0,  -2,  12,  -7, -17, -11,  11, -18,  -1,   4,  19, -14,  30,   9, -28,  10,   7,  19,  23,   0,  24, -30,  19, -19,   8,  21, -17, -27,  12, -25,   1,   3,   7,   6, -13,  31,   0,  18,  -7, -18,  -6, -10,  38, -29, -21, -19, -29,  -4,  -4,   7,  -6,  -5,  17,  -5,  18, -11,  -4,  11,   5,  -5,  21,   1,  -6,   8 },
     {    0,   0,   4,  16, -44,   3,  19, -21,   1,  -5,  -3,   6, -36,   1,  23,   1,  -2, -13,  18,  -5,   0,  31,  51, -11, -10,  12,  19, -15,  -1, -22,  24,  10,   3,   7,  -8, -16,   1, -13, -10,  19, -31,   8,   4,  -7,  -4, -12, -18,   5,   3, -12, -16,  -8, -24,   5, -17,   8, -11,   4, -21,  16,   0,  -4,   2, -19 },
   },
   {
     { -110,  35,  10,   3,   3,   1,   1,   0,  33, -30,   2,   0,   0,   0,   0,   0,  28,   0,  -9,   0,  -1,   0,  -1,   0,  -3,   5,   1,  -1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   12, -25,   6,   0,   1,   0,   1,   0, 100,  -2, -26,  -2,  -5,  -1,  -2,   0, -41,  49,   3,  -4,   0,  -1,   0,   0, -21,  -6,  14,   1,   2,   0,   1,   0,  -4,  -3,   1,   1,   0,   0,   0,   0,  -4,  -2,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0 },
     {  -37,  14,   1,   3,   0,   1,   0,   0,  -9,  46,  -8,  -5,  -1,  -2,   0,   0, -73, -17,  42,   1,   6,   0,   2,   0,  38, -54, -12,  10,   0,   3,   0,   1,  11,  12, -15,  -4,  -1,   0,   0,   0,   4,  -1,   1,  -1,  -1,   0,   0,   0,   1,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {   36,  98, -42,  -9,  -9,  -3,  -3,  -1,  27, -26,  30,  -6,   1,  -2,   0,   0, -25, -18,   5,  10,   1,   2,   0,   0, -12,   3,  -4,   1,   2,   0,   0,   0,  -1,  -5,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    6, -30,   9,   2,   2,   1,   1,   0,  39, -16,   9,  -4,  -1,  -1,   0,   0,  -1, -60,   3,  25,  -1,   5,   0,   1,  25,  26, -65,  -1,   3,   0,   0,   0, -17,  45,  12, -27,  -1,  -3,   0,  -1,  -1,  -7,  17,   5,  -5,   0,   0,   0,  -3,   2,  -2,   3,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   -3,  13, -21,   8,  -3,   2,  -1,   1,  24,  85, -18, -25,  -3,  -5,  -1,  -2,  52, -24,  36,   0,  -6,  -2,  -1,   0, -38,  11,  -3,  12,   4,   0,   0,   0, -14, -11,  12,  -3,   4,   1,   1,   0,   0,  -9,  -1,   4,  -1,   1,   0,   0,   1,  -2,   1,   0,   1,   0,   0,   0,  -1,  -2,   0,   2,   0,   1,   0,   0 },
     {   16,  -2,  14,  -7,  -1,   0,   0,  -1,   6, -34,  -2,   7,   4,   0,   1,   0,  38,  -2,  35, -12,  -9,   0,  -2,   0, -17, -47,  -5,  55,   0,   2,   0,   1, -11,  12, -63, -13,  22,   3,   2,   0,   2,  12,  13, -26, -11,   3,   0,   0,   1,   3,   0,   8,  -5,  -4,   0,   0,  -1,   4,   2,  -1,   3,  -1,  -1,   0 },
     {   -9, -30,   9,   4,   4,   0,   1,   0,  -2,  -7,  32,  -8,   0,   0,   0,   0, -25, -33,   0,  35,  -2,   3,  -1,   1, -65,  -5, -16,  -1,  17,   0,   3,   0,  48, -47, -11,  -8,  -4,   6,  -1,   1,   7,  25, -28,  -3,  -6,  -1,   0,   0,   2,  10,   6,  -8,  -1,  -4,   0,  -1,   2,   2,   1,   1,   0,  -1,  -1,   1 },
     {  -14, -34, -81,  51,   1,   7,   0,   2,  -6, -25,  27, -43,  14,  -1,   2,   0,  -1,  27,  27,  -8, -18,   0,  -1,   0,  17,   9,  -3,  15,  -4,  -4,   0,   0,   3,   2,   1,   1,   5,  -1,   0,   0,  -2,   2,  -4,   1,   2,   1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0 },
     {  -14,  13,  -6,   1,  -1,   2,  -1,   0, -35,   5, -19,   7,   5,   1,   1,   0, -26,  43,  -7,  -1,  -5,   1,  -1,   0, -47,  25, -30,   7,  16,  -5,   3,  -1,   1,  15,   9, -58,  13,   2,   2,   0,  15, -18,  44,  -7, -26,   3,  -1,   0,   8,  -8,   1,  18,  -4,  -6,   0,   0,   2,  -4,   1,   2,   3,  -1,  -2,   0 },
     {    0,   1, -18,   2,   3,   0,  -1,   1,  10,  24,   0,  -6,  -7,   1,  -1,   0,   8,  -1, -28,   6,   7,  -2,   1,  -1,   7,   4,   4, -24,  -1,   3,   0,   0,  34,  26, -32,   2,  -9,  -2,   0,   0, -29,  53,  34, -35,   4,  -7,   0,  -1,   3, -34,  48,  18, -13,   3,  -2,   0,  -2,  -6, -15,  18,   5,   0,   0,  -1 },
     {    1,   0,   8,   4,   2,  -5,   1,   0,   5,  25,  57, -25,  -9,   1,  -2,   0,  18,  35, -45,  45,  -4,  -4,  -1,  -1,  10, -52,  -5,  -8,  27,   4,   0,   1, -41,   4,   0, -16,  -5,  10,   2,   0,   8, -10,  -1,   1,  -8,  -3,   2,   0,   4,   4,  -3,   1,  -1,  -2,  -1,   0,  -1,   1,   1,   0,   1,   0,   0,  -1 },
     {    7,   4,   1,   0,  -2,  -1,   0,   0,  -2,  -8,   4,  -2,   0,   0,   1,   0,  24,  15,  11,  -1,   1,  -3,  -1,   0, -23, -39,  -8,  -9,   2,   7,   1,   1,  61,  44,  16,  -5, -29,  -8,  -4,  -1, -37, -12,  16,  48,  18,  -6,  -1,  -1,   1, -10, -31, -12,  18,   9,   2,   1,  -4,   4,   7, -10,  -7,   1,   1,   1 },
     {   -5,   3,  44, -22,   2,  -6,   1,  -1,  -7,  14,  54, -22, -11,  -2,  -1,   0, -16,  26,  26,   9, -22,  -2,  -2,   1,  -8,  44,  -4,  33, -17,  -9,  -2,  -1,   2,  18,  -3,  34,  18, -18,  -1,  -1, -11, -10,   5,  -6,  34,   2,  -3,   1,   3, -22,   5,  -3,   1,  13,  -2,   2,   0,  -5,  -7,   4,  -2,   3,   0,   1 },
     {  -10, -13, -18,  13,   9,  -2,   1,   0,  -9,  -8,   2,   0,  -2,   6,   1,   1, -19, -30, -11,  -5,  16,  -1,   2,   0, -56, -14,  28, -19, -18,  12,  -1,   2, -38,  49, -13,  26, -16, -16,   4,  -2,  38, -21,   7, -10,  27,  -8,  -4,   0,  18,  -4,  -1,  -1,  -7,  13,  -3,   0,  -2,   6,  -3,   2,  -1,  -2,   3,  -1 },
     {   10,  15,  39,  71, -35,   0,  -7,  -2,   7,  12,  13,  -8,  35, -14,   3,  -1,  -5,  -5, -36, -46,   6,  22,   1,   2,  -9, -13, -26,  10, -29,   0,   6,   1,   2,  -7,   8,   3,  11, -12,  -3,   1,  -1,   9,   0,  -1,   2,   5,  -2,  -1,   1,  -1,   1,  -3,  -1,  -1,   1,   0,  -3,   0,   1,   2,  -1,  -1,  -1,   0 },
     {   -2,  11, -10,  -6,   1,   1,  -1,   0,  -9,   5, -24,   1,   0,   0,   1,   0,  -1,  19, -12,   4,  -2,   1,   0,   0,  -8,   7, -29,   3,   7,  -2,  -1,   0, -12,  30, -13,  -2,   4,  -2,   0,  -1, -15,  19, -38, -16,  22,  -3,   3,  -1,   3,  31,  27, -68,   5,   2,   0,   0,  11, -18,  51,   0, -23,   5,  -2,   0 },
     {  -12, -19, -40, -34,   2,   9,   3,   2,  14,  31,  43,  43,  -4, -12,  -4,  -3,  -6, -12, -38, -30,  11,  14,   5,   2,  -2,   6,  10,  22, -22, -11,  -1,  -2,  11,   4, -14,  -7,  26,   1,  -1,  -2,  -9,  -4,  10,  -1, -13,   5,   3,   1,  -5,   7, -27,  -6,  19,  -5,   2,   0,   3,   3,  17, -28,  -8,   5,  -2,   2 },
     {   -2,  -5, -11, -17,  10,   4,  -2,   0,   2,   0,  31,  44, -16, -10,   2,  -3,  13,  28,  36, -30,  22,   3,  -4,   0,  -6, -17, -53, -22, -14,  15,   6,   1, -18, -20,  21,   9, -23, -13,   2,   0,  24,  29,   2,   6,   7,  -8,  -5,  -1,  -1,   4,  20,   7,   1,   6,   0,  -2,  -8, -14, -10,  18,   3,   0,   1,   0 },
     {    6,   6,   8,  -2,   1,  -2,   0,  -1,   9,   7,   8,  -9,   1,   1,  -1,  -1,  11,   4,  -4,  -2, -11,  -4,   4,   0,   9,   3,  23,   6,  -6,  -9,  -2,  -1,  42,  32,   5, -10,   1,  -3,  -3,  -2,  77,   4,  -4,   8,  -6,   0,  -2,   0, -57,  32,  21,   7,   6,  -1,   0,  -2, -15, -27,   9,  12,   5,  -1,   1,   2 },
     {   -3,  -1,   6,   3, -16,  -5,   3,   1,   0,   3,   5,   4,  18,   4,  -3,  -1, -11,  -9,  -1, -16, -21,  -5,   3,   0,   4,  18,  17,  21,  28,   2,  -6,  -2, -28, -12, -18, -18, -41,  17,  10,   2,  15,  37,  20,  32,   5, -44,  -4,  -5, -10, -25,  -3, -14,  39,  26,  -7,   1,   8,  -2,  -1,  -4, -30,   2,   8,   0 },
     {    1,  -2,   7, -56,  44,  -8,   6,   0,   0,  -3,   0, -69,  48,  -4,   5,  -1,  -3,  -4,  -9, -36,   4,  13,  -2,   2,  -2,  -5,  -7,  -8, -22,  10,   0,   5,  -3,  -7,   4,  -2, -19,   8,  -1,  -1,   2,   8,   9,   4, -19,   0,   0,  -2,  15,   9,   1,   1,  -8,  -8,   0,   2,  -1,   1,  12,   2,  -5,   0,  -1,  -2 },
     {    0,   4, -24, -13,  15,  -6,  10,  -2,   0,   5, -39,   6,   8,   5,  -2,   1,   4,   8, -38,  29, -36,  18,  -7,   2,   6, -15, -41,  33, -23,   1,   2,   1,  13, -22,   0,  23,  -7, -22,   4,   0,  24,  -5,   7,  -6,  32, -12,  -1,   1, -15,  -4, -12,   2,  -2,  15,  -4,   1,   4,  30, -27,  -3,  -7,  -5,   2,  -1 },
     {    6,  10,   6,  34,  70, -32,  -2,  -1,   2,   3,   3,   2, -45,  37,  -2,   4,  -4, -10, -12, -37, -17, -10,   6,   0,   1,  -1, -10,  17,  28, -11, -10,  -1,   6,  -2,  17,  13,  -2,  11,   0,  -4,  -1,  12,   6, -12,   2,  -5,   2,  -2,   8,   1, -15,   8,  -3,  -3,   4,   0, -13,  -6,  19, -14,   5,  -4,  -3,   3 },
     {    5,   7,   6,  16,   2,  -1,  -4,   0,   7,  11,   9,  -7,  -5,   8,  -3,   1,   8,  16,   5,  -6,  34, -12,  -1,  -1,  15,  34, -23, -24,   6,  30,  -3,   1,  21, -20, -67,   8, -22,   5,  11,   4,  17, -42,  15,  -6,   0,  -8,  -2,   3,  -8,  16, -12, -12,  -8,   5,   1,  -2,  -5,  31,  14, -15,  -3,  -4,   2,   0 },
     {   -6,   6,  -2,   2,  -6,  15,  -6,   5, -11,   7,  -9,  -5,  -6,   8,  -2,   2, -13,   9,  -9,  10, -12,  -5,  -2,   3, -19,  16, -15,   5, -13,  10,  -4,   2, -27,  20, -27,  32, -22,   0,   2,   3, -28,  34, -14,  35, -23,   4,  -4,   0, -36,  28, -21,  34,   2, -28,   9,  -3,  -4,  -6,   0, -12,  35, -20,  -1,  -3 },
     {   -2,  -3,  -5,  -4, -14, -10,   7,   6,   3,   6,  11,  20,  -1,   7,  -6,  -5,  -3,  -5,  -9, -11, -21,  16,   0,   2,   4,   6,  11,  29,   1,   0,  -6,  -3,   9,   2,  -2, -10, -31,   8,  12,  -1, -24, -32,  -7,  13,   0, -24,   1,   4,  37,  47,  25,  11, -25,   2, -11,  -3, -28,   6,   1,  55,  12, -15,   7,  -5 },
     {   -2,  -5,  -5, -17, -43,  13,   5,  -5,  -2,  -5,  -8, -31, -25,   3,   5,   1,   0,   1, -16,  -6,  59, -38,   4,  -2,   4,   3,  -6,  52,   9,   8,  -6,  -5,   9,   1,  30,   2, -18,   5,   5,  -5,  14,  22,  -2, -19,  15,  -3,  -1,   2,  17,   3, -17,  10,  -5,   6,   2,  -1,  -2,   8,  10,  -2,  14,  -8,  -4,   0 },
     {   -2,  -6,  -2,   3,   7,  -4,   4,   1,  -4,  -9,   2,  -9, -14,   5,   2,   2,  -4, -11,  -8,  -7,  20,   3,  -6,  -2,  -8, -11,   3,  17,   0,  17,   0,  -2,  -4, -20,   8,   5,  -6,  -1,  11,  -4, -32, -35,  28,  -5,   6,  -6,  -4,   2, -51,  15,  18,   3,  -5,   9,  -3,   1,  69, -47,  -5,  14,  -7,   1,   6,  -4 },
     {   -3,  -5,  -5, -12, -20,   3,   6,   3,   1,   1,   4,   0,   4,  15,  -3,  -2,  -3,  -6,  -2, -11, -22, -22,  12,  -3,  -1,  -5,  -4, -19,  13,  -3,  -3,   4,  -5, -10,  -2,   4,  17,  -5, -18,   2,  -1,  12,  21,   5,   3,  37, -18,  -5, -14,   2, -29, -10, -43,  36,  29,  -4,  -7,  -2,  15,  31, -27, -51,  24,   7 },
     {    0,   0,  -3,  -1,   8,  12,   5,  12,   2,   3,   6,  21,  27,  17, -12,   5,   0,   3,  -4,  -8,  12, -19,  13,  -3,  -6,  -1, -14,  24,   5, -11,  -8,  15,   3,  13,   3,  16, -22,  52, -36,   9,   7, -25, -19,   8,  -5,  16,  22,  -8, -13, -39,  29, -23, -15, -24,  11,  10,   8,   5, -27,   0,   2, -20,  -5,  -3 },
     {    4,   4,   2,  11,  24,   3,  17,  15,   8,   9,   8,  26,  42,  -4,   2,   2,   1,  -1,  -1,  12, -12, -52, -25,   0,   9,   8,  12,   9,  -9,  46,   0,   5,   6,  -2,   4, -26,  -2, -19,  14,  16,   2,   6, -10, -11,  27,   5, -13,  13,  19,  -3, -17,   6,  -6, -30,  -4,  -6,   9, -32,   4,  -8,   3, -17,   7, -10 },
   },
   {
     {  109, -36, -26,   2,  -3,   0,  -1,   0, -26,  36,  -1,  -9,   0,  -1,   0,   0, -14,  -1,  14,   1,  -1,   0,   0,   0,  -3,  -1,   0,   3,   1,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   12,  83, -33, -20,  -2,  -4,  -1,  -2, -39,  -1,  62,  -9,  -8,  -1,   0,   0,  11, -36,   0,  27,  -1,  -1,   0,   1,   1,   1, -12,   0,   7,   0,   0,   0,   2,  -5,   0,   1,   0,   1,   0,   0,   1,   0,  -3,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {   38,  -7,  46, -23,  -5,  -2,  -2,  -1, -14, -59,  19,  56,  -8,   1,  -1,   1,  -4,  10, -58,   8,  24,   0,   1,   0,  -3,  13,   1, -25,   0,   5,   1,   0,   0,   2,  -1,   0,  -4,  -1,   1,   0,  -1,   3,   1,  -3,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 },
     {   31,  45, -29, -23,   2,  -2,   0,  -1,  91, -23,  -7,  -3, -12,   1,  -2,   1, -36,  33,   6,   2,  -1,  -4,   0,  -1, -13,  -7,  16,   2,   3,   0,   0,   0,  -6,  -3,   0,   5,   1,   1,   0,   0,  -4,  -1,   1,   0,   1,   1,   0,   0,  -2,  -1,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   10,  51, -13,   5, -13,   1,  -3,   0, -31,  -2, -45,  25,  35,  -8,   3,  -1,   4,   3,   3, -69,  19,  17,  -1,   1,   8,  -9,  31,  -4, -38,   9,   3,   0,   0,  -1,  -1,  13,  -3,  -9,   2,   0,   1,  -2,   4,   0,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0 },
     {    4,  24,  77, -45, -18,   0,  -3,  -1,   6,  60,  -9,  22, -18, -11,  -1,  -2, -18, -11,  28,   9,  11,  -5,  -2,  -1,   8, -27,  -2,  18,   3,   5,  -1,   0,  -3,   0, -12,   3,   7,   0,   2,   1,   1,  -5,  -1,  -1,   1,   1,   0,   0,  -1,   0,  -3,   1,   2,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 },
     {   21,  10,  36, -13, -16,  -3,   3,  -2,  -7, -42,  -4, -22,  20,  13,   0,   4,  23,   5,  13, -11, -66,  10,   9,   1,  -6,  12,  -1,  51,  -6, -43,   0,   3,  -1,   3,  -6,  -1,  29,   1, -15,  -2,  -2,   1,   1,   2,   0,   5,   1,  -3,   0,   0,  -2,  -1,   3,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   0 },
     {   16,  10,   7, -12, -11,   2,  -1,   0,  38,  33, -26,  -8,   3,  -8,   0,  -2,  84, -26, -30,   4,   5,   3,  -4,   0, -29,  41,   0, -17,   1,   3,   1,  -2, -13,  -3,  25,  -3,  -6,   1,   2,   0,  -5,  -3,   2,   7,  -1,   0,   0,   1,  -2,  -1,   2,   1,   1,   0,   0,   0,  -1,  -1,   0,   1,   0,   0,   1,   0 },
     {   10,  -2,  11,  46, -26,  -2,  -5,  -1,  37, -16,  18,  17,  42, -26,  -6,  -3, -20, -52,  20,   1,  18,  36, -10,   1,  -5,   5, -48,  10, -11,  10,  13,  -2,  -3,   8,  -5, -19,   8,  -4,   3,   3,   1,   2,  -1,  -4,  -8,   3,   0,   0,  -1,   1,   1,  -2,   0,  -1,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0 },
     {    5,  36,   3,  47, -21,  -9,  -4,  -1, -21,   3, -46,  17,   9,  -1,   2,   1,  -7,  30,  -5,  25,  -2, -25,   1,  -4, -12,  -2, -14,  -2,  61,  -7,  -6,  -1,   5, -10,   9, -42,   7,  24,  -3,   1,   1,  -3,   6,  -1, -12,   5,   2,   0,   1,  -1,   0,  -3,  -2,   1,   1,   0,   0,   0,   1,   0,  -1,   0,   1,  -1 },
     {   -2,  -8, -27, -51,  31,  14,   1,   4,  -9, -33, -57,  21, -16,  12,  11,   0,  -4, -49,  30,   3,   3, -14,   1,   5,   8,  11, -34,   6,   8,   4,  -9,   0,  -4,  22,   1, -23,  -3,   4,   3,  -5,   4,   1,  12,  -3, -10,  -2,   1,   0,  -1,   4,   0,   2,  -2,  -4,  -1,   0,   1,   0,   3,   0,  -1,  -1,  -1,   0 },
     {  -13,   2, -12, -17,  -4,   4,   8,   0,   2,  37, -12,   9,  13,   9,  -7,   2, -53, -22, -47,   4, -24,  25,  27,  -1,  15,  19,  -5, -29,  -9, -53,  10,   8,   5,  -5,  27,  -4,  14,  -2, -27,   2,   2,   0,   0,   3,  -7,   7,   0,  -5,   0,   0,   3,   1,   2,  -2,   0,  -1,   1,  -1,   0,   0,  -1,   0,   0,  -1 },
     {  -15, -18, -42,  -7, -23,  20,  -1,   3,   0,  12,  17,  55, -26, -39,   3,  -1,  18,  16,   4,  -5,  13,  -7, -10,   0, -11,  13,   9,  33,  -2, -23,   2,  -4,   0,  -3, -10,  -1,  56,  -3, -14,   1,   0,  -5,   4, -19,  -2,  25,  -1,  -3,   1,  -1,  -1,   1,  -1,  -1,   4,   1,   1,  -1,   0,  -3,  -1,   2,   0,   0 },
     {    3,  -7,   1,  11, -13, -12,   4,   3,  15,  -4,  14,   6,   8,  -8, -12,   0,   2, -32,   4, -21,   7, -29,  17,   9,  17,  -5,  20,  -8,  13, -15, -65,  18,  -5,   9,  -4,  25,  -4,  54, -19, -39,  -1,   3,  -3,   1,  -2,   2,  31,  -8,  -1,   0,   0,   0,   1,   1,   0,   6,   0,   1,  -1,   0,   0,   1,   2,  -1 },
     {  -10,  -8,   5,  -7,  -1,   1,   2,   1, -15, -10,  -2,  17,   8,  -7,  -3,   1, -26, -18,  41,  10,  -6,   2,  -1,   1, -79,  30,  32, -10,   4,   0,   0,  -1,  21, -41,  18,  20, -14,   7,   1,   1,   9,   0, -26,  15,   8,  -8,   1,   2,   3,   1,  -4,  -8,   7,   2,  -3,  -2,   2,   1,  -1,   0,  -1,   1,   1,  -1 },
     {   -8, -18, -20, -50,   0,   0,  13,  -1,   1,   7,  18,  15,  52,  -3, -24,  -5,  24,  28,   3,  -2,   1,  37,  15,  -8, -13, -37,  -6,   0,  23,   2,   8,  12,   5,  -8, -15, -28, -19,  30,   3,   0,  -1,   3,   4,   6, -26, -16,  12,   1,   1,   0,   2,   2,   2,  -5,  -4,   1,  -1,   0,   0,   1,  -3,  -1,   2,   0 },
     {    6,   9,   5,  30,  76, -45, -19,  -2,  11,  20,  10,  45, -11,  37, -19, -14,  15,  -8,   6, -12, -11,  10,  26,  -2, -17,  -4,   2,   9,  -3,  -7,   4,   7,  -2,  -4, -13,  -2,  13,   0,   3,   1,   0,  -2,   2, -13,  -1,   6,  -2,   4,   0,   0,  -2,   1,  -4,   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   1,   1 },
     {    7,   8,   9,  13,  16,  15,   5,  -7,  -7, -14, -14, -13, -24, -20,  -6,  12,  20,  14,  25,  19,  28,  16,  13, -16, -18, -28, -19, -27, -19, -23,  17,  46,   5,   7,   0,   6,  -8, -12, -58, -15,  -3,   3,   0,   7,   9,  16,  13, -26,   1,   1,   2,   0,  -2,  -4,  -1,   5,   0,   0,   0,   1,   2,   3,   0,  -2 },
     {   -2,  -6,  -6,   2, -11,  -3,   5,   7,  -1,  -3, -20, -30,  -4,  17,   9,   2,  -7, -41, -51,  10,  25,   0,   6,   6, -52, -53,  22,  31,   1,  -2,  -3,   3,  13,   2, -34,  -2,  14,  -6,   2,   0,   5,  14,   4, -21,  -2,   4,  -5,   0,   3,   2,   1,   3,  -5,  -1,  -2,  -1,  -3,   3,   0,  -2,   0,  -2,  -1,   0 },
     {   -3,   2,   9,  -4,  -7,  10,  13,  -5,  -1,   9,  27, -18, -13,  27,  36, -21, -11,  20,  -1, -58,  30,  -8,  25,  28, -32,  22, -44,  10,   7,   0, -16,  23,  -6, -13,   7, -18,  -2,  -7,  -9,  -1,  13, -11,  12,   9, -13,  -2, -10,  -1,  -1,   1,   0,   2,   2,   0,   3,  -5,   0,  -3,   4,  -1,  -1,   0,  -1,   0 },
     {    1,   5,  13, -17,  59, -29,   7,  -3,  -9,   6,  -2, -29,  27, -48,  17,   9, -16,  15, -19,   1,  10,  10, -47,  17,  -8,  15,  -7,  10,  -5,  10, -21, -15,  14,  -5,   6,  -6,  21,  14, -14, -15,   0,  -6,   7,  -1,  -9,  25,   8,  -8,  -1,   0,   2,   4,  -7,  -1,   8,   2,   0,   0,   0,   0,  -2,   1,   1,   0 },
     {    6,   5,   7,  11,  31,  76, -30, -19,   9,   7,   8,  15,  40,  -7,  46, -10,   7,  -1,  -3,  21, -22, -15,  15,  35,   1, -18,  12, -12,  -2,   1, -13,   5,  -1,  -6, -10,   3,   1,  -6,   2,   2,  -6,   3,  -4,  -5,   0,   1,  -5,  -5,   3,   0,   1,  -3,  -4,   1,   2,  -3,   0,   3,   0,  -2,   0,   3,  -1,   0 },
     {   -6,  10,  -6,   8,  -7,  27, -14,   1, -10,  15, -11,   9, -16,  18, -26,  11, -12,  25, -18,  18, -16,  26,  -8,  11, -23,  20, -29,  20, -32,  20, -16,  -6, -10,  10, -16,   8, -10,   5,  18, -54,   4,   2,   8,  -8,   9, -22,  41,  -8,   6,  -3,   6, -10,  10, -10,   5,   9,  -1,  -1,   3,  -3,   0,  -2,   3,  -2 },
     {   -6,  -2,   3,   3,   8,   3,   2,   9, -11,  -7,   8,  -2,   8,  -4,  -7,   1, -10,  -2,  -2,   8,   0,  -8,  -6,  -5, -21, -23,  20,   4, -10,  -5,   2,   4, -97,  21,  36, -15,   5,   0,   6,  -3,  25, -37,  -5,  10,  -9,   6,  -1,   1,   4,   4,  -9,  -1,   0,   0,  -1,   2,   4,  -4,   2,  -1,   1,   4,  -2,  -1 },
     {    4,   0,   8,  10,   5,  28, -37,  -8,   2,  -6,   9, -11, -45,   7,  13,  -1,  -2, -16,  -5, -30,  -4,  31, -21, -16,   5, -11,  19,   0,  14,   0,  20, -25,   8,  -9,  12, -13,   0,  34,  -4,  12,  -3,   1,  -7,  20, -44,   6,  33,  -9,  -3,   6,  -7,   4,  10, -31,   4,  14,   1,   1,   0,   2,  -1,   0,  -8,   2 },
     {    1,   0,  -4,   7,   6,  -7,  -1,   6,   4,   6,  12,  22, -18, -43,   3,  13,   1,  -5, -17, -27, -37, -15,   0,   9, -21, -36, -11,  13,   6,  -7,  -3, -15,  28,  40,  33, -11, -32, -21,  -5,  -9,   6,   9,  16,  30,  14, -26,  -8,   3,  -2,  -4,   2,   0,  20,   8, -11,  -7,   1,   1,  -3,   0,   0,   5,   2,  -8 },
     {    0,  -4, -12,  -3,   4, -13, -12, -14,   0,  -7, -19,   8,  -6, -10,  10, -11,   6,  -4, -20,  -4,   0,  20, -10,   3,  19, -25, -36,  22,  29,   0,  -9,   0, -32, -56,   5,  46,   0, -12, -14,   7,   3,   8, -25,  14,  14, -18,   2,   1,  10,  10,  -8, -34,  15,  16,   3,   2,  -4,  -2,   7,  -1,  -6,  10,   1,  -1 },
     {    2,   6,   7,  12,  21,  11,  33,   1,   1,   6,   3,   2, -11, -26,  27,  33,  -6,  -8,  -3, -12,  -5,   7,  -2, -28,   7,  28,  18,   7,  13, -13,  10,  11, -36, -20, -42, -16, -23,   7,  -2,   1,  18,  30,   9, -23,   3, -35,  -3,   5, -14, -10,  23,   9,  10,   4, -20,   3,   4,  -3,  -2,   1,  -1,   1,  -2,  -7 },
     {   -5,  -5,  -9,  -7, -17, -20, -15,  -3,   4,  10,   7,  18,  12,  28,   6,  18,  -2,  -5,  -1,  -2, -20, -11, -34,  10,  -4,   7,  10,  -1,   8,  21,   2,   9, -14,   6, -24,  -8, -24, -11, -38,  17,  12,  12,  11,  11,   5,  12, -11, -67,  -4,  -2,   2,   2,   5,   1,  45,   4,   2,   1,  -3,  -6,   1,  -4,  -3,   7 },
     {   -2,  -2,  -2,  -1,   6,  -4,  -9, -49,  -1,  -1,  -3,  -3,   5, -14, -17, -22,  -6,   0, -11,  16,  26, -23,  25,  -4,  17,  17,  16,  61, -26,  -1,   6,  14,   4,  -9,  31, -26, -35,   7,   3,  12,  -6,  21, -12,   0,   5,   5,   3, -17,  -4, -15,  -2,  -2,  15, -13,   0,  14,  -2,   4,   6,  -8,  -5,  -5,   1,   3 },
     {   -2,  -6,  -5,  -2, -13, -36, -33,  38,   0,  -5,  -4,   1, -10, -12,  35, -24,   2,   0,   4,  19,   3,   9,  -8,  23,  10,  -1,  12,  -6, -23, -34,  -1,  25,   0, -23, -19, -25, -16,  -6,  32, -19, -26, -21,  10,  21,  -4, -11,  -3, -19,  -2,  10,   6,   5,  12,   0, -19,  17,   9,  -8,  -8,  -6,   1,  10,  -1,  -4 },
     {    2,   1,  -4,  15,  -8, -13,  79, -52,   2,   0,  -5,  22, -20,  16,  22,  -4,   9,  -3,   6,  15, -13,  18, -22,  38,   2, -18,   5,  -9, -17,  -5,  -5,  -5,   1,   2,   2,  -1,   1,   6,  13,   7, -14, -14, -12,   9, -14,  -5,  11, -11,   4, -14,  -6,   5,  -1,  -4,  -6,   3,  -2,   3,   0,   2,   9,  -4,   0,   2 },
   }
 },
 {  //33
   {
     {  112, -12, -21,  -4,  -5,  -1,  -2,   0, -44,  28,  10,  -1,   0,  -1,   0,   0, -12, -11,   7,   3,   1,   1,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,  -3,  -1,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   16, -54,   8,   7,   2,   2,   1,   1,  79,  40, -36,  -9,  -5,  -2,  -2,  -1, -43,  27,  27,  -4,   0,  -1,   0,  -1,  -2, -18,   2,   8,   1,   1,   0,   0,  -6,   5,  -2,  -1,   2,   0,   0,   0,  -1,  -4,   1,   0,   0,   1,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {    7,  91, -28, -13,  -5,  -4,  -2,  -1,  45,   8,  30,  -7,  -6,  -3,  -1,  -1, -49, -22,  19,  16,   5,   2,   1,   0,  10, -16, -10,   3,   2,   1,   0,   0,  -5,   1,  -1,   0,   1,   0,   0,   0,   2,  -3,  -1,   0,   1,   1,   0,   0,  -2,  -1,   0,   0,   0,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0 },
     {  -44, -24,  -7,  16,   3,   3,   1,   1, -32,  70,  13, -26,   0,  -4,   0,  -1,   0, -58,  34,  20,  -2,   2,   0,   1,  15,  -5, -36,   4,   6,   1,   0,   0,   2,   7,   6, -11,  -2,   0,   0,   0,   4,  -3,  -2,   4,  -2,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   1,  -1,  -1,   1,   0,   0,   0,   0 },
     {    6, -37,   4,  -1,   6,   1,   1,   1,  26, -27,  22,  14,  -7,   1,  -1,   0, -38, -40, -54,  26,  16,   5,   3,   2,  21,  40, -27, -45,  -1,   0,   0,   0,  -4,   9,  29,  -1, -13,  -2,  -1,  -1,   2,   2,  -4,   5,   2,  -3,   0,   0,  -2,   2,   3,  -1,   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {   -2,  14,  63, -10,  -8,  -3,  -1,  -1, -39, -17, -45,  28,  14,   3,   2,   1, -55, -12,  15, -23,   3,   3,   2,   0,  47, -16, -12,   8,  -9,  -1,  -1,   0,  -6,  16,  -4,  -6,   3,  -1,   0,   0,   8,  -4,   4,   1,  -3,   0,   0,   0,  -1,   3,  -2,   0,   1,  -1,   0,   0,   2,  -1,   1,   0,   0,   0,   0,   0 },
     {   25,   9,  54,   0,  -8,  -5,  -2,  -2,  21, -22,  -2,   0,   4,   2,   0,   1,  34, -13,  11,  35,  -7,  -1,  -3,   0, -46, -19, -54,   9,  28,   7,   4,   1,  10,  17,  -5, -40, -13,   3,   2,   1,  -4,   3,  12,  12,  -7,  -6,  -1,   0,   1,   2,  -4,  -2,   4,   0,  -1,   0,  -1,   0,   2,   1,   0,   0,   0,   0 },
     {   -9, -30, -54,  14,  11,   2,   2,   1, -20, -62,  20, -21,   1,   7,   2,   2, -28,  34,  26,  13, -15,  -5,  -1,  -1,  24, -28, -27,  15,  15,   1,   0,   0,   2,  23, -10, -23,  -4,   2,   0,   0,   3,   2,  10,   4,  -5,  -3,  -1,  -1,   1,   3,  -1,   1,   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 },
     {   14,  18, -18,  30, -11,  -2,  -1,   0,  16, -25, -54, -39,  24,   6,   3,   1,  21, -19,  27, -10, -32,  -1,  -1,   1,  19,  59,  -9,   7,   2, -12,  -3,  -2, -16,  -6,  34,  -1,   3,   4,  -1,   0,   0,  -6,  -8,  12,   1,   1,   1,   0,  -4,   0,   2,  -6,   1,   0,   0,   0,   0,  -1,   0,   2,  -1,   0,   0,   0 },
     {  -10,  27, -25,  18,   0,  -1,  -1,   0, -25,  30, -47,   2,   1,   4,   2,   0, -17,  41, -24,   5,  15, -10,   2,  -2, -19, -12,  -6, -52,  23,   9,   3,   1,  19,   0,  35, -27, -33,   1,   0,   1,   2,  -5,  13,  21,  -2,  -9,  -2,  -1,   3,  -3,   1,  -2,   4,   2,  -1,   0,   1,  -2,   2,   1,  -1,   0,   0,   0 },
     {    0,   5,  52,  13,   5, -11,  -1,  -2, -18,   4,  25, -40, -13,  12,  -1,   2, -23,  39,   4,  50, -17, -19,  -2,  -2,  14,  15,  19,  -3,  34,   1,  -4,  -2, -12, -40,   5,  18,  -3,  15,   4,   1,   1,  -2, -27,  -3,   9,  -2,   3,   1,   0,  -2,   3,  -8,  -5,   2,   0,   0,   1,   0,  -3,   2,   0,  -2,   0,   0 },
     {   -5,  -7,  -3,  22,  -1,  -1,   0,   0, -12, -22, -19, -23,  16,   7,   2,   1, -49, -42,  -3,   1, -11,   6,   3,   2, -60,  -1,  26,   1,  -3,  -6,   1,   0,  56, -20, -25,  11,   1,   0,  -3,   0,  -5,  26,  -5, -13,   2,   0,   0,  -1,   7,  -2,   7,  -1,  -4,   0,   0,   0,   0,   3,  -1,   1,   0,  -1,   0,   0 },
     {    6,   1,  12,   2,   3,  -4,  -1,  -1,  -1, -13,   3, -22, -10,   5,   2,   1,  13,   0,  22,   3,   3,  -6,  -3,   0,   0, -26,  -1, -19, -32,  13,   2,   2,  20,  38,  21,  53, -25, -39,  -5,  -4,  -9,  -4,  -7,  22,  59,   7, -10,  -2,  -1,  -3,  -6, -17,  -5,  20,   7,   0,  -2,   0,  -1,   2,  -2,  -5,   3,   1 },
     {    9,  11,  20,  85,  -9, -13,  -5,  -3,   9,  13,  35, -20,  33,   4,  -3,  -1,   1,   7, -44, -32,  -7,  19,   9,   2,  12, -28,  -3,   2, -14,  -9,   3,   3, -11,  14,  -5,  -6,   5,  -3,  -4,   0,   0,  -3,   7,   0,  -5,   1,   0,  -1,   0,   1,  -2,   2,   1,  -1,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0 },
     {   -5,   7, -10,  23,  -4,  -4,  -1,   0,  -9,   8, -18,  30,  -9,  -3,  -1,   0,  -9,  16, -17,  35,   4,  -9,   0,  -1, -15,  18, -25,  22,   9, -18,   5,  -2,   2,  17,   2,   1,  61, -23,  -9,  -1,   3, -16,   9, -35,  33,  39,  -9,   0,   1,  -7,  -1, -12, -23,  16,  11,   0,   0,  -1,   0,  -1,  -2,  -6,   2,   1 },
     {   10,   4, -14,  34,  15,  -8,  -4,   0,  11, -11, -38,   7, -13,  12,   2,   0,  29, -34,  -6,  24,  32, -25,  -1,  -2,  37, -28,  20, -16,  26,  28,  -8,   1, -14,   2, -43,   9, -10,  12,  11,   2,  -5,  15,  -5, -31,   5,  -4,   0,   2,   1,   0,   9,  -3, -12,   3,  -1,   0,  -2,   2,  -1,   1,   0,  -2,   0,  -1 },
     {   -3,  -6,  -5,  41, -17,  -5,  -3,   0,  -1, -12,  21,  57,  28, -16, -10,  -7,  -3,  12,  57,  -4,  35,  16,  -8,  -4, -16,  16, -14, -16,  -2,  24,  13,   2,  -6, -36,   4,   8, -11,  -1,   8,   4,   6,   0, -29,  -1,   6,  -2,   1,   2,   0,   5,   2, -12,  -1,   1,   0,   1,   2,  -1,  -1,   1,  -3,   0,   0,   1 },
     {   -4,  -4,  -8, -10,   7,   4,   1,   4,  -7, -11, -14, -11,  -5,   9,   3,   1,  -4, -15, -20,  -8,  -1,   1,   4,   2, -38, -40,  -9,  12,  -4,   2,   1,   2, -62, -15,  24,   6,   1,  -6,  -2,  -3,  63,   2, -38,  -2,   4,   1,  -2,   0,  -4,  25,  14, -16,  -6,   0,   1,  -2,   7,  -3,   8,   7,  -3,  -1,  -1,   1 },
     {   -6,  -9,  -8, -27, -84,   7,  18,   2,  -1,  -1, -10, -23,  49, -25, -17,   3,   6,   6, -13,  43,   9,  20,  -1,  -8,  13, -15,  17, -11,  -2,   1,  12,   3, -11,   8, -12,   4,   1,  -5,  -4,   2,   4,   3,   2,  -5,   2,   0,   0,  -2,  -2,  -1,   5,   0,  -2,   1,  -1,   1,   0,   1,  -1,   1,  -1,  -2,   0,   0 },
     {   -5, -10, -14, -12, -25, -27,  -2,   3,   9,  12,  14,  44,  30,  31,  -5,  -4,  -1, -13, -12, -12, -41, -26,  13,   7,   5, -11,  14,  19,  39,  -7, -27, -10,  16,  13,  12,   2, -12,  22,  18,   0,  -5,  -7, -21,  13,  24, -10,  -3,   6,  -3,   5,  -2, -24,   0,  17,   3,  -1,   1,  -5,   2,   3,  -8,  -6,   4,   1 },
     {    3,   1,   2,  -3,   3,  11,   1,   1,  -1,  -8,  -3,  -8,   3,  -9,  -1,  -2,   7,   2,   5,   1,  13,   5,   6,   2,  -4, -23,  -6, -15,  -2, -10, -12,  -2,  24,  26,  30,  17,  23,  33,   1,  -6, -36, -44, -46, -24, -29,   7,  27,  13,  21,  26,  10,  -5, -17, -29, -15,  -2,  -6,  -2,   7,  12,   6,   4,  -2,  -2 },
     {   -3,  -2,   4,  18,  10,   1,  -3,  -4,  -7,  -7,   5,   6,  -1,  -4,  -5,   0, -11, -17,  14,  25,   7,  -3,  -3,  -4, -26,  17,  65,  -1,   0,  11,  -1,  -2, -23,  46,  -6, -17,   3,   4,   6,   1,  20, -57,  10,  26, -10,  -3,   0,   0,  -2,   8, -35,   3,  13,   1,   2,   0,   0,   3,   4, -14,   0,   1,   0,   1 },
     {    0,   1,  -1, -28,  66, -50,  12, -12,   0,   6,  -4, -18,  67, -24, -11,   3,  -1,   8,  -9,   4,  14,  17, -20,   4,  -2,   7,  -8,   6,   0,  23, -12,  -3,   0,   4,  -2,   1,  -2,  11,   9,  -6,   3,  -4,   5, -11,   8,  10,   0,   4,  -6,  -5,  -1,  -2,  -9,  12,   2,   0,  -1,  -2,  -1,  -1,  -4,  -4,   3,  -2 },
     {    3,   3,  -5,  -6,  42,   5,   0,  -1,   3,   6, -10,  28,  16,  29, -21,   4,  10,  -2,   7,  31, -45,  40,  24, -18,  12, -15,  11, -23,  -8, -27,  44,  14,  -5,  -8, -12,  -5,  -4, -11, -29,  15, -14,  -4, -15,   7,  -1,   0,   2, -14,   5,   2,   6,  -8,   0,  -7,   3,   4,   4,   5,   0,  -3,  -1,   1,  -2,   2 },
     {   -5,  10,  -7,  11,   1,  16,  -6,   3,  -9,  12, -16,  10, -14,   5, -18,   4, -10,  17, -17,  22, -16,  31,  -5,  -1, -14,  22, -22,  28, -27,  14, -10,  -8, -17,  27, -23,  24, -20,  16,  34, -23, -11,   8,  -4,  -6,  11, -46,  31,   9,   3, -13,  12, -17,  17,  -8, -14,   4,   6,  -1,   2,  -4,  -1,   4,   2,  -2 },
     {    3,   0,  -3,   9,   7,  14,   7,  -6,   1,  -4,  -8,  -1, -22, -42,  -2,  15,  -1, -10, -27,  -5,  14,  23, -34, -19,  20,  -9,  15,  51,  24,   9,  32,  -7,  24,  -9,  17, -16,  -6,  -2,  -1,  22, -18,   0, -29,  25,   8,   1,  -6,  -7,  -5,   6,  -5, -31,  14,   5,   3,   2,   3,  -9,   7,   1, -14,   3,   1,  -2 },
     {    6,   8,   8,   6,   9,  13,   0,  -1,   7,   9,   8,   6,   5,   1,  -5,   1,  11,  15,   9,  11,   1,  -6,   0,   0,  15,  17,  16,   4,  -8,  -4,   3,  -2,  47,  41,   5, -12,  -2,  -1,  -3,   4,  65,  25, -25, -14, -10,  -5,   0,  -2, -38,   7,  42,  11,   4,  -1,  -5,  -3,  -7, -20,  -4,  12,  10,   7,   2,  -2 },
     {    4,   5,   6,   6,  24,  88,   5,  -3,   3,   2,   7,  13,  37, -44,  20,  13,  -3,  -2,  -2,   0, -36, -30,  -4,   3,   0,  -7,  -1, -17,   9,  10, -15,   1, -10,  -4,  -3,   5,   1,   2,   4,  -9,  -2,   5,   6,  -3,  14,   2,  -2,  -2,   8,  -3,  -4,  -1,  -2,   7,   3,   0,  -8,   1,   3,   2,  -3,  -6,   1,   0 },
     {   -1,   1,   1, -12,  -4,  13, -15,   8,   1,   6,   4, -10,  14,  35,  20,  -9,  -5,   4,  -8,  -5, -14,  -1, -16,   6,   0,  13,  -1,   4,  -7,  44,  35, -24,   2,   4,  -3, -13,  -4, -38,  45,  35, -12, -23,  -7, -26, -12, -10, -35,   7,  14,  12,  11,   7, -21,  -9,   0,  -9,   0,   2,   3,   9,   4,  -4,  -6,   4 },
     {    1,   0,   2,   4,  10,  -8,  19,   0,  -1,  -5,  -2,   0,  -2, -24,   1,   0,   2,  -2,   5,   2,   7,   2,  23,   3,  -1,  -5,   1,  -1,   9, -38,   4,   8,  -2,  -6,   6,   5,   8, -14,  16,  12,  -3, -13,   4, -15,   8, -46, -41,  29,  10,   7,   9,  15,  24,  34, -50, -28,  -5,   3,  -4,   1,  -7,  25,  26, -10 },
     {    0,   3,   3,  -9,  -4,   4,  20,  -1,   0,   4,   7, -19,  -9,  -5,  -4,  -3,  -3,   6,  -1, -28,   9, -12,  17, -13,  -1,  29, -11, -11,  -8, -13,  -1,  26,   3,  -1, -53, -35, -28,   2, -15,  -5,   1, -36, -10, -11,  23,  -8,  -1,   0,  -7,  19,   9, -45, -25,   7,   1,  -2,  -3,  -6,  27,  16, -12, -16,  -1,  -5 },
     {   -5,   0,   7,  -3,   0,  -6,  -8,   6,  -7,   0,  13,  -2,  -7, -11,  -5,  -6, -10,  -4,  16, -24,  -2,   7,  -4,  -9, -19,  12,  23,  -7,  40,  -1,  21,   7, -38,  46,  20,   0,  -6,  -4, -15,   5, -23,  35,  10, -20,  -7,  -5,   9,  -7,   9, -27,  27,  -4, -25,   4, -11,  12,  -4,  -1,  -4,  34, -27, -11,  10,  -3 },
   },
   {
     {  112, -44,  -8,  -3,  -2,  -1,  -1,   0,   1,  17,  -9,   1,  -2,   0,  -1,   0, -29,  20,   1,  -1,   0,   0,   0,   0, -10,   2,   5,  -1,   0,   0,   0,   0,  -5,  -1,   2,   1,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {  -14,  -9,   5,   0,   1,   0,   0,   0, 106, -31, -20,   0,  -4,   0,  -2,   0,  -7,  43, -15,  -2,  -2,  -1,  -1,   0, -29,  15,  12,  -4,   1,  -1,   1,   0,  -6,  -3,   6,   1,   0,   0,   0,   0,  -5,  -1,   1,   1,   1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0 },
     {  -16,  41, -16,   0,  -3,   0,  -1,   0,  12,  56, -17,  -9,  -3,  -3,  -1,  -1, -66,  24,  39, -12,   4,  -3,   2,  -1,   5, -48,  28,   6,   0,   1,   0,   0,  13, -19,  -6,   9,  -1,   1,   0,   0,   4,  -4,  -5,   2,   1,   0,   0,   0,   2,  -1,  -2,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 },
     {   38,  80, -55,  -1,  -9,   0,  -3,   0,  23,  14,   6, -14,   1,  -3,   0,  -1,  43, -28,  10,  -2,  -2,   0,  -1,   0, -25,  25,   0,   1,  -1,  -1,   0,   0, -14,  -5,  14,  -1,   2,  -1,   1,   0,  -2,  -7,   3,   3,   0,   1,   0,   0,  -2,  -4,   1,   2,   0,   0,   0,   0,  -1,  -1,   0,   1,   0,   0,   0,   0 },
     {   11,  -3,  -4,   3,  -1,   1,   0,   0,  47, -18,  10,  -5,   0,  -1,   0,   0, -29, -60,  33,  15,   1,   3,   0,   1,  36, -25, -47,  22,  -1,   3,  -1,   1, -10,  41, -30, -10,   5,  -2,   1,   0,  -6,  11,   6, -12,   1,   0,   0,   0,  -3,   5,   4,  -2,  -2,   0,   0,   0,  -1,   1,   2,   0,   0,   0,   0,   0 },
     {   15,  30,  -1,  -7,  -2,  -2,  -1,   0, -16, -72,  53,   3,   4,   1,   1,   1, -40,   0,  -5,  16,  -2,   3,   0,   1, -38, -11,   7,   8,   3,   0,   1,   0,  29, -33, -19,  10,   1,   2,  -1,   1,   5,  17, -22,  -3,   1,   0,   0,   0,   3,   6,  -2,  -6,   0,   0,   0,   0,   2,   5,  -1,  -2,  -1,   0,   0,   0 },
     {   -8,  25,  -9,  -1,  -1,   0,   0,   0, -12,  -5,   2,   6,   0,   1,   0,   0, -70,   5, -30,  16,   6,   2,   1,   0,  14,  46, -33, -31,   8,  -4,   1,  -1,  -7,  18,  44, -31,  -3,  -1,   0,  -1,  11, -24,  25,   8,  -8,   1,  -1,   0,   3,  -7,   0,   9,  -1,   0,   0,   0,   3,  -4,  -1,   2,   1,   0,   0,   0 },
     {   16,  34, -31,   5,  -3,   0,  -1,   0, -12, -38, -32,  25,   4,   4,   0,   1,  11,  38, -35, -19,   7,  -3,   1,  -1,  53, -27,  14, -12,  -2,   0,  -1,   0,  13,  26, -25,  -3,  -2,   0,   0,   0, -20,  26,  20, -18,   1,  -2,   1,  -1,  -2,  -9,  16,   1,  -4,   1,   0,   0,  -2,  -1,   4,   3,  -1,   0,   0,   0 },
     {   19,  24,  47, -36,   1,  -6,   1,  -1,  16,  -5,  44, -12,  -7,  -1,  -1,   0,   6,  10,   3,  -2,  -7,   0,  -1,   0,  67,   5,  26,  -7, -16,  -1,  -3,   0, -18,   2,  23,  33, -11,   0,  -1,   0, -13, -13, -13,  28,   7,  -2,   1,   0,  -8,  -1, -15,   5,   9,   0,   0,   0,  -4,   0,  -6,  -2,   4,   1,   0,   0 },
     {    6,  30,  51, -36,   3,  -7,   1,  -1,  -7,  27,  23,  -7,  -5,  -1,  -1,   0,  10,  40,  -7,  16,  -9,   0,  -2,   0, -34,  -8, -19,  26,   9,   1,   1,   0, -13,  30, -31, -24,  16,  -2,   2,  -1,  -9,   1,  35, -26,  -2,   2,   0,   0,   7, -20,  12,   9,  -9,   3,  -1,   0,  -1,  -4,  -1,   7,  -1,  -1,   0,   0 },
     {   10,  11,  19, -16,   1,  -3,   1,  -1,  23,  39, -11,   9, -11,   0,  -3,   0,  13, -23, -54,  46,   2,   2,  -1,   0,  15, -22, -20, -20,  22,  -3,   3,  -1,  41, -38,   8, -15,  -3,   6,  -1,   1,  -2,  26, -22,  -1,  -6,   1,   0,   0, -11,  17,  11, -10,   0,  -3,   1,  -1,  -2,  -2,   7,   0,  -1,   0,   0,   0 },
     {   -7, -19, -29,  21,  -3,   2,   0,   0,   7,  18,  46, -44,   5,  -5,   1,  -1,   8,  41,  16,  -8, -14,   1,  -2,   0,  27,  33, -30,  13,  -7,  -3,  -1,  -1,  59,  -9,   0,  -6,   1,  -2,  -1,   0, -17,  21,   8,  -2,  -1,   0,  -1,   0, -15,  -2,  15,   3,   0,   0,   0,   0,  -3,  -8,   0,   6,   1,   0,   0,   0 },
     {   -9,   1,  -9,   6,   1,   0,   0,   0,  -2,  16,   6,   0,  -2,  -1,   0,  -1, -18,   7, -18,   7,   1,   1,   0,   0,  20,  40,  -1,   7,  -7,  -3,  -1,  -1, -58, -24, -23,  31,  22,  -1,   4,   0,   9,  16, -36, -48,  18,   2,   2,   0,   3,   6,  40, -25, -14,   3,  -1,   0,   6, -10,   6,   9,  -7,   0,   0,   0 },
     {   -6,  -7,   6, -17,   7,  -1,   2,   0,   3,  18,   7, -16,  -1,   0,   0,   0, -19, -40,  -9, -12,  10,   3,   3,   1, -29,  20,  38, -42, -14,   3,   0,   1,  21,  40,  -8,  25, -29,  -2,  -2,   0, -24,  44,   9,  -2,   9, -11,   2,  -2,   4, -27,  31,   5,   1,   1,  -2,   1,   0,  -6,  -8,  12,  -1,   2,  -1,   0 },
     {   10,  22,  42, -26,  -1,  -3,   0,  -1,  -4, -10, -50,  33,  -3,   3,   0,   0,  -5, -11,  40, -34,   5,   0,   0,   0,   4,  54, -19,  14, -14,  -1,  -1,   0,  42, -21, -10,   6,   9,  -5,   0,   0,   0,   3,   1,  -7,   3,   2,  -2,   0, -16,   9,   5,  -3,  -1,   1,   0,   0,   2,  -9,   6,   2,  -2,   0,   0,   0 },
     {    7,   9,  24,  69, -27,  -2,  -7,   0,  11,  17,  36,  64, -12, -11,  -4,  -2,   8,   6,  22,  12,  31, -14,   1,  -2,   2,  14,  12, -18,  24,   3,  -3,   0,   9,  12,  -9,   0,   1,   9,  -1,   0,  -5,  -3,  -4,  -2,   6,   4,   3,  -1,   0, -14,  -2,  -2,   2,   5,   1,   1,  -3,  -4,  -7,  -3,   0,   3,   1,   0 },
     {    4, -11,   2,  -3,   2,   0,   0,   0,  13,  -8,  10,   1,  -3,   0,  -1,   0,   6, -20,   6,  -3,   2,   1,   0,   0,  21,  -5,  23,  -8,  -2,  -1,  -1,   0,   3, -45,   0,  -8,  10,   3,   1,   1,  47,  -4,  38, -30, -22,   6,  -4,   1,  10, -23,  22,  54, -24,   0,  -3,   0,  -5,   6, -41,  23,  12,  -3,   2,   0 },
     {    3,   5,  20,   0,   0,  -1,   0,   0,  -2,  -8,  13, -20,  19,  -4,   3,  -1,   4,   6,  27, -41,  23,  -4,   3,   0,  -3, -10,   1, -40,  24,   7,  -1,   1, -32,  -7,  14, -44,  -7,  16,  -3,   2,  -6,  34,  -8,  -6, -32,   9,  -2,   0, -20,  46,  12,  -2, -17,  -6,   1,  -1,  -3,  -1,  21,  14,  -9,  -4,  -1,   0 },
     {   -7,  -7,  -3,   4,  -1,   2,   0,   0,  -8,  -9,  -6,   9,   1,   0,   1,   0,  -9, -10,   4,   7,   5,   0,   0,   0,  -7, -12, -11,  14,   6,   1,   1,   0, -28, -45,  19,   3,   6,   4,   1,   1, -90,   3,  21,   5,  -1,   2,   1,   0, -33, -29,   7,  12,   4,   0,   1,   0,  16, -19, -22,   0,   2,   2,   0,   0 },
     {   -9, -13, -35, -46,  35,  -1,   6,  -1,  -6, -10,  -3,  26, -22,   5,  -3,   0,   9,  16,  54,  54, -19, -11,  -3,  -1,   9,  16,  23, -20,  35,  -7,  -2,  -1,  -2,  10,  -2, -14,   1,  10,  -1,   1,   0,   9,  -9,   1,  -4,   2,   1,   0, -10,  -6,   7,  -5,   3,   0,   0,   0,  -7,  -6,   2,   1,  -2,   3,   1,   0 },
     {   -6,   7,  -4,  11,  -7,   1,   1,   0,  -7,  13,  -8,  11,  -9,   2,  -1,   0, -10,   8, -10,  14, -10,   1,  -1,   1,  -6,  20,   2,  19, -15,  -4,   1,   0, -23,  10, -20,   8,  -2,   0,   2,   0,   6,  51,  -4,  17, -28,  -6,   0,  -1, -29,  35, -22,  52,  20, -19,   3,  -3, -22,   8, -20, -22,  41,  -4,   2,  -1 },
     {   11,  18,  36,  46, -16,  -6,  -3,  -1, -12, -20, -45, -46,  21,   7,   2,   1,   2,   6,  22,  40, -35,  -5,  -3,   0,   6,   4,   3,   2,  17,  -5,   0,  -1,  -7,   0,   9,  -2, -16,   6,  -1,   1,  10,  22, -11,  17,   5,  -9,   2,  -2,  13, -25,  19,  -6,   6,   1,  -2,   0,  -2,   2, -13,  18,   2,  -1,  -1,   0 },
     {    1,   1,   2,   7,   2,   0,  -2,   1,  -1,  -6,   6,   5,  -4,   0,  -1,   1,   3,   1,  -7,   0,  -2,   0,  -1,   0, -10, -21,   8,  15,  -8,  -1,   0,   1,   1,  20,  17,   3,  -2,  -7,  -2,  -1,  47,  -8,   5,  10,   7,  -4,  -1,   0, -78,   4,  35,  -6,  16,   5,   0,   1,  30, -57,  -3,   8,  -4,   6,  -1,   1 },
     {    1,   1,   8,  18,   7,  -7,  -1,   0,   0,   1,  -5, -19,  15,   2,  -1,   0, -11, -23, -16,  11, -15,   5,  -1,   0,   7,  18,  63,  22,   5, -23,  -3,  -1,  22,  23,   0, -24,  43,   7,  -7,   1, -25, -23,  -7, -24, -10,  29,  -2,   3, -29,  15, -18,  -3, -23,   1,   7,   0,   8,  13,   5,  -6,  -3,  -8,   1,  -1 },
     {    4,   6,   0,   9,  69, -29,   2,  -3,   4,   6,  -3,   7,  73, -27,  -3,  -1,   2,   6,  -4,  16,  30,   1,  -5,   1,   1,   6,  -9,  10,  -8,  20,  -8,   2,   2,  -2, -14,  16, -25,  22,   0,   0,   5, -14,  -2,   5, -13,   6,   7,  -3,  -9,  -8,  -2,   2,  -6,   1,   1,   0, -12,  -4,  -2,  -3,   5,  -3,   2,  -1 },
     {   -1,  -2,   0,  -6,   0,   2,   1,   0,   6,   7,   5,  -5,  12,  -2,   0,  -1,  -3,  -6,  -5, -15,   0,   5,   1,   0,  11,  13,   7,  -3,  11,   0,  -2,  -1, -15, -20, -21, -24,   4,   5,   2,   2,  29,  37,  16,   7,   0,  -6,  -3,   0, -26, -54, -33, -20,  28,  13,   1,   2,  30,  31,   7, -45, -32,   6,  -2,   1 },
     {    0,  -2,  -1,  -1,  10,  -7,   1,   0,   4,   4,   9,  33,  18, -10,  -2,  -2,   1,   0,   6,   4,  -1,   9,  -4,   1,   2,  -7,  13,  43, -26,  -2,   0,   0,   0,   8,  68, -17,  15, -14,  -1,  -1,   6,  60,   1,  -5,   6,   5,  -4,   1,  41,  -3, -10,  -4,  -7,  11,   0,   0,   5, -19,  10,   8,  -9,   1,   3,  -1 },
     {   -6,  -9,  -8, -33, -57,  32,  -1,   3,  -4,  -6, -10, -20,  16, -13,   8,   1,   1,   3,  -2,  28,  62, -36,  -2,  -1,   5,  11,  16,  37,   7,  19, -13,   4,   6,   5,  10,  -1, -17,  18,   2,  -2,   5,  -1,   2, -13,   1,  -7,   9,  -3,  -8,   0,  -5,  -3,  -1,   0,   1,   1, -14,   1,  -2,   2,  -6,   2,   0,   1 },
     {    2,  -3,   2,   3,  -1,   8,  -2,   1,   4,   0,   7,   1,   3,  -5,   3,   1,  -4, -14,  -8,  -5,  -9,   3,   3,  -3,  11,  13,  27,  20,   8, -10,  -1,  -1,  -2, -15, -24, -15,   0,  24,   4,   1,  -6, -15,  19,  18, -41, -17,   8,  -1,  44,   6,  26,  -1,  44, -21,  -4,   0,   2, -49,  39, -12,   1,  20,  -7,   4 },
     {   -1,   0,   6,  15,   5,  -4,  -3,   0,  -3,  -2,   4,   4,  -3,  -2,  -2,   0,  -5,  -8,   5,  12, -15,   6,   0,   1,  -3,  -5,  14,  -1, -15,   7,   2,  -1,  -6, -18,  15,   4,  -6, -12,   6,   0,   0,  -3,  53, -16,  22, -19,   0,  -1, -27,  16,  15, -22,  11,   8,  -5,   2, -79,  25,  34, -11, -11,   7,   0,  -1 },
     {    4,   5,   3,  20,  48, -27,  -5,   5,  -3,  -5,  -2, -25, -60,  33, -10,   6,  -2,  -5,  -9, -10,  35, -40,  11,   2,   4,   8,   4,  33,  21,  -5, -10,   6,  -1,  -5,   6,   0, -20,  10,  -6,  -1,   2,  10,  14,   3,   9, -13,  -4,  -1,  10,  10,  -6,   0, -11,  20, -11,  -5,  10,  -1,   5,  -9,   3,  -9,   1,  -6 },
     {    1,   3,   2,  11,   5,  25, -14,   5,   0,   4,   4,  16, -22,  22,  -8,   3,   1,   4,   4,  15, -12,  30, -14,  -1,   2,   6,  10,  15, -36,  39,   2,   1,  -1,  -2,   0, -18, -44,  10,  19,  -2,  -6,  -9,  -5,  -6, -26, -21,  20,  -1, -11,  -9,   1,  -5, -25, -23,   4,  -2,  31,  27,  25,  21,  -9, -29,  -2,  -1 },
   },
   {
     { -103,  64,  -1,   1,   0,   1,   0,   0,   6, -22,  17,  -4,   2,  -1,   1,   0,  22, -17,   0,   3,  -1,   1,   0,   0,   5,   0,  -3,   1,   0,   0,   0,   0,   5,  -2,  -1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -32, -60,  58,  -8,   8,  -1,   3,   0,  47, -54,  -7,  19,  -5,   3,  -1,   1,  -5,  29, -36,   8,   1,  -1,   0,   0,  -8,  15,  -4,  -7,   3,  -1,   0,   0,  -1,   6,  -1,  -3,   0,   0,   0,   0,  -1,   3,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {   20,  28, -64,  33, -12,   5,  -4,   1,  85, -29, -27,   4,   2,  -2,  -1,   0,  -2,  13,   1, -11,   5,  -2,   1,  -1, -18,  10,  10,  -7,   2,  -1,   1,   0,  -7,   0,   8,  -2,   0,   0,   0,   0,  -5,   0,   3,   0,   0,   0,   0,   0,  -2,  -1,   2,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
     {   47,  28,   3, -26,   2,  -6,   0,  -1,   6, -52,  74, -17,  -4,   2,  -1,   1,   7, -23, -11,  36, -16,   5,  -2,   1, -16,  23, -27,  12,   4,  -4,   2,   0,  -5,   3,  -3,  -2,   5,  -1,   0,   0,  -3,   3,  -3,  -1,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   0 },
     {    8,  -5, -25,  42, -20,   7,  -4,   1, -54, -31, -16,  55,  -8,   3,   0,   1,  36, -22, -46,  12,  16,  -8,   3,  -1,   5,  16, -12, -26,  17,  -4,   1,   0,   4,   2,   7, -16,   1,   2,  -1,   0,   1,   4,   4,  -5,  -1,   0,   0,   0,   1,   0,   2,  -3,   0,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0 },
     {   13,  61,  12, -31,   1,  -6,   0,  -2, -16,   1,  -5,  26, -12,   5,  -2,   1, -52,  59, -42,  11,   7,  -6,   4,  -2,  -2, -10,  25, -21,  11,  -2,   1,   0,  13, -21,  10,   1,  -3,   3,  -1,   1,   4,  -5,   2,   1,   0,   0,   0,   0,   2,  -3,  -1,   2,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {   23,  23,  13, -61,  37, -15,   5,  -2,  16,   9, -45,   0,  10,   3,  -2,   1,  70, -22, -27, -12,   3,   5,  -3,   1,   7,   5,  12, -15,  -9,   8,  -3,   1, -10,   8,   5,   4, -11,   3,   1,   0,  -6,   2,   5,   3,  -4,  -2,   1,   0,  -2,   0,   1,   3,  -1,  -1,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0 },
     {   -1, -12, -21,   3,  22,  -8,   2,   0, -39, -52,  16, -32,  49,  -9,   4,   1,  -9,  18,  -1, -54,  16,  11,  -5,   3,  -7,  13,  25, -22, -26,  18,  -4,   1,   8,  -4,   8,  12, -26,   4,   2,  -1,   1,   5,  -2,  11,  -8,  -4,   2,   0,   2,   0,  -1,   4,  -1,  -2,   0,   0,   0,   1,  -1,   2,   0,  -1,   0,   0 },
     {   25,  27,  49,   3, -38,   9,  -7,   0,   5, -35,   3,  41,  -9,  -3,   4,  -1,   1, -21,  38, -40,  29, -10,   2,   0,  14, -31,  21,  10, -13,  13,  -6,   2, -10,  14, -28,  18,  -2,   0,   3,  -1,  -4,   6,  -5,  -3,   5,  -2,   1,   0,  -2,   2,  -4,  -1,   1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0 },
     {   -5,  -7,  -2,  -4,  15,  -7,   1,   0,  28,  29,  27,   9,   6, -19,  -2,  -2, -39, -49, -12,  12,  61,  -2,  -4,   3,   9,  -4, -20, -49,   8,  31,  -7,   2,  -5,   6,   7,  -3, -29,  12,   4,  -2,   2,   0,  -3,   2,  -7,  -3,   4,   0,  -1,   1,   2,   1,  -3,  -2,   1,   1,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {    0, -11, -34,  -1,  49, -25,   8,  -1,   1, -34, -11,  17,   9,   6,  -8,   4, -21, -13,  -1,  30, -17,   9,  -2,  -1,  53, -62,   5,  15,   1,  -3,   0,   1,  11,   5, -26,   8,   5,  -1,  -2,   1,  -5,  13,  -8, -10,   7,   1,   0,  -1,  -2,   4,   0,  -5,  -1,   2,   0,   0,  -1,   2,   0,  -2,  -1,   0,   1,   0 },
     {    6,  26,  25,  15,  15, -27,   2,  -2,  -7,  -1, -36,  17,  59, -32,   4,  -1, -26,  -4,   2,  -1,   9,  16, -15,   5, -21,  18, -45,  37,  -3,   6,   2,  -4,  -4,   6,  -8, -27,  26,  -4,   2,  -1,   7, -10,  13, -14,  -5,  12,  -3,   1,   1,  -2,   2,  -3,  -5,   3,   1,   0,   1,  -1,   2,  -1,  -2,   1,   0,   0 },
     {   12,   6,  10,  17, -37,  20,  -8,   1,  16,  13,  12, -42,  30, -20,   8,  -3,  27,  31, -49,  -1,  17,  -6,  -2,   1,  58, -25, -27,  14,  -4,   8,  -5,   1,  18,  -5, -10,   0,   1,   2,  -1,  -1, -10,   9,  -1,  -3,   1,   1,   1,   0,  -5,   2,   4,  -3,  -1,   0,   0,   0,  -2,   0,   2,   1,  -1,   0,   0,   0 },
     {    4,  -2,   8,   2,  13, -21,   7,  -1,   0, -10,   7, -14,  14, -20,   8,   0,  21,  12,  30,  -1,  15, -48,   9,  -1,  -4, -13,  16,  -9,  69, -23, -21,  11,   2,  -9,   4, -24,  14,  39, -28,   6,  -3,   0,  -2,   3, -17,  22,  -2,  -5,   0,   1,  -3,   2,  -7,   6,   4,  -2,  -1,   1,  -2,   1,  -3,   0,   1,   0 },
     {   -3,  -9,  -8,  22,  19, -25,   2,  -1,   6,  28,  44,  35,   6, -31,  -7,  -1,  41,  27, -12,  16,  20,   3, -17,   2, -37,  -9,  36,  19, -13,  19,  -1,  -5,  -5, -19, -18,  29,  12,  -6,   8,  -1,  -5,  -2, -10, -16,  18,   4,  -1,   1,   0,  -3,  -1,  -5,  -1,   5,   0,   1,  -1,   0,  -2,  -2,  -1,   1,   0,   0 },
     {  -10,  -3, -32, -26,  17,   9,   3,   0, -12,   3,   7,  -4, -22,  10,   3,  -1,  -5,  31,  -7, -25,  30, -28,  12,  -4,   4,  20, -27,   8,   8,  -1,  -8,   7, -31,  55, -49,   8,  14,  -6,  -1,  -2,   0,  -2,  18, -30,  15,   2,  -1,   0,   6, -10,  11,  -4,  -8,   7,  -2,   1,   2,  -3,   4,   0,  -3,   1,   0,   0 },
     {    5,   9,   9,  13,  10,   5,  -8,  -4, -12, -16, -19, -19, -12,  18,  -2,   0,  26,  38,  36,  36,  17,   1, -11, -12, -34, -40, -37, -16,  10,  43,  19,   2,   9,  14,  -1, -13, -37, -12,  14,   1,  -2,  -3,   2,   8,   7, -10,   3,   3,   4,   2,   0,  -4,  -3,  -3,  -2,   1,  -1,   1,   1,   1,   1,  -1,   1,  -1 },
     {    6,  11,  12,  28,  19, -35,   1,  -5,   6,   9,  27,  24, -14,  -5, -15,   2,  13,  37,  29, -16, -19,  18,  -8,  -7,  64,  30, -13, -21,   0,   1,  12,  -7, -26,  19,  29, -11,  -6,  -3,   4,   2,  -5, -10,  10,  11,  -3,  -1,   1,   3,  -8,  -1,   0,   5,   3,   0,   1,   0,   0,  -4,  -3,   1,   3,  -1,   1,   1 },
     {   -3,   1,   6,   7, -18,  17,  -7,   3,  -7,  -1,   2, -15,  27, -23,   9,  -2,  -6,  -3, -15,  32, -24,  12,  -2,  -1,  -7, -28,  47, -20,   3,   1,   2,   2, -69,  52,   6, -15,   4,  -6,   3,   2,   7, -12,  17,   1,  -6,   0,  -1,  -1,   6,  -8,  -3,   5,  -1,   1,  -1,  -1,   3,  -1,  -2,   2,  -1,   0,   1,   0 },
     {   -2, -11, -13, -25,  -4,  24,  -7,   3,   4,  -4,  21,   5,  -9,  15, -35,   8,   9,   2,  -1, -30,  25,  43, -33,   2,  -5, -34,   2,  -3,  11,  -2,  22, -17,   9,  -9,  11, -40,  40,  -7,   1,   6,   2, -13,  21, -13, -25,  25,  -9,   2,   3,   3,  -4,  11, -13,   0,   8,  -3,  -1,   1,   0,   1,  -4,  -1,   2,   1 },
     {   -6, -12, -25, -44,  -7,  32,  -2,   6,  -7, -18, -11,  19, -15, -63,  38,  -5,   1,  12,  29,  24,   3, -19, -31,  11,  25,  17,  -1,  10, -11,  22,  -3, -11,   5, -10,  23,  -3,   5, -13,   5,   2,   2,   1,   0,   8,  -1,   9,  -8,   3,  -4,   1,  -4,   3,   3,  -1,   2,  -2,   1,  -1,  -2,   0,   2,   1,  -1,   0 },
     {    2,   0,   3,   1,  10,   5,  21, -13,   4,   2,   8,   7,   4, -17,  34, -17,   5,   1,   4,  -4,  16,  11,  42, -14,  -7, -14, -11, -20, -23, -32,  51,  29,   3,  -3,   4,   2,   5, -38, -38,  40,  -4,   4,  -3,   3,   4,  14, -26,  -2,  -2,   1,   2,   0,  -2,   2,  -5,  -4,  -1,   0,   0,   1,   1,   2,  -2,  -2 },
     {    1,   1,   5,  -3,  33,  47, -74,  14,   1,   1,   4,  15,  37,  12, -29, -20,   1,   1,  14,  16,  11, -19,  33, -30,  10,  17,   8,  10,  -1,  -7,  16,  -8,  -3, -11,   9,   7,   3,   5,   8,   7,  -6,  -2,   0,  -3,   2,   0,   3,   6,  -7,   1,   0,  -3,   0,   1,   2,   1,   0,  -2,  -1,   1,  -1,   1,   0,   2 },
     {    9,  12,  23,  41,  37,   0, -18, -10,  -4,  -8, -16, -46, -42,   9,  20,   5, -10, -20, -16,   5,  22, -18, -35,   1,  14,  21,  31,   8,   2,  11,  18, -16,  -7, -11,  -5,   4,  16, -20,  -1,  19,   6,  -1,   1,  -8,   0,  17, -19,  -3,  -1,  -1,   1,   3,  -9,   1,  10,  -4,   1,   1,   0,  -1,   0,  -1,  -1,   2 },
     {    0,   3,   2,  -9, -18,  11,  -2,  -7,   0,   5, -12,  -4,  17, -16,  -9,   6,  -2,   2,  14,   6, -11,  39, -30, -17,   4,  32,   8, -38,  45,  -7,  27, -16,  17,   8, -51,  23,   5,   4,  -1,  19,  -2,  30, -34,  -3,  10,  -4,   1,  -5,  -2,  -4,   5, -13,   4,   2,  -1,   3,   3,   0,   0,  -1,  -5,   3,   1,  -1 },
     {    2,   5,   8,   5,   1,   0, -17,  17,  -6,  -6, -12, -14,  -7, -17, -33,  19,  -1,   1,   8,  14,  14,   5, -18,  10,  -8,  -6, -11, -25, -23, -32, -17,   8,  16,  27,  29,  38,  20,  -8, -46, -37, -14,  -9,  -2,   1,  31,  26,  17, -12,  -1,   4,   6, -11,  -6,   1,   7,  -2,  -3,  -3,   2,   4,   0,   5,   1,   1 },
     {    5,  -3,  -4,   0,  -5,  -6,   0,   5,   8,  -4,  -1,  -2,   4,  -4, -10,   0,  10,   0,   1,   9,  -1,  -3,   8, -17,  13,   7,   2, -10,  -5,   4,  -4,   6,  27,   5, -30,  13,   2,  -7,  -5,  -3,  84, -69,  -1,  17,  -5,   8,  -7,   0,   9,   2, -20,  15,   5,   2,   2,  -4,  -1,   8,  -6,  -5,   7,   1,   6,  -1 },
     {   -3,  -3,  -6,  -8,  -5,   9,  12,  -6,  -4,  -7,  -6,   3,  11,  15, -21,   1,  -1,   2,  -8, -11,  14,  -3, -18,  21,   5, -11, -26,  25,  28,  -5,  13,  -5, -63, -30,  26,  33, -13,  -5, -14,  15,  22, -13, -38,  22,  17,  -4, -11, -12,  15,  -3, -10, -20,   7,  16,   3,  -4,   1,   2,   0,  -2,  -5,  -2,   3,   1 },
     {    1,  -4,   4,  -7, -13, -21,   6,  19,   1,  -7,   0, -13,  15,   2,   2,   2,   0, -10,   8,  -5,   0, -19,   8,  14,   4,  -7,   2, -27,  18, -10,   6, -19,  11,  -9,  19,   7,  19, -25,  44,  -2,  -8, -24,  11, -24,  44, -21, -33,  44,  -6,   3,   6, -12,  -8,  27, -29,   1,  -9,   3,  -1,   8,  -8,   8,  -4,  -6 },
     {   -4,  -6,  -6, -10, -31, -18,  41, -22,  -6,  -9, -12, -13,   6,   1, -36,  16,  -3,  -3,   8,  34,  45,  -6,   0, -40,  13,  20,  24,  33, -11, -13,  30,  -4,   1,   4,  12,   8, -12,  17,   2,   7, -20, -11,  24,  -4,  -5, -12,   7,   8,   0,  -5,   4,  10,   1,  -5,  -1,   3,  -2,  -4,  -5,   4,   4,  -1,   2,   1 },
     {   -4,  -6,  -4,  -5, -14,  -9, -22,  23,  -2,  -1,   1,   7,   4,  11,  16,  -3,  -3,  -3,   2,   4, -14,  12, -25,  13,   2,   5,  -1,  -1,   7,   2,   9,  29,  -9, -26, -27,  13, -36,  21, -22,  12, -28, -44,  46, -13,   3,   4, -20, -12, -26,  14,  21,  -6,   1,   4,  22, -15, -17,  -6,   5,  11,  -4,   6,  -8,   9 },
     {    3,   3,   5,   8,  13,  42,  32, -65,   2,   0,   3,   7,  17,  33,  24, -40,   2,   3,  10,  13,   5,  19,  -9,  18,   8,   9,   6,  -8,   5,   7, -19,   9,   9,   5,  10,  18,  21,  17,   3, -16, -12, -32,   1,  -9,   6,   7,  14,  11,  -9,   2,  -3, -11, -12,   1,  -3,   2,  -3,   5,   3,   4,  -5,  -3,  -1,   1 },
   }
 },
 {  //34
   {
     { -114,  32,  13,   2,   4,   1,   2,   0,   7, -35,  10,   3,   1,   1,   0,   0,  18,  -7, -15,   4,   0,   1,   0,   0,   2,   4,  -3,  -6,   1,   0,   0,   0,   5,  -1,   0,  -1,  -2,   0,   0,   0,   1,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -3,  63, -18,  -2,  -5,  -1,  -2,   0, -61,  47,  37, -16,   2,  -4,   1,  -1,   2, -43,  39,  12,  -6,   2,  -1,   0,   5, -12, -15,  21,   2,  -1,   0,   0,   1,  -4,  -5,  -3,   9,   0,   0,   0,   3,  -3,  -2,  -1,  -1,   3,   0,   0,   1,  -2,  -1,   0,   0,   0,   1,   0,   1,  -1,  -1,   0,   0,   0,   0,   0 },
     {  -39,  -8,  -1,   3,  -1,   2,   0,   1,  36,  60, -42, -17,  -1,  -5,   0,  -1, -11,  32,  54, -30,  -6,  -3,  -1,  -1,   5, -22,  26,  30, -13,  -2,  -1,   0,  -3,   0,  -8,  12,  14,  -4,  -1,   0,   0,  -6,   1,  -2,   4,   6,  -1,   0,  -1,   0,  -1,   0,  -1,   1,   2,   0,   0,  -2,   0,   0,   0,   0,   0,   1 },
     {    2,  38, -21,  -6,   1,  -1,   0,   0, -53,   6, -31,  10,   4,   5,   2,   1,  -8,  31, -28, -44,  17,  -3,   4,  -1,   9,  -6,  50, -42, -17,   8,  -2,   1,   7,  -8,  12,  20, -28,  -2,   2,   0,   1,   1,   2,   8,   3, -12,   1,   0,   2,  -3,   2,   2,   3,   0,  -4,   0,   0,   0,   1,   1,   0,   0,  -1,  -1 },
     {   30,  83, -25, -13,  -4,  -5,  -1,  -1,  74, -15,  20, -10,  -7,  -1,  -2,   0,   7,  22,  -3,   8,  -6,  -3,  -1,  -1, -13,   7,  18,   3,   4,  -3,  -1,   0,  -4,  -8,   3,  10,   3,   3,  -1,   0,  -6,  -1,  -2,   1,   4,   2,   1,   0,  -2,  -3,   0,   0,   0,   2,   1,   0,  -2,   0,   0,   0,   0,   0,   1,   0 },
     {  -18,  -8, -31,  15,   6,   1,   1,   0,  31,  52,   1, -16,  -2,   0,  -2,   0, -37,  -9, -11,  32,  17,  -2,   4,  -1,  -2,  -8, -23, -54,  29,  11,   1,   2,   4,  -6,  23, -22, -39,  15,   2,   1,   0,   1,  -2,  13, -12, -15,   5,   0,   0,   0,   5,  -1,   2,  -4,  -5,   1,   1,   0,   0,   2,   0,   0,  -1,  -2 },
     {    2, -17,  11, -13,  -2,   1,   3,   0,  17, -12,  32, -26, -11,   5,   2,   1,   6, -29,  22,  -5, -11,   5,   4,   1,  11, -18,   6,  -6, -39,  19,  -1,   3,  -8,  16, -16,  43, -53, -18,  16,  -2,  -2,   8,  -7,  16,  23, -47,   2,   4,   2,   0,   4,  -2,  16,   0, -18,   2,  -1,   2,  -2,   2,   2,   4,  -2,  -3 },
     {   -3,   2,  59, -23,   2,  -7,   1,  -1, -13,   5,  42,  20, -22,   4,  -5,   1, -62,  40,  -3,  34,   2,  -9,   2,  -2,   2, -32,  24,   2,  19,  -3,  -1,   0,   4, -13, -13,  15,   4,   8,  -2,   0,  -1,  -1, -10,  -3,   6,   2,   2,   0,   1,  -1,  -4,  -2,   0,   2,   2,   0,   1,  -1,  -2,  -1,   0,   0,   1,   0 },
     {    7,  12,  17,   3, -11,   2,   1,  -1, -12, -25, -22, -10,  20,   5,   1,   2,  17,  24,  25,  -6, -24,   2,   3,  -1, -11, -21,  -1,  19,  37,   5,  -1,   2,  -6, -13, -22, -39, -30,  41,  13,   1,  -3,   5,  -3,  11, -39, -40,  17,   4,   3,   1,   2,  -2,  16, -16, -17,   3,   0,   2,   2,   2,  -2,   4,  -3,  -4 },
     {   11,  37,  36,  -8,  -4,   2,  -2,   0,  24,  -6, -36,  47,  10,  -3,   2,  -1, -37, -42,   2, -24,  34,  10,   2,   2,  10, -27, -33,  -3, -18,  18,   2,   2,  -8,  21, -18, -19,  -2,  -8,   6,   0,   3,  -4,  13,  -9,  -8,  -1,  -1,   2,  -2,   4,  -1,   3,  -2,  -4,  -1,   0,   0,   0,   2,  -1,   0,  -1,  -1,  -1 },
     {   -7,  -8, -65,  13,  12,   2,   2,   1,  -8, -57,  -2, -24,  15,   7,   1,   2, -72,  -5,   9,  -4, -14,   9,   2,   0,   0, -29,  -2,  13,  -5, -10,   3,   0,   7,   0, -18,   0,  11,  -5,  -6,   1,   3,   8,  -2, -12,   1,   5,  -2,  -2,   3,   2,   3,  -2,  -6,   0,   2,   0,   2,   2,   1,   0,  -1,  -2,   0,   0 },
     {    6, -14,  15, -23,  22,  -2,  -2,   0,  13, -17,  27, -27,  23,  -2,  -1,  -1,  15, -27,  26, -30,  29,   5,  -6,   1,  16, -25,  30, -25,   5,  13,  -8,   2,  14, -17,  15,  -3,  -1,  30, -19,   2,  -4,   4, -12,  12, -31,  43,   5,  -6,  -5,   4, -10,   8, -19,  -7,  22,  -2,  -1,   0,  -2,   1,  -1,  -6,   1,   4 },
     {    2,  -7, -31,  42,  -1,  -6,   0,   0,   5,  -3,  -5,  42,  -6,  -9,  -4,   0,  35, -14,   3,  25,  33, -23,   0,  -4,  28, -36,  25,   3,  30,  11, -12,   1,  -7,   1, -38,  25,   3,  18,   2,  -1, -10,  10, -19, -20,  17,   0,   6,   1,  -1,  -2,   1, -14,  -2,   6,  -1,   1,  -1,   2,  -1,  -4,  -3,   1,   1,   0 },
     {    9,  15,  19,  21,  12,  -8,  -3,  -2, -15, -27, -30, -28, -11,   6,   4,   1,  20,  30,  36,  28,  13, -17,  -8,  -3, -20, -34, -29, -19,   9,  19,   7,   2,  19,  19,  18,  -4, -16, -33, -10,   0,   4,  -2,  10,  15,  33,  23, -13,  -3,   0,   0,  -5,  -7, -12,  13,  11,  -2,  -2,   0,   1,   3,   2,  -3,   3,   1 },
     {    3,   3,  -5, -59,  20,   1,   4,  -2,   2,  16, -24, -33,  -7,  16,  -4,   2,  24,   5, -44,  12, -18,   2,   6,  -2,  59, -34, -23,   3,   7, -15,   4,   1,  16,  11, -26,  -7,   6,  -2,  -8,   3,  -3,  15,   2, -13,   3,   2,  -3,  -2,   0,   3,   9,  -3,  -4,   4,  -1,  -2,  -2,   2,   3,   1,  -2,   0,   2,   0 },
     {   -5,  -8, -17, -17,  27,   5,   2,   0,  -7,   7,  27, -13, -15,  16,   0,   0,  12,  28,  -3,  -9,  29,  -2,  -1,  -1, -32,   5,  -4,  -2,  -8,  46,  -3,   1, -23,  10, -38, -18,  -3,  -2,  28,  -2,   7, -22,  26, -48, -11,  -1,   4,   8,   3,  -3,   2,  14, -32,  -7,  -1,   4,   1,  -1,  -1,   4,   0, -11,  -2,   1 },
     {    2,  -6,   8, -10,  -5,  -7,   9,   2,   7,  -3,   8,  -9,  31, -12,  -2,  -1,   3, -16,   1, -17,   2,  -8,   7,   1,  -3, -12,  13, -10,  24, -12,  -3,   7,   2, -34,   9, -18,   8,  -8,   0,   5,  23, -37,  19, -23,  33, -18, -29,  15,  16,  -4,   4, -13,  -7,  48, -39,  -3,   2,   9,  -4,   9, -30,  14,  12, -13 },
     {   -5,  -4,  -7, -45,  21,  12,  -3,   1,  -4,   1, -43,   6, -22,  19,  11,   4,   3, -41,  -9,  16,  -2,  -6,   7,   1, -70, -12,   7,  11,   7,   5,  -2,   1,  -9, -40,   2,  23,   1,   5,   0,   1, -10,  -3, -23,   7,   5,   0,   1,  -2,   1,   1, -11, -12,   1,   0,   5,   1,   2,   3,  -2,  -8,   0,   1,  -1,  -1 },
     {   -6,  -5, -28, -53,  12,   5,  10,  -6,  -8, -23,  -9,  25, -25, -13,   5,   6, -10,   7,  41,  11,  10, -27,  -7,   0,  37,  28,  -5,   8,  11,   4, -21,  -1, -22,  27,  31,  -6,  -6,  20,   0,  -7,   3, -29,   8,  17,   6,  -4,  12,   2,   2,   0,  -9,   2,   8,  11,  -4,   2,   4,  -3,   4,  -2, -11,   7,   5,  -2 },
     {    7,  16,  33,  36,  28, -27,   9,   2,  -1,  -3, -30, -40,   4,  34,   8,  -1, -21, -25,  -7,   9, -17, -19,   5,   2,  26,  34,  11,  -1,   7,   6, -23,   0, -25,   2,   6,  16,   6,  22,  16, -10,  -7, -16,  -7, -18,  11,   0,  15,   7,   0, -15,  -6,  11, -17,   1,   0,   9,  -1,   5,  -9,   3,   6, -11,   0,   2 },
     {    3,   5,  12,   1,   4,  26, -16,   5,  -6,  -8, -16, -18, -10, -34, -23,  -2,   3,  -4,   9,  24,   5,  15, -11,  -4,  -8,  -5,  11, -24, -13, -20,  10,  -1,   3,  24,  17,  21,  21,   7,  -9,  -2, -12, -30, -17, -37, -21, -10,  19,  11,  20,  21,  32, -13, -13, -29, -30,  -4,   1,  -4,  14,  15,   4,  16,  -1,  -4 },
     {    0,  -2,   4,   4, -68,  23,  15,   7,   1,  -9,   1, -19, -61,  11,  22,   6,  -3, -12,   6, -30,  -1, -37,  22,  10,  -2,  -1,  -7, -22,   9, -19, -15,  10,   3,  -4, -11,  -6,   3,   9, -17,  -6,  10,   5,   5, -17,  -3,   7,   0,  -4,   9,   4,  13,   2,  -8,  -8,   8,   1,   1,   4,  -5,   5,   1, -10,  -2,   3 },
     {   -6,   2,   8, -19, -13,  -7,  18,   0,  -8,   7,   1, -24,  41, -23,  -3,  -1, -10,   6,  -6, -10,  21, -10,  15,   1, -29,  21,  -2,  -2,  22, -11,  -3,  11, -43,  49, -18,  10,  -8,   6, -14,  -2, -32,  37, -12,  -2,  12,  11,  -7,  -6,   1,  13,   6, -26,   0,  15,   8,  -3,  -4,   5,  11,  -2, -12,   3,   5,  -3 },
     {    2,   4,   6,  -8,  18,   5,  -8,   5,   0,  -2,  -6,   3,  -5, -14,   3,  -2, -10,   8,  26, -11,  13,   4, -11,  15,   2,  43, -26,  -9,   1,  10, -18, -22,  63, -27, -25,  11,   6,  14,  13,  -3, -31,  35,  -1,  -9,  10, -12,   2,  -2,   1, -15,  34, -13,  -6,  15,  -2,   2,   4,   7, -12,   5, -11,   4,   3,  -7 },
     {   -1,  -1,  -1,   1, -32,  20,  -5,   6,  -3,  -5,   5, -21,  14,  -7,  13, -10,  -4,  23, -23, -10,  29,   1, -13,  12,  24, -14, -39,  31,  -6,  37, -10,   0, -26, -40,  34,  13,   8,  -3,  17,  -3, -14, -16, -35,   1,   7,   0,   5,   1,   7,   2,   5, -25,   1,  -6,   5,   6, -15,   2,   0,  -1,   4,   4,  -9,   2 },
     {    2,   3,   1,  16,  46,  -1, -16,  -3,  -1,   0,  14,  17, -19,   1,  19,  18,   8,  15,  11, -28,  -5,  31,  38,  11,   6, -12, -31, -11,  22, -21, -16,   9, -42,  -8,   9,  23,   6, -15, -29,  11, -10,   0,   3,   7, -16,  -9,  -9, -14,  11, -25,  18,  -1,  -7,  -2,  -7,   3,  -7,   0,  -9,   3,  -3,   5,  -8,  -8 },
     {    1,   4,   6,  -3,  -7,   9,   7, -12,  -4,  -9, -21, -15, -13, -29, -12,  15,  -3,   4,  15,  25,  19,  27,  -6,  -1,  19,  29,   3,  -8, -12,  -4,  34,   9, -36, -39, -23,  -2, -19,  -2, -12,  18,  51,  26, -15, -13,  -3,  12,   0,  -2, -18, -17,  -8,   2,  -5,   8,  -3,   5,   0,   7, -13,  -7,   4,  -1,   9,  -8 },
     {   -2,  -4,  -9, -13, -27, -41,  25, -10,   1,   0,  -3,  27,   9,  19,   3,   9,   8,   9,  22,   9, -18,  33,  28, -14,   8,   5, -10, -20, -11,  -9,  16,   5,   3, -13,  -4,  -2, -15, -11,  24, -12, -20, -19, -34,  -6,  -3,  18,  16,   7,  25,  12,   2, -13, -34,   8,  15,   0,  16,   9,   1,  23,  -9,  -9,  -4,  17 },
     {    0,   3,  -2,   7,   2,  29,  -2,   7,  -3,   0,  -1,  11,  -9,   1, -31, -14,   0,   5,  -1,  -2, -20,  -6,   3,   2,   0,  -8,  -9,   0,  -9,   0, -18,  -6, -15,   1,   8,   0,  -3,   4,  15, -18,  -8,  17, -28,  13, -19,  -5, -32,  15, -26,  17,  -7,  27, -39,  28,  -8, -36, -16,   5, -13,  25,  -3, -11,  51, -10 },
     {   -3,  -5,  -2, -10, -11, -16,  -7,  16,  -4,  -7,  -5,  -3, -11, -31,  17,  -8,   1,  -7,   6,  15, -22,   2,   0,   3,   7, -16,  21, -23, -11,  19,   1,  -9, -32, -17,  13, -29,  19, -17,   8,   5, -59,  22,  28, -18,   2,   6,  -5,   2, -28, -18,  -6,   1,  23,  -3, -10,   3,  16,  -7,  -9,  27,   6, -13,  -1,   5 },
     {    0,   0,   2,   5,  -4,  12, -42,  17,  -2,  -5,   1, -17, -32,  18,  -9, -12,  -3,  -9,   1, -17,  16,  32,  -8, -15,   8,   2,  17,  -8,  27, -16,  37, -29, -18,  15, -10, -14,  16,   2,  39,  -5,  -8,  -9,  -3,  30,   9,  -4,   0,   0,   9,   3,   4, -21,   6,  28,  17,  17,  -2,  -8,  -6, -14, -16, -15,   8,   4 },
     {    0,   0,   0,   0,   3,  -3,  -4,  -4,  -1,  -5,  -1, -13,   3,  21,   6,  -2,  -3,   4,  13,   0,  27, -14,  11, -12,  15,  26,  -4, -13,  12,  -5,   1,   9,   3, -18, -33,   5,  13, -28,  -4,  11, -21, -29, -13,  14, -22, -10, -41,  13, -25,  23, -13,  -3,  22,  -9, -15, -26,  25,  -7,  23, -10,  34,  16,   4,  24 },
   },
   {
     {  109, -38, -11,  -3,  -2,  -1,  -1,   0, -25,  32,  -5,   0,  -1,   0,   0,   0, -31,  10,   8,  -1,   1,   0,   1,   0,  -2,  -5,   4,   1,   0,   0,   0,   0,  -3,  -1,   1,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {    6, -22,   9,  -1,   2,   0,   1,   0,  92, -23, -26,   3,  -5,   1,  -2,   0, -26,  60, -14,  -6,  -1,  -2,   0,   0, -29,   6,  23,  -5,   2,  -1,   1,   0,  -2,  -9,   6,   3,  -1,   1,   0,   0,  -4,  -2,   1,   2,   0,   0,   0,   0,  -1,  -1,  -1,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0 },
     {  -31,  17,   2,   2,  -1,   0,   0,   0,  10,  48, -17,  -7,  -2,  -2,  -1,  -1, -62,   5,  50,  -8,   4,  -2,   2,  -1,  21, -64,  12,  18,  -1,   3,   0,   1,  13,  -3, -25,   6,   1,   1,   0,   0,   2,   4,  -4,  -5,   2,   0,   0,   0,   2,   2,   0,  -2,  -1,   0,   0,   0,   1,  -1,   1,   0,  -1,   0,   0,   0 },
     {   30,  95, -54,  -4,  -8,  -2,  -3,  -1,  18,  -5,  26, -13,   1,  -3,   1,  -1,  -8,  -5,   8,   4,  -1,   0,   0,   0, -33,   7,  15,   0,   2,  -1,   1,   0,   4, -25,   9,   6,   1,   1,   0,   0,   0,  -2,  -6,   3,   1,   1,   0,   0,   1,  -3,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {   16,  -4,  -6,   3,  -1,   0,   0,   0,  58,  -8,   7, -11,  -1,  -2,   0,  -1, -19, -55,  23,  23,  -2,   4,   0,   1,   7,   8, -63,  14,   5,   2,   0,   0, -17,  42,  -4, -31,   5,  -2,   1,  -1,   0,  -3,  21,  -4,  -6,   0,  -1,   0,  -3,   2,   2,   5,  -2,  -1,   0,   0,  -1,   0,   0,   2,   1,   0,   0,   0 },
     {  -15,  -4, -10,   7,  -1,   2,  -1,   1,  16,  82, -36, -18,  -1,  -4,  -1,  -1,  37,  -6,  20, -11,  -2,  -2,   0,  -1, -12,  29,   4,  -9,  -1,  -2,   0,  -1, -20,  -6,  47,  -6,  -3,  -1,   0,   0,   7, -29,   6,  22,  -5,   2,  -1,   0,  -3,   0, -10,   6,   5,  -1,   1,   0,   0,  -5,  -1,   0,   2,   1,   0,   0 },
     {   20,  -5,   3,  -1,  -2,   0,   0,   0,  19,   1, -10,  -1,   1,  -1,   0,   0,  78,  -6,  30, -15,  -8,  -2,  -2,  -1, -28, -29,  15,  43,  -7,   4,  -1,   1, -18,   5, -50,  15,  15,  -1,   2,   0, -10,  14,  -5, -22,   9,   0,   1,   0,  -2,   0,   4,  -7,  -3,   2,   0,   0,  -2,   3,   2,  -2,  -3,   0,   0,   0 },
     {  -10, -39,  15,   6,   3,   1,   1,   0,  15,   7,  33, -15,  -1,  -3,   0,  -1, -14, -44,  11,  28,  -3,   4,  -1,   1, -33,  12,  -2,   3,   3,   0,   0,   0,   9, -51,   4,  29,  -2,   4,  -1,   1,   7,  10, -60,   8,  13,  -1,   1,   0,  -2,  16,  -3, -27,   6,   1,   0,   0,   3,   2,   3,  -5,  -6,   2,   0,   0 },
     {  -14,  -7,   9,   4,  -1,   1,   0,   0, -31,   5,  -4,   1,   5,   0,   1,   0,  -7,  20,   5,  13,  -6,   0,  -1,   0, -82,  -8, -27,  16,  25,  -1,   4,   0,  26, -19,  -9, -53,   9,   5,   1,   0,  22,  -2,  24,  -9, -24,   3,  -2,   0,   8,   0,  -1,  12,  -5,  -4,   0,   0,   1,   2,   1,   1,   2,  -1,  -1,   0 },
     {  -19, -39, -80,  48,   4,   8,  -1,   2, -12, -30, -10, -22,  20,  -1,   3,   0,  -8,   7,  32, -33,  -7,   3,   0,   1,  -1,  21,  -3,   5, -17,   0,   0,   0,  17,   4,  -1,   4,  -2,  -5,   0,   0,  -2,  11,   2,  -4,   3,  -2,  -1,   0,   1,  -1,   8,  -2,  -2,   1,  -1,   0,   2,  -1,   0,   3,  -1,   0,   0,   0 },
     {   -8,   1, -26,  12,   0,   2,  -1,   0,  11,  51, -15,   7,  -9,  -1,  -2,  -1,   0,  -6, -53,  20,  11,  -1,   2,  -1,  -1,  15,  -9, -21,   9,  -3,   1,   0,  14,  -1, -26,   1,  10,   0,   0,   0, -16,  41,   1, -47,  10,  -1,   1,  -1,   2, -14,  47, -13, -12,   4,  -1,   1,   1,  -6,   4,  10,  -7,   0,   0,   0 },
     {  -15,   2,   2,   0,   3,   0,   0,   0, -16,   5,  15,  12,  -3,   0,  -1,   0, -34,  12,  -1,  -3,   2,   1,   0,   0, -10,  40,   4,  27, -13,  -3,  -1,   0, -61,   4, -15,  19,  39, -10,   3,  -1,  24, -45,   3, -41,  19,  11,  -1,   2,  13,  -9,  -3,  -6, -17,   9,   0,   1,   5,   1,  -3,  -1,  -6,  -2,   2,   0 },
     {    4,  10,   4,   2,  -1,   0,  -2,   0,  -7, -23, -69,  41,   9,   4,   1,   1, -19, -61,  -5, -25,  30,   0,   3,   1, -36,  -4,  17, -27, -10,  11,   0,   1, -10,   5,  -5,   9, -17,  -1,   1,   0,  17,  -3,  -3,  -2,   3,  -6,   0,   0,   8,   7,  -3,  -6,   1,   0,  -1,   0,   3,   4,   4,  -4,  -3,   0,   0,  -1 },
     {    6,  23,   0,  16,  -9,  -1,  -2,   0, -13, -11, -40,  12,   8,   2,   0,   1,   6,  35,   5,  13,  -9,  -4,  -1,  -1,  12,  -4, -46,  12,  23,  -3,   1,   0,  -8,   4,  12, -14,   2,   2,   1,   0, -13, -13, -37,   8,  25,  -1,   3,   1,  -1,  17,  -2, -64,   6,   8,  -1,   0,   5,  -6,  30,  -2, -24,   3,   0,   0 },
     {    8,  19,  40, -29,  -1,   0,   0,  -1,  -3,  -2, -28,   5,   2,   1,   1,   0, -10,  -5,  35, -20, -19,   5,  -1,   0,   8,  64,  -8,  23, -19, -12,   0,  -2,  57,  -9,  -1,  11,  19, -10,  -4,   0, -19,  18,   5,  -7,   3,   8,  -2,   0, -10,  -8,  12,   3,  -2,   1,   3,   0,   0,  -9,  -2,   6,   1,   0,   0,   0 },
     {    6,   3,  -8,   0,  -4,   4,  -1,   1,   8,  -8, -19,  13,   7,  -2,   0,   0,   3,  -4,  -4, -11,  -2,   0,   1,   0,  46,   9,  -1,  14,   8,  -9,  -1,  -1, -32, -61,  -9, -31,  20,  17,   2,   2,  39,  25, -13, -10, -44,   3,  -1,   0, -22,  28,   9,  17,   7, -13,   2,  -1,   2,  -5,   4,  -8,   6,   3,   0,   0 },
     {   -6,  -2, -22,   8,   0,   3,  -1,   1,   2,  16,  -2,  29,  -8,  -1,  -1,  -1,  -6,  -1, -26,  10,   6,   0,   2,  -1,  -2,  14,  -3,  25,  -3,  -4,   0,   0,  15,   6, -31,   5,  14,  -4,   0,   0, -10,   4,   8,  25,   4,  -5,   0,  -1, -22, -16, -53,   4,  51,   2,   3,   1,   1,   1,   1, -66,   2,  13,   1,   1 },
     {  -11, -13, -40, -67,  36,  -2,   7,   0, -10, -12, -30, -10, -11,  12,  -2,   2,   0,  -3,  11,  53, -31,  -4,  -2,  -1,   0,   2,  34,  -1,  13, -13,   1,  -1,   0,  27,   0,  -4,   6,   5,  -4,   0,  22,   3, -11,   0,   0,   3,   0,   0,  -4,   6,   2,  -3,   1,   0,   0,  -2,  -8,  -2,   8,  -2,   0,   1,   0,  -1 },
     {    7,   4,  -3,   3,  -6,   2,  -1,  -1,  15,  12,  16,   2, -12,  -1,   0,  -1,  15,  10,  -9, -18,  13,  -1,   3,  -1,   8,   7,  -7, -12,  -7,   5,   0,   1,  53,  26, -23,   8,  -3,  -8,   1,  -1,  74, -31,  -8,  -8,   6,   3,  -4,   2, -16,  44,  -2,  -7,  -9,   2,  -1,   1, -26,   7,  21,   1,   1,  -4,   1,   1 },
     {   -6,  11,  16, -10,  -4,   5,  -1,   0, -11,  10,  15, -20,  10,  -1,   0,  -1, -15,  19,  -2, -17,   0,   3,   1,  -1, -26,  16,  -2,  11, -21,   8,  -3,   1, -36,  41,  -3,   5, -28,   2,   4,   0,  30,  53,  -8,  22,  -7, -22,   3,  -2, -20,   3,  28,  -8,  29, -12,   0,   0,  11, -32,  21,  -2,   1,   7,  -3,   1 },
     {    1,   2,  -3, -26,  23,  -1,   1,  -1,  -1,   1, -22, -57,  26,   3,   5,   1,  -7, -11, -61, -19, -11,  16,   0,   5,  -2, -26, -25,  35, -37,   5,   3,   1,   4, -20,  11,  12,  -1, -14,   2,   0,  -5,  -9,  15,  -5,   5,   1,  -5,   0,  -1,  20, -14,  -2,  -7,   2,  -1,  -1,   3,  10,   7,  -2,  -6,  -3,   0,  -1 },
     {    9,  12,  28,  55, -34,   1,  -3,   0,  -2,  -5, -21, -35,  41,  -8,   1,   2,  -6, -14, -18,  38, -32,   6,  -3,   2,   8,   4,  49,  11,   4,  -9,  -1,  -2,   9,  23,  -5, -12,  11,   2,   1,  -1,  20,  -8,  -5,   3,   1,   1,  -1,   2, -12, -14,   4,   1,   2,   5,  -3,   1,  -7,   6, -16,  -1,   0,   1,   3,   0 },
     {    0,  11,  -2,  21,  -3,  -1,  -1,   1, -10,  -3, -27,   8, -11,   5,   3,  -1,   1,  20,   1,  38,  -3,  -4,  -8,   1,  -8,  -3, -35,   7, -16,   9,   4,   0,  -5,   6,   3,  43,  -8,  -8,   2,  -4,  -7,  -1, -26,  -5,  -3,   4,   4,   1,   5,  27,   7,  40,  -2, -32,   3,  -5, -16,  -6, -26,  -3,  62,  -5,  -2,  -1 },
     {   -5,   0,   2,   5,  -9,   2,   1,   0,  -3,  11,  12,  -2,   4,   1,   0,  -1, -10,  -6,  -3,  -1,  -6,   5,   2,  -1,  -7,  19,  33,  -7, -14,  -3,  -3,  -1,  -4,  19, -19, -22,   4,   7,   0,   2, -58, -10,  -6,  -5, -34,   8,   5,   0,  17,  55, -17,  -5, -11, -25,   3,  -2,  -2,   1,  56,  -1,  14, -12,  -6,  -1 },
     {    1,   5,   7,  -9,  10, -11,   4,  -3,  -4,   0, -12, -34,  18,   7,  -1,   1,  -1,  10,  12,   6,  -7,  -1,  -1,   0,   7,  28, -19, -35,  18,  16,  -3,   1, -12, -25, -71,  15, -38,  21,   8,   1,  -7, -32,  13,  13,  -8, -23,   5,   0,  -5, -14,   5,  -1,   5,  -2,  -7,   1,   2,  21,   0,  -1,   2,   0,  -1,  -4 },
     {   -5,  -3,   0,   7,   6,  -6,  -3,  -2,  -8,  -9,   1,  -1, -11,  -2,   2,  -1,  -4,  -4,   4,   9,  11,  -1,  -5,  -1, -14, -12,   7,  -8,   1,   3,  -1,   2,  -2,  -5,   4,   4,   6,   1,  -3,   1, -25, -26,  20,  -6,   9,   7,  -1,   1,-105,  19,  16,   2, -10,   8,  -1,  -1,  23, -29,   0,   5,   2,  -2,   2,  -2 },
     {   -6, -12, -11, -30, -71,  34,  -1,   3,  -5, -11, -11, -32, -17,   0,   8,   0,  -4,  -6, -11,  -8,  35, -38,  12,   0,  -1,   3,  -2,  21,  30, -20, -13,   6,   4,   5,  -4,  20,   5,  18, -11,  -1,  -7, -14,   4,   6,  -3,   8,   7,  -6,   9,  10,  11,   4,  16,  -1,   3,   0,  11,  -3,  -9,   9,   0,   6,  -4,  -2 },
     {    3,   4,   9,  -9,  28, -12,   4,  -2,  -1,  -1,  -1, -30,  42, -13,   9,  -6,   4,  12,  11,  -2,  63, -10, -12,   3,  -1,   7,  -4, -11,  11,  18,  -7,   0,   6,  23,  11,  -3,  18,  27,  -8,  -5,  -5,  11, -34, -26, -14,  32,   7,  -2,  -7, -14, -17,   8, -10,   1,   5,  -2,  -6,   3,  -5, -31,  -1,   5,   6,   3 },
     {   -5,  -9, -17, -30, -10,  15,   4,   2,   7,  10,  13,  44,  48, -28,  -7,  -5,  -1,  -2, -12,  -8, -14,   6, -14,   6,  -2,  -7,  -2,  19, -19,  17, -10,  -3,  12,   5,  -1, -16, -10,  21,  14, -11, -15, -41, -27,   9,  -7,  -5,  17,   5,  -5,  -9,  40,   1,  10, -12,  -8,  -2, -17, -10, -25,   3,  -7,   1,   4,   2 },
     {   -4,  -8,  -7, -24, -49,   8,   9,  -5,  -1,  -4,  -4,   2,  16, -28,  -3,   1,   8,  15,  22,  42,  33,  21, -26,   4,   6,   1,   2, -13, -35,  24,  19,  -6,  -2, -17,   5,  -6,   6, -37,  12,  -5,   7,  15,  22,   0,  10,  -9,  -4,   7,  -2,   1,   1,  -5,  -5,   1,  -3,   3,  15,  32,  20,  -8,  -2,   0,  -6,  -4 },
     {    5,   0,  10,   0,  19,  -2,   2,   2,   0, -10,   1, -23,   6, -16,  -5,   3,   5,  -5,   8,  -2,  14,   3,   0,  -5,   4,  -6,  10, -21,  11,   1,   8,  -4,   5, -12,  -2, -13,  27, -25,  13,   3,  -7, -21,  15, -14,  21, -25, -16,   1,  33,  24,  24,  -2,  42,  -2,  -7,  -2, -12, -58,  -5, -20,  16,  35,  -8,  11 },
     {    2,   1,  -1,  -7,  -6,   3,   3,  -1,   3,   0,  -1,  -5,  -2, -11,  -3,   0,   5,   3,  -3, -16, -24,  18,  10,   6,  -7, -17, -17, -44, -14, -28,  29,   8,   2,  16, -18,   2,  31, -20, -28,   0,   6, -15, -33,   8, -23,  19,   1,  -3,   3, -14,   6,   4,   3, -22,   7,   5,  64,   1, -13,  -8,  -7,  12,  -7,   2 },
   },
   {
     { -105,  -8,  30,   8,   5,   2,   2,   1,  56, -15, -26,   0,   0,   1,   0,   0,  -1,  15,   1,  -7,  -2,  -1,   0,   0,   4,  -3,   2,   1,  -1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   27, -81,  -4,  21,   3,   3,   1,   1,  35,  60, -32, -25,  -2,  -2,  -1,  -1, -22,  11,  34,   0,  -7,  -2,   0,   0,  -2, -10,  -4,  11,   4,   0,   0,   0,  -4,   3,   0,  -4,   1,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0,  -1,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {  -24, -63,   0,  24,   7,   2,   1,   1, -59,  -4,   4,  -8,   6,   4,   2,   1,  73,  15, -28, -11,  -8,  -2,  -1,   0, -15,  20,  19,  -5,  -3,  -3,  -1,   0,   8,  -6,   0,   6,  -1,   0,   0,   0,  -3,   4,   0,  -1,   1,   0,   0,   0,   3,  -1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   45, -11,  45, -12, -12,  -1,  -1,  -1,   4, -50, -55,  34,  16,   2,   0,   2,  -8,  45,  -8, -42,  -3,   4,   1,  -1,   0,  -1,  23,  10, -13,  -4,   0,   0,  -3,   3,  -7,   4,   8,  -1,  -1,   0,  -1,   0,   4,  -2,  -2,   2,   1,   0,  -1,   1,  -2,   0,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
     {    0, -33,  46,   7,  -7,  -2,   0,   0, -16,  -1, -11,  16,  10,   0,  -2,   0, -37, -59, -34,  10,  15,  10,   4,   1,  34,  34, -25, -40, -10,   0,   1,   1,   1,  12,  26,   4, -14,  -6,  -1,  -1,   3,   1,  -4,   7,   6,  -3,  -1,   0,  -1,   1,   2,  -2,   1,   2,  -1,   0,   1,   0,  -1,   1,   0,   0,   0,   0 },
     {  -14,  38,  37,  16, -23,  -4,  -2,  -1, -50,  25, -43, -21,  31,   1,   2,   1,  15, -26,  50,  -9, -29,   3,   0,   0,  10,  -3, -18,  30,  10,  -9,  -1,   0,   1,   0,   1, -13,   7,   6,  -1,   0,   3,  -3,   1,   6,  -4,   0,   2,   0,   0,  -1,  -1,  -3,   2,   0,   0,   1,   1,  -1,   0,   1,  -1,   0,   0,   0 },
     {   16,  20,  60,   6, -18,  -7,  -3,  -2,  30,  47,   7,   9,   3,  -7,  -4,  -2,  31, -19, -32,  11,  19,   8,   0,  -1, -64, -16,  14,  -7,   0,   8,   6,   2,  10, -18, -16,   2,  -1,  -3,   0,   2,  -3,   5,  -2,  -6,   0,   0,  -1,  -1,   3,   0,   1,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0 },
     {   10, -20,  29,   3,   6,  -5,   0,  -1,  -6, -44,  11,   0,  -3,  11,   2,   2,   7,  18,  13,  44,  -4, -15,  -4,  -1, -18, -27, -49,  -6,  45,  22,   1,   0,   0,   8, -12, -43, -32,   7,  10,   3,   3,   7,  10,   9,  -8, -18,  -5,   1,   2,   4,   1,   0,   5,   3,  -3,  -2,  -1,   1,   1,   1,   0,   0,   1,   0 },
     {   26,  23,  -1,  57,   2, -19,  -4,   0,  35, -33, -11, -63,   6,  26,   4,   1,   3,  -4, -18,  15, -28, -13,   5,   2,  -3,  36,  19,  -9,  12,  -7,  -9,  -1,  -1,  -7,  14,  14,  -6,   5,   2,  -1,  -6,   1,  -7,   2,  10,  -2,   1,   1,  -1,  -2,   1,  -4,  -3,   3,   0,   0,  -1,   0,  -1,   2,   0,  -2,   0,   0 },
     {   -5,  25, -29,  22,   8,   0,  -2,   0, -17,  16, -43,  -7,  -1,   6,   6,   1,  -3,  13,  -3, -25,  21,  -6,   0,   0, -11, -36,   3, -47, -16,  26,   7,   2,  13,  10,  12,   1, -50, -32,   5,   3,   7,   6,  13,  22,  19, -16, -19,  -3,   1,   2,   1,   2,   6,  13,   1,  -5,  -1,  -1,   1,   1,   1,   0,   4,   1 },
     {   -3,  -4,  -7,  51,  14,  -8,  -2,  -1,  -2,  -6,  34,   3,  22,  11,  -3,  -2, -29, -13, -34, -44, -21,  25,  21,   6, -13, -46, -18,  19, -12, -28,  -2,   6,  11,  18, -12, -23,  11,   5, -10,  -6,   7,   8,  16,   0, -15,  -2,   5,   0,   2,   2,  -1,   4,   6,  -4,  -4,   0,   0,   1,   1,   0,   0,   2,   2,  -1 },
     {    4,  -7, -45, -21, -46,  11,  11,   2,  36, -14, -10,  16,  45, -19, -13,  -1,  30, -13,   7,  -4, -19,  27,   5,  -2, -31,  24, -26, -12,   0, -19,   5,   4,  11,   8,  27, -14, -12,   3, -10,  -2,  -5,   6,   4,  21,  -2,  -9,   2,  -3,   0,  -1,   0,  -3,  10,   5,  -3,   0,  -2,   0,   0,   1,  -3,   2,   2,   0 },
     {    8,  13,  16,  31,   0,  -7,  -3,   0,  13,  22,  34,  27,  13,  -3,  -5,  -3,  26,  44,   4, -17,  -4,  11,   3,   0,  34,  -2, -34, -11,   3,   3,   6,   4, -57, -26,  36,  10,  -8,  -4,   0,   1,  10, -13, -23,   9,  10,  -1,  -2,  -2,  -2,   5,  -1, -12,  -1,   5,   2,   0,   0,  -2,   1,   1,  -2,  -2,   1,   1 },
     {  -10,  14, -24,  15, -11,  -3,   2,   3, -15,  33, -33,  34,   1,   1,  -4,  -1, -23,  27, -40,  15,   2, -12,   7,  -2,   5,  11,  17, -14,  50, -12, -13,  -2,   6, -19,   9, -28,   3,  46,   4,  -4,  -3,  -5,  -2,  -5, -27,  -9,  19,   6,   2,  -1,   1,   1,   3, -10,  -7,   4,   0,  -1,   2,   0,   0,   2,  -3,  -3 },
     {   -5,  -1,  26, -64,   7, -11,   9,   1, -10,  26,  20, -31,   5,  19, -10,   0, -18,  32, -23,   1, -49,  -2,  15,   2,   1, -13,   5,  -3,  -1, -22, -13,   1,  20,  -2,  22,  18, -21,   4,  -2,  -7,  -6,   0,  -1,  30,  14, -15,   1,   2,   1,  -1,  -5,  -3,  12,   9,  -4,  -2,  -1,  -2,   1,   0,  -2,   1,   3,   0 },
     {  -12,  -7,  -2,  21, -51,  -7,  18,   1, -18,  -5,  23,  -8,  51,   2, -21,  -2, -34,  35,  13,  39,  13,  15,   3,  -8,  10, -11,  19, -18,  -9,   7,  16,   9,  17, -14, -29,  18,   7, -13,  -4,   5, -11,   3,   1, -15,  11,  10,  -5,  -5,   5,  -8,   2,  -5, -10,   6,   5,   1,  -1,   6,  -3,   0,  -2,  -6,   1,   1 },
     {   -7, -10, -12, -13,  17,   8,  -2,   0, -11, -17, -17,   5,  13,   9,   1,   2, -20, -19,  10,   0, -23,  -4,   3,   9, -36, -20, -11, -22,   0,  -5,  -8,  -4, -21, -58,  -2,  28,   3,   2,  -5,  -6,  39,  31, -38, -25,  -4,  -2,   4,   2,  -3,  18,  24,  -5, -17,  -2,  -3,  -1,   0,  -5,   3,  15,   5,  -4,  -4,  -1 },
     {    3,   6,   2,   9,  20,   7,  -2,  -1,  -5,  -3, -11,  11,  -4,   8,   7,  -9,   6,  16,   9,  19,   0,   0,   2,   0, -22, -20, -29, -19,   0,  -5,  -4,   4,  39,  34,  38,  23,  25,  -6, -20, -15, -25, -36, -17, -17,   7,  36,  28,  15,   2,   7,  -4, -14, -29, -11,   4,   6,   1,  -4,   5,   6,   1,  -2, -12,  -4 },
     {   -3,  -5,  -8, -17, -43, -11,  15,   7,  -5,  -5, -14, -40,   0,   4,  -5,   8,  10,  -1, -36,  15,  21,  -9, -11,   3,   2, -38,   3,   7,  -6,   1,   3,  -2, -46,  39,  19,  -1,  19,   4,  -7,  -4,  40, -23,  -7,  -8, -15,   8,   0,  -6, -12,  24,   7,  -1,   5,  -9,  -1,  -3,   5,  -9,   6,   3,  -5,   4,   2,   4 },
     {    3,   9,   0, -11,  31,  14, -22,  -8,  -2,   4, -21, -29,   0,  -9,   8,   8,   0,  20, -26,  23,  -4,  49, -25, -10,  11,   9, -19,  -4,  -7,  -6,  47,   6,  -6, -26, -10, -16,   6, -20, -37,  12,   1,   5,  19,   8, -20,  11,  20, -11,  -5,  -1,   3,  16,   1,  -5,  -1,  13,  -6,  -4,  -3,   1,   1,   3, -12,  -5 },
     {   -3,   2,   6,  14,  -4,  39,   4, -18, -11,  -9,   7, -13, -11, -60, -14,  12, -24,   6, -16,  -9,  -1,  13, -10,  -7, -13,   3,  16,  40,  34,   8,  15,  -5,  10,  13,  22,  -5, -23, -10,   7,  -2,  16,   9, -40,   4,  33,   7,   2,  13,  -3,  -4,   7,  -9,  -8,   3,  -3,   1,   2,   5,   1,   4,  -2,  -5,   1,  -5 },
     {    2,   1,   5,  -3,  13,  -3, -16, -20,   2,  -3,   6, -20,  22, -22,  -6,   9,   9,   3,  15,   0,  10,  41,  -5,  -5,   8,  -9,  11, -20, -18,  28, -26, -21,  -9,  19,  -3,  19, -12,  38,  25, -24,   0,   4,   5, -11, -20, -32,  32,  34,  -3, -13,  -8,   9,  -5, -14, -26,   5,  -5,  -3,  -3,   5,   8,   3,  -5, -12 },
     {   -8,  -7,  -4,   1,  -4, -56,   1,  22, -14, -13,  -6,  -6, -24,   0,  19, -20, -17,   5,  12,   7,  29,  32,  14,  -8, -28,   4,  -2,  17, -13,  -1,  12,  -8,  -1, -22,  30,  -7,  18,  -2,   5,  14,   0,   7, -25,  18,  -2,  -6,   3,  26,  -2, -16, -36,   1,  23,   9,  -8, -14,   0,  21,  13,   0, -10,   3,  -1,  -4 },
     {   -2,   0,  -1,  -8,  42,  10,  -2, -20,  -1,   3,  -1,  17,  42,  32,  -9, -23,   0, -10,  13,  30,   4,   7,  32,   3, -13,   5,  46,   9,   1, -12,  18,   7, -31,  21,  11, -24, -20, -18,  14,   9,  12, -17, -11,  -2,   4,  12,  -7,  12,  -5,  -4,  -7,  16,   2, -12,   0,  12,  11,  -6,   3, -17,  -7,   3,  -5,  -2 },
     {   -1,  -2,  -3,  -4,   4,   3,  19,   0,   6,   6,  14,   3,  39,  25,  29,   1, -11,  -4, -13, -20, -10, -18, -57,  -7,  -3,   4,  -5,  18,  -1,  36,  18, -29,  25, -13,  22,   3,  -1, -12,  26,  24,  -4, -24, -12,  -8, -37,  -3,  -9,  10,  -8,  -4,   8,   5,   4,  -7,  -6,  -3,  -2,   4,  -3,  -3,  -6,   0,  -3,   1 },
     {   -4,  -4,  -6, -13,   3,   1, -15,  -7,  -5,  -6,  -4, -18,   9,  11,  -9,   2,  -8,  -8,  -7, -14,  28,  17,  -3,   1,  -7,  -1,  -2,   9,   3, -13, -12, -15, -22, -15,  -9, -17,  -7,   3,  14,  11, -50, -32, -10,  16,   7,  -1,  -6,  -5,  64,  32,  19, -21, -26,   2,  -4,   0, -15,   2,  13,  22,   4,  -4,   3,   1 },
     {    3,  -6, -21, -14,   6, -30,  -6,  -8,  11,   2, -27,  13,   8,   7, -11,  -9,  23, -24, -32,   8, -19,   5,   2,  -4,  33, -34, -11,  14,  17,  16,  14,   4,  -4,   3, -28,   6,   7,  -9,  17,  -4, -24,  11, -23,  -7,  27,   8,   2,  21, -13, -27,   0, -26,   8,  11,  -5,  10, -17,  11, -24,  13,  -9, -22,  -2,   5 },
     {    4, -10,   4, -12,  15, -31,   2, -13,   9, -12,  10, -16,  29, -15,  -2,  19,  13, -17,  15, -24,  43, -18,  -8,   2,  21, -28,  20,  -9,  27, -14, -10,  12,  21, -28,  19, -12, -15,  16, -32, -10,  -1,   0, -12,  15,  -6,  31,   5, -22, -13, -16,  -8,  16,  -9,  -3,   8,   3,  -1,  10,  -7,  -4,  -1,   6,  -6,  -3 },
     {    0,  -8,   0,   3, -39,  15, -32, -27,   7,  -2,  13,   9, -11,  28,  59, -16,   6,  -5,   5,   2,   4,  -3,  -5, -44,  15, -16,  13,   3,  -1, -23,   2,  12,  10,  -2,  -2,   4, -12,   5,   9,   8,  18,   7,  -4,  22,  16,   0,  26,   7,   0,   3,  14,  -3,   4,  -1,  -6,  25,   6,   1,  15,  16,  24,   2, -11, -13 },
     {   -7,  -8,  -2,   3,  -9, -17, -47,   8,  -6,  -5,   7,  18,   0, -15,  21,  31, -12, -11,  10,   2, -24,   5,   6,  10, -17, -13,  23,  -9,  43,   6,  19, -15, -22,  22,  11,  47,   7,  -8, -10,  19, -14, -10,  12,  12, -14,  -4,  -4, -23,  19,   0,  -4,   7,  10,   8,  -3,   6,  -4,  -1, -29, -16,  -2,   3,   0,  -2 },
     {   -2,   0,  -2,   6,   3,   1, -30,  -3,   0,   2,   0,   9,  28, -31,  21, -19,  -4,   5, -11,  28,  -3, -38, -10,  47,   2,  -7,   4,  17, -38,   3, -11, -31, -10,   2,   3,  -7, -11,  19, -32,  23, -16,  19,   8,  12,  17,   2,  16,  12,  -8,   9,  -1,  -4,   8,   4,  17, -12,  12,   8,  11,  18,  -9, -12,   4,  -2 },
     {    0,  -5,  -2,  -5, -13,  -5, -37,  -4,   3,  -5,   2,  -9,   3, -30,  11,  -7,   2,  -1,  -6,  -3, -15, -12,  40,  -8,  11,  -3,  -2,  -9,  -7,   9,  16, -18,  19, -25,  -7, -22,   8, -10,  10, -36, -10, -42, -13, -19,   1,  -6, -29,  11, -25,  32,  -2,  31,   7,  19,   5,  -2,  -8, -22,  11,  -5,   1,  10,  -9,  -8 },
   }
 }
};

const int8_t g_lfnst16x16[ 35 ][ 3 ][ 32 ][ 96 ] = {
 {  //0
   {
     {  107, -39, -15,  -2,  -3,  -1,  -1,   0,  -1,   0,   0,   0, -48,  25,   5,   0,   1,   0,   0,   0,   0,   0,   0,   0, -15,   1,   5,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -31,  20,   3,  -3,   1,  -1,   1,   0,   0,   0,   0,   0, -95,  16,  24,   1,   4,   0,   2,   0,   1,   0,   1,   0,  61, -30, -10,   3,  -2,   0,  -1,   0,   0,   0,   0,   0,  15,   4,  -7,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {  -31, -96,  55,  12,   8,   2,   3,   2,   2,   1,   1,   1,  -3,  32, -25,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  22,  23, -12,  -7,  -2,  -1,  -1,  -1,  -1,   0,   0,   0,  -5,  -3,   5,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   3,  -2,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   1,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {  -31,   6,  -2,   6,   2,   1,   0,   0,   0,   0,   0,   0, -15,  61, -11, -14,  -1,  -3,   0,  -2,   0,  -1,   0,  -1, -69, -19,  40,   2,   3,   1,   2,   0,   1,   0,   1,   0,  54, -31, -14,   7,  -2,   1,  -1,   1,   0,   0,   0,   0,   6,   6,  -4,  -2,   0,   0,   0,   0,   5,  -4,   0,   1,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0 },
     {   14,  23,  -8,  -9,  -1,  -2,   0,  -1,   0,  -1,   0,   0,  35,  75, -34, -16,  -6,  -4,  -2,  -2,  -1,  -1,  -1,  -1,  44, -45,  12,   6,  -2,   1,  -1,   1,   0,   0,   0,   0, -51,  -8,  14,   7,   3,   1,   1,   1,   1,   0,   0,   0,  -1,  -4,   0,   1,   1,   0,   0,   0,  -6,  -3,   1,   1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -4,  -2,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,  -1,  -1,   0,   0 },
     {   21,  30,  92, -56,  -5, -11,  -1,  -6,  -1,  -3,  -1,  -2,  -3,  -5, -18,  23,  -3,   2,  -1,   1,   0,   1,   0,   0, -23, -26, -23,  17,   6,   4,   2,   2,   1,   1,   1,   1,   8,   7,  -1,  -6,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,  -4,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   1,   1,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
     {   -1,  39,   3, -13,  -4,  -3,  -1,  -2,  -1,  -1,   0,  -1, -28,  14, -31,  11,   5,   1,   2,   1,   1,   0,   1,   0,   0,  79,  -4, -30,   0,  -6,   0,  -3,  -1,  -1,   0,  -1, -16, -48,  42,   9,  -2,   2,   0,   1,   0,   0,   0,   0,  14, -22, -11,   9,   1,   1,   0,   1,   2,   1,   1,  -2,   0,   0,   0,   0,   3,  -4,  -1,   2,   0,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0,   0,   1,  -2,   0,   1,   0,  -1,   0,   0,   1,  -1,   0,   0,   0,  -1,   0,   0 },
     {  -14, -13, -15,   5,   5,   1,   2,   1,   1,   0,   1,   0, -32, -31, -57,  33,   9,   7,   3,   4,   2,   2,   1,   1, -25, -33,  43, -11,   2,   1,   1,   0,   0,   0,   0,   0, -37,  36,  17, -10,  -3,  -2,   0,  -1,   0,  -1,   0,  -1,  46,  -3, -11,  -1,  -2,  -1,  -1,   0,  -2,   7,   4,  -3,  -1,  -1,   0,   0,   8,  -1,  -1,   0,  -1,   0,   0,   0,   0,   3,   2,  -1,  -1,   0,   0,   0,   4,   0,  -1,   0,   0,   1,   1,  -1,   2,   0,   0,   0,   0,   1,   1,  -1 },
     {   13,   7, -12,   9,  -4,   1,  -2,   0,  -1,   0,   0,   0,   8, -11, -72,  23,  10,   3,   3,   2,   1,   1,   1,   1,  35,  -4,   5, -20,  -1,   0,   0,   0,   0,   0,   0,   0,  66,  11,   4,  -7,  -9,  -2,  -3,  -1,  -2,  -1,  -1,  -1, -55,   9,  16,   1,   3,   0,   1,   0,   2,  -6,   6,   0,   0,   0,   0,   0, -10,   0,   2,   0,   1,   0,   0,   0,   1,  -1,   3,  -1,   0,   0,   0,   0,  -5,   0,   1,   0,   0,   0,   1,   0,  -3,   0,   1,   0,   0,   0,   1,   0 },
     {   -4,  14,  -2,  53, -22,   2,  -7,   1,  -3,   0,  -2,   0,  -6,  28,  -7,  -7,   7,  -4,   1,  -2,   0,  -1,   0,  -1, -33,  -6, -69,  -5,  21,   4,   7,   2,   3,   1,   2,   1,  -2,  41,  30, -32,  -6,  -3,  -1,  -2,  -1,  -1,   0,  -1,   5, -24,  28,   7,  -5,   0,  -2,   0,   7,  -7,  -7,   4,   1,   0,   0,   0,   1,  -2,   3,  -2,   0,   0,   0,   0,   3,  -2,   0,   1,   0,   0,   0,   0,   1,  -1,   2,  -1,   1,  -1,   0,   0,   0,  -1,   1,   0,   1,  -1,   0,   0 },
     {   18,  16,  44,  75, -44,   0, -15,  -1,  -6,  -1,  -3,   0,   0, -24,   0, -15,  20,  -2,   4,   0,   2,   0,   1,   0,   9, -21,  12, -25,   4,   3,   2,   1,   1,   1,   0,   0,  -9, -35,  -8,  25,  -2,   3,   0,   2,   0,   1,   0,   1,   8,  26, -24, -11,   5,  -2,   1,  -1,  -3,   5,   6,  -3,  -2,   0,   0,   0,   0,   3,  -3,  -2,   1,   0,   0,   0,  -1,   2,   1,   0,  -1,   0,   0,   0,   0,   2,  -2,  -1,   0,   1,   0,   0,   0,   1,  -1,  -1,   0,   1,   0,   0 },
     {    5,  17,  33,  10, -10,  -4,  -4,  -2,  -2,  -1,  -1,  -1,   0,  16,  15, -20,   2,  -1,   0,  -1,   0,   0,   0,   0,   6,  31,  58, -26, -12,  -3,  -3,  -3,  -2,  -1,  -1,  -1,  -2,  67, -28,   5,   0,  -5,   0,  -2,   0,  -1,   0,  -1, -15, -51,   1,  11,   6,   4,   2,   2,   5,  -4,  -8,   3,   0,   1,   0,   0,  -1,  -7,  -3,   1,   2,   1,   0,   0,   1,  -2,  -3,   1,   0,   0,   0,   0,  -1,  -3,  -2,   1,   1,  -1,  -1,   1,   0,  -2,  -1,   0,   0,  -1,  -1,   1 },
     {  -12,  -8,   5,  -6,   2,   1,   1,   0,   0,   0,   0,   0, -22, -18,  -4, -27,  13,   2,   5,   1,   2,   1,   1,   0, -23, -12,  11,  14,  -7,   2,  -1,   1,   0,   0,   0,   0, -41, -11,  29,  12,   0,  -1,   0,   0,   0,   0,   0,   0, -73,   2,  11,   1,   6,   1,   2,   0,  65,  -5, -18,   1,  -5,   0,  -2,   0,  -4,   9,  -1,  -2,   0,  -1,   0,   0,  14,   0,  -3,   1,  -1,   0,   0,   0,  -2,   3,   0,   0,   7,   0,  -1,   0,   0,   1,   0,   0,   4,   0,  -1,   0 },
     {   -5,   0,   1,  16,  -7,   1,  -2,   1,  -1,   0,   0,   0,   6,  19,  30,  93, -32,  -7, -11,  -3,  -5,  -2,  -3,  -1,  -2,  -6,  12, -29,  26,  -3,   3,  -1,   1,   0,   0,   0, -21, -18, -20, -21,   9,   7,   4,   2,   2,   1,   1,   1, -28,  -1,   5,   1,  -1,   1,   1,   0,  21,  -1,  -8,  -8,   1,   0,   0,   0,   0,   3,   1,   0,  -1,   0,   0,   0,   4,   0,  -2,  -4,   1,   0,   0,   0,   0,   1,   0,   0,   2,   0,  -1,  -2,   0,   0,   0,   0,   1,   0,  -1,  -1 },
     {    5,   8,   1,  33,  35, -21,   0,  -7,   0,  -3,   0,  -2,  -8,  -9, -33,   2,  -8,  13,   0,   4,   0,   2,   0,   1,  -1,   3, -16,  37, -14,   0,  -1,   0,   0,   0,   0,   0,  -9, -27, -55,   4,  30,   0,   5,   1,   2,   1,   1,   1,   1, -52,  21, -20,  -5,   9,   0,   3,   9,  37,   3,  -6,  -6,  -4,  -1,  -2,  -1,   0,   8,  -4,  -1,   0,   0,   0,   1,   5,   3,  -1,  -1,  -1,   0,   0,   0,  -1,   3,  -2,   0,   3,   2,   0,   0,   0,   1,  -1,   0,   1,   1,   0 },
     {   -8,   8,  -6, -17, -41,  17,  -1,   5,   0,   2,   0,   1, -14,   8, -24,  -3,  12,  -5,   4,  -1,   2,   0,   1,   0,  -6,  33,  -4,  13,  12,  -8,  -1,  -3,  -1,  -2,   0,  -1, -29,  12, -56,  -2,  16,   1,   5,   1,   2,   0,   1,   0,   1,  45,  39, -34,  -6,  -3,  -2,  -3,   2, -37,  19,  17,  -5,   2,  -1,   1,   7,  -2,  -5,   1,   2,  -1,   0,   0,   1,  -6,   4,   2,  -1,   1,   0,   0,   2,  -1,   0,   0,   1,  -3,   2,   1,   1,  -1,   0,   0,   0,  -2,   1,   0 },
     {    2,   7,   3,  -5,  61, -27,   6,  -8,   2,  -4,   1,  -2,  -3,  -1,   1, -16,   2,   8,  -1,   1,   0,   1,   0,   0, -13, -12, -22, -80,  -5,  17,   7,   6,   3,   3,   2,   2, -15,   2, -24,  25, -12,  -2,   1,   1,   1,   0,   0,   0,  -1,  25,  32,   9,  -8,  -5,  -4,  -3,  -5, -11,   4,   6,   1,   0,   0,   0,   8,  -1,  -2,   4,  -2,   0,  -1,   0,  -1,  -1,   0,   2,   0,   0,   0,   0,   2,   0,   0,   3,   0,  -1,   0,   1,   1,   0,   0,   1,   0,   0,   0,   1 },
     {    8,  21,  15,  40,  52, -30,  -2,  -9,  -2,  -4,  -1,  -2,  -2,  11,   7,  13, -11,   6,  -4,   1,  -1,   0,  -1,   0,   2,  23,  15,  34, -33,   1,  -5,   0,  -2,   0,  -1,   0,   0,  19,  27, -29,  11,  -5,   1,  -2,   0,  -1,   0,  -1,   0,  47, -18, -24,   0,   1,   2,  -1,   4, -36,   7,   9,  -1,   3,   0,   1,  -3,  -3,  -6,  -3,   0,   1,   1,   1,   2,  -6,  -1,   0,   1,   0,   0,   0,  -1,  -1,  -2,  -2,   0,  -3,  -1,   0,  -1,  -1,  -1,  -1,   0,  -2,   0,   0 },
     {   -4,  -2,   9,  10,  -4,  -4,  -1,  -1,   0,   0,   0,   0, -12, -13,   5,  -5, -18,   8,   0,   3,   0,   1,   0,   1,  -8,  -6,  20,  14,   5,  -6,  -1,  -2,  -1,  -1,   0,   0,  -8, -17,  23,  -7,   4,   3,  -1,   0,   0,   0,   0,   0, -37, -11,  40,   8,  -6,   2,  -1,   0, -75,  -8,  -1,  10,  11,   0,   3,   0,  60,   1, -25,  -5,  -3,   0,  -1,   0,  -7,   7,  -1,  -1,   2,  -1,   0,   0,  12,   1,  -4,  -2,  -2,   2,  -1,   0,   5,   0,  -2,  -1,  -2,   1,   0,   0 },
     {   -9,   1, -14,   0,  -2,   1,   1,   1,   1,   0,   0,   0,  -9,  12,  -5,   8, -17,   4,  -1,   1,   0,   0,   0,   0, -18,   3, -31,   0,   0,   0,   4,   1,   2,   1,   1,   0,  -9,  29,  -6,  42,   7, -12,  -2,  -4,  -1,  -2,   0,  -1, -38,   1, -58, -16,  35,   1,   7,   1, -19,   7,  46, -31,  -9,   3,  -1,   0,  27, -17,   3,  11,  -5,   0,  -1,   0,   0,   2,   3,  -4,   0,   0,   0,   0,   6,  -3,   2,   2,   1,   0,   2,  -2,   2,  -2,   2,   1,   0,   0,   1,  -1 },
     {    2,   2,   2,   3, -13,   3,  -2,   2,  -1,   1,   0,   0,  -6, -11, -12, -27, -98,  37,   3,  11,   2,   5,   1,   3,   5,  -1,   2, -10,  23, -23,   4,  -3,   1,  -1,   1,  -1,   9,   8,  12,  18,  26,  -9,  -5,  -4,  -2,  -2,  -1,  -1,  14,  14,   6,   1,  -6,   2,  -2,  -1,  18,  -2, -10,   4,  10,  -3,   0,  -1, -15,  -1,   3,  -1,   0,   1,   0,   0,   1,  -1,  -1,   1,   4,  -1,   0,   0,  -3,  -1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0 },
     {    0,   2,   1,  -4,  14,  -5,   1,  -2,   0,  -1,   0,   0,  -5,  -6,  -8, -39,  -3,   3,   6,   2,   2,   1,   1,   1,   2,   2,   4, -13,  44,  -8,  -2,  -2,  -1,  -1,   0,  -1, -10,  -9, -22, -76,  -3,  24,   5,   6,   2,   3,   1,   1, -21,   3, -46,  17, -24,   0,   5,   2,   1,  12,  32,   9,   0,  -7,  -4,  -2,   6,  -5,   5,   9,  -5,   1,  -1,   0,   2,   1,   2,   6,   1,  -1,  -1,  -1,   1,  -1,   1,   2,   1,   0,   1,   3,   0,  -1,   1,   1,   0,   0,   1,   2 },
     {    0,   5,   7,  -3,  -5,  53, -19,   4,  -6,   2,  -2,   1,  -2,   6,   4,  -3, -11,   0,   7,  -1,   1,   0,   0,   0,  -7,  -1,   0, -18, -68,  -1,  11,   5,   4,   2,   1,   1,  -5,  -2,   4, -35,  18,  -8,   1,   1,   1,   0,   0,   0,  -4,  25,  20,  10,   6,  -8,  -3,  -3,  -1,  54,  13,  -6,  -2,  -6,  -2,  -3,  -1, -35,   2,  14,   6,   0,   0,   1,   3,   1,  -6,   2,   2,   0,   0,   0,   0,  -5,  -1,   2,   1,   0,  -2,   1,   0,  -2,   0,   1,   0,   1,  -1,   0 },
     {    0,  -5,  -6,  -3, -12,  16,  -5,   3,  -2,   1,  -1,   1,  -2,  -9,  -8,  -1, -15,   6,   2,   1,   1,   0,   0,   0,  -4, -14, -16, -15, -50,  15,   6,   5,   2,   2,   1,   1,  -1, -17, -19, -22,  25, -11,   5,   1,   2,   1,   1,   0,   3, -29, -27,  32,  21,  -3,  -2,   0,  -2, -62,   7,  15,  -2,   7,   0,   2,   1,  49,  -4, -13,   3,  -6,   0,  -2,   1,  -2,   9,   1,  -1,   0,  -1,   0,   0,   8,   1,  -1,   0,   0,   3,   0,   0,   3,   0,   0,   0,  -1,   1,   1 },
     {   -7,  -8, -11,  -8, -54, -78,  38,  -1,  12,   0,   5,  -1,  -1,   3,   1,  -4,  -4,  17, -20,   6,  -3,   2,  -1,   1,  -1,   4,   3,  -5, -21,  45,  -7,   1,  -3,   1,  -1,   0,  -1,  -3,   0, -22,  11, -16,  10,  -1,   2,   0,   1,   0,  -2,  13,   4,   9,  11,  -2,  -5,  -1,   0,  21,   0,  -1,   0,  -1,   1,  -2,  -2, -13,   6,   4,   4,   3,  -1,   0,   0,   0,  -2,   1,   0,   0,   1,   0,   0,  -2,   0,   1,   0,   0,  -1,   1,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {    0,  -8,   0, -12,  -4,  -3,   4,   2,   1,   1,   0,   0,   6,  -3,  16,  -1,   8,  -1,  -3,  -1,  -1,   0,  -1,   0,   1, -16,   3, -29,   0,  -1,   4,   2,   1,   1,   1,   1,  12,  -5,  31,  -4,  36,  -5,  -8,  -2,  -3,  -1,  -1,   0,   1, -29,   3, -65, -19,  30,   2,   6,   6,  -2,  35,  38, -37,  -7,  -1,  -2,  -7,   2, -24,  22,  16,  -5,   0,  -1,   1,   6,  -4,  -5,  -3,   1,   0,   0,  -2,   2,  -3,   4,   0,   2,  -2,   0,  -1,   1,  -1,   3,   0,   1,  -1,   0 },
     {   -6,  -3,   2,   3,   2,  -2,  -1,   0,   0,   0,   0,   0,  -9,  -4,   2,   3,   0,  -3,   1,   0,   1,   0,   0,   0,  -9,  -4,   4,   5,   5,   1,  -1,   0,   0,   0,   0,   0, -16,  -6,   3,   7,   2,   2,   0,   0,   0,   0,   0,   0, -13,  -8,   8,   9,   0,   0,   0,   0, -52, -15,  21,   5,   7,   2,   1,   0, -80, -17,  23,   1,   3,   3,   3,   1,  61,   1, -23,  -2,  -3,  -1,  -1,   0,  -4,   7,   1,  -3,  12,   0,  -3,   0,  -4,   2,   2,  -1,   7,   0,  -2,   0 },
     {   -2,   1,  -2,  -9,   1,   0,   2,   0,   0,   0,   0,   0,  -2,   3,  -4,  -5,  21,  -4,   1,  -2,   1,  -1,   0,   0,  -5,   0,  -5, -21,   2, -16,   6,   0,   3,   0,   1,   0,   0,   7,   0,  -9,  54,  -6,  -7,  -2,  -2,  -1,  -1,   0, -13,  15, -23, -24,  -6,  24,   2,   2, -31,   7, -71,  -8,   0,  -3,  10,   2,   7,  11,  51, -11, -12,   0,  -4,  -1,   3, -11,   1,  14,  -5,   0,   0,   0,   5,   1,   3,   1,   0,  -2,   1,   3,   1,   0,   1,   1,   0,  -1,   0,   2 },
     {    1,  -1,   0,   3,   3, -20,   2,   0,   2,   0,   1,   0,  -2,  -7,  -5,  -4,  -9, -82,  24,   0,   9,   0,   3,   0,   1,   0,   2,   1,   7,   4, -17,   3,  -2,   1,  -1,   1,   6,   7,   8,  12,  60,  22, -16,  -5,  -5,  -2,  -2,   0,   6,  10,  11,  41,  -5,   4,   0,  -4,   8,   2,  17, -19, -10,  10,  -1,   1,   0,  -2, -16,  -2,  -2,   1,   3,   1,  -3,  -2,   2,  -2,  -2,   4,   0,   0,  -2,  -1,  -1,   0,   0,   0,   0,  -1,  -1,   0,  -1,  -1,   0,   0,   0,   0 },
     {   -1,   0,  -1,  -3,  -3,  -4,  -2,   0,   1,   0,   1,   0,  -6,  -7,  -9,  -8, -44, -56,  26,   3,   7,   1,   3,   0,  -3,   0,  -4,  -9, -15,  29, -15,   6,  -2,   2,  -1,   1,  -3,  -2,  -3, -10, -43,  30,   3,   1,   1,   1,   1,   0,  -2, -12,  -6, -55,  26, -15,   8,   2,  -8,   4, -24,  23,  13,   4,  -5,  -1,   2,   1,  17,   5,  -1,  -2,   0,  -1,   1,   1,   1,   4,   5,   1,  -2,   0,   0,   0,   2,   2,   1,   1,   0,   1,   0,   0,   1,   1,   0,   0,   0,   0 },
     {   -1,   6,   4,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,  -2,   6,   7,   0,  -6,  -1,   0,   0,   1,   0,   0,   0,  -1,   8,   8,   0,   2,   5,  -2,  -1,  -1,   0,   0,   0,  -2,  12,  11,   1, -11,  -7,   3,  -1,   0,   0,   0,   0,  -4,  10,  16,  -8,  -6,  -5,  -2,   0,  -9,  41,  37, -12,  -8,  -2,  -3,  -3,  -5,  85,  17, -18,   0,  -8,   0,  -5,   4, -56,   0,  18,   2,   5,   0,   2,  -1,   0, -11,   0,   0,  -8,  -1,   1,   0,   2,  -2,  -1,   1,  -6,  -1,   1 },
     {    1,   2,   2,   0,  10,  33, -21,   1,  -5,   1,  -2,   1,   0,   0,  -1,  -1,  -1,   8,   6,  -3,   0,   0,   0,   0,   5,   8,   7,   5,  26,  90, -16,  -7,  -8,  -2,  -2,  -1,   3,   2,   5,   7,  25, -31,  20,  -3,   1,  -2,   0,  -1,  -4,  -3,  -8,  -7, -39, -31,   7,   7,  -3,  -3, -10, -10,   3,   5,  -3,   0,   3,  -5,  -3,   4,  -1,  -6,   2,   0,   2,   3,   2,  -1,  -2,  -1,  -1,   0,  -1,   0,   0,   1,   1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0 },
   },
   {
     {   96, -54,  -5,  -2,  -1,  -1,  -1,   0,   0,   0,   0,   0, -59,  18,  13,   1,   2,   0,   1,   0,   0,   0,   0,   0,  -6,  13,  -3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   57,  31, -40,  -3,  -5,  -1,  -2,   0,  -1,   0,  -1,   0,  56, -51,  -5,   8,  -1,   1,   0,   1,   0,   0,   0,   0, -60,  -1,  28,   0,   3,   0,   1,   0,   1,   0,   0,   0,  -9,  15,   0,  -4,   0,  -1,   0,   0,   0,   0,   0,   0,  -4,   1,  -1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0 },
     {    6,  85, -48, -12,  -4,  -2,  -2,  -1,  -1,   0,  -1,   0, -67, -10,  17,  11,   5,   2,   2,   1,   1,   0,   1,   0,  25, -16,   6,   2,  -2,   0,  -1,   0,   0,   0,   0,   0,  22, -10,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   4,   1,   0,   0,   0,   0,   2,  -1,   0,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 },
     {  -38, -25,   0,  17,   2,   4,   1,   2,   1,   1,   0,   1, -29, -55,  47,   7,   2,   1,   1,   1,   1,   0,   0,   0, -31,  59,  12, -21,  -1,  -4,   0,  -2,   0,  -1,   0,  -1,  39,  14, -29,  -4,   0,  -1,   0,   0,   0,   0,   0,   0,   5,  -6,  -3,   4,   0,   0,   0,   0,   6,   1,  -2,   0,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   3,   1,  -2,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,  -1,   0 },
     {   -8,  19,  42, -26,  -1,  -5,   0,  -3,   0,  -1,   0,  -1, -24,  40, -29,  -8,   5,   0,   2,   0,   1,   0,   0,   0, -80, -16,  14,  21,   5,   2,   2,   1,   1,   1,   1,   1,  48,   6,  -4,  -3,  -6,  -1,  -2,  -1,  -1,   0,  -1,   0,  19, -12, -10,   1,   0,   1,   0,   1,   4,  -4,   3,   2,   0,   0,   0,   0,   2,  -2,  -1,   0,   0,   0,   0,   0,   3,  -1,   1,   0,   0,   0,   0,   0,   2,  -1,  -1,   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0 },
     {   20,  33,   6,  -8,  -7,  -2,  -3,  -1,  -2,  -1,  -1,  -1,  33,  29,  22, -32,  -4,  -4,  -1,  -2,  -1,  -1,   0,  -1,  -8,  54, -54, -12,   9,  -1,   2,  -1,   1,   0,   1,   0,  18, -51, -21,  28,   3,   3,   0,   2,   0,   1,   0,   1, -16, -16,  19,   6,  -2,   1,   0,   1,  -2,  -2,   2,   0,   0,   0,   0,   0,  -4,  -2,   3,   1,   0,   0,   0,   0,  -1,  -2,   0,   1,   0,   0,   0,   0,  -2,  -2,   2,   1,   0,  -1,   0,   0,  -1,  -1,   1,   0,   0,  -1,   0,   0 },
     {   24,  25,  82, -52, -11,  -8,  -2,  -4,  -1,  -2,  -1,  -1,  -5, -54,   7,   5,   9,   4,   2,   2,   1,   1,   1,   1,  26,   2, -22,   6,   3,  -1,   0,   0,   0,   0,   0,   0, -16,  26, -11,  -4,   1,  -2,   1,  -1,   0,   0,   0,   0,  -7,   1,  -7,   6,   2,   0,   0,   0,  -1,   3,  -2,   1,  -1,   0,   0,   0,  -1,  -1,  -2,   2,   0,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   1,   0,   0 },
     {    7, -17,  -7, -15,   9,   0,   3,   0,   2,   0,   1,   0,  11, -17, -68,  34,  10,   3,   2,   2,   1,   1,   1,   1,  31,  24,  11,  13, -16,  -3,  -4,  -1,  -2,  -1,  -1,  -1,  61, -17,  -2, -12,  -8,   2,  -2,   0,  -1,   0,  -1,   0, -44, -17,  20,   3,   4,   2,   1,   1,  -7,  12,   9,  -5,   0,  -1,   0,  -1,  -7,   1,   0,  -1,   1,   0,   0,   0,  -1,   2,   3,  -2,   0,   0,   0,   0,  -3,   0,   1,   0,  -1,   1,   1,  -1,  -2,   0,   1,   0,   0,   1,   1,  -1 },
     {   11,  15,   5,  -6,   2,  -4,   0,  -2,   0,  -1,   0,  -1,  17,  45,   0,  11, -15,  -3,  -4,  -2,  -2,  -1,  -1,  -1,  26,  20,  47, -41, -12,  -1,  -2,  -1,  -1,  -1,  -1,  -1,  -8,  34, -52, -22,  17,   0,   4,   0,   2,   0,   1,   0,  10, -39, -23,  27,   5,   1,   1,   1, -12, -10,  13,   4,  -2,   1,   0,   0,  -1,  -4,   0,   1,   0,   1,   0,   0,  -4,  -2,   1,   1,   1,   0,   0,   0,   0,  -3,  -1,   2,  -2,  -1,   1,   0,   0,  -2,   0,   1,  -1,  -1,   0,   0 },
     {  -15,   6, -13, -33,  17,  -1,   5,   0,   2,   0,   1,   0, -24,  -1, -44,  25,  14,   0,   2,   1,   1,   0,   1,   0, -17,  62,   1,  -1, -10,  -6,  -1,  -2,  -1,  -1,   0,  -1, -65, -18,  17,   9,   1,   4,   2,   1,   1,   1,   1,   1,  38,  -7,   6,   4,  -6,   1,  -2,   0,  15,  -6,  -1,  -4,  -2,   0,  -1,   0,   6,  -4,   2,   1,   0,   0,   0,   0,   2,  -1,   2,  -1,  -1,   0,   0,   0,   3,  -2,   1,   0,   1,  -1,   1,  -1,   2,  -1,   0,   0,   1,   0,   1,   0 },
     {  -14, -30, -40, -83,  45,   8,  13,   3,   6,   1,   3,   1,  13,   3,  30, -17,   8,  -6,  -2,  -2,  -1,  -1,   0,  -1,   2, -21,  -3,  27,   0,  -3,  -1,   0,   0,   0,   0,   0,   4,  12, -30,  19,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   7,  10,  -8,   0,  -1,   0,  -7,   2,   5,   0,  -2,   1,   0,   0,   1,   2,   3,   1,  -2,   0,  -1,   0,  -2,   1,   0,   1,  -1,   0,   0,   0,   1,   1,   1,   1,  -1,   0,   0,   0,   0,   1,   1,   1,   0,   0,   0,   0 },
     {   -4,  -2,  25, -20,   7,  -4,   2,  -2,   1,  -1,   0,  -1,  -2,   8,  26,  34, -20,  -3,  -5,  -1,  -2,  -1,  -1,   0, -17, -19,  47, -25, -17,   7,  -1,   2,  -1,   1,   0,   1, -15, -64, -11,  -3,  21,   8,   3,   3,   2,   2,   1,   1, -34,  33,  19,   1,   5,  -5,   0,  -2,  21,  21, -20, -10,   0,  -2,   1,  -1,   2,  -1,  -8,   4,   1,   0,   0,   0,   5,   2,  -4,  -1,   0,   0,   0,   0,   0,   0,  -2,   1,   2,   1,  -2,  -1,   0,   0,  -1,   1,   1,   1,  -1,   0 },
     {    4,  11,   0, -19,   8,   1,   0,   0,   0,   0,   0,   0,  17,  22,  33,  -4,   6, -10,  -2,  -4,  -1,  -2,  -1,  -1,  12,  41,  18,  32, -28,  -7,  -5,  -2,  -2,  -1,  -1,  -1,  13,   1,  44, -42, -20,   7,  -2,   1,  -1,   1,   0,   0, -10,  26, -43, -30,  20,   1,   3,   0,   4, -22, -21,  20,   6,  -1,   1,   1,  -5,  -5,   2,   3,   0,   0,   0,   0,   0,  -6,  -3,   3,   0,   1,   0,   0,  -2,  -2,  -1,   0,   0,  -3,  -2,   2,  -1,  -1,   0,   0,   0,  -2,  -1,   1 },
     {   11,  -1,   8,   1,  16,  -6,   2,  -3,   0,  -1,   0,  -1,  19,  -4,  16,  64, -24,  -7,  -6,  -2,  -3,  -1,  -2,   0,   6, -11, -26,  -2, -15,   9,   1,   3,   1,   1,   0,   1,  28, -19,   0,   2,   0,   5,  -2,   1,  -1,   0,   0,   0,  74,   7,   0, -17, -10,  -1,  -3,   0, -41, -10,   1,  -7,   8,   2,   2,   1,  -9,   7,   6,  -1,   2,  -1,   0,   0,  -8,   2,  -1,  -4,   2,   0,   1,   0,   0,   2,   1,  -1,  -4,   0,   0,  -2,  -1,   1,   0,   0,  -2,   0,   0,  -1 },
     {    2,  -8,  10,   6, -32,  11,  -4,   4,  -2,   2,  -1,   1,   3, -17,  14, -30,  25,   2,   1,   0,   0,   0,   0,   0,  12,  -5,  43,  43, -10, -15,  -5,  -4,  -2,  -2,  -1,  -1, -11, -53,  -9,  -1, -15,  11,   3,   4,   2,   2,   1,   1,  33, -38, -10,   5,  10,   5,  -1,   1, -28,  26,  22,  -7,  -2,  -4,   0,  -1,   1,  16, -10,  -8,   4,  -1,   1,   0,  -3,   4,  -2,   1,   0,  -1,   0,   0,   1,   3,  -4,  -1,  -2,   2,   0,   0,   1,   2,  -2,  -1,  -1,   1,   0,   0 },
     {    4,  -5,  -4, -19, -10,   6,   1,   2,   0,   1,   0,   1,   1, -17, -33, -45,   9,  15,   5,   5,   2,   2,   1,   1,  19,   3,  27, -42,  19,   2,  -2,  -1,  -1,   0,  -1,   0,  26, -24,  -2,  20,  19, -10,  -1,  -1,   0,  -1,   0,   0,  33,  56, -29,   4,  -4, -11,   0,  -3, -11, -21, -12,  14,  -1,   1,   1,   1, -15,  -3,  11,   9,  -2,   0,   0,   0,  -2,   0,   2,   2,  -2,   0,   0,   0,  -3,   1,   2,   2,  -2,   0,   1,   1,  -2,   0,   1,   1,   0,   0,   0,   1 },
     {  -11, -15, -21, -19, -55,  30,   3,   8,   1,   4,   0,   2,  -3,  -3,  -6,  22,  -5,   0,  -2,  -1,   0,   0,   0,   0, -16, -12, -44,   5,  -1,   6,   2,   3,   1,   1,   1,   1,  -6, -19, -17, -48,  29,  11,   3,   3,   1,   1,   1,   1,  -6,   8, -26,  28,  28, -11,   0,  -2,   7, -17,  23,  21, -12,  -3,  -1,   0,  -1,   8,  13,  -5,  -2,  -1,  -1,  -1,   3,   0,   3,  -1,  -1,   0,   0,   0,   0,   3,   3,  -1,   1,   0,   2,   0,   0,   2,   2,   0,   1,   0,   1,   0 },
     {    4,   4,   6,  21,  73, -34,  -3,  -9,   0,  -4,   0,  -2,  -9, -23,  -7, -27,   5,  -8,   8,   1,   3,   1,   2,   0, -10, -15, -19, -15, -36,  20,   5,   5,   1,   2,   1,   1,  -4, -24,   8, -15,  12,  11,  -2,   1,   0,   1,   0,   0,  -7, -13, -34,  25,  20,  -4,   1,   0,  -4,   0,  24,  20, -11,  -3,  -2,  -1,   2,  13,   6,  -9,  -6,   1,   0,   0,   2,   2,   0,   2,  -1,   0,  -1,   0,   1,   3,   1,  -1,   1,   1,   1,   1,   0,   2,   0,  -1,   0,   0,   0,   1 },
     {   -8,  -8,  -6, -36, -11,   8,   6,   2,   2,   1,   1,   1,  -6,  -6,   7,  -8, -44,  17,   0,   4,   0,   2,   0,   1,  -5,   0,   2, -41,  20,  15,  -1,   2,   0,   1,   0,   0,   4,  -2,  65,  -1,   6,  -9,  -8,  -2,  -3,  -1,  -1,  -1, -10, -51,  -8,  -7,   9,   2,   3,   3, -33,  21,  -4,   9,  16,  -4,   2,  -2,  15,  14,  -9,   5,  -3,  -4,   0,  -1,  -1,   0,  -5,   4,   2,  -1,   0,   0,   4,   1,  -1,   3,  -1,   1,  -2,   1,   2,   1,  -1,   1,  -1,   0,  -1,   1 },
     {   -8,   7,  -1,   3,  -1,  -5,   1,  -1,   0,   0,   0,   0, -12,   6,  -9,  -3,  18,  -6,   1,  -1,   1,   0,   0,   0, -16,   9,  -3,  -7,  -3,  -7,   4,  -1,   1,   0,   1,   0, -17,   7, -22, -12,  -5,   7,   4,   2,   2,   1,   1,   0, -27,  38,  18, -21,   5,   0,   0,  -2, -90,  -3,  12,   9,  11,   0,   3,   0,  46,   6,  -8,   7,  -4,  -1,  -2,  -1,   8, -10,  -2,   2,  -3,   1,   0,   0,  11,  -2,   0,   1,   1,  -3,   0,   1,   4,  -1,   0,   1,  -1,  -1,   1,   0 },
     {    2,   0,  -7,  -8,  -9,   5,   0,   3,   0,   1,   0,   1,   4,   2,  -3, -34,  18,  -8,   4,   0,   2,   0,   1,   0,   1,  -2, -15, -21, -42,  14,   8,   4,   2,   2,   1,   1,  11,  20,   8, -42,  11,  25,  -5,   1,  -2,   1,  -1,   0,  34,   5,  49,   7,  19, -12, -10,  -3,   4,  39, -28, -19,   0,  -8,   3,  -1,  -7, -21, -24,  18,  10,  -1,   2,   0,  -6,  -3,   6,   9,  -3,  -1,   0,   0,  -1,  -3,  -1,   4,  -2,  -1,   1,   3,  -1,  -2,  -1,   2,  -1,   0,   1,   1 },
     {    1,   1,   6,  -3,  -1,  21,  -6,   3,  -3,   1,  -1,   1,   9,  11,  19,  27,  80, -33, -12,  -8,  -3,  -4,  -1,  -2,  -3,  -9,   1, -41,   7, -14,   8,   2,   3,   1,   1,   0,  11,   2,  41,   9,  -7,  -6,   1,  -3,  -1,  -1,  -1,  -1,  -7,  -4,  -6,  34, -18,  -8,   1,  -1,  -6,  -4,   9, -11, -18,   9,   2,   1,   0,  13,  -6,  -7,   1,   1,   0,   0,   1,  -3,  -6,  -1,  -3,   2,   1,   1,   0,   2,  -2,   1,   0,  -1,  -2,  -1,   0,   1,  -1,   0,   0,  -1,  -1,   0 },
     {    1,   7,   4,   2,  -4,  -2,   0,   0,   0,   0,   0,   0,   0,   5,   9,  -8, -28,   8,   7,   0,   2,   0,   1,   0,   4,  24,  13,  35, -13,  19, -10,  -2,  -4,  -1,  -2,   0,   8,  15,  29,   7,  43, -18, -12,  -3,  -3,  -2,  -1,  -1,  -1,  29,  14,  36, -21, -27,   1,  -3, -13,  -5,  37, -31, -33,  10,   2,   2,   6,   9, -24, -26,  12,   5,   2,   1,   3, -10,  -9,   5,   5,   0,   0,   0,   0,  -1,  -4,  -2,   0,  -4,  -2,   1,   0,   0,  -2,  -2,   0,  -2,   0,   1 },
     {    6,   9,  -1,  10,  32,  27, -13,  -3,  -4,  -1,  -1,   0,   4,   7,  -3,  15,   0,  14, -13,   2,  -2,   1,  -1,   0,   4,   9, -13,  10,  40, -24,  -4,   0,   0,  -1,   0,   0,   5,  -2,  -6, -10,  13, -19,   5,   0,   1,   0,   0,   0,  11,  17, -35,   2,  10, -10,   7,  -1,   3,  75,   0,   5, -15, -12,   1,  -5,  22, -29, -18,   0, -13,   6,   1,   2,  -9, -14,  11,   4,  -2,   2,   0,   1,   1,  -4,   3,  -2,  -3,  -2,   3,   0,   1,  -1,   1,  -1,  -2,   0,   2,   0 },
     {   -4,  -7, -10,  -5, -11, -86,  34,   0,  10,  -1,   4,  -1,   6,   6,  10,   7,  18,   7,   8,  -6,  -2,  -2,  -1,  -1,   4,   4,   1,   0,  57,  13, -18,  -4,  -3,  -1,  -1,   0,   1,  -9,   3, -40, -15,  -5,   4,   5,   2,   2,   1,   1,  10,   7,  11,  15,  -4,   5,  -1,  -2,   2,  10,   3,   4, -10,  -2,  -2,  -1,  -7,  -1,  -7,  -4,  -2,   6,   1,   1,  -1,   0,  -1,   4,  -1,   0,  -1,   0,  -1,  -1,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   1 },
     {    4,   5,  12,   6,  40,  34, -22,  -3,  -6,  -1,  -2,   0,  -2,  -3,   5, -11,  -3,  33, -16,   5,  -3,   2,  -1,   1,  -1,  -2,   7,   3,  32, -30,  -7,   2,  -1,   0,  -1,   0,  -4, -10,   6, -53,   5, -13,  12,   3,   2,   2,   1,   1,   8,  -8,  54,   8,  -6,  -2,   0,  -3,  -5, -39,  -3,  17,  -2,  -4,   1,   1, -19,  17,   2,  -3,  -3,  -1,   3,  -1,   7,  10,  -8,   1,  -1,  -2,   1,  -1,  -2,   2,  -3,   1,   2,   1,  -2,   2,  -1,   0,  -1,   1,   1,   0,  -1,   1 },
     {   -2,  -2,   1,  13,   1,  -6,  -1,   1,  -1,   0,  -1,   0,  -1,  -2,   4,   6,  26,  -7,   3,  -4,   0,  -1,   0,  -1,  -2,   1,   3,  31,   5,  26, -14,  -3,  -3,  -1,  -1,   0,  -5,  -6,  -2,   0,  54, -22, -17,   2,  -2,   1,  -1,   0,  -8, -32,  -4,   0, -13, -32,  12,   4, -25,  -7, -54,   7,  -2,  11,  13,   0,   0, -19,  26,  33,  -8,   1,  -3,  -1,   5,  16,  17, -16, -14,   2,  -1,   0,   3,   2,  -2,  -4,   1,   3,   2,  -3,   1,   0,  -1,  -1,   0,   2,   0,  -2 },
     {   -7,   3,  -2,  -1,   2,   3,  -1,   0,   0,   0,   0,   0,  -9,   6,   0,  -1,  -6,  -9,   2,   0,   1,   0,   0,   0, -10,   5,  -4,  -2,   3,  -7,   5,  -1,   1,   0,   0,   0, -14,   7,   2,  -3,  -1,  -2,   3,  -1,   0,   0,   0,   0, -13,   8, -12,  -7,  -1,   4,   3,   1, -32,  33,  19, -18,   1,   1,   0,  -1, -92, -14,  22,   6,  10,   3,   2,   1,  48,  13, -10,   0,  -3,  -1,  -2,  -1,   6,  -9,  -5,   2,   8,  -1,   1,   1,  -3,  -3,   1,   0,   7,   1,  -1,   0 },
     {    0,   2,   5,   3,  -6,  -6,  11,  -2,   2,  -1,   1,  -1,   2,   8,   7,  11,   4,  18,  -8,  -1,  -3,   0,  -1,   0,   2,   3,  13,   3,  -3,  -6,   3,  -2,   1,  -1,   0,  -1,   2,  17,  11,  38,  -6,  22, -12,  -2,  -4,  -1,  -2,  -1,   5,   2,  29,   9,  52, -27, -16,  -1,  -9,   8,  -6,  43, -35, -29,   9,  -1,  -5, -10,  33, -28, -29,  16,   2,   2,   3,  10, -16, -16,   9,   1,   0,   0,   3,  -5,  -2,  -1,  -1,   2,  -3,  -3,   1,  -3,   0,   0,   0,   1,  -2,  -1 },
     {    1,   0,   3,  -1,   4,   4, -15,   4,  -3,   1,  -1,   1,  -4,  -8,  -7,  -6, -28, -90,  30,  11,   8,   3,   2,   0,   9,   5,  14,  10,  40,  -9,  31, -12,  -1,  -4,   0,  -2,   3,  -2,   5, -10,  14,  31, -10,  -8,  -1,  -1,   0,   0,   5,   0,  12,   2, -10,   1,  -7,   1,   1,   2,  -1,  21,  -4,   1,  -2,  -2,   2,  -2,   1,  -8, -12,   3,   0,   1,  -6,   3,  -2,  -3,   4,   6,  -2,   0,   0,  -1,  -2,   0,  -1,   1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0 },
     {   -2,  -4,  -2,   2,  -4, -12,   5,  -1,   1,  -1,   1,   0,  -4,  -7,   0,  -3, -31,   6,  -7,   6,  -1,   2,   0,   1,  -5,  -7,  -4,   4, -10, -47,  15,   3,   3,   1,   1,   0,   1,  -8,  14,   1, -27,  11,  22,  -5,   1,  -1,   1,  -1,  -8,   0, -11,  58,  -5,   9,  -6,  -7, -25, -18, -30, -19, -14,  13,   3,   4,  11, -50,   7,   8,  15,  17,  -4,   2,  -9,  18,  23,  -3,   3,  -5,  -3,  -1,   5,   5,  -3,  -7,  -2,   3,   2,   0,   2,   0,  -2,  -2,  -1,   2,   1,   0 },
     {    0,  -1,   5,  -3,   6,  21,  -3,  -1,  -2,   0,  -1,   0,   0,  -3,   9,   1,   7,  -1,   4,  -4,   0,  -1,   0,  -1,   0,  -2,   8, -10,   9,  40, -14,  -3,  -2,   0,  -1,   0,  -4,  -9,   6,  -5, -21,   1, -16,   7,   0,   2,   0,   1,  -3,   0,  10, -18,  -3,  -6,   4,   5,   2, -18,  58,   8,  -3,  -1,  -6,  -2,   2, -76,  -7,  12,   1,   0,   2,   4, -23,  34,  15,  -6,   3,  -4,   2,  -2,   8,  12, -14,  -4,  -3,   1,  -2,   2,   4,   0,  -3,   0,  -2,   4,   0,   0 },
   },
   {
     { -109,  46,  19,  -2,   2,   0,   1,   0,   1,   0,   0,   0,  37, -17,  -5,   1,  -1,   0,  -1,   0,   0,   0,   0,   0,  18,  -7,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -36, -95,  60,  19,   4,   3,   2,   2,   1,   1,   1,   1, -15,  26,  -9,  -4,   0,   0,   0,   0,   0,   0,   0,   0,  18,  19, -17,  -4,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   2,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   4,   4,  -3,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
     {   18, -30,   2,  10,  -1,   2,   0,   1,   0,   0,   0,   0, 102, -26, -31,   2,  -4,   0,  -2,   0,  -1,   0,  -1,   0, -44,  21,   8,  -3,   2,   0,   1,   0,   0,   0,   0,   0, -17,   5,   6,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  -8,   3,   2,   0,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   27,  32,  65, -56,  -7,  -6,  -1,  -4,  -1,  -2,   0,  -2,  19,  60, -37, -12,  -1,  -2,   0,  -1,   0,  -1,   0,  -1,   2, -25, -14,  19,   2,   2,   0,   2,   0,   1,   0,   1, -11, -12,   6,   6,   0,   1,   0,   1,   0,   0,   0,   0,  -3,  -6,  -1,   4,   1,   0,   0,   0,  -3,  -3,   1,   1,   0,   0,   0,   0,  -1,  -2,  -1,   1,   0,   0,   0,   0,  -1,  -2,   1,   1,   0,   0,   0,   0,  -1,  -1,   0,   1,  -1,  -1,   0,   0,   0,  -1,   0,   1,  -1,  -1,   0,   0 },
     {  -20,  -7, -60,  39,   8,   3,   1,   3,   1,   1,   0,   1,  14,  84, -28, -30,   0,  -5,   0,  -3,   0,  -1,   0,  -1,   1, -17,  30,  -8,  -3,   0,   0,   0,   0,   0,   0,   0,  -1, -19,   9,   5,  -1,   1,   0,   1,   0,   0,   0,   0,   1,  -4,   5,   0,   0,   0,   0,   0,   0,  -5,   2,   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0 },
     {   21,  -7, -21,   9,   2,   1,   0,   0,   0,   0,   0,   0,  25, -14, -14,   8,   0,   1,  -1,   1,   0,   0,   0,   0, 102, -18, -29,   0,  -5,   0,  -3,   0,  -1,   0,  -1,   0, -43,  14,  12,  -3,   2,   0,   1,   0,   0,   0,   0,   0, -11,   3,   6,  -2,   0,   0,   0,   0, -12,   3,   4,  -1,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,  -6,   1,   2,   0,   0,   0,   0,   0,  -2,   0,   1,   0,  -3,   1,   1,   0,  -1,   0,   1,   0,  -2,   1,   1,   0 },
     {    8,   3,  10,  12, -12,   1,  -3,   0,  -2,   0,  -1,   0,  36,  21,  96, -53, -23,  -2,  -4,  -3,  -3,  -1,  -1,  -1,   5,   1, -22,   6,   4,   1,   1,   0,   0,   0,   0,   0,  -9,  -8, -22,  13,   7,   0,   1,   1,   1,   0,   0,   0,  -6,   0,  -4,   2,   2,   0,   0,   0,  -3,  -2,  -6,   3,   2,   0,   0,   0,  -2,   0,  -1,   0,   1,   0,   0,   0,  -1,  -1,  -3,   2,   1,   0,   0,   0,  -1,   0,  -1,   0,  -1,  -1,  -1,   1,  -1,   0,  -1,   0,  -1,   0,  -1,   1 },
     {   11,  38,   9,  31, -29,  -3,  -6,  -1,  -3,  -1,  -2,   0,  -6,  12, -16,  -2,   8,  -2,   2,   0,   1,   0,   0,   0,  11,  89, -34, -40,   9,  -5,   2,  -3,   0,  -1,   0,  -1,   1, -28,  15,   4,  -1,   2,   0,   1,   0,   0,   0,   0,  -2, -15,   3,   3,   1,   1,   0,   1,   0,  -7,   3,   1,   0,   0,   0,   0,  -1,  -5,   1,   1,   0,   0,   0,   0,   0,  -4,   2,   1,   0,   0,   0,   0,  -1,  -3,   1,   0,   0,  -2,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0 },
     {  -14, -18, -36, -81,  56,   5,  12,   0,   5,   0,   3,   0,   3,   7,  11, -10,  -5,   1,  -1,   0,  -1,   0,   0,   0,  10,  58,   1,   6, -18,  -5,  -3,  -2,  -2,  -1,  -1,  -1,  -9, -13,   4,   9,   0,   0,   0,   1,   0,   0,   0,   0,   1,  -6,   3,   8,  -3,   0,  -1,   0,  -1,  -3,   0,   2,   0,   0,   0,   0,   1,  -2,   1,   3,  -1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   2,   0,  -1,   0,   1,   0,  -1,   0,   1,   0,  -1,   0,   1 },
     {  -10,   6,  12,   5,  -5,  -3,   0,  -1,   0,   0,   0,   0, -22,   1,  14,  -1,  -2,   1,   0,   0,   0,   0,   0,   0, -23,   4,  28, -12,  -1,   1,   0,   0,   0,   0,   0,   0,-104,  14,  33,  -1,   5,   0,   3,   0,   1,   0,   1,   0,  32,  -8, -12,   2,  -1,   0,  -1,   0,   3,   0,  -4,   1,   0,   0,   0,   0,  17,  -3,  -7,   1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   7,  -1,  -3,   0,   1,   0,  -1,   0,   3,  -1,  -1,   0,   0,   0,  -1,   0 },
     {   13,   7,  29,  -1,  -2,  -8,  -1,  -2,  -1,  -1,   0,  -1,  12,   7,  13,  26, -17,  -1,  -4,   0,  -2,   0,  -1,   0,  37,  14,  91, -40, -29,   2,  -5,  -2,  -3,  -1,  -2,  -1,  14,  -4, -26,   1,   4,   1,   1,   1,   1,   0,   0,   0, -10,  -4, -18,   6,   6,   0,   2,   1,  -5,   0,  -5,   0,   3,   0,   1,   0,  -5,  -1,  -5,   2,   2,   0,   0,   0,  -2,  -1,  -3,   0,   2,   0,   0,   0,  -2,  -1,  -3,   1,  -1,   0,  -1,   0,  -1,   0,  -2,   0,  -1,   0,  -1,   0 },
     {   -4,   2, -10,   5,  10,  -4,   2,  -1,   1,   0,   0,   0,  11,  41,  24,  93, -47, -14, -10,  -4,  -4,  -2,  -2,  -1, -12,   2, -28,   7,  12,  -1,   2,   0,   1,   0,   1,   0,  -4,   1,  -6, -30,  12,   4,   2,   1,   1,   1,   1,   0,   6,  -5,   5,  -6,  -1,   1,   0,   0,  -1,  -4,  -1,  -8,   3,   2,   1,   0,   2,  -2,   2,  -1,  -1,   0,   0,   0,   0,  -2,   0,  -4,   2,   1,   0,   0,   1,  -1,   1,  -1,   0,  -1,   0,  -2,   0,  -1,   0,  -1,   0,  -1,   0,  -2 },
     {   -8, -20, -13, -23, -62,  44,  -1,  10,  -1,   5,   0,   3,  -1, -14,   3,  19, -13,   1,   0,   0,   0,   0,   0,   0,   6,  -9,  14,  19,  12, -12,  -2,  -2,   0,  -1,   0,  -1,  -7, -79,  17,  22,   5,   4,   0,   3,   0,   1,   0,   1,  -1,  14,  -5,  -2,   5,  -3,   0,  -1,   0,   6,  -2,  -4,   2,  -1,   0,   0,   0,  11,  -3,  -3,   2,  -2,   0,  -1,   0,   3,  -1,  -2,   1,   0,   0,   0,   0,   4,  -1,  -1,   0,   2,   0,  -1,   0,   2,   0,  -1,   0,   1,   0,  -1 },
     {  -10,  -5, -17, -30, -64,  45,   2,   9,  -1,   5,   0,   3,   0,  17,  -6, -10,  -7,   3,   1,   0,   0,   0,   0,   0,   5,  20,  11,   4,  18, -16,  -1,  -4,   0,  -2,   0,  -1,   7,  76, -15, -24,   5,  -8,   0,  -3,   0,  -2,   0,  -1,  -1, -13,   4,   4,   5,  -2,   0,   0,   0,  -7,   2,   3,   1,   0,   0,   0,   0, -10,   2,   4,   2,   0,   0,   0,   0,  -4,   0,   2,   1,   0,   0,   0,   0,  -4,   1,   2,   0,  -2,   0,   1,   0,  -2,   0,   1,   0,  -1,   0,   1 },
     {   -8,   2,   4,   2,  -2,   1,  -2,   0,   0,   0,   0,   0,  -6,  10,  12,  16,  55, -29,  -2,  -6,   0,  -3,   0,  -2,  -7,   9,  10,  19,  -2,  -4,  -2,  -1,  -1,   0,  -1,   0, -31,  -2, -21,  -6,  -3,   8,   4,   2,   1,   1,   1,   1, -86,  11,  29,  -9,   3,   2,   3,   0,  29,  -6,  -6,  -1,  -6,   2,   0,   0,   1,   0,   1,  -1,  -1,   0,   0,   0,  12,  -2,  -3,  -1,  -3,   1,   0,   0,   1,   0,   0,  -1,   6,  -1,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1,   0 },
     {   -3,  -1,   7,  -1,   3,  -3,   1,  -1,   0,   0,   0,   0, -10,  -5,   3, -14, -62,  27,   4,   5,   0,   3,   0,   1,  -5,  -7,   4, -14,  -6,   5,   2,   1,   0,   1,   0,   0,  12,  10,  73,  -9,  -7,  -7,  -5,  -3,  -3,  -1,  -2,  -1, -62,  -3,  23,   7,   3,   0,   1,   0,  19,   0, -12,   3,   6,  -3,  -1,   0,  -3,   0,  -7,   2,   3,   0,   0,   0,   7,   0,  -6,   1,   4,  -1,   0,   0,  -1,   0,  -3,   1,   3,   0,  -3,   0,  -1,   0,  -1,   0,   2,   0,  -1,   0 },
     {    2,  -9,   1, -19,  -1,  11,   1,   2,   0,   1,   0,   1,  18,   9,  35,  13,  55, -34,  -5,  -7,  -1,  -3,  -1,  -2,   4, -18,   0, -42,  21,   7,   0,   2,   1,   1,   0,   1,  24,   4,  59, -21, -40,  11,  -3,   1,  -3,   1,  -1,   0,  14,  -2,  -9,  15,  -7,   0,   0,   0,  -8,   2,  -9,   1,  -2,   2,   1,   1,  -3,   2,  -6,   6,   0,  -1,   0,   0,  -4,   1,  -4,   1,  -2,   1,   0,   0,  -2,   1,  -2,   2,  -2,   0,  -2,   1,  -1,   0,  -1,   1,  -1,   0,  -1,   0 },
     {   -7, -10, -10, -26,   0,   1,   7,   1,   2,   0,   1,   0,  -6,  -5, -11,  -2, -12,   6,   0,   2,   0,   1,   0,   1, -12, -30, -21, -88,  38,  16,   5,   5,   3,   3,   1,   2, -18,  -8, -44,  18,  15,  -1,   2,   1,   1,   0,   1,   0, -18,   1,  14,  18,  -3,  -3,  -1,  -1,   5,   4,   3,   2,  -2,  -1,  -1,   0,   3,   3,   7,   6,  -4,  -1,  -1,   0,   3,   2,   2,   2,  -1,  -1,   0,   0,   2,   1,   3,   3,   1,   1,   1,   1,   1,   1,   1,   2,   1,   1,   1,   1 },
     {    1,   9,   0,  -7,  -1,   2,  -1,   1,   0,   0,   0,   0,  -1,   9,  -1,  -5,  -6,   1,   3,   0,   0,   0,   0,   0,   0,   9,  -2, -17,  -4,   4,   1,   0,   0,   0,   0,   0,   0,   8,   6, -12,   3,  -1,   0,   0,   0,   0,   0,   0,   8, 114,  -8, -35,  -1,  -9,   0,  -5,  -1, -26,   4,   8,   0,   1,   0,   1,   1,  -4,  -1,   3,   0,   0,   0,   0,  -1, -12,   1,   5,   1,   0,   0,   0,   0,  -2,   0,   2,   0,  -5,   0,   2,   0,  -1,   0,   1,   0,  -2,   0,   1 },
     {    8,  14,  16,  15,  48,  87, -56,   2, -15,   1,  -6,   1,   0,   4,   1,   8,   5,  20, -10,   0,  -2,   0,  -1,   0,  -3,  -6,  -5, -11,  -8, -26,  17,  -1,   5,   0,   2,   0,  -2,   4,  -2,  14,  -3, -12,   4,  -1,   1,   0,   0,   0,  -1,   0,   0,   2,  -5,  -7,   4,   0,   4,   1,  -1,  -2,  -1,  -2,   1,   0,   0,  -1,   0,  -2,  -2,  -2,   1,   0,   1,   0,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -5,  -3, -10,  -5, -28,  13,  -2,   5,   0,   2,   0,   1,   0,   4,  -1,   8,   4,  -8,   1,  -1,   0,  -1,   0,   0, -16, -10, -32, -16, -99,  43,  14,   6,   3,   3,   2,   1,  -3,   3,  -5,   3, -13,  -1,   4,  -1,   1,   0,   1,   0,   0,  -5, -10,  13,  24, -10,   0,  -1,   4,  -4,   4,  -2,   3,  -1,   0,   0,   0,   1,   4,   1,  10,  -4,  -1,  -1,   1,   0,   3,  -1,   3,  -1,  -1,   0,   0,   1,   2,   0,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   1,   0 },
     {   -2,   1,  -4,   3,  -7, -13,  11,  -2,   3,  -1,   1,  -1,   3,  12,   6,  29,   5,   3,  -9,  -2,  -2,  -1,  -1,   0,   0,   6,  -1,   7,  11,   2,  -4,  -1,  -1,   0,   0,   0,   5,  36,  13, 105, -23, -25,  -4,  -7,  -2,  -3,  -1,  -2,  -2,  13,  -1,   8,   2,  -5,   0,  -2,   1,  -3,  -3, -16,   3,   1,   1,   1,   1,  -5,   0, -11,   2,   3,   0,   1,   0,  -3,  -1,  -8,   0,   2,   1,   1,   0,  -2,   0,  -4,   0,  -2,   0,  -4,   0,  -1,   0,  -2,   0,  -1,   0,  -2 },
     {    1,   2,   4,   3,  10,  12,  -8,  -1,  -2,   0,  -1,   0,  -7, -12, -10, -12, -35, -97,  51,   2,  11,   1,   4,  -1,  -1,  -3,  -1,  -4,  -6, -23,   5,   2,   1,   1,   0,   0,   2,  10,   5,  22,  13,  20, -16,  -2,  -4,  -1,  -1,   0,  -1,   0,   0,   1,   1,   5,  -2,   0,  -7,   0,   5,  -2,   3,   9,  -4,   0,  -2,   0,   0,  -1,  -1,   2,   0,   0,   1,   0,   0,   0,   2,   5,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -1,   0,  -9,   1,  10,  -2,  -1,   0,   0,   0,   0,   0,   0,   1,  -7,  -1,  -1,  -1,   1,  -1,   1,   0,   0,   0,   1,   0, -11,   4,  21,  -2,  -3,  -1,   0,   0,   0,   0,  -1,   3,   4,  -6,   9,  -2,  -2,   0,   0,   0,   0,   0, -29,  -1,-109,  15,  32,  -4,   8,   2,  34,   0,   3,  -3,  -4,  -1,  -2,   0,  -3,   0,   5,   0,  -3,   1,   0,   0,   4,   0,   9,  -2,  -4,   0,   0,   0,  -3,   0,   5,   0,   1,   0,   4,  -1,  -1,   0,   2,   0,   1,   0,   1,   0 },
     {   -8,   1,   0,   1,   2,   0,  -2,   0,   0,   0,   0,   0, -10,   2,   4,   1,   3,   6,  -4,  -1,  -1,   0,   0,   0, -14,   1,   1,   2,   3,  -1,   1,  -1,   0,   0,   0,   0,  -4,   0,   5,  -3,  -3,  -3,   1,   0,   0,   0,   0,   0, -39,   0, -17,   1,  11,  -1,   3,   0,-105,   4,  49,  -4,   3,   1,   3,   0,  11,  -1,  -1,   0,  -1,   0,   0,   0,   6,  -1,   0,   0,  -1,   0,   0,   0,  13,  -1,  -4,   0,   2,   0,   0,   0,   5,   0,  -1,   0,   0,   0,   0,   0 },
     {    6,   9,   8,  15,   9,  44,  90, -52,   0, -10,   3,  -7,   0,   2,   1,   1,   0,  11,  17,  -6,  -2,  -1,   0,  -1,  -4,  -3,  -6,  -6, -11, -24, -25,  18,   1,   3,  -1,   2,  -4,  -3,  -7,  -5, -19,   2,  -1,   4,   1,   1,   0,   1,  -2,   2,  -4,  -5,   0,  -1,  -5,   3,   1,  12,   1,  -4,   3,  -3,  -2,   0,   1,  -2,   2,   0,   2,  -1,  -3,   1,   0,   0,   1,   1,   1,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -1,   3,  -2,   4,  -1,   1, -15,   6,  -1,   1,  -1,   1,  -6,  -1, -11,  -6, -29,  -4,  -2,   7,   2,   2,   1,   1,   1,  10,   2,  14,  13,  16,  -8,  -4,  -1,  -2,  -1,  -1, -13,  -1, -27,  -8, -86,  41,  19,  -1,   4,   1,   2,   0,  -4,  14,  -8,  26, -30, -14,  13,  -3,   1,  21,  -5,  -6,  26, -10,  -1,  -2,   1,  -3,   5,  -4,   7,  -3,   0,   0,   1,  -2,   3,  -1,   8,  -2,  -1,   0,   1,  -3,   2,  -1,   1,  -1,   1,   0,   1,  -1,   1,   0,   0,   0,   0,   0 },
     {    2,  -1,   4,  10,   7,  15,   7,  -8,  -2,  -1,  -1,  -1,  -2,  -9,  -1,   4,  -2,   0,   7,  -2,   1,   0,   0,   0,   3,  -2,   5,  14,  13,  28, -20,   0,  -3,   0,  -1,   0,  -1,  -5,  -2,   0,   3,   4,  -2,   0,   0,   0,   0,   0,   1,  -8,   4,  44,  -7, -17,   1,  -1,  -6,-100,   1,  34,   6,   6,  -1,   4,   0,  10,  -1,  -7,  -2,  -4,   1,   0,  -1,   4,   0,  -5,   0,   0,   1,   0,   0,  10,  -1,  -6,   1,   2,   0,  -2,   0,   4,  -1,  -2,   0,   0,   0,  -1 },
     {   -2,  -7,  -5,  -9,  -7, -20,  -2,   2,   4,   1,   1,   0,   0,  -6,  -2,   0,  -8,   2, -14,   4,   0,   2,  -1,   1,  -3, -14,  -7, -16, -11, -60,  26,   4,   5,   1,   2,   0,  -5,  -4,  -9,   6, -48,  -4,  14,   3,   2,   1,   1,   0,  -4, -23, -14, -51,   5,  32,   0,   4,  -2, -50,  -8,  24,  10,   4,   0,   1,   0,   6,   2,   4,   6,   5,  -3,  -1,  -2,   4,   2,   3,   3,   0,  -1,  -1,   0,   6,   2,   1,   0,   2,   1,   1,   0,   2,   1,   1,   0,   1,   0,   0 },
     {   -1,   0,  -2,   4,  -6, -26,   3,   4,   2,   0,   0,   0,   0,   2,   0,   2,   7,   7,   3,  -2,  -1,   0,   0,  -1,  -2,  -4,  -4,   2, -15, -68,  28,   2,   5,   1,   2,   0,  -2,  -5,  -4, -15,   4, -18,   1,   4,   1,   1,   0,   1,   3,  29,   9,  85,   9,  -8, -13,  -7,  -7,   4, -18,  -4,   9,   2,   1,   0,   1,  -2,   1,  -5,   0,   9,  -2,   0,   0,  -2,  -1,  -5,   0,   3,   0,   0,   1,  -1,   2,  -3,   0,  -1,   0,  -2,   0,  -1,   1,  -1,   0,   0,   0,  -1 },
     {   -1,   2,  -3,   2,   4,   8,   3,  -5,   0,  -1,   0,  -1,  -6,  -2, -12,  -5,   2, -17, -28,  15,   2,   3,  -1,   2,  -3,   3,  -6,   1,   3,  11, -14,   2,   2,  -1,   0,   0,   2,   2,  -2,   2,  19,  -1,   7,  -5,  -1,  -2,   0,  -1,  -6,   0,  -5, -10,  20,  -6,   2,   0, -39,   9, -99,  15,  35,  -5,   9,  -1,  10,  -2,  -3,  -3,  -1,  -1,   0,   0,  -1,  -1,   7,   0,  -4,   2,   1,  -1,   4,  -1,   9,  -1,   0,   0,   3,   0,   2,  -1,   2,  -1,   0,   0,   2,   0 },
     {    1,   1,   3,   3,   0,   4,   9,  -5,   2,  -1,   0,  -1,  -5,  -6,  -5, -12, -10, -33, -89,  45,   2,   8,  -3,   6,  -2,  -3,  -1,  -3,  -9, -17, -29,  12,   3,   2,   0,   1,   3,   2,   9,   2,  14,  13,  23, -16,   0,  -3,   1,  -2,   3,   5,   6,  15,   0,   3,   6,  -5,  13,  -2,  34,  -3, -15,   4,   6,  -4,  -8,  -4,   3,   1,   0,   2,   2,   0,  -1,   0,  -2,  -1,   0,   1,   5,  -2,  -2,   0,  -3,   0,   0,   0,  -1,   0,  -1,   0,  -1,   0,   0,   0,   0,   0 },
   }
 },
 {  //1
   {
     { -119,  22,  17,   4,   4,   2,   2,   1,   1,   1,   1,   0,  24, -20,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  18,  -1,  -7,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   3,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    3, -74,  11,   6,   4,   3,   2,   2,   1,   1,   1,   1,  89,   1, -33,   0,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0, -16,  35,   1,  -7,   0,  -1,   0,  -1,   0,   0,   0,   0, -10,  -1,   8,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   2,   0,   1,   0,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {   17, -36,  31,   0,   1,  -1,   0,   0,   0,   0,   0,   0, -25, -86,   7,  24,   4,   6,   2,   4,   2,   2,   1,   1,  42,   7, -45,   2,   2,   0,   0,   0,   0,   0,   0,   0,  -1,  31,  -2, -12,   0,  -1,   0,  -1,   0,   0,   0,   0,  -2,   6,   4,  -1,  -1,   0,   0,   0,  -2,   6,   1,  -1,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0,  -1,   4,   0,  -1,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0 },
     {   32,  83, -13, -17,  -6,  -6,  -3,  -3,  -2,  -2,  -1,  -1,  69, -41,   8,   7,  -5,   0,  -2,   0,  -1,   0,  -1,   0,  -7,   4, -14,   7,   2,   0,   0,   0,   0,   0,   0,   0, -23,  11,   7,  -5,   3,   0,   1,   0,   0,   0,   0,   0,  -6,  -8,   4,   2,   0,   1,   0,   1,  -5,   0,   0,   0,   1,   0,   0,   0,  -3,  -3,   1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,  -2,  -1,   1,   0,  -1,   0,   0,   0,  -1,  -1,   0,   0,  -1,   0,   0,   0 },
     {   -2,  28,   4,   2,  -3,  -1,  -2,  -1,  -1,  -1,  -1,   0, -34,   2, -56,   9,  11,   2,   4,   2,   2,   1,   1,   1,   3,  71,  -5, -46,   3,  -4,   0,  -3,   0,  -1,   0,  -1, -10,  -8,  53,  -3, -12,   0,  -2,   0,  -1,   0,  -1,   0,   4, -19,   1,  15,  -1,   0,   0,   0,   1,  -3,   1,   1,   1,   0,   0,   0,   2,  -5,  -1,   2,   0,   0,   0,   0,   1,  -2,   2,   0,   0,   0,   0,   0,   1,  -3,   0,   1,   1,  -1,   0,   0,   1,  -2,   0,   1,   1,  -1,   1,   0 },
     {    4,  -4, -51,   6,   4,   2,   2,   1,   1,   1,   1,   1,  14,  29,  -8, -27,   1,  -2,   0,  -1,   0,   0,   0,   0,  99,  -3,   0,  -6, -14,  -1,  -4,  -1,  -2,  -1,  -1,   0, -15,  26,   4,   3,  -2,  -3,   0,  -1,   0,  -1,   0,   0, -20,  -3,  15,   1,   2,   0,   0,   0,  -5,  -6,   1,   4,   0,   1,   0,   0,  -6,  -2,   2,   1,   1,   0,   0,   0,  -3,  -2,   0,   1,   0,   0,   0,   0,  -3,  -1,   1,   0,  -1,  -1,   0,   1,  -2,  -1,   1,   0,  -1,  -1,   0,   0 },
     {  -13,  -7, -83,  27,  12,   4,   4,   3,   2,   1,   1,   1,  -5, -19, -23,   6,  11,   3,   2,   1,   1,   0,   1,   0, -43, -10, -22, -11,  14,   3,   4,   2,   2,   1,   1,   1,  30,  33,  -6, -35,  -6,   0,   0,  -1,   0,   0,   0,   0,   7,  11,  27,  -6, -11,  -2,  -2,  -1,   2,  -1,   7,   4,  -2,  -1,  -1,   0,   1,   0,   6,   1,  -1,   0,  -1,   0,   1,   0,   3,   1,  -1,   0,   0,   0,   1,   0,   3,   0,   1,   0,   1,   1,   1,   0,   2,   0,   0,   0,   1,   0 },
     {    2,   9,  42, -10,  -9,  -1,  -4,  -1,  -1,  -1,  -1,   0,   5,  65,  -8,  25,  -8,  -8,  -3,  -4,  -2,  -2,  -1,  -1,   2, -10, -65,  10,  27,  -1,   5,   0,   2,   0,   1,   0,  -1,  24,  -3, -45,   4,   5,   1,   1,   0,   0,   0,   0,  -9,  -3,  30,  -3, -13,   1,  -1,   0,   1, -11,   2,   7,  -1,  -1,   0,   0,  -1,  -3,   2,   1,   0,   0,   0,   0,   0,  -5,   0,   1,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,  -2,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0 },
     {  -14, -21, -31, -12,  11,   3,   5,   2,   2,   1,   1,   1, -13,   6,  71,  -9, -23,   0,  -5,  -1,  -2,  -1,  -1,   0, -17,  50, -14,  25,   1,  -9,   0,  -3,   0,  -1,   0,  -1, -45,   7,  19, -14,  14,   0,   1,   0,   1,   0,   0,   0,  14, -35,   8,  10,  -6,   5,  -1,   1,   6,   0, -17,   2,   3,  -1,   1,   0,   5,  -3,   0,  -2,   0,   0,   0,   0,   3,   0,  -5,   0,   1,   0,   0,   0,   2,  -2,   0,  -1,   1,   0,  -2,   0,   1,  -1,   0,   0,   1,   0,  -1,   0 },
     {    3,  -7,   2, -42,   5,   3,   2,   1,   1,   1,   1,   0,  15,   2,  38, -11,  -9,   1,  -2,  -1,  -1,   0,  -1,   0,   2,  -9,  -5, -29,  -2,   5,   2,   2,   1,   1,   0,   1,  74,   5,  37,  -3, -38,  -2,  -7,  -1,  -3,   0,  -2,   0,  -9,  -7,   1,  49,  -2,  -8,  -1,  -1, -11,   1, -17,   4,  13,  -1,   1,   0,  -5,   0,  -2,   2,   2,   1,   0,   0,  -4,   0,  -4,   0,   2,   0,   0,   0,  -2,   0,  -1,   2,  -2,   0,  -2,   0,  -2,   0,   0,   1,  -1,   0,  -1,   0 },
     {    3,  15,  -2, -10,  -2,  -2,   0,   0,   0,   0,   0,   0,  -1,   4,   5,  -6, -20,   2,  -1,   0,   0,   0,   0,   0,   9,  72,  -3,  25,  -9, -18,  -2,  -6,  -2,  -3,  -1,  -2,  67,  -2, -44,   6,  26,  -4,  -3,  -1,   0,  -1,   0,   0,  -7,  18,   2, -27,   4,   7,  -1,   0, -16,  -6,  19,   0,  -4,   1,   1,   0,  -4, -11,   1,   3,   0,   1,   0,   1,  -4,  -3,   2,   0,   1,   0,   0,   0,  -1,  -4,   0,   0,  -2,  -2,   1,   0,  -1,  -2,   0,   0,  -1,  -1,   1,   0 },
     {    1,  -4, -35,   6,   0,   2,   0,   2,   1,   1,   0,   1,   9,   5,   6,  52,  -9, -10,  -2,  -4,  -1,  -2,  -1,  -1,  30,  -3, -25,   3,  38,  -5,   0,  -1,   0,  -1,   0,   0,  10, -72,   4,  -9,   2,  20,  -1,   5,   0,   2,   0,   2,  12,   3, -40,   2,  -6,   2,   4,   1,  -5,  26,  -1, -22,   2,  -3,   0,  -1,  -3,   4,  10,  -2,  -5,   0,   0,   0,  -3,   5,   2,  -3,   0,   0,   0,   0,  -2,   1,   3,   0,  -2,   2,   1,  -1,  -1,   1,   2,   0,  -1,   1,   1,  -1 },
     {   -5,  -8, -12, -96,  25,  13,   8,   4,   3,   2,   2,   1, -10,  -7, -43, -15, -13,   9,   5,   3,   3,   2,   2,   1,   2, -13,   3,  26,  -3,  -7,   0,  -1,   0,  -1,   0,   0, -16, -11,   6, -12,   5,   3,  -1,   2,   0,   1,   0,   1,  28,  20,   2,  -5, -17,  -2,  -1,  -1,   1,   7,  14,   9,  -4,  -5,  -1,  -1,   1,   0,   3,   7,   1,  -1,  -1,   0,   0,   1,   2,   3,   1,   0,   0,   0,   0,   1,   1,   2,   0,   1,   1,   1,   0,   0,   0,   1,   0,   0,   1,   1 },
     {   -9,  -6, -18, -27,  -9,   8,   2,   4,   2,   2,   1,   1,  -7,   8, -19,  63, -20, -10,   1,  -3,   0,  -1,   0,   0, -10,  -8,  15,  22,  -1, -12,  -2,  -1,   0,  -1,   0,   0,   3,  45,  12,  29,   9, -16,  -5,  -5,  -2,  -2,  -1,  -2, -38, -17, -20,   6,  44,   2,   1,   0,   8, -11,  -6, -32,   2,  12,   0,   2,   5,   3,   1,  -4,  -6,   1,   1,   0,   3,  -2,   1,  -7,   0,   0,   0,   0,   2,   0,   1,  -1,   2,  -1,   0,  -3,   1,   0,   0,  -1,   1,   0,   0,  -2 },
     {    3,  -1,   8,  12,  -9,  -2,  -3,   0,   0,   0,   0,   0,   1, -16,  -9,  26, -15,  -9,   2,  -1,   1,   0,   0,   0,  15,  -5,  71,  22,  23, -21, -11,  -4,  -4,  -2,  -3,  -1,  16,   1,  13, -38,  20,  25,  -9,   1,  -2,   1,  -1,   0,  -7, -20,  27,   6, -36,  13,   4,   1,  -1, -13, -21,  31,   3,  -9,   3,   0,  -3,   5, -17,  -8,   9,   1,   0,   0,   0,   2,  -3,  -2,   0,   0,   0,   0,  -1,   3,  -5,  -3,   0,   1,  -1,   0,   0,   1,  -2,  -1,   0,   0,   0,   0 },
     {    3,   2,   2,  25,   7,  -5,  -1,  -2,  -1,  -1,   0,  -1,  -2,  -7, -33, -40, -10,  21,   6,   5,   2,   2,   1,   1,  -6, -21, -39,  46, -19, -23,   9,   0,   3,   0,   1,   0,  23, -26,  31,  13,  32, -12, -13,   0,  -3,   0,  -1,   0, -17, -34,   5,  11,   4,  20,  -2,   1,   3,  13, -29,  -3,   6,  -2,   4,  -1,   4,  11,   6, -16,  -1,   2,  -1,   0,   0,   4,   2,   0,  -2,  -1,   0,   0,   0,   3,   4,  -4,   0,   2,   1,   0,   0,   1,   1,  -2,   0,   1,   1,   1 },
     {   -6,  -4,  -5, -19,  22,   0,   2,   0,   0,   0,   0,   0,  -7, -13,  10,  -5,  12,   3,  -1,   0,   0,   0,   0,   0, -10,   5,  -6, -12,  11,   8,   0,   1,   0,   0,   0,   0, -11, -42,  -7,   0,  -4,  14,   3,   4,   1,   1,   1,   1,-100,   7,  26, -10,  -6,   0,   6,   1,  17, -18,   8,   4,  -9,  -2,  -1,   1,  18,   2,  -9,   7,  -5,  -2,  -1,   0,   4,   6,   0,   0,   0,  -2,  -1,   0,   6,   1,  -1,   2,   3,   2,   0,   0,   3,   1,   0,   1,   1,   1,   0,   0 },
     {   -2,   4,  -2,  26,  -7,   3,  -1,  -1,   0,  -1,  -1,   0,  -8,   0,  -3,   6, -47,  10,   7,   0,   1,   1,   1,   0,  -4,   3,  -3,  23,  -9, -14,   4,  -1,   1,  -1,   1,  -1, -21,  -9,  18,   5, -18,  -5,   1,   1,   1,   0,   0,   0, -10,  73,  12,  30,  -1, -33,  -1,  -7, -19, -11,   3,   6,  46,  -3,  -5,   0,   6, -18,  -8, -12,   6,  12,  -1,   1,   3,   1,  -3,  -5,   4,   1,   1,   0,   2,  -3,  -1,  -4,   1,  -1,  -1,  -1,   2,  -1,  -1,  -2,   1,   0,   0,  -1 },
     {   -6, -11, -14, -16,-106,  20,  18,   5,   5,   4,   2,   2,  -1, -11,   2, -19,  -7, -20,   8,   4,   2,   1,   1,   0,  -3,  -5, -14,  -9,  17,  -7,  -3,   2,   0,   1,   0,   0,  -1, -16,   7,  17,   4,  -9,  -4,   0,  -1,   0,   0,   0,   1,   1,  20, -10,   8,   3,  -7,   0,  11,   0,   5,  17,  -9,  -1,  -1,  -2,   0,   2,  -7,   7,  12,  -2,  -1,  -1,   1,   2,  -1,  -2,   1,   3,  -1,   0,   0,   1,   0,   1,   1,   1,  -1,   0,   0,   0,   0,   1,   0,   1,   0,   0 },
     {    2,  -3,  -6,   2,  21,  -7,  -3,   0,   0,  -1,   0,   0,   7,   3,  14,  44,   4,   5,  -8,  -4,  -2,  -2,  -1,  -1,   6, -13,  -6, -11, -34,   0,  19,   2,   4,   1,   2,   1,   4, -31,  24,  13,  12, -39, -10,   5,  -1,   1,   0,   1,  26,   6,  39, -16,  16,  20, -23,  -2,  -9, -30,  11,  27, -22,  15,   6,  -2,   0,  -3, -33,  16,  11,  -7,   4,   0,  -1,   9,  -7, -13,   4,   3,   1,   0,  -1,   2,  -3,  -1,  -1,   3,  -2,  -3,  -1,   1,  -1,   0,  -1,   1,   0,  -2 },
     {    0,  -1,   8,  11,   2, -10,  -1,  -2,   0,  -1,   0,   0,   1,   2,  20, -19,  -9,   1,   1,  -1,   0,   0,   0,   0,  -4, -13,   9,  -9,  14, -22,  -4,   2,  -1,   1,   0,   1,   1,  21,  62, -12,  24,  10, -23,  -2,  -6,  -1,  -3,  -1, -12,  38, -23, -29,  -2,  30,   5,  -3,  -9,  14,  31, -18, -29,  -1,   7,   0,   2, -24,  15,  29, -11,  -4,   0,   1,   1,  -4, -16,   8,   5,  -2,   0,   0,   1,  -3,  -2,   2,   1,  -1,  -5,   2,   0,  -1,  -1,   1,   0,   0,  -3,   1 },
     {    3,   3,   2,   8,   3,  12,  -8,  -1,  -2,   0,  -1,   0,   1,   1,   0,  10, -75,  25,  18,  -1,   2,   1,   1,   0,   1,  -8,   2, -47, -30, -12,  15,   6,   2,   2,   2,   1,  -2,   1, -20, -12,   6,  -6,  -4,   3,   1,   1,   1,   1,  -6, -14,  -7,  -7, -11,  -6,   2,   1,  51,  25,   8,   6,  -8, -15,  -7,  -2,  -4,  16,   6,  16,  10,  -6,  -3,  -3,  -5,  -2,   4,   1,   9,   0,  -2,   0,  -2,   0,   0,   3,  -1,  -1,   0,   1,  -1,   1,   0,   2,  -1,   0,   1,   0 },
     {    3,   0,   5,   4,  19,   4,  -2,   0,  -2,  -1,  -1,   0,  -2, -12,  -6, -29, -42, -12,  17,   4,   4,   2,   2,   1,   2, -11,  -1, -27,  45, -17, -12,   7,  -1,   1,   0,   1,  -5,  -8, -20,   1,  10,   8,  -9,   0,   1,   1,   0,   1,  19, -18,  16,  14,  29,  -2, -11,  -3, -24, -42,  10, -20,   2,  42,  -1,   2,   4,  11,  -7,   6, -35,  -1,  11,  -2,   2,   9,   1,   8,   3,  -7,   0,   0,   1,   3,  -1,   1,   1,   2,   1,   2,   1,   1,  -1,   0,   0,   1,   0,   1 },
     {   -7,  -4,  -1,  -8, -12,  -1,   7,   1,   1,   0,   1,   0,  -7,   1,  -2,  10,  -7,   7,   3,  -1,   0,   0,   1,  -1,  -8,  -9,   6, -28, -17,   5,  14,   3,   2,   1,   1,   0, -17,  -2, -22, -12,  20, -16,  -5,   6,   1,   2,   0,   1, -22, -21,   4,  15,  -8,  14,  -5,   0, -93,  35,  12,   3,   4, -11,  10,  -4,   4,  -6,  17, -10,   8,  -4,  -5,   2,  18,  -6,   0,   5,  -6,   2,  -2,   1,   7,   1,  -3,   1,   5,   0,   1,   1,   2,   1,   1,   0,   2,   0,   0,   0 },
     {    0,  -2,  -1,  16,   3,  24,  -5,  -4,  -2,  -1,  -1,   0,  -1,  -2,  -7,   0, -16,  14,   2,  -1,   1,   0,   0,   0,  -2,  -8, -10,  30,   4,  12,  -9,  -4,   0,  -1,   0,  -1,   7,   2,  12,  23, -30,  18,  22,  -7,   0,  -2,   1,  -1,   0, -49, -12, -12, -13, -24,  19,  10, -27, -17,  51,  15,  -8, -18, -13,   4,   1, -16, -12,  35,  12,  -5,  -3,  -2,   5,   7, -21, -11,  16,  -1,  -1,   0,   1,   5,  -1,  -8,   0,   3,  -2,  -2,   0,   3,   1,  -2,   0,   2,  -1,  -2 },
     {    1,   0,   7,   6, -14,   5,   3,  -2,   0,   0,  -1,  -1,   0,  -6,  -2,  -4,   0,   3,  -4,  -2,   1,   0,   1,  -1,   7,   4,  26,  28, -11,  20,  16, -12,  -1,  -3,  -1,  -2,   7, -18,  13, -55, -20, -33,  26,  15,   2,   4,   1,   1,  -8, -13,  16, -10,  23, -26, -32,  10,   7,  24,  11, -23,  -2,  31, -11,  -7,  -4,  -6,  22,  14, -26,   0,   6,   0,  -2,  -3, -12,  20,   8,  -5,   1,  -1,   0,   1,  -5,  -3,  -1,  -1,  -2,   4,  -1,   1,  -2,  -1,   1,   0,  -2,   2 },
     {    0,  -1,   4,   2,  24, -44,   5,   4,   2,   0,   0,  -1,  -3,  -1,  -2,   4,  -8, -49,   1,   7,   2,   2,   1,   0,  -3,  -1,   0,   4,  16,  -9, -23,   3,   1,   1,  -1,   1,   0,   0,  -1,  37,  -5,  -1,  -6,  -6,   1,  -2,   0,  -1,  -1, -10,  46,   6,   0,  -9, -12,  -2,  14,  59,  25,   4,   8,   3, -14,  -9, -18,  -5,  11,   9,  16,  15,   3,  -4,   0, -18,  -9,  -5,   1,  12,   3,   2,   1,   0,  -7,  -1,   1,  -2,  -3,  -4,   0,   1,  -1,  -2,   2,  -1,  -1,  -1 },
     {    3,   7,   8,   8,  19,  96, -20, -17,  -6,  -4,  -1,  -1,   2,   3,   9,   5,  21,  -4,  21,  -9,  -5,  -1,  -1,   0,  -2,   5,   3,   4,   9, -29,  -2,  10,  -3,   0,  -1,   1,  -3,   1,   3,  16, -11,   0,  -6,   0,   3,  -2,   0,  -1,  -2,  10,  26,  -4,   3,   1,  -2,  -2,  -7,  30, -11,  -4, -11,   5,   7,  -1, -23,   6,   4,  -7,   0, -14,   8,   3,   0, -13,   6,   1,   1,   1,  -3,   3,   1,  -5,  -6,   1,   1,  -2,  -1,   0,   1,  -1,  -1,   0,   1,  -1,   0,  -1 },
     {    1,  -3,   0,  -3,   5,   8,   8,  -7,  -1,  -2,   0,   0,   4,  -1,   6,  -3,  38, -10,  20,  -2,  -2,  -1,  -1,  -1,   0,  -6,  -2,  -8,  -3, -55,  14,  16,   3,   2,   0,   1,  -2,  -2, -12, -24,  12, -19, -20,  12,   2,   4,   1,   2,  -8, -18, -37,  11,   3, -14, -15,   2,  17, -14,  33,  21,  31,   3, -26,  -7,  -9, -27,  -5,  -2,   7,  37,  -1,  -3,   3,   8, -13,  -7, -12,   4,   6,  -1,   1,   1,   2,  -1,   1,   2,  -2,   0,  -1,   1,   1,   0,   1,   1,   0,   0 },
     {    0,   4,   1,   6,  17,   7,  18,  -5,   0,  -2,  -1,  -1,  -5,  -3,  -2, -11, -18, -67,  10,  12,   2,   5,   2,   3,  -7,   0, -10,   1, -27,  10, -18,   1,   5,   0,   2,   0,   1,  -2,  17, -30,   0,   6,  17,  -5,  -2,   1,   0,   1, -13,   9, -36, -22,  23,  -9,  -6,   9, -18,  -9, -13,  32, -11,   6,  -4,  -3, -24,  32,  -4,  -9,  21, -11,  -1,  -3,   9,   2,  14, -12,   1,  11,  -4,  -2,   6,  -5,   5,   5,   2,   0,   2,   0,   2,  -2,   1,   2,   1,   0,   1,   1 },
     {    5,   3,   3,   5,   2,  17,   7,  -6,   0,  -3,   0,  -1,   5,   2,  -3,   2,   5, -41,   0,   8,   0,   1,   1,   1,   4,   1,  -2,   4, -33, -15, -10,   6,   4,   1,   1,   1,   4,   0, -10, -13, -12,  13,  -8,  -6,   2,   0,   0,   0,  10,  -1,   6,  14,  -1,   8,   0,  -7,   4,  16,   4, -17,  -2,  -2,   8,   0,  98,  -1, -12,  12,  -1,  -6,  -9,   2,  -3,  10, -10,   4,  11,   7,  -1,  -3, -14,   3,   6,  -5,  -2,  -2,   2,   4,  -4,   0,   0,  -2,  -4,   0,   0,   1 },
     {    2,   3,   2,  -1,  14,  16,  -4,  -4,  -3,   0,  -2,  -1,   2,   3,  -1, -13, -12,  15,  -8,  -5,  -2,   0,   0,   1,   4,  -1,   3, -14,  45,  27,   4, -16,  -8,  -4,  -2,  -1,   8,  16,  10,   8,  30, -17,  24,  11,  -9,  -3,  -2,   0,  -3,   3, -14, -18,   5, -12, -25,  18,   1,  34, -12,  16,   7,  11, -10, -13,  39, -15, -42, -16,  18,  13,   9,  -1,   6,  11,   0, -25, -18,   7,   2,   2, -11,  -1,  14,   3,  -1,  -2,   0,   1,  -1,   0,   1,   2,  -2,  -1,   2,   1 },
   },
   {
     {  116, -31, -14,  -5,  -3,  -2,  -2,  -1,  -1,  -1,  -1,   0, -37,  -1,   5,   2,   2,   1,   1,   1,   1,   0,   0,   0, -14,   5,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   39,  35, -15,  -8,  -5,  -2,  -2,  -2,  -1,  -1,  -1,  -1, 103,  -9, -24,  -6,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0, -19, -23,   0,   5,   3,   2,   1,   1,   1,   1,   0,   1, -26,  -5,   8,   4,   2,   1,   1,   0,   0,   0,   0,   0, -13,   0,   5,   1,   0,   0,   0,   0,  -6,   1,   2,   0,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0,  -4,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   12, 109,   0, -16, -11,  -7,  -4,  -4,  -3,  -2,  -2,  -2, -50,   8,  -1,  -2,   3,   1,   2,   1,   1,   0,   1,   0, -17, -26,  -2,   6,   4,   2,   2,   1,   1,   1,   1,   0,   8,  -8,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   6,  -6,   0,   1,   0,   0,   0,   0,   4,  -2,   0,   1,   0,   0,   0,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0 },
     {   12,   5,   8,  -8,  -2,  -2,  -1,  -1,   0,   0,   0,   0,  12, 102,   8, -18,  -9,  -8,  -4,  -5,  -2,  -2,  -1,  -2,  45,  21, -19, -12,  -4,  -2,  -1,  -1,  -1,   0,   0,   0,  -9, -32, -13,   3,   4,   3,   2,   2,   1,   1,   1,   1, -10, -16,   0,   6,   3,   1,   1,   1,  -6,  -7,   2,   3,   1,   0,   0,   0,  -3,  -3,   1,   2,   0,   0,   0,   0,  -2,  -5,   1,   1,   0,   0,   0,   0,  -2,  -2,   1,   1,  -1,  -2,   0,   0,  -1,  -1,   0,   0,  -1,  -2,   0,   0 },
     {  -15, -21,  24,   4,   1,   0,   0,   0,   0,   0,   0,   0,  -1,  49,  17,  -7,  -6,  -5,  -3,  -3,  -1,  -1,  -1,  -1,-102, -13,  14,   4,   6,   3,   3,   2,   2,   1,   1,   1, -23,  -7,   3,   2,   3,   2,   1,   0,   1,   0,   0,   0,  17,   7,  -4,  -2,  -1,  -1,  -1,   0,  15,  -1,  -6,  -1,  -1,   0,   0,   0,  10,   0,  -4,  -1,   0,   0,   0,   0,   5,  -2,  -2,   0,   0,   0,   0,   0,   4,   0,  -1,   0,   2,  -1,  -1,   0,   2,   0,  -1,   0,   1,  -1,  -1,   0 },
     {  -19,  -3,-106, -13,  13,  11,   9,   6,   4,   3,   2,   2, -12,  25, -37,  -1,  10,   2,   2,   0,   1,   0,   0,   0, -15,  21,  27,   8,   0,  -5,  -3,  -3,  -1,  -1,  -1,  -1,   0,  -1,  16,   2,  -3,  -2,  -1,  -1,   0,   0,   0,   0,   4,  -4,   9,   0,  -2,  -1,  -1,   0,   5,  -3,   4,   0,  -1,   0,   0,   0,   3,  -2,   3,   0,  -1,   0,   0,   0,   2,  -2,   2,   0,  -1,   0,   0,   0,   1,  -1,   2,   0,   1,  -1,   1,   0,   1,  -1,   1,   0,   1,  -1,   1,   0 },
     {    6,  30,   2,  -1,  -6,  -2,  -2,  -1,  -1,  -1,  -1,   0,  21, -10,  40,   1,  -7,  -3,  -3,  -1,  -2,  -1,  -1,   0, -21,  89,  35, -20, -10,  -9,  -4,  -5,  -2,  -2,  -1,  -1,  16,  31, -18, -19,  -3,  -2,   0,  -1,   0,   0,   0,   0,  -6, -17, -19,  -2,   5,   3,   2,   2,  -3, -12,  -7,   4,   3,   2,   1,   1,  -2,  -9,  -2,   4,   2,   1,   0,   0,   0,  -4,  -2,   2,   1,   0,   0,   0,   0,  -4,  -1,   1,   0,  -2,  -1,   1,   0,  -2,  -1,   1,   0,  -1,  -1,   0 },
     {    8,  -8,  13,  -5,  -1,  -1,  -1,   0,   0,   0,   0,   0,  15,  23, -43, -16,   4,   2,   3,   1,   1,   0,   1,   0, -18,  -8, -22,   0,   9,   4,   3,   2,   1,   1,   1,   0,  95,  38,  -7,  -4,  -8,  -6,  -5,  -3,  -2,  -2,  -1,  -1,  24,   4,  -6,  -2,  -3,  -1,  -1,   0, -17, -16,   3,   5,   2,   2,   1,   1, -12,  -5,   5,   3,   1,   0,   0,   0,  -6,  -2,   4,   1,   0,   0,   0,   0,  -3,   0,   2,   1,  -3,  -1,   2,   1,  -2,  -1,   1,   0,  -1,  -1,   1,   0 },
     {    1,  -5, -41,  -3,  -3,   3,   4,   2,   1,   1,   1,   1,  18,   8,  83,  22, -10,  -7,  -8,  -5,  -4,  -2,  -2,  -1,  10, -54,  22,  12,  -4,   2,  -1,   1,   0,   1,   0,   1,  41,  -9, -24,  -8,  -3,   2,   1,   1,   0,   0,   0,   0,  11,   6,  -8,  -4,   0,   0,   0,   0,  -8,   3,  -7,  -2,   2,   0,   1,   0,  -7,   5,   0,   0,   1,   0,   0,   0,  -4,   2,  -4,  -1,   1,   0,   0,   0,  -2,   2,  -1,   0,  -2,   1,  -2,  -1,  -1,   1,   0,   0,  -1,   0,  -1,   0 },
     {    8,  15,  -4,  98,  16, -12, -10,  -7,  -4,  -3,  -2,  -2,   0,  22, -15,  57,   3, -14,  -5,  -4,  -1,  -1,  -1,  -1,   2,  -5,   2, -11, -14,  -3,   2,   2,   2,   1,   1,   0,   0,  10,  15, -23, -10,   2,   2,   1,   0,   0,   0,   0,  -2,   5,   6, -15,  -3,   3,   1,   0,  -4,  -4,   1,  -6,   1,   2,   1,   0,  -2,  -2,  -1,  -5,   0,   1,   1,   0,   0,  -1,   0,  -3,   1,   1,   0,   0,   0,  -1,   0,  -2,   0,  -1,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,  -1 },
     {   -3,   0, -23,  24,   9,   2,  -1,   0,   0,   0,   0,   0,   1, -26,   3,   2,   5,   1,   0,   1,   0,   0,   0,   0, -33,  17, -59, -29,   5,   4,   8,   3,   4,   1,   2,   1,  19, -54, -53,   7,  14,   9,   5,   4,   2,   2,   1,   1,  -7, -23,  13,  22,   7,   1,  -2,   0,   3,   9,  18,   6,  -4,  -4,  -2,  -1,   1,   7,  12,  -2,  -4,  -2,  -1,   0,   2,   5,   4,  -2,  -2,   0,   0,   0,   1,   3,   4,  -1,   1,   3,   2,  -1,   0,   1,   2,   0,   0,   1,   1,   0 },
     {   -3,   6, -28,  -6,   7,   4,   2,   1,   1,   0,   0,   0,  -2,   6,  31, -13,   0,   0,  -2,  -1,  -1,  -1,   0,   0,   1,  -7, -65, -34,   4,   6,   9,   4,   3,   2,   2,   1, -36,  71,  -3, -13,   5,  -3,   1,  -3,   0,  -1,   0,  -1,   8,  36,  15,   3,  -1,  -6,  -3,  -3,   3, -11,  -4,   4,  -1,   0,   1,   1,   2, -11,   3,   6,   0,   0,   0,   0,   1,  -6,   1,   3,   0,   0,   0,   0,   2,  -3,   3,   2,   1,  -3,   0,   1,   1,  -2,   1,   1,   0,  -1,   0,   1 },
     {   -6,  -5,  -3,  16,   7,   3,   0,   0,  -1,   0,   0,   0, -11,   2,  10, -40, -14,   4,   5,   3,   1,   1,   1,   1, -11, -29,  15, -28, -10,   8,   4,   3,   1,   2,   1,   1,  15,  22,  25,   2,  -2,  -1,  -3,  -3,  -2,  -1,  -1,  -1, -84, -39,  13,  15,  12,   5,   2,   1, -21,  -6,   3,   8,   5,   0,   1,   0,  12,  15,  -1,   0,  -2,  -3,  -1,  -1,   8,   2,  -5,   2,   0,  -1,  -1,   0,   5,   2,  -2,   1,   4,   1,  -1,   1,   2,   1,  -1,   0,   1,   0,  -1,   1 },
     {   -7,  -6,   7, -36, -13,  -2,   4,   3,   3,   2,   1,   1,  -1,  12,   1,  67,  23,  -4,  -5,  -5,  -3,  -3,  -2,  -2, -13,   9, -33,  53,  19,  -5,  -2,  -3,  -1,  -2,  -1,   0,  -3,  21, -11, -13, -11,  -8,   1,   0,   1,   0,   0,   0, -52, -23,   8,  -5,  -2,   3,   4,   2, -12,  -5,  10,  -5,   0,   3,   1,   1,   9,   5,   6,  -4,  -2,   0,  -1,   0,   7,   0,  -1,  -6,  -2,   0,   0,   0,   5,   0,   0,  -2,   2,   0,   0,  -2,   2,   0,   0,  -1,   1,  -1,   0,  -1 },
     {   -6,  -9, -15,  18, -69, -13,  10,   6,   4,   2,   2,   1,  -8,  -1, -25,   1, -53,  -4,  12,   4,   3,   2,   2,   1,   4, -14,   1,   8,   8,  12,   3,  -2,  -1,   0,   0,   0, -20,  20, -34, -17,  24,  10,   2,  -2,   0,   0,   0,   0,   7, -24, -41,  -8,  21,   7,   1,   2,   7, -10,   1,   9,  11,   0,  -3,   0,   4,   4,  10,   2,   2,  -3,  -2,  -1,   1,   2,   7,   0,   1,  -1,  -1,  -1,  -1,   1,   3,  -1,   0,   0,   3,   0,   0,   0,   2,  -1,   0,   0,   1,   0 },
     {   -2,  -2,  -6,   0, -55,  -7,   9,   4,   3,   2,   1,   1,   3,  -5,  20,  20, -41,  -4,   7,   1,   0,   0,   0,   0,  -7,  23, -28,  -4,  -5,   6,   6,  -1,   1,   0,   0,   0,  20, -17,  69,  31,   0,   2,  -9,  -5,  -4,  -2,  -2,  -1,  14,  -4,  40,   8,  -3,  -2,  -7,  -3,  -5,  -6,  -9, -15,   2,   2,   1,   2,  -7,  -4,  -5,  -6,   6,   2,   0,   1,  -2,   2,  -5,  -4,   4,   0,  -1,   0,  -2,   1,  -1,  -1,  -1,   0,  -4,  -2,  -1,   0,  -2,  -1,   0,   0,  -1,  -1 },
     {    4,   5, -10,  45,  20,   0,  -7,  -4,  -2,  -2,  -1,  -1,   3,  -4,  24, -39,   4,  -1,  -3,   0,   0,   1,   0,   0,   6,  11, -22,  61,  36,  -8,  -6,  -5,  -2,  -3,  -1,  -1,  -4,  16,   7,  52,   7, -15,  -6,  -4,  -2,  -2,   0,  -1,   9, -30, -24,  -9, -13,  -1,   5,   5,   8, -13, -11,  -5,  -4,   4,   3,   2,   5,   2,   5,  -8,  -5,   2,   0,   0,   0,   2,   1,  -2,  -2,   1,   1,   0,  -1,   1,   3,  -4,  -1,   0,   0,  -1,  -1,   1,   1,  -2,  -1,   0,   0,   0 },
     {   -1,   7,  -2,  21, -22,  -6,   1,   1,   0,   1,   0,   0,  -3,   6,  -2, -25, -19,  -2,   4,   2,   1,   1,   1,   0,  -2,  18,  11,  30,  15,  -1,  -3,  -4,  -3,  -3,  -1,  -1,  15,  -1, -30,   8,  12,   0,   0,  -1,   1,  -1,   0,   0, -46,  75,  28, -17,  -3,  -9,  -3,  -4,  -1,  42,  14, -11,  -5,  -5,  -1,  -2,  -4,  -8,  -9, -10,   0,   2,   2,   1,   0, -10,  -3,   2,   2,   1,   0,   1,   0,  -7,  -1,   0,   2,  -4,   0,   0,   1,  -3,   0,   0,   1,  -2,   0,   1 },
     {    5,   0,  -1,  -7,   2,  -1,   0,   1,   0,   0,   0,   0,   9,   7,  -3,   9,   0,  -3,  -2,   0,   0,  -1,   0,   0,   7,  -3, -10, -11,  -7,  -1,   1,   1,   1,   1,   0,   1,  22,  17,   9,   1,  -2,  -2,  -1,  -2,  -2,  -1,  -1,  -1, -11, -18, -13,   7,   7,   5,   3,   1,  92,  52, -19, -22, -12,  -5,  -4,  -3,  28,  13,  -9,  -7,  -3,  -1,  -1,   0, -18, -16,  -2,   3,   4,   2,   1,   1, -13,  -7,   3,   4,  -5,  -3,   2,   1,  -2,   0,   1,   1,  -2,  -1,   1,   0 },
     {   -2,  -4,  -7,  16, -50, -11,  -1,   6,   2,   1,   1,   1,   4,   1,  12, -19,  60,  27,  -2,  -3,  -5,  -3,  -2,  -1,   2,   1,  -2, -25,  57,  14,  -8,  -2,  -3,  -1,   0,   0,   3, -24,  23, -34, -21,  -8,  -3,   4,   2,   2,   1,   1,  -8,   6, -13, -23,  -5,  -2,   5,   3,   8,  -1, -14,  10,  -1,  -2,   2,  -1,  -1,   4,   5,  11,  -3,  -2,   1,  -1,   1,   1,  -1,   5,  -6,  -3,   0,  -1,  -2,  -1,   0,   3,   0,   0,   0,   3,  -1,   0,   0,   2,   0,   1,   0,   2 },
     {   -1,   1,  -8,  -9,  19,   5,   3,   0,  -1,  -1,   0,   0,  -4, -10,  -1, -20, -25, -11,   1,   3,   3,   2,   1,   1, -10,   7, -39,  19, -35, -11,   7,   3,   4,   1,   1,   0,   6, -30,  18, -29, -32,  10,   8,   5,   2,   2,   1,   1, -19,  16, -32, -56,   1,  16,  11,   3,   0,  -1, -33,   1,  25,  10,   2,  -1,   3,   1,   6,  16,  14,  -1,  -4,  -2,   3,   3,   4,   9,   4,  -2,  -2,  -1,   1,   0,   3,   2,   1,   0,   2,   3,   1,   0,   2,   1,   1,   1,   1,   1 },
     {   -2,   1,  -3,  -7,   2,   6,   1,  -3,   0,   0,   0,   0,  -1,   3,  -9,  32,  15,  -1,  -1,  -3,  -1,  -1,  -1,  -1,  -3,   0,   2, -30,   3,   2,   0,   2,   0,   1,   0,   0,   7,  -8,   1,  53,  39,  -6,  -9,  -5,  -3,  -2,  -2,  -1, -30,  38, -51,   5,  22, -11,   2,  -1, -11, -20, -50, -19,   1,   3,   8,   5,   1,  -4,   4,   5,  -1,   1,   1,   0,   5,   5,  11,  -2,  -5,  -1,   0,   0,   1,   1,   6,   1,   2,   1,   3,  -2,   0,  -1,   2,   0,   0,   0,   1,  -2 },
     {    1,   7,   3,  -9,  28,  28,   7,  -3,  -4,  -2,  -2,  -1,  -2,   4,   4,  13, -26,   3,  -3,  -1,  -1,   0,   0,   0,  -2,   9,  -6, -15,  38,  21,  -4,  -5,  -3,  -2,  -1,   0,   0,   3,  29, -16,  34,  10, -11,  -5,  -3,  -1,  -1,   0,  12, -11, -28, -26,  -7,  -6,   0,   4, -39,  57,  23, -10,  -5,  -9,   1,  -1,  -9,  28,  18,   0,  -8,  -8,  -1,  -2,  -2,  -6,  -5,   0,   0,   1,   1,   0,   3,  -8,  -4,   4,   2,  -3,  -2,   2,   1,  -3,   0,   2,   1,  -3,  -2,   1 },
     {   -6,  -2,  -8,   0, -29, -37,  -8,   4,   4,   3,   1,   0,  -1,   5,   4, -10,  33,  -6,  -2,   1,   1,  -1,  -1,  -1,  -7,   0,  -6,  -8, -45, -16,   9,   5,   5,   1,   2,   1, -11,  14,   1,  35, -34,  -9,  10,   0,   0,  -1,   1,  -1,  10,   1, -33,   5,   4,  12,   9,  -1, -33,  44,  27, -15,  -3,   1,  -1,  -5, -13,  27,  21, -10,   4,   1,  -3,  -3,  -4,  -5,  -2,  -7,   0,   2,  -1,   1,   0,  -5,  -2,   0,   3,  -4,  -2,   0,   3,  -2,   1,   1,   1,  -3,  -1,   0 },
     {    4,   6,   6,  18, -19,  79,  30,  -7,  -9,  -5,  -3,  -2,   2,   8,   1,   6,  -2,  61,  20,  -8,  -6,  -3,  -2,  -1,  -3,  -3,  -6,   4, -25, -21, -10,  -1,   3,   3,   2,   1,  -5,   2,  -9,   6, -21, -36,  -9,   4,   5,   3,   2,   1,   2,   1,  -4,   5,   9, -12,  -2,   2,   1,   7,   4,   2,   8,  -5,  -2,   0,  -6,   3,   4,  -3,   5,  -3,  -1,   0,  -2,   1,   0,  -2,   2,  -2,   0,   1,  -2,  -1,   0,  -2,  -1,  -1,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   -4,  -5,  -1,  -4,  -3,   3,   2,  -1,  -1,   0,   1,   1,  -5,  -2,  -4,   2,  -3,   1,   1,   0,   1,   0,   0,  -1,  -5, -11,   4, -20,   3,  11,   8,   2,   1,   1,   0,   0,  -9,   4, -17,  14,  -6,  -7,   2,   0,   0,   0,   0,   0, -15, -32,   2, -27, -43,  -1,   7,   6,   8,  10,   0, -39, -14,  10,   5,   2, -63, -51,  -8,  14,  24,  13,   3,   3, -14,  -8,   9,  17,  12,   0,  -2,  -1,  11,  14,   6,   4,   6,   4,   1,   0,   4,   4,   0,   0,   3,   2,   1,   1 },
     {   -2,  -1,   2,   4,   4,   5,   2,  -2,   0,   1,  -1,  -1,  -2,   0,   9,  -1,   8, -18,  -5,   4,   0,   1,   0,   1,  -4,  -2,  -6,  19,   3, -21,  -5,   2,   0,   0,  -1,   0,   3,  -3,  33, -11,  12,   9,  -2,  -2,  -3,   0,  -1,   0, -12,   8, -20,  29,  47,   7,  -1,  -3,  24,  11,  23,  39,   5, -17,  -9,  -3, -61, -38,  19,  14,  -7,  -4,   1,   1, -16,  -6,   3,  -3,  -8,   4,   3,   2,   7,  10,   4,  -8,   4,   0,  -5,  -3,   1,   1,  -2,  -2,   0,   1,  -1,  -2 },
     {   -2,  -2,  -5,  -1,  17, -28, -29, -16,   2,   4,   4,   2,   2,   4,   4,   5,  -6,  49,  12,  -8,  -5,  -4,  -1,  -2,   1,  11,  -2,  25, -19,  76,  38,  -1,  -4,  -7,  -3,  -3,  -2,  -3,   2,  -5, -15,  -6, -11,  -2,   4,   1,   1,   0,  -1,   0, -13,  15,  12, -27, -16,  -1,   0,  10,  -8,  10,   8, -17,  -4,   2,  -9, -11,  -9,  -5,   1,  -5,   2,   4,   1,  -2,   5,   0,   1,  -4,  -1,   1,   3,   1,   6,  -1,   1,   0,  -1,  -2,   2,  -1,  -1,  -3,   1,   0,   1,  -1 },
     {   -1,   0,  -3,   6,   0,   2,   3,   5,  -1,  -1,   0,  -1,  -2,   0,  -6, -13,  10,  -5,   0,   3,   0,   0,   0,   0,  -2,  -1,  -8,  13,  12, -19,  -6,  -2,   0,   0,   0,  -1,  -6,   3,  -5, -32,   1,   6,   0,   3,   2,   1,   0,   0,  15, -17,   8,  36,  33,  10,  -5,  -4, -37,  36, -56, -31,  10,   1,   7,   0, -13,  -9, -49, -28,   2,   5,   9,   3,   9,   6,  10,  11,  -1,   1,  -1,  -1,   9,   1,   9,   6,   3,  -1,   5,   3,   0,   0,   1,   0,   0,  -1,   2,   2 },
     {    0,   2,   0,   0,   4,  11,  11,   0,  -3,  -1,   0,   0,  -3,  -2,  -7,  10, -29, -39,  -3,  -1,   3,   3,   1,   2,   3,   2,   4, -15,  41,   6,  14,  -1,  -2,   1,  -1,  -1,  -4,  -3,  -3,  10, -58, -22,  15,   3,   3,   2,   3,   0,  -7,  13, -23,  45, -36, -18,   1,   1,   1,  14,  -9,  32,  31,  10,   1,  -4,  -3,   4,  -5,  -9,   3,   3,  -2,   0,   1,   3,   3, -10,   3,   4,   0,   1,   0,   1,   1,  -3,   1,   1,   4,  -3,   2,   0,   1,  -1,   0,  -1,   1,   0 },
     {    1,  -4,  -5,   0,  -4,  -6,   2,  -7,   2,   3,  -1,   1,   4,  -2,   2,   9,   6,   1,  15,  -4,  -5,   2,  -1,   0,   1,  -6,  -8,  -9,  -1, -16,   6,  -3,   1,  -1,   0,   1,   5,  -5,   7,  14,  11, -12,  -7,  -4,  -2,   1,  -2,   1,   5,  -7, -21, -15,  -4,   9,   3,   2, -19,  21,  28,  33,   6,  -6,  -7,  -3,  49, -55, -58,  17,   8,   2,   2,   4,  15, -25, -21,   2,   0,   6,   0,   4,   8,   7,   9,   4,  -6,   3,   6,  -6,  -6,   3,   8,  -2,  -6,   1,   2,  -4 },
     {   -1,  -1,  -3,   2, -24,  53, -57,  -4,   1,  -1,   3,   3,  -2,  -1,  -7,  -1,   1,   3, -68,  -9,   3,   4,   2,   2,   2,   2,   0,   4,  -2,  15,   6,   7,   0,  -4,  -1,  -2,   3,   6,   1,  14,  -9,  34,  36,   1,  -4,  -5,  -3,  -3,   2,   0,  -1,   2,   4,   7,  22,  -6,  -2,   0,   3,   0,  -3,  -8,   6,   0,   4,  -5, -12,   3,   4, -14,   1,   4,  15,  -3,  -9,   3,   3,  -5,   4,   4,   8,  -1,   0,   1,  -5,  -3,   0,  -1,  -4,   2,   2,  -2,  -5,   0,   3,  -2 },
   },
   {
     { -118,  38,  13,   1,   3,   2,   2,   0,   1,   0,   1,   0,  25,  -6,  -5,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  14,  -5,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   31, 111, -36, -27,  -3,  -3,  -3,  -4,  -2,  -1,  -1,  -1,  -2, -20,   3,   6,   1,   1,   0,   1,   0,   0,   0,   0, -13, -16,  10,   5,  -1,   0,   0,   1,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,  -4,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,  -2,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {  -14,  14,  21,  -9,  -2,  -1,   0,  -1,   0,   0,   0,   0,-111,  24,  29,  -4,   2,   1,   2,   0,   1,   0,   1,   0,  32,  -5, -15,   2,   1,   0,   0,   0,   0,   0,   0,   0,  17,  -7,  -7,   2,   0,   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,   0,   0,   0,   0,   0,   5,  -1,  -2,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   2,  -1,  -1,   0,   2,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {  -25, -21, -97,  33,  30,   1,   3,   4,   3,   2,   2,   1, -25,  -8,  27,   2,  -7,   0,   0,   0,   0,   0,   0,   0, -20,  12,  36, -11,  -9,   1,   0,  -1,  -1,   0,   0,   0,  15,   1,  -9,  -1,   3,   0,   0,   0,   0,   0,   0,   0,   6,   0,   2,   0,  -1,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   2,   1,   2,  -1,  -1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   1,   0,   1,   0,   0,   0,   1,   0,   1,   0,   1,   0,   0,   0 },
     {    0, -11,  20,   2,  -6,   1,   0,   0,  -1,   0,   0,   0, -17,-112,  17,  39,   3,   4,   2,   4,   1,   2,   1,   2,   9,  16,  -6,  -7,   1,   0,   0,  -1,   0,   0,   0,   0,   4,  22,  -6, -10,   0,   0,   0,  -1,   0,   0,   0,   0,  -1,   3,  -1,   0,   0,   0,   0,   0,   1,   7,  -1,  -2,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   1,   4,  -1,  -2,   0,   0,   0,   0,   0,   2,   0,   0,   0,   2,   0,  -1,   0,   1,   0,   0,   0,   2,   0,  -1 },
     {    8,  -5, -45,   3,  12,   1,   2,   1,   1,   1,   1,   0,  10,  -4, -20,   9,   3,   0,   0,   0,   1,   0,   0,   0, 103, -17, -28,   5,  -5,  -1,  -3,   0,  -1,   0,  -1,   0, -28,   1,  17,  -2,  -2,   1,   0,   0,   0,   0,   0,   0, -19,   6,  12,  -3,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -6,   1,   4,  -1,   0,   0,   0,   0,  -3,   1,   2,   0,   0,   0,   0,   0,  -3,   1,   2,   0,  -1,   0,   1,   0,  -1,   0,   1,   0,  -1,   0,   1,   0 },
     {   -5,   5,  -1,  23,  -3,  -5,   0,  -1,   0,  -1,   0,   0, -38,  -3,-101,  21,  35,  -2,   5,   2,   4,   1,   2,   1, -15,  12,  23, -13,  -5,   2,   0,  -1,   0,   0,   0,   0,  -5,   0,  31,  -7, -11,   2,   0,  -1,  -1,   0,   0,   0,  10,  -2,  -4,  -1,   1,   0,   0,   0,   3,   0,   6,  -1,  -1,   0,  -1,   0,   3,  -1,   1,  -1,   0,   0,   0,   0,   2,   0,   4,   0,  -1,   0,   0,   0,   2,  -1,   0,  -1,   1,   0,   2,   0,   1,   0,   0,   0,   1,   0,   1,   0 },
     {   11,  31,   8,  73, -20, -23,  -4,  -3,  -3,  -2,  -2,  -1,   8,   4,  14, -19,  -6,   4,   0,   0,   0,   0,   0,   0,  12,  61, -28, -52,  12,   6,  -1,  -2,   0,   0,   0,  -1,  -8, -16,  -3,  11,   3,  -1,   1,   1,   0,   0,   0,   0,  -5, -13,   6,   3,  -2,   1,   1,   1,   0,  -1,  -1,  -1,   1,   0,   0,   0,  -2,  -5,   1,   0,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 },
     {    5,   5,  12,  76, -17, -20,  -3,  -3,  -2,  -2,  -1,  -1,  -2,  -6,   6,  -3,  -3,   5,   0,   0,   0,   0,   0,   0,  -6, -91,   5,  11,   7,  11,   1,   5,   1,   2,   1,   2,   6,  13,  -4,  -8,   0,  -1,  -1,   0,   0,   0,   0,   0,  -4,  17,  -5, -11,   2,   1,   1,  -1,   0,   1,   0,   1,   0,  -1,   0,   0,   0,   6,  -1,  -4,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   3,   0,  -2,   0,   1,   0,   0,   0,   2,   0,  -1,   0,   1,   0,   0 },
     {    5,  -3, -10,  -1,  -7,   3,   1,   0,   0,   0,   0,   0,  11,  -5, -32,  -4,   7,   4,   1,   1,   1,   1,   1,   0,  10,   2, -24,   8,   8,  -3,   0,   0,   0,   0,   0,   0, 103, -15, -39,  11,  -1,  -4,  -3,   1,  -1,   0,  -1,   0, -23,   0,  19,  -2,  -4,   0,   0,   0, -18,   7,  13,  -4,  -1,   1,   0,   0,  -3,   1,   1,  -1,   1,   1,   0,   0,  -5,   1,   5,   0,  -1,   0,   0,   0,  -3,   0,   2,   0,  -2,   1,   2,   0,  -2,   0,   1,   0,  -1,   0,   1,   0 },
     {    9,   3,  25,  -2,  14, -10,  -5,   0,  -2,  -1,  -1,   0,   5, -19,  -4, -55,   6,  19,   1,   3,   0,   2,   1,   1,  40,   2,  77, -10, -43,   2,  -2,  -2,  -4,   0,  -2,  -1,   7, -11, -18,  23,   6,  -5,  -1,   1,   0,   0,   0,   0,   2,   1, -24,   4,   9,   0,   0,   0,  -8,   3,   3,   5,  -2,  -1,   0,   0,  -4,   0,  -6,   1,   2,   0,   0,   0,  -3,   2,  -1,   2,   0,  -1,   0,   0,  -2,   0,  -3,   1,  -1,   1,   0,   1,  -1,   0,  -1,   0,  -1,   1,   0,   1 },
     {   -6,  -6, -16,  -4,   5,   4,   1,   2,   1,   1,   1,   1, -13, -36, -11, -84,  16,  29,   3,   5,   2,   3,   2,   2, -23,  -3, -40,  29,  23, -10,   0,   1,   2,   0,   0,   0, -16,  -1,  24,  31, -13,  -9,   1,  -1,  -1,  -1,   0,   0,   6,   5,  11,  -7,  -7,   3,   0,  -1,   7,   4,  -3,   6,   1,  -2,  -1,   0,   3,   2,   3,  -1,  -2,   0,  -1,   0,   3,   2,   0,   3,   0,  -1,   0,   0,   1,   1,   1,   0,   1,   1,   0,   1,   1,   0,   1,   0,   1,   1,   0,   1 },
     {   -6,  -9, -23,  -8,-102,  26,  34,   0,   4,   3,   2,   0,   0,   0,   7,  -3,   8,   8,  -7,   0,  -1,   0,   0,   0,  14,   2,  30,   2,  30, -13, -17,   1,  -2,  -1,  -1,   0,  -4, -11,   0,   2,  -5,   0,   4,   0,   0,   1,   0,   0,  11,   0, -14,   3,   9,  -1,  -1,  -1,  -3,   2,   0,   1,   0,  -2,   0,   0,  -3,   1,   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    0,  -7,   3,   0,   9,   0,  -2,   0,  -1,   0,   0,   0,   2, -15,   6,  28,  -9,  -8,   0,   1,   0,  -1,   0,   0,  -6, -18,  -4,   0,  -2,   7,   2,   1,   0,   1,   0,   0,  -5,-105,  20,  41,  -3,   7,   0,   4,   0,   2,   0,   2,  19,   7,  -7,  -3,  -2,  -1,  -1,   0,  -4,  24,  -3, -14,   4,   1,   0,  -1,  -4,   5,   1,  -1,   0,  -1,   0,   0,   1,   7,  -1,  -4,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   3,  -1,  -1,   0,   1,   0,  -1,   0,   2,   0,  -1 },
     {   -5,   1,  -1,   3,  -8,   7,  -1,  -1,   1,   0,   0,   0, -15,  -4, -33,   4, -43,   8,  19,   0,   4,   1,   2,   0,  -9,   5,   2,  17,   5, -11,   0,   0,   0,  -1,   0,   0, -45, -11, -66,  12,  57,  -5,  -1,   2,   3,   0,   2,   1, -20,  14,  20, -13,  -6,   1,   2,  -1,   6,   2,  22,  -4,  -6,   1,  -2,   0,  12,  -3,  -3,  -2,   1,   0,  -1,   0,   4,   0,   5,   0,   0,  -1,  -1,   0,   3,   0,   0,  -1,   1,   0,   2,   0,   1,   0,   1,  -1,   1,   0,   1,   0 },
     {    2,  -5, -11, -15,  10,  -3,   5,   2,   2,   1,   0,   0,   3,  -1, -11,  -4,  20,   0,  -5,   0,   1,   0,   0,   0,   4, -23, -31, -42,  13,  22,   1,   2,   2,   2,   1,   1,  -9,  12, -43,   8,  18,  -7,   3,   1,   1,   0,   0,   0,  72,  -7, -38,  23,  -3,  -9,  -2,   1, -15,  -2,  24,  -1,  -9,   3,   1,  -1, -16,   6,  16,   0,  -3,   0,  -1,   0,   1,   0,   2,   1,  -2,  -1,   0,   0,  -2,   1,   3,   2,  -2,   0,   3,   1,  -1,   0,   2,   1,   0,   0,   1,   0 },
     {    6,   7,   1,  16,   4,   4,  -4,  -3,  -1,  -2,   0,  -1,   0,  -1, -14,   0, -76,  10,  21,   1,   3,   1,   2,   0,  13,  19,  -4,  40,   3, -18,  -5,  -3,   0,  -1,  -1,  -1,  23,  11,  32, -14,  10,  -2, -10,   0,  -3,  -1,  -1,   0,  53, -11, -35,  -5,   3,   2,   0,   1, -16,  -3,   0,   4,  10,  -1,  -1,   0, -13,   2,   8,  -5,  -1,   1,   1,   0,  -3,  -1,  -1,  -1,   5,   0,  -1,   0,  -3,  -1,   1,  -2,  -2,   0,   0,   0,  -1,  -1,   0,  -1,   0,   0,   0,   0 },
     {    4,   8,   2,  28,   7,  -8,  -4,  -3,  -1,  -1,  -1,  -1,   7,   5,  14,  11,  59, -13, -21,   0,  -2,  -2,  -2,   0,   6,  28,  -5,  68, -13, -34,   4,  -4,   0,  -2,  -1,  -2, -12, -11, -34, -13,  -6,  13,   9,   1,   2,   1,   1,   0,  28,   4, -14, -27,   5,   9,  -2,   0,  -6,  -3,  10,   2,  -9,   0,   2,   0,  -6,  -4,   4,  -8,  -1,   4,   0,   1,   0,  -1,   1,  -1,  -3,   0,   1,   0,  -2,  -1,   1,  -3,   0,  -1,   1,   0,  -1,  -1,   1,  -1,   0,   0,   0,  -1 },
     {   -1,  -4,   2, -19,  11, -72,  11,  24,   1,   1,  -1,   1,   1,  -4,  -1,  -8,  -2, -12,   3,   5,   1,   0,   0,   1,   9,   8,  25,   9,  59,  23, -33,  -9,  -3,  -2,  -2,   0,   2, -12,  -2, -28,   2,  23,  -3,   1,   0,   1,   0,   1,   9,  16,  12,  -9, -36,   5,   9,  -2,  -6,  -2,  -2,  11,  -1,  -1,   1,  -1,   1,  -3,  -6,   1,  -3,   3,   2,  -1,  -2,   1,   0,   2,  -1,  -1,   0,  -1,   1,   0,  -3,   1,  -1,   0,  -1,   1,   0,   0,  -1,   0,  -1,   1,   0,   0 },
     {   -4,  -2, -11, -15, -26, -20,   7,  12,   2,   1,   1,   1,  -1,  -3,  -1, -15,  -5,   1,   7,   3,   0,   0,   0,   0, -10,   9, -27, -13, -48,  22,  22,  -5,   2,   1,   2,  -1,   7, -14,  10, -34,  13,  18,  -6,  -1,   0,   1,   0,   1,   2,  70, -12, -28,  32, -10,  -8,  -1,  -4, -12,   3,  22,  -4,  -6,   2,   1,   0,  -9,   9,   5,   2,   1,  -3,   0,   1,   2,  -1,   3,   1,  -1,  -1,   0,  -1,  -2,   2,   4,   0,   0,   0,   1,   0,  -1,   1,   1,   1,   0,   0,   0 },
     {   -5,  -4,  -8, -10, -20, -63,   4,  24,   5,   1,   1,   1,   1,  10,   3,  30,   0,  22,   0, -12,  -1,  -2,   0,  -1,  -2,   9,  -4,  22, -15,  18,  10, -13,  -2,  -1,   0,  -1,  -1,  29,   6,  62,  -6, -46,  -3,   1,   0,  -2,  -1,  -2,  -2,  -2,  -8, -11,  14,   6,  -2,   0,   0,  -1,   0, -27,   2,   8,  -1,   1,   0,  -5,   2,   2,   2,   2,  -1,   0,   1,  -3,  -1,  -6,   1,   1,   0,   1,   0,  -3,   0,  -1,   0,  -1,  -1,  -3,   0,  -1,   0,  -1,   0,  -1,   0,  -1 },
     {   -4,  -9,  -7,  -5, -19, -39,  10,  13,   4,   1,   0,   0,  -2,  -7,  -4, -12,  -2,  12,   4,  -2,   0,   1,   0,   0,  -7,  -9, -15,  20, -34,   8,  12,  -3,   1,   0,   1,  -1,  -5, -21,  -5, -40,  11,  19,  -2,   3,   0,   1,   1,   1,  -9, -73,   7,  45,  18,   5,  -6,   2,  14,  17,   0,   3,  -6,  -7,   0,   0,   0,  17,   3, -12,   3,   2,  -2,  -1,  -1,   3,   2,   3,   0,  -2,  -1,   0,   2,   4,   1,  -2,   0,   2,   1,   1,   1,   1,   0,  -1,   0,   0,   1,   1 },
     {    1,   0,   1,   3, -21,   2,  14,   0,  -2,   0,   0,  -1,  -2,  -8,  -3, -18, -10, -79,   3,  25,   5,   2,   0,   1,   3,  -2,   3,   8, -34,  17,   9,  -8,   0,   0,   1,   0,  -1,  19,  -1,  29,  -3,  16,  -3, -11,  -1,  -3,   1,  -1,  34,  -8,  53,  -3, -18,  -1,  -8,   2,  -6,  -4,  -4,  -5,  14,  12,   0,  -2,  -6,   9,  -6,  -1,   8,  -2,   0,   0,   0,  -2,   0,   0,   0,   5,   0,  -1,  -1,  -1,  -1,   1,  -1,  -1,  -1,  -1,  -1,   0,  -1,  -1,   0,  -1,   0,   0 },
     {   -4,   1,   1,   0,   1,   1,   1,   1,   1,  -1,   1,   0,  -7,   1,  10,  -4,  -2,   0,  -5,   2,  -1,   0,   0,   0,  -4,   1,   1,   1,   0,   3,   3,   0,   0,   0,   1,  -1, -14,   5,  20,  -4, -10,   9,   2,  -4,  -1,  -1,   0,  -1, -18,  -6,   4,   4,   5,   3,  -1,   0, -92,  15,  65, -10, -11,   1,   4,  -1,  18,   6, -13,  -4,   1,  -2,   1,   0,  23,  -6, -20,   4,   5,  -2,  -1,   0,   4,   0,   0,   1,   3,  -1,  -3,   0,   0,   0,   1,   0,   0,  -1,  -2,   1 },
     {    1,   0,   3,  -2, -11,   3,  -9,  -1,   3,   1,   0,   1,   4,   7,   4,  13,   3,  64,  -6, -20,  -3,  -2,   0,  -1,   3,  -4,   5, -10, -26,  -6,  11,   4,   0,   1,   1,   0,  -2,  -5,  -2, -22,  -8, -14,   6,   8,   1,   0,   0,   1,  46,   5,  64, -10, -45,   3,  -2,  -3, -15,   4,  -7,   6,  10, -10,  -1,   2,  -6,  -1, -17,   6,  14,  -2,  -1,   0,   0,  -2,   0,   2,   0,  -3,   0,   1,  -2,  -1,  -2,   0,   1,   0,  -1,  -1,  -1,  -1,  -1,   1,   0,   0,   0,   0 },
     {   -1,   2,  -1,   5,   0,  11, -16,   2,   2,  -2,   1,   0,  -4,   0, -10,  -1, -25,   6,  -5,   1,   5,   1,   1,   0,   0,  10,   1,  21,   4,  42,   1, -20,  -2,  -2,   1,  -2, -15,  -5, -36,  -9, -70,  -2,  44,   1,   2,   2,   3,   0,  -5,  15,  -3,  25,  10, -39,  -1,   4,  -3,   4,  -5,   1,  30,  -2,  -9,  -1,  -2,   0,   5, -13,  -1,  -1,   1,   4,   1,  -2,   6,   2,   5,   1,  -2,   0,   1,   0,   0,  -2,  -1,  -1,   2,  -1,   0,   0,   1,  -1,   1,   0,   2,   0 },
     {    2,   5,   4,   8,   6,  32,  11, -10,  -9,  -1,   0,   1,   2,   6,   5,  -2,  18,  13, -11,   6,  -3,  -2,  -1,   0,   2,  12,   5,  24,   5,  70, -12, -26,  -1,  -4,   0,  -1,   5,   7,  16, -10,  44, -10, -14,   3,  -1,  -1,  -3,   1,   0,  -8,   7, -21,   3, -32,   3,  12,   3,  46, -11, -16, -17,  -5,   5,  -7, -16,  -4,   1,  -1,  -4,  -6,   3,   2,   2,  -8,  -7,   1,  -5,   0,   2,   0,  -1,  -2,  -2,  -1,   1,  -3,  -1,   1,   0,  -1,  -3,   0,   0,   0,   0,   0 },
     {   -5,  -5, -10,  -5, -29,   3,-107,  16,  29,  -3,   5,   3,   0,   0,   3,  -3,   8, -10,   6,  -3,  -4,   5,  -3,   1,   5,   1,   7,   2,  14,   6,  22,  -9,  -9,   3,  -2,   0,   2,   4,   6,   5,  10,  18, -21,  -3,   2,  -4,   0,  -1,   3,  -3,  -3,   4,  -7,  -7,  12,  -1,   4,   5,   5,   2, -10,  -3,   3,   0,  -5,   4,  -2,   1,  -1,  -2,   4,   0,   0,   3,   0,  -3,  -1,   1,   1,  -1,   0,  -2,   0,  -1,   0,  -1,  -1,  -1,  -1,  -1,   0,   2,   0,  -1,   0,  -1 },
     {   -2,  -3,  -5, -12,  -2,  -4, -10,  -1,   2,   3,   2,   0,  -1,   2,  -6,  -9, -12,  -6,   6,   6,   0,   1,   1,  -1,  -1,  -7,  -4, -23,  -1,  -2,  -2,   5,   3,   0,   1,   1,  -6,   8, -19, -15, -26,   9,  15,   1,   0,   2,   1,   0,   1, -26, -12, -66,  16,  44,   5,   1,   4,  58,  -6, -20,  22,  -7,  -8,  -2, -11,  -9,   3,  31,  -3, -12,   2,   1,   4,  -8,   4,   6,  -1,   3,  -1,  -1,   2,   1,   1,   3,  -2,  -2,   2,   3,   1,   1,   0,   3,   1,  -1,   0,   2 },
     {    0,  -5,  -1,  -1,  -2,  18,   0,   2,   0,  -4,   0,   1,  -1,  -8,   0,   6,  -1,   7,  -4,  -3,   3,  -1,   2,   1,   0,  -5,  -1,   2,   1,  42,   7, -16,   0,   0,   0,   0,  -1, -16,  -2,   6,  -1,  -4,   7,   4,  -1,  -1,   0,   1,  -7, -49, -12, -51,  -8,  15,   8,  13,  -7, -70,   9,  43,   4,   2,   1,   4,  13,   3,  -5,  10,   2,  -6,  -4,   3,  -9,   9,  -4,  -8,   4,   0,   2,   1,  -1,   7,   2,   1,   0,   3,   1,  -2,   0,   2,   2,   1,   0,   2,   0,   0 },
     {    2,  -1,   2,   1,  -1,   2,   7, -20,  -8,  -1,  -1,  -1,   7,   3,  13,   3,  11,   2,  79,   2, -19,   4,  -4,  -2,   3,   1,   5,   4,  -2,   9,  -7,  -1,   4,  -1,   1,   0,   1,  -4,  -2,  -4, -36,  -6, -22,   5,  11,   0,   1,   1,   7,  -2,  12,  -2,  -2,  -5,   3,   1,  37,  -1,  53,   1, -26,  10, -15,  -3, -13, -16,  10,   5,   0,   0,   0,   2,   8,   8, -19,  -4,   5,  -2,  -6,  -1,   1,  -1,  -6,   1,  -3,   2,  -3,  -1,   0,   0,   0,   0,  -1,  -1,   0,  -1 },
     {   -3,   0,  -1,   0,   1,   1,   3, -10,  -3,   0,   1,   3,  -2,   1,   3,   2,  26,  -1,  62, -16, -20,   4,  -1,  -3,  -4,   1,   1,   2,   5,   2,   3,   4,  -3,   0,   1,   0, -10,  -1, -10,   3,   9,   1, -24,   8,  11,  -1,   0,   2, -13,  -3,  -3,  -1,   4,  -2,  -1,  -1, -52,  -9, -50,   5,  43,  -6,  -9,   6, -18,  23,  20, -15, -10,   2,   1,   1,   5, -14,  13,   4, -17,   3,  -1,  -1,   9,  -1,  -2,   3,   1,   1,   3,  -1,  -1,  -1,  -2,  -1,   2,   1,   0,  -1 },
   }
 },
 {  //2
   {
     {  114, -45,  -9,  -5,  -2,  -2,  -1,  -1,  -1,   0,   0,   0, -32,  -3,   5,   2,   2,   1,   1,   1,   0,   0,   0,   0,  -9,   7,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -42, -42,  27,   6,   6,   2,   2,   1,   1,   0,   1,   0, -94,  23,  32,   3,   3,   0,   1,   0,   1,   0,   0,   0,  15,  29,   1,  -8,  -4,  -2,  -1,  -1,  -1,  -1,   0,   0,  10,   6,  -8,  -5,  -1,   0,   0,   0,   0,   0,   0,   0,   9,   2,  -6,  -1,   0,   0,   0,   0,   5,  -1,  -4,   0,   0,   0,   0,   0,   4,  -1,  -2,   0,   0,   0,   0,   0,   3,  -1,  -2,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   1,   0,  -1,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   19,  27,  11, -18,  -2,  -4,  -1,  -2,  -1,  -1,   0,  -1,  -2,  93, -19, -25,  -5,  -5,  -2,  -3,  -1,  -1,  -1,  -1,  35,  -3, -46, -10,   3,   2,   1,   1,   1,   1,   0,   0,  -7, -27, -15,  13,  10,   4,   2,   1,   1,   1,   1,   1,  -1, -11,   2,  12,   2,  -1,  -1,   0,  -3,  -7,   4,   5,  -1,  -1,   0,   0,   0,  -3,   3,   3,  -1,   0,   0,   0,  -2,  -4,   2,   1,  -1,   0,   0,   0,   0,  -2,   1,   1,  -1,  -2,   1,   1,   0,  -1,   1,   0,  -1,  -1,   1,   0 },
     {   12,  92, -28,  -9,  -9,  -3,  -3,  -2,  -2,  -1,  -1,  -1, -66,  -2,  -3,   0,   5,   3,   3,   1,   1,   1,   1,   0, -43,  -6,  14,   8,   6,   2,   2,   0,   1,   0,   0,   0,   0,   6,   7,   1,  -2,  -2,  -1,  -1,   0,   0,   0,   0,   3,   0,   1,  -2,  -1,  -1,   0,   0,   6,  -1,  -1,  -1,   0,   0,   0,   0,   3,  -2,  -1,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 },
     {   14,  30,  15, -10, -12,  -2,  -3,  -2,  -2,   0,  -1,   0,  29,  -5,  55, -20, -13,  -3,  -3,  -2,  -1,  -1,  -1,  -1,  -7,  61,  16, -43, -12,  -3,  -1,  -2,   0,  -1,   0,  -1,   6,   8, -39, -31,   5,   8,   4,   2,   2,   1,   1,   0,  -9, -10, -22,   2,  19,   7,   2,   1,  -1,  -5,  -9,  10,  10,  -2,  -2,   0,  -3,  -5,  -3,   7,   3,  -3,   0,   1,   0,  -2,  -2,   5,   1,  -1,   1,   0,  -1,  -2,  -1,   3,   0,  -1,  -1,   2,  -1,  -1,   0,   1,   0,   0,  -1,   1 },
     {    6,  17, -59,  13,   4,   2,   0,   2,   0,   1,   0,   0, -11, -28, -20,  23,  10,   1,   1,   1,   1,   0,   0,   1,  84,  27,  -4,   5,  -8,  -7,  -5,  -2,  -2,  -2,  -1,  -1,  39,  -4, -18,  -8,  -9,  -2,   0,   0,   0,   0,   0,   0,   3, -16,  -6,   1,   1,   3,   2,   1,  -2,  -7,   4,   4,   0,   1,   1,   0,  -5,  -4,   6,   1,   0,   1,   0,   0,  -3,  -1,   4,   0,  -1,   0,   0,   0,  -2,  -1,   3,  -1,  -1,   0,   1,  -1,  -1,   0,   1,  -1,   0,   0,   0,   0 },
     {  -15, -17, -41,  -2,  11,   9,   3,   3,   1,   2,   1,   1,  -3, -28,   1, -16,  19,   5,   3,   1,   2,   0,   1,   0, -37,  33, -46, -17,  23,   3,   4,   0,   1,   0,   1,   0,  -8, -23, -32,  37,  32,   3,   0,   0,   0,   0,   0,   0,  -2,   4,  19,  39,   3, -15,  -7,  -3,   5,   4,  11,   8, -17, -12,  -1,   1,   2,   2,   7,   0, -10,  -1,   3,   1,   3,   1,   2,  -1,  -5,   1,   1,   0,   0,   1,   3,  -1,   1,   0,   1,  -1,   0,   1,   1,   0,   0,   0,   0,  -1 },
     {  -10, -21, -37,  -5,  10,   3,   5,   2,   3,   1,   1,   1,  18,  50, -46,   1,   8,  -3,   0,  -1,   0,   0,   0,  -1, -42,  57,  13,  15,   2,  -9,  -3,  -4,  -1,  -1,   0,  -1,  17,  44,   6,  -6, -16, -12,  -3,  -2,  -1,  -1,   0,  -1,   3,   1, -16, -16, -11,  -1,   3,   2,   0, -10,  -8,  -1,   3,   8,   4,   1,   2,  -7,   0,   4,   3,   3,  -1,  -1,   2,  -5,   1,   2,   0,   0,  -1,   0,   1,  -4,   1,   1,   1,  -3,   1,   0,   0,  -2,   1,   0,   0,  -2,   0,   0 },
     {    2,   6, -29,  52,  -4,  -7,  -4,   0,  -1,   0,  -1,   0,  13,  29,  55,  36, -13, -21,  -7,  -5,  -2,  -2,  -1,  -1,  -9,  20,   2,  12, -19, -17,   0,  -1,   0,   0,   0,   0, -21, -35,  15,  21, -18,  -4,   4,   3,   1,   1,   0,   1, -11,  -2,  26,  -8, -25,  -1,   5,   2,  -3,  -2,   3, -19, -12,  11,   7,   1,   0,   1,   1, -12,   1,  11,   1,  -2,   0,  -1,  -4,  -8,   3,   5,  -2,  -1,   0,   0,  -1,  -4,   0,  -1,  -2,  -2,   0,   0,  -1,  -1,  -1,  -1,  -1,  -1 },
     {  -15, -16, -43,  -3,   4,   9,   5,   4,   2,   2,   1,   1,  -9,  21, -17,  12,  -3,   7,   1,   1,   0,   0,   0,   0,   4, -12,  44, -16, -16,   4,  -2,   0,  -1,   1,  -1,   0, -64, -16,  -4, -37,  10,  18,   7,   3,   3,   2,   2,   1, -27,  -6,   4,  13,  35,  15,  -2,  -2,  -1,  16,  17,  16,  11, -12, -14,  -5,   3,   6,   2,   0,  -5, -13,  -3,   2,   0,   3,   0,  -1,  -2,  -4,   3,   2,   2,   1,  -1,   0,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    3,  -4,  21,  47,   3, -10,  -6,  -3,  -1,  -1,  -1,  -1,  -3,   3, -27,  43,  -6,  -9,  -3,   0,   1,  -1,   0,   0, -44, -19, -24,   5, -15,   1,   6,   4,   3,   2,   1,   1,  36, -26, -41, -22,  -6,  16,  10,   5,   2,   1,   1,   1,  13, -27, -19,   4,  20,  20,   4,   1,  11,  -5,  10,  16,  12,   1,  -8,  -5,   3,  -1,   9,   3,  -3,  -7,  -6,  -1,   1,  -1,   7,  -2,  -3,  -2,   1,   2,   1,   1,   4,  -3,   1,   1,   3,  -2,   1,   0,   1,  -1,   1,   0,   1,  -1 },
     {   -8, -19, -58,  12, -18,  14,   7,   6,   2,   2,   1,   1,  -1,   9,  30, -48, -23,   7,   6,   1,   1,   0,   1,   0, -10, -36, -22, -19,  -5,  17,   9,   4,   2,   1,   1,   1,  35, -10,  21, -13,  -5,  15,   2,   0,  -2,   0,  -1,   0,  20,  13,  -7, -28,   8,  12,   0,  -2,   6,  -6, -13,   1,  22,   8,  -3,  -2,   1,   4,  -2,   8,  14,  -3,  -7,  -3,   0,   0,  -1,   9,   6,  -6,  -4,   0,   0,   2,   2,   5,   0,   1,   0,   3,   0,   2,   1,   1,   0,   0,   0,   1 },
     {   -7,  -3,  15, -18, -10,   1,   5,   1,   1,   0,   1,   0, -11, -25, -10, -29,  -1,   4,   9,   3,   3,   1,   2,   1,  -6,  32, -25,  34,   0,   0,   2,  -2,  -1,  -1,   0,  -1, -10, -34,  25,   7, -32,   5,   1,   2,   0,   2,   0,   1, -52, -37,   0, -23,   4,  26,  13,   6, -19,  -5,  13,  12,  25,  19,  -2,  -4,  -2,  10,  13,   5,   9,  -4, -13,  -5,   0,   5,   5,   0,   1,  -7,  -5,   1,   2,   3,   2,  -1,   2,   3,   0,   0,   1,   2,   0,  -1,   1,   1,   0,   0 },
     {    3,  13,   5,  49,  -4,  -9,  -5,  -3,  -3,  -1,   0,   0,  -9, -15, -18,   4,  -5,  -5,   3,   2,   1,   0,   1,   1,  11,  12, -62, -20,  -4,   6,  12,   3,   3,   1,   1,   0, -48,  53,   3, -15,   0,   7,   7,  -1,   1,  -1,   1,  -1, -15,  39,  -2, -22,   4,   5,   1,  -3, -12,   9,  -5,  -2,   9,   4,   0,  -3,  -2,   1,   0,   1,   5,   0,  -4,  -1,  -2,  -2,   1,   4,   3,  -1,  -3,   0,   3,  -1,   0,   2,   0,  -1,   0,   1,   1,  -1,   1,   0,   0,   0,   1,   0 },
     {   -6,  -7,   2, -33, -37,  -9,  13,   5,   2,   1,   1,   1,  -9, -14, -37,  11, -57,  -4,  14,   8,   3,   3,   1,   1,  -4,  10,   5, -33, -43,   2,  16,   6,   2,   1,   1,   1,   0,  -2, -15,  13,  -6,   9,  14,   0,   0,  -1,   1,   0,  13,   2,  29,  13, -24,   9,   4,  -4,   5,   4,   6, -27,  -9,  21,   7,  -2,   0,  -3,  -1,  -6,  13,  19,  -2,  -5,   1,   4,   2,  -1,  12,   6,  -9,  -5,   0,   0,   2,   2,   0,   0,   0,   0,   0,  -1,   1,   2,   1,   1,   1,   0 },
     {    5,   6,  18,  35, -27,   4,  -3,  -2,  -2,   0,  -1,   0,   1,   9, -16, -10, -21,  17,   4,   0,   0,   1,   1,   0,   1,  31,  12,   0,  -2,  21,   0,  -3,  -2,   0,   0,   0,  42,   0,  41,  -1,  10,  12, -14,  -8,  -5,  -1,  -2,  -1, -14,  11,  39,   8,  20,  -9, -22,  -9,  -1,  20,  28,   8,  -8, -34, -18,  -1, -14,  -1,  -2, -15, -15, -16,   4,  12,  -4,   0,  -4,  -4,  -3,   1,   9,   5,  -3,  -1,  -5,  -3,   0,  -3,  -6,  -2,  -2,  -2,  -4,  -2,  -2,  -1,  -2,  -2 },
     {   -4,  -7,  -8,  28, -15,  -4,   0,   3,   1,   0,   0,   1,  -7,   8,   7,  -8, -23,   0,   1,   4,   1,   1,   1,   1,  -2, -32,  20, -16,  18,   7,   5,   2,   0,  -1,   1,   0,  13,  43, -18,  20,  20, -14,  -4,  -7,  -2,  -2,   0,  -1, -51, -40,  -9,  18, -17, -12,   3,   4, -40, -27,  16,  10,  -3,   8,  14,   5, -10,   6,  19,   6,  -2,   7,   1,  -4,  -5,   5,   8,  -3,  -2,   3,  -4,  -2,   2,   4,   2,  -3,   0,   3,   1,  -2,   1,   1,   1,  -1,   1,   1,  -1,  -1 },
     {    0,   3,   4,  -7,  56,  -3, -10,  -5,  -4,  -3,   0,   0,  -5,   1, -12,  -7,  33, -10, -11,  -1,   0,   0,   0,   1, -12, -11,   1, -50, -13, -11,   0,   9,   3,   3,   1,   1,  34, -18,   6, -31,  -9,  -2,   9,   9,   1,   2,   0,   0, -36,  18,  36, -12, -10,  -3,   9,   3, -28,   6,   5, -16,  -7,   3,  11,   1, -22,   4,   0,  -1,   8,  10,   8,  -2,  -4,   6,  -2,   5,   8,   3,  -1,  -4,  -2,   4,  -3,   5,   0,   2,  -2,   3,  -1,   0,  -1,   2,   0,   1,   1,   1 },
     {    4,   1,  12,  57,   2,   0,  -9,   1,  -5,  -1,  -2,   0,  -5, -12, -38, -46,   5,   1,   5,   6,   2,   1,   2,   2,   4,  15,  30, -34,  12,  -5,  -4,   2,  -2,   0,   0,   0, -24, -36,  21,   2,   6, -14,  -3,   2,   2,   2,   2,   1,  27, -26, -23,   0, -12,  -6,  10,   8,  16,  -8, -13,  -3, -12,   6,  13,   6,   8,  -6, -11,   7,   5,  14,  11,   0,  10,   3,  -1,   7,   2,   4,  -2,  -7,  -5,   1,   0,   2,   1,  -2,   3,   4,   0,   0,   1,   2,   0,   1,   2,   1 },
     {    6,   7,   8,   5,  30,  21,   6,  -7,  -3,  -2,  -1,  -2,   4,   5,  -2,  31,  -4,  33,   0,  -6,  -8,  -1,  -2,  -1,   4,   8,  15, -39,  32,  20,   3, -10,  -2,   0,   0,  -1,  -2,   0,  -6,  39,   7,   2, -10,  -9,   0,  -1,   0,  -2,  13, -13,  25, -26, -18,  21,  -7,  -3,   0,   8,  -6,  -7,  41,  18, -16,  -8,   9,  -2,   5,  13,  11, -23, -30,  -4,  -4,  -1,  -7,  -4,  -2, -19,  -4,  11,  -1,  -3,   0,   2,  -2,   1,   0,  -1,  -1,  -2,   0,   2,  -2,  -1,  -1,  -2 },
     {    4,   3,  -2,   8,  27, -14,  -3,  -2,  -1,  -1,   0,  -1,   3,   0,  12, -41,   7, -17,   0,   2,   2,   0,  -1,   0,   6,  -9,   6,  -5, -16, -18,  12,   4,   3,   1,   1,   0,  -1,  35,  -4,  12, -36, -10,  14,  -3,   0,  -2,   1,  -1,  25, -14,  16,  16, -19,  23,   9,  -5,  27,  31,  49,  33,  -3,   7, -10, -14,  10,  11,  14,  -8, -18,  -9, -16,  -2,   0,   0,  -3,  -7,  -5,   1,  -2,   6,   3,  -1,  -3,  -1,  -2,  -3,  -6,  -3,  -1,  -1,  -3,  -2,  -3,  -1,  -1,  -2 },
     {    1,   2,   3,  -8,  22,   7,   4,  -1,  -2,  -1,   1,  -1,   0,  -7,   9,  11,   1,   6,   2,  -1,  -2,  -1,   1,   0,  -5,   4, -10, -10, -33,   2,   4,   2,   3,   0,   1,   0, -18,   8,  39,   0, -12,   8,   1,   1,  -1,   0,  -1,   0,  44, -15,  15,  20,   7,  -5, -15,  -4, -36, -62,   2,  29,   1, -10,  -6,   5, -19, -30,  21,  22,  -1,  -1,   5,   6,  -2,  -2,  19,   6,  -5,  -1,   1,  -1,  -3,   0,   7,  -4,   0,   0,   1,  -4,   3,   2,   0,  -3,  -1,   0,  -2,  -3 },
     {    1,   1,   2,   5,  -5,  51,   3,  -5,  -5,  -1,  -1,  -1,   1,   3,   6,  -7,   2,  57,   2, -10,  -5,  -2,   1,  -2,  -6,  -7, -15,   8,  -1,  21, -10,  -6,   0,  -1,   1,  -1, -17,   6, -36, -32, -40, -13, -13,   3,   9,   3,   3,   2,  10, -21,   9, -12, -23,  -8,  -5,  11,  -1,   4,  25, -18, -20, -11,   1,  11,  -7,  -5,   0, -20,   3,   9,  20,  11,  -2,   3,   4,   0,  16,  11,   3,  -5,  -3,   4,   2,  -1,   0,   0,   2,   2,  -1,   1,  -1,   0,  -1,  -1,   1,   1 },
     {   -4,  -7,  -9, -15,   0, -24, -15,  11,   6,   0,   4,   2,  -4,  -6,  -5,   4, -28, -19, -21,  14,   4,   2,   2,   1,  -2,  -6,  -8, -22,  19, -18,  -7,   9,   4,   2,   2,   1,  -6,   3,  -7,  23, -27, -20,  -8,   8,   1,   0,   3,   2,   2, -20,   9, -36,  10,   9,  -7,  12,   2,   1, -11,   0,  20, -42, -10,  15,  -1,  -6,  -6,  -5, -23, -20,  41,  24,   1,   0,   0,   6,   7,  18,  33,  -7,   7,  -1,   4,   3,   1,   4,   0,   0,   3,   0,   1,   0,   4,   3,   0,   2 },
     {   -2,   0,   3,  15, -10,   1,   3,   5,  -1,   0,   0,   1,  -3,   3,  -4, -23,   6,  -9,  15,   9,   1,   1,   1,   2,   1,   1,  27,  21,  30, -13,   4,  -6,  -6,  -2,  -2,  -2,  -4, -11, -38,  17, -15,  11,  13,   0,   1,  -1,   2,  -2,  26,  39,   2,   3,  11,  37,   8,  -5, -53,   2,   5, -15,   5,   3, -17,  -8, -45, -21,  -3,  -6,   1,   2,   4,   2, -15,  -2,   7,  -6,   1,   2,   2,  -2,  -9,   1,   1,  -8,  -3,   2,   8,   3,   0,   3,   1,   2,   2,   3,   2,   1 },
     {   -5,  -6,  -3, -14, -49, -30,  -4,   4,   5,   3,   2,   1,   1,   2,   1,  19,  15,  -7,  -8,  -4,   3,   1,  -2,  -1,  -2,   6,  -3, -17,  35,   5, -11,  -6,   0,   2,  -1,   2,  -6, -18,   5, -35,   7, -20,  -4,   4,   4,   2,   0,   2,  31,  11,   6, -24, -29, -20,   7,   8, -12,  12,  38,  26,   5,   4,  14,   0, -18,  12,  32,  23,   1,  -6,  -5,  -7,  -3,  10,   2,   0, -11,  -6,  -7,   0,  -5,   1,  -3,   2,   1,   1,  -5,  -4,  -3,   0,  -3,  -1,  -3,  -1,  -3,  -1 },
     {    1,   6,  -2,  -5,  33,  26,  -6, -12,  -2,  -6,  -2,  -3,  -5,  -4,  -4,   2, -53,  12, -14,   1,  -1,   0,   2,   1,  -3,   6,  -1,   4, -29,   2, -10,   6,   1,   0,   0,  -1,   1, -20,   3,  18,   9, -29, -11,  -2,   2,   0,  -1,   0,   2,  33, -49,  -6,  -5, -18,  16,   8, -21,  34,   2,  26,   1,   4,   7,  -6, -22,   9,  18,   4, -16,  -4,  -3,   0,   0,   5,   6, -15,  -4,   5,   6,   3,   3,   1,   0, -10,   6,  -2,   4,  -4,   4,   1,  -1,  -2,  -2,  -1,   0,  -3 },
     {   -1,   0,  -9,  -6,  34, -35,   2,  -1,   8,  -2,   2,   0,   1,  -3,   7,  -3, -14, -19,  -2,   4,   3,  -3,   2,  -2,  -1,   9,  -9,  18,   4,  36,  -3,  -1,   4,  -3,   1,   0,  -5,   8,   8, -15,  32,  21, -17,  -7,   1,  -2,  -1,   0,  31, -40, -10,  -7,   3,   7,  -9,   6, -40,  27,  18, -25,  -8,   5,  -2,   3,  -3,   4, -12, -24,   8,  11,  -1,   6,  26,  13, -17,   0,  16,  -2,  -2,   5,  -4,   3,  -5,   9,   4,  -9,   1,  12,   1,  -1,   0,   6,  -7,  -4,   0,  -1 },
     {    1,   1,   1,   3,   3,  -6,  12,  -3,   2,  -4,   0,   0,   1,   4,   1,   3, -13,  24,   5,   5,  -5,   0,  -2,  -1,   1,  -1,   2,   2,  16, -30,   6,  -1,  -3,  -5,   4,  -3,   1,   9,  -6,   8, -20,  19,   0,   5,  -5,   0,  -1,  -1,  16, -18,  30, -19,  51, -25,   8,   6, -15,  31, -26,  23, -14, -14,  43, -12,   9,   7,   5,   7,  -2,  38,  -7, -23,   0,   9,   4,  -4,  -2, -11, -20,  -9,   1,   2,  -1,  -7,   0,   3,   0,   1,   3,  -2,  -1,  -1,   2,   1,  -1,   0 },
     {   -1,  -2,  -2,  -3,  33, -37,   3,   3,  -2,   0,   3,  -1,   0,  -1,   4,   1, -24,   0,  14,   6,  -8,   0,   4,   0,   4,   1,  11,   9,  26,  42,  33, -14, -11,  -5,  -2,  -2,  -4, -15, -12, -10, -16,  13,  -7, -13,  -6,   2,  -1,   0, -23,  22,  -1, -15, -13, -12,  -3,   1,  33, -10,   2,  31, -17,   4,   2,  -2,   0, -48,   7,  -2,  -3,  17,   0,   7,  -8, -13,  16,   8,  10,   6,  -4,  -8,  -3,   5,   6,   3,   1,   0,  -2,   1,   0,   1,  -5,   0,   5,  -2,  -3,  -1 },
     {    0,   1,  -2,  -7,  -4,  -2,  48, -10,  -7,   0,   0,   0,   2,   4,   6,  26,   5,   0,  55, -20,  -9,   3,   0,  -6,  -2,  -4, -14, -30,  25, -18,  25,  -1,   5,   4,  -3,  -3,   4,  -5,  28,  -7, -19, -20,  -4, -10,   0,   3,   2,   0,  -4, -10, -29,  21,   3,  12,  -1,  -3,  -4,  13, -11,   9,  -6,   1, -11,  -3, -23,  14,  -9, -19, -21,  11,   1,   9,   0,  -1, -12,   3,  13,  18,   8,  11,  -3,  -2,  -3,   8,   0,  -6,   4,   2,  -5,   2,   5,   7,  -3,  -6,   2,  -1 },
     {   -4,  -3,  -4,  -1,   0,  -8,  -3, -15,   0, -13,   2,   0,  -3,   0,   7,   5,  -7,   5, -10,   3,  -1,  -2,   2,  -1,  -5,   0,  -9,   6,   2,   0, -24,  15,  -7,   0,  -5,  -1,  -4,   9,  13,  -2,  15, -22,   5,   9,   1,  -2,  -3,   1,   2, -26,   9,  14,   9,  -2,  16,  -6,  19,  24,   4,  -2,   6,  14,   7, -10, -40, -18, -36,  -1,  12,  -5,  -4,  -9, -53, -34,  24,  32,   9, -13,   3,   0, -20,   4,  24,  13,  -4,   8,   8,   5,   1,   5,   6,  -6,  -3,  -1,   1,   0 },
   },
   {
     {  -71,  65, -17,   3,  -2,   2,  -1,   0,   0,   0,   0,   0,  61, -48,   2,   6,  -2,   1,  -1,   1,   0,   0,   0,   0, -14,   1,  15,  -9,   2,  -1,   1,   0,   0,   0,   0,   0,   2,   5,  -8,   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,  -2,   2,   2,  -1,   0,   0,   0,   1,   1,  -1,  -1,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   57,  -7, -32,   9,  -3,   1,  -2,   1,  -1,   0,  -1,   0,  10, -56,  52, -12,   2,  -1,   1,   0,   1,   0,   0,   0, -40,  53, -17, -12,  10,  -5,   3,  -2,   1,  -1,   1,  -1,   8,  -7, -13,  19,  -7,   1,   0,   1,   0,   0,   0,   0,  -4,   1,   9,  -6,  -2,   3,  -1,   0,   1,   0,  -4,   1,   2,  -1,   0,   0,  -2,   1,   2,   0,  -1,   0,   0,   0,   1,   0,  -2,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   1,   0,   0,   0,  -1,   0 },
     {   61,  -7,   7, -23,   2,  -3,   0,  -2,   0,  -1,   0,  -1,  34, -28, -36,  33,  -6,   3,  -1,   2,  -1,   1,   0,   1,  -8, -31,  58, -19,  -3,   4,  -2,   1,  -1,   0,   0,   0, -20,  34, -20, -12,  16,  -8,   3,  -2,   2,  -1,   1,  -1,  -2,  -4,  -3,  16,  -9,   2,   1,   0,  -3,   4,   3,  -7,   2,   3,  -2,   0,  -2,  -1,  -1,   3,   1,  -2,   1,   0,  -2,   3,   1,  -2,   0,   1,   0,   0,   0,   0,  -1,   1,  -1,   1,   0,  -1,   0,   0,   0,   1,   0,   1,   0,  -1 },
     {   24,  56, -72,  23, -10,   3,  -3,   1,  -1,   0,  -1,   0, -50,   1,  19,   9,  -9,   7,  -3,   2,  -1,   1,   0,   1,  44, -33,  16, -15,   7,  -1,  -1,   1,  -1,   0,   0,   0, -10,  -2,   7,  -3,   4,  -5,   3,  -2,   1,  -1,   0,   0,   1,   4,  -2,  -1,  -1,   1,   0,   0,  -1,  -3,   1,   1,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   39,  17, -17,  12, -13,   0,  -3,   0,  -2,   0,  -1,   0,  44,  -3, -26, -19,  11,   0,   1,   0,   1,  -1,   0,   0, -10, -22, -16,  51, -16,   3,  -1,   2,  -1,   1,   0,   1,  13, -26,  53, -33,  -7,  12,  -6,   3,  -2,   2,  -1,   1, -15,  14, -13, -10,  27, -14,   4,  -1,  -1,   0,   2,  12, -13,   0,   4,  -2,  -4,   2,   0,  -6,   4,   4,  -3,   1,   0,  -1,   2,   3,  -1,  -2,   1,   0,  -2,   1,   0,  -2,   0,   0,   1,   1,  -1,   1,   0,  -1,   0,   0,   1,   1 },
     {   10,  43,   0, -32,   8,  -3,   0,  -2,   0,  -1,   0,  -1, -33,  17, -45,  47, -10,   1,   1,   0,   1,   0,   0,   0, -40,  33,   7,  -3, -10,  10,  -5,   2,  -2,   1,  -1,   0,  46, -40,   0,   7,   2,  -1,  -3,   2,  -1,   1,  -1,   1,  -7,  -6,  21, -12,   0,   0,   2,   0,   4,   3,  -9,  -1,   7,  -3,   0,  -1,  -2,  -3,   4,   3,  -4,   0,   1,   0,   2,  -1,  -1,  -3,   2,   1,  -1,   0,   0,  -2,   2,   1,   1,   0,  -1,  -1,   0,  -1,   1,   1,   1,   0,   0,  -1 },
     {  -22, -17,   8,   7, -10,   6,   0,   2,   0,   1,   0,   1, -44, -34,  20,   9,  21,  -4,   2,   0,   2,   0,   1,   0, -10,  21,  36,   1, -30,   6,  -3,   0,  -2,   0,  -1,   0,  17,  -3,   6, -53,  33,  -1,  -4,   2,  -1,   1,   0,   0, -13,  24, -27,  28,   7, -24,  13,  -6,   7,  -1,   0,   2, -22,  17,  -3,  -1,   2,   0,  -2,  -3,  10,  -2,  -4,   3,   3,   1,   0,   1,  -4,  -1,   3,  -1,   0,   1,  -2,   0,   2,   0,   0,   0,   0,   1,  -1,   0,   1,   0,   0,   0 },
     {    7,  25,  26, -62,  33, -17,   8,  -8,   3,  -3,   2,  -1, -17, -45,  17,  22,  -3,  -2,   4,   0,   2,   0,   1,   0,  41,  11, -36,  24, -17,   8,  -5,   1,  -1,   0,  -1,   0, -36,   9,  17,  -6,  -4,   7,  -4,   3,  -2,   1,  -1,   1,  18,  -2,  -6,  -9,  11,  -4,   0,   1,  -7,   1,   0,   8,  -3,  -3,   3,  -2,   3,   2,  -1,  -2,  -2,   4,  -2,   0,  -2,  -1,  -1,   1,   1,  -2,   0,   1,   1,   0,   0,   0,  -1,   0,  -1,   0,   1,   0,   0,   0,  -1,   0,   0,   0 },
     {   20,  34,   3, -21,  -6,   3,  -1,  -1,  -1,   0,  -1,   0,  27,  36,   3, -21,  11, -17,   2,  -4,   1,  -2,   0,  -2,  14,  15, -39, -17, -14,  14,   1,   0,   1,   1,   0,   0, -10, -17, -23,   6,  44, -11,   1,   1,   1,   0,   0,   0, -11,  11, -15,  46, -33,  -7,  11,  -6,   6, -17,  12, -10,  -9,  26, -13,   4,  -1,  -2,   2,   1,  11, -12,  -1,   3,   0,  -4,   1,   1,  -5,   2,   3,  -2,  -2,   0,   1,   2,   0,  -2,   1,   0,  -1,   0,   1,   1,   0,  -1,   1,   0 },
     {   -8, -41,  -4,  -3,  17,   3,   1,   2,   1,   1,   1,   0,  31, -14,  10,  34, -32,   4,  -3,   1,  -2,   0,  -1,   0,  43,  -7,  -4, -36,  12,   9,  -6,   3,  -1,   1,  -1,   1,  19, -52,  27,   8,  14, -17,   6,  -1,   1,  -1,   0,   0, -28,  39,  -2, -12,  -7,   3,   5,  -5,   2,  -2, -12,   9,   5,   0,  -4,   2,  -5,   5,   6,  -2,  -5,   0,   2,   0,  -1,   2,  -5,   0,   3,   1,  -1,  -1,  -2,   2,   2,   0,   0,   0,  -2,   0,  -1,   2,   1,   0,   0,   0,  -1,   0 },
     {  -14, -20, -24,   6,   4,  12,  -1,   2,   0,   2,   0,   1, -16, -27, -16,  15,  15, -17,  11,  -1,   3,  -1,   2,  -1, -21, -18, -15,  39,  -8,  24,  -6,   3,  -1,   1,   0,   1,  -6,   9,  32,  20,   2, -26,   0,  -1,  -1,  -1,  -1,  -1,   6,   8,   4,   2, -51,  25,   1,  -3,  12, -14,  23, -33,  27,   6, -20,  10,  -2,   9,  -4,   1,   2, -20,  15,  -2,   3,  -1,   2,  -3,  -3,  10,  -2,  -4,  -1,   3,  -1,  -1,   1,  -1,   1,  -2,   0,   1,   0,   0,   1,  -1,   1,  -1 },
     {    7,  32,  44, -23,   8, -12,  -1,  -3,  -1,  -2,  -1,  -1, -21,  -2,  24, -31,  -7,   7,   2,   0,   1,   0,   1,   0, -35, -35,  12,  -4,  40,  -8,   1,   1,   1,   0,   0,   1,  10,   6,  12,   9, -16, -17,  12,  -8,   3,  -2,   1,  -1, -32,  39, -33,  -2,  -5,  20,  -5,  -1,  20, -15,  -1,  13,   3,  -4,  -8,   7,  -4,   0,   2,  -7,  -2,   0,   5,  -1,   4,   1,  -5,   5,   2,   0,  -1,  -1,  -2,   1,  -1,  -1,   2,   0,  -1,   2,  -1,   0,   0,  -1,   1,   0,  -1,   1 },
     {   13,  16,  41,   1, -43,  25, -15,   5,  -5,   2,  -3,   1,   0, -22,   1, -15,  42, -26,   9,  -3,   3,  -1,   1,  -1,  50,   2,  18,  -1, -19,  -2,   3,  -6,   1,  -2,   0,  -1,  23, -18, -19,   2, -15,   5,  -3,   3,   1,   0,   0,   0, -15,  -2,  16, -20,   0,  20,  -5,   3,   4,   1,  -8, -11,  30, -16,  -4,   4, -11,   7,  -8,  13, -10,  -7,  14,  -2,   2,   0,  -1,  -3,   0,   9,  -3,  -2,  -3,   2,  -1,   3,   1,   0,   0,  -2,  -1,   0,  -1,   2,   0,   0,   0,  -1 },
     {   -5,  -9,   2,  -9,  -8,   3,   6,   0,   1,   0,   1,   0, -20, -37, -28, -12,  36,   6,   0,   3,   0,   2,   0,   1,   0, -40, -15,   1,  22, -12,  15,  -3,   3,  -1,   2,  -1,   3, -16, -10,  51,  12,  12, -11,  -1,  -2,  -1,   0,   0,  -6,  15,  35,  16,  -7, -37,   3,   0,  -5,  18, -13,   6, -33,  16,  12,  -9,   6,  -5,  12,  -2,   2,  11, -17,   4,   1,   4,   1,  -4,  -3,  -5,   3,   2,   2,   1,   3,  -1,   0,   1,   0,  -1,   1,   0,   2,   0,   0,   1,   0,   0 },
     {    5,  11,  27,   3,  -5,   0,  -5,   0,  -1,   0,  -1,   0,   1,   6,  44, -12,  -9,  -3,  -2,  -1,  -2,   0,  -1,   0, -29, -37,   0, -10,  -4,  26,  -7,   6,  -1,   3,   0,   2,   0, -19,  24, -13,  39, -18,  -2,   4,  -4,   2,  -2,   1,  30, -20,  26, -11,  -8, -10,  11,  -6, -36,  42, -23, -12,  15,  -1,   1,  -4,   8,  -2, -12,  19,  -9,  -3,   2,   0,  -4,   2,   3,  -6,  -1,   4,  -2,  -1,   3,   1,  -5,   4,  -2,   1,   0,   0,   0,   1,  -2,   2,  -2,   0,   0,   0 },
     {   -1, -11, -25, -18,  48, -34,  17,  -6,   5,  -2,   2,  -2,   8,   3,  24, -26,   0,   3,   2,  -5,   2,  -1,   1,  -1,  18, -25,  12,  32, -25,  11, -12,   6,  -3,   1,  -1,   0,  62,  13, -37,  -4,  -5,  -4,   0,   2,  -2,   1,  -1,   1, -26, -18,  19,   2,  -7,   2,   3,   0,   6,   7,  -5,  -7,  -2,   5,   0,   0,  -8,  -2,   0,   4,   4,  -1,  -1,   1,   2,   2,  -1,   0,  -1,  -1,   2,   0,  -4,   1,  -1,   2,   1,   1,   0,   0,  -2,   0,   0,   0,   1,   1,  -1,   0 },
     {    4,   2,  16,  16,  -4, -20,   6,   0,   0,  -1,  -1,   0,  -9, -32, -26, -15, -17,  45,  -9,   7,   0,   3,   0,   1,  20,   9, -19,   3,  35, -19,  -5,   4,  -4,   3,  -2,   1,  28, -23,   1, -23,   8,  -6,  16,  -8,   4,  -2,   2,  -1,  16, -25,  -8,  45,  -2,  14, -17,   1, -15,  11,  27, -25,   4, -15,  -4,   7,   1,   0, -15,   9,   4,  -7,  12,  -3,  -2,   2,   8,   4,  -9,   2,   1,  -2,   1,   1,  -3,  -2,  -2,   1,   3,   2,   1,   0,  -1,  -1,   0,   1,   2,   1 },
     {   11,  15,  29,  23, -16, -10,   2,  -4,   0,  -2,  -1,  -1,   8,   3,  12,  24, -11,  13, -14,   1,  -3,   0,  -2,   0,  20,  15,   9,  14,   3, -21,   8,  -3,  -1,  -2,  -1,   0,  25,  19,  20,   8, -28, -10,  -6,  -6,   0,  -3,   0,  -1,   9,  -6,  -6, -21, -44,  -6,  18,   1, -13,   1, -11, -17, -11,  57,  -4,  -8,   4, -12,   3, -11,  27,  -2, -28,  12,  -5,  -2,  -3,  -4,   2,  -8,  11,   3,   2,  -1,  -1,  -4,  -2,  -1,  -2,  -2,   1,   0,   0,  -1,  -1,   0,  -1,   0 },
     {   -4,  -7,  -5,  -8,  -4,   2,  15,  -4,   3,  -1,   1,   0,  -8, -11,  -9, -11,  -9,  11, -15,  10,  -2,   3,  -1,   2,  -8, -15,  -9, -15,  -8,   6,  22,  -5,   3,  -1,   2,   0,  -7, -11, -11,  -9,  -4,  26, -18,   9,   2,   2,   1,   1,  -4, -10,   0,  -9,  22,  24,  17,  -2,   0,  11,   3,  23,  29,  21, -52,  -6,  10, -12,  17, -10,  17, -55,  -2,  30,  -3,   7,  -3,   8,  -5,  -2,  29, -16,   2,   0,   2,   0,   0,   2,  -1,   1,   2,  -1,   2,   0,   0,   1,   0,   1 },
     {    4,   6,  30,  30,  -7, -13,   3,  -1,  -1,  -3,   1,  -1,  -1,  -1,  27,  28, -26,  18, -13,  -1,  -1,   0,  -1,   0, -15, -27,  -6,   9, -29,  -1,   0,   6,   1,   2,   0,   1, -23, -24, -17,   7,   5,  32,  -4,   1,  -1,   3,   1,   2, -15,  -3,  33,   2,  10, -11, -16,   7,  31, -34,  20, -27,  -6,  -2,   9,   3, -21,  24,  -5,  -9,  19,  -6,   2,  -4,   9,   0,  -4,   5,  -8,  -1,   2,   1,   0,   2,   2,  -5,   2,   0,  -3,  -1,  -1,   0,   0,  -2,   0,  -1,  -2,   1 },
     {   -3,  -8, -15, -44,  14,  20,  -7,   5,   1,   1,   1,   1,   9,  13,  27,   6,  15, -23,   5,  -3,  -2,  -2,  -1,   0,  -5, -14,  28, -16,  15, -20,  18, -11,   5,  -4,   1,  -2,  19, -28,  21,   8, -26,  28, -16,   4,  -1,   0,   0,   0,  32, -15, -19,  15,   5,  -9,  -8,   4, -11,  -8,  30, -26, -12,   3,   3,   1, -11,  18, -24,  -1,  18,  -9,   2,   1,  -8,   2,  -2,  11, -11,  -1,   6,  -2,   4,   2,  -4,  -4,  -3,   1,   0,   3,   2,   2,  -1,  -1,  -2,   0,   0,   2 },
     {   -4,   0, -12,  -8,  15, -34,  31, -11,   7,  -4,   2,  -2,   4,  13,  -3, -26,  33,  24, -22,   1,  -1,  -1,  -1,   0,  -2,  25,  41,   3, -20, -28,   2,   3,  -6,   2,  -2,   0, -45, -34,  18,   2,  -2,   6,   8, -10,   7,  -2,   2,   0,  -7,  27,  13,  -7, -22,   8, -13,   6,  -8,   6, -16, -20,  15,  -2,   3,   2,  11,  -3,  -5,  14,  -2,  -5,   0,   2,  -4,   0,   0,  -8,  -2,   4,   5,   0,   3,   0,  -4,   3,   0,  -1,   0,  -2,   1,   1,  -3,   2,   0,  -1,   1,   0 },
     {   -2,  -5,  -5,  -2,  -7, -18,  15,   1,   2,   2,  -1,   0,  -9, -18, -24, -32,  -7,   2,  12,  -4,  11,  -1,   5,   0,  -6, -11, -13, -46, -25,  16,   2,  16,  -2,   6,   1,   3,  13,   1,   5,  -5,   2,  14,  -3,  -5,   4,  -1,   2,  -1,  21,   0, -22, -22,  -8,   9,  -2,  18,   2, -21,   8,   3,  30,  23,  27, -22, -14,  26,  -9,  15,   8,  16, -42,  -9,   6, -12,   4,  -3,   5, -12,  -9,  19,   1,   9,   0,   4,   4,  -6,   1,  -1,   0,   4,   1,   1,   2,  -2,   1,   0 },
     {  -10, -12, -19, -42, -29,  34,  -6,  11,  -3,   3,  -1,   3,  -2,   5,  16,  17, -14,   9, -13,   5,  -2,   0,  -2,   1, -11, -12, -22,  22,   3, -32,  19,   1,   1,   2,   0,   0, -11,  -1, -10, -30,  -2,   4,  20, -14,  11,  -5,   5,  -2, -46, -16,   2,   9, -10,  18,  -3,  12,   1,  20, -10,  -7,  17,  17,  16,  -5,   6,   3,  -1,  23,  -1,  11, -22,  -9,   3,  -4,  10,  -8,   4,  -5, -10,  12,   1,   1,   4,   2,  -2,   2,   3,  -2,   1,   2,   0,   0,  -1,   0,   1,  -2 },
     {    2,   4,  10,  34,  22, -13,  -5,  -4,  -2,  -1,   1,  -1,  -3,  -6,   3,  20,   8, -30,  27, -15,   6,  -3,   3,  -2,  -9, -15,   1,  -1, -16,   6,  -8,   0,   3,   1,   1,  -1, -18, -41, -22,   2, -36,   9,  23,   5,   1,   2,   1,   1,  -9,  -6, -18,  23,  -2,  37,   4,  -8, -24,  25,  20,  11, -10,  13,  -4,   0,  27, -14,   6,  -3,  -1,  18, -19,  -5,  -5,   4,   9,  -9,   9, -12, -12,   8,   2,   1,  -4,   1,   2,   0,   3,   1,   2,   0,  -1,   0,   0,   0,   3,  -2 },
     {   -4,  -1,   2,  -5,   9,   9, -11,   1,  -4,   3,  -3,   0,  -8,  -5, -12, -31, -11,   6,  21,   0,   2,   0,   2,   0,   7,  20,  31,   9,  28,  14, -25, -10,   0,  -3,   0,   0, -37, -24,  10,   3,  -2, -25,  -1,  13,  -3,   5,   0,   1, -33, -53,  15,  -8,   2,  -5,  32,  -5,  10,  -2,   6,   7,  -8,  22,  -3, -11, -25,  29,  -9,  -5,  12,   2,  -9,   3,   6,  -2,  -1,   8,  -2,  -6,  -3,   2,  -2,   2,   2,  -2,   1,   3,  -1,   3,  -2,   1,   1,  -3,   3,   1,  -2,   3 },
     {    0,   8,   4,  14,  14, -10,   4,  -8,   0,   0,   0,  -1,  -5,   3,  -8,   2,  -8, -21,   0,   3,   5,   1,   0,   0,   2,  13, -13, -27,   6,  15,   9,   2,   0,   0,   0,   0,  -3,  36,  24,   8, -18,  12, -21,   6,  -3,   1,  -3,   0, -60,  15,   8,   8,  20, -20,   7,   7, -13,  30,  21, -36,   8,   0,   7,  -9,   7, -10, -33,  27,   2, -11,   3,   1,   5,  -2,  13,   0, -15,   6,   0,  -4,  -2,  -2,  -9,  -7,   4,  -2,   5,   2,  -2,   1,  -4,  -3,  -1,  -2,   2,   3 },
     {    3,   4,   2,  -5,   6,  28, -18,   9,  -6,   1,  -4,   2,   6,   9,   9,  -2,  -1,  25,  17, -22,   4,  -5,   2,  -3,   8,  13,  16,  19,  12,  25, -24,   0,  -5,  -1,   0,  -1,   4,   9,   5,  12,  15,  12,   0, -15,  -1,  -2,   1,  -2,   9,  28,  20,  11,  16,  12, -32,   1,   9,  11,  14,  -8,   6, -10, -19, -18,   5, -13, -11,  -5,  -3, -16, -50,  17,   0,  -9,   2, -14,   4, -36,  19,  30,  -4,   2,  -8,  -2,   1,  -5,  -3,  -5,  -2,   1,  -1,   0,   2,  -2,  -1,   1 },
     {   -4,  -5,  -9, -24, -26,   3,  -7,   7,   2,   3,   2,   0,  -4,  -4, -10, -28, -34,   3,  27,   1,   3,   3,   3,   1,   1,   3,   2,  -7,  -5,  18, -15,   0,   6,  -3,   3,  -1,  -5,   1,  -4, -12, -14,  15,  10,  13,  -6,   7,  -3,   5,   7,  27,  21,  -8,   1,   8,  -8,  -8,   1,   4, -14, -39, -13,  12,   3,  26,  17, -27,   6,  -9,  46,  20,  20,  -9, -14,  22,  -7,  18,  17, -13, -11, -12,   7,  -7,  -2,   8,  -3,   4,   2,   5,   3,   3,   1,   1,   0,   4,   1,   4 },
     {    5,   7,   6,  20,  45,   2, -33,   4,   1,   2,  -2,   1,  -3,  -4, -10, -14, -19, -18,  25,  -1,  -2,   0,   2,   0,   3,   4,  -3,  -9,  -2, -29,  28, -16,  10,   0,   3,  -1,  -1,   7,  24, -26,   6,  32, -14,   3,   3,  -7,   1,  -1,   9,   0,  22,  26, -30,   9,  -6,   3,  33,  16, -34,   4,   8,  -2,  -3,   0, -24,   1,  16,  -1,  -5,   0,  -3,  -1,   5,   4, -12,  -9,   6,   1,  -2,  -2,  -6,  -8,   2,   4,   3,   2,   2,  -5,  -1,   1,  -2,   2,   0,   4,  -2,  -1 },
     {    0,   0,  -2,   7,   7,  -2,  -8,  11,   5,  -5,   5,  -1,  -1,  -1,  -4,   6, -18,   9,  38, -19,   1,   6,   1,   0,  -2,  -4,   1,  27,  24, -12, -18,  -7,   6,  -5,   4,  -2,  -9, -13, -38,  -5,  -4,  22, -10,   5,  -4,   4,   5,   1,  11,  24, -29, -16,  -4, -35,  18,  -5, -15, -21, -23, -13,  21,   9,  -3,   0,  -2,   0,   0,  47, -12, -14,   5,   8,   2,  -8,  21, -10, -18,  10,   7,  -8,  -3,   6,   6,   8,  -2,   2,  10,   2,   0,  -1,   3,  -2,   2,   2,   3,   3 },
     {    2,  -1,   7,   0,   1,  22,   7, -10,   0,  -3,  -1,   0,  -1,  -6,   8, -15, -21,  10,  -8,  10,   3,   1,   1,   1,   5,   2,  18,   9,  -4,  22,   9,  -3,  -9,   1,  -2,   1,   0, -15,  10,  17,   6,  16, -41,  -2,   6,   0,  -2,  -2, -11, -30, -25,   5, -18, -12,   2,  19,  19, -20,  15,  -1,  10, -12,  15,  -6,  23, -47,  22,   2, -18,  17,   5,  -7, -25,  25,  14, -31,   9,   3,  -7,  -3,  -4,   9, -15,   5,  -1,   0,  -1,  -1,   0,   1,  -2,   4,  -1,   0,   3,   2 },
   },
   {
     {   96, -63,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0, -48,  12,  19,  -3,   1,   0,   1,   0,   0,   0,   0,   0,  -6,  18,  -6,  -5,   1,   0,   0,   0,   0,   0,   0,   0,  -3,   0,  -5,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   45,  34, -47,   2,  -6,   1,  -2,   0,  -1,   0,  -1,   0,  38, -76,  14,  19,  -2,   3,  -1,   2,   0,   1,   0,   1, -30,   9,  37, -15,  -2,   0,   0,  -1,   0,   0,   0,   0,  -4,  13, -10, -10,   6,   0,   0,   0,   0,   0,   0,   0,  -5,   2,  -1,   3,   3,  -2,   0,   0,  -1,   3,  -2,   0,  -1,  -1,   1,   0,  -2,   0,   1,   1,   0,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },
     {  -15,  61, -43,  -4,   1,   0,   0,   0,   0,   0,   0,   0, -84,   4,  32,  11,   3,  -1,   1,   0,   1,   0,   0,   0,  15,  27,  -8, -14,  -3,  -1,   0,  -1,   0,   0,   0,   0,  17, -16, -17,   8,   4,   1,   0,   1,   0,   0,   0,   0,   6,  -6,   7,   5,  -4,  -1,   0,   0,   4,  -4,   0,  -2,  -1,   2,   0,   0,   2,  -2,   2,   0,   0,   0,  -1,   0,   2,  -2,  -1,   0,   0,   0,   0,   0,   1,  -1,   1,   0,   1,  -1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0 },
     {  -44, -40,  18,  21,   1,   5,   0,   3,   0,   1,   0,   1,  17, -25,  64, -25,  -5,  -1,  -1,  -1,  -1,  -1,   0,   0, -21,  51, -19, -36,  17,  -3,   2,  -2,   1,  -1,   1,  -1,   2,   5, -27,  18,  10,  -6,   1,  -1,   0,   0,   0,   0,   1,   3,  -2,   7,  -7,  -3,   1,   0,   3,  -1,  -4,   2,  -2,   3,   0,   0,   1,   1,   0,   1,  -1,   0,  -1,   0,   1,   0,  -2,   1,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },
     {   28,  28, -38,  12,  -4,   0,  -1,   0,  -1,   0,  -1,   0,  39,  23,   2, -37,   0,  -3,   0,  -2,   0,  -1,   0,  -1, -23,   6, -62,  20,  23,  -3,   4,   0,   2,   0,   1,   0,   6, -27,  15,  45, -19,  -4,  -2,   0,  -1,   0,   0,   0,  -2,  -4,  18, -14, -16,   8,   1,   1,   0,  -5,   3,  -2,   5,   5,  -3,   0,  -1,   0,   5,  -3,   0,  -1,  -2,   0,  -1,  -2,   1,   1,   1,   0,   0,   1,  -1,   0,   2,  -1,   0,  -1,   0,   0,   0,   0,   1,  -1,   0,  -1,   0,   0 },
     {  -11, -30, -59,  58,   0,   6,   1,   3,   0,   1,   0,   1,  11,  14,  29,  11, -31,   2,  -4,   0,  -2,   0,  -1,   0,  48, -38,   7, -20,  -6,  12,  -2,   2,  -1,   1,   0,   1, -10, -13,  26, -10,  11,   1,  -4,   1,  -1,   0,   0,   0,  -3,  11,   3, -15,   5,  -4,   0,   0,  -5,   2,  -3,   0,   6,  -2,   2,   0,  -1,   5,  -1,  -2,   0,  -2,   0,   0,  -3,   0,  -1,   0,   1,  -1,   1,   0,  -1,   2,   0,  -1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {  -16, -11, -33,  36,   4,   0,   1,   1,   0,   0,   0,   0,  -5,  55, -28,   6, -15,   0,   0,  -1,   0,   0,   0,  -1, -80,  16,  32,  -3,  -5,   6,  -1,   1,   0,   0,   0,   0,  20,  17, -12, -15,   3,   2,  -3,   0,   0,   0,   0,   0,  12, -16,  -9,   6,   4,   0,  -1,   1,   6,  -6,   8,   2,  -2,  -2,  -1,   0,   3,  -4,   1,  -2,   0,   0,   1,   0,   3,  -2,   2,  -1,   1,   0,   0,   0,   1,  -2,   1,  -1,   1,  -1,   1,   0,   1,  -1,   0,   0,   0,   0,   1,   0 },
     {  -18, -14,   2,  27,  -6,   2,  -2,   1,   0,   0,   0,   1, -32, -42,  17, -15,  14,   5,   5,   1,   2,   1,   1,   1, -20,   5,   6,  55,  -1, -11,  -3,  -2,   0,  -1,   0,   0,  25,   1,  44, -20, -48,   6,  -1,   2,  -2,   1,  -1,   0,  -4,   6,  -6, -38,  18,  18,  -1,   0,   3,   4,  -3,   6,  12,  -8,  -6,   0,  -1,   4,  -5,  -5,   0,  -4,   3,   1,   2,   2,  -1,   2,   0,  -1,   1,  -1,   1,   1,  -2,  -2,   1,   1,  -1,   1,   0,   1,  -1,  -1,   1,   0,   0,   1 },
     {  -10, -32,  -9,  -3,   9,   4,   3,   0,   1,   0,   1,   1,  18, -15, -30,  58,   1, -10,   1,  -2,   1,  -1,   0,  -1,  13,  49, -15,  10, -41,   4,   2,  -1,  -1,   0,   0,   0,  27, -46,  -6,   6,  -7,  21,  -5,   3,   0,   1,   0,   1,  -3, -12,  37,  -4,  -3,   1,  -7,   2,   1,   3,   3, -18,   3,   2,   0,   1,  -3,   0,   3,  -2,   7,  -2,  -1,   0,  -2,   2,   1,  -4,   1,  -1,   0,   0,  -1,   1,   1,  -1,  -1,   1,   0,  -2,   0,   0,   1,   0,   0,   0,   0,  -1 },
     {  -19, -21, -17, -11,  32,  -2,   5,   0,   2,   0,   2,   0, -21, -18,  -6,  41,  -3,  -8,   2,   1,   1,   0,   0,   0, -15, -22,   2,   9,  14,  -1,   6,  -1,   1,   0,   1,   0, -35,  42,   9,  50, -16, -20,   2,  -7,   1,  -3,   0,  -1,  15,   6,  10, -10, -52,  14,   4,   0,   0,  -4,   1, -16,   8,  21,  -6,   0,   7,   0,   4,  -3,   1,  -2,  -8,   2,   1,   0,  -1,  -5,   2,   1,   0,   3,   2,   1,   2,  -1,   0,   0,  -1,  -2,   1,   1,   0,   0,  -1,   0,   0,  -1 },
     {    7,  -1,  26,  30, -24,  -5,  -6,  -1,  -1,   0,  -1,   0, -13, -19, -24,  -6, -18,  25,   0,   6,   0,   2,   0,   1,  10, -27,  27, -20,  40,   4, -13,   1,  -2,   1,  -1,   0,  60, -15, -20,  25,  -3, -27,   2,   1,  -1,  -1,   0,   0,  -8, -24,  29,  -2, -20,   8,   7,   2,  -1,  12,   6, -19,   7,   2,  -1,  -2,  -7,   2,  -3,   0,   4,  -2,   1,   1,  -1,   6,  -1,  -1,   1,  -2,   1,  -1,  -2,   1,  -2,   1,   0,   2,   0,   0,  -1,   0,   0,   1,   0,   1,   0,   0 },
     {   -1,  -7,  -2, -60,  46,   2,   4,   0,   0,   0,   1,   0,  12,  14,  50,  -2, -10, -18,  -2,  -2,  -1,  -1,  -1,  -1, -25, -42,  17,  -1, -11,  14,   5,   2,   0,   2,   1,   1,  55, -18,   3,   1,   6,   4,  -9,   0,  -1,   0,  -1,   0,  -9,  -4,   5, -14,   0,  -3,   0,   3,  -6,   8,  -7,   0,  12,  -1,   1,  -1,  -4,   6,  -5,   4,  -2,  -5,   1,  -1,  -1,   2,  -4,   2,   0,   1,   2,   0,   0,   1,  -2,   1,   0,   1,  -2,   1,   0,   0,   0,   1,   0,   0,  -1,   1 },
     {   10,  28,  42,  -3, -17,  -2,  -4,  -4,  -1,  -2,  -1,  -1,  -3,  12,   8,  32, -52,  11,  -4,   2,  -1,   0,   0,   0, -24,  17, -10, -36,   0,  21,  -2,   2,   0,   1,   0,   1, -28,  -4,  29,  -4,  -4,   5,  -7,   2,   0,   0,   0,   1,   5,   3,   9, -46,   9,  14,  -3,   4,   0,   0, -20,  -3,  38,  -9,  -5,   0,   3,  -1,  -3,   7,   1, -13,   4,   2,  -1,   0,  -5,   1,   3,   0,   4,  -1,   1,  -1,  -1,   2,   0,   0,  -2,   0,   1,  -1,   0,   1,   0,  -1,   0,   0 },
     {  -10,  -5,   8,   6, -16,   8,  -2,   3,  -1,   2,   0,   1, -25, -24, -14,  -4,  17,   2,   4,   0,   3,   0,   1,   1, -44, -23, -14,  -1,  -1,  25,   1,   4,   1,   2,   1,   1, -10, -34,  40,  10,  52,   1, -16,   2,  -4,   0,  -2,   0,   3,  33,  20,  14,  -2, -46,   0,   1,   4,   6,  -7,  -6, -21,   2,  15,  -1,   2,   7,   0,   2,   1,   4,  -2,  -5,   3,   2,  -2,  -2,  -5,   0,  -2,   0,   0,   3,   0,   0,   2,   0,  -1,  -1,   0,   1,   0,   1,   0,   0,   0,  -1 },
     {    9,   7,  17,  17,   7, -12,  -1,  -4,   0,  -2,   0,  -1,  -2, -25,  11,  22, -39,  -1,   7,   0,   2,  -1,   1,   0,   1, -19, -54,  20,  -7,  33,   4,   1,   1,   1,   1,   1,  26,  29,  -1, -10,   9,   9, -16,  -1,  -2,   0,  -2,   0,  31, -35, -23,  24,   8,  -8,  -3,   3,   1, -15,  32,  10, -16,   0,  -1,   1,   0,   1,  10, -20,  -2,   2,  -1,   1,  -3,   1,   2,  -3,   6,   0,   0,   1,  -2,   1,   1,  -4,  -2,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0 },
     {   10,  17,  29,  25, -51,  10,  -2,  -1,  -3,   0,  -1,   0,   8,  21,  45,  12,  15, -23,  -8,  -5,  -3,  -2,  -1,  -1, -11, -23,  15,  14, -40, -19,  16,   1,   1,   0,   1,   0,  -2,   3, -10,  -1, -11,  20,   6,  -8,   1,   1,   0,  -1,  12,   2,  12,   7, -29,   0,  -4,  -3,  -6,  -3,   6, -38,  -9,  27,   0,   0,  -4,   3,  -7,  -9,  21,   3, -10,   0,   0,  -2,  -4,  -4,   1,  -1,  -1,   3,  -1,   1,  -3,  -2,  -1,  -1,  -1,  -1,   0,   0,  -1,   0,  -1,  -1,   1,  -1 },
     {   -3, -11,  -7, -21,  -8,  16,   2,   4,   0,   0,   0,   1,   4,  -5,   4,  -5,  17,  13,  -8,   3,  -2,   0,  -1,   1,  -1,  -2,  20, -18,  18, -17,  -3,   1,   0,   2,   0,   0, -16, -37,  25,   3,  -7,   5,   6,   2,   2,   1,   0,   0,  89, -41,  -8,   7,   6,   6,  -6,   0, -19,  -7,  22,   0,   6, -11,  -2,   2,  -6,  14,   2,  -1,   0,  -3,   4,  -2,  -5,   6,  -5,   2,   1,  -3,   2,  -3,  -2,   2,   0,   2,   0,   2,  -2,   1,   0,   1,   0,   0,  -1,   0,  -1,   0 },
     {    5,   2,  -2, -10, -21,  19,  -1,   1,  -2,   1,  -1,   0,  13,  11,  -2,  21,  13, -10,  -8,  -2,  -1,  -1,  -1,  -1,  16,  32,   0,  -4,   9,   1, -11,   0,  -2,  -1,  -1,   0,  39,   7,  38, -16,  10, -45,  -9,   2,  -2,   0,  -1,   0,   5,   7, -47, -15, -33, -18,  33,   1,  -6, -20, -10,  11,  -2,  36,   7,  -7,   0,   0,   4,   4,   3,   0, -11,  -3,  -2,  -5,  -1,  -2,  -1,   3,  -1,   3,  -1,   1,   1,   0,  -2,  -2,  -1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0 },
     {   12,  19,  22,  32,  55, -58,  -8,  -3,  -2,  -3,  -1,  -2,   3,  -3,  -3, -31,  -6, -12,  24,   1,   3,  -1,   2,  -1,   9,  22,  33,  -7, -27,  15,   2, -12,  -1,  -1,   0,   0,  -8, -28,  22,   8,  -7,  -8,  -7,   2,   4,  -1,   0,   0,  18,   8, -14,   6, -22,   0,   6,   1,  -6,   0,   0,  -9, -10,  15,  -4,   0,  -6,   1,  -3,  -1,   5,   5,  -5,   1,   1,  -3,  -2,   1,   1,   0,  -2,   3,  -1,  -1,   0,  -2,   0,  -1,  -2,   0,  -1,   1,  -1,   0,  -1,   0,  -1,   0 },
     {   -6,  -4, -13, -26,  -2,  23,  -2,   2,  -4,   2,   0,   0,  -1,   1,   8, -34, -30,  20,  10,   1,   0,   0,   1,  -1,   4,  21,   7,  28,  -8,  17, -21,  -2,  -2,   0,  -1,   1, -27,   8,  16, -31,  16, -20, -12,  13,   1,   1,   0,   0, -14, -30,  25,  19, -16,   5,  17,   1,  20,  14,  21, -37,   6,  15, -15,  -3,  -4,   0, -20,  -6,  25, -13,  -3,   3,  -3,   1,   2,  11,   4,  -9,   5,  -1,   1,  -1,  -3,   2,   2,   0,  -1,   3,  -2,   0,  -1,   0,   0,   1,   1,   1 },
     {    1,   4,   9,  -8,   7, -14,   6,  -1,  -2,  -1,   1,  -1,   7,  15,  26,  38, -39,  28,  -4,  -4,  -2,  -4,  -2,   0, -10,   5,   7,  25,  26, -40, -18,   2,  -3,   2,   0,   0,  -7, -29,  12,  -7, -13, -23,  25,   4,   2,   1,   0,   0,  -7,  21,   8,  21,   5,  -4,   5, -12,   2,   8,   3,   6, -48,  -9,   4,   0,   2,  -7,   0, -18, -10,  31,   2,   0,   3,  -1,  -1,  -4,   9,   1,  -7,   0,  -1,  -4,   1,  -5,   1,  -1,  -2,  -1,  -1,   1,   0,  -3,   0,   0,  -1,  -1 },
     {    5,   7,   0,   1,   9,  -3,  -4,  -3,  -2,  -1,   0,   0,  10,  13,   8,  -5,   5,   8,  -5,   0,  -4,  -1,  -2,   0,  20,  24,  12,  13,  12,  -4,  -3,  -6,   0,  -3,  -1,  -1,  12,  50,  13,   9,  -1,   5, -30,  -3,  -2,  -2,  -1,  -1,  23,  -9,  34, -25,   2, -48, -23,  16, -13,   7, -20, -24, -18, -30,  36,   9,  -2,  -8, -14,   1,   0,  19,  16,  -6,  -4,  -2,  -2,   3,   1,   2,  -2,  -6,   0,  -5,  -2,  -1,  -1,  -1,  -2,   0,  -1,  -2,  -1,   0,  -2,   0,   0,   0 },
     {    0,  -1,  -2,  -8,  -8,  29, -11,   1,   0,   2,  -1,   2,   0,  -3,  -6, -18, -35,  31,   5,  -1,   0,   1,   1,   0,   6,   6,  17,  42, -26,  15,  -8,  -1,   0,  -1,  -1,  -1,   3,  -9, -47,  29,  14,  -4,  -5,   3,   1,   0,  -1,   1,  31,  51, -20, -23,   3,  -3,  -1,  -1, -29, -15,   1,   3,   9,   0,  -2,   0,  -1,   4,  13,  -3,   4,   1,   2,   1,   6,  -2,  -6,  -4,   6,  -3,  -1,   1,  -1,  -6,   3,   4,  -1,  -1,   5,  -3,   2,  -1,  -1,  -2,   1,   0,  -1,   0 },
     {    0,   7,   3,  14,  13,   8, -10,   3,  -2,   0,   1,  -1,  -5,   0,  -4,  12,   6,  -7,  -7,   5,   2,  -1,   0,   0,  -9,   2, -20,   1,  -1, -11,   3,  -4,   3,   1,   0,   0,  -7,   0,  -8,  -9,   7, -15,   0,   6,   3,  -1,   1,   0, -12,  -6,  -9,  -1,   2,  -1,  14,   2, -82,  57,  23,  -2,   5,  -3,   2,  -6,  16,  23, -37,   8,  -3,  -4,   0,  -2,  13, -18,  -3,  15,  -7,   1,   0,   3,   3,  -3,  11,   0,   0,  -2,   1,  -2,   1,   0,   2,  -1,   0,  -1,   1,   0 },
     {    8,  10,  12,  18,  15,  26, -24,  -8,   1,   0,  -3,   1,   7,   8,   9,  17,  23, -31,  -1,  -3,  -1,  -2,   0,  -2,   5,   6,   5,  12,  10,  11, -15,  -1,  -5,   2,  -3,   0,   5,  -2,   5,   0,  17, -28,   1,  -1,   0,   3,  -2,   1,   6,  24,   5,  17,  -7,  12, -14, -11,  11, -13,  16, -13,   7, -43, -37,  15,  -4,  -7, -12,  -8, -24, -29,  43,  20,  -1,  -1,   0,  -1,  -5,  18,  15, -11,  -4,   0,  -3,  -2,  -2,  -1,   0,  -1,   0,  -1,  -2,  -1,   1,  -1,   0,   0 },
     {   -1,  -2,  -5,   7,   0, -51,  19,   6,   0,   1,   3,   0,   6,   1,  10,   8,  23,  14,   7, -13,  -2,  -1,  -3,   0,  -7, -12, -24,  20,  -1,  -3, -39,   9,   5,   4,   1,   1,   1,   5, -28, -37,  23, -14,  19,  22,  -5,   4,  -4,   2,  35,   4,  24, -10, -11,   5,  15, -13,  -3,  -1, -33,  -3,  22,   2,   2,  -2,  -5,   4,   6,  27,  -1, -13,   4,  -6,  -7,   0,   7,  -1, -18,   2,   4,  -2,   2,   0,   0,   3,  -1,   0,  -1,  -1,  -1,  -1,  -1,   1,  -1,   1,   0,  -1 },
     {   -5,  -6, -13, -21, -32, -40,  49,  -1,   5,   2,   1,   0,   5,   6,   1,   2,  12,  25,   1, -21,   1,  -1,   1,   0,   3,   8,   2, -18,   1,  27, -23,  -1,   6,   0,   2,  -1,  18,  14,  19,  16, -13,  17,  -2,   0,  -2,  -4,   0,  -2,  -9,  31, -19,  16,  -3,   8, -29,  -3, -20,   3,  25, -19,  -3,  -8, -20,  17,   9,   5, -15, -15,   9, -13,   4,   5,   4,  -9,  -8,  11,   3,  -6,   5,   6,   0,   3,   2,   0,   0,  -2,  -2,   1,   0,  -1,   2,   0,   2,  -1,  -1,   1 },
     {   -2,   0,   6,   6, -19,  -9,  21, -11,  -4,   5,   2,   2,   3,  10,  20,  17,  28, -15,  10, -14,  -4,  -2,  -1,  -1,  -2,  -2,  23,  18,   4,  38, -28, -11,   0,  -7,   1,  -3, -32, -16, -20,  28, -14, -20, -34,  18,   4,   8,   2,   2, -18, -30,   2, -11,  25, -12,  25,  18,  -4,  -1,  11,  22,   0,  -4,   8, -24,  -2,   2,  10, -15, -16,   3,  -5,   6,  -3,   3,   1,  -6,  10,   6,   0,   1,   0,   5,  -6,   0,  -3,   1,   3,   0,   2,  -1,   2,   1,   0,  -1,   1,   1 },
     {    4,   8,  11,  12,  30,   0, -25,  15,  -4,   1,  -2,   0,   2,   7,   9,   6,  29,  64, -29, -14, -13,   2,  -2,   0,  -5,  -1,  -2,  -8, -34, -21,  -9,  15,   3,   7,   1,   1,  -1,  10,  13,  18,  13, -10,   6,   4,  -9,   1,  -5,  -1, -28, -16,   0,  -6,   6, -12,  -8,   0,   2, -24,  25,   5,  12,  -1,   4,  -1,   1,  18,  15, -16,   0, -19, -11,  -7, -10,   5, -10, -13,  -8,  -8,   9,  11,   0,   3,  -5,   7,   0,   1,  -3,  -3,   0,  -2,   2,   0,   2,  -2,   1,  -1 },
     {    5,   3,   4,  10,   3,  14,  12, -12,   6,  -5,  -3,   1,   4,   3,   5,   3,   2,  24, -18, -14,  -3,  -3,   3,   0,   7,   4,   6,  13,  -6,   2,  11,  -9,   3,   4,   0,   0,   8,   3,   9,  21,   3,   8,  -8, -11,  -2,   1,  -2,   0,  19, -10,  -9,  20,   9,  -8,  -3, -10,  13,  55, -41,   9,  -5,  10, -33, -16,  10, -34,  -2,  10, -20, -27, -29,  23, -12,  -5,  17, -11, -11,  -3,  17,  18,   7,  -9,   0,  -3,  -3,   1,   2,  -4,  -3,  -2,   0,  -7,   1,   2,   0,  -4 },
     {   -3,  -5,  -6,  -4, -20,  -6,  -4,  15, -14,   7,   2,   3,  -2,  -5,   2,  -2, -10,   4, -11,  -7,   5,  -5,  -1,  -1,  -4,  -4,  -4,   6, -42,  12,  -2,  12,   2,  -8,   0,   0,  -2,  -4,   8,  33,  -4, -29,  25,  -4,   1,   4,   2,  -2,  -6, -26, -41,  14,  -4,  10,  -4,   4,  21,  20, -20, -13,   6, -30,  26,  15, -11,   5, -19,  29,  -5,  20,  27, -15,  -5,  -7,  22,  -3,  -3,  18, -13, -11,   0,   7,   0,  -8,   6,  -5,   3,  -2,   2,   2,  -1,   1,  -1,   4,   1,   1 },
     {    5,   5,   5,  14,  29,  14,   0,  -7,   3,  -6,  -3,   0,   1,  -2,  -2,  20,  13,   3,  -7, -25,  -5,  -5,  -2,   1,  -2, -10, -11,   1,  31,   1,  -8,   4,   4,   5,   0,   4,  -4, -19, -24,   0, -15,   6,  -8,   6,   5,   6,   2,   2,   6,  15, -41,  -5,  21, -15,   3,  12,  37,   1,   2, -39,  12,   4,  17,  -4,  -2,  -7, -30,   7,  43,   1,  -7,  -3,   0,   4,  -4,  30,  -6, -24,   3,  -3,  -8,  -6,   5,   6,   4,  -5,  11,   4,  -2,  -3,   2,  -8,   4,   3,   1,  -3 },
   }
 },
 {  //3
   {
     {   92, -57,   1,  -4,   1,  -1,   0,   0,   0,   0,   0,   0, -58,  16,  18,  -3,   3,  -1,   1,   0,   1,   0,   0,   0,   2,  20, -12,  -4,   0,  -1,   0,   0,   0,   0,   0,   0,  -3,  -5,  -5,   5,   1,   0,   0,   0,   0,   0,   0,   0,   1,   3,   2,   1,  -1,   0,   0,   0,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -36, -22,  26,   2,   4,   0,   2,   0,   1,   0,   1,   0, -30,  82, -30,  -9,  -2,  -2,   0,  -1,   0,  -1,   0,  -1,  34, -35, -28,  26,  -4,   4,  -1,   2,  -1,   1,   0,   1,   1, -12,  31,  -2,  -8,   2,  -2,   0,  -1,   0,   0,   0,   4,   1,  -5, -10,   3,   1,   1,   0,  -1,  -2,   3,   3,   2,   0,   0,   0,   2,   0,  -1,  -2,   0,   0,  -1,   0,   0,  -2,   2,   1,   0,   0,   0,   0,   1,   0,  -1,  -1,   0,  -1,   1,   0,   1,   0,   0,   0,   0,  -1,   1,   0 },
     {   48,  -7,  11,  -6,  -4,  -3,  -1,  -1,  -1,  -1,   0,   0,  28,  -8, -49,  11,   0,   5,  -1,   2,   0,   1,   0,   1,   7, -55,  53,  16, -10,   3,  -4,   2,  -2,   1,  -1,   1, -12,  22,  18, -45,   8,  -1,   1,  -1,   1,   0,   0,   0,  -7,   4, -20,   9,  18,  -5,   2,  -1,  -3,   6,   5,   4,  -7,  -5,   1,  -1,  -2,   0,  -5,   0,   0,   2,   2,   0,  -1,   2,   3,   1,  -1,   0,   0,  -1,  -1,   0,  -2,   0,   0,   1,   1,   0,   0,   0,  -1,   0,   0,   1,   1,   0 },
     {  -38,  -7,  25, -12,   2,  -1,   2,   0,   1,   0,   0,   0, -49,   8,   3,  18,   7,   2,   0,   1,   0,   0,   0,   0,  23,   8,  41, -36,  -9,  -3,  -1,  -1,  -1,   0,  -1,   0, -13,  24, -47, -17,  35,  -5,   6,  -2,   2,  -1,   1,  -1,   5,   0, -11,  45, -11, -12,   2,  -2,   2,  -1,   3,  -9, -18,   9,   1,   1,   4,  -3,  -2,   3,   5,   3,  -2,   0,   1,   0,   1,  -2,  -3,  -1,  -1,   0,   1,  -1,  -1,   2,   0,   0,   1,  -1,   0,  -1,  -1,   1,   0,   1,   0,  -1 },
     {   22,  -6,   2, -17,   4,  -1,   2,  -1,   1,  -1,   0,   0,  43,  27, -30,  14,  -7,  -4,  -3,  -2,  -1,  -1,  -1,   0,  -2, -25, -11, -21,  11,   4,   5,   1,   2,   0,   1,   0, -21,  15, -26,  55,  10, -13,   2,  -4,   1,  -2,   1,  -1,   8, -12,  22,  12, -55,  11,  -1,   2,  -1,  -5,   7, -26,  14,  23,  -7,   2,   2,  -2,   5,   4,   7, -10,  -7,   1,  -2,   0,   1,  -5,   0,   0,   4,   2,   1,  -1,   2,   2,  -1,   0,   0,  -3,   1,   0,   1,   1,   0,   0,   0,  -2 },
     {  -17, -74,  41,   6,   5,   2,   1,   2,   1,   1,   0,   0,  64,   6,  24, -34,   1,  -4,  -1,  -2,  -1,  -1,  -1,  -1, -19,  14, -14, -13,  10,  -3,   2,   0,   1,   0,   0,   0,  -5,   9,  -7,  -8,  12,  -1,   2,   0,   0,   0,   0,   0,  -6,   9, -19,  13,  13, -11,   2,  -2,  -1,   3,  -3,  15, -13,  -4,   4,  -1,  -1,   1,  -3,  -1,  -4,   6,   0,  -1,  -1,   0,   0,   3,  -1,   1,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   1 },
     {  -26, -44,  20,   4,   7,   6,   2,   1,   1,   0,   1,   1, -16, -24,  11,  21, -15,   1,  -2,   2,  -1,   1,   0,   1,  -2,  -6,  38,  -7,   5,   6,  -2,   0,  -1,   0,  -1,   0,   3,  29,   7,   2, -29,  -9,  -3,  -1,   0,  -1,   0,  -1,   9, -14,   6, -46,  -6,  37,  -3,   6,   1,   5,  -1,  -9,  53, -13, -15,   1,   0,   4,  -5,   6,  -7, -25,  11,   3,  -1,   4,  -4,  -1,   2,   5,   8,  -4,   0,   1,  -2,   2,   0,   1,  -2,  -1,   0,   0,  -1,   1,   0,   1,   0,  -1 },
     {   16,  28,   6, -19,   7,  -3,  -2,  -3,  -1,  -1,   0,   0,  16,  31,  -2,  -6, -18,   1,  -3,   2,  -2,   1,  -1,   0,  26,   4,  -3, -38,  14,  -1,  -2,  -1,  -1,   0,   0,   0, -10, -11, -39,  -2,   0,   6,   5,   4,   2,   2,   1,   1,   6, -24,  13,  -6,  45,  13, -11,   2,  -6,   9,  -5,  20,  18, -54,   4,   0,  -4,  -1,  -6,  11, -30,   3,  28,  -4,  -2,   3,   0,   5,   2,  13,  -5,  -9,   0,  -2,   0,   2,  -1,   1,   0,   1,   0,  -1,   0,   2,  -1,   0,   0,   1 },
     {    2,  16,  42, -14,  -1,  -7,   0,  -3,   0,  -2,   0,  -1, -17,  10, -46, -14,  15,   0,   7,   0,   3,   0,   2,   0, -66,  33,  -1,  37,   7, -10,   2,  -3,   1,  -2,   1,  -1,  17,  27, -25,   4, -31,   3,   2,   0,   0,   0,   0,   0,   8, -24,  -3,  12,   4,  14,  -4,   2,   6,   0,  11,   2,  -7,  -7,  -4,   1,   2,  -8,  -2,  -5,  -4,   7,   3,   1,   2,   1,   4,   0,   2,   2,  -3,  -1,   1,  -3,   0,  -1,   1,   0,   2,   0,   0,  -1,  -1,  -1,   0,   1,   1,   0 },
     {  -14, -25,  -8,   2,  17,  -4,   2,   0,   2,   0,   0,   0, -15, -23, -13,  21,  -6,   9,   5,   3,   0,   2,   0,   1,  -7, -24,   9,   4,  26,  -9,   0,  -2,   1,   0,   1,   0, -11,   8,   8,  37, -11,  -1,   6,  -1,   0,  -1,   0,   0,  14,   3,  33,  -9,   5, -21,  -9,  -1, -11,  15, -17,   4, -35, -20,  38,  -5,   1,   4,   1,   0, -18,  55,  -5, -17,   0,   3,   2,  -6,  10,  -3, -28,   6,   2,  -1,   1,  -3,   0,   1,   0,  -2,   1,   0,   0,  -1,   0,   1,   0,  -1 },
     {    1, -21,   3,   1,  -6,  -5,   3,  -1,   2,   0,   0,   0,  29,  10, -19,  31,  -9,   8,  -1,  -3,  -1,  -1,  -1,   0,  30,  41,  13,   1, -39, -13,  -4,  -3,  -1,  -1,  -1,  -1,  21, -23, -20, -20, -25,  34,   0,   3,   0,   2,   0,   1,   2, -26,  24, -17,  11,   0,  -7,   5,  -8,   4,  -7, -17,  -2,  20,  14,  -5,   0,  -4,   4,  -1,  27,  16, -29,  -2,  -2,   0,   0,  -4,   7, -23,  -3,  12,  -2,  -2,   0,   0,  -1,   1,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  -1 },
     {   14,  36,  16,  -4,  -2,  -9,  -2,  -2,  -1,  -1,  -1,  -1,  -2,  26,  35, -17, -21,   3,  -2,  -1,  -2,   0,  -1,  -1,  -2,  -7,  11, -21,  15, -10,   3,  -1,   2,  -1,   1,   0, -11,  40,   2,  -3, -22,  -5,   4,  -4,   1,  -1,   0,  -1,  -1,  -3, -36, -35,  -5,   3,  12,   4,  -9,  -1, -25,  25,   4,  33,  16, -11,   6,  -2,   5,   4,   7,  21, -39,  -4,  -1,  -6,  -2,  -2,   9, -18,  -8,  18,   1,  -1,   2,   1,  -1,  -1,  -2,   1,   0,  -1,   1,   1,  -1,   0,  -1,   0 },
     {   10,  16,  70, -26,  -8,  -8,  -4,  -3,  -2,  -1,  -1,  -1,  -1, -52,   2, -29,  21,   3,   6,   2,   1,   1,   1,   0,  61,  -2, -16,  17,  11, -13,  -1,  -3,  -1,  -2,  -1,  -1, -14, -16,   1,  11, -14,  -7,   7,  -1,   2,   0,   1,   0,  -3,  10,  10,  -1, -10,   8,   2,  -2,  -3,   3,  -4, -13,   1,   5,  -1,   0,  -2,   4,  -3,   4,   8,   3,   1,  -1,   0,   3,  -1,   1,  -1,  -3,  -4,  -1,  -2,   1,  -2,   1,   0,   2,  -1,   0,  -1,   0,  -1,   0,   0,   1,   0,   0 },
     {    8,  11,   5,  15, -11,   2,  -5,  -1,  -1,   0,  -1,   0,  15,  22,  17,   7,  -1, -25,   7,  -1,   0,  -3,   0,  -1,   6,  23,  12,  10, -24,   2,  -7,  -2,  -2,   0,  -1,   0,  11,   6,  12, -15,  -4, -26,  11,   1,   1,   0,   0,  -1,  -2,   7, -14,  -4, -37,   7,   0,  -2,  -1, -22,  -2, -37,   3, -12,  23,  16, -10,  12, -18,  14,  -5,  34,  39, -28,   1,  -1,   1,   1,   3,  27, -40, -10,  -2,  -1,  -3,  -1,   1,  -2,   1,  -1,  -1,   0,  -2,   0,   0,  -1,   0,  -1 },
     {   -3, -15, -48,  30,   7,  -2,   2,  -1,   0,   0,   1,   0,   8,  -7,  -3, -11,  -2,   1,   5,   0,   1,   0,   0,   0,  50,   9,  -5,  25,  15, -16,  -1,  -7,   0,  -2,   0,  -2, -16,  31, -23,   4, -39,  -9,  10,  -3,   4,   0,   2,  -1,  -3, -13, -26,  23,  -3,  32,  -1,  -2,   5, -11,  29,  16, -17,  -2, -18,   3,   1,  -5,   6, -21,  -6,  10,   1,   8,   0,  -1,   9,   0,  10,   2,  -4,   0,  -1,  -1,   2,  -6,   0,  -1,   3,   0,  -1,   0,   0,  -3,   0,   0,   1,   0 },
     {   -3,  -7,  -9, -49,  25,  -1,   7,  -1,   3,  -2,   2,   0,   9,  -5, -28,  38,  17, -19,  -3,  -1,   1,  -1,   1,  -1,  20,  38,  -7,   5,  -6,  -4,   1,  -3,   1,  -1,   0,  -1,  14,   6,  11,   7,   1, -23, -11,   1,  -1,   0,  -1,   0, -11,  17, -27, -16, -12, -16,  23,  -1,  -5,  -2, -28,  40,  -2,   7,  -2,  -5,   2,  -9,  13,   4, -32,  -5,   5,   8,  -1,  -2,  -3,  -7,  -1,  23,  10,  -1,   3,  -3,   8,   2,   1,  -1,  -1,  -3,   0,  -1,   3,   2,   0,   0,   0,  -1 },
     {    3,  -7,  -3,  16,   0,   4,  -4,   1,  -2,   1,  -2,   0,  -6, -14, -23, -27, -17,   7,   2,   4,   1,   2,   0,   1,  17,  -6,   3, -11,  32,  22, -11,   4,  -4,   1,  -1,   1,  80, -11, -11,   0,  19, -38, -10,   1,  -1,   1,  -2,   1, -13, -36,  -6,   2,  -7, -10,  23,   7,  -3,   8,  14,   5,   9,  10,  15, -12,  -7,   6,  -5,   8,  -3,   5, -10,  -9,   0,   1,   5,   3,   0,  -3,  -6,   5,  -3,   1,  -1,   1,   0,   0,   2,   1,  -1,   1,   0,   0,   1,   1,   1,   1 },
     {  -10, -12, -14, -19,  -2,  13,   1,   3,   1,   2,   0,   1,  -8, -17, -20,  -2,  11,  -5,  20,  -5,   4,  -2,   2,  -1, -13,  -3, -22,  -3, -17,  21,   1,   2,   0,   2,   0,   1, -21, -16,  -7,  -9,  15,  11,  24,  -4,   3,  -1,   2,   0, -14,   7, -16,  13,   6,  41,  -5,  -5,   6,  -5,  10,  19,  40,   5,   5, -10,   0,  21,   3,  34, -12,  14, -30, -28,   6,  -7,   8,  -9,  -2, -13, -39,  24,   2,   2,   2,   2,   3,  -2,   5,  -3,   0,   2,   0,   1,   2,  -1,   2,  -1 },
     {   -7, -10, -42, -20,  -9,  11,   7,   3,   2,   1,   0,   1,   1,   5, -11, -48,  45,  10,  -6,   3,  -2,   1,   0,   1,  14,  13,  11,  13,  10, -18,   8,  -2,   0,  -1,   0,   0,  -9,  34,   1,  -1,   7,  12, -11,  -5,  -5,   0,   0,   1,  25, -13,   3, -11,   3, -30,  -9,   9, -12,  -6, -15, -17,  22,  -2,  24,   1,  -3,   5, -12,  34,  14, -15,   3, -13,  -1,  -1,   3,  11, -24,  -3,   4,  -3,   1,   0,   0,   3,  -1,   0,   1,   1,   1,  -1,   0,   2,  -1,  -1,   1,   2 },
     {   -1,  -5, -13, -59,  26,   5,   3,   4,   2,   2,   0,   0,   6,  10,  54,  -4,  22, -21,  -6,  -5,  -4,  -1,  -2,  -1,  -7, -46,  19,  24, -16, -21,  12,   2,   0,   1,   0,   0,  43, -15, -16,   3, -12,   7,  12,  -7,   1,  -2,  -1,  -1,  -1, -18,   2,  12,  -2,  12,  -3,  -4,  -4,   5,  16,  -2,  -7,   2,  -1,  -5,  -1,   5,  -4,  -9,  -5,   4,  -3,   0,  -1,   3,  -1,  -1,   6,   0,  -1,   1,  -1,   4,  -2,  -1,  -3,   1,  -1,  -1,  -1,   1,   0,   1,   0,   1,   0,   0 },
     {    7,   4,  19,   8, -10,   2,   0,  -2,  -2,  -1,   1,   0,   0, -24,  10,  23,  -4, -12,  -5,  -4,   1,  -1,   0,  -1,  15, -20, -45,   2, -30,  25,  13,   3,   4,   0,   1,   1,  26,  47,  -6,   9,  14,  35, -24,  -8,  -3,  -3,  -2,  -1,  26, -19, -22,  -2,   3, -20, -30,  13,  -5, -26,  14,   7,   2, -13,  13,  10,   1,   4,  12,  -4, -10,   0,   1,  -5,  -5,  -6,   2,  -9,   2,   3,   3,   6,   1,   3,   4,   0,   0,  -2,   0,  -4,   1,   0,   1,  -1,   1,  -1,   0,  -1 },
     {    8,  13,  11,  14,   9,   4, -10,   0,  -3,   0,  -1,   0,  14,  21,  16,  13,  19,   7,   3, -13,   0,  -4,   0,  -3,  13,  24,  23,  18,  15,   2, -26,  -1,  -6,   0,  -2,  -1,  14,   9,  18,  11,  10, -13,  -1, -15,   0,  -5,   0,  -3,   5,  21,  11,  15,  -4,  -2, -35,   1,  12,   9,   9,  -7,   3, -32,  -6, -10,  -8,   8, -16,   5, -34, -13, -37,   8,   1, -17,   2, -24, -11, -23,  13,  41,  -3,   1,  -5,  -3,   0,  -4,  -1,  -4,  -2,   0,  -3,   0,   0,  -2,   1,  -2 },
     {    2,   3,  11,  44, -20,   4,  -5,  -1,  -5,   1,  -1,   1,  -1,   0,  12,   7,  27, -26, -16,   1,  -2,  -1,  -1,   1,  -4, -13,   0,  17, -25, -14,  14,  -4,   2,   0,   2,   1, -30,  -9, -22,  -2,  -8,  -9, -13,  13,   2,   5,   3,   1,  -1, -28,  20,  -2,  -3, -17,  43,  13,  -1,  23,   9,  14,   7,  10,  12, -32,  -3,   9,  -6,  11, -38,  -8, -12,   1,  -1,  -4,   4, -14, -11,  23,   5,   6,   3,  -2,   2,  -3,  -2,   0,   0,  -6,   1,  -1,  -1,   0,  -1,  -2,   1,  -2 },
     {    1,  -5,  -1,  35, -33,  -6,   2,  -3,   4,  -1,   0,   1,  -4,   3, -12, -11,  15,  -2,   3,   1,  -1,   1,  -1,   0,   2, -24,   8,  -8,  -7, -22,  20,  -2,   1,  -1,   0,   0,  41,  -1, -29,  15,  -8,  22,  17, -17,   4,  -3,   5,  -1, -30,  39,  -1,  -4,  -9,  10, -23,  -8,   1,  10, -45,  10,   2,   2, -11,  18,   3, -23,  10,  23, -12,  -2,   5,   6,  -2,   2,  11, -11,  -7,  14,   2,   1,   2,  -3,   2,   2,   1,   0,   1,  -3,  -3,  -1,   2,   1,   0,  -1,   1,  -2 },
     {    7,   7,  15,  19,  21,  -2,   5,  -9,   3,  -4,  -2,  -1,   8,  11,  14,  33,  16,  22,  -5,   6, -10,   0,  -5,   2,   5,   1,   7,  12,  22,  -6,   6, -16,   3,  -6,   0,  -2,  10,   4,   2,  11,   1,   7,  -5,   6, -12,   0,   0,   1,  -9,  -6,   3,   7,  14,  -3,   5, -21,   9,  -5,   7,  20,   3,  12, -13,   6,   6,  12,  29,   8,  18, -24,   8, -42, -11,  13, -18,  10, -47,  -4, -43, -10,   1,  -8,   5, -21,  -1,   3,  -5,   1,   0,  -4,   2, -10,  -1,  -1,  -5,   0 },
     {   -1,  -2,   5,  15,  29, -13, -17,   6,  -2,   0,   0,   0,  -4,  -5,  -2,  20,  14,   7, -10,  -5,   1,   0,   2,  -1, -13, -14,  -9,  -2,  10, -32, -24,  15,   0,   5,   0,   2, -14, -31, -16,  -8, -14, -18,   7,   1,  11,  -2,   1,   1, -24, -12, -30,  -2, -10, -10, -21,  36,   4, -24,   3,   3,  10,  -9,  41,  29,   0,   8,  18,  14,  13,  -6,   9,  -4,  -1,   1,   6,   2,  -9, -14,  32,   5,   2,   8,   6,   6,  -1,   1,   1,   2,  -1,   2,   1,   3,   0,   3,   2,   0 },
     {    0,   6,  -2,  -1,  34,  13,  -2,  -6,  -3,   1,  -1,  -1,  -4,   2,   4,   6,   8, -11,  -4,   1,   5,  -4,   2,  -1,   5,   2, -24,   3,  -2,  19,  12,  -5,  -1,  -1,   3,   1,  -8,  40,  -1, -13,   0,   2, -16,  -9,   0,  -1,   5,  -1, -84,  -9,  28,  -1,   4,  -5,  -8,  11,  13,  33,  -8, -13,   2,   0,   1,   9,   7,  -4, -27,   1,  14,   7,   3,   5,  -1,  -7,   6,  15,   5, -10,  -5,  -3,  -2,   1,  -2,   3,   3,  -2,   0,   3,   0,   2,   0,   0,  -2,  -2,  -1,   2 },
     {   -7,  -7,  -3, -35, -86,  18,  14,  12,   5,  -1,   4,   1,   1,   9,  13,  16,  -3,  29,  -3,  -3,  -4,   2,  -2,   1,   0,   4,   4,   6,  16,   0, -17,   2,  -3,  -1,  -2,  -1,  -7,   3,  11,  10,  -7,  -2,  -6,   2,   2,  -1,  -6,   0, -41, -16,  -3,   6,  -2,  -3,  -7,  18,   6,  14,  17,   9,  -3,  -1,  13,  14,  11,   7,  14, -14,  -7,   0,   6, -10,   1,  -1, -10, -17,   3,   8,   3,  -2,   2,  -3,  -3,  -3,  -4,   1,  -4,   2,   0,   2,  -1,   3,   0,   1,   3,  -1 },
     {    0,  -2,  -4,  -2, -22,   9,  -5,   7,   1,   0,  -3,   1,   2,  -1,   2,  39,  32,   7,  -8, -11,  -2,  -2,  -3,   0,  -6, -16, -53, -23,  20, -21, -19,  12,  -1,   0,   0,   3,  19,  29,   2, -40, -11, -26,  33,  16,  -2,   3,   0,   0,  21,   4,  17,   5,  10,  14,  14, -20, -10,   3, -10, -12,  -6,  -2,  -7, -11, -11,  -5, -12,   9,   7,   1,  -2,   7,  -7,  -1,  -1,  10,  -1,  -8,   4,   1,  -4,  -2,   3,   1,  -6,   1,   0,   3,  -2,  -1,   2,   1,  -2,  -2,   1,   0 },
     {    2,   0,   2,   2,   4,  51,  -6, -15,   0,  -4,   1,   1,   4,   3,  11,   0,  29,  25,   5, -11,  -4,  -2,  -3,   0,  -5, -16,   4,  14,  -7,  33, -39, -26,   4,  -5,  -1,   0,  -8, -15, -32,   4,  -3, -15, -40,  22,   5,   3,   3,   4,  15,   0, -11, -15,   1,  -1,   0,  -2,  -8,  -4, -24,   2,  -5,   0, -23,  19,   3, -13,   6,   7,  11,  16,   5,  18,   2,   1,  19,  -3,  17,  -5, -16,   7,   3,   3,   1,   0,   0,   3,   2,   0,   0,   0,   2,   1,   1,   1,   1,   0 },
     {   -5,  -7,  -8, -20, -18, -32,  -4,  -7,   3,   3,   4,   2,   1,  -2,  15,  21, -57, -12,   6,   4,   5,  -1,   0,   3,  -2,  -5, -14,  56,  25,   9,  -4, -12,   0,  -3,   0,  -1,  -7,  -3, -38, -14,   9,  -2,  -8,  -1,   4,   0,   4,   2,   8,  17,  -8,  -3,   3,  -6,   5,   2,   3,  11, -15, -14,  -1,   1,  -1,  -1,  -7,   3, -17,  29,   5,  -7,  -1,   8,   1,   1,  18,  12, -30,  -2,   5,   8,  -1,   0,   3, -13,   2,  -3,   0,   2,   0,   2,   0,  -2,   2,  -2,  -2,  -2 },
     {    0,  -1,  -1,  -4,  -5, -24,   5,  16,   0,   5,   5,  -2,   5,  10,   8,   2,  24,  -4,  17, -10,  -6,   0,   3,  -1,   4,  -3,   9,   1,  16,  36,  11,  -9,  -7,   6,  -3,  -2,   5,   9, -12,  -8, -12,   3, -12,   6,  10,   7,  -4,  -1,  -5,  32,  26,   0,   7,   4,  18,  21,  42, -40,   3,   1,   8,  -7,  23, -35, -18, -16,  30,  -1,   6,   6,   0,  -4, -12,  18,  16,  -7,  17, -14,   8, -22,  -8,   1, -10,   3,   1,   4,  -1,   4,  -1,  -2,  -3,  -3,  -3,   5,  -1,  -3 },
   },
   {
     {   99, -54,  -2,  -4,   0,  -1,   0,   0,   0,   0,   0,   0, -56,  17,  10,   1,   2,   0,   1,   0,   0,   0,   0,   0,   1,  12,  -6,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,  -4,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   3,   1,  -1,   0,   0,   0,   0,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   50,  -3, -14,  -1,  -3,   0,  -1,   0,  -1,   0,   0,   0,  71, -58,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -62,  11,  24,  -2,   3,  -1,   1,   0,   1,   0,   0,   0,  -3,  23,  -9,  -5,   0,  -1,   0,   0,   0,   0,   0,   0,  -6,  -3,  -3,   3,   1,   1,   0,   0,  -1,   4,   2,   0,  -1,   0,   0,   0,  -3,   0,  -1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0 },
     {   26, -17,   3,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,  50,  21, -32,  -1,  -6,   0,  -2,   0,  -1,   0,  -1,   0,  45, -73,   3,  10,   0,   3,   0,   2,   0,   1,   0,   1, -45,   5,  35,  -6,   3,  -1,   1,  -1,   0,   0,   0,   0,  -3,  19,  -9,  -6,   1,  -1,   0,   0,  -7,  -2,   1,   3,   1,   0,   0,   0,  -1,   4,   0,  -1,   0,   0,   0,   0,  -4,   0,   1,   0,   1,   0,   0,   0,  -1,   2,   0,   0,  -2,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 },
     {   30,  40, -34,  -1,  -6,   0,  -2,   0,  -1,   0,  -1,   0,  24, -12,  -3,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  26,   9, -34,   4,  -3,   1,  -1,   1,   0,   0,   0,   0,  26, -75,  12,  20,  -2,   5,  -1,   3,   0,   1,   0,   1, -29,   4,  42, -11,   1,  -2,   0,  -1,  -2,  13,  -9,  -8,   3,  -1,   1,  -1,  -7,  -1,   3,   3,   1,   0,   1,   0,   0,   2,   0,  -1,  -1,   0,   0,   0,  -3,   0,   2,   0,   0,   1,  -1,  -1,  -2,   0,   1,   0,   0,   0,   0,   0 },
     {   27,  82, -43,  -6,  -9,  -2,  -3,  -1,  -2,   0,  -1,   0, -49, -21,  -6,  15,   3,   5,   1,   2,   1,   1,   0,   1,   9, -31,  36,   1,  -1,  -1,  -1,   0,   0,   0,   0,   0,  -7,  23,   2, -17,   2,  -3,   1,  -1,   0,   0,   0,   0,   7,   0, -18,   5,   4,   0,   1,   0,   3,  -9,   4,   6,  -3,   0,   0,   0,   1,  -1,   3,  -2,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   20,  17,   2, -11,  -1,  -3,   0,  -1,   0,  -1,   0,  -1,  22,  49, -45,  -4,  -2,  -1,  -1,  -1,   0,   0,   0,   0,   5,   1,  -7,   0,   0,   0,   0,   0,   0,   0,   0,   0,  21,   7, -47,   4,   2,   1,   1,   1,   0,   0,   0,   0,  12, -66,   4,  35,  -2,   5,  -1,   3,  -7,  -4,  41,  -6,  -7,  -1,  -2,   0,  -2,   1,  -3,  -9,   2,   1,   1,   0,  -4,  -1,   5,   1,   2,   0,   0,   0,   0,   0,   0,  -1,  -2,  -1,   3,   0,   0,  -1,   0,   0,  -1,  -1,   2,   0 },
     {    2,  -9,  -1,   3,  -1,   1,  -1,   1,   0,   0,   0,   0,  13, -44,  37,   2,  -1,   0,   0,   0,   0,   0,   0,   0,  79,  26,  -1, -35,  -5,  -7,  -1,  -3,   0,  -1,   0,  -1, -35,   9, -44,  -2,  13,   2,   5,   0,   2,   0,   1,   0,  -6,  -7,  -5,  27,   0,   0,  -1,   0,  -2,  -4,  12,   0,  -9,   0,  -1,   1,  -3,   2,   2,  -4,   1,   2,   1,   0,  -2,   0,  -2,   0,   1,   0,   0,   0,  -2,   1,   1,   1,  -1,   1,   0,   0,  -1,   1,   0,   1,   0,   0,   0,   0 },
     {  -11, -25,  16,   3,   0,   1,   0,   1,   0,   0,   0,   0, -21, -32, -11,  26,   0,   6,   0,   3,   0,   1,   0,   1,   8, -34,  48,   2,  -4,   0,  -2,   0,  -1,   0,  -1,   0,  16,  -3,  -2,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,  -4, -23,  52,  -1,  -6,   0,  -2,   0, -14,  50,   7, -44,   4,  -5,   2,  -2,  -1,   8, -32,   2,  10,   0,   3,   0,  -2,   5,   0,   4,  -1,  -1,   0,  -1,   2,   1,  -5,   0,  -1,   3,   0,   0,   1,   1,  -3,   0,  -1,   2,   1,   0 },
     {   -2,  16, -14,  -1,   1,  -1,   1,   0,   0,   0,   0,   0, -10,   5,   5,  -1,   2,  -1,   1,   0,   0,   0,   0,   0, -30,   1, -23,  -3,  11,  -1,   4,   0,   2,   0,   1,   0, -78,   1,  -4,  50,   4,   2,   1,   0,   0,   0,   0,   0,  44,   2,  39,   7, -30,   0,  -6,   0,   5,  -1,   4, -32,  -2,   6,   1,   1,   6,   2,  -5,   0,   8,   1,   0,  -1,   1,  -1,   1,  -1,  -1,  -1,   0,   0,   2,   1,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0 },
     {  -14, -25, -29,  26,   2,   5,   1,   3,   1,   1,   0,   1, -10, -26,  17,   4,   3,   0,   1,   0,   1,   0,   0,   0,  -9, -20, -16,  26,   2,   2,   0,   2,   0,   1,   0,   0,   9, -17,  28,   4,  -7,   0,  -2,   0,  -1,   0,  -1,   0,   2, -10,  -7,  20,   0,   0,   0,   1,   8, -19,  65,   6, -25,   1,  -4,   0,  -6,  28,  12, -53,   0,   2,   1,  -1,  -2,   5, -19,  -2,  14,   1,   2,   0,  -1,   8,   0,   0,   0,   0,  -3,   0,   0,   3,   0,  -2,   0,   0,  -2,   0 },
     {   -2,   9,  25, -11,  -3,  -3,  -1,   0,   0,   0,   0,   0, -12,  12, -16, -10,   8,   0,   2,  -1,   1,   0,   1,   0, -25,  -2,   9,   0,   0,   1,   0,   1,   0,   0,   0,   0, -27, -20, -14, -12,  23,   2,   6,  -1,   2,   0,   1,   0, -63,  37,   5,  50,   9, -10,  -1,  -4,  41,  21,   8,   0, -35,  -3,  -5,   0,   4, -12,  -8, -14,   0,   8,   1,   2,   5,   2,   2,   2,   3,   0,  -1,   0,   2,  -1,  -3,   0,   2,   0,   1,   1,   0,  -1,  -1,   0,   2,   0,   1,   1 },
     {    0,   7, -39,  29,  -4,   5,  -1,   2,   0,   1,   0,   0,  15,  53,  61, -29, -15,  -6,  -5,  -3,  -2,  -1,  -1,  -1, -14, -11,  17, -33,   1,   2,   1,   1,   1,   0,   1,   0,   3,   0,   2, -17,  12,   0,   2,  -1,   1,   0,   0,   0, -10,   1,  -7,   0,   7,  -2,   1,   0, -13,  20,   5, -16,   4,  -2,   1,  -1,  -3,  15, -39,  -2,  14,  -2,   3,  -1,   2, -11,  -7,  24,  -1,   0,   1,   0,   1,   0,   2,   1,   0,  -3,  -2,   4,   0,   0,   0,   0,   0,  -2,   0,   2 },
     {  -19, -36, -68,  21,  17,   8,   6,   3,   2,   1,   1,   1,   2,  20, -19,  41,  -9,  -1,  -2,   0,  -1,   0,  -1,   0,  -8,  -4,  34, -25,  -6,   0,   0,   0,   0,   0,   0,   0,  -4, -14, -22,  12,  12,   1,   3,   1,   1,   0,   1,   0,  -6,   3,  11,   6,   0,  -1,   0,  -1,   8,  -3, -14,  23,   0,  -2,  -1,   0,  10, -20,  34,  10, -22,   1,  -3,   1,  -3,   6,   9, -22,   0,   3,  -1,   0,   0,  -1,   0,   0,  -1,   2,   2,  -3,   1,   0,   1,   0,   0,   1,   1,  -3 },
     {    0,   3,  17, -14,   1,  -1,   0,   0,   0,   0,   0,   0, -14, -15,  31, -26,  -4,   4,   0,   2,   0,   1,   0,   0, -17, -58, -24, -16,  34,   6,   7,   3,   3,   1,   2,   0,   6,   3, -24,  33,  11,  -3,   0,  -1,  -1,  -1,   0,   0, -21, -20,   6,   3,  11,   2,   1,   1, -27,  21,  -2,  41,  11, -13,   1,  -3,  13,   7,  21,  11, -31,  -4,  -1,   0,   1,   4,   3, -15,  -3,   5,   2,   0,   2,   2,   0,   1,   0,   3,   0,  -2,   1,   1,   0,   1,  -1,   1,   0,   0 },
     {   -3,  -9, -37,  19,   6,   1,   1,   1,   1,   1,   0,   0,  -2,  -4,  -6,  -9,  10,  -4,   3,  -1,   1,   0,   1,   0,  19,  -1, -31,  32,  20,  -7,   4,  -3,   1,  -2,   0,   0,  22,  62,  -2,  26, -20, -18,  -3,  -5,   0,  -2,  -1,  -1, -43,   5,   1,   4, -28,   2,   4,   2,  15,  13,  -4, -12, -11,  11,   1,   0,  13, -21, -18,   9,   4,   4,  -2,   0,   4, -14,  23,   9, -11,   2,  -3,   1,   1,  -1,   5, -15,   0,  -4,   2,   0,   1,  -1,   1,  -4,   0,  -1,   1,   0 },
     {    6,  -4,  13,  -8,  -1,   1,  -1,   1,  -1,   0,   0,   0,   6,  -4,  16,  -2,  -7,   0,  -3,   0,  -1,   0,  -1,   0,   0, -41,   5, -18,   7,   7,   0,   2,   1,   1,   0,   1,  52, -10, -20,  16,  26, -12,   1,  -3,   0,  -1,   0,  -1,  26,  33,  -5,  22, -38, -12,   0,  -3,  22, -42, -11, -27,  -9,  22,   3,   5, -22,  -8,  10,   9,  25,   0,  -1,  -2,  -3,  11,   8,  -1,  -5,  -5,  -2,  -1,  -6,   3,  -2,  -5,  -1,   2,  -2,   0,  -3,   1,  -1,   0,   0,   2,   0,   0 },
     {  -12, -18, -22,   6,  -1,   5,   1,   2,   0,   1,   1,   1, -13, -26, -21, -13,  18,   3,   5,   1,   2,   1,   1,   0, -11,  -7, -18,  16,  -3,   3,  -1,   1,   0,   1,   0,   0, -13, -32,  -2, -21,  22,   8,   2,   2,   1,   0,   0,   1,  13,  -8, -25,  30,  17,  -6,   1,  -2, -20,  -6,   6,   3,  11,  -1,   2,   0, -10,  12, -18,  51,  13, -11,   1,  -3,  14,  -9,  39,  23, -37,  -1,  -6,   0,   2,   6,  12, -22,   1,   1,   2,   2,   0,   3,   3,  -4,   2,  -1,   2,   1 },
     {   -6,  -2,   3,   1,  -8,   4,  -1,   1,   0,   0,   0,   1, -10,   2,  -5,   3,  -6,   4,   2,   0,   0,   0,   0,   0,  -9,  -5,   6, -10,   1,   5,   2,   0,   1,   0,   0,   0, -22,   1,  -8,   1, -11,   3,   2,   0,   1,   0,   1,   0, -63, -23,   2, -21, -10,  31,   2,   6,  19, -41,  13, -15,  50,  18,  -8,   2, -19,  22,  17,  13,  29, -26,  -5,  -4,   5,  19, -10,   7,  -6, -11,   2,  -1,   1,   1,  -2,   6,   3,   3,   0,   2,   0,   1,   1,   2,   2,   0,   0,   0 },
     {   -6,  -1, -12, -41,  23,  -1,   6,   0,   1,   0,   1,   0,  -3,  -1, -30,  23,   7,  -5,   1,  -1,   0,   1,   0,   0,  -2,  17,  -6, -25,   6,   4,   3,   1,   0,   1,   0,   0,   9,   8,   5,  15,  26, -16,   1,  -5,   0,  -2,   1,  -1, -19,  30, -14,   5, -26, -11,   5,  -1, -54,   1,   8,   1,   3,  15,   4,   0,   8,  44, -13,  -9,   0,  -2,  -4,  -2,   9, -15, -31,  14,   4,   4,   2,   2,   7,  -8,  12,  16,   4,  -4,   0,  -4,   2,  -1,   3,   2,   1,  -2,   0,  -2 },
     {    9,  18,  27,  27, -14,  -1,  -6,  -1,  -3,  -1,  -1,   0,   8,  18,  20,  16,   2, -12,  -3,  -3,   0,  -1,  -1,  -1,   3,  18,  22,  13, -15,  -4,  -4,  -2,  -1,  -1,  -1,   0,   0,   6,  23,   4,   7, -11,  -5,  -2,  -1,  -1,  -1,  -1, -13,   5,  10,  18, -27, -12,   1,  -1, -21,  -4,  21,  -4,  -1,   2,   0,  -1,   3,  29,  18,  18, -30, -17,  -3,  -3,  15,  -7,  14, -46, -40,  13,   0,   4,  -4,   1, -17, -33,   1,  -4,  -9,  -2,  -4,   0,  -4,  -4,   1,  -1,  -2,  -2 },
     {    4,   3,  14,  73, -37,   0, -10,   0,  -4,   0,  -2,   0, -10, -10, -47, -33,   6,  15,   5,   5,   1,   2,   1,   1,   4,  15,  15, -38,  38,  -1,   1,  -5,   2,  -1,   0,  -1,   1,  -7,  13,  20,   6, -12,  -3,  -3,   0,  -1,   0,   0,   3,   1, -12,  -2,  -7,  -5,   4,  -2, -17,   5,  -4,   4,  -6,   5,   0,   0,  -8,  -5,   7,  -7,   5,   9,  -2,   1,   4,   6,  -6,   4,   9,  -6,  -3,   0,   1,  -5,   2,   8,   0,   3,   3,  -2,  -2,   3,  -1,   2,   1,   0,   0,   0 },
     {    1,  -3,  -8,  -5,   2,   3,  -2,   1,   0,  -1,   0,   0,   3,  -1, -18, -10,   6,   6,  -2,   1,   0,   0,   1,   1,  11,   6,   4, -11,  22,   0,  -2,  -1,  -1,  -1,  -1,   0,  14,  14,  -2,  13,   4,  -4,  -3,  -2,   0,  -1,   0,  -1,  20,  23,  13,  -2,  18,   5, -13,   1,  57,   0,  12,   4,  22, -46,  -9,  -1, -27,  28, -21,   5, -34, -34,  22,  -2,   3, -11, -23,   4, -22,  21,   9,  -1,  -1, -13,   8,  -1,  -1,  -3,  -1,  -2,  -3,  -1,   1,   1,   0,  -2,  -2,   0 },
     {   -6,  -8, -19, -29,  11,   3,   1,   2,   0,   2,   0,   0,   4,  11,   7, -52,  45,  -1,   3,  -1,   0,   0,   0,  -1,  10,   8,  52,  37,  17, -29,  -3,  -7,  -1,  -2,  -1,  -2,  -5, -19, -17,  18, -30,  -9,   5,   1,   2,  -1,   1,   1,   2,  11,  -7,   3,  -6,   5,   0,   0, -21,   0,  -6,  -1,  -5,   0,   1,   0, -22,  10,  11,   0,   1,   5,   2,   1,  -4,  27,  -5,  -7,   0,  -2,   0,  -2,   2,  -2, -16,   3,   1,  -1,   1,   3,   1,   0,  -2,   2,   1,   1,   1,   2 },
     {   -3,  -1, -13, -14,  29,  -3,  -1,   0,   2,   0,   1,   0,  -2,   3,   1, -11,  -5,  10,  -1,   1,  -2,   2,   1,   1,   3,  14, -15, -11,  13,  16,  -7,   2,  -2,   0,  -2,   0,  -3,  32,  43, -18,  31, -15, -13,  -1,  -3,  -2,  -1,   0,  -6, -30,  27,  13,  16, -26,  -2,   1,   0,   1, -17,  -4, -20,  -9,  15,   0, -48,   2,  22,   9,  12,   6,  -2,  -2,   9,  39,  -3,  -9,  -3, -14,   4,  -3,   6,  -3, -21,   3,  -2,   1,   5,   4,   1,  -3,  -5,  -2,   3,   0,   1,   3 },
     {    1,   4,   1,  37,  -7, -12,  -1,  -2,   0,  -2,   0,   0,  -3,   6,  -6,  12,   0,  -6,   0,   0,   0,   0,   0,   0,  -7, -13, -24,  23, -21,  -8,   7,   0,   2,   1,   1,   0,   2,  26, -49, -11, -11,  33,   4,   3,   1,  -1,   0,   0,  -4,  41,  28,  -8,  26,  -7, -13,  -5, -39, -15,  -6,  -4, -12, -13,   4,   2, -37,  14,  10,  -1,   4,   7,   6,   2,  14,  13, -13,   2,   0,  -3,   1,  -1,   4, -14,   2,   7,   5,   2,   3,  -2,   2,  -6,   2,  -1,  -2,   0,   4,  -1 },
     {    4,   8,  21,   7,  28, -15,   4,  -4,   0,   0,  -1,  -1,   1,   6,  12,  41, -20,   0,  -8,  -1,  -3,   0,  -1,   0,  -5,  -1,  15,   1,  21, -17,  -3,  -3,  -1,  -2,   0,  -1, -13, -18,   0,  23, -27, -23,   4,   1,   1,  -1,   1,   0, -10, -18, -15,  -6,  -2,   1,   4,   5,   4,  -7,  -9,  18, -21,  -8,  16,  -2, -51,   6, -11,   0,  -8,  27,   5,  -1,  12,  -8,  15,  34,  -3,   9, -12,   2,   5,  -1,  34, -19,  -1,   8,  -5, -11,   0,   0,   5,  -3,  -1,   3,  -1,  -3 },
     {    3,   1,  -3,  21,  20, -25,  -4,   2,   2,  -2,  -1,   1,  -1,  -6,  -2, -10,   8, -13,   8,  -3,  -1,   0,   3,  -1,  -3, -13, -20,   0, -30,  -1,  -3,   4,  -1,   4,   0,  -1,  -1,  -9,  -9, -20,  -6, -19,  11,   1,   5,  -1,   1,   1,  14,  -8, -25,  13, -27,  35,  13,   6,  25,  43, -21,  -9,   6,  13,   4,  -7, -20,  23,  10,   8, -14,   4, -18,  -7,  29,  -5, -16, -24,  11,   0,   4,  11,   3, -18,   0,  24,  -8,   5,  17, -16,   6,  -3,  -7,   0,   5,   0,   6,   3 },
     {   -3,  -1,  -1,   2, -14,  13,  -1,   1,  -2,   3,  -1,   0,  -5,  -4,   0,   0,  10, -18,   3,  -2,   4,  -2,   1,  -1,  -6,  -1,  -8,   4,  -8,   9,   0,   4,  -2,   1,   0,   1,  -6,  -2,   3,  -8,  12, -17,   4,   0,   0,  -1,   1,  -1, -16,  -7,  -5,   1, -20,   1,  -3,   4, -13, -28,   0, -16,  18, -24,  10,   6, -26, -13, -27,  10, -34,  27,  34,   4, -11,  -5,  20, -17,  42,  39,  -3,  -8,  -8,  28, -15,  34,  12, -11,  10,   8,  -3,   5,   2,   3,   3,  -3,   3,  -1 },
     {   -2,   0,  -8, -17,  22,  -8,  -1,  -1,  -1,   0,   0,   0,  -5,  -5, -15, -17, -45,  11,   5,   3,   2,   3,   0,   0,   0,   7,  -5, -30,  17,   0,   0,   2,   2,   1,   0,   0,   6,   1,   4, -10, -29,  10,  15,  -2,   5,   0,   0,  -1,   5,  33,   0,   0,   7,  36, -12,   3, -10,   6,  28, -15,   7, -10,  -6,   0,  -3,   7,   1,  -2,   0,  27,  -2,  -2,  27,  -3,  34, -26,  13, -10, -22,  -1,  -8,  13, -25, -27,  -1,   0, -12,  17,  -2,   6,   0,   0,   2,   0,  -4,   7 },
     {   -5, -10,  -9, -19, -41,  13,   2,   5,   0,   0,   0,   2,   1,   1,  -5, -16,  10,  10,  -1,   1,   0,   2,   0,   1,   2,  -6,   6,  -7, -14, -13,   8,   0,   3,  -3,   0,  -1,   2,   8,  -4,   0,   4,  14,   2,   4,  -2,   2,   0,   0,  -3, -16,  20, -20,   5, -24,   4,   3,  31, -18, -12,  13, -13,  14,  35,  -5,   9,  50,  -8,   8,  -1,  46, -17, -15,  21, -23,   3, -15,  10,   1, -30,   7,  -6,  -6,   7,  -6,  -5,   3,  -4,  -3,  -1,   2,   0,   1,   1,   1,  -5,   0 },
     {   -5,  -6,  -9, -26, -71,  32,  13,   0,   5,   3,   1,   3,   4,  10,   6,  41,   5,  -3,  -4,  -7,   1,  -5,   0,  -1,  -3,   1, -21,  -7,  17, -19,   8,   2,   2,  -1,   1,   0,   0,   5,   0,  -1,  -7,  -7,   5,   1,  -2,   3,   1,   0,   9,   7, -23,   0,   2,   4,   4,  -2,   2,  44,   8,   0,   0,  11, -10,  -5, -35,  -6,   9,  -2,   1,  -6,  -5,   4,  11,  23,   4,   4,  -2,  -4,   2,  -1,  -4,   6, -19,  -3,   3,  -7,  -1,  -1,   7,   2,   5,  -1,   8,  -3,  -2,  -1 },
     {   -4,  -6,  -8, -11, -31, -10,   6,  -1,   2,   4,   2,   2,   2,   3,  10,  -4, -48,  20,  36,  -7,   4,   1,  -1,   0,   8,  12,  20,  49,   6,  53, -11, -16,   0,  -2,  -2,   0,  -4,  -4,  -9,  15,  23,   0, -25,  -7,   3,   0,  -2,   1,  -1,   1, -14,   3,  -5,   8, -19,   6,  -6,  10,  -4,  -6,   4,  -9,   0,  17,  -3,  -3,  10,   1,  -3,   7,   2,  -7,  20,  -5,  -8,   5,   7,  -3, -11,   4,   1, -18,  13,   7,  -3,  -3,   8,  -8,  -4,   4,   2,  -2,   0,   0,   2,  -2 },
   },
   {
     {  -83,  64, -13,   6,  -3,   2,  -1,   1,  -1,   1,   0,   0,  60, -36,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -12,  -2,   8,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   5,   1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   71,   7, -39,   8,  -9,   3,  -4,   1,  -2,   1,  -1,   0,  34, -65,  32,  -7,   5,  -2,   2,  -1,   1,   0,   1,   0, -42,  33,  -1,  -3,   2,  -1,   1,  -1,   0,   0,   0,   0,   6,  -5,  -3,   2,   0,   0,   0,   0,   0,   0,   0,   0, -10,   6,   2,   0,   0,  -1,   0,   0,   1,  -1,  -1,   0,  -1,   0,   0,   0,  -4,   2,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   43,  -3,  12, -21,   5,  -7,   2,  -3,   1,  -2,   0,  -1,  63,   4, -61,  20, -11,   6,  -4,   3,  -2,   1,  -1,   1,  -2, -53,  37,  -1,   4,   1,   1,   1,   0,   0,   0,   0, -20,  24,  -2,  -3,   1,  -1,   1,  -1,   0,  -1,   0,   0,   0,  -9,   3,   2,   1,   0,   0,   0,  -7,   8,   1,  -1,   0,  -1,   0,   0,  -1,  -3,   1,   1,  -1,   0,   0,   0,  -4,   3,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,  -2,   2,   0,   0,   0,  -1,   0,   0,  -1,   1,   0,   0 },
     {   25,  75, -65,  11, -14,   3,  -5,   1,  -3,   1,  -1,   0, -54,  12,  -7,  10,   1,   3,   0,   1,   0,   1,   0,   1,  18, -36,  20,   2,   1,   0,   0,   0,   0,   0,   0,   0,  -9,   4,   7,  -4,   1,  -1,   1,  -1,   0,   0,   0,   0,   6,  -4,   2,   0,   0,   0,   0,   0,  -1,   0,   2,  -2,   0,   0,   0,   0,   3,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {   20,  17, -14,   6, -10,   3,  -4,   1,  -2,   1,  -1,   0,  45,  44, -17, -35,   8, -11,   3,  -5,   1,  -2,   1,  -2,  19,  -6, -65,  32,  -8,   8,  -1,   3,  -1,   1,   0,   1,   4, -45,  35,   5,   1,   2,  -1,   1,  -1,   1,   0,   1,  -9,  10,  -2,   0,  -1,  -1,   1,  -1,  -1, -11,   9,   2,  -1,   2,  -1,   1,  -5,   4,   1,  -1,   1,   0,   0,  -1,  -1,  -5,   4,   1,   0,   0,  -1,   1,  -2,   2,   1,  -1,  -1,  -2,   2,   1,  -1,   1,   1,   0,   0,  -1,   1,   0 },
     {   16,  38,   6, -32,   9, -11,   4,  -5,   2,  -3,   1,  -2,  12,  35,   1,   1, -20,   5,  -7,   2,  -3,   1,  -2,   0,  12,  42, -21, -48,  17,  -9,   6,  -3,   2,  -1,   1,  -1,  -4, -11, -58,  37,   2,   6,   1,   2,   1,   0,   1,   0,   5, -25,  22,   6,  -1,   1,  -1,   1,  -5,   1,  -8,   5,   0,   0,   1,  -1,   2,  -9,   9,   2,  -1,   1,  -2,   1,  -2,   0,  -2,   2,   0,  -1,   1,   0,   1,  -4,   4,   1,  -1,   0,  -1,   0,   0,  -2,   2,   1,  -1,   0,  -1,   0 },
     {    1,  30,  23, -35,  12, -11,   4,  -5,   2,  -2,   1,  -1, -30,  18, -38,  14,   3,   4,   2,   1,   1,   0,   1,   0, -89,  16,  -1,  24,   1,   5,  -1,   2,   0,   1,   0,   1,  26, -13,  17,   2,  -2,  -2,  -1,  -1,  -1,   0,   0,   0,  -9,  10,   1,   1,  -4,  -1,   0,  -1,  14,  -5,   1,  -2,  -2,   1,   0,   0,  -1,   1,   0,  -1,   0,   1,   0,   0,   7,  -3,   1,  -1,  -1,   0,   0,   0,  -1,   0,   0,  -1,   3,  -1,   0,   0,  -1,   0,   0,   0,   2,  -1,   1,   0 },
     {  -15, -30, -33,   7,  23,  -5,   9,  -2,   4,  -1,   2,   0, -10, -31, -23,  35, -12,  15,  -6,   6,  -2,   3,  -1,   2,   2, -15, -26,   9,  24,  -9,   7,  -3,   3,  -1,   2,   0,   0, -26,  20,  55, -25,   7,  -8,   3,  -3,   1,  -1,   1,   9,   0,  45, -28,  -7,  -4,  -1,  -1,  -6,  12,  -7,   0,  -4,   3,   2,  -1,   5,  -1,  13,  -9,   0,   1,  -3,   1,  -2,   5,  -3,  -1,   1,  -2,   1,   0,   3,   0,   5,  -4,  -1,   3,  -2,  -1,   2,   0,   2,  -2,   0,   1,  -1,   0 },
     {   -5,   0,   6,  16, -12,   6,  -5,   2,  -2,   1,  -1,   1,   1,  24,  32, -49,  10,  -7,   3,  -2,   0,  -1,   0,   0, -55, -33, -15,   7,  23,  -3,  10,  -1,   4,  -1,   3,   0, -50,  40,  -2,  30,  -8,   2,  -3,   0,  -1,   0,  -1,   0,  24, -12,  12,  -9,  -4,  -2,  -1,   0, -10,  13,  -6,   1,  -1,   1,   1,   0,  12,  -7,   0,  -2,   0,   1,  -1,   0,  -3,   4,  -2,   1,   1,   0,   0,   0,   5,  -2,   0,   0,  -2,   1,  -1,   1,   2,  -1,   0,   0,  -1,   1,   0,   1 },
     {  -14, -34, -53,  56, -13,  16,  -5,   8,  -2,   3,  -1,   2,  19,  58, -23,   9,  -9,  -2,  -3,  -1,  -2,  -1,  -1,   0, -32,  28,  31, -17,  -4,  -7,   2,  -3,   1,  -1,   0,  -1,  24,   2, -12,  -8,  -4,   3,   0,   1,   0,   0,   0,   0,  -5, -11,   4,  -3,   0,   1,   0,   1,   3,  -2,   0,  -2,   2,   1,   1,   0,  -2,  -3,   2,   0,   1,   1,   0,   0,   1,  -1,   1,  -1,   1,   0,   0,   0,  -1,  -2,   2,   0,   1,  -1,   1,   0,   0,  -1,   1,   0,   1,   0,   0,   0 },
     {   12,  18,  19,   0,   4, -14,   4,  -5,   1,  -2,   1,  -2,  11,  19,  37,  -5, -32,  10, -12,   4,  -4,   2,  -2,   1,   5,  -6,  19,  -1,   9, -14,   5,  -4,   2,  -1,   0,  -1,  35,   2,  29, -17, -45,  12,  -9,   3,  -3,   1,  -2,   1,  -8,  -3, -11, -60,  32,   2,   6,   2,  -7,   6, -27,  23,   9,  -2,   2,  -2,  -1,  -5,   1,  -7,   5,   0,   0,   2,  -3,   3,  -9,   8,   3,  -2,   2,  -2,   0,  -2,   0,  -2,  -1,   1,  -4,   3,   0,  -2,   0,  -1,  -1,   1,  -2,   2 },
     {   -3,  -7,  11,  10,  -6,   1,   0,   1,  -1,   1,  -1,   0,   0,   7,  33,   1, -27,   7,  -7,   2,  -3,   0,  -1,   0, -22, -57,  12, -48,  18,   7,   5,   3,   1,   2,   1,   1,   2, -59,  -3,   7,  31,   2,   7,  -1,   2,   0,   2,   0, -27,  34,   3,  20,   6,  -3,  -2,  -3,  15,  -3,   6,   2,  -3,  -5,  -3,   0,  -8,  12,   0,   0,  -2,  -2,   2,   0,   6,  -1,  -1,   1,  -1,   0,  -1,  -1,  -3,   5,   0,   0,   2,   0,  -1,   1,  -1,   2,   0,   0,   1,   0,  -1,   0 },
     {   -6,   1, -18, -14,  23, -10,   9,  -5,   4,  -2,   2,  -1,  12,  21,  21,  35, -29,   8,  -8,   0,  -3,  -1,  -1,  -1,  -6,  37,   6,  -8, -14,  -1,  -6,   1,  -3,   1,  -2,   0, -80,  -2,  42, -11,   6,  -4,   5,  -2,   3,   0,   1,   0,   9,  28, -15,  -8,  -8,   5,  -3,   1,   0, -13,   3,  -7,   6,   2,   0,   1,   7,   8,  -8,  -1,   1,   0,   0,   0,  -1,  -7,   0,  -2,   2,   0,   0,   2,   2,   2,  -1,   0,   0,  -3,  -1,   0,   1,   1,  -1,   0,  -1,  -2,   0,   0 },
     {  -12, -14, -29, -17,  17,   0,  10,  -2,   4,   0,   2,   0, -10, -12, -26, -27,   3,  19,  -2,   7,   0,   2,   0,   2,  -9,   6, -13,  -4,  12,  -3,   6,  -3,   1,   0,   1,   0, -24,  -9,  -6, -30,  11,  22,  -3,   6,  -1,   2,   0,   2,  -1,  -7, -17,  15,  69, -14,   2,  -3,  -3,  14,   6,  50, -15, -25,  -3,  -5,   6,  -5,   7,  -4, -13,  -3,   5,   2,  -2,   4,   1,  10,  -5,   1,   3,  -5,   2,  -1,   3,  -2,  -1,   2,   1,   4,   2,   0,   2,  -1,  -1,   1,   0,   2 },
     {    7,  15,  43,  40, -33,   7, -12,   2,  -5,   0,  -2,   0,  -6,  -8,  -5,  33,   3, -14,   3,  -1,   1,  -1,   1,   0, -10,  10,   2,  10, -37,   6,  -2,   1,   0,   0,   0,   0, -36, -43,  -4, -17,   3,  19,  -1,   9,  -1,   4,  -1,   2,   3, -33,  35,  -8,  26,   0,   4,  -2, -14,  26,  -5,  -1,   2,  -4,  -2,  -2,   9,  -7,   6,  -4,   0,  -4,   1,   0,  -2,   9,  -2,  -5,   0,  -1,   1,  -2,   4,  -2,   0,  -2,   0,   3,  -1,  -2,   1,  -1,  -1,  -1,  -1,   2,   0,  -1 },
     {   -4,  -1,   2, -29, -17,  23,  -9,   8,  -2,   3,  -1,   1,  -2,   5, -14, -37,  43, -10,  11,  -4,   2,  -1,   1,   0,  19,  26,  60,  -7,   9, -11,  -1,  -8,   0,  -4,   0,  -2, -23, -23,  19,  26,  -8,   1,  -9,   2,  -4,   1,  -1,   0, -31,  30,  25, -14,   1, -12,   1,  -4,   4,   9, -20,  -1,  -9,   1,  -1,  -1,  -5,   1,   1,  -3,   3,  -1,   0,   0,   0,   0,  -9,   4,  -3,  -1,   0,   0,  -3,  -2,   1,  -1,   1,   1,  -4,   1,  -2,  -1,   0,   1,  -1,   1,  -2,   1 },
     {    2,  -1, -12, -11,  41, -19,  11,  -7,   3,  -3,   2,  -2,  17,  24,  48,  13,  15, -10,  -8,  -4,  -4,  -1,  -3,  -1,   3,  -8,  31,  47, -19,  -6,  -8,  -4,  -2,  -1,   0,  -1,  16, -11,  -9,   6,  10,   2, -11,   6,  -4,   1,  -1,   0,  11, -10,  20,  32,   3, -31,   0,  -4,  -4,  15,   7,  -4, -45,  -2,   2,   0,   4,  -7,   0, -18,  -3,   5,   4,   3,  -2,   1,  -4,  -2,  -8,   3,   2,  -1,   1,  -5,  -1,  -7,  -2,   0,  -2,  -1,   0,  -2,  -1,  -3,  -1,   0,  -1,   0 },
     {  -12, -20, -36, -55,  15,   9,   7,   5,   4,   2,   2,   1,   0,   2,  17, -13,   3, -19,  13,  -7,   4,  -2,   2,  -1, -13, -23,   6,   1, -38,  30,  -9,   9,  -2,   3,   0,   2,   1, -25, -16, -13,   2,   1,   5,   3,   1,   2,   0,   2,  -4, -33,  10, -24,  -6,  28,   3,   4, -11,   8, -12,  -3,  43,   9,  -4,  -1,   3,   1,   4,  20,   6,  -9,  -6,  -2,   0,   6,   1,   1,   7,  -4,  -2,   1,   3,   2,   2,   6,   0,   1,   0,   0,   1,   2,   0,   3,   0,   1,   0,   0 },
     {   -2,   2,   3,  19,  19, -16,   7,  -4,   2,  -2,   0,   0,  -5,   1,   6,   1,   7,   4,  -3,   2,   0,  -1,   0,   0,   2,  -3,   7,  33,  14, -22,   0,  -4,  -1,  -1,  -1,   0, -29,  -5, -26,  11, -36,  -1,  10,  -1,   5,  -1,   3,   0, -79, -27, -15,   9,  -2,  30,   5,   7,   3, -12,  30,   6,  11,   5,  -3,  -2, -13,  13,   4,   3,   0,  -3,  -4,  -3,   8,   1,   8,  -1,  -2,  -2,  -3,   0,  -2,   6,   1,  -5,   3,   0,   1,  -1,  -2,   2,   0,  -2,   1,  -1,   1,   0 },
     {    5,   5,   7,   4,  18,  14, -13,   7,  -6,   2,  -3,   1,   5,   7,  10,   5,   7,   1,   2,  -6,   1,  -3,   0,  -1,   6,  11,  13,   8,  32,  11, -17,   5,  -8,   1,  -3,   1,  10,  -1,  21,  13,  10, -23,   7, -10,   1,  -4,   1,  -2,  14, -18,  18,  -1,  12,  21, -25,   6,   1,  13,  17,   1,  32, -66,   0,  -1, -10,  13, -18,   9, -48,  -5,  18,   1,   3,  -7,   5, -11,   1,   7,   3,   0,  -7,   5,  -8,   1,   0,  -4,   1,  -3,  -2,   1,  -3,   1,  -1,  -2,   1,  -2 },
     {    2,  -5,   2,  -8, -12,  13, -11,   6,  -1,   1,   0,   2,   7,  -3,   2,  16,  19, -12,   7,  -1,   0,  -1,   0,   0,  -3, -10,  19,  15,  10, -26,   8,  -5,   3,  -2,   1,  -1,  -8, -44, -26,  15, -33,  10,   4,   5,   3,   0,   1,   1,  63,  -5, -47,  -2,   3,   8,   0,   4,  27, -41,   1,  12,   2,  -1,   2,   2,   6,   4,  -3,   3,  -6,   2,   2,   1,   1, -11,   4,   4,   4,   2,   2,  -2,  -1,   3,   3,   1,   0,  -2,   2,   2,   0,   2,   1,   0,   1,  -3,   1,   0 },
     {   -1,  -4,   4,  -3, -16, -15,  15,  -8,   4,  -2,   2,  -1,  -6, -13,  -8, -54, -37,  39,  -3,  11,  -1,   5,  -1,   1,   3,  13,  30,   1,   5,  -6,   0,  -3,  -2,   0,  -1,   0,   4, -26,  29,  -4,  -5,   9,   0,  -3,   1,   0,   0,   0,  27, -45,  10,  23, -14,   8,  -8,   3,  -5,  -2,  32, -26,  -4,  11,   1,   1,   4,   7, -15,  -3,  19,   6,  -6,   1,   1,   0,   5,  -2,   5,  -4,  -1,   2,   1,   4,  -9,  -1,   1,   0,   0,   2,   2,   2,  -3,   1,   1,  -1,   2,   2 },
     {    7,   9,   7,  35,  60, -36,   8, -14,   3,  -6,   0,  -3,  -2,  -5, -14, -16,  51, -27,   2,  -1,   3,  -1,   2,  -1,  -4,  -4,   3, -50,   4, -16,   4,   6,   2,   1,   1,   1,   2, -20,  22, -20,  -5, -10,  12,   3,   3,   1,   1,   0,  16,  -8,  -3,   0, -15,  10,   8,   1, -18,   3,  -6,   1,   3,  11,   3,  -2,   8,  -5,   1,   8,  -1,   4,  -2,  -2,  -3,   4,   1,   2,  -2,   2,   0,  -3,   2,   1,   0,   6,  -3,   2,   1,   1,   0,  -1,  -1,   2,  -3,   0,   0,   2 },
     {    1,   1,  -5,   5,   9, -26,   7,  -7,   4,  -4,   1,  -3,  -2,  -3, -11, -25, -12, -16,   9,   3,   1,   0,   0,   1,   3,  -6,  -5,  -6, -41,  -5,   6,   4,   0,   4,   0,   2,  -4,  14, -27,  -9, -34,  28,  14,   4,   3,   0,   3,  -1,  11,  41,  38, -10,  15,  14,  -9,   0,  42,   3,  14, -25,  -2, -24,   2,  -2, -14,  11, -21, -14,  -9,  12,  12,   0,  12,  -9,   0,  -2,  11,  11,   0,  -4,  -8,   3,  -4,   0,   5,  -3,   3,   3,  -3,   3,  -1,   0,   3,  -1,   2,   1 },
     {    2,   1,   1,  27,   5, -27,   7, -10,   3,  -5,   2,  -2,  -2,  -4, -11, -27, -16,  29,  -8,   7,  -1,   3,  -1,   2,   5,   7,  14,  42,   7,  10, -24,   8,  -9,   4,  -4,   1, -11, -21, -35,  -3,  23, -31,  20,  -5,   7,  -2,   2,   0,  10,  25,  -5, -17, -12,   4,   3,   1,  12,  10, -52,   5,   7,  -4,   9,  -1,   6, -24,  14,  20,  -6,   8,   0,  -2,   3,   1,   4,  -1,   1,  -3,  -1,  -1,   2,  -4,   5,   4,   2,  -1,   4,  -3,   0,  -2,   1,   1,   2,   2,   0,  -1 },
     {   -5,  -5,  -4, -15,  -6,  15,  -9,  13,  -4,   4,  -5,   5,  -6,  -6,  -4,  -2, -23, -17,   7,   0,   4,   0,   1,  -1,  -6,  -5,  -9,  -9,  -4, -19,  -5,   7,  -1,   1,  -2,   2,  -6, -10,   4, -10, -14, -31,  17,   6,   3,   1,   2,   0, -17, -20,  -9,  -1, -12,  -8,   7,   8, -26,  -5, -19,  -9, -37, -16,  34,  10, -16, -16,  -9, -13, -15,  57,  31,   0,   2,   9,   9,   4,  28,  24,  -9, -13,   1,   3,   3,   2,   2,   1,   3,   2,  -1,   2,   3,   2,   2,   2,   2,   1 },
     {   -2,   0,  -4, -16, -22,  13,  17,  -2,   3,  -1,   3,  -1,   1,   7,   4,   7,   5,   0, -17,   4,  -4,   1,  -3,   1,  -3,   3,  -4,   8,  14, -26,   8,  -8,  -1,   0,  -1,   0,   8, -13, -12, -19, -22,   5,  13,   0,   8,  -1,   2,   1,  12,  44, -15,  16, -18,  32,   5,   0, -40,  64,  23, -15,   6,  -4,  -9,  -5,   4, -13,  12,  -5, -11,  -5,  -3,  -1, -21,  13,   7, -10,   4,  -1,   2,  -3,  -6,   0,  -1,   3, -10,   2,   5,  -2,  -2,   0,   0,   0,  -7,   9,   1,  -2 },
     {   -2,  -3,  -4, -11,  -4,   4,  -3,  -3,  -3,   3,  -2,   1,   4,   5,   8,  12,  30,  27, -28,   6,  -5,  -1,  -1,   0,  -3,  -8, -17,   6,  16, -37,  10, -12,   4,  -2,   2,  -1,  -3,  -2,  -8, -34,   5,  16, -13,   5,  -2,   0,   0,   1,  -5,  -1,  22,  -3,   8, -25,  -6,   8,  -5, -10, -21, -39,  15, -11,  30,  -7,  -9,   7, -29,  49,  11,  11,  -9,  -3,  11, -10,  19,   4,  -9, -14,  -5,   1,   2,   3,  -7,   4,   0,  -3,   2,   0,   0,   1,  -4,   2,   1,  -2,   2,   0 },
     {   -1,   5,   6,  15,  14,  -1,  -1,  -8,   0,   0,   2,  -2,  -6,  -2,  -7,  -9,  -9,   5,  13,  -9,   4,  -1,   3,   0,  -5,   2,  15,   7, -13,  29,  -5,  -6,   0,   0,   0,   0,   2,  -2, -20,  20,  -4,  29, -12,   2,  -4,   4,  -1,   3,   9,  29,  -9,  -4,   2, -17, -11,  -2, -85, -33,   7,   6,   7,  -8,  12,   1, -31,  23,   2,  10,   0,   6,  -1,  -5,  -6,   2,  10,  -2,   2,  -3,  -3,   2,  -8,   9,  -2,  -2,   1,  -5,   4,  -2,  -3,   0,   1,  -2,  -2,  -1,  -1,  -1 },
     {    3,   3,   4,   2,  14,  22, -10,  -3,  -6,   1,  -2,  -1,   4,   5,   7,   8,   6,  21,  31, -13,   9,  -6,   0,  -3,   2,  -1,   0,  -6,  -2,  18,  -2,   0,  -4,   7,  -1,   1,   3,   3,   5,  -2, -15,  15,  21, -17,   9,  -6,   3,  -3,  -3,   0,   0,   4,   1,   8, -22,  -4,  14,  20,  -1,  16, -24,   3,  19, -12,  20,   6,   9,   6,  -2,  43, -51, -12,  -3, -16,  14, -13,  20, -39, -28,  22, -10,  -1,   4,   1,  -5,  -2,   0,   4,  -7,  -3,   0,   5,  -1,  -4,   2,   0 },
     {   -3,  -4,  -4,  -2, -40, -41,  40,  10,  14,  -5,   3,  -1,   2,   2,   2,   4,  16,   0,  -3,  -7,  20,  -3,  -7,  -5,  -1,  -3,   1, -13, -24,  -8,  -7,  27,  -1,  -5,  -8,   0,   3,   4,  11,  24,  -3,   4, -22,  -2,  14,  -9,  -5,  -6,  -9,  -3, -16,   9,   0, -22, -10,  27,   2,   1,  -5,  10,  15, -13,  18, -12,   7,  -8,   3,  -2, -31,   7, -21,  19,   4,   9,  14, -17,  10, -10,  22,  -3,   8,  23,  10,   0,  -5,  -1,  -9, -11, -12,   3,  -5,   7,   4,   0,   2,  -6 },
     {   -3,  -4,   1, -12, -13, -11, -19,   6,  -6,   5,   4,   2,   2,   3,   6,  -1,  21,  13, -33,   3,   1,   2,  -3,  -1,  -2,  -1,   1,  -2,   1,  -5,  -7,   7,   1,  -2,   5,  -3,  -2,   4,  -2, -26,   6, -19,  16,   1,   4,   3,   1,  -2,  18,   8,  34, -15,  -6,  13, -11,  14,  -4, -27,  34,  31,  -1,   4,  15,  -1, -33,   5,  53, -20,  10,  14,  -6,  11,  18,   8, -18, -11,   0, -12,  -6,   5,  19,   0,   4,  -4,   4,   6,  -9, -13,   5,   0,  -4,  -2,   8,   4,  -1,   2 },
   }
 },
 {  //4
   {
     {  -93,  54,  -1,   4,   0,   1,   0,   0,   0,   0,   0,   0,  60, -19, -14,   1,  -2,   0,  -1,   0,  -1,   0,   0,   0,  -5, -17,  13,   1,   1,   0,   0,   0,   0,   0,   0,   0,   2,   7,   1,  -4,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,  -3,  -2,   1,   1,   0,   0,   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -42, -15,  23,   2,   4,   0,   2,   0,   1,   0,   1,   0, -36,  78, -26,  -4,  -3,  -1,  -1,  -1,  -1,   0,   0,   0,  45, -41, -19,  18,  -4,   4,  -2,   2,  -1,   1,   0,   1,  -5, -10,  29,  -6,  -3,   0,  -1,   0,   0,   0,   0,   0,   5,   4,  -8,  -8,   4,   0,   1,   0,  -2,  -3,   2,   4,   1,  -1,   0,   0,   3,   0,  -1,  -2,  -1,   0,   0,   0,  -1,  -2,   2,   1,   1,   0,   0,   0,   1,   0,  -1,  -1,   0,  -1,   1,   0,   1,   0,   0,   0,   0,  -1,   1,   0 },
     {   53, -17,   5,  -3,  -2,  -2,  -1,  -1,   0,   0,   0,   0,  56,  -4, -42,   4,  -4,   2,  -2,   1,  -1,   1,  -1,   0,  -1, -62,  35,  16,  -2,   4,  -1,   2,   0,   1,   0,   1, -16,  25,  25, -27,   2,  -3,   1,  -1,   0,  -1,   0,  -1,  -7,   5, -19,   2,   9,  -1,   2,   0,  -5,   4,   5,   6,  -3,  -3,   0,  -1,  -2,   1,  -4,  -2,  -2,   1,   1,   0,  -2,   2,   2,   1,   0,   1,   0,   0,  -1,   1,  -2,   0,  -1,   1,   1,   0,   0,   0,  -1,   0,  -1,   0,   1,   0 },
     {   22,  13, -30,  10,  -3,   2,  -1,   1,  -1,   0,   0,   0,  43,   3,   7, -14,  -2,  -4,   0,  -2,   0,  -1,   0,  -1, -10,  -3, -58,  26,   3,   5,   1,   2,   1,   1,   0,   1,  15, -49,  49,  26, -17,   5,  -5,   2,  -2,   1,  -1,   1,  -7,  12,  19, -38,   1,   1,   0,   0,  -2,  -2,  -9,   4,  12,  -1,   1,   0,  -3,   5,   5,  -2,  -3,  -4,   0,  -1,  -1,  -2,  -3,   0,   1,   1,   2,   0,  -2,   2,   2,  -1,   0,  -1,  -1,   0,  -1,   1,   1,   0,   0,  -1,  -1,   0 },
     {   27,  32,  -7, -18,  -1,  -4,   0,  -2,   0,  -1,   0,  -1,  26,  35, -39,   8,  -3,  -1,  -2,  -1,  -1,   0,  -1,   0,  20,  -8, -19, -17,   2,  -1,   2,   0,   1,   0,   1,   0, -18, -14, -46,  40,  14,   1,   4,   1,   2,   0,   1,   0,   9, -23,  35,  33, -30,   2,  -5,   1,  -1,  -4,  14, -27,  -5,  10,  -1,   2,  -1,  -5,   4,   2,  10,   1,  -2,   0,  -1,   0,   4,  -4,  -1,  -2,   0,   0,   0,  -2,   1,   1,  -1,   0,   1,  -2,   0,  -1,   0,   1,   0,   0,   1,  -1 },
     {  -16, -69,  38,   1,   6,   1,   2,   1,   1,   0,   0,   0,  64,  24,  15, -31,   0,  -7,  -1,  -3,   0,  -1,   0,  -1,  -8,  32, -30, -11,   3,  -1,   1,  -1,   1,   0,   0,   0,  -2,  -1, -24,   0,   8,   1,   2,   0,   1,   0,   0,   0,  -8,  -1, -16,  19,   7,  -2,   2,   0,   2,  -1,   6,  13, -11,  -1,  -1,   0,  -2,  -1,   1,  -2,  -4,   4,   0,   0,   0,  -1,   2,   2,   1,   1,  -2,   0,  -1,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   1 },
     {   18,  48, -19,   5,  -8,  -2,  -3,  -1,  -2,   0,  -1,   0,  -8,   8,  -3, -26,  10,  -2,   4,  -1,   1,  -1,   1,  -1,  11,  16, -19,  18,  -7,  -1,  -2,   0,  -1,   0,  -1,   0,   7, -23,  -5, -38,  12,   4,   4,   1,   1,   1,   1,   0, -20,  10, -35,  47,  30, -12,   5,  -4,   8,  -8,  19,  26, -44,  -3,  -1,  -1,  -2,  -5,   2, -17,  -1,  17,   0,   2,   3,  -3,   6,   5,   3,  -2,  -6,   0,  -1,  -1,   0,  -5,   1,  -2,   2,   2,  -1,   0,   0,  -2,   1,  -1,   1,   2 },
     {  -19, -34,  -2,  19,  -2,   1,   2,   1,   1,   1,   0,   1, -14, -23, -14,   5,  15,   1,   4,   0,   2,   0,   1,   0, -36, -11,  -4,  41,  -8,   2,   0,   1,   0,   0,   0,   0,   1,  -4,  22,  12,   4,  -1,  -1,  -1,   0,  -1,   0,   0,   4,  22,  13,  33, -36, -12,  -2,  -3,  13, -10,  21, -40, -37,  27,  -2,   5,   1,   1,  -2, -21,  31,  12,  -7,   1,   4,   2,   4,   0,   3, -11,  -3,   0,   0,   2,  -2,  -5,   1,   0,   1,   0,   0,   1,  -1,  -2,   1,   0,   1,   0 },
     {    6,  26,  15, -13,   3,  -4,   0,  -2,   0,  -1,   0,  -1, -14,  43, -27, -16,   3,  -4,   2,  -2,   1,  -1,   1,  -1, -91,  11,  -8,  20,  11,   1,   4,   1,   1,   0,   1,   0,  15,  35,  -3,   8, -15,  -3,  -3,  -1,  -1,  -1,   0,  -1,   7, -19, -16,  -7,  -2,  10,   0,   2,   5,  -5,   4,  11,  11,  -2,  -3,   0,   4,  -5,   2,   3,  -9,  -5,   1,   1,   3,  -3,   3,  -1,  -1,   4,   1,   0,   2,  -2,   1,   0,   1,  -2,   2,   0,   1,  -1,   0,   0,   1,  -1,   1,   0 },
     {   11,  10,  15,   0, -16,   0,  -3,   0,  -2,   0,  -1,   0,  16,  16,   0,  -8,   1,  -2,   0,  -2,   0,  -1,   0,   0,  21,  32,  17,  11, -34,   2,  -7,   1,  -3,   0,  -1,   0,  41,  -3,   1, -41, -10,   8,  -3,   1,  -1,   0,  -1,   1,  -7, -17, -14, -15, -12,  12,   2,   3,   3, -23,   4, -36,  29,  31,  -6,   5,  -4,   1,  -5,   9,  48, -23,  -9,  -1,  -2,   1,  -2,   9,  -8, -19,   9,   0,  -4,   0,  -4,   3,  -1,   0,   0,   2,  -2,   0,  -2,   1,   0,   0,   0,   1 },
     {   -6, -24,   6,   1,   7,  -2,   1,  -1,   1,   0,   0,   0,  -1, -15, -53,  29,  -2,   5,   2,   2,   1,   1,   0,   1,   8,  25,  12,  29, -10,  -1,  -5,  -1,  -2,   0,  -1,   0,  48, -25, -16,  -7, -26,   9,  -3,   4,  -1,   2,  -1,   1,  -1, -37,  33,  -5,  28,   5,  -3,   1,  -1,  21,   6,   6, -11, -26,   3,  -4,  -4,  -7,   0, -10, -19,  18,   7,   1,   0,   5,   4,  -5,  11,   5,  -8,  -2,  -2,  -2,   1,  -2,  -1,   2,   1,  -2,  -1,  -1,   0,  -1,   0,   1,   1,  -1 },
     {   12,  19,  20,  -1,  -1,  -9,  -1,  -3,   0,  -2,   0,  -1,  13,  21,  20,   4, -24,   1,  -3,   0,  -1,   0,  -1,   0,  10,  16,  14,  -4,   3, -13,   1,  -5,   0,  -2,   1,  -1,  -3,  23,   7,   8, -36,  -1,   0,   0,   0,   0,   0,   0,  10,   8,   4, -28,  -9, -11,   2,  -2, -14,   0, -36,  -7, -45,  15,  18,   1,   3, -19,   2, -23,  15,  57,  -8,   2,   0,   0,   0,   2,  31,  -9, -19,   1,   0,  -5,   0,  -2,  -2,   0,  -2,   1,   0,  -3,   0,  -1,  -1,   0,  -1,   0 },
     {    7,  23,  83, -30, -10,  -6,  -5,  -2,  -2,  -1,  -1,  -1, -11, -17,   5, -24,   5,   4,   2,   1,   1,   0,   0,   0, -21, -18,   2,  -4,  -3,   5,  -1,   2,  -1,   2,   0,   1,  -9, -43,   5,  -9,  18,  12,   1,   4,   1,   2,   0,   1,  -7,  13,  31,   1,  18, -11,  -3,  -2,  -2,  24, -15, -16,  -1, -11,   5,  -2,  -2,   5, -13,   9,  10,   1,   1,  -2,   2,   7,  -3,   5,  -3,  -7,  -1,  -1,  -1,   0,  -5,   4,   1,   2,  -1,   1,  -1,   0,  -2,   2,   0,   1,   0,   1 },
     {    9,  11,  41, -13,   2,  -4,  -2,  -2,  -1,  -1,   0,   0,  -9, -52,  -3, -29,  14,  -1,   5,   2,   3,   0,   1,   1,  45, -14, -38,  28,  27,  -4,   4,  -2,   1,  -1,   1,   0,  19,  36,  -6,  26, -25, -13,  -1,  -3,   0,  -2,  -1,  -1,   9, -17, -21,   0, -16,  17,   1,   3,   2, -13,  16,   8,   5,   4,  -4,   1,  -4,   3,   3,  -3,  -5,   0,  -1,  -1,   0,   0,   3,   1,   1,   1,  -4,   0,  -1,   1,   1,  -2,  -1,   0,   1,   0,  -1,   0,   0,  -1,   0,   1,   0,   0 },
     {    5,  13,  17,  -4,  -7,   2,  -1,  -2,  -1,   0,  -1,   0,  10,   7,  -6,  16,  22, -22,   4,  -5,   2,  -2,   1,  -1,  14,  42,   6,  26, -27,  -2,  -6,  -1,  -2,  -1,  -1,  -1, -19,  10,  20,  -4,   6, -23,   0,  -3,   1,   0,   1,   0,   0,  28,  -2,   5, -34,  -2,   2,  -2,   2,  -6,  -8, -22,   3, -23,   3,   3, -14,   5, -23,  17, -41,   5,  33,  -4,   3,  -8,   3, -11,  -6,  46, -10,  -4,  -1,  -2,  -1,  -1,   1,  -4,   1,  -3,  -1,  -1,  -2,   0,   1,  -2,   0,  -2 },
     {    8,   6,   1,  33, -10,   6,  -5,   2,  -4,   0,  -1,   0,   3,  20,  50, -17, -19,  -6,  -6,  -2,  -3,  -1,  -1,  -1,  -4, -40,  16, -11,   7,  11,  -1,   4,  -1,   2,   0,   0,  59,   2,  -5,   4,  10, -16,  -3,  -3,   0,  -1,  -1,  -1,   2, -22,  11,  12, -12,  -3,  -1,   3,   5,   8,  24, -23,  -7, -21,   3,   1, -13,   4, -28, -17, -17,   4,  22,   0,  -2,   0,   0,   1,   9,  22,   1,  -4,  -2,   4,  -6,  -1,  -1,  -1,   0,   1,  -1,   2,  -2,  -1,   0,   0,   0,   0 },
     {   -1,  -7, -10, -27,   8,   2,   2,   1,   2,   2,   1,   0,  -2,  -2, -28,  -1,  15,   3,  -1,   1,   0,   1,   0,   0,   3,  -7,  -3, -21,  13,   4,   1,   2,   0,   1,   0,   1,  73,   4,  -3,   8,  37, -15,  -6,  -3,  -1,  -3,  -1,  -1,  -6,  32, -11,   1, -16, -36,   8,  -3,  -7,  -6, -45,   9,  -9,  13,  16,   0,  -4,  -9,   7,  27,  -1,   5,  -4,  -2,  -1,  -2,   3,   1, -11,   2,   0,   0,  -2,  -3,   5,   3,   0,   0,   2,  -1,  -1,  -1,   1,   1,   0,   0,   0,  -1 },
     {   -9, -13, -31,  -2,  -3,  12,  -1,   2,   1,   1,   0,   1,  -9,  -8, -14, -38,   0,   0,  15,   1,   4,   0,   2,  -1,  -7,  -9,  -7,  -7,   2,  20,   0,   1,  -1,   1,  -1,   0, -11,  -5, -11, -15,   1,  -3,  15,  -1,   4,   0,   2,   0,  -1,  -7,  -4,  -4,   1,  30,  -2,   0,  -9,  -6,  -7, -20,  24,  -9,  16,   0, -10,  13, -21,  36,  13,  49,   6, -18,  13,  -4,  22,  15,  23,  -3, -51,   0,  -3,   1,  -2,   4,   2,   0,   6,   5,  -1,   1,  -1,   0,   2,  -1,   2,   1 },
     {    3,   3,  30,   9, -11,  -4,  -1,  -1,  -1,  -1,   0,   0,  -3, -10,  10,  52, -28, -15,   2,  -3,   1,  -1,   0,  -1, -16, -23, -14, -12, -32,  18,   2,   5,   1,   3,   1,   2,  11, -30, -13,  -3,   4,  -4,  13,  -1,   2,   0,   2,   0, -18,   1, -27,   3, -23,  26,  10,  -2,   3, -21,   3,  24,  11,  23,  -3,  -1,   6,   5,  30,   5, -18,  17, -13,  -4,   8,  -5,  11, -24,  15,   6, -17,   6,   0,   4,   5,  -3,   1,   0,   1,  -7,   0,   2,   1,   0,   0,   0,  -1,  -2 },
     {   -5,  -2,  -5, -20,  10,   1,   3,   2,  -1,   1,   0,   1,  -5,  11,  -1, -21,   9,   0,   3,   1,   0,   0,   1,  -1,   4,   8,  38,   3,  27, -10,  -4,  -4,  -2,  -1,  -2,  -1,  -4, -15,   6,  13, -14, -33,   6,  -3,   1,  -1,   1,   0, -89,   0,  10,  -2, -22,  16,  19,   1,  -7,  20,  16,   6,   5,  21,  -6,  -5,  -4,   4,   6, -11,  -6,   1, -10,   0,   3,   2,  -7,  -7,   7,   1,   0,   4,  -1,   2,  -2,   4,   2,   0,  -2,  -1,  -1,   1,   1,   1,   1,   1,  -2,   0 },
     {    7,  10,  19,  99, -17, -11,  -8,  -6,  -3,  -2,  -2,  -2,  -7,  -1, -29,  -7,  -5,   2,   1,   0,  -1,   1,   1,   1,   3,   9, -11,  -6,  14,  -4,   5,  -1,  -1,  -2,  -1,   0,  -5,  16, -13,   5,   2,   2,   1,  -5,  -1,  -2,   1,  -1, -29,  14,   7,  -3,   2,   0,   0,  -4,  -7,  23, -19,  -2,   3,   2,  -1,  -1,  -1,  -8,  -1,  24,   3,  -6,  -8,   0,   1,   1,  12,  -3, -20,  -7,   1,   2,   1,  -6,  -2,  -6,   0,  -1,   5,  -1,  -1,  -2,  -1,  -3,  -1,   1,   3,  -1 },
     {    1,   0,  -4,   8,   6,  -4,  -2,   2,  -1,   1,   0,   1,   7,  11,  38,   0,  44, -12,  -3,  -6,  -2,  -2,   0,  -1,  -2, -17,  21,  41,  -7, -23,  -4,  -3,  -2,  -1,   0,   0,  -6, -14, -10,   4,  -6,   0,  -1,   1,   0,   2,   1,   1,  15, -41,   3,   7,  -3, -21,   7,   1, -34,   1,  -9,  13,  -6,   8,  22,  -6,   7,  11,  16,  42,  -6,   8, -18,  -7,   4,   4,  17, -16, -29, -11,  -6,   5,  -2,   0,  -2, -11,   1,   0,   4,  -7,  -1,   1,  -3,  -4,  -1,  -1,   1,  -1 },
     {   -6,  -3,  -4,  -9,  -2,  -7,   7,   1,   3,   1,   2,  -1, -10, -12,  -1,  20,  -9, -11,  -8,   6,  -1,   3,   0,   0,  -8,  -9, -43,  -4, -20,  -4,  13,   5,   6,   2,   3,   0, -14,  22,  -3, -12,   6,   6, -10,   9,  -3,   2,   0,   0, -44, -40,  -4, -12,   4, -32,  18,   8, -39,   6,  -2, -11,  -9,  15,  18,   5,  -5,  14, -10,  -5,  17,   2,  27,  -1,  -5,   9,  -5,  27,  -3,  31,  11, -20,   9,   1,   5,   4,   0,   5,  -1,   3,   2,   2,   2,   0,  -1,   2,  -1,   1 },
     {   -3, -10,  -4,  10, -16,  -3,   5,   1,   0,   1,   0,  -2,  -5,  -7, -11, -31,   1,   4,  -7,   8,   0,   0,   1,   1,  -6, -18,  23,  -2,  -6, -37,  11,   2,   7,   0,   2,  -1,  -8, -41, -21,   8, -29,   1,  13,   9,   5,  -1,   2,   1,  14,  -3, -16,  -4,  -7,  10,  19,  -1,  22, -24, -31,   5,  -6,  15,  -8,   6,  -4, -29,   6,   1,   2,  -8,  33,  11,  -6,  14,   5,  19,  -4,  41,  14, -19,   2,   2,  14,   3,  -2,   5,   3,   0,   1,   0,   4,  -1,  -1,   1,   1,   0 },
     {    0,   1, -13, -16,  30, -10,   7,  -2,   0,   1,   1,  -1,   5,   8,  25,  24,  16, -21,  -3,  -8,  -1,  -2,   0,  -2,   0, -28, -25,  21, -12, -16,   9,   0,   3,   0,   2,  -1,  13,  17, -39, -16, -28,  14,  15,  -1,   3,   2,   2,   0, -17,  38,  10,   6,   7,  25,   0,  -7,  21,  37, -19, -11,   1,   0, -21,  -8,  -2, -16, -27,   9,   6,  -5,  -7,   9,   6,  -6,   6,  12,  -3,  -3,   2,   1,   1,  -2,  -1,  -1,   1,  -3,   3,   2,  -1,  -2,   1,  -1,   0,  -2,   1,   0 },
     {    5,   6,  14,  11,  63, -13, -19,  -5,  -3,  -2,   0,   0,   3,   1,   1,  41,  -6,  -5,  -9,   0,  -3,   0,  -1,   0,  -6,   6,   1,  -7,  48,  -6, -13,   0,  -5,   3,   0,   1,  -5, -30,   7,  -4,  -8, -30,   3,   2,   3,   1,   2,   1,   2,  -2, -26,   1,   6,  -1,  -2,   4,   5, -24,  10, -15,  -2, -16,   8,   9, -15,  -1,  -8,  12,  21,  -1,  18,  -6,   0,  11,   7,  28, -14, -13,  -6,  -7,   0,   0,   2,  -9,   2,   2,  -1,   6,  -2,  -1,   1,  -3,   0,   1,   0,   1 },
     {    4,   3,   0,   7,  25,  -1,  -6,   1,   0,  -2,   0,  -2,  -1,  -9, -11,  -9,   2, -22,  -4,   0,   1,   3,  -1,   1,   3,  11,  -6, -15,   2,  14,   5,   8,   0,   2,  -2,  -1,  -9, -13,  18, -21,  14, -22,  -2,   0,   1,   1,  -2,   2,  37, -32,  -5, -10, -18,  -4,  43,  -1,  28,  50,   1,   5,  -5,  37,  -4, -23,  17,  -7, -27,  -1,  -8,   3, -22,   8,  -3, -16,  10,   6,   7,   9,   7,   9,   0,   5,   9,   1,  -4,  -2,  -3,   1,   0,   2,   3,   1,   0,  -1,   3,   0 },
     {   -1,  -2,  -4, -20,   3,  -3,  14,  -6,   4,  -2,   2,  -1,   0,  -1,  17,  -3, -53,  16,   6,   4,  -1,   1,   0,   0,   8,   8,   6,  19,  -1,  32,  10, -11,  -1,  -3,   0,  -3,  -6,  12,  23,  12,   2,   8, -25,  -3,  -5,  -1,   2,   0, -17, -19,  10,  21,   6,  -3,   2,  -7,  34,   3, -11,   3,  -6,   8, -10,   4,  -7, -37,   3,  32,   2,  -6,   3,  -5,   2,  14,  27,   1, -32,   8, -18,  -3,   0,  -7,   0, -32,   1,   1,   3,   1,   1,  -2,   1, -11,  -1,   1,   2,   1 },
     {    0,   1,   2,  -6, -22, -11,  19,  -6,   6,  -2,   0,  -2,   4,   8,  -1,   9, -15,  12,   8,   7,  -4,   1,  -3,   1,  -2,   8,   4,  -7,  17, -11,  16, -13,   2,  -3,  -1,  -1,  16,  -2,   3,  -8, -14,  -3,  16,   6,  -1,   1,   0,   2,  32,  32,  12,   8,   3,  10,  27, -18, -38,  17,  35,  12,   0,  19, -14, -17,  -3,  30,   1,   2,   7,  -3,   1, -21,  -4,   3, -15,  22, -33,  20, -30,  -9,   2, -11,  11, -21,  -2,   1,  -5,   2,   2,  -2,   1,  -3,   0,  -1,  -3,  -1 },
     {    4,   5,   4,   9,  57,  -4, -11,  -1,  -2,   0,  -1,  -2,  -3,  -4, -10, -27,   5, -32,   3,   5,   2,   1,   1,   0,   3,   1,  14, -35, -46, -10,  32,   6,  -1,   2,   1,   1,  13,  10,  32,  22,  -8,  30,  24, -18,  -4,  -4,  -2,  -1,  -8,  -8,   7,  11,   0,   8, -18, -19,   0,  -8,   3,   5,  -9,  -4,  -3,   3,   9,  17,   9,  -2,  -1,   2,   3, -10,   0,   3,  -8,   4, -16,   2, -10,  -1,   5,  -1,   7, -10,   2,   4,  -7,  -2,   4,   1,  -1,  -2,   1,   3,  -3,   1 },
     {    4,   4,   8,   9,  -6,  14,   4,   3,   1,  -4,   0,  -1,   4,   8,   5,  11,  60,  -7, -24,  -3,  -8,   0,  -1,   0,   2,  -5,  -4, -16,  17,  17,   3,  -7,  -1,  -3,   1,   0,  -1,  16,   0, -24,  15,  10, -23,   0, -10,   5,  -2,   2,  -7,  -5,  40, -23,  -2,  10,  11, -11,  34, -21,   7,  16,  -3,   6, -13,  -5,   6,  -9,  30, -11,   7,  10,  20, -11,   4,  19,   2,  13,   3,   6, -29, -14,   1,  13,  -2, -12,   0,  -1,  -5,   5,  -2,   3,   2,  -8,   0,   1,  -4,  -2 },
     {    0,  -3,   1,  -8, -26,   0,   4,  11,   7,   2,   2,   0,   2,   3,  -3,  14,   9,  22,  -2,  -7,  -2,  -1,  -1,   1,   0,  -3,  -8, -23,  14, -30,   4,   2,   3,   0,   4,  -2,   2,  -1,   7,  -4, -26,  17,   8,  -6,   4,  -4,  -1,  -2, -17,  -7, -10,  -4,  -3, -17, -17,  18,  60,   5,  11,  -7,  -2,   0,  44,   6,  23,  24, -18,   3,  -4,   7,  -8, -27,  -3, -22,   2,  -4, -25,  11,  -7,  21,  -7,  -3,  13, -10,   0,   5,   0, -10,   0,   3,  -3,   4,  -1,   4,   5,   1 },
   },
   {
     {  -72,  60, -11,   2,  -1,   1,   0,   0,   0,   0,   0,   0,  67, -48,   0,   4,  -1,   0,  -1,   0,   0,   0,   0,   0, -18,   2,  13,  -6,   2,  -1,   1,   0,   0,   0,   0,   0,   2,   7,  -8,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,  -3,   2,   2,  -1,   0,   0,   0,   2,   1,   0,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -55,   3,  33,  -8,   5,  -2,   2,  -1,   1,   0,   1,   0,  -6,  53, -45,   7,  -3,   1,  -1,   0,  -1,   0,   0,   0,  46, -58,  15,  10,  -5,   3,  -2,   2,  -1,   1,   0,   1, -19,  16,  13, -16,   4,  -1,   1,  -1,   0,   0,   0,   0,   7,  -1, -13,   6,   2,  -1,   0,   0,  -2,  -1,   6,   0,  -3,   1,   0,   0,   3,  -1,  -3,  -1,   1,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0,   0,   1,   0,  -1,   0,  -1,   0,   1,   0,   1,   0,  -1,   0,  -1,   0,   1,   0 },
     {  -27,  49, -55,  24,  -6,   4,  -2,   2,  -1,   1,   0,   1, -56,   5,  47, -13,   0,   0,   0,   0,   0,   0,   0,   0,  51,  -8, -24,  -4,   5,  -2,   0,   0,   0,   0,   0,   0,  -2, -19,  12,   8,  -3,   0,   0,   0,   0,   0,   0,   0,   2,   7,   1,  -8,   2,   0,   0,   0,   0,  -4,  -2,   4,   1,  -1,   0,   0,   1,   1,   2,  -1,  -1,   0,   0,   0,   1,  -2,  -1,   1,   0,   0,   0,   0,   0,   1,   1,  -1,   0,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
     {  -54, -22,  30,   7,   1,   2,   1,   1,   0,   1,   0,   1,   1,  32,  -2, -21,   6,  -4,   2,  -2,   1,  -1,   0,  -1, -26,  38, -45,  24,  -3,   1,   0,   1,   0,   0,   0,   0,  44, -48,  26,   3, -10,   5,  -3,   2,  -2,   1,  -1,   1, -13,  12,   3, -15,   8,  -2,   1,  -1,   5,  -2,  -6,   8,  -1,  -2,   1,   0,  -1,   0,   2,  -2,  -1,   1,   0,   0,   3,  -2,   0,   1,   0,   0,   0,   0,  -1,   1,   1,  -1,   1,  -1,  -1,   1,   0,   0,   0,  -1,   1,  -1,   0,   0 },
     {   23, -18, -16,  34, -15,   5,  -4,   3,  -2,   1,  -1,   1,  51,  -8,  18, -49,  15,  -5,   2,  -3,   1,  -1,   1,  -1,   5, -31, -22,  40,  -3,  -1,   1,   0,   0,   0,   0,   0, -27,  16,  39, -23,  -8,   5,  -2,   0,   0,   0,   0,   0,  -6,  17, -30,   0,  14,  -5,   2,  -1,   1,  -9,   9,  12, -11,   2,  -1,   0,  -2,   4,  -1,  -9,   4,   1,  -1,   0,  -1,  -1,   1,   4,  -1,  -1,   0,   0,  -1,   2,  -1,  -3,   0,  -1,   0,   2,   0,   1,   0,  -1,   0,  -1,   0,   1 },
     {   34,  37, -32,  12, -16,   4,  -4,   1,  -2,   0,  -1,   0,   3,  18, -42,  13,   1,   2,  -1,   1,   0,   1,   0,   0, -18, -28,  38,   1,  -6,   3,  -1,   1,  -1,   1,   0,   1,  51, -40,  18, -21,   9,  -1,   0,   0,   0,   0,   0,   0, -40,  23,  -3,   4,   7,  -7,   4,  -2,   9,   0,  -5,   3,  -6,   2,   0,   0,  -6,  -2,   3,  -1,   1,   1,   0,   0,   3,   0,   0,  -1,   0,   0,   0,   0,  -3,   0,   1,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0,   0 },
     {   23,  30,  18, -60,  25, -12,   6,  -5,   2,  -2,   1,  -2,   4, -32,  -7,  39, -12,   0,   0,   0,   0,   0,   0,   0,  24,  16, -36,  14, -17,  12,  -5,   3,  -3,   1,  -1,   1, -22, -27,  43, -16,  10,  -1,  -1,   1,   0,   1,   0,   0,   2,  22, -19,  -3,   6,  -7,   4,  -2,  -3, -10,   1,  10,  -5,   2,   0,  -1,   2,   3,   3,  -4,  -1,   1,  -1,   0,  -3,  -1,  -1,   1,   1,   0,   0,   0,   1,   1,   1,  -1,  -1,  -1,   0,   1,   0,   1,   1,   0,  -1,  -1,   0,   0 },
     {  -22,  -3,  16,   4, -13,   4,  -1,   2,  -1,   1,   0,   0, -49, -29,  13,  -4,  28,  -5,   5,  -1,   3,   0,   2,   0, -13,  27,  25,  15, -29,   4,  -3,   0,  -1,   0,  -1,   0,  11,  19,  -8, -47,  23,  -1,   0,  -1,   0,   0,   0,   0,  -5,  10, -35,  41,  -2, -10,   4,  -2,  16, -15,  21,  -9, -18,  12,  -4,   2,  -3,   5,  -8,  -5,  15,  -4,  -1,   0,   4,  -2,   1,   4,  -6,  -1,   1,   0,  -1,   1,  -2,  -2,   2,  -1,   1,   1,  -1,   1,  -1,  -1,   1,  -1,   1,   0 },
     {  -21, -50,   5,   4,  16,  -2,   4,   1,   2,   0,   1,   0,   8, -22,  26,  16, -21,   7,  -4,   2,  -2,   1,  -1,   1,  23, -14,  22, -41,  21,  -2,   0,  -1,   0,   0,   0,   0,   4,  -5,   1,  15,   0, -11,   5,  -3,   1,  -1,   1,  -1, -45,  59, -21,  -5,  -7,   5,   1,  -2,  24, -20,  -7,  11,   1,   2,  -3,   2,  -5,   3,   9,  -6,  -2,   0,   0,   1,   3,   1,  -6,   1,   3,  -1,   0,   0,  -2,   2,   1,   0,   2,   0,  -2,   0,  -1,   1,   1,   0,   1,   0,  -1,   0 },
     {   14,   7,  13,  -3, -29,  14,  -6,   2,  -3,   1,  -1,   1,  34,  12, -12, -16,  34, -20,   4,  -4,   1,  -2,   1,  -1,  50,  16, -20, -13, -24,   4,   1,  -1,   0,  -1,   0,   0, -11, -29, -35,  24,  21,   0,   0,   2,   1,   1,   1,   1, -18,  12,  19,  24, -28,   1,   0,   0,  14, -21,  11, -27,   7,  12,  -5,   3, -13,   7,   0,   8,   9, -12,   3,  -1,   2,  -2,   0,  -2,  -6,   4,   2,  -1,  -4,   1,   1,   2,   1,  -1,   0,  -1,  -2,   1,   1,   1,   1,  -2,   0,   0 },
     {    7,  26, -18, -11,  32, -17,   5,  -4,   2,  -2,   1,  -1,  11,  53,  11, -10, -40,   9,  -3,  -1,  -2,  -1,   0,  -1, -34,   6, -24,  -5,  20,   3,  -2,   2,   0,   1,   0,   1,  -2,  15, -15,   4,   5, -12,   7,  -2,   2,  -1,   1,  -1,  -4,  -4, -28,  30, -15,  11,  -2,   1,  25, -38,  35, -10,  -2,   5,  -6,   3, -10,  10,  -4,  -8,   9,  -5,   2,  -1,   2,  -4,  -1,   7,  -3,  -1,   1,   0,  -2,   0,   1,  -2,   2,  -3,   0,   1,  -1,   0,   1,  -1,   0,  -1,   0,   1 },
     {   -9, -22, -24,  31, -24,  22,  -7,   6,  -3,   3,  -1,   2,   9,   6, -21,  32,  -6,  -3,  -2,   1,  -1,   0,  -1,   0,  -9,  25,   6, -39,  15,  -6,   4,  -2,   0,   0,   0,   0, -21, -31,  41,  -2,  10,  -6,   3,  -2,   2,  -1,   0,   0,  48,   0, -26,  14, -15,   3,  -1,   0, -30,  -3,  24, -12,   4,   4,  -3,   1,   9,  13, -14,  -2,   4,  -6,   2,  -1,  -5,  -8,   4,   2,  -3,   2,   1,   0,   4,   3,  -1,  -2,  -2,  -2,   2,   0,   1,   2,  -1,  -1,  -2,  -2,   1,   0 },
     {   12,  37,  35,  -3, -33,   5,  -7,   0,  -2,   0,  -1,  -1, -16,   5, -23, -11,  21,   2,  -1,   0,   0,   1,   0,   0,  -6,  21,  10, -20,  31, -35,  13,  -7,   3,  -3,   1,  -2, -22,   6,  -9,  22, -42,  25,  -3,   0,   1,   0,   1,   0,   3,  19, -27, -15,  28,  -1,  -8,   4,  11, -28,   5,  25, -11,  -9,   7,   0,   1,   2,   7, -13,  -2,   9,  -3,   0,  -1,  -3,  -5,   5,   4,  -4,   0,   1,   1,  -1,   2,   1,  -1,  -1,  -1,   0,   0,   0,   1,   0,   0,  -1,  -1,   1 },
     {   -7,   4,  -3,  -4,  -5,  16,  -5,   4,  -1,   2,   0,   1, -29, -14, -11,  14,  24, -21,  11,  -4,   4,  -1,   2,   0, -49, -33,  -2,  24,  -8,  20,  -1,   3,   1,   1,   0,   1, -24,  12,  26,  38,  -4, -11,  -3,  -1,  -2,  -1,  -1,   0,   4,  24,  18, -26, -30,  14,  -2,   1,  24, -26,   6, -19,  27,  -3,  -5,   2, -10,  16,  -2,   4,  -7, -10,   8,  -3,   6,  -2,  -3,  -5,   0,   7,  -2,   0,  -1,   2,   1,   2,   2,  -1,  -1,  -2,  -1,   1,   0,   0,   1,  -1,   0,  -1 },
     {   10,  23,  43, -26, -10,   4,  -2,  -2,  -1,   0,  -1,   0,  -4,  -1,  39, -29,  10, -12,   4,  -2,   0,  -1,   1,  -1, -19, -37,   0, -19,  23,   2,   4,   0,   2,   0,   1,   0,  18, -17,   8,  11,   3, -11,   3,  -2,   1,  -1,   1,  -1,  -3,  14, -28,  20, -21,  12,   2,  -1, -40,  34,   3, -11,  12,   2,  -5,   0,  28, -15, -16,  12,   0,  -8,   2,   1,  -8,  -2,   9,  -1,  -4,   4,   0,   0,   5,   0,  -6,   2,  -3,   1,   1,   1,   2,   0,  -2,   1,  -2,   1,   1,   1 },
     {   -3, -14,  -1,   9,  10, -21,  13,  -3,   3,  -1,   2,  -1, -16, -46, -29,  -1,   0,  35,  -7,   6,   1,   2,   0,   2,   6, -28, -21,  23,  29, -26,   8,   4,   0,   0,   1,   0,   7, -26,  -1,  22, -21,  20,   5,  -4,   2,   0,   0,   0,  -9,  -8,  16,  33,   0,  -1, -11,   2,   4,   8,  38, -34, -16,   2,   5,   1,   2,   5, -12,  -3,  19,   1,  -3,  -3,  -1,   5,   6,   6, -10,  -6,   1,   1,   1,   3,  -1,  -5,   0,   1,   2,   2,   0,   1,   0,  -2,   0,   2,   1,   1 },
     {    8,   1,  11,  -7,  -7,  12, -11,   3,  -2,   2,  -1,   1,   8, -12,   4,   8,  -8,  -6,   7,  -2,   1,  -1,   1,   0,  36,   8,   2,   4,  12,  -8,  -6,  -1,  -1,   0,  -1,   0,  48,  22,  15, -22, -30,  -6,   4,  -3,   0,  -2,   0,  -1,  -4, -27, -17, -17,   2,  28,  -9,   4,   7,  -7,  18, -17,  37, -27,   1,   3, -30,  26, -32,  19, -14,  -6,  16,  -8,  12, -12,   4,   3,  -4,  13,  -9,   0,  -8,   2,  -1,  -3,   3,  -2,   0,   2,  -4,   2,  -2,   0,   2,  -2,   1,   1 },
     {    7,  19,  28,  30,  -6, -41,  21, -11,   3,  -4,   1,  -2,  -1,   7,   7, -20,  -8,  45, -25,   6,  -4,   3,  -2,   1,  -1,  23,  26,   9,  -7, -24,  -6,   4,  -3,   1,  -2,   0, -29,  -8,  21, -23,  12, -13,  17,  -8,   3,  -1,   1,   0,  -1,  16,  19, -17, -28,  18, -10,   1,   1,  -1, -25, -20,  22,  -1,   0,   6,  -6,   4,  -5,  13,   4,  -7,   5,  -3,   1,  -5,  -3,  -3,  -4,   2,   1,   0,  -1,  -1,  -2,   2,   0,  -2,  -2,   0,  -1,   0,  -2,   1,   0,  -1,   0,   0 },
     {    5,   2,   9,  -2,  -2, -13,  17,  -7,   3,  -2,   1,   0,  12,  11,  31,  12,  -3,  -3, -17,   3,  -3,  -1,  -1,  -1,   5,  -9,  26,  -6, -29,  13,   3,   2,  -3,   0,   0,   0,  25,   3,   6,  18, -24,  21, -20,   2,  -3,   2,  -2,   1,  19,   0,   4, -40,   6, -26,   9,   5, -12, -21,  15, -21,  -9,  35,   5,  -4,   1,  17, -20, -12,  45, -18, -13,   5, -11,  -2,  -5,  18, -12, -11,  16,  -2,   4,   3,  -2,  -5,  -4,  -2,  -2,   3,   2,   0,   1,  -1,  -2,  -1,  -1,   2 },
     {    0,  -2, -28,   1,   7, -14,   6,   1,   2,  -1,   0,   1,   4,   3, -26,   2,   9,   3,   2,  -3,   2,   0,   2,  -1,  19,  32,   1,  22, -15,  -5,  -5,  -2,   0,  -1,   0,  -1,  24,  36,   2,  19, -24,  -3,  -4,  -1,  -3,  -2,  -2,   1,  -5,  27, -27,  -7, -23,  12,  -2,  -1, -15,  -4,  11, -20,  14,   3,  -7,   5,  45, -48,  12,   3,  -6,  -3,   4,   3, -29,  19,   3, -14,   7,   3,   2,  -3,   7,  -5,  -6,   5,  -6,   1,   3,  -2,   2,  -3,  -2,   2,  -1,   0,   2,  -2 },
     {    5,   8,  19,   7, -26,  29, -20,   3,  -1,   1,  -1,   1,   0,   1,   8,  30, -19,  -3,  -3,   4,  -5,   2,  -2,   0,   2,  -2, -29,   4,  31, -24,  15,  -9,   4,  -2,   3,  -1,  13,  31,  26, -19,   1,   1,  -7,  -4,   0,  -2,  -2,   0,  -7, -12,  14,   8, -27,  -7,   5,   0,  22,   8, -27, -17,  -4,  31,  -9,   1,  -8, -23,  29, -11,  26, -14,  -8,   7,  -3,  17, -18,   1,  -2, -10,  11,  -4,  -5,  -9,   4,   4,   1,   4,  -3,  -3,  -1,  -3,   2,   0,   1,   2,  -1,  -1 },
     {    7,   4,  19,  28, -29, -14,  19, -13,   3,  -5,   1,  -3,  -9, -28, -27,  -8,  -3,  19, -11,  10,   0,   2,   0,   2,   5, -17, -34, -16,   0,  37, -12,  10,  -6,   4,  -2,   2,  30,   6, -16,   7,  21, -36,  11,  -3,   1,   1,   1,  -1,  19, -18, -17, -19,  -2,  13,  10,  -3,   2, -18,   6,  39,   8,   9, -12,   0,   3,   4,  17, -13,   2, -12,  -1,   2, -11,   8,  -4,   8,   3,  -1,   6,  -3,   6,   1,   2,   2,  -5,   2,   1,   0,   3,   1,   1,  -1,  -3,   1,   0,   0 },
     {    3,  -1,  -4, -18,  17, -14,  -4,   2,  -2,   1,   2,   1,   5,  -3,  -2, -23,  16,   6,  17,  -7,   5,  -4,   1,  -2,  18,   4,  28,  17,  13, -13,  -7,  -7,   2,  -2,   1,  -1,  31,   1,  16,  26,   0, -13,   1,   0,  -5,  -2,  -3,   1,  44,   4,   0,  15, -39,   0,   5, -11, -13,   0, -17,  12, -28,  10, -12,   5, -32,  16,  21, -39,  -4,  14,   6,   3,  19, -16, -16,  -3,  15,   3,  -8,   0,  -5,  -5,   5,  -2,   0,  -1,  -3,  -4,  -2,  -1,   1,  -2,   4,   0,  -2,  -1 },
     {   10,  15,  29,  37, -11,  -1, -17,   5,  -5,  -1,  -2,   1,   7,   8,  18,  38,  -7,  -4,  -1,  -9,  -4,   0,   0,  -2,   4,   3, -10,  15,   0,   9, -21,  10,  -4,   1,  -3,   1,   8,  16,  -5,  17,   3, -32,  20,  -9,  -1,  -1,   0,  -1,   1,  20,  18,  -2,   4, -14,  -9,  -7,  -4,  -2,  13, -14, -43, -12,  13,   0,   5,  -8, -28, -11, -12,  44,  -5,  -2,  -2, -10,  22, -15,  16,  -8, -12,   7,  -5,   5,  -9,  -5,  -1,  -3,   0,  -1,   0,  -1,  -3,  -2,  -1,  -3,   1,   0 },
     {    4,   7,  19,  24,  23,   0, -44,  14,  -4,   1,  -1,   0,  -7, -13,  -8, -17, -35, -14,  54, -12,   4,   0,   4,  -1,   5,   7,  21,   6,   5,  19, -39,   2,   3,  -2,  -1,   0, -19, -28,  -9, -10, -24,  -3,  12,  14,  -5,   4,   0,   1,  -3,  -9,   6,  -7,   5,  13,   4, -10,  -5,  -7,   2,  -6,   6,  11, -13,  11,   8,  -7,   7,  -7,  27, -10,   7,   0, -11,  16, -10,  12,  -3, -10,   1,  -3,   5,  -2,  -1,  -2,  -1,   2,   0,   2,   1,  -2,   0,  -1,  -2,   1,  -1,   1 },
     {   -2,   7, -11,  -2,   7,  -1,  10,  -8,   5,  -2,   1,  -2,  -9,  -1, -35, -23,   3,  -4,  14,   5,   1,   0,   1,   1,   5,  28, -10, -21,  28,  21,  -9,   3,  -4,   0,  -1,  -1, -10,  40,  21,   9,  -2, -14,  -4,   4,  -5,   1,  -1,   0, -32,  13,  13,  -7,   3, -14,  24, -10,  -7,  32, -10,   2, -15,  26, -13,  -5,  -3,  13, -31,  12,  17, -11,  -6,   2,  14, -32,  23,   7, -16,   3,  -1,   1,  -3,   5,  -1, -12,   2,  -4,  -2,   9,   0,   1,   3,  -3,   2,   0,   0,   2 },
     {    6,   5,   9,  18,  13, -17,  -6,  10,  -5,   0,  -6,   2,  -2,  -9, -11, -19, -12,  10,  21, -20,  11,  -3,   4,  -1,   1,  -7, -14, -26,   6,  14,   4,   9,  -5,   5,  -4,   1,  18,  10,   3,  -8,   9,  27, -32,  12,   1,   2,   0,   0,  39,  28,   0,  16,  -1, -34,  10,   4,  21,  -8, -19, -19,  20,  -2,  18, -12, -11,   5, -10,  35, -21,  18, -19,  -5, -13,   8,  -3, -11,   7,  -9,   3,   5,   5, -14,   2,  13,  -8,   3,   4,   2,   0,  -3,   7,   3,  -4,   0,   4,  -4 },
     {    7,   9,  19,  29,  12, -16,  17, -22,   1,  -5,   2,  -3,   4,   4,  13,  11,   8, -19, -13,  12,  -1,   2,  -3,   2,   6,   9,  18,  -3,   4,  21,   6, -18,  -3,  -1,   2,  -2,  -8, -10,  22,  16, -19,  -6, -26,  20,   1,   3,  -1,   1,   4, -15,   3,  27,  16,  -9,  23,  -8,  36,  20,  11,  -2,  -8, -11, -29,  -1, -11, -15,   1, -10, -29, -19,  27,   6, -10,  21,  -4, -12,  -7,  31,  -3,  -9,   3,  -4,  -9,  -1,   2,   1,   5,   0,  -1,   3,  -2,  -5,   0,   1,   2,  -1 },
     {    0,   1,   2,  -6,   5,  11, -26,  17,  -8,   3,  -3,   3,  -1,  -1,   0,   0, -11,  18,  -4,  -5,   0,   0,  -1,   1,   2,  -1,   1,  24,  13, -38,  15,  -9,   7,  -2,   1,  -3,   8,   5, -16,  -1,  39,  -4, -10,   9, -10,  -2,   0,   1,   2,  16,  21, -18,   4, -17,  25, -19, -11, -18,  10,  19,  -6,  -1, -21,   3,  19,  23, -28,   0, -18, -23,  21,   0, -32,   8,   0,   7, -21,  30,  -2,  -1,  18, -13,   8,  -6,  -6,  -3,   0,  -2,   4,  -1,   0,  -4,  -3,  -1,   2,  -1 },
     {   -5,  -7, -11, -38, -23,  18,  12,  -5,   2,  -1,   7,  -1,   2,   5,   5, -26,   5,  11,  -1,   4,   2,   0,  -4,   0,   0,   6,  12, -17,  12,   9, -27,  27, -13,   5,  -4,   4,   1,   6,  27,   6,   9, -12,  33, -15,  -4,  -4,  -1,  -2,   6,  -1,  17,  -3,  34, -17,  -6, -17,   8,  -7,  14, -28, -14,  -9,  -1,   7, -11, -11,   6,  10,   6,   9,  14, -11, -16,  38, -24,   2,  -1,   5, -12,   0,  12, -28,   7,   6,  -3,  11,  -3,  -4,   2,  -5,  -2,   5,   0,   4,   1,   2 },
     {   -3,  -4,   0, -23, -24,  26,   2,   7,  -9,   4,   5,  -1,  -5,  -8,  -1, -31, -37,  23,  -2,  24,   5,  -1,  -5,   1,   4,   4,  17,   8,  -4,  11,  14, -31,  -4,  -1,   0,  -1,  -6, -12,   8,   2,  -1,  -4,  -8,  11,  -4,  -3,   0,   0,  -4, -16,   7,  -9,  -6,  -2,  18,  -6,  13, -14,  15, -10,  -1,  11, -17, -13,   8,   1,  -6,   1, -13,  21, -16,  26,  -4,  -2,  21, -24,  37, -17,  -1,  -4,  -6,  22, -28,  17,   1, -12,  10,  -9,  -2,   4, -11,  10,   3,  -3,   3,  -1 },
     {   -1,  -3,  -3,   1, -17,  -9,  17, -17,   0,  -1,   3,   0,  -1,  -3,  -3,   4, -18, -12,  -1,   7,   0,  -3,   8,  -6,  -8, -14, -17, -10, -21, -31, -29, -17,  28,  -1,   4,   4,  -2,  -3, -10,   1,  -2,  15,   1,  22, -10,  -3,  -5,   8,   4,  -1, -19, -13,  -7, -10,  18, -16,   9,   7,  -5,  -1,  -5,  19,  -4,  33,  13,   6, -11,  17,  13,   9,  32, -18,  28,   1,  -5, -23,  17,  -3, -41,  -3, -12, -17,   2,  -4,   9,  16,   9,   1,   6,  -7,   5,   0,   2,   3,   8,  -5 },
   },
   {
     {  -97,  57,   5,   2,   0,   1,   0,   0,   0,   0,   0,   0,  49,  -9, -19,   1,  -2,   0,  -1,   0,   0,   0,   0,   0,  13, -22,   4,   6,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   8,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   38,  37, -37,  -3,  -5,   0,  -2,   0,  -1,   0,  -1,   0,  43, -76,  11,  15,   1,   3,   0,   2,   0,   1,   0,   1, -39,  11,  39, -13,  -1,  -1,   0,  -1,   0,   0,   0,   0,  -3,  20, -14, -12,   5,  -1,   1,   0,   0,   0,   0,   0,  -5,   0,  -5,   6,   3,  -1,   1,   0,   0,   3,  -1,   1,  -2,   0,   0,   0,  -2,  -1,   0,   1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },
     {  -15,  64, -39,  -6,  -1,  -1,   0,  -1,   0,   0,   0,   0, -81,  -4,  26,  16,   2,   1,   1,   1,   1,   0,   0,   0,  20,  24,  -2, -12,  -6,  -1,  -1,  -1,   0,   0,   0,   0,  26, -16, -21,   5,   3,   2,   0,   1,   0,   0,   0,   0,   5, -13,  11,   7,  -3,  -1,   0,   0,   5,  -3,   2,  -5,  -2,   1,   0,   0,   1,  -3,   2,  -1,   1,   1,   0,   0,   2,  -1,   0,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   1,  -1,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0 },
     {  -48, -38,  18,  21,   3,   4,   1,   3,   0,   1,   0,   1,  13, -18,  53, -21,  -4,  -2,  -1,  -1,   0,  -1,   0,   0, -28,  62, -18, -31,  11,  -4,   3,  -2,   1,  -1,   1,  -1,   8,   4, -36,  19,   7,  -3,   1,  -1,   0,   0,   0,   0,   2,  -4,   2,  11,  -8,  -1,   0,   0,   5,  -2,  -1,  -1,  -3,   2,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },
     {   26,  11, -50,  20,  -2,   2,  -2,   2,   0,   0,   0,   0,  43,  -2,  24, -16,  -8,  -2,  -2,  -1,  -1,  -1,  -1,   0,  32,  -3, -52,  -6,  12,   0,   3,   1,   1,   0,   1,   0, -11, -45,  18,  40,  -4,   1,  -1,   1,  -1,   0,   0,   1,  -9,   7,  32, -17, -12,   1,  -1,   0,  -6,   2,  -5,  -9,   7,   3,   1,   0,  -3,   4,   3,   0,   2,  -2,  -1,   0,  -2,   0,  -1,   1,   0,  -1,   1,   0,  -1,   1,   2,   0,  -1,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0 },
     {  -13, -34, -11,  10,   8,   5,   2,   2,   1,   1,   1,   1,  -5, -18,  27,  30, -16,   0,  -4,   1,  -1,   1,  -1,   0,  73, -23,  32, -43, -13,   6,  -1,   1,  -1,   0,  -1,   0, -41,  10,  -4, -17,  29,  -1,   2,  -1,   1,   0,   1,   0,  -3,  19,  -5,   8,   2,  -9,   1,  -1,  -3,   5,  -8,  -2,  -2,   0,   2,   0,   0,   2,  -3,   3,   0,   0,   1,  -1,  -2,   2,  -2,   0,   0,   0,   0,   0,   0,   1,  -1,   1,  -1,   0,  -1,   0,   0,   1,   0,   0,  -1,   0,  -1,   0 },
     {   18,  16,  67, -52,  -8,  -7,   0,  -3,  -1,  -1,   0,  -1,   5, -47, -16,   8,  20,   2,   4,   1,   2,   0,   1,   0,  46,  26, -33,   6,  -1,  -7,  -1,  -2,  -1,  -1,   0,  -1, -13, -12, -17,   8,   0,   1,   3,   1,   1,   0,   0,   1, -18,  -1,  13,  14,  -3,   1,   0,  -1,   0,   7,   2,  -6,  -6,   1,  -1,   0,  -3,  -2,  -1,   2,   2,   2,   0,   0,   0,   2,   1,   0,  -1,   0,  -1,   0,  -1,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   13,   8, -13,  -9,   7,  -2,   1,  -1,   1,   0,   0,  -1,  29,  51, -46,  18, -11,  -3,  -3,  -2,  -1,   0,  -1,   0,  -3,  10,   7, -31,  -3,   6,   1,   0,   0,   0,   0,   0, -13,  -1, -52,  19,  28,  -3,   2,   0,   1,   0,   1,   0,  14, -30,  15,  42, -19,  -5,  -2,   0,  -2,  -7,  22, -14, -15,   8,   0,   1,   0,  -5,   5,  -5,   5,   4,  -1,   0,  -3,   0,   4,  -3,   3,  -1,  -1,   0,   0,  -2,   1,   0,  -1,   0,   2,  -1,   0,  -1,   0,   1,   0,   0,   1,   0 },
     {  -11, -31, -11,   1,   8,   4,   4,   2,   1,   1,   1,   1,  14,  -5, -25,  49,  -9,  -6,  -2,  -1,  -1,  -1,   0,   0,  -8,  37,   4,   4, -42,  12,  -4,   2,  -3,   1,  -1,   1,  42, -57,  19, -31,   9,  17,  -6,   3,  -1,   2,  -1,   1, -22,   0,  18,  -6,  20,  -9,  -2,   0,   4,   9,  -4,  -6,   1,  -6,   3,   0,  -6,   3,  -2,   1,   2,  -1,   2,  -1,   1,   1,   0,   0,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   1,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   17,  12,   2,  21, -17,  -1,  -4,  -1,  -2,   0,  -1,   0,  22,  18,  18, -50,   6,   3,  -1,  -1,   0,   0,   0,   0,  32,  -9,   1,  -1,   3,  -6,  -2,  -1,  -1,   0,   0,   0,  43, -13, -23, -40,  11,   4,   3,   2,   1,   0,   0,   0, -32, -12, -27,  30,  37,  -9,   3,  -2,  -2,   7,  13,  25, -22, -12,   2,  -2,  -8,   0,  -3,  -7,  -7,   8,   4,   0,  -1,   0,   1,   4,   1,   2,  -3,  -1,  -2,   0,   0,  -1,   0,   0,   1,   2,  -1,   0,  -1,   0,   0,   0,   0,   1 },
     {    0, -18,  20, -12,   7,   1,  -1,   0,   0,  -1,   0,   0,  -2, -16,  16,   3, -11,   3,  -2,   1,   0,   1,   0,   1,  -3, -61,  24,  -6,  18,   6,  -1,   3,   0,   1,   0,   1,  67,  -4, -19,  24,   1, -14,  -1,  -2,   0,  -1,   0,   0, -17, -16,  34,  -2, -26,   3,   3,   1, -12,  24,   5, -27,   9,   9,  -1,  -2,  -4,   8, -10,   0,   9,  -4,  -3,   0,  -2,   5,  -4,   0,   2,  -4,   2,   1,   0,   2,  -2,   3,   0,   2,   0,  -1,   0,   1,  -1,   1,   0,   1,   0,   0 },
     {    4,  19,   6, -68,  30,  -1,   3,  -1,   2,   0,   1,  -1,  12,  29,  45,   6,  -9, -19,  -3,  -4,  -1,  -2,  -1,  -1, -31, -19, -20, -15, -10,  19,   4,   4,   1,   2,   1,   1,   3, -10,   7,  -2,  33,  -2,  -6,   0,  -1,   0,  -1,   0,  -8,  33, -22,   2,   2, -21,   4,  -2,  -2,  -6, -20,  17,  -1,   3,   6,   0,   1,   1,   3,  11,  -8,   0,  -1,  -2,   0,  -3,  -2,   1,  -3,   3,   0,   0,   1,   0,   2,   3,   0,  -2,  -2,   0,   0,   0,   1,   1,   0,  -1,  -1,   0 },
     {    7,  22,  36,  37, -46,  -3,  -7,   0,  -2,  -2,  -2,   0, -20,  -4, -23,  -7,  -8,  18,  -1,   3,   0,   2,   0,   1, -27,  -3,   6, -41,  28,  16,  -6,   3,  -1,   2,   0,   1,  -8, -37,  26,   1,  39, -20,  -2,   4,   0,   0,   0,   0,  -1,  29,   5,   5,  -5, -21,  10,  -4,   6,   4, -16,  -4,  -6,   4,   5,  -4,  -2,  -1,  -3,   9,   3,   2,  -2,  -1,   2,   2,   1,   0,  -4,  -2,  -1,  -1,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0 },
     {  -16, -19, -22,   3,  -1,   8,   2,   3,   1,   1,   1,   1, -16, -23,   3, -12,  39,  -9,   5,  -2,   3,   0,   1,   0, -20, -27,  -8,  43, -15,  -2,   6,   1,   2,   0,   1,   0, -18, -16,   3,   8,  19,   8,   1,  -1,   1,  -1,   0,   0,  -4,  30,  10,  50, -12, -24,  -6,  -4,  11,  -2,  25, -19, -45,   9,   3,   3,   2,   4,  -4, -18,  10,  13,   0,  -1,   3,   2,   3,  -1,   0,  -2,  -4,   0,   0,   2,  -2,  -4,   1,   0,   1,   0,   0,   2,   0,  -2,   1,  -1,   1,   0 },
     {    2,  11,   0,   5, -24,  11,   0,   1,  -1,   1,  -1,   0,   6,  33,   5, -14,  25, -11,  -4,  -2,   0,  -2,   0,  -1,   1,  10,  25,   6, -25, -14,   7,  -3,   0,  -1,   1,  -1,  -9,  16, -29,  -8,  -7,  15,   3,  -2,   2,   0,   1,   0, -54,  35,  24, -15, -16,   7,  -3,   0,  21,  -1, -35, -28,  17,  18,  -3,   3,   5, -17,  -9,  25,  21,  -9,  -5,   1,   7,  -3,   3,   2,  -9,  -5,   3,   1,   1,  -5,   0,   3,   2,  -1,   1,   0,  -1,  -2,   0,   1,   1,   0,   0,   0 },
     {   -7,  -8, -22,  13,  36, -14,   1,  -2,   3,  -2,   1,  -1, -13, -12, -37,  14, -20,   7,  10,   1,   2,   2,   1,   1,   3,   3, -28,   5,  29,  12,  -6,   0,  -2,   1,  -1,   0,  12,  33,  -9,  12,   5, -26,   0,  -3,   2,  -2,   0,  -1, -65,  24,  -5,   4,  -6,   3,  10,  -3,  27,   1,  -6,  27, -10,  -2,  -5,  -2,   7, -16,   9,   3, -19,   7,  -1,   3,   4,  -3,   8,  -3,   1,   4,  -2,   2,   1,  -2,   7,  -4,   0,  -2,   2,  -2,   0,   0,   2,  -1,   1,  -1,   0,   0 },
     {   -5,  -5,  -7, -16,  51, -26,   2,  -1,   2,   0,   1,   0,  -7,  -2, -21, -46,  17,   7,   9,   1,   2,   1,   1,   0,   2,  11,  55,  -9,   8, -14,  -1,  -5,   0,  -2,   0,  -1, -13, -42,  -8,  27, -10,   5,   4,   1,   2,   0,   1,   0,  10,  14,  11,  -5,  13,   1,  -4,  -2,  -6,  27, -24,  15,  -6, -19,   3,  -1,  -3,  -4,  -7,  10, -15,   8,   6,   1,   2,   2,   1,   3,  -3,   5,  -3,  -2,   2,  -4,   1,   1,   0,   2,   1,   0,   1,  -1,   0,   0,  -1,   1,   0,   1 },
     {    5,  -4,  -7,   4,  12,  -2,  -6,   0,  -2,   0,  -1,   0,  11,  -6, -15, -12,  30, -16,   7,  -2,   1,  -1,   1,  -1,  30,  24,  17,  -2, -11,   2,  -5,  -3,  -3,   0,  -2,  -1,  42,  -3,  19,   8,   1, -32,  -4,  -1,   1,  -2,   0,  -1,  22,  25, -49,   3, -46,   2,  20,  -1, -14, -31,  -3,  -5,   4,  37,  -5,  -2, -11,   2,  11,   3,  12,  -6,  -9,   1,  -2,  -2,   0,  -3,  -2,  -2,   2,   2,  -1,   1,  -1,   2,  -2,  -2,   1,   0,  -2,   1,   0,   0,   0,  -1,   1,   0 },
     {  -16, -26, -33, -39, -46,  43,  14,   5,   4,   3,   2,   2,  -1,   2, -18,  18,  38,   8, -21,  -1,  -4,   0,  -2,   0,  -2,  -6,  -5, -17,  40, -37,  -2,   6,   1,   0,   1,   0,  11,  -8,  -8,   5,  -1,  -6,  21,  -1,   1,  -2,   1,   0,   0,   6, -10,   1,   5,   3,  -1,  -7, -12,   3,  -7,  22,   8, -10,   0,  -2,   0,   6,   9,  15, -19,  -5,   0,   0,   0,   0,   2,  -6,  -9,   8,   1,   0,   1,   3,   3,   1,   0,   0,   0,  -1,   0,   1,   2,   1,   1,   0,   1,  -1 },
     {    4,   3,   3, -18,  -1,  -1,   6,   1,   0,  -1,   1,   0,  15,  19,  22,  16, -33,  28, -12,  -4,  -2,   0,  -2,   1,   9,  18,  17,  16,  25, -36, -19,   6,  -3,  -1,  -2,   0,  10,   4,  19, -14, -21, -19,  23,   1,  -2,   1,   0,   0,  17,  17,  16,   4,  -9,  -3,   2,  -8,  25,  -6, -21, -16, -44,   5,  10,   1,  -8, -20,  -4, -21,  11,  36,  -1,   2,  -5,   1,   2,   0,  20,  -6,  -9,   1,  -2,   0,  -2,  -4,  -2,   1,  -1,   1,   0,  -2,  -2,  -2,   1,   0,  -2,   1 },
     {   -6,   1, -16, -17,  11,   9,   6,  -2,   0,   1,   0,   0,  -3,   9,   0, -17,  20, -18,   2,  -1,   0,   0,   1,   0,   0,  25,  -2, -14,  17,   3,  -4,  -2,   0,  -2,   0,  -1, -16,  23,  53, -24,   8, -23, -10,  -1,  -2,   0,   0,  -1, -24, -48,   8,  14, -12, -10,  18,   3,  -8,  49,  -1, -17,  -5,  11,   4,  -6,   3,   5, -34,   4,  11,  -1,  -2,  -1,  -4,  -2,  -2,  17,   0,  -2,   1,   1,   2,   1,  -2,   4,   0,  -1,  -1,   2,   1,   0,  -1,   1,   0,   1,   0,   1 },
     {    6,   4,   5,  18,  -2,  -1,  -3,  -3,   0,  -3,   0,  -2,   4,  -1, -12,  10,  27, -25,   0,   2,   0,   0,  -1,   0,  11,   0, -15, -19,  -5,  -1,  24,  -4,   3,   0,   1,  -1,  33,  46,   5,  15,  -2,  25, -22,  -9,  -3,  -2,  -3,  -1,  26,   8,  24, -11,  12, -41, -13,   6, -21,   8, -34,   5, -27, -11,  30,   0,   2, -12,   1,   4, -10,  29,  -1,  -6,  -6,  -1,   5,  -3,   2,   3,  -8,   1,   0,  -3,   2,  -2,  -2,  -2,   1,   0,   0,  -1,   1,  -1,  -1,  -1,   1,  -2 },
     {   -4, -15, -21, -18,  11,   0,   0,   1,   1,   3,  -1,   1,  -3, -21, -18, -39, -16,  39,  -3,   1,   0,   2,   1,   1,   7,  -5, -25,  17,   2,   7, -10,   5,  -2,   2,  -1,   1,  12,  13, -11, -31,  32,   3,  -3,   3,  -1,   1,  -2,   0,  37, -10,  16,   5,   8, -18,  -3,   4,  17, -16, -28, -15,  19,   3,   7,   1,   0, -15,   4,  39,  22, -15,  -3,  -1,  -7,   7,  15,   5, -19, -14,   6,  -1,  -3,   3,   0,  -1,  -1,   2,   2,   1,  -1,  -1,  -1,   0,   0,   1,   1,   1 },
     {    3,  -3,   7,   1,   8,  -2,   6,   0,  -2,   2,  -1,   0,   0,  -7,   3,   1,  18, -21,  -3,   1,   0,   2,   1,   0,  -2, -21,   1, -35,   3,  -1,  20,   0,   3,   0,   1,   1,   2, -12,  26,   4, -23,  17,   1,  -9,   1,  -1,   0,   0,  -1, -35,  -8,  23,  -4,  11, -14,   6,  81, -17, -18,   4,  -2,   1,  -1,   6, -30, -10,  25,   5,   1,  -6,   0,   0,  -8,  15,   7,  -6,  -3,  -1,   3,  -1,  -4,   5,  -1,  -5,  -2,   4,  -1,  -1,  -2,   2,  -2,   1,   1,   1,  -1,   1 },
     {    3,   0,   4,  -8, -19,  48,  -8,  -3,  -1,   0,  -1,   1,   6,   5,   9,  -9,  -5, -10,   6,   1,   1,   0,   2,  -1,  15,  13,  30,  39,  -6,  25, -17,  -7,  -3,  -3,  -2,  -1,  -3,   5,  -4,  27,  25, -22, -33,  13,  -4,   0,  -2,   1,   9, -14,  14, -20,  -3, -21,  18,   8,  39,   4,   2,  14,  -2, -15,   0,  -9, -12,  -7,   8,  -5, -34,  -7,  13,   2,  -3,   4,  -1, -15,   5,  19,   4,  -3,  -1,   1,  -3,   0,  -1,  -3,   0,  -1,   0,  -1,   0,   1,   1,  -2,   0,  -2 },
     {   -8, -11,  -6, -16, -14, -16,  18,   1,   4,   3,   2,  -1,  -4,  -4,  -3, -12,   8,  32, -11,  -2,   1,  -1,  -1,   1,  -2,  -4,  12, -10, -15,   5,   2,   2,   1,  -2,   2,   0, -11,  -5,  -2,  15,  -5,   4, -17,   8,   1,   3,  -1,   0, -30, -31,   2, -32,  11, -25,  21,  13,  -2, -28,  -1,  27, -16,  37,  22, -10,  23,  13,  28, -20,  21,  14, -30,  -5,  -7,   9, -15,   0,  17, -19,   4,   4,   2,   2,   1,   5,  -2,   3,  -2,   3,   0,  -1,   2,   0,   0,   3,  -2,   0 },
     {    0,   0,  -2,  -8,  -1, -38,  15,   2,   2,   3,   0,  -1,   8,   9,   5,  -3,  36,  12,   9, -15,  -1,  -1,   0,  -2,   5,   7,   7, -18,  10,   8, -24,   1,   0,  -2,   0,   1,  13,  19,  17, -15,  14,   0,  -2,   5,  -4,  -3,   1,   1,  -2,  14,  41,  -6, -11,   3, -17,   0,  15, -46,  23,   2,   3, -33,  -7,  12,  14,  28, -14, -14, -12, -18,  21,   0, -11, -11, -26,   1,   1,  10,  11,  -2,  -8,   8,   4,   7,  -2,   1,  -3,   0,  -3,   2,  -2,   1,   0,   0,  -2,  -2 },
     {    4,   4,  13,  13,  25,  28,  -7, -13,  -7,  -2,   1,  -1,  -3,  -7,   1,  -2,  -7, -44,  10,  13,   7,   0,   2,   0,  -3,  -4,   6,  10,  26, -26,  17,  -4,  -3,  -1,   1,  -1,  -8, -20,  -5, -43,  14,   1,   5,   0,   1,   3,   2,   0,   2, -11,  41,  -4, -21,  10,  11,  -2, -14, -38,  -6,  34,  -2,  16,  -4,  -6,  35,   2,   5,   4,  -8,   6,  -8,   1, -13,  12,   4,  -5,  -1,  -1,  -4,   1,   1,   3,  -2,   1,  -2,   3,  -1,  -2,   2,  -1,  -1,   2,   1,   1,   0,   0 },
     {   -1,   1,  -4, -24,   7,  50,  -8,  -3,  -6,   1,  -1,   1,  -7,  -6, -21, -25, -33,  -6,   2,  22,   2,   4,   0,   2,   7,  14,  11, -23,  25,   4,  15,  -9,  -3,  -2,  -1,   1,  10,   6,  13,  10, -10,  31, -21, -12,   1,  -2,   0,  -1, -13,  31,  -9,  -4,  13,  -7, -21,  13,   3, -21,  35, -15,   2,  -7,   5,   8,   1,   0, -13, -21,  27,  -7,   7,   0,   0,   8,  -8,  21,  12, -13,   8,  -3,   4,  -1,  -4,   3,   0,  -3,  -2,   3,   2,   0,   0,   0,   0,  -2,  -2,   1 },
     {    6,  13,  16,  20,  36,   9, -24,  -7,  -5,   0,   0,  -1,   1,   8,   8,  14,   6,  37, -40,  15,  -3,   0,  -3,   1,   0,  10,   9,  19,  -7, -49,  24,  14,  -3,   1,  -4,   1,   0,   8,  19,  21,  28,  -3,   7, -20,  -2,  -1,   1,  -2, -20, -12,  -6,   6,  -1, -22,  -9,   6,  -3,  -8,   8,  -6,   8,  -8,  14,   0, -10,  26,  10,  11,  -9, -19,   0,  -6,  -4,  -4,  -9,  -5, -20,   2,  13,   0,  -2,   2,   0,   1,   0,  -5,  -1,  -3,  -1,   1,  -2,  -2,   1,  -1,  -4,  -1 },
     {    6,   1,   5,   4,  -5,  -9,  13,  -2,  -6,   3,   3,  -3,  10,   4,  12,  15,   4,   2,  -8,  -4,   5,  -1,  -3,   1,   9,   4,  11,   5,   7,  -1,  -1,   9,   0,  -4,   0,   2,  13,  -6,  13,  23,  -3,  12,  -7,  -5,  -4,   4,  -3,  -3,  13,   6, -12,  15,  10,  -6,   4,   0,  18,  18,  20,  11,  -1,  -1, -14,  -7,  76, -47, -18,   9,   5, -26, -19,   6, -28,  -3,   9,  -5, -21, -16,  16,  11,   1,  -3,   4,  -8,   0,  -3,   1,   1,  -2,  -1,   1,  -3,  -3,  -1,  -2,   0 },
     {   -2,  -1,  -2,   6,  -1, -26,   2,  -2,   3,   1,  -2,  -1,   6,  14,  12,  35,   6,  -7,  24,  -1,  -7,   1,  -3,   0,   3,   4,   7,  24,  55,  -9,   2, -36,   9,  -3,   2,  -3,  -6,  -7, -18,  -9,  -8,   7, -40,  13,  16,   2,   1,   1,   5,  10,  -4,   4,  16, -21,   3,  19,   1,   6,   0,  -2,  -2,  17,  -1,  -7, -30,  20,   6,  12,  11, -22,  -5,   2,  -1,   9, -11,  17, -22, -20,  12,   1,  -5,  -4,   8,  -8,  -5,   3,   1,  -1,   0,   2,   3,  -6,   3,   0,  -4,   1 },
   }
 },
 {  //5
   {
     {  111, -48,  -5,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0,   0, -38,   3,   6,   1,   2,   1,   1,   0,   0,   0,   0,   0,  -5,  10,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -5,  -1,  -2,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -3,   2,   1,   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   46,  21, -22,  -2,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0, 101, -28, -27,  -1,  -4,   0,  -2,   0,  -1,   0,   0,   0, -10, -29,   2,   6,   2,   2,   1,   1,   1,   0,   0,   0, -11,  -4,  10,   3,   0,   1,   0,   0,   0,   0,   0,   0, -10,   0,   6,  -1,   0,   0,   0,   0,  -7,   2,   3,   0,   1,   0,   0,   0,  -6,   2,   2,   0,   0,   0,   0,   0,  -4,   2,   1,   0,   0,   0,   0,   0,  -3,   1,   1,   0,  -2,   1,   1,   0,  -2,   1,   0,   0,  -2,   1,   0,   0 },
     {  -13, -27,   6,  10,   1,   3,   1,   2,   0,   1,   0,   1,  -2, -80,  28,  11,   4,   3,   2,   2,   1,   1,   1,   1, -55,   7,  47,  -1,   1,  -1,   0,  -1,   0,   0,   0,   0,   5,  39,   8, -18,  -4,  -3,  -1,  -2,  -1,  -1,   0,  -1,   1,  13, -12, -11,   2,   0,   1,   0,   4,   7,  -8,  -1,   3,   0,   0,   0,   2,   3,  -5,   1,   1,  -1,   0,   0,   3,   3,  -3,   0,   0,  -1,   0,   0,   1,   1,  -2,   0,   2,   1,  -2,   0,   1,   1,  -2,   0,   1,   1,  -1,   0 },
     {  -22, -79,  20,  14,   7,   5,   4,   2,   2,   1,   1,   1,   9,  -3, -10,  14,  -3,   2,  -1,   1,  -1,   0,   0,   0,  49, -53,   2,  15,  -5,   3,  -2,   2,  -1,   1,   0,   1,  -9, -11,  40,   3,  -5,   0,  -2,   0,  -1,   0,   0,   0,   5,  15,  14, -19,  -5,  -1,  -1,  -1,  -6,   7,  -3, -12,   5,   1,   1,   0,  -1,   8,  -3,  -2,   4,  -2,   0,   0,  -3,   5,  -2,  -1,   1,  -1,   1,   0,   0,   4,  -2,   0,  -1,   2,  -1,  -1,   0,   2,  -1,   0,  -1,   1,  -1,   0 },
     {   -5, -45,  14,   3,   2,   2,   1,   1,   1,   1,   1,   1,  47, -15,  21,  -7,  -3,  -3,  -2,  -2,  -1,  -1,   0,  -1,  51,  55,  -6, -26,  -7,  -6,  -3,  -3,  -1,  -1,  -1,  -1,  29,   1, -46, -15,   5,   1,   2,   1,   1,   0,   0,   0,  -5, -21, -19,  15,  11,   2,   2,   1,  -4,  -6,   4,  13,   0,  -2,  -1,   0,  -6,  -3,   4,   4,  -3,   0,   0,   0,  -5,  -1,   2,   1,  -1,   1,   0,   0,  -4,  -1,   1,   1,  -2,   0,   1,   1,  -2,   0,   1,   0,  -2,   0,   1,   0 },
     {  -10,  -7, -45,  20,   3,   3,   2,   2,   1,   1,   0,   1, -19, -56, -14,  27,  10,   3,   4,   2,   2,   1,   1,   1,   3,  15, -28,  19,   1,  -2,   0,  -1,   0,   0,   0,   0,  15, -37, -16,  37,  -3,   2,   0,   1,   0,   1,   0,   1,   2,  -9,  39,  27, -15,  -3,  -3,  -1,   9,   7,  24, -12, -22,   0,  -2,   0,  -1,   1,   8, -18,  -5,   7,   0,   1,   2,   4,   3,  -9,   3,   3,  -2,   0,   0,   3,   3,  -5,   1,   3,   1,  -3,  -1,   1,   1,  -2,   1,   2,   0,  -2 },
     {  -14, -32,  41,  -8,   4,   0,   1,   0,   1,   0,   0,   0,  23,  18,  29, -23,   1,  -4,   0,  -2,   0,  -1,   0,   0, -56,   7, -20, -16,  10,   1,   3,   1,   2,   1,   1,   0, -52, -31, -12,  22,  18,   4,   4,   2,   1,   1,   1,   1, -18,   9,  26,  25,   0,  -5,  -2,  -2,   4,  16,  10,  -3, -14,  -5,  -1,  -1,   4,   6,  -3,  -8,  -5,   2,   1,   0,   4,   2,  -4,  -1,   0,   2,   0,   0,   4,   1,  -2,   0,   3,   0,  -2,   0,   2,   0,  -2,   0,   2,   0,  -1,   0 },
     {   15,  37,  17,   5, -16,  -3,  -5,  -2,  -3,  -1,  -2,   0,   4, -18,  69,  -5, -14,  -2,  -3,  -2,  -1,  -1,  -1,   0,  24,  16,   1, -13, -13,   1,  -1,   0,   0,   0,   0,   0,  12, -40,  22,   4, -11,   6,  -2,   2,  -1,   1,   0,   1,   1,  10,  34, -33, -11,   3,  -2,   0,  -1,   1, -19, -36,  11,   8,   3,   2,  -7,   0, -13,  -3,  22,   1,   0,   0,  -2,   2,  -9,   5,   8,  -6,   0,  -1,  -3,   1,  -5,   4,  -1,   0,  -4,   3,  -1,   0,  -2,   2,   0,   0,  -3,   1 },
     {   -9, -30, -23,   5,   7,   2,   3,   2,   2,   1,   1,   0,  25,  33, -21,  -3,   3,  -4,   1,  -2,   0,  -1,   0,  -1, -38,  52, -31,  12,  -1,  -3,   0,  -2,   0,   0,   0,   0,  27,  21,  18,  20, -15,  -3,  -4,  -1,  -2,  -1,  -1,   0,  23,  25,   9, -30, -23,   0,  -2,   0,   2, -14, -26, -24,   8,  13,   3,   3,  -3, -11,  -5,   7,  18,   3,  -2,   0,   0,  -6,   2,   7,   3,  -4,  -2,   0,  -2,  -4,   3,   2,  -1,  -4,   1,   0,  -1,  -2,   2,   1,   0,  -2,   1,   0 },
     {  -13,  -5, -61,   8,   7,   8,   4,   4,   1,   2,   1,   1,  -3,  13,  -4,   9,   1,   0,  -1,   0,  -1,   0,   0,   0,  18,  44,  28,  -8, -14,  -7,  -5,  -3,  -2,  -1,  -1,  -1, -62,  -3,  18, -21,  -3,   3,   3,   2,   1,   1,   1,   0, -51,   7,  10, -13,  13,   6,   3,   0, -15,  16,   3,  -1,   9,  -3,  -1,  -2,  -1,   6,  -5,   0,   2,  -5,  -1,  -1,   3,  -1,  -6,   2,   0,  -2,   2,   0,   4,  -1,  -2,   3,   3,  -1,  -1,   2,   2,  -1,   0,   0,   1,  -1,   0,   0 },
     {   -9,  -9, -28, -21,   9,  11,   3,   3,   1,   1,   1,   1, -13, -30,   1, -39,  16,  10,   3,   2,   2,   1,   1,   1, -10, -13, -42, -17,  18,  11,   2,   3,   1,   2,   1,   1,  10, -38,   9,  -1,  14,   7,  -1,   2,   0,   1,   0,   1,  -7,  -2, -12, -24,  25,   2,   1,   0,  -5,  -8, -19,  27,  40,  -5,  -1,  -1,   3,  12,  13,  35,   4, -24,  -5,  -4,   3,   6,   3,  11, -18, -12,   3,   0,   1,   4,   5,   4,   1,   2,   1,   3,   2,   3,   3,   2,   1,   1,   1,   1 },
     {   -1,   5, -13, -55,  12,   6,   4,   2,   2,   1,   1,   0,  -5, -20, -13, -47,  21,   9,   4,   4,   3,   1,   1,   1,  36,   1,  11, -29,  22,  -2,  -1,  -2,  -1,  -1,  -1,   0,  -9,  31,   2,   8,  19, -12,  -2,  -3,  -1,  -2,   0,  -1,  22,  33,  21,  18, -12, -17,  -4,  -3,   8,  16,  -8, -21, -22,  -5,   3,   0,  -4,  -8, -20,  -7,   5,  10,   6,   2,  -3,  -3,  -4,  10,   8,   1,  -1,  -2,  -2,  -4,  -2,   8,  -2,  -2,   0,   4,  -1,  -2,   0,   4,  -1,  -1,   1,   2 },
     {    8,  15, -10,  46,   6,  -9,  -5,  -2,  -3,  -1,  -1,  -1,   5,  -5,  29,  23,   7, -16,  -3,  -3,  -2,  -2,   0,  -1,  18,  -2, -45,   0,   6,  -8,   3,   1,   2,   0,   1,   0, -21,  15,  -1,  -1,  15,  -6,   2,  -2,   1,  -1,   1,  -1,  19,  64, -24,   3,  13, -15,  -1,  -6,  13,  12, -24,  24, -11, -17,  -1,  -3,   1,  -2,  -8,   2, -26,  -3,   6,   1,   0,  -8,  -3,  -1, -10,  11,   4,   1,   0,  -3,   4,  -2,  -1,  -5,   2,  -3,  -1,  -2,   2,  -3,  -1,  -3,   1,  -2 },
     {    1,   0,   0, -14,   1,   6,  -1,   2,   0,   1,   0,   0,   7,  10,  24, -35,  14,  -2,   0,  -1,   0,  -1,   0,  -1,   1,  20,  10,  39,   8,  -7,  -7,  -2,  -2,  -1,  -1,  -1,  40,  -8,  60,   4, -13, -17,  -9,  -1,  -3,  -2,  -2,  -1, -24,   5,  -3,   3,   6, -14,   4,  -1,   1,   4,  24,  34, -21, -15,   0,  -2,  -1,   9,   9, -21, -36,  -3,   5,   1,  -2,  -6,  -9, -16,  -4,  15,   6,   1,  -2,   0,  -4,  -6,   0,  -2,  -5,  -1,  -1,   0,  -3,  -2,   0,  -2,  -3,   0 },
     {    6,  19,  54,  -2,  15, -18,  -5,  -6,  -1,  -2,  -1,  -1,  -8, -27, -21,  16,  26,  -6,   1,   0,   1,   0,   1,   0,  19,  24, -26,  11,  23,  -7,   2,  -2,  -1,  -1,   0,  -1, -19,  35,  11,  27,  -5, -14,  -1,  -3,  -1,  -1,   0,  -1, -44, -11,  11, -10, -16,  -3,   5,   2, -39,  -8,  -6,   1,  11,   7,   8,   1,  -6,   7,   9,  18,   0,  -3,  -1,  -3,   2,   7,   2,  -1, -12,  -3,   0,  -1,   1,  -1,  -2,  -2,   2,   0,   0,  -3,   2,  -1,   1,  -1,   1,   0,   1,  -2 },
     {   -6,  -8, -23,   5, -13,   4,   2,   1,   1,   0,   1,   0,  -3,  -9,  34, -28, -11,  -1,   6,   0,   2,  -1,   1,   0,   7, -13,  -8,  16,   2,   0,   4,   1,   1,  -1,   0,   0, -32,   6,  11,  13,  -5,  -4,   3,   0,   1,  -1,   0,   0,  -5,  -7, -34,  16, -16,  -3,   6,   1, -23, -59,  -9,   9, -27,  26,   7,   6, -11, -24,  18,  -3,  21,  34,  -2,   1,   4,   6,  20,   8,  17,  -2, -15,  -3,   4,   7,   8,  -8,   1,   3,   0,  -2,   2,   5,   0,  -3,   1,   2,   0,   0 },
     {    1,  -2, -20, -14,  27,  -4,   2,   0,   0,   0,   0,   0,  13,  16,  35,  26,  23, -11,  -8,  -3,  -4,  -1,  -2,  -1,   8,  -1,  47,  11,  26, -18,  -6,  -5,  -2,  -1,  -1,  -1, -19, -10,  -9,  43,   6, -15,  -2,  -1,   0,   0,  -1,   0,  39, -31,   3,  -3, -17, -10,  -2,   3,  10, -12, -18,   7,  15,  -5,   5,   0,  14,  -1,   9,  27, -13, -15,   1,  -2,  -2,  -1,   2,  -8, -23,  -1,   7,   1,   0,  -2,   1,  -5,  -2,   0,  -1,  -6,   0,   0,  -1,  -2,  -1,   0,  -2,  -3 },
     {    3,   3,  28, -28,  -8,  -4,   4,   1,   2,   0,   1,   0,   2, -22, -26,   3,  11,  -1,   5,   1,   1,   0,   1,   0,  10,  27,  -1,  49,  -8,  -7,  -6,  -4,  -2,  -2,  -1,  -1, -31, -39,   5, -11, -14,   1,   2,   2,   2,   2,   1,   0,   5,  -3, -38,  -3,   4,   6,   6,   1,  55,   5, -22,   2,  -5,   6,  -3,   0,  24, -14, -21,  -1,  12,   7,  -2,   2,   8, -12,  -1,  11,   6,  -2,  -3,   1,   0,  -4,   5,   4,   1,   0,   5,   2,  -1,   0,   2,  -1,   0,   0,   1,   0 },
     {   -3,  -2,   1, -20, -54,   6,  16,   2,   3,   2,   1,   0,  -3, -13,  -3,  -4, -68,  11,  15,   1,   2,   2,   1,   0,  11,  15,  -6,   1, -20,  10,   6,  -3,   0,   0,   0,  -1, -19,  29,  12,  45,  -2,   7,  -6,  -6,  -1,  -2,  -1,   0,  16,  -3,  12,   7,   3,   1,  -9,  -1,   6,   4,  -2,  22,  20,  -8,  -7,  -1,  17,   3,   9,  10, -16, -16,  -4,   0,   1,  -6,  -6, -15, -11,   3,   6,   4,   2,  -5,  -1,  -4,  -1,  -3,  -1,  -5,   0,  -2,   0,  -1,   0,  -1,  -1,  -2 },
     {   -7,  -5,  -2, -30, -24,   2,  14,   3,   2,   2,   2,   0,  -4,   9,  -4,  26, -18,   0,   8,  -1,  -1,  -1,   0,  -1,  -8,  -7,  23,  29,   5,  -6,  -3,  -2,  -2,  -2,   0,  -1,  29, -26, -29,  16,  25,  -9,   3,   1,   1,  -2,   1,   0, -35,  30,  -9,   7,   7, -15,   3,  -3, -37,  15, -43,  -1,  -4,  -8,  10,  -2, -33,  -9, -15,  16,  -5,  12,  11,  -2,  -6,   7,   1,  10,   0,  13,  -2,  -3,   0,   1,   4,   5,   2,   2,   3,   0,   2,   1,   3,  -1,   2,   1,   1,  -3 },
     {    4,   7,  18,  26,  -5,  19,  -4,  -4,  -2,  -2,  -2,   0,   2,   5,  -9,  10,  13,  26,   0,  -5,  -2,  -1,   0,  -1,   4,  23,  25,  13,   7,  27,  -9,  -6,  -3,  -2,  -2,  -1,   4,  -8,  28,   2,  43,  16, -13,  -6,  -3,  -1,  -2,  -1,   9,  21,   6,  39,  23, -10, -21,  -8,   1, -27,  14,  -4,   1,  -7, -18,   4, -15, -13,   5,  14,  31, -19, -10,   3,  -4,   0,  11,  12, -15, -32,  -3,   2,  -7,  -2,  -1,  -9,  -1,  -1,   0,  -2,  -2,   0,  -2,  -5,   1,   1,  -1,  -2 },
     {    2,   1,  16,   6,  23,   1,  -1,  -3,  -1,  -2,  -1,  -1,  -4, -10, -28,  12, -10,   7,   1,   2,   1,   1,  -1,   1,  -8,  12,  10, -43, -14,   6,   4,   3,   3,   2,   1,   0,  -5, -24,  29,  -8,  -5,   6,  -2,   4,  -1,   1,   0,   1,  15, -17,  26, -10,  13, -13,  -3,   1,  -9, -28, -29,  24, -16, -25,  10,   3,  -8, -47,  -3,   4, -34,  19,  18,   6,   0,  -9,  11,  -1,  12,  25,   2,  -2,   4,   2,   9,   3,   0,   3,  -1,  -3,   2,   2,   0,   1,   1,   2,   0,  -1 },
     {   -3,   3,   6, -58,  18,  -7,   3,  -3,   2,  -1,   2,  -1,   2,  -6,   9,  24,   7, -10,   0,  -3,   0,  -1,   1,  -1,  -3,   6,  -9,  14, -33,  -9,   0,   0,   0,  -1,   0,  -1, -13,  -7,  -6, -29, -25,  15,  12,   4,   4,   0,   2,   1,  33,  41,  14,   6,  -1,  18,  -1,  -4, -25, -13,  28,  14,   5,   8, -14,  -3, -29,  -5,  30,   7,  -4,  -7,  -9,   0, -11,   9,  10, -14,  -9,  -2,   1,   2,  -1,   5,  -2,  -5,  -3,  -1,  -4,  -1,   0,   0,  -4,   1,   1,  -1,  -3,   0 },
     {    4,   3,   0, -18,  28,  20,   1,  -8,   0,  -4,  -1,  -2,   7,   9,  11,  32,  -2,  33,   2,  -6,  -3,  -3,  -1,  -2,   6,  -8,   5, -23,   3,  21,  12,  -2,   1,  -2,   1,  -2,  21,   3,   0,  15, -18,  20,   7,  -5,  -2,  -1,   0,   0, -33,  23,   3,   2, -10,   3,   3,  -6,  15,  10, -17,  28, -22,  19,  -3,  -5,  46,   2,  23,   6,   4,  18, -28,  -3,  17,  -3,  -2,   1,  10, -19, -26,   3,   6, -13,   3,  -3,   0,  -4,  -3,  -3,  -1,  -7,   2,   0,  -3,  -2,   0,   1 },
     {    0,  -2,  -1,   8,  -5, -36,   7,   5,   0,   0,   0,   1,  -3,  -2, -19,  -7,  -5, -47,  13,   7,   4,   2,   1,   2,  -2,  -3,  17, -36, -31, -35,  17,   5,   1,   4,   3,   1,  19, -17,  26,   0,   4, -19,  16,  -2,  -1,  -1,   1,   0, -13,  18,   0,  41, -15,  -9,  10,  -4,  16, -17,  -3,   1,   0,  19,   5,  -2,  16,  16,   1,  27,  13,  -4,  -4,  -9,   8,   7,  -1,   2, -25,  -4,  -3,  -3,   0,  -3,  -7,  -5,   2,  -2,  -2,  -2,   0,  -3,  -3,   0,   1,  -2,   1,  -1 },
     {    0,   4,  10,  14, -14, -14,  -2,  -1,   4,   0,  -1,   0,  -4,  -2, -25, -10, -19,  -4,   1,   6,   2,   4,   0,   2,   2,  22,  21, -15,  37, -12,  -4,  -1,   0,  -1,   1,  -1,  -9, -26,   0,   1,  13,  -4,  -5,   2,   0,   0,   0,   0,  28,  14, -16, -40,   5,  -7,   3,  -1,  -8,  29,   6,   5, -21,  13,  11,  -3,  -2,  23,  55, -11,   1,  22,  -8,  -6, -11,  19,  24, -12,   1, -14, -13,  -1, -12,  -5,   1,  -7,  -6,   5,   0,   2,  -1,   0,  -1,   0,  -2,   0,  -3,   0 },
     {   -1,  -2,  -3,  17,   9, -35,  -5,   6,   2,  -1,   0,   1,  -2,   3,   5, -22,   0, -25,   5,   8,   0,   0,   1,   0,   2,  -4,  11,  17,  15, -13,  12,  -7,   1,  -3,   0,  -1,   6,  12,   2,  17, -24,  26,   7,  -7,  -4,  -2,  -1,  -2,  10, -13,  29,  11,  33,  25,   5, -16,  -7,  39, -10,  38,   0,   6, -14, -12, -15, -30, -15,   7,  16,   5, -17,   2, -14,   1,   1,  28,  17, -20, -10,   0,  -1,   4,   6,   3,  -3,   2,   2,  -4,  -2,  -3,   3,  -5,  -2,   2,   1,  -2 },
     {   -3,  -7, -12, -34, -10, -40,  10,  11,   6,   4,   2,   3,   5,   0,  14,  34,  -4, -36,   1,   7,   1,   4,  -2,   2,  -1,   1, -20,   9,  22,   9,   0,   2,   0,   2,  -3,  -1,   8,  13,  17, -25,  43,  20, -12,  -1,  -4,   2,  -2,   0,   1, -20,  14,  -9,  35, -10, -14,  -4,  -9, -17,   6, -15, -12,  -6,   4,   6,  39,   0,   2, -13,   6,  17,  -1,  -1,  16,  -7,  -5,   0,   7,  -4,  -6,  -3,   3, -10,  -4,   2,   2,  -5,   1,   6,   4,  -2,   2,   2,   1,  -1,  -2,   1 },
     {   -3,  -7,  -1,  -3, -23, -11,  -1,   7,   4,   1,   0,   1,  -2,   0,   3,   5,  -2, -14,  -1,   5,   4,   1,  -1,  -1,   4, -10,   7, -11,  37,   8,   3,   1,  -3,   1,  -2,  -1,   4,   0, -12,  -3,   5,  -5,   3,  -6,  -2,  -1,   1,   0, -37,  24,  12, -26, -20,  15,   5,   2,  44, -14,  35,  -2,  13,  10,  -6,   0, -13, -52,  16,   8, -11, -12,   2,   5, -17, -38,  22,   4,  -6,  10,   9,   4,  -2,  -4,  17,  -1,   0,  -2,   1,  -6,   0,   3,  -3,  -1,   2,   1,  -4,  -1 },
     {    2,  -1,   2,  15, -26,  12,   9,  -4,  -2,  -2,   2,  -2,   3,   3,   5, -23,  12,   3,   7,  -7,   1,  -2,   0,   1,   3,  -3,   0,  27,  -4,   3,   2,   0,  -1,  -1,   2,   0,  -2,   5, -21, -40, -23,  -2,   0,   1,   1,   2,  -1,   2,   7,  -2,  40,  -6, -10, -29,  -7,   5, -11,  18,   3,   1, -16, -16,  12,   3,  27, -10,   7,  42,   4,   7,  10,  -2,  32,   4,  35,  15, -31,   7,  -8,  -3,  14, -20,   2, -16,   1,  -4,   4,  -3,   2,  -2,  -1,  -5,  -1,  -1,  -1,  -2 },
     {   -1,  -3,   0, -10,   4,   4, -40,  -6,  11,   3,   3,   3,  -4,  -4,  -4,   0, -25,  -2, -68,   0,  13,   8,   4,   3,  -2,  -8,  -5,  -1,  -7, -35, -45,  10,  15,   4,   4,   1,   4,   6,  -6,   9,  -9, -18, -18,  23,   2,   3,   1,   0, -10,  10,   8,   7,  21, -13,  -1,  14,   5,   1,  -3,   8,   1,  -1,   8,  13,   8, -18,  10,  -7,  20,  -7,   2,  12,  10,   1,   3,   6,   0, -21,  11,   0,  12,  -8,  -1, -11,   3,  -5,  -4,  -1,   4,  -4,   2,  -1,   2,  -2,  -2,   0 },
     {   -2,   2,  -2, -12,  13,  11,  -2,   5,  -7,  -3,   1,  -3,  -2,   0,  17,  17, -10,  11,  -4,   6,   1,  -3,  -3,   1,  -7,  -1, -20,  -1, -21,  -5,   0,  10,   2,   1,   2,   0,  -2,  12,  28,  -6,   8,  -9,  -5,  10,   0,   0,   1,   2,  -1, -30,  -5,  17,   3, -12,   3,   4,  10,  33,   2, -12,  -8,   8,  15,  -3, -15,  18,   9,  24,   6,  14,   1, -14, -54, -35,  32,  24,  -9,   5,  -8,   4, -24, -16,  29,  11,  -1,  -1,  16,   0,   2,  -1,  -3,  -3,   2,  -1,  -1,   1 },
   },
   {
     {  -80,  55,  -4,   3,  -1,   1,   0,   0,   0,   0,   0,   0,  68, -36,  -9,   2,  -2,   0,  -1,   0,   0,   0,   0,   0, -17,  -7,  18,  -4,   2,  -1,   1,   0,   0,   0,   0,   0,   3,  12,  -8,  -3,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,  -6,  -1,   4,   0,   0,   0,   0,   2,   2,   1,  -1,  -1,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -50,  -6,  22,   2,   3,   0,   1,   0,   1,   0,   0,   0, -14,  65, -34,  -1,  -4,   0,  -1,   0,  -1,   0,   0,   0,  47, -59,   1,  16,  -4,   4,  -2,   2,  -1,   1,   0,   1, -13,   6,  28, -16,   2,  -2,   0,  -1,   0,   0,   0,   0,   4,   5, -18,   0,   5,  -1,   1,   0,  -2,  -3,   6,   6,  -3,   0,   0,   0,   3,   0,  -3,  -3,   0,   1,   0,   0,  -1,  -2,   2,   1,   1,   0,   0,   0,   2,   0,  -1,  -1,  -1,  -1,   1,   0,   1,   0,  -1,   0,   0,  -1,   1,   0 },
     {   60, -22,   8,  -6,  -1,  -2,  -1,  -1,   0,  -1,   0,   0,  45, -13, -33,   7,  -2,   3,  -1,   1,  -1,   1,   0,   0, -15, -43,  47,   1,   1,   0,   0,   0,   0,   0,   0,   0, -24,  47,  -8, -22,   6,  -4,   2,  -2,   1,  -1,   1,  -1,  -2,  -6, -19,  20,   1,   1,   0,   1,  -3,   3,  12,  -2,  -8,   1,  -1,   0,  -1,  -1,  -4,  -3,   3,   2,   0,   1,  -3,   3,   3,   1,   1,  -2,   0,   0,  -1,   0,  -2,   0,  -1,   1,   1,   0,   0,   0,  -1,   0,  -1,   1,   1,   0 },
     {  -28,  -1,  21, -10,   3,  -1,   2,  -1,   1,   0,   1,   0, -53,   5,  -6,  18,   3,   3,   1,   1,   0,   1,   0,   0,  13,  20,  35, -28,  -1,  -5,   0,  -2,   0,  -1,   0,  -1, -10,  34, -64,   7,  12,  -2,   4,  -1,   2,  -1,   1,   0,  18, -28,  14,  32, -18,   4,  -4,   2,   3,   3,   9, -25,   0,   3,  -1,   1,   4,  -5,  -5,   8,   7,  -2,   0,   0,   1,   2,   2,  -3,  -4,  -2,   1,   0,   2,  -3,  -2,   2,   1,   1,   2,  -2,   1,  -2,  -1,   1,   1,   0,   1,  -1 },
     {  -31, -44,  26,  13,   3,   2,   1,   1,   1,   1,   0,   1,  -6,   0,  27, -29,   7,  -4,   3,  -2,   1,  -1,   1,  -1, -34,  36,  -5,  13,   0,  -2,   0,  -1,   0,   0,   0,   0,  15,  24,   0, -34,   0,  -2,   1,  -1,   0,   0,   0,   0, -10,   9, -51,  26,  17,  -5,   4,  -2,  13, -11,  17,  23, -28,   3,  -3,   1,   2,  -2,   0, -19,   4,   8,  -2,   2,   3,  -2,   1,   7,   4,  -4,  -2,   0,   1,   0,  -1,  -3,   2,  -2,   1,   3,   1,   0,  -1,  -2,   1,  -1,   0,   2 },
     {  -14, -61,  39,  -4,   7,   0,   3,   0,   1,   0,   1,   0,  53,  35, -17,  -8,  -3,  -5,  -1,  -2,  -1,  -1,   0,  -1, -28,  14, -28,   1,   4,   1,   2,   0,   1,   0,   0,   0,   5,  -2, -12,  34,  -9,   2,  -2,   1,  -1,   0,  -1,   1,  12, -24,  30, -14, -15,   7,  -4,   3,  -6,   7,  -1, -16,  20,  -2,   1,   0,  -1,   1,  -4,  11,  -2,  -8,   2,  -1,  -3,   1,   1,  -3,  -3,   3,   2,   0,   0,   0,   0,   1,  -1,   1,   0,  -1,   0,   0,   0,   1,  -1,   1,   0,  -1 },
     {   21,  32, -10,   5, -10,  -2,  -3,  -1,  -2,   0,  -1,   0,  23,  29, -23, -19,   7,  -2,   1,  -2,   1,  -1,   0,  -1,   9,  13, -33,  19, -10,   0,  -2,   1,  -1,   0,   0,   0,   2, -35, -20,  -2,  12,   6,   2,   2,   1,   1,   1,   1,  -6,  -6,   5,  53, -13,  -4,  -3,  -1,  10, -19,  35, -30, -29,  18,  -5,   4,  -6,  -3,  -1, -15,  35,   1,  -3,   1,   0,   1,   5,   9,  -5, -13,   1,  -1,  -4,  -1,  -1,  -4,   0,   0,   3,   3,  -2,   0,   0,  -2,   0,   0,   2,   1 },
     {   -5, -27, -15,  10,   4,   3,   2,   2,   0,   1,   0,   1,  36, -14,  48, -15,  -5,  -3,  -2,  -2,  -1,   0,  -1,   0,  66,   1, -23, -35,  10,  -4,   2,  -2,   1,  -1,   1,  -1, -56,   8, -12,  17,  13,  -3,   4,  -2,   2,  -1,   1,   0,   6,  16, -13,  12, -10,  -1,  -1,  -1,  -4, -10,   2,   7,  -2,   4,  -1,   1,   1,   4,   4,  -2,  -2,  -1,  -1,   0,  -5,  -2,  -2,   0,   0,   1,   1,   0,   0,   3,   1,   0,  -2,  -1,  -1,   0,   0,   1,   1,   0,  -1,  -1,   0,   0 },
     {  -18, -32,   5,  15,  -7,   3,   1,   2,   1,   1,   0,   1, -14, -41,  12,   5,  12,   3,   1,   1,   1,   0,   1,   0,  -6, -16,  15,  20, -13,   3,  -2,   2,  -2,   0,   0,   0,  -6,  -4,  40,  -8,   9,   1,  -1,  -1,  -1,   0,   0,   0,   7,  30,  14,  -3, -30,  -5,  -3,  -1,   7, -11,   4, -55,  13,  23,  -3,   5,  -5,  14, -14,   9,  42, -22,  -1,  -3,   1,   5,  -5,   9, -19,  -9,   7,  -2,  -1,   3,  -4,  -1,   1,   1,  -1,   1,   0,   2,  -3,  -1,   1,   0,  -1,   1 },
     {  -10, -30,  24,  -6,  11,  -4,   3,  -1,   2,   0,   1,   0,   2, -33, -21,  25, -11,  10,  -1,   4,   0,   2,   0,   1,  22,   2,  12,   1,   3,  -4,  -2,  -2,  -1,  -1,   0,  -1,  -2, -32,   0,   8, -11,  10,   0,   3,   0,   2,   0,   1, -10,  13,  35,   2,  18, -13,  -1,  -4,  -2,  22, -14,   4, -48,   2,   5,  -1,   5, -19,  15, -33,  15,  37, -11,   6,  -1,  11,  -2,   1,  28, -25,  -6,   0,  -2,  -1,  -1,  -1,   0,   5,  -3,   0,   0,   0,   0,   0,   0,   2,  -1,   0 },
     {   -7,   7, -38,  22,   6,   1,   1,   0,   0,   0,   0,   0, -22,   8,  16,  -3, -10,   2,  -1,   1,   0,   0,   0,   0, -45, -30, -17,  16,  26,  -1,   5,  -1,   3,   0,   1,   0, -32,  34,  11,  42, -25,  -1,  -2,   0,  -1,   0,   0,   0,  32, -11,  13,  -1, -13,   5,  -5,   1,   5,  -2,  19,  -4, -23,   1,  -1,   0,   8,  -5,   4, -33,   4,  17,  -3,   3,   0,   1,   0,   3,  20,  -8,  -3,   0,   4,   1,   0,  -3,   0,   0,  -1,   0,   2,   0,   0,  -2,   0,   0,   0,   0 },
     {   -9, -33, -61,  41,   3,   9,   2,   3,   1,   1,   0,   1,  10,  26,  24,   0, -20,   0,  -5,   0,  -2,   0,  -1,   0,   4, -24,  30, -25,   5,   3,  -1,   1,   0,   1,  -1,   0,  56, -11, -13,  -9,  10,  -6,   1,  -1,   0,   0,   0,   0, -27, -16,  14,   9,   2,  -3,   2,  -1,  -1,  20,   4,  -9,  -3,  -1,   0,  -2,  -4,  -4, -11,  -1,   5,   2,   1,   1,  -2,   2,   4,   4,   1,  -1,  -1,  -1,  -2,   1,  -2,  -3,  -1,   1,   0,   1,  -1,   1,  -2,  -1,  -1,   0,   0,   1 },
     {  -12, -21,  -8,   3,   2,   6,   3,   2,   0,   1,   0,   1, -13, -28, -19,  12,  22,  -8,   6,  -2,   3,  -1,   1,   0, -12, -19, -11,  16,  -1,   9,   0,   2,   0,   1,   0,   1,  -9, -27,   1,  19,  28, -13,   6,  -2,   2,  -1,   0,   0, -14,  13,  12,  45, -11,  -2,   0,  -2,  13,  12,  27,  13,  -8, -25,  -3,  -3,  -5,  15,  -4,   8, -57,  -1,  13,  -3,   7,  -8,  10, -23,  -6,  34,  -8,   3,   1,   5,   1,  -4,   2,  -2,   2,  -7,   0,   3,  -1,  -2,   1,  -1,   1,  -3 },
     {   -1,   9,   9, -20,   3,  -2,   1,  -1,   1,  -1,   0,   0,   1,  36,   5,  14, -10,  -5,  -3,  -1,  -2,  -1,  -1,  -1, -45, -12,   9, -40,   8,   6,   2,   3,   1,   1,   1,   1,  -6, -10,  -4,  21,  31, -15,   4,  -4,   2,  -2,   1,  -1, -26,  59, -10,   6, -31,  -8,   5,  -4,  13, -15, -31,   6,  -6,  24,  -4,   3,   6, -12,  22,   1,   5,   0, -10,   4,   3,   0,  -6,  -8,   4,  -5,   4,   2,   1,  -2,   1,   5,   1,   0,  -2,  -2,   1,  -2,   1,   2,   1,  -1,  -1,  -1 },
     {   14,  27,  51,  -1, -15,  -4,  -9,  -1,  -3,   0,  -2,  -1,   2,   3,  44, -40,  -1, -11,   6,  -3,   2,  -2,   0,  -1,   8, -21,   4,   3,   7,   0,  -1,   0,  -1,   1,  -1,   0,  15,   9,   3,   0,  -6, -19,   7,  -4,   2,  -2,   1,  -1, -16,   8,  16,   1, -20,   9,  -4,   0,   1,  23,  -9, -32, -12, -13,   9,  -1, -11,  -3, -36,   0, -13,  35,   9,  -1,   7,  -7,  10,  -3,  25,  11, -17,   3,  -3,   2, -10,   3,   1,  -3,   1,   1,  -2,   0,  -4,   2,   1,  -1,   1,   1 },
     {   -1,  -6, -27,   9,  -7,   7,  -1,   1,  -2,   0,   0,   0,  18,  22, -12,  40,  -8, -17,   2,  -4,   3,  -1,   0,   0,  11,  50,  20,   7, -40,   9,  -9,  -1,  -4,   1,  -2,   0,  -3,  11,  16, -11, -11,  -7,   1,  -2,   1,   0,   1,  -1,  22,  18, -11, -12, -28,  10,  -5,   2,   3, -23,  -2, -22,  -5,  -6,   6,   3,  -8,   0,  -9, -12, -20,  30,   8,   1,   6, -15,   5, -16,  33,  15, -15,   4,  -3,   3,  -2,   5,   0,  -5,   1,  -5,  -3,   2,   0,   1,  -1,  -3,   0,  -2 },
     {   -5,   4, -10,  15,  -7,   2,   0,   1,  -1,   0,   0,   0,   1,  27,  10, -17,   6,  -1,  -1,   0,  -1,   0,   0,   0, -34,  13,  16,  -9, -18,  -3,   0,   0,   1,   0,   0,   0, -60, -24,   6, -30,  23,  20,   0,   5,   0,   3,   0,   2,  25,   3,  22,   7,  34, -26,  -1,  -4, -19,  31, -22,  -6, -11, -14,  13,  -4,   6,   1, -21,  23,   1,   9,   0,  -3,   5,  -2,   6,   2, -11,  -2,  -4,   1,   3,  -2,  -5,  -1,   1,  -2,   2,   3,   1,  -1,  -2,   0,   1,   0,   1,   1 },
     {    2,   5,  11, -55,  33,   3,   4,  -1,   0,   0,   0,  -1,   5,  -1,  40,  42, -31, -11,  -6,  -2,   0,   1,  -2,  -1, -12, -16, -18, -18, -15,  21,   0,   5,  -2,   0,   0,   2,  11, -14,  20, -17,  23,   0,  -6,   1,  -1,   1,   0,   0,  27, -26,  -8,  10,   6, -11,   2,   1, -10,   5,  29,  -7,  -4,  -5,   3,  -4,  -5,  15, -17, -19,   3,  -3,   6,  -1,  -6,  -2, -11,  14,   5,   3,   3,  -2,   2,   2,   1,   0,  -3,   0,  -3,   1,   0,   4,  -1,   1,  -1,   0,  -1,   2 },
     {   -8, -12,  -8,  -6,  -5,  18,  -4,   5,  -5,   2,  -1,   1, -12, -19, -13, -13,  -1,  -3,  10,   0,   5,   1,   2,  -1, -17, -20, -14, -22,   4,  27,  -9,   8,  -3,   2,  -1,   2, -16, -30, -14, -17,   2,  -2,  15,   0,   6,   1,   3,   0,  -4, -18, -20,  -5,  10,  41,  -8,   6,  -1, -23,   3,  -5,  34,  -5,   6,   0,  -7,   8,   9,  20,  15,  30, -13,  -3,  13,   0,  22,  -2,  37, -35, -21,   3,  -5,  11,  -8,  13,   3,   1,   6,   1,  -3,   6,  -2,   3,   2,  -1,   3,   0 },
     {   -5, -11, -36, -40,  27,   5,   6,  -2,   2,   1,   0,   0,   0, -11, -10, -10,  18, -10,   6,  -1,   5,  -1,   2,  -1,   8,  -7, -32,  26,   1,  -1,   6,   0,   0,  -2,   1,   0,  26,  31,  -6,  -2,  14,  -3,   0,  -3,  -1,  -1,   0,  -1,  15,   4,  -8,  10, -19, -22,   3,  -2, -25,  13, -39,  -2, -15,  18,  18,  -4,  12, -18,  -8,  35,   6,  24, -14,  -2,   4, -11,  23,   3,  -6,  -7, -11,   3,   1,  -1,   2,  -1,   0,  -5,   7,   2,   0,   0,   2,   0,   0,  -1,   4,   0 },
     {   -7,  -1, -21, -30,  39,  -3,   2,  -1,   2,  -1,   1,   0,  -2,  10,  -2,  10, -23,   0,   1,   0,   1,   1,   0,   0,   0,  19, -10,  20,   6,   0,  -3,  -1,   0,   0,  -1,   0, -36,  22,  -3, -11, -17,  -4,   1,   2,   2,   1,   0,   0, -65,  12,  16,  -4,  17,  20,  -2,   2,  42,  15,   1, -15,  10, -15,  -8,   2, -13,  -7, -21,   8,  10,  -6,  12,  -2,   9,   2,   6,  17,  -8,  -2,  -6,  -3,   1,  -5,  -1,  -8,   3,   0,   1,   6,   0,  -3,  -2,  -2,   2,  -1,   1,   2 },
     {    2,   3,  -1,   8,  16, -12,  -1,  -2,   1,   1,   0,   0,  -3,  -4, -40,  -7, -18,  21,   3,   4,  -1,   1,   0,   1,   3,  32,   6, -32,  60,  -7, -11,   0,  -3,   0,  -2,   0,   7, -12,  31,  11,   1, -36,  14,  -3,   0,  -2,   0,  -1,  25,  16, -23,   9,  -3,   7,   4,  -5, -19,  14,   9,  -6,  -4, -15,  -1,   2,  -5,   8, -33,   6,   6,   5,  10,  -5,   3,  -6,   3,  25, -12,  -3,  -6,   0,   1,  -5,   3,  -7,  -1,  -2,   2,   3,   0,  -2,   1,  -1,   0,  -3,   1,   3 },
     {   -1,  -9, -12, -26,  13,  -3,   1,   0,   1,   1,  -1,   1,   5,   5,   1, -23,  15,   5,  -4,   1,  -1,   1,   0,  -1,   1, -18,  17,  -3,   9, -15,   5,  -1,   0,   0,   0,   0,  22,  -7, -22,   2,  -7,  21,   6,   1,   0,   0,   1,   0,  41,  30,  11,  -2,  22,  -8, -25,   2,  40, -44, -10,  -4,  -6, -34,  15,   6, -33,  10,   3,   1,  -3,   9,  21, -10,   0,   8, -18,  17,  -5,   9,  -7,  -3,  -3,  -8,   7,   9,  -3,   1,   1,   1,   0,  -2,   3,   2,   1,   0,   1,   1 },
     {    2,   5,  -1, -19,  36,  -2,  -5,  -6,  -5,   0,  -2,   4,  10,  17,  30,  -9,   9, -13,   1,  -3,   0,  -6,   1,  -2,   7,  10,  48,  36,  10,  -2, -18,  -2, -11,  -1,  -3,  -2,   3, -15,   6,  18, -12, -12,  15,  -2,  -3,  -3,   1,  -1,  10,  12,   2,  10,   7,  17, -20,   0, -20,   1,  23,   3,   4, -11,  -6,  -2,  19,   5,  20,  10,  -5, -11, -25,  -4,   4,  -4,   4, -33, -16, -39,   3,   9,  -9,   7, -18,  -3,   4,  -2,  -2,  -4,  -5,   2,  -5,  -3,  -1,  -1,  -1,  -1 },
     {   10,  10,  23,  45,   1, -11, -11,   0,   2,  -1,   0,   1,   6,  -2,  11,  43, -27,  18,   2,  -6,  -6,  -6,   0,   0,   2,  -6, -36,   1,   1,  12, -13,   7,  -1,   4,  -2,   1,  18,  27,   2,  -8,  16,   5,   9,  -3,  -6,  -2,  -1,   0,  22,  13,  12,  12,  20,  -1, -17,  -6,  28, -11,  -9,   4,   0, -17,   0,   5,  -1, -16,   7,  28,   0,   1,  -6,  -4,   1,   4,  29, -18, -20, -22,  -9,   7, -10,   0, -11, -20,   2,   0,   1,  -1,  -3,  -5,  -4,  -5,   0,  -1,   0,  -1 },
     {   -2,  -2,   8,  15,  26, -39,   8,  -5,   6,  -3,   2,   0,  -1,  -2,  19,  20, -22,  19,  -2,  -1,  -7,   2,   0,  -1, -12, -17,   6,  21,  -2, -30,   1,   1,   1,  -1,   1,   0, -12, -36, -25,  -8, -33,   6,  18,   2,   3,   2,   2,   1,  -8,  -9, -25,   2, -14,   8,  13,   2, -30, -28,  -9,  -6,  -7,  18,  10,   6,  -3,  -2,   0,  22,   0,  15,  13,  -6,   4,   5,  12,  -2, -14,  35, -10,  -2,  10,   2,  11, -13,  -3,   7,   0,  -1,   5,   2,   1,  -5,  -2,   3,   0,  -2 },
     {    0,   2,   2,   3,   7, -12,   0,   5,   1,   0,   2,   1,  -3,   3,  -4, -11, -12,  18,   1,  -3,  -4,   1,  -3,   0,   3,  12,  10,  -1,   8,   0,  -9,   3,  -1,   1,   0,   0,  -7,  13,  18,  11,  15, -10, -20,   3,  -3,  -2,  -3,  -1, -28, -34,  10, -10,  -7, -47,  11,  12,  10, -31,   6,   4,  -8,   9,  44, -15, -28,  50,  10,   8,  -5,  19, -25, -14,  15,  -2,  -6, -16,  -6, -11,  -8,  13,  -6,   1,  -8,  -5,   4,   3,  -1,   0,   0,   0,  -1,  -3,   4,  -3,  -3,   5 },
     {   -2,   1,   4,  -4, -31,  20,  -6,   4,   2,  -2,  -2,   0,   1,   6,   2,  39,  23, -10,  -9,  -6,   2,   2,  -2,   0,  -4,  -6,  -9, -17,  -2, -24,  13,   5,   3,   0,   0,   1,   8,   9,  -8,  27, -23,  31,   3, -11,   2,  -2,   2,  -1, -31,  23, -17,  -6,  17,  -8, -28,   5, -26,   2,  29,  -5,  -6, -12,  21,   0,  15,  28, -25,  10,   8,  11, -12,  -9,   3, -17,   6,  18, -11, -15,  -6,   9,  11,  -1,   9,  -9,   0,  -7,   0,  -1,   1,   1,   2,  -5,   0,  -4,   0,   1 },
     {   -1,  -5,  -1, -20, -30,  21, -11,  -2,  -2,   5,  -9,  13,   5,   7,  13,  20,  20, -20,  -9,   5,   0,  -8,  -2,  -3,   1,   2,   5, -16,  25, -26,   2,  -2,  -1,   0,   2,  -7,  -8, -13,  38, -14, -28,   0,  21,   4,  -9,  -2,   4,  -2,   6, -40,  -6,  22, -20,   9,   0,   7,  38,  14, -20,   4,  -9,   5,  -1,  -6, -24, -16,  20,   6,   3,   8, -12,  -8,   1,   5,  10,  -5, -26,   4,  -6,   3,   4,  -6,   9,  -9,  -1,   2,   1,  -2,   1,  -7,  -4,  -1,  -1,   0,  -5,  -4 },
     {   -5,  -5, -10, -16, -23,  -8,  14,   1,   6,   2,   1,  -1,   0,   2,   3,   2, -18,  45,  -6,   0,   1,   4,  -3,   0,  -4,   6, -17,  11,  17, -40,  -3,  14,  -2,  -3,   1,  -1,   2,  11,  14, -32,  15,  12,  26, -13,  -2,  -1,   0,  -2, -16,  -5,  16,  21, -18,   4, -30,  -9, -21,  -6,  -7,  -5,  16, -22,   5,   9,  13,   1,  11, -12,  -8,   3,  -2,   1,  -5, -10, -23, -16,  14,  -2,  25,   0,   1,   3, -11,  36,  -2,  -6,   8,   2,   4,   5,   0,   7,   7,  -2,   1,   1 },
     {    2,   0,  -2,  10,   2, -15,  29, -13,  -5,   3,   5,   5,   3,   1,  -1,  17,  10,   2,   0,   3,  -2,  -3,   1,   4,   0,  -9, -14,   2,  11,   0,  -3,   4, -15,  -8,   3,   2,   1,  -7, -23,  -4, -26,   6,  -4,  -5,  -1,  -1,   3,   1,  17,  15, -18,  11,  18,   2,  17,   9,  23,  29, -12,  -9,  12,  24,   4, -26, -24,  11, -27,  -7,  -4,  -5, -33,  10,  21, -27, -29, -17,   6,   3,  24,  16, -23,  10, -19,  23,   5,  -4,  -2,  -5,  -5,   2,   0,  -8,  -1,  -5,   3,  -2 },
     {    3,   7,   7,  22,  30,  19, -12, -20,  -5,   7,  17,   8,   0,   5,   2,  21,  38, -20,   6,  15, -17, -17,  10,  10,  -1,   2,  -8,   1,  21,  -7,   3, -13, -19,  -1,   1,   2,  -3,   7,   9,   7,  15,  -5,  -4,   0, -18, -16,  -5,  -2, -18,  -9,  -6,  11,  22, -12,   0,  -4, -20, -11,  -7, -11,  21,  -5,  22,  15,  -6, -18,   1,  -9,  24,   9,  26, -10,  -9,  23, -21, -15,  12,  16,  -8,  -5,  -6,   7, -16,  22,   4,  11,   8, -11,   5,   0,   5,  -2,  -2,   6,  -4,  -7 },
   },
   {
     {  -98,  62,  -2,   4,   0,   1,   0,   0,   0,   0,   0,   0,  47, -21,  -7,   1,  -2,   0,  -1,   0,   0,   0,   0,   0,  -1,  -8,   7,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   4,   3,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -3,   0,   1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   53,   6, -26,   1,  -5,   1,  -2,   0,  -1,   0,  -1,   0,  74, -65,  10,  -2,   1,   0,   0,   0,   0,   0,   0,   0, -47,  19,  11,  -3,   3,  -1,   1,  -1,   1,   0,   0,   0,   2,   4,  -6,   1,   0,   0,   0,   0,   0,   0,   0,   0, -11,   2,   2,   1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,  -5,   1,   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   1,   0,   0,  -2,   1,   0,   0,   0,   1,   0,   0 },
     {   12, -34,  29, -10,   5,  -3,   2,  -1,   1,  -1,   1,   0,  66,  27, -54,   7,  -9,   2,  -4,   1,  -2,   0,  -1,   0,  28, -57,  17,   6,   0,   2,   0,   1,   0,   0,   0,   1, -27,  21,   7,  -7,   2,  -2,   1,  -1,   0,   0,   0,   0,   1,  -4,  -5,   4,   1,   1,   0,   0, -10,   6,   4,   0,  -1,  -1,   0,   0,  -1,   0,  -1,   0,  -1,   1,   0,   0,  -6,   3,   2,  -1,   1,   0,   0,   0,  -1,   0,  -1,   0,  -3,   2,   1,   0,   0,   0,   0,   0,  -2,   1,   1,   0 },
     {  -44, -80,  65,   1,  11,   0,   4,   0,   2,   0,   1,   0,  15, -28,  25,  -7,   3,  -2,   1,  -1,   0,   0,   0,   0, -14,  37, -11,  -5,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   6,   5,  -9,   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   7,  -4,   0,  -1,   0,   0,   0,   3,   2,  -4,   0,  -1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   1,   1,  -1,   0,   0,   1,  -1,   0,   1,   0,  -1,   0 },
     {    5,  -2, -11,   7,  -2,   2,  -1,   1,  -1,   0,   0,   0,  43,  17,   3, -23,   2,  -7,   1,  -3,   0,  -1,   0,  -1,  44,  20, -71,  15,  -4,   3,  -1,   1,  -1,   1,   0,   0,   8, -63,  25,  19,  -3,   4,  -1,   2,   0,   1,   0,   1, -12,  12,  13, -10,  -1,  -1,   0,  -1,  -3,  -8,   1,   1,   2,   1,   0,   0,  -7,   6,   4,  -1,   0,  -2,   0,   0,  -2,  -3,   1,   1,  -1,   1,   0,   0,  -4,   3,   2,  -1,  -1,  -1,   0,   0,  -2,   2,   1,  -1,  -1,  -1,   0,   0 },
     {   -4, -17, -22,  22,  -4,   6,  -1,   3,  -1,   1,   0,   1, -14, -66,  44,   9,   3,   2,   0,   2,   0,   1,   0,   1,  71, -39,  13,  -4,  -2,  -1,  -1,   0,  -1,   0,   0,   0, -24,   7,  19,  -7,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  17,  -2,  -9,   0,  -2,   0,  -1,  -7,   6,  -4,  -2,   2,   0,   1,   0,  -1,   8,  -3,  -1,   0,  -1,   1,   0,  -3,   4,  -3,   0,   0,   0,   1,   0,   0,   4,  -2,   0,  -2,   2,  -1,   0,   0,   2,  -1,   0,  -1,   1,  -1,   0 },
     {  -17, -20, -13,  27,  -3,   7,  -2,   4,  -1,   2,   0,   1, -13,   1, -11,  -1,   7,   0,   3,   0,   1,   0,   1,   0, -63, -34,   3,  42,  -2,   8,  -1,   3,   0,   2,   0,   1,  -9, -16,  70, -10,  -5,  -2,  -2,  -1,  -1,   0,  -1,   0,  -6,  35,  -4, -20,   2,  -4,   1,  -2,   7,  -1,   0,  -2,   1,   0,   0,   0,   2,  10,  -4,  -5,   0,  -2,   1,  -1,   5,  -1,  -2,   0,   0,   0,   0,   0,   2,   4,  -2,  -2,   3,  -1,  -1,   0,   1,   2,  -1,  -1,   2,   0,  -1,   0 },
     {  -21, -47, -63,  64,  -5,  15,  -2,   7,  -1,   3,   0,   2,   6,  13, -36,  29,  -9,   6,  -3,   2,  -2,   1,  -1,   1,   8,  26,  16, -16,  -1,  -4,   0,  -2,   0,  -1,   0,  -1,  25,   0, -17,  -2,   1,  -1,   1,   0,   0,   0,   0,   0,   1, -17,  10,   5,  -1,   1,   0,   1,   2,   0,   9,  -5,  -1,   0,   0,   0,  -1,  -2,   5,  -3,   0,   0,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,  -1,   0,   3,  -2,  -1,   1,   2,  -2,  -1,   0,   2,  -1,   0,   1,   1,  -1 },
     {    4,   2,  22, -31,  15,  -8,   5,  -3,   2,  -2,   1,  -1, -23, -49, -49,  53,   0,  12,   0,   5,   0,   2,   0,   2,   7, -12,   1,  15,  -4,   2,  -2,   1,  -1,   1,   0,   0,  46, -35,   2,  17,  -7,   2,  -2,   1,  -1,   0,   0,   0,  -6,  -6,  37,  -7,  -9,   1,  -2,   0,   4,  12,  -1, -14,   2,  -1,   0,   0,  -5,   3,   4,  -2,   2,  -1,   0,   0,   1,   6,  -1,  -4,   1,  -1,   0,  -1,  -2,   1,   1,   0,   0,   3,  -1,  -2,  -1,   1,   0,   0,   1,   2,   0,  -2 },
     {    9,   2,   8,  -1,  -7,   2,  -4,   1,  -2,   0,  -1,   0,   6,  -2,   3, -14,   7,  -5,   2,  -2,   1,  -1,   0,   0,  21,  -6,   8,  23, -19,   6,  -6,   2,  -3,   1,  -1,   0,  85,  11,   7, -54,   1,  -5,   1,  -2,  -1,  -1,   0,  -1, -12,   2, -51,  11,  12,   0,   5,   0,   9, -15,   8,  12,  -3,   1,  -1,   1, -11,   1,  -6,   1,   1,   1,   1,   0,   1,  -5,   4,   3,   0,   1,  -1,   1,  -6,   2,  -2,   0,   0,  -2,   2,   2,  -4,   1,   0,   0,   0,  -1,   1,   1 },
     {   -2, -12,   3,   8,  -4,   3,  -2,   2,  -1,   1,   0,   0,  11,  24,  47, -32,  -7,  -4,  -2,  -2,  -1,  -1,  -1,  -1, -17, -58,  36, -20,  14,   1,   4,   1,   1,   0,   1,   1,  45, -41, -21,  26,   3,   1,  -1,   1,  -1,   0,   0,   0,   0, -16,  29,   2,  -5,   0,  -2,   0,  -2,  14,  -4, -11,   3,  -2,   1,  -1,  -6,   4,  -3,   1,   1,  -1,   1,  -1,  -1,   8,  -4,  -1,   1,   0,   0,  -1,  -2,   3,  -3,   1,   0,   4,  -3,  -1,  -1,   1,  -2,   1,   0,   2,  -2,   0 },
     {   -2,  -6,  -4,  19,  -9,   5,  -2,   2,  -1,   1,  -1,   0, -13, -27, -11, -26,  29,  -5,   8,  -1,   4,   0,   2,   0, -12, -36, -50,  41,   6,   5,   3,   2,   2,   0,   1,   1,  -4,  17, -34,  20,   4,   3,   0,   0,   0,   0,   0,   0,  21, -30,  -2,  47,  -6,   0,  -3,   0,   6,   2,  35,   2, -20,  -1,  -5,   1,   4,  -1,   4,  -9,  -3,   4,  -1,   1,   0,   2,  10,  -4,  -1,   0,  -2,   0,   1,   2,   1,  -5,   0,   2,   4,  -2,   1,   2,   0,  -3,   0,   2,   2,  -1 },
     {    3,  -3, -12,   0,   8,  -4,   3,  -2,   1,  -1,   1,   0,   8,   4,   6,  -7,  -1,  -1,  -2,   0,  -2,   0,  -1,   0,   2,  -4, -24,  21,   9,  -6,   4,  -2,   1,  -1,   1,  -1,  34,  53,   8,  20, -41,   2,  -9,  -1,  -4,   0,  -2,   0,  54, -17,   9, -57,  -5,   4,  -2,   2, -15,   3, -33,   2,  18,   0,   5,   0,   8, -12,   3,   6,  -1,  -1,  -1,   0,  -7,   1,  -4,   1,   1,   0,   2,   0,   1,  -4,   3,   1,  -5,   0,  -1,   0,   0,  -2,   2,   1,  -3,   1,   0,   0 },
     {  -20, -28, -63, -63,  68, -13,  22,  -4,   9,  -1,   5,  -1,   2,   4,   5, -18,  19,  -9,   6,  -4,   2,  -1,   1,  -1,   1,   0,  15,  15,  -6,   0,  -5,   1,  -2,   1,  -1,   0,  -2,  -2,  -5,   4,   1,  -1,   0,   0,   0,   0,   0,   0,  -5,   6,  -6,  12,  -4,   0,  -1,   0,   2,  -5,   2,   8,  -3,   1,   0,   1,   1,   2,   3,   3,  -4,   1,  -1,   0,   1,  -1,   1,   3,  -3,   1,   0,   1,   2,   1,   2,   1,   1,   0,   1,   1,   0,   1,   1,   0,   0,  -1,   1,   0 },
     {    4,  -2,  -1,  -6,   0,  -2,   0,   0,   0,   0,   0,   0,   3, -11, -12,   4,  -5,   6,  -1,   2,  -1,   1,   0,   0,  -5,  -5,   1, -34,   6,   7,   0,   3,   0,   1,  -1,   1,   4, -32,   7,   4,  30,  -6,   2,  -2,   1,  -1,   0,   0,  93,  24, -32,  13, -16,  -6,  -3,  -1,  13, -37,  -2,  -5,   2,   6,   0,   3,   4,   0,   0,  10,   3,  -5,   1,  -1,  -5,  -5,   4,   3,  -5,   0,   0,   0,  -2,   3,   3,   0,  -2,   0,   1,   0,  -2,   3,   1,   1,  -2,   0,   1,   0 },
     {    0,   1,   0,   3, -11,   9,  -5,   6,  -3,   2,  -2,   1, -10,  -8, -42, -48,  45,  -7,  12,  -3,   4,   0,   2,   0,  11,   2,   4, -31,  32,  -8,   4,  -3,   2,  -1,   1,  -1,  16,  29,   0,  19,   5, -11,   0,  -5,   0,  -2,   0,  -1, -25,  50,  14,   0, -23,  -4,  -3,  -1,  14,  -1, -20, -21,   4,   2,   3,   3, -13,   1, -11,  14,   8,  -1,   2,  -1,   8,  -5,  -2,   5,  -5,  -1,  -1,   1,  -5,  -2,   0,   6,   4,  -4,   2,   2,  -2,  -1,   1,   3,   2,  -2,   1,   1 },
     {    4,   6,   6,   1,  -8,   3,  -3,   1,  -1,   1,  -1,   0,  11,  14,  16,   0,  18, -17,   4,  -6,   0,  -2,   0,  -1,  12,  27,  31,  37, -41,   4, -11,   1,  -5,   0,  -2,   0,  -9,   7,  11,  -7,  -4,  -7,   0,  -1,   0,   0,  -1,   0,  24, -10,  29,  11, -42,   3,  -5,   1,   4,  -4,  17, -65,  -2,  12,  -2,   4,  -9,   2, -24,  -2,  23,  -2,   4,   0,  -1,  -2,   1,  -2,   1,  -1,   0,   0,  -4,   0,  -9,   0,   0,  -2,   0,  -1,  -3,  -1,  -4,   0,   0,  -1,   0,  -1 },
     {   -2,  -4,   5,  -5, -15,  12,  -4,   4,  -2,   2,  -1,   1, -10, -19, -31, -48,  39,  -1,  12,   0,   4,   0,   2,   0,   5,  20,  32,  -4,  -6,  -2,  -2,  -1,  -1,  -1,  -1,   0, -15, -35,  19, -20,  -4,   7,  -2,   5,   0,   2,   0,   1,  19, -53,   1, -14,  25,   5,   3,   1, -29,  17,   0,  18,  19,  -7,   0,  -2,  11,   0,   2,  15,  -3,  -8,  -1,  -1,  -6,   7,   3,   2,  -6,  -2,   1,   0,   4,   1,  -1,   3,  -2,   3,   1,   2,   3,   1,  -2,   2,  -1,   1,   0,   1 },
     {   -8,  -9, -12, -14, -10,  12,   1,   5,  -2,   2,  -1,   1,  -5,  -3,   1,   8, -12,   0,   3,   0,   0,  -1,   1,   0, -18, -22, -38,  -9, -19,  24,  -2,   7,   1,   3,   0,   2,  -6,   7, -25, -27,  27,   0,   6,   0,   1,   0,   1,   0, -20, -12,  -1, -13, -14,  17,   1,   4, -20, -11,  -7, -24,  60,   7,   1,   3,  -4,  16,  -4,  46,  17, -24,  -1,  -5,   5,  -1,   9,   6, -13,  -5,   1,   0,   2,   4,   2,   7,   4,  -1,   4,  -1,   2,   2,   1,   3,   2,  -1,   2,   0 },
     {   -2,  -3,  -9,  -9, -16,  15,  -4,   2,  -2,   0,   0,   1,   4,   5,  -2,   3,   7, -13,   1,  -3,   1,  -1,  -1,   0,  -3,  -2, -13,   7, -37,  16,  -8,   3,  -2,   2,  -1,   1,   4,  -2, -25, -25,  34,   5,  -2,   4,  -3,   3,  -1,   1,  20,  38,  23,   4,  19, -26,   3,  -8,  -3,  59, -42,   2, -22,  -8,   2,  -3,  24, -26,  -2,   0, -11,   8,   0,   2,  -7,   5,  -3,  -3,   3,   1,  -2,   1,   7,  -8,  -2,   3,  -3,   0,   1,   0,   2,  -3,   0,   2,  -2,   0,   1,  -1 },
     {    5,   9,   4,  26,  49, -41,  12, -12,   2,  -5,   1,  -2,  -5,  -7,  -3, -22, -16,  12,  -1,   4,  -1,   2,   0,   1,  -6,  -1, -27, -41,   6,  15,   1,   8,   0,   3,  -1,   2,  13,  16,  38, -33,  22,  -5,  -1,  -1,  -1,  -1,  -1,   0,   2, -18,  24,  10,   1,  -6,  -4,  -1, -21,  23,  18, -23, -13,  -1,   0,   1,   6,  -6, -19,  -9,   9,   8,   2,   2,  -7,   2,  -1,  11,   8,  -5,  -1,   0,   2,  -2,  -5,   3,  -1,   1,   0,   5,   2,  -1,  -2,   2,  -1,   1,   0,   2 },
     {  -11, -16, -19, -54, -46,  37,  -2,  12,   1,   5,   0,   3,  -4,  -6,  -2, -28, -51,  40, -12,  10,  -2,   4,  -1,   3,   3,   9,  -6,  -1,  15,   6,  -2,  -2,   0,  -1,  -1,  -1,   5,  18,  25,  -3,   8,  -3,  -3,  -2,  -2,  -1,  -2,  -1,   3, -11,  20,   5,   9,  -6,  -2,   0,  17,   0,  19,  -9, -10,  -5,   1,   0,  -4,   2,  -7, -21,   6,   4,   1,   1,  -1,   2,  -5,   1,   9,  -2,   1,   0,  -1,   1,  -3,   1,  -1,   0,  -1,   5,  -1,  -1,  -1,   2,   0,   0,   0,   3 },
     {    2,  -2,   2,  17,  12, -15,   6,  -5,   2,  -3,   2,  -1,  -7, -14, -18, -37, -27,  11,   8,   3,   3,   1,   1,   1,   1, -11,   4,   1, -35,  11,   4,   2,   1,   0,   0,   0, -18, -14, -27, -12, -24,  24,   2,   5,   0,   3,   1,   2,  -5,  -4,   4, -21,  11,   7,   2,   1,  57, -25, -27,  -5,  -8,  23,   1,   4,  -1, -19,   8,  -6,  35,  17,  -5,   2,   3,   6,   1,  17,  16, -13,  -7,  -4,  -2,  -3,   5,   1,   2,   5,   1,   4,   0,   1,   3,   1,   0,   3,   0,   1 },
     {    6,   4,   6,  11,  17,  -1, -10,  -1,   0,   1,   0,   0,   2,  -2,   0,  -7,  -8, -10,   0,  -1,   3,  -1,   1,   0,   7,   2,   3,  -6, -26,  21,  -9,   5,  -2,   2,   0,   1,   0,  17,  22,  19,  17, -15,  -1,  -6,  -1,  -2,   0,  -2,   2, -21,  19,   8,   8,   9,  -7,   1,  72,   7,  -5,  19,  26, -29,  -8,  -2,   0,   7,  -4,  15, -44, -24,   5,  -1,   5, -14,  -7, -21, -14,  18,   7,   4,  -5,   1,  -2,   1,   2,  -3,  -3,  -6,  -2,   0,  -2,   1,   0,   0,  -1,  -2 },
     {    1,   7,   2,  19,   7, -20,   2,  -1,   3,  -3,   0,  -1,  -9,  -7, -16, -36, -42,  38,  -6,  10,   0,   3,   0,   2,   5,  15,  21,  29, -28,  16, -12,   3,  -4,   1,  -1,   0,   3,   0, -17,  47,   3,  -5,  -1,  -5,   0,  -1,   1,  -1,   0,  32, -29,   8,  -1,   5,  -2,  -1, -41,  24,   3,  -2,  16,  -9,   1,  -3,  -7,  15,  -7,   5,  -4,  -9,   1,   0,  -1,   0,   5,  -7,  -4,   2,   3,   1,   0,   0,  -1,  -4,   2,  -4,   1,  -2,   1,  -1,  -1,  -2,   0,  -2,   1,  -1 },
     {   -1,  -1,   7,   3,  -4,   3,  -2,   0,   0,   1,   0,   1,  -4,  -4,   2,  -5,  -2, -10,   2,  -1,   3,   1,   0,   0,   0,   2,  12,   3, -23,  28,  -5,   4,  -1,   1,  -1,   0,   7,  19,  22,  26,  39,  -7, -10,  -1,  -6,   1,  -4,  -1, -18, -15,  25,   2, -10, -31,   5,  -5, -35, -58, -16,  24, -18,  16,   4,   3,  -6, -28,  40,   7,   0,   8,  -1,   2,  -2,  12,   8,  -6,  -4,  -1,  -4,  -1,  -1,   1,   3,  -4,  -1,   3,  -3,  -4,  -1,   0,   0,  -2,   0,   2,  -1,  -3 },
     {    2,   2,   6,  15,   7, -21,  10,  -9,   5,  -2,   1,  -1,   3,   3,   7,  -7,   1,  35, -17,   8,  -5,   2,  -2,   1,   7,   7,  11,  52,  46, -25,  -2,  -5,  -3,  -2,  -4,  -2,  -7,  -5, -23, -37,  32,  -5,   4,   1,   4,   0,   1,  -1,   9,   9,  39,  -6,  13, -17,  -6,   1,  10, -22,  -7,   6,   7, -14,   4,  -3, -16,  -5,   2,  21,   0, -13,   1,  -1,   1,  -4,   4,   4, -25,  -3,   7,  -1,  -3,  -2,  -1,  -6,  -1,  -3,  -2,   0,  -3,  -1,   0,  -3,   0,   0,  -2,  -1 },
     {    0,   3,   2,  -3,  -2,  -6,   1,  -1,   0,   0,   0,  -1,   0,   3,  -5,  -6,   6, -16,   7,  -3,   5,  -1,   1,  -1,  -2,  -3,   3, -19, -28,   5,   5,  -1,   2,   0,   0,  -1,   5,  16, -20, -10, -11,  32, -11,  10,  -6,   3,  -2,   1,  17,  36,  38, -23,  31, -12, -10,  -1, -12, -22,  55,  13,   3,  -6,  -9,  -1,  -7,  45,  16, -21,  -3, -18,  -1,  -3,   5,  -1, -14,  -3, -10,  -4,   6,  -1,  -3,   6,  -3,  -5,   0,  -2,  -7,   3,   1,   1,  -1,   2,  -1,  -3,  -3,   2 },
     {   -4,  -5,  -3,  -3, -22, -26,  14,  -1,   2,   1,   0,   0,  -7, -11, -11, -15, -30, -34,  23,  -5,  10,  -1,   5,  -3,  -2,  -2,   1,   4,   3, -12,  -3,   1,  -4,   2,   0,   1,  -5, -11, -17, -25,   3, -31,  20,  -4,   5,  -1,   3,  -2,   3,   8,  16, -26, -35,  24,   2,   5, -12, -10,  16,  15,  -1,  -5,   6,  -3,  -4, -10,  -4,  -6, -41,  33,  10,   1,  -4, -10,   5, -17,  27,  30,  -5,   1,   2,   6,  -2,  10,   0,  -5,   2,  -2,  -1,   1,   0,   2,   1,  -2,   1,   1 },
     {    5,   8,   8,   4,  54,  56, -50,  15, -20,   4,  -6,   4,  -3,  -2,  -4,  -4,   2,   5, -11,  -1,  -2,   3,   1,   1,  -4,  -4,  -4,   4, -11, -19,  -1,   9,  -1,   2,  -1,   1,  -7,  -3, -18,   2,  22, -29,  10,  -3,   0,   1,  -1,   0,   4,  -5,  -5, -36,   5,  -7,   3,  -6,  -7,  -4,  12,  -2,  -4, -14,   5,   2, -30,   0,  -8, -16,   2,  14,  12,   0,   2,  -5,  -2,  15,  20,   4,  -4,  -2,   0,   3,   2,   9,  -2,  -1,  -2,   2,   0,   2,   4,   0,   0,   0,  -2,   0 },
     {    5,   4,   6,   4,  18,  19, -14,   5,  -6,  -3,   0,  -2,   1,  -2,  -1, -14, -15,   5,   0,  -2,   0,   2,   2,   2,   2,   1,   1,   5,   0, -30,  -2,  -1,   7,   1,   1,   2,   8,   1,   3, -12,   3, -19,  12,   1,   9,   0,   3,  -1,  -4,  15,  12,   6, -32,  21,  19,  -2,   4,   8,  14,  21,   9,  40, -30,   3,  53, -10,  29, -11,  10, -38, -22,   1, -14,  14, -16, -12,  -9, -12,  13,   3,  12,  -2,  -7,   3,  -9,   1,  -5,  -2,   1,  -4,  -2,  -2,  -2,   1,  -3,   1 },
     {   -1,  -1,  -4,   0, -11, -24,  -2,  -2,   3,   2,   3,  -2,  -1,   0,  -7,   0,  -5, -22,  -1,  -4,   5,  -3,   2,  -2,   3,   6,   2,   8,  23,  26, -24,   5, -10,   4,  -4,  -2,  -5,  -2, -14,   3,  15, -37,  -9,  12, -11,   3,  -3,  -1,  -3, -22, -20, -32, -15, -39,  23,   2,  18,  19,  -3,   3, -19,  19,  12,   1,   8,  51,   2, -15,  17, -16,   7,  -7,   3,  13, -27,  13, -11,   0,   1,  -6,   1,  -3,  10,  -3,  -3,   6,  -9,  -3,   4,   3,   5,   0,  -1,   1,   0,   0 },
   }
 },
 {  //6
   {
     {   93, -51,   0,  -4,   0,  -2,   0,  -1,   0,   0,   0,   0, -62,  24,   9,   0,   2,   0,   1,   0,   0,   0,   0,   0,  13,  10, -11,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  -4,  -9,   3,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   5,   2,  -2,   0,   0,   0,   0,  -2,  -1,  -2,   0,   1,   0,   0,   0,   0,   2,   0,   1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -69,   2,  16,   2,   4,   1,   2,   1,   1,   0,   1,   0, -66,  59,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  49,  -6, -22,   3,  -4,   1,  -1,   0,  -1,   0,   0,   0,   3, -17,   8,   2,   0,   1,   0,   0,   0,   0,   0,   0,   9,   2,  -1,  -4,   0,   0,   0,   0,   1,  -5,  -2,   1,   2,   0,   0,   0,   4,   0,   0,   1,  -1,   0,   0,   0,   2,  -3,   0,   0,   0,   1,   0,   0,   3,   0,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   1,  -1,   0,   0 },
     {   20, -14,   4,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,  46,  37, -39,   1,  -7,   0,  -3,   0,  -1,   0,  -1,   0,  34, -80,   5,  14,   0,   4,   0,   2,   0,   1,   0,   1, -31,   8,  37,  -8,   1,  -2,   0,  -1,   0,   0,   0,   0,  -1,  11, -10,  -9,   1,  -1,   1,   0, -10,   0,   2,   4,   3,   0,   1,   0,   0,   3,  -1,   0,  -2,   0,   0,   0,  -4,   0,   2,   0,   1,   1,   0,   0,   0,   2,  -1,   0,  -2,   0,   1,   0,   0,   1,   0,   0,  -2,   0,   1,   0 },
     {   22,  40, -30,  -1,  -5,  -1,  -2,   0,  -1,   0,  -1,   0,  34,   9,  -5,  -5,  -2,  -2,  -1,  -1,   0,  -1,   0,   0,  26,  15, -54,   4,  -3,   2,   0,   1,   0,   0,   0,   0,  13, -72,   5,  30,  -2,   6,  -1,   3,   0,   1,   0,   1, -11,  -1,  40,  -8,  -4,  -1,  -1,   0,  -3,   0,  -3,  -8,   3,   0,   1,   0,  -6,   1,   6,   1,   1,  -1,   0,   0,  -2,   0,  -1,  -1,  -1,   0,   1,   0,  -3,   0,   3,   0,  -1,   0,   0,  -1,  -2,   0,   2,   0,  -1,   0,   0,   0 },
     {   20,  30,   3, -11,  -2,  -3,  -1,  -2,   0,  -1,   0,  -1,   4,  39, -39,   1,  -2,   1,  -1,   0,   0,   0,   0,   0,  13,   5,   6, -10,   1,  -2,   0,  -1,   0,  -1,   0,   0,  -5,  19, -63,   3,   7,   1,   3,   0,   1,   0,   1,   0,   8, -48,  -1,  51,  -5,   5,  -2,   2,   3,  -4,  35,  -8, -17,   1,  -3,   1,   0, -10,  -1, -11,   5,   5,   0,   1,  -2,   0,   7,   1,   3,  -3,  -1,  -1,   0,  -3,  -1,  -1,  -1,   0,   3,   0,   0,  -2,   0,   0,  -1,   0,   2,   0 },
     {   22,  85, -37,  -1,  -9,  -2,  -4,  -1,  -2,  -1,  -1,  -1, -55,  -5, -18,  15,   1,   3,   1,   2,   1,   1,   0,   1, -28, -30,  20,  15,   2,   3,   0,   1,   0,   0,   0,   0,   3,   9,  26,  -8,  -4,  -1,  -1,  -1,  -1,   0,   0,   0,   6,   6,  -5, -12,   2,   0,   0,   0,   4,  -4,  -4,   4,   4,  -1,   0,   0,   3,   0,   3,   2,  -3,  -1,   0,   0,   3,  -2,   0,  -2,  -1,   1,   0,   0,   2,  -1,   0,  -1,   1,  -1,   0,  -1,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   11,  17,  -6,  -2,   0,  -1,   0,  -1,  -1,   0,   0,   0,  21,  34,  11, -24,  -1,  -7,   0,  -3,   0,  -2,   0,  -1,  -8,  28, -30,  -1,   2,   0,   1,   0,   0,   0,   0,   0,  15,   9,   2, -20,   0,  -4,   0,  -1,   0,   0,   0,  -1, -17,   8, -65,  -2,  23,   0,   7,  -1,   1, -21,  -6,  62,  -1,  -4,  -1,  -1,   4,  -4,  18,   2, -29,   0,  -2,   0,   1, -12,   1,  -5,  -2,  11,   0,   2,   1,  -2,   6,  -1,   0,  -4,   0,   0,   0,   0,   2,   0,   0,  -3,   0,   1 },
     {  -12, -23,   8,   4,   4,  -1,   2,   0,   1,   0,   0,   0,   2,  47, -23,  -4,  -2,  -2,   0,  -1,   0,  -1,   0,   0, -94,  -1, -31,  23,   8,   5,   4,   2,   2,   1,   1,   1,  -4,   1,  13,  32,  -7,   2,  -3,   1,  -1,   0,  -1,   0,  12,   1,  18,  -2, -13,   1,  -3,   1,   9,  -2,  10,  -7,  -4,   2,  -1,   0,   6,  -1,   4,  -9,  -1,   2,  -1,   0,   4,  -2,   1,  -3,   3,   1,   0,   0,   3,   0,   0,  -2,   2,   0,   0,  -1,   2,   0,   0,  -1,   1,  -1,   0,  -1 },
     {   10,  18,  21, -23,   0,  -4,   0,  -2,   0,  -1,   0,  -1,  14,  36,  -2,   0,  -7,  -1,  -4,  -1,  -2,   0,  -1,   0,  -6,  17,  15, -31,   3,  -3,   2,  -1,   1,   0,   1,   0,  -2,  23, -20,   8,  -3,   1,  -1,   0,  -1,   0,   0,   0,  17, -10,  14, -44,  -6,   3,  -1,   1, -20,   7, -54, -11,  42,   0,   8,   0,   2,  -6,  -7,  48,   2, -10,   0,  -2,  -1,   0,   0,   2, -15,   0,   1,   0,   2,  -7,   0,   6,  -1,  -1,   1,   0,   1,  -3,  -1,   3,  -1,   0,   0,   0 },
     {    3, -19,  -2,   8,  -1,   3,   0,   0,   0,   0,   0,   0,   8,   5, -25,   7,  -3,   2,   0,   0,   0,   1,   0,   0,   4,   1,  11,  24, -11,   4,  -4,   2,  -2,   1,  -1,   0,  96,   1,  15, -25, -21,  -1,  -6,   0,  -3,   0,  -2,   0,   5, -40,  -6, -25,   6,   6,   1,   3,  -5,  -9,  14, -11,  16,   1,   0,   0, -10,   4,  -6,  -5,  13,  -5,   2,  -1,  -6,   4,  -1,   7,   5,  -4,   2,  -1,  -4,   2,  -3,   2,  -3,   1,   0,   1,  -2,   1,  -1,   0,  -1,   0,   0,   1 },
     {  -11, -26, -17,  -2,  10,   2,   4,   1,   2,   0,   1,   0, -14, -30, -33,  31,   3,   4,   1,   3,   0,   1,   0,   1,   5, -19,  -2,  -2,   9,   0,   3,   0,   1,   0,   1,   0,   8,  -2, -19,  33,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0, -29,  17,  -4,   6,   3,   2,   1, -12,  18,  -4,  46,  13, -14,   2,  -4,  11,   0,  31,  16, -53,  -5,  -3,   0,  -1,  -1,   6, -32,  -7,  19,   1,   3,   3,   1,   7,  -2,  -1,   3,   2,  -8,   1,   1,   2,   0,   0,   2,   1,  -3 },
     {  -12, -23, -80,  32,   1,   9,   2,   4,   1,   2,   1,   1,  10,  15, -29,  33,  -4,   2,  -1,   0,   0,   0,   0,   0,  15,  49,   9,   4, -17,  -3,  -4,  -2,  -2,  -1,  -1,  -1, -29,  23,   6,  -6,  -7,   1,   0,   0,   0,   0,   0,   0,  10,  12,  -5,  -8, -11,   2,  -2,   1,   5, -21,  -7, -11,   3,   8,   0,   3,   3,  -9,   4,   5,  11,   0,  -2,   0,   1,  -5,   5,   1,  -1,  -3,  -1,   1,   0,  -3,   3,  -2,   0,  -3,   2,   0,   0,  -1,   2,  -1,   0,  -1,   1,   0 },
     {    7,  10,   6,  -4,  -1,  -2,   0,  -1,   0,   0,   0,   0,  11,  19,  23,  19, -23,  -4,  -6,  -1,  -2,  -1,  -2,   0,  12,  10,  23, -13,  -2,  -4,  -1,  -2,  -1,  -1,   0,  -1,   6,  14,   8,  14, -26,  -1,  -4,   0,  -2,   0,  -1,   0,  15,   2,  19, -28,  -3,   0,   0,   0,   2,  18, -11,   9, -30,  -6,   0,  -1,   9, -24,   5, -56, -21,  33,   0,   8,  -7,  -2, -17, -15,  54,   8,  -3,   1,   1,   0,  -1,  11,  -4,   0,  -7,  -3,   1,  -1,  -1,   3,  -2,   0,  -3,  -1 },
     {    8,   7,  13,  -8,   3,  -4,   1,  -2,   0,  -1,   0,  -1,   0, -34,   5,  -4,   5,  -1,   3,  -1,   1,   0,   1,   0,  24, -13, -45,  11,  12,   1,   4,   0,   2,   0,   1,   0,  14,  41,  -4,  38, -11,  -6,  -4,  -3,  -1,  -1,  -1,  -1,  51,  11, -20,  -5, -36,   2,  -4,   1,   3, -49,  -5,  -8,  -2,  18,   1,   5,   0,  -9,  23,  -1,  11,   3,  -4,   1,  -3,  -1,   6,  -4,   3,  -4,  -1,   0,  -3,   1,   4,  -2,  -3,   0,   0,  -1,  -1,   0,   2,   0,  -1,  -1,   0,   0 },
     {    6,  15,  55,   1,  -8,  -6,  -3,  -3,  -1,  -2,  -1,  -1,  -6, -22, -63,   7,   6,   2,   6,   2,   2,   1,   2,   1,  18,  48, -19,  30,  -3,  -1,  -3,  -2,  -2,  -2,   0,  -1, -12,  26,  21,  -2, -13,  -7,  -3,  -3,   0,  -1,   0,  -1, -39,  17,   7,  -5,   0,   0,   2,   0,   7,  23,  -5,  -4,  -4,  -2,  -1,  -1,   0,  -9, -11,  -1,  -6,   3,   1,   2,   4,  -5,   2,  -2,  -1,   4,  -1,   2,   1,  -5,  -2,  -2,   2,  -3,   1,  -2,   0,  -2,   0,  -1,   1,  -1,   1,  -1 },
     {    5,   1,   2,   2,  -9,  -3,   0,   0,  -1,   1,   0,   0,  12,  18,   7,  11,   5, -13,   0,  -4,   1,  -2,   0,  -1,   5,   0,  23,   8, -16,  -3,  -5,   1,  -2,   0,  -1,   0,  44,   0,  -1,  -1,  20, -14,   3,  -4,   0,  -2,   0,  -1,  20,  50,   9,  36, -30, -16,  -5,  -5,  21,   7,  -8, -18, -19, -11,   2,   0,  -4,   4, -15,  19, -38,  -1,   8,  -2,   7, -19,   9, -28, -25,  27,  -1,   4,  -3,  -5,  -4, -10,   1,  -6,   2,  -6,  -3,  -2,  -3,  -3,   0,  -3,   1,  -2 },
     {   -7, -17, -18, -36,  13,   3,   7,   1,   2,   1,   2,   0,  -8,  -9, -48, -13,   7,  10,   3,   4,   2,   2,   1,   1,  -8, -14,  -1, -20,  13,   1,   4,   1,   1,   1,   0,   0,  28, -19, -22,  -7,  13,   2,   3,   2,   1,   0,   1,   1,  -2,  38,  -5,   7,  15, -12,   2,  -4,  12,   3, -45,   8, -26,   5,   4,   1,  -3, -32,   1,   1,  22,  26,   0,   5,   1,   8,   7,  23,  20, -14,  -5,  -3,  -1,  -4,   7,  11,  -1,   3,   3,   4,   0,  -2,   3,   3,   0,   2,   2,   2 },
     {    2,   1,  42, -16,  -5,  -4,  -3,  -1,   1,  -1,   0,   0,   5,  -3,  -8,  38,  -9,  -1,  -2,   0,  -1,  -1,  -1,   0, -11,  11,  30,  16, -36,   1,  -4,   1,  -2,   1,   0,   0, -20, -53,  -1, -19,  -4,  18,   5,   5,   1,   2,   1,   1,  21,   0, -11,   2,   5,   8,  -2,   0,  15, -42,  -9,   7,   1,  11,  -4,   3,   0, -10,  42,   5,   7,   3,  -6,   0,   5,  15,   9, -23,  -2,  -9,  -3,   0,  -5,   3,  -3,  -6,  -1,   6,  -2,  -2,  -1,   1,  -1,   0,   0,   2,  -1,  -2 },
     {    6,   4,  22,  78, -27,  -7,  -9,   0,  -3,  -1,  -1,   0,  -2,   2,   4,  43, -37,   5,  -6,   4,  -2,   1,  -1,   1,  -8, -22, -27, -29,   4,  10,   4,   4,   2,   1,   1,   1,  18,   9, -24,  -6,  16,   5,  -1,   0,   0,   0,   1,   0, -22,  16,   4,   1,  11,  -6,   0,  -2,  -6,  -3, -13,   1,  -5,   4,   2,   0,   1,  -9,  12,   2,  13,   4,  -2,   0,   1,   5,   6,   8,  -5,  -9,  -2,  -1,   3,  -4,   4,  -6,   1,   2,   1,  -3,   0,  -2,   1,  -3,   0,   0,   1,   0 },
     {    2,  -1,   4,  16, -17,  10,  -4,   2,  -2,   0,  -1,   1,  -2, -21, -17, -11,  -7,   1,   7,   1,   3,   1,   1,   0,  -6,  -3, -36, -15, -28,  18,   3,   7,   0,   3,   0,   1, -17, -14,  -8, -34,   7,  10,   5,   1,   2,   0,   0,   1,  60, -11, -12,  -9,  13,  16,  -1,   2,  20,  19,   1,  -8,  17, -17,   1,  -5,   1,   5, -37,   6, -22,  15,   5,   0,   6, -22,   7,  -2,  29,  17,  -7,   2,  -5,   8,  -1,  17,  -1,  -6,   4,   3,  -2,   3,   2,   4,   0,  -2,   3,   1 },
     {   -6,  -5,  -6,  -6, -11,   1,   2,   2,   1,   2,   0,   1,  -7,   0,  12,   4,   4, -16,   7,  -2,   3,  -2,   2,  -1,  -8, -18,   0,   9, -18,   3,   1,   2,  -1,   2,   0,   0,  -3,  -7, -17,   6,  -6, -12,  10,  -3,   5,  -1,   1,   1, -63, -12, -22,  -3, -42,  23,   7,   6, -10, -18, -11, -22,  15,  20,   7,   3, -15,  -7, -12,  17, -10,  31,   1,   3,  11, -13,  23, -12,  39,  20, -13,   2,   1,  12,   4,  15,   6,  -2,   6,  -6,  -1,   6,  -2,   3,   4,   1,   2,  -1 },
     {    1,   0,  -6,  13,   5, -14,   1,  -3,   1,  -1,   1,  -1,   4,   1,  25,   9,  12, -14,  -2,  -2,   0,  -1,   0,  -1,   2, -14, -10,  68,  -9, -24,   3,  -5,   1,  -3,   0,   0,  -8,  -4, -32,  -6, -41,  10,   5,   5,   2,   1,   1,   2,  17,  -9,  -8,   0,  15,  -2,   5,  -1, -16,  33, -27,   7, -26,   4,   4,  -2,   9, -18, -19,  15,   7,  15,  -2,   3,   0,   1,  16,  17, -18,  -1,  -5,  -1,   6,  -8,   4, -16,  -3,   0,   4,   2,   2,  -1,   2,  -6,  -2,   1,   2,   0 },
     {   -2,  -3, -12, -45,   5,   1,   8,  -2,   2,  -1,   3,  -1,   3,   2,  13,  32, -15,  -5,  -5,  -1,  -1,  -1,  -1,   0,   2, -18, -30,   3, -22,   0,   4,   2,   1,   3,   1,   1,   8,  21, -29, -16, -11,  15,   3,   2,   0,   2,   0,   0, -23,  24,   3, -11,   0,  19,   4,   4,  54,  25,  14,  -4,  24,  14, -14,  -1,  25,   9,   4, -12,  10, -27,  -9,  -4,   2,  -5, -19,  -7,  -7,  -5,   7,   1,  -1,  -5,  -3,  -1,  -2,  -2,  -5,   6,   0,  -3,   3,   3,  -1,  -1,  -1,   0 },
     {   -7,  -2,  -1, -43,  -9, -12,   4,   3,   3,   1,   1,   1,  -2,  -5,  -4,  45, -10,  -5,   2,   1,  -1,   1,  -1,   1, -11,   6, -23,  -1, -18,  -1,   2,   4,   1,   1,   0,   2,  -1,   5,  -2, -13,  13,   5,   8,   3,   2,   2,  -1,  -1,   8,  13,  -4,   5,   4, -13,   3,  -5, -75, -13,   9,  -3, -13,  -3,  10,   4, -18,  30, -13, -15,   0,   3,  14,  -2,  -8,  -6, -27,  17,  -4,  21,   3,   0,  10,  -5,  15,   3,   2,  -1,  -1,  -4,   3,  -1,   6,   0,   1,  -1,   0,  -1 },
     {   -7, -11, -14, -17, -63,  -3,  11,   8,   3,   2,   1,   1,  -9,  -9,  -5, -57, -63,  11,  10,  11,   4,   4,   1,   2,   2,  -2,   7,   1, -27,  -1,  10,   0,   3,  -1,   0,  -1,   1,  10,   5,   3, -18,  15,  -2,   1,  -2,   1,  -1,   0,  -4,  -1,  16,   6,   1,   0,   3,  -4, -13,   0,   2,   3, -10,  14,  -2,  -2,   5, -12,  12,   5,   2,   1,  -4,  -2,   2,  -1,   6, -12, -17,  11,  -1,   3,   3,  -7,   1, -19,   0,   3,   0,  -2,   0,  -3,   1,  -5,   0,   2,  -1,  -1 },
     {    1,  -3,   7,  13,   2, -26,   8,  -3,   2,  -1,   2,  -2,  -4,  -9, -20,   2, -21, -12,  -1,   2,   0,   2,   0,   2,  -2,  -1,  19, -22,   8, -24,  11,   0,   3,   0,   1,  -2,  -6, -34,  -1,  23, -13, -13,   1,   0,  -1,  -1,   1,   1,   2, -16, -37, -17, -25,  -6,  19,   6,  34,  -2,  -4,   1,  -8,  20,   4,  -1,   4,   7, -23, -15,   0, -13,  10,  -5, -11, -15, -27,  28, -24,  32,  12,  -3,   6, -14,  31,  -5,  -2,   4,   1,  -2,   1,   1,   6,  -3,   0,   4,  -2,   0 },
     {   -1,  -4,   0, -28,   5,  -1,   4,   1,   1,  -1,   1,   1,   2,  -4,  21,  28, -14,   4,  -5,  -2,  -3,   0,  -2,   1,  -1,  -3, -15,  -6,   5,   2,  -2,   1,  -1,   1,   0,   2,  -3,  12,  38,  -3,  17,   6,  -7,   0,  -3,   3,  -4,  -1, -13, -39,  15,  13,  18, -11,  -6,   1,  23, -29,  -8,  -6,  -9, -23,   4,   3,  10, -54, -24,   9,  -6,  -4,  16,   3,   2, -24,  29,  20,  -9,   9,   7,   1,  -1,  10,  18, -22,  -7,   5,  -6,   1,   0,   3,   4,   2,  -1,   0,  -1,  -1 },
     {   -3,  -3,  -1,  13,  -8,  -2,  -2,   2,  -1,   2,  -1,   1,  -5,   1,  -3, -14,  15, -12,   2,   3,   2,   2,   1,   1,  -2, -12,   1,   9,  -3, -21,   6,  -1,   1,  -1,   1,   2,  -8,   9, -26,  -3, -23,   9,  10,   4,   3,   2,   0,   0,  -9,  19,  23,  -5,  25,  11,  -6,   0,  15, -39,  -2,  -4,  13, -34,  -4,   3, -54, -27,  -1, -12, -14, -18,  30,   3, -22,  18, -23,   1,  -5,  25,  23,  -1,  -4,   7,  -5,  19,   4,   5,   3,   2,  -1,  -3,   0,   5,   6,   0,   2,   1 },
     {    2,   4,   0,  14,  44, -20,   1,  -8,  -2,  -2,   1,  -1,  -5,  -5, -10, -19,  25, -32,  -1,  -3,   1,  -2,   1,   1,  -8,  -9, -20, -49, -48, -27,   7,   7,   4,   3,   2,   1,   3,  12,  21, -33, -18,   5,   7,   5,  -2,   3,   0,   1,  -5,  -9,  24,   3, -13,  -4,  11,  -1, -14,   0,   3,  -3, -20,  19,   5,   1,  11, -10,  16,   2,  -2,   4,  -6,  -2,  -5,   6,   6, -14, -11,   6,   5,   0,   3,   4,  -5, -10,  -3,   5,  -4,   3,   0,  -1,  -2,   2,  -3,   2,  -1,   1 },
     {   -3,  -6,  -7,  -1, -42, -19,  -2,   5,   3,   4,   0,   3,   1,   0,   5,  -1,  19,  -7,   5,   3,   1,  -4,  -1,  -1,  -1,   0, -14,  23,   8, -11,  -5,   0,  -2,   0,  -1,   0,  -6,   2,   1, -27,  41, -20,   0,   3,  -1,   0,   1,  -1,   0, -31,   6,  -4, -11, -40,  16,   0,  12,   2, -21,  -6,  -6,  13,  19,  -1,  28, -12,  13,  -2,   2,  -5,  -4,  -4, -24,  12, -31, -20,   3,  -2,  20,  -2,  -6,   5, -14,  37,   4,  -1,  14,  -2,   3,   0,   0,   8,   0,   0,   6,  -2 },
     {   -1,  -2,   0,  -7,  -8,   3, -14,  31, -10,  13, -15,   2,   2,   3,   0,  12,   8,   0,  10,  29, -15,  10,  -7,   2,  -2,   3,  -4, -13,  43,  -2,  11, -15,  -3, -16,  11,  16,   2, -12,   9, -22, -24, -21,  24, -16,  20,   7,   5, -12,  10,   5,   2,  16,   6,  32,   3,  -8,  -9,  10,  -2, -11,  11,  21,  13,  -6,   1, -12,  13, -12,   7,   2, -10,  -2, -19, -30,   9,  -7,  -9,  15,  14,  22,  -7,   9, -16, -17,  13,  13,   1,   6,  11,  -2,   3,  -7,  -1,  -9,  -3,  -2 },
     {   -2,  -1,  -9, -12,  -1, -19,  15,   6,   1,   9,   3,  -3,  -2,   1,   3,  14, -32, -28,   1,  -2,  -3,   5,   5,  -1,   0,   7, -17,  -3,  11, -21,   9,   5,   1,   2,   1,   1,  -3,   8,  22,  -2,  14, -21,  12,   1,   3,  -3,  -6,  -1,   9, -18, -16,   7,  -4,   9,   2,   1,  22,  32,  -6,  -6,  -2,  -9,   3, -10, -66,   7,  18,   7,  -8,  11,  -1,  -6,  16,  44,   8,   3,  -2, -20, -11,   1,   1,  -4, -15,  -3,  -6,   5,  -5,   6,   6,   5,  -5,  -8,  -5,  -3,  -2,  -1 },
   },
   {
     {  -72,  54,  -8,   3,  -2,   1,  -1,   1,   0,   0,   0,   0,  69, -47,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -27,   9,  11,  -5,   2,  -1,   1,  -1,   1,   0,   0,   0,   8,   5, -11,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -2,  -5,   4,   2,  -1,   0,   0,   0,   2,   2,  -1,  -2,   0,   0,   0,   0,  -1,  -2,   0,   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -59,   3,  24,  -2,   4,  -1,   1,   0,   1,   0,   1,   0,  -2,  49, -37,   5,  -5,   1,  -2,   0,  -1,   0,  -1,   0,  45, -60,  19,   3,  -1,   1,   0,   1,   0,   0,   0,   0, -26,  28,   5, -12,   4,  -3,   1,  -1,   1,  -1,   0,   0,  13,  -7, -14,   9,  -1,   1,   0,   0,  -4,  -1,  10,  -2,  -2,   1,  -1,   0,   4,  -1,  -5,  -1,   2,   0,   0,   0,  -2,   0,   3,   1,   0,   0,   0,   0,   2,   0,  -2,   0,  -1,   0,   1,   0,   1,   0,  -1,   0,  -1,   0,   1,   0 },
     {   63, -24,  10, -10,   0,  -2,   0,  -1,   0,  -1,   0,   0,  52, -19, -32,  10,  -3,   2,  -1,   1,  -1,   1,   0,   0, -30, -27,  43,  -6,   3,  -1,   1,   0,   1,   0,   0,   0, -20,  44, -21,  -6,   4,  -3,   1,  -1,   1,  -1,   0,  -1,   4, -18,   0,  15,  -5,   3,  -1,   1,  -6,   8,   9, -10,   1,   0,   0,   0,  -1,  -1,  -6,   3,   2,  -1,   1,   0,  -3,   3,   3,  -1,  -1,   0,   0,   0,  -1,  -1,  -2,   1,  -2,   2,   1,   0,   0,   0,  -1,   0,  -1,   1,   1,   0 },
     {   33,  25, -43,  18,  -8,   3,  -3,   1,  -2,   0,  -1,   0,  35, -17,  20, -19,   1,  -2,   0,  -1,   0,  -1,   0,   0,  29, -40, -19,  19,   0,   2,   0,   1,   0,   1,   0,   1, -30,  -5,  49, -16,   0,  -1,   1,  -1,   0,   0,   0,   0,  -8,  29, -33,  -2,   9,  -3,   2,  -1,   0, -17,  11,  14,  -9,   3,  -2,   1,  -4,   9,   2, -13,   5,   0,   0,   0,  -3,  -3,  -2,   7,   0,  -1,   1,   0,  -2,   4,   1,  -3,  -1,  -2,  -1,   2,  -1,   2,   1,  -2,  -1,  -1,   0,   1 },
     {   12,  45, -38,   0,  -3,   0,  -1,   0,  -1,   0,   0,   0, -45, -32,  29,  14,  -1,   4,  -1,   2,   0,   1,   0,   1,  40,  -3,  17, -28,   7,  -5,   2,  -2,   1,  -1,   0,  -1, -39,  28, -34,  19,   2,   0,   1,   0,   1,   0,   0,   0,  27, -21,  17,   2, -10,   4,  -3,   2, -10,   5,  -2, -10,   8,  -1,   1,   0,   6,  -1,  -1,   7,  -2,  -1,   0,   0,  -2,   0,   0,  -3,   0,   0,   0,   0,   3,  -1,   0,   2,  -1,   0,   0,  -1,   1,  -1,   0,   1,  -1,   0,   0,  -1 },
     {  -20, -15,   4,  16,  -2,   2,   0,   1,   0,   1,   0,   1, -42, -20,  31, -18,  11,  -1,   3,  -1,   2,   0,   1,   0, -33,  19,  20,  12,  -3,  -2,  -1,  -1,   0,   0,   0,   0,   2,  47,  -4, -38,   5,  -4,   1,  -2,   1,  -1,   0,  -1,   4,   5, -51,  33,   2,  -1,   1,   0,  16, -17,  28,  -2, -16,   6,  -4,   2,   0,   6, -12, -11,  13,  -2,   1,   0,   5,  -3,   2,  10,  -5,  -2,   0,  -1,   1,   2,  -3,  -5,   2,  -2,   0,   4,   1,   1,  -1,  -2,   2,  -1,   0,   2 },
     {   24,  56, -20,   0, -13,   0,  -3,   0,  -2,   0,  -1,   0, -10,  26, -37,  -1,   9,  -1,   2,  -1,   1,   0,   0,   0,   5,  -1,   7,  11, -13,   3,  -4,   1,  -1,   1,  -1,   0,  34, -31, -23, -14,  16,  -1,   2,   0,   2,   0,   1,   0, -25, -10,   7,  44, -14,   1,  -1,   0,  21, -14,  29, -31,  -7,   7,  -3,   2, -15,   2,  -9,   3,  17,  -7,   3,  -1,   5,  -1,   7,   3, -10,   1,   1,   0,  -5,  -2,  -1,  -3,   2,   0,   3,   1,  -3,  -1,  -1,  -1,   2,  -1,   2,   0 },
     {   -3, -24,  -1,  18,  -4,   2,   0,   0,   0,   0,   0,   0,  42,  27,  21, -42,   8,  -7,   1,  -3,   0,  -2,   0,  -1,  21,  31, -53,  10,  -2,  -3,   1,  -1,   0,   0,   0,   0, -23,   4, -27,  13,  -3,   6,  -1,   2,   0,   1,   0,   0,  27, -43,   6,  24,  -5,   4,  -3,   2, -13,  -1,  27, -17,  -2,   0,   0,   0,  -1,   2, -13,   3,   8,  -2,   1,   0,  -3,  -5,   9,   4,  -5,  -1,   0,   0,  -2,   2,  -3,  -2,  -1,  -2,   3,   2,   0,   0,  -1,  -1,  -1,  -1,   2,   1 },
     {    7,  30, -38,  25,  -5,   1,  -3,   0,  -1,   0,  -1,   0, -11,  49,   5, -30,  -6,   1,  -1,  -1,   0,   0,   0,   0, -49, -26,   1,  16,  16,  -4,   3,   0,   2,   0,   1,   0,  27,  31,  -3,   5, -24,   4,  -5,   0,  -2,   1,  -1,   0,   5, -33,   3, -24,  19,   1,   1,   1,  -7,  18, -13,  23,  -2,  -7,   2,  -2,   9, -17,   5,  -4, -10,   8,  -2,   2,  -1,   1,   2,  -1,   7,  -1,   0,   0,   3,  -4,   1,  -1,  -1,   0,   1,   1,   2,  -2,   1,  -1,   0,   0,   0,   0 },
     {  -14, -18,  -7,  27, -12,   7,  -2,   3,  -1,   2,  -1,   1, -16, -18,  10, -13,  17,  -4,   3,  -1,   2,   0,   1,   0, -26, -36,   7,  31, -11,   6,   0,   3,   0,   1,   0,   1,  -5,  -9,  33,   2,   6,  -1,  -1,   0,  -1,   0,   0,   0,  13,   8,  43,  -9, -23,   0,  -4,   0,  -4,  19,  -3, -53,  24,  -2,   3,  -1,  -3,  16, -29,  29,   5, -10,   4,  -3,   6,  -3,  10,  -8, -13,   7,  -2,   1,   0,   5,  -9,  -1,   2,  -1,   1,   0,   0,   3,  -4,  -1,   1,   0,   1,   0 },
     {   -9, -41, -17,  37,  -5,   6,  -2,   3,   0,   1,   0,   1,  17,   0,  31, -25,  -3,   1,  -2,   0,  -1,   0,  -1,   0,  43, -21,  31, -23,  11,  -9,   1,  -2,   1,  -2,   0,  -1,  36,   8, -35,   4,  -5,  -2,   3,  -2,   1,  -1,   1,  -1, -49,  11,  14,  -2,  -1,   5,  -3,   1,  29,  -2, -15,  -3,   2,  -2,   0,   0, -19,  -2,   7,   2,   3,   2,   1,   0,   3,   8,  -5,   0,   2,  -2,  -1,   0,  -6,  -2,   0,   3,   1,   3,  -1,  -1,  -3,   0,   0,   1,   1,   2,  -1,   0 },
     {   13,  24,  13, -11, -12,   4,  -5,   0,  -2,   0,  -1,   0,  16,  30,  15, -15,  10, -12,   1,  -5,   0,  -2,   0,  -2,  10,  25,   8, -15, -16,   8,  -5,   1,  -2,   1,  -1,   0,   9,  24,   0, -24,  14, -13,   3,  -4,   1,  -2,   1,  -1,   1,  13, -33, -21, -13,   8,   0,   1,  -6, -17, -30,   1,  47,  -6,   3,   0,  -6,  10, -12,  49, -26, -10,   3,  -4,   5, -19,   9, -16,  -6,  16,  -4,   4,  -3,   2,  -4,   9,   1,  -7,   3,  -2,  -2,   0,  -2,   4,   0,  -3,   1,  -1 },
     {   -9,  -2, -13,  -9,  12,   2,   1,   1,   1,   1,   0,   0,   1,  33,  10,   5, -13,  -5,   1,  -2,   0,  -1,   0,  -1, -48, -14, -14, -27,  21,   6,   3,   1,   2,   1,   1,   1, -40,   1,  -9,  48,   5,  -9,   4,  -3,   2,  -1,   1,   0,  -9,  32,   2,   6, -27,  14,  -6,   2,  37, -38,  17,  -4,   9,  -3,  -4,   2, -14,  18,   6,  -7, -10,  -2,   4,  -2,  11,  -6,  -6,  -4,   5,   5,  -1,   1,  -1,   2,   5,   1,   5,  -2,  -2,  -3,  -1,   2,   2,   2,   3,  -2,  -1,  -1 },
     {  -13, -29, -37,  12,   1,  11,  -1,   5,   0,   2,   0,   1,  -9, -27, -27,  20,  13, -10,   7,  -1,   3,   0,   1,   0,   4, -11, -13,  12,  -6,  12,  -2,   4,  -1,   1,  -1,   1,  15,  -9,  -7,   4,  11,  -9,   4,  -3,   1,   0,   1,  -1,  -4, -24, -14,  20,   0,  13,  -1,   2,  -6,  -1,  29,  35,  14, -20,   2,  -4,   2,  12,  14,  11, -56,   9,  -2,  -1,   5, -14,  20, -31,  14,  14,  -6,   5,  -2,  10,   1,   6,   3,  -3,   6,  -9,  -1,   5,   1,   1,   1,  -1,   3,  -4 },
     {   10,  26,  61, -32,  -1,  -5,  -1,  -3,  -2,  -1,  -1,  -1, -11, -20,  35, -31,  15,  -9,   4,  -4,   4,  -2,   1,  -1,  -2, -39, -29,  21,  -3,  11,  -1,   3,   0,   2,   0,   1,  -9,   5, -15,   1,  14,  -3,   2,   1,   1,   0,   0,   0, -32, -13,  14,  18,   3,   2,   3,   0,  21,  23,  -4,  12,  -6,  -8,  -1,  -3,  -5,  -9,  12,  -4, -22,   6,   0,   0,   6,   8,  -3, -16,  11,   4,  -3,   1,  -3,  -1,  -4,   7,   3,   2,   0,  -3,  -1,  -1,  -2,   2,   1,   2,   1,  -1 },
     {    1,  -2,  -7, -33,  45, -10,   4,  -3,   1,  -1,   1,   0,   7,   0,  41,  29, -55,   7,  -4,   0,  -3,   0,  -1,   0,   5, -19, -18, -10,  18,  10,  -3,   2,  -1,   3,   0,   1,  42,   5,  20, -10,  -1, -11,   1,  -1,  -1,  -1,  -1,   0, -11, -26, -13,   4,  -8,  12,   1,   3,  -8,  12,  25, -21,  16,  -6,  -2,   0,   3,   4, -26,   5,   6,  -5,   5,  -2,  -5,   1,   8,   6,  -3,   3,  -2,  -1,   2,   2,  -8,  -5,  -3,   0,   1,   5,   1,   1,  -3,  -1,  -2,  -1,   1,   2 },
     {   -6,  -7, -18,   7,  -3,   8,  -2,   2,   0,   1,  -1,   1,   7,  27,  26,   0,  -3,  -8,  -1,  -2,  -1,  -1,   0,  -1, -19,   2,  18, -30,  -6,   6,   1,  -1,   1,   0,   0,   0, -25, -33,   1, -16,  37,  -8,   5,   0,   2,   0,   1,   0, -18,  17,  -6,  38, -10,  -7,   6,  -5, -27,  41,  -8,   2, -13,   7,  -7,   2,  44, -33,   3,   4,  -5,   1,  -3,   3, -17,  10,   9, -13,   5,   0,   2,  -1,  10,   2, -12,   6,  -2,  -3,   3,  -1,   5,   1,  -2,   1,  -2,  -1,   0,  -1 },
     {    8,   9,   9,   7, -15,   7,  -5,   0,  -3,   1,  -1,   0,  12,  14,  16,  17,   1, -16,   1,  -3,  -1,  -2,  -1,  -1,  14,  19,  19,   8, -21,   9,  -5,   0,  -2,   0,  -1,   0,  18,  16,  21,   5, -10, -11,   0,  -3,  -2,  -1,  -1,  -1,  18,   7,  11, -10, -30,   9,  -6,   0,   7,   1,  13, -27, -24,  -9,   2,  -2,  -1, -12, -12, -50, -19,  35,   2,   6,  -2, -11,  -8, -21,  55,   3,  -4,   3,  -4,   7, -11,  15,   0,  -5,  -2,  -6,  -2,   2,  -4,   4,   0,  -2,  -2,  -2 },
     {    7,  10,  41,  11, -32,  15,  -8,   0,  -4,   1,  -1,   0,  -3,  -4,  17, -17,   5,  -2,   0,   0,   0,   0,  -1,   0,  -6, -23,  31, -38,  18,  -5,   4,  -1,   1,   1,   0,   0,  16, -49,  -6,  15, -14,  11,   3,  -1,   2,   0,   1,   0,  21,  -8, -17,  -8,  19,   8,  -5,   3, -26,  -8,  38,  10,   8, -11,   1,  -3,  11,  33, -15, -12,  -5,  -4,   1,  -3,  -7,  -8,  -1,   5,   1,   1,   0,   2,   1,  11,  -2,  -7,  -4,  -2,  -6,   6,   1,   3,   1,  -1,  -1,  -1,  -2,   2 },
     {   -3,   5,  -4,   3,  12, -13,   0,   0,  -1,   1,   0,   0,   0,  14,   2,   2, -19,  10,   1,  -3,   2,  -1,   0,   0,   0,  29,  29,   4,   4,  -8,  -3,  -2,   0,   0,   0,  -1, -50, -17,  10, -27,  -9,  13,   5,   2,   2,   1,   1,   1, -34, -19,  17,  -3,  38,   1,  -4,   2,  13,  23,   2,   4,  -8, -23,  11,  -6, -24,  25, -31,   3, -15,  18,  -1,   1,  32, -21,   9,   3,   7,   1,  -5,   5,  -9,  11,  -4,  -5,  10,  -3,  -4,   4,  -3,   0,   0,  -2,   3,  -2,  -3,   2 },
     {    5,   9,  13,  11,   6, -19,  -3,   1,  -3,   0,  -1,   0,  -8, -19, -34, -29, -13,  20,   4,   1,   3,   2,   2,   1,   8,   6, -16,  -2,  48, -23,   1,   0,   1,  -2,   0,   0,   6,   3,  -1,  15, -23,  -3,  10,  -5,   3,  -2,   1,  -1,   4,  41, -19,  18,  -7,  13, -13,   2,   3,  12,  -7, -15,  -4, -16,   9,   1,  16, -28, -22,  21,  -5,  25,  -4,   0,  -5, -11,  32,  -6,  13, -11,  -6,   1,  -4,   6, -16,   7,   3,  -7,  12,   3,  -1,   3,   0,  -4,   1,  -1,   2,   1 },
     {   -9, -12, -19, -29,  -6,  21,  -6,   5,   0,   0,   1,   1, -10, -13, -15, -22,   7,  -9,  12,   0,   3,   2,   1,   1, -11, -14, -13, -20,  -7,  26,  -4,   3,   2,   1,   0,   2, -11, -13, -12, -21,   8,   3,   5,   2,   2,   2,   1,   1, -10, -18, -10, -21,  10,  20,   1,   3, -11, -16, -12, -13,  21,   4,   7,   4, -10,  -9,  -7,  -2,  39,  28,  -3,   4,   1,   6,   6,  25,  50, -26,  -5,  -2,  -4,  10,  -3,  30,   5,  -1,   6,   4,  -2,   4,   1,   9,   1,   0,   3,   3 },
     {    4,  10,  30,  58, -10, -14, -10,  -2,  -3,   0,  -3,   0,  -6,  -4,   1,  26, -30,   6,   8,  -5,   1,  -1,   0,  -1,  -7,   7,  -3,  18,  -4,  -1,  -6,   1,  -1,   1,  -1,   0, -25, -10,   3,  -7, -19,   4,   9,   3,   0,   1,   0,   2, -35, -28, -17,  -6, -16,  32,  -3,   3, -14, -23,  -1, -16,  29,   5,  -1,   7,   9, -16,  35,   1,   8,  -3,  -1,  -2,  -4,  20,  -4, -11,   4,  -4,  -4,   0,   8,   0,   0,   1,   1,   5,  -1,  -8,   3,   0,  -5,   0,  -1,   1,   0,  -3 },
     {    0,  -5,  -7,  -3,  25, -40,  10,   0,  -1,   1,   1,   1,   4,   1,  12, -30, -18,  45,  -7,  -6,   1,  -2,   1,   0,  -5,  -8,  24,  11,  -4, -10,  -6,   6,  -1,   2,  -2,   0,   7, -30,  -2, -25,  26,  12,   8,   2,  -2,   0,  -1,   1,  54,  -5,   8,  18,  -6,   2,  -8,  -2,   8, -27, -14,   1,   9,  -5,   6,   2, -10,   3,  25,   0, -14,   6,  -5,   1,  -1,   8, -18, -11,  19,  -1,   5,   2,  -1,  -2,   7,  16,  -3,   4,  -2,  -8,  -4,   3,   0,   8,   2,   2,   0,  -4 },
     {   -7,  -3, -18, -45,   8,   1,   1,   3,   2,  -1,   2,   1,  -3,   1, -12, -45,  16,   6,   9,  -4,   4,  -1,   1,  -1,  13,  28,  37,  25,  15,  -6,  -8,  -3,  -3,  -4,  -1,  -2,  -9,   4,  22,  16, -17, -10,  -3,   2,  -4,  -1,  -1,   1, -32,  -6,   8,  -7, -13,  15,  -3,   2,  -9,  -6,  30,  -1,  21, -10,  -6,   5,  30,   6,   6,  -8,  -7, -12,   9,  -3, -22,  27, -18,  -2,  -2,   6,   2,  -2,  17,  -7,  -5,   3,  -5,   6,  -4,  -1,   5,  -1,  -3,   3,  -3,   0,  -1,   0 },
     {   -1,   0,   1,  11, -12,  -1,  -4,   9,  -2,   1,   2,   0,  -9, -12, -29, -28,  -8,  17, -16,   6,   4,   3,  -2,   3,  -1,   5, -22, -40,   8, -12,   4,   2,   2,  -1,   4,   0,   0,   6,  22, -19,   5, -18,   0,  -5,   6,  -1,   2,  -2,  -6, -19,  -2,   5, -36,  14,  11, -10,  -6,  32, -13,   2,   4,  10, -25,  10, -10,   7,  -7, -27,  -9, -20,  19,  -1,  17, -13, -30,   8,   8,  28,   1,  -4, -12,   6,  27,  -3,  11,  10, -12,  -1,  -3,   0,  14,   4,   1,   6,  -3,  -6 },
     {    4,   5,   6,   7,  16,  24, -21,  -4,  -3,  -1,  -4,   0,   4,   7,   6,   5,  15, -14,  21,   0,  -3,   1,   0,  -2,   7,  12,  22,  10,  38,  19, -23,  -1,  -3,  -2,  -3,  -1,   5,  -5,  27,  18,   9, -44,  11,  -2,  -3,   0,  -2,   0,  25, -14,  -4,  26,   0,  13,  -6,  -6,   1,  19, -19,  11,   6, -10,  -4,   4, -24,   1,  28,   3,  19,  -3,  -4,   0,  24,  10,  -3,   9,  -9, -32,  -1,  -2, -28,   8,  -5,  -8,   7,  -5,  -3,  -1,  -8,   0,  -3,  -6,   3,   2,  -2,   2 },
     {    1,   5,   5,  16,  24, -25,   0,  -6,  -1,  -3,   2,  -2,  -3,  -3,  -4, -13, -13,  -3,   1,   8,  -2,   2,   0,   2,  -1,  -4,   1,  21, -21,  23, -19,   0,   0,   0,   0,   1,   3,   2, -36,   1,   9, -48,  20,  -4,   4,   1,   0,   0,  -9,  21,  24, -11, -10,  19,  -3,  -6, -38, -10, -11,  24, -19,   0,  -5,   7,  19,  37, -14, -11,   9,  11,   6,  -5, -11, -23,   3,  20,  -5,  -4, -12,   0,   9,   7,   3, -12,   2, -11,   0,   0,   5,   0,   6,  -2,  -4,  -1,  -1,   2 },
     {    3,   1,  -1,  18,  -6,   2,   6, -16,   7,  -4,   1,  -1,   1,  -4, -10,   3, -10, -13,   5,   6,  -4,   2,  -1,   1,   2,   5,  -6,  -4, -12,  23, -11,  -1,  -1,   1,  -2,   0,  -8,  -9,  -6, -13,  -1, -13,  14,   6,   4,   0,   1,   1,  28,  10,   2,  -2,  26,  31,  -2,  -2,  61,  29,   7,  10,  27, -25,  -8,  -2,   1, -15, -19,  -9,  -3, -12,  13,  -6, -36,  11, -13,  21,  -4,  10,  -3,   5,  18, -25,   0,   6, -17,   5,   6,   4,   8,  -6,  -5,  -2,  -6,  -4,   6,   3 },
     {   -2,  -8,  -3, -17,   4,  -2,   6,   3,   0,   1,  -1,  -1,   1,  -6,  11,  -6,  -3, -19,  -6,   7,   6,   0,  -1,   0,   1,  -8,  31,  26,  -7,  -2,  11,  -2,  -2,  -2,   0,   0, -11, -38,  -7,  20, -23, -12,   0,   1,   0,  -1,   2,   0,  11, -27, -32,  -8, -23,  12,  18,  -4,  19, -10, -26,   4, -22,  23, -12,   3,  -9, -33, -18,  11,  -3,   9,  14,   3,   3,  -5,  30,  15,  -2,  33, -12,  -3,  -2,   3,  22, -20,  -4,   5,   0,  -1,   5,   2,   5,  -2,  -3,   2,   3,  -5 },
     {    0,  -1,   1,  -3, -17, -11,  22,   2,   5,   2,   2,  -2,  -1,  -5,  -4, -19, -15,  28, -15,   6,   6,   3,   1,   2,   1,   3,   5, -25, -45,  19, -14,  12,   1,   4,   1,   2,  -2,   3,  18,  17,  -6, -17,  -4,  -7,   0,   2,   0,  -1,   0, -16,   2, -15,  -2,  -3, -18,   6,  29,   0,   9,  21,   0,   5, -10,   3,  18,  -8,   1,  -7,  20, -17, -14,   0,  15,   4,  39, -27, -10, -22,  17,  -4, -20,  15, -24, -15,   2,  -4,  13,  -6,  -8,  11,  -3, -14,   4,   2,   0,   3 },
     {   -2,  -3,  -7, -23,   1,  12, -31,  17,  -8,   5,   0,   3,  -2,  -3,  -3, -16, -39, -46,  17,   7,  10,  10,  -2,   3,  -2,  -5,  -4, -10, -17, -14,  -8,  -3,   3,  -1,   6,   1,  -2,  -2,  -7, -34, -49,   3,  13,  -5,  10,   3,   2,  -1,  18,  12,  17,   9,  10,   7,  -7, -14,   8,  -1,   0, -15, -16,  11,  -8,  -3,  12,   9,  26,  14, -12,  -7,  -5, -10,  -3,   2,  -9, -22, -12, -12,   1,   0,  -6,  13,   9,  -8,   5,  -3,  -1, -12,  -8,  12,   4,  -2,   2,   5,  -6,  -1 },
   },
   {
     {  100, -56,  -3,  -3,  -1,  -1,   0,   0,   0,   0,   0,   0, -48,  12,  13,   0,   2,   0,   1,   0,   0,   0,   0,   0, -11,  18,  -3,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,  -5,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -43,  -6,  17,   1,   4,   0,   1,   0,   1,   0,   0,   0, -79,  60,   2,  -2,   0,  -1,   0,  -1,   0,   0,   0,   0,  51,  -1, -25,   3,  -3,   1,  -1,   0,  -1,   0,   0,   0,  15, -24,   3,   7,  -1,   1,   0,   1,   0,   0,   0,   0,   5,  -4,   7,  -2,  -1,   0,   0,   0,   2,  -2,   1,  -2,   1,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   1,  -1,   0,   0 },
     {   10,  77, -53,  -3,  -7,  -1,  -2,   0,  -2,   0,  -1,   0, -54, -30,  23,  12,   5,   3,   2,   1,   1,   1,   1,   0, -16,  18,  16,  -5,  -2,  -2,   0,  -1,   0,   0,   0,   0,  36,  -1, -19,  -5,  -1,   1,   0,   0,   0,   0,   0,   0,   2, -15,   1,   6,   2,   1,   0,   0,   4,  -3,   4,   1,  -2,  -1,   0,   0,   0,  -3,   3,  -1,   0,   0,   0,   0,   3,   0,   0,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 },
     {   46,  39, -30, -10,  -4,  -2,  -2,  -1,  -1,  -1,  -1,  -1,   6,  21, -35,   9,  -2,   2,  -1,   1,  -1,   0,   0,   0,  43, -74,  11,  17,  -4,   4,  -1,   2,  -1,   1,   0,   1, -17, -12,  35,  -7,  -2,   0,   0,   0,   0,   0,   0,   0,  -8,  10,   3, -12,   3,  -1,   0,   0,  -7,   4,  -3,  -1,   4,  -1,   1,   0,  -2,   3,  -1,   1,   1,  -1,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   1,   0 },
     {   -7,   6,  22, -12,   2,  -4,   0,  -1,   0,  -1,   0,   0, -19,  46, -49,   4,   1,   2,   1,   1,   0,   0,   0,   0, -68, -18,  28,  22,   0,   2,   0,   1,   0,   0,   0,   0,  19,  38,   4, -18,  -8,  -3,  -2,  -2,  -1,   0,   0,  -1,  27, -16, -29,  -1,   4,   3,   2,   1,   2, -15,   8,  11,   1,   0,  -1,   0,   5,  -3,   5,  -1,  -4,  -1,  -1,   0,   2,  -2,   2,  -2,   0,   1,   0,   0,   2,  -1,   0,  -1,   1,  -1,   1,   0,   1,  -1,   0,   0,   1,  -1,   1,   0 },
     {  -14, -14,  25, -12,   4,  -1,   1,   0,   1,   0,   0,   0, -38, -43,  24,  14,   3,   3,   1,   1,   1,   1,   0,   1,  38, -27,  43, -12,  -4,  -2,  -2,  -1,  -1,  -1,   0,   0, -30,  59,  -6, -30,   8,  -4,   2,  -2,   1,  -1,   0,  -1,  -2,  16, -32,   5,   8,  -2,   1,   0,  -1,   4,  -7,  10,  -2,  -2,   0,   0,   6,   0,  -4,   3,  -4,   1,   0,   0,   1,   1,  -1,   0,  -1,   1,   0,   0,   2,   0,  -2,   1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   1,  -1,   0 },
     {  -21, -25, -70,  59,  -1,  11,   0,   5,   0,   2,   0,   2,  -2,  34,  20,  -1, -16,  -1,  -4,   0,  -2,   0,  -1,   0, -17,  -2,  29, -28,   0,   3,   1,   0,   0,   0,   0,   0, -32,  -1,  15,  -6,  11,   0,   2,   0,   1,   0,   0,   0,  26,  18, -11,  -8,   3,  -6,  -1,  -2,   4, -11, -10,   5,   2,   0,   1,   0,   3,  -2,   4,   2,  -2,   0,   0,   0,   1,  -1,   1,   0,   0,   1,   0,  -1,   2,   1,   0,  -1,   0,  -1,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0 },
     {    3, -21, -10,  15,  -3,   3,   0,   1,   0,   1,   0,   1,  22,  12, -36,  23, -11,   4,  -2,   2,  -1,   1,  -1,   0,  38,   3,  40, -29, -10,  -2,  -3,  -1,  -1,  -1,  -1,   0,  29,  -8, -45, -20,  16,   2,   3,   1,   1,   0,   1,   0,  -6, -47,   5,  37,   9,  -1,  -1,   1,  -6,   7,  33,  -5, -13,  -4,  -1,  -1,  -6,   3,  -2, -13,   1,   3,   2,   0,  -6,   3,   1,   0,   4,   0,   0,  -1,  -3,   1,  -1,   1,  -2,   2,   2,   0,  -2,   0,   0,   0,  -2,   0,   1,   0 },
     {  -16, -33,   6,  20,  -6,   6,  -2,   3,  -1,   1,   0,   1, -10, -27,  -1,  12,   1,   3,   1,   1,   1,   1,   0,   0, -18, -13,  31,  10,  -7,   4,  -3,   1,  -1,   0,  -1,   0,  66, -34,  41, -27,  -8,   0,  -3,   0,  -1,   0,  -1,   0, -34,  19,  10, -28,  15,  -2,   3,  -1,  -7,  24, -12,  -1,   9,  -6,   2,  -1,  -5,  12, -12,   2,  -2,  -3,   2,  -1,   3,   2,  -6,   3,  -1,   1,   1,   0,  -1,   2,  -2,   1,   1,   2,  -2,   0,  -1,   2,  -1,   0,   1,   1,  -1,   0 },
     {    1,  13,  10,  -3,  -9,   1,  -2,   0,  -1,   0,   0,  -1, -11, -11, -54,  47,  -2,   5,   0,   3,  -1,   1,   0,   1,   4,  64,  -2,  -7, -20,   1,  -5,   0,  -2,   0,  -1,   0, -43, -20,  30, -28,  11,   6,   1,   1,   1,   1,   0,   0,  -5,  12,   1,  -5,  16,  -6,  -1,  -1,  26,  -5, -13,   4,  -1,  -5,   2,   0,   4, -16,   1,   5,  -2,   2,   1,   0,   1,  -3,   7,  -2,  -1,   1,  -1,  -1,  -1,  -1,   2,  -2,   1,   0,   0,  -1,   0,  -1,   0,   0,   1,  -1,   0,   0 },
     {   10,   6,  -8,  21, -14,   6,  -6,   1,  -3,   1,  -1,   0,  20,   8,   9, -27,   2,  -2,   2,  -1,   0,  -1,   0,  -1,  28,  20, -39,  13,  -1,  -2,  -1,   0,  -1,   0,   0,   0,  27,  15,  13, -35,  -4,  -3,   0,  -2,   0,  -1,   0,  -1, -27,  -3, -56,  -1,  26,   1,   7,   0,   0, -29,  10,  50,  -1,  -4,  -2,  -1,   6,   1,  18,  -4, -19,  -1,  -1,   0,  -1,  -8,   2,  -6,   0,   6,   0,   1,  -3,   1,   3,  -1,  -1,  -3,   1,   1,  -2,   1,   1,  -1,   0,  -2,   0,   1 },
     {    4,   3,  16,  70, -47,   6, -12,   2,  -5,   1,  -2,   1, -21, -32, -30, -18,   6,  16,   4,   5,   2,   2,   1,   1,   8,  -3, -15,   7,  29,  -6,   2,  -2,   0,  -1,   0,  -1, -11,  28,  21,  11, -13, -12,  -2,  -2,  -1,  -1,   0,   0,  11, -31,  17,  -7, -13,   9,   2,   1,   2,   1,  21, -19,   4,   4,  -2,  -1,   1,   5,  -4,  -8,  10,  -4,  -1,  -1,  -3,   4,  -1,   1,   0,  -4,   1,   0,   0,   1,  -2,  -1,   0,   1,   1,   0,   0,   0,  -1,  -1,   0,   1,   1,   0 },
     {   11,  31,  30,  16, -23,  -4,  -8,  -2,  -2,  -1,  -2,  -1,   0,  31, -10, -33,  -1,   3,   0,   1,   0,   0,   0,   0, -10,   1,  19, -51,  28,  -2,   3,  -2,   0,  -1,   1,   0,   5,   1, -20,   7,  20, -16,   1,  -1,   2,  -1,   1,  -1, -49,  29,  -6,   9,  -4,  -7,   6,  -1,   9,  19, -33,   2,   0,   3,   3,  -2,   3, -14, -17,  20,   1,   3,  -1,   0,   6,  -9,   4,  10,  -7,   0,  -2,   1,   2,  -6,   3,   1,   2,  -3,   1,  -1,   0,  -2,   0,   1,   2,  -1,   0,   1 },
     {  -14, -11, -30,   9,  11,   2,   1,   2,   2,   0,   1,   0, -11,   9,  -5,  -9,  19,  -7,   5,  -2,   1,  -2,   1,   0, -19,  -7, -11,  43, -17,  -2,  -1,   2,  -1,   0,   0,   0, -23,  23, -11,  -7,  -6,   9,  -1,   2,   0,   1,   0,   0, -68,  12,  19,  30,   4,   1,  -2,  -1,  32,  19,  22, -13, -25,  -3,  -4,   0,  15,  -7, -16, -22,   6,   5,   3,   2,   5,  -7,   4,   6,   5,   0,  -1,  -1,   2,   0,   3,  -3,   2,  -2,   1,   0,   2,   0,   0,  -1,   2,  -1,   0,   0 },
     {    1,   7,  23,   6,  -2,  -5,  -3,  -1,  -1,  -1,  -2,   0, -13, -20,   3, -30,  19,  -6,   9,  -1,   3,   0,   1,   0, -16, -17,  13,  -3,   2,   5,   4,   0,   1,   1,   0,   0, -17, -60,  20,   3,  20,   2,   1,  -2,   1,   0,   1,   1,  23,  17, -32,  47,   4, -12,   2,  -1,  10,  -2,  32,   8, -35,   1,   0,   0, -20,  12,   6, -27,  -3,  11,   0,   1,   2,   8,  -3,  -2,   8,   1,  -3,  -1,   0,   6,  -4,  -1,   2,   1,  -1,   1,   0,   2,  -1,   0,   1,   0,   0,   1 },
     {   -6,  -9,  -3,   2,   8,  -5,   5,  -3,   1,  -1,  -1,   0, -12, -16,  -9,  54, -37,  14,  -8,   3,  -2,   1,  -1,   1, -22, -33, -44, -29,  28,  17,   3,   7,   2,   2,   1,   1,  10,   6,   4,  19,  23, -16,   0,  -4,  -1,  -1,   0,  -1, -24,  14,   7,  15, -14, -11,   5,  -1,  -2, -17,  16,  20,  -9,   7,   2,  -1,  18,   6,  23, -10, -18,   6,  -6,  -1,   2,  -3,  -2, -14,   6,   5,  -2,   3,   2,   3,   6,  -1,   0,   1,  -1,  -3,   0,   2,   2,   0,   0,   2,  -1,  -2 },
     {    3,   7,  17, -18, -64,  38,  -2,   7,  -4,   2,  -2,   1,  12,  22,  42,  24, -16,   2, -14,  -1,  -5,   0,  -2,   0,  -9, -11, -11,  19,  15, -14,  -5,   4,  -1,   1,   0,   1,   0,  -3,  -9, -47,  24,  -1,   9,   1,   2,   0,   1,   0,  13,  12,  21,   2,   7,  -6,  -2,  -3,  12, -17,   9, -19,  -9,   2,   1,   0,  -2,  -3, -14, -10,  15,   1,   2,   1,  -2,   4,  -3,   7,   7,  -5,   0,  -2,  -4,   6,  -5,   2,   0,   0,  -4,   2,   0,   0,  -1,   1,   0,  -1,  -1,   1 },
     {   -5,   4,   0, -19,  -4,   8,   1,   1,   2,   2,   0,   0,  -7,   8,  15,  -5, -12,   7,  -6,   4,  -3,   0,  -1,   0, -17,   0,   3,  -6,  34, -18,   3,  -1,   1,   0,   1,   0, -40, -18,   8,  -5, -11,   1,   6,   1,   1,   1,   1,   0, -11, -48,  12,  -9,  32,  10,   2,   0, -22,  50,   8,  37,   9, -27,  -2,  -7,  15,  15,  -9,  -9, -35,  -1,   3,   3,   0,  -1,  -7,  -4,   4,  14,   3,   0,   7,   0,   0,   4,   1,   0,  -1,   1,   3,   1,  -1,   0,   0,   1,  -1,   0 },
     {   11,  23,  20,  16,  26, -28,  -6,  -6,  -1,  -2,  -1,  -1,   5,  14,  27,  -2, -28,  -9,   4,  -3,  -1,  -1,   0,  -1,  -9,   6, -21, -26, -35,  37,   3,   4,  -1,   2,  -1,   1,  -1,  22,  39, -27,  17,   6, -11,  -3,  -2,  -1,  -1,   0,  -1, -28,   1,   8,   5, -12,  -4,   2, -13,  13,   8, -27,   4,  -1,   6,   1,   3,  22, -30, -12,  17,   1,   5,  -3,  -1,  -5, -15,  14,   6,  -5,   1,   0,   1,  -2,   1,   7,   0,  -4,  -1,   3,   1,  -1,  -1,   0,   0,  -1,  -1,   0 },
     {    4,   7,   2, -17,  18,  -3,   3,  -5,  -2,   1,   1,   0,   7,  12,   9,  -5, -42,  17,  -5,  -1,  -1,   1,  -1,   0,  17,  33,  41,   9,  27, -14, -13,  -2,  -3,  -1,  -1,  -1,   4,   3,  26,  16, -22, -24,   1,  -3,   1,  -2,   1,  -1, -15,   7, -12, -19, -29,  10,  12,   1,  22,  -9,  16,  -6, -15,  18,  -6,  -1,   5,   4,  -2, -43,   9,   8,  -7,   3, -12,  -1, -22,   0,  28,  -4,   2,   2,  -2,   2,  -6,   6,   2,  -1,  -6,   2,   0,  -3,   0,   0,   0,  -2,  -1,   2 },
     {    3,  13,   7,  16, -11,   4,  -3,   0,  -1,   1,   0,   0,  -3,   5, -12,   7,  -5,  -1,   4,  -1,   1,   0,   1,  -1,   0,  20,   6,  18, -17,  -4,  -2,  -5,   0,  -1,   0,   0,  -9,  -1, -29,   9,   0,   9,   1,   1,   2,   0,   0,   0,   1,  50,  -4, -17,  -2,  -7,  -1,  -1, -82,   1,  31,   0,  -7,   1,   1,   1,  24,  20, -10, -13,  -2,   5,  -1,   0,  23, -13, -19,   1,   2,   3,   0,   1,   3, -14,   3,   6,   0,  -4,   8,  -1,   1,   0,   0,  -4,   0,  -1,  -1,  -1 },
     {   13,  16,  36,  43,   6, -19,  -3,  -7,  -3,  -2,  -2,  -2,   2,   0,  19,   2, -43,   8,   0,   3,   1,   1,  -1,   1,  -1, -15,  15,  26, -26,  18,  -1,   1,  -3,   1,   0,   1,  -6, -12, -39,  15,  -8,  19,  -6,   0,   0,   1,   0,   1,   3,   5,  15, -26,  30,   3, -11,   3,  32,  -3,  -2,  12,   2, -16,  -3,   1,  -2, -27,  22, -10, -17,   2,   4,   4,  -5,   9,   4, -15,   9,   3,   1,  -2,  -1,   3,  -3,  -4,  -1,   2,  -3,  -3,  -1,  -2,  -1,  -1,  -1,   1,  -1,  -1 },
     {   -1,   4,  12,  -5, -41,   1,   7,   4,   0,   2,  -1,   1,   5,  15,  16,  27,  45, -32,   6,  -7,  -1,  -3,   0,  -2,  -6,  -7,  15, -40, -25,  11,   8,  -3,   4,  -1,   0,   0,  -8,   0,   8,  25, -32,  22,  -8,   2,  -3,   2,   0,   0, -27,  -8,  -5,  -9,  -6,  17, -12,   3,  11, -21,  11,  -1,  16,   5,  -7,   8,  11,  20,  19, -17,  -6, -10,   2,  -1, -11,  -2, -20, -17,  14,   2,   3,   0,  -3,   4,   0,  12,   4,   3,  -4,  -1,   1,   1,  -2,   3,   1,   0,  -2,  -1 },
     {    5,   1,   2,   2, -16,  24, -21,   5,  -1,   2,   0,   1,   9,   6,   8,  12,  10, -16,  -5,   1,  -3,   0,  -1,   0,  10,   7,   1,   8, -19,   6,   6,  -7,   0,  -1,   0,  -1,  22,  21,  23,  24,  -7, -11,  -8,   2,  -1,  -1,  -1,  -1,  26,  -3,   3,  17,  -8, -30,   2,   0,  -8,  39, -32,  -1, -40,   0,  10,  -2,  18, -31,   2, -15, -33,  29,   3,   2, -14, -10,   5, -17,  17,  21,  -6,   3,   7,  -7,   3,  -4,  -4,   1,   0,  -5,   0,  -4,   1,   0,  -2,   0,   0,  -3 },
     {    2,  -1,  -7,   2,   3, -27,  17,  -5,   9,  -2,   3,  -2,   3,  -2, -12,  12,  11,  10, -11,   2,  -1,  -1,  -1,   1,   2,  -7, -27, -34,  27, -18,   7,  -2,   3,   1,   1,   1,  16,  10,  -6,  -8, -49,  26,   1,   5,   2,   1,  -1,   1,  25,  31,  -6,   0,  37,   4, -15,  -1,   4,  10,  -5,  -2, -17, -16,  -6,   2,  -8, -17, -21, -31,  15,   9,   7,   6, -10,   0,  -1,  32,  19,  -7,   0,  -5,   4,  -1,   7,   4,  -4,  -1,  -1,   3,   1,  -3,   2,  -1,  -1,   0,   0,   3 },
     {    2,   3,   2,  20, -14, -21,  21,  -4,  -3,  -1,   1,   0,   8,  11,  18,  48,  -5, -13,   8, -17,   3,  -4,   1,  -2,   3,   5,  -5,  25,  21,  14, -26,   2,  -1,   1,  -2,   0,  -2,  -8,  -2,   3, -17,  -8,  -1,   4,   0,   0,   2,   1,   2,   3, -46,  23, -25,  25,   1,   7,   1,  43,   6, -10,  14,  -3, -10,  -5, -31,  -5,  -9,  14,   0,  -8,   0,   4,  15, -13,  16,  -7, -13,   5,   1,   2,  11,  -6,   5, -10,   0,  -7,   4,  -6,  -3,   2,   0,  -3,   1,   0,   0,  -2 },
     {   -4,  -6, -11, -20, -15,  48,  -7,   6,  -2,   2,  -3,   3,  -8, -11, -15,  -9, -26,  10,  -5,   3,   0,   5,   1,   2,  -6,  -8, -19, -14, -32,   8,   7,   1,   4,   0,   0,   0,   1,   0,  -9,  34,  15,  -1,   3,   2,  -1,   0,  -2,   0,  -3,  -1, -52, -28,  21,   5,   4,   1,  15,  38,  23, -21,  12,   5,  -6,   1, -12, -13,   6,  -1,  14,  -8,   4,  -3,  -7, -10,  -6,  22,   7, -12,   2,  -1,  10,   1,   2,   3,  -2,   0,   1,   2,   1,  -1,   4,   1,   0,  -2,   0,  -1 },
     {    0,   5,  -5,   2,  13,   8,  -7,   3,  -3,   1,  -1,   1,  -1,   4, -14,   2,  10, -23,  16,  -2,   3,   1,   3,  -1,   3,   7, -11,  -2,   6,  -6,   0,   1,  -3,   2,   0,   1,   4,  25, -13,  -1,  25, -23,  -3,   0,   3,  -1,   1,  -1,   6,  22,  17, -20,  -2,   2,  -1,   5,  16,   2,   1,  -4,   6, -33,   3,  -5, -70,  33,  -4, -19, -34,  -3,  18,  -1,  -5,   0, -15,  -7,  13,  28,   0,  -5,   3,  -2, -13,  20,   8,  -9,   6,   4,  -3,  -3,   3,   2,   4,  -3,   2,  -4 },
     {   -8, -11, -15, -20, -36, -46,  23,   9,   6,   7,   8,   1,  -1,   0,  -1, -19,  16,  22,   4,   3,   7,   5,  -3,   1,  -3,  -1,  -4,  -5, -11,  41, -50,  -1,  -5,   0,  -4,   3,  -8,  -1, -11, -10,  -2, -37,   0,  24,   0,   3,  -2,   0,   0,  -5,   8, -18,   6, -29,  22,   2,  -9,   4,   2,  -4,  -1,   0,   4,  -9,  -3, -18,  18, -12,  -5,   1,   4,  -9,   2,   7,  -4,   5,  13,  -2,   2,   0,   4,  -8, -12,   4,   5,   6,   2,   7,   1,   2,   2,   4,  -3,   1,   1,   1 },
     {    2,  -4,  -1,  -9,   0,  -8, -16,  12,  -1,   3,  -2,   1,   3,  -8,   1, -11,   3,  13,  -8,   5,  -4,   0,   0,   0,   3,  -5,   3,  -9, -13,  10,  -6,   8,  -3,   3,  -2,   1,   9, -10,   3,   5,  -1, -20,  14,   2,   3,   1,  -1,   1,  18,  14,   0,   9,  -8,  27,  -1,  -7,  15, -22,  24, -13,  30, -27, -14,  -4,  38, -14, -50,   3, -33,  -8,  15,   9,   6, -35,  25,   1,  13,  29,   0,  -1, -16,  15,   5,   1,   1,   5,   3,  -3,  -6,   4,  -6,  -3,  -1,  -3,   3,   4 },
     {    1,   2,   3,  -3,  -8,  -1,  -3,  19,  -2,  26,  15,   5,   3,   6,   5,   3,  -5,  32, -11,  18, -12,  11,  -9,  -1,   0,   2,   3,  -1, -17, -15, -18,   7, -23,   7,  -9,  14,   3,  10,  10,  23, -17,  12,   5, -37,  27,   0,   3,  -4,  -2,   3,   2,  24,   9, -28,   0,   0,   4,  -6, -12,  -2,  -8, -22, -14, -10, -12,  31,   7,   7,   7, -32, -13,  -9,   7, -11,  -2,  -1,  -5,  11,  30,  -8,  13,   0,   0, -28,   3,  -4,  16,  11,  -4, -14,  -1, -12,  -1,   4,  -2,   8 },
     {    0,  -1,  -1,  -5,  -6,  23,  -8, -11,   8,   2,   7,  -1,  -6,  -9, -13, -16, -34, -29,  27,  21,  18,   8,   1,   2,  -1,  -3,  -4,  -7,  -5,  17, -11,   7,   0, -10,  -5,  -1,  -1,  -4,  -6, -23, -24,   7,  24,   0,  -7,  11,  -6,   1,   6,   6,  12,  25, -31,   9, -13, -10,  -6,   6, -29,  18,  18,   0, -27,  17,   2,   3,  19, -32,  -1, -13,  12,  -8,   1, -14,   7,  19,  -1,  -7,   7,  -4,  -1, -23, -16, -10,   9,  18,  -3,  12,  -1,   8,   0,   2,  -3,   2,  -3,  -5 },
   }
 },
 {  //7
   {
     {  115, -48,  -6,  -5,  -2,  -2,  -1,  -1,   0,  -1,   0,   0, -24,   2,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -9,   4,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -6,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   2,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {   27,  18, -18,  -1,  -4,   0,  -2,   0,  -1,   0,   0,   0, 106, -37, -21,  -1,  -4,   0,  -1,   0,  -1,   0,   0,   0, -13, -31,   7,   4,   2,   2,   1,   1,   1,   0,   0,   0, -16,  -1,  11,   1,   1,   0,   0,   0,   0,   0,   0,   0, -12,   2,   5,  -1,   1,   0,   0,   0,  -7,   4,   2,   0,   1,   0,   0,   0,  -5,   2,   1,   0,   0,   0,   0,   0,  -4,   3,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,  -2,   1,   0,   0,  -2,   1,   0,   0,  -2,   1,   0,   0 },
     {   12,  10,  -6,  -4,  -1,  -2,   0,  -1,   0,   0,   0,   0,  10,  59, -28,  -3,  -5,  -2,  -2,  -1,  -1,   0,  -1,   0,  82, -24, -36,   3,  -4,   1,  -1,   1,  -1,   0,   0,   0,  -3, -46,   2,  12,   2,   3,   1,   2,   0,   1,   0,   1,  -7, -14,  15,   6,   0,   1,   0,   0,  -8,  -3,  11,   0,  -1,   0,   0,   0,  -4,   1,   5,  -2,  -1,   0,   0,   0,  -5,  -1,   2,  -2,   0,   0,   0,   0,  -3,   1,   2,  -1,  -2,   0,   1,   0,  -1,   1,   1,   0,  -2,   0,   1,   0 },
     {   26,  70, -32,  -9,  -6,  -3,  -2,  -2,  -2,  -1,  -1,  -1,   1,  10,  -8,  -8,   2,  -2,   1,  -1,   0,  -1,   0,   0, -31,  54, -20,  -5,   2,  -2,   1,  -1,   0,   0,   0,   0,  30,  -8, -41,   8,   0,   1,   0,   1,   0,   0,   0,   0,  -3, -34,  -5,  20,   2,   2,   1,   1,   4, -15,  13,  10,  -4,   1,  -1,   0,   0,  -7,  11,   1,  -4,   1,  -1,   0,   2,  -3,   5,  -2,  -2,   1,   0,   0,  -1,  -4,   3,  -1,   1,  -2,   3,   0,  -1,  -2,   2,  -1,   0,  -1,   1,   0 },
     {  -11, -59,  13,   8,   4,   2,   2,   2,   1,   1,   1,   1,  36, -38,  20,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  47,  49,  -5, -16,  -6,  -5,  -2,  -2,  -1,  -1,  -1,  -1,  42,  -1, -36,  -7,  -1,   0,   0,   0,   0,   0,   0,   0,  -9, -25,  -7,  10,   4,   2,   1,   1,  -9,  -5,   9,   6,   0,   0,   0,   0,  -8,   2,   6,   0,  -1,   0,   0,   0,  -5,   3,   1,  -1,   0,   0,   0,   0,  -4,   2,   0,   0,  -2,   2,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0 },
     {    3, -10,  24, -12,   2,  -3,   0,  -2,   1,  -1,   0,  -1,  30,  62, -16, -16,  -6,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   1,   2,   9, -11,   1,  -2,   0,  -1,   0,  -1,   0,   0, -13,  59, -14, -23,   2,  -4,   1,  -3,   0,  -1,   0,  -1,  15,   0, -57,   0,   6,   2,   2,   1,  -5, -23, -14,  27,   5,   2,   1,   1,   4, -10,   4,  13,  -6,   0,  -1,   0,  -2,  -8,   5,   2,  -3,   2,   0,   1,   1,  -4,   2,   1,  -2,  -3,   2,   1,   0,  -2,   1,   0,  -1,  -2,   1,   1 },
     {    9,  23, -41,   8,  -2,   3,   0,   1,  -1,   1,   0,   0, -33, -44, -22,  22,   4,   7,   2,   4,   1,   2,   1,   1,  45, -40,  23,  12,  -2,   2,  -2,   0,   0,   0,   0,   0,  34,  33,   9, -15,  -8,  -5,  -2,  -2,  -1,  -1,  -1,  -1,  15, -11, -31,  -6,   2,   0,   2,   0,  -9, -22,   2,  19,   4,   2,   0,   0,  -1,  -1,  19,   7,  -7,  -1,  -2,  -1,  -3,   3,   9,  -6,  -5,   1,  -1,   0,  -2,   2,   3,  -4,  -2,   2,   2,  -2,  -1,   2,   1,  -1,  -1,   1,   1,  -1 },
     {  -19, -39,   0,   7,  10,   2,   4,   1,   2,   1,   1,   0,  -8,   1, -36,  14,   5,   2,   2,   1,   1,   1,   1,   0, -32, -27,  -3,  17,   7,   2,   3,   1,   1,   0,   1,   1, -25,   6, -22,  20,   4,   0,   2,   0,   1,   0,   1,   0,   0, -35,   2,  44,  -3,   2,   0,   1,   5,   1,  52,  21, -20,  -3,  -5,  -1,  12,  10,  24, -21, -21,   1,  -3,   0,   1,   3,   4, -20,  -1,   5,   0,   1,   3,   3,   1,  -8,   1,   2,   1,  -4,   2,   2,   0,  -2,   1,   2,   1,  -2 },
     {  -11,  -8, -41,   1,   8,   4,   4,   3,   1,   1,   1,   0, -10, -22,   7,  -8,  12,   0,   3,   1,   1,   0,   1,   0,  14,  26, -30, -10,  10,  -3,   1,   0,   1,  -1,   1,   0, -61,  -9,  -1,  11,  12,  -1,   4,   0,   1,   0,   1,   0, -37,  33, -18,   3,  12,  -6,   4,  -1,  -3,  -2, -24,  39,   4,  -1,   0,  -1,  10,   1,  23,  29, -17,  -2,  -4,  -1,  10,   0,  10,  -3, -18,   2,  -2,   0,   5,  -3,   7,  -6,   3,  -1,   2,  -3,   2,   0,   4,  -1,   2,  -1,   1,  -1 },
     {   15,  41,  41, -12, -10,  -4,  -4,  -2,  -2,  -1,  -1,  -1, -14, -17,  48, -12,  -5,   0,  -1,   1,  -1,   0,   0,   0,  26, -23,  33, -28,   1,  -1,   1,   0,   0,   0,   0,   0, -41,   2, -25, -19,  13,   1,   5,   0,   2,   0,   2,   0, -29, -24,  -7,  21,  13,   3,   2,   1, -11,  16,  23,   8,  -2,  -4,  -2,  -2,   1,  17,  -3,  -9,  -4,  -2,   0,  -1,  -1,   7,  -8,  -4,   2,   0,   1,   0,   1,   3,  -8,   1,   0,   1,  -4,   1,   1,   0,  -3,   2,   0,   0,  -2,   1 },
     {   -5,  -2, -47,   4,   4,   4,   3,   2,   2,   1,   0,   1,  -2,  12, -34,  19,  -1,   2,   0,   1,   0,   1,   0,   0,  27,  43,  42,  -8,  -9,  -8,  -4,  -4,  -2,  -2,  -1,  -1, -43,  39,  -7,  -9,  -6,  -2,   0,  -1,   0,   0,   0,  -1, -12,  -9,  23,   3,  -9,   3,  -2,   1,  -2,  21,   9, -34,  -5,   1,   0,   0,   2,  -5, -28, -15,  14,   5,   2,   1,   1,  -4,  -9,   8,  13,  -1,   0,  -1,   3,  -4,  -3,   8,   1,  -2,   1,   3,   1,  -3,  -1,   1,   0,  -1,   1,   1 },
     {    2,   1,  19, -24,   7,  -1,   2,  -1,   0,   0,   1,  -1,  -9, -28, -30,  -5,   9,   7,   3,   4,   1,   2,   1,   1,   7, -24, -14, -26,  17,   2,   3,   0,   2,   0,   1,   0, -11,  12, -55,  27,  16,   2,   3,   1,   1,   0,   1,   0,  41,  -4,  24,  12,  -7,  -3,  -4,  -1,   1,  12, -31, -21,   9,  -3,   3,  -1,  -4, -21, -14,  38,  11,   1,   1,   0,  -1,   1,  19,  15, -10,  -5,  -2,  -1,   1,  -4,   6,   3,  -1,   1,   7,  -1,   0,  -1,   2,   1,   0,   2,   4,   0 },
     {   11,  15,   3,   5,  -7,   0,  -2,  -1,  -2,   0,  -1,  -1,   8, -17,  25, -10,   1,  -4,   0,  -2,   0,   0,   0,  -1,  29,   8, -32,   5,   2,  -2,   1,  -1,   0,   0,   1,   0, -11,  29,   1,  21,   0,  -5,  -1,  -3,   0,   0,   0,  -1,  42,  47, -10,  24, -19,  -9,  -4,  -2,  36,  -7,  14,  -4, -40,   0,  -4,   1,   1, -13,  -5, -43, -10,  12,   3,   4,  -6, -13, -13,  -9,  16,  10,   2,   2,  -3,  -4,   0,   1,  -3,  -4,  -2,   0,  -3,  -1,   1,  -1,  -2,  -2,  -1,   0 },
     {    3,  -1,  -8,  45, -15,  -1,  -3,   1,  -1,   0,   0,   0,  11,  10,  36,  43, -28,  -2,  -5,  -1,  -3,  -1,  -1,   0,  -2,   4,  20, -12, -23,   0,  -2,   0,   0,   0,   0,   0, -37, -27, -12,  -1,  -8,  10,   4,   4,   1,   2,   0,   1,  51, -20,   5, -10,  -4,   7,  -1,   2,  18, -25, -10,  -2,  10,   4,   0,   0,   7, -20,  21,  13,  -2,   0,  -2,   0,  -1,   3,  18, -12,  -6,   0,  -1,   0,   0,   2,   4, -11,  -2,   4,   0,  -7,   0,   2,  -2,  -3,   0,   1,  -2,  -3 },
     {   11,  20,  45,  30, -10, -15,  -4,  -7,   0,  -3,  -2,  -1,   1,   1, -20,  60, -13, -10,  -1,  -3,   1,  -2,  -1,   0,   6,  11, -13,  21, -11,  -6,   4,  -2,   1,   0,   0,   0,  14,  25, -18,  10,  -9,  -3,   1,  -1,   1,  -1,  -1,   0, -31,  22,  11,  -8,  -4,  -1,   1,  -1,  -8,  41, -18,   9,  -1,  -6,   2,  -2,  14,   3,   2,   9, -25,  -2,  -1,   0,   3,  -9,   0, -21, -14,   8,   0,   2,  -2, -10,  -2, -10,   1,  -2,   2,  -5,   0,  -4,   0,  -2,   0,  -2,   0,  -3 },
     {   -2,  -4,  -4,  38, -12,  -1,  -4,  -1,  -1,   0,  -1,   1,  -8, -13, -18,  15,  -8,  -1,   1,   0,   1,   0,   0,   1, -11, -18, -39,   6, -14,   8,   4,   3,   1,   2,   1,   1,  -7, -18, -29, -43,  12,  15,   6,   3,   3,   1,   1,   1, -18,   9, -43,   8,  20,  12,   2,   1,  23,   1,   8, -11,   7,   4,  -2,   2,  -3,  -3, -33, -10,  31,   1,   4,   0,  -2, -16,  -3,  35,  16,  -5,  -1,  -1,   0,   1,  10,  14,   2,   0,   8,   7,   1,   2,   5,   1,   0,   1,   4,   2 },
     {    0,  -9, -31, -10,   8,   6,   2,   2,   1,   0,   0,   1,   7,  17,   2,  -3,   1,   0,  -4,   1,  -1,   0,  -1,   0,   0, -15,  42, -21,   1,  -2,  -3,   1,   0,   0,  -1,   0,  32, -31, -16, -13,  15,  -1,   3,   1,   1,   0,   0,   0,  12,  17,  -6,  13,  21,  -7,   2,  -3,  46,  52, -12,  24,  -6, -17,  -4,  -6,  20,  -6, -17,   2, -32,  -6,   0,   1,  -4, -20, -10, -10,  -7,  11,   4,   3,  -5,  -9,   1,   2,  -3,   0,   3,   0,  -2,   0,   1,   0,  -1,   0,   1,  -1 },
     {    5,   8,  49, -42,   3,  -3,  -1,   0,   1,   0,   0,  -1,  -4, -29, -46,   9,   7,   9,   4,   3,   2,   0,   1,   0,  11,  36,  22,  21, -13,  -2,  -6,  -2,  -3,  -1,  -1,   0, -21, -39,  31, -13,  -8,   0,  -1,   1,   0,   1,   0,   0,   5, -11, -20,   3,  -1,   4,   1,   2,  42, -16,  -5,   7,  -3,   3,  -2,   2,  12, -14,  -1,   1,  -2,   0,   0,   2,   3,  -7,   5,  -1,   0,   2,  -1,   0,  -2,   0,   3,   0,  -1,   3,   2,  -1,  -2,   2,   0,  -2,   0,   2,   0,  -1 },
     {    2,   2,  -1,  15,  -2,   0,  -2,   0,  -1,  -1,  -1,   1,   6,   9,  15, -12,  13,  -8,   0,  -4,   1,  -1,  -1,  -1,   7,   5,  14,  40,  -5,  -8,  -5,  -1,  -2,   0,  -1,   0,   5,   3,   9, -12,  11,  -7,   0,  -2,   1,  -1,   0,   0,   2,  13,  -3,  47, -10,  -8,  -3,  -1,  10,  -1,  21, -26,  -9,  -3,  -2,   1, -12,  18, -21,  38,  12, -19,   6,  -4,  13,   2,  30,  13, -59,  -3,  -4,   1,  -3,  -5,  -2, -36,   2,  -4,  -1,  -4,  -1,  -2,  -4, -10,   1,  -2,  -1,   0 },
     {   -2,  -3,   4,   4,   4,   1,   0,   2,  -4,   1,   1,  -1,  -2,  11,   6,   5,  -3,   3,  -3,   1,  -5,   0,   1,   0,  -4, -19,  32, -11,  -2,   3,  -4,   2,  -1,   1,   1,   0,  18,  -7, -11,  15,  -5,  -9,   4,  -1,  -2,   0,   2,  -1, -62,  19,   2,  19, -21, -11,   8,  -2,  12, -30, -10,  -8, -30,  13,  -1,   1, -12, -60,   8,   3,   5,  28,   0,   4,  -1,   7,  28,   9,  17,   5,  -5,  -3,  10,   7,  11,  -3,   1,   4,   5,  -6,   0,   0,   0,  -3,   1,   1,   2,  -4 },
     {   -4,  -6, -15, -29,   0,   1,   5,   1,   5,   0,   1,   1,   4,   3,  36,  11,   5, -12,  -1,  -4,  -1,  -1,  -2,  -1,   3, -12,   8,  67,  12, -25,  -1,  -6,  -2,  -2,   0,   0, -11, -11, -28,  34, -16, -14,   2,   1,   3,  -1,   1,   1,  -3, -29, -34, -14, -20,  10,   8,   5,  -4,  15, -27,  -4,  -2,  12,  12,   2,   4,   7, -16,  -3,   4,   9,   2,  -1,  -6, -16,  -8,  12,   9,   5,  -2,   1,  -4,  -3,  12,  11,   2,   2,   7,  -1,   2,   1,   3,  -1,   0,   1,   1,  -1 },
     {   -4,   2,  -3, -33,   8,  -2,   7,  -2,   2,   0,   1,   0,   0,   7,  -9,  31,   7,  -3,   0,  -3,   0,  -1,   0,  -1,  -6,   7,  21,  13, -10,  -9,   0,   0,  -1,  -1,  -1,   0,  -3, -28, -38, -15,   7,   3,   6,   2,   2,   1,   1,   1,  22,  59, -12,   0,   7,  -2,  -1,  -2, -36,  -1,  15,   1,  -3,   1,  -3,   1, -55,   8,  22, -10,   7,  -1,   0,   0,  -3,  23,  -6,  -7,   4,  -3,   0,  -1,   3,   8, -11,   0,   4,  -1,  -7,   4,   4,  -2,   0,   6,   3,   0,   0,   1 },
     {   -9,  -8,  -7, -15, -51,  -3,   9,   1,   9,  -1,   1,   2,  -9,  -8,  -5, -17, -49,   1,  11,   0,   7,   2,  -1,   2,  -5,   5,   5,  11, -28,   6,  13,  -6,   0,   1,   0,  -1,  -1,   3,  17,   4,  16,  18,   1,  -2,   3,  -3,   0,  -1,   3,  -4,  16,  27,   6,  10,  -8,   0, -33,  -3, -15,  21,  -8, -11,   8,   5, -36, -10, -18,   3, -18,   1,  12,   7, -14, -34,   7,  -2,   2,  22,  -1,   6,  -9,  -3,  19,   7,   3,   4,   7,  -5,   3,   2,   1,  -4,   1,   1,  -3,  -2 },
     {    6,  11,  13,  41,  41,  -4, -12,  -3,  -6,  -1,  -3,  -1,   1,   1,   2,  16,  48,  -9,  -8,  -4,  -1,  -3,   0,  -1,  -3,  10,   5,  -6,  10,  -8,  -5,   1,  -1,  -1,   1,  -1,  -7, -13,  28,   5,  12, -12,  -8,  -1,   0,  -1,   0,   0,  19, -20,  -9,  16,   7, -16,  -3,  -3, -39,  10,  -6,  19, -10, -18,   3,   0, -29, -23, -22,   1, -16,  -3,   8,   5, -16, -15,  -1,  20,  12,   6,   4,   3,   5,  13,  15,  17,   4,   6,   4,  -5,   3,   5,   1,  -4,   0,   1,   0,  -1 },
     {    0,  -2,  -1,  14,  -9,   2,  -5,   2,  -4,   2,   0,   1,  -2,  -5, -16, -10,  15, -19,   2,   0,   2,   1,   2,   0,  -5,  -3,   7, -31, -14,   0,  -2,   7,   0,   3,   0,   1,   3, -24,  -3,  -2,  15, -16,   8,  -3,   2,   0,   1,  -1,   2,   8, -26,   6, -55,   5,   8,   2, -23, -17,  -1, -30,   7,  17,  13,   2,  13,  31, -13,  15, -22,  16,   1,  -2,  24, -25,  -6, -39,   9,  25,  -5,   6,  -3,  -3, -14,  17,   4,  -6,   9,   5,  -4,   3,  -2,   3,   0,  -2,   4,   2 },
     {   -2,  -1,  -8, -27,  18,   0,  -2,  -2,  -1,   0,   0,   0,   0,  -4,  17, -10,  30,  -9,  -4,  -3,   0,   0,   1,  -1,  -4, -11, -31,   5, -27,  -9,   7,   6,   0,   2,   1,   1,  -8,  10,  -4, -54, -38,   6,  10,   5,   0,   3,   2,   2,   0,  -5,  22, -19, -15,  22,   0,   2,  -3,  22,  24,   0,   1,   6,  -3,  -3,  -5, -37,   0,  11, -17,   6,  -1,   1,  -8, -24,  30, -19,  -1,  18,  -1,   1,  -2,  11,   1,  -9,  -1,   3,   0,   0,   2,   6,  -1,   3,   0,   2,   1,  -1 },
     {    0,   0,   6,  48, -41,  -5,  -7,  -1,   2,   3,   1,   0,  -7,  -1, -26, -56,  15,   3,   2,   5,   2,   3,   3,   1,   3,   5,  14,  16,  18, -25,   2,   0,   1,  -2,   0,   0,  -8,  -6, -14,   3, -21,  -8,   0,   4,   1,  -1,  -1,  -1,   3,  -4,  -6, -18,   5,  10,   9,   1,  23,  25, -10,   3,   6,   0,   3,  -5, -40,  -5,  11,   2,  -8,  11,   7,  -2,  -9,  23,   5, -17,   9,   5,   3,  -1,   8,  20, -13,  -4,  -3,  -2, -10,  -3,  -5,  -3,  -8,   6,   0,   0,   1,   6 },
     {   -3,  -1,   1,  -5,  -9,  20,  -2,  -1,  -2,   0,  -3,  -1,   0,   5,   9,  17,  11,  16,  -5,  -4,  -4,  -3,  -1,  -1,   0,   1,  -4,  14,  23,   6, -17,   0,  -1,   1,   0,   0,  -5,  13,  -5,  -4,  43, -18, -12,  -2,   2,   0,  -1,  -1, -23,  -2,  10, -14,  -7,  -4,  -8,   2,  33, -24,  12, -23,  34, -21,   0,   6, -27,  -6,  15,  11, -15, -28,   5,   0, -43, -26,  -5, -23,   3,   4,   8,   3, -22,   2,  -5,  20,   0,   4,   3,   7,  -1,   5,  -3,   6,   5,   3,   2,  -2 },
     {    1,   1,   2,  11,  39, -23,   0,  -9,  -2,  -3,   3,  -3,  -4,  -2,  -3,  -7,  -1, -19,  -8,   2,  -4,   4,   0,   1,  -1, -12,  10, -12, -19, -17,  15,   0,   1,   0,   1,   0,  -3,   6,  -8,  13, -38,   5,  13,  -5,   5,   2,   1,  -2, -12,   4,  26,   8,  -6,  17,   5,   4,  32, -34,  -7,  25,  10,   2,  -6,  -5, -39,  27,  -3,   7,   8,  -7,   3,  -2,  -3, -16, -35,   7, -15,   5,   8,   5, -18, -29,   4,  11,  -2,   2,  10,   7,   0,  -4,   5,   2,   2,   6,   1,   4 },
     {    4,   5,   7,  11,   3,  37,  -4,  -3, -11,   0,   2,  -2,   2,   3,   2, -16,  14,  27,   4,  -4,  -8,  -1,   5,  -1,   5,   5,  18,   7, -20,  52,  -5,  -4,  -5,  -1,  -3,  -1,   1,  -2, -11,  32, -26,  17, -15,  -7,   1,   0,   1,   1,   0,  20, -10,   3, -17,  27, -19,   4,  -7,  15,  14,  14,  30,  11, -27,  -4,  -2,   5,  -2,   2,   2, -17, -20,   6,   1, -22,  10,   5,   9, -18,   0,  10,  15,   3,  11,   9,   8,  -6,   3, -21,  -2,  -7,  -4,  -9,  -3,  -1,   0,   1 },
     {   -2,   0,   2,  -3, -15, -11,   8,   4,  -1,   1,  -1,  -3,  -3,   1,  -5,  13, -17,   9,  -3,  -5,   1,  -1,  -1,  -1,  -3,   7,  -1, -38,  36,  -1,   1, -11,   1,   6,   1,  -1,  -1,  -6,  17,   3, -11, -13,   0,  -4,   2,   3,  -1,   2,   5,   0, -20, -23, -23,  -5,  15,   4,   5,  35,  38,   3, -20,  23,   6,  -8, -26,  14,  19,  13,  12,  10, -15,  -4, -18, -24,  19,  24, -11,  -4,  -6,   1, -24, -26,  23,  -3,   9,   7,   2,  -4,   8,   3,  -5, -15,   0,   3,  -6,  -4 },
     {    1,   2,   0,  -5,   5, -34,  19,  14,  17,   9,  -6,   6,   1,   1,   1,  -2,  16, -29, -12,  -2,   9,   0,  -2,   0,   2,   1,   4, -12, -21,  20,   1,   3,  -7, -11,  -2,   0,   3,   9,   7,  22,   0, -11,  21,  -2, -17, -16,   3,   2,  -3,   1, -10,   8,  22,   2,   7,  -6,  15,  10,  -8, -22,  29, -11,  19,  11,  -8,  21,  18, -31,   9, -17,  -4,  -2,   0,  -7,  46,  -8,   9,  10,  -8, -16,  10,   7,  36,   2,  -1,   4, -13,  -4,   1,   6,   2,  13,   6,  -6,   0, -10 },
   },
   {
     {   83, -49,  -1,  -3,   0,  -1,   0,   0,   0,   0,   0,   0, -72,  31,  10,   1,   2,   0,   1,   0,   0,   0,   0,   0,  18,   9, -14,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  -2, -12,   4,   3,   0,   1,   0,   0,   0,   0,   0,   0,   1,   6,   2,  -3,   0,   0,   0,   0,  -2,  -2,  -3,   1,   1,   0,   0,   0,   1,   1,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -60,   5,  14,   2,   4,   0,   1,   0,   1,   0,   0,   0, -34,  56, -16,  -1,  -3,   0,  -1,   0,  -1,   0,   0,   0,  66, -44, -12,   7,  -2,   2,  -1,   1,  -1,   0,   0,   0, -18,  -7,  25,  -5,   2,  -1,   1,  -1,   0,   0,   0,   0,   5,  13, -11,  -5,   2,  -1,   1,  -1,  -1,  -6,  -1,   6,   0,   0,   0,   0,   4,   1,   1,  -1,  -2,   0,   0,   0,   0,  -2,   1,   0,   1,   1,   0,   0,   2,   1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0 },
     {  -42,  34, -10,   2,  -1,   1,   0,   0,   0,   0,   0,   0, -50,  -8,  26,   1,   4,   0,   2,   0,   1,   0,   0,   0,   5,  58, -26,  -7,  -3,  -2,  -1,  -1,   0,   0,   0,  -1,  42, -48, -16,  16,  -2,   4,  -1,   2,  -1,   1,   0,   1,  -7,  -4,  30,  -6,  -1,  -1,   0,   0,   6,   6, -12,  -7,   3,  -1,   1,   0,  -1,  -4,   2,   6,   1,  -1,   0,   0,   4,   0,  -1,  -1,  -2,   0,   0,   0,   0,  -2,   2,   1,   2,   0,   0,  -1,   0,  -1,   1,   0,   2,   0,   0,   0 },
     {   26,  10, -22,   1,  -3,   1,  -1,   0,   0,   0,   0,   0,  44, -21,  12,  -2,  -1,  -2,  -1,  -1,   0,   0,   0,   0,  20, -11, -29,   4,  -1,   3,   0,   1,   0,   1,   0,   0, -11, -49,  47,  10,  -1,   1,  -1,   1,   0,   0,   0,   0, -24,  49,   8, -33,   5,  -5,   2,  -2,  -2,  -1, -34,  18,   6,   0,   2,   0,  -2,  -1,  16,   6, -10,   1,  -2,   0,  -1,   0,  -3,  -7,   2,   2,   0,   1,  -3,   1,   3,   2,   0,   0,  -2,  -1,  -2,   1,   2,   1,   0,   0,  -1,   0 },
     {  -32, -73,  41,   5,   8,   2,   3,   1,   2,   1,   1,   1,  34,  36,   2, -20,  -2,  -5,   0,  -3,   0,  -1,   0,  -1, -37,  32, -41,  13,   2,   3,   1,   1,   0,   0,   0,   0,   9, -10,  12,  14,  -8,   2,  -3,   1,  -1,   0,  -1,   0,   8,  -1,   1, -15,   0,   1,   0,   0,  -1,  -4,  -2,   5,   4,  -1,   0,   0,   2,   2,   1,  -2,  -3,  -1,   0,   0,  -1,  -2,  -1,   1,   2,   1,   0,   0,   1,   1,   1,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },
     {   19,  -5,  10,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0,   0,  37,  19, -38,   5,  -3,   2,  -2,   0,  -1,   0,   0,   0,  33, -16,   0,  -1,  -3,   0,  -1,   0,   0,   0,   0,   0,   4, -33, -23,  17,   3,   4,   1,   2,   0,   1,   0,   1,  -5, -32,  60,   7,  -9,   2,  -3,   1, -14,  32,   2, -42,   9,  -3,   2,  -2,  -8,   0, -23,  20,  12,  -3,   2,  -1,  -3,   4,   9,   4, -11,  -1,  -1,   0,  -1,  -2,  -3,  -3,  -2,   2,   3,   0,  -1,   0,  -2,   0,  -1,   2,   2,   0 },
     {   17,  25, -25,   8,  -4,   0,  -2,   0,  -1,   0,  -1,   0,   7,  43, -22,  -8,  -4,  -1,  -1,  -1,  -1,   0,   0,   0, -31, -24,  -4,  23,  -1,   5,  -1,   2,   0,   1,   0,   1,  70, -25,  22, -20,  -7,  -2,  -2,   0,  -1,   0,  -1,   0, -26, -20, -18,   5,  16,   0,   4,   0,   2,   4,  29,  13, -13,  -2,  -3,  -1,  -4,   3,  -6, -23,   1,   3,   1,   1,  -1,  -3,  -2,  10,   8,  -1,   0,   0,  -1,   1,   2,  -2,  -1,  -1,   0,   1,  -1,   1,   1,  -1,   0,  -1,   0,   1 },
     {  -11,  -7,  17,  -6,   1,  -1,   0,   0,   1,   0,   0,   0, -30,   2, -17,  14,   1,   2,   1,   2,   0,   1,   0,   0, -45, -28,  42,   3,   1,   1,   0,   1,   1,   0,   0,   0,  28,  14,  17,  -8,  -3,  -3,  -1,  -1,  -1,   0,   0,   0,  -5,  16,  16, -36,   4,  -3,   1,  -1, -13,  33, -52,   1,  20,  -6,   5,  -2,   9, -11,  -7,  37, -14,   0,  -1,   0,   6,  -1,  10, -12,  -9,   5,  -1,   1,   3,  -3,  -4,   0,   1,   1,   1,  -1,   1,  -1,  -2,   1,   1,   1,   1,  -1 },
     {  -18, -55,  16,  14,   3,   4,   1,   2,   0,   1,   0,   1,   0, -39,  18,   1,   2,   1,   1,   0,   0,   0,   0,   0,  33,  -6,  28,  -4,  -4,   0,  -2,  -1,   0,   0,  -1,  -1,  31,  -9,  -3, -29,   9,  -3,   2,  -2,   1,  -1,   1,   0, -59,  24,  -7,  24,   9,  -5,   3,  -1,  16,  21,   7,  -7, -19,   2,  -3,   0,  -8,  -3, -13, -10,  13,   5,   1,   1,   1,   6,   4,  11,  -1,  -5,  -1,  -1,  -3,   2,  -4,  -6,   1,   3,  -1,   3,  -1,   1,  -2,  -1,   1,   1,  -1,   1 },
     {   10,   0,  16, -17,   2,  -3,   1,  -1,   1,  -1,   0,  -1,  21,  21, -26,  11,  -1,   0,  -2,  -1,  -1,   0,   0,   0,  28,  21,   8, -24,  -4,  -2,  -1,  -2,   0,  -1,   0,  -1,  20,   5, -47,   2,   7,  -2,   1,   0,   0,   0,   1,   0, -26, -12, -17,   7,   7,   1,   3,   0,   2,  -5, -21,  48,  -1,  -2,   0,  -1,  11, -26,  41,   6, -35,   8,  -6,   2,  -5,   0,   5, -34,  13,   7,  -1,   2,  -3,  -3,  -1,  10,  -2,   0,   2,  -4,  -2,  -2,   1,   3,  -1,   0,   1,  -2 },
     {   14,  28,  69, -36,  -4, -11,  -2,  -4,  -1,  -2,   0,  -1, -13, -33, -48,   3,  16,   5,   6,   2,   3,   1,   1,   1,   5,  28, -15,  42, -16,   0,  -4,  -1,  -2,  -1,  -1,   0, -10,  -7,  22, -26, -10,   7,  -1,   3,  -1,   1,   0,   0,  -1,   2,  -8,   1,  17,  -2,   0,   0,  10,  -1,   5,   1,  -8,  -4,   0,  -1,  -7,   0,  -5,  -4,   5,   3,   2,   0,   3,   3,  -1,   4,  -2,  -2,  -1,  -1,   0,  -2,  -1,  -1,   1,   1,   1,   1,  -1,  -1,  -1,   0,   1,   0,   0,   1 },
     {    5,  20,  -2,   2,  -4,  -1,  -1,  -1,   0,  -1,   0,  -1,   5,  46, -13, -19,   4,  -3,   1,  -2,   0,  -1,   0,  -1, -30,  14,   5,  -4,  -1,  -2,  -1,   0,   0,   0,   0,   0, -21,  -6, -17,  -8,  18,  -1,   5,  -1,   2,   0,   1,   0, -25,  50,  -8,  31, -12,  -6,  -2,  -2,  42, -24, -22, -25, -11,  11,  -2,   4,  -7, -22,  -5,  25,  26,  -3,   1,   0,   8,   2,  15,   7, -21,  -6,  -1,  -1,  -3,  -8,  -3,  -8,   4,  -2,   4,   4,   0,  -3,   0,  -2,   1,  -1,   1,   2 },
     {  -14, -18,  -7,   3,   5,   3,   2,   1,   1,   0,   1,   0, -16,  -5, -39,  36,   0,   5,  -1,   2,   0,   1,   0,   1, -33,  -7,  31,  -9,  -3,   4,  -1,   2,  -1,   1,  -1,   1, -25, -42,   8,  20,  13,   3,   2,   1,   0,   1,   0,   0,  12,  16,  30,  12,  -8,  -6,  -2,  -2,  16,  -3,  14,   0, -24,   3,  -2,   0,   4,  -1,  29, -43,  -1,  10,  -4,   2,  -5,  13, -21, -14,  28,  -7,   2,  -1,   0,   1,  -1,  15,   1,   3,  -1,  -5,   2,   1,   0,   3,   1,   2,  -1,  -3 },
     {    4,  18,  20, -21,   6,  -2,   1,  -2,  -1,   0,   0,   0, -12,   3,  29, -20,  -1,  -5,   2,   0,   0,   0,   0,   0, -14, -33, -27,  10,  16,   1,   4,   0,   2,   0,   1,   0, -21,  30, -11,  36, -18,  -1,  -4,   1,  -2,   1,  -1,   1, -49,  -6,  29,  -6,  -1,   9,   0,   3,  37,  32,   4,   3,  -7,  -5,  -5,  -3,  14, -25,  -5, -27,  -7,   7,   0,   3,  -5,   4,  -5,   1,  23,  -2,   2,   0,   1,   3,  -2,   7,  -1,   0,  -2,  -2,   2,  -1,  -1,   2,   2,   0,   0,   0 },
     {   -7,  -1,  -1,   8,  -3,   0,   0,   0,   0,   0,   0,   0, -14, -13,   0, -13,  13,   1,   4,  -1,   1,   0,   0,   1, -26, -16, -19,   8,   9,   1,   0,   1,   2,   1,   1,   0, -34, -32,  -8,   4,  24,   0,   6,   1,   2,   0,   1,   0, -16,  -5,  -6,  46,   8,  -4,   3,  -2, -27,  29,  23,  29,  -8,  -6,  -2,  -3,  23,  23,  -2,  24, -32,  -5,  -3,  -1,  13, -20,  25, -17, -27,  14,  -4,   4,   3,   6,   6, -19,   4,  -1,   0,  -2,   1,   3,  -1,  -4,   2,  -1,   2,  -1 },
     {    5,  15,   5,  -6,   2,  -4,  -1,  -2,   0,  -1,   0,   0,  11,  26,  13, -11,  -9,  -1,  -1,  -1,  -1,   0,  -1,   0,   1,  35,  19, -30,   1,  -4,   0,  -3,  -1,  -1,   0,  -1, -26, -11,   6, -24,   7,  -1,   2,   1,   1,   1,   1,   0,  14,   9,  -9,  -4,  17,  -9,   4,  -3, -34,  49, -17,   7, -26,  -4,   2,  -2,  21, -15, -41, -25,   5,  22,  -1,   5,  -2, -23,  -8,  23,  27,  -5,  -2,   1,   5,   0,   7,   8,  -2,  -4,  -4,   2,   1,  -1,   2,   4,  -2,  -2,  -1,   0 },
     {    7,  15,  49, -47,   0,  -4,   1,  -2,   1,  -1,   0,  -1,   4,   4,  28,   4,   0,  -4,  -1,  -2,  -1,  -1,  -1,   0,  -9, -30,   5, -25,   2,   6,   2,   2,   2,   0,   1,   0,  31, -19, -19,  20,  18,  -8,   2,  -2,   1,  -1,   0,   0,   1,  33, -13,   8, -26,   0,   0,  -1, -35, -23,  10, -10,   8,  12,  -2,   3,   7,  33,  -2,  -8,   7,  -9,  -1,  -1,  -5, -10, -23,   7,   2,   2,   4,   1,   3,   2,   7,   8,  -1,  -1,  -4,  -2,  -1,   1,   1,   4,   0,   0,  -1,  -1 },
     {   10,  10,  29,  63, -33,  -4, -10,  -1,  -3,  -1,  -2,   1,  -7, -13, -21, -65,   9,  14,   5,   7,   2,   2,   1,   0,   8,  -6,  21,  -2,  34, -18,   0,  -5,   0,  -2,   1,  -1,  10, -12, -15,  23, -22,  -4,   5,   0,   1,   0,   0,   0,  15,  17,  -4,  -6,  -5,  12,  -1,   0, -14,  -2,  -2,   6,   0,  -1,  -3,   0,   1,  -4,  10, -10,   2,   2,   2,   1,  -3,   7,  -7,  -2,   5,  -3,   1,  -1,   1,  -2,  -2,   5,  -1,   1,   2,  -1,  -1,  -1,  -1,   0,   0,   1,   0,   0 },
     {    1,   9,  13,   9,  -4,   3,  -4,  -2,  -2,   0,   0,   0,   0,  20,   3,  -6,  -5,  -2,  -1,   2,   0,   0,   0,  -1,  -3,  22,  32, -28,   2,  -1,  -4,   0,  -1,   0,  -2,   0, -29,  -8,  13,  -9,   3,  -3,   3,  -2,   2,  -2,   0,  -1, -37, -19,   1, -22,   4,  14,   1,   3,  29,   1,  16,   9,  35, -15,  -3,  -4, -38,  50, -13,   3, -22, -19,   7,  -3,  15,  -2, -31,  -3,  -3,  17,   1,   1,  -2,  -8,   2,  14,   9,  -2,  -3,  -3,   0,  -1,  -1,   2,   3,  -1,  -1,   1 },
     {    5,   5,   1,  10, -18,  10,  -8,   2,  -2,   3,  -2,   1,   8,   5,  10,  -5,  19, -14,   6,  -6,   2,  -2,   1,  -1,  10,  13,   4,  16, -25,   7,  -7,   2,  -2,   1,  -1,   0,  16,  12,  22,  -3,  12, -19,   1,  -4,   0,  -2,   1,  -1,  16,  19,   3,  14, -40,   8,  -4,   2,  13,   8,  16, -31,   5,  -8,  -3,  -1,   5,  -5, -21,  -6, -38,  10,   2,   3,   2, -27,  16, -51,  29,  17,  -2,   5, -14,  15, -25,  13,  -1,  -8,   2,   3,  -5,   5,  -4,   4,   2,  -5,   3,  -1 },
     {    2,   3,   5, -15,  12,  -1,   1,  -1,   1,  -1,   0,   0,   8,  15,  52, -37,   1,  -9,   0,  -2,  -1,  -1,   1,  -1,  -2, -16,  32,  40, -28,   0,  -3,   0,   0,  -2,  -1,   0,  -8, -23, -22, -28,  -4,  20,  -1,   6,   0,   2,   0,   1,  27,  -5,  20,   7,  28,  -8,  -5,  -4,  -1,  -1,  -7,   4, -10, -12,   5,  -3, -18,   4,  23,  -1,  -5,   3,   1,  -2,  -1,  28,  -9, -23,  -3,  -1,  -1,   0,  -2,  -7, -21,   6,   1,   2,   7,   4,   2,  -2,  -5,  -4,  -1,   1,  -1,   1 },
     {   -7,  -8, -34,  -5,  11,   0,   3,   1,   2,   1,   0,   1,   0,   6, -23,  -9,  19,  -6,   3,   0,   2,  -1,   2,  -2,   7,  19,  14,  45, -18, -11,  -3,  -2,  -3,   0,  -1,   0,  -5,  22, -33,   2, -18,  14,  -3,   3,   0,   1,  -1,   0, -32,  36,  21, -17,   6,   2,  -5,  -2, -27, -18,  10,  -2,  10,  -3,   3,   2,  43,  30, -11,  -8,  -1,  -7,   0,  -4,  -8, -20, -19,   6,   1,   4,   1,   1,   1,  -3,  13,  11,   6,  -3,  -1,  -4,   2,  -2,   2,   1,  -3,  -2,   1,  -1 },
     {    0,   7,  14,   8,  -7,  -3,  -1,   0,  -1,   1,   0,  -1,  -1,   4,  15,   6, -22,   7,  -5,   1,   0,  -1,   0,   0,  -7,  13,  17,  -3,  -2,   1,  -4,  -2,   1,   1,   0,  -1, -18, -20,  20, -17,  -4,   1,   4,  -2,   3,   0,   0,   1, -28, -23,  -9, -12,  -5,  13,   2,   1, -15, -16,  20, -15,  27,  12,  -5,   3,  59, -20,  11,   4,  17, -20,  -3,  -2, -43,  23,  15, -15,  -6, -12,  10,  -3,   1,  24, -28,  -1,  -2,  -6,  -1,  10,   4,   4,  -5,   1,  -1,  -1,  -2,   3 },
     {    6,  14,  17,  14,   4, -18,   3,  -5,   0,  -1,   2,  -1,   9,  21,  17,  31,   0, -12,  -6,  -1,  -1,  -1,   1,  -1,   8,  22,  21,   6,   9, -28,  -2,  -1,   1,  -1,   1,   0,   6,  23,  27,  26, -22, -10,   0,   0,  -2,   0,  -1,   0,  -7,  10,  19,   3,  -5, -21,   8,  -3, -11,   4,  13,   6, -36,   2,   3,  -5,  -6,  12,   9, -12,  -6,   0,  -8,   0, -11,   1,  -2, -21, -40,  -3,  -1,   3,   0, -22,  -4, -49, -13,   5, -13,  10,   3,  -6,   1,  -2,  -4,   0,  -5,  -4 },
     {    5,   5,   2,   9,  79, -22,  -9,  -6,  -4,  -4,   0,  -4,  -4, -11, -11, -18, -51,   2,  15,   2,   6,   2,   0,   2,   5,   5,  19,  13,  -6,  29,  -8,  -2,  -3,  -2,   0,  -1,   1, -14, -13,  22,   5, -19,  -8,   1,   1,  -1,   1,  -1,  10,  12, -11, -19, -10,   6,  13,   2,   9,  -5,  18,   8,   3,  -3, -12,  -3,  -2, -10, -16,   3, -11,   1,   5,   6,   6,  -5,  18,   5,  -2,   4,  -5,  -3,  -7,   3,   0, -16,  -3,   1,  -5,   3,  -1,   0,   4,   0,   2,   1,  -1,   0 },
     {    3,  -3,  13,  -3, -15,  15,  -3,   4,   0,   1,  -1,   2,   8,   2,  15,  22,  14,  -1,  -8,   5,  -5,   0,   0,   0,   2, -12,  25,  -8,  -8, -11,   3,  -3,   2,   0,  -1,   1,   3, -26, -12,  16, -29,  22,   7,   1,   0,   2,   1,   2,  23,   3,  -2, -11,  31,  18, -14,   1,  43, -24,   8,   6,   0, -22,   3,   0,  18,  -8, -20,   3, -10,   2,   7,  -2,  -8, -34,  32,  19,   0,  10,  -5,   1,  -6,  25,  23, -26,  -7,   5, -15,  -7,   0,   5,   0,  -1,  -1,   0,  -1,  -3 },
     {   -3,  -7, -18, -36,  31,  11,  -4,   0,   0,   3,  -3,   0,   3,   0,  -5, -11,  42, -13, -16,   4,  -2,   3,   2,  -5,   7,   8,  19,  -3,  13, -27,  -4,  11, -12,   8,  -3,   4,   6, -13,  21,  22, -24,   6,   9,  -1,  -4,  -1,  -2,  -6,  12,   1, -28,  17, -21,  35,  -3,  -3,   3,  28,   5,  10,   7,   1, -12,   5,   6, -14,  -4,   6,   2, -20,   4,   5,   3,  28,  -1,   7,  -6,   0,   2,  -8,  -6, -16,  -8,   4,  24, -12,   7,   5,   3,  -2,   7, -14, -14,   1,  -9,  -9 },
     {   -4,   0,   3, -11,   1,  20,   1,   2,   0,   1,   0,   3,  -5,   2,   4,  -2,  -6, -19,   2,  14,  -3,   3,   1,   0,  -4,   4,  18,   7,   6,  -2,  11,  -2,  -2,   2,  -3,   1, -16,  -6,   9,   5, -25,  -4,  -8,   2,  -2,   0,  -2,   0, -30, -18,   4,   2, -12,  16,  -6,   9, -43, -28,  -6,   0,  -3,   8,  13,   0, -26,  -3,  15,  -2,  22,  21,  -7,   2,  52, -21,  36,  -4,  29, -19, -11,   1, -21,   9,  -2,  -7,  11,   8,  -7,  11,  -5,   9,   0,  -3,  -1,  -1,   0,  -4 },
     {    1,   1,   3,  11,   6,  -9,  -7,   0,   1,   2,   0,  -2,  -1,  -3,  -2, -28,  57, -25,  12,   3,   1,  -2,   1,  -2,   2,  -5,  17, -15, -48,  29,  -2,  -6,  -3,   0,   1,  -1,   0,   3,  15,  26,  26,  -3, -15,   0,  -4,   2,   0,   0,  -5, -35,  -1, -16,  -3, -18,  19,  -2,  12,  -1, -15,  13, -14,  24,   1,  -3,  12,   6,   9,   1,  26,  -8,  -8,  -1,  -8, -16,  -3,   6,  -9, -10,   4,   2, -11,  11,  10,  -2,  -1,   3, -10,   2,  -1,   4,  -1,   1,  -2,   1,   0,  -1 },
     {   -3,  -6, -10, -24,  20,   1,   0,  -1,   1,   3,  -2,   2,  -3,  -7,  -6, -24,  11,   1,  -3,  -2,  -3,   1,   3,   1,   2,  -1,   6,   6,  51, -20,   1,  -3,   3,  -3,   2,  -3,   1,  -9,   7, -32,   7, -24,  12,  -9,   4,   1,   2,   2,   3, -23,  29,   5, -18,  -1,   7,   1,  19, -34, -30,  15, -12,  20,  -3,   6,   0,  26, -20,   3, -15,  16, -13,   7, -40, -13,   6,  -1,  10,  -5,   3,   1,   1,  10,  -4,  -8, -13,   3,  -7,   4,   1,  -4,   4,   7,   0,  -1,  -3,   0 },
     {   -1,   3,  -6,  -1,  13, -26,  18,   5,   5,   6,   5,   1,  -3,   0, -13,  -5,  13,  -5, -14,  33,  -4, -15,   2,   3,  -3,   1, -14, -15,  -9, -29,  28, -12,  18,   0,   1,   7,   6,   9,  -4,  14,  29, -10,   0, -11,  -5,  -1, -17,   9,   0,  16,  -5,  -7,  28,   5, -12, -13,  -4,   0,   9,   0,   5, -28,  17, -18, -14,  15, -12,   5,   3,   2,  -6,  -1, -11,  22,  12,  -4,  19, -25,  -9,  -1,  -1,  28, -37,  -8,  14,  -3, -19,   2,  -8,  -2, -16,   4,  -4,  -2,   1,   5 },
     {    0,   4,   4,   4, -14,  15,  13, -16,   8,   3,  -5,  11,   3,   9,  15,  13,  11,   2,  -6,   7,  -4,  -9,   5,  -5,   2,  12,  17,  42,  -3,  -3, -25,   8,   4,  -1,  -3,  -1,  -5,   3,  -3,  15,  40, -28,  21, -17,   3,   2,   6,   4,  -3,   4,   6, -11,  -7,  19,   8, -10, -12,   7,  13,  17,  11,   2,  -3,   1, -28, -18,  10,   6,   0,   4, -11,  11,  -7,  11,  14,  33,   7,  -7,   9,   4,  29,  14,  24,   0, -18, -16,  13, -18,  17,   0,  -4,   7,   2, -11,   6, -11 },
   },
   {
     {   93, -62,   7,  -6,   1,  -2,   0,  -1,   0,  -1,   0,   0, -53,  30,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   5,   4,  -7,   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -4,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -61,  -2,  30,  -4,   6,  -1,   3,  -1,   1,   0,   1,   0, -62,  61, -16,   4,  -2,   1,   0,   0,   0,   0,   0,   0,  53, -26,  -6,   3,  -3,   1,  -1,   1,  -1,   0,   0,   0,  -3,  -1,   5,  -3,   1,   0,   0,   0,   0,   0,   0,   0,  11,  -3,  -5,   1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   5,  -1,  -1,   0,   0,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,  -1,   0,   0,   2,   0,   0,   0,   0,  -1,   0,   0 },
     {    4,  57, -58,  20, -12,   5,  -5,   2,  -2,   1,  -1,   1, -73,  -9,  40,  -4,   6,   0,   2,   0,   1,   0,   1,   0,  -6,  26,  -7,  -5,   1,  -2,   1,  -1,   0,   0,   0,   0,  25, -22,  -2,   4,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   6,  -3,   0,   0,   0,   0,   8,  -4,  -3,   0,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   4,  -2,  -1,   0,   0,   0,   0,   0,   1,  -1,   1,   0,   2,  -1,   0,   0,   0,   0,   1,   0,   2,  -1,   0,   0 },
     {   42,  46, -51,   6,  -9,   1,  -3,   0,  -2,   0,  -1,   0,  19,  14, -27,   8,  -5,   2,  -2,   1,  -1,   1,  -1,   0,  44, -67,  21,   1,   2,   1,   0,   1,   0,   0,   0,   0, -32,   8,  14,  -6,   3,  -1,   1,  -1,   1,   0,   0,   0,   1,   1,  -3,   1,   0,   0,   0,   0,  -8,  -1,   4,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   1,   0,  -2,   0,   1,   0,  -1,   0,   1,   0,  -2,   0,   1,   0 },
     {    6,  27,   1, -15,   2,  -4,   1,  -2,   0,  -1,   0,  -1, -15,  43, -54,  21,  -7,   4,  -2,   2,  -1,   1,   0,   0, -78,  -3,  40,  -1,   3,   1,   1,   0,   1,   0,   0,   0,  27,  17, -15,  -6,   0,  -2,   0,  -1,   0,   0,   0,   0,   7, -22,   1,   7,  -1,   2,   0,   1,   6,   0,   6,  -4,  -1,  -1,   0,   0,   3,  -6,  -1,   0,   0,   0,   0,   0,   3,   0,   2,  -1,   0,  -1,   0,   0,   2,  -3,   0,   0,   2,   0,   1,  -1,   1,  -2,   0,   0,   1,   0,   1,   0 },
     {   10,  11, -10,  -1,  -2,   1,  -1,   0,  -1,   0,   0,   0,  39,  47, -36, -10,  -2,  -5,   0,  -2,   0,  -1,   0,  -1,  11,  24, -47,  15,  -6,   3,  -1,   1,  -1,   0,   0,   0,  39, -69,  20,   8,  -2,   4,  -1,   2,   0,   1,   0,   1, -21,   3,  14,  -5,   2,  -1,   1,  -1,   3,  -8,   1,   2,   0,   1,   0,   0,  -9,   0,   6,  -1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,  -5,   0,   3,   0,   0,  -1,   1,   0,  -3,   0,   2,   0,   0,  -1,   1,   0 },
     {  -20, -33, -30,  66, -29,  22, -10,   9,  -4,   4,  -2,   2,  15,  44,  16, -46,  13, -12,   4,  -4,   1,  -2,   0,  -1, -37, -11,  -2,  12,   4,  -1,   3,  -1,   2,   0,   1,   0, -17,  17,  17,  -9,  -2,   0,  -1,   0,   0,   0,   0,   0,  11,   4, -17,   0,   3,  -1,   1,   0,   2,  -7,   8,   4,  -2,   1,  -1,   0,   4,   3,  -3,  -4,   1,  -1,   1,   0,   1,  -4,   2,   2,   0,   1,   0,   0,   2,   1,  -1,  -1,   1,  -2,   1,   1,   1,   0,   0,  -1,   0,  -1,   1,   0 },
     {   -3,  13,  21, -21,   7,  -5,   2,  -2,   1,  -1,   0,  -1, -20, -19,   1,   6,   6,   0,   3,   0,   1,   0,   1,   0, -40, -28,  -5,  35, -11,  10,  -3,   4,  -1,   2,   0,   1, -17, -19,  68, -22,   4,  -2,   0,  -1,   0,   0,   0,   0, -13,  57, -21, -15,   4,  -5,   3,  -2,  12,  -5,  -8,   6,  -2,   1,  -1,   1,   0,  10,  -3,  -2,   0,  -1,   0,   0,   6,  -2,  -3,   0,   0,   1,   0,   0,   1,   4,  -2,  -1,   3,  -1,  -2,   1,   1,   2,  -1,   0,   2,  -1,  -1,   0 },
     {  -17, -48, -14,  39, -10,  12,  -4,   5,  -1,   2,   0,   2,  -9, -30, -16,  30,  -8,   8,  -2,   4,   0,   2,   0,   1,  10, -32,  49, -14,   3,  -1,   0,   0,   0,   0,   0,   0,  54, -32,   9,  -8,   3,  -2,   0,   0,   0,   0,   0,   0, -28,  11,  16,  -5,   2,  -2,   1,  -1,   6,  12,  -2,  -6,   1,  -2,   0,  -1,  -7,   6,   0,  -1,   1,  -1,   0,   0,   2,   7,  -1,  -2,   1,  -1,   0,   0,  -2,   3,   0,  -1,   1,   4,  -1,  -2,  -1,   2,  -1,   0,   1,   2,  -1,  -1 },
     {   14,  10,  14, -11,  -4,  -1,  -3,   0,  -1,   0,  -1,   0,   6, -11,  11, -27,  19, -11,   7,  -4,   3,  -1,   1,  -1,  23, -14,   4,  25, -16,   6,  -5,   2,  -2,   1,  -1,   1,  69,  25, -11, -38,   4,  -6,   0,  -2,   0,  -1,   0,  -1,  -1, -12, -54,  24,   4,   0,   3,   1,   6, -27,  23,   8,  -5,   4,  -3,   2, -12,   6,  -2,  -3,   3,  -2,   1,  -1,  -1,  -4,   5,   2,  -2,   1,  -1,   0,  -6,   3,  -2,  -1,  -1,  -2,   2,   2,  -3,   2,  -1,   0,  -1,  -1,   1,   1 },
     {   -7, -12, -28,  15,  -2,   6,  -1,   3,   0,   1,   0,   1,   0,  -6, -40,  46, -17,  10,  -4,   4,  -1,   1,   0,   1,  21,  67,  -1, -14, -14,  -1,  -5,   0,  -3,   0,  -1,  -1, -18,  22,  21, -34,   8,  -6,   4,  -3,   1,  -1,   0,  -1,  12,  17, -39,   2,   3,  -1,   1,  -1,   7, -28,   1,   9,  -1,   3,   0,   2,   0,  -3,   3,   0,   1,  -1,   1,   0,   1,  -9,   5,  -1,   1,   1,  -1,   0,  -1,  -1,   2,  -1,   0,  -4,   1,   1,   0,  -1,   2,  -1,   0,  -3,   1,   0 },
     {   -5, -12, -10,  16,  -4,   6,  -3,   3,  -1,   1,   0,   1, -19, -43, -33,  21,  14,   2,   3,   3,   2,   1,   1,   1,  -8, -12, -43,  56, -15,  10,  -4,   3,  -1,   1,   0,   1,  -7,  -1,  10,   5,  -5,   5,  -4,   2,  -2,   1,  -1,   1,  34, -50,  20,  13,  -2,   1,  -1,   1, -13,   9,  34, -13,  -3,  -1,  -2,  -1,   6,   5,   2,  -7,   1,  -2,   1,  -1,  -2,   6,   8,  -5,  -1,  -1,   0,  -1,   2,   4,   0,  -4,  -1,   3,   3,  -2,   1,   2,   0,  -2,   0,   2,   2,  -1 },
     {   13,  29,  55,  21, -64,  30, -25,   9,  -9,   4,  -5,   1, -12, -12, -28,  -1,  25,  -9,   8,  -1,   3,   0,   1,   0,   6,  14,   1, -11,   4,  -4,   3,  -2,   2,  -1,   1,  -1, -28,  -6, -15,  -3,  10,   0,   2,   1,   1,   0,   1,   0, -42,   5,   5,  19,   2,  -1,   2,  -1,   8,   1,  12,  -7,  -4,   1,  -1,   0,  -4,   3,  -5,   1,   3,  -1,   1,   0,   6,   0,   2,   0,  -3,   1,  -1,   0,   0,  -1,  -1,   0,   2,   0,   1,  -1,   0,   0,   0,   0,   2,   0,   1,   0 },
     {   15,  15,  33,   8, -37,  15, -13,   3,  -5,   2,  -2,   0,   6,  -6,  -3,   1,   0,   0,  -1,   0,  -1,   1,  -1,   0,   8,   2,   1,  -9,  10,  -8,   4,  -3,   2,  -2,   1,  -1,  33,   6,  27,  -4, -16,   2,  -4,   0,  -2,   0,  -1,   0,  69,  -1,   2, -49,   0,  -3,   0,   0, -25,   5, -41,  -2,  16,  -1,   5,   0,   3, -12,  -6,  13,   2,  -1,   0,   1, -10,  -1,  -1,   2,   2,   0,   0,   0,   1,  -3,   1,   3,  -5,   0,   0,   0,  -1,  -1,   0,   2,  -3,   0,   0,   0 },
     {   -6,  -8, -23, -19,  12,  -1,   6,  -1,   3,   0,   1,   0,  -3,  -2, -26, -34,  55, -29,  19,  -8,   7,  -4,   3,  -2,  17,  30,  41,  19, -40,  10,  -9,   0,  -4,   0,  -2,  -1,  -8,   7,   1, -15,  -1,   5,  -3,   2,  -2,   1,  -1,   0,  -2,   6,  27, -11,   0,  -5,   1,  -2,   0,  28, -27, -19,   0,   4,   0,   1,   2, -10, -15,  14,   8,  -1,   1,   0,   2,   2,  -1,   5,  -3,   0,  -2,   0,   0,  -5,  -2,   5,   0,   0,   1,   1,   0,  -3,  -2,   2,   0,   0,   0,   1 },
     {    2,   5,  -2,  19,  -5,  -4,   2,  -1,   0,  -1,   0,  -1,  -2,   5,  -5,  11, -18,  15,  -8,   5,  -2,   1,  -1,   1,   5,  -5, -38,  39,  -5,   1,  -1,   2,   0,  -1,   1,   0,  21,  68, -16,  -5, -17,  -2,  -2,  -1,  -2,   0,  -1,   0, -52,  16,  26, -30,   9,  -5,   3,   0,  14,  12, -31,   3,   3,   0,   3,  -1,   0, -19,   7,   3,  -1,   2,   0,   1,   2,   1,  -1,  -2,   1,   1,   1,   0,   1, -10,   3,   1,   1,   0,   1,  -2,   0,  -5,   2,   1,   0,   1,   1,  -2 },
     {    7,  14,  18,  14,  -3,  -7,  -1,  -2,  -1,   0,  -1,  -1,  13,  25,  45,  16, -38,  12, -11,   0,  -3,   0,  -2,  -1,   4,  18,  15,  12, -23,  10, -10,   3,  -4,   1,  -2,   0,  -5,  -5,  26, -48,  18,  -8,   4,  -1,   1,   1,   0,   0, -10, -29,  18,   8,   0,   1,   0,   0, -16,  34,  15, -35,   2,  -2,   1,  -2,  -2,   7, -34,  -3,  14,  -2,   3,   0,  -1,  -1,  -2,   6,   0,  -2,   0,   0,   1,  -1,  -9,  -1,   0,  -2,   0,   1,   0,  -1,  -4,  -1,   0,  -1,   0,   1 },
     {    2,  -1,  -2,   1,  -2,   0,  -2,   0,   0,   1,  -1,   0,   6,   3,   0,   5,  -5,   2,  -2,  -2,   1,  -2,   0,   0,   4,  -7,  -9,   4,   0,   0,  -1,   2,  -2,   1,   0,   0,   9,  13, -12,  24,   4,  -9,   4,  -4,   2,  -1,   1,  -1,  33,  44,  21,  11, -40,   7,  -8,  -1,  50, -24,  11, -62,   2,   4,  -3,   2, -23,   7, -32,   2,  24,  -4,   4,  -1,   6, -11,   4,   9,  -5,  -3,  -1,   0,  -8,   1,  -3,  -4,   1,  -4,   2,   1,  -4,   1,  -2,  -1,   1,  -2,   0,   1 },
     {   -6,  -5, -19, -18,   5,   2,   4,   2,   2,   0,   1,   0, -11, -12, -29, -31,   0,  14,  -2,   7,   0,   2,   1,   1,  -5,  -3, -37, -32,  42, -17,  15,  -4,   5,  -1,   2,  -1,  16,  16,  14, -20,  10,  -8,   4,  -5,   2,  -2,   0,  -1, -12,  20, -11,  21,  -5,   3,  -1,   0, -38,  46,  12, -21,   4,  -6,   2,  -2,  12,   3, -29,  18,   5,  -2,   1,   0,   1,  -4,  10,  12,  -4,  -2,   0,   0,   7,  -2,  -4,   5,   0,  -4,   4,   4,   0,   0,  -2,   3,   0,  -2,   3,   3 },
     {   -3,  -6,  -5,   1,  -5,  10,  -2,   4,  -1,   1,   0,   0,   1,  -2,   4,   7, -10,   4,   1,  -2,   3,  -1,   0,  -1, -15, -24, -24,   0, -25,  27,  -7,  10,  -2,   4,  -1,   2,  -8, -25, -46, -22,  57,  -9,   9,  -1,   3,   2,   0,   1,  16,   5,  -9,  13,   4,  -5,   4,  -4,   1,   3, -42,   7,  24,  -2,   1,   0,   0, -15,   4,  47,  -3,  -6,  -1,  -1,   8,   0,  22,  -2, -14,   3,  -2,   0,   3,  -2,   4,   1,   1,   0,   7,  -3,   0,   2,   2,   1,   2,   1,   4,  -1 },
     {    7,   8,  18,  40,  18, -45,  20, -13,   4,  -5,   2,  -3, -11, -21, -27, -53,  -9,  38,  -7,   9,   1,   3,   0,   2,   4,   4,   9,  -1,  18, -10,   0,   1,  -3,   1,  -1,   0, -15, -15,   9, -17,   1,  -5,   7,  -3,   4,   0,   0,   1,   3, -30,   5,  -5,  10,   7,  -3,   2,  37, -20, -17,   6,   4,  -3,   2,  -1, -17,  -6,  -5,  12,  10,  -1,   2,   0,   4,   4,   5,  12,  -6,  -4,   0,  -1,  -5,  -2,   1,  -1,   2,   4,   3,   3,  -1,   1,  -2,   0,   1,   3,   1,   2 },
     {   -7, -14, -16, -48, -13,  36, -13,  13,  -4,   6,  -2,   4,   2,   0,   9,  -2,  -8, -11,   6,  -3,   0,   0,  -1,   0,   1,  -8, -15, -16,  23, -13,  11,  -3,   4,  -2,   2,  -1,  -3,   7,  26, -39,   8,   4,  -2,   0,   0,   1,   0,   0,   4, -37,  27,   6,   9,  -2,   0,   0,  67,   0, -14,   7, -11,  -2,  -3,   0,  -8,  -7,  10,  -5,  -1,   0,  -1,   0,   0,  10, -10,  -1,   6,  -3,   1,   1,  -3,  -2,   1,   3,  -1,   4,  -2,   2,   0,  -1,  -1,   3,  -1,   3,  -1,   0 },
     {   -7,  -6,  -7, -14,   5,   8,  -4,   3,  -3,   4,  -1,   2,  -3,   3,   8,  13,  10, -15,   5,  -5,   3,  -2,   1,  -1,  -4,   0,   3,   1,  15,  -5,   0,   1,  -1,   0,   0,   1,  -8,  10,  19,   5, -13,  -5,   0,  -1,  -1,  -1,   1,   0, -36, -34,   3, -26, -12,  18,  -6,   8, -25, -48,   2, -12,  47,  -5,   9,  -1, -28,  27,  -5,  51,   4, -11,   2,  -4,  11,  -3,  17,   1, -17,  -1,  -2,  -2,  -3,   9,  -2,  -1,   5,  -1,   4,  -2,  -1,   4,  -2,   0,   2,   0,   1,   0 },
     {    6,   5,   7,  20,  16, -16,  -2,  -3,   0,  -1,   1,  -2,  11,  10,  14,  37,  27, -23,   0, -11,   3,  -1,  -1,  -2,  10,  10,  11,  38,  48, -37,   1,  -9,   2,  -3,   0,  -2,   4,  -6,  11,   6,   1, -12,  -1,  -3,   1,   0,   1,  -1,   4,  -2, -14,  33, -12,   5,  -1,  -2,  23,  25, -18,   6,   0,  -9,  -2,  -2,  23, -18,  -7,  27, -23,  -2,  -4,   1,   6,  -4,  10, -13, -22,   5,  -2,   1,   0,  -7,  -4,  -8,   0,  -1,   2,  -4,   0,  -2,  -1,  -4,  -2,   0,   3,  -2 },
     {   -9, -12, -15, -40, -35,  24,  -3,  11,  -1,   4,   0,   3,  -3,  -1,   2, -23, -35,  39, -15,  12,  -4,   4,  -1,   2,   7,  16,  35,  63,  15, -24,  -2,  -6,   1,  -5,   0,  -2, -10, -19, -16,  15,  -7,   0,   2,   1,   0,   2,   0,   1,  -1,   8, -17,   5,   3,   2,  -1,   0, -13,  -2,  -2,  -2,  -2,  -1,   3,   1,  -3, -15,   5,   6,  14,   1,   1,   1,   0,   5,   8,   6,  -5,  -6,  -1,  -1,   1,  -3,   1,  -6,  -1,   3,   3,   1,   1,   0,   1,  -2,   0,   1,   0,   0 },
     {    2,   3,   4,  -5,  -8,  -1,  -4,   4,  -4,   3,  -1,   0,  -1,  -2,  -6, -31, -19,  11,   3,   3,   1,   0,   0,   1,   9,  15,  15,   4,  -5,  17,  -4,  -3,  -2,  -2,  -2,   0,  12,  22,  22,  14,  37, -33,   9, -11,   1,  -4,   1,  -3,  10,  11,  38,   4, -24,   3,  -6,  -1,   2,   5,  21,  25,  15,  -8,   0,  -5,  14,  28,  29,   6, -41,  -9,  -5,  -3,   7,  -8,  -3, -39, -15,  10,   0,   3,  -9,   1,  -8, -10,   1,  -4,  -8,  -6,  -4,  -1,  -4,   1,   1,  -3,  -2,  -3 },
     {   -1,   3,  -2,   2,   8,   0,   1,  -3,   2,  -1,   1,  -1,   2,   9,  -1,  13,  20,   5, -11,  -5,   2,  -2,   0,   1,   0,   4,  -2,  17,  26, -20,  -1,  -4,  -1,  -2,   0,   0,  -5,   1, -30, -26,  14,   7,  -1,   2,  -1,   2,   2,   0,  24,  20,  19, -12,  31, -16,   6,  -6, -14,  14,   5,  18, -11, -14,   7,  -4, -62,  56,   4, -10,  -5,  -5,   4,  -4,  24,  -3, -15,   7,  -9,   2,   0,   0,  -4,  -4,   2,  -4,   4,  -1,  -9,   6,  -4,  -5,   2,   0,   2,  -3,   0,   3 },
     {    2,  -1,  -5,  -6,  -6,   7,  -5,   3,  -3,   1,   1,  -1,   2,  -2,  -8,  -2, -12,   6,   5,  -2,   6,  -3,   2,  -3,   4,   0,  -2,  -1,  -9,  15,  -9,   8,  -6,   5,  -2,   1,  -5,  -9, -17,  18, -33,  -5,   9,   0,   0,  -2,   1,   0,   0, -15, -41, -26,  17,  13,  10,  -1,  46,  46,   2,   3,  36, -16,  -2,  -5,  19,  45, -29,  -4, -13,  -5,  -1,  -3,  12, -24,  -3,  -1,   4,   9,   0,   1,   1,   1,  -2,  -1,  -1,  -5,   0,   5,   0,   2,  -1,   1,   0,  -4,  -1,   2 },
     {   -3,   0,   0,   0, -16,  -4,  12,  -7,   0,   3,   1,   0,  -2,   3,   3,   0,  -5,   9,   6,  -5,  -2,  -2,   1,   1,  -3,   0,  -2, -15, -17,  17,  -4,   7,  -6,   2,   0,   1,   0,  10,  30,  29, -21,   4,  -8,  -1,  -2,   1,   0,  -1,   1, -12, -27,  38, -12,  -1,   5,  -2,  -2,  40, -23, -11,  -5,  -4,   7,   0, -73, -16,  33,   5,  -3,   3,   2,  -3,  20,   1,   3, -15,  -5,   8,  -4,   0,  -6,  -3,   3,  -8,   7,   0,   1,   4,   1,  -2,   1,   0,  -2,   3,  -1,  -3 },
     {   -1,  -3,  -2,  -8, -24, -21,  21,  -2,   1,   1,   4,  -3,   5,   4,   8,   1,   3,  15, -15,   6,  -1,   1,  -1,   2,  -1,  -4,  -1, -17, -34, -25,  25,  -2,   4,  -1,   2,  -1,   3,   2,   2, -10, -25,  49, -22,   8,  -2,   4,  -1,   2,   7,  13,  12,  12,  29, -29,   5,   1,  13,  -9,  22, -12,   4,   0,   6,  -7,  22,   9,  -1,  30, -17, -12,  -2,   1,   5,   8,  23, -21, -31,  12,   3,  -2,  -1,  -2,  -6, -19,  -2,   2,  -3,  -3,   0,  -2,  -6,  -3,  -4,   0,   1,   0 },
     {    6,   8,   9,   9,  51,  28, -53,  26, -12,   5,  -7,   2,   0,   0,  -1,  -7, -12,  14,   3,  -3,   6,   3,  -3,   2,   4,   6,   9,  -4,   1,  24,  -9,  -2,   2,  -1,   2,  -2,   7,  13,  19,  32,   0,  16,  -5,   3,  -5,  -2,   0,   0,   2,   3,   6,  16,  37, -25,   2,  -6,   2,  -1,   4,  14,  -8,   1, -11,   0,  -3,  -3, -14,  14,  17, -16,   2,  -2,  -2,   1,  13,  19, -30,  -5,  -1,  -1,   3, -15,   5, -11,  -8,   4,   7,   2,   4,  -3,  -1,  -7,  -3,   2,   1,  -1 },
     {    2,   2,   4,   9,  16, -13, -27,  18, -12,   4,  -1,   0,  -3,  -6,  -8, -11, -41,  -3,  53, -32,  16,  -2,   3,  -3,   0,  -2,   1,   7,   0, -16,   3,  -5,   1,  -3,   2,  -2,  -5,  -8, -17, -33, -20,  35, -31,  11,  -5,   3,  -1,   1,   4,  14,   4,  11, -33,  14,   3,  -4, -10,  -1,   1,  -3,  26,  -3,  -7,   3,   1,  -1,  16,  -4,  -8,  13,  -9,  -1,  19,   0, -19,   0,   3,   5,  -4,   0,   7, -11,   2,   4,   1,  -1,  -3,   0,   1,   0,   1,  -1,   2,   1,   0,  -1 },
   }
 },
 {  //8
   {
     {  108, -52,  -3,  -5,  -1,  -2,   0,  -1,   0,  -1,   0,   0, -41,   8,   6,   1,   2,   0,   1,   0,   0,   0,   0,   0,  -5,   9,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0,  -2,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   40,   4, -13,  -1,  -3,   0,  -2,   0,  -1,   0,   0,   0,  95, -51,  -8,  -1,  -2,   0,  -1,   0,   0,   0,   0,   0, -42, -13,  17,   2,   3,   1,   1,   0,   1,   0,   0,   0, -14,  15,   4,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,  -9,   4,  -1,  -1,   0,   0,   0,   0,  -3,   4,  -1,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0 },
     {   13, -11,   2,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,  35,  40, -28,  -1,  -6,  -1,  -2,  -1,  -1,   0,  -1,   0,  71, -60, -17,   8,  -2,   3,  -1,   2,   0,   1,   0,   1, -36, -26,  30,   4,   3,   1,   1,   0,   1,   0,   0,   0,  -8,  16,   8,  -6,   0,  -1,   0,  -1,  -9,   3,  -2,  -2,   1,   0,   0,   0,  -3,   4,   0,   0,   0,   0,   0,   0,  -4,   0,   1,   0,   1,   0,   0,   0,  -1,   2,   0,   0,  -2,   0,   0,   0,  -1,   1,   0,   0,  -2,   0,   0,   0 },
     {  -38, -90,  55,   1,  12,   1,   4,   1,   2,   1,   1,   1,  13,  -7,  18,  -7,   1,  -2,   0,  -1,   0,   0,   0,   0,  -2,   3,   8,  -5,   1,  -1,   0,  -1,   0,   0,   0,   0, -27,  35,   4,  -7,   2,  -2,   1,  -1,   0,  -1,   0,   0,   5,  21, -20,  -2,  -1,  -1,   0,   0,   1,   0,  -8,   5,   0,   1,   0,   0,   4,   1,  -3,   2,  -1,   0,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   2,   2,  -2,   1,   0,   0,  -1,   1,   1,   1,  -1,   0,   0,   0,  -1,   0 },
     {    3,  32, -12,  -1,  -3,  -1,  -1,  -1,  -1,   0,   0,   0, -40,  10, -12,   7,   0,   3,   1,   1,   0,   1,   0,   0, -31, -55,  37,   9,   4,   3,   1,   1,   1,   1,   1,   0, -28,  51,  31, -18,   0,  -5,   0,  -2,   0,  -1,   0,  -1,  22,  26, -34, -13,   1,  -2,   1,  -1,   3, -13, -13,  12,   3,   2,   1,   1,   8,  -3,   4,   4,  -3,   0,  -1,   0,   3,  -2,   1,  -3,  -1,   0,   0,   0,   3,  -1,  -1,   0,   1,  -1,   0,   0,   2,   0,   0,   0,   1,  -1,   0,   0 },
     {    9,  -3,   9,  -7,   2,  -3,   1,  -1,   0,  -1,   0,   0,  29,  64, -39,  -7,  -6,  -3,  -2,  -1,  -1,  -1,  -1,  -1,  -9,   7,  -7,   0,   1,   0,   0,   0,   0,   0,   0,   0,  11,  35, -37,  -4,  -1,  -1,   0,  -1,   0,   0,   0,   0,  20, -48, -33,  25,   1,   5,   0,   2,  -4, -26,  29,  17,  -5,   2,  -2,   1,   2,  -3,  13,  -9,  -5,   1,  -1,   0,  -5,  -3,   3,  -3,   4,   1,   1,   0,   0,  -1,   1,   1,  -2,  -2,   3,   0,   0,  -1,   1,   0,  -2,  -1,   2,   0 },
     {  -10, -11,   9,  -1,   2,  -1,   1,  -1,   1,   0,   0,   0,   7,  55, -20,  -6,  -3,  -2,   0,  -2,   0,  -1,   0,  -1, -81,   6, -26,  18,   3,   5,   2,   2,   1,   1,   1,   1,  -4, -32,  26,  22,  -4,   3,  -1,   1,  -1,   1,   0,   1,  12,  20,  32, -17,  -8,  -2,  -2,  -1,   9,   6, -14, -17,   3,   0,   1,   0,   2,  -3,  -6,   4,   5,   0,   1,   0,   5,  -1,   0,   2,  -1,  -1,   0,   0,   2,  -2,   0,  -1,   2,  -1,  -1,   0,   1,  -1,   0,   0,   1,  -1,   0,   0 },
     {  -12, -31,   8,  12,   1,   2,   1,   1,   1,   0,   0,   0, -17, -19, -29,  22,  -2,   6,   0,   3,   0,   2,   0,   1, -12, -51,  34,  16,  -1,   5,  -1,   2,   0,   1,   0,   1,  39,  -1,   8,  -1,  -5,  -1,  -2,   0,  -1,   0,  -1,   0,   8, -41,  33,   6,  -3,   1,  -1,   1, -10,  25,  40, -25,  -3,  -3,  -1,  -2,  -3,  19, -15, -22,   5,  -1,   2,  -1,  -5,   8,  -8,   1,   7,  -1,   1,  -1,   1,   5,  -4,   1,  -1,   3,  -1,  -1,   1,   2,  -2,   0,   0,   2,   0,  -1 },
     {    2,  24,  14, -11,   0,  -3,   0,  -1,   0,  -1,   0,  -1, -15,  -3,   2,  -1,   4,   0,   1,   0,   1,   0,   0,   0,  -5,   4,   1, -10,   4,  -3,   2,  -1,   1,  -1,   0,   0, -87,  16, -34,  20,   8,   3,   3,   1,   1,   1,   1,   0,   1, -11,  31,  34, -10,   2,  -4,   1,  13,  26,  23, -28, -14,  -2,  -3,  -1,   7,  -2, -22, -13,   7,   3,   2,   1,   1,  -1,  -2,   8,   5,  -1,   0,   0,   5,  -1,  -1,   2,   2,  -1,   1,   0,   2,  -1,  -1,   0,   1,   0,   0,   0 },
     {  -22, -39, -87,  46,  -1,  14,   3,   7,   1,   3,   1,   2,   0,   7,  -3,  21,  -6,   3,  -2,   1,   0,   0,   0,   0,  -5,  20,  14,  -7,  -5,  -1,  -2,  -1,   0,   0,   0,   0, -34, -13,   6,  -1,   5,   3,   2,   0,   0,   0,   0,   0,   3,   3,  -1,  12,   2,  -1,   1,  -1,   9, -12,  18,  12,  -9,  -1,  -1,   0,   4,   4,  17, -16,  -7,   1,   0,   0,   0,   0,  -1, -10,   5,   2,   0,   0,   1,   2,   1,  -1,   1,   0,   0,  -1,   0,   1,   2,  -1,   0,   0,   0,  -1 },
     {   -1,  -6, -22,   9,  -2,   2,  -1,   1,   0,   0,   0,   1,  12,  32, -19,  11,  -8,   2,  -3,   0,  -2,   0,  -1,   0,  15,  23,  44, -23,  -7,  -4,  -3,  -2,  -1,  -1,  -1,  -1,   8,  21, -18, -26,   3,  -1,   1,   0,   0,   0,   0,   0,   3,  -3,  12, -17,   3,  -1,   0,  -1, -13,  29, -40, -32,  16,  -3,   5,  -1,  -7, -15, -43,  35,  18,   1,   4,   0,   2,  -7,  12,  24, -10,  -2,  -2,   0,   2,  -8,   2,   0,   0,  -2,   4,   2,   0,  -3,   0,   1,  -1,  -1,   2,   1 },
     {    2,  10,   0,   2,  -4,   1,  -2,   0,  -1,   0,   0,   0,   9,  42,  53, -30,  -6,  -6,  -3,  -4,  -1,  -1,  -1,  -1, -16, -10,  26, -21,   2,   0,   1,   0,   0,   1,   0,   0,  11, -15,  11, -30,  10,  -3,   3,  -1,   1,  -1,   0,  -1, -45,   9, -18,   9,  14,  -2,   4,  -1,  -2,  22,  20,   6, -13,  -3,  -2,  -1,   1,  21, -24, -34,   0,   1,   2,   0,  -2,  -7, -29,  12,  17,   3,   3,   1,   2,   1,  -3,  14,   1,  -3,  -3,   4,   1,   0,  -1,   2,   0,  -2,  -2,   2 },
     {   -6,   0,  15,  -4,   4,  -3,   1,  -2,   0,   0,   0,  -1, -15,  -6, -60,  18,   4,   5,   3,   2,   2,   0,   1,   1,  -9,  18,   3,  16,  -2,  -1,   0,  -1,   0,  -1,   0,   0,  -9,  -2,   4, -15,   1,  -1,   2,  -1,   1,  -1,   1,   0, -84,   2,  -7,  10,  19,   1,   5,   0,   6,  29,   0,  31,  -3,  -4,  -2,  -2,  16,   0,   7,  11, -20,  -1,  -2,   0,  10,  -4,   9,  -6, -10,   5,  -2,   1,   3,  -5,   5,  -8,   3,   1,   1,  -2,   1,  -1,   1,  -2,   1,   1,   1,  -1 },
     {    6,   9,   1,   7,  -4,   1,  -2,   0,  -1,   0,   0,   0,   4,  -8, -26,   9,  -2,   2,   1,   0,   1,   0,   0,   0,  21,  44, -24,  27, -13,  -1,  -4,  -1,  -2,  -1,  -1,   0,   6,  38,  40, -12, -21,  -4,  -5,  -2,  -2,  -1,  -1,  -1,  -3,  16,  -7, -21, -11,   2,  -1,   1,  24, -10,  16, -26, -12,   6,  -2,   2,  -2,  -4, -25, -43,  19,   5,   4,   2,  -6, -12, -24,  18,  27,  -2,   3,  -1,  -1,   0,   2,  16,   0,  -5,  -1,   2,  -1,  -2,   1,   1,   0,  -3,   0,   1 },
     {    1,  -3, -11,  62, -20,   6,  -7,   0,  -3,   0,  -1,   1,  -3,   4,  18,  -9,  -7,  -1,   0,  -1,   0,   0,   0,   0,  -5, -35, -52,   2,  18,   3,   7,   1,   3,   0,   1,   0,  24,  48, -21,  21,   1,  -4,  -2,  -3,   0,  -2,  -1,  -1, -35,  13,   6,  12, -12,  -2,  -1,  -1,  -6,   7,  -2,  -9,  -9,   5,   0,   1,   6, -12, -11,  10,   4,   4,  -1,   1,   6, -11,  17,   9,  -8,   0,  -3,   0,   2,  -3,   8, -11,   0,  -2,   3,  -3,   0,   0,   1,  -2,   0,  -1,   2,   0 },
     {    0,   9,   5,  23, -11,   2,  -2,   0,  -1,  -2,   0,  -1,  -4,   1,  16,  -9,  -5,   3,   1,   1,   0,  -1,   0,  -1,  -4,  19, -19,   6,   0,   1,   2,   0,   0,  -1,   0,  -1, -35,  15,  27,  -2, -12,   3,  -3,   0,  -1,   1,  -1,   0, -10, -52,   1, -40,   8,   7,   2,   4, -55,  -2,  12,   2,  37,   0,   5,  -1,   7,  43,   3,  18,   4, -15,  -1,  -5,   3,  10, -13,  -1, -15,  -4,   1,   0,   4,  -2,  -6,   0,   2,  -2,   0,   3,   2,   0,   0,   0,   0,  -1,  -1,   0 },
     {   11,  16,  31,  24, -15,  -4,  -5,  -3,  -2,  -1,  -1,  -1,  12,  22,  18,  47, -26,  -1,  -8,   0,  -3,  -1,  -2,   0,   9,  13,  26, -19,  -6,  -5,  -2,  -1,   0,  -1,   0,  -1,   6,  19,   0,  18, -10,  -2,  -2,  -1,  -1,  -1,  -1,   0,  -2,   5,  33, -21, -14,  -2,  -1,  -1,  -2,  12, -14,   5,  -8,  -3,   0,  -1,  18, -12,  33, -20, -31,   6,  -5,   2,  -5,   2, -13, -47,  12,   8,   2,   2,  -2,  -4, -12,   2,  -3,   1,  -6,  -4,   0,  -4,  -3,   1,  -1,   0,  -2,  -3 },
     {  -10, -10, -46, -54,  36,  -4,  12,   1,   4,  -1,   2,   0,   0,   8,  -1, -25,  25,  -8,   5,  -3,   1,  -1,   1,   0,   8,   1,   4,   8,  12,  -7,   2,  -3,   1,  -1,   1,  -1,  14,  38,  -8,  40, -17,  -7,  -2,  -3,  -1,  -1,  -1,  -1, -15,  16,  19, -10, -20,   1,  -1,  -1, -36,  19,  -5,   5,   1,   3,   1,   0,  12,  10,   7,  -7, -14,   3,  -3,   1,   0,  -2, -19, -12,   8,   3,   2,   1,   2,  -6,  -3,  13,   0,  -2,   0,   1,   1,  -3,   1,   2,  -2,  -1,   0,   0 },
     {    1,   3,  -4, -11,   8,  -3,   2,  -1,   0,   1,   1,   0,  11,  21,  46,  14,  10, -12,  -1,  -5,  -1,  -3,  -1,  -1,   6,   4,  26,  70, -30,  -6,  -7,  -2,  -3,  -1,  -2,   0,  -2,   6,   5,  -6, -24,   2,  -2,   0,  -1,   1,  -1,   0, -20,  -5,   2,  13, -16,   4,  -2,   2,  15, -15,  17, -19,  -7,   4,  -3,   1,  -8,  -1,  -2,  15,   4,  -4,   0,  -1,  12,   0,  29, -11, -29,   2,  -4,   0,  -4,   3,  -3, -34,   0,   1,  -4,  -5,  -1,   1,  -5,  -7,   0,  -1,  -2,  -1 },
     {    4,   3,  -6,  -4,   3,   1,   5,  -3,   1,   0,   1,   0,   8,  -3,  14,  25,  -5,  -1,  -6,   0,   0,   0,  -2,   1,   2,   3, -33,  30, -11,   4,  -1,   2,   0,   1,   0,   0,   4,   2,  -3, -18,   2,   6,   2,  -1,   1,  -1,   0,   1,  53,   9,  -8,  10,  27,  -5,  -1,  -3, -14,  53,   3,  24, -11, -22,  -1,  -7,  22,  17, -28,  -1, -37,   1,   1,   0,   4, -33,  -4,   2, -12,  15,  -1,   5,   1, -11,  11,  -7,  -4,  -7,  -1,   0,  -1,  -1,   2,  -1,  -3,  -1,   2,   0 },
     {    7,  10,  13,  29,  -8,  -3,  -4,   1,  -3,  -1,  -1,   0,   1,   0, -24, -48,  15,   0,   7,   1,   1,  -1,   1,   1,  14,  31,  29, -10,  26,  -8,  -5,  -3,  -1,  -1,  -1,  -1,  11,   2,  40,  16,  18, -18,  -2,  -5,   1,  -2,   0,  -2,  24,  20,   0,  29, -18, -17,  -1,  -3,  -7,  10,  22,  -6, -19,  -3,   0,  -2,  -6,  21,  -3,  11, -15,  -4,  -1,  -1,   7,  -4,   7, -18, -30,   5,  -3,   2,  -3,  -5, -10, -24,  -1,  -2,  -6,   1,  -2,  -2,  -5,  -3,   0,  -2,  -1,   1 },
     {    2,  -6, -10,   8,   0,   6,  -3,   2,  -1,   1,  -1,   0,   2,  -2,   8, -49,   8,  -3,   4,  -1,   0,   0,   1,  -1,   9,  -7,   7,   8,  10,  -4,   2,  -1,   0,   0,   0,   0,  -3,   3,  16,  12, -16,  -2,  -5,   0,  -1,   1,  -1,   0,   6, -54,  -2, -28,   8,   5,   2,   4,  70,  34, -17,  13,   8,  -5,  -2,  -2,  27, -26,  -1,   2, -14,   0,  -4,   0, -14,  -8,  10,  -8,   5,   1,   0,   1,  -4,   2,   0,   1,  -4,   4,   1,  -2,   1,   2,  -4,   0,   0,  -1,  -1,   2 },
     {   -2,  -3,  -6,   1, -39,  10,   1,   3,   3,   1,  -1,   0,  -3,   0, -13, -46,  -8,  13,   6,   2,   3,   1,   2,   1,   3,  -2,  -4,  29, -17,   3,  -5,  -2,  -2,   0,   0,   0,   1,  11, -40, -37,  -5,  17,   1,   3,   0,   1,   0,  -1,   7,  34,  23,  -6,  26,   7,  -6,   0,  10,   8,   6, -12,  15, -11,  -3,  -2, -11,   9,  22,   1,  -2,  -8,   5,   0,  -2,  26, -13, -30,   0,  -1,   5,  -1,  -2,  -7, -34,   6,   3,  -5,   2,  17,  -3,  -6,  -3,   5,   0,   0,   4,   3 },
     {    2,   4,  13,  42,  -6, -13,  -4,  -1,   0,  -2,  -1,   0,  -2,  -1,  -8,  -7,  22,   1,   2,   1,   0,  -1,   0,   0,   0,   3,  27,  39,  15, -11,  -1,  -5,  -1,  -3,  -1,   0, -12, -37, -37,  21, -43,  -3,   4,   1,   2,   1,   2,   1,   9,  14, -42, -15, -14,  13,   8,   3, -33,  23,  -7,   9,  -6,   9,  -1,   1,  -7, -13,  -2, -11,   7,  11,  -3,   2,  -5,   3,  -2,  10,  21,  -3,   0,  -2,   2,   0,   4,  11,  -2,   1,   5,   0,   1,   0,   2,   0,  -2,   0,   1,   0 },
     {   -4,   0,  -1,  -6,  -5,   6,   2,   3,   0,   1,   0,   0,  -5,  -2,   8, -20,  -1,   3,   2,   2,   1,   2,   0,   1,  -5,   8,  -7,  14,  -7,   5,  -1,  -1,   0,   0,   1,   0,  -8,   8,  32, -19,   9,  -1,  -3,  -1,  -3,   1,  -1,   0,   1, -23,  18,  17,  21, -16,   1,  -4, -53,   5, -17,   7, -35,  -9,   4,   2, -19, -58,   5, -15,  -3,  25,   5,   7, -16,  20,  32,  -3,  23,   3,  -2,  -3,   7,  18,  -8,  -2,  -3,   2,  -2,   5,   3,   3,  -2,   0,  -2,   3,   0,  -1 },
     {   -7, -11,  -5, -38, -65,  28,  -1,  10,   4,   4,   1,   2,  -2,  -7,   0, -11, -47,  29,   0,   3,   2,   2,  -1,   2,   1,   4,  14,  14,  -5,  13,  -5,  -3,  -1,  -2,   1,   0,   0, -14,   3,  23,  10,  -9,   1,  -3,   3,  -1,   0,   0,  -5,  -4, -20,   9, -23,   2,   5,  -1, -16,  -4,  -3, -13,   1,  16,   3,   3,  40,  -2,  -5,   3,   6,   6,  -4,  -1,   6, -30,   7,  -4,   9,   5,  -2,   3,  -2,  -1,  12,   0,  -1,   1,   2,  -2,   0,   4,  -3,   0,  -2,   1,  -1,   1 },
     {    2,   2,   1,  22,  24, -24,  -1, -10,   4,   6,   1,   2,  -4, -11, -12, -24,  10, -22,   6,   2,   8,   1,  -3,  -2,  -5,  -7,  -4, -20, -62,  29,   8,   6,   3,   1,   0,   1,   0, -10,   3, -12, -11,  31,   1,  -9,  -2,  -3,   3,   0,  -2,  -7, -12,  19, -20,   7,  -3,   2, -14,  -9, -18, -25,  -3,  10,   2,   7,  33,   4,  -8,   6,  -5,   9,  -5,   3,   9, -26,   0, -32,  11,  17,   0,   4,  -6,   5, -10,  11,   2,  -1,   3,   7,  -5,   2,  -3,   4,   1,   1,   5,   2 },
     {   -3,  -5,   0,  -6, -45,  25,  -9,  14,  -3,   1,  -4,   1,   4,   6,   0,  24,  77, -11,   0,  -9,  -3,  -2,   4,   5,  -4,  -6, -14,   1,   9, -25,  -4,   0,   3,   4,   1,   1,  -3,  -2,   3, -25,   2, -16,   3,   2,   6,  -1,  -1,  -1,  -2, -18,   2,  20, -15, -10,   4,  -2,   4,  -6, -34, -12,  -5,  11,  11,   3,   1,  19,  -2,   3,  -8,  10,   3,  -3,  -4,  -2, -17, -17,   3,  11,   0,   2, -11,   0,  -3,  21,  -2,   0,   6,   3,  -1,   5,   0,  -3,   0,   4,   2,  -1 },
     {   -3,   0,  -6,   0,  -3,  -7,  13,  -7,   5,   1,   1,   3,  -6,  -1,   0,   5, -34,   3,  -4,   1,   3,   2,   0,   4,  -2,   0,  -6,   9, -19,  12,  -1,   0,   2,   0,   0,   1,  -3,   7,   8,  34,   1, -10,  -2, -10,   3,  -6,   2,  -2,  -7, -17, -29,  25, -15,  -2,   3,  -2,  20,  19, -35,  -3,   3,  -6,   8,  -4, -68,  25,  -8,   7, -20,   6,   9,  -3, -18,  21, -18,  -5,  -2,  18,   6,   6,   5,   8, -11,  11,   6, -17,   2,  13,  -2,   2,   9,  -4,  -2,   5,   5,  -5 },
     {   -1,   0,  -9, -14,   2,  -9,  -4,   1,   4,  -2,  -1,  -1,   1,   1,   2,   9, -17,   6, -16,  -6,   9,   3,   5,   5,  -2,   1, -22, -26,  19, -14,  -1,   1,   4,  -2,   0,   3,   4,   0,  19, -31, -43,   3,  20,   2,   3,  -2,   1,  -1,   9,   3, -13,  28, -26,  31,   6,  -4,   4,  41,  -5,  -5,   1,  17,  -3,   4,  17,  14,  24,  -2,  24,   8,  -6,   0,  -3,  27,  21,   3,   1, -17,   0,   0,   4,   8, -20, -25,  -9,  -5, -17,  14,  -6,  -3,   1,   4,  -7,   3,  -5,   1 },
     {   -1,  -3,   0,  -9,  -9,  -6,  11,  17,   0,   8,  -4,   4,   0,  -2,  -2,  -1,   8,  12,  -3,  12, -15,  19,   3,   9,  -4,  -6,  -6, -27,   8, -30,   0,  -3,  -1,  12,   0,   1,  -2,  -6,   2,  -1, -24,   0, -16, -12,   6,  -7,   3,  -3,  -3,   2,  -6, -11,   5,  27,  -8,   1,  -5,  -8,  12, -11,  31, -23, -15,   2, -24, -14, -13, -28, -12,  15,  17,  11,  23, -27,  24, -21,  -9,  33,   4,  11,  14,  10,   2, -18,  28, -20,  -4,   1,  19,   1,   2,   3,  18,  -1,   7,   4 },
     {    1,  -2,  -1,   2,  -1,   2,  -9, -23,  -3,  19,   1,  14,   0,  -3,  -8, -12,  -3,  -6, -28,  -2,   0,  15,   1,  16,   3,  -3,   3,   4, -23, -30,   9,  -7,   6,   9,   1,  10,   2,   1,  -6,  16,   5,   5, -25, -14,  15, -11,   8,  -4,   5,  -7,  -2,   2,  29, -16,  -6, -10,  -1,  -2, -30,   0, -15,  22,   3,   4,  19,  14,   3, -29,  17,   4,   1,  12,  29,  31, -15,  11, -13, -10,  12,   7,   3, -20,   9, -25,  -4, -11, -10, -20,  22,   8,  -2, -19,   3,   6,  -5,   4 },
   },
   {
     {   72, -52,   8,  -4,   2,  -2,   1,  -1,   0,   0,   0,   0, -70,  46,  -2,   1,   0,   1,   0,   0,   0,   0,   0,   0,  31, -12,  -8,   3,  -2,   1,  -1,   0,   0,   0,   0,   0, -12,  -2,   9,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   4,   5,  -5,   0,   0,   0,   0,   0,  -3,  -3,   1,   1,   0,   0,   0,   0,   2,   2,   0,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -68,   7,  22,  -1,   4,   0,   2,   0,   1,   0,   0,   0,  -8,  46, -30,   6,  -5,   2,  -2,   1,  -1,   0,  -1,   0,  46, -55,  19,  -1,   1,   0,   0,   0,   0,   0,   0,   0, -28,  30,   0,  -7,   3,  -2,   1,  -1,   1,  -1,   0,   0,  18, -11,  -9,   7,  -2,   1,  -1,   1,  -6,   2,   8,  -3,   0,   0,   0,   0,   6,  -1,  -5,   1,   0,   0,   0,   0,  -2,   0,   3,   0,   0,   0,   0,   0,   3,  -1,  -2,   0,  -1,   0,   1,   0,   2,   0,  -1,   0,  -1,   0,   1,   0 },
     {   55, -31,  14, -10,   2,  -3,   1,  -1,   0,  -1,   0,   0,  63,  -8, -34,   7,  -6,   2,  -2,   1,  -1,   0,  -1,   0, -30, -30,  36,  -5,   4,  -1,   2,   0,   1,   0,   0,   0, -19,  43, -19,  -3,   1,  -1,   0,  -1,   0,   0,   0,   0,   6, -22,   2,   9,  -3,   2,  -1,   1,  -9,  11,   7,  -8,   1,  -1,   0,   0,  -1,  -2,  -6,   3,   1,   0,   0,   0,  -4,   4,   3,   0,  -1,   0,   0,   0,   0,  -1,  -2,   0,  -2,   2,   1,   0,   0,   0,  -1,   0,  -2,   1,   1,   0 },
     {   30,  26, -45,  16,  -9,   4,  -3,   1,  -2,   1,  -1,   0,  38, -21,  24, -17,   2,  -3,   0,  -1,   0,  -1,   0,   0,  44, -39, -19,  11,  -1,   1,   0,   1,   0,   0,   0,   0, -47,  -2,  39,  -7,   2,  -1,   1,  -1,   1,   0,   0,   0,   1,  26, -23,  -3,   2,  -2,   1,  -1,  -6, -15,   8,   9,  -3,   2,  -1,   1,  -2,  10,   1,  -8,   1,  -1,   0,   0,  -4,  -3,  -2,   4,   1,   0,   0,   0,  -1,   4,   1,  -2,  -2,  -2,  -1,   1,  -1,   2,   1,  -1,  -1,  -1,   0,   1 },
     {  -15, -45,  33,  -3,   6,   1,   2,   0,   1,   0,   0,   0,  52,  38, -28,  -9,  -3,  -3,  -1,  -2,  -1,  -1,   0,  -1,  -9,  -4, -26,  21,  -7,   5,  -2,   3,  -1,   1,  -1,   1,  25, -41,  38, -12,   2,  -1,   0,   0,   0,   0,   0,   0, -30,  31, -15,  -3,   5,  -3,   2,  -1,  11, -13,   2,   8,  -4,   2,  -1,   1,  -9,   6,   3,  -6,   1,  -1,   0,   0,   2,  -2,  -2,   3,   1,   0,   0,   0,  -4,   3,   2,  -1,   1,  -1,  -1,   1,  -2,   1,   1,   0,   1,  -1,   0,   0 },
     {   16,  17,  -3, -18,   5,  -4,   1,  -2,   0,  -1,   0,  -1,  25,   8, -26,  23, -12,   5,  -4,   2,  -2,   1,  -1,   0,  47,  -5,  -6, -22,   5,  -3,   1,  -1,   0,  -1,   0,   0,  -4, -48, -12,  34,  -4,   5,  -1,   2,   0,   1,   0,   1,  -7,  -3,  54, -27,   1,  -2,  -1,  -1, -16,  23, -36,   3,   9,  -4,   3,  -1,   4, -15,  17,  10,  -9,   3,  -2,   1,  -6,   9,  -4, -10,   4,  -1,   1,   0,   0,  -5,   2,   6,  -3,   3,   0,  -4,  -1,  -2,   1,   2,  -2,   2,   0,  -2 },
     {  -30, -67,  23,   5,  12,   1,   4,   1,   2,   0,   1,   1,  12, -27,  41,  -9,  -2,  -1,   0,  -1,   0,   0,   0,   0,   3,  22, -25,  -3,   6,  -3,   2,  -1,   1,  -1,   1,   0, -39,  18,   8,  16, -11,   5,  -3,   2,  -1,   1,  -1,   0,  36,  -1,   5, -26,   6,  -3,   1,  -1, -24,  15, -19,  16,   2,  -1,   2,  -1,  18,  -6,   8,  -4,  -7,   2,  -2,   1,  -9,   4,  -6,  -2,   6,  -1,   1,  -1,   6,   0,   0,   4,  -3,   1,  -1,  -1,   3,   0,   0,   1,  -2,   1,  -1,   0 },
     {    3,  -4, -16,  26, -12,   6,  -3,   2,  -2,   1,  -1,   1,  27,  35,  10, -41,  12,  -9,   2,  -4,   1,  -2,   0,  -1,  21,  19, -41,  22, -11,   1,  -2,   1,  -1,   0,   0,   0,  12,  -8, -23,   0,   1,   4,  -2,   2,  -1,   1,   0,   1,  17, -56,  21,  21,  -4,   3,  -1,   2, -16,  16,  24, -26,   4,  -3,   1,  -1,  -4,   0, -22,  14,   5,  -2,   2,  -1,  -2,  -4,  17,   1,  -8,   2,  -2,   0,  -2,   1,  -5,  -5,  -2,  -1,   5,   2,  -1,   1,  -2,  -2,  -1,  -1,   2,   1 },
     {    9,  31, -26,   7,  -1,  -1,  -1,  -1,   0,   0,   0,  -1,   3,  62,   6, -20,  -8,  -2,  -3,  -1,  -2,   0,  -1,   0, -49, -11, -13,   6,  13,  -2,   5,  -1,   3,   0,   1,   0,  10,  24,   9,   9, -20,   5,  -5,   1,  -2,   1,  -1,   0,  10, -14, -11, -32,  18,  -4,   3,   0,  -9,   5, -23,  37,  -6,   0,   1,  -1,  18, -23,  22, -14,  -8,   6,  -4,   2,  -7,   4, -10,   1,  10,  -3,   3,  -1,   6,  -7,   5,   3,  -3,   0,  -2,  -1,   2,  -3,   3,   2,  -1,   0,  -1,   0 },
     {  -13, -17, -18,  31, -10,   9,  -3,   4,  -1,   2,  -1,   1, -23, -11,   7, -15,  12,  -2,   4,  -1,   2,   0,   1,   0, -42, -57,  18,  25,  -2,   6,   1,   3,   1,   1,   1,   1,  -5, -17,  27,   3,   3,   1,  -1,   0,   0,   0,   0,   0, -13,  14,  48, -11,  -5,  -2,  -1,  -2,   4,  40, -14, -26,   8,  -5,   1,  -2,   2,   1, -15,  17,  -2,  -2,   1,  -1,   7,   5,   6,  -9,  -5,   2,  -2,   0,   0,   2,  -7,   0,   3,   2,   1,  -1,   0,   1,  -3,   0,   2,   1,   0,   0 },
     {    9,  37,  23, -36,   8,  -8,   2,  -4,   1,  -2,   0,  -1, -13,  -5, -30,  22,   0,   2,   1,   1,   1,   0,   1,   1, -38,  13, -28,  24, -13,   9,  -3,   3,  -1,   2,   0,   1, -37, -26,  40,  -6,   8,   0,   0,   1,   1,   1,   0,   1,  46,  -7,  -2,   4,  -5,  -1,  -1,   0, -29,   9,  16, -12,   3,  -1,   1,   0,  18,   6, -15,   4,   0,  -1,   0,   0,  -4,  -7,   9,   0,  -3,   1,  -1,   0,   6,   4,  -4,  -3,  -2,  -3,   2,   0,   3,   1,  -1,   0,  -1,  -1,   1,   0 },
     {   -8, -18, -19,   2,  17,  -5,   5,  -1,   2,   0,   2,   0, -13, -26,  -9,  20, -18,  15,  -4,   5,  -2,   2,   0,   1,  -4, -32, -17,   3,  22,  -3,   4,   0,   2,   0,   1,   0,  15, -24,   1,  21, -17,   7,  -3,   2,  -1,   1,   0,   1,  -5, -37,  10,  15,  18,  -1,   1,   0,   1,   8,  55,  17, -27,   1,  -5,   0,   4,  11,  12, -49,  10,  -1,   0,   0,  -8,  14, -12,  15,   3,  -5,   3,  -2,   5,   1,   8, -11,  -5,   6,  -5,   2,   3,   1,   3,  -4,  -2,   3,  -2,   1 },
     {  -10,  -9, -28,   2,  13,   2,   2,   2,   1,   0,   0,   1,   8,  40,  15,  14, -24,   1,  -4,  -1,  -2,  -1,  -2,  -1, -33,   8,  -5, -35,  19,  -4,   5,   0,   2,  -1,   1,   0, -44, -11, -12,  43,   3,  -2,   3,  -1,   2,   0,   1,  -1,   3,  14,  14,   1, -21,  12,  -7,   3,  23, -34,  20, -18,  17,  -7,   1,   1, -16,  27, -13,   5,  -3,  -2,   2,  -3,  15, -16,   0,   3,  -2,   2,  -3,   2,  -4,   7,   2,  -3,   5,  -5,  -2,   1,  -2,   2,   2,  -1,   3,  -3,  -1,   0 },
     {   10,  19,  28, -13,  18, -17,   4,  -7,   2,  -3,   1,  -2,  10,  15,  45,   0, -32,  14, -12,   3,  -5,   1,  -3,   1,  12,   1,  13, -11,  19, -16,   4,  -4,   2,  -2,   1,  -1,  19,   9,  23, -10, -21,   5,  -3,   0,  -1,   0,  -1,   0,  -1,  12,   7, -17,  12,  -7,   1,  -1,  -3,  26,  -3, -31, -12,   1,   1,  -1, -10,   0, -48,  -5,  32,  -3,   4,   0,  -2,  -2,  -4,  34, -15,  -3,   0,  -1,   1,  -6,  -2, -16,  -3,   0,  -6,  11,   0,  -2,   0,  -4,  -1,   0,  -3,   5 },
     {   -9, -27, -62,   3,  29,   3,   3,   3,   1,   1,   0,   1,   3,  -5, -13,  36, -27,   9,   0,   0,  -1,   0,  -1,   0,   9,   3,   2,  -5,  12,   2,  -3,   2,  -1,   0,   0,   0,  49,   8,  26, -16,   0, -12,   2,  -4,   0,  -2,   0,  -1,  23,  -3, -27, -10,  -9,   9,  -3,   0, -28,   1,  -7,  -8,  21,  -8,   3,   1,   4,  -6, -12,  28, -10,   3,   2,  -2,  -4,  -7,  20,  -7,  -1,   3,  -3,   1,  -3,   4,  -5,   3,   0,  -4,   7,   0,  -1,   3,  -1,  -1,  -1,  -2,   2,   0 },
     {  -10, -13, -38,  12,   9,  -5,   7,  -2,   2,   0,   3,   0,  -8,  -4, -52,  -6,   7,  11,  -1,   6,   0,   3,  -1,   1,  -1,  30,  -5,  14,   3, -11,   6,  -7,   0,  -2,   1,  -1, -35,   7, -11,  -6,  -8,  17,  -6,   5,  -2,   2,  -1,   0,  -8,  16,   3,   3,  21, -16,   6,  -4,  10,  -1, -22,   2, -17,  15,  -3,   1,  -2, -16,  -4,  17,  34, -10,   3,   0,   5,   7,  12,  36, -36,  -2,  -2,  -3,   8, -16,  16, -21,  -2,   5,   0,   8,   4,  -5,   5,  -6,   0,   1,   1,   2 },
     {   -5,  -1,  -7, -13,  20,  -7,   3,  -2,   2,  -1,   1,  -1,   0,  12,  29,   0, -18,   4,  -4,   0,   0,   0,   0,   0, -11,   1, -10, -22,   9,   1,   1,   2,   0,   1,   0,   0, -29, -17,  -9, -22,  28,  -8,   9,  -2,   4,  -1,   2,  -1, -56,  -2, -23,  47,  -3,   4,   3,   0, -11,  37,   2,   5,  -4,   3,  -2,   0,  33, -32,   4,  15,  -6,   2,  -2,   1, -11,  17,  12,  -9,  -1,  -1,  -1,  -1,  11,  -5, -10,   3,  -3,   2,   6,  -1,   4,  -1,  -3,   1,  -1,   0,   3,   0 },
     {   -4, -10, -21,  29, -24,  15,  -6,   5,  -2,   2,  -3,   1,  10,  24,  -2,  14,  10, -10,  -1,  -3,  -2,  -2,  -2,  -1,   2,  37,  58, -18, -23,  -4,  -2,  -2,  -2,  -2,   0,  -1,  -6, -24,  17, -10,  11,   0,  -2,   1,   0,   1,   0,   0,  18,  14,   4,   6,  -8,  -6,   2,  -2, -22,   9,  16, -15, -15,   6,  -2,   1,  22,  -8,  -3, -40,  11,  -1,   1,   2, -24,  11, -27,   4,  16,  -2,   4,  -1,  12,  -5,  -3,   6,  -7,   0,  -5,  -1,   2,  -1,  -1,   2,  -3,  -1,  -2,   0 },
     {    2,   2,  25,  24, -29,  12, -11,   6,  -5,   3,  -3,   1,  -6,  -7,  -3, -14,   8,  -1,   4,   0,   2,   0,   1,   0, -11, -17,  18, -35,  30, -11,  10,  -4,   3,  -1,   2,  -1,   5, -42, -14,  11, -11,  16,  -3,   5,  -1,   2,   0,   2,   3,  -2, -35,  -6,  22,  -5,   3,   0, -48, -17,   4,  21,   9,   4,   4,   1,  12,  19, -10,  30,   3,  -6,  -1,  -2,  -4,  -7,  23,   6, -20,   2,  -3,   1,   7,   6,   1, -12,  -4,   1,   1,   6,   4,   3,   2,  -5,  -1,   0,   0,   3 },
     {   -2,  -7,  -7, -36,   0,  19,   5,  -1,   1,  -1,   2,  -1,   1,  -4,  12, -13,  18, -16,  -4,  -1,   1,   0,  -1,   0, -10, -37, -23, -21, -13,  19,   7,   3,   0,   1,   2,   1,  27,  -4, -19,  13,  28, -10,  -4,   1,   0,   1,   0,   1,  28,  16,  -4,   3, -26,   1,   7,  -2,  -3, -20, -10, -17,  12,  15,  -5,   3,   8, -20,   4,   1,  32, -16,   7,  -2, -24,  24, -13,  35,  -9,   0,   1,  -4,  14, -10,  17,   0, -10,  10,  -3,   1,   6,  -3,   4,   3,  -5,   4,  -1,  -1 },
     {    8,   8,  18,  35,   0, -25,   7,  -4,   0,  -2,  -1,  -2,   3,  -3,   0, -12, -27,  36, -11,   6,  -4,   1,  -2,   1,  -1,  -5,  -5,  21,  13, -29,   9,  -4,   2,  -1,   1,  -1,  14,  -5, -10,   6, -11,  13,  -1,   2,   1,   0,   0,  -1,  37,  38,   4,  30, -20,  -1,  -9,   0,  11, -24,   8, -23,  -5,   3,   0,   3,  17, -26,  29,   9,  -3,  -3,  -4,  -1, -16,  19,   6, -31,  -5,  -1,   1,   0,   5,  -6,  -8,   3,  -5,   4,   6,  -9,  -1,  -3,  -6,  -2,  -2,   2,   1,  -2 },
     {   -3,  -4,  -6,  -5,  25, -16,  10,  -5,   4,  -3,   1,  -1,   2,   8,  32,  -4, -17,  16, -13,   3,  -6,   2,  -4,   0,  -8,  -9,  32,  18, -27,  -6,  -2,   3,  -1,   1,   0,   1,  -8, -29,  -8, -39,  29,   7,   3,   2,   3,   1,   0,   1,  17, -32,  13,   4,  10, -13,   2,  -1, -20, -26, -20,  17,   5,   8,   2,   2, -29,  24,  33,  10,   7, -11,  -3,  -1,  18,   8, -13,  -1, -12,  -3,   2,  -2, -10,   2,  12,  -7,   3,   7,  -8,  -1,  -3,   0,   5,   0,   2,   5,  -3,  -2 },
     {   -5,  -3,  -9, -16,   9,  -9,   2,   3,   0,   2,  -2,   1, -13, -14, -29, -65, -16,  30,   7,   5,   3,   2,   3,   1,  -2,   5,   5, -13,  14,  -1,  -4,   8,  -2,   2,  -2,   1, -11, -13,  -6, -22,  -1,   6,  14,  -4,   3,   0,   1,   0,  -6,  -5,   0, -17,  11,  16,  -6,   4,  -8, -15,   0, -17,  21,  -6,   7,  -1,  -8,  -5, -17, -11,  15,  13,   3,  -2, -13,   3, -15,   9,  45,  -9,   2,  -2,   2,  10,  -7,  43,   3,   1,   9,   2,   2,   4,  -1,  10,   2,   1,   3,   2 },
     {    4,   0,  15,  47, -25,  -4, -12,   2,  -2,  -1,  -1,  -1,   0,  -7,   2,  42, -28, -10,  -3,  -2,  -2,   1,  -2,   2,  -9, -15, -25, -12, -16,  10,  -7,   8,  -1,   1,   1,  -1, -14, -23, -21, -33,  -5,  -2,  15,   0,   3,   3,   2,   1,  20, -13, -19, -12,  12,  23,   2,   6,  33,   9, -14,  -5,  11,  -5,   3,  -2, -12, -32,  -5,  -3,  15,  10,   2,   1,   7,  -7,   0,  23,  16,  -6,  -1,  -1,  -4,  -6,  15,   7,   5,   3,   3,  -3,  -4,   6,   2,   3,   4,   1,   1,  -2 },
     {    0,  -6, -17, -22,   6,   4,   6,  -3,   3,  -2,   1,   2,   7,   4,   0, -17,  11,   8,  -7,   2,  -2,   3,  -1,   0,  -1, -11,   1, -23,  -1,   6,   1,  -1,   0,   2,   0,   1,  10, -17, -28, -10,  17,  15,  -1,   4,   0,   2,   1,   0,  52,  35,  -2,  18,  28, -22,   2,  -7,  27,  38,   1,  21, -32,  -4,  -3,  -4,  -2,   0, -16,  -7, -14,   8,  -5,   4,  13, -32,  12, -12,   8,   2,   3,   3, -19,  12,  -8,   5,   6, -11,   1,   4,  -7,   3,   2,   0,   3,  -4,   0,   2 },
     {    0,   2,   2,  30, -20, -31,  37, -19,   8,  -3,   5,  -3,  -2,  -2, -11,   9,  14,  32, -41,  22,  -7,   2,  -1,   3,  -3,  -3, -23,  -3, -12,  -7,  23,  -7,   0,   2,   1,   0,   9,  16, -11,   6,  34,  -8,  -5,  -4,   1,  -2,   1,  -2, -19,   9,  -2,  -5, -25,   0,   7,  -1, -32,  12,  -8,   7,   9,   6,  -1,  -3,   9,  18, -16, -17,  -9,   3,  -3,   4,  10, -27, -13,  17,  10,   8,   3,  -1,   1,  12,  13,   6,   4,  -8,  -6,   2,   2,   1,   7,   1,  -2,   0,  -2,  -1 },
     {   -5,  -7, -10, -25, -13,  21,  24, -14,   6,  -1,   0,   0,   2,   2,  15,  24,  18,  -3, -31,  21,  -7,   1,  -4,   2,  -3,  -7,   1,   7, -21, -17,  33, -18,   8,  -2,   3,   0,  -9,  -7,   1,  -4, -25,  34, -31,  16,   0,  -1,   0,   0, -15, -20, -15, -15,  12, -17,  12,   1,  14, -20,   8, -22,   3,  11,   2,  -6,  -5, -19,  -7,  12,   8,   6,  -4,   3, -10,   7,  16, -20,  21,   0,   4,  -2,  -4,  12, -18,  19,   1,   1,  12,  -1,  -2,   8,  -4,   3,   3,  -1,   1,   5 },
     {    3,   4,   2,  11,  41,   0,  -2, -10,   0,  -2,  -2,  -1,   2,   1,  14,  11,  -2,  -9,   1,   4,  -1,   0,   0,  -2,   1,  -8,  18,  49, -25,  18, -18,   4,  -4,   1,  -2,   1,  -1, -11, -47,  22,  -8, -17,  15,  -2,   8,  -2,   3,   2, -12,  20,   4, -19,  12,   3,  -7,   5, -16, -26,   4,  13, -19,   3,   0,   3,  27,  -5, -31,  -3,  10,   6,   4,   2, -20, -19,  23,   6,  11,   1,   0,   0,   9,   5,  -2,  10,  -4,  -6,  13,   0,   5,   3,   3,  -4,   0,  -2,   5,  -1 },
     {   -1,   4,   5,   5,   1,  33, -25,  20,  -5,   4,  -5,   1,  -5,   1,  -1,  22, -22,  -6,  26, -16,  -1,  -5,   0,  -3,  -7,  -6, -28,   6,   9, -24,   0,   7,  -2,   5,   0,   1,   2,  22, -11,  -3,  15,  17,  -7,   0,  -2,  -3,   0,   1, -13,  18,  15,  21,   8, -36,  11,   0, -48, -11,  -8, -20, -22,  18,   5,  -6, -23,  -8,   2,  -6,   6,   4,  -4,   4,  12, -12, -11,   1,  16,   3,   7,  -2, -13,  12,   9,  25,  12,  -4,   1,  -6,  -2,   0,   4,   2,   4,   0,  -1,  -4 },
     {   -4,  -5, -12, -17, -25, -11,  -7,   3,   5,   1,   0,   1,  -5,  -6, -11, -16, -57, -23,   4,   1,   1,   3,   0,   1,  -3,  -6, -17, -17, -50,  -5,   2,   5,   5,   1,   2,   0,   3,   9, -23, -27, -26,  -3,  -3,   7,   2,   1,   0,   1, -10,  15,   8, -22,  -4,  -8,  10,  -1,  -4,   4,  13,  -6,  -3,  13,   0,  -1,  34,  36,   2,   0,  -3,  -5,   0,  -2,   6,   2,  -7, -21, -13,  13,  -2,  -2,  10,  -4, -11, -15,  -4,  -5, -10,   5,   4,  -5,   2,   2,  -1,  -2,  -2,   2 },
     {    0,  -1,   6,  10, -10, -14,   7,  -7,   4,  -2,   2,  -1,  -2,  -5,  -4, -14, -15,   0,  -1,   3,   0,  -4,   2,   2,  -1,   1,   4, -33, -11,  26, -15,   4,  -7,   5,  -3,   1,   2,   9,  31,  17,  10, -14,   4,  -4,  -2,  -5,  -3,  -2,  -2, -18,  19,  -3,  -7,   0,   9,  -4,  -4, -21,  -6,   4, -27,  33, -10,   9,   5, -30,  14,  -9,  34,  -5,  -3,   2,  14,  -9,  44, -10,   2, -16,   2,  -5, -18,  40, -20, -12,  14, -11,   7,   7,  -8,  10,  -7,  -8,   1,  -7,  -5,   5 },
     {   -3,  -7,  -2, -26, -36,  25,   8,   3,   5,   5,  -1,   0,  -2,  -6,   4, -18, -43,  11,  12,   5,   1,   4,  -1,  -2,   6,   5,  26,  39,  13, -16,   2,  10,   3,   1,   0,  -1,  -5, -17,   2,  23,   8, -21, -11,   4,   2,  -3,   1,   2,   0, -19, -22,   1, -25,  -5,   8,   3,  16,   9, -13,  -1,   2,  11, -15,  -7,  12, -16,  -4,  -7,  -8,   2, -11,  11,  -1, -36,   7,   9,  -3,  10,  10,   3, -17,  -1,  32,  -8,   0,   9,  -2, -12,   5,   3,   8,  -1,  -4,   6,  -7,  -3 },
   },
   {
     {   87, -67,   7,  -1,   1,  -1,   0,   0,   0,   0,   0,   0, -53,  28,  13,  -7,   2,  -1,   1,  -1,   0,   0,   0,   0,  -1,  14, -14,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -2,  -3,   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   1,   0,   0,  -1,  -1,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -48, -19,  39,  -1,   5,  -2,   2,   0,   1,   0,   1,   0, -22,  75, -37,  -9,   0,  -1,   0,  -1,   0,   0,   0,   0,  31, -34, -17,  29,  -8,   3,  -1,   1,  -1,   1,   0,   1,  -1,  -9,  27, -10,  -8,   4,  -1,   1,  -1,   0,   0,   0,   6,  -1,  -7,  -8,  10,   0,  -1,   1,  -1,  -1,   4,   3,  -1,  -2,   1,  -1,   2,   0,  -2,  -2,   0,   1,   0,  -1,   0,   0,   1,   1,   1,  -1,   0,   0,   1,   0,  -1,  -1,   0,  -1,   1,   1,   0,   0,  -1,   0,   0,  -1,   1,   0 },
     {  -45,  21,  -1,   5,  -2,   3,   1,  -1,   0,   0,   0,   0, -59,   6,  46, -12,   6,  -3,   2,  -1,   1,  -1,   0,   0,  24,  42, -43, -19,  14,  -5,   0,   0,   0,  -1,   0,   0,  20, -20, -22,  34,   0,  -4,   0,  -1,   0,   0,   0,   0,   0,   0,  17,   4, -19,   3,   1,  -2,   4,  -4,  -8,  -4,   3,   4,  -1,   3,   0,  -2,   3,   2,   0,  -2,  -1,   1,   3,  -4,   0,  -2,  -1,   2,  -1,   0,   1,  -1,   1,   1,   1,   0,   0,  -1,   1,   0,   1,   1,   1,   0,  -1,   0 },
     {  -33, -68,  47,  10,   5,   1,   2,   2,   1,   1,   1,   0,  59,  -6,  20, -31,   0,  -3,  -1,  -2,   0,  -1,   0,  -1, -31,  30, -21,  -7,  14,  -4,   4,  -2,   2,  -1,   1,  -1, -11,  14, -13,  17,  -5,  -2,   1,  -1,   0,   0,   0,   0,   3,   1,  -4,   0,  -4,   4,  -2,   1,   0,   0,   0,   2,   1,   0,  -1,   0,   1,   1,  -1,  -1,  -1,   0,   0,   0,  -1,   1,  -2,   2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -32, -27,  28, -12,   6,   1,   2,   2,   2,   0,   1,   0, -43, -10,   6,  28,   0,   2,  -2,   2,   0,   1,   0,   0,   7,   9,  57, -32, -18,   2,  -1,  -1,  -1,   0,  -1,   0,  -8,  33, -34, -35,  30,  -2,   3,  -1,   2,  -1,   1,   0,   6,  -2, -20,  28,   5, -11,   2,   0,   3,   2,   3,  -1, -11,   4,   1,  -2,   4,  -1,  -4,   2,   2,   2,  -2,   0,   2,   2,  -2,   0,  -1,  -1,   1,   1,   1,  -1,  -2,   1,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   1,   0,  -1 },
     {   -4, -30, -31,  27,   4,   5,   0,   2,   0,   1,   0,   1,  22,  -5,  35,   2, -22,   2,  -5,   1,  -2,   0,  -1,   0,  74, -39,   2, -37,  12,   5,  -2,   1,   0,   0,   0,   0, -36,  -5,  16,   5,  17, -11,   3,  -1,   1,  -1,   1,  -1,  -7,  21,  -2,  -6,  -5,  -3,   4,  -2,  -8,   2,  -8,   4,   4,   1,   1,  -1,  -1,   5,   1,   2,  -4,  -1,   0,   0,  -3,   0,  -2,   1,   1,   1,   0,   0,   0,   2,   1,   1,  -1,   0,  -1,   0,   0,   1,   0,   0,  -1,   0,  -1,   0 },
     {  -28, -16, -48,  35,  15,   0,  -1,   6,   2,   2,   0,   0, -19,   8,  39,   1, -27,   4,  -3,   1,  -1,   1,   0,   0, -39,  16,  13,   7, -12,   8,   0,  -4,   0,   0,   1,   0,  13,  -6,  27, -28, -17,  10,  -2,   6,  -2,   0,  -1,   1,  13,  -8, -16, -35,  41,   1,  -4,   5,   1,   2,   8,  15,   5, -16,   2,  -6,   5,   5,  -3,  -8,  -6,   3,   5,  -3,   0,   4,  -6,   8,   4,  -4,   2,   0,   0,   1,   0,  -3,   0,  -1,  -2,   4,   0,   0,  -1,  -2,   0,   0,   1,   0 },
     {   -2,  -3,  38,   8, -22,   1,  -1,  -2,  -2,   0,  -1,   1, -23, -66,  15, -20,  33,   6,   6,   1,   3,   1,   1,   1,  24,  -2,   1,  32,   5, -15,  -4,   3,  -2,  -1,  -1,  -1,   9,   6,  38, -28, -21,   7,  -2,  -2,  -1,   2,   0,   0, -18,  32, -10, -15,  17,  -3,   1,  -5,   0,   7, -24,  12,  -3,  -8,   5,   4,  -5,   0,  -2,   3,  -1,   3,   0,   0,   3,  -3,   2,  -1,  -3,   2,  -4,   0,   1,   1,  -1,   0,   1,   2,   0,  -1,   0,   1,   0,   1,   1,   1,  -1,   2 },
     {   13,   6,  -1,  -2,  -7,  -1,  -1,   1,   0,   0,   0,   0,  35,  14,   6, -18,   8,  -2,  -3,  -2,  -1,  -1,   0,   0,  28,  34, -19,  -6, -29,  -1,  -1,   0,   0,   0,   0,   0,  19, -50, -18, -39,  27,  22,  -2,   5,  -1,   3,   0,   2, -20,  -6,  -6,  29,  40, -29,   2,   0,  -2,   4,   5,  15, -27,  -4,  11,  -4,  -5,  -1,  -1,  -2,   2,  10,  -4,  -2,   0,  -1,   1,   4,  -3,  -3,  -1,   2,  -3,  -1,   1,  -1,  -1,   0,   1,   2,  -2,   0,   0,   0,   0,   0,   1,   1 },
     {  -11, -39,   4, -11,  11,   5,   4,   1,   2,   0,   1,   0,   9, -27, -25,  60,  -2,  -4,  -1,   0,   0,   0,   0,   0,  22,  26,   8,   4, -34,   7,  -2,  -1,  -2,   0,  -1,   0,  18, -39,   8,   9, -10,  14,  -6,   4,  -1,   2,  -1,   1,  -1, -10,  48, -28,  -4,   7,  -6,   3,  -3,  11,  -7, -23,  22,  -4,   0,   0,  -4,   2,  -1,   7,   0,  -8,   3,  -1,  -2,   4,  -2,  -5,   0,   1,   1,  -1,  -1,   2,   0,   0,  -1,   2,  -1,  -2,   0,   0,   0,   0,   0,   1,  -1,  -2 },
     {    9,  24,  17, -66,  27,  -7,   1,   2,   2,  -1,   1,  -1,   2,  -8,  24,  10, -10, -10,   0,   0,   0,   0,   0,  -1,  21,  16, -24,   6, -16,  -2,   8,  -8,   2,  -1,   2,   0, -61,   9,  -3,   1, -17,  13,   4,   3,   3,  -1,   0,   1,  21,  -8, -17, -18,  25,  15, -11,   8,   1, -10,  11,  16,   9, -15,  -3,  -3,   4,   5,   2,  -4, -10,   0,   7,  -3,  -3,   5,  -8,   7,   5,  -1,   3,  -2,   0,   0,   1,  -1,   0,  -2,  -1,   5,  -1,   0,   0,  -2,  -1,   0,   1,  -1 },
     {    7,  -2,  -5,  29,  -5,  -6,  -3,  -1,  -1,   0,   0,   0,   4,  -5, -17, -20,  -4,  14,   1,   3,   0,   2,   0,   1,  47,   6,  15,   5,  18, -21,  -5,  -5,  -1,  -2,   0,  -2,  18,  24, -45,  -7, -44,   5,  12,   0,   3,   0,   2,   0,  -3, -50,   4,   4,  18,  37, -12,   2,  -4,   3,  24,  -2,  12, -23, -10,   5,  -4,  -2,  -4,  -6,  -1,  -1,  11,   1,  -2,   2,   3,   3,   3,  -1,  -1,  -4,  -2,  -1,   0,   0,   0,   1,   2,   1,  -1,  -1,  -1,  -1,  -1,   1,   1,   1 },
     {   15,  20,  43, -25,  -3,  -1,  -2,  -3,   0,  -1,  -1,  -1,   8,  17,  15,  -8, -28,   5,  -4,   1,  -1,   0,  -1,   0,  -8, -17,  14, -43,  21,  16,  -6,   4,  -1,   2,   0,   1,  46,  -8,  -4,  -7,  12, -14,  -3,   2,  -1,   1,   0,   0, -13,  -7,  -1, -34,  17,   5,   4,   0,  -9,  12, -29,  15,  36, -22,   0,  -2,  -4,   1,  -7,  22, -18, -10,  11,  -1,  -2,   0,  -3,   2,  -2,   8,   0,  -3,  -1,  -1,  -1,   5,  -1,   1,  -1,   0,   0,  -1,  -1,   3,  -1,   0,  -1,   1 },
     {    7,  18,  -1,  34, -43,   5,   3,  -4,  -2,  -1,  -1,   0,   9,  32, -26,  10,   3,   2,  -3,  -4,   0,  -1,  -1,  -1,   6,  58,   6, -24, -11,  -5,  -5,   4,  -1,  -1,  -2,  -1, -36,  19,   7, -18,  -6,  11,  -4,  -2,   0,   1,   1,  -1,  11,  21, -12, -22,  -6,   6,  -3,  -1,   4, -23, -35,   1,  25,   1,   0,   8,  -3, -14,   6,  17,   4, -10,   1,   1,   2, -10,   5,  -5,  -5,   6,   0,  -1,   3,  -4,   0,   3,   1,  -3,   2,  -2,   1,  -1,   2,   2,   0,  -1,   0,   0 },
     {    4,   0,  -9, -20,   7,   7,   0,   1,  -2,   1,   0,   0,  22,  20,  43,  -5,  12, -27,  -2,  -5,  -1,  -3,  -1,  -2,  18,   2,  12,  23, -37,  -4,   4,  -4,   0,   0,   0,   0,  38,  18,  -1, -22, -15,  -9,  -2,  -1,   2,  -1,   0,  -1, -15,  -2, -31, -12, -36,  17,  15,  -1,  -4, -15,   1, -20,  20,  40, -16,   2,  -2,   2,  -6,   0,  23, -18,  -7,   5,  -3,  -3,  -3,   3,  -2,  -2,   6,   1,  -2,   1,  -2,   2,  -1,  -2,   0,   1,  -1,   0,   0,   1,  -1,  -1,   0,   0 },
     {    6,  17,  31,  18, -37,   3,  -6,   4,  -2,   1,  -1,   0,  -7,   3,  34, -24,  -7,   9,  -4,   2,  -2,   1,  -1,   0, -16, -27,  14,  -5, -22,   6,   7,  -1,   4,   1,   2,   0, -34, -23,  -2,  -5,   3,  28,   1,   3,   3,  -1,   0,   1,  11, -10,  32,   0,   4,   3, -15,   5,   3,  13,  14, -53,  10,   5,  -4,   2,  -2,   7, -23,   0,  34,  -9,   2,   0,   1,   8,  -3,  12,  -3, -12,   6,  -1,   0,   2,  -5,  -3,   0,   0,  -2,   5,  -1,   1,  -3,  -1,  -1,   0,   1,   0 },
     {  -11, -11, -35, -28,  11,   6,   9,   3,   2,   1,   1,   1,   3,  29,  13, -18,  45,  -6,  -5,  -5,  -3,  -2,  -1,  -1, -12, -17,  34,   6,   0, -38,  11,   1,   2,  -1,   0,  -1, -13,  -2, -29,  11, -14,  26,   9,  -7,   2,   0,   2,  -1, -15,  21,  27,  -5,  11,  -2, -15,  -1,  -3,  22, -39,   4,   3, -15,  12,   4,   4,  -8,  -2,  22, -10,   5,  -1,  -2,   2,  -3,   5,  -5,  -7,   4,  -4,   1,   3,  -1,  -1,   3,   1,   1,   0,  -1,   1,  -1,   1,   2,   0,   0,  -1,   0 },
     {  -11, -11, -11,  -4,  -8,  17,  -3,   1,  -1,   1,  -1,   1,  -8,  -6,  -9,  -4,  14, -12,  10,  -1,   2,  -1,   2,  -1, -18,  -5,  -2,   2,   0,  12,  -7,   2,  -1,   1,   0,   1, -18, -39,  -4, -14,  14, -10,  16,   4,   4,   2,   1,   0, -31,  23, -12,  29,   4,  46,   0, -12,  24,  -8,  26,  11,  31, -19, -42,   9,  -1,   6,   3,   4, -16, -36,  20,   7,   4,  -2,   5,  -1,  -7,  12,   5,  -8,   1,   3,   0,   1,   3,   0,   2,   0,  -1,   2,   0,   0,   2,   0,   0,   0 },
     {    7,  18,  25,  20,  39, -34,  -6,  -6,  -1,  -4,   0,  -2,   1,  16,   9,  12, -36,  -1,  10,  -1,  -1,   0,   0,  -1,  -6,  18,  25,   8,   1,  -9, -15,  -4,  -2,   0,  -1,   0, -22, -21,  -2,  -4, -27, -19,  16,   4,   4,   1,   1,   0, -51,  24,  11,   1, -10,  18,  10,  -7,  16,  16, -12,  -6, -26,   8,  -6,   0,   1, -15,  -2, -15,   4,  23,  -3,   3,   5,   1,   0,  -2,   8,  -3,  -8,   0,  -2,  -5,   0,  -5,   2,   1,   0,  -1,  -1,  -3,  -1,  -3,   2,   0,  -1,   0 },
     {   -3,  -3, -11, -26,  13,  -2,   0,   6,   1,   0,   1,   0,  -4,  -5, -25, -47,  13,  23,  -1,   6,   1,   2,   1,   1,  12,  36,  48,  -8,  22,  -7,  -6,  -8,  -2,  -1,   0,   0,  -9, -44,  26,  13,   0, -31,  10,   3,   3,  -3,   0,   1,  28,  -7, -21, -11,  -7,  -6,  14,   1,  -9,   5,   6, -11,   2,  14,  -1, -10,  -2,   8, -12,   6,  13,  -3,  -1,  -4,  -1,   6,  -7,  19,  -8,  -9,   5,   0,   0,   0,  -2,  -3,   0,  -2,   0,   8,  -3,   0,  -2,  -2,  -1,   1,   1,   1 },
     {    0,  -2,  14, -11, -49,  27,  -1,   5,   0,   1,   1,   2,   8,  14,  16,  47,   3,   5, -29,   3,  -8,   1,  -3,   0,  -9,  -7,  11, -15,  22, -48,  13,   4,   1,   1,   1,   1,  10, -30,   8,   6, -24,  -5,  26,  -9,   2,  -1,   1,  -1,   4,   2, -17,  -2,  -1,   8,  -2,  -3,  -3, -12,  20,  14, -17,   7,   2,  -2,   1,  12,  13, -28,  -4,  15,  -6,   1,  -5,   2, -10,  -1,  13,  -4,   0,   3,   0,   3,   3,  -2,  -2,   0,  -3,  -1,   0,   1,   1,  -2,   0,  -1,  -1,  -2 },
     {   11,   8,  13,  13,  17,  -4, -10,  -2,  -3,  -1,  -1,  -1,  14,   9,  18,   1,   4, -15,   6,  -1,  -2,   0,  -1,  -1,  11,   0,  24,  10, -11,   1, -10,   0,  -2,  -1,  -2,   0,  27,  -7,   5,  15,   1,  -5,  -1,  -2,   0,   0,   1,  -1,  60,  37,  -4,  19,  -8,  -4, -28,  -7,   6, -24,  -2, -19, -20, -42,  -4,  24, -15,  -5,   0, -11, -14,   3,  38,   1,  -1,   0,  -2,  -5,  -4,  15,  -2, -11,  -4,  -3,  -1,  -4,  -3,   1,  -1,  -3,   0,   0,  -2,  -1,  -1,  -1,  -1,   1 },
     {   -5,  -2, -18, -31,  -7,  15,  10,  -2,   5,  -2,   2,   0,   8,  17,   8,  15,   3,  16,  -8,  -4,   1,  -2,  -2,   0,   2,  24, -15,   2,  33,  -2, -23,   5,  -6,   1,  -1,   1,   0,  36,  31, -23,   6, -20,   0,  -3,  -3,   0,  -1,  -1, -34,   0,  21,  11,  -1, -16, -12,   3,  13,   5,  13, -38,  -6, -29,  12,  10,  -6,   7, -39, -10,   9,   1,  23,  -6,  -4,  -5,  -6,  16,  -1,   5,   1,  -5,  -1,  -1,  -5,   1,   4,  -4,   0,   6,   0,  -1,  -2,   1,   0,  -1,   0,   3 },
     {   -6, -14, -15, -28,   3, -10,  14,   1,   5,   0,   4,   0,  -3, -12, -17, -10, -33,  43,  -8,   2,   0,   1,   0,   1,  -1, -13, -14,  -8,   6,   7,  -3,   6,   2,   0,   0,   3,  23,  -4, -20,  -5,  -5,  22,   9,  -8,   1,  -1,   1,   0,   0,  42, -10,  -2,   9,  16, -25,   6,  21, -38, -15,  -9,  -2,   9,   0,  12,  -7, -18,  15,  -1,  44,   4,  -5,   2,   5,   7,   3,  13,  10, -26,   5,  -2,  -5,   2,   3,   3,   0,   1,   0,   4,  -1,  -1,   3,   3,   1,   1,   1,   2 },
     {    3,   5,   3,  12,  35, -17,   4,  -6,   0,  -2,   1,  -1,  -6,  -6, -22,  -4,  22, -36,   5,  -1,   2,   2,   0,   2,  -5,  -1, -11, -66,  -2, -11,  34,   1,   5,   1,   2,  -1,  14,   4,  40,   0, -13,  35,   7, -20,   3,  -4,   1,  -3, -13,  -3, -16,  12,  -3,  15, -22,   0,  -6,   7,   7,  -6,  -1,   1,  -1,  10,  -3,   8, -21, -10,  12,   1,  -3,   1,  -2,  -5,  -2,  18,   4,  -1,   0,   0,   1,   0,   2,   2,   1,   3,   0,   1,   0,   1,  -1,   1,   0,  -1,  -1,   2 },
     {   -1,  -4,  -5, -12,   8, -10,  11, -14,   0,  -3,   2,   1,   6,   1,  14,  12,   1,   8,  21, -10,   0,  -1,  -2,  -1,  -9,  -6, -17,  -7,  18, -13, -38,  21,   0,   0,  -1,   0,  -3,  -9,  13, -39,  -7,  10,  15,  12,  -7,  12,  -1,   0,  50, -19,  16,  29,  -4,  24,  19, -32, -25,  17, -27,   6,   1,  10, -15,   8,  -2,  -2,  -4,   1,  15,  -1, -15,  12,  -1, -15,  13,  -7,   5,  -2, -10,   1,   4,   4,  -3,   7,  -3,   4,   0,  -7,   1,   3,   1,   5,   0,   1,  -2,   5 },
     {   -3,  -5,  -3,   0,  -5,  16, -12,   8,   0,   1,  -2,   0,  -4,  -9,  -3,   3, -11, -22,  22,  -3,   1,  -3,   1,  -1,  -7,  -9, -11,   3,  -4,   0,  -2,  -4,   5,   0,   2,   1,  -5, -15,  -9, -10,  -3, -13,  23,   4,   3,  -1,   1,  -1,  -2, -11, -20,   0, -11,  11,  -3,   2,  -4, -25,   0, -17,  11,  -4,  39,   6, -11,  14, -13,  48,   0,  50,   7, -33,  10,  -4,  28,  10,  -4,  -3, -44,   4,  -2,   1,   6, -11,   3,  -3,   5,  -5,  -2,   1,  -2,  -3,   2,  -2,   2,  -1 },
     {   -4,  -5,   1,  -9,  -4, -12,  23, -17,   1,  -1,  -2,   0,  -2,   1,  12, -17,  13,   2,   5, -10,  10,  -6,  -1,   0,  -5,  -8,  27, -13,  -3,  16, -25,  20,  -2,  -3,  -2,   1, -15, -17,  14,   4, -18,  16, -31,   5,  -4,   5,   0,   0, -24, -41,   1,   5, -22,  -8,   0,  15,  14, -44,  -2,  13,   1,  -8,  28,  14, -21,   7,  26, -22,  -1,   0,   7,  -4,  -4,   6,   6, -21,  20,  12,  -5,  -4,  -2,  10,  -8,   9,  -2,   9,  -4,  -4,   5,   8,  -2,   5,   1,   1,  -2,   4 },
     {   -4,  -5, -12, -19, -29,  27,   2,  -1,  -1,   0,   1,   2,  -3,  -2, -17, -29,  -7, -24,  14,   5,   7,  -1,   3,   0,   4,   5,   5, -15, -17,  32,  -5,   1,   1,  -1,   1,  -1,   3,   7, -14,   2, -11,  -1,  -6,  10,  -3,   2,   1,   0,   8,  29,   8, -16,   2,  18,  11,  -5, -11,  26,   7,  -4,  -9,   1,  -4,  -4,   1,  -2, -14, -38,  -7,  32,   1,   2,   1, -17, -17,  -7,  53,   8, -23,   3,   3,  -3,  -2,  24,   3,   1,   3,   3,   1,  -2,   1,   3,   0,   1,   0,   3 },
     {   -2,   1,   1,   4, -15, -48,  34,  -7,   6,  -6,   1,  -1,   2,   3,   5,  15,  10,  30,   9, -19,   0,  -5,   0,  -3,  -2,   1,   4,   5,  -9,  13, -15,   0,  -1,   1,   0,  -1,   3,  -6, -16,  20,  -5,   9,  -9,   3,  -2,   1,   0,   1, -11,  29, -21,  -1,  -4,   3,  -9,   3, -55,  -2,  26,   3,   7,  -5,   8,   3,  32,  34, -25,   1,  -6,  -5,  -1,  -6,  -3, -28,   1,   7,  -9,   4,  -6,   4,   4,   2,   7,  -5,   2, -10,   0,  -5,  -1,   2,   1,  -3,  -2,  -2,   0,  -2 },
     {    2,   3,   3,   5,  18,   2, -33, -11,  10,  -3,   2,   5,   0,  -1,  -2,  -9,  -4,  22,   9, -16,   3, -12,   4,  -4,  -1,   3,  -3,   4,  -2, -38,  24,   8, -21,   3,  -2,   2,   1,  -5,   0, -26,  32, -11, -18,   7,  11,   6, -13,   1,   7,  12,  14, -10, -22,  28, -27,   7,  -2,  -7,  19,  10,  10,  -2,  39, -22,   8,   2, -12,  -1,  -4, -14,  -4,   7, -11, -11,  -4, -33,  24, -10,  -3,  12,   1,   0,  -6,  13,  -2,   6, -11,  10,  -4,   8,  -4,   6,   6,  -1,   0,   4 },
     {    1,   5,   7,   9,  22,  55, -18, -16,  -1,  -7,  -2,   2,  -2,   3,  -1,  15, -12, -11,  11,  12,  -3,  10,  -1,  -1,  -1,   7,  12,  24,  35,   9,   7,  -8,  -8,   2,  -3,  -3, -11, -13, -15,   3,   1,  20, -40,  -3,   7,  -2,   5,  -1, -13,   1, -13,   9,  -3,  -5,  -3,   9, -47,  -5, -19,   9,   2, -10,  -2,  13,  -1,   8,  -1, -13,  32, -10,   3,   8,  -2,  -4,   0,  23,   0, -15,  -4, -11,  13,   3,   7,   9,   5,   5,   2,  -5,   4,   6,   6,   3,   1,  -1,  -1,   0 },
   }
 },
 {  //9
   {
     {  115, -50,  -6,  -6,  -2,  -2,  -1,  -1,   0,  -1,   0,   0, -17,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0, -10,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -8,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {   22,  21, -17,  -1,  -4,   0,  -2,   0,  -1,   0,  -1,   0, 108, -42, -17,  -2,  -4,  -1,  -1,   0,  -1,   0,   0,   0, -18, -21,   8,   2,   2,   1,   1,   1,   1,   0,   0,   0, -16,   0,   6,   1,   1,   0,   0,   0,   0,   0,   0,   0, -12,   2,   4,   0,   1,   0,   0,   0,  -6,   4,   1,   0,   0,   0,   0,   0,  -5,   2,   1,   0,   0,   0,   0,   0,  -4,   2,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0 },
     {   13,  10,  -7,  -2,  -1,  -1,  -1,  -1,   0,   0,   0,   0,  14,  50, -30,  -1,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0,  91, -38, -25,   3,  -4,   1,  -1,   1,  -1,   0,   0,   0, -12, -35,   8,   7,   2,   2,   1,   1,   1,   0,   0,   0,  -8,  -8,  11,   3,   0,   1,   0,   0, -10,  -1,   8,  -1,   0,   0,   0,   0,  -4,   3,   3,  -2,   0,   0,   0,   0,  -5,   0,   2,  -1,   1,   0,   0,   0,  -2,   1,   1,   0,  -3,   0,   1,   0,  -1,   1,   1,   0,  -2,   0,   1,   0 },
     {  -29, -71,  32,  10,   8,   3,   3,   2,   2,   1,   1,   1,  -1, -30,   9,  11,   0,   3,   0,   2,   0,   1,   0,   1,  22, -53,  24,   4,   0,   1,   0,   1,   0,   0,   0,   0, -37,  18,  32,  -6,   1,  -2,   0,  -1,   0,   0,   0,   0,   5,  29,   1, -10,  -2,  -3,   0,  -1,  -2,  13,  -6,  -7,   0,  -1,   0,   0,   3,   8,  -7,  -2,   1,   0,   0,   0,  -1,   4,  -3,  -1,   0,   0,   0,   0,   2,   4,  -3,  -1,   0,   2,  -2,   0,   1,   2,  -2,   0,   0,   1,  -1,   0 },
     {   12,  56, -11,  -7,  -5,  -3,  -2,  -2,  -1,  -1,  -1,  -1, -27,  42, -15,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -40, -45,  18,   6,   6,   3,   3,   1,   1,   1,   1,   0, -62,  25,  24,  -2,   4,  -1,   1,  -1,   1,   0,   0,   0,  10,  22,  -8,  -5,  -1,  -2,  -1,  -1,   6,  -1,  -9,   0,   0,   0,   0,   0,  10,  -4,  -5,   2,   0,   0,   0,   0,   3,  -4,  -1,   2,   0,   0,   0,   0,   5,  -3,   0,   1,   2,  -2,   1,   0,   2,  -2,   0,   0,   1,  -2,   0,   0 },
     {   -3, -26,  41, -14,   4,  -3,   2,  -2,   1,  -1,   0,  -1,  40,  53, -12, -15,  -6,  -4,  -3,  -2,  -2,  -1,  -1,  -1,   7,  17,   9, -15,   1,  -3,   0,  -2,   0,  -1,   0,  -1,  -8,  60, -32, -12,   0,  -3,   0,  -1,   0,   0,   0,   0,  24, -19, -39,   7,   1,   2,   0,   1,  -8, -19,  -4,  12,   2,   2,   0,   1,   3, -12,   0,   7,  -1,   1,   0,   1,  -6,  -5,   4,   3,   0,   0,   0,   0,   0,  -4,   1,   2,  -3,  -2,   2,   1,   0,  -2,   0,   1,  -2,  -1,   1,   1 },
     {  -10, -13,  47, -13,   2,  -3,   1,  -2,   0,  -1,   0,   0,  23,  38,  28, -25,  -3,  -6,  -1,  -3,  -1,  -1,   0,  -1, -28,  35, -25,  -9,   1,   0,   1,   0,   0,   0,   0,   0, -39, -46,   7,  13,   8,   4,   2,   2,   1,   1,   1,   1, -36,  21,  27,  -2,   3,  -2,   0,  -2,   7,  23, -10,  -8,  -2,  -2,   0,  -1,   3,   2, -12,  -1,   1,   0,   0,   0,   5,  -2,  -7,   3,   0,   0,   0,   0,   1,  -2,  -2,   2,   3,  -1,  -3,   1,   1,  -2,  -1,   1,   1,  -1,  -1,   1 },
     {   13,  27, -10,   0,  -7,  -1,  -3,  -1,  -1,   0,  -1,   0,   5, -16,  41, -14,   0,  -5,   0,  -2,   0,  -1,   0,  -1,  47,  30,  -4, -14,  -6,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   1,  12,  20, -21,   1,  -3,   1,  -2,   0,   0,   0,   0,  -2,  60, -33, -26,   4,  -4,   2,  -2,   9, -15, -46,   6,   6,   2,   2,   1,  -8, -14,  -8,  15,   4,   1,   1,   1,   0, -10,  -1,   9,  -2,   1,   0,   0,  -4,  -4,   3,   3,  -1,  -4,   0,   2,  -2,  -2,   1,   1,   0,  -2,   0,   1 },
     {   22,  43,  49, -21, -11,  -6,  -5,  -3,  -2,  -2,  -1,  -1,  -8, -14,  56,  -9,  -8,  -1,  -3,  -1,  -1,   0,  -1,   0,  21, -25,  46, -20,  -2,  -2,   0,  -1,   0,  -1,   0,   0,  11,   4, -13, -12,   0,   1,   1,   0,   0,   0,   0,   0,  -3, -36,  13,   2,   3,   2,   0,   1, -16,  22,  13, -13,   1,  -3,   0,  -1,  -2,  11, -13, -10,   3,  -1,   1,   0,  -5,   8, -12,  -2,   5,  -1,   1,   0,  -1,   1, -10,   2,  -2,   2,  -6,   1,   0,   0,  -5,   2,  -1,   1,  -3,   1 },
     {   -4,   4, -11,   4,   0,   2,   0,   1,   0,   0,   0,   0, -13, -33,  31, -11,  10,  -1,   3,   0,   1,   0,   1,   0,  16,   9, -21,  -7,   7,  -1,   2,   0,   1,   0,   0,   0, -67,   1, -22,   6,  11,   1,   3,   1,   2,   0,   1,   0, -24, -16, -30,  35,   6,   2,   2,   2,  -7, -28,  32,  34,  -6,   1,  -2,   1,   4,  15,  31,  -4, -11,  -2,  -2,  -1,   4,   8,   4,  -9,  -5,   1,  -1,   0,   3,   6,   1,  -6,   2,   1,  -1,  -3,   1,   2,   0,  -1,   1,   1,   0,   0 },
     {   -7,  -4, -43,   7,   8,   4,   3,   2,   1,   1,   1,   1,  -7,  -2, -23,  16,   2,   2,   1,   1,   0,   1,   0,   1,  26,  45,  34, -22,  -2,  -8,  -1,  -4,  -1,  -2,   0,  -1, -39,  33, -24,  -4,  -1,  -2,   1,  -1,   0,   0,   0,   0, -33, -12,  32,  -3,   0,   2,   1,   0, -15,  45,  -8, -20,   3,  -3,   2,  -2,   8,  -5, -23,   4,   4,   2,   1,   1,   0,  -2,  -4,   7,   2,   0,   0,   0,   4,  -8,  -2,   6,   1,  -1,   2,   2,   2,  -5,   0,   2,   0,  -1,   2,   0 },
     {   -2,  -5, -20,  46,  -9,   4,  -4,   1,  -1,   0,  -1,   0,  19,  41,  33,  21, -12,  -8,  -5,  -4,  -3,  -2,  -1,  -1,  -3,  27,  29,  21, -19,  -3,  -6,  -2,  -3,  -1,  -1,  -1,   2,   7,  45, -14, -16,  -4,  -3,  -2,  -1,  -1,  -1,  -1, -12,  -4,  -8,  -7,  -8,   2,   0,   1,  -3, -20,  38,  -7,  -8,   3,  -2,   1,  -5,  24,   6, -34,  -2,  -2,   0,  -1,  -1,  -1, -17, -13,   6,   1,   1,   0,  -2,   3,  -9,  -5,   0,  -4,  -7,  -1,  -1,   1,  -3,  -1,   0,  -3,  -3,  -1 },
     {   -7,  -6,   5,  -2,   6,  -1,   2,  -1,   1,   0,   0,   0,  -7,  15, -17,  11,  -3,   2,   0,   0,   0,   0,   0,   0, -15, -15,  39,  -3,   0,   0,   0,   0,   0,   0,   0,   0,  21, -21,   4, -26,   3,   0,   3,   1,   1,   0,   0,   0, -77,  -6, -26, -11,  18,   3,   6,   1, -29, -11, -22,  28,  17,   3,   3,   0, -11,   0,  20,  32,  -2,  -2,  -3,  -1,   8,  14,  16,   0, -10,  -3,  -2,  -1,   2,   7,   2,  -6,   5,   4,  -1,  -3,   1,   1,  -1,  -1,   2,   2,   0,   0 },
     {   10,  14,  33,  42,  -8,  -7,  -6,  -4,  -3,  -2,  -1,  -1,   4,  -4,  13,  62, -16,  -2,  -6,  -1,  -3,  -1,  -1,   0,   4,  12,   2,  14, -14,  -3,  -2,  -1,  -1,  -1,   0,   0, -24, -10, -17,  22, -14,   6,  -1,   3,   0,   1,   0,   1,  21, -11,  21,  -7,  -8,   2,  -1,   0,  -3, -10, -32,  -1,   3,   2,   2,   1,   4, -32,  24,  25,  -7,   4,  -3,   1,   3,   7,  29, -17, -12,   0,  -3,   0,   1,   0,   4, -17,  -3,   5,   1, -11,   0,   0,  -2,  -6,  -1,   2,   0,  -5 },
     {    8,  16,  41, -29,   5,  -7,   2,  -4,   0,  -2,   0,  -2,  -5, -16, -49,  35,  -8,   6,  -1,   1,   0,   1,   0,   1,  14,  22,  19,  -2, -11,  -4,  -3,  -2,  -1,   0,  -1,   0,   3,   0,  -5,  -2,  -4,  -3,  -1,  -1,   1,   0,   0,   0, -22,  35,  -8,  11,  -1,   1,  -2,   0,  53,  -3,  14,   9, -17,   0,  -5,   1,  13,   7,  20, -33, -11,   0,  -1,   0,   3,   3, -22, -18,   9,   0,   3,   1,  -6,  -7,  -5,   2,   0,  -1,  -5,  -1,  -2,  -3,   0,   2,   0,   0,  -1,  -1 },
     {   -9, -22, -41, -19,   0,   8,   4,   3,   2,   1,   1,   1,   6,  10,  13,  -8, -25,   7,  -3,   1,  -2,   1,  -1,   1,  -5,  -5,  47, -42, -11,   1,   0,  -1,   0,  -1,   0,   0, -15, -52, -16,   1,   0,  10,   2,   3,   1,   1,   0,   1,  40, -11,  -5, -14,   5,   6,   1,   2,  32, -13, -19,  10,   5,   2,  -1,   1,  10,  -8,  23,  -6,   0,   0,  -2,   0,  -2,   9, -15,  -9,  11,  -3,   3,   0,  -3,  -2,  -3,   9,  -2,   6,  -2,   3,   0,   1,   0,   4,  -1,   2,  -1,   1 },
     {    4,  -2,  -5,  59,  -6,  -1,  -7,  -1,  -4,  -1,  -2,   0,   4,  12,  13,  21, -10,  -6,  -5,  -1,  -1,  -2,  -1,  -1, -10, -34, -17, -20,  -5,   4,   5,   4,   3,   1,   1,   1,  11, -10, -56, -28,  11,   5,   8,   3,   3,   1,   1,   0, -13,  21, -29,  16,  11,   2,   2,   0,  34,  33,   8,   0,  -4,  -5,  -3,  -3,  10,   6, -25,  -5,  -1,  -3,   0,   0,   1, -15,  -3,  12,   2,   2,   1,   1,  -2,  -4,   6,   4,   0,  -3,   6,   1,   0,  -1,   4,  -1,   0,  -1,   3,  -1 },
     {   -1,   3,  20,   0, -24,   4,  -2,   1,   0,   1,   0,   0, -11, -24, -29,  13, -22,   9,   2,   3,   1,   2,   1,   2,  -8,  12, -34,  -6, -30,  14,  -1,   3,   1,   2,   0,   1, -29, -13,   5, -42, -12,  15,   3,   4,   1,   2,   0,   1,   7, -29, -19, -37,  12,  16,   5,   4, -14, -18,   6, -21,  25,   8,   3,   1, -10,  11, -15,  -8,  29,  -1,   2,  -1,  -3,  -2, -13,  20,  18,  -5,   2,  -1,   3,   4,   3,  16,   4,   0,   3,   6,   3,   2,   3,   4,   1,   0,   1,   3 },
     {    0,  -3,   4,  -8,  -8,   1,   2,   1,   0,   1,   1,   0,   1,  -8,  -5,  -7,  12,  -3,   5,  -3,   2,  -2,   1,  -1,   9,  12,  37,  32,  -9,   2,  -5,  -1,  -3,   0,  -1,   0, -12, -27,  20,  -7,   7,  -4,   1,  -1,   0,   0,   1,   0,  -1, -29, -21,  42,  -5,   3,   0,   2,  46, -14,   5, -20,  -6,   2,  -2,   0,  -2, -17, -47,  20,  10,   0,   3,   2,   7, -23,  31,  26, -12,   5,  -3,   1,  -2,   8,  13, -12,   1,  -1,   7,  -7,  -4,   5,  -1,  -8,   0,   0,   1,  -1 },
     {    3,   5,   3,   0,  58, -10,   0,  -7,   0,  -3,  -1,  -1,   5,   3,  11,  36,  58, -10,  -3,  -7,  -2,  -2,  -2,  -1,  -1,  -3,   5,  12,  34, -12,  -1,  -2,  -1,  -2,  -1,   0,  -9, -20, -14,  11,  -3,  -6,   1,   1,   1,   0,   0,   1,   0, -14, -22, -20, -11,   0,   4,   3,  -6, -16, -17,  -9,  -5,   6,   4,   3, -16,  -2, -10, -22,   9,   5,   3,   1, -15,  -2, -26,  20,  19,  -1,   2,  -1,  -2,   0,  10,  26,   2,   4,   6,   5,   2,   2,   6,   4,   1,   2,   2,   0 },
     {    5,   8,   4,  -4,  18,  -4,   1,  -3,  -2,  -2,  -1,  -1,   6,  -1,  -8,  12,  21,  -3,   0,  -3,  -1,  -2,  -1,  -1,   4,  23,   1,  17,   4,  -5,  -3,  -2,  -1,  -1,  -1,  -1,  -5, -23,  18, -26,  10,  -4,  -1,   1,   2,   1,   0,   0,  45, -11, -26, -14,  21,  -8,   3,  -2,  -9,  53,   5,  26,  15, -11,  -1,  -6,  33,  43,   1,  22, -21, -12,  -5,  -4,   7,   0,   1, -11, -24,   0,  -1,   1,   2,  -5, -11, -12,  -5,  -5,  -3,   2,   0,  -4,  -3,   0,  -2,  -3,   0,   1 },
     {    6,   7,  15,  36,  23, -12,  -5,  -4,  -3,  -1,  -3,  -1,  -2,  -2, -23, -13,  39, -13,   3,  -3,   0,  -1,   0,  -1,  -3,   7,   4, -67,   9,  -1,   8,  -1,   1,   1,   1,   0,  -1,  -6,  23, -43,  13,   0,   5,  -1,   1,   0,   0,   0,  16,   6,  35,  10,   2,  -2,  -4,  -3,   7, -26,  21, -10,  -2,  -1,  -2,  -2, -16, -11,   9,   9,   0,   0,  -1,   0,   3,   8,  15,  -6,  -7,  -1,  -1,   0,  -4,  11,  -8,  -7,  -1,   3,  -8,   1,  -3,   2,  -7,   3,   0,   1,  -2,   2 },
     {    4,  -2,   3,  23,  -1,   3,  -9,   1,  -4,  -1,  -1,  -1,   2,  -5,   2, -33,  20,   0,  -4,  -2,  -1,   0,  -1,   0,   6, -12, -19,   6,  11,  -1,  -1,  -2,   2,  -1,   0,   0,   3,  19,  18,  -9, -23,  -2,  -1,  -2,   0,   0,   0,  -1, -29, -51,   5, -30,  -4,   5,   6,   3,  49,   1, -28,   6,  -2,   5,   1,   1,  55, -17,  10,  -5,  -5,   3,  -4,   1,  16,   4, -11, -15,   4,  -1,   1,   0,   1,  -7,  -7,   3,  -4,   3,  -2,   4,  -2,   1,   0,   4,  -2,   1,   0,   1 },
     {   -4,  -3, -18, -39,  26, -11,   6,  -4,   1,   0,   1,  -1,   4,   7,  19,   1,  11, -15,   0,  -4,   0,  -3,  -1,  -2,  -3, -13, -15,  20, -27,   0,  -2,   3,  -1,   1,  -1,  -1,  -7,   6, -34, -37, -33,  -1,   5,   3,   3,   2,   1,   1,  -6,  11,   4, -16, -36,  12,   0,   6,  19,   1,  12, -35,   2,   8,   3,   1, -15,  16,  10,  18,   2,   4,  -1,   1,  17,  14,  26, -22, -10,   4,  -2,   0,  -8,   5, -20, -21,   1,  -2,  -9,   5,  -2,  -1,  -5,   3,   2,  -2,   0,   2 },
     {   -1,  -2, -14, -35,  47, -18,  13,  -6,   3,  -1,   2,  -2,   6,   2,  29,  43, -31,   5,  -5,   1,   0,   0,  -1,  -1,  -4,  -9, -28, -25, -16,   1,   3,   8,   2,   2,   1,   1,   7,  16,  33, -15,  -5,   1,  -1,   2,   0,   0,   1,   0,  -8, -15,  22,  10,  16,   3,  -3,  -1,   7,  -5,  13,  29,   2,  -4,  -7,  -1,  22, -21, -14,   3,  -6,   0,  -1,   0,  -3, -24,   4,  21,   1,   3,  -2,   2,   7,   0,  20,  -3,  -2,  -1,   1,  -8,   0,   1,   3,  -4,  -1,  -1,  -2,  -2 },
     {    3,   2,  -1,  -1,   6,  12,  -6,   5,   0,  -3,   1,   1,   7,  10,  13,  17,  -5,  23,  -5,   2,  -3,  -2,   0,   1,   8,   3,   1,  10,  36,  16,  -5,  -6,  -2,  -1,   0,  -1,   3,  22,  -9,  18,  28,  13,  -5,  -4,  -3,  -3,  -2,   2,  -4,   2,  23, -11,  27,  14,  -9,  -5,  42, -26,   8, -11,  39, -15,  -7,  -1,  -8,  39,   4,  24,  23, -24,   0,  -1,  15,  17,  -1,  19, -12, -23,   0,  -3,  -8,   0,  -7,   1,   1,   0,  -8,   4,   0,   3,   1,   2,   0,   1,  -3,   0 },
     {    0,   2,   1,  -4,   6,  35,   8,   7,  -2,  -1,  -1,  -1,   3,   6,  13,  -4,  24,  26,   1,   0,   1,  -3,  -1,   1,   1,   4,   3,  39,  -6,  23, -13,   0,  -2,   1,  -1,   1,  -3,   2,  -5, -37,  20,  22,  -7,   1,   0,   1,   0,   1,   1,   8,  10,  -1,  31,   4,  -9,  -3,   0,  23,  17,   3,  -5,  -9, -13,  -2,  -4, -53,  30, -12,   5,   4,  -5,   6, -32,  -1,  19,  -6,  23,  -6,   4,   1, -12,   2,  -3,  12,  -5,   9,   0,   7,   7,   4,  -2,   5,   2,   2,   1,   1 },
     {    0,   2,   2,  12,   1, -36,   7,   4,   4,  -3,   3,  -1,  -4,  -1,  -5, -25,  -4, -29,   5,  11,   2,   0,   1,   2,   3,   3,  10,  12,   2, -34,  19,  -2,   5,  -3,   2,  -3,  -4,   5,   0,  14, -44,  19,  -1,   8,  -3,   2,   0,   1,   4,   7,   2, -14,  21,  18,   5,  -3,  21,  26,  19,  27,  12,  12, -12,  -7, -21,  -5,  16,  -4,  25,  -9,  -8,  -1, -30,  13,   9,  34,  -4,  -8,   1,  -2,   0,   6,  21, -12,  -3,  10, -13, -12,   5,  -3,  -4,   0,   1,  -1,  -3,   1 },
     {   -7, -10, -12, -14, -28, -55,   2,  10,   9,   5,   5,   4,  -1,  -2,   2,  13,  -9, -60,   5,   3,   7,   2,   4,   1,   0,  -2,  -6,  21,  15,  -9,   1,  -2,  -2,  -2,   0,  -1,  -1,   6,   0, -18,  52,  13,   9, -10,  -2,  -4,  -2,   1,   0, -14,   9,  -3,  29,   8,   1,  -7,   3,  -7, -16, -12,  -8,  -6,   2,  -1, -11, -10,   2, -15, -16,   8,   3,   4,   6,   0,  -7, -17,  -3,   5,   5,   3,  -9,  -8,  -9,   7,   0,   2,  -1,  14,   1,   3,   3,   5,   0,   2,   0,   0 },
     {    2,   2,   4,  11,  11,  -2,   2,  -6,  -1,   0,   1,   1,   1,  -1,   0, -12,  23, -24,  11,  -7,   3,  -4,   2,  -2,   0,  -5,   9,  21, -56,   5,  -3,   2,   4,   1,   2,   1,   3,   4,  -5,   9,  -9,   5,  15,  -2,   2,  -1,   2,   1,   5,  11,  24,  37,   3,  25,  -1,  -1,  -7, -12,  -8,  27,  40,  -2,  -3,  -4,   5,   1,  -7,  24,   0, -11,  -6,  -1,  15,  -5, -36,  -9,  21,  -7,   1,   2,   0, -11, -16,  37,   4,  -3,  14,   3,  -1,   2,   2,   3,   0,   1,   6,  -4 },
     {    1,   5,   1,  16,  41, -13, -13,  -4,   4,   1,   1,   0,  -5,  -2,  -4, -16, -14,  31,   4,   0,   5,   9,  -2,   2,  -4,   0,  -7,  -6, -33,  11,   2,   8,   6,  -4,   2,   0,  -4,  12,  16,  23,  -1,  12,   8, -10, -10, -14,  -3,   1,  -4, -23, -20,  11,  21,   7,   2, -11,  14,  16, -30, -14, -12, -21,   8,  -6, -47,  17,   0,  -4, -16,   8,   8,   2, -20,  -2,   1, -25, -12,   0,   3,   5, -31,  -2,   0,  -2,   2,   8,  -5,  14,   0,   4,  -1,  -3,   0,  -1,  -7,   0 },
     {    6,   7,  -1,  12,  41,  -7,   1,  -3,  -3,  -3,   1,  -2,   1,  -4,  -5, -18, -19,  -8,   8,  -3,   0,   1,   1,  -2,   4,   3,   3,   9, -21,  -6,  -4,  -4,  -5,   2,  -2,   1,   6,  -2,  -6,  13,  19,   7, -14,  -4,  -4,   0,  -3,  -2,   4,  11, -16,   2,  23,   2,  -8,  -4, -17,  -7,  -4, -24,  -4,  -1,  -5,   8,  37,  -7, -12, -27,  10,   4,   7,   5,  36,  46,  25,   4,  18, -13,  -1,  -1,  29,  31,   4,   4,   7,   5, -13,  -5,   0, -14, -19,  -7,   1, -10,  -9,   1 },
   },
   {
     {  -79,  44,  -1,   4,   0,   1,   0,   1,   0,   0,   0,   0,  74, -35,  -4,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0, -31,   3,  10,  -1,   2,   0,   1,   0,   0,   0,   0,   0,  11,   9,  -8,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,  -8,   2,   2,   0,   0,   0,   0,   2,   4,   1,  -2,   0,   0,   0,   0,  -1,  -2,  -1,   0,   0,   0,   0,   0,   1,   1,   1,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {  -73,   7,  13,   3,   4,   1,   2,   1,   1,   0,   1,   0, -23,  50, -14,   2,  -3,   0,  -2,   0,  -1,   0,   0,   0,  63, -45,  -2,   1,  -1,   1,   0,   1,   0,   0,   0,   0, -28,   7,  14,  -4,   2,  -1,   1,  -1,   0,   0,   0,   0,  14,   6, -12,   0,  -1,   0,   0,   0,  -3,  -8,   4,   2,   0,   1,   0,   0,   4,   3,   0,  -2,   0,   0,   0,   0,  -1,  -2,   0,   1,   1,   0,   0,   0,   3,   1,   0,   0,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0 },
     {  -42,  30,  -6,   2,  -1,   1,   0,   0,   0,   0,   0,   0, -60,  -5,  22,   0,   5,   0,   2,   0,   1,   0,   1,   0,  11,  55, -24,  -2,  -4,  -1,  -1,  -1,  -1,   0,  -1,   0,  45, -50,  -7,   8,  -2,   3,  -1,   1,   0,   1,   0,   1, -15,   6,  20,  -6,   2,  -1,   1,  -1,  10,   3, -14,  -2,   0,   0,   0,   0,  -2,  -6,   4,   4,   0,   1,   0,   0,   6,   0,  -1,  -2,  -1,   0,   0,   0,  -1,  -2,   1,   0,   3,   0,  -1,   0,  -1,  -1,   1,   0,   2,   0,   0,   0 },
     {   19,  11, -18,   1,  -3,   1,  -1,   0,  -1,   0,   0,   0,  45, -19,   8,  -3,  -1,  -2,   0,  -1,   0,  -1,   0,   0,  34, -14, -25,   4,  -2,   2,  -1,   1,  -1,   0,   0,   0, -24, -52,  42,   4,   4,   1,   1,   1,   1,   0,   0,   0, -27,  58,   1, -18,   2,  -4,   1,  -2,   7, -14, -25,  14,   0,   2,   0,   1,  -6,   0,  19,   1,  -3,   0,  -1,   0,   0,   2,  -6,  -5,   1,   0,   0,   0,  -4,   1,   3,   2,   1,   1,  -2,  -1,  -2,   1,   2,   0,   0,   0,  -1,   0 },
     {    0, -41,  22,  -3,   4,   0,   2,   0,   1,   0,   0,   0,  41,  32, -27,   0,  -6,  -1,  -2,  -1,  -1,   0,  -1,   0,  21,  -3,  -9,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  17, -41, -14,  11,  -1,   3,   0,   2,   0,   1,   0,   1, -15, -26,  53,  -3,   0,   0,   0,   0, -14,  44,  -9, -25,   6,  -5,   2,  -2,  -3,  -9, -21,  18,   3,   1,   1,   1,  -1,   2,  13,   1,  -5,   0,  -1,   0,  -1,  -1,  -3,  -4,  -2,   2,   2,   1,   0,   0,  -2,   0,  -1,   1,   1,   0 },
     {   33,  74, -31,  -4,  -8,  -3,  -3,  -2,  -2,  -1,  -1,  -1, -13, -32,  -9,  12,   1,   5,   0,   2,   0,   1,   0,   1,  47, -25,  25, -10,  -1,  -3,   0,  -1,   0,  -1,   0,  -1, -19,   2, -25,   3,   3,   1,   1,   1,   1,   0,   0,   0,  -1, -25,  27,  11,  -2,   1,  -1,   1, -10,  22,   1, -17,   2,  -2,   1,  -1,  -1,  -6, -10,  11,   2,   1,   0,   0,  -2,   2,   8,  -2,  -4,   0,  -1,   0,  -1,  -1,  -3,  -1,  -2,   2,   2,   1,   0,  -1,  -1,  -1,  -1,   1,   1,   0 },
     {   16,  15, -21,   4,  -3,   1,  -1,   1,  -1,   0,   0,   0,  27,  21,  -2,  -9,  -3,  -3,  -1,  -1,  -1,   0,   0,  -1,  25,   6, -32,   8,  -3,   2,  -1,   0,  -1,   0,   0,   0,  31, -22,  -2,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0, -11, -38, -21,  22,   3,   5,   1,   2,   2, -25,  65,   4,  -9,   1,  -3,   0, -12,  28,  -5, -39,   8,  -4,   2,  -1,  -7,  -6, -17,  18,   8,  -1,   2,   0,  -2,   5,   9,   1,  -1,  -3,  -2,  -1,  -2,   2,   3,   0,  -1,  -1,  -1,   1 },
     {    9,  32,  -7,   2,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0, -19,  30, -27,   1,  -2,   1,   0,   0,   0,   0,   0,   0, -44, -53,  20,  22,   0,   6,  -1,   2,   0,   1,   0,   1,  49,  -9,  37, -23,   0,  -4,   0,  -2,   0,  -1,   0,  -1, -37,  18,  -9,  -3,   6,  -2,   2,  -1,  13,  17,  -1,  -4,  -3,  -1,  -1,  -1,  -3,  -4, -20,   3,   4,   1,   2,   0,   7,  -5,  10,   9,  -4,   0,  -2,   0,   0,   0,   0,  -9,   1,  -1,  -1,   3,   0,   1,  -1,  -1,   1,  -1,   0,   1 },
     {  -19, -33,   6,  12,   3,   3,   1,   2,   0,   1,   0,   1, -19, -55,  31,  -5,   5,   1,   3,   0,   2,   0,   1,   0,   6, -10,   7,   9,  -1,   2,  -1,   0,  -1,   0,   0,   0, -23,  -4,  25, -10,   2,  -1,   1,  -1,   1,   0,   0,   0, -12,  16,  12,   9,  -1,   0,  -1,   0,  10,  25,  31, -33,  -6,  -4,  -1,  -2, -13,  27, -51, -13,  18,  -3,   5,  -1,   5,  -8,  -3,  32,  -6,  -1,  -1,   0,   2,   7,   4, -11,   1,  -1,  -5,   2,   1,   3,   0,  -1,   1,   0,  -2,   1 },
     {    6,  31, -31,   7,  -5,   1,   0,   1,   0,   1,   0,   0,   5,  41,  17, -18,   1,  -5,  -2,  -2,  -1,  -1,   0,  -1, -37,  13, -40,  10,   3,   1,   3,   1,   2,   0,   0,   0, -39,  -5,  16,  16,   0,   2,   0,   1,  -1,   1,   0,   1,  48,  -1,  10,  -9,  -6,   0,  -3,   0, -12,  -5,  -4, -21,  11,  -2,   3,  -1,  -4,   8, -33,  24,   6,  -2,   3,  -1,  10, -17,  18,  17, -17,   3,  -3,   1,   3,  -1,   6, -16,   0,  -4,  -1,   4,   1,   0,   2,  -2,   0,  -3,   1,   1 },
     {   -5,   8,   0,  -4,   1,   0,   1,  -1,   0,   0,   0,   0, -17,   0,  -4,  15,  -2,   3,  -2,   1,   0,   1,   0,   0, -38, -24,   1,   1,   6,   2,   3,   1,   1,   1,   1,   0, -19, -23,  23,  29,  -4,   4,  -3,   2,  -1,   1,  -1,   1,  20, -14,  45, -23,  -2,  -2,  -1,  -1, -14,  18,  13,   4,   0,  -1,   0,  -1,  15,  19,  21, -43,  -8,   0,  -1,  -1, -15,  14, -47, -10,  27,  -2,   6,  -1,   6,  -2,   0,  30,   1,   3,  -1, -11,   3,  -1,  -1,   5,   0,   2,  -1,  -2 },
     {  -13, -33, -78,  30,   2,  10,   2,   5,   1,   3,   1,   2,   4,   5,  55,  -6,  -8,  -4,  -3,  -1,  -2,   0,  -1,   0,   8, -30,   5, -15,   7,   1,   5,   0,   2,  -1,   1,   0,  30,  17,  -3,   5,  -3,  -1,  -1,  -1,  -1,  -1,  -1,   0, -10, -16,   2, -18,   5,   1,   1,   0, -23,  15, -11,  16,   8,  -2,   1,  -1,  11,  -2,   6,   6, -12,   1,  -2,   1,  -4,  -1,   4,  -5,   2,   1,   0,   0,   1,   4,   0,  -2,  -2,   0,   0,  -1,   1,   2,   0,   0,  -2,   0,   0,   0 },
     {  -15, -23, -29,  18,   3,   5,   1,   3,   1,   1,   1,   1, -16, -13, -26,  10,   7,   1,   3,   0,   1,   0,   1,   0, -26, -18,  -3,  17,  -1,   5,   0,   1,   0,   1,   0,   0, -27, -31, -13,  16,  11,   4,   2,   1,   1,   1,   1,   1, -12,  -7,  14,  43,  -8,   4,  -3,   1,  18, -22,  32,   3,  -3,   1,  -3,   1,   1,   6,  36,  19,  -9,  -4,  -3,  -1,  16,  11,  32, -32, -20,  -1,  -4,   0,  -5,   8, -10, -20,   5,   3,   3,   1,  -1,   5,  -3,  -6,   2,   1,   1,  -1 },
     {    4,   2,  33,  -4,  -1,   0,  -2,  -2,  -2,  -1,  -1,  -1,  -8, -46,  -5,   1,   6,   2,   2,   1,   2,   1,   0,   1,  11, -18, -28,  26,  -2,   3,  -1,   2,  -1,   1,   0,   0,  31,  -7,  35,  -6,  -9,   1,  -4,   1,  -1,   0,  -1,   0,  35, -29, -15, -18,   8,  -2,   3,   0, -53,   0,   5,  19,  13,  -2,   3,  -1,   9,  17,   8,  23, -15,  -2,  -3,  -1,   3,  -4,  22, -16, -13,   3,  -4,   2,  -1,  10,  -3, -16,  -2,   1,  -3,   2,  -1,   5,  -2,  -2,  -1,   0,  -1,   0 },
     {    2,  15,  22,  -2,  -6,  -3,   2,   1,   1,   0,   0,  -1,  -5,   4,  25, -36,   6,  -2,  -2,   0,  -2,   1,   1,   0,  -5,   2, -41,  23,   4,   1,   7,  -1,   5,  -1,   1,   0, -38,  19,   3,  -1,  -3,   0,   2,  -1,  -3,   0,  -1,   1, -48, -26,  -4,  10,  22,   6,   2,   2,  14,  45,   6,  32, -12,  -7,  -3,  -2,  27, -27,   1, -10, -18,   6,  -3,   2,  -1,   5,  12,  -8,   7,  -1,   1,  -1,   6,   0,  -7,  -3,   0,  -1,   3,  -1,   3,  -1,  -1,  -2,  -1,  -1,   1,   1 },
     {    5,   8,  35, -37,   1,  -2,   4,  -2,   1,  -1,   1,  -1,   0,  -2,  46,  -7,  -1,   1,  -6,   1,  -2,   0,  -1,   0,  -6, -54, -33, -11,  18,   3,   7,   1,   5,  -1,   1,   1,  25,  16, -31,  41,   5,  -5,   2,  -3,  -2,  -1,  -1,   0,  -9,  10,  17,   5, -20,   4,  -3,   2,  18, -24,  -1, -11,   7,   2,   1,   0, -25,   8,   8,   9,   6,  -7,   1,  -2,   7,   9,  -4,   0, -11,   0,   1,   0,   0,  -5,   0,  -7,  -3,   3,  -2,   1,  -1,   0,   2,   2,   0,   1,  -2,  -1 },
     {    6,   8,  16,  18, -20,   4,  -6,   2,  -3,   1,  -2,   0,   9,  14,  32, -27,  11,  -9,   1,  -3,   1,  -2,   1,  -1,   4,   3,  13,   9, -20,   4,  -6,   3,  -2,   1,  -1,   1,   8,  -1,   6, -24,  14,  -8,   5,  -3,   1,   0,   0,  -1,   7,   8,   4,  14, -21,   3,  -5,   2,   2,  -2,   9, -35,  13,  -5,   4,  -2,   2,  15,  -6,   7, -18,   3,  -3,   0,   0, -10,  -5, -59,  20,   3,   3,   2, -17,  11, -45,  34,   7,  -8,  15,   8,  -5,   4, -11,   1,   3,  -4,   4,   2 },
     {    4,  -7, -14,  36,  -4,   2,  -2,   0,   0,   1,  -1,   0,   0, -17, -27, -18,   5,   4,   2,   1,   1,   1,   1,   1,  20,  -1, -24,  37,   6,  -1,   0,  -1,   0,  -1,   0,   1,  23,  36,  20,  17, -26,  -1,  -6,  -1,  -3,   0,  -2,   0,  30,  -4,  22, -17,  -7,   0,  -3,   1,  50,  -7,  -8, -21, -11,  -2,   0,  -1, -14, -35,   7, -15,  13,   4,   3,   2, -10,  20,   4,  -3,   7,  -7,   2,  -2,  -2,  -4, -13,  -1,  -3,   2,   5,   3,  -1,  -1,  -2,  -2,  -3,   1,   1,   0 },
     {   -6,  -4, -17, -14,   9,   0,   3,   1,   1,   1,   0,   0,  -7,  -6, -34,  15,   3,  -4,   3,   0,   0,   1,   1,   0,   5,  19, -18,   4,  -1,   3,   0,   0,   0,   1,   0,   0, -10,  41,  17,   9,   1,  -9,   0,  -3,  -1,   0,  -1,  -1, -53, -20,  10, -32, -14,  12,  -4,   5,   9, -16,  -4,  -2,  41,  -8,   4,  -2, -20,  37, -10,  23, -22,  -6,   2,  -2,  10, -17, -10,  -3,   3,  11,  -1,   2,   1,   0,   4,  24,   7,  -4,  10,  -7,   0,   0,   0,   0,   5,  -3,   4,   0 },
     {   -2,  -5,  11,  44, -19,   4,  -6,   2,  -3,   1,  -1,   0, -13, -15, -17, -46,  17,   3,   7,   1,   3,   1,   2,   0, -12, -24,  -6,  -3,  14,  -2,   2,   0,   3,   0,   1,   0,  -9, -27, -29,  -6,   9,   2,   7,   1,   1,   1,   1,   1,  -3, -14, -19,   3,  12,   6,   1,   1, -15, -14, -22,   6,  21,   1,   3,   0, -24,  -5,  -1,  21,  18,   5,   1,   0,  -4,   8, -12,  30,  16,  -8,   1,  -2,   6,  -3,  18,  43,   7,   4,  23,  -9,   0,   0,   2,   0,   4,   2,   8,   0 },
     {    8,  15,  33,  58, -13,  -6,  -9,  -1,  -5,  -2,  -2,  -1,   3,  20,  14, -15, -11,   2,   1,   0,   1,   0,  -1,   0,   4,  17,  39, -12,  21, -18,  -1,  -4,  -2,  -1,   0,  -1,  -9,   4,  15,  15, -27,   1,  -5,   0,   0,  -2,   0,  -1, -24, -11,  10, -20,   1,  -1,   1,   1,  -6,  -7,   6,   4,  -2,  -2,  -1,   0, -19,  32,  21,   1,  -1,  -5,   2,  -3,  11,  18,  -7,  -5, -32,   0,  -3,   1,   1, -15,  -2, -33,  -3,   8, -13,   3,   2,  -5,  -2,  -5,   1,   3,  -6,   1 },
     {    4,   4,   8,  36, -12,  -1,  -4,  -2,  -3,   0,   0,   0,  -3,  -8, -31,  -8,   2,   1,   6,   1,   3,   0,   0,   0,   3,   3, -20, -29,  24,   0,   2,   1,   1,   1,   1,   0,  16,  15,  -6,  44,   0, -16,  -2,  -4,   0,  -2,   0,  -1,  -9,  40,  -9,  16, -35,   2,  -1,  -1, -23,  14,  12,  -6,  -8,   8,  -2,   2,  49,  -1, -12,  -8,  -8,   1,  -6,   1,  -6, -38,   3,  -5,   2,   4,   2,   2,  -7,   7,  16,   0,   2,  -2,  -5,  -6,  -1,  -1,   1,   4,  -1,  -2,   2,  -3 },
     {   -4,  -1, -19, -14,   4, -11,   2,   1,   4,  -2,   2,   0,   0,   6, -13, -30,  22,   6,   2,   3,  -1,   0,  -2,   0,   4,  10, -13,  31, -17,  -5,   1,  -1,   4,  -2,  -1,   0,  10,  33, -11, -18,   3,   9,  -1,   0,  -1,  -2,  -1,   1, -11,  38,  27,  23,   4, -16,   0,  -3, -50,   6,   2,  -7, -15,  10,   1,   3,   7,  26,  20,   6,  15, -11,   0,  -6,  -3,  30, -22,   7, -24,  -5,   0,  -4,  16, -21,  -3,   3,   2,   6,   6,  -3,   2, -11,  -7,  -4,  -1,  -2,   0,   1 },
     {   -5,  -3, -17, -23,  31, -15,  -2,  -4,   4,  -1,   4,   1, -11, -11, -26, -41, -36,   8,   7,   4,   3,   2,   0,   1,  -7,  -1, -13, -37,  26,  -5,   2,  -1,   0,   2,   0,   0,  -6,  -8,   1,  -2, -10,  -5,  -3,   1,   2,   0,   3,  -1,  -3,   0, -20, -11,   9,  10,   3,   0, -13,   7,   1, -12,  -5,   4,  -5,   0, -11,   7,  -2,  -8,  16,   2,   6,   4,   3,  11,  -3, -17,   5,  -2,   0,  -2, -13,   5, -45,  -2,   0,   0,  -3,  56,   5,  -5,  10,  -4,   1,  -2,  -2,   7 },
     {   -2,   5,   9,   6,  -3,   1,   1,   0,   0,   1,   0,   0,  -3,   3,  14,  23,  -6,  -5,   0,  -2,  -2,  -1,  -1,   0,  -8,   7,  -6, -11, -16,   9,   1,   1,   0,   1,   1,   1, -10,   6,  28,   1,   9,  -6,   0,  -1,   1,  -1,   0,  -1, -36, -19,  -1,  -6, -27,   5,   2,   4, -35, -31,   6, -30,   8,  27,   2,   7,  15, -38,   8,   2,  42,  -3,   0,   1, -29,  35,  21,  12,   7, -12,   5,  -5,  13,  23, -11,   3,  -2,  -3,  -4,   1,   7,   5,  -1,   3,  -1,  -3,   0,   1 },
     {   -1,   0,  -1, -18,  23,  -7,  -5,  -8,  -2,  -1,  -1,   0,   0,   0,   6, -28,   8,  -9,   8,  -4,   5,  -3,   1,   0,   2,  -6,  31,  48,   7, -17, -17,  -2,  -6,   0,  -1,  -1,  -6, -14, -38,   6, -39,  12,   1,   6,   2,   2,   1,   0, -14,  10,  10, -35,  -3,   2,  -3,   4,  -3, -38,  10,  -1,  13,  -6,   2,   0,  42,   0, -12,  -4,   1,   7,   4,  -1, -18, -16,  16,   6,  12,   1,  -3,  -1,  -1,  20,   3,  -3,   4,  -1, -12,   4,   1,   4,   4,   2,   0,   2,  -2,   2 },
     {   -4,  -8,  -9, -33, -39,   4,   8,  14,   6,   0,   1,   2,  -2,  -3, -12, -51,   8,   3,  -2,  10,  -9,   3,   3,   1,   8,  11,  36,  -7,   2,   4,  22,  -9,  16,  -7,  -2,  -2,   9,  -6,  27,  24,  43,  -1,   8,  -2, -11,  -2,  -8,   1,   7, -11,   3,  -4,   1,  12,  -9,   3,  16, -11,   2,  11,   7,  -8,  -8,  -4,  17,  -2,  -9,   6,  -3, -28,   3,  -6, -25,   9,   0,   1, -13,   6,  11,   2,   4,   8,   5, -10, -14,   1, -19,  -5,   2,  -4,   1,   5,  -4,   5,   2,   2 },
     {    4,   8,  -2,  -6,  68,  -1,   3,  -6,  -1,  -3,   1,   0,   4,   6,  12, -12,  -1,  -5,   5,  -3,   1,   1,   1,  -1,   9,  16,  31,  30,  25,   5,  -2,  -5,  -1,  -3,  -2,   0,   5,  -8,  22,  40,  12, -17,  -1,  -4,   0,  -3,   0,  -1,   7,   3, -27,  23, -24,   9,   0,   1, -13,  23,   9,  -1,   7,  -9,  -5,  -3, -22, -16,   5,  10, -11,   3,   0,   0,   5,  21,   0,   7,   4,  -2,  -3,  -2,   5,  -2,   7,  14,   3,  -7,  15, -22,  -3,   5, -11,  -4,   0,  -3,   5,  -2 },
     {    6,   2,   5,  10,  40, -11,  -3,  -9,   2,  -3,  -1,  -3,   3,  -6,   7, -12, -27,  -3,   4,   6,   2,   4,  -3,   0,   2,  -9,  -6,  -7, -18,  -6,   1,   8,  -4,   3,  -2,   0,   4,  -6,  11, -12,  11,   3,  -9,   3,  -3,   2,  -3,   0,   7, -28,  -5,   7,  -3,  -3,   7,   0,  44,   2, -20, -16,   7,  21,   0,   6,  62,  24,  18,  23,  22, -14,  -7, -11,   1, -11, -22,  11, -23,  -3,   3,   3,  -7, -14,   6,   7,   5,   6,   9, -12,  -4,   0,  -8,  -2,   1,   1,   2,  -2 },
     {    3,   3,   2,  13,  -2,   1,   1,  -1,  -3,  -1,   0,   0,   2,  -1,  -2,  22,   4,  -4,  -4,  -1,   1,   1,  -2,   0,  -1,  -2, -13, -13,   2,   2,  -8,  -2,  -5,  -1,   0,  -1,   3,  -8, -17,  -9, -19,   9,   2,   1,   2,   1,   1,   2,  11,  29, -16, -11,   8,  21,  -3,   3,  16,  19,  36,   8,  46, -17, -14,  -2,  14, -11, -10,  15,  -9, -39,  12,   0, -31,  33, -12,   1, -26,  14,   4,  -7,  28, -11, -21,   5, -11, -19,  14,   2,   4,   4,   0,  -9,  -8,  -3,  -3,   4 },
     {   -2,  -2,  -3,  -6, -14, -31, -15,  -5,   4,   2,   3,  -2,  -3,  -5,  -4, -29, -45,   0,   5,   3,  -3,   8,  -3,   1,  -3,  -4,  -5, -13, -39, -17, -23,  -2,   1,   0,   1,   1,   1,   3,   0, -19,  -8, -15,   5,   2,  -1,   1,   1,   2,   9,  10,  12,   1, -21,  25,   1,   2,  -2,  10,  10,  -5,  24, -11,  -8,   2,  -9, -12,  21, -14, -13,   1,  18,   4,  31,  12,  12,   3,  15,  11,  -7,  -8,  13,  15,  38, -11,   0,   0,  -8, -20,  -3,   7, -19,  14,  -3,  -4,   1,   0 },
     {   -1,  -1,   0,  12,   7, -31,  13, -11,   4, -17,  -6,  13,  -1,  -1,  -2,   6,  17, -26,  13,  -9,   2, -17,  14,  -4,  -3,  -8,  -1,   4, -40,  10,   6, -25,  20, -10,  11,  -1,  -2,  -5, -13,  10,  -2, -27,  31, -16,  -5,  12,  -7,   7,   2,  -5,  -7, -18, -13,  29,  10, -10,   0,   9, -11,  11, -31,   6,  -6, -30,  -8,  12,   1,   4,  19,   8,   8, -13, -13,  -4,  -6,  -7,  -4,  22,  -4,  -4,  -5,  -7,   0,  -5,  -1, -22, -10,   4,   0,   6,  11, -17,  -1,   6,   2,  16 },
   },
   {
     {  -79,  60,  -8,   1,   0,   1,   0,   0,   0,   0,   0,   0,  66, -42,  -2,   3,  -1,   0,  -1,   0,   0,   0,   0,   0, -10,  -4,  11,  -3,   1,   0,   1,   0,   0,   0,   0,   0,  -3,   7,  -3,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -58, -28,  54,  -8,   8,  -1,   3,  -1,   2,   0,   1,   0,   5,  60, -44,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0,   0,  35, -36,  -6,  15,  -5,   3,  -1,   2,  -1,   1,   0,   1, -15,   3,  18,  -7,   0,   0,   0,   0,   0,   0,   0,   0,   6,   4,  -8,  -2,   2,  -1,   0,   0,  -2,  -1,   1,   2,   0,   0,   0,   0,   3,   0,  -2,   0,  -1,   0,   0,   0,  -1,  -1,   2,   0,   0,   0,   0,   0,   1,   0,  -1,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0 },
     {   24, -57,  44, -10,   4,  -2,   1,   0,   1,   0,   0,   0,  69, -14, -21,  -5,   1,  -2,   0,  -1,   0,   0,   0,   0, -63,  30,  -3,   8,   0,   1,   1,   0,   0,   0,   0,   0,  -3,   9,  -3,   2,  -3,   1,  -1,   0,   0,   0,   0,   0,   2,  -3,  -1,   1,  -1,   1,   0,   0,  -2,   1,   2,  -1,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   57,  10, -15, -14,   2,  -4,   0,  -2,   0,  -1,   0,  -1,  33, -17, -21,  17,  -5,   2,  -2,   2,  -1,   0,   0,   0,  26, -61,  34,   1,  -4,   4,  -1,   1,  -1,   1,   0,   1, -48,  39,   8, -18,   8,  -3,   2,  -2,   1,  -1,   1,  -1,  -1,   8, -17,   8,   2,  -2,   1,   0,  -3,  -3,   5,   2,  -3,   1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,  -4,   2,   1,   0,   1,   0,   0,   0,  -1,   1,  -1,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0 },
     {    6,  31,   2, -15,  -2,  -1,  -1,  -1,   0,  -1,   0,  -1, -24,  25, -51,  29,  -3,   3,   0,   2,   0,   1,   0,   1, -51, -16,  52,  -3,  -4,   3,  -1,   1,   0,   1,   0,   0,  60,   0, -11, -19,   5,  -3,   0,  -1,   0,   0,   0,   0,  -5, -21,  -3,  10,   3,  -1,   2,   0,  -2,   7,   8,  -3,  -2,   0,   0,  -1,   1,  -1,  -3,   0,   0,   0,   0,   0,   3,   0,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   1,  -1,   0,  -1,   0,   0,   1,   0,   0,   0 },
     {    2,  20,  48, -72,  21, -11,   6,  -6,   2,  -2,   1,  -2, -31, -44,  -6,  45,   0,   2,   0,   2,   1,   1,   0,   1,  23,  34, -21,   6, -17,   4,  -3,   1,  -2,   0,  -1,   0, -12,   7,  -6,  -6,   5,   1,   0,   0,   0,   0,   0,   0,   1,  -6,   6,   4,   0,  -1,  -1,   0,   6,   1,  -2,  -3,   0,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0,   0,   1,  -1,  -1,   1,   0,   0,   0,   0,  -1,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -29, -20,   6,  11,   5,  -1,   1,   1,   1,   0,   1,   0, -34,   0,  28,  -9,   1,   1,   1,   0,   0,   0,   0,   0, -18,   6,  23, -12,   5,  -2,   1,  -1,   1,   0,   0,   0, -25,  75, -45,   0,   3,  -4,   1,  -2,   1,  -1,   0,  -1,  37, -30, -16,  20,  -9,   4,  -2,   2,   3, -10,  17,  -4,  -3,   2,  -1,   0,   6,   1,  -2,  -6,   2,  -1,   0,   0,  -1,   0,  -2,   2,   1,  -1,   0,   0,   4,  -1,  -1,   1,   0,  -1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {    7,  35,  -1,  -7,  -3,  -3,  -1,  -1,  -1,  -1,   0,  -1, -18,   3,  11, -27,  15,  -2,   4,  -1,   2,   0,   1,   0, -44, -10, -37,  64, -10,   4,   0,   2,   0,   1,   0,   1,  -6,  18,  46, -27, -14,   5,  -4,   0,  -2,   0,  -1,   0,  27,  -1, -27, -11,  16,  -3,   1,   0,  -2, -15,   7,  15,  -4,  -2,   1,  -1,   4,   5,   1,  -7,  -2,   1,   0,  -1,   0,  -2,  -1,   2,   1,   0,   0,   0,   2,   1,   0,  -1,   1,  -1,   0,   1,   1,   0,   0,  -1,   0,  -1,   0,   1 },
     {    8,  21, -27,   1,   0,   2,  -1,   0,   0,   0,   0,   0,   8,  32, -31,  30, -20,   4,  -5,   2,  -2,   0,  -1,   0, -11,  30, -14, -25,  10,  -2,   1,   0,   0,   0,   0,   0, -44, -24,  19,  17,   5,  -1,   1,   1,   1,   0,   0,   0,  71, -26,   7,  -5, -11,   3,  -2,   0, -21,  -7,   3,  -6,   8,   0,   1,   0,   1,   4,   3,   2,   0,  -2,   0,   0,  -4,  -1,   0,  -1,   0,   0,   1,   0,   3,  -2,   1,   1,  -2,  -1,   1,  -1,   1,   0,   1,   0,  -2,   0,   1,  -1 },
     {   -8,  -9, -30, -29,  70, -29,  14,  -6,   7,  -3,   3,  -1,  12,  28,  38,  11, -52,  11,  -5,   0,  -3,   0,  -2,   0, -25, -24, -13,   6,  -3,  19,  -6,   5,  -1,   2,   0,   2,  10,   6,  -6,   9,  10, -11,   0,  -1,   0,  -1,   0,   0, -10,  11,  10, -11,  -9,   1,   2,  -1,   2,  -2, -11,   3,   7,   0,   0,   0,   4,  -2,   5,   2,  -4,   0,  -1,   0,   0,  -1,  -2,  -1,   2,   0,   0,   0,   0,   1,   1,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,  -1,   0 },
     {   20,  16,  -5,  -6,   2,  -7,   0,  -2,   0,  -1,   0,  -1,  30,  15,  -2, -24,   3,  -1,   1,  -1,   0,   0,   0,   0,  33,  -9, -35,  28,  -9,   1,  -2,   1,  -1,   0,   0,   0,  26,  11, -32,   5,  -9,   4,  -2,   1,  -1,   1,   0,   0,  -2, -72,  33,  12,   2,   5,  -1,   2, -18,  22,  27, -17,   1,  -3,   1,  -1,  -4,   6, -14,  -3,   5,  -1,   2,   0,  -6,   0,   2,   5,  -1,  -1,   0,   0,  -1,  -1,   0,  -1,  -3,   1,   2,   1,  -1,   0,   0,  -1,  -2,   0,   1,   0 },
     {  -10, -30, -33,  17,  11,   0,   2,   2,   1,   1,   0,   1,  14,  -5,  -3,  21, -18,   5,  -4,   2,  -1,   0,   0,   0,  28,  39,  -6,   1, -23,   7,  -5,   1,  -3,   0,  -1,   0,  22,  -4,  20, -57,  26,  -7,   3,  -1,   1,   0,   0,   0,  -4, -15, -38,  29,  14, -10,   4,  -1,   3, -17,  28,  12, -18,   5,  -1,   0,  -7,  10,   2, -15,   2,   3,  -2,   0,  -1,  -4,  -2,   3,   4,  -2,   0,   1,  -3,   2,   1,  -1,   0,  -2,   1,   1,  -1,   1,   1,  -1,  -1,  -1,   0,   0 },
     {   15,  33,  -6, -14,  -8,   3,  -4,   1,  -1,   0,  -1,   0,  18,  47, -16, -11,  10, -11,   2,  -4,   0,  -2,   0,  -2,   4,  28, -15,  -1,  -9,  -1,   1,  -1,   0,  -1,   0,   0,  -6,  14, -40,  15,   1,   3,  -1,   0,   0,   0,   0,   0, -26,  21, -10,   9,  -4,   1,  -1,   1,  47, -63,   3,   9,  -4,   5,  -2,   3,  -8,  -6,  24,  -7,   1,  -1,   0,   0,  -3,   0,  -3,  -4,   3,  -1,   1,   0,  -5,   0,   0,   2,   2,  -3,   1,   1,  -2,  -2,   2,   0,   1,  -2,   1,   0 },
     {   14,  13,  28, -18,   9, -12,   1,  -3,   0,  -2,   0,  -1,  12,  -5,  16, -18,  -6,   3,   1,  -1,   0,   0,   0,   0,  21,  -8,   2, -28,  29,  -8,   3,  -2,   2,  -1,   1,   0,  42, -12,  -4,  25, -19,  -3,  -1,  -1,   0,  -1,   0,   0,  24,  15, -59,   6,  -7,   8,   0,   1, -48, -20,  26,  18,   4,   2,  -2,   2,   7,  15,   9, -13,  -6,  -1,   0,   0,  -3,  -6, -10,   1,   4,   1,   1,   0,  -2,   1,   2,   4,  -3,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   1 },
     {   -1,  -7,   1,   4, -14,  23, -11,   4,  -2,   1,  -1,   1,   0,  -7,  -7,  32, -34,  12,  -6,   5,  -3,   2,  -2,   1,   4,  -8, -32, -11,  59, -31,  13,  -5,   4,  -2,   1,  -1,  20,  41,  26,  -6, -27,  -2,   2,  -5,   0,  -2,   0,  -1, -20, -26,   6, -20,   2,  14,  -4,   3,  26,  -9,  -9,  18,  10,  -8,  -1,   1,  -7,  -1,  12,  -3,  -9,   0,   3,  -1,  -3,   6,  -2,  -4,   4,   1,   0,  -1,  -1,  -1,   0,   0,   1,   1,   0,   1,  -1,   0,   1,   0,   0,   1,   0,  -1 },
     {    7,  11,  17,  42,  13, -54,  14, -11,   4,  -5,   1,  -3, -14, -24, -35, -30, -24,  61,  -8,  10,  -1,   5,   0,   2,   8,  12,  13,  20,   9, -10, -15,   3,  -5,   1,  -2,   0,  -8,  -9,  -7,   4,  -2, -15,  15,  -3,   3,   0,   1,   0,   4,   2,  12, -10,   0,  13,  -3,  -1,  19, -14,   7,   5,  -3,  -5,  -3,   1, -10,   5,   3,  -9,   2,   3,   1,   0,   3,   4,  -1,   0,   2,  -3,   0,  -1,  -3,  -1,  -1,  -1,   1,   2,   1,   0,  -1,   0,   0,  -1,   1,   0,  -1,   0 },
     {   13,  29,  17,  -2,  -1, -12,   2,  -4,   0,  -2,   0,  -1,  12,  34,  15, -24, -15,  14,  -6,   1,  -2,   0,  -1,   0,   6,  36,   4, -15,  17, -21,   4,  -3,   0,  -1,   0,  -1, -11,  13,  -1, -37,   7,   1,   4,  -1,   1,   0,   1,   0, -12,   3, -22,  15,  10,  -3,  -1,  -2, -12,  47, -40,   8,  -9,  -4,   3,  -1,  16, -45, -11,  18,  -2,   9,  -2,   3,  -1,  -3,  17,   2,  -3,  -1,  -2,   0,   2,  -6,  -4,  -3,  -2,  -1,   0,   3,   1,  -3,  -1,   1,  -1,  -1,   1,   1 },
     {   -1, -16, -19,  -2,   4,  -1,   5,   0,   1,   1,   0,   1,   9,  -6, -13,  -1,  12,  -3,  -1,   0,   2,   0,   0,   0,  28,  17,  24,  15, -30,   1,  -1,  -4,  -1,  -2,  -1,  -2,  29,  23, -18,  -3, -12,   9,  -6,   1,  -3,   1,  -1,   0,  34,  -3,   9, -54,  23,  -6,   0,   2, -21, -13, -43,  34,  14,  -9,   6,  -1,   5, -12,  27,  16, -18,   5,  -2,   0,  -5,   3,   0, -15,   1,   3,  -1,   0,  -1,  -4,   0,   4,  -3,   2,   2,   0,   0,  -2,   1,   1,  -2,   1,   0,  -1 },
     {  -11,  -8, -31,  -4,  22,  -9,   8,  -1,   2,   0,   1,  -1, -11,  -1, -27,   6,   9,  -1,   2,   2,   2,   0,   1,   0,  -7,  18,   6,   1,  -1,  -4,   2,  -2,   0,   0,   0,   0, -29,   1,   1,  34, -44,  25,  -7,   3,  -2,   1,  -1,   1, -43,  -5, -17,  -3,  33,   0,  -3,   3, -12,  21,  36,  15,  -8, -11,   2,  -2,  45, -14,  -2, -18, -12,   8,  -2,   1, -10,  -3,   2,   0,  10,  -2,  -1,   0,   5,   4,   2,   0,   0,  -1,  -1,  -1,   4,  -1,   2,   0,  -1,   0,   1,   0 },
     {    5,  10,  16,  38, -22,  -7,  -2,  -3,  -1,  -2,   0,  -1,  -6, -11, -15,  14,  -3,   1,   4,  -2,   1,   0,   1,   0,  -8, -10, -45, -20, -16,  44, -15,  11,  -3,   5,  -2,   3,  10,  23,   0,  14,  33, -30,  -3,  -1,  -1,  -2,  -1,   0, -16,   1,  -2,  -8,  -9,  -8,  16,  -4, -24,  -4,   4,   8,   3,  12,  -4,  -1,  49, -24,  13,   9,  -6,  -2,  -5,   2, -13,   4,   9, -14,   2,   1,   1,   2,   1,   1,   0,   1,  -3,   2,   0,  -2,   1,  -1,   1,   1,  -2,   1,   2,  -1 },
     {   -4,  -8,   7,  -5,  -5,  12,  -3,   1,  -2,   2,   0,   0,  -6, -12,   9,   0,  -7,   7,  -8,   4,  -1,   1,  -1,   1, -14, -34, -10,  12,   9, -14,   9,  -1,   3,  -1,   2,   0, -17, -47, -34,  -6,  17,   9,   3,   3,   1,   2,   1,   1,   8, -26,   3,  34,  14, -11,   0,  -1,  -1,   1, -12,  38, -25,   1,   3,  -1,   4, -17,  55, -14, -13,  14,  -6,   2,  -4,  23,   2, -21,   9,  -4,   1,   0,   2,   1,   0,   1,  -1,   7,   0,  -3,   1,   1,   0,   0,   0,   4,   0,  -2 },
     {    9,  20,  34,   8,   4,   2, -10,  -4,  -1,  -2,   0,  -1,   3,  15,  26, -13, -11, -14,  11,  -2,   0,   0,  -1,  -1,  -3,   8,  26, -36, -13,  30, -11,   4,  -3,   2,  -1,   1, -20,  -4,  20,   2,   6,  -4,  -5,   2,  -1,   1,   0,   0, -19, -40,   9, -30,  29, -10,  12,   0,  14,   9,  -1,  33,  -4,  -5,   1,  -6, -17,  38,   3, -24, -13,   2,  -1,   0,   4,  -4, -26,   0,  12,   2,   1,   1,   2,  -1,   1,   7,   1,   0,  -5,  -1,  -1,  -1,  -1,   2,   1,   0,  -2,   0 },
     {    4,  10,  15,  10,  19,  35, -55,  15, -10,   4,  -5,   3,  -5,  -4,  -2,  -7, -46,  -6,  27,   0,   1,   1,   2,  -1,   4,  11,  15,  30,   6, -16,  12, -12,   2,  -3,   1,  -2,  -9,  -7, -22, -20,  19,   7, -10,   4,   2,   1,  -1,   1,  -4,  22,  23, -18,  -5,  -1,   2,   1, -24, -17,  21,   2,   2,  -3,   1,   1,  32,   7, -22,  -3,  -3,   0,   2,  -1, -14,  -2,   2,   1,   4,   2,  -1,   1,   0,   6,  -1,   0,   0,  -2,  -3,  -1,   1,  -1,   0,   0,  -1,  -1,   0,  -1 },
     {    4,   0,  -4,   9,  42,  12, -30,   4,  -3,   2,  -2,   0,  -8, -21, -29, -30, -15, -30,  40,  -4,   7,  -2,   5,  -1,   9,   1,   0,  -1,   1,  23,  -9,  -3,   0,  -1,  -2,   1,   9,  -1,  14,  19,  -7,   0, -15,   6,  -2,   0,   0,   0,  16, -14, -29,  13, -21,   6,   8,   1,  35,  25,  -9,  -4,   7,   1,  -7,   0, -24, -35,  18,   5,  -2,  -1,   5,   2,  10,   6,  12,  -7,   0,   0,  -3,   0,  -5,  -1,  -3,  -4,   0,   1,   2,  -2,  -4,   1,   0,   0,   0,   1,   1,   0 },
     {    8,   4,  23,  15,  -9,  -3,  -6,   1,  -2,  -1,   0,  -1,  11,   4,  28,  21, -22,  13,  -7,  -1,  -4,   0,  -2,   0,   5, -12,   5, -13, -18,   8,  -4,   1,  -1,   1,  -1,   0,  18,  -4,   5,   7, -28,  31, -11,   7,  -3,   2,  -1,   1,  34,  11,   3,  -8,  39, -27,   0,   0,  34,  -9,  23, -39, -19,  -6,   6,  -3,   8, -39,  -6,  -2,  15,   9,  -2,   0, -15,  22,   2,  -2,   2,  -8,   0,   0,  -5,   2, -12,  -6,  -3,   0,   0,   4,   1,  -2,  -2,  -1,  -2,   1,  -1,   0 },
     {    1,   2,   6,  -2,  -9,  -4,   5,   1,  -2,   1,  -1,   1,   7,  12,  24,   1,   3,  -1,  -6,  -3,   2,  -1,   2,  -2,   8,  16,  34,  21, -15, -10,  -2,   0,  -3,   1,  -3,   0,   0,  -4,  34,  -7,  -8,  -9,   3,  -4,   0,   0,   0,  -1, -10, -13,   5,   9, -43,  28, -10,   3,   4,   4,   9, -26,  27,  -3,  -7,   3,  32, -11,  39, -14, -12, -14,   3,  -2, -22,  25, -28, -36,   8,   2,   6,  -1,   1,   1, -19,  13,  -2,   0,  -1,  -2,   0,  -2,  -6,   2,   1,   1,  -2,  -2 },
     {    2,   0,  -7,  -7,  23, -10,   8,  -2,   2,  -2,   1,   0,  -2,  -7, -21, -26,  20,  -9,   1,   0,   1,   3,   0,   1,   3,  -2,  -8, -27,   8, -15,  21,  -8,   4,  -3,   3,  -1,  19,  17,  22,  -5,  14,   7, -12,  -1,  -1,   0,  -1,  -1,  19,  14,  39,  11, -15, -17,   2,  -2,  12,   3, -19,   7, -34,  20,  -1,  -5,  23,   0,  -7, -42,  15,   3,  -8,   4, -37,   3, -27,  15,  21,  -9,   8,   0,   3,   0,   2,  19,  -1,   0,  -4,  -1,   1,  -2,   2,   4,  -4,  -1,  -1,   1 },
     {    2,   7,   1,  13, -17,   8,  -9,   3,  -3,   1,  -3,   2,   4,  12,  -2,  26, -16,   4,  -7,   6,  -3,   1,   0,   0,   6,  14,  -5,  32,  -4,   8,  -2,  -4,  -3,  -1,   0,  -1,   4,  24, -13,  19,  -7,  -8,  -3,   1,   0,  -2,   0,   0,   8,  37,   3,   2,  -8,  -9,   3,  -3, -16,  49,   2,  -7, -23,   4,  -1,  -2, -23,   2,  30, -43,   2,   6,  -5,   5,  40, -15, -33,   4,  12,  -5,   3,  -1,  -6, -16,  12,   2,   2,  -1,  -1,  -8,  -4,  -5,   1,  -5,   1,   0,  -2,  -2 },
     {    3,   9,  13,  38,  27, -27,  -2,  -4,  -5,  -4,   0,  -1,  -4,  -1,  -4,  15,   3,  -5,  10,  -5,   4,   0,   0,   1,  -5,  -2, -12, -13,  -3,  -2,  11,  -6,   4,  -3,   3,  -3,  -5,  10,   4, -21, -22,  54, -26,   8,  -6,   3,  -2,   1, -18,   0,  11,  34,  -5, -20,  -3,   5, -30, -19, -22, -17,  23,   3,  13,  -3,  -7,  26,  25,  13,   3,  -7,  -4,  -3,  21,  -4,  -4, -13,  -9,   3,  -1,   1,  -2, -12,  -3,   7,   4,   5,   4,   2,  -3,   1,   0,   1,   1,   0,   1,  -2 },
     {    3,   5,   7,   8,  15,  -4, -17,  23,  -7,   1,  -3,   1,  -1,   0,   2,   0,  10,  22, -31,   7,  -9,   4,  -3,   2,  -4,  -7,  -6, -11, -51, -28,  65, -26,  16,  -2,   5,  -2,   3,   5,  14,  27,  30,  -3, -16,  -4,   3,  -6,   1,  -3,  -7, -13, -18, -16,  -9,  11, -10,  13,   6,   8,  13,   5,  -4,  -1,   9,  -3,  -8,  -5,   0,  10,  -1,  -1,  -4,   0,  17,  -4,   6,  -4,  -5,   1,   3,  -1,  -8,   5,   2,  -8,   4,  -1,  -6,   2,  -1,   0,   4,   0,   2,   0,  -2,  -2 },
     {    9,   9,  14,  24,  22, -12,  -1,   1,  -2,  -2,  -3,  -3,  11,   8,  16,  27,  26, -19,  10,  -9,  -2,  -5,  -1,  -1,   7,   2,   6,   5,   4, -10,  -2,   8,  -8,   4,   0,   1,  12,   5,  10,   4,  -7,  -1,  24, -13,   5,  -1,  -2,  -2,  22,  12,  28,  21, -11,   5, -16,  -2,  23,   9,  15,  30, -21,  -5,  -7,   7,  17,  12,  -2,   5, -42,   9,   2,  -1,   4, -29,  30, -14, -11,   9, -10,   3,  -8,  13,   7, -34,  -2,  -3, -10, -10,  -5,  -1,  -2,  -6,  -5,   0,  -8,  -3 },
     {   -8,  -9, -13, -28, -15, -36, -28,  45, -11,   9,   0,   3,   3,   8,   6,   0,  15,  31,  37, -36,   8,  -4,   0,   0,   2,   7,  10,   4,  22,  -9, -15, -10,   8,  -2,  -1,  -2,   1,  11,  19,  16,  10, -13,  -5,  14,  -9,   1,  -2,  -1,  -5,  -4,  11,   3,  12, -17,  20, -12,  -8,  -3,  10,  -3, -21,  19,   0,  -8,  -7,   8,  11,  24,   1,   8, -17,   4,   6,   8,   3, -22,  -7,  -6,   1,   5, -10,   3,  -8,   4,  10,   0,  -2,   5,  -3,  -4,   0,  -2,   1,  -2,  -2,  -3 },
   }
 },
 {  //10
   {
     {  116, -45,  -6,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0, -20,  -3,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0, -16,   5,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -7,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   3,  -1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   15,  12,  -9,   0,  -2,   0,  -1,   0,  -1,   0,   0,   0, 107, -41, -10,  -3,  -3,  -1,  -1,   0,  -1,   0,   0,   0, -34, -21,  12,   2,   3,   1,   1,   1,   1,   0,   0,   0, -23,  11,   8,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  -9,   8,   0,  -1,   0,   0,   0,   0,  -4,   3,  -1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0 },
     {   16,  -7,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  25,  30, -16,  -1,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0,  92, -42, -19,   1,  -3,   0,  -1,   0,  -1,   0,   0,   0, -29, -40,  17,   4,   4,   2,   2,   1,   1,   0,   1,   1, -20,   8,  14,  -1,   1,   0,   0,  -1,  -9,   7,   1,  -3,   0,   0,   0,   0,  -5,   3,  -1,   0,   0,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {   24,  40, -21,  -2,  -6,  -1,  -2,  -1,  -1,   0,  -1,   0,  23,  -2,  -9,  -1,  -2,  -1,   0,  -1,   0,   0,   0,   0,   8,  39, -18,  -1,  -4,   0,  -1,  -1,  -1,   0,   0,   0,  67, -40, -29,   6,  -3,   2,  -1,   1,  -1,   0,   0,   1, -19, -51,  20,   9,   3,   2,   1,   1, -11,   4,  21,  -3,   1,  -1,   0,  -1,  -8,   5,   1,  -5,   1,  -1,   0,   0,  -3,   2,  -1,   1,   1,   0,   0,   0,  -3,  -1,   2,   1,  -1,   0,   1,   0,  -2,  -1,   1,   0,  -1,   0,   1,   0 },
     {   29,  90, -34,  -4,  -9,  -3,  -4,  -2,  -2,  -1,  -1,  -1, -22,  34, -24,   3,  -2,   1,  -1,   0,   0,   0,   0,   0, -17, -39,   4,   7,   3,   2,   1,   1,   1,   1,   1,   1, -12,  11,  12,  -2,   1,  -1,   0,  -1,   0,   0,   0,   0,  26,   4, -15,  -3,  -1,   0,   0,   0,   5, -21,  -4,   6,   1,   1,   0,   1,   1,  -8,   7,   2,   0,   0,   0,   0,   1,  -2,   3,  -1,   0,   0,   0,   0,   1,  -2,   1,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0 },
     {   -1, -30,  10,  -2,   3,   0,   1,   0,   1,   0,   0,   0,  31,  41, -15,  -5,  -4,  -2,  -2,  -2,  -1,  -1,  -1,  -1,  21,   4,  -5,  -3,   0,  -1,   0,  -1,   0,   0,   0,   0,   3,  47, -26,  -5,  -3,  -2,   0,  -1,   0,   0,   0,   0,  37, -41, -42,  14,  -3,   4,   0,   2, -13, -47,  25,  18,   1,   4,   0,   2,  -3,   4,  22,  -6,  -2,  -1,  -1,   0,  -8,   2,   2,  -5,   2,   0,   0,   0,  -1,   2,  -1,  -1,  -3,  -1,   1,   1,   0,   1,   1,   0,  -2,  -1,   1,   0 },
     {    7,   9,  -2,  -5,  -2,  -1,  -1,  -1,   0,   0,   0,   0,  23,  19,   7,  -9,  -1,  -4,  -1,  -2,   0,  -1,   0,  -1,  13,  57, -23,  -9,  -4,  -3,  -2,  -2,  -1,  -1,  -1,  -1,   1,  -6, -12,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -7,  42, -20, -10,   0,  -2,   1,  -1,  21, -22, -57,  14,   1,   3,   1,   2,  -3, -48,  11,  31,  -1,   5,  -1,   2,   3,  -6,  26,  -1,  -7,   0,  -2,   0,  -3,  -2,   5,  -9,  -1,  -1,   0,  -1,  -2,  -2,   2,   1,   0,  -1,   1,   0 },
     {   -8, -18,  15,  -3,   3,  -1,   1,   0,   1,   0,   1,   0,  17,  81, -22,  -7,  -5,  -3,  -2,  -2,  -1,  -1,   0,  -1, -58,  11, -22,   6,   1,   3,   1,   1,   1,   1,   0,   1, -26, -27,  -3,  16,   1,   4,   0,   2,   0,   1,   0,   0,   4,  -8,  29,   4,  -3,   0,  -1,   0,   8,  21,   6, -16,  -2,  -3,   0,  -1,   9,  -2, -21,  -3,   3,   0,   1,   0,   1,  -9,  -1,   9,   1,   1,   0,   0,   4,  -2,   4,   0,   1,  -2,   1,  -2,   2,  -1,  -1,   0,   0,  -1,   0,   0 },
     {   12,  35,  17, -10,  -2,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   0,   5,  13, -10,   0,  -2,  -1,  -1,   0,   0,   0,   0,  29,  27,  20, -19,   0,  -5,  -1,  -3,   0,  -1,   0,  -1, -37,  49, -33, -11,   0,  -3,   1,   0,   0,   0,   0,   0, -21, -19,  -7,   9,   2,   3,   0,   2,  -4,  41,  -1, -14,   0,  -3,   0,  -1,  15,  -1, -48,  -1,   4,   2,   1,   0,   0, -25,  -9,  23,   3,   3,   1,   1,   5,  -7,  10,   7,   0,  -3,   3,  -3,   1,  -3,  -1,   1,   0,  -2,   0,   1 },
     {    3, -16,  17,  -3,   0,   0,   0,   0,   0,   0,   0,   0,  11,  15, -19,   0,  -3,   0,  -2,   0,  -1,   0,   0,   0,   4, -39,  18,   5,   0,   1,  -1,   1,  -1,   0,   0,   0,  74,  21,  11, -14,  -4,  -5,  -2,  -2,  -1,  -1,  -1,  -1,  26,   6,   2, -21,  -1,  -4,   1,  -2, -22,  24, -32, -16,   8,  -2,   3,  -1,  -6, -15, -37,  23,   7,   2,   2,   1,  -3, -14,  15,  21,  -4,   1,  -1,   1,   1,  -1,  10,  -4,  -4,  -1,   1,  -4,  -2,   0,  -1,   1,  -2,   0,   1,   1 },
     {   12,  22,  49, -12,  -5,  -5,  -3,  -3,  -1,  -1,  -1,  -1,  14,  20,  48, -24,  -3,  -6,  -2,  -2,  -1,  -1,   0,  -1,  -4,   5, -21,  -5,  -1,   1,   1,   1,   0,   1,   0,   0,  25,  -6,  27, -11,  -1,  -2,  -1,  -1,   0,   0,   0,   0,  -3,  35, -21, -17,  -1,  -2,   1,  -1,  -5, -20,   7,  -2,   1,   1,   0,   1, -12,  37,  -7, -28,   6,  -4,   0,  -2,  -1,  -1, -46,   4,  12,   0,   2,   0,  -3,  -6,  -7,  18,   3,  -5,  -2,   6,  -2,  -2,   1,   2,   1,  -2,  -2,   2 },
     {   10,  21,  82, -21,   0,  -9,  -3,  -4,  -1,  -2,  -1,  -2, -11, -22,   4, -13,   4,   0,   2,   1,   1,   1,   1,   0,  -9, -40, -33,  -2,  10,   5,   4,   2,   2,   2,   1,   1,  -4,  -3, -35,  19,   5,   3,   1,   1,   0,   0,   0,   0,  -4, -14,  11,  17,  -1,   1,  -1,   1,  -7,   5,  -4,   9,  -1,   1,   0,   1,   4, -19,  21,  18,  -5,   3,  -3,   1,   1,  14,  26, -13,  -7,  -2,  -2,  -1,  -1,   7,  -6, -11,  -2,   5,  -2,  -1,   1,   1,  -2,   0,   0,   2,   0,  -1 },
     {   -5,  -2,   6,  -1,   1,  -2,   1,   1,   0,   0,   0,   0,  -3,  34, -18,   2,  -1,   0,  -1,   0,   0,   0,   0,   0, -16,  -9,  32,   5,   1,  -2,   1,  -1,   0,  -1,   0,   0,  17,  10,   8, -22,   1,  -5,   2,  -3,   0,  -1,   0,  -1, -88,   1, -25,   1,  12,   1,   5,   0, -19,  12,  -4,  32,   1,   1,  -1,   1,  20,  10,  25,   8, -17,  -1,  -4,   0,  11,   4,   1, -22,  -7,   2,  -1,   0,   3,  -4, -11,  -7,   1,   0,   0,   2,   2,   0,  -1,   0,   1,   0,  -1,   0 },
     {    7,  11,  16,   2,  -6,  -2,  -2,  -1,   0,  -1,  -1,   0,  11,  22,  24,  13,  -9,  -2,  -5,  -1,  -1,  -1,  -1,  -1,  12,  18,  54,  -6, -11,  -6,  -5,  -2,  -2,  -1,  -1,  -1,  -8,   3,   3,  -5,  -8,   0,  -2,   1,  -1,   0,   0,   0,  30, -14,  39, -23,  -7,  -3,  -1,   0,  -7,  14, -17, -21,   3,  -3,   3,  -1,   2,  -8,  23,   2,  -4,   0,  -1,   0,  -6,  36,  -5, -46,   0,  -3,   1,  -2,  -7,  -1, -43,  -3,  -3,  -2,  -8,  12,   2,  -5,  -5,   5,  -1,  -1,  -1,   0 },
     {   -6,  -6, -36,   5,  -1,   0,   3,   1,   1,   2,   0,   1,   2,  15,  73, -13,  -5,  -7,  -2,  -3,  -1,  -2,  -1,  -1, -15, -26,  10, -30,   7,  -1,   5,   1,   2,   0,   1,   0, -11, -32, -29,  -3,  14,   4,   5,   1,   2,   1,   1,   1,  13, -10, -15,   4,  10,   0,   3,  -3, -46,   4, -12,  16,  10,   0,   4,  -1,  -5,  14,  -2,  24,  -2,  -1,  -1,   0,  10,   0,  15,   9,  -9,   1,  -2,   1,   7,   1,   8,  -7,   1,   1,  -1,  -4,   0,   5,   0,   1,   0,   0,  -2,   1 },
     {   -5,   2, -13,  -8,   5,   0,   2,   0,   0,   0,   1,   0,  -8, -12, -33,  -7,   6,   4,   4,   2,   1,   1,   1,   1,  -1,  23, -43,  -5,   6,   2,   3,   0,   1,   0,   1,   0, -14,  27,   7,  20,   1,  -1,  -2,  -1,  -1,  -1,  -1,  -1,   4,   7,  20, -26,  -2,  -2,   0,  -2, -78,  -1, -20,  -2,  19,   2,   5,   1,   7,  24,   7,  19,   3,  -1,  -3,  -1,  17,  10, -17,  -5,  -3,  -3,  -2,  -1,   6, -18, -11,  15,   5,  -4,   9,   9,   3,  -4,   4,   2,   1,  -1,   4,   0 },
     {    4,  13,  16,  18,  -6,   0,  -4,  -1,  -2,   0,  -1,  -1,   1,   2,  22,  10,  -2,  -1,  -3,  -1,   0,   0,   0,   0,   2,  24,  -6,  30, -11,   1,  -4,   1,  -2,   0,  -1,   0, -12,  11,  48,   2, -15,  -4,  -5,   0,  -2,  -1,  -1,   0,  -9, -14,   5,  -9,  -9,   1,  -1,   2, -22, -19,  27, -13,   2,   1,   2,   1,  -7,  16, -11,  14,   4,  -1,   1,  -2,  19,  -8,  44,  10, -27,   1,  -6,   0,   4,  19,  13, -55,  -6,   2, -26, -13,  -2,   3,  -6,   0,   1,  -2,  -6,   0 },
     {    0,   3,  34, -23,   4,  -2,   0,  -2,   0,   0,   0,  -1,  -7, -16, -35,   5,   0,   4,   2,   4,   1,   1,   0,   1,  -5,  20,  32,   1,  -9,   0,  -3,  -1,  -1,   0,  -1,   0, -28, -54,  16, -29,  -1,   6,   2,   2,   1,   1,   1,   1,  15, -36, -33, -19,  21,   7,   4,   2,  -4, -12, -18,  13,  25,   1,   2,   0, -18,  13,  -7,  19,  12,  -3,   1,  -1,  -3,   9,  -4,  21,  -1,  -3,  -1,   0,   3,   0,  12,  14,   4,   3,   6,  -2,   1,  -1,   2,   1,   1,   2,   1,   0 },
     {    8,   7,  27,  87, -11,  -8,  -7,  -3,  -4,  -2,  -2,  -1,   2,  -1,  -3,  63, -18,  -5,  -4,   0,  -2,   0,  -1,   0,  -4,  -2, -19, -22, -10,   8,   4,   1,   1,   2,   1,   0,  -3,  -2, -12, -21,   7,   8,   2,   0,   0,   2,   0,   1,   3,   3,   3,  -8,  10,  -1,  -1,  -2,  -7,  -2,  -4,   8,   3,  -2,   0,   0,   7,  -1,   5,  -3,  -7,   0,   0,   1,   2,  -9, -10,   0,   3,   4,   1,   2,   4, -12,  17,  11,   4,   3,  11, -11,  -1,   1,   0,  -6,  -1,   1,   0,  -2 },
     {    6,   3,   7, -15,  -1,  -2,   2,   0,   0,  -1,   0,  -1,   9,  -7,  -1,   5,   8,   0,   1,  -1,  -1,  -2,   0,   0,  11,  15,   6,  20,   3,  -8,  -2,  -2,  -1,  -1,  -1,   0,   6, -21,  10,  11,  -3,  -5,   0,  -1,   1,  -1,   0,   0,  47,  10, -15,   9,  -4,  -5,   0,  -2,  -3,  37,   5,  33, -15,  -7,  -2,  -1,  51,  19,  10,   0, -37,  -3,  -4,   1,  21, -38,  -7, -22, -16,  11,   1,   3, -12, -24,  14,  -9,  -9,   5,   8, -14,  -4,   5,  -6,  -4,  -1,   0,  -4,  -1 },
     {    1,  -5, -11,  14,   7,   5,  -4,   1,  -1,   1,   0,   0,   1,  -5,  28, -18,  -2,   1,  -4,   2,   0,   1,   0,   1,  -2, -16, -36,   2,  -2,   4,   0,   2,   0,   1,   0,   0,  -1,  10,  35,   7, -13,  -3,  -6,   1,  -1,   0,  -1,   0,  -9, -62,  -8, -37,   6,   3,   3,   7,  35,   3, -25,   4,  19,   2,   2,   2,  52,  -7,  -1,  14,   1,  -4,  -2,  -3, -19,  -6,  -6,  -3,   2,  -5,   2,  -1, -13,   2,  -6,  12,  -4,   1,   9,  11,   3,   1,   3,  -1,  -1,  -2,   1,   1 },
     {    2,  -1,  -4,  22,  -5,   1,  -2,   0,  -2,   1,   1,   0,  -1,  -7, -23, -22,   3,   5,   4,   1,   1,   1,   0,   0,  -1,   1, -22, -24,  -7,   6,   1,   2,   2,   1,   0,   0,   2,  -7,  -6, -39,  17,   4,   4,   0,   3,   0,   1,  -1,  15,  11, -21,  16,  15,  -1,   2,   0,  21,  14,  15, -14,   5,  -4,   1,  -3,   6,  27, -26,  18,   4,  -3,   0,  -4,  19,   1,  16, -19, -20,  -1,  -5,   1, -18,   0, -50, -23,   0, -14, -11,  47,   2,  -3,  12,   7,   0,  -4,   1,   3 },
     {    3,   4,  10,  35,  10,   2,  -7,   1,  -3,  -3,   1,   0,  -4,   0, -20, -57,  -9,   7,   8,   6,   2,   0,   2,   1,   2,  13,  24, -59,   9,   5,   9,   1,   0,   1,   1,   0,   5,   2,  43,  30,  33,  -7,  -4,  -6,  -3,  -2,  -1,  -1,   4, -11,   8,  21,   0, -16,  -5,  -1,  -2,   1,  -9,   7, -15,  -4,   2,   2, -10,   0,  13,  -3,  -7,   4,   2,   2, -10,  -5,   0,   2,   0,   1,   1,   2,   0,   8,  13,  -4,  -4,  13,  -2, -14,  -1,   4,  -8,  -3,   0,   3,  -6,   1 },
     {   -5,  -6,  -5, -51, -22,  20,   0,   7,   2,   1,   0,   0,  -2,  -6,  -4,  13, -33,  13,  -9,   3,  -1,   1,   0,   2,  -6,  -5, -13, -51, -28,  10,  -2,   5,  -2,   1,   0,   3,   1,   4,   9, -33,   9,   5,  -1,   5,  -3,   1,  -2,   1,   0,   3,  26,  -6,  20,   0,  -3,   5,  -6, -18,  16,   1,   8, -10,   0,   3,  26, -23,   4, -11,  -1,  -1,   3,  -3,  16, -14, -11,  -6,   4,   6,   5,  -1, -10,   9,  15, -14, -18,  11, -20, -14,  -1,   2,  -9,  17,   1,   2,  -2,   3 },
     {   -2,   3,   0,   8,   0,   2,  -4,  -1,   0,   0,   0,  -1,  -4,  -6,   2,  -5,   0,   1,   2,   0,   0,   0,  -1,   1,  -5,  12, -12,  15,  -5,   1,  -1,   0,   0,   0,   0,   0,  -7, -15,  19, -13,  -1,   3,  -1,   1,   0,   0,   1,   0,   9,   4, -30,   3,   8,  -2,   1,  -1, -53,  26,  16,   9, -16,   0,  -1,   1,   9, -54,  -7, -33,  -3,  11,   1,   6, -38, -11,  16,  -8,  34,   3,   4,   0,   3,  37, -20,   9,   6,  -7,  -8,  18,   2,   2,   5,   3,   0,  -1,   0,   2 },
     {    1,  -1,   8,  17,   1,   2, -11,   2,  -3,   2,  -1,   0,  -2,  -5,  -6, -12,   2,   7,  -2,   6,  -1,   2,   0,   2,   0,  -6,  35,   1,  -5,   8,  -2,   3,  -4,  -1,  -3,   1,   2, -27, -20,  22,  -7,   0,  -2,   5,   0,   1,  -1,   1, -11,  13, -12,   1, -29,  -1,  -5,  10, -22, -42,   8, -37,   3,  22,   4,   8,  55, -16,  -7,  -5,  32,   7,  -3,  -1,  21, -16,  -4,   5,   2, -10,  -2,  -1, -29,  -6,  -1,   2,  -8,   8,  13,  12,   2,   3,   7,  -4,   2,   1,   2,  -2 },
     {    3,   5,   2,   1,  60,  -6,   4, -11,  -3,   0,  -2,  -2,   2,   2,   0,  11,  60, -20,  -1, -10,  -2,  -2,   1,  -1,  -5,  -7,  -5, -30, -29, -43,  -1,   0,   7,   2,   4,   1,   2,  -2,  15, -19, -29, -11,  12,   2,   3,  -2,   0,   1,   0,  -1,  19,  27,   0,   8,  10,  -7,  -7, -21,  -6,   1,  -5,  -6,   2,  -3,  10,  -6, -11,  -3,  -4,  -4,   6,   3,   7,   3,   4,  -1,   3,   5,   8,  -2,  -2,   1,   1,   9,  -5,   1,  -1,   1,  -2,  13,  -5,   9,  -5,   0,   0,   6 },
     {    2,  -1,  -3,  13,  16,  -6,  -7,  -6,  -2,  -1,  -2,  -1,   1,  -1, -13,   4,   7,   5,   0,  -1,  -2,   3,   1,   0,   2,  -4,  -2, -10,  10,  -7,   5,   1,   0,  -1,   1,   0,   2,  -7, -19,  24, -12,   1,  -3,   1,  -2,  -2,   0,   1,   5,  -1, -28, -24,   1,   5,   2,  -2,   7,  23, -16,   2,  -1,   8,  -2,   2,  10,   4,   9, -29,  11,  -7,  -1,  -1,  15,  23, -18,  17,  13,  -4,   1,  -3,  19,   0,   5, -35, -20,   0, -59,  -7,   4, -12,   0,  55,   0,   0,  14,   1 },
     {    2,   4,  -1,  29,  34,  15, -11,   2,  -6,  -3,  -3,  -1,  -4,   1,  -4, -56,   2,   5,  -5,   9,   1,   3,   3,   1,   1,  -2,   7,  26, -19,   6,  -5,   5,  -9,   0,  -2,   2,  -5,   4, -30, -23, -23,  14,   0,   8,   0,   5,  -1,   2,   6,  21,  23, -31,  22,  10,  -5,  11,  -1,  10,  30,  20,  27,  -6,  -6,  -1, -14, -13,  15,   6,  -8,  -2,   2,  -2,  -3,  -5, -16,   2, -12,  -1,   6,   0, -17,   6,  13,  -5,  -8,  10,   1, -12,  -1,  -5,  -6,  -3,   3,  -1,  -3,   0 },
     {   -1,   0,  -3, -11,  24,  -7,   0,  -5,   0,   1,   1,  -2,  -3,  -4,  -3,   7, -35, -12,   0,   1,   3,   0,  -2,  -2,  -5,  -9,  -1, -22, -59,   6,   6,   3,   2,   4,   3,  -1,   0,   1,   1,  36, -29,  10,  -2,   5,  -1,   1,  -2,   1,  -5,  -3, -28,  -7, -15,  18,   3,   3,   5,  23,   0, -21,   7,   9,   1,   0, -36,   5,  24,   1,  -4,   2,  -1,   1,  15, -43,   5, -15,   5,  10,   2,   2,  11,   1,  -8,  -1,   9,  -4,   9,  12,  -5,  -1,   9, -10,   5,   0,   0,  -8 },
     {   -1,   4,  -1, -10,   5, -10,   2,  -3,   2,   1,   3,   0,  -1,   3, -10,  20,  15,   3,   1,  -7,  -2,  -3,   0,  -2,  -2,   6,   7, -24,   6,   7,   5,  -2,   2,  -2,  -2,   0,   3,  -1, -16,  13,  -3,   4,  -1,   1,   1,   2,   2,   1,  -3,  25,   2, -15, -13,   1,  -4,   0,   1,  -4,   7,   0,  24,   5,   1,  -3,  19,  42,   3,  15,  -2, -12,  -5,  -2, -78, -22,  16,  -8,  -9,   2,   3,   3, -20,  36,   7,  -2,  17,  10, -19,  -3,   1, -13, -13,   5,   0,  -3,   2,   2 },
     {   -4,  -5,  -1,   3, -24,  -8,   1,  -4,  -1,   6,  -5,   0,   0,   3,   2,   5,  49,  14,  -3,  -8,  -7,  -2,  -2,   0,  -2,  -2,  -2, -15,  40,  21,   5,   3,   3,  -5,   1,  -2,  -7,  -4,   5, -10, -14, -14,  -9,  -8,  -5,   1,   1,   7,  -6, -15,  -4, -14, -25,  -8,   7,   1,   1,   6,  12,  -5,  20,   8,  -1,   1, -45, -25,   3,   8,  -7,  -1,  -7,   4,  25, -39, -15, -29,   6,   7,   5,   3, -23,   2,   4,   5,  -7,  10,  -2,   6,   1,   8, -14,  19,   4,   4,  11,   6 },
   },
   {
     {   85, -51,   2,  -4,   0,  -2,   0,  -1,   0,   0,   0,   0, -67,  35,   2,   2,   1,   1,   0,   0,   0,   0,   0,   0,  24,  -4,  -7,   1,  -1,   0,  -1,   0,   0,   0,   0,   0, -10,  -3,   5,   0,   1,   0,   0,   0,   0,   0,   0,   0,   3,   4,  -2,   0,   0,   0,   0,   0,  -3,  -2,   1,   1,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {  -71,  15,  13,   2,   3,   1,   1,   0,   1,   0,   0,   0, -47,  44,  -9,   2,  -2,   1,   0,   0,   0,   0,   0,   0,  70, -33,  -6,   0,  -2,   0,  -1,   0,   0,   0,   0,   0, -18,   1,   8,  -2,   2,   0,   1,   0,   0,   0,   0,   0,  13,   2,  -6,   0,  -1,   0,   0,   0,  -2,  -5,   2,   1,   0,   0,   0,   0,   5,   1,   0,  -1,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0,   0,  -1,   0,   0 },
     {   22, -25,   9,  -3,   2,  -1,   1,   0,   1,   0,   0,   0,  67,   1, -24,   1,  -5,   0,  -2,   0,  -1,   0,  -1,   0,  20, -51,  17,   0,   3,   0,   1,   0,   1,   0,   0,   0, -64,  40,   7,  -3,   2,  -1,   1,  -1,   0,   0,   0,   0,  20,  -1, -13,   3,  -2,   1,  -1,   0, -13,  -3,   8,   1,   0,   0,   0,   0,   2,   4,  -2,  -2,   0,   0,   0,   0,  -7,   1,   0,   1,   1,   0,   0,   0,   1,   1,   0,   0,  -3,   0,   1,   0,   1,   1,   0,   0,  -2,   0,   0,   0 },
     {   36,  60, -56,   8, -12,   2,  -5,   1,  -3,   1,  -1,   0,   5, -38,  20,  -2,   3,   0,   1,   0,   1,   0,   0,   0,  52, -17,  -8,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0, -20, -28,  18,   1,   3,   0,   1,   0,   1,   0,   0,   0, -23,  27,   0,  -4,   1,  -1,   0,  -1,   7,  -9,  -4,   3,  -1,   1,   0,   0,  -8,   4,   6,  -2,   1,   0,   0,   0,   1,  -1,  -3,   0,   0,   0,   0,   0,  -4,   2,   2,   0,   1,   0,  -1,   0,  -2,   1,   1,   0,   0,   0,  -1,   0 },
     {   11,  51, -33,   2,  -7,   1,  -3,   0,  -1,   0,  -1,   0, -46,  -9,   9,   8,   1,   3,   0,   1,   0,   1,   0,   1, -30, -16,  36, -10,   7,  -3,   2,  -1,   1,  -1,   1,   0,  -8,  57, -26,  -3,  -2,  -1,   0,  -1,   0,   0,   0,   0,  38, -36,  -9,   9,  -3,   3,  -1,   1, -12,   1,  13,  -3,   1,  -1,   0,   0,  11,   0,  -5,  -1,   0,   0,   0,   0,  -2,  -2,   1,   1,   0,   0,   0,   0,   5,  -2,  -1,   0,  -1,   0,   1,   0,   3,  -1,   0,   0,  -1,   0,   0,   0 },
     {   23,  20,  -9,  -6,  -2,  -2,  -1,  -1,   0,  -1,   0,  -1,  25,  33, -42,   8,  -8,   3,  -3,   1,  -2,   1,  -1,   0,  26, -32,  13,  -2,   1,   0,   0,   0,   0,   0,   0,   0,  55, -19, -21,   3,  -2,   1,  -1,   1,   0,   0,   0,   0, -24, -47,  32,   4,   3,   2,   1,   1, -18,  35,   4, -11,   1,  -2,   1,  -1,   0, -12,  -9,   5,   1,   1,   0,   1,  -7,   4,   8,  -1,   0,   0,   0,   0,  -1,  -2,  -3,  -1,  -3,   2,   2,   1,  -1,  -1,  -1,   0,  -2,   1,   1,   0 },
     {    9,  31, -18,   2,  -4,   0,  -2,   0,  -1,   0,  -1,   0,   7,  65, -28,  -5,  -5,  -2,  -2,  -1,  -1,  -1,   0,   0, -55, -28,  -2,  20,   0,   6,   0,   3,   0,   2,   0,   1,  13, -14,  40, -14,   3,  -4,   0,  -2,   0,  -1,   0,  -1,   1,  29, -29,  -3,   1,  -2,   1,  -1,  14, -31,   7,   9,  -2,   2,  -1,   1,  -5,   9,   3,  -6,   1,  -1,   0,  -1,   6,  -9,  -2,   3,   0,   1,   0,   0,  -1,   1,   3,  -2,   2,  -3,  -1,   1,  -1,   0,   1,   0,   2,  -2,   0,   0 },
     {    8,   5, -16,   5,  -2,   1,  -1,   0,   0,   0,   0,   0,  28,  16,   5, -17,   2,  -5,   0,  -2,   0,  -1,   0,  -1,  28,  44, -49,   6,  -7,   1,  -2,   0,  -2,   0,  -1,   0,  15,  -6, -11,   2,  -3,   1,  -1,   0,  -1,   0,   0,   0,  46, -42, -24,  12,  -1,   3,  -1,   1, -27, -37,  37,   7,   0,   2,   0,   1,  -6,  22,   2, -12,   1,  -2,   1,  -1,  -5, -11,  -4,   7,   1,   1,   0,   1,  -2,   3,   7,  -1,  -3,  -3,   0,   0,  -1,   2,   2,   0,  -2,  -2,   0,   0 },
     {  -10, -36,  -5,  17,   0,   5,  -1,   3,   0,   1,   0,   1,  -4, -30,  16,  -3,   4,   0,   1,   0,   1,   0,   0,   0,  17, -36,  26,   1,   0,   1,   0,   0,   0,   0,   0,   0,  58,  25,   1, -24,   2,  -7,   1,  -3,   0,  -1,   0,  -1, -34,   2, -38,  21,  -1,   4,   0,   2,  13, -22,  36,   1,  -4,   2,  -2,   1, -18,  25,  -6, -14,   4,  -3,   1,  -1,   4,  -3,  -4,   7,   0,   0,   0,   0,  -6,   7,   1,  -4,   2,  -1,  -2,   2,  -4,   4,   1,  -1,   1,  -1,  -1,   0 },
     {   22,  40,  60, -60,  12, -17,   4,  -8,   2,  -4,   1,  -3,  -6, -24, -37,  28,  -4,   8,  -1,   3,   0,   1,   0,   1,  18,  16,   5,  -2,  -6,  -1,  -2,   0,  -2,   0,  -1,   0,  14,  11, -10, -10,   0,  -2,   0,  -1,   0,   0,   0,   0,   1,  10, -24,   4,   0,   1,   0,   0,  14, -29,  -4,  10,  -2,   2,  -1,   1, -14,  -1,  15,  -3,   0,   1,   0,   0,   1,   3,  -7,  -2,   2,  -1,   0,   0,  -4,  -3,   2,   3,   1,   1,  -1,  -1,  -3,  -2,   1,   1,   1,   0,  -1,   0 },
     {    1,  12,  34, -20,   2,  -5,   1,  -2,   1,  -1,   0,  -1, -18, -34,   6,   3,   7,   1,   3,   1,   2,   0,   1,   0, -19, -25, -14,  30,  -7,   7,  -1,   3,  -1,   1,   0,   1, -10, -32,  42,  -7,   3,   0,   0,   0,   0,   0,   0,   0,   9, -12,  10,   2,   2,   0,   1,   0, -39,  36,  35, -15,   0,  -4,   0,  -2,  11,  38, -34, -13,   2,  -3,   1,  -1,   0, -14,  -5,  12,   0,   2,   0,   1,   5,   9,   2,  -4,  -1,  -1,  -4,   0,   2,   4,  -2,   0,   0,  -1,  -1,   1 },
     {    7,  24,  18,  -9,  -3,  -3,  -1,  -2,  -1,  -1,   0,  -1,   5,  30,  16, -29,   8,  -8,   2,  -3,   1,  -2,   0,  -1,   0,  23, -18,   8,  -5,   1,  -1,   1,   0,   0,   0,   0, -36,  27, -21,  -8,   4,  -1,   1,   0,   1,   0,   0,   0, -58, -10,  -5,  31,  -3,   6,  -1,   2,  37,  15,  23, -22,  -5,  -1,  -2,  -1,  -4,  -2, -38,   5,   7,  -1,   2,   0,  12, -17,  14,  12,  -6,   2,  -1,   1,  -1,   3,  -2,  -8,   5,  -6,   0,   5,  -1,   1,   0,  -1,   3,  -3,   0,   1 },
     {   12,  13,  18, -10,  -4,  -3,  -1,  -1,  -1,  -1,  -1,   0,  13,   4,  41, -34,   6,  -7,   2,  -4,   1,  -2,   0,  -1,  15, -12, -18,  17,  -5,   3,  -1,   1,   0,   0,   0,   0,  41,  34,  13, -21,  -5,  -5,  -1,  -2,  -1,  -1,  -1,  -1,  21,  21, -27, -19,   3,  -4,   2,  -2, -24,  21, -42,   5,   3,  -1,   1,   0,  15, -33, -19,  24,  -2,   4,   0,   2,  -8, -10,  20,   4,  -4,   2,  -1,   1,   1,   1,  -3,  -4,  -5,  -4,   2,   3,   1,   0,  -1,   0,  -3,  -2,   1,   2 },
     {    5,  13,  29, -18,   4,  -7,   0,  -2,   0,  -1,   0,  -1,   5,  19,  56, -30,  -1,  -6,  -2,  -3,   0,  -1,   0,  -1, -14, -40, -11,  -6,  13,  -1,   6,   2,   3,   0,   1,   1,   1,  -2, -26,  33, -11,   7,  -4,   2,  -2,   1,  -1,   1, -24,  -9,  13,  -8,   3,   0,   1,   1, -11, -12,   6,  24,  -3,   3,   0,   0,   5,  13,  39, -26,  -4,  -2,  -2,  -1,  -7,  20, -35,  -6,   8,  -2,   3,  -1,   4,  -8,   3,  11,  -2,   6,  -4,  -4,   2,  -3,   0,   2,  -1,   3,  -2,   0 },
     {   -5,  -8, -18,   1,   6,   0,   2,   1,   1,   0,   0,   1, -14, -25, -24,   6,   9,   1,   4,   1,   2,   1,   1,   0,  -7, -15, -57,  54,  -7,   6,  -2,   4,  -1,   1,   0,   1,  -3,  36,  24,  -5, -13,   2,  -4,   1,  -3,   0,  -1,   0, -20, -39,  13,  -4,  12,   0,   3,   1,  26,  -1,  -2,  21,  -7,   1,  -1,   0,   6, -23,  31,  -4,  -6,   3,  -1,   1,  -7,  16,  11, -12,   3,  -2,   1,  -1,   3,  -1,  -2,  -1,   0,   3,   5,   0,   1,   1,   1,  -3,   0,   1,   2,   0 },
     {   -5,   0, -15,   3,   4,   0,   3,   0,   1,   0,   0,   0,   0,  10, -14,  14,  -3,   1,  -2,   1,  -1,   0,   0,   0,   6,  38,  11, -16,   0,  -4,  -1,  -2,   0,  -1,   0,   0,  -8,  15,  15, -21,   0,  -3,  -1,   0,  -1,   0,   0,   0, -49,   4, -31,  -3,  17,  -3,   6,  -1, -44,  24,   2,  37, -11,   4,  -2,   0,  51,  -3,   6, -26,  -7,   2,  -4,   1, -15, -11, -25,   7,  11,   1,   3,   1,  12,  -9,  13,   9,  -4,   2,  -2,  -4,   6,  -4,   2,   4,  -2,   1,   1,  -1 },
     {    2,  -1, -15,  -8,  11,  -3,   4,  -1,   1,  -1,   1,   0,   8,   8,   0,  11, -14,   3,  -4,   2,  -2,   1,  -2,   0,  12,   9, -13,  -6,   7,  -4,   2,  -2,   0,  -1,   1,  -1,  19,  40,  18,   7, -16,   1,  -6,  -1,  -3,   0,  -2,  -1,  18,  11,  31, -38,   2,  -4,   1,  -2,  40,  19,  -9, -15,  -3,  -3,  -1,  -1, -15,  24, -31, -33,   9,  -4,   3,  -1,   0, -15, -48,  27,  11,   1,   3,   0,   1, -16,  16,  11,  -2,   0,  -7,  -2,  -1,  -7,   3,   4,  -1,  -1,  -2,  -1 },
     {  -13, -17, -48, -60,  61, -17,  17,  -2,   7,  -2,   4,  -1,   5,   7,  32,  35, -32,   6,  -8,   0,  -3,   0,  -1,   0,  -8, -12, -13,   0,   1,   8,  -2,   3,  -1,   1,  -1,   1,  -3,  -7, -12,   2,   9,  -2,   3,   0,   1,   0,   1,   0,  -6,   1,  -6,  11,  -2,   0,  -2,   1,  -7,  -4,  -3,   1,   2,   1,   0,   1,   4,   0,  -9,  15,  -1,   2,  -1,   0,   1,  -8,  18,   6,  -4,   1,  -2,   1,   0,  10,   3,  -9,   0,  -2,  -3,   3,   1,   2,   4,  -1,   0,  -1,   0,  -1 },
     {   -1,  -3,   3,  12,  -9,   3,  -3,   2,  -2,   1,  -1,   0, -10, -22, -36,   3,   9,   3,   3,   2,   1,   0,   1,   1, -12, -29, -49, -22,  30,  -4,   9,   0,   5,   0,   2,   0,   8,  -1, -39,  43,  -5,   4,   0,   1,   0,   0,  -1,   0,  -5,  29, -13,   3,  -6,   0,  -1,   1,  -5,  -9,   0,   6,  -1,   2,   0,   2,  26,  -7, -19,   7,   7,   1,   1,   0,  -5, -28,  12,  33,  -4,   2,  -1,   0,   8,   3,  33,  -8,  -3,   4,  -2,  -5,   3,   0,   8,   4,  -1,   0,   3,  -3 },
     {   -5,   1,  -2,   1,  -2,  -3,   1,   0,   2,   0,   1,   0,  -9,   1, -13, -11,   3,   1,   2,   2,   0,   1,   1,  -1,  -4,   6,  -2,  -7,   7,   1,   1,   0,   1,   1,   0,   0,  -9,  26,   9,   3,  -4,  -2,   1,  -1,  -1,  -1,   0,  -1, -34, -16,  14, -30,  10,   1,   3,  -1, -50, -19, -28,  13,  26,  -5,   6,   0, -53,  39,   2,  35,  -6,  -1,   1,  -1,  34, -13,   7,   8,  -9,   4,  -4,   1,  -6,   4,  14,  -3,  12,   2,  -4,  -1,  -4,   0,   2,   2,   5,   0,   0,   0 },
     {    8,  12,  31,  31, -26,   6,  -9,  -1,  -3,   0,  -1,   1,   8,   9,  33,  61, -47,  11, -15,   4,  -6,   2,  -4,   2,  -3,   0, -28, -35,  19,  -1,   4,   1,   1,   1,   0,   0,   4,  15,  31,  -3,   5,  -8,   1,  -4,   0,  -1,   0,  -1, -11, -13,   8,  -8,  -4,   1,  -1,   1,  -1, -12,  11,  -5,   5,  -1,   1,   1,   7,  -7,   1,   1,  -1,   0,  -1,   0,  -2,   0,  21, -12,  -3,  -1,  -1,  -1,  -4,  11,  -9, -16,  -3,  -3,  -6,   3,   0,   2,   0,  -3,  -1,  -1,  -3,  -2 },
     {    1,  -2,  10,   6,   1,  -8,   3,  -2,   1,  -2,   0,  -1,  -1,  -4,  13,  -9,  -9,  10,  -5,   3,  -2,   2,   0,   1,  -2,  -5,  22, -11,  10,  -6,   3,  -4,   2,   0,   1,  -1, -15, -35,  15,  -5, -12,  10,  -3,   4,  -2,   1,   0,   1,   9, -48, -34, -11,  33,  -3,   7,  -1,  30,  11, -14,  31, -11,   1,  -3,  -1, -27, -26,  -5,  -4,  14,   3,   2,   1,   0,   0,  10,  43,  -6,   0,  -2,   0,   7,  -1,  34, -23,  -6,  10, -10,  -1,   3,   0,   7,   0,  -1,   5,  -4,  -2 },
     {    1,   4,   2,   2,   6,  -6,   5,   0,   2,   0,   1,  -1,   2,   7,   4,   3,   4,   3,  -4,  -2,  -1,   1,  -1,   1,  10,  19,  36,  37, -12,  -9,  -7,  -4,  -1,  -2,   0,   0,   4,  15,   5,  30, -36,  13, -11,   5,  -4,   3,  -3,   1,  -9,  27,  29, -19,   4,  -3,   0,  -3, -20, -30,  31,  -7,  -9,   0,   0,   1,   9, -16,   0, -11,  10,  -4,   3,   0, -12,   8,  27,  12, -17,  -1,  -3,   0,   7,  11,  11, -47, -11,   6, -24,  -2,   8,  -3,   3,  -3,  -5,   1,  -4,  -4 },
     {   -4,   1,  -6,  -7,  10,  -6,   2,   0,   2,   0,   2,   0,  -9,  -5, -11, -27,  -6,  16,   0,   4,   0,   2,   0,   1,  -6,   7, -13, -23,  23,  -2,   3,  -1,   2,   0,   1,   0,   0,   8,  32,   1,  12,  -7,   2,  -4,   1,  -2,   0,  -1, -12,   2,  -2,  19, -13,  -2,  -1,   0, -25, -14,   4, -33,   5,   6,   0,   4, -24, -21,  -7,   4,  37,  -6,   8,  -3, -38,  58,  -7,  23,  -3,  -8,  -1,  -4,  26, -18,  -8,  10,  -8,   8,  10,   2,   8,  -2, -10,  -1,  -6,  -1,   2,   6 },
     {   -2,  -3,   5,   4, -25,  17,  -7,   3,  -2,   1,  -1,   0,   4,   6,  19,  36,  11, -11,  -2,  -2,  -1,  -2,   0,   0,  -6, -10,  -3,  24, -40,  13,  -6,   4,  -2,   1,  -1,   1,  -6,  -4, -28, -43,  29,   0,   7,   0,   3,   2,   0,   1,  -4,  -1,  28,  11,   0,  -7,   1,  -2, -13, -29, -21,  -9,   3,   9,  -1,   5,  -3,  -9,   0,   2,  19,  -4,   2,  -1, -16,  -3, -17,  33,  -1,   0,   3,  -1,  10, -13,  39,   9,  -5,  17,   2, -18,   4,  -2,   1,   4,  -1,   4,   3,  -6 },
     {   -7, -11, -10, -51, -32,  29,  -2,   4,   0,   3,  -1,   1,   3,   2,   4,   5,  43, -30,   5,  -5,   4,  -4,   2,  -1,   1,   0,  15, -30,  -2,   5,   4,   0,  -1,   0,   1,  -1,   2,  -3,  32,  17,  -9,   2,  -6,  -1,  -1,   1,  -1,   0, -11, -29,  -1, -22,  -7,  11,  -1,   3,  20, -27,  -6,  -5,  24,   0,   0,   3,  28,   2, -12,   8,   4,  -7,   1,  -1, -28,  -7,  12,   2,   8,   2,   3,   0,   1,  21,  -5,  10,   2,  -6,  -1,  -2,  -4,   4,   1,   1,   0,  -3,  -2,  -1 },
     {    4,   4,  11,  31,  26, -28,   3,  -4,  -1,  -1,  -1,   0,  -4, -11,  -2, -21, -31,  28,   0,   6,  -1,   2,   0,   2,   0,  -1,  16,  21,  -3,  -1,  -2,  -2,  -1,  -1,   1,  -1,  -6, -15, -12,  -8,  -1,  -7,   3,   1,   0,  -1,   0,   0, -13, -18,   4, -34,  -2,  17,  -1,   3,  11, -44,  -9,  -6,  26,  -7,   4,  -1,  30,   4, -27,   6,  -8,   2,   2,  -1, -35, -20,   1,   1,  19,   5,   2,   2,   6,   4,  -8,  29,   4, -10,  22,   1,   1,  10,   4,  -1,   0,   1,   1,   1 },
     {    3,   2,   5,   5,  17, -15,   2,  -1,   2,  -2,  -1,   1,  -1,  -4,  -4, -31,  -2,   7,  12,  -2,   2,  -1,   1,   0,   4,   4,  13, -17,  -9,  23, -11,   4,  -5,   0,  -2,   0,   8,  11,  32,  21,  48, -40,   7,  -8,   1,  -2,  -1,   0,   4, -15,  30,  20, -41,   0,   3,  -5,   7,  -7, -13,  15, -19,  14,  -8,   4,  13,  -6,  13, -13,  -5,   1,  -2,   1,   7, -27,   5,  -2,  -1,   2,   0,   2,  -4,  10,  26, -19,   0,   6, -15, -17,  -5,   1,  -3,   5,   1,   5,  -3,  -5 },
     {    1,   2,   3,  16,   3,  -2,  -4,  -4,   1,   0,   0,   0,   2,   4,  10,  31,  -5, -16,   4,  -8,   3,  -3,   1,  -1,   3,   5,  11,  29, -19,  14,  -9,   5,  -3,   2,  -1,   1,   0,  -1,  12,  57, -17, -11,   4,  -1,  -2,  -2,  -1,   0, -10,  -8, -46,  -1, -43,  22,  -5,   8,  -6,  21,   2, -14,  21,  -3,   0,   1, -20, -18,   2,   9,  -8,  -1,   3,  -1,   8,   2, -15,   1,   4,   5,   1,   3,   2, -22,  20,  18,   0,   6,  18, -20,   1,   1,  -2,  -1,  -3,   3,   4,  -7 },
     {    3,   6,   2,  23,  23, -18,  -6,  -6,  -1,   0,  -1,   0,  -1,   0,  -7,   9,  18, -10,   8,  -6,   4,  -2,   3,   0,  -7,  -8, -13, -34, -64,  41,  -4,   7,   1,   5,   1,   1,   0,   4, -14,  16,  23,  -6,  -7,   4,  -4,   1,  -2,   0,  -3,   0, -17, -41,  19, -18,   9,  -2,   2,   7,  11,   0,  -5,   6,   5,  -2,  -3,   5,  -7,  -1,  14,  -5,   2,   2,  -3,  25,   1,   6,  -3,  -3,   2,  -4,   7,  10, -23, -11,   4, -17,  -8,  19,   2,  -1,   7,   5,  -2,   1,  -1,  -2 },
     {    4,   1,   1,   3,   5,  -7,  -6,   5,  -3,  -3,  -2,   0,   5,  -1,  -1,   1,   0,  -3,   2,  -5,  -1,   0,   2,   0,   6,   1,   0,   2,   3,  -9,  -4,  -4,   2,  -1,   1,  -1,   5,  -3,   3,  -7,  -1, -12,   8,   0,   1,   2,   0,   1,   4,  -7,  -1,   0,  -2,  13,   3,  -4,  24,   0,  -3,  14,  18,   1,  -5,   1,  49,  41,  -1,  40,   1, -28,   3,  -9,  17,  31, -15,   3, -56,   2,  -7,  -2,  14, -42,   3, -25, -15,   0,  -2,  -8,   1,  -3,  -6,  -7,  -4,  -2,   1,   1 },
     {    0,  -5,   0,  -7,   5,  10,  -6,  -1,  -2,   1,  -1,   1,  -1,  -9,  -1, -13, -19,   9,   5,   1,   3,   2,   1,   0,   0,  -7,   3, -12, -21,   7,   8,   1,   1,  -2,   1,   1,   1,  -1,  16,   5, -14,  26, -16,  -2,  -7,   5,   1,   0,  -4, -13,   1,   2,   7, -17,   5,  -1, -12, -24,   2, -29,  -6,  12,   0,   9,  33, -29,   6,   0,  35,  -4,  -1,   1,  63, -23, -13,  -1,  -1,  -6,  -1,  -4, -26, -28,  -2,  13,  -1,  12,  17,  20,  -7,   3,   2, -16,  -2,  -2,  -2,   6 },
   },
   {
     {  -84,  57,  -3,   3,  -1,   1,   0,   0,   0,   0,   0,   0,  66, -31, -11,   2,  -2,   1,  -1,   0,   0,   0,   0,   0, -10, -14,  16,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   2,   9,  -2,  -5,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,  -3,  -2,   2,   1,   0,   0,   0,   1,   1,   1,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   54,   5, -21,  -2,  -4,   0,  -2,   0,  -1,   0,  -1,   0,  29, -72,  27,   3,   3,   0,   1,   0,   1,   0,   0,   0, -49,  44,  17, -17,   4,  -4,   2,  -2,   1,  -1,   1,  -1,   5,  11, -29,   6,   2,   0,   0,   0,   0,   0,   0,   0,  -5,  -5,   6,   7,  -4,   1,   0,   0,   2,   2,  -1,  -3,  -1,   1,   0,   0,  -3,  -1,   1,   1,   0,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   1,  -1,   0 },
     {   51, -18,   7,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0,   0,  44,   1, -38,   5,  -4,   2,  -2,   1,  -1,   1,   0,   0,  -5, -63,  41,  11,  -1,   2,   0,   1,   0,   1,   0,   1, -25,  36,  20, -29,   4,  -4,   1,  -2,   1,  -1,   0,  -1,  -4,   6, -25,   6,   7,  -2,   2,   0,  -5,   2,   6,   6,  -5,   0,   0,   0,   0,   1,  -3,  -2,  -1,   2,   0,   0,  -2,   1,   2,   1,   1,   0,   0,   0,   0,   1,  -2,   0,  -1,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   1,   0 },
     {   23,  -4, -19,   8,  -1,   2,  -1,   0,   0,   0,   0,   0,  56,   5,   7, -15,  -5,  -3,  -1,  -1,  -1,  -1,  -1,   0, -14,  -9, -51,  21,   5,   4,   1,   2,   1,   1,   0,   1,   8, -41,  52,  22, -17,   4,  -5,   2,  -2,   1,  -1,   1,  -8,  16,  16, -44,   5,   1,   0,   0,  -6,   0, -15,   9,  16,  -4,   2,  -1,  -3,   5,   5,   2,  -7,  -4,   1,  -1,  -2,  -1,  -3,   0,   1,   2,   1,   0,  -2,   2,   2,   0,  -1,  -1,  -1,   0,   0,   1,   1,   0,   0,  -1,  -1,   0 },
     {   31,  75, -57,  -3,  -7,  -1,  -2,  -1,  -2,   0,  -1,   0, -37, -20,   4,  21,  -2,   4,  -1,   2,   0,   1,   0,   1,  46, -30,  14,  -7,  -4,   1,  -1,   1,  -1,   0,   0,   0,  -7, -19,  14,  -3,   5,   0,   0,   0,   0,   0,   0,   0,  -5,  11,   8,  -8,   1,  -3,   1,  -1,  -3,   0,  -9,  -1,   3,   0,   1,   0,  -1,   1,   2,   3,  -1,  -1,   0,   0,   0,   0,  -1,  -2,  -1,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
     {  -26, -28,  -6,  18,   3,   4,   0,   2,   0,   1,   0,   1, -22, -37,  54, -13,   1,   0,   1,   0,   0,   0,   0,   0,  12,   3,  11,  -2,   3,  -2,   0,  -1,   0,   0,   0,   0,  -2,  15,  34, -35,  -2,  -2,   0,  -1,   0,   0,   0,   0, -13,  34, -46, -18,  28,  -6,   6,  -2,   1,  -1, -15,  39,  -5,  -6,   1,  -2,   2,   3,   4,  -5, -12,   4,  -1,   0,   1,  -1,  -4,   2,   3,   2,  -1,   0,   1,   2,   0,  -1,   0,   0,  -2,   1,   0,   1,   0,  -1,   0,   0,  -1,   1 },
     {    8,  37,   6,  -2,  -4,  -4,  -2,  -1,  -1,  -1,  -1,  -1, -32,  13, -33,  -3,   9,   3,   4,   0,   2,   0,   1,   0, -59,  15,   8,  41, -11,   2,  -3,   1,  -1,   1,  -1,   0,  55, -12,  18, -34, -14,   7,  -4,   2,  -2,   1,  -1,   1,  -9,  -9, -14,  -1,  29,  -4,   1,  -1,   7,   5,   6,  14, -15,  -8,   1,  -1,   1,  -5,  -2, -10,  -1,   8,   1,   2,   3,  -1,   3,   3,   3,  -1,  -1,  -1,   0,  -1,   0,  -2,   2,   0,   1,   1,   0,   0,  -1,  -1,   1,   0,   1,   1 },
     {  -16, -16,  19,  -9,   3,   2,   1,   1,   1,   0,   1,   0, -30, -19, -10,  32,  -3,   3,  -1,   2,   0,   1,   0,   1, -24, -13,  40,  -7,  -2,   4,  -1,   1,   0,   0,   0,   0,  12,  16,  18,  10, -11,  -3,  -4,   0,  -2,   0,  -1,   0,  17,  -8,  28, -54,  -9,  10,  -3,   3, -10,  16, -27, -15,  49, -10,   3,  -1,   2,   7,  -7,  28, -12, -16,   5,  -3,   0,   2,  -3,  -7,  -7,   8,   3,   0,   2,   0,  -1,   5,   0,   1,  -1,  -3,   1,   0,   0,   2,   0,   1,   0,  -2 },
     {   -3,  26, -36,  13,   3,   1,   0,   0,   0,   0,   0,   0, -19,  37,  25, -15, -11,  -3,  -2,  -1,   0,  -1,   0,   0, -52, -19, -16,  -4,  23,   2,   6,   0,   3,   0,   2,   0, -23,  54,  -1,  28, -10,  -9,  -2,  -3,  -1,  -1,   0,  -1,  28,  -6, -25,  -8, -18,  10,  -1,   3,   5, -22,   9,  12,  10,   4,  -4,   1,   9,  -2,   7,  -6,  -7,  -6,  -1,   1,   0,  -6,  -1,  -1,   3,   3,   3,   1,   3,   1,   2,  -1,   0,  -3,  -1,  -1,   2,   0,   1,   0,   0,  -1,   0,   0 },
     {  -21, -34, -37,  36,  -4,   6,   0,   3,   1,   2,   1,   1, -15, -15,   4,  13,   3,   4,   0,   1,   0,   0,   0,   0, -28, -19,  15,  21, -13,   3,  -2,   3,  -1,   1,   0,   1,   3, -15,  29,  -1,   8,   3,  -2,   1,   0,   0,   0,   0,   5,  24,  18,  18, -28,  -7,  -2,  -1,  13,  -8,  18, -48, -18,  25,  -3,   5,  -2,   6,  -5, -16,  41,  -4,  -6,   0,   2,   5,   0,   6,  -3, -14,   3,  -1,  -1,   4,  -3,  -3,   1,   0,   1,   1,   0,   1,   0,  -2,   1,   0,   1,   0 },
     {    5,  27,  57, -30,  -7,  -7,  -2,  -1,   0,  -2,   0,  -1, -23, -47,  -9, -17,  26,   1,   7,   0,   3,   0,   2,   0,   1,   3, -31,  38,   2,  -4,   2,  -2,   1,  -1,   0,  -1, -46,  15,  22,  14, -12,   4,  -1,   0,   0,   0,   0,   0,  16,  25,  -2,  -1, -15,   0,  -3,  -1,   9, -13,   1, -13,  -5,  11,  -1,   3,   3,   1,   4,  -4,  13,   1,  -2,   1,   2,   2,  -1,   1,  -2,  -5,   1,  -1,   0,   1,   0,  -1,   2,   1,  -1,   0,   0,   0,  -1,   0,   1,   0,   1,   0 },
     {   -2,  12,  12, -31,   4,   0,   3,   0,   1,  -1,   0,   0,   4,  33,  -3,  26,  -2,  -7,  -4,  -2,  -2,  -1,  -1,   0, -39,  14,   8, -34, -11,   8,   1,   3,   1,   1,  -1,   1, -20, -24,   5,  -1,  41,  -5,   4,  -1,   2,  -1,   1,   0, -15,  62,  -8,   7, -13, -18,   4,  -6,  13,  -8, -39,   8,  -7,  14,   1,   1,   5, -11,  11,  15,  -2,   0,  -5,   0,   5,  -3,  -1,  -6,  -4,   1,   0,   1,   0,  -2,   2,   3,   2,  -1,  -2,  -2,   0,  -1,   0,   2,   1,  -1,   0,  -1 },
     {   -6, -10, -37,  -3,   4,   4,   4,   1,   2,   1,   1,   0,  14,   9, -26,  38,   6,  -4,   1,  -3,   1,  -2,  -1,  -1,  10,  58,   0,  19, -37,  -3,  -5,   0,  -1,  -1,  -1,  -1, -47,   3,  10, -25,  -8,  19,  -4,   3,  -2,   2,  -1,   1,  42, -11, -13, -12,  15,   1,  -6,   2,  -6, -25,   7,   5,  10,  -6,   2,   2,   2,   5,  10,   3,  -2,  -5,   2,   0,   0,  -3,   0,  -3,  -3,   3,   1,   0,   0,  -3,   4,  -1,   0,   0,   0,  -1,   0,   0,   1,  -1,   0,  -1,   1,   0 },
     {  -12, -18, -11,  -1,  19,  -2,   3,  -1,   3,   0,   0,   0, -13, -18, -20,  33, -13,   1,   1,   2,   1,   1,   0,   0, -12, -13,  -1,   5,  17,  -2,   1,  -1,   1,   0,   1,   0, -22,  -5,  -1,  35, -12,   1,   2,   1,   0,   0,   0,   1,   9,   3,  31,   4,  12,  -7,  -5,  -2,  -4,  26,  -3,  29, -45, -18,   9,  -4,  10, -14,  16, -30, -30,  43,  -3,   3,   2,   1,   2, -13,  30,   0, -13,   3,   3,   0,   2,  -1,   0,   2,   0,  -3,   1,  -1,   1,   0,   0,   1,   0,  -1 },
     {    1, -11,   0, -31,  18,   4,   1,   1,   1,   0,   0,  -1,   4, -19, -16,  45,  -6, -10,   2,  -1,   1,  -1,   1,   0,  21,   0, -17,  -1,  -3,  12,  -3,   4,  -2,   0,  -1,   0,  55,  13,   8,  17,  -1, -22,  -1,  -4,  -2,  -1,   0,  -1, -12,   1, -30,  -2, -34,  11,   9,  -1,   6, -36,  21,  16,   6,  21, -12,   3,  -1,   6,  28, -25,  -4,  -8,  -3,   3,  -5,   0, -11,  -7,  11,   3,   6,  -1,  -4,   4,   6,   2,   0,   1,  -3,  -5,  -2,   3,   1,   1,   0,  -1,   0,  -1 },
     {   -2,   1,   3, -60,  15,   6,   3,  -1,   0,   0,   2,   0,   5,   2,  56,  15,   3, -12,  -7,  -6,  -1,  -1,  -2,   0, -15, -31, -17,   3, -29,   7,   6,   6,   2,   2,   3,   1,   9,  -2,   7, -14,  23,  21,  -9,   4,  -4,   2,  -1,   0,  28, -23,  10,  12,  12, -21,  -2,  -1, -28,   9,  18,  -6,  -8,  -5,  10,  -4,   3,  26, -17,  -9,   0,   3,  -3,  -3,  -5,  -4,  -6,   6,   2,   1,  -3,   3,   5,   7,  -1,   1,  -4,  -1,  -2,   0,   1,   3,   1,   1,  -2,   0,  -3,   0 },
     {  -12, -14, -18,   2,   3,   7,   3,   1,   0,   1,   0,   0, -16, -22, -21,  -5,  26,  -6,   5,  -1,   4,  -1,   2,   0, -19, -18, -20,   9,   2,   9,   1,   4,   0,   2,   1,   2, -20, -22,  -4,   5,  30, -10,   2,   0,   1,   0,   1,   0, -21,  -1,   0,  36,   4,   8,   1,   0,  14,   8,  33,  20,  24, -25, -13,  -2,  -5,  25,  -1,  19, -41, -32,  16,  -4,   5,  -2,   4, -10, -25,  28,   3,  -2,   1,   3,   1,  -2,   2,  -1,   4,  -1,   1,   2,   2,  -1,   0,   0,   1,  -1 },
     {    7,   7,  26,  42, -28,  -5,  -6,  -3,  -1,  -2,  -2,  -1,  -3, -14, -18,  -9,  -5,  12,   0,   2,   1,   1,   0,   2,  -3,   3,   0, -33,  13,  15,  -8,   5,  -1,   0,  -1,   0,  10, -26,  21,   3,  37, -15,  -7,   1,  -2,   0,  -1,   0,  62, -16, -16,   4,  -7, -26,   9,  -2, -33, -12,   7,  10,  -6,   6,  10,  -4,  -5,  17,  10,  -4, -11,   5,  -6,  -1,  -3,  -2,  -5,  -8,   1,   3,  -1,   5,   0,   2,   4,   3,  -2,   3,  -2,  -3,   0,   3,  -1,   0,   0,  -1,   0,   0 },
     {   -6, -18, -38, -15,   3,   5,   6,   1,   2,   2,   1,   1,   1,  -9, -22, -24,  41,  -4,  -3,  -2,   3,   1,   0,   1,  11,  -6, -13,  24,   1, -18,   4,   1,   3,  -2,   1,  -1,  36,  27, -15,  12,  12,   1,  -3,  -4,  -3,  -1,  -1,  -1,  21,  12,  -8,   7, -15, -29,   3,  -1, -25,   1, -41,   3, -12,  10,  16,  -2,   3, -14,   0,  42,  -5,  13,  -8,  -1,  -1,  -4,  25,  -6, -12,   1,  -6,   3,   0,  -4,   0,  -2,   0,  -1,   6,   2,   1,  -1,   2,   0,   0,   0,   1,   0 },
     {   -3,  -6,  -6, -31,   3,   6,   1,  -1,   2,   0,   1,   1,   4,   9,   8, -32,  27,   1,   1,   0,  -1,   0,   0,  -1,  -3, -10,  54,   5,  12, -31,   1,  -3,   0,  -1,   0,  -1,   4, -52, -18,  15, -24,   3,  19,  -3,   5,   0,   2,   0,  32,  22, -14,  -1,  -2,  27, -11,  -1, -13, -24,  10,   7,   3,  -8,  -7,   6,  -8,  14,  18, -12,  -3,   1,   6,   0,  -3,  12, -15,  -7,   8,  -2,  -1,  -3,  -3,   0,  -3,   7,   0,   4,  -2,   1,  -1,   0,   0,   2,   0,   2,  -2,   0 },
     {   -7, -13,  -5,  -6,  -7,  -5,   9,  -1,   3,   0,   3,   0, -12, -16, -16,  -1,  -4,  17,  -6,   2,  -1,   1,   0,   2, -13, -29, -18, -16,   2,  -3,  12,   1,   3,   1,   3,   2,  -8, -12, -29,  -2,   2,  34,  -7,   6,   0,   0,   0,   1,  -5, -19,   2, -12,  35,  -8,   3,   4,   2,  -6,  -5,  32,  12,  27,  -3,  -1,   9,  -7,  46,   4,  37, -25, -25,   4,  -2,  21,  -5,   5, -11, -32,  19,  -3,   1,   1,   4,   2,   0,   5,  -1,   3,   1,   0,   3,   3,   0,   3,  -1,   0 },
     {    8,   7,  27,  46, -42,  -5,   2,  -2,  -3,   0,  -3,  -1,   8,   7,  30,  31,  19, -17,  -5,  -6,   0,  -4,   0,  -1,   1, -24,  -2,  17, -45, -11,  15,  -1,   1,   1,   2,   0,  13,   7, -26,   5, -12,  29,   2,  -4,  -1,   0,  -1,   0,   0,  14,   8,   3,  -5,  -1, -12,  -1,   7, -17,  -2,   5,  -1,  -3,   2,   1,   2,  -6,  27,   3, -25,   2,   0,   0,   0,   2,   4, -34,  -2,  15,  -1,   1,  -4,   0,  -4,  -1,   1,   1,   1,  -7,  -3,   2,   0,  -1,   1,   0,   0,  -4 },
     {   -3,   5,   0,   3, -11,  11,   3,  -2,  -5,   1,   2,   4,  -4,  11,  -1,  -8,  28, -10, -17,  -5,  -3,   2,  -2,   0,  -3,  20,  24,   4, -14, -16,  14,  -7,  -2,   1,   2,   1, -22,  -1,  19,  15,  -8, -10,  -7,  -6,   0,   0,   0,  -2, -46, -24,   2,  -7, -21,  -6,  24,   4, -46,  24,  12,  16,  -7,  38,  -2, -11,  19,  20,   2,  -8,  -2,  -8, -28,   6,   5, -13,  -8, -20,   8,   2,   8,   8,  15,   1,  -5,   7,  -4,  -8,   0,  -2,  -1,  -3,   2,   2,  -1,   3,  -3,  -2 },
     {   -3,  -8, -12, -18, -46,  31,  -6,  11,  -2,   5,   0,   1,   1,   5, -16, -18,  40,   6, -11,   3,  -5,  -3,  -1,   3,   5,   3,   0, -32,   2,  -7,   3,  -1,   4,   0,   0,  -1,   7,  24,  13,   0,   7,   6, -14,   7,  -2,   3,  -4,   0,  -5,  -2,  26, -11,  12, -24,  -4,   5,  47, -16,   2,  -4,   1, -13,  10,  -2,   3,   1,  -7, -28,  -8,   0,   5,  -3, -10, -10, -32,   5,  26,   6,   1,   1,  -5,   2,   6,  22,   1,  -1,  -1,  -3,   3,   2,  -3,   7,   0,  -4,   2,   0 },
     {   -7,  -6,  -3,  -9, -51,  -6,   6,   3,   3,   3,  -2,   2,  -2,  -1,  -6,  40,   8,  17,  -7,   1,   0,  -2,   0,   1, -11,  -3, -27, -17,  37, -26,  -4,   9,   3,   0,  -1,  -2,   2,   0,   1, -14, -17,   4,  29,  -5,   2,   0,   1,   2, -20,  12, -10,  11,   8,  22, -16,  -9, -40, -11,   7, -10,   6,  -6,   6,  10,  23,  20, -17,  11,  10,  17,   4,  -2,  -7, -21,  11,  18,  -1,  -4, -10,   5,   4,   4,  14,  -8,   3,   0,  -5,  -2,  -1,   3,   0,  -1,   3,  -2,   0,   5 },
     {    6,   6,   6,   4,   7, -11,  -6,   3,  -5,  -8,   3,  -5,  10,  13,   6,  23,   3,  23, -17,  -1, -10,   0,  -2,   2,  13,  15,  11,   9,  25, -36,   7,  -8,   0,  -1,   5,  -3,  18,  19,  16,  21,   3,  10,   1,   1,  -6,   2,  -1,  -1,  19,  19,  13,  10,   9, -15,  -2, -14,   6,  21,   4,  15, -21,  -6, -12,  -1,   4,   3,  -3, -19,  -2, -41,   4,  -2,   2,   2, -23,  13, -56,   4,  18,  -4,  14,  -6,   4, -20,  -5,  -5,  -9,  -2,   2,  -8,   2,  -6,  -1,   1,  -2,  -1 },
     {    1,   5,   0,  20,  27, -25,   3,  -8,   1,  -3,   1,   0,  -6,  -6, -29, -16,  -9, -10,  11,  -7,   5,   2,   5,   0,   1,   4,  -2, -18, -26,  10,  19,  -1,   1,   0,   1,   0,   9,  16, -17,  26,   0,  35, -16,  -1,  -6,  -2,   0,  -2, -18,  41,  -7,  -9,   8, -12, -20,   9, -36, -13,  11,  -2,   4, -18,   8,   8,   8,  33, -22,  -9,   8,   0,  13,  -9,  -1,   3, -25,  18,  10,   0,  -6,  -9,  13,   1,  -1,  15,   4, -10,   4,   2,   1,  -7,   4,   2,   0,   2,  -1,   1 },
     {    5,   3,   4,  15,  17,   4,   4,  -6,   2,   2,  -5,  -4,   2,  -5,  -7,  12,  38, -32,   4,  -2,  10,  -5,   1,   0,   0,  -8,  -5, -35, -17,  -4,  19,   4,   5,  -3,   1,  -3,   2, -18,  14,  17, -19,  13,   6,  -8,   0,   1,  -1,   2,  22, -12, -40,   7,  -5,  13, -15,  -6,  23,  49,  -4,   3,   5,   9,  -7,   2,  11, -22, -21,  -2,   7,  -2,  -5,  11, -14, -16,   6,  29,   3,  -5,   7,   1, -15,   5,  20,   2,  11,   4,  -4, -10,   0,   8,  -1,  -2,   0,  -4,   4,  -1 },
     {    1,   3,   6,   2,  27, -23,   9,   2,   0,  -3,  -1,  -2,  -2,   2,   5,   0,   7,  -7,   1,  -2,  -4,   3,  -2,  -2,  -1,   5,  17, -20,  15, -12,  13,  -6,   3,   3,   0,   0,  -2,   0,  48,  16,  -6,  22,  10, -16,   7,  -8,  -3,  -1, -25, -40,   4,  19,   3, -13, -24,  12,  18, -44,   0,  -2,  -9,  -6,  34,   4, -19,   7,  14,  26,  -1,  14,  -6, -15,  -7,   7,  13,  -3, -17,   1,  -5,   6,  -8,  14,  -5,  -7,   2,   3,  -7,  -2,  -7,   0,   0,   3,   2,  -1,   3,  -7 },
     {   -1,  -5,  -8, -17,  -4, -31,   2, -11,   6,  -3,   4,   2,   3,  -5,  -1, -31, -24,  30,  12,  -2,   4,  -1,  -5,   0,   7,   6,   8,  -8, -16,   5,  -9,   1,  -3,  -6,  -1,  -1,  12,   0,  16,  -5,   3,  21,  -3,  -6,  -4,   0,   2,   1,   7,   6,   8,  13, -11,  12, -24,   9,  17,   5,   6,  -2,   5,   3, -13,  12,  64,  -4,  -1,   0, -20, -10, -17,  12, -19, -31,  13, -36,   7,  -8,   7,  13,  -3,   0,  -2, -16, -12,  -1,  -5,  -3,  -2,   4,  -2,  -1,  -2,   2,  -3,   0 },
     {    4,   3,   6,  12,  27,  44, -29, -23, -32, -11,  12, -11,   0,  -7,  -7,   2,   3,   5,  -9,   7, -16,   5, -17,  17,  -2,  -5, -16, -10,  -1,  -5,  10, -13,  -9,   9,  17, -14,   5,  -2,   0, -19,   3,  -1,   8,  -5,  13,   6,   2,  -5,   6,   0,   5,   2,  -6,  30, -20, -11,   3, -11,  -6,   3,  -5,  -7,  15,  -7,  23,   4,   5,   6,   7,   7,  -3,   4,  19, -10,  -9,  -2, -11, -11,  -5,  15,  11,   0, -29,  20, -21, -21,  25,  -3,  11,  11,  11, -12,  -8,   3,   6, -15 },
     {    0,   0,   7,  -9,  -6,  49, -19,  -1, -13,   8,   1,  -1,  -3,  -3,  -2,  -8, -20,   5,  -2,  -2,  10,  -3,  -3,  -2,  -1,  -2,   8, -24, -20,   0,   5,  10,  15,  -9,   7,  -1,   0,  -2,  12,  16, -43, -12, -36,  30,  -8,  -6,   6,  -2,  -7,   4,   0,  21,   6, -20,  -7,  -4, -18, -15,   4,   0,  -8, -10,  -8,   9,   5,  -5,   8,   9,   5,  -9,  25,   4,  19,  -3,  35,  -3,  -9,   0,   8,  -9, -11,  -9,  16, -25,   7,   8,  -5,   2,   3,   5,   5,  -5,  -3,   3, -16,   0 },
   }
 },
 {  //11
   {
     {  117, -46,  -6,  -5,  -2,  -2,  -1,  -1,   0,  -1,   0,   0, -11,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -16,   3,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -9,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {  -10, -15,  10,   0,   2,   0,   1,   0,   1,   0,   0,   0,-112,  40,  11,   3,   3,   1,   1,   0,   1,   0,   0,   0,  18,  20,  -7,  -1,  -2,  -1,  -1,   0,  -1,   0,   0,   0,  23,  -3,  -6,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  11,  -5,  -2,   0,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   2,  -1,   0,   0,   2,  -1,   0,   0,   2,  -1,   0,   0 },
     {   14,  -3,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  14,  28, -15,   0,  -4,   0,  -2,   0,  -1,   0,  -1,   0, 105, -32, -18,   0,  -4,   0,  -2,   0,  -1,   0,  -1,   0, -10, -37,   9,   3,   2,   1,   1,   1,   1,   0,   0,   0, -21,  -3,  11,   1,   1,   0,   1,   0, -12,   4,   5,  -1,   0,   0,   0,   0,  -6,   5,   1,  -1,   0,   0,   0,   0,  -6,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   1,   0,  -3,   1,   1,   0,  -2,   1,   0,   0,  -2,   0,   1,   0 },
     {   25,  47, -24,  -3,  -6,  -1,  -2,  -1,  -1,   0,  -1,   0,  18,   9,  -9,  -3,  -2,  -1,   0,  -1,   0,   0,   0,   0,  -4,  39, -17,   0,  -3,   0,  -1,   0,   0,   0,   0,   0,  80, -23, -27,   3,  -4,   1,  -1,   1,  -1,   0,   0,   0,  -3, -47,   5,   6,   2,   2,   1,   1, -12, -12,  12,   2,   1,   0,   0,   0, -10,  -3,   7,   0,   1,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0,   0,  -5,  -2,   2,   0,  -2,   0,   2,   0,  -3,  -1,   1,   0,  -2,   0,   1,   0 },
     {   25,  80, -32,  -5,  -9,  -3,  -3,  -2,  -2,  -1,  -1,  -1, -12,  54, -23,  -2,  -4,  -1,  -1,  -1,  -1,   0,   0,   0, -17, -21,  -2,   4,   2,   2,   1,   1,   1,   0,   1,   0, -46,  11,  10,   1,   2,   0,   1,   0,   1,   0,   0,   0,  19,  13, -10,  -1,  -1,   0,  -1,   0,  13, -16,  -8,   2,   0,   1,   0,   1,   6, -13,   2,   3,   0,   0,   0,   0,   3,  -8,   4,   1,   0,   0,   0,   0,   3,  -4,   2,   0,   1,  -4,   1,   0,   2,  -2,   1,   0,   1,  -2,   1,   0 },
     {    4,  35, -18,   3,  -4,   1,  -2,   0,  -1,   0,  -1,   0, -31, -33,  12,   5,   4,   1,   2,   1,   1,   1,   1,   0, -24, -11,   4,   5,   1,   2,   1,   1,   0,   0,   0,   0,   0, -53,  21,   5,   3,   1,   1,   1,   0,   1,   0,   0, -59,  17,  41,  -5,   4,  -2,   1,  -1,   6,  47,  -2, -12,  -1,  -3,   0,  -2,   8,  12, -15,  -4,   0,  -1,   0,   0,   9,   1,  -9,   1,   0,   0,   0,   0,   3,  -1,  -1,   2,   5,   1,  -1,   0,   2,   0,  -1,   0,   3,   1,  -1,   0 },
     {    4,   0,   5,  -3,   0,  -1,   0,  -1,   0,   0,   0,   0,  19,   7,  11,  -9,   0,  -3,  -1,  -2,   0,  -1,   0,   0,  19,  59, -18,  -9,  -4,  -4,  -2,  -1,  -1,  -1,  -1,  -1,   9,  -5,  -5,  -3,   1,  -1,   0,   0,   0,   0,   0,   0, -22,  58, -10, -10,   0,  -3,   0,  -2,  41,   4, -52,   1,  -3,   1,   0,   1,   0, -43, -11,  15,   1,   4,   1,   2,   0, -17,   9,   8,   0,   1,   0,   0,  -5,  -8,   7,   2,  -1,  -2,   2,   0,  -3,  -3,   2,   1,   0,  -2,   1,   1 },
     {   14,  31, -26,   7,  -7,   2,  -3,   1,  -1,   0,  -1,   0, -27, -77,  23,   8,   6,   3,   2,   2,   1,   1,   1,   1,  37, -28,  15,  -3,   0,   0,   0,  -1,   0,   0,   0,   0,  22,  18,   7,  -6,  -1,  -2,  -1,  -1,   0,  -1,   0,  -1,  22,  22, -22,  -6,  -2,  -1,   0,  -1,   8, -30, -12,   9,   0,   2,   0,   1, -16, -11,  20,   5,   1,   1,   0,   0,  -4,  10,   7,  -5,  -1,  -1,   0,   0,  -5,   6,   0,  -3,  -2,   3,  -2,  -1,  -3,   2,   1,   0,  -1,   2,   0,   0 },
     {  -14, -24, -19,  13,   2,   4,   1,   2,   1,   1,   0,   1,  -3,  -7, -12,   9,   1,   2,   1,   1,   1,   1,   0,   0, -21,  16, -31,  14,  -1,   2,   0,   2,   0,   1,   0,   1, -30, -54,   4,  19,   4,   4,   2,   3,   1,   1,   1,   1,  -8,   1,   3,  12,   0,   2,   0,   1,  22, -49,  11,  17,  -2,   4,  -2,   2, -14, -11,  55,   1,  -2,   0,  -1,   0,   4,  24,  20, -17,  -2,  -3,  -1,  -2,  -1,  13,   0, -11,   3,   7,  -3,  -5,   0,   4,  -1,  -2,   3,   3,  -1,  -1 },
     {   -4, -26,   0,   5,   1,   1,   1,   1,   0,   1,   0,   0,   6,  15, -33,  10,  -4,   3,   0,   1,  -1,   2,  -1,   0, -23, -57,   1,  19,   3,   5,   2,   2,   1,   1,   0,   1,  55, -30,  24,   5,  -1,   0,  -1,  -1,   0,  -1,   0,   0,  45,  17,  15, -12,  -3,  -3,   0,  -2,  22,   9, -21, -10,   0,  -1,   0,  -1,   1, -27, -12,   8,   2,   2,   0,   1, -11,  -9,  17,   6,   1,   1,   0,   1,  -6,   5,   9,  -3,  -5,   4,   3,  -3,  -4,   3,   0,  -2,  -3,   2,   0,  -1 },
     {   -9, -16, -48,  17,   3,   7,   3,   3,   1,   1,   0,   1, -14, -25, -47,  22,   3,   7,   2,   3,   1,   1,   1,   1,  11,   6,  -2,   6,   1,   1,   0,  -1,   0,   0,   0,   0, -20,  30, -31,   8,  -1,   1,   0,   0,   0,   0,   0,   0, -17, -33,   9,  14,   1,   2,   0,   1,  -4,  27,  -4,   2,  -2,  -1,   0,  -1,  22, -36, -15,  18,  -3,   4,   0,   1,  -4, -21,  35,  11,  -2,   2,  -1,   1,   6,   4,  20,  -7,  -3,   4,   6,  -7,   2,   3,   2,  -2,  -1,   1,   4,  -1 },
     {   -8, -20, -67,  23,   1,   8,   1,   4,   1,   1,   0,   1,   6,  -3, -21,  10,   1,   1,   1,   0,   0,   0,   0,   0,  11,  36,  -7,   8,  -5,  -1,  -2,  -1,  -1,  -1,  -1,   0,   3,   7,  25,   2,  -3,  -1,  -2,  -1,   0,   0,  -1,   0,  33,  25,   6,  -9,  -5,  -2,  -2,  -2,  13,   0,  17, -14,  -2,  -2,  -1,  -1,  -1,  40, -12, -25,   0,  -4,   1,  -2,   8, -10, -44,   0,   4,   2,   2,   1,  -5, -19,  -7,  15,   1,  -9,   5,   7,  -2,  -4,   5,   1,   1,  -3,   1,   1 },
     {   11,  16,  30,  -9,  -2,  -4,  -1,  -2,  -1,  -1,  -1,  -1,  -1, -33,  22,  -5,   4,   0,   1,   1,   0,   0,   0,   0,  10,  11, -32,  -7,   0,   0,   1,   1,   0,   0,   0,   0, -34, -30, -24,  20,   4,   5,   1,   2,   0,   1,   1,   2,  65, -25,  21,  12,  -4,   1,  -3,   1,  35,  14,  13, -16,  -4,  -4,  -1,  -2,   8,  -4, -22, -10,   1,   0,   0,   0,  -8, -21,  -5,   9,   4,   3,   1,   1,  -7,  -3,  13,   4,  -5,   4,   3,  -3,  -4,   2,   1,  -1,  -2,   1,   1,  -1 },
     {    4,   5,  -3,   3,  -6,   1,   0,  -1,   0,  -1,  -1,   0,   8,  13,   8,  16, -10,   0,  -3,  -1,  -2,   0,  -1,   0,  16,  25,  49,  -8,  -6,  -5,  -3,  -3,  -1,  -1,  -1,  -2,  -5,  12,   4,   3, -10,   0,  -3,   0,  -1,   0,   0,   0,  24,  -9,  51, -12,  -7,  -2,  -3,  -1,  16,  29,  -4, -17,  -5,  -3,  -2,   0,   8, -26,  28,  -6,  -4,   0,   0,   1,  -9,  43,  10, -36,  -1,  -5,   0,  -2,   4,   6, -36, -16,  -5,  -2, -14,   0,   1,  -5, -11,   1,  -3,  -1,  -4,  -1 },
     {   -8, -14, -41,  24,  -8,   9,  -2,   4,   0,   3,   0,   2,   6,  17,  56,  -8,  -5,  -3,  -3,  -2,  -2,  -1,  -1,  -1,  -5,   9,  37, -20,  -3,  -2,  -1,   0,   0,   0,  -1,   0, -19, -39,   9, -16,   4,   2,   1,   2,   1,   1,   0,   1,  21, -21,  -3, -15,   8,   0,   2,   0, -38, -10, -20,  -5,  13,   1,   4,   0, -16,  -9, -18,  19,   9,   1,   2,   0,  -4, -12,  12,  22,   0,   1,   0,   1,   2,   8,  21,  -1,   1,  10,   1,  -9,   1,   4,  -4,  -2,   0,   1,  -2,   0 },
     {   -7,  -6,  -7, -37,  10,   3,   6,   1,   3,   0,   2,  -1, -13, -16, -39, -17,   7,   8,   5,   4,   1,   1,   1,   1,  -6,  -2,  -2, -22,  10,   1,   3,   2,   1,   0,   2,   0, -15, -19, -23,  -4,   8,   4,   3,   2,   2,   1,   1,   0,  11, -10,   4, -15,  10,   0,   2,  -1, -25,   1, -42,  -1,  12,  -1,   3,   0, -23, -15,   2,  10,  12,   0,   5,   1, -17,  36, -31,   2,  13,  -4,   3,  -2,   0, -21, -24,  38,   1,  -6,  20,  21,   5,  -1,  10,   7,   0,   2,   8,   2 },
     {   -2, -14, -21,   7,  -1,   3,  -1,   2,   0,   0,   0,   1,   6,  11,  15,  -8,   0,  -1,  -3,  -1,  -2,   0,  -1,   0,   0, -30,  38, -26,   4,   1,   0,  -1,  -1,   0,   0,   0,   7, -17, -34, -20,   8,   1,   2,   1,   1,   1,   1,   1, -20, -23, -27,  18,  10,   4,   1,   3,  70,  -6,   3,  12,  -5,   0,  -3,   0,  37,   0,   5, -12,  -4,  -3,  -1,   0,   2,  -1, -21,  -5,   6,   0,   2,   0,  -5, -16,  -9,  16,  -3,  -5,  12,   8,  -2,   0,   4,  -2,  -3,   0,   1,   0 },
     {    8,  13,  47,  30, -12,  -6,  -3,  -2,  -4,  -2,  -1,  -1,  -2,  -9, -28,  59, -17,   2,  -4,   3,  -1,   0,   0,   1,   5,  21,  38,  20, -12,  -4,  -5,  -3,  -3,  -1,  -1,  -1,  -9, -13,  20,  -1,  -9,   0,   1,   0,   0,   0,   0,   1, -10, -23, -12, -13,   1,   5,   1,   2,   4, -23, -17,  -3,   6,   4,   3,   1,  -2,  -3,  -2,  -8,   6,   2,   0,   0,   4,  -5, -21,   9,  11,   2,   0,  -1,   0, -19,  20,  22,   4,   8,  22, -10,  -2,   4,   3, -11,   0,   4,   0,  -6 },
     {   -3,  -2,   0, -64,  10,   3,   7,  -1,   3,   2,   1,   0,  -6, -13, -34, -38,  13,   4,   4,   0,   3,   2,   1,   2,   9,  24,  44,  -8,  -1,  -1,  -3,  -3,  -2,  -1,  -1,   0,  -3,  -5,  49,  -8,   0,  -6,  -2,  -2,  -1,  -1,  -1,  -1,  -6, -20,   6,   0,   2,   0,  -4,   1,  25, -24,  18,  -1,   0,  -1,   0,   0,  -3,   4, -11,   1,   3,   1,   1,   0,   2, -19,  12,  11,   0,   3,   0,   1,  -8,  16,  14, -11,   0,   8, -14,  -7,  -4,   6,  -5,   3,   2,   1,  -2,   3 },
     {   -5,  -3,  -7,   3,  -1,   3,   0,   3,   0,   1,   1,   0,  -2,  15,  17,  -6,  -4,   0,   2,  -2,  -2,   0,  -1,  -1,  -4, -14,   3,   3,  -1,   3,   1,   1,   1,   1,   0,   0,   2,  36, -11,  15,   2,  -2,  -1,  -3,  -1,  -1,   1,   0, -26,   5,  41,  17,  -4,   0,  -1,   0,  24, -17,  18, -19,  -9,  -1,   1,   0, -64, -19,  -5, -19,  13,   2,   5,   0, -40,   4, -15,  20,  16,  -1,   2,   2,  -1,   6,  20,  23,   8,  17,  16, -12,   4,   6,  -6, -12,   1,   3,  -5,  -4 },
     {    0,   0,   4,  24,  -5,  -4,  -4,   0,  -3,   0,  -1,   0,  -4,  -5,  -8,   9,  -4,  -8,   0,   2,   0,   0,   1,  -1,  -4,   0, -15,   5, -12,  -2,  -1,   0,  -1,   1,   0,   0,  -9,  -4,   3, -32,  -7,   4,   4,   1,   1,   1,   0,   0, -13, -20, -24, -15,  -1,   7,   6,   0,  24,  -7,  10, -31,   5,   3,   0,  -1, -31,  -6, -42,  -8,  18,   4,   5,   0, -12,   6,  18,   6,   9,   0,   0,  -1,  -6,  35, -44, -14,  -2, -23, -27,  38,   3,  -4,  16,  19,   3,  -6,   7,   4 },
     {    0,  -4,  -4,  11,  -5,  -3,  -1,  -1,  -3,   1,   0,   0,   0,  -5,  26, -27,   2,  -4,  -1,   0,   0,  -1,  -1,  -1,  -1,  -6, -42,   7,  10,   3,   3,   3,   0,  -1,   1,   1,   0,  26,  52,   2, -21,   2,  -7,   0,  -3,   0,  -1,  -1, -15, -42,  21, -32,  -2,   6,  -1,   2,   1, -23, -16,  -1,   8,   4,   2,   2,  50, -23,   6,  -3,   4,   0,  -1,   0, -16,  -4, -10,  -7,  12,  -1,   3,  -1, -11,  -8,  -8,  16,  -3,   3,  12,  10,   1,   5,   3,  -1,  -2,   3,   0,   0 },
     {    3,   2,  18,  42, -31,   6, -11,  -2,  -4,   1,  -1,  -1,  -6,  -8, -38, -26, -25,  12,   4,   4,   2,   4,   1,   1,  -5,   2, -17, -76,  -1,  10,   6,   4,   2,   2,   2,   1,   3,   5,   2, -33,   9,   5,   2,  -3,   1,   0,   0,  -1,   3,  12,  17, -15,   7,   2,   1,  -1,   1,   0,  13,  -1,   5,  -2,   0,   1,   4,   8,  11,  -4,   4,  -3,  -1,   0,   6,   6,   4,   4,   8,  -3,  -1,   0,   1,   7,  17,  -2,  -2,  10,  -5, -19,  -3,   4, -18,   0,   0,  -1,  -6,   5 },
     {    4,   6,  12,  43,  11,  -7,  -2,  -6,  -1,  -7,   1,  -3,  -3,   0,  -9, -23,  22, -13,   5,  -1,   3,  -3,   2,  -1,   1,  11,  12, -21,  31, -22,  -6,   3,   4,  -5,  -1,   0,   2,  -3,  41,  28,  26, -13,  -8,  -4,  -3,  -5,   0,   0,   5,  -4, -24,  33,  11,  -4,  -2,   4,  -5,  26,  16,   8,  -2,  -8,  -1,   2, -20, -18,   2,   4, -13,  -3,   7,   4, -14,  -4,  12,  -8, -11,   4,   3,  -4, -11,  -7, -19,   9,  -1, -11,  25,  19,   2,   9,  15, -21,   1,   2,  -7,  -7 },
     {    5,   1,   5,  19,  15,   2,  -6,   1,  -1,   2,  -2,  -1,   6,   4,  -1,  -9,  31,  -5,   7,  -2,   2,  -3,   1,  -3,   6,  -6,  -6,   4,  13,  -3,   0,   4,   1,   0,   0,   1,   2,  16,   0,   1,   1,   0,  -2,   1,   0,   1,   0,   0,   0, -23,  24,  -3,  -6,  -2,   1,   0,  34, -13, -25, -18,  15,  -2,   2,   1, -22,  34, -11,  47,   8,  -7,  -4,  -4,  51,  28,   8,  16, -31, -10,  -7,   0,  23, -18,  -1, -10,   0,  -7,  15,  -9,  -4,   0,  -6, -10,   0,  -1,  -4,   2 },
     {   -3,  -7,   0,  -9, -63,  28,   6,   6,  -1,   2,   1,   2,  -3,  -6,   4, -38, -64,  19,  15,   4,   3,   2,   2,   1,   3,   0,   4,  28, -26,   5,  -1,  -1,   0,   1,  -2,   0,   1,   2,   4,  27,   1,  -5,  -5,  -3,   1,  -1,  -1,  -1,   1,  -5, -10,  19,   0,  -3,  -5,  -3,   4,   6,  -3,  -1,   8,  -3,  -2,  -2,  -7,   2, -10,  12,   1,   0,  -1,   2,  20,   6,  10,  -6,  -5,   2,  -2,   1,  10,  -7, -13,   4,   4, -15,  21,   9,   0,   0,  10, -12,   0,  -4,   0,  -6 },
     {   -1,   3,  -7,  -4,   6, -11, -10,   0,   3,  -3,   3,   1,  -1,   1,   2,  -6, -11,   3, -11,  -2,  -1,   0,   2,   0,   1,   8, -14,  21, -17,  10,  -2,  -5,   0,  -1,  -3,  -3,  -4,   6,  16, -12,  14,   1,   6,   3,   2,  -3,  -3,  -1,   5, -22, -33, -27,   6,  -1,   9,   6,  31,  53,  -8,  25,  17,  -4,  -2,  -7, -16,  29,  27,  14,  -3, -13,  -5,  -6, -48,  16,   1,  10,  -7,   2,  -1,  -2,   8,  13,  17,   6,   9,   8,   0, -18,   8, -11, -10,  -7,  -5,  -1,  -1,   0 },
     {   -4,  -3,  -8, -39,  -4, -13,  12,   1,   0,  -3,   1,  -2,   3,   0,  16,  52, -19,  -1, -11,  -3, -11,   1,  -2,   3,  -4,   1, -25, -20,  -8,   4,   4,   6,   3,   3,   3,   2,   0,   7,  21, -52,  22,  -5,   1,  -5,   2,  -6,   1,   0,   3,  -1,  19,  13,  21,  -6, -10,   9,   2,  -2,  18,   4,   1,  -5,  -1,   2,  -8,  -4,   0,  18,  -1, -10,  -1,   4,   6,  -9,  18, -10, -18,   8,   2,   3,  -1, -14, -10,  10,  -7, -13,  24,  10,  -7,  15,  11, -30,  -8,   9,  -9,  -7 },
     {    1,   2,   1,  12,  20, -22,   3,  -2,   0,   0,  -2,   1,  -4,  -7,  -7, -44,   4, -17,  12,   2,   3,  -2,   2,   2,   2,   1,   7,  39, -12,   5,  -6,  -4,  -3,   1,   0,   3,  -7, -11, -17, -55, -21,   9,   2,   4,   0,   5,   3,   0,   5,   9,  -5, -12, -17,   7,  -5,   6,  -9,  11,  22, -25,  -4,  14,   2,   3,  -9, -18,  15,  -8,  -9,  12,  -4,   0,  12, -12,  12, -21,   5,  12,   1,   1,  10, -16,   0,   9,   9,   2,  34,  -6, -12,   7, -13, -22,  -2,   5,  -8,   9 },
     {   -1,  -1,  -3,   0,  24,   7, -11,  -5,  -1,  -6,  -4,  -1,  -2,  -1,   3,  -1,  -4,  12,   1,  -8,   0,   2,  -1,   0,  -5,  -9,  -6, -23, -32,  -6,   3,  25,   8,  12,  12,   1,  -1,  -1,  15,   6, -18,  -1, -20, -21, -16,  -4,   4,   4,  -2, -19,  -4,  10, -22,  27, -10,  13,   4,  13, -23,  20, -18,   6,  10,   5, -33,   3,   3,  -3,  -1,  10,  -1,  10,   7, -27,  -8, -37,  23,  -4,   0,  14,  32, -14,   9,  -5, -20, -22, -15,  -9, -13,  -3,   5,   7,   1,   2,  12,  -5 },
     {    8,   8,   2,  10,  68,   4,   5,   4,   2,  -3,  -2,   2,   4,  -4,   5,  -1, -35,  32,  30,   5,   6,   1,  -2,  -9,   4,   4,  -1,   0, -15,   9, -13,  10,   7,   5,   1,  -2,   6,   0,  11,  -4,  25,  18,  -9,  -6,   1,   0,   1,   0,   4,   4,   3,  25,   9, -10,  14,   3,  -5, -13,  -1,  -7,  12,   6,  -3,   6,  27, -13,  -9, -10,   8,   0,  -3,  -1,  -1,  12, -13,  14,  -3,  -2,  -2,  -6,  32,  25,  -2,  -6,  25,  -4,   8,   3,  10, -12,   2,  -1, -10,  -4, -15,  -4 },
     {    0,  -3,  -5,   0,  -6, -16,   8,   8,  -9,   4,   6,   1,   2,   0,   5,   1,  -1,   7,  -3,   1, -23,   2,  -7,   5,   2,  -2, -13,   5,  20,   9,   2, -25,  -6, -11,  -8,   1,   1,   4,  21,  -7,  10, -12,  18,   6,   6,  -1,  -9,  -5,   5, -17,  -7,  17,   3, -10,  10,  11,   1,  14,  -2,  28,  -6,   0,  -5, -14, -18, -35, -14, -27,   1,  18,   2,  -2,  46,  21, -16,  12,  19,  33,  -9,  -6,  16,  -6,  -7,  -1,  -2,   4, -16, -19, -11, -16, -24,  10,  -3,  14,   9,  22 },
   },
   {
     {   68, -41,   3,  -4,   1,  -1,   0,   0,   0,   0,   0,   0, -72,  40,  -1,   3,   0,   1,   0,   0,   0,   0,   0,   0,  43, -19,  -5,   0,  -1,   0,  -1,   0,   0,   0,   0,   0, -25,   4,   7,  -1,   1,   0,   1,   0,   0,   0,   0,   0,  12,   4,  -6,   1,  -1,   0,   0,   0,  -6,  -5,   4,   0,   0,   0,   0,   0,   3,   4,  -2,  -1,   0,   0,   0,   0,  -2,  -3,   0,   1,   0,   0,   0,   0,   1,   2,   0,   0,  -1,  -1,   0,   0,   1,   1,   0,   0,  -1,  -1,   0,   0 },
     {   86, -22,  -9,  -3,  -3,  -1,  -1,  -1,  -1,   0,   0,   0,  13, -32,  11,  -2,   2,  -1,   1,   0,   1,   0,   0,   0, -54,  39,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,  33, -20,  -4,   1,  -1,   1,   0,   0,   0,   0,   0,   0, -26,   6,   9,  -2,   2,   0,   1,   0,   8,   4,  -7,   1,  -1,   0,   0,   0,  -8,  -3,   4,   0,   0,   0,   0,   0,   2,   4,  -2,  -1,   0,   0,   0,   0,  -4,  -1,   0,   1,   1,   1,   0,   0,  -2,   0,   0,   0,   1,   1,   0,   0 },
     {   34, -23,   7,  -3,   1,  -1,   0,   0,   0,   0,   0,   0,  63,   2, -20,   0,  -4,   0,  -2,   0,  -1,   0,  -1,   0, -21, -41,  24,  -1,   5,   0,   2,   0,   1,   0,   1,   0, -36,  54, -11,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,  26, -32,  -2,   5,  -1,   1,   0,   1, -24,  16,  10,  -5,   2,  -1,   1,  -1,   8,  -2, -11,   3,  -1,   1,   0,   0,  -9,   1,   7,   0,   0,   0,   0,   0,   2,   2,  -4,  -1,  -5,   0,   2,   1,   1,   1,  -1,  -1,  -3,   0,   1,   0 },
     {   25,  17, -20,   2,  -4,   1,  -2,   0,  -1,   0,   0,   0,  49, -25,  11,  -4,   0,  -2,   0,  -1,   0,   0,   0,   0,  34, -18, -16,   1,  -1,   1,  -1,   1,   0,   0,   0,   0, -49, -19,  32,  -2,   5,  -1,   2,   0,   1,   0,   1,   0,  -7,  50, -22,  -3,  -2,  -1,   0,  -1,  12, -40,   4,   8,  -1,   2,  -1,   1, -17,  22,  10,  -9,   2,  -2,   1,  -1,   4,  -6, -12,   5,   0,   1,   0,   0,  -6,   3,   8,   0,   1,   0,  -5,   0,  -3,   1,   3,   1,   1,   0,  -2,  -1 },
     {   -7,  12, -11,   4,  -2,   1,   0,   1,   0,   0,   0,   0, -38, -20,  25,  -3,   4,   0,   2,   0,   1,   0,   1,   0, -50,  24,   0,   2,   1,   1,   0,   0,   0,   0,   1,   0,  -8,  44,   1,  -5,  -1,  -2,   0,  -1,   0,  -1,   0,  -1,  47,  -5, -39,   7,  -5,   2,  -2,   1,   1, -43,  30,   4,   1,   1,   0,   1,  -3,  32,  -5, -11,   2,  -2,   1,  -1,   9, -17,  -8,   9,  -1,   2,   0,   1,   0,   5,   8,  -5,   3,  -4,  -6,   2,   1,   1,   4,  -1,   3,  -2,  -2,   0 },
     {   29,  70, -34,   0,  -8,  -1,  -3,   0,  -2,   0,  -1,   0, -18, -51,  16,   4,   4,   2,   2,   1,   1,   0,   1,   1,  41,  11,   5, -10,  -1,  -3,   0,  -1,   0,  -1,   0,   0, -14,   1, -22,   8,  -2,   2,   0,   1,   0,   1,   0,   0,  15, -35,  23,   1,   0,   1,   0,   0, -26,  24,   0,  -7,   3,  -2,   1,  -1,   9,  -4,  -8,   6,  -2,   1,   0,   0,  -5,  -2,   6,  -1,   0,   0,   0,   0,   1,   1,  -1,   0,  -2,   0,   0,   0,   1,   0,   0,   1,  -1,   1,   0,  -1 },
     {   -3,   8,   9,  -5,   0,  -1,   0,   0,   0,   0,   0,   0, -26, -22,  -4,  10,   1,   3,   1,   2,   0,   1,   1,   0, -24,  -2,  28,  -9,   4,  -3,   1,  -1,   1,  -1,   0,  -1, -42,  25,  -9,   6,   2,   1,   0,   0,   0,   0,   0,   0,  13,  38,   8, -10,  -2,  -2,  -1,  -2,  21,  -4, -50,  11,  -2,   3,  -1,   1,   8, -38,  41,   7,  -3,   2,  -1,   1,  -5,  34, -13, -20,   4,  -4,   2,  -1,   6, -19,  -9,  18,   1,   6,  11,  -8,   2,  -4,  -6,   3,   1,   3,   3,  -1 },
     {   11,   5,   4,  -6,   0,  -2,  -1,  -1,   0,   0,   0,   0,  22,  21, -20,   4,  -4,   1,  -2,   0,  -1,   0,  -1,   0,  27,  16,  -5,  -6,  -2,  -2,   0,  -1,   0,  -1,   0,   0,  37,  -3, -26,   6,  -3,   1,  -2,   1,  -1,   0,   0,   0,  23, -36,  -7,  -1,  -1,   0,   0,   1, -16, -48,   8,  17,   2,   4,   0,   2, -14,   9,  53, -16,   0,  -3,   0,  -2, -12,  30, -34, -14,   6,  -3,   2,  -1,  -1, -19,   1,  19,  -4,   7,   7,  -9,  -2,  -4,  -3,   5,  -3,   3,   3,  -2 },
     {  -26, -63,  22,  -1,   6,   2,   3,   1,   3,   0,   1,   1,  -3, -47,  18,   8,   3,   2,   1,   1,   0,   0,   1,   0,   9,  23,   0,  -8,   0,  -2,   0,  -1,   0,   0,  -1,  -1, -32, -22,   0,  14,   0,   3,  -1,   2,   0,   1,   1,   1,   6,  -8,  29,  -9,   2,  -2,   0,  -1, -31,  21,  11,   0,   2,   0,   1,  -1,   7,  38,   4, -16,  -1,  -4,   0,  -2,  -2,   7, -33,   5,   2,   0,   1,   0,   9,  -6,  15,   7,  -2,   9,  -6, -10,   3,  -2,  -2,   6,   0,   4,  -1,  -3 },
     {   -8, -37,   2,   5,   2,   1,   1,   2,   1,   1,   0,   0,  23,  -7,  14,  -6,   0,  -2,   0,  -1,   0,   0,   0,  -1,  30,  58, -29,  -2,  -6,  -1,  -3,  -1,  -1,  -1,  -1,  -1, -10,   3, -12,   3,  -1,   1,   0,   1,   0,   0,   0,   1,  39,   1, -16,   0,  -2,  -1,  -1,  -1,  -6, -17, -24,   3,   0,   2,   0,   0,   5, -41,   1,  26,  -1,   4,  -1,   2,  -5,  -9,  47,  -9,  -4,   0,  -1,   0,  -7,  20, -12, -15,  -4,  -8,  -2,  10,  -1,   4,   4,  -4,  -2,  -2,  -1,   1 },
     {  -11, -23,  -8,  13,  -2,   5,   0,   2,   0,   0,  -1,   0, -14, -39,  18,  -6,   4,   0,   2,   1,   1,   0,   1,   0,   7, -30,  10,  13,   1,   2,  -1,   0,   0,   1,   0,   1,   6,   0,   8, -12,   3,  -2,   1,  -2,   1,   0,   0,   0, -24, -20,  -4,  21,  -1,   4,  -1,   2,   7, -12,  33,  -7,   2,   0,  -1,   0, -15,   7,  24,   7,  -1,   0,  -1,  -1,   4,  39,  25, -37,   0,  -6,  -1,  -2, -10,  20, -53,   3,  10, -13,  19,  15,  -2,  11,  -6, -16,   2,  -2,  -2,   6 },
     {   -1,  13, -31,  12,  -2,   3,  -1,   0,   0,   1,   0,   0,   3,  36,  27, -17,  -2,  -3,   0,  -2,  -1,  -1,  -1,  -1, -40, -12, -40,  16,   3,   4,   1,   2,   0,   2,   0,   1, -13, -33,  41,   9,  -2,   3,  -1,   0,   0,   0,   0,   0,  30,  -5,  11, -17,   5,  -4,   1,  -1, -41,  16,  -4,  11,   1,   0,   0,   0,  30,  -3,  19,  -5,  -7,   0,  -2,   0, -13,  16,  -2, -16,   5,  -3,   1,  -1,   7,   4, -18,   8,  -1,  -7,   8,   2,   4,   3,  -2,  -5,  -1,  -3,  -1,   1 },
     {   -3, -13, -26,  19,  -3,   2,   0,   1,  -1,   0,   0,   1,  14,  16,  43, -28,   1,  -6,   0,  -3,  -1,  -1,  -1,  -1,  24,  22,  -9,  -1,  -1,  -2,   0,  -2,   0,   0,   0,   0,   7,  44,  -6, -15,  -3,  -3,   0,  -2,   0,  -1,   0,  -1, -15,   6, -12,   0,   2,   0,   0,   1,  16,  25, -13, -10,  -1,  -2,   0,  -1,  -1,   3, -26,  -5,   1,  -1,   1,   0,  -4,  -1, -41,   2,   8,   1,   2,   1,  -1, -29, -17,  41,   7, -13,  35,   0,  -4,   1,  -8,  -6,   1,  -5,   6,   9 },
     {    9,  27,  46, -22,  -1,  -5,  -1,  -3,  -1,  -2,  -1,  -1,   3,  14, -39,   0,   4,   0,   2,   0,   0,   0,   0,   0, -10,  37,   0,   8, -11,   2,  -4,  -1,  -2,   1,  -1,   0, -30, -26,  -2, -10,   8,   1,   3,   0,   2,  -1,   1,  -1,  18,  13,  10,  12,  -9,   1,  -3,   1,  -5,   1,  12, -30,   8,  -4,   2,  -1,  -6,  25, -23,  13,  -2,   0,   0,   0,  12,  -4,  -8, -21,   3,  -1,   0,   0, -11,  -2, -37,  27,  13, -19,  26,  10,  -5,   3,  -5,  -9,   4,  -7,   3,   2 },
     {    4,  11,  20, -16,   2,  -5,   0,  -2,   1,  -1,   0,   0,   4,   5,   9,   5,  -7,   1,  -1,   0,   0,   0,   0,   0,  11,  13,   9, -16,   4,  -4,   1,  -2,   1,  -1,   0,  -1,  12,  24,  11,  -1,  -7,  -1,  -3,  -1,   0,   0,  -1,   0, -10,  15,  -9, -23,   7,  -3,   3,  -1, -10,  15,  -4,  10,  -5,   2,  -1,   1,  11,  11,   0, -27,   2,  -3,   2,  -1,  -6,  15, -19,  -7,   1,  -1,  -1,  -1,  -6,   5, -44, -26, -12, -14, -41,  59,  15, -14,  40,   6,  -5,   0, -10, -14 },
     {    5,  -1,  31, -11,   0,  -3,  -1,   0,   0,   1,   0,   0,  -5, -46, -12,   6,   5,   1,   2,   2,   1,   1,   1,   1,  16, -27,   0,  18,  -3,   1,  -1,   1,  -2,   0,   1,   1,  58,   9,  37, -20,  -3,  -5,   0,  -2,   0,  -2,   0,  -2,  29,  25, -26, -18,   3,  -6,   0,  -3, -35,   5, -18,   7,   2,  -2,   2,   1,  14,  -3, -13,   8,  -5,   3,   0,   1, -16, -14,  -4,   1,   7,   2,   2,   0,  -2,   2,   2,  21,   0,  -2,  18, -10,  -5,   6, -10,  -1,  -1,  -2,   4,   5 },
     {    0,   9,  10,  -7,  -2,  -2,   0,  -1,   0,   0,   1,   0,  -5,  -5, -22,   2,   5,   1,   1,   1,   1,   0,   1,   0,   0,  28,  -8,   6,  -7,   0,  -2,  -1,   0,   0,  -1,  -1, -25,   5,   1, -17,   4,  -3,   2,   0,   1,   0,   0,   0, -58, -10, -45,  25,   7,   2,   3,   2, -17,   3,  23,  29, -10,   8,  -2,   3,  59, -17,  14, -16,  -3,   0,  -4,   0, -26,  -8,   4,   1,   6,  -1,   3,  -1,  15,   4,   8,   6,  -1,  -3,   8, -10,   2,   5,  -4,  -2,  -1,  -2,   2,   3 },
     {    8,  18,  61, -22,   1,  -6,   1,  -5,  -1,  -2,   0,  -2,  -4, -14,  18, -28,   2,  -1,   3,  -1,   1,   0,   1,   0,   1,  -9, -60,  36,   7,   3,   1,   1,   0,   2,   1,   0, -15,  19,  14,   7, -14,   4,  -3,  -1,  -1,   1,  -1,   0,  -5, -35,   9,  -6,  10,   1,   2,   1,  32,   5,  -1,   5,  -7,  -1,  -3,  -1, -21, -16,   2,  -9,   9,   2,   2,   0,   1,  18,  -1,  14,  -1,   0,  -2,   1,  -2, -10,  10, -10,  -2,   7,  -1,  -2,   0,   0,  -3,  -1,  -2,   3,   1,   1 },
     {   -6, -13, -24,  -1,  11,   2,   3,   0,   2,   0,   1,   0,  -8, -19, -31,  23,  -2,   7,  -4,   3,   0,   1,   1,   1,  -9, -16, -22,   2,  10,   0,   2,   1,   1,  -1,   0,  -1,  -5, -19, -12,  19,  -2,   0,  -2,   2,   0,   0,   0,   2,   7, -11,  -8,   2,   7,  -1,   3,   0,   9,  -7,  -3,  10,  -7,   4,   0,   1,  -6, -16, -13,  -2,  13,  -1,   4,  -1, -10,  -6,  -9,  34,  -4,   3,   3,  -1,   0, -17,   1,  22,   5,  -8,  40,  44,  12,   2,  46, -41,  -7,   8, -14, -13 },
     {    1,   8,  21, -57,  16,   3,   4,  -3,   2,  -2,  -1,  -1,   7,  14,  61,  40, -24,  -8,  -7,  -3,  -4,   0,  -1,  -1,  -5, -13,  -2, -41,  18,   8,   1,  -1,   0,   2,   1,  -1,   3,  -1,  -6,  25,   6,  -6,   2,  -1,   1,  -1,   0,   0, -16,  -2,  -8,   1, -14,  10,   0,   5,  -5, -11,   0,   3,  17,  -5,  -2,  -1,   3,   1,  -4,  10, -13,   1,   0,   0,  -8, -10,  10,  -6,   6,   3,  -1,   1,   2,  12,   4,  13,   9,  -4,  20,  -9,   1,  10,  -2, -17,  -1,  -1,  -4,  -1 },
     {   -5,   0,  -1,   1,   1,  -5,   4,   1,   0,   0,   0,  -1,  -9,   1,  -9, -18,   9,  -1,   1,   1,   2,   0,   0,   0,  -8,  -7, -19,   3,  -1,  -2,   4,   3,   1,  -1,   2,   3, -10,  19, -15,   7,   1,   1,   3,  -1,   2,  -1,  -1,  -1, -42, -10, -11, -15,   8,  -1,   0,   1, -54,  -9, -35,  27,  14,  -2,   6,   1, -24,  23,  12,  41, -10,   5,  -4,   2,  47, -21,   5,  -7,  -9,   3,   0,  -1, -15,  -7,  -1,  11,  13,   8,  -1,  13,   1,  -6,  10,   9,  11,   1,   7,  -7 },
     {   -3,  -9, -38, -24,  19,  -7,   6,   2,   0,   2,   1,   0,   2,  -7, -38,  43, -10,  -4,  -4,   1,   1,   0,  -1,   1,   9,  14, -33,  14,   4,   2,   0,   3,   0,   0,  -1,   0,   6,  34,  26,  22, -10,  -9,  -2,  -2,  -3,   1,  -2,  -1,   4,  -3,  16, -10,  -7,   5,  -5,   0,  34,  22,   9,  12,  -2,  -4,  -2,  -2,  17,  17,  14,  -5, -22,   5,  -4,   2,  29, -13,  -1, -25,   1,   2,   1,  -1, -15,   0,  -8,  -3,   5,  -6,  -4, -14, -12,  -4, -11,   0,   3,  -3,   1,   8 },
     {   -2,   5, -13,  -9,  18,  -5,   3,  -2,   0,  -1,   0,   0,  -2,   8,  -6,  22, -27,   7,  -6,  -1,   0,   1,  -1,   1,   2,  32,  19, -11,  -1,  -8,  -1,  -4,   0,   0,   1,  -2, -10,   7,  51, -27, -10,  -3,  -1,   1,  -2,  -1,   0,   0,  -7,  -9, -13,  -7,  10,  -1,   1,   1, -33,   7,   3,  -5,  -9,   1,   1,   2, -42, -14,  -6,   7,  17,   3,   0,   2,  16,  38,   6,  23, -16,  -5,  -5,   0,  11, -19,   6, -18,   5,  12,  10,  -6,  -1,  -2, -19, -22,  -2,   3, -12,  17 },
     {   10,  11,  33,  43, -19,   2,  -5,  -1,   0,  -1,  -1,  -2,   6,   8,  10, -10, -15,   3,  -2,  -1,  -3,   0,  -1,   0,  10,   6,  19,   0,   5, -11,  -2,  -1,   1,  -1,  -2,  -1,   9,   8,  -1,  17, -14,   3,  -1,  -2,  -1,   0,  -4,   0,   5,  15,  10,  -2,   1,  -3,  -2,   0,   2,  12,   9,  18, -14,   2,  -1,   0,  23,  17,  23,  -2, -13,  -1,  -3,  -2,  27, -13,   9, -11, -17,   1,  -4,   0, -12,  -8,  17, -18,  11,   4,  23,  -7,  -4,   2,   4, -62, -14,   4, -35,  -6 },
     {   -3,  -7, -17, -56,  29, -19,  12,  -1,   1,   1,   2,  -1,   2,  -1,  12, -55,  -4,   8,  -6,   0,  -4,  -2,   0,   0,   5,  -3,  45,  14, -14,   0,  -1,   2,  -2,  -2,  -1,   1,  -2, -30,   3,  -1, -13,  -1,   1,  -1,   2,   1,  -1,   1,  20, -16, -23,  18,   5,  -2,   3,  -2,  16,  17,  -6,  14, -12,  -1,  -2,   1,  11,  -4,  13,  -4,  -4,   3,   0,   0,  20, -17,  -4,  -3,  -2,  -2,   5,  -4, -23,  -8,   3,   2,   7,  14,  -1,   8,  -6,  -8,   3,  -4,   2,   5,  -3,  -1 },
     {   -1,  -1, -12, -31,  12,  -8,   7,   4,  -3,   1,   0,   0,   3,   0,   7,  -4,  -2,   7,   2,   0,   1,  -2,  -1,   3,   3,  -8, -23,  32,  15,  -6,  -3,   0,   1,   2,   5,   2,  14,   9, -45,  -2,  17,   4,   4,  -2,   0,   3,  -1,  -2,   6,  51,   7,  30,  -8, -10,  -6,  -3, -23,   5,  20,  -1, -10,  -3,   2,  -2,  15,  12,   0,  14,  10,   0,  -2,  -4,   8,  28,   9,  13, -21,  -6,  -6,   1,  22, -37,  -6, -26, -14,   0,   1,  -3,   0,   2,  -7,  -2,  -3, -11,  -1,  -5 },
     {    5,   2,  15,  30,  -3,  -3,  -5,   1,  -2,   2,  -1,   2,   4,  -2,  16,  43,  -4, -15,   3,  -2,   0,  -3,  -1,  -2,  -1, -10, -13,  -5,  22,  11,  -3,  -1,  -2,   0,   0,  -1,  -1, -25,  -2, -28,  30,  -6,   5,  -1,   3,   2,   1,  -1,  19, -18, -40,  40,  -2,   2,  -1,  -2,   6,  32, -14,  -4,  -3,  -7,  -2,  -3,   4, -12,  10,   0,  -8,  -2,   4,  -2,  38,   4, -21,  -7,  -4,  -9,  -2,   1, -25, -19,   7,  -9,  -3,   6, -10,   8,  -3,  -1,   9,  16,   2,  -4,  12,   0 },
     {    0,   0,   2,  13,  14,  -4,   7,   2,  -5,  -3,   1,   1, -10, -16, -28, -44, -49,  13,   3,   7,   7,   5,   1,   4,  -7, -11, -30, -62,  31,  -9,   4,  -2,   6,   4,   3,   0,  -1, -12, -20,  -8,   1,   0,   9,   1,   4,   2,   1,  -3,   9,   9, -15,  -7,   5,   6,  -3,  -1,   5,  15,   9,  16,   5,  -5,   4,   1, -10,   4,  -6, -10, -10,  11,   5,   3, -11,   0,   3,  -3,  13,   3,   1,   4,  -4,  -2, -12, -11,   6,  -7,  -3, -10, -11,   3, -17,   3,  -1,  -6,  -4,  20 },
     {   -2,  -6,  -6, -30, -48,  42,   5,  -2,   6,  -7,  -2,   2,  -1,  -5, -15, -20,  31, -24,   8,  -6,   0,   5,   2,  -3,   0,  -3,  -9, -48, -27,  13,  13,   3,   0,   4,   3,  -4,   5,  -1,  20,  10,   7,  -8,  -6,  -6,  -5,   2,  -1,   1,   6,  -7, -13,   3,  -6, -11,   3,   0,  12,  12,   2,   4,  18,   9,  -6,   0,  11,   9,   0,  18,  14, -11,   5,  -5,  12,  16,   3,  -1,  -5,   5, -15,   4,   0, -30,   1, -21, -17,  -3,  11,  -6,   4,   4,   1,  -4,  -7,   0,  -2,  -6 },
     {   -3,   2,   1, -10, -31,  -3, -14,   5,  -4,   1,   4,   1,  -2,   7,   1,  10,  11, -18,  -1,   0,  -6,   1,  -5,   1,  -2,   4,   4,  11, -18,   7,  11,   1,   3,  -4,   1,   3,  -3,   0, -20,  27,  17,  -9,  -5,  -2,  -4,   0,  -2,   5,  11,  20, -20,  -1,  10,  -8,   3,  -3,  -8,  48,  22,  34, -20,  -2,  -3,  -3, -60,  -6,  17, -15,  -6,  -8,  -6,   9, -16,  -8,  10,  13,   1,  -8,  13,  -3,  -1,  19, -10,  14,   5,  -9,  -6,  -9,  -3,  -1,  -2,  -3,   6,  -2,  -2,  21 },
     {    6,   7,   9,  27,  51,   2, -16, -13,   0,  -2, -10,  -8,   1,  -2,  -9, -20,  16,   2,   8, -23,  -7,  -1,   8,  -7,   7,  10,  17,   5,  16,  -5,   6,  14,   0,  -1,  -8,  -7,   5,  -1,  18,  54,  23, -38,  -1,   0,   2, -12,  -3,  -3,   1,  -3, -16,   8, -13, -11,   6,   6,   3,   3,   4,  11,  19, -16,  -4,  -6,  -1,  -3, -12,  11,  -9,  -3,   1,  -3,   9,  20,  -1,  12,   7,  -2,  -8,  -3,  25,   0,  -7,   4,  -6, -15, -10, -13,   2,   9,   5,   1,   5,   8,   7, -13 },
     {    3,   9,   6,  16,  47,  10,  23,  -5,  11, -16,   3,  -4,  -3,   0,   0,  -4,  -3, -24,  -3,  -7,   6,   2,   9,  -1,  -4,  -3, -15, -14, -26,   0,   1,   6,   5,  -4,   3,  12,   3,  13,   3, -16,  -4, -13, -19,  16,   0,   8,  -5,   2,   2,  13,  22,  28, -38,  -7,   0,   6,  -9,  14,   5,   9,  -8,  30, -23,  10, -11,   0,  16,  10,  26, -16,   8,   7, -19, -10,  -7,  -2,  -3,   8,  10, -29, -23,  15,   5,   1,   5,  15,   3,   6,   6,   3,   0,   4,  -4,  -1,  -1, -11 },
   },
   {
     {   99, -58,   0,  -3,   0,  -1,   0,   0,   0,   0,   0,   0, -51,  20,   9,  -1,   2,   0,   1,   0,   0,   0,   0,   0,  -1,  11,  -7,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,  -4,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   51,   9, -24,  -1,  -5,   0,  -2,   0,  -1,   0,  -1,   0,  66, -65,   8,   2,   1,   0,   0,   0,   0,   0,   0,   0, -56,  19,  20,  -6,   3,  -1,   1,  -1,   1,   0,   0,   0,  -1,  15, -11,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -7,  -2,   1,   3,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0 },
     {  -21,  30, -17,   4,  -1,   1,  -1,   1,   0,   0,   0,   0, -60, -20,  39,   0,   5,   0,   2,   0,   1,   0,   1,   0, -30,  67, -12, -11,   0,  -3,   0,  -2,   0,  -1,   0,  -1,  42, -18, -28,  11,  -3,   2,  -1,   1,   0,   0,   0,   0,   2, -13,  12,   1,  -2,   1,   0,   0,   9,  -1,  -3,  -2,   0,   0,   0,   0,   2,  -4,   3,   1,   0,   0,   0,   0,   5,  -1,  -1,  -1,   0,   0,   0,   0,   1,  -2,   1,   0,   3,  -1,   0,   0,   1,  -1,   1,   0,   2,   0,  -1,   0 },
     {   38,  81, -62,  -2,  -9,  -1,  -4,  -1,  -2,   0,  -1,   0, -32,  -1,  -9,  15,  -2,   4,  -1,   2,   0,   1,   0,   1,  25, -42,  18,   2,  -1,   1,   0,   1,   0,   0,   0,   0,  -4,  -7,  12,  -5,   2,  -1,   1,   0,   0,   0,   0,   0,  -3,   2,   3,  -2,   2,  -1,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   2,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
     {   13,   1, -17,   5,  -3,   1,  -1,   1,   0,   1,   0,   0,  43,  -1,   8, -11,  -2,  -3,  -1,  -1,   0,  -1,   0,  -1,  36,  17, -57,   6,  -1,   2,  -1,   1,   0,   0,   0,   0,  10, -72,  23,  25,  -4,   5,  -1,   2,  -1,   1,   0,   1, -23,  15,  33, -19,  -1,  -1,   0,  -1,  -5,   4, -12,  -3,   6,  -1,   1,   0,  -8,   2,   5,   3,   0,  -1,   0,   0,  -2,   1,  -1,  -1,   0,   0,   0,   0,  -4,   2,   2,   0,  -1,   1,  -1,   0,  -2,   1,   1,   0,  -1,   0,   0,   0 },
     {   13,  21,   4, -10,  -1,  -3,   0,  -2,   0,  -1,   0,  -1,  19,  68, -60,   0,  -4,  -1,  -1,  -1,  -1,   0,   0,   0, -57,   4,  -7,  18,  -1,   4,   0,   1,   0,   1,   0,   1,  39, -16,  -6,   5,  -6,   2,  -2,   1,  -1,   0,  -1,   0,   6, -33,  11,   5,   0,   2,   0,   1,  -1,   0,  12,  -4,   0,   0,  -1,   0,   1,  -6,   1,  -2,   1,   0,   0,   0,  -1,  -1,   4,   0,   1,   0,   0,   0,   0,  -3,   1,  -1,   0,  -1,   2,   0,   0,  -2,   1,   0,   0,  -1,   1,   0 },
     {  -13,  -4, -11,  17,  -3,   4,  -1,   2,  -1,   1,   0,   1, -15,   8,  18, -12,   2,  -2,   1,  -1,   1,   0,   0,   0, -61, -14,  -4,  23,   2,   5,   1,   2,   0,   1,   0,   1,   5,  -5,  66, -15,  -7,  -1,  -2,  -1,  -1,  -1,  -1,   0,   1,  48, -25, -37,  11,  -5,   3,  -2,   7,  -9, -25,  20,   7,  -1,   1,   0,   4,   3,   5,   4,  -6,  -2,   0,   0,   5,  -3,  -3,  -1,  -1,   2,   0,   0,   2,   1,   0,  -1,   3,  -2,  -2,   1,   1,   0,   0,  -1,   2,  -1,  -1,   0 },
     {    1,   9, -43,  31,  -8,   6,  -3,   2,  -1,   1,  -1,   1,  22,  48,  29, -35,  -3,  -7,  -1,  -4,  -1,  -2,   0,  -1,  -1,  22, -31, -14,  12,  -4,   4,  -2,   1,   0,   1,   0, -46,  22, -16,  12,   3,  -1,   2,  -1,   1,   0,   1,   0,  20,  -3, -30,  18,  -6,   2,  -1,   1,   6, -31,  18,  13,  -8,   4,  -2,   2,   3,   0,   7,  -9,  -1,   1,  -1,   0,  -1,  -9,   1,   2,   1,   1,   0,   1,   1,  -1,   3,  -2,   0,  -4,   1,   1,   0,  -1,   2,  -1,   0,  -3,   0,   1 },
     {  -24, -43, -51,  48,  -1,  11,   0,   4,   0,   3,   0,   2,  -4,  13,   1,  21, -15,   4,  -5,   2,  -2,   1,  -1,   0, -20, -21,  34, -24,   8,  -1,   3,   0,   1,   0,   0,   0, -12, -10,  -5,  23,   0,   2,   0,   0,   0,   0,   0,   0,  12,  -8,  49, -12, -12,   2,  -3,   0,  -3,  25,  -5, -28,   9,  -2,   1,  -1,   4,   0, -10,   5,   6,  -3,   0,   0,   0,   5,   1,  -1,   0,  -2,   1,  -1,   3,   1,  -2,   0,   1,   2,   0,  -1,   2,   0,  -1,   0,   0,   1,   0,  -1 },
     {    5,  30,  32, -40,   4,  -9,   2,  -4,   1,  -2,   0,  -1, -15,   0,  -7,   3,   2,   1,   2,   0,   1,   0,   1,   0, -27,   6, -17,  -3,  10,   0,   3,  -1,   2,   0,   1,   0, -74,   5,  -3,  41,  -3,   3,  -1,   1,   0,   1,   0,   0,  27,  13,  34, -17, -14,   1,  -3,   0,   2,   9, -26, -14,  11,   0,   2,   0,   8,  -8,   0,  13,   0,  -1,   0,   0,   2,   1,  -2,  -2,  -1,   0,   0,   0,   4,  -3,   0,   3,   2,   0,  -1,  -1,   3,  -2,   0,   2,   1,   0,   0,  -1 },
     {   -7, -11,  -6,  10,   2,   1,   1,   1,   1,   1,   0,   0, -20, -43,  -5,  -3,  17,   0,   6,   1,   3,   0,   1,   0, -14, -27, -38,  56,   1,   5,   1,   2,   0,   1,   0,   1, -10,   0,  20,  19, -14,   3,  -3,   1,  -2,   0,  -1,   1,  23, -25,  11,  18,  -9,   3,  -3,   1,   3,  -9,  56, -15, -15,   2,  -4,   1,  -2,  22,  -2, -28,   8,  -1,   1,   0,  -1,   4,  -1,   3,   3,  -1,   1,  -1,   1,   8,   1,  -6,   0,   2,  -1,   0,   1,   4,   0,  -2,   1,   1,   0,   0 },
     {  -12, -14, -25,   4,   6,   2,   4,   0,   2,   1,   1,   0,  -7,  -9, -56,  50,   2,   4,   0,   2,   0,   1,   0,   1,   4,  57,   8,   9, -26,   2,  -7,   0,  -3,   0,  -1,   0, -43, -22,  13, -22,  10,   5,   2,   2,   1,   1,   1,   0,  -6,  21, -10,  18,   8,  -5,   1,  -2,  22, -16,   5,   6, -13,   1,  -2,   1,   0,  -4,  13,  -6,   1,   2,   0,   0,   5,   0,   3,  -3,  -1,  -1,  -1,   0,   0,  -2,   3,  -1,   2,  -1,   2,  -1,   0,  -1,   1,  -2,   2,   0,   1,  -1 },
     {   -2,  10,  19,   3, -10,  -2,  -2,   1,  -1,   0,  -1,   0,  -8,  11,  13, -27,   7,  -3,   2,  -1,   1,  -1,   1,   0, -27, -19,  11,  -9,   7,   2,   2,   1,   1,   0,   1,   1, -18, -29,  -9,  -6,  28,  -3,   7,  -1,   3,  -1,   1,   0, -64,  33,   9,  48,  -7,  -7,   1,  -3,  28,  20,  18, -21, -28,   6,  -3,   1,   0,   0, -17, -13,  15,   4,   0,   1,   8,   1,  -1,   8,   0,  -3,  -1,   0,   1,   1,  -3,  -1,   5,   0,  -1,   2,   1,   0,  -2,   0,   3,   0,   0,   1 },
     {    8,  14,   9,   9, -18,   3,  -4,   1,  -2,   0,  -1,   0,   9,  11,  11, -17,   4,  -3,   1,  -1,   0,  -1,   0,  -1,  13,  32,  14,  15, -25,  -1,  -5,  -1,  -2,   0,  -1,  -1,  -4,   8,  23, -39,  -5,   6,  -2,   1,  -2,   1,   0,   0,  26,  -2,   6, -12,  -6,  -1,  -3,   0,   0,  -4,  -1, -61,  10,   7,   1,   3, -18,   5, -45,   7,  45,  -7,   5,  -2,   0,  -3,  -2,  27,  -7,  -8,   1,  -1,  -3,  -1,  -5,   0,  -1,  -3,  -1,   5,  -1,  -1,  -2,   0,   0,  -2,   0,   2 },
     {   -7,   1, -20,  -7,  15,   0,   3,   0,   1,   1,   1,   0,  -3,  10,   7, -24,  17,  -7,   3,  -2,   1,  -1,   1,  -1,   2,  19,   5,  33, -13,  -7,  -3,  -2,   0,  -2,   0,  -1, -34,  16,   2, -21, -23,  11,  -2,   4,  -1,   2,   0,   1, -45, -36,  13, -15,  36,   5,   3,   1, -23,  53,   4,  19,   5, -16,   1,  -5,  21,   3, -16,   1, -17,   1,   1,   1,   3,  -4,   0,   0,   0,   3,   0,   1,   8,  -2,  -3,   0,   1,  -2,   0,   1,   4,  -1,  -2,   1,   1,  -1,   0,   1 },
     {   11,  16,  38,  73, -59,   7, -17,   3,  -7,   1,  -4,   1, -12, -18, -39, -13,   4,  15,   0,   5,   1,   1,   1,   1,  11,  12,  -8,   0,  26,  -9,   3,  -3,   0,  -1,   1,  -1,  -6,  17,   4,   4, -15,  -6,   2,  -2,   1,   0,   0,   0, -16,  -7,   8, -10,   6,   7,   1,   1,   9,   5,  -8,   8,   4,  -6,   0,  -1,   3, -13,   7,   6,  -9,   1,   1,   1,   1,   1,   5,  -7,  -2,   1,   0,   0,   1,  -5,  -1,  -1,   1,   0,   3,  -3,   0,  -2,   0,   0,   1,   0,   2,  -1 },
     {    6,  14,  34,   8, -21,   2,  -6,  -1,  -2,   0,  -1,   0,   1,   3,  26,  15, -25,   8,  -6,   1,  -3,   1,  -2,   1, -12,   1,  16, -32,  -1,   4,  -1,   2,   0,   1,   0,   0, -24, -51,  20, -23,  17,   7,   2,   3,   1,   2,   0,   1,  32, -31,  -4,   1,  26,  -6,   3,  -1, -34,   8,  25,  20,  -3,  -9,   3,  -3,   4,  29,   5, -21,  -9,   2,   0,  -2,  -1,   4, -16,  -6,   7,   0,   0,   0,   2,   5,  -3,   1,   0,   2,  -5,  -1,   1,   2,  -2,   0,   0,   1,  -4,  -1 },
     {    5,  13,  14,   2,   9,  -9,   4,  -5,   1,  -2,   0,  -1,   7,  19,  22,  50, -44,   4,  -9,   1,  -3,   1,  -3,   0,   5,  17,  -6,  -4,   2,   0,  -1,   0,  -1,  -1,   0,   0,   5,  39,  25,  14, -24,  -8,  -3,  -4,  -1,  -2,  -1,  -1, -45,   2,   7, -22, -20,  12,  -2,   3,  10,  -4,  20, -12,  -3,   6,  -4,   1,   6,  10,   0, -43,   3,   2,  -2,   1,   0,  -4, -26,  -4,  19,  -2,   2,   0,   1,  -4,  -3,   4,   0,  -3,  -5,  -2,   1,  -3,  -1,   0,   1,  -2,  -3,   0 },
     {    6,   1,   2,   4,   3,  -7,   3,  -3,   1,   0,   0,  -1,   9,   5,   9, -12,   1,   4,  -3,   1,  -2,   1,  -1,   0,  12,   1,  20,  20,   1, -17,   2,  -4,   0,  -2,   0,  -1,   8,  -6,  -9,   0, -25,  14,  -2,   3,  -2,   1,  -1,   1,  42,  17,  14,  -3,  18, -16,  -1,  -5,  25,  23, -21,  10, -56,  -2,  -1,  -1,   5, -40,   6, -47,   5,  27,  -2,   6, -16,   7, -12,   2,  26,  -5,   0,  -1,  -2,  -7,  -1,   4,  -6,   2,  -3,  -1,  -1,  -4,  -1,   0,  -4,   1,  -2,   1 },
     {   -5,  -5,  -8, -18,  18,  -6,   4,   1,   2,  -2,   2,   0,  -8,  -5, -32, -50,  25,   5,   6,   1,   2,   0,   1,   1,   8,  18,  47, -28,  30, -19,   0,  -5,   0,  -2,  -1,  -1,  -2, -16,  19,  22, -26, -10,   4,  -2,   1,  -1,   0,   0,  -8,  -1,  -6, -18,  -7,  16,   1,   2,  -1, -22,  14,   4,   9,   2,  -3,   1, -20,  29,   9, -22,   7,  -3,   1,  -1,   2,  11, -29,   8,  10,  -5,   3,  -1,  -5,   0,   0,  10,   3,   3,  -7,   3,   0,   0,  -1,   5,   2,   1,  -2,   2 },
     {    2,   1,  17,   8, -10,   1,  -1,  -1,  -1,   0,  -2,   0,   6,  11,  34,   2,   1, -11,  -2,  -4,   0,  -3,   1,  -2,  -1, -13,  41,  38, -34,  -4,  -5,  -1,  -3,   0,  -1,   0, -17, -25, -24,   6, -24,  14,   1,   3,   1,   2,   1,   1, -17, -19,   4, -11,  -3,  19,  -4,   5,  30, -42,   0,   0,  20,   2,  -2,   2, -11,  -8,  38,  11,   0,  -6,  -2,  -1,   4,  26,  10, -24,  -6,  -5,   2,  -1,  -3,   4,  -9,  -4,   1,   6,   3,  -5,  -1,   3,  -5,  -2,   1,   3,  -2,  -1 },
     {   -3,  -8,  -6,  -8,  17, -10,   4,  -3,   2,  -1,   2,   0,  -8, -15,  -7,  -8, -45,  18,  -3,  10,   0,   3,   0,   2,  -9, -19, -29, -21,  19,   8,   4,   2,   1,   1,   1,   1, -12,  -7, -12, -22,  -9,  14,   4,   4,   0,   1,   0,   2,  -9, -29,  -6, -20,  42,  12,   0,   1,  53,  -9,  -2,   8,  13,  -8,  -3,  -2, -23, -15, -10,   3,  25,  -2,   4,   0,  -6,   7,   9,  34,   3,  -9,   1,  -3,   0,   2,   6,  -3,  -1,   1,   0,   7,   1,   1,   3,   0,   0,   2,   1,   3 },
     {   -2,  -4,  -2,  -9, -30,  30,  -6,   5,  -2,   3,  -2,   3,   2,   6,  -8,  15,  32, -17,  -5,   0,   0,   0,   2,  -1,  -6, -17, -19, -12, -24,  -4,  13,   4,   2,  -1,   2,  -1,   5,  12, -42, -10,  -2,  32,   0,   0,  -1,   1,   0,   0,   1,  30,  20, -20,  27,  -1, -10,  -2,  14, -12,   6,   8,  12, -14,   0,   0, -10,  37,  -8, -17,  -4,  -8,   4,  -1,   6, -12, -41,   6,  10,   5,   1,   1,   0, -12,   9,  20,   1,  -3,   1,   0,   0,  -3,   4,   6,   1,  -2,   0,  -2 },
     {  -11, -16, -16, -53, -51,  56,  -9,  11,  -3,   8,   0,   2,   1,   4,  11,  13,  14,   1,  -6,  -1,  -2,   0,  -2,   1,  -1,   0, -10,  -2,  29, -23,   2,   0,  -1,  -2,  -1,   0,  -3,   5,  26, -19,   4, -15,   8,  -4,   2,   1,   1,  -1,  -9, -33,   4,   8, -12,   3,   8,  -3,   9,   0, -15,  -3,  -4,  15,  -4,   0,  -3, -19,  13,  -5,   9,   6,  -5,   4,  -7,  16,   5,  -7,  16,  -5,   2,  -2,   1,   6,  -8,  10,   0,   1,   3,   1,   1,   1,   0,   3,   0,   1,   0,  -1 },
     {    7,   3,   1,  -6,   6,   5,  -6,  -1,  -3,   0,  -1,   1,   9,   3,   6,  -6,   3, -13,   5,  -7,   3,  -2,   0,  -1,  11,   7,   9,  -5,   1,  10, -13,   3,  -2,   0,  -1,   1,   8,   0,  26,   0,  13, -18,   3,  -4,   1,  -2,   1,  -1,  26,  -9,   7,  11,  -8,  18,  -8,   1,  74,  22,  11,   0,  16, -33,  -5,  -5,  -5,   8, -27,   3, -51,  -5,  10,  -3,   0, -16,   0, -27,   3,  19,  -2,   2,  -8,   4,  -8,   1,  -1,  -4,   0,  -2,  -5,  -1,  -4,   1,  -2,  -2,   1,  -1 },
     {    2,   3,   0,  15,  19, -13,  -1,  -4,   2,  -1,  -1,  -1,  -4,  -4,  -8, -24,  15, -23,  16,  -5,   4,  -2,   2,  -1,  -2,  -3, -17, -34, -34,  35,  -3,   6,   1,   2,  -1,   1,   2,  29,  28, -19,  33,   8, -11,   2,  -3,   0,  -1,   0,  -6, -26,  35,   3,  -3, -12,  -1,  -3,   6, -10, -16,   1,  -6,  11,   0,   4,  -1,  -9,  36,  -7,   9,  -1,  -3,   0, -13,  32, -19, -17,  12,  -8,   3,  -3,   2,  -2, -14,  14,   1,   2,  -1,  -1,   1,  -2,  -4,   8,   0,   2,   1,   1 },
     {    5,   4,  14,  17,  13,  -2,  -5,  -3,  -2,   0,  -1,  -1,   1,  -3,   4,  33,  45, -59,   6,  -5,   2,  -2,   0,  -2,  -4, -11,   0, -35,  -5,  -5,  20,  -7,   5,   0,   3,   0,  -5, -18,  19,  13, -31,  15,   0,   0,   1,  -1,  -1,   1,  -7, -27, -25,   1,  -9,   8,   3,   3,   9,   0,  -9,   8,  -9,  18,  -4,   1,  10, -16, -22,   2,   9,   6,   0,   3, -12, -21,  16,  24,   1,   7,   1,   1,   2,   2,  20,  -2,  -2,   3,   1,  -5,   0,   3,   2,  -2,  -1,   3,   0,  -2 },
     {   -2,  -7,  -7,  -7, -21,  11,  -5,   5,   1,   1,  -1,   1,  -5, -15, -15, -26, -21,  10,   6,   1,   2,   0,   2,   2,   0,  -5,   2,  -1, -39,  32,  -9,   5,  -1,   3,   0,   1,  -2, -20, -16,  15,   3,  -6,  -2,   3,  -1,   2,  -1,   0,   2,  -2, -39, -18, -19,  20,   9,   2,   7,  18,   2,  -7,  20,  22,  -8,   4,  58, -13,  -6, -13,  10,   1,  -5,   1, -12, -16, -13,  -2,  27,  -4,   3,   1,   0,   1,   4,  26,   0,  -4,  11,  -1,  -2,   2,   3,   1,  -2,   0,   4,  -1 },
     {    1,   4,  -1,  11,  14, -20,   7,  -2,   3,  -1,   0,  -1,  -1,   0,   0,  10,  13,  -7,  10,  -5,  -3,  -2,   2,   0,  -9,  -8, -25,  -9,  12, -12, -12,   4,   1,   1,  -1,   1,  -4, -11, -20, -55, -27, -12,  34,   3,   7,   4,   1,   0,  13,  28,   7,   7, -24,  53,   4,  -8, -21,  15,  12,  -1,  21,   1, -23,   0,  -9, -19,   6,  -4,  -3,  -2,   0,   2, -11,  12,  -1, -16,  18,  -4,  -1,   0,   3,   6, -18,   9,   0,  -1,   1,   3,   1,  -1,  -2,   4,   0,  -2,   3,   0 },
     {    5,   3,   1,   1,   5,  -2,  10,  -9,   4,  -3,   2,  -1,   7,   5,   3,   4,  13,  15, -16,   6,  -5,   4,  -1,   1,   7,   2,  -3,  -4,   8, -23,  14,  -1,   3,  -4,  -1,  -2,   8,   1,   5,  -5,   2,  20, -11,   4,  -4,  -2,   1,  -1,  13,   3,  -6,   2,  -6,  -2,  12,  -4,  31,  38,  10,  13,   8,  30, -26,  -1,  31,  16,  24,   4,  21, -53, -14,   4,  -8,   6, -14,  -6, -46, -11,  20,  -1,  -1, -14, -10, -17,   0,   0,  -8,   3,   3,  -5,  -2,  -7,   0,   0,  -4,   2 },
     {   -3,  -4,  -7, -17, -23, -15,  17,   1,   5,   2,  -1,   0,  -1,  -3,  -4, -12,  -5,  19,  -5,  -2,   1,   1,  -3,   1,  -1,  -1,  -6, -32, -26, -14,  21,  -2,   5,  -1,   2,  -1,  11,  10,  27,  12, -25,  48,   6, -21,   2,  -2,   2,   0,   0,  11,  -3,  26,  15,  16, -46,  12,  -9,  -3,  10, -13,   3, -22,  -2,   9,  21, -12,   1,   5,   2,  10,  -2,  -1,  12,   0,  24, -21,   1,  -7,   1,   5,  -8,  18, -15, -10,   7,   1, -10,   4,  -3,  -1,  -6,  -5,   1,  -2,  -6,   3 },
     {   -6,  -8, -10, -10, -54, -54,  60,  12,  11,  -5,   3, -12,   3,   6,   6,  13,  15, -11,  25, -12,  -1,  -7,   2,  -6,   3,   2,   6,   2,  31,  27, -33,  -8,  -7,   1,   7,   6,   1,   1,  -5,  -1,   5,  11, -19,   9,   2,   2,  -3,   3,  -2,  -1,  -2, -13,   8,   0,   0,   2,   8,   1,   6,  -2,  -6,  10,   6,  -5,   1,   7,   0,   5,  -3,   0,  -1,  -2, -15,  -2,  -1,   2,  -1,   2,   0,   3,   2,   7,   1,   0,  -1,   0,   6,   1,   2,  -1,   5,  -3,   0,  -1,  -4,  -2 },
   }
 },
 {  //12
   {
     {  108, -42,  -3,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0,   0, -46,  14,   2,   2,   1,   1,   0,   0,   0,   0,   0,   0,  13,  -5,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0, -16,   1,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   3,  -1,   0,   0,   0,   0,   0,  -5,  -1,   1,   0,   0,   0,   0,   0,  -1,   3,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {   49,  -6,  -6,  -2,  -2,  -1,  -1,   0,  -1,   0,   0,   0,  83, -42,  -1,  -3,  -1,  -1,   0,   0,   0,   0,   0,   0, -68,  10,   8,   1,   2,   1,   1,   0,   1,   0,   0,   0,  10,   4,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -16,   0,   4,   0,   1,   0,   0,   0,   2,   6,  -2,   0,   0,   0,   0,   0,  -5,  -3,   1,   0,   0,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0,  -3,  -1,   0,   0,  -1,   2,   1,   0,  -2,   0,   0,   0,  -1,   1,   0,   0 },
     {    3,  -6,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  65,   8, -13,  -1,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0,  58, -52,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0, -71,   2,  17,   0,   4,   0,   2,   0,   1,   0,   1,   0,   1,  18,  -3,  -1,  -1,   0,   0,   0,  -8,  -2,   0,   0,   0,   0,   0,   0,  -1,   4,  -1,   0,   0,   0,   0,   0,  -5,  -1,   1,   0,   0,   0,   0,   0,  -1,   3,   0,   0,  -3,   0,   0,   0,  -1,   1,   0,   0,  -2,   0,   0,   0 },
     {   14,  10,  -7,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  16,  -6,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  56,  15, -17,   0,  -4,   0,  -2,   0,  -1,   0,  -1,   0,  45, -67,   0,   2,   0,   1,   0,   1,   0,   0,   0,   0, -65,  -2,  25,  -1,   5,   0,   2,   0,   2,  22,  -4,  -2,   0,  -1,   0,  -1,  -8,  -3,  -1,   1,   0,   0,   0,   0,  -1,   3,   0,   0,   0,   0,   0,   0,  -5,  -1,   2,   0,  -1,   3,   0,  -1,  -3,   0,   0,   0,   0,   1,   0,   0 },
     {   -8,   8,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -28, -20,  14,   1,   3,   0,   1,   0,   1,   0,   0,   0, -23,   8,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0, -42, -14,  20,   0,   4,   0,   2,   0,   1,   0,   1,   0, -34,  75,   3,  -6,   0,  -2,   0,  -2,  54,   7, -36,   1,  -5,   1,  -2,   0,   2, -24,   4,   5,   0,   1,   0,   1,   8,   3,  -1,  -1,  -1,   0,   0,   0,   1,  -5,   0,   1,   5,   1,   0,  -1,   1,  -2,  -1,   0,   3,   0,  -1,  -1 },
     {   16,  28, -14,   0,  -4,   0,  -1,   0,  -1,   0,  -1,   0,  13,  -4,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  21,  19, -14,  -1,  -3,   0,  -1,   0,  -1,   0,   0,   0,  17, -10,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  30,  15, -26,   0,  -4,   0,  -1,   0,  23, -81,  -2,  13,  -1,   4,   0,   3, -39,  -5,  45,  -3,   4,  -1,   1,  -1,   0,  23,  -5,  -8,   0,  -1,   0,  -1,  -8,  -4,   0,   1,  -1,   3,   0,  -1,  -4,  -1,   1,   1,   0,   2,   0,  -1 },
     {    4, -14,   5,  -1,   2,  -1,   1,   0,   0,   0,   0,   0,  17,  32, -16,   0,  -3,   0,  -1,   0,  -1,   0,  -1,   0,  14,   4,  -2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,  23,  20, -20,  -1,  -3,   0,  -1,   0,   0,   0,   0,   0,  15,  -5,  -3,   0,   0,   0,   0,   0,  24,  11, -36,   1,  -3,   0,  -1,   0,  14, -81,  -1,  21,  -1,   6,   0,   3, -26,  -1,  48,  -5,   2,  -1,   0,  -1,   0,  16,  -8, -10,  -7,  -6,   1,   4,  -1,   1,   1,  -1,  -4,  -1,   2,   1 },
     {   34, 106, -32,  -3, -10,  -3,  -4,  -2,  -3,  -1,  -2,  -1,  -8,  17, -11,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  -7, -14,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,  -7,   6,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,  -7,   0,   0,   0,   0,   0,   0,  -2,  14,  -3,  -4,   0,  -1,   0,   0,  14,  -6, -21,   4,  -1,   1,  -1,   0,   2, -26,   4,   8,   0,   2,   0,   1,  -1,  -5,   9,   0,   0,  -2,   2,  -1,   0,  -1,   1,   0,   0,  -2,   1,   0 },
     {    1,  22,  -2,   0,  -2,   0,  -1,   0,   0,   0,   0,   0, -23, -13,  -7,   6,   0,   2,   0,   2,   0,   1,   0,   0, -17, -41,  20,   2,   3,   2,   2,   1,   1,   0,   1,   0, -13, -10,   0,   4,   1,   1,   0,   0,   0,   0,   0,   0, -15, -27,  25,   2,   3,   1,   1,   0, -15,  11,   5,   1,   1,   0,   0,   0, -10, -14,  41,   1,   1,   1,   0,   0, -13,  70,  11, -27,   1,  -6,   1,  -4,  12,  12, -45,  -1,   1,  -8,  -1,  11,   7,   2,  -1,   0,   1,   1,   0,   0 },
     {    2,  -4,   8,  -6,   1,  -1,   1,  -1,   0,   0,   0,   0,  19,  76, -20,  -5,  -6,  -2,  -3,  -1,  -2,  -1,  -1,  -1, -17,   8,  -4,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   9,   0,  -5,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,  -1,  14,   7,  -6,   1,  -2,   0,  -1,   7,  13, -17,  -2,  -1,   0,   0,   0,   1,   9,   4,  -6,   1,  -2,   0,  -1,   7,  24, -48,  -6,   1,  -2,   1,  -1,   9, -52, -22,  29,  -2, -12,  30,   4,   2,  -1,  -2,  -6,  -2,  -4,   1,   2 },
     {    7,  12,   4,  -5,   0,  -1,  -1,  -1,   0,  -1,   0,   0,  -3, -61,  11,   2,   4,   2,   2,   1,   1,   0,   1,   1,  50,  35,  19, -12,  -1,  -4,  -1,  -3,  -1,  -1,  -1,  -1,  -8,  50, -30,  -3,  -3,   0,  -1,   0,  -1,   0,   0,   0,   4, -10,  -3,   0,   0,   1,   0,   1,   1,  21, -16,  -1,  -1,   0,   0,  -1,   9,  -8,  -2,  -1,   0,   0,   0,   0,  -4,  13, -25,  -3,   2,  -1,   1,   0,   5, -35, -12,  21,  -3,  -7,  25,   4,   1,   1,   2,  -6,  -3,   0,   0,  -1 },
     {    0,   7,   0,  -3,   0,  -1,   0,  -1,   0,   0,   0,   0,  10,  25,   9,  -7,   1,  -3,  -1,  -2,   0,  -1,   0,  -1,  -4,  81, -26,  -4,  -4,  -2,  -1,  -1,  -1,  -1,   0,  -1, -66, -12, -19,   9,   0,   3,   0,   2,   1,   1,   0,   1,  -8, -29,  10,   9,   0,   3,   0,   1,   2,   1,  17,  -5,   1,   0,   0,  -1,   2,   3,   3,  -1,   0,  -1,   1,   0,   4,   7,  14,  -6,  -1,  -1,   0,   0,   2,  19, -16, -12,   5, -11, -15,   9,   3,  -5,   6,   3,   2,  -2,  -1,   0 },
     {   -4,  -3,   6,  -1,   2,   1,   0,   0,   1,   0,   0,   0,  -8,  -5, -20,   8,   0,   3,   0,   1,   0,   0,   0,   0, -12,  -1,  -1,  -1,   2,  -1,   1,   0,  -1,   1,   0,   0, -30, -24, -19,  16,   0,   4,   0,   2,  -1,   1,   0,   1,   5, -30,  26,   5,   1,   1,   0,   0,  -3,  -6,  -6,   8,   1,   1,   0,   1,   0, -25,  25,   5,   0,   1,   0,   1,  -7,  -1,  -4,   9,   0,   1,   1,   1,   6, -25,  59,  12,  -5,  49,  26, -42,  -1,  10, -32,  -6,  -3,   4,  -1,   4 },
     {    0,   7,  14,  -6,   0,  -1,  -1,  -2,   0,   0,   0,   0,  -3,  25,  -2,  -4,  -1,   0,   0,  -1,   0,  -1,   0,   0,   0, -12,  11,  -1,   2,   0,   2,  -1,   1,   0,   1,   0,   1,  57, -18,  -9,  -1,  -3,   0,  -1,  -1,  -1,   0,  -1, -89, -15, -29,  13,   3,   4,   2,   3,  16, -27,  19,  14,  -1,   3,   0,   1,  -3,   4,  11,  -3,  -1,  -2,   0,   0,   2,  -1,   2,   6,   0,   1,   0,   0,   3,   6,  15,  -4,   3,  10,  -6,  -9,   2,  -5,  -7,   1,   3,  -2,   0,   1 },
     {   -8, -13, -30,  10,   1,   5,   0,   3,   0,   1,   1,   1,  -8, -11,  -2,   1,   1,   0,   1,   1,   0,   0,   0,   0, -11, -13, -19,  11,   0,   3,   1,   2,   1,   0,   0,   1, -12, -14,   0,   2,   2,   1,   0,   0,   1,   0,   0,   0, -15, -21, -15,  16,   0,   4,   0,   1,  12, -18,  18,   3,   1,   1,   0,   1,  -8,  -9, -12,  11,   0,   2,   1,   1,   0, -18,  19,   9,  -1,   0,   0,   0,  -4, -11, -10,  18,  10, -28,  64,  20,  -1,  34,  32, -45,  -4,   5, -21,  -8 },
     {    3,  17,  42,  -8,   0,  -3,   0,  -2,  -2,  -2,   0,  -1,   1,   6,  31, -11,   2,  -4,   1,  -2,  -1,  -1,   0,   0,  -5,  15,  -7,  -5,   1,   1,   1,   0,   0,   0,   0,   0,  -1,  -4,  16,  -9,  -1,   0,   0,   0,  -1,   0,   0,   0, -12,  21, -25, -13,   3,  -2,   2,   0, -84,  -8, -31,  17,   6,   3,   3,   2,   7,  -9,  13,  23,  -2,   5,  -2,   1,  10,  -1,  19,  -5,  -3,  -1,  -1,  -1,  -2,  -1,   0,  -1,   7,   4,  19,  -1,  -1,  17,  -4, -17,  -1,  -1, -14,   1 },
     {  -13, -24, -98,  18,   0,   7,   1,   4,   2,   2,   1,   2,  -2,   6, -19,  12,  -2,   2,   0,   1,   0,   1,   0,   0,   0,  18,  18,  -6,   1,  -2,   1,  -1,  -1,   0,  -1,   0,  -2,  10,   0,  -4,   0,  -2,   1,  -1,   0,   0,   0,   0, -16,  12,  -6,  -6,   2,  -3,   2,   0, -43,  -3, -20,  16,   3,   3,   2,   2,  14,  -1,  18,   7,  -4,   2,  -2,   0,   9,   0,   2,  -3,  -1,  -1,   0,   0,   2,  -8,   5,  -3,  -3,   2, -11,   0,   2,  -6,   2,  13,   1,  -1,   9,   0 },
     {    1,   0, -27,  10,  -3,   2,   0,   2,  -1,   2,   0,   1,   8,  13,  30,  -9,   0,  -3,  -3,  -1,  -1,  -1,  -1,  -1,  10,  11,  16,   2,  -3,  -1,   0,   0,  -1,   0,   0,  -1,   9,  15,  16, -15,  -3,  -3,   0,  -1,   0,   0,   0,  -1,   8,  11,   7,   2,  -3,   1,  -1,   0,  16,  14,  21, -19,  -3,  -2,  -1,  -1,  -5,  11,  -6,  -8,  -2,  -2,  -1,  -1,  10,  13,  18, -17,  -3,  -4,   0,  -2,  -1,  13, -13,  -9,   3,  14,  16, -28, -12,  27, -62, -30,  -4, -23, -33,  38 },
     {    3,   0,  -2,  -1,  -1,   0,  -1,   0,  -1,   0,   0,   0,   6,  -3,  21,  -7,  -2,   0,   1,   0,   0,   0,   0,   1,   0,  -5, -16,   2,   0,   1,  -2,   0,   1,   1,   1,   1,   7, -10,   9,   8,  -1,   3,  -1,   1,  -1,   1,   0,   0,   1,  -1,  -8,   0,   1,   0,  -1,  -1,  13,  -9,  18,  25,  -5,   3,  -3,  -1,  98,  -1,  36, -28, -13,  -4,  -3,  -2, -24, -11, -17, -25,   8,  -2,   3,   0,  -6,   5,  -6,   6,  -4,  -1,   1,  -3,  -2,   3,   1,  -3,  -4,   2,  -6,  -3 },
     {    3,   4,  24,   3,   1,  -2,   0,  -1,  -1,   0,  -2,   0,  -7, -21,-100,  13,   6,   6,   4,   3,   2,   2,   1,   1,   5,  27, -14,  24,  -1,  -1,  -2,   1,  -1,   0,   0,   0,   3,  12,  37,  -6,  -1,  -5,   0,  -1,  -1,  -2,   0,  -2,  -4,  11,  -7,  -3,  -4,   1,   0,  -1,  -3,  -1,  13,  -4,   2,  -1,   0,   1,  19,   8,   3,   4,  -3,  -2,   0,   0,  12,   2,  10, -11,  -6,  -1,  -1,  -1,  -5,   5,  -9,  -4,   2,  -2,   4,  -3,  -3,   1, -13, -10,  -2,  -6,  -7,   8 },
     {   -2,  -3, -15,   8,   1,   4,   5,   0,   0,   2,   0,   1,  -5, -16,   7, -24,  11,  -1,   6,  -2,   2,   0,   1,   0,  -5, -21, -97,  12,   5,   4,   3,   3,   2,   1,   0,   1,   4,  31, -12,  20,   3,   2,  -1,   1,   0,   0,  -1,   0,  -7,  12,  13,  -4,  -3,   0,  -1,   1, -16,  -1,  -2, -10,   3,  -3,   1,  -2,  -7,  -8,  -7,   3,   4,   0,   3,   2,  -5,   7,  -7,   1,   3,  -1,   3,   0,  12,   2,  -8,  11,   2,  -2,   1,  -7,  -2, -11, -22,  13,   4, -14,  20,  21 },
     {   -3,   4,  -6,   2,   2,   1,   0,   2,   1,  -1,   0,   0,  -4,   5,  -1,   1,  -3,   1,   1,  -1,   1,   0,   2,  -2,  -3,  22,   7,   1,  -2,  -1,  -2,   0,   0,   0,  -1,   0,  -3,   7,  42, -14,  -1,   1,  -1,   0,  -1,  -2,  -1,  -1,  -7,   1,  -2, -18,   3,   0,   3,  -1, -13,  -3,   4,  -9,   3,   1,   0,   1, -20,  -9, -14, -20,   9,  -3,   3,  -1, -94,   4, -20,  20,  17,   0,   3,   0,   6,  24,  10,  27,  14,   2,   5,  -2,  12,  -1,   0,  -4,  -1,   0,  -2,  -1 },
     {   -4,  -5,  -5, -33,   9,   2,  -1,   1,   0,  -1,   0,  -1,  -7, -13, -17, -31,   6,   3,   0,   1,   1,   1,   1,   1,  -8,  -9,  12, -18,  -1,   0,   2,   0,  -1,   0,  -1,   0, -11, -40, -21, -36,  10,   2,   6,  -1,   5,   0,   2,   1,  -1, -18, -45,   3,   9,   6,   4,  -1,   5, -14, -12,  -1,  14,   1,   2,  -1,  13,   1, -24,  12,   8,  -1,   1,  -2,  -7,  -1,  -4,   4,   9,  -1,   3,  -2, -14,   3, -23,  20,   2,  -2,   2,   4, -14,   1, -32,  25,  10, -17,  23,  35 },
     {   -6,  -3, -19, -94,  13,   0,   5,   0,   4,   3,   1,  -1,  -2,  -3, -17, -20,   3,   8,  -6,  -2,  -1,   1,  -1,   1,   1,  -4,  -8, -18,  11,   0,   3,   0,   1,   0,   2,  -3,   5,   9, -17,   9,  -6,   4,   2,   2,   0,  -1,  -1,  -2,   3,  14,  22, -17,   4,   2,  -1,  -2,  -1,   1,  -6,   4,  -5,   0,   0,   0,   0,  10,   9, -11,   5,  -2,   1,  -2,   0,   4,  -1,  15,  -5,  -5,   0,   0,   3,  17,  10, -12,   6,  -2,  -8,   6,   8,   5,  10, -27, -10,   3, -38,  -2 },
     {    1,   5,   9,  21,   3,  -2,  -2,  -3,  -1,  -1,   0,   0,  -2,   1, -20,  18,   1,  -1,   0,  -1,   1,   0,  -1,   0,  -4,  -1,   7, -33,   7,  -7,   3,  -1,   2,   1,   0,  -1,   4, -12, -64,  10,  17,  -3,   4,  -1,   2,   0,   3,   0,   3,  58,  -4,  45, -10,  -2,  -5,  -3, -27,   3,  31,  -8, -15,   0,  -4,   1,   0,  -3,  -4, -16,  -1,  -1,   1,   1, -23,   0,  -1,  -2,   3,   0,   3,   0,  -4,  15,  -1,   7,  12,  -3,   1,   0,  -1,   7,  -2,  -6,   2,  -4,  -9,   1 },
     {    1,   0,  -4,   2,  -1,  -2,   2,  -1,  -1,   1,  -1,   0,   0,  -4,  -2,  -5,   0,  -4,  -1,   2,   2,  -1,  -1,   2,   1,   2, -18,   5,  10,  -2,   5,   0,   1,   0,   0,   1,   0, -14,   8,  -6,  -2,   2,   5,   0,   1,  -1,  -1,   0,  13,  -1, -61,  15,  17,   2,   4,   1,   7,  52,  14,  49, -14,   2,  -3,  -2, -18,  -5,   6, -17,  -7,   0,  -1,   0,  -3,  25,   2,  24,  -6,   3,  -3,   1,  43, -14,  16, -22, -14, -10,  -9,  -8,   4,   0,  -1,  -3,  -5,  -6,  -9,   1 },
     {    5,   3,   2,  51,  31,  16,  -3,   1,  -2,   0,   0,  -3,  -4,  -1, -17, -87,  28,  18,   4,   2,   2,   2,   2,   0,   1,  -2,  10, -35,  -2,   1,   1,   0,   0,  -1,   2,  -1,   0,   0,   3,  -9,  -3,   5,  -3,   3,  -1,   0,   0,  -1,   0,  -4,   8, -15,   1,   0,  -5,   0,   3,  -2,   2,  -5, -12,  -3,  -2,   4,   0,  -2,   6,  -3,   8,   3,   6,   2,   5,  -2,  -2,  -1,  -3,  -2,   0,   2,   5,  -4,   3, -10,  -1,   9,  -5,   1,   5,  -4,  11,  -9,  -9,   6, -15, -17 },
     {    2,   0,   7,  25,   1,  -8,   2,   4,  -4,  -4,  -2,   1,   0,  -2,  -6,  44, -20,   4,  -1,  14,   0,   7,   2,   8,  -5,  -7, -30, -84,  14,   5,   2,   2,   6,   2,   0,   3,   0,   6,  11,  -7,   5,   4,   3,  -1,   0,   2,  -2,   0,  -3, -21,   0, -23,   4,   2,  10,   6,  16,   3, -25,   7,   3,   3,   1,   0,   0,  17,   6,  -4,   1,   6,  -2,   0,   6,  -1,   1,   9,  18,   3,  -2,  -2,  -5,   4,  -4,  -6,  11,   3,  -6,  10,  -6,  11,  -1,  -4,   2,  -6, -20,  17 },
     {    5,   0,   2,   9,   4,  -5,   3,  -3,  -3,   0,   2,   0,   7,  -2,   5,  13,   1,   1,   0,  -3,   0,  -1,   0,   2,   5,   1,  11,  -3,  -2,   0,   4,  -1,  -2,   2,   0,   0,   6, -11,   0,   6,  -1,   0,   0,  -1,  -3,   0,   1,  -1,  -2,   0,   2, -10, -15,  -1,  -3,   2,  -1, -37,  -1, -29,   6,   7,   0,   3,  43,  -6, -26,   5,  23,  -9,   8,   0,  18,  37,   3,  43,  -9,  -1,  -1,  -3,  69,   6,  12,  -4,  12, -25,   1, -13,   0,  -4,  -1,  -2,   4,   0,  -3,   1 },
     {    3,   1,   0,  10,  -5,  -2,   2,  -1,   0,  -2,   1,   2,   3,   3,  -5, -12,   4,  -4,   5,  -2,   2,   2,   0,  -5,   5,  -1,   9,   6,   6,  -5,   1,   1,  -2,   0,  -3,   2,   5,  -2, -17,  25,   1,   3,   4,   0,   0,   1,  -1,  -1,   7,  -3,  -7,  -5,  -8,  -3,  -4,   4,  21,  14, -28,  20,  14,  -1,   1,   4,  -1,  52,  13,  47, -30,  -7,  -4,  -3, -22,  -4,   9, -15, -13,   3,   3,   0,  19,  25,   7,  14,  57, -12,   8, -19,  -1, -15,  -2, -11,   0,  -3,   9,  -5 },
     {    2,   4,  -4,  -6,  50, -10,   3,  17, -13,  16, -22,  -2,   4,   3,   5,  15,  10,  -9,  15,  17,   1,  -4,  16,  20,   0,  -2, -12,   5, -36,   5,  -5,  -6,  35,  11,  21, -33,   2,   4,   1, -39,   3,  10,   2,  18,  15,  18, -23,  -9,   5,   1,   8,  29, -11,  -2,   9,  -2,   3,   6, -11,   8,  10,  -2,   3,   7,   1,   9,   0,   0,  11, -17,   3,   6,   2,   1,   1, -18,  -5,   1,  -1,  -3,  11,  11,   6,   4,   3,   7,   7,  -7,  10,   3,   5,   6,  -1,   7,   2, -12 },
     {   -3,  -6,  -2,  -1, -66,   9, -26,  -3, -11,   4,  -5,   5,  -1,  -4,  -6, -14, -25,   8,  -9,  -8,   1,  -4,   0,   3,  -3,  -4, -20,  -9, -37, -12,  -9,  -3,  -7,   3,  -1,   0,  -1,   1, -15, -47,   5, -31,   7,  -7,   1,   1,   5,  -3,   5,   3,  -3,  -7,  14, -17,  14,  -4,   4,   7,   3, -15,   5, -13,   3,  -2,   6,  -1,   3,  11,   4,  -3,   0,   4,   6,  12, -12, -13,  -7,   4, -10,   1,   9,  19,  14,   3,   4,  14,   3,  -8,  16,   6,  14,  -3,   9,   7,  -7, -23 },
   },
   {
     {  -73,  54,  -9,   4,  -1,   1,  -1,   1,   0,   0,   0,   0,  69, -46,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0, -27,   9,   9,  -3,   2,  -1,   1,   0,   0,   0,   0,   0,   8,   4,  -8,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -3,  -4,   3,   1,   0,   0,   0,   0,   3,   1,   0,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   68,  -8, -22,   3,  -5,   1,  -2,   0,  -1,   0,  -1,   0,  12, -47,  28,  -6,   5,  -1,   2,   0,   1,   0,   1,   0, -54,  54, -13,  -1,   1,  -1,   0,  -1,   0,   0,   0,   0,  29, -24,  -5,   7,  -3,   2,  -1,   1,  -1,   0,   0,   0, -17,   5,  10,  -4,   2,  -1,   0,   0,   5,   1,  -6,   1,   0,   0,   0,   0,  -6,   0,   3,   0,   0,   0,   0,   0,   2,   1,  -2,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   1,   1,  -1,   0,  -2,   0,   1,   0,   1,   0,  -1,   0 },
     {  -39,  44, -31,  13,  -5,   4,  -2,   2,  -1,   1,  -1,   1, -68,  -2,  40,  -8,   6,  -1,   2,  -1,   1,   0,   1,   0,  36,  29, -31,   1,  -3,  -1,  -1,   0,  -1,   0,   0,   0,  19, -37,   9,   5,  -2,   2,  -1,   1,   0,   0,   0,   0,  -5,  12,   4,  -7,   2,  -1,   1,  -1,   8,  -7,  -5,   3,   0,   0,   0,   0,   0,   1,   4,  -2,   0,   0,   0,   0,   5,  -2,  -3,   0,   0,   0,   0,   0,   1,   0,   1,   0,   2,  -1,  -1,   0,   0,   0,   1,   0,   2,  -1,  -1,   0 },
     {   48,  41, -52,   8,  -8,   1,  -3,   0,  -2,   0,  -1,   0,   7, -39,  27,  -1,   0,   1,   0,   1,   0,   0,   0,   0,  39, -34,  12, -10,   5,  -2,   1,  -1,   1,   0,   0,   0, -53,  31,  -1,   2,   2,  -1,   1,   0,   1,   0,   0,   0,  16,  -1,  -7,   2,  -3,   1,  -1,   0, -10,  -3,   5,  -1,   1,   0,   0,   0,   1,   2,  -1,  -1,   0,   0,   0,   0,  -5,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0 },
     {   -5,  18,   1,  -8,   2,  -3,   1,  -1,   0,  -1,   0,   0, -56,  -7,  -4,  18,  -1,   5,  -1,   2,   0,   1,   0,   1, -27,  16,  40, -23,   6,  -5,   2,  -2,   1,  -1,   0,  -1,  18,  42, -55,  10,  -2,   0,   0,   0,   0,   0,   0,   0,  22, -46,  21,  10,  -7,   4,  -2,   1,  -4,  17,   2, -14,   5,  -2,   1,  -1,   7,  -8,  -7,   7,   0,   0,   0,   0,   1,   3,   3,  -2,  -1,   0,   0,   0,   4,  -4,  -2,   1,   1,   1,   1,  -1,   2,  -2,  -1,   1,   1,   1,   0,   0 },
     {  -27, -27,   3,  18,  -2,   4,   0,   2,   1,   1,   0,   1, -24,  -4,  38, -25,   9,  -5,   3,  -2,   1,  -1,   1,  -1, -54,  15,  -9,  15,   1,   1,   1,   0,   1,   0,   0,   0, -37,  47,  19, -12,  -1,  -1,  -1,  -1,   0,  -1,   0,   0,  35,  16, -40,   4,   0,  -1,   0,   0,   8, -22,  12,   7,  -5,   2,  -1,   1,   7,   9,  -4,  -7,   2,  -1,   0,   0,   4,  -6,  -2,   4,  -1,   0,   0,   0,   4,   3,  -1,  -2,   3,  -3,  -1,   2,   3,   1,   0,  -1,   2,  -2,  -1,   1 },
     {   15,  34, -34,  22, -15,   5,  -5,   2,  -2,   1,  -1,   0, -12,  31, -15, -12,   4,  -2,   2,  -1,   0,   0,   0,   0, -34, -46,  31,  14,   0,   2,   0,   1,   1,   1,   0,   0,  52,   6,  12, -27,   4,  -4,   0,  -2,   0,  -1,   0,  -1, -33,   6, -28,  18,   2,   0,   2,   0,  18, -13,  20,   0,  -7,   4,  -2,   1,  -8,   5,  -3,  -7,   5,  -2,   1,   0,   6,  -2,   1,   4,  -1,   0,   0,   0,  -2,   0,   0,  -2,   3,  -1,   1,   1,  -1,   0,   0,  -1,   2,  -1,   1,   1 },
     {    3,  24,  23, -43,  17, -11,   5,  -4,   2,  -2,   1,  -1, -31, -29, -20,  46, -12,   9,  -2,   3,   0,   1,   0,   1, -20,  -3,  29, -17,   1,   2,  -2,   1,  -1,   0,   0,   0, -23,  -6,  19,  -4,   9,  -4,   3,  -2,   1,  -1,   1,  -1, -15,  59, -14,  -7,  -4,  -2,   0,  -1,  25, -18, -20,  10,  -1,   2,  -1,   1,  -3,   0,  13,   0,  -3,   0,  -1,   0,   7,   3,  -8,  -4,   2,   0,   0,   0,   0,  -2,   2,   3,   4,   0,  -2,  -2,   0,  -1,   1,   2,   2,   0,  -1,  -1 },
     {   19,  32,  18, -36,  10, -11,   2,  -4,   1,  -2,   0,  -1,   4,  -9, -22,  16,  -2,   2,   0,   1,   0,   0,   0,   0,  19,  35, -27,  18, -17,   6,  -5,   2,  -3,   1,  -1,   1,  15,  -3,   7, -28,  11,  -5,   2,  -1,   1,   0,   0,   0,  31, -29, -36,  30,  -4,   3,   0,   1, -16, -29,  45,  -8,  -5,   4,  -2,   2,  -6,  19, -11,  -8,   8,  -3,   2,  -1,  -3,  -9,   3,   7,  -3,   1,   0,   0,  -2,   4,   0,  -4,  -2,  -4,   0,   2,  -2,   1,   1,  -1,  -1,  -2,   1,   1 },
     {   13,  44, -23,  -9,   3,  -5,   2,  -2,   0,  -1,   0,  -1,   3,  47, -21,   8, -14,   3,  -4,   1,  -2,   0,  -1,   0, -47,  -3, -25,   8,   7,   1,   2,   1,   1,   1,   1,   0, -14, -15,  30,  13, -10,   4,  -2,   1,  -1,   0,  -1,   0,  46, -13,  10, -30,   9,  -4,   1,  -1, -28,  12, -22,  16,   4,  -3,   3,  -1,  22, -12,   9,   2,  -8,   4,  -3,   1,  -5,  -3,   0,  -4,   5,  -1,   1,   0,   7,  -2,   1,   2,  -2,  -2,  -1,   0,   4,  -2,   2,   1,  -1,  -1,   0,  -1 },
     {   -5,   1,  17,  -7,   0,  -1,   0,   0,   0,   0,   0,   0, -26, -42,   1,  -4,  17,  -1,   5,   0,   2,   0,   1,   0, -14, -27,   0,  41, -19,  10,  -4,   3,  -1,   1,  -1,   0,  -9, -13,  36, -29,  12,  -1,   0,   0,   0,   0,   0,   0, -14,  -7,  20,   7,   4,  -6,   3,  -2, -27,  61,   2, -20,  -1,  -1,   0,  -2,  21,  -3, -31,  12,   2,   0,   0,   1,  -2,  -1,  15,   0,  -5,   1,  -1,   0,   6,   5,  -7,  -3,   0,   1,   3,   1,   4,   2,  -3,  -1,   0,   1,   1,   0 },
     {   12,  30,  14,  16, -35,  12, -10,   4,  -5,   1,  -2,   1,   3,  27, -31, -24,  34, -16,   8,  -5,   4,  -2,   1,  -1,  -8,  30,   7,   9, -22,   3,  -3,   0,  -1,  -1,   0,   0, -52, -18, -22,   0,  17,   3,   2,   2,   1,   1,   1,   1,   3,  16,  21,  20, -13,  -1,  -2,  -1,  32,  -7,  -7, -28,   2,   2,  -2,   1, -12,  -9, -10,  18,   6,  -4,   2,   0,   8,   0,   8,  -4,  -9,   3,  -1,   0,  -2,  -5,  -3,   0,   4,   0,   3,   0,  -1,  -2,  -2,   0,   2,  -1,   2,   0 },
     {  -13, -15, -31,   4,  21,  -4,   5,   0,   1,   1,   1,   1,   0,  17,   8,  24, -30,   9,  -7,   2,  -2,   1,  -1,   0, -26,  -5,  -2, -24,  18,   0,   3,   1,   1,   0,   0,   0, -36, -38,   6,  20,   7,  -3,   6,  -1,   3,   0,   1,   0, -17,  -6,  19,  21, -14,   9,  -4,   1,   3,   4,  47, -32,   5,  -3,  -1,   1,  -6,  36, -29,  -8,   7,  -7,   4,  -2,  11, -12,   0,  10,  -4,   2,  -1,   1,   1,   7,   1,  -6,   4,  -1,  -3,   2,   1,   3,   0,  -2,   3,  -1,  -1,   0 },
     {  -10, -29, -14,  36, -33,  20,  -9,   7,  -3,   4,  -1,   2, -11, -33, -38,  20,  23,  -6,   4,   1,   2,   1,   1,   1,   2,  -4,  38, -15,  -5,   0,   2,  -3,   1,  -1,   1,  -1,  11, -41,  13,   6,   6,   1,  -1,   1,   0,   1,   0,   0,  48,   6,  -5,  -7,  -2,  -4,   1,  -1, -35, -19,  10,   6,   3,   2,   1,   1,   1,  20,  13,  -3,  -2,  -1,  -1,   0,  -3,   0,  -3,  -4,   2,   0,   1,   0,   1,   4,   5,   3,  -2,   3,   0,  -3,  -1,   3,   1,   1,  -1,   2,   1,  -2 },
     {   -5,  -8, -26,  25,   0,  -4,   3,   0,   1,   0,   1,   0,   2,  10, -24,   8, -13,  15,  -6,   3,  -3,   1,  -1,   0,  12,  53,  19, -16,   2, -11,  -1,  -3,  -1,  -1,   0,  -1, -25,  -1,   4, -27,   6,   0,   3,  -2,   1,   0,   0,   0, -30,  -1, -31,  24,  16,  -4,   5,  -2, -10,  34,   3,  20, -25,   5,  -4,   1,  37, -35,   5, -20,   3,   6,  -4,   3, -13,   6,  -1,   6,   6,  -5,   3,  -1,  11,  -7,   0,   1,  -3,  -1,   1,   0,   6,  -5,   0,   0,  -3,   0,   1,   0 },
     {   -8, -22, -43,  -1,  16,   3,   4,   2,   3,   1,   1,   0,  13,  12,   8,  20, -16,  -1,  -1,  -3,   0,  -1,   0,  -1,   7,  12,  -1,  -6, -14,  12,  -8,   4,  -3,   1,  -2,   1,  22,  12,   2, -28,  26, -18,   9,  -6,   3,  -3,   1,  -1,  20,  24, -10,   9, -28,   6,  -3,   0, -10,  -6, -34, -26,  25,   0,   1,   1,  -1, -23, -15,  50,  -6,  -1,   1,   1,   2,  -9,  29, -14, -10,   6,  -3,   2,  -8,   5,  -8,  -1,   0,  -4,   6,   2,  -1,   0,  -1,   1,   0,  -2,   3,   0 },
     {   13,  24,  40,  -4, -21,   6,  -5,   0,  -2,   0,  -1,   0,  14,  24,  32, -13,   4, -11,   0,  -3,   0,  -1,  -1,  -1,  11,  17,  30, -41,  13, -10,   4,  -3,   0,  -1,   0,  -1,  18,   1,   8,   6,  -8,  -1,  -1,  -2,   0,  -1,   0,   0,   2,  34, -21, -15,  -2,   0,  -1,   0, -32,  21,   1,  -8,   7,  -4,   1,  -1,  22,  20, -45,   0,  -2,   0,   0,   0,  -1, -31,  -3,  11,   0,   3,  -1,   1,  -1,   8,  -2,  -2,   0,  -9,  -5,   4,   0,   2,  -1,   0,  -1,  -4,  -3,   2 },
     {    6,   1,   6,  32,   4, -21,   4,  -6,   2,  -1,  -1,  -1,  -4, -19, -29, -10, -23,  39, -12,  11,  -3,   3,  -1,   2,  14,   8,   1,  17,  33, -34,  10,  -8,   2,  -2,  -1,  -1,  12,   4,  18,  -3, -33,  16,  -3,  -1,   1,  -2,   0,   0,  30,  13,   5, -11,  13,  -5,  -7,   2,  34,   1,  -5, -22,  -8,  -5,   6,  -3, -18, -12, -33,  13,  14,   2,   1,   1,   5,  -9,  17,  14, -13,  -1,   1,   0,  -4,  -3,   2, -11,  -2,  -1,   2,   4,  -3,   0,   1,  -4,  -1,   0,   0,   1 },
     {    3,   6,  17,   6,  18, -35,  19, -11,   6,  -3,   3,  -2,   5,  13,  41, -34, -17,  29, -17,   6,  -4,   1,  -2,   1,  -5,  -8,  40,  10,  -8, -14,   0,   0,  -2,   0,   0,   0,  -5, -40, -18, -28,  20,   6,   3,   0,   2,   0,   1,   1,  24,   3,   7,  18,  -1, -11,   2,  -3, -27,  -9, -11,  -2,  -9,  10,  -1,   2, -13,   4,  25,   1,  10,  -4,  -2,   1,   2,  26, -12,  -6,  -5,  -1,   2,  -1,  -2,  -4,  -6,  -1,   2,   6,  -3,   1,   1,  -3,  -2,   0,   0,   4,  -2,  -1 },
     {    0,  -7, -21,  -6,   8,   9,   1,   0,   1,   0,   0,   3,  14,  16,  12,  22,   1, -14,   1,  -1,  -3,  -2,   0,  -2,  14,  19,  25,  15, -25,  11,  -8,   2,  -3,  -1,  -2,   0,  13,  19,  20, -22,   9,  -6,  -3,   1,  -1,   1,  -2,   0,  31,   9,  36, -22,   4, -13,   2,  -5,  33,  25,   0, -14, -15,   1,  -1,  -3, -14,  17, -13, -36,   9,  -1,   3,   0,  -1,   3, -49,  13,  12,  -4,   2,  -1,  -2, -19,   4,  11,  -1,   2,  -5,  -3,  -2,  -4,  -1,   4,  -1,  -1,  -3,   0 },
     {    3,  -3,  13, -11,  -7,  13,  -3,   4,  -4,   0,  -1,   1,   6,  -5,  23,   9,   3, -10,   1,  -3,   0,   0,   0,  -1,  -7, -32, -19, -24,  10,   3,   7,   0,   3,   0,   1,   1,  18, -27,  -2,  17,   6,   0,   0,   0,   0,   1,  -1,   0,  38,  11, -14,  32, -14,  -1,  -3,  -1,  36,  13,  12, -20, -17,   8,  -8,   2,  26, -50,  -1, -13,  15,  -6,   4,   1, -35,  19,   1,  10,   0,  -1,   2,  -3,  12, -14,   4,   1, -15,   8,   3,  -1,   2,  -1,  -1,  -2,  -7,   3,   1,   1 },
     {   -7, -12, -37, -33,  24, -13,  14,  -4,   5,   0,   2,   0,  -1,  -4, -23, -28,  19,   5,   7,   1,   1,   0,   0,   0,  -1, -10,  -1,  23, -17,   3,  -3,   1,  -1,   1,   0,   0,   6,  -5, -55,  24,  -4,  14,  -4,   4,  -2,   1,   0,   1,   7,  43,  -3,  -2,  10, -14,   5,  -4, -13,  -3,   2,   7, -16,   8,  -1,   1,  28,   6, -21,  -6,   9,   0,  -2,   1,  -6, -29,  -1,  23,  -5,   2,   1,   1,   2,   0,  20,  -3,  -2,  -1,  -1,   0,   1,  -3,   5,   2,  -1,  -1,   1,   1 },
     {   -7,  -9, -13, -39,  -7,  32,  -7,   9,   4,   3,   0,   3,   3,  10,  22,  21,  10, -20,  -2,  -4,  -6,   2,  -2,   0,  -4,  -3,  15,  -9,   0, -14,  15,  -4,   6,  -3,   3,  -1, -14, -18,  -3, -19, -11,  29, -12,   5,   0,   1,   0,   2,   3, -23, -12,  -2,  38, -24,   8,   0,   1,   1, -11,  16, -27,   7,   5,  -2, -43, -14, -12,  14,  17,   4,  -7,   3,  13,  -4,  30,  16, -11,  -1,   3,   0, -12,  12,   0, -17,   5,   1,  -1,   7,  -3,   3,   4,  -2,   4,   1,  -2,   1 },
     {   -4,  -2,  -5,  -8,  -7,   4,  -2,   1,  -2,   0,  -2,   3,  -8,  -5, -19, -15,   1,  -2,   1,   0,   3,   0,   3,  -1,  -6,  -1, -26, -26,  14,  -6,  13,  -5,   5,  -2,   1,  -2,   0,  15,  11, -10,   9,   1,   1,   2,   1,  -1,   0,  -1, -14,  -2,  -8,   8,  -5,   0,   2,  -3, -32,   5, -27,  -9,  12,  13,  -3,   6, -28,  14,  -2,  16,  29, -18,  10,  -4,   2,  20, -35,  37, -27,   7,   0,  -3,  22, -45,  31,   5,  -8,  17,   4,  -5,   3,  -9,   1,   5,  -2,   7,   4,   1 },
     {    0,  -4,  -6, -18,  22,   4, -13,   7,  -2,   2,  -2,   1,  -5, -12, -15, -50, -12,  -6,  29,  -7,   4,  -1,   2,   1,   3,  -4,  11, -15,  24,  14, -19,   4,  -4,   1,  -2,   1,  -5, -29,  -4,  -9,   6, -22,  13,   0,   1,   1,   3,   1,  25, -20, -30,   3,  -5,  23,   1,  -1,  30,  38, -14,  24,   8, -15,  -1,   1, -22,   7,  -3,  -4, -21,  10,   1,  -3,  19, -16,  -5,   3,  12,  -1,  -4,   3, -11,  -3,   5,   4,   2,   3,   3,  -4,  -4,   0,  -2,   2,   2,   0,   1,   2 },
     {    6,   4,  22,  39, -17,  -6,   5,  -8,   2,  -2,   0,  -2,   1,  -5,  10,  29, -13,   9, -20,   8,  -7,   2,  -4,   3,  -5, -18, -25,  -6, -23,  11,   6,  -3,   3,   3,   2,   0,  -1,  -3, -23, -14,  34,  -9,   1,   2,   2,  -4,   3,   0,   6, -12,   1,  -3,  -4,   4,   9,  -2,  13,   9, -22,  24,  -8,   2,  -4,  -1,   8,   0,   3, -10,   2,  -1,  -4,   0,  28, -49,   2,  27,  -6,   4,   0,  -1, -12,   4,  41, -17,   5,   1,  -7,  -6,  -6,   3,   9,  -2,   3,   0,  -3,  -4 },
     {   -5,  -9, -15, -29, -28,   3,  23,  -9,  13,  -4,   3,   1,   4,   2,   4,   4,  28,   8, -39,  12, -13,   5,  -2,   3,   1,  -2,  -3,  16,   4, -39,  37, -13,   6,  -3,   3,  -4,  -1,  -8,   0,  -9,   1,  19,  -6,  -5,   1,   0,   1,   0,  -1,  -8, -18,  18, -18,   9, -15,  10,  19,  16,  -3,  -3,  22, -19,  17,  -3,   9,  14,   6,   8, -32,   9,  -4,  -1,   3, -18, -25, -19,  12,   6,   2,   3,  -1, -11,   6,  19,  -3,  -1,   5, -10,   0,  -1,  -5,   5,  -2,   1,   1,  -1 },
     {   -2,  -4,  -4,  -4,  -2, -18,  27, -23,  15,  -7,   6,  -6,  -2,  -2,   3, -17,  16, -10, -19,  12,  -6,   3,  -2,   1,   0,   4,  26, -12, -15,  19,  -3,  -4,   0,  -2,  -2,   1,  -8, -15,  17,   0,   5, -22,   8,  -3,   3,   1,   0,  -1, -13, -37,  -5, -40,  10,  10,   2,   5,  25, -33,  -2,   6,  18,   2,  -1,   0,  11,  -7, -29,  14,   7,   0,   7,   1, -44,   6,  12,  23,   4,   4,  -4,  -2,  12,   6,  16,  -2, -12,  13,  -6,  -9,   4,   0,  -3,   8,  -4,   7,   3,  -3 },
     {    4,   3,  11,  15,  23,  15, -32,  17,  -7,   0,  -1,   0,   3,   0,  14,  16, -28, -12,  28, -13,   3,  -2,  -1,  -1,   2,   0,  16,  26,   5,  -5,  -7,  -6,   2,  -1,   5,  -2,  -7, -16, -10, -18, -18,  25,  -9,   9,  -3,   2,  -1,   3,  -7, -16,  -6,   7,  14, -19,   5,  -2,  12, -28, -10,  -9,  20,  -5,   1,   4,  41,  13,  -9,  22, -19,  -2,  -4,  -3, -14, -12, -15,  -9,  -2,   9,  -2,   3,   5, -28,  30,   6,  -3,  18,   7, -13,   0,  -5,  -4,   5,  -5,  10,   2,  -2 },
     {   -4,  -6,  -4, -19, -23,   6, -17,  12, -18,   9,  -9,   6,  -3,  -4,  -6, -16, -17,  21,   7,  -1,  13,  -5,   6,  -3,  -5,  -5, -15, -17,  15, -15,   8,  -5,   0,   1,  -2,   1,   0,   4,  -8, -43,  27,   4,   9,   0,  -2,   1,   2,   0,  -1,   0,  39,   6,   5,  -4,  -9,  -1,   8, -24, -23,  10,  -1,  -2,   5,   0,   7,  23,  -6, -21,   2,   6,   1,   4, -43, -20,  11,   3,  16,  -9,   7,  -4,  29,  21, -17,   9, -12, -16,   1,   7,  -3,  16,   4,  -6,   4,  -2,  -7,   5 },
     {    1,  -3,  -6, -10,   5,   2,   0, -14,   0,  -6,  11,   0,   1,  -2,  -8, -16,  -2, -13,   1,  21,  -6,   7,  -5,   7,   8,   8,   9,  -1,  26,  17,  20, -20,   1,  -8,   1,  -6,   4,   4,  14,   1,   6,   0, -26,  17,   1,   1,  -4,   6,   1, -13,   4,  -2, -31, -29,  24,  -9,  19,   6,  -7,   6,  -2,  33, -23,  -9,  22,  17,  20,  -6,   5, -27,  11,  10,   3, -18,   3, -30, -33,  24,  -2,  -3,   4,   7, -13, -26,  -2,  -2,  -4,  21,  -1,   3,  -3,  -7,  -5,   0,  -3,   3 },
     {   -3,  -2,  -2, -16,  -4,   4,  -1,  -4,   8,   3,   4,  -1,  -1,   2,   4,  -6,   5, -18,   2,  -1,   8,   2,  -5,  -1,   1,   7,  17,  18,  25, -19, -11,  -2, -10,  -5,  -5,   3,  -2,   5,  24,  -2, -31, -19,  11,  -1,   0,   3,   1,   5,  -4,  -3,  27,  29, -23,  25,  -9,   1, -19, -16, -14, -13, -13,  19, -12,   2,   5, -23,  19, -10,  21,  -1,   3,   2,   1, -30,   0,   8,  18,  -4,   4,  -2, -24,  27,  42,  11,   9,  16,  -8, -14,  -6,   4,   5,  -4,   7,   4,  -3, -13 },
   },
   {
     {  109, -49,  -9,  -3,  -2,  -1,  -1,   0,   0,   0,   0,   0, -18,  -7,   8,   1,   1,   0,   1,   0,   0,   0,   0,   0, -35,  17,   7,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  -3,  10,  -5,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,  -2,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    4,   7,  -7,   1,  -2,   1,  -1,   0,   0,   0,   0,   0, 101, -46,  -9,  -1,  -2,   0,  -1,   0,  -1,   0,   0,   0, -35, -15,  18,   1,   3,   0,   1,   0,   1,   0,   0,   0, -37,  22,   8,  -3,   1,  -1,   1,   0,   0,   0,   0,   0,  -2,  11,  -7,  -1,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0 },
     {   -4,  65, -36,   1,  -6,   0,  -2,   0,  -1,   0,  -1,   0, -25, -10,   0,   7,   2,   2,   1,   1,   0,   0,   0,   0, -72,   0,  31,   2,   3,   0,   1,   0,   1,   0,   0,   0,  48,  14, -11,  -8,  -3,  -2,  -1,  -1,  -1,   0,   0,   0,  23, -17, -10,   2,   1,   1,   0,   0,  -2,  -8,   6,   3,   0,   0,   0,   0,   1,   0,   3,  -2,   0,   0,   0,   0,   3,   0,  -1,  -1,   0,   0,   0,   0,   2,  -2,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0 },
     {   46,  47, -40,  -3,  -7,  -1,  -2,  -1,  -1,   0,  -1,   0,  -5,  17, -24,   9,  -3,   2,  -1,   1,   0,   1,   0,   0,  43, -71,  16,   9,   0,   3,   0,   1,   0,   1,   0,   1, -10, -21,  28,  -6,   2,  -1,   1,   0,   0,   0,   0,   0, -14,  18,   5,  -9,   2,  -2,   1,  -1,  -6,   4,  -5,   0,   2,   0,   0,   0,  -3,   1,   1,   2,   0,   0,   0,   0,  -2,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {  -18,  -9,   8,   2,   2,   0,   1,   0,   1,   0,   0,   0, -42,  22,   3,   2,   0,   0,   0,   0,   0,   0,   0,   0, -19, -22,  18,  -1,   3,   0,   1,   1,   1,   0,   0,   0, -65,  47,  18,  -8,   4,  -2,   1,  -1,   1,  -1,   0,  -1,  37,  37, -32,  -6,  -3,  -2,  -1,  -1,  24, -22, -17,  10,  -1,   2,   0,   1,   7, -10,   7,   4,  -2,   1,  -1,   0,   3,  -1,   2,  -2,  -1,   0,   0,   0,   4,   0,  -1,   0,   2,  -2,  -1,   0,   3,  -1,   0,   0,   2,  -1,   0,   0 },
     {   -2,   0,  -4,   6,  -3,   2,  -1,   1,   0,   0,   0,   0, -23, -68,  41,   2,   6,   1,   2,   1,   1,   1,   1,   0,  51, -12,  32, -23,   3,  -5,   0,  -3,   0,  -1,   0,  -1,  14,  43, -26, -14,   4,  -4,   1,  -2,   0,  -1,   0,  -1, -30,   7, -25,  13,   3,   1,   1,   1,  -9,  -1,   3,  11,  -5,   1,  -1,   1,   2,   5,   5,  -3,  -4,   1,  -1,   0,   1,   3,  -3,  -3,   1,   1,   0,   0,  -1,   0,  -1,   1,  -1,   2,   0,   0,  -1,   1,  -1,   0,   0,   1,  -1,   0 },
     {  -14, -21, -72,  53,  -8,  12,  -1,   6,   0,   2,   0,   2,   8,  21,  -3,  16, -13,   2,  -5,   0,  -2,   0,  -1,   0,  12,  41,  36, -40,  -2,  -5,  -1,  -2,  -1,  -1,  -1,  -1, -20, -23,  -1, -10,  14,  -1,   4,   0,   2,   0,   1,   0,  -1,  -8,   0,  11,   3,  -1,   1,  -1,  17,   2,  -2,  -1,  -4,   0,  -1,   0,  -2,  -3,   1,  -2,   0,   1,   0,   0,   0,  -2,   1,   0,   1,   0,   0,   0,   1,  -1,   1,   0,   1,  -1,   0,  -1,   0,  -1,   1,   0,   1,  -1,   1,   0 },
     {  -14,   4,   0,  -1,   1,   0,   1,   0,   1,   0,   0,   0, -25, -34,  21,   3,   4,   1,   1,   0,   1,   0,   1,   0, -33,  10,   7,   4,  -1,   1,   0,   1,   0,   0,   0,   0,  -5, -31,  26,  -1,   0,   2,   0,   1,   0,   0,   0,   0, -44,  46,  29, -20,   5,  -2,   1,  -2,  22,  45, -39, -11,   4,  -4,   0,  -1,  14, -16, -23,  16,   1,   0,   1,   1,   6,  -7,   7,   6,  -4,   1,  -1,   0,   3,   0,   3,  -3,   4,   0,  -2,   0,   2,   0,   0,   0,   2,   1,  -1,   0 },
     {   21,  62, -11, -16,  -5,  -4,  -2,  -2,  -1,  -1,  -1,  -1,  -9, -10,   9,  -8,   8,  -2,   2,  -1,   1,   0,   0,   0,  15,  41, -33,   4,   0,  -2,   0,  -1,  -1,  -1,   0,   0, -57,   9, -32,  23,  -3,   4,   0,   2,   0,   1,   1,   1, -14, -16,  15,  16,  -8,   4,  -2,   1,  40,  -7,  11, -10,  -5,   1,  -3,   0,   6, -14,   0,  -6,   3,   0,   1,   1,  -2,  -3,   1,   1,   2,  -1,   1,   0,   1,  -1,   1,   1,   1,  -2,   0,   0,   0,  -3,   1,   1,   1,  -1,   1,   0 },
     {   13,  15,   3,  -2,  -4,   0,  -1,   0,  -1,   0,  -1,   0,  18,   4, -27,  10,  -3,   2,  -1,   1,   0,   1,   0,   0,  18,  42, -24,   2,  -7,  -2,  -2,  -2,  -1,   0,  -1,   0,  40,  11,  14, -25,   1,  -5,   0,  -2,   0,  -1,   0,  -1, -15,  28, -29, -12,  11,  -3,   3,  -1,  20, -37, -41,  29,   1,   1,   0,   1,  -4, -36,  25,  18,  -9,   3,  -1,   1,  -2,   1,  16, -12,  -5,   3,  -1,   0,  -6,  -3,  -1,  -4,  -1,  -1,   0,   2,  -3,  -2,   1,   1,   0,  -2,   1,   0 },
     {   -5,  -4,  13,  -7,   4,  -2,   1,  -1,   0,  -1,   0,   0,  -6,  25, -54,  20,  -3,   3,  -1,   2,   0,   0,   0,   0,  -7,   8,  20,  13,  -7,   0,  -2,   0,  -1,   0,   0,  -1,  -8,  45,   4, -25,  -2,  -3,   0,  -2,  -1,  -1,   0,  -1, -60, -34,  -6,   5,  12,   4,   3,   2,   2,  29,  15,   5,  -2,  -3,  -2,  -2,  39,   5, -19,  -6,  -5,   0,  -1,  -1,   0, -18,  -1,   3,   3,   2,   0,   1,   2,  -3,   2,   1,   2,  -2,   2,   0,   3,  -1,   1,   0,   1,  -1,   0,   0 },
     {   -2,  -5,   3,  22, -16,   4,  -5,   1,  -1,   0,   0,   0, -25, -61, -33,  42,  -4,  13,   0,   5,   1,   2,   1,   1,  12,  12, -20,  16,  -4,   2,  -2,   0,  -1,   0,   0,   0, -12,   7,  47, -17,  -7,  -1,  -3,  -1,  -1,   0,  -1,   0,  31, -33,  28, -22,   4,   1,   1,   0,   6, -10,   9,  -9,   9,  -4,   1,   0, -16,  14,   3,  -2,   3,  -4,   2,  -1,  -3,  11,  -3,  -2,  -1,  -1,   1,  -1,   1,   3,  -2,  -1,  -1,   3,  -1,  -1,  -1,   2,  -1,  -1,   1,   1,   0,   0 },
     {   11,  16,  36,  53, -48,   7, -14,   1,  -5,   0,  -3,   0,  -1,   9,  -4,  -6, -17,  10,  -3,   3,  -1,   1,   0,   0, -16, -22, -36, -50,  44,  -1,   9,   1,   4,   1,   1,   1,   8,  16,   6,   5,  19, -13,   2,  -4,   0,  -1,  -1,  -1, -14,   4,  15,  18, -12,  -5,   0,  -1,  -3,  -1,  -6,  -4,  -7,   4,   0,   1,   9,  -1,  -8,  -2,   2,   3,  -1,   1,  -1,  -5,   4,   2,   1,  -1,   0,   0,   0,  -1,   2,   0,   1,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   1 },
     {  -13, -11, -15,   7,  -1,   6,   2,   1,   1,   0,   1,   1,  -9,  -3,  11,   8,  -2,   1,  -1,   0,   0,   0,   0,   0, -28,  -8, -24,  18,   0,   3,   0,   1,   0,   2,   0,   1,  -6, -22,  16,   7,   0,   1,   2,   0,  -1,   0,   0,   0, -54,  22,  -4,   8,   0,   1,   0,   1,   3, -44,  36,  19,  -8,   4,  -3,   2,  16,  29,  44, -30,  -5,  -2,  -2,  -1,  13,  16, -26, -22,   8,  -1,   0,   0,  -3,  -5,  -8,   9,   4,   4,   3,   2,   1,   1,   0,  -1,   3,   1,   0,  -1 },
     {  -10, -24, -50,  26,   4,   6,   2,   4,   0,   1,   1,   1,   4,   5,   0, -21,  12,  -7,   2,  -2,   1,  -1,   2,  -1,   5, -15, -23,  24,   4,  -2,  -1,   0,   1,   0,   1,   0,  23,  52,  -2,  39, -32,   2,  -7,  -1,  -2,  -1,  -1,  -1, -12,   5,  20, -28, -20,  10,  -2,   1,   4, -19,  -4, -22,  19,   7,   0,   2,  20, -12, -10,  10,  14,  -7,  -2,   0,  -7,  -7,  10,   8,  -5,  -4,   1,   0,  -3,   4,   5,  -4,  -2,   0,  -1,  -1,   0,   0,   1,  -1,  -1,  -1,   0,   1 },
     {    0, -13,  17,   4,  10,  -6,   1,  -3,   0,  -1,   0,   0,  -5, -12, -21, -36,  23,  -4,   7,   1,   3,   1,   1,   0,  -2, -20,  30,  -9,  16,  -7,   3,  -1,   2,  -1,   0,  -1,  15, -21,  24,  21, -13,  -2,  -2,  -1,  -1,  -1,   1,   0, -10, -34, -18,  16, -12,  11,   1,   2,  71,  -1,  -4,   2,  -6,   4,  -4,   1, -12,  -9,  16, -10,   2,  -2,   0,   1, -35,  13,   8,  -5,   7,  -2,   2,  -1,  -4,  12,  -5,   2,   1,   4,  -3,   1,  -2,   2,   0,   1,  -1,   3,   0,   0 },
     {   -6, -22,   1,  -8,   5,  -2,   4,   1,   1,   1,   1,   0,   2,  -6, -12,  48, -22,   7,  -5,   1,  -3,   1,  -1,   0,  -7, -37,  -5,   8, -17,  12,  -3,   4,  -1,   1,   0,   1,  16,   0, -52, -14,  19,   2,   3,   3,   2,   1,   0,   1,   6,  20,  33,  22,   1,  -7,  -4,  -3,  40,  -9,  19, -15, -17,   0,  -3,   0,  -1, -40,  -2,  -6,   7,   5,   2,   3, -14,  10,  16,   0,   5,  -3,   0,  -1,  -2,   9,  -1,  -5,  -2,  -1,   0,   1,   0,   1,   1,  -1,  -1,   1,   2,  -2 },
     {   -7, -13,  -5, -39, -62,  48,  -6,  10,  -2,   4,   1,   3,   5,   9,  11,   6,  -2,  18, -11,   3,  -4,   0,  -2,   0,   6,   4,  16,  39,  47, -42,  -1,  -6,  -3,  -2,  -1,  -2,   0,  -8, -14, -14,  -8, -15,  13,  -1,   3,   1,   1,   1,   1,  -1,  -1, -11, -13,  14,   4,  -2,  11,   2,   3,   5,   3,   2,  -4,   0,  -7,  -1,  -5,   2,   1,  -4,  -1,   1,  -1,   2,  -2,  -1,  -2,  -2,   2,   0,   1,  -2,  -1,  -2,  -1,  -1,   1,   0,  -1,   0,   0,  -1,   0,  -1,  -1,   1 },
     {   11,  11,  10,   7,  -7,  -2,   0,  -1,  -1,  -1,  -1,  -1,  20,  21,  47,   1, -13,  -8,  -5,  -3,  -3,  -2,  -1,  -1,  10,   4,  -3,   5, -19,  10,  -5,   2,  -1,   1,  -1,   1,  23,  15,  23, -33,  -1,   4,  -2,   0,   0,  -1,  -1,   0,   8,   7,  -4, -19,  12,  -5,  -1,   0,  52,  -6,   9, -15,   3,  -3,  -1,   0,  -1,  26, -32, -29,  11,  -5,   2,  -2, -18, -14, -39,  20,  11,  -2,   3,   1,  -8,  -8,   6,  18,   0,   0,   2,  -4,  -2,  -1,  -2,  -1,  -1,  -2,  -3,   1 },
     {    0,   3,   8, -24,  12,   0,   4,  -1,   0,   1,  -1,   0,  -3,   0, -45,  10,  11,  -4,  -2,   2,   2,   0,   1,   0,   6,  23,  25, -27,   7, -14,   2,  -2,   1,  -1,   0,  -1,  14,  16,   5,  20,  -2,  -1,   8,  -5,   1,  -3,  -1,  -1,  -6,  46,  14,  14, -30,   5,  -6,   0,  -3, -14,   0, -38,  -3,  10,  -3,   2, -33,  18,  -9, -12,  21,  -1,   0,  -1,  20,  14, -29,  13,   2,  -7,   0,  -1,   4, -25,  -8,  17,   3,  -7,   8,  -2,   1,  -3,  -1,   0,   1,   2,   0,   0 },
     {    0,   8,  18,  -5,   4,  -8,   0,   0,   0,   1,  -1,  -1,   5,  29,  31,   5, -14,  -2,  -4,  -2,  -3,   1,  -1,   0,  -7,  -3,  28, -12,  -4,   2,   2,  -1,  -1,   0,   0,   0,  -9,  17,  -2, -10,   3,   3,  -1,   0,   1,  -2,   0,   0, -33, -32,  40, -25,   9,   1,  -1,   2,   0, -28, -18, -13,  21,  -3,   4,   0, -26,   8,  16,  27,   4,  -4,   2,  -2,   0,  54,  10, -10, -13,  -4,  -1,  -4,   9,   2, -26,  -8,   0,  -4,  -5,   4,   4,   0,  -4,   1,   0,   1,  -3,   1 },
     {   -8, -12, -36, -56,  18,  -1,   9,   4,   4,   2,   2,   0,   2,   0,  15,  14,  16,  -8,   0,  -4,   0,  -1,   0,   0,  -5, -10, -33, -31,  12, -10,   9,   0,   4,   1,   1,   1,   4,  14,  30, -19,  21,  -9,   9,  -3,   2,  -3,  -1,  -1,   5, -23,  25,  35, -11,  -3,  -3,  -2,   0, -10, -35,  13, -13,   8,   0,   4,  16,  25,  -8,   4,  -3,   4,  -4,  -1, -19,  -6,   6,  -2,   1,   1,  -2,   1,  -3,   3,   8,   1,   3,   1,  -4,  -2,   1,   1,  -2,   3,  -2,   1,  -1,   0 },
     {    7,   9,  22, -11,  -5, -13,  10,   0,   0,  -2,  -1,  -1,   7,   1,  20,  33, -30,  11,  -5,   2,   0,   0,   0,   0,  11,  16,  36,  -4,  -2,   4, -13,   1,  -2,   2,   0,  -1,   4, -12,  21,  33,  -4, -11,  -1,   0,  -2,  -2,  -1,   0,  11,  -4,  -2,  -5, -34,   1,   4,   1, -14, -14,  -5, -20,   1,  14,  -1,  -1,  63,  -8,  20,  -7,  12,   0,  -6,  -2, -15, -12,  13, -10,   5,  -9,   0,   1, -27,  10,   4,   0,   5,  10,  -6,  -6,   6,   4,  -6,  -2,  -3,   1,  -3,  -3 },
     {    4,   3,  -3,   3,  -3,   4,  -4,   3,  -2,   1,  -1,   1,   6,   8,   2, -17,   7,   3,  -3,   1,  -1,   1,  -1,   0,  11,  11,   8,   2,  -4,  -8,   1,  -7,   0,   0,   1,  -1,  10,  25,  13,   7, -14,   1,  -1,  -1,  -1,   2,   0,  -1,  26,  16,  42,   3,  10,  -9,  -3,  -3,  -7,  37,  -3,  16, -24, -13,  -3,  -1,  24, -19,  13, -32, -26,  11,  -1,   3, -14,  27, -18, -44,  18,   3,   4,   0,  16,  -2, -39,  10,  -5, -10,  -4,  16,   1,  -9,   3,  -1,   0,  -6,   3,   1 },
     {   -2,   5,  -9,  -7,   9,  10,  -7,   5,  -4,   2,  -1,   2,  -3,   3,  10, -26, -20,  19,   1,   5,  -1,   4,   0,   0,   1,  18,  -1,  23, -11,   6,  -3,  -3,  -2,  -1,   0,   0,   5,  16,  39,   3,  36, -31,  -2,  -3,  -1,  -2,   0,  -2, -12,   9, -13,  12,  -8, -15,   7,   1, -13,   4,  33, -17, -29,  10,   2,   0, -29,  -9, -11,  -6,  15,   7,   0,   0,  -7,   8,  27,  12,   2,  -7,   0,   1,  17,  42,  -8, -25,   5, -10, -22,   4,   1,  -2,   3,   4,   3,  -1,  -2,   1 },
     {    1,   6,  10,  14, -18, -20,  15,  -4,   6,  -2,   0,   1,   4,  11,  12,  20,  44, -30,  -3, -11,   2,  -2,   2,  -1,  -2,   8,   5,  -2,   8,  -3, -11,   0,   1,  -3,   1,   0,  -5,  -8,  14, -36, -48,  28,   5,   5,   3,   2,   1,   0,  -4,  16,   5,  30,  -9,  17,  -5,  -3, -23, -12,  20,   6,   2,   5, -13,   1,  -5, -17, -17, -12,  13,   0,  -1,   4, -12,   6,  21,   8,   0,  -6,   3,   0,  12,  32,  -2, -13,   5,  -3, -18,  -4,   1,  -3,   3,   2,   2,  -3,   0,  -2 },
     {    6,   8,  16,  12,  31,  52, -30, -11,  -8,  -4,  -2,  -2,  -3,  -5,  -2, -12,   0, -10,  -9,   6,  -1,   2,  -1,   3,  -2,  -3,   9,  -5, -29, -54,  27,   4,   8,   1,  -2,   1,  -2,  -8,   4,  -3,  10,  24,  16, -12,  -1,   0,  -2,   1,   2,  -1,  14,  -6,  18,  19, -15, -11, -12, -29,   4,  -3, -16,  -1, -10,  12,  22, -11, -11,   1,   0,   1,   3,   7,  -4,   0,   4,  14,   7,   2,   5,  -4,  -8,   8,  11,   0,  -2,  21,  -6,  -6,  -3,  -5,  -9,  -2,   0,  -8,  -1,   5 },
     {    2,   6,   1,  10,  10,  20,  -9,  -9,  -4,   3,   0,  -1,   5,  13,  22,  52, -12, -10,  -6,  -2,   3,  -4,  -1,  -2,  -6,  -8, -22,   1, -21, -24,  20,  12,   6,   2,  -2,   1,   3,  15,   7,  25, -23,   4,   6,  -5,  -4,   1,  -2,   1,  -8,  -4, -31,  13, -11,  18,  -7,   4,   4,  46, -11, -16,   1,   1,  -3,  -1, -38,  10,  32,  -3,  -6,  -4,  -1,  -2, -10, -16,  14, -12,   2,   4,   0,   2,  11,  -3, -11,   1,   7,  -9,   1,   4,   3,   5,  -4,   0,   1,   3,  -2,  -3 },
     {   -7,  -5, -10, -26, -17,  14,  -2,   1,   9,   2,   2,   2,  -8,  -6,  -9,  -6,  -9,  13,  -4,   0,  -7,  -2,  -4,   3,  -9,  -5, -17, -30, -11,  12,  -1,   3,   2,   0,   0,   2,  -9,  -6,  -3,  -5,  -4,  11, -11,   4,   2,   5,   5,  -2, -11, -11, -25, -15,  16,   4,  -1,  -4, -27,   4, -15, -17,  22,   7,   6,   3,   4, -39,  17, -16,  33, -10,   1,   4, -35,  10, -34,  21,  26, -10,   6,  -1,  24,  20, -10,  25,  -1,   1,   4,   4,  -7,  -7,   8,   4,   4,  -5,   4,   5 },
     {   -3,  -7,   1,   3, -34,   2, -15,   5,  -3,  -6,   2,   1,   1,   2,   5,  -4,   4,  10,   1,   4,  -8,   6,  -4,   1,   1,  -2,  11,  31, -24,  16,  -9,  -7,  10,  -2,   2,  -1,   1,   1,   9,  12, -14,  23, -16,  -1,  -4,   2,  -1,  -1,  -2, -10,  -4,  54,  20, -21,  -1,  -2,  -5, -12, -39, -22, -30,  -3,   4,  10,   3,   8,   6,   3,  20,   3,   5,   1,   5,   1,   0,   6, -14,   2,   2,   0,  20, -36,   5, -18,  -5,  24,   7,  -2,  -9,  -4,  -5,  -4,  -3, -15,   5,  14 },
     {   -3,   0,  -2, -21, -37,  11,  21,   5,   5,   0,   2,   0,  -3,   2, -15,  -4,   0,  -7, -14, -10,  -1,  -4,   2,   2,  -2,   4,   3, -21, -14,  -4, -21,   1,  -5,   3,   1,   2,   6,  17,   4,  38,  12,  27,   3,   8,  -4,  -3,  -1,  -4,   0,  18,   4,   5,  45, -10,   0,  -1,   7,   6,  11,  -2,  22, -27,  -2, -16, -12,  22,   8,  -2, -19,  -6,   1,  -6,  -8,   3,  15,  -5,   0,   7,   1,   4, -23,  30,  25,  -2,   2,  21, -23, -13,  12,  -1, -14,   3,   4,  -4,  -7,  -1 },
     {    3,   0,  -3,  -4,  -6,   4,  -5,   8,   0,  -3,  -4,   1,   3,  -6, -11,   0,  11,   7,  -1,   3,  -3,   4,  -2,   3,   1,  -5, -18, -15, -14,  -5,  -5,  -4,  -3,   0,   3,   0,   5, -12, -12,  -8,   3,   3,  -5,   3,  -1,  -2,   3,  -2,  13,  10, -31,   3,  10,  20,  -5,   3,  18,   6,   2, -27,   8,   5,  -5,   4,  38,  37, -16,  23,  15, -15,  -4,  -1,  16,  17,   1, -15, -12,  -7,   1,  -8, -14,  14, -52, -28,  -1, -17, -21,  35,  -3,  -4,  12,  -3,   0, -14,  -2,   7 },
   }
 },
 {  //13
   {
     {  118, -42,  -7,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0,  -7,  -4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0, -16,   2,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -9,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -6,   3,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {   -3,  -9,   5,  -1,   1,   0,   1,   0,   0,   0,   0,   0,-114,  34,   8,   4,   3,   2,   1,   1,   1,   1,   1,   0,  14,  18,  -6,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  32,  -4,  -6,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0,  14,  -7,  -2,   0,   0,   0,   0,   0,   4,  -5,   1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   5,  -2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   2,  -1,   0,   0,   1,  -1,   0,   0,   2,  -1,   0,   0 },
     {   13,  -6,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  11,  18,  -9,   0,  -3,   0,  -1,   0,  -1,   0,  -1,   0, 108, -23, -13,  -2,  -4,  -1,  -2,   0,  -1,   0,  -1,   0,   2, -36,   5,   1,   2,   0,   1,   0,   0,   0,   0,   0, -33,  -7,  11,   1,   2,   0,   1,   0, -18,   9,   5,  -1,   1,   0,   0,   0,  -6,   6,  -1,   0,   0,   0,   0,   0,  -5,   1,   0,   0,   0,   0,   0,   0,  -4,   2,   1,   0,  -3,   1,   0,   0,  -2,   1,   0,   0,  -2,   1,   0,   0 },
     {   13,  20,  -9,  -1,  -3,   0,  -1,   0,  -1,   0,   0,   0,  28,  -7,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -17,  29,  -7,   1,  -2,   1,  -1,   0,  -1,   0,   0,   0,  97, -21, -18,  -1,  -4,   0,  -2,   0,  -1,   0,  -1,   0,  -5, -46,   7,   2,   2,   1,   1,   1, -27,  -5,  15,   0,   2,   0,   1,   0, -13,   8,   6,  -2,   1,  -1,   0,   0,  -3,   6,  -1,  -1,   0,   0,   0,   0,  -3,   0,  -1,   0,  -3,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   1,   0 },
     {   15,  18,  -9,  -3,  -2,  -1,  -1,  -1,   0,   0,   0,   0,  16,  24, -12,   0,  -3,  -1,  -1,   0,  -1,   0,   0,   0,  29, -13,  -5,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,  -6,  29,  -8,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  81, -13, -26,   1,  -5,   1,  -2,   0,   1, -60,   4,   5,   1,   2,   1,   1, -29, -12,  21,   1,   3,   0,   1,   0, -12,   9,   7,  -3,   1,  -1,   1,   0,  -3,   4,  -1,  -1,  -4,  -2,   0,   1,  -3,   0,   2,   0,  -3,   0,   1,  -1 },
     {   32, 101, -37,  -4, -10,  -2,  -5,  -2,  -3,  -1,  -2,  -1, -13,  23, -12,   0,  -1,   0,  -1,   0,   0,   0,   0,   0, -13, -16,  -3,   3,   1,   1,   1,   0,   0,   0,   0,   0,  -8, -22,   9,   1,   2,   0,   1,   0,   0,   0,   0,   0, -17,  15,   7,  -1,   1,   0,   0,   0,  26,   4,  -9,  -1,  -2,   0,  -1,  -1,   6, -13,  -3,   1,   0,   0,   0,   0,   4,  -8,   1,   1,   0,   0,   0,   0,   0,  -4,   2,   0,   2,  -2,   0,   0,  -1,  -3,   1,   0,   1,  -2,   1,   0 },
     {    0,  18,  -6,   0,  -2,   0,  -1,   0,   0,   0,   0,   0, -19,   1,  -3,   3,   0,   1,   0,   1,   0,   0,   0,   0, -19, -44,  12,   4,   4,   2,   2,   1,   1,   1,   1,   0, -32,   4,   8,   2,   2,   1,   1,   0,   0,   0,   0,   0,  13, -39,   9,   2,   1,   0,   1,   1, -66,  11,  35,  -3,   5,  -1,   2,  -1,   1,  58,  -1,  -8,  -1,  -2,  -1,  -2,  16,  13, -19,  -2,  -1,   0,  -1,   0,  11,  -3,  -9,   2,   3,  -4,  -1,   2,   4,   1,   0,   0,   2,   0,   0,   0 },
     {   -4,   6, -13,   4,   0,   3,  -1,   1,   0,   0,   0,   0, -27, -77,  19,   8,   7,   3,   2,   2,   2,   1,   1,   1,   6, -21,   9,   2,   2,   1,   0,   0,   0,   0,   0,   0,   9, -18,  15,  -2,   1,   0,   1,  -1,   0,   0,   0,   0, -19,  25,   0,  -1,  -1,   0,   1,   0,  14, -35,   3,   5,   0,   2,   0,   1, -48,   8,  32,  -1,   4,  -1,   1,   0,   2,  38,  -1,  -7,   0,  -2,   0,  -1,   1,  11,  -9,  -2,   6,   2,  -5,   2,  -2,   1,   2,   1,   3,   4,  -1,  -2 },
     {  -12, -25,   7,   0,   3,   0,   2,   0,   1,   0,   1,   0,  11,  63, -16,  -4,  -4,  -2,  -1,  -1,  -1,   0,  -1,  -1, -36,  13, -20,   5,  -1,   3,   0,   1,   0,   1,   0,   0, -25, -57,   5,   9,   3,   3,   2,   2,   1,   1,   1,   1, -21,  -5,  13,   4,   1,   1,   1,   0,  15, -27,  11,   2,   0,   1,   0,   0, -28,   3,  29,  -3,   4,  -1,   1,  -1,   3,  37,   2,  -8,   0,  -2,   0,  -1,   9,   7, -13,  -2,   4,  -3,  -4,   2,   2,  -2,   1,   1,   3,   1,   0,   0 },
     {    5,  19, -11,  -2,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   6,  -5,  18,  -4,   1,  -4,   0,  -2,   0,  -1,   0,   0,  22,  77, -12,  -8,  -4,  -4,  -1,  -2,  -1,  -1,  -1,  -1, -30,  24, -18,  -3,   0,   0,   1,   0,   0,   0,   0,   0, -26,   1, -20,   6,  -1,   2,   1,   1,   3, -26,  10,   6,   1,   1,   1,   1,  -2,  38,   6,  -5,  -1,  -2,  -1,  -1,  37,  -1, -30,  -1,  -3,   0,  -1,   0,   2, -30,  -5,   7,   3, -12,   5,   4,  -3,  -4,   6,   0,   3,  -1,   0,  -1 },
     {   -1, -16,  -1,   1,   1,   1,   2,   1,   1,   0,   0,   0,  13,  14,  -5,  -2,  -2,   0,  -2,   0,   0,   0,   0,   0,  -5, -20,  -5,   5,   0,   2,  -1,   1,   0,   0,   0,   0,  35, -26,  29,  -1,   0,  -1,   0,  -1,   0,  -1,   0,   0,  36,  56,  -1, -12,  -3,  -4,  -2,  -2,  14,   0,  -7,  -8,   1,  -1,   0,  -1, -18,  44, -21,  -6,   0,  -2,   0,  -1,  34, -23, -38,   9,  -4,   2,   0,   1,  -9, -35,   7,  10,  -3,  -4,   8,   2,  -7,  -2,   5,   0,  -1,   0,   0,  -1 },
     {    8,  15,  37, -12,   0,  -5,  -1,  -3,  -1,  -1,   0,  -1,   8,  31,  13,  -8,  -3,  -2,  -2,  -2,  -2,  -1,   0,  -1,   6,  -9,  24,  -7,   1,  -3,   0,  -1,   0,   0,   0,  -1,  25,  41,   0, -11,  -3,  -3,  -2,  -2,  -1,  -1,  -1,  -1, -14,  32,  -6, -10,   0,  -2,   0,  -1, -12,  23, -20,  -4,   0,   0,   0,   0, -10,  -5,   3,  -1,   1,   0,   1,   0,  -5,  51, -12, -12,   0,  -3,   0,  -1,  28,  -2, -47,   2,   2, -32, -10,  13,   3, -13,   4,   6,  -2,  -4,   3,   0 },
     {   13,  29,  59,  -6,  -7,  -9,  -2,  -4,  -2,  -1,  -1,  -2,   1, -23,  42,  -5,   0,  -2,   2,  -1,  -2,  -1,   0,   0,   4,  33, -27,  -6,  -4,  -1,  -1,   0,  -1,   0,   0,   0, -24, -48,   0,  10,   3,   1,   1,   2,   1,   0,   1,   1,  49,   3,  15,  -8,   1,  -3,  -2,   0, -17,  26, -13, -10,   1,   0,  -1,  -1, -14,  -3, -10,   0,   0,   2,   0,   0,  -8,   6,   4,  -1,  -1,   1,   0,   0,   2,  10,   1,  -1,  -1,   2,  -3,  -3,   4,  -2,  -5,  -3,  -2,  -1,  -5,   2 },
     {    5,  16,  60, -17,   2,  -7,  -2,  -3,   0,  -2,   0,  -1,   2,  22,  34, -10,   0,  -4,  -2,  -2,   0,   0,   0,  -1, -11, -28,   8,  -1,   4,   0,   2,   1,   1,   0,   0,   1,  10,   9,  -2,  -4,   3,  -2,   0,  -1,   1,  -1,   0,   0, -40,  11, -29,   6,  -1,   2,   0,   1, -18, -42,   5,  14,   0,   3,   1,   2,  -6,   7,   5,   4,  -2,  -1,   0,   0,  11, -31,   9,   6,   0,   2,   0,   2, -23,  11,  31,  -6,   6,  30,  -4, -11,  -1,   9,  -9,  -1,   5,   1,  -4,   0 },
     {    8,   3,  48, -13,   0,  -5,  -2,  -3,  -1,  -1,   0,   0,  -1, -17,  -1,   0,  -1,   3,  -1,   2,   1,   0,   0,   1,   3, -36,  -3,   2,   1,   2,   2,   1,   1,   0,   0,   0,   2,  -3, -28,   8,   0,   1,   0,   1,   0,   0,   0,   0,   1, -42,  11,  16,  -1,   2,   0,   1,  66,  -2,  19,  -1,  -4,  -2,  -1,  -1,  23,   1,   7, -10,  -2,  -3,  -2,  -1,   6,   9, -19,  -7,   1,  -1,  -1,  -1,   2, -39, -11,  11, -16, -12,  25,   6,  -8,  12,  12,  -4,  -4,   8,   2,  -4 },
     {   -3,  -7, -34,   7,   2,   4,   0,   4,   2,   1,   1,   1,  10,  26,  77,   0,  -6,  -5,  -8,  -2,  -1,  -3,  -2,  -1,   3,  -2,  63,  -7,  -2,  -6,  -4,  -3,  -1,  -2,  -2,   0,  -6, -13,  -9, -10,  -3,   1,   2,  -1,   1,   1,   1,   0,   6, -27,   6,  -5,  -1,   2,   3,  -2,  29,   9,   3, -12,   0,  -1,   2,  -1, -23,   5, -19,  -3,   4,  -3,   1,   1, -16,  -2, -10,   8,   6,   1,   2,   1, -12,   4,  -2,   4,   3,   7,   0,   1,  -3,   6,  -4,   0,   4,  -3,   0,   0 },
     {    6,   5,  14,  -1,   0,  -1,  -2,  -2,  -1,  -1,   0,   0,   6,  -1, -12,   3,   0,  -1,   0,   0,   1,   0,   0,   0,  10,  10,  -2,   6,  -3,   0,  -2,   0,  -1,   0,  -1,   0,   8,  16,  18,  -1,  -4,  -1,  -2,  -1,  -1,   0,   0,   0,   8,  -2,  15,  -1,  -5,  -1,  -2,  -1,  46,   3,  34, -11,  -3,  -3,  -1,  -2,  19,  39,  -2, -23,  -2,  -4,  -2,  -2,  -4,  -1,  -5, -13,   2,  -2,   1,   0, -15,  47,  -6, -18,  17,  16, -55,  -5,   2, -33, -22,  18,   0, -15,   8,   8 },
     {    1,   9,  17,  -1,   2,  -3,  -1,  -1,   0,   0,  -1,   0,  -6,   4, -38,   8,   2,   2,   1,   1,   0,   1,   0,   0,   3,  16,  13,  -2,  -1,  -2,   0,  -2,   0,   0,   0,   0,  -7,  31,   9,  -8,  -4,  -1,  -1,  -1,  -1,   1,  -1,   0,  -8,  -9,  28,  -6,   2,  -1,   1,   1,   8,  17,   0, -15,   1,  -2,  -1,  -2, -56,  -3,  -8,  -2,   6,   0,   2,   0, -10,  17, -12,   9,   0,   1,   0,   0,   2, -27,  23,   8,  -9,  49,  20, -21,  14,  13, -43,  -8,   0, -16,  -8,  10 },
     {    7,   5,  20,  54, -12,  -2,  -4,   0,  -5,   0,  -1,  -1,   6,   5,   3,  44, -10,   0,  -2,   1,  -4,   1,  -1,   0,   6,  15,  39,  11, -16,  -1,  -5,   0,  -1,   0,   0,  -2,   3, -10,  44,  -8, -12,   1,  -3,   0,  -2,   0,  -1,   1,   5,  -5,   6, -20,  -4,  -1,  -2,  -1,  -9, -18,  -7,   1,  -1,   3,  -1,   2,  45, -10,  31,  -5,  -3,   2,  -1,   0,  20,   9,   5, -20,  -4,  -3,  -2,  -1,  -1, -16,  -2, -11, -15,  10,   8,  -8,   4,   9,  -6,  -9,  -3,   6,  -7,  -3 },
     {    2,  -2, -10,  74, -17,   4,  -6,   0,  -3,   0,  -1,  -1,   4,  11,  39,  14, -10,  -4,   0,  -3,  -1,  -1,   0,   0,  -7, -26, -67, -13,  -4,   3,   5,   3,   3,   2,   2,   1,   7,  29, -28,  -1,   1,  -3,   2,   1,   0,   0,   1,   0,  -8,  14,  20,   6,   2,  -3,  -2,  -1,   2,   3,   9,  -8,  -1,   1,  -3,   0,  -6,   1,   4,  -8,   1,   2,   0,   0,  -6,   2,  -3,  -3,   2,   0,   1,  -1,   8,  -2,   5,   2,   1,   5,   3,  -4,   5,  -1, -10,  -2,  -1,  -6,  -1,   3 },
     {    2,   7,  29,  29,  -5,  -4,  -3,  -1,   1,  -1,  -1,  -2,  -4,  -1, -21,  27,  -6,   1,   0,   2,   1,  -1,   0,   0,  -4,  11,   0,   6,  -7,   1,  -1,   1,  -1,  -1,  -1,   1,  -7,  12,  26, -13,  -3,  -1,  -1,   1,   0,  -1,  -1,  -1, -17, -18,   4, -11,   2,   0,   2,   0,   8, -13,   6, -17,   7,  -2,   1,   0, -39,   8, -34,   0,   7,   1,   3,   1, -29, -31,   4,  21,   3,   4,   1,   1,   0,  16,  -3,   5,  16, -37,   4,  18,  -9,   1,  47,  -5,   1,  27,  -1, -16 },
     {   -3,   1, -14, -42,  12,   3,   4,   0,   2,   0,   1,   1,   4,   8,  31,  -9,   3,  -2,  -4,  -1,   0,   0,  -2,   0,   3,  10, -26,  32,   2,  -1,  -1,   1,   0,   0,  -1,  -1,   7,  29,  42,  21,  -8,  -3,  -5,  -2,  -2,   0,  -1,   0,  -1,  13,  45,  -5, -11,  -1,  -2,  -3,  -5, -12,  24, -14,  -7,   1,   1,   0,  15,  -4,   6, -13,  -4,   0,  -1,   0, -19,   1,   4,  -8,   3,  -1,   1,  -1,  -6,   5,  -8,   1,   1,  -4,  21,   4,  -4,  42,   5, -21,   8,   8, -26,  -7 },
     {   -1,   4,   1,  13,  -4,   4,  -1,   1,  -3,   2,  -2,  -1,  -6,  -2, -11, -19,   0,   2,   4,   0,  -2,   0,   2,  -1,  -6,   5,  -3, -20,   5,  -2,   1,  -1,   0,   1,   1,   0,  -1, -13,  12,  -8,  12,  -2,   3,  -1,   1,  -2,   1,  -1,  -4,   8, -37,  -3,   8,  -4,   1,   2,  -5,  -2,  -4,   9,   2,   1,  -2,   2,  26,  17,  10, -25,   5,  -3,   0,  -3, -75,  10, -37,  -5,  15,  -1,   5,  -2, -38,   1, -12,  33,  -1,  10,  20,   9,  13,   5,  -4,  -8,  -2,  -4, -12,   8 },
     {    0,   5,   5,  -6,  -1,   3,   2,   3,   2,   0,   1,  -1,   0,   6, -13,  46,  -7,   6,  -1,   3,  -2,   0,   1,   0,   1,   8,  29,  23, -17,   3,  -8,  -4,  -4,   0,  -2,   1,  -3,  -8, -58,  17,  -3,  -1,  -2,   4,   0,   0,   0,   1,   7,  48,  -1,  21,  -3,   0,  -5,   1, -13,  24,  43,  -1, -13,   0,  -4,  -2, -13,  -4,  20, -19,  -6,  -1,  -1,   2, -16, -13,   1, -17,   3,  -1,   3,   1, -18,  -1,   3,  -2,   5,  -1,   3,   0,   0,   6,  11,  -7,   7,  11,  -2,  -7 },
     {    4,   2,   2,  18,  -9,   1,  -3,   0,  -1,   1,   0,  -1,   4,   2, -12, -15,   7,   0,   4,  -2,   1,  -1,   0,   0,   4,   5,   7, -25,   3,   3,   3,  -1,   0,  -1,  -1,   0,   7,  -5,   6, -18,   6,  -4,   3,   1,   2,   0,   0,  -1,   9,   6, -34,  25,   7,   1,   0,   1,  22,  25,  21,  11,  -5,  -6,  -4,  -3,  -5,  22,  -7,   2,  -5,  -2,  -2,  -1,  22,   5,  32,  -4,  -5,   1,  -2,  -1,  23,  34,  -6, -12,   6, -14,  21,  -8,  -6,  50, -10, -30,  10, -10, -45,   5 },
     {   -2,  -5,  -8, -22, -29,   5,  12,  -1,   2,   4,   3,  -2,   3,   1,  13,  26, -20,   1,   8,  -5,  -9,   1,   1,   0,   1,  -4, -16,  68,  -5,  -4,  -2,  -3,  -4,   1,  -1,   0,   2,   1,   0,   4,  17, -14,   0,   2,   0,  -5,   1,   1,  -6, -14, -49, -14,  17,  -6,  -3,   7,  14,  10, -21, -15,   5,   0,  -6,   5,   0,  19,  -4, -14,   4,   4,  -1,  -4,  -6,  15,   6,   2,  -2,   7,  -4,  -2,  28,   3,  21,   0,   1,   2,   6, -11,  22, -14,  -4, -13, -16,   6, -11,   8 },
     {    5,   7,   1,  13,  75,  -6,   0,  -3,   2,  -4,  -2,  -4,   7,   7,   6,  46,  67,   0,  -7,  -1,  -4,  -5,  -2,  -3,   0,  -3, -15,  17,  25,  -7,  -2,  -8,  -3,  -3,  -2,   1,   0,  -5,  -4,   3, -10,  -9,  -5,   0,   3,  -4,   1,  -1,   1,  -7, -19,  -5, -12,  -6,   4,   4,   4,   6, -14,  12, -11,   1,  -1,  -3,   3,  11,  -7,   8,  -8,  -2,   0,   2,  -6,   9,  -3,   8,  -1,  -2,   4,   2,   4,   4,  -3,   4,   6,   1,   3,  -5,   1,   1,  -4,  -3,   5, -11,   0,   0 },
     {    1,   2,   7,  42, -19,  11, -10,  -1,  -7,  -5,  -2,   2,  -3,   0,  -2, -53,  37,  -7,  13,   0,  -3,  -4,  -3,  -8,   4,   6,  15,  55,  14, -17,  -2,   5,   8,   2,   5,  -6,  -1,   0,   9,  35, -12,   7,  -5,   1,  -4,   4,   0,   4,  -3,   2, -15,  30, -13,   9,  -1,  -3,  -4,   3,  11, -13, -18,   2,  -2,   8, -11,  -6, -13,  -1,   5,   8,   8,   2,  -2,  -1,   9,   0,   2,   1,  -3,  -5,  -8, -12,  -8,  -4, -20,  -2,  -1,   9,   2,  -4,  -3,   8,   2,  -4,  15,   3 },
     {    2,  -1,   3,   0, -10,   1,   4,  -4,  -5,   1,   1,   4,   3,  -2,  -2,  12,   7,   2,   4,  -4,  -4,   3,   1,   0,   3,   3,  22, -23,   2,  -7,  -1,   0,   1,   1,   0,  -3,   7,  -9,   1,  34,   7,  -2,  -3,  -1,   2,  -3,  -2,   1,   0,  13,  14,  34, -12,   1,  -5,   0,  -8, -35,   5,  -9,  -9,   4,   3,   3,  24, -11, -26, -15,  10,   3,   7,   1, -10,  14, -27,  24,  11,   5,   4,   0,  56,  18,  32,  22,  16, -12,   7, -25, -12, -17,  -8,  -7,  -2,  -4,  -6,   5 },
     {   -2,  -3,   2,  -9,  -4,  -3,  -6,   0,   3,   0,   3,   1,  -3, -10,  -6,  14,  -8,   1,   2,   5,   3,   1,   1,   0,  -5,  -6,  -7,  14, -35,  -7,   0,  -2,   1,   4,  -1,   3,  -5, -16, -17, -13, -19, -11,  -1,   0,   3,  -3,   4,   0,  -8,   4, -10, -19, -13,  10,  -1,   1,  -8, -32,   8, -27,  11,   7,   5,  -1,   6,  -7, -34,   7,   2,   2,   1,   4,  -1, -14,   6,  -6,   9,   1,   3,  -1,  -2,   1, -39,   8,  -3, -24,  10,   1,  -8,  23, -42,  10,  13, -50,  11,  29 },
     {   -2,  -2,  -5,  -8,  16, -21,  -8,  -8, -14, -12,  -1,  17,   0,  -1,  10,   6,   5,  -6,   4,   9, -23,   1, -10, -20,   0,   4,  -3,  10,  -3,  20,  22,  15,  30,  23,  15,  -8,  -2,  -5,  15, -47,   9,  -8, -13,  -3,  -4,   7,   7,   9,   2,   4,  -1,  16,  20,  10,  -1, -11,  -2,   2,  33,  22,   9,   6,   6,  -1,  -7, -27,  -9, -21, -12,   9,   5,   3,   2,  -4, -13,  -3, -12,   9,  -3,  -3,   9,   6,  13,   8, -38, -14,   0,  13,  11, -22,  -9,  -3,  -3,  -2,   0,   1 },
     {    3,   4,  -2,   0,  55,   8,  -5,  -7,   5,  -6,  -4,   0,   0,  -3,   4,  -2, -33,  15,  -1,   0,   5,   2,   6,   0,  -2,  -4,  -8, -11, -34, -21,  -5,  22,  -5,   1,   5,   0,   2,   1,  16,  10, -21,  16,  20,  -7,   5,   8,  -7,   7,  -1,  -5, -21,  25,  25,   1, -19,  -1,  -1,   6,  -2, -20, -21, -22,   2,  11,  -2,  -5,   3, -10,  29,  18,  29,   6,  14,  -4,   3,  -4,  26,   2,  -1,  -4,   7,  -1,  -2,   0,  -1,   9,   1,  15,  20,  -4,  -1,   8,   3,  11,  -5,   2 },
   },
   {
     {   75, -42,   2,  -4,   0,  -1,   0,  -1,   0,   0,   0,   0, -73,  36,   1,   3,   0,   1,   0,   0,   0,   0,   0,   0,  38, -13,  -5,   0,  -1,   0,  -1,   0,   0,   0,   0,   0, -21,   1,   5,   0,   1,   0,   1,   0,   0,   0,   0,   0,  10,   3,  -4,   0,  -1,   0,   0,   0,  -6,  -3,   2,   0,   0,   0,   0,   0,   3,   3,  -1,   0,   0,   0,   0,   0,  -2,  -2,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,  -1,  -1,   0,   0,   1,   1,   0,   0,  -1,  -1,   0,   0 },
     {   80, -23,  -7,  -3,  -2,  -1,  -1,  -1,   0,   0,   0,   0,  23, -31,   7,  -2,   2,   0,   1,   0,   0,   0,   0,   0, -65,  35,   0,   2,   0,   1,   0,   0,   0,   0,   0,   0,  37, -13,  -7,   1,  -2,   0,  -1,   0,   0,   0,   0,   0, -23,  -1,   8,  -1,   1,   0,   1,   0,   6,   6,  -4,   0,   0,   0,   0,   0,  -7,  -3,   2,   1,   0,   0,   0,   0,   1,   3,  -1,  -1,   0,   0,   0,   0,  -4,  -1,   0,   0,   0,   2,   0,   0,  -2,   0,   0,   0,   0,   1,   0,   0 },
     {  -31,  23,  -5,   2,  -1,   1,   0,   0,   0,   0,   0,   0, -67,   2,  16,   1,   4,   0,   2,   0,   1,   0,   0,   0,  12,  42, -18,   1,  -4,   0,  -2,   0,  -1,   0,  -1,   0,  53, -49,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0, -34,  19,   9,  -3,   2,  -1,   1,   0,  25,  -3, -10,   1,  -1,   0,  -1,   0,  -8,  -3,   5,   0,   0,   0,   0,   0,   9,   1,  -3,  -1,   0,   0,   0,   0,  -3,  -2,   1,   0,   5,   0,  -1,   0,  -1,  -1,   1,   0,   3,   0,   0,   0 },
     {   25,  13, -14,   0,  -4,   0,  -1,   0,  -1,   0,   0,   0,  40, -25,   7,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,  41,  -9, -14,   0,  -3,   0,  -1,   0,   0,   0,   0,   0, -42, -32,  27,   0,   5,   0,   2,   0,   1,   0,   1,   0, -25,  59, -11,  -4,  -1,  -1,   0,  -1,  28, -34, -10,   7,  -2,   2,  -1,   1, -23,   8,  16,  -4,   1,  -1,   1,   0,   7,   1, -10,   0,   0,   0,   0,   0,  -8,  -1,   6,   2,   3,   2,  -3,  -1,  -4,  -1,   2,   1,   1,   1,  -1,  -1 },
     {    6,  42, -21,   1,  -5,   0,  -2,   0,  -1,   0,  -1,   0, -38, -37,  27,   0,   5,   1,   3,   0,   1,   0,   1,   0, -31,  29,  -4,  -2,   0,  -1,   0,   0,   0,   0,   0,   0, -24,  26,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  57,   0, -25,   2,  -5,   1,  -2,   0,  -5, -42,  17,   4,   2,   1,   1,   1,  -7,  27,   5,  -7,   0,  -2,   0,  -1,  12, -11, -10,   3,   0,   1,   0,   0,  -3,   3,   6,   0,   5,  -2,  -4,   0,  -1,  -1,   3,   1,   3,  -1,  -2,  -1 },
     {  -36, -69,  33,   2,   8,   1,   4,   1,   2,   1,   1,   1,  -3,  34,  -3,  -5,  -1,  -2,  -1,  -1,  -1,   0,   0,   0, -49,   3,  -4,   7,   0,   3,   0,   1,   0,   1,   0,   1,   0,  12,  16,  -7,   1,  -2,   0,  -1,   0,   0,   0,   0,   5,  37, -27,   1,  -3,   0,  -1,   0,  29, -36,   4,   6,  -2,   2,  -1,   1, -12,  13,   8,  -6,   2,  -2,   0,  -1,  11,   0,  -8,   1,  -1,   0,   0,   0,  -3,  -1,   3,   0,   4,   0,  -2,   0,  -1,   0,   1,   0,   3,   0,  -1,   0 },
     {   -6,   9,   5,  -1,   0,   0,   0,  -1,   0,   0,   0,   0, -35, -27,   3,   5,   3,   3,   0,   1,   1,   1,   1,   1, -32,  -9,  25,  -6,   5,  -1,   1,  -1,   1,   0,   1,   0, -45,  28,  -3,   2,   1,   1,   0,   0,   0,   0,   0,   0,  -8,  36,  13,  -7,   1,  -2,   0,  -1,  34,  30, -43,   0,  -4,   0,  -1,   0,  19, -50,   7,  11,  -2,   2,  -1,   1,  -9,  21,  11, -10,   1,  -2,   1,  -1,  10,  -5, -12,   2,  -1,  -1,   6,   0,   4,  -1,  -4,   0,   1,   0,   1,   1 },
     {  -22, -47,   8,   9,   2,   4,   1,   1,   1,   1,   1,   1,  -4, -46,  35,  -7,   5,  -2,   2,  -2,   0,   0,   1,   0,   5,  22, -20,   3,  -1,   2,  -1,   0,   0,   0,   0,   0, -35, -21,  19,   0,   2,   1,   0,   0,   0,   1,   1,   1,  -5,  10,  13,  -6,   2,  -1,   0,   0, -30,  40,   8,  -9,   1,  -3,   1,  -1,  26,  22, -36,   2,  -1,  -1,  -1,   0,   3, -29,   9,   8,  -3,   2,  -1,   1,  -1,  17,   5,  -8,   3,  -5,  -7,   3,   0,   5,   4,  -1,   2,   0,  -4,   1 },
     {    9,  33, -21,   9,  -4,   1,  -3,   0,  -1,   0,  -1,   0, -18,  11,   0,  -7,   0,   0,   0,   0,   0,   0,   0,   0, -31, -60,  15,  13,   4,   4,   1,   2,   1,   1,   1,   1,  12,  14,  28, -18,   2,  -5,   1,  -3,   0,  -1,   0,   0, -45,  13, -18,  10,   1,   2,   1,   1,  14,  16,  27,  -9,  -1,  -2,   0,   0,   3,  23, -39,  -3,   1,  -2,   0,   0,  13, -29,   9,  13,  -4,   2,   0,   1,  -4,  12,   5, -10,   6,  -6,  -6,   5,  -1,   3,   3,  -2,   3,  -3,  -2,   1 },
     {    9,  31, -15,  -1,  -3,   0,  -2,   0,  -1,   0,  -1,   0,  17,  61,  -5,  -9,  -4,  -4,  -1,  -2,   0,  -2,  -1,  -1, -32,  16, -32,  11,  -2,   2,   0,   1,  -1,   1,   0,   0, -11, -37,  29,   3,   1,   0,   1,   1,   1,   0,   0,   0,  36,  25,  -2, -12,   1,  -3,   0,  -2, -21,   5, -19,   5,   1,   1,   0,   0,  34, -23,  -7,   8,  -4,   3,  -1,   1, -16, -17,  18,  -2,   2,   1,   0,   0,   6,  12,  -5,  -3,  -1, -11,   0,   4,   1,   2,   2,  -1,  -1,  -5,  -2,   1 },
     {    0, -18, -23,  19,  -1,   3,  -1,   2,  -1,   1,   0,   1,  11, -13,  36, -17,   3,  -2,   1,  -2,   0,   0,   0,   0,  30,   5, -22,  10,  -2,   1,  -1,   1,  -1,   0,   0,   0,  31,  34,   1, -15,  -2,  -2,  -2,  -2,  -1,   0,   0,   0,   4,  -5, -38,  11,  -1,   3,  -1,   1,  23, -17,   7,   0,  -4,   1,  -2,   0,   5, -42,  -6,   7,   2,   1,   0,   1, -27,   2,  43,  -6,   1,  -2,   0,   0,  -2,  28, -24, -10,  -1, -20,   7,  11,  -5,  13,   2,  -8,  -1,  -6,  -3,   5 },
     {    8,  12,  17,  -9,   1,  -3,   0,  -1,   1,  -1,   0,  -1,  13,  26, -21,   0,  -2,  -1,   1,   0,   1,  -1,   0,   0,  12,  52,   2,  -6,  -5,  -4,  -2,  -2,  -1,  -1,  -1,  -1, -30,  14, -30,   0,   3,   0,   1,   1,   1,   1,   0,   1,   0,  -9,   9,  17,  -5,   3,  -1,   1,  51,   7,   2, -25,   1,  -4,   0,  -1, -25,  -4, -44,  16,   2,   3,   2,   1,  24, -34,  21,  11,  -7,   3,  -2,   1, -13,   3,   5, -10,   7,  -3,  -4,   7,  -4,  -4,   3,  -2,   2,  -2,   0,   1 },
     {    8,   9,  53, -26,   4,  -8,   0,  -4,   0,  -1,   0,  -1,  -6, -39, -51,  27,   3,   7,   1,   3,   1,   2,   1,   1,  14,  -2,  28,  -1,  -5,  -1,  -1,  -1,  -1,   0,  -1,   0,  32, -10,   2, -14,   2,  -3,   1,  -1,   1,   0,   0,   0,  20,  33, -23,   1,  -2,   0,  -2,   0, -17, -14,  -2,   4,  -1,   1,   0,   1,  20,  -7, -10,   3,  -1,   2,  -1,   0, -17, -23,  19,   3,   1,   0,   0,   1,  -8,  24,   5,  -8,   1,  -6, -12,   3,  -4,   0,   9,   2,   0,   3,  -3,  -3 },
     {    2,  -5, -20,   7,   0,   1,  -1,  -1,   1,   0,   0,   1,  12,  11,  15,  -1,  -3,  -2,  -3,  -1,  -2,   0,   0,   0,  19,  13,   0, -18,   3,  -1,   1,  -2,   0,  -1,   0,   0,  25,  42, -16,   6,  -6,   0,  -4,   0,  -2,  -1,  -2,   0,  -3,  23, -15, -19,   3,  -2,   1,  -1,   6,   8, -26,  20,  -4,   2,  -2,   0,  37,  -8,  -8, -18,  -2,  -2,  -1,  -1,  -8, -41, -32,  25,   4,   3,   1,   2,  -2, -22,  50,   5,  -8,  18, -14, -15,   2, -15,   3,  11,  -5,   5,   3,  -6 },
     {  -13, -31, -58,  19,   2,   6,   0,   3,   2,   2,   2,   2,  -8, -15, -25,  11,   1,   4,  -1,  -1,  -2,   3,   1,   1, -10, -38,   5,  -3,   5,   1,   2,   1,   0,   0,   1,   1,   9, -28, -17,  17,  -3,   6,  -1,   1,   1,   1,   0,   1,  34,  -5,  29, -10,  -2,  -1,  -3,   0,  15, -15, -16,  -4,   3,  -1,   1,   0, -21, -14,   3,  15,   0,   3,  -2,   0,  10, -24,  25,   2,  -1,   1,   0,   1, -23,  19,  24, -12,   3,  12, -12,  -1,  -4,  -5,  10,   4,  -1,  11,   1,  -3 },
     {   -1, -12, -28,  31,  -4,   2,  -3,   2,  -1,   1,  -1,   1,  -3, -19, -52, -14,  13,   2,   4,  -1,   1,   0,   2,   1,  10,  15,   2,  30,  -8,   3,  -5,  -1,  -1,   2,  -1,   1,   7,  -8,  18, -33,   1,   0,   3,  -2,  -1,   2,   0,   1,  27,  11, -11,  12,   1,  -4,  -1,   0,  10,  16,  10, -16,  -8,   1,   0,   0,  10,  -1, -20,   0,   7,  -2,  -1,   0,   5,   8, -16,   3,  -4,  -1,  -1,   1,   8, -51,  -2,  19, -14,   6,  38,  -8,   0,  10, -17,  -7,  -5,  -5,   6,   8 },
     {  -10, -13, -48,  11,   0,   5,   1,   2,   0,   2,   0,   2,  -4,  11, -35,  13,  -2,   4,  -2,   2,  -2,   1,   1,   1,  -5,  33,  31, -22,   4,  -3,  -1,  -2,   0,  -2,   0,   0, -28,  -9, -28,   1,   4,   3,   3,   1,   1,   1,   1,   1, -42,   5,  -8,  18,  -1,   3,  -2,   0, -13, -24,  31,   6,  -4,   2,  -2,   1,  39,   6,  11, -17,   2,  -2,   0,  -2, -25,   6,   0,   0,   3,  -2,   1,   0,  13,  18, -10,  -2,   7, -18,  -6,   5,   1,   1,   9,  -1,   2,   1,  -4,   0 },
     {    0,   6,  23,  -3,  -1,  -3,   2,  -2,   0,  -1,  -1,  -1, -10, -17, -20,  -3,   8,   2,   1,   2,   2,   1,   1,   0, -13,  -8, -46,  20,   1,   4,   2,   2,   0,   1,   0,   1, -26, -18,  10,   8,  -6,   5,  -2,   3,   0,   0,  -1,   1, -13, -45, -10,  12,  10,   2,   3,   2,  32,  -9,  25,  12, -11,   3,  -3,   0,  10, -27,   8, -21,  10,  -2,   2,  -1, -35,   0,   3,  24,   1,   1,   1,   1,  15,   3,  38, -11, -15,  24, -12, -12,   8,  -5,  -2,   8,  -5,   4,   6,  -2 },
     {    0,  -5, -16, -26,  18,  -6,   8,   0,   1,   0,   1,   0,   7,  -1,  10,  47, -25,   8, -11,   2,  -4,   1,  -2,   1,   4,  -2,   1, -24,  10,  -1,   5,   1,   1,   0,   1,  -1,  15,  -5,  29,  18,  -7,   1,  -3,  -1,  -2,  -1,   0,  -1,  34,   6,  15, -11,  -3,  -3,   0,  -2,  46,  23,  25, -14,  -8,  -4,  -2,  -1,  -5,  10, -19, -27,   7,  -4,   3,  -1, -29,  16, -22,  14,   2,   0,   2,   0,  20,   2, -13,   0,  -4, -23,   0,   7,  -4,   7,   8,  -7,  -4,  -2,  -8,   0 },
     {   -7,  -7, -31, -30,  15,   4,   1,   2,   0,   0,   1,   0,  -5,  -6, -42,  30,  -6,   2,   3,   1,   2,   1,  -1,  -1,   4,  27, -28,  10,  -2,   7,  -3,   2,  -1,   1,  -1,   0,  -4,  42,  51,  -8, -13,  -6,  -3,  -1,  -2,  -1,   0,   0, -24, -19, -11, -23,   7,   0,   7,   1, -19,  12, -19,  24,   6,   0,   3,  -2, -24,   1,  10,  10,  -8,   0,  -1,   3,  22,   3,   9,  -2,  -2,   1,  -1,  -1,   1,   8,  -4,  -4,   7,  -1, -11,   4,   2,  -8,   6,   2,  -3,   0,   1,   0 },
     {   -4,  -4,  -5, -41,  12,  -5,   9,   0,   2,   1,   2,   1,  -2,  -8,   5,  33, -10,   1,  -4,   0,   0,  -1,  -1,   0,  -9, -19, -37, -14,   3,   5,   7,   2,   0,   2,   0,   0,  -4,  -2, -20,  20,   6,  -3,  -1,  -1,   0,   0,   1,  -2, -19,  -3,  -9,  14, -12,   6,  -4,   3,  -4, -20,   5,  -4,  11,   2,   1,  -1,  24,  -8,  -2,  22, -10,   2,  -2,   1,  12, -30,   5, -13,   3,   3,   1,   0, -11, -26,   6,  26,   0,  -4,  52, -10,  -8,  36, -13, -13,   4, -11,  -4,   8 },
     {    6,   5,  10,   4,  -7,  -1,  -1,  -1,  -3,   1,   2,   2,   0, -11, -34, -28,   8,   6,  -5,  -4,  -3,   3,   3,   2,   4, -13, -55,   5,  14,   4,  -1,   2,   2,   1,   3,   3,  10,  23, -22,  16,  -8,   8,  -2,  -2,   1,   3,   1,   3,   5,  11,  25,  -9,   2,  -7,  -3,   1,  25,   8,   7,   1, -10,  -1,  -1,  -1,  43,  29,  16, -11,  -2,  -7,  -7,  -1,  19,   4, -12, -26,  -3,  -7,   2,   0, -21,   8, -24,  -1,  15, -25, -11,  13, -12, -12,  15,   2,   4,   5,  -1,   2 },
     {   -1,   3,   0,   0,   7,  -6,   2,  -2,   3,   0,   2,   0,  -6,  -6, -12, -16,  -2,   1,   0,   0,   0,   1,   0,   1,  -4,  -9, -31,  -8,  21,   2,   5,  -1,   0,   0,   1,   0,   4,   8, -33,  19,  -2,   6,  -4,   1,  -1,   1,  -1,  -1,  -5,  33,  -3,   3,   5,  -5,  -1,   1, -38,  -8,  -2,   2,  -2,   2,  -1,   0, -34, -12, -26,   8,  20,  -3,   4,   0,   6,  23,   7,  42, -18,   0,  -2,   0,  44, -22,   6, -16, -12, -23,   2,  17,   3,   2,  27,  -6,  -2,  10,  -7,  -3 },
     {    3,   2,  24,  53, -34,   1, -10,   1,   1,  -1,  -4,  -2,  -3,  -2, -16,  -2,   6,  13,   5,   9,  -1,  -1,  -2,   0,   0,   5,  13, -13,  19, -18,   7,  -3,   4,  -2,  -3,  -2,  -3,  -1,  22,  40, -18,  -8,   0,   3,  -3,  -6,  -1,  -2,  -7, -19,  -8, -28,   0,  14,   5,  -1,   5,   1,  -8,  22,   7,  -4,   2,  -1, -15,  -3,   8,  -7, -11,  10,   6,   0,  -1, -28,  -7,  13,  13,   4,   2,  -1,  -1,   2,  -5,  12,   0, -27,  27,  11,  -1,  32,  13, -18,  -1,  -5, -17,  -2 },
     {   -2,   2,   1,   5, -23,  13,   0,   4,  -5,   5,  -3,   1,  -3,   1, -13,   2,  30, -17,   2,  -6,   1,  -1,   2,   1,  -3,   0, -19,   2, -30,  17,  -3,   6,   1,   0,  -1,   2,   5,  27,  -2,   3,  15, -15,   2,  -4,   4,   0,   1,   2,  -9,  14,  39, -12, -30,  11,  -5,  -1,  -1, -17,  -9, -19,  29,   2,   6,   1, -19,  23, -14,  13, -10,  -6,   0,  -1, -60,  -1,  -7,   1,   4,   7,   1,   0,  18,  16,   6,  17,   4,  -4,  -1, -16,   6,   8,  -9,  12,  -1,  -9,   2,   0 },
     {   -5,  -8, -11, -41,  12,  -5,   2,   2,   0,   2,   1,  -3,  -2,  -4,   6, -26,  32,  -8,   6,   0,   6,  -4,   3,  -2,  -2, -11,  19,  20, -32,   4,  -4,  -4,  -2,   2,   1,  -2,  -6, -17,  -3,   2,   2,   1,   6,   3,   1,  -3,  -2,   1,   0, -26,  -7,   6,   6,  -4,  10,   0,  10,  -8, -26,  15,  -1,   8,  -3,   1,  -2,  -7, -13,  -4,   1,  -3,  -1,   3,  -6, -23, -30,   9,  18,   6,   1,   2,   3,  -7, -23,  26,  -5, -38,   2,  19, -13, -10,  44,  13,  -6,  11,  13,  -9 },
     {   -2,   4,  -8,  -6,  55, -47,   3,  -6,   5,  -2,   1,   1,  -5,  -1,  -4, -39, -26,  31,  11,  -1,   2,   1,   4,  -1,   2,   7,  10,  40,  22, -18, -14,   8,  -4,   0,  -3,   1,  -4,   6,  -8,  14, -22,  -9,  15,  -3,  -3,  -1,   0,  -1,  -6,   7,  11, -10,  -2,   8, -10,  -1,  -6,  -4,  -6,  -1,  -2,  -4,   5,  -2, -12,  10,  -5,   0,  -3,   5,   2,  -4, -33,  -5,  -9,   3,   3,  -7,  -2,   0, -14,  17,   2,  10,  10,  10,   9,  -7,  -1,  12, -16,   1,   6, -10,  -3,   4 },
     {    0,  -2,   0,  -6, -15,  -5,  -6,  -4,   0,  -2,  -3,  -4,   5,   4,  10,  22,   4,  10,   7,   8,  13,  -3,   3,   0,   5,   2,   9,  52, -10,  -3,   1, -13, -13,   2,   2,  -6,   3,  17,   2,  -3,   8,  11,  -5,  13,   3,  -3,  -9,  -5,  -6,   2,  38, -10,  -2, -12,   9,  12,  -8, -21,  -6, -30, -11,   6, -15,  -2,  11,  -9,  25, -26,  30,  11,   3,  -5,  22, -24,   2,  15,   3, -15,   6,   2,  28,   7,   1,  -7,  -6,  -8,   0,  -3, -18,   7, -11, -24,  20,   4, -21,   7 },
     {   -2,  -5, -11, -51, -16,   9,  19,  -2,   2,   2,  -3,  -2,   2,  -1,  -6, -49,  31, -11,   2,  -2,  -5,   4,   1,   4,   7,   4,  27, -16,   8,  -3,  -4,  -1,   5,   0,  -1,   0,   8,  -2,  15,  23,  -5,   0, -12,  -2,  -1,  -4,   0,   0,   5,  -5,   7, -14,  10,  -3,   0,   0,  16, -12,  -4,  20, -10,  -7,   0,  -4,  16,  23,  -3,   6,  -3,  -5,  -2,  -3,   7,   5,  25,   6, -12,  -3,   0,  -3,  21,  -2,   3, -27, -13,  12,   6,  -4,  14,  18, -30, -17,  -3, -14, -20,   6 },
     {    4,  -2,   7,  18,   6,  -8,   7,  -4,  -1,   0,  -6,   0,   2,  -8,   7,   0, -35,  19, -12,   6,  -2,   0,   2,   0,   4,  -2,   9,   6,  -1,  -6,   4,   6,  -3,  -3,  -2,   1,   2,  -7,  14, -13, -17,  -1,   2,  -3,  -1,   1,   1,  -2,  -7, -22,   6,  -9,  -9,   8,  -7,   0,  18, -32, -16, -11,  33,  -8,   6,  -1,  34,  18,   3,  39, -12,   1,  -4,  -7,   7,  -6,   6, -10, -22,   5, -10,   1,  47, -25,   2, -18, -15, -21, -13,   7,   2, -12,   7,  17,  -4,   2,  19,  -9 },
     {    3,   0,  -2,  -5,  30,  21, -41,   8,   3,  -1,  -5,   5,   4,   0,  -2,  -3,  11,   2,  31,   0, -17, -13,   1,  -1,   3,  -4,  -2,   8, -10,  -2,  16, -13, -17,  14,  17,  -1,   3,  -9, -19,  13,  -8,   4,  -5,   4,  10,  -5,  -2,  -1,   7,   8, -27,   5, -32, -13,  -9,  10,  13,  33,   4,   5,  26, -14,  16, -16,  -3,  10,  25, -10,  -3,   8,   0, -11,  15,  -8,   8,   2,  12,   1,  -9, -12,  37,  16,   0,  -6,  14,  -4,   7, -15,  16,   5,  -8,   5, -11,   4,  11,  -4 },
     {    2,   0,   3,  -1,   0,  24,   6,  -6,   1,   8,  -9,   1,   0,  -1,   0, -15, -16,   1,  10,  -5,   6, -12,  15,  -8,   6,   7,  18,  13,  -6,  33, -15, -15,  -2,  11,   4,  -4,   6,   8,  24,  56,  -5, -20,  18,  -7,  -7,   5,  -4,  -1,   2,  12,  11,   8, -17,  12,  -1,   5,  -7,  -1,  41,   2,   2,   2, -11,   7,   8, -29,   0,  10, -11, -10,   0,   1,  12,   7,  10,   3,  10,  14,  -2,  -1,  -2, -24,   0,  -1,  16,   9, -14,  -9, -31,  -6,   8,  11,  17,   9,  14,  -6 },
   },
   {
     {   95, -48,  -4,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0, -64,  24,   8,   2,   2,   0,   1,   0,   0,   0,   0,   0,  -4,  10,  -4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   9,  -5,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -47,  -1,  16,   2,   3,   0,   1,   0,   1,   0,   0,   0, -60,  52,  -9,   1,  -1,   1,   0,   0,   0,   0,   0,   0,  78, -27, -15,   1,  -3,   0,  -1,   0,  -1,   0,   0,   0,  -6, -15,  11,   0,   2,   0,   1,   0,   0,   0,   0,   0,   0,   4,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0 },
     {  -12,  70, -48,   4,  -7,   2,  -3,   1,  -1,   0,  -1,   0, -61, -24,  31,   6,   5,   1,   2,   0,   1,   0,   1,   0, -12,  14,   9,  -5,   1,  -1,   0,  -1,   0,   0,   0,   0,  52,  -7, -15,  -1,  -3,   0,  -1,   0,  -1,   0,   0,   0,  -5,  -8,   4,   1,   1,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   5,  -1,  -1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   2,   0,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 },
     {  -48, -42,  41,   0,   9,   0,   3,   0,   2,   0,   1,   0, -11,  23,  -4,  -4,   0,  -1,   0,  -1,   0,   0,   0,   0, -44,  61, -20,   2,  -3,   1,  -1,   0,   0,   0,   0,   0,  52, -23, -12,   5,  -3,   1,  -1,   1,   0,   0,   0,   0,   2, -15,   9,   0,   1,   0,   0,   0,   1,   4,   2,  -3,   0,   0,   0,   0,   1,   1,  -2,   0,   0,   0,   0,   0,   4,  -2,  -1,   1,   0,   0,   0,   0,   1,  -1,   1,   0,   2,   0,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0 },
     {  -17,   7,  12, -10,   4,  -2,   2,  -1,   1,  -1,   0,   0, -34,  27, -19,  10,  -2,   2,  -1,   1,   0,   0,   0,   0, -48, -23,  31,   4,   3,   1,   1,   1,   0,   0,   0,   0, -17,  62,  -9, -12,  -1,  -3,   0,  -1,   0,   0,   0,  -1,  61, -17, -26,   5,  -3,   1,  -1,   1,  -4, -17,   9,   3,   0,   1,   0,   0,   7,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   5,  -1,  -1,   0,   0,  -1,   1,   0,   4,  -1,  -1,   0,   0,  -1,   0,   0 },
     {  -15, -16,  15,  -8,   7,  -1,   2,  -1,   1,   0,   0,   0, -28, -54,  48,  -1,   7,  -1,   3,  -1,   1,   0,   1,   0,  24,  51, -28,  -5,  -4,  -2,  -2,  -1,  -1,  -1,   0,   0, -57,  36, -17,   8,  -1,   2,   0,   1,   0,   0,   0,   0,  13,  -1,  -2,   3,  -2,   1,  -1,   0,  16,  -9,  -1,  -1,  -1,   0,   0,   0,   1,  -2,   2,  -1,   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   2,   1,   0,   0,   2,   0,  -1,   0,   2,   0,   0,   0,   1,   0,  -1,   0 },
     {   -9,  12,  13, -13,   4,  -3,   2,  -1,   0,  -1,   0,   0, -21,   4,   1,   3,   2,   0,   0,   0,   0,   0,   0,   0, -44,  -8,  -7,  16,  -2,   5,  -1,   2,   0,   1,   0,   1, -20, -14,  37,  -7,   3,  -1,   1,   0,   0,   0,   0,   0, -24,  74, -11, -12,   2,  -4,   1,  -2,  51, -13, -33,   8,  -3,   2,  -1,   1,   0, -19,   8,   6,  -1,   1,   0,   1,   7,  -1,   5,  -2,  -1,   0,   0,   0,   0,   2,   1,  -2,   6,  -1,  -2,   0,   1,  -1,   0,   0,   3,  -1,   0,   0 },
     {   18,  25,  60, -64,  14, -15,   5,  -7,   2,  -3,   1,  -2,  -8, -38, -32,  32,   2,   5,   0,   3,   0,   1,   1,   1,  17, -10,  -6,  21, -11,   5,  -4,   2,  -2,   1,  -1,   0,  29,  19,  -4, -14,  -3,  -2,  -1,  -1,   0,   0,   0,   0, -29, -12,   2,   3,   5,   1,   1,   0,  -7,   8,   4,   0,   0,   0,   0,   0,   2,   0,  -3,   1,  -1,   0,   0,   0,  -1,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,  -1 },
     {   24,  52,  -4, -18,  -3,  -6,  -1,  -3,  -1,  -1,  -1,  -1,   8,  19, -25,   4,  -1,   0,   0,   0,   0,   0,   0,   0,  -3,  38, -49,  16,  -8,   5,  -3,   1,  -1,   1,   0,   0, -29, -39,  30,   4,   0,   2,   0,   1,   0,   0,   0,   1,  57, -22,  11, -11,   1,  -3,   0,  -1,  -5,  -7,   0,  -2,   2,  -1,   1,  -1,  -7,   0,   0,   1,   1,   0,   1,   0,  -1,  -1,  -1,   0,  -1,   0,   0,   0,   0,  -3,   1,   0,  -1,  -2,   1,   0,  -1,  -1,   0,   0,  -1,  -1,   1,   0 },
     {    1,  -6, -33, -43,  28,   0,   9,  -1,   5,   0,   2,   0,  32,  39,  49,  -1, -10, -10,  -4,  -6,  -3,  -3,  -1,  -2,   8, -16, -43,  31,  -9,   8,  -4,   4,  -1,   2,  -1,   1,  29,  29, -28,  -3,  -7,   3,  -3,   1,  -1,   0,   0,   0,   5,   8,  -3,  -8,  -1,  -1,   0,  -1,   1, -27, -10,   2,   3,   2,   1,   1,  -5, -11,   5,   6,   1,   0,   0,   0,  -5,  -5,   4,   3,   1,   0,   0,   0,  -2,  -1,   4,   1,   0,  -2,   1,  -1,  -1,  -1,   1,  -1,  -2,  -2,   0,   0 },
     {    2, -11,  19,  23, -16,   4,  -6,   2,  -3,   1,  -1,   1,  -1, -34,  12, -26,  12,  -3,   5,   0,   2,   0,   1,   0,  11, -13, -20,  20,   2,   2,   1,   0,   0,   0,   0,   0,  38,  -1,  40, -15,  -5,  -5,  -1,  -2,  -1,  -1,   0,  -1,  25,  17, -46,  -5,   1,  -1,   0,   0, -26, -43,  16,  17,   0,   3,  -1,   2, -25,  30,  18, -11,  -1,  -2,   0,  -1,  10,   5, -11,  -3,   0,   0,   0,   0,  -6,   3,  -1,   1,   1,  -1,  -1,   1,  -4,   2,   2,   0,   1,   1,  -1,   0 },
     {   -6,   5,  16,   7,  -3,  -2,  -1,  -1,   0,   0,   0,   0, -15,  -5,  37, -32,   8,  -3,   3,  -1,   0,   0,   1,   0, -35, -42, -31,  34,   6,   6,   2,   3,   2,   1,   1,   1, -18,   5,  20,   9,  -9,   3,  -2,   1,   0,   0,   0,   0, -11, -11,  29, -15,   5,  -1,   1,   0, -40,  49,   5,  -8,   4,  -3,   1,  -1,  30,  15, -21,  -1,  -2,  -1,   0,   0,   2,  -4,  -3,   4,  -1,   1,   0,   0,   9,   1,  -1,  -2,   0,   1,  -1,   0,   5,   2,  -2,   0,   0,   0,  -1,   0 },
     {   17,  30,  22,  38, -35,   6, -13,   1,  -5,   0,  -2,   0,   6,  23, -24, -43,  17,  -5,   6,  -1,   3,   0,   1,   0,   5,  10, -33,   7,   9,  -4,   3,  -1,   1,  -1,   1,   0,   9,  48, -13,   6, -13,  -1,  -4,  -1,  -1,   0,  -1,  -1, -33, -22,   0,   2,   3,   4,   1,   2,  42, -15,   3,   0,  -2,   1,  -2,   0,   8, -20,  -1,  -2,   2,  -1,   0,   0,  -8,  -4,   6,   2,   1,   0,   0,   0,  -2,   0,   1,  -2,   0,  -2,  -1,   1,  -1,  -3,   1,   0,   0,  -2,   1,   0 },
     {   10,   3,  30,   9, -11,  -2,  -5,  -1,  -1,  -1,  -1,  -1,   4, -17,   2, -11,   0,   3,   1,   1,   0,   1,   0,   0,  15,  -9,   2, -22,  15,  -8,   5,  -2,   2,  -1,   1,   0,  28,   1, -19,  29,  -9,   1,  -4,   0,   0,   0,  -1,   0,  50,  32,  10, -23,  -6,  -2,  -2,  -1, -25,  15, -48,   2,   6,   0,   2,   0,   8, -53,   6,  20,  -2,   4,  -1,   2, -17,   7,  19,  -4,  -1,   0,   0,   0,   4,   1,  -1,  -2,  -8,   3,   0,  -1,   0,  -1,  -1,   0,  -4,   1,   1,   1 },
     {    2,   4, -18,  -1,  -6,   6,   0,   1,   1,   0,   0,   0,   8,  12,  -8,   2,   6,  -5,  -1,  -3,   0,  -1,   0,  -1,  16,  46,   4,   3,  -8,  -4,  -3,  -1,  -1,  -1,  -1,   0,   3,  20,  18, -33,   0,  -3,   0,   0,  -1,  -1,  -1,  -1,  -9,  26, -48,   1,   5,  -2,   2,  -1, -39,  26, -10,  14,  -2,   1,   0,   1,  58,  -9, -23,   3,  -7,   1,  -2,   1,   2, -39,   3,   5,   0,   3,  -1,   2,  -3,  -7,   9,   1,  -1,  -3,   2,  -1,   2,  -2,  -2,  -1,  -2,  -5,   2,   1 },
     {   13,  33,  31, -16,  17, -17,   2,  -4,   0,  -4,  -1,  -2,   2,  22,   9, -48,   2,  -4,   5,  -2,   2,   0,   1,   0,  -2,  10,  27,  -5,  -8,   2,  -3,  -2,   0,   1,   0,  -1, -17, -24, -49,  39,   2,   4,  -2,   3,   0,   1,   0,   2,  -8,  35, -23,  11, -11,   2,  -1,   1, -21, -10,  30,   3,  -1,   1,  -2,   0,   2,  14,   3, -12,   0,  -1,   0,   0,  20,  -7, -13,   0,   2,  -1,   1,   0,  -6,  -4,   3,   2,   4,   0,  -4,   0,   1,  -2,  -1,   2,   2,  -1,   0,   1 },
     {    3,   4,  -9,  33,  66, -44,   9, -11,   3,  -6,   2,  -3,  -5,  -7,  -9, -26, -50,  25,  -1,   7,   0,   5,   0,   2,   3,   9,   4,   2,  -5,  15,  -8,   1,  -2,   0,  -1,   0,   3,  12,  22,  12,   4, -14,  -1,  -2,   0,  -1,   0,   0, -13, -24, -13, -21,   7,  -1,   6,   0,   0,   3, -12,  24,   1,   0,   0,   0,   8, -10,  20,   0,  -9,   2,  -2,   1,  -6,  12,   3,  -7,   3,  -1,   0,   0,   2,   0,  -6,  -2,  -2,   1,   0,   3,   2,   0,   2,  -1,   0,   1,  -1,  -1 },
     {   -6, -17, -18,  -2,   4,   6,   5,  -1,   1,  -1,   0,   1,  -4, -16, -18, -38,  38, -11,   9,  -1,   3,   0,   1,  -1,  14,   5,  41,  65, -46,   6, -14,  -1,  -4,   0,  -2,   0,  -1, -13, -14,  -9, -10,  11,   0,   3,  -1,   1,   0,   2,  18,  -9,   1, -19,  19,  -7,   3,  -1,  21,  17,  -1,   3,  -1,  -5,  -1,  -2, -13, -18,  -7,   7,   0,   1,   2,   1,  -8,   7,  10,  -2,  -2,   1,   0,  -1,  -1,   3,  -5,  -2,   0,  -3,   1,   2,   0,   1,  -1,   0,  -1,   1,   1,   1 },
     {   10,   3,  16,  11,  16, -14,   2,  -6,   1,  -2,   0,  -1,   9,  -8,  20, -12, -16,   4,   0,   1,  -1,   0,   0,  -1,  11, -13,   2, -10,  -9,  16,  -5,   3,  -1,   1,  -1,   0,  26,  -1,  15,  -7,  19, -11,   0,  -2,  -1,  -1,  -1,   0,  39,   7,  -7,  25, -17,  -2,  -1,  -2,  54,  29,  13, -33, -11,  -1,  -3,  -2,  -3,   6, -50,  -9,   9,  -2,   3,   0, -10, -35,  -2,  17,   1,   1,   0,   1,  -8,   0,  10,  -1,  -1,  -1,   0,  -3,  -6,   2,  -3,   0,  -3,  -1,   0,   1 },
     {   -1,  -5, -15,   2,   9,   5,  -2,   0,   0,   0,   0,   0,  -2,  -4, -31,   4,  -9,   2,   0,   4,   0,   1,   0,   0,  11,  14,   8,   9,  15, -10,  -1,  -3,  -1,  -1,   0,  -1,   9,  34,   3,  29, -24,   2,  -4,  -2,  -1,  -1,   0,   0,   5,  25,  30, -51,  -9,   2,  -2,   0,   7,  -2, -13, -10,  17,  -5,   3,  -1, -15,  46, -25,   4,   1,  -4,   1,  -2,  35, -18, -29,  10,  -8,   0,  -1,   1,  -7, -26,  10,   4,   3,  -2,   1,  -2,  -3,  -4,   1,   0,   2,  -3,  -1,   0 },
     {   -8, -19, -32,  18,  22,  -7,   1,  -3,   2,   0,   2,   0, -15, -34, -53,  11,  15,   6,   7,   4,   1,   0,   1,   1, -11, -25, -48,  11,   5,  10,   3,   5,   0,   1,   1,   1,  -2, -14, -35,  22,  13,  -1,   2,   1,   1,   1,   1,   0,   5,  20,  -1,  19,  -5,   1,  -2,  -1,  -7, -22,  25,  -1,  -1,   2,  -1,   2,  29, -12,  -9,   1,   1,  -2,   1,   0,  -1, -20,  13,  11,  -2,   0,   0,   0,  -6,   4,  14,  -2,   0,   1,   3,  -1,   0,   1,   4,  -1,   1,  -3,   4,   1 },
     {   -5, -11, -14, -41,  10,  14,  -3,   6,   2,   1,   0,   0,  -3,  -9,  -3, -16,  -2,   1,   3,   0,   0,   1,   0,   1,   0,  -7,   2,   3,  51, -33,  13, -10,   4,  -3,   1,  -1,   0, -19,  20,  35, -43,  12,  -6,   3,  -1,   1,   0,   0,  -1, -27, -47,   5,  -8,  20,  -4,   4,  26,  19,   7,   7,  11, -10,   0,  -2,   2,  -6,   4,  -1,  -4,  -1,   0,   0, -12, -14,  10,   6,   0,   1,   0,   0,  -1,   5,   5,  -2,  -3,   4,  -1,  -1,   1,   2,   0,   0,  -2,  -1,   0,   0 },
     {    0,  -6,  -3, -11,  -1,  -6,   7,   0,   0,   1,   0,   1,   5,  -1,   2,   8,  42, -16,   0,  -5,   1,  -3,   0,  -1,  -3, -16,  -5, -47, -44,  28,   2,  10,   3,   2,   1,   2,  10,   1,  14,  42, -11,   7,  -8,  -6,  -3,  -1,   1,  -1,   5, -17, -12,  -9,  24, -14,   0,   0,  21,   0, -19,  10,  -8,   0,   4,   3,  37,  22,  21, -12,  -6,   0,  -6,  -2, -10,  -6, -27,  -6,   4,   0,   3,   1,   0, -17,   1,  11,  -4,   0,   4,  -1,   4,  -3,   3,   2,  -4,  -2,   0,   0 },
     {   -3,   2, -10,   1,  18,  -8,   3,  -5,   2,   0,   2,   0,  -5,   4, -14, -13,  27, -16,   9,   0,   1,  -1,  -1,   0,  -8,  -2, -18, -25, -13,  16,   5,   3,   2,   1,   1,   0,  -7,  15,   6,   3,   3,   6,  -4,  -1,  -1,   0,   1,   0, -15, -11, -28,  32,  -5,  -4,  -1,   3, -24,  26, -38, -24,  11,   0,   4,   1, -62,   0,  -1,  32,   8,   1,   3,  -1,  35,  -6,  20,   9, -13,   1,  -2,   1,   1,   0,   3,  -6,   4,  -1,  -1,   2,  -1,  -1,   1,   2,   6,   0,   1,   1 },
     {    3,   4,  23,  20,  -5,  -5,  -6,   2,  -1,  -2,  -1,   1,   1,   0,  19,  21,  -6,  -7,  -2,  -1,  -1,   1,   0,   0,  -4, -14,   8,   1,  -2,   5,   2,  -1,  -2,   1,   0,   1, -13, -22, -24, -11,  -6,   3,  -1,   4,   0,   2,   0,   2,  -1, -28,  -2, -23,  23,  -2,   6,   2,   6, -17, -18,  40,  11,  -5,   4,  -4,  -9,  12,  -7,  16, -18,   3,  -3,   0,  12, -55,  22,   9,  -3,   7,   0,   3, -29,  18,  43, -10,   7,  21,  -9,  -8,  -4,   1,   0,   2,   2,   3,   1,  -1 },
     {    3,  -2,   1,   6,  -5,  10,  -6,   3,  -1,   0,  -1,   1,   2,  -4,  -5,   2, -13,   4,   0,   0,   0,   2,  -1,   0,   1, -13, -22,  15,  13, -10,   1,   0,   1,  -2,   1,   0,  -2,  -7, -34, -22,  13,  -3,   4,   0,   1,   0,   0,   2,  16,  -4, -13,  15,   5,  -5,   6,  -2,  19,  49, -19,  31, -21,   0,  -5,  -3,  -8,  -8,  31, -24,  -8,   6,  -3,   3,  44,   1, -37,  -8,   7,  -4,   1,   0,  -2, -42,   3,  16,  -3,   1,  15,   0,  -4,  -3,   3,   0,  -2,   2,   1,  -1 },
     {    4,   6,   9,   3,  43,  52, -59,   9, -13,  -4,  -3,  -1,  -4,  -7,  -5,  -3, -28, -49,  27,   4,   6,   5,   1,   2,  -1,  -3,   0,  -5, -23,  -8,  30, -15,   6,  -3,   2,  -3,   1,   0,   5,  -7,  18,  20, -26,   4,  -2,  -2,   1,   1,  -1,  -1,   6,  -1,  -6,  -9,  -2,   6,   0,  -7,   1,   0,   3,  -2,   3,  -1,  10,   1,   0,   0,  -2,  -2,   4,   0,  -6,   4,   3,  -2,   1,   1,  -3,   1,   2,  -5,  -4,  -1,  -2,  -1,   1,   2,   0,   0,   0,  -1,  -2,   0,   0,   0 },
     {   -5,  -7, -10, -41,  -2,   0,  -2,   1,   0,   3,   0,   2,  -5,  -7, -11, -60,   9,  14,   4,   6,   0,   3,   2,   0,   2,   0,  10,  -8,  19, -13,   7,   0,  -1,  -2,   0,   1,   2,   0,   8, -17,  41, -12,   9,  -7,   4,  -2,   1,  -3,   4, -14,  36,  26,  -6,  -2,  -4,  -4,  -2, -27, -43,  18, -16,  13,  -2,   3,  18,  25,   1,  10,   1,  -2,  -3,  -1,  -6, -12,   0,  -3,  -5,   0,   1,   2, -14,  -2,   9,   3,  -1,  12,   1,   1,   1,   2,  -6,   2,  -1,  -1,   0,   1 },
     {   -4,  -1,  -8, -18,  -4,   0,   2,  -2,   4,   1,   2,   1,  -7,  -2, -13, -24, -14,   9,   2,   3,  -1,   3,   0,   1,  -6,  -2, -17, -38,  13,   5,   4,  -2,   1,   1,  -1,  -1,  -4,   3, -15, -23,   0,  23,  -2,  -1,   0,   0,   1,  -1,  -2,  16,  -5, -21,  57, -25,  -4,  -5,  -5,   6,  15,  22, -16, -18,   7,  -2, -34,   3, -36, -11,  -6,  13,   5,   2, -34,   9,  -7,  23,  13,   0,   0,   2,  22,   2,   2,   7,   3, -10,  -3,   1,   1,  -3,   1,   1,   0,   1,   2,   2 },
     {   -5,  -7,  -5, -23, -43,  27,  -5,   1,  -2,   5,  -2,   0,  -4,  -5,  -3,  -4, -30,   2,   5,   1,   2,   1,   2,   0,  -1,  -1,  -2,   5,  -4,  31, -15,   7,  -2,   2,  -1,   1,   0,  -1,  -8,  34,  56, -33,   0,  -7,   2,  -3,   3,  -1,  -9, -13, -30, -37, -14, -10,  20,   2,   6,  11,  10,   6,  10,  12,  -3,  -1,  -5,   3,   1,  20,   4,  -5,  -2,  -2,   2,  -6,  11,   7,  -8,   4,  -2,   2,  23,   2,  -4,  -7,  -4, -10,   0,   1,   0,  -1,   8,  -1,  -3,   2,  -1,  -2 },
     {   -3,  -2,  -4,  -4,  -9,  26,  15,   4,   8, -17,  -6,   5,  -3,   1,  -1,   1, -11,  19,   0,   1,  -6,  14,  -6,   2,   0,   7,   8,  10,  -8,  13,  -6,  -5,  16,   7,  10,   1,   2,  18,  25,  27,   3,  19, -11,   1, -14,  -2,  -2,  15,  -7,   2,  13,  10, -13, -35,   1,  -3, -16,   2,  13,  -2, -27,  -9,   9,  -3, -17, -27,  19, -23,  12,  15,   7,   8,  -1, -15,   0,   0,   2, -20,  15,   1, -19,   3,  42,   6,   3,  23,  -3,  -7,  12,  10, -21,  23,  -8,   8,  13,   0 },
     {    3,   3,  -4,  -4,   8,  22,  -1,  15,  -2,   1,   2,  -2,   6,   6,  -3,   2,   7,  35,   7,  -8,   0,  -5,  -1,  -4,   5,   4,  -7,  -1, -11,  -6,  26,  -2,   4,   1,  -7,   0,   9,  18,   2,  15,  16, -11,  -4,  -8,   6,  -2,   1,  -2,  12,  14,  10,  -1,  33,  23,  -9, -10,   4,  24,  12,  -5, -17,  -5,  -9,  -1,  -5,   9,  20,  -5, -34,  -4,  -6,   3,  21, -16,  13, -21,   7,   0,  -4,  -1,  -5,  48, -26,  -8,  28,   5, -31,   1, -11,  -9,  -7,   0,   8, -10,  -6,   0 },
   }
 },
 {  //14
   {
     { -114,  42,   4,   5,   1,   2,   1,   1,   0,   1,   0,   0,  36,  -8,  -2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   31,  -2,  -5,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0,  93, -35,  -3,  -4,  -1,  -2,  -1,  -1,   0,   0,   0,   0, -70,  10,   6,   2,   2,   1,   1,   0,   1,   0,   0,   0,  14,   3,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -10,  -1,   1,   0,   0,   0,   0,   0,   3,   2,  -1,   0,   0,   0,   0,   0,  -5,  -2,   1,   0,   0,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0 },
     {  -12,  10,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -59,  -3,   9,   1,   3,   1,   1,   1,   1,   0,   0,   0, -58,  40,  -1,   2,   0,   1,   0,   0,   0,   0,   0,   0,  81, -15,  -9,  -1,  -3,  -1,  -1,   0,  -1,   0,   0,   0, -20,  -7,   5,   0,   2,   0,   1,   0,  11,   4,  -2,   0,  -1,   0,   0,   0,  -3,  -2,   1,  -1,   0,   0,   0,   0,   7,   1,  -2,   0,   0,   0,   0,   0,  -1,  -3,   1,   0,   3,   1,   0,   0,   0,  -1,   0,   0,   2,   0,   0,   0 },
     {  -15,  -6,   5,   1,   1,   0,   1,   0,   0,   0,   0,   0, -30,   9,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0, -70,  -3,  10,   1,   3,   1,   1,   0,   1,   0,   1,   0, -43,  46,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,  74,  -6, -13,   0,  -3,   0,  -1,   0,  -8, -12,   3,   0,   1,   0,   0,   0,  10,   2,  -1,   0,   0,   0,   0,   0,  -1,  -3,   1,   0,   0,   0,   0,   0,   5,   2,  -1,   0,   0,  -2,   0,   0,   3,   1,   0,   0,   0,  -1,   0,   0 },
     {    2, -10,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,  19,  14,  -7,   0,  -2,   0,  -1,   0,  -1,   0,   0,   0,  20,  -8,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  51,   8, -12,   0,  -3,   0,  -1,   0,  -1,   0,  -1,   0,  35, -59,   3,   0,   1,   0,   1,   1, -79,  11,  20,  -1,   4,   0,   2,   0,  14,  22,  -7,  -1,  -1,   0,  -1,   0,  -5,  -6,  -1,   0,   0,   0,   0,   0,   2,   2,  -1,   0,  -4,  -2,   1,   0,   1,   2,   0,   0,  -3,  -1,   0,   0 },
     {   15,  26, -11,   0,  -3,  -1,  -2,   0,  -1,   0,  -1,   0,  11,  -6,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  26,  13,  -9,  -1,  -2,   0,  -1,   0,  -1,   0,   0,   0,  29,  -6,  -2,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,  57,   2, -17,   0,  -4,   0,  -1,   0,  23, -65,   4,   2,   1,   1,   0,   1, -63,   8,  22,  -2,   4,   0,   2,   0,   9,  14,  -6,  -1,  -1,  -1,   0,   0,  -9,  -4,   1,   1,   2,   2,   0,   0,  -6,  -1,   1,   0,   0,   1,   0,   0 },
     {   34, 109, -32,  -2, -11,  -3,  -5,  -2,  -3,  -1,  -2,  -1, -16, -17,   2,   1,   1,   2,   0,   1,   0,   1,   0,   1,  -5, -20,   6,   0,   2,   0,   0,   0,   0,   0,   0,   0, -13,   4,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -5,  -7,   3,   0,   1,   0,   0,   0, -18,  13,   2,   0,   1,   0,   0,   0,  16,   1,  -6,   0,  -1,   0,   0,   0,  -1,  -8,   2,   1,   0,   0,   0,   0,   0,   2,   1,   0,   2,  -2,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
     {   -8,  -7,   0,   1,   1,   1,   0,   0,   1,   0,   0,   0, -11, -21,  10,   0,   3,  -1,   1,   0,   1,   0,   0,   0, -14,   7,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -31, -17,  11,  -1,   3,   0,   1,   0,   1,   0,   1,   0, -27,  11,   1,  -1,   1,   0,   0,   0, -40, -10,  19,   1,   3,   0,   1,   0, -28,  74,   0,  -4,  -1,  -2,   0,  -2,  59,  -3, -29,   2,  -5,   1,  -2,   0,  -8, -18,   5,   2,   7,   5,   0,   0,  -2,   0,   1,   0,   5,   2,  -2,   0 },
     {   -5,   0,   2,   0,   1,   1,   1,   0,   0,   0,   0,   0, -13,   1,  -5,   2,   0,   1,   1,   0,   0,   0,   0,   0, -17, -31,  13,   0,   3,   1,   1,   0,   1,   0,   0,   0, -11,  -1,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0, -28, -17,  12,   1,   2,   0,   1,   0, -25,   0,   4,   0,   0,   0,   0,   0, -50, -12,  27,   0,   4,   0,   2,   0, -28,  75,   4,  -7,   0,  -2,   0,  -1,  43,   3, -27,   1,  -7, -10,   5,   2,   7,   5,  -2,  -1,  -2,   1,   1,   0 },
     {    8,  17,   2,  -4,  -2,   0,  -1,  -1,   0,   0,   0,   0,  19,  99, -31,  -2,  -8,  -2,  -4,  -2,  -2,  -1,  -1,  -1, -21, -28,  -2,   3,   2,   2,   1,   1,   1,   0,   0,   0,  12,  -5,   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0, -10,   9,  -2,   0,   0,   0,   0,   0,  16,  -3,  -3,   0,   0,   0,   0,   0,   2,   8,  -1,  -1,  -1,   0,  -1,   0,  30,   0, -16,   1,  -2,   0,  -1,   0,   8, -35,   0,   4, -15,   1,  11,  -1,   4,   2,  -3,  -1,  -2,  -3,   1,   0 },
     {    2,  -4,   4,   0,   1,  -1,   0,   0,   0,   0,   0,   0,  -3, -44,   2,   5,   3,   2,   1,   1,   0,   0,   0,   0,  24,  -7,  17,  -4,   1,  -2,   0,  -1,   0,   0,   0,   0,  19,  53, -15,  -3,  -4,  -2,  -2,  -1,  -2,   0,  -1,  -1,   7,  -2,  -3,   1,   0,   1,   0,   0,  23,  19, -13,   0,  -3,   0,  -1,   0,  16,  -4,  -2,   0,   0,   0,   0,   0,  34,  18, -24,  -2,  -4,  -1,  -1,   0,  23, -64,  -8,   8, -33,  -3,  25,   0,   5,   9,  -3,  -3,  -6,  -2,   1,   0 },
     {    5,  15,  -4,  -1,  -2,  -1,  -1,   0,   0,   0,   0,   0,  12,  22,   5,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0,  -1,   9,  96, -23,  -4,  -6,  -3,  -3,  -2,  -1,  -1,   0,  -1, -38, -19, -14,   7,  -1,   3,   0,   2,   0,   1,   0,   0,  12, -24,  15,  -3,   2,   0,   0,   0,  -3,  17,  -6,  -1,   0,  -1,   0,   0,   7,  -5,   0,   0,   0,   0,   1,   0,   1,  20,  -8,  -2,  -1,  -1,   0,   0,  27, -18, -13,   3,  -6, -18,   8,   1,   2,   2,   2,  -1,   0,  -2,  -1,   0 },
     {    2,  -2,  -1,   2,   0,   0,   0,   1,   0,   1,   0,   0,   5,   1,   4,  -2,   1,  -1,   0,   0,   0,   0,  -1,   0,   5,  -7,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,  20,   5,  10,  -3,   0,  -1,   0,  -1,   0,  -1,   0,   0,  10,  33, -13,  -3,  -3,   0,  -1,  -1,   5,   4,   1,   0,   0,   0,   0,   0,  18,  20, -11,  -2,  -1,   0,   0,   0,  11,   3,  -4,  -1,  -1,   0,   0,   0,  38,  21, -35,   0,  28, -79, -12,  15, -31, -13,  38,  -1,   2,  16,  -1,  -6 },
     {    3,   9,  15,  -6,   1,  -3,   1,  -1,   0,   0,   0,  -1,   3,  17,  -6,  -2,  -1,   0,  -1,   0,   1,  -1,   0,   0,  10,  21,   6,  -3,   0,  -3,  -1,  -1,   0,  -1,   0,  -1,  -7,  75, -23,  -8,  -4,  -1,  -2,  -1,  -2,  -1,  -1,  -1, -50, -42, -16,  12,   2,   4,   0,   2,  26, -24,  24,  -1,   0,   0,   0,  -1,  -8,  21,  -4,  -4,   0,   0,   0,   0,  -4,  -4,   0,   2,   0,   1,   0,   0,  -8,  23,   1,  -4,  23,  -2, -16,   2,   0, -15,   2,   2,   3,   0,   1,   0 },
     {   11,  26,  94, -28,   1, -10,  -1,  -5,  -1,  -3,  -1,  -1,   3,   0,   9,  -8,   2,  -1,  -1,  -1,   0,   0,   0,   0,   3,  14, -18,   1,   0,   0,  -1,   0,   0,   1,   0,   0,   9,  15,   9,  -6,  -2,  -2,   0,  -1,  -1,  -1,   0,   0,  -4,  45, -17,  -4,  -1,  -1,   0,  -1, -38, -11, -12,   8,   1,   3,  -1,   1,   8, -10,   6,   3,   1,  -1,   0,   0,  -5,   9,   1,  -1,  -1,   0,   1,   0,   2,   2,   1,  -2,  -4,   9,  -1,  -3,   6,   6,  -9,   0,   1,  -4,  -4,   1 },
     {   -3,  -2,  -3,   1,   0,   0,   1,   0,   0,   0,   0,  -1,  -4,  -2,   6,  -1,  -1,   0,  -1,  -1,   1,  -1,   0,   0,  -7,   0,  -8,   4,   1,   2,   0,   0,   0,   2,   0,   0,  -8,  -4,   5,  -1,   1,  -1,   0,   0,   0,   0,   0,   0, -22, -13, -16,   7,   1,   2,   1,   1,  -5, -40,  17,   5,   3,   2,   0,   1,  12,  -5,   4,  -1,  -2,   0,   0,   0, -24, -22,  10,   2,   2,   1,   1,   1,  -9, -13,   1,   4, -32, -28,  42,   2, -26,  71,  19, -17,  19,  14, -30,  -3 },
     {   -6, -11, -66,  16,  -3,   8,   1,   3,   0,   2,   0,   0,   5,  18,  15,  -3,   0,  -1,  -1,  -2,   0,  -1,   0,  -1,   6,  26,   9,  -7,   0,  -1,   0,  -1,  -1,   0,  -1,  -1,  10,  36,   3,  -5,   0,  -2,  -1,  -1,   0,  -1,   0,  -1, -20,  51, -24,  -4,  -2,  -2,  -1,  -1, -46, -27, -19,  14,   1,   4,   0,   2,  19, -13,  20,  -1,  -1,   0,   0,  -1,  -7,   9,  -2,  -4,   0,   0,   0,   0,   4,   0,   1,   0,  -2,  12,  -3,  -3,  13,  -4, -10,   1,  -1,  -8,   1,   2 },
     {    1,   1,  -5,   6,  -3,   1,   2,   1,  -2,   1,   0,   1,  11,  24, 101, -19,  -4,  -7,  -4,  -3,  -1,  -2,  -1,   0,   0, -17, -26,  -3,   2,  -1,   1,   2,   1,   2,   1,  -1,   7,  21,  -1,   1,   0,   2,   0,  -1,  -1,  -1,   0,   0,   4,   0,  16,  -4,   1,   0,   0,  -2,  10,  26,  -6,  -9,   2,  -3,  -1,   0, -29,   9, -19,   5,   1,   1,   1,   1,   3,  -5,   9,   2,  -3,   1,  -1,   1,   4,   8,  -3,  -4,  -8,   4,   0,   0,   2,  17,  -7,  -4,   7,  -6, -11,   2 },
     {   -3,   0,  -4,  -6,   3,  -1,   0,   0,   1,   0,   0,   1,  -3,   7, -30,   7,  -1,   3,   0,   1,   1,   0,   1,   1,   2,  16,  20,  -6,  -2,  -1,  -2,  -1,   0,   0,   0,   0,  -4,  26, -10,  -2,  -1,  -2,   1,  -1,  -1,   0,   0,   0,  -2,  14,   9,  -1,  -2,  -1,   0,  -1, -28,  35, -30,  -3,   0,  -3,   3,  -1, -69, -30, -20,  24,   0,   6,   1,   3,  20, -30,  26,   2,  -2,   1,   0,   0, -22,   0,   4,  -2,   1, -16,  12,   4, -20,   7,  17,  -4,   3,  15,  -3,  -4 },
     {   -2,  -2,   8,  -3,   2,  -2,   0,   0,   1,   0,   0,  -2,  -1,   2,  37,   1,  -2,  -1,  -2,   0,   0,  -1,  -2,  -1,  -8, -17, -14,  -3,   3,   1,   1,   2,   0,   1,   0,   1, -12, -10, -23,   7,   0,   3,   0,   2,   0,   1,   0,   1, -14, -25,   4,   1,   1,  -2,   1,  -1, -18, -28, -21,  15,   1,   3,  -1,   2,   7, -35,  22,   0,   1,   1,   0,   0,   5, -10, -13,   2,   0,   1,   0,   0, -16, -31,  16,   6,  -3, -10,  -9,   7, -18, -46,  37,  10, -24,  42,  25, -12 },
     {    2,   5,  12, -21,   5,  -3,   1,   0,  -1,   2,   1,   0,   6,  17,  32,   6,  -8,  -1,  -4,   0,  -1,   0,  -2,   0,   6,  11, 100, -25,   3,  -6,  -1,  -3,  -3,  -1,  -1,  -1,  -2, -26, -30,  -4,   1,   1,   3,   1,   0,   0,   1,   1,   9,   8,  -9,   1,   3,   1,   0,   0,   6,   8,   0,   2,   0,  -2,   1,   1,   7,  15,   7,  -8,  -3,  -1,  -2,  -1, -14,   9,  -8,  -5,   2,  -2,   1,  -1, -18,   3,  -3,   5,   0,   5,  -2,   2,  -7,   7,   5,  -3,   3,  11, -10,  -2 },
     {   -2,   1,  -5,  -4,   4,  -2,   1,  -2,   1,   0,   1,  -1,  -3,  -1, -10,   1,   4,  -1,   2,   1,   1,   1,   0,   0,  -2,   2, -10,  -7,   1,  -3,   2,  -1,   0,   0,   0,   0,   3,   6, -10,   2,  -2,   2,   2,  -1,   1,   0,   0,   1,   1,  27,   3,   0,   0,   1,   0,  -3,   0,  11,   6,  -3,   2,  -1,   0,  -1, -27,  29, -29, -15,   2,  -3,   1,  -3, -80, -23, -45,  26,   2,   6,   0,   2,  21, -24,  25,  17,  -4,   2,   7,  -5,  -1, -14,   9,   4,  -4,   6,  12,  -3 },
     {    6,   5,  32, 107, -37, -13, -12,  -1,  -3,  -3,  -3,   0,   0,   4,  -1,   0,   0, -10,   2,   4,  -1,  -1,   1,  -1,   0,   6,  18, -26,  -1,   0,  -1,   1,  -1,  -2,   1,  -1,   0,   1,  17,   5,  -4,  -6,   3,  -2,  -1,   0,  -2,   1,  -3,  -3,   0, -11,   1,   2,  -2,   1,  -1,   0,   3,  -2,  -3,   2,  -4,   2,  -4,  -1,   1,   0,   0,   1,   2,   1,  -7,  -5,  -4,   0,  -1,   1,  -1,   1,  -7,  -3,  -2,   1,  -3,  -2,   4,   2,  -1,  -4,   1,  -2,  -7,   3,   3,   1 },
     {    0,   2,  -2,   5,   6,   4,  -6,   0,   1,  -1,   0,   1,   0,   4, -12,   6,  -5,   2,   0,  -1,  -3,   2,   2,  -1,  -1,  -5, -28,  -5,   4,   5,  -1,   0,   2,  -2,   2,   0,   4,  23, -22,  16,   0,   0,  -1,   0,   1,   0,   0,   0,   8,  20,  49,  -9,  -5,  -4,  -3,  -1,   2,  20,   0,  -4,  -5,   0,   1,   2,  10,  12,  38, -20,  -2,  -1,  -1,   1, -12,  24, -28, -24,  13,  -4,   3,   0, -64,   3, -21,  14,  20,  -3,   6,   7,  -8,  13,   8,  -6,  11,   9, -16,  -4 },
     {   -1,   0,  -3, -15,   2,   1,   0,   0,  -1,   0,   1,   0,   1,   5,   9,   3,  -3,   5,  -3,   2,   0,  -1,   0,  -1,   3,  13,  11,  21,   3,   0,   1,  -2,  -1,  -1,  -1,  -1,   3,  11,  93, -22,  -8,  -3,  -7,  -2,  -2,   0,   0,  -1,   1, -31, -24, -19,  15,  -3,   3,   0,   7,  26, -24,   9,   2,  -3,  -1,  -2,  -7,   4,  18, -12,  -3,   0,   1,   1, -16,   5, -17,  -7,   6,   1,   2,  -1, -22, -12,   2,  10,  11,   1,  -1,   8,  -5,  -2,   9,   1,   1,  10,   1,  -6 },
     {    1,   3,   2,  -5,  12,  -9,   9,  -4,   1,  -3,  -2,   1,   4,   1,   6,  94, -37,   0, -12,   2,   0,  -2,   0,   0,   0,   6,  -5, -27,  19,  -7,   2,  -6,   0,   0,   2,  -2,   4,  10,  21,  15,  -5,   4,  -5,   1,  -4,  -1,  -1,   0,  -2,   2,  30, -13,  -4,   0,   1,   1,   1, -14,   1,  -9,  -1,   0,   1,  -1,   3,   4,   7,   2,  -2,   0,   0,   1,   5, -13,   4,  -5,   3,   1,   2,  -1,  26,  12,  11,  -9, -18,   7, -12,  -1,   5,   2,   6,   4,   7,   5,  -3,  -7 },
     {   -1,  -1,   1,   6,   5, -14,   6,   3,   1,   2,  -6,  -1,   0,  -1,  10,  44,  -5,  -2, -14,  -6,   2,  -2,   3,   8,  -4,  -4, -16,   2,  27,   0,  -2,   0,  -2,   7,   4,  -7,  -5, -15, -15, -28,  -5,   8,   4,  -1,   0,   0,  -1,   2,  -3,  -2, -34,  31,   2,  13,  -4,  -4,  -4,   2, -12,  -1,  12,   3,  -7,   3,  -5,   5, -16,  14,  -2,  -6,  -1,   0, -15,  -4,   7, -12,  -1,  -2,   2,   0, -33, -21, -43,  15,  15, -23,  19,  10,  17, -12, -23,   3,  -8, -31,  16,  20 },
     {    0,  -1,   4, -11, -11,  15,  12,   3,  -5,   4,  -4,   0,  -1,  -1,   1,  -8,  -2,  -7,   0,  -2,   5,   1,   1,   6,   1,  -1,  21,   6,  -9,   1,  -2,   4,  -1,   4,   4,   0,   2,   6,  17,  20,  14,  -3,  -6,   6,  -1,  -6,  -4,   0,  -6,  -9,  47, -29,  -7,   3,   1,  -3,  -3, -46, -36, -27,  21,  -3,   1,  -5,  15,  10, -33,  35,   7,  -4,   3,   1, -12,  11,  -1,   2, -19,   4,  -2,  -1,  -9, -10,   0,  10,  32, -12,  20,  -9,  -7,  -1, -17,  -7, -12, -12,   6,  10 },
     {    0,  -2,   2,  -4,   0,  -1,  -3,   4,   1,   1,  -3,  -2,  -1,  -5,  -9,  -4,  10,  -8,   4,  -3,   2,  -2,   0,  -1,  -4, -16, -16, -55,   2,   5,   9,   1,   1,   2,   5,   1,  -2,  -8, -17,   9,   1,  -1,   4,   0,   1,  -1,   1,   1,  -2, -25, -24,  -7,  13,   0,  -1,  -2,   5,   8, -63,  22,   5,   3,   0,   0,  -1,  13,  33,  10,  -8,   0,  -1,   1,   3, -17, -20,  -1,  -4,   1,   1,   2,  24,  14,  18,   7,  28,  -5,   2, -11,   0,  15, -10,  -5,  23, -17, -18,   5 },
     {   -2,   0,  -4, -17, -12,  -3,   7,  -4,  -5,  -8,  -1,   5,  -4,  -4, -12, -22,  10,  -4,   0, -10,  -3,  -2,   5,   9,  -2,   1,  -2, -29, -12,  10,   8,   0,  -1,  11,   3,  -4,   1,   2,   5,   9,  -5, -14,   2,  -1,   0,  -2,  -5,  -1,  -2,   0,  12, -19,  14,   7,   2,   9,   8,  -7, -15,  18,   9,   4,  -8,   1,   0,  34,   3,  13,  -4,  -6,  -3,   4, -14, -11,  22,  -6,  -7,  -2,  -3,  -2, -20,   2, -35, -11, -64,  -7, -28,  26,  29,  -5,  19,  14,  -6,  -6,  16,  -5 },
     {   -5,  -9,   1, -28,-102,  19,  -7,  10,  10,   3,   6,  -5,   1,   0,   2,   7,  -4,  27,  -4,   0,  -3,  -4,  -5,  -7,   2,   1,   1,   8,  37,   3,   4,   0,  -4,  -2,  -2,   6,   1,   0,  -3,  17,  -4, -15,   9,   6,  11,   1,   0,  -4,   0,   1,   0,  10,   1,  11,  -6,   2,  -1,   3,   5,  -3,   6,   3,   0,   6,  -5,  -9,   4, -10,   0,  -4,  -3,  -7,   2,  -6,  -7,   3,   7,   4,  -4,   0,   2,  -2,   2,   2,   5,  -5,  -8,  -1,  10,   2,   5,  -3,   4,  -5,  -7,   3 },
     {   -2,  -2,  -4, -22, -17,   0,  26,   3, -27,  -1, -24,   9,  -1,  -1,   2,  -7,  -9, -26,   4, -13,   7,   2,  11,  19,  -1,   1,  -2, -40, -25,   4, -11,  12,  -2,  21,  13,  -7,   3,   2,  24,  37,   3,   7,  -5,   2,  -4, -12, -13,   2,   0,   3, -14,  27, -21,   6,  -8,  17,  -1,  14,  36,   4,  -2, -11,  -7,  -1,  -4, -19,   9,  -3,   5,   1,   4,   7,   1,  -8, -15, -19,  -1,  -4,   2,  -1,  -1,  -3,  -5, -17,  12,   1,  23,  -1,  -7,  -9, -15,  -5,  -9,  11,  11,   5 },
   },
   {
     {   76, -62,  18,  -7,   3,  -2,   1,  -1,   1,  -1,   0,   0, -62,  46,  -9,   2,   0,   0,   0,   0,   0,   0,   0,   0,  16,  -6,  -5,   4,  -2,   1,  -1,   0,   0,   0,   0,   0,  -2,  -3,   5,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   75,   9, -42,  12, -10,   2,  -3,   1,  -2,   0,  -1,   0,   9, -52,  40, -13,   6,  -2,   2,   0,   1,   0,   1,   0, -44,  37,  -8,  -1,   3,  -2,   1,  -1,   1,   0,   0,   0,  20,  -7,  -7,   6,  -4,   2,  -1,   1,   0,   0,   0,   0,  -8,  -1,   5,  -3,   1,   0,   0,   0,   1,   1,  -2,   1,   0,   0,   0,   0,  -4,   1,   2,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    9, -52,  53, -25,  14,  -6,   4,  -2,   2,  -1,   1,  -1,  72,  -2, -35,  11,  -7,   1,  -2,   0,  -1,   0,   0,   0, -51,  24,  -2,   5,  -3,   3,   0,   1,   0,   0,   0,   0,   7,  -4,   2,  -2,   1,   0,   0,   0,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {   49,   5, -16,   0,  -4,  -1,  -1,  -1,   0,   0,   0,   0,  49, -14, -13,   5,  -4,   2,  -2,   1,  -1,   0,  -1,   0,  25, -56,  32, -10,   5,  -1,   1,   0,   0,   0,   0,   0, -56,  47,  -9,  -1,   3,  -2,   2,  -1,   1,  -1,   1,   0,  15,  -6,  -6,   6,  -4,   2,  -1,   1,  -4,  -3,   5,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0 },
     {   -7, -34,   9,  10,   0,   2,   1,   1,   1,   0,   0,   0,  27, -21,  37, -30,  11,  -7,   2,  -3,   1,  -1,   0,  -1,  53,  25, -57,  23, -11,   2,  -2,   0,  -1,   0,   0,   0, -51, -15,  28,  -2,   1,   3,   0,   1,   0,   1,   0,   0,   7,  14,  -5,  -5,   3,  -2,   1,  -1,   1,  -6,  -1,   4,  -2,   1,   0,   0,  -1,   3,   1,  -2,   0,   0,   0,   0,  -2,  -2,   0,   1,   0,   0,   0,   0,  -1,   2,   1,  -1,   0,  -1,  -1,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {   19,  10, -20,  16, -12,   7,  -5,   3,  -2,   1,  -1,   1,  42,  25, -21,  -6,   1,  -4,   0,  -1,  -1,   0,   0,   0,  26, -31,   4,   5,  -2,   2,  -1,   1,  -1,   1,   0,   0,  37, -47,  26, -12,   5,  -1,   0,   0,   0,   0,   0,   0, -59,  40,  -8,   1,   3,  -3,   3,  -1,  22, -11,  -3,   4,  -4,   2,  -1,   1, -12,   1,   3,  -2,   1,   0,   0,   0,   3,  -1,  -1,   0,  -1,   0,   0,   0,  -5,   1,   1,   0,   2,  -1,   0,   0,  -3,   1,   0,   0,   1,  -1,   0,   0 },
     {   21,  35,  17, -57,  31, -22,  10,  -8,   3,  -3,   1,  -2, -13, -44, -12,  50, -26,  17,  -6,   6,  -1,   1,   0,   1,  43,  20,  -7, -13,   0,   0,  -4,   0,  -2,   0,  -1,   0,  -6, -22,   6,   0,   6,  -4,   4,  -1,   1,   0,   1,   0, -18,   9,   6,   0,  -3,   2,  -1,   0,   7,   3,  -6,  -1,   2,  -1,   0,   0,  -4,  -3,   2,   2,  -2,   1,   0,   0,   0,   1,  -1,  -2,   1,   0,   0,   0,  -2,   0,   1,   1,   0,   1,  -1,  -1,  -1,  -1,   0,   1,   0,   1,   0,  -1 },
     {    7,  36,  -5, -13,   2,  -4,  -1,  -1,  -1,   0,  -1,   0, -20,  11, -14,   8,   0,  -1,   2,   0,   1,   0,   1,   0, -57,  -5, -10,  24, -10,   9,  -3,   3,  -1,   1,   0,   1, -42,  -7,  49, -24,  13,  -4,   3,  -1,   0,   0,   0,   0,  30,  38, -36,   7,  -1,  -4,   1,  -2,  11, -27,   5,   3,  -5,   4,  -2,   1,  -4,   6,   3,  -4,   2,  -1,   0,   0,   5,  -3,  -1,   1,   0,   0,   0,   0,   1,   2,   0,  -1,   2,  -2,   0,   0,   0,   1,   0,   0,   2,  -1,   0,   0 },
     {  -23, -35,  16,   5,   5,   0,   3,  -1,   2,   0,   1,   0, -24, -29,  23,  -2,   2,   0,   0,   1,   0,   0,   0,   0, -20,  -6,  26, -18,  10,  -3,   2,  -2,   1,   0,   0,   0, -27,  26, -19,  11,  -2,   0,   2,  -1,   1,  -1,   1,   0, -50,  41,  -6,   5,  -3,   2,  -1,   1,  60, -22,  -4,  -1,  -2,   1,  -2,   1, -19,   1,   7,  -2,   2,  -1,   1,   0,   8,   5,  -4,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   5,   0,  -1,   0,  -1,   1,   1,   0,   3,   1,  -1,   0 },
     {    4,   6,  35, -19,  -4,   0,   0,  -2,   0,  -1,   0,   0, -11, -40,   5, -16,  32, -16,  11,  -4,   5,  -2,   2,  -1,  13, -27,  -2,  42, -34,  19,  -9,   5,  -2,   1,   0,   1,  47,  38,  -2, -35,  12,  -7,  -1,  -1,  -2,   0,  -1,   0, -10, -14, -22,  19,  -1,  -2,   4,  -1,  -4,  -7,  17,  -4,  -3,   5,  -3,   1,  -7,  10,  -5,  -2,   4,  -3,   2,  -1,   3,  -1,   2,   1,  -2,   1,   0,   0,  -1,   2,  -3,   1,   0,   0,   1,   1,  -1,   1,  -1,  -1,   0,   0,   0,   0 },
     {   14,  34,  31,  -1, -43,  29, -22,   9,  -7,   2,  -3,   1,  -9,  -1, -44,  -3,  43, -29,  18,  -7,   4,  -1,   2,  -1,   1,  25,   3,   3, -16,   6,  -2,  -2,   1,  -1,   0,  -1, -36,  -7, -12,  13,  -2,   6,  -4,   4,  -1,   1,   0,   1, -16,  -9,  32, -12,   6,  -3,   2,  -2,  22,  18, -17,  -1,  -2,  -1,  -1,   0,  -1, -13,  -1,   6,  -1,   1,   1,   0,   0,   3,   3,  -4,   0,   1,  -1,   0,   0,  -4,  -1,   1,   0,   2,   1,  -1,   0,  -2,  -1,   1,   1,   0,   1,   0 },
     {    0,  14, -11, -16,  28, -23,  14,  -9,   5,  -3,   2,  -1,   1,  23,  20, -18, -13,   9,  -7,   3,  -3,   1,  -1,   0, -28, -31, -27,  41, -11,  11,  -4,   5,  -1,   2,   0,   1, -19,  18,   1,  -4,   0,   0,   0,  -1,   0,   0,   0,   0, -45, -17,  37, -14,  11,  -4,   3,  -1,  12,  41, -25,   6,  -4,  -1,   0,  -2,  17, -21,   0,   2,  -3,   3,  -2,   2,  -9,   5,   2,   0,   2,  -1,   1,   0,   5,  -2,   0,   0,  -2,   1,   0,   1,   2,  -1,   0,   0,  -1,   1,   0,   0 },
     {   11,  24,  -6,  -9,  10, -11,   6,  -3,   1,  -1,   0,  -1,  18,  39,  -5, -10,  -8,   4,  -5,   0,  -1,  -1,   0,  -1,  14,  21, -31,  11,  -3,  -1,  -1,   0,   0,   0,   0,   0,  18,  29, -27,  10, -11,   2,   0,   0,  -1,   0,   0,   0,  27, -22,   8,  -7,   0,   3,  -3,   2,  60, -44,   1,  -2,   2,  -2,   1,   0, -45,   8,   9,  -2,   2,   0,   1,  -1,  13,  -1,  -5,   2,   0,   0,   0,   0,  -8,  -4,   2,   0,   3,  -2,   0,   0,  -4,  -2,   1,   0,   2,  -1,   0,   0 },
     {    5,   3,  33, -11,  -5,  -1,  -3,  -1,   0,  -1,   0,  -1,  -3, -24,  13,  -9,   8,  -1,   2,   0,   0,   0,   1,   0,  -4, -52,  -7,   2,  18,  -8,   9,  -2,   4,   0,   2,   0,  27, -13,   7,  27, -25,  13,  -8,   2,  -2,   0,  -1,   0,  47,  31,  11, -37,  11,  -6,  -2,   0,  -1,  -4, -38,  21,  -1,  -3,   4,  -1, -13, -22,  23,   0,  -3,   5,  -2,   1,  -2,  14,  -2,  -3,   1,  -2,   1,   0,   4,  -2,  -1,   0,  -4,   2,  -2,   0,  -1,  -1,   1,   1,  -1,   2,  -1,   0 },
     {    0,  -7, -10,  10,  -7,   2,  -1,   1,  -1,   1,   0,   1,   5,  -1,  -6,   5,  -5,   3,  -1,   1,   0,   0,   0,   0,  15,  25,   3, -10,  -2,  -1,  -1,  -2,   0,  -1,   0,  -1,  21,  22,  10, -28,  11,  -7,   2,  -2,   0,   0,  -1,   0,  23,   5, -28,  16,  -5,   0,   0,   0,  48,  17, -24,   5,  -9,   3,  -2,   0,  36, -66,  11,  -2,   2,   1,  -1,   1, -39,  15,  11,  -4,   3,  -3,   2,  -1,  13,  -3,  -8,   2, -10,  -4,   5,   1,   6,   0,  -2,   0,  -5,  -1,   2,   0 },
     {   12,  32,  31, -12,  -7,  -3,  -5,   0,  -2,  -1,   0,  -2,   2,  16,   0, -23,  -2,   6,   1,   0,   0,   0,   1,   1,   0,   9, -29,  -7,  36, -29,  14,  -8,   5,  -2,   2,  -1,   4,  30,   2,  22, -39,  21, -10,   4,  -3,   1,  -1,   0, -33,  16, -33,   8,   2,   3,  -4,   3, -15, -12,  29,  -6,   3,  -2,   2,   0,  40,  -2, -15,  -3,   2,   0,  -1,   0, -20,  -7,   4,   6,  -1,  -1,   1,  -1,   6,   5,  -1,  -2,  -3,  -3,   1,   2,   2,   1,  -1,  -2,  -1,  -2,   0,   1 },
     {   -3,  -8,   9,  23, -26,  14, -12,   8,  -5,   3,  -3,   2, -10, -23, -15,  19, -19,  22,  -9,   7,  -2,   3,   0,   2,  -5, -13, -25,   5,  40, -32,  19,  -9,   6,  -2,   2,  -1,   1,  14,  49, -31,  -7,   4,  -2,  -1,   0,  -1,   0,   0, -11, -48,   4,   7,   4,   3,  -2,   4,  17,  20,  -6,   1,  -3,  -2,   2,  -2, -32,  12,   2,  -4,   3,   1,   0,   0,  15,   2,  -2,   3,  -1,  -2,   0,   0,  -2,  -3,   3,  -1,   2,   3,  -2,   1,  -2,   0,   0,   0,   1,   1,   0,   0 },
     {   -9, -15, -29, -48,   1,  44, -24,  22,  -9,   8,  -2,   3,  10,  17,  30,  54,   0, -45,  21, -16,   5,  -5,   1,  -2,  -6, -12, -30, -17,   2,  11,   4,  -3,   5,  -2,   3,  -1,   7,  21,   8,   3,  -3,   6,  -9,   5,  -5,   3,  -1,   0, -11,  -3,  -2,  -1,   2,  -6,   4,  -1,  -4,  -7,  -5,   3,   1,   4,  -1,   0,   3,   4,  10,  -1,  -2,  -1,   1,  -1,  -3,   4,  -6,  -2,   2,   1,  -1,  -1,   4,  -4,   2,   3,  -4,  -2,   1,  -1,   2,   2,   1,   0,  -1,  -1,  -1,  -1 },
     {   -4,  -5, -17,   9,   0,  -5,   6,  -2,   3,   0,   1,   0,  -4,  -1, -26,   6,   0,  10,  -4,   5,  -1,   1,  -1,   1,  10,  32,  15,  12, -15,  -3,  -2,  -5,   1,  -2,   0,  -1,  -2,  34,  -1,  -9, -13,   8,  -5,   3,  -3,   0,   0,   0, -23,   4, -17, -31,  36, -16,   8,  -1, -37, -18, -19,  47, -22,  10,  -1,  -1,   2,   8,  42, -26,   2,   4,  -5,   2,   8,  13, -22,   2,   4,  -7,   4,  -2,  -1,  -7,   2,   2,   0,   2,  -1,   0,   1,  -1,   2,   1,   0,   1,  -2,  -1 },
     {   -5, -20, -30,  10,  10,   2,   0,   5,  -1,   3,   0,   0,  -6, -26, -38,  18,   3,   5,   6,   0,   2,  -1,   1,   1,   5,  -6, -10,  20,  -6,   4,  -4,   0,   1,   0,   1,  -1,  13,  16,   5,  11, -20,   6,  -4,   2,  -2,   1,  -1,   1,  16,  16,  13, -28,  16,  -5,  -2,   1,  31,  -2,  -2,   0,   3,  -7,   4,  -2,  49,  22, -38,   8,  -8,   2,  -1,  -1,  10, -50,  14,   5,  -3,   3,  -2,   3, -15,   6,  11,  -4,   5,  -3,  -4,   1,  -3,   0,   3,   1,   2,  -4,   0,   0 },
     {   -7, -16, -19,   8,   4,   0,   7,  -4,   2,  -2,   2,   0, -11, -25, -27,  10,   3,   1,   0,   3,   2,   1,   1,   1,  -9, -22, -14,  15,  -6,   6,  -5,   4,   0,   1,   0,   0,  -9, -31,  -6,  17, -12,  11,  -4,   4,  -2,   2,   1,   0, -10, -37,   8,  -3,  12,  -3,   4,   1,   5, -50,  38,  -8,   6,   1,   0,   1,   3, -12,  13,  -4,   5,  -2,   1,   0, -55,  35,  11,   0,   2,  -2,   2,  -2,  19,  11,  -9,   1,  -6,  -2,   3,   1,   7,   3,   1,  -1,  -2,   2,   0,   0 },
     {   -2,  -4, -12, -29,  27,  -6,   0,   3,   0,   1,   1,   0,   3,   5,  13, -14,  14, -11,   9,  -9,   3,  -3,   1,  -1,   6,   7,  49,  10,  -1, -16,   4,  -6,   2,  -4,   0,  -1,  -8, -27,  12,  -4, -37,  39, -17,  14,  -5,   3,  -1,   1,  11, -23, -29,   8,  35, -23,   9,  -2,  26,  33,  17,  -4, -17,   2,   1,  -4, -19,  -2, -12,  -2,   4,   3,  -3,   1,   3,  -3,   2,   3,  -2,  -2,   2,   0,  -2,  -1,   1,  -4,  -1,   1,  -1,   3,  -1,   0,  -1,  -3,   0,   0,   0,   1 },
     {   -4,  -7, -23,   8,  31,   2, -20,  15,  -9,   6,  -3,   2,  -6,  -6, -30,  -7, -12,   0,  16, -12,   9,  -3,   4,   0,   5,  16,  -1,  31,   4,  -9,   7,  -5,  -1,  -2,  -1,  -1,   2,  24,  -4,   1,  -8,   6, -11,   8,  -4,   0,   0,   0,  -4,  37,   2,  -3, -15,   9,  -3,   0, -29,  16,  -2, -27,  22,  -8,   6,  -2, -48, -22,   1,  26,  -6,   1,   2,  -1,   5,  15,  26,  -8,  -2,   1,  -2,   1,  20,  19, -21,  -6,   2,  -9,   4,   3,   0,   5,   0,  -3,   1,  -1,   0,   1 },
     {   -6,  -9, -19, -24, -39,   3,  45, -27,  19,  -7,   5,  -2,   4,   8,   5,  23,  43,   3, -45,  29, -19,   6,  -4,   3,  -1,   4,  -5,   1, -18, -21,  24, -14,   7,  -1,   2,  -2,   2,   9,   0,  -9,  -3,  14,  -2,  -1,   3,   0,  -1,   0,  -1,  16,   5,   3,   8,  -7, -10,   6, -11,  -3,  -7, -17,   5,  -5,  10,  -4, -14,  -5, -13,  23,  -1,   4,  -4,   4,   0,   2,  20,  -9,  -1,   3,   0,  -1,   6,   5,  -5,  -3,  -5,  -2,   3,   0,   1,   3,  -3,  -1,   0,   2,  -3,   0 },
     {    6,  12,  27,  11,  -6,   5, -10,   5,  -5,   3,  -3,   1,   4,  12,  33,   7, -26,   1,  -2,   0,  -1,   0,   0,   0,   2,  13,  27, -11, -10,   9,  -4,  -1,   0,  -1,   0,  -1,   2,  14,  19, -18,   9,  -8,   0,  -4,   1,   0,   0,   0,  -4,  -2,  13, -45,  22, -13,   7,  -1,  -2, -18,   0,  15,  -7,   2,   0,  -1, -20,  18, -34,  18,  -5,   3,   0,   1, -47, -11,  38,  -4,   3,   0,   0,   1,  12,  30,  -6, -11,   1,  -3,  -7,   4,   1,   6,  -1,   0,  -1,  -3,   0,   0 },
     {    0,  -1,   2, -22,  -6,  22, -27,  16, -10,   8,  -3,   3,   1,   1,  12,  -5, -20,  19,  -8,   3,  -1,   2,   0,   1,   3,   1,  18,  47,   1, -40,  32, -22,   8,  -7,   1,  -2,  -5, -20, -42, -37,  21,   9,  -5,  10,  -1,   1,   5,   0,  10,  15,  28,   5,  -6,   5,  -8,   2,   4, -18,   0,   0,  -8,   7,   4,  -4,  25,  14,   5,  -9,   3,  -5,  -2,   2, -12,  -1, -11,   8,  -1,  -4,   5,   0,   6,  -6,   3,  -3,  -1,   0,  -1,  -1,  -3,   0,   0,   1,   2,   1,   1,   0 },
     {   -3,  -2,  -5,  -9,   5,  -2,   4,   2,  -2,   0,   0,  -1,  -4,  -2,  -6,  -9,   7,  -5,   2,  -7,   8,  -3,   2,  -2,  -6,  -8,  -9, -10,   1, -10,  10,  -5,   2,   1,   0,  -1,  -9, -12, -17, -10,   9,   7,  -9,   8,  -4,   2,  -1,   2,  -9, -21, -16,  13, -13,  10,   0,   2, -15, -16, -25,  21,   1,   1,   5,   0, -11, -32,  17,   0,   5,   6,  -6,   4, -13, -45,  26,   1,   3,  -1,   4,  -2, -54,  50,  15,  -6,  31,   5, -17,   3, -10,  -9,   4,   2,  -1,   6,   1,   0 },
     {    6,   7,  21,  38,  -3,   5,  -3,  -9,   1,  -2,   1,   0,   1,   0,  11,  28, -31,  -4,   9,   6,  -6,   0,   1,  -1,  -3,  -8, -20,   0,  -5,   5, -12,   6,   4,   1,   0,   0,  -5,  -6, -34, -24,  -8,  23,   2,   4,   3,   4,   0,   0,   2,   9,  -1,  12,  35, -39,  15,  -4, -11, -12,   8, -31, -14,  19,  -3,   3,   7, -33,  -4,  19,   1,   5,  -1,   4,  37,   5,  13,  -7,   0,  -2,  -1,  -4, -25,  -7,   0,   0,  -3,   5,   7,   0,   0,  -2,  -7,  -5,   2,   3,  -1,   5 },
     {   -1,  -2,  -8,  -3,  -3,   4,  -1,  -1,   0,  -2,   2,   0,  -3,  -5, -17,   2,   4,   7,  -9,   0,   0,   2,   0,  -1,  -3,  -8, -32, -14,   7,   0,   0,  12,  -3,   5,  -1,   3,   2,   3, -29,  12,  19,  -1,   3,  -6,   4,  -2,   1,   0,  10,  26,   5,  13,   7, -17,  15,  -8,   0,  36,  21,  14, -50,  25, -10,   0, -31,   5, -12, -34,  28,  -2,  -5,   1, -17, -22,   7,  24,  -4,  -5,   5,   1,  29,  18,  -3,  -5,  -8,  -7,  -7,   5,  -5,   2,   5,   2,   2,  -4,  -5,   1 },
     {   -4,  -5,  -1, -16, -49,   4,   9,   7,  -7,   5,  -4,   3,   3,   5,  14,  23, -13,  10,  -7,  -9,   8,   1,   1,  -1,   2,   2,   7,  43,  21,   5, -31,  16, -11,   4,  -5,   1,  -5,  -9, -23,  10,  -6, -19,  26, -10,   4,  -2,   2,   1,  -2,   0, -36, -35,  11,  25,  -7,   5,   5,  23,  32,   2,   8, -12,   6,  -5,  -7, -22,   8,   2,  -6,   7,  -3,  -1,   7,  -7,  -6,   3,   7,  -7,  -1,   3,  -7,  11,  -5,  -2,  11,  -8,   1,   0,  -4,  -2,  10,   3,   3,   1,  -2,  -1 },
     {   -1,   0,   1,  12, -12,  -3,   6,  -7,   3,  -2,   0,   0,   0,   2,   3,  16, -12,  -6,  -1,   4,   4,   0,   1,  -2,   2,   7,   6,  19,  -5,  10,  -4,  -1,  -1,  -1,   0,   0,   3,  10,   4,   7,  -7,   8, -14,   8,  -4,   4,  -3,   2,   2,  17,  -1,   2,  -6,  -2,   4,  -6,  -7,  11,  -4, -22,  19,  -6,  -1,   9, -20,  -4,  -4,  -4,  16, -16,  10,  -3, -41,   1, -21,  33, -20,   4,  -3,   2, -13, -22,  61, -21,   6,  44, -27,  -9,  11, -15,  -7,   7,  -8,   8,   7,   0 },
     {    1,   2,   4,   3, -11,  -6,   3, -19,   8,  -7,   1,   0,   3,   5,   6,   7,  -3, -15,  -3,  -2,   8,   7,   3,  -3,   6,   9,  13,  11,  18,   7, -10,  18, -15,   4,  -5,   2,   5,   7,  14,  -5, -22,   7,   3,   3,  -2,  -7,   5,  -4,   8,   8,  26,  24,   3, -13,   2,  -7,   4,   6, -20, -44,  12,  22,   0,  -2,  14,  27,  36, -29, -16, -12,  -2,  -3, -12,  -2,  -7,  -2,  18, -12,   8,   2,   1,  43, -10,  -5,  19, -24, -14,  -3,  -4,   9,  17,   3,  16,   8, -10,  -1 },
   },
   {
     {  108, -47,  -7,  -4,  -2,  -2,  -1,  -1,   0,   0,   0,   0,  35, -21,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -18,  -1,   4,   1,   1,   0,   0,   0,   0,   0,   0,   0, -16,   6,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   32, -29,  10,  -3,   2,  -1,   1,   0,   0,   0,   0,   0,-108,  36,  14,   2,   3,   1,   1,   0,   1,   0,   0,   0, -16,  20,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  21,   0,  -6,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   9,  -4,  -2,   1,   0,   0,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   43,  58, -47,  -1,  -8,  -1,  -3,  -1,  -2,   0,  -1,   0,  -2,  56, -37,   4,  -6,   0,  -2,   0,  -1,   0,  -1,   0,  40, -26,  -6,   4,  -1,   1,  -1,   1,   0,   0,   0,   0,  16, -36,  11,   1,   1,   1,   0,   0,   0,   0,   0,   0,  -6, -11,   9,  -1,   1,   0,   0,   0,  -7,   1,   3,  -1,   1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,  -1,  -2,   1,   0,  -1,  -1,   1,   0,  -1,   0,   0,   0,  -1,  -1,   0,   0 },
     {   -2, -54,  30,  -1,   6,   0,   2,   0,   1,   0,   1,   0,   5, -12,  10,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  97, -10, -18,  -3,  -5,  -1,  -2,   0,  -1,   0,  -1,   0,  39, -17,  -9,   0,  -2,   0,  -1,   0,   0,   0,   0,   0, -10,  -8,   2,   1,   1,   1,   0,   0, -14,   1,   4,   0,   1,   0,   0,   0,  -7,   4,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0 },
     {    7,  15,  -9,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  26, -16,   1,  -2,   1,  -1,   0,   0,   0,   0,   0,   0, -30,  30,  -4,   0,   0,   0,   0,  -1,   0,   0,   0,   0,  93,   0, -25,  -1,  -6,   0,  -2,   0,  -1,   0,  -1,   0,  44, -36,  -9,   2,  -1,   1,   0,   1, -12, -17,   8,   2,   1,   1,   0,   0, -11,  -1,   6,  -1,   1,   0,   0,   0,  -2,   1,   1,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,  -2,  -1,   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {    5,  14, -17,  10,  -5,   2,  -2,   1,  -1,   0,   0,   0, -31, -61,  27,  11,   3,   4,   1,   2,   1,   1,   1,   1,  -5, -76,  44,   1,   6,   2,   1,   1,   1,   1,   0,   0,  27, -14,  22,  -8,   1,  -2,   0,  -1,   0,  -1,   0,   0,  11,  18,  -1,  -8,   0,  -2,   0,  -1,  -9,  12,  -5,  -2,   1,  -1,   0,   0,  -8,   7,  -2,   0,   0,   0,   0,   0,   1,   4,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   2,   0,   0,   0,   2,  -1,   0,   0,   1,  -1,   0 },
     {   11,   9,   3,  -8,   0,  -2,   0,  -1,  -1,   0,   0,   0,   1,  12,   4,  -7,  -1,  -1,   0,  -1,   0,   0,   0,   0,  37, -20,   6,  -9,   3,  -2,   0,  -1,   0,   0,   0,   0, -19,  44,  -8,  -5,   0,  -2,   0,  -1,   0,  -1,   0,   0,  80,  19, -33,   0,  -6,  -1,  -2,   0,  41, -34, -15,   5,  -2,   1,  -1,   1,  -3, -23,   5,   3,   1,   1,   0,   1,  -6,  -3,   4,   1,   1,   0,   0,   0,  -1,   0,   0,   0,   0,  -2,  -1,   1,   1,  -2,   0,   1,   0,  -2,   0,   0 },
     {   19,  42,  64, -60,   6, -13,   1,  -6,   0,  -3,   0,  -2,   9,  14,  43, -46,   9,  -9,   3,  -3,   1,  -2,   0,  -1, -12, -25,  -9,   3,   6,   0,   3,   1,   1,   0,   1,   0,   2, -17, -13,  14,   0,   2,   1,   2,   0,   0,   0,   0,  -8,   3,  -3,   7,  -1,   1,  -1,   0,  -8,   8,  -2,   1,   0,   0,   0,   0,   1,   2,  -4,   1,   0,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,  -1,   0,  -1,   1,  -1,   0,  -1,   1,  -1,   0,  -1,   0,   0,   1,  -1,   1 },
     {  -19, -45,  32,  -2,   8,  -3,   3,  -1,   2,   0,   1,   0,  16,  33, -34,   8,  -3,   0,  -1,   0,   0,   0,   0,   0, -33, -41,  -2,  19,   0,   5,   0,   2,   1,   1,   0,   1,  -3, -50,  33,   5,  -1,   3,   0,   1,   0,   0,   0,   0,  50, -20,  17,  -8,  -2,   0,  -1,   0,  11,   0,   5,  -6,   1,  -1,  -1,   0,   0,   4,  -2,  -2,   1,   0,   0,   0,  -2,   2,   1,   0,   1,   0,   0,   0,   1,   3,   0,  -1,  -1,   1,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0 },
     {   -4,  -1,  -2,  -2,   1,  -1,   1,   0,   1,   0,   0,   0,  -9,   7,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   2, -18,   5,  -1,   3,   0,   1,   0,   1,   0,   0,   0, -37,  31,  -8,   5,   0,   1,   0,   0,   0,   0,   0,   0,  15, -36,  -8,   9,  -1,   2,   0,   2, -82, -28,  32,   5,   4,   2,   2,   1, -37,  29,  20,  -4,   1,  -1,   1,  -1,   1,  20,  -2,  -4,   0,  -1,   0,  -1,   6,   5,  -5,   0,  -1,   0,   0,   1,   0,   2,   0,   0,  -1,   1,   0,   0 },
     {   -5, -15,  19,   6,  -4,  -1,   0,  -1,   0,   0,   0,   0,  22,  54, -12,   2, -13,  -1,  -4,  -1,  -2,  -1,  -1,   0,  -9,  -9,  42, -22,   2,  -1,  -1,  -1,   1,  -1,   0,  -1,  34,  39,  11, -30,   5,  -4,   0,  -3,  -1,  -2,  -1,  -1, -11,  37, -24, -12,   7,  -3,   2,  -2, -41,  29, -20,   3,   4,   0,   1,   0,   9,   2, -16,   5,  -1,   0,   0,   1,   5, -10,  -3,   3,  -1,   0,   0,   0,   4,  -6,  -1,   2,  -2,  -2,   0,   1,   1,  -1,  -1,   1,  -1,  -1,   0,   1 },
     {    0,  -3,  -5,  21, -14,   4,  -4,   2,  -1,   1,  -1,   1,  12,  18,  53, -17, -11,  -2,  -3,  -2,  -2,  -1,  -1,  -1,  10,  28,  56, -50,   3,  -6,  -1,  -3,  -1,  -1,  -1,  -1, -20, -44,  17, -20,  12,   1,   3,   1,   1,   1,   1,   0,   6, -41,  -4,   6,   7,   1,   1,   2,  14, -11,  -2,   7,   0,  -1,   0,   0,  -4,   6,   0,   1,   0,  -1,   0,   0,  -3,   3,  -4,   1,   0,   0,   0,   0,   0,   1,  -2,   1,   1,   0,  -2,   1,   0,   1,  -1,   1,   0,   1,  -1,   0 },
     {   15,  25,  49,  48, -50,   8, -16,   1,  -6,   0,  -3,   0,   3,   2,  24,  48, -46,  13, -13,   4,  -4,   0,  -2,   0,  -5,  -9, -34,   6, -10,  10,   1,   4,   1,   1,   0,   1,  -8,  -3, -22, -12,   9,   3,   3,   2,   1,   0,   0,   0,   6,   2,   2,  -9,  10,  -3,   1,  -1,  -1,   2,   7,  -6,   5,  -3,   0,  -1,   0,   1,   2,  -3,   2,  -1,   0,   0,  -1,  -1,  -1,  -2,   2,  -1,   1,   0,   0,   0,  -1,  -1,   1,  -1,   0,  -1,   1,   0,   0,  -1,   0,  -1,   0,  -1 },
     {   -3,   0,   6,  -3,  -1,   2,   1,   0,   0,   0,   0,   0,   2,   6,   1,  -1,  -6,   1,   1,  -1,   0,  -1,   0,   0,  -9,  20,  -5,   6,  -8,   1,  -2,   1,   0,   0,   0,   0,   9, -24,  27,  -5,   1,  -2,   0,  -1,   0,   0,   0,  -1, -17,  53,  -9,  -8,   2,  -2,   1,  -1,   2, -33, -16,   5,   2,   1,   1,   2, -85, -15,  24,   6,   6,   1,   1,   1, -20,  25,  14,  -3,   0,  -1,   0,   0,  -4,  14,  -1,  -3,   2,   3,  -3,   0,  -5,   1,   0,   1,   0,   3,   0,   0 },
     {    7,  18,   0, -20,  -5,   2,   2,   1,   0,  -1,   0,   0,  -2, -15,  13,   1,  -4,   2,   1,   0,   1,   0,   0,   0,  21,  49,  -5,  19, -18,  -3,  -5,  -2,  -1,  -1,  -1,   0, -12,   3,  48,  -2, -13,  -1,  -5,   0,  -1,   0,  -1,   0,  47,  13,  25, -31,  -2,  -3,  -2,  -1, -22,  33,  -9, -23,   9,  -3,   2,  -1, -12,  33, -28,   0,   5,  -1,   1,   0,  13,  -8, -16,   6,   0,   0,   0,   0,   0,  -8,   2,   2,   1,  -2,   0,  -1,  -1,  -1,  -1,  -1,   1,   0,  -2,  -1 },
     {  -14, -27, -39, -13, -43,  45,  -8,  11,   0,   4,   0,   3,   6,  14,  18, -39, -29,  29,  -8,   6,   0,   0,   1,   0,  -5,  -7,  -6,  18,   7,   1,  -2,  -3,  -1,   1,   0,  -1,   2, -13, -23,  46,   6, -11,   0,   0,  -1,   0,  -1,   0,  11,  25, -20,  17,  -2,  -7,   1,   1,  -2,  15,   0,  -5,  -1,  -2,   1,   0,  -2,  21,  -3,  -5,   1,  -1,   1,   0,   5,  -2,  -5,   2,  -1,  -1,   1,  -1,  -1,  -3,   3,   3,   0,  -2,   3,   0,  -1,   1,   3,   0,   1,   1,  -1,   1 },
     {    1,  -2, -17,  48,  25, -23,  -1,  -5,   0,   0,  -1,  -1,   4,   7,  13, -12,  31, -23,   6,  -5,   1,  -2,   1,  -2,  -8,  -3, -45, -44,  24,   0,  10,   0,   3,   0,   3,  -1,  -8, -24, -25,   1,  10,   7,   4,   1,   2,   0,   1,   1,  30,  39,   4,   9,  -7,   1,  -3,  -2, -17,  31,   4,   0, -10,  -1,  -2,  -1, -16,  12,  -8,  -1,  -3,   1,   0,   1,   2,  -5,  -4,   5,  -2,   2,   0,   0,   0,  -5,   4,   3,  -1,  -2,   3,   0,   1,   0,   2,   0,   0,   0,   0,   1 },
     {   -2,   2,   4, -18, -15,  10,  -1,   3,   1,   2,   0,   1, -13, -18, -27,  -7,  -8,  14,   0,   5,   1,   2,   1,   2,   1,   5,  -1, -25,  10,   0,   1,  -1,   1,   0,   0,   0, -13,   1, -12,  -1,  14,  -3,   1,   0,   1,  -1,   0,   0,  13, -26,  10,   4,   8,  -4,   0,  -1, -19,  45, -19,   0,   0,  -4,   0,  -1, -15, -40, -17,  12,   1,   2,   1,   2, -72,  -9,  26,   9,   4,   1,   1,  -1, -14,  17,  11,   1,  -5,   9,   0,   0,   2,   2,  -2,   1,  -5,   1,   2,   1 },
     {   -2,  -4, -16,  17,  26, -20,   7,  -5,   3,   1,   0,  -1,  11,  25,  39,  27,   5, -20,   1,  -5,  -2,  -2,  -1,  -2,   2,  -2,  30,  47, -34,   6,  -8,   0,  -2,  -1,   0,  -2,  -6,   0, -30,  26, -34,  16,  -3,   4,   0,   0,   0,   1,  -2,  -7, -18,  -4, -10,  10,   0,   3,   3,  28,  -2, -17,   4,   0,  -1,   0, -12,  -6,  -2,  -5,   7,  -2,  -1,   2, -39,  -5,  11,   1,   4,  -1,  -1,   0,  -7,   4,   4,  -2,  -2,   2,   3,  -3,   1,   3,   0,  -2,  -1,   1,  -1,   0 },
     {    0,  -1,   0,  18,  -9,   6,  -5,   0,  -1,   0,   0,   0,   5,  19,  12, -29,   4,   0,  -1,   1,   1,   0,  -1,   0,  -1, -16,  -2,  -9,  13,   1,   4,   0,   2,   0,   0,   0,  20,  49,  13,   0,  -9,  -5,  -1,  -1,  -2,   0,  -2,  -1, -12,  -3,  60,  -7,  -7,   1,  -3,   3,  35,  -1,  31, -25,  -3,  -3,  -2,  -1, -13,  18, -12, -19,   5,  -2,   3,  -1, -41,  30, -15,  -1,   6,  -1,   2,   0,  -2,  13, -10,   3,   7,  -6,  -4,   1,  -1,  -3,   1,   0,  -3,   1,  -1,   2 },
     {   -3,  -8, -18,  23,  -6,   5,  -1,  -2,  -1,   0,   0,   1,   4,   9,  40, -27,   4,  -4,   1,  -3,  -1,   0,  -2,  -1,  -6,  -6, -26,  18,  -1,  -1,  -2,   1,   0,   0,  -1,   0,   3,   3,  36,   1, -10,  -2,  -4,   3,  -1,  -1,  -1,   0,   1,   5,  22, -18,  -2,   1,   0,   0, -43, -39,  -1,   6,   5,   4,   0,   2,  38, -56,  -1,  10,  -1,   3,   0,   2,  -7, -26,  15,   5,  -1,   2,   0,   1, -11,  11,  12,  -4,  -6,   6,   1,  -2,  -1,   7,  -1,  -2,  -2,   0,  -2,   0 },
     {   -4,  -7, -15,   7,  -5,  -3,   2,   0,   2,   3,  -1,   0,  -2,   1,   6, -24,  13,  -6,   4,   0,   2,  -1,   1,  -2,  -7, -17, -39,  17,   1,  11,  -2,   3,   1,   0,   2,  -1,  -2,   7,   5, -38,   1,   4,   1,   4,  -1,   1,  -2,   0, -15, -32, -39, -32,  25,   3,   7,   0,  16,  -7, -38,  12,  25,  -3,   2,  -3,  -5,  47, -18,  23,  -1,  -8,  -1,  -1, -20,   4,  -3,  12, -10,  -1,   1,   1, -14,  -2,   6,   4,  -2,  -5,   6,   2,  -2,   2,   4,   0,  -3,   1,  -1,   1 },
     {    7,  12,  33,  40, -22,   0,  -8,   3,  -4,  -3,  -2,   1,  -6,  -9, -26, -13,  17, -10,   6,   1,   2,  -2,   1,   1,   2,   5,   8,  -6,   1,  -2,   1,  -1,   0,   0,   0,   0,   6,   7,  36,  44, -21,   5,  -6,  -2,  -4,  -1,  -1,   0, -10,  -7, -27,  37, -26,   9,  -5,   3,   1, -14, -16,  -1, -11,   8,   0,   1,  -4,  25, -15, -10,   2,   3,   1,  -1,  -8, -27,   2,   6,   3,   3,  -2,   1, -33,  -9,  21,   3,  -4,  14,   8,  -5,  -3,   4,  -1,  -3,  -1,   2,   2,   0 },
     {    6,   7,   8,  26,  -6,  -5,  -6,   3,  -1,   0,  -3,  -1,   0,  -4,  -2, -16,   9,  -6,   1,   3,   1,   1,   0,   0,   6,  11,  -7,  17,   0,  -2,  -2,  -3,  -2,   0,  -1,   0,   3,   1,  25,  19,  -3,  -1,  -4,  -2,  -1,  -1,   0,   1,   9,  -5, -16,  16,   2,  -4,   1,  -1,  -9,  26, -32,  20,  -5,   0,   0,   3,  31,  -1,  21,   1,  -5,   1,  -1,  -1,  -9,  61,  19, -12,  -3,  -3,  -1,  -3,  49,  22, -27,  -7,  12, -19, -16,   0,  14, -12,  -6,   1,   3,  -5,   2,  -2 },
     {    1,   5,   2, -11,   4,   7,  -1,   2,  -2,   2,  -1,   0,   2,   1,   4,  13,  11,   0,  -1,  -1,   1,  -1,  -1,  -1,   4,  16,  -1,  18,  23, -19,   2,  -4,   0,   0,   1,  -1,   1, -11,  16, -26,  19, -13,   5,  -5,   2,  -1,   3,  -2,   8,  23, -30,  -8,  -5,   0,   6,  -4,   0,   9,  48,   7, -14,   4,  -1,   0,  35,  32,  35, -17, -16,  -1,  -6,   1, -41,  -8,   9, -18,   2,   0,   1,   2, -36,  24,   2,  -7,   4,  12, -11,  -2,  -2,   3,  -5,   0,  -1,   2,  -2,   0 },
     {    0,   1,  -5,  -2,  10, -17,  19,  -7,   5,  -1,   3,  -1,  -5, -10, -11, -29, -32,  12,  14,   3,   2,  -1,   1,   0,  -4,  -2,  -7, -33, -70,  46,  -5,  10,   1,   4,   0,   1,   2,  -1,   7,  -6,  -6,  28, -12,   3,  -1,   1,  -2,   0,   7,  19,  -4,  19,  24,   2,  -6,  -3,  -2,  -2,  20,   5,   8, -11,  -3,  -1,  21,  13,  10, -14,  -4,  -8,   0,  -2,  -9,   4,   6,  -6,   3,  -1,  -2,   0, -16,  10,   1,  -3,   2,   1,  -5,  -3,   1,  -1,  -2,  -2,  -1,   0,  -2,   1 },
     {   -6,  -9,  -8,  -6, -54, -46,  48,  -6,   7,   0,   5,  -3,  -3,  -4,  -4,  -5, -28, -40,  47, -13,   5,  -1,   4,  -4,   4,   5,   5,  16,  33,   4,   3, -10,  -2,  -1,   1,  -1,   1,   2,   2,   5,  20,  19, -18,  -2,   0,  -1,   1,   1,  -2,   1,  -2,  -9,   1,  10, -13,   0,   3,   5,   8,   0,  -7,   1,  -5,   3,  -4,  -9,   7,  -4,  -3,   1,  -1,   3,  -1,  -3, -10,  -3,  -1,   1,   0,   2,   6, -12,  -7,  -1,  -3,   1,   6,  -3,   3,   2,  -1,  -2,  -2,  -2,   1,   1 },
     {   -2,  -1, -13, -12,  28, -22,  20,  -7,  10,  -7,   4,  -2,   6,  10,  30,  35, -35,   6,  -6,   0,  -1,   0,  -2,  -1,  -3,  -6, -16,   4,   6,   0,  -5,   1,   0,   1,  -1,   2,   9,  20,  23,  14,  35, -26,   2,  -7,   0,  -5,  -1,   1,   0,  -9,  22,  47,   1, -16,   3,  -6,   4,  -3, -26,  29, -20,   3,  -1,   1,  -4,  22, -15,   7, -12,   7,   1,   3,  -5,   2,  11,  -4,  -5,   4,   0,   1, -11,   0,  23,  -4,  -1,  -2,   0,  -9,  -3,  -2,   3,  -3,  -1,   0,  -2,  -4 },
     {    0,  -1,   1,  10,  -4,  12,  -6,   4,  -8,  -5,   4,   4,   1,   4,   8, -22,  14,  -5,  -2,  -3,  -4,   5,  -6,   2,   0,  -7,  -9,  20, -11,   4,  -5,   6,  -2,   0,  -6,   5,   1,  19,  22, -26,  -2,  -6,  -6,   3,   5,   6,   0,  -1,   1, -18,   0,  25,  19, -14,   3,   6,  17,  43,  19,  22,  -9, -19,   3,  -1, -34, -14,  40, -10,  -2,   0,   3, -10,  16, -43,   9,  -6,   1,  14,  -5,  -2,  25, -25,   2, -15,   2,  13,   5, -12,  -9,   7,  -1, -10,  -8,  -3,   2,  -4 },
     {    1,   0,   0,   6,  -8,  -4,  13,  -2,   3,   3,   1,   1,   3,   3,  -2,  -2,  11,   0,   8,   5,   4,  -1,  -1,   0,   2,  -2,   8,   7,   0,   3,   0,   3,   0,  -2,   1,   2,   1,   2, -12,   3,  10,  -2,  -2,  -1,   1,  -2,   1,   1,  -1,  -3,   4,   0,   9,  -2,   5,  -1,   2,  -6,   1, -14,   5,   0,  -1,  -3, -11,  10, -12,   4,  -2,   3,  -2,  -4,  23, -31,  15,  23,  -7,   9,  -4,   1,  10,  59,  16,  -9,  64,  19, -45, -11,  11, -34, -13,   8,   0, -11,   8,   6 },
     {    1,   2,   4,   4, -10,  14,   5, -14,  15,  -2,  -2,  -4,   0,  -2,  -2,   2,  21,   5,   3,  -5,   0,  -9,   1,  -2,   1,   6,   6,  31,  -5,  12, -12,  -3, -12,  -5,   0,  -1,  -2, -11,   1, -35,   8,   0,   3,   5,  -4,  -8,  -3,   9,   2,  13,  -8,  50,  -6,   2, -14,  -4,  -4, -11,  31,  19,  -6,   0, -19,   5,  -5, -10, -54,  13,   9, -14,  10,  -1, -15, -10, -14,  13,  -7,  -9,   8,  -5,  16,   9, -24,   7,  -1, -28,   4,  17,   2,  -4,  17,   6,   0,  -1,  -1, -11 },
     {   -4,  -7, -11, -22, -18,   4,  -1,  10,   0, -13,   6,  -2,   2,   4,   8,  12,  -9, -10, -23,  11,   1,   8,  -1,   6,  -3,  -6, -15, -15,   9,   7,  -2, -14,  -3,   2,  -4,   4,   0,   3,   3, -19, -39,  22,  -2,   4,  -6,  -3,   1,   0,   2,  -1,  10,   9, -38,  50, -12,  14,   3,  18,  -8,  39,   1,  19, -10,  -9,  -4,   2,   5,  17,  16,  -8,   0, -19,  14,   2,  -2, -21,  10, -15,   9, -10, -14,  18,  -1, -13,  14,  14, -10,   5,   0,  -1, -10,  -1,  -1,  -1,  -7,  -2 },
   }
 },
 {  //15
   {
     {  119, -42,  -5,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0,  -6,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -16,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -6,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {    5,   8,  -3,   0,  -1,   0,  -1,   0,   0,   0,   0,   0, 118, -28,  -9,  -4,  -3,  -2,  -2,  -1,  -1,  -1,  -1,   0, -13, -10,   3,   0,   1,   0,   1,   0,   0,   0,   0,   0, -30,   2,   3,   1,   1,   0,   1,   0,   0,   0,   0,   0, -11,   3,   1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -4,  -1,   1,   0,   0,   0,   0,   0,  -5,   2,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,  -2,   1,   0,   0 },
     {   13, -10,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  12,  11,  -5,   0,  -2,   0,  -1,   0,  -1,   0,   0,   0, 118, -20, -10,  -3,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   0,   4, -20,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0, -26,  -3,   4,   1,   1,   0,   0,   0, -13,   1,   2,   0,   1,   0,   0,   0,  -7,   1,   1,   0,   0,   0,   0,   0,  -6,   0,   1,   0,   0,   0,   0,   0,  -4,   1,   0,   0,  -3,   0,   0,   0,  -2,   1,   0,   0,  -2,   0,   0,   0 },
     {    6,   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  25, -11,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0,   0, -13,  20,  -3,   1,  -1,   0,   0,   0,   0,   0,   0,   0, 112, -19, -11,  -2,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0,  -8, -26,   4,   0,   2,   0,   1,   0, -29,   0,   5,   1,   1,   0,   1,   0, -13,   4,   2,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0,  -3,  -1,   1,   0,  -2,   1,   0,   0,  -2,   0,   0,   0,  -2,   1,   0,   0 },
     {   -3,  12,  -4,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0, -14,  -9,   3,   0,   1,   0,   1,   0,   0,   0,   0,   0, -20,   6,   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,  -4, -23,   4,   0,   1,   0,   1,   0,   0,   0,   0,   0,-109,  15,  16,   2,   4,   1,   2,   0,   6,  34,  -4,   0,  -1,   0,  -1,   0,  31,   4,  -7,   0,  -2,   0,  -1,   0,  13,  -5,  -3,   0,  -1,   0,   0,   0,   1,  -3,   1,   0,   3,   2,   0,   0,   3,  -1,   0,   0,   2,   0,   0,   0 },
     {   37, 110, -40,  -2, -13,  -2,  -6,  -2,  -3,  -1,  -2,  -1,  -7,  18,  -7,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1, -19,   1,   0,   2,   1,   1,   0,   1,   0,   0,   0,  -1,  -5,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   8,  -5,   0,   0,   0,   0,   0,   0,   2,  -7,   1,   0,   0,   0,   0,   0,  -6,  -2,   2,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0 },
     {   -4,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -9,   3,   0,   1,   1,   1,   0,   0,   0,   0,   0,   0, -10, -19,   6,   0,   2,   0,   1,   1,   0,   0,   0,   0, -28,   7,   3,   1,   1,   1,   0,   0,   0,   0,   0,   0,   3, -29,   4,   0,   1,   0,   1,   0,-104,  16,  19,   2,   4,   1,   2,   0,   6,  42,  -3,  -1,  -1,  -1,  -1,  -1,  24,   3,  -7,  -1,  -2,   0,  -1,   0,   8,   0,  -2,   0,   3,   0,   0,   0,   4,   1,  -1,   0,   1,  -2,  -1,   0 },
     {   -5,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -16, -28,   6,   2,   3,   1,   2,   1,   0,   0,   1,   0,  -9,  -5,   1,   2,   1,   1,   0,   0,   0,   0,   0,   0, -13, -21,   7,   1,   2,   0,   1,   0,   0,   0,   0,   0, -26,  11,   3,   2,   1,   0,   0,   0,   2, -32,   6,   1,   1,   0,   0,   0, -97,  14,  22,   0,   5,   0,   2,   0,   4,  44,  -2,  -1,  -1,  -1,  -1,  -1,  19,   6,  -8,   0,   8,   0,  -2,   0,   0,   0,   0,   0,   3,   3,  -1,   0 },
     {   -4,  -4,   2,  -4,   2,   0,   0,  -1,   0,   0,   0,   0,  20,  95, -18,  -7,  -6,  -3,  -5,  -3,  -2,  -1,  -1,  -1,  -8,  66, -17,  -4,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0, -18,  -8,  -8,   1,   1,   2,   1,   1,   0,   0,   0,   0, -16, -13,   0,   2,   2,   2,   0,   0,  -6, -15,   3,   1,   1,   0,   0,   0, -19,  -1,   5,   1,   2,   0,   1,   0,   4,   5,   0,   0,   0,  -1,   0,   0,   5,  -2,  -2,   0,   2,  -1,   0,   0,   2,  -2,  -1,   0,   0,  -3,   0,   1 },
     {   -4,  -2,   3,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -8,  14,  -3,   1,  -2,   1,   0,   0,   0,   0,   0,   0, -14, -24,   6,   1,   3,   0,   1,   1,   1,   0,   0,   0,  -7,  -1,   0,   1,   1,   1,   0,   0,   0,   0,   0,   0, -14, -23,   9,   1,   1,   1,   1,   0, -21,  10,   2,   0,  -1,   1,   0,   0,   2, -40,   7,   0,   1,   1,   1,   1, -97,  17,  25,   0,   5,   0,   2,   0,   4,  41,  -3,  -2,   7,   4,  -6,   0,   7,   3,  -2,   1,   0,  -1,   0,   0 },
     {   -5, -21,  10,   0,   2,   1,   2,   0,   0,   0,   1,   0,  10,  54, -21,   1,  -4,  -1,  -2,   0,  -1,   0,   0,  -1, -25, -81,   7,   9,   5,   3,   2,   3,   1,   2,   1,   1,  18, -53,  19,   2,   3,   1,   1,   0,   0,   0,   1,   0,  11,  17,   6,  -2,  -1,  -2,   0,  -1,  17,   8,   0,  -2,  -2,  -1,   0,  -1,   0,  10,  -2,  -1,  -1,   0,   0,  -1,  16,   4,  -5,  -1,  -1,   0,  -1,   0,   7,  -7,  -3,   1,  -5,  -3,   1,   1,  -3,   1,   1,   0,  -1,   2,   1,   0 },
     {   -4,  -4,  -7,   1,   0,   1,   0,   1,   0,   0,   0,   1,  -7, -13,   0,   1,   2,   0,   0,   0,   0,   0,   0,   0, -10,   7, -11,   4,  -1,   1,   0,   1,   0,   0,   0,   0, -20, -55,   7,   4,   3,   3,   2,   1,   1,   1,   1,   1,   5, -22,   7,   2,   2,   0,   0,   1,  -2, -26,  11,   1,   2,   0,   1,   0,  -9,   7,   3,   1,   0,   1,   0,   0,  -5, -41,   8,   2,   2,   1,   1,   1, -82,   8,  28,   0,  -4,  37,   2,  -3,  -2,  11,  -2,  -1,   5,   8,  -4,   0 },
     {    1,   9,  29,  -9,   1,  -4,   0,  -2,   0,  -1,  -1,  -1,   2,  25,   6,  -6,  -1,  -2,  -1,  -1,   0,   0,   0,  -1,  -4, -19,   5,  -3,   2,   2,   1,   0,   1,   0,   0,   0,  19,  56,   2, -10,  -3,  -4,  -1,  -3,   0,  -1,  -1,   0, -21,  69, -22,  -7,  -3,  -1,  -1,  -2, -29, -11, -11,   1,   1,   3,   0,   1, -28,  -2,  -5,   2,   1,   1,   1,   0,  -7, -23,   0,   3,   2,   1,   1,   1, -45,   5,  11,   1,   2,  14,  -3,   0,  -1,   0,   0,   0,   1,   4,  -1,   0 },
     {  -16, -33,-105,  27,   1,  12,   1,   7,   2,   3,   1,   3,  -2,   5, -32,   9,  -4,   4,   0,   1,  -1,   1,   0,   1,   0,   0,  27,  -2,  -2,  -2,   1,  -1,   0,  -1,  -1,   0,   7,  21,  11,  -6,  -1,  -3,   0,  -2,   0,  -1,  -1,   0,  -6,  16,  -3,  -3,   0,  -1,  -1,   0,  -6,  -2,  -1,   1,   0,   1,   1,   0,  -8,   0,   1,   1,   1,  -1,   1,   0,  -4,  -6,   0,   1,   1,   0,   0,   1,  -5,  -3,   4,   1,  -2,   3,   1,   0,   2,   0,   1,   0,  -1,   0,   1,  -1 },
     {   -1,   1,   7,   1,  -2,   0,  -1,   0,   0,   0,   0,   0,  -4,   3,   1,   0,  -2,   2,  -1,   0,   0,   1,   0,   0,  -3,  -9,   6,  -2,   1,   0,   0,   1,   0,  -1,   0,   0,  -5,  16,  -8,   1,  -2,   1,   0,   0,   0,   0,   0,   0,  -7, -21,   8,  -2,   1,   0,   1,  -1,  -2,  15,  -9,   2,  -1,   0,   0,  -1, -17, -27,   8,   2,   2,   0,   1,   1,  -5,  13,  -7,   0,   0,   1,   0,   0,  -1, -60,   8,   5, -79,  30,  31,  -2,  25,  32, -13,  -3,  -3,  -6,  -2,   1 },
     {   -2,   1,  -5,   1,   0,   2,   1,   1,   0,  -1,   0,   0,  -6, -12,  -4,  -2,  -3,   0,  -1,   2,   2,   0,   0,  -1,  -6,  12, -26,  -3,  -2,   0,   0,   1,   0,   1,   0,   0,  -8, -30,   9,   6,   3,   1,   0,  -1,   0,   0,   0,   0,   4,  32,  -8,  -3,  -1,  -1,  -3,  -1, -35, -14,  -9,   6,   0,   2,   0,   1,   6, -33,   3,   6,   1,   1,   1,   1, -14, -12,   2,   2,   2,   0,   1,   0,   8, -23,  -8,   4, -24, -54,  20,   8, -53,  17,  34,  -2,  -3,  32,   1,  -5 },
     {   -2,  -8,   5,  -1,   0,   0,   0,   1,   0,  -1,   0,   1,  -2,   3, -13,   2,  -1,  -1,   1,   1,   0,   1,   0,   0,  -6, -35,  -1,   6,   2,   0,   2,   1,   1,   0,   0,   0,   4,  34, -18,  -3,  -3,  -1,   0,   0,  -1,   0,   0,   0, -36, -46, -22,  12,   1,   3,   1,   3,   9, -77,  15,  13,   1,   3,   1,   2,  26, -16,  21,   1,  -1,   0,   0,   0,  27,  -1,   9,  -4,   0,  -1,   0,  -1,   3,   6,   2,  -3,  -4,  -8,   2,   0, -17,   4,   8,   0,   0,  10,   1,  -1 },
     {   -2,  -6, -25,  10,   2,   2,  -1,   0,   0,   1,   1,   1,  11,  21, 108, -20,  -4,  -8,  -3,  -5,  -3,  -1,  -1,  -2,   0,  -7,  34,  -9,   0,  -3,  -1,  -2,   0,   0,   0,   0,  -2, -20, -19,  -3,  -1,   1,   2,   2,   0,   0,   0,   1,  -1,   0, -17,  -2,  -1,  -1,   1,   1,  -5, -14, -10,   3,   1,   2,   1,   1,   7,  -4,   1,   0,  -1,   1,   0,   1,   1,   2,  -2,  -1,   0,   1,   0,   0,   4,  -1,   2,   0,   0,  -2,   0,  -2,  -2,   3,   0,  -2,   3,   1,  -3,  -1 },
     {    1,   0,  -7,   1,   0,  -1,   1,  -2,  -2,   0,   0,   0,   1, -10,  -5,   3,  -2,   0,   0,   1,   0,   0,   1,   0,   0,   8, -29,   4,   0,   2,   1,   1,   0,   0,   0,   0,   3, -27,   5,   5,   1,   2,   1,   0,   0,   0,   0,   0,  11,  45,  -9,  -2,   1,  -1,  -3,  -1, -31, -30, -16,  14,  -1,   4,   1,   2,  37, -38,  29,   6,   0,   0,  -1,   2,  23,  31,  13, -10,  -4,  -3,  -1,  -2,  10,   9,  -2,  -7,   4,  33,  -5,  -4,  42,  -5, -23,   0,   1, -26,   1,   3 },
     {    4,   9,  26,  -9,   4,  -6,   0,  -1,   0,  -2,  -1,   0,   0,  -2, -15,   9,   4,   1,   1,  -1,  -1,   0,   0,   1,   9,  27,  90, -10,  -7,  -4,  -5,  -2,  -3,  -1,  -1,  -1,   2, -14,  63, -15,  -3,  -4,  -2,  -1,   0,   0,   0,   0,   1,   0,  -4, -11,   2,   1,  -2,   0,  -4, -23,   6,  -3,   1,   2,   0,   1,  17,  -6,   7,  -3,  -1,  -1,  -2,  -1,  -1,  10,  -3,  -9,  -2,  -2,   1,  -1,  -3,  -2, -12,  -1,  -7,   2,  -1,   0,  -4,   6,  -1,   0,  -1,   5,  -3,  -2 },
     {    0,  -1,   7,  -3,   0,  -1,   1,   1,   1,   0,   0,   0,  -1,  -2,  -8,   8,   3,   2,  -3,   0,  -1,   0,   0,   1,   1,  -2,  26,  -1,  -3,  -2,   0,   0,  -1,   0,  -1,  -1,  -5,  -9, -11,  -3,   1,   0,   2,   1,   0,   0,   1,   1,   2, -17,  -3,  -1,   1,  -1,   1,   1, -11,  26, -27,   2,  -1,   1,   0,   0, -35, -61, -17,  22,   1,   7,   2,   3,  46, -43,  33,   9,  -2,   2,   0,   1,  30,  23,  22, -11,   6,  11, -10,  -6,  -9,  -9,  -3,   2,   3,   7,   3,   1 },
     {   -6,  -4, -27,-100,  34,  -5,  15,  -2,   6,   1,   3,   2,  -3,  -5,  -8, -58,  19,   0,   4,  -1,   2,   0,   0,   1,   1,  -2,   5,   8,   2,   5,  -3,   0,   0,   0,   0,   0,   1,  -1,  -3,   7,  -2,   0,  -2,  -1,  -1,   1,   0,   0,   2,   1,   2,   4,   0,  -3,  -2,  -1,   0,  -1,  -2,   1,   0,   0,  -1,   1,   2,   2,  -2,   4,   2,   0,   0,  -1,  -1,   3,   0,   5,   1,   1,  -1,   0,   5,   0,   5,   2,  -2,   4,   1,   1,   2,   0,   2,   1,  -2,   4,   1,  -2 },
     {    2,   2,  -3,   0,   2,   2,   0,  -1,   0,  -1,   1,  -1,   3,   2,  -3,  -2,   2,  -2,   2,   1,   2,   0,   0,   0,   2,   4,  -3,   2,  -2,  -3,  -3,   0,  -1,   1,   1,   0,   5,   3,  -3,  -4,   2,  -2,   1,  -1,   1,   0,   0,   0,   2,  13,  -1,  -2,   0,   4,  -1,   2,   7,   1,  11,  -2,   2,   0,  -2,  -2,   3,  17,  -9,   1,   0,   2,   0,   0,   8,  -6,  20,  -2,   3,  -1,  -2,   0,  16,  38,  -6,  -6,  -8,  -3,  13,  -1,  -4,  84, -10, -11,  63, -14, -39,   1 },
     {    0,  -2,  -5,   8,   4,  -2,   1,  -1,   0,  -1,  -1,   0,   3,   3,  26, -11,   5, -14,   2,   0,  -2,  -2,   2,   0,  -1,  -2, -42,  27,  10,   0,   2,   2,   0,   0,   0,   0,   5,  19,  85,  -8, -12,  -6,  -4,  -4,  -3,   0,  -1,  -2,  -6, -23,  21, -20,  -6, -11,   1,   0,  -1, -19, -22, -10,  -4,  -2,   7,   3,  -6,  -6, -17,  -2,   1,   6,   2,   1,  -1, -22,  -8,  14,   1,   5,   2,  -1,  22,  -3,  12,   5,  10,   0,   4,   0,  14,   5,  -2,  -1,   2,  -5,   1,   1 },
     {    1,   2,   8,   1,   6,  -2,  -4,   2,   1,  -1,   1,  -3,  -3, -10, -20,   6,   2,   4,  -1,   1,  -1,  -1,  -1,  -1,  -2,  -1,   7, -23,   3,  -5,   3,   0,   0,   0,   0,  -1,  -3, -24, -17,   3,  -1,   2,   2,   0,   2,   1,   1,   0,  -7,   2, -47,   1,   4,   6,   5,   0,  -4, -30, -20,   9,   5,   0,   2,   1,   0,  24, -27,  -1,   2,  -2,   0,   1, -39, -41, -23,  23,   1,   4,   2,   2,  45, -29,  17,  12,  15,  10,   5,  -3,  22,  -1,  -9,   0,   9, -19,  -3,   4 },
     {   -2,   0,  -9, -47,   7,   0,  12,  -3,   0,   0,   3,  -2,   6,   5,  19,  92,  -3,  -2,  -7,   0,  -5,  -3,  -1,  -2,   0,   3,  -1,  54, -17,   1,  -5,   5,  -1,   1,  -1,  -3,  -2,  -1, -11, -17,   0,   5,   5,   1,   3,  -1,   0,   1,   0,   7,   3,  -3,  -4,  -1,   1,   3,  -2,  -5,   1, -10,  -1,   2,   0,   0,   1,   6,   4, -16,   5,  -3,  -4,  -1,  -9,  -7,  -4,   0,   2,  -3,  -2,   2,   9,  -6,  10,  -7,  -5,   2,  -5,  -3,  -7,  -1,   1,   3,   1,  -3,   1,   1 },
     {   -1,   0,  -5, -10,   5,  -4,  -6,  -2,  -3,   1,   0,   0,   0,   0,   1,   5,  -6,   1,  -3,  -1,  -1,  -2,  -2,   1,  -2,  -1, -19,  -3, -14,  -3,   9,   1,   0,   1,   0,   1,  -1,  -2,  30, -24,  -5,  -5,  -1,  -2,  -2,  -3,   0,  -2,  -2, -25, -86,   5,  13,   0,   3,   2,   8,  44, -29,  21,  -3,   1,  -2,   1,  -5,  13,  35,  -5,  -4,   1,  -1,   3,   5,  13,  12,  -4,   3,  -1,   4,   2, -16,   6,  -4,   0,   1,  -5,   1,  -1,  -3,   9,   1,  -1,  -4,   3,   2,   1 },
     {    6,   9,   1,  24,  87,   3,   7, -18,  -1, -13,  -1,   3,   0,   0,  -1,   8,  31,  -6, -20,  -7,  -2,   3,   6,   3,  -3,  -5, -11, -37, -16,  -1,  -1,   4,   8,   4,   2,  -2,  -1,   0,  -7, -33,   0, -10,  15,   7,  14,  -5,  -1,   0,   1,   0,  17,  -8, -14,  -6,  -6,   8,   1,   0,  -5,  -2,  -6,  17,   8,   0,   2,   5,  11, -12,   3,   2,   0,  -7,   2,  -3,   1,  -7,   3,  -7,  -4,   1,   5,  10,   3,   4,  -9,  10,   6,  13,  -8,  -3,   9,   9,  -5,  -7,   6,  -7 },
     {    1,   3,   0,  10,  -1,   7,  -2,   6,   0,   2,  -1,   0,  -1,  -2,   6, -14,  -3,   3,   1,   0,  -1,   0,  -1,   0,  -1,   8,  -8,  11,   3,   0,   1,   0,  -2,  -3,   0,   1,   1,  -6,  12,  -4,  -5,   3,  -3,   0,  -3,   0,  -1,   1,  -2,  18, -21,   3,   3,   0,  -1,   1,   6,  -5,  42,   0,  -6,   0,  -1,   2,   3,  28, -10,  -2,   5,  -1,   0,  -2,  -6, -18,  31,  12,  -5,   5,  -3,  -1,  30,  34,   9, -26, -72,  11, -21,  -5,  -6, -20,  -4,   8, -34,  -2,   7,   4 },
     {    2,   3,   2,  23,  21, -21,  -3,  -1,  -1,  -4,  -4,   4,   0,   1,  -5,  -4,  31, -32,  -3,  -3,  -3,   1,   1,   2,   3,   3,  15,  46, -18, -10,   5,   3,  -1,  -2,   4,   1,   0,  -2,  -3,  62, -35,   0,  -3,   7,  -8,   1,  -1,   1,  -2,   4, -22,  34, -30,  -8,   8,  -1,  -3,   6,   5, -13, -22,  10,   2,  -1,  -5,  -3,   4,  -5,  -9,   2,   0,  -5,  -1,  -9,   7,  -8,   4,   5,   1,   0,  -5,  -2, -15,  10,   0,  -8,  10,  -6,  15,  -3,   4,  -1,   5,  -3,   1,  -4 },
     {   -3,  -4,  -2, -10, -17,  10,  -6,  -3,  -3,   1,   3,   7,  -3,  -6,  -3,  -8,   3,  -3, -13,   0,  -2,   3,   4,   6,  -3,  -3,   0,  12,   6, -13,  -7,  -4,  -1,  -4,   0,  -2,  -4, -11,  -7, -29,   5,  -8,  10,  -2,  13,  -1,   2,   3,  -4,  -7, -13,  -6,  -7,   5,  -6,   6,  -3,  -8,   2, -10,  -4,  11,   6,   1, -12, -15, -43, -12,   9,   5,   5,  -3,   8,  -7,  10, -12,  16,  -7,  -4,  -1, -25,   5, -49,  12,  -9, -31,   7,  22,  12,   0,  -6,   8, -10, -61,  23,  13 },
     {   -1,  -1,   2,  -3, -34,  22,  11,   5,  -7,  -3,  -2,   1,  -1,   1,  -9,   0,  34, -14, -22,   1,  -3,   1,  -3,  -1,   1,   2,   4,  -1,   8, -15,   8,  -2,   3,   1,   1,  -1,  -2,  -1, -15,  19,  -4, -20,  -5,  -5,   8,   2,   3,  -3,  -2,  -5,   2, -43, -12, -17,   4,  11,   0,  -8, -53, -27,  -4,   3,  11,   2,  11,  32,  15,  18,   7,   6,   1,  -7,  -4,  15,  17,  -2,   1,   1,   2,  -1,   0,  10,   6,  12, -13,  12,   3,  -8, -37,  -6, -16,   1,  -5,  -2,   0,   1 },
   },
   {
     {  -80,  45,  -2,   4,  -1,   2,   0,   1,   0,   0,   0,   0,  75, -37,  -2,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0, -28,   7,   6,  -1,   1,   0,   1,   0,   0,   0,   0,   0,   8,   4,  -5,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,  -4,   2,   0,   0,   0,   0,   0,   2,   2,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   64, -11, -11,  -2,  -2,  -1,  -1,   0,  -1,   0,   0,   0,  21, -34,  11,  -1,   2,  -1,   1,   0,   1,   0,   0,   0, -71,  40,   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,  50, -17,  -8,   1,  -2,   0,  -1,   0,  -1,   0,   0,   0, -26,   0,   9,  -1,   1,   0,   1,   0,   6,   6,  -5,   0,   0,   0,   0,   0,  -4,  -3,   1,   1,   0,   0,   0,   0,   2,   2,   0,  -1,   0,   0,   0,   0,  -3,  -1,   0,   0,   1,   1,   0,   0,  -2,   0,   0,   0,   1,   1,   0,   0 },
     {  -36,  47, -20,   5,  -4,   1,  -2,   1,  -1,   0,   0,   0, -72, -17,  27,   1,   6,   1,   3,   1,   1,   0,   1,   0,  22,  30, -14,  -2,  -3,  -1,  -1,  -1,  -1,   0,   0,   0,  40, -32,   0,   2,  -1,   1,   0,   1,   0,   0,   0,   0, -31,  13,   7,  -2,   2,  -1,   1,   0,  23,  -3,  -7,   1,  -1,   0,  -1,   0,  -7,  -2,   4,   0,   1,   0,   0,   0,   7,   1,  -2,   0,  -1,   0,   0,   0,  -2,  -1,   1,   0,   4,   0,  -1,   0,  -1,  -1,   1,   0,   3,   0,  -1,   0 },
     {  -49, -54,  44,  -2,   9,   0,   4,   0,   2,   0,   1,   0,  -4,  60, -29,  -1,  -5,  -1,  -2,  -1,  -1,   0,  -1,  -1, -26,   7,  -4,   8,  -2,   3,  -1,   1,   0,   1,   0,   0,  50, -23,   7,  -6,   0,  -1,   0,   0,   0,   0,   0,   0, -21,   7,  -1,   0,   2,   0,   1,   0,   7,   3,  -1,   1,  -1,   0,   0,   0,   2,  -3,  -1,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },
     {   14, -22,   3,   2,   0,   1,   0,   0,   0,   0,   0,   0,  50,  -8,   4,  -8,   0,  -3,   0,  -1,   0,  -1,   0,  -1,  45,  10, -28,   4,  -5,   1,  -2,   0,  -1,   0,   0,   0, -28, -44,  28,  -1,   4,   1,   2,   1,   1,   0,   1,   0, -38,  49,  -6,  -4,   1,  -2,   1,  -1,  34, -22,  -8,   5,  -2,   1,  -1,   1, -24,   5,   9,  -2,   1,   0,   0,   0,   8,   2,  -5,   1,   0,   0,   0,   0,  -7,  -1,   3,   0,   2,   1,  -1,   0,  -3,   0,   1,   0,   1,   0,   0,   0 },
     {  -22, -28,   9,   2,   5,   1,   2,   1,   1,   0,   1,   0, -15, -15,  25,  -7,   3,  -1,   2,  -1,   0,   0,   0,   0, -22,  66, -37,   5,  -4,   1,  -2,   0,   0,   0,   0,   0, -35, -24,  18,   7,   1,   3,   1,   2,   0,   1,   0,   1,  64,   3,  -6,  -8,  -1,  -2,  -1,  -1, -28,  -6,   2,   3,   1,   1,   0,   1,   7,   9,   1,  -1,  -1,   0,   0,   0,   2,  -5,  -2,   0,   0,   0,   0,   0,   0,   2,   2,   0,   1,  -1,  -1,   0,   1,   1,   1,   0,   1,  -1,  -1,   0 },
     {    4, -14,  13,  -7,   3,  -2,   1,  -1,   1,   0,   0,   0,  22, -16,  -1,   4,  -1,   1,  -1,   0,   0,   0,   0,   0,  60,  24, -17,  -7,  -4,  -3,  -1,  -1,  -1,  -1,  -1,  -1,  16, -20, -14,   9,  -2,   3,  -1,   1,   0,   1,   0,   0, -20, -39,  32,  -5,   4,   0,   1,   0, -39,  51, -12,  -4,   2,  -2,   1,  -1,  33, -22,  -7,   7,  -4,   2,  -1,   1, -22,   3,   9,  -3,   1,   0,   0,   0,   7,   0,  -3,   0,  -7,   1,   2,   0,   3,   0,  -1,   0,  -4,   0,   1,   0 },
     {    4,   5, -55,  33, -10,   8,  -3,   4,  -1,   2,  -1,   1,  25,  40,  37, -36,   3,  -8,   0,  -4,  -1,  -2,   0,  -1,   7, -24, -29,  18,   2,   3,   1,   2,   0,   0,   0,   0,  40,  -9,  21,  -9,  -4,  -1,  -1,  -1,  -1,   0,   0,   0,   7,  -3, -17,  -1,   3,  -1,   1,   0, -38,  -8,  15,   5,  -1,   2,   1,   1,  18,  15,  -7,  -6,   0,  -2,   0,  -1,  -7, -13,   1,   4,   0,   1,   0,   1,   1,   6,   1,  -2,  -1,  -4,  -1,   1,  -1,   2,   1,  -1,   0,  -2,   0,   1 },
     {   24,  57,  20, -34,   2,  -8,   0,  -4,  -1,  -2,  -1,  -1,   4,   6, -50,  28,  -6,   5,  -2,   3,  -1,   1,   0,   1,   3,  -1,   5,  -1,  -3,   2,  -1,   1,  -1,   1,   0,   0,  19, -46,  25,  -7,   6,  -2,   1,   0,   1,   0,   0,   0,  23,  27, -21,   0,  -1,  -2,   0,  -1, -42, -14,   7,   6,   1,   2,   0,   1,  20,   7,   2,  -3,  -1,  -1,  -1,   0,  -7,  -5,  -3,   1,   1,   0,   0,   0,  -1,   1,   3,   1,  -1,  -1,  -1,  -1,  -1,   1,   0,   1,   0,  -1,   0,   0 },
     {   -4,  16, -18,   8,  -4,   3,  -1,   0,   0,   0,   0,   0, -11,  23,   3,  -6,   1,   0,  -1,   0,   0,   0,   0,   0, -45, -13,   1,   4,   4,   1,   2,   1,   1,   0,   0,   0, -52, -19,  19,   6,   2,   2,   1,   1,   1,   0,   0,   0, -28,  31,  11,  -9,   4,  -3,   1,  -2,   5,  33, -23,   5,  -3,   1,  -1,   0,  48, -39,   5,   2,  -3,   2,  -1,   1, -30,  15,   7,  -3,   4,  -2,   1,  -1,  21,   0,  -7,   1,  -5,  -3,   3,   0,   7,   1,  -1,   0,  -2,  -1,   1,   1 },
     {   10,  23, -29,   5,  -4,   2,  -1,   1,   0,   0,  -1,   0,  19,  57, -14,  -1,  -8,  -2,  -3,  -1,  -2,  -1,  -1,   0,   3,  16,   1, -19,   6,  -5,   2,  -1,   1,   0,   0,  -1,   2, -27, -29,  27,  -6,   5,  -1,   3,   0,   1,   0,   1,  43, -16,  36, -16,  -2,  -1,  -2,  -1,  20,  -2, -29,  -1,   5,  -2,   2,   0, -38, -16,  20,   7,  -2,   2,   0,   1,  12,  17,  -6,  -7,   1,  -1,  -1,   0,  -1, -13,   0,   4,  -1,   4,   1,  -2,   0,  -5,   0,   1,  -1,   2,   0,  -1 },
     {   16,  43,  27, -11,  -5,  -4,  -2,  -3,  -1,  -1,  -1,  -1,   1,  21, -15, -22,  10,  -5,   3,  -2,   2,  -1,   1,  -1,  -8,  15, -48,  44, -13,   7,  -3,   3,  -1,   1,   0,   0, -16,  16,  31, -22,   0,   0,  -1,  -1,   0,   0,  -1,   0,  -8, -48,  -4,   9,   8,  -1,   3,   1,  30,  28,   8,  -6,  -5,  -2,  -1,  -1, -21,  -5, -14,   3,   2,   1,   1,   1,  12,  -6,   8,   2,  -3,   1,  -1,   0,  -4,   2,  -2,  -3,   3,  -2,   1,   3,  -1,   0,  -1,  -1,   3,  -1,   0,   1 },
     {    3,  -8,  16,  -5,  -1,  -2,   0,   0,   0,  -1,   0,  -1,   2, -26,  11,  -7,   3,   0,   2,   0,   1,   0,   0,   0,  14, -19, -15,  19,  -1,   3,  -1,   2,   0,   0,   0,   0,  41,  19,  16, -13,  -6,  -2,  -2,  -1,  -1,  -1,  -1,   0,  46,  11, -21,  -6,   5,  -3,   1,  -2,  27,  -6, -21,  14,  -7,   3,  -2,   1,  -5, -50,  27,  -5,   3,   1,   0,   1, -37,  39,  -1,  -5,   3,  -3,   2,  -1,  22, -12,  -9,   5, -15,   1,   8,  -2,   6,   1,  -4,   0,  -6,   1,   1,   1 },
     {    6,   2,  -7,  30, -21,   7,  -6,   2,  -3,   1,  -2,   1,   6,  10, -30, -24,  20,  -3,   3,   0,   2,   0,   1,   0,  18,  48,  31,  -4, -13,  -7,  -1,  -3,  -2,  -2,  -1,  -1,  -1,  -4, -28, -17,   8,   1,   2,   1,   1,   1,   1,   0,  13,   8, -21,  30,  -5,   1,  -1,  -1,  31, -18,  25, -18,  -5,   3,  -3,   1,  23, -18, -24,   7,   7,  -2,   2,   0, -32,  -7,  23,   2,  -4,   1,   0,   0,   9,  13, -12,  -4,   1, -13,   3,   5,  -1,   3,   1,  -3,   0,  -4,   1,   1 },
     {    4,   5,   1, -33,  19,  -6,   4,   0,   0,   0,   1,  -2,  11,  11,  19,  19, -17,   0,  -4,  -1,   0,   0,   0,   0,   2,  -4, -30, -11,   4,   5,   1,   1,   1,   1,   1,   0,  14,  22, -10,  18,  -3,  -1,  -2,  -1,  -2,   0,  -1,   0,  18,  12,  10, -18,  -4,   1,  -3,   1,  48, -11, -14,   1,   6,  -5,   1,   0,  41,   0, -18,   4,  -9,   2,  -2,   1, -21, -51,  16,   1,   7,   1,   0,   2, -17,  34,   7,  -5,  15, -11, -15,   3,  -9,  -1,   9,   2,   4,  -1,  -3,  -3 },
     {   -4, -16,  -9,  18, -12,   9,  -3,   4,  -1,   1,   0,   0,  -7, -28,  -9,  -3,   8,   1,   1,   1,   1,   0,   0,   1,   0, -36,  27,  17,  -6,   4,  -1,   0,  -1,   0,   0,   0,   9, -37,  21, -13,   7,  -2,   3,   0,   1,  -1,   1,   0,  38,  12,  12,   0,  -2,  -1,  -3,  -2,   9,  49, -17, -13,  -3,  -3,   0,  -1, -17, -19, -21,  16,  -1,   3,   0,   2,  25, -28,  21,  -2,  -2,   2,  -1,   1, -33,  21,   2,  -7,  15,  -2,  -7,   4, -11,  -1,   3,   0,   3,   3,  -1,  -1 },
     {  -11, -18, -59, -37,  29,  -1,  10,   1,   6,   0,   3,   2,   6,  10,  12,  55, -19,   1,  -8,   0,  -4,  -1,  -2,  -1,  -1,  10,   3, -12, -10,   5,  -2,   3,  -2,   1,   0,   1,  -4, -10,  24, -28,  20,  -7,   3,  -3,   1,  -1,   1,   0,  -1, -23, -21,  31,  -7,   3,   1,   0,  11,  14,  19, -15,  -5,   1,  -2,   0, -18, -14,  -4,   3,   7,  -1,   3,   0,  -9,  12,   6,   2,  -2,   0,  -1,   0,   9,   0,  -7,  -1,  -3,  -7,   6,   1,   0,   5,  -1,  -2,  -1,  -1,   1,   1 },
     {    2,  10,  33, -22,   4,  -2,   1,  -4,   0,  -1,   1,   0,   0,   1,  42, -10,   6,  -9,  -1,  -1,  -1,   0,   1,   0, -13, -35, -10,  11, -11,   8,  -1,   2,   1,   2,   0,   0,  -6, -34, -34,  10,  14,   1,   4,   1,   2,   1,   2,   0,   3,   9,  23,  18, -21,   4,  -5,   1, -12, -15,  19, -31,  15,  -1,   1,  -1, -31,   9, -22,  19,   0,  -3,   3,  -1, -27,  12,  21,  -7,  -5,   3,  -1,   1,  23,  12, -19,  -3, -12, -12,   8,   7,   4,   7,   0,  -3,  -5,  -2,   0,   0 },
     {   -5, -13, -17,  19,  10,   0,   0,  -1,   0,  -1,   0,  -1, -13, -26, -39,   0,  -7,  12,   2,   5,   0,   2,   2,   2,  -8, -18,   3,  14,  13,  -3,  -2,   0,   0,   0,  -1,   0,  -8, -34,  20,  22, -16,   3,  -1,   0,  -1,   1,   0,   0,  -5, -50,  15, -23,  12,   2,   1,   1,  16, -30,  13,  17,  -2,  -3,   1,  -1, -16,  19,   3, -14,  -2,   2,  -2,   0, -48, -12,   9,   8,   6,  -1,   2,   1,  20,  22,   5,  -4, -11,  -7,  -4,   2,   2,   7,   7,  -1,  -2,   1,  -3,  -1 },
     {   -8, -14, -24, -25, -28,  24,  -6,   5,  -2,   4,   0,   1,  -1,  -5, -22,  21,  57, -28,  10,  -7,   3,  -2,   2,  -2,  -4,  -8, -12,  17, -48,  19,  -6,   4,   0,   0,   0,   2,   1,   9,  -3, -12,  25,   0,  -2,   0,  -2,   1,   0,   0,   1,  -4,  40,  -1,  -4,  -9,   4,   0,  -1, -24, -20,  -1,   0,  10,  -4,   3,  17,  19,  13,   2,   1,  -7,   1,   0,  -7,  -2, -14,  -5,  -2,   3,   0,   0,   1,  -8,  10,   6,   0,   3,  -4,  -5,   2,  -6,   1,   5,  -4,   2,   4,  -4 },
     {    3,  -3,  -4,   0,  20, -14,   1,   0,   0,   0,  -1,  -1,   3,  -9,   3,  -8, -18,  13,   2,   3,  -1,   3,   1,   0,   3, -13,   6,  26,   1,  -2,  -2,   0,   0,  -1,   0,   0,   4, -25, -12,  -7,   0,   6,   0,   3,  -1,   1,   0,   1,  22,  -2,  17,  12,   0,  -5,  -1,  -4,  42,   8,  13, -14,  -6,  -2,   0,   1,  50,  29, -15,  -9,   3,  -4,   0,  -1,  32,  -4, -22,   9,  -5,   1,  -2,   0,  24, -46,  10,   0, -33,  16,  11,  -1,  18,  -1,  -6,  -1, -11,  -4,   4,   0 },
     {    1,  -2,  15,   9, -48,  26, -11,   7,  -4,   2,  -2,   1,   2,  -2,   4,  31,  24, -17,   2,  -5,   1,  -3,   0,  -1,  -6, -16, -24, -52,  13,   1,   9,  -2,   4,   1,   2,   0,   6, -10,  24,  46,  -9,   1,  -7,   1,  -2,  -2,  -1,   0,   5, -12, -28,  -4,  -3,   3,   2,   1,  22,  16,  33, -16,  -3,  -2,   1,  -2,   7,  -1, -17,   3,   4,  -2,   1,   1,   7,   3,  -2,  -1,  -4,   1,  -1,   0,   8, -17,   4,   1, -11,   7,   3,  -3,   8,  -1,  -3,   0,  -5,  -1,   3,   0 },
     {    0,  -2, -17,  10,   7,  -3,  -2,   1,  -1,   0,   1,   3,  -2,  -5, -30,   6,  -3,   3,  -2,   1,   2,   2,   0,   0,   7,  17,  -8,   1,   7,   0,   0,   4,  -1,   0,   0,  -1,  14,  38,  32,   5,   0,  -3,   0,  -7,  -1,   0,  -2,  -1,   0,  29,  19, -33, -15,   4,  -4,   4,  -6,   3, -11, -15,  22,  -6,   7,  -5,  -9,  21, -21,  32,  -8,  -2,  -1,   0,  23,   2,  22, -11,  -9,  -1,  -1,   2,  44,   0, -34,  -3, -21, -26,  18,   6,   4,  12,  -4,  -6,  -3,  -9,   0,   4 },
     {    3,   4,  12, -27,  -9,   4,   4,   2,   5,  -3,   3,   0,  10,  15,  35,   3,  21, -16,  -3,  -4,  -4,   1,  -2,   0,   8,  11,  51,   1, -16,  -3,  -2,  -2,  -1,   0,   2,  -1,   1,  -6,   5,  -2, -17,  18,  -3,   6,  -2,   3,  -1,   1,   6,  -2, -11, -29,  33, -19,  11,  -4,  11,  25,  -4,  26, -16,   9,  -5,   0,  -6,  27,  -1, -17,   8,  -4,  -1,  -1,  10,  -1,  -4,  14,  -6,   0,   1,  -1,  40,  22, -13, -12,  -2, -25,  -7,   6,  -1,  10,   9,  -2,   5,  -2, -11,   3 },
     {    8,  10,  34,  47, -10,  -1, -11,  -5,  -6,   0,   0,  -2,   4,   4,  24,  13, -27,  -2,   0,   3,   0,  -1,   1,   1,  -1,  -1,  13, -35,  -8,  13,  -4,   1,   1,   2,   1,   0,   1,  -8,  29, -32,  39, -17,   6,  -3,   1,   2,   1,  -2,   3, -31,  14,  11, -21,   0,   0,   3,  12, -11, -28,   2,   9,   7,   0,   2,  17,  20,  26,  -6,  -8,  -4,  -3,  -1,   1,  -3,  -8, -11,   4,  -2,   0,   1,  10,  10, -13,   9,   5, -19,   3,   2,  -6,   3,   4,  -1,  -2,   0,  -1,  -2 },
     {    2,   9,  10,  20,  -5,  -5,   2,  -7,  -4,   0,  -1,  -2,   2,  12,  15,  17, -12,  -2,  -6,   3,   1,  -1,  -2,   0,   2,  17,  17,   0, -13,   4,   3,  -4,   0,  -3,  -1,   0,   2,  18,  18, -13,  -2, -14,  -5,   2,   0,  -1,  -1,   1,   0,  20,  17, -16,  -5,   6,   3,   0,  -8,  11, -10, -14, -14,   1,  -1,   1, -34,   6, -27,  -2,  14,  -4,   5,   0, -47, -29,  -7,  18,  -1,   9,  -2,   2,   7, -14,  33, -10, -23,  42, -17,  -8,  20, -25, -11,   3, -13,  -1,   6,   1 },
     {   -2,  -4,   4,  -5, -39, -13,  11,  -5,  13,  -3,   0,  10,   3,   2,  11,  19,   8,  35, -27,   1, -10,   2,  -6,  -3,  -3,  -4,  -9,  -1,  23, -45,  16,  -5,   1,  -5,   1,  -4,   0,   2,  17, -30, -22,  33,  -2,   0,   4,   2,  -2,  -1,   1,  -4,  29,  17,  25, -19,  -3,  -3,  -4, -27, -15, -28, -19,   1,  14,  -3,   8,  10,  -3,  11,  13,   2,  -8,   1,  -4,  10,  14,  -5,  -9,  -8,   6,   0,  -9,   0, -10,  -4,  12,   4,   0,   2,  -2,  -2,  -3,   0,   2,   1,   1,   4 },
     {   -1,   0,   6,   1, -14,  14,  -4,   7,   1,   4,  -3,   4,   2,   4,  18,  26,  -2,  -3,   2,  -9,  -1,   1,   1,  -4,  -2,   2,  13,  14, -10,   6,  -1,   6,  -1,   0,  -2,   0, -11, -14, -14, -24, -29,   3,   7,  -4,   1,   4,   1,   0, -11, -20, -24, -15,  14,  20,   2,   1,  -5, -32, -19,  26,  29,  -8,   1,   1,   3, -25,  -8,  15, -20,  11,   0,   3,  18, -31,  26, -20,   3,   2,  -4,   0, -17, -17,  -9,  -3, -39,   8,  16,  10,  24,   4,  -8,  -4,  -9,  -6,   7,   4 },
     {    2,   3,   6, -11,  10,  27, -11,   5,  -4,   3,  -2,   1,   3,   4,  15,   2, -15, -35,   9,  -5,   7,  -3,   1,   0,   9,  14,  39,  31,  13,  -2,  -6,  -2,   1,  -2,  -2,  -2,   1,   1,  34,   5, -36,  -9,   0,   2,   1,   0,   2,  -1,   2,  -5,  18, -11,  -7,   9,  -8,   2,   9,  -9,  25, -25,  20, -13,   4,   1,  19,  10,   6,  12, -16,  -6,   1,  -1, -12,  37,  -3, -22,  -1,   5,  -1,   2, -22,   1,  -2,   1,  36,  19,  -6,  -4,   0,  -6,  -4,   1,  -3,  -2,   1,  -2 },
     {    3,   3,   6,  17,  12,  46, -31,  -4,  -6,   5,   0,   2,  -2,  -4,  -9,  31, -28, -35,  21,  -4,  -2,   3,   2,  -1,   0,  -1, -23,   7,  33,  10,  -9,   1,   4,  -3,   1,  -1,  -2,   4, -19, -17, -18,  -2,   0,   0,   1,   0,   2,  -1,   4,  15,  25,  23,  21, -11,   5,  -8, -10,  -7,  -6,   5, -20,  14,  -8,   1,  -5,  -9, -17, -31,  16,   4,   5,   2,  12,   4,  12,  25, -11,  -4,   1,  -3,   8,   3,  -5, -12,   5, -16,   4,   5,   0,   2,   6,  -2,  -7,   1,  -1,   3 },
     {    4,   3,   6,  35,   9, -23,   9,  -5,   8,  -3,  -2,   0,   2,   0,   6,  25, -13,  12, -13,  -5,  -4,  -2,   0,  -1,  -1,  -5,  -2,   6, -47,  22, -12,  11,  -9,   5,  -3,   1,   1,   3,  -3,  16,  10,  -7,  -3,   5,   4,  -3,  -1,   1,  -1,   0,  13, -32,  33, -15,  22, -18,  -4, -22,  35,   0, -18,   0,  -5,   6,   9, -25, -23,  13,   5,   4,  -5,   8,  18,  28,   5, -10,   7,   0,   4, -10,  -8,  -6,   5,  -9,  18, -15, -12,  -3,  -7,  -3,  -1,   0,   1,  -7,   8,   5 },
     {    0,   3,   1,   5,   7,   9,  -6, -15,   0,   3,  -4,   0,  -1,   3,   3,   4,  -5,  -9,  -2,  -2,   4,   4,   5,  -1,   0,   4,   9,   1, -18,  -1,  12,   5,   1,  -1,  -9,   1,   1,   7,   7,  10,  24,  23, -24, -10,  -6,   0,  -2,  -2,   1,   7,   8, -11,  15, -23, -25,  14,   0,   4,   6,  -8,   3, -31,  22,  -3,  -6,  -1, -13, -37,   2,   8,   6, -13, -22,  -3, -16,   1,   3, -14,  -3,  13, -43, -19, -21,  14, -23,  -5,  31,  19, -14,  34,  -1, -10,  20,  -1,  -7,   7 },
   },
   {
     {  105, -46, -11,  -1,  -2,  -1,  -1,   0,   0,   0,   0,   0, -12,  -8,   8,   1,   1,   0,   1,   0,   0,   0,   0,   0, -46,  23,   8,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   6,  -5,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   4,  -3,  -1,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    0,   9,  -6,   0,   0,   0,   0,   0,   0,   0,   0,   0,  98, -42, -11,  -1,  -2,  -1,  -1,   0,   0,   0,   0,   0, -33, -18,  15,   2,   4,   1,   1,   0,   1,   0,   1,   0, -46,  25,   9,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   5,   9,  -7,  -1,  -1,   0,   0,   0,   7,  -4,  -3,   2,   0,   0,   0,   0,  -2,  -3,   2,   0,   0,   0,   0,   0,  -3,   2,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {   34, -38,  11,  -5,   1,   2,   0,   1,   0,   0,   0,   0,  13,   6,  -9,  -1,   2,   0,   0,   0,   0,   0,   0,   0,  77, -29, -25,   9,   0,  -2,  -1,  -1,   0,   0,   0,   0, -48, -24,  21,   9,   3,   1,   0,   1,   1,   0,   0,   0, -39,  20,  14,  -5,   0,  -1,   0,   1,   7,  10,  -9,  -3,   0,  -1,   1,   0,   0,   0,  -2,   2,   0,   1,   1,  -1,   1,  -1,   1,   0,   0,   1,   0,   0,  -1,   1,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -2,   1,   1,   0 },
     {  -42, -68,  48,   9,   6,   2,   1,   2,   1,   1,   1,   1,  13,  -6,  14,  -8,   1,  -2,   0,  -1,   0,   0,   0,   0,  -8,  68, -32,  -9,   1,  -3,   0,  -2,   0,  -1,   0,  -1, -21,  11, -13,  10,   0,   1,   0,   0,   0,   0,   0,   0,   3, -16,   6,   5,  -2,   2,  -1,   1,   9,   1,   2,  -3,  -1,   0,   0,   0,   3,   3,  -2,  -1,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   -7,  -5,  -2,   7,   1,  -2,   1,  -1,   0,   0,   0,   0, -44,  34,   4,   1,  -3,  -1,   0,   0,   0,   0,   0,   0,  -5, -10,  14,  -9,  -2,   3,   0,   1,   0,   0,   0,   0, -66,  35,  14,  -7,   4,   0,   2,  -1,   0,   0,   0,   0,  47,  32, -25,  -1,  -2,  -2,  -1,  -2,  39, -31, -10,   6,  -2,   1,  -1,   1,  -5, -13,  10,   3,   0,  -1,   0,   1,  -7,   5,   3,  -2,  -1,  -1,   0,   0,   3,   1,  -1,  -1,   3,  -1,   0,   0,   1,  -1,  -1,   0,   2,  -2,   0,   0 },
     {    8,   7, -17,  13,  -3,  -2,   0,  -1,   0,   0,   0,   0,  -4, -40,  39,   3,  -3,   0,  -1,   1,   0,   0,   0,   0,  54,  17,  30, -40,  -5,  -3,  -2,  -1,  -2,  -1,   0,  -1,   3,  48, -41, -16,   5,  -2,   3,  -1,   0,   0,   0,  -1, -32,   0, -21,  21,   2,   2,   2,   0,   4, -15,  11,   5,  -5,   2,  -2,   0,   8,  -1,   4,  -5,   0,  -1,   0,   1,  -6,   4,  -3,   0,   1,  -1,   0,   0,  -2,  -2,   0,   2,   2,  -1,   1,   0,  -1,  -1,  -1,   0,   2,  -1,  -1,   0 },
     {   12,   4, -28,  29,  -2,  -8,   0,  -2,   0,   0,  -1,   0,  51,  66,  -8, -12, -18,  -3,  -5,  -2,  -2,  -2,  -1,  -1,  21,  40,  12, -36, -15,   5,  -3,   3,  -2,  -1,   0,   0,  14, -36,   0, -14,   2,   3,   6,   1,  -1,   1,   1,   0,  16,  -4, -11,   0,   5,  -1,   3,  -4,   7,  -4,  -6,   4,   0,   1,  -2,   0,   1, -18,   2,   5,   0,  -2,  -1,   2, -20,   1,   6,   3,  -2,  -2,   0,   1,  -3,  -1,   1,  -1,   0,  -2,   0,  -1,  -1,  -2,  -1,   0,   5,  -4,  -1,   0 },
     {   18, -14,  28, -12,  -1,   0,  -1,  -1,   0,   0,   0,   0,  22,  11,  -9,  -5,   0,  -2,   0,  -1,   0,   0,   0,   0,  38, -26, -20,   9,   1,  -1,   0,   0,   0,   0,   0,   0,  25,  29,  -8,  -3,  -5,  -2,  -1,  -2,   0,   0,   0,   0,  60, -26, -20,   0,  -4,   1,  -1,   0, -41, -39,  19,   5,   3,   3,   2,   1, -30,  19,  12,  -4,   1,  -1,   0,   0,   5,   9,  -8,  -2,   0,   0,   0,   0,   1,  -3,  -3,   3,  -4,  -2,   3,   1,  -2,   0,   0,   0,  -2,   0,   0,   0 },
     {  -29, -68,  -8,  27,  10,   5,   3,   4,   1,   2,   0,   1,  -3,  -6,  -9,  19,  -4,   2,  -1,   1,   0,   0,  -1,   0,  -2, -22,  58,   1, -12,   1,  -3,   0,  -1,  -1,  -1,   0,  23,  -1,  33, -32,   1,  -1,  -1,  -1,   0,   0,   0,   0,  -5,  17, -22, -12,  11,  -3,   2,  -1, -33,  14,  -5,  10,   3,  -1,   2,  -1,   0,   6,   1,   3,  -4,   0,  -1,   0,  12,  -2,   0,  -3,  -1,   1,   0,   1,   1,  -1,   0,  -2,  -4,   2,   1,   0,   0,   3,   0,  -1,   0,   2,   0,  -1 },
     {  -11,  -8, -60,  31,   5,   6,  -1,   4,   1,   1,   0,   1, -11, -30,  13,  12,   3,   3,   2,   1,   0,   0,   0,   0,  13,   3,   7, -16,   8,  -3,   2,  -4,   1,   0,   0,   0, -30, -34,   0,  38,   4,  -1,  -2,   1,   2,   0,   0,   1,  42, -26,  25,   5, -12,  -1,  -4,   2, -10,   6,  11, -20,   1,  -1,   3,  -1, -34,  24,  -2,  -4,   5,   1,   1,  -2,  15,  -1,  -7,   2,   0,   1,  -1,   0,   3,  -3,   2,   1,  -2,   1,   0,   0,   0,   2,   2,  -1,  -4,   2,   1,   0 },
     {    0,   8,  -7, -13,   4,  -1,   1,  -1,   0,   0,   1,   0,  -9,  33, -43,  10,   3,  -1,   2,   0,   0,   0,   0,   0,  -4,   6,  21,   7, -13,   1,  -2,   0,  -1,   0,  -1,   0, -48,  16,  -3, -18,   3,   5,   1,   1,   1,   1,   0,   0, -23, -63,  -5,  22,  10,   2,   2,   3, -17,  11,  33,   3,  -2,  -1,  -1,  -1,  26,  38,  -9, -12,  -2,  -2,  -2,  -1,   4, -12, -11,   2,   2,   1,   1,   1,  -1,  -6,   3,   5,   2,   2,   2,  -1,   2,   1,  -1,   0,   0,  -1,   0,   0 },
     {   -4,  -7,  11,  23,  -8,   0,  -2,  -1,  -1,   0,   0,   0,   4,  39,  31, -35, -13,  -2,  -1,  -1,  -1,   0,  -1,  -1, -27, -44,  -7, -38,  27,   5,   7,   2,   3,   1,   1,   1, -12,   7, -24,  28,  15,  -5,   2,  -2,   1,  -1,   1,  -1, -14,  24,   2,  14, -14,  -4,  -1,  -1, -48,  11,  16,  -7,   0,   4,   1,   1,  24,   6, -10,  -2,   2,  -1,   0,   0,  12,  -7,  -6,   6,  -1,  -1,   0,   0,  -2,   0,   3,   1,  -1,   1,   1,  -1,   3,   1,  -1,   0,   1,  -2,  -1,   1 },
     {   -9, -23,   6, -43,  22,  10,   1,   3,   1,   1,   2,   0,   7,   6,  -4,  -2,   7,  -2,   1,   0,   0,   0,   0,  -1,  -9, -21,  27,  11, -10,  -1,  -1,  -3,   0,   0,  -1,   0,   4, -43, -56,   6,  33,   2,   0,   0,   4,   0,   1,   1,   2,  12, -18,  40,   3, -11,  -2,   0,  29,   2,  22,  -2, -24,  -2,  -1,   2, -25, -14,   9, -10,   1,   5,   2,   1,  -1,   7,  -6,  -3,   7,   3,   0,  -1,   5,  -1,  -6,   4,  -4,  -2,   3,   2,  -1,   3,   2,   1,  -2,   2,   1,   0 },
     {    1, -25,  -2,  29,  -8,   8,  -1,   3,   0,   0,   0,   0,   3,   2, -41,   9,   7,   5,   0,   0,   0,   1,   0,   1,   8, -31,   4,  -7,  26,  -6,  -1,  -1,   2,   1,   1,   0,  37,  31,   3,  21,   1, -13,  -1,  -3,   0,  -2,  -1,  -1,  11, -21,  15,  22, -23,   3,  -3,   0,  54,   6,   5, -23,  -7,   0,  -2,   2,  26, -13, -19,  -8,   3,   1,   3,  -1, -29, -16,   7,   8,   3,   2,   1,   0,  -3,   8,   7,  -1,   5,   1,  -2,  -5,  -1,  -3,   0,   0,  -4,   0,   2,   2 },
     {   14,  21,  59,  -4, -23,  -6,  -4,  -2,  -2,  -2,  -2,   0,  -6, -23,   2,   7,  -3,   6,  -2,   2,  -1,   1,   0,   1,  10,  -4,  18, -40,  -2,   6,  -1,   2,   0,   1,   0,   0, -11, -17,   6, -30,  22,  -1,   3,   3,   0,   1,   0,   1,  41, -11,   3,  16,  12,  -9,   0,  -4,  -2,  51, -20, -10,  -5,  -6,   1,  -2,  -9,   1, -26,   7,   0,   4,   2,   1,   4, -22,  12,   6,  -2,   1,  -1,   0,  -1,   4,   5,  -3,  -2,   6,  -6,  -3,   0,  -1,  -3,   2,   0,   0,   1,   0 },
     {    0,  -6, -17, -40,  31,   2,   5,   0,   2,   0,   1,   0,  10,   9,  34,  -3,   5, -10,  -4,  -1,  -1,  -1,  -1,  -1,  20,   3,  20,  29, -16, -15,   2,  -2,   0,  -1,  -1,  -1,  -3,  14, -21,  19, -24,   6,   5,  -2,   2,   0,   1,   0,  37,   2,  -5, -23,  -6,  10,  -5,   2,  -7,  32,  -5, -11,  15,  -5,   0,  -1,  45, -12, -21,  11,  -5,   0,   0,  -2,  -2, -34,   8,   6,  -5,   2,   0,   2, -13,   8,   9,  -6,   2,   3,  -7,  -2,   3,  -4,  -2,   2,   0,  -2,   1,   0 },
     {    0,   2, -12,  -5,  19,  -6,   0,  -1,   0,  -1,   1,  -1,   3,   1, -32,  54,  -7,  -6,   1,   0,   0,  -1,   0,   0,   2,  15, -21,  -5, -28,  19,  -1,   2,   0,   1,   0,   0,   2,  21, -16,  -9,  37,   0,  -6,  -1,  -1,   0,   0,   0,  10,  37,  43,  14, -14, -22,  -3,  -4, -30, -18,  -5, -18, -12,   6,   4,   4,  19, -16, -14,  11,   3,   1,   3,   2,  25,   2,   4,   0,  -4,  -1,  -1,  -1, -11,  -3,   7,  -2,   0,   3,   2,  -3,   2,  -1,  -1,   1,   2,  -2,   0,  -1 },
     {   -2,  -8,  13,  -1,  21, -12,  -1,  -4,   1,   0,   0,  -1,   2,  10,  38,  13, -29, -11,   0,  -1,  -1,   0,   0,  -1, -13, -32,   8,  -7, -35,  19,   4,   6,   0,   1,   0,   1,  -1,   6,  -4, -18,  -8,  17,  -1,   1,  -1,   1,   0,   0, -12, -23,  49, -20,   7,   4,  -5,   1,  13, -11,  11, -23,  17,  -2,   0,   0, -23,   8, -18,   1,  11,  -5,   1,  -1, -45,  27,   6,   7,   2,  -3,   1,   0,  16,   7,  -1,  -2,   9,  -5,  -2,   0,  -4,   2,   1,   1,  -1,   2,   0,  -2 },
     {   -3,   2,  10, -22, -17,  12,   0,   2,   1,  -1,   0,   0,  -2,  19, -25,   4,  11,   0,  -1,  -3,   0,   0,  -1,   0,   2,  18,  41,  -4,  21, -29,  -6,  -1,  -1,   0,   1,  -1, -10,  12, -19,   7, -14, -17,  10,   0,   2,  -1,   2,   0, -21,   2,   9,  -9, -40,  22,   6,   0, -23,   0, -22, -30,  19,  14,  -3,   3, -43, -13,  -9,  27,  18,  -5,   1,   3,   1,  17,  15,   7, -11,   0,   1,  -1,  17,   5,  -4,  -5,  -7,  -2,  -1,  -3,  -4,   1,   1,   2,   2,   2,  -1,  -1 },
     {   -6,  -7, -14,  -3,  11,   1,  -2,   3,   2,   0,   1,   0, -14, -16, -16,   3,   4,   1,   1,   0,  -1,   0,   1,   1,  -4,  -9, -17, -10,  11,   0,   5,   0,   0,   0,  -1,  -1, -16,  -7, -10,   7,   7,   0,   3,   1,   1,   0,   0,   0,   0, -26,  -9,  14,  14,  -4,  -1,   1, -43,  -3, -16,  30,   2,  -7,   5,   0,   5, -37,  15,  15, -15,   6,   1,   1, -51,   9,  42,  -9,  -1,   2,   1,   0,  -4,  45,  -2, -11,  18,  -5, -17,   5,  -1,  -8,   5,   5,  -8,   5,   6,  -2 },
     {    7,  11,  25,   3,  46, -38,  -4,  -5,   2,  -1,  -1,  -3,  -7, -14, -18, -17,  24, -10,   6,  -4,   2,   0,   2,   0,   1,   6,  17, -23, -39,  25,   6,   5,  -2,   1,   1,   0,   4,  -5,  32,  45, -28,  -3,  -3,   1,  -4,   0,   0,   0,  -6,   7, -22,  26, -20,   3,  -4,  -2, -11,   1,   0, -24,   3,   9,  -7,   7,   4,  -2,  19, -14,   5,  -4,   5,   4,  -5,   7,   4,   2,   0,  -4,   4,  -2,  -1,  -1,  -2,   5,   3,   1,  -1,  -1,   0,   0,  -2,   1,   1,  -2,   1,   0 },
     {   -4,  -5, -11, -50,  15,  13,   1,   0,   1,   1,   0,   1,   6,   6,  27,   8, -13,  -6,   1,  -1,  -1,   0,  -2,  -1,   0,   2,   0, -21,  16,  14,   0,  -1,   0,   0,  -1,   0,   9,   1,  38,   4,  42, -32, -11,  -4,  -1,   0,  -1,  -2,  -5, -23,  -9,  -2, -39, -18,  19,  -3,   8,  -5, -34,  12,  -2,  27,   2,  -2,  10,  31,   8,   3,   1,  -4, -11,   1, -17,  -2,  -4,  -9,  -1,   1,   0,   0,  -3, -10,  -3,   3,   4,  -1,  -1,   2,   1,   2,   3,   0,  -4,   2,   0,  -1 },
     {   -4,  -2, -25, -18,  20,  -3,  -2,   1,   1,   2,  -3,   0,   1,  10,  -4, -53,  25,  13,   5,   0,   1,   1,   0,   1,   5,  15,   2,   3,  11,   9,  -5,  -2,  -3,  -1,  -1,   0,   9,  34,  21,   0,  15,   4, -18,   1,  -2,  -1,   0,  -1,  -8,   9,  12,  11,  32, -33, -10,   1,  -5,   2, -11,  -6, -14, -30,  10,   2, -37,  10, -35,   2,  -5,   6,  10,  -1,  -4,  -6,   3,  25,  -7,   0,   1,  -1,  16,   5,  16,  -3,  -1,   6,  -7,  -7,  -6,  -2,  -9,   6,   5,  -2,   1,   3 },
     {   -6,  -3,   1,  -2, -28,  -2,  10,  -3,   1,   0,  -1,   1,  -3,  -8,  11,  -4,   9,   3,  -3,   0,  -1,  -2,   1,   0,  -5,   4,   2,  18,  -6,   4, -19,   9,  -2,   2,   0,   0,  -8, -20,   5, -15,  -4, -13,   4,   3,   0,   2,  -1,   1,  -6,  10,  -6,  -3,  -9,  -8,  19, -11, -14, -30,  20, -22,  -7,  18,  -8,   3, -14,  -2,  -5, -15,  12,  -3,   1,   8, -32, -29, -10,  21,   4,   1,   9,  -2, -53,  18,  33,   2,  16,  41,  -3, -18,   6, -14, -13,   5,  -4,  -4,   8,   5 },
     {    3,   9,   3,   0, -14,   3,   5,   1,   0,  -3,  -2,  -2,   7,  12,  23,  27,  15,   6, -13,  -6,  -3,  -1,   0,   1,   6,  19,   4,  19,  10,  -8, -28,   9,   1,   5,  -1,   0,  -2,   7,  23,   0,   2, -19,   5,   9,  -3,   0,  -1,   0,  11,  23,   6,   9,  -5, -12,   4,  -7, -19,  28,  28, -12, -21, -14,  -4,  10,  12,   2,  10, -33,  -5,   0,  21,   6, -28,  21, -15,  -4,   8,   2,   7,  -8,  33,  26, -23,  10,  14, -26,   0,  -1, -17,   0,  10,  -8,   6,   8,  -6,  -2 },
     {    0,  -1,  -5,  16,   6, -37,  15,  -9,   0,  -3,   1,  -1,  -3,  -1, -12,   4, -16,   8,   5,   2,   3,   0,   0,   1,   3,  -2, -11,  17,   0,  48, -25,   7,  -7,   1,  -1,   0,  -4,  16, -29,  15,   8, -36,  -3,   3,  -3,   0,   0,   0,  -4,  -4, -19, -32,   2,  -9,  37,  -5,   7,  34,  14,   6,   9,  16,  -3, -17, -24, -11,   2,  15, -13,   2, -13,  -1,  -1, -18,  -6,  -2,   2,   6,  -3,   1,  27,   3,   3,  -2, -11,   2,   3,  -1,  -8,  -5,   0,   1,   5,  -1,   0,   0 },
     {    1,  -1,  18,   1, -21, -12,  11,  -3,  -5,   4,  -1,   0,  -2, -11,  -4, -19,  19,  17,  -2,   2,   2,   2,  -2,   1,   8,   4,  37,  10,  12,  21, -26,   4,  -2,  -4,  -3,   0,  -3, -19,  -6,  20,  19,  -5,  -6,   0,  -4,   3,  -1,   0,   8,  -9,  13, -22,  13, -30,   0,  -6,   3, -31,  12, -12,  17, -19,  12,   2,  41,   4,  -1,  33,  -3,  -7,   3,  -4,  -3,  34,   2, -11, -24,  -1,   1,  -1, -18, -15,  -4,  -6,   6, -20,   8,   7,  14,   4,   5,   1,  -5,  -4,  -2,  -4 },
     {    3,   5,   6,  29,  -6, -19,  -3,  -2,   1,   0,   2,   0,   2,   8,  20,  -9,  10,  -7,  -2,  -2,   0,  -1,  -2,   0,   2,  -3,   9,  37, -14, -26,  13,  -1,   0,  -3,   0,   0,  -4,   6,   0,  -4, -12,   4,   2,   1,   0,   1,   0,   0,  -1, -16,   7,  12, -18, -43,  15,   2,  -4,   7, -12,   3, -50,  34,  35, -12,   0,  -6,  -2,   1,  18,  32, -23,  -4,   0,  10,  24,  -9,  -8,  -9,  -7,   7,   4, -10,  -4,  -9,  -7,   2,   0,   9,   6,   5,   0,  -2,  -5,  -9,   3,  -1 },
     {    9,  17,  21,  18,  16,  27, -35,  -1,  -7,  -6,  -4,  -1,  -4,  -2, -11,  -2, -43,   5,   6,   8,   4,   5,   1,   1,   5,  15,  16,  12, -10, -20,  26, -14,   5,  -1,  -3,   0,   0,   4,   6,  28,  37,  -7, -22,   1,   1,  -8,  -2,  -1,  -8,  -2, -24, -42,   5,   0,  -6,  15,  -2,  -1,  17, -20, -24,  -4,  13,   1,  -4,  -2,   2,   6,   3,   1,   7,  -4,   0,   1, -13,   2,  13,   1,   2,  -2,  -6,  26,   7,  -7,  17,  -4,  -9,  -5,   4,  -7,  -3,   5,  -5,   3,  -1,   0 },
     {   -1,  -5, -13,   2, -13,  13,   5,  -2,   1,   4,   1,   1,  -2, -13, -12, -26,  22,  -7,   4,   6,   4,   0,   1,   0,   3,   1, -16,   3, -14,   9,  16,  -5,   1,  -4,   2,   0,   3,  -5,  -3, -20,  15,  26,  -8,  -4,   1,   2,  -1,   0,  10,  13, -16,   2,  -2,   0,  13,   2,   4,  10,  12,  -7,  16,  41, -14,  -8,  15,  13, -12,  12,  45, -23, -21,  -2,   4,   5,  -4,  10, -14, -14,   7,  -5,  28,  21,  -5,  -6,  36, -28, -13,   2, -11, -19,   7,   7,  -7,  17,   3,  -3 },
     {   -2,  -2,  -8, -10, -32,  22,   3,  16,   0,  -2,  -1,  -2,  -2,  -4,  13,  -5,  20, -22,   5, -10,  -4,   2,   1,   0,  -2,  -4, -10,  -9, -33,  -2,  11,  -3,  -2,   3,  -2,   1,   0,   8,  30,  10,  21,   2, -15,  11,  -1,  -4,  -5,  -2,  -2, -18,  -6,  -3,   5,   6,  12,  -4,  -5,   0,  34,  -8,  -2,   5, -10,  -8,   2, -48,  -6,   9,  -3, -21,   0,   3,  -2,   1,   6,   5,  -1,   8,   4,   3,   8,  -1,   0,  -7, -57,   0,  10,   3,  12,  26,   4,   6,  13, -13,  -9,  -2 },
     {   -4,  -7,  -3, -25, -22,  -3,  21,   2,   3,   0,   4,  -1,  -1,   1, -13,  17,   6,   7,  -1,  -3,  -3,   0,  -2,   0,  -3,  -4,  -1, -36,   1,  -4,  -4,   9,   0,  -1,  -2,   1,   1,   3,  -5,  16, -13,  35, -17,  -1,  -8,   1,  -2,   0,  -4,  -4,  -5, -28,  14, -24, -12,  11,  -5,  -1, -11, -18,  -1,   0,  42,   1,   3, -14,  32,  -4,  18,  20, -18, -16, -12, -27, -29,  13,  14, -20,  -3,   6,  10,   3, -10,  23,  -4,   3,  16,  -6, -14,  -7,  12,  -4,   7,   4,  -6,   5 },
   }
 },
 {  //16
   {
     {  117, -42,  -6,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0,  23,  -9,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  -7,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -8,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   18, -15,   3,  -1,   1,   0,   0,   0,   0,   0,   0,   0,-115,  26,   9,   5,   3,   2,   1,   1,   1,   1,   1,   0, -36,  14,   2,   1,   1,   1,   0,   0,   0,   0,   0,   0,   8,   3,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  12,  -1,  -1,   0,  -1,   0,   0,   0,   6,  -1,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   12, -10,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0, -25,  17,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0, 106, -11, -10,  -3,  -3,  -2,  -2,  -1,  -1,   0,  -1,   0,  54, -19,  -4,  -2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0, -11,   1,   0,   1,   0,   0,   0, -15,  -2,   2,   0,   1,   0,   0,   0,  -7,   1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    4,  10,  -3,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,  27,  -3,  -3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -39,  21,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,  91,  -1, -11,  -2,  -3,  -1,  -2,  -1,  -1,   0,  -1,   0,  65, -19,  -6,  -2,  -2,  -1,  -1,   0,  10, -15,   0,  -1,   0,   0,   0,   0, -12,  -6,   2,   0,   1,   0,   0,   0,  -7,   0,   1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,  -1,   0,   0,   0,  -1,  -1,   0,   0,  -1,   0,   0,   0 },
     {  -10, -19,   6,   1,   2,   1,   1,   1,   1,   0,   0,   0,   0, -20,   5,   1,   2,   0,   1,   0,   0,   0,   0,   0, -35,   9,   4,   1,   1,   1,   0,   0,   0,   0,   0,   0,  47, -19,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -73, -11,  10,   1,   4,   1,   1,   0, -69,  20,   9,   2,   2,   1,   1,   0, -12,  20,   0,   0,   0,   0,   0,   0,  12,   7,  -3,   0,  -1,   0,   0,   0,   6,  -1,  -1,   0,   1,   1,   0,   0,   1,   1,   0,   0,   0,   1,  -1,   0 },
     {  -36,-100,  39,   3,  12,   2,   5,   2,   3,   1,   2,   1,   1, -45,  17,   1,   4,   0,   2,   1,   1,   0,   1,   0,   8,   1,   3,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,  -2,  15,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  23,   6,  -6,   0,  -1,   0,  -1,   0,  14,  -2,  -3,   0,  -1,   0,  -1,   0,  -1,  -3,   0,   0,   0,   0,   0,   0,  -4,   1,   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0 },
     {   -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   4,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   2, -15,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0, -41,   8,   4,   1,   1,   1,   0,   0,   0,   0,   0,   0,  58, -22,  -3,  -1,  -1,   0,   0,   0, -76, -14,  13,   1,   4,   1,   2,   0, -59,  22,   7,   1,   2,   1,   1,   0,  -3,  16,  -1,   1,   0,   0,   0,   0,  11,   6,  -2,   0,   4,   0,  -1,   0,   3,   1,  -1,   0,   0,  -1,   0,   0 },
     {    1,   8,  -4,   0,   0,  -1,  -1,   0,   0,   0,   0,   0, -10, -23,   4,   1,   2,   0,   1,   1,   0,   0,   0,   0,  -4,  -9,   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,   3, -20,   4,   0,   1,   1,   1,   0,   0,   0,   0,   0, -33,   8,   4,   2,   0,   1,   1,   0,  40, -23,  -1,   0,  -1,   0,   0,   0, -79, -17,  16,   1,   4,   1,   2,   0, -60,  24,  10,   1,   2,   0,   1,   0,  -9,  22,  -1,   0,  10,   7,  -4,   0,   2,   0,  -1,   0,   2,   2,   0,   0 },
     {  -12, -38,  15,   0,   3,   1,   2,   0,   1,   0,   0,   0,  26,  91, -27,  -4,  -9,  -3,  -3,  -2,  -2,  -1,  -1,  -1,  -4,  55, -19,  -2,  -4,   0,  -1,  -1,  -1,   0,  -1,   0, -10,   4,  -6,   1,   0,   1,   0,   0,   0,   0,   0,   0, -18,  -7,  -1,   2,   1,   1,   1,   1,  -1, -11,   1,   2,   1,   0,   1,   0, -15, -10,   4,   1,   1,   0,   0,   0, -19,   2,   5,   1,   1,   0,   0,   0,  -2,   3,   0,   0,  -1,   1,   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0 },
     {    3,   4,  -2,   1,  -1,  -1,   0,   1,   0,   0,   0,   0,   2,  -3,   4,  -2,   0,  -2,   0,   0,   0,   0,   0,   0,   9,  29,  -5,  -2,  -3,  -2,  -1,  -1,   0,   0,   0,   0,   3,   9,  -4,  -1,  -1,  -1,   0,  -1,   0,   0,   0,   0,  -8,  26,  -4,   1,  -1,   0,  -1,  -1,  34, -14,  -3,  -2,  -1,  -1,  -1,   0, -53,  26,   4,   0,   1,   0,   1,   0,  80,  16, -16,   0,  -4,   0,  -2,  -1,  39, -26,  -8,   0,   4, -15,   2,   0,  -8,  -7,   3,  -1,   1,   1,   1,   0 },
     {    2,   2,   0,  -1,   1,  -2,  -1,   0,   0,   0,   0,   0,   1,   3,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   3,  -5,   3,  -1,  -1,   0,   0,   0,   1,   0,   0,   0,   8,  19,  -2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   4,  -1,  -1,   0,  -1,   0,  -1,  -2,  21,  -6,   0,  -1,   0,   0,   0,  23, -10,  -3,  -1,   0,   0,   0,   0, -36,  33,   0,   0,   0,   0,   0,   0,  90,  17, -22,   0,  42, -32,  -9,   0,   1, -18,   2,   1,  -9,  -5,   4,   1 },
     {    7,  24,  -8,  -1,  -2,   0,  -1,   0,   0,   0,   0,   0,  -6, -32,  22,  -2,   3,   0,   2,   0,   1,   0,   0,   0,  22,  67,  -7,  -8,  -3,  -3,  -1,  -3,  -1,  -1,  -1,  -1,   1,  75, -20,  -6,  -5,  -2,  -1,  -2,   0,  -1,   0,  -1,  -9,  24, -13,  -1,  -2,   0,  -1,  -1, -27,  -1,  -6,   1,   0,   0,   0,   1,  -3, -12,  -2,   2,   0,   0,   1,   1, -20, -12,   5,   1,   1,   1,   0,   0, -21,   7,   4,   1,   6,   5,  -1,   0,   7,   0,  -2,   0,   2,  -3,   0,   0 },
     {  -14, -31, -95,  31,   0,  12,   4,   4,   1,   3,   2,   2,  -8, -19, -58,  22,  -1,   9,   2,   3,   1,   3,   1,   1,   2,   7, -12,  10,  -1,   2,   1,   0,   0,   0,   0,   0,   0,   4,   3,   3,  -1,  -1,   0,   0,   1,  -1,   0,   0,   3,   6,   7,  -3,   1,  -2,   0,   0,  -1,   2,   4,   0,   0,  -1,   0,   0,   1,  -4,   4,   0,   0,   0,   0,   0,  -1,  -2,   4,   0,   0,   0,   0,   0,   3,  -3,   3,  -1,   1,  -2,   3,  -1,   0,   0,   4,   1,  -3,   1,   1,  -1 },
     {   -2,  -6,  -1,   0,   0,  -1,  -1,   0,   0,   0,   0,   1,   4,  13,  -6,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2, -12,   3,   3,   0,   0,   1,   1,   0,   0,   1,   0,   8,  -2,   7,  -2,   1,  -1,  -1,   0,   0,   0,  -1,   0,   6,  29,  -3,  -3,  -2,   0,   0,  -2,   2,   7,  -2,  -3,  -1,  -1,   0,   0,   0,  27,  -5,   0,  -1,   0,  -1,   0,  21,  -6,  -2,  -2,   0,  -1,   0,   0, -37,  35,   3,  -1,  81,  22, -23,  -1,  41, -36,  -9,   1,  -2, -14,   4,   1 },
     {    2,   7,   3,   2,  -4,  -1,  -1,   1,   0,   0,  -1,   0,  -5, -34,   5,   0,   3,   2,   2,   0,  -1,  -1,   1,   1,   8,  47, -26,  -1,  -1,   0,  -1,   0,  -2,   0,   0,   0, -23, -35, -11,   7,   3,   2,   2,   3,   0,   0,   0,   1,  -3, -75,  11,   7,   3,   3,   3,   3,  11, -35,  18,  -1,   4,   2,   0,   0,  26,  -1,  10,  -1,   0,  -1,  -1,  -1,  17,   8,   5,  -3,  -1,  -2,   0,  -1,   7,  19,  -2,  -3,  28,   6,  -6,  -2,  13,  -4,  -3,  -1,   2,  -4,  -1,   0 },
     {    0,   2, -10,   7,  -3,   0,  -2,   0,   0,   1,   1,   0,   3,  -7,  27,  -4,   1,   1,  -3,  -1,   1,   0,   0,   0,   7,  54,  -2,  -4,  -4,  -2,  -2,   0,  -1,  -1,  -1,   0, -14, -60,  24,   2,   1,   0,   2,   1,   2,   1,   1,   1,  27,  17,  16,  -8,  -1,  -4,  -1,  -2,   2,  63, -19,  -9,  -5,  -2,   0,  -2, -10,  23, -21,  -5,   0,  -1,   0,   0, -24,  -6,  -8,   3,   2,   1,   2,   2,  -6,  -8,  -4,   1,   2, -10,  -1,   0, -10,   1,   5,   4,   0,   3,  -3,   1 },
     {   -1,  -2,   2,   0,  -1,   0,  -1,  -1,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,  -2,  -1,   2,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,  -6,   3,   1,   0,   0,   0,   0,   1,  -1,   0,   0,  -3,   6,  -7,   0,  -1,   0,  -1,   0, -11, -26,   3,   2,   1,   0,   1,   0,   2,  -4,  -2,   0,   0,   1,   0,   0,   3, -28,   7,  -1,   1,   0,   1,   0, -21,  15,   2,   0,  38, -42,  -5,   2, -89, -17,  26,   1, -19,  37,   4,  -2 },
     {   -7, -16, -48,  24,   0,   4,  -2,   3,   1,   2,   0,   1,  10,  26,  69,  -4,  -1, -10,  -5,  -4,  -3,  -1,  -1,   0,   4,   3,  73, -13,  -9,  -6,  -3,  -4,  -1,  -1,  -1,  -1,   0,   0,  12, -10,  -6,   0,  -2,  -2,   0,   0,   0,   0,  -8, -19,  -4,  -1,   2,   1,  -1,  -1,   1, -28,  -1,  -2,  -1,   0,   2,   3,   4, -10,  -1,  -1,  -1,   2,   3,   1,   5,   2,  -3,  -1,   1,   1,   0,   0,   1,  -1,  -3,  -2,   2,  -2,  -2,   0,   8,  -2,  -2,  -1,  -1,  -1,  -1,   0 },
     {    0,   1,   1,  -4,   2,   3,   0,  -1,   1,   0,  -1,   0,  -1,   3, -16,   2,  -2,   3,   1,  -1,   1,  -1,   0,   1,   5,  -4,  23,  -5,   0,  -1,   0,   1,   1,  -1,  -1,   0,   3,  48,  -1,  -7,  -3,  -4,  -3,   0,  -1,  -1,  -1,  -1,  -9, -50,  16,   2,   2,   3,   1,   0,  34,  28,  23,  -9,   1,  -2,  -1,  -2,  -4,  74, -13, -10,  -1,  -3,   2,  -1, -22,  13, -13,  -2,  -2,   1,   1,   0, -20,  -3,  -4,   2,   2,  -6,  -2,   1, -17,   5,   1,  -1,  -2,  13,  -5,  -1 },
     {   -8,  -9, -35, -79,  28,   1,   9,   1,   3,   2,   3,   1,  -3,  -2,  -7, -72,  26,   1,   6,  -1,   3,   3,   2,   0,   0,   2,   4, -20,  17,  -5,   1,  -1,   3,   0,   0,  -1,   0,  -3,   6,  11,   3,  -2,   0,   1,   0,   0,   0,   0,   1,  -4,   0,  18,  -1,   0,  -1,   0,   1,   1,  -3,  17,  -5,   1,  -1,   2,  -1,   0,  -2,   9,  -2,   2,   0,   1,   0,  -2,   3,   6,  -2,   1,  -1,   0,   1,  -4,   3,   1,   1,   2,   0,   1,   2,  -4,   2,   3,  -9,   2,   1,   1 },
     {   -1,  -3,  -9,  -5,   5,   2,   2,  -1,   2,  -1,   2,  -1,   3,   5,   5,  -9,   0,   1,   1,   0,   0,  -2,   1,   0,  -1,  -1,  -7,   0,  -2,   4,  -1,   0,  -1,   0,   1,   1,   4,  -2,   2,   4,   1,  -1,  -2,   2,   1,  -1,   0,   1,   2,  21,  -2,  -2,  -5,   0,   2,   1,  -1, -14,   8,  -2,   0,   5,   0,  -1,  12,  18,  -2,  -2,   1,   1,  -3,   0,   2,   5,  -1,   2,   2,  -1,  -1,  -1,   5,  37,  -9,   1,  12,  -9,   1,   2, -15,  58,  -3,  -2,  93,  -1, -27,  -1 },
     {    3,   8,  25, -22,  12,  -4,  -4,  -4,   0,  -3,   0,  -1,  -5, -14, -48,  16,   5,   3,   1,   3,   1,   1,   1,   1,   8,  27,  49,   3, -15,  -2,  -1,  -2,  -2,  -1,  -1,  -1,   4,   2,  79, -24,  -8,  -1,  -5,  -3,  -3,  -2,  -1,   0,   0,   0,  15, -23,  -1,  -1,  -2,   0,  -7, -18,  -8,  -4,  -1,   0,   0,   1,   6, -16,   2,   2,   1,   1,   0,   1,   7,   2,   0,  -4,   0,  -1,   0,  -1,   4,   5,  -7,  -2,  -2,   5,  -2,   1,   0,   3,  -2,  -2,   4,  -2,  -3,   0 },
     {   -1,   0,  -9,  -8,   7,  10,   2,  -2,  -1,  -2,  -3,   1,   3,   4,  22,   3,   6,  -4,   0,  -3,  -3,  -1,  -3,  -1,  -1,   6, -27,  13,   7,  -2,  -6,   2,  -1,   2,   0,  -1,   4,  -3,  20,   4,   0,   2,  -4,   0,  -4,  -2,  -2,  -1,   5,  36,  13, -10,   2,  -4,  -4,   1,  -7, -28,  24,   4,   2,  -1,  -5,  -1,  32,  21,  32, -15,  -4,  -5,  -1,  -4, -10,  64, -31, -20,  -2,  -4,   1,  -2, -25,   6, -17,   5, -25,  -4,  -3,   3,  -9, -13,   0,  -4, -22,  -9,  10,   1 },
     {   -1,  -3,  -3,   4,  -2,  -1,   0,   1,   1,   0,   3,   0,   2,   6,  28, -15, -10,   2,   2,   0,   0,  -2,   1,  -1,  -3, -12, -36,   6,  -8,   4,   4,   1,   0,   3,   3,   0,   7,  28,  39,  19, -13,   2,  -5,  -2,  -3,  -1,  -1,   0,   1,  -3,  78,   2, -14,  -1,  -6,   1,   3,   4,  36, -17,  -4,   0,  -2,   1, -13, -25,  -4, -12,   6,  -1,   0,   1,   7, -35,  -3,   4,   4,  -3,   3,  -3,  11,  -1,   8,  -2,   7,   2,  -2,  -2,   2,   3,  -5,  -1,  -4,  -4,  -2,   0 },
     {   -2,   1,  -6, -50,  23,  -3,   3,   3,   5,   4,   2,   1,   4,   2,   9,  58,   1, -28,  -7,  -1,   2,   6,   1,  -2,   3,   4,  15,  58, -20, -17,  -6,  -3,   1,   0,  -1,  -1,  -3,  -5, -23,  21, -13,  11,  -1,  -3,   3,   0,  -1,  -3,   2,   2,  15,  35, -15,   0, -16,  -5,  -2,  -1,  -2,  -2, -18, -10,  -4,   3,  -2,   4,  -7,  -5,  -8,   5,   7,   4,  -4,  -5,  13, -10,  -1,   7,  -2,   1,   2,  -7,   4,  -6,   3,  -5,   3,  -3,   3,  -4,   4,  -3,   9,  -5,   4,  -1 },
     {   -2,  -2,  -2,  -2,  -8,  -3,   5,  -3,   2,   0,   1,   0,  -1,  -3,   5,   1,   6,  -3,   4,  -6,  -1,   0,   0,  -1,  -4,  -6, -25,  17,   8,   5,   1,   1,  -2,   1,  -1,   1,   2,  -1,  28,  -1,  -5,   2,   2,   0,  -1,  -1,  -1,  -1, -11,  -5, -26,  -6,   3,  -3,   4,   5,  -3, -47, -19,  -1,   3,   0,  10,   0,   4,  28, -27,  -5,  -1,   7,   2,  -4, -34, -30, -32,  15,   2,   8,   0,   3,   5, -55,   5,  16,  22, -18,  18,   1,  14,   1,  -2,   2,   6, -30,   2,   4 },
     {   -5,  -7,  -3, -42, -73,  39,  15,   9,   1,   3,   0,   3,  -2,  -5,   3,  -4, -67,  26,   6,   6,  -2,   0,   1,   4,   2,   5,  15,  13,  -9,  11,  -3,   3,   1,  -2,  -3,   2,   1,   0,   3,  17,   9,  -1,  -2,   0,  -1,  -1,  -1,  -3,   1,   3,  -7,   5,  18,   0,  -2,  -3,   0,  -1,  -1,   1,  17,  -1,  -4,  -3,   1,   0,   2,   0,  12,  -4,  -2,  -1,   0,  -1,   1,  -3,   3,  -1,   4,   0,   2,   2,  -4,  -1,  -2,   1,  -1,  -1,  -1,  -3,  -1,   1,  -1,  -6,  -2,   2 },
     {   -2,  -3, -18, -18,  21,   3,   2,   6,  -1,   0,   1,   0,   2,   5,  25,   7, -20,  -3,  -6,   0,  -1,  -1,   0,   1,  -5, -13, -45,  29, -17,   8,   2,   3,  -2,   4,   0,  -1,   1,  15,  27, -49,   3,  -4,   2,   6,   0,   2,   1,   1,  -5, -34, -26, -10,   6,   4,   4,   1,   5,  24, -34,   4,   3,  -1,   2,   4, -11, -18, -16,   7,   0,  10,   2,   0,  20,  16,  -7,   8,   5,   1,  -2,  -4, -13,  29,  -6,  -8,  -8, -11,  -4,   8,  -3,   6,   1,   6,  -5,  -2,   4,   5 },
     {    3,   2,   9,  29, -16,  15,   4,   6,  -4,   3,  -3,   0,   0,   4,  -3, -34,  29,  10,   1,  -3,   0,   0,  -1,   2,   5,   4,   9,  68,  43, -13, -10,  -4,   2,  -3,  -2,  -1,   2,  11,  18,  44, -10, -18,  -8,   4,   2,  -5,  -1,  -2,  -5, -16, -20,  -4, -13, -10,  -1,   1,   0,   9, -18, -13, -12,  -7,   4,   4,  -5,  -8,  -5,  -7,  -5,   1,   2,   2,   9,   9,  12, -12,  -5,   2,   4,   0,  -6,  10,  -9, -13,  -6,  -1,  -4,   1,  -4,  -1,  -5,  -2,   1,   2,  -5,  -3 },
     {   -1,  -2,   4,  -4,  -9,  -7,  10,   3,   1,   2,  -1,   2,  -2,  -2, -14,  -5,   3,  -4,   1,   0,  -2,   2,  -1,   0,  -1,  -8,   8, -13,  -7, -14,   5,  -3,   2,  -1,   1,   3,  -3,   4, -29,  16,  -9,   5,   0,   2,   4,   0,  -2,   1,   1, -12,  39, -18,   1,   1,  -5,  -3,  -7,   1, -37, -15,  15,  -3,   8,   2,  -4, -35, -12,  -1,   7,   6,   3,   2,   9,  28, -33,  -5,   0,   0,  10,  -2, -36, -16, -34,  20,  16, -36,  19,  16,  20,  -2,  14,   3,  17,  13,  -3,  -3 },
     {    0,   0,   5,  12,   1,  16, -19,  -4,  -3,  -5,   5,   2,  -4,  -6, -15, -32,   4,  -2,   2,  -6,  -7,  -6,   1,   1,  -1,   0,  11,  16, -31, -11,   1,  -5,  -5,   3,   3,   0,  -4,  -7, -17, -19,  -2,  22,  -7, -16,  -5,   1,   2,  -1,  -1,  -2,  -8,   9,  19, -17, -16,   2,   2,   9,  30,  -5, -23, -10,   2,  -2,  -9,  -8,  -6,   5,  -7,  -6, -13,  -7,   0, -22, -16,  -4,  -3, -25,  -9,   4,  -7,  18, -28,  13, -12, -19,  12,   6, -11, -14,  -2, -10,   8, -57,  24,   3 },
     {   -1,  -2,   1,  -4, -20,  -4,  -5,   3,  -2,   5,  -3,   1,  -1,  -2,  -2,   5,  -4, -24, -21,   9,   8,   4,  -5,   7,   0,   0,  -2, -10,   5, -19,   5,  -6, -16,   6,   4,   3,   0,   1,  23,   1, -29,   9,  -4,   6,  -1,  -3,   5,  -4,   2,  -6, -36,  25,  -9,  -6,  21,   8,   2,  32,  36,  13,  -2, -24,   8,  -4,  -1, -25,  38,  -2,  -4,  15, -16, -10,   3,  14, -15,   4,  -7, -15,   3,  -9, -12, -25,  20,   9,  24, -18,   2,   4,   5,  14, -12,  -6,   6,   9, -13, -10 },
   },
   {
     { -101,  41,   3,   5,   1,   2,   1,   1,   0,   1,   0,   0,  55, -20,  -2,  -2,  -1,  -1,   0,  -1,   0,   0,   0,   0, -25,   7,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,  13,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -6,   0,   1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },
     {   61, -20,  -4,  -3,  -1,  -1,   0,  -1,   0,   0,   0,   0,  76, -27,  -2,  -3,  -1,  -1,  -1,  -1,  -1,   0,   0,   0, -61,  17,   3,   2,   1,   1,   1,   1,   1,   0,   0,   0,  33,  -7,  -3,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0, -19,   2,   2,   0,   1,   0,   0,   0,   9,   0,  -1,   0,   0,   0,   0,   0,  -6,  -1,   1,   0,   0,   0,   0,   0,   3,   1,   0,   0,   0,   0,   0,   0,  -3,  -1,   0,   0,   2,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0,   0 },
     {    9, -16,   4,  -1,   1,   0,   1,   0,   0,   0,   0,   0,  69,  -2, -10,  -2,  -3,  -1,  -2,  -1,  -1,   0,   0,   0,  31, -26,   3,  -1,   1,   0,   1,   0,   0,   0,   0,   0, -67,  25,   2,   2,   1,   1,   1,   0,   0,   0,   0,   0,  50, -13,  -4,  -1,  -1,   0,  -1,   0, -34,   5,   4,   0,   1,   0,   1,   0,  17,   0,  -3,   0,  -1,   0,   0,   0, -12,  -1,   2,   0,   0,   0,   0,   0,   6,   2,  -1,   0,  -5,  -1,   1,   0,   3,   1,  -1,   0,  -3,  -1,   0,   0 },
     {   19,  35, -15,  -1,  -4,   0,  -2,  -1,  -1,   0,  -1,   0,  29, -36,   8,  -1,   2,   0,   1,   0,   1,   0,   0,   0,  75,   4, -12,  -3,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0, -20, -23,   7,   1,   2,   1,   1,   1,   0,   0,   0,   0, -35,  24,   0,   0,   0,   0,   0,   0,  40, -16,  -4,   0,  -1,   0,  -1,   0, -35,   5,   5,   0,   1,   0,   1,   0,  18,   1,  -4,   0,  -1,   0,  -1,   0, -11,  -2,   2,   0,   5,   1,  -2,   0,  -5,  -1,   1,   0,   2,   1,  -1,   0 },
     {   25,  70, -30,  -1,  -9,  -1,  -4,  -1,  -2,   0,  -1,  -1, -29, -53,  21,   3,   6,   2,   3,   1,   2,   1,   1,   1, -35,  34,  -4,  -1,  -1,   0,  -1,  -1,   0,   0,   0,   0, -30,   1,   3,   3,   1,   1,   0,   1,   0,   0,   0,   0,  40,  -5,  -2,  -2,  -1,  -1,   0,  -1, -24,   1,   1,   1,   1,   0,   0,   0,  12,   1,  -1,   0,  -1,   0,   0,   0,  -3,  -2,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   1,  -1,   0,   0,  -1,   0,   1,   0,   1,   0,  -1,   0 },
     {  -11, -20,   8,   2,   2,   1,   1,   1,   1,   0,   0,   0, -12,  14,  -3,  -2,   0,  -1,   0,  -1,   0,   0,   0,   0, -56,  -2,   6,   3,   2,   1,   1,   1,   0,   0,   1,   0, -75,  13,   8,   1,   2,   1,   1,   0,   1,   0,   0,   0,   4,  26,  -6,   1,  -1,   0,  -1,   0,  52, -22,  -3,   0,  -1,   0,  -1,   0, -36,  10,   4,   0,   1,   0,   0,   0,  27,  -1,  -4,   0,  -1,   0,  -1,   0,  -9,  -3,   3,   0,   6,   3,  -2,   0,  -2,  -2,   1,   0,   3,   1,   0,   0 },
     {    8,  26,  -9,   0,  -3,  -2,  -1,   0,  -1,   1,   0,   0,  -2,   5,  -3,   0,  -1,   0,   0,   1,   0,   0,   0,   0, -14, -21,   8,   0,   3,  -1,   1,   1,   1,   0,   1,   0, -38,  15,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -69,  -2,  11,   2,   3,   1,   1,   1,  21,  32,  -8,   0,  -2,   0,  -1,  -1,  34, -32,   0,   1,   0,   0,   0,   0, -44,  21,   6,   0,   1,   0,   1,   0,  36,  -7,  -7,   1, -22,   0,   5,   0,  13,   3,  -4,   0,  -5,  -2,   1,   0 },
     {   24,  68, -23,  -1,  -7,  -2,  -4,  -1,  -2,  -1,  -1,   0,  12,  57, -22,  -1,  -5,   0,  -2,  -1,  -1,  -1,   0,   0,  -8, -51,  15,   1,   4,   0,   2,   1,   1,   1,   1,   0,  25,  29,  -9,  -3,  -3,  -1,  -1,  -1,  -1,   0,  -1,   0,  17, -20,  -2,   3,  -1,   1,   0,   0,  15,  -6,   1,  -2,   0,   0,   0,   0, -26,   3,   1,   0,   1,   1,   0,   0,  18,  -4,  -1,   0,  -1,   0,   0,   0, -10,  -2,   1,   0,   4,   0,   0,   0,  -2,  -2,   1,   0,   0,   1,   0,   0 },
     {   -3, -12,  10,  -1,  -1,   1,   1,  -1,   1,   0,   1,   0,  -4, -29,   2,   4,   2,   1,   1,   1,   0,   1,   0,   0,  14,  11,   6,  -2,   0,  -1,  -1,  -1,   0,   0,   0,   0,  25,  25, -17,   3,  -3,   0,  -2,   0,  -1,   0,  -1,   0,  56, -22,   2,  -1,   0,  -1,  -1,   0,  65,   2, -14,  -3,  -3,  -1,  -1,   0, -22, -34,   7,   2,   2,   1,   1,   0, -18,  27,   0,  -1,   0,  -1,   0,   0,  29, -21,  -5,   1, -29,   6,   7,   0,  14,  -1,  -5,   0, -10,  -2,   3,   0 },
     {   -6, -12, -63,  24,  -6,  10,   0,   4,   0,   2,   0,   2,  15,  42,  54, -26,   2, -10,  -2,  -5,  -2,  -3,  -1,  -1,  -3,   3, -44,  10,  -1,   5,   1,   1,   1,   1,   1,   1,   8, -16,  24,  -2,   0,  -1,  -1,   0,  -1,   0,  -1,   0,  19,  18, -13,  -3,  -1,  -1,   0,  -1,  17, -17,   0,   5,  -1,   1,  -1,   1,  14,  -2,   3,  -4,  -1,  -2,   0,  -1, -22,  -2,   0,   2,   1,   2,   1,   1,  19,   0,   0,   0, -10,  -4,   0,   1,   3,   4,   0,  -1,   0,  -4,  -1,   1 },
     {  -11, -28, -30,  18,  -1,   6,   1,   2,   1,   1,   0,   1, -11, -44,  47, -13,   7,  -4,   2,  -1,   0,   0,   1,  -1,  -2, -41,  -8,   8,   3,   3,   2,   1,   1,   0,   1,   0,   3,  53,  -2,  -5,  -3,  -2,  -2,  -2,  -2,  -1,  -1,   0, -26, -43,   4,   4,   4,   1,   2,   1, -10,  25,   4,  -2,  -2,  -1,  -1,   0, -25,  10,  -2,   1,   1,   0,   0,   0,  23,  -5,   1,  -1,  -2,  -1,   0,   0, -15,   6,   1,  -1,   8,   3,  -2,   1,  -2,   0,   1,   0,   1,   1,  -1,   0 },
     {   -2,   0,  -4,   3,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   9,   0,  -1,   1,   0,   0,   1,   0,   0,   0,   0,  -8,   0,  -2,   1,   0,  -1,   0,   0,   0,   0,   0,   0,  -6, -24,   4,   2,   1,   1,   1,   1,   1,   0,   0,   0, -11,  -3,   5,  -2,   1,  -1,   0,   0, -60,  -8,   4,   2,   1,   2,   1,   1, -78,   8,  17,   1,   4,   0,   1,   0,  -7,  43,  -7,   1,  -2,   0,  -1,   0,  34, -21,  -2,   0, -33,  11,   6,   0,  18,   3,  -4,   0,  -7,  -1,   2,   0 },
     {    1,   1,  13,   0,   1,  -3,   0,  -2,   0,   0,   0,   0, -12, -43,  -5,   4,   2,   4,   4,   2,   2,   2,   1,   1,  -8, -64,  14,   3,   6,   1,   1,   1,   1,   0,   1,   1,  21,   5,  16,  -4,   0,  -2,   0,  -3,   0,   0,   0,   0,   6,  42, -24,   1,   0,  -2,  -1,   0,   1, -47,  20,   1,   1,   0,   1,   1,   0,  27,  -7,  -4,   0,   0,   0,   0, -41,  -6,   8,   5,   3,   0,   1,   0,  28,  17,  -2,  -3,  -5, -12,  -4,   0,  -5,   9,   4,  -1,   5,  -2,  -3,   1 },
     {    0,   2,  10,  -2,   2,  -4,  -1,   0,   0,  -1,   1,   0,   2,  14,   2,  -5,   2,  -1,  -1,  -1,   1,  -1,  -1,   0,   3,  28,  -5,   3,  -4,  -1,  -2,   1,  -1,   0,  -1,   0,  -9,   4,  -6,  -5,   1,   1,   0,  -1,   0,  -1,   0,   0,  -2, -33,   0,   3,   0,   2,   1,   2,   5,   4,   3,  -4,   2,   0,   1,  -1, -43,  -8,  -4,   3,   1,   1,   1,   1, -51,   0,  18,  -1,   3,   0,   2,   1,  15,  46, -13,   0,  40, -40,  -4,   2, -44,  16,  11,  -2,  27,   5, -10,   1 },
     {    6,   7,  28,  53, -27,   0, -10,   0,  -4,   0,  -3,  -1,  -5,  -5, -31, -72,  31,   1,  11,  -2,   5,   0,   3,  -1,   3,   6,   2,  50, -16,  -1,  -9,   0,  -4,  -1,  -2,   0,  -1,  -5,   7, -26,   6,   2,   2,  -1,   0,   1,   1,   0,   1,  -4, -12,  10,   4,  -1,   0,   1,   0,   5,  12,  -2,  -7,   0,  -3,   1,   3,  -6,  -9,  -2,   6,  -2,   2,   0,   5,   4,   6,   3,  -4,  -1,  -1,   0,  -3,  -7,  -4,  -1,  -1,   4,   3,   2,   3,  -2,  -3,  -1,  -2,  -1,   3,   3 },
     {   -8, -21, -71,  28,  -7,  13,   3,   6,   0,   3,   0,   3,  -2,  -3, -50,  12,   6,   3,   2,  -1,   3,   2,   1,   0,  11,  35,  46, -10,  -2,  -5,  -2,  -3,  -1,  -1,   0,  -2,  -2,  18, -25,  -2,  -1,   1,   1,   0,   1,   0,   1,  -1, -12,  -5,   0,   3,   1,  -1,   1,   1,   1, -25,  13,   2,   0,   3,   2,   0,   6,  26,  -7,   1,  -1,   0,  -1,   0,  -1, -10,   3,   3,   1,   0,   0,   1,  22,   0,  -3,   0,  -1,  -2,   1,   0,   3,  -1,  -2,   1,   1,  -5,   4,   0 },
     {   -2,  -3,  -8,   6,  -2,  -1,   0,   0,   0,   0,   0,   0,  -2,   1,  -2,  -2,   0,   1,   2,   1,   0,   0,   0,   0,  -1,   4,   4,  -4,   2,  -1,   1,   1,   0,  -1,   0,  -1,  -3,   6, -10,   0,   0,  -1,   1,   0,   0,   0,   0,   1,  -4,   2,  -1,   1,   0,  -2,   1,  -2,  -6, -22,  -4,   4,  -1,   0,  -1,   0, -18, -21,   9,  -3,   2,   1,   2,   1, -69,  -7,   6,   1,   1,   0,   2,   0, -74,  12,  15,   1,   9,  45,  -6,   0,  32, -17,  -6,   1, -18,   4,   4,  -1 },
     {    7,  16,  36, -16,   4,  -3,  -2,  -4,   0,  -1,  -2,  -1,   7,  17,  18,  -2,  -3,  -3,   1,  -1,  -1,   0,  -2,   0,   9,  48, -26,  -4,  -5,   1,   0,  -1,   0,  -1,  -1,   1,   4,  71,   0, -17,   0,  -6,  -3,  -2,   0,  -2,  -1,  -2, -24,  -2, -27,   3,  -2,   4,   1,   1,  -7, -33,  13,   3,   3,  -2,  -1,   0,   2,  26, -10,  -4,   1,  -1,  -2,  -1,  -1, -18,   0,   4,  -1,   0,  -2,   0,  12,  -8,   5,  -3, -27,   4,   0,   2,  17,   3,  -1,   1,  -4,  -6,  -2,   0 },
     {   -3,  -5, -14, -20,  -5,   6,   6,   1,   1,   1,   1,   0,  -3,  -6, -35,  32,  -3,   0,  -2,   2,  -3,   2,   1,   3,  -6,  -1, -30,   6,  -4,   6,   3,   1,   2,   0,   1,   0,  -4, -26,  56,  -9,   5,  -3,   0,  -2,  -2,  -2,   0,  -1,  -1, -45, -31,  22,   4,   2,   0,   2,  17,  14,  42, -15,  -5,  -4,  -1,  -2,  -7,  -1, -31,  11,   3,   0,  -1,   2,  -3, -12,  15,  -1,   1,   0,   2,   0, -10,  -3,   5,  -3, -21,   7,   1,   2,  13,   4,  -1,  -1,  -2,  -4,   1,  -1 },
     {    1,   2,   1,  -5,   0,   2,  -3,  -1,  -1,  -1,  -1,   0,   1,   0,  -5,  10,  -1,  -5,   3,  -1,   1,   1,  -1,   0,   1,   2, -14,   1,  -3,   2,   2,  -1,   1,  -1,   0,   1,   4,  10,  15,  -5,   0,  -1,   1,   0,   0,   0,   1,  -2,   2,   6,  -9,   7,  -1,   1,   1,   1,   0,   8,   4,  -2,  -4,  -1,  -2,   1,  -1,   5,  -5,  -2,  -1,  -1,   1,   1,   4,  24,  -7,   3,  -3,  -1,  -2,   0,  40,   7,  -5,  -2,  81,   7, -19,  -1,  35, -56,  -7,   3, -35, -11,  10,  -1 },
     {   -3,  -6, -18,   3, -15,  -3,   5,   3,   3,   1,   2,   2,  -2,  -3, -29,  16,  12,  -2,   2,   0,  -4,  -1,   2,   0,   4,   9, -11,  -7,  -4,   9,   4,  -2,   0,   0,   0,  -1,  14,  48,  34,  -1,  -4, -12,  -6,  -4,  -2,  -3,   0,   0,   2,  57, -29,  -7,  -3,   2,  -2,  -1, -22,  25, -10,  -1,  -1,  -4,  -2,   1, -10, -45,  10,   9,  -1,   1,  -1,   2,   7,  27,   0,  -8,  -5,   0,   1,   0, -19,  -7,   1,   3,   6, -11,   5,   1, -18,  12,   6,  -2,  21,   0,  -3,   0 },
     {    1,   1,   6,  -1, -16,   9,  -5,  -1,  -4,   3,  -2,   0,   6,  13,  47,  -4,   0,  -1,  -8,   2,   4,   0,   0,  -1,   5,   7,  63, -28,  -7,   4, -10,   1,  -3,   1,  -3,   1,   2, -10,   0, -17,  14,  -6,   5,  -3,   4,   1,   2,   0,   7,  28, -25,  22,  -6,  -5,   2,   2,   1,  38,  13, -24,   5,   2,   3,  -3, -15,  -1, -34,  12,   4,   1,   2,  -1,   2,  -9,  15,  -6,   1,  -1,  -1,  -3,  -4,  13, -22,   3,  -8,   0,   2,   5,  20,   2, -11,  -2,  -4, -14,  -2,   4 },
     {   -2,  -3,  -2, -10,   6,   4,   3,  -4,  -4,  -6,   2,  -2,   1,   0,  21,   4,   3,   3, -12,   0,   7,  -2,   0,  -4,  -1,  -5,  25,  -7, -16,   5,  -9,   7,  -2,   1,  -2,   3,  -4,  -7,  -3, -14,  11,   2,   6,  -2,  -1,   3,   3,   6, -11, -26, -40,  17,  -8,  -3,  11,   7,  -7, -41,  17,  -5,   2,  10,   8,  -6,  17, -36,   9,  -9,  15,  -1,   5,  -5,   3,  23, -15,   5,  -3,  -1,  -1,  -3,  -5, -47,   3,   0,  22,   5,   1,  -6, -50,   2,   3,   2, -18,  12,   4,   3 },
     {    2,   4, -11,  17,  71, -15,  -3, -15, -10,   0,   4,   0,  -2,  -2,   0, -13, -59,  14,  11,   7,   3,  -3,  -1,   0,   5,  12,  15,  31,  41,  -8, -12,  -5,  -7,  -2,   2,   0,   1,   4,  15, -16, -23,  -8,   6,  -6,   3,   3,   0,   1,   3,  10,  -7,  11,   1,   4,  -1,   3,  -1,   9,  -2,  -4,  -3,   1,  -5,  -1,  -4,  -2, -15,  11, -14,   1,   2,   7,  -1,   6,   7,  -3,   9,  -3,  -4,  -3,  -3, -13,   0,  -1,  -1,   4,  -4,   1, -10,  -6,   6,  -1,  -2,  11,  -2,   1 },
     {   -7,  -6, -24, -81,  10,  11,   7,   2,  -1,  -2,   1,  -1,   0,   0,   0,  -7,  27,  -2,  -2,  -5,   2,   0,   2,  -4,   2,   0,  13,  65, -16,  -3,  -5,  -1,  -5,  -1,  -1,   2,   0,   4, -11, -30,  -6,   4,   1,   7,  -2,   1,   0,   2,   4,  18,  14,  12,  12,   1,   3,  -7,  -5,   3, -17,  -3,  -3, -13,  -4,  -3,  -5,   1,   1,   1,   2,   3,   0,   1,  -5, -12,   1,  -7,  -1,  -2,   1,   0,   0,   0,   2,  -5,  -2,  -6,  11,   2,  -1,   1,  -2,   3, -12,  -6,   8,   2 },
     {   -1,  -1,   1,   7, -16,   4,  -1, -11,   5,   0,  -3,  -4,  -1,  -1,  -4,   0,  -3,  -7,   0,  -6,  -5,   1,  -3,   0,  -1,   0,  -7,  -4,   5,   9,  14,   2,   5,   0,   0,  -5,  -1,   1, -12,  19,   3,   6,  -8,  -8,  -8,  -1,  -3,   0,  -1,  17,  -1,  21,  -1, -13,   4,   2,  -9,  13, -12,   0, -25,   8,  -2,   6, -19, -15, -17,   6,   5,  -3,  -6,  -2,  -5, -51,   4,  -4, -11,  -7,   2,  -1,  18,  -8,  12, -10, -21, -10,  -6,  -6, -30, -40,  29,  -2, -41,  47,   7, -13 },
     {    2,   1,   4,  32,  -2,  10,  -3,  -5,   3,   2,   1,  -2,   1,  -1,  -3,  41,  -5,  12,  -4,  -5,   1,   4,  -4,   3,  -5,  -9, -41,   3,  -4,  14,   5,  -1,   2,  -6,  -1,   5,  -4,  -5, -45, -59,   6,   9,   5,   2,   6,   8,   3,  -1,   4,  14,   8,  34,   3,  -4,   2,  -5,  -2,   3, -18, -45,  10,  -3,  15,  -4,   5,  16,   6,  17,   4,   2,   0,   1,  -7,   7,   0,  -7,  -2,  -1,  -5,  -1,  -5, -11,  -4,   6,   2,   0,   3,  -2,  -3,   3,   4,  -3,  11,   0,   5,   1 },
     {    2,   3,  10,  17, -16,   3,   7,   2,   8,   0,  -6,  -3,   4,   5,  15,  32, -10,   3,  -8,  -4,  -5,  -1,  -4,  -1,   5,   9,  34,  28, -17,  -8,   2,   7,   0,  -4,  -2,  -1,   3,   7,  49, -17,  -7,  -1, -16,  -6,  -1,   0,  -5,  -5,   4,  -6,  22,   2, -13, -21,   4,   6,   5,  15, -14,  -2, -12,  11,   4,   2,   1,  22,  20, -17,  14,  -1,  -6,  -3, -18, -10, -31,  17,  -8,   1,   5,   4,   5, -10,  17, -18,   0,  15, -10, -11,  -7, -13, -12,  10,  37,  11,  -5, -12 },
     {    1,   2,   2,  -7,  -3,  18,   7,   2,  -7,  -9,  -3,   0,   1,   3,   1, -12,  24,   2, -25,  -6,   8,  -1,   6,   1,   1,   2,  -9,   3,  14,  -6,  14, -14,  -1,   1,  -2,   0,   0,   3, -21,  35, -26,   9,   2,  11,   3,  -3,   0,   1,   3,  12,  -6,   7,  -3,  17,  -3,  -6,   6,  22,  21, -16,  19, -14,  -2,  -7,  -2,  26, -35,  -1, -21, -13,   3,   1, -19,  25, -15,  17,   0,  11,  11,  -6,  -6, -25,  12, -12,  -4,  21, -28,   3, -30, -16,  -6,  15,  23,  -1,   4, -10 },
     {   -3,  -4,  -4,  10, -22, -59,  14,   5,  11,   3,   7,  -2,   2,   3,  12,  16,   9,  48, -19, -20,  -7,  -4,  -2,   2,  -1,   1,   4,   6,   4, -35,  23,  -5,  -6,   3,   1,   5,  -3,  -7,  -6, -17, -27,   4,  17,  11,   8,   3,   5,   1,  -2,  -4, -18, -13,   5,  28, -25,   3,  -1,   1,  -6,  12, -19, -11,  10,  -1,  -3,  -8, -11, -13,   8,   4,  -2,   1,   4,  -3,  -6,   0,  -9, -10,  -9,  -2,  -2,   6,   4,  -9,  -8, -13, -14,   4,  -1,  -3,   5,  -5, -18, -17,   4,   3 },
     {    1,   1,   1,   8,   9,  -6,  11,  11, -21,   0,  -2,  -1,  -2,  -4,  -2,   2, -22, -24,  19,   1,  12,  -4,   7, -12,  -3,  -8,   1, -14, -31,  20,  -2, -23, -16,   3,  -4,   5,  -2,  -8, -10,  -7, -11,   9,  10,  31,  13,   1,  -3,   8,   1,  -2, -19,  -1,  39,  14,  16,  -5,  -6,   1, -19,  16,   1,  -1, -13, -18,  -9, -23, -10,  -3,   5, -16,  -7,   7,   2, -28,  -2,  28, -16,   2,  -2,  -8,  14,  -4,  25, -18,   3,  -1,  -1,   2,  12,  -4,  -6,   4,  34, -10,  16,  10 },
     {   -1,  -2,   9,  -2, -48,  27,  -7,  -4,   8,   6,  -2,   7,   1,   2,   6,   5, -26, -18,  25,  18,  -4,   0,   7,   0,   2,   6,  10,   1,  19,  21,  -8, -19,   3,  -4,   1,  -2,  -3,   0,   8,  -4, -50, -18,  11,  -1,   7,  10,   4,   9,  -1,  -1,  21,   9,  19,  14, -13,  -6,   0,  -6,  21, -14,   3, -15,  -7,  -6,   4,  -1,  26, -11,   8,  -6,   7,   7,   1,  10,   1,  -2,  14,  -5,  -9,   4,  -7,   7,  12, -17,  -6, -14,  -7,  -1, -19,   1,   6,  -8, -30, -20,   2,   4 },
   },
   {
     { -107,  31,   8,   5,   3,   2,   2,   1,   1,   1,   0,   0,  45,  -6,  -5,  -2,  -2,  -1,  -1,  -1,   0,   0,   0,   0,  30, -12,  -3,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0, -23,   3,   4,   1,   1,   0,   1,   0,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0,   8,  -1,  -2,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   30,  -7,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, 100, -26,  -7,  -4,  -3,  -2,  -1,  -1,  -1,   0,   0,   0, -57,   2,   8,   2,   3,   1,   1,   0,   1,   0,   0,   0, -25,  14,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  22,  -2,  -3,  -1,  -1,   0,  -1,   0,  -1,  -4,   1,   0,   0,   0,   0,   0,  -4,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {  -43,  37,  -6,   2,  -1,   0,  -1,   0,   0,   0,   0,   0, -28,  -8,   6,   1,   2,   1,   1,   1,   0,   0,   0,   0, -84,  11,  13,   2,   4,   1,   2,   1,   1,   0,   0,   0,  57,   7, -11,  -2,  -3,  -1,  -2,  -1,  -1,   0,  -1,   0,  23, -12,  -3,   0,  -1,   0,   0,   0, -24,   2,   5,   1,   1,   0,   1,   0,   1,   3,  -1,   0,   0,   0,   0,   0,   7,  -1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   30,  73, -45,   0,  -9,  -1,  -4,  -1,  -2,   0,  -1,   0,  -7, -26,   9,   4,   2,   2,   1,   1,   0,   0,   0,   0,  20, -63,  34,  -4,   5,   0,   2,   0,   1,   0,   0,   0,  -8,  29,  -9,  -3,  -1,  -1,   0,  -1,   0,   0,   0,   0,  -2,  16, -12,   2,  -2,   0,   0,   0,  -2, -12,   4,   2,   0,   0,   0,   0,   0,  -3,   4,  -1,   1,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   12,   8, -10,   1,  -2,   1,  -1,   0,   0,   0,   0,   0,  45, -21,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  26,   9,  -2,  -2,  -1,  -1,   0,  -1,   0,   0,   0,   0,  77, -21, -11,  -3,  -3,  -1,  -1,   0,  -1,   0,   0,   0, -67,  -9,  13,   2,   3,   1,   2,   1, -17,  19,   1,  -1,   0,   0,   0,   0,  27,  -3,  -7,  -1,  -1,  -1,   0,   0,  -3,  -7,   1,   1,   0,   0,   0,   0,  -7,   1,   2,   0,   0,   3,   0,   0,   1,  -1,   0,   0,  -1,   0,   0,   0 },
     {   -4,  11,  -7,   1,  -2,   0,   1,   0,   0,   0,   0,   0, -24, -36,  17,   2,   5,   1,   1,   1,   1,   0,   0,   0, -38,  36,   2,  -2,   0,   0,  -1,   0,   0,   0,   0,   0, -37,  16,  -4,   2,   0,  -1,   1,   0,   0,   0,   0,   0, -63,   2,  10,   4,   3,   2,   1,   0,  66,   3,  -7,  -2,  -3,  -1,  -2,  -1,  18, -11,  -2,  -1,   0,   0,  -1,   0, -24,   3,   4,   0,   2,   0,   1,   0,  -1,   3,   0,   0,   6,   0,  -1,   0,   1,  -1,   0,   0,  -1,   1,   0,   0 },
     {   12,  27,  46, -25,  -1,  -6,  -2,  -3,  -1,  -1,  -1,  -1,  -8, -63,   2,   8,   7,   3,   3,   2,   2,   1,   1,   1,  28,  21, -48,  16,  -4,   2,   0,   1,   0,   0,   0,   0,   7,  43,   4,  -5,  -4,  -2,  -3,  -2,  -1,  -1,  -1,   0,  21, -37,  16,  -6,   1,   0,  -1,   1,  -8, -11,   0,   2,   1,   1,   2,   1, -13,  15,  -3,   1,   0,   0,   0,   0,   6,   1,  -1,  -1,   0,  -1,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0 },
     {  -13, -34, -48,  24,   2,   8,   4,   4,   1,   2,   1,   1,  -2, -27,  35,  -3,   2,  -1,  -1,   0,   0,   0,   0,   0,  28,  42,  35, -27,  -2,  -8,  -4,  -4,  -1,  -2,  -1,  -1, -11,  19, -39,   6,   1,   1,   2,   1,   1,   0,   0,   0,  14, -35,  -7,  13,  -1,   4,   0,   2, -20,  -4,  23,  -5,   0,  -1,  -1,   0,  -8,  21,   1,  -5,   0,  -2,   0,  -1,  10,   0,  -9,   2,   0,   0,   0,   0,   1,  -7,   0,   2,  -3,   1,   3,  -1,   0,   2,   0,  -1,  -1,   0,  -1,   0 },
     {   -5, -15,  20,  -6,   2,  -1,   1,   0,   0,   0,   0,   0, -13,   2, -11,   3,   1,   1,   0,   0,   0,   0,   0,   0, -15, -21,   6,   3,  -1,   1,   0,   1,   0,   0,   0,   0, -42,  32,   5,  -4,   0,   0,  -1,   0,   0,   0,   0,   0, -27,  -9,  -2,   2,   3,   0,   1,   1, -65,  16,  11,   3,   2,   1,   1,   0,  70,  13, -14,  -1,  -4,  -1,  -3,   0,   8, -17,  -1,   1,  -1,   0,   0,   1, -24,   3,   7,   1,   5,   6,  -2,  -2,   7,   0,  -2,   0,  -2,  -1,   1,   0 },
     {    7,   8,  17,  63, -34,   2,  -8,  -4,  -3,   0,  -2,   0,  -1,   5,   6, -32,   6,   2,  -1,   2,   1,   1,   0,   0,  -6, -14, -35, -51,  32,  -2,   9,   5,   2,   0,   1,   0,   0,   8,   0,  42,  -9,  -3,   0,  -4,  -2,  -1,   0,   0,   1,   5,  28,  11, -14,   2,  -7,  -4,  -1,  -8,  -6, -22,   5,   2,   0,   3,   2,   6, -16,   4,   5,  -2,   3,   1,   5,  -3,   4,   7,  -2,  -1,  -1,  -1,  -2,  -2,   8,  -4,  -5,   6,  -2,  -1,   4,  -3,  -4,   0,   3,  -2,   1,   0 },
     {  -16, -64,  32,   5,   4,   1,   1,   1,   1,   1,   1,   0,  -4, -29,   7,  -3,   2,   1,   1,   1,   1,   1,   0,   1,  -4, -45,  17,   1,   7,   1,   2,   1,   1,   1,   0,   0,  37,  46,  -8,  -7,  -5,  -4,  -1,  -2,  -1,  -2,  -1,  -1, -16,  35, -19,   3,  -3,   1,   0,   0,  15, -34,   5,   3,   1,   3,   0,   1,  -9,  -7,  11,  -5,   0,  -1,   0,   0,  -1,  17,  -4,  -1,   0,  -2,   0,  -1,   3,  -2,  -4,   1,  -3,  -3,   1,   1,  -1,   1,   1,   0,   0,   0,   0,   0 },
     {    4,   2,  -1,   0,   1,   0,   1,   0,   0,   0,  -1,   0,  10,  39, -12,  -3,  -3,  -1,  -1,  -1,  -1,   0,  -1,   0,  15,   1,   5,  -1,  -2,   0,   1,  -1,  -1,  -1,   0,   0,  27,  37, -19,  -2,  -3,  -1,  -1,  -1,  -1,   0,   0,   0,  42, -32,  -6,   3,  -2,   1,   0,   1,  46, -14,   7,  -2,  -2,   1,   0,   0,  54, -11, -11,  -3,  -1,  -1,  -1,   0, -51, -14,  10,   2,   3,   0,   1,   1, -14,  10,   4,   0,  14,   0,  -4,  -1,   5,  -4,   0,   1,  -4,  -2,   2,   0 },
     {   -3,  -2, -21, -24,  17,  -3,   6,  -1,   2,  -1,   2,   0,   9,  37,  45, -17,  -4,  -6,  -4,  -4,  -1,  -2,  -1,  -1, -10, -17, -33,  37, -15,   8,  -2,   2,   1,   2,   0,   0,  -5,  13, -47,  12,   2,   4,   2,   2,   0,   1,   0,   1, -16,   9,  47, -23,   4,  -4,  -2,  -2,  -3, -30,  14,  -3,   2,   0,   0,   1,  -9,  15, -20,   9,  -1,   0,   2,   0,   8,   7,  -4,   1,  -2,   0,   0,   0,  -4,  -6,   8,  -4,  -4,   4,   0,   0,   5,   1,  -3,   0,   0,  -2,   0,   0 },
     {    6,  17,  10,   7,  -5,  -3,  -3,  -2,   0,   0,   0,  -1,   6,  48, -49,   3,   0,   1,   0,   0,   0,  -1,   0,   0,  -7,   7,  13,  -4,   0,   2,  -1,   0,  -1,  -1,   0,  -1,  -1,  29,   6, -11,  -2,  -5,  -3,  -1,   0,  -1,   0,   0, -48, -42, -10,  18,   4,   3,   3,   2,   6, -32,  17,   4,   1,   1,  -1,   1, -35,  40,   6,  -9,   1,  -2,  -1,   0,  11,  14, -11,   2,   0,   0,   0,   0,  11, -20,  -2,   3,  -6,  -1,   5,   0,   1,   5,   0,  -1,   1,   0,  -1,   1 },
     {   -7, -18, -52,  29,  -5,  14,  -1,   4,   1,   1,   0,   2,  -2,   2,   7, -15,   2,  -2,  -2,   1,  -1,   0,   0,   1,   0,  -4, -19,  27,   6,  -5,   3,   0,   0,  -1,  -1,  -1,   7,  28,  60,  -7, -14,   0,  -4,  -1,  -2,  -1,  -1,  -1,  -6, -21,   5, -38,   7,  -2,   1,   3, -13,  -5, -39,  21,   9,  -2,   5,  -1,   2,  13,  17,  11,  -8,  -2,  -2,  -3, -19,  -1,  13, -12,  -1,   2,   0,   0,  15,   4, -11,  -5,   0,  -4,  -1,   5,  -6,   2,   3,  -1,   0,  -1,  -2,  -1 },
     {   -1,  -3, -14,   3,  -1,   1,   0,   2,  -1,   0,   1,   1,   1,  -7, -10,   1,   0,   1,  -1,   0,   1,   0,   0,   0,   2,   2,   3,  11,  -4,   1,  -1,  -1,   0,   0,  -1,   0,  11,  -3,  18,  -1,  -5,   1,   2,   0,  -1,   0,   0,   0,  15,  10,  -4, -12,   1,   3,  -2,   1,  51,  -6,  -9,   1,   7,   0,  -1,  -1,  30,  29,  -2,   0,  -2,  -4,   1,  -2,  79, -29, -14,  -2,  -6,   0,  -2,  -1, -37, -28,  10,   1, -17,  15,   3,  -1,  10,   0,  -3,  -2,   4,  -5,  -1,   0 },
     {   -6, -22,   6,  -3,   6,  -1,   1,  -2,   1,   0,   1,   0,  -7, -15,  -9,   5,  -2,   3,   1,   2,   1,   1,   0,   0,  -6, -50,  32,  10,   0,   3,   0,   2,  -1,   0,  -1,   0,   0,  -5,  -2,  -1,   0,   0,   1,  -1,   0,   0,   1,   0,   2, -43,  37,  -1,   2,   1,  -1,   0,  24,  55,  -5, -18,   2,  -3,   0,  -2, -29,  22, -30,  10,   2,  -2,   2,  -1, -10, -28,  12,  12,  -2,   2,  -1,   1,  22,   1,   9, -10,  -8,  12,  -4,  -3,   0,  -2,  -4,   4,   6,  -2,   1,  -1 },
     {    5,   5,  -4,  32,  64, -31,   5,  -8,  -2,  -3,  -1,  -3,  -2,  -6,  -2,  -9, -23,   6,   1,   2,   2,   1,   1,   1,  -3,  -5, -10, -27, -59,  35,  -4,   8,   2,   1,   0,   2,   6,  11,  17,  18,  34, -10,  -5,  -2,  -4,  -1,  -2,  -1,   0,  -2,  -5,   6,  21, -18,   4,  -2,  -1,   0,  -6,  -6, -18,   6,   3,   0,   0,  -4,   3,   0,  -5,   5,  -4,   1,   4,  -6,   3,   0,   7,  -2,   0,   0,   2,   5,   0,  -1,  -1,  -2,   0,   1,   0,  -3,   1,   0,   1,   0,   1,  -1 },
     {   -4,  -6,  -1,   8,  -2,   4,  -1,   3,  -1,  -1,  -1,  -1,  -6, -21,   8,   2,  -5,  -1,   1,   1,   0,  -1,   0,   0,  -7, -21,  -4,   2,   0,  -1,  -4,   0,   2,   2,   0,   2,  -2, -44,  18,  -3,  11,   2,   1,   2,   0,   1,   0,   0,  -2, -18,  -3,   1,   2,  -2,   3,   1,  -2, -35,  15,   7,  -1,   1,   0,   0, -23,  29,   4, -11,   2,  -2,   0,  -1, -36,  -4,  -5,   1,   2,   0,   1,   1, -70,   7,  17,   7,  40,  21,  -8,  -2,  17,  -5,  -3,  -4, -16,  -1,   5,   2 },
     {   -9, -18, -46,  -5,   8,  -1,   4,   5,   4,   0,   1,   2, -11, -24, -73,  32,   0,   7,   0,   5,   0,   1,   1,   1,  -1,  -1, -15,  12,   6,  13,   3,   1,   0,   0,  -1,  -1,  -1,  -5,  -7,  25,   7,  -1,   0,  -1,   0,  -1,   1,   0,   4,   1,  32,   5, -26,   5,  -7,   1,  -7, -23,  10, -21,   9,  10,  -4,   3,   8, -26,  -3,  10,  15,  -3,  -2,   1,   0,  25,   8,   4,  -5,  -6,  -2,   0,  -6,  -1,   2,  -8,  -4,  -6,   1,  -2,   6,   3,  -2,   6,  -1,   0,   1,  -1 },
     {    5,   9,  20,  22,   5,  -6,   3,  -5,  -3,  -3,   0,  -2,  -7, -17, -26, -55,   9,   7,   2,   3,   4,   3,   1,   3,   6,   7,  24,  41,  -7,  -6,  -8,  -1,  -1,  -1,  -1,   0,  -4, -19,  -9,  43, -10,  -3,   5,  -2,  -2,  -1,   0,  -1,   0,   0, -25, -50,   5,  10,   1,   3,   2,   1,  40,  -1,   3,  -1,  -4,   1,   9,   3,   0,  17,  -3,  -4,   0,  -2, -12,   5, -18,  -3,   3,  -1,   2,   1,  14,  -3,   1,  -4,   1,  -4,   9,   3,  -4,   1,   0,  -2,   0,   3,  -2,  -1 },
     {    0,   3,   1,  -1,  -2,  -2,  -5,   0,  -2,   0,   0,   1,   2,  14, -14,  -6,   7,  -2,   0,   0,   0,  -1,   1,   1,   8,  28,  19,   7,   1,   0,   6,  -3,   2,  -3,   0,  -2,   1,  38,  -9,   8,  -5,  -9,  -4,  -3,  -1,   0,   0,  -1, -13,  13, -14, -18,   1,  -1,  -3,   3, -12,   9, -15,  -7,   1,  -3,   5,   0, -47, -39,  -6,  25,   1,   7,   0,   1,  13, -30,  23,   4,  -4,   3,  -3,   1, -39,  32,  19, -13,   6,  30, -17,  -2,  19, -18,  -6,   6,  -3,  -8,   8,   0 },
     {    0,   1,  -9,  14,  -1,   2,  -5,   5,   5,  -1,   2,   0,   1,  -2, -14,  -1,  -3,   1,  -2,   1,  -1,   1,   2,   2,   4,  17,   0,  11,   0,   0,  -3,  -1,  -5,   0,  -6,  -1,   6,  14,  10, -22,   5,  -4,   3,   2,   4,   1,   0,   0,   4,  41,  14,   7,  -7,  -8,  -4,   2,   2,   1,  19,  -6,   0,  -1,  -5,  -2,  -3,  33, -32,  -9,   4,  -2,   2,  -2,  12, -19,  -7,  13,  -6,   0,  -2,   2,  17,  32,  11,  -4,  63, -27, -11,  -6, -42, -18,  10,   0,   4,  20,  -2,   2 },
     {   -2,  -8, -17,  14, -13,  20,  -3,   4,  -1,   0,  -1,   1,  -2,   1,  -4, -12,  12,  -8,  -3,   1,   0,   1,   0,   2,  -8, -25, -46,  25,   1, -12,   3,   5,   3,  -1,   1,   0,  -2,   6, -10,   0,   9,  11,   4,  -1,  -1,   1,   1,  -1,  -5, -29, -19,  26,  10,  -8,   2,   1,  13,  30,  37,   6, -29,   5,  -6,  -1, -11, -36,  20, -20,   0,  10,  -9,   5,  29, -14, -11,  -6,  16,  -9,   2,   3,  -7,  26,  -6,   6,   1, -11,   4,   5,  -6,  -7,   8,  -6, -14,  12,   4,   0 },
     {    5,   6,  10,  17,  22,  58, -19,  -1,  -8,  -5,  -3,  -2,  -3,  -5, -18,   4,  -4, -14,  -1,   3,   2,   3,   0,   1,  -2,  -1,   4,  -4, -41, -62,  16,   2,   7,   5,   0,   0,   0,   1,  -6,  -7,   5,  34,   7,   0,   1,  -2,  -2,   0,   4,   9,  16,   2,  29,  24, -16,   0,  -4, -13,  -9,  -4,  -4, -26,  -1,   6,   1,  -4, -13,   1,  -8,  -4,  11,   0,   0,  10,  12,   1,   4,   7,  -4,  -2,  -2,  -4,  -1,  -1,  -4,  -2,  -1,   2,   2,   5,   0,   1,  -1,  -3,   2,   1 },
     {    2,   1,  13, -17,  13,   9,   5,  -2,   3,  -2,   0,   0,   3,  11,   4,  18, -13,  -7,  -5,   2,  -1,  -1,   1,  -1,  -2, -13,  13, -30,   5,  -5,   3,   4,  -3,   3,  -4,  -2,   2,   6, -12,  41,  -6,   0,  -1,  -2,   2,   1,   0,   0,  -5, -24,  15, -30,  -5,  12,   0,   4,   9,  17, -18,   1,  18,  -3,   2,  -2,  -1,   0,  36,  16, -12,   1, -12,   0,  26,  19, -11, -13,  -7,  -3,  -4,  -1, -20,  13, -25,   9,  51, -27,   2,  -2, -15,  -3,  16,  -3, -10,  16,  -8,  -3 },
     {    2,   1,  13,  38, -12, -26,  -5,   6,   2,  -2,   0,  -2,  -3,  -4, -14, -16, -11,  14,   5,   4,  -1,   3,  -1,   2,   3,   1,  21,   9,  12,   7,  -9, -12,  -5,  -1,  -4,   0,  -7,  -7, -25, -44,  12, -15,   1,   9,   6,   0,   2,   2,   8,  -8,  43,  -5,  13, -12,  -1,   7,   1,  -2, -13,  42,  -8, -13,   6,  -2,   2, -24, -14,   0,  -5,   9,   7,  -4,  17,  26,  11, -29,   8,   5,  -2,  -2, -16,  -4,  -8,   5,   4, -14,   5,  12,   9,   2,   3, -12, -14,   8,   0,  -4 },
     {   -3,  -6,  -4,   5, -34, -17,   6,  14,   0,   2,  -3,   2,   3,   0,  -4,  26,  56,   3,   5,  -9,   0,  -4,  -3,   0,  -1,  -6,   3,  -5, -36,  15, -20,  -2,   0,  -1,   2,   2,   0,  -4,   2,  13, -46,  13, -13,  13,   0,   1,   3,   3,   6,   3,   4,   1,  55, -27,  11,   3,   0,  -6,   0, -19,   1, -15,  10,   0,   1,  -5,  -8,   3, -19,  10,   1,   2,   0,  12,  -5,  10,  -3,   5,   1,  -3,  -3,   1,   0,   3,   8,  -3,  -1,  -5,  -4,   5,   5,   1,  -3,   3,  -5,  -2 },
     {   -2,  -3,   1,  -5,   2, -15, -12,   4, -13,   9,  -8,   6,  -2,  -3,   3,   1, -12,  19,   1,   3,   1,  -8,   0,  -2,  -5,  -6,  -2,  -2,   6,  -6,  -1,   0,   4,   0,  10,   2,  -6, -14,  -7,  -5,   0,  -8,   0,   5,   1,   0,  -2,   1,  -9, -21, -12,  -8,   1,   5,  -9,   4,  -5, -37, -10, -17,  10, -19,   9,  10,  -2, -16,  -6,   8,   6,  10,  18,   3,  -3, -42,   4,   6,  -5,   4,   2,   0, -20,  16,  -1,   9, -33, -22,  21,   8, -55,  13,  31,  -5,  27,  24,  -9,  -7 },
     {   -4,  -5, -13, -44,  16,  -4,   4,   1,  -5,   3,   1,   4,  -4,  -6, -14, -65,  37,  16,  -3,   9,   1,  -3,   2,  -3,   1,  -2,   2, -11,  24,  -9,  11,   3,   5,   4,   2,  -4,   0,  -5,  -1,  -1,  22,  -1,   8, -10,  -5,  -5,  -1,   1,   1,  -8,  -5,  22,  18, -24, -24, -10,   1,  -5, -25,  -3,  -8, -10,  10,   6,   6,   8,   5,   1, -12,   9,  10,  -6,   7,  13,  11,  25, -21,   7,  -7, -10,   0,  -4,  -5,   1,  13,  -8,  -4, -12,   6,  -8,  -4,   7, -10,   8,  -1,   5 },
     {   -3,  -4, -12, -12,  -9,  -1,  -5,  -8,  24,   2,  11,   2,  -5,  -8, -29,  -7, -27,   0,   2,  10,  -3,   3,   6,  -1,  -2,   1, -18, -22,  14,   5,  14,  -2, -14,   0, -11,  -2,   0,   7, -27,   5, -10,  13,  -6,  -4,   9,   7,   4,   1,   3,  21,   2, -10,  37,  -7,   1,  -5,  -2,  14,   2,  38, -17,  -5, -12, -15, -11,  12,  12,   1, -17,  11, -19,   5, -22, -24,  -7,   6,   3, -11,  -1,  -1, -15, -21,   4,  -8,   1,   3,  15, -10, -21,  25, -10,  13,  27, -18, -14,  -2 },
     {   -1,  -2,  -5,  -3,  17, -13,   9,   4,  -6,  -2,   2,  -5,  -2,  -5, -10,   4, -16,  16, -11,   5,   0,   8,  -3,   3,  -3,  -6, -11, -14,  14,  -3, -12,   5,   6,   0,   3,   4,  -6, -12, -20,  -3, -38,   1,  12,   6,   5,   1,  -4,  -1,  -2,  -7,  -8,  -9,  20,  19,   1,  -7,  -3, -18, -14,  16,  11,   6, -23,  14,   5,  -7,   6, -24, -17,  -7,  11,  20,   9, -13,   8,  11,  17, -10,  11,   4,  29, -11,  12, -13,   7,  27, -29,  13, -12, -40,  12,   7, -33,  10,  19,   0 },
   }
 },
 {  //17
   {
     { -120,  42,   4,   6,   2,   3,   1,   1,   1,   1,   0,   0,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   7,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0, 124, -25,  -8,  -5,  -3,  -3,  -2,  -1,  -1,  -1,  -1,  -1,  -2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -11,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   0,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   -5,   8,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,-125,  19,   7,   4,   3,   2,   2,   1,   1,   1,   1,   0,   1,   5,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  10,   0,  -1,   0,   0,   0,   0,   0,   6,   0,  -1,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   -1,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -10,   5,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,-125,  17,   8,   4,   3,   2,   2,   1,   1,   1,   1,   0,  -1,   8,  -1,   1,   0,   0,   0,   0,  10,   0,  -1,   0,   0,   0,   0,   0,   7,   1,  -1,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   1,   0,   0,   1,   0,   0,   0 },
     {   -3, -13,   2,   0,   2,   0,   0,   0,   0,   0,   0,   0,   6,   4,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   9,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   9,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0, 124, -15,  -8,  -3,  -3,  -2,  -1,  -1,   3, -11,   1,  -1,   0,   0,   0,   0, -12,   0,   1,   0,   0,   0,   0,   0,  -5,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0 },
     {  -39,-114,  33,   3,  13,   3,   6,   3,   4,   1,   2,   1,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -6,   7,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -11,   4,   0,   0,   0,   0,   0,   0, -10,   4,   0,   0,   0,   0,   0,   0,   7,   2,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   3,   1,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
     {    2,  10,  -2,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -6,  -5,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -9,   3,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   6, -11,   1,  -1,   0,   0,   0,   0,-124,  16,   8,   3,   3,   1,   2,   1,  12,  11,  -2,   0,  -1,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0,   0,   6,   2,  -1,   0,   1,  -1,   0,   0,   1,   1,   0,   0,  -1,   0,   0,   0 },
     {   -3,  -3,   1,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -4,  -4,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -5,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -8,  -5,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0, -12,   2,   2,   0,   0,   0,   0,   0, -13, -11,   2,   0,   1,   0,   1,   0,-122,  12,   9,   2,   3,   1,   1,   0, -15,  15,   0,   1,   0,   0,   0,   0,   3,   4,   0,   0,   6,   2,  -1,   0,   0,   0,   0,   0,   1,   2,   0,   0 },
     {    1,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   5,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,  -7,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   5,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,  -6,   1,   0,   1,   0,   0,   0,  -2,   4,  -1,   0,   0,   0,   0,   0,  15, -17,   0,  -1,   0,   0,   0,   0,-123,  12,  10,   2,   3,   1,   1,   0,   8,  15,  -2,   1,   4,   0,   0,   0,   6,   2,   0,   0,  -3,  -1,   0,   0 },
     {   -1,   0,  -2,   1,   0,   1,  -1,   0,   0,   0,   0,   0, -22,-117,  26,   4,   8,   4,   5,   3,   3,   2,   2,   1,   1, -11,   6,   0,   1,   0,   1,   0,   0,   0,   0,   0,  -1,   6,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   5,   3,  -1,  -1,  -1,   0,   0,   0,  -2,   3,   0,   0,   0,   0,   0,   0,   4,   0,   1,   0,   0,   0,   0,   0,  -5,  -2,   1,   1,   0,   0,   0,   0, -31,   5,   3,   0,   2,   5,  -1,   0,  -2,   2,   0,   0,   4,   1,   0,   0 },
     {   -1,  -2,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   5,  30,  -7,  -1,  -2,  -1,  -1,  -1,   0,   0,   0,   0,  -3,   5,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,  -2, -10,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,  -3,   0,   1,   0,   0,   0,   1,  -6,  -6,   2,   0,   1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -7, -16,   2,  -1,   1,   0,   1,   0,-119,  12,   9,   2,   4,  18,  -1,   1,  13,   1,  -1,   0,   4,   1,   0,   0 },
     {   -1,  -4,   7,   3,   1,   1,   0,  -1,   0,  -1,   0,   0,  -1,  10,  -6,  -1,  -2,   0,  -1,   1,   0,   0,   0,   0, -18,-121,  18,   6,   8,   3,   4,   2,   3,   1,   1,   1,   5, -19,   7,   0,   2,   0,   1,   0,   1,  -1,   0,   0,   1,  11,   1,   1,   0,   0,   1,   0,   5,   1,   2,   0,  -1,   0,   0,  -1,   1,   7,   0,  -1,  -1,   0,   0,  -1,   6,   1,   0,   0,  -1,   0,   0,   0,  -1,   2,   0,  -1,   0,   1,   0,   0,  -1,   0,   0,   0,  -2,   1,   1,   0 },
     {    0,   1,   3,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -2,  -2,   3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,  -2,   1,   0,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   4,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -4,   1,  -1,   1,   0,   0,   0,  -1,   3,  -1,   1,   0,   0,   0,   0,  -6,  -6,   1,   1,   1,   0,   0,   0,  -3,   6,  -1,   1,   0,   0,   0,   0,  -1, -22,   2,   0,-118,  18,  10,   2,  34,  14,  -4,   0,  -1,  -4,   0,   0 },
     {   13,  29, 119, -25,   0, -13,  -4,  -6,  -3,  -4,  -2,  -3,   0,  -4,  -6,  -3,  -1,   2,   0,   2,   0,  -1,   0,   0,   2,   4,  -7,   1,  -1,  -1,   1,   0,   0,   0,  -1,   0,   0,   2,   2,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   1,   1,  -8,   2,  -1,   0,  -1,   1,   1,   5,  -1,  -1,   0,   0,   0,   0,  -2,  -3,  -4,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0,  -1,   0,  -1,   0,   4,   1,  -1,   0,   5,  -2,  -1,   0,   0,   0,   0,   0 },
     {    0,   0,   7,  -2,   1,   0,   0,   0,   0,  -1,   0,   0,  -1,   2,   1,  -1,  -1,   1,   1,   1,   1,  -1,   0,  -1,  -2,   5,  -2,   2,   0,  -1,   0,   0,   0,   0,   0,   0,  -6, -29,   5,   2,   1,   1,   0,   1,   0,   0,   1,   0,  -1,  -9,   1,   0,   1,   0,   1,   0,  -3, -11,   0,   0,   0,   1,   0,   0,   1,  -4,   2,  -1,   0,   0,   0,   0,  -9,  -8,   2,  -1,   0,   0,   0,   0, -11,  -4,   1,  -1, -34, -17,   5,  -1,-112,  13,  11,   1, -13,  21,  -1,   1 },
     {   -1,  -3,   5,   2,  -3,   2,   0,  -1,   0,   0,   0,  -1,  -2,  -5,   1,   1,  -3,   1,   1,   2,   1,   0,   1,   0,  -2, -10,   5,   0,   3,   1,   1,   0,   0,   0,   0,   0,  -8,  -4, -18,   5,  -3,   1,   1,   1,   0,   1,   0,   0, -15,-117,  14,   7,   6,   5,   1,   3,   8, -24,  12,   2,   3,  -1,   0,   0,   1, -10,   5,   2,   0,   0,   1,   1,  10,   1,   4,  -1,   1,  -1,   0,   0,   5,   4,   1,  -2,   8,   6,   0,   0,   9,   0,   0,  -1, -10,   0,   1,  -1 },
     {    0,   0,  -1,   0,   0,  -2,  -1,  -1,   0,   0,   0,  -1,  -1,  -9,   2,   0,  -1,   1,   0,   1,   1,   0,   0,   0,  -1,  17,  -9,  -2,  -1,   0,   0,   1,   0,  -1,   0,   0, -14,-117,  17,   5,   6,   5,   2,   3,   2,   1,   1,   1,  11,   1,   6,  -1,  -1,  -1,   0,   0,   1,   4,  -1,   2,   0,  -1,   1,  -1,   7,   6,   1,   0,   0,  -1,   0,   0,  -4,   7,  -2,   0,   1,   0,   0,   0,  11,   0,   0,   0,   4,   4,  -1,   0,  27,   2,  -2,   1,  17,  -4,  -2,  -1 },
     {    1,   0,   3,   5,  -2,   0,  -1,   0,   0,  -1,   0,   0,   2,   4,  -1,  -2,  -2,   1,   0,  -1,   1,   1,   0,   0,   0,  -4,   2,  -2,   2,   1,   1,   0,   0,   0,  -1,   0,   2,  14,  -1,  -2,  -1,   1,   0,   0,  -1,  -1,   0,   0,  -2, -12,  -1,   2,   0,   1,  -1,   1,   1,   1,   3,   1,   1,   0,   0,   0,   0,   8,  -1,   0,  -2,  -1,  -1,   0,  -2,  -4,   1,  -1,   0,   0,   0,   1,   4,  10,  -2,   1,  -3,  -5,  -1,   0, -14,  28,  -2,   2, 120,  -9, -13,  -1 },
     {    1,   1,   5,  -9,  -1,   0,   0,   1,   1,   0,  -1,   0,  11,  24, 120, -22,  -7,  -8,  -6,  -4,  -4,  -3,  -2,  -2,   1,  -2,  16,  -5,   1,   0,  -3,  -1,  -1,  -1,  -1,  -1,   1,  -1,  -2,  -1,   2,  -1,  -1,   0,   1,   0,   1,   0,   0,  -1,  -2,  -1,   1,   0,   1,   0,   0,  -1, -10,   2,  -1,   1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,  -1,  -6,   1,   1,   1,   1,   1,   0,   0,  -1,   1,   3,  -2,  -2,   0,   1,   0,   0,   0,   0,   0,  -2,   0 },
     {    0,   0,   2,  -8,   1,  -3,  -3,   0,   0,  -1,   0,  -1,   0,  -3,   0,   0,  -3,  -1,  -1,  -1,  -1,   0,   0,   0,  -1,   0,   1,   1,  -3,   2,  -1,   0,   0,  -1,   0,   1,   0,  -2,  -5,   2,   0,   0,  -1,   0,  -1,   2,   0,   0,  -6,  28, -16,   3,  -2,  -1,  -2,   1, -17,-116,  16,   7,   3,   6,   2,   2,  13, -17,  17,   0,   2,   0,  -1,   1,   0,   1,   1,   0,   2,  -1,   1,   0,   8,   1,   6,   1,   0,  11,  -2,  -1,   7,   3,   0,  -1,   5,  -4,   0,   1 },
     {   -4,  -2, -10, -68,   5,  -7,   7,   0,   3,   1,   1,   0,  -1,   0, -17,  -1,   9,   0,   0,   2,  -1,   1,   2,   0,   7,  13, 100,   5,  -3,   0,  -8,  -4,  -4,  -1,  -2,  -1,   0,  -4,  16, -12,  -1,   1,   1,  -1,  -1,   0,  -1,  -1,   0,  -3,  -8,   1,   1,  -3,  -2,  -1,   1,   5,   5,   2,  -1,  -1,  -2,   0,  -1,  -4,  -4,   2,   0,   0,   1,   1,   2,   5,  -2,  -1,   0,   0,  -1,   0,  -1,   2,  -2,   3,  -2,  -3,  -2,   2,   0,   1,  -3,   1,   2,  -2,  -1,   1 },
     {    6,   4,  21,  96, -27,   3, -12,  -1,  -4,  -1,  -4,  -1,   1,   2,   0,  19,  -8,   0,  -4,   0,   0,   0,  -1,   0,   5,  12,  64,  -1,  -1, -11,  -3,  -5,  -2,  -1,  -1,   0,   1,  -1,  22,  -8,   0,  -4,   0,  -1,  -1,   0,   0,   0,   0,   1,  -2,  -5,  -5,  -1,  -1,  -1,  -1,  -6,   4,  -1,  -1,  -1,  -2,  -1,   1,  -2,  -3,  -6,   1,   0,   0,   2,  -2,  -2,  -5,   0,   0,  -1,   0,   0,   1,  -5,  -2,  -2,   1,   0,   0,  -1,   1,   1,  -2,  -2,  -5,  -1,   0,   0 },
     {   -1,  -2,  -2,   6,  -4,  -2,   0,  -2,   1,  -1,   0,   0,  -1,   1,  -1,  -3,  -3,  -2,   0,   0,  -1,   0,  -2,   0,  -1,  -4,  -2,   4,  -4,   4,  -1,   0,   0,  -1,   2,  -1,  -2,  -5,   2,  -1,   0,   3,  -1,   0,  -1,   1,  -1,   1,   2,   6,  -4,  -1,   4,  -1,  -1,   1, -10,  21, -16,   2,  -2,  -3,   1,   0, -11,-114,  27,   7,   1,   4,   0,   2,  18,   5,  22,  -2,   3,   0,   0,  -1,  -2,  17,  -2,   2,   1,  -6,  -2,  -2,   0,  -4,   1,  -3,   9,  -2,   5,   0 },
     {    1,   2,   4,   7,   1,   3,  -1,  -1,  -1,  -2,   2,   0,   1,   1,  -1,  25,   3,   4,  -2,  -1,   1,  -1,   1,   1,   1,   0,  25,  -4,  -8,  -3,   1,   1,  -1,  -2,   0,  -1,  -7, -16,-115,  15,  12,   6,   4,   5,   5,  -1,   1,   1,   3,  17,   0,   3,   1,  -2,   2,   2,  -2,   6, -10,  -5,  -1,   0,   2,  -1,  -1,   4,   1,   1,   1,   1,   0,  -3,  -1,   6,  11,  -4,  -1,  -2,   1,   1,  -2,  -5,  -2,   1,   0,  -4,   2,  -3,  -3,  -5,  -2,   1,   4,   3,   2,  -1 },
     {    2,   1,   1,  23,  10,   0,  -1,  -1,  -1,  -2,  -1,  -2,  -6,  -5, -19,-117,  11,   9,   9,   4,   3,   3,   0,   0,   1,   3,  12,  -9,   2,   0,  -4,   1,  -1,   2,   0,   1,  -1,  -2, -20,  -2,   4,   5,   1,   0,  -1,   0,   0,   0,   0,   2,  -1,  -4,  -3,   1,  -3,  -2,   0,   0,  -2,  15,   0,   0,  -3,   1,   1,   2,  -1,  -1,   0,  -1,   1,   0,  -1,   0,   2,   8,   0,  -1,   0,   0,   1,   0,   1,   1,   0,   2,  -2,   5,   0,  -1,  -2,   1,  -2,   2,  -1,   2 },
     {   -1,  -2,  -3,  -4,   4,  -1,  -4,   0,   0,   1,  -1,   1,  -1,  -3,  -1,   1,  -2,   2,  -2,   2,  -2,   2,   0,   1,  -1,  -3,   2,   6,  -3,  -1,  -1,   1,   1,   0,   1,   1,  -1,  -5,  -3,  -5,   3,   3,   4,  -2,  -1,   1,   1,  -1,  -3,  -3, -13,   2,  -3,   0,   0,  -2,  -3,   1,  -8,   2,  -1,   0,   1,   1, -14, -13, -13,   3,   2,   3,   2,   0, -14,-108,   4,  12,   4,   0,   0,   0,  13, -46,  21,   3,   9,  -4,   9,   1,  11,   6,   6,   1,   2,   7,   2,   1 },
     {   -6, -10,   1, -23,-109,  34,  -2,   3,  -4,   7,   3,   3,  -1,  -1,  -5, -12,  -5,  14,   4,   4,  -4,  -3,  -2,  -2,  -1,  -2,  -7,   3, -11, -13,   2,   6,   1,   4,   0,   3,   0,   0,  -3,  -5,  12, -13,  -2,  -7,   1,   4,  -1,  -1,   1,   2,  -3,  15,  14,  -1,  -7,  -7,  -1,   2, -14,   4,   6,   5,  -2,   0,   1,   7,  14,   0,  -1,  -3,   0,  -1,  -2,   0,   1,   0,   4,  -2,   0,   0,   0,  -7,   3,  -2,   1,   6,  -5,   3,  -2,  -3,   4,   3,   1,   0,   2,   1 },
     {    0,   0,   0,   0,   7,  18,  -1, -12,  -3,   3,   1,   2,   0,   0,  -6,  13,  10,  -7,   3,  -6,  -1,  -2,   3,   1,  -3,  -4,  -4, -86,  23,   6,  -2,  -4,  -3,   6,   2,   5,  -1,  -3,   1, -36,  27, -13,   7,  -1,   5,   0,   1,   0,  -4, -10, -47, -18,  11,   2,   6,  -3,  -1,  -3, -32,  15,   3,   4,   1,   8,   0,   0,  -7,  12,   9,   4,   0,   2,   2,   6,  -7,  14,  -2,  -6,   2,  -1,  -2,   3,  -5,  10,   0,  -5,   0,   4,  -1,   5,  -2,   4,  -3,  -5,   3,   3 },
     {   -1,  -1,  -7,   9,   3, -17,  -2,   3,  -1,  -1,  -2,   2,  -1,  -1,   2,  -2, -19,   2,   1,   1,   6,  -3,  -1,   0,   1,   1,  -8,  55,   1,   4,   0,   4,  -1,  -3,  -1,  -1,  -2,  -5,  -6, -23,  -6,   3,  -2,   5,  -1,  -2,   1,  -1,  -5, -12, -99,  18,   9,  11,   9,   8,   2,  11,   1,   8,   2,   0,   0,  -1,   1,  10,   4,   5,  -1,  -5,   1,  -3,  -1,  19,   1,   0,   1,  -1,   0,   1,  -3,  -8,   3,   2,  -1,  -5,  -3,   0,   0, -13,   2,  -1,   2,   2,   2,  -1 },
     {    0,   0,   1,   2,   2,  -4,  10,   0,  -1,  -2,  -4,  -3,   0,   1,  -2,   1,  19, -15,   5,  -2,   2,  -5,   2,   0,   0,  -4,   0,  -3,  -2,   9,   3,   2,  -1,   0,   3,   3,  -1,   2,   2,   5,  -2,   3,  -3,  -2,  -2,   1,   0,  -1,   1,  -4,  11,  -6,   7,  -3,  -6,  -1,  -3,   1,   1,   3,   6,  -1,  -2,   2,   3, -19,   4,   6,   2,   1,   4,  -1,  -9,  39, -48,  11,  -8,   6,  -3,   0, -10, -87,  21,   5,  17,  15,   1,  -7,  -8, -30,   8,   4,  19,  18,   0,  -2 },
     {   -1,  -1,   2,   3, -10,  -2,   3,  -4, -10,  -5,  -8,  -4,   3,   5,   6,   4,  29,  10,  31,  11,  11,  -7, -17,   3,   1,   1,   1,  -1,  -1,  32,  19,  15,  10, -13,  -7,  -7,  -1,  -1,   1, -31, -15,  25,   3,  12, -14,  -3, -13, -11,   0,   3,   9, -36,  42,   0,  12,  10,  -1,  -7,   6,  -1,  -1, -20,   9, -15,   0,   7,   1,  11, -26,   1,   0, -22,  -1,  -3,  34,  -8, -10,   1, -12, -13,  -2,  -9,   4, -30,  -1, -15,   4,   7,   1, -10,  -3,   1,  -1, -18, -10,   2 },
     {   -2,  -3,  -2,  -5,  -9,  -1, -29,  -7,  -9,  12,  17,  -1,  -1,  -1,  -1,  -4,  -2,   0,   0,  -9,  12,  -3,   8,   8,  -1,  -2,   0,   5, -19,   7,  -2, -12,  12,   6,  11,   5,   1,   0,   6,  41,  -5, -11,   5,   7,   1,  -5,  -2,  -2,  -3,  -5,  -3, -14, -24,  24,  11,   5,  -1,  -9, -11,   1,   0, -11,  15,  16,  -1,  -1, -16,   5,  10,   7,   3,   3,  -2,   8,  -2,  19, -19,  -5,  -4,   2, -14, -18, -18,  -1,  -4, -62,   9,  10,   2, -11,   1,   0,   1, -52,  18,   6 },
   },
   {
     {   96, -67,  23, -18,   9,  -7,   4,  -3,   2,  -2,   1,  -1, -31,  22, -10,   7,  -3,   2,  -1,   1,   0,   0,   0,   0,   8,  -4,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -61,  -4,  32, -15,  14,  -7,   6,  -3,   3,  -1,   1,  -1, -73,  53, -24,  18,  -9,   6,  -3,   2,  -1,   1,   0,   0,  35, -20,   7,  -5,   1,  -1,   0,   0,   0,   0,   0,   0, -11,   7,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   7,  -4,   1,   0,   0,   0,   0,   0,  -3,   2,  -1,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   37,  46, -64,  37, -31,  18, -14,   8,  -7,   4,  -3,   2, -69,   7,  19,  -7,   9,  -3,   3,  -1,   1,   0,   1,   0,  20,  -7,   1,  -2,   2,  -2,   1,  -1,   0,  -1,   0,   0,  -5,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -12,  -7,  10,  -5,   5,  -2,   2,  -1,   1,   0,   0,   0, -54,   2,  14,  -4,   6,  -2,   3,  -1,   2,  -1,   1,   0, -74,  48, -19,  15,  -7,   6,  -3,   2,  -1,   1,   0,   0,  50, -27,   9,  -7,   3,  -2,   0,   0,  -1,   0,   0,   0, -23,  12,  -4,   2,   0,   0,   1,   0,  13,  -6,   2,  -1,   0,   0,  -1,   0,  -6,   3,  -1,   0,   0,   0,   0,   0,   6,  -2,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   2,  -1,   0,   0,  -1,   1,   0,   0,   2,   0,   0,   0 },
     {  -15, -23,  20, -10,  11,  -5,   5,  -3,   2,  -1,   1,  -1, -22, -49,  57, -34,  27, -15,  11,  -6,   5,  -2,   2,  -1,  56,  21, -37,  18, -16,   7,  -6,   2,  -2,   0,  -1,   0, -25,  -7,  12,  -4,   4,   0,   0,   1,   0,   1,   0,   0,   7,   6,  -6,   2,  -1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   2,   3,  -2,   1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   2,  -1,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {    7,   5,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  15,   3,  -8,   4,  -4,   2,  -1,   0,   0,   0,   0,   0,  69, -11,  -3,  -4,  -1,  -1,  -1,   0,  -1,   0,   0,   0,  85, -26,  -2,  -3,  -1,  -2,   0,  -1,   0,  -1,   0,   0, -47,   8,   1,   1,   1,   1,   0,   0,  23,  -5,  -1,   0,  -1,   0,   0,   0, -13,   1,   2,   0,   1,   0,   0,   0,   7,  -1,   0,   0,   0,   0,   0,   0,  -5,  -1,   1,   0,   2,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0,   0 },
     {   28,  66,  -9, -44,  39, -42,  28, -23,  14, -11,   6,  -6,  -8, -20,  -7,  27, -26,  23, -15,  12,  -7,   5,  -2,   2,   9,  23, -15,   2,  -2,   0,  -2,   1,  -1,   1,  -1,   1, -14, -12,   8,  -2,   3,  -2,   2,  -1,   2,  -1,   1,   0,   6,   0,   0,  -1,   0,   1,  -1,   1,  -2,   0,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,  -1,   1,  -1,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   1,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
     {    1,  -3, -11,  17, -16,  14, -11,   8,  -5,   4,  -2,   2,  25,  47, -37,  11,  -9,  -1,   0,  -2,   0,  -1,   0,   0,  17,  36, -41,  27, -23,  14, -10,   5,  -4,   2,  -2,   1, -39, -32,  39, -22,  19,  -9,   6,  -2,   2,   0,   1,   0,  -8,  17, -12,   7,  -4,   1,   0,  -1,   9, -11,   7,  -3,   2,   0,   0,   1,  -4,   6,  -3,   1,   0,  -1,   1,  -1,   1,  -4,   3,  -1,   0,   0,   0,   0,  -3,   2,  -1,   0,   0,  -2,   1,   0,   0,   1,   0,   0,  -1,  -1,   1,   0 },
     {   -1,   2,   3,  -3,   3,  -3,   2,  -2,   1,  -1,   0,   0,  -7,  -8,   6,  -1,   1,   0,   0,   0,   0,   1,   0,   0, -10, -14,  13,  -8,   6,  -2,   2,  -1,   1,   0,   0,   0, -55,  17,  -6,   8,  -3,   2,  -1,   1,   0,   0,   0,   0, -75,  17,   7,   0,   2,   1,   0,   0,  67,  -8,  -7,   1,  -2,   0,  -1,  -1, -34,   2,   4,   0,   1,   0,   0,   0,  15,   1,  -2,   0,  -1,   0,   0,   0, -10,   0,   1,   0,   7,   1,  -1,   1,  -3,  -1,   1,   0,   2,   1,   0,   0 },
     {    0,   1,  -2,  -2,   3,  -2,   1,   0,   1,   0,   0,   0,   0,   1,   1,  -4,   2,  -2,   2,  -1,   1,  -1,   0,   0,  -2,  -1,   0,   1,  -1,   1,  -1,   0,   0,   0,   0,   0, -14,   0,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0, -85,   3,   8,   1,   3,   1,   1,   0, -79,  19,   7,   0,   3,   0,   2,   0,  37,   0,  -4,   1,  -1,   0,  -1,   0, -20,   4,   2,   0,   0,   0,   0,   0,  11,   1,   0,   0,  -7,   0,   1,   0,   3,   2,  -1,   0,  -1,   0,   0,   0 },
     {    6,  14,   2,  -5,   0,  -1,  -2,   1,  -2,   0,   0,   0,   9,  30,  -6, -13,  11, -14,   9,  -8,   5,  -4,   1,  -1,   4,  31, -30,  22, -18,  11,  -8,   4,  -3,   1,  -1,   0,   4,  27, -30,  16, -11,   7,  -6,   4,  -3,   2,  -1,   1,  -6, -55,  45, -25,  18, -10,   6,  -3,  18,  34, -27,  13,  -9,   4,  -3,   0,   2, -19,  10,  -4,   2,   0,   0,   1,  -7,   7,  -3,   1,   1,   0,   0,   0,   0,  -6,   2,  -1,  -2,   3,  -1,   0,  -1,  -3,   1,   0,  -1,   1,   0,   0 },
     {  -19, -47, -25,  32,  -7,   9,   1,   2,   3,   0,   2,   0,  -7, -29,   0,  35, -34,  34, -24,  20, -12,   9,  -6,   4,  16,  51,  -6, -24,  20, -20,  12, -10,   5,  -4,   2,  -2,  -8, -11,  -7,  14, -10,   8,  -3,   2,   0,   0,   1,  -1,  -3,  -8,  16, -13,   9,  -5,   2,  -1,   2,  11, -13,   6,  -4,   0,   1,  -1,   6,  -5,   5,  -5,   2,   0,   0,   1,  -1,   3,  -3,   2,   0,  -1,   1,  -1,   2,  -2,   2,  -1,  -1,   2,  -1,   0,  -1,   0,   1,  -1,   1,   1,  -1,   0 },
     {    2,   4,   4,  -2,  -2,   1,  -1,   0,   0,  -1,   0,  -1,   1,  -1,   2,   1,   2,  -2,   0,  -1,   0,   0,  -1,   0,  -1,  -4,   5,  -5,   3,  -2,   2,  -1,   1,  -1,   1,   0,   2,  -4,   6,  -4,   2,  -1,   1,  -1,   0,   0,   0,   0,  -1,  10,  -9,   4,  -2,   1,   0,   0,  61,  -2,  -2,  -2,  -1,   0,  -1,   0,  88, -18,  -9,   2,  -3,   0,  -1,   0, -58,  -2,   8,  -1,   2,   0,   1,   0,  24,   1,  -4,   1,  -5,  -3,   1,  -1,   7,   0,  -1,   0,  -4,  -2,   1,  -1 },
     {   12,  29,  57, -14, -31,  29, -36,  24, -21,  13, -11,   6, -17, -55, -45,  33,   8,  -4,  15,  -8,  10,  -4,   4,  -1,   7,  19,   9,  -5,  -1,  -1,  -1,  -1,   0,  -1,   1,  -2,  -5,  -1,  -3,   0,   1,   0,   0,   1,   0,   2,  -1,   1,   1,   2,   3,  -4,   3,  -1,   2,  -1,  -5,   6,  -2,   2,  -1,   0,  -1,   0,  -1,  -2,  -1,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,  -1,   0,   1,  -1,   1,   0,   0,   2,   0,   0,  -1,  -3,  -1,   0,   0,   2,   1,   0,   0 },
     {   -1,   0,  -1,  -2,  -1,   0,   0,  -1,   0,   1,  -1,   1,  -1,   4,   0,  -2,   2,   0,   0,  -1,   0,  -1,   0,   0,   1,   7,  -2,  -5,   2,  -3,   3,  -3,   2,  -2,   2,  -1,   1,  -1,  -5,   4,  -5,   4,  -1,   0,   0,   1,   0,   0,   2,   4,  -3,   0,  -2,   1,  -1,   2,  -8, -11,   6,  -4,   2,  -2,   2,   0, -64,  -1,   8,   1,   2,   0,   0,   0,-103,  16,   9,   0,   5,  -1,   2,   0,   8,  11,  -2,   1,  21,   1,  -2,   0,   8,  -1,  -1,   0,  -5,  -1,   1,   0 },
     {    1,   2,  -6,   3,  -1,   2,   0,  -1,   2,  -2,   1,  -1,   3,   5, -11,   9,  -2,   3,  -3,   1,  -1,   1,   0,   1,   9,  30, -23,  17, -13,   7,  -2,   2,  -1,   1,  -1,   0,  11,  68, -49,  28, -25,  13, -11,   5,  -4,   2,  -1,   0,  -1,  46, -32,  16, -12,   6,  -5,   2,  -5, -27,  21, -10,   9,  -4,   2,  -1,  10,   2,  -2,  -2,   1,  -1,   1,  -1,   6,  -3,   0,   0,  -1,  -1,   0,   0,   3,   1,  -2,   0,  -6,  -3,   0,   1,  -1,   0,   1,  -1,  -2,  -1,  -1,   0 },
     {   -8, -22, -20,  17,   1,  -2,   7,  -6,   5,  -3,   3,  -1, -14, -45, -21,  26, -14,  15,  -8,   8,  -3,   3,   0,   1,  -9, -41, -14,  43, -29,  27, -15,  13,  -7,   6,  -3,   3,  15,  43,  15, -30,  19, -17,   9,  -7,   2,  -3,   1,  -1,  -7, -26,   1,   6,  -1,   2,   1,   0,   8,  11,   1,  -5,   3,  -1,   0,   0,  -6,   0,  -2,   3,  -2,   1,  -1,   0, -11,   3,   4,  -3,   1,   0,   0,   0,   3,   1,   0,  -1,   0,   2,   0,   0,   6,  -1,  -1,   0,  -3,   1,   0,  -1 },
     {    1,   2,   2,   1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,  -1,   0,   1,   2,   0,   2,  -1,   2,   0,   0,   0,  -1,   0,   2,  -1,   2,  -2,   2,   0,   1,   0,   0,   0,  -1,  -6,   2,   0,   0,  -1,  -1,  -1,   0,   0,   0,   0,  -1,   2,  -3,   1,  -1,   0,   0,  -1,   5,   2,  -2,   2,  -1,   0,   0,   0, -29,  -4,   4,  -1,   3,  -1,   1,   0,  10,  10,  -2,   1,   0,   0,   0,   0,  84, -16,  -8,   1, -83,   0,   8,  -1,  26,   5,  -3,   1, -10,  -3,   1,  -1 },
     {    8,  16,  37,  13, -25,   6,  -1, -11,   9,  -9,   4,  -4,  -1,   1,  -2, -21,   1,  18, -26,  28, -21,  17, -11,   8,  -8, -34, -45,  27,  13, -18,  30, -21,  18, -11,   8,  -4,   4,   8,  31,   4, -19,  20, -22,  13,  -9,   4,  -2,   0,   0,  11,   0, -16,  19, -19,  12,  -7,  -3,  -4,   1,   7,  -9,   6,  -1,   0,   0,   1,  -2,   1,   2,  -1,   1,   1,   0,  -3,   1,   1,   0,   1,   1,  -1,   2,  -2,   3,  -2,   2,   0,  -1,   1,   1,  -2,   2,  -1,   0,  -1,   0,   0 },
     {    1,   3,   4,  -1,   1,   1,  -4,   4,  -3,   3,  -3,   1,   1,   2,   3,  -2,  -4,   2,  -2,   1,  -1,   0,  -1,   1,   2,  10,  -6,  -7,   7,  -6,   5,  -5,   3,  -2,   2,  -1,  -1,   6,  -4,   4,  -6,   2,  -1,  -1,   1,  -1,   1,  -1,  -6, -52,  18,   0,   3,   2,  -1,   5,   9, -26,  32, -22,  14, -10,   6,  -3,   0,  63, -50,  28, -17,   8,  -3,   1,  -2, -37,  23, -10,   5,  -1,   1,   1,  12,  12,  -7,   3,  -4,  -6,   3,   0,   1,   5,  -1,   0,   3,  -4,   0,   0 },
     {    5,  12,  20, -13,  -2,  12, -21,  17, -15,  10,  -8,   3,   7,  19,  37,  -5, -20,  14, -17,   9, -10,   5,  -5,   4,   2,  14, -14,  -9,  14, -17,  19, -18,  11,  -9,   6,  -4,   6,  32,   4, -26,  18, -18,  12, -11,   6,  -4,   3,  -3,  -7, -29, -27,  42, -26,  23, -11,   8,   5,  34,  -3, -10,   7,  -7,   3,  -2,  -7, -16,   6,   0,   1,   1,   1,  -1,   4,   8,  -7,   2,  -2,   0,   0,   1,  -5,  -3,   2,   0,   1,   3,  -1,  -1,   3,  -2,  -1,   0,  -4,  -1,   2,  -1 },
     {   -4,  -8, -26, -40,  29,  13, -29,  36, -30,  24, -16,  13,   4,   4,  31,  34, -33,  -1,   9, -19,  15, -12,   7,  -5,  -9, -34, -37,  10,  11,   0,   3,  -2,   4,  -1,   2,  -1,   1,  -2,   5,  10, -10,   7,  -8,   8,  -4,   2,  -2,   1,   2,  13,   4, -15,  11, -11,   6,  -5,  -3, -14, -12,  13,  -6,   4,   0,   2,   4,   7,   4,  -2,  -1,   1,  -3,   1,  -3,  -4,   2,   0,   0,   0,   1,   0,   0,   4,  -1,   0,  -4,   2,  -3,   1,   3,   6,   1,   0,   3,   0,   0,  -1 },
     {    1,   0,  -2,  -2,   1,   0,  -1,   1,  -1,   0,  -1,   0,   2,   2,   0,   0,  -2,  -2,  -1,  -1,   2,  -1,   1,  -1,   2,   2,   0,   1,  -1,   0,  -1,   0,   1,  -1,   1,   0,   1,   2,   0,   1,   2,   0,   2,  -2,   1,   0,   1,  -2,  -1,   1,   2,   0,   0,  -1,   2,  -1,  -1,  -1,   0,   3,  -1,   0,  -1,   2,   3,   4,   0,  -2,   1,  -1,   0,   0,  27,  14,  -6,   2,  -1,   1,   0,  -1,  76,   3,  -7,   1,  90,  -3, -10,   0,  32, -11,  -5,   0,  -1,  -7,   2,  -1 },
     {   -7, -13, -39, -13,  33, -13,   6,   4,  -8,   5,  -1,   5,  -7, -16, -56, -19,  30, -21,  20, -11,  13,  -6,   6,  -2,   4,  11,  -5,   3,  21, -28,  30, -26,  19, -12,   8,  -8,   3,  12,  29,  -8, -10,   9, -19,  17, -13,   8,  -7,   2,  -2, -16, -17,  13,  -5,   5,  -2,   3,   5,  12,   8,  -3,   2,  -1,   2,  -5,   0,  -7,   4,   3,  -3,   1,  -2,   2,   4,   5,  -3,   3,  -3,   3,  -2,   0,  -4,   0,   5,   1,   1,   2,   1,   1,   5,   2,   0,   0,   4,   0,   1,   0 },
     {    0,   0,  -1,   0,   0,   0,   4,  -1,  -2,   1,   1,   2,   0,   1,   1,  -2,  -6,  -2,  -1,   1,   3,  -1,   0,  -1,   0,  -1,   0,   1,   2,   2,  -1,  -2,   0,   0,   0,   0,   0,   4,   2, -10,  -3,   1,  -2,   4,  -1,  -1,  -1,  -1,   0,  -2,  -4,   5,   2,  -2,   4,  -2,   0,   6,   1,  -1,   1,   3,  -1,  -1,  -6, -12,   0,  -5,   1,  -1,   0,   1,  -3,  -3,   2,   1,   0,  -2,   2,   0,  42,  11,  -8,   4,   5, -23,   4,  -1, -98,  16,  14,  -3,  55,  10,  -8,   3 },
     {    0,   2,  -3, -14,   0,  -1,  -2,   3,  -3,   2,  -3,   2,   3,  10,  10,  -8,   0,  -3,  -1,   1,   0,  -1,  -2,  -1,  10,  35,  35, -14,   6,  -5,  -2,   0,  -5,   0,  -2,   1,  13,  62,  46, -41,  13, -16,   2,  -7,   2,  -3,  -1,  -1,   0,  23,  17, -38,  23, -20,  13,  -9,  -6, -25,  -4,  12, -10,   5,  -2,   5,  -1,   3,  -8,  -5,  -1,  -1,   2,   1,  -3,  -6,   0,   3,   0,   3,   0,   0,  -1,  -4,   0,  -2,  -5,   0,  -3,   2,   1,  -5,   2,   1,   1,  -2,  -1,  -1 },
     {    1,   1,  -2,   7,   4,  -2,   3,  -4,   5,  -2,   1,   0,   1,   3,  -4,   1,   4,  -3,  -1,   1,   5,  -5,   3,  -3,   1,   2,   3,  -1,  -1,  -1,   0,   1,  -3,   4,   0,   2,   3,   7,  -3,  -2,  -5,   2,  -1,   0,   1,  -1,   0,   1,   6,  36, -22,   7,  -8,   0,   1,  -2,   2,  67, -54,  31, -19,   7,  -5,   1, -10,  42, -31,  16, -10,   7,  -4,   2,  -7, -32,  22, -11,   8,  -3,   1,  -1,   3,   6,  -4,   2,   3,  -2,   2,  -1,   4,  -2,   0,  -2,   2,  -4,   0,   0 },
     {    8,  11,  30,  60,  24, -23,  -4,  -6,  -2,   1,  -7,   1,  -2,  -2,   1,  -7, -34,  -1,  18, -22,  24, -18,  15, -10,  -2,  -1, -21, -29,  24,   8, -12,  21, -18,  13,  -8,   8,   4,  14,  36,  19, -20,   1,  -8,  -8,   0,  -1,  -2,   3,  -3, -12, -24, -14,  -2,   6,   0,   6,   2,   8,  13,   1,   2,   0,  -2,   1,  -2, -11,  -2,  -5,   3,  -1,   3,  -1,   0,  -1,  -2,  -1,   0,   1,   1,  -1,  -1,  -3,   4,   1,   3,   2,  -7,  -1,   0,   0,   2,  -2,  -7,  -1,  -3,   0 },
     {    1,   2,   1,  10,  37, -15, -22,  30, -31,  22, -23,  16,  -5,  -4, -13, -52, -19,  23, -19,  28, -14,  10,  -7,   8,   4,   6,  39,  24, -25,  24, -13,  -5,   6, -12,   5,  -5,  -2, -13,  -4,  22,  -7,  -8,  -1,   2,  -4,   3,  -2,   1,   2,   7, -11, -14,   8,  -2,   8,  -5,   3,  10,   5,  -6,   5,  -3,   1,  -4,  -1,  -7, -12,   4,  -4,   2,   0,   3,  -1,   3,   2,   0,  -2,   4,   3,  -4,   1,  -2,   2,  -2,   0,   3,  -1,   3,  -2,   1,   1,  -1,  -3,   0,   0,  -1 },
     {    2,   3,   7,  12,   7,   2,   0,  -4,  -3,  -2,  -2,  -1,   1,   0,  -2, -10,  -8,   8,   0,  -3,   9,   3,   3,   1,   1,   1,   0,   5,   3,  -6,   0,   7,  -5,  -5,  -3,  -3,  -2,  -1, -14, -13,  -2,   2,   6,  -7,   6,   0,   1,   3,   1,   0,  -2,   9,  13,  -8,  -4,   1,   1,  -4,   1,  -4, -14,  11,   2,   9,   4,   7,   6,   0, -12,   1,  -5,   0,  -4,  13, -13,  17,   8,   4,  -5,   0, -13,  10,  -4,  -3,  -8,   5,   8,   2,  49,  12, -11,  -1,  99, -20,  -1,  -1 },
     {    3,   4,  11,  14,  -2,   1,   6, -11,  -5,  -9,   2,   5,   6,  10,  35,  45,  -5, -13,  -1, -11,  -2,  -9,  10, -11,   5,  10,  47,  33, -31,  -6,  -4, -14,  12,  -4,  -2,  -2,  -2,  -6,  18, -12, -32,  15, -28,  33, -24,  15, -10,   3,  -2, -12, -11,  10,   0,  -8,   1,  -3,   1,  12,   5, -23,  -3,   1,   6,  -3,  -3,  -3,  -2,  -8,   6,   1,   2,  -2,  -2, -11,  -5,   2,  -6,   2,   3,   4,  -1,  -7,   2,  -2,   4,  17,  -7,   4,   8,   8,  -6,  -7,   5,  11,  -2,  -2 },
     {   -1,  -2,  -3,  -8, -11,  -5,   1,   0,  -2,  -1,   1,   3,   0,   0,   0,  -3,  -1,   4,  -3,   1,   3,   4,  -1,   5,   1,   2,   6,   6,  10,  -4, -15,   7,   0,   0,   3,  -5,  -1,   0,   2,   2,  -3,  -5,  -1,   0,  -2,   4,  -4,  -1,  -3,  -9, -14,  -2,  13,  -2,   4,  -4,  -2,   5,   7,   1,  -4,  -3,   3, -10,  -6, -35,  23, -17,   9,   2,  -2,   5,   8, -55,  12,  -4,   4,  -8,  10,  -3,  -1,  74, -16,   1,  -3, -47,  15,  -3,  24,   0, -17,   6,  -7,   4,   1,   2 },
   },
   {
     { -112,  45,  11,   5,   2,   2,   1,   1,   0,   1,   0,   0, -20,   9,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  32, -11,  -5,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -13,   8,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0, 109, -38, -12,  -4,  -2,  -1,  -1,   0,  -1,   0,   0,   0,  27, -14,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -38,  12,   6,   1,   1,   0,   1,   0,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {  -28,  30,  -7,   1,  -2,   0,  -1,   0,   0,   0,   0,   0,  17, -11,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,-109,  24,  16,   4,   4,   2,   2,   1,   1,   0,   1,   0, -11,  12,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  33,  -6,  -6,  -1,  -1,  -1,  -1,   0,   2,  -2,   0,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   37,  83, -49, -11,  -9,  -2,  -3,  -1,  -2,  -1,  -1,  -1,  -8,  16, -11,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   5, -63,  26,   7,   4,   1,   1,   1,   1,   0,   1,   0,  -3,  -6,   7,   0,   0,   0,   0,   0,   0,   0,   0,   0,   4,  12,  -4,  -2,  -1,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    0,  10,  -8,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  40, -10,  -3,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   5,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0, 107, -31, -13,  -3,  -3,  -1,  -1,   0,  -1,   0,  -1,   0,   6, -13,   1,   0,   1,   0,   0,   0, -39,   9,   6,   1,   1,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {    3, -16,   4,   2,   2,   1,   1,   0,   0,   0,   0,   0,  -2,   5,  -2,   1,   0,  -1,   0,   0,   0,   0,   0,   0,  29, -13,   1,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,  -4,   8,   0,   0,  -1,   0,  -1,   0,   0,   0,   0,   0, 113, -20, -14,  -3,  -4,  -1,  -2,  -1,  -5, -14,   2,   0,   1,   0,   1,   0, -34,   7,   5,   0,   1,   0,   1,   0,   3,   1,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0 },
     {    5,  10,  21, -12,  -2,  -3,   0,  -1,   0,   0,   0,   0,  28,  86, -39, -12,  -8,  -3,  -2,  -2,  -1,  -1,  -1,  -1,  -7,  20, -31,   6,   0,   2,   1,   1,   0,   0,   0,   0, -19, -53,  22,  12,   3,   3,   1,   1,   0,   1,   0,   1,   4, -15,  13,   0,   0,   0,   0,   0,  -1,  11,  -4,  -3,   0,  -1,   0,   0,  -2,   2,  -2,   1,   0,   0,   0,   0,   0,  -4,   2,   1,   0,   1,   0,   0,  -2,  -1,   1,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },
     {  -15, -21, -79,  36,  10,  10,   3,   4,   1,   2,   0,   2,   5,  19, -16,   4,  -2,   1,  -1,   0,  -1,   0,   0,   0,  17,  43,  51, -28, -11,  -8,  -4,  -3,  -2,  -1,  -1,  -1,  -3, -13,  10,  -5,   1,  -1,   1,   0,   0,   0,   0,   0,  -3, -18, -11,   7,   3,   2,   1,   1,  23,   0,  -3,   0,  -1,   0,  -1,   0,   3,   0,   1,  -1,   0,   0,   0,   0,  -4,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },
     {    4,   5,  14,  -9,  -2,  -2,  -1,   0,   0,   0,   0,   0,   8,  -4,   0,  -2,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1, -10,   7,   1,   2,   0,   0,   0,   0,   0,   0,  39, -11,  -2,   0,   0,   0,  -1,  -1,  -1,   0,   0,   0,  11,  10,  -1,  -2,  -1,   0,   0,   0, 110, -19, -13,  -3,  -4,  -1,  -2,   0,   7, -15,   1,   0,   1,  -1,   0,   0, -30,   5,   5,   0,   1,   0,   0,   0,   0,   2,   0,   0,   3,  -1,   0,   0,   0,  -1,   0,   0,  -1,  -1,   0,   0 },
     {   24,  60,  -9, -13,  -9,  -4,  -3,  -2,  -2,  -1,   0,  -1,  -5, -14,  21,  -6,   1,  -2,   0,  -1,   1,   0,   0,   0,  26,  68, -35,  -7,  -4,  -1,  -1,  -1,  -1,   0,   0,   0,  -9,  23, -27,   3,  -1,   1,   0,   1,   0,   0,   0,   0,  -1, -50,  13,  10,   2,   2,   1,   1,   6,  -9,  11,   1,  -1,  -1,   0,   0,  -5,   8,  -1,  -2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   1,   0,  -1,  -3,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0 },
     {    2,  -1,   1,  -3,   1,   0,   1,   0,   0,   0,   1,   0,   0,   4,  -4,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   8,  -5,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  16,   3,  -1,  -3,   0,  -1,   0,  -1,  -1,   0,   0,  32, -16,   3,  -2,   1,  -2,   0,   0, -12,  10,  -2,   0,   1,   1,   1,   0, 113, -14, -14,  -2,  -4,  -1,  -2,  -1, -19, -16,   4,   0,   1,   0,   1,   0, -24,   5,   5,   0,   4,   2,  -2,  -1,   4,  -1,  -2,   0,  -1,  -2,   0,   1 },
     {    1,  -2,  15,   9,  -8,  -4,  -1,  -2,   0,   0,   0,   0,  22,  67,  -1, -10, -10,  -5,  -3,  -3,  -1,  -1,  -1,  -1,  -2,  -5,  14, -17,   5,   1,   0,   0,   0,   0,   1,   0,  18,  70, -40, -10,   1,  -2,   0,  -1,   0,   0,  -1,  -1, -14,  14, -21,   9,  -1,   0,   0,   1,  -6, -45,  15,   6,   1,   2,   1,   1,   1,  -6,   8,  -1,  -1,   0,   0,   0,  -2,   5,  -2,  -1,   0,  -1,   0,   0,   3,  -1,  -2,   0,   1,  -2,   1,   1,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   -6,  -6, -12, -57,  24,   8,  10,  -2,   2,   0,   3,   1,  -5,  -2, -52,  17,  13,   4,   2,   0,   1,   1,   0,   0,   6,  19,   5,  55, -21,  -8,  -8,   0,  -2,  -1,  -1,   0,  14,  38,  33, -21, -16,  -5,  -2,   0,  -2,  -1,  -1,  -1,  -8,  -5,  -1, -21,   4,   2,   3,   0,  -2, -21,  -3,   7,   5,   1,   0,   0,  -4,  -1,   1,   6,   0,  -1,  -1,   1,  16,  -2,   0,  -1,  -1,   0,   0,   0,   0,  -3,  -2,   0,  -4,   0,   1,   1,  -3,   0,   2,   0,   1,   1,   0,   0 },
     {    8,  11,   2,  57, -27,  -8,  -7,   1,  -3,   0,  -2,   0, -10, -15, -57,  31,   3,   7,   0,   3,   0,   2,   0,   1,  -2,   4, -41, -33,  23,  10,   5,   0,   2,   1,   1,   1,  13,  26,  38, -23,  -3,  -4,  -1,  -2,   0,  -1,   0,   0,   4,  14,  22,   5,  -8,  -5,  -2,  -1,  -2,  -8,  -7,   3,   2,   0,   0,   0,  -2,  -7,  -3,  -1,   2,   1,   0,   0,   9,   1,   2,   0,  -1,   0,  -1,   0,  -1,   1,   1,   0,  -8,   2,   1,  -1,  -1,   1,   0,  -1,  -1,   0,   1,   0 },
     {   -1,  -4,   6,  -5,  -2,   0,  -1,   1,   0,  -1,  -1,  -1,  -8,  -6, -20,   7,   0,   4,   1,   1,   1,   1,   1,   1,  -4,  -9,   7,  -1,  -1,   1,  -1,   0,   0,   0,   0,   1,  -5,   4,   7,  -2,   0,  -2,   0,  -1,  -1,  -1,   0,   0, -11, -31,   3,   4,   4,   3,   0,   1, -25,  -4,   2,   2,   1,   1,   0,   1, -21,  -3,   7,   0,   0,  -1,   1,   0,-109,  23,  12,   0,   2,   1,   1,   0,   3,  15,  -3,   2,  19,  -1,  -3,   0,   3,   0,   0,   0,  -4,  -1,   1,   0 },
     {  -13, -28, -41,  15,  14,   4,   2,   1,   1,   1,   1,   1,   0,   4,  -6, -15,   3,   3,   4,   0,   0,   0,   1,   0, -17, -43, -34,  22,   7,   2,   4,   3,   3,   1,   0,   0,   3,  13, -21,  33,  -2,  -1,  -4,   0,  -1,   1,  -1,   0,  -8, -42,  55,   1,  -8,   1,  -3,   1,  12, -22,  20, -10,  -2,   0,   0,   0,   3,  28, -12,  -7,   3,  -1,   0,  -1,   8,   9,  -5,   1,   1,   0,  -1,   1,  10,  -8,   3,   1,  -2,   1,   1,   0,   2,   2,  -1,   0,   0,  -3,   3,   0 },
     {   -4,  -3, -47,   3,   7,   7,   3,   1,   0,   1,   1,   0,   4,   9,  22, -12,  -2,  -4,   1,   0,   0,  -1,  -1,  -1,   4,  25, -44,  20,   0,  -4,  -2,   1,   0,  -1,   1,   1,  -3,   2, -10,  12,  -2,   4,  -2,   1,   0,   0,   0,   0,  17,  70,  20, -28,  -6,  -1,  -1,  -3, -21,  -4,  -6,  -1,   3,  -1,   0,   1,  -5, -43,   2,  11,   2,   1,   1,   2, -26,   3,   4,   0,   0,   0,   1,   0,  10,  12,   0,  -3,   9,  -5,  -2,   0,  -3,  -3,   2,   0,  -2,   2,   0,  -1 },
     {    2,   2,   6,  -1, -18,  12,   2,   1,  -1,   0,  -1,   0,   0,   2,  -2,   4,  -3,   2,  -1,  -1,   0,   1,   0,   0,   4,   3,   7,   5,  12, -10,  -5,  -2,   0,   0,  -1,   0,   0,   3,   1,   0,  -1,  -3,   1,   1,   1,   0,   1,   0,   3, -10,  -6,  -3,  -1,   3,   1,   1,   1,  17,  -8,   2,   1,  -2,  -1,   0,  24,  -6,   0,   2,   2,  -1,   0,   0,  -1,  13,  -1,  -2,  -1,   0,   0,   1, 114, -16, -13,  -2, -13, -18,   4,   0, -15,   5,   3,   0,  -3,   1,   0,   0 },
     {    7,  11,   7,  20,  70, -38, -13,  -6,  -2,  -3,   0,  -2,   0,   0,  -1,  10,   0, -15,   2,  -1,   1,   0,   0,  -1,  -5,  -8, -11, -14, -68,  33,  15,   5,   1,   3,   1,   0,   0,   1,  -3, -10,   1,  13,  -2,   0,  -1,   2,   0,   1,   3,   3,   3,   9,  30, -13,  -8,  -1,  -1,   1,  -3,   2,  -2,  -4,   2,   1,   4,  -7,   1,  -4,  -8,   3,   4,   0,  -2,   4,   0,   1,   1,   1,  -1,   0,  24,  -3,  -2,   1,  -8,  -4,   2,   0,   0,  -1,   0,  -1,   2,  -1,   0,   0 },
     {   -2,  -5,  -5,   4, -12,   3,   3,   2,  -1,   3,   0,   0,  -2,  -4, -33,  58,  -9,  -4,  -3,   0,  -3,   0,  -1,   1,  -1,  -3,  -9,  25,  -3,  -4,  -2,  -2,  -1,  -1,  -1,   1, -15, -23, -69, -29,  23,  11,   7,   3,   4,   3,   2,   1,   4,   2,  -8,  -4,   6,   4,   1,   1,   8,  18,  47,   7, -15,  -5,  -5,  -2,   4, -13,  11,   5,  -2,   1,  -1,   2,   2,   1,  -2,  -8,   0,   2,   1,   1,  -5,   7,   1,  -5,  -4,  -1,  -1,   0,   0,   1,  -1,   1,  10,  -2,   0,  -1 },
     {    0,  -2,  -2,  17,  -9,   0,  -2,   2,  -2,   3,  -3,   1,  14,  25,  52,  40, -26,  -8, -11,   0,  -5,  -4,  -1,  -1,  -2,  -8,   1,  20,  -3,   0,  -2,   0,   1,  -1,   0,  -1,   3,  13,  17, -51,   9,   7,   2,   0,   0,   1,   1,   0,  -2, -20,  28, -37,  10,   4,   2,   0,   6,  14, -30,   8,   3,  -2,   1,  -2,  -8,  22, -18,  12,  -4,   0,   0,  -1,  -2, -14,   1,   3,  -1,   1,   0,   1,  -2,  -3,   3,  -1,  32,  -1,  -6,  -1,  -5,  -3,   1,   0,  -5,  -3,   1,   1 },
     {   -2,  -4,  -4,  -1,  -5,   2,   0,   2,   0,   1,  -2,   1,   3,  10,  12,   0,  -6,  -1,  -2,   1,  -1,  -1,   0,   1,   0,  -4,   1,   7,   0,   1,   1,  -1,   2,  -1,   0,   0,   2,  17,  -5,  -4,   4,  -1,   0,  -1,  -1,   1,   0,   1,  -2,  -9,   6,  -8,   0,  -3,   0,   0,   1,  30, -21,  -1,  -1,   0,   0,   0, -13, -16,  -5,   4,   2,   1,  -1,   0, -22,  -9,   7,   3,   1,   1,   1,   1, -17, -20,   5,   2,-107,  15,  11,   0,  12,  17,  -3,  -1,   5,  -1,  -2,   1 },
     {   -1,  -2,  -1,  -1,  -6,   1,   5,  -5,   1,  -2,   2,   0,   3,  -4,  34,  15,   1,  -7,  -4,  -4,  -2,  -1,  -2,   0,  -3,  -3,  -3,  -2,   4,  -1,  -7,   4,  -1,   2,   1,   1,  -9, -42,  27, -24,   2,   6,   4,   6,   2,   0,   0,   0,  -3, -12,   5,  -1,   2,   1,   2,  -2, -12, -82,  14,  20,   5,   3,   0,   0,  20, -14,  19,   0,   0,  -4,   0,   0,   4,  30,  -9, -10,  -1,  -3,   2,  -2,  -2,   9, -11,  -4, -35,   0,   4,   3, -10,   8,   3,  -1,   8,   5,  -3,  -1 },
     {   -8, -10, -16, -62,   5,  17,  10,   6,   0,   6,   0,   2,   0,   0,   1,   3, -17,   6,   0,   0,   0,  -1,   0,   1,  -3,  -3,  -5, -43,   4,  13,   0,   5,   2,  -1,  -1,   1,   2,   2,   2, -33,  44,  -4,   4,   0,  -2,   0,   0,   0,   9,  19,  32,  50,  -5, -20,  -5,  -6,  -1,   2,  -1,   2, -21,   5,  -1,   1,  -2,   6, -23,  -4,   8,   2,   1,   0,  -3,   0,  -4,   4,   2,   0,  -2,   0,   5,  -8,   3,   4,   3,   0,   4,  -1, -11,   3,   0,   2,  -9,  -2,   1,   1 },
     {   -1,  -2,  -6,  -5,  -2,   2,   7,  -4,   2,   0,   1,   0,   1,   1,   5,   7,   2,   4,  -3,  -2,   0,  -1,  -1,   0,   0,  -1,  -8,   2,   5,   3,  -4,   3,   1,   0,  -1,  -1,   0,  -4,   3,  -4,   5,   2,   0,   4,   0,   2,  -1,   0,   0,  -4,  -8,   9,   6,  -2,   1,   0,  -2,  -5,  -9,   8,  -2,  -3,  -2,  -1,   1, -12,   5,   0,  -3,   0,   0,   0,  11,  14,   4,  -2,  -4,   0,   0,   0,  16,  -2,   0,  -3,   9,  22,   1,   0, 113, -16, -10,  -1, -27, -16,   3,   0 },
     {    7,  11,  20,   5,  22,  53, -30,  -6,  -8,   0,  -2,   0,   4,   6,   3,  29,  21,  -8, -13,  -2,  -3,   0,   1,   0,  -1,   1,  11, -12, -19, -64,  26,  11,   6,   2,  -1,   0,   0,  -1,  -6,   4, -24,   6,  14,  -2,  -2,   1,   1,   0,   4,  13,  26,  14, -15,  22, -16,  -6,  -2,  -4,  -3, -16,   3,   5,  -4,   1,  -1,  -1, -19, -10,  12,  -3,   4,   0,  -2,  14,  -3,   1,   5,  -1,  -2,  -1,  -4,  -4,   4,   4,  -2,  -1,   2,   0,   8,   1,  -5,   1,   0,  -6,   0,   0 },
     {   -4,  -7, -26,  -2,   7,  21, -25,   9,  -3,   2,  -6,   2,   2,   4,  10,  -6,   3,  -3,   6,  -1,   0,   1,   5,   0,  -8, -14, -54,   8,   6, -21,  25,  -5,   6,  -1,   0,   0,   2,   4,  10, -15,   4,  -1,  -4,  -2,  -2,   1,   2,   1,  -6,  -8, -59,  26,  13,  12, -12,   4,   0,  11, -21,  28, -14,   8,  -7,   2,   5,  20,  34, -17, -11,  -4,   3,  -2,  -2,  -4,  14,  -7,  -1,   0,  -2,  -1,  -8, -10, -13,   3,   3, -10,  -1,   2,  -9,   6,   8,   0,   3,  -9,   3,   3 },
     {   -2,  -5,   1,  -5,   9,   4,  10,  -4,  -1,   1,   4,   0,   0,   1,  -1,   3,   6, -14,   0,   0,   3,  -2,   1,   0,  -3, -17,   5,  -5,  12,  -6,  -1,   2,   2,   0,   0,  -1,  -1,   6, -12,   4,  -3,   8,   2,  -4,  -3,  -1,   1,   0,  -6, -41,  18,  27,  -5,   2,  -2,   1,  -1,   7, -37,   7,   2,  -3,   3,   3, -14, -73,  16,  14,  -4,   5,  -1,  -1,  27, -15,  32,  -4,  -5,   0,  -1,  -1,  -1,  43, -10, -10,  10,  -1, -10,  -1, -17,  -9,   8,  -1,  -4,   7,  -1,   1 },
     {   -4,  -5, -11, -18,  11, -37,  22,   9,   2,  -2,   2,  -3,   5,   9,   7,  11,  54, -42,   3,  -9,  -1,  -4,   2,  -3,  -1,  -1,  -4, -16,  32,  11, -11, -12,  -2,  -2,   0,   1,  -3,  -2,  -8, -25, -52,  34,   8,   8,   1,   2,   0,   3,   2,   9,   2,  12, -26,  16,   2,   6,   3,  15,   3,   3,   9,  -4,  -7,  -2,   1,  19,   0,  -2,   4,  -5,  -1,  -2,  -9,   1,   2,   4,  -3,  -1,   1,   1,   2, -10,   0,  -1,  -1,  -5,   3,   1,   1,   3,   2,   0,  -3,  -2,   0,   1 },
     {   -1,  -1,  -4,  -6,   1, -18,  -2,   0,   1,   1,  -2,   2,   7,   9,  20,  65, -16, -10,   0,  -3,   5,  -1,  -2,  -2,   5,   9,   3,  19,   9,  14,   7,   5,  -2,  -2,   5,  -2,   5,  10,  22,  45,   4,  -5, -13,   2, -10,   1,   1,   0,  -2,   1, -11,  20, -27, -17,  -9,  -2,  -2,   5,   7, -37, -11,  19,   9,   1,   3,  -3,  27, -29,  17,   7,  -5,  -3,  -2,  -2, -12,   5,   6,  -9,  -3,   1,  -2,   2, -11,  10,  -1,   0,   4,  -3,  -5,  -6,  12,  -5, -25,   5,  -3,  -1 },
     {   -1,   0,  -2,  -8,   3, -18,  -5,  25, -10,   5,  -2,   5,   1,  -2,  -6,  20,  -2,  -3, -17,   3,  -1,   1,   2,   1,   3,   7,   6,   3,  10,   4,   1, -13,   8,  -2,   1,  -4,   0,  -3, -16,  19,  11, -16,   6,   1,  -2,  -7,  -1,   1,   2,   8,   5,   7,  -7, -12,  -3,   1,  -6, -14, -41, -10,  12,  16,   5,  -2,   5,  19,  -6,   9,  -9,   5,   0,  -3,   1,   9,  14,  -6, -12,  -6,  -4,   3,   3,   7, -17,  -4,   7,  -7,   0,  -1,  21,  21,   0,   5,  86, -15,  -7,   3 },
     {    5,   8,   7,  10,  55,   5, -16, -15,   1,  -1,  -5,   2,   4,   6,   8,  -3,  26,  23,  -3,  -8,  -9,  -2,  -6,  -1,   0,  -2,  -3,  -9,  23,   1, -15,   0,   3,  -6,   2,   1,  -2,  -2,   0, -22,  13, -46, -12,  16,   8,   6,   2,   2,  -4, -10,  -7, -24, -51, -10,  24,   6,   2,   8,   8,   6, -34,  15,   2,  -6,   1,  -3,  11,  12,  17,   6,  -6,   7,  -2,  -5,  -4,  -3,   7,   8,   1,  -3,   5,   8,   5,  -3,   1,   7,  -4,   5,   4,   1,   4,   4,  10,   5,   4,  -4 },
   }
 },
 {  //18
   {
     {  117, -34,  -5,  -5,  -2,  -2,  -2,  -1,  -1,  -1,  -1,   0,  33,  -9,  -2,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   0, -12,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0, -11,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -7,   2,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   30,  -9,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,-115,  22,   7,   4,   3,   2,   2,   1,   1,   1,   1,   0, -36,   7,   2,   1,   1,   1,   1,   0,   0,   0,   0,   0,  12,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   9,  -2,  -1,   0,   0,   0,   0,   0,   6,  -1,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,   2,   0,   0,   0 },
     {   21,  -6,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0, -27,   3,   1,   1,   1,   1,   0,   0,   0,   0,   0,   0, 113, -16,  -8,  -3,  -3,  -2,  -2,  -1,  -1,  -1,  -1,   0,  41,  -6,  -4,  -1,  -2,  -1,  -1,   0,   0,   0,   0,   0,  -6,   0,   0,   0,   0,   0,   0,   0, -10,   2,   1,   0,   0,   0,   0,   0,  -9,   2,   1,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,  -3,   1,   0,   0,  -2,   0,   0,   0,  -1,   0,   0,   0 },
     {  -27, -92,  33,   2,  10,   3,   4,   2,   3,   1,   1,   1, -10, -69,  22,   2,   6,   2,   2,   2,   1,   1,   1,   1,  -3,  -6,   1,   1,   0,   1,   0,   0,   0,   0,   0,   0,   7,  12,  -5,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   4,  12,  -4,   0,  -1,   0,   0,   0,   2,   8,  -2,   0,  -1,   0,   0,   0,   1,   5,  -2,   0,  -1,   0,   0,   0,   1,   5,  -2,   0,   0,   0,   0,   0,   1,   3,  -1,   0,   1,   2,  -1,   0,   0,   2,  -1,   0,   0,   2,  -1,   0 },
     {    3,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  25,   0,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -35,   4,   1,   1,   1,   0,   1,   0,   0,   0,   0,   0, 110, -14,  -9,  -3,  -3,  -2,  -2,  -1,  -1,   0,  -1,   0,  41,  -5,  -4,  -1,  -1,  -1,  -1,   0, -16,   1,   1,   0,   0,   0,   0,   0,  -9,   1,   1,   0,   0,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0,   0,  -3,   0,   0,   0,  -2,   0,   0,   0,  -3,   0,   0,   0,  -1,   0,   0,   0 },
     {   15,  58, -14,  -1,  -6,  -2,  -3,  -1,  -2,  -1,  -1,   0, -14, -73,  18,   4,   5,   3,   2,   2,   1,   1,   1,   1, -10, -77,  17,   4,   5,   3,   2,   2,   1,   1,   1,   1,  -1,  -8,   1,   2,   0,   1,   0,   1,   0,   0,   0,   0,   3,   7,  -2,   0,  -1,   0,   0,   0,   3,   8,  -2,   0,  -1,   0,   0,   0,   0,   6,  -1,   0,   0,   0,   0,   0,   0,   6,  -1,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   3,  -1,   0,   0,   1,   0,   0,   0,   2,   0,   0 },
     {    7,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   4,   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  20,   0,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -27,   4,   1,   1,   0,   1,   0,   0,   0,   0,   0,   0, 106, -11,  -9,  -3,  -4,  -1,  -2,   0,  59,  -6,  -6,  -1,  -2,  -1,  -1,   0,   2,  -1,  -1,   0,   0,   0,   0,   0, -11,   1,   1,   0,   0,   0,   0,   0,  -7,   1,   1,   0,  -4,   0,   0,   0,  -2,   0,   0,   0,  -3,   0,   0,   0 },
     {    8,  22,  73, -46,   3,  -7,   0,  -4,  -1,  -2,   0,  -1,   7,  20,  69, -44,   2,  -6,  -1,  -4,  -1,  -2,   0,  -1,   2,   2,  20, -13,   0,  -1,  -1,  -1,   0,   0,   0,   0,  -1,  -4,  -4,   3,  -1,   1,   0,   1,   0,   0,   0,   0,  -2,  -3, -11,   7,  -1,   1,   0,   1,  -2,  -3,  -9,   6,   0,   1,   0,   0,   0,  -2,  -7,   4,   0,   0,   0,   0,   0,  -2,  -5,   3,   0,   0,   0,   0,  -1,  -1,  -4,   2,   0,  -1,  -3,   1,   0,  -1,  -2,   1,   0,  -1,  -2,   1 },
     {   -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -6,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -33,   5,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,  53,  -4,  -3,  -1,  -1,  -1,  -1,   0,-110,  11,   9,   2,   3,   1,   2,   1,  -4,   1,   1,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   4,   0,   0,   0,   1,   0,   0,   0,   3,   0,   0,   0 },
     {   -8, -38,   7,   3,   4,   1,   2,   1,   1,   0,   1,   0,   8,  41, -11,  -1,  -3,  -2,  -1,  -1,  -1,   0,   0,   0,  -7, -60,  14,   4,   4,   2,   1,   2,   1,   1,   0,   1, -10, -87,  20,   7,   4,   4,   1,   3,   1,   1,   0,   1,  -1, -20,   4,   3,   0,   1,   0,   1,   2,  12,  -2,  -1,  -1,   0,   0,   0,   1,  15,  -3,  -1,  -1,  -1,   0,  -1,   1,   7,  -1,  -1,   0,   0,   0,   0,   0,   5,  -1,   0,   0,   3,  -1,   0,   1,   3,  -1,   0,   0,   1,  -1,   0 },
     {   -5, -15, -57,  15,   3,   6,   3,   3,   1,   1,   1,   1,   2,   4,  28, -17,   1,  -1,   1,  -1,   0,  -1,   0,   0,   8,  15,  91, -37,  -2,  -6,  -3,  -3,  -2,  -1,  -1,  -1,   4,   8,  29, -13,  -2,  -1,  -2,  -1,  -1,   0,   0,   0,   1,   1,  -8,   2,  -1,   1,   0,   0,  -1,  -2,  -9,   4,   0,   0,   0,   0,   4,  -1,  -8,   3,   0,   0,   0,   0,   2,   0,  -6,   3,   0,   0,   0,   0,   0,  -1,  -3,   2,  -1,  -1,  -2,   1,  -1,   0,  -1,   1,  -1,   0,  -1,   1 },
     {   -5,  -1,  -1,   2,   0,   0,   0,   1,   0,   0,   0,   0,  -6,   2,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -7,   1,   5,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -8,   4,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -7,   0,   0,   1,   0,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0,   0,-113,  12,  10,   2,   4,   1,   2,   0, -53,   5,   5,   1,   2,   0,   1,   0,   0,   0,   0,   0,  11,  -1,  -1,   0,   7,  -1,  -1,   0,   1,   0,   0,   0 },
     {    4,  20,  -6,  -3,  -2,   0,   0,   0,   0,   0,  -1,   0,  -7, -51,  11,   1,   2,   2,   1,   2,   1,   1,   0,   1,   8,  73, -11,  -5,  -5,  -3,  -2,  -2,  -1,  -1,  -1,  -1,  -8, -68,  11,   7,   4,   3,   1,   2,   1,   1,   0,   1,  -2, -44,   7,   3,   2,   2,   0,   1,   1,  24,  -3,  -1,  -1,  -1,   0,  -1,  -2,   9,  -1,  -1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,  -1,   0,  -1,   0,   0,   0,   1,   1,   0,   0 },
     {    8,  11,  36,  75,  -4, -19, -10,  -6,  -4,  -2,  -2,  -1,   7,  12,  41,  70,  -4, -20, -10,  -6,  -3,  -2,  -2,  -1,   3,   7,  19,  17,  -1,  -7,  -3,  -2,  -1,  -1,  -1,  -1,  -1,  -2,  -2, -10,   1,   1,   1,   0,   0,   0,   0,   0,  -1,  -4,  -6, -15,   1,   3,   1,   1,  -1,  -1,  -5, -10,   0,   3,   1,   1,   0,  -1,  -5,  -7,   0,   2,   1,   1,   0,  -1,  -4,  -5,   0,   2,   1,   1,  -1,  -1,  -2,  -4,  -2,   0,  -1,  -2,   0,   0,  -1,  -2,   0,   0,  -1,  -1 },
     {    1,   3,  -8,  -1,  63, -47,  17, -10,   4,  -3,   1,  -1,   0,  -1, -10,  -4,  67, -49,  18,  -9,   4,  -2,   1,  -1,   0,   0,  -4,  -6,  32, -20,   7,  -3,   1,   0,   0,   0,  -1,  -1,  -2,  -3,   4,   1,  -1,   1,  -1,   0,   0,   1,   0,  -3,   1,  -1, -11,   8,  -4,   2,   0,  -3,   3,   1, -11,   7,  -3,   2,  -1,  -1,   3,   1,  -8,   6,  -2,   1,   1,   0,   2,   1,  -6,   4,  -1,   1,  -1,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -3,  -8, -35,   5,   4,   5,   1,   1,   1,   1,   0,   1,   3,   8,  27,  -8,  -1,  -2,  -1,  -1,  -1,   0,   0,   0,   1,   2,   9,  19,  -3,  -4,  -3,   0,  -1,   0,   0,   0,  -8, -15, -90,  44,   5,   1,   2,   2,   2,   1,   1,   1,  -4,  -2, -47,  13,   5,   0,   3,   0,   1,  -1,  19, -10,   0,   0,   0,   0,   2,   2,  18,  -8,  -2,   0,   0,   0,   2,   1,   6,  -3,  -1,   0,   0,   0,   0,   1,   3,  -2,  -1,   1,   2,  -2,  -1,   1,   2,  -1,   1,   0,   1,  -1 },
     {   -1,   0,   0,   1,   0,  -1,   1,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   3,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -9,   0,   1,   0,   1,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  54,  -4,  -4,  -1,  -2,   0,  -1,   0,-107,  10,   9,   1,   3,   1,   1,   0, -24,   1,   3,   0,  31,  -2,  -2,   0,  -3,   0,   0,   0,   1,   0,   0,   0 },
     {    1,   6,   0,  -3,   1,  -2,   0,   0,   0,   0,   0,   0,   1,  11,  -3,  -1,   1,  -2,   1,   0,   0,   0,   0,   0,   1,  12,  -1,  -2,   2,  -2,   0,   0,   0,   0,   0,   0,  -1, -12,   1,  -5,   2,   1,   0,   0,   0,   0,   0,   0,   6,  77, -15, -15,  -2,  -4,   0,  -2,   8,  90, -13, -13,  -3,  -5,   0,  -3,   2,  19,  -1,  -3,  -1,  -1,   1,  -1,  -1, -13,   3,   2,   0,   1,   0,   1,  -2, -13,   2,   3,   0,  -8,   1,   2,   1,  -6,   1,   1,   1,  -3,   1,   1 },
     {   -2,   0,  -5, -36,   7,   2,   5,   1,   2,   1,   1,   0,   1,   1,   0,  15,   4,  -2,   0,  -1,   0,   0,   0,   0,   6,  11,  31,  87,  -2, -12, -10,  -5,  -4,  -2,  -2,  -1,   6,  11,  39,  54,  -7, -11,  -9,  -4,  -2,  -2,  -1,  -1,   1,   5,  14,   1,  -5,  -3,  -1,  -1,   0,   0,  -4, -14,  -2,   2,   1,   1,  -2,  -3,  -7, -10,   0,   3,   2,   1,  -1,  -1,  -3,  -6,   0,   2,   1,   0,   0,  -2,  -2,  -5,   2,  -1,  -2,  -3,  -1,   0,  -2,  -2,  -1,  -1,  -1,  -1 },
     {   -4,   0,   1,  -1,  -1,   1,   1,   1,   0,   0,   0,   0,  -5,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,  -6,   1,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -5,   1,   3,   0,   0,   0,   0,   0,   0,  -1,   0,   0,  -6,   1,   1,   0,   0,   0,   0,   0,  -7,   0,   3,  -1,   1,   0,   0,   0,  -6,   1,   2,  -1,   0,   0,   0,   0,  -5,   4,   1,   0,   0,   0,   0,   0, -92,  11,   7,   1, -84,   6,   8,   1, -18,   1,   2,   0,  -3,   0,   0,   0 },
     {    4,  10,  51, -16,  -5,  -2,  -1,  -3,  -1,  -2,   0,  -1,  -6, -14, -77,  22,   3,   4,   4,   2,   1,   1,   1,   1,   5,  10,  63, -15,  -3,  -3,  -3,  -2,  -1,  -1,  -1,  -1,  -1,  -3, -18,   0,   0,   2,   0,   0,   0,   0,   0,   0,  -4,  -5, -37,   9,   1,   1,   2,   1,   0,   0,  19,  -3,  -1,  -1,  -1,   0,   1,  -3,   2,   0,  -1,  -1,   0,   0,   0,   0,   6,  -1,  -1,  -1,   0,   0,   1,   0,  -1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,  -1,   1,   1,   0 },
     {    0,  -3,   1,   1,   3,   0,  -1,   0,   0,   0,   0,   0,   1,  10,  -3,  -5,   1,   1,   0,   0,   0,  -1,  -1,   0,  -2, -16,   5,   4,  -3,   2,   0,   0,   0,   1,   0,   0,   5,  50, -12,  -3,  -2,  -1,  -1,  -2,  -1,   0,   0,  -1,  -7, -80,  10,   4,   4,   3,   2,   3,   6,  63,  -5,  -9,  -2,  -4,   0,  -3,   5,  44,  -6,  -5,  -1,  -3,   0,  -2,   1, -19,   2,   3,   0,   1,  -1,   1,  -3,  -3,   1,   1,   0,  -3,   0,   1,   2,  -1,   0,   0,   0,  -3,   1,   0 },
     {   -6,  -8, -22, -71, -21,   7,  11,   8,   5,   2,   2,   1,   5,   8,  21,  74,  16,  -9, -11,  -7,  -5,  -3,  -1,  -2,   0,  -1,  -1,  -4,  10,   3,   0,  -2,  -1,   0,   0,   0,  -4,  -5, -18, -47, -14,   7,   6,   4,   3,   2,   1,   1,   0,  -2,   4,  21,  -3,  -1,  -3,  -1,   0,   2,  -1,   4,   1,  -1,   0,   1,   0,   2,  -1,   4,  -1,  -1,   0,  -1,  -1,   1,  -1,  -1,   0,   1,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,  -1,  -1,   1,   0,   0,   1,   0,   0,   0 },
     {    0,   0,   1,   1,   1,   1,  -1,   1,  -1,   0,   0,   0,   1,   0,  -1,   2,   0,   1,  -1,   0,   0,  -1,   0,   0,   1,   0,   0,  -2,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,  -1,   1,   1,   0,   0,   1,   1,   0,   0,   0,   2,   2,   0,   1,   0,  -1,   0,   0,   0,  -3,   1,   1,   1,   0,   1,  -1, -10,   4,   0,   2,   0,   0,   0,   0,  38,  -3,  -6,   0,  -1,   0,   0,  -1, -80,   7,   5,   1,  89,  -6,  -6,  -1, -13,   2,   0,  -1,   2,   1,   0,   0 },
     {    0,   0,   5,  -6,  -1,   1,   1,  -1,   1,   0,   1,   0,   1,   0,  10,  -7,  -6,   1,   2,  -1,   0,  -1,   0,   0,   2,   1,  19,   0,  -4,   0,  -1,   0,  -1,   0,   0,   0,  -1,  -5, -14, -15,   3,   2,   0,   2,   2,   0,   0,   0,   5,   5,  60, -40,  -7,   2,  -2,  -1,   7,   5,  86, -28, -15,   1,  -5,  -1,   2,  -4,  30,  -4,  -8,   0,  -2,   1,  -1,  -7,  -3,   6,  -2,   1,   0,   0,   0,  -1, -10,   5,   1,   0,  -8,   3,   1,   0,  -6,   3,   0,   0,  -4,   2 },
     {   -1,  -3,   5,   2, -39,  29,  -9,   3,  -2,   3,  -1,   2,   0,   0,   0,  -4,   5,   3,  -2,   1,   0,   1,   0,   0,   2,   2,   0,   0,  79, -43,   8,  -7,   0,  -2,   0,  -1,   2,   3,   0,   1,  62, -36,   5,  -4,  -1,  -1,   0,   0,   0,  -3,   0,  -2,   8,  -3,  -2,   1,   0,  -1,  -2,  -3, -12,   9,  -2,   1,   0,   1,  -4,   1, -10,   5,   0,   0,   0,   0,  -2,   2,  -6,   2,   1,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   1,   0 },
     {   -5,  -8,  -8, -11, -50, -58,  22,   8,  11,   2,   3,   1,  -4,  -6,  -6,  -9, -53, -61,  22,   9,   9,   3,   3,   2,  -1,  -3,  -2,  -3, -17, -31,  11,   2,   2,   1,   1,   1,   1,   2,  -1,   4,   4,  -2,   3,   0,  -2,  -1,  -1,   0,   1,   1,   1,   0,  10,  13,  -3,  -1,   1,   2,  -2,  -1,   7,  12,  -4,  -1,   0,   2,  -2,   0,   8,   7,  -4,  -1,   1,   2,   0,  -1,   5,   5,  -3,  -1,   0,   2,   1,  -2,   2,   1,   0,  -2,   1,   0,   0,   1,   2,   0,   0,   0 },
     {   -1,  -6,  -1,   3,  -4,  -1,   0,  -1,   3,   0,   0,   0,   0,  -2,   3,   1,   2,   0,  -1,  -1,   0,   1,  -1,   0,  -1, -11,  -1,  -1,   1,  -1,  -1,   1,   0,   2,   0,   0,   0,  -2,   6,   3,   0,   2,  -1,   0,   2,   1,   0,  -1,  -2, -17,  -4,  12,   4,  -1,   1,   0,   2,  33,  -5,   0,   0,  -3,   1,   0,  -9,-100,  16,   7,   3,   8,   1,   3,  -4, -59,  14,  11,  -1,   4,   0,   2,  -2,   9,  -1,   1,   2,   6,  -2,  -2,  -4,   9,  -3,  -1,  -3,   5,  -2,  -1 },
     {   -1,  -1,  -8, -17,  22,   5,   5,  -3,   2,   0,   0,   0,  -1,  -3,  -3, -11,  -8,   5,   4,   0,   2,   0,   0,   0,   0,  -1,   4,  40, -18,  -2,  -2,  -3,   0,  -2,  -1,   0,  -2,  -6,  -8, -52,  29,   2,   4,   0,   1,   1,   0,   0,  -5, -14, -33, -75,  35,   8,   6,   2,  -3,  -9, -22,  -3,   8,   5,   2,   1,  -2,  -8,  -3,  16,  -2,   1,  -1,   0,   0,  -1,   0,   9,  -2,   0,  -2,  -1,  -2,   1,   2,   4,   1,  -1,   2,   5,   2,   2,   2,   2,  -2,   0,   1,   0 },
     {   -2,   0,   0,  -1,   1,   1,  -1,  -1,   0,   1,  -1,   0,  -2,   1,  -1,   1,  -1,  -3,  -1,   0,   0,   2,  -1,   1,  -2,  -1,  -1,  -2,  -1,   1,  -1,  -1,   1,   0,   1,  -1,  -3,   1,  -1,   3,   1,   2,   1,   0,  -1,  -1,   1,  -2,  -2,   1,   2,  -5,   2,   1,   1,   1,  -1,   1,   0,   0,   1,   0,   0,  -1,  -5,   6,  -2,   2,  -1,   0,  -1,   1,  -4,  -1,  -5,   2,  -1,   0,   0,   1,  22,   1,  -6,   1,   4,   3,  -2,   1,-118,   7,  11,   0, -38,   2,   3,   1 },
     {    0,  -2,   5,  16, -53,  38,  -5,  -1,  -1,  -1,  -2,   1,   0,   0,  -5, -28,  43, -24,  12,  -3,   3,  -2,   0,  -2,   1,   2,   1,  22,  14, -20,  11,  -3,  -1,  -1,   0,  -1,  -3,  -3,  -2, -14, -63,  33,  -2,   5,   0,  -1,   2,   2,  -2,  -4, -10, -24, -19,  16,   4,   2,  -1,  -1,  -6,   5,   8,   1,   2,  -2,   1,   0,  11,  -7,   6,  -1,   0,   0,   1,   1,   1,  -5,   3,   1,   0,   1,   0,   1,   0,  -4,   2,  -1,   0,   2,   1,   1,   0,   2,  -5,  -2,   1,   1 },
     {    0,   0,   1,   5,   4,   0,  -5,   4,   3,   1,  -1,   0,  -2,  -2, -12,  -7,  -3,   0,  -8,   5,  -1,   2,   0,  -1,   1,   3,  19,  10,   1,  -7,  -7,   1,  -1,   0,   0,  -1,  -4,  -8, -54,   1,   4,   6,   6,   0,   2,   1,   0,   0,   3,   7,  57, -10, -32,   5,  -6,   1,  -2,   4, -26,  19,   1,  -6,  -5,   3,  -7, -15, -69,  28,  19,  -6,   4,   1,   0,  -1,   2,   3,   3,  -4,  -1,   0,   0,  12,   7,  -5,  -1,  -6,   3,   3,   4,  -3,   4,   0,  -6,  -2,   4,   1 },
   },
   {
     {  -72,  63, -19,   6,  -2,   2,  -1,   1,  -1,   1,   0,   0,  61, -49,   9,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -18,   9,   6,  -5,   2,  -1,   1,   0,   0,   0,   0,   0,   4,   1,  -5,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -2,  -1,   2,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   74,   4, -43,  14,  -9,   2,  -3,   1,  -2,   0,  -1,   0,   6, -47,  45, -15,   5,  -1,   1,   0,   1,   0,   0,   0, -39,  42, -17,   1,   3,  -2,   1,  -1,   1,   0,   0,   0,  16, -14,   0,   6,  -4,   2,  -1,   1,   0,   0,   0,   0,  -9,   4,   4,  -4,   2,   0,   0,   0,   2,   0,  -3,   2,   0,   0,   0,   0,  -4,   1,   2,  -1,   0,   0,   0,   0,   1,   0,  -1,   1,   0,   0,   0,   0,  -2,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    7, -50,  57, -27,  11,  -4,   3,  -2,   2,  -1,   1,   0,  62,  -3, -36,  12,  -3,  -1,   0,   0,   0,   0,   0,   0, -55,  30,   1,   3,  -3,   2,   0,   0,   0,   0,   0,   0,  20, -10,  -1,  -1,   1,   0,  -1,   0,   0,   0,   0,   0,  -8,   2,   1,   1,   0,   0,   0,   0,   1,   2,  -1,  -1,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   58,   8, -13, -11,   1,  -3,   0,  -2,   0,  -1,   0,  -1,  52, -21, -20,  17,  -8,   3,  -3,   2,  -1,   1,  -1,   1,  19, -44,  38, -15,   4,   0,   0,   0,   0,   0,   0,   0, -46,  42, -16,   0,   5,  -3,   2,  -1,   1,  -1,   0,   0,  12, -12,   0,   6,  -5,   2,  -1,   1, -11,   6,   3,  -4,   2,   0,   0,   0,   1,  -1,  -1,   2,   0,   0,   0,   0,  -6,   3,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0 },
     {  -15, -51,  10,  23,  -3,   4,   0,   2,   0,   1,   0,   1,  46, -10,  36, -43,  12,  -6,   2,  -3,   1,  -1,   0,  -1,  46,   5, -45,  23,  -6,  -1,   0,   0,   0,   0,   0,   0, -43,   7,  19,  -3,  -4,   4,  -1,   1,   0,   0,   0,   0,   9,   3,  -9,  -2,   4,  -2,   0,   0,  -6,  -1,   2,   3,  -2,   1,   0,   0,   0,   4,  -2,  -2,   1,   0,   0,   0,  -4,   0,   0,   2,  -1,   0,   0,   0,   0,   2,  -1,  -1,  -2,   0,   0,   1,   0,   1,   0,   0,  -1,   0,   0,   1 },
     {    0,  12,  32, -59,  35, -18,   9,  -6,   3,  -3,   2,  -2, -32, -53,  17,  37, -16,   8,  -1,   2,   0,   1,   0,   1,  31,  36, -26,  -7,  -3,   2,  -3,   0,  -1,   0,   0,   0, -27,   0,   3,   4,   3,  -2,   1,   0,   0,   0,   0,   0,  18,  -4,   0,   0,  -3,   2,  -1,   0,  -7,   3,  -2,   0,   2,  -1,   0,   0,   5,  -1,   0,   1,  -2,   1,   0,   0,  -2,   1,  -1,  -1,   1,   0,   0,   0,   2,   0,   0,   1,  -1,   1,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {  -27, -30,   5,  17,  -1,   2,   1,   1,   1,   0,   0,   0, -31,  -4,  25, -14,   4,   1,   0,   1,   0,   0,   0,   0, -55,  10,  15, -10,  11,  -5,   4,  -2,   2,  -1,   1,   0, -26,  48, -35,  16,  -6,   1,   0,  -1,   0,   0,   0,   0,  46, -39,  15,  -2,  -4,   4,  -3,   2, -12,  16,  -3,  -4,   4,  -3,   1,  -1,  13,  -6,  -2,   3,  -2,   0,   0,   0,  -1,   3,   0,  -1,   0,   0,   0,   0,   6,  -2,  -1,   1,   0,   1,   0,   0,   3,  -1,   0,   0,   0,   1,   0,   0 },
     {   14,  44, -10,  -8,  -5,   0,  -2,  -1,  -1,   0,  -1,   0,   3,  37, -26,  -2,   3,  -4,   1,  -1,   0,  -1,   0,   0, -41,  15, -28,  33, -14,   6,  -1,   2,   0,   1,   0,   0, -46, -10,  44, -23,   7,   2,  -1,   1,   0,   1,   0,   0,  45, -11, -17,   4,   3,  -4,   2,  -1, -18,   0,  10,   0,  -3,   2,   0,   0,  11,  -3,  -3,  -1,   2,   0,   0,   0,  -3,  -3,   4,   0,  -1,   0,   0,   0,   4,  -1,  -1,  -1,  -1,  -2,   1,   0,   2,  -1,   0,   0,   0,  -1,   1,   0 },
     {   -7, -15, -42,  14,  29, -15,   9,  -2,   3,  -1,   2,   0,  26,  42,   1,  21, -55,  27, -14,   4,  -5,   2,  -2,   1,  10,  21, -14, -33,  36, -16,   3,  -1,   0,   0,   0,   0,  -2, -26,   6,  23, -13,   0,   4,  -2,   1,   0,   1,   0,  14,  -6,  12, -15,   1,   6,  -4,   1, -16,   5,  -6,   4,   6,  -6,   3,  -1,   5,  -3,   5,  -1,  -5,   3,  -1,   0,  -5,  -1,  -1,   0,   3,  -2,   0,   0,   1,  -1,   2,   0,  -2,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0 },
     {    1,  -3,   9,  25, -52,  38, -20,  10,  -6,   4,  -3,   2,  -4, -13, -41,  22,  27, -18,   8,  -1,   2,   0,   1,   0,  31,  50,  13, -38,   3,  -7,   4,  -5,   1,  -2,   0,  -1, -19, -29,  -3,  19,  -3,   5,  -2,   2,   0,   1,   0,   1,  18,  -1,   8,  -8,   0,  -1,   0,   0, -14,   3,   0,  -1,   2,   0,   0,   0,   5,  -1,  -1,   1,   0,  -1,   1,   0,  -4,   0,   1,  -2,   0,   1,   0,   0,   1,  -1,   0,   1,  -1,   0,   1,  -1,   0,   0,   0,   0,  -1,   0,   1,  -1 },
     {   15,  21,  11, -11,  -4,  -3,   0,  -2,   1,  -1,   0,  -1,  16,   8,  -3, -16,  12,  -5,   1,  -1,   0,   0,   0,   0,  33,  -5, -21,  21, -11,   2,   0,   0,   0,   0,   0,   0,  67,   2, -17,   1,  -8,   4,  -4,   1,  -2,   0,  -1,   0,  30, -48,  27,  -8,   5,  -1,  -1,   1, -44,  34, -11,   1,   2,  -3,   3,  -2,   8, -13,   1,   4,  -2,   2,  -1,   1, -11,   5,   2,  -2,   1,   0,   0,   0,   0,  -3,   0,   1,  -6,   2,   1,   0,  -1,  -1,   0,   1,  -3,   1,   1,   0 },
     {    9,  32,  10, -13,  -1,  -5,   1,  -3,   0,  -1,   0,  -1,   7,  36, -10, -17,   8,  -5,   1,  -1,   0,   0,   0,  -1,   5,  29, -25,   9,  -6,   0,   0,  -1,   0,  -1,   0,   0, -37,  21, -33,  27, -10,   5,  -1,   1,   0,   0,   0,   0, -48, -22,  44, -17,   5,   2,  -1,   2,  42,  -4, -19,   4,   2,  -3,   1,  -1, -16,  -5,  10,   0,  -1,   1,   0,   0,   9,  -2,  -3,   0,   1,   0,   0,   0,  -4,  -4,   3,   1,   4,  -1,   0,  -1,  -1,  -2,   1,   0,   2,  -1,   0,   0 },
     {  -13, -29, -47,  21,   8,   6,  -1,   5,   0,   2,   0,   1,  -2,  -6, -24,  35, -12,   3,   0,   0,   0,   0,   0,   0,  14,  36,  -9,  18, -32,  19, -11,   3,  -3,   1,  -2,   0,  11,  33,  -1, -44,  33, -17,   4,  -2,   0,   0,   0,   0, -16, -32,  11,  22, -11,   1,   4,  -2,  13,   8,   3, -13,   2,   3,  -3,   1, -10,  -2,   1,   3,   2,  -3,   2,   0,   4,  -1,   3,  -3,  -1,   1,  -1,   0,  -3,  -1,   1,   0,   1,   0,   1,  -1,  -1,  -1,   1,   0,   1,   0,   1,  -1 },
     {   -5, -10,   3,  22,  -8,  -7,   5,  -2,   1,  -1,   0,   0, -19, -35, -25,   9,  -3,  22,  -9,   8,  -2,   3,   0,   2, -14, -23,  -1,  20,   6, -11,   6,  -2,   1,   0,   0,   0, -22,  -8,  34, -12, -11,  11,  -3,   1,  -1,   1,   0,   0, -46,  -5,  25, -23,  24,  -9,   2,   0, -20,  44, -30,  16,  -8,  -3,   5,  -3,  26, -21,  10,  -3,  -3,   5,  -4,   2,  -5,  12,  -2,  -2,   3,  -3,   2,  -1,   7,  -3,   0,   0,   0,   4,  -1,  -1,   4,  -1,   0,   0,   0,   3,   0,   0 },
     {   -9, -14, -20, -38,   2,  49, -24,  13,  -5,   5,  -1,   3,   8,  15,  34,  40,  -2, -53,  23, -11,   2,  -4,   0,  -2, -16, -31, -23, -15,   1,  28,  -4,   2,   2,   1,   1,   1,  -3,   1,  15,  11,  -2,  -4,  -6,   5,  -3,   1,  -1,   0, -14,  -2,  10, -13,   3,  -3,   6,  -3,  -7,  15, -14,   5,   2,   2,  -3,   0,   9,  -5,   5,   2,  -5,  -1,   1,   0,  -1,   4,  -2,  -3,   2,   1,  -1,   0,   2,   0,   1,   1,   0,   0,  -1,  -1,   1,   1,   0,   1,   1,   0,  -1,  -1 },
     {   -7, -10, -23,  -9,  28, -22,  21, -10,   7,  -4,   4,  -2,   4,  10,   9, -22,  24, -14,   2,  -4,   2,  -2,   1,  -1,   1,  11,  47,   0, -47,  29, -15,   6,  -5,   2,  -2,   1, -22, -46, -21,  11,  28,  -9,   3,   3,   0,   2,   0,   1,  -2,  18,  18,  -9,  -8,  -3,   4,  -3, -25,   9, -15,   5,   1,   4,  -2,   1,  21,  -7,   3,   3,  -3,  -1,  -1,   1,  -7,   4,  -2,  -2,   0,   1,   0,   0,   5,  -1,  -1,   2,  -2,   0,   0,  -1,   2,   0,  -1,   1,   0,   0,   0,   0 },
     {    2,  15,  14,  -5,  -6,   1,  -4,   1,  -2,   0,  -1,   0,   2,  24,  19, -13,  -5,  -1,   0,  -2,   0,  -1,   0,  -1,   2,  29,   1, -20,  12,  -8,   3,  -2,   0,  -1,   0,  -1,  -3,  23,  -7, -11,   4,  -2,   2,  -1,   1,  -1,   0,   0, -62,   3, -29,  36, -10,   5,   1,   1, -50,   9,  26, -17,   4,   2,  -1,   1,  44, -20, -12,   7,   1,  -2,   0,   0, -15,   3,   5,  -1,  -1,   0,   1,   0,  10,  -5,  -5,   2,  -3,  -2,   2,   0,   4,  -2,  -1,   0,   0,  -1,   1,   0 },
     {    6,  10,  33,   3, -12,   3,  -8,   2,  -3,   0,  -1,   0,   6,   9,  37, -15, -14,   4,  -1,  -2,  -1,   0,  -1,   0,   2,  -4,  23, -26,  12,  -6,   1,   0,   0,   0,   0,   0, -11, -39,  29, -29,  30, -15,   8,  -2,   1,   0,   1,   0,   1, -42,   8,  29, -20,  10,  -1,   0,  29,  36, -27,  -9,   5,  -1,  -2,   0, -23,  -6,   2,   8,  -1,   0,   1,   0,  12,   4,  -6,  -1,  -1,   1,  -1,   0,  -7,  -1,  -1,   2,   2,   2,  -3,   0,  -2,   0,  -2,   1,   1,   1,  -2,   0 },
     {    1,   7,  20,  26,  -7,  -4,  -8,   3,  -3,   1,  -2,   0,  -8, -10,  -3,   6, -16,  10,   4,  -2,   2,  -1,   1,   0, -14, -24, -27,   7,  -1,  18, -10,   8,  -1,   2,   0,   1, -13, -15, -13,  16,  20, -20,  11,  -3,   2,   0,   0,   0, -18,   3,  28,  -4, -20,  15,  -4,  -1, -60,  -8,  16, -12,  19,  -6,   0,   2, -34,  38, -12,   3,  -2,  -3,   4,  -3,  31, -20,   4,  -1,  -1,   2,  -3,   1,  -7,  10,   0,  -2,   9,  -3,   0,   0,  -1,   3,   0,  -1,   5,  -2,   0,  -1 },
     {  -10, -11, -26, -34,  13,   1,  12,  -3,   5,  -1,   3,   0,  -2,   5,   0,  -8,  27, -11,  -2,   2,   0,   1,  -1,   0,   4,  16,  32,  -3,   6, -24,  13,  -9,   2,  -2,   0,  -1,  -5,   5,  22,  -8, -27,  23, -12,   2,  -1,   0,   0,   0, -16, -24, -10,   6,  19,  -9,   0,   4, -37,  11,   7,   6, -11,   3,   3,  -3, -51,  24,  -3,  -2,   5,   1,  -1,   1,  42, -16,  -3,   5,  -4,   1,   0,   0, -14,   5,   3,  -2,  10,  -1,  -4,   3,  -3,   1,   1,   0,   5,  -1,  -1,   1 },
     {    3,  10,  23,   7, -37,   9,  10, -10,   3,  -2,   1,  -1,   8,  24,  39,  27, -18,  12, -30,  12,  -9,   3,  -4,   1,   2,  18,  12,  -5, -27,   5,   5,  -3,   0,   0,   0,   0,   5,  30,  -8, -25,  18,  -5,   0,  -3,   2,  -2,   1,  -1,   9,  31,  -5, -27,  19, -15,   4,  -1, -24,  -8, -28,  32, -12,   4,   3,  -2, -16,  -3,  18, -14,   4,   2,  -3,   3,  10,  -5, -11,   4,   2,  -2,   3,  -1,  -7,  -3,   3,  -1,   1,  -2,  -4,   1,  -1,  -2,   1,   1,   1,  -1,  -2,   0 },
     {    7,  12,  16,  22,  38,  -4, -43,  21, -12,   3,  -4,   2,  -5,  -4, -12, -27, -34,  -9,  53, -25,  12,  -2,   4,  -2,   8,  23,  28,   2,  11,  13, -30,   7,  -2,  -2,  -1,  -1,   4,  19,   8, -10,  -9,  -8,   4,   4,  -5,   2,  -2,   1,   3,   6, -12, -20,  18,  -3,   6,  -7, -10,   3, -12,  26,  -7,   0,  -3,   3,  -6,  -2,   8, -11,  -4,   5,   2,  -2,   6,  -4,  -4,   4,   4,  -2,  -3,   1,  -5,  -1,   1,  -1,   3,  -1,  -2,   2,  -2,  -1,   1,   0,   1,   0,  -1,   1 },
     {   -3,  -6,  -4, -24,  -4,  30, -34,  27, -12,   9,  -4,   4,   2,   2,  17,   3, -32,  19,  -1,  -2,  -1,   2,  -1,   1,   5,   2,  29,  55,  -4, -36,  15, -13,   5,  -4,   1,  -2, -11, -30, -33, -19,   5,  21,  -6,   6,   1,   1,   1,   1,  -1,  -5,  22,   1,   4,  -7,  -1,   1,   0, -20,   5,  -4,   4,  -2,   5,  -2,  19,  20, -13,   1,   0,  -1,  -1,  -2,  -9,  -4,   4,   1,   0,   0,   1,   0,   5,   3,  -2,  -3,  -4,   0,   1,   0,   1,   1,  -1,  -1,  -1,   0,   0,   0 },
     {    4,   4,  17,  19,  -3, -11,  10,  -8,   2,  -2,   1,  -2,   4,   1,  21,  16,  -9,  -4,  -7,   2,  -1,  -1,  -2,   0,   1,  -1,  18, -15, -20,  25, -12,   7,  -5,   2,  -1,   1,   3,  -6,  23,   8, -12,   0,   0,  -2,   0,   0,  -1,   0, -13, -51,   7, -25,  42, -15,   6,   1,  15, -22,  36,   0, -19,   7,   1,  -2,  22,  31, -43,   2,   7,  -3,   0,   1, -14,  -9,  13,   2,  -3,   1,   1,   0,   5,   7,  -8,  -3,  -4,  -2,   0,   1,   2,   2,  -2,  -2,  -2,   0,  -1,   0 },
     {   -2,  -4, -10, -27,   0,   6,   4,   0,   1,   0,   0,   0,  -1,  -4, -12, -29,  21,   0,  -2,   2,  -1,   0,   1,   1,   1,   4,   1, -23,  26, -13,  10,  -3,   1,   0,   0,   0,  11,  23,  34, -29,  25, -28,  10,  -7,   1,  -1,   0,  -1,   0,   0,  44, -20, -19,  19,  -7,   1,  -6, -41, -17,  29,   3,  -2,  -1,   3,  30,  27,  -9,  -3,  -6,   0,   0,  -1, -11, -15,   1,   8,  -2,   3,  -1,   1,   6,   5,  -1,   0,  -4,  -2,  -1,   4,   1,   1,   0,   2,  -2,  -1,  -1,   2 },
     {    4,   1,   0,   3,  -1,   1,   3,  -2,   1,   0,   1,  -1,   6,   1,  -1,   3,  -1,   0,  -2,   1,  -3,   0,  -1,   0,   7,   1,  -1,   4,  -1,   1,  -3,   1,  -1,   0,   0,   0,   5,   3,  -2,   2,  -5,   3,   1,  -1,   0,  -1,   0,   0,  10,   9,  -1,   1,  -1,  -2,  -1,   0,  17,  15,  -6,  -9,   3,  -4,   2,  -1,  71,   0, -10,   3,  -4,   1,  -2,  -1,  80, -27,  -3,   0,  -3,   2,  -2,   1, -50,  10,   5,  -3,  11,  -7,  -2,   1, -12,   1,   2,   0,   5,   0,  -1,   0 },
     {   -5,  -7, -10, -28, -40,   3,  10,   7,  -1,   5,  -1,   2,   2,   4,  10,   3,   0,  22,  -4, -10,   4,  -4,   1,  -1,   5,  14,  29,  35,   2,   6, -38,  19, -13,   5,  -5,   1,   3,  12,  26,  22, -27, -23,  28, -15,   2,  -2,   1,  -1,  -3,   3, -16, -23,  -4,  30, -18,   3,   4,   7,  -1, -14,  25, -26,   6,   4,  -8,  -9,  -6,  19, -16,   8,   5,  -6,  -4,   0,   5, -13,   8,  -1,  -4,   3,   1,  -1,  -7,   5,  -2,  -1,   2,  -4,  -1,   0,  -2,   2,  -2,   0,   0,  -1 },
     {    6,  10,  17,  37,  24, -26,  -1,  -7,   2,  -5,   1,  -3,   6,  12,  22,  39,  21, -30,  11, -14,   0,  -3,  -1,  -2,   3,   8,  17,  11,  -3, -22,   9,  -6,   2,  -2,   1,   0,   3,  15,  24, -21, -24,  26, -15,   1,  -2,   0,   0,  -1,   7,  29,  37,  -7, -21,   9, -10,   3,  -1,  -6, -10, -32,  23,  -8,   3,   0,  -5,  -3, -17,  18,  -5,   0,   2,  -1,  -7,   0,   1,  -8,  -2,   4,   0,   2,   6,  -2,  -5,   1,  -2,  -2,   1,  -2,  -1,   0,  -3,   0,   0,  -2,   0,  -1 },
     {    1,  -2,   7,  14,   0,  -9,   5,  -1,   1,  -1,   0,   0,   3,  -2,  10,  24,  13,  -6,   4, -10,   2,  -4,   1,  -2,   1,  -6,   9,  16,  -1,   0,  -5,   2,  -2,   1,  -1,   0,  -1, -15,   5,   4, -14,   4,   0,   0,   1,   1,   0,   0,  -9, -44, -11,  13, -16,  24, -10,   3, -13, -53,   5,  14,   9,  -8,   0,   5,  19, -10,  53, -39,  -1,   5,  -2,  -1,   8,  21, -38,  12,   6,  -6,   2,   1,  -6,   0,  14,  -5,   8,   4, -10,  -1,  -1,   1,   3,  -1,   3,   2,  -4,  -1 },
     {   -3,  -5,  -7, -10, -21, -37,   9,  36, -19,  14,  -6,   2,   5,   8,   8,  20,  28,  27,  14, -57,  26, -17,   7,  -4,  -5,  -9, -19, -23,   0, -18, -15,  31,  -7,   5,  -1,   1,  -3,  -5, -18, -20,   4,  14,  13,  -6,  -1,   4,  -1,   1,   5,   7,  13,  18,  20, -10,  -7,  -4,   0,   3,  -5,   6, -24,   5,   6,   4,   2,   4,  -1,  -3,   7,   6,  -6,  -4,   0,  -3,   5,   3,  -3,  -4,   0,   4,   3,   3,  -1,  -1,   4,  -1,   1,   2,  -2,   1,   0,   2,   2,   0,   1,   0 },
     {    0,   0,   1,  -1,   9, -12, -15,  40, -28,  22, -14,   6,   0,   0,   0,  -7,   3,  28, -14, -14,   6,  -6,   5,  -1,  -2,  -3,   0, -20, -52,  -1,  40, -20,  17, -10,   7,  -3,   7,  15,  29,  34,  18,  -3, -31,  15, -16,   6,  -5,   1,  -4,  -6, -15, -17,  -8,  13,   3,   2,   2,   0,   4,  -3,  14, -18,   5,  -3,   0,  -6,   1,   7,  -6,   5,   1,  -2,   4,  -1,  -4,  -3,   0,   2,  -1,   3,  -1,  -3,   6,   3,   0,   1,  -5,   0,  -2,   0,   1,   1,  -2,   1,   0,  -1 },
     {   -3,  -5,  -2, -17, -21,  10,   9, -10,   6,  -1,   0,  -2,  -3,  -6,  -3, -27, -26,  23,  11,  -5,   2,   2,   1,   2,   0,  -1,   3, -19,  -4,  19, -16,  10,  -8,   5,   0,   2,   2,   7,  21,  -7,  -5,  28, -27,  13,  -7,   1,  -4,   2,   8,   8,  43,  12,   7, -29,  24, -13,  -9, -34,  18, -28,  13,   2, -11,   6,  -5, -25,  13,  20,  -4,  -1,   2,   0,  14,  27, -18,  -9,   5,  -8,   2,  -4, -12,  -9,  -3,  13,   3,   5,  -1,  -5,   2,  -3,  -2,   2,  -3,   6,  -1,   2 },
   },
   {
     {  116, -43, -11,  -5,  -2,  -2,  -1,  -1,  -1,   0,   0,   0, -21,   5,   3,   1,   1,   0,   0,   0,   0,   0,   0,   0, -21,   7,   3,   1,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {  -13,   8,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,-114,  37,  12,   3,   3,   1,   2,   0,   1,   0,   0,   0,  32,  -7,  -3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  17,  -7,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0,   5,  -2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   1,  -1,   0,   0 },
     {   -7,  49, -24,   0,  -5,   0,  -2,   0,  -1,   0,  -1,   0, -18,   3,   3,   0,   1,   0,   1,   0,   0,   0,   0,   0,-103,  16,  17,   3,   4,   1,   2,   0,   1,   0,   1,   0,  40,  -6,  -5,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   0,   5,  -5,   0,   0,   0,   0,   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   2,  -1,   0,   0,   1,  -1,   0,   0,   2,  -1,   0,   0 },
     {   38,  89, -55,  -5, -11,  -1,  -4,  -1,  -2,   0,  -1,   0,   8,  -7,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,  39, -39,   9,   1,   1,   0,   0,   0,   0,   0,   0,   0, -20,   6,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -5,  -3,   3,   0,   1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,  -4,  -1,   2,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -2,  -1,   1,   0,  -1,   0,   0,   0,  -1,  -1,   1,   0,  -1,   0,   0,   0 },
     {   -7,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -21,  11,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0, -33,   8,   7,   0,   1,   0,   1,   0,   0,   0,   0,   0,-106,  23,  12,   2,   4,   1,   2,   0,   1,   0,   1,   0,  48,  -7,  -5,  -1,  -2,   0,  -1,   0,   3,  -2,  -1,   0,   0,   0,   0,   0,  11,  -2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   5,  -1,   0,   0,   2,  -1,   0,   0,   2,   0,   0,   0,   1,  -1,   0,   0 },
     {    3,   2,  -3,  -2,   1,   0,   0,   0,   0,   0,   0,   0,  30, 107, -52,  -7,  -8,  -4,  -3,  -2,  -2,  -1,  -1,  -1,  -2,  -4,  -3,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0, -29,  12,   2,   2,   1,   1,   1,   0,   0,   0,   0,  -3,  -1,   2,   0,   0,   0,   0,   0,  -3,  -8,   4,   1,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,  -2,  -4,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,  -1,  -2,   1,   0,   0,  -1,   0,   0,  -1,  -1,   1,   0 },
     {  -21, -41, -90,  64,  -2,  15,   0,   6,   1,   3,   0,   2,  -1,   2,   0,   4,  -2,   0,   0,   0,   0,   0,   0,   0,  10,  20,  28, -21,  -1,  -4,  -1,  -2,  -1,  -1,   0,  -1,  -1,  -3,  -3,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -3,   2,   5,  -3,   0,  -1,   0,   0,   2,   0,   1,  -1,   0,   0,   0,   0,   0,   1,   2,  -1,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0 },
     {   -6,   8,   4,  -1,  -2,   0,  -1,   0,   0,   0,   0,   0, -10,   2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0, -11,  23,  -9,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -38,   3,   8,   0,   2,   0,   1,   0,   1,   0,   0,   0,-103,  14,  13,   2,   4,   1,   2,   0,  51,  -7,  -6,  -1,  -2,   0,  -1,   0,  -6,  -1,   1,   0,   0,   0,   0,   0,  12,  -2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   5,  -1,   0,   0,   2,  -1,   0,   0,   2,   0,   0,   0 },
     {  -12, -30,   5,   4,   5,   1,   2,   1,   1,   0,   1,   0,  -4,   3,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   0, -26,-104,  44,  10,   5,   4,   2,   3,   1,   1,   1,   1,  -1,  15,   0,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0, -18,  30,  -7,  -2,   0,  -1,   0,  -1,   8,  -2,  -1,   0,   0,   0,   0,   0,   3,   7,  -3,  -1,   0,   0,   0,   0,   2,   2,  -1,   0,   0,   0,   0,   0,   2,   4,  -2,   0,   1,   1,   0,   0,   1,   2,  -1,   0,   0,   1,   0,   0 },
     {   -4,   2,   2,   4,  -3,   0,  -1,   0,   0,   0,   0,   0, -13,  -4, -20,  10,   1,   3,   1,   1,   0,   0,   0,   0,  -7,   6,  -3,  -1,   2,   0,   0,   0,   0,   0,   0,   0,  -8,  18,   0,  -3,  -1,  -1,   0,   0,   0,   0,   0,   0, -45,   2,   8,   0,   3,   0,   1,   0, -96,  14,  14,   1,   4,   0,   1,   0,  56,  -6,  -5,  -1,  -2,   0,  -1,   0,  -9,  -1,   2,   0,   0,   0,   0,   0,  11,  -1,  -1,   0,   1,  -1,   0,   0,   3,   0,   0,   0,   1,  -1,   0,   0 },
     {    0,   3,   1,   1,  -1,   0,  -1,   0,   0,   0,   0,   0,  18,  38,  97, -50,  -6, -11,  -3,  -5,  -2,  -2,  -1,  -2,  -2,  -2,   3, -10,   2,   0,   1,   0,   0,   0,   0,   0, -11,  -7, -33,  16,   2,   4,   1,   1,   1,   0,   0,   0, -10,   3,   2,   1,   0,   0,   0,   0, -21,  -2,  -4,   4,   1,   1,   0,   1,   9,  -1,  -2,   1,  -1,   0,   0,   0,  -2,  -2,  -4,   2,   0,   0,   0,   0,   2,  -1,  -1,   1,   0,  -1,  -2,   1,   1,   0,  -1,   1,   0,  -1,  -1,   1 },
     {   14,  17,  45,  84, -59,   4, -17,   0,  -7,   0,  -4,   0,   0,   1,  -2,   4,  -6,   2,   0,   0,   0,   0,   0,   0,  -6, -15, -27, -32,  24,   0,   7,   1,   3,   0,   1,   0,   1,   3,   4,   5,  -1,   0,  -1,   0,   0,   0,   0,   0,   4,   2,  -1,  -3,   2,  -1,   0,   0,   3,  -1,  -1,  -2,   2,   0,   0,   0,  -5,   1,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 },
     {   -1,   1,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -7, -27,   8,   0,   3,   1,   1,   1,   1,   0,   0,   0,  -6, -10,  -2,   4,  -1,   1,   0,   1,   0,   0,   0,   0, -20,-109,  39,  16,   3,   6,   1,   3,   1,   1,   1,   1,   1,  22,   2,  -3,  -1,  -1,  -1,  -1, -13,  21,  -4,  -3,   0,  -1,   0,  -1,   7,   1,  -1,   0,   0,   0,   0,   0,   2,   8,  -2,  -1,  -1,   0,   0,   0,   2,   2,  -1,   0,   1,   3,  -1,   0,   1,   1,  -1,   0,   1,   2,  -1,   0 },
     {    8,  14,  36,  -4,  -7,  -6,  -1,  -3,  -1,  -1,   0,  -1,   2,   1,  -4,   8,  -6,   1,  -2,   0,   0,   0,   0,   0,  17,  28,  98, -45, -10,  -6,  -4,  -3,  -2,  -1,  -1,  -1,   0, -10,   0,  -7,   2,   2,   1,   0,   0,   0,   0,   0,  -6,  -6, -34,  11,   3,   2,   2,   1,   4,   3,   1,  -1,   0,  -1,   0,   0,   8,  -4,  -9,   3,   0,   1,   0,   0,  -2,   1,  -2,   1,   0,   0,   0,   0,   0,  -2,  -4,   2,  -1,   0,  -1,   0,   0,  -1,  -2,   1,  -1,   0,  -1,   0 },
     {    0, -10,   2,   2,   3,   0,   0,  -1,   0,   0,   0,   0,   1,  -3,  -1,  -3,   2,  -1,   0,   0,   0,   0,   0,   0,  -2, -24,   3,  -3,   4,   0,   1,   1,   1,   0,   0,   1,   1, -13,  -1,   9,  -3,   2,   0,   1,   0,   0,   0,   0,  -8,-101,  32,  15,   0,   6,   0,   3,  29,  22,  -7,  -4,  -2,  -1,  -1,  -1,  43,  10, -11,  -2,  -2,  -1,  -1,   0, -18,   3,   2,   0,   1,   0,   0,   0,   2,   5,  -1,  -1,  -3,   3,   0,   0,   0,   3,  -1,   0,  -2,   1,   0,   0 },
     {   -6,  -5,   4,  -3,  -2,   1,   0,   1,   0,   0,   0,   0,  -8,  -2,  -5,  -9,   6,   0,   2,   1,   0,   0,   0,   0,  -8,  -7,  13,  -4,   5,  -2,   0,   1,   0,   0,   0,   0, -11,  -1,   2,   9,  -3,   1,  -1,   0,   0,   0,   0,   0, -16, -46,  11,   7,   1,   3,   0,   2, -47,  14,   9,  -2,   2,   0,   1,   0, -93,  21,   9,   0,   3,   0,   1,   0,  35,  -3,  -4,   0,  -2,   0,   0,   0,  -4,   3,   0,   0,   8,   0,  -1,   0,   1,   1,  -1,   0,   3,   0,  -1,   0 },
     {    0,   0,  -3,  -3,   4,  -1,   0,   0,   1,   0,   0,   0,  -9, -15, -37, -92,  53,   0,  13,   2,   4,   1,   2,   1,   3,   1,   2, -22,  16,  -3,   2,  -1,   0,   0,   0,   0,   6,  11,  15,  35, -19,  -2,  -4,  -1,  -2,   0,  -1,   0,   2,  10,  -8,   3,  -3,   0,   0,   0,   6,  -3,   1,   7,  -3,   1,  -1,   0,   9,  -2,   1,   3,  -2,   0,   0,   0,  -2,   1,   1,   4,  -3,   0,   0,   0,   1,  -1,   1,   2,   0,   0,   1,   2,  -1,   0,   0,   1,   0,   0,   1,   1 },
     {    9,  15,  11,  45,  84, -57,  -3, -13,   1,  -6,   0,  -3,  -1,   0,   0,   2,   2,  -6,   2,  -1,   1,   0,   1,   0,  -5,  -6, -12, -26, -40,  27,   2,   6,   1,   3,   0,   1,   0,   3,   4,   2,   5,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,  -1,  -3,  -1,   0,  -1,  -2,   2,   0,   0,  -5,   0,   1,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    0,   1,   2,  -1,   2,  -2,   2,  -1,   0,   0,   0,   0,  -6, -10, -32,   9,   0,   4,   0,   2,   0,   1,   1,   1,  -2,  -3,  -5,  -4,   3,  -1,  -1,   1,   0,   1,   0,   0, -15, -25,-103,  43,  18,   1,   5,   2,   3,   1,   1,   1,   0,  10,  -4,   7,   0,  -1,   0,  -1,   8,   6,  29, -10,  -5,  -1,  -2,   0,  -2,  -1,   1,   0,  -1,   0,   0,   0,  -3,   3,  10,  -3,  -1,   0,   0,   0,   3,   0,   2,  -1,   1,   1,   3,  -1,   1,   0,   2,   0,  -1,   0,   2,  -1 },
     {   -5,  -1,  -3, -13,   2,   0,   2,   2,   0,   0,  -1,   0,  -6,   1,   3,   6,   0,  -1,  -2,  -1,   0,   0,   0,   0,  -8,  -1,  -4, -26,  13,  -1,   4,   0,   1,   0,   1,   0,  -4,   1,   7,  -4,   1,  -1,   1,   0,   0,   0,   0,   0,  -7,   8,   8,  11,  -4,  -2,  -1,  -1,  -4,   2,  -4,  -2,   1,   1,   0,   0, -39,   5,   7,   3,   0,   0,   1,   0,-108,  10,  13,   1,   4,   0,   2,   0,  32,  -2,  -3,   1, -10,   0,   0,   1,   8,  -2,  -1,   1,  -3,  -1,   1,   0 },
     {    3,   5,  13,  35, -15,  -6,  -5,   1,  -2,   0,  -1,   0,  -4,  -2,  -5, -26,  -1,   8,   3,   1,   0,   1,   0,   0,   5,  12,  26,  82, -45,   2,  -8,  -3,  -3,  -1,  -1,   0,  -3,  -1,  -4,   7,  -6,  -1,  -1,   0,   0,   0,   0,   0,  -5, -12, -10, -43,  18,   3,   5,   1,   0,  -1,   8,   8,  -1,  -1,  -1,   0, -10,   3,   0,  -8,   3,   0,   1,   0, -34,   4,   4,   0,   3,  -1,   1,   0,   6,  -1,  -2,  -3,   0,   1,   0,  -1,   2,  -1,   0,  -1,   0,   0,   0,   0 },
     {   -2,  -3,  -5, -11,  -6,   8,   5,  -1,   2,   0,   1,   0,  -8, -12, -13, -43, -86,  54,   6,   8,   1,   4,   1,   2,  -2,  -4,  -6, -19, -17,  21,  -1,   3,   0,   2,   0,   1,   4,   3,  11,   6,  39, -18,  -5,  -3,  -1,  -1,   0,   0,   0,  -2,   3,   4,   4,  -5,  -1,  -1,  -1,  -8,   1,   7,   4,  -3,   1,   0,   1,   4,   0,   2,   2,  -2,   0,   0,   2,   2,   0,   2,   3,  -3,   0,   0,  -2,   1,   1,   2,   2,   0,   0,   1,   1,   0,   0,   1,   0,   0,   0,   1 },
     {    0,  -2,   1,   2,   1,   2,   0,   0,  -1,   1,   0,   0,  -2, -13,   2,   2,  11,  -4,  -1,   0,   0,   1,   0,  -1,  -1,  -4,  -2,  -3,   4,  -2,  -1,   1,   0,   1,   0,   0,  -4, -22,   0, -16,   5,   1,   3,   1,   2,   1,   1,   0,  -4, -20,  -5,   7,  -3,   1,   0,   2, -10,-109,  35,  23,  -4,   8,  -1,   4,   6,  24,  -2,  -2,  -1,  -1,   1,  -1,  -1,   9,  -3,   0,  -2,  -1,   0,  -1,   6,   3,   1,   0,   1,   8,  -4,  -1,   1,   0,   0,   0,  -1,   3,  -1,   0 },
     {    3,   4,  14,  -5,   3,   4,  -5,   1,   0,  -2,   0,   0,   1,  -2,  -6,  -1,   1,   2,   1,   1,   0,   0,   0,   0,   5,   6,  26, -11,   1, -10,   3,  -1,  -1,  -1,   0,   0,   2,   2,  -6,   1,   1,  -1,   1,  -1,  -1,   0,   0,   0,  14,  20, 100, -48, -11,  -1,  -4,  -1,   0, -13,  -4,  -1,  -2,   2,   0,   0,  -5,  -8, -32,  13,   3,   1,   1,   1,  11,   3,   2,   0,  -2,   0,   0,   0,  17,  -4,  -6,   3,  -7,   1,  -2,   0,   3,  -1,  -3,   1,  -4,   0,   1,   0 },
     {    0,   4,   7,   0,  21,  37, -32,   7,  -8,   2,  -3,   2,  -4,   0,   0,   1,   3,  -3,  -4,   1,   0,   1,  -1,   1,  -4,  -2,   1,  -8, -10, -20,  14,  -1,   3,  -1,   1,   0,  -3,   2,   2,   3,  10,   3,  -2,   0,   0,  -1,   1,   0,  -1,   3,   9,  -7,  -1,   1,  -1,   0,  -6,  -2,   4,  -1,  -2,   1,   2,   0,  -1,   1,  -8,   4,   1,  -1,   1,   0, -31,   4,   5,   0,   1,  -1,   1,   0, -95,   6,  10,   0,  39,  -4,  -4,   1, -17,   2,   2,   1,  11,   0,  -1,  -1 },
     {    6,   7,   8,   6,  40,  64, -54,  13, -12,   3,  -4,   1,   1,  -1,  -1,  -2,  -3,  -1,  -2,   4,  -1,   1,  -1,   0,  -1,  -4,  -7,   0, -20, -39,  29,  -5,   8,   0,   1,   0,   3,   1,   2,   3,  12,   3,  -1,  -2,   0,  -1,   0,  -1,  -1,  -4, -17,   3,   2,   4,   2,   0,   3,   6,   0,   4,  -3,  -3,   2,  -1,   1,   3,   6,  -3,   1,  -1,   0,   0,  14,  -2,  -5,  -1,  -2,   0,   0,   0,  49,  -4,  -4,   0, -25,   0,   4,   0,  12,  -1,   0,   0,  -9,   0,   1,   1 },
     {    0,   0,   0,  -5,   1,   0,  -6,   3,  -3,   2,  -2,   1,   3,   3,   9,  28, -20,  10,  -8,   2,  -1,   1,  -1,  -1,   2,   4,   6,   5, -10,   0,   1,   2,   0,  -2,   0,  -1,   6,  13,  23,  91, -48,  -6,  -3,  -5,  -2,  -2,  -2,  -1,  -1,   2,   0,   9, -15,   5,   1,  -2,  -3, -30,  -8, -38,  18,   4,   2,   3,   4,   4,   5,   0,   5,  -3,   0,  -1,  -1,  -3,   0,  -6,   3,   1,   2,  -1,   2,   1,  -1,  -3,   1,   3,  -4,  -3,  -3,  -1,   0,  -1,   0,   0,  -1,  -2 },
     {    2,   4,   3,  15,  35, -24,   6,  -9,   3,  -3,   1,  -2,  -1,   0,   2,  -3, -20,  19,  -1,   0,  -2,   2,   0,   1,   5,   7,   8,  24,  75, -47,   0,  -5,  -1,  -2,  -1,  -1,   0,  -1,   4,   2,  16, -21,   5,  -2,   0,  -1,   1,  -1,  -6,  -8, -18, -26, -43,  27,   3,   6,   0,   2,   0,   3,   9,  -2,   0,   0,   0,  -6,   4,   1,  -4,   5,  -1,   1,  -3,   3,  -1,  -1,  -1,   1,   0,   0,  -5,   3,   1,  -3,  -2,   0,   1,   1,   0,   0,   0,   1,   0,   1,   0,   0 },
     {    3,   0,   0,   1,   2,   3,  -5,   1,   0,  -1,   0,   0,   4,   0,   0,   0,  -1,   0,   2,   1,   0,   0,   0,  -1,   4,  -1,  -1,   0,   3,  -1,  -1,   2,  -1,   2,  -1,   0,   3,  -1,   0,  -3,   3,   3,  -1,   2,  -2,   1,  -1,   0,   3,  -2,  -2,   0,  -3,   0,   1,   0,   4,   1,  -2,  -1,   3,  -3,   0,  -1,   2, -10,   2,   1,   0,  -1,   2,   0,   2,   5,   1,   0,  -1,   1,   0,   0,  54,   3, -10,  -1, 100, -12,  -8,  -1, -52,   2,   6,   0,  11,  -2,   0,   0 },
     {   -2,  -7,   0,  -3,  -2,   1,  -7,   3,   0,  -1,   1,   1,  -1,  -3,   0,   4,   1,  -1,  -3,   0,  -1,   0,  -1,   1,  -2, -10,   1,  -3,  -2,   6,   3,   0,  -1,   1,   0,   0,  -2,  -4,  -1,   7,   1,   1,  -6,   3,  -1,  -1,   0,   0,  -5, -20,  -3,  -9,  11,  -4,   2,   1,  -6, -19, -14,  15,  -2,   5,   1,   1, -10,-104,  34,  27,  -6,   9,  -2,   4,   3,  29,   5, -10,   1,  -4,  -1,  -1,  -4,  13,  -3,  -3,  -9,  -1,   3,   1,   2,   9,  -3,  -2,   0,   1,   1,   0 },
     {    0,   0,  -1,  -1,   8,   1, -25,  13,  -6,   7,   0,   4,   4,   7,   8,   5,  34,  72, -66,   9, -16,   5,  -3,   3,   0,   1,   0,  -3,  11,  26, -15,  -2,  -2,   1,   1,  -3,  -3,  -5,  -7, -11, -21, -32,  27,  -1,   6,  -2,   2,  -1,   0,  -1,  -2,  -1,  10, -12,   7,  -4,   1,   4,  -1,   8,  -4,  -5,   7,  -3,  -2,   1, -12,  -1,  -2,   0,   3,   1,   2,   1,   0,   0,   0,  -5,   0,   1,   2,   0,   3,  -1,   0,   0,   1,  -2,  -3,   0,   1,   1,  -2,  -3,   0,   0 },
     {    0,  -2,  -1,  -1,  -6,  -3,   4,   6,   0,   1,   1,   2,   4,   5,  18,   2,  10,   5,   8,  -3,   0,  -3,   0,   0,   0,  -3,  -4,   0,  -8,  -6,   2,  -1,   0,  -1,   1,   0,   5,   7,  27,   2,  14, -25,   1,  -5,  -1,  -3,   0,   1,   1,   0,   1,   6,  -1,   8,  -2,  -3,  13,  18,  98, -31, -21,   8,  -9,   0,  -3, -26, -18,  19,   6,  -4,   1,   1,  -2,  -9, -12,   5,   2,   2,   2,   2,   4,  12,  -1,  -5,  -5,  -8,  -8,   4,   0,   2,   0,  -1,  -1,   3,  -4,  -1 },
   }
 },
 {  //19
   {
     {  118, -48,  -4,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0,   0,   9,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -8,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    6,  -5,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,-117,  32,   6,   4,   2,   2,   1,   1,   1,   1,   0,   0, -31,   4,   3,   1,   1,   0,   0,   0,   0,   0,   0,   0,  14,  -6,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  10,  -3,   0,   0,   0,   0,   0,   0,   6,  -1,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,   2,  -1,   0,   0 },
     {   -8,   7,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  21,  -3,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,-107,  27,   6,   3,   2,   1,   1,   1,   1,   0,   0,   0, -58,   8,   4,   1,   1,   1,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   8,  -4,   0,   0,   0,   0,   0,   0,   7,  -2,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   1,   0,   0,   0,   2,  -1,   0,   0,   1,   0,   0,   0 },
     {    2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -27,   9,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,  41,  -4,  -4,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   0, -83,  23,   4,   2,   1,   1,   0,   0,   0,   0,   0,   0, -77,   7,   6,   2,   2,   1,   1,   0,  -5,  -7,   1,   0,   0,   0,   0,   0,  14,  -5,   0,   0,   0,   0,   0,   0,   7,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   1,  -1,   0,   0,   2,  -1,   0,   0 },
     {    1,  -7,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -6,  -3,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,  33,  -6,  -3,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0, -55,   6,   4,   2,   1,   1,   1,   0,   0,   0,   0,   0,  75, -23,  -3,  -2,  -1,  -1,   0,   0,  75,  -8,  -7,  -1,  -2,  -1,  -1,   0,  10,   6,  -2,   0,  -1,   0,   0,   0,  -5,   5,   0,   0,   0,   0,   0,   0,  -6,   2,   0,   0,  -1,   1,   0,   0,  -2,   1,   0,   0,   0,   1,   0,   0 },
     {    2,   5,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -9,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  12,   2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -43,   8,   3,   1,   1,   1,   0,   0,   0,   0,   0,   0,  47,  -3,  -4,  -1,  -1,  -1,  -1,   0, -71,  24,   3,   2,   1,   1,   0,   0, -78,   3,   7,   1,   2,   1,   1,   0,   4,  -8,   1,  -1,   0,   0,   0,   0,   6,  -2,   0,   0,   2,   0,   0,   0,  -2,  -1,   1,   0,   2,  -1,   0,   0 },
     {    5,  15,  -4,   1,  -2,   0,  -1,  -1,   0,   0,   0,   0,   0,   9,  -2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -8,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   9,   7,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -35,   6,   3,   1,   1,   0,   0,   0,  51,   0,  -4,  -1,  -1,  -1,  -1,   0, -76,  27,   3,   2,   1,   1,   0,   0, -71,   2,   7,   1,   2,   0,   1,   0, -12,  -6,   2,   0,  -1,  -5,   1,   0,   4,  -2,   0,   0,   0,  -1,   0,   0 },
     {  -42,-105,  46,   0,  11,   2,   5,   1,   3,   0,   2,   1,  -3, -23,   7,   1,   2,   1,   1,   0,   0,   0,   0,   0,  -4,   3,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   3,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   0, -10,   6,  -1,   0,   0,   0,   0,   0,   6,   2,  -1,  -1,   0,   0,   0,   0, -18,   4,   1,   0,   0,   1,   0,   0,  -1,  -4,   1,   0,   0,   0,   0,  -1,  12,  -2,  -1,   0,   5,   0,  -1,   0,   4,   1,  -1,   0,   1,   0,   0,   0 },
     {   -6, -14,   7,   0,   2,   0,   0,   0,   0,   0,   0,   0,  -3,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   9,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -11,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  17,   6,  -3,   0,  -1,   0,   0,   0, -42,   8,   2,   1,   1,   0,   0,   0,  45,   1,  -4,  -1,  -1,   0,  -1,   0, -77,  26,   4,   2,   1,   0,   1,   0, -70,  -3,   7,   0,   3,  -8,   1,   0,   3,  -3,   0,   0,   2,  -1,   0,   0 },
     {    4,   7,  -3,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -2,  -8,   2,   1,   0,   0,   1,   0,   0,   0,   0,   0,   4,  -6,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,  -9,   2,   1,   1,   0,   0,   0,   0,   0,   0,   0,   8,  -3,  -1,   1,   0,   0,   0,   0, -11,  -7,   3,   0,   1,   0,   0,   0,  22,  -4,  -2,   0,  -1,   0,   0,   0, -50,  -6,   5,   0,   2,   0,   1,   0,  77, -31,  -4,  -2,  75,   1,  -9,   0,   9,   8,  -1,   0,   0,   4,   0,   0 },
     {    7,  17, -11,   1,  -2,   1,  -1,   1,  -1,   0,  -1,   0, -28,-106,  32,   5,   8,   3,   3,   2,   2,   1,   1,   1, -12, -43,   8,   3,   2,   1,   1,   1,   1,   0,   1,   0,  -3,   6,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -3,  10,  -2,   0,   0,   0,   0,   0,  -1,   8,  -2,   0,  -1,   0,   0,   0,  -1,   3,  -1,   0,   0,   0,   0,   0,  -6,   5,   0,   0,   0,   0,   0,   0,  -6,   2,   0,   0, -12,   5,   0,   0,  -6,   0,   1,  -1,   2,   2,  -1,   0 },
     {   -2,  -2,   2,  -1,   1,   0,   1,   0,   0,   0,   0,   0,  -1,   7,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   2,  13,  -4,  -1,  -2,   0,  -1,  -1,   0,   0,   0,   0,  -2,  10,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   4,   5,  -1,   0,  -1,   0,   0,   0,  -9,   1,   1,   1,   0,   0,   0,   0,  16,   7,  -2,   1,  -1,   1,   0,   0, -35,   4,   2,   1,   1,   0,   0,   0,  51,   3,  -5,   0, -68,  32,   4,   2, -76,  -1,   9,   0,  -3, -10,   1,  -1 },
     {   -2,  -7,   9,  -4,   2,  -1,   1,  -1,   0,   0,   0,   0,  11,  33,  -3,  -2,  -3,  -2,  -3,   0,  -1,   0,   0,  -1, -14, -90,  26,   3,   7,   1,   3,   1,   2,   1,   1,   1, -17, -60,  15,   3,   4,   2,   0,   1,   2,   1,   1,   1,  -4,  -8,  -1,   1,   0,   1,   1,   0,  -6,  -5,  -1,   1,  -1,   0,   1,   0,   0,   1,  -3,  -1,  -1,   0,  -1,   0, -12,   1,   1,  -1,   1,   0,   0,   0,   7,   4,  -2,   0, -27,   2,   2,   1,  14,  -7,  -1,   0,  19,   1,  -3,   0 },
     {    2,   3,  -5,   1,  -1,   1,   0,  -1,   0,  -1,   0,   0,  -4, -12,   0,   1,   3,  -1,   1,   0,   0,   0,   0,   0,   9,  33,  -8,  -1,  -2,  -2,   0,  -2,   0,   0,   0,  -1,   2,   6,   1,   0,   0,  -1,   0,   0,  -1,  -1,   0,   0,   6,   0,  -1,   0,   1,  -1,   0,   0,  -5,  -6,   3,  -1,   0,  -1,   0,   0,  10,  -3,   0,   0,   0,  -1,   0,   0, -16,  -9,   3,   0,   1,  -1,   1,   0,  32,   0,  -4,  -1, -54,  -7,   7,   0,  79, -33,  -6,  -1,  52,   8,  -8,   1 },
     {   13,  29,  76, -25,  -6,  -7,  -3,  -4,  -3,   0,  -1,   0,  -3, -17,  25,  -6,   1,   0,   0,   0,   0,   1,   0,   0,   7,  32,  -5,  -2,  -2,  -1,  -1,   1,  -1,   0,  -1,  -1,  -1, -43,  13,   1,   3,   0,   0,   1,   0,  -1,   0,   0, -13, -63,  10,   0,   3,   3,   2,   2,  -7, -15,  -4,   1,   0,   2,   1,   1,  -6,  -4,  -2,   1,  -1,   1,   1,   0,  -3,   4,  -1,   0,   0,   0,   0,   0,  -4,   5,  -2,   1,  -3,   2,  -1,   0,  -3,   0,   0,   1, -12,   0,   1,   0 },
     {   11,  24,  70, -28,   0,  -6,  -2,  -2,  -1,  -1,   0,  -1,   6,  20,  16,  -7,  -2,  -2,  -1,   0,   0,   0,  -1,   0,  -8, -33,   6,   3,   2,   1,   0,   0,  -1,   1,   1,   0,   9,  67, -16,  -3,  -4,  -3,  -3,  -2,  -1,  -1,   0,  -1,  13,  34,  -9,  -2,  -2,  -1,   1,  -1,   4,  -7,   3,   0,   1,   1,   1,  -1,   4, -16,   3,   1,   1,   0,   0,   0,   2, -10,   2,   0,   1,   2,   0,   0,  -1,  -8,   0,   1,   7,  -5,   0,   1,  -3,  -5,   2,   1,  31,  -9,  -2,   0 },
     {    2,   6,   5,  -5,  -1,   0,   1,   0,   0,  -1,   1,  -1,   0,  -7,  -2,   2,   0,   0,  -1,   1,  -1,   0,   1,   0,   5,  26,  -5,  -4,  -2,   0,  -2,   0,   0,   0,   0,   0,  -7, -49,   9,   3,   4,   2,   0,   1,   0,   0,   0,   0,  -5,  33, -14,  -1,  -2,   0,   0,   0,  18,  51,  -9,  -5,  -2,  -2,  -1,  -1,  -1, -12,   4,   0,   1,   0,   0,   0,  11,   0,   3,  -1,  -1,   0,   1,   0, -12, -14,   3,  -1,  22,   5,  -2,  -1, -35, -12,   6,   0,  76, -14,  -7,  -1 },
     {    6,  14,  29, -12,   1,  -5,  -1,  -2,  -1,  -1,   0,  -1,  -1,   3,   1,   1,  -2,   0,   1,   1,  -1,   0,   0,   1,   3,   8,  -7,   0,   0,   1,   0,   0,   0,  -1,   0,   0, -11, -25,   7,  -1,   2,   0,   1,   1,   1,   0,   0,   0,   5,  62, -19,  -1,  -5,   0,  -2,   0,  11,  54, -12,  -2,  -4,  -2,   0,   0,  12,  14,   2,   0,  -1,  -1,   0,   0,  -1,  -7,   4,   0,   1,   0,   0,   0,  14,   3,   0,  -1, -17,  -1,   2,   1,  31,   9,  -5,   0, -67,  14,   6,   1 },
     {   -1,  -1,  -5,  -1,  -1,   1,  -2,   0,   0,   2,   0,   1,  -2,  -6,   9,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   6,  17,   7,  -1,  -1,  -2,   0,  -1,  -1,   0,   0,  -1,  -6, -30,  15,   1,   3,   1,   0,   3,   0,   0,   1,   0,  10,  53,  -7,   0,  -3,  -3,  -1,  -1,   3, -71,  24,   1,   1,   1,   0,   1, -20, -69,  10,   5,   1,   1,   0,   1,  -8,  -6,  -3,   2,  -2,   1,   0,   0, -10,  -9,   0,   2,  -8,   6,  -3,   0,  -4,   7,   0,  -1, -14,   6,   1,   0 },
     {   -2,  -7, -19,  19,  -9,   7,   0,   4,   1,   2,   0,   1,  10,  24, 100, -31,  -4, -11,  -3,  -4,  -3,  -3,  -1,  -2,   6,  16,  51, -13,  -7,  -3,  -2,  -3,  -3,   0,   0,  -1,  -1,   0,  -9,  -5,  -2,   6,  -1,   2,   0,   0,   0,   0,  -1,   2,  -8,  -2,  -1,   3,   1,   1,   1,   9,  -5,  -1,  -1,   3,   0,   0,   4,  10,   1,  -2,   0,   1,   1,   1,   0,  -4,  -3,   3,   0,  -1,   1,   0,   1,   2,  -2,   1,   1,   1,  -3,  -1,   2,   2,  -2,   0,   0,   1,  -2,   1 },
     {    2,   5,  10,  -1,  -2,   2,   0,  -3,   0,  -1,   0,   0,  -1,  -1,  -2,  -1,  -2,   0,   1,  -1,   1,   0,   0,   1,   2,   8,   2,   0,  -3,   1,   0,   0,   0,  -1,   0,   0,  -4, -12,  -6,   3,  -2,   1,  -3,   2,  -1,   0,   0,   0,   6,  35, -18,   0,   0,   1,   0,  -2, -10, -47,   0,   5,   3,   3,   1,   1,  -6,  66, -24,  -1,  -3,  -3,  -3,   1,  21,  75,  -8,  -6,  -3,  -3,   0,  -3,  11,   8,   5,  -2,  13,   2,   6,  -1,   2, -14,   3,   1,  10,  -3,   1,  -1 },
     {    6,   6,  29,  63, -27,   5,  -6,   0,  -5,   0,  -1,   0,  -2,  -8, -30,  40, -17,  11,   0,  -3,   0,   2,   1,   1,   5,   8,  66, -23, -11,  -7,  -1,  -2,  -2,   0,  -1,  -1,   4,  14,  42, -11,  -8, -12,   0,  -2,   0,   0,   0,  -1,   0,   0,  -5,  -3,   2,  -2,   0,   1,   0,   4,  -1,  -1,  -1,  -1,   0,   2,   2,   0,  -3,  -1,   1,  -2,   1,   2,   0,  -2,   1,  -2,   2,   1,   2,   1,   0,   1,  -5,   1,  -1,  -1,   0,  -1,  -2,  -1,   1,   1,   0,   1,   0,   0 },
     {   -5,  -2, -18, -77,  36,  -1,   9,   4,   4,   0,   2,  -1,  -4,  -9, -26, -27,   6,   7,   4,   2,   4,   1,   3,   0,   5,  13,  64, -18, -12,  -3,  -2,  -4,  -1,   0,   0,   0,   4,  11,  38,  -5,  -6,  -4,   0,  -1,  -2,   1,   0,  -1,   0,  -2,   8,   3,  -4,   2,   0,  -1,   1,   6,  14,  -2,  -4,  -2,   1,  -1,   1,  11,   0,   4,  -1,  -1,   0,   1,   1,   0,  -9,   3,   1,   0,   1,  -1,  -2,  -4,   0,   2,   2,   6,   1,  -1,   3,   3,   1,  -1,  -1,  -2,   0,   0 },
     {    1,   1,   3,   1,  -4,   2,  -1,   2,  -2,   2,   0,  -1,  -1,  -5,  -2,   0,   0,  -2,   1,   5,  -1,   1,  -1,   0,   2,   3,  -1,  -3,   4,  -2,   4,   1,   1,   0,  -1,  -1,  -4, -13,  -4,   2,  -3,   1,   3,   1,   0,   0,   0,   1,   4,   9,   3,  -8,   7,  -2,  -1,   0,  -8, -32,   7,   0,   2,   6,  -2,  -3,  11,  60,   4,  -8,  -6,  -2,  -2,  -2,  11, -69,  30,  -1,   4,   0,  -1,   1, -22, -65,   9,   7,  -9,   3,  -4,   1, -10,  -6,  -4,   0,  -1,  -2,  -1,   2 },
     {    0,   1,   2,  -7,  12,  -4,   3,  -4,  -2,   1,  -1,   1,  -2,  -4, -23,  20,   3,  -3,   3,   3,  -1,  -1,   0,   0,   4,   9,  44,  -2,  -2,  -4,   0,   0,   3,  -1,  -1,  -3,  -6,  -9, -84,  34,   1,   3,   0,   3,   3,   1,   1,   0,  -6, -18, -53,   6,   8,   9,  -1,   3,   0,   2,  -7, -12,   2,  -3,   4,  -1,  -3, -17,  -4,  -1,   0,   4,   0,   1,  -1,   5,   0,  -2,   0,   3,   2,  -2,  -1, -19,   0,   5,  -9,   1,  -1,  -2,   2,   9,  -1,  -1,  -5,   5,   3,  -1 },
     {    1,   0,  10,  22, -30,  15,  -9,   2,  -3,   2,   0,   0,  -4,  -5, -30, -75,  55,  -3,  11,   2,   4,   1,  -1,   2,  -2,  -3,  -4, -55,  24,   1,   5,   3,   6,   0,   4,  -1,  -1,  -1,  -9,  -2,   1,  -1,  -2,   1,   2,   1,   1,  -1,   0,   2,  -9,   2,   5,  -4,   0,   2,   0,   0,  -9,  12,  -6,  -1,  -2,  -1,  -1,  -5, -12,  11,   0,   2,  -4,   0,  -2,  -3,  -1,   1,  -2,   3,   2,   0,  -1,   4,   1,   1,   1,   9,  -2,  -1,   3,   9,  -4,  -1,   2,   3,   3,   1 },
     {    0,   2,   2,  -6,  -7,   6,   2,   1,   3,   0,   0,  -1,  -1,   0,  -4,  10,  -7,  -1,  -3,   2,   0,  -1,   0,   0,   0,   3,   0,   3, -12,  -4,  -3,   2,  -1,   0,   1,   0,  -2,  -4, -14,  17,  -3,   1,  -2,   0,   2,   1,   0,  -3,   2,  10,  -1,  -1,   3,   0,  -9,   0,  -3, -11,   9,   0,  -3,  -6,   0,   0,   3,  19, -12,  -5,   4,   3,   1,  -2, -10, -44,   0,   7,   3,   7,   1,  -1, -12,  70, -26,  -5,  23,  73,  -9,  -9,  11,   7,   4,  -2,  14,  10,   6,  -1 },
     {   -6, -10, -10, -37, -80,  27,   8,   5,  -2,   3,   2,   1,  -3,  -4,  -5, -15, -52,  14,   8,   1,   1,   5,  -1,  -1,   1,   0,   2,  16,   4,  -5,  -4,  10,   8,  -2,   0,  -4,   1,   1,   0,  10,  18, -12,  -6,   2,   5,  -2,  -3,  -2,   0,  -1, -10,  15,  28, -16,  -7,  -2,   0,   2,  -9,   9,  22,  -8,  -2,  -5,   1,  -2,  -9,   5,   9,  -1,   5,  -5,  -3,  -6, -15,  -1,   0,   6,   0,  -1,   3,  -1,   9,   1,  -3, -17,   3,  -4,  -3,  -2,   3,  -6,  -2,  -4,   2,  -1 },
     {    0,   0,  -6,   3,  19,  -7,  -7,  -1,  -3,  -1,   3,   1,   0,  -1,   3,  -4,  -2,  -4,   4,  -4,   2,   0,  -3,  -2,  -2,  -3, -17,   6, -10,   5,  -3,   0,   3,   2,   3,  -2,   3,   9,  36, -17,   7,   0,  -2,  -8,   2,  -1,   1,  -1,  -6, -15, -86,  50, -10,   4,   5,  -3,  -4, -15, -27,  24,  -4,  12,  -3,   1,   2,   8,  18,   2,  -1,  -9,   2,  -4,  -3, -19, -14,   3,   2,  -1,   2,  -4,  -3,   6,  -1,  -9,   2,   0,   3,  -3,  -2,   8,   0,  -3,   4,  -2,  -4,  -1 },
     {   -3,  -3,  -8, -24, -17, -18,  19,  -7,  14,   5,   1,  -6,   1,   1,   7,  45, -13, -17,  -3,  -2,  -8,  -8,  -4,   6,  -3,  -4, -13, -74,  16,  10,   8,   3,   4,  -1,   0,   1,  -2,  -1, -12, -45,  20,  25,  -2,  10,  -2,  -9,  -2,   6,  -1,  -3, -10, -10,  -4,   0,  12,   1,   0,  -2,  18,   6,   9,  -6,  -3,  -8,   1,   6,   7,  12, -14,   3, -11,   4,  -2,   0, -12,   8,   0,   0,  -3,   9,   1,   6,  -1,   8,  -2,  -2,  -1,  -2,   2,  11,  -4,   4,   3,  -8,   1,   5 },
     {   -1,  -1,   1,   1, -17, -12,   7,  -5,   4,  12,  -2,   3,   1,   5,   2,   2,   8,  -2,   4,   5,  -1,   2,   0,  -5,  -1,  -1,  -8,  21, -22,  -2,   3,  -3,   4,   0,   2,  -1,   2,   7,  12,  -8,   9,   2,   0,  -1,   2,  -1,  -2,   0,  -2,  -4,   4, -11,   2,  -1,   3,   3,   0,   7,  -4,   1,   1,   0,   3,   3,  -3,   2,  -4,   5,  12,  -9,  -1,   0,   6,  26, -18,   8,  -3,  -4,   1,  -6,  -7, -46, -14,  14, -17,  43, -33,   3,  20,  74,  -6, -12,  18,  13,   2,  -8 },
     {   -1,   0,  -5,  -6,  19, -14, -11,  -2, -10,  19,   0, -10,  -1,  -2,   0,   2, -24, -17,  -6,   0,  -2,   6,  -1,   2,  -1,  -1,  -1, -24, -25,  17,   8,  -6,   4,   6,  -4,   3,   1,   2,   1,   3,  -8,  14,  14,  -2,   1,   2,   3,  -1,   2,   7,  21,   0,   3,   7,  -4,   8,  -4,  -8, -63,  30,  18,  -7,   3,  -3,  -4, -15, -52,  18,  18,   4,   9,  -3,   0,  -6,  21, -26,  11,  -3,  -1, -10,  -1,  -2,  10,  -8,  -3,  -1, -13, -17,  -3,  -2, -10,   1,   1,  11,  -4,   2 },
   },
   {
     {  -90,  23,   6,   5,   3,   2,   1,   1,   1,   1,   1,   0,  77, -15,  -6,  -4,  -3,  -2,  -1,  -1,  -1,  -1,  -1,   0, -35,   6,   3,   2,   1,   1,   1,   0,   0,   0,   0,   0,  13,  -3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -72,  26,   2,   3,   0,   1,   0,   0,   0,   0,   0,   0, -49,   4,   5,   2,   2,   1,   1,   1,   1,   0,   0,   0,  72,  -9,  -6,  -3,  -3,  -2,  -1,  -1,  -1,   0,  -1,   0, -45,   7,   4,   2,   2,   1,   1,   0,   1,   0,   0,   0,  23,  -5,  -2,  -1,  -1,   0,   0,   0,  -8,   2,   1,   0,   0,   0,   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   37, -14,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  71, -14,  -4,  -3,  -2,  -2,  -1,  -1,   0,   0,   0,   0,  24,   1,  -3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -72,   7,   6,   3,   3,   1,   1,   1,   1,   0,   0,   0,  50,  -6,  -4,  -2,  -2,  -1,  -1,   0, -29,   5,   3,   1,   1,   1,   1,   0,  11,  -3,  -1,   0,   0,   0,   0,   0,  -6,   2,   1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,  -2,   1,   0,   0,   1,   0,   0,   0,  -2,   1,   0,   0 },
     {   23,  73, -22,  -4,  -7,  -4,  -3,  -2,  -2,  -1,  -1,  -1,  -8, -78,  21,   4,   7,   4,   3,   2,   2,   1,   1,   1,   3,  48, -12,  -3,  -4,  -3,  -2,  -1,  -1,  -1,   0,  -1,  -1, -24,   6,   1,   2,   2,   1,   1,   1,   0,   0,   0,  -3,   8,  -2,   0,   0,  -1,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,  -3,  -2,   1,   0,   0,   0,   0,   0,   2,   2,  -1,   0,   0,   0,   0,   0,  -2,  -1,   1,   0,   1,   1,   0,   0,  -1,  -1,   0,   0,   1,   1,   0,   0 },
     {  -10,   9,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0, -42,   7,   3,   2,   0,   1,   0,   0,   0,   0,   0,   0, -69,   9,   4,   2,   2,   1,   1,   1,   0,   0,   0,   0,   2,  -5,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  57,  -3,  -6,  -1,  -2,  -1,  -1,   0, -61,   8,   6,   1,   2,   1,   1,   0,  42,  -8,  -4,  -1,  -1,  -1,  -1,   0, -19,   4,   2,   0,   1,   0,   0,   0,  10,  -3,  -1,   0,  -2,   1,   0,   0,   2,  -1,   0,   0,  -1,   0,   0,   0 },
     {    6,  -5,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,  23, -10,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  64,  -9,  -3,  -2,  -1,  -1,  -1,   0,   0,  -1,   0,   0,  77,  -9,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0,   0,  -2,   8,  -1,   0,   0,   0,   0,   0, -52,   2,   4,   1,   2,   1,   1,   0,  40,  -4,  -3,  -1,  -2,  -1,  -1,   0, -28,   5,   2,   1,   1,   1,   0,   0,  10,  -3,  -1,  -1,  -6,   2,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0 },
     {   -4,  -4, -66,   8,   8,   7,   4,   4,   2,   1,   1,   1,  10,  21,  77, -10, -11,  -9,  -6,  -5,  -2,  -2,  -1,  -1,  -5, -21, -53,   7,   8,   7,   4,   4,   2,   1,   1,   1,   7,  13,  29,  -4,  -6,  -4,  -3,  -2,  -1,  -1,   0,   0,   1,  -9, -11,   2,   4,   2,   1,   1,   1,   4,   2,   0,  -2,  -1,  -1,   0,  -2,  -2,   2,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   1,   0,   0,  -1,   0,   3,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0,   0,  -1,   0 },
     {   -3,  -8,  -2,   1,   1,   0,   1,   0,   1,   0,   0,   0,  -2,   2,   5,   0,  -1,   1,   0,   0,   0,  -1,   0,   0,  -9,  13,  -4,   1,   0,   0,   0,   0,   0,   0,   0,   0, -40,  -3,   5,   1,   2,   0,   1,   0,   0,   0,   0,   0, -60,  13,   2,   1,   1,   0,   1,   0,  16, -17,   1,   0,   1,   1,   0,   0,  49,   9,  -6,  -1,  -2,   0,  -1,   0, -62,   0,   7,   1,   2,   1,   1,   0,  49,  -6,  -5,  -1, -27,   4,   3,   1,  17,  -5,  -2,   0,  -6,   3,   0,   0 },
     {   20,  67,  -5,  -4,  -9,  -2,  -4,  -2,  -2,  -1,  -1,  -1,   7,   8, -16,   2,   1,   1,   1,   1,   0,   0,   0,   0,  -7, -52,  17,   1,   3,   1,   1,   1,   1,   0,   1,   1,  -2,  58, -11,  -3,  -3,  -2,  -2,  -1,  -1,  -1,   0,  -1, -22, -46,   8,   5,   3,   2,   2,   1,  -2,  31,  -3,  -3,  -2,  -2,  -1,  -1,  12, -18,   1,   2,   1,   1,   0,   0, -10,   7,   0,  -1,   0,  -1,   0,   0,   4,  -2,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    4,   7,   1,   1,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   3,  -2,  -4,  -1,   0,   1,   1,   0,   0,   0,   0,   0,   4,  -9,   1,   1,   0,   1,   0,   0,   0,   0,   0,   0,  29,  -6,   0,   0,   0,  -1,   0,  -1,   0,  -1,   0,   0,  73, -14,  -4,  -3,  -2,  -1,   0,   0,  82,  -2,  -6,  -2,  -1,  -1,  -1,   0,   6,  11,  -2,   1,  -1,   0,   0,   0, -46,   3,   4,   0,   1,   0,   1,   0,  19,  -3,  -2,  -1, -14,   3,   1,   1,   4,  -2,  -1,   0,  -2,   2,   0,   0 },
     {  -15, -60,  20,   2,   5,  -1,   2,   1,   2,   0,   2,   0, -13, -56,  11,   2,   5,   3,   2,   1,   1,   1,   0,   1,  -1,  27, -13,   1,  -2,   0,   0,   0,  -1,   0,   0,   0,   5,  33,   4,  -2,  -3,  -3,  -1,  -1,  -1,   0,   0,   0,  -8, -52,   0,   5,   4,   2,   1,   2,   8,  49,  -4,  -5,  -3,  -2,  -2,  -2,   5, -29,   5,   3,   1,   2,   1,   1,  -3,  16,  -4,  -3,   0,   0,   1,   0,   5,  -5,   1,   1,   0,   0,  -1,   0,   2,   1,   0,   0,   0,   1,   0,   0 },
     {    0,  -3,   0,   0,   1,   0,   0,   1,   1,   0,   1,   0,   1,   1,   0,   0,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   1,   1,   0,   0,   0,   0,  -9,  -6,   0,   1,   2,   1,   0,   0,   0,   0,   0,   0,  -8,  -1,   0,  -1,   0,  -1,   0,   0,  31, -12,  -1,   0,  -1,   0,   0,   0,  77,  -4,  -4,  -1,  -1,  -2,  -1,  -1,  -2,  15,  -2,   1,  -1,   0,  -1,   1, -72,  -2,   7,   0,  57,  -7,  -5,  -1, -19,   6,   1,   1,   3,  -3,   0,   0 },
     {   -4,  -4,  -1, -64,  -7,  20,   9,   1,   1,   2,   2,   0,   5,   1,   5,  77,   8, -23, -10,  -1,  -1,  -3,  -2,  -1,  -3,  -5,  -8, -54,  -4,  18,   7,   2,   1,   2,   0,   1,   1,   2,   5,  33,   1, -11,  -4,  -1,  -1,  -1,   0,  -1,   0,  -1,  -4, -15,   3,   6,   1,   0,   2,  -1,   3,   4,  -3,  -2,   0,   0,  -1,   1,  -3,   1,   2,   0,   0,  -1,  -1,  -2,   1,  -3,  -1,   1,   0,   0,   1,   2,   0,   2,   0,  -1,   0,  -2,   0,  -1,   1,   1,   2,  -1,   0,  -2 },
     {   -8, -20, -44,   2,  -1,   5,   4,   4,   2,   1,   1,   1,  -3, -25,  32,   3,   4,   0,  -2,  -3,   0,   0,   0,  -1,  -2, -14,  43,  -8,  -2,  -6,   1,  -2,   0,   0,   0,   0,  -5,   9, -70,   5,   6,   7,   2,   2,   0,   1,   0,   0,   7,   9,  53,  -2, -10,  -4,  -3,  -2,  -4,  -9, -33,   4,   7,   1,   3,   1,   3,  11,  14,  -4,  -3,   0,  -1,   0,  -3,  -5,  -3,   4,   0,  -1,   0,   0,   1,   6,  -3,  -1,   0,  -1,   2,   2,  -3,   1,  -2,  -1,  -1,  -1,   1,   0 },
     {    1,   5,   2, -16,  60, -28,  -7,   2,   0,  -1,   0,   0,  -2,  -3,  -1,  18, -75,  33,   8,  -3,   1,   1,   0,   1,   2,   2,   1, -13,  50, -24,  -8,   3,   0,  -1,   1,   0,  -1,  -1,  -3,  10, -26,  14,   5,  -1,   1,  -1,   0,   0,   0,   0,   4,  -5,   8,  -7,  -3,   0,   1,   0,  -4,   2,   2,   0,   0,   0,   1,  -2,   2,   1,  -5,   2,   1,  -1,   2,   2,  -2,  -1,   4,  -2,   0,   0,  -1,  -1,   0,   1,  -3,   1,   0,  -1,   2,   1,   0,   2,   0,   1,   0,  -1 },
     {   -2, -14,  22,   9,   4,  -1,  -2,  -1,  -2,   0,   0,  -1, -12, -67,   2,  -3,   2,   3,   3,   1,   3,   0,   1,   0, -13, -67,  -5,   4,   7,   4,   2,   3,   1,   0,   1,   1,   0,  31,  20,   3,  -3,  -4,  -4,  -2,  -1,  -1,  -1,  -2,   2,  22, -20,  -9,   3,   1,   2,   2,  -6, -37,  15,   6,  -1,   1,  -1,   1,   0,  37,  -8,  -5,   1,  -2,  -1,   0,   1, -20,   3,   4,  -1,   1,  -1,   1,   2,  10,  -2,  -2,   9,   2,  -2,   0,  -2,  -5,   1,   0,   4,   1,   0,   0 },
     {   -2,  -2,   2,   0,   3,   0,   0,   1,  -1,  -1,   0,   0,  -4,   0,   0,   0,  -1,   0,  -1,   0,   1,   0,   0,   0,  -4,   4,   0,   0,   0,  -1,   1,   1,   0,   0,   1,   0,  -5,   7,   2,   1,   1,   1,   0,  -1,   0,   0,   0,   0,  -8,  11,  -1,  -1,  -1,   0,   0,   1, -21,  13,   0,   1,   0,   1,   0,   0, -57,  14,   1,   2,   1,   1,   1,   1, -90,  15,   5,   0,   1,   1,   0,   0, -58,  -6,   5,   0,   6,  -6,  -1,  -1, -13,  -1,   2,   0,   0,  -6,   0,  -1 },
     {    0,   0,   0,   1,   2,   1,   0,   2,   0,  -1,   0,   0,   1,   4,   0,  -1,  -3,  -3,  -1,   0,   1,   1,   0,   1,   1,   3,   2,   0,   2,   0,  -1,   1,   0,   1,  -2,  -2,  -3,  -4,  -1,   1,  -1,   1,   1,   1,  -1,  -1,   1,   0,  -2,  -1,   0,   0,   1,   0,  -2,  -1,  -2,   2,   1,   1,  -3,   1,   0,   1,  -9,  -2,   1,   1,   0,   0,   1,  -2, -10,  -3,   2,   0,   1,  -2,   2,   1,  51, -18,  -2,   0,  46,  18,  -7,   0, -88,   2,  10,   1,  50, -14,  -5,  -1 },
     {   -1,  -4,   3,   1,  -1,   4,  -1,  -1,  -1,  -1,   0,   0,  -2, -15,   9,  -1,   1,  -2,   0,   2,   1,   1,   0,   0,  -6, -54,  11,   7,   2,   1,   3,  -1,  -2,   0,   0,   0, -10, -44,   0,   8,   3,   0,   1,   3,   1,   1,   0,   1,  -4,  48,  -6, -13,  -3,  -2,  -2,  -1,  11,  35,   1,  -2,  -2,   1,   0,  -1,  -5, -57,  -3,   8,   4,   2,   1,   1,   3,  49,  -2,  -8,  -2,  -2,  -2,  -2,   6, -28,   3,   6,  -3,   9,  -4,  -3,   9,  -1,   1,   1,  -3,   1,   0,   0 },
     {   -7, -17, -83,  32,   9,   6,   0,   2,   4,   0,  -1,   2,  -7, -14, -73,  24,   4,   1,   7,   5,  -1,   3,   0,   1,   0,   0,  14,  -5,   3,   7,   0,   0,   0,  -1,   0,   0,   1,   1,  30,  -1,  -9,  -6,   0,   0,  -1,   1,   0,  -1,   0,   4,  -4,   4,   3,  -2,   1,   0,   1,   6,  10,  -4,  -7,  -2,   0,   1,   0,  -3,   3,   4,   1,  -3,   1,   0,   0,   0,   3,  -3,  -1,   2,   0,   1,   0,   2,   3,  -1,   0,   0,   2,  -4,  -1,   4,  -2,  -1,   0,   0,   0,   1 },
     {   -3,  -5,  -5, -43,  -1,   0,  11,   2,   2,   1,   0,  -1,   1,  -8,   5,  25,  -2,   1,  -7,  -3,  -1,   0,  -1,   0,   1, -10,  10,  26,   3,  -2,  -2,   2,   0,  -1,  -1,   0,  -4,  -8,  -4, -62,   1,   7,   5,   1,   0,   2,   1,   1,   3,   4,  -7,  68,   1, -11,  -3,  -3,  -1,  -7,  10, -50,  -4,   6,   3,   3,  -3,   3,  -6,  25,  -2,  -4,   1,  -2,  -2,  -2,   6,  -8,   0,   3,   0,   0,   2,  -3,  -2,  -5,   9,  -3,   1,   6,   8,   4,  -1,  -4,   2,   0,  -2,   0 },
     {    1,   2,   0,   4,  -2,  -1,  -5,  -4,   1,  -1,   2,  -2,   1,   4,  -2,  -1,  -1,  -1,   2,  -1,  -1,   0,  -1,   1,   2,   4,  -2,  -4,   1,   3,  -2,  -1,   0,   1,   0,   1,   3,   2,   1,   3,  -4,  -2,   1,   1,   1,   1,   0,  -1,   2,  -6,   9,  -9,   2,   1,   2,   1,  -6, -11,  -6,   2,   2,   0,   1,   0, -20,  -8,   5,   1,   2,   2,   0,   0, -14, -10,   0,  -2,   0,   0,  -1,  -1,  34, -23,  -2,   2,  93, -10,  -7,  -2,  64,   6,  -7,   0, -16,  11,   2,   0 },
     {   -1,  -3,   1,   4,  -2,  -4,   1,   4,   0,  -2,   0,  -1,  -1,  -8,   1,   1,   0,   1,  -9,  -4,   1,   3,   0,  -2,  -3, -35,   5,   5,   2,  -2,   2,   5,  -3,  -3,   0,   2,  -8, -82,  13,  11,   5,   7,  -1,   3,   1,   2,   1,   1, -13, -74,  20,   0,   4,   0,  -3,   4, -11,  -1,  12,   7,  -5,  -4,   0,   4,  -5,  22,   6,  -1,  -4,   2,  -2,  -2,  -7,  -7,  -6,  -2,  -2,  -1,  -1,   2,  -1,  15,   2,  -3,  -1, -10,  -2,  -1,   2,   8,  -2,  -2,   6,  -1,   0,  -2 },
     {   -5,  -7, -11,  -8, -23, -54, -14,  13,  10,   7,   3,   2,   7,  10,   6,  20,  36,  64,  17, -14, -13,  -7,  -4,  -1,  -5,  -9, -12,  -8, -31, -43, -14,   9,  10,   6,   2,   2,   3,   4,  10,   9,  16,  27,   6,  -7,  -4,  -3,  -2,  -2,   0,   3,  -4,  -8,  -6, -13,   1,   3,   0,   1,  -3,  -3,  -2,   3,  -2,   3,   0,   1,  -3,   6,   8,   1,   3,  -1,   1,  -1,   5,  -4,  -6,  -3,  -3,   1,   1,  -1,  -3,   5,   3,   6,  -1,  -4,  -2,  -3,   2,   2,   0,   3,  -1,  -2 },
     {    1,   4,  15,  -7,  -8,  -3,   3,   1,   0,  -1,   1,   0,  -1,  -2, -25,  12,   7,   2,  -4,   1,   0,   4,  -1,   1,  -4,  -9, -47,  15,   5,   8,   3,   3,   0,  -5,   3,  -1,   0,   1,  27,  -8,  -8,   1,   1,  -1,  -2,   2,  -2,   0,   4,  14,  43,   6,  -3,  -3,  -7,  -1,  -5,  -8, -57,  -4,   2,   3,   6,   4,   5,   1,  62,  -3,  -9,  -3,  -4,  -3,  -2,   5, -38,   4,   6,   4,   3,   2,   0,  -9,  17,  -3,  -7,   5,   0,   0,  -3,  -4,  -2,   2,   4,  -1,   3,  -3 },
     {    0,  -2,  -4,  -2,  -7,   0,  -1,  -1,   1,  -1,   1,   1,   2,   1,   1,  -2,  -2,   2,   4,  -2,  -2,   1,   0,   0,   2,   3,   6,   4,   2,  -1,  -5,   0,  -1,   1,  -3,   0,   2,   2,  -1,   3,   2,  -4,   1,  -2,  -1,   0,   0,   0,   2,  -1,  -2,   3,   2,   6,   1,   0,   1,  -6,   1,   7,  -1,   0,  -1,   0,   3, -13,   0,  -5,   2,   4,   4,  -1,  -2,  -8,  -2,   3,   2,  -1,  -1,   2, -16,  -8,  -3,   1, -11,  -7,   3,   0,  47, -29,  -5,   3, 110,  -1,  -6,  -2 },
     {    4,   4,  18,  67,  -9, -22,  -8,  12,  -1,   1,  -8,   0,   2,   2,  15,  26,  -8, -16, -16,  -5,   2,   1,   0,   0,  -1,  -3,   6, -64,  -2,  17,  14,   8,  -3,  -5,   2,   2,  -1,  -7,   0, -31,  -5,  21,   7, -10,  -4,  -1,  -1,   1,   0,   1, -11,   8,   7,  -4,   8,   6,  -1,   4, -18, -25,  19,   9,  -7,  -4,   0,  -2,  -9,  10,   0,  -4,  -5,   2,  -1,  -6, -17,  12,  10,  -1,   1,  -2,  -1,   1,   2,   5,  -1,   1,  -3,   2,   2, -11,  -2,   1,   5,  -6,   3,   6 },
     {   -1,  -2,  -4,   4, -16,   0,   4,   3,  -3,   0,   0,   0,   0,   1,  -1,   3,   0,  -4,   5,  -7,  17,  -1,   2,  -2,   0,   8,   0,  -5,   9, -10,   7,   3, -10,   2,  -2,   0,   2,  10,  -4, -13, -10,   0,  -5,   7,   5,   0,   2,   0,   2, -22,  27, -10,  10,   2,   0,  -5,  -5, -40,  33,  -3, -12,  -4,   2,  -1,  -6,  18, -15,  -1,   4,  -1,   1,  -1,  10,  47, -20,  -6,   4,   0,  -6,  -3,  -8, -50,   4,  13,  -8,  52,  -4,  -7,   3, -30,   7,   8,  -9,   8,  -6,  -5 },
     {   -1,  -1,  -5, -18,  18,   7,   0,  -3,  -3,  -3,   1,  -4,   2,   3,  22, -11,   7,   6,   4,   1,  -4,  -1,   1,  -1,   4,   6,  68,  -9, -17, -11,  -1,  -3,  -3,  -3,   0,   1,   3,   7,  46, -14,   0,  -6,  -3,   4,  -4,   0,   1,  -4,  -3, -13, -33, -17, -16,   9,  -2,   7,  -1, -26,  -9,   0,  29,   8,   4,   3,  -1,  -9,  38,   1, -12,   4,  -7,   0,  -2,  12, -26,  -6,   8,  -5,  -1,  -1,   2,  -5,  14,  -6,  -4,  -2, -15,  13,  -2,   1,  10,  -4,  -3,   2,   1,  -1 },
     {   -1,  -1,   1,   0, -18, -10,   2,  10,   3,   2,   2,   2,   1,   3,   9, -22,  24,  10,   9, -10,  -2,  -2,   1,   1,   1,   3,  13,  -8,  15,   9, -10,   0,  -1,   3,   0,  -2,  -3,  -4,  -5,  42, -55, -23,   4,   9,   3,  -2,   1,   0,   2,   6,  -6,   3,  48,   7,  -4,  -7,  -2,   3,   0, -32, -16,   5,   4,  -1,   2,   4,   6,  39, -16, -18,   1,   6,  -1,  -7,  -8, -14,  22,  15,  -2,  -4,  -2,   4,   9, -16,   0,  -3,   5,  21,  -4,  -2,  -9,  -9,   0,   7,   7,  -9 },
     {    0,   0,  -2,  -5,  -1,  18,  10,  -4,   2,  -1,  -1,   1,  -1,  -1,  -1,   8, -14,  -9, -10,  -9,   3,  -1,   4,   1,   1,   0,  13,  17, -11, -29,   3,  16,   1,   2,  -1,  -3,   0,  -4,  16, -24,  23,  14,   2,  11,  -3,  -2,   1,   4,  -3,   7,  -3, -42,  21,  19,  -6,  -8,   0,  40, -30,  -2, -19,  -6,   6,  -6,   5,  27, -28,  10, -13,  -6,   5,   2,   5, -33,  -7,   7,  30,   6,  -1,   1,  -5, -22,   7,  -8,   3,  10,  17,  12,   1,  -6,  -4,  16,  -2,  23,  -8, -10 },
     {    0,   1,  -2,   0,  -1,  -5,  40, -20,  -6,  -9,   3,  -7,   0,   1,  -1,  -6,   1,  24, -60,  27,   0,  10,  -3,   6,   1,   3,   2,   1,  -6, -23,  56, -27,  -1,  -7,   5,  -6,   1,   4,  -1,   9, -13,  12, -32,   8,  -1,   1,  -3,   4,   1,   4,  -1,  -8,   3, -17,  19,  -1,   1,   4,   6,  -2, -14,  15,  -1,  -2,   2,  -1,   4,  13,  16,  -3,  -6,   0,  -2, -12,  -5, -16, -12,   1,  -2,  -1,   2,   8,  12,  15,   3,   1,  -6,   0,   3, -10,  -1,   0,   4,  -5,   4,   5 },
   },
   {
     {  119, -32, -12,  -6,  -3,  -2,  -2,  -1,  -1,  -1,  -1,   0, -13,   5,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -27,   6,   3,   1,   1,   0,   1,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -8,   3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,-116,  22,  11,   6,   4,   2,   2,   1,   1,   1,   1,   0,  24,  -8,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  38,  -5,  -4,  -2,  -1,  -1,  -1,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },
     {   27,  -9,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  12,  -5,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 111, -19, -12,  -4,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   0, -31,  11,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0, -36,   4,   4,   1,   1,   1,   1,   0,  12,  -2,  -1,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   -5,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0, -40,   9,   3,   2,   1,   1,   0,   0,   0,   0,   0,   0, -19,   8,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,-107,  12,  10,   4,   4,   2,   2,   1,   1,   1,   1,   0,  26, -14,  -1,  -1,   0,   0,   0,   0,  39,  -3,  -4,  -1,  -2,  -1,  -1,   0,  -6,   3,   0,   0,   0,   0,   0,   0,  -3,  -1,   1,   0,   0,   0,   0,   0,   4,  -2,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },
     {  -25,-108,  26,  11,  12,   6,   6,   3,   3,   1,   2,   1,   2,  14,  -4,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0,  10,  49,  -9,  -6,  -5,  -3,  -2,  -2,  -1,  -1,  -1,  -1,   4,  -9,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   6,  -7,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,   1,   2,  -1,   0,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   -9,  -4,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -6,   7,  -1,   0,   1,   0,  -1,   0,   0,   0,   0,   0, -37,  11,   4,   2,   1,   1,   0,   0,   0,   0,   0,   0,  -5,   6,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,-109,  10,  10,   3,   4,   2,   2,   1,  34, -14,  -2,  -1,  -1,   0,   0,   0,  32,  -1,  -3,   0,  -1,   0,   0,   0,  -5,   2,   1,   0,   0,   0,   0,   0,  -4,   0,   1,   0,   2,  -2,   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0 },
     {    0,   4,   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,  -9,  12,  -2,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,  -2,   8,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0, -41,   1,   5,   2,   2,   1,   0,   1,   0,   0,   0,   0, -14,  11,   1,   0,  -1,   0,   0,   0,-104,   7,  11,   3,   3,   1,   2,   1,  48, -17,  -3,  -1,  -1,  -1,   0,   0,  18,   3,  -3,   0,  -1,   0,  -1,   0,  -9,   1,   1,   0,  -1,  -1,   1,   0,   3,  -2,   0,   0,   1,   1,   0,   0 },
     {    2,  12,  -6,  -3,   1,  -1,   0,   0,   0,   0,   0,   0,  19, 106, -17, -13, -10,  -6,  -5,  -3,  -2,  -1,  -1,  -1,   5, -11,   6,  -1,   2,   0,   1,   0,   0,   0,   0,   0,  -4, -57,   7,   6,   5,   3,   2,   2,   1,   1,   1,   1,   0,   5,  -3,   0,  -1,  -1,   0,   0,   9,   6,  -2,   0,  -1,   0,   0,   0,  -6,   2,   1,   1,   0,   0,   0,   0,  -4,  -3,   0,   0,   1,   1,   0,   0,  -1,   1,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,  -1,  -1,   0,   0 },
     {   -9,  -1,-101,  15,   6,  13,   5,   6,   2,   3,   1,   2,   1,  -3,  13,  -2,   0,  -1,   1,   0,   0,   0,   0,   0,  13,  40,  52, -12,  -8,  -9,  -5,  -4,  -2,  -2,  -1,  -1,   2,  -3,  -8,   3,   1,   1,   0,   1,   0,   0,   0,   0,  -1, -19,  -9,   2,   3,   2,   2,   1,   1,   1,   3,  -1,  -1,  -1,   0,   0,   5,   2,   3,   0,   0,   0,  -1,  -1,  -1,   3,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,  -1,   0,   0,  -1,   1,   0,   0 },
     {    7,  34,  15,  -4,  -5,  -3,  -4,  -2,  -1,   0,   0,  -1,  -2,   1,  -3,   0,   3,   0,   2,   0,   1,   1,   0,   0,   4,  67, -21,  -6,  -5,  -3,  -1,  -2,  -1,  -1,   0,   0,  -4,  -4,   8,   0,  -1,   0,   0,   1,   0,   0,   0,   0, -32, -29,   7,   5,   4,   2,   1,   1, -22,  17,  -3,   0,  -1,   0,   1,   0, -76,   8,   8,   1,   2,   1,   2,   0,  28, -15,  -1,  -1,  -1,   1,   0,   0,   8,   0,  -1,   0,  -6,   2,   1,   0,  -1,  -1,   0,   0,   2,  -2,   0,   0 },
     {  -13, -37, -20,   7,   7,   4,   4,   3,   2,   1,   1,   1,  -1,   6,  -1,  -3,  -2,   0,   1,   0,   0,   0,   0,   0, -16, -58,  27,   4,   3,   3,   0,   1,   0,   1,   0,   1, -13,  24,  -9,   1,  -1,   0,  -1,  -1,   0,   0,   0,   0, -22,  42,  -4,  -2,  -2,  -2,  -2,  -1, -22,  -2,   5,   0,   2,   0,   0,   1, -71,  -2,   9,   3,   1,   0,   1,   1,  26, -10,  -3,   0,   0,  -1,  -1,   0,  17,   4,  -3,   0,  -3,   1,   0,   0,  -4,   0,   1,   0,   4,  -1,  -1,   0 },
     {   -2,   3,  -2,  -6,   1,   1,   0,   1,   0,   0,   0,   0,  -4,   3,  -3,   1,  -1,   1,   0,   0,  -1,   0,   0,  -1,  -4,   8,   0,   3,  -1,   1,   0,  -1,   0,   0,   1,   0,  -4,  13,   1,  -1,  -1,   0,   0,   0,   0,   0,  -1,   0,  -7,  12,   2,   1,  -1,   0,   0,   0, -30,   4,   0,   2,   1,   1,   0,   0, -26,  18,   3,   1,   0,  -1,   0,   0,-114,   7,  12,   2,   3,   1,   2,   0,   5, -17,  -1,   0,  17,   0,  -3,  -1,  -1,   1,   0,   0,  -6,   0,   1,   0 },
     {    0,  -5,   0,   9,  -2,   3,  -2,   0,  -1,   0,   0,   0,  -9, -52,  14,  -1,  10,   1,   2,   0,   0,   0,   0,   0,  -2,  -9,   1,   2,  -1,  -1,   0,   1,   0,   0,   0,   0, -15, -93,   4,  15,   8,   7,   3,   2,   2,   1,   1,   1, -12,  25, -10,  -2,   1,   1,  -1,  -1,   4,  48,   0,  -9,  -5,  -3,   0,  -1,  -4, -13,   6,   2,   1,   2,   0,   1,  -8,  -3,  -1,   1,   0,   0,   0,   0,  10,   1,  -3,  -1,   9,   2,  -1,  -1,   0,   1,   0,   0,   1,   1,   0,   0 },
     {    0,  -1,  -3,  11,   5,   0,   0,  -2,   1,   0,   0,   0,  11,  16,  98, -12,  -9,  -8,  -6,  -4,  -2,  -3,  -2,  -3,  -6,  -7, -33,   5,   1,   4,   1,   3,   1,   1,   1,   1,  -5,  -1, -58,   6,   6,   4,   4,   3,   1,   2,   1,   1,  -1, -12,  28,  -5,  -1,  -1,  -1,  -1,  -4,   2,  -1,   2,  -1,   2,   0,  -1,  -4,   6,  -7,  -1,   0,   0,   1,   0,  -1,  -2,   0,   1,   1,   0,   0,   0,  -8,  -1,   1,   1,   3,   0,  -1,  -1,  -1,   0,   0,   1,   0,  -1,  -1,   0 },
     {    2,   6,  40, -55,   2,   3,   4,  -1,   1,   0,   1,   0,   3,   4,  22,  11,  -3,  -3,  -5,  -2,   0,  -1,   0,  -1,  10,  18,  60,  35, -16, -11,  -8,  -3,  -3,  -1,  -3,  -1,  -3,   3, -37,  -9,   8,   4,   4,   3,   1,   1,   1,  -1,  -3,   5, -50, -15,   8,   6,   6,   2,   4,   1,  20,   1,  -5,  -1,   0,  -1,   1,  -8,   9,   7,  -1,   1,  -2,  -1,  -1,   0,  -6,  -1,   1,   1,   1,   0,  -2,   3,  -4,  -2,  -8,   3,   1,   0,  -1,  -1,   2,   0,   3,  -2,  -2,   0 },
     {   -4,  -6, -20, -45,   2,   6,   8,   3,   5,   1,   1,   1,   2,   0,   0,   8,  -4,   0,  -1,   0,  -1,   0,   0,   0,   0,  -6, -18,  26,   2,  -2,  -2,   1,  -1,   0,   1,   1,   2,  -7,   7, -13,   1,   1,  -1,   0,   1,   0,   0,   0,  -1, -20,  19,   0,   3,   0,  -1,   0,  -5,  -8,  -5,   6,   0,   0,   0,   1,  18,   0,  -5,   0,   0,  -1,   0,  -1,   5, -16,   4,  -1,   0,  -1,   0,   0,  98, -14, -10,  -2,   9,  17,  -3,  -1, -22,   5,   4,   0,  -1,  -2,   0,   1 },
     {  -10, -18, -26, -61,   8,   5,  13,   3,   5,   3,   3,   2,  -2,  -7, -11,  10,  -1,  -1,   0,   0,   0,   1,   1,   1,  -5, -25, -15,  43,   4,   4,  -3,   0,  -1,  -1,  -1,   1,  -1,  -9,  16, -11,   2,   2,   1,  -1,   0,   0,   0,   1,  -7, -40,  21,  -4,   0,   2,   1,   1,  -7,  25, -12,   5,  -1,   1,   0,  -1,  -9,  19,  -2,   1,  -1,  -1,   0,   0,   5,   1,   3,   0,   1,   0,  -1,   0, -62,   6,   9,   1,   9, -11,  -2,   0,  11,   2,  -1,   0,  -1,   3,   1,   0 },
     {    3,  -7,  27,  33,  -2,  -4,  -4,  -4,  -3,  -3,  -2,   0,   0,   0,  -1,  -9,   2,  -1,  -1,   0,   1,   1,   2,   0,  -4, -42,  35, -20,  -2,   2,   3,   0,   1,   0,   1,   0,   1,   2,  -3,   7,   3,   2,  -1,  -2,  -1,  -1,  -1,   1, -16, -85, -23,  15,   7,   7,   2,   3, -14,  23,   1,  -6,  -2,   0,  -1,   0,   7,  32,   4,  -1,  -4,  -5,   0,  -1,  -2,  -6,   2,   2,   1,   1,  -1,  -1,  13,  -8,  -4,   2,  15,   4,  -3,  -2, -10,   5,   2,   1,   4,  -1,  -1,   0 },
     {   -1,  -3,  -2,  -1,  -3,   3,  -2,  -2,   0,   1,   1,   0,  -3,  -3,  -2,  -2,  -1,  -3,   1,   0,   0,   2,   3,  -1,  -3,  -9,  -4,   1,   3,  -1,   5,   0,  -1,   0,   1,   1,  -3,  -2,  -3,  -3,  -1,   2,   1,   0,  -1,   1,   0,   1,  -2,  -7,   5,   4,  -1,   2,   0,   0,  -2,  16,   0,   2,   2,   1,   0,  -1,   7,  21,   0,   0,  -2,   0,   0,  -1, -19,  -7,   2,   2,  -1,   1,   0,   0,  15,  14,  -3,   1,-113,   6,  10,   2,  36, -15,  -3,  -2,  11,   4,  -1,   0 },
     {    2,   1,   0,  18,  16,   0,  -5,   4,  -7,   0,  -6,   0,   5,  -2,  10,  88,  12, -10,  -6,  -8,  -6,  -7,  -6,  -4,  -1,  -6,  -2,  -8,   2,   7,  -1,  -2,   6,   2,   4,   0,  -7, -32,   1, -50,  -5,   8,   3,   5,   4,   2,   1,   3,  -3,  -6, -11,  14,  -4,   1,  -3,  -2,  -9, -47,   0,   4,  10,   7,   2,   3,  -9,  15,  -3,   1,   2,  -3,  -1,   3,   1,   8,   0,  -6,   0,  -1,  -1,  -1,  -4,  -2,   2,  -1,  -2,  -3,  -2,  -3,   7,   0,   0,   2,   4,   6,  -2,   0 },
     {    3,   8,  -5,   8,  70,  -7,  -2,  -5,  -6,  -5,  -3,  -2,   2,  10,  -9,  29, -13,  -9,   1,  -2,   0,  -3,  -1,   1,  -5,  -3, -15,  -8, -56,   5,   5,   7,   4,   3,   4,  -1,   3,  25,  -2, -13,  12,   4,  -1,  -1,  -2,   1,  -1,   0,   1,   9,   3,   8,  14,  -3,  -1,  -4,  10,  54,   2, -13, -12,  -3,   0,  -4,  10, -20,   6,   0,  -1,   3,   1,   2,  -4, -18,  -2,   5,   4,   0,   0,   2,  -1,   9,  -5,  -3,  -3,   6,   1,  -4, -20,   0,   6,   0,   2,  -7,  -2,   0 },
     {    3,   5,   5,  -5,  57,  -2,  -6,  -6,  -4,   0,  -2,  -1,  -6, -16,  -2, -41,  -8,   2,   8,   3,   0,   4,   0,   1,  -1,  -5,   2,   6, -51,   0,   3,   2,   2,  -1,   4,   3,  -2, -26,  12,  26,  19,  -1,  -5,   0,  -1,   0,  -1,   0,   0,  -4,   5,  -7,  16,   3,   0,  -2, -10, -61,   6,  15,  -1,   3,   2,   1,  -9,  26, -11,  -4,  -3,  -1,   1,  -1,   5,  21,   2,  -4,   0,   0,   0,  -1,   1,  -5,   7,   2,  -5,  -1,  -1,   4,  10,   1,  -3,  -1,  -1,   3,   0,   1 },
     {   -1,  -2,   2,   2, -16,   7,  -1,   4,  -4,   2,  -2,   1,   0,   0,  10,   5,  13,   1,  -2,   0,   1,  -4,  -3,   3,  -1,   0,   3,   9,  -3,  -4,   0,  -1,   1,  -1,  -2,  -3,  -1,  -5,  19,  11,   3,  -3,  -3,   0,  -3,  -1,   0,   0,  -1,   4,   2,   0,   1,  -3,  -4,  -2,  -3,  -8,  -7,   1,  -5,  -2,   1,   4,  -2,  13,  -6, -15,  -2,   1,   2,  -1,  -2,  24,   1,   3,  -1,  -3,   0,   1, -22,   9,  -1,   2, -34,  31,  -1,   1,-103,  -4,  13,   3,   1, -17,   1,  -2 },
     {    1,   3, -12,  30,  -2,   1,   6,   6,  -2,  -3,  -1,  -1,  -4,   0, -50,   7,  17,   1,   5,   3,   5,   0,  -1,   1,  -1,   2, -29,  34,   8,  -1,  -2,  -2,  -1,  -3,   1,  -4,  -4,  -4, -56,   7,   9,   7,   2,   4,  -1,   1,   1,   1,  -2, -11,   0, -32,  -7,   1,   3,  -5,   4,   1,  63,   7, -20,  -3,  -3,  -3,  -1,  12, -13,  14,  -6,  -6,  -3,   0,   0,  12, -10,  -6,   1,   4,   0,  -1,   2,  -2,   5,  -6,   2,  -1,  -2,  -1, -11,  -3,   3,   2,   6,  -4,  -3,   3 },
     {    2,   1,  -5,  46,  -6,   1,  -4,   5,   0,  -4,  -2,  -1,   4,  10,  19,  -1,  16,   6,   1,  -6,  -5,  -5,  -3,  -2,   0,  -2, -15,  59, -17,  -7,  -3,  -3,   2,   4,   2,  -1,   6,  10,  40,  -6,  -6,  -4, -11,   2,   2,   0,   2,   1,  -4,  -8, -33, -58,  10,  12,   7,   2,  -1,   1, -31,  17,   2,  -2,   2,   1,   2, -14,  20,   0,   2,  -3,  -2,  -3,  -3,  -4,   6,   1,   0,  -1,   0,   1,   5,   3,  -8,  -1,   2,   1,  -1,   1,  16,  -6,   2,   1,   7,  -6,  -2,  -1 },
     {   -1,  -5,   2,  -5,   3, -13,   7,   1,   5,  -8,   0,  -1,   0,   1,  -3,  -9,  19,   6,  -9,   5,   3,   2,   4,   1,  -2, -12,   6,  -8, -11,   5,   1,  -1,  -3,   5,   0,  -4,  -1,  -9,  -6,  -9, -12, -13,   7,   2,   1,   0,  -2,   0,  -4, -40,  21,   6, -11,   2,   7,   2,  -1, -16,   3,  -3,  -2,   2,   0,  -5, -18, -90,   7,  11,   6,   5,  -2,   0, -18,  44,  -1,  -1,  -3,  -1,   1,  -3,   3,  19,   1,  -7,  -9,   5,   0,   0,   2, -10,  -1,  -2,  16,  -9,  -1,   2 },
     {   -6,  -8, -20, -22, -19,  10, -15,  -4,   3,   0,   0,   4,  -1,   0,   7,  -2, -14,  -8,   6, -20, -14,   1,   8,  -5,  -7,  -7, -49, -18,  -6,   2,   7,  12,   0,   5,   9,   9,   3,   5,  10,  23,  -5,  22,   0,   6,   2,  -3,   4,   4,  -3,  -8, -55,  26,  20,  14,   5,  -7,   2, -13,  35,   3,   9,  -6,   2,   1,   0,  -7,  41,   1,   7,  -4,   1,   6,  -5,  11, -15,  -3,   5,  -3,   0,   0,  -2,  17,  -9,   3,   0,  -6, -10,  -5,  -1,   2,  16,   5,   2,  -5,  -1,   2 },
     {    1,   0,   1,   7,   3,  -5,   9,   0,   3,   2,   1,  -3,  -1,   1,  -8,   0,  11,   1,  -2,  14,   0,   5,  -6,   9,  -1,  -2,  -4,   8,   5,   2,   3,  -1, -10,  -8,  -7,  -4,  -2,  -1, -18,  -7,   4,   5,  14,   0,   8,  -1,  -4,  -2,  -1,  -4,  -2,   7,  22, -13,  -6,   2,  -3,  -4, -13,   4, -19,  -6,  -1,   2,   0,  -5,  39, -18, -17,   4,   3,  -3,   4,  16,  19,   1,  -9,  -8,  -5,   2,   3,  13, -17,  -4,  -9,  -5,   4,   5,   2,  14,   7,   0, -94,  23,  11,  -1 },
     {    0,   0,  -1, -11,  20, -26,   9,   8,  -1,   3,  -6,  -2,   4,   8,  10,  -4,  82, -20,  -1, -17,  -6, -13,  -9,  -4,   1,   2,   3,   9,  -2,   8,  -4,  -2,   7,   3,   4,   1,   0,   6,   5,  32, -53,   6, -17,   3,   0,  -1,   2,   5,   3,   5,  -4,  20,  -5,  -2,   0,  -3,   2,   6,  11,  -2,   0,   0,   8,  -3,   4,  11, -12,  -5,   7,   6,   1,  -2,  -1, -11,  -6,  -8,   0,   3,  -1,   8,   9,  14,  11,  11,   2, -11,   1,   2,   1,   0, -12,  -6, -10,   9,   3,  -4 },
     {   -4,  -5, -12,  -5,   2, -77,  -9,   6,  15,   7,   1,   2,  -1,  -3,   1,   0, -21,  21,   9,  -4,  -2,   5,  -2,  -5,   2,   6,  -3,   2,  10,  64,   7, -23, -10,  -2,   0,   0,  -3,  -7,  -6,   0,   1, -43,  -7,  13,   2,   2,  -1,   1,  -2,   0, -21,  -7,   7,  -7,   0,   5,   0,  -2,   4,  -2,   3,  18,   2,   0,   2,   6,   2,   3,  -3,  -3,   1,  -2,  -1, -11,   6,   8,   3,  -9,  -1,   1,  -1,   6,   3,  -5,  -4,   7,  -2,  -3,  -9,  -2,   2,   5,   3,   2,   2,   1 },
     {    0,  -1,  -2,   5,   6, -11,  -2,  -5,  -1,  12,   9,  -2,  -2,  -7,  -5,  -3,  -9, -41,  -5,  27,  14,   9,  -6,  -1,   0,  -1,  -3,  12,  11, -19,   4,   1, -10,  -9,  -5,  -5,   0,  -7,   0,  10, -15,   8,  32,  -6,  -9,  -9,   2,  -3,   1,  -6,  -1,   0,  -2,   4,  20,  -3,  -4, -33,   4, -28,  -5,  -2,   9, -11,  -1, -11,   7, -16, -16, -10,  -4,  -3, -10, -56,  10,  21,  12,  -1, -12,   0, -12,  26,  -1,   7,  10,  33,  13,  -7,  -1, -21,   8,   3,  11,  10,  -5,  -7 },
     {    1,   2,   2,  -1,   7, -15,  31,  26,  -2,   1,  -9,  -3,   1,   1,   7,  20,  -4, -24,  11,  14, -16,   1,  -4,   2,  -1,  -1,  -5,  -7,  27, -13, -33,   7,   1,   1,   3,   3,   5,   7,  14,  37,  46,   3, -22, -13,   9,  -4,  -1,   3,  -1,  -4, -10,  -3,   3,  12,  11,  -7,  -4, -10, -15, -33, -31,  16,   5,   7,  -3, -11, -11,  34,  -3,   9,  -4,  -3,  -2,   2,   3,  13,  -2,  -1,   3,   1,   4,  -6,  10,  -5, -12, -23, -16,  -3,   0,   1,   1,   3,  -5,  -2,   6,   9 },
   }
 },
 {  //20
   {
     {  122, -35,  -6,  -5,  -2,  -2,  -1,  -1,  -1,  -1,  -1,   0, -10,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0, -11,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   1,   0,   0 },
     {   -5,   9,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,-118,  15,   9,   4,   3,   2,   2,   1,   1,   1,   1,   0,  28, -14,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,  31,  -2,  -2,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0 },
     {   10,  -5,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  17, -10,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0, 111,  -8,  -9,  -3,  -3,  -2,  -2,  -1,  -1,  -1,  -1,   0, -34,  21,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0, -40,  -1,   4,   1,   1,   0,   1,   0,   7,  -4,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,  -1,   1,   0,   0,  -2,   0,   0,   0,  -1,   0,   0,   0 },
     {    0,   8,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0, -32,   5,   3,   1,   1,   0,   1,   0,   0,   0,   0,   0, -19,  12,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,-108,   7,  10,   3,   3,   1,   2,   1,   1,   0,   1,   0,  31, -27,   0,  -2,   0,   0,   0,   0,  33,   0,  -4,   0,  -1,   0,  -1,   0,  -2,   1,   1,   0,   0,   0,   0,   0,   5,  -1,   0,   0,   0,   0,   0,   0,   3,  -2,   0,   0,   4,   0,  -1,   0,   1,  -1,   0,   0,   3,  -1,   0,   0 },
     {    5,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8, -12,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,  42,  -3,  -5,  -1,  -2,  -1,   0,   0,   0,   0,   0,   0,   7, -14,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0, 103,  -2, -12,  -2,  -4,  -1,  -2,   0, -40,  32,   0,   2,   0,   0,   0,   0, -25,  -5,   4,   0,   1,   0,   1,   0,   2,   3,  -2,   1,  -1,   0,   0,   0,  -6,   1,   0,   0,  -2,   3,   0,   0,  -5,   1,   1,   0,  -1,   0,   0,   0 },
     {   -1, -15,   2,   1,   0,   0,   0,   0,   1,   0,   0,   0,  13,  -1,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   9,  -9,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  34,  -3,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  26, -19,   0,  -1,   0,   0,   0,   0, 100,   0, -14,  -1,  -4,  -1,  -2,  -1, -31,  38,  -1,   1,   0,   0,   0,   0, -30,  -3,   5,   0,   1,   0,   0,   0,   3,   1,  -1,   0,  -6,   2,   0,   0,  -2,   3,   0,   0,  -4,   0,   1,   0 },
     {    0,  11,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   8,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -8,   0,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -5,  15,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -26,   3,   3,   1,   1,   0,   0,   0, -15,  19,   0,   0,   0,   0,   0,   0, -98,  -1,  16,   1,   4,   0,   2,   0,  18, -46,   3,  -1,   1,   0,   1,   0,  46,  -3,  -9,   0,   3,   8,  -1,   0,  -3,   1,   1,   0,   2,  -3,   1,   0 },
     {   33, 117, -22,  -3,  -9,  -4,  -5,  -3,  -3,  -1,  -2,  -1,   9,  -7,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0, -11,   1,   1,   0,   0,   1,   0,   0,   0,   0,   0,   6,   2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   1,   0,   0,   0,   6,   0,   0,   0,   0,   0,   0,   0,   3,   6,  -1,   0,  -1,   0,   0,   0, -21,   6,   3,   0,   1,   0,   0,   0,  -4,  -5,   2,   0,  -3,   0,   0,   0,  -1,  -1,   1,  -1,  -1,  -2,   1,   0 },
     {   -5, -12,   1,   1,   1,   1,   1,   0,   0,   0,   0,   0,  -9,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,  15,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0, -10,  -6,   3,   0,   0,   0,   1,   1,   0,   0,   0,   0,   1,  17,  -3,   0,  -1,   0,  -1,   0, -27,  -6,   5,   0,   1,   0,   0,   0,  11,  25,  -6,   1,  -1,   0,  -1,   0, -93,  -2,  18,   0,   4,   0,   2,   0,  47, -43,  -5,   0,  13,  21,  -6,   0,  -9,  -7,   5,  -1,   7,  -2,  -2,   1 },
     {    6,   6,  -2,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0,   7, -15,   1,   0,   0,   1,   1,   0,   0,   0,   0,   0,  10,  -5,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   8, -15,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  16,  -4,  -2,   0,   0,   0,  -1,   0,  17, -24,   1,  -1,   0,   0,   0,   0,  45, -14,  -5,  -1,  -2,   0,  -1,   0,  42, -20,  -6,   0,  -2,   0,  -1,   0,  83,  10, -19,  -1,  -3,  43,  -5,   0, -17,  16,   3,   0, -13,   8,   2,   0 },
     {   -1,   0,  -1,   0,   0,   1,   0,   0,   1,   0,   0,   0,  -3,   6,  -1,   1,  -1,  -1,  -1,   0,   0,   0,   0,   0,   1,   5,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -6,  -5,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,  12,  -2,  -1,   0,   0,   0,   0,  -6,  -6,   2,   1,   1,   1,   0,   0,   3,  13,  -1,   0,   0,   0,   0,   0,  -8,  -7,   2,   0,   0,   0,   0,   0,  10,  34,  -6,   1, -98,   8,  24,  -1,  16, -55,   2,   1,  27,   5, -11,   0 },
     {    0,   1,   8,  -2,   1,  -1,   0,  -1,  -1,   0,   0,   0, -18,-110,  10,   6,   6,   5,   5,   3,   2,   2,   1,   1,  -7,  51,  -8,   0,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0,   9,  11,   2,  -2,   0,  -1,  -1,   0,   0,   0,   0,   0, -10, -11,   0,   1,   1,   1,   1,   0,   4,  12,  -2,  -2,  -1,  -2,   0,   0,  -5,  -3,   1,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -2,   4,   2,  -1,  -5,  -6,   2,   1,   9,  -3,  -3,   0,  -1,   3,   0,  -1 },
     {   -3, -13,  -5,   2,   1,   0,   0,   0,   1,   0,   1,   0,  -3, -47,   9,   1,   2,   3,   1,   2,   1,   0,   0,   0, -19, -99,   5,   8,   5,   4,   2,   3,   2,   1,   2,   1, -17,  27, -11,  -1,  -2,   0,  -1,  -1,  -1,   0,   0,   0,   6,  44,  -2,  -4,  -2,  -3,  -1,  -1,   6,   1,   5,   0,   0,   0,   0,  -1,  -2,  -8,   1,   2,   1,  -1,   0,   0,  -8,   6,  -1,   1,   0,   0,   0,   0,   2,   1,   1,  -1,   0,   1,   1,  -1,   0,  -3,   0,   1,   6,   2,  -3,   0 },
     {    3,   1,   0,   0,  -1,   0,   0,  -1,   0,   0,  -1,   0,   4,   2,   1,  -2,   0,  -2,   0,   0,   0,  -1,   0,   0,   7,  -9,  -1,   2,   1,  -1,   1,   1,   1,   0,   0,   0,   0, -14,   1,   0,   2,  -1,   0,   0,   0,   0,  -1,   0,   7,  -6,   0,   0,   0,   0,   0,   0,  -2, -13,   2,   1,   0,   0,   0,   0,   6,   1,   0,   0,   0,  -1,   0,   0,  -4, -24,   2,   0,   2,   0,  -1,   0,  20,  -8,  -3,  -1,  16, -39,  -2,   0,  97,   0, -28,   1,  -5,  51,  -3,   0 },
     {  -10, -16,-111,   1,   5,  11,   7,   7,   3,   3,   2,   2,  -2,  -9, -10,   2,   3,   2,   1,   1,   2,   0,   0,   0,   4,   9,  39,   0,  -5,  -4,  -1,  -1,   0,  -2,  -1,  -1,  -3, -29,   5,   6,   1,   0,   0,   2,   0,   1,   0,   0,  -8,   9, -13,   0,  -1,   1,  -1,   1,   6,   8,   0,   1,  -2,  -1,  -1,   0,  -3, -10,   6,   3,   0,   0,   0,  -1,  -4,   0,  -5,   1,   1,   0,   0,   0,  -3,   4,   0,  -2,   2,   0,   0,  -1,   0,   2,   1,  -1,   0,   1,   1,   1 },
     {    3,   0,  22,  -3,   0,  -3,   1,  -1,  -1,   0,   0,   0,   1, -17,  18,   3,  -3,   0,   0,  -1,   0,   0,   0,   0,   0, -37,  10,   3,   3,   1,   0,   0,   1,   0,  -1,   0, -12, -90,   3,   5,   7,   4,   4,   2,   3,   1,   1,   0, -31, -41,  -9,   6,   0,   3,   3,   2, -13,  31, -13,   1,  -2,  -1,   0,  -1,  -4,  28,  -5,   0,   0,  -1,   0,   0,   7,   9,   2,  -1,   2,  -1,  -1,  -1,   1,   0,   1,  -1,   3,   3,  -3,   1,  -5,  -6,   3,  -1,  12,  -7,  -3,  -1 },
     {    1,  -6, -17,   6,   2,   1,  -1,   0,   0,  -1,   0,   0,   8,   1,   4,  -4,  -1,   0,   0,   0,   0,   0,   0,   0,  -1, -14,   6,   2,   1,   0,   0,   0,   0,  -1,   0,   0,  20,  40,   1,  -4,  -2,  -2,  -1,  -1,  -1,   0,  -1,   0,   3, -70,   5,   2,   4,   1,   0,   1,  -4,   3,   2,   0,  -1,   0,   0,  -1,  10, -19,   3,   2,   0,   1,   1,   1,   9,   7,  -2,  -2,   1,   1,   0,   0,  11, -32,   4,   2,  14,   4,  -6,   0,   0, -36,   5,   2,  69,   0, -22,  -1 },
     {   -5,  -8, -24,   2,   0,   3,  -1,   2,   0,   0,   0,   1,  -3,   0,   9,   1,  -1,   0,   0,  -1,   0,   1,   0,   0,  -9, -20,   1,  -1,   1,   0,   1,   1,   1,   0,   0,   0,   3,  46,  -2,  -3,  -3,  -1,  -2,  -1,  -1,   0,   0,   0,  -1, -53,   8,   6,   2,   2,   2,   2, -36,  13, -22,   4,  -1,   1,   0,   0,  20,  57,   4,  -5,  -3,  -3,   0,  -2,  -6, -20,   4,   1,   2,   1,   0,   1,  -2,  21,  -5,   0, -18,   2,   5,  -2,   0,  24,  -1,  -2, -47,   1,  16,  -1 },
     {    0,   1,  -5,  -3,   4,  -1,  -1,  -2,   0,   1,   0,   0,   0, -15,   5,   4,  -1,   0,   0,   1,   0,  -1,  -1,   0,   4,   3,  24,  -3,  -1,  -3,   0,   0,   0,  -1,   0,   0,  -2,  -5,  -8,   0,  -1,   2,   0,   1,   1,   0,  -1,   0,  13,   6,   4,  -2,  -1,  -1,  -2,   1, -20, -89,  -2,  13,   3,   3,   3,   3, -27,  54, -21,  -5,  -4,   0,   0,  -1,  34,  12,  12,  -5,   1,   0,   1,   0, -15, -19,  -2,   3,   9,   3,  -2,  -3,  -2,  -6,   2,   0,  13,  -5,   0,   0 },
     {   -2,  -3, -27,  -8,   1,   0,   4,   5,   1,   0,  -1,   1,   8,  10, 101,  11,  -9, -16, -10,  -6,  -6,  -2,  -2,  -1,  -3,   8, -48,   1,   1,   6,  -3,   3,  -1,   1,   1,   1,   0,  -8, -18,  -5,   1,   0,   2,   0,   2,   0,   2,   0,   3,   9,  27,  10,  -2,  -4,  -1,  -1,   4,  -9, -15,  -3,   5,   4,   2,   1,  -5, -11,   6,   1,   0,  -2,  -1,   0,   0,   5,  -4,   2,   1,   3,   0,   0,  -1,  -5,   0,   1,   0,   4,  -1,   0,   3,  -4,  -1,   1,  -2,  -5,  -1,  -1 },
     {    0,   3, -11,   8,   3,  -1,  -2,   1,   0,   1,  -1,   1,   0,   3,  -7,  -3,  -1,  -1,  -3,   2,  -1,   0,   1,   2,   0,  14, -19,  -2,  -3,   2,  -1,   2,   1,   0,   0,   1,  -3,   9, -21,  -2,   2,   0,  -2,   1,   1,   0,   0,  -1,   0,  45,  -7,   0,  -6,   1,  -3,  -1,  11,  54,  -4,  -6,  -1,  -3,   0,  -3,  30,  60,   2,  -9,   2,  -2,   0,   0,  46, -18,  19,  -2,  -1,   1,   0,  -1,   0, -36,   5,   6,  13,  -2,   0,   2,   1, -20,  -3,   2,  19,   1,  -8,   2 },
     {    3,   3,  28,   3,   1,  -2,  -8,  -3,  -3,   2,  -3,  -1,   5,   7,  61, -10,  -8,   1,  -1,  -1,  -4,  -1,  -1,  -1,   6,  13,  92,   0, -19,  -5,  -3,  -4,  -5,  -1,  -1,  -2,   1,  19,   3,   5,  -4,   0,   1,  -3,   0,   0,  -1,   1,   1,  13, -26,  -4,   8,  -2,   1,  -1,   6,  16,   2,  -1,   1,  -3,   1,   0,  11,  -1,   3,   2,  -1,   1,  -1,  -2,  -2, -16,  -6,   1,   1,   1,   0,  -1,  -6,  -6,  -9,  -1,  -3,   3,  -5,   1,   1,   0,  -2,  -2,  -3,   5,  -1,   1 },
     {    2,   1,   4, -10,   3,   4,   4,  -2,   1,   0,   2,  -1,   2,  -7,  -7,  12,   9,   2,  -4,   1,  -2,  -1,  -1,   0,   3,  -5,   0,   7,  -6,   1,  -4,  -1,   0,  -1,   1,   0,   0, -16,  -1,   4,  -1,   2,   1,   1,  -1,   0,   1,   1,   8,   1,  29,  -6,   0,   1,   2,   0,   1, -16,  -9,   6,  -4,   3,   1,   2,  20, -10,  20,  -3,   4,   2,   1,  -1, -15, -88,  18,  14,   2,   2,   3,   4, -41,  20, -27,   2,  24,   9,  -1,  -5, -11, -14,   3,   2,  18, -14,  -7,   1 },
     {   -5,  -1,   1,-105, -14,   8,  11,  -3,   0,   0,   1,   3,  -1,   2,  -7,  13,  -8, -19, -10,   6,   1,   0,   3,   0,   2,   4,   7,  51,  20, -12, -11,  -1,  -1,  -2,   0,   2,   0,   9,  -6,  -7,   4,   9,   4,  -2,  -1,  -1,   1,  -1,   0,  -1, -10,  -1,   2,  -5,  -4,   0,   0,   8,   8,  -1,   4,   0,   2,  -3,   1,   6,  -6,   6,   1,  -1,  -1,   3,   3,   3,  -4,   3,   3,   3,  -2,  -2,   2,  -5,  -1,   1,  -3,   1,  -2,   1,   1,   6,  -2,   0,   5,   1,  -1,  -1 },
     {   -1,  -1,  -1,  -7, -17,   2,   7,   3,  -1,   2,  -1,   1,   1,  -1,   7,  11,  -7,  -5,   4,  -4,   3,   3,  -2,  -1,  -2, -10, -15,  -7,  -1,   6,  -3,   2,  -1,   0,  -2,   3,   7,   4, 110,  11, -15,  -8, -12,  -2,   1,  -2,  -1,  -1,  -1,  13, -15,  23,   6,   5,  -1,  -1,   3,  -4, -14, -11,   6,   3,   7,   1,   1,   6,   2,   0,  -8,   1,   1,  -1,  10,  -8,  -2,   4,   3,  -1,   5,   1, -10, -23,  -3,   4,  -6,   4,  -5,  -2,  -3,  -2,  -3,  -2,  -2,   4,  -2,  -2 },
     {    1,  -1,  -2,   9,  -1,   7,  14, -16,  10,   0,  -8,  -5,   1,  -1,   8,  -1,   4,  -5, -14,  27, -19,  -1,   8,   7,   2,  -2,   1,   0,  -3,   7, -21, -23,  25,   1,   0,  -2,   1,  -3,  13,  -1,  -5,   8,  34,  11, -26,  -3,  -9,   1,   4,  -1,   9,  -5,   6,  -2, -32,   5,  -1,  -7,  15,  -4,  -9,  -8,  19, -17,   6,   6,   0,   3,  14,  10,   2,  23,   0, -13,   8,   3, -12, -12,  -9, -17,  19,   4,  24, -14, -11, -57,  -7,  20, -26,  10,   2, -14,   5,  -7,  -4,   1 },
     {   -1,  -3,  -5,  -2,  10,  -5,  -4,   8,  -1,   7,  -1,   2,  -2,  -4,  10, -59,  -4,  23,   4,   3,  -2,   1,  -2,   1,  -1,  -5, -16,  28,   2,   2,   1,   4,   2,   0,   1,  -2,   2,  -9,   5,  28,  10,   2,   0,  -7,  -2,  -2,   1,   2,  -2, -15,   3, -33,   7,  13,  10,  -3,   3,  -6,  46,   0, -15,  -8,   1,  -1,  -2,  -9,  13,  -2,  -1,   3,  -3,   1,   9,  -9,  24,  15,  -5,   8,  -4,   1, -17, -45,  -4,  14, -30,   8,   1,   3,  -1,   8,   0,   4, -25,  -1,   2,  -6 },
     {   -2,   0,  -5, -16, -19,  -7,   1,  -7,   1,   6,   1,  -1,  -4,  -2,   2, -81, -35,  17,  13,   8,   4,   4,   1,  -1,  -1,   5,  -6,  -5,  -7,  14,  -8,  -1,  -3,   0,   0,   1,   0,   2,  12,   9,   2,  -3,   2,  -6,   1,   3,   0,  -1,   1,  12,  24,   1,  -8,  -6,  -6,  -2,  -4,   3, -15,  14,   7,  -5,   2,  -1,  -1,  18,  -3,   9,   5,  -3,   1,   0,  -6,  12, -22,   4,   4,   4,   0,   3,   7,  42, -17, -12,  29,  -4,   3,   3, -11, -11, -10,   1,  22,  16,   3,  -4 },
     {   -1,  -2,  -2,  -3,   2,  -9,  -3,  -5,  -1,  -4,   0,   2,  -1,  -3,  -9,   6, -18,  -7,   1,  -3,   1,   7,  -1,   0,   1,  -3,  28, -11,  -5,  -4,   1,  -1,  -5,   5,   1,   1,  -1,  -5,  12, -26,   4,   1,  -4, -10,   2,   3,   1,   0,   3,   2,  85,   9,  -6,  -9,  -4,  -3,   4,  25,   6,  10,   8,   2,  -2,  -4,  -4, -14, -54, -13,   5,   5,   7,  -1,   4,   3,  23,   7,  -2,  -7,  -3,   1,  -1, -20,  -3,  14, -18,   2,   0,  -6,   1,   9,   0,  -1,  -8,   9,   0,  -1 },
     {    1,   5,   1, -36,  65,   4,  13,  -3,  -4,  -4,   5,   0,   0,  -1,  -1,  13, -37,   1,  -6,  -3,   8,  -2,  -2,   2,  -3,  -6,  -6, -57, -34,  24,   8,  13,  -6,  -6,   0,   0,   1,   3,   2,  -1,  18,   2,  -4,  -5,   2,   4,  -1,  -1,   0,  -6, -12,  14,  -5,  -1,   0,   2,   3,  -1,  41,   7,  -6, -10,   7,   8,   2,   6,  13,   7,   5,  -3,  -8,  -6,  -1,   0,   3,   4,   7,  -4,  -2,  -7,   3,   7,  -1,   7,   1,  -9,  -3,  -1,  -3,   1,  -3,  -1,   0,  -2,  -1,  -2 },
     {    4,   5,   5,  16,  55,  17,  -3,  -9,   5,  -8,   8,  -2,  -1,  -1,  -7,  -8,  -4,  11, -23,  -2,  -2,  -2,   6,   2,   1,   1,   2,  46, -21,  -9, -19,  13, -13,   9,  -1,   1,   0,  -3,  17, -20, -23,  26,   5,   4,   3,  -7,  -3,   0,   0,   5,  20, -11, -29,  -7,  13,  -2,  -2,  -1, -16,  -8,  19,  10,  -7,   4,   1,   9,  18,  26, -24,  -3,  -1,   4,   0,   2, -31,   2,   6,   1,   1,   3,   7,  -2,  22,  16,  -1,  -1,  -7,  -3,   3,  -1,   3,   1,  -4,  -5, -11,  -1 },
     {    1,  -1,  -1,  10,   8,   2, -16,  -3,  10,   2,   4,   3,  -1,  -3,  -1,  -9, -30,  -4,   4, -20,  12,   3,   5,   2,   1,   0,  -1,  17,  13, -16,   9,  13, -14,  -1,  -4,  -5,   0,   0,  -2, -10,   7,  -9,   7,   5,  -9,  -3, -11,  13,   0,  -4, -16,  20,   6,  -9, -18,  -4,   3,  -3,  12, -15,  25,  29,  12,  -5,  -2,  -4,   6, -50, -30,  -6,  13,  23,  -2,   1,  14,  28,   7, -27, -31, -28,   0,   7, -11, -14,  14,  -5,  22,   0,  -1, -13,   5,   4, -11,  -9,  -9,   1 },
   },
   {
     {  -84,  66, -23,  14,  -6,   4,  -2,   2,  -1,   1,  -1,   0,  47, -40,  15,  -8,   3,  -2,   0,   0,   0,   0,   0,   0,  -4,   3,   1,  -2,   1,  -1,   1,   0,   0,   0,   0,   0,  -3,   3,  -3,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -77, -10,  47, -19,  16,  -6,   5,  -2,   3,  -1,   2,  -1, -31,  48, -36,  19, -10,   5,  -3,   2,  -1,   1,  -1,   0,  41, -24,   5,  -3,  -1,   1,  -1,   1,  -1,   0,   0,   0,  -9,   4,   2,  -2,   2,  -1,   1,   0,   0,   0,   0,   0,   3,  -2,  -1,   1,  -1,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   0,   0,   4,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    4,  41, -48,  27, -18,   9,  -7,   3,  -3,   1,  -1,   1, -86,  10,  30, -13,  11,  -3,   4,  -1,   2,   0,   1,   0,  40, -26,   5,  -4,   2,  -2,   1,  -1,   0,   0,   0,   0,   8,   5,  -6,   3,  -3,   1,  -1,   0,   0,   0,   0,   0,  -3,   2,   1,  -1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },
     {   33,   4, -14,   1,  -2,  -1,   0,  -1,   0,   0,   0,   0,  44,  -7, -12,   6,  -6,   3,  -2,   1,  -1,   0,  -1,   0,  56, -46,  25, -15,   7,  -4,   1,  -1,   0,   0,   0,   0, -66,  42, -13,   8,  -1,   0,   1,  -1,   1,   0,   0,   0,  12, -11,   2,   0,  -1,   1,  -1,   1,  -2,   2,  -1,  -1,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,  -5,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0 },
     {   -9, -33,  17,   4,   0,   2,   0,   1,   0,   0,   0,   1,   1, -30,  35, -25,  17,  -9,   6,  -3,   2,  -1,   1,  -1,  65,  38, -57,  24, -18,   6,  -5,   1,  -2,   0,  -1,   0, -31,  -8,  22,  -7,   4,   0,   0,   1,   0,   1,   0,   0, -12,   2,  -1,  -1,   3,  -2,   2,  -1,  10,  -4,  -1,   2,  -2,   1,  -1,   0,  -1,   5,  -2,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
     {  -16, -10,   2,  10,  -5,   5,  -1,   2,  -1,   1,   0,   0, -33,   6,  13, -10,   8,  -4,   3,  -2,   1,   0,   1,   0, -52,  22,  -4,   4,  -1,   1,   0,   0,   0,   0,   0,   0, -68,  17,   1,   4,   0,   2,   0,   1,   1,   0,   0,   0,  71, -30,   2,  -4,   0,  -1,  -1,   0, -10,   7,   0,   0,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   7,  -3,  -1,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   1,   0,   0,   0 },
     {  -15, -37, -10,  51, -43,  37, -22,  16,  -9,   7,  -4,   4,  34,  49, -14, -34,  27, -25,  13, -10,   4,  -4,   1,  -2,  -3, -19,  10,   1,   1,   0,   1,  -1,   2,  -1,   1,   0,   5, -15,  10,  -1,  -1,   3,  -2,   2,  -1,   1,   0,   0,   1,  16, -11,   3,   0,  -1,   1,  -1,   0,  -3,   3,  -1,  -1,   1,  -1,   1,  -1,   2,  -1,  -2,   1,  -1,   1,  -1,   1,  -2,   1,   1,   0,   1,   0,   0,  -1,   2,  -1,   0,  -1,  -2,   1,   0,   0,   1,  -1,   0,   0,  -1,   0,   0 },
     {    3,  24, -10,   0,  -1,   0,  -1,   0,  -1,   0,   0,   0, -11,  24, -15,   6,  -5,   1,   0,   0,   0,   0,   0,   0, -32,  10,   2,   4,  -2,   2,  -2,   1,   0,   0,   0,   0, -63,  -6,  24,  -8,   8,  -3,   3,  -1,   1,   0,   0,   0, -67,  11,  -3,   6,   0,   1,   1,   0,  59, -24,  -1,  -1,  -1,   0,  -1,   0,   2,   8,  -2,   0,  -1,   0,   0,   0,   1,  -4,   2,   0,   1,   0,   0,   0,  -3,   1,   0,   0,   7,  -4,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0 },
     {   23,  46, -14, -10,   1,  -5,  -1,  -2,  -1,  -1,  -1,   0,  12,  19, -28,  13,  -9,   4,  -2,   1,   0,   0,   0,   0,  17,   1, -16,  14,  -9,   6,  -5,   3,  -2,   1,  -1,   1, -10, -54,  51, -25,  16,  -8,   5,  -1,   2,   0,   0,   0,  48,  21, -29,   9,  -7,   0,  -1,  -2, -21,   4,   7,  -1,   0,   1,  -1,   1,  -7,   0,  -2,   1,   0,   0,   0,   0,   6,  -2,   1,  -1,   0,   0,  -1,   0,  -1,   2,  -1,   0,  -2,   1,   1,   0,   0,   1,  -1,   0,  -1,   0,   0,   0 },
     {  -11, -27, -22,  27,  -4,   4,   2,   1,   1,   1,   1,   1,  -5,  -7,  -3,  24, -25,  21, -14,   9,  -6,   4,  -3,   1,  18,  49,  17, -55,  36, -30,  15, -10,   4,  -4,   1,  -2, -19, -39,   8,  17,  -9,   7,  -1,   1,   1,   0,   1,   0, -11,  10,  -3,   2,  -3,   3,  -3,   2, -22,  10,  -3,   0,   3,  -3,   2,  -1,  15,  -7,   3,  -2,  -1,   1,  -1,   0,  -2,   3,  -1,   1,   0,  -1,   1,   0,   1,  -3,   2,   0,  -2,   1,  -1,  -1,   2,  -1,   1,   0,  -1,   1,   0,   0 },
     {    5,  16,   5,  -7,  -3,   0,  -2,   0,  -1,   0,   0,   0,   0,  11,  -4,  -7,   9,  -7,   5,  -4,   2,  -1,   0,   0, -12,   3,  -9,  16, -10,   7,  -4,   3,  -1,   1,   0,   0, -20,  15,   1,  -4,   2,   0,   0,   0,   0,   0,   0,   0, -46,  13,  10,  -5,   5,  -2,   2,  -1, -88,   0,   9,   3,   2,   1,   1,   1,  56, -26,  -3,   0,  -2,   0,  -1,   0,   0,   6,  -1,   0,   0,   0,   0,   0,   5,  -2,   0,  -1,  -4,   0,   2,   0,   5,  -4,   0,   0,  -1,   0,   0,   0 },
     {  -12, -27,  12,  -5,  12,  -9,   8,  -5,   4,  -2,   2,  -1, -15, -39,  30,  -7,   3,   1,   0,   1,   1,   0,   1,   0, -17, -37,  28,  -5,   5,  -1,   1,   2,  -1,   2,   0,   1, -15,  -1,   5,  -5,   5,  -4,   3,  -2,   1,   1,   0,   0,   4,  67, -47,  19, -13,   4,  -3,  -1, -13, -30,  26,  -9,   5,   0,   0,   1,  -9,  -1,  -4,   0,   1,  -1,   1,   0,  12,  -5,   0,   1,  -1,   1,  -1,   0,   3,   4,  -3,   0,   2,  -2,   1,   0,   0,  -1,  -1,   0,   1,  -1,   0,   0 },
     {  -12, -21, -46,  15,  37, -32,  31, -17,  13,  -6,   6,  -3,  23,  52,  42, -24, -32,  24, -25,  11,  -9,   3,  -4,   0,  -8, -19, -22,  18,   2,   2,   0,   2,   0,   1,   0,   2,   1,  -1,  -1,   0,   0,  -4,   4,  -3,   2,  -2,   1,  -1,   0,   2,   9, -11,   5,   1,  -1,   0,  -6,  -5,  -1,   7,   0,  -2,   0,   0,  -2,   5,  -2,  -2,  -2,   3,  -1,   0,   2,  -4,  -1,   1,   1,  -1,   1,   0,   1,   2,   3,   0,  -1,   0,  -1,   1,   2,   1,   0,  -1,  -2,  -1,   0,   1 },
     {    1,  -6,   0,  -4,   4,  -3,   1,  -1,   0,   0,   1,  -1,   2, -11,  10,  -5,   0,   0,   1,   0,   0,   0,  -1,   0,   2, -21,   9,   2,   0,   0,  -1,   0,   0,   0,   0,   0,  11, -25,   5,  -2,   2,   0,   1,  -1,   0,   0,   0,   0,  23,  -7,  -4,  -1,   1,  -1,   0,   1,  41, -14,  -4,   0,  -2,   0,  -1,   1,  99,  -2, -12,   1,  -4,   0,  -1,  -1, -37,  25,   3,   0,   1,   0,   1,  -1,  -6,  -5,   1,   0,   0,   3,  -1,   0,   7,   1,  -2,   0,  -5,   6,  -1,   0 },
     {    9,  22,   2, -13,  -1,   0,  -3,   1,  -2,   1,  -2,   1,   8,  27,  -7,  -4,   0,  -5,   3,  -4,   0,  -1,   0,   0,  12,  36, -17,  -7,   6,  -3,   0,  -1,   0,  -1,   0,  -1,   4,  10, -35,  38, -26,  14,  -7,   4,  -1,   1,   0,  -1,  22,  19,  -5, -13,   7,  -4,   1,  -3,   3, -64,  42, -14,  11,  -3,   2,   0,  10,  23, -24,  10,  -5,   2,  -1,  -1,  16,  -2,   3,   0,   1,   1,   0,   0, -13,   7,   0,  -1,  -2,  -6,   2,   0,  -3,   2,   0,   0,  -1,  -1,   2,   0 },
     {   10,  26,  35, -23,   6,  -9,   5,  -8,   4,  -3,   1,  -2,   5,  16,  23, -31,  17, -12,   4,  -1,   0,  -1,   1,   0,  -3, -14,   3,  -5,  11,  -9,   7,  -4,   4,  -2,   3,  -1, -20, -44, -12,  44, -28,  23, -13,  10,  -4,   2,  -1,   1,  -1,  29,   2, -20,  13,  -9,   5,  -2,   6,  31, -22,  13,  -8,   4,  -3,  -1, -16, -16,  20,  -7,   5,  -2,   0,   2, -20,   3,  -3,   2,  -1,   0,   0,  -1,   2,  -8,   3,   0,  -1,   4,  -2,   2,   0,  -4,   1,   0,  -1,   1,  -1,   0 },
     {    1,   8,   6,  13,  -9,  -2,   2,  -3,   1,  -1,   0,  -1,  -4,   6,  -6,   0,  -2,   8,  -8,   5,  -3,   2,  -2,   1,  -4,  13, -13,   1,   4,  -4,   4,  -2,   1,   0,   0,   1,   5,  28,   2,  -8,   0,  -3,   1,  -3,   1,  -1,   1,  -1,  -1,  19, -11,   9,  -6,   4,  -1,   2, -14,  -3,   9,  -2,   0,   2,   0,   1, -23,  29,  -2,   0,   1,  -1,   0,   0,-107,   4,  17,   0,   3,  -1,   2,   0,   2, -19,   0,   0,   0,  -1,   0,   1,   5,  -7,   1,  -1,  -3,   0,   1,   1 },
     {    8,  13,  38,  17, -29,   9,  -6,   0,  -5,   2,  -1,   0,  -2, -12,  -1,  -1, -14,  22, -19,  16,  -9,   6,  -4,   3, -13, -38, -53,   5,  43, -34,  31, -16,  14,  -5,   4,  -2,   5,  16,  29, -12, -17,   9,  -6,   1,   0,  -2,   1,  -1,   3,   1,  15,  -6,   2,   2,  -6,   5,   2, -14,  -2,   6,   1,  -4,   5,  -4,   7,   2,  -8,  -2,   2,   0,  -2,   1,  19,  -4,  -2,   0,  -1,   1,   0,  -1,   4,   5,   1,  -3,   4,   1,  -2,   1,  -2,   2,   0,   1,   2,  -2,  -1,   1 },
     {   -3,  -2,  -4,  -3,   6,   3,  -6,   6,  -4,   3,  -3,   1,  -1,   5,   5,   6,  -5,  -3,   3,  -1,   1,  -1,   1,   0,  -4,   2,  -7,  -1,   3,  -2,   3,  -3,   3,  -2,   1,   0,  -1,  18,   3,  -7,   1,   2,  -4,   2,  -2,   1,  -1,   0,   5,  15,  -4,   2,   0,  -2,   2,  -3,  -3,  15,  -6,  -2,  -1,   2,  -2,   1,  -3, -12,  11,  -3,   3,  -2,   2,  -1,   2,  31, -12,   3,  -3,   1,  -1,   0,-107,  -1,  21,  -4,  28, -27,  -5,   3,   7,   3,  -1,  -1,  -2,  -2,   1,   0 },
     {    5,  10,  13,  32, -25, -25,  40, -39,  27, -20,  10,  -6,  -4,  -2, -35, -32,  30,  12, -24,  24, -15,  11,  -3,   4,   6,  23,   8,  -1, -10,   7,  -3,   5,  -7,   3,  -3,   2,   8,  23,   1,   7,  -9,   1,   2,  -6,   4,  -3,   1,  -1,   2,   0, -29,  15,  -1,   1,  -1,   4,  14,  14,   4,  -1,  -2,  -1,   1,  -4,  10, -23,  10,  -2,   3,   0,  -4,   2,  15,  10,  -6,   3,  -3,   0,   1,  -2,  -8,   9,  -3,  -1,  -7,   4,   1,   0,  -5,  -3,   1,   0,  -2,   0,  -1,   1 },
     {   -3,  -5, -21,  -3,   3,  13, -13,  15,  -8,   7,  -3,   3,   2,   9,  -7,  27, -19,   3,   3,  -5,   3,  -3,   1,  -2,   2,  18, -24,  14,  -2,   2,   1,  -1,   0,   0,   0,   0,   7,  40, -12,  -2,  -6,   6,  -7,   3,  -2,   1,  -1,   1,  16,  53, -13,  -4,  -1,  -4,   0,  -2,  27,   3, -19,  12,  -5,   3,  -1,  -1,   8, -51,  35, -11,   5,   0,   0,   1,  -1,  16, -16,   3,   0,  -3,   0,  -1,  33,   7,  -1,  -2, -23,   6,   1,   0,  -2,  -5,   3,   1,  -5,   2,   0,   0 },
     {    7,  20,  19, -20,  12,   0,  -7,   8,  -6,   3,  -2,   1,  12,  35,  40, -20,  -2, -11,   7, -10,   5,  -5,   0,  -3,   5,  16,  -5, -20,  20, -19,  11,  -5,   3,  -1,   0,  -2,   7,  14,  -6, -13,   8,  -4,  -2,   0,  -1,   1,   1,  -1,  -7, -28, -34,  57, -38,  28, -14,   8,   4,  12,  -2, -20,  12,  -9,   4,   1,   8, -10,   5,  -5,   3,  -4,   3,  -1,   6,   5,  -4,   4,  -3,   2,  -4,   0,  10,   2,  -1,   4,  -2,   4,  -1,  -3,  -1,   1,  -1,   0,  -3,   3,  -1,   0 },
     {   -9, -15, -37, -43,   7,   9,   2,   1,   8,  -1,   5,  -1,  -2,  -3, -10,  -4,  15,  -3,  -1,   4,  -5,   0,  -4,   3,   2,   3,  23,  32,   2, -24,  23, -21,   6,  -3,   2,  -3,   6,  14,  38,  -6, -34,  34, -31,  17,  -8,   5,  -1,   1,  -9, -29, -38,  -2,  19, -13,   7,   1,  -7,  -5,   6,  -2,  -6,   2,   5,  -5,  -1,  15,  11,  -1,   3,  -3,   0,   0,  -2,   3,  -1,   5,  -2,   0,  -2,  -1,   6,  -2,   4,  -2,  -5,  -3,  -2,   1,  -3,   3,   4,   0,  -3,  -1,  -2,   1 },
     {    0,   5,  11,  -3,   1,  -1,  -4,  -4,   1,  -1,   0,  -2,  -2,   5,  14,  -4,   2,  -6,   3,  -1,  -1,   1,  -2,   1,  -3,   6,  12, -16,   2,  -4,   4,  -5,   4,  -3,   3,  -2,  -2,   9,   9, -16,  12,  -9,   4,   0,   0,  -1,   2,  -2,  -1,   8,   6, -10,   7,  -2,   3,  -4,  -3,   5,  -9,  19, -16,   8,  -7,   5,   3,  39,  -5,  -3,   3,   0,   2,  -4,   7,  17,   1,  -2,  -3,   0,  -1,   3, -22,  33,   2,   3, -96,   3,  11,   3, -19, -14,   7,  -1,  -5,  -4,   0,  -1 },
     {   -4,  -8, -16,   0,   1,   4,  -4,   6,  -4,   2,   0,   1,  -5, -10, -15,  12,   1,   8,  -4,   4,   0,   2,   1,  -1,  -7, -14, -21,  32,  -8,   1,   4,  -3,   2,   2,  -1,   2,  -8,  -7, -17,  30, -12,  10,  -9,   5,  -4,   4,  -1,   1,  -7,  15,  -3,  11, -21,  13, -13,   6,   9,  58,   3, -46,  26, -19,  13,  -7,  10,  12, -25,  17, -13,   5,  -2,   0,   0, -22,  20,  -5,   0,   0,  -3,   2, -13,  14,   2,  -2, -29,  -1,   9,  -4,   2, -10,  -2,   3,   4,  -2,   2,  -3 },
     {   -1,   0,  -4, -31, -24,  27,  -2,  -5,  12,  -7,   7,  -4,   4,  11,  12,   3,  13,   4, -23,  23, -19,  12, -10,   6,   8,  19,  21,  26, -11, -19,  23, -19,   8,  -4,   0,   0,  -6,  -3, -41, -25,  24,  -6,   8,  -4,   4,  -3,   4,   0,   9,  27,  13,  -3, -14,   7, -13,   7,   1,   6, -15,  20, -15,   9,   0,  -3,  13,  14,  -7,  -9,  19,  -9,   4,   1,   9, -27,  20,  -8,  -4,  -2,  -1,  -1,  -6,   2,  -9,   6,  24,   8,  -4,  -3,  -2,   5,  -1,   1,   4,   0,  -2,   1 },
     {   -2,  -5,  -6, -12, -10,  18,   2,  -8,  12, -10,   7,   0,   1,  -1,   5,  14,  14, -11, -12,  18, -15,  11,  -6,   3,   0, -11,  -2,  12,  -9,  -2,  11, -13,   6,  -7,   3,   0,  -6, -30, -13,  -1,  13,  -7,   5,  -1,   1,  -2,   1,   1,  -3, -17,  27,  -1, -10,  15, -14,   9,   0, -26,  33, -18,  15,  -8,   7, -10, -24, -31,   9,  -2,   5,   1,  -1,   3, -17,  42, -36,  18, -11,   4,  -1,   0,   6,   5,  12, -11, -25, -18,   7,   2,  15, -19,  -1,  -1,  17,   4,  -3,   0 },
     {   -1,  -2,   1,   5,  -4,   3,  -4,   3,   3,  -1,   1,   1,   0,   1,   5,   7, -11,  -1,   0,   2,  -6,   3,  -1,   0,  -2,  -3,  -2,   7,  -4,   3,  -3,  -1,   0,   1,  -2,  -1,  -4,  -7,  -8,   1,   4,   4,  -3,   2,  -2,  -1,  -1,   1,  -2,  -1,   0,   5,   2,  -4,   4,  -2,  -2,   6,  13, -12,   2,   0,  -2,  -1,   1,   0,  -7,   3,   3,  -2,   2,   2,  -3,  25, -11,  -3,  -1,   0,   0,   0,  -1, -17,  13,  -3,  16,  40, -10,   3,-102,  -3,  27,  -5,  13, -24,  -7,   0 },
     {   -6,  -8, -15, -36, -32,  22,  13, -24,  23, -14,   4,  -5,   5,  11,  26,  37,  14, -38,  -6,   4, -20,  11,  -4,   2,  -7, -10, -25, -36,  15,  23, -17,  21,  -5,   2,  -4,   3,   3,  10,  19,   5, -14,  -2,  15,  -1,   5,  -8,   6,  -7,   0,   5,  -4,   2,   6,  -1,   4, -11,  -1,  10, -11,  -5,  -1,  -8,   4,  -1,   3,   5,   0,  22, -12,  11, -12,   6,  -4,  -5,  -3, -10,   9,   1,   6,   0,   4,  11,  -6,  -6,   9,  -7,   7,  -1,  -8,  15,   1,   3, -15,   4,  -1,   0 },
     {    2,   2,  12,  22, -22,  -3,   4,  -1,   4,   1,   1,   2,   4,   4,  21,  42, -22,  -1,  -4,   5,  -4,   9,   0,   5,  -3,  -9, -13,   4, -16,  14, -11,   6,  -2,   3,  -5,   4,  -4, -11, -24,  -2,  -8,  27, -16,  15,  -5,   7,  -2,   3,  -5, -13, -17,  35,  17, -22,  17,  -9,  -1,   9,  20,  -5, -17,  26, -14,   1,   8,  21,   6, -25,  33, -22,   3,  -1,   5, -12, -19,   8,  -7,  10,   0,  -1,  -4,  -7, -10,   7,  -5,   6,   7,  -5,  24, -13,   4,  -6,   7,   6,  -6,   4 },
     {   -2,  -4, -12,   9,   8,  -8,   7, -11,   2,  -3,   7,   0,  -3,  -9, -21,  16,  15, -21,  15,   3,   1,  -6,   7,   0,  -3,  -9, -18,  11,  10,  15,  -1,  -2,   0,   3,  -3,   1,  -1,  -6,  -2,  31,  -7,  -1,  -1,  -5,  -2,  -4,   3,   2,   0,  -4,  -5,   6, -37,  31, -23,  12, -15, -30, -44,  27,  10,  -7,   4,  -6,   9,  38,  27, -35,   9,  -8,   8, -11,   5,  -6, -15,  11,  -3,   4,   0,   4,  -7, -16,   7,   1,  -4,   4,  -4,  -7, -10,   4,   6,   4,   1,   3,  -4,   2 },
     {    1,   2,   5,   3,   3,  -8,  -6,  18,   2,  -7,   3,  -5,   0,  -1,   2,  -2,  -7,  -5,   4,  -2,   0,  -3,   0,   8,   1,   1,   3,  -1,   2,  -2,   1,  -5,  -2,   3,   2,   1,   1,   0,   7,   3,   0, -18,   5, -15,   9,  -9,   9,  -8,  -1,  -7,  -8, -34,  19,   7, -17,   4,   1,   6,  20, -25,   0,  -2,   7,   1,  11,  -3,  33, -26,   0, -12,  13,  -1, -20, -38, -21,  19,  -8,  20, -17,  13,  -8,  56, -35,   5,  18, -10,  10, -10, -24,  -4,   2,   7, -12,   0,  -3,   3 },
   },
   {
     { -110,  63,  -5,   7,  -1,   3,   0,   1,   0,   1,   0,   0,  -8,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   5,  -4,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   4,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0 },
     {  -10,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0, 111, -52,  -2,  -4,  -1,  -2,  -1,   0,   0,   0,   0,   0,  31,  -7,  -4,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  -3,   7,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -8,   5,  -1,   0,   0,   0,   0,   0,  -7,   3,   0,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -2,   1,   0,   0 },
     {  -32, -61,  56, -13,  13,  -4,   6,  -2,   3,  -1,   2,  -1, -31, -27,  28,  -4,   6,  -1,   2,   0,   1,   0,   1,   0,  56, -30,   0,   0,  -1,   0,  -1,   0,  -1,   0,   0,   0,  33, -10,  -6,   1,  -2,   0,  -1,   0,   0,   0,   0,   0,  15,   0,  -6,   1,  -1,   0,  -1,   0,   4,   4,  -4,   1,  -1,   0,   0,   0,   0,   4,  -2,   1,   0,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,  -1,   0 },
     {  -35, -43,  44,  -9,  10,  -2,   4,  -1,   2,  -1,   1,   0,   8, -25,  14,  -3,   3,  -1,   1,   0,   0,   0,   0,   0, -82,  34,  -1,   4,   1,   2,   0,   1,   0,   0,   0,   0, -43,  14,   1,   2,   1,   0,   0,   0,   0,   0,   0,   0,  -7,   0,   0,   0,   0,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0,   9,  -4,  -1,   0,   0,   0,   0,   0,   5,  -2,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   3,  -1,   0,   0,   3,  -1,   0,   0,   2,  -1,   0,   0 },
     {    8,  10, -13,   5,  -4,   2,  -2,   1,  -1,   1,   0,   0,  13, -31,  12,  -1,   2,   0,   1,   0,   1,   0,   0,   0, -59, -12,  15,   0,   4,   0,   1,   1,   1,   0,   1,   0,  78, -39,  -1,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0,  50, -10,  -7,   0,  -2,   0,  -1,   0,  20,   1,  -5,   0,  -1,   0,  -1,   0,   6,   5,  -3,   0,  -1,   0,   0,   0,   3,   4,  -2,   0,   0,   0,   0,   0,  -1,   4,  -1,   0,   0,   2,  -1,   0,  -2,   2,   0,   0,   0,   1,   0,   0 },
     {  -18, -28,  23,  -4,   5,  -1,   2,   0,   1,   0,   1,   0,  30,  64, -50,   9, -11,   2,  -4,   0,  -2,   0,  -1,   0,   2,  52, -31,   4,  -6,   0,  -2,   0,   0,   0,  -1,   0,  49,   8, -12,  -1,  -1,  -2,  -1,  -1,   0,   0,   0,   0,  18,   8,  -3,  -2,   0,  -1,   0,  -1,  -2,   3,   2,  -1,   1,  -1,   1,   0,  -9,   3,   3,  -2,   1,   0,   1,   0,  -8,  -1,   3,  -1,   1,   0,   0,   0,  -7,   0,   2,   0,  -5,  -1,   2,   0,  -3,   0,   1,   0,  -2,  -1,   1,   0 },
     {    1,  -5,   0,   2,   0,   0,   0,   1,   0,   0,   0,   0,  12,  14, -16,   3,  -4,   1,  -1,   0,   0,   0,   0,   0,  24,   0,  -6,   1,  -2,   0,  -1,   0,   0,   0,   0,   0, -48, -12,  13,  -1,   3,   0,   2,   0,   0,   0,   0,   0,  66, -43,   4,  -2,  -1,   0,  -1,   0,  64, -19,  -5,   0,  -3,   0,  -1,   0,  37,  -4,  -5,   2,  -2,   0,   0,   0,  13,   4,  -4,   0,  -1,   0,   0,   0,   7,   4,  -3,   0,   3,   3,  -1,   0,   1,   3,  -1,   0,  -2,   2,   0,   0 },
     {  -18, -41, -47,  65, -25,  21,  -8,   9,  -3,   4,  -2,   3, -23, -41, -24,  44, -12,  11,  -3,   5,   0,   2,   0,   1,   4,   7, -20,  12,  -4,   2,   0,   0,   0,   0,   0,   0,  -4,  19, -14,   1,   0,   0,   0,  -1,   0,   0,   0,   0,   8,  12,  -6,  -4,   1,  -2,   0,  -1,   3,  10,  -2,  -4,   1,  -2,   0,  -1,   0,   6,   2,  -4,   1,  -2,   0,   0,  -3,   4,   1,  -2,   1,  -1,   0,  -1,  -3,   3,   1,  -2,  -2,   2,   2,  -1,  -1,   1,   1,  -1,  -1,   1,   0,  -1 },
     {  -18, -39,  -2,  19,  -3,   5,   0,   2,   1,   1,   0,   0,  18,  33, -39,  16,  -8,   3,  -3,   1,  -1,   1,  -1,   0, -16, -55,  35,  -5,   7,  -1,   4,   0,   1,   0,   1,   0,  -4, -52,  35,  -9,   6,   0,   2,   1,   1,   0,   1,   0, -38, -12,  12,  -2,   4,   2,   0,   2, -14,  -9,   4,   0,   1,   1,  -1,   1,   5,  -4,  -3,   2,  -1,   0,   0,   0,   8,  -2,  -2,   1,  -1,   0,   1,   0,   6,   0,  -2,   0,   5,  -1,  -2,  -1,   4,   2,  -2,   0,   1,   0,  -1,   0 },
     {    4,   1,  -5,   0,  -1,   2,  -2,   0,  -1,   0,   0,   0,  -3, -13,   4,   1,  -1,   1,   0,   1,  -1,   0,   0,   0,  10,  22, -17,   5,  -5,   1,  -1,   0,   0,   0,   0,   0,  40,   5, -10,   1,  -2,  -1,  -1,  -1,   0,   0,   0,  -1, -64,  -5,  12,   0,   3,   0,   2,   0,  35, -46,   8,   0,   1,   0,   1,   0,  59, -22,  -3,   1,  -1,   0,  -1,   0,  38,  -7,  -5,   0,  -2,   0,   0,   0,  17,   1,  -3,   0,   8,   2,  -1,   0,   4,   1,  -1,   0,   3,   1,   0,   0 },
     {    5,  15,  16, -26,  11,  -9,   3,  -2,   1,   0,   1,  -1,   0,   5, -35,  19,  -4,   5,  -3,   3,   0,   2,   0,   1, -20, -60,  -8,  27,  -5,   7,   0,   3,   0,   2,   1,   1,  -4,  26, -29,  14,  -6,   2,  -1,   0,   0,   0,   0,   0, -16,  41, -25,   3,  -3,   1,   0,  -1,  47,  17, -15,  -3,   0,  -2,  -1,  -1,  16,  21,  -6,  -4,   0,  -2,   0,   0,  -3,  16,   2,  -3,   0,  -2,   1,   0, -12,  11,   2,  -1,  -6,   4,   2,  -2,  -4,   3,   2,  -1,  -2,   1,   1,  -1 },
     {   -8, -17, -15,  21,  -7,   6,  -3,   2,  -1,   1,   0,   1,   7,  18,  25, -26,   6,  -6,   3,  -3,   1,  -2,   1,  -2,   3,  15,  34, -32,  12,  -8,   3,  -5,   1,  -2,   0,  -1,  10,  19,  11, -13,   1,  -2,  -1,  -1,  -1,   0,   0,   0, -33,  -2,   4,   1,   1,   1,  -1,   2,  72,   6, -15,   2,  -5,   1,  -3,   0, -27,  30,  -4,   2,  -1,   0,   0,   0, -32,   8,   1,   0,   1,   0,   1,   0, -24,   2,   3,   0, -10,  -2,   1,   0,  -8,  -1,   1,   0,  -3,  -2,   1,   1 },
     {   -7, -17, -12,  22, -10,   7,  -2,   3,  -1,   2,  -1,   0,  14,  34,  20, -31,   2,  -6,   3,  -3,   0,  -2,   0,  -1,   0, -13,  40, -25,  11,  -6,   2,  -2,   0,   0,   0,  -1,   0,  38, -19,  -3,  -2,  -1,  -1,   0,  -1,   0,  -1,   0,  29,  46, -24,   5,  -5,  -3,  -1,   0, -22,   7,  -7,   4,   2,   0,   1,   0,  45, -17,  -8,   0,  -2,   1,   0,   1,  33,  -3,  -7,   0,   0,   1,  -1,   0,  18,   0,  -2,   1,   5,   1,  -1,   1,   2,   2,   0,   0,  -2,   0,   0,   1 },
     {   14,  22,  57,  30, -59,  19, -16,   5,  -9,   2,  -3,   1,  12,  22,  44,  28, -47,  12, -13,   3,  -5,   1,  -2,   1,  -2,  -7,   9,  20, -15,   3,  -4,   2,  -1,  -1,   0,   0,   0,   3,  -6,  10,   0,  -3,   0,   0,   0,  -1,   1,   0,  -2,  -7,  -1,  -2,   7,  -4,   4,  -2,   5,  -7,   0,  -1,   4,  -2,   3,   0, -10,  -6,  -1,  -4,   5,  -1,   1,   1,  -2,  -4,  -3,  -2,   2,   0,   0,   1,   0,  -2,  -3,   0,   4,  -2,  -3,  -2,   2,  -1,  -1,  -1,   1,  -2,  -1,  -1 },
     {    1,   5,   4,   0,  -8,   2,  -3,   2,  -1,   0,   0,   0,  -3,  -1,  13,   1,  -9,   2,  -3,   2,   0,   1,   0,   1,   8,  30,  -4,  -4,  -4,   0,  -1,   0,  -1,  -1,   0,   0, -12, -49,  12,   5,   3,   3,   0,   2,   0,   0,   0,   1,  -9,  23, -15,   3,   0,   1,  -1,  -1, -23,  10,  -9,   0,   0,   0,  -1,   0,  67,  21, -17,  -2,  -4,  -1,  -2,  -1, -37,  42,  -3,   0,  -1,   0,   1,  -1, -36,  13,   3,   0, -23,   5,   4,  -1, -10,   0,   2,  -1,  -6,  -1,   2,   1 },
     {    9,  20,  26, -18,  -2,  -8,   2,  -2,   0,  -1,   0,  -1, -11, -29, -30,  42, -16,   8,  -3,   3,   0,   1,   1,   0,  10,  24,  40, -29,   2,  -9,   0,  -4,  -1,   0,   0,   0,  10,  23,  53, -37,   8,  -8,   1,  -3,  -2,   0,  -1,  -2,  10,  31,   9, -18,   2,  -1,  -2,   0,   8,  16,   5,  -6,   0,   0,  -2,   1,  11,   6,  -4,  -2,   0,  -2,  -1,   0,   7,  -8,   1,   0,   0,  -2,   1,   1,  10, -11,  -2,   0,   9,  -4,  -3,   1,   7,  -5,  -2,   2,   1,  -1,   0,   1 },
     {   -1,   2,   7,  -1,   1,  -1,   0,   0,   1,  -2,   1,  -1,  -3,  -2,  -2,   8,   1,   0,  -1,   0,   1,   0,   1,   0,  -4,  -8,  11,   1,   3,  -3,   0,   1,   0,   1,   0,   0,   6,  34,  10, -13,   0,  -1,  -2,  -1,  -2,   0,  -1,   0,   4, -24,  21,  -8,   2,   3,  -1,   0, -18, -14,   7,   2,  -1,   4,   1,   1, -11, -13,  -7,   7,  -4,   3,   0,   1,  40,  27, -22,   4,  -4,   0,  -2,  -1, -38,  51,  -6,   0, -55,  21,   7,  -3, -39,   3,   8,  -1, -15,  -6,   4,   0 },
     {    0,  -1,   2,   1,   3,   0,   1,   0,  -1,  -2,   0,   0,  -2,  -3, -11,   0,   2,   4,   3,   0,   0,  -1,   0,   1,  -8, -25,  13,  -6,   6,   4,   1,   1,  -1,  -1,   1,   1,   9,  38, -10,  -2,   2,  -4,   3,  -2,   0,   0,  -1,  -1,  16,  -6,  23,  -4,   1,  -4,   2,  -2, -13, -57,  29,   2,   0,  -1,   4,   0,  25, -24,   9,   5,  -1,   3,  -2,   1, -73,  -1,  12,   4,   5,   1,  -1,   2, -31, -15,   4,   1,   0, -11,  -1,   1,  15,  -7,  -2,  -1,   4,   0,  -2,   0 },
     {   -7, -10, -30, -21,  36, -20,  13,  -2,   6,  -3,   3,  -2,   2,   1,  12,  34, -38,  17,  -7,   5,  -5,   2,  -1,   2,  10,  24,  62,   6, -28,  15, -13,   3,  -5,   1,  -3,   0,  -5,  -9, -21,  31,  -7,   6,  -3,  -1,   0,   0,   0,   0,   1,  -1, -26,  23,   0,  -2,   1,  -2,   1, -28,  -3,   5,   4,  -1,   4,   0, -17,  -6,  -4,   1,   2,   3,   0,   2,  -2,  -3,  -4,  -1,   3,   1,  -4,   2,   1,   3,  -4,   1,  -4,   5,   0,   0,  -3,  -1,  -1,   1,   2,   0,  -1,  -3 },
     {    8,  15,  33,   7, -31,  13,  -6,   2,  -6,  -2,  -1,   2, -12, -23, -52, -12,  16,   0,   8,  -1,   4,  -2,   2,   2,   4,  13,  15, -39,  28,  -5,   3,  -8,   2,  -2,   1,  -2,  -6, -14, -31,  14,   3,  -3,   0,  -1,   2,   2,   1,  -1,  -6, -13, -44,  34,  -4,  -4,   2,  -2,  -4, -26,  -6,  18, -11,  -2,   7,   0, -17,   1, -12,   6,   4,   1,  -1,   2,   1,  -2,  -6,   5,   3,   2,  -5,   1,   4,  11,  -9,   5,  -3,   7,   3,   5,   4,  -2,  -1,   2,  -1,   2,   1,  -1 },
     {    1,  -1,   1,   2,   0,   1,  -2,   1,  -1,   1,  -1,   2,   1,  -3,   0,   0,   2,   0,   2,   1,  -2,   1,   0,   0,   4,   3,  -2,  -6,   3,  -3,   0,  -2,   0,   0,   0,  -1,  -6, -21,   5,   7,  -1,   4,  -2,   1,  -1,   0,  -1,   0,   2,  20, -15,   7,  -5,   2,   0,   1,  18,  10,  14,  -6,   1,  -3,  -2,   0, -10, -45,  26,  -7,   4,   0,   1,   1,  39,  -9,   4,   0,  -2,   0,   1,   2, -79,  -7,  21,   0,  -1, -40,   3,   3,  24, -17,  -3,   3,  22,  -7,  -2,   2 },
     {    1,   2,   4,  11,   0,  -1,  -2,   1,  -1,   2,  -2,   1,  -1,  -1,   5,  -8,   8,  -4,  -3,   1,   0,   0,   0,  -1,   2,   0, -12,  -3,   0,  -2,   1,   0,   1,   2,   0,   0, -12, -34,  10,   9,   8,  -1,   3,  -4,   0,   0,   0,   1,  12,  70,   8, -12,   0,  -6,   4,  -5,  21, -43,  30,  -8,   5,  -2,   3,  -1, -22,  -5,  20,  -4,   3,   0,  -1,   1,  -2, -22,  19,   2,  -2,   2,  -3,   2,  23,   2,  -5,   3, -43,  26,   1,   3, -22,   3,   3,   0, -15,   0,   6,  -1 },
     {   10,  17,  25,  63,  31, -49,   8, -19,   5,  -8,   3,  -6,   6,   8,   9,  41,  40, -38,   7, -15,   2,  -9,   0,  -4,   2,   6,   7,   7,   9,  -4,  -3,   1,   2,  -2,  -2,  -1,  -1,  -4,  -8,  -7, -13,  17,  -8,   8,   2,   0,   1,   1,  -3,  -3, -13,  -8, -10,   8,  -3,   2,  -2,  -8,  -6,  -6,  -4,   3,   1,   4,   2,  -4,  -5,  -6,  -1,  -1,   1,   2,  -3,  -1,  -1,  -2,   2,   0,  -3,   0,  -4,  -4,   3,  -4,  11,  -2,  -3,  -1,  -6,  12,  -4,  -1,  -9,   2,   2,  -1 },
     {    1,   1,   7,  10,  -7,  17, -11,  -1,   1,  -1,   1,   2,  -3,  -4, -14, -12,  17,  -6,   4,  -3,   2,  -1,  -2,   4,   0,   1,   7,  14, -40,   1,   1,   4,  -2,  -5,   1,   3,   6,  13,  46,  34, -62,  27, -18,   6,  -5,   0,  -2,   1,   2,   5,  11,  29, -26,  11,  -9,  -2,   1,   8,  -8,  25, -11,   9,  -7,  -3,   9,  -3,   0,  22,  -2,  -5,   4,   0,  -9, -22,   3,   2,  -1,  -1,   5,  -4,   5,  -2,  -7,   4, -15,   0,   4,   0,  -4,  -8,   5,  -1,  12,  -8,  -3,   1 },
     {    2,   2,   4,   5,   9,   1,  -1,  -4,   1,  -8,   2,  -2,   1,  -1,  -1,   4,   4,   6,   3,  -5,   1,  -4,   2,  -2,   1,  -2,  12,  10,   6,  -3,   4,  -2,   3,  -1,  -2,  -1,   4,   5,  -4, -11,  14,  -6,   7,  -2,   2,   0,   0,   0,  -9, -34, -18,  18,  12, -13,  11,   2,   1,  36,  -8,  -6,  11, -10,   9,  -4,  28, -15,  27, -23,  14,   0,   1,   0, -13, -41,  23, -13,  13,   2,  -3,   3,  20, -11,  17,  -5, -61,  -1,  22,   2, -12, -21,  10,   2,  -1, -13,   3,   0 },
     {    1,   1,   9,   1,  -7,  -1,  -1,   7,  -3,   3,  -5,   2,   0,  -1, -11,   1,   8, -11,   3,   5,   0,   5,  -3,   1,   2,   7,  32, -22, -11, -11,  -1,  -3,  -4,   1,   0,   1,  -4,  -9, -48,  33, -22,   8,   0,   8,  -2,   2,   1,   1,  -1, -14,  36, -39,  -2,   7,  -3,   2,   0,  21,  42, -25,   1,   2,  -8,   1,  10,  31,  14,  -9, -11, -11,   4,  -8,  11,  13,  19,   1,  -9,  -7,   4,  -3,  14,  13,   4,  -1, -11,  -9,  10,  -3,  15, -19,  -3,   4,  12,   1,  -2,   4 },
     {    3,   4,   3,  15,  10,  -8,   5,  -5,   4,  -3,   2,  -3,   1,   0,   1,  -4,  -1,  13,  -1,  -7,   1,  -3,   2,   0,   2,   1,   0,  25,   4,  -6,   1,  -3,   2,  -3,  -2,  -7,  -1,  -5, -13, -11,  36,  -1,   1,  -4,   2,  -6,  -2,  -1,   7,  17,  27,  -6, -11,   7,  -5,   0,  -5, -21, -32,  28, -11,   6,  -4,   0,  -6,  24, -32,  14,  -9,   2,   4,  -1,   9,  -7, -10,   7,  -4,   1,   3,  -3,  10,   1,  -8,  -3, -31, -36,  28,   0,  26, -42,  -8,   1,  33, -10, -15,   4 },
     {    5,   7,  11,  30,   7, -17,  32,  -9,  -1,  -8,   1,   3,  -3,  -5,  -8, -29, -28,  30,  21,  10,  -2,   0,   6,  -5,  -7, -11, -18, -43, -41,  50, -22,  17,   4,   8,  -2,   1,  -2,  -4,  -9, -32, -11,  18,   0,  -4,  12,   2,  -1,   5,  -3,  -3,   1,  -6,  -2,   3,   7,  -3,   1,   7,   1,   2,   7, -10,   6,  -1,  -1,  -2,  -7,   8, -12,  -5,  -3,   3,   3,  -3,   6,  13,  -5,  -5, -13,  -3,   0,  -2,   8,   4,  -5,  -6,   5,   6,  -1,   1,   1,   8,  -8,  -3,   8,   2 },
     {    0,   2,  -1,  -3,  10,  13, -10,  -6,   1,   4,   2,  -1,  -2,  -1,   2,  -2,  -7,  -4,   4,  -1,  -5,   2,  -1,   0,   0,   2,   1,   6,  -2, -14,  -2,  10,  -2,   1,   0,  -3,   0,   1,   2,  -5,  -5,   7,  -3,  -3,   5,  -4,   3,   2,   0,  -1,   3,   2,   6,   3,   4,  -5,  -4,  -4, -10,   7,  13,   3,  -1,  -7,   2,  17,   2,   0,  -6,  -2,  -4,   3,   3, -32,  11, -11,   2,  -3,   3,  -4, -13,  36,  -8,  -2, -15, -15,   3,   0,  68,  31, -21,  -1, -63,  29,  15,  -1 },
     {   -5,  -6, -10, -11, -46, -24,  57, -40,  26,  -8,  10,  -2,  -2,  -3,  -3,  -2, -33, -16,  39, -28,  17,  -5,   2,   2,  -1,  -1,  -5,   1,   7, -13,  20, -15,  -4,   4,  -2,   0,   1,   0,   2,  17,  -3,  -2,  -1,  11, -18,   3,  -3,   0,   1,   4,  14,  -1,  -3,   4,  -1,  17,   0,  -1, -10,   5,  -1,   7,   2,   4,   2,   0,   8,  -6,  14,   1,   2,   2,   1,  -2,   7,   3,  12,   8,   1,  12,  -2,   4,   3,  -7,  -7,   6,   0,   1,   5,  14,  -4,  -3,   7,   7,   5,  -3 },
     {   -1,  -3,  -3, -11,  -5,  14,  -7,  -2, -10,   2,  -1,   3,   1,   1,   3,  10,  13, -18,  -8,  14,  -9,   8,  -6,  -5,  -2,  -5,  -4, -17, -22,  22,  19,   2, -11,  13,  -3,   4,   0,   6,   4,  -2,  -9, -11,  18,  10,  -9,   3,   0,  -1,   3,  -5, -12, -16,   2,  -4, -13,  23, -11, -32, -11,  23,  -9,   9,   0,  17,  -4,  27, -23, -15,  29,  -9,   7,  -6,  22,   8,  30, -27,  13,   3,  -2,  -7,  -5, -30,  43,  -4, -13, -20,   1,   2,  -9,  10,  -4,  -2,  -9,  -4,   5,  -7 },
     {   -2,  -3,  -6,  -7,   6, -14, -12, -12, -17,   5,  -3,  12,   2,   3,  14,  23, -29, -13, -11,  -2,   0,  -8,  -2,  -6,  -5, -11, -24, -24,  27,  -7,  -5, -22,  -7,  21,  -1,  -2,  -1,   0,  -4,  -5, -17,  -7,   0,   4,   3,  -3,   2,  -4,   3,   1,  24,  61, -30, -14,  -7,   8,  -1,   1,  20, -10, -12,  -6,  -2, -17,   4,  20,  -1,   3, -17,  -2,   2, -15,   8,  -1,  -2,   2,  -6,   0, -21,  -3,   3, -14,  -5,   6,  -4, -20,  11,  12, -10,  -7,  -4,   8, -21, -11,  -1,   3 },
   }
 },
 {  //21
   {
     {  116, -47,  -4,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0,   0,  15,  -2,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -13,   7,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -9,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -7,   3,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -2,   1,   0,   0,  -2,   1,   0,   0,  -1,   0,   0,   0 },
     {  -14,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0, 112, -32,  -6,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0,  33,   5,  -6,   0,  -2,   0,  -1,   0,  -1,   0,   0,   0, -26,  15,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -18,   5,   1,   0,   0,   0,   0,   0,  -8,   2,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0,  -3,   1,   1,   0,  -3,   0,   1,   0,  -2,   0,   1,   0,  -2,   0,   0,   0 },
     {  -13,  11,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,  20,   6,  -3,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   0, -98,  29,   7,   3,   2,   1,   1,   0,   1,   0,   1,   0, -60,  -6,  10,   0,   3,   1,   1,   0,   1,   0,   1,   0,  10, -20,   3,  -1,   1,   0,   0,   0,  17, -12,   0,   0,   0,   0,   0,   0,  16,  -5,  -1,   0,   0,   0,   0,   0,   7,  -2,  -1,   0,   0,   0,   0,   0,   6,  -1,  -1,   0,   3,   0,  -1,   0,   3,  -1,   0,   0,   2,   0,  -1,   0 },
     {   -4,  -8,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,  39, -11,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -43,   0,   5,   1,   2,   1,   1,   0,   0,   0,   0,   0,  76, -31,  -5,  -2,  -1,  -1,  -1,   0,   0,   0,   0,   0,  66,   9, -13,   0,  -3,  -1,  -2,   0, -14,  25,  -3,   1,  -1,   0,   0,   0, -19,   9,   2,   0,   0,   0,   0,   0,  -9,   1,   1,   0,   0,   0,   0,   0,  -3,   2,   0,   0,  -5,   2,   1,   0,  -2,   1,   0,   0,  -3,   1,   1,   0 },
     {    3, -11,   3,   0,   1,   0,   1,   0,   0,   0,   0,   0,  -7, -15,   4,   1,   1,   1,   0,   0,   0,   0,   0,   0,  36,  -9,  -5,   0,  -1,   0,  -1,   0,   0,   0,   0,   0, -36,  -4,   4,   1,   1,   0,   1,   0,   0,   0,   0,   0,  69, -37,  -6,  -1,  -1,   0,  -1,   0,  70,  10, -17,   0,  -4,   0,  -2,   0,  -4,  31,  -5,   0,  -1,   0,   0,   0, -18,  14,   3,   0,   0,   0,   0,   0, -12,   3,   3,   0,  -5,   2,   1,   0,  -5,   1,   1,   0,  -1,   1,   1,   0 },
     {   28,  77, -31,  -1,  -7,  -1,  -3,  -1,  -2,   0,  -1,   0,  -2,  30,  -9,  -2,  -2,  -1,  -1,   0,   0,   0,   0,   0,  14,  11,  -3,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0, -30,   9,   4,   0,   1,   0,   1,   0,   0,   0,   0,   0,  34,   0,  -3,  -1,  -1,   0,   0,   0, -36,  32,   4,   0,   1,   0,   0,   0, -47,  -4,  13,  -1,   3,   0,   1,   0,  -3, -16,   5,   0,   1,   0,   0,   0,   1,  -9,   1,   0,   3,  -4,   0,   0,   1,  -4,   1,   0,   2,  -2,   0,   0 },
     {   26,  65, -28,  -2,  -7,  -1,  -3,  -1,  -1,   0,  -1,   0,  19,  27, -11,  -3,  -3,  -1,  -1,  -1,   0,   0,   0,   0,  -7,  -7,   4,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,  42,  -9,  -3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -31,  -5,   6,   0,   1,   1,   0,   0,  51, -27,  -6,   0,  -1,   0,   0,   0,  34,  24, -11,   1,  -2,   0,  -1,   0, -24,  21,   3,  -1,   1,   0,   0,   0, -10,   3,   4,  -1,  -8,   1,   3,   0,  -4,   1,   2,   0,  -4,  -1,   2,   0 },
     {  -15, -20,  10,   1,   2,   1,   1,   0,   1,   0,   0,   0,   1,  15,  -4,   1,   0,   0,   0,   0,   0,   0,   0,   0, -19,   9,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   9,  22,  -6,   0,  -1,   0,  -1,   0,   0,   0,   0,   0, -41,   5,   6,  -1,   2,   0,   0,   0,  38,  14,  -8,  -1,  -2,  -1,  -1,   0, -71,  39,  14,  -1,   2,   0,   1,   0, -41, -33,  18,   0,   3,   0,   1,   0,  11, -25,  -1,   1,  12, -10,  -4,   0,   8,   1,  -3,   0,   2,  -2,   1,   0 },
     {  -13, -26,  12,   0,   3,   1,   1,   0,   1,   0,   0,   0,   0,  51, -16,  -2,  -3,  -1,  -1,  -1,  -1,   0,   0,   0,  12,  57, -13,  -3,  -4,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -4,   9,   5,  -3,   1,  -1,   0,  -1,   0,  -1,   0,   0,  16,  14,  -2,  -1,  -1,  -1,   0,   0, -26,  -2,   6,   1,   2,   0,   0,   0,  30,   6,  -6,   0,  -1,   0,  -1,   0, -51,  32,  11,   0,   2,   0,   1,   0, -39, -23,  16,   0,   4, -20,   1,   1,   3,  -7,  -1,   0,   2,  -2,   0,   0 },
     {    6,  23,  -8,  -1,  -3,   0,  -1,   0,   0,   0,   0,   0, -23, -75,  26,   5,   7,   2,   2,   2,   2,   0,   1,   0, -24, -26,   5,   4,   3,   2,   1,   1,   1,   0,   1,   0, -14,  17,  -7,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -5,  35,  -7,  -2,  -1,  -1,   0,   0, -16,  15,   2,  -1,  -1,   0,   0,   0,  14,  15,  -3,  -3,  -1,  -1,  -1,   0, -45,  25,  11,  -2,   2,  -1,   1,  -1, -24, -22,  15,  -1,   9, -20,   1,   2,  11,  -8,  -3,   1,   4,   0,  -2,   0 },
     {    4,  -3,  12,  -3,   1,  -2,   0,   0,   0,  -1,   0,   0,  11,  19,   9,  -5,  -2,  -1,  -1,  -2,  -1,  -1,   0,   0,  11, -44,  22,   0,   2,   1,   1,   0,   0,   0,   0,   0, -18, -68,  14,   7,   5,   2,   2,   2,   1,   0,   1,   0,  -3, -18,  -7,   5,   0,   2,   0,   1, -24, -11,  -1,   2,   2,   0,   1,   1,  13,   1,  -5,   1,   0,  -1,   0,   0, -37,  -9,  11,   1,   2,   0,   1,   0,  29, -43,  -7,   3,  36,   5, -19,   1,   4,  17,  -6,  -2,  -1,   8,  -1,  -1 },
     {    8,  14,  -4,  -3,  -2,  -1,  -1,   0,   0,   0,  -1,   0, -10, -42,   6,   3,   1,   1,   2,   0,   0,   0,   1,   0,  11,  59, -20,  -3,  -5,  -2,  -1,  -1,   0,   0,  -1,   0,  21,  24,   2,  -3,  -1,  -1,  -2,  -2,   0,  -1,   0,   0,  20,  -9,  10,  -2,   1,   0,  -1,   0,   1, -32,  11,   2,   1,   1,  -1,   0,  17,  -8,  -2,   3,  -1,   0,  -1,   0, -24, -23,   6,   3,   2,   1,   1,   0,  53, -28, -21,   3,  14,  34, -14,  -2, -11,  14,   3,  -2,  -7,   6,   3,   0 },
     {    7,  14,  18,  -9,  -1,  -3,  -1,  -1,   0,   0,  -1,   0,  -2, -28,  25,  -1,   2,  -1,   1,  -1,   0,   0,   0,   0,   5,  34,   1,  -2,  -4,  -2,  -1,  -1,   0,  -1,   0,   0,  22, -21,  20,   0,   1,   0,   1,   0,   0,   0,   0,  -1, -18, -72,  16,   6,   4,   2,   1,   1, -13, -15,  -7,   6,   0,   2,   0,   1, -28, -10,  -3,   4,   0,   2,   1,   1,   7,   6,  -9,   0,   0,   0,   0,   0, -36,   4,  10,   0,  10, -37,  -2,   2,  30,  -2, -17,   2,   4,  15,  -4,  -2 },
     {    5,  17,  53, -17,   0,  -7,  -2,  -4,  -1,  -1,  -1,  -1,  -7,  -2,  30,  -6,  -1,  -3,  -1,  -1,  -1,   0,   0,   0,   0,  18,  -4,   2,  -3,   0,   1,   1,   0,   0,   0,   0, -10, -28,   2,   6,   1,   2,   1,   2,   1,   0,   0,   0, -19,  -5, -11,   5,   0,   2,   1,   1, -21,   8, -11,   1,   1,  -1,   1,   1,  -8,  28,  -8,  -2,   0,  -1,   1,   0, -20,  11,   3,  -3,   2,  -1,   1,  -1,  19,  21,  -6,  -4, -50,  34,  22,  -3, -29, -33,  20,   1,   9, -22,  -2,   2 },
     {    6,   9,  47,  -8,  -2,  -5,   0,  -1,   0,  -1,  -1,  -1,  13,  29,  36, -11,  -4,  -5,  -3,  -2,  -1,   0,  -1,   0, -10, -43,  16,   7,  -1,   1,   0,   2,   1,   0,   0,   0,   7,  66, -14,  -1,  -3,  -1,  -2,  -2,  -1,  -1,   0,  -1,  31, -11,  23,  -5,   3,  -1,   0,  -1, -14, -46,   8,   7,   1,   2,   0,   2,  -9,  -4, -10,   6,   0,   1,   0,   1,  -5,   3,  -2,   0,   0,   0,   0,   0,  -4,  -3,   0,   0,   0,  -2,  -4,   1,  -4,   5,  -3,   0,  -1,  -3,  -1,  -1 },
     {  -14, -26, -63,  22,   1,   9,   2,   4,   1,   2,   1,   1,  -5,  -3,  -7,  -3,   4,   1,   0,  -1,   0,   1,   1,  -1, -10, -14,   4,  -3,   1,   1,   1,   0,   0,   0,   0,   0,   2,  18,   7,  -3,  -2,  -2,   1,   0,  -1,   0,   0,   0,  -8, -48,  15,   6,   1,   1,  -1,   2, -42, -11, -12,   9,   2,   2,   0,   2, -16,  34, -16,  -2,  -1,   1,   0,  -1, -10,  32,  -2,  -5,   1,  -2,   0,  -1,   3,   7,   5,  -4,   2,  25,   0,  -2, -28,   7,  19,  -2,  -2, -22,   4,   4 },
     {   -4, -15, -34,  13,   1,   3,   1,   2,   0,   1,   0,   0,   0,  -3, -11,   1,   0,   2,   1,   0,   0,   0,   0,   0,   6, -10,  16,  -6,   0,   0,   0,   0,   0,   0,   0,   0,  -6,  -9,  11,  -3,   0,  -1,   0,   0,  -1,   0,   0,  -1,  16,   9,   9,  -2,  -1,  -1,  -1,  -1,  -6, -46,  19,   3,   2,   1,   0,   1,  -7,   1,  -3,   4,   0,   2,  -1,   1, -27, -24,   3,   3,   2,   1,   1,   0,  18,  14, -10,  -1, -49, -13,  20,   2,  42, -53, -16,   6,  21,  23, -16,  -1 },
     {   -3,  -6, -16,   2,  -2,   3,   1,   1,   0,   0,   1,   1,   0,  -5,  19, -10,   1,   1,  -1,  -1,   0,  -1,   1,   0,   2,   3,  32,  -9,  -1,  -2,  -2,   0,  -1,   0,   0,  -1,  -1, -14,  24,  -1,   0,   0,  -1,   0,   0,  -1,   0,  -1,  -3,  14,   6,   3,  -2,  -2,  -1,  -1,  27, -12,  21,   1,   2,   0,   0,   1, -29, -69,  17,   8,   3,   4,   1,   3, -36,   0, -13,   7,   0,   1,   1,   1, -40,   4,  -4,   1,  -3,  25,  -4,  -3, -39,   3,  16,   0,   4, -24,   0,   1 },
     {   -2,  -9, -26,  15,   0,   1,  -2,   3,   1,   2,  -1,   1,   9,  17,  71, -24,  -1,  -5,  -3,  -3,   0,  -1,  -1,  -2,   7,  20,  51, -12,   0,  -4,  -4,  -2,   0,  -1,   0,   0,   9,  21,  20,  -4,  -3,  -2,  -3,  -1,   0,   0,  -1,  -1,  -2,  -1,  -2,   2,  -1,  -1,   0,   0,   4,  43, -28,   3,  -3,   1,  -1,  -1,  36,   9,   6,  -6,   0,   0,  -2,  -1,  10, -22,   1,   3,   0,   0,   0,   0,  11,  -9,  -4,   2,  -8,  -7,  -2,   0,   8,  -7,  -5,   3,   0,   4,  -6,   1 },
     {    3,   2,  18,  -1,   3,  -2,   0,  -2,  -1,  -1,   0,  -1,   1,  -7, -44,   1,   3,   2,   1,   2,   2,   0,   0,   1,  -7, -29, -11,   0,   0,   1,   1,   3,   2,   1,   1,   1,  16,  27,  21,  -9,   0,  -4,  -3,  -3,  -1,   1,  -1,  -1,  -6, -48,  12,   3,   2,   1,   2,   1,  -1,  40,  -3,  -9,  -4,  -4,   0,  -2,  36, -24,  26,   2,  -2,  -1,  -2,   0, -15, -27,   7,   5,   1,   3,  -1,   1,  -5, -22,  -1,   2, -10,   1,   5,   2, -15, -31,   3,   3,  34, -16, -18,   2 },
     {    1,   7,  24, -29,   1,  -4,   4,  -2,   1,   0,   0,  -1,  -8, -15, -48,   9,   1,   2,   2,   2,   0,   1,   2,   1,   4,   5,  61, -34,  -2,  -4,  -1,  -1,  -2,  -1,  -2,   0,  -2,  11,  57,  -5,  -6,  -3,  -3,  -3,  -1,  -1,  -1,   0,   4,  19,  24,   9,  -4,  -4,  -3,  -1,  -8,   4,   4,   9,  -3,   1,   1,  -1,   0,  31, -14,   8,  -3,  -2,   1,  -1,   7,  -1,   0,  -4,   2,   1,   0,   1,  -3,  15,  -9,   0,  -1,  -1,  -4,   2,  -6,  19, -10,  -1, -16,   2,   9,  -1 },
     {    7,   7,  29,  95, -38,   2, -12,  -2,  -5,  -1,  -1,  -1,   1,   0,   1,  52, -15,  -3,  -3,   1,   0,   0,  -1,  -1,   1,   1,   4, -17,   9,  -8,   0,   1,   1,   1,  -1,   1,  -1,   0,  11, -16,  12,  -2,   2,  -1,   0,   1,   0,  -1,  -2,   0,  -8, -11,   4,   3,   3,   0,  -3,   0,  -4,  -7,  -3,   3,   1,   1,   0,   1,  -2,   0,  -4,   0,   2,   0,  -8,  -3,  -4,   2,  -2,   2,   0,   1,  -4,  12,  -8,  -2,   0,   2,  -2,   1,  -1,   8,  -4,  -1,  -6,  -2,   3,   0 },
     {    5,   3,   5,  16,  -8,   0,   2,   0,  -2,   1,   0,  -1,   5,  -5,  -2,   3,   6,  -4,  -1,   0,   0,  -1,   1,   0,   8,  13,  19,  -4,  -2,   0,  -1,   0,   0,  -1,  -2,  -1,   6,  -9,   7,   7,  -6,  -1,  -1,  -1,   2,  -1,  -1,   0,   8,  17,   2,   2,  -1,  -3,  -2,  -3,  14, -22,  10,   9,   1,  -1,  -1,   3, -14,  13, -15,  -6,   4,  -1,  -1,   1,  52,  19,  20, -11,   0,  -3,  -1,  -2,   4, -48,  18,   6,  28,  -6,  -3,   1, -17, -42,  16,   8,  49, -28, -20,   6 },
     {    0,   2,   6,   4,   2,  -6,  -2,   4,   2,   1,  -3,   1,   1,   7,  -9,  -7,  -7,  -3,  -4,   0,  -1,  -1,   2,   2,  -3,  -2,   1,   8, -23,  -3,  -4,   5,   2,   1,   1,   1,   5,  18,  22,   4,   2, -10,  -1,   0,  -1,  -1,  -2,   1,  -5,   7,  -6,   8,   1,  -3,  -1,   0,  13,  25, -20,  -6,   0,  -2,   0,   0, -18, -33, -19,  11,  -2,   6,   0,   0, -16,  65, -34, -16,  -2,  -1,   1,  -2,  45,   8, -11,  -7,  17,  14,  -2,  -6,  36,  -7,  -8,   8,  23,  12, -10,   3 },
     {   -1,   0,   2, -10, -11,  -2,   1,   1,  -1,   0,  -1,   0,  -3,  -5, -27,  10,  -1,   1,   4,   2,  -2,   0,  -1,   1,   3,  11,  54, -22,  -4,   2,  -1,   0,  -1,   0,   0,  -1,  -1,  17, -58,  15, -11,   9,   0,   2,   0,   1,   0,   0,  -3, -23, -68,  19,   3,   5,   0,   4,  -4, -11, -24,   6,   8,   2,  -2,   1,   0, -20, -10,  -6,  10,  -3,   3,  -1, -10,  -9,   5,  -4,   0,  -1,   3,  -1,  -4,   3,   7,   2,   3,   7,   5,   0,   4,  -2,   4,  -2,   8,   8,   1,   2 },
     {    2,   2,  13,  13, -15,  21,  -9,   2,  -1,   2,   1,  -2,  -3,  -4,  -6, -71,  10,  14,   9,   9,   2,   2,   1,   0,  -3,  -2, -16, -76,   9,  13,  10,   7,   6,  -1,   1,   0,  -4,  -7, -21, -25,  -4,   2,   2,  -1,   0,   3,   1,  -2,  -2,  -1,   6, -24, -15,   0,   5,   0,   0,   2,   5,   3, -15,  -1,   1,  -1,  -4,  -3,   0,  -2,  -8,  -3,   3,   0,  -2,  11,   3,   6,  -9,  -2,  -3,   2,   9,  10,   3,   7,   8,  -1,  -3,   5,   4,  -1,  -7,   2,   4,  -1,  -1,   1 },
     {   -1,  -5,  -2, -10,  -9,  -7,   9,   6,   3,   4,  -1,   1,   2,   0,   2,  24,  -7,   0,  -1,  -6,   1,  -1,   0,  -2,   5,   7,  33,  -5,   8, -10,  -1,  -3,  -1,  -1,   3,  -3,   0,  -6, -34,   2,   6,  -1,  -1,   6,   2,  -3,   2,   1,  -1, -20,  32, -19,   7,  -5,   0,   4,   7,  21,  47, -16,  -2,   1,  -4,   0,   1, -12,  25,  -7,  -1,  -3,   1,   1,  -6,  34,  10,   0,  -7,  -1,   2,  -3,  49,   4,  28,   5, -13, -36,  11,   8, -14,   2,  -5,  -1, -16, -12,  10,   4 },
     {   -1,   0,  -4,   3,  17, -19,   7,  -1,  -6,  -6,   2,   3,   0,   1,   5,  -5,  13, -12,   2,  -4,  -6,  -4,   0,   5,  -3,  -5,   0,  -5, -10,  -1,  -3,   5,  -6,   1,  -3,   2,   2,   9,  18, -40,   6,  -3,  13,   2,   6,   1,  -3,   0,  -8, -11, -49,  -4,   3,  13,   3,   2,   2,  -4,  29, -30,  -1,  -4,   4,   2, -10,   3,   6,  36, -16,   4,   2,   4,  19,  23,  31,  -2,   4,  -7,  -2,  -2,  -4, -28,   3,  12, -44,  10, -27,   1,   3,  18,  -8,  -4,  -5,  13,  -1,  -5 },
     {    1,   0,  -4, -18,  23,   1, -10,  -6,  -8,  -5,   6,   0,   3,   2,   9,  19,  -7,  -1,  -7,   4,   5,  -8,   0,   2,   0,  -6, -16, -22,   7, -12,   6,  -4,  -2,   2,   3,  -1,   2,   6,  27, -40,  11,  -9,  10,  -2,  -2,   6,  -2,   2,   6,   9, -30,  26, -11,  19,   1,  -1,   0, -31, -34,   8,  -1,  -1,   6,   3,  -2,  -7,  41, -17,   6,   0,   2,   1,  -3,   8,   4,   6,  -2,   3,  -4,  -5,  27,  14,  17, -10,  21, -32,  18,   7, -11, -10,  -1,   7, -15, -24,   2,   1 },
     {    0,   2,  -4, -18,  40,  -7,  14,  -4,  -1,  -2,   1,   1,   5,   8,  12,  43,  22, -25,   2,  -5,  -1,  -4,   0,   1,   1,   2,   6, -20,  29,  -2,   0,   0,  -2,   3,  -2,   0,  -2,  -3, -20, -40,  -6,   7,   1,   8,   8,   5,   1,   1,  -1,  -3,  18, -28,  -5,  -8,  -7,  -3,  -2,   9,  -3, -17, -10,   4,  -8,   1,  -1,  -4, -16, -15,  -9,   6,   2,  -3,  -5,   2,  17, -12,  -9,  -2,   0,   2, -16,  19, -26,   0,  31,  29,   1,  -7,  10, -16,   5,  16,  11,  14, -16,   1 },
     {   -4,  -6, -10, -29, -30,  12,   6,  12,  -1,   1,  -3,   2,   2,   3,  20,  19, -66,  18,   4,   6,  -1,  -1,  -4,   0,  -3,  -5, -13, -23, -15,   0,  16,   4,  -8,  -3,   4,   4,   1,   2,  -1,  -6,   9,   3,  10,   4,  -7,  -3,   0,   3,   0,  -2,   3,  24,  -8,  11,   7,   8,   3,  12,  29,  -9,  28,  -7,  19,   0,   5,  12,  -8,  20,  -8,   9,   7,  -3,   2,  -9,  11, -10,   2,   8,   0,   3,  -9,  17,   3,  19,  20,  18,  -5,  12,  15,  -6,  11,   2,   9,  -2, -22,   9 },
     {    2,   3,   5,  14,  17,  -7,   2,   4,  -1,   0,  -8,   0,   0,  -1,   1, -17,   3,  14,   3,   6,   0,  -4,   3,  -8,   1,   4,  10,   4,  -8,   8,  -1,   7,  15,  -9,   3,  -8,   2,   4,   8,  34,   0, -23, -11,  -7,   1,  -6,   5,  -2,   1,  -3, -20,  -6,  23, -17,  -2,  -8,  -6, -15,  18, -37,  14,  18,   2, -16,   5,  27,  59,  -1,  -8,   2,  -5, -11,   0,  11,  12,   4,  -9,   9,   8,  -4,  -7,  35, -15,  -2,  34,  10,  -5,  -7,  -5,  -9,  -5,   6,  10,  13,  -5,   4 },
   },
   {
     { -108,  39,   6,   5,   2,   2,   1,   1,   1,   1,   0,   0,  51, -19,  -3,  -2,  -1,  -1,  -1,   0,   0,   0,   0,   0,  -6,   6,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {   41, -25,   2,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,  90, -15,  -9,  -3,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0, -68,  18,   5,   2,   2,   1,   1,   1,   1,   0,   0,   0,  19,  -8,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -9,   9,   0,   0,   0,   0,   0,   0,  -5,  -1,   1,   0,   0,   0,   0,   0,  -1,   3,  -1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {   24,  -8,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,  51, -27,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  65,   9, -11,  -2,  -4,  -2,  -2,  -1,  -1,   0,  -1,   0, -81,  13,   9,   3,   3,   1,   2,   1,   1,   0,   1,   0,  25,  -9,  -4,  -1,  -1,   0,  -1,   0, -14,  10,   1,   0,   0,   0,   0,   0,  -1,  -3,   0,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,  -2,   2,   0,   0,  -1,   0,   0,   0,  -2,   1,   0,   0 },
     {   28,  77, -33,  -1,  -9,  -2,  -4,  -1,  -2,  -1,  -1,   0, -25, -56,  25,   2,   6,   3,   3,   2,   1,   1,   1,   1, -31,  40, -10,   0,  -2,  -1,  -1,  -1,   0,   0,   0,   0, -10, -23,  10,   0,   2,   1,   1,   0,   0,   0,   0,   0,  20,   1,  -6,  -1,  -1,   0,   0,   0,  -8,   3,   3,   0,   1,   0,   0,   0,   4,  -6,  -1,   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   1,   1,   0,   0,   0,  -1,   0,   0,   1,   1,   0,   0 },
     {  -19, -30,  15,   1,   5,   1,   2,   1,   1,   0,   1,   0, -23,  34,  -8,   1,  -2,   0,  -1,   0,   0,   0,   0,   0, -51,  10,   8,   0,   2,   0,   0,   0,   0,   0,   0,   0, -40, -21,   7,   2,   3,   1,   1,   1,   1,   0,   1,   0,  76,  -1, -11,  -2,  -4,  -1,  -2,  -1, -39,   4,   8,   1,   2,   1,   1,   0,  23, -10,  -4,   0,  -1,   0,   0,   0,  -4,   4,   1,   0,   0,   0,   0,   0,   7,  -4,  -1,   0,   0,   2,   0,   0,   3,  -2,   0,   0,   0,   0,   0,   0 },
     {    3, -11,   9,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   7, -28,   1,   1,   2,   1,   1,   1,   1,   0,   0,   0,  49,  -2,  -3,  -3,  -2,  -1,  -1,  -1,  -1,  -1,   0,   0,  72, -26,  -8,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  36,  34, -12,  -1,  -4,  -1,  -2,  -1, -58,   5,  10,   2,   2,   1,   1,   1,  17,   0,  -6,   0,  -1,   0,  -1,   0, -18,  10,   3,   0,   1,   0,   0,   0,   1,  -2,   0,   0,  -7,   5,   1,   0,  -2,  -1,   0,   0,  -4,   2,   1,   0 },
     {  -21, -62,  38,  -2,   8,   0,   3,   0,   1,   0,   1,   0, -29, -54,  10,   4,   6,   3,   2,   2,   1,   1,   1,   1,   1,  61, -11,  -4,  -4,  -2,  -2,  -2,  -1,  -1,  -1,  -1, -11, -28,   9,   3,   3,   1,   1,   1,   1,   0,   0,   0, -29,  11,  -4,  -1,  -1,   0,   0,   0,  25, -13,   1,   0,   0,   0,  -1,   0,   1,   6,  -4,   0,   0,   0,   0,   0,   4,  -1,   1,   0,  -1,   0,   0,   0,   3,   0,  -2,   0,   1,   1,   0,   0,   2,   0,  -1,   0,   1,   1,   0,   0 },
     {   -4, -19,   1,   0,   2,   0,   1,   1,   1,   0,   0,   0,   1, -20,  12,  -2,   2,  -1,   1,   0,   0,   0,   0,   0,   0,  -7,  -6,   2,   0,   1,   0,   1,   0,   0,   0,   0,  27,  -1,   0,  -2,  -1,  -1,   0,  -1,  -1,   0,   0,   0,  57, -29,  -8,   1,  -2,   0,  -1,   1,  34,  46, -12,  -2,  -3,  -1,  -2,  -1, -68,  -5,  14,   2,   3,   1,   1,   0,  36,  -8, -10,   0,  -2,   0,  -1,   0, -17,  13,   3,   0,   4,  -4,  -1,   0,  -5,   5,   0,   0,   1,  -1,   0,   0 },
     {  -10, -30, -11,   7,   3,   3,   0,   2,   1,   1,   1,   1, -10, -43,  47,  -7,   5,  -2,   2,  -1,   0,   0,   0,   0, -23, -29, -21,  10,   2,   3,   1,   2,   1,   1,   1,   1,   7,  62,  -1,  -9,  -4,  -3,  -2,  -2,  -1,  -1,   0,  -1,  -7, -42,  11,   6,   3,   1,   1,   1, -26,  15,  -7,  -3,   0,   0,   0,   0,  37, -15,   1,   1,  -1,   0,  -1,   0, -13,  11,  -4,   0,   0,   0,   0,   0,  12,  -4,   1,   0,  -2,   5,  -2,   0,   2,  -1,   1,   0,   1,   1,  -2,   0 },
     {   12,  26,  67, -22,  -1,  -9,  -3,  -5,  -2,  -2,  -1,  -1, -14, -37, -63,  21,   4,   9,   4,   5,   2,   2,   1,   1,  -9, -10,  41, -11,  -1,  -2,  -1,  -1,   0,   0,   0,   0,   1,  27, -19,   5,  -1,   0,   0,   0,   0,   0,   0,   0,  -1, -26,   5,  -2,   1,   1,   0,   0,  -6,  16,   1,   0,   0,  -1,   0,  -1,   6, -11,  -2,   0,   0,   1,   0,   1,   0,   4,   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   1,   1,   1,   0,   1,   0,  -1,   0,   0,   0,   1,   0 },
     {    7,   1,   1,  -1,  -1,   0,   0,   0,   0,  -1,   0,   0,   8, -11,   3,  -1,   1,  -1,   0,  -1,   0,   0,   0,   0,   6, -22,   8,   2,   2,   1,   1,   0,   0,   0,   0,   0,  10, -20,  -1,   1,   0,   1,  -1,   0,   0,   0,   0,   0,  43, -25,   2,  -1,  -2,  -1,   0,   0,  74, -23, -17,   3,  -3,   0,  -1,   0,  46,  41, -14,  -2,  -5,  -1,  -2,  -1, -42,  15,   8,   2,   2,   0,   1,   0,   3,   0,  -3,   0, -10,   6,   2,   0,  -1,   0,  -1,   0,  -4,   3,   0,   0 },
     {   -2,  -7,   3,   0,   0,   1,  -1,   1,   0,   0,   0,   0,  -6, -43,  21,  -3,   4,  -1,   2,   0,   1,   0,   0,   0, -17, -77,  23,   4,   5,   2,   1,   1,   1,   1,   1,   0, -30, -15, -19,   6,   0,   2,   1,   2,   1,   0,   0,   0,   2,  56,   0,  -6,  -2,  -2,  -1,  -1,  -1, -30,   4,   5,   3,   1,   0,   0, -14,   8,  -7,  -1,   0,   0,  -1,   0,  28,  -5,   0,   0,  -1,   0,  -1,   0,  -7,   7,  -3,   1,   6,  -1,   0,   0,   1,   3,  -3,   0,   2,   0,   0,   0 },
     {   -3,   0,  -1,   4,  -2,   1,   0,   0,   1,   0,   0,   0,  -6,  -6,   1,   0,   2,   0,   2,   0,   0,  -1,   0,   0,  -6,   6,   3,   0,   0,  -1,   0,  -1,   0,   0,  -1,   0,  -4,  20, -11,   0,   0,   0,   1,   1,   0,   0,   1,   0,   8,  10,  11,  -2,  -1,  -3,   0,  -1, -15,   2,  -8,   3,   1,   2,   0,   1, -42,  44,  15,  -5,   2,  -3,   1,  -2, -45, -57,  14,   4,   3,   2,   2,   2,  59,   9, -18,  -4, -30,   2,  14,   1,  17, -10,  -7,   1,  -5,   4,   2,  -1 },
     {    7,   8,  25,  63, -15,  -5, -10,  -3,  -3,  -1,  -2,  -1,  -5,  -5, -24, -80,  14,   5,   9,   3,   3,   1,   2,   1,   0,  -1,  -2,  49,  -9,  -1,  -4,  -1,  -1,  -1,  -1,  -1,   0,   3,   9, -22,   5,   1,   1,   0,  -1,   0,   0,   0,  -2,  -2, -19,   5,  -2,   0,   0,   0,   2,  -1,  13,   6,   2,  -1,  -1,  -1,  -3,   0, -10,  -6,  -1,   1,   2,   1,  -2,   0,   5,   6,   0,   0,  -1,   0,  -2,   1,   1,  -4,   1,  -3,   0,   3,   0,   0,   0,  -2,   1,   0,  -1,   2 },
     {   -7, -21, -54,  39,  -5,   9,  -2,   3,  -1,   1,   0,   0,  -5, -12, -18, -11,   3,  -1,   3,   1,   2,   1,   1,   1,  10,  23,  58,  -7,   0,  -7,  -3,  -4,  -1,  -2,  -1,  -1,  -3, -14, -57,  15,   1,   4,   4,   2,   1,   2,   0,   1,  -1, -17,  37, -10,  -1,  -2,  -3,   0,  -8,   4, -21,  10,   1,   3,   1,   1,  10,  -9,   9,  -6,   0,   0,  -1,   0,   9,   4,  -5,   2,   0,   0,  -1,  -1,  -2,   3,   1,  -3,   5,   0,   1,   3,   0,   2,  -1,  -1,   2,   1,   1,   0 },
     {    7,   4,   5,  -1,   1,  -2,  -3,   0,  -1,   0,   0,   0,  10,   2,  -3,  -2,  -3,   0,   0,   0,   0,   0,   0,   0,  12,  -3, -10,   0,  -5,   0,  -2,   1,   0,   1,   0,   0,  13, -16,  16,   0,   2,  -2,  -1,  -1,  -1,  -1,   0,   0,   6, -36,   8,   2,   1,   2,   0,   1,  -8, -46,  -4,   4,   1,   3,   1,   2,  28, -15,   0,  -2,  -2,   0,  -1,  -1,  68, -37, -14,   6,  -2,   2,  -2,   1,  39,  40, -21,   0, -21,  12,   8,   1,   4,   6,  -4,   2,  -7,   6,   3,   0 },
     {   -2,  -9, -17,   7,   0,   2,   0,   1,   0,   0,   0,   0,   1,  -8,  -7,   5,  -1,   1,   0,   0,   0,   0,   0,   1,   1, -37,  23,  -4,   6,  -2,   1,  -1,   1,   0,   0,   0,  -8, -62,  35,   8,   0,   2,   0,   2,   1,   1,   1,   1, -35, -27, -34,   9,   1,   2,   2,   2,   0,  47,  11, -10,  -3,  -2,  -1,  -3,  -1, -25,   4,   7,   1,   2,   1,   1, -20,   3,  -2,  -3,   0,  -1,   1,   0,  36,  -5,  -5,   2, -15,   8,   0,   0,  10,  -4,  -2,   0,  -1,   6,  -2,  -1 },
     {   -8, -19, -56,  23,  -3,   9,   0,   5,   0,   2,   1,   2, -10, -23, -68,  31,   0,   8,   0,   4,  -1,   2,   1,   2,  -7,  -3, -20, -10,   1,   0,   1,   1,   1,   1,   1,   0,   6,  38,  27,  -2,  -4,  -2,  -3,  -3,  -2,  -2,  -1,  -1,  12,  20, -28,   1,   3,  -1,   2,   0,  15, -13,  26,  -6,  -2,  -1,  -1,   0,   2,   4, -14,   3,   1,   1,   1,   0,   8,   1,   5,  -5,  -2,   0,   0,   0,   1,   8,   0,   2,  -3,   4,   1,  -2,   1,   1,   2,   0,  -3,   3,   2,  -2 },
     {    4,   6,   8,   2,   3,  -3,  -1,  -3,   0,   0,   1,   0,   4,   8,   7,  -8,  -3,   2,  -1,   1,   0,   1,   0,  -1,   2,   9, -11,   5,   0,   0,   0,  -2,   0,   0,   0,   0,   0,  18, -31,   5,  -1,  -2,  -1,   0,  -1,   1,   0,   0,  15,  49,   7, -12,  -4,  -3,  -1,   0,  46,  16,   9,  -4,   2,  -4,   0,  -2,  10, -68,   9,   9,   0,   4,   1,   2,  -2,  20, -16,  -4,  -2,   0,   0,   0,  47,  -2,   0,   0, -32,  13,   1,  -1,  12,  -6,   2,   0,  -4,   4,   0,   0 },
     {   -1,   1,   2,   4,   2,  -4,   1,  -1,   1,   0,  -1,   0,  -2,   1,   3,   0,  -2,   4,   1,   0,   0,  -1,   0,   0,  -3,   0,  -1,  -8,   2,   3,  -2,   1,   2,   0,   0,  -1,  -3,   1,   1,   4,   1,  -4,   0,   1,  -1,   1,   0,   1,  -5,   6,  -8,  -4,  -3,   1,   2,   0,  -1,  13,  -1,  -2,   3,  -3,  -1,  -1,  18,   3,   9,  -1,  -2,  -2,  -1,  -1,  -8,   1, -17,   6,   2,   1,   0,   1, -30,  48,  24,  -7, -32, -71,   9,   9,  53,  20, -28,  -7, -25,  12,  16,   1 },
     {    0,  -1,  -2,  -7,   1,   4,  -1,  -2,  -1,   1,   0,   0,   5,   6,  40, -23,  -1,  -2,  -1,  -1,  -1,   0,   0,   0,   8,  21,  77, -36,  -1,  -6,  -1,  -5,  -1,  -2,  -2,   0,  10,  37,  24,  -1,  -9,   0,   0,  -3,  -3,  -1,   0,  -1,   9,  -1, -42,   6,   0,   1,   0,   2,   9, -22,  30,  -7,  -5,   0,  -1,   0,  -6,  -8, -16,  10,   3,   1,   1,   0,   1,  -5,   5,  -3,  -2,   0,  -1,   2,   2,   3,  -5,   5,  -1,  -5,  -2,   0,   0,  -4,   0,   1,   2,   1,  -2,   1 },
     {   -1,   0, -17, -28,  47,   0,   6,  -2,   2,  -1,   0,  -1,  -4,  -8,  -8,   4, -49,   1,   1,   2,   0,   1,   1,   1,   5,   8,  21,  51,  18,  -2,  -9,  -4,  -1,  -2,  -2,  -1,  -3,  -2, -11, -59,   2,   6,   8,   2,   2,   3,   1,   1,   1,   2,   4,  44, -15,  -6,  -4,  -2,   1,  -1,   4, -21,  14,   5,   2,   1,  -1,   1,  -4,   6, -10,  -1,  -1,   0,   3,   2,  -1,   3,   6,   2,   0,   0,   1,   3,  -2,  -4,  -1,  -2,  -1,   2,   5,   0,  -2,  -3,  -2,   3,  -2,   0 },
     {   -4,  -6,  -7, -25, -45,  21,  10,   0,   2,   0,   2,   1,   3,   4,  -3,  13,  63, -24,  -7,   1,  -1,  -1,  -1,  -1,   0,   0,  11,  10, -55,  13,   5,   0,   1,  -1,   1,  -1,  -3,  -3,  -6, -37,  30,  -3,  -2,   0,   1,   4,   3,   1,   0,   3,   9,  31, -15,  -3,   1,  -1,   1,  -1,   1, -21,   7,   6,   1,   1,  -1,  -3,  -3,  15,   4,  -7,  -2,  -1,   0,   6,  -4,  -5,  -4,   4,   2,   1,   2,   8,   3,   0, -10,  -2,  -4,   2,   3,   0,   3,  -3,  -1,  -3,   0,   1 },
     {    0,   4,  -1,  -5,  12,   3,   0,   0,  -1,  -3,   3,  -2,   0,   6,   1,  -8,  10,  -7,   3,   0,   2,   0,   1,   0,   0,   6,   3,  -4,  -7,   0,   1,  -7,   1,  -2,  -1,   0,  -5,  10, -13,   1,  -7,   0,   0,  -2,   3,  -2,  -1,  -1,  -7,  33, -42,   5,   3,  -2,   0,   1,   9,  57, -31, -15,   2,  -1,  -2,  -2,  52,  30,  13, -10,   0,  -1,   1,  -2,  26, -32,   5,   0,  -2,   1,   0,   3,   9,  21,  -4,  -5,  22,   4,   4,  -2, -31,   6,   5,   0,   6, -11,   4,   2 },
     {   -3,   2,   2,   8,   1,  -6,  -3,  -3,   2,   1,   1,   1,  -3,   4,   4,   3,   1,  -7,   3,   1,  -2,  -2,  -2,  -1,  -4,   4,   3,  -3,  13,   4,  -1,   2,  -1,  -2,   1,   0,  -5,   1,  -3,   2,  -1,  -3,  -1,   5,   2,   3,  -1,   0,  -8,   0,  -1,   5,   5,  -1,   1,   0,  -8,  11,  -7,   0,  -3,   0,   2,  -2,  -3,  17, -18,  -2,  -1,  -4,   0,   1,  13,  30,   3,  -1,  -2,   1,  -1,  -2, -28,  35,  -1,  -4, -68,  19,  31,  -4, -16, -64,  15,   6,  26,  -4, -20,   1 },
     {    0,  -1,   3,   1,  -6,   5,   1,  -1,  -1,  -2,   2,   2,  -4,  -8, -15, -17,  -1,   1,   7,   4,   4,   3,  -1,  -1,   0,  -3,  16, -11,  10,  -8,  -3,  -4,  -1,   1,   0,   0,   3,   6,  58,  -3, -11,   3,  -9,  -1,  -1,  -2,   0,   1,   8,  24,  19,  17,   7,   1,  -6,  -2,  -5, -21, -66, -15,   2,   6,   4,   3,  -2,  -9,  49,   5,  -6,  -2,   1,  -1, -10,   9, -34,   3,   5,  -1,   2,   0,   2,  -4,  23,  -3,  -5,   1,  -9,   6,  -8,  -8,   4,  -3,   9, -10,  -5,   2 },
     {   -5,  -3, -25, -54,  47, -31,   4,   7,  -3,   2,   2,  -1,  -4,  -3, -23, -50,  45, -38,  16,  -2,   1,  10,  -1,   0,  -2,  -1,  -9,  -1,  10,  -5,  12,  -4,   2,   0,  -6,   2,   1,   0,   3,  20,   3,   6,   9, -11,   1,  -5,   2,  -2,  -2,  -6,   3, -17,  -6,  -1,  -7,   2,  -3,  -5,   4,   8,  -7,   9,  -4,   3,  -6,   1,  -4,   8,  -5,   1,   5,  -3,   0,  10,   5,  -9,   6,  -3,  -3,  -2,   5,   0,   8,  -3,  -2,  -1,   6,   6,   1,  -2,  -5,  -4,   1,   7,   6,   0 },
     {   -3,  -6,  -6, -17, -18,  -5,  -2,  -4,  -4,  -1,   2,  -1,  -1,  -6,  -2,   2,  -7,   2,  -3,   6,   0,   1,  -1,   2,   2,  -3,   2,  20,   6,   7,   9,   4,  -2,  -4,  -3,   2,   0,  -8,  -1,  -1,  -7,   8,  -6,  -5,   0,   0,   2,   1,   2,  -5,  11, -21,   8,  -4,  -4,  -2,   8,  -5,  48,   6, -12,   6,  -4,   5,  10, -21,  23,   9,   7,   0,   2,  -2, -34, -41, -24,  16,   5,   4,   2,   1, -13,  36,  14, -16,   9, -17,  -2,  12, -40,  -7,  18,  -3,  37, -18, -11,   3 },
     {    3,  -3,  -3,  -1,   3,  -6,   9,   2,  -4,  -2,  -4,  -4,   5,  -3,  -2,   1,   7,  15,  -9,   3,  -1,  -2,  -2,   5,   4,  -5,  -1,   5, -15,  -3,   4,  -4,   1,   3,   3,   2,   7,   0,   3,   1,   5,  -6,  10,  -6,   3,  -7,   1,  -3,   7,   0,   2,  -1,  -3,   0, -11,   5,  10,  -2,   4,  -9,   4,  -5,   3,  -4,  20, -12,  35,   0,  -5,   5,  -2,   2,  14, -31,  33,  -4,  -2,  -8,  -1,   4, -19, -36,  -5,   9,  13, -18,  14,   4,  37, -54, -25,   8,  36,  29, -29,  -6 },
     {    3,   1,   3,  15,  14,  -3,   2, -11,   6,  -5,  -4,   1,   1,  -1,  -4, -15,  18,   0,   8,  -8,   0,  -1,   0,   0,  -1,  -7,  -9, -65,  28, -16,   9,   3,   1,  -2,   3,   1,  -2,  -9,  10, -61,   3, -11,  -2,  15,   3,   8,  -1,   1,   3,   9,  45,   4,   3,  -3,   1,  -2,   7,  24,  24,  10,   4,  -4,  -3, -10,  11,   1, -13,   0,  -4,  -4,  -3,   1,   5, -23,  -5,  12,  -4,  10,   1,   1,  -6, -11,  -2,   6,  -1,  -3,  -8,  -6,  -1,  -2,   8,   1,  -4,   3,  -3,  -1 },
     {    3,   3,  10,  13,  -5,  18, -13,  10,  -9,   4,  -5,  -2,  -2,  -1,  -6, -19, -11, -24,   5,  20,   7,   1,  -4,   2,  -4,  -4, -28, -40,  12,  13,  -2,  -1,  -3,   0,   1,   3,  -4,  -7, -35,  20, -17,   7,   4,   0,  -4,   8,   0,  -1,   1, -11,  -9,  40, -11,   6,  -6,   6,  -2, -17,  13, -56,   8,   1,  -9,   6,  -6,   2,  15,  31,  -3,  -1,   1,  -2,  -7,   8,   2,  21,   0,  14,  -2, -10,   3,   7,  -2,  -2,  14,  -4,  11,   2, -11,  -3,  -5,   7,   5,  21,   2,  -4 },
     {    0,   2,  -4,  17,  41, -35,   3,  -8,   2,  -3,  -1,  -3,  -2,  -1,   3,  -6, -14,  40,  -1,   6,  -5,   6,   1,   3,  -5,  -8,  -2, -18, -50,  -9,  -3,   1,  -1,   1,   5,  -2,   0,   0,  -1,  17,  58, -20,   2,  -4,   1,  -2,  -3,  -2,  -1,   1,   7,   4, -11,  15,  -4,   5,  -2,  -4,  -2, -29, -13,   3,   4,   2,  -2,  -2,  -2,  20,  11,  -2,  -2, -11, -13, -10,  -4,   1,   8,  -5,  -5,   7, -13,  10,   3,  -6,  -7,  13, -12,  -4, -15,  -1,   7,   3,  -7, -13,   8,  -3 },
   },
   {
     { -102,  36,   9,   4,   3,   1,   1,   1,   1,   0,   0,   0,  23, -14,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  55, -17,  -8,  -1,  -2,  -1,  -1,   0,   0,   0,   0,   0, -16,  11,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0, -10,   1,   4,   0,   1,   0,   0,   0,   5,  -4,   0,   1,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   25,  21, -14,  -3,  -4,  -1,  -1,  -1,  -1,   0,   0,   0,  92, -26,  -6,  -4,  -3,  -2,  -1,  -1,  -1,   0,  -1,   0, -31,  -6,  10,   4,   3,   1,   1,   1,   0,   0,   0,   0, -63,  16,   5,   2,   3,   1,   1,   0,   1,   0,   0,   0,  12,  -3,  -3,  -1,  -1,   0,   0,   0,  15,  -3,  -3,   0,  -1,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,  -6,   2,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0 },
     {  -10, -77,  31,   9,   6,   2,   2,   2,   1,   1,   1,   1,  30,  -2,  -9,   2,  -1,   0,  -1,   0,   0,   0,   0,   0,  23,  72, -30, -10,  -6,  -2,  -2,  -2,  -1,  -1,  -1,  -1, -16,   1,  13,  -3,   0,   0,   0,   0,   0,   0,   0,   0, -11, -27,  11,   5,   2,   1,   0,   1,   0,   1,  -6,   2,   0,   0,   0,   0,   1,   8,  -3,  -2,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   59,  -4,  -5,  -7,  -3,  -1,  -1,   0,  -1,  -1,   0,   0, -15,  -2,   4,   1,   1,   0,   1,   0,   0,   0,   0,   0,  81, -31, -10,   3,  -2,  -1,  -1,   0,   0,   0,   0,   0, -33,  28,  -1,   0,   1,  -1,   0,   0,   0,   0,   0,   0, -48,   9,   9,  -1,   2,   1,   1,   0,  20, -11,  -2,   1,  -1,   1,   0,   0,   3,   3,  -3,   0,   0,   0,   0,   0,  -7,   3,   1,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,  -1,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {   14,  -7,   6,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,  55, -36,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   9,  -8,  -9,   3,  -1,   1,   0,   1,   0,   0,   0,   0,  75,  18, -20,  -5,  -4,  -2,  -3,  -1,  -1,  -1,  -1,   0, -29,  32,   7,  -3,   0,  -1,   0,  -1, -51,  -7,  14,   4,   3,   1,   1,   0,   5, -12,  -4,   3,   1,   0,   0,   1,   6,   5,  -4,  -1,   0,   0,  -1,   0,  -3,   4,   2,  -1,  -4,   1,   1,   0,  -2,   0,   0,   1,  -2,   1,   0,   0 },
     {   -4,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  34,  63, -27,  -7,  -6,  -3,  -3,  -2,  -1,  -1,  -1,  -1,   0, -12,  17,  -1,   2,  -1,   0,  -1,   0,   0,   0,   0,   3, -69,  19,  10,   3,   4,   2,   2,   1,   1,   0,   1, -54,  24, -11,   3,   3,   1,   1,   1,   6,  13,  -4,  -7,   0,  -2,   0,  -1,  27,  -7,   1,   0,  -3,   0,  -1,   0,  -6,  -2,   2,   2,   0,   0,   0,   0,  -5,   0,   0,   1,   1,   0,   0,   0,   1,   1,   0,   0,  -1,   0,   1,   0 },
     {    9,  19,  73, -39,  -8,  -4,  -2,  -4,  -1,  -2,  -1,  -1,  -1,   3, -11,   5,  -1,   0,   0,   0,   0,   0,   0,   0, -15, -23, -66,  39,  10,   3,   3,   3,   1,   1,   1,   1,  -3, -14,  16,  -8,   4,   1,   0,   0,   1,   0,   0,   0,  13,   9,  25, -16,  -5,  -1,  -1,  -1,   6,  10, -10,   5,  -2,  -1,   0,   0,  -4,  -2,  -6,   5,   1,  -1,   0,   1,   1,  -2,   3,  -2,   0,   0,   0,   0,  -1,   1,   1,  -1,   0,   0,   0,   0,   1,  -1,  -1,   1,   1,   1,  -1,  -1 },
     {  -16,   8,   3,  -1,   0,  -1,  -1,   0,   0,   0,   0,   0, -18, -10,   6,   3,   2,   1,   0,   0,   0,   0,   0,   0, -55,  15,  -2,   5,   2,   2,   1,   0,   0,   0,   0,   0,   4,  29,  -5,  -5,  -2,   0,   0,  -1,   0,   0,   0,   0, -63,  -6,  24,  -2,   3,   0,   1,   0,  38, -50,   1,   5,   0,   1,   0,   1,  44,   4, -19,   2,  -1,  -1,  -1,   0, -10,  17,   1,  -4,   1,   0,   0,   0,  -5,  -4,   6,  -1,   5,  -4,  -1,   1,   3,   0,  -2,   0,   0,   0,   0,   0 },
     {  -19, -66,  16, -12,  17,   4,   4,   2,   2,   0,   2,   0,   7,   3,  -6,   2,  -3,   1,   1,   0,  -1,   0,   0,   0,  -6, -40,  26,  19,  -9,  -1,  -1,   1,   0,   0,   0,   0,  10,  14, -16,  -1,   2,  -2,  -1,   0,   0,   0,   0,   0,  24,  44, -23, -12,   1,  -2,   0,  -2,  52, -20,   9,   0,  -4,   0,  -1,   0, -21,   2,   3,   6,   0,   1,  -1,   1, -22,   7,  -2,  -1,   2,   0,   1,   0,   5,   1,  -1,  -1,   2,  -1,   1,   0,  -3,   2,   0,   0,   0,   0,   1,   0 },
     {   13,  19,  -2,  15,  -6,  -3,  -6,   0,  -1,  -1,   0,   0,  20, -18,   3,  -2,   2,   1,   0,   0,  -1,   0,   0,   0,  11,  -6, -10, -19,   5,   2,   4,   1,   1,   0,   0,   0,  55, -25, -12,   4,  -4,   0,   0,   0,   1,   0,   0,   0,   2, -36,   8,  14,  -3,   1,  -3,   1,  63,  18, -17,  -2,  -4,  -1,  -2,  -1, -18,  47,  -3,  -9,   1,  -2,   1,  -1, -33,  -2,  13,   1,   2,   1,   0,   1,  -4, -12,   0,   5,   2,   2,  -3,  -1,  -2,   5,   1,  -1,  -2,   1,   1,   0 },
     {   -3,   4, -36, -63,  24,   9,   8,   3,   3,   1,   2,   1,   0, -16,   6,   3,  -5,   1,   2,   2,   1,   0,   0,   0,  13,  26,  22,  67, -27, -12,  -9,  -3,  -3,  -2,  -2,  -1,  13, -10,   9,  -6,  14,  -2,  -2,  -2,  -1,  -1,   0,   0,  -7, -27,  -8, -27,  10,   7,   4,   2,  -1,  13, -14,   4,  -9,  -1,   1,   0,   8,  13,   4,   5,  -3,  -3,  -1,  -1,  -6,  -3,   5,  -1,   4,   1,  -1,   0,  -4,  -4,   0,  -1,   1,   1,  -1,  -1,   1,   1,  -1,   1,   0,   1,  -1,   0 },
     {    1,  11,  -1,  -9,   3,   0,  -1,   0,   1,   0,  -1,  -1,  -6, -12, -54,  24,   6,   3,   1,   2,   1,   0,   1,   1,   0,   3,  13, -15,  -2,   2,   2,   0,   0,   0,   0,   0,  19,  26,  76, -31, -14,  -2,  -3,  -4,  -1,  -1,  -1,  -1,  -4,  10, -22,  24,   1,  -4,  -1,   0,   4, -15, -36,  14,  11,   1,   2,   1, -17,  -6,  17, -11,  -1,   3,   0,   0,   0,   2,  10,  -3,  -4,  -1,   0,   0,   7,   1,  -5,   3,  -1,  -1,  -1,   0,  -1,  -1,   2,   0,   0,  -1,   0,   0 },
     {   -2,   4, -22,  -6,  -3,   2,   1,   1,   0,   2,   0,   1,  22,  80, -22, -10,  -8,  -2,  -2,  -1,  -1,  -1,  -1,  -1,   3,  -9,  -5,  13,   3,  -1,   0,  -1,  -1,  -1,   0,   0,  16,  35, -11,  -9,  -2,  -1,   0,  -1,   0,   0,   0,  -1,  12, -34,  34,  -6,   0,   2,  -1,   1, -11, -37,   4,  13,   2,   3,  -1,   2, -35,  19, -13,  -3,   3,   2,   2,   0,  14,   1,   0,  -6,  -2,  -2,   0,  -1,   7,   0,   3,   1,  -4,  -2,   1,   1,  -3,   0,   0,  -1,   0,  -1,   1,   0 },
     {   -8,   9,   4,  -6,  -9,   2,   2,   0,   0,   1,   0,   0,  -3,   0,  11,  -3,   0,   2,  -1,   0,  -1,   0,   0,   0, -22,  22,  -1,   5,   2,  -3,   1,  -1,   0,   0,  -1,   0,  -6,  12, -15,   4,   0,  -3,  -1,   1,   0,   0,   0,   0, -58,  16,  -2,  -3,   4,   0,  -1,   0,  15,  25,   7, -10,  -2,  -1,   1,  -2, -76,  -6,  28,   1,   3,   1,   2,   0,  18, -40,  -4,   6,   1,   0,   0,   1,  30,   1, -14,  -1,  -2,   6,   2,  -3,  -1,  -1,   4,   0,   3,  -4,   0,   1 },
     {   -7, -16, -26,   3, -56,  38,   4,   3,   0,   3,   1,   2,  -3, -12,  -8, -19,  16,   1,   4,   2,   1,  -1,   0,   0,   6,   6,   1,  19,  58, -38,  -8,  -3,  -1,  -3,   0,  -1,   6,   2,  19,  26, -26,   4,  -3,  -2,  -1,   0,  -1,   0,   8,  12,  17, -23, -25,  17,   1,   0,   9,   2, -13, -10,  11,  -6,   1,   2,   3, -10,  -5,   8,   6,  -4,   1,   0,  -4,   0,   5,   2,  -4,   2,  -1,   0,   0,   4,   0,  -4,  -2,   1,   0,   0,   0,   0,  -2,   0,   0,  -1,   0,   1 },
     {    9,  35, -25,   1,   8,  -5,  -1,  -1,  -1,   0,  -1,   0,  -1,   9, -19,   2,   0,   1,   0,   1,   0,   0,   0,   0,  18,  60, -23,  -8, -12,   4,   0,  -1,  -1,   0,   0,   0,   5, -14,  -3,  -7,   7,   1,   1,   0,   0,   1,   0,   0,  23,  49,  12, -17,  -4,  -5,   0,  -3,  26, -16,  32,   7,  -8,  -2,  -3,   0, -13, -32, -10,  17,   6,   4,   0,   3, -35,  14,  -8,  -6,   6,   1,   3,   1,   5,  -1,   7,  -8,   5,  -3,   3,   3,  -1,  -2,  -2,   2,  -1,   2,   1,  -1 },
     {    7,  20,  43, -30, -37,  14,   2,   1,  -2,   0,  -1,  -1,   2,  12,  -4,  26,  -7,  -7,  -2,  -2,  -1,   0,   1,   0,  12,  25,  26,  -5,  31, -19,  -5,  -5,   0,  -1,  -1,  -1,  -2,   3, -30, -19,   9,   9,   2,   1,   0,   0,   0,   1,   0, -14, -54,  14, -12,   8,   5,   2,  -8, -22,  33,   1,  -1,  -4,   0,   1,   6,  18,  15,  -2,   2,  -3,  -2,  -1, -14,  11,  -5,  -1,   2,   2,   0,   0,   1, -15,  -1,   3,   1,  -1,  -2,   0,   0,   2,   1,  -1,  -1,  -1,   1,   0 },
     {    0, -11, -18,   9,  -2,   6,  -3,   2,   0,   2,   0,   1,   8, -11,  21,  36,  -7,  -3,  -4,  -2,  -2,   0,  -2,   0,   5, -10,  -4,  -2,  15,  -9,   4,  -3,   2,  -1,   1,   0,   2, -37,  -8, -57,  11,  13,   6,   3,   2,   1,   3,   0,  -2, -14,   7,  -3, -16,   6,  -2,   2,  29,  -8,  -4,  37,   0,  -7,  -2,   0,  -5, -26,   1,   8,   7,  -1,   0,   1,  53,   9, -22, -13,  -5,   1,   0,   0,   1,  28,   0,  -6,  -8,   1,   7,   2,  -3,   0,   0,   1,   0,   1,  -2,  -1 },
     {   11,  15,  27,  -4,  -2,  -6,   1,  -2,  -2,  -1,  -1,   0,   3, -14, -11, -30,   5,   6,   3,   2,   0,   1,   1,   1,  13,   7,  13,  -3, -13,   3,  -4,  -2,   1,  -1,   0,   0,  16, -14,  14,  55,   4,  -9,  -4,  -1,  -2,  -1,  -1,   0,   6, -27, -21,  -6,  17,   0,   4,   1,  24, -41,   5, -20,  -5,   8,   2,   3, -10, -19,  -5,  13,  -8,   1,  -1,   1,  50,  19, -21,   3,   1,  -4,   0,   0,   8,  27,   2,  -9, -13,  -2,  11,   1,  -6,  -3,  -2,   4,   0,   1,  -2,  -1 },
     {    0,   4,   2, -13,   2,  -1,   1,  -1,   2,  -1,   2,   1,   5,  20, -28,  10,  10,   2,   2,  -2,   0,   0,   0,   0,   1,   5,   1,   1,  -3,  -1,  -3,   3,   0,   0,   0,   0,  14,  44, -23,  -5, -17,  -2,  -2,  -4,   2,  -1,   0,   0,   9, -17,   4,  16,   5,   1,  -1,  -4,  35,  50,  14, -20,   0,  -4,  -1,   0,  34, -42,  18,  -5,  -1,   1,   0,   1,   9, -20, -16,  12,   0,   2,  -1,   1, -41,  26,  -2,   1,  -7,  -2,   7,  -2,  10,  -3,   0,   0,   1,   4,  -2,  -1 },
     {    6,   9,  12,   7,  30,  56, -20, -13,  -8,  -3,  -4,  -2,  -1,  -1, -15,  14,  27, -13,   3,  -3,   1,  -2,   1,   1,  -6,  -8,  -5, -13, -39, -51,  23,  13,   5,   3,   3,   2,  -8,  -6, -11,   4, -40,  17,  -3,   4,   3,   1,   2,   1,  -3,  -5,   0, -11,  26,  18,  -9,  -4,  -3,   5,  16,  10,  17,  -7,   2,  -3,  -1,  12,  -6,  11,  -6,  -3,   1,  -2,   4,   3,   1, -10,  -6,   3,  -2,   0,  11,  -5,   1,  -9,  -2,   4,  -1,   3,  -4,  -2,  -1,   2,   0,  -1,   4,   3 },
     {    5,  13,   9,  13,  -2,  11,  -8,  -5,  -5,   0,   0,   0,  15,  35,  74, -26,  -6, -10,  -4,  -3,  -1,  -1,  -2,  -2,   3,  15, -17,   6, -11, -12,   5,   4,   4,   0,   0,  -1,  14,  19,  12,  -9, -11,   3,  -2,  -3,  -1,   1,   0,   1,   7,  13, -33,  -3,  16,   2,   0,  -1,   8, -11, -46,   9,  10,  -2,   2,  -1,   6, -19,  31,  -2,  -6,   1,  -1,   1, -11,  10,   4,   2,  -5,   1,   1,   0, -14,   6,  -5,  -1,   1, -10,  -1,   1,   4,   0,   0,   0,  -2,  -1,  -2,   0 },
     {    4,  -3, -10,  -7,   0,   1,  -1,   3,   0,   0,   1,   1,   6,   3,   4,  -3,   4,  -2,   0,   0,   3,   0,  -2,   1,   5,  -4, -15,   3,   3,   4,   1,  -2,   2,  -1,  -1,   1,   7,   0,  -3,  -1,   3,   4,   5,   3,  -1,  -2,   1,   0,  16,  -4,   3,  20,   6,  -2,  -1,   0,   7, -20,   4,  -2,   1,   1,  -1,  -1,  47,  -7,  17,  -2, -11,  -2,  -2,   2, -13, -43,   1,   5,   1,   2,   2,   2,  80,   4, -28,  -5, -22,  37,   9,  -4, -12, -15,   5,   3,   8,   0,  -7,   1 },
     {   -4,  -4,  -9, -67,  26,  -4,   0,   3,   4,   0,   2,   1,   5,   2,  34,  11,  -6,   7,  -1,   1,   3,  -3,  -2,  -1,   2,   4,   1, -24,  16,  -3,  -3,   1,  -1,  -2,   0,  -1,  -5, -16,  -2,  24, -42,   3,   1,  -4,   0,   3,  -1,   2,   4,  11,  20,  51, -11, -15,  -4,  -1,   0, -12,  -1, -11,  31,  -7,   2,  -1, -14,  -4, -16, -12,   4,  11,   1,   1,   0,   6,  -7, -11,  -7,   2,   1,   1,  -5,   4,   5,   3,  -4,   1,  -2,  -3,   8,   2,   1,   1,   2,   3,  -1,   2 },
     {    0,  -2,   4,   4, -25, -38,   8,   2,   3,   3,   3,   0,   4,   2,   3,   8,  53,  -8,  -9,  -1,   0,  -1,  -1,  -1,   6,   5,  15,  16,  -3,  44,  -9,  -3,  -4,  -4,  -3,  -1,  -4, -10, -11, -13, -65,  11,  12,   2,   1,   1,   1,   1,   0,  -8,  -9, -31,  19, -26,   6,   4,   1,   2,   1,  12,  21,  -4,  -5,  -1,   1,   3,  -1,  17,  -5,  12,  -4,  -2,   1,   2,   6, -10,  -5,   2,   0,   0,   8,  -9,   0,  -2,  -2,  -1,  -5,   2,  -2,   1,   0,   1,   1,  -2,   1,  -1 },
     {    5,   7,  40, -11, -21,   4,   4, -16,  -3,   0,   2,  -2,   0,   1,   6, -23,   6,  -1,  -3,   1,  -2,   1,   1,   1,   9,  13,  65, -14, -23,  -5,   1,   4,  -1,   1,  -2,   0,   0,   3, -11, -13,  16,  -2,   4,   0,   4,   0,   0,  -2,   7,   7,  37,   6,  -5,   5,  -7,  -5,   1,  13, -33,  25,  -4,  -6,   0,   1,   0, -24, -42,  -1,  13,  -3,   1,   3,  -8, -17,  20,  -7,  -2,   0,   0,   1,  15,  10,   8,   4,   0,   8,   2,   3,   0,  -3,  -7,   0,  -1,  -4,  -1,  -1 },
     {    0,  -7,   3,   2,   3,   5,   8,  -6,  -6,  -1,   2,   1,   6,   0,  16,   7,   6,   3,   3,  -1,  -1,   0,  -7,   2,  -2, -22,   2,   0,   4,   0,   0,   3,   8,  -4,   1,   5,   6,  -3,  24,  -3,  -1,  -6,   3,   1,  -2,  -4,   5,   1,  -8, -46,  -1,   7,  -4,   3,   3,   0,  -7,  -3,  23,  -4,   4,  14,  -2,   6, -15, -51,  -5,  24,  -6,  -2,   7,   7, -38, -10, -11,  11,   6, -11,   4,   2,  14,  -9,   8, -15,  66,   0, -11,   2,   1,  13, -16,   7,  -8,   9,  -5,  -6 },
     {    5,  10,  13,   8,  26, -10, -12,   0,  -2,  -4,  -2,  -1,   6,  11,  11,  55, -28,   3, -10, -11,  -3,   0,   0,   0,   7,  11,  19, -17,  18,  -5,  -1,  -3,  -3,   1,   2,   1,   6,  12,   4,   4,  -8,  -9,   7,   1,  -2,   0,   0,   0,   7,  10,   8, -45, -14,   3,   5,   1,   2,  -3, -26, -44,   8,  17,   3,   4,   9,  13, -15,  38,  -9,   0,   1,  -3,  13, -22,  13,  19,   3,  -7,  -2,   2,   3,   0,   0, -13,  -1,   9,  -6,  -2,   0,   4,  -4,   3,  -4,   1,   6,   0 },
     {   -3,  -2,   1, -10,  -4,   4,  22,  -3,  -1,   0,   3,  -3,  -1,   7,  19, -16, -29,   2,  -4,   1,  -4,   3,  -1,   1,  -7,  -8,   0, -25,   4,   8, -13,   7,   0,   4,  -2,   3,   5,  22,  39,  -6,  -2, -10,   1,  -6,  -1,  -1,   0,   0,  -9,  -7, -10, -14,   7,   0,   3,  -6,   3,  44,  31,  14,  11,  -6, -12,  -3,   5,  25, -17,  39,  12,  -7,   2,   4, -14,  22, -24, -22,  -7,   2,   8,   0,  14,  13,   7, -17, -32,   6,  15,  11,  -5, -15,   1,   1,   7,  -6,  -7,  -1 },
     {   -4,  -2,  -8,   2, -14,  -7,   5,   6,   1,   3,   0,  -1,   0,   9,  -2,  34,  18,   1,   2,   1,  -3,   0,  -4,   3,  -5,  -2,  -8,  11,  -2, -14,   7,   0,  -3,  -1,   1,   0,  -5,   3, -24,   2,  11, -24,  -2,  -1,   1,   0,   4,  -4,  -7,  -5,  -2,  16,  12,   1,  -5,   3,  -7,  24, -31, -27,  -3,  27,   5,  -3, -10, -19,  -9,  -6, -11,   4,   7,   3, -18,  67,   7,  -2,   1, -13,  -9,  -1,  33,   9,  12,  -2, -21,  -7,   8,   2, -31, -15,  -5,   3,   7,   5,  -1,  -4 },
     {    2,   2,  -1,  -2,  -9,   8,  18, -15,   7,   0,   0,  -3,   5,   5,   7,  -1,  -1,  -7,  -7,   5,  -2,   5,   1,  -2,   2,   3,  -1,  -2,  -9,  -5, -23,  16, -11,   8,  -3,  -3,   5,   6,   9,  -9,  -1, -18,   0,  -5,  -3,  -1,   0,  -4,   9,  12,  10,   6,   7,  -2,  -2, -11,  14,  12,   7, -15,  -6,   1,  -6,  -7,  14,  -2,  -5, -11,  -2,   7,   5,   4,  38,  21, -11,   2,   0,   3,  -1,   1,  10, -44, -13,   4,  29,  22, -20,  -1, -32,  49,  35, -13,   4, -31,  13,   8 },
     {    4,   7,   5,  16,  20,   7,  41, -24,  10, -13,   4,  -7,  -4,  -7, -14,  -1,  -8,   1, -21,   6,  -7,   5,  -7,   3,  -1,   1,  -4,  10,   1,  -9, -49,  31,  -5,  12,  -6,   9,  -4, -11, -19,   4, -11, -18,  50, -14,  10, -10,   9,  -6,  -2,   1,   3,   7, -10,  15,  18, -15,  -4, -13, -10,  15,  21,  12, -20,  13,  -4,   4,  19,  -8,  -5,  -3,   3,   1,  -8,  -2,   3,  13,  -6,  -1,   2,  -1,  -3,  13,  -4,   1,  -1, -13,  10,  -2,  -1, -15,  -5,   6,   0,   1,   3,  -3 },
   }
 },
 {  //22
   {
     {  120, -35,  -6,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0,  -5,   7,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0, -18,   7,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {    0,   5,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,-116,  15,   9,   4,   3,   2,   2,   1,   1,   1,   1,   0,   4, -24,   4,  -1,   1,   0,   1,   0,   0,   0,   0,   0,  40, -11,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,  10,   1,  -2,   0,  -1,   0,   0,   0,   6,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   5,  -1,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   2,   0,   0,   0,   1,   0,   0,   0,   2,   0,   0,   0 },
     {   13,  -7,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4, -13,   2,   0,   1,   0,   1,   0,   0,   0,   0,   0, 110, -14, -11,  -2,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0,  -2,  37,  -6,   1,  -2,   0,  -1,   0,   0,   0,   0,   0, -42,  10,   7,   0,   1,   0,   1,   0,  -6,  -4,   2,   0,   0,   0,   0,   0,  -5,   1,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -4,   0,   1,   0,  -1,   0,   0,   0,  -2,   0,   1,   0,  -1,   0,   0,   0 },
     {    1,  -9,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  35,  -8,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   1, -17,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  95, -12, -15,  -1,  -4,  -1,  -2,   0,  -1,   0,  -1,   0,  12,  48, -10,   0,  -3,   0,  -1,  -1, -44,  20,  10,  -1,   2,   0,   1,   0, -14,  -5,   6,   0,   1,   0,   0,   0,  -5,  -3,   1,   1,   0,   0,   0,   0,  -3,   1,   0,   0,  -4,   1,   1,   0,  -2,   0,   1,   0,  -2,   0,   1,   0 },
     {    6,  -8,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1, -19,   3,   0,   1,   0,   1,   0,   0,   0,   0,   0,  37,   0,  -7,  -1,  -2,   0,  -1,   0,  -1,   0,   0,   0, -19, -14,   6,   0,   2,   0,   1,   0,   0,   0,   0,   0,  85, -22, -19,   0,  -4,   0,  -1,   0,   8,  58, -13,  -1,  -3,  -1,  -1,  -1, -38,  15,  15,  -2,   2,   0,   1,   0,  -4,  -9,   3,   1,   0,   1,   0,   0,  -3,   1,  -1,   0,  -2,   3,   0,  -1,  -4,   1,   2,   0,  -1,   0,   1,   0 },
     {   -3, -27,   4,   0,   2,   1,   1,   0,   1,   0,   0,   0,  18,  -1,  -4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1, -20,   3,  -1,   2,   0,   0,   1,   0,   0,   0,   0,  42,  -4,  -8,  -1,  -2,   0,  -1,   0,  -1,   0,   0,   0,  -9, -21,   4,   1,   2,   1,   0,   1,  78, -14, -27,   1,  -5,   0,  -2,   0,   3,  64, -11,  -3,  -2,  -1,  -1,  -1, -28,  11,  15,  -2,   2,  -1,   1,  -1,  -5,  -1,   3,   1,  -5,   0,   1,   0,  -2,   3,   1,   0,  -3,   0,   2,   0 },
     {  -19, -85,  17,   0,   7,   2,   3,   2,   2,   1,   1,   1,  -6, -21,   3,   2,   1,   1,   0,   1,   0,   0,   0,   0,  12,   8,  -5,   1,  -1,   0,  -1,   0,  -1,   0,  -1,   0,  -6, -19,   3,   1,   1,   0,   1,   0,   1,   0,   0,   0,  27,   0,  -7,   1,  -1,   0,  -1,   0, -12, -18,   6,   2,   1,   1,   0,   1,  51, -21, -23,   3,  -4,   1,  -1,   1,   8,  42, -14,  -4,  -2,  -1,  -1,  -1, -12,  11,   9,  -2,  -1,   2,   1,   1,  -1,   1,   0,   1,  -1,   4,  -2,   0 },
     {  -27, -77,  19,   2,   6,   3,   3,   2,   2,   1,   1,   1,  -8,  16,  -3,   1,  -1,   0,  -1,   0,  -1,   0,   0,   0, -16,  10,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,  20,  -2,   0,  -1,  -1,   0,  -1,   0,   0,   0,   0, -29,  -2,   9,   1,   1,   0,   0,   0,   1,  14,   0,  -1,   0,   0,   0,   0, -48,  -4,  24,   0,   4,   0,   1,   0,  15, -54,   1,   5,   0,   2,   0,   1,  23,  -3, -17,   2,   5,   7,  -5,  -2,   1,   3,   1,  -1,   2,  -3,   0,   1 },
     {    8,   9,  -3,   1,   0,   0,  -1,   0,  -1,   0,  -1,   0,  18,   4,   1,  -2,   0,  -1,   0,  -1,   0,  -1,   0,   0,   1, -37,   4,   1,   2,   1,   1,   1,   1,   1,   0,   0,  15,   0,  -2,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -2, -30,   3,   2,   1,   1,   0,   0,  28,   1, -10,   0,  -1,   0,  -1,   0,  -7, -33,   3,   4,   1,   2,   0,   1,  62,  -3, -43,   3,  -6,   0,  -2,   1,  -7,  68,  -6,  -9, -18,   4,  19,  -2,  -3,  -1,   2,   2,  -1,   0,   0,   0 },
     {   -9, -11,  -4,   0,   1,   1,   0,   1,   0,   1,   1,   0,   8, 103, -12,  -6,  -6,  -3,  -3,  -2,  -2,  -1,  -1,  -1,  14,   6,   8,  -1,   1,  -1,  -1,   0,   0,  -1,   0,   0,  -3,  -1,   2,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,  -3,  -4,   0,   0,   0,   0,   0,  -6,  22,  -2,  -3,   0,  -1,   0,  -1,  -8,   0,   3,  -1,   1,   0,   1,   0,  -2,  29,   0,  -5,  -1,  -2,   0,  -1, -36,   0,  31,  -2,   2, -32,   5,   5,   1,  -6,  -2,   2,   2,  -6,  -2,   1 },
     {   -7,  -3,   1,   1,  -1,   0,   0,   1,   0,   0,   0,   0, -15, -47,   3,   3,   4,   2,   1,   1,   1,   1,   1,   0, -35,   4, -18,   2,  -2,   2,  -1,   1,   1,   1,   0,   0,   1,  69,  -9,  -6,  -2,  -3,  -2,  -1,   0,  -1,  -1,  -1,  -3,   8,   9,  -3,   1,  -1,   1,  -1,   6,  25,   1,  -3,   0,  -1,   0,  -1, -20,   1,   8,   0,   1,   0,   0,   0,   1,  28,  -3,  -6,   0,  -1,   0,  -1, -39,   7,  35,  -5,   1, -43,  11,   7,   8, -11,  -9,   3,   4,  -3,  -5,  -1 },
     {    0, -11,   2,   1,   1,   1,   1,   0,   0,   0,   0,   0,   9,  -4,  17,  -3,   1,  -1,   0,  -1,   0,   0,   0,   0, -13,-102,  17,   7,   6,   4,   2,   3,   2,   1,   1,   1, -40, -13, -18,   6,   0,   2,   0,   2,   1,   1,   1,   1, -10,  20, -13,  -1,  -1,   0,   0,   0,   1,  16,   4,  -3,   0,  -2,   1,  -1,   4,  -8,   6,   2,   0,   0,   0,   0,  -7,   2,   2,   1,   0,   0,   0,   0,  -4, -24,   6,   4,  20,  -9, -17,   2,   2,   9,  -4,  -2,   3,   2,  -2,  -1 },
     {   -4,  -2,   8,  -1,   1,   0,  -1,   0,   0,  -1,   0,   0,  -9,   6,  -4,   1,  -1,   0,   0,   1,   0,   1,   0,   0,  -3,  -4,   4,   2,   1,   1,   1,   0,   1,   0,   0,   0, -33,  -7, -18,   4,   0,   2,   0,   1,   0,   0,   0,   0,   1,  59, -11,  -8,  -2,  -2,  -1,  -1,   2,   2,   6,  -4,  -1,  -1,   0,  -1,   6,  32,   0,  -4,  -1,  -2,  -1,  -1, -15,  -9,   7,   1,   1,   0,   1,   0,   9,  41,  -9,  -9, -41,   9,  50,  -5,   3, -53,   9,  12,   6,  -5, -13,   2 },
     {    1,   2,  12,   0,   1,   1,   0,   0,   0,   0,   0,   0,  -6, -33,   2,   4,   2,   1,   1,   1,   0,   1,   0,   0,   6,  22,   8,  -5,   1,  -2,  -1,  -1,   0,  -1,   0,  -1,   1, -72,  18,   8,   2,   3,   1,   2,   1,   1,   1,   1, -60, -12, -25,  10,   1,   4,   1,   2,  -3,  37, -12,  -5,  -1,  -2,   0,  -1,  -1,   5,  14,  -4,   1,  -1,   1,   0,   4,  -4,   0,   0,   0,   1,   0,   0, -14,   7,  12,  -1,   0, -35,   9,   7,  18,  -7, -21,   3,  -1,  12,  -3,  -3 },
     {    4,   0,  -1,   0,   0,   0,  -2,  -1,  -1,   0,  -1,   0,   4,   2, -12,   4,   2,   0,   1,   1,   1,   0,   0,   0,  14,   8,   8,  -2,   0,  -1,   0,  -1,   0,   0,   0,   0,  -3, -14,  -6,   0,   2,   1,   0,   0,   0,   0,   0,   0,  31,  34,  15, -12,  -2,  -4,   0,  -1,  37, -42,  27,   6,   0,   2,   0,   0, -18, -21,   1,  11,  -1,   3,  -1,   2,  -5, -26,  -7,   7,   0,   2,   0,   1,   1,   1,   0,  -1,   3, -48,  -2,  13,  27,  10, -51,   2,  -6,  38,   1, -13 },
     {    1,  -5,   3,  -4,   0,  -1,  -1,   1,   2,   0,   0,   0,   4,   2,  18,  -5,   0,  -3,   0,  -1,   0,   0,   0,   0,  -1, -26,   7,   3,   0,   0,   2,   0,   0,   0,   0,   1,   5,  10,   7,  -2,  -1,  -1,   0,   0,   0,  -1,   0,  -1,  15, -49,  25,   8,   2,   2,   0,   1, -64, -30, -30,  16,   2,   5,   1,   3,   4,  31,  -7, -13,   0,  -3,   1,  -1,   0, -21,  15,   5,   0,   1,   0,   0,  -1,   3,  -6,   0,  -7, -25,  20,   6,  20, -26, -26,  10,   5,  22, -13,  -6 },
     {   12,  21, 119, -17,   0,  -8,  -5,  -4,  -3,  -3,  -1,  -2,   4,   9,  14,  -3,   1,   0,   0,  -2,  -1,   0,   0,   0,   1,  -3, -12,  -4,  -1,   2,   2,   2,   2,   1,   0,   1,   5,   3,   4,  -2,  -1,   0,  -1,   0,   1,   0,   0,   0,   5,  -3,  -1,  -1,  -1,   0,   0,   0,   7,  -5,   5,   4,   1,   0,   0,   0,  -8, -12,  -5,   2,   0,   1,   0,   0,  -3,   5,   1,   1,   0,  -1,   0,   0,  -2,  -7,   1,   3,  -1,   1,  -2,   1,  -3,  -5,   3,   2,   3,  -5,  -5,   2 },
     {    0,  -7, -20,   4,  -3,   1,   1,   2,   1,   2,   1,   0,   5,  -9,  25,  -5,  -1,  -1,  -1,  -1,   0,   0,   0,   0,   3,  -6,  18,   0,   1,   0,   0,   0,   0,  -1,   1,   0,  12, -13,  32,   1,   0,   0,  -1,   0,   0,   0,   0,   0,  -3, -23,   5,   4,   0,   1,  -1,   1,  16, -19,  30,   2,  -1,   0,  -1,   0, -42, -40,  -8,  17,   0,   3,   1,   2, -33,  14,  -5,  -3,   1,  -2,   1,   0,   0, -23,   5,   2,  -8,   6,  11,  -3,  -6, -51,  16,  15,  24, -17, -38,   7 },
     {   -3,   1,   3,   1,  -4,  -3,   1,   2,   0,   1,  -1,   1,  -8,  -9, -17,   6,   1,   2,   2,   1,   1,   1,   0,   1,  -3,  -3,   6,   0,   1,   1,   0,  -1,   0,   0,   1,   0, -16, -22, -11,  12,   3,   3,  -1,   1,   1,   1,   0,   1,  -6,  11,   1,   4,   0,   0,  -1,  -1, -24, -29, -16,  18,   2,   6,   0,   2, -75,  19, -50,  -1,   4,   0,   2,   1,   5,  35, -13, -23,   3,  -4,   3,  -3,  15,  14,  17, -12,   6,   2,   0,  -5,  -4,  23,   7,  -9,  -3,  -8,  20,   0 },
     {    0,   1,   7,   1,   6,  -3,  -2,  -1,   1,  -1,   0,   1,  -6,  -6,-112,  13,   9,   4,   5,   2,   2,   2,   0,   2,  -3, -28, -22,  -8,   1,   0,   4,   3,   3,   0,   1,   0,   0,  -1,  19,  -5,  -1,   0,   1,  -1,   0,  -1,   0,   0,  -2, -20,   7,   5,  -1,   2,   0,   0,  -6,   8,  15,   4,  -2,   0,  -1,   1,   5,  -9,   0,   1,   2,   2,   2,   0, -13,   1,   5,   2,   3,   0,   0,   0,   3,  -5,  -3,   2,   2,   5,   7,  -2,  -1, -13,  -2,   7,   5,   5,  -9,  -1 },
     {   -3,   0,   5,  -2,  -3,   1,   1,   1,   0,   1,  -1,   0,  -5,   2,   3,  -2,  -1,  -3,   0,  -1,  -1,   0,   0,   0,  -4,   0,   7,   5,  -2,   0,   1,   0,  -2,   0,  -1,   1,  -2,   8,  -1,  -1,   1,  -3,   1,   0,   1,   0,   1,   1,  -4, -11,  -2,   4,  -1,   6,   1,  -1, -12,   3, -11,   4,   1,   2,   1,  -1,   4, -10,  12,  10,   0,   3,  -1,   2, -84,  -9, -58,  18,   9,   3,   4,   1, -19,  25, -25, -31,  13,   1,  14, -13,   6,  24,   3, -11,  -1,   1,  20,  -2 },
     {   -1,  -1, -12,  18,   0,   3,  -1,  -4,  -1,   1,  -1,   1,   4,  18,  27,  15,   0,   0,  -2,  -2,  -1,  -1,   0,  -1,  -6,  -9,-107,  15,   3,   6,   4,   4,   3,   3,   2,   1,   0, -18,  -7, -16,  -1,   0,   3,   0,   2,   0,   0,   0,  -9,  -5,   2,   1,   2,   2,   0,   1,   1,   7,   8,  -4,  -2,   1,   1,   0,  -6,  -2, -10,   7,   1,  -1,   1,   1, -13,   5,  -8,  -7,   2,   0,   0,   1,  25,   4,   4,  -1,  -2,   1,  -1,  -2,  10,  -5,  -9,   2,   1,  17,  -2,  -7 },
     {   -6,  -3, -15,-114, -18,   3,  12,   7,   3,   3,   3,   1,  -2,  -1,  -5, -20, -13,  -4,   2,   2,   2,   0,  -1,   1,  -1,  -4, -19,  24,   1,  -2,  -3,   0,   0,  -1,   0,   1,   0,  -2,  17,   1,  -1,  -1,  -1,   1,  -2,   2,  -1,   2,  -1,  10,  -3,  10,   5,  -2,  -2,  -1,   4,  -3,  -6,   4,   0,   2,  -3,  -4,  -1,   1,   9,   4,   1,   2,  -1,   0,   5,   4,  -6,   1,   2,   1,   1,   1,   8,  -6,   1,   0,  -3,  -4,   3,  -2,   4,   3,   3,   2,  -2,  -1,  -2,   0 },
     {    0,  -2,  -1,   5,   2,   4,  -1,  -1,  -2,   1,   0,   0,  -2, -10,   0,  -5,  -6,   2,   0,  -1,   2,   0,   2,   2,   2,  -5, -25,   0,   3,   3,   1,   3,   3,   0,   1,   1,  -9, -39,   7, -15,   4,   9,   1,   3,  -1,   0,   1,   0,   5,  14,  61,  -2,  -1,   0,  -5,  -2,  -6, -23,   9,   0,  -4,  -1,  -3,   0,  -7,  17,  29,  -8,  -6,   1,  -1,  -2,  10, -12,   2,  21,   0,   2,   0,   2, -55,   5,  -9,   9,  -1, -13, -13,  -5,  -9,   5,  24, -11,   4, -47,   9,  14 },
     {   -1,  -2,  -3,   8,   7,   0,  -4,  -2,   4,  -2,  -3,   0,   1,   5,  14,   3,  -5,   3,   4,  -2,  -3,  -2,   1,  -1,  -5, -16,   3, -38,  -6,  -1,   6,   1,   0,   1,   1,   1,  -3,  14,  94, -28,  -8,   1,  -6,  -2,  -2,  -3,  -1,   0,   0,  36,  13,  12,  -6,   1,  -4,  -1,   4,  17, -17,  13,  -2,   3,   4,  -1,   4,   4, -18,   0,  -2,   0,   1,   0,  -3,  -1,  -4,  -8,  -2,  -2,   3,   1,  10,   7,   0,   5, -19,  -2,  -6,  -1,   0,  12,  -4,  -5,  -9,   9,  14,  -5 },
     {    3,   0,   4,   6,   3,  -4,   1,  -5,   0,  -2,   0,  -1,   1,  -4,  -3,  -5,  -2,  -2,   3,   0,   0,  -1,   1,   2,   5,  -2,  10,   8,  -3,   4,   1,   1,  -1,   0,   3,   0,   1,  -6,  -2, -14,   6,   2,   1,  -1,   0,   2,  -1,  -1,   6,  -4,  21,   7,  -6,  -3,  -4,   2,  -5,  -7,  -7, -18,   1,   0,   2,   1,  14,   5,  38,   3,  -6,   3,  -4,  -1, -12,  -1, -22, -11,   8,   0,   2,   2,  64,   5,  54, -18,  -7, -18,   0,  44, -37,   4,  -9,   0,   4, -28,   1,   9 },
     {   -1,   0,   1, -18, -17,   5,  -7,  13,   5,   6,  -4,   2,   5,   4,   7, 109,  15, -10,  -1, -10,  -6,  -3,  -3,  -1,   3,   1,  16,  12,   8,   4,   8,  -2,   2,  -2,   2,   1,   0,   2,  -3,  -6,  17, -12,  -4,  -3,   1,   0,   2,   0,   0,  -2,   4,  -2,  -3,   2,  -7,  -5,  -2,   3,  -3,  -1,  -9,  12,   3,   1,   4,   0,   0,  -1,   3,  -1,   6,  -1,   1,   0,  -7,   3,  -2,   1,   0,  -3,  -2,  -8,  -6,   9, -21, -11,  -8, -11, -15,   0,  -6,  -1,   1, -11,  -1,  -1 },
     {    2,  -1,   0,  19,  -9,  -6,  -3,   2,   1,   0,   0,  -1,   2,   1,  -8,  11,   5,   1,  -2,   1,   4,  -3,   1,   0,   3,  -1, -16,  23,   0,   3,  -1,  -2,  -1,   1,  -2,   1,   2,   5,  25,  11,   5,  -5,  -3,  -4,   0,   0,   0,   1,  10,  25, -37,  32,   4,  -2,  -5,  -2,  10, -40, -56, -26,  18,   5,   6,   0,  -7, -24,  25, -37,  -7,   1,   0,   1,   2,  -3,  13,   6,  -7,   0,  -4,   1,  -6,   0,   1,  -5,  35,  -6,  15,  -8,   1,  -4,   8,   8,   0, -10, -11,   6 },
     {   -2,  -2,  -7, -14,  19, -16,  -2,   0,   7,  -4,  -1,   1,  -1,   1,   5,   6,  -4,  10,   0,   2,  -3,  -5,   2,   1,  -5,  -7, -12, -60, -21,  -1,  13,   5,   1,   5,   2,   1,  -3,   5,  -3,  -9,   2,   5,   5,  -2,   1,  -7,   2,  -1,   0, -14, -60, -13,  13,   6,  -1,   4, -13, -39,  20, -18, -10,  11,  10,   2, -13,  17,  36,  14,  -8,  -4,  -5,  -6,   0,   8,   3,   7,   4,  -8,   1,  -3,  -7,  12,  -4,  13, -12, -10,  -9,   0, -17,  13,   2,   1,  11,   1,   1,  -5 },
     {   -1,  -1,   2,  20,  -8,  -5,   6,  -1,   0,  -6,  -3,   5,  -4,  -1,  -9, -19,   2,  -4,   8,   2,  10,   1,   0,   1,  -2,  -4,  -2,  45, -10,  -4,   1,   4,   0,  -1,  -3,  -1,  -1,   8,   4,  15,   3,  -3,  -7,   0,   2,   0,   0,   3,   0,  -7, -27,  22,  10,   8,   1,  -8,  -3, -10,  -1,   5,   3,   1,  -5,   2,   0, -10,  -1,   6,   7,   8,   3,  -3,  -9, -11,   6,  -2,   8,   3,  -3,  -4,  -7,  -7,   1,  14, -85, -18, -32,   4,  11,   8,   9, -28,   3,  -5,  18,   8 },
     {   -3,  -8,   2,  14,-100, -30,  14,  23,   1,  -5,   5,   0,   1,   0,   1,  -6,  13,   0,  12,   1,   0,   5,  -3,   1,   1,  -3,  -4, -26,  28,   8,   3,  -4,  -1,  -6,   3,  -1,  -1,  -1,   4, -13, -10,  -8,  -1,   2,   0,   2,  -1,   1,   1,  -4,  -9,  -5,  13,  -2,   6,   1,  -2,  -6,  14,   8,   4, -14, -13,   6,   4,   8,   9,   6,  11,   3,   3,  -4,   4,  -5,   2,   0,  -1,   7,  -3,  -2,  -2,   0,   9,  -7,   5,   3,  10,   9,   4,   2,   7,  -1,  -9,   1,   8,   3 },
     {   -2,  -2,  -5, -10, -18,   0,   5,  -1,  -6,   3,   4,   2,  -1,  -1,  -3, -11,  24,  -9,  -3,  -2,  -4,  -5,  -5,  -2,  -2,   3,  -2, -45, -23,  -7,  10,   6,   1,   5,  -1,   2,  -4, -12, -35,  -9,  10,   1,   5,   0,   4,   2,   2,   0,  -3,   7,  34, -13,   7,   9,   3,   1,   4,   9, -61,  -2,   7,  13,   2,  -2,  -4, -45,   9,  -3,  -5,   4,   1,   2, -16,  17,  21,   8,   6,  -1,   7,  -4,   5,  -2,   2,  15, -29,  12,  -8,  -2,   7,  -1,  -2,  -2,   8,  10,  -7,  -8 },
   },
   {
     {   97, -32, -12,  -4,  -2,  -1,  -1,  -1,  -1,   0,   0,   0, -67,  26,   8,   2,   1,   1,   1,   0,   0,   0,   0,   0,  -7,  -7,   4,   1,   1,   0,   0,   0,   0,   0,   0,   0,  18,  -3,  -4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -5,   4,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {  -39,  30,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,   0, -75,   4,  14,   2,   4,   1,   2,   1,   1,   0,   1,   0,  80, -28, -11,  -2,  -3,  -1,  -1,   0,  -1,   0,   0,   0,  -8,  18,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -11,  -4,   6,   0,   1,   0,   0,   0,  10,  -4,  -2,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 },
     {   24, -59,  17,   3,   4,   1,   1,   1,   1,   0,   1,   0,  31,  13, -14,  -1,  -3,  -1,  -1,   0,  -1,   0,   0,   0,  59,  21, -15,  -4,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   0, -74,   9,  18,   2,   3,   1,   1,   0,   1,   0,   0,   0,  -1, -16,  -1,   2,   1,   1,   1,   1,  13,   5,  -6,  -1,  -1,   0,   0,   0,  -5,   5,   1,  -1,   0,   0,   0,   0,  -4,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {  -55, -62,  32,  10,   7,   4,   3,   2,   2,   1,   1,   1, -24,  64, -22,  -4,  -4,  -2,  -1,  -1,  -1,   0,   0,   0,  -7,  10,   6,  -4,   0,  -1,   0,  -1,   0,   0,   0,   0,  34, -38,   3,   3,   0,   1,   0,   0,   0,   0,   0,   0,  -2,  12,  -7,   0,   0,   0,   0,   0,  -1,   1,   3,  -1,   0,   0,   0,   0,   4,  -1,   0,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   1,  -1,   0,   0,   1,   1,   0,   0,   0,  -1,   0,   0 },
     {   10, -19,   3,   1,   1,   0,   0,   0,   0,   0,   0,   0,  44,   6,  -5,  -5,  -2,  -2,  -1,  -1,  -1,   0,   0,   0,  26, -35,   2,   3,   1,   1,   0,   1,   0,   0,   0,   0,  43,  34, -23,  -3,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0, -77,  26,  21,  -1,   3,   0,   1,   0,   9, -30,   1,   4,   1,   1,   0,   1,   4,   9,  -8,   0,  -1,   0,   0,   0,  -5,   2,   2,  -1,   0,   0,   0,   0,  -3,   2,   1,   0,  -1,  -1,   0,   0,  -2,   1,   0,   0,  -1,   0,   0,   0 },
     {   -1,  15, -10,   3,  -3,   1,  -1,   0,   0,   0,   0,   0,  14,  61, -12, -11,  -5,  -4,  -1,  -2,  -1,  -1,  -1,  -1, -22, -79,  30,  10,   7,   4,   2,   3,   2,   1,   1,   1, -45,  18, -13,   3,   1,   1,   1,   1,   1,   1,   0,   0,  33,   7,  -2,  -7,  -2,  -2,  -1,  -1,  10,  -8,   3,   2,  -1,   0,   0,   0,  -7,   3,  -1,   1,   0,   0,   0,   0,   1,  -3,   0,   0,   1,   0,   0,   0,   1,   1,   0,   0,   1,  -1,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0 },
     {   19,   4,  44, -15,  -4,  -5,  -1,  -2,  -1,  -2,   0,  -1,  13, -18, -31,  11,   4,   3,   1,   1,   1,   1,   0,   1,  36, -29, -12,  -1,   1,   1,   0,   1,   0,   0,   0,   0,  20,  -3,   6,  -1,  -3,   0,  -1,   0,   0,   0,   0,   0,  41,  45, -19,  -4,  -4,  -2,  -2,  -1, -63,   9,  23,   1,   3,   0,   1,   0,   3, -22,  -6,   3,   1,   1,   0,   1,   1,   9,  -5,  -2,  -1,  -1,   0,   0,  -1,   3,   1,   0,  -5,   1,   2,   0,  -1,  -1,   0,   0,  -3,   1,   0,   0 },
     {   11,  33,  71, -29,  -9,  -8,  -3,  -4,  -2,  -2,  -1,  -1, -19,  -2, -58,  24,   4,   5,   1,   3,   1,   1,   1,   1, -27,  -1, -10,   1,   4,   2,   2,   1,   1,   0,   0,   0, -10,   7,  25,  -9,  -2,  -2,   0,  -1,   0,   0,   0,   0, -24, -14,   3,   5,   0,   1,   1,   1,  42, -14, -10,   1,  -1,  -1,  -1,   0,  -1,  11,  -4,  -2,   0,   0,   0,   0,  -1,  -3,   6,  -1,   0,   0,   0,   0,   1,  -2,  -1,   1,   2,  -1,   0,   0,   0,   0,  -1,   0,   2,   0,   0,   0 },
     {  -14, -47,  23,   4,   2,   1,   2,   1,   1,   0,   1,   0,  -8, -25,  13,   1,   4,   0,   1,   0,   1,   0,   0,   0, -20, -25, -24,   9,   3,   4,   1,   2,   1,   1,   0,   1,  27,  67,  -6, -18,  -6,  -5,  -2,  -3,  -1,  -1,  -1,  -1,  39, -44,  18,   8,  -1,   1,   0,   1,  -8,  11, -20,   4,   1,   1,   0,   0, -17,  17,   6,  -6,   1,   0,   0,  -1,   8,  -8,   0,   1,  -1,   0,   0,   0,   0,   4,  -1,   0,  -1,   1,  -1,   0,  -1,   2,   0,   0,   0,   0,   0,   0 },
     {   12,   9,  -7,  -3,  -2,  -1,   0,  -1,   0,   0,   0,   0,  23,  -3, -11,   2,  -1,   1,  -1,   0,   0,   0,   0,   0,  33, -18,  16,  -8,  -1,  -2,   1,  -1,  -1,  -1,   0,   0,  42, -28,  -1,   4,  -2,   1,  -1,   0,   0,   1,   0,   0,  13,  -8, -24,   6,   0,   2,   0,   1,  47,  51, -15, -11,  -5,  -3,  -2,  -2, -59,  18,  27,   2,   2,   0,   1,   0,   0, -20,  -2,   5,   1,   1,   0,   1,  -2,   5,  -6,   0,  -1,   4,   1,  -1,  -4,   2,   2,   0,  -1,  -1,   0,   1 },
     {   -9, -39,  17,  10,   1,   1,  -1,   2,   0,   0,   0,   0, -20, -67,   7,   3,  11,   2,   3,   1,   3,   1,   1,   1,  -9, -19,  34,  -7,  -6,   0,   0,   0,   0,   0,   0,   0, -19, -17, -30,  20,   3,   3,   1,   1,   1,   1,   0,   0,  15,  36, -10, -20,   0,  -4,   1,  -1,  45, -23,  13,   2,  -6,   1,  -2,   0,  -2,  13, -15,   5,   0,   0,   0,   0,  -9,  11,   4,  -4,   1,  -1,   0,   0,   5,  -1,  -1,   1,   0,   3,  -1,   1,   0,   1,  -2,   0,   0,   2,   0,   0 },
     {    1,  -2,  11, -11,   1,  -1,   1,   0,   0,   0,   0,   0,   9,   3,  53,  -8,  -7,  -3,  -4,  -2,  -2,  -1,  -1,  -1, -11, -28, -78,  26,  11,   5,   5,   3,   3,   2,   1,   1,  -7, -41,  24, -12,   0,   2,   1,   1,   0,   0,   0,   0,   4,  36,  16,  -3,  -6,  -3,  -2,  -2,  26,  -1, -17,   4,   1,   0,   0,   0,  -5,  -2,   0,   0,   1,   0,   0,   1,  -9,   3,   1,  -2,   0,   0,   0,   0,   2,   0,   2,   0,  -1,   2,  -2,   1,   0,   1,   0,  -1,   0,   0,  -1,   0 },
     {   16,  22,  23,  79, -20,  -7, -12,  -4,  -4,  -2,  -3,  -1,   1,   3, -16, -64,  16,   4,   8,   1,   2,   1,   2,   1,   5,   5, -18, -14,  -2,   2,   1,   2,   1,   1,   0,   1,  11,  -5,   7,  30,  -3,  -1,  -3,   0,  -1,  -1,   0,  -1,  13,  -6,  17, -15,   3,  -1,   0,  -1,  -2,  -7, -11,  10,  -1,   1,  -1,   1,   4,  19,  -8,  -8,   2,  -1,   1,   0, -12,  -1,  11,   1,   0,   0,   0,  -1,   2,  -6,  -1,   1,  -2,   2,   1,   0,  -1,  -1,   0,  -1,  -1,   0,   1,   1 },
     {    6,   2,  -7, -22,   5,   1,   3,   0,   2,   0,   0,   0,  14,   4,   2,  13,  -5,  -1,  -1,  -1,  -1,   0,  -1,   0,  21, -13,   9,   6,  -1,  -1,   0,   0,  -1,   0,   0,   0,  23, -26,  -1,  -7,   4,  -1,   0,   1,   0,   0,   0,   0,  28, -36,  -6,   7,  -1,   4,  -2,   1,   3,  -8, -14,   4,  -2,   1,   0,   0,  62,  50, -25,  -8,  -4,  -3,  -1,  -1, -46,  23,  22,  -3,   1,   0,   1,   0,  -7, -18,   1,   7,  -1,   6,  -5,  -2,   0,   3,   1,  -1,  -5,   2,   1,   0 },
     {   -6, -14, -32,  28,   3,   2,  -1,   0,   0,   0,   0,   0,  -5, -14, -10,   0,   4,   0,   1,   1,   0,   1,   0,   0,  -9, -23,  -1, -16,   9,   0,   4,   0,   2,   1,   1,   0,   6,  19,  77,  -7, -18,  -4,  -7,  -2,  -3,   0,  -2,  -1,  -8,  13, -58,  29,   8,   0,   2,   0,   9, -21,  -1, -17,   6,   2,   2,   1,  10,  -1,  13,  -2,  -8,   1,  -2,   0,  -7,  17,  -4,   1,   0,  -2,   1,  -1,  -2,  -7,   4,   1,   1,  -1,  -5,  -1,   1,   1,   1,  -1,   0,   2,  -1,   0 },
     {    3,  -7,  27,  11,  -2,  -6,   0,  -2,  -1,  -2,   0,   0,   0, -17,  13, -24,   2,   4,   1,   2,   1,   1,   1,   0,  -7, -35,  12,  23,  -3,   1,  -1,   0,   0,   0,   0,   0,  -7, -16,  -4,   8,   8,   0,   2,   0,   0,   0,   0,   0, -34, -39, -17,  -3,   4,   5,   4,   3,   1,  53,   2, -17,  -6,  -4,   0,  -2,  49, -38,  10,  12,  -3,   0,  -3,   0,   4,  13, -24,   5,   2,   1,   0,   0, -19,  17,   7,  -9,   7,  -7,   1,   3,   1,   4,  -2,   0,   0,   1,   0,   0 },
     {   -7,  -7,   7,  -5, -26,   8,   1,   2,   0,   1,   0,   1,  -9,  -7,   7,   5,  20,  -8,   0,  -1,  -1,  -1,   1,   0, -13,  -8,   3,  -1,   7,   3,  -1,   1,   0,   0,   0,   0, -23,   2,  -6,   3, -11,   4,   1,   1,   0,   0,   0,   0, -28,  20, -10,   6,   6,  -3,   1,   0, -35,  30,  -8,  -9,   2,   0,   3,   0,  12,  27,  29, -11,  -2,  -5,  -1,  -2, -45, -57,  21,  16,   2,   2,   1,   1,  29, -19, -27,   2,   2,  14,   1,  -5,   6,  -3,   2,   0,   0,  -1,   0,   1 },
     {   -3,  -4, -12, -14,   5,   6,   2,   0,   0,   0,   0,   0,  -7, -10, -19, -39,   3,   6,   7,   4,   2,   2,   1,   0,   8,  13,  19,  90,  -9, -14, -11,  -6,  -5,  -2,  -2,  -1,   5,   4,  23, -22,  13,   1,  -1,   1,   0,  -1,   0,  -1,  -2,   5, -19, -37,  -5,   7,   7,   1,  -5, -14,  -7,  25,  -1,  -1,  -4,   1, -10,   9,  10,  -4,   5,  -3,   0,   0,   1,  -9,   6,   2,  -1,   0,  -1,  -1,   3,   0,  -6,  -4,   0,   3,   1,   0,   0,  -1,   0,   0,  -1,   1,   0,   1 },
     {    3,  -6,   5,  -7, -16,   6,   1,   2,   2,   1,   1,   0,   7, -15,  13,   6,  16,  -7,   0,  -2,  -1,  -1,   1,  -1,   9, -25,  29, -11,   6,   4,  -2,  -1,   0,  -1,   0,   1,  -3, -45,  27,   5,  -7,   0,  -2,   2,   0,   1,   0,   1,  -5, -25,  20,   8,   5,   0,  -1,   0, -27, -41, -34,  10,   2,   8,   1,   2,   1,   9,   0, -11,  -1,  -4,   0,  -1,  58, -24,  -6,   8,  -5,   0,  -2,   0,   6,  27, -16,  -2, -15,  10,   6,  -2,  -2,  -2,  -1,   1,   0,   1,  -2,   0 },
     {    3,   3,   4,  18,  75, -25,  -3,  -8,  -4,  -4,  -1,  -1,  -9, -17, -10,  -5, -65,  26,  -2,   8,   4,   4,   2,   2,  -5, -15,   5, -10,  -3,  -4,   3,   3,  -1,   0,   0,   1,  -9, -10,   0,  -3,  28,  -2,   2,  -2,   0,   1,   0,  -1, -12,   4,  10,  12, -20,   3,  -5,   1, -18,   1, -14,   0,  10,  -4,   3,  -1,  -3,  10,   9, -15,  -6,   1,   0,   1,   1, -15,   6,   8,  -1,  -1,   0,   0,   9,  -8, -12,   4,  -1,   8,   1,  -3,   3,  -1,   2,   0,   1,   1,   0,   1 },
     {   -5, -11, -41,  27, -13,  13,  -2,   0,   0,   2,  -1,   1,  -7, -18, -58,  23,  17,  -1,   4,   3,  -1,   1,   1,   1,  -1, -10,  -6,   7,   7,  -7,  -2,   0,   2,  -2,   0,   0,  -6, -15,  -8, -37,   1,   8,   0,   0,   2,   2,   0,   0,   4,  20,  48,  13, -17,  -5,  -3,  -2,   9,  29, -32,   9,   8,  -2,   2,   0,  18, -24,  -1, -13,   4,   1,   0,   1,  -2,   8,   8,   3,  -4,   0,  -2,  -1, -11,   8,   2,  -4,   4,  -3,   0,   0,   1,   2,   0,  -2,  -1,   1,   2,   1 },
     {   -4,  -3,   7,   9, -15,   6,  -4,   3,   0,   1,   0,   1,  -5,  -8,   8,   0,   3,  -1,   1,   2,   0,   1,   0,   0,  -6, -15,  10,   7,  13,  -3,   1,  -2,   1,  -1,   1,   0, -14, -21,   5, -14,   1,   9,   1,   0,   1,   1,   0,  -1, -28, -16,  -1,  16,  -8,   0,   2,   3, -32,  16,  20,   9,   1,  -2,   1,   0, -48,  19, -31,  -9,  10,   1,   4,   0,  14,  23,  25, -12,  -3,  -4,  -1,  -3, -31, -50,  23,  18,  25,  -8, -28,   0,   0,   6,   2,  -4,   7,  -4,   0,   0 },
     {    8,  13,  32,  24, -10,  -1,  -2,  -1,  -5,  -4,   0,  -3,   3,   6,  25, -11,  -7, -10,  -2,  -1,   2,  -1,   0,  -1,   9,  21,  19,  10,  18,  -5,  -3,  -3,  -2,  -1,   0,  -1,  -4,   1, -33, -59, -14,  25,  12,   2,   2,   2,   1,   2,   9,  13, -25,  41, -11, -15,  -3,   0,  12, -14,   4,   5,  21,  -4,  -3,  -1,  10,  10,  20, -30, -13,   8,   1,  -1,   8,   7, -16,  17,   0,  -3,   0,   1, -11,   8,   0,   0,  -9,  -6,   9,   1,   2,  -3,   0,  -2,  -1,   1,   2,   1 },
     {    1,   4,  22, -19,   7,   1,   6,  -1,  -1,  -4,   0,  -1,   2,   6,  25, -15,  14,  -9,  -8,   0,   4,   0,   1,   0,   1,   7,  43, -15, -52,  10,   1,   5,  -2,  -1,   0,   1,   2,  12,  26,   2,   6, -10,   5,  -1,  -2,   0,  -1,   0,   6,  26,  26,   7,   1,  -1,  -9,  -1,  -4,   4, -51,   7,   3,   0,   9,   0,  -8, -23,  22, -13,  -2,   4,   0,   2, -25,  24,   2,   6,  -4,  -2,   1,   0, -13, -16,  11,   5,  21, -15,  -8,   4,   2,   0,  -5,  -2,   4,   2,  -1,   2 },
     {    2,   1,  -4,  16,  13,  -8, -15,   1,   0,   0,   0,   0,   3,   2,  -2,   4,  32,  -6,   3,  -4,  -2,  -3,   0,   0,  -3, -11, -11,   3, -69,  26,   7,   6,   1,   1,   0,   1,  -1, -13,  -1, -41,  34, -17,   1,   3,   0,   0,   0,   1,  -3, -11,  -4,  29,  20,  -3,  -2,  -2,   3,  -4,  41,   9, -26,   8,  -3,  -2,  -5,  -2, -12, -11,  11,  -8,   5,   1,  -4, -14,  -3,   0,  -1,   3,  -1,   1,  20,   7,  -5,   2,  -9,   8,   3,   0,   1,  -3,   0,   0,  -1,   1,   0,   1 },
     {    3,  -3,   9,   1,  -8,  -1,   0,  -2,  -1,   2,   3,   1,   5,  -2,  12, -12,  -8,  10,   8,   3,   0,  -3,   0,  -1,   9,  -2,  30,   2,  31, -11, -10,  -3,  -2,  -2,  -1,   0,  10,  -6,  21, -19, -10,  -1,   0,   0,   1,   0,   0,   0,   6, -23,  30,  14, -11,   1,   1,  -1,  11,  -7,  21,  -5,   7,  -6,  -2,   0, -16, -53, -23,  12,   0,   5,   1,   4, -39,  -1,  -2,  -3,   3,   0,   0,  -1,  62, -13,  -8,   3,  -8,  17, -10,   2,  -5,  -3,   2,  -1,   0,   2,  -5,   3 },
     {   -6,  -8, -12,  -4, -27, -82,  17,  10,   6,   0,   1,   1,   1,   4,   1,   1,  19,  69, -22,  -7,  -3,   0,   1,   1,   1,   4,   3,   9,   3,  -6,  10,  -1,   1,  -3,  -2,   0,  -4,  -4,  -6,  -3, -12, -23,   3,   0,   3,   5,   0,   1,   3,   7,   6,  10,  11,  18,  -9,  -3,  -1,  -4,  -1,  -1,   1, -13,   8,   2,  -1,   8,   1,  -5,  -8,   7,  -1,  -3,   0,  -3, -10,   4,   2,  -5,   2,   1,  -6,  -1,   8,  -3,   2,  -3,  -1,   0,  -1,   7,   0,  -2,  -1,  -2,   0,  -1 },
     {    2,   4,  12,  44, -16, -17,   2,   3,  -2,   2,   1,   0,   5,   9,  23,  58, -16, -10,  -1,  -6,   1,  -1,  -1,  -2,   2,   1,  14,  -7,  12,  10,  -1,  -2,  -1,   1,   0,   0,   2,   6,  14, -20,  16, -14,   4,  -5,   2,   1,  -2,   0,  -9,   0, -18, -50,  -9,  21,   0,   2,  -3,  10, -14,  43, -14, -11,   0,   4,   0,  -3,   5,   0,  18,  -7,   0,  -2,  -9,   5,   0, -19,  -4,  10,   2,  -1,   9,   6,  -3,   9, -12, -11,  10,  -1,   7,  -3,  -7,   0,   0,   1,  -2,  -4 },
     {    3,   0,   1,   6,  -4,  -5,  -5,   4,   1,  -6,   1,  -1,   8,   5,   1,  12,   1,   5,  -4,  -3,  -1,   1,  -2,   2,   4,   2,  -7,   3, -12,   9,  -1,   4,  -1,   3,   1,  -1,   7,   7, -11,   4,  -3,   9,   7,   0,  -2,  -2,   2,  -1,   8,  -1, -10, -15,   6,  -6,   2,   5,  17,  -5,  -4,   8,   1,  -6,  -5,  -1,  24, -30,  23,  -6,  -5,   6,   1,   0,  39,   4,  10,   1,  -7,  -3,  -7,   2,   1, -53, -35,  23,  16,  53, -27, -17, -23,  20,  25,  -3,   1, -13,  -2,   5 },
     {    2,  -1,   2,  -8,   9,   6, -12,  -2,  -2,   0,  -3,   4,   5,   0,  13,  -5,   7,   8,   8,  -2,  -2,   2,  -2,   0,   3,  -1,   8, -26,   7, -20,   2,   6,  -4,  -1,   6,   0,   6,   2,  25, -16,  -4,  -1, -17,   1,  -1,  -2,   0,  -1,  12,   0,  22, -18, -17,  18,  15,   1,  16,  -6,  50,  18,   4, -12,  -1,   2,  17,  -3,  10,  -8,  15,   1,  -7,   5,  -8, -46,   3,  23, -10,   7,  -3,   0, -52,   8, -11,  -6,  34,  -7, -11,  -2,   6,   8,  -5,   8,  -3,   1,   0,   1 },
     {   -5,  -4,  -7, -36,   1,   0,   4,   2,  -2,  15,   5,  -9,  -2,   1,   2, -39,  11,  16,  23, -12,   7,   0,  -2,  -4,  -6,  -1,  -2, -36,  11,   3, -15,   7,  -5,   2,  -4,  -2,  -3,  11,   4,  -5,   7,  -6,   6, -11,   3,  -7,   2,  -1,  -9,  12, -15,   0,  -9,  10,  -7,   2,  10,  33,   1,  30,  10,  -5, -17,  -9,  13,   2, -17, -37,  14,   9,   5,  -5,  19,  14,  20, -13, -17,  11,  -2,   3,  14,  19,   1,  28, -27,   3,   2, -13,   2, -14,  10,  -2,  -8,   0,  -5,   4 },
     {    0,   2,  -2,   0,  36, -25,   6, -25,  -6,   5,   0,  -5,   0,   1,   1,  -1,  12,  -5,  -6,  17,   0,   0,  -1,   4,  -3,  -4,   5,  -9, -11,  -7,   8,  -4,  10,  -5,   4,  -2,  -4,  -6,   7, -24, -53,  40,  -3,  -3,  -2,   6,   1,  -1,  -4,  -6,   9, -30,  55, -29,   5,   0,  -4,  15,   2,  20,  15,  -2,  -8,  -5,  -1,   3,  -3,  18,  -7,  11,   1,   4,  -3,  -1,  14, -18,  -2,   4,   1,  -8,  -1,   5,  -4,  -2, -18, -11,  -6,   3,   3,   3,   9,   3,   1,   0,  -3,  -2 },
   },
   {
     {  109, -58,   2,  -6,   0,  -2,   0,  -1,   0,  -1,   0,   0,  16,  -5,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -21,  15,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0, -12,   6,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -7,   4,   0,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
     {  -25,  -7,  12,   0,   2,   0,   1,   0,   1,   0,   0,   0, 105, -54,  -1,  -3,  -2,  -1,  -1,   0,   0,   0,   0,   0,  14,   5,  -9,   2,  -2,   0,  -1,   0,   0,   0,   0,   0, -27,  18,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0, -13,   6,   1,   0,   0,   0,   0,   0,  -6,   2,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -2,   1,   0,   0 },
     {  -28, -71,  67, -16,  14,  -3,   5,  -1,   3,  -1,   1,   0, -37, -29,  29,  -1,   5,   1,   1,   0,   1,   0,   0,   0,  30,   2, -21,   8,  -5,   2,  -2,   1,  -1,   0,   0,   0,  23,  11, -15,   2,  -3,   0,  -1,   0,   0,   0,   0,   0,   3,   7,  -4,   0,   0,   0,   0,   0,  -1,   3,  -1,   0,   0,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,  -1,   0 },
     {   34,  16, -26,   5,  -6,   1,  -2,   0,  -1,   0,   0,   0,  -4, -11,   3,   1,   1,   0,   0,   0,   0,   0,   0,   0,  95, -57,   4,  -4,  -1,   0,  -1,   0,   0,   0,   0,   0,  20,   5,  -9,   2,  -3,   1,  -1,   0,  -1,   0,   0,   0, -19,  19,  -4,   0,   0,   0,   0,   0, -14,   8,   1,   0,   0,   0,   0,   0, -10,   4,   1,   0,   0,   0,   0,   0,  -5,   3,   0,   0,   0,   0,   0,   0,  -4,   2,   1,   0,  -3,   1,   0,   0,  -3,   1,   0,   0,  -2,   1,   0,   0 },
     {   -7,  -8,   3,   2,   2,   0,   1,   0,   0,   0,   0,   0,  -9,  50, -40,  13,  -7,   2,  -2,   1,  -1,   0,   0,   0,  52,  51, -40,   1,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0, -44,  31,  12, -10,   5,  -3,   2,  -1,   1,  -1,   0,   0, -28, -11,  21,  -5,   3,  -1,   1,   0,   0, -13,   7,  -1,   0,   0,   0,   0,   3,  -6,   3,   0,   0,   0,   0,   0,   1,  -5,   3,   0,   0,   0,   0,   0,   1,  -4,   2,   0,   1,  -3,   1,   0,   1,  -2,   1,   0,   1,  -2,   1,   0 },
     {    7,  20,  -7,  -2,  -2,   0,  -1,   0,  -1,   0,   0,   0, -42, -41,  43,  -8,   8,   0,   2,   0,   1,   0,   0,   0,  -7, -18,  11,   0,   1,   1,   0,   1,   0,   0,   0,   0, -83,  43,  -6,   7,   1,   0,   1,   0,   0,   0,   0,   0, -12,  -7,   3,   0,   2,   0,   1,   0,  20, -14,   0,   0,  -1,   0,   0,   0,  13,  -5,  -3,   1,  -1,   0,   0,   0,   7,   0,  -3,   0,   0,   0,   0,   0,   4,  -2,  -1,   0,   3,   0,  -1,   0,   3,   0,  -1,   0,   2,   1,  -1,   0 },
     {    4,  26, -26,  10,  -9,   5,  -3,   2,  -2,   1,  -1,   1, -12, -24,   7,   6,   1,   2,   1,   1,   0,   0,   0,   0, -24,  37, -33,  11,  -4,   1,   0,   1,   0,   0,   0,   0,  44,  59, -44,   1,  -5,  -2,  -1,  -1,  -1,  -1,   0,  -1, -10,  40,   3, -11,   4,  -4,   1,  -2, -25,  -1,  22,  -6,   3,  -1,   1,   0,  -7, -13,  12,   0,   1,   0,   0,   0,   2,  -7,   2,   0,   0,   0,   0,   0,   2,  -4,   2,  -1,   1,  -3,   2,   0,   0,  -3,   2,   0,   1,  -2,   1,   0 },
     {   22,  53,  43, -70,  28, -23,   8,  -9,   2,  -3,   1,  -2,  17,  32,  22, -37,   9,  -9,   2,  -3,   0,  -1,   0,  -1,   0,  -2, -14,  21, -15,   9,  -5,   3,  -1,   1,  -1,   1,   2,  -3, -11,  14,  -6,   3,  -1,   0,   0,   0,   0,   0, -10,   2,   2,   1,   1,   0,   0,   0,  -9,  -3,   6,   0,   1,   1,   0,   0,  -3,  -6,   2,   2,   0,   0,   0,   0,  -1,  -3,   0,   1,   0,   0,   0,   0,  -1,  -2,   0,   1,   0,  -1,   0,   1,   0,  -1,   0,   1,   0,  -1,   0,   1 },
     {   19,  25, -13,  -5,  -2,  -1,  -1,   0,  -1,   0,   0,   0,  -4, -31,  11,   0,   3,   0,   1,   1,   0,   0,   0,   0,  28,  37, -36,   8,  -7,   0,  -1,   0,  -1,   0,   0,   0,   2,  -4,   3,  -1,   1,  -1,   1,  -1,   0,   0,   0,   0,  83, -46,  -1,  -2,  -3,   2,  -2,   1,  17,  17, -12,   2,  -3,   1,  -1,   0, -14,  16,   0,  -2,   2,  -1,   0,   0, -11,   8,   2,  -1,   0,   0,   0,   0,  -6,   2,   2,   0,  -6,   3,   1,   0,  -3,   0,   1,   0,  -3,   1,   1,   0 },
     {    9,  12,  -4,   3,  -7,   2,  -1,   1,  -1,   0,  -1,   0,  -1, -29,  34, -16,  10,  -4,   3,  -1,   1,  -1,   1,   0,   8,  35,   4, -20,   7,  -7,   1,  -3,  -1,  -1,   0,  -1,  44, -15,  32, -18,   2,  -3,  -1,  -1,  -1,   0,  -1,   0, -54, -41,  39,   3,   0,   5,  -1,   2,   8, -35,  -6,  12,  -4,   5,  -1,   1,  14,  -1, -15,   6,  -2,   1,  -1,   0,   4,   4,  -6,   2,  -1,   0,   0,   0,   0,   3,  -3,   1,   0,   2,  -2,   1,  -1,   2,  -1,   1,   0,   1,  -2,   1 },
     {   -8, -16, -10,   9,   1,   2,   0,   2,   0,   1,   0,   0,   8,  28,  34, -49,  18, -14,   5,  -5,   1,  -2,   0,  -1,  19,  34,  46, -52,  14, -11,   1,  -4,  -1,  -1,   0,  -1,  -7,  21,   2,   3, -10,   4,  -4,   1,  -1,   1,   0,   0,  30,   7, -12,  12,  -9,   3,  -2,   0, -33,   5,  13,   1,   2,   1,   1,   0, -19, -15,  12,   2,   1,   1,   0,   1,  -1, -10,   0,   4,  -1,   1,   0,   0,   3,  -6,  -1,   2,   1,  -2,  -1,   1,   1,  -3,   0,   1,   0,  -1,  -1,   2 },
     {    1,  -9,   5,  10,  -7,   4,  -3,   2,  -1,   0,  -1,   0,  24,  36,   4, -11,  -5,  -1,  -2,  -1,  -1,  -1,  -1,  -1,   2, -16,  11,  -8,   4,  -2,   1,   0,   0,   0,   0,   0,  26,  47, -41,   7,  -6,   0,  -1,   0,  -1,  -1,   0,  -1,  11, -14,  -1,   1,   0,   0,   1,   0,  78, -40,  -7,   1,  -3,   2,  -2,   1,   6,  18,  -8,   1,  -2,   0,  -1,   0, -11,  10,   1,  -1,   1,  -1,   0,   0, -11,   6,   2,  -1,  -6,   1,   1,   0,  -4,   1,   1,   0,  -3,  -1,   1,   0 },
     {  -13, -26, -54, -27,  64, -29,  20,  -5,   7,  -2,   4,  -1,  -8, -21, -42, -14,  36, -12,  10,  -1,   3,   0,   1,   0,  -3,  -8,  11,   2, -15,  13,  -7,   4,  -3,   2,  -1,   1,  10,  16,  13,  -3, -15,   5,  -5,   1,  -1,   0,  -1,   0,   9,   1,  17,  -3,  -4,   0,  -2,   0,  11, -12,   9,   4,  -3,   1,  -2,   0,   6,  -4,  -2,   6,  -3,   2,  -1,   0,   1,   4,  -3,   1,  -1,   0,   0,   0,  -1,   3,   0,   0,  -1,   2,   0,   0,   0,   2,   1,   0,  -1,   1,   0,   0 },
     {   10,  10,  34,  -8,  -8,   0,  -3,  -2,  -1,  -1,  -1,   0,   5,  -3, -14,  20, -12,   5,  -2,   1,  -1,   1,  -1,   1,  -3, -33,  10,  -8,  14,  -6,   5,  -1,   1,   0,   0,   1,   7,  28,  26, -38,  16, -10,   1,  -2,   0,  -1,   0,   0,  49,   2,  40, -21,  -3,  -1,  -3,   0, -20, -33,  36,   6,  -3,   5,  -2,   1,   4, -35,  -6,  14,  -3,   4,  -1,   1,   8,  -1, -13,   5,  -1,   1,   0,   1,   2,   3,  -6,   0,  -1,   3,  -3,   2,   0,   2,  -3,   1,  -1,   3,  -2,   0 },
     {    5,  16,  21,  -8,  -7,  -1,  -3,  -1,  -1,   0,   0,   0,  -8, -13, -28, -21,  35, -13,  10,  -3,   4,  -1,   2,  -1, -10,  -8, -35, -45,  49, -19,  12,  -2,   4,  -1,   1,   0,  -2,   0,  13, -28,   6,   2,  -2,   3,  -1,   2,   0,   1,  -9,  36,   3,  -4, -14,   6,  -5,   2,  37,  33,  -7,  -2,  -9,   0,  -2,  -2,  -6,  16,  12,  -3,  -3,   1,  -1,  -1,  -9,  -6,  12,   1,  -3,   2,  -2,   0,  -3,  -4,   4,   2,   0,  -1,   2,   1,  -2,   0,   2,   1,   0,  -1,   1,   0 },
     {   -4,  -1,   5,  -6,   9,  -5,   2,  -2,   0,  -1,   0,   0,  -1,   5,   3,  17, -10,   2,  -2,   0,  -1,   0,  -1,   0, -16, -19,  21,   8,  -5,   4,  -2,   1,  -1,   0,   0,   1,   9,  42,  14, -24,   4,  -4,  -2,  -1,  -1,   0,  -1,  -1, -20, -42,  17,   2,   2,   4,  -2,   2,  -7,  31, -20,   2,  -2,   1,   1,  -1, -72,  28,  22,  -7,   6,  -2,   3,  -2,  -7, -29,  13,   1,   1,   0,   1,   1,   7, -12,  -1,   3,   6,  -7,  -5,   0,   4,  -1,  -1,   0,   3,  -2,  -1,   0 },
     {    9,  26,  24, -22,  10, -12,   6,  -7,   2,  -3,   1,  -2,  -8, -18, -20,  39, -18,  15,  -6,   3,  -2,   2,  -1,   1,  11,  36,  50, -24, -13,   4,  -6,   0,  -3,   0,  -1,   0,  11,  26,  20, -16,  -2,  -5,   0,  -3,   1,  -1,  -1,  -1,  -1,  23, -46,  21,  -3,  -3,   1,   0,   7,   1, -26,   4,   4,  -3,   2,   0,  23,   8, -12,  -7,   3,  -2,   0,   0,  -7,  14,   5,  -6,   2,  -2,   0,   0,  -7,  -2,   3,  -1,  -4,  -1,   3,  -1,  -2,  -2,   1,   0,  -1,   0,   1,  -1 },
     {    4,  10,  23,  11,  -1, -10,   2,  -5,  -1,  -1,   0,  -1,  -7,  -9, -40,  14,  22, -10,   8,  -2,   1,   1,   1,   0, -10, -15, -12, -26,  26,  -8,   5,   0,   1,   1,   1,   0,  -2,   9, -39,  27, -20,  12,  -5,   4,  -1,   0,   0,   0,  -5, -51,  -9,  34, -14,  11,  -2,   3, -50, -22,  -2,   5,   6,   1,   1,   2,   1,  12, -17,  -6,   3,  -3,   1,  -1,   0,  16,  -2,  -7,   0,   0,   0,  -1,   3,   5,   2,  -3,   0,   2,   3,  -2,   0,   1,   0,  -1,   1,   0,   1,   0 },
     {  -11, -14, -19, -45, -29,  45, -11,  12,   1,   2,   0,   2,  -5,   3, -15, -47, -19,  28,  -4,   7,   2,   2,   1,   2,  -9, -13,  -8,   5,  -2,  -1,   6,  -3,   3,   1,   0,  -1,   3,  27,  14,  -2,  -7,  -8,   4,  -5,   1,  -1,   0,  -1,   6,   2,  19,  15,  -1,  -7,   1,  -3, -28,  14, -28,  22,   4,  -4,   1,  -1,  24,  24, -25,  -3,   4,  -5,   2,  -1,  -3,  19,   4,  -6,   2,  -2,   1,  -1,  -4,   0,   8,   0,  -3,   0,   2,   1,   0,  -1,   1,   0,  -1,  -1,   2,   0 },
     {    4,   3,  11, -17, -28,  22, -12,   7,  -2,   2,  -2,   1,  -6, -21, -33,  -8,  -5,  14,   0,   6,   2,   3,   0,   1,   5,   3,   4,  -5, -15,   4,   0,  -1,   0,   1,   1,   0,   9,  -1,  24,  48, -35,   9,  -6,  -1,  -2,  -1,   0,  -1,  -2,  -4,   1,  31,  -7,  -3,   3,  -2,  35,  -3,  16, -12,  13,  -7,   3,  -2, -26, -37,  22,   3,   7,   1,   0,   2,  24, -33, -17,   9,   0,   2,  -1,   2,  12,   5, -14,   1,   1,   6,  -1,  -2,  -2,   5,  -3,  -1,   1,   1,  -2,   0 },
     {    4,   2,   7,  39,   3, -18,  -1,  -4,  -2,  -1,   0,  -1,  11,  20,  15,  -3,  17, -22,   4,  -5,  -1,  -1,  -1,  -1,  -5, -16, -24,  13,  -6,  10,  -5,   5,  -1,   2,   0,   1,  12,  35,  33,  -6, -11,   3,  -6,  -2,  -2,  -1,   0,   0,   5,  -4, -10,  16,  -1,  -1,   2,   0,  -3,  32, -48,   5,   7,  -4,   3,  -1,  27, -30,  -5,  -1,   3,   0,   0,   0,  49,   3, -23,  -2,  -3,   1,  -2,   1,  -2,  19,   1,  -6,  -5,   1,   0,  -1,  -6,   4,   4,  -1,  -3,   0,  -1,   0 },
     {    4,   9,  22,  39, -11,  -9,  -2,  -5,  -1,   0,  -1,   0,  -3,   3,   5,   2,  14, -13,   5,  -4,   1,  -1,  -1,   0,  -4,   1,   2,  10, -25,  18, -10,   6,  -2,   0,  -1,   1,  -5,   1,  35,  20, -44,  22, -14,   5,  -3,   0,  -1,   0,  12,  36,  41,   0,  -9,  -5,  -2,  -3,  -3, -12,   2,   5,  15,  -8,   4,  -2,  -9,  30, -24,  -3,  11,  -7,   3,  -2, -42,  10,  12,  -6,   3,  -1,   1,   0,  -2, -20,   7,  -3,   5,  -6,  -1,   1,   5,  -5,  -1,   0,   1,  -3,   0,   0 },
     {   -5,  -3, -17, -13,   5,   8,  -3,   5,  -1,   2,   0,   1,   0,   6,  14,  14,   1,  -9,   2,  -5,   2,  -2,   2,   0,  -2,   3, -19,  24,  13, -13,   9,  -7,   1,   0,   1,   0,  -9,  -9,   7, -17,   7,   3,  -2,   4,  -1,   1,  -1,   0,   1,  32, -23,  35, -29,  12, -10,   3,   3, -43, -20,  46,  -8,   4,   0,   0, -55, -13, -21,  11,  10,  -1,   4,   0,  -7,   6, -21,  -7,   7,  -3,   2,   1,   9,  16,  -5,  -7,  -2,   3,   2,  -4,   1,  -2,   2,  -3,   1,   1,  -1,  -2 },
     {   -2,  -3,  -1, -22, -10,  25, -11,   9,  -4,   1,  -2,   1,   3,   2,  -2,  23,  37, -31,  12, -12,   1,  -1,   2,  -1,   8,  10,  30,  56,  30, -52,  13, -14,   0,  -3,   0,  -2,   6,  14,  20,  22,  -5,   4, -10,   2,  -3,   1,  -2,   0,   1,  -4,  -8, -28,   6,  13,  -4,   2,  -4,  -3,  -9, -26,   6,   0,   3,   0,  13,   1,   2,  -9,  -6,   1,   0,   1,  -5,   0,   7,  -4,  -2,   2,  -1,   1,  -2,  -4,   3,  -2,  -2,  -4,   2,   2,  -1,  -1,   0,  -1,  -2,  -1,  -1,  -1 },
     {    0,  -8, -16,  -6,   5,  -6,  10,   2,   1,   0,   0,  -1,   8,   5,  21,  13,  -8,   3,  -4,   2,  -5,   1,  -1,   0,  -2, -18, -30,  -1,   7,  -2,   3,   1,   1,   1,   2,   1,   7,   5,  19,  -3,   0,  -3,  -3,  -2,  -1,  -3,   0,  -1,  11,  -3, -13,  37, -17,  10,  -6,   2,  -7, -29,  -6,  -4,   9,  -1,   7,   0,  45,   0,  38, -24,   4,  -3,  -3,  -2, -28, -42,  32,   9,  -2,   4,  -2,   1,  16, -35, -15,  13,   9,  10, -15,  -1,   2,   1,  -4,   0,  -3,   3,  -4,   0 },
     {   -5,  -5,  -5,  -6,   2,   5,  -3,  -1,  -1,   1,  -3,   1,  -3,   1,   9,   1,  -9,  -3,   0,  -3,   3,  -1,   0,   0, -10, -12, -17,  -7,   5,   8,   1,   0,   2,  -1,   2,   0,   3,   8,  28,   3,   4,   1,  -1,  -2,   0,  -3,  -1,   0, -10, -14, -15,  20,   9,  -4,   6,  -1,   1,  -1,  16, -27,   6,   0,   3,  -1, -15, -36,  31,  -8,  -3,   3,  -2,   3, -43,  58,  -9,  -1,  -2,  -3,   0,   0, -55,   5,  28,  -3,  14, -22,  -2,   3,   3,   0,   0,  -1,   6,  -6,   0,   2 },
     {   -4,  -4, -17, -24,  18,  -5,   6,  -5,   8,  -3,   3,  -1,   6,  11,  30,  43, -34,  -1, -14,  -4,   1,  -5,  -1,   1,   1,   6,   1, -13,  31, -13,   3,  -6,   0,   0,   0,  -2,  -2,  -2, -13,  11, -22,  25,  -2,   7,  -2,   2,   1,   1,   1,   7,  48,  21, -33,   9,  -7,  -1,  10,  37,  18, -10,   8, -17,   4,  -5,  11,   8, -13,  -5,   1,  -2,   1,   3,  17,   8, -12,   1,  -6,   4,  -2,   3, -11,  15,  -1,   6,  -4,  -4,   1,  -1,  -2,   2,  -1,   0,   2,  -3,   2,  -2 },
     {    5,   7,  10,  15,  56,  18, -63,  40, -24,  12, -10,   6,   2,   4,   5,   7,  23,  16, -31,  16,  -7,   3,  -1,   0,  -1,  -3,  -2,  -8, -24,  -5,  32, -26,  16,  -4,   4,  -1,  -2,  -2,  -2,  -7, -17,  -6,  22, -13,   3,   2,   0,   1,  -1,  -1,   3,   3, -10,  -1,   4,   2,  -2,   2,   4,   0,  -2,  -3,  -1,   1,   5,   3,   5,  -2,  -3,  -1,   1,  -1,  -1,   0,   3,  -1,   3,  -1,   0,  -2,  -4,   5,  -1,   0,  -2,   0,   1,   0,   0,   1,   1,   0,  -2,  -2,   3,   2 },
     {    1,   3,   9,  16,   0, -14,  -3,   0,   0,   0,   1,   1,  -5,  -4, -16, -11,  -7,  -5,   8,   5,  -2,   5,   2,   1,   3,   7,  21,  23,  -3, -12,   7,  -6,   1,  -2,   2,  -1,   2,   5, -21,  18,  20, -15,   1,   0,  -1,  -1,  -1,  -1,  -6,  -8,  21,  16, -24,  16, -14,   4,   8,  43,  -6,  34, -27,   8,  -6,  -1,  23, -41,  -4,  23,   8,  -1,   2,  -2, -52, -13, -12,   7,  14,   1,   4,   0, -14, -15, -11,  -5,   9,   0,  -6,  -5,   4,   4,  -3,  -3,   1,   1,   0,   0 },
     {    2,   5,   3,  20,  30, -18,  13,  -2,   1,  -2,  -3,  -2,  -3,  -2,   0, -15, -19,  18,  -3,   6,   0,   1,  -1,  -1,  -3,  -1,  -7,  -8,   5, -33,   7,  -4,   1,   2,  -1,   1,   4,   7,  23,  55,  40, -48,  -3,  -7,  -2,   0,  -6,  -1,   6,  20,  18,  18,  24,  20, -12,   2,  -1,  -7,  -4, -21,   6,  14,   5,   2, -16,  16, -14, -16,  -4,   5,   2,   2,   7,   5,  -5, -12,  -4,   1,   2,   0,   8,   4,   4,   0, -10,   9,   6,   0,  -3,  -5,   1,  -2,   0,   1,   0,  -5 },
     {    3,   2,   6,  11,   6,  -2, -20,  -3,   6,  -6,  -3,   2,  -1,  -8, -11, -15, -12,  -1,   6,  -1,   8,  -2,   5,  -3,   6,   5,  10,  20,   1,  -1,  -7,  -2,   5,  -3,  -3,   3,   1,  -4,  -4,   5,  20,  11, -15,   0,  -4,  -1,   1,  -1,   4,   5,   3,  12, -26,   4,   0,  -1,   4, -11,   6,  17, -20,   1,   7,  -6,  -8,   0,   2, -12,  -6,   3,   9,  -1,  58,  16,  39,   0, -13,  -7,  -8,   2, -29, -44,  25,  22,  23, -17, -25,   6,  -3,  21,  -1,   3,  -3,  -3,   4,  -1 },
     {    0,   0,   5,   9, -15,  -4,   0,   4,   5,  -1,   1,   5,   0,   3,   1,  -7,  18,  -1,   4,  -7,   8,   3,   4,  -5,   1,   2,   9,  30,  -4,   7, -10,   9,  -5,   7,  -5,   1,   3,   8,  16, -12,  14, -17,  15,  -3,  -9,   7,   1,   0,  -5,  -6, -11,  33, -29,  -2,  15,  -4,   4,  11,  64,   5, -26,   0, -13,  -5,  13,  38,  16, -33,   4, -12,   5,  -5,   6,   7, -24,  -8,   6,  -2,   1,  -4,  11,  12,   0,  -8, -21,   1,  23,   5,  -6,  -3,   6,   1,  -2,  -6,   1,   1 },
   }
 },
 {  //23
   {
     {  117, -41,  -6,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0, -18,  12,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -17,   4,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {  -11,  12,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,-115,  21,  12,   3,   4,   1,   2,   1,   1,   1,   1,   0,  13, -29,   4,  -1,   1,   0,   1,   0,   0,   0,   0,   0,  33,  -8,  -4,  -1,  -1,   0,  -1,   0,   0,   0,   0,   0,   6,  -1,  -1,   0,   0,   0,   0,   0,   8,  -1,  -1,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0,   6,  -1,  -2,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   2,   0,  -1,   0,   1,   0,  -1,   0,   2,   0,  -1,   0 },
     {  -13,   2,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0,   4,  16,  -5,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,-107,   8,  16,   2,   5,   1,   2,   1,   1,   0,   1,   0,  29, -43,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  34,   3,  -9,   0,  -2,   0,  -1,   0,  -1,   2,   1,   0,   0,   0,   0,   0,   6,  -2,  -1,   0,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   3,   1,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
     {   -1,  11,  -2,   0,  -1,  -1,   0,   0,   0,   0,   0,   0, -29,   3,   6,   1,   1,   0,   1,   0,   0,   0,   0,   0, -16,  28,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -93,  -3,  24,   1,   6,   0,   2,   0,   1,   0,   1,   0,  27, -56,   1,   1,   1,   1,   0,   1,  31,   1, -13,   0,  -2,   0,  -1,   0,   5,   2,   0,  -1,   0,   0,   0,   0,   6,  -1,  -1,   0,   0,   0,   0,   0,   3,  -2,  -1,   0,   3,   1,  -2,   0,   2,   0,  -1,   0,   2,   1,  -1,   0 },
     {  -10,   1,   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -4,  25,  -4,   0,  -1,  -1,  -1,  -1,   0,   0,   0,   0, -35,   1,   7,   1,   2,   1,   1,   0,   1,   0,   0,   0,  -6,  29,  -3,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0, -82,  -3,  31,  -1,   6,   0,   2,   0,  22, -65,   2,   4,   1,   2,   1,   1,  26,   0, -16,   1,  -2,   0,  -1,   0,   5,  -1,  -1,  -1,   0,   0,   0,   0,   6,   0,  -3,   0,   3,  -2,  -1,   0,   3,   1,  -3,   0,   1,   0,  -1,   0 },
     {   14,  55, -14,  -3,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0, -12,  -7,   6,   1,   2,   0,   1,   0,   0,   0,   0,   0,  -7,  18,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0, -26,   3,   7,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -4,  31,  -1,  -3,   0,  -1,   0,  -1, -63,  -6,  37,  -1,   5,  -1,   2,   0,  16, -64,   1,   8,   0,   2,   0,   2,  18,   2, -17,   1,  -2,   0,  -1,   0,   4,   0,   0,  -1,   3,  -1,  -1,   0,   3,  -3,  -1,   1,   1,   1,  -2,   0 },
     {   31,  95, -30,  -3, -10,  -3,  -4,  -2,  -2,  -1,  -1,  -1,  15,  10,   0,  -2,   0,   0,   0,  -1,   0,  -1,   0,   0, -10, -19,   4,   0,   1,   1,   1,   0,   1,   0,   0,   0,  13,  13,  -4,  -3,  -2,   0,  -1,  -1,   0,   0,   0,   0, -12, -11,   3,   1,   1,   0,   1,   0,  23,  17, -14,  -2,  -2,  -1,  -1,  -1, -27,  30,  16,  -5,   2,  -2,   1,  -1, -12, -17,  17,   2,   1,   0,   1,   0,   1, -10,   0,   2,   1,  -3,   0,   0,  -1,  -2,   1,   0,   0,  -3,   1,   1 },
     {   18,  28,  -6,  -4,  -3,  -2,  -2,  -1,  -1,   0,  -1,   0,   1, -38,   5,   2,   2,   1,   1,   1,   1,   0,   0,   0,  14,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   7, -32,   1,   3,   0,   2,   1,   1,   0,   0,   0,   0,  18,   3,  -8,  -1,  -2,   0,  -1,   0,   9, -36,  -3,   7,  -1,   2,   0,   1,  52,  24, -45,  -1,  -6,   0,  -2,   0, -23,  62,  10, -11,   1,  -3,   0,  -2, -11,  -5,  17,   1,  -4,   1,   1,   2,  -1,  -2,   0,   0,  -3,   4,   1,  -1 },
     {   -9,  -4,   6,  -1,   2,  -1,   0,   0,   1,   0,   0,   0, -19,  16, -12,   3,  -2,   1,   0,   0,   0,   0,   0,   0,   7,  53,  -2,  -7,  -2,  -3,  -1,  -2,  -1,  -1,   0,  -1, -13, -12,   4,   2,   1,   1,   0,   0,   0,   0,   0,   0,  -3,  38,   0,  -5,  -1,  -1,   0,  -1, -25,  -7,  12,   1,   3,   0,   1,   0,  -4,  38,   2,  -9,   0,  -3,   0,  -2, -42, -23,  51,   1,   4,   0,   1,   0,  18, -52,  -8,  12,   3,   3, -10,   0,   5,  -7,  -2,   0,   0,   0,  -1,   0 },
     {    2,  -2,  12,  -3,   1,  -1,   0,  -1,   0,   0,   0,   0, -22,-106,  19,  10,   8,   4,   3,   3,   2,   2,   1,   1, -29,  11, -14,   4,   0,   1,   0,   1,   0,   0,   0,   0,  10,  19,  -4,  -2,  -2,   0,   0,   0,   0,   0,   0,   0, -15,   0,   3,  -1,   1,   0,   1,   0,   7,   0,  -4,   1,  -1,  -1,   0,   0,  -9,   6,   7,  -2,   1,   0,   1,  -1,   1, -28,   3,   7,   0,   1,   0,   1,  16,   1, -21,   1,  -1,  11,   0,  -3,   3,   1,  -2,   1,   0,   4,   0,  -2 },
     {    2,  -7,  11,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   9,  -4,  14,  -3,   0,  -1,   0,  -1,  -1,   0,   0,   0,  -7, -89,  25,   8,   6,   4,   2,   3,   1,   1,   1,   1, -53, -12, -24,   9,   0,   4,   1,   2,   1,   1,   0,   1,  13,  33,  -3,  -7,  -1,  -2,  -1,  -1,  -4, -14,   7,   3,   0,   0,   0,   0,   4,   6,  -3,   0,  -1,   0,  -1,   0, -10, -18,   9,   6,   1,   1,   0,   0,  16, -15, -19,   7,   4,  13, -11,  -3,   3,   6,  -3,  -1,   0,   5,  -1,  -1 },
     {  -13, -12,  -1,   2,   2,   0,   1,   0,   0,   1,   0,   0, -10,  -6,  11,  -1,   1,   0,   0,   0,   1,   0,   1,   0, -33, -17, -18,   7,   1,   3,   1,   2,   1,   1,   1,   0, -14,  61, -14,  -7,  -2,  -2,  -1,  -1,  -1,   0,   0,  -1,   0,  14,  -3,  -4,   0,  -1,   0,   0,  -6,  32,   4,  -9,   0,  -2,   0,  -1,  -5,   2,   3,  -2,   1,  -1,   0,   0,  -8,  44,  15, -16,   1,  -4,   0,  -2, -24, -30,  51,   6,  12, -35, -13,  12,   0,   0,  -2,  -1,   5,  -7,  -3,   1 },
     {   -4,   5,  -9,   2,  -1,   0,   0,   0,   0,   1,   0,   0,  -9,   6, -25,   3,   1,  -1,   2,   0,   1,   0,   0,   0,  -6,  14,  -8,   1,  -3,  -1,  -1,   0,   0,   0,   0,   0, -15,  41, -32,  -4,  -1,   0,   0,   0,   0,  -1,   0,   0,  43,  44,  16, -18,  -2,  -5,  -1,  -3,  11, -17,   3,   6,  -1,   1,  -1,   1,  -1,  25,  -5,  -6,   0,  -2,   0,  -2,   2,  -4,  -3,   1,   1,  -1,   0,   0,   0,  51,  -5, -18, -21,   2,  48,  -5,   0, -25,   9,   9,  -2,  -7,   3,   3 },
     {    3,  -4, -10,   1,  -2,   1,   0,   1,   0,   0,   0,   0,  10,   8,   9,  -4,  -1,  -2,   1,  -1,   0,   0,   0,   0,   0, -10, -11,   0,   0,   1,   0,   1,   0,   1,   0,   0,  27,  47,  12, -16,  -3,  -5,  -3,  -2,   0,  -1,   0,  -1,  58, -38,  46,   6,  -1,  -1,   0,   0, -34, -37,  -5,  18,   0,   5,   0,   3,  -1,  -6, -18,   1,   1,   1,   0,   0,  -3,  -9,   8,   1,  -2,   1,   0,   0,  10, -22, -16,  12,   5,  18, -23,  -4,  -3,  10,   5,  -5,  -2,   2,  -1,   0 },
     {   -8, -17, -58,  19,   0,   7,   3,   2,   1,   1,   0,   1,  -4, -13,  15,  -8,   6,  -1,   1,   0,   1,   0,   0,   0,  -1,  -7,  27,   0,   1,  -1,  -1,   0,  -1,  -1,   0,   1,  -8, -18,   4,   2,   1,   0,   1,   1,   1,   0,   1,   0, -11, -13,  15,  11,   1,   2,   1,   1, -51, -11, -32,   9,   4,   5,   2,   2, -10,  22,   2, -14,   0,  -3,   0,  -1,  -3, -13,   3,  -1,   1,   1,   1,   0,   1,  19,  20, -11,  -5, -40,  20,  19,  13,  -9, -35,   6,  -1,  13,  -2,  -7 },
     {   10,  13,  46, -16,   0,  -6,   2,  -2,  -1,  -1,   0,   0,   3,  -6,  -5,   4,  -1,   1,  -1,   0,  -1,   0,   0,   0,  12,  10,  12,  -4,  -2,  -2,  -1,  -1,   0,   0,   0,   0,   5,  -9,  -3,   6,   1,   0,   0,   0,  -1,   0,   0,   0,  25,  10,  22,  -9,  -2,  -3,  -3,  -1,  30, -30,  20,  11,   0,   2,   0,   1, -34, -14,  -2,  10,   0,   4,   2,   2,   4, -15,  -6,   3,   0,   0,   0,   1,  -8,  -2,  20,   2,   3, -55, -10,  25,   8,  13, -51,  -4,  -2,  20,   4, -11 },
     {   -7, -20, -88,  19,   4,   9,   4,   4,   1,   2,   0,   1,   1, -13,  -9,   1,  -1,   2,   1,   1,  -1,   1,   1,   1,   4,   9,  14,  -8,   2,   0,  -2,   0,   0,  -1,   0,  -1,   3,  -1,   4,   2,  -2,   0,  -2,   1,   0,   0,   0,   0,   9,  24,   1, -11,  -1,  -3,  -1,  -1,  55,  -8,  32,   0,  -2,  -2,  -1,  -1, -15, -22,  -5,  14,   1,   4,   0,   1,   5,   0,  -3,  -2,  -1,   0,   0,   0,   0, -18,  -3,   9,   5,   7, -21,  -1,  -3,  12,   0,  -5,   0,   5,   3,  -2 },
     {    2,   1,  -8,   5,  -2,   1,   1,   0,   0,   0,   0,   0,  -1,  -3, -35,  -2,   2,   2,   1,   0,   0,   0,   0,   1,   4,   1,  -4,   3,  -3,   1,   0,   2,   1,   0,   1,   0,   4,  11, -12,  -9,   1,   0,   0,   0,   0,   0,   1,   0,  11,   3,   9,   1,  -5,  -1,  -2,   1,   5,  15, -14, -14,   0,  -3,   2,  -1,  81,   2,  47,  -5,  -6,  -3,  -4,  -2, -12, -22, -10,  26,   0,   4,   0,   1,  -2,   4,  -3,  -1,   0, -19, -22,  11,   4,  35, -17, -16,  -8,   6,  23,  -3 },
     {    1,  -5,  -2,  -5,   2,   0,  -1,   1,   1,   0,   1,   1,  13,  17,  89, -14,  -8,  -6,  -4,  -5,  -2,  -2,  -1,  -1,   3,   7,  -8,  -1,   4,   2,   2,   0,  -1,   0,   0,   0,  11,   8,  14,  -3,  -4,  -3,  -1,   0,  -1,  -1,   0,   0,   9,   4,   0,  -4,   1,   0,  -1,   0,  24,   3,  22, -12,  -4,  -1,  -1,  -1,  36,  -6,  22,  10,  -5,   1,  -3,   2, -25, -13,   3,  13,   1,   3,   1,   1,  -3,   0,  -8,   0, -10,  -8,  21,   3,   6, -31, -12,  19,  -1,  10, -21,  -2 },
     {    3,  -4,   9,   5,  -5,   1,  -3,   1,  -2,  -1,  -1,   0,  -1, -15, -57,   9,   9,   6,   4,   4,   2,   2,   1,   0,   3, -19,  49,  -5,  -4,  -2,   0,   0,   0,  -1,   1,   0,  10,  20,  28,   2,  -6,  -1,  -3,  -1,  -1,   0,   0,   0,   4, -26,  -6,   2,   2,   2,   3,   0,   3,   7,  23,  -7,  -4,  -2,  -2,  -2,  18, -16,  11,   5,  -1,   1,   0,   0,  -9,   1,  17,   4,  -1,  -2,  -2,   0,   3, -18,  -2,   1,   0,  -3,  24,   5,   0, -47,  -7,  30,   8,   6, -40,  -1 },
     {    1,   4,  14,  -2,   2,  -2,  -1,   0,  -1,   0,   0,  -1,  -3,  -5,  30, -17,   7,  -4,   1,  -1,   0,  -1,   0,   1,   6,  31,  99,  -6,  -8,  -6,  -6,  -4,  -4,  -2,  -1,  -1,   2,  25, -16,  12,  -8,   3,  -2,   0,   0,  -1,  -1,   0,   5,   3,   1,   1,  -1,  -1,   0,   0,  -4,   5,  -6,  -1,  -2,   3,   0,   0,  -2,   6,   5,   0,  -2,   2,  -1,   0,  -7,  11, -15,   3,   0,   1,   2,  -2,  -5,   3,   3,  -4,   3,   7, -10,  -5,  -8,  19,  23, -13,  -1, -21,  13,  13 },
     {   -3,  -1,   0,   6,  -2,  -1,   0,   0,   0,   0,  -1,   0,  -5,   0, -12,  -1,   3,  -1,   2,   2,   1,   1,   0,   0,  -5,  -7, -10,  10,  -1,   1,  -1,   1,   1,   1,  -1,   1,  -3,   0,  12,  -6,   1,   0,   1,   1,   0,  -1,   0,   0,  -5,  -7,   6,   8,  -1,   1,  -3,   0,  -5,   3,  12,   1,  -2,   0,  -2,   0, -17,   2,  -5,  42,  -1,   3,  -1,   1, -82,   0, -58,  -3,  24,   0,   7,   0,  23,  -2,   8, -40,   8,   3,   7,   3,   2,   6,   4,  -5,   2,  -5,   7,   3 },
     {   -1,  -3,   2, -29,  11,  -5,   1,   0,  -2,  -1,  -1,  -1,  -4, -14,  -6, -31,   5,   0,   4,   3,   3,   0,   1,   1,  -1, -16,   0, -14,   7,   1,   0,  -2,   2,   0,   3,   1,  -7, -26,  56, -29,   2,   1,   0,   2,   1,  -1,   0,   0,  -6,  24,  64,   1, -12,  -5,  -8,  -6,   2,  10,  11,  11,  -5,   3,  -1,   3,  -3,  15,  24,   1,  -5,  -2,  -6,   0,   9,   8,   1,   6,  -4,   3,   1,   0, -22,   8,  10,   3,  -5,  -2,  -1,  -2,  -2,   0,  27,   1,   4, -26,  -1,  19 },
     {    8,   7,  27, 113, -17, -12, -11,  -7,  -5,  -2,   0,  -1,   0,  -3,   6, -11,   8,  -6,   0,   0,   0,   0,   0,   1,   0,   2,  -5, -24,   2,   4,   3,   1,   3,   2,  -2,  -2,  -2, -12,  -4, -12,   1,   2,   3,  -2,   0,   1,   0,   0,   0,   6,  23,  -2,  -3,  -1,   0,   1,   6,   7,   2,   5,  -1,  -2,  -3,  -1,   0,  -2,   1,  -1,  -1,  -2,   2,   1,   3,   3,   0,  -1,  -1,   1,   0,   1,  -4,   0,   2,   4,   0,   1,  -5,   3,  -4,  -5,   5,   0,   2,  -3,  -6,   3 },
     {    0,   1,   7,  10, -11,   1,   0,  -2,   0,   1,   1,   1,   0,   4,   3,  20,  -2,  -1,  -2,  -1,  -3,   0,   0,   1,  -5, -11,  -1, -35,   0,   1,   2,   2,   2,   0,   0,   0,  -1,  29,  76,  18, -22,  -3,  -8,  -2,  -1,  -1,   0,   0,   9,  39, -40,  29,   4,   2,   1,   2,   6, -21, -33,  -9,  15,   1,   2,   1,  -5,   1,   0,  -5,  -3,   3,   0,   0,  -1,  -3,  -4,  -1,   2,   0,   1,   1,  -4,   9,  -7,  -3,  -6,  -8,  -4,  -2,   4,   9,  -1,  -9,  -6,   1,  14,  -1 },
     {    3,  -1,   4,  -5,  -4,  -2,  -2,  -2,  -1,   2,   0,   0,   4,  -1,   2, -20,   1,  -1,  -2,   0,   0,   1,   1,  -1,   6,  -2,   3,  -5,   8,   0,  -1,  -1,   1,   0,   1,  -1,   4,   3,   9, -18,   0,  -2,   1,  -1,   1,   1,  -1,   0,   5,   2,  -8,  -2,   5,   0,   1,   1,   6,  -2,   6, -26,   0,  -1,  -1,  -1,   8,   7,   3,   2,  -2,  -2,   0,  -2,  22,  -8,   4, -38,   5,  -2,  -1,  -1,  87,  11,  47,   9, -22,  14,  -7,  35, -11,   2,   0,  -1,  -6,  -6,   3,   7 },
     {   -2,  -4,  -5,  -5, -18, -17,  -5,  -3,   0,  -3,   3,   3,   7,   4,  16, 107,  15, -10, -11,  -6,  -8,  -4,  -3,   1,   5,   0,  12,   2,   6,   2,   0,  -3,   3,   5,  -3,   0,  -2, -12,  -5, -24,   2,  11,   1,   2,   5,   1,  -2,  -1,   1,   1,  20,   1, -14,  -3,  -7,   0,  -1,   0,  -4, -10,  -6,   0,  -2,   3,   6,   2,   8,   6,   5,  -1,  -1,   0,   5,  -7,  -5, -11,   4,   5,   3,   3,   2,  -3,  11,   2,  -5,  -9,  -3,  -1,   1,  -2,  15,  -3,  -1, -16,  -8,  13 },
     {    3,   4,   3,   7,   5,   7,  -4,  -1,  -2,  -1,   1,   1,  -1,  -4,  -4, -11, -23,  -8,  -2,   8,   3,   0,  -1,   1,   7,  11,   4,  59,   4,   4,  -4,  -7,  -4,   0,  -1,  -1,  -1,  -7, -10, -22,  -6,   0,   7,   6,   2,   0,  -1,   1,   9,  26,   4,  42,  -8, -11, -10,  -7,  16, -19, -49,  -7,  22,   9,  15,   0,  -6, -19,  18,   7,  -3,  -9,  -6,   1,  17,   8,  -8,   5,  20,   4,   2,  -2,   2, -33,   1, -12,   4,   5,   4,   4,  -8, -17,   1,   2,  12,  -1, -24,   3 },
     {    1,  -2,  -3,   6,  -2,   0,  -6,   2,   1,  -3,   0,   1,   0,  -5,  -3,  -5,   9,   4,   6,  -2,  -2,   0,   1,   2,   1,  -2, -11,  -7,  13,   7,   1,  -5,  -2,   0,  -1,   1,   1,  -5, -11,   0,  12,   2,   4,   1,  -1,   0,  -2,   0,   9, -18, -27, -11,   9,   3,   5,   0, -17, -56,  19, -55, -10,   7,   5,   7, -25,  25,  49,  18, -36,  -7,  -6,   2,  10,  16,  -7,  22,  -3,  -3,   1,  -1, -14,  -4,   2,   5,   4,  -5,  -2,  -5,   7,  -2,  14,  -3,  -3, -10,  -1,  12 },
     {    2,   4,   5,  26,  17,  -2,   3,   2,  -4,  -1,  -1,  -3,  -3,  -1,  -1,  10, -38, -12,  -8,   1,   2,   6,  -2,   0,   2,  -2,  13,  76,  19,  -6, -14,  -5,  -6,  -2,   3,  -1,   4,  10,  42,  -6,   4,  -3, -10,   1,   1,   2,   0,  -4,  -5,  -4, -15, -39,  -5,  11,  -1,   2,  -8,   3,  25,   4,  -7,   2,   3,   3,   0,  21,  -9,   0,  -5,  -5,   1,  -1,   2,  -6,   5,  -8,  -4,   4,   3,  -1, -12,  20,  -6,  10, -16,  -3, -11,   0,   4,   6,  -3, -15,  -4,   0,   9,  -3 },
     {    3,   6,   0,   8,  83,  -2, -20,  -5,   0, -22,   9,   3,   0,   0,  -2,  -2,  13,  -2, -16,   0, -11,  -5,   5,  10,  -2,  -4,   0, -14, -17,  -4,   4, -18,   0,  20,  -9,   2,  -2,   3,  -4, -13,   8,  -3, -13,  -7,  12,   1, -16,  -9,  -1,  -6, -11,  10, -35,  -1,  -7,  -7,  -3,  -7,  -2, -15,   6, -23,   6,   1,   2,  -1, -12,  17,  13,  12,  -1,   9,   0, -14,   7,   1,  -6,   5,  15,  14,  -9,  -4,   6,   3,  -6,  -1,  16,  11, -10,   1,   9,  -9, -12,   3,   5, -15 },
     {   -2,   0,   0,   0, -11,  20,   4,   3,  -4,  -5,  -1,  -4,  -4,   0,  -4,  -8,  16,  25,  -7,   2,   3,  -1,   3,   3,  -3,   0,  -6,  14, -20,  -1,  -4,   1,   2,   4,   1,  -2,  -4,  -5, -13,   9,  -6,  -3,   7,   4,   1,   1,   0,   0,  -3,  -3,  -5,  14,  -1,   7,  -8,  -2,   2, -11, -11,   6,   1,   1,   1,   5,  -9, -24,   5,   8,   9,   3,   6,   0, -39,  -5,   6,   0, -18,   3,  -2,   0,  -7,  26,   4,  49, -66, -16, -29,   1,   6, -11,  19, -12,  -1, -16, -23,   9 },
   },
   {
     {   92, -46,  -2,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0, -65,  38,   0,   2,   1,   1,   0,   0,   0,   0,   0,   0,   2,  -9,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   9,  -3,  -3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   44, -31,   7,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,  74,  -8, -19,   0,  -4,   0,  -2,   0,  -1,   0,  -1,   0, -74,  39,   5,   0,   2,   0,   1,   0,   0,   0,   0,   0,   4, -17,   6,   1,   1,   0,   0,   0,   0,   0,   0,   0,   2,   4,  -5,   1,  -1,   0,   0,   0,  -7,   4,   1,  -1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   43,  65, -60,   9, -11,   1,  -4,   0,  -2,   0,  -1,   0, -10, -49,  47, -11,   7,  -2,   2,  -1,   1,   0,   1,   0, -29,  -4,   0,   7,   1,   1,   1,   1,   0,   0,   0,   0,  18,  10, -13,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,  -2,   3,   5,  -3,   0,   0,   0,   0,  -3,   0,   1,   1,   0,   0,   0,   0,  -2,  -2,   1,   0,   0,   0,   0,   0,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   43,  -8,  -5,  -4,   0,  -2,   0,  -1,   0,   0,   0,   0,  32, -29,   7,   1,  -1,   1,  -1,   0,   0,   0,   0,   0,  54,  12, -24,   1,  -4,   0,  -2,   0,  -1,   0,  -1,   0, -80,  36,  14,  -3,   3,  -1,   1,  -1,   1,   0,   0,   0,   8, -26,   5,   4,   0,   1,   0,   1,   1,   8,  -8,   0,  -1,   0,   0,   0,  -3,   3,   2,  -1,   0,   0,   0,   0,  -4,   2,   1,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0 },
     {   13,  12, -11,   3,  -3,   1,  -1,   0,  -1,   0,   0,   0,  42,  35, -23,  -4,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   0,  -5, -79,  46,  -1,   6,   1,   2,   1,   1,   1,   0,   0, -17,  38, -36,  11,  -3,   1,   0,   0,   0,   0,   0,   0,   4,  15,   8, -10,   2,  -2,   0,  -1,  -3, -12,   8,   2,  -1,   1,   0,   0,  -4,   2,  -5,   2,   0,   0,   0,   0,  -3,   0,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0 },
     {   -9,  23, -15,   7,  -4,   1,  -1,   0,  -1,   0,   0,   0, -25,  29,  -2,  -2,   0,   0,   0,   0,   0,   0,   0,   0, -32,   1,  22,  -7,   4,  -2,   2,  -1,   1,   0,   0,   0, -52, -32,  17,   9,   3,   2,   2,   1,   1,   0,   1,   1,  78, -21, -26,   2,  -2,   0,  -1,   1, -10,  26,   2,  -7,   1,  -2,   0,  -1,   1, -11,   9,   2,  -1,   0,   0,   0,   2,  -2,  -4,   2,   0,   0,   0,   0,   4,  -1,  -1,  -1,   1,   1,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
     {  -17, -36, -54,  59, -16,  16,  -4,   7,  -1,   3,   0,   2,  18,  37,  28, -48,  14, -12,   3,  -5,   1,  -2,   0,  -1,  17,  22,   8,  -2,  -7,  -1,  -1,  -1,  -1,  -1,  -1,   0, -11, -21,  -4,  15,   0,   3,   0,   1,   0,   1,   0,   1, -25,  -1,   1,  -6,   5,   0,   2,   0,  15,  -5,  -2,   0,  -2,   0,   0,   0,   0,   6,   2,  -2,   0,  -1,   0,   0,   0,  -4,   1,   2,   0,   1,   0,   0,  -1,   0,  -1,  -1,   2,  -1,   0,   0,  -1,   1,   1,   0,   1,  -1,   0,   0 },
     {   -9,  18,   2,  -6,   0,  -1,   0,  -1,   0,   0,   0,   0, -21,  24,  -8,   7,  -4,   1,  -1,   1,   0,   0,   0,   0, -48,   9,   7,   1,   4,   0,   1,   0,   1,   0,   0,   0, -24,  26,  10,  -9,   2,  -3,   1,  -2,   0,  -1,   0,   0, -37, -57,  29,   7,   2,   3,   1,   2,  57, -22, -30,   4,  -2,   1,  -1,   1,  -1,  21,   0,  -7,   0,  -2,   0,  -1,   5,  -7,   6,   1,  -1,   0,   0,   0,   1,  -2,  -1,   2,   5,  -2,  -2,   0,   1,   1,  -1,  -1,   3,  -1,   0,   0 },
     {   22,  51, -14, -11,  -5,  -1,  -3,  -1,  -1,  -1,  -1,   0,   8,  14, -38,  24,  -7,   4,  -1,   1,  -1,   1,   0,   0,  31,  20,  24, -30,   9,  -8,   2,  -4,   1,  -2,   0,  -1, -13, -45,  26,  13,  -6,   4,  -1,   2,  -1,   1,   0,   0, -44,  25, -27,   7,   5,  -1,   2,   0,  21, -14,  15, -12,   2,  -1,   0,   0,   0,   3,  -2,   4,  -2,   1,   0,   0,  -5,   2,   3,  -1,   1,  -1,   0,   0,  -2,  -3,   0,   1,   1,  -1,   1,  -1,  -1,  -1,   0,   0,   1,  -1,   2,  -1 },
     {  -10, -32, -20,  -2,  16,  -4,   6,   0,   2,   1,   1,   1, -10, -39,   1,  29, -15,  12,  -4,   3,  -1,   1,   0,   1,  12,  22,  59, -47,  11, -12,   2,  -4,  -1,  -2,   0,  -1,  12,  10, -44,  30, -10,   3,  -1,   1,   0,   0,   0,   0,   1, -32,   0,   0,   4,   3,   0,   2,   5,  26,  -6,  -7,   1,  -3,   0,  -1,  -2,   4,   5,   2,  -3,   0,   0,   0,   2,  -1,  -1,   0,   1,   0,   0,   0,  -1,   0,  -1,   1,   0,   2,   0,  -1,  -1,   0,   0,   0,   0,   1,   0,  -1 },
     {  -11,   0,  -2,   3,   0,   2,   1,   1,   0,   1,   0,   0, -24,   2,  -6,   5,   1,   1,   1,   0,   0,   1,   0,   0, -17,  43,   7, -13,  -2,  -2,  -1,  -1,   0,   0,   0,   0, -33,  27, -15,  10,  -1,   0,   0,   0,   0,   0,   0,   0,  -3,  19,  22, -18,   4,  -4,   1,  -2, -42, -58,  35,  10,   2,   4,   1,   2,  39, -21, -31,   6,   0,   0,   0,   1,   2,  15,   3,  -8,   0,  -1,   0,   0,   6,  -7,   3,   2,   1,  -2,  -1,   1,   4,  -2,  -2,   1,   1,   1,   0,  -1 },
     {   -4, -23,  21, -13,   9,  -3,   3,  -2,   1,   0,   1,   0, -11, -38,  32,  -2,   3,  -1,   1,   0,   1,   0,   0,   0, -15, -18,   5,  11,  -3,   3,  -2,   1,  -1,   1,   0,   0, -39, -35, -19,  18,   2,   4,   1,   2,   1,   1,   1,   1,  11,  41, -13, -19,   4,  -4,   1,  -1,  60, -28,  10,   5, -10,   3,  -2,   1, -16,  23, -18,   5,   2,  -2,   1,   0,  -3,   0,   5,  -5,   2,   0,   0,   0,   2,   2,  -3,   2,   2,   3,   0,  -1,   0,   3,  -2,   1,   0,   1,   0,   0 },
     {    8,   3,  45,  36, -54,  23, -17,   4,  -6,   2,  -3,   1, -15, -34, -31, -18,  50, -20,  16,  -4,   5,  -1,   3,   0,  -8,  -7,   3, -28,   4,   3,   1,   1,   1,   1,   1,   0,   3,  19,   9,  24, -26,   6,  -6,   0,  -2,   0,  -1,   0,   2,  -6, -11,   4,  15,  -7,   3,   0,   4,   3,   1,  -9,  -2,   4,  -1,   1,  -5,   0,   3,   0,  -2,  -1,   1,   0,   4,  -3,   0,   3,   0,   0,   0,   0,  -1,   1,  -3,  -1,   1,   0,   1,   0,   0,   1,  -1,   0,   1,   1,   0,   0 },
     {  -13,  -9,  -4,  -5,   6,  -2,   2,   1,   0,   1,   0,   1, -11,   6,   9,   4,  -6,   4,  -2,   0,   0,  -1,   0,   0, -21,   8,  -9,   2,   7,  -2,   3,   0,   0,   0,   1,  -1, -21,  37,   2,  -4,  -7,   4,  -3,   1,  -1,   1,  -1,   0, -30,  29, -15,   4,   2,  -3,   3,  -2,  14,  22,  33, -23,   5,  -5,   0,  -2, -41, -52,  35,  19,  -4,   5,   0,   2,  28, -24, -32,   5,   4,  -2,   0,   1,   5,  12,  -1, -10,   5,  -3,   5,   1,   0,  -2,   2,   2,   3,  -3,  -3,   1 },
     {  -13, -32, -45,  16,   2,   6,   1,   4,   1,   2,   0,   2,  -6, -20, -16,  21,  -6,   3,   1,   2,   0,   0,   0,   0,  -6, -12,   5,  -8,   9,  -2,   2,   0,   0,   0,   0,   0,  11,  33,  45, -35,  -4,  -5,  -2,  -2,  -1,  -1,  -1,  -1,  23,  16, -40,  33,  -8,  -1,  -1,  -1,   5, -42,   5,  -6,   9,   2,   0,   2, -10,  31,  -2, -12,  -1,  -1,  -1,   0,   4,   4,   6,   2,  -2,  -1,   0,  -1,   4,   0,   1,   1,  -4,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,  -1 },
     {    3,  19,  -8,   1,  -6,   3,  -1,   0,   0,   0,  -1,   0,  -3,  17, -29,   5,   2,  -3,   0,   0,   0,   0,   0,   0,  14,  45,  -9,   0,  -9,   1,  -4,  -1,  -2,   0,  -1,   0,  13,  18, -26,  -3,   4,  -3,   3,   0,   0,   0,   0,  -1,  37,  23,  22, -28,   0,  -3,  -2,  -2,   6,  -4,   5,  16,  -8,   3,  -2,   0, -63,  33,  10,  -9,   9,  -4,   3,  -1,   6, -35,  23,   0,  -2,   3,   0,   1,   4,  -8, -10,   5,  -2,  -1,   4,  -2,  -3,  -3,   1,   0,   1,  -5,   2,   0 },
     {    2,   2,   5,  21,  -1,  -3,   0,  -2,  -1,   0,  -1,   0,   9,  18,  25,  14, -29,  12, -10,   2,  -3,   1,  -2,   0,  -8, -14, -57, -53,  55, -18,  16,  -1,   5,   0,   3,   0,   1,  -3,   3,  22, -34,  13,  -7,   0,  -1,   0,   0,   1,  15,  15,  26,   0,  -3,  -4,  -3,  -2,   4,  -6, -14,  -1,  11,  -4,   2,   0,  -3,   2,  -1,   0,  -3,   4,  -1,   1,  -7,   1,   8,  -5,   1,   0,   1,   0,   3,  -2,   1,   4,  -2,  -1,   0,  -1,   1,  -1,   0,   1,  -1,   0,   0,   0 },
     {    4,   5,   7,  57,  24, -45,  13, -12,   2,  -3,   0,  -2, -18, -28, -41, -33, -15,  43, -12,  12,  -1,   3,   0,   2,  -6,  -8,   4,   3,  -9,  -2,   3,   0,   0,   2,   0,   1,  -9,  -1,  13,  14,  19, -15,   3,  -2,   1,  -2,   0,   0,  -8,  22,  11,  -8, -12,  11,  -6,   2,   1,   4, -21,  10,   3,  -5,   5,  -2,   5, -12,  15, -15,   1,   2,  -1,   1,  -8,  -4,   2,   8,  -1,   0,   0,   0,   7,   2,  -8,  -3,   2,   4,   2,  -1,   2,   1,   1,  -1,  -1,   0,   0,   2 },
     {  -11,  -6, -12,  -8,  -5,  12,   0,   3,   2,  -1,   0,   0, -11,   0,  -3,  18,   5,  -8,   2,  -1,  -1,  -1,   0,   0, -18,   0, -13,   8,  -5,   5,  -2,   3,  -1,   0,  -1,   0, -20,  15, -15,  -3,  -6,   4,   0,   1,   2,   1,   0,   1, -20,  32, -10,   2,   5,  -4,   2,  -3,  -6,  38,  -3, -14,   5,  -2,   0,  -1,  41,  23,  28, -13,  -4,  -2,  -3,  -2, -60, -25,  29,  15,   2,   1,   1,   0,  12, -35, -16,   5,   4,   8,  -6,  -7,   8,   0,   4,  -1,  -3,  -1,   1,   1 },
     {   -1,  -6,   8, -12,  -8,  14,  -4,   4,  -2,   0,  -1,   0,   0, -11,  20,  12,   7, -10,   5,  -2,   0,   0,   1,   0,   2, -11,  26,  -7,   7,  -6,  -1,  -1,   1,   0,  -1,   0, -19, -31,   8, -11,   4,   5,   1,   2,   2,   2,   1,   0, -13,  11,  30,  16,  -9,  -1,   0,  -3, -41, -26, -33,  15,  13,   2,   3,   2,  -6,  -3,  24, -24,   1,  -2,   0,  -1,  25, -44,  -3,  27,  -9,   4,  -3,   1,  15,  23, -37,  -1,  -8,  10,  10,  -8,   4,  -2,  -4,   4,  -1,   1,  -3,   0 },
     {   -8, -18, -41,   6, -22,  22,  -7,   8,  -1,   3,   0,   2,  -6, -18, -45,  45,  12, -11,   5,  -1,   0,   0,   0,   0,  -3, -15, -18,   8,   3,   4,  -2,   1,   0,   0,   0,   0, -11, -25,  -4, -19,   7,  -3,   4,   0,   2,  -1,   1,   0,   2,  15,  45, -16, -11,  -1,  -4,  -2,  21,  25, -14,  17,  -4,  -1,   1,  -1,   4, -34, -12,   4,   7,   0,   2,   2,   5,  22, -10, -12,  -2,   0,  -2,  -1,  -6,  13,   9,  -3,   3,  -1,   2,   0,  -2,   3,   1,   0,  -2,   2,   0,  -2 },
     {   -6, -10, -28, -25,  20, -10,  11,  -4,   4,  -2,   2,  -1,   2,   0,  -4,  11,   9,  -7,   0,  -4,   2,  -1,   1,   0,  -7, -18, -16, -12, -28,  25,  -6,   9,  -3,   3,  -1,   2,   9,   5,  47,  52, -16, -13,  -1,  -5,   0,  -2,  -1,  -1,  -2,  -7,  -1, -40,  46,  -7,   4,   0, -14,  -7, -18,  -2, -28,  15,  -2,   3,  -3,  -5,   7,   7,   1,  -6,   0,   1,   6, -10,   5,   0,   0,   1,   0,   1,   8,   7,  -9,   3,  -3,   4,   0,  -5,   1,   0,   1,   4,  -1,   0,  -2,   0 },
     {    5,  10,   8,  19,  32,  -1, -13,   2,  -4,  -3,  -1,  -1,   4,   9,   9,  14,   3, -23,  17, -11,   3,  -2,   0,  -1,  -6,  -3, -17, -55, -52,  34,  -8,  15,   1,   4,   1,   2,  -1,   0, -19,   1,  34, -23,   3,  -3,   1,   0,   0,   0,   7,  13,  12,  47,  -5,   3,  -4,  -5,  10,   2,   1, -23,   4,   8,   0,   1,   1,  -3, -11,  -5,  -1,  -7,   4,   0,   5,   1,  -4,   5,  -1,   1,  -2,   1,  -5,   3,   3,   1,   2,  -2,   2,  -2,  -2,  -1,   1,   1,  -1,  -2,  -2,   2 },
     {    4,   5,  -7,  20,   3,  10, -14,   4,  -5,   2,  -2,   1,   0,  -1, -24,  18, -21,  10,   2,   2,   0,   0,   1,   1,   1,   0, -35,  31,  -6,  -3,   5,  -1,   1,  -2,   1,   0,   2,  -5, -35,  20,   0,   5,  -2,   2,  -1,   2,  -2,   1,   5, -18, -21,  -5,   6,  -3,   6,  -1,   2, -15,   6, -25,   5,  -2,  -2,   2,  36,  31,  -1,  15,  -7,  -3,   0,   0,  18, -27, -26,  12,   5,   2,   0,   1,  -7,  43, -16, -26, -13,   7,  27,  -5,   5,  -5,  -3,   4,  -4,   4,   1,  -4 },
     {    5,   1,  -3, -21,  -2,   0,   5,   9,   0,  -2,  -1,   0,   7,   4,  -4, -20,   5,   6,  10,  -2,   3,  -6,   2,  -1,  15,  17,  14,   0,   0,   2,  -6,   0,  -1,   0,  -4,  -1,  21,   7,   3,  -4, -12,  -2,   2,   3,   3,   2,  -2,  -1,  37,   7,  27,  10,   7,  -9,   5,  -9,  38, -31, -16, -13,   8,   6,  -4,   5,  11, -32,  21,  14, -23,   5,  -2,   0, -52,  10, -29,  26,  13,  -7,   0,  -4,  12,   5, -15, -20,  -5,  16,   7,   0,   4,  -3,   0,   2,  -8,   5,  -1,  -2 },
     {   -4,   5,  -4, -20,   5,  -3,   4,   0,  -3,   2,   0,   0,  -6,   9,  -6,  -8,   0, -12,   6,  -4,   5,  -1,   1,   0,  -3,  19,   0,  16,  -8,   2,  -5,   0,  -1,  -2,   0,  -1, -10,  15, -19,  25, -23,  18, -10,   2,  -3,   2,  -1,   1,   2,  36, -18,  14,  -4,  -9,   4,  -3,  -3,   1, -60,   1,  14,  -3,   4,   2,   2,   2,  18, -36,  -7,   6,  -4,   0,  25,  43,   5,  -1,  -2,  -8,   0,  -3, -28,  16,  24,  -7,   1, -23,  10,   5,   1,  -1,  -2,  -2,   1,  -5,   5,  -1 },
     {    2,   3,  18,  22, -53,  -6,  15,  -6,   2,   0,   1,   0,   9,  16,  33,  49,  -4,  10, -35,  13,  -7,   2,  -3,   1,   2,   8,   4,  -2, -10,  20,  -9,  -4,   1,   0,   0,   0,   5,  10,  -9,   4,  12, -34,  20, -11,   3,  -3,   2,  -1,   2,   1, -32, -17,  -6,   9,   0,   6,   7, -15, -13,  15,  -9,   6,  -4,  -1,   4, -15,  20, -14,  14,  -3,   4,  -1, -14,  15, -10,   9,  -6,  -2,   0,   1,   0,  12,  -5,  -3,   0,  -8,  -2,  -2,  -3,   4,   1,  -4,  -2,   1,   3,  -3 },
     {   -4,  -6, -18, -26, -19, -36,  38, -21,  16,  -2,   6,  -2,   6,   5,  -6,  -3,  65,  16, -32,  12, -10,  -1,  -2,   1,   1,  -2, -13,  -9, -12,  -6,  27,  -4,   4,   1,   0,   1,  -2,  -9, -16,  -3, -13,  -5,  -1,   2,   0,   3,  -1,  -1,   8,   2,  15,  25,   0,  16, -17,   6,   4,  -6,  11, -19,   3, -15,  11,  -3,  15,  13,   1,   3,  -3,   4,  -7,   2,  17, -17,  -3,   1,  -5,   1,   3,   2, -11,   2,   7,   1,   2,  -4,   0,   8,   1,   4,  -8,  -6,   0,   1,   4,   2 },
     {   -5,   0,  -1,  13,  -4, -10,   3,  -2,  -3,   4,  -3,  -1,  -6,   1,  -1,  18,   9,   2, -13,   3,  -1,  -2,   3,   0,  -5,  -1,   0,  24, -10,  14,  -3,  -3,   2,  -2,   1,   0, -10, -12, -10,  17, -12,   8,   0,   0,  -1,   0,   1,   1, -16,  -8,   6,  30,  -1,  -7,   3,   2, -30,  -3, -15, -15,  11,   5,   3,   0, -49,  12, -20,  11,  -5,   5,   2,   1, -34,  13, -12, -32,  13, -11,   6,  -3,  59,   1,  16, -16, -18,   6,   8,  16,   5,  -7,  -1,  -4,   1,   0,   1,  -2 },
     {    3,  -1,   4,  26,  -4,  -7,   6,  -2,  -1,  -1,   0,   0,   1,  -7,  -5,  20, -16,   2,  -8,   5,   1,  -2,   0,   0,   3,  -5,  -1,  24,  -2,   1,  -6,   7,  -4,   3,  -3,   2,  -5, -21, -15, -12,  -9,   7,   7,   0,   3,  -1,   1,   2,   4,  -7,  15,   8,  33, -23,  -1,  -3, -13, -21,  -5, -34,  -6,  25,  -1,   7, -12,  -2,  22,   6, -25,  -5,   1,   1,  26,   9,  -5,  14,  10,  -4,  -3,  -4, -27, -54,   1,  10,  24,   6, -35,  -1,  -6,  17,   3,  -9,   2,   2,  -3,  -3 },
     {    3,   2,  -6,   1,   8,  -4,   6,  -8,  -2,   0,   4,  -2,   6,   5,  -9,   5,   7, -13,   0,  -2,   7,  -3,   0,   0,   5,   3, -15,  17,  -5,   2,   3,   4,   3,  -1,   2,  -1,   8,   3, -18,  17, -13,  18, -10,   3,  -4,   0,  -1,   0,   3, -16, -26,  13, -18,  11, -12,   6,  18, -11,  -2,   4,  14,  -4,   4,  -4,  -1, -41, -13, -14,   9,   0,   3,  -1,  18,  12,  20, -11, -15,   4,  -2,  -1,  24, -19, -47,  30,  -3,  46,  -5, -35,   7,   4,  20,   1,  -5,   7,   1,  -1 },
     {   -6,  -8,  -7, -28, -29,  15, -13,  20,  -6,   4,  -2,  -1,  -4,  -4,  -3, -15, -16,  16,   5,   7,   3,   0,  -1,   1,  -2,  -2,  -2,  10,   0,  -2,   2, -12,   6,   3,   5,   2,   1,   3,  11,  42,  20, -23,   2, -12,   3,  -6,   2,  -5,  -1,   4,  16,   0, -44,  25,  -5,   4,  -8,  -6,  -5, -54,  12,  -2,   3,   3, -13,   5,  -5,  22,  12, -17,   4,  -3,   4,  11,  21,  19,  -4,   4,  -3,  -3,  -1,  -2,  -2,   0,   3, -10, -24, -10,  -3,  16,   3,  -8,   1,   5,   8,  -1 },
   },
   {
     {  105, -46,  -4,  -5,  -1,  -2,   0,  -1,   0,  -1,   0,   0,  33,  -2,  -8,   0,  -2,   0,  -1,   0,  -1,   0,   0,   0, -22,  22,  -4,   1,  -1,   0,  -1,   0,   0,   0,   0,   0, -23,  13,   3,   0,   1,   0,   0,   0,   0,   0,   0,   0, -13,   3,   4,   0,   1,   0,   0,   0,  -5,  -1,   2,   1,   0,   0,   0,   0,  -3,  -1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   41,   0,  -7,  -3,  -2,  -1,  -1,   0,  -1,   0,   0,   0, -99,  43,   7,   2,   2,   1,   1,   0,   0,   0,   0,   0, -27, -10,  13,  -1,   3,   0,   1,   0,   1,   0,   0,   0,  25, -26,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  18,  -9,  -5,   2,  -1,   0,  -1,   0,   7,   1,  -4,   0,   0,   0,   0,   0,   2,   2,  -1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {    9, -57,  28,  -1,   5,   0,   2,   0,   1,   0,   1,   0, -28, -43,  18,   6,   3,   3,   1,   1,   1,   1,   0,   1,  55, -30, -20,   7,  -3,   2,  -1,   1,  -1,   0,   0,   0,  41,  14, -30,   2,  -4,   0,  -1,   0,  -1,   0,   0,   0,  -3,  33, -13,  -4,   0,  -1,   0,  -1, -13,  18,   4,  -5,   1,  -1,   0,  -1,  -8,   4,   5,  -1,  -1,   0,   0,   0,  -3,   1,   1,   1,  -1,   0,   0,   0,  -3,   2,   0,   0,  -2,   2,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0 },
     {   40,  61, -44,   0,  -9,   0,  -3,  -1,  -2,   0,  -1,   0,  15,  28, -12,  -7,  -1,  -2,   0,  -1,   0,  -1,   0,   0,  63, -40,  13,  -7,   1,  -2,   0,   0,   0,   0,   0,   0,  14,  -1,  -2,   3,  -3,   1,  -1,   1,  -1,   0,   0,   0, -27,  20,  -1,   2,   0,   0,   0,   0, -19,   6,   7,  -1,   2,  -1,   0,   0,  -9,  -3,   6,   0,   1,   0,   0,   0,  -5,  -2,   2,   1,   0,   0,   0,   0,  -4,  -1,   1,   0,  -2,  -1,   1,   0,  -2,  -1,   1,   0,  -1,  -1,   1,   0 },
     {   -9,  -1,   6,  -1,   1,   0,   1,   0,   0,   0,   0,   0, -43,  29,  -7,   3,  -1,   2,   0,   0,   0,   0,   0,   0,  37,  45, -30,  -2,  -4,  -1,  -1,  -1,  -1,   0,  -1,   0, -28,  59,   1, -12,   2,  -4,   1,  -2,   0,  -1,   0,  -1, -32,   2,  35, -11,   4,  -3,   1,  -1,   1, -32,  24,   3,   0,   1,   0,   1,   7, -22,   2,   7,  -1,   1,   0,   1,   6,  -9,  -4,   3,   0,   0,   0,   0,   3,  -5,  -2,   1,   3,  -3,  -1,   0,   2,  -2,   0,   0,   1,  -1,   0,   0 },
     {   15,  10,  -1,   1,  -3,  -1,  -2,   0,  -1,   0,  -1,   0, -26, -44,  33,   0,   5,   0,   1,   0,   1,   0,   1,   0,  26, -41,  13,   3,   1,   1,   1,   0,   0,   0,   0,   0, -74,   5,   5,  10,   2,   3,   1,   1,   1,   1,   0,   0,  -8, -35,  -3,   8,   2,   2,   1,   1,  33, -15, -22,   4,  -1,   1,  -1,   1,  14,  13, -19,  -2,   0,  -1,   0,   0,   2,  15,  -4,  -5,   0,  -1,   0,  -1,   1,   5,   1,  -2,   1,   3,  -1,   0,   0,   2,  -1,   0,   0,   2,  -1,   0 },
     {    2, -25,  -5,   9,   1,   1,   1,   1,   0,   1,   0,   0,   5,  15, -26,   5,  -3,   2,   0,   1,   0,   0,   0,   0,  62,  34, -18, -14,  -3,  -4,   0,  -2,  -1,  -1,   0,  -1,  11, -26,  34,  -9,   1,  -2,   0,  -1,   0,   0,   0,   0,  33, -53,   7,  12,  -4,   3,  -2,   3,   2,  11, -29,  13,  -3,   2,  -1,   1, -18,  27, -14,  -3,   3,  -2,   2,  -1, -11,  11,   5,  -8,   2,  -1,   1,   0,  -6,   2,   6,  -3,  -4,   1,   3,   0,  -2,   0,   1,   0,  -2,   0,   1,   0 },
     {   12,  24,  76, -51,   3, -11,   1,  -5,   0,  -2,   0,  -2,  19,  36,  46, -24,  -9,  -5,  -3,  -3,  -2,  -1,  -1,  -1,  15,   9,  -7,  21, -10,   3,  -4,   1,  -2,   0,  -1,   0,   5, -20,  -6,  18,   1,   1,   0,   1,   0,   0,   0,   0,  -8, -18,  -4,   7,   5,   0,   1,   1, -13,  -7,  -7,   4,   3,   2,   0,   1,  -3,  -4,  -8,   3,   1,   1,   0,   1,  -1,  -2,  -4,   0,   1,   1,   0,   0,  -1,   0,  -2,   1,  -1,   0,  -1,   0,  -1,   0,  -1,   0,  -1,   0,  -1,   0 },
     {   22,  59,  -8, -11,  -5,  -3,  -2,  -2,  -1,  -1,  -1,  -1, -10, -36,   7,  10,   0,   1,   0,   1,   0,   0,   0,   0,  -6,  29, -50,  18,  -5,   3,   0,   1,   0,   0,   0,   0,  -7,  21, -25,  -3,   5,  -2,   2,  -1,   1,   0,   0,   0,  58,  -5,  -7, -16,   2,  -3,   0,  -1,   0,  31,   6, -11,  -1,  -1,  -1,  -1, -17,   8,  16,  -3,  -1,  -1,   0,  -1,  -8,  -2,  10,   0,   0,   0,   0,   0,  -2,  -8,   5,   1,  -1,  -3,   2,   1,   0,  -3,   2,   0,   0,  -1,   1,   0 },
     {  -16, -22,   8,   1,   4,  -1,   1,   0,   1,   0,   1,   0,  -5,  44, -20,   0,  -2,  -1,  -2,  -1,  -1,   0,   0,   0,  14,  -1,  -1,   4,   0,   0,  -2,   0,   0,   0,   0,   0, -63, -11, -11,  14,   4,   3,   2,   2,   1,   1,   1,   1,   0,  19, -45,  -1,   6,  -1,   3,  -1,   2,  48,  -6, -27,   7,  -5,   2,  -3,   0,   8,  35, -20,  -1,  -1,  -1,  -1,   6, -17,  23,   4,  -5,   1,  -1,   1,   4, -13,   3,   8,   3,  -4,  -2,   2,   2,  -2,   0,   0,   1,  -2,   0,   0 },
     {    1,   0,  -6,   7,  -5,   2,  -2,   1,  -1,   0,  -1,   0,  -7, -18,  50, -23,   3,  -3,   1,  -1,   0,  -1,   0,  -1,  14,  39,  58, -41,  -6,  -6,  -1,  -4,  -1,  -1,  -1,  -1,  -5,  21,  31,   5, -14,   0,  -3,   0,  -1,   0,  -1,   0,  29,  15, -14,  21,  -8,  -1,  -2,   0, -21,  17,   4,   5,   6,  -2,   1,  -1, -13, -23,  13,   3,   4,   2,   0,   1,   0, -20,  -5,   9,  -1,   3,   0,   1,   2,  -7, -10,   5,   0,   1,  -5,   0,   0,  -1,  -3,   0,   0,   0,  -2,   0 },
     {    0,  -9,   2,   0,  -3,   3,  -2,   1,  -1,   1,   0,   1,  20,  39,  13, -19,  -3,  -3,   0,  -1,  -1,  -1,   0,   0,   3,  -9,  23, -12,   3,  -4,   2,  -1,   0,  -1,   0,   0,  16,  51, -28,  -3,  -4,   0,  -2,  -1,  -1,   0,   0,  -1,  35,  13,  -3,  -8,  -2,   0,  -1,  -2,  70, -15,  -3,  -1,  -6,   1,  -2,   1,  -2,  32,  -4,   2,  -3,  -1,   0,   0, -27,  15,  12,  -1,   2,  -1,   0,   0, -15,  -4,  12,   2,  -5,  -8,   3,   3,  -4,  -3,   1,   0,  -2,  -3,   1,   1 },
     {    4,  19,  10,  50, -29,   3, -10,   0,  -4,   0,  -1,   0,  -3,  -2,  25,  32, -14,  -4,  -4,  -2,  -1,  -1,   0,   0,  -8,  39,   0,  -9,   6,  -7,   2,  -5,   1,  -1,   0,  -1,  22,   9, -24,   3,   2,   2,   1,  -1,   0,   0,   0,   0, -53,  -6, -31,  15,   4,   4,   4,   0,  -2,  -3, -36,  -6,  11,  -3,   3,   0,   5,  24,  -6, -26,   8,  -2,   1,  -1,   0,   7,  17, -15,  -1,   0,  -1,   0,   1,  -3,  10,  -2,   2,  -4,   3,   0,   1,  -2,   2,  -1,   0,  -3,   1,  -2 },
     {   -9,   0, -34, -55,  27,   1,  10,   1,   3,   0,   2,   0, -15, -21, -19, -51,  17,   5,   8,   4,   2,   1,   2,   0, -12,  21,  -6,  -3, -20,   7,  -3,   2,  -1,   1,   0,   0,  15,  16,   1,  27, -19,  -1,  -5,  -1,  -1,   0,  -1,  -1, -34,   3,  -2,  24,   1,  -7,   0,  -2,   3,   8, -24,   3,  12,  -5,   1,  -1,   8,  26,  -3, -18,   7,  -2,   0,  -1,   1,   8,  20, -12,  -2,   1,  -1,   0,   0,  -5,  13,   1,   1,  -3,   4,   3,   0,  -2,   2,   1,   1,  -2,   2,   0 },
     {    8,  19,  29, -20,   1,  -4,  -2,   0,  -1,  -1,   0,   0,  -7, -13, -22,  15,   3,   1,  -3,   2,  -1,   0,   1,   0,  11,  23,   4, -22,   9,  -2,   0,  -3,   1,  -1,   0,  -1, -10, -24,  29, -21,   2,   4,   1,   1,   0,   1,   0,   1,   9,  63, -17, -11, -13,   4,   0,  -1,   8,  -4,   0,   4,  -4,  -1,   2,  -1,  52,  -3, -27,  -1,  -1,  -2,   0,   0, -10,  36,   6, -10,   2,  -2,  -1,  -1, -12,   1,  20,  -2,  -3,  -8,   9,   4,   0,  -6,  -1,   3,  -3,   0,   3,   1 },
     {   -6, -11, -26,  10,   5,  -3,   1,   2,   1,   2,   1,   0,   2,  18,  -7,   6,  -3,  -3,  -3,   1,   0,   0,   1,   0,   6,   9,   3,  27, -13,   0,  -4,  -1,  -2,   1,  -1,   0, -13,  11, -31,  43,  -3,   0,  -1,  -2,  -1,   0,   0,   0,  37,  12, -32,  13,  12,  -6,   3,  -1, -31, -26,  -6,   8,  16,   1,   3,   2,   0, -30, -42,   1,  11,   5,   3,   1,   4,  24, -28, -25,   6,   1,   1,  -1,   1,  17,   9, -21,   0,   5,  12,  -7,   1,  -2,   6,  -1,   1,  -1,   3,  -1 },
     {    3,  -7, -25,   3,  -6,   8,   1,   3,   0,   1,  -1,   1,   5, -11,  11, -31,   5,   0,   3,   1,   1,   0,   0,   0,  19,  32, -11, -24,   6,  -3,   2,  -1,   0,  -1,  -1,   0,   9, -48, -39,  15,   4,   4,   1,   3,   1,   1,   1,   1,  -5, -23, -22, -12,   7,   2,   2,   1,  24, -25,  34, -28,  -3,   1,  -2,   1,  34, -17,  22,   5, -20,   5,  -5,   2,   2,  13,  -7,  22, -11,   0,  -1,  -1,  -7,  11,  -9,   6,  -4,   7,   0,  -1,  -2,   3,   3,   0,  -1,   2,   0,   2 },
     {    1,  -1, -18,  -7,  20,  -6,   4,  -2,   1,  -1,   1,   0,   3,  -7,   9,  41, -11,   1,  -5,  -1,  -3,   0,  -2,   0,  15,  20,  37,  63, -37,  -1, -13,   0,  -3,   0,  -2,   0,  20,   0,  36,  12,   9, -14,   1,  -4,   0,  -1,   0,  -1,  -6,   2,   0, -21,  28,  -6,   3,  -2,  26,   2,  -6, -27,   3,   7,   0,   0,  18,  -2,  10,  -3, -14,   6,  -3,   2,  -6,  -3,   0,  12,  -6,   0,   0,   1,  -7,  -3,  -8,   7,  -4,   1,  -6,   0,  -2,   2,  -1,  -2,  -2,   0,   0,  -1 },
     {   -9, -15, -40,   8,  27, -11,   4,  -1,   2,   1,   1,   0,   6,  19,  62, -17,  14, -14,  -3,  -5,  -3,   0,  -1,  -1,  -3,   5,  -8,   8,   6,   4,  -5,   1,   0,   0,   0,   0,  -9,  -8,  -3, -32,  19,   7,   2,   1,   0,   2,   0,   0,  -6,   6,  14, -44,  -2,  12,   0,   3, -35,  18, -10,  -1, -21,  11,   0,   1,  10,  17, -16,   4,  -9,  -3,   1,  -1,   4,  16,  11, -12,   1,  -3,   1,  -1,   0,  -2,  23,  -4,   3,  -6,   8,   6,   3,  -6,   2,   4,   0,  -1,  -1,   1 },
     {   -9, -15, -18, -20, -68,  37,   1,  11,   1,   3,   0,   2,  -4,  -1,  12, -23, -63,  17,   7,   4,   2,   1,  -1,   1,  -3,  -4,  -9,  17,   0, -23,  12,  -3,   4,  -1,   0,  -1,  -4,  -2,  11,   4,  25, -23,  -1,   0,   1,  -1,   1,  -2,   1,  11,  14,   4,  14,  -1, -10,  -2,  -6,  19,   2,  11,   2,   3,  -4,  -4,  14,   3,  -3,   7,   1,  -2,   0,   0,   3,   9,   0,   1,   3,  -3,   1,   0,  -5,  -2,  10,   2,  -1,  -7,   2,   2,  -1,  -2,   1,   1,  -2,  -1,  -1,   1 },
     {    2,   0,  -7,   1,   2,  -5,   3,  -1,   0,   0,  -1,  -1,  10,  16,  15,  -5,   2,   2,  -1,   1,  -1,  -1,  -1,  -1,  -6, -18, -39,  -2,   4,  11,   2,   2,   2,   1,   0,   1,  17,  30,  28, -28, -11,   1,  -1,  -2,  -3,   1,  -1,   0,  -2, -13, -24,  20,  -7,   4,   0,   2,  23,  32, -34, -11,   7,  -6,   2,  -2,  34, -46,  14,  -2,   2,   3,  -3,   2,  24,   4, -31,   6,  -7,   2,  -2,   2, -18,  29,   0,  -7,  -5,   1,  10,  -5,  -3,  -2,   8,  -1,  -1,   0,   0,   1 },
     {    3,   1,  13, -27,  -2,   4,   2,  -2,   1,   0,   1,  -1,   0,  -9, -25,  -3,  -6,  11,   0,   2,   0,   1,   1,   0,   1, -13,  30, -19,  10,  -8,  -2,   0,  -2,   1,   0,   0,  10,  31,  -3, -27,  28, -14,   1,  -2,   0,  -1,   0,  -1,  15, -31, -26, -24,  17,   6,   0,   3, -33, -29, -15, -26,  -5,  17,  -2,   3,  -5, -10,   7, -23, -19,  10,   0,  -1,  36,   9,  22,  -6, -19,  -4,   0,  -1,  15,  12,  15,   8,  -1,   0,   8,  10,  -4,  -1,  -1,   6,   0,   0,  -2,   3 },
     {    3,  -4,   6,  17, -11,   1, -12,   5,  -4,  -2,   2,   1,   7,   6,  10,  -2,  28, -17,  -6,   1,  -1,  -2,   1,   1,   0, -14, -17,  13,  35, -21,   0,  -2,   0,  -1,   2,   0,   9,  10,  28,  47, -12,   4, -11,  -3,  -2,  -1,   0,  -1,  21,  -4,  38,  17,  -7,   5,  -1,  -3, -15,  -4,  17, -31,  13,   2,   8,   1,  29,  -2,  -1, -39,  -9,   8,   3,   2,   7,   9,  22, -10, -17,   4,   1,   1,  -4, -14,  13,  14,  -6,  -6, -11,  14,  -4,   0,  -9,   1,  -3,   1,  -3,  -1 },
     {    5,  -2,  -8,  14,   2,   2,  -1,   1,   1,  -1,   1,   0,   9,   5,   8,  -5, -18,  15,  -1,   1,   1,  -3,   0,   1,   6, -10, -16,  -2, -30,  28,   5,   1,  -1,  -3,   0,   0,   0,  -7, -10, -17, -17,   5,  10,   0,  -1,  -2,  -1,   0,  30,  13,   5,  -2,  -1, -15,   5,   1, -13, -36, -28,  21,  18, -10,  -1,   3,  46,  10,  -5, -10,   9,  -8,  -5,   1,  21, -36,  30,  16,  -3,   5,  -6,   2,  17, -17, -17,  27, -13,  15, -15,   7,  -7,   3,  -3,  -4,  -6,   3,   0,   1 },
     {   -5,  -8, -19, -13,   6,  -4,  10,   1,  -3,   3,   2,   1,   0,  -2,  17,   5,  -4,   1,  -2,   5,  -1,   0,   0,   0,  -6, -13, -33,  -2,  16,  -3,  -1,   3,   3,   1,   2,   1,   1,  -7,  25,  -8,  20,  -4,   0,  -1,   0,   0,   1,  -1,  -4,   5, -37,  15,   9,   5,   2,   1,   5, -42,  -9,   4,   5,   6,   2,   2, -28, -39,   8,   2,   4,   5,   0,   2, -56,  -5,  29,  -6,  -2,   1,   0,   0,  -9, -34,   8,  10,  12,  -3, -19,   5,   9,   2, -10,  -1,   3,   2,  -4,  -3 },
     {    2,   2, -12, -21,   4,  20,   2,  -4,   4,  -3,  -1,   0,   4,   0,   0,   5,  28, -11,  11,  -9,   0,  -2,  -2,   0,  10,  18,  -2,  20,  51, -35,   0,  -2,  -1,  -1,   0,  -3,   5,   0,  -9,   5,  32,  -1, -15,   7,  -2,   1,   0,   1,   3,  21,  -8,   0, -10,  22, -14,   3,  21,  -3, -31,  18, -13,   7,   1,  -3,   1,   2, -16,  -1,   3,  -5,   6,  -3,  21, -39,  -8,   8,   2,  -1,  -1,   3,  24,   2, -30,   2,  -6,  23,  -6,  -7,  -4,  10,   7,  -4,  -4,   2,   3,   0 },
     {   -5,  -8, -22, -44,  16,  -2,  -9,   8,   1,   4,   2,   1,   5,   7,  13,  55, -16,   3, -21,   4,  -6,  -1,  -1,   0,   2,  -1,   8,  -1,   8, -12,   6,  -6,   0,  -2,   1,   1,   0,  10, -21,  -6, -13,   0,   7,   1,  -2,   3,  -1,   1,   2, -23, -13,  34, -39,  14,  -5,   6, -15,  -2,  28,  15, -12,  -3,  -1,  -1,  23,  15,  23,   8,   9, -15,   4,  -5,  14,  17,  -2,   3,  12,  -4,   3,  -4,  -2, -12,   2,   7,   0, -12, -13,   2,  -3,   2, -14,  -2,  -4,   3,  -4,  -7 },
     {    2,  -1,  -8,  -9,  13,   1,   1,   2,   1,  -2,   2,   2,   7,  -3,  12,  19, -18,   4,   6,  -6,   2,   1,  -4,  -2,   6,   4,  -5, -20,   1,   8,   8,  -1,  -7,   0,   0,  -2,   4, -21,  -6,  12,  -8,   5,   2,   1,   5,  -4,   2,   1,  -2,  10,  14, -10,   9,  -3, -11,   3,  41,   5,   3,  -6,   9, -11, -10,  -5, -35, -29, -20,  -4,  13,  -1,  -6,   5,  53, -15,   3, -25,  -1,   0,  -2,  -2,   7,  -9,  29,   6,  -6, -33,   4,  25,   0,   8, -21,  21, -15,  14, -17,  -3 },
     {    3,   4,   7,   6,  24,  67, -40,   3, -10,   4,  -4,  -1,   4,   6,  10,  -3,  35,  51, -30,  -7,  -6,   4,  -1,  -1,  -1,  -1,  -5,   5,  -1, -11,  16, -13,   5,   1,   4,   1,  -2,  -4,   6,   3, -16, -37,  28,  -4,   6,  -1,   2,  -1,  -4,   1,  -7, -13,  -4, -25,   6,   5,  -1,   0,  -2,   2,  -9,  -3,  -4,   5,  -7,  -5,  -1,  -1,   0,   1,   1,   0,  -5,  -1,   3,  -3,   3,  -1,  -2,   0,  -1,   5,   3,  -9,   5,  -2,   3,  -2,   2,   0,   0,   2,   1,   0,  -1,   1 },
     {   -6,  -5, -14, -27, -11,   6,   5,  -2,   1,   1,  -1,   2,   3,   9,  26,  32,  -4,  -8,  -8,  -3,  -2,  -2,  -5,  -3,  -4,  -4, -24, -27, -11,   0,   2,   5,   7,   4,   0,   0, -10, -12,  19,   1, -19,  -4,  -7,   6,   1,   2,   2,  -3,   5,  31,  11,   7,   5,  -5,   0,  -4,   1, -31,  -3, -21,  10,   6,   5,   2, -32,  33,  16, -27, -15,   4,  -1,   3,   0,   6, -14,  12,  -9,  -3,  -2,   3,  24,  40, -11,  -7,   0,  14,  27,  -6,   4, -11,  18,   1,   0,  -4,   5,   7 },
     {   -2,  -1,  -4, -14,  -3,   6,  -2,   2,   1,   4,  -2,  -4,   1,   3,   1,  10, -12,   5, -17,  18,   1,   3,   6,   0,  -5,  -4, -10, -26,   5,   1,  -8,  -7,   7,  -4,   4,   6,   3,   9,   7,   7, -20,  17,   6, -25,   4,  -1,  -1,  -4,  -8,  -6, -12, -15,   6,   5,  10,  -2,   3,  20,   8, -39,  12,  24,  -4,  20,  -5,   8, -54,  29,  -1,  10,  15,  -1,   6,  -9, -11,   5,   2,  -6,  11,   1,   8, -26,   3,  23,   0,  32, -23,  -6, -10,   2,  15, -22,   7,  -4,  17,   1 },
     {    0,   4,   2,  10,  32, -31, -13,  -5,  15,   1,  -6,   1,  -5,  -5,  -3,  -8, -24,  14, -15,  20,   1, -11,   3,   3,  -3,  -1,  -4, -18,   4, -14,  37, -14,  -4,  -1,   9,   1,  -5,  -2,  -6,  19,  27, -11,   0, -24,  -4,   9,  -4,   2,   3,  14,  21,  10,   0,   1, -23,  18, -10,  -8, -28, -28,   0,  12, -25,  14,   5,   6,   6,   6,  -1,  -8,  11,  -8, -13, -18,  -9,  15,   2,   1,  13,  -8, -21,  13,  -5, -10, -12, -17,   8, -13,  -3,  -4, -16,  -5,  -4,   5,  -3, -22 },
   }
 },
 {  //24
   {
     {  110, -37, -11,  -4,  -3,  -1,  -1,  -1,  -1,   0,   0,   0, -39,  26,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0, -22,   0,   7,   0,   1,   0,   1,   0,   0,   0,   0,   0,   5,  -3,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   4,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   21, -26,   5,   0,   1,   0,   0,   0,   0,   0,   0,   0,  96, -11, -21,  -2,  -5,  -1,  -2,   0,  -1,   0,  -1,   0, -46,  47,   1,  -2,   0,  -1,   0,  -1,   0,   0,   0,   0, -27,  -7,  14,   1,   2,   0,   1,   0,   0,   0,   0,   0,   2,  -5,  -3,   2,   0,   0,   0,   0,  -6,   3,   0,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -3,   0,   1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {  -32,   4,   3,   3,   1,   1,   0,   1,   0,   0,   0,   0, -10,  39,  -9,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0,   0, -73, -11,  35,  -1,   6,   0,   2,   0,   1,   0,   1,   0,  49, -57,  -9,   8,  -1,   2,   0,   1,   0,   1,   0,   1,  15,  19, -20,  -3,  -1,  -1,   0,   0,  -1,   2,   6,  -4,   0,  -1,   0,   0,   4,  -1,   0,   1,  -1,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   2,   1,  -1,   0,   1,   0,   0,   0,   1,   1,   0,   0,   1,   0,   0,   0 },
     {   32,  96, -38,  -9,  -9,  -4,  -4,  -2,  -2,  -1,  -1,  -1,   6, -35,  28,  -1,   3,   0,   1,   0,   0,   0,   0,   0, -37, -13,   4,   8,   2,   3,   1,   1,   0,   1,   0,   0, -15,  -3,   2,   0,   2,   0,   1,   0,   1,   0,   0,   0,  17, -12,  -3,   0,   0,   0,   0,   0,  -1,   6,  -3,  -1,  -1,   0,   0,   0,   1,  -3,   2,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 },
     {   13,   2,  -4,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  46, -13,   0,  -5,  -1,  -2,   0,  -1,   0,   0,   0,   0,  -1, -49,   3,  11,   1,   4,   1,   2,   0,   1,   0,   1,  41,  35, -50,  -2,  -3,  -1,  -1,   0,  -1,   0,  -1,   0, -31,  53,  23, -19,   0,  -4,   0,  -2,  -6, -20,  22,   9,  -2,   2,  -1,   1,  -6,   2,  -8,   5,   2,   0,   1,   0,  -4,  -1,   2,  -3,   1,   0,   0,   0,  -2,   1,   1,   0,  -1,  -2,   1,   0,  -1,   1,   0,   0,  -1,  -1,   1,   0 },
     {   11, -16,  13,  -4,   2,  -2,   1,  -1,   0,   0,   0,   0, -11, -75,  10,  15,   3,   5,   2,   2,   1,   1,   0,   1,  16,  21, -21,  -6,  -1,  -1,   0,  -1,   0,   0,   0,   0,  34, -28,  -1,   5,  -3,   2,  -1,   1,  -1,   0,   0,   0,  20,  38, -41,  -3,  -2,  -1,   0,   0, -24,  33,  23, -15,   1,  -3,   0,  -1,   0, -14,   8,   9,  -3,   2,  -1,   1,  -5,   8,  -5,   1,   2,   0,   1,   0,   0,  -1,   1,  -1,  -3,   3,   0,  -1,   0,   0,   1,   0,  -1,   2,   0,   0 },
     {   12,   4,  -7,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  29,  33,  -4, -11,  -3,  -4,  -1,  -1,   0,  -1,   0,   0,  50, -43,  29,   4,  -3,   0,  -2,   0,  -1,   0,  -1,   0, -50, -35, -16,  27,   4,   6,   1,   3,   1,   1,   0,   1,  12,  16, -43, -12,   5,   0,   2,   0,  -1,  29,  12, -19,  -4,  -2,   0,  -1,  -5,   0,  15,   5,  -4,   0,  -1,   0,  -4,  -1,   0,   5,   1,  -1,   0,   0,  -3,   0,   0,  -1,  -1,   1,   1,  -1,  -1,   0,   1,   0,  -1,   0,   0,   0 },
     {   -9, -49,   3,  10,   5,   1,   1,   2,   1,   1,   1,   0,   5, -30,  31,  -6,   2,   0,   0,   0,   0,   0,   0,   0, -30, -55,  -6,  23,   2,   6,   1,   3,   1,   1,   0,   1,  -5,  16, -21,  -9,   3,  -1,   1,   0,   0,   0,   0,   0,  38, -31,  -4,   6,  -5,   3,  -1,   1,  10,  32, -42,  -1,   0,  -1,   1,  -1,  -7,  22,  13, -17,   1,  -2,   0,  -1,   4,  -1,   3,   4,  -4,   1,  -1,   0,  -2,   6,  -1,   0,   0,   2,  -2,  -1,  -1,   3,   2,  -1,   0,   1,  -1,   0 },
     {   14,  15,  -8,  -3,  -2,   0,  -1,  -1,   0,   0,   0,   0,  27,   4,  -3,   1,  -2,  -1,  -1,   0,  -1,   0,   0,   0,  28,  -8,   6,  -9,   2,  -3,   0,   0,   0,   0,   0,   0,  59, -13,  24,   2,  -9,   0,  -4,   0,  -2,   0,  -1,   0, -51, -22, -17,  40,   5,   5,   1,   2,  -4,  16, -46, -19,  13,   0,   4,   0,  -4,  20,  21, -24,  -7,   0,  -1,   0,  -2,  -4,   9,  10,  -6,  -1,  -1,   0,  -5,   2,   0,   2,  -1,  -2,   1,  -1,  -2,   1,   1,   0,  -1,  -1,   0,   0 },
     {   15,  27,  88, -32, -10,  -9,  -3,  -4,  -2,  -3,  -1,  -1,   2,  12, -28,  26,  -4,   3,  -1,   1,  -1,   0,   0,   0, -12, -35, -45,   8,  16,   3,   5,   2,   3,   1,   1,   1, -12,  -9,  14, -18,   0,   1,   1,   0,   0,   0,   0,   0,   0,   8,  -1,   4,  -8,   0,  -1,   0,  11,   5,  -4,   0,   2,  -3,   0,   0,  -4,   2,   4,  -2,   2,   1,  -1,   0,  -1,  -4,   3,   0,  -1,   1,   0,   0,   0,   0,  -1,   2,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0 },
     {   -8,   5, -26,   4,   2,   1,   1,   0,   1,   1,   1,   0,   7,  49,   3,  -7,  -7,  -2,  -2,  -1,  -1,  -1,  -1,   0,  -8,   6, -42,   7,   4,   0,   2,   1,   1,   0,   1,   0,  16,  40,   7, -29,  -1,  -4,   0,  -2,  -1,  -1,   0,  -1,  17,   5,   1,   4,  -3,   0,  -1,   0, -26,  43,  20, -22,   3,  -5,   1,  -2,  -7, -35,  47,  14,  -5,   2,  -2,   1,   1, -13, -17,  13,   4,   1,   1,   0,  -2,  -3,   3,  -5,   2,  -5,  -1,   2,   0,  -2,   3,   0,   1,  -2,  -2,   2 },
     {  -11,  -5,   1,  -1,   2,   1,   1,   0,   0,   0,   0,   0, -20, -20,  -7,  10,   1,   1,   2,   2,   1,   0,   1,   0, -33,   6,   1,  -1,   1,   0,   1,   0,   0,   0,   0,   0, -21,  25, -15,  15,  -1,   1,  -1,   0,  -1,   0,   0,   0, -60,  -5, -29, -14,  20,  -1,   7,   1,  57,   6,  15, -45, -11,   0,  -2,   0,   5,   5,  31,  16, -16,  -2,  -3,  -1,   5, -10,  -5,  12,   7,  -2,   0,   0,   1,   0,  -1,  -1,   3,  -2,   0,   0,   0,   1,   0,   0,   2,  -1,   0,   0 },
     {    3,  15,  -9,  11,  -4,   2,  -1,  -1,  -1,   0,  -1,   0, -15, -14, -83,   8,  14,   5,   5,   3,   3,   2,   2,   1,   1,   8,  22, -24,  -8,  -1,  -1,   0,   0,   0,   0,   0,   3,  29, -14,   3,  -2,  -5,   0,  -1,   0,  -1,  -1,   0,  36,  11,   6, -14,   2,  -3,  -2,  -1,  -4,  20, -10,   1,   1,   1,   0,  -1, -17,  25,  20, -26,   2,  -3,   2,  -1,   2, -20,  31,   8,  -8,   2,  -2,   1,  -2,  -3,  -4,   7,  -1,  -4,   7,  -1,   0,  -2,  -1,   1,   0,  -2,   3,   1 },
     {   -2,   8,  -4,  -4,  -2,  -2,   0,   0,   1,   0,   0,   0,  -4,  16,  28,  -3,  -4,  -2,  -4,   0,  -1,   0,  -1,   0,   6,  47, -36,  25,   1,  -2,  -1,  -1,  -1,  -1,   0,   0,  -8, -35, -53, -13,  24,   4,   6,   2,   3,   1,   2,   1, -17,  17,  12, -27, -11,   2,   0,   2,   4,  11, -16,   5,   1,  -4,   2,   0,   5,  30,  15, -32,   1,  -3,   0,  -2,  -2, -13,  37,   9, -10,   0,  -2,   0,  -1, -10,  -5,   9,   1,  -3,   3,  -1,   2,  -2,   0,   1,   0,  -1,   2,   2 },
     {   -8, -13, -65,   7,  11,   9,   4,   2,   0,   2,   1,   1,   1,  -1, -29,  23,  -3,   2,  -1,   2,   0,   0,   0,   0,  -4, -27, -48,  -6,  17,   2,   5,   2,   3,   0,   2,   0, -13, -24,  29, -25,  -7,   4,   0,   1,   0,   0,   1,   1,  -6,  29,  -4,   5,  -9,  -4,  -2,  -1,  30,   1,  -2,   0,   0,  -4,  -1,  -2,   9,  20, -21,   0,   4,   0,   0,  -1, -11,  23,  13, -21,   2,  -1,   0,  -1,   1,  -7,  17,   3,  -1,   3,  -1,   0,   0,  -1,   4,   0,  -1,   1,   2,  -1 },
     {   -7,  -3,   3,  17,   2,  -1,  -2,   0,  -1,  -1,   0,   0, -14,  -4,  -4,  -4,   3,   1,   1,   1,   1,   1,   0,   0, -20, -12,  -4,   3,   2,   3,   0,   3,   0,   1,   1,   0, -36,  14,  -3,   5,  -3,   1,   1,   2,   0,   0,   0,   0, -16,  28, -17,  22,   8,  -7,   1,  -2, -55, -29, -14, -21,  35,   3,   7,   0,  44, -23,   3, -43, -18,   9,  -1,   3,  13,  15,  11,   9, -19,  -5,  -3,  -2,   7,   1,   2,   6,   1,   2,   1,   1,   3,  -1,   0,  -1,   1,   2,   1,   1 },
     {    4,  12,  14,  29, -11,  -7,  -5,  -1,  -1,   0,  -1,   0,   5,  23,   0,  -6,   4,  -1,  -1,   0,   0,   0,   0,   0,  -5,   8, -35,  -6,  -9,   7,   1,   2,   1,   1,   1,   1,  19,  27, -19,  34,   5,  -8,  -1,  -3,  -1,  -2,  -1,  -1,   6, -28, -37, -15,  26,   6,   4,   1, -15,   9,   1, -14, -14,   4,   0,   1,   0,  15, -37,   1,   8,  -6,   3,  -1,  -3,  36,  18, -39,   3,  -4,  -1,  -3,   1, -12,  33,   5,   0,  -3,   0,   2,   0,  -4,   7,   1,   0,  -1,   1,   0 },
     {   -8,  -7, -17, -83,  18,  11,   7,   3,   4,   1,   2,   1,  -1,   6,  -8,  42, -22,  -1,  -2,   0,  -1,   0,  -1,   0,   4,   9,  29,  42,   1, -14,  -3,  -2,  -1,  -2,  -1,  -1,   5,  27,  -9, -13,  11,   3,  -2,  -2,  -1,   0,   0,   0,   7, -12, -28,  -3,   2,   4,   5,   1, -19,  -5,   0,  -6,  -1,  -1,   1,   1,   5,   5, -12,  -4,   2,  -2,   0,   0,   0,  13,   8, -14,   0,  -1,  -1,  -1,   0,  -4,  10,   4,   1,  -2,   1,   1,  -1,  -1,   1,   1,   0,   0,   0,  -1 },
     {   -4,  -3,   5, -21,   5,   0,   2,   1,   1,   0,   0,   0,   6,  17,  55,  12,  -6,  -6,  -3,  -4,  -3,  -1,  -1,  -1,  -5,   9,   2, -45,  10,   5,   1,   1,   1,   1,   0,   0,  -6,  26,  17,  16, -33,   1,  -3,  -1,  -1,  -1,  -1,   0,  28,  39,  -9,  21,  17, -12,   1,  -5,  17, -14,   2,  -9,   9,   8,  -3,   3, -26,  25,   1,  -7,  -4,   1,   2,   0,  -9, -19,  31,  -1,  -4,  -1,   0,   1,   5, -18,  -7,  16,  -3,  -1,  -4,  -3,   1,  -4,  -1,   4,  -1,  -1,  -2,   0 },
     {    8,   2,   4,  -2,   1,  -3,  -2,  -2,   1,   0,   1,   0,   7,   1,  -7, -14,   0,   3,   0,  -1,  -1,   1,  -1,   0,  13,   2,  -6,   5,   1,  -4,  -1,   1,  -1,  -1,   0,   0,  16,  -6,  -6, -18,  -2,   0,   4,   1,   0,   0,   1,   0,  41, -11,  12,  -6,  12,  -3,  -3,   0,   5, -42,   2, -37,  -6,  19,  -1,   4,  38,  36,  12,  19, -53,  -7,  -4,   0, -28,  22,  -3,  33,  19, -12,   1,  -3,  -4,  -4,   2, -11,  -6,   2,   2,   7,  -1,   1,   0,  -3,  -3,   0,   2,   4 },
     {   -2,  -2,   2, -23,   0,   8,   6,   0,  -1,   0,   0,   0,  -8,  -6,  -8, -70,  -4,  13,   6,   4,   1,   1,   1,   1,  -5,  -1,  -9,  50, -27, -15,   0,   1,   1,  -1,   0,   0,   6,   7,  50,  34,  13, -19, -10,  -4,  -2,  -2,  -1,  -1,   3,  34,   5,  -8,   7,   7,  -6,  -4,  16,   3, -12,  -2,  -3,   4,   3,   1,  -7,  -3,  -2,  -5,   0,  -2,   2,   1,  -5,  -3,  16,   1,   1,   0,  -1,   0,   4, -13,   5,   7,   0,  -1,  -6,   4,   0,  -1,  -2,  -1,   0,   1,  -2,   2 },
     {    2,  12,   5,  21,  -1,  -2,  -2,   1,  -1,   0,  -1,  -1,  -7,   3,  -8,   1, -23,  -1,  -1,   1,   3,   2,   1,   0,   3,  26,   4,  29,   6,  -9,  -2,  -1,  -2,  -2,   0,   0,  -8,   7, -26,   3,  -4,   7,   2,   2,   1,   1,   0,   0,  12,  33,  -9,  39,   1, -14,   0,  -4,  20,  -6, -35,  -5,  29,   5,   1,  -1, -32,   7,  -5,   4,   4,   0,   1,  -1,  -6,   2, -38, -20,  26,  -2,   5,   0,   4,  20,  18, -44,   2,  -9,  20,   7,   0,  -4,   2,  -5,   1,  -3,   5,   1 },
     {   -4,  -4, -16,  -2,   6,   6,  -3,   3,   0,   1,   0,   0,  -2,   2,  -3,  31,  -4,  -1,  -3,   1,  -1,   0,   0,  -1,  -6, -10, -33,   5,  -6,   6,   1,   0,   2,   1,   1,   2,  -4, -10,  12,  40,   6, -14,   0,  -2,   0,  -2,  -1,  -2,  -3,   1,   2, -15,  29,   9,   0,  -2, -44, -23, -18,   1, -23,   9,   6,   3,  -6,  24,   5,  14,  13, -18,  -3,  -5,   2, -40, -14,   7,  19,  10,   0,   2,  18,  19, -36, -23,   0,  10,  18, -10,   2,   1,  -2,   2,  -1,   2,   3,  -6 },
     {    4,   0,   1,  49, -44,  15,  -1,  -2,   1,   2,  -2,   0,   6,   2,   7,  23,   5,  -8,  -8,   1,   0,   2,   1,   0,   8,   2,  10,  44,   8, -18,  -5,  -1,  -1,  -1,   0,  -1,  11,  12,  28, -31,  11,   7,  -6,  -1,  -2,   0,   1,  -1,  11,   1, -43, -18,   1,   8,  10,   2,  14, -37,  11,   2,   2,  -5,   2,   1, -18,  -4,   0, -10,   4,   8,   0,   1,   0, -16,   9,   6, -10,   2,   0,   0,   2,  -8, -22,  10,  -2,  11,  -6, -15,  -1,   0,   4,  -2,  -2,   1,   0,   0 },
     {    5,  -1, -13, -19,   9,   9,   1,   1,   0,  -1,   0,   1,  12,   3,  -6,   9,   5,   1,  -2,   1,  -1,   1,  -1,   0,  10,   0, -29,  -6,  -7,   4,   4,   2,  -1,   3,   0,   0,  16, -15, -14,  23,   8,  -6,   1,   2,   0,   0,   0,  -2,  26, -12,   7, -13,  24,  -6,   0,   0,  36, -26,  -3, -18,  11,  10,  -4,   1, -27, -43,  -4, -29, -14,  19,   1,   3,  39,  10, -14,   7, -36,  -6,  -2,   0, -11,  35,  -3,   9,  -1,   2,   9,  -4,  -4,   5,   2,   3,  -1,   0,   3,  -1 },
     {  -10, -14, -14, -35, -96,   6,   8,   8,   4,   6,   4,   4,  -1,   0,  -4,  -7,  34, -12,  -1,   0,  -1,  -3,  -1,  -1,   1,   7,  -2,   5,  41,  22,  -6,  -6,  -2,  -1,  -3,   0,  -1,   0,  -5,  20, -17,  -5,   5,  -1,   0,   0,   2,   0,   0,   2,  10,   3,  -1, -10,  -4,   0,  -3,   8,  -7,  -1,  -1,   1,  -2,   0,  11,  -2,  -1,   1,   2,  -1,   2,   0,   0,   4,  -5,   3,   4,   0,   0,  -1,  -3,   6,   6,  -9,   2,  -6,   3,   5,   0,   2,  -3,  -1,   1,  -2,   1,   0 },
     {   -2,  -2,  -1,   3,  10,   3,   6,  -5,   0,  -1,   1,  -2,  -7, -11,  -9, -19,   6,  -4,   5,   0,   0,   0,   1,   1,  -1,  -3,   2,  14,  40,  -1,   2,  -4,  -3,  -3,  -4,   1,  -9, -18,  -4, -13,  -6,  21,   7,   0,   1,   1,   0,   1, -12, -13,  18,   8,   3,   5,  -2,   1, -33,  -8,  12, -33,  -9,  19,   4,   8, -65,   4, -12,   6, -25, -14,   2,   4,  37, -10,  14, -10,  15,  -8,  -7,  -2,  16, -16,  14,  14,  -1,   6, -20,   7,  -2,   4,   1,  -3,   2,   1,  -5,   2 },
     {    1,   1,  -3,  13, -11,  14,   4,   6,  -3,   1,  -2,   1,  -2,  -4, -15,  17, -47,   4,  -7,   7,   4,   3,  -3,  -1,   1,  -2,  -9,  29,  -5,  -1,   1,   0,  -1,  -1,   1,  -3,   2,  -9, -21,  32, -23,  -8,   2,   1,   3,   2,   2,  -1,   2, -23,  25,  35,  15, -23,  -6,  -1,   5,  20,  25,   0,  23,   7,  -8,  -5,  10,  -8, -16,   6,  -1,  14,   6,  -5, -17,  -8,  11,   8,   5,  -6,   6,  -1,  -5, -28, -29,  20,   0,  16, -27, -18,   4,   7,   6, -11,  -2,   3,   0,  -1 },
     {    2,   1,  -7, -18,  14,  -4,   4,   0,   2,  -2,   3,  -1,  -1,  -3, -15, -60, -12,   8,   5,  -1,   2,   2,   4,   3,   1,   1, -26,  -9,  29,  -4,  -8,   1,   0,   3,   0,   0,   7,   8, -19,   0, -24,  22,   0,   0,  -1,   1,  -1,  -1,   6, -22, -47,  10, -26, -14,  13,   3,   3, -25,  16,  17,   8, -17,  -5,   3,  14,  13,  -5,   2,  19,   1,  -7,  -1,  15, -23,   2,   1,   5,  11,   1,   2,  -3,  15, -25,   1,  -3,  10,  10, -14,  -1,  -1,   4,   1,  -1,   2,   3,  -2 },
     {   -1,   1,   0, -10,  15,  -1,  -6,  -5,   2,   2,   3,   0,   1,   6,  -2,  -6,  61,   2,  -9,   4,  -6,  -1,  -1,  -1,   1,   5,  -1,  26, -45,  31,  13,  -4,   2,  -1,  -3,   2,  -5,  -9,  -7, -36, -41, -30,  15,   9,   3,   2,   6,   2,  -9,  -4, -21,  10,  22, -20, -18,   1,  -1,   1,  13,  16,  12,  17,  -5,  -7,  -8,  14,  -1,   2,   1,   6,   8,   4,   5,  -6,  -5,  -9,   4,  -6,   0,   1,   5,  -1,  -9,  -5,   1,   4,  -4,  -6,   0,   1,  10,  -2,   3,  -1,   3,   3 },
     {    6,   0,   1,   8,  32, -29,   5,  -6,   4,  -3,   5,  -1,   7,  -1,  -3,   4,  27, -11,   7,  -2,  -2,   0,   1,   4,  11,   3,   8,  20,  41,  18, -27,  -5,  -4,   0,  -5,   3,  11,  -1,   2,   2,  12,  -3,  17,  -9,  -4,   0,  -1,   3,  14,  -1,  11,   5,  33,   3,   7,   0,  23,   4,   2, -15,   8,  11,  11,   4,  35, -25,  -4, -16,  24, -11, -14,   8, -18, -32,   1, -23,   9,  32, -11,   6,  10, -12,   2,   5,  -1,  -1,  -5,   6,  -3,  -2,  -8,   3,  -9,   4,  -7,   4 },
     {    1,  -6,  -1,  -1, -28,  11,  -7,   0,   5,   1,   0,   0,   2,  -8,   3,  -3, -33,  16,   0,   6,   0,  -2,   2,   3,   2,  -9,  -1,  -8, -31,  -4, -12,   1,   0,  -3,   2,   2,   0, -18,  -4, -15,   0, -23,  12,   0,   3,  -5,   2,   1,   7, -10,   6,  21,  19,  -6,   6,  -6,   2, -25,  31,  -3,   3, -17,   4,  -1,   4,  13,  43,  -8,  26, -23,  -9,  -4,  26,   1,   8, -25,   4,  24,  -4,   0,  19,   9,  25,   8,  -2, -16,  -7,  32,  -6,  -3, -18,  -1,  -3,   2,  -5,  -4 },
   },
   {
     { -119,  33,   7,   5,   3,   2,   1,   1,   1,   1,   0,   0,  12, -12,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  22,  -7,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   5,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   6,  -4,   0,   0,   0,   0,   0,   0,   4,  -2,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   1,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    4,  -9,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0, 115, -13, -13,  -4,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0,  -6,  32,  -6,   0,  -2,   0,  -1,   0,  -1,   0,   0,   0, -34,  12,   4,   0,   1,   0,   0,   0,   0,   0,   0,   0, -11,   2,   3,   0,   1,   0,   0,   0,  -8,   1,   3,   0,   1,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,  -4,  -1,   2,   0,   0,   0,   0,   0,  -2,  -1,   1,   0,  -2,  -1,   1,   0,  -1,  -1,   0,   0,  -2,   0,   0,   0 },
     {   15, -10,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -7, -20,   3,   0,   2,   1,   1,   1,   1,   0,   0,   0, 107,  -9, -17,  -2,  -5,  -1,  -2,   0,  -1,   0,  -1,   0,  -7,  46,  -7,  -1,  -2,  -1,  -1,  -1,  -1,   0,   0,   0, -33,  10,   9,  -1,   2,   0,   1,   0,  -7,  -1,   3,   0,   0,   0,   0,   0,  -7,   1,   3,   0,   1,   0,   0,   0,  -3,  -1,   2,   0,   0,   0,   0,   0,  -3,  -1,   1,   0,  -1,  -1,   2,   0,  -1,  -1,   1,   0,  -1,   0,   0,   0 },
     {    3,  22,  -4,   0,  -2,   0,  -1,   0,   0,   0,   0,   0, -28,   3,   5,   1,   1,   0,   1,   0,   0,   0,   0,   0,   8,  27,  -4,  -1,  -1,  -1,  -1,  -1,   0,   0,   0,   0, -89,   4,  26,   0,   6,   0,   3,   0,   2,   0,   1,   0,   9, -63,   7,   4,   2,   2,   1,   1,  30,  -6, -17,   2,  -2,   1,  -1,   0,   5,   4,  -2,  -2,   0,   0,   0,   0,   5,  -2,  -1,   1,   0,   0,   0,   0,   2,  -1,  -3,   0,   3,   1,  -2,  -1,   1,   0,  -1,   0,   2,   1,  -1,   0 },
     {   13,  82, -18,   0,  -7,  -2,  -4,  -1,  -2,   0,  -1,  -1,   1,  27,  -2,  -3,  -1,  -1,   0,   0,  -1,   0,   0,   0, -21,  -4,   5,  -1,   1,   0,   1,   0,   1,   0,   0,   0,   6,  29,  -1,  -3,  -1,  -1,   0,  -1,   0,   0,   0,   0, -54,  12,  28,  -3,   4,  -1,   2,  -1,  -7, -43,  19,   3,   2,   1,   1,   1,  10, -21,  -6,   5,  -1,   1,   0,   1,   5,  -5,  -3,   0,   0,   0,   0,   0,   3,  -4,   1,   1,   2,  -4,  -2,   2,   1,  -2,  -2,   1,   1,  -2,  -1,   1 },
     {   24,  74, -20,  -2,  -8,  -2,  -3,  -2,  -2,  -1,  -1,  -1,   5, -14,   3,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  21,   4,  -5,  -2,  -2,  -1,   0,   0,   0,   0,   0,   0,  -5, -22,   3,   2,   2,   1,   1,   1,   0,   0,   0,   0,  48,  17, -28,  -2,  -4,  -1,  -1,   0, -26,  56,  11,  -9,   1,  -3,   0,  -2, -17, -10,  25,  -1,   2,   0,   0,   0,  -3,  -8,   5,   3,   0,   1,   0,   0,  -2,  -6,   3,   1,  -1,  -2,   2,   1,  -1,  -4,   1,   1,   0,  -1,   0,   1 },
     {  -16, -31,  10,   1,   3,   1,   2,   1,   1,   1,   1,   0, -24,  11,   1,   2,   1,   0,   1,   1,   0,   0,   0,   0,   2,  43,  -3,  -3,  -2,  -1,  -1,  -1,  -1,  -1,   0,  -1, -30,  -5,  14,   0,   3,   0,   1,   0,   0,   0,   1,   0,   8,  30,  -1,  -4,  -1,  -2,   0,  -1, -48,  -8,  45,  -1,   5,   0,   1,  -1,  12, -68,   2,  14,   0,   4,   0,   3,  12,  -4, -26,   4,  -1,   1,   0,   1,   5,   2,  -8,  -3,   2,   2,  -2,  -1,   2,  -1,  -1,   0,   1,   1,  -2,  -1 },
     {   17,  16,   3,  -2,  -2,  -2,  -1,  -1,  -1,   0,   0,   0,   6, -75,  12,   5,   3,   2,   2,   1,   1,   1,   1,   1,   3, -20,   0,   4,   1,   1,   0,   1,   0,   0,   0,   0,  -5, -28,   0,   7,   1,   3,   1,   1,   0,   1,   0,   0,  11,   0,  -8,   2,  -1,   0,   0,   0,  -4, -34,   2,   8,   0,   2,  -1,   1,  31, -12, -48,   8,  -4,   2,  -1,   1,   0,  49, -23, -13,   0,  -3,   1,  -2,  -3,  17,   9,  -9,  -2,   7,   5,  -1,  -1,   2,   2,  -1,  -1,   5,   1,  -1 },
     {    0,  -4,   5,   2,   1,   1,  -1,   0,  -1,   0,   0,   0,  -2, -61,  21,   3,   5,   2,   3,   1,   1,   0,   1,   0, -36, -27,  -7,   7,   2,   4,   1,   2,   0,   1,   0,   0,   4,  42,  -6,  -3,  -2,  -1,  -1,  -1,   0,  -1,   0,   0, -12, -21,   8,   4,   2,   2,   0,   0,  13,  24,  -1,  -6,  -2,  -1,   0,  -1, -22, -31,  34,   9,   1,   2,   0,   1,  19, -36, -32,  17,  -1,   4,  -1,   2,   2,  19, -22,  -5,   3,   6,  -4,  -4,   0,   6,   0,  -3,   2,   1,   0,  -1 },
     {   -6,   5,   7,  -3,  -1,  -1,   0,   0,   0,   0,   0,   0, -30, -61,  -2,   8,   3,   5,   2,   3,   1,   1,   1,   1, -27,  64, -17,  -3,  -2,  -1,  -1,  -1,  -1,   0,   0,  -1,   5,  22,  -4,  -4,  -1,  -2,   0,   0,   0,  -1,   0,   0,  -7,  26,   7, -10,   1,  -3,   1,  -1,  -8,  -6,   3,   0,   2,   0,   1,   0,   5,  33,   0, -16,   0,  -4,   1,  -2, -13,  -6,  44,  -1,   0,   0,   0,  -1,   6, -30,   1,  13,   0,  -2,  -3,   3,   3,  -6,  -3,   2,   0,   0,  -1,   1 },
     {  -12,  -9,  -5,   2,   0,   1,   0,   1,   0,   0,   0,   1,  -7,  12,   8,  -6,   1,   0,   0,   0,   0,   0,   0,   0, -28, -33, -20,   8,   2,   4,   1,   2,   2,   1,   0,   1, -37,  51, -21,  -5,   0,  -1,   0,   0,   0,  -1,   0,  -1,  10,  19,  -4,  -9,   1,  -2,   0,  -1,  -2,  30,   6, -10,   0,  -3,   0,  -1,   3,  -4,  -3,  -1,   1,   0,   1,   0,   0,  44,   4, -25,   1,  -5,   0,  -3,  -8,  -8,  58,  -4,   5, -28,   2,  19,  -1,  -4,  -3,   3,   4,  -5,  -1,   2 },
     {   -1,   3, -23,   2,  -1,   0,   1,   0,  -1,   1,   0,   1,  -3,   3, -18,   4,   0,   0,   1,   1,   1,   0,   0,   0,   3,  59, -20,  -5,  -2,  -3,  -1,  -1,   0,   0,   0,   0,  57,  18,  27,  -9,   0,  -4,  -1,  -3,  -1,  -2,  -1,  -1,  -4, -37,  -2,  10,   2,   2,   0,   1,  10,  17, -12,  -4,  -1,  -1,   0,   0, -12, -15,  -6,   5,   1,   1,   1,   2,   7,  28, -28, -15,   0,  -3,   1,  -1, -10,  27,  30, -19,   2, -14,  20,   5,  -2,  -3,   4,   5,  -1,  -4,   0,   1 },
     {   13,  18, 104, -24,   4,  -9,   0,  -5,  -2,  -3,  -1,  -2,   9,   9,  29, -10,   3,  -2,   1,  -1,   0,  -1,   0,  -1,   7,   6,  -2,  -2,  -4,   0,   1,  -1,   0,   0,   0,   0,  21,   4,  15,  -5,  -2,  -2,  -2,  -1,   0,   0,   0,   0,   8, -19,   6,   3,  -1,   2,   0,   1,   6,   3,   8,   2,  -3,   1,  -1,   0,  -3, -12,   0,   8,  -2,   3,   1,   1,  -1,   5,   3,  -2,  -2,  -1,   1,   0,  -2, -17,  16,  11,   1, -13, -15,  14,  -3,   2,  -9,   1,  -1,  -1,  -6,   2 },
     {    5,   3,  -6,  -2,   0,   0,   1,   1,   0,   0,   0,   0,   8,  14,  -8,   1,  -4,  -2,  -2,   0,   0,   0,  -1,   0,   2,  -4, -28,   7,  -3,   2,   1,   1,   0,   0,   0,   0,  14,  46, -13, -14,  -1,  -6,   0,  -1,   0,  -2,   0,   0,  88,  -7,  40, -11,  -2,  -4,  -3,  -2,  -4, -35,   6,  14,  -2,   5,   0,   2,   1,   4, -13,   4,  -2,   1,   1,   1, -10,  -6,  -2,   6,   2,   1,   1,   0,  -1,   3, -22,  -2,  -6,  17,   2, -10,  -1,   0,  10,   0,  -3,   0,   1,   2 },
     {   -1, -15, -46,   8,  -2,   3,   2,   3,   2,   2,   1,   0,   9,  -6,  23,  -1,  -2,   1,   1,  -1,  -1,   0,   0,   0,   2, -11,   2,   1,   2,   1,  -1,   0,   0,   0,   0,   0,  20,   2,  33,  -3,   0,  -2,  -1,  -3,  -1,  -1,   0,  -1,   4, -29,   3,  14,  -1,   3,  -2,   0,  -5,   2,  14,   1,  -1,   0,   0,   0,  -4, -23,  -9,  11,   0,   3,   0,   1,   1,  14,  21,  -4,  -3,  -1,  -1,  -1,  -5, -53,   5,  41,   1,   5, -61,   2,  -4,  16,  -2, -14,   1,   4,  -5,  -3 },
     {   -3,   4,  -2,  -2,   6,  -3,   2,  -1,  -1,   1,   0,   1,   4,  25,  77,   0,  -1,   0,  -2,  -4,  -3,  -2,  -1,  -2,  -8,  12, -42,   5,   0,   4,   2,   1,   0,   1,   1,   0,   1,   8,  -8,  -1,  -5,   0,  -3,  -1,   0,   1,   1,   0, -19,  -1, -30,  12,   0,   2,   2,   0, -25,  11, -21,  -4,   4,   1,  -1,   0,  11,  13, -18,  -4,   2,   0,   3,  -1,  -7,   6,  -7,  -8,  -1,   0,   1,   1,   1,   0, -23,  -4,   4,  31,   7, -32,   5,  -4,  36,   0,   0,  -4,   2,   8 },
     {   -2,   2,  17,  -1,   1,   0,   0,  -2,   0,   0,   0,  -1,  -7,  -5, -39,   4,   1,   2,   3,  -1,   0,  -1,   2,   0,  -9, -24,  -1,   4,   1,   3,  -1,   2,   0,   1,   1,   1,  -4,   9,  17,  -1,   1,  -1,   0,  -1,  -1,   0,  -1,  -1,  -6, -31,   6,  22,  -3,   5,   0,   1, -95,  -1, -40,  11,   6,   3,   4,   1, -10,  10,  -5, -21,   4,  -4,   1,  -2,  14,  -2,   8,  -6,   0,   0,   0,   0,   8,   2,   2,  -2,   3,  -2,  -2,   0,   2,   4,  -1,  -4,   1,   2,   3,  -2 },
     {    1,   9,  25,  -5,   3,  -5,  -2,  -1,  -2,  -2,  -1,   0,  -6,   4, -74,   5,  -1,   7,   1,   4,   1,   2,   1,   2,  -4,  -8, -42,   6,   3,   3,   3,   1,   1,   0,   0,   0,  -2,   1, -11,   3,   2,   1,  -2,   1,   0,  -1,   1,   1, -15,  -1, -29,  -7,   3,  -3,   2,   0,  26,  16,  14, -12,   3,  -4,  -2,  -2,   5, -13,  -9,   3,   0,  -1,   1,   1,   5,  16,  -1,  -8,   0,  -1,  -2,   1,  -3, -11, -18,  11,  -2,  30, -27, -27,   0,  11,  34, -21,   5,  -2,  14,   3 },
     {   -4,  -1,   5,  -2,  -6,  -2,   3,   0,   1,   0,   0,   0,  -5,  -3,   0,   2,  -5,   1,   1,   2,   1,   1,   0,   0,  -6,  -5,  20,  -6,   4,   2,   1,   1,  -1,  -1,   0,  -1, -11,   6,  23,  12,  -2,  -1,  -2,   0,  -1,   1,  -1,   0,  -1,  17,  -3,   3,   1,  -1,   0,  -1,   8,  -1,  17,  17,  -3,   4,  -3,   0, -93,   6, -49,  16,  14,   1,   3,   0,  -1,  -5,   1, -29,   8,  -6,   3,  -1,  14,  -5,  -2,  -4,   4,  10,  12, -15,   2, -13,  14,  10,   3,  -2,  -9,   3 },
     {    0,   3,  -5,   9,   1,   3,  -5,   1,  -2,  -2,   2,   0,  -1,   6,  12,  10,   4,  -2,  -3,  -1,   0,  -1,   0,  -2,   3,  17, -47,  11,  -1,  -1,   0,   0,   3,   2,   3,   1,  -3, -34, -54,  10,   4,  10,   4,   2,   2,   0,   2,   0,  -6,  -7,  12, -13,   4,  -3,  -2,   3,  -6, -11, -10,  15,  -7,  -1,  -2,   0, -37,   3, -11,   7,   5,   0,  -1,  -1,   7, -17, -17,  -9,  11,   1,   1,  -1,  17,   3,  16,  -8,  -1, -16, -30,  19,  -2,  33, -30, -26,  -1,   9,  22, -15 },
     {   -4,  -5, -16, -63,  15,  -7,   8,   2,   1,   1,   1,  -1,   1,  -1,  -4,  15,  -5,  -6,   4,  -5,   2,  -1,  -1,   0,  -4, -26, -67,   5,   6,   6,   6,   7,   0,   2,   3,   1,  -3, -26,  41,  -2,  -5,  -6,   2,  -3,  -2,   1,   0,   1,  -8,  11,  16,  21,  -4,   7,   0,  -2,  10,  -1,  20,   7,   1,   2,   0,  -1,   3,   6,   3,   2,   0,   1,  -1,   1,   5, -15,  13,   3,  -3,  -1,   3,   2,  -5,   2,  -2,  -5,  -4, -13,  19,   7,   2, -13, -20,  14,   2,   9,  -8, -11 },
     {    5,   0,  19,  94, -20,   8,  -9,  -3,  -5,  -1,   1,  -1,   2,   2,  -1,  16,  -3,   8,  -7,  -3,  -1,   0,   1,  -1,  -3, -17, -44, -28,  10,   3,   6,  -1,   1,   0,   4,   0,   1, -10,  30,  -7,   0,   5,   3,   1,   2,   1,  -1,  -2,  -5,   2,   3,  -1, -11,   4,  -1,   0,   2,   0,  10,   8, -10,   0,  -3,  -1,   6,   5,   6,   9,  -3,  -2,  -2,  -2, -15,  -6,   0,   2,   2,   0,   0,   0,   1,  -4,   1,   2,  -3,  -3,  13,   1,  -2, -14,  -9,  15,   3,   8, -14,  -3 },
     {    0,   7,   3,   0,   7,  -2,   7,  -2,   1,   2,   1,   0,  -3,   4,  -4,   9,  -5,  -2,   0,  -4,   1,   0,   1,   1,  -3,   4,  10,  -6,   3,  -2,  -1,   0,  -2,   1,   3,   0,  -5,  31,  40,   7,  -7,   1,   2,  -3,  -2,  -1,  -1,  -1,  11,  43, -33,  38,  -4,  10,  -2,   1,  11, -11, -26,   4,   0,  -2,   1,   0,  10,  16,  -2,  12,  -2,  -2,   0,   0, -36, -13, -32,   5,  15,   2,   8,   2,  -3,   2,   0, -18,  -4, -10, -27,   1,  -4,  20, -21, -36,   6,   7,  26, -15 },
     {    2,   3,   4,   5,  -9,  -5,  -3,   2,   5,   1,   3,  -3,   7,  10,   5,  65, -13,  -2,  -8,  -6,  -1,  -3,   2,  -1,   3,   5,   5,   5,  -1,  10,  -3,   0,  -3,  -1,  -1,  -1,   4,  23, -13,  29,  -8,   7,   1,  -1,   2,   0,   0,  -4,  13,   1, -40,  10,  -2,   6,   5,  -2,  13, -11,  -6,   5,   6,  -6,   3,  -3,   3, -24, -19, -31,   5,  -7,   3,   0,  42,  -8,  27,  -1, -19,  -1,  -4,   2,  -5,  20,  -6,  23, -10, -15,  -1,  -2,  -5, -11, -21,   7,   2,  10,  -9, -12 },
     {    4,   2,   3,  18,   3,   5,  -5,  -1,  -1,  -2,  -5,   3,   2,   0,   5, -43,  27, -18,   9,   2,   6,   3,  -1,   1,   2,   2, -16, -14,  -3,  -7,   7,  -3,   5,   3,  -4,   2,  -3,  -7,  28,  14,  -8, -12,  -6,   1,  -2,   0,  -2,   3,  11,  45,  27,   4,  17, -10,  -4,  -2,  18,   0, -44,  -7,   6,   3,   2,   2,  -2, -13,  -4, -19,  13,   1,   0,   2,  59,   4,  11,  -6, -11,   0,  -3,  -1,   1,  13,  -1,  13,  -4,   7,  -8,   5,  -1,   8,   9,  -7,  -4, -10,   7,  -1 },
     {    0,  -3,  -4,  15,  -6,  -7,  -1,   4,   0,   0,   2,   1,  -1,  -2,  -6, -67,  16,   0,  10,   0,   2,   1,   2,   0,   0,   0, -18,  43,   0,   7,  -1,   0,  -4,  -2,   0,   0,   7,   8,   0,  38,  -6,   3,  -1,   0,  -1,   0,  -2,  -3,   7, -18, -51,  10,   3,  -2,   6,   0,   1, -39,  14, -10,  15,   0,  -1,   0, -12,  -6,   9,  -6,  -7,   9,  -4,   0, -11, -10,  12,  10,   1,   1,  -1,   0,   6,  -6,   4,  -5,   1,  -8,  14,   7,   4, -14, -12,   7,   1,   9,  -8,  -7 },
     {   -1,   4,   3,  16,   8,  -5, -20,  -4,   2,   0,  -6,   3,  -6,   2,  -5, -13, -31,   9,  -1,   8,  -6,   5,  -5,   7,  -1,  11,   4,  53, -17,  -2,   4,   0,   1,   3,  -6,   0,  -1,  -2, -10,  31,  -2,  -7,  -7,   2,  -4,   1,  -1,   4,  -8,   8,  31,  27, -10,   8, -10,  -1,   3,  37,  -2,  49,  -2,  -8,   0,  -4,  11,   1, -14,  10,  17,   2,   3,   4,  -4,  15,  -5,  17,   3,   8,   2,   3, -28,  -6, -21,   7,  -2,  -7,   3,   0, -12, -14,  -7,   1,   0, -13,  -6, -14 },
     {    2,   0,   2,  18,   0, -14,  -1,  -1,  -5,   3,  -6,   3,   1,  -3,  -1,  25, -23,   6,   8,  -1,   6,   0,  -5,   2,   6,   2,   2,  47, -22,  18, -10,  17,  -5,   5,  -6,   1,  -2, -14,   8, -14, -12, -14,  -8,   6,  -4,   5,   0,   3,   5,   9,  13,  20, -10, -12,   2, -22,   5, -11,  -1, -35,  29,   4,  10,   3,   0,  16,   1,   9, -22,   7,   4, -13,  11, -26, -21,  -9, -13,   2,  -2,  11,  19,  -4,  31,  15,   5,  -1,   1,  13,  -1,   0,  28,  -7,   7, -13,  -3,  28 },
     {    4,   7,  -5,  25, 100, -50,  12, -14,   1,   1,  -2,   3,   0,   0,  -1,   2,  14,   8,   1,  -3,  -4,   0,   2,   3,  -2,  -2,   4, -12, -28,   1,  -5,   2,   3,   0,   2,  -1,  -1,   2,  -5, -13,  13,  -7,   3,  -2,  -3,   0,  -1,  -2,   0,  -3,  -9,  -1,  -4,   1,  -8,   0,  -1,  -7,   7,  -4,  -7,  -1,  -1,  -2, -10,   1,   1,   3,   0,   2,  -2,  -1,   6,   5,   7,   0,  -2,   6,  -1,  -3,  -6,   3,   0,   4,  -6,   8,   5,   2,  -1,  -3, -10,   3,  -2,   5,  -3, -12 },
     {   -2,   1,   1,  -8,   8,  26, -28,  -2,   5,  -9,  -7,   8,  -1,   2,   7,  16,  15,  12, -21,  -7,  -2,  11, -11,  12,  -1,   1,   0,  20,   0,   2,   8,   9,  19,   7, -13,   2,  -2,   1,  25,  15,  -5, -15,   0,   1,   1,   0,  -8,  10,  -2,   1, -16, -41,   2,  24,   2,   0, -15, -24,   9,   0, -27,   3,  -8,   6, -18,  28,  32,  18,   0,  -1, -10,  -9,  14,  21, -10,  10, -14,  -4,   5,  -1, -24,  12,   3,  23, -18,   8,   3,   3,  -3,  -2,   0, -11,  14, -20,  18,   7 },
     {   -1,   1,   1, -12,  -5,  21,  14,   1,  -3,  -4,  -2,   0,  -1,   1,  -4,  10,   6,  35,  -7,  19,  -2,   0,  -1,   1,  -3,   2, -10, -37, -22,   2,   4,  -2,  13,   1,   0,   5,  -4,  -7, -26,  23,  25,  10,   5,  -7,   0,   1,  -2,  -4,   5,   8,   5,  52,   6, -14,  -9,  -3,   2, -25, -14, -11,   7,   1,  -8,   6, -23,  -3,  25,  -5,  -3,  13, -11,  -4, -10,  17, -21,  13, -12,   1,  -5, -12, -25, -18,   3,  17,  13,   1,   6,  -1, -11,  -5,   7,   7,   4, -11, -24,   3 },
     {    1,   3,   3,   0,   0,  25,  -9,   4,   4,   6,   1,   0,  -2,   0,  -3, -10, -14,  -5,  14,  -2,   4,   1,  -1,   3,   2,   6,   0,  19,  10,  -3,  -5, -11,   2,   2,   0,  -1,  -7,  -6, -16, -81, -20,   9,   5,   4,   6,   2,   0,   4,  -2,  13, -20,  25, -18,  14,  -3,  -2,   8, -16, -19,   8, -13,   7,  10,  -6, -25, -14,   8,   2, -19,  -7,  -4,   2,  21,  26,   6,  21,  12,  11,  -1,  -4,  -2, -13,  -6,   8,  -8,  10,  12,  -3,   2,  -4, -14,   2,  -2,   3,  -6, -10 },
   },
   {
     {  109, -59,   1,  -5,   0,  -2,   0,  -1,   0,  -1,   0,   0, -10,  11,  -4,   2,  -1,   1,   0,   0,   0,   0,   0,   0, -19,  13,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0, -10,   6,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -7,   4,   1,   0,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   -1, -16,   9,   1,   1,   0,   0,   0,   0,   0,   0,   0, 103, -57,  -2,   0,  -2,   0,  -1,   0,  -1,   0,   0,   0,   2,  16, -15,   4,  -2,   1,  -1,   0,   0,   0,   0,   0, -25,  23,  -6,   0,   0,   0,   0,   0,   0,   0,   0,   0, -15,  10,   1,  -2,   1,   0,   0,   0,  -9,   3,   2,  -1,   0,   0,   0,   0,  -4,   2,   1,   0,   0,   0,   0,   0,  -4,   2,   1,   0,   0,   0,   0,   0,  -2,   1,   1,   0,  -2,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   39,  77, -75,  19, -18,   5,  -7,   2,  -3,   1,  -2,   0,  26,  11,  -9,  -7,   2,  -3,   1,  -1,   0,  -1,   0,   0, -22,   0,  20, -11,   6,  -3,   2,  -1,   1,   0,   0,   0, -13,  -4,  13,  -3,   0,   0,   0,   0,   0,   0,   0,   0,  -9,  -5,   7,   0,   0,   0,   0,   0,  -2,  -4,   3,   1,   0,   0,   0,   0,  -2,  -3,   2,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0 },
     {   29,  11, -18,   3,  -5,   1,  -2,   0,  -1,   0,   0,   0, -19, -43,  36,  -5,   7,  -1,   2,   0,   1,   0,   1,   0,  77, -60,   3,   2,  -2,   1,  -1,   1,  -1,   1,   0,   0,  10,  15, -20,   8,  -4,   2,  -2,   1,  -1,   0,   0,   0, -12,  19, -11,   2,   0,   0,   0,   0, -11,  14,  -3,  -2,   1,  -1,   0,   0,  -8,   6,   1,  -1,   1,   0,   0,   0,  -4,   4,   0,   0,   0,   0,   0,   0,  -3,   2,   1,   0,  -2,   2,   0,   0,  -2,   1,   0,   0,  -1,   1,   0,   0 },
     {   -4,  -6,  -8,   6,   1,   1,   0,   1,   0,   0,   0,   0,  16,  63, -62,  15, -10,   3,  -3,   1,  -2,   0,  -1,   0,  72,   9, -24,  -6,  -1,  -3,   0,  -1,   0,  -1,   0,   0,  -4,  21,   5, -11,   3,  -3,   1,  -1,   0,   0,   0,   0, -18,  10,  12,  -7,   2,  -1,   0,   0, -13,   0,  12,  -3,   1,   0,   0,   0,  -7,  -3,   8,  -1,   1,   0,   0,   0,  -4,  -5,   7,  -1,   0,   0,   0,   0,  -2,  -4,   4,   0,  -1,  -3,   3,   0,  -1,  -3,   2,   1,  -1,  -2,   1,   1 },
     {   -1,  24,  -5,  -5,   1,  -3,   1,  -1,   0,  -1,   0,   0, -47, -24,  22,   3,   4,   1,   1,   1,   0,   1,   0,   0,   8,  47, -42,  10,  -6,   1,  -1,   0,  -1,   0,   0,   0, -59,  56,  -4,  -5,   4,  -3,   2,  -2,   1,  -1,   1,  -1,  -6, -13,  17,  -8,   5,  -2,   1,   0,   9, -14,  11,  -3,   0,   0,   0,   0,  11, -15,   6,   0,  -1,   1,   0,   0,   7,  -6,   1,   1,  -1,   1,   0,   0,   5,  -5,   0,   1,   3,  -1,  -1,   0,   3,  -2,   0,   1,   2,   0,  -1,   0 },
     {   -5,  -5, -34,  32, -12,   9,  -4,   4,  -2,   2,  -1,   1, -17, -28, -23,  31,  -5,   7,  -2,   3,   0,   1,   0,   1, -28,  21, -43,  19,  -1,   2,   1,   1,   0,   0,   0,   0,  63,   9, -35,  -5,   4,  -5,   1,  -1,   0,  -1,   0,   0,   7,  33, -10, -15,   5,  -4,   2,  -2,  -5,  16,   9, -13,   3,  -2,   1,  -1,  -7,   5,  12,  -8,   1,  -1,   1,   0,  -1,  -1,   8,  -4,   0,  -1,   0,   0,  -2,  -1,   6,  -2,   0,  -2,   4,  -1,   0,  -1,   3,  -1,   0,  -1,   3,  -1 },
     {   25,  50,  29, -64,  25, -22,   9,  -9,   3,  -4,   1,  -2,  15,  15,  18, -27,   5,  -5,   0,  -2,   0,  -1,   0,  -1,   5,  29, -26,  14, -14,   7,  -5,   3,  -2,   1,  -1,   0,  48,  -8,  -7,   2,  -6,   1,  -1,   0,  -1,   0,   0,   0,   0,  14,   2,  -1,  -2,  -1,  -1,   0, -14,   6,   9,  -1,   0,  -1,   0,   0, -10,  -3,   7,   2,   0,   0,   0,   0,  -5,  -5,   6,   2,   0,   0,   0,   0,  -3,  -6,   4,   2,  -1,  -4,   2,   3,  -1,  -3,   0,   2,  -1,  -2,   0,   2 },
     {  -10,   4,  15,  -8,   1,  -2,   1,  -2,   1,  -1,   0,   0,   3,  35,   2,  -8,  -4,  -2,  -1,  -1,  -1,   0,   0,  -1, -43, -44,  24,  12,   1,   4,   0,   1,   0,   1,   0,   1, -13,  47, -47,  14,  -5,   1,  -1,   0,   0,  -1,   0,   0, -45,  44,  -2,  -6,   7,  -5,   3,  -2,   0, -13,  14,  -9,   6,  -2,   2,  -1,   9, -12,   9,  -4,   1,   0,   0,   1,   9, -12,   3,   1,  -1,   1,   0,   0,   5,  -4,   0,   1,   4,  -4,  -1,   1,   3,  -2,  -1,   0,   2,  -2,  -1,   0 },
     {   16,  32,  18, -32,   9, -11,   4,  -5,   2,  -2,   0,  -2,  -3, -15, -45,  58, -24,  15,  -8,   7,  -2,   3,  -1,   1, -11, -43, -17,  38, -10,   8,  -2,   2,   0,   2,   0,   1, -32, -16,  -4,   9,   7,  -2,   4,  -1,   2,   0,   1,   0,  17, -17, -15,   3,   5,  -2,   1,   0,   9,   5, -12,  -3,   4,  -1,   1,  -1,   6,   8,  -9,  -4,   2,  -1,   1,   0,  -2,  10,  -2,  -5,   3,  -1,   0,   0,   0,   6,   0,  -4,  -1,   3,   2,  -3,   1,   1,   1,  -2,   0,   2,   2,  -2 },
     {   -5,  -3, -10,   8,  -4,   4,  -3,   2,   0,   1,   0,   0,   6,  29,   9, -25,   7,  -7,   2,  -3,   0,  -1,   0,  -1,   5,  19,  -6,   6,  -7,   2,  -4,   2,  -2,   1,  -1,   0, -45, -10, -27,  34,  -8,   6,  -2,   3,   0,   1,   0,   1,  49,  15, -58,   9,   0,  -3,   2,  -2,  -9,  44, -12, -15,   8,  -7,   2,  -2,  -4,   8,  16, -14,   6,  -4,   0,  -1,  -3,  -1,  15,  -6,   1,   0,   0,   0,   1,  -7,  10,  -1,  -1,  -4,   6,   0,   1,  -4,   3,   0,   0,  -3,   2,   1 },
     {   -5,   3,   0, -27,  22, -10,   7,  -3,   3,  -1,   1,  -1, -25, -37, -31,  17,   9,   2,   3,   3,   1,   1,   1,   1,  -4,  15,  32, -45,  18,  -9,   4,  -2,   1,  -1,   0,  -1, -19, -15,  25, -14,   1,   3,  -2,   1,   0,   1,   0,   0, -10,  51, -22,  -1,  -6,   2,  -1,  -1, -34,  22,  23,  -9,   0,  -2,   2,  -2,   8, -18,  12,   2,  -4,   2,  -1,   0,   6,  -8,   7,   2,  -3,   1,   0,   0,   7,  -9,   0,   4,   3,  -2,  -1,   2,   4,  -3,  -2,   3,   1,   1,  -1,   0 },
     {    9,  22,  46,  24, -53,  25, -19,   6,  -8,   2,  -3,   1,  -4,   6,  27,  21, -34,  14,  -8,   2,  -5,   0,  -1,   0,   9,  28,  -3,   6,  -4,  -7,   3,  -5,   1,  -1,   1,  -1,  -4, -33,  13,  11,   3,  -3,   4,   0,   2,   0,   0,   0, -45,  22, -20,  14,   4,  -3,   3,   0, -17,  -4,  -8,   4,   7,  -3,   2,   0,   0,  -9,  -9,   0,   7,  -2,   1,   0,   3,  -5,  -3,  -4,   4,  -1,   0,   0,   3,  -2,  -4,  -1,   2,  -1,  -3,  -1,   2,   0,  -3,  -1,   1,   0,  -1,  -1 },
     {   14,  20,  40,  -3, -34,  15, -13,   4,  -5,   1,  -2,   1,   9,   4,   8,  14, -24,  11,  -5,   2,  -2,   0,  -1,   0,  -1,  -7,  24, -32,  19, -13,   9,  -5,   3,  -1,   1,  -1,  13,  49,  -9, -26,   9, -10,   4,  -3,   1,  -1,   0,  -1,  68,  -5,  11, -13,  -1,  -1,  -3,   1,  -2,  17,  10,   0,  -5,   2,  -1,   0,  -8,  -3,  11,   6,  -2,   2,  -1,   0,  -8,  -3,   2,   6,  -1,   0,   0,   0,  -2,  -5,   1,   5,  -4,  -3,   0,   3,  -1,  -3,  -1,   3,  -2,  -1,  -2,   2 },
     {   10,   7,  18,  16, -17,   5,  -7,   3,  -3,   0,  -2,   0,   3,  -4, -16, -24,  27, -17,  10,  -1,   3,  -1,   1,  -1,   7, -14, -27,   7,  14,  -5,   4,  -2,   1,   0,   0,   0, -15, -39, -19,   8,   5,   6,   2,   4,   0,   1,   0,   1,  35,  26,   7, -28,   2,  -2,  -1,   0,  -8, -47,  48,   1,  -5,   7,  -1,   1,  24, -24, -19,  10,  -9,   5,  -2,   2,  -5,  13, -10,   6,  -4,   1,  -1,   0,   1,   8, -10,   0,  -7,   8,  -1,  -2,   0,   2,  -1,  -1,  -3,   4,   1,   0 },
     {   -4, -12, -29,   0,  14, -10,  15,  -5,   5,  -1,   1,   0,   4,   0,  14,  19, -42,  23, -12,   7,  -6,   3,  -2,   1,  16,  26,  47,  11, -45,  19, -13,   2,  -5,   0,  -2,   0,   7,   3, -11,  29, -16,   0,   3,  -3,   2,  -1,   0,   0,  30,  10,   9,   5,   0,  -9,   1,  -1, -18, -28,  28,  10,   7,  -1,   0,  -1,  13, -20, -14,  10,   3,   1,   0,   1,  -4,   1, -12,   2,   4,  -2,   0,  -1,   0,   3,  -8,  -1,  -5,   3,  -3,  -2,  -1,   1,  -2,  -1,  -2,   1,  -1,  -1 },
     {   -8, -14, -12,   8,   5,  -2,   1,  -1,   2,  -1,   1,   0,   5,  16,  35,  -1, -11,   3,  -3,  -1,  -2,  -1,  -1,   0, -15, -32, -20,  39, -11,   9,  -2,   4,  -1,   1,   0,   1, -16, -16,  35, -32,  20,  -9,   6,  -2,   2,  -1,   0,   0,   9,  -1,  24, -22,   4,   0,   1,  -1, -56,  32,  16,  -8,  -1,   0,   1,  -1, -12, -21,  23,   3,  -1,   1,   0,   0,   5, -17,  -5,   8,  -3,   3,  -1,   1,   9,  -5,  -6,   4,   3,   0,  -8,   4,   3,   2,  -4,   0,   0,   1,  -3,   0 },
     {    0,  -7, -12, -15,   3,  -2,   4,   0,   1,   0,   1,   0,   9,   4,  25,   1, -23,  15,  -8,   1,  -2,   0,  -1,   0,  15,   7,  12,  -1,  -3,   2,   3,  -2,   0,  -1,   0,   0,  -9, -31,   3, -20,  31, -13,   9,  -4,   3,  -2,   2,  -1,  -7,  22,  -5, -43,  24,  -9,   3,  -1,  72,  -3,  15, -32,  -1,   0,  -5,   2,  10,  18,  21, -10, -12,   5,  -4,   1,   0,   2,  14,   6,  -8,   2,  -1,   0,  -7,   1,   4,   8,  -1,  -4,   3,   6,  -2,  -1,  -1,   5,   1,  -2,  -2,   3 },
     {   -4,   3,  -6,   4,   8, -16,  10,  -5,   3,  -1,   1,  -1,  -3,  11,  16,  -3,  -8,   1,   1,   0,  -1,   1,   0,   0,  -5,   5,  -1,   5,   7,  -1,   0,  -2,  -1,  -1,   0,  -1, -12,   8,   3, -32,  24,  -9,   3,   2,   0,  -1,   0,   0,  11,  47,   9, -15,  -9,   5,  -6,   0, -23, -14, -39,  31,  -3,   5,  -3,   3,  17,  41, -52, -13,   9,  -6,   2,   0, -22,  23,  14, -20,   5,  -4,   1,  -2,   7,   0,  14, -10,  -5,  -9,  13,  -1,   4,  -5,   3,   0,  -3,  -4,   4,   1 },
     {   -6, -10, -12, -36, -28,  47, -21,  16,  -7,   7,  -2,   5,  -8, -11, -25, -45,  -8,  19,   0,   9,  -1,   2,   1,   1,  -7, -14, -13,   8, -39,  21,  -5,   5,   0,   2,   2,   0,  -2,  11,  33, -23, -24,   9,  -1,  -5,   5,  -2,   1,  -1,   8,  29,   0,  17,   1, -11,   1,   0,  18,   5,  -3,  15,  -4,  -6,   0,  -2,  12,   1,  -3,  12,   2,  -7,   2,  -3,   2,   8,  -1,   1,   7,  -5,   0,   1,  -8,  -1,   6,   1,   1,   2,   0,   1,  -3,  -3,   4,   1,   1,   2,   0,   0 },
     {   -3, -10, -27, -17,  14,  -2,   8,   0,   1,   0,   1,   0,  16,  21,  31,  16, -23,   1,  -3,  -2,   0,  -1,  -1,   0,  -2, -13, -39, -12,  36, -17,  14,  -8,   8,  -3,   3,   0,  12,  14,  14, -15,  -1,  -2,   0,   0,  -2,   1,  -1,   0,   5,   6, -47,  33,  -9,   2,  -4,   4,  15,  -9,   3,   1,   6,  -3,  -1,  -1,  29, -44,  -9,  13,  -1,   5,  -2,   3,   8,  19, -21,  -5,   1,  -2,  -1,   0,  -2,   7,   0,  -4,  -3,   7,  -3,  -5,  -2,   2,   4,  -2,  -3,   2,   0,  -3 },
     {  -10, -14, -26, -52,  -7,  34, -10,  12,  -2,   5,  -1,   4,  -4,  -5,  -3, -19, -21,  22,  -6,   6,  -2,   1,  -1,   1,   0,   3,  -6, -21,  19, -16,  14,  -6,   5,  -2,   1,  -1,  -5, -21, -53,  10,  26, -15,  12,  -2,   5,  -1,   2,   0, -14, -35,  12,  -6,  13,  -1,   2,   2, -41,  -6,  -8,   5,   5,   3,   0,   1,   5,   6,  -3,  -9,  -4,   3,  -1,   1,   0,   1,   2,  -1,  -3,   2,   0,   1,   8,   3,  -2,  -3,   0,   5,   4,  -1,   3,   2,   2,  -1,   0,   1,   2,   0 },
     {    6,   4,   8,   6,   3,  -1,  -8,   4,  -3,   0,   1,   1,   2,  -5, -19, -13,  26, -15,   4,   1,   5,  -4,   0,   0,  11,  12,  32,  24, -18,  -6,  -3,   1,   0,  -1,  -1,   0,  16,  14,   3,  -7,  44, -29,   3,  -8,   5,  -5,   3,  -2,   3,  -5, -17,  -7,  45, -12,   2,   0,  -4,  12, -31,   4,   3,   9,  -8,   7,  25, -46,  -2,  -8,   2,   4,  -4,   3,  30,  10, -20, -16,   6,   2,  -1,   3,  -6,  11,   5, -14,   2,   6,   5,   1,  -4,   2,   7,  -3,   0,   1,   6,   2 },
     {   10,   7,  17,  31,   7, -19,   5, -10,   1,  -3,   1,  -4,   3,  -8,  -7, -26, -13,  34,  -6,   6,  -1,   2,  -1,   0,   0, -14, -17, -34,  -3,  34, -13,   6,  -2,   3,  -2,   2,  12,  -1,   4,  -8,  -7,   7,  -1,  -3,   1,  -1,   2,  -1,   0, -21,   2,   2,   1,  -2,   4,  -3, -13,  -9, -13, -21,  22, -10,   8,   1,  63,  -6,  17, -35,   9,  -5,  -2,   3,   5,  -6,  22,  -8,  -9,   4,  -3,   0,  17,  -4,   5,   4,  -8,   0,  -1,   6,   5,   1,  -4,   3,  -4,   2,   0,   2 },
     {   -1,  -9, -11, -29, -13,  20, -13,  13,  -4,   4,  -1,   2,   9,   1,   9,  34,  22, -29,   7, -13,   5,  -6,   2,  -2,  10,   3,  12,  33,  14, -17,   8,  -6,   0,  -3,  -1,  -1,  11,   5,  23,  15, -11,   1,  -5,   1,  -5,   0,  -2,  -1,  18,  10,  14,  14,  -3,  -3,   1,  -3,  -5, -28, -14,  -3,  13,  -2,   3,   0,  41,  15,  21, -33,   5,  -2,  -2,  -1,   6, -34,  33,   7,  -4,   8,  -3,   2,  12,  -8, -12,   7,  -7,   2,  -2,   6,   0,   4, -11,   1,  -2,   2,   1,   0 },
     {    4,   3,  -2,  10,  13, -13,  -3,  -1,   0,   0,   1,   1,   0,  -7, -11, -16,  -8,  17, -15,   3,  -1,  -1,  -1,   1,   5,  -1,  -9,  -3,   7,   5,  -2,   8,  -1,   1,  -2,   1,  -8, -14, -20,  -9,  21, -20,   7,   1,   4,  -2,   2,  -1,  15,  28,   1,  18, -28,  11,  -3,  -4,  21, -22, -38,  40,  -6,   6,   1,   0, -23, -16,  29,   1,   6,  -4,  -2,  -1,  16, -53,  -7,  21,   7,   4,   0,   3,   4,   2, -30,  -1,   1,   8,  -8,   0,  -7,   9,  -8,  -8,   3,   1,   4,  -5 },
     {    2,   4,  -1,  -1,  -2,  12,  -3,  -3,   2,   0,  -1,  -1,   5,   9,   5,  19,   3,   3,   7, -12,   3,  -2,  -1,   2,   5,  12,   6,   4,  -8,   0, -13,  11,  -6,   1,  -1,  -1,  -2,  -9, -28,  -7,  -3,  26, -19,  11, -10,   5,   1,   2, -11,   0,  13, -21, -27,  25,  -8,   2,   0,  33, -24,  13, -31,  -2,   3,  -2,  59,  -2,  20,  33, -26, -12,  -1,  -8,   6,  13, -18,  31,  -3,  -5,   3,  -5,  -3,  13,   1,  -5,  -2,  -6,   2,   6,  -3,   2,  -1,   0,  -2,  -3,   0,   3 },
     {    0,  -2,  -3,  -9,   0,   6,  -6,   5,  -6,   4,   0,   0,   8,   8,  14,  34,  15, -25,  11, -11,   6,  -8,   1,  -4,   2,   0,  -2, -29, -19,  40,  -9,   5,   4,   0,  -2,   3,  -4, -12, -28, -49,  -9,  35,   0,   3,  -1,   6,   0,   0,   3, -10, -22, -13,  30, -21,  18,  -8,  -6, -22, -13,  16,   9,  -2,  10,   0, -16,  -4,  -6,  15,   7,  -9,   1,   0,  15,  -9,   9,  -5,  -4,  -2,   0,   1,   2, -24,  -4,   1,   6,   9, -10,  -9,  -1,  -3,  -3,  -3,   5,   4,  -2,  -8 },
     {   -2,  -3,  -4, -14, -13,  -6,   6,  -2,   8,  -3,   2,  -3,   2,   2,   7,  10, -10,  -1,  14, -13,   7,  -4,   0,   0,  -2,  -5, -20, -32,  16,   3,  -5,   8,  -6,   5,  -2,   2,  -5,  -8,  12,  36, -17,  10, -13,   4,  -4,  -1,   0,   0,  13,  24,  48, -23,  17, -18,   7, -10,   9,  17, -23,  -5,  24, -12,   0,   5, -12, -11,  -9,  -4,  21,  -7,  -1,   4,  32, -11, -27, -14,   6,   3,   0,   1,  -7,  32,   5, -23,   3, -10,  15,   2,  -7,   5,   1,  -1,   4,  -5,   7,   4 },
     {    5,   7,   4,  12,  52,  28, -36,  11, -13,   0,  -2,   0,   6,  11,  12,  11,  35,  39, -42,  15, -11,  -1,  -1,   2,  -1,  -1,  -8,  -5,  20,   4,  -6,   1,   0,  -2,  -1,   0,   4,   9,  17,  30,  -2,   5,   5, -15,   1,  -2,   2,  -2,   4,   9,   0, -14,  24, -15,  12,  -9,  -1,   8,  -5,  -1, -12,   0,   2,  -2,   6,  16, -23,  21,  -8,  -3,  -1,  -3,   0,  -8,  -6,   8,   1,   4,   2,  -5,   9, -12, -13,   4,   1,   5,  -2,  -9,   3,  -1,   1,  -3,  -1,   0,   1,  -5 },
     {   -2,   1,  -1,  -5,  -7,  -4,  -6,   1,   7,  -4,   1,  -2,  -3,   2,   1,   2,   0,  -2,  17, -12,   9,  -5,  -2,   4,  -6,  -2,  -7, -18, -14,  11,  -7,  18, -12,   6,  -1,   3,   2,  14,  16,   1,  15, -13, -10,   1,  -5,  -2,   3,   0,   0,   0, -29,   6,  17,  -5,  -5,  -3, -21, -32,  -6, -32,  -5,  -3,  -6,  14,   0,  33,  18,   6, -23,   4,  -5,  -2, -44, -21, -41,  24,  12,   4,   3,   1,   5,  19, -18, -31,  -2,   1,  16,  -6,   7,   6,  12,  -1,   3,  -7,   9,   7 },
     {   -3,  -2,  -5,  -2, -20,  -4,  26, -12,  17, -15,   6,  -1,   0,   4,  -1,   6,  -1, -14,  13,  -1,  -5,   4,  -3,   0,  -9,  -9, -21, -28, -19,   1,   1,  14,  -3,   1,   4,  -1,   4,   7,  10,  32,   5, -21,  13,  -1,   9,  -4,  -4,   2,  -6,  -1,  -9, -10,  27, -10, -11,  13,   1,  22,  -4,  29, -19,   6, -22,   3,  17,  -1, -26,  13, -11,  13,  -5,  -4, -12, -21,  30,  18,  -5,  18,  -7,   0, -16, -12, -13,  26, -19,   6, -27,   7, -12,  11, -10,  -9,  -4,   9,  -1,  -3 },
   }
 },
 {  //25
   {
     { -112,  45,   6,   4,   2,   2,   1,   1,   1,   0,   0,   0,   0, -18,   9,  -1,   2,   0,   1,   0,   0,   0,   0,   0,  28, -20,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  10,  -1,  -6,   0,   0,   0,   0,   0,   0,   0,   0,   0,   6,   0,  -2,  -1,   0,   0,   0,   0,   3,   0,   0,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   15,  24, -12,  -1,  -3,   0,  -1,   0,  -1,   0,   0,   0, -98,  39,  17,  -2,   4,   0,   2,   0,   1,   0,   0,   0,  -1, -42,  23,   0,   3,   0,   1,   0,   1,   0,   0,   0,  28, -23,  -8,   6,  -2,   1,  -1,   1,   0,   0,   0,   0,   9,   0,  -9,   0,   1,   0,   0,   0,   6,   1,  -3,  -2,   0,   0,   0,   0,   2,   1,  -1,   0,  -1,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   25, -24,   5,   1,   1,   0,   0,   0,   0,   0,   0,   0, -20, -51,  23,   5,   4,   2,   2,   1,   1,   0,   1,   0,  64, -40, -32,  13,  -4,   4,  -2,   2,  -1,   1,   0,   1,   1,  48, -36,  -3,  -1,  -1,   0,  -1,   0,   0,   0,   0, -13,  23,   6, -12,   2,  -2,   1,  -1,  -5,   5,   7,  -2,  -2,   0,  -1,   0,  -5,   2,   3,   1,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,  -2,   1,   1,   0,  -1,   1,   1,   0,  -1,   1,   1,   0,  -1,   1,   0,   0 },
     {   30,  84, -45,  -3,  -8,  -2,  -3,  -1,  -2,  -1,  -1,  -1,   5,  11,   7, -10,   1,  -2,   0,  -1,   0,  -1,   0,   0,  29,   8,  -8,  -6,  -2,  -1,  -1,  -1,   0,   0,   0,   0, -35,  33,  22, -12,   1,  -3,   0,  -1,   0,  -1,   0,  -1,  -8, -25,  30,   4,  -1,   1,   0,   0,   0, -12,   1,  10,  -1,   1,   0,   0,  -1,  -8,   1,   3,   2,   0,   0,   0,   0,  -4,  -1,   1,   1,   1,   0,   0,   0,  -4,   0,   1,   0,  -2,  -1,   1,   0,  -2,   0,   1,   0,  -1,  -1,   1 },
     {  -23, -34,  21,   1,   5,   1,   2,   0,   1,   0,   1,   0, -64,  -2,  -9,  13,   1,   4,   1,   2,   1,   1,   0,   1,   4,  53, -40,  -8,   2,  -3,   1,  -1,   1,  -1,   0,  -1, -12,  25,  32, -29,   5,  -5,   1,  -2,   0,  -1,   0,  -1,  17, -31,  30,   2,  -7,   3,  -2,   1,   9,  -8,  -4,  13,  -3,   0,   0,   0,   4,  -5,  -4,   3,   3,   0,   0,   0,   4,  -1,  -3,  -1,   0,   1,   0,   0,   3,  -1,  -1,   0,   2,   0,  -1,   0,   1,  -1,   0,   0,   1,   0,   0,   0 },
     {   18,  15,  16,  -8,  -1,  -3,  -1,  -2,  -1,  -1,   0,  -1,  -7, -66,  45,   5,   2,   2,   0,   1,   0,   0,   0,   0,  -5,  -1,   0,   8,  -3,   1,  -1,   0,   0,   0,   0,   0, -20, -31,  30,  15,  -2,   3,  -1,   2,   0,   1,   0,   1,  19, -36, -33,  34,  -2,   5,   0,   2,  -3,  21, -39,   0,   9,  -2,   2,  -1,   1,   9,  -9, -12,   4,   1,   0,   0,  -1,  10,  -4,  -6,  -1,   0,   0,   0,   0,   3,  -1,  -3,  -1,   4,  -1,  -2,   0,   1,   0,  -2,   0,   2,   0,  -1 },
     {   -2,  22, -11,   3,  -1,   0,  -1,  -1,   0,   0,   0,   0,  -9, -34,  18,   6,   4,   2,   0,   1,   0,   0,   0,   0, -87, -10, -23,  26,   5,   6,   2,   2,   1,   1,   0,   1, -18,  14, -38, -14,  14,  -2,   5,  -1,   2,   0,   1,   0,  13,   0,  14, -36,   4,   0,   1,   0,  17,  -8,  14,  -7, -14,   3,  -2,   1,   9,   2,   2,   5,  -8,  -2,   0,   0,   5,   3,   0,   2,  -1,  -2,   0,   0,   3,   1,   1,   0,   2,   0,   1,   0,   2,   0,   1,   0,   1,   0,   0,   0 },
     {    8,  24, -32,   6,  -1,   1,   0,   0,   0,   0,   0,   0, -19, -30, -40,  22,   5,   6,   2,   3,   1,   1,   1,   1,  11,  38, -29, -19,   8,  -5,   1,  -2,   1,  -1,   0,  -1,  14, -22,   8,  -2,   3,   1,   0,   1,   0,   0,   0,   0,   6,  26, -44, -13,   8,  -1,   2,  -1, -11,  26,  27, -43,   2,   0,   0,  -1,   4, -10,  28,  -2, -17,   3,  -2,   0,  -2,   0,   9,   8,  -5,  -3,   1,  -1,   2,  -5,   5,   5,  -1,   0,   2,   2,   1,  -2,   2,   2,   0,   0,   1,   1 },
     {    8,   4, -32,   8,  -2,   3,   0,   1,   0,   1,   0,   1,  15, -24, -12,   0,   2,   2,   1,   1,   0,   0,   0,   0, -18, -18, -16, -12,   9,   0,   3,   1,   2,   0,   1,   0,  77,  -2,  15, -41,  -9,  -1,  -3,   0,  -2,   0,  -1,   0,  17,   7,  30,  19, -31,   4,  -6,   1, -12,  16, -16,  34,  -2,  -8,   2,  -2, -12,  12,  -3,  -1,  15,  -5,   1,  -1,  -4,   1,   2,  -4,   5,   2,  -1,   0,  -4,   2,   3,  -1,  -2,   0,   1,   0,  -2,   0,   1,   0,  -1,   0,   0,   0 },
     {  -17, -32, -59,  40,   1,   9,   0,   4,   0,   2,   1,   1,   0,  16, -29,   9,   5,   0,   0,   0,   0,   0,   0,   0,  -3, -27, -10,  17,   5,   3,   1,   1,   0,   1,   0,   0, -28,  15,  -9,   1,   8,  -4,   3,  -1,   1,  -1,   1,   0,  -1, -28,  -9,  15,  11,   0,   2,   0,   4,  13, -44, -17,  20,  -2,   3,   0,   1,  13,  10, -38,   2,   4,  -1,   0,   5,   3,   9,  -9, -12,   2,   0,   0,   1,   3,   9,  -4,   2,   0,   4,  -1,   0,   0,   5,  -2,   1,  -1,   2,   0 },
     {   -8, -31, -52,  19,   3,   8,   2,   3,   1,   2,   0,   1, -12, -42, -10, -17,  14,   2,   5,   1,   3,   0,   2,   0,   8,  -5,  51, -34,  -4,   0,  -2,   0,  -1,  -1,  -1,   0, -31, -15,  14,  11,  -9,   4,  -1,   2,   0,   1,   0,   1,  14,   4,   8, -16,  -7,  -1,  -1,   0,  13, -13,  24,  18, -22,   3,  -4,   1,   4,   1, -21,  30,  -2,  -5,   0,  -1,  -2,   9, -11,   1,  10,  -3,   1,   0,   0,   4,  -5,  -1,   0,   3,  -1,  -2,   0,   2,  -1,  -1,   0,   2,   1,   0 },
     {  -14, -18, -23,  -3,   6,   3,   2,   1,   2,   1,   1,   0,  -5,  -4,  51, -34,   1,  -3,  -1,  -1,   0,   0,   0,  -1, -28,  10,  15, -16,  -3,   3,  -1,   1,   0,   0,   0,   0,  19,  45,  24,  -2, -23,  -2,  -5,  -1,  -2,  -1,  -2,  -1, -24,  15,  -5,  35,   8,  -7,   2,  -3,  -5,  -2,   5, -29,  22,   0,   1,   0,  12, -29,  29,   0, -17,  12,  -4,   3,   6,  -6, -12,  19,  -9,  -1,   2,  -1,   1,  -2,  -3,   6,   2,  -1,  -6,   3,   0,   1,  -2,   1,   1,   0,  -3,   1 },
     {   -1, -17,  -8, -16,   8,   1,   3,   0,   1,   1,   1,   0,  18,  29,  29, -42,   3,  -3,   0,  -1,   0,  -1,   0,  -1,  19,  -7, -11,   2,  -3,   4,  -1,   2,  -1,   0,   0,   1, -10, -10,  -5, -23,   9,   2,   1,   0,   2,   0,   1,   0,  58, -25,  -3, -23, -13,   9,  -3,   3,   0,  45,   3, -15, -28,   2,  -2,  -1,  -9,   7,  35,  15, -17,  -7,  -1,  -2,  -5,  -3,   2,  20,   0,  -5,  -1,   0,  -3,  -2,   3,   8,  -3,  -1,  -3,   7,  -2,   0,   0,   3,  -2,   0,  -3,   3 },
     {    3,  -2,  22, -22,  13,  -4,   3,  -3,   1,  -1,   0,  -1,   8,  -4, -26,   9,   9,   2,   3,   1,   1,   0,   1,   0,  -4, -57, -15, -25,  27,   4,   7,   2,   2,   1,   1,   0,  -9, -12,  35, -33,  -6,   8,  -3,   3,  -1,   1,  -1,   1, -42, -21,   4,   4,  -1,   9,   1,   3,  25, -15,  -3, -31, -16,  10,  -2,   3,  10,  12,   4,  15, -35,  -1,  -1,   0,   2,  11,  -6,  14,  -2, -11,   2,  -1,   0,   5,   0,   3,   1,   3,  -2,   2,  -1,   3,   1,   1,   0,   1,  -1,   1 },
     {   -3,   8,   6, -68,  23,  -1,   8,   1,   1,  -1,   1,  -1, -16, -16, -46, -46,  16,  11,   6,   4,   2,   3,   2,   1, -13,  -1,   6, -16, -23,   9,  -1,   2,  -1,   2,  -1,   1, -16,  14, -15,  14, -19,  -2,   2,   1,   0,   1,   0,   0,   2,  12,   4,   4, -10, -11,   1,  -1,   4,  11, -15,  11,  10, -10,   0,  -2,   0,  18,   7, -27,  10,  -2,  -2,  -1,   7,  -8,  28,  -1, -10,   5,  -3,   1,   3,  -3,   4,   8,   1,  -2,   4,   3,   0,   0,   1,   4,   1,  -1,   2,   3 },
     {    7,   1, -11,  -8,   9,   0,   1,  -2,   0,   0,   0,   0,   6, -12,   1, -13,   7,   2,   2,   0,   1,   0,   0,   1,  13,  29,  10,  -1, -14,   0,  -3,   0,  -2,   0,  -1,   0,  59,  10, -15,  24,   6, -11,   1,  -4,   1,  -2,   0,  -1,   3, -43,   8, -18,  41,   5,   2,   1,  23, -34, -18, -22, -10,  27,  -4,   7, -10,  28, -15,  -4, -27,   2,   5,  -1, -11,  10,   9,   1,  -5, -10,   2,  -1,  -5,   3,   3,   1,  -3,   0,   1,   1,  -2,   1,   1,   0,  -1,   0,   2,   0 },
     {   -8,  -4, -20, -27,   3,   5,   3,   3,   1,   1,   1,   1,   2,  17,  -5,  14, -20,   8,  -4,   1,  -2,   1,  -1,   0, -13, -12, -33,   6, -16,   5,   2,   3,   1,   1,   1,   1,   5,  27,  16,  31, -27, -11,  -2,  -4,  -1,  -1,   0,  -1,   0,  -9, -31,  17,  19,  -9,   5,  -2,  19,  21,  12,  -2, -20,  -7,  -3,  -2, -14,   9, -13,  49,  22, -19,   4,  -4,  -3,   4, -11, -28,  36,   0,  -1,   1,   0,  -4,  12, -11,   0,  -1,   0,  -6,   0,  -2,   5,  -1,   1,  -1,   0,  -3 },
     {  -10, -17, -46, -43,  35,   0,  11,   0,   3,   0,   3,   1,  -2,   5,  21, -16,  -6,   6,  -2,   1,  -1,   1,  -1,   0,  12,  12, -29,  45,  -3,  -3,   1,  -2,   0,  -1,   0,  -1,   0, -50,   4,  -9,  22,  -1,   2,   2,   1,   1,   1,   0, -36,  11,   7,  -3,  -6,   7,  -2,   1, -20, -19,   5,   6,   2,   0,   3,   0,  14, -10, -20,  -3,  -1,   3,   0,   2,   0,  16,  -8, -11,  -2,  -2,   1,  -1,   5,   2,   7,  -4,   1,   3,   2,   0,   2,   1,   2,  -1,   0,   2,   1,  -1 },
     {    6,  -1,   4, -18,   1,   0,   2,   2,   1,   0,   0,   0,   6,  -8, -36,  10, -13,   6,   2,   4,   0,   0,   0,   1,   4, -31,   3,  38, -23,  -4,   3,   2,   1,   0,   0,   0,   6,  -2,  19,  11,   2, -16,   1,  -2,   1,  -1,   0,  -1,  55,  -9,  23,  13,  14, -18,  -4,  -2,  -5, -14,  25,  -7,  34,   6,  -8,   2,   9, -32,   6,   4,  -9,  25,  -7,   6,  -7,   6, -30,  20,  -8,   4,   4,   0,  -5,  10, -13,  -6,  -4,   5,  -3,  -3,  -3,   4,  -1,  -3,  -1,   2,  -1,  -1 },
     {   -5,  -1,   0,  -1,  -6,   5,   1,   2,   0,   0,  -1,   0, -12, -19, -16,  13, -22,  12,   2,   5,   0,   2,   0,   1,  -4,   2,  37,  37, -30,  -8,  -2,  -1,   0,  -1,   0,   0,   6,   0,  15,   4,  22, -13,   0,  -1,   0,  -1,   0,   0, -47, -15,  20, -11,   3,  14,  -2,   3, -16,  34, -11,   4, -24,   3,   5,  -4, -17,   3,  38,   9,  -2, -19,   5,  -3,   9, -34,  -4,  25,  14,  -2,  -2,   3,   3,  -2, -17,  -2,   4,  -3,  -4,   0,   1,   3,  -5,  -2,   2,   0,  -2,  -1 },
     {   -3,  -9, -18, -39,  -9,  13,   1,   3,   2,   3,   2,   0,   7,   1,  25,  50, -43,   1,  -4,   0,  -1,   1,  -3,   0,   7,  -2,  29,  -2,   4, -20,   4,  -5,   1,  -1,   0,  -1,  -2,  12,  23, -19,  13,  -9,  -2,  -1,  -1,  -1,   0,   1,  15,  24,  -2, -15,  -1,  -3,  -3,   1,  44,   1,   3,  -2,   6,  -3,  -6,  -1,   6,  21,  -3, -18,   5,   9,  -4,   1,  -1,  -1,  34,  -5, -18,   7,  -3,  -1,  -2, -10,   4,  25,  -4,  -2,  -5,   7,  -3,   0,  -6,   7,  -2,   1,  -3,   1 },
     {    1,   4,  -1,  13, -24,  15,  -3,   3,  -3,   1,   0,   1,  -5,   2,  -1,  -5, -41,  27,  -4,   3,  -2,   0,  -1,   1,   7,  16,  -7,  15, -21,   9,  -1,  -2,   0,  -1,   1,   0, -12, -14, -20, -11, -32,  13,   5,   3,   1,   1,   1,   0,   9,  24,   8,  29, -26, -17,   8,  -5,   9, -26, -20,  -9,   5,  -2,   7,   1,  16,  33,   5,   9, -38, -20,   2,  -3,  -8,   8,   6,  27,  26, -25,   0,  -4,   0,  -1,   0, -10,  -1,  -1,   6,   6,   0,  -1,  -1,  -2,  -1,   0,   2,   2 },
     {   -3,   2,   7,   1,  13,  -4,  -2,  -1,  -1,   1,  -1,   0,  -5,   5,  -3,   7,   7,  -2,  -7,   1,  -1,   1,   0,   1, -11, -13,  13,   9,   1,   3,  -7,  -1,   0,  -1,   0,   2, -12,  25,  33,  -4,   3,  -6,  -3,  -4,   0,  -1,   1,  -1,  26,  32,   2,  -1,  22,  13,  -9,   0, -70,  -2,  -9,  -6, -12,  34,   6,   1,  -6,  13, -26,  -1, -29, -12,  18,  -4,   2,  12,  23, -13,   7, -19,   1,  -1,  11, -18,  20,   4,   6,  -3,  -1,   8,   1,  -4,  -2,   4,   0,   0,  -3,   1 },
     {    1,   5,  -8, -23,  49, -22,   3,  -7,   2,  -2,   1,  -1,  -1,  -1,  -6,  31,  19, -22,   1,  -3,   1,  -1,   0,  -1,   7,  20,  35,  15,  30, -10, -10,  -5,  -2,  -1,  -1,  -1,  -4,  19, -28, -27,   0,  23,  -1,   5,  -1,   1,   0,   1,  18,  10, -10,  20,  -7,   5,  -1,  -3,  12,  -5, -20,  -9,   8,   7,   4,   2, -14,   4,  11,  12,  -3,  -5,   0,  -2,  -9, -17, -29,   2,  28,   5,   1,   0,   4,   5,  -6, -31,   1,  -2,   9,  -5,   3,   0,   5,  -4,   1,  -2,   3,  -1 },
     {    6,  10,  18,  27,  -1, -15,   6,  -5,  -1,  -2,  -1,  -2,  -4,  -5, -21, -52,  -6,  11,   5,   4,   2,   2,   1,   2,   2,   3,  16,  33,  -5, -20,  -3,  -3,   0,  -3,   0,  -1,   7,  23,  23, -19,  39,  -4,  -6,   0,  -3,   0,  -1,   0,  -6,  20, -24,  -2, -13,  10,  -1,   3,  45,  16,  -2,  -3,   7,  -7,   1,  -3,   8,  -5, -14,   3,  -1,   5,   2,   3,  -9,  30, -11, -18,   8,  -6,   3,  -2,  -4,  -5,  29,  -9,   1,  -7,  -1,  12,  -2,  -2,   3,   1,  -1,  -2,  -1,   3 },
     {   -4,  -5,  -3, -38, -46,  30,   2,   4,   1,   3,   1,   1,  -2,  -4, -11,   0, -25,  13,   5,   1,   2,   1,   0,   0,  -3,   1,  20, -23,  12, -16,   4,   1,   2,  -1,   1,   0,   2,  21, -38, -20,  24,   2,  -6,   2,  -1,  -1,  -1,   1,  -1, -35, -27,  13, -13,  20,   0,   2, -36,  -9,   7,   2,  -6,  -4,  10,  -3,  -2, -24,  -8,  -1,  -4,   0,   5,   2,   3,  24, -15,  -5,  -4,  -8,   1,  -1,   4,   9,  15, -12,   4,   2,  10,   0,  -1,   1,   6,   4,   2,  -1,  -1,   3 },
     {   -3,  -3,   0,  -9, -36,   8,  14,   2,  -3,   2,   1,   0,  -6,  -7,   0, -35, -15, -21,  18,   0,   4,   0,   1,  -1,  -2,   0, -17, -10,  46, -34,   4,  -3,   4,  -2,   1,  -1, -13, -19,  -3,   7,  17, -22,  -8,   4,  -1,  -1,  -2,   0,   0,  22,  22,  18,  37, -12, -26,  -1,   4,   6,  -3,   1,  14,  28, -15,   0, -18,  26,   9,   6,   1,   5,   3,  -7,  -3, -19,  -6,  -4,  17,  10,   5,   1,   4,  -4,  -8, -15,   3,   2,   2,  -9,   2,  -1,   1,  -4,   2,  -2,   3,  -1 },
     {   -5,   0,  -9, -13, -23,  14,  -9,   9,   2,   3,   2,   4,   0,  10,  -1,  10,  27,   9, -12,  -1,  -7,  -1,  -1,   1,  -5,   4, -18,  10,  -2,   4, -16,   7,  -1,   1,   1,   2,  -3,   6,   8,  31,   4,  32, -23,  -2,  -4,  -2,   0,  -2,  -2,  22,  -1,   7, -19,  35,  -5,  -7,  29,  14,  -3,  24,  -3,  23,   5,  -7, -36,  -5,  -9,  -4, -21,  27,  24,  -1,  -8,   0, -11,  20, -27, -23,  23,  -3,   0,   1, -14,  -1,   3,   3,   3, -10,   2,  -3,   3,  -3,   4,   0,   0,   0 },
     {   -2,   2,   2,  11,  18, -15,   6,  -5,   4,  -1,   1,  -3,  -4,   0,   7,   8, -17,  12,   0,   0,   3,   1,   0,   2,  -4,   3,   9, -15, -26,   4,  11,  -1,   3,   0,   0,   1, -10, -12, -21, -30, -10,  -3,  13,   3,   2,  -1,   1,   1, -18, -16,   6,   6,  -2, -35,   2,   5,  -2,  28,  19,   3,  26,  14, -27,   2, -60,   4, -17,   6, -17,  35,   8,  -1,   0,   3,  -4, -14, -12, -17,  16,  -4,  13, -10,  18,   0,   4,  -3,  -4,   7,   2,   6,   3,   5,  -2,   0,  -3,   5 },
     {    0,   2,   0,  -9, -13, -11,  -1,   6,   0,   3,   4,   2,   6,  13,  12,  22,  35,   7, -14,  -8,  -6,  -4,  -4,  -1,   1,   4,   1, -31, -43,  -5,  16,  -4,   5,   2,   3,   3,  -3,  -1,  -6, -12,  12, -36,   0,   4,   1,  -3,   0,  -1,  -8,  20,   7,  -9,  15,  -6, -21,   0,   2,  29, -25,   9,  -2,   4, -11,   1,  44,   9,  -3,   1,  -5,  10,  -7,  -5,   3,  22, -23,  14,   8,  -3,   4,  -7,   1,  19,   3, -28,   2,  -2,  26, -10,  -2,  -4,   8,   2,   1,  -5,   5,  -1 },
     {   -5,  -5,   1, -18, -37,   8,   3,   4,   3,   2,  -3,   2,  -2,  -2,   0,  14,  37, -25,   2,  -3,   6,  -2,  -2,  -3,   1,   9,  16,  26,  22,  -3, -12,  -3,  -6,  -1,  -1,  -1,   5,  -4,   2,  15, -25,  20,   6,  -2,   0,   0,   0,  -1, -15, -16,  40, -12, -10, -23,  15,   1,  -4,  35,   8, -16,  17,  -9,  -5,   4,  18,  -1,   1,   2,  -4,  -4,   0,   1, -29,  27,   4, -10,  11, -10,  -2,   3,  -4, -19,  30,  -3,  -2, -18,  -5,  16,  -3,  -3,  -3,   1,   0,  -5,  -2,   2 },
     {   -2,  -1,  -8,   7,   8, -61,  28,   0,   7,   3,  -1,   0,  -3,   1,  -6,   9, -29, -33,  12,   9,   0,   3,   2,   2,  -6,  -1,  -7, -11, -14, -26,  -6,  19,  -6,   7,   1,   1,  -1,   5, -23,  20,  -6,   5, -35,   5,   1,   3,   2,   0,  -5, -17,   7,  15, -19,  23, -26,  -3,  -7,  16,  12,   1,   4,   5,   2,  -7,  27,   6,  -9,   5,  -8,  -1,   2,  -2,  -7,  21,  16,  13,  -1,  -3,   1,  -6,   1,  -5, -13,  17,  -5,   1, -15,   3,  -4,  -3,  -4,  -5,  -1,   3,   2,  -3 },
   },
   {
     { -112,  47,   5,   5,   2,   2,   1,   1,   0,   1,   0,   0,  16, -18,   4,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  27, -10,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   7,  -5,   0,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    5,  -9,   3,  -1,   1,   0,   0,   0,   0,   0,   0,   0, 105, -32, -12,  -3,  -3,  -1,  -1,  -1,  -1,   0,   0,   0, -26,  39,  -8,   1,  -2,   0,  -1,   0,   0,   0,   0,   0, -38,  13,   8,   0,   2,   0,   1,   0,   0,   0,   0,   0,  -7,  -3,   3,   0,   1,   0,   0,   0,  -6,  -1,   2,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,  -1,  -1,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {  -22,   5,   5,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,  18,  -8,   1,  -2,   0,  -1,   0,   0,   0,   0,   0, -93,  19,  21,   0,   5,   0,   2,   0,   1,   0,   1,   0,  38, -57,   7,   2,   1,   1,   1,   1,   0,   0,   0,   0,  29,   4, -18,   1,  -3,   0,  -1,   0,  -1,   5,   1,  -2,   0,  -1,   0,   0,   4,  -1,  -2,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   3,   1,  -2,   0,   1,   1,   0,  -1,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   -1,  18,  -7,   0,  -2,   0,  -1,   0,   0,   0,   0,   0, -36,  12,   7,   0,   1,   0,   0,   0,   0,   0,   0,   0, -15,  30,  -5,  -1,   0,  -1,   0,  -1,   0,   0,   0,   0, -72,   0,  35,  -3,   6,  -1,   2,  -1,   1,   0,   1,   0,  28, -69,   9,   7,   1,   2,   0,   1,  24,  -3, -21,   3,  -2,   1,  -1,   0,   6,   1,  -3,  -1,   0,   0,   0,   0,   5,   2,  -5,   0,   0,   0,   0,   0,   2,   0,  -1,  -1,   3,   1,  -2,  -1,   2,   0,   0,  -1,   2,   0,   0,   0 },
     {   -7,  29, -10,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  -2,  42, -12,  -2,  -3,  -1,  -1,  -1,  -1,   0,   0,   0, -35,   4,   9,   0,   2,   0,   1,   0,   0,   0,   0,   0,   5,  34,  -7,  -6,   0,  -2,   0,  -1,   0,  -1,   0,  -1, -56,   5,  46,  -8,   5,  -2,   2,  -1,  16, -65,  14,  12,   0,   3,   0,   2,  10,  -8, -19,   5,   0,   1,   0,   1,   6,  -3,  -3,  -3,   0,   0,   0,   0,   3,   0,  -2,   0,   3,  -3,  -1,   0,   1,   0,  -2,   0,   2,  -2,   0,  -1 },
     {  -38, -95,  51,   0,  12,   1,   5,   1,   3,   0,   2,   0, -10,  -1,  -7,   4,   0,   1,   0,   1,   0,   0,   0,   0,  -1,   9,  -6,   3,  -1,   0,   0,   0,   0,   0,   0,   0,   9,   7,  -2,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,  -9, -17,  13,   4,   0,   1,   0,   1,  19, -21, -20,   9,  -1,   2,  -1,   2,   2,  20, -17,  -4,   0,  -1,   0,  -1,   0,   5,   0,  -3,   0,   0,   0,   0,   0,   6,  -3,  -2,   1,   2,   1,  -2,   0,   2,   0,  -1,   1,   0,   0,  -1 },
     {   14,  19, -14,   1,  -3,   1,  -1,   0,   0,   0,   0,   0,  26,  -3,   1,  -3,   0,  -1,   0,  -1,   0,   0,   0,   0, -13, -50,  15,   5,   3,   2,   1,   1,   1,   0,   0,   0,  17,   1, -16,   3,  -2,   1,  -1,   0,   0,   0,   0,   0, -10, -30,   4,   8,   0,   2,   0,   1,  35,   7, -56,   4,  -3,   1,  -1,   1, -15,  63,   1, -25,   2,  -6,   1,  -3,  -4,  -5,  30,  -1,  -2,   0,  -1,   0,  -4,   0,   1,   5,   0,  -1,   2,  -1,  -2,   2,   0,   1,   0,  -1,   2,   0 },
     {   18,  11,  -4,  -2,   0,  -1,  -1,  -1,   0,   0,   0,   0,  -8, -66,  25,   5,   5,   2,   2,   1,   1,   0,   1,   1,  -6, -17,   0,   2,   0,   1,   0,   0,   0,   0,   0,   0,   1, -20,  -1,   8,  -1,   2,   0,   1,   0,   0,   0,   0,   9,   2,  -4,   2,   0,   0,   0,   0,   0, -39,   2,  18,  -1,   4,  -1,   2,  16,   1, -65,  10,   1,   1,   1,   1,  -8,  46, -13, -26,   3,  -5,   1,  -2,   1,   5,  14,  -7,  -2,   6,   2,  -1,   0,   1,   3,  -2,  -1,   3,   1,  -2 },
     {   -3,   6,  -3,   1,  -1,   1,  -1,   0,   0,   0,   0,   0, -32, -66,  31,   3,   8,   1,   2,   1,   1,   0,   1,   1, -46,  17, -33,  11,  -3,   4,  -1,   2,   0,   1,   0,   1,  26,  55, -14, -13,  -2,  -4,   0,  -2,   0,  -1,   0,  -1,   1,  -2,  16,  -5,   1,   0,   1,  -1,   8,  15,  -1,  -5,   0,  -1,   0,  -1, -10,  -4,  27,  -2,  -1,  -1,   0,   0,   6, -21,  -1,  13,  -1,   2,   0,   1,   1,   0, -10,   2,   2,  -1,  -2,  -2,   0,   0,  -1,  -2,   2,   0,   0,   0 },
     {   -5, -19,   2,   6,   1,   2,   0,   1,   0,   0,   0,   0,  16,   8,  16,  -8,   2,  -3,   0,  -2,   0,  -1,   0,  -1, -26, -70,  22,  13,   3,   4,   1,   2,   1,   1,   1,   1, -26,  14, -17,   3,  -1,   2,   0,   1,  -1,   1,   0,   0,   3, -12,  -4,   6,   0,   1,   0,   0,   8,  14,   1,  -6,   0,  -1,  -1,  -1,  -1, -36,  14,  23,  -2,   4,  -1,   2,   8,   1, -59,   8,   7,   1,   2,   1,  -5,  27,  -1, -26,   4,  -1,   5,   0,  -1,   5,   1,  -4,   2,   0,   1,  -1 },
     {   11,   9, -10,   1,  -2,   0,   0,   0,   0,   0,   0,   0,  18,   4,   2,  -2,  -2,   0,  -1,   0,   0,   0,   0,   0,  24,  22, -16,  -3,  -1,  -2,   0,   0,   0,   0,   0,   0,  68,  -9,  43, -16,   0,  -5,   0,  -2,   0,  -1,  -1,  -1, -25, -48,  10,  27,  -1,   6,  -1,   3,  -6,   6, -22,   2,   1,   0,   0,   0,  -6, -17,   3,  14,  -1,   2,   0,   1,  -3,   8, -41,   2,   6,   0,   2,   0,  -7,  18,   4, -19,  -1,  -1,   6,   1,  -3,   2,   3,  -2,  -1,  -1,   0,   1 },
     {    9,   9,  25, -10,   1,  -4,   0,  -1,   0,  -1,   0,  -1,  -9, -37,  15,   2,   2,   1,   1,   1,   1,   0,   0,   0,  10,   3,  34,  -9,   0,  -2,   0,  -1,   0,  -1,   0,   0, -14, -46,  15,  18,   0,   4,  -1,   2,   0,   0,   0,   1, -49,   4,  -6,   8,   1,   2,   0,   1,   8, -18,  -2,   1,   1,   1,   0,   0,  13,  -1,  19,   1,  -5,   1,  -2,   0,   6, -38,  -5,  36,  -2,   5,  -1,   2,   1,   8, -52,   5,  -2,  13,  -3, -16,   3,   3,  -4,  -3,  -1,   3,   1,  -2 },
     {   14,   7,  21, -12,   0,  -2,  -1,  -1,   0,  -1,   0,   0,  18,   7,   3,  -6,  -1,  -2,  -1,  -1,   0,   0,   0,   0,  12, -18,  10,  -4,  -1,  -1,   0,   0,   0,   0,   0,   1,  19,  10,   4, -12,   0,  -3,   0,  -1,   0,   0,   0,  -1,  76,  -9,  49,  -6,  -6,  -1,  -3,  -1, -25, -27,   6,  34,  -3,   6,  -1,   3,  -6,  -3,  -6,   3,  -1,   1,   0,   1,  -7, -21,   3,  26,  -4,   3,  -1,   2,  -4,   1, -37,   9,  -7,  10,  -4, -10,  -2,   1,  -4,   0,  -3,   2,   0,  -2 },
     {   12,  30,  86, -46,   6, -12,   0,  -6,   1,  -3,   0,  -3,   5,  15,  30, -10,   2,  -3,  -2,  -2,   0,  -1,   0,  -1, -10, -18, -19,   7,   2,   2,   1,   1,   0,   0,   0,   0,  -2,   3,  11,  -6,   0,   0,   0,   0,   0,   0,   0,   0, -16,  -7,  -4,  10,   0,   2,   0,   1, -16,  10,  -2, -10,   1,  -1,   0,   0,   4,  -9,   3,  -3,  -2,   0,   0,   0,   9,   8,  12, -13,  -5,  -1,  -1,   0,   7, -13,  27,  15,   0,  -8, -14,  17,  -1,   0,  -4,   1,   0,  -1,  -6,   3 },
     {   15,  29,  41, -16,  -1,  -6,  -2,  -4,  -1,  -1,  -1,  -1,  -2,  -4, -33,  10,   0,   3,   0,   2,   0,   1,   1,   1,  13,  24,  -9,  -1,  -2,  -1,  -1,  -1,  -1,   0,   0,   0,   1,  -8, -29,   4,   3,   1,   1,   0,   1,   1,   0,   0,  18,  12,   0, -21,   1,  -4,   0,  -2,  59,  -3,  13,  -6,  -5,   1,  -2,   0, -21,  21,   7,  17,  -4,   3,  -1,   0, -17,  -4, -25,   3,  12,  -1,   3,  -1,  -4,  23,   7, -36,   2,  -9,  35,   6,  -2,  -4,  -5,  10,  -2,  -2,   4,   0 },
     {   -1,  13,  10,  -3,  -1,  -1,   0,  -1,  -1,   0,   0,  -1, -17,  -8, -45,  15,   3,   5,   2,   3,   1,   1,   0,   0, -11,  -5, -15,   6,  -1,   0,   1,   1,   0,   0,   0,   1, -17,   4,  -7,   0,   3,  -2,   1,   0,   1,   0,   0,   0,  -5,  13,   6,  11,   0,   0,  -1,   0, -69,  -2, -45,   7,  14,   2,   5,   0,  15,   4,  -2, -36,   3,  -4,   1,  -3,  10,  -7,  -9,  -2,   2,   0,   0,   0,  10,  29,  -3, -33,   5,  -3,  37,  -3,   5,  -3,   1,   5,   1,  -1,   6,   0 },
     {    3,   2,  10,  -4,   3,  -1,   1,  -2,   1,  -2,   0,   0,  -9, -32, -79,  36,  -3,   9,   1,   5,   0,   2,   0,   1,  -9, -36,  17, -27,   5,  -2,   3,   0,   2,   0,   1,   0,   5,  20,  46, -13,  -6,  -3,  -3,  -2,  -1,   0,  -1,   0,   1,  -7,  -1,   5,   1,   0,  -1,   1,  15,  10,  18, -10,  -2,  -1,  -1,   0,   2, -12,   3,  10,  -2,   1,  -1,   1,  10,   7,   8,  -1,  -3,   0,  -2,  -1,  -3, -10,   5,  19,  -2,   2, -18,   5,  -2,   3,  -2,  -5,   0,   1,  -1,  -1 },
     {    3,   7,   4,   2,  -2,   0,   0,   1,   0,   0,   0,   0,   8,  20, -14,  27,  -8,   4,  -2,   0,  -2,   1,  -1,   0,  -5, -18, -74,  26,   5,   5,   2,   4,   2,   1,   0,   1,  -1, -12, -11,   5,   4,  -1,   1,   0,   0,   0,   1,   0,  -7,  -8, -21,  -2,   7,  -2,   2,   0,  12,   5,   7,   1,  -1,  -1,  -1,   0,  -4,  -9, -21,   5,  11,   1,   2,   1,   5,   1,   1,  -2,  -1,   1,   0,   0, -12,  -6, -42,  10,   0,  26, -12, -51,   4,   0,  28, -11,   3,   3,   4,  -1 },
     {    3,  -2,  -6,   1,   0,   0,   1,  -1,   1,   0,  -1,   0,   6,   4,   0,  -2,  -3,   1,   1,   0,  -1,   0,  -1,   0,   0, -10, -24,   4,  -1,   2,   1,   1,   0,   1,   0,   1,   6,   2,   0, -13,   2,  -1,   1,   0,   0,   1,   0,   0,  12,   2,  -3,   4,   0,  -1,  -1,   0,  11, -18,  -1, -33,   4,  -4,   2,   0,  95,  13,  29,  -4, -24,   1,  -7,   0, -33,  24,   3,  31,   0,   3,   0,   0,  -9,   1,  -1,   0,  -7,  -1,   4,   7,  -1,  -2,  -5,   1,  -4,   4,  -1,   1 },
     {    1, -10, -15,   7,  -2,   4,  -1,   0,  -1,   0,   0,   0,   2, -14,  -9,  -3,   1,   0,   2,   0,   1,  -1,   0,  -1,   4, -18, -41,  17,   6,   3,   3,   2,  -1,   1,   1,   2, -11, -63,  -6, -28,  13,   0,   4,   2,   3,   0,   0,   1, -10,   2,  56, -26, -10,   2,  -2,  -2,   8,   7,   7,   8,  -9,  -1,  -3,   0,  -1,  -5,  13,  -2,  -2,  -2,   0,   0,  35, -10,   1,   2,  -7,  -2,  -4,   0,   7,   9,  21,  11,   0, -12,  -9,  30,  -5,   6, -20,  -2,   0,   2,  -1,  -3 },
     {    8,   6,   9,   2,  -9,   1,  -3,   0,   0,  -1,   0,   1,   5,  -6, -15,  10,  -4,   3,   0,   0,   0,  -1,   0,   1,  14,  22,  24,   6,  -9,  -2,  -4,  -1,  -1,  -1,  -1,   0,  18,   2, -60,  28,   1,   2,   2,   1,   0,   0,   0,   1,   9, -44,  -9, -13,  12,  -2,   3,   0,  -1, -10,  -2, -10,   8,  -1,   2,   0,  16, -34,   7, -24,  -5,   2,   0,   0,  49,  20,   5,   3, -19,   0,  -3,   0, -19,  15,   6,  21,  -2,  -2, -16,   4,  -6,   7,   0,  -9,   2,  -1,   2,   1 },
     {    4,  10,  19,  43, -10,   5,  -6,  -1,  -2,  -1,  -2,   0,  -5,  -1, -12,  23,  -4,   4,   0,   0,  -1,   2,  -1,   1,   1,   9,   3,  25,  -8,  -1,  -3,   0,   0,   0,  -1,   0,  -5,  -5, -22,   9,   5,   1,   2,   2,   0,   0,   1,  -1,   0,   0,   5,  19,  -1,  -1,   0,   2, -17,  -8, -17,   1,   9,  -1,   2,   1,  -7,  10,   5,  20,  -3,  -1,   0,   0, -43, -15, -22,  -2,  19,   1,   6,   0,  14, -25,  10,   3,  -6,  -7, -52,  19,   2,  16, -13, -49,   2,  -1,  14,  -5 },
     {   -8, -10, -31, -63,  29,  -6,  13,  -2,   4,   2,   2,   0,  -3,  -6, -17,  -9,  13,   6,   3,   2,  -1,  -1,   0,   1,  -3,  -8,  -6,   3,   0,   2,  -2,  -2,   1,   0,   0,   0,   3,  -7, -24,  18,   0,   2,  -3,   3,   0,   1,   1,   1,  -8, -48,  -3, -12,   9,   6,   1,   3, -34, -10,  32, -17,  -6,   2,   0,   2,  -8,  23,   5,  24, -10,   4,  -3,  -2, -20, -14,   1,  -4,   8,   2,   3,   0,  29,  -8,   0,  -8,   0,   5,  -3,   4,   4,   1,  -3,  -6,   1,   1,   5,  -1 },
     {   -2,  -5, -28, -45,  19,  -3,   6,   2,   4,   1,   1,   0,   5,   4,   3,  11, -17,   5,  -4,  -1,  -1,   2,  -3,   0,   8,   3, -10,  27,  -2,   1,  -3,   0,   0,   1,   0,   0,   1,   7,  16,  -2,  -3,   1,   0,  -1,   0,   0,  -1,  -1,  24,  34, -27,  35,   1,  -4,  -1,  -1,  29, -39, -25,  -7,  18,  -3,   3,   2, -11, -14,   2, -12,  -7,   2,   1,   1,  21, -27,  -8,  -9,  -4,  -1,   0,   0,  15,  24,   6,  14, -16,  -2, -12,  17,  -3,   8, -13, -16,  -2,   0,   7,  -6 },
     {    1,  -6,   4,  25, -19,   1,  -3,   3,  -2,   1,   0,   1,  10,   6,  34,  74, -31,   2,  -9,   0,  -5,  -1,  -1,  -1,  -1, -13,  -1,   6,   1,  -5,   4,   0,  -1,   1,   0,   0,   6,   1,  17,   8,  10,  -2,   1,  -4,  -1,   0,  -1,  -1,   3, -22, -10, -19,  18,   2,   7,   0, -13, -22,  20, -23,   1,   1,  -1,   0, -11,   4,   3,  -5,  -6,   5,   1,   1,  -1, -18,  14,  -4,  -2,  -2,   0,   0,  22,   2,   3,  -9,   1,   0,  25,  19,   0,  -7, -19,  28,  -5,   5, -11,  -7 },
     {    7,   2,  12,  47,   0,   2,  -9,   0,  -2,   0,  -2,   0,   2,  -2,  -6, -49,  20,  -4,   6,  -1,   2,   1,   1,   0,  -3, -13, -35, -51,   4,  -5,   6,   1,   1,   2,   1,   1,   4,  -3,   7,  20,  -9,   3,  -5,  -1,  -2,   0,   0,   0,  14,  -4, -15,   4,  17,  -3,   3,  -1,  -2, -40,   3, -36,   7,   5,   3,   0, -18,   5,   9,  -1,  -8,  -3,  -1,  -1,  20, -15,   1,  -8,   0,  -1,   0,   0,  27,  11,   5,  -1,   0,  12,  -3,   5,   2,   4,   8,  -4,  -1,   4,   4,   9 },
     {    0,   0,   0,  -3,   8,  -9,  -3,  -3,   6,  -3,   1,  -1,   3,   0,  17,  25,  10,   8, -11,   0,  -2,  -7,   0,   2,   4,   7,  17, -30,  51, -18,   7,  -3,   0,   1,   0,  -2,  -7,  -6, -23, -68,  12,  -3,  14,   5,   3,   3,   3,   2,  -2,  -4,  -3,  -6, -19,  -3,  -7,   2,  -5,  -8,   4, -20,  22,   4,   3,   3,  -8,   8,   5, -12, -11,   7,   2,  -4,  -6,  -2,  -8, -15,  -7,   5,   3,   0,   0,  14,   5,   2,  -3,   7, -10, -21,   8,  13,  26, -14,   3, -12,  10,  22 },
     {    2,  -2,   6,   7, -23,  -8,   3,  -4,   4,  -4,   1,   0,  -5, -15, -18, -39, -25,  12,   1,   3,   0,  -1,   1,   1,   6,  -3,  23,  61,  -9,  -7,  -9,   5,  -4,   2,  -6,  -2,   4,  -4,   8, -20,  24,   5,  11,  -6,   1,   2,   2,   0,   2,  -5,   9,   7,   6, -15,   3,  -4,   6,  -7,   9, -22,  15,   2,   8,   1,  -9,  -9,   7, -11,  -2,   1,  -1,   2,   0,   7,   8,  -8,   9,   5,   6,  -6,  31, -23,   0, -14,  -4,  27,   6,  -9,  -1,   0,  29,  19,  -6,  -5, -15,  28 },
     {    4,   2,   3,   0,  -8,   7,  -8,   2,  -1,   3,  -1,   0,   6,   3,   1,   0,  -9,   1,  -4,   0,  -3,   1,   1,   0,   4,   6,   5, -19,  -4,  -7,  -3,  -2,  -3,   1,   0,  -1,   9,  10, -15,   3,   0,  -7,   1,  -3,   1,   1,  -1,   0,   1,  -3,   6,  -3,  -6,   3,  -1,   1,  18,  39,  -4,  17, -11,  -1,  -2,  -2,  55,  -6, -21,  11,  20,  -3,   3,  -1,   2, -43,  -5, -41,  -7,   6,   3,  -1,  58,  15,  -2,   5, -18,  28,   0,   5,   2,   1,   5,   3,  -8,   5,   1,   4 },
     {    0,   7,   6,  12,   7,  -1,   2,  -4,   0,   2,   0,  -2,  -4,   5,   4,  -4, -14,   3,  -5,   4,   3,   1,   0,  -1,   1,  10,  15,  26,  -9,  -1,   3,  -3,   0,  -3,  -1,   2,  -2,  13,  19, -10,   3,  -6,  -8,   0,   1,  -2,  -1,  -1,   1,  11,  -2, -22,  12,   2,   6,   3, -24,   8,   4, -23, -21,  -1,  -2,   2,   5,  58, -11,  38,  -5,  -7,  -7,  -2,  45,   9, -17,   6,   4,   0,  -2,   0,  -7,  37,  -1,  22, -35,  -8,  -8, -12,   6, -11,  10, -14,  -7,   2,   4,   5 },
     {    1,   3,   3,  10,  -3,   1,  -5,   7,  -4,  -1,  -2,   1,  -4,  -3, -10,  -9, -29,  -7, -13,   1,   3,   5,   2,  -5,   1,   6,  -3,   0,  26,   9,  15,   6,  -4,  -6,  -1,   2,  -6,   0, -35, -36, -35,  27,  -5,  -3,  -1,   1,  -4,   0,  -3, -14, -21,  47, -28,  10,   7,   0,  -6,  -8,  12,  10, -14, -18,  -5, -11,  -6,   4,  -8,  17,   8, -16,  -9,   7,  16,   8,   5,  24,   6, -10,  -2,   1,  18,  -9, -10,   6,  -6,  -1,  12,  11,   2, -18, -19,  14,  -9,  17, -15,  -8 },
     {    4,   5,   1,  27,  66, -59,  -3,   1,   2,   1,   1,   0,   0,  -1,   0,   1,  -4, -14,   7,   2,   2,   4,   1,  -2,   1,  -1,   9,  -2, -34,  40,  18,  13,  -4,  -4,  -2,   3,   1,  -5,  -5, -24,  15,   1,   9, -10,  -1,   5,  -5,   0,  -3,  -7,  -3,  -1,  24,  -7,   1,  -4,   2,   1,   6,   4,  -9,  -4,  -9,  -9,  -1, -13,   2, -19,  14,  -1,  -3,  12, -17,  -3,   3, -12,  -3,   2,  -5,  -3,   1,   5,  -3,   6, -14,  -4,   9,  -2,  -2,  -4,  -5,  -6,   3,   5,   4,   2 },
   },
   {
     {  -98,  31,  17,   3,   3,   1,   1,   0,   1,   0,   0,   0,  65, -29,  -9,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   5,   9,  -6,  -2,  -1,   0,   0,   0,   0,   0,   0,   0, -11,   1,   5,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   4,  -4,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   40, -37,   8,   0,   1,   0,   0,   0,   0,   0,   0,   0,  73,   3, -27,  -1,  -5,  -1,  -2,   0,  -1,   0,  -1,   0, -72,  38,  12,  -2,   2,   0,   1,   0,   0,   0,   0,   0,  -3, -16,   7,   3,   1,   1,   0,   0,   0,   0,   0,   0,   3,   4,  -6,   0,   0,   0,   0,   0,  -6,   1,   3,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {  -29, -84,  39,  19,   6,   4,   2,   2,   1,   1,   1,   1,  -5,  52, -27,  -7,  -1,  -1,  -1,  -1,   0,   0,   0,   0,  47,  10,  -5, -11,  -4,  -2,  -1,  -1,  -1,  -1,   0,   0, -18,  -7,  10,   5,  -1,   1,   0,   0,   0,   0,   0,   0,  -2,  -3,  -3,   2,   1,   0,   0,   0,   1,   1,  -2,  -1,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,  -1,   1,   1,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -43,  -1,   8,   9,   1,   2,   0,   1,   0,   0,   0,   0, -24,  38,  -9,  -8,   1,  -1,   0,  -1,   0,   0,   0,   0, -49, -30,  34,   3,   4,   1,   2,   0,   1,   0,   1,   0,  73, -31, -23,   6,  -3,   1,  -1,   1,  -1,   0,   0,   0,  -3,  25,  -5,  -6,   0,  -1,   0,  -1,   1,  -8,   7,   1,   0,   0,   0,   0,   3,   0,  -4,   1,   0,   0,   0,   0,   3,  -1,  -1,  -1,   0,   0,   0,   0,   1,   2,  -1,   0,   1,  -1,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 },
     {   10,  14, -23,   5,  -4,   2,   0,   1,  -1,   0,   0,   0,  44,  52,  -8, -23,  -4,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   5, -71,  29,  20,   1,   3,  -1,   2,   0,   1,   0,   1, -43,  33, -22,   3,   4,   0,   1,   0,   0,   0,   0,   0,   2,   1,  11, -11,   1,   0,   0,  -1,   6,  -8,   1,   4,  -2,   0,   0,   0,  -5,   3,   1,   0,   1,  -1,   0,   0,  -1,  -3,   0,   0,   0,   1,   0,   0,   0,   0,   1,  -1,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   1 },
     {   13, -26,  26,  -8,   2,  -1,   0,   0,   0,   0,   0,   0,  25, -31,  -7,   9,  -1,   2,   0,   1,   0,   0,   0,   0,  16, -16, -21,   8,   1,   2,   1,   1,   1,   1,   0,   0,  49,  55, -26, -15,  -6,  -3,  -2,  -1,  -1,  -1,  -1,  -1, -58,  22,  32,  -2,   0,  -1,   0,  -1,   2, -24,   1,   9,   2,   1,   0,   1,  -3,   8,  -7,  -2,   1,   0,   0,   0,  -1,   0,   3,  -2,   0,   0,   0,   0,  -3,   1,   1,   0,   0,  -1,   1,   1,  -2,   1,   0,   0,   0,  -1,   0,   0 },
     {   12,  28,  75, -48,  -9,  -7,  -1,  -5,  -1,  -2,   0,  -1, -16,   7, -49,  38,   0,   3,   0,   2,   0,   1,   0,   1, -15, -35, -11,   5,  13,   3,   2,   2,   2,   1,   1,   1, -13,   4,  24, -12,  -4,   1,   0,  -1,   0,   0,   0,   0,  20,  -4,  -8,   6,  -3,  -1,   0,   0,   2,   4,  -1,  -1,   1,  -1,  -1,   0,  -1,  -1,  -1,   2,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   1,   1,   1,   0,   0,   0,  -1,   1,   0,   0,   0,  -1 },
     {   16, -12,   9,  -1,  -3,   0,  -1,   0,   0,   0,   0,   0,  15, -41,  -5,   4,   7,   1,   1,   1,   1,   1,   1,   1,  38, -22,  17,  -5,  -5,  -1,  -2,   0,  -1,   0,   0,   0,  -1, -20, -27,  22,  -2,   4,  -1,   2,   0,   1,   0,   1,  32,  63, -36, -19,   1,  -4,   0,  -2, -36,  17,  36,  -6,  -3,  -1,   0,  -1,  -2, -12,  -1,  10,   1,   0,   0,   0,  -7,   5,  -1,  -2,   2,   0,   0,   0,   0,   1,   0,   0,  -4,   2,   2,   0,   0,   0,   0,   0,  -2,   1,   1,   0 },
     {   15,  48,  -8,  -4, -11,  -2,  -3,  -1,  -1,  -1,  -1,   0,   4,  21, -48,   6,   7,   1,   2,   1,   1,   1,   0,   0,  35,  35,  35, -31,  -8,  -4,  -1,  -3,  -1,  -1,  -1,  -1,  -1, -21,  10,  24,  -2,   0,  -2,   1,  -1,   0,   0,   0, -58,  11,  -2,  -4,  11,  -1,   2,   0,  13, -26,  16,  -3,  -3,   2,   0,   0,   3,  -1,  -6,   3,   0,   0,   0,   0,  -1,  -1,   5,  -2,   0,   0,   0,  -1,  -1,  -4,   0,   2,   1,  -3,   1,   1,   0,  -1,  -1,   1,   0,  -1,   1,   0 },
     {   10,  25,   7,  12, -10,  -4,  -2,  -2,  -1,   0,  -1,   0,  18,  39,  23, -24,  -6,  -3,  -2,  -3,  -1,  -1,   0,  -1,   4,  -2, -77,  21,  14,   2,   3,   2,   1,   1,   1,   1,   3, -34,  30, -12,  -1,   4,   0,   1,   1,   0,   0,   0, -16,  45,   8,  -3,  -6,  -3,   0,  -2, -13,  -7,  17,  -1,   5,  -1,   0,   0,  10, -19,  -1,   5,  -1,   2,   0,   1,  -5,   1,  -2,  -1,   0,   0,   0,   0,   1,  -3,   3,   0,  -1,  -1,  -1,   1,   0,  -2,   0,   1,  -1,  -1,   0,   0 },
     {   10, -10,   5, -20,   5,   0,   2,   0,   1,   0,   1,   0,  26, -12,  29,   8, -13,   0,  -3,  -1,  -2,   0,  -1,   0,  18, -35,   3,  11,   8,  -1,   0,   0,   1,   0,   0,   0,   9, -53,   5,  14,   1,   6,  -1,   2,   0,   1,   0,   1, -40, -10, -32,  13,   1,   5,   2,   2,  44,  31, -35, -18,  -1,  -3,   0,  -1, -13,  30,  13,  -6,  -4,  -2,  -1,  -1,  -3,  -1,   4,   4,   1,  -1,   0,   0,  -6,   4,  -2,   0,   0,   2,   0,  -3,  -2,   3,   1,   0,   0,   1,   0,   0 },
     {   12,  14, -11,  37, -10,  -2,  -5,   0,  -2,   0,  -1,   0,  16,   8, -29, -13,  10,   0,   2,   0,   1,   0,   1,   0,  13, -15,  -8, -24,   2,   4,   2,   1,   1,   1,   1,   1,  46,  31,  25,  -5, -19,  -5,  -4,  -1,  -2,  -1,  -1,  -1,   3, -23, -24,  31,   7,  -1,   0,   0, -29,  48,  -8, -25,   8,  -2,   3,  -2, -10, -14,  36,  -4,  -5,   1,  -1,   0,   4,  -7,  -5,   8,  -1,   0,   0,   0,  -4,   0,   3,  -1,  -1,  -1,   0,   2,  -2,  -1,   2,   1,   0,   0,   0,   1 },
     {   12,   6,  36,  64, -30,  -8, -10,  -3,  -4,  -1,  -2,   0, -11, -34,  -5, -40,  33,   1,   7,   0,   4,   0,   2,   0, -19, -22, -10, -21,  -7,  11,   3,   4,   1,   2,   1,   1, -16,  -7,   7,  30,  -5,  -4,  -1,   0,   0,   0,   0,   0,   5,   2,   8,  -7,   7,  -1,  -1,  -1,  27, -16, -18,  10,  -5,   3,  -1,   1,  -3,  23, -17,  -9,   2,  -1,   1,  -1,  -5,   2,   9,  -2,  -1,   0,   0,   0,   1,   2,  -2,   0,   0,   2,   0,  -2,   0,   2,  -1,  -1,   0,   1,   1,   0 },
     {    3,  16,  23,  -2,  -8,  -7,  -2,  -2,  -1,  -2,  -1,   0,  -7,   6, -17, -24,   8,   3,   1,   1,   0,   1,   1,   1,   2,  33, -11,  32, -16,  -4,  -3,   0,  -2,   0,  -1,  -1,   1,  -2, -47,   2,  25,  -2,   3,   0,   2,   0,   0,   0,  -1,  -3,  20, -36,  -7,   6,   1,   0,  26,  50,   6,   2, -13,  -8,  -1,  -3, -32, -10,  37,   3,   9,  -2,  -1,  -1,  11, -28,  -5,  11,   0,   3,   0,   1,  -4,   2,  -4,  -6,   3,  -3,   5,   1,  -2,   0,   1,   1,   3,  -3,   0,   2 },
     {    4,   0,  45, -18,  -2,  -4,   2,  -4,   0,  -1,   0,  -1,   2,  -3,  42, -45,   7,   2,   1,   0,   1,   0,   0,   0,  -2,   3,  32,  22, -17,   2,  -5,  -1,  -2,   0,  -1,   0,   4,  14,  20,  19,   5,  -5,  -2,  -1,  -1,  -1,  -1,  -1, -24, -34, -21,   3,  12,   7,   2,   3, -15,  -6,  22,  -5,  -9,   2,   0,   1,  13, -40,  -9,  25,  -3,   4,  -1,   3,   7,  16, -33,  -1,   2,  -1,   1,   0,  -1,   4,   1,  -3,   1,   0,  -6,   3,  -1,   1,  -2,  -1,   0,   1,  -3,   0 },
     {  -20, -22, -33,   1,   1,   7,   4,   3,   1,   1,   1,   1, -22, -18, -21,   2,  10,   4,   2,   1,   0,   1,   1,   1, -27,  -8,  -9,  19,   0,  -2,  -2,   1,   1,   0,   1,   0, -21,  30,  40,  -5,  -9,  -3,  -3,  -2,  -1,   0,  -2,  -1, -21,  35, -37, -13,  12,  -2,   6,  -1,  50,  -4,  26, -21,  -5,  -2,  -2,   0, -18, -16,  16,  19,  -4,   1,  -2,   1,   8,  -5, -11,   1,   3,  -2,   1,   0,   0,   1,   2,  -2,   7,  -2,  -2,  -1,   0,   0,   2,   0,   3,  -1,  -1,  -1 },
     {   -4,   1,  18,   9,   2,  -1,  -5,  -1,  -1,   0,  -1,   0,   0,  14,  38,  34,  -8, -10,  -3,  -4,  -2,  -2,  -1,  -1, -21,  -4, -13, -66,  19,   9,   8,   2,   2,   1,   1,   1, -25,   9, -42,  18, -18,   6,  -1,   2,   1,   2,   0,   1, -13,   2,  15,  13,  11, -12,   1,  -2,  18,  15,  24, -25,  -1,   3,  -3,  -1,   1, -24,  10,  20,  -7,   3,   1,   1,   4,  -9, -17,   4,   6,   0,   1,   1,  -2,   0,  -1,  -2,   4,  -1,  -1,  -1,   1,   0,   0,   1,   1,   0,  -2,   0 },
     {   -6, -11,  -6, -26, -77,  41,   5,   8,   0,   4,   0,   2,   7,   5,  15,   7,  50, -38,   2,  -5,   1,  -2,   0,  -2,   7,  12,  10,   9,  30,   0, -14,  -3,  -3,  -1,  -1,   0,   5,  -1,   1,  -6, -31,  16,   5,   2,   0,   0,   0,   0,   5,  -3,   6,  -9,   1,  -8,   5,   0,  -4,  -2, -10,   5,   3,   3,  -3,   1,  -8,   0,   5,   1,   2,  -1,   0,  -1,   4,  -5,   0,   2,  -3,   2,  -1,   1,   0,   2,  -4,   1,   0,  -1,  -1,  -1,  -1,   1,   0,   0,   1,  -1,  -1,   1 },
     {    1,  -8, -17,  36, -21,   6,  -2,   2,  -1,   0,  -1,   1,   0, -16, -32,  26,  -4,   2,  -1,   3,   0,   1,   1,   1,   3, -17,  -6,  19,  20,  -8,   1,  -1,   0,  -1,   0,   0,  -3, -27, -15, -17,  14,   8,   3,   2,   1,   1,   1,   1, -17, -33,  12,   1, -16,   6,   0,   4,   7,  25,  14,  11,  -5,  -7,  -4,  -2,  11, -31, -37,  18,  11,   3,   0,   1,   2,  38, -29, -23,   5,  -1,   2,  -1,  -8,   7,  21, -13,   3,   1,   1,   0,  -1,   2,   4,  -3,   1,   2,   1,  -2 },
     {    8,  -2,   1,   1,  15,  -4,   1,  -1,   2,   0,   1,  -1,  10, -11,   8,  -7,  -6,   7,   2,   0,  -2,  -1,  -1,   0,  15, -10,  29, -26,   2,  -5,   0,  -3,   3,  -2,   1,   1,   9, -19,  35, -23,  -7,   7,  -3,   1,  -2,   2,   0,   1,  19,  18,  39,  10, -22,  -1,  -1,  -1,  19, -11, -12,  34,  11,   1,  -2,   4, -55, -34,   4,  11,   4,   9,   1,   2,  15, -14, -32,   7,  -3,   1,   3,   2,   3,  22, -15,  -4,  -2,   0,   3,   2,  -4,   3,  -3,   2,   0,  -1,  -4,   0 },
     {   -4,  -2,  11,  12,  -8,  -9,  -7,  -2,   0,  -1,   0,   1, -17, -19,  -8, -15,   2,   7,   5,   0,   2,   3,   2,   1,  -8,   3,  30,  18,  16, -11,   2,  -1,  -3,  -1,  -2,  -2, -34,  -9, -15, -30,   1,  12,   0,   4,   2,   2,  -1,   1, -27,  25,   7,  33, -24, -10,   0,  -2, -15,  -3, -15, -14,  33,  -8,   3,  -4,  45,  -8,  36, -23, -17,   8,  -5,   2,  -1,  -9, -12,  28,   0,  -1,   1,   0,   0,   4,  -5, -14,   1,  -3,   1,   3,   4,   0,  -3,  -4,   0,   1,   2,   1 },
     {    9,  14,  23,  38,  -7, -14,  -2,  -4,  -1,  -2,  -1,   0,   6,  10,  23,   6, -17,   2,  -5,  -2,   0,   1,   0,   0,  10,  18,  16,   4,  29, -13,   0,  -3,  -1,  -2,  -1,  -2,   5,  13, -11, -51,  -5,  17,   2,   1,   0,   0,   0,   1,   7, -15, -53,  -5, -20,  -3,   6,   3,  16, -22,  38,  13,   2,  -8,  -6,  -3, -21,  32,   2,   4,   5,   3,  -2,  -1,   0,  -6,   9,  -4,   1,  -1,   1,   0,   0, -10,  -3,   7,  -3,   1,   0,   1,  -1,   0,  -1,   1,   0,  -2,  -1,   1 },
     {    0,   6,   1,  15,  -3,  -2,  -2,  -1,   1,  -1,   1,  -1,  -3,   3,  -6,  19, -21,   1,   4,  -3,  -1,   1,   1,  -1,   1,   8, -18,  23,  14,   3,  -6,   2,  -1,   0,  -1,   0,  -3,  -1, -16,  34,   2,  -2,   3,  -5,   0,   0,   1,  -1,   3,  -5, -22,  -8,  18,  -5,   5,   3, -29, -44, -28, -11,  14,  13,  -2,   4, -31,  10,  11,  -2, -10,  -3,   5,  -1,  42,  -1, -36,   1,  10,  -3,  -1,   2,  -7,  37,   7, -32,   0,  -9,  20,   2,  -2,   1,  -1,  -2,   1,  -3,   4,  -2 },
     {   -3,   5,   4, -20,  -1,  -6,  -3,  -1,   0,   0,  -1,   1,  -5,   7,  -1, -30,  -3,   5,   5,   3,   2,   2,   1,   1,  -2,  18,   1,   1,   3,   7,  -1,   1,  -3,   0,  -1,  -1,  -9,  11, -24,   3, -13,   3,   1,   1,   0,   0,  -1,   0,   3,  24, -12,  42,  -7, -11,   0,  -2,  13,  13, -15,  -6,  24,  -3,   1,  -4, -37, -15, -25, -23,   6,   6,   4,   1,  30,  36,  21, -36, -11,  -8,  -1,  -2,  -2, -10,  44,  13,   1, -10, -13,   9,  -1,  -2,   5,  -1,   3,  -1,   0,   1 },
     {   -2,  -2,  -1,  -9,   6,   2,   2,   2,   0,   1,   1,  -1,  -8, -10, -10, -31, -43,  18,  11,   1,   1,   0,   1,   0,   0,   3,   3,  -7,  74, -31,  -7,  -3,   2,  -3,   0,  -1,   5,  15,  13,  44, -21,  17,  -6,  -4,  -1,  -2,  -1,   0,   1,  -6,  14, -37, -19,   5,  11,   2,  -2,  12,  -7,   1,   6,  -7,  -2,   2,  11,  -2,   3,   8,   2,   1,  -2,  -2,  -6,   0,  12,  -2,   0,   2,   1,   1,   6,  -7,   2,   8,   0,  -2,  -6,   0,   0,   2,  -1,   0,  -1,  -1,  -1,   1 },
     {    8,   0,  -6, -17,  12,  -2, -14,  -1,   1,   1,   0,   2,   6,  -6, -21, -34,  -2,  -9,  15,  -1,   4,   0,   3,   0,  17,   8,  -7,  13,  10,   0,   3,   3,  -5,  -1,  -3,  -1,  17, -13, -17, -12, -19,   8,  -5,   2,   5,   1,  -1,   0,  39, -13,  20,  30,  20, -17,  -4,  -6,  35, -21,  18, -32,  -8,  20,  -4,   1,   7,  17,  -3,  28, -27,  -7,   0,   0, -29,   5, -14,   9,  24,  -6,  -1,  -2,  -8,  -1,   9, -20,   1,  -2,  -1,   4,   1,   6,   1,  -2,  -3,  -1,   3,   2 },
     {    7,  13,  18,  19,  26,  78, -21, -10, -11,  -2,  -6,   0,  -3,   0,   0,   4, -18, -41,  27,   2,   4,  -3,   3,  -1,  -4,   0,  -1,   1, -25, -43, -11,  10,   5,   2,   1,   2,   0,   8,   0,   0,  21,  30, -11,  -7,  -2,   2,   0,  -1,   0,   8,  -6,   7,   4,   4,  11,  -3,   3,  -1, -12,  -2,   5,  -5,   0,   5,  -5, -10,  -4,  -1,  -5,  -2,   1,   0,  -8,   4,   8,  -1,   2,   1,  -2,  -1,   3,  -2,  -2,   1,   4,   5,  -1,  -2,  -1,  -1,   0,  -1,   0,  -2,  -1,   1 },
     {    2,   5,   9,  21, -33, -18,  -1,   2,  -1,   0,  -1,   1,   6,  14,  14,  34, -25,   5,  -9,   1,   1,  -1,  -1,   0,   9,  20,  18,  21,  -6,   8,   2,  -2,  -5,  -2,  -2,  -1,  15,  23,   5,  15,  12, -15,   2,   0,   1,  -2,   0,  -2,  22,  23,  -2,  -5,   3,   7,  -3,  -4,  30,  -5, -41, -13,  15,  -1,   2,   0,   3, -34,   3,   7,  -6,   2,   1,   2, -56,  22,   2,   1,   2,  -2,  -2,  -1,  11, -13,  -3,  -9,   2,   6,  -2,   2,  -2,  -1,   0,  -5,   0,  -3,   1,   2 },
     {   -2,  -3,  -3,  12, -14,  -2,  -4,   7,  -1,   1,  -1,   0,  -5,  -8,  -6,   8, -28,   2,   3,   1,   1,   1,   1,   1,   1,   0,   2,  40,   5,  11,  -1,  -1,  -3,  -1,   1,   0,  -3,  -4,  -1,  25,  -6,   3,  -1,   0,   0,   0,   1,   0,  -4, -10,  11,  42,  15, -18,   4,  -3,  -5,   5,  29,  -6,   7,  11,  -3,   3, -20, -15, -33, -14,   5,   5,   7,   1,   7, -36,  21,  11, -11,   1,  -4,   1,   6,  -3, -52,  21,  -7,  25,  -7, -28,   3,  -2,   8,  -4,   1,   1,  -1,  -1 },
     {    5,   4,   9,  20,  41, -14, -17,   6,  -3,  -1,  -3,   2,   0,  -5,  -2,  12,   1, -31,  22,  -6,   6,  -2,   2,  -2,   4,   2,  15,  21,   7,  37, -15,  -3,  -6,   1,  -2,   0,  -4, -12,   5, -19, -36,  -2,  16,   3,   0,  -2,   4,  -1,  -8,  -4,   6, -37,  26, -33,   3,   4,  -2,   9, -17,  14,  -1,  21, -14,   0,  11, -19,  20,   5,  -9,  -6,   7,  -4,  16,  19,  17, -10,  -1,   5,  -6,   1,   5, -12,   0,  19,  -9,   2, -11,   0,   1,  -1,   0,  -4,   3,   1,  -3,   1 },
     {    1,  -7,   1,   4, -15,  -6,  -1,   2,   1,   9,   3,  -1,   4,  -9,   3,   8,  -4, -16,   8,  -8,   3,   1,   1,  -3,   6, -10,   8,  -7,   9,  16,  -2,   0,   5,   0,  -3,   0,  10,  -5,  19,   2,  66,   0,   0, -15,   3,  -2,   1,  -2,  10,  -7,  22, -12, -22,  -9,   6,   9,  -1, -20,  16, -47,  -5,  10,  -8,   6,  -8,  -3,  12,  -6, -16,   5,  -3,  -6,  30,   5,  18,  19,  -6,  -8,   0,   6, -14, -15,  23,  21,  -1,  -9, -35,   6,  -5,   5,   2,  -9,  -1,   1,   2,   2 },
     {   -1,   2,   7,   9, -16,  11,  12,  -1,   0,   0,  -3,   0,  -6,  -6,   2,  -9, -21,   5, -16,   0,  -2,   2,   0,   1,  -1,   2,  15,  -3,  21, -17,  18, -17,  -3,  -3,   0,  -2,  -9,   0,   3, -25,  19, -15,  -3,   1,  -1,   2,   3,   0,  -6,  23,   5,  20,  31,   1, -13,   3, -16,   8, -26,  -1, -44,  32,  15,   0,   1,   3,   1,  30,  15, -39,   9,   3,  18, -21,  10, -21,  23,  12, -15,  -4,   1, -10,   4,   6,   3,  -2,  -7,  12,  -5,  -1,  -2, -10,   9,  -1,   3,   4 },
   }
 },
 {  //26
   {
     {  105, -30, -15,  -4,  -3,  -1,  -1,  -1,  -1,   0,  -1,   0, -56,  28,   5,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -8,  -8,   5,   1,   1,   0,   0,   0,   0,   0,   0,   0,   3,   2,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   2,   2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -34,  40, -10,   0,  -2,   0,  -1,   0,   0,   0,   0,   0, -85,  -3,  27,   2,   6,   1,   2,   1,   1,   0,   1,   0,  58, -40,  -9,   2,   0,   1,   0,   0,   0,   0,   0,   0,  10,  12,  -8,  -2,  -1,  -1,  -1,   0,   0,   0,   0,   0,   4,  -4,   2,  -1,   0,   0,   0,   0,   4,   2,  -3,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   1,  -1,  -1,   0,   0,   0,   0,   1,   0,   0,  -1,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   34,   4,  -4,  -5,  -2,  -2,  -1,  -1,   0,   0,   0,   0,  19, -51,   8,   6,   2,   2,   1,   1,   0,   0,   0,   1,  59,  34, -37,  -3,  -6,  -1,  -2,  -1,  -1,   0,  -1,   0, -57,  39,  22,  -7,   1,  -2,   1,  -1,   0,   0,   0,   0,  -2, -22,   8,   5,   1,   1,   0,   1,  -5,   6,  -4,   0,   0,   0,   0,   0,  -2,  -3,   3,   0,   0,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0,  -1,  -2,   1,   1,  -2,   1,   0,   0,   0,  -1,   1,   0,  -1,   1,   0,   0 },
     {   29,  91, -31, -13,  -9,  -4,  -4,  -2,  -2,  -1,  -1,  -1,   2, -43,  22,   6,   3,   1,   1,   1,   1,   0,   0,   0, -52,  10,   1,   1,   4,   1,   2,   0,   1,   0,   1,   0,  14, -18,   9,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   6,  -2,  -7,   3,   0,   0,   0,   0,   0,   2,   3,  -2,   0,   0,   0,   0,   1,  -4,   0,   2,   0,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   15,   1, -12,   2,  -1,   1,  -1,   0,   0,   0,   0,   0,  41, -19,  15,  -8,  -1,  -2,   0,  -1,   0,  -1,   0,   0,  -1, -53, -12,  20,   1,   5,   0,   3,   0,   1,   0,   1,  34,  58, -44, -14,  -3,  -3,  -1,  -1,  -1,  -1,   0,   0, -34,  26,  35,  -9,  -1,  -2,   0,  -1,   0, -15,   1,   9,   1,   1,   0,   1,  -7,   6,   1,  -3,   1,   0,   0,   0,  -1,  -3,   1,   1,  -1,   0,   0,   0,  -3,   2,   1,   0,   0,  -1,   0,   1,  -2,   1,   1,   0,   0,  -1,   0,   0 },
     {   -5,  33, -37,  10,  -3,   3,  -1,   1,   0,   1,   0,   0,  22,  75,   7, -21,  -6,  -6,  -2,  -3,  -1,  -2,  -1,  -1,  -9, -18,  14,   9,   2,   1,   0,   0,   0,   0,   0,   0, -46,  21,  14,  -9,   4,  -2,   2,  -1,   1,   0,   0,   0,   3, -47,  25,   9,   0,   2,   0,   1,   9,  -8, -20,   5,   1,   1,   0,   1,   0,   0,   5,  -5,   0,   0,   0,   0,   2,  -6,  -2,   3,  -1,   1,   0,   0,   0,  -1,   0,   0,   1,  -2,  -2,   1,   0,   0,   0,   0,   0,  -1,  -1,   1 },
     {   17,  11, -22,   6,  -5,   1,  -1,   1,   0,   0,   0,   0,  31,   4,   8, -13,   1,  -3,   0,  -2,   0,  -1,   0,  -1,  36, -34,  28,   8,  -4,   0,  -2,   0,  -1,   0,  -1,   0, -41, -34, -34,  21,   6,   6,   2,   3,   1,   2,   0,   1,  35,  41, -30, -25,  -3,  -3,   0,  -1, -24,  18,  28,  -2,  -4,  -2,   0,  -1,   3,  -8,  -1,   6,   2,   0,   0,   0,  -7,   4,   1,  -1,   1,   0,   0,   0,   1,  -1,   0,   0,  -4,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0 },
     {    9,  44,  11,  -5,  -9,  -2,  -3,  -1,  -1,  -1,  -1,  -1,  -4,  24, -55,   6,   4,   2,   2,   1,   1,   0,   0,   1,  32,  37,  25, -22,  -6,  -6,  -2,  -3,  -1,  -1,   0,  -1,   2,   3,  -8,  11,   1,   0,   0,   0,   0,   0,   0,   0, -38,  27,  26, -19,   4,  -4,   1,  -2,   4, -44,  29,  15,  -2,   2,  -1,   1,   0,  -5, -15,   4,   3,   1,   1,   1,   0,  -6,   6,  -1,  -1,   0,   0,   0,   0,  -3,  -4,   2,   0,  -3,   2,   1,   0,  -1,  -2,   1,   0,  -2,   1,   0 },
     {   -6, -13, -75,  15,   5,   9,   3,   4,   2,   2,   1,   2,  17,  16,  22, -12,  -5,  -4,  -2,  -2,  -1,  -1,  -1,   0,  35,  33,  -3, -10,  -1,  -5,  -1,  -3,  -1,  -1,   0,  -1,  47, -14,  18,   4,  -7,   0,  -3,   1,  -1,   0,  -1,   0, -41,   9, -21,  16,   7,   1,   2,   0,  -1,  11,   6, -23,   3,  -1,   1,  -1,  -5,  -6,  16,   6,  -5,   0,  -1,   0,   0,  -1,  -2,   0,   2,  -1,   1,   0,  -2,  -2,   5,   2,  -1,  -1,  -1,   1,  -2,  -1,   2,   1,   0,  -1,  -1,   0 },
     {   25,  24,  45, -17,  -8,  -6,  -4,  -3,  -2,  -1,  -1,  -1,  28,   5, -23,   9,   2,   0,   0,   0,   0,   0,   0,   0,  30, -47,  20, -16,   4,   1,   1,   1,   1,   1,   1,   0,  20,  -9,  20,  19, -11,   1,  -4,   0,  -2,   0,  -1,   0, -31, -24, -35,  26,  10,   4,   2,   2,  12,  26, -26, -21,   0,  -1,   0,   0, -13,  14,  15,  -2,  -4,  -2,  -1,  -1,  -1,  -1,  -2,   4,   2,   0,   0,   0,  -6,   3,   0,  -1,   0,   0,   0,  -1,  -3,   2,   0,   0,   0,   0,   0,   0 },
     {   -2,  16,  16, -16,   0,  -4,   1,  -2,   0,  -1,   0,   0,   3,  41, -13,  20,  -7,  -1,  -3,  -1,  -2,   0,  -1,   0, -13, -10, -55,  -6,  13,   3,   5,   2,   2,   0,   1,   1,   3,  30,  20, -21,  -9,  -3,  -2,  -2,  -1,  -1,   0,   0,  12,  23, -16,  -4,   4,  -3,   1,  -1, -17,  20,  49, -28,  -3,  -3,  -1,  -2,   5, -37,  25,  25,  -5,   3,  -2,   1,  -1,  -3, -11,   4,   3,   0,   1,   0,   1,  -7,   4,   3,   0,  -1,  -4,   2,   1,  -3,   1,   2,   0,   0,  -1,   0 },
     {  -10,  -2, -24,  27,  -6,   4,  -2,   1,   0,   1,   0,   1, -22, -21, -49,   1,  12,   4,   5,   2,   2,   1,   1,   1, -23,   8,  18,   2, -10,  -1,  -2,   0,  -1,   0,   0,   0, -18,  35, -36,  13,  10,  -4,   1,  -1,   1,   0,   0,   0, -11, -16, -19, -28,  12,   1,   4,   1,  30,  40,   4, -28,  -9,  -4,  -1,  -2, -10,  -2,  40,  -3,  -1,  -2,  -1,  -1,  11, -12,   2,   8,  -1,   2,  -1,   1,  -3,   1,   5,  -3,   5,  -4,   3,   1,  -1,   0,   3,   0,   3,  -2,   1,   1 },
     {   -9, -14,  11, -27,   7,   0,   4,   1,   1,   0,   1,   0,  -2,  -2,  56,   9,  -5,  -5,  -4,  -3,  -1,  -1,  -1,  -1, -16,  -4,  14,  -8,   1,   4,   1,   1,   1,   0,   0,   0, -33,   3,  -3,  28,  -4,   2,  -2,   0,   0,   0,   0,   0, -45, -10, -16, -18,  19,   1,   5,   1,  47, -18,  31,  -9, -21,   1,  -4,   1,  -5, -10, -17,  35,   2,  -1,   0,   0,   8,   4, -17, -11,   5,   1,   1,   0,   0,   2,  -1,   2,   4,   1,  -5,  -2,   0,   2,  -1,   1,   2,   1,  -2,  -1 },
     {   -4, -16, -28,  13,   1,   5,   0,   1,   1,   1,   1,   1,  -1, -26, -28,  -1,   8,   2,   4,   2,   2,   1,   1,   0,  -7, -30,  42, -28,  -1,   0,   1,   0,   1,   0,   1,   0,  13,  41,  43,  15, -17,  -7,  -6,  -3,  -3,  -2,  -1,  -1,  28, -14,  -4,  13,   9,   1,  -1,   0, -21, -13,  19,  12,  -4,   5,  -1,   1,   8, -21, -32,  25,   4,   2,   1,   1,  -5,  22, -20, -14,   5,  -2,   2,  -1,  -1,   1,   4,  -4,  -1,   4,  -2,  -4,   0,   0,   2,  -1,  -1,   2,  -1,  -1 },
     {    5,   6,  -2,  69, -12,   0,  -7,  -1,  -2,   0,  -2,   0,   1,   3, -17,  -6,   6,   3,   2,   1,   0,   0,   0,   0, -10, -24, -61,   1,   2,   4,   5,   3,   3,   1,   1,   1, -12, -20,  17,  14,   2,  -2,   0,   0,   0,   0,   0,   0, -19,   7, -20,  -9,   8,   0,   3,   0,  21,  -9,   2,   7,  -2,  -2,  -1,  -1, -10,  11, -38,  -8,  14,  -2,   3,  -1,   5,  17,  14, -32,  -3,  -3,   0,  -1,   0,  -2,  14,   3,   2,   2,   2,  -6,   0,  -1,   6,   0,   1,   0,   1,  -1 },
     {   -9, -11, -46, -51,   0,   9,  10,   5,   4,   3,   2,   1,   7,   7, -21,  78,  -6,  -3,  -9,  -3,  -3,  -1,  -2,   0,   1, -13, -13, -17,  15,   3,   5,   1,   2,   0,   1,   0,  -6,  -4, -11,   0,  -9,  -2,   1,   0,   1,   0,   1,   0,  13,   1,  -2,  -7,   7,  -1,   0,  -1,  12,   1, -12,   7,  -3,   0,   0,   1, -12,  25, -15, -18,  12,  -5,   2,  -2,   2,   1,  26, -18,  -7,   1,  -1,   1,  -1,  -1,   5,   7,   1,   0,   5,  -5,   0,   0,   2,   3,   0,   0,   3,  -2 },
     {    9,   1,   3,  -5,  -2,  -2,   0,   0,   0,   0,   0,   0,  14,   5,  -5,  -2,   1,   0,   0,   0,   0,   0,   0,   0,  20,   6,  -4,   6,  -3,  -1,  -1,   0,   0,  -1,   0,   0,  37, -21,  -4, -18,  17,  -1,   4,  -1,   0,   0,   1,   0,  21, -38,  25, -31, -21,   8,  -2,   5,  30,  41,  18,  37, -22,  -8,  -7,  -3, -38, -21, -10,  25,  24,   2,   3,   0,   6,   5, -20, -13,  -1,   3,   0,   1, -10,   5,   7,   0,   1,   1,  -4,   2,  -5,   1,   2,   0,   0,   0,  -1,   0 },
     {   -1, -11, -18,  12,   1,   2,  -1,   0,   1,   0,   0,   1,  -3, -24, -21,   6,   1,   3,   0,   3,   1,   1,   0,   1,   2, -31,   3, -16,  23,  -2,   4,  -2,   2,   0,   1,   0, -28, -40,  24, -45,  -9,   9,   2,   3,   0,   2,   1,   1,  -7,  27,  31,  12, -19,  -7,  -5,  -2,  43, -15,   1,  -5,   3,   4,  -1,   1,  -7,  -1,  24,  18, -19,   3,  -3,   1,   1, -10, -26,  25,   9,  -1,   2,   0,  -4,  12,  -8, -13,   3,  -1,   0,   4,  -2,   5,  -3,  -1,   1,   0,  -1,   1 },
     {   -1,   4,  -5,  -6,   6,  -1,   2,  -1,   3,  -1,   1,   0,  -5,  -1, -26,  22, -25,   6,  -2,   4,   0,   2,   0,   1,   5,   8,   5,  67,  11,  -7,  -5,  -3,  -2,  -1,  -1,  -1,  -7,  -8, -25, -15,  11,   6,   5,   2,   3,   1,   1,   0, -21, -11, -14,  23,   0,  -3,   2,   0, -14,  -7, -10, -13,  16,   1,   2,   0,   8, -14, -19,  18,   0,   1,   1,  -1,  -9,  22, -40, -27,  23,  -6,   7,  -2,   7,   6,  21, -36,   3,  -1,   9,   1,   2,   0,   7,  -5,   0,  -1,   2,   0 },
     {   -7,  -4, -11, -47,  15, -13,   3,  -1,   1,   0,   0,   0, -10,  -9, -22, -49, -17,  12,   4,   2,   1,   1,   1,   1,  -4, -11,  11,  36,  -4, -13,  -8,  -2,  -3,  -2,  -2,  -1, -10,  11,  36, -18,  11,   9,  -1,  -1,  -4,   0,  -1,   0, -14,  22, -23,   3, -18,   0,   5,   2,  22,   8,  13,   9,  12, -11,  -4,  -1, -16,  -3, -11, -21,   5,  13,   3,   1,   8,   6,  29,  -9, -16,  -3,  -4,   0,  -2, -15,   6,  25,   2,   3, -10,  -2,   0,  -3,   5,   4,   0,   1,  -3,   3 },
     {   -7,  -1,   6,  10, -23,   4,  -3,   2,   1,   1,   0,   0,  -5,   5,  19,  22,  14,   1,  -3,  -2,  -1,  -1,  -1,  -1, -11,   4,  15, -20, -11,   2,   3,   2,   0,   0,   0,  -1, -21,   5,  -9,   4, -12,  -1,  -3,   2,  -1,   2,  -1,   1, -27,  24,   5,  37, -18,  -8,  -3,  -2, -10,  36,  -5,  30,  41, -16,  -1,  -5, -22, -45,  -3, -24,  25,  24,   3,   4,  18,  -3, -16,  -9, -20,  -1,   3,   1,  -3,   4,   2,  -5,   8,  -3,   2,  -1,  -1,   1,   1,   0,   4,  -2,   0,  -1 },
     {    1,   9,   1,  -9,  10,  -6,  -1,   0,  -1,  -1,  -1,   0,  -2,  11,   0, -22, -11,   3,   0,   0,   0,   0,   0,   0,   0,  18, -11,   1, -17,   2,  -4,   3,  -1,   0,  -1,   0,   1,  21,   0,  29, -26,  -5,  -1,  -2,  -3,   0,   0,  -1,  42,  42,  -6,  23,  16,  -8,  -1,  -5,  29,  -7, -23,   7,   0,   7,   2,   3, -54,  20,   5,  14,  11, -10,   4,  -3,  -7, -23, -21, -12,  25,   3,   1,   2,  -5,   8,   6, -25,   3,  -7,   9,  -2,  -1,   1,   3,  -2,   2,  -4,   3,   0 },
     {   -8, -11, -18, -34, -31,  -3,   9,   3,   2,   3,   2,   2,   0,   1, -12,  -2,  35,   6,   1,   1,  -2,  -1,  -1,  -1,  -7, -17, -29,  18, -49,   5,  -2,   4,   3,   2,   0,   1,   1, -13,  27,  36,  32, -15,  -5,  -5,  -3,  -4,  -1,  -1, -15,   4,   7, -15,  -3,  17,   6,   3, -10,  -7,  -6,  14,   1, -14,  -1,   0,  17,   2,  11,   0,   5,   7,  -4,  -1, -14, -10, -25,  22,   5,   0,   5,   0,   2,   8, -16, -29,   0,   2,  12,  -5,   1,   3,   1,  -2,   0,   1,   4,  -1 },
     {    1,  -4,   8,  22, -60,   7,  -8,   4,  -3,   1,  -1,   1,   8,   5,  18,  37,  10,  -3,  -9,  -3,  -1,  -1,  -2,  -1,  12,  11,  25,  51,  11,  -4,  -2,  -4,   1,  -3,  -1,  -1,  15,  19,  29,  -8,  11,  -2,   7,  -3,   0,  -1,   1,  -1,  25,   9, -22, -12,   8,  -4,  -4,   0,  24, -23,   8,  -5,  13,   6,  -4,   2, -20,  -1,  -1,  -9, -16,   9,   3,   3,   9,  -8,   6,  17,  -7,  -7,  -1,  -2,  -8,   1, -18,   6,  -1,   4,  -9,  -2,   0,   2,  -3,  -6,   1,   1,   2,  -4 },
     {    4,   5,  -2,  28,  50, -14, -10,  -8,  -2,  -3,  -1,  -2,   0,  -4,   4,  34, -73,  14,  -1,   4,  -2,   3,   0,   1,   0, -11,   1,   0,  -8, -16,  -1,   0,   0,  -1,   0,   0,   4,   5,  16,  30,  31,  -3,  -2,  -5,  -3,  -1,  -1,   0,  -1,   0,   1, -11,  -5,   5,   3,   4,  -1,   0,   1,  15,  11,  -6,  -4,   0,   0, -11,   5,  -9,   5,   8,   3,   1,  -4, -23,  -4,  22,  -3,   3,  -1,   2,  -5,   5, -31,  -1,   0,   4,   0, -15,   4,   2,  -2,  -2,   0,   2,   1,  -3 },
     {   -5,   0,   3,  14, -23,   6,  -8,   4,  -2,   2,   0,  -1,  -7,   2,   7,   8, -11,   1,   3,  -3,   1,  -1,   1,  -1,  -7,   6,  17,   3,  11,   0,   0,   2,  -3,  -2,  -1,  -1, -10,  16,  13, -27,  -8,   3,   0,   3,  -2,   1,   0,   0, -24,  13, -24,   3, -31,  10,   1,   0, -17,  19,   2,  33, -33, -24,   3,  -3,  15,  42,   5,  26,  36, -20,  -5,  -5, -31, -28,   8,  -5,  24,  15,   2,   2,   9, -10, -13,  -4,  -3,  -2,   0,  -3,   6,  -3,   1,  -1,  -1,  -1,   0,  -1 },
     {   -6,  -6,  -3,  -5,  -7,  -5,  -5,   3,  -7,   1,  -3,  -1, -10, -12,  -8,  -7, -11, -18,   5,  -9,   4,  -5,   2,  -1,  -3,  -6,   1,  23,  31,  14,  -5,  -3,  -4,  -5,  -4,  -1, -10, -18,  21,  11, -20,  -8,  10,   0,  -2,   2,   1,   1, -21,   4,  18, -25,  33,  11,  -8,   1, -53,  -2, -19,   4, -32,  11,   7,   8, -37, -24,  12, -10,   5, -18,   3,   2,  29, -24,  -8,  -7,  -5,  12,  -9,   3,   2,   3,  -2,   7,   9,   2, -14,   1,   4,   7,   0,  -5,   4,   3,   3,   3 },
     {    2,   1,   6,   2,  20,  -4,   5,  -8,   6,  -2,   2,  -2,   3,   2,  15,   1,  26,  -5,   3,  -2,  -1,  -2,   2,   0,   0,  -1,  11, -28,  23, -27,  10,   1,  -1,   1,   3,   0,   0,  13,   9, -26,  56,  -8,  -4,   4,   2,   0,  -3,  -1,   4,   9, -28, -21,  -8,  26,   6,   3, -14, -35, -16, -13,  -3,  -6,   6,  -2, -24,   1,  17, -14,  19,   8,  -4,   4,   7,  -4, -16, -23,  -9,   8,   2,   3,   5,  14,  22, -16,   3,  -9,  23,   9,  -2,  -2,   0,  11,  -2,  -3,  -3,   7 },
     {    3,   3,   9,  12,   6,   5,   4,   1,  -2,  -2,   0,  -1,   1,  -2,   4,  21, -11,   0, -10,  -2,  -3,   3,  -1,   2,   3,  -4,  21,  15, -50,  12,   7,   2,  -1,  -2,   0,   0,  -2,  -8,  18, -10,  15, -35,  -1,  -2,   1,  -1,  -1,   0,   5,  29,  31,  13,  18,  18, -11,  -6,  18,   8, -15, -26, -18,  -5,  15,   0,  20, -11,  12,   8,  18, -12,  -9,   1,  -8,   8,   8, -28, -11,  14,   2,   2,  10,  -6,  29,  28, -12,  -4, -14,  23,   1,  -3,  -6,  -4,  -3,  -1,  -3,   1 },
     {    2,  -5,   2, -12, -34,  19, -16,   9, -10,   4,  -3,   1,   0,  -9,  -3, -21, -37, -24,   9, -14,   3,  -6,   4,  -1,   4,   0,  -6, -13,  27,  -9,  -1,   4,  -4,  -6,  -3,  -1,   9,  -1,  -1,  12, -11,  11,   1,   0,  -9,  -1,  -2,   1,  12,  -2,   5, -23,  12,  -8,  15,  -4,  25,  -8, -17,  -5,  12, -20, -11,   7,  38, -19,  11, -19,  19,  27,  -1,   3, -39, -15, -16, -19, -12,  14,   5,   8,  -2,   2,  20,  -2,  -1,  -4,  -2,   5,   8,   5,   4,   0,  -5,   4,   0,   8 },
     {    4,   1,  -2,   7,  47, -13,   5,  -3,  -3,  -4,   0,  -1,   3,  -3,  -8,   3,  46, -27,  -1,  -7,  -1,   0,  -2,  -1,   6,  -2,   1,  27,  -1,  15,   5,   1,  -3,   1,   0,   1,   4,  -9,  -5, -17, -23, -14,   7,   6,   0,   0,  -1,   2,   2, -18, -14,  15,  24,  -6,  -4,   0,  14, -10,  24,   2,  11,  -5,  -5,  -1,   5, -11,  -5, -13,  26,  13,   3,   4, -31, -42,   2,  -4, -14,  22,   2,   5,  20,   9, -26,   2,  -6,   6,   4, -13,   8,   6,  -2,   3,  -7,   1,   2,  -4 },
     {   -3,  -6,  -6,   5, -23, -77,  16,  -2,   3,   3,   2,   1,   3,   2,   7,  -3,   4,  66, -17,   0,  -8,  -1,  -3,  -1,   2,   2,   4,  -1,  19, -27,  24,  -7,   9,  -1,   1,  -1,  -1,  -5, -10,  -3, -18,  -1, -17,   1,   1,  -2,  -1,   0,   3,  -4,  14,  -3,  30,  10,   3,  -4,  -2,   5,  -1,  -5,  -5, -13,   4,   4,   2,   0,  -3,  -1,   4,  13,  -5,  -2, -15,  -6,   0,  -4,  -4,   0,   6,   1,   5,  -8,   1,   2,  -1,   0,  -2,  -2,   5,   3,   2,   2,  -2,   0,  -4,   1 },
   },
   {
     { -110,  43,   5,   5,   2,   2,   1,   1,   0,   1,   0,   0, -15,  -6,   7,   0,   2,   0,   1,   0,   0,   0,   0,   0,  30, -24,   3,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  19, -10,  -3,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   9,  -1,  -5,   1,  -1,   0,   0,   0,   3,   2,  -3,   0,   0,   0,   0,   0,   2,   1,  -1,   0,   0,   0,   0,   0,   1,   1,  -1,  -1,   0,   0,   0,   0,   1,   0,   0,  -1,   1,   0,   0,  -1,   1,   0,   0,   0,   1,   0,   0,   0 },
     {  -23,  -3,   5,   0,   2,   0,   1,   0,   0,   0,   0,   0, 102, -40,  -7,  -4,  -2,  -1,  -1,  -1,  -1,   0,   0,   0,  25,  22, -16,   0,  -4,   0,  -2,   0,  -1,   0,   0,   0, -32,  30,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0, -15,   5,   7,  -2,   1,  -1,   0,   0,  -8,   0,   6,   0,   0,   0,   0,   0,  -3,  -2,   2,   1,   0,   0,   0,   0,  -4,   0,   1,   1,   0,   0,   0,   0,  -2,  -1,   0,   1,  -2,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {  -31,   8,   5,   0,   1,   0,   1,   0,   0,   0,   0,   0,  25,  14, -12,   0,  -3,   0,  -1,   0,  -1,   0,   0,   0, -85,  47,  13,  -2,   2,   0,   1,   0,   0,   0,   0,   0, -20, -44,  25,   1,   5,   1,   2,   0,   1,   0,   1,   0,  28, -26,  -9,   4,  -1,   1,   0,   1,   7,   0,  -8,   1,   0,   0,   0,   0,   8,  -1,  -4,  -1,   0,   0,   0,   0,   3,   0,  -1,  -1,   0,   0,   0,   0,   4,   0,  -1,  -1,   1,   0,   0,  -1,   2,   0,   0,  -1,   1,   0,   0,   0 },
     {    2,  31, -11,  -1,  -3,  -1,  -1,   0,  -1,   0,   0,   0, -41,   7,   9,   0,   2,   0,   1,   0,   0,   0,   0,   0,  32,  33, -15,  -2,  -4,  -1,  -2,  -1,  -1,   0,   0,   0, -71,  33,  32,  -5,   5,  -2,   1,  -1,   1,  -1,   0,  -1,   1, -53,  22,   5,   2,   1,   1,   1,  13, -14,  -8,   6,  -1,   1,  -1,   1,   5,  -8,  -3,   2,   0,   1,   0,   0,   4,  -1,  -5,   1,   0,   0,   0,   0,   2,  -3,  -1,  -1,   2,   0,  -2,   0,   1,  -1,   0,   0,   1,   0,  -1,   0 },
     {   12, -27,   7,   0,   2,   0,   1,   0,   0,   0,   0,   0,  -8, -55,  16,   5,   4,   1,   1,   1,   1,   1,   1,   0,  29,  -9,  -9,   3,  -1,   1,  -1,   0,   0,   0,   0,   0, -19, -38,  10,   9,   2,   3,   1,   1,   0,   1,   0,   1,  38, -26, -47,  13,  -4,   4,  -1,   2,  -3,  50, -33, -10,   0,  -2,   0,  -1,  -4,  19,   7, -12,   1,  -2,   0,  -1,  -3,  10,   5,  -4,  -1,  -1,   0,   0,  -1,   3,   4,  -2,  -1,   4,   2,  -1,  -1,   1,   2,  -1,  -1,   3,   1,   0 },
     {  -24, -67,  25,   5,   7,   2,   3,   2,   2,   1,   1,   1,  -4, -10,   6,   0,   1,   0,   0,   0,   0,   0,   0,   0, -18, -25,   3,   8,   1,   2,   0,   1,   0,   1,   0,   0,   3,  22,  -8,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -19, -33,  27,  10,   2,   2,   0,   1,  23, -35, -41,  20,  -1,   5,   0,   2,  -1,  29, -36,  -9,   4,  -2,   2,  -1,   4,  11,  -1, -15,   2,  -2,   1,  -1,   2,   8,   3,  -7,   3,   2,   3,  -3,   0,   4,   1,  -1,   1,   1,   1,   0 },
     {   16,  40, -20,   1,  -5,   0,  -2,   0,  -1,   0,  -1,   0,  34,  41,  -8,  -8,  -3,  -3,  -1,  -2,  -1,  -1,  -1,   0,   0, -32,  10,   1,   2,   0,   1,   0,   1,   0,   0,   0,  11,   7,  -1,  -6,   0,  -1,   0,  -1,   0,   0,   0,   0, -15, -33,   5,  10,   0,   3,   0,   2,  11,  18, -50,  -8,   1,  -1,   1,  -1, -10,  39,  27, -34,  -1,  -5,   0,  -3,   6, -15,  33,  10, -10,   1,  -3,   1,  -4,  -1,  -1,  13,  -1,  -3,  -1,   3,  -2,   1,   0,   2,   0,  -2,   1,   1 },
     {  -33, -58,  10,   5,   6,   2,   3,   2,   2,   1,   1,   1, -19,  23,  -9,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,  -9,  21, -18,   3,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,  37,  -5, -13,  -1,  -3,   0,  -2,   0,  -1,   0,  -1,  -1,   0,  -1,  -3,   0,  -1,   0,   0,   7,  38,   3, -26,   0,  -5,   0,  -2,   1,  -8,  63,  -3,  -8,   0,  -3,   0,   7, -27,   2,  30,  -4,   4,  -2,   2,  -3,  -1, -11,   6,   2,  -4,  -3,   1,  -1,   0,  -1,   0,   1,  -2,  -1,   2 },
     {    3,  37,  -9,  -2,  -3,  -1,  -2,  -1,  -1,   0,   0,   0, -15, -58,  25,   0,   5,   1,   2,   1,   1,   0,   1,   0, -67, -39, -16,  15,   1,   6,   2,   3,   1,   2,   0,   1, -28,  38, -33,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,  -4,  12,  -1,  -9,   2,  -2,   2,  -1,  12,   9,   2, -10,   0,  -1,   0,  -1,   6,  -1,  14,  -3,  -1,   0,  -1,   0,   8,  -4,   0,   1,  -1,   1,   0,   0,   5,   1,   2,   0,   4,   0,   0,   0,   2,   1,   1,   0,   2,   0,   0,   0 },
     {   -7,   1,   0,  -5,   1,   0,   1,   1,   1,   0,   0,   0, -27, -20, -25,   8,   0,   3,   1,   2,   1,   1,   1,   1,   2,  47, -17, -10,  -2,  -2,  -1,  -1,  -1,   0,  -1,   0,   1,   6, -15,  -1,   1,   0,   1,   0,   0,   0,   0,   0,   7,  33,   4, -18,   0,  -4,   0,  -1, -12, -13,  -4,   5,   3,   1,   2,   1,   4,  32, -15, -42,   4,  -6,   1,  -2,   9,   1,  67, -17, -17,  -1,  -5,  -1,   7, -15,  11,  30,  -4,  -1,   0,   9,   1,  -4,  -2,   6,  -1,   0,   1,   0 },
     {   12,  22,   0,  -4,  -1,   0,  -1,  -1,   0,  -1,   0,   0,  -4, -49,   5,   6,   2,   2,   1,   2,   1,   1,   0,   1,  -8,  38, -22,  -4,  -4,  -1,  -1,  -1,   0,   0,   0,   0,  85,  21,  30, -16,  -4,  -6,  -3,  -3,  -1,  -2,  -1,  -1,   4, -31,  22,  14,  -1,   3,  -1,   1,   2,   1, -11,   7,  -1,   1,   0,   1,  -6,  -2,   7,   7,  -2,   1,  -1,   0,  -4,  -3, -11,  12,   1,   1,   1,   1,  -9,   4,  -9,  -1,  -2,   0,  -2,  -2,  -2,   1,  -1,  -1,   0,   0,   0,  -1 },
     {   -8,  -1, -39,  10,   1,   3,   1,   1,   0,   0,   0,   0,   1,  26, -19,   4,  -3,   0,  -1,   0,  -1,   0,  -1,   0,  -9,   9, -41,   6,   2,   1,   1,   1,   1,   1,   0,   1,  14,  27,  -2,  -5,  -1,  -2,   0,  -2,   0,  -1,   0,  -1, -10,  -6, -28,   5,   6,   2,   3,   1,   2,  32,  -4, -16,   1,  -4,   0,  -2, -18,  -5, -20,   7,  12,   0,   4,   0,  -2,  26, -14, -54,  11,  -5,   3,  -2,   7,  -2,  52, -22,   5,  -5,  10,  11,   0,  -2,  10,   2,   1,  -2,   1,   3 },
     {   -3,  12,   7,   0,   0,  -1,  -1,   0,   0,   0,   0,   0, -19, -18, -26,  13,   4,   3,   0,   2,   1,   1,   0,   0,  -1,  27,  12,   0,   0,  -3,   0,  -2,   0,  -1,   0,  -1,  -7, -18,  -3,  19,   3,   4,  -1,   1,   0,   1,   0,   0, -86,  10, -28,  15,   5,   4,   3,   1, -11,  -6, -24, -14,  10,  -3,   4,  -1,   8,  11,  10, -11,  -2,  -2,   0,  -1,  -7,  -9, -15,  11,  13,   0,   3,   0,   0,   5, -13, -37,  10,  -4,  22, -13,   4,   0,   2,   1,   3,   0,   6,   0 },
     {    2,   3, -27,  12,  -4,   2,  -1,   1,  -1,   1,  -1,   0,   0,   4, -40,   8,   2,   2,   1,   1,   0,   0,   0,   1,   6,   2, -15,  11,  -2,   1,  -1,   0,   0,   0,   0,   0,  -6,   3, -35,  -8,   6,  -2,   3,  -1,   1,  -1,   1,   0,  59,  15,  26,  -7,  -4,  -4,  -2,  -2, -12, -11, -24,  15,   9,   3,   3,   1,  10,  20,   4,  -8,  -3,  -3,   0,  -1, -20,  -5, -23,  20,  17,   0,   4,   0,  -6,  16, -19, -48,   3,  -1,  30, -11,   1,   0,   1,   0,  -1,  -1,   8,   0 },
     {  -12, -30, -93,  23,  -2,   9,   3,   5,   2,   3,   0,   2,  -2, -25, -23,   8,   1,   2,   2,   2,   1,   1,   1,   1,  -4, -25,   5, -16,   7,  -2,   3,  -1,   1,   0,   1,   0,  -2, -16,  29,  -8,   1,  -2,   0,  -1,   0,   0,   0,   0, -17,  -8,   5,   1,   1,   2,  -1,   1,  -9, -11,  12,  -3,  -2,   1,   0,   0,  20, -11,  -1,  -3,  -5,   1,   0,   1,   9,   4,   8,  10, -10,   0,  -3,   0,  -1,  -1, -15,  19,  -1,   8, -11, -10,   2,   3,   2,  -2,   0,   3,  -1,  -1 },
     {   -8,  -5,  -2,   2,   1,   2,   0,   1,   0,   1,   0,   1,  -4,  10,  34,  -5,   3,  -4,  -1,  -2,   0,   0,   0,   0, -14,  -3,   6,   6,   1,   1,  -1,   1,  -1,   1,   0,   0,  -6,  17,  14,  -1,   0,  -2,  -2,  -1,  -1,  -1,   0,  -1,   2,  10,   9,  26,  -6,   4,  -3,   2, -96,  13, -24,  25,  15,   2,   5,   1, -21, -25,  -9, -18,  16,  -2,   6,   0,   8, -15,   2,  -7,  -1,   1,  -1,   1,   4,  -6,  -9,  -4,   3,  -1,  -2,  -9,   6,  -3,   2,  -4,   3,   1,  -2,  -2 },
     {   -4,   0, -20,  10,   0,   1,   1,   1,  -1,   0,   0,   0,   1,  12,  19,  19,  -5,   1,  -2,  -2,  -1,   0,  -1,   0,  -2,  18,  -4,  10,  -3,   1,  -2,  -1,   0,   0,   0,  -1,   1,  13, -15,  31,  -3,   1,  -3,  -1,  -1,   0,   0,   0,  -2,   2, -22, -10,   6,  -2,   3,  -1,  30,  21,  10,  26,  -8,   0,  -3,  -1, -41,  14, -27,  13,  21,   0,   6,   0,  -2, -13,  11,   4,   4,  -1,   2,   0, -20,  -8, -48,  10,  -1,   6,  -4, -59,   4,  -5,  17,  -5,   1,   0,   1,  -6 },
     {   -5,  -9, -42,  11,   0,   5,   1,   2,   2,   1,   1,   1,  -4, -14,  42, -30,   3,  -2,   1,  -2,   1,  -1,   0,  -1,   5,  36,  71, -10,  -8,  -5,  -5,  -3,  -2,  -2,  -1,  -1,   2,   9, -14,   8,   0,   4,  -2,   0,  -1,   0,   0,   0,   3,  14,  13,  -3,  -6,   0,  -1,   0,  18,   8,  -6,   4,  -2,   1,  -1,   0, -22,  19,  10,   6,  -3,  -1,   1,  -1, -13, -14,  -5,  -7,  11,  -1,   2,   0,   9,  -2,  22, -15,   3,  -8,   7,  31,  -2,   0, -11,   2,   0,  -4,   0,   4 },
     {    0,  -2, -19,  13,   0,   2,   0,   2,   0,   0,   0,   0,  11,  23,  69,  -4,  -7,  -3,  -5,  -3,  -3,  -1,  -2,  -1,   4,  25, -26,  26,  -7,   0,  -1,   1,  -1,   0,  -1,   0,  13,  -2, -19,   7,  -1,   1,  -1,   1,  -1,   0,  -1,   1,  -2,   3, -19,   9,   5,  -1,   1,  -1,  -5, -23, -10, -18,   7,  -3,   2,   1,  69,   8,  -2,  -7, -12,  -1,  -4,  -2,  -7,  26,   4,  12,   1,  -1,   0,  -1,  -7,  -2,  -4,   3,  -7,   3,   0,  -5,   1,  -2,   3,  -3,  -4,   2,   0,   0 },
     {    4,   6,  16,  20,  -7,  -2,  -1,   0,  -1,  -1,  -1,   0,  -1,   0, -31,  13,  -2,   2,   1,   1,  -1,   1,   0,   0,  -3,  -7,  54, -10,  -6,  -3,  -4,  -1,  -2,  -1,   0,   0,   5,  65,  29,  31, -10,  -4,  -8,  -5,  -3,  -2,  -1,  -1,  28,  11, -35,  20,   2,   4,   0,   1,   1,   4,  -3,  -5,   4,  -2,   3,  -2,  42,   1, -11,  -4,   3,   1,  -1,   0,  -5,  17,  -2,   1,  -2,  -1,   0,   0,  -7,   3,  -1,   7,  -6,   0,  -3,  -6,  -2,  -1,   2,  -3,  -3,  -2,   1,   0 },
     {    6,   5,   5,  51, -13,   4,  -6,   0,  -3,   1,  -2,   0,   0,  -9, -21,  29,  -5,   3,  -3,   0,  -3,   1,  -1,   1,  11,  17,  14,  47,  -4,   1,  -2,  -1,  -3,   0,  -1,   0,   5, -13, -45,  14,  10,   5,   3,   2,   1,   0,   0,   0,   0, -19,   7,   9,   3,  -2,   2,  -1,  -1,  -6, -12,   3,   5,  -2,   2,   1,  -3, -29,   9,   0,   8,  -1,   0,  -1,  24, -16, -15, -13,  -4,   4,   0,   2,  11,   7,  10,  27,  -5,   0, -39,  16,  -3,  10,  -6, -29,   2,  -1,   1,  -2 },
     {    3,  -3,  -5, -20,   6,  -3,   1,   1,   1,  -1,   1,   0,   7,  -1,   4, -11,  14,  -2,   5,   1,   0,  -1,   0,  -1,   8,  10,  21, -13,   7,  -2,  -1,  -1,  -1,   1,  -1,   0,  13,   6, -29,   6,  10,   0,  -1,   0,   0,   0,   0,   0,   4, -28, -11, -31,   2,   2,  -1,   1,   8,   3,   1,   2,   4,   0,  -2,  -1,   9, -51,  -7, -39,   5,   5,   0,   3,  56,   9,  -2,  -3, -26,   4,  -8,   1, -15,  23,  -4, -17,  10,   5,  33,  -9,  -4,   0,   5,  26,   1,   4,  -2,   4 },
     {    6,   5,  25,  89,  -5,   4, -14,  -2,  -8,  -1,  -3,   0,   4,   3,  26,  40,   3,  -3,  -8,  -1,  -4,  -1,   0,   0,  -4,  -5,  -3, -33,   7,   2,   5,  -2,   3,  -1,   2,  -1,  -3,  -5,  16, -21,   9,  -1,   3,   0,   1,   0,   1,   0,  -6,   3,  12, -25,   1,  -1,   1,  -1,   0,  10,   8,  -1,  -7,   0,  -2,   0,  -7,   4,   1,  -9,  -5,   0,   2,   1,  -2,   7,   7,   0,  -6,  -2,  -2,   0,  -9,  -1,   3, -13,  10,  -6,  24,   2,  -1,  -4,  -6,  24,   0,   3,  -6,   2 },
     {    4,   9,   8,  27,  11, -13,  -1,  -3,  -1,   0,   0,   0,  -7,  -4, -11, -54,  10,   0,  10,   1,   1,   1,   1,   0,   2,  12,   2,  -6,   8,  -4,   3,  -1,   1,  -1,   0,  -1, -10, -11,  -5, -40,  -6,  11,   2,   4,  -1,   0,   0,   2,  -8,  35,  21,  34, -16,  -4,  -8,   0,  19,  26, -20,   3,   0,   4,  -1,   0,  11,  -4,  -9,   7,   6,  -5,  -1,   2,  25,  25,  -8,  -5,   2,   2,  -1,   2, -32,  12,   1,  16,  -8,  -5,  -9, -30,   1,  -3,  23,  -9,   3,  -3,  -3,  10 },
     {   -1,   1,   9,   5,   0,   1,  -2,  -1,  -1,   0,   0,   0,  -6,  -5, -21, -26,  11,   1,   1,  -1,   1,   0,   1,   0,  -4,   2,  25,  12,  -1,   1,   1,   0,  -1,   0,   0,   0,   4,   4, -34,  -9,  12,   4,   1,   0,  -1,   0,  -1,   0, -11, -51,   9, -29,   2,   0,   2,   4, -41,  23,  23,  -3, -14,   3,  -1,   2,  13,  11,  -6,  15, -10,  -2,  -2,  -1, -34,  34,  12,   9,   5,  -4,   1,  -2, -19, -44,   3,   2,  11,   3,  -7, -14,   4,  -4,   7,   2,   4,   0,  -4,   5 },
     {    7,   2,   7,  21,  26,  -5,   3,  -4,   1,  -3,   0,   0,   3,  -6,   1, -29,  10,  -4,   1,   3,   0,   0,  -1,   0,   6,  -1,   6,  -6,  26,  -7,   2,  -2,   1,  -2,  -1,  -1,   7,   5,   0, -23,   9,  -1,  -3,  -1,   2,  -2,   0,   1,  18, -11, -23,  -2,  26,   5,   1,  -1,  -3, -40,   8, -40,   9,   6,   1,   1,  -5,   6,   8,  -4,   9,   2,  -3,  -1,  -8, -42,   1, -33,   5,  13,   3,   4,  35,  -9, -11,  -8,  -1,  14,  -5, -31,  12,   4,  30,  -3,  -4,   0,  -1,  16 },
     {    0,   2,  -3, -16,  23,  16,  13,   2,  -1,   2,   1,   2,   1,  -1,   5,  33,  14,   7,   0,  -3,  -2,   0,  -2,   0,   5,   2,  26,  70,  33,  -5,  -5,  -6,  -3,  -2,  -5,  -3,  -2,   7,  33, -39,  24,  -1,  -4,  -3,  -1,  -2,  -1,  -1,   2,  24,  -7, -22, -20,   2,  -2,   5,  11, -10, -14, -10,  -1,   0,   5,   3, -17,  -4,   4,  -8, -10,   0,  -5,   3,  -1,  21,  -1,   4,  -6,  -5,   0,   0, -13, -15,   4,  -2,   4,  -4,   1,  -6,   1,   0,   4,   1,   2,  -2,  -6,  -1 },
     {    1,   1,  -1, -23,  21,   2,   1,  -4,   1,   1,   0,   0,   3,   3,   7,  34,   4,  -3,  -8,   4,  -4,   4,   0,   0,   3,   2,  11, -18,  26,   9,   1,  -2,  -1,   0,  -1,   0,   0,   2, -11, -31,   8,   5,   6,   5,   1,   1,   0,   0, -14,  -6,   6, -12, -23,   3,   7,   3,   4,  39,   5,  31, -11,  -5,  -3,   1,  57,   6,  -8,  -3,  26,  -8,   6,  -4, -24, -35, -11, -26,   3,  11,   3,   6,  15,  17,  -3,   6,  -9,   2,  -7,   1,   2,  -5,   0, -15,  -4,  -4,   3,  -1 },
     {    3,  -1,   3,  18,  15, -12,  -2,  -1,  -1,  -3,   0,  -2,   1,  -4,  -7, -26,  12,  -4,   8,  -5,   2,   0,  -2,  -1,   2, -12,  -8,  51, -19,   2,  -8,   1,  -3,   4,  -1,   0,   7, -13,  34,  37,   5,  -1,  -3,  -7,  -5,  -1,  -1,  -1,  -6,   3,  34,   3,  -2, -13,  -5,  -1,  22,  26,  25,  15,  17, -11,  -3,  -5,  25, -11,   2, -26,  21,   3,   0,   2, -10, -16,  17, -13,   6,  -1,  -3,   3,  20, -19,  -3, -18,   3,   5,  17,   1,   3,  -7,   1,  17,  -2,   4,  -2,   5 },
     {   -1,  -1,  -2,   0, -14,  -9, -13,  -6,   1,  -1,  -1,  -1,  -2,  -4,  -4,  -6, -22,  -4,  -1,   1,   1,  -1,  -2,   3,   4,   1,   5,  34,   8,   7,  -4,  -4,  -3,  -4,  -2,   1,   1,   1,  15,  -1, -10,  15,  -1,   2,   0,   2,   2,   0,   1,  -6,  -5,  -8,   1,  -4,   7,   2, -28,  -4,  19,   3, -37,  -8,  -7,   0,  13,  47,   5,  33,   3, -20,  -5, -12,  59, -20,  -4, -12,  12,   0,   4,  -1,  -6,  22,   1,   1,  -4, -15,  18,   0,   4,  -1,   1,  31,  -3,  -1, -13,   2 },
     {    1,   5,  -4,  10,  56, -11, -12,  -8,  -2,  -8,  -7,  -3,   1,   6,  -4,   1,  26,   3,   6,   3,  -3,  -5, -13,  -3,  -2,   2,  -5,  -7,  -5,  27, -21,  -9, -11,  10, -17,   1,  -1,   4,   0,   5, -26,  23,  -1,  -1, -11,   3, -18,  -6,   0,   1,  -1, -10,   2,  -8,  -7,   2,  -9,   5, -13, -19, -12, -23,  -6,  11,  -8,   7, -25,  10, -13, -12, -15,  10,  15,  -3, -13,  14,  -8,   5,  -9,   2,   9, -13, -13,   1, -10,  35,   6,  22, -27, -11, -17, -17,  12,  11,   3, -15 },
     {    1,   2,   4,  -5, -18,  16,  34,  20, -16,  25,   2,  -7,   1,   3,   1,   1,  24, -19, -21,  26, -25,   6,  13,  -7,  -1,   0,   2,  -5, -17,   1,  14,  -6, -11,  14,  -1,  -3,   4,   2,   7,  17,  43, -11,   8,  13,  -4,  -6,  10,   2,  -2,   7,  19,  -3,   3,  -9,  -8,  -5,  -4,  13,  -8, -22,  -5,  10,   4,  22,   2,   9,   2,  23,  -6,  -4,  -7,   6,  19,   9,   3, -19,   2,   4,  12,  -2,   9,   9, -14,   6, -18,  32,  18,  -5,  15, -16, -14, -15,  -1,   3,   2,   8 },
   },
   {
     {  -99,  62,  -2,   2,   0,   1,   0,   0,   0,   0,   0,   0, -18,   1,  10,  -3,   2,  -1,   1,   0,   0,   0,   0,   0,  34, -28,   5,   0,   0,   0,   0,   0,   0,   0,   0,   0,  14,  -7,  -3,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   7,  -3,  -2,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   25,  -4,  -1,  -4,   0,  -1,   0,  -1,   0,   0,   0,   0, -99,  54,   4,   0,   2,   0,   1,   0,   0,   0,   0,   0, -13, -14,  15,  -3,   2,   0,   1,   0,   1,   0,   0,   0,  35, -28,   2,   2,  -1,   1,   0,   0,   0,   0,   0,   0,  12,  -4,  -6,   2,  -1,   0,   0,   0,   7,  -1,  -3,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   2,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   30,  70, -71,  13, -10,   2,  -4,   1,  -2,   0,  -1,   0,  25,  37, -28,  -5,  -1,  -3,   0,  -1,   0,  -1,   0,   0, -19,   1,  25, -13,   4,  -2,   1,  -1,   1,   0,   0,   0, -21,  -7,  19,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -5, -11,   9,   1,   1,   0,   0,   0,   1,  -8,   5,   0,   1,   0,   0,   0,   1,  -5,   4,   0,   0,   0,   0,   0,   0,  -4,   2,   1,   0,   0,   0,   0,   0,  -3,   1,   1,   0,  -2,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,   0 },
     {   47,   8, -27,   5,  -4,   0,  -1,   0,  -1,   0,   0,   0, -15, -22,  18,  -1,   4,   0,   1,   0,   1,   0,   0,   0,  73, -60,   2,   5,  -4,   2,  -2,   1,  -1,   1,   0,   0,  -1,  28, -26,   8,  -4,   1,  -1,   0,   0,   0,   0,   0, -26,  26,  -3,  -5,   3,  -2,   1,  -1, -10,   6,   6,  -4,   1,  -1,   0,   0,  -6,   1,   4,  -1,   0,   0,   0,   0,  -3,   2,   1,   0,   0,   0,   0,   0,  -3,   1,   1,  -1,  -1,   1,   1,   0,  -1,   0,   1,   0,  -1,   0,   0,   0 },
     {  -12, -17,   3,   5,   3,   1,   1,   0,   1,   0,   0,   0, -12,  48, -54,  18,  -6,   3,  -2,   1,  -1,   1,  -1,   0,  58,  36, -42,  -8,   0,  -4,   0,  -2,   0,  -1,   0,  -1,  -4,  28,  16, -22,   5,  -4,   0,  -2,   0,  -1,   0,  -1, -21,  -1,  28,  -8,   0,   0,   0,   0,  -5, -14,  16,   2,  -1,   0,   0,   0,  -1, -10,   6,   2,   0,   0,   0,   0,   0,  -6,   3,   1,   0,   0,   0,   0,   0,  -4,   1,   1,   0,  -2,   1,   1,   0,  -2,   0,   1,   0,  -2,   1,   0 },
     {   23,  37,  49, -72,  18, -15,   5,  -6,   2,  -3,   1,  -2,  14,  35,  28, -38,   0,  -5,  -1,  -2,  -1,  -1,   0,  -1,  15,   2,  -5,  18, -14,   4,  -4,   2,  -2,   0,  -1,   0, -23,   1,   5,  12,  -2,  -1,   0,   0,   0,   0,   0,   0, -12, -16,  11,   5,   2,   0,   1,   0,   1, -14,   1,   6,  -1,   1,  -1,   1,   2,  -7,  -2,   6,  -1,   1,  -1,   0,   0,  -3,  -3,   5,  -1,   1,   0,   0,  -1,  -2,  -4,   4,  -1,   0,  -3,   2,   0,  -1,  -2,   1,   0,   0,  -2,   1 },
     {   -1, -13,  16, -12,   7,  -4,   2,  -1,   1,   0,   0,   0,  52,  33, -40,   3,  -7,   0,  -3,   0,  -1,   0,  -1,   0,  12, -29,  15,   3,   0,   0,   0,   0,   0,   0,   0,   0,  54, -55,   0,   9,  -3,   3,  -2,   2,  -1,   1,   0,   1, -15,  34, -23,   6,  -1,   0,   0,   0, -23,  16,   3,  -6,   4,  -2,   1,  -1,  -8,   3,   5,  -3,   0,   0,   0,   0,  -4,   0,   3,   0,   0,   0,   0,   0,  -4,   2,   0,   0,  -2,   0,   1,   0,  -1,   1,   0,  -1,  -1,   0,   1,   0 },
     {   -7, -37,   9,  11,   1,   2,   1,   1,   0,   1,   0,   0,   2,   9,  -6,   1,   0,  -1,   1,   0,   0,   0,   0,   0,  38, -25,  45, -26,   4,  -4,   1,  -2,   0,  -1,   0,  -1, -56, -21,  46,   5,  -4,   5,  -2,   1,  -1,   1,   0,   0,   0, -35, -12,  27,  -6,   5,  -2,   2,  12,   3, -32,  10,   2,   0,   0,   0,   4,  14, -15,  -3,   3,  -1,   1,   0,  -1,   9,  -4,  -3,   0,   0,   0,   0,   0,   4,  -2,  -3,   0,   3,   0,  -1,   0,   2,  -1,  -1,   0,   2,  -1,   0 },
     {   11,  22,  -7, -22,  19, -11,   5,  -4,   2,  -2,   1,  -1, -16, -22, -38,  42,  -7,   9,  -3,   4,  -1,   2,   0,   1,   1, -13, -46,  46, -13,  10,  -4,   4,  -1,   2,   0,   1, -30, -27,   7,   2,  13,  -2,   4,   0,   2,   0,   1,   0,  22, -18, -23,  -3,   7,  -1,   1,   1,   4,  32, -22, -14,   3,  -2,   1,  -1,  -1,  15,   4, -15,   1,  -2,   1,  -1,  -1,   6,  11,  -9,   0,  -1,   0,  -1,   1,  -1,   8,  -5,   1,   0,   4,  -3,   1,  -1,   3,  -2,   0,   1,   2,  -1 },
     {  -26, -24,   5,  -2,  15,  -5,   5,  -1,   2,  -1,   1,   0,  -1,  34, -27,   8,  -3,   2,  -2,   0,  -1,   0,  -1,   0, -42, -46,  20,  20,  -2,   6,   0,   2,   1,   1,   0,   1, -47,  32, -37,  11,   2,  -1,   2,   0,   1,   0,   0,   0, -21,  26,  -5, -17,  11,  -6,   3,  -2,  26, -21,  12, -14,   1,   1,   0,   0,  15,  -2,   2,  -2,  -4,   1,   0,   0,   6,  -2,   3,  -2,   0,  -1,   0,   0,   3,   0,   2,  -1,   3,  -2,   2,   0,   2,   0,   1,   0,   2,  -1,   1,   0 },
     {   11,  16,  26,   3, -24,   6,  -5,   0,  -2,   0,  -1,   0,   4, -16,   0,  40, -26,   9,  -5,   3,  -2,   1,  -1,   1, -21, -40,  -3,  19,   9,  -4,   4,   0,   1,   0,   1,   0,  23,  -3,  26, -31,  15,  -6,   1,  -1,   0,  -1,   0,   0, -35,  -8,  43,  -2,  -4,   6,  -3,   1,   8, -32, -10,  28,  -7,   4,  -1,   1,   2,   4, -32,  11,   4,  -2,   2,   0,   2,  10, -14,  -8,   5,  -2,   1,  -1,  -1,   8,  -3,  -6,   2,   2,  -1,  -3,   0,   3,   0,  -2,   1,   1,   0,  -1 },
     {    8,  13,  34,  35, -64,  26, -17,   7,  -7,   3,  -3,   1,  12,  38,  28,  16, -37,   5,  -6,  -1,  -3,   0,  -1,   0,  10,  10, -12,  13,  13, -14,   6,  -4,   2,  -2,   1,  -1, -21,  -6, -32,  20,  15,  -3,   2,   0,   0,   0,   1,   0,   7, -11, -25,  -2,  11,  -1,   1,   0,  -5,  10,  -5, -13,   5,  -1,   1,  -1,  -2,   1,   7,  -8,   1,  -1,   0,   0,  -1,  -3,   3,   0,   1,  -1,   0,   0,  -1,  -2,   0,   0,  -1,  -1,   0,  -1,   0,  -2,   1,  -1,  -1,  -1,   0,  -1 },
     {   -4, -23, -24,   8,   7,   4,  -1,   3,   0,   1,   0,   1,  26,  32,   4, -27,   6,  -5,  -1,  -2,   0,  -1,   0,  -1,  -2, -28, -22,  19,  -5,   7,  -1,   2,   0,   1,   0,   1,  34,  24, -28,   0,  -6,  -2,  -1,  -1,   0,  -1,   0,   0,  25, -57,  30,   2,   1,   2,   0,   1,  11,   3, -35,   9,  -5,   2,  -1,   2, -22,  34,  -1, -10,   5,  -4,   2,  -2,  -4,   1,   7,  -7,   1,   0,   1,   1,  -5,   2,   6,  -2,  -1,   0,   2,   1,  -3,   2,   3,   0,   0,  -1,   1,   1 },
     {  -12,  -6, -25,  16,   1,   3,   1,   2,  -1,   2,   0,   1, -10,  -1,  13, -12,   6,  -2,   1,   0,   0,   1,   0,   0,  -4,  33, -18,  26, -25,  10,  -7,   2,  -1,   1,  -1,   0,   5, -19,  -1,  42, -13,   0,  -2,   0,   0,   0,  -1,   0, -73,   1,  -9,  28,  15,  -8,   5,  -1,  -7, -23, -25,   4,  18,  -3,   3,   0,   9,   5, -23, -16,  10,   0,   1,   0,   7,   8,   1, -18,   2,   0,   1,  -1,   4,   2,   8,  -9,   4,  -1,   5,  -4,   2,   0,   3,  -2,   2,  -1,   2,  -1 },
     {   12,  31,  31, -15,  -3,  -6,   2,  -4,   0,  -2,   0,  -1,  -5, -10, -32,  14,  18,  -9,   6,  -1,   2,   0,   1,   0,  -9,   6,  -2, -42,  39, -15,   7,  -3,   3,  -1,   1,  -1,  16,  38, -24,   6, -14,  10,  -6,   2,  -3,   1,  -2,   0, -14,   1, -27,  35, -15,   6,   0,   1,  18, -11, -29,   2,  11,  -6,   3,  -1,  -7,  24,  -6, -22,   9,  -3,   0,   0,  -3,   2,  22, -11,  -5,   2,  -1,   0,   1,  -4,   9,   3,   0,  -3,   4,   1,  -1,  -2,   2,   0,   0,  -2,   2,   0 },
     {    0,  -4,   4,  10, -10,   5,  -4,   4,  -4,   2,  -2,   1,  -3,  -6,  -4, -36,  46, -23,  13,  -5,   5,  -2,   1,  -1,  -6, -21, -50, -12,  51, -13,   7,  -2,   5,   0,   1,   0, -23, -50,  -8,  -2,  10,  13,  -6,   6,  -1,   3,   0,   2, -22,  -4,  17, -25,  -5,  11,  -1,   1,   2,   0,  13,  -8, -14,   5,   0,  -1,  14,   5,  -2,   2,  -9,   0,  -1,   2,   6,   4,   4,   1,  -5,   1,  -1,   0,   3,   3,   4,   1,   2,   1,   3,   2,   1,   1,   1,   2,   0,   0,  -1,   3 },
     {    9,   2,   0,  -2,  -3,   1,   0,   0,   0,   1,   0,   0,   2, -13, -27,   4,   3,   1,   4,   1,   1,   0,   0,   0,  23,  15,   9,   0,  -7,  -3,  -2,   0,  -1,   0,   0,   0,  -1, -23, -33,  40,  -4,   3,  -1,   2,   0,   1,   0,   1,  48,  -2,   3, -14,   4,  -6,  -1,  -1,  16, -54,  30,  16,  -2,   7,  -5,   3,  -4,   7, -47,  16,   1,   3,   1,   0, -19,  27,  -8, -15,   7,  -5,   3,  -2,  -2,   5,   4,  -9,  -5,   4,   5,  -3,   0,   1,   1,  -2,  -2,   2,   3,  -1 },
     {  -11, -12, -16, -50, -35,  57, -12,  12,  -3,   5,  -1,   2, -15, -17, -33, -41, -21,  36,   0,   7,   1,   3,   1,   1, -11, -13,  -8,  -1,   3, -15,  11,   0,   3,  -1,   1,   0,  -8,   8,   7,  19,  -8, -14,   2,   0,  -1,   0,   0,  -1,   7,  11,  12,  21,  -8,  -6,  -3,  -1, -17,  15,   5,  15,   4,  -5,  -1,  -2,  -2,  -2,   0,   1,  10,  -3,   1,  -2,   6,  -3,   1,  -3,   3,  -1,   0,  -1,   4,   0,   0,   0,   1,   1,   2,   1,   2,   0,   0,   1,   0,   0,   1,   1 },
     {    2,  -4,  -9, -21,  -4,  17,  -3,   2,  -1,   2,   0,   1,   1, -18,  -1, -13, -17,  18,  -2,   3,  -1,   1,   0,   1,  16,  18,  22,  -6,  -7,  -7,   3,  -3,   0,  -2,   0,  -1,  20, -15,   3, -11,  27, -17,   5,  -2,   1,  -1,   1,   0, -25, -10,  -7, -29,  25,  -3,   0,   2,  58, -18,  -6, -31, -11,  10,  -4,   2,   9,  34,  20, -12, -23,   4,  -2,  -1,  -8,   4,  27,  15, -14,  -1,  -1,   0,  -6,  -2,   2,  19,  -3,  -1,  -4,  10,  -3,   1,  -4,   4,  -1,   0,  -1,   2 },
     {    1,  -6,  -8,  -7,   2,   2,   4,  -2,   0,   1,   1,   0,   4, -10,  -8,  11,  -7,   6,   0,   0,   0,   0,   0,   1,  -2, -12, -15,  -6,  12,  -3,   4,   2,   0,   0,   0,   0,  18,   5,  -4,  20, -11,   5,  -7,   4,  -1,   1,  -1,   0, -15, -51, -19,  32,   0,   2,   0,   3,  24, -10,  23, -34,   6,  -5,   2,  -1,   7, -36,  32,  27, -14,   9,  -5,   3,  -6,   8, -49,  22,   8,  -2,   5,   0,  -9,  16, -12, -18,   3,   5,   1, -10,  -4,   5,   4,  -3,   1,   2,  -1,  -2 },
     {   -9, -15, -40, -32,  37,   2,  -1,   4,   2,   2,   0,   1,   7,  15,  34,  14, -20,   6, -10,   4,  -4,   2,  -2,   0,   5,   9,  -3,  -2,  37, -31,  19,  -8,   4,  -3,   2,  -2,  -4,  -3, -21, -24,  42, -11,   5,   1,   1,   0,   0,   0,  -3,   2, -28,   8, -13,  24, -10,   6, -16, -13,  -4,   6,  -6,   3,   3,  -1,  -2, -13, -18,   2,   2,  -3,   3,   1,   0,  14,  -4, -23,   3,  -3,   0,   0,   3,  -2,  15, -10,   4,  -1,   5,   1,   1,  -2,   3,   2,   1,   1,   0,  -1 },
     {   -2,   3,  11,  -6,  14,  -8,   0,   0,   1,  -1,   1,  -1, -13, -18, -19,  18,  -1,   2,   0,   3,   1,   0,   0,   1,  -2,  11,  41, -11,   2, -11,   5,  -4,   0,  -1,   0,   0,  -4,  -7, -39,  20,  14,  -6,   3,  -1,   0,   1,   0,   1, -26, -49,  31, -16,  14,   5,  -2,   3, -61,  15,  10,  -6,  -3,   9,  -1,   1, -15,  14,   2,  -5,  -6,   1,   1,   0,  26, -16,   1,   8,  -3,   0,  -1,   1,  12,   2, -11,   1,   4,   3,   1,  -2,   2,   3,  -2,   1,   2,   0,   0,   0 },
     {    2,  -1,   2,  -2,  12,  20, -27,  14,  -6,   4,  -3,   1,   3,   0,  -1,  18,  39, -31,   3,  -6,   3,  -3,   1,  -1,   7,   6,  22,  58,  17, -51,  19, -14,   4,  -4,   1,  -2,  12,  20,  33,  27,  -5, -14,  -5,  -1,  -4,   0,  -2,  -1,  15,  10,  11,  -5,   6,   6,  -8,   3,  -1,  -5,  -9, -14,  11,   4,  -1,   0,  10,  -7,   3, -11,  -5,   7,  -1,   2,  -7, -10,   0,   8,  -7,   4,   1,   2,  -4,   2, -10,   0,  -3,  -1,  -1,  -4,   0,  -2,  -1,  -5,   0,  -2,   0,  -2 },
     {    1,  10,   6,   4,  -1,   9,  -4,  -2,   1,  -2,  -1,  -1,  -5,   1, -20, -15,  17,  -5,   9,  -3,   5,  -2,   2,  -1,   4,  22,  16,  13,  -4,  -2,   0,  -2,  -3,  -1,   0,  -1,  -2,   5, -17,  -6,  19,  -4,   6,   2,  -4,   1,   0,   1, -20,  14, -17, -14,   2,   7,  -1,   3,  27,  36, -32,  27, -32,  12,  -3,  -2, -18,   7,  -4,  44,   3, -10,   6,  -4, -10, -17, -33,   9,  35,  -8,   5,  -1,  -5,   1,  -7, -37,   5,  -3,  12,  -9,  -1,  -1,   5,  -4,   1,  -4,   5,  -1 },
     {    1,   1,  -8,   0,  19,  10, -13,   8,  -4,   3,  -2,   1,   1,  -2,   7,  18, -10,  14, -16,   7,  -6,   2,  -2,   1,   9,  13,   2,  11,   0,   2,   2,  -5,   1,  -2,   0,  -1, -14, -31, -36,  -4, -20,  25,  -3,   4,   1,   3,   0,   1,   2,  16,  28,  15, -51,  11,  -3,  -2,  15, -19, -15,  22,  -3, -17,  12,  -3,  19,  15,  31, -11,   4, -11,  -3,   0,  -2, -23,   4,  36,   5,   4,  -3,   2,  -6,   8, -31,   1,  -3,   1,   1,  -7,  -1,   2,  -4,  -5,  -2,   1,   2,  -1 },
     {    5,   1,   6,  21,  -5, -20,  24, -15,   7,  -4,   2,  -2,  -2, -10, -17, -29,   4,  -1,   8,   0,   2,   1,   1,   1,   4,  -7,  -1,   5,   8,   0, -11,   8,  -3,   2,   0,   1,   0,  -5, -10,  -2,  58, -46,  13,  -4,   3,  -1,   2,   0,  19,   9,   3,  35,  11,  -9,  -7,   1, -10, -27, -20,  23,  16,   7,  -7,   4,  16, -19,  25, -22,   1,   9,  -6,   1,  -3, -10,  -3,  15,  -9,   9,   2,   0,  -3,   9, -20,  -6,  -2,   2,   2, -10,   0,   1,   1,  -6,   0,   2,   1,  -1 },
     {   -9, -11, -20, -18, -30, -16,  58, -30,  15,  -5,   7,  -3,  -3,   0,   7,   9, -31, -20,  34, -13,   6,  -4,   4,  -2,   3,  10,  10,  22,  24, -21, -27,  18,  -8,   4,  -4,   2,   1,   3,   8,  23,   7,  17, -34,  12,  -5,   4,  -4,   1,  -2,   7,  16,  -4, -16,  12,  -4,  -3,  10,  13,   9,  -1,  -3,  -7,   4,  -1,  -9,  13,   0,   0,   6,  -4,   0,   3,  -5,  -8,   5,   2,   0,   2,  -2,   1,  -5,  -3,  -6,   4,   4,  -1,   0,   1,  -1,   0,  -3,   1,   0,  -1,  -1,  -1 },
     {    5,  -1,  -1, -12, -16,  13,  10,  -2,  -2,   0,   1,   0,   6,  -1, -13,  -2,   5,  -3,  10,  -3,   0,   2,  -1,   2,  10,   8,  18,   6,  -2,  11, -14,   3,  -4,   2,  -2,   2,  15,  10, -19, -17,  -4,  18,  -4,   0,   1,  -4,   3,  -1,   6, -18,  -4, -23,  -1,   7,   7,  -3, -13,  13, -29,   1,  -5,  -1,   8,  -5,  90, -35, -13,  -8,   2, -10,   3,   1,  14,  20,   1,  -4,   1,  -1,  -5,   1, -11,   6,  10,  -3,  -8,  -2,   7,   4,  -5,   0,   2,   5,  -6,   1,   2,   2 },
     {   -4,  -6,  -2, -11, -26,  24,  -7,   4,   2,  -1,   2,  -1,  -6,  -7, -12,  -7,  20, -24,  11,  -6,   5,  -2,   3,  -3,  -1,  -1,  12,  31, -17,  13, -13,   9,  -6,   3,  -2,   0,  -5,   1,  -1, -47,   5,  30, -10,   8,  -1,   3,  -2,   0,  -3, -18, -46,  -2,  -8,   7,   0,   1, -29, -49,   5,   7,  -5,  -7,   7,   0, -36,  -4,  18,  -6,  -5,  -5,   1,  -3,  -1,   1,   7,   8,  -3,  -4,   1,   0,  15,   0,  -5,   5,   4,   1,  -4,   0,   3,   5,  -2,  -2,   2,   4,   0,   0 },
     {    3,   1,  10,  19,   8,   1,  -7,  -2,  -5,   3,   0,   0,   1,  -5,  -9, -10,  12,  15,  -6,   2,  -2,   0,  -2,   2,   3,   1,  12,  18,  -8,   0,  -3,  -4,  -2,   0,  -2,   1,   7,  -1,  -3,  15,  21,  -1,  -5,  -3,   1,  -5,   0,   0, -13, -20,  -5,   2, -25,  26,  -4,   6,  38,  22,  33,  21, -22,  -3,   5,  -4, -20, -47,   1, -22,  39, -11,   9,   2,  19,  11,  18, -22, -16,   6,  -6,   5, -11, -11,  -1,  27,   2,  -2, -16,  -7,  -5,   0,   4,  -5,  -1,   3,  -3,   1 },
     {   -2,   3,   0,  15,   3, -10, -11,   1,   0,  -2,   3,  -3,  -3,   7,  -4,  -6,   0,  10,  -4,   9,  -2,   3,   0,  -3,  -8,  -3, -11,   8,  -8,  -1, -15,  10,   0,   1,   0,  -1,   2,  20,  10,  20,  13,  -3,  -5,   0,  -2,  -3,   3,  -1,  -3,   3, -14,  28, -25,  25, -11,  -5, -23,   5,  10,   1, -38,  26,   1,  -5,  19,  24, -20,   9, -45,   1,  16,  -8,  -5,  30,  11,  38,   0, -32,  10,  -7,  -3, -15,  -6,  27,  -8,   5,  -8,   2,  -2,  -2,  -2, -10,   0,   0,   3,  -5 },
     {   -1,  -6, -16, -20,   0,  16,  -4,  10,   8,  -3,   0,  -4,   9,   6,   8,  33,   2, -24,   9, -18,  13, -10,   6,  -3,   0,  -7, -22, -28, -15,   6,  13,  10,   1,   3,   1,   0,   4,  -2,  -6,  15,   6,   4,  -8,  15, -13,   8,  -4,   1,   3,   1, -13,  -1,  31, -11,  -8,  12,  14,   5, -11,  43,  -8,   2, -18,   9,   7, -15, -11,  14,  -7,  14,   1,   2,  34, -22,  13,  16, -18,  -1,   5,  -3,   4,   1, -22,  29,  -4,  12, -17, -22,  -5,   1,   5,  -8,   5,  -2,  -1,  -9 },
   }
 },
 {  //27
   {
     {  110, -34, -15,  -4,  -3,  -1,  -1,   0,  -1,   0,   0,   0, -41,  30,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0, -13,  -1,   6,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   12, -52,  18,   1,   3,   1,   1,   1,   1,   0,   0,   0,  86,   6, -37,   0,  -6,   0,  -2,   0,  -1,   0,  -1,   0, -41,  46,   9,  -6,   0,  -2,   0,  -1,   0,  -1,   0,   0,  -9, -11,  11,   4,   0,   0,   0,   0,   0,   0,   0,   0,  -6,   4,  -3,   1,   1,   0,   0,   0,  -3,  -2,   3,   0,   1,   0,   0,   0,  -2,   1,  -1,   1,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,  -1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {  -30,   2,  -2,   8,   0,   2,   0,   1,   0,   0,   0,   0, -11,  77, -10, -16,  -1,  -5,   0,  -3,   0,  -1,   0,  -1, -34, -33,  58,   1,   2,   0,   1,   0,   0,   0,   0,   0,  26, -40, -19,  19,   1,   4,   0,   2,   0,   1,   0,   1,   2,   9, -10,  -7,   3,  -1,   1,   0,   6,  -6,   1,   0,  -1,   1,   0,   1,   1,   3,  -3,  -1,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,   1,  -1,  -1,   1,  -1,   0,   0,   0,   1,  -1,   0,   1,  -1,   0,   0 },
     {   25,  92, -28, -13,  -8,  -4,  -3,  -2,  -2,  -1,  -1,  -1,  13,   0,  -1,   2,   0,  -1,   0,   0,   0,   0,   0,   0, -29,  34,  14, -18,   3,  -3,   1,  -2,   0,  -1,   0,  -1,  -7, -29,  42,   5,  -4,   1,  -1,   0,   0,   0,   0,   0,   4, -18, -10,  16,   1,   1,   1,   1,  -1,  -3,   0,  -2,   3,   0,   0,   0,   2,  -6,  -3,   4,   0,   1,   0,   0,  -1,  -1,  -1,   0,   1,   0,   1,   0,   1,  -2,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {   24,  31, -24,   0,  -3,  -1,  -1,  -1,  -1,   0,   0,   0,  58,  -8,  36, -14,  -5,  -4,  -3,  -2,  -1,  -1,  -1,  -1, -35, -47,  -5,  39,   4,   6,   0,   3,   1,   1,   0,   1,   5,  26, -44, -19,   9,   0,   3,   0,   1,   0,   1,   0, -11,   9,  25, -12,  -7,   0,  -1,   0,   0,  -3,  -3,   7,   0,   0,   0,   0,  -4,   2,   4,  -3,   0,   0,   0,   0,  -1,  -1,  -2,   2,   0,   0,   0,   0,  -2,   1,   2,  -1,   0,  -1,  -1,   1,  -1,   1,   1,  -1,   0,   0,  -1,   0 },
     {  -25, -12, -29,  10,   5,   3,   2,   2,   1,   1,   1,   1,  -8,  54,  -6, -11,  -5,  -2,  -1,  -2,  -1,  -1,   0,  -1, -16,   9, -30,  -5,  12,  -2,   4,   0,   2,   0,   1,   0,   5,  33,  36, -53,  -6,  -2,  -1,  -2,   0,  -1,   0,  -1,   6, -31,  47,  25, -17,   0,  -4,   0,   4,  -8, -16,  13,   6,  -1,   1,   0,   2,  -3,   6,   0,  -1,   1,  -1,   0,   1,  -3,  -4,   4,   1,   0,   0,   0,   1,  -1,   2,   0,   1,  -1,  -2,   2,   0,  -1,   1,   0,   1,  -1,  -1,   1 },
     {    5,   7, -39,  17,  -4,   2,   0,   2,   0,   1,   0,   0,  31,  34, -10, -30,   3,  -5,   1,  -2,   0,  -1,   0,  -1,  76,   7,  28,   0, -19,  -2,  -6,  -1,  -3,  -1,  -2,  -1, -51,  20,  -2,  23,  11,  -3,   1,  -1,   1,   0,   0,   0,  -4,  -9,   8, -14,   7,   3,   1,   0,  -8,  -4,   6,   6,  -4,   2,   0,   1,  -2,  -4,   1,   0,   2,   0,   1,   0,  -3,  -2,   1,   3,  -1,   1,   0,   0,  -1,  -2,  -1,   0,  -1,  -1,   0,   0,   0,  -1,   0,   0,  -1,  -1,   0,   1 },
     {   -2,  23,  -2,   9,  -6,  -1,  -2,  -1,  -1,  -1,   0,   0, -25,  -2, -55,  14,  13,   3,   3,   2,   2,   1,   1,   1, -18,  30, -11,  -2,   0,  -3,   1,  -1,   1,  -1,   0,   0,  -8,   3, -33, -19,  29,  -2,   5,  -1,   2,  -1,   1,   0,   9,  13,  31, -64, -12,   7,  -1,   2,   8, -13,  33,  18, -25,  -2,  -2,   0,   2,  -2,  -3,   5,   5,  -4,   0,   0,   3,  -4,   8,   3,  -3,   1,  -2,   1,   1,  -1,  -1,   2,   1,  -2,   3,   1,   1,  -1,  -1,   1,   1,  -1,   2,   1 },
     {   15,  24,  19,  -8, -10,   0,  -3,  -1,  -2,   0,  -1,   0,  15,   6, -52,  20,   7,  -2,   2,   0,   1,   0,   0,   0,  21, -32,   4, -42,   2,   3,   2,   2,   1,   2,   0,   1,  44,  44,   4,   9, -31,  -5,  -7,  -2,  -3,  -1,  -1,   0, -38,  19,  17,  14,  17,  -9,   2,  -3,   3, -16,   6,  -6,   5,   6,  -2,   1,  -9,   3,   1,   4,  -2,   1,   1,  -1,   1,  -5,   2,   0,   1,   1,   2,   0,  -4,   1,   0,   2,   0,  -2,   1,   0,  -2,   0,   0,   1,   0,  -1,   1,   0 },
     {  -14,  -4, -34,   1,   4,   4,   2,   2,   1,   1,   1,   0,   2,  28,   9,  36,  -8,  -6,  -3,  -3,  -2,  -1,  -1,  -1, -19,   1, -41,   5,  13,   5,   3,   1,   1,   1,   1,   1, -10,  21,  -1,  19,   2, -12,   2,  -4,   1,  -2,   0,  -1, -13,  -9, -34, -20,  52,   4,   4,   1,   7,   2,  19, -55, -20,  15,   1,   2,   4,  -7,  14,  14, -12,  -7,   1,  -1,   2,   0,   1,  -8,   2,   0,  -1,   0,   2,  -3,   5,   3,   1,  -1,   0,  -2,   1,  -2,   2,   1,   1,   0,   0,  -2 },
     {   -7, -22, -90,  20,   8,   8,   4,   4,   2,   2,   1,   1,   2, -30,  -7,  -5,   1,   2,   2,   0,   1,   1,   1,   0,   7,   5,   3, -22,   9,  -2,   2,  -1,   0,  -1,   0,  -1,  36, -21,  -2, -11, -17,   8,  -3,   3,  -1,   1,  -1,   1,   0,  45, -11,   4, -16,  -9,   0,  -3, -10,   9,  19,  -2,   8,  -7,  -1,  -2,  -1,  -1,   4,   3,   1,   3,  -1,  -1,  -2,   2,   4,  -1,   0,   0,   1,   0,   0,  -1,   2,   1,  -1,   0,   2,   0,   0,   0,   1,   1,  -1,   1,   1,   0 },
     {   -8, -12, -20,  25,  -1,   1,   0,   1,   0,   0,   0,   0, -21, -39,   3, -34,  19,   5,   4,   2,   3,   1,   1,   1, -34,  24,  34,   1, -33,   1,  -4,   1,  -2,   0,   0,   0,  28,  36,   1,  31,   9, -17,  -3,  -5,  -1,  -3,   0,  -1, -24, -38,  11,  -2,  24,  13,   0,   2,  19,  -6, -23,  -1, -12,   6,   3,   2,  -4,   5,   1,  -3,   1,  -5,   1,   0,   7,  -3,  -4,   0,   0,   0,  -2,   1,   0,   4,   0,  -2,   3,  -1,  -1,   0,  -1,   1,   0,   0,   2,   0,  -1,   0 },
     {   -1,  17,   1,  39,  -5,  -4,  -2,  -1,  -1,  -1,  -1,  -1, -16,   1, -37,   6, -10,   8,   1,   2,   2,   1,   1,   1,  -8,  18,  -7,  40,  -5, -12,   0,  -3,   0,  -1,   0,  -1, -16,   7, -25,   3,   7,   0,   5,   0,   2,   0,   1,   0,  -4,  17,  -7,  29,   9, -26,   3,  -6,  -7,  -6, -19, -23,  62,   2,   0,   1,   7,  -5,  13, -41, -12,  17,  -1,   3,   1,  -3,   7,   3,  -4,  -3,   2,   0,   3,  -3,   4,  -9,   1,  -2,   3,   1,   1,  -2,   2,  -4,   1,  -1,   1,   0 },
     {   10,   0,  -6,  39,  -9,   2,  -2,   0,  -3,   1,  -1,   0,  10, -16, -49,  10,  22,   2,   2,   2,   1,   1,   1,   0,   0, -54,  -3,  -3,  12,   3,   4,   1,   2,   1,   1,   0, -12, -16,  16, -16,  16,   7,  -2,   1,  -1,   1,   0,   0,   7, -37, -26,  -3, -14,  25,   1,   5,   0,  39, -31, -11, -19, -12,   8,  -2,  -1,  13,  16, -10,   3, -11,  -2,  -1,   0,   4,   2,  -1,  -2,   1,  -2,   0,  -1,   2,   5,  -4,   1,   2,   1,   1,   0,   1,   2,  -1,   0,   2,   1,   0 },
     {    0,   7,   9,   8, -17,  -7,   0,   0,   0,   0,   1,   1,   5,  26,  12,   0,   5,   5,  -7,  -1,  -3,   0,  -2,   0,   5,  16, -27,   9, -27,  -5,   7,   0,   1,   0,   1,   0,  33,  12,   5,  15,  18,  -7,  -8,  -1,  -2,  -1,  -1,   0, -29,   1, -29,  -6, -25,  11,  15,   2,   5,  29,  16,  15,   9, -55,  -6,  -3,  -3, -15,  15,  -1,  46,  22, -16,  -2,   0,  -5,  -4,  -9, -11,  14,   9,   0,  -2,  -4,   3,   1,   0,  -2,  -2,  -3,  -1,  -2,   2,  -1,   0,  -1,   0,  -1 },
     {  -10,  -7,  -9,   0, -13,   7,   1,   1,  -1,   1,   0,   1,  -8,  -7,  -3,   9,  15,  -4,   2,  -2,   0,   0,   0,   0, -43, -30,   0, -30, -22,  21,   4,   6,   2,   3,   1,   2, -58,  28,   2,  26, -34, -14,   3,  -2,   1,  -1,   1,  -1,  37,  13,  -3,   9,  -1, -19,  -1,  -5,   2,   6,  12,  23,  13, -17,  -2,  -4,   8,  -8,  -9,  -2,  29,   5,  -5,   1,   2,  -1,   2,  -9,  -3,  10,   0,   0,   3,  -2,  -3,   1,   0,   0,   2,  -2,   2,  -1,  -1,   1,   0,   0,   1,  -2 },
     {   -8, -10, -32, -37, -15,  13,   5,   7,   1,   2,   1,   2,   2,   1,  -1,  71, -12, -12,  -5,  -4,  -1,   0,  -1,   0,  11,   9,  11,  10,   5,   2,   1,  -2,  -2,   0,  -1,   0,   2, -13, -17,  14,   9,  -5,   3,   0,   2,   0,   0,   0, -16, -17,  16,   3,   2,   5,   0,   0,   4,  -8, -42,  28,   3,  -5,   7,   0, -11,  19, -18, -31,  27,  -7,   2,   0,   7,  -3,  14, -15,  -7,   7,  -3,   2,  -2,   4,  -1,  -2,   3,  -2,   4,  -4,  -1,   2,   0,   0,   1,  -1,   2,  -2 },
     {   10,   3,  10,  71, -18,  -8,  -8,  -2,  -1,  -2,  -2,   0,  13,   9,  25,  17,   7,  -2,  -5,  -2,  -2,  -2,  -2,   0,   7,  -4, -30, -13, -14,   1,   6,   2,   3,   1,   1,   0,  25, -24,  13,  14,  11, -25,  -3,  -4,   0,  -1,   0,   0,  28,  11,  17,  -4,  25,   9, -13,   1, -23, -20,  -3,  21,   0,  23,   2,   0,  -5,   4, -34,  -6,   5,   3,  10,   0,  -6,   7,   2, -14,  -2,  -2,   2,   1,  -1,   0,  -3,  -2,  -2,   2,   2,  -5,   0,  -1,  -1,  -1,  -1,   1,   0,  -2 },
     {   -2,  -8,  -7, -39,   2,  -5,   9,   1,   3,   0,   1,   0,   8,   4,   8,  -5,  47,   2,  -7,  -3,  -1,  -3,   0,  -1,  -2,  -9,  -4, -27,  -9,   4,   8,   3,   1,   2,   1,   1,   8,  -2,   4,  -3,  28, -12, -10,   0,  -2,   0,  -1,   0,  17,  -1,   3, -15,   5,  18,  -4,   2,  -9,  -2,  -2, -14,  20,  12, -10,   4,   2, -10,  11, -44, -27,  42,   0,   1,   2,   3,  14,  22, -48,  -8,   7,  -1,  -3,   1,  -4,  14,   0,   2,  -1,   6,  -2,   0,  -2,   4,   0,   1,   0,   2 },
     {   -3, -10,  -2,  27,   3,  -9,  -5,   2,  -2,   1,   0,   1,  -1, -14,  12,  25,   0,   0,  -3,  -1,  -2,  -2,  -1,  -1,  -3, -17,  21,  -4,  29, -13,  -1,  -1,   0,   0,   1,   0, -22,  -1,  17,  -6,  15,  21,  -5,   2,  -2,   0,  -1,   1, -35, -23,  -1,  -9,  -8, -15,   9,   5,  39, -17,  25,   8,  20,   5, -35,   2,  -7,  -8, -28,  15,  -6,  48,   5,  -7,   3,   8, -17, -13, -12,  -9,  18,   1,  -2,   5,   0,  -2,   4,   4,  -3,  -2,  -1,   1,   1,  -1,   1,   1,  -1,  -1 },
     {    2,  -5, -11, -30,  -8,   5,   2,   2,   1,   2,   1,   1,   4,  -8, -31,  11, -24,   9,   0,   2,   1,  -1,   0,   2,   6, -16,  20,  44, -20, -20,   0,  -2,   1,  -1,   1,   0,  31,  17,  20,  -1,  17, -24,  -1,  -5,  -1,  -3,  -1,  -2,  53, -19,  -2,   1,  13,  -9, -10,   3, -13,  24,  25,   5,  10,  12, -16,  -3,  -7, -15, -12,  21,   1,  16,   9,  -2,  -7,   9, -15, -10,   9,  -6,   6,   2,  -1,   0,   3,  -4,  -3,   3,  -2,  -2,   1,  -1,   1,  -3,   0,   2,   0,  -1 },
     {    2,  -2,   8,  20, -50,   0,  -4,   2,   0,   1,   0,   1,   7,   5,  26,  37,  17, -14,  -9,  -3,  -2,  -2,   0,  -2,  10,  27,  31, -20,  -4,  15,  -4,  -2,  -3,   0,  -2,   0,  10,   3, -35, -23, -17,  -1,  11,   2,   4,   1,   2,  -1,  25, -19,   8,   4,  -4, -16,  -5,   0,  12,  28,   1, -15,   2,  -1,  -9,  -1,  -9,  -4,  38,  18, -15,   4,  -1,  -2,  -5,  -5, -18,  26,   9,  -4,   3,   1,  -5,   1,  -3,  -5,   0,  -2,  -3,   3,  -2,  -1,  -1,  -3,   0,  -1,  -2,   1 },
     {   -4,  -7,   2,   8, -15,  -5,   8,  -2,   1,   0,  -1,   0,  -5, -11,  11,  30,  -8,  -9,  -3,   0,   1,   1,  -1,   0,  -6,  -9,  40,   5,   3, -17,   8,  -3,   1,  -2,  -1,   0, -11,  18,  24, -14,  27, -12, -19,   4,  -3,   2,  -1,   2,  -3,  27,   5,   9,  -2,  36,   6,  -6, -32, -39,   2, -16,  -6, -25,  21,  11,  33, -11,   9,  10,  -4,  -4, -35,   2,  -6,   2, -10,  18,   9,  18,   5,  -7,   6,   0,  -9,  -7,  -2,   1,   0,   3,   1,  -2,  -3,  -1,  -1,   1,   0,  -1 },
     {   -1,  -6, -10, -13, -27,  16,  11,  -3,   0,   3,  -2,   1,   1,  -4, -13,  -5,   0,  23, -12,   3,  -5,   2,   0,   1,   5, -18, -29,  16, -49,  16,  19,   1,   4,   0,   1,  -1, -10, -47,  11,   6,  -7, -23,   3,  -2,   3,   2,   2,   1, -18,   1,  21,   2,  -6, -11,   7,  -3,  33, -29,  -4, -13,  -1, -12,  -6,   7,   7,  22,   8,  23, -17,   3, -14,  -7,  -2,   8, -15,  21,  11, -12,  16,  -4,  -3,   4,  -3,  -9,   2,   2,   3,   2,  -2,   2,   1,  -4,   1,   1,   0,   0 },
     {    8,   1,   5,   4,  29,   4,  -1,  -2,  -1,  -2,   0,   1,   3, -12,  -7, -17, -12, -40,   7,   6,   4,   0,  -1,   0,  11,  -3,  -2,  -2,  21,   1,  -3,  -1,  -2,   0,  -1,  -1,   6, -26,  -4,  -6,  -2, -28,   9,   7,   1,   1,   1,   1,  13, -13,  22,  -4,  36, -18, -19,   3,  23,  -1,   0, -11,   4, -11, -12,   9,  10,   4,  19,  -1,  32,  17, -43,   2,  -5,  -3,  -3,  -1, -14,  53,   0,  -7,  -4,   6,  -3,   7,  -3,   2,  -2,   2,  -2,   4,  -1,   1,  -2,   0,   0,   1 },
     {   -5,  -6,   0, -14, -49,   2,  12,   1,   3,  -2,   2,   1,  -5,  -1,  -9, -21,  66, -26, -10,   6,  -2,   1,  -1,   0,   2,   6,   0,  42,  24,  21,  -8,  -8,  -2,  -4,  -1,  -2,   0,  -6,  12, -20, -12,  -1,  15,  10,   0,   3,   1,   2, -26,   5, -13,   8,  15,  -6, -19,   3, -19,  -8,  12,   6,  -2,  11,   3,  -5,  11, -10, -21,   4,   9,  -3,   0,   3,  -2,   5,  -4, -23,   3,   9,   0,  -1,   7,  -5,   6,  -3,   1,   1,  -1,   1,   1,   0,   1,  -1,  -1,   1,   0,   0 },
     {    1,   6,  -5,  -7,  63, -31,   5,  -2,   6,  -2,   1,  -3,   0,   8,  -7,  21,  52,  -9, -18,   0,  -1,  -3,  -2,  -1,   0,   6,  -4,   2, -14,   1,  -8,   4,  -1,  -1,   1,   1,  -6,  -2, -21,   6,  12,  -3, -16,   1,   6,  -2,  -1,   2,   4,  12,   8,  29,   0,  17,  -8, -12,   5,  11,  -9,   4,   2,   4,   3,  -4,  -7,  14,   1,  28,  -1,  10,   1,  -7, -17,  -4, -22, -12,  29,   3,  23,  -3,   6,  -6,   7, -28,   2,  -6,   4,  -5,   1,  -2,   2,  -4,  -1,  -1,   0,   0 },
     {   -6,   1,   5,   0, -13,  -6,   5,   4,   1,   1,   3,   0, -10,  -2,   5,  -3, -22,  -5,  -4,   2,  -2,   5,   2,   1,  -7,  -4,   9,  -4,  -1,  -8,   7,  -6,   2,  -2,   1,   0, -27, -10,  24, -13,  15, -37, -11,   7,   2,   0,  -1,   1, -35,  48,  16,  -2,  12,  13, -33,  -4,  18,  52,  -8,  10,   1,  27,   6, -13, -30, -13,  16,   2,   1,   0,  15,   2,  19, -15,  -4,   4,  -3,  -2,   0,   2,  -4,   0,   2,  -2,   3,  -6,  -5,   0,   1,   0,  -1,   1,   5,  -2,  -2,   1 },
     {    0,   1,   5,  12,  43,  24,   4,  -1,  -7,  -2,  -1,  -2,  -3,  -3,  -1,  25,   4, -13,   8,  -6,   7,  -1,  -3,  -2,  -3,  -8,   5,   8, -36,  19,  -3,  -4,  -2,   3,  -1,   0,  -2, -23,  -1,  -6, -18,   1,  11,   8,   0,   0,   1,  -1, -35, -28,  23,  -3,   0,  -3,  -2,  -3, -46,  13,  27,   1,   7,  -8,  17,  -1,  -3, -43,   2,  -2,  -7,  -3,   4,   9,  22,  16, -13,  11,  -8,  -8, -15,   0,   2,   6,  -2,  10,   3,   3,  -9,  -4,   1,   2,   2,   3,   1,   0,  -1,   1 },
     {    3,   8,   7,  10,   7,  49,  31,   4,  -2,  -1,  -7,  -5,  -7,  -3,  -4,  -6, -30, -39, -39,   2,   1,   9,   4,   6,  -5,   2,  -3,  -3,   3,  38, -22,  -5,  -2,  -3,  -1,   1,   7,  16,  14,  18,  10,  17,  15,  -2,   1,  -2,  -3,  -3,   3,   5,  -9, -15, -18,  10, -16,   2, -15,  -4,  -7,   8,   5,  11,  -2,  -8,   8,  19,  15,   4,  -8,  17,   6, -22,  -6,  -8,  -5,   0,  -6,  -8,  36,  -6,  -2,  -9,   5,  -9,   4,  -4,   3,   3,  -1,  -2,  -1,  -4,   1,   0,  -2,   1 },
     {   -3,  -4,  -2,  -5,   0, -23,  18,  -9,  17,   6,  -8,   0,   0,   1,   8,   7,   9, -14,  33,   4, -10,   4,   8,   2,  -7,  -8,  -5,  -9,  -3, -30, -15,  12,   2,   0,   4,  -1,  -1,  -5,  14,  12,   6,  -6,  24, -30,  -2,  -1,   3,   3, -19,  -1,   9, -16,  -6, -23,   1,  32, -33,  13,   2,   6,  11,   4, -40,  -9,  29,  28,  14,  -7,  -2, -24,   5,  -5,   3, -13, -13,  16,   8,   2,  -3,  -1,  -4, -16,  -5, -16,  14,  -4,   6,  -5,  -1,  -3,   1, -19,   5,   5,   1,  -5 },
     {    1,  -1,  -3,   8,  13, -28, -13,  -3, -14,   8,   4,   6,   3,   3,   1,  20,  10, -33,   7,  -5,  -9,   5,  -5,  -4,   4,   0,   4,  43, -24,  14, -28,   3,   4,  -5,  -4,   4,  10,   7,  42,   1, -31,  19, -13,  -7,   1,  -1,   3,   5,  15,  17,  -8, -28, -11,   8,   4,   2,  39,  -9,   4,   5,  11,   5,  12, -12,  -8,  14,   9,  -9, -10, -10,   2,  15,   1,  -5,  11,  22, -12,   8, -15,   2,  -6,   1, -17,   6,  -6,   0,  -3,  -3,  -1,   2,  -3,   4,  -2,  -1,   1,  -1 },
   },
   {
     {  113, -46,  -6,  -4,  -2,  -2,  -1,  -1,   0,  -1,   0,   0,   5,  13,  -8,   1,  -2,   0,  -1,   0,   0,   0,   0,   0, -25,  19,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0, -14,   6,   3,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  -7,   1,   4,   0,   0,   0,   0,   0,  -3,  -2,   2,   1,   0,   0,   0,   0,  -3,   0,   1,   1,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   16,  10,  -6,  -1,  -2,   0,  -1,   0,   0,   0,   0,   0,-104,  38,  13,   1,   3,   1,   1,   0,   1,   0,   0,   0,  -8, -34,  18,  -1,   3,   0,   1,   0,   1,   0,   0,   0,  31, -26,  -3,   3,  -1,   1,   0,   1,   0,   0,   0,   0,  12,  -3,  -8,   3,  -1,   1,   0,   0,   6,   2,  -6,   1,   0,   0,   0,   0,   2,   2,  -3,  -1,   0,   0,   0,   0,   4,   0,  -1,  -1,   0,   0,   0,   0,   2,   1,   0,  -1,   2,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {  -16,  35,  -9,   0,  -2,   0,  -1,   0,   0,   0,   0,   0,  19,  45, -18,  -3,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0, -71,  44,  22,  -9,   4,  -3,   2,  -2,   1,  -1,   0,  -1, -13, -39,  35,  -2,   4,   0,   1,   0,   1,   0,   0,   0,  15, -32,   4,   7,  -2,   2,  -1,   1,   7, -10,  -6,   5,  -1,   1,   0,   0,   5,  -3,  -5,   2,   0,   0,   0,   0,   2,  -2,  -2,   0,   1,   0,   0,   0,   2,  -2,  -2,   0,   1,  -1,  -1,   0,   1,  -1,  -1,   0,   1,  -1,   0,   0 },
     {  -18, -60,  35,   1,   7,   0,   2,   0,   1,   0,   1,   0,  18, -29,   6,   5,   0,   2,  -1,   1,   0,   0,   0,   0, -48, -14,  13,   9,   2,   2,   1,   1,   0,   1,   0,   0,  27, -49, -17,  15,  -1,   4,  -1,   2,   0,   1,   0,   1,  17,  15, -41,   5,  -1,   1,   0,   1,  -3,  23, -15,  -9,   3,  -2,   1,  -1,   0,  13,  -1, -10,   1,  -1,   0,   0,   0,   5,   3,  -5,  -1,   0,   0,   0,   1,   4,   1,  -2,   0,   2,   2,  -1,   0,   2,   0,   0,   0,   1,   1,   0 },
     {   18,  44, -30,   3,  -7,   0,  -2,   0,  -1,   0,  -1,   0,  43,  26, -19,  -5,  -2,  -3,  -1,  -1,   0,  -1,   0,  -1, -15, -38,  22,  -1,   4,   0,   2,   0,   1,   0,   0,   0,  42, -12, -34,   6,  -3,   2,  -1,   1,   0,   0,   0,   1, -16,  54,  -7, -16,   1,  -4,   1,  -2, -11,   1,  30,  -8,   0,  -1,   0,  -1,  -3,  -6,  10,   4,  -2,   1,  -1,   0,  -3,  -4,   4,   4,   0,   0,   0,   0,  -2,  -2,   0,   2,  -2,  -2,   1,   1,  -1,  -1,   0,   0,  -1,  -1,   1,   0 },
     {   37,  53, -21,  -4,  -6,  -2,  -3,  -1,  -1,  -1,  -1,  -1,  22, -39,  32,  -7,   4,  -2,   1,  -1,   0,   0,   0,   0,  22, -31,  17,   0,   0,   1,  -1,   0,   0,   0,   0,   0, -21, -28,  18,  14,  -2,   4,  -1,   2,   0,   1,   0,   1,   7, -31, -30,  24,  -2,   5,   0,   2,  -7,  23, -37,  -2,   4,  -1,   2,   0,  -5,  13,  -3, -14,   3,  -2,   1,  -1,  -2,   7,   3,  -5,  -3,   0,   0,   0,  -2,   2,   2,  -1,  -1,   3,   0,   0,  -1,   1,   1,   0,  -1,   2,   0,   0 },
     {    2,  23, -26,   6,  -2,   1,  -1,   0,   0,   0,   0,   0, -30, -26, -21,  16,   3,   5,   1,   2,   1,   1,   0,   1, -13,  39, -41,   1,   2,   0,   1,   0,   0,   0,   0,   0, -23,  -3, -10,  -5,   8,  -1,   3,   0,   1,   0,   1,   0,  19,  15, -39, -19,   7,  -3,   3,  -1,  -4,  38,  18, -42,   3,  -5,   1,  -2,  11,  -9,  40, -13,  -9,   1,  -3,   0,   5,  -5,  15,   7,  -9,   1,  -2,   0,   2,  -5,   6,   7,   1,  -1,   3,   3,   1,  -2,   2,   2,   0,   0,   1,   1 },
     {    1, -22,   7,  -4,   3,   0,   1,   0,   0,   0,   0,   0,  12,  40, -27,  -5,  -4,   1,  -1,   0,  -1,   0,   0,   0,  75,  31,  10, -25,  -2,  -6,  -1,  -3,  -1,  -2,   0,  -1,  25, -39,  35,   2,  -4,   3,  -2,   1,  -1,   1,  -1,   0,   7,   6, -23,  15,  -7,   3,  -2,   1, -20,  26,   4,  -6,   0,  -2,   1,  -1,  -6,  -4,  15,   6,  -3,   0,  -1,   0,  -7,  -3,  -1,   9,   1,   0,   0,   0,  -4,  -2,  -3,   3,  -3,   0,  -1,   1,  -2,  -1,  -1,   0,  -2,   0,   0,   0 },
     {   15,  26,  34, -24,   0,  -7,   0,  -3,   0,  -1,   0,  -1, -13, -48,  35,  -6,   3,   2,   2,   1,   0,   0,   0,   0,  -2,  29,  13, -12,  -6,  -2,  -2,  -1,  -1,   0,   0,   0,   8, -22,  16,   6,  -4,   2,  -1,   1,  -1,   1,   0,   0,  14,  24,   1, -11,  -7,  -1,  -1,  -1,  -3, -11,  50,  14, -12,   2,  -4,   1,   4, -25, -13,  44,  -4,   3,  -1,   2,  -7,   4, -25,  11,  10,  -2,   2,   0,   0,  -1, -12,  -2,  -2,   2,  -4,  -2,   0,  -1,  -4,  -1,  -1,   2,  -2,  -1 },
     {  -10, -13,  24,  -8,   4,  -1,   0,   0,   1,   0,   0,   0,   2,  38,  24, -19,  -1,  -3,  -1,  -2,  -1,  -1,  -1,  -1,   8, -24,  34,   3,  -2,   2,  -1,   0,  -1,   0,  -1,   0, -79,  -5, -14,  23,   2,   6,   1,   2,   1,   1,   0,   1, -23,   4, -25, -11,   9,  -2,   3,   0,  11,  -3,  19, -27,  -2,   0,   0,   0,  15, -15,  21,   6, -13,   3,  -3,   1,   4,  -5,  -7,  18,  -4,   0,  -1,   0,   2,   0,  -8,   4,   2,   0,  -3,   2,   2,   0,  -2,   0,   1,   0,  -1,   1 },
     {   -6, -11, -33,  27,  -5,   6,  -1,   3,   0,   1,   0,   1, -15, -18, -41,  26,   2,   3,   1,   3,   1,   1,   1,   1,  17,   4,  16,  -5,   3,  -4,   1,  -2,   0,   0,   0,   0, -43, -33,  -6,  23,   6,   4,   1,   2,   1,   1,   1,   1, -10,  23, -14, -11,   5,  -3,   2,  -2,  -5, -25,  13,  20,   4,   1,   0,   0,   0,   1, -51,   1,  19,  -1,   5,  -1,  -2,  14,  -1, -37,   9,  -1,   2,  -1,   7,  -1,   8, -15,   1,   2,   7,  -5,   3,   0,   2,   0,   0,   2,   3,  -2 },
     {   12,  39,  66, -33,   2,  -9,  -1,  -4,  -1,  -2,   0,  -2,  -5,  20,  -2,   8,  -3,  -1,  -1,   0,  -1,   0,   0,   0,   9,  33, -16,  16,  -7,   0,  -2,  -1,  -1,   0,   0,   0,   0,  -4, -28,  20,   5,   1,   2,   0,   1,   0,   0,   0, -17,  19,  -8,   0,   8,  -3,   2,  -1, -16, -16, -19,   9,  16,  -2,   4,   0,  -6,  10, -18, -37,  14,  -1,   2,   0,   4,  -5,  30, -19, -10,   2,  -3,   1,   4,  -4,   4,   7,   0,  -3,   4,   1,   1,  -2,  -1,   2,   0,  -2,   2,   0 },
     {    9,   1,  11,  -8,   2,  -1,   0,  -1,   0,  -1,   0,   0,  13,  11, -13, -10,   3,  -1,   1,   0,   0,   0,   0,   0,  20, -19,  -5,   3,   2,   1,   0,   0,   0,   0,   0,   0, -22, -22, -29, -18,  14,   0,   5,   0,   2,   0,   1,   0,  85,  -9,  26, -44, -11,  -2,  -4,   0,  15,   5,   8,  19, -17,   5,  -5,   2,   0,  23,  -8,  -2,  -3,  -1,   0,  -1,  -6,   7,  19,   7,  -8,  -1,  -2,   0,  -3,   2,  -1,  12,  -5,   1,   1,   4,  -2,   0,  -3,   3,  -2,   1,   1,   0 },
     {   -3,  -9, -10,  -3,   2,   2,   1,   1,   1,   0,   0,   0, -14, -37, -11, -22,  14,   0,   5,   1,   3,   0,   1,   0,  -2,  12,  50, -44,   0,  -3,   1,  -1,   0,  -1,   0,  -1,   2,   0,  11,  -7,  -5,   4,  -2,   1,  -1,   1,  -1,   0, -18,  15,  20, -10, -10,   3,  -1,   0,   2, -20,  -6,   5,   1,   2,   0,   2,  15,  13,   8, -39, -11,   4,  -4,   1,  13,  -6,  41,  21, -35,   3,  -6,   1,   0,   3, -13,  36,  -1,   2,  -4,   6,  -1,   1,  -6,   5,   0,   1,  -1,   2 },
     {   10,   9,  26, -18,   5,  -6,   1,  -3,   0,  -1,   0,  -1,   4, -16, -44,  27,  -1,   5,   0,   2,   0,   1,   0,   1,   8, -29, -26,  -3,  13,  -2,   4,   0,   1,   0,   1,   0,  14, -33,  16,  -9,   8,   1,   1,   0,   0,   1,   1,   0, -41, -34,   0, -18,  13,   4,   3,   2,  44, -27,   7, -31, -13,   8,  -5,   3,   4,  14,   2,  11, -25,   4,  -4,   0,   1,  12,  -5,   8,  -6,  -4,   1,  -2,  -5,  10,   1,   4,   0,   4,  -2,   3,  -1,   5,   1,   2,   0,   1,   0,   0 },
     {    2, -10, -34,  27,  -2,   6,   0,   1,   0,   0,   0,   1,  12,  10,  57,  -8,  -4,  -2,  -3,  -1,  -2,  -1,   0,  -1,  21,  45,  -4,  12,  -6,  -2,  -4,   0,  -1,  -1,  -1,   0,  33, -18, -31,  21,  13,  -2,   2,  -2,   1,   0,   1,   0,   1, -21, -11, -19,  22,  -2,   5,   0,  27, -41,   1, -21,   1,   5,  -1,   3,   9,   5,  -9,  -7, -13,   7,  -3,   2,  -1,  13,   2,   2, -12,  -1,  -1,  -1,  -4,   4,   1,   5,  -2,   1,   0,   0,  -2,   1,   2,   0,  -1,   1,  -1,   1 },
     {    3,   2,  10,  38, -21,   3,  -5,   1,  -3,   0,   0,   0,  -1,  -7,   2,  33, -16,   7,  -3,   1,  -1,   0,  -1,   1,   6,  -7,  -5,  38,  -3,  -2,  -1,  -1,   0,  -1,   0,   0, -10, -36,   0, -17,  13,   0,   4,   1,   2,   0,   1,   1,   2,  11,  35,  13,  -7,  -1,  -3,  -1, -46, -26, -13,   2,  19,   3,   4,   1,  19, -14,  21, -16, -19,   3,  -3,   0,  -3, -10, -13,  40,  -2,   0,   0,   0,   1,   6, -25,  -7,   0,   3,  -2,  -4,   2,   1,  -6,  -7,   0,   1,   1,  -1 },
     {    7,  15,  20,  22, -17,   0,  -6,   1,  -2,  -1,  -1,   0,  -7, -13, -34,  17,  -6,   2,   1,   2,   1,   0,   1,   0,   7,  26,  51,  -4, -15,  -5,  -3,  -2,  -1,  -1,   0,  -1,   7,  30, -32,  25,  -2,   0,  -3,  -1,   0,   0,   0,   0,  11,  -6,  -3,   6,   4,  -6,   1,  -2,  37,  10, -22,   1,   0,  -3,   1,   0,  -3,  33,  20,  15,  -5,  -8,  -2,  -2, -13, -12, -16,  25,  26,  -3,   4,  -1,  -7,   5, -12, -26,  -1,  -7,   9,  -6,  -2,  -1,  -1,  -8,   0,  -3,   3,  -1 },
     {   -5,  -4, -32, -49,  36,  -9,   9,   0,   4,  -1,   3,   0,  -8, -10, -25, -55,  21,   4,   6,   2,   2,   1,   1,   0,   1,  11, -14,  15, -13,   7,  -4,   2,  -1,   1,  -1,   0,  -5, -22, -30,   0,  -5,   7,   2,   3,   1,   1,   1,   0,  -4,   7,  13,  21,  -7,  -3,  -1,  -2, -11, -20, -23,  -4,  10,   3,   3,   1,  16,  11,   8,   5, -16,  -1,  -2,  -1,  -9,   2, -23,  12,  17,  -1,   4,  -1,   3,   5,   3, -29,   2,   0,   9,   4,   3,   1,   2,  -3,   0,   0,   4,   3 },
     {   -6,  -2, -12,   7,  -6,   1,   2,   0,   0,   1,   1,   1,  -6,   1,  26,  -3,  -5,   6,  -3,   0,  -2,   0,  -1,   1,  -1,  17, -10,  14,  -8,   7,  -6,  -1,   1,   1,  -1,   0, -13, -40,  -8, -10,   5,   4,   1,   3,   2,   2,   1,   1, -54,   8,  27, -22, -23,   5,  -2,   1,  12,  37,  -3,  23, -27,  -3,  -3,  -1, -39,  38,  15,  17,   7, -15,   4,  -7,   1, -18,  17,   3,  15,  -7,   1,  -1,   0, -14,   0,   3,   6,  -5,  -5,   0,   2,  -3,  -3,  -2,   3,  -2,  -1,   0 },
     {   -3,  -8, -17, -26,  11,   7,   1,   2,   1,   1,   0,   1,  -1, -10, -22,  14,  -1,   7,   0,   1,   0,   1,  -1,   1,   0,  -4,   9,  22, -29,  11,  -5,   1,  -3,   0,   0,   0,  12,  -4, -19,  47,  -6,   0,   0,  -3,   1,  -2,   1,   0,  -2, -54,   9, -12,   5,   0,   5,   1, -18,  12,  22,  20,  -4,  -8,  -2,  -1, -27, -39,   8,  -5,  20,   2,   1,   2,  23, -20,  -3,   9, -19,   5,  -5,   2,  -5,  12, -22,  16,   2,   3,  -6, -10,  -2,   6,  -2,  -1,   2,   0,   0,  -5 },
     {    3,   3,  17,  24,  18, -15,   4,  -4,   2,  -2,   0,  -1,  -2,  -6,   1, -10,  47, -24,  11,  -6,   3,  -3,   2,  -2,   1,   8,  22, -19,  40, -15,   5,  -5,   1,  -1,   0,  -1,   6,  -1, -24, -23,  39,  -1,   3,   0,   0,   1,   0,   0, -13, -31,   6, -26,  17,   6,  -3,   5, -39,  11,   2, -10,   3,   9,  -3,   3, -27, -13,   4,  -5,   2,   7,   0,   1,  -2, -12, -17, -12,   3,   6,   3,   2,   6,   1,   5, -24,   5,  -2,   5,   2,   1,   2,   1,   0,   1,   0,  -1,   4 },
     {    4,   6,  23,  41, -46,  11,  -6,   3,  -3,   2,  -2,   1,  -3,  -2, -17, -59, -14,  20,   6,   7,  -1,   4,   0,   1,  -5,  -2, -35, -36,   1,  18,   5,   7,   1,   2,   2,   0,  -1,  -9, -31,   1, -11,  14,  -2,   3,   1,   0,   0,   0, -14, -21, -12,   6, -18,   7,   3,   0,   3,   4,   0,  20,   3, -11,   1,  -3,   1, -24,  -8,   1,  -1,   4,   2,   2,   3,   8, -10,   4, -12,  -4,   1,   0,   0,  10,  -1,  -2,   1,   7,   0,   5,   3,   3,   1,   1,   0,   2,   2,   2 },
     {    4,   2,  -3,  15,  30, -24,   3,  -5,   2,  -2,   0,  -1,   8,   8,   9,  21,  37, -29,   2,  -5,   2,  -3,   2,   0,   4,   0, -29, -13,  24, -10,   3,   1,   0,  -1,   0,   0,  -6, -14,  -9,   9,  -6,   4,  -7,   4,  -2,   0,  -1,   1,   1,  19,  -2,  23, -22,   1,  -3,   2,  56,   1,  -3,  35,  12, -11,  -2,  -3,   8, -22,  13, -12,  21,   7,  -2,   2,  17, -13,  -8,  15,  -8,   9,  -1,   2, -12,   7, -23,  -4,   2,   3,   3, -16,  -5,   4,   0,  -7,   0,   2,   4,  -5 },
     {   -5,   1,  -7,  -2,   8,   0,   1,   3,  -2,   1,   0,  -1,   1,  15,   8,  36,  28,  -4,  -4,   0,  -3,   0,  -1,   0, -10,  -1, -22, -31,   8,  10,  -4,   4,  -1,   1,  -1,   1,  -4,  17, -17,  21, -20,  22,  -8,   5,  -3,   1,  -1,   0,   1, -13, -25,   4, -20,  22,  -4,   3, -44, -19,  -4,   6, -37,  12,   4,   1,  -3,  22, -13,  15, -28, -12,   5,  -3, -22,   9,  16,  28,   3, -15,   3,  -4,   4,  -2, -20,   7,  -4,   4,   3, -11,   6,  -1,  -2,  -3,  -1,   1,   3,  -3 },
     {   -4,  -3, -15, -47, -15,  15,   8,   5,   3,   3,   0,   1,   3,   2,   8,  34, -32,   9,  -5,   4,  -1,   3,  -2,   1,   6,  14,  26,  -6,  15,  -5,   4,  -1,   0,  -2,   0,  -2,  -1,   5, -37, -49,  27,  15,  -2,   6,  -1,   1,  -1,   1, -12,  -2, -17,   4, -26,  21,  -4,   3,  10,  12, -19,  15, -11,   2,   3,  -3,  13, -30,  -6,  -3,  -1,   1,   4,   1,  10,  19, -19,   0, -15,  -2,   2,   1,  -4,  10,   5,  -3,   3,  -1,   7,   2,  -1,  -2,   3,   4,   0,  -2,   0,  -1 },
     {    7,  13,  14,  40,  49, -12,  -9,  -3,  -1,  -3,  -2,  -1,  -1,   5, -12,  -7,  14,   7,  -4,   4,  -1,   1,   0,   0,   5,  12,  21,  49, -13,  17,  -3,   2,  -6,   0,  -1,   1,   7,  12,   7,  -9, -15,  17,  -3,   1,  -5,   1,  -2,  -1, -12,  -3,   0, -11, -28,  12,   1,  -1,  -1,  30,   5,   0, -14,  -1,   2,  -4,  26,  -4, -22,  -4, -19,  -1,   3,   5,  23,  42,  -7,  -2, -15, -12,   2,  -1,  -1,   4,  11,   7,   1,   3,  -1,   0,  -1,  -4,   6,   0,  -1,  -3,  -2,   2 },
     {   -5,  -3,  -7, -13, -21,  -5,   8,   1,   6,   1,   0,   0,  -2,   6,  -3, -16,   7,  -5,  12,   2,   3,  -1,   2,  -1,  -7,  -3, -12,   4,   1,  -4,   4,  -4,   4,  -2,   1,  -2,   4,  10,  25,  27,  50, -26,   3,  -8,   2,  -8,   0,  -1, -13,  16,   0, -22,  29,   0,   3,   2,  -2,  34, -14,  24,   0,  -3,  -8,   3,  14,   0, -23,  -8, -10,  -1,   2,  -1,  -3,  38,   0,  38,  -5, -18,  -1,  -2, -12,   5, -22, -10,   0,   4,  13, -28,   2, -10,   5,  -4,  -2,   3,   5,   1 },
     {    3,  -2,   2,  -3,  -3,  -1,   3,   0,   1,  -3,   1,  -1,   5,   2,  10,  12,  -5, -13,   8,  -2,   1,  -4,  -1,  -2,   2,  -3,  -6, -22,  -7,  -9,   3,   2,   1,  -1,  -1,   0,   9,  -8, -13,  12,  -6,  -9,   1,  -2,   2,   1,   1,   2, -16, -26,  21,  -9,  -2,  -9,  -1,   1, -18,  26,  13,   3,  -9, -11,   0,  -2,  95,  16,   0,   4,  20, -13,  -4,  -1,   7, -16,   6, -17,  19,  11,  -7,  -1,   3,   3,   4,  -9,  -9,  -7,  13,   8,  -3,   4,  -3,   7,  -4,  -2,  -3,  -2 },
     {   -3,   3,   0,  -3,  15,   0,  -7,  -1,   2,   0,  -2,  -1,   0,   8,  17,  30,  -3,   3, -12,  -4,  -1,  -2,   1,   1,  -5,   2,  -3, -47, -31,  24,  16,   5,   2,   3,   1,   1, -11, -10, -13,  -4, -25,  -8,   9,   2,  -1,  -3,   0,   0,  -3,   3,  30,   3,  11, -29,   2,   0,  -8,  24,   8,  -9,  32,  -4, -14,   6, -23,  16,  -7, -10, -11,  27, -11,   1,  -3,  35, -22,   7, -19,   5,   7,  -3, -13,  18,  15, -12,   9,  -5,  17,   2,  -3,   2,  -1,   6,   1,  -4,  -3,   0 },
     {   -1,   3,   1,   4,  10,  23,  -1,  15,  -5,   3,  -5,   0,  -7,  -4, -10,  -8, -11, -19,   0,   6,   2,  -1,  -1,   0,  -1,   8,   2,  19,  37, -34,  -1,   0,   4,  -3,  -1,  -1,  -7,  -5, -29,  -4, -10,  -9, -15,   2,   3,  -6,   4,   0,  -8,  -7,   7,  45,  25, -25,  -4,  -2,  -9,  19,   5,   9,   3,   4, -25,  13,   0,   8, -10,  38,  -6,  10, -22,   0,  -5,  20,  19,   7,  -9,  13,  -1,  -8,  -8,  -2,  -3,  41,  -4,  13,  -9,   2,   3,  -1,   0,  -8,   2,   1,  -2,  -9 },
     {    2,  -2,   1,   5,  12, -19,  12, -11,  20,  10,   7,   3,   1,  -3,  -1,   3, -12,  -2,   1,  13,  12,  13,  -3,   7,   2,  -5,  10,   4, -10,  20,  -1,  14,   3,   6,   1,  -5,   6,   1,  -3, -17,   9,   1,  -7,  18,  -4,   3,   0,  -8,  -5, -21, -12,   7,  -4, -22,   2,  10,   4, -11,  16,   9,   5, -27, -22,   4,   8,  -7,  15,   1,  19,   8,   3, -18, -44,  14,  29,   6, -16,   4,   8,   5,  22, -38,  -5, -20,  16,   5,  -4, -18,  26, -31,  -4,   3, -12,   4,  10,  17 },
   },
   {
     { -102,  40,  14,   3,   2,   1,   1,   0,   0,   0,   0,   0,  48, -32,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  25,   0, -11,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,  -9,   8,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -3,   2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -26,  17,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0, -86,  17,  22,   1,   4,   0,   2,   0,   1,   0,   1,   0,  62, -46,  -4,   3,   0,   0,   0,   0,   0,   0,   0,   0,  26,   8, -17,  -1,  -2,   0,  -1,   0,   0,   0,   0,   0,  -8,   9,   3,  -3,   0,   0,   0,   0,   3,  -3,   1,   1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },
     {   21,  90, -52, -12,  -6,  -2,  -2,  -1,  -2,  -1,  -1,  -1,   5, -14,  27,  -8,   0,  -2,   0,  -1,   0,  -1,   0,   0, -38, -34,  28,  14,   1,   2,   0,   1,   0,   1,   0,   0,  -1,  -8, -11,   7,   3,   1,   1,   0,   0,   0,   0,   0,   9,  -1,  -6,  -4,   2,   0,   0,   0,   2,   2,  -1,  -1,  -1,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   46,   8, -19,  -4,  -3,  -1,  -1,   0,  -1,   0,   0,   0,  11, -40,  18,   3,   1,   1,   0,   1,   0,   0,   0,   0,  55,   3, -33,   5,  -4,   1,  -2,   1,  -1,   0,   0,   0, -57,  55,   4, -10,   2,  -3,   1,  -1,   0,  -1,   0,  -1, -14, -18,  23,   0,   1,   0,   0,   0,   4,  -4,  -5,   5,  -1,   1,   0,   0,  -3,   1,   0,  -1,   1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,  -1,   0,   0,   0,  -1,  -1,   0,   0,  -1,   0,   0,   0 },
     {   -3,  10,  -3,   0,  -2,   1,  -1,   0,  -1,   0,   0,   0,  23,  81, -30, -17,  -5,  -5,  -1,  -2,  -1,  -1,  -1,  -1,  31, -27,  35,  -4,  -2,  -2,  -1,  -1,   0,   0,   0,   0, -58, -12,  18,  16,   2,   2,   1,   1,   0,   1,   0,   0,  -9, -18,  -2,   4,   5,   1,   1,   1,  14, -10,  -8,   0,   0,   2,   0,   1,   0,   5,  -2,  -1,  -1,   0,   0,   0,   0,  -2,   2,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },
     {   13, -18,  11,  -6,   2,  -1,   0,   0,   0,   0,   0,   0,  50,   5, -14,  -4,  -4,  -1,  -1,   0,  -1,   0,  -1,   0,   2, -52,   9,  18,   0,   3,   0,   2,   0,   1,   0,   1,  27,  30, -46,   0,   0,   0,   0,   0,   0,   0,   0,   0, -36,  52,  16, -23,   1,  -3,   1,  -2,  -5, -21,  24,   6,  -4,   1,  -1,   0,  -3,  -1,  -6,   7,   1,   0,   0,   0,  -3,  -1,   0,   0,   1,   0,   0,   0,  -2,   2,   0,   0,  -1,  -1,   1,   0,  -1,   1,   0,   0,  -1,  -1,   1,   0 },
     {  -18, -28, -78,  54,   6,   8,   0,   5,   1,   2,   0,   2,   6,  -1,  10, -26,   8,  -1,   3,  -1,   1,  -1,   0,   0,  13,  28,  39, -31, -14,   0,  -1,  -1,  -1,  -1,  -1,  -1,  20,  20, -13,  14,  -6,  -4,  -2,  -1,  -1,   0,   0,   0, -11,  -4,   9,   2,   8,  -3,   0,  -1,  -7,  -9,   5,   1,   0,   3,   0,   1,  -1,  -2,  -1,   0,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    5,  46,  -8, -17,  -4,  -2,   0,  -1,  -1,  -1,   0,  -1,  -4,  27, -31,  11,  -4,   1,  -1,   1,   0,   0,   0,   0,  -2,  42, -18, -14,   6,  -5,   1,  -2,   1,  -1,   0,  -1,  42,   7,  31, -23,  -3,  -3,  -1,  -2,  -1,  -1,   0,  -1, -48,  -4,  43,  10,  -4,   1,  -1,   0,  -1, -38,   4,  18,   4,   0,   0,   1,   3,  -6, -14,   2,   4,   1,   0,   1,   1,  -1,   0,  -4,   0,   1,   0,   0,   0,  -2,   0,   1,   1,  -2,   0,   0,   0,  -1,   0,   1,   1,  -2,   0,   0 },
     {  -27, -18,   2,  12,   1,   3,   0,   1,   0,   1,   0,   1, -17,  23,  10,  -7,  -1,  -1,   0,  -1,   0,   0,   0,   0, -58, -19,  -8,  20,   6,   3,   2,   1,   1,   0,   1,   1, -13,  48, -16, -27,   7,  -4,   2,  -2,   1,  -1,   1,  -1,  26, -35,  46,  -6,  -8,   1,  -2,   1,  23, -15, -23,  22,  -3,   2,  -1,   1,  -2,  15,  -8,  -7,   5,  -2,   1,  -1,   5,  -5,   2,  -2,  -1,   0,  -1,   0,   2,   1,   0,  -1,   3,  -1,   0,   0,   1,   1,   0,  -1,   2,  -1,   0,   0 },
     {   -9,  -2, -25, -22,  21,   3,   5,   1,   2,   0,   1,   0, -25, -23, -69,  38,   3,  10,   2,   4,   1,   2,   1,   1,  -5, -12,  33, -16,  -3,  -4,   1,  -1,   1,  -1,   1,   0, -17,  36,  23, -20,  -7,  -4,  -1,  -1,  -1,  -1,   0,  -1,  39,  27, -11,  -6,  -5,  -5,  -1,  -2,   5,  -3,  21,   0,   0,  -1,  -1,   0,  -4, -10,   2,   7,   1,   2,   0,   1,  -1,  -2,   2,  -1,   1,   0,   0,   0,   1,  -1,   0,   2,   0,  -1,   1,  -1,   1,  -1,   0,   1,   0,   0,   1,   0 },
     {   -3,  14, -12,   8,  -7,   2,  -2,   0,  -1,   0,   0,   0,  -3,  40,   5, -16,  -3,  -4,  -1,  -1,   0,   0,   0,   0,   4,  33, -39,  13,  -2,  -2,   0,   0,   0,   0,   0,   0, -18,  -3, -18,  -8,  15,  -2,   4,   0,   2,   0,   1,   0,  45,  21,  12, -39,  -5,   0,  -1,  -1, -28,   9,  54,   0, -13,  -1,  -1,  -1,   3, -33,   9,  26,   0,   0,  -1,   1,  -2,  -3, -11,   6,   5,   1,   0,   1,   0,  -5,  -1,  -1,  -1,  -1,  -1,   0,   1,  -3,   1,   0,  -1,  -1,   0,   0 },
     {   -7,   3,  27, -10,   1,  -2,   0,  -2,   0,  -1,   0,   0, -28, -22,  17,   5,   1,   1,   1,   1,   0,   0,   0,   0, -12,  34,  32, -10,  -7,  -1,  -2,  -2,  -1,  -1,  -1,  -1, -39, -13, -24,  32,   7,   2,   1,   1,   1,   1,   0,   1, -18,   6,   1, -29,  12,  -2,   3,  -1,  32, -47,  26,  10, -15,   6,  -3,   2,  11,   8, -40,  17,   2,  -1,   1,   0,   0,  10,   1, -11,   4,  -2,   1,  -1,   3,  -2,  -1,   1,   2,   2,   0,  -2,   2,   0,  -2,   1,   1,   1,   0,  -1 },
     {  -12, -23, -26, -62,  53,  -3,  11,  -2,   5,   0,   3,   0,   4,   6,  22,  -9, -28,  11,  -4,   2,  -1,   1,  -1,   0,  13,  18,  17,  56, -34,  -8,  -4,  -2,  -3,  -1,  -1,   0,  13,  -6,  12,  -6,  25, -11,   1,  -3,   0,  -1,   0,  -1,  -5,  -9,  -1,  -3,   4,   8,  -3,   2,  -4,   0,  -8,   3,  -5,   3,   1,   0,   2,  -1,   0,   3,  -2,  -2,   0,  -1,  -1,   2,  -3,   2,   3,  -1,   0,   0,  -1,   0,  -1,  -1,   0,  -1,  -2,   0,   0,   1,   0,   0,   0,   0,  -1,   0 },
     {  -16,  -8, -40,  16,   0,   9,   2,   3,   1,   1,   1,   1, -13,   1, -20,  15,   1,   1,   0,   1,   0,   1,   0,   1, -19,   3, -41,  24,   7,  -2,   0,   1,   1,   0,   1,   0, -21, -18,  -4, -12,  12,   3,   2,   1,   1,   0,   1,   0, -51,  20, -15, -22,   7,   1,   6,  -1,  53,  12,  -1, -35,  -5,  -3,   1,  -1,  15,  10,  28,  -9, -12,  -2,  -3,  -1,  -2,  -2,  13,   8,  -3,  -2,  -1,  -1,   1,  -4,   6,   3,   3,  -2,   1,   1,   0,   0,   2,   1,   1,  -1,   1,   1 },
     {    7,  23,  22,  -9,  15, -17,  -2,  -5,   0,  -2,   0,  -2,  -3,  10, -24,  19, -15,   6,  -2,   3,  -2,   1,   0,   0,   3,  27,  14, -18,  -9,   3,   0,  -1,  -1,   0,   0,   0,  14,  30, -61,  14,  11,  -2,   2,   0,   1,   0,   1,   0,  -7, -46,  -5, -11,  15,   4,   2,   1,  -8,  21,   4, -32,  -6,   1,   2,   0,  -8,  13,  33, -11, -12,  -3,  -1,   0,   7, -20,  12,  16,  -5,  -1,  -2,   1,  -1,  -1,  -6,   6,   1,  -3,   2,   1,  -1,   0,   0,   1,   1,  -2,   1,   0 },
     {   -2,   5,  22, -21, -17,   6,   1,   0,   0,   1,   0,   0,  -9,   0,   0, -59,  33,  -6,   6,  -1,   3,   0,   1,   0, -15,   7,   0,  10, -22,   7,  -5,   2,  -1,   1,  -1,   0,  11,  42,  33,  26, -27, -12,  -2,  -4,  -2,  -2,  -1,  -2,  -5,  10, -11,   1,  27,  -7,   0,  -1,  29,  -2,   3, -20,   0,   7,  -3,   1, -12, -17,  30,  16, -12,   4,  -1,   1,  -3, -11, -13,  20,   2,  -1,   1,   0,   0,   0,  -7,  -2,   2,  -1,  -2,   0,   0,  -1,  -1,   0,   1,  -1,  -1,   0 },
     {   -6,  -4,   7,   6,  36, -22,   1,  -5,   2,  -1,   1,  -1,   3,  17,  42,  19, -26,   6,  -7,  -1,  -3,   0,  -1,   0, -23, -19,  -4, -38,   3,  11,   7,   1,   2,   1,   1,   1, -14,  22,  10, -18,  -7,  -1,   0,   1,   0,   0,  -1,   0, -24,  -6, -18,  35,  -1,  -6,   1,   0,   1,  -3,  28, -22,   5,  -2,  -2,   0,  29, -32,   6,  25, -13,   6,  -4,   2,  -2,  13, -31,   5,   8,  -3,   3,  -1,  -1,   2,   4,  -7,   2,  -1,  -3,   1,   1,   0,   1,   1,   0,   1,  -2,   0 },
     {    2,  11,   5,  20,  39, -26,  -3,  -4,   1,  -2,   0,  -1, -15,  -7, -28, -27,  12,  13,   1,   2,   0,   1,   1,   1, -16, -13, -30,  -3, -49,  30,   4,   4,   1,   2,   0,   1, -18,  -8,  -1,  20,   7, -19,   6,  -2,   2,  -1,   1,   0,  -3,  16,  11,   9,  17, -10,  -4,  -1, -36,  -5, -10,   2,  13,   7,   1,  -1,  32,  -3, -16, -34,   1,   3,   2,   0,   8,   6,  16,  -9, -13,  -1,  -2,   0,   2,   0,   7,   5,   0,   1,   4,   2,   2,   0,   0,   1,   0,   1,   2,   0 },
     {  -14, -11,  -7, -22, -40,  31,   3,   6,   0,   3,   0,   1, -10,   6,   0,  12,   7, -12,   3,   0,   1,   0,   0,   0, -12,   9,   9,  29,  25, -17,  -6,  -3,   0,  -1,   0,  -1, -14,  26,  -9,  10,  -6,   8,  -4,  -2,  -1,  -1,   0,   0, -15,  -8, -21,   8,  11,  -2,   6,  -1, -58, -18,  -2, -19,  21,   0,   6,   0,  42, -17,  -5, -25,  -8,   7,   0,   1,  17,  15,   0,  -3, -12,  -2,  -1,  -1,   1,   5,  10,   0,   2,  -1,   3,   3,   1,  -1,   0,   2,  -1,  -1,   0,   0 },
     {   -5,  -2, -13, -35,  11,   4,   5,   0,   2,   0,   1,   1,   5,  18,  32,   1, -14,  -1,  -3,  -2,  -2,  -1,  -1,   0,   0,  16, -19, -29,  14,   0,   1,   0,   1,   0,   0,   0, -12,   1, -16,  12, -20,   8,   1,   3,   1,   0,   0,   0,  27,  45,   4,  24,  -6, -20,   1,  -6,   9, -30,  -5,  -5,  31,  -1,   0,  -1, -17,  27,  13, -30, -10,  11,  -2,   2,  -3, -19,  30,  14, -15,   1,  -1,   0,   3,  -9, -17,  19,  -4,   2,   0,  -2,   1,  -2,   0,   1,   0,  -1,   1,  -1 },
     {   -6,  -8,   2, -31,   5,  10,   2,  -2,   2,   0,   1,   0, -11, -17,  -7, -25,  11,  -3,   5,   0,   3,   0,   1,   0,  -4,   0,  23, -19,  11,  -8,   2,   1,  -1,   0,  -1,   0, -20, -23, -17,   0, -14,  11,   2,   4,   1,   1,   0,   1, -23,   4,  58,   4, -30,  -6,  -3,  -1,   1,  52,   3,  16,   8, -14,  -3,  -4,  37, -16,  14,   0,   2,   4,  -4,   2,   6, -18, -12,  21,  -3,   7,   0,   2, -10,   8, -14,  -1,   1,   0,   0,   1,   1,   1,  -2,   1,   1,   0,  -1,   2 },
     {   -6, -16, -30, -41,   6,  11,   6,   2,   4,   1,   1,   1,   8,   3, -11,   7,   7,  -4,   3,   0,   0,   0,   0,   0,  -5, -25, -43, -22,  13,  10,   2,   2,   1,   1,   1,   1,  13,  -8,  -1,  46, -28,  -1,  -4,   1,  -1,  -1,  -1,   0,   7, -37,  15,  -5,  19, -12,   1,   0,   0,  15,  20, -16, -14,   6,  -6,   2,  -4,  19, -29,  24,   5,  -2,   2,  -2,   5,   8,  -2, -28,   6,  -2,   2,   1,  -2,   1,  17,   1,  -2,   2,   1,   0,  -1,   1,   3,   3,   0,   0,   0,   0 },
     {   -5,   4,   1,  11,   4,  19, -17,   4,  -3,   0,  -1,   1,  -1,  15,  14,  59,   7, -11,  -4,  -7,  -2,  -4,  -1,  -1,  -8,  11,   3,  20, -30,   7,   3,  -2,   1,   0,  -1,   1, -13,  -1,  -6,  21, -19,  -3,   6,   3,   1,   0,   1,   0, -14,  20,  15,   7,  26, -21,  -2,  -3,  -1,  25, -19,  12,  17,   6,  -3,  -3, -38, -27,  -1,   1,  23,   2,   8,   0,  19, -28, -22, -16,   7,   6,   2,   2,   3,   9,  -6, -18,   4,  -2,   7,  -8,   1,  -1,   3,  -3,   1,  -1,   1,  -3 },
     {   -6,   1,  -2,   5,  -8, -21,  11,   4,   0,  -1,   2,  -1, -12,  -5, -10, -25, -47,  30,  -1,   8,   1,   2,   0,   2,  -9,   4,  -5,   3,  43, -10,   0,  -7,   2,   0,   0,   0,  -4,  22,   6,  48,  30, -16,  -8,  -3,  -3,  -1,  -1,  -1, -12,  16,  -3,  14, -19,  18,  -1,  -1,   8,  19,  -4,  -3,   7, -11,   6,  -3, -28, -29, -13,  -6,  26,  -2,   2,   3,  10, -12,  -1, -26,  -7,   3,   0,   2,   6,   2,   2,   0,   1,   0,   2,  -3,   1,  -1,   4,   2,   1,   0,   0,   0 },
     {    0,  -5,  -1,  12, -24,   7,  -2,   4,  -3,   3,  -1,   0,  -1, -12,  -1,  23, -33,  10,  -6,   3,  -1,   1,   0,   1,   0, -16,   0,  -3,   0, -10,   9,   1,   2,  -1,   0,   0,   2, -10,  24,  32,  25, -24,  -4,   1,  -2,  -2,  -1,   0,  15,  11,  43,   2,  22,  12, -16,  -3, -15, -15,  -9, -19,  -8,  24,   5,   1,  21,  21,  11,  11, -31,  -8,   4,  -2,  -9, -13, -15,  37,  19, -11,   0,   0,   1,   4, -25, -16,  -2,   1,   8,  -9,   1,   0,  -3,  -4,  -1,   0,   0,  -1 },
     {   -4,  -5,   2, -10, -11,  -1,   1,   4,   1,   1,  -1,   0,   7,  14,  33,  23,   9, -12,  -9,  -3,  -1,  -2,  -1,  -1,  -8,  -8,   4, -33, -10,   3,  -1,   0,   1,   3,   0,   0,   6,  21,  36,   6,  12, -13,   2,  -5,  -2,  -2,   0,  -1, -11,  11,  -8, -39,   8,  24,  -2,   3,  -8,  26,  -5,  15, -35,  -1,   5,  -1,  25,  -7,   0, -10,  24, -12,   4,  -1,   1, -15,  38,  -1, -24,   9,  -7,   4,  -2,  -6, -16,  36,   1,  -1, -10,  -8,   1,   0,  -1,   3,  -2,   2,  -1,  -1 },
     {    3,   5,   7,  -5,  10,  65, -32,  -6,  -4,  -1,  -2,   2,  -1,  -2,  -3, -14,   1,  -5,  17,  -5,  -1,   0,   2,  -1,  -8, -11, -14, -23, -30, -45,  25,  10,   3,   1,   1,   0,  -3,   1, -11,  -6,  32,  -9, -21,   6,   2,   0,   0,   2,  -4,  -2, -11,  32,   1,  26, -13,  -8,   6,  -3,  24,   4,  -6,   0,  11,  -6, -11,   9,   4, -11,  13,  -5,   4,   4,  11,   3, -10,  -7,  -4,   4,  -2,   1,   1,  -8,   2,   5,  -3,   3,  -5,   1,   0,   1,   0,   0,   2,   1,   2,   0 },
     {    8,  11,  10,  21,  50,   8, -28,  -9,  -2,  -2,   0,  -2,   4,   5,   5,  -1,   5,  15,  -1,  -3,  -2,  -1,  -3,   1,   2,   4,  -1,  12,  19, -49,  31,   1,   0,  -2,   0,   0,   4,   9,  10,  17, -47,   9,  -8,   8,  -1,   2,   0,  -1,   7,   3,   6, -39,  -9,  20,  -8,  -1,  -2,  -3, -29, -22,  16,  -8,   7,  -1,  11, -10, -13,  14, -13,  -7,  -1,   3,  -7,   5,  17,  -1,   0,  -3,  -2,   0,  -2,   3,   2,   1,   2,  -3,  -1,   1,   0,  -2,   0,   0,  -1,   1,   1,   0 },
     {    5,   7,   3,  -2, -20,   9,  -6,   1,   1,   1,   0,   0,   8,  12,  -5,  11, -15,  -9,   1,   2,   1,   1,   1,   1,  12,  16,   5,   4, -16,   3,   3,  -1,  -2,  -1,  -1,  -1,  17,  17, -19,  10,   6,  -7,  -1,  -1,  -1,  -1,   0,  -1,  25,  14,   7,  23,   8,  -8,  -9,  -2,  42,  31, -17,  -1,  10,   4,   0,  -3,  44, -15, -22,  -3,   2,   6,   3,   1, -59,  26,   9, -11,   2,   0,   3,   0, -16, -19,  24,  -1,   5,  -7,  -8,   7,   1,   0,  -1,  -3,  -2,   0,   0,  -3 },
     {   -4,   7,   0,   7,   6, -14, -11,  -1,   4,   1,  -1,   0,  -8,   8,  -1,  -5,   5, -11,   5,  -4,  -2,  -2,  -1,   0,  -7,  11,  -2,  18,   1,  -3,  -3,   0,  -7,   1,  -1,   0,  -6,  24,   9,   8,  -4,  13, -10,   3,   0,   1,  -1,   1, -12,  14, -21,  21, -15,  -5,   7,  -1, -23,  17,  15,   7,  -3, -16,   1,   0,  24,  66,  -9,  19,  24, -16,  -1,  -6, -16, -35,   8,  -4,  42,   9,   1,   1,  -7, -17,  -9, -27,   4,   0,   1,  -4,   3,  -7,   0,  -4,  -1,   2,   3,  -4 },
     {   -1,   1,  -7, -22,   8,  -4,  22, -16,  10,  -2,   5,  -1,   4,  11,   1,   5,  41,  -1,  -4,   3,  -7,  -3,   1,   0,  -2,   6, -12, -35,   6,  -2, -11,   9,   3,   2,  -3,  -1,  -3,   4, -18,  12,  24,  -5,   3,   0,   8,  -4,   0,   0,   4,  20, -11, -13, -16,  23,  -5,  -4,  -5, -19, -54,   2, -13,  -4,   9,  -3,   3,  -4,  23,   4,  14,  -8,  -1,   4,  -4,  18, -26,  -4,  22,   5,  -3,   0,  -8,   6,  14, -41,   0, -13,  21,   6,   1,  -2,  -4,   1,   2,   2,   7,   0 },
     {   -1,  -3,   1,   1,  22, -34, -11,  -7,   7,  -3,   1,   0,   1,  -1,   4,   7,  41, -50,  12,  -1,  -3,  -1,   1,  -1,   1,   2,  11,  11,  23,  -3, -11,   3,  -6,  -1,  -1,  -1,   1,   2,  12,   2,  37,  31, -18,   4,  -2,  -2,  -1,   0,   4,   9,  30,  17,   6,  16,  -5,  -6,   1,   7,  11, -40,  -8,  18,  13,   0, -18,  -7, -21, -17, -14,  16,  10,   5,  -5,  -7,  -8,  -5, -10,  -1,  -2,   2,  -1, -13,   2,   8,   0,   6,  -9,  -1,   3,   0,   0,  -3,   3,   1,   2,  -2 },
   }
 },
 {  //28
   {
     {  103, -28, -20,  -2,  -4,  -1,  -2,   0,  -1,   0,  -1,   0, -51,  35,   7,  -3,   1,  -1,   0,   0,   0,   0,   0,   0, -14, -13,  12,   3,   1,   1,   0,   0,   0,   0,   0,   0,   6,  -2,  -6,   3,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   3,   0,  -2,   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -16,  59, -18,  -6,  -3,  -2,  -1,  -1,  -1,   0,   0,   0, -66, -24,  44,   1,   4,   1,   2,   0,   1,   0,   1,   0,  36, -48, -18,  15,   1,   3,   0,   2,   0,   1,   0,   1,  10,  22, -19, -10,   1,  -1,   0,  -1,   0,   0,   0,   0,   3,   1,   8,  -4,  -3,   0,  -1,   0,   3,   2,  -1,   2,  -1,  -1,   0,   0,   1,  -1,   1,   0,   1,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0,   1,   0,   1,   0,   1,   1,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0 },
     {   51,  64, -26, -20,  -5,  -5,  -2,  -3,  -1,  -1,  -1,  -1,  49, -44,  22,   6,  -5,   1,  -2,   1,  -1,   0,  -1,   0, -42,  19, -16,  12,   5,   0,   1,   0,   1,   0,   0,   0, -20,   2,  16, -10,   4,   0,   1,   0,   0,   0,   0,   0,   5, -13,   0,   8,  -3,   2,   0,   1,  -4,   5,  -4,  -1,   2,  -1,   1,   0,   0,  -3,   1,   0,   0,   0,   0,   0,  -2,   2,  -1,   0,   1,   0,   0,   0,   0,  -2,   1,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,  -1,   1,   0,   0 },
     {   21, -35,  16,  -2,   1,   0,   0,   0,   0,   0,   0,   0, -23, -39,  -9,  22,   3,   5,   2,   3,   1,   1,   0,   1,  48,  40, -54, -19,   2,  -1,   1,  -1,   0,   0,   0,   0, -17,  35,  38, -22, -10,  -3,  -2,  -2,  -1,  -1,   0,  -1,  -5, -18,  12,  21,  -4,  -2,  -1,   0,  -3,   3,  -9,   2,   7,  -1,   0,   0,  -1,  -2,   1,  -3,   1,   2,   0,   0,  -1,   2,   0,  -1,  -1,   0,   0,   0,   0,  -1,   1,   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0 },
     {    7, -36, -26,  20,   4,   3,   2,   2,   1,   0,   1,   1,  49,   8,  16, -31,  -2,  -2,  -2,  -1,  -1,  -1,  -1,  -1,  22, -23, -11,  34, -11,   3,  -2,   2,  -1,   0,  -1,   1, -31,  49, -36, -16,  17,  -4,   3,  -1,   2,  -1,   1,  -1,  -4,  -3,  40, -18,  -9,   4,  -2,   0,  -1,  -4,  -4,  19,  -3,  -2,   1,   0,  -4,   2,   2,  -4,   5,  -1,   0,   0,  -1,  -1,  -1,   2,  -1,   1,   0,   0,  -2,   0,   1,  -1,   0,  -1,   0,   1,  -1,   0,   1,  -1,  -1,   0,  -1,   1 },
     {    3, -32,  34,  -9,   3,  -1,   1,   0,   1,   0,   0,   0,   2, -45,  -3,  26,  -2,   4,   0,   2,   0,   1,   0,   1, -50, -18, -30,  10,  16,   4,   4,   2,   2,   1,   1,   1,  51,   9, -36, -27,  -3,   4,   0,   1,   0,   1,   0,   0,   1,  40,   8, -15, -15,  -4,   0,  -2,  -2,   0,  16,   6,  -2,  -5,  -1,  -1,  -1,   2,  -1,   4,   2,   1,  -1,   0,   0,   1,   1,  -1,   1,   0,   0,   0,  -1,   2,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {   26,  25,  74, -27, -17,  -5,  -5,  -3,  -3,  -2,  -1,  -1,  21,  17, -33,  26,   2,  -5,  -1,  -1,   0,  -1,   0,   0,  26, -50,  -5,   3,  12,   6,   0,   2,   0,   1,   0,   1, -44,   6, -11,   9,   0,   5,   2,   1,   1,   1,   1,   0,  -2,   1,  -7,  -9,   6,  -2,   2,   0,   5,  -1,   4,  -4,  -6,   2,  -1,   1,  -3,   6,  -3,   1,   0,  -2,   1,  -1,   0,  -1,   3,  -1,   1,   0,  -1,   0,  -1,   1,  -1,   1,  -1,  -1,   1,  -1,   0,   0,  -1,   0,   0,   0,   1,   0 },
     {  -17,  17,   9,  -7,  -2,  -1,   0,   0,   0,   0,   0,   0,  -4,  61,  -9,   0, -10,  -3,  -1,  -1,  -1,  -1,  -1,   0, -34, -17, -16,   1,  23,   0,   4,   0,   2,   0,   1,   0,  13,   9,  29, -48, -10,   5,   0,   0,   0,   0,   0,   0,   8, -19,  40,  36, -26,  -7,  -2,  -1,   1, -12, -26,  22,  21,  -5,  -2,   0,   0,   0,  -2, -12,   4,   7,  -1,   0,   3,  -3,  -2,   2,  -3,   1,   2,   0,   1,   1,   1,  -1,   1,  -1,  -1,   1,   0,   0,   0,   0,   1,  -1,  -1,   1 },
     {  -12, -31,  19,  -2,   8,  -3,   1,   0,   1,   0,   1,   0, -14,   8,  43,   3, -19,   2,  -4,  -1,  -1,   0,  -1,   0, -11,   0, -26,  43,  13,  -6,   1,  -2,   0,  -1,   0,   0, -36, -42,   0, -14,  30,  12,   2,   4,   1,   1,   1,   1,  28, -21, -34,   0, -14,  13,   4,   2,   4,  34, -15, -20,  -3, -10,   4,  -1,   7,   2,  16,  -6,  -6,  -2,  -4,   1,   0,   3,  -2,   4,   0,  -1,   0,  -1,   1,   0,   2,  -2,   0,   2,  -1,   0,   1,   0,   1,  -1,   0,   1,  -1,   0 },
     {   -4,  22,   2,   6,  -7,   0,  -1,   0,  -1,  -1,  -1,   0, -34,   4, -36,   1,  13,   3,   3,   2,   1,   1,   1,   1, -12,  41,  -2,  -3,  -8,  -5,   1,  -3,   0,  -1,   0,  -1, -30, -11,  -9, -17,  30,  -1,   3,   1,   2,   0,   1,   0,  33,  -2,  26, -51, -25,  13,  -1,   2,   5,  -4,  28,  35, -27, -13,   2,  -1,   2,  -4, -10,  10,  19,  -5,  -4,   1,   1,  -3,   3,  -1,  -1,   6,  -1,  -1,   2,  -2,  -3,   0,   0,  -1,   2,   1,   1,  -1,  -1,   0,   1,  -1,   1,   1 },
     {    1,   7,  -6, -50,  18,   4,   1,   3,   1,   1,   1,   0,  11,  26,  19,  23, -41,  -1,  -4,  -1,  -2,   0,  -1,   0,  39,  35,  -1,  25,  14, -23,  -3,  -4,  -2,  -2,  -1,  -1,  21, -16,  -4,  -7,  19,  16,  -5,   0,  -1,   0,  -1,   0, -47,  19,   6, -11,  -8,   6,   8,  -1,   2, -29,  18,   7,  -2,  -4,   0,   3,  -2,  -4, -14,  11,   3,   2,  -1,   1,  -2,  -1,  -2,  -6,   3,   1,   1,   0,   0,  -1,  -1,   1,   0,   0,   0,  -1,  -1,  -1,   0,   1,   0,  -1,   0,   0 },
     {   -8,  -7, -59, -12,  15,  10,   4,   5,   1,   2,   1,   1,   6,  11, -25,  19, -13,  -3,   3,   0,   1,   0,   0,   0,   9, -23,  -9, -20,  27,  -1,   0,   1,   0,   1,   0,   0, -31,   3,   6, -24, -26,  15,   2,   2,   1,   1,   0,   1,  31,  33, -18,   0, -10, -23,   4,  -3,  -1,  16,  42, -20,   5,  -2, -10,   0,   0, -10,  15,  23, -10,   4,  -1,  -2,  -1,   0,  -5,   3,   8,  -3,   1,   0,   0,  -2,   2,   1,  -1,   0,   1,   0,   1,  -1,   1,   1,  -1,   0,   0,   0 },
     {   -8, -21, -12, -60,  18,  15,   8,   4,   2,   2,   1,   0, -15, -35, -28,  15, -13,   2,   4,   5,   3,   1,   1,   1,  -8, -15,  61,  23, -12, -16,  -8,  -2,  -2,   0,  -1,   0,  -9,  26,   6,   4,   9,   1,  -5,  -2,  -1,   0,  -1,   0,  25, -17,  -3,  12,  -8,   5,   0,   1,  -5,   0, -24,  15,   8,  -3,   2,  -1,  -3,   9,  -7, -18,   8,   3,   2,   0,   5,   1,   5,  -3,  -7,   1,   0,   1,   1,   4,  -2,   0,   0,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   1,   0 },
     {  -18,  -4, -12, -17,   3,  10,   2,   1,   1,   1,   1,   0, -13,  21,  -5,  26, -11,  -9,  -1,  -1,   0,  -1,   0,   0, -38,   5, -21,   4,   6,   4,   1,   0,   1,   1,   1,   0,   0,  44,   2,   9,   4, -21,   2,  -6,   0,  -3,   0,  -2, -13, -34,   7, -22,  48,  15,  -5,   1,  23,  -7,   1, -34, -37,  22,   8,   1,   3,   5,  12,  18, -14, -22,   4,   2,   5,  -4,   0,   0,   9,  -1,  -6,   1,   0,  -1,   2,   2,   2,  -2,  -1,  -1,   0,  -1,   1,   1,   2,  -1,   0,  -1 },
     {   -6, -12, -19,  13,   6,   3,   0,   2,   0,   0,   0,   1,  -6, -21,  14,  27,  -2,  -5,   2,  -3,   0,   0,   0,   0, -13, -10,   6, -29,  21,   8,   1,   2,   1,   1,   0,   0, -38, -32, -13,   4, -16,  15,   8,   4,   2,   2,   1,   1, -39,  -6,  17,  -4,  -6, -15,   3,   2,  57, -12, -23,  16,  -6,  -3,  -9,   1,   1,  39, -16, -14,  13,  -2,   1,  -4,   1,  -5,  19,  -9,  -5,   4,  -1,   2,   0,   4,  -3,   5,   3,   1,   1,  -3,   0,   2,   0,   2,   1,   0,   1,  -2 },
     {   -4, -19,  21, -28,   3,   5,   2,   1,   2,   0,   1,   0,   5,   5,  62,  -9,  -9, -10,  -2,  -1,  -1,  -1,  -1,  -1,  -7,  -3,   3, -20,  -5,  14,  -2,   0,   0,   2,   0,   1, -12,   0,  19,  18, -29, -13,   3,   0,   0,   0,   0,  -1,  40,   4,  12,   4,  12, -11, -13,   1,  -7,  -4,  21,  34, -17,  13,  -5,  -2,  -7,  -7, -36,  31,  21,  -6,   8,  -2,  -5,   0, -10, -18,  10,   7,   1,   3,   0,  -2,  -5,   1,  -1,   1,  -3,  -5,   1,  -1,  -2,   2,  -1,   1,  -2,  -1 },
     {   -2,   5,   6,  22, -14,   4,  -4,  -1,  -2,   0,  -1,   0,  -3,  10,  22,  47,   1, -24,  -1,  -4,   0,  -3,   0,  -1,   2,  39,  24,  -4,  26,   8, -15,  -3,  -4,  -2,  -1,  -1,  -7,  17, -52,  -4,  -7,  17,  18,  -4,   2,   0,   2,   0,  25, -17,   1,  -1,  10, -22,   3,   4, -41, -14,  -5,  -5,  23,  17,  -7,   1,   4, -10,  -3, -21, -11,   9,   9,  -1,   4,  -1,   4,   0,  -6,  -5,   1,   3,   1,   0,  -1,  -3,   0,  -1,   2,   0,   1,  -1,  -1,  -1,   0,  -1,   1,  -1 },
     {   -4,   4,  10,   3,  16, -13,  -6,  -1,   1,   0,   0,  -1, -24,  -7, -11, -26, -22,  18,   5,   1,   1,   2,   1,   0, -19,  11, -18,  28, -15, -10,   3,   1,   0,   0,   0,   0, -37,  -6,   1,  12,   5,  -9,   2,  -1,   2,  -1,   1,   0,   9,  43,  19,  12,  27, -25, -12,  -7,  12, -37,   6, -12,  34,  33,  -6,  -1,   3,  -2, -15, -15, -28,  12,  17,   1,   3,   2,   7,   0,  -2, -13,   1,   2,   5,  -1,   1,  -1,   0,  -1,   1,   2,   2,  -1,  -1,  -1,   0,  -1,   1,   1 },
     {   10,  14,  24,  25,  76, -20, -19,  -6,  -6,  -3,  -1,  -2,   1,   0,   8,   2, -34,  34,   9,  -5,   0,  -1,  -1,   0,  -3,  -2,  19, -37, -17, -19,  11,  10,   3,   2,   0,   0,   0,  16, -23, -11,   2,   5, -12,   1,   2,   1,   0,   1,   3, -31,   1,   5,  -8,  -1,   8,  -2,   2,   7,  11,  -3,  -2,  -4,  -7,   3,  -2,   6,   7,  18,  -8,   2,  -1,  -4,   1,  -1, -10,   4,  10,  -3,   3,   1,  -1,   1,  -2,  -3,   1,  -1,   0,   1,  -1,   0,  -1,   0,   0,  -1,  -1,   0 },
     {   -3,   1, -22,  30,  13, -12,  -4,   1,   0,   1,  -1,   0,   6,  15,  -3,  47, -20,  13,  -5,  -3,  -2,   0,  -1,   0,  -6, -19, -26,  -7,  -5,  -5,  19,   3,   3,   0,   2,   0,  -4,   9,  10,  31,  18, -19, -11,   1,  -2,  -2,  -1,  -1,  11,  20, -21,   2,  10,  24,  -2,  -7, -23, -10,  -7,  12,   8, -21,  17,  -1,  -7, -14, -23, -28,  35,   0,  -5,   8,   7,   4,  13, -33, -15,  11,  -2,  -1,   3,  -1,   7,   3,   1,   2,   3,  -4,   1,   1,   3,   1,   0,  -1,   2,  -3 },
     {    5,   3, -12, -42,  24,   8,   2,   1,   2,  -1,   2,   0,  11,  12,  -5,  10,   4,  -3,  -1,   0,  -1,  -1,  -1,   0,  18,   4, -24, -31, -30,  21,  12,   1,   2,   2,   2,   0,  10, -41, -37,  -5,   4, -32,   8,  10,   3,   1,   1,   1,  16,   4,  27,   7,  10,   3, -29,  -3,  -9,  24, -15,   3,   0,  17,  11, -10,   0,   5,  23, -26,  -6,  -9,   3,   6,  -2,  -9,  13,  18, -14,   1,  -6,   0,  -3,   3,  -5,   6,  -1,  -2,   2,   1,  -1,   1,  -2,   1,  -1,  -1,   1,   0 },
     {   10,  -2,  -8,  25,  16,  -5,  -8,   1,  -4,   0,   0,   0,   9, -17, -11,   0,  -6,  -4,   9,  -1,   3,  -1,   2,   0,  21, -22,   8,   1,  32,  -2, -12,  -5,  -2,   0,   0,  -1,  16, -27,  26, -23,  10,  14,   2,   1,   0,   0,   0,   0,  24,   2,  11,  -8,   6,  23, -11,   0, -28, -26, -19,   1, -33,  34,  21,   1,   7,  18, -24,   2, -27, -34,  15,   8,  -3,  14,   8,   0,  12,  -8, -16,   0,  -3,   3,  -2,   0,  -3,   3,  -2,   1,   0,   0,  -2,   0,  -2,   2,   0,   0 },
     {   -7,  -7,   4, -11,  12,  18,  -1,  -2,  -2,   0,   1,   0,  -5,   1,  16,   4,  34,  -8,  -5,  -3,  -3,   0,  -2,   0, -16, -12,   9, -13, -31,   9,  11,   5,   2,   2,   1,   0, -28,  -1,   4, -35,   5,  -6,  -1,   5,   1,   1,   0,   1, -34,   0, -14,  10,  -3,  11,   3,   1, -33, -33,  23, -10,  -4,  -8,   8,   4,  61,  -4, -14,  -3,  -8,   0, -16,   1,   5,  28,   4,   9, -13,  -1,   2,  -6,   3,  -3,   4,   8,  -1,   5,  -2,   3,   3,  -3,  -1,   0,  -1,   2,  -1,   2 },
     {   -2,  -5,   2, -28,  -5,  -6,   7,   8,  -1,   0,   0,   0,   1,   0,  15, -39,  -9,  13,   9,  -7,   3,  -1,   2,   0,  -7,  -3, -18, -42,  33,  -6, -10,   9,   0,   4,   0,   2,   0,  12,   5,  30,   0,  17, -15,  -5,   2,  -3,   0,  -2,  14,  -2,   1, -34, -10,  12,  27,  -2, -18, -26,  -5,  -6,  -4, -11,  -1,  11,  -1,   2,  15, -26, -14,  13, -10,  -3,  17,  -1,  23,  23, -32,  -5,  -1,  -2,  -1,   6,  -9,  23,  -1,  -1,   1,   2,   0,   0,  -3,   6,   0,   0,   0,   1 },
     {   -2,  -3, -12,  -7,  -2,   9,   1,   3,   0,   0,   1,   0,  -4,  -7, -31, -20, -17,  10,  17,   5,   2,   3,   1,   1,  -4, -14, -32,  12,  11, -10,  -6,   4,   2,   1,   0,   0,   6, -26, -19,  26, -13,   0,   8,  -5,   2,   0,   1,   0, -22, -51,  13,  -4,   4, -21,   3,   8, -41,  23,  10,   8,   7,   3, -23,  -2,  30,  -5, -25,  12,  12,   4,  12, -10,  10,  23,   5, -26,   5,  -1,   0,   5,   1,   3,  13,  -1,  -1,   2,   1,   2,   2,   1,   3,  -1,   0,   1,   1,  -1 },
     {    3,  -5,  -7,  -5,  14, -16,   9,  -4,   0,  -2,   2,   0,  13,   6,   2,   9,  30,   6, -17,   1,  -2,   0,  -2,   1,   8, -11,  -1,   0,  -5,  12,   9,  -3,   3,  -1,   0,   1,  23,   0,  17, -22,  21,   1,   3,   0,  -3,   2,  -2,   1,   9, -30, -19, -31,   5,  -2,  -2,  12,  30,   7,  17,   0,  14,  22, -26,  -8, -16, -11, -34, -10, -19,  37,  25,  -3, -10,   7,  10, -17, -35, -23,   9,  11,  -1,  -1,   4,  17,   0,   4,  -3,   0,  -3,   2,   0,   4,  -1,   0,  -1,  -1 },
     {   -7, -11,  -9,  -8, -27, -63,  20,  16,   5,   3,   2,   1,  -3,  -6,   3,  20,  -4,  31, -28,  -8,  -2,   0,   1,   1,   1,  -4,  22, -10,  20,  17,  12,  -8,  -4,  -3,   0,  -1,  -8,  -2,  -1,  19,  12, -23,  -4,  10,   1,  -1,   0,   0,  -5,  -2,  38,   3, -20,  20,  -4,  -5,  -6,  11,   0, -15,  14,  -5,   2,   1,  26, -11,  -7,  20, -20,   5,  -1,  -1, -16,   8,  -9,  10,  16,  -9,   6,   0,   2,   2,  -3, -10,   1,   0,  -3,   0,   0,   0,   0,  -3,  -1,   1,  -1,   0 },
     {    0,   2,  -7, -12,   6, -35,  13,   4,   5,   1,  -1,  -1,   1,   9,  -9, -11,   7,  36, -20,   2,  -4,   3,   1,   1,   8,  14,  -8,  24,  -6,  32,  15,  -7,  -3,  -2,  -2,   0,   5,   4, -13, -13, -35, -11,  25,   9,   3,   0,   2,   1,  12, -13, -42,  -6,  -6, -17,   0,   9, -11, -42,  -9,   4,  -6, -13,  -7,   2, -13,  39,   3,   6,   6, -19, -13,  -8,  15,   1,  15,  -5,  14,   7,  -3,  -5,  -1,  -2,   5,   4,  -1,  -2,   2,   0,  -1,   1,   1,  -1,  -1,  -1,   2,   1 },
     {   -1,  -6, -12,  -1, -40,  15,   7,   0,   1,   1,   0,   0,   8,   2,   4,  21,  10,   5,  -2, -11,  -2,   1,   1,  -1,   2, -10, -19, -11, -33, -31,  13,  19,   4,  -2,   1,   2,   9, -14, -10,   4,  26,   8, -34, -10,   4,   0,   0,  -1,  26, -14,   1,  19,   9,   1,  21,  -9,  -4, -26,  10,  10,   5,   1, -14,  14, -10,  51,   1,  13, -17,  12,   2, -14,  -4,  -6,  -6,   4,  27, -14,   9,   4,  -1,  -4,  -2, -15,  -3,   0,   3,   3,   1,   2,   0,  -4,  -1,   0,   0,  -1 },
     {    2,   1,   0,   5,  34,   7,  -6,   4,  -4,   4,  -2,  -2,   4,   4,   3,  33,  16,  -8,   6,  -4,  -3,  -1,   0,  -1,   1,  -7,  -6,  24, -29,  26,  -5,   1,   1,  -1,   0,   2,   8, -10,  36,  29,  -6,  10,   6, -14,  -4,  -5,  -2,  -2,   8,  11,  31, -37,  -4, -31,  20,  13, -12,  -5, -32,  -7,  -7, -25, -25,  15,   9, -10,  11,   4, -14,  11, -13, -12, -10,   4,  -5,  16,  14,  -3,   4,  -6,  -1,   0,  -6, -11,  -2,  -1,  -1,  -3,   3,   1,   0,  -2,  -1,   2,  -1,   0 },
     {   -1,   5,   2,  -9,  10,  25,  -1,  -6,  -5,  -1,   4,  -2,   0,  12,   5,   1,  43,  31, -15, -24,  -1,  -1,   2,  -2,  -2,  15,  -2,  15,   2, -34,  23,   7,  -8,  -1,  -1,  -4,  -9,   5, -12,   3, -50,  14, -22,  16,  13,   0,   0,   1,   0,  12,   9, -16,  -9,  17,  11, -20,  -4,  23, -27, -11,   1,   3,  20,   1,  -4,  -9, -19,   5,  -9,  -8,  18,  -1,  -6,  -2,  16, -20,   8, -16,  -1,   5,  -7,   4,   7,  -2,   3,  -5,  -2,  -5,   5,  -7,  -5,   0,   0,  -1,  -2,  -2 },
     {    7,   6,   2,  -4,  14,  19,  20, -12,  -3,  -6,  -3,   0,   6,   5,  -6, -12,   9,  -3,  14,  21,  -2,  -1,  -2,   0,   7,   9, -10,  -8,   7,  14, -29,  -4,   7,  -1,   1,   1,  11,   5, -26,  10,  13,  -7,   3, -17,  -4,   5,  -1,  -1,  31,  -1,   4,  22, -18,   7,   4,  -5,  31, -15, -15, -43,   9, -20,  13,  18,  16,   0, -36,  22,  -1,   4, -17,   7, -32,   6,  19, -21,   2,  15,   9, -13,   2,  -8,  15,   1,  -3,   3,  -3,   8,   1,  -1,   0,  -3,  -2,   2,   3,  -1 },
   },
   {
     {  113, -37,  -9,  -5,  -3,  -2,  -1,  -1,  -1,  -1,   0,   0, -33,  25,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0, -19,   1,   4,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -5,   1,   3,   0,   0,   0,   0,   0,  -1,  -1,   1,   1,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   14, -26,   7,  -1,   2,   0,   1,   0,   0,   0,   0,   0,  99,  -8, -23,  -1,  -5,  -1,  -2,   0,  -1,   0,  -1,   0, -43,  49,  -1,  -2,   0,  -1,   0,  -1,   0,   0,   0,   0, -19,  -6,  14,  -1,   2,   0,   1,   0,   0,   0,   0,   0,  -2,  -3,   1,   2,   0,   0,   0,   0,  -5,  -1,   2,   1,   0,   0,   0,   0,  -1,  -1,   0,   1,   0,   0,   0,   0,  -3,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {  -24,   4,   4,   3,   1,   1,   0,   0,   0,   0,   0,   0, -10,  49,  -9,  -3,  -2,  -2,  -1,  -1,  -1,  -1,   0,   0, -76, -20,  40,   1,   7,   0,   2,   0,   1,   0,   1,   0,  39, -55, -11,   9,   0,   3,   0,   2,   0,   1,   0,   1,   8,  14, -17,  -2,  -1,  -1,   0,   0,   5,  -2,   0,  -3,   1,   0,   0,   0,   3,   3,  -1,  -2,  -1,   0,   0,   0,   3,  -2,   1,   0,  -1,   0,   0,   0,   1,   1,  -1,   0,   2,  -1,   0,   0,   1,   1,  -1,   0,   1,  -1,   0,   0 },
     {  -15, -73,  20,   4,   6,   3,   3,   1,   2,   1,   1,   1,  25,  11, -13,  -3,  -2,  -1,  -1,   0,   0,   0,   0,   0,  20, -43,  -1,   8,  -1,   3,   0,   2,   0,   1,   0,   0,  29,  32, -48,  -2,  -2,   0,  -1,   0,  -1,   0,   0,   0, -20,  40,  17, -16,   0,  -4,   0,  -2,   1, -10,  11,   6,  -2,   1,  -1,   0,  -5,   7,  -2,   0,   1,   0,   0,   0,   0,  -2,   2,  -1,   0,   0,   0,   0,  -2,   3,   0,   0,  -1,  -1,   1,   0,  -1,   2,   0,   0,   0,  -1,   1,   0 },
     {   30,  76, -31,   0,  -9,  -2,  -4,  -1,  -2,  -1,  -1,  -1,  34, -24,  15,  -6,   2,  -2,   0,  -1,   0,   0,   0,   0, -30, -36,  -4,   9,   2,   4,   1,   2,   1,   1,   0,   1,  25,  29, -26,  -8,  -2,  -3,   0,  -1,   0,   0,   0,   0, -18,  23,  26, -12,   0,  -3,   0,  -2,   1, -13,  13,   9,  -3,   1,  -1,   1,  -4,  -1,  -2,   5,   2,   0,   0,   0,  -1,  -2,   0,   0,   1,   0,   0,   0,  -1,   0,   1,   1,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0 },
     {   19,  -4,   9,  -5,   0,  -2,   0,  -1,   0,   0,   0,   0,   2, -68,   7,  10,   3,   4,   1,   2,   1,   1,   0,   1,  10,  -4,  -2,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   6, -43, -22,  22,  -1,   5,   0,   3,   0,   1,   0,   1,  13,  43, -58, -14,   3,  -3,   2,  -1, -10,  30,  25, -24,  -2,  -4,   0,  -2,   4,  -3,   9,   6,  -5,   0,  -1,   0,  -5,   7,   2,   1,   1,  -2,   0,  -1,   1,   0,   1,   1,  -2,   3,   1,   0,   0,   0,   0,   0,  -1,   2,   1,   0 },
     {    2, -21,  -7,   7,   2,   1,   1,   1,   0,   0,   0,   0,  18, -19,  32,  -5,   0,  -1,   0,  -1,   0,   0,   0,   0, -21, -48, -14,  18,   3,   6,   1,   3,   1,   1,   0,   1,  14,  14,   6,  -5,  -5,  -1,  -2,   0,  -1,   0,   0,   0,  -3, -39, -21,  37,   3,   5,   0,   2,  -1,  32, -57, -26,  12,  -2,   4,  -1,   0,  13,  26, -26,  -8,  -1,  -2,   0,   4,   2,   0,   6,  -7,   0,  -1,   0,  -2,   5,   4,   0,   0,   1,  -2,   1,  -1,   2,   3,  -1,   0,   1,  -1,   0 },
     {  -14, -22,  22,  -3,   4,  -1,   2,   0,   1,   0,   0,   0, -43, -56,   4,  12,   4,   5,   1,   3,   1,   2,   1,   1, -53,  39, -37,  -1,   0,   0,   1,   0,   1,   0,   0,   0,  48,  15,  12, -22,  -3,  -4,  -1,  -2,  -1,  -1,   0,  -1,  -4,   5,  24,   7,  -4,   0,  -2,   0,   6, -14,   0,  10,   2,   2,   0,   0,   0,   0,  -8,   0,   2,   1,   0,   0,   3,   0,  -2,  -4,   0,   1,   0,   0,   2,   1,   0,   0,   1,   0,   0,   0,   1,   1,   0,   0,   1,   0,   0,   0 },
     {  -12,  -1, -25,   3,   2,   2,   1,   1,   0,   0,   0,   0,  -5,  33,  -5,  -2,  -6,   0,  -2,   0,   0,   0,   0,   0,  -9,  15, -31,  -4,   5,  -1,   2,   0,   1,   0,   1,   0,  13,  27,  11, -17,  -4,  -3,  -2,  -1,  -1,  -1,  -1,  -1, -20,   4, -18,  -6,  10,  -2,   4,  -1,  12,  20,  39, -53,  -9,  -2,  -2,  -1,   8, -18,  49,  36, -25,   0,  -6,   0,  -2,  -3, -18,  22,  14,  -3,   2,  -1,  -1,  -5,   5,  -5,   0,  -2,  -3,   3,   0,  -2,   3,   0,   1,  -1,  -1,   2 },
     {  -15,  -8, -17,   9,   1,   3,   1,   2,   0,   1,   0,   1, -20,  -5, -16,   0,   7,   0,   3,   0,   1,   0,   0,   0, -42, -10, -12,  13,  -1,   6,   0,   2,   1,   1,   0,   1, -64,  26, -34, -14,  12,  -3,   6,  -2,   3,   0,   1,   0,  55,   6,  10, -31,  -3,  -4,  -2,  -1,  -9,  15,   7,  -2,   3,  -3,   0,  -1,  13, -12,  15, -23,   0,   2,  -1,   2,   4,  -1,  14,   4, -17,   3,  -3,   1,   4,  -2,   2,   8,   2,   0,   1,   2,   2,  -2,   0,   1,   1,   1,   1,   1 },
     {   -7, -27, -12,  -2,   4,   1,   2,   1,   1,   1,   1,   0,   8,  -5,  57,  -7,  -5,  -3,  -2,  -2,  -1,  -1,   0,  -1, -15, -27,   0,  13,   1,   3,   0,   1,   0,   0,   0,   0, -18,  -4,  14,  -2,  -5,   0,  -1,   0,   0,   0,   0,   0,  33, -15,   4,   1,  -7,   1,  -2,   1,   8,   7,  25,  23, -20,   0,  -5,   0,  -4,  -8, -29,  57,  12,  -4,   2,  -1, -10,  11, -34, -18,  29,   2,   5,   0,   2,  -2,   3, -14,   0,   4,  -6,  -2,   0,   0,   1,  -2,  -1,   2,  -3,  -1 },
     {  -13, -26, -85,  23,  -4,   8,   1,   5,   1,   3,   1,   2,  -2,   1,  46, -21,   2,  -5,  -1,  -2,   0,  -1,   0,  -1,  12,  46,   3,  15,  -4,  -1,  -2,  -2,  -2,  -1,  -1,  -1,  10, -22, -18,   2,   7,   2,   1,   1,   1,   1,   1,   0, -16,  18,   3,  -4,   0,  -1,   1,  -1,  -4, -13,  -9,   2,   7,   0,   2,   0,   1,   5,  -4, -19,   2,   1,   0,   0,   3,   0,  12,  -2,  -9,  -1,   0,   0,   1,   0,   1,   3,   0,   0,   2,  -1,   0,   0,   1,   0,   0,   0,   1,   0 },
     {    5,  -5, -38,   1,   2,   2,   0,   3,   0,   1,   0,   1,  12,   3, -36,  -3,   1,   0,   1,   0,   1,   1,   0,   0,  24, -13,  -3, -23,   6,   1,   1,  -1,   1,   0,   0,   0,  47,  -5,  37, -31,  -8,  -3,   0,  -3,  -1,   0,  -1,  -1,  59,  19,  12,  18, -19,   0,  -8,   1, -33,  27,   4,  18,   8,  -2,   1,  -1,  -6, -16,   3,  -5,   5,   3,   1,   2,  -8,   2,   0,   1,  -7,   3,  -1,   1,  -4,   0,   2,   4,  -3,   1,  -1,   2,  -2,  -1,   1,   1,  -1,   0,   0,   1 },
     {    8,   5,  29,   0,  -4,  -4,  -1,  -2,  -1,  -1,   0,  -1,   3,  -3,  20, -19,  -1,   4,  -3,   0,   0,   0,   0,   0,  13,   7,   6,  20,  -4,  -4,   0,  -1,  -1,  -1,  -1,   0,  -7, -31, -16, -29,   5,   3,   2,   3,   1,   1,   1,   1,  24,  16,  37,  14, -20,  -3,  -5,  -2, -16, -24, -15,  -6,   9,   9,   0,   3,  24,  17,  25,   8, -41,  -1,  -5,  -2, -14,   0, -14,  53,  22, -11,   3,  -3,   0,   1, -14, -19,  -4,   0,   2,   6,   1,   2,  -4,  -4,  -2,   0,   0,   3 },
     {   -1, -12, -23,  15,   1,   7,   0,   3,  -2,   1,   0,   1, -11, -46, -35, -24,   8,   0,   5,   3,   4,   1,   2,   1,  -2,   4,  80, -16,  -7,  -5,  -5,  -3,  -2,  -1,  -1,  -1,   9,  24,  -4,  16,  -4,   3,  -1,  -1,   0,   0,  -1,   0,   1, -32,  10,  -8,   6,   1,   0,   1,  22,   3,  -4,  -2, -14,   5,  -2,   2,   0,   4,  16,  18, -13,  -4,  -3,   0,  -3,  -1, -14,  18,   8,   1,  -1,  -1,  -3,   3,  -4,  -5,   0,   0,  -3,   3,  -2,   2,  -2,  -2,   0,   1,  -1,   2 },
     {    0, -10,  15, -31,   7,   1,   2,   0,   0,   0,   2,   0,   9,  -3,  46, -22,   2,  -6,   0,  -2,   0,  -1,  -1,  -1,   1, -12,  15, -31,   0,   7,   1,   2,   1,   1,   0,   1,   5,  -2,  26,  -6,  -7,   2,  -4,   1,  -2,   1,   0,   0,   0, -11,   6, -23,  -3,   4,  -1,   3,   5, -16,  28,  -3, -12,   3,  -5,   2,  11, -24,   5, -15, -17,  16,  -5,   5,  10,   7,  15,  26, -54,   0,  -4,  -1,  -6,   8, -13,  47,  -1,   4,  -7,   1,  -1,   1,  -5,   9,  -1,   1,  -2,   1 },
     {   -1, -11, -37, -12,   1,   2,   0,   1,   2,   0,   1,   1,   5, -14, -29,  36,  -1,   5,  -1,   1,  -1,   0,   0,   0,   4, -32, -31, -16,  16,   1,   8,   0,   3,   0,   2,   0, -12, -42,  10, -21,   0,   2,   1,   1,   1,   1,   1,   1,  -2, -18,  12, -19,   2,  -1,   1,   0,  45, -26,  -1, -10, -21,   7,  -5,   3,   0,  47,  -7,   7, -15,  -5,  -2,  -3,   0,   2,  18,   8,   2,  -4,  -1,  -2,  -3,   7,  -5,   6,  -2,   1,   3,  -1,  -1,   3,  -2,   2,   0,   1,   2,  -1 },
     {    8,  10,  19,  -7,  -4,  -1,  -1,  -1,   1,   0,  -1,   0,   8,  14,  14,  -9,   2,  -5,   0,  -1,   0,   0,   0,   0,  13,  27, -15,  10, -10,   1,  -4,  -1,  -1,   0,  -1,   0,  31, -14, -31,  -4,  13,  -4,   3,   0,   1,  -1,   1,  -1,  22, -29,  14, -37, -10,   5,  -2,   1,  34,  53,  -5,  20, -24,  -7,  -5,  -2, -35,  19,  25,   3,  18,  -8,   1,  -3,   7, -20,   8,   8,  -2,   5,  -1,   2, -10,   1,  -5,   6,   0,  -5,   1,  -2,  -4,   0,  -1,   1,   0,  -2,   0,   0 },
     {   -8,  -1, -20, -90,   7, -11,   8,   2,   4,   0,   4,   1,   2,   8,  18,  50, -11,   7,  -5,  -2,  -3,   0,  -1,   0,   0,  21,  29,  -7,   7,   0,   1,  -1,   0,  -1,  -1,  -1,   5,  28, -27,   5, -10,   2,  -2,  -1,   0,   0,   0,   0,  18,   4,  -2,  16,   1,  -6,  -1,  -2,  -5,   1, -13,  -7,   4,  -1,   1,   1,   4,  12,   5,   2,   0,  -3,   0,  -1, -10,  -5,   6,  -1,   8,   2,   0,   1,   1,  -7,  -3,  -3,   1,  -1,   2,   0,   0,  -3,  -1,   1,   0,  -1,   0,  -1 },
     {   -2,  -1,  -8, -45,  25,  -9,   5,  -1,   1,  -1,   1,   0,  -6, -13, -25, -26,  -7,   5,   0,   4,   0,   2,   0,   1,   6,  -5,  10,  43,   7,   1,  -3,  -2,   0,  -2,   0,   1, -15, -28,  -7, -38,  11,   6,   5,   3,   2,   1,   1,   1, -33,  -1,  11,  15, -20,   2,  -4,   0,  28,  16,   8,   2,  19,  -6,   1,  -1, -14, -36,   0,  -6,   2,  11,  -1,   5,  16,  13, -22, -17, -11,   2,   0,   1,  -4,   8,  20,  -3,   5,  -1,  -3,  11,  -2,   2,   3,  -1,   2,  -1,   0,   3 },
     {    7,   9,   3, -11,  26, -11,   4,  -3,  -1,  -2,   0,  -2,   4,   8, -15, -52, -10,   1,   2,   3,   2,   3,   1,   2,   8,  11, -16,  36,  -7,   1,  -6,   0,  -3,  -1,   0,   1,  25,  19,  -6,  11,  -2,   4,  -2,  -2,   1,   0,   0,   0,  48, -15, -27,   0,  17,  -5,   3,  -1,  17, -35,   7, -26,   0,  10,  -1,   4,   0,  22, -31,  -2, -20,  -3,   0,   1, -12,  17,  14, -16,  -3,  -5,  -1,  -2,   3,  -4,  15,   9,  -6,   5,  -1,   4,  -1,  -3,   4,   2,  -2,   0,  -1,   2 },
     {   -2,   2,   3,  -1,  10,  -2,   4,  -1,  -1,  -1,   0,  -1,  -6,   0,  11, -36,  -2,  -2,   2,  -1,   2,   0,   0,   0,  -9,  -1,   1, -27,  -2,   0,   0,   0,   0,   0,   1,   0, -30,  14,  23,   3, -29,   7,  -5,   0,  -2,   2,  -1,   0,   8,  61,  -9,  33, -12, -11,  -2,  -5,  53,   4,  -8,   8,   6,  -3,   0,  -2, -32,  26,  14,  -1,   7,  -6,   2,  -2,   9, -26,  14,   3,   7,   0,  -1,   2,  -8,  -4,  -5,   2,   3,  -7,   2,   0,  -2,   0,  -2,   2,   1,  -3,  -1,  -1 },
     {    5,   3,  13,   6,   5,   9,  -1,  -2,   0,   0,  -1,  -1,  -1,  -7,   5, -11,  -1, -19,   1,   0,   1,   0,   2,  -2,   8,   7,  11,  -8,  37,   6,  -1,  -2,  -4,  -1,  -2,   0,  -5, -17, -15, -29, -23,  -5,   2,   3,   3,   1,   2,   1,   8,   4,  21,   7,  37,   2,  -5,  -2, -19, -13, -13, -34, -23,  13,   6,   4,   9,   6,  22,  19,  28,  -9,  -5,  -5,  -1, -31,  -9, -44, -15,  29,   2,  10,   9,  13,  19,  20,  -5,  -8,  -6,   5,   2,   6,   2,   1,  -3,  -4,  -4,   2 },
     {   -8,  -3, -14, -37,  31, -14,   1,  -6,   5,  -1,   0,  -2, -10,  -3, -17, -43,  13,   0,   6,   2,   1,   0,   2,   1, -12, -10, -23,   0, -12,  11,   2,   6,  -3,   1,   1,   2,  -9,  -9,  10,  30, -10,   3,  -9,   2,  -3,  -1,  -1,  -1, -26,  -7,   8,  -8, -10,  14,   6,   0, -61,  10,   5,  12, -29,  -1,   0,  -2,   2,  30,   8,   4,   5, -10,   3,  -8,   9, -21,   1,   7,   9,  -4,   6,   4,  17,   5, -11,  -5,   2,  -2,  11,  -2,   5,   1,  -2,   0,  -1,  -1,   6,  -1 },
     {    2,   5,   2,  29,  50, -13,  -3,  -7,  -1,  -2,  -1,  -1,  -3,  -5,   9,   7, -75,  31,  -6,   4,  -1,   0,  -1,   1,  -1,   2,  14,   2,  40, -36,   5,  -3,  -1,   0,   2,  -1,  -2,   9,   2, -14,   2,  18,  -2,   1,  -1,   0,  -1,   0,  -8,  -5,  -5, -14, -10,  -8,   8,   2, -16,  14,   4,  10,  -2,   9,  -9,  -2,  -2,   9,  -4,  -4,  -2,   0,   5,  -6,   6,  -9,  13,   7,   4,  -5,   0,   2,   2,  -4,  -8,  10,  -3,  -1,   0, -11,   3,  -1,   2,   2,  -1,   0,  -2,  -2 },
     {    0,   0,  -4,   2,  -2,   0,  -9,  -4,   0,   3,  -2,   2,   2,   1,   1,  -1,  -7,  12,   0,   0,   1,  -2,   1,  -1,  -1,   4, -21, -40,  17,  -5,   9,   6,  -1,   0,   0,  -1,  -2,  -9, -32,  34, -18,  11,  -9,   6,  -3,   0,   2,   0,  20, -11,  20,   5,   3,  -6,   5,  -6,  -9, -11,  -9, -14,  10,  11,  -6,   0, -49, -19,  -3,  -1, -17,  22,   8,  -2,  41,  13, -35,   3,  -5, -28,   8,  -2,   2,  25,  22, -10,   0,  -8,  10,   4,   0,   4,   1,  -1,   1,  -1,   2,   1 },
     {    1,   1,   5, -10,  20,  10, -12, -15,  11,  -1,   7,   5,   3,   3,  14, -19,  39,   1,   3,  -9,   5, -14,  -2,  -7,  -4,  -2,   3, -43,  13, -26,  -6,  -9,  16,  -2,  -1,   2,   1,   9,   1, -25,  49,  22,   9,   5,  -6,  -5,  -3,  -3,  -5,   5,  -7, -18,   7,   6,  -6,  15,  -5,  12,  -8, -16,   6,  10, -13,   6,   1,  13, -24, -10,   2,  12,  -2,   3,  -4,   7,   5,  -4,  18,  18,   5,  -8, -15,  -7,  17, -30,   0, -18,  10,  10,   9, -10,   1,   3,  -3,   3,   0,  15 },
     {   -5,   1,   6,  -7, -45,  30, -12,  -2,   1,   1,   3,   2,  -8,   0,  -5, -12, -16,  -7, -11,  -1,   3,  -1,   1,   0,  -6,   6,   9,  -4,  -5,  11,   4,  -4,   1,  -1,  -2,   2,  -9,   4, -11, -32, -31,  13,   6,   4,  -1,  -1,   1,  -1, -17,  -3, -11,   1, -15,   0,  -3,   2, -26,  21,   0,   8, -16,  14,   6,  -2, -13,  23, -19,   9, -22, -16,  11,   1,   9,  35,  22,  -1,   8, -20,   1,  -7,  -9, -21,  34,  26,   4, -12, -17,  10,   4,  -6,  11,  -2,  -2,  -5,  -4,   4 },
     {    3,   4,   0,   4,  32,  -1,   2,  -2,  -1,   0,  -1,   1,   1,   3,  -3, -14,  -9,  -2,  -3,  -5,   0,   2,   0,   0,  -1,   3, -21, -41,  -3,   6,   0,   2,   0,   0,   1,   1,   0, -16, -28,  24, -29,  -1,  -2,   2,   0,   0,   0,   0, -10,  -5,  32,   9,  10,  -5,   1,  -6,  26,  32, -10,  14,   8,  -4,   8,  -1,  55, -12,  -1,  -3,   2,   1,  -3,   4, -34,  33,   3,  -4,   4,   2,   2,  -3,   5, -30,  12,   3,  -5,  10,  -7,   1,   3,  -4,   8,  -2,  -4,   4,  -2,   2 },
     {    0,   5,  -2,  16,  60, -15, -15,  -6,  -5,  -8,   2,   0,  -1,   5,   2,  30,  21, -11,  -8,  -6,   0,  -2,   1,   0,   0,   5,   3,   6, -18,  31,  -8,  -2,   1,   1,  -2,   3,  -3,  -1,   6,  -1, -17,   3,  19,  -1,   4,   3,   2,   0,  -4,  15,  -5, -23, -11,  -5,  -8,  12, -10, -13, -40,  -4, -19,   3,   5,   9, -34, -21,  -1,  12, -20, -24,  -6,  10,  -4,  21,  -7,   8,   1,  19,  -9,  -7, -16,  -6,   0,  12,   9,  -1, -13,   0,  -3,  -4,   0,  -7,   5,  -2,  -2,   2 },
     {   -2,  -3,  -1,   0,   5,  15, -11,  -6,  12,  -8,   1,   1,  -3,  -4,  -2,   2,   8,  11,  -4, -10,  -1, -10,  -3,  -4,  -2,  -4,   1,  18,  -2,   7, -10,  -3,   7,   1,   1,   1,  -4, -13,  -1,   5, -27, -12,  14,  -2,   1,   1,  -3,  -4,  -4, -17,  19,  13,  43, -22, -13,  -5, -14,  16,  35,  -6,  40,  10, -12,  15, -38,  13, -14,  -3,  -3,  24, -21,  -3,  -9, -11,  11,  20,   3,  28,   8, -10,  -2, -21, -24,   3,  10,  -3, -14, -15,   1,  -9,   2,   0,  -4,   1,   2,   6 },
     {    0,   0,  -1,  -6,   3,  12,  22,  -4,   2,   3,   0,  -1,  -3,  -5,   0,   4, -36,   5,  -5,   6,   3,   3,  -1,   3,  -3, -10,   1,   1, -45, -18,   2,  -2,   4,   0,   1,   2,  -2,  -9,  17,  25,  50, -20, -14,  -7,  -1,  -2,  -1,  -5,   7,  20,  33,   9,  20,  10,   4,  -7, -10,  -2,  -9, -16,  -7,   3,  -2,   4,  -7,   9,  20,  14,   4,   2,  -9,   1,  22,  24,   7, -12,   0,  11,   7,   3, -29, -17,  15,  16,  19, -10, -17,   0,  -5,  -1,   2,  -3,  -3,  -1,  -7,   3 },
   },
   {
     {  -97,  64,  -5,   2,   0,   1,   0,   0,   0,   0,   0,   0, -26,  15,   6,  -5,   2,  -1,   1,  -1,   0,   0,   0,   0,  18, -21,  10,  -3,   0,   0,   0,   0,   0,   0,   0,   0,  20, -17,   2,   2,  -2,   1,   0,   0,   0,   0,   0,   0,  12,  -7,  -3,   2,  -1,   0,   0,   0,   5,  -1,  -3,   1,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   50,  19, -28,  -1,  -4,   0,  -2,   0,  -1,   0,  -1,   0, -73,  67, -12,  -3,   2,  -1,   1,  -1,   0,   0,   0,   0, -24,   4,  17, -11,   5,  -2,   2,  -1,   1,   0,   0,   0,  13, -23,  18,  -5,   0,   0,   0,   0,   0,   0,   0,   0,  12, -16,   6,   3,  -3,   1,  -1,   1,   7,  -6,  -1,   3,  -2,   0,   0,   0,   2,  -2,  -1,   2,   0,   0,   0,   0,   2,  -2,   0,   1,   0,   0,   0,   0,   1,  -1,   0,   1,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0 },
     {   -2, -44,  58, -22,   8,  -3,   2,  -2,   1,  -1,   1,   0, -63, -17,  52,  -8,   2,   1,   1,   0,   0,   0,   0,   0,  -3, -24,   5,  15,  -7,   4,  -1,   1,  -1,   1,   0,   1,  20, -11, -21,  17,  -5,   1,   0,   1,   0,   0,   0,   0,  11,   6, -23,   8,   1,  -1,   0,   0,   4,  10, -14,  -1,   3,  -1,   0,   0,   2,   7,  -5,  -4,   2,   0,   0,   0,   2,   4,  -2,  -2,   0,   0,   0,   0,   1,   3,  -1,  -1,   1,   2,  -1,   0,   1,   2,  -1,   0,   1,   1,  -1,   0 },
     {   44,  25, -39,  11,  -8,   2,  -3,   1,  -2,   0,  -1,   0,  -1, -20,  10,   0,   4,  -1,   1,   0,   1,   0,   0,   0,  63, -69,  20,  -2,   0,   1,  -1,   1,   0,   0,   0,   0,  16,  -3, -13,  12,  -7,   3,  -2,   1,  -1,   1,  -1,   1, -18,  23, -14,   5,   0,  -1,   1,  -1, -16,  15,  -2,  -4,   4,  -2,   2,  -1,  -9,   5,   4,  -5,   2,  -1,   0,   0,  -4,   2,   3,  -2,   0,   0,   0,   0,  -3,   1,   2,  -1,  -2,   1,   1,   0,  -2,   1,   1,   0,  -1,   1,   0,   0 },
     {   16,  25,  17, -39,   7,  -6,   2,  -3,   0,  -1,   0,  -1,  51,   7,  35, -50,   6,  -5,   1,  -3,   0,  -1,   0,  -1, -16,   2,  49, -22,  -6,   4,  -3,   1,  -2,   1,  -1,   0,  -4, -23,  27,  13, -17,   7,  -4,   2,  -2,   1,  -1,   1,   0, -14,  -3,  28, -13,   3,  -1,   1,  -3,  -3, -14,  20,  -2,  -1,   1,   0,  -3,   2, -12,   9,   3,  -2,   1,   0,  -3,   2,  -8,   4,   4,  -1,   0,   0,  -2,   1,  -5,   1,  -1,   0,  -3,   0,  -1,   0,  -2,   0,  -1,   0,  -2,   1 },
     {   18,  50,   2, -33,   6,  -5,   1,  -3,   0,  -1,   0,  -1, -29,  -8,  35, -26,   9,   0,   2,  -1,   1,   0,   0,   0,  17,  19, -25,   4, -10,   5,  -3,   1,  -1,   1,  -1,   0, -40,  62, -20,   1,  -3,  -1,   2,  -2,   1,  -1,   0,  -1, -14,   2,  18,  -9,   4,  -4,   2,  -2,  10, -21,  18,  -1,  -2,   1,  -2,   1,   9, -17,   5,   8,  -4,   2,  -2,   1,   3,  -5,  -3,   7,  -1,   1,   0,   0,   2,  -4,  -1,   3,   1,  -2,  -1,   1,   1,  -2,  -1,   1,   0,  -1,  -1,   1 },
     {    5, -13,  23, -46,  28, -10,   7,  -4,   3,  -1,   1,  -1,  13,  37, -32, -20,   9,   0,   0,   0,  -1,   0,   0,   0,  62,  12, -41,   4,  -8,   7,  -4,   1,  -1,   0,   0,   0,  46, -12,  -2,  -8,  -2,   2,  -1,   0,   0,   0,   0,   0,  12,  -4,  14,  -6,  -4,   2,  -1,   0, -14,   5,  12,   2,  -6,   2,  -1,   0, -15,   3,   7,   6,  -4,   0,   0,   0,  -9,  -1,   5,   4,  -2,   0,   1,   0,  -5,  -2,   2,   3,  -4,  -1,   1,   2,  -2,  -1,   0,   2,  -2,   0,   1,   1 },
     {   11,  36,  -1, -15,  -2,  -4,   1,  -2,   0,  -1,   0,   0,  26,  28,  -7,   7, -16,   3,  -3,   0,  -1,   0,  -1,   0, -38,   6, -17,  48, -23,   8,  -3,   2,  -1,   1,  -1,   1,  11, -13, -42,  45,  -5,  -3,   3,  -1,   1,  -1,   1,   0,  -1,  13, -34,   7,  18, -12,   6,  -3,  -5,   7,  -2, -22,  22,  -8,   2,  -1,  -2,  -3,  13, -21,  10,   1,  -2,   0,   1,  -6,  11,  -9,  -1,   4,  -1,   0,   0,  -3,   5,  -3,   0,  -2,   4,  -2,   0,  -2,   2,  -1,   0,  -2,   1,  -1 },
     {   26,  39,  -3,  -4,  -8,  -2,  -1,  -1,  -1,  -1,  -1,  -1,  -5, -56,  16,  28, -12,   7,  -1,   3,  -1,   1,   0,   1,   4,  15, -35,  28, -12,   2,  -1,   0,  -1,   0,   0,   0,  12, -30,  28,  -7,  11,  -7,   3,  -1,   1,   0,   0,   0,  34, -50,  14,   1,   2,   2,  -3,   2,  -1,   7, -16,   9,  -3,   1,  -1,   0, -10,  15, -14,   1,   4,  -2,   1,  -1,  -8,  13,  -3,  -7,   5,  -1,   1,  -1,  -2,   3,   1,  -5,  -2,   3,   1,  -3,  -1,   1,   1,  -2,  -1,   2,   0,  -1 },
     {   -1,  -8,  31,  17, -46,  23,  -9,   4,  -4,   2,  -2,   1,   3,  40,  11,  20, -56,  19,  -8,   2,  -3,   1,  -2,   0,  45,  13,  12,   9, -27,  -2,   2,  -3,   0,  -1,   0,  -1, -11,  19,  17,  10,  -7, -11,   5,  -3,   0,  -1,   0,  -1, -11,  -5,   8,  18,   4, -12,   3,  -1,  -3,  -9,  -9,  17,  11,  -8,   1,   0,  -2,  -4, -13,   7,  15,  -6,   0,   0,  -3,  -1,  -8,  -3,  13,  -2,   0,   0,  -1,  -2,  -3,  -6,  -1,  -1,  -1,  -4,  -1,  -2,   0,  -3,   0,  -1,  -1,  -2 },
     {  -11, -13, -17,   1,  31, -15,   7,  -3,   3,  -1,   1,  -1, -24,  -7, -28,   6,  24,  -8,   3,  -1,   1,   0,   1,   0,  32,  28,  -6,   0,  -8,   6,  -9,   2,  -3,   0,  -1,   0, -55,  -5,  24,  31, -21,  13,  -5,   1,   0,   1,   0,   1,  -9, -14, -27,  40, -10,   2,   3,  -2, -10,  13, -25,   2,  15, -11,   7,  -3,   4,   1,  -1, -20,  16,  -6,   2,   0,   5,  -2,   8, -14,   2,   3,  -2,   1,   6,  -3,   5,  -4,   2,   0,   3,  -2,   2,  -1,   2,  -1,   1,   0,   1,  -1 },
     {   -8, -20, -12,   2,  34, -23,   9,  -4,   4,  -2,   2,  -1,  10,  21,  -5,  15,   6, -11,  -3,  -2,   0,  -1,   0,  -1, -12, -57,   6,  50, -22,  16, -10,   5,  -2,   2,  -1,   1, -16,   8,   0,  11,  -9,   5,  -3,  -1,   0,   0,   0,   0, -10, -26,  47,  -6,   2,   1,   0,  -1,  16, -31,   1,  16,  -5,   5,  -2,   1,   0,   6, -21,  10,   3,  -2,   3,  -2,   0,   7, -14,  -5,   7,  -2,   3,  -1,  -2,   7,  -3,  -8,   1,   2,  -1,  -5,   0,   3,  -1,  -3,   1,   1,   0,  -2 },
     {    5,  24,  45, -25,  11, -18,   5,  -5,   2,  -3,   0,  -1,  -8,  -1, -13,  33,  -7,  -2,  -1,   2,  -1,   1,  -1,   0,  -4, -14,  12,  13,   3,  -4,  -3,   1,   0,   0,   0,   0, -26, -29,  44, -19,  24,  -7,  -1,   0,   1,   0,   0,   0, -52,  28,   5, -20,  13,   2,  -3,   1, -27,  17,   8, -12,   0,   6,  -2,   0,   7, -11,   4,  -1,  -7,   9,  -2,   0,  12, -10,   0,   3,  -7,   4,   0,   0,   7,  -4,  -6,   4,   3,   0,  -5,   2,   3,   1,  -4,   1,   2,   0,  -2,   0 },
     {   10,   1,  25,   7,  -5,  -5,  -3,  -1,  -2,   1,  -1,   0,  16,  19,   2,   5,  14, -16,  -1,  -3,  -1,  -1,   1,  -1,  35,  -5,  13,  -2,  33, -23,   2,  -4,  -1,  -2,   0,   0, -38, -21, -19,  17,  38,  -8,  -1,   2,  -1,   1,   1,   1,  27, -34, -24, -12,  29,   5,  -9,   5,  30, -28,   0, -21,   4,  22, -11,   4,   8,  -2,  -2,  -9, -13,  16,  -1,   0,  -8,   8,   1,  -1, -12,   6,   3,  -1,  -6,   5,   0,   1,  -4,   4,   1,   0,  -3,   2,   0,   1,  -3,   2,   0,   0 },
     {    6,  15,  25, -36,  -9,  31, -15,   4,  -3,   1,  -1,   0,  -6, -15, -59,   3,  -6,  42, -10,   9,  -1,   4,   0,   2, -13, -36, -13, -16,  14,  11,   6,   1,   2,   1,   1,   0, -23,  -3, -18,  -2,  -5,   9,   5,  -1,   2,   0,   1,   1,   3,   4, -19,  11, -22,   5,   6,  -2,  24, -19,  -6,  21, -16,  -5,   7,  -1,  11,   3, -13,   9,  -3, -14,   6,  -1,  -5,  10,   3,  -4,   6, -11,   1,   0,  -2,   5,   4,  -3,   0,   2,   5,  -2,   0,   0,   2,  -1,   0,   0,   2,   0 },
     {   -6,  -6,  -6, -13,  -4,  16,  -3,   4,  -1,   2,  -1,   1,  11,  30,  20, -20,  -4,   4,  -4,  -2,  -1,  -2,   0,  -1, -13, -36, -34,   2,  25,  -4,   8,  -1,   4,  -1,   2,   0, -12,  29,   6, -35,  25, -11,   3,  -2,   1,  -1,   0,  -1, -15, -33,   2,  15,   4,  11,  -3,   5, -15,  30, -52,  -2,   6,   0,   3,  -2,  -9,  13,  13, -29,   4,  -1,   1,   0,   9,  -5,  14,  -9,  -8,   5,  -2,   2,   2,  -5,  10,   2,   2,  -2,   1,   5,   1,  -1,   2,   2,   2,  -1,   0,   1 },
     {   12,  28,  37,  20,   1, -27,   1,  -7,   0,  -3,   0,  -2,   5,   8,  -6,  35,  25, -37,   9,  -6,   2,  -2,   0,  -2,  -6,   6,  -6,  -1,  35, -27,   4,  -2,  -1,   0,  -1,   0,  32,  35, -11,   0,   1,   2, -12,   4,  -4,   1,  -2,   1,  -6,  19,  -2,  28, -18,  12,  -9,   1,   3,  -7, -18,  33,  -3,   3,   1,  -1,  -6,   2, -16,   3,  17,  -4,   4,  -2,  -3,  -4,  -1, -17,  16,   1,   0,   1,  -2,  -3,   4, -13,   0,  -4,   4,  -6,  -2,  -1,   2,  -4,   0,  -2,   1,  -2 },
     {   -2,   6,  -4,   3,   1, -15,  10,  -3,   1,  -2,   1,   0,  -1,  17,   1,  -5,  -8,  -7,   3,   1,  -1,  -1,   0,   0,   2,   9,  14,   3, -18,   6,  -3,   4,  -2,   0,   0,   0, -42,  -1, -27, -12,  15,   2,   5,   3,   1,   0,   1,   0,  32,  25, -13, -40,  -2,   7,  -7,   2, -14,  44, -12,  21, -27,   9,  -2,  -3, -25,  39, -27,  23,  -1, -13,  10,  -6,  -8,   0,   9,  -6,  19, -12,   4,  -1,   5, -13,  10,  -7,   2,  -6,   5,   1,   2,  -5,   2,  -1,   1,  -2,   2,   0 },
     {   -7,  -5,  -2,   1,  -6,  35, -24,   6,  -4,   2,  -1,   1, -13, -11,  -7, -29,  24,   6,  -3,  -3,   3,  -1,   2,  -1,  -4,  11,   1,  17,  25, -26,   9, -10,   2,  -1,   1,  -2,  -8, -27, -21,  36,  14, -23,   8,  -1,  -1,   1,   0,   0, -25,  -4,  38,   3,   5,  -9,  -5,   2, -44,  -9,  22,  25,   3,   6,  -6,   2, -28,  13, -13,  -1,  13,   1,   2,  -3,   2,  -5,   3, -13,   7,   3,   1,  -1,  11,  -7,  -1,  -7,   6,  -1,  -1,  -2,   3,   1,  -2,  -2,   2,   0,   0,  -1 },
     {   -2,  -3, -13, -24,  15,  -5,  18,  -6,   2,   1,   2,  -1,  -8, -21, -29, -14,  -3,   9,  13,   0,   1,   3,   2,   0,   1,   5,  29, -10, -16,  -3,   1,  -1,  -3,   0,  -1,  -1,  19,  21,   6, -14,  26, -37,   8,  -1,   0,  -1,   0,  -1,   7,  20,   3,  -2,  43, -35,  -7,   3,  21, -13, -30,  17,  33,  -2, -15,   5,   3,   8, -25, -11,  18,  12, -12,   1,  -4,  -2,   9, -13,   0,  14,  -5,  -2,  -5,   2,   6,  -5,  -2,  -2,   6,   1,  -2,   0,   2,   2,   0,   0,   1,   2 },
     {    8,  15,  45,  20, -30,  -9,   2,   1,  -3,  -1,  -1,   0,  -6, -10, -33, -18,   8, -13,  17,   0,   7,  -2,   2,   0,  -6, -13,  13, -15,  -9,  15,   1,  -1,   3,   0,   0,   0,  -6,  15,  -1,   3, -36,  25,  -4,   1,   1,  -1,   0,   0,  20, -17,  37, -25,  -1,  -7,   5,  -1, -12,   4,  -2, -26,  25,  -3,   0,   3, -19,  26,   1, -36,   9,   3,  -4,   1,   4,  -5,  11,   1, -15,   6,  -4,   1,   2,  -3,  -2,   9,   1,   0,  -2,   9,   0,   1,  -1,   4,   2,  -2,   0,   3 },
     {    5,   9,   8,  14,  30,  18, -46,   4,  -4,   2,  -2,   1,   5,   7,  11,  17,  36,  32, -53,   6,  -5,   3,  -2,   1,   6,  20,  26,   3,   5,  32, -31,   0,  -3,   0,  -2,   1,  17,  18,   4,  -3,   5,  14,   3, -17,   4,  -4,   1,  -1,   4,   4,   2, -17,  14,   2,  18, -14,  -4,   1, -21,  -3,   0,  11,  13,  -2,  -5,   1, -14, -13,  -6,   8,  11,   0,  -1,  -5,   3,  -8,  -9,   2,   7,   1,  -1,  -3,  -1,   0,  -2,  -1,   1,   1,  -1,  -1,  -1,   1,  -1,  -1,   1,   0 },
     {    6,   3,   4,  34, -22,  -3,   2,  -2,  -2,  -1,   1,  -1,   2,   3,   7, -40,  13,  -5,  11,  -1,   2,   1,   0,  -1,  12,  14, -27,  21, -10,  15,  -7,   7,  -3,   1,  -1,   0,  -2, -46, -11, -24,   6,  14,   2,   1,   4,   0,   2,   0, -27,  29,   7, -18, -13,   5,   7,  -4,  19, -20, -36,  26,   5,   0,   4,   3,  39, -10, -19, -19,   1,  -3,   0,   2,   6,   2,   6,  -1,  -9,  -3,  -3,   0,   0,   8,   0,   2,  -4,   3,   4,   5,  -3,   1,   1,   1,  -2,   0,   3,   0 },
     {   -6,  -3,  -9, -12, -20,   0,  12,  -3,   1,   2,   1,   0,  -7,  -4,  -2,   8,  -6,   0,  13,   3,  -3,   3,  -3,   1,  -2,  19,  -4, -23,  -6,  10,  -9,  13,  -5,   5,  -2,   2,  21,   1, -11,  19,  -7,  21, -23,  11,  -3,   0,  -2,   2, -69, -26, -10,  -8,  30,  14,  -6,   2,  17,  -9, -19, -12, -14,  18,  -2,  -3, -25,  30, -18,  14, -24,   6,  10,  -7, -10,  11,   2,  -5,  -9, -12,  10,  -4,   3,  -3,  10,   2,   6,  -6,   1,   1,   3,  -4,   1,   1,   3,  -4,   1,   0 },
     {   -4,  -7, -16, -29,  -1,  -3,  16,  -2,   4,   0,   2,   1,  -3,  -9,   4,  21, -34,   3,   1,   5,   0,  -1,   0,   0,   6,  12,  31,  -3,  -1,   1,  -9,   3,  -2,  -2,   0,   0,   1,  -4, -38, -18,  21,  28, -21,   7,   1,   2,   0,   3,  10, -10,   8,  -4, -18,  37, -16,   0, -29, -29,   3,  21,   8,   4,  15,  -9,  -8, -26,   0, -33,  19,   0,  11,   3,  13, -16,   9, -10, -22,   8,  -3,   5,   4,   2,  -7,   8,   0,   4,   0,   2,   2,   3,  -2,   2,   0,   1,   1,   0 },
     {    3,   3,   5,  11,  -4,  -3,  -7,  -1,  -1,  -2,   2,  -2,  -6, -14, -30, -14,   1,   8,   4,   3,  -2,   1,   2,   1,   4,   9,  31,  36, -28, -16,   7,  -1,  -2,  -1,   1,  -2,  12,  20,  -4,  16,   5, -19,   5,   0,   0,  -1,   0,  -1,   1, -10,   4,  15, -18,  16, -16,  11,   6,  14, -29,  12, -33,  31, -13,   2,   9,  -7,  35,  -7, -38,  11,   5,  -7,  17, -21,  11,  33,  -9, -11,   9,  -4,   2,   0,  -4,   6,  -1,   1,  -6,   4,  -3,   3,  -2,  -3,   1,  -1,  -1,   1 },
     {   -3,  -3, -15, -36,   5,  34, -13,   5,  -4,   4,  -2,   2,   3,   5,   6,  15,   5,  -6,   0,  -3,  -3,  -4,   0,  -2,   8,  14,  14,  29,   1, -54,  21,  -4,  -3,   0,   2,  -2,   0,   3,   9,  -2, -27,  10,  -6,   9,  -6,   5,  -1,   2,  -1,  22,  17, -21, -25,  24,  -9,   1,  31,  15, -10, -31,  18,  -6,   5,  -1,  14,  23, -24, -15,   1,  -1,  -3,   6,  -1,  16,  -7,   4,  -5,  -1,   1,  -2,  -8,   4,   6,   4,  -4,  -3,   4,   1,   1,  -8,   0,   2,   1,  -2,   1,   3 },
     {    0,   8,   5,  13,  26, -21,   7,  -7,  11,  -3,   1,  -2,  -4,   1,  11,   9, -15,   7,  -9,  -1,   3,  -1,   1,  -1,   3,  14,  14, -11,  -6,   4,  13, -14,   4,  -3,   2,  -2,  -4, -11, -20, -17,   8, -25,  43, -10,   6,   3,   0,   1, -26,   3,   7,  19, -38,  -5,  18,   9,   7, -26,   3, -18,  -9,  -9,   6,  10, -28,  34,  12, -29, -12,  -9, -10,  14, -22,  29,   9,  10,  -9,  -5,  -9,   6,  -9,   4,   2,  -2,   2,  -6,   9,  -1,   4,  -3,  -2,   1,   3,  -2,   0,  -1 },
     {    3,  -4,  -3, -13,   3,  -3,   2,  13,  -3,   3,  -3,   4,   9,  -2,   9,  21,   6, -14,   4,  -2,  -4,  -2,  -5,   2,   9,   2,   7, -15,  -1,   1,  33, -19,   5,  -2,   3,  -2,  11,  -6,  -6,  13, -16,  -4,  37, -10,  -3,   3,   0,  -1, -13, -26,  -3, -34,   2, -12,  26,   3,  21,   3,  -6,   9, -31, -12,   1,   6,  12,  -9,  -4,   0,  14, -17,  -3,   4,  42, -39,  29, -25,   7,  -2, -13,   7,  13, -15,   5,   9,  -3,   9,  -5,   2,  -3,   6,  -2,   3,   0,   2,   0,  -2 },
     {    3,   2,   9,  24,  24, -22, -28,  10,  -5,   1,  -2,   2,  -3,  -7,  -5, -42, -24,  22,   5,   9,  -4,   4,  -1,   3,   1,  -1,   5,  13,  -7, -12,  11,   3,   0,  -2,  -3,  -1,  10,   7,  23,  -2,  22,   4, -23,   6,  -1,  -2,   1,   0, -12,  -4, -14,  -8,  -1,  30,  -9,  -2,  31,  -7,  26, -22,  -3,  -2,   5,  -9, -15,   2,  10,  -7,  29, -19,  14,  -4,  -5,  -8,   9, -33,  30,  -3,  -8,   8,   7,  -8,  13,  -8,  -2,   6,  -7,   9,  -6,   4,   1,   6,   1,   0,  -3,   3 },
     {    5,   6,   9,  16,  19,  35,  23, -52,   3,  -2,   2,  -4,   4,   4,   6,   8,  12,  38,  39, -61,   2,  -3,   3,  -5,   3,   5,  10,  10,   8,  24,  19, -21,  -9,   2,   1,  -2,   2,   2,   1,   0,  13,  22,  -2,   8, -16,   8,  -4,   0,   2,  -2,   4,   1,   9,   4,  -8,  13,   1,   4,   0,  -7,   2, -11,   1,   7,  -1,  -3,   6,   6,   1, -14,   7,   6,   3,  -6,  -1,   5,   7, -11,   0,   7,  -1,  -2,   0,  -3,  -2,   2,  -1,  -2,   1,   0,  -1,  -1,   1,   1,  -1,   2 },
     {    3,   3,   1,  11,  15,  15,   6,   0,   1,  -3,   0,  -1,   4,   6,  15,   6, -15,   2,  -4, -11,   5, -10,   4,  -1,  -6, -12, -30, -32, -20,  15,   9,  -7,   8,   3,  -1,   4,  -2,   0,  25,  33, -11, -25,  -1,   5,  -6,  10,  -9,   1,  20,  35,  11,   7,  -2,  21, -29,   4,  -8, -20,   5,  -4, -22,  36,  -5, -11,  13,   8, -18, -23, -10,  17,   6,  -4,   8,   2,  23, -28, -18,   9,   9,  -3,   2,  -5,  16,   9,   0,   1,   3,   4,   1,   0,  -1,   1,   2,   2,   3,  -2 },
   }
 },
 {  //29
   {
     { -112,  45,   5,   5,   1,   2,   1,   1,   0,   1,   0,   0,  -8, -16,   9,  -1,   2,   0,   1,   0,   1,   0,   0,   0,  25, -22,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  11,  -3,  -7,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   6,   1,  -4,  -1,   0,   0,   0,   0,   3,   1,  -1,  -2,   0,   0,   0,   0,   3,   0,   0,  -1,  -1,   0,   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {   21,  25, -11,  -1,  -3,   0,  -1,   0,  -1,   0,   0,   0, -99,  42,  16,  -2,   3,   0,   2,  -1,   1,   0,   0,   0, -10, -34,  24,  -1,   3,   0,   1,   0,   1,   0,   0,   0,  21, -26,  -1,   6,  -2,   1,  -1,   1,   0,   0,   0,   0,   8,  -5,  -9,   4,   0,   1,   0,   0,   5,   0,  -5,  -1,   0,   0,   0,   0,   3,   0,  -2,  -1,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   2,   0,  -1,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {  -24,  31,  -7,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  27,  54, -22,  -5,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0, -51,  50,  26, -14,   3,  -4,   1,  -2,   1,  -1,   0,  -1,  -9, -33,  44,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   9, -28,   6,  15,  -4,   3,  -1,   1,   2,  -9,  -8,  10,   1,   0,   1,   0,   2,  -4,  -6,   1,   4,   0,   1,   0,   1,  -4,  -2,  -1,   1,   1,   0,   0,   2,  -2,  -2,   0,   1,  -2,  -1,   0,   1,  -1,  -1,   0,   0,  -1,   0,   0 },
     {   18,  57, -33,  -2,  -6,   0,  -2,  -1,  -1,   0,  -1,   0, -17,  12,  -6,  -5,   1,   0,   1,   0,   0,   0,   0,   0,  39,  18, -28,  -8,  -3,  -2,   0,  -1,   0,   0,   0,   0, -20,  56,  14, -26,   2,  -5,   1,  -3,   0,  -1,   0,  -1,  -4, -13,  50, -10,  -5,  -1,  -1,   0,   2, -17,  16,  18,  -8,   2,  -2,   1,  -3,  -8,  -2,  14,   2,  -2,   1,   0,  -1,  -4,  -2,   3,   5,  -1,   0,   0,   0,  -4,   0,   0,  -1,  -2,  -1,   1,   0,  -2,   0,   1,   0,  -1,   0,   1 },
     {   35,  58, -24,  -3,  -7,  -2,  -3,  -1,  -2,   0,  -1,  -1,  61,   9,  17, -15,  -2,  -6,  -1,  -3,   0,  -1,   0,  -1,  31, -46,  39,   0,  -2,   0,  -2,   0,  -1,   0,   0,   0,   9, -17,  -5,  22,  -8,   4,  -2,   2,  -1,   1,  -1,   1, -15,  10, -17,  14,   2,  -1,   1,   0, -11,   5,  -7,   0,   6,  -1,   1,   0,  -8,   2,  -1,  -2,   3,   0,   0,   0,  -5,   0,  -1,  -1,   2,   1,   0,   0,  -4,   0,   0,  -1,  -3,   0,   0,   0,  -2,   0,   0,   0,  -2,   0,   0,   0 },
     {   -9,   5, -17,   5,  -1,   2,   0,   1,   0,   1,   0,   0,  14,  41, -33,  -2,  -1,  -1,   0,  -1,   0,   0,   0,   0, -32,  -9,   5,   1,   5,   1,   1,   0,   1,   0,   0,   0,   9,   4, -49,  -2,   5,   0,   2,   0,   1,   0,   1,   0, -11,  48,   3, -49,   8,  -6,   3,  -3,  10, -13,  51, -20, -14,   3,  -4,   1,   4,  -5,  14,  17, -14,   1,  -2,   0,   0,  -4,   3,  10,   1,  -3,   1,  -1,   0,  -2,   1,   3,   1,  -2,   2,   2,   0,  -1,   0,   1,   0,  -2,   1,   1 },
     {   13,  45, -24,   4,  -2,   0,  -2,   0,  -1,   0,  -1,   0, -14, -26, -13,  18,   3,   4,   0,   2,   1,   1,   0,   1, -10,  31, -32,   6,   3,  -1,   1,   0,   1,   0,   0,   0, -29,  -6,  -1,   9,   9,   0,   2,   0,   1,   0,   1,   0,  -6,  -1, -47,  -1,  18,  -1,   5,   0,  -7,  19,  -5, -51,  19,  -2,   3,  -1,  11, -11,  29, -32, -14,   8,  -3,   2,   4,   1,  12,   4, -19,   0,  -1,   0,   2,  -2,   6,   5,   1,   0,   3,   2,   2,  -2,   2,   1,   1,   0,   2,   0 },
     {   -8, -35, -15,  11,   2,   2,   1,   2,   1,   0,   0,   1,  -5,  42, -54,   7,  -2,   1,   0,   0,   0,   0,   0,   0,  76,  17,  -3, -26,   3,  -7,   1,  -3,   0,  -1,   0,  -1,  30, -11,  19,  -1,  -5,   1,  -3,   1,  -1,   0,  -1,   0,  -4,  16, -17,  13,  -2,   0,   0,   0, -15,  14,  -1, -10,  10,  -3,   2,  -1,  -5,  -1,  11, -10,   1,   2,  -1,   0,  -3,  -2,   7,  -1,  -4,   1,  -1,   0,  -2,  -1,   3,   1,  -2,  -1,   2,   0,  -1,  -1,   1,   0,  -1,  -1,   1,   0 },
     {  -17, -20, -14,  14,   0,   2,   1,   2,   1,   0,   0,   1,  10,  54,   4,   4,  -5,  -5,  -3,  -2,  -1,  -2,  -1,  -1, -20, -30, -15,  30,   1,   3,   1,   2,   0,   1,   0,   1, -11,  50, -23,   2,   2,  -6,   0,  -2,   0,  -1,   0,  -1, -24, -13,  11,  24,  13,  -4,   3,  -1,  -2,  -5, -41,  -4,  28,  -3,   5,  -1,  -1,   0,   6, -38,  14,   5,   0,   1,   7,  -8,   6, -14, -10,   7,  -2,   2,   1,   1,   6,  -4,   3,  -3,   1,  -1,   0,   1,   3,  -2,   1,  -1,   0,  -1 },
     {   -5,  -9,  48, -13,   1,  -5,   0,  -2,   0,  -1,   0,  -1, -10,  32,  28, -10,  -3,  -4,  -2,  -1,  -1,  -1,   0,  -1,  40,  27,  29,   7, -12,  -2,  -5,  -1,  -3,  -1,  -1,   0, -61,   9,  -5,  41,   4,   2,   1,   1,   0,   0,   0,   0, -21,  -3, -16,  -8,  21,  -5,   5,  -1,   6, -23,  17, -14,   3,   6,  -2,   2,   6, -10,  -6,   9,  -8,   5,  -1,   1,   0,  -1,  -5,   3,   1,  -1,   2,   0,   2,  -3,  -4,  -1,   1,   0,  -1,   0,   1,  -1,  -2,   0,   0,   0,  -1,   0 },
     {   -4, -20,  -2, -17,   8,   1,   3,   0,   2,   0,   0,   0,  13,  11,  29, -33,   6,  -1,  -1,  -1,   0,  -1,   0,   0,   1, -24,   2, -22,   6,   6,   2,   1,   1,   1,   1,   0,  10,   7,  27, -31,  -3,   3,  -2,   0,  -1,   0,   0,   0,  -3, -28,  11,  -1,  -3,  10,  -1,   3,   8,  -5,   2, -38,  -6,  11,  -2,   3,  11, -13,  32,  -4, -48,  17,  -6,   3,   0,   8,  -7,  41, -34,  -9,   3,  -2,  -4,   7,  -5,  19,   0,   3,  -5,   7,  -1,   2,  -1,   3,   0,   1,  -3,   3 },
     {   10,  28,  57, -30,  -2,  -7,  -1,  -4,   0,  -2,   0,  -1,   9,  28,  33,  -2, -13,  -1,  -4,  -2,  -2,   0,  -1,   0,   0,  35, -34,  22,  -9,   0,  -2,  -1,   0,   0,   0,   0,  64,   9, -11,  -7,   2,  -6,  -1,  -2,  -1,  -1,   0,  -1,  12,  17,  -2,   6,  -1,  -1,  -3,   0, -17,  17,   2,  -1,   7,  -2,   1,  -2,  -8, -10,  16,  -3,   4,   1,   1,   1,  -2, -10,   2,   5,   0,   2,   0,   1,  -3,  -5,  -1,   3,  -2,  -3,  -2,   1,  -2,  -2,  -1,   1,  -1,  -2,  -1,   0 },
     {   -9,  -6, -17,   4,  -2,   3,   1,   1,   1,   1,   1,   0, -17, -26,  16, -17,   0,   1,   3,   0,   0,   1,   0,   0, -26,  21,  34, -25, -12,  -1,  -2,   0,  -1,   0,   0,   0,  12,  36,  19,  11, -33,   2,  -6,   0,  -3,   0,  -1,   0, -31,  52,   3,  38,  -9, -12,   1,  -4, -15,  -2,  20,   5,  27, -15,   3,  -3,  11, -27,  12,  -1,  11,   5,  -4,   3,   5,  -9,  -7,   7,  -1,   6,  -1,   1,   1,  -2,  -6,   1,   2,  -2,  -3,   0,   2,  -2,  -2,   0,   1,   0,  -1,   1 },
     {   -7,  -9, -20,   4,   2,  -1,   2,   1,   1,   0,   0,   0,   4,  14,  20,  10, -11,  -3,  -4,   0,  -2,  -1,  -1,   0,  -8,  -9, -28,   7,   6,  -2,   3,   1,   1,   0,   1,   0,   5,  16,  26,   4, -17,  -3,  -2,  -2,  -1,   0,  -1,   0, -18, -14, -33,  23,  14,  -2,   5,   0,  11,  21,   8, -30, -23,  -3,  -2,  -1,  -2,  10,  12,  60,  -4, -21,  -1,  -6,  -4,   4, -19,  -8,  53,  -6,   1,  -1,   2,  -9,  11, -18,   0,  -1,  -5,  -4,   0,  -2,   4,  -4,   0,   0,  -2,  -1 },
     {    2,   8,  23, -23,  -2,   2,   0,   1,   0,  -2,  -1,   0,   4,   0,  -8,  38,   8,  -1,  -5,  -1,  -1,  -2,  -1,   0, -12, -25, -20, -40,  12,   5,   5,   3,   2,   2,   1,   2,  18,  14,  30,  20,  -4,  -3,  -4,  -2,  -4,   0,  -2,   0, -41, -13, -16, -14,  15,  12,   5,   3,  20,  -6,  21,   8, -14,  -1,  -2,   0, -11, -15, -24,  -2,  25,  18,   2,   2,   8,   0,   4, -30, -36,   9,   3,   2,   1,   5,   1,  20,   4,   1,   0,  -7,  -2,   0,   1,   5,   2,   0,   0,  -4 },
     {  -10, -22, -49,  -7,   6,   8,   2,   4,   2,   1,   1,   1,   6,  16,  46, -18,   6,  -4,   0,  -3,  -1,  -2,  -1,  -1,   5,  13,  -4,   1,  -7,   8,  -2,   3,  -1,   2,  -1,   1, -15,   2,  -4,  17,   0,  -1,   2,  -1,   0,  -1,   1,   0,  43,  -7, -16, -27,  -6,   5,   0,   1,  -5,  46,  22,  18, -17, -13,  -2,  -5, -10, -18,   6,  -1,  32,   4,   2,   2,   3,  -9,   7, -17, -25,  11,  -2,   3,  -2,  -3,   1,  18,   0,   0,  -2,  -3,   0,  -2,   1,   4,   0,   0,  -2,  -2 },
     {   -1,  11,  27, -49,  13,  -3,   6,  -1,   0,  -2,   1,  -1,  -6,  11, -32, -41,  11,   6,   4,   1,   3,   0,   1,   0,  -9, -19, -17, -27,  -5,  12,   3,   3,   0,   2,   0,   1, -42,   8, -10, -11, -23,  12,   2,   4,   0,   1,   0,   1,  -5,  17, -18,  10, -36,   0,   4,   1,   4,  32, -18,  13, -16, -17,   6,  -2,  -2,  20,  12,  -2,   5, -19,  -1,  -1,   4,  -1,  16,   1,   6,  -4,  -4,   0,   4,  -4,   8,   3,   3,  -4,   4,   4,   1,  -1,   2,   2,   1,  -2,   2,   2 },
     {    1,  -1, -22, -30,  11,   1,   7,   0,   2,  -1,   1,   0,  -2, -13, -23, -38,   5,   4,   4,   2,   2,   1,   1,   1,  -2,  20,   2, -10, -28,   8,  -3,   3,  -1,   1,   0,   0,  38,  22, -23,  42, -21,  -7,  -2,  -3,   0,  -1,   0,  -1,  26, -27,  -6,   7,  29, -14,   3,  -3,  39, -23,  -9, -12,   8,   7,  -8,   4,  -7,  22, -17,  -1,  -8,   8,  -2,   0,  -8,  12,   4,  -8,  -1,  -3,   3,  -2,  -4,   2,   9,  -3,  -2,   0,   4,   1,  -2,   1,   3,   1,   0,  -1,   1,   1 },
     {    8,  10,  28,  53, -27,  -1,  -7,  -1,  -3,  -1,  -1,   0,   6,   5,  11,  21,   9, -11,   0,  -4,  -1,  -1,   0,  -1,   0,  11,  32, -35,  21,  -3,  -1,  -1,  -1,   0,   0,  -1,   4,  48, -11,  -9, -10,  11,  -5,   1,  -2,   1,  -1,   0,  40,   3, -23,  -3, -11,   7,   1,   0,  41,  13, -16,  -9, -11,  -3,   2,  -1,   0,  16,   5,  -4,  -2,  -6,   0,  -1,  -4,  -2,  11,   4,  -2,  -2,   0,  -1,  -5,  -4,   6,   5,  -4,  -6,   0,   4,  -3,  -3,  -2,   1,  -1,  -2,  -1,   0 },
     {    2,   0,   6, -10,  37, -27,  12,  -8,   3,   0,   2,  -1,   2,  -5,   8, -23,  43, -29,  21,  -8,   5,  -2,   3,  -2,   0,  -2,  10, -21,  43, -30,  10,  -4,   2,  -3,   1,  -1,   2,   2,   6, -20,  27, -21,   4,  -1,   1,   0,   0,   0,   7,   4,   4,  -9,  34, -14,  -6,   2,  -9,   0,   2, -15,  15,   2,  -7,   3,   5,   2,   3,  -6,  18,  -3,  -5,   3,   3, -14,   6, -23,  23,  22, -13,   5,   3,  -6,   4, -14,  -3,  -1,   0,   6,   1,   1,  -2,   2,  -2,   1,  -1,   2 },
     {    6,   2,  24,  -7, -11,  -1,   1,   1,   0,   1,   0,   0,   7,   2, -30,  15, -14,   4,   6,   1,   1,   0,   1,   0,  -1, -43,  -7,  14, -12,  -4,   2,   2,   1,   1,   1,   1, -20,   4,  24,   3, -14, -14,   5,  -4,   1,  -1,   0,  -1,  63,  -4,   5,  17,   8, -28,   0,  -3,  12,   1,  34,  -4,  31, -18,  -9,   1,  15, -24,  17,   4,  14,  14, -13,   6,  -5,   0, -13,   8,   3,  16,  -8,  -1,  -6,   3,  -7,  -2,  -6,   2,  -3,  -2,   0,   1,  -2,  -2,  -2,   1,  -1,   0 },
     {   -1,   3,  18,  10,  -9,   0,  -5,   2,  -1,   2,   0,   0,  -8, -13, -30,  -3,   3,  10,   4,   3,   0,   2,   1,   0, -10,  -9,  39,  10, -31,   2,  -6,   2,   0,   1,   0,   0,   8,  24,  18,  21,   7, -18,   2,  -6,   0,  -3,   0,  -1, -12, -26,  24, -24,  31,  -1,  -7,   4, -37,  50,  -3,  -7, -24,  11,  -5,  -1, -21,  12,  26,   2,  -1,  -8,   7,  -5,  12, -25,  10,   8,   3,  -2,  -2,   0,   2,  -1,  -6,   5,   3,  -4,  -5,   1,   2,   1,  -4,   0,   3,   0,  -1,   0 },
     {    2,  -5,  -3,  -7, -24,   7,   7,  -1,  -3,   1,   1,   1,   1, -11,   7,  -7, -24,  -7,  -3,   1,   1,   4,   0,   0,  13,   8,  10,   3, -17,  -3,   8,  -7,   0,  -5,   1,  -2,  14, -21,   1,  -9,  14, -13,  13,  -1,   5,  -1,   0,  -1, -34, -18,  10, -21, -11, -10,  14,   0,  54,  -7,  -5,  -7,   3, -35,  10,  -2,  27,   4,  35, -10,  35, -27, -16,   3,   8, -13,  19,   3,  13,  20,  -7,   4,   0,  -4,   2,  12,   3,   4,  -7,   7,  -5,   2,  -2,   1,  -1,  -1,   0,   3 },
     {    0,   3, -12, -21,  64, -26,  -9,   2,  -1,   0,   1,  -1,   1,   1,  -6,  40,  19, -14, -19,   6,   1,  -1,  -4,  -2,   5,   3,  26,  45, -19,   0,  -8,  -6,  -1,  -1,  -2,  -1,   2,  13,  28,   1,   6,   0,  -5,   0,  -4,   0,  -1,   1,   1,  27,   5,   0, -18,  11,  -6,   1,  44,  11,  -5,   2,  -8,   1,   1,  -4,   2,   9,   8,  -8,  -8,   4,   1,  -2,   0,   9,   1,   4,  -5,  -2,   3,  -1,  -5,  -4,  -1,   5,  -2,  -7,  -4,   3,  -6,  -4,  -2,  -1,  -1,  -2,  -2,  -1 },
     {    6,   4,  12,  35,  22, -12,  -6,  -2,  -1,  -2,  -1,  -1,   2,  -1, -14,  -6,  58, -16,  -4,  -3,   0,  -2,   1,  -1,   2,   3, -16,  12,  10,  16,  -9,   3,  -3,  -1,  -2,   0,  12, -10, -22,  35, -26,  23,  -4,   1,   1,   1,  -1,   1,  -9, -42,  10,   9, -30,   8,  11,  -1,  -6, -12,  19,   1,  -8, -13,  10,  -2,   5, -26,  26,   1,   7, -10,  -1,   2,   6, -12, -11,  10,   6,  10,  -4,   2,  -4,   7, -10,  -6,   1,   6,  -5,  -3,  -1,   3,  -1,  -3,   1,   1,   0,   0 },
     {   -1,  -2,   2, -14, -35,  11,   1,   3,  -5,   2,   1,   3,  -4,  -6, -28, -33,  16, -16,  -6,  -4,   5,   2,   0,   0,  -2,  -3,   9,  22,   3,  -4,   6, -10,  -3,  -2,  -1,  -2,   8,  23,   7,   4,  49, -10, -11,   5,  -4,   1,  -2,  -1,   3,   5, -27,   0,  -6,  38,  -8,   1,   5,   1, -10,  13,  -8,  13,  20,  -8,  19, -41,   5,   6,   9,  -4,   7,   5, -10,  24, -33,  -5,   2,   1,  15,   3, -11,   6,  10, -12,   9,  -1,   4,  -2,  -5,   0,   5,  -3,   0,   0,   4,   0 },
     {   -3,  -2, -13, -45,  15,  -7,  14,  -5,   9, -10,   0,  -1,  -1,  -9,  -7,  37, -42,   9,  -5,  -3,   4,  -6,   0,   4,   1,   4,  34, -13,  21, -25,   6,  -1,  -1,   5,   0,   2,   2,  20, -38,  -1,   4,  13, -15,   8,  -7,  -1,   1,   2,  13, -34,  -4,  11, -16,  28,   0,  -1, -24, -11,   5,  -9,  -5,  -9,   8,  -1,   0, -16,  13,   6,   0, -11,  -4,   1,  10, -12,   3,  -1,   5,   0, -18,   4,   0,  15,  -8, -11,  -3,   4,   6,  -7,   1,   4,  -2,  -2,   1,  -1,  -2,   0 },
     {    2,   5,   5,   0,  13,  22,  14,  -7,  -9,  -7,  -2,  -1,  -6,  -8,  -8,   1, -21, -41, -14,   1,   4,   1,   1,   0,   3,   5,   6,   7,  20,  32,   5,  -9,  -8,  -3,  -6,  -3, -10,  -8,  -2,  -5, -38, -24,   6,   5,   8,   2,  -1,   0,  15,  17,  14,   2,  15,  16,   5,  -7, -16, -17, -24, -19, -18,  15,  20,  13, -10,   7,  15,  20,  16,   0, -21,   0, -12,  -8, -10, -10, -20,  29,  27,   4,   2,  -1,   1,  16,  22,   4,  -6,  -6,  -5,   1,   4,  -2,   5,  -6,   5,  -4 },
     {    4,   4,   4,   9,  34,   3,  -4,  -3,   1,   2,   1,   0,   5,   3,   0,   5,   8,  45,  -6,  -1,  -6,   2,  -3,   0,   4,   1,  18,  -9, -23,  41,  -2,  -3,   0,  -2,  -1,   0,   7,   9,  -1, -38,  -3,  -5,  35, -20,   4,  -2,   1,  -3,   7,  -2, -27, -17,  16, -13,  23,   1, -15, -17, -22,  -6, -17, -10,   8,   7,  22, -35, -15,  -3,  12, -17,   6,   7,   2,   1, -21, -10,   5,  -5, -10,   5,  -4,  10,  -2,  -8,  -6,   0,   7,  -4,   1,  -2,   2,   2,  -3,  -1,  -2,   1 },
     {   -2,  -6, -10, -14, -14,  -3,  12,   1,  -2,   4,   1,  -1,   5,   7,  15,  24,  15, -19, -15,  -3,  -7,   6,   1,   1,  -5, -13, -20, -28, -25,  18,   7,  10,   3,  -2,  -1,  -2,   0,   4,  22,  34,   7,  -9, -18,  -6,  -1,   0,  -1,   2,  26,  20,   1, -15,   3,  17,   2,   2, -27, -36, -11,   8, -12,   9,   5,  -2,  39,  14,  14, -24, -10, -21,   3,  -4,  -6, -17,  21,  11,  28,  -5,  -4,  -1,   9,   0,  -8,  -8,  -8,   2,   1,   9,   0,   3,  -2,  -2,  -4,  -2,   5,   2 },
     {    0,  -3,   5, -13,  -7,  18, -32,   9,  -2,   3,  -1,   6,  -1,  -4,   0, -17,  -4,  22,  -4,  -2,   5,   1,   1,  -1,   1,  -5,   8,  22,  25,   8, -14,  -1,   4,  -3,  -3,  -2,   3,   4,  25,  -1,  -9,  45,  -7,  -7,   0,  -1,   2,  -4,  12,   3,  21,  -3,   5,  19,  13, -16,  -5,  -6,  28, -39,   4,   2,  34, -13,   3,  13, -16, -23,   8, -20,  20,   7, -11,  -5,  16, -29,   6,   5,  -1,   8,  10, -13,  18,  -3,  -6,   1,  -7,  16,   1,  -2,  -9,   4,  -3,   1,  -4,   3 },
     {   -1,  -1,  -2,  -8,   1, -16,  -2,   2, -26,  34,  15,  -4,  -1,   0,  -7,   3, -15, -29,  38,   5, -15,  30,   7,   0,   0,   3,   3,  14,  20,  -2, -21,   8,  -3, -13,   1,  -3,   4,   8,   0,  -2,  -7,  10, -13,   8,   5,  11,  -6,   1,   0,  -6,  -4,   2,   9, -16,  18,   5, -18,  -2,  -1,  -3, -17, -22, -18,   1,  27,   1, -14,  -8,  14, -14,  11,  -4,   6,  16,  -2,   7,   2, -30,  25, -10, -22,  14,  -8,  38,  -1, -14,  -4,  -1,   5,   6,  -5,  11,  -3,   2,   6,   4 },
   },
   {
     {  100, -11, -21,  -5,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   0, -67,  17,  19,   1,   2,   1,   1,   0,   0,   0,   0,   0,   9, -17,  -5,   4,   1,   1,   0,   1,   0,   0,   0,   0,  -3,  11,  -2,  -4,   0,   0,   0,   0,   0,   0,   0,   0,  -1,  -3,   4,   1,   0,   0,   0,   0,  -1,   1,  -2,   1,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   41, -57,   8,   6,   2,   2,   1,   1,   0,   0,   0,   0,  55,  48, -29, -10,  -5,  -3,  -2,  -1,  -1,  -1,  -1,  -1, -58,  10,  29,   2,   1,   0,   1,   0,   0,   0,   0,   0,   6, -18,  -8,   6,   2,   1,   1,   1,   0,   0,   0,   0,  -6,   9,   0,  -4,   0,   0,   0,   0,   0,  -5,   2,   2,   0,   0,   0,   0,  -2,   2,  -1,  -1,   1,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,  -1,   0,   0 },
     {  -50, -20,  -5,  17,   4,   4,   1,   2,   1,   1,   1,   1, -32,  59,  12, -20,  -2,  -4,   0,  -3,   0,  -1,   0,  -1,   6, -69,  17,  18,   2,   3,   1,   2,   0,   1,   0,   1,  24,  12, -36,  -7,   1,   1,   0,   0,   0,   0,   0,   0,   1,   8,  17,  -5,  -5,  -2,  -1,  -1,   4,  -5,  -7,   7,   2,   0,   0,   0,   0,   4,   2,  -5,   0,   0,   0,   0,   3,  -2,  -1,   2,  -1,   0,   0,   0,   0,   2,   0,  -1,   2,  -1,   0,   1,   0,   1,   0,  -1,   1,  -1,   0,   0 },
     {  -15, -65,   5,  14,   6,   4,   3,   3,   2,   1,   1,   1,  -8,  29, -10, -10,  -1,  -1,   0,  -1,   0,   0,   0,   0,  69,  27,  -4,  -6,  -8,  -5,  -3,  -2,  -2,  -1,  -1,  -1, -58,  -3,  20,   7,   5,   1,   1,   0,   1,   0,   0,   0,  21, -15, -11,   2,   0,   1,   0,   1, -10,   9,   1,  -3,   0,   0,   0,   0,   5,  -4,   1,   1,   0,   0,   0,   0,  -4,   2,   0,   0,   1,   0,   0,   0,   2,  -2,   0,   0,  -2,   1,   0,   0,   1,  -1,   0,   0,  -1,   0,   0,   0 },
     {  -10,  30,  12,  -4,  -4,  -1,  -2,  -1,  -1,   0,  -1,   0, -49,  19, -18,   7,   5,   2,   2,   1,   1,   0,   1,   0,  -7,   9,  45, -17,  -2,  -4,  -1,  -2,   0,  -1,   0,  -1,  -1, -71,  -2,  34,   3,   6,   0,   3,   0,   1,   0,   1,  18,  18, -40, -18,   5,   0,   2,   1,   0,   1,  21,  -4,  -8,  -2,  -2,  -1,   6,  -2,  -9,   5,   3,  -1,   0,   0,   0,   1,   4,  -3,  -1,   1,   0,   0,   3,  -1,  -3,   2,   0,   1,   1,  -1,   1,   0,  -1,   1,   0,   0,   1,  -1 },
     {  -13,  21, -41,   2,   8,   2,   2,   1,   1,   1,   0,   0,   5,  44,  42, -10, -15,  -4,  -4,  -3,  -2,  -1,  -1,  -1, -31,   5, -12,   7,  13,  -2,   3,   0,   1,   0,   1,   0,  -3, -21,  49, -11,  -7,   2,  -1,   0,  -1,   0,   0,   0,   5, -37, -27,  37,   6,   2,   1,   2,   2,  14, -16, -29,   6,   1,   2,   0,   7,  -6,  14,   7, -11,  -1,  -2,   0,  -2,   2,  -8,   1,   6,  -1,   0,   0,   2,  -3,   3,  -1,   0,   1,  -3,   0,   1,  -1,   1,   0,   0,   0,  -2,   0 },
     {   -9, -24,  42,  -4,  -3,   0,  -1,   0,   0,   0,   0,   0, -32,  -1, -38,  16,   8,   3,   2,   1,   1,   1,   1,   0,   0,   2,  10, -25,  -1,   1,   1,   0,   0,   0,   0,   0,  43,  24,  25,  12, -13,  -6,  -5,  -2,  -2,  -1,  -1,  -1, -43, -39, -15,  21,  20,   6,   2,   2,  26,  12, -21, -25,  -6,   2,   2,   0,  -8,   5,  16,   3,  -5,  -4,  -2,  -1,   7,  -2,  -5,   2,   2,   1,  -1,   0,  -4,   2,   2,  -1,   3,  -1,  -1,   0,  -2,   1,   1,   0,   2,   0,  -1,   0 },
     {   10,  61,  26,   8, -16,  -5,  -6,  -3,  -3,  -1,  -2,  -1,   2,  36, -49, -15,  11,   1,   4,   0,   2,   0,   0,   0,  -2,  -1,  21,  -2,  -8,  -1,   0,   0,  -1,   0,   0,   0, -29,  22,  21, -12,   7,  -4,   2,  -2,   0,  -1,   0,   0,  22, -33,  36,  13, -11,   2,  -3,   1, -14, -12, -21,  23,  11,   1,   2,   1,   5,  -5,  -5, -16,   3,   3,   1,   1,  -1,  -2,   2,   6,  -4,  -1,   0,   0,   2,  -1,  -2,  -2,  -1,  -2,   0,   1,   0,  -1,  -1,  -1,   0,  -1,   0,   1 },
     {  -14,  -2, -45,   2,   7,   3,   3,   2,   1,   1,   1,   1, -21,  12,   9,  -7,  -6,   0,   0,   1,   0,   0,   0,   0,   9,  71,  18,  -7,  -6,  -9,  -3,  -5,  -2,  -2,  -1,  -1,  22, -16,  -3,  -7,   9,   3,   2,   1,   1,   0,   0,   0, -36,  -4,  46,  -6, -10,   0,  -1,   0,  26, -27, -13,  28,   3,  -1,  -2,   1, -14,  10,  -4,  -9,   7,   2,   2,   0,   8, -12,   2,   3,  -2,   0,   0,   1,  -3,   3,  -3,  -1,   3,  -4,   1,   1,  -1,   1,  -2,   0,   2,  -2,   0,   0 },
     {   18,  30, -19,   5,  -7,  -1,  -2,   0,  -1,   0,  -1,   0,  38,  -2,   7, -24,   5,  -1,   1,   0,   1,  -1,   0,  -1,  38, -12,  42,  23, -13,  -3,  -5,  -2,  -2,   0,  -1,   0, -17,   2, -33,  22,  15,   3,   2,   2,   1,   0,   0,   0, -39, -30, -17, -17,   6,   7,   4,   2,  28,  32,  -4, -14, -16,  -8,  -2,  -1, -14,  -1,  20,  11,   4,  -2,  -3,  -1,   1,  -3,  -9,   1,   5,   3,   1,   1,  -4,   0,   3,  -3,   1,   1,  -1,   1,  -2,   0,   1,   0,   0,   0,  -1,   0 },
     {   -4,  17, -22, -26,  -4,   7,   2,   1,   1,   1,   1,   0,  16,  47,  -6,  31,  -3, -11,  -4,  -4,  -2,  -2,  -2,  -1,   6,  -8, -24, -36,   6,   9,   5,   3,   2,   1,   1,   0,  -4,  32,  17,  14, -10, -11,  -3,  -4,  -2,  -1,  -1,  -1,  -9,  -2,   6, -33,  17,   5,   3,   1,  12,  -8,  46,  12, -33,  -1,  -4,   0,  -8,  -8, -16,  31,  20,  -4,   0,  -1,   0,  -5,   0, -16,   1,   6,   2,   1,   0,  -2,   0,   8,   0,  -2,  -1,  -4,   0,  -1,   0,   3,   0,  -1,   0,  -2 },
     {   -4,   4,  -4,  54,   9, -11,  -6,  -3,  -2,  -1,   0,  -1, -25,  -8, -29, -55,  -3,  18,   8,   5,   3,   2,   1,   1, -25,  22, -23,  22,  -4, -12,  -1,   0,   0,   0,   0,   0,  11,  23,   9, -18,  11,  -1,  -2,  -4,  -2,  -1,   0,  -1,   4,   0,  -9, -26,  -4,   0,   5,   2,  10,  10,  45, -10, -14,  -5,  -6,  -1,   4, -15,   6,  28,  -3,   0,  -1,  -1,   1,   3,  -6,  -2,   9,  -1,   1,   0,   2,  -5,   5,   0,   1,   0,  -2,   2,   1,  -2,   2,   0,   0,   0,   0,   1 },
     {   14,  23, -13,  18,   7,  -6,  -5,  -2,  -1,  -1,  -1,   0,  19,  18, -24, -14, -12,   7,   3,   1,   0,   1,   0,   1,  40,  -3, -14, -12,   6,  -4,   0,   0,   0,  -1,   0,  -1,  15,  -1,   3,  10, -13,   5,  -1,   1,  -1,   0,   0,   1, -21,  44,  -6,  10,  23, -10,   0,  -4, -20, -48,  -3, -37,   3,  17,   6,   5,  14,  27,  15,  -2, -32, -12,  -3,  -1,  -7,  -7,  11,  19,  15,  -4,  -4,  -2,   0,  -1,  -4,  -5,  -3,  -2,   5,   3,   1,   0,  -1,  -1,  -3,  -2,   2,   2 },
     {   12,  -1,  20,  28,   9, -10,  -5,  -4,  -1,  -2,  -1,  -1,  16, -10,  33, -43,  -8,   9,   1,   2,   1,   0,   0,   0,  18, -23,   3, -10,  -9,  -3,   2,   1,   1,   1,   0,   0,  28,  -7,  50,  35,  -6,  -2, -10,  -2,  -4,  -1,  -2,   0,  -9,   0, -14,   8,  -3,  10,   3,   2,  -7, -14,  -5,  42,  -5,  -6,  -2,   1, -13,  10, -39,  -4,  27,   2,   4,  -1,   2,   0,   6, -15,  -7,   2,   2,   1,  -4,   4,  -6,   5,  -1,   0,  -1,  -6,  -1,   2,  -2,   1,  -1,   0,   0,  -2 },
     {    5,  13,  43,  18,   7, -11,  -5,  -6,  -1,  -2,  -1,  -1,   0,  18,   7,   1,  -8,   4,  -1,   2,  -1,   0,  -1,   0,  -5, -10, -58, -10,   8,   6,   7,   3,   3,   1,   1,   1, -36, -40,  -3,   0,   2,   3,   3,   3,   2,   2,   1,   1, -18,  22,  16,  -7,  -6,  -5,  -2,  -2,  55,   8, -15,   2,   3,  -4,  -2,  -2, -34,   0,  19,  -6,   0,   5,   2,   0,  16,  -9,  -8,   5,  -1,  -1,   0,   0,  -6,   7,   2,  -3,   4,  -3,   1,   1,  -2,   2,   0,  -1,   3,  -1,   0,   0 },
     {    3,  -1,  43, -28,  21,  -2,  -7,  -3,  -1,  -1,   0,  -1,  -4,  -8,  21, -13, -33,  -4,   8,   1,   2,   1,   1,   1,   4,   2,  32,  10,  29,  -5, -12,  -1,  -3,  -1,  -2,  -1, -14,  -4,   2, -28, -24,  20,   8,   2,   1,   1,   0,   1,   2, -10,  22,   9,  -3, -15,  -5,   0,   8, -14,  10,   6, -26,  17,  -5,   1,   7,  17,  -8,  51,  -8, -19,   3,  -4,  -9,  -5, -14,   2,  32,   3,   0,  -1,   1,   1,  -6,  -6,  -1,  -1,  -2,   2,   0,   1,  -3,  -1,  -1,  -1,  -1,   0 },
     {   -9,  -8,  12,  10, -19,   0,   3,   0,   0,   0,  -1,   0, -11,  -7,  39, -13,  34,  -5,  -7,  -2,   0,   0,   0,  -1, -36,   7,  -4, -14, -34,  14,   6,   3,   0,   2,   1,   1, -27,  13,  -1,  31,  15, -17,  -4,  -2,   0,  -2,   0,  -1,   6,  -9,  11, -24,  13,  16,   0,  -2,  -3, -30, -21,  -8, -27,   1,   6,   6,  26,  34,  17,  12,  -7, -15,  -9,  -4, -10, -13,  -4,  16,  16,   4,   1,  -1,   5,   2,  -8, -11,  -1,  -2,   3,   4,   3,   1,  -2,  -2,  -1,  -1,   1,   2 },
     {   -6,   7,  12,  -5,  33,   4,  -7,  -4,  -4,  -1,  -1,  -1, -15,   2,   6,   3, -43,  -1,   8,   5,   2,   1,   1,   1, -24,  13,  -2,  25,  26,  -3,  -8,  -2,  -3,   0,  -1,   0, -36,  35, -15,  24,   3,   8,   6,  -3,   1,  -1,   0,   0, -25, -15, -28, -20,  21,  -6,  10,   4,  -2, -19,  -3,  -3,   8,   1,  -2,  -1,  16,   9, -14, -41,   7,   4,   3,   2,   7,  -6,  27,  -6, -25,   3,  -5,   2,   4,   3,  -7,  12,   1,  -2,   7,  -5,   1,   0,  -2,   4,   2,   0,   2,  -2 },
     {    2,  11,  53, -27, -28, -15,   6,   1,   2,   1,   1,  -1,   5,  42,  38,  -3,  19,  11, -10,  -6,  -5,  -3,  -2,  -1,  15,  38, -18,  29, -21,  -9,   5,   1,   0,  -1,   0,  -1,  25,   8, -18, -12,  12,   6,  -5,   0,   0,   1,   0,   0, -24,   6, -16,   8,  -5,  -1,   9,  -1, -19,  -4,  22,  -5,  13,  -7,  -4,   1,  10, -15,  -9,   1,  -3,  12,   3,   0,  -6,   2,   1,  -3,  -1,  -8,   0,   0,   4,  -6,   2,   1,  -3,  -2,  -1,   1,   1,  -2,   0,  -1,  -1,  -1,  -1,   0 },
     {    1,  -3,   6, -43,  33,  -4,   2,  -1,   2,  -1,   1,   0,   4,  -4,  16, -36,   3,  -2,   8,   2,   0,  -1,   1,  -1,  -1,  -9,  17, -43,  -5,  -4,  -1,   3,   1,   0,   2,   0,   2,  12,  15,   4,  16,   4, -10,  -3,  -2,  -2,  -1,  -1,   8,   0,   8, -26, -17,   8,  10,   3,   5,  -7,  15, -11,  14,   4, -11,   2, -10, -27,  15, -11, -47,  22,   3,   4,  15,  15,  -5,  35, -20, -16,  -3,  -1,  -7,   3,   1,  14,   3,  -1,  -5,   5,  -2,   1,  -1,   3,   1,   1,  -2,   3 },
     {   12,  13,  20,  18,  37,  32,  -8,  -8,  -4,  -1,  -1,  -2,   6,  12,   6,  12, -38, -30,   9,   7,   2,   1,   1,   0,  12,  17, -20, -19,  -3,   9,  -8,  -2,   0,   0,   1,   0,  42,  -3, -21,  20,  20, -11,   4,  -4,  -1,  -2,  -1,  -1,  22, -16,   7, -18, -18,   5,  -2,   5, -25,  27, -13,   8,   5,  -9,   1,  -3,  17,  -1,  25,   4,   8,  -3,  -7,  -1, -23,  -8, -12,   6,   9,   7,   4,   0,   4,  -4,   1, -13,  -5,   1,   1,   4,   0,  -1,   0,  -3,  -2,  -1,   0,   1 },
     {   -2,   9,   5,  12,  10,   7,  -3,  -2,  -2,  -1,   0,  -1,  -7,   4,  17, -16,  -6, -11,   5,  -2,   1,   1,   1,   0, -11,  25,  17, -14, -16,  14,  -5,   2,  -2,  -1,  -1,   0, -12,  27, -23,  32, -20, -15,  -1,  -2,   1,   0,   0,   0,  10,  36,  17,  36,  29,  -8,  -7,  -7,  -6,  11, -13, -16,   6,  15,  12,   3, -28, -48, -10,  10,   6, -14,   2,   3,  21,  19,  -5, -22,  10,   2,  -5,  -1,  -3,  -9,  14,   2,   4,  -2,  -6,  -1,  -1,  -2,   4,  -2,   2,  -1,  -2,   0 },
     {   -5,  -1,  -5, -38, -12,  -5,   7,   2,   1,   1,   0,   0,  -5,   5, -13,  -1, -16,   9,  -7,   3,   2,  -1,   0,   0,   1,   9,   3,  43,  12, -11,   0,  -4,  -1,  -2,  -1,   0,   8,  27,  27,  27,   7,  -1,  -5,   0,  -1,  -1,  -1,  -1,  31,  31, -20, -17, -16,   3,   5,   0,  -5, -12, -28,   7,   3, -16,   1,   3, -39,   8,  38,  11,  12,  -2,  -4,  -5,  19, -18, -23,  -4,  13,   3,   4,   1,  -5,   9,   5, -16,   5,  -7,   3,   6,  -3,   2,   1,  -4,   3,  -4,   0,   0 },
     {    2,   1,   8,  -4, -22,  10,   1,  -7,   0,   0,   0,   0,   0,   1,   3, -34,  21, -39,  -1,   9,   1,   0,   2,   0,  11,  18,   0,  14,  39,  57,  -4, -11,  -5,  -3,  -3,  -1,  13,  -5,   5, -16, -34, -34,  20,  14,   4,   2,   1,   1,   4,  -2, -17, -21,   8,   5, -28,  -3,   3,   1,   9,  10,   2,  11,  14,  -7,  -6,   2,   9, -16,  -5,  -3,  -5,  11,   5,  -6,   3,  11, -17,  10,  -2,  -5,  -6,   5,  -3,   8,   2,  -1,   2,   3,  -2,   1,   0,   2,   1,  -1,   0,   1 },
     {    6,  -4,   6,   3,  -8,  -2,  -3,  -1,   0,   1,   2,  -1,   9,  -1,   0,  12,  -7,   6,   3,   1,  -1,  -1,  -1,  -1,  12,   0, -18,  30,   3,  -8,   0,  -2,  -1,  -1,   3,   0,  26, -26, -10,   5,  23,  -4,   5,  -1,   3,  -1,   1,   0,  39, -44,  17, -22,  36,  10,   1,   3,   0, -33, -12, -20, -22,  22,   2,   7, -24,  -9, -20,   2, -15,  -9,   1,   0,  26,  32,   9,  -8,  -7,  -6,  -6,  -3, -19,  -1,   2,  10,   6,   4,  -6,   0,  -8,   2,   2,   0,   1,   1,  -2,   0 },
     {   -1,   0,   0,  23, -26,  37,   1,   0,  -4,   0,  -4,   0,  -2,  -3,   0,  27, -16, -34,  -5,   6,  -2,   2,   1,   1,  -3,  -5,   5,  20, -19,  19,   0,  -4,  -2,  -1,  -1,   0,  -6,  11,  19,  15,  10,   3,   1,  -1,  -2,  -2,  -1,  -1, -14,   0, -17,  -1, -34, -16,  10,   3,  21, -12,  14,  13,  24,   7,  -8,  -8,  -9,  -2, -26,   7, -37,  -9,  15,   4,   9,  23,   8,  22,  35, -22,  -5,  -8,  -9, -10,  -5, -11,   3,  -1,  -2,  -8,   0,   0,  -2,  -3,   1,  -1,   1,  -2 },
     {    1,   0,  -3,  29, -18, -38,  -9,   9,   0,   5,  -2,  -1,   8,   7,  23,  18,   6,  29,  -2, -13,  -1,  -7,  -1,  -2,  -1,  14,   9, -25,  23, -15,   3,   2,  -5,   2,   0,   1,  13,  22, -15,  13, -27,  11,   2,   1,   7,  -1,   1,  -2,  32, -17, -12, -20,  10, -32,  -6,   1,   6,  22,   0,  20,  11,  21, -14,  -8, -14,   5,   6, -16, -13,  -2,  26,   6,  -5, -16,   1,  20,  17, -14, -15,   4,  -2,   3, -10, -12,   0,  -2,   7,   0,  -3,   1,  -2,   0,   2,  -3,   2,  -1 },
     {   -1,   1, -13, -48,   8,   1,  15, -11,   0,   1,   3,   1,  -2,   5, -24, -40,  -5,   5,  -4,  17,   4,   3,  -2,   2,  -4,   4, -35,   2, -29,   0,   9,  -3,   3,   3,   2,   0,   8,  -3, -26,  19, -15,   3,  -4,   5,  -1,  -1,  -2,  -1,  21, -11,  -5,  28,   2, -10,  -5,  -3,  18,  30,   2,  17, -15,  -3,   1,  -4,   4,  27, -16,  -7,   1, -15,   6,  -3,   4,   3,  36,  21,  19, -12,  -5,  -6,  -6,  -6, -12,  -4,   3,   3,   9,  -9,  -7,  -3,   0,   2,   0,  -3,   3,  -2 },
     {   -7,  -7, -10, -15,  -8, -38, -31,   8,   5,   1,   2,   2,  -3,   2,  -8,   2, -15,  32,  21,  -6,  -5,  -1,  -2,  -1,  -2,  -2,  -9,   6,  19,  25,  -2,   5,   1,  -1,   0,  -1,  -7, -14,  -2,  23, -18, -25,   0,   2,   5,   1,   0,   0, -30,  -5,  12, -15, -17,  19,   0,  -4, -20,   6, -14,  23,   4, -14,   6,   0,  28, -12,   8,   5,  -9,  -3,  -4,  -2,   4,  42,  -4,   4,  23,  -9,  -2,  -7,  -4, -33,  12,  -3,   5,  13, -12,  -3,   3,  -5,   4,  -3,  -2,   0,  -1,   0 },
     {    2,   8,   4,  18,  24,  12,   1,  -7,   1,   2,  -2,   0,   2,   9,   9,  15,  16,   5,   2,  12,   1,  -6,   0,  -1,  -3,   6,   5, -24,  -6, -11,  -9, -10,   3,   7,   2,  -1,  -8,  10, -18, -41, -11,   8,  -1,  15,   0,  -2,  -2,   0,  -2,   0, -37,   3,  -2,  21,   3, -12, -25,  -7,   0,  21, -19, -25,   4,   8, -15,  28,   6,  -9,   0,  -4, -13,  -6,  49,  15,  -5,   5,  17,   0,  -2,  -8, -37, -13,  13,  -2,  17,   0,  -7,  -3,  -4,   3,   2,   0,   5,  -4,   0,   1 },
     {   -4,  -4,  -8, -10, -14,   4, -19, -16,   1,   3,   4,   2,   0,   2,  -3, -17,  36, -19,  28,  21,  -3,  -2,  -6,  -1,  -6,  -2, -18, -14,  10,   4, -25,   3,   8,   6,   5,  -1,   0,  -3,  -1,  12,  15,  41,  15, -10,  -7,  -2,   0,  -1, -13,  -4,  -3,  -6, -18, -40,  -4,  21, -24,  -3, -19,  -6,   5,  23, -23, -17,   6,   8,  11,  19,  16,   2,  21,  -5,  21,   3,   2, -32,  -6,   1,  -5,   8, -13,  -9,   4,   3,   8,   6,   1,   1,  -3,  -3,   2,  -1,   2,   0,   1,   0 },
     {   -3,   0,   2,  -9,  -6,  27,   3,  10,   9,  -2,   1,   0,  -3,   1,   0,  -1,   7,   4,   6,   9,  -8,   6,  -1,  -4,  -8, -11, -11,  12, -11, -38, -58, -12,   6,   0,   2,   2,  -2,  -9,  21,   1,  -3,   3,  32,  -2,   0,  -3,  -2,  -5, -18,  14,  11, -10,  40,  28, -17,  -2, -15,  21,   0,  23, -10,   5,  -7,  -5,  -7, -18,   3, -21, -20,  -2,  19,   2, -17,  -9,  -9,   9,  16,  -2, -15,  -6,   9,  -2,  -5,  -7,   3,   4,   2,   3,   3,   1,   2,   3,   1,  -2,   3,   0 },
   },
   {
     {  104, -42, -15,  -1,  -3,  -1,  -1,   0,  -1,   0,   0,   0, -41,  35,  -3,  -2,   0,   0,   0,   0,   0,   0,   0,   0, -23,   3,  11,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   3,  -6,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -14,  37, -15,  -1,  -3,  -1,  -1,   0,   0,   0,   0,   0, -85,  16,  35,  -6,   5,  -1,   2,  -1,   1,   0,   1,   0,  44, -53,   4,  10,  -1,   2,   0,   1,   0,   0,   0,   0,  22,   4, -21,   3,   0,   0,   0,   0,   0,   0,   0,   0,   3,   4,  -1,  -5,   1,   0,   0,   0,   3,   1,  -1,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0 },
     {  -32, -83,  57,  10,   4,   3,   2,   2,   1,   1,   1,   1, -35,  11, -23,  15,   2,   2,   1,   1,   0,   1,   0,   0,  40,  13, -25, -12,   2,  -1,   0,  -1,   0,   0,   0,   0,  16,   7,   0,  -6,  -4,   0,  -1,   0,  -1,   0,   0,   0,  -5,   9,   1,   0,  -1,  -1,   0,   0,   1,   0,   2,  -1,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   1,   0,   1,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
     {   40,  -5,  -4,  -6,   0,  -2,   0,  -1,   0,   0,   0,   0,   3, -68,  32,  13,  -1,   5,  -1,   2,   0,   1,   0,   1,  34,   3, -52,  18,  -1,   2,   0,   1,   0,   1,   0,   0, -29,  53,  -9, -20,   5,  -3,   1,  -1,   1,  -1,   0,  -1,  -8,  -2,  17,  -5,  -2,   1,  -1,   0,  -2,   2,   0,   2,  -1,   0,   0,   0,  -2,  -1,   2,  -1,   1,  -1,   0,   0,  -2,   3,  -1,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,  -1,   1,   0,   0,  -1,   0,   1,   0,  -1,   1,   0,   0 },
     {    2,  11, -31,  15,  -5,   3,  -1,   1,   0,   1,   0,   0,   2,  38, -25, -14,   6,  -3,   1,  -1,   1,   0,   0,  -1,  66,  16,  10, -28,  -5,  -2,  -2,  -2,  -1,  -1,  -1,  -1, -53,  17,  43,   3,  -7,  -2,   0,  -1,   0,   0,   0,   0,  -7, -34,  16,  18,   1,   0,   0,   1,   1,  -9, -12,   9,   4,   1,   0,   0,  -1,  -2,  -4,  -1,   4,   0,   1,   0,  -1,  -2,   0,  -1,   0,   1,   0,   0,   0,  -2,   0,   0,   0,  -1,  -1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0 },
     {   -1,  34,   9, -18,   0,  -4,   0,  -2,   0,  -1,   0,  -1, -50, -23,   7,  24,  -2,   5,   0,   2,   0,   1,   0,   1, -18,  57, -30, -18,   9,  -5,   3,  -2,   1,  -1,   1,  -1,  15, -31,  50, -17,  -5,   2,  -2,   0,  -1,   0,   0,   0,  23, -25, -11,  25,  -7,   2,  -1,   1,  -2,   9, -10,  -2,   6,  -2,   1,   0,   4,  -3,   0,  -1,  -1,   1,   0,   0,   1,   2,   0,  -1,   0,  -1,   0,   0,   2,  -2,   0,   1,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   1,   0,   0 },
     {  -20, -33, -64,  63,  -1,   6,   1,   4,   0,   2,   0,   2, -15, -37,  22, -22,  19,  -1,   3,   0,   2,   0,   1,   0, -12,  33,  17, -32,  -8,   4,  -2,   0,  -1,   0,   0,   0,  22,  10, -11,   8, -11,  -2,   0,  -1,  -1,   0,   0,   0,  13,   4,  -3,   0,   3,  -4,   0,  -1,  -6,   6,   4,  -3,   2,   0,   0,   0,   1,  -2,   5,  -2,  -1,   1,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0,   1,   0,   1,   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,   1,   0,   0 },
     {   11, -15,  21,  -5,   0,  -1,   0,  -1,   0,  -1,   0,   0,   4, -25,  30,  -5,   1,   0,   0,   0,   0,   0,   0,   0,  30,  -3,  22,   2, -13,   2,  -4,   0,  -2,   0,  -1,   0, -43, -39,   2,  49,  -1,   3,   0,   2,   0,   1,   0,   1,  12,   0, -63,   8,  19,  -2,   3,   1,  -3,  29,  -6, -31,   5,   2,   1,   0,   4,   2,   8,  -4,  -7,   1,   0,   0,  -2,   5,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   2,   0,  -1,   0,   0,   0,   0,  -1,   2,   0,  -1 },
     {   24,  26,  -9, -12,  -4,  -1,  -1,   0,  -1,   0,  -1,   0,   5, -33, -45,  28,   6,   3,   3,   2,   1,   1,   1,   1,  34,  -2,  15, -40,   6,  -2,  -1,  -1,   0,   0,   0,   0,  35,  -5,   7,  12, -20,   5,  -4,   1,  -2,   1,  -1,   0, -42,  49, -16,   1,   7,  -7,   4,  -3, -10, -11,  32,  -9,   1,   0,  -2,   0,  -1,  -6,  -2,  11,  -3,   2,  -1,   0,  -2,   1,   1,   0,   1,  -1,   1,  -1,  -2,   0,  -1,   2,  -1,  -1,   1,  -1,  -1,   0,  -1,   1,   0,   0,   1,  -1 },
     {    8, -10,  10,  -8,   2,  -2,   1,  -1,   1,  -1,   0,   0,  33,  10,  54, -28, -11,  -4,  -3,  -2,  -2,  -1,  -1,  -1,  16,  -5,  -2,   0,  -3,   1,  -1,   1,   0,   0,   0,   0,  55,  -1,  22,   2, -21,   1,  -6,   0,  -2,   0,  -1,   0, -42,  -3,   7,  53,  -2,  -3,  -1,  -1, -14,   0, -34,   6,  23,  -1,   3,   0,  -4,  -2,   0,  -9,   3,   4,   0,   0,  -2,   0,  -7,  -1,   1,   1,   1,   0,  -2,   1,   0,  -2,   0,   0,  -2,   0,  -2,   1,   0,  -1,   0,   0,  -2,   0 },
     {   -3, -12,   5,   0,  -2,   3,  -1,   1,   0,   0,   0,   0,  -6, -27, -24,  27,   8,  -1,   2,   0,   1,   0,   1,   0, -29, -63,  29, -23,  22,   6,   1,   3,   2,   1,   1,   1,  -9,  39,  17, -14, -27,   9,  -4,   0,  -2,   1,  -1,   0,  21, -16, -12,  36,  -8,  -6,   2,  -2,  12,  12, -32,  -7,  15,  -4,   1,   0,  -4,  12,   7, -16,  -2,   2,  -1,   0,   3,   2,   3,   0,  -5,   0,   0,   0,  -1,   3,   1,   1,   1,   1,  -1,   0,   0,   1,   1,   0,   0,   1,   0,   0 },
     {    4,  10,  48, -20,  -9,  -1,  -1,  -2,  -1,  -1,   0,  -1, -20, -28,  12, -15,  14,  -1,   4,   0,   1,   0,   1,   0, -12,  27,  50, -18, -21,   5,  -4,   0,  -2,   0,  -1,   0,  -5,  25, -16,  37,  -9,  -6,   1,  -1,  -1,   0,   0,   0,  -6, -17,  25,  -4,  10,  -3,  -2,   0,  17, -36,  -5,  32,  -7,   7,  -3,   1,  -3,   9, -32,   3,  11,  -2,   4,  -1,   2,   0,  -1,  -6,   2,   1,   0,   1,   1,   2,  -5,  -2,   2,   0,   0,  -1,   0,   1,  -3,   0,   1,   0,   0,   0 },
     {  -17, -28, -43, -60,  58,  -1,  11,   2,   4,   1,   3,   1, -11, -13, -14,   3, -23,  20,  -2,   5,  -1,   2,   0,   1,   6,   6,  32,  29, -38,  -8,   2,  -3,   0,  -1,   0,  -1,   8,  12,  19,  -3,  15, -16,  -2,  -2,  -2,  -1,   0,  -1,   6,   7,   0,   6,   5,   3,  -4,   0,   1,   0,  -2,   4,   1,   2,   1,  -1,  -3,  -3,  -2,   1,   0,   1,   1,   1,   2,   0,  -2,  -2,   1,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 },
     {  -13,   9,  -8,  -1,   0,   1,   0,   1,   0,   1,   0,   0, -10,  25,   2,  25, -12,   1,  -3,  -1,  -2,   0,  -1,   0, -24,  21, -16,   4,   1,   2,   1,  -1,   0,   0,   0,   0,   1,  43, -15,  24,  -1, -13,   1,  -4,   1,  -2,   0,  -1, -45, -24, -15,   7,  46,  -7,   4,   0,  25, -17,  -3, -55,   7,  15,  -1,   3,   4,  -3,  19,  -5, -23,   5,   0,  -1,   7,  -2,   1,   3,  -3,  -4,   0,   0,   0,  -1,   4,   2,   3,  -2,   0,   0,   0,  -1,   2,   0,   3,  -1,   0,   0 },
     {   -7,  -8, -14,   2,  -6,   9,   2,   0,   1,   0,   1,   0, -17, -29, -27, -32,  28,   4,   5,   3,   1,   1,   1,   1, -17, -21, -30,  21, -11,   6,   0,   3,   0,   2,   0,   2, -14, -44,  25,  13,  -3,  -3,   1,   0,   2,   0,   1,   0, -51,   9,  20, -10,   4,  -2,   0,   1,  41,  -3, -29,   2,  -1,  -1,  -1,   0,  18,  19,  -5, -17,  -2,  -1,  -1,  -1,   2,   9,  11,   0,  -8,   0,  -2,  -1,   1,   5,   3,   2,   2,   1,   2,   2,   0,   2,   0,   0,   1,   1,   1,   1 },
     {   10,  10,  -1,  31, -12,  -5,  -2,  -1,  -2,   0,  -1,   0,   4, -15, -25,  40, -16,   8,  -3,   2,  -1,   1,   0,   0,  24,  14,  29,  21,  -5, -14,  -2,  -5,   0,  -3,   0,  -2,  23, -18, -34,  -3,  44,  -5,   1,   0,   1,  -1,   1,   0, -16, -30,   3,   0,  -7,  15,  -2,   2,  -9,  10, -40,   7,   9,  -3,   8,  -3,  -7,  17,   2, -40,  10,   2,   2,   1,   3,  -1,  14,  -4, -11,   2,  -1,   1,  -2,   2,   2,  -2,  -1,   0,   3,   0,  -1,   1,   0,  -2,   0,   0,   1,  -1 },
     {   11,  12,   8, -20, -23,  14,  -3,   2,  -1,   1,   0,   0,   4,  -3, -45, -38,  31,  -5,  10,  -1,   2,   0,   2,   0,   7,   2, -22,  17, -27,   7,  -1,   3,  -1,   1,   0,   1,  35,  20,  -5,  25, -13, -17,   0,  -3,   0,  -2,   0,  -1,  38, -19,  10,  -1,  21,  -5,  -7,   1, -33,  34,  -5, -11,   3,   6,   0,  -2, -15,  -2,  24,  -9,  -2,  -1,   2,  -1,  -1,  -9,   4,   9,  -5,   1,   0,   1,  -2,  -2,   1,   2,  -1,  -2,   1,   1,  -1,   0,   1,   0,   0,   0,   0,   1 },
     {   -7, -20, -16, -34, -39,  39,   1,   7,  -1,   3,   0,   2,   7,  -8,  12,   4,   7, -19,  10,  -3,   2,  -2,   1,  -1,  10,   3,  22,   6,  33, -29,  -5,   0,  -1,   0,  -1,   0,  18,  -1,   9,  -6,   3,  16, -13,   1,  -3,   0,  -1,   0, -18, -38,   0, -22,  -5,  22,   6,   2,  20,  27,   4,  -8, -39,  -1,   1,   0,  -6,   6,   5,  33,  -1, -13,   0,  -2,  -1,   3,  -9,   3,  11,  -1,  -1,   1,  -3,  -1,   1,   2,   0,   1,  -4,   0,  -1,   1,   1,   0,   1,   1,  -2,   0 },
     {  -12,  -7, -17, -19, -43,  42,   0,   7,  -1,   4,   0,   2, -10,   3,   0,   4, -13, -11,   9,  -1,   2,   0,   1,   0,  -4,  18,  -1,  32,  28, -34,  -5,  -1,  -1,  -2,   0,   0, -19,   8,  -4,  22,  -8,  14, -10,   1,  -1,   0,  -1,   0,  15,  41,   2,  14,  -3, -11,   5,  -9, -17, -27,  -3,   6,  40,  -6,   1,  -1,   8, -10, -11, -26,  10,  13,  -2,   2,  -1,  -4,   7,  -9, -14,   5,   2,   0,   3,   0,  -1,   0,   0,  -1,   2,   0,   3,  -1,   0,  -1,  -1,   0,   2,   0 },
     {   -3, -10, -18,  15, -15,   4,  -2,   3,  -1,   1,   1,   0,  12,   8,  11,  52,  -6, -10,  -9,  -3,  -1,  -1,  -1,   0,  -4, -23, -25,   7, -14,  12,   5,   2,   3,  -1,   1,   0,  14,  -6,  29,  31, -11, -24,   2,  -4,   0,  -1,   0,  -1,  16,  -3,  -2,  -8,  37,  -1, -10,   2,  -6,   1,   0,  20, -16,  15,  -2,  -4, -10,  11, -42,   9,  32, -12,  12,  -4,   0,   3,   5, -37,  10,   8,  -1,   3,   1,  -1,   4,  -2,  -1,   0,   1,  -8,  -1,  -1,   1,   0,  -1,   1,   0,  -3 },
     {   -5, -10, -26, -45,  30,   2,   4,   3,   1,   0,   1,   0,   9,   9,   8,  24,  18, -16,   1,  -6,  -1,  -2,   0,  -1,   7,   7, -20, -30,  15,  24,  -3,   3,  -2,   2,  -1,   1,  -8, -31, -55,  11, -30,  19,  12,   3,   4,   1,   2,   1,  -1, -26,  20,  -1,  -5, -23,   6,   2, -12,   3,  -4,   7,   6,  -7,  -9,   1,  -1,  21,   1, -15,   6,   0,  -3,  -3,   5,  -2,  16,  -3, -11,   4,  -3,   0,   0,   4,   1,   8,   0,   0,   3,   0,   0,   1,   0,   2,   0,   0,   1,   0 },
     {   -9,  -3,   6,   3,   0,  -1,   3,   1,   0,   1,   0,   1, -11,   2,   1,  23,  42, -17,   0,  -4,   1,  -2,   1,  -1, -16,   0,  19,  26, -16,  18, -11,   2,  -4,   1,  -1,   0, -16,   6,   8, -13,  -6,   3,   8,  -1,   0,  -1,   1,   0, -39, -15,  -1,  -3,  -4,   2,   6,   3, -59,  17,   8,   3,   3,  -9,   4,  -1,  43, -43,  -7,   1,  10,   2,  -4,   3,  18,  12, -23,  -5,  -3,   4,   0,   1,   3,   4,   2,  -9,   1,   2,  -1,   0,   2,  -1,   1,  -1,   1,   1,  -1,  -1 },
     {    7,   4,   3,   6,  12,  -1, -10,   6,  -1,   0,  -2,  -1,  11,   6,  11,  18,  43, -39,   6,  -7,   1,  -2,   0,  -2,  17,  17,  18,  33, -15,  10, -12,  -1,  -5,   0,  -2,   0,  15,   1,  20, -40, -16,  26,   7,  -3,   0,   0,   1,  -1,  19,  15, -20, -30,  21, -12,   9,   3,  21, -34, -25,  -3,   6,  10, -11,   7, -16,   9,   6,  -4, -15,   5,   0,  -2, -14,  -2,  12,   9,  -8,  -2,  -1,  -2,  -1,  -3,  -3,   4,  -2,  -2,  -2,   0,  -1,   1,  -1,   0,  -1,   0,   0,   1 },
     {    3,  -8,   0,  -4,  17,   3,   6,  -4,  -1,  -1,   1,   1,   6,  -9,   7, -10,  12,   5,   0,   0,   0,   0,   0,   0,   1, -23,  -8, -19,  24, -17,  -1,   5,   1,   0,   2,   1,   9,  -8,  17,   8,  21,  -2, -10,   3,  -2,  -1,  -1,   0,  15, -20,   6, -16,  10,  34,  -8,   0, -48, -53,   1, -19, -12,  29,  10,   2,  15,   7, -20, -19, -37,   3,   6,   5,   5,  18,   6,  10, -21, -13,   0,   0,   3,   5,  -3,   9,  -3,   4,   0,   1,   0,   2,  -2,   2,  -1,   2,   0,   0 },
     {    0,  -6,   1,  11, -19,   6,  -3,   3,  -2,   1,  -1,   1,  -4, -17,  -7,   8, -31,   0,   5,   0,   2,   0,   1,   0,   7,  -6,  19,  19,   1,   6,  -2,  -2,  -2,  -1,   0,  -1,  -8, -34,  -9, -22, -16,  -9,  15,   4,   3,   0,   2,   0,   9, -10,  45,  20,  22, -40, -13,   3, -13,  -5,   8, -13,  12,  35, -14,   4,  10,  10,  -1,  25, -35,   0,   9,  -7,  -5,   5, -23,  17,  27, -14,   4,  -1,   0,   0,   1, -16,   0,   0,  -1,   2,   0,   2,  -1,  -3,   0,   2,  -1,   0 },
     {   -5,  -3,   1, -21, -32,  14,   5,   0,   1,   1,   1,   1,   4,  11,  29,  27,   9,  -8,  -6,  -3,   1,  -3,  -1,  -1,  -3,   1,  12, -40, -21,  15,   9,   1,   3,   0,   1,   0, -14,  -5,   9, -14,  21, -36,   2,   1,  -2,   1,  -1,   0,   1,  42,  29,  -6,   7,  11, -20,   0,   7,  23, -14,   4,  -8,  18,   9,  -5,   3, -10,  26, -25, -10,   4,   7,   3,   7, -16,   3,  29, -20,  -5,   0,   5,  -5,   3, -12,  11,   1,  -3,  -1,   2,  -1,   0,  -2,   1,   2,  -2,   0,   1 },
     {   -2,  -2,  -4,   0,  -3, -57,  50, -11,   7,   0,   3,  -4,   0,   0,   1,   5,   9,   3, -25,  10,  -3,   4,  -1,   2,   6,   9,  12,  17,  45,   5, -43,   1,  -2,   2,  -1,   1,   4,   5,  14,  29,  -9,  -4,   8, -10,   1,   0,   0,  -2,  11,  15,  34, -16,  -6,  13,   4,  -2,   4,  16, -16, -21,   8,   6,  -5,  -2,   4,  -2,   2,  10,   5,   4,  -4,  -3,  -9, -16,   1,   2,   3,   9,   2,   1,  -5,  -2,  -5,  -2,  -3,  -1,  -2,   1,   3,   0,  -2,  -2,  -1,  -1,   0,   0 },
     {    6,  10,  12,  21,  48,   9, -39,   2,   0,  -2,  -3,   1,   0,   1,   5,  16,  -2,  -5,   5,  -7,   3,  -1,   1,  -1,  -1,   2,   4,   1,   0, -41,  36,  -2,  -2,   0,   1,   0,  -4,   0,   4,  26, -29,  18, -22,  12,  -2,   1,  -1,   2,   3,  12,  44, -25,   5,  16,  -9,  -6,  -1,  29, -25, -16,  10,  -3,  10,  -2,   7, -15,   1,  13,  -5,  -3,  -2,   3,   2, -15,   0,   2,   3,   4,  -1,   0,  -1,   3,  -1,  -4,   1,  -2,  -2,   1,   0,   1,  -1,  -2,   0,  -1,  -1,  -1 },
     {    2,  -7,   3,   1,   2,  -3,  -9,   4,  -2,   2,   1,  -1,   6,  -7,   7,   0,   8,   5,  -6,  -3,  -1,   1,   1,  -1,   3, -13,   5,  -1,  11,  -3,  11, -11,   1,  -1,   1,   0,  10, -10,  16,   3,  18,  -7,  -1,   4,  -1,   1,  -1,   1,  -1, -29,  13, -18,   4, -18,  -4,  10,  23,  11,  38,  -1,  35,  -2, -32,   0, -29, -28, -11,  -5,   5,  62,  -5,  -2,   7,   2, -16,  -6, -41,   8,  18,  -3,  -3,   6,  -6,   9,   2,   5,  -4,  -1,  -1,   2,   0,   2,   1,   1,  -2,  -1 },
     {    0,  -3,  -3,   1, -24,  23,  -4,   4,  -2,   3,  -2,   1,   5,   3,   7,  17, -18,  53, -30,   3,   0,   0,  -3,   2,   3,  -3,   0,   5, -25,  28,   8, -15,   2,  -2,   2,   0,  10,   1,  15,  22, -27,  16,  26, -10,  -5,   0,  -1,   2,   8, -10,   5, -32, -34,   1,  28,   7,  -4, -19, -13,  -8,   1, -28,   0,   9,  -1, -11,   2,   6, -14,  -5, -10,   5,   2,   8,   5,  24,  -6, -23,   1,  -3,  -7,   1,  -2,   3,   3,   1,  -2,  -2,  -2,   1,  -1,   0,   1,   0,   0,   1 },
     {    4,  -2,  -6,  -4,  -7,   3,   8,  -7,  -2,  -4,   2,  -2,  11,   6,  -3,  20,  44,  -8, -17,  10, -12,   2,  -4,  -1,  10,   1,  -5,   7, -17,   0,  24,  -7,   9,  -2,   3,  -2,   9,  -5, -21,  11,   8, -13, -11,   9,   0,   3,   0,   3,  22,  -4,   7,  34, -23,  23, -24,  -5,  33, -19,  10, -18,  17, -14,  15,  -9,  27,  -1,  -7,  18,  -3,   6, -12,   3, -37,  30, -16,   6,   7,   4,   8,  -7,  -5,  -3,  15, -19,  -1,  -1,   3,   0,  -1,   0,   4,   1,  -6,   0,  -2,   2 },
     {   -2,  -3,  -2,   8,   2,   1,  14,  -5,   8,   0,   0,   1,  -3,  -5,  -2,  17,   3,  -7,  15,  -7,   3,   0,   1,  -2,  -4,  -7,  -7,   8, -18,  19, -19,  24,  -7,   0,  -4,   5,  -7, -19,   2,  14,   6,  19, -34,  10,   7,  -2,   0,   2,   1,  -3,  22,  20,  -1,  11,  26, -27,   4, -20,  14,  -6,  -9,  -8,  33,   2, -61, -16,  15,  -8,  -5,  -3,  -5,  17,  15, -10, -34,   3,   3,  -5,  -3,   5,  11,   1, -12, -13,   2,   1,  -2,  -1,  -2,   3,   0,  -5,   2,   2,  -1,   1 },
   }
 },
 {  //30
   {
     { -114,  31,  11,   4,   3,   2,   2,   1,   1,   1,   0,   0,  23, -32,   5,   0,   1,   0,   1,   0,   0,   0,   0,   0,  24,  -7,  -9,   1,  -1,   0,  -1,   0,   0,   0,   0,   0,   3,   5,  -4,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,  -2,   0,   0,   0,   0,   1,   1,   0,  -1,  -1,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    1, -33,   9,  -1,   3,   0,   1,   0,   1,   0,   0,   0,  92,  -8, -30,   1,  -6,   0,  -2,   0,  -1,   0,  -1,   0, -21,  62,  -6,  -9,  -1,  -3,   0,  -2,   0,  -1,   0,  -1, -18,  -2,  27,  -4,   1,  -1,   0,   0,   0,   0,   0,   0,  -4,  -8,   3,   9,  -1,   1,   0,   0,  -6,  -1,  -1,   2,   2,   0,   1,   0,  -2,  -1,   0,   0,   1,   0,   0,   0,  -3,  -1,   1,   0,   1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {   31, -12,  -3,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,  -3, -63,  10,   8,   3,   3,   1,   2,   1,   1,   0,   1,  46,   3, -58,   7,  -4,   2,  -1,   1,  -1,   1,   0,   0, -17,  60,  -7, -24,   2,  -5,   1,  -2,   0,  -1,   0,  -1,  -3,  -1,  27,  -7,  -4,   0,  -1,   0,  -3,   3,   4,   7,  -3,   0,  -1,   0,  -3,   0,   3,   2,   2,  -1,   0,   0,  -2,   2,   0,   0,   1,   0,   0,   0,  -1,   0,   2,   0,  -1,   1,   0,   0,  -1,   0,   1,   0,  -1,   1,   0,   0 },
     {  -11, -71,  15,   7,   5,   3,   2,   2,   1,   1,   1,   1,  27,   6,  -9,  -3,  -1,  -1,  -1,   0,   0,   0,   0,   0, -10, -41,   7,  19,   0,   4,  -1,   2,   0,   1,   0,   1,  13,   8, -64,   5,   5,   2,   1,   1,   1,   1,   0,   0,  -8,  38,   4, -35,   2,  -4,   1,  -2,   5,  -4,  18,  -1, -10,   1,  -2,   0,  -3,   8,   1,   4,  -2,  -2,   0,  -1,   0,   0,   2,   1,   1,  -1,   0,   0,  -1,   3,   0,   0,   0,   0,   1,   0,  -1,   2,   0,   0,   0,   0,   1,   0 },
     {   -7,  28, -27,   4,  -3,   1,  -1,   0,   0,   0,   0,   0,  20,  55,  -1, -20,  -3,  -4,  -1,  -3,   0,  -1,   0,  -1, -20, -14,   0,   7,   1,   2,   1,   1,   0,   0,   0,   0,   2,  34,  -5, -36,   3,  -5,   2,  -2,   1,  -1,   0,  -1,   5, -12,  67,  -6, -20,   0,  -5,   0,   4, -15,  -3,  43,  -6,  -3,  -1,  -1,  -5,   0,  -6,   3,  17,  -3,   2,  -1,   2,  -7,  -1,  -1,   3,   5,  -1,   1,  -2,  -1,  -1,   0,   1,  -3,  -1,   1,  -1,  -1,   0,   0,   0,  -2,  -1,   1 },
     {   36,  74, -18,  -3,  -7,  -4,  -3,  -2,  -2,  -1,  -1,  -1,  60, -25,  24,  -2,   0,  -2,  -1,  -1,  -1,  -1,   0,   0, -17, -28,   7,  22,  -1,   4,  -1,   2,   0,   1,   0,   0, -16,  -7, -30,  17,   8,   2,   2,   1,   1,   0,   0,   0,  -9,   9, -18, -16,  10,   0,   2,   0,  -3,   3,   8, -14,  -4,   2,   0,   0,  -1,   1,   3,   2,  -5,  -1,   0,   0,  -2,   2,   1,   0,   1,  -1,   0,   0,  -1,   0,   1,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   -3,  25,  -7,  -6,  -1,   0,  -1,   0,  -1,   0,   0,   0, -26,  -1, -41,   7,   5,   3,   3,   2,   2,   1,   1,   1,  -5,  40,  -4, -25,   1,  -3,   1,  -1,   0,  -1,   0,  -1, -15,  -9,  -7,  -2,   8,   0,   3,   0,   2,   0,   1,   0,   7,   9,   2, -60,   3,   0,   1,   0,  14,  -9,  54,  -3, -42,   4,  -7,   1,   2,   1,  -2,  37,  -6, -12,   0,  -3,   1,  -2,   5,   2,  12,  -3,  -2,  -1,   2,  -3,  -1,   4,   1,  -1,   2,   1,   1,  -1,  -1,   2,   0,  -1,   1,   0 },
     {   13,   4,  11,  -3,  -5,  -1,  -2,   0,  -1,  -1,  -1,   0,  17,  20,  -9, -13,  -1,  -3,   0,  -2,   0,  -1,   0,   0,  91,   2,  49, -13, -16,  -4,  -7,  -2,  -3,  -1,  -1,   0, -16,   4,   0,  38,  -5,   3,  -2,   1,  -1,   0,  -1,   0, -14,   7,   2,  -1,   3,  -2,   1,  -1,  -6,  -8,   4,  27,  -2,  -2,  -1,   0, -11,   2, -17,   5,  19,  -3,   2,  -1,  -3,  -4,   0,  -8,   5,   5,   0,   1,  -3,   0,  -5,  -1,  -2,  -1,   1,  -2,  -2,   0,  -2,   0,  -1,  -1,   0,   0 },
     {  -10,  -5, -41,  11,   4,   4,   2,   2,   1,   1,   0,   1,   9,  53,  -7,  -7,  -3,  -3,  -1,  -2,  -1,  -1,   0,  -1,  30,  -4, -17,  -1,   8,   0,   1,   0,   0,   0,   0,   0, -16,  33,   5,  -1,  -1,  -1,  -1,  -1,   0,   0,   0,   0, -23, -11, -11,   2,  31,  -2,   6,  -1,   3, -12,   5, -60,   6,  15,   0,   4,   4,  -7,  24,   0, -42,   6,  -1,   1,   0,   3,  -2,  16,  -4, -13,   2,  -1,   0,  -3,   6,   1,   0,  -1,   0,   2,   0,  -1,   2,   1,   0,   0,   0,   1 },
     {    5,  -9,  -7, -15,   2,   1,   2,   1,   1,   0,   0,   0,  22,  -3,  23,  -1, -11,  -2,  -2,  -1,   0,   0,  -1,   0,   2, -34,  -7, -28,   7,   3,   4,   2,   2,   1,   1,   0,  50,   6,  37,  -3, -33,   0,  -8,   0,  -3,   0,  -2,   0, -11,  -2,  -9,  20,   0,   4,   0,   1,   6,   8,  12,  -5, -37,   1,  -1,   0,  -5,  11,  -4,  57,   0, -23,   0,  -5,  -4,   3, -11,  -5,  33,  -2,  -2,  -1,  -2,  -1,   2,   0,  -2,   2,  -5,  -2,  -1,  -1,   1,   1,  -1,   1,  -2,  -1 },
     {    5,  16, -46,   8,  -1,   3,   1,   2,   0,   1,   0,   0,   1, -29, -56,   0,   9,   4,   5,   2,   3,   1,   1,   1,  -1,  -3,  13, -30,  -3,  -4,   2,  -1,   0,   0,   1,   0,  53,   8,   0,  -2, -20,  -1,  -4,   0,  -2,   0,  -1,   0,  -8,  50,   9,  18,  -3, -14,  -1,  -4,  -8,  -5,  11,   1,  31,  -3,  -1,  -1,  -1,  -8,   4, -20,   0,  15,  -1,   2,  -2,  -2,   6,   4, -14,   0,   2,   0,  -1,   1,   0,   1,   0,  -1,   3,   1,  -1,   1,   0,   0,   0,  -1,   2,   0 },
     {  -11, -34, -85,  15,   6,   8,   5,   4,   2,   2,   1,   1,  -3, -29,  12, -32,   9,  -2,   5,  -1,   2,   0,   1,   0,  -1,  -5,  28,  -2, -15,   2,  -2,   1,  -1,   0,  -1,   0, -35, -14,   3,   5,  -3,  -2,   3,   0,   2,   0,   1,   0,  29, -18,  -8,  -6, -11,   2,  -2,   3,   6,  27,  -8,   8, -14,  -7,  -2,  -2,   2,   8,   7,   5,   7,  -8,  -2,  -2,   0,   4,   1,   3,   3,   2,  -2,  -1,   1,   2,   2,   0,   0,   2,   0,   1,   1,   1,   1,   0,   0,   1,   0,   1 },
     {  -10,  -2, -26,   8,  -3,   3,   1,   1,   1,   1,   1,   0,  -2,  25,   2,  40,  -6,  -8,  -2,  -2,  -1,  -1,  -1,  -1, -11,   5, -32,   5,   3,   2,   2,   2,   2,   0,   1,   0,  -2,  23,   0,  30,  -2, -14,   0,  -5,   0,  -2,   0,  -1, -11,   4, -30,  -3,  23,  -2,   5,  -1,   2,  23,  12,   6,   1, -23,   1,  -5,  -3, -12,   0,  -5,  65,   0,  -5,   0,   5, -12,   6, -41,  -5,  32,  -1,   5,  -1,  -3,   3,   3,   2,  -2,   0,  -9,   1,  -2,   2,   1,   1,  -1,   1,  -5 },
     {    7,   5,  -8,   7,   2,  -1,   0,  -1,   0,   0,   0,   0,  16,  29,  -7,   7,  -6,   1,  -2,   0,  -2,  -1,  -1,   0,  45,  -7, -38,   8,  20,  -3,   2,  -1,   1,  -1,   1,  -1,   3, -57,  -5, -46,   9,  13,   3,   5,   1,   2,   1,   2,  19,  22, -18,  -1, -40,   2,  -3,   0, -20,  34,   9,   3,   3, -18,   2,  -5,   2,  -6,  17,   0,   4,  -2,  -4,   1,  -5,   3,   2,   4,   2,   2,  -2,  -1,   0,  -2,   4,   0,  -2,   1,   0,   2,   0,  -1,   2,  -1,  -2,   0,   0,   1 },
     {   -5,  -5, -16,  -2,   2,  -2,   1,   0,   1,   0,   0,   0,  -6,   7,  70, -20,  -8,  -3,  -5,  -1,  -1,   0,  -1,   0,   2,  62,   8,  28, -14,  -8,  -5,  -4,  -1,  -1,  -1,  -1,  33,   7,  -7,   2,   7,  -2,  -2,  -2,  -1,  -1,   0,   0,   7,  38,   1,   0,  -1,  -3,   0,  -3, -17,  -7,  30,   1,  16,   2,   1,   0,  -1, -25,  -2,  13,  -5,   8,   2,   3,  -6,   0, -11,   4,   9,  -3,   2,   0,   1,  -7,  -3,  -5,  -2,  -1,  -3,   2,   0,  -3,  -1,  -1,  -1,   0,  -2,   0 },
     {   -4,  11,   1,  19,  -5,  -3,  -2,  -1,  -1,  -1,   0,  -1, -15,   3, -26,   5, -22,   9,   1,   1,   2,   0,   1,   0,  -9,  15,  -4,  31, -13,  -5,  -1,  -2,  -1,  -1,   0,  -1, -24,   7, -22,   7, -23,  -1,   5,   1,   4,  -1,   1,   0,   0,  25,   0,  34,  -6, -26,   0,  -6, -13,  -2, -25,  -5,   7,  -1,   9,  -3,   8,  25,   6,  15,  -3, -44,   3,  -7,   2,  -7,   8,  -2,  62,  -8, -13,  -2,   5,  -8,   5, -20,   1,  -1,   2,   0,   0,  -3,   1,  -6,   1,  -1,   1,   0 },
     {    5,   3, -17,  35,  -9,   1,  -4,   0,  -2,  -1,  -1,   1,   1,  -9, -15,  13,   0,   2,  -3,   0,  -1,   0,   0,   0,  17,  24,   4,  32,   4,  -7,  -4,  -5,  -2,  -2,  -1,  -1,  68,  -8,  -8,   6,  37,   0,  -1,  -2,  -1,   0,  -2,   0, -17, -40,   3, -23,   8,  24,   1,   5,  -4,   3, -34,  -2, -23,   8,   8,   0, -12,  23,  -1, -10,  -2, -16,   6,   0,   0,  -2,   4,  -4,   6,  -4,  -2,   1,  -3,   3,   1,  -5,   0,  -1,   2,  -2,  -2,   0,   1,  -2,   0,  -1,   1,   0 },
     {    6,  12,  37,  -1,  -3,  -3,  -1,  -1,  -1,   0,  -1,   0,   1,   8, -11, -38,   9,  -1,   2,  -3,   0,   0,   1,   0,  -5,  10,  14, -10, -28,  -2,  -2,   1,  -1,   0,   0,   1,  20,  38, -12,   1,  -4, -10,   1,  -3,   0,  -1,   0,  -1,  15, -12, -15,  -6,  -9,   0,   1,   2,   9,  55,  -3, -11,  -4, -23,  -3,  -5,  -2,  -4,  63,   0,  10,  -4, -10,  -2,   1, -14,   0,  26,  -2,   9,  -2,   2,  -6,  -1,  -3,   3,  -1,  -5,  -3,   5,   0,   1,   0,  -2,   0,  -3,  -1,   3 },
     {    4,   1,  18, 105, -12, -11,  -8,  -3,  -5,   0,  -2,   1,   1,  -5,  22,  10,  28, -10,  -1,  -4,  -2,  -1,  -1,   1,  -6,   0,  12, -31,  18,   4,  -2,   0,   0,   0,   0,   0, -14,   3,   5,  -8,  -4,  12,  -1,   2,   0,   1,   0,   0, -17,   4,   4,   1,  -6,  -2,   3,   0,  10,   1,  13,   3,   9,  -3,  -3,   1,   0,  -8,   3,  12,   0,   4,  -2,   2,  -2,   0, -11,   2,   7,   3,  -1,   0,   1,   0,  -3,  -7,   2,   0,  -4,   0,  -1,   1,  -1,  -3,   0,  -1,  -2,   0 },
     {   -8,  -2,  -6, -14,  -3,   1,   2,  -2,   1,   0,   1,   0, -12,  -6,   2, -33,  -3,  10,  -2,   3,   0,   0,   1,   1,  -9,   9,  -8,   1,  -8,  11,   0,   0,  -1,   0,   1,   0,  -7, -11,  -5,  -3,   2,  14,   2,   3,   2,   0,   1,   1,-102,  -8,  -8,  -1, -37,  -1,   9,   3,   7,   1,   4,   8,  -3, -28,   1,  -5,  17,  -4,  -1,  -4,   6,  -1,  -4,   1,   9,   7,  -2,   1,  -3,  -6,  -3,  -2,   5,  -1,   7,   0,   7,   2,   2,   1,   1,   0,   1,   2,   1,   1,   0,   0 },
     {    1,  -2,  -2, -19,  22,  -2,   0,   2,  -1,   1,  -1,   0,  -6, -21,  -4,   3, -11, -24,   3,   0,   2,   2,   1,   0,   8,   5,  38,  18,  50, -16,  -8,  -4,  -4,  -1,  -3,  -1,  -9, -10,  -4, -29,   7,   9,  -2,   2,   0,   2,   1,   1, -17,   3,  24,  10,  21,   3,  -4,  -5,  12, -11,  -2, -23,   3,  -2,   8,   4,  15,  -1,  34,  15,  22,  -3, -23,   1,  -4, -18, -17,  -3,   5,  50,   0,   4,   0,   8, -12,  -3,  -2,   0,  -2,   1,  -1,   1,  -3,  -2,   0,  -1,  -1,   2 },
     {   -5,  -2, -13, -30, -14,   3,   2,   4,   2,   2,   1,   0,   5,   8,  32,  62,  22,  -5,  -5,  -7,  -3,  -3,  -1,  -2,   1,  24,  30, -37,  17,  12,   3,   2,  -1,   0,   0,   0,  -3,  21, -31, -17, -22,  21,   3,   2,   2,   1,   0,   0,  -3,   2,  -8,  -5, -24,   4,   9,   0,  -7,   0, -22,  -1, -12, -10,   7,   2,  -4,  15,   7, -20,  -9, -16,   2,  -1,   5, -10,  17,  15,  -2, -11,  -8,  -1,   1,   2,  -1,   5,   1,  -4,   5,   3,  -1,  -2,  -1,   1,   0,  -1,   2,   0 },
     {   -3,  -2,  -4,  -5,   0,   5,  -3,   0,   2,   0,  -1,   2,  -2,  -5, -19,  41, -18,   4,  -2,  -1,  -4,   1,  -1,   1,  -5, -19,  35,  39,  -4, -26,  -3, -10,   1,   0,   2,   1,  -9,  34,  32, -11,  25,  18, -10,  -5,  -3,   2,  -2,  -2,   0,  -3, -23,  -3, -33,   6,  10,  -2,  25,  15,  21,   8,  19, -13,   1,  -4, -21, -19,  -5,  12, -12,  11,  12,   7,   5,   2, -20,   8,  10, -23,   2,   2,  -1,  -3,  -2, -12,   4,   1,  -3,  -1,   0,   0,  -1,  -1,   2,   0,  -1,   0 },
     {   -5,  -5,   4,   3, -25,   7,  -5,   3,  -1,   0,   0,  -1,  -6,  -4,  -9, -11,   4,  -4,  -5,   4,   4,   1,   2,  -1, -12, -19,   7,   1,   9,   9,   0,   1,  -2,   1,   1,   1, -11,  20,  33,  12,  23,   6, -10,  -4,  -4,  -2,  -4,   1,   3,  23,  -3, -11,  -7,  31,   3,   1, -83, -14,  -3,  -7, -25, -10,  27,   3,  13, -22,   5,  -6,   0, -20,  -8,   7,   5,  -3,  -3,  10,  -1,  11,  -9,  -5,   8,   4,  -5,   2,   1,   3,   1,  -2,   1,  -4,   2,   2,  -1,   1,  -2,  -2 },
     {   -6, -10, -11, -20, -50, -25,   0,   5,   3,   3,   2,   3,   3,   2,   7,  29,  30, -12,  -8,  -5,  -5,  -2,  -3,  -1,   0, -11, -18,  -5, -33, -13,   4,   2,   2,   2,   0,   2,  -6, -31,  16,  23,  30, -21, -13,  -5,  -3,  -1,  -1,   0, -14,  16,  37,   2,  10,   6,  -5,  -2,  17,   4,  -1,   5,  13,  -4,  -3,  -2,   6,  14,  28,   6,  -3,   5,  -5,  -2,  -6,  -7,  -9,  28,  18,   3,  -1,   1,   0,   0, -14, -15,  -3,   0,  -3,   1,   1,   3,  -5,  -7,   0,  -2,   1,   3 },
     {    1,   8,   2,  -2,  75, -12,   7,  -3,  -2,   0,   1,   0,  -1,   9,   1,   0,  54,  -3,   0,  -3,  -2,  -1,  -1,   1,  -4,   0,   0,   1, -13,  -9,   6,   3,   5,   0,   0,   0, -10,   9,  17, -10,  24, -13,  -2,   3,  -1,   3,   1,  -1, -17,  44, -11,   3,   1,  29, -12,   4,   5,  11, -26,   4, -15,  16,   2,  -7,  -7,  10,  -6,   1,   0,  -3,  10,  -6,  -6,  -8,   8,  -9,   5,  -4,   7,   6,   7,  -6,   8,  -7,   0,  -3,  -2,   5,   1,  -1,  -1,  -4,   1,  -3,   0,   1 },
     {   -3,   0,   3,  -5, -44,  -5,  -6,   3,   1,   4,   0,  -1,  -4,   0,   6, -31, -18,  -6,  -3,   6,   2,   0,   3,   2,  -1,   7,  -1,   4,  52,   4, -13, -10,  -7,  -1,   4,  -2,  -9,  10,  25,  10,   6,   8,  -2,  -3,  -4,  -4,  -2,   2,  -3,  45,   6,  -7,   0,   7,  -8,  -2,  40,  37, -14,   0,  -7,  27,   3,  -5,  -8,  14,  -6, -17,  -4,  -7,  20,   2,   1,  -8,  26, -11,  -7, -10,  -3,   3,  -3,  -6,  11,  15,  -2,  -6,  -1,   3,  -2,  -5,  -1,   3,  -1,  -2,   0,  -1 },
     {    5,   3,   2,  24,  17,  19,  -2,  -5,  -5,  -2,  -3,   1,   4,   0,  19,  10, -53,  14,   1,   1,  -1,   2,   0,   0,   3,  -5,  -1, -11, -32, -24,   9,   2,   2,   1,   0,   1,  -2,  -2,  19,  -4,   7, -10,  -7,   0,  -1,   3,  -1,  -3,   4,  25,   0,  -5,  -8,  16,   4,  -1,  36, -32,  -6, -10, -16, -28,  11,   9,  17,  14,   2, -19,   4, -11, -18,   2,  -4,  -1,  34,  24, -14,  11, -11,  -3,  -7,   2, -13,  24,  -7,   1,  -2,  -1,   1,  -2,  -4,   1,  -2,   1,   1,  -3 },
     {   -2,   6,  -1,  12,  -8,  17,   6,  -4,  -4,   2,   1,  -2,  -7,   1,  -9,  27, -29,  13,  -2,   4,   3,   2,   1,  -1,   0,  11,   9,  12, -17,  38,  -7,   8,   3,  -1,  -3,  -2, -12,  -7, -15,   2, -17, -14,   6,   4,   5,  -3,   3,  -1, -11,  18,  17,  24,   3,  31, -15,  -4,  -1,  22, -15,   4, -32,  12,   9,  -3,  -7,  -9,   9,  28,  -2,  22,  -1,  -5, -24,   1, -35,  11, -37, -11,  29,  -6,   4,   9,   8,   8,  -1,  -4,  -3, -10,   0,   0,   1,  -2,   2,  -3,  -2,  -6 },
     {    3,   3,   1,  -5, -16,   1,   4,   5,   2,   4,   2,   0,   4,   7, -11, -14,  40,   4,  -7,   8,   0,  -3,   0,  -1,   5,  10,  -7,  21,  19,  24,  -6,   2,   0,  -2,   0,  -2,   8,  16,  -2,  15,  -1,   6,  -9,  -3,  -1,  -3,  -2,   2,  28,   6, -16,   9, -24,  19,  12,  -8,  35, -33,  -9, -11,   2, -34,   9,   7,  24,  18, -15,  15,   9,  12, -29,   0, -31,  31,  -7,   2,   0,   9,  10, -13,   2,  -3,  19, -14,  -4,   3,   4,   0,  -3,  -4,   5,  -1,  -1,   2,  -3,  -1 },
     {    3,   1,   2,   8,   8,  12,  19,   3,  12,   7,  -2,  -1,   4,   4,  -3,  12,  16, -16,  19,   8,  15,   6,  -1,  -2,   4,   2,  -7,  28,   1,   1,  17,  -7,  11,   2,   3,  -1,   6,   1,   4,  26, -15,  21,  -7,  -7,   5,   3,  -3, -10,   1,  -5,   8,  15, -12, -10, -15,   9,   6,  18,  13,   6, -31,   9, -15,   2,  40, -15, -25, -16, -19,  16,   2,  -4,  15, -36,   3,  34,  11,  27,  -5,  -4, -20,   8,   4,   7,   8,  -3,   8,  13,   1, -16,   1,   9,   0,  14,  -3, -10 },
     {    2,   2,   4,   1, -22,  29,   1,  16,  12,   0,  -5,  -1,   2,   0,   0,  -4,  10, -19, -14,  23,   3, -10,  -9,   3,   3,  -1,   1,   3, -25,   1,  12, -42, -23,  -5,  29,   1,  -2,  -3,   1, -31, -21, -16,   3,   0,   9,   3, -17,   3,  -3,   5, -21, -10,   8,   7, -23,  16,  11, -15,  -6,  -2,  -1,  -1,  15,   7,   4, -10,  -2,  -5,   0,  -8,  21,   8,   1,   0,  -6, -15,   4,  10,  22,  18, -11,  27,  -6, -13,  16,  -7,   6,  22,   2,   5,   3, -15, -10,  -5,   2,  30 },
   },
   {
     { -106,  47,   7,   4,   2,   1,   1,   1,   0,   0,   0,   0,  45, -27,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  13,  -2,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   19, -39,  17,  -1,   3,   0,   1,   0,   1,   0,   0,   0,  89, -15, -26,   0,  -5,   0,  -2,   0,  -1,   0,  -1,   0, -52,  47,   1,  -3,   0,  -1,   0,  -1,   0,   0,   0,   0, -15,  -7,  11,   1,   1,   0,   1,   0,   0,   0,   0,   0,  -3,   2,  -1,   1,   0,   0,   0,   0,  -3,  -1,   2,   0,   0,   0,   0,   0,  -2,   1,  -1,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
     {   48,  56, -42,  -4,  -8,  -2,  -3,  -1,  -2,  -1,  -1,   0,  21, -67,  38,  -2,   4,   0,   1,   0,   1,   0,   0,   0,   8,   9, -23,   6,  -2,   2,  -1,   1,   0,   1,   0,   0, -26,  30,   4,  -8,   1,  -2,   0,  -1,   0,   0,   0,   0,  -5,  -8,  11,   2,  -1,   0,   0,   0,  -4,   2,  -2,   3,   0,   0,   0,   0,  -2,  -2,   1,   1,   1,   0,   0,   0,  -2,   2,  -1,   0,   1,   0,   0,   0,  -1,  -1,   1,   0,  -1,   1,   0,   0,  -1,  -1,   1,   0,  -1,   1,   0,   0 },
     {    8,  60, -47,  10, -10,   3,  -4,   1,  -2,   1,  -1,   0,   0,  29,  -4,  -8,   0,  -3,   0,  -1,   0,  -1,   0,  -1, -70,  -1,  36,  -4,   7,  -2,   3,  -1,   1,   0,   1,   0,  37, -39,   3,   4,   0,   1,   0,   1,   0,   0,   0,   0,   0,   2,  -3,   1,   0,   0,   0,   0,   6,  -6,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   3,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0 },
     {    1,  22, -27,  10,  -5,   3,  -2,   1,  -1,   1,  -1,   0,  -4,  46, -36,   5,  -3,   0,  -1,   0,   0,   0,   0,   0,  35,  39,  -2, -24,   1,  -6,   1,  -3,   0,  -2,   0,  -1, -55,  -9,  54,  -5,   2,  -1,   1,  -1,   0,   0,   0,   0,  23, -37,   2,  12,  -1,   2,  -1,   1,  -6,   2,  -2,   1,   1,   0,   0,   0,   5,  -8,   1,   2,   1,   0,   0,   0,  -3,  -1,   0,   0,   1,   1,   0,   0,   2,  -4,   0,   1,  -2,   0,   0,   0,   1,  -2,   0,   0,  -1,   0,   0,   0 },
     {    8,  -5, -31,  21,  -5,   5,  -2,   3,  -1,   1,  -1,   1,  52,  46, -18, -21,  -3,  -5,  -1,  -2,   0,  -1,   0,  -1,  22, -58,  23,  10,   0,   2,   0,   1,   0,   0,   0,   1, -15,  44, -28,  -5,   2,  -1,   1,  -1,   0,   0,   0,   0, -20,   3,  30,  -9,  -1,  -1,   0,  -1,   0,  -9,  -2,   9,  -1,   0,   0,   1,  -6,   2,   3,  -2,   2,  -1,   0,   0,  -1,  -3,   0,   2,   1,   1,   0,   0,  -2,   0,   2,  -1,  -1,  -1,   0,   1,  -1,   0,   1,  -1,  -1,  -1,   0,   1 },
     {   16,  -3,  -3,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  21, -15,  18, -16,   7,  -4,   2,  -1,   1,  -1,   0,   0,  32, -25,  50,  -7,  -6,   0,  -3,   0,  -2,   0,  -1,   0, -32, -47,  -9,  45,  -3,   8,  -1,   3,   0,   2,   0,   1,  19,  19, -56,  -2,   6,   0,   2,   0, -20,  23,   7, -13,   0,  -1,   1,   0,   4,  -1,  -3,   0,  -1,   0,   0,   0,  -7,   6,   1,  -2,   0,   0,   0,   0,   2,   0,  -2,   0,  -3,   3,   1,  -1,   1,   0,  -1,   0,  -2,   2,   0,   0 },
     {  -17, -26,   3,   2,   7,   0,   5,   0,   2,   0,   1,   0,   3,  17,  68, -42,   3,  -9,   0,  -4,  -1,  -1,  -1,  -1, -33, -17,   1,  22,  -1,   6,  -1,   2,   0,   1,   0,   0, -17,   3,  18, -13,  -2,   0,   0,   0,   0,   0,   0,   0,  32, -46,   4,  29, -10,   4,  -4,   2,  -3,  17, -35,   8,   6,   0,   1,  -1,   5,   0,   0,  -4,   1,   0,   0,   0,   0,   4, -10,   1,   1,   0,   0,   0,   2,   1,   0,  -2,   0,   2,  -4,   0,   1,   0,   0,  -1,   0,   1,  -2,   0 },
     {  -26, -36, -55,  61, -13,  17,  -5,   7,  -1,   3,  -1,   2, -19,  -5,  10,  -8,   4,   0,   1,   0,   1,   0,   1,   0, -14,  32, -24,  16,  -5,   0,   0,   0,   0,   0,   0,   0, -18,   5, -32,   6,  13,  -3,   4,  -1,   2,  -1,   1,   0,  18,  14, -11, -30,  10,  -5,   3,  -2,  -2,   2,  23, -18,  -2,  -1,  -1,   0,  10,  -8,  10,  -2,  -4,   0,  -1,   0,   1,   0,   5,   0,  -2,  -1,  -1,   0,   4,  -2,   3,   0,   0,   0,   2,   1,   2,  -1,   1,   0,   0,   0,   1,   0 },
     {    0,  31,  39, -29,   3,  -9,   1,  -4,   0,  -2,   0,  -1, -18,  10,  -9,   7,  -3,   3,  -1,   1,   0,   1,   0,   0, -38,  -3,  13,  19,  -5,   4,  -2,   1,   0,   1,   0,   0, -60,  15, -40,   9,  13,   0,   6,   0,   2,   0,   1,   0,  42,  -1,  16, -35,  -3,  -2,  -1,   0,  -2, -13,  22,  14, -11,   1,  -2,   0,  14,  -7, -11,   7,   2,  -2,   1,   0,   0,   0,   2,   0,   1,   0,   0,   0,   5,  -3,  -3,   0,   0,   0,   1,   0,   3,  -1,  -2,   0,   0,   0,   1,   0 },
     {   -6, -12,  -1,   5,  -4,   3,  -1,   2,  -1,   1,   0,   1, -10, -12,  22, -44,  24,  -9,   8,  -3,   3,   0,   1,   0,  22,  43,  41, -26, -12,  -2,  -6,  -2,  -3,  -1,  -1,  -1,   5, -11,  -1,   7,  -3,   4,  -1,   2,  -1,   1,  -1,   0,  -9,  15,  43, -13, -18,   3,  -5,   1,   6, -34,   5,  50, -13,   2,  -2,   1,  -6,   9, -28,   8,  10,  -3,   3,  -1,   1,  -3,  -1,   2,   2,   0,   0,   0,  -2,   2,  -7,   0,   0,  -1,  -1,   1,   0,   1,  -3,   0,   0,  -1,   0,   1 },
     {   -8, -13,   3,  25, -16,   9,  -6,   3,  -2,   2,  -1,   1, -24, -42, -26,  23,  11,   3,   3,   2,   1,   1,   1,   1, -23,  -8,  56, -39,   7,  -4,   2,  -2,   1,  -1,   1,   0,  -9,  43,  -8,  21, -10,  -1,  -2,  -1,  -1,   0,  -1,  -1,  -2, -43,   0,  12,  11,   3,   1,   1,  17,   9, -27,  -9,   1,   2,   2,   0,  -1,   5,   5,  -9,  -2,  -2,   0,   0,   8,   2,  -3,   2,  -2,   0,  -1,   0,  -1,   2,   0,  -2,   3,   1,  -1,  -1,  -1,   1,   0,  -1,   2,   1,  -1,  -1 },
     {    3, -16, -14,  13,  -2,   3,   0,   1,   0,   1,   0,   1,   4, -28, -23,  19,   1,   2,   1,   1,   0,   1,   0,   1,  -1, -34,   0, -21,  20,  -4,   6,  -1,   2,   0,   1,   0,  25,  -9,  10, -39,   1,  -1,   0,   1,   0,   0,   0,   0,  67,  12,  20,  -7, -30,   4,  -8,   1, -43,  19,   9,  28,  -3,  -3,   1,  -2,   4,  -3, -13,   4,   6,   0,   1,   0, -11,   9,   1,  -2,   0,  -1,   0,   0,   3,   1,   0,   0,  -5,   2,   0,   0,   1,   0,  -1,   0,  -3,   1,   0,   0 },
     {    5,   7,   5,  23, -17,   6,  -6,   2,  -3,   0,  -1,   1,  -4,  -9, -17,  24, -20,  13,  -6,   4,  -1,   2,   0,   1,  12,  12,   5,  52, -24,   0,  -6,  -2,  -2,  -1,  -1,  -1,  -4, -31, -20,   9,  15,  -1,   4,   0,   2,   0,   1,   0,   0,   1,  16,  19,  -9,  -6,  -1,  -1, -15,  -8, -51,  26,  34,  -8,   6,  -2,   0,  17, -10, -45,  16,   2,   3,  -1,  -1,  -3,   9,  -5,  -3,   3,   0,   0,   1,   3,  -1, -10,  -1,  -1,   5,  -1,   1,   1,   0,  -4,  -1,  -1,   2,   0 },
     {   -4,  -7, -40, -43,  60, -27,  23, -10,  10,  -3,   5,  -2,   0,  -3, -19,  -9,   4,   5,  -1,   3,   0,   0,   0,   0,  14,  29,  21,  10, -25,   2,  -8,   1,  -4,   0,  -2,   0,  31,  12, -30,  10,  10,  -2,   0,  -1,   0,  -1,   0,   0,  14, -29,   7,  -8,   1,   3,   0,   2, -19,  22,  -6,  -8,   3,  -3,   1,   1,   0,  -8,  18, -12,  -5,   4,  -2,   1,  -2,   0,   4,   7,  -6,   1,   0,   0,   0,  -2,   3,   1,  -1,   1,   1,   3,   0,  -1,   1,   0,  -1,   0,   1,   1 },
     {    3,  18,  21,  11,  -2,  -7,  -4,  -4,   0,  -1,   0,  -1,  -6,  12,  -5, -31,  22,  -8,   8,  -3,   4,  -2,   1,  -1,  -9,  17,  -4, -18,   4,  -1,   3,   1,   0,  -1,   0,   0,   5,  39,  -7,   9, -26,   4,  -5,   2,  -3,   0,  -1,   0,  10,  40,  -5,  43,  -3, -12,   1,  -5, -48, -12,   3, -11,  40,  -2,   6,   0,  21, -26,  15, -25,  -7,  10,  -2,   3,  -3,  -8,   7,   9,  -8,   2,  -1,   1,   9,  -5,  -2,  -1,  -1,  -3,   2,   4,   2,  -2,  -2,  -1,  -2,  -2,   2,   1 },
     {  -16, -17, -34, -25,  16,   5,   6,   3,   3,   1,   2,   0,   3,  15,  41,  51, -54,  11, -13,   2,  -6,   1,  -2,   0,  -6,   9,  15, -19,  15,  -7,   5,  -2,   1,  -2,   1,  -1,  -9,  23,  -2,  17, -21,   3,  -1,  -1,   0,   0,   0,   0,  -2,  23,  -6,  21,  -1,  -7,   2,  -3,  -7, -19,   7,  15,  12,  -8,   2,  -2,   5,  -6, -26,   3,  20,  -2,   2,  -1,   0,  -1,   3, -20,   5,   3,   1,   1,   2,  -3,  -1,   3,   0,  -1,   0,  -5,   1,  -1,   0,   1,   0,   0,   0,  -3 },
     {    9,  15,  21,  10, -37,  14,  -9,   4,  -4,   2,  -2,   1,  13,  22,   7,   1,  -4,  -1,  -3,  -1,  -2,   0,  -1,   0,   9,  19, -17,   1, -26,  11,  -2,   1,  -1,   0,   0,   0,  42,  23, -14,  38, -18, -10,  -1,  -3,  -1,  -2,   0,  -1,  16, -30,  -9,  -6,  10,   3,   1,   1, -22,  35,  11,  18, -21, -12,   2,  -4, -14,  -9,  -8,  21,  35, -14,   3,  -3,  -3,   1,   0, -22,  14,   7,  -1,   1,  -3,  -6,   4,   0,  -1,   0,  -2,  -4,  -1,  -3,   1,   0,  -1,   0,   0,  -2 },
     {    6,  -2,   1,  -6, -10,   6,  -2,   4,  -2,   1,  -1,   1,  15,   7,  15,  11,   1,  -7,  -1,  -3,  -1,   0,  -1,  -1,  10,  -4, -18, -16,  11,   3,   2,   1,   1,   2,   0,   0,  21, -11, -10,  24,  -8,  -5,   0,  -1,  -1,   1,   0,   0,  22, -32,  18, -14,  38,  -2,  -1,   1, -27, -54, -10, -25,  -7,  25,  -1,   8,  25,  35, -18, -11, -36,  -3,  -1,  -2, -20,  11,  10,  14,  -4,  -9,   0,  -3,   3,   6,  -4,   1,  -7,   2,   5,   2,   1,   3,  -2,   2,  -3,   1,   2,   0 },
     {   -6,  -7, -20,   5,  28, -15,   9,  -5,   3,  -2,   2,  -2, -10, -14, -30, -13,  20,  -1,   5,   1,   1,   1,   2,   0, -23, -34, -26,   9, -41,  27,  -8,  10,  -1,   4,   0,   2,  -8,  12,  46,  32, -19, -10,  -5,  -3,  -1,  -1,  -1,  -1,   7,  13, -13,   1,  19,  -1,   4,  -3,   5, -23,   2,  16,  -6,  -3,   2,   0,  -2,  19, -30,  -6,  20,  -6,   1,  -1,   3,   1,  16, -18,  -4,   5,  -1,   0,   1,   2,   0,   3,   1,   1,   5,  -4,   1,   1,   0,   2,   0,   0,   2,  -2 },
     {   -5,  -4,   2,   2, -11,   7,  -2,   5,  -2,   1,  -1,   1, -10, -11,  -9,  -6, -11,  14,  -5,   5,  -2,   3,  -1,   1,  -1,   2,  37,  43, -21,  -3,  -9,  -4,  -2,   1,  -1,  -1,   9,  16,  27, -28,  12,  -4,  -1,  -2,   0,  -2,  -1,  -1, -36, -11, -14,   4,   7,   3,   2,   3, -40, -11,  19,  -7,  -9,   7,   0,   1,  47, -19, -27,  25, -14,  -3,   1,  -1, -16,  26, -12,  -3,  12, -10,   2,  -3,   7, -10,   2,   1,  -4,   5,  -3,  -2,   4,  -2,   2,  -1,  -2,   3,  -1,   0 },
     {    7,   5,   6,  19,   9,  -3,  -7,  -4,   0,   0,  -1,  -1,  -2, -11,  -8, -19, -46,  31,  -7,   9,   1,   1,   0,   2,   9,   4,  15,  17, -21,   2,  -4,   2,  -1,  -2,  -1,   0,  10,   6,   8,   2, -27,  12,   0,   0,  -1,   0,   1,   0,  43,  12,  11,  19,   1, -19,   5,  -6,  21, -19,   7, -19,   1,   4,   0,   3, -27,  15,  23,  35, -22, -15,   1,  -2,   0, -13, -17,  19,  32,  -8,  -1,   0,  -2,  10,  -4, -14,   0,  -4,   0,   4,  -3,   2,  -1,  -3,  -1,  -2,   0,   0 },
     {   12,   8,  21,  31,  19, -19,  -4,  -9,   2,  -4,   0,  -1,   9,  -2,  13, -10, -14,   8,   1,   2,   1,  -1,   0,   0,  12,   1,  15,  10,  18, -36,  11,  -6,   2,  -3,   1,  -1,  24,  20,  22,   3,  44, -13,   1,  -5,  -2,  -2,  -1,  -1,  25, -14, -18, -16,  21,  -1,   8,   1,  18, -25,  21,   1,  21,  -5,  -7,   6, -20, -19, -18, -17,  13,  29,  -6,   6,   7,   4,   3, -14, -25,  10,   0,   2,  -7,   4,  -4,   7,   0,   3,  -2,  -3,  -6,   2,  -1,   2,   0,  -1,  -1,  -1 },
     {    6,   7,  16,  49,  35, -30,   4, -11,   0,  -5,   1,  -3,  -1,  -2,  12,   5, -23,  19,  -7,   3,  -3,   1,   0,   0,  -7, -14, -15, -34, -15,  11,   4,   5,   2,   2,   0,   2, -10, -32,  -5,  22,  -3,  -5,   2,   0,   2,  -1,   0,  -1, -28,  -6,  49,  -5,  11,   1,  -4,  -3, -12,  21,  -1,  -4,   3,   0,   4,   1,  15, -39,  -5,   9,  -5,   4,   0,   4,   1,  17, -14,  -1,   1,  -2,   2,  -1,   3,  -5,   2,   0,   2,   4,  -6,   0,   1,  -1,   0,  -1,   1,   1,  -2,   1 },
     {   -6, -13,  -4, -27, -29,  19,  -4,   5,   2,   4,   1,   1,   4,  -1,  19,  30,  11, -10,  -5,  -3,  -1,  -3,   0,  -1,  -1, -11,  18,   4, -38,  17,   1,   3,   0,  -3,  -1,   0,   4,  -3,  40,  10,  13, -24,   6,  -9,   4,  -2,   0,  -2,   9,  16,  31, -21,  22,  -9,   5,  -5,  -8,  -4,   6,  -4,  12,   0,   2,   3, -11, -37,  36, -10,   0,  12,  -5,   6,  11, -17, -18,  26, -14,   6,  -1,   2,   1,  10, -11,  -3,   4,   0,  -3,   6,  -1,   3,  -6,  -1,   0,  -1,  -2,   1 },
     {   -1,   9,  11,  13,  14,  -5,  -4,  -1,  -1,  -2,   0,   0,  -6,   7,  15,  -9, -27,  15,  -6,   5,  -1,   1,  -1,   1,  -9,   5,   6, -22,  -5,  -7,   6,   2,   3,   0,   1,   0,  -2,  22,  23,  -1,  35, -20,   0,  -5,  -1,  -2,   0,  -1, -12,  26, -16, -29,  -1,  24,  -4,   4, -34,  16, -31,   4, -31,   2,   4,   0,   7,  39,  22,  -9,   1, -17,   1,  -5,  -5, -38,  19,  15,   5,   2,  -2,   2,  11,   0,  -9,  -3,  -3,  -8,   8,   4,   5,   1,  -3,   0,  -2,  -3,   3,   2 },
     {   -4,  -7, -10, -39, -21,  39, -14,  13,  -3,   2,  -1,   1,  -8, -14, -21, -58, -40,  43,  -7,  11,  -2,   6,  -2,   4,  -6, -13, -23, -18,   7, -14,   8,  -2,   3,   0,   3,  -1,   0,  -5,   1,  27,   5,  -9,   4,  -6,   2,  -3,   1,  -1,  -4,  -5,   5,  -6, -10,   0,   5,   1,   5,   4,  -2,   9,   4,  -7,  -1,  -5,   8, -14,   3, -16,  -4,  12,   0,   1,   6,  20,   5,   5, -19,  -8,   3,  -2,  -5,  -1,   2,  14,   1,   2,   0,  -2,  -1,  -3,   3,   2,   1,   1,   0,   0 },
     {   -2,  -1,  -2,  -6,  15,  12,   0,  -4,   0,  -4,   1,  -1, -11, -15, -18, -21, -22, -37,  21,   0,   3,   1,  -1,   3,   2,   3,   3,  26,  54,  -8,  -3,  -9,  -2,  -4,   1,  -1,  -6, -10,   4,   6, -17, -18,  20,  -1,   5,  -3,   3,  -1,  -7,   6,  15,  -2,  48, -21, -15,   1, -24,   9, -14,   6, -14,  24,   5,  -1,  -6, -12,   5,   6,  19, -15,  -4,   1,  22, -10, -11, -12,  11,  22,  -7,   2,   1,   4,  -5,  -6,   3,  -2,  -1,  -3,  -4,   3,   2,  -4,   0,  -1,   2,  -1 },
     {    4,  -5,  -4,  -2, -10, -13,  13,  -7,   7,  -4,   1,  -1,   5,  -8,  -6,  -7,   3,  -1,   0,   3,  -5,   5,  -6,   5,   7,  -6,  -5, -12,  -2,   1,  21,  -4,   3,  -7,   2,   1,   6, -11, -16,  -1,  13, -34,  18,  -5,   5,  -7,   6,  -3,   6,  -7,  -9,  20,  -5,   7,  -2,  -3,  28, -32,   0,   3,  -1, -22,   8,  -6,  55,   1,  32,  17,  38,  -5, -22,   4, -25,  -8, -19,  -2,  16,  31,  -4,   5,  13,   4,  -8, -13, -15,   2,  -3,   1,  -1,   3,  -1,  -6,  -5,   2,   2,  -3 },
     {   -1,  -2,   6,   7,   7,   2,  -3,   0,  -4,   0,  -2,   1,  -3,  -6,   6,  -1,  -5,  12,   4,   4,   1,  -2,   0,  -1,  -3,  -6,   7,  15,  22,   3, -14,   3,   1,   2,  -1,  -2,  -1,  -8,  18,   3, -38,  38, -11,   1,  -2,   4,  -3,   3, -21, -35,  11, -34, -11, -19,  15,  -2, -15,   4,  38,   1,  20, -32,  -9,  -3,  16,  23,  25, -11,  24,   4, -10,  -4, -16, -24,  14,   9, -11,  21,   0,   5,   6,  13, -13,   4,  -5,  -5,   7,   7,   7,   6,  -7,   6,  -1,  -1,   0,   0 },
     {   -1,  -1,  -5,   4,  14,   5,  -4,   3,   1,  -2,   2,  -3,   0,   0,  -6,  15,  14,  -8,   7,  -2,  -1,  -5,   3,  -1,  -3,  -4, -16,  23,  23, -28,   3,   7,  -4,   0,   0,   0,   5,   7,  14,  51,  14,  38, -31,   3,  -6,   2,  -1,  -1,   0,   2,  -3,  -7, -31,  27,  10,   0, -13, -28, -25,  12, -23, -11,   9,  -2,   3, -29,  19,  12,  -4,  -5,   1,  -2,   2,   8, -34,  14,  13,  -6,   8,  -2,  -9,   9,   3, -13,   0,   0,  -3,  -2,  -4,   1,   3,  -7,  -2,  -1,   2,  -1 },
     {    1,   4,   2,  -1,   0,  -4,   4,   6,  -3,   7,  -2,  -2,   1,   4,   1,   2,  -2, -17,   9,  -2,   4,   3,   3,  -3,   2,   9,   8,  10,   6, -14,   2,   6,   0,   1,  -3,  -2,   8,  14,  22,  12, -13,   5,   0,  -3,  -1,  -5,   1,  -2,  10,  30,  13, -30,   9, -25,  -1,   2,  40,  28, -39,   6,  11,  -2, -24,   2,  20,  -6,   8,   5,  -8,   5,  -2,  -4, -56,  29,  18,  -5,   8,  -5,   2,  -7,  14, -24,  11,   5, -19,   5,  -1,  -3,  10,  -6,   0,   8,  -3,   1,  -2,  -2 },
   },
   {
     {  105, -47, -16,   1,  -3,   0,  -1,   0,  -1,   0,   0,   0, -32,  35,  -6,  -2,   0,  -1,   0,   0,   0,   0,   0,   0, -20,   4,  11,  -2,   1,   0,   0,   0,   0,   0,   0,   0,  -2,  -2,   2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   0,   2,   1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   16,  73, -41, -12,  -3,  -3,  -1,  -1,  -1,  -1,  -1,  -1, -53,   6,  51, -14,   1,  -2,   1,  -1,   1,  -1,   0,   0,  15, -46,  12,  19,  -3,   2,  -1,   1,   0,   1,   0,   1,   7,   0, -16,   5,   3,   0,   0,   0,   0,   0,   0,   0,   3,  -3,  -1,  -1,   1,   1,   0,   0,   2,   1,  -3,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0 },
     {   47,  46, -26, -21,   0,  -5,   0,  -2,  -1,  -1,   0,  -1,  69, -48,  14,   6, -10,   2,  -3,   1,  -1,   0,  -1,   0, -39,  26, -15,  13,   1,  -2,   1,   0,   0,   0,   0,   0, -22,   5,  12,  -8,   7,  -2,   0,   0,   0,   0,   0,   0,  -3,  -7,   4,   4,  -1,   2,  -1,   1,  -5,   1,  -1,   1,   1,   0,   1,   0,  -2,  -2,   0,   1,   1,   0,   0,   0,  -3,   1,   0,  -1,   1,   0,   0,   0,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 },
     {   17, -24,  31,  -9,   1,  -3,   0,   0,   0,   0,   0,   0, -41, -59,  28,  36,  -2,   5,   0,   2,   0,   1,   0,   1,  29,  -1, -64,  13,  10,   1,   2,   1,   1,   0,   0,   0,  -3,  32,  -9, -25,   3,   0,   1,  -1,   0,   0,   0,   0,   0,   4,   5,  -4,  -4,   0,   0,   0,  -1,   6,   1,  -4,   0,  -1,   0,   0,   0,   1,   2,  -1,  -2,   0,   0,   0,   0,   3,   0,  -1,  -1,  -1,   0,   0,   0,   1,   1,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0 },
     {   -2, -42,   9,  12,   4,   0,   3,   1,   1,   0,   1,   0,  40,   0,  36, -26, -11,   0,  -4,   0,  -2,   0,  -1,   0,   0, -30,  12,  55, -24,   0,  -3,   1,  -2,   1,  -1,   0, -12,  18, -55,  24,  25,  -9,   3,  -1,   1,  -1,   1,  -1,  -8,  12,  -2, -25,  15,   4,  -1,   0,  -1,   1,   0,  -4,  -3,   3,   1,   0,  -2,   3,   2,  -5,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,  -2,   0,   0,   0,   0,  -1,   1,   0,  -1,   0,   0,   0,   0 },
     {   -1, -17, -70,  46,   5,   5,   1,   3,   1,   1,   0,   1,  10, -18,  -8, -33,  23,   2,   3,   1,   2,   0,   1,   0,  52,   7,  -8, -24, -16,   8,  -3,   1,  -1,   1,  -1,   0, -28,  37,   9,  -7, -10,  -5,   3,  -2,   0,  -1,   0,  -1,  -3,  -7,  23,   0,  -4,  -5,  -1,   1,  -4,   0,   3,   8,  -2,  -1,  -2,   0,  -1,  -2,   4,   2,   1,  -1,  -1,   0,  -2,   1,   0,   2,   0,   0,   0,   0,   0,  -1,   1,   0,  -1,   0,   0,   1,   0,   0,   1,   0,  -1,   1,   0,   1 },
     {  -25, -16, -50,  38,  11,   2,   1,   2,   1,   1,   1,   1, -11,  -7,  -2,  12,   3,   2,  -1,   0,   0,   0,   0,   0, -65, -10, -14,  12,  31,  -3,   5,   0,   2,   0,   1,   0,  33, -11, -21, -42,  15,  13,   0,   2,   0,   1,   0,   0,   9,  12,  12, -20, -23,   7,   2,   0,   7,   1,   7,  -2,  -7,  -6,   1,   0,   3,   3,   5,   0,  -5,  -1,  -1,   1,   3,   1,   1,   1,  -2,  -2,   0,   0,   2,   1,   2,   0,   1,   0,   1,   0,   1,   1,   1,   0,   1,   0,   1,   0 },
     {   -2,  15,   5, -23,   2,   0,   0,  -1,   0,   0,   0,   0,  16,  65, -22,  -1, -19,  -2,  -4,  -1,  -2,  -1,  -1,  -1,  28,  -6, -31,  31,  16, -16,   3,  -3,   1,  -2,   1,  -1, -21,  17,   6, -44,  32,   4,  -3,   1,   0,   0,   0,   0,   0, -18,  39, -12, -20,  16,  -2,   0,  -2,  -9,   1,  17,  -8,  -3,   4,   0,  -1,  -4,   1,   3,   2,  -1,   0,   1,  -1,  -4,  -1,   2,   1,   1,   0,   0,  -1,  -2,   0,   0,   0,  -1,  -1,   1,   0,  -1,   1,   0,   0,  -1,   0,   1 },
     {  -22, -25, -17,  -6,  21,  -1,   7,   1,   2,   1,   1,   0, -25,   7,  51, -31, -29,  13,  -4,   3,  -1,   1,  -1,   1,  -4,  63,  -8,  21, -21, -20,   3,  -5,   0,  -2,   0,  -1,   0, -24,  33,   1,  19,  -9,  -4,   0,  -2,   0,  -1,   0,  16, -20, -10,  21,  -3,  10,  -5,   1,  -2,   6, -14,   1,   8,  -3,   4,  -2,   4,  -4,  -1,  -3,   2,   2,  -1,   1,   0,   0,  -3,  -1,   1,   1,   0,   0,   2,  -2,  -1,   0,   0,   1,  -2,   0,   1,  -1,   0,   0,   0,   0,  -1,   0 },
     {    5,  29,   4,  32, -22,  -8,  -4,  -1,  -2,   0,  -1,  -1, -30, -12, -35,  16,  15,   2,   1,   2,   1,   1,   1,   1,   3,  45,   3,   9, -18, -12,   0,  -4,   0,  -1,   0,  -1, -15, -11, -15,  22,  51, -19,  -1,  -2,   1,  -2,   0,  -1,   9,  -8,  -1, -48,  26,  21,  -6,   1,   4,  -6,  15,  -6, -20,  12,   3,  -1,   1,  -2,  -2,  -3,  -2,   0,   3,   1,   2,  -2,   5,   0,  -4,   1,   0,   1,   1,  -1,  -1,  -1,   1,  -1,   2,   0,   0,  -1,   0,   0,   0,   0,   1,   0 },
     {    5,   6,  17,  67, -49,  -6,  -5,   0,  -3,   0,  -2,  -1,  19,  24,  46,  17,   5, -28,  -5,  -5,  -2,  -3,  -1,  -2,   8,   2, -11,  -4,  40,   2, -10,  -1,  -1,  -1,   0,   0, -24, -15,   8,  -1,   1,  19,   0,  -1,   0,   0,   0,   0,   3, -19, -17,  14,   3,   4,   6,   0,  -3,   5, -17,  -9,  10,   1,   4,   2,   0,   0,   1, -10,   1,   3,   0,   1,  -1,   1,  -1,  -4,  -2,   3,   1,   0,   0,  -1,   0,  -2,  -1,   0,  -1,  -1,   0,   0,   0,  -1,   0,   0,  -1,  -1 },
     {   -3, -13, -47, -14,  27,   8,   0,   3,   1,   2,   1,   1,   8,   0, -17,  50, -36,  -3,   2,   0,   0,   0,   0,   0,  39, -10,   7,  11,  27, -24,  -3,  -2,  -1,  -1,   0,  -1, -20, -23,   5,  15,   8,  17,  -7,   1,  -1,   1,   0,   1,  -9,  16, -41,  -2,  16,  -1,   7,  -3,  -3,   5,  10, -36,   8,   6,   1,   1,   3,   0,   6,  -3, -12,   3,   0,   1,  -2,   2,   5,  -4,  -2,  -1,   0,   0,   0,   1,   1,   1,  -1,   1,   2,  -2,   0,   0,   1,   0,  -1,   0,   1,  -2 },
     {  -19,   0, -18,  -2,  -5,   8,   4,   2,   2,   0,   0,   0,  -9,  31,  -1,  28, -16, -11,  -1,   1,  -2,   0,  -1,   0, -38,  -1, -26,   5,  -8,  15,   4,   2,   3,   1,   1,   0,  -7,  45,   7,  11, -22, -39,  14,  -7,   3,  -3,   1,  -2,   2, -23,  11,  10,  55, -27, -13,   3,  11, -10,  -2, -23,  11,  24, -10,  -1,   4,  -3,   5,  -1,  -4,   4,   4,  -2,   4,  -3,   0,  -6,   0,   3,  -1,   1,   2,  -1,   2,   1,   2,  -2,   0,  -2,   1,  -1,   1,   1,   1,  -1,   0,  -1 },
     {  -11, -12,   8, -17,  23,   1,   2,  -1,   1,   0,   1,   0, -18, -10,  -4,  -2,  24,   2,   0,  -1,   1,   0,   0,   0, -28, -27,  18,   1,   4,  19,  -3,   2,  -1,   1,  -1,   1, -85,  -2,  11,  -6,  -1,  12,  10,  -2,   2,   0,   1,   0,  45, -31, -14, -11, -12,   3,   6,   3,  14,  11, -10,  -1, -14,  -4,   2,   3,  10,   5,  -2,   0,  -6,  -6,  -1,   1,   2,   7,   0,   1,  -3,  -4,  -1,   0,   4,   1,   0,   2,   2,   2,   0,   1,   2,   1,   0,   1,   1,   1,   0,   0 },
     {    9,   6,   4,  24,   0, -14,   2,  -2,  -1,  -1,  -1,  -1,   5,  -1, -38, -18,  18,  23,  -8,   4,  -1,   2,   0,   1,  -2, -47, -39,  28, -25,   3,  20,   0,   4,   0,   2,   0,   9, -21,  34,  -2,  11, -28,  -7,   6,  -2,   1,  -1,   0,   5,   5, -27,  24,   3,  12, -13,  -3, -10,  27, -14, -18,  20,  -7,   9,  -5,   3,  -2,  17, -19,   1,   8,  -5,   3,  -1,   5,   2,  -1,  -7,   1,   1,  -1,   0,   1,   4,  -2,  -1,   2,   0,  -1,   1,   0,   2,  -2,   0,   1,   1,   0 },
     {    4,   7,   1, -15,  18,  -3,  -1,  -2,   0,  -1,   0,   0,   8,  25,  11, -40,  25,   0,  -3,  -1,  -2,  -1,  -1,  -1,   5,  19, -61,  -9,  12,  32,  -3,   1,   0,   1,   0,   0,  -4, -42, -18,   8, -18,  10,  19,   0,   3,   0,   2,   0,  -6,  12, -15, -24,   9, -14,   6,   6,   0,  -6,  30, -28, -16,   9,  -9,   4,   2,  -3,   9,  16, -23,  -3,   0,  -1,  -2,   1,   2,   9,  -2,  -6,   0,  -1,   0,   0,   1,   6,  -1,   0,   1,   1,   0,   0,   0,   3,   0,  -1,   1,   1 },
     {    2,  17,   6,  21, -13,   6,  -5,   0,  -3,  -1,  -1,   0, -16,  -1, -26,  -6, -31,  19,  12,   2,   0,   2,   1,   1,  -2,  20,  -3,  28, -12,  -3,  -1,  -1,  -2,  -1,  -1,   0, -28, -11, -30,  -2, -17,  13,  25,  -1,   5,   0,   3,   0,  12,  18,   3,   8, -24, -59,  15,   5,   1,  -1,   7,  10,  46, -22, -23,   3,   2,  -5,   1, -14,  10,  12,  -7,  -4,   1,  -3,   5,  -2,   6,   2,  -1,  -2,   3,  -2,   2,  -5,   0,  -1,   2,   0,   1,  -1,   1,  -2,   0,  -1,   1,   0 },
     {    8,  19,  19,  39,  73, -57,  -8,  -7,  -2,  -3,  -1,  -3,   5,  19,  14,  30,  -6,  32, -25,  -4,  -4,  -1,  -1,   0,   3,   7,  15, -14, -13,  20,  11,  -6,   1,  -1,   0,  -1,   2,  18,  -5,  -7,  -1,   5,   6,   1,  -1,   1,   1,   0,  -9,  -1,  -1,   0,  -2,   0,   8,  -1,   2, -14,  -1,  -5,   3,  -6,   3,   4,  -2,  -1,  -3,   0,  -4,   0,   0,   0,   0,  -2,   0,  -3,   3,  -1,   0,   1,   0,  -1,  -1,  -1,   0,  -1,   0,  -1,   0,  -1,   0,   0,   0,  -1,   0,   0 },
     {   -3,   7,  14,  -1,   4,  -9,  -3,  -1,   0,  -1,   1,  -1, -15, -10, -16, -42,  15,   3,  11,   0,   4,   1,   2,   0, -16,   9,   8,  13,  23, -25, -16,   5,  -3,   2,  -1,   0,  -2,  40,   5,   2,  11,  32, -23,  -8,  -2,  -2,  -1,  -1, -41, -12,  -3,  14,   5,   6,  23,  -7,  21, -20,  -8, -38,  22,   4,  -2,  10,   7,  -4,  16,  -9, -34,  20,  -1,   1,   4,   2,  -4,  11, -11,  -8,   4,  -2,   1,   0,   2,   2,   3,   0,  -2,   3,   1,   0,   1,   1,   2,   0,   0,   2 },
     {    9,  -1,   2,  -7,  28, -13,   0,  -5,   0,  -1,   0,   0,   9,  -7,  -9,  -2,  33, -31,  -2,   5,   1,   0,   0,  -1,  20,  12,  20,  24,  18,   2, -25,   1,  -4,  -1,  -2,   0,  30,  -4, -10, -21,   6,   3,   8,  -1,   1,   1,   1,   0,  54, -10,   6,   5,  15, -19, -23,   8, -22,   6, -14,   5,  22,  37, -20,  -5, -12,  -1,   0, -15, -15,  26,  12,  -7,  -8,  -1,  -1,   3, -11,   3,   7,   2,  -2,   2,  -1,  -1,  -4,   1,   0,   1,  -1,   0,   1,   0,  -1,  -1,   0,   1 },
     {   -9,  -5, -17, -18,   5,   9,   5,   0,   2,   0,   1,   1,  -6,   5,  -3,  34,  51, -24, -18,   1,  -4,   0,  -2,  -1,   3,  20,   1,  39, -18,  37, -17,  -7,  -2,  -1,  -2,   0,   3, -14,   4,   8,  -2,  10,  19, -10,   0,  -1,   2,  -1, -49,  -1,  10,   4,  -8,   3,  18,   3,   5,  -4, -24,   8,   5, -22,  13,   3,   8,  -5,   1, -27,  20,  -1,  -6,   6,   5,  -2,   1,  -8, -11,   7,  -2,   2,   1,   0,  -1,  -3,   3,   0,   1,  -2,   1,   1,   0,  -1,   2,   0,   0,  -1 },
     {    1,  -7,   7,  12,  10,  -3,  -4,   3,  -2,   0,  -1,   1,  -5, -20,   6,  -3,  -7,   4,   9,   1,   2,   0,   1,   0,   3, -21,   5,   2,   7,  -7,   9,   0,   2,  -1,   0,   0, -22, -54,  18, -11,   1, -13,   5,  11,   2,   2,   1,   1, -48,   5,  43,  -4,  13, -14, -25,   9,  25, -25, -19,  15,   4,  35, -16,  -4,   8,  17, -29,  -1, -15,   9,  13,  -5,   3,   9,  -5,  -3,  -2,  -5,   4,   1,   1,   5,  -1,  -4,   2,   1,  -1,   1,   1,   2,  -1,   0,   1,   0,   0,   0 },
     {   -6,  -2,   3,  -1, -15, -10,   7,   3,  -1,  -1,   0,   0,  -3,   8,  19,   6,  27, -10, -15,   1,  -4,   1,  -2,   0,  -5,   3,  18,  21,  -5,  -1,   8,  -4,  -2,  -2,  -1,  -1, -20,  15,  44, -16,   1,  -7, -12,   7,   0,   1,  -1,   0,  11,  72,  -1,  -3,   0,  -4,  -4, -11,   0, -18,  49,  -1,  13,   6,   2,   1,  12, -25, -12,  20,  -3,  17,   0,   3,  -3,   0, -13,   4,   1,   2,   7,   1,   4,  -5,  -6,   0,  -3,   0,  -4,  -1,   1,  -3,  -3,  -1,  -1,  -1,   0,   0 },
     {    3,   8,   6,  23,  29, -15,  -3,  -7,   4,  -4,   1,   0,  -9,  -6, -17, -18, -17, -26,  16,   4,   2,   4,   1,   0,  -9,  -6, -23,  33,  13,  -9, -38,  11,  -1,   1,  -1,   2,   2,   5,  27,  34, -28,  24, -11, -16,   2,  -2,   1,  -1,   9,  15,   5,  -3,  10,   5,  -2, -14,   2,  -6,   4,  23, -18,   7,  19, -11,  -7,   2, -16,   5,  26, -30,  13,   2,   1,  -8,  12, -20,  17,   9,  -9,   5,  -1,   2,   0,   3,   0,  -2,   0,  -5,   0,   1,   0,   0,   1,  -1,   1,  -1 },
     {    5,   2,   8,  18, -17,  20,  -9,  -2,  -2,   1,   0,   0,   0, -10,  -8,   1, -13,  12,  -4,   0,  -2,   3,   0,   1,  11,   1,  29,  43, -20,  15,  -8,  -9,   1,  -2,  -1,  -1,  20,   2,  29, -39, -15,  13,  31, -11,  -1,   0,   1,  -1,  -8, -19, -15,   0,   6, -11,  12,  15,  -6,   5,   4, -22, -34,  10,   0,   5,  -8,   9,   4,  36, -24, -26,   6,  -1,  -2,   3, -11,   9,  23, -15,  -4,   1,   1,  -2,   0,  -3,   0,   0,  -1,   1,  -1,   0,   0,   0,  -1,   0,  -2,   0 },
     {    0, -10,  -5,  -3,  -4, -34,  14,   3,   6,  -2,   1,  -1,   7,  -2,   5,   4,   5, -11,   4,   0,   2,   1,   1,  -1,   1, -20,  -8,   0,  -5, -15,  -9,   4,  -1,   1,   1,   1,  19,  -2,  32,  14,  17,   2,   1, -12,  -2,  -5,  -3,  -2,  -6, -33,  -7, -28,   1, -20,  24,  21,  20,  14,  23,   8,  -3, -27, -53,  16, -11,   7, -12,  23,  22,  31, -18, -16,   3,   4,  -8, -10,   0,  10,   7,  -5,  -1,   2,  -1,   1,   0,   3,  -2,  -3,  -1,   1,   0,   0,   0,   2,   0,  -1 },
     {   -6,  -6, -12, -15, -33, -65,  48,   6,   4,   3,   3,   0,  -8,  -7, -15, -12, -35, -19, -23,  29,   0,   5,   0,   2,   2,   1,   5, -13,  -6,  30, -30, -10,   6,  -3,   2,   1,  -8,  -4, -14, -21,   5,   3,   7,  -7,  -2,   4,  -1,   0,  -7,   9, -10,   8,   2,   7, -14,  -5,  13,   0,  -6,   3,   4,   4,   8,  -9,   4,  12,   9,   1,  -9,  -6,  -2,   4,  -2,   2,   7,   7,   6,  -4,  -1,  -6,   0,  -1,   4,   2,   0,   0,  -1,   1,   2,   0,   1,   0,  -1,  -1,   2,   1 },
     {   -4, -10, -13, -10, -30, -12,  19,  -7,   4,   5,   4,  -1,   2,  -3,  -3,   3,   7,  62, -37,   3,  -2,  -2,  -3,   2,   3,  -5,  -3,  16,  33,   2,   1,  -8,   3,  -3,   0,   0,  10,   0,   9,  43, -13,   7,   7,  -8,  -2,  -2,   1,   0,  23, -13,  24,  -4,  -4,   3,  16,  -3, -26, -27,  -2,   1,   2,   2,   8,   9,  -9,   2, -32,  -8, -15,   3,   9,   0,  -4,  10,   0,  -5,  -2, -18,   4,  -2,   0,   7,  -2,  -5,   0,   0,   5,  -4,   0,   0,  -1,   0,   0,  -2,   3,  -1 },
     {    7,   2,  -4,   1,  -2,  20,   7,  -3,  -2,   1,  -1,  -1,  11,   5,   1,   6,   8,  -9,  -2,   0,   2,   1,   0,  -1,  10,   1, -13,   0, -26,   3,   7,   9,  -1,   2,  -1,  -2,  12,  -4,   0, -11,   6,   3, -17,   2,   4,  -1,   3,   0,  36,  12, -14,   5,   6,  -4,  17, -21,  63, -45, -21,   0,   6,  -9,   8,  19, -39,  35, -18,  -2, -13,  -2, -12,   5, -16,   8,  18,  -3,   4, -10,   0,  -6,  -4,  -2,   5,   1,  -1,   1,   4,  -1,  -2,   0,   1,   2,   0,  -1,   2,   0 },
     {    4,   2,   7,   7,  10,  30, -14, -12,   1,  -1,  -3,   0,   1,  -5,   4,  -5, -39, -14,  10,   9,  -6,   5,  -1,   1,  -5, -19,  -7, -29, -12,  35, -27, -13,  10,   1,   3,   0,  10,   4,  30,  12,  43,   6,  27, -29, -12,  -3,  -5,  -2,  13,  21,  16, -10,  -4,   6,  19,  16, -11, -11,  -1,  -2,   9,  -3,   5,   5,   7, -15,  -5, -21, -13,  14,   3,  14,  -4,   9,  -3,  14, -27,  -9,  -2,   0,   2,   4,  -7,   1,  -2,   1,   3,  -2,   0,   1,  -3,   2,  -2,  -2,   1,  -2 },
     {    5,   0,   9,  10,  16,   2,  22, -16,  14,  -3,   3,  -6,  -4, -15,  -5,   0,  -7, -39,  -3,  13,  -9,   8,  -3,   0,   2,  -6,   6,   4,   3, -10,  14,   3,  -4,   2,  -2,  -3,  -1, -24, -10,   4, -11, -19,  -9,   2,   8,   3,   0,  -1,   8, -21,  37,   5,  10,  -3,   9, -16, -31, -21,   8, -20,   5, -28,  20,  27,  16,  -2,   8,  17, -10,  -3, -49,  16, -15,   9, -20,  11,  13,  11,  -7, -10,  10,  -3,  -8,  -5,   2,   1,  -3,  -3,   2,  -1,   3,  -2,   0,   1,  -1,  -1 },
     {   -2,  -2,  -4,   2,  -1,   7, -14,   7,   8,   4,  -8,   2,   1,   4,  -2,  18,  -3,  -5, -12,  -5,   5,  -1,   3,  -2,  -6,  -5, -13,  -5, -42,  12, -23,  22,   5,   3,  -1,   3,  -1, -14, -15,   2,   1,  18, -45, -14,  18,  -2,   5,  -5,  -3, -27,   7,  11, -11,  -1,   8, -28, -10,  16,  28,   9,  12,  13,   3, -13,  17,  -1, -18,   9, -16,  17,  27,   0,  -9,   7, -24,   1,  11, -30,  16,  -4,   1,  -5,  22, -15,   1,  -5, -14,   0,   7,   3,   8,  -7,  -1,   5,   4,   1 },
   }
 },
 {  //31
   {
     {  108, -15, -19,  -5,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0, -55,  25,  12,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -5, -12,   3,   3,   1,   1,   0,   0,   0,   0,   0,   0,  -2,   3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -2,  -2,   2,   1,   0,   0,   0,   0,  -1,   1,  -1,   0,   1,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   23, -59,  10,   5,   3,   2,   1,   1,   1,   1,   0,   0,  66,  43, -36,  -8,  -6,  -2,  -3,  -1,  -1,  -1,  -1,   0, -47,  28,  30,  -5,   0,  -2,   0,  -1,   0,   0,   0,   0,   0, -21,   2,   9,   1,   2,   0,   1,   0,   0,   0,   0,  -6,   6,  -3,  -1,   1,   0,   0,   0,  -1,  -5,   1,   1,   0,   0,   0,   0,  -2,   2,  -1,   0,   0,   0,   0,   0,  -1,  -2,   1,   1,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0 },
     {  -36,  -3,  -8,  12,   1,   3,   1,   2,   0,   1,   0,   0, -18,  59,  15, -19,   0,  -5,  -1,  -3,   0,  -1,   0,  -1, -18, -71,  30,  24,   3,   4,   1,   2,   0,   1,   0,   1,  23,  -3, -48,  -2,   5,   2,   2,   1,   1,   1,   0,   0,   0,  12,  12, -15,  -4,  -2,  -1,  -1,   6,  -5,  -3,   5,  -2,   0,   0,   0,   0,   4,   1,  -3,   0,  -1,   0,   0,   3,  -2,  -1,   2,   0,   0,   0,   0,   0,   2,   0,  -1,   1,  -1,   0,   1,   0,   1,   0,  -1,   1,  -1,   0,   1 },
     {   19,  92,  -2, -13,  -9,  -6,  -4,  -3,  -2,  -2,  -1,  -1,  16, -11,  -1,   9,   2,   1,   0,   0,   0,   0,   0,   0, -61,  -2,  26,  -5,   6,   1,   2,   0,   1,   0,   1,   0,  23, -34,   3,  13,  -2,   2,  -1,   1,   0,   0,   0,   0,  -7,   5, -14,   0,   5,   0,   1,   0,   4,  -6,   4,  -4,  -1,   1,   0,   0,   0,   0,  -4,   2,   0,   0,   0,   0,   1,  -2,   1,  -1,   0,   0,   0,   0,   0,   0,  -2,   1,   1,  -1,   1,  -1,   0,   0,  -1,   0,   0,  -1,   0,   0 },
     {  -19,   1,  15,   2,  -2,   1,  -1,   0,   0,   0,   0,   0, -50,  23, -24,   1,   7,   1,   3,   0,   1,   0,   1,   0,  39,  21,  48, -23, -10,  -6,  -4,  -3,  -1,  -1,  -1,  -1, -17, -46,   7,  46,   5,   4,   0,   1,   0,   0,   0,   1,   7,   0, -38, -17,  13,   4,   4,   1,   1,   0,  13,  -8, -11,   0,  -1,   0,   3,   0,  -7,   5,   1,  -2,   0,   0,   0,   0,   3,  -2,  -1,   1,   0,   0,   2,  -1,  -3,   1,   0,   0,   1,  -1,   1,   0,  -2,   1,   0,   0,   1,   0 },
     {   23, -13,  29,  -1,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   0,  10, -59, -18,   9,  10,   4,   3,   3,   1,   1,   1,   1,   1, -11,  -1,  12,  -9,   4,  -3,   1,  -1,   1,   0,   1,  -9,   1, -62,  12,  17,   1,   5,   0,   2,   0,   1,   0,   1,  31,   1, -58,  -3,   2,   1,   0,   4,   1,  25,  12, -24,  -4,  -3,  -1,  -2,   8,  -5,   5,   8,  -6,   0,  -1,   0,   0,   7,  -1,   0,   2,  -2,   0,  -1,   3,  -2,   1,   0,   0,   3,   0,  -1,   2,  -1,   0,   0,   0,   2,   0 },
     {    6,  35, -32,  10,  -7,   0,  -2,   0,  -1,   0,   0,   0,  14,  19,  -4, -27,   2,  -1,   1,   0,   0,   0,   0,   0,  22,  32,  16,  16, -11,  -6,  -4,  -3,  -2,  -2,  -1,  -1, -44, -16,  -7, -23,  24,   4,   5,   1,   2,   1,   1,   0,  28,  -1,  36, -15, -38,   0,  -5,   1,  -9,   1,   3,  47,   9, -14,  -2,  -4,   2,  -7,  -4, -11,  19,   8,  -1,   1,  -3,  -3,   1,   3,  -5,   5,   2,   0,   0,  -2,  -2,  -2,  -1,  -1,   1,   2,   0,  -1,  -1,  -1,  -1,  -1,   0,   1 },
     {    7,  -4, -44,  -5,   7,   4,   3,   2,   2,   1,   1,   1,  44,  -1,  67,  -3, -14,  -6,  -5,  -2,  -3,  -1,  -1,  -1,  17,  -8,  -4,  27,   5,   3,  -2,   1,  -1,   0,   0,   0, -40, -23, -10,  20,   5,   5,   2,   2,   1,   1,   1,   1,   2,  -1, -43,  -8,  16,   1,   4,   1,  -5,  11,   8, -30,  -9,   4,  -1,   1,   2,  -1,   3,  11, -10,  -4,   0,  -1,  -4,   4,  -2,  -3,   5,  -3,   0,   0,   1,  -1,   2,   1,  -2,   2,  -1,  -1,   0,   0,   1,   0,  -1,   1,  -1,  -1 },
     {  -11,   9, -35, -10,  -1,   6,   1,   3,   1,   1,   0,   1,  -3,  34, -11,  15,   1,  -7,  -1,  -2,   0,  -1,   0,  -1, -16,  10, -28, -25,   6,   6,   3,   2,   1,   1,   1,   0,  -2,  27,  15,  -6,   3, -11,   1,  -4,   0,  -2,   0,  -1,  -7, -28,   9, -46,  11,  15,   1,   5,  18,   5,  27,   2, -57,  -6,  -2,   0,  -3,   1,   4,  35,  18, -20,  -5,  -4,   4,  -4,   3,  -8,   8,  10,  -3,   0,  -1,  -1,   1,   8,   1,  -2,   0,  -3,   0,  -1,   1,   4,   1,  -1,   0,  -1 },
     {  -25, -33, -26, -10,  11,   5,   6,   2,   3,   1,   1,   1, -43, -15,  25,  21, -10,   2,  -2,   1,  -1,   1,   0,   1, -45,  37, -21,   0,  17,  -3,   3,  -2,   1,  -1,   1,   0,  14, -50,  -5, -28,  -3,   7,   3,   4,   1,   1,   0,   1,  25,  33,   0,  -5, -18,  -9,  -4,  -2,   1,  -5,  16,   7,   5,   0,  -3,  -1,   5,   2,  -7,   3,  -4,   2,   1,   0,   2,   1,   3,  -1,   1,  -3,   0,   0,   3,   1,  -1,   1,   1,   0,   1,   0,   2,   0,   0,   0,   1,   0,   1,   0 },
     {    5,   7,  64, -13,  -1,  -4,  -2,  -3,  -1,  -2,  -1,  -1,   7,  28,   1,  22,  -1,  -3,  -2,  -2,  -2,  -1,  -1,   0, -11, -42, -35, -11,  15,   9,   6,   4,   3,   2,   1,   1, -48, -17,  18, -13, -12,   1,   2,   1,   1,   1,   1,   1,  49,   5, -13,   5, -11,  -5,  -3,  -1, -15,  14,  10,   9,  -2,  -6,  -1,  -2,   8,  -7,  -5,   8,  12,   0,  -2,  -1,  -4,   3,  -1,  -6,   0,   6,   1,   0,   3,  -1,  -1,   2,  -1,   1,  -1,  -1,   1,  -1,  -1,   1,  -1,   1,   0,  -1 },
     {    1, -22,  -1, -23,  -5,   4,   2,   2,   1,   1,   1,   1,  13, -18,  42,   5,   4,  -8,  -3,  -3,  -1,  -1,   0,  -1,   9, -17,  22, -21, -13,   6,  -1,   2,   0,   1,   0,   0,  34, -10,  18,  24, -13,  -6,  -7,   0,  -3,  -1,  -1,   0,  -9,  -3,  -6,  12, -20,  10,  -2,   3,   4,  21,  -7,  51,  -7, -28,   0,  -6,  -9,   1, -15,   2,  55,   4,  -6,  -2,   0,  -1,  -3, -14,  -9,  21,   4,   1,  -4,   3,  -6,   4,   0,   0,  -1,  -5,  -2,   1,  -2,   1,   0,   1,  -1,  -2 },
     {   -6, -10,  32,  28,   1,  -6,  -1,  -3,  -1,  -2,  -1,  -1, -24, -18,  13, -40,   5,   9,   3,   3,   1,   2,   1,   1, -43,  11,   1,  37, -15,  -6,  -1,   0,   1,   0,   1,   0,  -5,  -1,  10,  -8,  29,   1,  -1,  -2,   0,  -1,   0,   0,   0, -54, -27,  -4, -16,  16,   5,   6,  15,  35, -12,  -6, -12, -18,   0,  -2,   3,  -5,  14,   6,   6,  -1,  -5,  -1,   4,   4,  -9,   2,   3,   1,   0,  -1,   1,  -1,   3,  -2,   2,   2,  -3,   1,   0,   0,   1,  -1,   2,   1,  -2,   0 },
     {    9,  -2,  14, -11,   9,  -2,  -1,  -1,  -1,  -1,   0,   0,   6, -17,   9, -36,  -9,   6,   5,   2,   2,   1,   1,   1,  18,  -8,  28,  -7,   7,  -2,  -3,  -1,  -2,   0,  -1,   0,   8, -22,  10, -25, -23,  12,   2,   4,   0,   2,   0,   1,  14,  -6,  23,  13, -13,   1,  -8,   1,  -6, -12, -13,   8, -43,  20,   7,   4,   1,  23,  -7,  39, -26, -47,   5,  -5,  -2,   1,  -1,  11,  39, -10, -11,  -2,   0,   2,  -2,   0,  -1,   1,  -1,   4,   0,   1,  -1,   0,  -1,   1,  -1,   2 },
     {    6,  -3, -41,  63,   9,  -1,  -4,  -1,  -1,  -1,  -1,   0,   6, -21, -29, -42,   2,  10,   6,   2,   2,   1,   1,   1,  -4, -37, -15, -25,  -9,  -3,   4,   3,   3,   1,   1,   1,  16,  -3,  31,   6,  -9,  -5,  -6,  -1,  -3,   0,  -1,   0,  28,  30, -21,  -1,  -2,  -1,   0,  -1, -17,  11,  27,  -1,   1,  -4,  -3,  -2,   4,  -5,  -4,   8,   2,   5,   0,   1,  -5,   6,   2,  -4,  -1,  -2,   1,   0,   2,  -1,   2,   3,  -3,   3,   0,  -1,   0,  -1,   1,   1,  -1,   2,   0,   0 },
     {   -4,   0,  17,  24,  -3,   1,  -1,   1,   0,  -1,  -1,   0,  -8,   4,  45, -17,  11,  -5,  -1,  -3,  -1,   0,   0,  -1, -29,  21, -15, -16, -30,   7,   1,   4,   1,   2,   1,   1, -11,  16,   4,  43,  10, -24,  -5,  -6,   0,  -3,  -1,  -1,  23,  16,  24,  -5,  29,  16,  -6,  -3, -25, -44, -22,   7,  -9,  12,  15,   6,  16,  11,  -9, -13,   1, -13,  -5,   1,  -3,  -7,   3,   4,   1,   7,  -3,  -2,   3,   2,  -5,  -3,   0,  -2,   1,   0,   2,   0,  -2,   0,   0,  -1,   0,   0 },
     {   12,  19,  41,  41,  22,  -8,  -7,  -6,  -4,  -3,  -2,  -2,   6,  30,  25, -21, -31,   4,   0,   1,   0,   1,   0,   0,  28,  26, -42,   3,  11,  -8,   0,  -2,   0,  -1,   0,   0,  27, -22,  -5,  -1,  -3,   7,   0,   3,   0,   1,   0,   1, -43,  22,   1,  -7,  -3,  -8,   3,  -1,  22, -10,  18,   6,   1,  -5,  -4,  -1, -14,  -1,  -4,  10,  12,   2,  -1,   0,   4,  -6,  -1,  -8,   3,   7,   1,   2,  -5,  -1,  -1,   3,   1,  -2,   0,  -2,  -2,   0,   0,   1,   0,  -1,   0,  -1 },
     {   -1,   5,  -9,  15,   9,   6,  -5,  -1,  -2,  -1,  -1,   0, -11,  -9, -25,   5, -26, -13,   5,   3,   4,   1,   1,   0,  -3,  11,  10,  29,  19,   2,  -9,  -5,  -3,  -1,  -2,  -1, -12,   5, -19,   0, -26, -11,  10,   1,   3,   1,   1,   0,   8,  13,  -3,  26,  24, -14,  -9,  -5,  -9,   6, -14, -13,   0, -15,  14,   0,  13,   7,  27,   3,  46, -29, -27,  -1,   0, -11,   1, -13,  19,  55,  -7,  -2,  -2,   0,  -2,  -3,   1,  -2,   1,  -3,   0,   1,   0,  -2,   0,  -2,   0,  -1 },
     {    3,  -3, -12,  44, -24,   6,  -2,   1,   0,   0,  -1,   0,   8,  -3,  -2,  27,  39,  -3,  -2,  -1,  -2,  -2,   0,  -1,   8, -11, -20, -37,  -6,  20,   8,   2,   3,   2,   1,   1, -12, -55, -21, -14,  -2, -14,   4,   5,   2,   3,   1,   2, -21, -14,  15,   7,  -4,   8,  -9,   1,  21,  15, -38,  -9,  -6,  -4,   8,  -2,  -3,  17,  25, -11,  -4,  -9,  -6,   1,   4,  -3,  -2,  10,   0,   3,   0,  -2,  -3,   4,   3,  -6,   2,   0,   0,   3,  -1,   3,   2,  -2,   1,   0,   0,   1 },
     {   -4,  -3,  -2,  16,  19,   8,  -2,  -1,  -2,  -2,  -1,   0, -10, -18,   8, -16, -16, -12,   9,   3,   4,   2,   2,   0, -24,  -4,  17, -21,  14,  14,  -7,   1,  -3,   1,  -1,   1, -47,   9,  -4,  17, -34,  -3,  15,   3,   4,   1,   1,   0,  -6, -15,  20,   1,  17, -29, -13,  -4,  49,  13,   5,  19,  11,  28,  -7,  -6, -30, -16, -21,  -7,  -2,   4,  18,   6,  15,   5,   0, -10,  -8,  -9,  -4,   3,  -5,   1,   0,   7,   6,   1,  -2,  -5,  -2,   1,   0,   3,   4,   1,  -1,  -1 },
     {    4,  -2,  -7,  17,  49,  12,  -6,  -2,  -4,  -2,  -1,  -1,   0, -13, -12,  30, -45,   0,   9,   1,   1,   1,   1,   1,   1, -26,   1,  -3,  32, -16,  -6,   0,   1,   0,   0,   0,  -4, -11,  20,   9,  28,  17,  -1,  -4,  -2,  -1,  -1,   0,  -3, -29,   0, -24,  -8,  16,   6,   8, -10, -25, -34,  14,  -7,   0,   0,   3,   9,  28,  -9, -29,  -1,   1,  -2,  -1,  -1,   0,  11,   2, -17,   1,   1,  -1,   1,   7,  -5,   2,  -1,  -1,   3,  -1,   1,   3,  -3,   1,  -1,   0,   2,  -1 },
     {   -6,   4, -13,  15, -27,   3,  -2,   2,   0,   0,   0,   0,  -6,  12, -14,  31, -28,  22,  -1,   1,  -1,   0,  -1,   0, -10,   3, -19,  19, -31,  -4,   1,   4,   2,   1,   0,   0, -10,  13,  -7,  31, -10,  11,   0,  -6,  -1,  -1,   0,  -1, -12,  10, -17,  17, -27,  -8,  15,  -2,  -5,  14, -18,  23,  -2, -12,   9,  -8,  -9,   7, -19,  -4,   1, -37,  27,  -2,   7,   1,  16, -10,  34, -39, -16,   3,   4,  -2,   7,  -7,   2,  -1,   5,  -3,   1,  -1,   3,  -3,   1,   0,   3,  -1 },
     {    5,  -1,   9,  50, -28,   7,  -3,  -2,  -2,  -1,  -2,   0,   7,   1,  12,  38,   5, -19,  -3,   0,  -1,  -1,   1,  -1,  12,  14,  20,  22,  23,  14,  -4,  -4,  -2,  -2,  -2,  -1,  34,   3,  -6, -20,  -1, -13,  12,   5,   1,   0,   2,   0,  19, -34,  -7,  -4,  17,   0, -16,   2, -34,   5,  18,  12,  -4,  27,   3,  -3,  -3, -19, -35,   0,   0,  -5,  12,   4,  -5,  13,   2, -14,  -3,  -6,  -4,   0,   3,  -3,  -1,   3,  -3,   5,   1,  -4,   0,  -2,  -1,   1,  -2,   2,   0,  -3 },
     {    5,  10,   5,   1,  51,  10, -13,  -6,  -2,  -2,  -2,  -2,   4,   7,  11,  20, -31,   1,   9,  -2,  -3,  -1,   1,   0,   0,  14,  12, -46, -28,  -3,   3,   7,   4,   0,   0,   1,  19,  11, -42,   6,  12,  -9,  -4,  -3,   0,   0,   2,   0,  24,  -9,  14,   8, -11,   3,  -1,  -1, -20,  47,   2, -12,   7,   0,   4,  -2,   7, -25,  22,  11, -10,   0,   3,   0,  -8,   6, -14,   8,   8,  -8,   1,   0,  -3,  -7,   8,  -6,   1,   0,  -5,   3,  -1,  -2,   2,  -1,   0,   0,  -2,   1 },
     {    3,   0,  -5, -34,   3,  -1,   4,  -1,   0,   0,   2,   0,   4,   3, -25, -13, -19,   6,   3,   2,   0,   0,  -1,   1,  14,  -8, -31,  34, -19,  -3,   4,  -1,   2,   0,   2,   0,  26, -38,   9,  12,  18, -35,   1,  -1,   2,   0,   0,   0,  29,   0,  19,  -1,  34,  17, -18,  -1,  16,  26, -15,   2,  -6,  24,  14,  -1, -31, -14,  -7, -17,   2,  -9,   6,   8,  11,   4,   6,  -6, -12,  -2,  -7,   0,  -6,   3,   5,   3,   2,   1,   1,  -1,  -1,   2,   1,   2,   0,   0,   0,   0 },
     {    0,   6,  -1, -24,  11,  18,   5,  -4,   0,  -1,  -1,   1,  -3,   7, -10, -41,   6, -27,   1,   9,   5,   2,   1,   0,   1,  14, -25,  -7,  23,  52,  -4,  -3,  -4,   1,   0,   0,  13,   2, -14,  12, -27,  -1,  19,   6,   1,   0,   2,   1, -15,  -8, -29, -17, -19,  -1,  -9,   4, -30,  15,  -7,  13,  -3,  -6,  11,  -6,  18,  -3,   4, -37, -12, -12,   4,   6,  -8,   3,  17,  22,  -8, -10,  -8,  -4,   5,  -2,  -2,   0,  -1,   2,   4,   1,   0,  -3,   1,   2,  -1,   0,   1,   0 },
     {    2,   5,   2,   8,   4,  33,  15,  -5,  -4,  -3,  -2,  -1,  -7,  -7, -15,  -1, -23, -16, -30,   5,   8,   2,   1,   2,   2,   5,   5,  19,   3,  43,  20,  -7,  -3,  -2,  -3,   0,  -7,  -3,  -3,   7, -14, -17, -18,   9,   6,   0,   0,   2,   4,   7,  12,  11,  -6,  30,  15, -19,   3, -11, -10,  -6, -14, -19,   1,  13,  -5,   6,  -9,  29,  -1,  27,   1, -15, -11,  -8, -27, -14, -10, -29,  41,   9,   7,   7,  15,  -5,  -1,  -3,  -4,   1,   2,   0,   4,  -4,  -1,   0,   0,   0 },
     {    1,  -1,   4,   7, -33,   6, -12,  -1,   0,   1,   1,   1,   1,  -1,   6,   3,  -6, -36,  11,  -3,   1,  -1,   1,  -1,   3,   6,   0,   0,  29,   5, -22,   0,   0,  -2,   1,   0,  16,  11,  23,  27,  12,  27,  13, -10,  -6,  -3,  -1,  -2,  30,  26,   6, -24, -11,  -8,   7,  15,   1,   7, -28,  -2,   0,  -3, -13,   0, -42, -10,  41,   8, -10,   3,   3,  -4,  16,  -8, -31,  10,   8,  -3,  -2,   1,  -8,   6,  10, -10,   3,  -5,  -7,   5,  -5,   1,   4,  -3,   3,  -2,  -2,   3 },
     {   -1,   4,   4,   9,  20,   0,   6,   4,   1,   1,  -1,  -1,  -3,   1,   5,   9,  14,  -9, -11,   0,   0,   2,   0,   1,  -3,   0,  21,  -3,  15,  -3,  -2,  -6,   0,  -2,  -1,  -1, -19,  17,  14, -21,  17,   9, -11,  10,  -3,   0,  -2,   2, -20,  48,  -9,   9, -10,  47,  -7, -14,   3,  10,  -4,  -4, -20,  -5,  45,   4, -22, -39, -11, -17,  -7, -22, -12,  13,  24,  16,   3,   1, -10,  -1, -14, -13,  -2,  -5,   6,   7,   5,   2,  -3,   0,  -3,  -1,   2,   1,   4,   0,  -4,   0 },
     {    2,   1,   6,  -3,   0,  55,  12, -16,  -1,  -1,   1,  -1,  -2,  -4,  -5,   3,  -3, -60,  -5,  26,   3,   3,   0,   4,  -2, -15,  -5,   1, -40,   2,  -7, -12,   3,   1,   3,  -1,  -1,  -7,  17,  -4,  42,  -1,   8,   0,  -5,  -2,  -2,  -1, -16,   4,   5,  13, -18, -26,  -6,  11,   9, -13,  14, -16,  12,  10,  -5,  -8,   9,  -2, -11,  16,   2,  -9,  -6,   5,   2,  12,  16,   0,  12,  11, -11,  -5,   2,  -4,  -7,   6,  -3,   1,  -3,  -1,   1,  -3,   0,  -1,   0,   1,  -1,   2 },
     {    5,   8,  -4,   4,  63,  12,  16, -22,  -3,  -4,  -1,  -3,   5,  10,  -4,  14,  63,   1,  -8,  10,   0,   1,  -1,   1,   1,   7,  -6,  22, -26,  -3, -20,   1,   3,   1,  -1,  -1,   8,   3,  -3,  -9, -25,  20,   5,  -9,  -4,  -1,   3,   0,  22,  -5, -19,  -2,   8,  -8,  13,   6,  -1, -12, -11,  10,  -1,   2,  -1,  -2, -35,  13,   3,   3,   6,  -2,   9,  -3,   7,  -8,  -6, -11,   3,  -3,   8,   4,  -6,   3,  -5, -11,   6,   0,   7,  -2,  -3,  -3,  -3,   3,   2,   0,   1,  -4 },
     {    2,   2,   8,   6,   2,  15,  16,  -8,   1,  -4,   1,  -3,   2,   2,  12,  21, -18,  14,   5,   2,   4,   0,   1,   2,   3,  -3,  23,  12, -21,  -2,  12,  -9,  -1,   1,   1,  -3,  -1,  11,  32, -20, -22, -54,   4,  21,   3,   2,   0,   3,  -3,  10, -23, -39,  -3,   2, -17, -14,  11,  -4,  11,  22,  24,  -2,   0,  -5,  -5,  18,  38,  -7, -18,  -3,  -1,  10,  -1, -13,  -7,  31,  13,  -2,  -8,  -5,  -1,   4,  -6,  -1,  -1,  -3,  -2,   4,  -3,   0,  -2,  -4,   2,   0,  -1,   0 },
   },
   {
     {   96, -61,   3,  -2,   0,  -1,   0,   0,   0,   0,   0,   0,  21,   0, -15,   6,  -3,   1,  -1,   1,  -1,   0,   0,   0, -29,  33, -10,  -1,   1,  -1,   0,   0,   0,   0,   0,   0, -17,  10,   8,  -6,   2,  -1,   1,   0,   0,   0,   0,   0,  -7,   0,   6,   0,  -1,   0,   0,   0,  -4,   0,   1,   1,   0,   0,   0,   0,  -3,   0,   1,   0,   1,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   38,  17, -17,  -4,  -2,  -1,  -1,  -1,  -1,   0,   0,   0, -77,  68,  -5,  -5,   1,  -1,   1,  -1,   0,   0,   0,   0, -15, -12,  29, -10,   4,  -2,   1,  -1,   1,   0,   0,   0,  24, -34,  14,   4,  -4,   2,  -1,   1,  -1,   0,   0,   0,   8,  -8,  -6,   9,  -3,   1,   0,   0,   4,  -2,  -5,   2,   2,  -1,   0,   0,   2,  -1,  -2,   0,   1,   0,   0,   0,   3,  -2,  -1,   0,   0,   1,   0,   0,   1,  -1,  -1,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0 },
     {   -9,  48, -52,  14,  -5,   3,  -2,   1,  -1,   1,   0,   0,  54,  25, -43,  -5,   0,  -2,   0,  -1,   0,   0,   0,   0, -13,  33,  12, -27,   6,  -4,   2,  -2,   1,  -1,   0,  -1, -16,  -5,  41, -16,  -2,   0,  -1,   0,   0,   0,   0,   0,   0, -19,  20,   9,  -8,   3,  -2,   1,  -2,  -8,   1,  11,  -2,  -1,   1,   0,  -1,  -5,   0,   4,   3,  -1,   0,   0,  -1,  -4,   1,   1,   1,   1,   0,   0,  -1,  -2,   0,   1,  -1,  -2,   0,   1,   0,  -1,   0,   1,   0,  -1,   0,   0 },
     {   53,  31, -50,  11,  -9,   2,  -3,   1,  -2,   0,  -1,   0,  16, -15,  14, -11,   5,  -3,   1,  -1,   1,  -1,   0,   0,  45, -57,  26,  -3,  -2,   3,  -2,   2,  -1,   1,  -1,   0, -16,  23, -22,  15,  -7,   3,  -1,   1,   0,   0,   0,   0, -26,  23,  -6,  -3,   5,  -4,   3,  -2,  -6,   0,   8,  -6,   3,  -1,   0,   0,  -3,  -1,   4,   0,  -1,   1,  -1,   0,  -3,   1,   0,   1,   0,   0,   0,   0,  -2,   1,   1,   0,  -1,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   -5,  -6, -32,  31,  -4,   4,  -1,   2,   0,   1,   0,   1, -45,   0, -39,  35,   5,   2,   1,   1,   1,   1,   0,   1,  16,  -5, -46,   3,  18,  -4,   4,  -1,   2,   0,   1,   0,  -3,  28, -17, -36,  21,  -4,   3,  -2,   1,  -1,   1,  -1,   3,   1,  27, -40,   3,   6,  -2,   1,   8,  -8,  22,  -8, -18,   8,  -2,   1,   2,  -2,   7,   6, -12,  -1,   2,   0,   2,   0,   4,   2,  -2,  -5,   1,   0,   1,  -1,   3,   1,   1,   0,   2,   1,   0,   0,   1,   1,   1,   0,   1,   0 },
     {   19,  51,  -7, -16,  -4,  -3,  -1,  -2,  -1,  -1,  -1,  -1, -29, -24,  56, -19,   5,  -1,   1,   0,   0,   0,   0,   0, -35,  25, -16,  16, -10,   4,  -2,   1,  -1,   1,   0,   0, -37,  34,  -4,  -1,   5,  -6,   3,  -3,   2,  -1,   1,  -1,  20, -34,  24,  -6,   3,   0,  -1,   1,  14, -13,  -4,  12,  -6,   3,  -2,   1,   4,  -2,  -8,   5,   2,  -2,   1,  -1,   3,   0,  -3,   0,   2,   0,  -1,   0,   2,  -2,  -1,   0,   1,   0,  -1,   0,   1,  -1,  -1,   0,   1,   0,  -1,   0 },
     {  -18, -13, -25,  52, -18,   9,  -5,   4,  -2,   2,  -1,   1,   1, -17,   5,  21,   1,  -5,   1,  -1,   0,   0,   0,   0, -72,  -8,  37,   2,  16,  -6,   3,  -1,   1,   0,   1,   0, -22, -12, -21,  22,   8,   2,   0,   1,   1,   0,   1,   0,   5,   3, -29,  -7,  18,  -3,   3,  -1,  17,  -6,  -1, -25,   9,   4,  -2,   1,   9,   2,   1, -12,  -8,   8,  -1,   1,   5,   3,   0,  -3,  -7,   0,   1,  -1,   3,   2,   1,  -2,   2,   1,   1,  -1,   2,   1,   0,  -1,   2,   0,   0,  -1 },
     {    2, -22, -16,  27, -10,   8,  -2,   3,  -1,   1,  -1,   1, -26, -63,  17,   7,  10,   2,   3,   2,   2,   1,   1,   1,  11, -13,  11, -37,  15,  -2,   2,   0,   1,   0,   0,   0,   6, -11,  40, -29, -11,   9,  -4,   3,  -2,   1,  -1,   1,  18, -14,   3,  24, -31,  10,  -4,   2,  -6,  22, -26,  25,  -8,  -8,   5,  -3,  -4,  12, -10,   0,  10, -11,   3,  -1,  -1,   6,  -1,  -4,   5,  -1,  -2,   1,   0,   2,  -1,  -1,  -1,   3,  -1,   0,   0,   1,   0,   0,   0,   2,   0,   0 },
     {   27,  41, -13,   8,  -9,  -4,  -2,  -2,  -1,  -1,  -1,  -1,   9, -25,  -1,  40, -16,   2,  -4,   2,  -1,   1,  -1,   0,   6,  11, -40,  42,  -8,  -3,  -1,  -1,   0,   0,   0,   0,  35, -40,   9,   7,  18, -11,   2,  -2,   1,  -1,   0,   0,  13, -13, -21,   7,  17,  -1,  -2,   1, -21,  27, -18, -15,  18,   1,   1,  -2,  -9,   4,   9, -22,   6,   6,   0,   0,  -3,   0,   8,  -9,  -4,   6,   0,   1,  -2,  -1,   4,  -3,  -2,   1,   2,  -2,  -2,   0,   1,  -1,  -1,   1,   1,  -1 },
     {    8,   3, -19, -24,  30,  -7,   5,  -2,   2,  -1,   1,  -1,  27, -14,  -1, -33,  28,  -1,   2,   0,   0,   0,   0,   0, -40, -28,  -6,   4,   6,  16,  -3,   5,   0,   2,   0,   1,  38, -38, -24, -16,   3,  13,   0,   2,   1,   1,   0,   1,  17,  -4,   9, -34, -12,  13,  -1,   2,  -3,  12,   7,   0, -35,  10,   1,   1, -10,  17,   0,  15, -21, -10,   7,  -3,  -2,   5,   2,   8,   0, -14,   2,   0,  -1,   3,   3,   3,  -1,   2,   1,   3,  -2,   2,   1,   2,   0,   1,   0,   2 },
     {   -3, -18,   5,  29, -33,  11,  -4,   3,  -2,   1,  -1,   1,  30,  40,  25,   7, -36,   0,  -5,  -1,  -3,  -1,  -2,   0,  -7, -35,  11,  32,  -6,  -3,   0,   0,   0,   0,   0,   0,  22,  11, -18, -17,  12, -12,   2,  -2,   1,  -1,   0,  -1,  15, -17,  42,  -5,  -6,  -1,  -7,   1,  -1,   0, -19,  36,  -4,  -3,  -1,  -2, -12,  13, -11,   3,  22, -10,   2,  -3,  -2,  -2,  -3,  -7,  12,   5,  -3,   1,  -4,   2,  -1,  -5,  -1,  -1,  -1,  -2,  -1,   1,   0,  -2,   0,  -1,  -1,  -1 },
     {   -3,  -5, -22,  22, -23,  17,  -5,   4,  -2,   2,  -1,   1,   0,  11,  14, -39,  13,  -8,   7,  -3,   1,  -1,   1,  -1,   1,  44, -25, -25,   7,  -5,   2,   0,   0,   0,   0,   0,  49,  29, -45,  12, -23,   6,  -3,   0,  -1,   0,  -1,   0,  26,  13, -11,  14, -10,  -8,   3,  -4,  -7,  15,   6,  -3,  14, -14,   3,  -1,  -7,  -4,  21,  -8,   9,  -1,  -5,   2,  -2,  -8,   9,   4,  -2,   6,  -4,   0,  -3,  -5,   4,   4,  -2,  -2,   2,   1,  -2,  -2,   2,   1,   0,  -2,   0,   1 },
     {   19,  29,  47,   2, -42,  10, -10,   1,  -4,   0,  -2,   0,  13,  -2,  26,   6, -14,  -6,   2,  -3,   0,  -1,   0,   0,  12,  19,  18, -35,  37, -30,   9,  -6,   2,  -2,   1,  -1,  16, -10,   1, -16,  30,  -1,  -8,   5,  -3,   2,  -1,   1,  -4,  12, -13, -24,   7,  20,  -9,   3,  -3,  -6,  17,  -9, -21,  19,  -3,  -1,  -3,  -5,  -4,  19, -19,   3,   6,  -3,  -4,   0,  -5,   9,  -2,  -7,   5,   0,  -2,  -2,  -3,   5,  -1,  -2,  -1,   2,  -1,  -1,  -2,   2,  -1,   0,  -1,   1 },
     {   -4,  20,   1,  -9,   0,  -6,   3,  -1,   0,   0,   0,   0,  -6,  -4,  14,  10, -17,   8,  -3,   1,  -1,   0,   0,   0, -51, -11,  -9,   6,   0,   5,   1,   0,   3,   0,   1,   0,  18,  14,  13, -43,  -2,  -1,   0,  -1,   0,   0,  -1,   0, -49,  60,   4,  11, -23,  -1,   2,  -3, -15,  12,   4,   7,   7, -16,   9,  -5,  19, -28,  20,  -4,  10,  -5,  -4,   3,  11,  -9,  -4,   7,   2,   6,  -5,   1,   3,  -1,  -4,   2,   4,  -3,  -1,   0,   2,  -2,   0,   0,   1,  -1,  -1,   0 },
     {    2,   9,  -1,  -2, -31,  26,  -5,   4,  -3,   2,  -1,   1, -13, -23, -24,   6, -37,  38,  -2,   6,  -1,   2,   0,   2,  -3,  20,  11,  -6, -47,  20,   0,   1,  -1,   0,   0,   0,   2, -21,  -8,  23, -24,   0,  14,  -2,   3,  -1,   1,  -1,   0,  16,  -4, -23,  -6, -17,  14,  -3,   5,  -1,  25,  12, -22, -15,   1,   0,  -1,   9, -14,  27,   6, -25,   4,  -2,  -1,   2,   2,   0,  24, -16,  -5,   1,   0,   0,   0,  -1,   0,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -1 },
     {  -13, -31, -33,  29,   4,  -2,   4,   2,   2,   0,   1,   1,  -8,   1,  45, -25,   1,  -6,   1,  -1,   0,  -1,   0,  -1,  18,  32,  10,  14, -28,  12,  -9,   3,  -3,   0,  -1,  -1,  14, -23,  34,  -2,  10, -11,   3,  -3,   1,  -1,   0,   0, -29,   5,  25, -19,  29,  -9,  -1,   1, -26,   6,  13, -13,   1,  17,  -8,   3,   4, -13,   0,  14, -19,  14,   0,  -1,   2,   2, -17,  12,  -3,  -3,   7,  -2,   3,  -1,  -5,  -1,   2,   1,  -4,   0,   1,   0,  -1,   0,   0,   1,  -2,   0 },
     {    1,  -8, -11, -24,  -7,  25,  -2,   2,  -1,   1,   0,   0,   9,  -7,   5,  -2, -30,  31,  -6,   4,  -2,   1,   0,   0, -11, -28, -17, -17,   1,   5,   8,   2,   2,   0,   1,   1,  35,  20,  22, -19, -20, -11,   3,   0,  -2,  -1,  -1,   0,   8, -18,   1,  29,  21, -17,   4,   0,  33, -32,   3, -33,  20,  11,  -9,   7,  -1,   2,   1,   7, -30,  19,  -4,   1, -13,  18, -11,   9,  -8, -10,   9,  -4,  -6,   5,   4,   1,  -2,   2,   0,   1,  -2,   1,   1,   2,  -1,   2,  -1,   1 },
     {    7,  11,  19,  11,  -3,  -9,  -1,  -1,  -1,   0,  -1,  -1,  -3, -25, -13,  21,  31, -20,   4,  -3,   2,   0,   0,   0, -12,  -4,  -1,   8,  18, -17,  -5,   1,   0,   0,   0,   0,  35, -10,   1,  39, -25,   7, -14,   1,  -2,   1,   0,   0, -23,  -9,  54,  13,  -4, -10,  -1,  -6,  -2, -30,  41,  11,  17,  -6,  -6,   0,  -8,   1, -13,  13,  11,  13,  -8,   2,   0,   6, -15,  -1,   6,  12,   0,  -2,   0,   6,  -8,  -7,   1,   2,  -2,  -6,   0,   2,  -2,  -2,   1,   0,  -1,  -2 },
     {   -9, -10,  -5, -11, -28,  44, -10,   6,  -3,   3,  -1,   2,  -5,  10,   2, -28,  -1,  18,   4,  -2,   0,  -1,   0,  -1,  -5, -17, -45,   7,  30, -13,  16,  -2,   3,  -1,   2,   0, -37, -33,  -9,  20,   1,   5,  -1,   5,   1,   1,   1,   1, -43, -27,   9,  20, -10,   9,   1,   2, -34,   6,   0,  -8,  -5,  -4,  10,  -1,  -2,  -1,   6,  -2,  -8, -10,   4,   1,  12,  -1,  -2,   0,   0, -10,  -1,   1,   7,   2,   1,  -2,   4,   3,  -1,   0,   3,   1,   1,  -1,   2,   1,   0,   0 },
     {    3,  -5,  -5, -20,  -3,  17,  -1,   1,   0,   1,  -1,   1,  -6, -27, -21, -35,  -1,  18,   8,   2,   3,   1,   1,   1,   8,   1,  20,   3,  -3, -21,  15,  -7,   2,  -2,   0,  -2,  -4, -32, -19, -14,  41, -29,  12,   2,   0,   3,   0,   1,  17,  13,  18, -11,  -6, -16, -14,   6,  10, -13,   6,  27,  28, -10, -14,   1,  19, -17,   8, -26,  15,  25, -16,   4,  -3,   6,   0,   7, -21,  22,  -3,  -2,  -5,  10,  -6,   1,   0,   1,   2,   3,   0,   3,  -2,   1,  -1,   2,   1,   2 },
     {    3,  -6, -21,  17,  13,  -5,  -5,   0,   0,   0,   0,   0,   9,   7,  20,   2,   2,  -2, -11,   3,  -2,   0,  -1,   0,  19,  18, -22,  23,   7,  -1,  -3,  -2,   0,  -2,   0,   0,   5, -27, -15,  -1,  10,   8,  -1,   1,   0,   2,  -1,   1, -10,  10, -31,  26, -26,  19,   0,  -2,  44, -48,  -2,  23, -18,   2,   9,  -3,  31, -27, -10,  11,   2, -10,   5,   3, -17,  21, -15,  -7,  15,  -8,   1,   0, -10,  10,   4, -13,  -4,   3,   4,  -5,  -2,   2,   2,  -3,  -2,   1,   1,  -2 },
     {    7,   2,  19,  18,  -9, -11,   0,  -2,   1,   0,  -1,  -1,  11,  12,  18,   8, -12,  -5,  -5,   3,  -1,   0,  -1,  -1,  12, -14,  -6,   3,  -3,  16, -11,   4,  -2,   0,  -1,   1, -31, -31,   2, -21, -25,  32,   1,  -2,   3,   0,   1,   1,  46, -10,  12, -11, -31,  -6,  15, -10,  -8,  -2,  37, -26,  19, -21,  11,   1,  16, -20,  28,  -6,  -8,   4, -11,  11,  -4,   3,  -9,  24, -11,   3,  -5,   0,  -1,   4, -10,   8,  -4,   5,  -4,   6,   0,   1,  -3,   2,  -2,   2,  -2,   2 },
     {    8,  19,  30,  32,  -4,   6, -15,   1,  -4,   0,  -2,  -1,  -8,  -5, -26, -19,  14,  24, -15,   6,  -2,   3,  -1,   1,   2,  11,   9,  10,  -5,  30, -23,   2,  -5,  -1,   0,   0,   2,   4, -20, -17,   2,  27,   0,  -3,   2,  -2,   1,   0,  -7,  12,  15,  10,  -5,   0,   9,  -7, -18, -17, -36,  11,  26,  21,  11,  -1,  -8,  -1, -25, -22, -33,  16,  22,   0,   0,   1,  -2,   7, -13, -25,  18,  -2,   3,  -2,   4, -10,   1,  -1,   3,   1,   1,   1,   0,  -2,   0,   0,   1,   1 },
     {    6,   5,  21,  23, -27, -21,  40, -23,   8,  -6,   4,  -4,   3,  -2,  -8, -13,   9,  -3,  25,  -7,  -1,   0,   1,   0,  -4, -15, -23, -38,  -3,  40, -18,  20,  -7,   5,  -1,   3,   6, -11,  -8,  -7,  16,  -7,   7,  -3,   4,  -3,   4,   0, -26, -30, -14,  -2,  30, -22,  15,   1,   8, -11, -15,  16, -11,  -8,  -8,   9,  12, -16,  10,   2,  12,  -6, -13,   5,   2,  -1,   2,  -5,  -2,  12, -12,   2,  -6,   9,  -1,   4,   0,   2,   2,   0,  -1,   3,  -1,   2,   0,   2,   0,   1 },
     {   -6,  -5, -15,  -4,  24, -27,  25, -13,   7,  -3,   3,  -1,  -7,  -7,  10,  16, -34,  26, -15,   7,  -6,   4,  -1,   2,  -1,  11,  21, -22,  10,   7, -14,   1,  -2,   0,   0,   1,   1,   2, -34,   6,  32, -14,   9,  -6,   3,  -3,   2,   0,  -2, -20,   1,  39, -18,   8,  -2,  -1, -35, -21,  33,   2,  -6,   7,   6,  -2, -38,   5,  19,  -8,  -3,  -3,   8,   2,  -3,  -6,  -1,   5,   1,  -5,   2,   1,   9,  -6,  -7,   1,   3,  -1,  -1,  -3,   3,   0,  -1,   0,   1,   2,  -2,   0 },
     {    2,   0,   8,   1,  16,  35, -40,  12,  -6,   3,  -2,   1,   6,   5,   8,  15,  34,  35, -47,   8,  -5,   0,  -1,   1,   1,  -3,  10,  -7,  10,  29, -20,  -1,   2,  -1,   0,   1,  16,  16,  13,  17,  21,  -3,   5, -14,   3,  -4,   1,  -1,  -6, -23,  -5, -12,   9,  11,  16,  -3,   4,  12,   9,   6, -19, -14,  10,  -3,   7, -18,  17,   0,  22,   2,  -7,   2,   4,  -5,  -7,   1, -20,  26,   2,   0,  -7,   6,  -8,  10,  -3,   2,  -2,  -2,  -3,   1,   1,   1,  -1,   0,  -1,  -1 },
     {    1,  -5,   3,  -5, -21,  -5,   7,   2,  -1,   3,   0,   0,  -2, -14, -25,   0,  -2, -11,  18,  -1,   2,   2,   1,   1,   0,  -4,  25,  19, -34,  -9,  11,  -5,   0,  -1,  -1,  -1,  26,  28,   2,  19,  -3,  -1,  -3,   3,  -4,   0,  -1,   0,  -3, -37,   0,   5, -19,  33, -13,  10,  -9,  -2,  -9,  -9, -20,  17,   6,  -5,  32, -39,  26, -19, -17, -14,  13,  -1,  17,  -4,  -6,  22,  -7, -16,  -6,   4,  -7,  17, -14,  12,  -6,  11,  -3,   0,  -3,   3,   2,  -3,  -2,   1,   0,   1 },
     {   -8, -10, -24, -33, -23,  10,  20,  -6,   2,   3,   1,   2,  -4,  -2,   2,   8, -22,  -3,   8,  -1,  -2,   3,  -2,   1,   1,  11,   2,  -1,  -8,  12, -26,  16,  -7,   0,  -2,   2,   5,   2,  -4,  -1,  10,  48, -47,  14,  -1,  -2,   1,   2, -20, -11,  -4, -26,  -6,  37,  -9,  -6,  17,   1,   1,   2,  17, -17,   7,  -9, -10,   9, -14,  -6,  20,  26,   0,  -2,  -4,   1,  -5,  -5, -26,  12,  19,  -5,   4,  -4,   2,   2,   3,   0,  -1,   0,   0,   1,   0,   1,   0,   1,   2,   0 },
     {    3,   1,   3,   4,  15,   4,  -4,  -7,   1,  -2,   0,   1,   1,  -7,   9,  -5,   3,   3,   7, -15,   3,  -2,   2,   0,  13,  18,  24,  11,   2, -12,  18,  -8,  -2,  -1,  -2,   0,  18,  -2,  -1, -23,   7,  10,  20,   2,  -3,  -1,   2,   0, -34, -21,   9, -18, -10,   0,  28,   8,  24, -25, -13, -40,  10, -33,  23,  15, -16,  22,  -7, -26,   3, -33,  -9,  19,   9, -10,  11,  -1,   4,   1, -28,   9,   5,  -9,   4,   4,   0,   0,  -5,   8,  -3,   3,  -5,   1,   1,   1,  -1,   4 },
     {   -1,   6,  19,  23, -20,  -2,  -8,   3,  -2,   0,  -4,   1,  -7,   3, -13, -30,  19,  -5,   6,   0,   3,  -3,   0,   0,  -9,  -8,  -5,  20, -31,   1,  -3,   5,  -5,   4,  -1,   0,  -8,   8,  25,  -8,  16, -13,   1,   0,  -3,  -1,   1,  -1, -12,   6, -16,  14, -21,  17,  -3,  -2,  20,   2,  27, -12, -16,   6,  -8,  -4, -40,  32,  33, -17,  14,   4,  -2,  -9, -40,  33, -13,   2,   0,   7,   5,  -4,   4,  -7,   2,   4,   7,  -7,  -3,   6,   2,  -3,   0,  -1,   4,  -1,   0,   1 },
     {   -1,  -7,  -2,  -7, -15,  -5, -11,  22,  -5,   1,   3,   2,  -1,  -9,   7,  -7, -10, -15, -10,  20,  -1,  -1,   0,   1,   5,  -3,  -6,  -1,   5,  11, -22,   7,   3,   0,  -1,  -1,  -2, -22,  15,   7, -12,  -2, -10,  -8,   6,   1,  -3,   1,   2,  20,  20,   6,  11,   0, -12,  -9,  28, -33,   1,   6, -17,  24,   2,  -7,  -9,   7,  32, -29,  -4, -10,  14, -10,  18, -42,  50, -13,  -5,   7,  -6,   7,   1, -10,   4,  16,  -1,   7, -12,  12,  -5,   9, -13,   5,  -3,   2,  -2,   0 },
     {   -2,  -2,  -1,   0,  -3, -40, -38,  35,  -5,   6,  -1,   4,  -3,  -4,  -1,  -4, -13, -32, -34,  26,   0,   3,   2,   2,  -6,  -8, -16, -35, -32, -12,  16,  -6,  11,  -1,   5,   1,  -1, -12, -23, -12, -16, -13,  15,  -1,   7,   2,   0,   1, -29, -29, -11, -19, -13,   4,   2,   5,   1,  -1, -10,   1,  10,  -4,  -5,   4, -16, -12, -12,   2,   8,  24,  -5,  -1,  -2,  -8,  -9,  -3, -10,  -5,   8,  -2,   1,  -3,   2,   0,   3,   3,   3,  -1,   1,   1,   5,  -1,   4,   1,   1,  -1 },
   },
   {
     { -108,  50,  10,   1,   3,   1,   1,   0,   1,   0,   0,   0,  26, -25,   6,  -2,   1,   0,   0,   0,   0,   0,   0,   0,  22, -11,  -4,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   4,   0,  -3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
     {    7,  57, -31,  -4,  -5,  -1,  -2,  -1,  -1,   0,  -1,   0, -82,  33,  26,  -9,   5,  -2,   2,  -1,   1,   0,   1,   0,  23, -45,  20,   1,   1,   1,   0,   1,   0,   0,   0,   0,  17, -11,  -7,   5,  -2,   1,  -1,   1,   0,   0,   0,   0,   5,  -4,  -2,   1,   0,   0,   0,   0,   4,  -1,  -2,   1,  -1,   0,   0,   0,   2,  -1,  -1,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   1,   0,  -1,   0,   1,   0,   0,   0,   1,   0,   0,   0 },
     {    8,  52, -47,   5,  -4,   0,  -1,   0,  -1,   0,   0,   0,  51,  42, -32, -14,  -1,  -4,  -1,  -2,  -1,  -1,   0,  -1, -44,  19,  37, -17,   4,  -3,   1,  -1,   1,  -1,   0,   0,  -2, -28,  27,   2,  -1,   1,  -1,   1,   0,   0,   0,   0,   0, -11,   3,   8,  -2,   2,  -1,   1,  -1,  -6,   0,   4,   1,   1,   0,   0,   0,  -3,   0,   1,   1,   1,   0,   0,  -1,  -3,   0,   1,   0,   1,   0,   0,   0,  -2,   0,   1,   0,  -2,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0 },
     {  -48, -55,  37,  11,   6,   3,   2,   1,   1,   1,   1,   1, -30,  68, -37,  -5,   1,  -2,   0,  -1,   0,   0,   0,   0,  -6,   4,  23, -16,   4,  -3,   1,  -2,   0,  -1,   0,   0,  25, -29,   5,   5,  -3,   2,  -1,   1,  -1,   1,   0,   0,   6,   2,  -5,   1,  -1,   0,  -1,   0,   6,  -5,   0,   2,  -1,   1,   0,   0,   2,   2,  -2,   0,  -1,   0,   0,   0,   2,  -2,   0,   1,   0,   0,   0,   0,   1,   1,  -1,   0,   1,  -1,   0,   0,   1,   1,  -1,   0,   1,  -1,   0,   0 },
     {    8, -13,  22, -13,   1,  -1,   1,  -1,   0,   0,   0,   0,  43,   8,  38, -29,  -7,  -3,  -2,  -2,  -1,  -1,  -1,  -1, -21, -50,  48,  28, -15,   7,  -4,   3,  -2,   1,  -1,   1,  10,  -9, -44,  37,  -3,   2,   0,   1,   0,   1,   0,   0, -14,  21, -21,   0,  10,  -4,   2,  -2,  -1,   2,  -6,  -4,   5,   0,   1,   0,  -3,   6,  -3,  -4,   2,   0,   0,   0,   0,   1,  -2,  -1,   0,   1,   0,   0,  -2,   3,  -1,  -1,   0,   0,  -1,  -1,  -1,   1,   0,  -1,   0,   0,  -1,   0 },
     {   10,  31,  62, -54,   3,  -9,   0,  -4,   0,  -2,   0,  -2, -20, -19,  26,   7,  -8,   5,  -2,   2,  -1,   1,   0,   1, -31,  37,  -2,  -7,  -2,  -1,   0,  -1,   0,   0,   0,   0,  13, -37,  31,  13,  -9,   2,  -3,   1,  -1,   1,  -1,   0,   8,  -9, -26,  26,   0,   1,   0,   0,   0,   2,  -9,  -1,   6,  -1,   1,  -1,   2,  -1,  -7,   2,   2,   0,   0,   0,   1,   1,  -2,  -2,   2,   0,   0,   0,   1,   0,  -3,   1,   0,   0,  -1,  -1,   1,   0,  -2,   0,   0,   0,  -1,   0 },
     {   13, -10,   7, -14,   7,  -1,   1,   0,   0,   0,   0,   0,  21,  13,   2, -33,   9,  -3,   1,  -1,   0,   0,   0,   0,  85,  -1,  18, -31, -14,   3,  -5,   0,  -2,   0,  -1,   0, -33,   4,  38,  19, -22,   5,  -4,   1,  -2,   1,  -1,   0,  -8, -11,  -9,  32,  -5,   0,   0,   0, -12,   8, -12,   8,   6,  -3,   2,  -1,  -4,   0,  -2,   1,   5,   0,   1,   0,  -5,   1,  -2,   1,   0,   1,   0,   0,  -2,   0,  -1,   1,  -2,   1,  -1,   0,  -1,   0,  -1,   0,  -1,   0,  -1,   0 },
     {    8,  21,  36, -44,   7,  -4,   0,  -2,   0,  -1,   0,  -1,  22,  63,  -2, -10, -17,  -1,  -5,  -1,  -2,  -1,  -1,  -1,   8, -16, -41,  38,   4,  -1,   2,   0,   1,   0,   0,   0, -13,  23,  -7, -30,  16,  -4,   3,  -1,   1,  -1,   1,   0,   0, -19,  37,  -5, -10,   4,  -3,   1,   0, -10,  -1,  14,  -2,   0,  -1,   0,  -3,  -3,   3,   2,   2,   0,   0,   0,  -1,  -4,  -1,   3,   0,   1,   0,   0,  -1,  -2,   1,   1,   0,  -2,  -1,   1,  -1,  -1,   0,   0,   0,  -1,  -1,   1 },
     {   23,  33,  29,  -4, -20,  -1,  -7,  -1,  -2,   0,  -1,   0,   9,   2, -65,  52,  -6,   3,  -1,   1,   0,   0,   0,   0,  36, -16,  -3,   2,   6,  -7,   1,  -2,   1,  -1,   0,   0,  -4, -17, -16,  27,   8,   0,   1,   0,   0,   0,   0,   0, -15,  22, -26, -21,  22,  -4,   5,  -1,  -3,  -1,  20, -21,   1,   2,  -1,   1,  -3,   0,   3,  -3,  -2,   2,   0,   0,  -2,  -1,   7,  -4,  -1,   1,   0,   0,  -1,   0,   0,   0,  -1,   0,   3,  -2,  -1,   0,   0,   0,  -1,   0,   2,  -1 },
     {   13,   0,  12, -15,   3,  -1,  -1,   0,   0,   0,   0,   0,  15, -24, -25,  11,  17,  -2,   2,   1,   1,   0,   1,   0,  -9, -58,   3, -34,  26,   8,   3,   3,   1,   1,   1,   1,  55,   9,  14, -41, -27,  13,  -7,   3,  -3,   1,  -2,   1, -13,  16,  17,  22, -28,   3,  -4,  -1,  -5,   9,  -8,  16,  -3,  -4,   2,  -2,  -8,   8,  -1,   4,   2,  -4,   1,  -1,  -2,   2,  -1,   2,   2,  -1,   0,  -1,  -3,   3,   0,   1,   0,   0,  -1,   1,  -2,   2,   0,   1,   0,   0,   0,   0 },
     {    4,   9,  10, -38,  26,  -7,   4,  -3,   1,  -2,   1,  -1, -15, -26, -33, -41,  37,   3,   8,   2,   3,   1,   2,   1,  -2,  15,  27, -18, -17,   7,  -5,   1,  -2,   1,  -1,   0, -15,   4, -24,   7,   8,   3,   2,   1,   0,   0,   0,   0,   8,  10,  21, -48, -11,   9,  -2,   2,  10, -12,  26,  22, -34,   4,  -4,   1,   2,   5, -10,  16,  -4,  -6,   2,  -2,   1,  -1,   3,   7,  -1,  -5,   0,   0,   1,   1,  -3,   3,   0,  -1,   2,   3,   0,   0,  -1,   2,   0,   0,   1,   2 },
     {   -3, -15, -51, -41,  47,  -2,   8,   1,   4,   1,   2,   0,  10,   3,  -2,  -8, -14,  12,  -4,   1,  -1,   1,  -1,   0,  28,  24, -30,  21, -11,  -3,  -1,  -2,   0,  -1,   0,   0,  56, -37,  -8,   3,   0,   0,  -1,   1,   0,   0,   0,   0, -15,  23,  -8,  -1,  -1,  -1,   1,   0, -14,   7,  13, -12,   2,  -1,   1,   0,  -7,   1,   8,   2,  -6,   0,   0,   0,  -4,   0,   4,   0,   1,  -2,   1,   0,  -2,  -1,   3,   1,  -2,   0,   2,   0,  -1,   0,   1,   1,  -1,  -1,   1,   0 },
     {    3, -15,  -4,   3,   5,  -3,   3,   1,   0,   0,   1,   0,  13,  -6,  20,  13,  -4,  -7,  -3,  -1,  -1,  -1,   0,   0,  13, -26,   4,  10,  34,  -8,  -1,  -2,   0,  -1,   0,   0, -36, -65,  10, -30,  31,  18,  -1,   7,   0,   3,   0,   2,  38,  -2, -21, -24, -27,  20,  -6,   5, -11,  19,   1,   7, -24,   1,   1,  -2,   4,   8,  -5,  10,  -6,  -6,   2,  -2,  -6,   8,  -1,   1,   0,  -3,   1,  -1,   1,   3,  -1,   1,  -3,   3,   0,   1,   1,   1,  -1,   0,  -1,   1,  -1,   0 },
     {   16,  16,  23,  41, -49,   8,  -9,   1,  -3,   0,  -2,   0,   9,  -2,  15, -22,  -6,   3,   1,   0,   0,   0,   0,   0,  29,  39,   0,  -1,  -9, -10,  -1,  -4,   0,  -2,   0,  -1,  44, -26, -24,  -5,  17,   0,  -1,   1,   0,   0,   0,   1,  -4,  14,  23, -11, -24,   6,  -3,   1, -13,  -2,  -4,  41,   2, -13,   4,  -3,  -9,   3,  -7,  -7,  20,  -2,   1,   0,  -4,  -4,   1,   0,   2,   4,   0,   0,  -3,   0,  -2,  -3,  -1,  -2,   1,   0,  -2,   0,  -1,  -1,  -1,  -1,   0,   1 },
     {   -5, -11, -11, -19,  11,   9,   0,   1,   1,   1,   0,   0, -11, -25, -27,  35, -40,  22,  -2,   5,   0,   2,   0,   2,   7,  -7,  57,  35, -33, -17,   2,  -6,   0,  -2,  -1,  -1,   1,   8,  15,  -3,  24, -13,   1,  -4,   0,  -1,   0,   0,  10, -19,  23,   8,  -9,   3,  -2,   1,  -1,   2, -25,  29,  11,  -9,   3,  -1,  -4,   5, -15, -12,  18,  -2,   0,   2,   1,   1,  -4,  -5,   4,   2,  -1,   0,   0,   2,  -2,  -5,   0,   1,   0,  -1,   0,   1,   0,  -2,   0,   1,   0,  -1 },
     {    4,  -7,   4,  -7,  -2,  -1,   2,   0,   1,   1,   0,   0,  15,   0,  30,  15,  -5, -16,   1,  -4,   0,  -2,   0,   0,  19,   7,  26, -23,  22, -24,   4,  -3,   0,  -2,   0,  -1,  39,  16,  13,  -2,  31, -16,  -9,   0,  -4,   0,  -2,  -1, -24, -39,   7, -21,  39,  25, -10,   7,  14, -19,  -1, -16, -29,  30,  -6,   5, -11,   4,  -6,  19, -20,   5,   3,  -1,   0,   3,  -9,   3,  -5,  -2,   3,   0,  -4,   2,  -2,   4,   0,   1,  -4,   0,  -2,   1,  -1,   2,   0,   0,  -2,   0 },
     {   12,   5,  14,  33,  15, -25,   1,  -6,   0,  -3,   0,  -2,   2, -15, -30, -30,  49,  -3,   0,   3,   1,   0,   1,   0,  -2, -19, -20,  38, -22,   8,  -2,   2,   0,   1,   0,   1,  26, -15,  19,  11,  16, -19,   0,  -3,   0,  -1,   0,   0,  22, -45,  -1,   3,  18,  10,  -8,   5, -13,  20, -22, -11,   2,  10,  -2,   0,  -4,   4,   7, -15,  -7,   8,   0,   0,  -4,   7,   0,   1,  -7,   2,   1,   0,  -2,   2,   4,  -4,  -1,   2,   0,   1,  -1,   1,   1,  -1,  -1,   2,   0,   1 },
     {   -5, -12, -14, -20, -52,  38,  -4,  10,  -1,   4,  -1,   3,   5,   2,  -6, -12, -20,  25,  -6,   4,  -1,   0,   0,   1,  -9, -21, -27, -10, -27,  17,  10,   0,   4,   1,   2,   1,  11,   1,  13,  18, -17, -18,   7,  -5,   1,  -1,   1,  -1,  21, -26, -15, -12,  10,   4,   4,   2,   3,  25,  11,   4, -40,  -7,   2,  -2,  -8,   8,  -1,  30,  11, -25,   4,  -4,  -2,   4,   1,  -6,  14,  -8,  -1,   1,  -2,   0,   3,   3,   0,   1,   0,  -2,  -1,  -1,   2,   2,   1,   0,   0,  -1 },
     {   -4,  -1,   3,   7, -14,  19,  -5,   3,  -3,   1,  -1,   0, -16, -18, -15, -41, -12,  19,   6,   5,   2,   2,   1,   1,  -7, -11,  -9,  28,  12, -13,  -1,  -3,   1,  -1,   0,   0, -21, -27,  30,  -4,   8,  11,  -1,   0,  -2,   2,  -1,   1, -81,   8,   9,   0,  -7,   6,   5,   2,  34,  -4, -16, -10,  -5,   5,   0,   1,   5,   6,  -1,   2, -17,  -4,   3,  -1,  10,   4,  -2,   5,   4, -11,   1,  -1,   3,   2,   2,  -1,   3,   2,  -2,   1,   1,   2,   1,   0,   2,   0,   0,   0 },
     {   -2,   2,  -1,  22,  29, -32,   3,  -5,   1,  -2,   1,  -2,   4,  13,  19,  48,  14,  -7, -16,  -3,  -5,  -1,  -3,  -1,  -4,  -1, -12,  -4, -26,  39,  -9,   5,  -1,   2,   0,   1,  -6, -18,   5,  18, -16,   7,   4,  -1,   1,  -1,   1,   0, -48, -20,  11, -17, -10,  12,   8,   1,  19,   3,  -5,  24, -24, -15,   3,  -4,  -8,   8, -13,  10,  30, -13,   2,  -3,   7,  -1,   1, -16,   6,   4,  -1,   1,   0,   1,  -1,   0,   2,   0,   1,  -4,   0,   0,  -1,  -1,   1,  -1,   0,  -2 },
     {   12,  14,  24,  46,  35, -21, -11,  -4,  -4,  -3,  -2,  -1,   3,   1,   9, -11, -47,  43,  -9,   5,  -1,   2,  -1,   1,   3, -10,   7, -12, -30,  15,   9,  -3,   1,   0,   1,   0,   3,  -1,  21, -25, -12,  -7,  10,   0,   0,  -1,   0,   0,  14,  20,   6,  -8,   9, -11,   0,   0, -15, -18,  22, -16,  -5,  17,  -2,   4,   5,  -1,  -6,  22, -26,  -7,   5,  -2,  -2,   2,  -7,   9,  14, -13,   0,  -1,   1,  -2,  -3,   2,  -1,   0,  -3,   1,  -1,   0,  -1,   1,  -1,  -1,  -1,   0 },
     {    1,  14,   9,  13,  18, -10,  -6,  -2,  -1,  -2,   0,  -1,  -6,   8, -12, -13, -15,   9,  -1,   4,   1,   2,   0,   1,  -2,  22,  20,   7,   9,  -8,  -7,  -4,  -1,  -1,  -1,  -1,  10,  45, -18, -17,  14,  22,  -8,  -1,  -2,   0,   0,   0, -17, -17, -53,   3, -25,  25,  10,   1,  -9,  48,  -4, -15, -22, -18,  14,  -6,   0, -13,  23,  10,   8, -19,   2,  -1,   6,  -7,  -1,  -3,  11,   0,  -3,   1,   0,  -7,   6,   0,   1,  -2,   0,  -1,   1,  -3,   3,   0,   1,  -1,  -1,   0 },
     {    3,   3,  -3,  -3,  -4,  13,  -7,   4,  -1,   1,  -2,   1,   1,  -2, -10,  13,  -1,   8,  -9,  -2,   0,   1,   0,   0,  10,  19,   7,   1,   6,  26, -14,   1,  -2,  -1,  -1,   0,  -2, -13, -51,  -3, -21,  30,   9,   1,   4,  -1,   2,   0,  10, -36,  13,  23,  -6,  -6,   4,  -3,   0, -20, -40, -15,  -6,  22,  11,   0,  -6,  30,  -1,  -5, -41, -24,  18,  -4,  -1,   4,   8,   9,  17, -26,   1,  -2,   1,   3,   6,  -8,  -2,  -1,   2,   2,   1,   1,   2,  -3,   0,  -1,   1,   1 },
     {   -5,  -3,   1,  -4, -31,  -1,  13,   3,   4,   0,   1,  -1,  -4,   4,   3,  13,  22, -41,   7,  -1,  -2,  -1,   0,  -1,   5,  15,  25,  58, -14,  10, -25,   0,  -8,   3,  -2,   1,   8,  10,  34, -21, -27,  15,  -1,  -3,   0,   0,   0,   0,  -6,  22,   6,  -4,   8,  -2,   0,  -5, -35,  -2,  20,  -9, -14,  15,   8,  -1,  16,  -8, -11,  22,  -9, -20,   8,   0,  -3,  -1,  -7,   2,  20,  -4,  -4,   1,   2,  -2,   0,  -5,  -1,  -2,  -3,   1,   2,  -1,  -1,  -1,  -1,   0,  -1,  -1 },
     {    1,   2,  -5,  13,  11, -35,  18,  -4,   4,  -2,   2,  -1,  -7,  -9, -18, -22, -45, -31,  31,   2,   6,   1,   2,   1,  -7,  -8, -25,   1,  25, -28, -11,   9,  -3,   2,  -2,   2,  -2,   6,  -4,  33,  -6,   6, -21,   5,  -4,   1,  -2,  -1,  12,   3,   0,  19,  13,   5,  -8,  -3,  -7, -16,  -3,  23, -12,  12,   1,  -1,  -3,  25, -23,  12,  15, -25,   7,  -5,  -3,  -4,  13, -19,  24,   4, -10,   2,   2,   1,   3,  -4,   0,  -1,   3,  -4,   1,   1,   1,  -1,   0,   0,   3,  -2 },
     {   10,  -1,   2,   9,   4,  -3,   4,  -1,   0,  -2,   1,   0,  10, -10,  -5,  -7,  -7, -15,  14,   1,   2,   2,   1,   0,  17,   3,  10,  26,   7,  -8,  -2,  -2,  -1,   0,  -1,   1,  17,  -6,  -3, -16, -31,  13,  14,  -1,   2,   1,   0,   0,  21, -36, -18,  -4, -13, -32,  14,   4,  72, -19,  29,  -5,  13, -23, -15,   7,  -4,  -7,   0,  19,   9,   6, -16,   5,  -5,  10,  -7,   2,   8,  10,  -7,  -2,  -9,   5,  -1,   2,  -3,   3,  -3,   1,  -4,   2,   0,   0,  -1,   2,  -1,  -1 },
     {    3,   5,  -6,   6,   2,  12, -10,  -2,  -2,   2,   0,   1,   7,  12,   1,   4,  16,  15, -24,   1,  -1,   1,  -2,   0,   3,   4, -11,  19,  -4,   6, -14,   2,   1,  -1,  -1,  -1,  12,  29,  26,  11,   9,  16, -22,  -7,  -2,  -1,  -2,  -2,  29,  39, -35,   2, -11,  29,   3, -10,  46, -26, -13,  17,  -7,  14,  10,  -4, -15,  10, -29,  -9,  -7,  15,  19,  -3,  -9,   3,   6,  -6, -12, -22,  18,   0,  -7,  -1,   4,   0,  -1,  -2,   5,  -3,  -2,  -3,   0,  -1,   1,   0,   2,  -2 },
     {   -1,   0,  -6, -12, -29, -40,  41,  -4,   6,  -2,   4,  -1,   0,   1,  -5,  -5, -20, -24,   7,   6,  -1,   4,  -1,   2,   5,   3,   2, -20, -34,  22,  -9,   5,   0,   2,   0,   1,  -4,  -4,  -1, -39,  12, -18,  12,   1,   2,   0,   0,   0,   4,  29, -15, -10,  11,  -1, -14,   4,  39,  14, -40, -10,   0,  11,  -2,  -3, -12,   6,  22, -13,   0,  -6,  -2,   0,   5,  -9,  11,  12,   0,   1,  -7,   0,  -7,   1,   3,   2,   0,  -5,   2,   4,  -3,   1,   0,   1,   0,  -2,   0,   2 },
     {    6,  10,   7,   8,  34,  44, -24,  -3,  -8,   0,  -3,   1,   3,   6,   3,   3,  24,  -7,   6, -11,   2,  -4,   0,  -1,   5,  10,   2,  11,   3, -21,  28,  -7,   2,  -3,   1,  -1,  -7,  -3,  -6, -26, -25, -36,  16,  12,   2,   1,  -1,   2,  -1,  22, -20,  -4,  23, -10, -25,   6,  10,  -3, -35,  11, -17,  11,   4,  -4,   6,  15,   1,   2,  17, -27,  -4,  -3,  -1, -10,  12, -12,  16,  25, -12,  -1,   2,  -2,   1,  -4,  -1,  -2,   2,   0,   0,  -1,   0,  -1,  -1,  -1,   0,  -2 },
     {    2,   4,   7,   3,  19,  26, -29,   1,  -6,   4,  -2,   0,  -6,  -7,  -2, -13, -34, -22, -13,  16,   2,   6,  -1,   2,   1,   6,   7,   2,  34,  36, -46,  -1,  -5,   1,   0,   0,   5,   1,  17,  13,  -9,  17,  16, -12,   0,  -6,   0,  -1,  11,   6,  24, -27,  23, -14,  -6,   4,   4,  35,  -8,   2,   1,  18,  -5,  -4, -23,  -7,  11, -12,  13,   6,   2,   2,   0,  -1,   5,  -2, -19,  20,   7,  -4,  -5,  -3,  -2,  10,   0,  -2,   0,   3,  -2,  -1,  -2,   1,   1,   0,  -1,   0 },
     {    1,  -4,  -2,  -5,  -1,   0,  -9,   5,  -1,  -1,   1,   1,   1,  -6,  -1,  -9,  -1,  -1,  -7,   6,   1,   1,   1,   0,   2,  -7,  -3,  -5,   8,  24,  -9, -13,   5,  -3,   1,  -1,  10,   3,  18,   2,  30,  11,  10, -17,   2,  -2,   0,  -2,   7,   5,  -8,  -1,  11,   6,  36,  -2, -15, -61, -16, -14, -15, -31,  22,  12,  29,  19,  22, -12,  19, -27, -31,  14,   2,  -5,   0,  10,  -3,  28,  -9,   0,  -3,  10,  -2,  -2,  -1,  -2,   1,   3,   0,   7,  -1,   0,  -4,  -2,   1,   1 },
     {   -5,  -4,  -9,  -9, -14, -26,  -1,   0,   9,   6,   4,   1,  -2,   3,  -3, -11,   1,  17,  -6,  -5,  -3,   4,   0,   0,   1,   6,   5,   6,  28,  24,   6, -21,  -2,  -7,   1,   0,   5,  16,   3, -15,   6,  -8,  43,  -3,  -4,  -5,  -2,   0, -18, -11, -37,  -8,   5, -26,   4,  32, -22,   1,   7,  27,   7,   8, -29,   8,  18,  33, -39, -15,   0,  14,   5, -12,   1,   3,  17, -17,  -5,  -7,   6,  -1,  -2,  -5,   2,  -2,  -1,   2,  -1,   0,   2,  -1,   2,  -3,   1,  -3,   2,  -2 },
   }
 },
 {  //32
   {
     {  114, -37, -11,  -4,  -3,  -2,  -1,  -1,  -1,   0,   0,   0, -14,  31,  -9,   0,  -2,   0,  -1,   0,   0,   0,   0,   0, -21,  11,   8,  -3,   1,  -1,   0,   0,   0,   0,   0,   0,  -3,  -3,   6,   2,  -1,   0,   0,   0,   0,   0,   0,   0,  -5,   0,   0,   3,   1,   0,   0,   0,  -2,  -1,   0,   0,   1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    0, -53,  15,   0,   5,   0,   2,   1,   1,   0,   1,   0,  82, -41, -28,   8,  -4,   2,  -2,   1,  -1,   1,  -1,   1,  -2,  46, -32,  -3,   0,  -1,   0,  -1,   0,   0,   0,   0, -11,  16,  10, -14,   2,  -2,   0,  -1,   0,   0,   0,   0,  -2,   3,   7,   0,  -4,   1,  -1,   0,  -4,   3,   2,   2,   0,  -1,   0,   0,  -2,   1,   2,   0,   1,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0,  -1,   1,   1,   0,  -1,   0,   1,   0,  -1,   0,   0,   0,  -1,   0,   1,   0 },
     {  -28,  14,  -3,  -2,  -1,   1,   0,   0,   0,   0,   0,   0,  31,  60, -32, -11,  -5,  -3,  -1,  -2,  -1,  -1,   0,  -1, -27,  39,  46, -28,   2,  -6,   1,  -3,   0,  -1,   0,  -1,   8, -37,  43,  14, -11,   3,  -3,   1,  -1,   1,  -1,   1,  -1,  -7, -10,  24,   0,  -1,   0,   0,  -1,  -6,  -3,   2,   9,  -1,   0,   0,   0,  -3,  -3,   1,   1,   3,   0,   1,   0,  -3,  -1,   0,   0,   1,   1,   0,   0,  -1,  -1,   1,   0,  -2,   0,   0,   0,  -1,  -1,   0,   0,  -1,   0,   0 },
     {   -8, -43,  26,   4,   2,   1,   1,   1,   0,   0,   0,   0,  35,   9,  10,  -4,  -1,  -3,  -1,  -2,  -1,  -1,   0,   0,  -3, -33,  45,  22,  -7,   4,  -3,   2,  -1,   1,  -1,   1,   0, -21, -48,  50,   6,   3,   2,   1,   1,   0,   1,   0, -11,  17, -33, -22,  27,  -1,   4,  -1,   3,  -4,   4, -19,  -5,  10,  -1,   2,  -2,   7,  -5,  -1,  -7,   0,   3,   0,   0,   0,   0,  -3,  -1,  -2,   0,   1,  -1,   3,  -1,  -1,   0,   0,   0,  -1,  -1,   2,  -1,  -1,   0,   0,   0,   0 },
     {   33,  38,   6, -11,  -5,  -5,  -2,  -2,  -1,  -1,  -1,  -1,  26, -50,  33,   2,  -4,   0,  -1,   0,   0,   0,   0,   0,  32,  -6,  11,  -6,  -9,  -1,  -3,   0,  -1,   0,  -1,   0,   4,  -6,  28,  32, -23,   2,  -6,   1,  -2,   1,  -1,   0, -10,   5, -32,  49,  10, -11,   2,  -3,  -8,   6, -12, -10,  33,  -4,   0,  -1,  -2,  -4,  -3,  -7,   4,  13,  -3,   1,  -4,   3,  -4,  -3,  -1,   4,   4,  -1,  -1,  -1,  -2,  -2,  -2,   1,  -1,  -2,  -1,   0,  -1,  -1,  -1,   1,  -1,  -1 },
     {   25,  72, -45,   8,  -7,   0,  -3,   0,  -2,   0,  -1,   0,  69,  17,   0, -10,   1,  -5,  -1,  -2,  -1,  -1,   0,  -1,  -9, -14,   1,  17,   1,   2,   0,   1,   0,   0,   0,   0, -12,  13, -28,  -8,  15,  -2,   4,  -1,   2,  -1,   1,   0,  -5,  -1,  16, -26,   1,   4,  -1,   1,  -1,  -4,   7,   2, -12,   2,   0,   1,  -3,   0,   4,   2,   0,  -4,   0,   0,  -1,  -2,   2,   1,   0,   0,  -1,   0,  -2,  -1,   2,   0,  -1,  -1,   1,   1,  -1,  -1,   1,   0,  -1,  -1,   0,   0 },
     {    3,  29,  -4,   4,   1,  -2,  -1,  -1,  -1,   0,  -1,   0, -26, -11, -29,  31,   4,   3,   1,   2,   0,   0,   0,   0,   0,  44, -21,  -8,   7,  -4,   0,  -1,   0,  -1,   0,  -1, -19,   9,   6,  27,  13,  -5,   3,  -3,   1,  -2,   1,  -1,  -7,  -5, -25, -24,  52,   0,   5,   0,   5, -14,  19, -48,  -6,  30,  -4,   5,   1,  -3,  -1,   3, -29,   5,   8,   0,   0,  -1,   5,  -3,  -4,  -9,   3,   1,   1,  -2,   1,   1,   1,   0,   2,  -2,   1,  -1,   0,   1,   0,   0,   1,  -1 },
     {  -18, -16, -17,  14,   4,   0,   1,   1,   1,   1,   1,   0,   5,  37,  16,  10, -13,  -3,  -6,  -2,  -2,  -1,  -1,   0, -34, -12, -31,  32,  10,  -2,   3,   0,   2,   0,   1,   0,   7,  31, -10, -13,  -7,  -7,  -1,  -2,   0,  -1,   0,  -1,  -8,  12,   6,  37,  20, -22,   1,  -6,   2,  -8,  -6, -27,  55,   4,  -7,   2,   2,  -9,  11, -19,  -9,  32,  -6,   1,  -1,   0,  -6,  -1,  -9,   5,  11,  -2,   1,  -3,   4,  -5,   0,  -1,  -2,  -1,   1,  -1,   2,  -2,   0,  -1,  -1,  -1 },
     {    3,  23,  47, -20,   2,  -6,  -1,  -3,   0,  -1,   0,  -1,  -2, -21,  49,   4,  -4,   3,  -3,   1,  -1,   0,  -1,   0, -85,  24,  -3,  28,  -2,   3,   1,   1,   0,   1,   0,   0, -14, -20,   4,  -4,  15,  -2,   3,  -1,   2,   0,   1,   0,   7, -19,   0,  -5,  -4,   7,  -2,   2,   2,  -2,  -8,   7,  -7,   2,   1,  -1,   3,  -1,  -6,   1,   1,  -3,   2,  -1,   3,  -1,  -4,   0,   1,  -1,   0,   1,   2,   0,  -2,  -1,   1,   0,  -2,   0,   2,   0,  -1,   0,   1,   0,  -1,   0 },
     {    1,  15,   8,  15, -13,  -4,  -2,  -1,  -1,  -1,   0,   0, -19,  -7, -32,  11,   3,  -2,   2,   2,   1,   1,   0,   1,   0,  31,   6,  17, -20,  -9,  -2,  -2,   0,  -1,  -1,  -1, -22,  -8, -29,  21,  19,  -9,   5,   0,   2,   0,   1,   0,   0,  19, -12, -15, -17, -15,   3,  -2,   0,   8,  13,  19,  33, -45,  -4,  -4,   8,  -8,  10, -24,  46,  16, -27,   5,   0,  -2,   5,   0,  -9,  31,  -4,  -5,   0,  -1,  -2,  -4,   0,  -1,   3,  -1,   0,  -1,  -1,  -2,   0,  -1,   1,   0 },
     {    4,   8, -21,  16,   2,   1,   0,   1,   0,   0,  -1,   1,   1, -37, -39,  30,   7,   6,   2,   3,   1,   1,   0,   1, -46, -36, -11, -16,  28,  -2,   8,   0,   4,   0,   2,   0,  44, -46,  -8, -22,  -6,  12,  -3,   4,  -1,   2,  -1,   1,   1,  24, -28,   2, -15,   2,   1,  -1,   3,  11,   3, -15,   1,  -8,   4,  -1,   0,   6,   6,  -4,  -4,  -3,  -2,   1,   1,   4,   3,  -2,  -1,  -2,  -2,   0,  -1,   3,   2,   0,   0,   2,   2,  -1,   0,   1,   1,   0,   0,   1,   1,   0 },
     {    6,   2,   8,   2,   3,   0,   1,  -3,   0,  -1,   0,   0,  -1, -13,  -4, -17,  21,   7,  -4,   0,  -1,   0,   0,   0,  13,   3,  19,  -1,   1,   1,  -1,  -4,  -1,  -1,   0,  -1,   0, -19,  -1, -15,  27,  13,  -6,   2,  -2,   1,  -1,   1,  12,   3,  15, -16, -20,   9,  -1,  -2,   7,  -3,  17,  14,  25,  16, -20,   0,  -6,   2, -12,  -2, -29,  69,  16, -10,  -8,   2,  -9,   8, -30, -24,  48,   0,   0,  -3,   0,  -6,  -2,   2,  -2,   4,   0,  -1,   0,  -1,  -1,   0,  -1,   2 },
     {  -14, -29, -81,  30,  -5,  10,   1,   5,   1,   2,   1,   2, -13, -45,  22, -30,   7,  -2,   4,   0,   3,   0,   1,   1, -31,  16,  34,  -5, -11,   2,  -1,  -1,  -1,   0,   0,   0, -30,   4,   6,  13,  -3,  -4,   3,   0,   1,   0,   1,   0,   9, -12,  -1,   2,   4,  -3,  -1,   0,   3,   3,  -3,  -2,  -1,   0,  -1,   1,   4,   0,   1,   1,  -2,  -1,   0,   0,   2,   3,  -2,   0,   0,  -2,   0,   0,   3,   1,   1,  -1,   1,   1,   0,   0,   1,   0,   0,  -1,   1,   1,   0,   0 },
     {    1,  12,  22, -24,   0,   1,  -1,   0,   0,   0,   0,   0,  -6, -17, -18, -24,   6,   3,   4,   0,   2,   0,   1,   0, -37, -10,  18, -41, -11,   8,   0,   3,   0,   2,   0,   1,   0,  47, -15,   8, -46,   2,  -2,  -1,  -1,   0,   0,  -1, -18,  37,  26,  -2,   1, -29,   4,  -6,  15, -18,  33,   5,   1,  -3, -12,   2,   1,   1,   0,  20,  -3,   4,  -4,  -3,   2,  -4,   1,   5,   4,   0,   3,  -1,   0,  -1,  -3,   4,   1,  -2,   0,   2,   0,   0,  -1,   2,   1,  -1,   0,   1 },
     {   -5,  -9, -12, -14,  -3,  12,   2,   1,   1,   0,   0,   1,   7,  11,  16,  19,   9, -11,  -7,  -1,  -2,  -2,  -2,  -1, -12, -16, -17, -21,  -2,  17,   3,   1,   2,   1,   1,   0,  14,  20,  22,  19,   1, -17,  -9,  -4,  -3,  -1,  -1,  -1, -13, -19, -20, -14,  12,  35,   4,   1,   8,  14,  10,  -3, -31, -31,  -6,   0,   2,   7,   8,  25,  36,  33, -20, -15,  -4,   1, -10,  -9, -21,  19,  43,  -2,  -1,  -3,   1,   1,   0,   0,  -2,  -4,   0,  -1,   2,   2,   0,   0,  -1,  -2 },
     {   -1,   0, -23, -43,  13,  -1,   5,   0,   2,   0,   2,   0,  -4,   6,  14, -51,   0,   2,   1,   3,  -1,   2,  -1,   0,   6,  43, -30,  12, -29,   6,  -3,   1,   0,   0,   1,   0,  60, -17, -25,   6,  -4,  -7,   1,  -1,   1,  -1,   0,  -1,  21,  21, -19,  -6,   1,  -2,  -4,  -1,  -3,  18,  12, -12,  -4,  -1,  -1,  -2,   0,  -1,  16,   3,  -7,  -2,  -2,  -1,  -3,  -3,   8,   7,  -1,  -1,   0,  -2,   0,  -3,   5,   4,  -2,   0,   2,   4,   0,   0,   1,   1,   0,  -1,   1,   1 },
     {    5,  -8,   7, -14,  19, -14,   4,   1,   1,  -1,   0,   0,   9, -11,  14, -17,  19, -20,   6,  -1,   1,  -1,   1,   0,   9, -16,  14, -18,  23, -17,   1,   2,   0,   0,   0,   0,  10, -19,  17, -22,  18, -18,   2,   2,   1,   0,   1,   0,  15, -20,  21, -19,  25, -11, -10,   7,  13, -25,  16, -24,  10,   1, -10,  10,  13, -15,  16,  -9,   0,   5, -39,  16,   2,   4,  -1,  15, -14,  44, -29,  -9,  -1,   5,  -6,  10,   0,   1,  -2,   4,  -1,   2,  -2,   4,   0,   0,  -1,   2 },
     {    4,   2,  -1,  10, -14,   1,  -1,   1,  -2,   0,   0,   0,  -1, -18, -23,  19, -14,  -5,   1,   1,   0,  -1,   0,   0,  -8,  -7,  10,  12, -22,  -8,   0,  -3,   0,   0,   0,   0,  53,   4,   3,  22,   6, -38,   3,  -6,   0,  -4,   1,  -2,  31, -14,  41,   2,  40, -10, -21,   1,   3,   0, -16,  34,   4,  29, -10,  -1, -11,  12, -20,   0,  18,   3,  11,  -6,   0,  -6,   1, -11,   6,   8,   5,   3,  -2,   1,  -4,  -5,  -2,  -1,   0,  -3,  -1,   1,  -1,  -2,   0,  -1,   1,  -2 },
     {    1,   6,  17,   5,   6,   0,  -3,  -1,  -2,  -1,  -1,  -1,  -5,  -8, -28, -31,  42,   2,   1,   2,  -2,   2,  -1,   1, -16,  -7,   6,  -9, -16,  13,  -4,   0,   1,   1,   0,   0,  18,  34,  -2,  10,  21,  -3,  -6,  -4,  -2,  -1,  -2,  -1, -26, -11,   4,  11,   4,  40,   2,   1, -19,   1, -45,  -8,   1,   2,  30,   0,   0,  -6,  15, -45,  -8,  -5,   1,  12,   2,  -2,   3,   9, -25,  -1,  -6,   0,  -1,   2,   3,   0,   3,  -1,   1,   3,   0,   0,   2,  -1,   1,  -1,   0,   2 },
     {    3,   2,  12,  59, -15,  -1,  -9,  -2,  -2,   0,  -1,   1,   1,   2,  50,  22,   5, -11,  -5,  -5,  -1,  -2,   0,   0,   7,  47,  26, -20,  19,  -5,  -3,  -4,  -1,  -2,  -1,  -2,  55,  23, -14, -13,   1,  13,  -2,  -1,  -1,   1,  -1,   0, -16,  17,   2, -12,  -7,   7,   6,   0, -10,  -9,   8,  -2,  -4,  -3,   3,   3,  -5,  -6,  -1,   3,   1,  -3,  -2,   3,   1,  -6,  -1,   4,   2,   2,  -4,   0,   0,  -1,  -3,   2,   1,  -1,  -2,   0,   0,  -1,  -2,   0,   0,  -2,  -1,   0 },
     {   -9,  -7, -30, -65,   9,   5,   8,   1,   4,   0,   1,   0,  -8, -15,  -2,  31, -34,   1,  -1,   2,   0,   0,   0,   0,   2,   6,  41,  28,  12, -18,  -2,  -7,  -2,  -3,  -1,  -2,  12,  14,  16, -10,  26,   9,  -6,   0,  -3,   0,  -1,   0, -42,  29,  15,  -8,  -2,  16,   5,  -1, -17,  -7,   2,   4,   0,  -1,   9,   2,  -2,  -3,  -5,  -9,   0,  -1,  -2,   4,   1,  -4,   1,   0,  -5,   2,  -1,   1,   0,  -1,  -2,   2,   1,  -2,   0,   2,   0,  -1,  -1,   1,   1,  -1,   0,  -1 },
     {    5,   4,  13,  26,   7,  -9,  -3,  -3,   1,  -2,   1,  -1,  -4, -11, -26, -27, -26,  24,   5,   3,   3,   1,   1,   1,   2,  -6,  16,  44, -17, -24,  -4,  -2,   0,  -2,   1,  -1,  28,  20,  18,  -5,  12,  12,  -5,  -3,  -1,   1,  -2,  -1, -23, -43,  -8,   4, -13,  -5,  17,   2,  32,   4,  11, -23,  -4, -20, -13,   4,   1,   4,  15,  31,  -9,   0, -11,  -5,   1,   3, -12,   5,  21,  -2,   2,  -4,  -3,   1,   3,  -2,  -1,   2,  -4,   3,  -2,   0,   1,  -1,   0,   1,  -4,   1 },
     {   -5,  -9, -11, -13, -16, -12,   4,   7,   2,   4,   1,   0,   9,  10,  15,  14,  17,   5,   0,  -1,  -2,  -2,  -1,  -2, -11, -16, -16, -18, -16, -17,   3,   7,   4,   4,   2,   0,  16,  19,  22,  21,  19,   7,  -3,  -4,  -4,  -4,  -2,  -1, -12, -23, -21, -24, -17, -18,   8,  10,  24,  23,  29,  23,  21,   3, -14, -10, -14, -25, -17, -24,  -4,  -1,  24,  19,  10,   6,  11,   1,  -5, -24, -40,   3,  -2,  -3,  -2,   5,   4,   4,   2,  -2,  -3,  -1,  -1,   3,   1,   2,   1,  -2 },
     {    0,   3,  17,   4, -31,   8,  -6,  -8,   4,   0,   1,   0,   2,   5,   2,  -2, -40,   7,   8,  -9,   1,   0,   0,   0,  -8, -14,  19, -24, -15, -19,  11,  -5,   0,  -1,  -2,   2, -17,  27,  12, -21,  22, -19, -14,   1,  -2,   6,  -5,   2,  45,  28, -18,  -7,   5,  15, -15,  -4, -17,  50,   0, -18,   2,   6,   1,   1,  -8,   2,  23,   2, -14,   4,  -6,   5,  -4,   4,   2,  17,   2,  -2,  -6,  -9,  -6,  -4,   6,   8,  -1,  -3,  -4,   7,  -2,  -2,   2,   0,   2,  -1,  -4,   2 },
     {    2,   2,  -6, -33,   4,  11,  -1,  -1,   0,   0,   1,   0,   4,  -2,  -7,  26, -13,   3,   1,  -6,   0,   0,   1,   0,   3,  13,  25, -12,  20,   6,   3,  -3,  -3,  -2,  -1,   0,   9,  32, -42,   0, -16,  25,   4,   2,   1,   2,  -1,   1,  59, -46, -16,   6, -14,   1,  18,  -1,  30, -14, -23,  -5,   5, -12,   3,  10,  -2,   3,  -2, -15,  -4,   3,  -6,   4,  -3,   3,   7,  -1, -11,  -3,   2,  -1,  -4,   1,   5,   1,  -2,   0,   1,   3,  -2,   1,   1,   0,   0,  -1,   0,   1 },
     {    1,  -5,   0,   5, -41,  35,  -8,   6,  -2,   1,  -1,   2,   4,  -1,   8,  14, -31,  33, -13,   6,  -3,   3,  -2,   2,   5,  -9,  -9,  -5, -46,  36,  -4,   2,   3,   2,   0,   0,   7, -29,  13,  -2, -11,  -1,   4,  -7,   4,  -1,   2,  -1,   1,  -6,  26, -14,  -1,   9,  16,  -6,   5, -40,  14, -16,   5, -14,  21,   6,  14, -16,   7, -17, -13,  -4,  -8,   9,   3,   1,   0,  14, -17,  -8,  -2,  -4,   0,   4,  -5,   7,   1,   3,  -1,   2,   0,   0,   0,   1,  -1,   1,   0,   1 },
     {    1,   1,  -5,  11,  52, -31,  -3,  -5,  -1,  -2,   2,  -2,   4,   6,  -2,  37,  31,  -7,  -9,   4,  -8,   2,  -3,   0,  -2,  -7,  18,  28, -35,  29, -21,   7,  -2,   1,  -1,   1,  -1,  15,  33,   2, -11,  13,  -1,  -9,   1,  -1,  -1,  -3,  44,  28,   4,  -6,  -8,   9,   7, -13,   7,   7,  16,  -4,  -1,   3,  18,   1,   2, -15,   6,   7,  -4,  -6,  -2,   7,  -6,  -2,  -1,   6,   9,   2,  -2,  -7,  -1,  -6,  -3,  -5,  -3,  -1,  -5,   0,  -1,  -1,   1,  -3,  -1,   0,  -1,   0 },
     {    2,  -2,   3,  -5,  -5,  -8,  20, -31,  18,  -8,   6,  -2,   3,  -2,   3,  -6,  13, -21,  32, -39,  10,  -4,   6,  -4,   2,  -4,   0,  -5,  -6,  -4,  29, -24,  10,  -9,  -1,   0,   3,  -8,   7, -10,  10, -14,  20, -20,   6,   7,  -8,   1,   5,   3,  10, -10,  15,  -9,  45, -25,  -1,  -1,  -2, -11,  15, -26,  25,  14,   8,   1,  -7,   5,  11,  -7,  13,   6,   6,  -7,  -3, -15,  13, -27,  13,  27,   4,  -7,   3,  -3,   1,  -2,  -7,   4,   0,  -3,   4,  -2,   1,  -1,  -3,   1 },
     {   -1,   2,   4,  -3,  13,   8, -12,  -3,   3,  -3,   1,   1,  -1,  -1,  -2,  10,  -3,  -1,   3,  -2,   2,  -3,   1,   1,   1,   2,  15,   2,  -3,  33,  -9,   8,  -5,   0,  -2,   0,   2,  -8,   6, -15, -29,  16,  22,  -6,   6,  -3,   5,  -1, -18, -30,  14, -33,   9, -39,  16,   8, -43,  55, -14, -12,   4,   4, -30,   6, -10,   7,  23, -15,   2,   8,   5,  -8,  11, -14,   7,  -1, -10,   5,   2,   5,  -1,   6,   0,  10,   1,  -4,  -2,   3,   2,   1,  -1,   3,   1,   0,  -2,   0 },
     {   -4,  -2,  -8, -11, -28, -27,   8,  -3,   5,   3,   1,  -1,   0,   5,  -1,  17,  22, -18,   4,   2, -10,   6,  -1,   2,  -2,   2,  -1,  -2, -27,   8,  -6,   5,   3,   1,   1,  -1,   2,   2,   0, -11,   7, -11,   4,   4,  -3,  -2,  -1,   0,  -9,   2, -11,  -5,  -7, -15,   9,   8, -18, -15, -30,   2,   0,  -9, -22,  23,  -3,  38, -30,   6, -27,   1, -42, -10,   1,  30,   3,  16,  21, -17,  -1, -43,   3,   6,   5,   0,   0,   7,   1,   8,   0,  -4,   2,  -5,   1,   2,   0,   0 },
     {    3,   3,  -1,  18,  49,   0,   5,   0,  -1,   0,  -6,  -4,   1,   0,  10,   5, -25,   2,  16,  -8,  12, -14,  15,  -4,   0,  -1,  -6, -17, -22, -19,  -6,  12,  -7,   7,  -3,   6,  -6, -11,  -4,  -5,   3, -12, -33,   6,  -9,  -3,   5,  -4,  -8,   7,   0,   0,   6,   0,  11, -13,  16, -16, -18,  10, -11,  -6, -19,   7,  -4,   1,   8, -19,   8,   1,  -4, -12,   9, -14,  43, -15, -14, -28,   6, -24,  19,  -6,   9,  38,  -5,  -6, -14,   1,  -4,   3, -10,   7,   1,   0,  -4,   0 },
     {   -2,  -2,  -5,   5,  24,  -2, -11,  -3,  10,  -4,   4,  -3,  -1,   0,   2,  -7, -10,  26, -20, -12,   8,  -4,   3,  -3, -11, -11, -22, -16,  14, -32,  -1,   0,   0,   1,  -1,   7,   3,  -2,  -5,  47,  11,  22,  -5, -16,  -7,  -1,  -3,  -2,  19, -13,  24,   2,  -9,   2,  24,  -7, -66,  -7,  17,   2,   1,  -9,   3,  21, -15,  -1, -13,   6,  -3,   0, -16,  -2, -16,  -1,  -8,  12,   6,   3,  -2, -15,  -6,  -3,   1,  -8,  -3,   0,  -1,  -1,  -2,  -1,  -6,   1,   4,   0,   1,   3 },
   },
   {
     {  101, -40, -14,  -1,  -3,  -1,  -1,   0,  -1,   0,   0,   0, -51,  37,   1,  -2,   0,   0,   0,   0,   0,   0,   0,   0, -13,  -7,  12,   0,   1,   0,   0,   0,   0,   0,   0,   0,   5,  -4,  -3,   3,  -1,   0,   0,   0,   0,   0,   0,   0,  -4,   2,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -26,  32,  -7,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -80,   4,  31,   0,   4,   0,   2,   0,   1,   0,   1,   0,  56, -54,  -9,   9,   0,   2,   0,   1,   0,   1,   0,   0,  11,  19, -20,  -4,   0,  -1,   0,   0,   0,   0,   0,   0,   2,   2,   7,  -5,  -1,   0,   0,   0,   4,   0,  -1,   2,  -2,   0,   0,   0,   1,   0,   0,   0,   1,  -1,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0 },
     {   35,  87, -54,  -8,  -8,  -3,  -3,  -2,  -2,  -1,  -1,  -1,  20, -17,  39, -14,  -1,  -3,  -1,  -1,   0,  -1,   0,   0, -33,  -8,  16,  16,  -2,   1,   0,   0,   0,   0,   0,   0,  -7, -11,  -1,   7,   6,   0,   1,   0,   0,   0,   0,   0,   3,  -7,  -4,   1,   2,   2,   0,   1,  -2,   1,  -2,  -1,   1,   1,   0,   0,   0,  -3,   0,   0,   0,   1,   0,   0,  -1,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
     {  -38,   6,   7,   5,   0,   2,   0,   0,   0,   0,   0,   0,  -2,  59, -15, -12,  -1,  -5,   0,  -2,   0,  -1,   0,  -1, -35, -29,  57,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  34, -45, -23,  27,   1,   4,   0,   1,   0,   1,   0,   1,  -1,  18, -21, -11,   8,   0,   2,   0,   6,  -5,   7,  -6,  -4,   2,   0,   1,   1,   3,  -2,   2,  -2,  -1,   0,   0,   2,  -2,   1,   0,   0,  -1,   0,   0,   0,   1,  -1,   0,   1,  -1,   0,   0,   0,   1,  -1,   0,   1,  -1,   0,   0 },
     {   -2, -24,   9,   3,   4,  -1,   2,   0,   1,   0,   0,   0, -11, -55,  26,   8,   3,   4,   0,   3,   0,   1,   0,   1, -58,  -5, -24,  23,   4,   4,   2,   2,   1,   1,   1,   1,  49,   0, -45, -19,   9,   0,   2,   0,   1,   0,   1,   0,  -1,  32,   7, -27, -11,   2,  -1,   0,   3,   2,  13,   6, -11,  -4,   0,  -1,   1,   4,   0,   4,   2,  -4,  -1,  -1,   2,   2,   0,   0,   1,   0,  -1,   0,   0,   2,   1,   0,   1,   1,   0,   0,   0,   1,   0,   0,   1,   1,   0,   0 },
     {  -12,  17,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0, -57, -24,   3,  18,   6,   4,   2,   2,   1,   1,   1,   1, -11,  53,  -9, -32,   6,  -4,   2,  -2,   1,  -1,   0,  -1,  34, -44,  44,   4, -15,   3,  -4,   1,  -1,   0,  -1,   0,   8,   0, -33,  25,   3,  -3,   1,   0,  -3,   7,   3, -15,   8,   0,   0,   0,   5,  -5,   2,   2,  -6,   3,   0,   0,   0,   3,  -1,   0,   0,  -2,   1,   0,   2,  -1,   0,   1,   0,   1,   0,  -1,   1,  -1,   0,   1,   0,   0,   0,   0 },
     {  -24, -17, -67,  46,  -1,  11,   1,   4,   0,   2,   0,   1,   0,  21,  -4, -34,  14,  -4,   3,  -1,   1,  -1,   1,   0, -12,  10,   0, -16,  -1,   3,   0,   0,   1,   0,   0,   0,  35,  21,  14, -30, -14,  -3,  -1,  -1,   0,   0,   0,   0,  -4,  -4,  41,  17, -18,  -5,  -4,  -1,  -2,  -6, -11,  22,  11,  -6,  -1,  -1,  -1,  -4,   3,  -6,   7,   4,  -2,   0,   1,  -2,  -2,   2,  -3,   3,   1,   0,  -1,  -1,   2,   0,   1,  -1,  -1,   1,   0,   0,   1,   0,   0,  -1,  -1,   1 },
     {    7,  43,  32, -38,   1,  -8,   0,  -3,  -1,  -2,   0,  -1,  -3,  34, -37,  34, -13,   4,  -4,   1,  -1,   0,  -1,   0, -31, -19, -21,  -3,  27,  -3,   6,  -1,   3,   0,   1,   0,  17,   7,  16, -46,  -4,   5,  -1,   1,   0,   1,   0,   0,   2,  -3,  32,  15, -27,  -2,  -2,  -1,   4, -10,  -3,  20,   9,  -9,   0,   0,  -2,  -1,  -2,  -3,   7,   4,  -3,   0,   2,  -4,   0,   0,  -1,   3,   1,  -1,  -1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },
     {   18,  24, -20,  -3,  -5,   1,  -2,   1,  -1,   0,  -1,   0,   1, -33, -56,  18,  10,   4,   4,   2,   2,   1,   1,   1,  25,   6,  25, -47,  -4,  -1,  -1,   0,   0,   0,   0,   0,  15,  15, -16,  19, -14,  -1,  -3,  -1,  -1,   0,   0,   0, -25,  30,  17, -28,   8,  -5,   2,  -1,   6, -21,  33,  15, -17,   2,  -4,   1,  -7,   5, -13,  14,   9,  -6,   1,  -2,   1,  -4,   7,  -5,   4,   3,  -2,   1,  -3,   1,  -3,   2,   1,  -1,   3,  -1,  -1,   1,  -1,   1,   0,  -1,   2,   0 },
     {   14,  -1,  23,  -9,  -2,  -2,   0,  -2,  -1,   0,   0,   0,  30,  -3,  31,  -3,  -7,  -4,  -4,  -1,  -2,   0,  -1,  -1,  15, -18,  -4,   0,  -3,   5,  -1,   1,  -1,   1,   0,   1,  54,  20,   9,  18, -28,  -4,  -6,  -1,  -2,   0,  -2,   0, -59,  17,  -6,  44,  23, -11,   1,  -3,   0, -18, -14, -14,  27,  10,  -1,   1,  -7,   3,  -6,  -6,  -7,  11,   3,   1,   0,   0,  -4,  -4,  -1,  -2,   5,   0,  -3,   1,   1,  -1,  -1,   0,  -2,  -1,  -2,   0,   0,  -1,   0,  -1,  -1,  -1 },
     {    1,  13,  33, -24,   2,  -8,   1,  -3,   0,  -1,   0,   0, -12,  13,  23, -27,  -3,   2,   0,   1,  -1,   0,  -1,   0,  19,  65,  22,  18, -34,  -8,  -7,  -3,  -2,  -1,  -1,  -1,  19,  -6, -15,  12,  25,  -7,  -1,  -1,   0,  -1,   1,  -1,  -8, -10,  33,  -5, -12,  13,  -4,   1,  -8, -18,  -7,  41,  -1,  -6,   3,  -2,  -2,  -2, -16,  -8,  20,   1,   0,   0,   0,  -6,   1,  -1,  -3,   7,   2,   1,   0,  -1,  -6,  -2,   1,  -1,   0,   0,   0,   0,  -2,  -1,  -1,   0,  -1,   1 },
     {    5,  11, -11,  21, -11,   2,  -3,   1,  -1,   0,  -1,   0,  15,  38,  10,  -7,  -2,  -6,  -2,  -2,  -1,  -1,   0,   0,  30,  19, -58,  18,  16,  -2,   2,  -1,   1,  -1,   0,  -1,  -2, -44,   6, -21,  17,   7,   3,   2,   1,   2,   0,   1, -43,  28,  -8, -28,  -4,   6,   5,   1,  13, -19,  28,  -9, -17,  -3,  -1,   2,  -1,   0,   0,  15,  -5,  -4,  -3,   0,  -2,   1,   1,   0,   5,  -1,  -2,  -1,   1,  -1,   2,   2,  -1,  -1,   0,   0,   0,  -1,   0,   1,   0,  -1,   1,   0 },
     {  -21,  -8, -35, -34,  33,   2,   7,   1,   3,   0,   2,   1, -16,  17,  -5,  20, -24,   9,  -3,   1,  -1,   1,  -1,   0, -23,  21,  -6,  15,  -3,  -8,   3,  -1,   1,  -1,   0,  -1,   6,  34,   6,   5,  16, -16,  -3,  -3,   0,  -2,   0,  -2, -25, -34,   9, -14,  40,  13,  -2,   1,  21, -21,  -1, -29, -22,  26,   3,   2,   0,   6,   0,   8, -21, -11,   8,   0,   6,  -2,   3,  -1,   3,  -8,  -3,   3,   0,   0,   1,   3,   3,  -1,   0,  -1,   1,   0,   1,   2,   2,   0,   0,  -1 },
     {    5,  23,  42,  48, -49,   5, -11,   1,  -4,  -1,  -2,   0, -13,   4,  -4,  -3,  38, -17,   5,  -3,   2,  -1,   1,  -1, -28,   7, -20, -15,   7,  18,  -2,   2,  -1,   1,   0,   1,  -3,  12,  -8,  14,  -1,  -5,   7,  -4,   2,  -1,   0,  -1,  -9, -39,  15, -15,  28,   3,   0,   3,  17,  -9, -22,  -5, -15,  18,   1,   2,  -4,   9,  -1, -12,  -8,  -7,   8,   0,   7,  -4,   6,   1,  -6,  -4,  -5,   2,  -1,   3,  -1,   2,   3,  -1,   2,  -1,  -1,   1,   0,   1,   2,  -1,   0,   0 },
     {   -7,   6,   4,  16,  -4,   1,  -3,   0,  -2,   0,   0,   0,  -5,  24,  28,  19, -12,  -7,  -5,  -3,  -1,  -2,  -1,  -1, -12,  34,   0,  -6,  20,  -7,  -2,  -1,  -2,   0,  -1,   0, -25,  32, -26,   6, -13,   7,   0,  -1,   1,   0,   1,   0,  38,  30,   9,  -1,  17, -30,   4,  -9,  -9, -31,  36,  -9,  36,  15, -12,   5,  -1, -16, -12,   6, -11,  30,   3,  -1,  -2,  -5,  -7,  -5,  -3,  -2,  13,   0,   3,  -5,  -1,  -4,  -1,  -2,  -1,   0,   2,  -1,   0,  -1,  -1,  -1,   0,  -1 },
     {   10,  13,  14,   0,  -6,  -2,  -1,  -2,  -1,  -1,  -1,   0,  17,  22,  -1, -12,   5,  -2,  -1,  -1,  -1,  -1,   0,   0,  10,   3,  -9,  -7, -10,   6,   1,   1,   1,   1,   0,   0,  45,  33,   3,  -4,  -2, -12,   1,  -4,  -1,  -2,   0,  -1,  26, -16, -11, -10,   9,  13,  -5,   3, -55,  29,  33, -15, -27,   2,   6,  -1,   6, -38,  19,  33,  -5, -13,  -2,   1,  -7,   7, -18,   7,  14,  -1,  -3,  -1,   1,  -7,   4,  -4,  -2,   1,  -3,   0,  -1,  -2,   2,   0,  -1,   0,  -1,  -1 },
     {   -9, -10,   9, -34,  -2,   9,   5,   0,   2,   1,   1,   0,  -4,   9,  28, -36,  11,  -9,   0,  -2,   1,   0,   0,  -1, -24, -12, -19, -12, -14,  17,   1,   5,   0,   1,   0,   1, -20,  10,  40,  15, -41, -13,   1,  -2,   0,   0,   0,   0,  16,  18,  -3,  -4,   6, -10,  -4,  -1,  34,  -6,  20,  23, -28,  -6,  -3,  -1, -13,   7, -15,  31,  26, -13,  -3,  -2,   1,  -3,  -2, -13,  10,  12,  -3,   0,  -1,  -1,  -2,   7,   0,  -1,  -2,  -2,   0,  -1,  -1,   3,   1,  -1,   0,  -1 },
     {    2,  12,   3,   0, -13,   4,   0,   0,   0,  -1,   0,  -1, -14,  -9, -54, -30,  17,   8,   5,   5,   2,   2,   1,   1,  -6,  -5, -14,  55, -40,  -2,  -1,  -1,   0,   0,   0,   0,   2,   9,  30,   4,  17, -27,   2,  -4,   0,  -1,  -1,  -1,  14,  41,  -3,  -2,  18,  -7,  -9,  -3,   6,  -5,   5,  -9,  27,  11,  -2,  -3,  -2, -11,   9, -15, -11,  19,   3,   3,  -6,   0,  -4,   7,  -9,  -5,   5,   0,   1,  -2,   3,  -3,  -1,  -2,   0,   3,   1,   0,   0,  -2,   0,  -1,   1,   0 },
     {    3,   5,  14, -18,  29, -13,   0,  -7,   3,  -2,   2,   0,  -8, -10,  -8, -56,  17,  10,  12,   2,   0,   1,   1,   1,  -4,  -1,  -2, -18,  -1,  10,  -2,   5,  -2,   1,   0,   1, -19, -18, -21, -20, -18,  20,   3,   5,   2,   1,   2,   1,  -5,   2,  20,  17, -13, -13,  -2,   1, -19, -24,   3, -16,  -4,  28,  -4,   2,  18,  10,   0,   9, -51, -16,  12,  -4,   0,   4,   1,  21,  16, -27,  -4,   3,   3,   2,   0,  -2,   1,   3,   0,   4,   2,   1,   1,  -1,  -1,   1,  -1,   4 },
     {   -7,   9,   7,  16,  20, -22,   0,  -6,   2,  -2,   0,  -1, -16,   6, -12, -11,  -3,  18,  -3,   4,  -3,   2,   0,   2, -28,  -1, -18, -13, -29,  11,   9,   2,   3,   2,   1,   1, -26,  16, -21,  19,   1,  -7,  -2,   3,  -1,   0,   1,  -1, -51, -23, -15,  10, -18,  -7,  10,  -1,   5,  15,  29,  -6,  17, -37, -10,  -6,  32, -22,   1,   4,  19,  18, -17,  -6,   7,  10,  -4,  -2,  -1,   6,  11,  -2,   3,  -6,   6,  -3,   6,   3,   0,  -1,   1,  -2,   2,  -1,   1,   1,  -2,   0 },
     {   -5,  -3,   6, -25, -33,  24,   0,   4,  -1,   1,  -1,   2,  -8,  -9,  -4, -26,  15, -20,  11,  -2,   5,   0,   1,  -1,   0,  19,  34,  16,  23, -22, -15,  -3,  -2,  -2,  -1,   0,   0,  34,  -1, -23,   5,  35, -12,  -3,  -3,   0,   1,   0, -21, -11, -24,   2, -22,   0,  23,  -1,  33,  10,  16, -17,   0, -18,  -6,   8, -19,  -2,  22,  14, -10,  13, -11,  -2,   6,  -3,  -8,  14,   1,  -2,   7,  -5,  -4,  -2,  -1,   1,   1,  -2,  -3,   2,  -1,   0,   1,   0,   2,   0,  -1,   2 },
     {    6,   8,  -4, -18,  11,  -2,   6,  -1,  -1,  -1,   1,  -1,   8,  10,  -3, -10, -16,   5,  -6,   4,   1,   1,   0,  -1,  13,  17, -23, -16,  -3,   6,   1,   0,   0,   0,   0,   0,  13, -16, -33,  18, -26,  -8,  13,  -2,   4,   0,   1,   0,  12,   0,  26, -17,   8, -26,  -5,   0,  38,  48, -17, -10,  13,  -4, -12,  -1, -27,  -2,  54,  -7,  -4,  11,  -3,  -1,   0, -13,  -2,  23,  -7,   5,  -1,  -2,  -7,   4,  -3,   2,   0,  -6,   1,   1,  -4,   1,  -1,   2,   1,  -2,   2,   1 },
     {  -11, -11, -22, -37, -49,  25,   3,   6,   2,   5,   1,   3,  -8,  -1, -20,  -6, -18, -23,  17,  -1,   4,  -1,   2,   1,  -5,   3, -29,  18,  19,  -4, -12,   4,  -2,   2,  -1,   1,  -6, -13, -24,  36, -10,   6,  -8,  -5,   2,  -2,   2,  -1,  -6,  -2,  22,  21,   9, -13,   6,  -9, -38,   6,  -2,  -4,   4,  -5,   5,   1,  34,   0,  -1,  -7,  16, -22,  -6,  -4,   3,   2,   8,  -8,  12,  12,  -6,   0,   3,   0,   4,  -4,   3,   2,   3,  -1,   1,  -1,   2,  -2,  -1,   0,   0,   1 },
     {   -6, -13, -16, -45,  12,   6,   7,   1,   1,   0,   4,   1,   8,   6,   7,  14,  66, -25,   1,  -7,   0,  -1,  -1,  -3,   2,  -8,  -8, -10,   4,  25,  -5,   4,  -5,   2,   0,   1,   1, -16,   1,  -1,   8,   3,   1,   1,  -2,  -1,   2,   0, -11, -13,   4, -21,   5,   3,  -2,   5, -22,  11,  10,   8,   5,  11, -15,   2,   1, -16, -12, -18,  -3,  43,   5,  -3,   4,   9,  -2,  -4, -41,  -7,  20,  -1,   0,   1,   4,   9,   2,   0,  -1,   0,   0,   2,   2,   2,   0,   3,  -1,   1 },
     {    1,  -6,   7,   5, -13, -18,  17,  -3,   4,  -1,   1,  -1,  -1, -14,   0, -40, -30,  22,  -7,  11,   0,   4,   1,   2,  -4, -21, -12, -21,  43, -36,   4,  -3,   5,  -2,   1,   0,   7,  -6,  23,  36,  26,   6, -24,  -2,  -5,   0,  -2,   0,   4,  -8,  27, -14,   2,  22,   0,  -3,   5,  17,  10,  10,   4,  11,  -6,   2,  -5,  -7, -20,   8, -12,  30,   9,   0,   0,  10,  -8,  -6, -13, -11,  12,   1,  -2,  -2,   3,   3,  -1,   4,  -2,   1,  -2,  -1,   0,   3,   0,   1,  -1,  -1 },
     {    4,   3, -10,  -4,   8,  -1,  -1,   0,   0,   3,   1,   2,   7,   9, -11, -16,  -4,  -3,  15,  -4,   0,   0,   0,   1,   2,  -3, -38, -17, -11,   5,   1,   6,   1,   2,   2,   2,  15,  11, -25,  12,  25,   2, -22,   3,  -3,  -1,   1,   0,  31, -14, -39,  29, -14,  11,   6,  -4,  -6, -18,  24,  -4,   0,  -3,   1,  -3, -45,  35,  -8, -22,   9,  -1,   4,  -6,  18, -23,  31,   1, -11,  -6,   4,   7,  -4,   4, -12,   6,   7,   1,   7,  -8,   0,   0,   3,   1,   0,   1,  -2,  -1 },
     {   -6,   0,  -1, -15, -15,  -8,   8,   0,   1,   0,   1,  -1,  -1,  11,  18,   5,  10, -19,  -2,  -2,   0,  -3,   1,  -1, -10,   4,   1, -48,  -5,   2,   7,   0,   0,   3,  -2,   2, -18,  21,  -1,   5,  38, -19, -18,   4,  -4,   3,  -4,   1,  -6,  56,   6,  14, -15,  36, -12, -12,  10,   6, -17, -15,  -5,   2,  27,  -5,  23, -11,  25, -15,  -6, -13,  -4,  11,  -7,  -7,   3,  22,  -7,  -1,  -6,  -1,   0,  -4,  -3,   3,   1,  -3,   0,   5,   0,   1,  -2,  -1,   1,  -2,   0,   2 },
     {    1,  -9,   2,  -4, -28,  19, -11,  10,  -2,   6,  -1,   1,   4, -10,   6,   8, -15, -12,   9,  -3,   4,  -2,   2,   0,   5, -13,  11, -19, -23,   0,   5,  -2,   4,  -2,   1,   1,  -2, -28, -14, -31,  34, -45,   4,   6,   2,   3,   1,   2,   0, -21,  16,  35,   8,  -8, -29,   4,  18,   5,  30,  -6,   3,  18, -22,  -8,   2,  18,  -5,  31,  -3,  18,  -3,  -9,  -2,   3, -16,   3,  11,  10,  16,  -5,  -3,  -2,   2,  -8,   0,   0,  -6,   3,  -1,   2,  -1,  -3,   1,   0,   0,   0 },
     {    7,   1,   6,   1,  20,   3,  -1,   2,  -1,   0,  -2,  -3,   4,  -7,   0, -10,   3, -13,   5,   1,   2,  -2,   1,   1,   4,  -9,  -5,  -7,   7, -10,  -1,   1,   2,   1,  -1,   0,  14,   0,  12,   8,  -1,  -5,  -8,   2,   3,   3,  -5,  -1,  21, -11,   0, -22,  12,   9,  10,  -1,  -7, -48, -13, -13,  -5, -26, -16,  12,  41,  36,  26,   3,   4,  22, -39, -12, -28,  -8,  -3,  24,  -6,  35,  21, -11,   4,   1,  -2,  -7,  -6,   2,   6,   8,  -2,   0,  -8,   2,  -2,   0,   2,   1 },
     {   -5, -10, -10, -18, -50,  11,  -7,  10,  -1,   5,   0,   2,   9,  10,  17,  23,  -4,  35,  -9,  -3,  -2,  -4,   0,  -1,  -1,  -8,  -8, -14, -43, -11,  22,   3,   4,   0,   1,   1,   9,   3,  11,  14,  -8,   1,   3,   4,   0,  -1,   0,  -2,   7, -12,  -3, -28, -30, -13,  20,   5,  -2, -19,  -5,   7,   6, -21, -14,   3,   7,   6,  -3,  -6, -32,   3,   6,   0,  -7,   8,   6,  21,   8, -43,  -6,  -2,   3,   5,   0,   0,  -3,   0,  -1,   5,   0,   2,  -2,   0,  -1,   0,   3,  -2 },
     {    6,   6,  10,  13,  40,  42, -49,   7,  -4,   0,  -3,   2,   4,   2,   4,  16,  10, -12,  33, -18,   4,  -5,   1,  -3,  -5,  -9, -11,   0, -29, -36,  12,  18,  -3,   5,  -2,   1,   0,  -9,   8,  16,   4,  28, -37,   3,   2,  -3,   0,  -1,   3,   5,  29,   4,   3,   4,  15, -18,   9,  16,  -2,  -1,  -2,  -1,   4,   7,  -1, -14,  -3,   8,  -3,  -3,   3,   5,   9,  -3, -16,  -3,   5,  -8,   1,   2,  -5,   1,   0,  -7,   6,  -2,   2,  -2,  -2,   0,   1,  -1,   0,   2,  -2,   2 },
     {   -1,   3,   6,  -8,   2,  26, -24,   6,   0,  -1,   1,  -3,  -7,  -5,  -1, -28, -37, -16,   5,   0,   6,   2,   2,   0,  -7,  -3,  -3, -26,  -3, -29,  23,   3,   2,  -1,   0,  -1, -10,   2, -11, -31,  -9,  -9,   8,   5,   4,   3,   1,   3, -14,  -1, -34, -19,  27,  -7, -16,   3, -20, -13,  -9,  27,  -1,   3,   1,  -1, -13, -23,   6, -29,   3,   3, -10,  10,   2,  29,   7,   4, -25,   4, -21, -10,   8,   1,  14,  19,  -2,  -1,   0,   9,   3,  -1,  -2,   3,   3,   1,   2,  -2 },
   },
   {
     {   99,  -6, -31,  -7,  -4,  -2,  -2,  -1,  -1,   0,  -1,   0, -65,  19,  23,   0,   1,   0,   1,   0,   0,   0,   0,   0,   2, -16,   0,   6,   2,   1,   0,   1,   0,   0,   0,   0,   0,   5,  -3,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   2,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   20, -85,  15,  24,   4,   4,   1,   2,   1,   1,   0,   1,  27,  60, -39, -21,  -1,  -1,  -1,  -1,  -1,   0,   0,   0, -20,  10,  27,  -4,  -6,  -2,  -1,  -1,   0,   0,   0,   0,  -4, -10,  -3,  10,   3,   0,   0,   0,   0,   0,   0,   0,  -2,   4,   0,  -4,   1,   0,   0,   0,  -1,  -2,   1,   2,   0,   0,   0,   0,  -1,   1,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,  -1,   1,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },
     {   39,  43,  -5, -26,  -7,  -2,  -1,  -2,  -1,  -1,  -1,  -1,  55,  -8, -10,  11,  -6,  -3,  -2,   0,  -1,   0,  -1,   0, -82,  -4,  28,   8,   8,   1,   1,   0,   1,   0,   1,   0,  23, -16, -14,   3,   1,   3,   0,   1,   0,   0,   0,   0, -10,   9,   0,  -3,   0,   0,   1,   0,   3,  -4,   1,   1,   0,   0,   0,   0,  -3,   1,  -1,   0,   1,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   1,  -1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0 },
     {   41, -16,  29, -16,  -8,  -2,  -2,  -1,  -1,  -1,   0,   0,   5, -47, -35,  38,   9,   3,   1,   3,   1,   1,   0,   1,   6,  63, -24, -38,   2,   1,   1,  -2,   0,  -1,   0,  -1, -16,  -5,  38,   5, -12,  -3,  -1,   0,  -1,   0,   0,   0,  -1,  -3, -11,  10,   7,  -1,   0,   0,  -3,   2,   5,  -5,   0,   2,   0,   0,   0,  -1,  -2,   2,   0,   0,   0,   0,  -1,   1,   2,  -2,  -1,   0,   0,   0,  -1,   0,  -1,   1,  -1,   0,   1,  -1,   0,   0,   0,   1,   0,   0,   1,  -1 },
     {   -6,   9,  73,  -9, -26,  -4,  -3,  -2,  -2,  -1,  -1,  -1, -42,   6, -47,  23,  27,  -2,   1,   1,   1,   0,   1,   0,  -6, -48,   8, -10,   1,  12,   3,   2,   1,   2,   0,   1,  30,  21, -20,  -4,  -9,  -5,   1,   0,   0,   0,   0,   0,  -3,   5,  11,  -4,   2,  -2,  -2,   0,   5,  -2,  -2,   4,  -2,   1,   0,   0,  -1,   2,  -1,  -3,   2,  -1,   0,   0,   3,   0,   1,   1,  -2,   0,   0,   0,   0,   1,   0,   0,   2,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0 },
     {    3, -53,   3,  -1,  15,   3,   3,   1,   1,   0,   1,   0,  18, -28,  23,  13, -15,   1,  -2,   1,  -1,   0,  -1,   0, -26, -12, -55,  20,  19,   3,   3,   2,   1,   1,   1,   1,  28,  44, -11, -47,  -8,   2,   1,  -1,   0,   0,   0,   0, -12,   3,  29,  13, -12,  -5,  -3,  -1,   2,   1, -10,   2,   7,   0,   0,  -1,  -3,   2,   5,  -2,  -3,   0,   0,   0,   2,   1,  -2,   1,   1,   0,   0,   0,  -2,   1,   2,   0,   1,   0,  -1,   0,  -1,   1,   1,   0,   0,   1,  -1,   0 },
     {   24,  25,  34,  -5, -15,  -8,  -5,  -2,  -1,  -1,  -1,  -1,  43,  37,   1,  -5,  -6,  -5,  -4,  -3,  -2,  -1,  -1,  -1,  12, -39, -36,  19,  18,   7,   2,   2,   1,   1,   0,   1, -62,  10,  26, -17,  -4,   4,   4,   1,   1,   0,   1,   0,  26, -15, -11,  13,  -3,  -4,  -1,   1, -15,   8,   1,  -5,   5,   0,   0,  -1,   6,  -6,   0,   2,  -1,   2,   0,   0,  -6,   2,  -1,  -1,   1,   0,   1,   0,   2,  -2,   0,   0,  -3,   1,   0,   0,   1,  -1,   0,   0,  -2,   0,   0,   0 },
     {  -21,  12, -29, -15,   4,   9,   1,   3,   1,   1,   0,   1,  -7,  53,  -4,  15, -11, -14,  -1,  -3,  -1,  -1,  -1,  -1,  -8,  -3,  -3, -37,  21,   6,   2,   0,   2,   0,   1,   0,  36,   5,  50,  -4, -39,  -4,  -3,  -1,  -2,   0,  -1,   0, -18, -18,   0,  45,  18,  -9,  -3,  -1,   6,  -6, -11, -16,  11,  10,   1,   1,  -3,  -3,   2,   4,  -6,   0,   1,   0,   3,  -2,  -2,  -2,   2,   1,   0,   0,  -2,  -1,   1,   2,   1,  -1,  -1,  -1,   0,  -1,   0,   1,   1,  -1,  -1,  -1 },
     {  -18, -24,  24, -52,  19,  12,   3,   1,   2,   1,   1,   0, -19,  21,  32,  49, -32, -15,  -3,  -2,  -2,  -1,  -1,   0,  -3,   1,  -5,   8,  30,  -1,  -6,  -1,  -1,  -1,   0,   0,  -8, -48, -19,   6,  -1,  17,   6,   3,   1,   2,   1,   1,  14,  24, -24, -17,   0,  -6,   3,   1,   1,  -2,  15,  -3,  -6,   2,  -3,   0,   3,   4,  -6,   4,   0,  -2,   1,  -1,   0,  -1,   3,  -4,   0,   1,   0,   1,   2,   1,  -3,   2,   0,   0,   1,  -2,   1,   1,  -1,   1,   0,   0,   1,  -1 },
     {   -1,  14,  33,  60,   7, -27,  -9,  -5,  -3,  -3,  -2,  -1, -24,   4,  24, -32,  -4,  22,   0,   0,   0,   1,   0,   0, -39,   6, -44, -14,   8,   2,  12,   3,   3,   1,   2,   1,  18, -47,  18,   3,  -9,   0,  -2,   4,   1,   1,   0,   1,   7,  25, -13,   1,  -2,  -4,  -1,  -2,   5,  -6,   6,  -2,   4,   1,  -1,   0,   1,   4,  -4,   2,  -3,   1,   1,  -1,   2,  -1,   1,   0,   1,  -1,   1,   0,   1,   1,  -1,   0,   1,  -1,   0,   0,   1,   1,  -1,   0,   1,   0,   0,   0 },
     {  -12,  19, -26,  -2,   5,   5,   0,   0,   0,   0,   0,   0, -27,  18, -51,   4,  -6,   5,   6,   2,   1,   1,   0,   1, -13,  25,  -4,   2,  15, -17,   0,  -3,   2,  -2,   1,  -1,  -9, -13,   8, -53,  15,  15,   2,   1,   1,   1,   1,   0,  37,  11,  33,  -3, -45,  -7,   0,   1, -11,  -7,  10,  30,  13, -13,  -5,  -2,   7,  -6,  -3,  -6,   6,   8,  -2,   0,  -3,  -3,   2,   4,  -2,   0,   1,   0,   4,  -1,   0,  -1,  -1,  -1,   1,   2,   2,  -1,   0,  -1,   0,  -1,   1,   1 },
     {    9,  20,  47, -16,  11, -14,  -5,  -4,  -1,  -2,  -1,  -1,  -5,  21,   5, -15, -28,  12,   3,   2,   0,   1,   0,   0,  23,  39,   7,  34,  -1, -27,  -6,  -4,  -2,  -2,  -1,  -1,  14,  -8,  -2,  -9,  27,  16,  -5,  -2,  -1,   0,  -1,   0, -49, -32,  12,  18, -11,   6,   9,   3,  23,  -2, -26,  -4,  13,  -4,  -2,   1, -10,   2,   8,  -7,  -6,   6,   0,  -1,   5,  -4,  -5,   4,   1,  -2,   2,   0,  -3,   0,   1,  -3,   2,  -1,  -2,   1,  -2,   0,   1,  -1,   1,  -1,  -1,   1 },
     {   -9, -15,  11, -22,  23,  -1,   2,  -1,   1,  -1,   0,   0, -27, -27, -11, -25, -15,  21,   9,   5,   1,   3,   1,   1, -29, -10,  30,  39, -13, -23,  -3,   2,   1,   0,   0,   0,   9,  20,  32,  -2,  -2,   3, -12,  -4,  -2,  -1,  -1,  -1,  37,  -9, -43,  30,   9,   1,   3,  -3, -27,  10,   6, -23,   6,   3,   4,   1,  14,  -5,  -6,   5,  -8,  -2,   1,   1,  -5,   6,   1,  -2,   4,  -2,  -2,  -1,   5,  -2,  -1,   2,  -2,   2,   0,  -2,   2,  -1,  -1,   1,  -1,   1,   0,   0 },
     {   14,  -2,  13, -14,  10,  -3,  -4,  -1,   0,  -1,   0,   0,  16, -19,  11, -23, -10,   0,   6,  -2,   1,   1,   0,   0,  33, -12,  27, -15,  -7,   2,  -6,   2,  -2,   1,  -1,   0,  -1, -12, -12,  12, -33,   9,   2,   4,   1,   2,   0,   1,   6,  53,  24,  42, -10, -31,  -7,  -6, -18, -22, -12,  20,  44,   8,  -3,  -2,  -1,   1, -12, -17,  -4,  12,   6,   0,  -5,  -1,   0,   4,   0,  -3,   2,   1,   1,   2,  -3,  -3,  -3,  -1,   0,   2,   0,   1,  -1,  -1,  -1,   0,   0,   1 },
     {   -9,  -9,  16, -46, -21,   4,  20,   3,   4,   1,   1,   0,  -2,  19,  41, -14,  33,  -7, -22,  -3,  -3,  -1,  -1,  -1, -19,  19, -12,  -4, -47,  13,  18,  -1,   0,   1,   0,   0,  10,  -5,  18,   6,  15, -29, -10,   4,  -1,   0,   0,   0,  19,  -5,  13,  13, -17,  19,  -9,  -4, -11,   2,  -7,  28,   4, -19,   7,  -1,   3,  -4,  -4,  -6,  20,  11,  -9,   0,  -2,  -1,  -5,   1, -10,   4,   7,  -1,   1,  -3,  -1,   0,  -1,  -1,  -1,   1,   1,   0,  -1,   0,   0,   0,  -1,   1 },
     {   10,  10,   0, -22,  47,  -6, -11,  -1,   0,  -2,   0,   0,  14,  11, -19,  -5, -40,  22,  13,  -2,  -1,   2,   0,   0,  18, -25, -26, -22, -16, -19,   9,  11,   4,   2,   2,   1,  19,  23,  20,  28,  12, -17, -21,  -7,  -3,  -1,  -2,  -1,  -8,  20,   8, -29,   1,  20,   5,  -4,  -5, -15,  23,  24, -19, -12,   1,   2,  -5,   0, -13,   8,  19,  -1,  -2,  -2,  -2,  -3,   5,  -5,  -2,   3,   0,   1,  -1,   0,  -4,   3,   0,  -1,   2,  -2,  -1,   0,  -1,   2,  -1,  -1,   1,   0 },
     {   -5,   6,  10,  -3,  48,  14, -12,  -8,  -3,  -3,  -1,  -1, -18,  -3,  -1,   3, -19,  -8,  13,   5,   1,   1,   1,   0, -34,  -2,  -1, -27, -22,  14,   2,   4,   3,   2,   2,   1, -46,   0, -15,  14,  12, -20,   5,  -1,   2,   0,   2,   0,  22, -19,  40,   8,  11,   1, -17,   0,  14,   4, -41, -11,  -4,  21,  11,  -2,   0,   6,  18,  -9, -17, -15,   3,   4,   6,  -2,  -1,  12,   4,   0,  -6,  -2,   0,   1,   0,  -3,   2,  -2,  -1,   2,   1,   1,   0,   0,   2,   0,   0,   1 },
     {    1,  15,   5,  13, -12,  -3,  -3,   1,  -1,   0,  -1,   0,   1,  25,   6,  18,  -4,  -2,  -7,  -3,  -2,  -1,   0,   0,   4,  40, -10,  26,  -3,   0,  -2,  -4,  -2,  -2,   0,  -1,  15,  27, -14,   3,  12, -17,   2,  -1,   2,  -2,   0,  -2,  17,  13,  14, -13,  41,   3,  -7,  -5, -48, -54,   3, -14,   3,  33,  11,   6,  18,   1, -15, -11, -22,  -8,   4,   3,  -6,   0,   9,   2,   2,  -2,  -3,  -2,   5,  -2,  -1,  -2,  -3,  -2,   2,   0,   1,  -2,  -1,  -1,  -2,  -2,   1,   0 },
     {   -8, -10, -16, -42, -28,  -3,  17,   5,   5,   2,   2,   1, -10, -11, -29, -46,   7,   0,   0,   8,   3,   3,   1,   1,  -1,  -6, -25,  36,  21,   4,  -7,  -8,  -1,  -1,  -1,   0, -12, -17,  30,  13,  -8,  -8,   5,   0,  -2,  -1,   0,  -1, -10,  30,  16, -18,  15,   2,  -9,  -2,  29, -22, -19,  -4, -16,  17,   7,   2, -13,  19,  10,  -4, -12, -15,   1,   2,  10,  -6,   5,   9,   4,   0,  -6,  -1,  -3,   5,   0,   2,   1,  -2,   1,   1,   0,   2,   0,   1,   2,  -1,   1,   0 },
     {    7,   9,  25,   1,   9,  24,  -9, -10,  -4,  -1,  -2,  -1,  -4,  -4,  15, -45, -12, -44,  19,  14,   6,   2,   2,   0,  -1,  26,  11,  -3,  29,  26, -33,  -9,  -5,   1,  -1,   0,  17,  20,   5,   1, -26,   4,  24,  -4,  -2,  -3,  -1,  -1,  11, -21,   3, -17,  -4,  -4,  -8,  12, -11,   3,  10,  31, -25,   0,   0,  -7,   1,  -6, -11,  13,  22, -16,  -2,   2,  -4,   1,  -1,  -4,   2,  14,  -5,  -1,   1,  -1,  -3,   4,  -2,   0,   0,   0,   1,  -1,   0,   1,   0,   0,  -1,   1 },
     {    6,   1,   1, -23,  36, -36,   8,   4,   5,  -2,   2,  -1,  10,   3,  -2, -15,  30,  24,  -9, -11,  -1,  -2,  -1,   0,  20,   0,  12, -11,  -1,   5,  19,   1,  -4,  -2,  -1,  -1,   4, -26, -21, -50, -22,  -6,   7,  21,   6,   2,   1,   3, -10, -18,   3,   8,   5,   8, -23,  -4, -16,   1,  15,   0, -38,  16,   9,  -2,   9,  11,  -8,  17, -10, -29,   0,   2,  -5,   3,   7,   5,  12,   2, -10,  -3,   0,   3,  -4,   5,  -2,   1,   1,   1,  -1,   1,  -3,   2,  -1,   0,   2,   1 },
     {    9,  -3,   5,  19,   3,  34,   1, -12,  -2,  -1,  -1,  -1,   9, -13,   6,   9,  12, -29,  -1,   8,   3,   0,   0,   0,  13, -27,   8, -14,   6, -13,  -5,  -1,   0,   3,   0,   0,  16, -25,  29, -13,  30,  25, -17,  -2,  -6,   0,  -2,   0,  -6,   9,  -7,   6, -22,  36,   8,   0, -30, -22, -39,  -4, -21,  -5,  24,   6,  14,  21,   1, -11,  -9, -24, -10,   2,  -6,   4,   8,  12,   3,   2,  -5,  -6,   1,   3,  -2,  -2,  -1,   1,   0,   1,   0,   2,  -1,   0,  -1,   1,  -1,   0 },
     {   -9, -14,   1,  -1, -13, -60,  -7,  20,  10,   3,   2,   1,  -2,  -6,  24,  11,  -4,  27,   2, -20,  -6,   0,   1,  -2,  -9,   0,  35, -19,  50,   0,  -4,   4,  -3,  -2,  -3,   0, -11,  20,  17,  15,  -5,   2,  -4,  -1,  -1,  -2,  -2,  -1,  20, -19,   9, -19, -11,  20,  15,   4,  -6, -18, -35,  24,  -2,  -9,  -3,   7,  -4,   4,  -1, -12,  -1,  -3,   3,  -2,   6,   1,  -1,   1,  -3,  -2,  -7,   2,   1,   4,  -2,   0,  -1,   0,  -1,   0,  -2,   1,  -1,   0,   0,  -1,   0,   1 },
     {   -6,   1,   4,   3,  24,  12,   6,  -7,   0,  -1,   1,  -1,  -9,   1,  13,   4,  33,  -5,   0,   2,  -5,  -3,  -2,   0, -24,  -4,   6,   2, -17, -17,  -5,   7,   6,   3,   1,   1, -38,  17,  15, -17, -17,  21,  -1,  -9,   2,   1,   1,  -1, -31, -12, -14,  -7,  16, -15,  14,   7,  18, -42,  16,  28,  -2,   5, -18,   1,  11,  37, -27, -25,   7,  13,   8,  -7,   8, -10,  14,   0, -19,  -7,   5,   5,  -6,   3,  -4,   4,   6,  -2,   2,  -2,   0,   2,  -2,   1,   3,  -2,   1,  -1 },
     {    2,  -2,   4, -24,   6,   1,   4,   1,   1,   0,   1,  -1,   6,   3,   9, -30,  40, -15,  10,  -7,   1,  -2,   1,  -1,  -3,  -1, -16, -34,  13, -24,  -5,   9,   1,   2,   0,   1,  11,   6,  -7,  10,  30,  36, -13, -14,  -3,  -1,  -1,  -1,  15,  -5,  17, -10,  -2,  -6,  17,   9,  -6,   7,  11, -20,   0,  30, -24,   7, -13, -23, -10,  20, -38,  15,  24,  -5,   2,   8,  -8,  -4,  -3, -25,   1,   8,  -1,  -1,   4,   4,   3,   4,  -3,   1,  -2,  -1,   2,   2,  -1,   2,  -1,   0 },
     {    7,   5,   8,  15,  45,  -3,   9, -12,  -5,  -1,  -2,  -2,   5,  -2,   3,  18,  21, -15, -16,   8,   2,  -2,  -1,   0,  13,  -2,  26,  23,   8,  18,   2, -13,  -5,  -2,  -1,  -1,   5,   2,  32, -21,   2, -26,  -4,  13,   0,   0,   0,   1,  14,  18, -12, -22,  -3,  15, -12,  -3,  17, -11,  -4,   1,  13,  16,   4,  -1, -56, -25,   4,   5,  -9,  23,   6,   6,  14,  -2, -21,  -7, -15,  -6,  12,  -1,  -7,   4,   6,   0,   3,   1,  -4,   0,  -2,   1,   3,   0,   3,   0,  -2,   1 },
     {    0,  10,  -1,  -9,  32, -20,  16,  -8,   2,   1,   2,   0,  -2,  13,  -6,  -7,  23,  -7, -13,   5,   5,  -1,  -1,  -1,  -4,  16,  -8,  -6,  -2,  20,  -4,  -7,  -7,   0,  -1,   1,  -8,  31, -16,  12,  -9,  17,   4,  -2,   1,  -2,  -1,  -1,   0,  25, -18,  -1, -23, -15,  32,   2,  -5, -28, -13, -34,   0, -34,  -4,  16,   7,   6,  38,  25,   4,  -3, -33, -12,   1,  -5,  -4,  10,  14,  22,   9,  -8,   2,  -3,   0,  -5,  -3,  -3,  -2,   2,   0,  -1,   0,  -1,   0,  -3,  -1,  -1 },
     {   -1,   3,  11,   6, -32,  19, -12,   3,  -2,   1,  -1,   2,   3,   8,  23,   6,  -4,  12,  11,  -8,  -4,   0,   0,  -1,  -3,  12,  15,  -8, -21, -43,   2,  17,   2,   0,   1,   0, -13,  15,  -1, -25, -19,  17, -17,  -7,   4,   1,   1,  -1,  21,  27,   6,  10,  11,  -3,   1, -10,  19, -12,   2,   1, -27,   4,  14,   4, -38,  -8,  15,  30,  16, -29,  -4,   2,  24,  -8, -17,  -7,  20,  13, -10,  -4, -10,   5,   5,   0,   5,  -4,  -4,   0,  -5,  -1,   1,  -1,   3,  -1,   0,   1 },
     {    3,   1, -10,  -8,  11,  16, -28,   1,   1,   2,  -2,  -1,  12,  11,  -7,  12,  31,  28,  31,  -5,  -9,  -2,  -2,  -2,   9,   1, -25,  22, -20,   0,  -3,   9,   4,   0,   1,   1,  26, -19,  -5,  12, -25,   5,  17, -10,  -1,   1,   1,   0,  18, -21, -23, -20,  -7, -21,   4,  19, -29,   3, -36,  16,   7,  -1, -11,  -7, -26,   9,  29, -18,   2,  -9,  12,   0,  12,  -4,   1,   0,  -2,  -5, -13,   3,  -2,   6,   1,   0,  -1,  -2,   3,   2,  -1,   1,   0,   0,   0,  -2,   0,  -1 },
     {   -1,  -2,  -3,  -2, -18,  -9,  18,  -9,  -6,   4,   2,  -1,  -3,  -4,  -3,  -2, -41,  12, -47,  23,   3,   5,   0,   2,   0,  -9,   2, -19, -30,  28,   5, -12,   5,  -1,   2,   0,   3,   6,   5,  -3,  19,  20,  23,   1, -11,  -1,   0,  -3,   0, -15, -17,  -2, -23, -22,   0,  24, -14, -22,   7,  10,  -3,  12,  -7,  -9, -21,   7,  -5,   5, -12, -22,  16,   0,  33,   7,   0,  -1,  14, -21, -10,   4,  -9,   5,  17,   3,   4,  -4,  -1,   0,  -4,  -2,  -5,   4,   1,   3,   1,  -1 },
     {   -4,  -5,  -6,   7,  12, -17, -11,   2,   9,  -1,   5,   1,  -4,  -4,  -7,  16,  16,  -2,   4,  -5,  -5,   1,  -1,   0,  -6,  -7,  -3,  26,  -9,  12,   2,   0,  -3,  -1,   0,   1,  -7, -16,  13,  16,  15,  20,  19,  -7,  -4,  -3,  -3,   0, -13,   3,  31,  20,   4, -18,   5,  12,  -7,  -9, -21, -12, -14,   0,   2,   8,  -5, -39, -41,  20,  42, -14,  -6,   0,   9,  32,  14, -42, -10,  12,  -6,  -2,   2,  -5,  -2,   6,   4,   8,  -2,  -5,   0,  -1,   0,   0,   0,   2,   2,  -3 },
     {    3, -10,  -5,  -3,  -9, -38, -18, -11,   9,   7,   5,  -1,   9, -10,   1,   1, -10, -29,  22,   5, -11,  -3,   4,   5,  13, -12,   5,  -8, -12,   1,  -3,   3,  10,   0,  -4,  -1,  21, -16,   9, -22,  38,  -3,  13,  -8,  -5,   1,  -1,   0,  31, -10,   4,  11,  35, -20,  -8,  18,  11, -17,  10, -14,  -8,  -4, -26,  -4,  -5,  29,  11,   1,  18,  20, -21,  -8,   0, -20, -13,  13,  -9,  19,   6,  -5,  -3,   8, -11, -10,   1,   3,   0,   5,   0,   6,   1,  -6,  -4,  -3,   2,   5 },
   }
 },
 {  //33
   {
     {  111, -17, -19,  -5,  -5,  -2,  -2,  -1,  -1,  -1,  -1,   0, -46,  28,   9,  -1,   0,   0,   0,   0,   0,   0,   0,   0, -10,  -9,   7,   3,   1,   1,   0,   0,   0,   0,   0,   0,  -2,   1,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -3,  -1,   1,   0,   1,   0,   0,   0,  -1,   0,  -1,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   10, -64,  12,   8,   3,   3,   2,   2,   1,   1,   0,   1,  66,  42, -43, -10,  -6,  -2,  -2,  -1,  -1,  -1,  -1,  -1, -36,  29,  31,  -7,  -2,  -3,  -1,  -1,   0,  -1,   0,   0,   0, -18,   3,  13,   1,   1,   0,   0,   0,   0,   0,   0,  -7,   5,  -4,  -2,   3,   0,   1,   0,  -1,  -4,   2,   0,   0,   1,   0,   0,  -2,   2,  -1,   0,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,  -1,   0,   0,  -1,   1,   0,   0,   0,  -1,   0,   0 },
     {  -32,  19, -18,  11,   0,   2,   0,   1,   0,   1,   0,   0,  -8,  66,  25, -26,  -2,  -7,  -1,  -3,  -1,  -2,   0,  -1, -15, -59,  43,  28,   0,   3,  -1,   1,   0,   1,   0,   1,  12, -13, -40,   8,  10,   3,   2,   1,   1,   1,   0,   0,  -1,   3,   4, -13,  -1,   1,   0,   0,   5,  -4,  -4,   3,  -3,   0,   0,   0,   0,   2,   1,  -2,   1,  -1,   0,   0,   2,  -2,  -2,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   1,  -1,  -1,   1,   0,   0,   0,  -1,   1,  -1,  -1,   0 },
     {   -3, -60,  -5,   9,  10,   3,   4,   2,   2,   1,   1,   1,  17,  -8,  22,  -1,  -8,   1,  -3,   0,  -1,   0,  -1,   0, -13, -27, -52,  32,  13,   5,   3,   2,   1,   1,   1,   1,   3,  35, -25, -54,   4,   2,   3,   0,   1,   0,   1,   0,   3,   5,  32,   0, -24,  -3,  -3,  -1,  -1,   5,  -6,   9,   4,  -6,   0,  -1,   0,   2,   7,  -3,   0,   0,  -2,   0,   0,   2,  -2,   2,   1,   0,   0,  -1,   0,   1,   3,  -1,   0,   1,  -1,   1,   0,   1,   2,  -1,   0,   1,  -1,   0 },
     {   33,  72, -18, -12,  -5,  -5,  -3,  -3,  -2,  -1,  -1,  -1,  70, -16,  16,   6,  -8,  -1,  -3,  -1,  -2,  -1,  -1,   0, -56,  -6,  -9,  10,  12,   3,   3,   1,   1,   1,   1,   0,   7,  -3,   1, -14,  -4,   1,   1,   0,   0,   0,   0,   0,  -8,   1,   6,   9,  -2,  -3,   0,   0,  -1,  -2,  -2,  -2,   6,   1,  -1,   0,  -3,  -1,   2,   1,  -2,   2,   1,   0,  -1,  -1,  -1,  -1,   1,  -1,   1,   1,  -1,  -1,   1,   0,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {  -22,   1, -27,  -5,   7,   2,   2,   2,   1,   1,   1,   1, -17,  48,   0,  -4, -10,  -3,  -1,  -2,  -1,  -1,   0,  -1,  -9,  -1, -11, -33,  16,  -1,   5,   0,   2,   0,   1,   0,  26,   7,  55, -24, -40,   0,  -6,   0,  -2,   0,  -1,   0,   0,  -7,  10,  58,  -1, -16,  -3,  -3,   1,  -3, -10,  -6,  25,   4,  -2,   0,   1,  -4,   3,   3,  -5,   6,   1,   0,   1,  -1,  -4,  -1,   3,  -1,   2,   0,   0,  -1,   2,   2,   0,  -1,  -2,   0,   0,  -1,   1,   1,   0,   0,  -1,   0 },
     {   -3,  26,  -6,   9,  -5,  -3,   0,  -1,   0,  -1,   0,  -1, -24,   7, -40, -17,  10,   8,   3,   3,   1,   1,   1,   1, -13,  30,   9,  -6,  -1,  -8,   3,  -3,   1,  -2,   1,  -1,  -4, -30,   1, -44,  19,  13,   2,   4,   1,   2,   1,   1,  15,   2,  24, -16, -62,   1,   1,   2,   2,   6,   8,  39,   4, -32,  -5,  -3,   3,  -2,   2,  -4,  15,   9,  -9,  -1,   0,   0,   2,   8,  -3,   2,   2,  -2,   1,  -1,   0,  -2,   0,   0,   1,   4,   1,   0,   0,  -1,   0,   0,   1,   2 },
     {    4,  -1, -60,   5,   5,   2,   3,   2,   2,   1,   1,   1,  33,  13,  38, -30, -12,  -4,  -3,  -2,  -1,  -1,  -1,  -1,  57,  36,   3,  24,  -6,  -7,  -5,  -3,  -3,  -2,  -1,  -1, -49,  -6,  15,   0,  13,   7,   3,   1,   1,   0,   1,   0,   8, -14,  -6,   6,  -7,  -1,   1,   1, -10,   5,  -1,   0,   7,  -2,  -1,  -1,   2,  -5,   1,   0,  -1,   4,   1,   0,  -5,   2,  -1,   0,   1,  -1,   1,   0,   0,  -3,   0,  -1,  -2,   1,   0,   0,   0,  -2,   0,   0,  -2,   0,   0,   0 },
     {  -13,   6, -42,   1,  -1,   9,   2,   3,   1,   1,   0,   1,  -6,  27, -11,  20,   9,  -7,   0,  -2,   0,  -1,   0,   0, -20,  15, -29, -15,   8,   8,   4,   1,   2,   0,   1,   0,   0,  18,  18,  -7,  22, -15,   2,  -4,   0,  -2,   1,  -1,  -4, -33,   5, -41,  18,  38,  -1,   7,   4,   0,   3, -11, -58,   0,  16,   2,   2,  -1,   6,  12,   5, -29,  -6,   3,   3,  -1,   4,  -3,  -1,   6,  -9,  -2,   1,  -1,   3,   5,   1,   0,   1,  -1,   0,  -1,   1,   2,   1,   0,   1,  -1 },
     {  -14, -16,  14, -29,  11,   1,   4,   1,   2,   0,   1,   0, -20,  10,  48,  39, -18,  -4,  -6,  -2,  -2,  -1,  -1,   0, -31,  25, -13,  11,  33,   1,   3,  -1,   1,  -1,   0,  -1, -17, -65,  -2, -10,  -2,  16,   6,   6,   2,   2,   1,   2,  17,  11, -33,  -6,  -6,  -5,   0,   1,   2,   4,  12, -11,  -5,   0,  -2,  -1,   4,   0,  -5,   7,  -3,  -3,   1,   0,   1,   0,   1,  -5,   2,  -1,  -1,   1,   2,   0,  -2,   2,   1,   1,   1,  -2,   1,   0,  -1,   1,   0,   0,   0,  -1 },
     {    8,  32,  50,  10,  -4,  -6,  -7,  -4,  -4,  -2,  -2,  -1,   8,  56, -28,  24,  -2,  -1,   0,  -3,   0,  -2,   0,  -1,  29,   1, -30,   7,  20,   0,   0,   0,   0,   0,   0,   0, -46,   8, -11, -19,   5,   3,   9,   0,   3,   0,   1,   0,  12,  -5,  21,  -7,  15, -12,  -5,   0,  -7, -13,   3,  -8,  11,  28,  -3,   0,   0,  -4,  -2,  -7, -18,   3,  15,   1,  -2,  -4,   3,   1,   1,  -7,  -1,   4,   0,  -2,  -1,  -2,   0,  -2,   1,   0,   0,  -1,   0,  -1,   0,  -1,   1,   0 },
     {   12,   1,  48, -23, -10,  -1,  -1,  -1,   0,  -1,   0,   0,  29,  26,  24,   4,   9,  -6,  -3,  -3,  -3,  -1,  -1,  -1,  27, -37, -10, -22,  -9,  12,   4,   3,   2,   2,   1,   1, -11,   2,  24,   6,  -8,   1,  -5,   2,  -3,   1,  -1,   1,  -6, -20, -20,   7, -31,  18,  10,   3,  -4,  18,  -7,  18, -18, -45,   5,   0,   1,  -2,   3,   8,  33,  -4, -21,  -1,  -3,   3,  -4,  -1,  -4,  12,   2,  -5,   0,   0,  -1,   2,  -1,   1,  -2,  -1,  -1,   0,  -1,   1,  -1,   1,  -1,   0 },
     {    5,  -6, -42, -17,  -2,  11,   1,   3,   1,   1,   1,   1,  19, -10, -19,  23,   7, -10,   4,  -3,   1,  -2,   1,   0,  32, -31,  10, -47,  21,  17,  -1,   5,  -1,   1,   0,   1, -15, -16, -26,  -6, -40,   5,   7,   4,   2,   3,   1,   1,  14,  45,  13, -12,  -5, -14,  -8,  -4,  -8,  -3,  19,  14, -14,   6,   6,  -1,  -2,   5,  -7,   6,   4, -20,   1,   4,  -2,  -1,   7,   1,   3,   3, -10,   0,  -1,   2,  -2,   3,  -1,   0,   3,  -1,  -1,   1,  -1,   2,  -1,   0,   2,   0 },
     {   -8,  -4, -29,  -2, -16,  -1,   4,   3,   2,   1,   1,   1,   2,  16,  -4,  36,  11,  -4, -10,  -4,  -2,  -2,  -1,  -1,  -5,  -4, -22, -10, -10,  -4,   9,   1,   2,   1,   1,   1,  -2,  17,  -6,  21,  11, -11,  -9,  -4,  -1,  -2,  -1,  -1,  -5,   8, -12, -26,   7, -16,  13,  -1,  15,   6,  35,   1,  29, -20, -39,   0,  -2,  -2,   0,   3,  17,  62,  -2, -17,  -2,  -1,   0,  -7, -17,  -2,  31,   4,   0,  -3,   1,  -1,   0,   0,   1,  -2,   0,  -2,   0,   0,   0,   0,   0,  -1 },
     {    2, -10,  11, -29,   9,  -5,   3,  -1,   2,  -1,   1,   0,  10,  -6,  26, -16,  27,   4,  -5,  -2,  -2,  -1,  -1,  -1,   4,  -9,  17, -25,   6,   2,  -3,   1,  -2,   1,  -1,   0,  13,  -3,  22, -11,  15,  13, -10,  -2,  -3,  -1,  -2,   0,  10, -18,  22, -20,  -8,  20, -11,   7,   3,   5,   4,  27, -18,  44, -19,  -8, -11,   3, -18,  12, -23,  25,  45, -12,  -1,  -1,  -4,  -1, -11, -31,  17,  16,  -2,   0,  -4,   3,  -1,   0,  -2,   1,  -1,   0,  -1,   2,  -1,   1,  -1,   1 },
     {   -6,   2,  10, -59, -16,   6,   6,   2,   2,   1,   2,   1,  -3,  15,  25,  21,  -9, -16,  -6,  -1,  -1,  -1,  -1,  -1,   8,  50,  20,  11, -14,   8,  -5,  -3,  -1,  -2,  -2,  -1,  19,  25, -21,   5,  -4, -20,   6,  -2,   1,   0,   1,   0, -12,  10,  47,   1,   9,  -1,  -9,  -3,  -4, -27, -11,  31,   6,  -3,   9,   2,  -1,   0,  -7, -10,  19,  -3,  -9,   3,   1,  -7,   1,   3,  -4,  12,  -2,  -3,  -1,   0,  -4,  -1,   0,  -3,   1,   0,   0,   0,  -2,  -1,   0,  -2,  -1,   1 },
     {    7,   4,  -6, -12,   8,  -6,  -2,   1,   1,   1,   0,   0,   9,   3, -21,  28, -19,   4,   0,  -1,   0,  -1,   0,   0,  47,   3,  15,  18,  41, -20,  -4,  -5,  -2,  -2,  -1,  -1,  52,   0, -11, -19,  13,  34,  -4,   2,  -2,   0,  -1,   1, -58, -12, -17,  -5, -13,   4,  15,   4,  11,   3,   0, -13,  -4,  -5,  -6,   2,  -9,   4,   6,   4,  -7,   1,   2,  -2,   1,   0,  -1,  -1,   2,  -3,  -1,   2,  -4,   2,   2,   1,   0,  -1,   0,  -1,  -2,   1,   0,   0,   0,  -1,   0,  -1 },
     {   10,   6,   9,  71,  -5,  -5,  -3,  -4,  -1,  -1,  -1,  -1,  11,  14,  44,   1,  21,   1,  -7,  -3,  -4,  -1,  -2,   0,  19,  28, -29, -30, -10,   7,  10,   0,   2,   0,   1,  -1,  44, -31, -13,   3,   0,  -6,  -2,   4,   0,   1,   1,   1, -19,  30,   9,  -3,   3,  10,  -5,  -3,  -3, -14,   0,   6,  -3,   1,   9,  -1,  -6,   2,  -4,  -5,   0,  -2,   1,   1,   0,  -1,   2,   0,  -1,  -2,   2,   1,  -2,   1,  -1,  -2,   0,  -2,   0,   0,  -1,   0,  -1,  -1,   0,  -1,   0,   1 },
     {    8,  11,  18, -16,  44,  -2,   0,  -6,  -1,  -3,   0,  -2,   0,   2,  -3, -55, -34, -10,  -2,   4,   4,   4,   1,   2,  11,   3, -16,  -3,   9,  -5,  -3,  -5,  -1,   0,   0,   0,  15,  -1,   4,  -5,  -4, -10,  -4,   4,  -1,   1,  -1,   0,  10,  23,   9, -19,  18,  -2,  -4,  -2,  -1, -21,  34, -14, -21,  -4, -17,  13,   4,   7,   2,  35,   0,  14, -33, -17,  -3,   2,   1,   4,   6,  23,  22, -13,  -2,  -3,   0,   5,  -1,  -1,   0,   1,   0,  -1,  -2,   2,  -1,   0,   1,   0 },
     {    1,  -9,  -3,  12,  -9,  -5,   2,   1,   3,  -1,   1,   0,   3, -18,   0,  22,  34, -19,  -4,   0,   1,   1,  -1,   0,   7, -11,  33,   1,  29,   8,   1,  -3,  -1,   0,   0,   0, -14, -24,   5, -25,  10, -19, -10,   9,   0,   3,   0,   2,   7, -17,  18,  12,  16,  22, -14,  -5,  -5, -18, -34, -20,   3,   1,  -3,  18,   7,  20,   9,  -4, -10,  28, -41, -13,  -3,   5,  -5,  12, -10,  24,  30, -18,   1,   0,   0,  -7,  -1,   2,  -2,   3,  -1,   1,  -1,  -2,   0,   1,   0,   2 },
     {    1,  -2, -12, -28,   0, -15,   4,   2,   1,   3,   1,   1,   7,   9, -18,  14,  -2,  17,  -6,  -1,  -3,   0,  -1,   0,  11, -23, -18,  18, -15, -22,  10,  -1,   3,  -1,   1,  -1,  25, -12,  25,   9,  43,  -7, -23,  -3,  -6,   0,  -2,   0,  34,  46,  -7,   9,  -3,  34,   4,  -5, -42, -18,  -3,  -4,  11,   0,  22,   6,   7,  -4, -10, -18, -11,  -1,  -1,   8,  -6,   3,   7,   2,  -4,  -7,  -4,  -1,   2,  -1,   1,  -3,  -3,   0,   2,   0,   1,  -1,   0,  -1,  -2,   1,   1,   1 },
     {   -2,   0,  -7, -48,  -3, -10,   9,   2,   5,   0,   2,   0,   2,  10, -15, -23,  37,  31,  -8,  -1,  -5,   0,  -1,  -1,   8,   4, -30,   0, -41,  10,  16,   3,   4,   1,   1,   1,  -5, -36, -32, -13,  -8, -18,  -1,   4,   2,   4,   3,   2, -29,  -4,   1,  13,   4,  -2,  -3,  -6,  28,  13,  -8, -29,   0,   0,   1,   1,  -3,  11,  20,   6, -22,  -3,  -1,  -2,   3,   3,  -1,  11,   8, -11,   1,  -2,  -2,   2,   5,  -2,   2,   0,  -1,   3,  -1,   1,   3,   0,   1,   0,   0,   2 },
     {   -2,  -6,   3, -14,   2, -21,  -7,   3,   3,   2,   1,   0,   5,   2,  22, -15,  33,   0,  15,  -4,  -2,   1,  -1,  -2,  -7,  -1,   3, -30,   9, -19,  -9,   8,   3,   0,   2,   1,   3,  15,   2,  -5,  16,  14,  11,   2,  -4,   0,  -4,  -1,  -5, -10,  -7, -23,  -2, -24,  -1,  15,   1, -17,  20, -16,  22,   0,  -8,   9,   3, -23, -10, -24, -23,   2, -29,  35,  -1,  17,   3,  17, -16,  13, -48, -28,   1,  -2,  -2,   6,   3,   4,   2,   5,   0,   0,  -1,   3,  -1,   1,   1,   1 },
     {   -6,   1,   1,  -3, -14,  -2,  -3,   3,  -3,   0,  -1,   1,  -7,   4,  10, -11,  24,  15,  -9,  -6,   1,  -1,   0,  -1, -16,  18,   4,  -8,  10,  18,  -6,   4,  -5,   2,  -2,   0, -19,  35, -13,   9, -16,  49, -13,  -9,   0,  -3,   1,   0,  -4,  18,  -4,   5, -19,  13,  47,  -8, -23, -43,  -4, -22, -13, -19,   3,  15,  18,  14,   5,  10, -21,  -1,   2, -16,  -2,  -6,   2,   7,   8, -17,  17,   7,   4,  -2,   1,  -2,  -1,  -3,   0,   0,   2,  -2,   0,  -1,   0,  -1,   2,   1 },
     {    1,   4,  -4,   4,  70,  12,   1,  -4,  -2,  -1,  -2,  -2,  -2,  -1,   9,  28, -29,  14,  10,   1,   2,  -2,   0,   1,  -9, -12,   1, -38, -28, -31,   0,   6,   5,   4,   2,   0, -31,  -5,  -9,  12,  17,  12, -13,  -6,  -1,   0,   1,   0, -18,  -2,  27,   5,  -5,  10,  16,   0,  24, -13, -29,   5,   4,  -6,  -1,   9,  -3,  12,  -4, -17,  -2,   0,  -3,  -3,   3,  -4,   6,   6,  -6,  -1,   1,   0,   2,   2,  -5,   0,  -1,  -1,   2,  -1,   0,   1,  -2,   0,   1,   0,   2,  -1 },
     {    5,   9,  -1, -11,  57,  27,  11,  -7,  -5,  -4,   0,  -3,  -1,   3, -15, -13,  21, -38, -13,   2,   3,   2,   3,   1,   1,  14, -13,  -8,  14,  42,  -7,  -5,  -6,  -2,   0,   0,  17,   8, -26,  11,   4,   5,  14,   6,   1,   0,  -2,  -1,  13,  -7, -24,   2,  -8,  11,  -3,  -1, -22,  17, -17,  -9,  18,  -4,  12,  -8,   1, -14,  16, -31,   3,  15,   8,   8,   2,  -4,   0,  -3, -27,  -5,   6,   7,   0,  -2,   6,  -1,   0,   2,   1,   2,  -1,   0,   2,   1,   0,   0,   0,   0 },
     {    4,   6,   3,  10,  34,   0,  -5,  -3,  -2,  -2,   0,   0,   9,  12,  12,  38,  17,  42,   9,   1,  -1,  -3,  -2,  -1,   6,  18,  36,  -1, -22,  -5,   3,   4,  -1,   2,  -3,   0,  18,  29, -26, -23,  -9, -14, -10,   2,  -1,   2,   0,   2,  43, -14, -11,  10,  -7,  -3,  -5,  -6, -16,  26,  22, -14,  -4,   0,   3,   3,   2, -15,   0,  32,  -4,  -3,  -8,   1,  -3,  10,  -9,  -2,  17,   1,   6,  -7,  -1,  -6,   5,  -5,  -1,   2,  -5,   2,  -3,  -2,   0,  -1,  -1,   0,   0,   0 },
     {   -5,   2,   6,   4,   7, -17,   3,  -7,   4,  -5,   2,  -2,  -8,  -1,   4,  -9,   1,   7, -10,   3,   2,  -1,   0,  -2,  -7,   3,  20,   3,  19,  14,   5,   1,   1,  -2,   1,   0, -33,  16,   7, -17,   2,  -6, -29,   8,  -5,   8,  -3,   4, -36,  33,   2,  11,  11,  34,  -3, -33,  25,  22,  38,  -1,   6,   1,  40,  11, -16, -27, -10,  -4,  10,   8,   5,  22,   2,  -5,  -7,  -9, -22,  -2,   2,  -4,   0,  -9,   2,   5,   7,  -1,  -3,  -3,  -1,  -1,   1,   2,   1,  -1,  -3,  -1 },
     {    2,   0,   5,  -4, -45,  15,   9,   5,   0,  -1,  -2,   2,   3,   0,   4,  -9, -28, -38,   1,   1,   5,   4,   1,  -3,   0,   6,  -8, -38,  -4, -30, -25,  -3,  -2,   1,   0,   3,  10,  -1, -34,   1,  22,  23, -15,  -9,  -5,  -4,  -4,  -1,  36,  -3,  13,  18,   5,  18,  12, -11,  19,  35,   1, -20,   5,   3,  14,   6, -11,  -7,  19,  13, -13,   5,  -3,   8,  -1,  -2, -12,   6,   2,  -2,   2,  -3,  -2,   2,   2,  -4,  -1,   0,  -1,   5,  -2,   0,   3,  -1,   1,   2,  -1,   1 },
     {    4,  -3,   6,  -1,   7, -19,  22,  -4,   2,  -6,   3,  -1,   3, -10,   5, -13,  10,   0,  -4, -15,   8,  -3,  -1,  -3,   8,  -1,  14,  -8,  19, -15,  28,   4,   0,   0,   2,  -2,   5, -12,  -5, -18,   8, -22,   1, -11,   1,   4,  -2,   4,  20, -13,  18,  -1,  28, -15,  44,   3,  -3, -10,   3, -18,   6, -22,  -2,  15,  12,   7,  14,  -1,  36,  -7,  22,  11,   0, -16, -14, -24, -10,  -3, -25,  49,   2,   7,   3,   9,  -1,  -6,  -3,   0,  -1,   5,  -1,   2,  -1,  -1,  -1,   1 },
     {    1,   0,   7,   4, -22,  50,  19, -11,   3,  -5,   0,   1,  -1,  -5,   6,  -1, -36,  -1,  -4,   3,   8,   7,  -1,  -1,   0,  -8,  14, -15, -24,  -1,  12, -10,   6,   3,   1,   1,  -1,  -5,  -4, -38, -10, -25,  -5,   8,  -5,   9,   1,   5, -22, -22, -18, -14,   3,   6,   5,  -6, -32, -14,  23,  -4,   8,  -2,   7,   5,  34,   2,  -6, -14, -18,   6,   6,  -2,   4,  18,  20,  19, -16, -25,  -1,   4,  -2,  -3,  -3,   9,   1,   1,   1,  -1,  -4,   1,  -7,   2,  -2,   4,   2,   1 },
     {    2,  -1,  -4,   5,  -6,  22,  12, -17,   3, -11,   6,  -9,   2,  -1,  -7,   9,  -7,  18,   2,   3,   7,   0,   9,  -4,   6,  -5,  -3,  15, -14,  48, -17,  -8,  15,   1,   5,   3,  11, -11,  13,   2,  14,  20,  11,   8, -10,  21, -18,   8,  19,  -9,  15,  12,   3, -15,  15,   2,  20, -23,   8, -20, -11,  -9,   0,  12, -38,  -2, -24,   7,  12,   5,   6,  13,  -5,  17, -20,  26, -20,   1,  -8,   0,  -4,  -4, -31,  -8,  -6, -10,  -6, -14,   3,   5,   2,  -4,   5,   7,   5,   3 },
   },
   {
     {  109, -46, -12,  -3,  -2,  -1,  -1,   0,  -1,   0,   0,   0, -33,  26,  -3,   1,   0,   0,   0,   0,   0,   0,   0,   0, -20,   7,   5,  -1,   1,   0,   0,   0,   0,   0,   0,   0,  -3,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -4,   0,   1,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {    6, -45,  21,   2,   3,   0,   1,   0,   1,   0,   0,   0,  90, -26, -30,   5,  -4,   1,  -2,   1,  -1,   0,  -1,   0, -32,  48, -11,  -4,   0,  -1,   0,  -1,   0,   0,   0,   0, -20,   6,  11,  -3,   3,  -1,   1,   0,   0,   0,   0,   0,  -6,   1,   3,  -1,   0,   0,   0,   0,  -3,  -1,   3,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   1,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {   15, -25,  23,  -8,   2,  -2,   1,  -1,   0,   0,   0,   0, -18, -69,  36,  16,   2,   4,   0,   2,   0,   1,   0,   1,  47,  -7, -53,  18,  -4,   4,  -1,   2,  -1,   1,   0,   1, -15,  43, -18, -11,   3,  -3,   1,  -1,   0,  -1,   0,  -1,  -2,   5,   5,  -7,   1,   0,   0,   0,  -4,   7,   1,  -3,  -1,   0,   0,   0,  -1,   1,   2,  -1,   0,   0,   0,   0,  -2,   4,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   2,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0 },
     {   44,  80, -61,  -3, -10,  -2,  -3,  -1,  -2,   0,  -1,  -1,  36, -33,  17,  -4,   2,  -2,   0,  -1,   0,   0,   0,   0,  -2,   4,  -1,  -2,   0,   1,   0,   0,   0,   0,   0,   0, -27,  14,  16,  -1,  -2,  -1,   0,  -1,   0,   0,   0,   0,  -5, -12,   6,   6,   3,   0,   0,   0,  -5,  -1,   0,  -1,   2,   1,   0,   0,  -1,  -4,   1,   1,   0,   1,   1,   0,  -2,   0,   0,   0,   0,   0,   1,   0,  -1,  -2,   1,   1,  -1,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0 },
     {    9, -10,  -3,   4,   0,   1,   0,   0,   0,   0,   0,   0,  46,   2,  27, -23,  -3,  -5,  -1,  -2,  -1,  -1,   0,  -1, -23, -64,  30,  37,  -5,   7,  -2,   3,  -1,   1,   0,   1,  13,  13, -62,  16,   5,   1,   2,   1,   1,   0,   1,   0, -15,  25,   0, -18,   6,  -3,   1,  -1,   1,  -1,   1,  -2,  -1,   1,  -1,   0,  -4,   5,   2,  -3,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,  -2,   3,   1,  -1,   0,   0,   0,   0,  -1,   1,   0,  -1,   0,   0,   0,   0 },
     {   -7, -27, -40,  36,  -1,   7,   0,   3,   0,   2,   0,   1,  20,  42, -37, -12,   3,  -5,   2,  -2,   1,  -1,   0,  -1,  72, -19,   2, -11,   0,  -3,  -2,  -1,  -1,  -1,   0,   0, -33,  33,   1, -10,   3,   1,   0,   0,   0,   0,   0,   0,  -9,  -2,  22,  -9,  -2,   1,  -1,   0,  -6,  -4,   7,   3,  -3,   1,   0,   0,  -2,  -3,   5,   1,  -1,   0,   0,   0,  -3,  -1,   1,   2,  -1,   0,   0,   0,  -1,  -1,   2,   0,  -1,  -1,   1,   1,   0,  -1,   1,   0,  -1,   0,   0,   1 },
     {   15,  -6,  27, -20,   1,  -3,   0,  -1,   0,  -1,   0,   0,  15, -11,  12, -15,   3,  -1,   1,   0,   0,   0,   0,   0,  55,  -1,  42, -21, -19,   1,  -6,   0,  -3,   0,  -1,   0, -20, -25,  16,  57, -20,   4,  -4,   1,  -2,   1,  -1,   1,  -7,   9, -47,  24,  15,  -6,   5,  -1,  -9,   6,  -2, -10,  11,  -1,   0,   0,  -1,   1,  -7,  -1,   2,   2,   0,   1,  -4,   3,  -1,  -2,   1,   0,   1,   0,  -1,   1,  -3,   0,  -2,   1,   0,  -1,   0,   0,  -1,   0,  -1,   1,   0,  -1 },
     {  -15, -12,   8, -20,  13,  -1,   4,  -1,   2,   0,   1,   0,   4,  41,  61, -47, -13,  -5,  -3,  -2,  -1,  -1,  -1,  -1, -10,  20,  -5,  29, -15,   4,  -4,   1,  -2,   0,  -1,   0, -42,  19,  14,  -2,   2,  -4,   4,  -2,   1,  -1,   1,   0,  21, -41,  15,  20,  -2,  -2,  -2,   1,   2,  -3, -23,  11,   7,   2,  -1,   1,   5,  -3,  -4,  -1,   1,   1,   1,   0,   1,  -1,  -6,   1,   2,   0,   1,   0,   2,  -2,   0,   0,   0,   0,  -3,   1,   1,  -1,   0,   0,   0,   0,  -2,   0 },
     {    0,  22,  12,  -1,  -7,  -1,  -3,  -1,  -1,   0,  -1,  -1, -19,  11, -19,  11,   2,   2,  -1,   1,   0,   0,   0,   0,   0,  35,   2,  26, -18,  -4,  -3,  -2,  -2,  -1,  -1,  -1, -50, -10, -36,  34,  39, -10,   8,  -3,   3,  -1,   2,  -1,  15,  -3,  -8, -58,  25,   6,   2,   2,   6, -11,  23,  -9, -17,   9,  -4,   1,   7,  -3,  -2,   0,  -3,   1,   2,  -1,   1,  -3,   6,  -1,  -4,   0,   0,   0,   2,  -2,  -1,   0,   0,  -1,   3,   0,   1,  -1,  -1,   0,   0,  -1,   1,   0 },
     {   21,  42,  59, -56,  -1, -11,  -1,  -5,  -1,  -2,   0,  -2,  19,  44, -32,  15,  -9,   0,  -4,   0,  -2,   0,  -1,   0,  16, -20, -25,   8,   1,   4,   1,   1,   1,   0,   0,   0,  11,  24,   3,  -8,  -7,  -5,  -1,  -1,   0,  -1,   0,   0, -25,  -1,  20,   8,   7,  -2,  -1,  -1,  -1, -14,   4,   1,   4,   4,  -1,   1,  -5,  -3,  -1,   4,  -1,   2,   0,   0,   0,  -5,   0,   1,   1,   0,   2,   0,  -2,  -1,   0,   1,   0,  -2,   0,   0,  -1,  -1,   0,   0,   0,  -1,   0,   0 },
     {   15,   9,  25, -15,   2,  -6,   1,  -2,   0,  -1,   0,  -1,   6, -18, -12, -21,  15,   3,   2,   1,   1,   0,   0,   0,  17, -16,  22,  -1,  -1,  -1,  -1,   0,  -1,   0,   0,   0, -20, -40,  -4, -28,  19,  16,   0,   5,   0,   2,   0,   1,  29,   2,  22, -16, -57,  19,  -4,   3, -11,  10,  -7,  51, -18, -19,   7,  -4,  -1,   7, -10,   0,  15, -10,   1,   1,  -5,   3,   1,   4,   1,  -2,  -4,   1,   1,   2,  -3,  -1,  -2,   1,   0,   3,   0,   1,  -2,  -1,  -1,   1,   0,   2 },
     {    6,  -5, -12,   6,  -1,   1,   0,   1,   0,   1,   0,   0,  25,  23,  10,  41, -30,  -2,  -7,  -1,  -3,   0,  -2,   0,  17, -28, -35,  41,  19,  -8,   5,  -2,   2,  -1,   1,   0,  -5, -55,  17, -13,  23,   0,   3,   1,   1,   1,   1,   1,  23,  -8, -32,   5,   4,   4,  -2,   3, -19,  27, -10, -20,   9,   1,   1,  -1,  -2,   5,   5,  -9,  -2,   3,   0,   0,  -6,   5,   3,  -6,  -1,   0,   1,   0,   0,   0,   4,  -2,  -3,   2,   1,  -2,   0,   0,   2,  -1,  -2,   1,   0,  -2 },
     {   -4,  -3,  27,  19, -19,   4,  -6,   0,  -2,   0,  -1,   0, -12, -16, -23,  31,   9,   0,   2,   0,   1,  -1,   0,   0, -36, -59,  22, -11,  14,   5,   4,   2,   2,   1,   1,   1, -57,  27,  23,   4,  -6,  -7,   0,  -1,   0,   0,   0,  -1,  16, -30,  -7,  13,  11,   4,  -1,   1,  15,   0, -23,  -5,  -3,   4,   2,   1,   8,   8,  -5,  -6,   0,  -3,   1,   0,   3,   5,  -2,  -4,  -1,   1,  -1,   1,   2,   3,   0,  -2,   2,   2,  -1,  -1,   1,   1,   0,  -1,   1,   1,   0,  -1 },
     {   -4,  12,   2,  13,  -9,   0,  -3,   0,  -1,   0,   0,   0, -15,   3, -18,   4, -16,  14,  -1,   2,   0,   1,   0,   1, -11,  12,  -3,  19,  -5,  -5,   1,  -2,   0,  -1,   0,   0, -36,   5, -26,  -7, -15,  13,   5,   2,   2,   1,   1,   1,   7,  47,  -5,  33, -30, -37,   8,  -8,  -8,  -8,   2,   5,  60, -24,  -5,  -2,  13, -10,   8, -26,   8,  17,  -7,   3,  -1,  -4,   4,   2,  -1,   3,  -1,  -2,   4,  -3,   0,  -5,   0,  -2,   2,   1,   2,  -1,   1,  -2,   0,  -1,   1,   1 },
     {   10,   7,   8,  32, -23,  -2,  -4,  -1,  -2,   0,  -1,   0,  -1, -16, -28,   0,  -9,  11,   0,   4,   0,   1,   1,   1,  17,  15,  17,  33, -41,  -3,  -3,  -2,  -1,  -1,  -1,  -1,  45,   5,  -8,  17,  16, -19,   4,  -6,   1,  -2,   0,  -1,   2, -43,  22,  16,  11,  -1,  -1,   2,  -3,   8, -44,  15,  19,  -2,   0,   0, -14,  14,  -2, -24,  14,   2,   1,  -1,  -2,   0,   2,  -6,   1,   2,   1,   0,  -3,   1,   4,  -8,  -1,   0,   0,   0,  -2,   1,   1,  -4,   0,   0,   0,   0 },
     {   -7,   1, -11,  -7,  -7,   9,   2,   2,   0,   1,   0,   1,  -3,  13, -13,  -7,  -5,  13,  -4,   3,   0,   1,  -1,   0, -16,  -6, -27,   3, -33,  19,   3,   3,   2,   1,   1,   1,  -1,  21,   7,  18,  -8,  -4,   2,  -4,   0,  -2,   0,   0, -24,  -7, -41,  -8, -23,  15,  19,   2,   0,  41,   3,   4, -25, -53,  15,  -4,   1,  -9,  20,   5,  38, -24, -13,   3,   6,  -2,   5,  -7,   3,   8,  -9,   2,   1,  -6,   5,   2,   3,   0,   1,  -1,  -1,  -2,   2,   1,   2,   1,   0,   0 },
     {    9,  11,   9,   9,  25, -22,   0,  -4,   1,  -2,   0,  -2,   2,   0, -32, -48,  50,  -5,   6,   0,   3,   0,   1,   0,  -8,  -9, -25,   1,  -2,  12,  -5,   5,  -1,   2,   0,   1,  14,  13,  -7,   0,  -9,   3,  -3,  -1,   0,  -1,   0,   0,  56,  -8, -21,   3,   8,  -4,  -3,   0, -26,  29,   5, -26,   7,  12,  -5,   0,   0,  -6,  21,  -7, -20,  10,   1,  -1,  -5,  -1,   5,   8,  -8,  -1,   2,  -1,   0,  -2,   4,   2,  -1,   0,   1,   3,   0,  -2,   2,   1,  -1,   0,   0,   1 },
     {  -13, -20, -47, -59,  62, -10,  16,  -3,   6,  -1,   3,   0, -12, -24, -39,  12,  -5,  15,  -3,   6,   0,   3,   0,   2,  -4, -16,  15,  33, -23,   6,  -5,   1,  -1,   1,   0,   0,  -8,  -6,  12,   5,  -6,  -7,   4,  -1,   0,   0,   0,  -1,   1,   1,  13,   5,   6,  -7,  -2,  -1,   8,  -8,  -5,  -3,   2,   5,   0,   2,   4,   6,  -5,  -1,  -4,  -2,   0,   0,   1,   3,   2,  -3,   2,   0,  -1,   0,   1,   2,   1,  -2,   1,   1,   1,   0,   0,   1,   1,   0,   0,   1,   0,   0 },
     {    0,   7,  15,  16,  22, -21,  -3,  -4,   0,  -2,   0,  -1, -13, -11,  -7, -29,  12,   6,   2,   2,   2,   1,   0,   0,   0,   4,  13,  24,  18, -22,  -2,  -3,   1,  -1,   0,  -1, -14,  -7,  13, -20,  38,  18,  -9,   2,  -3,   2,  -1,   1, -74,  -8, -15,   6, -11,  20,   4,   3,  16,  11, -14, -25,   5,   0,   1,  -2,   2, -10,  21,  -9, -21,  11,   0,   1,  11,  -1,  -5,  11,  -8,  -4,   1,  -1,   0,   1,   1,  -1,   4,   0,  -2,   2,   0,   0,   1,   0,   2,   0,  -1,   2 },
     {    2,  -5,   9,  10,  -8,   0,  -1,   2,  -3,   0,  -1,   2,   0,  -9,   4, -17, -11,  13,   3,   3,   1,   1,   0,   1,  -6, -26,   5,   9, -30, -20,  11,  -2,   4,   0,   2,  -1,  12,   5,  54,  -6,  21, -28,   1,  -5,  -3,  -1,  -1,  -1,   6,   6,   0, -22,   4, -18,   2,   6,   4,   4,  50,  16,  16, -11, -33,   3,   3, -25, -11,  18,   6,  41,  -8,  -4,  -1,   6, -13,   3, -10,   0,  12,  -3,  -2,  -2,  -5,   5,   1,   3,  -4,   0,   0,  -1,  -2,   1,   1,   2,  -2,  -1 },
     {    2,   4,   7,  39,   2,  -8,  -2,  -1,  -1,  -1,  -1,  -1,  -3,   0,   0, -23,  21,   8,  -4,   2,  -1,   0,  -1,   1, -10, -18, -52,   9, -26,  32,  -3,   7,   0,   2,   0,   2, -16, -47,  28,  21, -22,  -4,   5,  -3,   3,   0,   0,   0, -29,  26,  27,  -4,  14,  -6,   2,  -3,   0, -29, -11,   9,  -7,   9,   6,   2,   5,  17, -14,  -1,  -4, -13,   5,   0,   1,  -1,   1,  -2,   6,  -2,  -2,   1,   2,   5,   0,  -4,   0,  -1,   2,   0,   1,   2,   0,  -1,   0,  -1,   1,   0 },
     {    2,   7,  12,  12,  -7,  12,  -9,   3,  -5,   2,  -2,   1,  -9,  -8,  -9, -20, -43,  25,   2,   4,   1,   3,   1,   1,  -2,   5,  24,  24,  -8,  -8,   1,  -8,   0,  -3,   0,  -1,   4,  22,  23, -21,   0,   0,   3,  -1,   0,  -1,   0,  -1,  13,  35,  -8,   9,  11,   8,  -5,  -2, -47,  -7,   0, -13, -28,  21,  24,   0,  15,  -4,   2,  20, -19, -43,  10,   2,  -2,  -6,  -3,   3,  23,  -6, -10,   4,   4,  -2,  -2,  -1,  -2,   0,  -2,   2,   2,   0,   0,   0,  -1,   0,   0,   1 },
     {    7,   0,   3,  -6, -18,  31, -14,   6,  -4,   2,  -2,   2,   2, -13, -17, -34, -30,  30,   3,   7,   2,   2,   0,   1,  11,  -5,  -9,   2,   8,   5,   4,  -3,   0,  -1,   0,   0,  -1, -25, -18, -11, -26,  15,  16,   2,   3,   1,   2,   0,  10, -35,  -2,  -3,  -7, -28,  12,   1,  58,  -3,  12, -24,  -4,  -2,  -9,   4,  -7,   9,  15,  22, -20,  -8,  -4,   0,  -7,  11,  -8,  13,  12,  -8,  -2,  -2,  -8,   3,   2,   1,  -1,   2,  -1,   3,  -3,   1,   2,   1,   0,   1,  -1,   1 },
     {    6,  10,   5,  37,  38, -31,  -5,  -4,  -1,  -2,  -1,  -2,   3,   7,  -3,  10, -11,  11,  -2,   2,  -2,   0,   0,   0,   1,   5,  -6,  21,  15, -11, -15,   2,  -1,   1,  -1,   0,  10,  22,  -3,  16, -13,  36, -14,   0,  -3,  -1,  -1,   0,   8, -16, -27,   7, -17,   4,  10,  -2,  20, -14,  22,  21,  -5,   2,  -2,  -5, -10,   0, -50,  19,   9,  -3,  19,   0,  -9,  20,   0, -28,  14, -12,   3,   2,  -2,  -7,   5,  -3,   0,   3,   0,  -8,  -1,  -4,   2,   0,   0,   2,   0,  -4 },
     {    2,  -5,  -2,  -3,  -7,   5,   2,   3,   0,   1,  -1,   0,  -1, -11, -12, -18,  14, -28,  17,  -3,   4,  -1,   2,  -1,   4,  -3, -10,  27,  31, -33,  12,  -4,   0,  -1,  -1,  -2,  10,  -6,   0,  20,  -1, -13, -15,  10,  -4,   4,  -1,   1,   7, -15,   3,  16,   2,  14,  -1,  -1,  -8, -52,  14, -12,  -3, -28,  14,  10,  31,  -6,   0,   6,  22, -12, -41,   9, -15,   7, -12,  -1,   6,  32, -13, -10,   5,   4,  -2,  -2,  -4,   4,  -1,  -2,   1,   2,   0,   0,  -3,   1,  -1,   0 },
     {   -4,  -5, -15, -29, -17,  43, -17,   7,  -6,   2,  -1,   5,  -2,  -1, -15, -36,   5,  19,  -2,   0,   5,   1,   1,   0,  -5,  -7, -32,  -4,  23, -32,  11,  -4,   8,  -1,   2,  -1,   2,  11,   4,  23,  22,  -6, -15,  -1,  -4,  -1,  -1,   1,   4,  22, -26,   5,   2,  17, -11,  -1,  13, -10, -22,  27,   3,  19,  -2, -10, -14,   9, -20, -23,   3,   7,  27,  -7,   6,  -3,  14, -13, -12, -14,   6,   6,  -2,   1,   5,   2,   2,  -1,   5,  -1,   0,  -1,   1,   2,   1,  -1,   2,   0 },
     {  -10, -12,  -7, -17, -39,  35, -10,   2,   1,  -1,   0,   2,  -6,  -5,  -7,   9,  30,  -8, -10,  -3,  -1,   1,  -2,   0,  -5,   0,   5,  34,   0,  17, -16,   0,  -3,   2,  -2,   0,   0, -10,  -1,  10, -17,  30, -13,  -2,   0,   1,   1,   1, -21, -30,  18,   9, -13,   5,   7,  -8, -41,  11,  31,  -5,   5,  10,   3, -10,  12, -32, -15,   6, -11,   9,  25,  -4,   5,  14,  -8,  -2,  -4, -22,  12,   5,   5,  -4,  -1,   3,   1,   6,  -2,  -2,   2,  -1,   1,  -2,   1,   2,   0,  -1 },
     {   -1,   5,   7,   6,  23,  12,  -3,   1,  -3,   0,  -1,  -1, -12,  -9, -11, -19, -18, -16,  21,  -1,   4,  -1,   4,   0,  -7,   1,  -3,  -2,  32, -20,   6,  -4,  -2,   0,   0,   0, -17, -15, -17,   3, -22, -13,  19,   7,   2,   2,   0,   0, -26, -11,  19,   2,  24, -30, -18,   9, -28,  49,   7,  15,  -6,  15, -18,  -6, -17,  -5, -10,  10,  32,  -5,   0,  -5,  14,  -8,   5, -27,   8,  22,  -3,   1,   6,  -9,   4,   3,   5,   3,   0,  -3,   0,  -3,   1,   2,   2,   2,  -1,  -3 },
     {    0,   2,   0,  -4, -30,   5,  16,  -2,   3,   0,   2,  -1,  -2,   1,  -7,   0,  25, -49,  12,  -5,   5,  -1,   2,  -1,   9,  17,  15,  52,   4,   3, -17,   0,  -7,  -1,  -3,  -1,   0,   9,  31, -19, -22,   8,  -3,   5,   2,  -1,   0,   0,  10,  39, -12, -10,   5, -10,  -6,  -5,  44,  17, -13,   1,  -9,   4,  -6,  -3, -22,  16,   5,  13,  13,   0,   2,  -5,   3,  -7,  -3,  -9,  10,   7,   6,   3,  -7,  -5,   3,  -3,   1,  -2,  -1,   0,  -3,  -1,   1,  -3,   1,  -2,   0,  -1 },
     {    3,   0,   4,  -3,   3,   3,   1,   0,   2,   1,  -2,  -2,   4,  -3,   4,  -7, -11,  -3,  17,  -6,   1,  -2,   2,   0,   6,   0,   8,   4,  19,  11, -13,   1,  -6,   2,   0,   1,  14,   5,  26,  -3, -11,  12,  17,  -5,   3,  -4,   0,  -3,   5,  -4,   2, -42,   6, -19,  17,   6,   5,  -6,  -5,  -6,   9,  -5,  16,   4,   3, -35, -21, -58,   0, -20,   2,  27,  19,  14,  23,   0, -29,   0, -41,   0,  -7,   0,  -4,  21,   4,   6,   3,  -1,  -2,  -3,  -3,   7,   1,   4,   0,   2 },
     {   -2,   2,   2,  -6,   5,  21,  -7,  -3,  -2,  -4,  -1,   3,  -4,   1,   3,  -6,   1,  11, -12,   1,   6,   1,   1,   1,  -2,   3,   8,   1,  17,   8, -20,  -3,  -2,  -1,  -1,  -2,   4,  13,  16,  -9,  -2,  28,   4,  -7,  -3,   0,   0,   0, -10, -10,  -8, -34,   4, -13,  32,   4, -40, -13, -12,  -7,   4, -19, -24,  18,   7,  66, -10,   0,   2,  19, -22, -16, -15, -12,  20,  -8,  -5,  11,  18, -12,  11,   1,   3,   2,  -3,  -5,   4,   3,   1,  -1,   1,   0,   1,  -5,   2,   1 },
     {   -4,  -2,   1,   0,  -7, -18,  18,  12,  -2,   7,  -2,  -3,  -6,  -3,   4,   1, -33,  -4,  30,   1,  -5,  -4,   2,  -3, -10,  -6,  -5,   2,  -4, -28,  12,  11,  -6,   1,   0,   1,  -7,   3,  15,  41, -22,  21, -36,   8,  -1,  -1,  -4,   1,   8,  -1,  35, -17, -22,  23,   0,  -2, -10,   9,  18, -19,   8,  -3,   3,  -1, -43,  29,  23, -13, -10,   3,   6,   0,   3, -10,   4,  17,  -3, -10,  -4,   2,   2,  -6,   1,   6,   2,   0,   2,   5,   0,  -1,  -2,   0,   3,  -3,  -2,  -1 },
   },
   {
     {  -97,  63,   1,  -3,   3,   0,   1,   0,   0,   0,   0,   0,  26, -33,  16,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,  24, -13,  -8,   7,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   5,  -4,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {   23,  58, -50,   2,  -6,   0,  -2,  -1,  -1,   0,  -1,   0, -57,  35,  29, -25,   8,  -4,   2,  -2,   1,  -1,   1,  -1,  17, -45,  32,   4,  -8,   3,  -1,   1,   0,   0,   0,   0,   8,  -6, -11,  14,  -2,  -1,   0,   0,   0,   0,   0,   0,   1,  -2,  -2,   0,   4,  -1,   0,   0,   2,  -2,  -2,   1,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {  -41, -43,  54,  -2,   2,   2,   1,   1,   1,   0,   0,   0, -76,  22,  -3,  15,   2,   1,   2,   0,   1,   0,   0,   0,  37, -25, -16,   2,   5,   0,   0,   1,   0,   0,   0,   0,  22,   2, -18,  -4,   0,   1,   0,   0,   0,   0,   0,   0,   5,   8,  -2,  -8,  -2,   0,   0,   0,   4,   2,   1,  -1,  -3,  -1,   0,   0,   2,   2,   0,   0,   0,  -2,   0,   0,   2,   1,   0,   0,   0,   0,  -1,   0,   1,   1,   0,   0,   1,   0,   0,   0,   1,   1,   0,   0,   1,   0,   0,   0 },
     {  -43,  -3,  -9,  16,   0,   5,   0,   2,   0,   1,   0,   1,  13,  60, -57,  -8,   9,  -6,   2,  -3,   1,  -1,   1,  -1, -10,  12,  44, -45,   7,  -1,  -1,  -1,   0,   0,   0,   0,  13, -31,  27,  11, -19,   7,  -2,   1,  -1,   1,  -1,   0,  -1,  -1,  -7,  13,  -1,  -5,   2,   0,   2,  -6,   1,   1,   3,  -1,  -1,   1,   1,  -1,  -2,   2,   0,   0,   0,   0,   1,  -3,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0 },
     {    6,  13,  58, -59,  10,  -8,   2,  -4,   1,  -2,   1,  -1,  26,  33,  18,  -7, -27,   4,  -5,   0,  -2,   0,  -1,   0, -19,  -6,  27,  32, -18,  -4,  -1,  -1,  -1,   0,  -1,   0,   5, -26, -12,  37,   7,  -7,   0,   0,   0,   0,   0,   0,  -7,   4, -22,   5,  20,  -2,  -1,   0,   1,  -4,  -3,  -9,   7,   7,  -1,   0,  -1,   1,  -5,  -1,  -2,   4,   2,   0,   0,  -2,  -1,  -2,   1,   0,   2,   1,  -1,   0,  -2,   0,   0,  -1,  -1,  -1,  -1,   0,  -1,   0,   0,  -1,  -1,   0 },
     {   19,  55,  11, -46,   8,  -7,  -1,  -3,  -1,  -1,   0,  -1, -27,  11, -24,  42, -19,   6,  -2,   2,  -1,   1,   0,   0,  -1,  13, -32,  -8,  29, -14,   7,  -2,   2,  -1,   1,  -1,   4, -15,  41, -39,   3,  11,  -6,   3,  -1,   1,  -1,   1,   3, -11,   6,  16, -20,   7,   1,  -1,  -1,  -1,  -1,   4,   3,  -6,   3,   0,   2,  -4,   1,   3,   0,   0,  -1,   1,   0,  -1,   1,   0,   1,   0,   0,   0,   1,  -2,   0,   2,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 },
     {  -20,   1,  -7,  25,  -4,   0,  -1,   1,   0,   0,   0,   0, -32, -23,  37,   6,   3,   0,  -1,   1,   0,   0,   0,   0, -80,  26,   0,   7,  13,  -4,   3,  -1,   1,   0,   1,   0,  45, -45,  -3,  -4,   7,   4,  -2,   2,  -1,   1,   0,   1,  13,   5, -20,  -2,  -3,   5,   0,   0,   6,   5,  -3,  -7,  -2,  -1,   2,   0,   4,   2,   0,  -2,  -3,  -1,  -1,   1,   4,   2,   0,  -1,   0,   0,   0,  -1,   2,   1,   0,   0,   2,   0,   0,   0,   1,   1,   0,   0,   1,   0,   0,  -1 },
     {    6,  -7,   9, -25,  13,  -4,   4,  -2,   2,  -1,   1,  -1, -23, -37,  27, -33,  15,   5,   2,   2,   1,   1,   0,   1,  23,  33,  -2, -33, -27,  14,  -4,   2,  -2,   1,  -1,   1,  11,   0,  36,  20, -48,   1,   1,  -1,  -1,   0,  -1,   0,  -1,   3, -16,  45,  -3, -22,   4,  -1,  -3,   5,  -6,   3,  22,  -9,  -6,   1,   1,  -2,  -3,   2,   5,   6,  -4,  -1,  -1,   2,  -2,   1,   2,   1,   1,  -1,   0,  -1,  -2,   1,  -1,   1,  -1,   0,   0,   0,  -1,   1,   0,   1,   0,   0 },
     {   20,  13,  13,   4, -13,   2,  -6,   2,  -3,   0,  -1,   0,   2, -51, -19,  53,   2,  -4,   4,  -1,   1,   0,   1,   0,   2, -42,  30, -34,  30,   0,  -2,   2,  -1,   1,   0,   1,   1,  -4, -17,  26, -23,  18,  -4,   1,   0,   0,   0,   0, -13,  27, -31,  -3,  14, -12,  10,  -3,   4,   0,  10, -20,   3,   3,  -4,   3,  -1,   4,  -4,   2,  -6,   2,   1,  -1,   0,   2,   2,  -4,   0,  -1,   0,   1,  -1,   3,  -2,   1,   0,   0,   1,  -2,  -1,   1,  -1,   0,   0,   1,   1,  -1 },
     {  -11, -25, -32, -43,  61, -10,  13,  -2,   5,  -1,   2,   0,   0, -21, -25, -11,  -5,  27,  -5,   6,  -1,   2,   0,   2, -29, -47,  16,   2, -16,  12,   8,   1,   2,   1,   1,   1,   9,   1, -10, -23,  -9,   0,   3,   4,   0,   1,   0,   1,  -3,  19,   7,  -4, -29,   1,   0,   1,   6,   5,   1,  11, -11, -15,   3,  -1,   1,   6,   3,   4,   4, -10,  -5,   2,   3,   2,   1,   3,   2,  -1,  -5,  -1,   0,   3,   1,   2,   1,   1,   1,   1,   0,   1,   1,   1,   1,   1,   0,   1 },
     {   13,  -4,  19,   5, -26,  10,  -4,   1,  -2,   0,   0,   0,  42,  17,  31, -30,   5, -13,   2,  -5,   0,  -2,   0,  -1,  14, -29, -13,  -4,  30,   0,  -3,   2,  -1,   0,   0,   1,  41, -17,  -9, -29, -12,  29,  -7,   4,  -2,   2,  -1,   1, -27,  27,   4,  15, -42,   5,   9,  -3, -11,   2,  -1,  17,   1, -21,   8,   0,  -5,   0,   2,   3,   8,  -5,  -5,   2,  -1,   0,  -3,   4,   3,   1,  -2,  -1,  -3,   1,   0,   0,  -1,   0,  -1,   2,  -2,   0,   0,   0,  -1,   0,  -1,   1 },
     {   -6,  -5,  32,  -2,  -9,   2,  -4,  -1,   0,   0,  -1,   0,  -4,   1,  24,  -2,  12, -15,   4,  -4,   1,  -2,   1,  -1, -35, -27,  32, -12,  20,   6,  -7,   5,  -2,   1,  -1,   1, -56,  23,   6,  -9, -11,  18,  -1,   0,   1,   1,   0,   0,  40, -45,   5,  16, -23,   6,   3,   1,   7,  13, -33,  18,  -3,  -8,   4,  -1,   5,   4,  -5,  -7,   8,  -6,   0,   0,   3,   2,  -3,   0,  -1,   1,  -3,   0,   2,   1,  -1,  -1,   1,   2,  -2,   0,   2,   0,  -1,   0,   0,   1,  -1,   0 },
     {   10,  34,  31, -24, -10,   6,  -3,  -3,  -2,  -1,   0,  -1,  -8,  -3, -28, -25,  52, -19,  12,  -1,   2,   0,   1,   0, -28,  19, -21, -15, -10,  31, -12,   7,  -1,   2,  -1,   1,   9,  10, -29,   5, -18,   5,  14,  -5,   2,  -1,   1,  -1,  -2,   2,  22, -41,   3,  -3,   4,   4,   7, -10,  17,  10, -30,   4,  -3,   3,   1,   0,  -1,  10,   2, -14,   0,   0,   2,  -2,   4,   4,   0,   1,  -6,   0,   0,   0,   1,   2,   1,  -1,   1,   2,   0,   0,   0,   1,   1,  -1,   1,   1 },
     {   10,  -3,   8,   2,  21, -19,   4,  -4,   1,  -2,   0,  -1,  -2, -24,   0, -11,  30,  -4,  -2,   1,   0,   0,   1,   0,  41,  15,  23,  -4,  -2,  -3,  -7,  -3,  -2,  -1,  -1,   0, -15, -48,  16,  10,  39,  16, -12,   4,  -3,   2,  -1,   1,  -1,  -6, -25, -21, -16,  47,  -3,   0, -11,  12,  -5,   6, -38,  -1,  20,  -5,   0,   3,  -4,   1,  -2, -18,   6,   5,  -2,   2,   2,   1,  -1,  -4,  -6,   3,   0,   2,  -3,   1,  -1,   2,   1,   0,   1,   1,  -1,   1,  -1,   1,   0,   1 },
     {  -10, -10,  -6, -14, -17,  40, -12,   5,  -1,   1,   0,   0, -11,  -6, -14, -14,  28, -28,  23,  -9,   8,  -2,   1,   0, -15, -22, -11,  44, -12,  -6,  -9,  12,   0,   1,  -1,   1, -33,  -6,  49, -16,   5,   0, -11,   3,   0,   1,  -2,   1, -31,  31, -22,  14,  10,   3,   0,  -7,  21, -10,  -1, -18,  -1,  11,   0,   2,   7,   5,   2,  -6,  -9,   0,   2,   1,   5,   5,   1,   0,  -2,  -3,   1,  -1,   2,   1,   3,  -2,   3,   2,   0,  -1,   1,   1,   1,   0,   2,   0,   1,  -1 },
     {   -5,  -2,  -3, -26, -31,  44,  -4,   6,  -3,   3,  -1,   1, -17, -19,   2, -15, -38,  11,  21,  -4,   6,  -1,   2,   0,   2,  32,  41, -21,  -3, -48,  12,   0,   0,  -1,  -1,  -1,   2,  25, -12,   1,  12,  -6, -19,   8,  -1,   2,  -2,   0,  -5,  15,  11, -11, -11,  12,  -4,  -8,  12, -16,  10,  21, -13,  -9,   9,  -2,   2,   1,  -8,   9,  12, -14,  -4,   5,   2,  -2,   1,   1,   7,   1,  -7,   0,   0,   0,  -2,   2,   1,  -1,   1,   0,   0,   0,  -1,   1,   1,  -1,   0,   1 },
     {   -8,   1,   1,   7,  42, -36,   8,  -6,   4,  -4,   2,  -2,  -5,  22,  40,  10,   3, -16,  -9,  -1,   0,  -2,  -2,  -1,  -6,  28,  20, -10,  17,   1, -16,   0,  -1,   0,  -2,   0, -19,  38,   2, -19,   0,  27,  -9,  -4,  -1,   1,  -1,   0, -31,  29,  -1,  -3,   0,   4,  13,  -7,  21, -40,  23,  -9,  -5,  13,  -1,   5,   4,  -3, -11,  15, -14,   4,   5,  -1,   3,  -2,  -4,  -1,   5,  -6,   5,   1,   1,  -1,  -2,   0,   2,  -2,  -1,  -1,   1,  -1,  -1,   0,   1,  -1,   0,   0 },
     {    5,  23,  24,  40, -15, -16,   0,  -4,  -1,  -2,  -1,  -1, -14,  -3, -21,  11,  -8,  16, -12,   8,  -3,   2,  -1,   2,  -4,  13,  17,  24, -43,  16,  -1,  -6,   2,  -2,   1,   0, -29,  -9,  -5, -10,   7,   0,  18,  -2,   2,   0,   1,   0, -16,  27, -11,   0, -34, -13,  13,   0,   9,  -9,   3,  20,  22, -43,   1,   3,   8,  -3,  -2, -13,  27,   2, -19,   4,   2,   0,   0,   0,   2,  10,  -2,  -5,   1,  -1,   0,  -5,   1,  -1,   1,   0,   1,   0,   0,  -3,   1,  -1,   1,   0 },
     {   -9, -10, -24, -15,   4,  18,  -2,   5,  -3,   3,   0,   1,  -6,  12,   4, -22, -17,  19,   2,  -2,   0,   1,   1,  -1,   6,  16, -35,   0,  29,  -5,   5,  -2,  -3,   1,  -1,   0, -55, -37, -35,  19, -15,  34,   8,   3,   4,   0,   2,   1,   8,  -7, -21, -15, -11, -21,  27,   0,   2,  -5,  15, -20,   2, -20,  -5,  10,  10,  -2,   6,   6, -11,  -2, -12,   1,  -2,   5,   3,   0,   2,  -2,  -5,  -3,   4,   1,   2,   2,  -1,  -1,   3,   1,   1,   0,   1,   1,   0,   0,   1,   0 },
     {    3,   7,  -4, -12,   6,   3,  -2,   1,   3,  -3,   2,  -2,   4,  10, -15,  -2,   9, -22,  12,  -5,   5,  -2,  -1,   1,   5,  22,  16,   9, -10,  -1, -14,   6,  -2,   1,  -2,   0,  15,  36, -26, -25,  13,  13, -11,   2,  -1,   3,  -3,   0,  50,  21, -46,   9, -10,   1,   6,  -8, -20,  31,   9, -26,  10,  -8,   2,   4,  -5, -10,  41, -13,  -9,   9, -10,   3, -10,  -2,   6,  16,  -5,   2,   3,  -4,  -2,  -3,   4,  -1,  -2,  -1,   2,   2,   0,  -2,   3,   0,  -2,   0,   2,   1 },
     {    7,   3,  15,   4,  18, -12,  -2,  -2,   0,  -1,  -1,   0,  -5, -23, -17, -28,  21,  -9,  12,  -3,   5,  -1,   3,  -1,   5,  -1,  18,  11,  18, -36,   3,   1,  -4,   1,  -1,   0,  -6, -28, -13,  -8,  18,   3, -12,  10,   1,   1,   0,   1,  18,   6,  39,  -4, -11, -22, -13,   0,  -1, -11,  11,  14,  40,  16, -34,   1,  -1,   5,  -9,  -2, -17,  48,  -2, -13,  -3,   2,  -4,   6,  -8,   0,  20,  -5,   2,   1,  -2,  -3,  -1,   1,  -1,   3,   1,   1,  -1,  -2,   0,   1,  -1,   1 },
     {   -4, -11, -26, -15,  20,  19, -17,   6,  -1,   2,   0,   1,   1,  -3,  -1,  43,   5, -36,   9,  -6,   2,  -2,   1,  -1,  16,  23,  17,  43,  -5,  -2, -31,   7,  -4,  -1,  -3,   0,  23, -10, -10,  17, -17,  33,   0, -15,   5,  -2,   1,  -1,  -6, -21,  27,  -3,  -2, -10,  25,  -6,  -5,  -3, -21,  22,   7,  -9,   0,   9,  -6,   8, -10, -13,  21,  -6,  -1,   1,   1,  -1,   1, -10,   1,   8,  -3,   0,   0,   1,   0,  -5,  -1,   0,   0,  -3,   0,   1,   0,  -2,   0,  -1,   0,  -2 },
     {   -5,   2,  -5,   1,  -2, -26,  28, -11,   8,  -4,   2,  -2,  -9,  -3, -24, -12,  12,  -6,   1,  10,  -2,   3,   1,   0, -19,  -5, -25,  -1,  10, -26,  -7,  13,   0,   2,   0,   3,   4,  22, -20,  45,   5,   4, -40,   5,   1,   0,  -1,   0, -38, -24,   4,  22,  -4,  31,   0, -19,   1,  -6, -10, -10,  17, -23,  22,  -6,   7,  -9,  14, -17,  11,  11, -15,   9,   7,   2,   1,  -4, -11,  11,   2,  -6,   1,   0,   3,  -2,   2,   0,   1,  -1,   1,   0,   3,  -1,   2,  -1,   0,   0 },
     {    5,  15,  17,  21,  16,  23, -24,   1,  -7,   0,  -4,   3,  -9,  -7, -22, -19,   8,  52, -12,  -5,   0,  -3,   1,   0,  -2,  11,  10,  26,  10,   5,   8, -13,   2,  -4,  -2,   0,   6,  17, -14,   2,  -6,  26,   1,  -1,   0,   1,  -1,   0,  -9,  -1,  13,  30,  -3,   4,  10,  -8, -15,  -6, -17, -13,   9,  25,  25,  -5, -13,   0,  -3, -18, -27, -21,  33,   5,   7, -10,   5,  -3,   4, -29,  -8,  15,   0,  -1,   0,  -2,   2,  -1,   1,   0,   2,   0,   2,  -1,   1,   0,   0,   1 },
     {    6,   8,  10,  15,  41,  15, -51,  16,  -9,   2,  -4,   2,  11,  15,  14,  14,  35,   6,  -5, -15,   1,  -4,   1,  -4,  -2,  -4, -12,   2, -13, -33,  54, -12,  -4,   2,   1,  -1,   5,  15,  10,  15, -23,  -6, -14,  23,  -8,   2,  -3,   1,   5,  -4, -13, -20,  -2,  11,  -2,   3,  11,   1,   4,  -2,  -2, -21,   0,   2,  -2,  -5,  -2,   7,  14,   8, -19,   1,  -4,   1,   2,  -4,  -4,  13,   2,  -5,  -1,  -2,   0,   2,  -1,  -1,   1,  -1,  -1,  -1,  -1,   0,  -1,  -1,   0,  -1 },
     {   -1,   5,   5,  15,  -6, -20,  16,  -5,   3,  -3,   1,  -1,  -4,   4,  -2,  -5, -20,  -4,  -7,  12,  -1,   2,  -1,   2,  -7,  -2,  -4,  22,  -5, -26,  -2,   8,   1,   1,   0,   1, -12,   0,  28,  18, -52,  19, -18,  -3,   1,   1,  -1,   0,  22,  44,  32, -34,   6,   7,  -3, -13, -25,  42,  -4,   1,  -5,   8,  11,  -2,  20, -12,   4,  12,  -2, -10,   7,   0,  -7,   3,  -4,   5,  16,  -7,  -1,   3,   2,  -3,   0,  -2,   1,  -4,  -1,   0,   1,  -2,  -1,   0,  -2,   0,   0,   0 },
     {   -5,  -1, -11, -14,  11,  10, -12,   3,   4,  -2,   0,   0,  -1,  11,  14,  26,  12, -19,   7,  -9,  -1,   1,  -1,  -1,  -3,  12, -10, -24,   2,  13,   8,  -6,  -3,   1,   3,  -2, -28, -18, -35,   6,   0, -21,   9,  12,  -5,   1,   0,   3,  -4,  45,  14,  20,  -5,   9, -39,   5, -27,  -5, -19,  16,  17,  17,  20, -21,  31, -15,   5, -29,   5,  -7,  15,   8,  -6,   5,   1,   0,  -9,   8,  -5,   4,   5,  -1,   0,  -4,  -5,  -1,   0,   2,   5,  -3,   2,  -1,  -2,   0,   0,   1 },
     {   -8,  -8,  -1, -18, -13,  29,   5,  -4,   2,   0,   2,  -1,  -3,   0,  11,  17,  30,  16, -20,   3,  -6,   1,  -1,   0,  -5,   1,  27,   8,   8,   7,  -3, -13,   4,  -4,   0,  -1,   3,   8,  11,   4,  14,  -6,  18,  -9,  -3,  -1,   0,  -2, -26, -19,  21,  -4, -13,   5,   8,   6, -33,  25,  33, -21,   9, -18,  -2,  -2,  35, -45,   1,  24, -11,  18,  -4,   3,  -5,  10, -26,   8,  -2,  -5,   9,   3,   6,  -4,  -5,  -5,  -2,   5,  -2,  -2,   2,  -1,   0,  -2,   0,   2,  -2,   1 },
     {    2,   5,  14,  15,  16,  -6, -13,   9,  -4,  -3,   2,  -1,  -6,  -6,   2,  -8, -32, -15,  32, -10,   6,   1,   2,   1,  -5,  -2,   8, -24, -34,  -4,  11,  12,  -4,   4,  -1,   1,  -4, -12, -12, -39,  -5,  16,  -4,   8,   1,   3,  -1,   2, -50, -31,   7, -15,  25,  -7,   6,   7, -30,  19, -25, -17,   0,   4,  -6,  11,  15, -11,  16,  -5, -16,   0, -10,   0,   9,   9,  -7,  13,  -2,   0,  -9,  -4,   3,   6,   4,   0,   4,   0,   0,   2,  -2,   2,   2,   1,   0,   1,   1,   1 },
     {    1,   1,  -1, -12,  -3,   3,  23,  -4,   6,  -6,   1,  -2,   1,   2,   5,  -8,  15,   6,  -8,  15,  -9,   3,  -4,   2,  11,  17,  17,   4,  33, -15,   9, -21,   1,  -1,   2,  -5,   5,  10,   8,  -6,  -7,  -4,  35,  -3,  -8,   0,  -1,   1, -24,  12,   3, -32,   5,  -6,  -1,  25,  41,  28, -49, -12,   9,  -3,  -4,   5,   5,   9,  20, -34,  -3,   5,  -6,   4,   0, -11,  19,  10, -17,   9,  -6,   0, -12,   2,  -2,  11,   4,  -4,  -1,   2,  -2,   1,   0,   0,   1,  -3,   3,  -1 },
     {    7,   2,  10,  15,  16,   0,  26, -19,   3,  -4,   0,  -6,   5,  -6,   0,  -4,   9,  -2,  29,  -1,  -5,  -1,   0,  -3,   8,  -4,  12,  19,  24, -10,   8,  12,  -6,   1,  -1,   1,   9,  -3,  10,  -3,  -3, -23,  15,   6,   2,   0,   0,   1,  25,   0,  13,  14,  15, -17,  -1,  16, -17, -42,   3, -21,  -4, -12,  20,  22,  12, -12,  -3,  -2,  -7, -37, -34,  25,  -9,   5,  -9,   2,  10,  19, -39,  -8,   5,   2,  -6,  -5,  -1,  -1,   2,  -2,   1,   2,  -2,  -2,  -1,   2,   1,   1 },
     {   -3,  -8,  -5, -14, -10, -38, -39,  47, -14,  15,  -9,   7,  -4, -13, -17, -30, -14, -15,  -9,   8,  14,   0,   6,  -5,  -3,  -8,  -2,   0,  17,  11,   3, -22,   1,  -1,   5,  -4,  11,  11,  11,   1,  24,  13,  12, -19, -10,  -7,   3,  -1,  12,  10,  -4,  -7,  22,   2,  17,  -4,   4, -12, -17,   3,  17,  -2,  16,   3,  24, -35, -19,   5,  10,  -3,  -2,  10,  -8,   3,  -9,  -1,  -5,  18,  -6,   2,   4, -13,  -2,   4,  -7,  -4,   0,   3,   0,  -3,  -1,  -2,   1,   0,  -2,   3 },
   }
 },
 {  //34
   {
     {  111, -35, -14,  -3,  -4,  -1,  -2,   0,  -1,   0,  -1,   0, -21,  39,  -7,  -3,  -1,  -1,   0,  -1,   0,   0,   0,   0, -16,   2,  15,  -3,   0,   0,   0,   0,   0,   0,   0,   0,  -3,  -3,   2,   6,  -1,   0,   0,   0,   0,   0,   0,   0,  -4,   0,   0,   1,   3,   0,   0,   0,  -1,  -1,   0,   0,   0,   1,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
     {   -1, -51,  11,   1,   4,   1,   2,   1,   1,   0,   0,   0,  70, -13, -52,   7,  -4,   2,  -1,   1,  -1,   1,   0,   0, -12,  58, -15, -26,   4,  -4,   1,  -2,   1,  -1,   0,  -1,  -2,  -1,  29, -12,  -8,   0,  -1,   0,  -1,   0,   0,   0,  -2,   3,   2,   9,  -6,  -3,   0,   0,  -2,   1,   4,   1,   4,  -3,  -1,   0,  -1,   1,   1,   2,   0,   2,  -2,   0,  -2,   0,   2,   1,   1,   0,   0,  -1,  -1,   0,   0,   1,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   1,   0 },
     {  -32,  17,  -5,   3,  -4,   2,  -1,   1,   0,   1,   0,   0,  25,  68, -20, -23,  -1,  -7,   0,  -3,   0,  -2,   0,  -1, -10,   5,  70, -12, -12,  -3,  -4,  -1,  -2,  -1,  -1,   0,   4, -26,   5,  44,  -5,  -2,  -1,   0,  -1,   0,   0,   0,  -5,   0, -12,   2,  19,  -2,   0,  -1,   2,  -8,   1,  -2,   2,   8,  -1,   0,  -1,   0,  -4,   0,   0,   1,   4,   0,   0,  -4,   0,   0,   0,   0,   0,   2,  -1,   0,  -1,   0,   0,  -2,   0,   0,   0,   0,  -1,   0,   0,  -1,   0,   0 },
     {   -3,  54, -27,  -8,   0,  -1,  -1,  -1,   0,   0,  -1,   0, -48,  22, -22,   5,   1,   3,   1,   1,   1,   0,   1,   0,  -6,  23, -14, -47,  15,  -6,   5,  -2,   2,  -1,   1,  -1,  12,  -5,  52, -30, -26,   5,  -5,   2,  -2,   1,  -1,   0,   5,  -4,   8,  30, -18, -10,   0,  -1,   2,   0,   4,   3,  12,  -7,  -5,   0,   2,  -3,   2,   5,   0,   7,  -4,  -1,   0,  -1,   1,   2,   1,   0,   3,  -1,   1,  -2,   1,   2,   0,   0,   0,   0,   1,  -1,   0,   1,   0,   0,   0,   0 },
     {    8,  44, -35,   9,   0,  -2,   0,  -1,   0,   0,   0,   0,  46,  31,   1, -11,  -2,   0,  -4,   0,  -2,   0,  -1,   0, -33,   9, -10,  33,  12,  -3,   4,  -2,   2,  -2,   1,  -1,  -8,  -8, -16, -42,  39,   8,   3,   2,   1,   1,   1,   0,   2,  -9,  19, -27, -33,  21,   2,   3,  -2,   2,  -2,  11, -13, -18,   8,   0,   1,  -3,   5,  -4,   3,  -4,  -9,   3,  -1,   0,   0,   4,  -3,   1,   0,  -4,   0,  -1,   2,  -1,   0,   0,   0,   2,   0,  -1,   1,  -1,   0,   0,   0,   1 },
     {   43,  66,   2, -24,  -5,  -6,  -3,  -3,  -2,  -1,  -1,  -1,  60, -38,  35,   4, -10,  -2,  -2,  -1,  -1,   0,   0,   0,  -1,   0,   2,   1,  -5,   1,  -2,   1,  -1,   0,   0,   0,   0,   1,  18,  24, -19,  -3,  -2,  -2,  -1,  -1,  -1,   0,  -6,   1, -12,  27,  20, -10,  -2,  -2,  -4,   0,  -3,  -8,  13,  11,  -1,  -1,  -3,  -2,  -3,   2,  -2,   3,   6,   0,  -3,   0,  -2,  -3,   2,  -1,   0,   3,  -2,  -1,  -2,   1,  -1,   0,  -1,  -1,  -1,   0,  -1,   0,  -1,   0,   0,  -1 },
     {    4,  27,  -7,   7,  -4,   2,  -4,  -1,  -1,   0,  -1,   0, -18,  -3, -29,  27,  10,  -4,   2,  -1,   1,   0,   0,   0,   1,  34, -20,  -5,   2,  -1,  -3,  -2,   0,  -1,   0,  -1, -12,  18,   7,  18,  30, -22,   4,  -7,   1,  -3,   1,  -2,   4, -16,   6, -35,  54,  26, -14,   4,  -2,  -5,   3, -15, -32,  43,  13,  -5,  -1,  -2,  -3,  -2, -10, -16,  21,   5,   1,  -3,   3,  -5,  -5,  -2,  -6,   9,   0,  -1,  -1,   0,   1,  -1,   2,  -2,   0,  -1,  -1,   0,   0,   0,   1,  -1 },
     {  -15, -17, -20,   5,  10,   2,  -1,   3,   0,   1,   0,   1,  18,  40,  10,  10, -17,  -4,  -2,  -4,  -1,  -1,  -1,  -1, -25, -28, -28,  15,  25,   2,   0,   3,   0,   1,   0,   1,   8,  25, -16, -25, -26,  -6,   4,  -5,   3,  -1,   1,  -1,   5,   7,  26,  18,  28, -33, -15,  -2,   3,  -4,   1, -11,  23,  44, -14,  -7,  -1,   0,   2,   1, -16,   8,  28,  -4,   2,  -4,  -1,  -5,   4,  -6,   1,  11,  -1,   0,   1,   0,   1,  -1,  -1,  -2,  -1,   0,   1,   0,   0,  -1,   0,  -1 },
     {    5,   8, -33,  18,  -8,  -5,   0,   0,   1,   1,   0,   0,  -3,  -5, -11, -28,  -2,   1,  -2,   2,   1,   0,   0,   0,  48,  31,   0,  18, -32, -17,  -5,  -2,  -1,  -1,   0,  -1, -17,  31,  -9,   9,  19, -18,  -2,  -3,   2,  -2,   1,  -1,   6,  -1,  20,  -8,   6, -27,  -3,   0,   2,   1,  11,   8,  37,  -3, -35,   2,  -1,  -1,   2,  -3,   4,  43,  -8, -16,  -2,  -1,   1,   3,  -3,   3,  20,  -5,   0,  -3,   2,  -4,  -1,   0,   0,   1,   0,  -1,   1,  -3,  -1,   0,   0,   1 },
     {    0,  15,  36,   2,   0, -10,  -1,  -2,  -1,  -1,  -1,   0, -21,  -9, -11,  39,  -8,   8,  -6,   1,   0,   1,   0,   1, -51,  20,   0,  20,  19, -14,   4,  -3,   2,  -1,   0,  -1,  -5, -39, -19,   7,  18,  11,  -1,   2,   2,   1,   1,   0,  -5,  10, -25, -17,   3, -22,  11,   0,   8,  -5,  16, -16,  25,  -1, -33,   6,   1,   5,  -6,   0, -10,  35,  -2, -17,   1,   0,   2,  -4,  -2,  -4,  18,  -3,   2,  -1,   0,  -2,   0,   0,   1,  -1,   1,   0,   0,  -1,   0,  -1,   0,  -1 },
     {    3,  -5,  14,  -1,  10,  -8,  -2,  -2,  -2,  -1,   0,  -1, -10, -15,  15, -23,  -1,  13,  -1,  -3,   1,  -1,   0,   0,  17,  23,  27,  10,   5, -13,  -6,  -4,  -4,  -1,  -2,  -1,  -3, -29,  13, -18,  19,  27,  -4,  -1,  -2,   0,  -1,   1,  12,  -2,   9,   6, -27,   6, -13,   3,  -6,  12,  -8,  30,   8,  41,   0, -22,  -2,  -2,  -9,  -6, -14,   3,  61,   1,  -2,  -3,  -3,   0,   0, -25,  -3,  32,  -1,   0,  -5,  -2,   0,  -1,  -1,   1,   0,   0,  -2,   0,  -1,   0,   0,   1 },
     {   17,  24,  -7,   4,  -2,   3,  -3,   1,  -2,   0,  -1,   0,  17, -20, -57,   5,  18,   2,   7,  -1,   4,  -1,   2,  -1,  27, -44,  12, -36,  15,   6,  -2,   3,   0,   2,  -1,   1,   1,  -5, -39,  -7, -20,  20,   5,   1,   2,   2,   1,   1,   0,  19,   6, -36, -15, -20,   1,   3,   8,   2,  18,  10, -14,   7, -18,  -4,  -5,  10,   1,   6,  -5,  -3,  16, -10,   1,  -1,   8,   1,   4,  -5,  -1,   6,  -3,   3,   0,   2,   1,  -1,   4,   0,  -1,   2,   0,   1,   0,   0,   2,   0 },
     {    9,  17,  78, -19,  -8,  -6,  -5,  -3,  -2,  -1,  -1,  -1,  11,  56,  -4,  26, -12,  -8,  -4,  -2,  -1,  -1,  -1,  -1,  56,   1, -14,   8,  13,  -4,  -4,  -1,  -1,  -1,  -1,   0, -14,  24,   2, -11,   7,  11,   0,  -1,  -1,   0,   0,   0,   5,  -6,  10,   2,  -8,   1,   7,  -1,  -1,  -5,  -2,   7,   2,  -5,   1,   4,  -4,   0,  -3,   0,   3,   1,  -5,   2,  -3,  -6,   1,   1,   1,   3,   1,  -2,  -4,   0,  -2,  -1,  -1,  -2,   0,   0,  -1,   0,  -2,  -1,  -1,  -1,   0,   0 },
     {    1, -13,   9, -20,   9,  -2,   0,   1,   2,  -1,   0,   0,  13,  -2,  24, -15,  20, -13, -11,   1,   0,   0,   1,   0,   2, -22,   9, -24,  16,  -1,  -2,   2,   3,   1,   1,   0,  22, -10,  26, -18,  17, -11, -20,   6,  -2,   1,  -1,   1,   9, -22,  14, -15,  12,  26, -19,   7,   1,  -3,   0,  -1, -27,  21, -33,  10,  -4,  14, -10,  24, -17,  34, -19, -31,   0,   6,  -2,   4,  -6,   6,  34, -23,   1,   2,   0,   4,  -1,   2,  -3,   2,  -1,   0,   0,   1,   0,   1,  -2,   0 },
     {   -5, -17, -53,  -5,  17,   4,  -1,   2,  -1,   1,   1,   1,   8,  11,  22,  18, -15,   0,  -1,  -1,   0,  -1,   0,  -1,  54,  12,  -7,   6,  41,  -8,  -3,  -3,  -1,   0,  -1,   0, -26, -24,   7,   4,  -2,  34,   0,   5,   0,   1,   0,   1,  -2,  -6, -45,  -7,   1,   0,  16,   1,   1,   2,   9, -35,  -8,   2,  -2,   9,  -1,   2,   1,   7, -16,  -4,  -4,  -3,  -2,   2,   4,  -4,   3,  -1,   0,  -3,   0,   1,   0,   2,   0,   1,   1,  -2,   0,   0,   2,   1,  -1,   0,   1,  -1 },
     {    6,   9,  11,   3,  14,  -1, -10,  -1,   2,  -1,   1,   0, -12, -18, -16, -25, -16,  -7,   9,   7,   1,   3,   0,   1,  10,  19,  19,  18,  20,  -2, -14,  -5,   0,  -2,   0,   0, -13, -25, -22, -19, -13,  -2,  10,   9,   2,   5,   2,   2,  14,  23,  23,  14,  11,  -9, -29,  -7, -10, -16, -16,  -8,   5,  27,  23,  13,  -3,   5, -10,  -7, -29, -21, -39,   9,   0,   6,   1,  12,   8,  21,  -2, -43,   3,   1,   2,  -3,   0,   1,   1,   4,   1,   0,   0,  -1,   1,   2,   1,   3 },
     {    0,   0,  10,  32, -15,   7,  -4,   2,  -2,  -1,  -1,  -1,   4,  -3, -12,  -4,  28,  -7,   4,   1,   2,   0,   1,   0, -13, -31,  -2,  -8,   3,  17,  -2,   5,   1,   0,   0,   1, -83, -10,  16,  -3,  -4,   3,   8,   2,   2,   1,   3,   0,  12, -45,  -8,  19,   0,  -4,  -5,   5,  -8,  17, -31,  -9,  12,   4,  -6,   0,   5,  -1,   7, -18,  -4,   9,  -1,  -3,   3,   0,  -3,   1, -11,  -2,   6,  -2,   1,   4,   0,  -6,   2,  -2,   0,   1,   1,   1,   1,   0,   0,   0,   0,   2 },
     {   -7,  -2,   8, -42,   3,   7,   2,   1,   1,   0,   0,   0, -10,   4,  28, -26, -13,   8,   6,   0,   0,   0,   0,   0, -24,  32,  -1,  -8, -15,  16,  -2,   5,  -2,   3,  -1,   0, -51,  12,  -5,   1, -30,   4,  20,  -1,   7,   0,   3,  -1,   4, -12,  13, -29,  -7, -30,   6,  10,   5,  -8,  20,  11, -33,  -1, -22,   4,  -1,   4,  -2,  28,   1, -19,   3, -10,   3,  -5,   6,   0,  16,  -2,  -7,   0,   1,  -1,  -1,   8,   1,  -2,  -1,   0,  -1,  -1,   0,   3,   1,  -1,   0,   0 },
     {   -4,  -4, -12, -64,  14,  11,   6,   0,   0,  -1,   2,   0,  -4, -13, -46,   0, -22,  14,   4,   6,   1,  -1,   2,   0, -13, -39,  22,  33,  -7, -33,   2,   4,  -1,   1,   0,   0, -19,  31,  28,  -4,  20,  -1, -17,  -1,  -2,  -2,  -1,   0,  24,  -2,  -8,   8,  -1,   6,   8,  -6,   9,  -3,   0,  -7,   5,  -3,   3,   7,   1,   5,  -1,  -2, -12,  -1,   1,   5,   5,  -2,   5,  -1,   3,  -7,  -6,   0,   1,   4,   0,   3,  -1,  -1,   1,   0,  -1,   1,  -1,   1,   0,   0,   1,   0 },
     {   -4,   4, -12,   8, -22,  13, -10,   1,   0,   1,   1,  -1,  -4,   7, -13,  20, -21,  20, -20,   6,  -1,   3,  -1,   0,  -5,   7,  -9,  16, -22,  17, -14,  -4,   1,   2,   0,   0,  -6,  13, -13,  16, -21,  17, -18,  -4,   2,   2,  -2,   0,  -4,  16, -11,  23, -19,  28,  -8, -14, -11,  16, -24,  16, -25,  10,   6,  -8,  -9,  13, -12,  14,  -6,  -3,  10, -43,   6,  -1,   5,  -6,  18, -15,  44, -27,   1,  -5,   6,  -5,   3,  -3,   2,  -3,  -1,  -1,   1,  -2,  -1,  -1,   2,  -1 },
     {    3,   5,  11,  35,   8, -11,   6,   0,  -3,   0,  -1,   0,   0,  -6,   3, -14,  16,  22,  -8,  -1,   4,  -3,   0,  -1,  -4,   5,  17,  -3,  -6,  -9,  19,  -8,   1,   4,  -2,   0,   8,  37,   4,  -6,   2,  20,  -5,  -3,   1,   1,  -1,  -1, -34, -10,  15,   7,  -5,   1,  43,   0,  -1, -31, -18, -31,  -6,   1,  -9,  31,  -4,   0,  -1,   5, -48, -11,   2,  -5,   4,  -2,  -2,   8,  11, -33, -14,  -8,  -1,  -3,   2,   2,   4,  -6,   1,  -2,  -4,   1,  -1,   2,   0,  -1,   1,   1 },
     {    6,   9,  14, -22,  23,   0,   6,  -8,   0,  -2,   0,  -1,   2,  20, -20, -51,  24,  26,  -5,   3,   2,  -2,   1,   1,  10,   2, -57,  13, -28,  14,  13,   2,   2,   1,   2,   1,  24, -32, -14,   6, -10,  -9,   2,   0,   4,  -1,   2,   0,  21, -10, -28,   0,   5,  -1,   3,  -5,   5,  16, -10, -27,   1,   2,   0,   3,   7,   8,   9,  -6, -16,  -2,   4,  -3,   1,   2,   3,   9,   2, -14,   3,   7,  -2,   0,   5,  -4,   2,   3,   0,   2,   0,   0,   3,  -1,   1,  -1,   2,   2 },
     {    4,  -4,  -4,  26, -23,   2,   2,   2,  -2,  -1,   1,   0,   6,  -4,  -7,  32, -31,  16,  10,   0,   3,   3,  -1,  -1,  18, -15,   6,  22, -24,  -2,  15, -10,   5,   2,   2,  -1,  25, -40,  23,  -6,   5, -23,  32,  -9,   7,   3,  -2,  -1,  30, -36,  13,   6,  -2, -25,   5,  14,  16,  -2,  -3,   6, -17,  -1, -15,  11,  -2,   6,  -5,  24,  -2, -21,   0,  -4,   7,   6, -14,  -2,  17,  -4, -11,  -6,  -3,  -2,   4,   0,   6,  -5,  -1,  -2,  -5,   5,  -2,  -2,   3,  -2,   0,  -3 },
     {    0,  -1,  -2,  -4,   8,  20,  10, -15,   4,   3,   3,  -3,  -3,  -2, -12,  -1,   1, -34,  13,  -2,   6,  -2,  -1,  -1,   9, -10, -25,  28, -20,  -5,  -8,   6,  -8,   2,   2,  -1, -14, -40,  23,  -1, -18,   5, -16,   2,   4,   0,   1,  -1, -57,   8,  18, -11,   6,  12,  14,  -9,  33, -40, -21,  20,   3,  11,  -6,   1,  -6,  -7,   5,   3,  11,   3,   4,  -8,  -6,  -7,   3,  -6,  -2,   1,  -4,   3,   1,  -9,  11,   5,  -1,   5,  -8,   2,  -2,  -3,   7,   0,  -1,   1,  -3,   4 },
     {   -3,  -2, -14, -44,  12,  24, -19,   4,  -3,   4,   1,   2,  -5,  -8, -10,  20,  16,  -8,   1, -10,  -2,  -1,  -1,   0,   6,  22,  22,  14,   7,  40,  -7,  -5,   6,  -3,  -1,   1,  32,   7, -17,  -5,  -1,  13,  30,  -9,   3,  -2,   1,   0, -17, -39,  11,   4,   1, -14,  12,  18, -12,  -6, -40,   0,   7,  -6,  -6,   2,   3,   7,  -9, -33,   7,   6,   3, -11,   5,  -2,  -2,   0, -20,   4,  10,   1,  -5,   5,  -5,  -2,   1,   2,   4,  -3,   0,   1,   0,  -1,   4,  -1,  -1,   0 },
     {    3,   4,   7,   5,  10,  12,  11, -12,  -6,   1,  -1,   0,  -4, -11,  -7,  -9, -13, -10,  -5,  -3,   8,   1,  -1,   0,   9,   7,  23,   6,  15,  13,  31, -19,  -8,  -1,   1,  -3, -12,  -7, -10, -25,  -5, -21,  -3,   2,   2,   1,   3,  -1,  13,  18,  13,  16,  20,  10,  16, -12, -21,  -8,  -5, -26, -16, -21,   9,  10,  10,   3,  27,  22,  22,   4,   9, -25,  -1, -24, -10,  -8,   8,  12,  31,  53,  -3,   3,   5,  -5,   4,   0,  -4,  -2,   0,   4,   0,  -2,  -2,  -3,  -1,   1 },
     {    2,   7,  15,  14,   9,  -8,  -6,   7,  -4,   3,   1,  -1,  -9,  -9, -22, -38, -27, -11,  -1,   3,   0,  -1,  -3,   3,  -5,   1,   3,  16,  21,   0, -20,  -4,   2,  -4,  -5,  -1,  15,  20,   1,  -5,   4,  16,   2,  -4,  -2,   8,  -1,   2, -46, -32, -17,   9,   5,  -5,  -8,   6,  27,  48,   4, -14, -11,   7,  -8, -10,  -9,   0,  11,  28,  22,  -2,   1,   4,   1,   5, -24, -13,   9,  23,  -4,  -3,  -7,  12,   9,  -6,   3,  -7,  -1,  -3,   3,   3,  -2,   3,   2,  -1,  -2,   2 },
     {    0,   1,  -7,   2,  60, -49,  15,  -5,   9,   7,  -3,  -3,   1,   2,   5,  26, -26,  22,   4,   3,  -3,   4,   1,   0,  -2,   0,   3, -25, -13, -13,  -2,   4,  -7,  -1,   0,   1,  -9,  -7, -18,  14, -16,   3, -17,   2, -10,  -2,   4,  -3,   9, -19,  27, -15,   9,   5,  15,  -9,   5,  24, -34,  17,  -1,   4,   4,  11,  -8,   9,  11, -11,  17,   6, -11,   5,   5,  -5, -13,   5,  -5,   1,   4,   1,  -7,  10,  11,  -5,   1,   1,  -4,   4,   3,   2,   2,  -3,  -1,  -2,  -2,   1 },
     {    2,  -2,   3,   6, -17,  24, -29,   7,   0,  -8,   6,   3,  -1,  -9,   2, -16,   1, -24,   2, -16,  12,  -3,   1,  -2, -10,  10, -13,  -2,  24, -31,   5,  16, -15,   7,  -2,   3,  21,  -2, -14,  30,  -7,  23, -37,   4,   8, -11,   3,   0,  44, -34,  18,  -7,  -2,  -5,  14, -22,  -1,  -2,  -9,   3,   8,  -4,   2,  14, -13, -19,  10,  -2,  13, -11,   4, -11,  15, -16,   9, -13,  15,   6,  -7,  -7,  -6,  -4,  -2,   4,   6,  -9,   1,  -5,   0,  -2,   3,   1,   1,  -1,  -2,   0 },
     {    3,   3,   7,  23,  16,  11, -15, -17,   3,   7,  -2,  -1,   2,   1,  12,   7,  29,  -5,   6,   2, -11,   2,  -4,  -2,  -8,   6,  13,  21, -10, -11, -21,   3, -12,  -4,   0,  -9,  10,  23,  35,  -4, -23,  24,  25, -12,  -7,  -7,  -3,   0,  38,  30,  -9, -22,  -6,   2,  -2,  11,  27,   1,  -7, -21,  -4,   6,  -5,  -5, -14,  27,  19,  -7,  17,  -5,  -5,  -6,   0, -12,   1,   9, -11,   8,   6,   2,  -5,  -3,  24, -12,  10,  -6,  -9,  14,  -7,   3,  -2,  -7,   5,  -4,   3,   0 },
     {   -1,  -3,   1,  -8, -33,  13,  20, -15,   2,   3,   5,   2,  -2,  -6,   1,  -5, -29, -23, -19,  11,  -8,  -3,   0,  -1,  -3,   1,   2,  -8,   3,  15,  -4, -22,   2,  -8,  -4,  -6,   2,   1,  -1,   7, -21,  11,  19,  -1,   9,   5,   1,   5,  12,  -8,  15, -23,  -2,  19,  10, -21,  26,  15,  17,  -2,  14,   3,  29,  14,  -8, -13,   9,  -7, -28,  24,  -5,  12,  11,  23, -25,  26, -23, -25,  12,  11, -19,   2,   6,  -1,  10,   5,   3,   1,  -3,  -6,   9,  10,   5,   0,  -3,  -5 },
     {   -2,  -1,  -2, -15,   7,   6, -23,  -2,  -1,   6,   1,   1,   1,   3,   5,  19,  32,  -8, -22, -29,  11,  -1,  -4,  -1,   6,  11,  16,   8,  -2, -19,  46, -19,   0,   6, -14,   5, -12,   0, -10, -12, -15, -27,  -8,  -3,   1,  -4,  -6,   7,  -6,  14,  -5,  -4,  -2,  -6,  -9, -13,  29,  34, -17,   2,  -5,  -3,  -1, -18,   0, -30,   5,   9,  -6,   0,   7,  -8, -15,   9, -24,  21,   1,  -3, -19, -26,  10,   6,   6,   9,   9,  -7,   6,  -6,  -3,  -5,  -6,  -9,   3,   2,  -6, -13 },
   },
   {
     {   94, -66,   6,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,  18,   1, -16,   7,  -2,   1,  -1,   1,   0,   0,   0,   0, -31,  32,  -8,  -2,   2,  -1,   0,   0,   0,   0,   0,   0, -14,   7,   8,  -5,   1,   0,   0,   0,   0,   0,   0,   0,  -6,   1,   4,   0,  -1,   1,   0,   0,  -3,   0,   1,   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0 },
     {  -31,  -2,   3,   7,   1,   1,   1,   1,   0,   0,   0,   0,  85, -67,   2,   4,  -1,   1,  -1,   1,   0,   0,   0,   0,   7,  16, -26,   7,  -3,   1,   0,   1,   0,   0,   0,   0, -30,  35,  -7,  -7,   4,  -2,   1,  -1,   1,   0,   0,   0,  -8,   3,   9,  -7,   1,   0,   0,   0,  -3,   0,   4,   0,  -2,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0 },
     {  -31, -52,  81, -25,  10,  -5,   4,  -2,   2,  -1,   1,  -1, -33, -16,  26,  11,  -9,   5,  -2,   2,  -1,   1,  -1,   1,  -2,   7, -31,  28,  -8,   2,  -1,   1,   0,   0,   0,   0,  19,  -2, -19,   2,   9,  -4,   1,   0,   0,   0,   0,   0,   9,   5, -10,  -6,   4,   1,  -1,   1,   2,   6,  -4,  -3,  -1,   1,   0,   0,   2,   4,  -1,  -2,  -1,   0,   0,   0,   1,   2,  -1,  -1,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0,   1,   0,   0 },
     {  -46,   3,   7,  -2,   6,  -1,   1,   0,   1,   0,   0,   0,  15,  21, -34,   4,  -2,   1,  -1,   0,   0,   0,   0,   0, -50,  64, -11, -17,   5,  -3,   2,  -2,   1,  -1,   1,  -1,  13, -31,  45, -19,  -1,   0,  -1,   0,   0,   0,   0,   0,  22, -27,   9,  13, -11,   3,  -3,   2,   2,   0,  -8,   9,   1,  -3,   0,  -1,   2,  -1,  -2,   1,   2,   0,   0,   0,   3,  -3,   0,   0,   1,   0,   0,   0,   1,  -1,  -1,   1,   1,   0,   0,   0,   1,  -1,   0,   0,   1,  -1,   0,   0 },
     {   18,  33,   7, -27,  -3,  -1,   0,  -1,   0,  -1,   0,  -1,   2, -16,  69, -46,   2,  -2,   1,  -1,   0,   0,   0,   0, -48,  23,  24,  10, -23,   7,  -3,   1,  -2,   1,  -1,   1, -20,   5,  -3,  29, -13,  -4,   2,  -1,   1,  -1,   0,   0,  17, -23,  -4,  15,   7,  -7,   1,   0,   8,  -5, -13,   8,   7,  -1,  -3,   1,   2,   0,  -8,   1,   5,   1,  -1,  -1,   3,   0,  -5,   0,   2,   1,   0,  -1,   1,   0,  -3,   0,   1,   0,  -2,   0,   1,  -1,  -1,   0,   0,   0,  -1,   0 },
     {    1,  13, -29,  42, -19,   7,  -5,   3,  -2,   1,  -1,   1, -55, -18,  17,  24,   4,  -4,   2,  -1,   1,   0,   1,   0, -28,  17, -34,  13,  17,  -8,   3,  -1,   1,   0,   1,   0, -20,  31, -19, -24,  26,  -5,   1,  -1,   1,   0,   1,  -1,  23, -25,  23, -28,   3,   9,  -4,   2,  11,  -6,   4,  -2, -12,   7,   0,   0,   3,   0,   1,   0,  -5,  -1,   1,   0,   3,   0,   1,   0,  -2,  -1,   0,   0,   4,  -2,   1,  -1,   2,   0,   0,   0,   2,  -1,   1,  -1,   1,   0,   0,   0 },
     {   31,  48,   0, -56,  24, -11,   4,  -5,   1,  -2,   1,  -1,   5,  23, -12,  -7,  -5,   7,  -3,   2,  -2,   1,  -1,   0,  33,   9, -55,  37, -20,   6,  -1,   1,   0,   0,   0,   0,   4,   3,   8, -18,  15,  -9,   4,  -2,   2,  -1,   0,  -1,   0, -17,  27, -14,  -1,   5,  -4,   2,  -8,  -1,   5,  10, -11,   4,   0,  -1,  -5,  -3,   3,   5,   0,  -3,   2,  -1,  -4,  -2,   2,   1,   1,   0,  -1,   1,  -2,  -3,   2,   1,  -2,  -1,   1,   2,  -1,  -1,   0,   0,  -1,  -1,   1,   0 },
     {   22,   2,  -8,  -4,   7,  -1,   0,   0,   0,   0,   0,   0, -19, -63,  24,  -2,  15,   3,   1,   2,   1,   1,   0,   1,  20, -16, -14, -24,  15,   4,   1,   1,   1,   1,   0,   0,  27, -20,  34, -36, -13,  17,  -6,   2,  -2,   2,  -1,   1,  16, -10,  -3,  24, -38,  10,   1,  -1, -15,  29, -21,  17,  -4, -15,   9,  -3,  -4,   7,  -2,   0,   6,  -8,  -2,   1,  -1,   4,  -1,  -1,   1,  -1,  -2,  -1,  -1,   3,   0,   0,  -2,   3,  -1,   0,  -1,   1,   0,   1,  -1,   2,  -1,   0 },
     {   15,   9, -15,  28, -24,   4,  -6,   3,  -2,   1,  -1,   1,  35,  -8,  10,  11, -11,  -5,  -1,  -1,  -1,   0,   0,   0, -10,  -2,  -4,  40,  -7,  -7,  -1,  -1,  -1,   0,   0,   0,  56, -53, -18,  19,  24, -12,   2,  -1,   0,  -1,   1,   0,  20,  -9, -13, -15,  21,   5,  -7,   4, -22,  27, -12, -13,   4,   8,  -2,  -1, -14,  11,   7, -14,   0,   2,   3,   0,  -5,   0,   6,  -5,  -2,   1,   0,   1,  -3,   1,   2,  -3,  -2,   1,   1,  -1,  -2,   1,   0,  -1,  -1,   1,   1,  -1 },
     {  -14, -40,  -9,   6,  -2,  12,  -2,   4,   0,   2,   0,   1,  19,  34,   7, -62,  30, -11,   7,  -4,   2,  -2,   1,  -1,  18,   0,  -2, -12,   8,   7,  -4,   4,  -2,   1,  -1,   1,  15,  10, -39,  -1,  -3,  15,  -2,   1,  -1,   1,   0,   1,  27, -28,  29, -19, -19,  13,   1,   0, -10,   7,  -2,  17, -25,   2,   8,  -1, -13,  11,  -1,   7,  -1, -14,   8,   1,  -3,  -2,   2,   3,   2,  -5,  -3,   4,  -1,  -1,   1,   2,  -2,   1,  -1,   2,  -1,   0,   1,   0,  -1,   1,   0,   1 },
     {  -10, -21,  -3,  -3,  32, -13,   5,  -3,   2,   0,   2,   0,  11,   0,   2,  16,   2,   4,  -8,  -1,  -2,   1,  -1,   0, -22, -53,  44,  33,  -9,   8,  -2,  -1,   0,   0,   0,   0, -35, -22,  26, -20,  30,  -4,   5,   0,   1,   1,   0,   0,  -2, -20,  19, -20,   0,  20,  -7,   5,   7,  -2, -23,  18, -25,  11,   4,   0,   4,   9, -16,   2,  -2,  -8,   6,   0,   3,   5,  -5,  -4,   0,  -3,  -2,   0,   2,   5,  -3,  -2,   1,   3,  -2,   0,   1,   2,  -1,  -1,   1,   1,   0,  -1 },
     {  -13, -16, -48,  -4,  52, -14,   8,  -3,   4,   0,   2,   0, -19, -23, -24,  -7,  39,   6,  -4,   2,   0,   1,   0,   1, -34,  -3, -22,  19, -25,  34, -10,   5,  -1,   2,   0,   1,  11, -10, -16,  19, -30,   8,   9,  -4,   2,  -1,   1,   0,  -5,  15,  -3,  -8,   5, -19,  13,  -3,  -2,   7,  11, -18,   3,  -7,  -1,   3,   2,   4,   8,  -3,  -7,   2,  -3,   0,   5,   0,   6,   0,  -4,  -1,   1,  -2,   3,   1,   3,   0,   3,   1,   2,   0,   2,   0,   1,   0,   1,   0,   1,   0 },
     {   -2,  12,  -4,  -9,   0,   1,   5,  -2,   1,   0,   0,   0,  18,  20,  29, -18,  -7,   2,   0,  -1,  -1,   0,  -1,   0, -52, -14,  -9,   7,  24,  -2,   4,   1,   2,   0,   1,   0,  24,   0, -19, -45,  20,  13,  -6,   5,  -1,   1,   0,   1, -39,  51,  -8,  -4, -29,  16,   1,  -2,  -2,  -3,  14,  -2,  -8, -13,   9,  -2,  12, -18,  15,   3,  -6,  -3,  -7,   4,   5,  -2,  -5,  11,  -2,  -2,  -2,  -2,   2,  -1,   0,   3,   2,  -2,  -1,   3,   0,  -1,   1,   1,   1,   0,  -1,   1 },
     {  -13, -22, -23,  18, -24,  19,  -2,   4,   1,   2,   0,   2,   0,  26,   7,   2, -44,  25,  -5,   4,  -3,   1,  -2,   2,   2, -14, -13,  28, -39,  15,   7,   1,   2,  -1,   1,   0, -11,  15,   2, -16, -16,  -8,  17,  -5,   5,  -2,   2,  -1,   1,  -6,   9,  22, -23, -25,  15,  -2,  -8,  12, -21,  17,  23, -39,   3,   3,   4,  -1,   4, -12,  26, -10, -14,   5,   2,  -2,  -1,  -2,  10,   2,  -6,  -3,   2,  -2,   2,  -2,   1,   1,   0,   0,   1,  -1,   1,  -1,   0,   0,   0,   0 },
     {   -9, -16,  -6,  22, -15,  -4,   1,   1,   0,   1,   0,   0, -17,  -4,  30, -22,   5, -16,   3,   0,   0,   0,   0,   0,  20,  29,   9,   1, -28,   8, -10,   5,  -2,   0,  -1,   0,   9,  -9,  49, -10,  -2, -15,   0,  -5,   2,  -2,   0,  -1, -46,  25,  37, -25,  27,  -9,  -4,   0, -31,  14,   9,   3,  -4,  16,  -8,  -1,   5, -11,  -3,  14,  -4,   6,   5,  -3,  10, -12,  -9,   6,   1,   1,   2,   2,   6,  -4,  -5,   3,   1,   1,  -4,   1,   2,   0,  -2,   1,   1,   0,  -2,   2 },
     {    0,  17,  17,   5,   3,  -2,  -6,  -2,   1,  -1,   1,  -1, -13,   6, -18,  -4,  19,  -3,   4,  -3,   1,  -2,   1,  -1,   6,  25,  12,   4,  -4,   0,  -6,  -2,   0,  -2,  -1,   0, -50,  -4, -16,  10,  20,   8,   8,  -1,   2,  -1,   1,   0, -12,   4, -25,  -6, -17,  21,   5,  -1, -39,  59, -23,  -9,  -4, -13,  19,  -6,  11, -14,  20, -25,   7, -11,   4,   9,  17, -17,   9,   4, -14,   6,  -8,   8,   9,  -6,   0,   4,   2,  -4,   1,   0,   1,  -2,   0,  -1,   0,   0,  -1,   1 },
     {    8,   9,  -1,  18,   8, -39,  18,  -7,   0,  -2,  -1,  -1,   2,  -2,   0,   9, -22,   8,  -7,   6,  -3,   2,   0,   1,  16,  32,  30, -20, -35,  41, -25,   3,  -2,  -2,   0,  -1,  26,  20, -15, -26,  18,   8,   6, -13,   6,  -3,   2,  -1,  26,  19, -28, -14,   7,   7,   7,  -6,  23,  -8, -14,  14, -16,   3,  11,  -1,   6,  -8,   0,   5,   3, -15,   8,   4, -10,  -1,   6,   0,   4,  -6,  -4,   4,  -8,   0,   2,   1,  -5,  -1,   2,   0,  -3,   0,   0,  -1,  -1,   0,   1,   0 },
     {   -7, -19, -20,   9,  28, -40,  14,   0,   4,  -1,   1,  -1,  10,  29,  30,  -4,  14, -40,   0,  -1,   1,  -1,  -1,  -2,  17,  -1, -12,   7,  12, -11,  -8,   3,   0,   1,  -1,   1, -10,  16,  10, -21,  17,  -8, -14,   0,   0,   2,  -2,   0,  17,  -4, -17,  27,   7,  -3, -14,   0,   7,  10, -20, -11,  37,   6, -19,   0,   0,  -6,  15, -20,   3,  37, -12,  -6,  -8,   0,   1,   0, -14,  16,  11,  -9,  -1,  -3,   4,   2,  -2,  -1,  -1,   0,   1,  -1,   2,   0,  -2,   1,   0,  -1 },
     {    5,  -2,   5,  15,  -4,  -7,   3,   1,  -1,  -1,  -1,   1,  -7, -21, -12, -28,   7,   4,   8,   4,   2,   0,   2,   1,  27,  23,  20,  22, -17, -20,   7,  -5,   1,  -1,  -1,  -2, -18, -49, -13, -12,  23,  11,   7,   5,  -1,   4,   0,   3,   5,  -3,  -1, -21, -36, -12,   7,   4,  25, -25,  29,   1,  18, -29, -14,   6,   9,  -5,   1,   4,  -6,  19, -28,   3, -10,  14, -12,  16,  -3,   5,   4, -15,  -2,   2,   1,  -4,  -1,   3,  -2,   3,  -1,   1,   0,  -1,  -1,   2,   0,   2 },
     {   10,  25,  40,  20,  -9, -42,   7,   2,  -4,   0,  -2,  -1,   9,   8,  -6,  27,  -2, -38,  12,   3,  -1,   0,   1,  -2, -17, -17,   9,   9,  -4,  13, -11,   4,   0,   1,   0,  -1,   6,   9,  -7,  -4, -35,  45, -11,  -3,   3,   0,   0,  -1,   2, -16,  38, -16, -14,  -7,  12, -11, -17,   9,  12,  -9,  20, -16,  -2,   5, -14,  11,  -9,   8,   2,   5,  -9,  -1,   8, -10,  -5,   5,   5,   3,   4,  -5,   4,  -1,  -4,  -2,   3,  -1,  -1,   0,   0,   0,  -1,   0,   1,   1,   0,   0 },
     {    0,   2,   5, -32,  -4,   3,   6,  -1,  -4,   2,  -1,   1,  -8, -21, -32, -19, -16,  11,  19,  -1,   4,   2,   2,   0, -22, -19,  30, -28,  -9,  -8,  13,   1,   0,   2,  -1,   2,  14,  25, -15, -35,  11, -28,   9,   6,  -1,   0,  -1,  -1,  18,  -3,   9, -23,  15, -26, -12,   9, -18,  32,  -2,   0,  11,   5, -33,   7, -11,   4,  12,  -3,  -2,  20, -17,  -7,   8,  -9,  11,   8,  -8,   7,   1,  -9,   0,  -2,  -1,  10,   2,   0,   0,   6,   1,   1,   0,   3,   1,   1,  -1,   1 },
     {    2,  -6,   5,  -4,  -4,  16, -13,   4,   2,   1,  -1,   1,  13,  11,  24,   5,  -8,  25, -14,  -1,  -2,  -2,  -2,   0,   7, -16,  -2, -16,   1,  29,   1,  -3,   1,  -2,   0,  -1, -15, -11,  21, -19,  -8,  14,  16,  -8,  -1,  -2,   1,   0,  33, -25,  -7, -12,  16,  -1,  16,  -8,  -6,   4,  31, -53,   6,   6,  12,   3,   9, -23,  31,   3, -32,   5,   3,  12,   2,  -2, -11,  24, -20,  -7,   4,   5,  -2,   5,  -8,  10,  -2,   3,  -5,   1,  -1,   1,  -3,   3,  -2,   3,  -2,   2 },
     {    9,  10,  26,  16, -31,   8,  -9,  11,  -7,  -3,  -1,   1,   7,   3, -23, -19,  36, -15,  14,  -8,   8,  -3,   5,  -2,  -8, -26, -28,   7, -10,  13,   1,   1,  -9,   8,  -2,   2,  -3,  -6,   9, -16, -17,   6,   4,   2,  -6,   2,   1,   0,  10,  10, -12,  -6,  29, -14,  -4,   3,  19,  -4, -37,  19,  -8,   4,  -5,   9,  54, -31,  -2,  -1,  -8,   1,  -6,  -1,  -4,   9,   6,  -3,  -4,   2,   1,  -4, -12,   9,   6,  -9,  -6,   5,   5,  -1,  -2,   1,   5,   0,  -2,   0,   0,   4 },
     {   -2,  -7,   0,  -2,  -7,  -2,   4,  -2,   0,   2,   0,  -1,   0,  -6,  -8,   5,  18, -15,  15,   3,  -4,   4,  -2,   0,  -7, -10,  11,   6, -21, -10,  -2,   3,   1,   1,  -1,   2,  37,  39,  25,  22,   7, -20,  -7,  -4,  -1,  -3,  -2,   0, -23, -45, -14,  -4,  -7,  28,   0,   8,  18,   9,  27,  -7, -13, -25,   6,   3,   5, -33,  23,  -5,  23, -14, -14,  12, -14,  20, -15,   9, -12,  10, -10,  -1, -10,   7,  -6,   1,   3,   0,  -1,  -3,   2,  -5,   1,   1,  -1,   5,  -4,  -1 },
     {    3,   7,  19,  26,  -8,  -6,  -3, -14,   2,   1,  -1,   1,  -9, -12, -28, -34,   2,  35, -15,   4,  -3,   4,  -1,   3,  -8, -10,   6,  21,   4,   6, -33,   8,  -2,  -4,   2,   2,  13,  15,  -4,  -5,  19,   4,  -8,  -9,   0,  -5,   3,   0, -21, -21,  16,  44, -10,  -1,   5, -10,  -6,  -4,   3,   2,  18,  26,  -3, -14,   1,  -6,   9, -16, -13,  24,  36, -11, -11,   5,  -8,   3,  -5, -15,  19,  16,   1,   4,  -2,  -5,  -1,   6,  -4,   2,   3,   1,  -2,  -3,   1,   0,   0,  -1 },
     {   -2,  -6, -18, -19,  18,  14, -29,  10,   2,   2,   2,   0,   2,  -4,  -4,  19,   0,  -1,  -4,  -4,   4,  -4,   1,  -1,  10,  21,  36,  42,   6, -34,  11, -11,  -3,  -1,  -1,   0,  23,  26,  -5,  -3, -22,  19,  -7,   0,  -6,   1,   0,   2,  24,  14,  13,  -3,  -9,  -3,  -4,  -6,   9,  23,  -8,   5,  17,  11,   3,  -8,  37, -22,  -5,  16,  -9,   0,  19,   0,   8, -14, -24,  17,   7, -10,   5,   9, -10,   3,  -5,  -2,  -6,  -1,   3,  -9,  -8,  -3,   0,  -2,  -4,  -1,  -2,  -2 },
     {   -4,  -1,  -6, -18, -24,  24,  -6,  17,  -6,   1,  -3,   1,  -2,   4,  -9, -14,   3,   6, -14,  10,  -4,   5,  -4,   1,  -4,   0,  -2,  10, -14,   0, -26,   7,   6,  -1,  -1,  -3,  -4,  15,  25,  -7,   0,  19, -37,  -1,   0,  -1,  -4,   1,  -3,   9, -21, -32,  -1,  17,  -1, -11,  33,  25,  -9, -29,   8, -16,   8, -14, -11,  37, -29,   6,  -1,  16,   4, -14, -14,   5,  12,  -6, -12,  -1,  17, -12,   9, -18,  23,  -2,   3,   0,  -5,   4,  -2,  -1,  -2,   6,  -2,   1,  -1,   3 },
     {   -8, -13, -22, -21, -24, -12,  36,   1,   6,  -1,   2,  -1, -10, -12, -13,  -9, -37, -21,  49,   6,  -1,  -4,   0,   3,  -3,  -5,  -6,   1, -24, -10,   1,  16,   0,  -2,  -3,   0, -18, -16,  -7,  -5, -15,   8, -20,  15,   7,  -2,  -2,   4,  11,   1,  -2,  12,  -3,  24, -22, -12, -17, -12,   4, -24,   1,  24,  17, -19,  22,   1,  11,  -4, -10, -17,  12,  -4,  -5,  15, -10,   1,   9, -10,   0,   7,  -1,   3,  -3,  -7,  -2,   1,   2,  -1,  -1,  -6,   3,   1,  -3,   1,   1,  -2 },
     {   -2,  -6, -11, -28, -11,   2,  11,   2,   0,   7,  -3,   0,  -1, -11,   1,   0, -11,   4,   4,   9,  -5,   1,   1,   3,   8,   8,   7,  -4,  38,   4, -28,  12,   2,  -3,  -4,   0,   7, -13,  -2,  12,   9,  14, -11,  -6,  11,   0,   0,  -1, -10,   2,  43, -17,  19, -13,  24, -15,  32,  -8, -27,   0,  11,  -7,  12,   5,  14,   2,  19, -47,  16,  -8,  -6,   9,  11,  -2,   4,   5, -18,  17, -16,   2,  -5,  19, -18,  14,  -5,  -5,  11,   6,  -1,   4,  -5,   2,   1,   0,   3,  -3 },
     {    0,   5,  10,   2,   9,   3, -36,  24,   0,  -2,  -2,   0,  -5,  -3,  -5,  -6,  24,  -7,  -6,  10,   8,   5,  -4,  -1,  -7,  -4,   7,  -2, -30, -15,  17, -14,  -2,   3,  -7,   6,   3,   8,   7, -20,   8,   3,   0,  -3,  -1,   2,  -2,  -4,  19,  29,  12,   2,  15,   9,  -6, -13, -24, -40,   1,  -1,  30,   2,  29,  -1, -26,  11,   3, -31,  18, -24,   9,   9, -23,  17,   4,   7, -16,  14,  -9,  12,   4,   4,  -6,  -2,  12,  -6,  -5,   1,  14, -11, -10,   3,  -2,   8,  -4,   0 },
     {   -9, -10, -14, -33, -44,   0,  -1,   6,   2,  -4,  -1,   5,  -2,  -2,   5,  28,  23,   5, -30,   2,  -5,   0,  -4,   0,   0,   7,  15,   8, -12,  15, -15,  -4,   1,  -3,  -1,   3, -10, -11, -15, -10, -13,  24,  -6,   8, -15,   1,   2,  -5,   5,   4,   4,  11,   2,  21, -10,   9, -15,  -9,  19,  28, -16,   8, -22,  -3, -11, -16,  34,   1,  18,  33,  -3,   0,  -2,  -2,  12, -29,  -1,  10,   8,  -5, -13,   4,  -2,  -6,  10, -15,  10,  -5,   3,  -2,  -5,  -6,  -5,   7,  -2,   0 },
     {    2,   2,   2,   6,   1,  -1, -14,   3,  11,   0,  -3,  -1,   2,   2,   7,   4, -12, -11,  -3,  -6,   4,  -2,  -1,   4,   2,   4,  -5, -15, -27,  11,  36, -15,   9,  -3,   5,  -1,   7,  -6,  -9,   3, -27,   3,  26, -14,   2,   1,   8,  -2, -28,  -3,  16,  -9, -32,  32, -22,  16,  43,  12,  -9,   0,  10,  32, -15,  -3,  -2,   7,   4, -37, -16,   8,   9,  -5,   5,   7,  22,  -1, -22, -11,   6,   0,  -1,   4,  -6,  13,  -5,   4,   0,   8,  -3,   6,  -1,   8,   6,   3,   0,   1 },
   },
   {
     { -104,  37,  14,   3,   3,   1,   1,   0,   1,   0,   0,   0,  55, -29,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   7,   5,  -4,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
     {    1,  82, -32, -10,  -8,  -2,  -3,  -1,  -1,   0,  -1,  -1, -57, -25,  45,   0,   6,   1,   2,   0,   1,   0,   1,   0,  37, -29, -12,   4,  -3,   2,   1,   1,   0,   0,   0,   0,  -1,  10,  -1,   0,  -3,  -2,   0,  -1,   0,   0,   0,   0,   4,  -7,   0,   3,   1,  -1,   0,   0,   0,   2,  -1,  -1,   2,  -1,   0,   0,   2,  -3,  -1,   1,   0,   1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
     {   48,  53, -45,  -5, -10,  -1,  -3,  -1,  -2,   0,  -1,   0,  68, -14,   6, -13,  -1,  -2,  -2,  -1,  -1,   0,  -1,   0, -51,  14,  25,  -1,  -5,   1,   1,   0,   0,   0,   0,   0,  -3, -12,   7,  15,  -2,  -2,   0,  -1,   0,   0,   0,   0, -10,  -3,  -1,   3,   7,  -1,   0,   0,  -1,  -4,   0,  -2,   1,   2,   0,   0,  -5,   1,  -1,   1,   0,   1,   1,   0,  -1,  -1,  -1,   0,   0,   0,   0,   1,  -2,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0 },
     {   30,  -9,  23, -20,   2,  -5,   1,  -2,   0,  -1,   0,  -1,   8, -84,  -2,  38,   1,   7,   1,   3,   0,   2,   0,   1,   4,  42, -50,  -9,   6,  -1,   1,   0,   0,   0,   0,   0, -11,  18,  21, -18,  -3,  -1,  -2,   0,  -1,   0,   0,   0,  -2,   0,  -1,   5,  -4,   0,   0,   0,  -2,   4,   2,  -2,   0,  -2,   0,   0,   0,  -1,   0,   1,   0,   0,  -1,   0,  -1,   2,   2,  -1,   1,  -1,   0,  -1,   0,   1,   0,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },
     {   -8,  -7,  -6,  -1,   9,  -1,   4,   0,   1,  -1,   1,   0,  -4, -16,  20,  11,  -5,  -1,   0,  -1,   0,   0,   0,   0, -65, -40, -18,  37,  26,   4,   3,   2,   1,   1,   1,   1,  48,  15, -35, -43,   8,   7,   0,   3,   0,   1,   0,   0,  -7,  14,  17,  -8, -19,   3,  -1,   0,  11,   3,  -4,   6,  -3,  -4,   0,   0,  -2,   1,   7,  -2,  -1,  -2,  -1,  -1,   6,   0,   0,   0,   0,  -1,   1,  -1,  -1,   3,   2,   0,   3,   1,  -1,   0,   0,   1,   2,  -1,   1,   0,  -1,   0 },
     {   11, -38,   0,  12,  -3,   2,  -1,   2,   1,   2,  -1,   1,  41, -13,  61, -38, -12,   1,  -4,   1,  -2,   0,  -2,   0,  30, -19, -20,  35, -24,   7,  -1,   0,   0,   1,   0,   1, -31,  30, -33,  16,   7, -12,   6,  -3,   3,  -1,   1,  -1,  -2,  -6,   4,  -8,  11,  -3,   0,   1,  -6,   6,  -5,  -4,   3,   0,   2,  -2,  -4,   3,  -1,  -1,   1,   1,  -1,   0,  -4,   2,  -4,   2,   0,   1,  -1,   1,  -1,  -1,   2,  -2,  -2,   0,  -1,   0,  -1,   0,   0,   0,  -1,   1,  -1,   0 },
     {  -12, -26, -87,  59,   2,  10,   0,   5,   0,   2,   0,   2,  -1, -21, -22,  -7,  17,   0,   6,   1,   2,  -1,   2,   0,  20,  14,  -3, -10,   9,  -7,   0,  -1,   0,  -1,   0,  -1, -13,   5,   5, -29,   8,   6,   2,   2,   0,   0,   0,   0,   9,   6,  12, -11, -19,   3,  -1,   1,  -4,   6,   6,   8,  -2,  -7,   0,   0,   3,  -1,   7,  -3,   0,  -2,  -3,   1,  -2,   1,   3,   1,  -1,   0,   0,  -2,   1,   1,   3,  -1,  -1,   1,   1,   1,   0,   0,   2,  -1,  -1,   1,   1,   0 },
     {   -8,  -9,  13,  21,  -6,   0,  -3,  -1,  -1,   0,  -1,   0, -38, -46,  20, -21,  19,   7,   5,   3,   2,   1,   1,   0, -22,  42,  30,   7, -30,   0,  -3,  -2,  -2,  -1,  -1,   0,  18, -44, -34,  32,  14,  -2,   5,   0,   1,   0,   0,   0,   7,  27, -12, -25,   4,   2,  -1,   2,   6,  -8,   6,   7,  -3,  -1,  -1,  -1,   0,   8,  -7,  -5,   3,   1,   0,   0,   3,  -1,   2,   1,  -3,   1,   0,   0,   1,   3,  -3,  -3,   1,  -1,   2,   0,   1,   2,  -1,  -1,   1,  -1,   1,   1 },
     {  -12, -21, -14,  15, -11,  10,  -3,   4,   0,   3,  -1,   2,  -8,  -3,  21, -23,   7,   0,   0,   0,   0,   0,   0,   0, -10,   8, -15, -45, -15,  11,   7,   2,   3,   2,   2,   1,  61,  26,  21,  17, -44, -17,  -4,  -5,  -1,  -2,  -1,  -1, -36, -10,   5,  32,  20, -14,   1,  -1,  13,  -3,  -7, -10,  12,   0,  -2,  -1, -12,   2,  -2,   8,   3,   5,  -2,   0,   3,  -1,  -5,  -2,   0,   2,   0,  -1,  -3,  -1,   3,   2,   2,  -1,  -1,  -1,  -3,   0,   0,   1,   2,   0,  -1,   0 },
     {  -11,   5, -10,   7,  -4,   1,   1,   2,   1,   1,   0,   0,  -5,  44,  47,  -7, -27,  -9,  -6,  -2,  -3,   0,  -1,   0,  -5,  63, -46,  28,  16, -14,   0,  -6,   0,  -2,   0,  -1,   8, -33,  31, -19,  24,   2,   1,  -1,   0,   0,   1,   0,  -2,  -9,  -3,   3,   3,   8,  -3,   2,   4, -13,   1,  -5,   1,   8,   3,  -1,  -3,   1,  -5,   0,  -4,   1,   4,   0,   0,  -6,   0,  -1,   0,  -1,   1,   3,   0,  -2,   0,   0,   1,  -2,   1,  -1,   0,   0,  -1,   1,   1,  -2,   0,  -1 },
     {  -19, -23,  -9, -16,  34,  -6,  12,  -4,   4,  -2,   3,   0, -17, -13,  38, -16,   5,  -5,  -3,  -2,   0,   0,   0,  -1, -43,  -3,  27,  -5,   3,   9,  -3,   6,  -2,   0,  -2,   0, -38,  -1,  34,  -6, -22,  11,   0,   3,  -1,   1,   0,   1,  47, -17, -25,  37, -12,  -4,  -2,   0,  -7,  16, -12,  -2,   4,  -2,  -2,   0,  12,  -5,   3,   3,  -2,  -3,   1,  -1,   1,   6,  -3,  -3,   1,  -2,   0,   0,   4,  -1,  -2,   4,   1,   3,  -3,   0,   3,  -1,   0,   0,   0,   2,  -1,   0 },
     {   10,  -1,  -1, -32,  16,  -1,   3,   0,   1,   0,   1,   0,  15,   1,   4, -34,  -7,  12,   2,   2,   0,   0,  -1,   1,  30,  -4, -10,  -2,  -6,   4,   1,   1,   1,   1,   0,   0,   8, -46,  -2, -19, -32,  22,   3,   7,   1,   3,   0,   2, -14,  48,   9,  19, -48, -25,   8,  -5,  -5,  -8,   8,  21,  26, -27,  -5,   1,  -1,   4,  -3,  -1,   5,   8, -12,  -1,  -2,   1,   2,   6,   3,   0,   2,  -5,   0,   2,  -1,   0,  -1,  -1,   2,   2,  -1,   1,   0,   1,  -1,   0,   1,   2 },
     {    1,  24,  22,   9, -15,  -6,  -8,  -2,  -2,   2,  -2,   0, -14,  11, -29, -41,  -1,  22,   1,  10,   0,   3,  -1,   2, -28,  22, -18,  -1, -34,  -3,  16,  -2,   5,  -2,   3,  -1,  -9,  23, -14, -23,   5, -23,  11,  -4,   8,  -2,   1,  -1,  33, -37,  38,   2, -21, -17,  -3,   4, -12,   4,  -3,  13,  22, -16,  -5,  -2,   2,  -3,  -2,   2,   4,  11, -12,  -1,  -5,   1,  -4,   7,   2,   4,  -1,  -2,   3,  -6,   3,  -1,  -2,  -1,   0,   3,   1,  -1,   1,  -1,   0,   0,   0,   1 },
     {    2,   6,  10,  33, -29,  16,  -8,   1,  -3,   1,  -3,   1,   3,   9,  -2,   6,  17,  -4,   1,   0,  -1,  -2,  -1,  -1, -29, -37, -58,  -9,   4,  13,  12,   7,   5,   3,   2,   0, -25, -23,  19,  16, -22, -15,   1,  -1,   1,   0,   1,   1,  36,  46, -14, -13,  15, -13,  -3,  -4,  -9,  -9,  19,  -8,   6,   5,  -1,  -1,   6,   3,  -4,   1,   1,   2,   0,   0,  -2,   0,   5,  -3,  -1,   2,   1,   1,   3,   2,  -1,  -1,  -1,   0,   3,  -2,   2,   1,  -1,  -1,   0,  -1,   2,  -1 },
     {   11,   9,  33,  28, -13,  -7,  -2,  -3,  -1,  -2,   0,  -1,   2,  -8,  -9, -49,  40,  -3,   7,  -3,   3,  -1,   2,  -2,   0, -20,  -9,  17,   3,   4,  -5,   3,  -3,   0,  -1,   1,  -3,  -3,  46,  -7,  15,  19, -10,   2,  -5,   1,  -2,   1, -32, -16, -25,  -2, -23,  32,  11,   3,  24,  13,  -6,   6, -27, -22,   6,   2,  -9,   0,   5,   3,   6,  -9,  -6,   2,   7,   4,  -2,   2,  -3,  -3,   0,  -5,  -4,   1,   0,   1,   4,   0,  -3,   0,  -2,   1,   1,   0,   1,   0,   0,   0 },
     {  -16, -11, -36, -39,  30,   1,   6,  -3,   4,   2,   1,   1, -16,  -3, -30,  -2, -15,  25,  -5,   9,  -2,   2,   0,   1, -16, -13, -18,  35, -35, -10,   9,   2,   4,  -1,   2,  -1, -11,   1,  31,  25,  18, -28,  -3,  -6,   1,  -3,  -1,   0, -23,   7, -13, -20,  17,  -7,   8,   1,  29,   2,   8, -10,   1, -19, -13,   3,   3,   4,   6,   2,   5,  14, -15,  -5,   5,   4,   2,  -1,  -2,   5,   6,  -6,  -3,  -2,   1,   0,   2,  -1,   1,   0,  -1,   1,   1,   0,   1,   0,   1,   0 },
     {    2,  -2, -14, -37,  23, -10,   9,  -3,   5,  -2,   3,  -1,  16,  20,  11,  -2,  34,  -5,  -5,  -1,  -5,  -2,  -2,  -1,   7,   7, -21, -26, -29,  23,   2,   6,   3,   3,   0,   1,  37,   7,  -1,  -1,  10, -13,   0,  -4,  -2,  -1,   0,  -1,  31,   3,  -7, -30, -10,  34,  -6,   5, -33,   9,  18,  15, -38,  -7,  14,  -4,   5, -10,  -1,   7,  17, -15,   0,   6,  -8,   0,   3,  -4,  -6,   2,  -4,   1,   3,  -1,  -2,   5,  -3,   2,   2,  -1,  -1,   0,   0,   4,  -3,   0,   1,  -1 },
     {    9,   1,   3,   1,  -9,  -2,   4,   1,  -1,  -1,   0,   0,   4, -15, -24, -25,  -4,  -4,  -1,   5,   2,   3,   0,   1,  20,  -2,  20,  53,  -4, -14, -15,  -4,  -3,  -2,  -1,  -1,  49,  26,  38,  -7,   5,  -1,   1,  -6,  -5,  -3,  -1,   0,  39,  29, -15,  18,  16,  -4, -10,   2, -29,  -1,  11,  -5,  18,  25,  -4,  -3,  -5,  -6,  -6,  -4, -11,   7,  14,   3, -10,  -1,   0,  -5,   1,  -7,   1,   4,   1,   0,  -2,  -2,  -3,   2,   2,   1,   0,  -1,  -1,   0,  -2,   0,   0,   0 },
     {  -13,  -7,  -8, -45, -34,  19,   4,   4,   5,   2,   2,   1,  -9,   5,  -2, -15,  40, -33,   7,  -6,   3,  -4,   2,  -1,  -9,  21,   7,  18,  11,  10, -19,  -2,  -6,   1,  -1,   0, -28,  35,   8,  -1, -16,   8,   1,  -2,   0,   0,   0,  -1,  -5,  20,  38, -12,   7,  -3,  -4,  -3,  23, -38,  -3,  24, -10,  10,   0,   2,   5,   9, -14,  -2,   6,  -4,   2,  -2,  -3,  -9,   3,   0,  -3,   5,   3,   0,   3,   2,  -4,   1,   1,  -3,   2,   2,   1,   1,  -1,   1,  -1,  -2,   0,   0 },
     {   -7, -20,  -8, -26, -76,  24,   1,  18,   1,   6,  -1,   5,   4,  -9,  13,  12, -28,   2,  -2,   4,  -1,   2,  -2,   3,   1, -12,  21, -10,  12, -21,  12, -11,   6,  -2,   3,  -1,  12,  -5,   8,  -1,  10,  -7,  -4,   2,  -3,   0,  -3,   1,  21, -13, -22,  -7, -17,  11,  10,   5,   1,  16,   7,   5,  -2, -35,   1,   3,   0,   0, -20,  11,  14,   9, -25,   0, -13,   3,   3,   4,  -3,   7,   2, -11,   1,  -5,   2,   4,  -1,   3,   1,  -1,   0,  -1,  -1,   3,  -2,   2,   0,  -1 },
     {   -1,  12,  -1,   5,  10, -17,   1,   0,   4,  -2,   1,  -1,  -6,  13,  -5, -18, -10,  -4,   5,  -3,   0,  -1,   1,   0, -10,  26, -11,  -7,  31,  -9, -15,   4,  -2,   2,  -1,   1,  -3,  50, -29,  22, -12,  33,  -6,  -5,  -3,  -2,   0,  -1,  -1,   5, -48, -18, -14, -16,  20,   0,   5,   0,  52, -14,   6, -18, -14,  -4,   2, -10,  -3,  11,   5,  12,   1,  -4,  -7,  -3,   5,  -4,   1,   0,   3,   2,   1,  -8,   2,   6,  -2,   0,   2,  -1,   0,  -1,   1,   2,   0,   0,   1,   0 },
     {   -2,  -7, -16, -38,  -7,  14,   1,   9,   2,   1,  -1,   1,  -4, -12, -30, -41,   8,   1,  24,  -1,   8,   2,   2,   0,   3,   0, -32,   6,  35, -10,   0,   3,   1,   2,   0,   1,   6, -34, -32,  18, -19,   6,  -3,   8,   1,   4,   1,   1,  13, -31,   1,   0,  12,   8, -11,  -1,   0,  30, -10, -32,  -5,  23,  13,  -4,  -8,   1,  13,   1, -21, -12,   7,   8,  -2,   2,   0,   6,   1,  -2, -10,   1,  -3,  -3,   6,   5,   2,   3,   1,   1,   0,   0,   2,   1,   0,   2,   0,   1 },
     {    0,   0,  -4,  -5, -14,  17,  -5,   0,  -5,   3,  -3,   1,   9,  14,   0,  35,  50, -27,   0,  -4,   0,  -6,  -1,  -2,   9,  14,   0,  33, -24,  28,  -9,   5,   2,   1,  -1,   0,   7, -13, -16, -23, -14,   0,  25,  -2,   1,   1,   1,   1,  -5, -26, -35,   6,   4, -22,   6,   7,  -3,  26,   4, -11,  21, -13, -27,  -2,  -4,   3,  20, -12,  -4,  20,  -9, -13,  11,   3,  -2,   6,  -4,  -2,   8,  -6,  -3,   3,   0,  -4,   0,  -2,  -1,   0,  -2,   0,   3,   0,   0,  -1,   1,   0 },
     {    5,  -9,   5,  -4,  14, -35,   4, -11,  12,  -1,   4,   0,  11,  -6,  15,  -8,  28, -18,  12,  -6,   2,  -3,   4,  -3,   7, -20,   1, -24,  19, -33,   2,  -4,   5,   1,   4,   1,   7, -16,   9,  -6,  18, -15, -11,   7,  -5,   3,  -3,   3,   8, -18,  11, -23,   4, -11, -22,  17,  12,  -9,  11,   7,  10,  18, -43,  -2,  -4,  14, -18,  16, -16,  50,  -1, -16,  -7,   6,  -1,  11,  -9,   7,  15,  -6,  -1,  -3,  -3,   3,  -2,  -1,   0,   1,  -1,   0,  -1,   1,  -2,   1,  -2,  -2 },
     {    4,   4,   2,  12,  14, -19,   5,  -7,   1,  -2,   2,  -2,  -2,  -7, -10,  -9,  -6, -22,   2,  11,   1,   4,   2,   1,   7,   4,   6,  15,  19,  16,  -7, -10,   1,  -2,   0,  -1,   4,  -6, -11, -12, -17, -25,   7,   8,   3,   2,   2,   1,   1,  -4,  -3,  12,  20,  15,  -3,  -3, -18, -40,  -3, -37, -17, -24,  13,  24,  21,  18,   0,  17,   6,  -3, -57,  -7, -13,   0,  -3,   2,   5,  28,  15, -24,   9,   4,   1,   3,  -7,  -3,   1,   0,   5,   3,   0,   2,  -3,   0,   2,  -2 },
     {   -5,  -5,  -8,  -6, -22, -34,   8,  12,   6,   0,   2,   2,   1,   7,   3,  14,  29,   0, -19,  -7,  -7,   1,  -1,   2,  -5,   2, -11,   0, -13, -26,   6,   9,  -3,   2,  -2,  -1,   2,   9,  -6,  32,  17,  21, -36,   1,   1,   2,   1,  -2,   4,  13,  14,  20, -23,  22,  11, -16, -23, -14, -26, -34,  20, -15,   6,   2,  15,  16,  35, -17, -10,  18,  -1,  -7,  -1, -11,  -7,  12,  -2, -11,   8,   1,  -1,   2,   6,  -6,   1,  -2,  -5,   1,  -2,   0,   3,  -1,   2,   0,  -5,   1 },
     {    1,   1,  -2,   6,  -8, -39,  25,   3,   0,  -3,   3,  -1,   0,   0,  -4,   5, -10, -17, -13,   6,   3,   6,   0,   2,   5,   6,  -4,   0,  19,  32, -31, -10,  -6,   1,  -2,  -1,   3,  -7,   8,  26,   6,  -6,  11, -10,  -3,   1,   2,   0,  21,  15,  26, -11,  -2, -13,  -1,  -4,  25,  36, -24,   3,   8,  -7,   6,   5, -40, -39,  15,   0,  26,  14, -12,   1,   7, -10, -15,  -3,  -3,  18,   9,  -5,  -6,  -2,   2,   8,  -1,   8,  -6,  -2,   0,  -3,   0,   2,  -2,   1,  -1,  -2 },
     {    3,   5,   6,  -7,  -8,   1,  -1, -15,  -6,   1,  -7,   5,   6,   7,  11,  -8,  -3,  12, -28,  13,  -7,   0,  -2,   0,  10,  11,  17, -17,  -7,  -7,  18,   4,  -2,  -5,   2,  -2,  -1,  15,   2, -30, -13,   7,  15,   6,  -2,   2,  -5,   2,  25,  31,   0, -22,   9,   5, -19,   6,  44,   1,   3, -32, -17,  -6,   1,  -4, -17,   6,  38,  12, -40, -10,  -9,  -3,  16,  12, -13,  24,   4, -13,  -1,  -8,  -8,  -7,   1,  -1,   9,  -5,  -1,  10,  -8,   4,   1,  -1,   4,  -4,   0,   6 },
     {    9,  10,  10,   8,  30,  58,  -2, -12,   0,  -4,   2,   0,   3,  -1,   2,   3, -22, -38,  37,  -5,   6,  -3,   0,  -2,   1,   1,  -5,  -6, -15, -37, -15,  13,  -1,   4,  -5,   0,  10,   8,   6,  11,  -8,  13, -16,  -7,  -1,  -1,   4,  -1,  25,   1,   6, -23,  -4,   5,   8,   1,  -7,  -7, -32,  -1,  -1, -19, -13,   8, -16,   4,  17,  -2,  20,   6,  -8,  -9,  19,   2, -16,   2,   0,   8,   3,  -3,  -9,   2,   8,  -8,  10,  -3,   0,   4,  -7,   1,   5,   1,   3,  -2,  -3,   4 },
     {    0,  -1,  -6,  -6, -17,  -7,  28, -11,   6,  -4,   3,  -5,  -1,  -2,  -8, -12, -27, -18,  27,   2,   2,   2,   3,   0,  -3,  -9, -14, -31, -14,  20, -31,   6,   2,   4,   0,   2,  -1,   4, -11, -11,  41, -12,   0,   0,   2,  -4,  -1,  -2,   5,   3, -31,  29,   3,  -7,   3,   4,  11, -19,   4,  21,  -4,   9,   4,  -1,  -5,  23,  20, -43, -15,  -7,  -1,  -1,   9,   2,  17,  30,  10,  -3,   1, -13,  -9, -13, -14,   1,  -1,   2,   7,  -3,   0,   1,   1,   4,   4,   0,  -3,   1 },
     {    3,  -3,  -5,   7,   5, -26,   2,  14, -12,   3,   2,   2,   8,   0,   0,  12,   3,  38,  -4,  -7,   0,  -2,  -1,   0,   8,  -4,  -4,  16,  12, -12,  -2,   4,   2,  -2,  -1,  -6,  11,  -3,   3,   8, -45,   1,  10, -16,  -2,   0,   3,   0,  27, -31, -20, -21,   8,   6,  15,  -1,  17, -36, -18,  42,   1,  -8,   1,   7, -21,  23,  12, -19,  -1,  -7,   4,   0, -10, -26,  -2,  15,  -2,   1, -13,   0,   8,   9, -12,  -3,   0,   1,  14,  -5,  -8,   5,   1,   9,   8,   7,  -5,  -5 },
     {   -2,   0,   3,   2,  -1, -10,  -1,  -4,  -1,  -1,  -4,  -2,  -1,   2,   6,   3,   3, -10,  -5,  12,   0,  -5,   7,  -5,   1,   7,  11,  11,  15,  -9,   7,  12,   1,   1,   2,   1,  -3,   6,   8,   6, -22, -39, -16,  30,   6,   6,  -3,   4, -12,  -8, -14, -26, -21,   2, -36, -26, -14,  -5,   6,  14,  28,   0,  19, -12,  -1, -12,   4, -30,   2, -13,   0,   8,  21,  17,  22,  -9,   4,  10,   2, -20, -34, -23,  16,   4,  17,  -5,  -9,  10,  -4,   2,  11,   3,  14,  -2, -17,   5 },
   }
 }
};
#else
#if EXTENDED_LFNST
/* 
The following list of non-CTC sequences are used for training  matrices under macro EXTENDED_LFNST:
"Rollercoaster2_UHD",
"Drums100_UHD",
"BuildingHall_UHD",
"CrossWalk4K_long_UHD
"TrafficFlow_UHD",
"Traffic_UHD",
"NebutaFestival_10bit_1600"
"SteamLocomotiveTrain_10bit_1600"
"PeopleOnStreet_1600"
"jets_720p"
"Spincalendar_720p"
"Shield_720p"
"TimeLapse_HD"
"tractor_HD"
"station_HD"
"Kimono_HD"
"bluesky_HD"
"ParkScene_HD"
*/
const int8_t g_lfnst4x4[35][3][16][16] = 
{
  { // 0
    {
      {-110,  26,  29,   0,  42, -22,  -8,   1,  18,   1,  -9,   0,   0,   0,   1,   0 },
      {   8, 105, -38, -19, -25, -29,  33,   2,   0, -30,   5,   8,   6,   2,  -4,   0 },
      {  31,  22, -24,  -3, 100,  -7, -29,  -4, -52,  11,  18,   0, -13,  -4,   7,   2 },
      { -20,  18, -14,   1,   5, 106, -11, -32,   7, -41,  28,   7,   0, -20,  -4,   9 },
      {  33,  32,  99, -44,   0,  14, -25,  14, -11, -21, -26,  12,  -1,  -1,   1,  -2 },
      {  28,  13, -15,  -5,  28,   7, -11,  -3,  96,  14, -34,  -5, -56,   2,  21,   1 },
      {  -6,  34,   4, -19, -24,  25,  -5,  -5,  -9, 102,   7, -30,  10, -40,  16,  11 },
      {  -4,  10, -20,   4, -33, -17,-107,  19,  13, -15,  27, -20,  19,  14,  22,  -3 },
      {  20,  34,  31, 108,   6,   4,   1, -12,   2,  -2, -20, -35,  21,   0,  -5,   0 },
      { -14,   2,  14,  24, -32, -11,   1,   1, -40, -14,  21,  -5,-101, -19,  37,   7 },
      {  15,   1,  38,   2,  10, -22,  16,  -4,  41,   6, 104,  -5,  -3,   0, -31,  17 },
      {   1,   2,   9, -10, -12, -18, -12,-111,  -6,  12,  -4,  14,   0,  50,  16,  14 },
      {   7, -13,  -3,  -2,  -1, -38, -19, -35,   9, -25, -19,  11,  13, -99, -18,  40 },
      {   4, -12,   9, -13,  15,  -5,  34,   0,   8, -29,  15, -36,  39, -10, 102,  12 },
      {   3,   7,   5,  35,  -2,   1,  -4,  12,  10,  23,  15, 108,  17, -12,  42, -13 },
      {  -3,   2,  -5,   7,  -1,   9,   0,  31,  -6,   5, -12,  14,   1,  34,  -1, 117 },
    },
    {
      {  94, -61,  -5,   1, -54,  23,  11,  -1,  -5,  12,  -3,  -2,  -1,  -1,  -1,   1 },
      { -59, -48,  63,  -1, -35,  50,  -4,  -9,  44,   1, -28,   2,   0,  -7,   2,   2 },
      {  -1, -75,  31,  18,  78, -19, -17,  -3, -43,  28,   4,  -5,  -7,   9,   0,  -1 },
      {  35,  21,  19, -24,  36,  62, -65,  -6,  -7, -56,   3,  23, -16,  -2,  21,  -1 },
      { -26, -35, -86,  58,  12,  41, -13,  -3,  25, -16,  20, -11,  -9,  -6,  11,  -1 },
      { -26,  12,  -5,   0, -25,  54,   0, -16, -87,  23,  23,  -2,  50, -17, -16,   4 },
      {  21,  27, -12,   0,  45,  38,  34, -35,  15,  49, -63, -10, -11, -43,  -3,  23 },
      {   2,   9, -10,  -7, -17, -13, -88,  37,  20,  70,  -6, -12,  15, -27,  12,  -1 },
      { -23, -35, -47,-103,   7,   7,  13,   3,   2,  15,  12,  26, -21,   1,  -1,   0 },
      {  10, -22, -11, -10,  24, -18,   1,   9,  27, -36, -20,  26, 101, -31, -19,   4 },
      { -17, -13, -19,  12, -30, -36, -14, -13, -46, -24, -66,  41, -17, -24,  55,  15 },
      {   1,  -7,  23,  -3,   1, -12,  21,  15,   2, -16,  57, -14, -21, -99,  27,  18 },
      {   3,  -2,   0,  -3, -12, -30, -31,-109,  21,   8,  35,   9,   7,   0,  -6,  24 },
      {  -6,  -9,  -9,  -5, -11, -12, -28,   9, -17, -33, -26, -30, -34, -20, -93,  44 },
      {  -2,  -7,  -6, -32,   0,  -5,  -3, -12,  -7, -25, -18,-104,  25,  17,  47,  14 },
      {  -1,  -3,  -5,  -6,  -4,  -9,  -8, -29,  -7, -10, -18, -21, -11, -35, -22,-112 },
    },
    {
      { 117, -28, -24,  -3, -18,  14,  -1,   0, -27,   4,   8,   0,  -2,   0,   1,   0 },
      {  16,  67, -35,  -7,  91,  -9, -23,  -4, -28, -13,  13,   3, -15,  -2,   6,   1 },
      {  -5, -90,  18,  15,  70, -22, -27,   4,  -3,  36,  -8,  -7, -12,   6,   5,  -2 },
      {  -5,  -1, -12,  10, -20,-105,  24,  21, -49,  -7,  13,   4,  17,  26,  -8,  -7 },
      {  32,  22, 102, -28,  23, -11,  34,   2,  -8, -12, -36,  14,  -2,   3, -11,  -1 },
      {  30,  11, -23,   0,   6, -46,  -8,  11, 101, -11, -37,  -1,  -8,  24,   1,  -6 },
      {  -3, -12,  22,   7, -22,  -6, -78,  23, -16, -81,  -9,  13, -28,  -4,  28,  -3 },
      {   7,  41,  20,   4, -25,  -7, -67,  14, -15,  74, -29, -30,  28,  10,  21, -11 },
      {  11,  13,  12, 107,  15,  22,  17,  38,   2, -19,  -7, -21,  28,  -2, -17, -11 },
      {  -1, -17,  -9, -31,  18,   9, -26,  -7,   2, -31, -10,  16, 107,   7, -34,  -3 },
      {   4,   4,  29,  -6,   3,  18,  -7,   5,  19,  -2,  80,  -8,   4,  87,  13, -19 },
      { -14,  -6, -30,  -4,  -1,  32,  15,  -4, -35,  -7, -73,   3, -22,  77, -10, -29 },
      {   5,   2,  12,  30,  -6, -19, -10,-106,  -4, -21,  -7, -47,   3,  14,   0,  25 },
      {   0,   6,   6,   2, -11,  -3, -37,   1,   1,  12,  14,   5, -37,   3,-114,   6 },
      {   2,   8,   3,  41,  -5,   0, -11, -27,   2,  25,  -2, 106,   5,  20,  13,  36 },
      {   2,   1,   4,  11,  -2,  -8,  -4, -39,   5,   4,   9,  28,  -3, -29,  -1,-114 },
    },
  },
  { // 1
    {
      {-112, -23,  36,  10,  37,  11, -12,  -3,  17,   6,  -6,  -2,   1,  -1,   0,   0 },
      {  33,-116,  -2,  24,   6,  27,  -2,  -6,   3,  22,   0,  -5,  -4,   1,   0,   0 },
      {  30,  14,   6,  -9, 102,  20, -42, -10, -45, -15,  10,   7,  -8,  -3,   5,   2 },
      { -33,  -1,-104,  -6,  16,  34,  40,  -8, -11,  -1,  32,   7,   1,  -8,  -8,   1 },
      {  11,  19,  26, -10, -18, 102,  14, -29,  24, -32, -32,   9,  -6, -22,  -1,   8 },
      {  17,  19, -30,  -5,  34,  -6, -21,   3,  93,  49, -26, -16, -30, -12,  13,   4 },
      { -11,  -3, -30,   0, -42,  12,-102, -19,  12, -20,  29,  17,  17,  11,  25,   4 },
      {   3,  33,  20,  43, -14,  31,  -9, -18, -25,  82,  48, -32,  13, -22, -15,   3 },
      {   0,  14, -19, 106,   4, -14,   3, -20,  -6, -38, -34, -19, -15,  -2,  24,   8 },
      {  12,   8,  -3,  20,  24,  12,   5,   6,  34,  -8,  -5,  -9,  96,  54, -36, -18 },
      {  11,   0,  28,  24,  11,  -9,  21,  -8,  51, -37,  90,  29, -31,   3, -17,  -3 },
      {  -1,   7,  -1,  17,  -5,  41,  -9, 106,  -4,  -8,  12, -20, -24,  20,  20, -33 },
      {  -1,   7,   7, -13,  -1,  12,  23, -43,  -7,  22,   6,  -8, -28,  93,  58, -18 },
      {   5,  -3,  10,  -3,  11,  -3,  27,  16,  10,   2,  21,  12,  56, -36,  95,  31 },
      {   1,   9,  -1,  30,   0,   4,   1,  12,  -7,  40, -23, 112,  -1,   6,  -1, -20 },
      {  -1,   3,   0,   5,  -1,   9,  -1,  27,  -2,   7,  -1,  10, -11,  42, -19, 115 },
    },
    {
      {  85, -69,   2,   2, -55,  24,  18,  -4,  -2,  18, -10,  -3,  -1,  -2,  -2,   2 },
      { -60, -33,  50,   1, -22,  76, -20, -15,  30, -14, -29,  10,   3, -11,   9,   3 },
      {   1, -72,  49,  13,  78, -27, -20,  -4, -30,   4,  10,   1,  -7,  15,   1,  -5 },
      {  61,  29,   2, -12,  24,  31, -69,   4,   7, -59,   6,  29,  -6,  -4,  28,  -7 },
      { -23, -39, -91,  48,  16,  41, -10,  -4, -12, -14,  36,  -6,   4,   1,   7,  -4 },
      {   7, -25, -24,  -2,  34, -28,  39,   3,  89, -38, -40,  -1, -22, -11,  13,   6 },
      {  23,  28, -15,  10,  53,  40,  14, -38, -28,  36, -66, -10,   4, -32, -13,  31 },
      {  -5, -13, -20,   7,  -7, -27, -77,  45,  33,  62, -27,  -6,  19, -30,   2,   3 },
      { -14, -27, -36,-107,  17,  17,   6,  19, -15,   8,  -3,  31,   8,   8, -17,   1 },
      { -18, -16, -18,  12, -32, -48,  -5, -23, -47, -29, -54,  56,  19,  -7,  44,   9 },
      {   9,  -5,  10,   7,   9,  -8,  10,  -8,  15, -29,  19,  11, 105, -32, -46,   7 },
      {  -4,  -8,   5, -33,  -2, -13,   2, -43,   1,  10,  50, -25,  -8, -80,  56,  19 },
      {  -2,  -6, -15, -14, -16, -24, -46, -87,  24,   0,   6,  -5, -17,  35, -49,  28 },
      {   9,  13,   0,   6,  16,  10,  15, -39,  35,  59,  19,  57,  26,  30,  41, -50 },
      {  -1,   1,   6,  25,  -1,   0,   9,  14,   4,  12,  27,  83, -50, -48, -48,  22 },
      {   3,   4,   7,   6,   5,   9,   7,  22,  13,  13,  22,  13,  19,  44,  36, 105 },
    },
    {
      {-122,  28,  16,   5,  16,   1,   2,  -1,  12,  -3,  -3,   0,   5,  -2,  -2,   0 },
      { -16, -30,  31,   3,-105,  -9,  48,   8, -10,  11,  -1,  -2,  16,  10, -13,  -4 },
      {  18, 107,   0, -18, -32,  26,  15,  -3, -33, -31,  -4,   8, -10, -11,   0,   1 },
      {  -6,   0, -20,  -4, -24,  90, -32, -21,  12,  69,  12, -22,   2, -10,   7,   0 },
      {  21,   0,  86,  -5,  44,  39,  55, -17,  24,   8, -16,  -9,   4, -12, -23,   5 },
      { -18, -42,  -1,  -2,  20,  34,  25,  -6, -90, -20,  39,  12, -38, -15,  12,   3 },
      {  -4,  -4, -67, -32,   7,  12,  72,  -4,  50, -13,  42,  18,  12, -14, -13,   7 },
      {  -7, -33, -19,  10, -16,  45, -14,   9,  17, -64, -76,  15,  -6, -39, -11,  11 },
      {  -2, -15,  32, -64, -15,  -5, -44, -60,   9, -36,  31,  20,  35,  -8,  17,  28 },
      {  -2,   1,  15, -20, -18, -32,  -1,  -1,  33,  22,   3,  -1, -92, -58,  25,  19 },
      {   8,   6,  29,  64,  -9,  17, -20,  45,  27, -27,  68,  18,  20, -37,  16,   5 },
      {   2,   3, -12,  57,  -9,  -5,  17, -74,  13, -37,   9, -67, -14,  14,  25,   3 },
      {   0,   5, -15,  14,   5, -33,  -2, -20, -33,  25,  -3, -27,  42, -83, -52,  24 },
      {   1,   2,  -4,  -1,  10,  -5,  32,  14, -13,  20, -32,   0,  45, -21, 101,  27 },
      {   2,   8,  -6,  51,  -2,  -4,   7, -46,  -1,  29, -12,  86,  -9,  20, -10,  51 },
      {   1,  -1,   1,  10,  -1,  -8,   4, -40,   2,   8,  -7,  42,   8, -35,  18,-106 },
    },
  },
  { // 2
    {
      {-119,  25,  18,   5,  26,  12,   1,  -3,  20,  -2,  -6,  -2,   8,  -4,  -2,   1 },
      {  26,  56, -24,  -5,  91, -12, -42,  -3,  -2, -38,  -4,   8,  -9, -10,   9,   2 },
      {  -5, -97,  -4,  10,  62, -37,  17,   9,  33,  -8,   7,  -2,   0,   1,   1,  -1 },
      { -28,   6, -28,   6, -30, -79, -16,  26, -39, -24,  50,  25,   0,  22,  30,  -5 },
      { -21, -31, -91,  -4,   7,  64, -19,   9, -36,  20,   9,   6, -14,  -1,   6,  -2 },
      {  -3, -11,  49,   6,  33,  20,  41, -19, -90, -14,  13, -10, -34,  21,   9,  -3 },
      {   8,  38, -18,  -6,  30, -15,  47,  20,  10,  71,  52, -26,   6,  18, -30, -32 },
      {  11,  10,   4,  21,  -6,  49,  45,  34,  35, -53,  56,  12,   7, -27,  43, -12 },
      {   7, -10,  46,  43,   8,  26, -71,  13,  16,  45,  27,  17, -16,  37,  32, -21 },
      {   2, -13,  14, -75,   8,  20, -11, -39,   2,  -9,  38,  41,  61,  34,  -7,  -8 },
      {   7,   3,  -6,  63,   9,   9,   6,  27, -28, -14, -29,  -1,  89,  34, -22,  -3 },
      {  -4, -10,  -7,   7, -15,   1, -32, -38,   2, -39,  22, -82,   7,  -3, -15, -67 },
      {  -1, -12,  25,   7,   4,  -5, -26,  15, -34,  15,  27,  29,  19, -94, -46, -17 },
      {  -3,  -4,  14, -34,  -6,  14, -10,  76,   5, -37, -15,  11, -36,  35, -61, -29 },
      {  -1,  -6,  16, -53,   8,  -4,  -7,  55, -20,  21, -34, -41,  32, -22,  67, -21 },
      {   1,  -5,  12, -10,   1,   9, -27,  25,  -3,  -6,  46, -59,  10,   4, -20,  93 },
    },
    {
      {  82, -63,   8,  -2, -62,  34,  11,  -4,   8,  11, -15,   1,  -2,  -4,   1,   3 },
      { -56, -15,  37,  -2, -13,  80, -40,  -5,  32, -39, -20,  17,  -1,  -3,  16,   0 },
      {  51,  16,   3, -13,   7, -14, -59,  25,  13, -60,  59,  11, -14,  20,  13, -26 },
      {   3, -69,  58,  -5,  69,  -6, -11, -16, -48,   7,   7,   1,  -5,  25, -11,  -2 },
      {  41,  26, -23,  12,  37,   8, -25, -29, -35, -17, -61,  46,  14, -30,  42,  18 },
      {  -4, -41, -45,  42,  33,  -1,  57, -21,  36, -57,   6, -11, -22,  12,  28,  -7 },
      {  24,  21, -13, -21,  28,  33,  -4,  16,   2, -24, -23, -55, -31,  16, -49,  69 },
      {  18,  23,  67, -31,  11, -43,  22, -27,  65,  -3, -37,   4, -33, -13,   7,  -5 },
      { -32, -30, -12, -17, -44, -56, -18,  15, -18, -20, -25,  30, -33,  42,  22,  53 },
      {   7,  47,  22,  28, -31,  19,  25, -64, -34, -17,   8,  16, -13,  66, -21,  -9 },
      {  -2, -10,  26,  40, -21, -37, -18, -35,   8, -40,  18, -21,  64, -36, -27,  47 },
      {  -2,  13,  28, -53, -13,   7,  56,  26, -46, -48,   1, -29,  36,  -8,  44,  -3 },
      { -16, -19, -35, -50, -17, -13, -21, -55, -26, -28, -14, -23, -28, -38, -38, -48 },
      {   3, -11, -30, -61,  19,   6,  16, -16,  36,  -3,   5,  52,  64,  36, -34,   9 },
      {   2,  -1, -15, -16,   1,  -4, -39, -41,  19,  31, -10, -68,  30,  47,  58,  -3 },
      {   4,  -3,   4,  31,  -2, -21, -10,  46,   2, -25, -73, -16,  33,  36, -35, -55 },
    },
    {
      {-100,  53,   1,   3,  53,   1, -19,   0,   0, -19,  -3,   5,   3,   0,   5,   3 },
      {  30,  32, -19,   0,  39, -88,   8,  10, -26,   8,  54, -10,  -2,  16,  -7, -22 },
      { -59, -38,  20,   5, -21, -20,  65,  -2,   3,  59,  12, -35,   5,   9, -29, -11 },
      {  -2, -84,  33,   7,  65,  -1,   9,  -6,   6, -46,  19,   3,  -7,   8,  31, -16 },
      {  29,   6, -20,  -1,  70,   8,  37, -23, -15,  18, -62, -33, -10, -21, -28,  38 },
      {  26,  43,  59, -21,  20,  34,   4, -17,  55,   3,  24, -31, -11, -11, -27, -50 },
      {   5,  11,  61, -10,  -3,  17, -16, -42, -84,  13,  -3,   9,  -2,  53,  -6,   4 },
      {   2,  -9, -51,  11,  25,  74,  -5,  50, -28,  19,  37, -18,  -9,  34, -20, -24 },
      {  -4, -37,  10, -12,  23, -22, -80,  10,  22,  54,  -9,  27, -15,  -4, -53,   8 },
      {   8,   6, -11,  45,  21,  12,   9, -31,  -1,  54, -13,  55,  40, -21,  34, -56 },
      {  15,  21,  45,   5,  20,   9,  31,  60,  23,  33,  16,  40,   7,  19,  27,  63 },
      {   4,   5,  42,  75,  -9, -11,  -4,  56, -31, -28, -26,  -9,   1, -35, -38, -21 },
      {  10,   1,  -7,  58,   1,  -6, -21, -24,  43, -10,  -5, -35,  58,  68, -13,  23 },
      {   2,  -9,   5, -59,   8,   3,   6,  25, -16, -19,  -4,   8, 100, -16, -30,  -8 },
      {  -1,  -4,  -2,  30,  -1,  20,  -2, -44, -13,  -5,  76,   8,  13, -58, -29,  52 },
      {   1,  -4,  16,  -4,   2,  -1, -45,  15, -17,  38,   5, -76,  21, -35,  68,  10 },
    },
  },
  { // 3
    {
      {-109,  44,  12,   4,  45,   4, -12,  -3,  11, -13,  -4,   2,   4,  -3,   2,   1 },
      { -44, -51,  27,   5, -65,  52,  28,  -5,  37,  30, -27,  -9,   7, -11,  -9,   5 },
      {  15,  94, -17, -10, -71,  18, -22,   2,  19,  -7,  -9,   8,  15, -17,  -3,   3 },
      { -33,   3, -17,   7, -44, -74,  31,  23,  -7,  40,  56,  -9,  15,  21, -13, -19 },
      {  16,  38,   0,  -4,  33,  -6,  48, -25, -11,  61, -42, -42,  -6, -19, -51,  11 },
      {  12,  17, 115,  -8, -17, -31,  -8, -21, -14, -11,  -4,  -3,  -5,  23,   0,  -5 },
      { -20,   3, -11,  -2, -25,  35,  17, -12,-112,  -8,  -5,   2,  15,  22,   7,  13 },
      {  14,  19,  18, -12,  22,  70,  26,  11,   9,   1,  72, -13,   2,  20, -22, -53 },
      {   3,  12,  -4,  39,  -9, -11,  78,   1,  13, -78,  -5, -39, -13,  -9,  21,   6 },
      {  11,  10,  13, 111,   5,  16, -33,  21,  -9,  32,   2, -20,   9,   1,  15,  -4 },
      { -13,   2,  -9,  -5, -22,   4, -24,   6, -13,   6,   8, -27,-118,  -3,  -2,  -8 },
      {   0,   1, -25,  11,  -5,   0,  -9,-108,  21,  11,  22, -18,   0,  40,  31,   2 },
      {  -7, -24,  -9,   6, -13, -10, -48, -20,  -9, -46,   7, -53,  26, -22, -78, -17 },
      {   2,   3, -11, -20,   3,   7, -14,  43,  20,  -4, -30, -53,  10,  94,  -1,  27 },
      {   0,  -3,  11, -38,   3,   2, -17,  12, -10,  18,  24, -75,  24, -51,  66,  13 },
      {  -3,  -2, -13,  -7,  -3, -11,  -1,  -1,  -7,   2, -58,  -9,   8,   7,  30,-107 },
    },
    {
      {  87, -58,   5,  -3, -64,  26,  12,  -3,   7,  14, -11,   0,  -3,  -4,  -1,   2 },
      {  53,  23, -34,   1,  20, -83,  31,   5, -37,  35,  24, -13,   1,   6, -15,  -1 },
      {  52,   4,   8,  -9,  28,  -6, -60,  14,   3, -70,  48,  14, -17,  23,  18, -19 },
      {   4,  77, -45,  -8, -68,  -3,  -5,  24,  52, -12,   5,   0,  -2, -15,   2,  -2 },
      {  38,  29, -36,   9,  38,  19, -11, -25, -20, -15, -73,  28,   6, -34,  35,  31 },
      {  -5,  30,  34, -16, -42,   4, -55,   3, -70,  37,   6,  38,  37,   0,   3,  -4 },
      {  29,  35,  44, -41,  31,  19, -26,   1,  22,  25, -18, -43, -18, -19, -67,  16 },
      {   3,  -6,  55, -10,   3, -62,  -3, -21,  56,  15, -18,  39,   3, -42,  32, -32 },
      { -12, -49, -39, -15,  17, -13, -46,  71,  13,  21,   5,  -8,  27, -51,   3,  21 },
      {  20,  15, -15,  41,  33,  51,   8, -10,  17,  32,  36,   6,  48, -26,  -9, -66 },
      {   7,  11,  35, -25,  -6,  -2,  42,   7, -16, -52,  11, -49,  73, -30,  25,  15 },
      {   5,  15,   8, -56,  20,  31,  38,  57, -11,  29,  -9,   4, -33,  19,  56, -34 },
      {  17,  18,  34,  46,  16,   5,   7,  36,  38,  28,  10,  20,  28,  49,  19,  68 },
      {   0, -13, -29, -65,  12,  -3,  11, -12,  26, -10, -18,  47,  58,  51, -39,  -7 },
      {   4,   2,  17,  46,  -6, -20,  -4,  58, -11, -26, -70, -12,  14,  24, -31, -51 },
      {   1,   4,  17,   6,   3,  10,  47,  36, -19, -30,  21,  72, -24, -48, -49,  18 },
    },
    {
      {  96, -54,  -1,  -3, -58,   8,  18,   0,   0,  20,  -4,  -4,  -3,  -1,  -5,   2 },
      {  45,  30, -26,  -4,  40, -85,   8,  11, -39,  12,  41,  -8,  -3,  17, -12,  -8 },
      {  46, -20,   3,   0,  51,  20, -47,  -2,   2, -77,   5,  21, -16,   3,  44,  -6 },
      {  23,  83, -33, -14, -58,   2, -37,  22,  34, -17,   4,  20,   7, -20,   7,  -3 },
      {  31,  39,  -7,  -4,  32,  31,  -8, -25, -55,  23, -70,   1,  12, -21, -13,  43 },
      {  -7,   5,  52, -11, -30,  22, -55,  -1, -65,  14,  36,  26,  20,  41, -17, -19 },
      { -28, -28, -63,  30, -32, -14,   7,  10, -51,  -9,  -6,  35,  30,   6,  58,  20 },
      {   0, -11, -62,   9,  23,  66, -11,  53, -15,  11,  36, -37,   0,  -7, -36, -16 },
      {  -2,  30,  -5, -22, -26,  22,  61,  -7, -28, -61,  -3, -31, -30,  60,  -5,   9 },
      {  16,  22,  20,  20,   9,  21,  48, -27,  -7, -20,  27, -10,  80, -37,  15, -45 },
      {  -8, -28,  12, -55,  -5, -28, -14,  42,   1, -41, -12, -30,  65, -11, -26,  44 },
      {  -8,   6,  29,  23, -27, -11, -21,   2, -36,  -6,  20, -74, -38, -56,  38,  18 },
      {   3,  10,  31, -51,  16,  14,  37,  73, -13,  31, -10,  22, -15, -12,  56, -19 },
      { -18, -21, -30, -67, -11,  -8,   3, -40, -31, -17,   1,  23, -29, -57, -25, -44 },
      {   2,   4,  18,  13,   4,  15,  33,   6,  -2, -13,  58,  56, -19, -37, -28,  73 },
      {   1,   2, -23, -58,  10,  21,  -9, -52,  22,  38,  47, -23,  16,  21,  48,  39 },
    },
  },
  { // 4
    {
      { 119, -27, -15,  -5, -22, -17,   2,   2, -17,   1,   7,   2,  -7,   2,   2,   0 },
      { -27, -60,  16,   6, -89,  -2,  35,   6,  12,  43,   7,  -6,  16,  11, -10,  -4 },
      {  25,  49,  10, -11, -34,  84,  -3, -17,  26,  21, -46, -10,   4, -25, -24,   7 },
      {  -6,  79,   9,  -5, -67, -31, -17,   2, -54, -16,  20,  13,   7,  19,  10,   0 },
      { -18, -44, -38,  -2, -37,  23, -50,   1, -11, -58, -36,  30,  -9, -25,  37,  25 },
      {  -2, -24,  53,   7,  14,  48,  42, -18, -80, -26,  -5, -13, -29,  15,   9,   3 },
      {  15,   2,  87,  27, -10, -40,   0,  -9,  49, -46, -31,  -1,   8, -15,  16,   6 },
      { -10,  -9,  -7,  -9,   2, -54, -16, -25, -38,  20, -66, -31, -18, -23, -61,  21 },
      {   6, -19,  28,  -7,  16,  10, -63, -41, -10,  34, -14,  19,  52,  67,   9,   5 },
      {   6,   8, -18,  71,  11,   5,  24,  57, -25,  -4, -40,  14,  65,   8, -13,   9 },
      {   1,   1,  25,  65,  -1,   7, -51,  32,  -6,  50,  10,  20, -68, -10,   2,  -3 },
      {   0,  19, -24,  18,   3, -16,  34, -30,   4,  42, -28, -20, -13,   6,  86,  48 },
      {   0,  -2,  26, -32,  15,  -8,  21,  -1, -17,  31,  12,  87,  20, -56,  -2,  39 },
      {   0,  -7,  -4,  45,  -1,   4, -21, -55, -21,   5,  43, -33,  45, -69,   9, -22 },
      {  -1,  -3,  18, -44,   6,  -6, -13,  46, -22,  27, -38, -18,  15, -30,  50, -72 },
      {  -2,   6, -24,  27,  -4, -10,  29, -56,   5,  -2, -31,  61, -21,  11,  -4, -75 },
    },
    {
      {  82, -59,   8,  -3, -67,  35,   7,  -2,  12,   8, -13,   2,  -3,  -5,   2,   1 },
      { -54, -13,  33,  -2, -10,  76, -41,  -1,  38, -49, -13,  15,  -4,  -2,  19,  -3 },
      {   3, -70,  56,  -5,  64,   3, -13, -15, -56,  13,  -4,   8,  -1,  20,  -3,   0 },
      {  49,  10,   4, -12,  14, -22, -48,  25,   8, -58,  67,   4, -19,  32,   5, -24 },
      {  43,  17, -42,  20,  42,  10,  -1, -26, -20, -43, -47,  33,   6, -24,  57,  11 },
      {  13,  46,  20, -25, -24,  13, -68,  18, -48,  39,  -4,  30,  43, -29,  -2,  -2 },
      {  -2,   2, -54,  29,   2,  66,   0,   4, -53, -14,  28, -38,  18,  27, -40,  20 },
      { -39, -34, -17,  32, -47, -24,  16,   0, -34,  -7,  36,  41,  35,  17,  52, -32 },
      {   3,  46,  26,  -2, -33,   2,  14, -49, -27, -10, -26,  35, -38,  74, -17,   2 },
      {  21,  21,  24,  48,  21,  20,  10, -59,  31,  14,  29,   7,  39, -18, -27, -59 },
      {   4,  10,  44, -32, -17, -19,  36, -17, -16, -65,   2, -33,  64, -16,  -5,  31 },
      {  10,  29,  17, -39,  18,  46,  59,  45, -15,  13,   1, -16,  -5,  13,  42, -56 },
      {   1,  20,  39,  55, -22,  -5, -18, -11, -29,   5,   9, -67, -46, -27,  47,   6 },
      { -16, -13, -20, -37, -16,  -3,   5, -27, -41, -31,   3,   7, -49, -64, -37, -50 },
      {   6,  11,  35,  56,   2,   4,  36,  64, -11, -21,   4,  53, -18, -30, -37,  22 },
      {   0,  -8,  -1,  30,  -8, -27, -24,  36,  -5, -25, -74, -33,  23,  24, -26, -61 },
    },
    {
      {-100,  51,   1,   3,  54,   1, -18,   0,   4, -20,  -2,   4,   3,   0,   6,   2 },
      {  45,  35, -23,  -1,  49, -76, -10,   9, -37, -13,  46,   1,  -3,  17,   5, -14 },
      { -44, -19,   3,   5, -11, -48,  60,   6,   1,  69,  22, -32,   9,   5, -39, -12 },
      {  -1,  87, -29,  -6, -78,   4, -23,   7, -11,  28,   6,  10,  11,  -3, -16,  -1 },
      {  34,  33, -27,  -4,  53,  15,  17, -21,  -5,  35, -64, -25, -10, -32, -37,  33 },
      {  14,  12, -23,  -1,   6, -17,  -2,  22, 106, -13,  18,  -9,  -7, -47,  -5, -24 },
      { -19, -16, -56,  14, -27, -47,  31,  12,  -7, -24, -34,   8,  13, -23,  55,  54 },
      { -10, -32, -83,  14,  21,  52, -17,  50, -19,  28,  27,   3,  -6,   8,  -1, -16 },
      {  -5, -31,  10,  -4,  10, -39, -70,   1,  14,  53, -24,  63,  28, -11, -11,  12 },
      {  19,  18,   5,  18,  22,  25,  31,  -4,  12,  13,   4,   8, 107,  10,  32, -16 },
      {  -9,  -8, -10, -18,   0,   6,  10, -58, -34,  12,  22,  21,  -8, -83,  28, -48 },
      {   7,  23,   6,  36,   8,   4,  16, -17,  26,  54, -10,  23, -56,  40,  68, -14 },
      {  -9, -18, -50, -15, -10, -12, -33, -81,  22,  -7,  -9, -42,  16,  48,  -1, -18 },
      {   2,  -3,   5, 117,  -9,  -6, -22, -19, -10, -15,   2, -13,   2, -22, -27,  -6 },
      {   2,   1,  -6,   2,   7,  21,  12, -43,  19,   6,  74,  30,  -3,   1, -15,  83 },
      {   2,  -1,  28,  -5,   1,   2, -49,  16,  -7,  34,  24, -81,   7, -27,  51,  34 },
    },
  },
  { // 5
    {
      {-111,  41,  10,   4,  45,   3,  -8,  -3,   8, -11,  -3,   0,   2,  -2,   2,   1 },
      { -51, -66,  25,   8, -65,  26,  32,  -1,  31,  38, -10,  -8,   9,  -2, -10,   0 },
      {   3, -90,  -1,   7,  88,   7,   8,  -3,  -4,  -8,  -2,  -2, -16,   4,   6,   0 },
      { -24, -22, -33,   7, -13, -94,   7,  30, -11,  13,  54,   5,   9,  26,   6, -14 },
      {  -7,   1,-109,  -2,   5,  47, -10,  19,  22,  29,   9,   0,   6, -19,  -8,  -1 },
      { -22, -16, -18,   1, -30,  42,  14,  -3, -93, -49,  20,  15,   3,  15,  20,  -2 },
      {   8,  31,   3,   2,  26,   9,  86,  -2, -32,  38,  16, -48,  -4,   5, -43, -15 },
      {  -1,  -1,  22, -20,  12,  21, -61, -25, -38,  82,  27,  13,   5,  38,  -9,  -6 },
      {  11,  10,  25,  92,   2,  34, -20,  43,  15,  -4,  45, -10,  -3,   5,   8, -34 },
      {   7,   3,  21, -66,   9,  30,  23,   6,  45, -27,  71,  34,  17,  10,  -1, -26 },
      {   8,  -4,  -5,  21,  13,  -1,   3, -11,   2, -13, -27,  10, 113,  31, -26,  -6 },
      {  -5, -13,  17,  -7,   0,  -9, -25,  37, -28,  -8,   4,  28,   4, -69, -85,  -9 },
      {  -2,   1,   2, -32,  -1,  10,  -7,  80,  -1,  -6, -53,  -5, -22,  65, -15, -30 },
      {  -5, -15, -11, -17, -15,  -2, -44, -38,  10, -41,  17, -88,  -5,  12, -35, -32 },
      {  -2,   4, -19,  35,  -7,  -3,  11, -49,  20, -17,  -2,  56, -47,  50, -61,  -1 },
      {   1,  -1,   7,   0,   1,  10,  -7,  30,   7, -13,  31, -24,   3,  28, -27, 109 },
    },
    {
      {  81, -59,   8,  -3, -67,  36,   8,  -2,  13,   7, -13,   2,  -3,  -4,   2,   1 },
      { -57, -14,  31,  -1, -10,  76, -37,  -2,  40, -48, -14,  13,  -6,   0,  18,  -3 },
      {  55, -10,  11,  -8,  43, -14, -50,  13,  -8, -63,  52,   5, -18,  32,   9, -15 },
      {  13,  74, -47,  -4, -57, -10,  -7,  26,  56, -25,  19,  -7, -12,  -9,   4,  -7 },
      {  38,  28, -39,  15,  36,  16,  -9, -28, -17, -23, -63,  34,  10, -36,  49,  15 },
      {   6,  38,  25, -22, -32,  17, -72,  14, -55,  38,  10,  27,  40, -13, -11,  -3 },
      {  29,  29,  14, -28,  40,  35, -16,   6,  18,  15, -30, -47, -35,  -9, -64,  41 },
      {   8,  -5,  61, -33,   9, -61, -17,   5,  58,  20, -26,  30, -12, -37,  22, -10 },
      {   7,  47,  47,   1, -30,  -7,  23, -62, -19, -13, -18,  23, -38,  61,  -3,   4 },
      {  24,  24,  18,  39,  29,  31,  16, -45,  36,  21,  35,   8,  35, -22, -26, -62 },
      {  12,  25,  47, -39,  -1,   9,  49,  15, -17, -37,   3, -54,  55, -15,  36,   3 },
      {  -3, -13,   8,  36, -18, -39, -24, -39,  18, -39,   4,  -6,  56, -11, -40,  64 },
      {  -1,   9,  34,  72, -16,  -6, -18,   1, -27,   5,  15, -49, -52, -48,  31,   4 },
      { -16, -19, -27, -53, -18, -17, -19, -63, -24, -27,  -5, -33, -19, -42, -20, -48 },
      {   3,  -3, -13,   8,   3, -10, -47, -22,  27,  36, -30, -64,  28,  59,  42, -17 },
      {   2,   1,  13,  37,  -8, -19,  -1,  48, -16, -36, -68,  -5,  12,  11, -46, -59 },
    },
    {
      {  97, -52,  -1,  -3, -60,   7,  17,   0,   2,  20,  -3,  -4,  -3,  -2,  -5,   0 },
      {  49,  32, -28,   0,  44, -82,   6,   8, -39,   5,  39,  -5,   0,  14,  -6,  -9 },
      { -51,  -7,  -2,   5, -35, -33,  57,   4,  -4,  78,   6, -25,  12,  -1, -35,  -1 },
      { -13, -93,  32,   6,  62,  -2,  31, -10, -25, -13,   7, -14, -10,  22,   7,  -2 },
      {  26,  24, -14,  -4,  46,  24,  20, -35, -29,  28, -78, -16,   1, -24, -17,  42 },
      {  -8,  11,  35,  -8, -31,  19, -31, -15, -94,  17,   2,  28,  21,  45,   5,  -1 },
      { -23, -34, -93,  28, -17,  -2,   3,  28, -31, -13, -16,  23,  14,   0,  37,  27 },
      {   7,   1, -42,   1,  28,  73, -15,  36, -12,  22,  37, -24, -12,  15, -55, -30 },
      {   0,  34,  -8, -14, -27,  22,  68, -11, -18, -60,   9, -58,  -6,  34,  17,   2 },
      {  16,  19,   8,  33,  22,  30,  43,  -3,   2,  25,   9,  25,  50, -23,  56, -61 },
      {   2,   7,  43,  21,  -6,  -5,  17,  86, -39, -30, -18,  -1,  -6, -56, -20,   3 },
      {  -3, -21,  -9, -53,   0,  -5, -22,  -9,  -7, -22,  13, -37,  93, -41, -15,   0 },
      {   0,   3,  -1, -97,  15,  13,  32,  40,   7,  22,   9,  43, -14,   5,  28,  22 },
      {  15,  11,  22,  37,  15,  11,  11,  29,  34,   2,  19,  11,  57,  50, -13,  73 },
      {   3,  -3,  -2, -15,   4, -21, -13,  43,  20,  -1, -76, -17,  23,  62,  -2, -53 },
      {  -3,  -3, -14,  -4,  -4,  -4,  41, -27,   2, -44, -10,  75,  14,   5, -73, -24 },
    },
  },
  { // 6
    {
      {-117,  31,  14,   4,  35,   4,  -4,  -2,  14,  -7,  -5,  -1,   6,  -1,  -1,   0 },
      { -40, -63,  19,   8, -82,   4,  37,   6,  14,  41,   4,  -7,  14,  10,  -9,  -3 },
      {   0, -92, -22,  11,  76,  -2,   7,   7,  36,   3,   1,  -2,  -7,  -8,   1,   1 },
      {  15,  12,  20,  -6,  10, 107,  14, -22,  24,  24, -34, -15,  -4, -25, -15,   5 },
      { -25, -17, -84,  -9, -18,  51, -36,   8, -47,   7,  18,  18, -14,  12,  27,   4 },
      {   6,  25, -46, -11, -19, -21, -50,   2,  86,  43,  -9,   7,  25, -11,  -8,   4 },
      {  -7, -34,  17,  -2, -34,   0, -46, -17,  11, -73, -59,  23,  -1, -27,  18,  30 },
      {  -3,  17, -34, -18, -16,   8,  65,  40,  31, -49,  31,  24, -12, -58,  -6,   5 },
      {  -8, -10, -15, -96,   5, -22,  24, -71, -10,   5,  -1,  -8,   2,  -2, -11,  16 },
      {  -3,  -2,  40,  -5,  -7,  -5, -28, -18,  16,  29,  51,  15, -84, -33,  37,  12 },
      {  -9,   4, -31,  19,  -6, -32,   4,  11, -31,  26, -58, -40, -57, -45, -42,   5 },
      {   0,  -8,  21,  -2,  15,  -3, -17,  11, -38,  35,   8,  78,  35, -43, -50,  28 },
      {   2,   7, -19,  55,   1, -11,  20, -53, -14,  12,  17, -26,  47, -50,  48,  38 },
      {   3,  14,   1, -13,  12, -10,  41,  34,   3,  34, -50,  33, -12,  34,  58,  60 },
      {  -1,  -9,  24, -52,   6,  -1, -22,  57, -20,  16,  -7, -42,  36, -49,  47, -28 },
      {   0,  -3,   9, -11,  -1,  10, -22,  32,   2, -14,  37, -56,   4,  17, -35,  92 },
    },
    {
      {  80, -60,   9,  -3, -66,  38,   6,  -3,  13,   7, -14,   2,  -3,  -6,   3,   2 },
      { -55, -13,  34,  -3,  -5,  75, -43,  -1,  36, -50, -10,  15,  -6,   1,  18,  -5 },
      {  23, -60,  51,  -9,  65,  -5, -28,  -5, -57,  -6,  20,   6,  -4,  28,  -4,  -6 },
      {  46,  34, -15, -13, -11, -22, -43,  32,  24, -56,  65,  -3, -17,  23,   2, -21 },
      {   2, -39, -35,  30,  37,  -7,  62, -23,  36, -53,  -6, -19, -39,  26,  13,   7 },
      { -45, -31,  40, -19, -32, -18,  23,  24,  32,  28,  43, -41, -18,  27, -51,  -1 },
      {  21,   6,  52, -40,  27, -51,  -6,   0,  64,   7, -40,  18, -26, -34,  18,   5 },
      { -30, -22,  15,  18, -48, -51,  13, -27, -11,  -7,  18,  59,   9,  18,  50, -41 },
      {  18,  54,  32,   7, -14,  12,   6, -62,  -6,   0, -18,  26, -36,  59, -45,  13 },
      {  18,   2,  16,  58,  30,  12,  -2, -40,  48,  21,  44,   9,  44, -34, -30, -33 },
      { -12, -33, -52,  30,   6, -11, -54,  18,  17,  39, -12,  49, -52,   5, -33,   2 },
      {   9,  22,   0, -15,  23,  44,  31,  25,   1,  48,   4,   2, -40,  17,  33, -79 },
      {   1, -12, -26, -55,  15,   6,  29,  19,  17, -20, -18,  55,  58,  27, -49, -23 },
      {  -5,   4,  12,  -7,  -7,  10,  38,   8, -38, -35,  26,  37, -48, -77, -39,  -5 },
      {   9,  13,  40,  73,  -4, -11,  14,  76,  -6, -21, -43,   5,  12,  15, -16, -11 },
      {  -6, -10, -12,  -6, -15, -24, -30, -32, -15, -27, -48, -51,  -5, -16, -34, -77 },
    },
    {
      { -99,  51,   1,   3,  57,  -2, -17,  -1,  -1, -21,  -1,   3,   3,   2,   6,   1 },
      {  57,  30, -28,   0,  58, -64, -14,   9, -38, -21,  36,   2,  -6,  17,   8, -10 },
      { -47, -28,   2,   7, -28, -73,  47,  13,  -7,  43,  41, -15,  11,  19, -18, -14 },
      {  -5, -93,  26,   7,  71,   6,  24, -11,   2, -25,  -5, -11, -15,   8,  17,   0 },
      {  22,  36,  10,  -6,  49,  14,  47, -23,   2,  59, -27, -45,  -4, -15, -51,   6 },
      {   9,  -3, -25,  10,  20, -23,  -7,  32, 104,  -4,  15,   1,   5, -48,  -9, -12 },
      { -16, -31, -99,  27,  10,  11, -13,  22, -29,  34, -31,  -1,  -5, -12,   1,  25 },
      {   3,  -1, -11,  10,  17,  69, -13,  22,   6,  36,  64,  -3,   7,  44,   0, -55 },
      {   1, -25,  34,  -1,  17, -26, -84, -20,   4,  57, -14,  38,  11,   8, -32,   4 },
      { -16, -17, -24, -71, -14,  -1, -30, -12,  -6, -22,  13, -36, -64, -12, -46, -27 },
      {  -2,   9,  37,  46,  -1,   7,   3,  36, -27,  17,  21,  27, -81, -52,   0,   2 },
      {  -1,  10, -14, -47,   6,  -4,  15, -35,  31,  47,  18,  25, -45,  17,  70,  35 },
      {  -2, -12,  -4, -50,  15,  23,  23,  11, -23,  -6,  49,  54,  41, -43, -34,  42 },
      {   4,   6,   5,  -4,   2,   3,  11,  56,  25, -16, -16,  20, -31,  77, -44,  51 },
      {   0,  -2,  -2,  46, -10,   8, -27, -46,  10, -14,  59, -48,  -3,   2, -15,  70 },
      {  -2,   2, -32,  37,  -4,   0,  30, -71,  14, -21,  -1,  65, -21,  15, -40, -31 },
    },
  },
  { // 7
    {
      { 115, -32, -13,  -4, -40,  -6,   6,   2, -14,   6,   5,   1,  -4,   1,   1,   0 },
      { -38, -61,  18,   6, -77,  37,  26,   1,  33,  38, -11,  -7,  11,  -7,  -9,   1 },
      {  -1,  97,   0,  -6, -81,  -3, -15,   0,  -5,   5,   0,   1,  15,  -1,  -1,  -1 },
      {  30,  25,  10,  -7,  25,  85, -11, -20,  40, -16, -52,  -7, -12, -39,  -7,  10 },
      {  17,  10,  47,   1,   8, -73,   0,   7,  86,  15, -20,  -5, -12, -12,  -4,   7 },
      {   4,  -8, 104,  19, -10,  -1,  -6, -25, -52, -32, -21,  -3,  -4,  15,   5,   2 },
      { -14, -31, -19,   2, -33, -10, -62,  10,  23, -76,  -5,  31,  -7, -15,  45,   9 },
      {   5,  -8,  17, -10,  13,  17, -93, -20,   4,  69,  20,  22,   8,  27,   4,   1 },
      { -10,  -9, -34,  38,  -5, -35, -30,  -2, -42,  28, -80, -18, -11, -32, -22,  23 },
      {  -7,   9,   1,  20, -11,   6,   3, -16,  -7,  20,  40, -23,-104, -27,  35,   7 },
      { -12,  -8,   1,-108,  -6, -18,  15, -29, -20,   7, -24,  29, -24, -12,   2,  24 },
      {   2,   0, -14,  22,   4, -14,  19, -92,   8,   6,  14, -10,  49, -19,  50,  33 },
      {  -1,  -6,  18,   2,   2,  -7, -12,   0, -14,  -3,  51,  26,  20, -96, -52,  -3 },
      {   5,  11,  18,  13,  13,  10,  25,  59, -16,  37, -14,  54,  18, -28,  70,  25 },
      {  -1,  -7,  12, -42,   4,  -1, -24,  35, -11,   3,   7, -96,  31, -30,  38,  -7 },
      {   1,   1,   4,   2,   0,   9,  -4,  25,   3, -12,  26, -17,   1,  22, -29, 115 },
    },
    {
      {  80, -60,   7,  -2, -66,  37,   9,  -3,  12,   8, -15,   2,  -2,  -5,   3,   2 },
      { -58, -15,  33,  -1,  -9,  75, -38,  -3,  40, -47, -12,  14,  -6,  -1,  18,  -5 },
      {  53,  -9,  15, -10,  34, -17, -53,  16, -11, -58,  55,   9, -19,  36,   8, -18 },
      {  15,  75, -48,  -3, -58,  -7,  -7,  23,  55, -26,  18,  -5,  -4, -13,   5,  -7 },
      {  39,  27, -30,  17,  39,  14, -15, -28, -20, -23, -60,  38,  13, -37,  52,   8 },
      {   0,  35,  38, -21, -39,   7, -62,  11, -50,  41,  11,  30,  46, -19,  -8,  -8 },
      {  17,  10,  62, -52,  18, -46,  -8,  11,  57,  18, -45,   8, -29, -18,   4,   4 },
      { -32, -34,   1,   8, -40, -59,  18,  -3,  -1, -11,   9,  53,   9,   7,  66, -36 },
      {   3,  46,  25, -13, -32,   7,  26, -40, -34, -17, -31,  12, -34,  79,  -2,   8 },
      {  22,  24,  42,  42,  18,   8,  27, -64,  41,   9,  35,  16,  34,  -7, -18, -43 },
      {   3,   6,  31, -45,  -5,  -4,  47,   4, -18, -60,   6, -41,  66, -21,  17,  21 },
      {  10,  14, -10, -37,  29,  47,  34,  47,  -1,  34,  -3,  -3,   8,  27,  34, -72 },
      {  -2, -20, -43, -48,  13,  -1,   3,  -6,  26, -11, -11,  66,  48,  31, -56,  10 },
      {   8,   2,   6,  33,   9,  -4, -20,  17,  38,  31,  -5, -18,  53,  64,  44,  52 },
      {  -8, -16, -33, -37,  -8, -17, -48, -67,   7,   9, -18, -62,  16,  13,  11, -42 },
      {  -2,   4, -14, -52,   8,  24,  11, -42,   6,  33,  64,  19, -27, -14,  44,  50 },
    },
    {
      {  93, -50,   0,  -4, -65,  14,  16,   0,   7,  18,  -6,  -3,  -2,  -5,  -4,   1 },
      {  54,  17, -18,  -1,  43, -81,  10,   6, -48,   9,  38,  -8,   1,  20,  -9,  -7 },
      {  45,  -3,   6,  -4,  34,  29, -50,  -1,   1, -83,   4,  23, -18,  15,  41,  -3 },
      {  19,  94, -30,  -4, -55, -12, -33,  15,  27,  -5,  11,  13,  17, -21,   0,  -5 },
      {  35,  37,  -9,  -4,  44,  32,   2, -26, -31,  25, -73,  -5,   3, -39, -15,  37 },
      { -14,  13, -11,   3, -40,  42,  -2,  -7, -90,  -6,   4,   7,  32,  56,  -2,  -2 },
      { -21, -33, -52,  18, -21, -47,   5,   1, -20, -18, -34,  28,  23, -34,  64,  29 },
      {   3,  15, -61,   1,   9,  31,  79,  24,  12, -46,   7, -47, -11,  -1,   0,  -5 },
      {   1,  37,  70,  -1, -29, -34,  46, -38,   0, -28, -28, -24, -17,  21,  38,  14 },
      {  15,   4,  34,  17,  24,  19,  35, -11,   9,  -4,  28,  22,  94, -32,  18, -28 },
      {  -8, -16,  24,  -3, -19, -19, -10,   6, -21, -63,  15, -15,  18, -43, -71,  50 },
      {  -4,  10,  33,  11,  -8,  17,  12,  51, -53,  10,  22,  11, -53, -66,  25, -27 },
      {   7,  -3,  18,  64,   3, -12, -30,  62,   2,  -4, -49, -55,  27,  17,  -2, -24 },
      {  -8, -13, -11, -52,  -8,  -6, -34, -41, -18,  -6,  -2, -71,  14, -35,  17, -56 },
      {   1,   0,  17, -63,   9,  11,  -8,  57,   4,  23,  19, -32,  27,  11,  43,  64 },
      {   3,   4,  -9,  69,   0,  19, -22, -44,   1,  20,  56, -46, -11, -12,  27,  46 },
    },
  },
  { // 8
    {
      {  76, -52,   5,  -3, -74,  39,   5,   0,  24,   2, -11,   1,  -5,  -6,   3,   1 },
      {  75, -10, -15,  -2,  33, -59,  17,   0, -62,  40,   5,  -3,  10,  -3,  -3,  -1 },
      {  15, -50,  49, -13,  59,  38, -58,   4, -30, -34,   8,  15,  -5,  12,  13, -10 },
      {  50,  55, -26,  -6,  -2, -23, -29,  16,  26, -66,  43,   1, -28,  27,   4,  -8 },
      {  30,  52, -19,   0,  25,  30, -20, -15,   5,  -4, -70,  23,  12, -59,  33,  10 },
      {  -1,  46,  51, -35, -46,  -4, -57,  29, -21,  39,  21,  -1,  38, -16, -14,  -3 },
      {  28,  22,  48, -18,  42,  11,  14, -14,  50,  27, -16, -46, -33,   7, -57,  17 },
      {  -6, -35,  23, -15,  10, -70, -15,  28,  64,  14,  -6,  16, -19, -39,  44, -11 },
      {  -4, -24, -74, -23,  26,  33, -26,  57,  17,  23,  25, -30,  11, -31, -26,   5 },
      {  17,   0,  14,  55,  22,   7,   9, -32,  38,  -4,  55,   5,  71, -33,  -9, -22 },
      {  -6, -25, -21,  30, -13, -40, -63, -32,   7, -11, -28,  23,   5,   1, -65,  42 },
      {   3,  12, -22,  32,   6,  20, -39, -21,  19,  75,   8,  30, -32,  48,  20, -40 },
      {  -2,   4,  15,  72, -12,   2, -24,   8, -26,   4,  17, -62, -47, -37,  29,  29 },
      {  -9,  -3, -11, -48,  -7,   7,   1, -70, -14,   6,  53,  22, -48, -54,  -6,   5 },
      {   1,   9,  17,  36,   0,  11,  27,  54, -16,  -4,   1,  64, -41, -35, -53, -30 },
      {   6,   7,   9,   5,  12,  15,  17,  21,  12,  20,  33,  49,   4,  19,  24, 100 },
    },
    {
      { -98,  60,   0,   3,  51, -14, -13,   1,   2, -12,   6,   1,   4,   1,   0,  -2 },
      {  63,  21, -38,   1,  72, -57, -16,   9, -28, -17,  24,   3, -10,   5,   6,  -2 },
      {  34,  36,  -7, -10,  28,  78, -50,  -9,  42, -29, -34,  11,  -6, -19,   3,   6 },
      {  20,  89, -36, -11, -44,  12,  23,  -6, -32,  48,  -3, -14,   5,  -3, -19,   4 },
      {  13,  -8,  30,  -5,  66,   8,  58, -30,  26,  48, -25, -37,  11, -12, -36,   1 },
      { -12, -19, -79,  20,  -8, -30, -15,  26,  70,  30, -22,  -6,  29, -12, -12,   3 },
      { -19, -29, -30,  20,  14,   3, -12, -24, -68,  13, -76,  22, -10, -22,   4,  36 },
      {   9,  28,  13,   4, -23, -44,  53,  -8,  27, -63, -34,   5,  -8, -57,  17,  23 },
      {   7,   8,  42, -61, -17, -54, -59, -23,   5,   9, -34,   4,  35,   8, -31,   4 },
      { -17, -21, -53, -72,  -2,   6,  14, -70,  11, -14,  23, -18, -24,   8,  24,   9 },
      {  10,  12,  11,   3,  11,  -1,  17, -24,  16,  37,   4,  66,  52,  15,  78,   7 },
      {  -7,  -8,  10, -44,   8,  -1,  -5,  38,   6,  46,  30,  49, -48, -68,  -9,  19 },
      {  -3, -10, -23, -23,   8,  22,  37,  12, -13, -41,  13,  59,  51,  15, -67,  15 },
      {  -3, -13,  -4,   0,   2,  15, -18, -11, -32,  -7,  34, -36,  70, -80,  12, -23 },
      {   3,   5,   2,  60, -13, -13, -21, -79,  16,   6,  36,  37, -20, -18, -44, -11 },
      {  -3,  -2, -12, -17,   1,  -3,  16,   0,  -7,  -2, -37,  30, -14,  -9,   0,-114 },
    },
    {
      { -94,  47,   2,   3,  67, -12, -14,  -1,  -6, -20,   5,   3,   1,   6,   4,  -2 },
      {  58,   7, -18,   0,  54, -67,   1,   4, -57,  -3,  34,  -2,  -4,  27,  -2,  -6 },
      {  49,  10, -15,   1,  42,  31, -45,   2,  37, -69, -17,  15, -22, -25,  35,   3 },
      { -12, -96,  32,   9,  59,   6,  31, -14,   9,   2, -13, -16, -23,   4,   4,   6 },
      {  11,  21, -12,   0,  17, -59,  26,   4,  83,  29,   3, -22, -10, -48, -24,   6 },
      {  23,  31, -22,  -3,  43,  51,   5, -22, -31,  48, -56, -17,   3, -18, -46,  20 },
      { -25, -33, -98,  40,  -9,  -7,   4,  30, -20,   5, -12,  -5,  -2, -29,  29,   7 },
      {  -3,  14,  20, -18, -19, -38,  49, -38, -35, -35, -46,  19,   3, -43,  35,  42 },
      {   1, -31,  11,   8,  10, -34, -60,  -6,  10,  18, -33,  41,  84,  -8,  -8,  10 },
      { -11,  -5,  13,  10, -22, -36, -59,  -1, -11,  23, -38,  19, -88,   9, -16,  12 },
      {  -1,  -3,  19, -12,  14,  23, -11,   0, -21,  49,  62,  50, -16, -70,  27,   4 },
      { -13, -28, -28, -64,  -4,   1,   3,  20,  -7, -44,  16,  31,  -5, -17, -79,  18 },
      { -10, -14, -36, -39, -10,  -3, -38, -90,   7,   6,  32, -43,   1,   7,  16,  20 },
      {  -2,  -5,  37,  35, -13,  -4, -34,  16, -34, -31,  16, -75,  11, -57, -29,   5 },
      {  -3,  -5,   7, -85,   6,  -7, -17,  57,  -8,  23, -30, -43,   4,  -4,  44, -14 },
      {   4,   6,   8,   8,   3,   9,  -4,  36,  12,   9,  23,  -8,   5,  31,  11, 113 },
    },
  },
  { // 9
    {
      {-117,  30,  13,   4,  29,  14,  -4,  -3,  19,  -4,  -8,  -1,   7,  -3,  -2,   1 },
      {  32,  40, -20,  -1,  99, -10, -32,  -1,  -6, -43,  -3,   7, -15, -10,  12,   4 },
      { -16, -83,  15,  12,  49, -65,  29,   6,  17,   1,  27,  -8,  -2,  13,   6, -10 },
      { -27,  20, -23,  10, -35, -55, -22,  21, -71, -29,  37,  22,  -2,  29,  29,  -1 },
      {  14,  44,  79, -11,  16,  -1,  59, -26, -43,   3,  19, -24,  -5,  21,  -3, -16 },
      { -10, -14, -52, -10,  35,  45,   4,   0, -46,  72,  32, -11, -25,  27,  -8, -13 },
      {  14,  52, -19,  -7,   3, -45, -10,   3,  49,  46,  28,  -8,  61,  32, -18, -23 },
      {   4,   7,  -7, -17, -11,  39,  23,   6,  49, -25,  75,  14, -16,  -7,  65, -18 },
      {   9,  21,   7,  96,   2,  11,  18,  71,   5,   9,   4, -11, -16, -14, -10, -19 },
      {   8, -23,  34,  30,   9,  38, -43,  -4,   3,  -3, -24,   1,  32,  81,  46,  -6 },
      {   0, -15, -33,  -3,  14,  33,  52,  18, -27, -45,  -1,   9,  86,  -2, -19,  -3 },
      {   5,   4,  31, -11,  17,   1,  13,  32,  -2,  41,   4,  92,  10,  -3,   6,  59 },
      {   0, -14,  30, -34,  -1,  19, -37,  38,   5, -30,  26,  29, -18,  22, -74, -48 },
      {   2, -17,  34,  27,   5,  11, -55, -31, -26,  18,  50,  -3,  45, -67,   0,  -7 },
      {   2,   3, -17,  55,  -6,   8,  10, -66,  18, -24,  37,  23, -16,  33, -50,  43 },
      {  -1,   2, -12,  22,   0,  -9,  18, -45,  -3,  10, -34,  69,  -5, -12,  13, -84 },
    },
    {
      {  76, -55,   7,  -3, -71,  40,   7,  -2,  19,   6, -17,   2,  -4,  -8,   5,   2 },
      { -48,  -9,  30,  -3,  -1,  69, -49,   3,  38, -60,   4,  16, -10,   8,  19, -11 },
      { -16,  68, -53,   8, -54,  -6,  24,   7,  64,  -9, -12,  -4,  -9, -22,  11,   2 },
      {  56,  16,  -6, -11,   1, -38, -29,  26,  15, -48,  69, -13, -24,  36,  -9, -20 },
      { -23,  28,  40, -35, -54,   2, -47,  33, -12,  53,  29,  -9,  27, -12, -36,   2 },
      {  45,  45, -28,   6,  11,  18, -53,   3, -47, -12, -18,  39,  36, -35,  40,  -6 },
      { -11,  10, -53,  26, -21,  57,  14,  -3, -51, -16,  19, -19,  15,  49, -46,  18 },
      { -32, -27,  -3,  33, -45, -28,  19, -14, -17,  -1,  42,  35,  40,  13,  43, -57 },
      {  -7, -46, -41,   7,  29,  13,  -1,  45,  21,  -3,  36, -42,  43, -64,  -1,   1 },
      {  25,  25,  24,  46,  24,  20,  -3, -69,  38,  17,  29,   6,  25, -22, -48, -24 },
      { -18, -34, -21,  57, -23, -45, -61, -13,   1,  -6,  -6,  23, -27, -15, -26,  50 },
      {   0,  -6, -36,  23,  23,  29, -24,  30,  14,  73,   3,  17, -55,  22,   9, -49 },
      {  -1,  21,  33,  42, -16,  15,   6, -12, -39,   3,  29, -62, -52, -35,  47,   9 },
      { -14, -10, -24, -51, -10,   5,  15, -28, -33, -20,  20,  32, -53, -59, -38, -33 },
      {   6,  11,  36,  52,  -3,  -6,  25,  67, -15, -32, -39,  13,  -5, -19, -45, -40 },
      {   6,  10,  20,  10,  12,  21,  43,  32,   6,  12,  53,  71, -10,  -8,  11,  63 },
    },
    {
      {  92, -61,   2,  -1, -55,  14,  20,  -3,  -2,  20,  -7,  -5,  -2,  -1,  -4,   3 },
      {  48,  38, -44,   2,  23, -81,  17,  16, -30,  10,  41, -14,   0,  11,  -9,  -8 },
      {  -7, -71,  56,  -6,  73, -30,  10, -13, -33,  18,   0,  -6,  -9,  13,  -4,  -2 },
      {  48,   2,  11,  -6,  26,  17, -70,   7, -14, -65,  19,  37,  -9,  12,  38, -14 },
      {  -9, -39, -72,  65,  33,  35,  31, -16,  -8, -34,   9, -16,   1,   2,  16,   0 },
      { -15, -32,  18,   0, -18, -40,  20,  16,  83, -40,  45, -12, -12,   6,  15, -35 },
      {  37,  23,  -3, -21,  45, -13,  30, -21,  50, -26, -65,  -4, -30, -27,  11,  38 },
      {   0,  29,  39,  24, -32, -18,  42, -81, -27, -36,   0,  15,  16,  31,  18,  -5 },
      { -17, -28, -16,  34, -20, -60, -20,  14,   1,  14, -41,  58,  33, -26,  45,  16 },
      {  25,  24,  29,  75,  19,   4, -38, -34,  36,  46,  15, -12,  -5, -36, -10, -23 },
      {  17,   1,   3, -25,  29,  15,  20,  -2,  15, -13,   6,  14, 104, -33, -25, -28 },
      {   5,  20,  36,  12,  -4,  11,  24,  47, -30,  -3, -10, -60,  12, -32,  75, -15 },
      {   3,  13, -16, -19,  18,  24,  38,  -2,   6,  46,  -7,  56, -29,  14,  37, -74 },
      {  15,  15,  17,  41,  12,   6,   3,  48,  23,   2, -45,  -3,  27,  89, -17,  -3 },
      {   2,  -3, -19, -31,  11,   8, -21, -37,  32,  44,  28, -26,  35,  43,  63,  41 },
      {   7,  14,  29,  22,  11,  24,  44,  35,   6,   9,  56,  55,  -6, -10,   4,  67 },
    },
  },
  { // 10
    {
      {  80, -50,   1,  -2, -74,  32,  13,  -1,  17,  14, -17,   1,  -2, -10,   2,   4 },
      {  46,  15, -21,  -2,   9, -73,  33,   3, -47,  58,  10, -17,   8,   2, -30,  11 },
      { -49,  52, -27,   0, -52,  -2,  33,   2,  45,  38, -37,  -8,  12, -39,  -3,  16 },
      {  44,  60, -28,  -7, -24, -36, -15,  19,  44, -53,  39,  -1, -23,  17,  11, -17 },
      {  32,   7, -43,  14,  44,  -3,  25, -13, -14, -26, -53,  23,   7, -38,  70,   2 },
      {  17,  53,  27, -12, -28,  16, -62,  -1, -50,  12,   0,  31,  54, -27,  11, -11 },
      {  37,  33,   3, -19,  47,  41, -26,   5,  16,  -1, -28, -26, -30, -26, -55,  51 },
      {   5, -17,  73, -26,  13, -60, -24,  14,  43,  22, -22,   6, -22, -33,  33,  -2 },
      { -12, -40, -41, -27,   7,   0, -20,  64,   2,  -6,  43, -34,  37, -52,  14,  20 },
      {  17,  -7,  -9,  57,  32,   6, -12, -37,  48,  28,  39,  14,  24, -44, -26, -46 },
      { -11, -17,   5,  63, -30, -41, -29, -31, -12, -41,   2,   1,   0, -20, -11,  73 },
      {  16,  21,  60,   2,   9,   9,  59, -13,  10, -35,  16, -54,  61,  -3,  10,  11 },
      {   4,  23,  12,  59,  -3,  27, -15,  21, -18,  42,  19, -56, -44,   6,  57,   3 },
      {  -8,   9,  15, -31, -11,  16,  41, -30, -31,  -7,  59,  28, -56, -60,   6,   8 },
      {  -8,  -5,   7,  17, -17, -14,   0,  18, -35, -41, -40, -43, -24, -43, -38, -70 },
      {  -6,  -9, -27, -47,  -7,  -9, -37, -86,   2,   6,   0, -62,   0,   3,  21, -11 },
    },
    {
      {-101,  59,   0,   2,  47, -10, -12,   1,   7, -12,   3,   1,   4,  -1,   0,  -1 },
      {  52,  28, -32,   2,  66, -67,  -2,   8, -41, -10,  29,  -4,  -9,  10,   3,  -3 },
      {  38,  21,  -7,  -4,  37,  59, -56,  -2,  32, -59, -21,  19, -18, -19,  19,   4 },
      {  22,  89, -58,  -4, -52,  14,   9,   4,   4,  27, -13,  -6,  12, -17, -13,   7 },
      { -17,   3,  -9,  -1, -68, -18, -36,  25, -30, -56,  48,  25,  -7,  26,  35, -12 },
      {  -1, -20, -34,  15,  -6, -50,  16,  25,  99, -25,  13, -10,   7, -18,   1,  -3 },
      {  19,  24,  37, -30,   7,  22, -25,  15,  38,  24,  61, -26,  13,  52, -36, -35 },
      {  14,  40,  59, -20, -14,  -8,  68, -27,   6, -50,  10, -22, -39, -17,   8,   1 },
      { -12, -14, -53,  36,  17,  56,  44,   3,  -3,   5,  34, -28, -49,  30,  29, -17 },
      {  15,  17,  13,  30,  10,   6,  30, -34,  14,   1,   0,  35,  78,  33,  62, -18 },
      {   4,  -1,  15,  50,   9,  31,  21,  47, -26, -24,  40,   5,  46, -58, -41,  -4 },
      {   0,   1,  13, -42,  10,   4,   2,  29,   8,  55,  39,  31, -17, -60,  63,  -5 },
      {  13,  27,  47,  82,  -9, -19, -20,  34,  12,  32, -26,  14, -47,  13,  10, -10 },
      {  -2,  -4,  -2,  27, -12, -13, -42, -62,  -6,   7,  13, -55,   4, -49,  14, -61 },
      {  -1,   0,  -4,  27,  -5,  -1, -11, -69,  17,  14,  58,  50, -21,  -6, -31,  54 },
      {  -4,  -3, -16, -13,   2,  -3,  23, -11,   1,  -5, -15,  73, -23,  -4, -41, -87 },
    },
    {
      {  92, -48,  -1,  -4, -65,   6,  19,   0,   2,  28,  -5,  -6,  -2,  -4, -10,   4 },
      {  52,  11, -11,  -2,  51, -75,   3,   4, -46,  -7,  45,  -4,  -5,  25,  -1, -16 },
      { -20,  76, -32,  -1, -65, -25,   3,  13, -12,  43,  20,  -4,  31,   4, -30,  -5 },
      { -47, -51,  24,   1,  20, -27,  53,  -9,   0,  65,   1, -29,  -1,   8, -43,  -1 },
      {   9,  11, -14,   5,  23,  41, -14, -12, -89,  30, -51,   3,  -3,  12, -16,  38 },
      { -36, -45,  11,  11, -43,  18,  -9,   7, -52, -18,  41,  16,  17,  60,  29, -30 },
      { -18, -28, -85,  26,  -3, -13,  43,  27, -10, -20,  -1, -22,   7, -28,  38,  36 },
      {  10,   0, -48,  19,  27,  64, -14,  21,  14,  24,  40, -28, -39,   6, -30, -52 },
      { -20,  -7,  -6,  -1, -28, -36, -38,  12,   2,  26,  -2,  17,-101,   9,  17,  27 },
      {  -3, -42, -22,  22,  16, -23, -73,  17,  11,  39,  -4,  44,  51, -29,  -7,  -8 },
      { -13, -11,  20,   5, -20,   0, -16, -31, -46, -28,  41, -18, -16, -89, -30,  -4 },
      {  11,  23,  29,  69,   6,   7,  -2, -38,   5,  48,  20, -26,   5, -10,  71,   4 },
      {   5,  -4,  14,  69,  -9,  -9, -29,  12,  17, -39,   2, -40,   7,  37, -54,  49 },
      {   7,   7,  28, -11,  17,  33,  22,  49,   3,  18,  66,  45,   5, -10,  -2,  66 },
      {   3,  13,  37,  51,  -3,  -8,  30,  76, -21,  -7, -40,  22, -15, -25,  -2, -45 },
      {  -1,  -4,  26, -48,   1,  -1, -40,  60, -10,  12,  -5, -82,  16,  -8,  30,   6 },
    },
  },
  { // 11
    {
      {-117,  29,  13,   4,  32,  14,  -4,  -2,  18,  -5,  -7,   0,   6,  -3,  -2,   1 },
      {  33,  47, -23,  -3,  94,  -8, -34,  -3, -18, -40,   0,   7, -15,  -4,  13,   3 },
      { -21, -69,   8,  15,  38, -81,  21,  13, -16,  -8,  34,  -1,  -3,  22,  11, -10 },
      { -12,  62,  17, -13, -42, -25,  13,  -7, -79, -18,  36,   1,   7,  34,   9,  -9 },
      {  19,   3,  69, -19,  48,  26,  59, -34, -12,  34,  -5, -28,  -8,  11, -29,  -7 },
      {  -7,   5, -59,   5,  21,  10, -20,  15, -24,  89,  15,  -3,   7,  42, -30, -12 },
      {  21,  54,  24,  27,  -1, -60,   0,  22,  56,  13, -11,  -7,  54,  11, -31, -10 },
      {   2,  10, -27, -62,   1,   0,  12, -40,  57,   3,  69, -13,  16,   6,  35, -17 },
      {  10,  13,   6,  62,   7,  51,  36,  59,  16, -12,  55,   2, -13,  10,  24, -23 },
      {   6, -19,  42,   9,   0,  23, -55, -17,  14,  -5, -20,   4,  14,  87,  44,  -8 },
      {   2, -26, -23, -20,  15,  33,  28,  12, -20, -43,  -8,  26,  92,  16, -29,  -1 },
      {  -4,  -8, -28,  20,  -9,   5, -10, -19,  -4, -39, -14, -90,  -5,   6, -27, -64 },
      {  -2,  -4,  13, -39,  -8,   2, -17,  24,  25, -31,  18,  39, -53,  33, -77, -23 },
      {  -2,  18, -44, -17,  -3, -13,  68,  16,  19, -10, -60,  -2, -33,  52,  27,  13 },
      {  -1,  -2,  23, -60,   6,  -2, -12,  70, -13,  20, -27,  -8,   9, -27,  37, -61 },
      {   0,  -4,  10, -30,   0,   6, -17,  52,   0, -10,  20, -70,   3,  14, -10,  81 },
    },
    {
      { -69,  56, -10,   2,  70, -49,   0,   1, -27,   5,  14,  -4,   7,   5,  -7,   1 },
      { -53,  -5,  29,  -3,  13,  57, -48,   4,  34, -67,  15,  11, -14,  19,  13, -12 },
      {  35, -54,  50, -15,  44,  -6, -37,   8, -63,   6,  35,  -2,  11,  21, -17,  -4 },
      {  53,  27, -28,  -1,  -1, -42,  -2,  18,  29, -48,  58, -19, -36,  39, -13, -10 },
      { -31,  25,  33, -33, -59,  -1, -27,  38,  12,  41,  33, -30,   9,  -2, -53,  16 },
      { -48, -54,  -5,  29, -19,   0,  64, -26,  -8,  -2,  22, -19, -12,  55, -32,  -1 },
      {  20,   2,  -9,   2,  55,  40,   1,  -8,  25,   8, -29, -33, -38,  -9, -64,  53 },
      { -15, -45,  46, -34,  22, -55,  21,  21,  55,  10, -16, -10, -40, -22,  30,   0 },
      { -17, -43, -52,   5,   3,   4,   0,  55, -13, -31,  21, -48,  32, -46,  20,  33 },
      {  -1,  24,  -1, -69, -10,  25,  30,  23, -50, -16, -34, -27, -30,  43,  35,  15 },
      { -15, -13,  -3,  11, -38, -47, -33, -11, -34, -39, -27,  42, -38, -10, -19,  62 },
      { -16, -28, -59,   5,   4,   6, -40,  47,  -4,  47, -19,  24, -47,  23,  -8, -45 },
      {   5, -15, -16, -37,  12, -20,   5,  25,  32, -24, -44,  34,  73,  42, -39,   3 },
      {   4,   2,   6,  42,   1,  -9, -37,   1,  24,  36, -10, -25,  19,  64,  54,  56 },
      {  10,  22,  48,  69,   0,   3,  30,  77, -15, -19, -32,   3,  -5,   2, -12, -14 },
      {   2,   4,  -3, -10,  15,  29,  38,  32,   9,  24,  55,  79,  -7,   0,  15,  51 },
    },
    {
      { -95,  58,   2,   1,  54,  -9, -20,   2,   5, -22,   4,   4,   1,   2,   5,  -2 },
      {  43,  57, -55,   3,  12, -73,  16,  16, -25,   4,  36,  -9,   5,   9,  -7,  -8 },
      { -12,  62, -38,  -1, -84,  34,   3,   6,  43,  -5, -14,   3,  15, -21,   1,   4 },
      { -52, -19,   0,  10, -16, -30,  70,  -8,   2,  67, -11, -31,   7,  -5, -36,  10 },
      {  -7, -27, -76,  67,  28,  45,  26, -16, -17, -24,   1, -17,   5,  -8,  18,   3 },
      {  -7,  19,  -2,   0, -14,  37, -40,  -2, -93,  47,  -3,   9,  39,   7, -17,   5 },
      {  37,  36,  -4, -16,  50,   8,  10, -26,   5,  19, -81,  -2, -11, -38,  -7,  40 },
      {  -7,  10,  19, -23, -24, -11,  62, -34, -44, -60, -30,  16,  24,  39,  29,   6 },
      {   7,  11,  42,  77, -18, -43, -25, -57,   6,   3,   5,   7,  28, -33,  22,   9 },
      { -27, -35, -39,  -1, -25, -48, -28,  29, -16,  14, -46,  45, -29,  -8,  50,  18 },
      {  -5, -30,  -9, -37,  14, -13,   6,  25,   4, -22,  11,   5,  85, -71,   3,   4 },
      {   9,  19,  15,  -6,  17,  20,  25,  11,   7,  50,   0,  -9,  11,  -4,  89, -60 },
      {   9,  -5, -12,  25,  21,  -4,  -3,  10,  42,  16, -35,  55,  58,  61, -28, -27 },
      {   6,   6,  29,  37,  -6,  -7,  -9,  76,  -5, -20, -39, -69,  20,  22,  10,  23 },
      {  -4, -15, -30, -39,   1,  -9, -38, -49,  28,  17,   9, -53,  31,  47,  33,  40 },
      {  -7, -12, -17, -12, -12, -21, -30, -29, -11, -21, -48, -45,  -5, -14, -27, -88 },
    },
  },
  { // 12
    {
      {  73, -47,   4,  -3, -76,  41,   5,   0,  30,   1, -14,   2,  -7,  -9,   6,   0 },
      {  51,   9, -20,  -1,  -6, -61,  33,   0, -46,  70,  -2, -11,  21, -18, -21,  11 },
      {  46, -55,  31,  -3,  45,   4, -30,   0, -62, -22,  30,   6,  -8,  42,  -7,  -8 },
      { -45, -44,  37,   0,   6,  45,  -6, -12, -36,  48, -41,  12,  41, -39,  11,   6 },
      { -40,   4,  33, -15, -57,   7, -15,  12,  -3,  28,  55, -25,  12,  38, -61,   1 },
      {  38,  61,  11, -12,   0,  20, -74,  19,  -8,  -9,  16,   3,  47, -45,   2,   4 },
      {   0, -35,  46, -30,  28, -43, -24,  24,  52,  19,  11, -19, -42, -50,  -6,  23 },
      {  25,  20,  -7,   6,  49,  54,   7, -30,  25,  20, -26,  -9, -19,   6, -79,  25 },
      {  -9, -50, -59,  28,  20,  12,  13,  16,  14, -23,  51, -32,  49, -37, -18,  12 },
      { -12,   8, -39, -27,  -8,  39,  -5,  55, -51,   6,  -9, -45, -61, -12,  15,  23 },
      { -12,   1,  -4,  65, -29,  -9, -27, -50, -30,  -6,  28,  26, -54, -44, -12,  22 },
      {  11,  26,  56, -12,   1,  17,  64, -37, -16, -33,  29, -48,   7, -22,  25,  30 },
      { -10, -10,   2, -36, -22, -28,  14,  16, -26, -68, -41,  34,  14, -19, -56,  37 },
      {   1,  -5,  11,  54,  -9, -25, -37,   2,   6, -10, -48, -60,  21,  38,  13,  56 },
      {  -7, -11, -41, -64,  -2,  -1, -25, -58,   8,  15,  21,  16,   9,  23,  29,  63 },
      {  -6, -11, -18, -32,  -9, -18, -29, -58, -10, -19, -28, -63,  -3, -19, -17, -66 },
    },
    {
      { -91,   7,  26,   2, -76,  15,  24,   1,  -3,  19,   1,  -2,  18,   6,  -6,  -2 },
      {  15, -84,  24,  11, -16, -69,  37,  10, -18,   3,  29,  -3,  -2,  23,   7,  -7 },
      {  54,  24,  -2, -12, -54,  33,  12,  -7, -83,  13,  19,   2, -32,  11,   8,   1 },
      { -53,  27, -27,  17,  20, -44, -25,  22, -54, -53,  23,  17, -30,  -6,  29,   3 },
      {  17, -36, -64,  27, -63,  -3, -61,  31,  13,   0, -27,  14,  15,  -8,   4,   3 },
      {  22,  -5,  35,  -7, -37,   3,  24,  -2,  24, -77,  -7,  13, -14, -74,   3,  17 },
      { -20, -42,  23,  39,  23,  38,  -5,  24, -11,  29, -37,  -1, -81, -23, -10,   5 },
      { -20, -33,  -1, -71,  -1, -19, -34, -64, -25,  -7, -46,   6, -15,   6, -20,  27 },
      {  12,  27,  20,  30,   7, -35,  12,  25, -44,   2, -58,   8,  40, -11, -70,   6 },
      {  -7,   9, -54,  -6,  -1, -34,  26, -21,   0,  52,  26, -32, -14, -77, -19,   9 },
      {  -9, -30,   1, -15,  27,  34,   3,  18, -25,  23,  32,  74,  47, -31,   8,  46 },
      {  10,  39,   0, -17, -19, -42,  19,  21,  44,  28,  -9,  62, -49,  29,  -2,  41 },
      {   8,  17,  33,  21,  -1, -25, -10, -18, -10,  35, -46, -12,  22, -22,  92,  20 },
      {  -7, -12, -61,  30,   9,  23,  75, -28,  -4, -28, -38,   2,   5,  28,  17,  34 },
      {  -4,  -4, -21, -53,   7,   0,  34,  34,  -7,   6, -41,  30,   1, -17,  25, -84 },
      {  -2,  -4,  -4, -54,   4,   6,   8,  77,  -5,  -6, -10, -66,   5,   8,  10,  53 },
    },
    {
      {  93, -46,  -4,  -4, -68,  16,  15,   0,   4,  21,  -6,  -3,   0,  -5,  -5,   3 },
      {  57,  -8, -10,   0,  60, -57,   1,   4, -64,  -7,  31,   0,  -8,  28,   3,  -9 },
      {  18,  84, -52,   1, -46,  12, -30,  17, -26, -24,  26,  10,  27,   0,   7,  -8 },
      { -46,  -5,   3,   1, -39, -14,  33,   0, -60,  67,  10, -11,  32,  30, -33,  -2 },
      {  14,  10, -18,   4,  42,  82, -27, -10, -47,  36, -44,  -2, -24,  -1, -12,  15 },
      { -36, -62, -23,  18, -24,  37, -11,  17, -34, -37,  33,   5, -26,  26,  53, -19 },
      {   5,  17,  81, -30, -24,   6,  -3, -26, -42, -46, -36,  28,  11,  28,  18,  12 },
      {  -3,  29,  17,   8, -30, -34, -38,  16,  12,  42,  -9,   5, -90,  33,  14,  -5 },
      { -15, -19, -10,   6, -15, -39, -21,   4, -41,   7, -16,  39,  -8, -87,  15,  45 },
      {   6, -22,  10,  63,   5, -17, -67,  21,  13,  13, -37,  10,  62,  28,   8,  -8 },
      {   9,  26,  37,  60,  -4,   5,  16, -26, -17,   8,  13, -73,  -1, -37,  47, -11 },
      {  -5,   7, -47,  30,  -9, -25,  54,  15,  -5, -37, -76, -23, -15,  26,   4,  27 },
      {   2,   6, -27, -44,  12,  -8,  13, -25,  16,  47, -16,  13,  28,  15,  93,   3 },
      {   7,  15,  26,   6,  13,  14,  44,  80,  -7,  13, -18,  38,   1, -26,  16, -60 },
      {  -9, -17, -11, -60,  -9, -20, -40,   9, -16, -12, -41, -63,   3, -23,  -8, -59 },
      {   2,   0,  23, -31,   7,  10, -10,  79,   2,   3,  18, -47,  12,  10,  15,  74 },
    },
  },
  { // 13
    {
      { 114, -18, -17,  -2,  32, -22,  -8,   1, -26,  -9,   8,   2, -16,   5,   6,   0 },
      {  27, -41,  11,  -2,-108,  -6,  23,   1, -14,  30,   5,  -1,  15,  13,  -6,  -2 },
      {  39,  70, -18, -11, -29,  80, -24,  -9,  15,  14, -26,   0,   5, -19,  -8,   4 },
      {  18, -56,   3,   0,  20,   7,  -2,  -4, 100,  27, -29,  -6,  20, -14, -15,   2 },
      {  -9,  -9, -69,  16, -37, -25, -58,  28,  25, -50,   9,  22,   7, -22,  28,   2 },
      {  17,  16,  52, -11, -18,  -5,  49,  -5,  30, -74,  14,   8, -14, -57,  11,   9 },
      {  16,  52,   5,   3,   3, -42,  13,   6,  19,  -8,  15,   3,  96,  34, -17, -10 },
      {  -1,   6, -31, -34,   1,  20,  22, -27,  36,  14,  86, -20, -15,  27,  45, -19 },
      {   9,  10,  13,  86,   4,  28,  22,  72,  16,  10,  25,   0, -18,  14,   5, -24 },
      {   5,   6,  35,  -5, -16,   1, -29, -12,  27, -45, -37,  -5, -34,  90,  15, -14 },
      {   3,  51, -15,  12, -18, -72,   5, -10,  26,  37, -16, -43, -55, -17, -16,  -8 },
      {  -2, -21, -21,  35,  -2,  21,   2, -36, -17, -43,  -4, -89,  22, -10, -22, -38 },
      {   5,   4,  47,  25,   1,  -9, -37, -18,  -8,  30,  -9, -24,  27, -25,  93,   2 },
      {   1,  -3,  51, -32,  -3,  -3, -71,  25,   1,   6,  48,  -9,  -5, -22, -44, -47 },
      {  -2,  -1, -17, -54,   5,  -1,  32,  53,  -9,   1, -48,  -9,   7, -11,  34, -75 },
      {   0,  -1,   3, -36,   1,   5,  -4,  65,   0,  -7,   4, -70,   5,  11,   4,  75 },
    },
    {
      {  69, -43,   4,  -3, -78,  42,   2,   1,  39,  -7, -11,   1, -12,  -7,   6,   0 },
      {  63,   5, -20,   0, -12, -52,  30,  -3, -55,  63,  -7,  -3,  29, -14, -10,   5 },
      {  52, -51,  29,  -7,  52,   8, -34,   3, -51, -35,  30,   3, -13,  37,  -5,  -4 },
      { -45, -47,  44,  -4, -10,  52, -14,  -6, -38,  46, -21,   7,  50, -22,  -9,   2 },
      {  22, -29,  -2,   4,  70,   3,  10, -14,  31,  14, -51,   6, -28, -65,  31,   8 },
      {  27,  43,  29, -29,   7,  -7, -75,  35,  29,  13,  28,  -9,  30, -48,  -9,   1 },
      { -26, -45,  50, -18, -12, -57,   8,  19,  27,  41,  22, -26, -56,   7, -19,   7 },
      {  20,   8,  -8,   0,  43,  35,  16,  -5,  51,  39, -10, -26,  22,  53, -65,   4 },
      {  -7, -47, -15,  14,   8, -24,  37,   2,  22, -50,  43, -22,  59, -43, -31,  11 },
      {  -6,   7, -44,  18,  -1,  51,  -4,  10, -34,  19,  44, -42, -55, -41, -28,  24 },
      {  -4,   2, -10, -86,   5,  16,  38,  66, -19, -22, -31, -28,   4,   3,   9,  13 },
      {  17,  49,  76, -15,  -1,  18,  54, -46, -13, -23,  13,   2, -15, -23, -22,  13 },
      {  -4,  -2,  -6,   4, -16, -26, -36,  -3, -16, -36, -64,  18, -17, -11, -77,  43 },
      {   8,  14,  36,  66,  -4,  -5,  -3,  28,  -8, -13, -33, -77,  14,  14,  32,  31 },
      {   6,   9,  20,  50,   6,  10,  30,  79,  -4,  -1,  -5,  46, -12, -13, -29, -52 },
      {  -2,  -1,  -4, -12,  -7,  -8, -13, -29, -12, -15, -27, -59,  -8, -17, -21, -99 },
    },
    {
      {  83, -64,   8,  -1, -61,  32,  11,  -4,   3,  15, -13,   1,  -1,  -4,   0,   3 },
      {  47,  20, -46,   7,  10, -72,  44,   5, -37,  41,  17, -20,   6,   5, -20,   5 },
      {   5, -61,  65, -12,  70, -22, -13,  -8, -43,  19,  -2,   4, -11,  18,  -5,   0 },
      {  57,   7, -20,   5,  23,  -7, -44,  17,  -2, -69,  52,  11, -19,  26,  23, -25 },
      {  20,  23,  45, -47, -10, -56,   9,  15,  71,   0, -24,   5, -44,   5,   5,  -1 },
      {   4,  39,  38, -42, -45,  -2, -57,  38, -44,  31,  25,   5,  34,  -3, -15,  -6 },
      { -41, -38,  18,   4, -40, -17,  29,  -1,  21, -13,  72, -32,   6,  44, -32, -20 },
      {   9,   4,  29,  -7,  -9, -35,  30, -48,  -5, -46,   1,  41,  77, -24,  19,  -7 },
      {  -5,  31,   3, -14, -29,  15,  25, -38, -48, -13, -25,  40, -40,  76,  -5,   0 },
      { -33, -30, -10,  33, -39, -45, -13,  19, -21,  20,  -4,  27, -24,  -9,  79, -27 },
      { -19, -33, -35, -75,   8,   9,  39,  64, -19, -32, -17,   4,  17,   6,  15,   7 },
      {  -3,  13,  31,   0, -21,  -3,  18,  -8, -45, -46,  12, -59, -43, -48,  19,  47 },
      {  -7, -21,  -8,  38, -19, -41, -41,  25,   3, -34, -41,   5,  20,  29, -36,  68 },
      {  16,  28,  37,  38,   9,  17,  24,  36,   3,   0, -31, -54,  40,  50,  45, -16 },
      {  -3, -10, -30, -48,   0,  -7, -38, -59,  15,  21,   7, -43,  21,  33,  51,  37 },
      {  -6, -11, -17, -14, -12, -22, -28, -24, -17, -27, -55, -50,  -2, -11, -32, -78 },
    },
  },
  { // 14
    {
      { 101, -41,  -6,  -3, -63,  10,  12,   2,  -6,  15,  -2,  -2,   2,  -2,  -3,   1 },
      { -58, -16,  14,   2, -75,  46,  10,  -1,  59,  16, -22,  -3,  13, -18,  -4,   4 },
      {  11, 100, -30,  -7, -54, -10, -18,  12,  -7, -24,  16,   7,  29,  -1,   2,  -4 },
      { -42, -19,   8,   3, -40, -49,  32,   3, -66,  45,  26,  -6,  28,  33, -15,  -6 },
      {   5,  -8,  -1,   0,  -9, -88,  19,  14,  80, -13,  27,  -6, -20,  10,   5,  -9 },
      { -21, -23,  12,   2, -39,  13, -16,   7, -31, -70,  27,  11, -65,  29,  37,  -6 },
      { -13,  -7, -98,  24,   6,  28,  17,  21,   3,  30,  42, -22, -35, -11,  -2,  -9 },
      {  -4, -47, -26,  10,  13,  11, -14,  10,  10, -52,  35,  10,  91,   7,  11, -12 },
      {   1,  -9,   7,   6,  -5,  -4,-104,   9,  15,  57,  21,  16,  -3,  28,  11, -11 },
      {  13,  21,  18,  -8,  14,  44,  24,  -6,  27,   0,  21, -22,   0,  95, -39, -22 },
      {  11,  17,  62,  25,   6,  22,  13,  -4,   1,  14,  84, -22,   2, -51,  14, -16 },
      {  -6,  -8,  -2,-102,   7,  11,   9,  23,   1,   9,  32,  54,  -8, -20, -22, -12 },
      {  -6, -12, -15, -12, -12, -12, -32, -82,  -1, -23,  28, -22,  -9, -17, -68,  32 },
      {   2,  10, -18, -11,   2,   5,  26, -82,  12,  24,  14,  40,   7,  21,  73,   9 },
      {   4,   5,   6,  64,   1,   4,  15,   8,   5,  -1,  12,  95, -10,   3, -45,  24 },
      {   1,   2,   6, -15,   4,   7,   2,  34,   6,   6,  23, -17,   7,  18,  18, 115 },
    },
    {
      {  73, -61,  14,  -4, -66,  47,  -2,   0,  18,  -1, -11,   3,  -4,  -4,   4,   0 },
      {  60,  11, -42,  10,  -9, -61,  51,  -7, -36,  50,  -9,  -8,  13,  -7, -10,   6 },
      {  12, -55,  54, -14,  57,  -2, -29,   3, -69,  31,   1,   5,  12,   3,  -3,  -1 },
      {  54,  21,  -2, -21,   8, -30, -37,  33,   0, -40,  70, -15, -23,  36, -16, -11 },
      { -17,  46,  19, -27, -64,   7, -39,  31, -30,  30,  11,  -1,  62, -18, -13,  -1 },
      {  13,   9, -44,  45,   7,  47,  -3, -33, -58, -52,  18,  14,  45,   8,   9,  -5 },
      {  54,  53,  14, -11,  33,  -4, -36,  -6,  16, -13, -50,  37,   2, -39,  46,  -5 },
      {  -8,  13,  48,   2, -40, -39,  13, -42, -19, -23, -24,  48,  -8,  68,   2, -19 },
      {   7, -30,  22,  12,  19, -41,   9, -11,  57, -26,  15,   2,  89, -10, -15,  -7 },
      { -26, -40, -20,   1, -27, -44,  -5,  22, -19, -16,  25,  18, -10, -39,  76, -40 },
      {   2,   7,  13,  73, -11, -10, -51, -49,  19,  47,  42,  11, -24, -21, -11,  -5 },
      {   0,  -8, -47, -14,  21,  24, -12,  19,  29,  53,   3,  52,  19,  51,   2, -56 },
      {   3,  14,  16, -61,   7,  24,  47, -50,   0,   5,  59,  34,  -9, -42, -10, -16 },
      {  12,  30,  47,  26,   9,  28,  41,   8,   6,  23,  18, -57,  12,  25,  59, -39 },
      {  -4, -10, -28, -48,  -4,  -9, -35, -62,  11,  17,   7, -37,  17,  34,  52,  45 },
      {  -1,  -4, -15, -21,  -3,  -8, -25, -40,  -7, -13, -36, -56,  -6, -17, -35, -84 },
    },
    {
      { -83,  78, -26,   7, -33,  31,  -5,  -2,  14, -15,   8,  -3,   9,  -8,   1,   1 },
      { -71, -35,  67, -13,  -5, -47,  47,  -9,  24,   2,  -9,   2,   7,   9, -10,   2 },
      {  -6,  31, -36,  16,  86, -63,   7,   2,  29, -35,  13,  -2, -12,   8,   0,  -1 },
      {   6,  20, -19,   6, -32, -43,  53, -12, -82,   7,  50, -14,  -8,  19,   0,  -2 },
      { -30,  24,  17, -18,  62,  18, -15, -12, -55,  68, -36,   3, -27,  12,  -1,   0 },
      {  42,  48,  34, -59,  17,  29,  47, -54,  19, -20,  10,   1,   4, -16, -10,  13 },
      { -21, -19,  33, -17,  14,  20, -25,  15, -40, -70,  31,  14, -55, -28,  36,   0 },
      {   8,  23,   6, -22,  -7, -39, -18,  22, -43, -30, -50,  53,  58, -32, -17,  14 },
      {  21,  29,   4,   4, -39, -18,  11,  17,  17, -11, -51,  28, -85,  36, -21,   8 },
      { -22, -36, -58, -73, -16, -20, -36, -54,   4,  -4,   1,  12, -11,  23,  -5,  12 },
      {  -1,  -5, -25,   9, -14, -25,  24, -22,  11,  32, -25,   1, -30, -93,  50,   4 },
      { -16, -37, -31,  42,  19,  39,  36, -25, -19, -24, -13,  16,  -9, -21, -66,  46 },
      {  -4,  -4, -16, -27,   9,  13,  26,  52,  21,  41,  53,  80,  -5,  -4,  -1,  16 },
      {   2,  12,  16, -24,  -8, -26, -38,  29,   6,  21,  32, -49, -23, -47, -72,  24 },
      {   7,  16,  35,  46,  -9, -21, -45, -48,   8,  16,  30,  31,   2,  13,  20,  72 },
      {  -2,  -3, -14, -26,   4,   8,  23,  46,  -2,  -6, -22, -48,  10,  20,  41,  88 },
    },
  },
  { // 15
    {
      { -96,  46,   2,   4,  66, -12, -12,   0,   1, -18,   5,   1,   0,   3,   3,  -2 },
      {  59,  -2, -11,   1,  75, -47,  -3,   0, -62,  -7,  20,   0, -12,  19,   1,  -3 },
      {  47,  27,  -9,  -5,  38,  40, -27,  -1,  64, -53, -13,   5, -37, -26,  19,   3 },
      { -19, -96,  31,  13,  44,  -8,  18,  -7,  40,  18, -13, -10, -34,   4,   1,   1 },
      {  16,  17,   6,  -5,  47,  48,   2, -17,  -6,  70, -44,  -3,  45, -32, -28,  14 },
      {  16,  23, -32,   4,   5, -76,  26,  15,  72,  29,  13, -11,  30, -17, -14,  -6 },
      { -11, -46, -78,  37,  10,  23, -48,  29,  -2,   1,  18,  10,  42,  -6,  22, -11 },
      {  13, -17,  56,   1,   9,  -8,  -6,  -8,  13, -58,  -6,  14,  92,  19,  -1,  -5 },
      {   9,  14,   3,  -8,   9,  39, -13,   3,  32,  27,  46, -18,  -1,  92, -32, -23 },
      {   3,   3,  40, -18, -13, -38,-101,   8,   8,  40, -11,  20, -12,  -1,  14,  -2 },
      {  11,  25,  20, 108,  -8,  -6,  -4,  20,  -3,   1, -43, -23, -11,  22,  -4,  -4 },
      {   5,   6,  50,  26,   4,  20,   1,  22,  -6,  13,  84, -25,   3, -62,   8, -23 },
      {   5,  16,  10,   2,   6,   7,  35,  -9,   6,  34,  -3,  34,   9,  23, 104, -27 },
      {   0,   0,  14, -15,   9,  10,  23, 101,  -3,   1, -16,  62,  -9,   1, -25, -14 },
      {   2,   3,   0,  42,  -2,  -1,   3, -49,  10,   8,  37,  93, -10,  -3, -29,  33 },
      {  -1,  -3,  -9,   2,  -2,  -4,  -2, -37,  -5,  -6, -20,  19,  -9, -18, -29,-113 },
    },
    {
      { -76,  56,  -6,   2,  71, -43,  -3,   1, -19,  -2,  11,  -2,   3,   6,  -3,  -1 },
      { -54, -20,  43,  -4,   8,  63, -48,   0,  39, -52,   1,  12, -17,   6,  14,  -6 },
      {  12, -65,  43,  -4,  54,  12, -21,  -7, -72,  31,  -2,   5,  20,  -2,  -3,   1 },
      { -58,  -2,  15,   2, -31,  15,  22, -14,  12,  62, -57,   8,  29, -51,   8,  12 },
      {   4, -46,  -7,  14,  50, -19,  48, -25,  43,   2, -38,   0, -67,   9,  12,   8 },
      { -21, -17, -66,  49,   5,  53,  38, -35, -32, -36,   8,   1,  31,   3,  -3,   3 },
      { -51, -35,  21,  -2, -47, -13,  30,   5, -11,  18,  33, -31, -18,  62, -47,   7 },
      {   4, -27, -21,   3,  37,  11, -10,  29,  57,  20,  40, -54,  32, -40, -41,  21 },
      { -11,  11, -37,   3,  -9,  37, -25,  33, -41,  32,   9, -31, -81, -28,  17,   8 },
      {  -4,  -1, -11, -82,  15,  26,  51,  56, -15, -30, -34, -12,  15,   9,   1,   5 },
      {  28,  55,  13,  -3,  25,  58,  -3, -30,   9,  34, -24,   2,  -9,  41, -57,  23 },
      {   9,  17,  49,   2, -11, -11,  24, -28, -27, -54,   4, -31, -19, -57, -24,  57 },
      {  -3,   3,   3, -47,   5,  18,  37, -33,   5,  17,  66,  51, -21, -40, -17, -42 },
      {  13,  28,  48,  27,  11,  27,  46,   3,   5,  16,  26, -55,  12,  12,  62, -35 },
      {   4,   6,  23,  56,   3,   2,  20,  60,  -8, -15, -24,  23, -12, -24, -56, -58 },
      {   2,   4,  14,  29,   5,  10,  26,  51,   7,  14,  35,  64,   2,  11,  25,  75 },
    },
    {
      {  99, -24, -20,  -1,  66, -19, -14,   0, -16,   1,   3,   1, -18,   4,   4,   0 },
      { -14, -91,  44,   6, -10, -64,  34,   3,  -1,  13,  -2,  -2,   2,  18,  -8,  -1 },
      { -43,  -2,  -3,   4,  76, -18, -22,   2,  85, -17, -21,   1,   4,  -3,  -1,   0 },
      {  30,  21,  75, -40,  18,  40,  49, -34,  23,  24, -27,  -2,   7,  -3, -22,   6 },
      { -24,  41,  29, -13,  36, -40,  38,  -7, -44, -68,  32,   5, -32,  -5,   9,   2 },
      {  50,   9,  -4,  -3, -52, -29,  16,   8,  53, -69,   1,  10,  41, -21, -10,   2 },
      {   0, -34, -17, -46, -27,  16,  -3, -45,  38, -10,  30, -12, -73, -19,  39,   5 },
      { -15,  -6, -25, -75,   2, -31, -37, -53, -32,  -4, -28,  17,  44,   4, -16,  19 },
      {   2, -13,  33,   7, -15,   9, -39,  26, -19, -30, -78,  40, -49, -32,  -2,  15 },
      { -12, -46,   6,  10,  27,  43,   6,  -2, -21, -16,  29,   6,  47, -85,  14,  15 },
      {  -1,   8, -14, -31,   3,  -7,  24,  44,  15,  36,  30,  95,  -4,   1,  11,  32 },
      {  -3,  21, -45,  15,  -3, -41,  43, -12,   1,  38, -26, -19, -27, -68, -43,  16 },
      { -10, -34, -41,  13,   9,  49,  33, -14,  -4, -42,  -2,  20, -16,  41, -71,  19 },
      {  -4,  -8,   7, -60,   1,   5, -21,  61,  -1,  -1,  25, -23, -12, -21, -60, -54 },
      {  -6, -12, -41, -26,   9,  14,  57,  19, -10, -12, -55,   2,  12,   5,  52, -62 },
      {  -1,  -3,  -8, -32,   3,   7,  14,  56,  -3,  -9, -15, -63,   5,  15,  20,  84 },
    },
  },
  { // 16
    {
      {-109,  30,  12,   5,  55,  -8,  -8,  -2,  16,  -9,  -2,   0,  -2,   0,   1,   0 },
      {  45,   9, -10,  -1,  95, -28, -13,  -2, -61,  -4,  10,   2, -16,  10,   3,   0 },
      {  37,  -5,  -4,  -2,  43,  13, -12,  -1, 101, -22, -14,  -2, -42, -12,   9,   1 },
      { -19,-105,  28,  12,  23,   4,  13,  -5,  -4,  53, -14,  -8, -19,   2,  -3,   2 },
      { -10,  11,  22,  -3, -37, -42,  10,   5, -14, -11,  13,   1,-104,  29,  15,   0 },
      {   8,  -2,  -4,   5,  -1,-104,  25,   7,  41,  20,  11,  -4,  40,  26, -18,  -4 },
      { -21, -30,-106,  11,  -1,   7,  14,  20,   2,  -9,  51,  -7, -16,   0,   2, -10 },
      {  -5, -53,  27,  -1,  -1,  -9,   5,   1,  -7,-107,  14,   9,  25,  17,   9,  -4 },
      {  -4, -17,   0,  10, -17, -24,-114,  14,   3,  13,  20,  10,   8,  -6,  33,  -6 },
      {  11,  17,  21,  94,   3,  28,  -2,   2,   7,   3,  28, -46,   7,  53,   1,  -7 },
      {  -4,  -3, -21, -47,   2,  25, -22,  -5,   7,   8, -28,  15,   5, 105, -18,  -7 },
      {   5,   5,  45, -46,  11,  25,   8,  23,  15,  21,  97,  12,   5,   2, -16, -18 },
      {   2,   4,   8,  13,   6,   6,  12, 115,  -5,   1, -35,  23,   1,   1,   7, -29 },
      {   1,   5,  -1,  -7,   5,   4,  34, -15,   6,  17,   5,  19,  19,  17, 116,  -9 },
      {   2,   1,  -2,  50,   0,   4,   6, -25,   4,   6,   9, 112,  -6,   3, -21,  -2 },
      {   0,  -1,   0,   0,   2,   4,   2,  31,   3,   2,  13,   9,   5,  12,   7, 122 },
    },
    {
      {  75, -47,   1,  -3, -75,  43,   3,   1,  29,  -8,  -8,   2,  -5,  -3,   4,   0 },
      { -43, -51,  51,  -2,  31,  66, -51,  -1,   3, -34,  11,   5, -10,   6,   6,  -4 },
      {  43, -39,   9,   1,  28,   5,  -6,  -5, -91,  39,   2,   3,  45, -17,  -5,   1 },
      { -44, -36, -16,  30,   4,  24,  52, -35,  14,  37, -63,  15,   0, -33,  21,   6 },
      {  35, -18, -54,  24,  54,  16,  36, -24, -15, -54,  11,   6, -43,  34,  -5,   0 },
      {  40, -21,  34, -19,  46, -39, -18,  14,  27,  42, -25,  -3, -71, -10,  21,   0 },
      { -39, -10,  -2,  -4, -55,   2,  11,   6, -64,  23,  20, -10, -78,  17,  -3,   5 },
      { -15, -48, -29,  -3,  16,   1,  16,  29,  35,  32,  46, -56,   8, -12, -56,  28 },
      {  -1, -28,  43,  48, -15, -48,  -2, -60,  16,  12,  18,  22,   9,  50, -38,  -1 },
      { -13, -33,  21, -63,   0, -20,  48,  39,  -7, -23, -44,   9,  21,  56, -16,  -5 },
      { -16, -52, -10,  -6, -13, -59,  11,   7,   0, -41,  43,  26,   5, -53,  41, -26 },
      { -11, -18, -60,   3,   1,   7, -41,  23,  15,  49,   6,  35,  11,  50,  20, -57 },
      {   3,  18,   4, -53,   9,  26,  22, -23,  10,  19,  30,  73, -16, -33, -56, -20 },
      {  -8, -18, -43,  -2, -12, -28, -64,   1, -11, -27, -54,  25, -11, -22, -52,  38 },
      {  -5,  -8, -25, -69,  -2,  -3, -17, -69,   6,  11,  18, -11,  11,  22,  44,  53 },
      {  -2,  -3,  -6, -28,  -3,  -6, -12, -50,  -5, -12, -24, -65,  -5, -15, -25, -84 },
    },
    {
      {  73, -87,  49, -19,   0,   2,  -5,   4, -14,  19, -11,   4,  -5,   5,  -2,   0 },
      {   6,   5,  -4,  -1, -72,  89, -48,  15,  -2,  -1,   5,  -4,  14, -18,  10,  -3 },
      { -82, -25,  72, -32, -31,  -2,  22,  -8,  27,   3, -13,   5,  17,   0,  -8,   2 },
      {  31,  -9,   3,  -8,  -7,  -8,   9,  -7,  73, -85,  43, -10,  15, -15,   1,   2 },
      {  11,  -1, -10,   8, -76, -30,  59, -23, -61, -11,  37, -11,   7,   6,  -3,   3 },
      {   1,  36,  44, -51,  25,  24,  -5, -11, -45, -38,  25,   0, -69,  -5,  13,   3 },
      {  44,  55,   9, -51,   6,  18,  31, -47,  11,  24, -31,  14,  53, -10, -17,  12 },
      { -17, -14,  10,  -3,  42,  -6, -20,   9, -54, -19,  28,  -2,  78, -51,  27,  -6 },
      {   1, -10,  -6,  14,  -3,   5,  36, -26,  24,  31,  -9, -10, -36, -85,  60, -13 },
      {  13,  22,  -1, -26, -40, -60, -26,  47,  -5, -24, -49,  53,  -5, -29,  19,  -3 },
      {  14,  31,  18, -35, -12, -34, -19,  34,  24,  54,  49, -49,  10,  21,  34, -31 },
      { -22, -40, -64, -68,  -8, -18, -42, -57,   0,   3,   4,  -4,  -8,   6,  10,   6 },
      {   5,  14,  24,  19, -12, -32, -46, -23,  -7,  -3, -23, -72, -10, -43, -53,  28 },
      {   5,  10,  22,  30,  -5,  -5, -19, -50,  -8, -32, -49, -23,  13,  45,  57, -51 },
      {   5,  14,  32,  40,  -8, -19, -42, -46,  11,  25,  44,  58,   0,   8,  23,  50 },
      {  -1,  -3,  -6,  -8,   2,   6,  16,  26,  -2,  -9, -26, -39,   9,  24,  57,  95 },
    },
  },
  { // 17
    {
      {-107,  22,  16,   6,  60,  -9,  -9,  -4,  15,  -5,  -3,   0,  -4,   0,   1,   0 },
      {  44,   1, -10,  -4,  92, -21,  -8,  -3, -68,   0,  10,   3, -23,   9,   2,   1 },
      {   3, 105, -21, -18, -26, -49,  12,  10, -18, -24,   9,   4,  17,   5,  -4,  -1 },
      { -40, -23,  -1,   9, -31,  10,   9,  -3, -92,  23,  13,   1,  61,   0, -10,  -1 },
      { -22, -28, -99,  -5,   2, -20,  51,  12,  15,  22,  28,  -5, -22,   9,  -7,  -2 },
      {  25,   0, -23,   1,  47,  10,   1,  -2,  43,  -4,   1,  -1, 103, -13, -13,  -2 },
      {  -4,  29, -34,   4,   4,  97,   9, -17, -17, -55,  14,   7, -17, -26,   1,   5 },
      {   6,   1,  -8, 109,  -7, -25,  -4, -44,   3, -27,  12, -26,  -1,  17,   1,   5 },
      {  10,  43,   9,  27,  11,  24,  44, -18,   0,  85, -37, -20, -12, -38, -11,   8 },
      {  -1,  19, -32,   6, -10,   8, -94,  -2,   6,  51,  47,  -1,  -4, -24,  20,   1 },
      {  -3, -23,  -3,  -2,  -5, -45,   5, -16,  -7, -28,  -7,  18,  -4,-109,  14,  13 },
      {   7,   3,  51,   0,   5,   1,  34,   3,  13,   7, 102,   4,  -7, -15, -41,  -4 },
      {  -1,  -1,  -4, -36,  -5,  -8,   1,-107,   7,  13,   3,  47,   1,  25,  -9,  20 },
      {  -3,  -5, -14, -11,  -5,  -6, -36,  -3,  -5, -11, -25, -31, -13, -10,-112,   8 },
      {  -2,  -5,   6, -42,   0,  -2,   6, -30,  -3, -10,  15,-107,   6,  -1,  32,  29 },
      {   0,   0,   1,  10,   1,   2,   0,  31,   2,   2,   3,  20,   3,  11,  -2, 122 },
    },
    {
      {  92, -31, -15,  -3,  75, -25, -13,  -2,  -2,   1,  -1,   0, -19,   5,   3,   1 },
      {  14,  84, -41,  -9,  26,  69, -37,  -9,  15,  -2,  -3,  -1,  -2, -17,   7,   1 },
      { -40,  -2,   7,   3,  50, -27,  -5,   1,  99, -29, -16,  -2,  24,  -3,  -7,  -1 },
      { -17, -28, -80,  32, -26, -30, -70,  29,  -6, -12,  -3,   5,   8,   0,  15,  -4 },
      {  55, -23,   1,   0, -47,  28,  -4,   0,  40,  37, -24,  -5,  76,   2, -19,  -4 },
      { -41, -26,   6,   8,  35,  27, -26,   0,  -2,  90, -34, -11, -25,  35,  -4,  -6 },
      {  18,  14,  23,  84,   0,  22,  22,  79,  17,   2,   6,  12, -23,  -3,   1, -12 },
      {  21, -20,  -5, -20, -54,  13,  -6, -18,  61,  -4,   0,  -7, -87,   7,  16,   1 },
      {  -8, -11, -38,  13,  10,   9,  36, -15,  20,  30,  94, -35,  17,  14,  30, -10 },
      {  14,  59, -23,   2, -15, -49,  18,   0,   6,   9, -10,  -5,  -5,  87, -37,  -6 },
      {   5,   4,  18,  37,  -3,  -2, -15, -30, -13, -25, -24,-101,  -3,  -9,  -4, -43 },
      {  -6,  -5, -58,  19,   3,  -7,  55, -17,   1,  20, -25,  -6, -21, -58, -62,  20 },
      {  10,  46,  15,  -6, -12, -66,  -5,   6,  10,  59,  -8,  -7,   0, -53,  53,  -2 },
      {  -3,  -1,   1, -65,   1,  -4, -15,  64,   2,   9,  29, -22,  -9, -22, -52, -56 },
      {   4,  10,  38,  27,  -6, -14, -59, -21,   6,  14,  55,  -4,  -7, -12, -59,  54 },
      {   1,   2,   6,  29,  -2,  -6, -11, -54,   3,   7,  14,  63,  -4, -11, -17, -87 },
    },
    {
      {  71, -62,  14,  -4, -65,  51,  -5,  -1,  15,  -4,  -9,   3,  -1,  -4,   3,   0 },
      { -47, -31,  68, -18,  34,  44, -65,  12,   0, -22,  11,   7,  -4,   4,   5,  -6 },
      {  33, -43,  26,  -7,  43, -27,   4,  -4, -74,  61, -14,   2,  22, -11,  -5,   4 },
      { -51, -33,  16,  20,   0,   8,  44, -40,  38,  32, -69,  20, -17, -14,  17,   6 },
      { -31,  -9, -42,  49,  -2,  60,   5, -34, -46, -17,  20,   3,  56, -22,  -3,   2 },
      {  27, -41, -15,  32,  59,  -2,  22, -34,   6, -45,  20,   2, -46,  50, -18,   2 },
      { -49, -17,   8,  -2, -58,   0,  18,   8, -48,  20,  23, -33, -48,  40, -36,  19 },
      { -15, -43,  11,  -2,   9, -25,  22,  14,  57,   5,  33, -47,  62,  -7, -43,  22 },
      {   0, -16, -53,  -1,  37,  40,  -8,  49,  11,  33,  -1, -49, -47, -40,  18,  15 },
      {  13,  15,  22,  70, -14, -24, -54, -43,  22,  40,  36, -14, -26, -28,   2,  10 },
      {  21,  51,   7,  -9,  21,  56,  -9, -18,  16,  41, -27,  -3,  11,  40, -61,  32 },
      {  16,  29,  47, -23,   7,  11,  40, -32, -20, -43,   5, -36, -20, -60,   1,  47 },
      {   8,   9,  19,  52,  -3, -10, -14,  29, -14, -28, -59, -61,  24,  40,  35,  21 },
      {  -6, -22, -33,  -1, -10, -33, -49,   8,  -9, -34, -43,  33,  -8, -29, -50,  60 },
      {   7,  13,  35,  59,   5,  10,  33,  64,  -4,  -9,  -6,  29, -14, -32, -52, -36 },
      {   1,   3,  12,  18,   2,   8,  23,  38,   6,  15,  38,  55,   7,  21,  45,  81 },
    },
  },
  { // 18
    {
      { 119, -18, -12,  -4, -31,   3,   3,   2, -27,   4,   3,   1,  -1,   0,   0,   0 },
      {  20,  -2,  -4,  -1, 115, -15, -12,  -4, -43,   4,   3,   2, -23,   3,   3,   1 },
      { -35,  -7,   0,   1, -29,   4,   3,   1,-112,  20,  12,   3,  36,  -3,  -3,  -1 },
      { -15,-114,  -4,  11,   3,  -9,   6,   4,  17,  49,   0,  -5,  -9,   7,  -1,  -1 },
      {  14,  -8,  -4,   0,  32,   0,  -3,  -1,  26,  -2,  -3,  -1, 119, -12,  -9,  -3 },
      {   0,  10,  -8,   2, -15,-121,   1,   9,   2,  -2,  10,   2,   7,  32,   0,  -2 },
      {  12,  -1, 117,  16,   0, -10,   5,  -3,  -6,   8, -44,  -6,   3,   2,  -4,  -1 },
      {   7,  51,  -4,  -1,   1,   2,  -3,   0,  15, 115,   9, -11,   0,  -8,  -4,  -2 },
      {   1,   7,  -7,  -1,  13,   1, 122,   9,  -1,  -2,  -4,  -9,  -3,   3, -31,  -4 },
      {   4,   9, -13, 119,   0,   4,  -4,  11,  -2,  -9,  13, -40,   0,  -3,   1,  -5 },
      {   4,  -5,  44,  -5,   2,   2,   9, -19,  10,  -7, 117,   4,   0, -10,   0,  -1 },
      {   1,   5,   3,  -1,   3,  33,  -3,  -7,   1,   4,   9,  -7,  12, 122,   2, -10 },
      {   1,   2,  12,  -5,   5,  10,  -7, 120,   2,   0,  14,  26,   0,   1,   3, -28 },
      {  -1,  -4,   3, -42,   0,  -3, -15,  20,  -4, -11,   5,-115,  -2,  -6, -20, -13 },
      {   0,  -1,  -2,   8,  -2,   1, -29,  -5,   0,  -3,   2,  22, -10,   2,-122,  -8 },
      {   0,   1,   3,   0,   1,   3,  -1,  29,   0,   2,   6,  -7,   3,  10, -11, 123 },
    },
    {
      {  93, -40, -13,  -1,  67, -27, -11,   0, -14,   8,   0,   0, -19,   8,   3,   0 },
      {  22,  83, -51,  -6,  30,  55, -39,  -5,  11, -17,   4,   1,  -5, -17,   9,   2 },
      {  40,   3,  -6,  -2, -64,  37,   7,  -1, -88,  31,  15,   0,  -9,  -2,   4,   0 },
      {  27,  31,  76, -48,  21,  33,  55, -37,  10,   3, -13,   3,  -1,  -7, -16,   9 },
      { -58,  22,  10,   0,  55, -13,   4,  -3, -56,  20,   8,  -1, -71,  21,   9,   0 },
      {  13,  28,  -5,  -3, -13, -51,  38,   3, -24, -89,  48,  11,  -6, -15,   3,   5 },
      { -27,  -3, -12, -34,  38, -19, -16, -28, -55,   6,  -3,   5,  89, -20,  -8,   6 },
      {  -7, -23, -32, -81, -33, -18, -29, -60,  17,  -8,  -4,   5, -39,  12,  10,  13 },
      {   0,  -9,  38,  -2,  -7,   2, -50,  27, -26, -45, -73,  41, -20, -38,  -2,  11 },
      { -18, -56,  10,   2,  20,  50,  -1,  -9,   5, -14,  51,  -9, -12, -71,  42,   6 },
      {  -6,  -8, -22, -29,   8,  13,  24,  47,  12,  24,  26,  99,   3,   9,  -7,  27 },
      { -11, -24, -53,  26,   8,  21,  45, -27,  -7, -13, -32,  -6, -20, -26, -77,  27 },
      {   8,  33, -13,   5, -12, -56,  23,  -3,  13,  56, -21,   3, -12, -79,  30,  13 },
      {   5,   6,  18,  62,  -4,  -3, -17, -62,   2,   1,   9,  34,   9,  25,  25,  74 },
      {   4,  11,  32,  -8,  -8, -19, -55,  15,   8,  21,  56, -17, -11, -24, -77,  32 },
      {  -1,  -2,  -8, -29,   3,   6,  15,  53,  -4,  -8, -17, -57,   7,  16,  26,  88 },
    },
    {
      { -71,  70, -23,   4,  57, -50,   7,   3,  -8,   0,  10,  -5,   0,   3,  -3,   0 },
      {  52, -21, -19,   9,  23, -56,  49, -14, -53,  57, -17,  -4,   8,  -4,  -7,   7 },
      { -19, -38,  76, -32,  56,  -1, -49,  16, -39,  21,   1,   6,   1,   3,  -1,  -3 },
      {  59,   1,  -7, -18,  25, -38, -14,  32,  16, -43,  59, -28, -32,  38, -21,  -1 },
      {  17,  34, -31,   7, -21,  29, -47,  28, -65,  12,  43, -18,  47, -27,  -1,  -3 },
      {  11,  43,  24, -67, -50, -37,  -2,  51,  22,  35, -30,   0,   1, -12,   6,  -2 },
      {  47,  12,  -5,  -1,  53,  -9, -17,  -4,  39, -27, -18,  33,  44, -52,  46, -21 },
      {  31,  45, -15,   3,  11,  29, -28,  -9, -36,   4, -38,  45, -71,  30,  20, -12 },
      {   8,  25,  37, -22, -27, -19,  22, -42, -37, -54,   6,  49,  46,  32, -24,  -5 },
      { -26, -34,  -2,  14, -37, -48,   3,  16, -33, -29,  22,  13, -31, -21,  67, -45 },
      { -12, -17, -34, -69,  28,  46,  59,  33, -30, -36, -11,  -7,  -2, -10,   9,   0 },
      { -16, -36, -49,   6,   2, -10, -23,  45,  13,  12, -23,  43,  42,  62, -16, -34 },
      {  -9, -12, -29, -51,  -1,   5,  -6, -45,  21,  45,  67,  52, -18, -19, -11, -15 },
      {   9,  26,  31,   6,  10,  32,  38,  -7,  12,  35,  26, -32,  19,  46,  39, -71 },
      {  -4, -11, -31, -44,  -7, -16, -40, -57,  -2,  -4, -19, -48,  19,  39,  53,  27 },
      {  -1,  -4, -13, -16,  -3,  -9, -23, -32,  -6, -16, -36, -41, -12, -31, -59, -81 },
    },
  },
  { // 19
    {
      { 106, -26, -14,  -5, -55,  10,   7,   3, -28,   9,   5,   1,   7,  -2,  -1,   0 },
      { -36,   6,   5,   2, -97,  23,  11,   3,  61, -13,  -7,  -2,  31,  -7,  -5,  -1 },
      { -25,-103,  28,  17,   9,  48,  -7,  -7,  -6,  38, -10,  -4,   4,  -8,  -1,   0 },
      {  43, -19,   3,  -1,  30, -14,  -1,  -1, 102,   3, -20,  -3, -46,  13,   6,   1 },
      { -21, -21, -24,   7, -24, -84,  30,  13,  -2,  71,  10, -13,   0,  28,  -8,  -6 },
      { -21,   8,   9,   0, -44,  14,  -4,   0, -29,   3,  -2,   3,-112,   7,  18,   2 },
      { -14, -17, -99,   9,  14,  37,  47,  -3,   9, -18,  38,  -3, -15,  -7,  -8,   1 },
      {   7,  55, -15,  -9,   9,  47,  -1,  -6,   9,  91, -24, -13,   2, -40,   7,   0 },
      {  -9, -14, -25, -56,  -6,  -4, -82,  31,  15,  14,  57,  22,   3,  -8,  22,  -7 },
      {  -8, -14,   7, -92,   7,   4,  53,  37,  -7,  -5, -42,  35,  -2,  -5, -12,  -7 },
      {   0,  16,  -7,  -9,   2,  44, -12,  -2,  -2,  14,  -8,   1,  12, 115, -10, -15 },
      {   8,  13,  54, -10,   8,  18,  29,  25,  10,  10,  78, -31, -12,  -3, -61,  -8 },
      {   0,  -2, -15,  26,   4,   7, -18,  98,  -7, -16, -35, -59,  -1,  -4,   1, -22 },
      {  -2,   1, -22,  21,  -5,  -6, -40,   5,  -2,   4, -29,  47, -14, -11,-102,   8 },
      {   6,  11,  16,  54,   6,   9,  24,  45,   7,  13,  19,  86,   4,  -1,  31, -37 },
      {   0,   4,   5,  10,   2,   7,   7,  38,   1,   7,   7,  14,   5,  14,  11, 119 },
    },
    {
      {  90, -67,  14,  -6,  41, -25,   1,   0, -27,  22,  -4,   1, -10,   5,   1,  -1 },
      {  11,  -8,   3,  -1, -77,  53,  -8,   3, -74,  44,  -3,   1,   1,  -7,   3,   0 },
      { -45, -61,  70, -13, -39, -33,  40,  -5,  20,   8, -18,   5,  17,   7, -12,   2 },
      {  52, -15,  -3,  -4, -37,  35,  -9,   1,  55, -39,   7,  -1,  69, -39,   0,   1 },
      {   9,   5,  -6,   3, -25, -42,  45,  -8, -49, -68,  64,  -9,  -8, -18,  10,   1 },
      {  30,  -8,  -2,   0, -50,  15,   2,   1,  54, -17,   4,   0, -93,  31,   5,   1 },
      {  25,  20, -51,  26, -34, -42,  30,  -2,   6,  20, -17,   4,  36,  61, -50,   9 },
      {  31,  59,  42, -62,  16,  26,  49, -45,  -3,   4, -10,  11,  -2,   5, -23,  14 },
      {   8,  14,  13, -14, -12, -23, -36,  37, -28, -56, -63,  62,  -9,  -7, -20,  10 },
      { -15, -29,  16,  -9,  21,  50, -27,   2, -18, -45,  33,  -8,  16,  76, -41,   1 },
      { -25, -47, -65, -42,   1,   5,  -5, -55,   0,   0,   2,  31, -15, -29, -29,  40 },
      {   2,   9,  -8, -78, -25, -44, -51, -12,   2,   8,  12, -15,  19,  35,  48, -22 },
      {   1,   0,   6,  15,  -7, -10, -25, -53, -15, -30, -56, -86, -10, -14, -23,  -9 },
      {  -7, -20, -32,   5,  10,  30,  48, -24,  -9, -29, -49,  21,  11,  29,  63, -47 },
      {  -7, -10, -31, -54,   5,   9,  31,  63,   0,   1,  -8, -33, -11, -23, -45, -59 },
      {   3,   6,  14,  22,  -5, -10, -25, -48,   4,  10,  25,  46,  -5, -14, -37, -90 },
    },
    {
      {  76, -64,  11,  -1, -62,  48,  -1,  -3,   3,   4,  -9,   3,   4,  -5,   2,   1 },
      {  50,  33, -73,  12, -24, -46,  63,  -3, -14,  13,   3,  -9,   5,   3,  -9,   3 },
      { -20,  35, -25,   5, -47,  27,   3,   0,  79, -67,   9,   3, -16,   7,   6,  -4 },
      {  38,  40,   3, -47, -12,  -6, -51,  53, -37, -30,  53,  -3,  16,   5,  -3, -13 },
      {  38,  -8,  31, -40,  22, -48,   8,  20,  40,   1, -26,   8, -70,  41,   5,  -5 },
      {  24, -33, -28,  46,  55,  18,  -7, -27, -20, -47,  52,  -8, -38,  31,  -4,  -6 },
      {   5,  41,  -1,  -8,  -8,  31,   3, -32, -59, -23, -48,  64, -32,   7,  30, -14 },
      {  48,   9,  24,   1,  46, -23,   0, -33,  33, -41, -17,  33,  74, -16,   0,   4 },
      {  31,  19, -31,  20,  33,  15, -37,  19,  26,  28, -13, -16, -26, -67,  60,  -4 },
      { -20, -43, -42, -60,  38,  31,  49,  46,  -4, -27, -13,   9,  16, -14,  14, -13 },
      {  10,  18, -39,  10,  24,  43, -34,  21,  23,  44, -31,  15,  27,  69, -33, -11 },
      { -19, -37, -16,  26, -25, -49, -20,  12,   3,   3,   9,  34,  25,  27,  59, -64 },
      {  -3,  -4, -17, -38,   2,   5,  -1, -39,  27,  46,  63,  69, -16, -26, -23,  -9 },
      {  11,  33,  39, -14,  12,  39,  48, -25,   9,  27,  30, -42,  14,  25,  40, -54 },
      {   7,  10,  28,  46,   5,   6,  24,  45,   0,  -6,  -6,  21, -16, -39, -60, -69 },
      {  -3, -11, -29, -41,  -4, -15, -38, -56,  -4, -13, -34, -48,  -6, -18, -35, -61 },
    },
  },
  { // 20
    {
      { -64,  63, -21,   4,  64, -58,  14,   0, -12,   5,   5,  -3,  -3,   4,  -3,   1 },
      {  56, -11, -24,   8,  -1, -41,  41, -10, -64,  63, -19,  -2,  20, -13,  -5,   5 },
      { -27, -42,  60, -19,  54,  21, -49,  12, -53,  31,  -2,   4,   9, -11,   5,  -3 },
      {  64, -12,  18, -22,  43, -45, -14,  21,  19, -13,  19,  -9, -57,  50, -12,  -3 },
      {  25,  38, -17,  -8, -17,   7, -35,  25, -58, -40,  70, -15,  37,   2, -16,  -2 },
      {  -5,  45,  37, -49, -59, -36, -28,  46,  18,  43, -21,  -3,   1, -12,   9,  -4 },
      { -35,   6,  -3,   5, -46,  21,  11,  -7, -60,   8,   0,   4, -83,  42,  -6,  -1 },
      {  33,  38, -33,   6,  27,  36, -27,  10,  -3,   9, -10,   1, -52, -61,  56,  -9 },
      {  22,  32,  21,  -4,   2,  -6,  -9, -27, -25, -42, -52,  69,  24,  35,  31, -29 },
      { -13, -36,  27, -17, -18, -53,  33,  -3, -21, -58,  18,   4, -22, -62,  34,  -4 },
      {  -6, -11, -33, -64,  21,  34,  47,  64, -15, -28, -42, -12,   9,   9,  -7,   0 },
      { -22, -46, -51,  19, -12, -25, -24,  36,   5,  15,  15,   4,  15,  35,  60, -52 },
      {  -6, -14, -36, -43,   1,   0,  -7,  -7,  13,  21,  35,  89, -13, -25, -46, -16 },
      {  -5, -22, -34,  33, -10, -33, -61,  23, -11, -31, -57,   6, -10, -18, -42,  41 },
      {   6,  11,  35,  66,   5,   8,  29,  62,   1,  -2,  -2,  18,  -7, -20, -40, -59 },
      {   1,   2,  14,  28,   3,   6,  22,  48,   3,   8,  26,  53,   6,  17,  39,  85 },
    },
    {
      { -91,  20,  17,   5,  80, -21, -14,  -4,  -2,   7,  -1,  -1, -10,   1,   3,   0 },
      {  33, -14,   3,   2,  41,  12, -20,  -5,-106,   8,  29,   5,  25,  -9,  -5,   1 },
      {  23,  83, -32, -20,  -4, -79,  22,  17, -19,   6,   5,   2,   1,  14,  -4,  -4 },
      { -63,  22,  12,   3, -65,   8,  -1,   2, -25, -29,  15,   1,  74,  -5, -13,  -1 },
      { -10,  -2, -71,   0,  29,  37,  59,  -9,  -6, -68,   3,  17,  -5,  14, -13,  -5 },
      {  42,   4,  23,   1,  44, -18, -38,  -7,  50, -52,  15,   6,  64,  21, -15,  -3 },
      {  -3, -40, -40,   2,  22, -18,  40,  11,  19,  65, -18, -14,  69, -19, -18,  -2 },
      {  -5,   8, -32, -24,   1,  10, -23,  18,  28,   6,  90, -17, -14, -61, -17,  17 },
      {  20,  58,   9,  28,  15,  30,  11, -36,   5,  -4, -41,  -9,  16, -79,  22,  24 },
      {  -2,   1, -25,  81, -14, -21,  -8, -74,   6,  22,  39,   7,  -6,  26,  -8,   8 },
      {   0, -48,  37,  16,  -5, -67,  41,  -4, -13, -47,   1, -14, -19, -45, -27,  24 },
      {  -7, -24, -46,  22,  -5, -25, -34,  24,  -7, -29,  -9, -47,  10,  -2,  86,  13 },
      {  -8, -22, -21, -22,  -9, -29, -24, -21,   3,  -3,  -8,  97,   3, -41,  30, -30 },
      {  -4,   1, -42,  29,  -7,  -4, -64,  28,  -9, -11, -52,  -3, -16, -16, -74,  -4 },
      {  -7, -11, -20, -70,  -6,  -8, -22, -63,  -5,   1, -22, -13,   3,  17, -12,  72 },
      {  -2,  -4,  -5, -31,  -3,  -7,  -7, -54,  -3,  -8,  -4, -59,  -5, -16,  -7, -92 },
    },
    {
      {  89, -32,  -9,  -2,  76, -25, -10,  -2,  -5,  14,  -6,  -1, -22,  11,   1,  -1 },
      {  36,  10,  -6,  -1, -43,  38,  -3,   2,-100,  27,  11,   2, -24,  -7,   7,  -1 },
      {  78,  -5,  -5,  -4, -69,  -2,  16,  -4,  35, -42,   7,  -3,  45, -11,  -8,   2 },
      {  19,  88, -26, -15,  37,  62, -11, -18,  25, -17,  14,  -2,  -8, -22,   7,   5 },
      { -11,   5,  22,  -6,  35, -35,  42,  -9, -49, -72,  40,   9,  12, -44,  -4,  12 },
      {   7,  19,   8,  -2, -31, -29,  37,  -7,  27, -17,  14,  -1,-104,  22,   9,   6 },
      {  10,  -2,  72, -12,  13,  34,  70, -24,  15,  52,  24, -13,  20,  13,   1,   2 },
      {  -2, -58,  19,  11,  -4,  40, -24,   6,  27,  -2,  17,  -7, -32, -82,  38,   5 },
      {   3, -22,  35,   3,   8,  57,   2,  17,  -3, -44, -49,  48, -21,  19, -59,  11 },
      {  13,  47,  52,  -7,  -9, -43, -12,   0,  -4,  22, -76,   3,  -2, -54,   4,   6 },
      { -11, -26,  -6, -92,  -9,  -2, -25, -75,  -5,  -5, -10,   1,  -4,   5,  -3,  23 },
      {  -5,  -8, -61,  -3,  -2,  -9,  48,   2,  12,  47,   0,  35,  -6, -50, -55,  31 },
      {   5,  11,  17, -44,  -3, -12, -19,  53,  10,  16,  37,  83,  10,  12,  37,  21 },
      {   6,  13,  37,  28,  -8, -18, -62, -16,   5,  17,  58, -11,  -6,  -1, -67,  37 },
      {  -1,   2,   6, -66,   1,  -1,  -3,  60,   3,   2,  16, -43, -11, -17, -47, -61 },
      {  -2,  -2,  -5, -18,   4,   8,   9,  52,  -4,  -7, -15, -60,   4,  13,  11,  95 },
    },
  },
  { // 21
    {
      { -68,  72, -32,   9,  50, -51,  15,   0,  -4,   0,   7,  -5,   1,   1,  -2,   0 },
      {  48, -26,   1,  -3,  35, -56,  34,  -9, -62,  60, -17,   2,   7,   0,  -5,   3 },
      { -51, -29,  77, -34,  49,   7, -49,  20, -10,  21, -10,   3,   0,  -1,   3,  -1 },
      { -35,  27, -16,   8, -52,  40, -15,   4, -49,  54, -25,   6,  47, -38,  11,   0 },
      {  31,  14, -20,  -2,  19,  13, -50,  35, -56, -20,  76, -30,  10,  -2, -11,  -2 },
      {  14, -16, -38,  46,  68,  45, -18, -18,  21,  19, -13,  -1, -19, -44,  42,  -1 },
      {  13,  37, -12, -10, -19,  11, -29,  25,  -8,  43, -22,   2, -91,  43,   8,  -7 },
      {  47,  36,   1, -34,   0, -37, -29,  48,  58,  19, -19,  -4,  37, -33,  22, -11 },
      {  27,  43,  20, -36,  15,  15,   7, -20, -44, -59, -28,  58, -11, -26,  26, -13 },
      { -22, -22,  14,  -5, -37, -39,  17,   7, -11,  -6,  34, -25, -50, -66,  56,  -5 },
      { -17, -38, -27,  34,  -9, -30, -23,  36, -17, -26, -20,  32,  21,  43,  49, -54 },
      { -15, -18, -38, -76,  18,  42,  56,  22,   9,  20,  27,   3,   8,  16,  22, -31 },
      {  15,  34,  43,  13,  -4,   3,  -3, -62,  11,  28,  43, -13,  23,  39,  55, -34 },
      {  -9, -19, -33, -24, -11, -26, -43, -40,  17,  26,  45,  80,  -6, -11, -16,  13 },
      {   6,  16,  39,  50,   4,  15,  39,  48,  11,  21,  36,  53, -10, -22, -33, -42 },
      {   2,   6,  12,  14,   2,   7,  21,  37,   0,   3,  18,  32,  12,  27,  52,  95 },
    },
    {
      { -88,  11,  14,   4,  85, -14, -15,  -3, -20,  12,   3,   1,  -8,  -2,   1,   0 },
      {  25, -65,   4,   8,  28,  66, -16,  -9, -65, -12,  24,   3,  24, -16, -10,   1 },
      {  40,  57, -18,  -5,  14, -53,  12,   3, -78,   6,   7,   2,  43,   4,  -8,  -2 },
      { -61,   9, -12,   4, -38,  18,  28,  -8,   2, -60,   0,  12,  73,  15, -21,  -7 },
      {  25, -19, -65,   2,  60,   4,  67,  -7,  23, -24, -31,  11, -14,  22,  -9,  -6 },
      {  33, -26,  32,   0,  42, -30, -34,  -1,  57,   7,  -6,  -5,  77,  19,  -7,   0 },
      {  26,  34,  40,   5,  19,   7, -32, -10,  -2, -84,   8,  15, -40,  48,  -9,  -3 },
      {  -8, -34,  42, -10, -13, -12,  14,  26, -39,  -4, -89,  -7,   0,  33,  39, -13 },
      {  -5, -19,  -8,  79, -20, -24,  -9, -75, -12,  27,  -5,  27,  -6,  37,   3,  -3 },
      {  16,  42,  22,  20,  15,  28,   5, -40,  12, -13, -49,  25,  15, -79,  31,  -1 },
      {   2,  50, -32,  -3,   7,  71, -27,  -2,   5,  36, -13, -27,  20,  56,  37,  -5 },
      {   1,   8,  17, -41,  -2,  18,   1,   2,   2,  40,  -4,  90,  -2,  14, -37, -53 },
      {   4,  21,  28,  43,   2,  25,  14,  23,  -1,  24, -37, -30,  -5,   1, -93,  23 },
      {   7,  10,  67,   2,   8,  16,  83, -16,   5,  20,  49, -13,   5,  20,  27,   3 },
      {  -3,  -4,   4, -68,  -5,   0,  -1, -59,  -6,   8, -19,   6,  -1,  13, -18,  85 },
      {   0,   3,  -4,  37,   3,   6,   2,  62,   3,   4,   8,  67,   8,  10,  28,  75 },
    },
    {
      { -85,  65, -17,   6, -49,  28,   1,  -1,  20, -25,  11,  -3,  13, -10,   0,   2 },
      { -43, -32,  39,  -8,  19, -69,  33,  -2,  59, -34, -15,  10,  17,   9, -19,   3 },
      { -26, -48,  48,  -9, -77,   3,  17,   6, -50,  38, -21,   8,  14,   2,  -7,  -2 },
      { -53,   9,  -6,   8,  42,  -3, -37,  20,   4,  70, -51,   8, -30,  35,   0, -12 },
      {  26, -25,  -3,   2, -22,  41, -44,  14,  73,  13, -34,   0,  62, -12,   3, -10 },
      { -36, -11,  37, -26,  58,  22,  -5, -18, -19,  23,  32, -25,  41, -59,  31,  -3 },
      { -25, -41, -42,  46,  -2, -39, -51,  44, -36, -36,   3,  -1,  13, -34,  11,  -4 },
      { -10, -39,  26,   9, -20,  22, -11,  13,  44,  -6,  36, -22, -79,  -3,  53, -13 },
      {  28,  53,  17,  -6, -16, -44,  21,  44,   9,  23, -28,  15,  -9, -61,  41, -27 },
      {   1,  -4, -47,  -2, -25, -52,  11, -34,  12,  45,  27, -60,  27,  28,  30, -27 },
      {   6,  13,  32,  59,  18,  18,  34,  59,  -6,   4,  37,  -1,  39,  50,  11, -27 },
      {   4,  19,  30, -59,  -2,  -8, -48,   7, -26, -51, -16, -18,   9,  44,  23, -52 },
      {  10,  16,  36,  53,   1,   3,  -4, -21,  -5, -13, -42, -84, -15, -22, -40, -14 },
      { -12, -27, -36,  15,  15,  34,  56, -31, -12, -27, -53,  17,   0,  -4,  40, -53 },
      {   7,  17,  30,  42,  -4, -15, -22, -45,  -8, -10, -24,  16,  16,  23,  71,  62 },
      {   4,   9,  23,  38,  -7, -15, -35, -58,   5,  12,  34,  58,  -4, -13, -23, -65 },
    },
  },
  { // 22
    {
      {  75, -73,  23,  -4, -51,  46,  -7,  -2,   2,   5,  -8,   3,   0,  -2,   2,   0 },
      {  49, -19, -18,  10,  35, -60,  41, -11, -56,  56, -12,  -6,   4,   1,  -7,   5 },
      { -28, -35,  77, -33,  57,   1, -49,  17, -29,  20,  -7,   7,  -3,   4,   3,  -4 },
      {  44, -15,   7, -15,  43, -42,   6,   8,  52, -56,  28,  -5, -42,  42, -11,  -4 },
      {  40,  30, -34,   1,  14,  24, -56,  30, -55, -14,  65, -21,   4,  -6,   1,  -4 },
      { -30, -54, -25,  66,  50,  44,   9, -46, -10, -18,  17, -10, -12,   3,  -7,   8 },
      { -33,   0,  17,  -5, -53,  -2,  20,  -4, -51,   4,  24, -11, -64,  59, -30,   7 },
      { -20, -29,  30, -10, -22, -35,  31,  -6, -18, -46,  52, -19,  69, -25, -20,   9 },
      { -24, -33, -17,  20, -12, -18, -21,  46,  50,  61,  26, -62,   7,  10, -11,   1 },
      {  11,  16,  -7,  -7,  15,  27,  -9,  -6,   7,   8, -21,  10,  63,  72, -69,  16 },
      {  21,  39,  38, -21,  21,  37,  30, -27,  20,  25,  19, -44, -24, -42, -34,  47 },
      {  21,  31,  53,  76, -22, -38, -49, -43,   8,   5,   3,  -1,   5,   6,  -1,   0 },
      {  -2, -11,  -9,  25,  -5, -18, -21,  49, -18, -43, -59, -15, -20, -35, -46,  51 },
      {  -9, -20, -29, -20, -10, -22, -34, -25,  18,  26,  39,  75, -18, -32, -44,  28 },
      {   6,  13,  29,  46,  10,  20,  42,  60,   5,  11,  20,  45,  -6, -19, -40, -52 },
      {  -2,  -4, -15, -24,  -4,  -7, -21, -39,  -3,  -8, -24, -37, -12, -28, -55, -87 },
    },
    {
      {  89,  -7, -16,  -4, -85,   7,  18,   4,  21,  -5,  -5,  -1,   6,   1,  -1,   0 },
      { -21,  65,  -2,  -4, -16, -75,  13,   6,  63,  19, -24,  -3, -22,  13,  10,  -1 },
      { -51, -46,   3,  10, -26,  55,   5,  -7,  81, -18, -10,  -1, -30,  -4,   4,   3 },
      {  41, -15,  36,   3,  20,  -2, -56,  -5,  18,  61,  22,  -6, -59, -37,  16,   9 },
      {   5, -42, -59,   3,  30,  -6,  66,   1,  -4,  64, -35,  -6,  -8, -17,   5,   1 },
      { -48, -17,  19,   0, -59, -15, -15,  10, -10,  60,  19, -19,  62, -31, -12,   8 },
      { -24,  47, -36, -20, -28,  23,  25,   8, -28,  -7,  44,  10, -54, -60, -20,   6 },
      {  20,   6, -33,  34,  35, -12,  -1, -39,  52,  -7,  58,  12,  43, -22, -47,   9 },
      {  -7,  16,  15,  75, -24,  -2,   8, -77, -29,  -2, -33,  18,  -9, -21,  27,  11 },
      {  18,  32,  27, -14,  28,  27,   9,  22,  21, -21, -51, -21,  41, -78,  13,  15 },
      {   0, -56,  47,  -5,  -3, -65,  34,   5,  -9, -43,  -2,   8, -25, -34, -42,  10 },
      {  -2, -13, -24,  21,   1, -23,   3,   7,  -4, -33,  36, -85,  -4,  -1,  58,  40 },
      {   7,  26,  32,  27,   3,  30,  23,   0,  -6,  21, -12, -65, -19,  34, -74,  26 },
      {   6,  11,  61, -21,  10,  18,  75, -20,   9,  18,  53,  11,  12,  16,  42,  -9 },
      {   3,   2,   7,  63,   2,   0,   9,  45,   0,  -4,  13, -19,  -4, -16,   3, -97 },
      {   2,   4,   7,  57,   2,   6,  10,  75,   1,   6,  10,  55,   2,  11,  10,  62 },
    },
    {
      { -92,  46,   5,   0, -65,  21,  10,  -1,  16, -20,   6,   0,  17,  -9,  -2,   1 },
      {  40,  27, -21,   0, -45,  59,  -7,  -2, -79,  17,  26,  -4, -16, -19,  15,   1 },
      {  -7, -79,  40,   6, -66, -36,  23,  10, -38,  21, -10,   1,  11,   9,  -6,  -3 },
      { -62,   3,   7,   7,  39,   4, -19,   6, -32,  74, -14,  -5, -52,  28,  17,  -6 },
      {  18, -17,  -9,  -1, -26,  54, -43,   3,  51,  58, -41,  -9,  47,  12,  -1,  -6 },
      {  22,  24, -51,  18, -50, -36, -15,  16,  15, -12, -28,   4, -52,  63, -14, -10 },
      { -29, -19, -59,  31,  15, -31, -62,  22, -42, -21,  -3,  -6,  49, -21,  -1,  -8 },
      {  -5, -42, -20,  28, -16,   4, -11,   7,  49,   2,  67, -34, -38, -21,  56,  -4 },
      {   4,  23, -30,  14,   7, -23,  51, -17,  -4,  44,  53, -30,  57,  50,  -6,  -3 },
      {  11,  17,  31,  86,  13,  23,  28,  75,   3,  -3,  -6,   8,   6,  -3,  -9, -16 },
      {  21,  53,  20, -15, -17, -65,  -6,  21,   7,  37, -20,  -5,  13, -46,  57,  -8 },
      {   4,   6,  46, -19,   6,  13, -29,   7, -17, -46,   1, -39,  20,  62,  51, -45 },
      {   3,   7,   8,  30,  -1,  -5,   7, -42,  -4,  -4, -36, -95, -18, -33, -37, -23 },
      {   9,  22,  52,   0, -10, -20, -66,  14,   5,  12,  53, -16,  -4,   9, -55,  42 },
      {   5,   7,  17,  63,  -2,  -3, -10, -60,  -6, -11, -23,  14,  10,  19,  44,  71 },
      {   5,   8,  19,  34,  -6, -11, -24, -59,   6,  11,  24,  56,  -1,  -9, -13, -80 },
    },
  },
  { // 23
    {
      { -89,  31,  14,   2,  77, -32, -11,  -1, -10,  11,  -1,   0,  -5,   0,   2,   0 },
      { -24, -77,  40,  12,  23,  73, -42, -10,   7, -12,  11,  -1,  -8,  -9,   3,   3 },
      { -51,   9,   3,   1, -45,   3,  15,   0,  97, -27, -22,   0, -22,  16,   4,  -1 },
      {  27, -18,  29, -10,   5, -32, -26,  20,  33,  83, -29, -22, -35, -22,  36,   5 },
      { -39,   8,  48, -16, -67, -13, -29,  25, -24,  16,  11, -14,  67, -18,  -8,   2 },
      { -36, -45, -45,  28, -10,  10,  64, -23,  -8,  50, -30,   1,  33, -30,  -1,   3 },
      { -20,  42, -24,   7, -32,  34,  -8, -12, -17,  14,  49, -10, -58, -64,  11,  25 },
      {  11, -13, -38,  -5,  25, -14, -17,   6,  57,  23,  75, -23,  38,  -1, -40,  19 },
      {  21,  24,  31,  81,  -3, -22, -26, -76,  16,  -1,  -5,   9,  22, -11,   0,  12 },
      {  24,  30,   8, -25,  38,  26,   9,  15,  31, -32, -37,   0,  43, -81,   7,   7 },
      {   2, -43,  40,  -6,  -3, -58,  35,   2,  -5, -28,  13,  19, -37, -46, -57,  16 },
      {   0,  -8,  -3,  36,   3, -10,   9,  25, -15, -35, -19,-100,  -4,   9,   6,  47 },
      {  10,  41,  48,   2,  10,  52,  40,   1,   4,  44, -10, -17,  -9,  26, -69,   7 },
      {   4,  -2,  41, -39,   7,   0,  56, -36,   4,  -1,  42,   0,  16,  20,  64,  46 },
      {   6,   3,  26,  44,   7,   0,  39,  31,   8,  -5,  45, -23,   1, -15,  28, -87 },
      {   3,   5,   5,  56,   4,   8,   8,  76,   3,   6,   9,  61,   6,  10,  16,  54 },
    },
    {
      { 112, -15, -10,  -2, -39,  32,   2,   0, -26,  -7,  12,   0,   3,  -8,  -1,   2 },
      {  13, -32,   4,  -2,  92,  25, -26,  -1, -44,  51,  13,  -5, -15, -15,  17,   1 },
      { -50,   9,   0,   3, -25,  63,   1,  -6, -57, -33,  48,  -2,  20, -44,  -7,  11 },
      { -15,-106,   2,   7,  -7,  32, -14,  -4,  46, -17,   0,   0,  25,  13, -17,  -1 },
      {  18,   8, -15,   3,  66,  -8,  31,  -9, -12, -76,   1,  28,  17,  23, -52,   2 },
      {  -7, -36,  27,  -1, -21, -63, -11,  18, -76,   1, -34,   4,  45,  -3, -14, -17 },
      {  16, -12,  48,  -2,   9, -50, -10,   6,  30, -24,  36,  -2, -22, -79, -10,  38 },
      { -16, -36, -10,   4, -17,   4,  54,   4, -27,  -4, -31,  34, -92, -14,  -5,  -4 },
      {  -3, -17, -80,  -1,   1, -33,  55, -14,   7,  39,  23,   5,  40, -36,   5,  16 },
      {   7,  12,  54, -10,  16,  40,  64,  26,  22,  16, -45,  15,  44, -34,  15, -13 },
      {  -1,  11,  28,   7, -15,   8,   4,   4,   1,  64,  28,  43,   2,  25, -83,  32 },
      {   1,  18, -45,  19,   2,  17, -51,  17,  13,   4, -61,   0,  -1, -59, -53, -17 },
      {   5,   4,  24,  92,  -1,  -3,   9, -83,  -4,   6, -15,  -4,   6,  -8,   6,  -4 },
      {   1,  -2,   0,  57,   6,  -2,  34,  64,  -2,   3,  21, -79, -12,  11, -27, -14 },
      {  -3,  -2,  -3, -16,   1,  11,   7, -12, -14,  -3, -51, -48,   2,  13,  -3, 103 },
      {   0,  -2,  14, -62,  -1,  -1,  19, -61,   0,  11,   0, -61, -11, -11, -46, -45 },
    },
    {
      { -85,  65, -10,   4, -55,  31,   6,  -2,   8, -20,  12,  -2,   8,  -9,   0,   1 },
      { -43, -25,  43,  -4,  13, -72,  42,  -1,  46, -44, -11,  11,  10,   7, -20,   7 },
      {  19,  54, -50,   4,  69,   1, -21,  -6,  64, -33,   8,  -4,  11,  -1,  -1,   2 },
      { -45, -10,  -8,  23,  13,  -4, -54,  33,  20,  48, -70,   8,   1,  38, -14, -13 },
      { -30,  41,   0, -16,  53, -23,  34, -26, -44,  37,  10,   5, -62,  32,   3,   6 },
      {   2,   9,  62, -48,  17,  51,  18, -37,  44,  44, -14, -17,  31,   3,  10,  -7 },
      {  46,  30, -18, -17, -40,   6,  32, -11,   4,  -3, -36,  42,   6,  50, -63,  22 },
      {  11, -10, -28,  23, -39, -23,  23, -13,  39,  25,  32, -56,  -5,  65,  27, -31 },
      { -19,  -4, -33,  11,  23, -21,  17, -21, -40,  33,  16, -13,  94,  -6, -31,   4 },
      {  -9, -15, -40, -69, -25, -36, -41, -65,  -4, -11, -32, -17, -11, -18,   9,  -5 },
      { -31, -57,  -7, -10,  19,  52, -26, -22,   1, -28,  46,  -3, -15,  45, -46,  23 },
      {  -8, -23, -31,  16, -15,  -5,  15, -10,  49,  60,  19,  25, -25, -48,  -5,  64 },
      {   7,   5,  12,  51,   3,  12,  14, -31,  -5,  -9, -40, -73, -30, -35, -56,  -1 },
      {  16,  38,  43, -22, -16, -41, -59,  26,   1,  19,  42, -36,   2,   6, -36,  39 },
      {  -4,  -7, -21, -44,   1,   1,  18,  43,  15,  21,  25,   1, -20, -33, -56, -76 },
      {   8,  12,  30,  47, -10, -15, -37, -64,   8,  11,  29,  57,  -4,  -7, -16, -53 },
    },
  },
  { // 24
    {
      {-100,  49,   0,   3, -51,   3,  14,  -1,  17, -26,   4,   0,  11,  -5,  -3,   1 },
      {  39,  18, -14,   0, -69,  71,  -4,  -3, -53,   2,  27,  -4,   6, -29,   9,   3 },
      {  11, -77,  38,   1, -66, -41,  20,  11,   2, -34,  -7,   6,  30,  -9, -13,   0 },
      { -46, -40,  13,   4, -19,  11, -43,  18, -29,  82, -37,  -8,  -7,  14,  20, -16 },
      { -27,  -4,  33,  -4,  28, -34,  42,  -7, -79,   1,  40,   0, -39, -32,  18,   7 },
      { -25, -35,  19,  -2,  42,  36, -36,   1,  28, -11,  33,  -9,  45, -72,  21,  11 },
      {  11,  12,  69, -35,  -4,  32,  58, -28,  41,  48,  12,  -9,  12,  19,  13,   2 },
      {  -2,  24,  16, -12,  26,   7,  13,  -5, -46,   9, -53,  26,  66, -17, -64,  10 },
      { -26, -44,  14,  10,  16,  58,  -6, -12,  -7, -24,  28,  10, -36,  48, -65,  19 },
      {  10,  16,  43, -13,  -9,  15, -17,  28,  23, -14, -51,  42, -69, -47,  -4,  24 },
      {  -4,  -7,  -8, -76, -11, -15, -37, -76,  -8, -16, -20, -40, -18,  -9, -10,   2 },
      {   2,  -6, -34,  30, -20, -28,  12, -26,  28,  60,  21,  -4, -15, -42, -48,  56 },
      {  16,  39,  54,  12,  -6, -23, -45,  41,  -4,   0,  25, -72,   1,   7, -38,   6 },
      {   2,  13,   4, -51, -10, -27, -43,  17,  -5,  14,  50,  67,  21,  32,  11,  44 },
      {  -4,  -8, -29, -59,   0,  -1,  18,  49,  14,  19,  28,   5, -14, -27, -47, -69 },
      {  -9, -16, -33, -42,   8,  15,  36,  58,  -5,  -8, -24, -48,   3,  10,  16,  70 },
    },
    {
      { -73,  78, -29,   6,  41, -41,  18,  -4,  10, -11,   3,   0,  -4,   3,  -2,   1 },
      {  51, -13,  -8,  -1,  72, -50,  21,  -7, -57,  41,  -8,   1, -18,   6,  -3,   2 },
      { -62, -48,  67, -19,  22, -26, -15,  13,  22,  37, -41,   6,   4,   8,   3,  -6 },
      { -40,  10, -15,  21, -51,  11,   4,  -1, -77,  59, -13,  -8,  29, -27,   9,  -1 },
      {  -9,  14,   0,  -5,  59,  60, -72,  20, -11,  17,  29, -19, -12, -29,  29,  -3 },
      { -17,  11,  43, -46, -16,  -3,  -5,  21, -70, -52,  25,  27, -42,  15, -13,  -1 },
      { -27, -42, -33,  44, -17, -20, -21,  10,  -3,   7,  37, -42, -53,  60,   1,  -9 },
      {  30,  57,  -5, -23, -25,   9, -29,  26,  11,  38, -54,   9, -24,  60,  11, -24 },
      {  16,  -2, -11, -11,  -8, -60, -43,  57,  -7, -23,  18, -19,  73,   2,   8, -23 },
      {  12,   7,  10, -22, -30, -39,  18,   4,  21,  14,  22, -10, -45, -48,  85,  -9 },
      {  -9, -30, -46,  32,   5,  -8, -30,   6, -13, -44, -61,  59, -21, -23,  30, -13 },
      { -10, -10,   4,  -2,  19,  28,  33, -30, -13, -10,  22,  24,  41,  53,  63, -57 },
      {  15,  25,  59,  64,   6,   2,   7,  -3, -17, -40, -40, -61,  -4,  -6,  13, -21 },
      {   3,   8,   8, -13, -10, -26, -57, -84, -11, -12,  -6,  -8,  16,  23,  27,  55 },
      { -13, -22, -41, -70,   4,  10,   8, -25,  -9, -21, -38, -67,  -7, -13, -12, -39 },
      {  -7, -12, -15, -14,  11,  23,  37,  51,  -8, -17, -25, -25,  12,  28,  44,  82 },
    },
    {
      { -90,   3,  22,   5,  82,  -9, -20,  -4, -13,  14,   2,   0,  -7,  -5,   4,   1 },
      {  11, -79,   6,  16,  16,  84, -19, -19, -24, -20,  22,   5,   7,  -8,  -7,   3 },
      {  41,  31, -10,  -9,  35, -20,  -2,   3, -95,  -5,  23,   7,  44,  -3, -15,  -3 },
      { -45,   0, -64, -11,   0,  20,  78,   2,  -7, -44, -16,  20,  16,  17, -18, -14 },
      { -17,  32,  56,  11, -24,   3, -34, -14,  -1, -85,  -8,  20,  18,  41,  -2, -12 },
      {  54,  16, -10,   5,  64,   8,  10, -12,   6, -28, -33,  14, -73,  23,  26,  -7 },
      {  18, -54,   7,   0,  27, -30,  -2,  14,  18,  21, -49,  -2,  55,  67,  -1, -22 },
      {  10,  -3, -31,   4,  20, -26, -18, -15,  43,  -5,  83,  19, -10,  39, -54,  -3 },
      {  -5,   7, -17,  77, -17, -14,   6, -82, -12,  31, -10,  32,  10,   5,  22,  -8 },
      { -22,   0, -29, -23, -29,  15, -24,  16, -36,  22,  32,  -4, -24,  67,  68, -11 },
      {  -9, -36,  31, -15, -27, -28,   9,  15, -35,  17,  -8,  66, -51,  -9, -34, -48 },
      {   9,  30,  -8, -50,  16,  36, -23, -10,  37,  22,  -2,  71,  41, -25,  27, -30 },
      {   1,  51,  26,  13,  -2,  65,  16,   5,  -4,  53, -12, -16,  -8,  42, -56, -15 },
      { -10,   6, -67,   5, -11,   2, -80,   5, -10,  -8, -52, -13, -12, -12, -39, -17 },
      {   5,   8,  -3,  57,   9,   4,   4,  49,  13, -11,  26, -23,   9, -24,  23, -89 },
      {   0,   6,  -9,  57,   2,   8,  -8,  74,   2,   4,  -4,  60,   5,   6,   4,  61 },
    },
  },
  { // 25
    {
      {  73, -50,   4,   0, -75,  45,   1,  -2,  27,  -9,  -7,   2,  -5,  -3,   4,   0 },
      {  26, -57,  30,  -1,  42,  20, -30,   1, -75,  36,  12,  -2,  33, -20,   0,   2 },
      { -62, -34,  45,  -3,  16,  54, -50,   3,  38, -39,  11,   3, -18,   6,   9,  -5 },
      {  15,  31, -29,   5,  -3,  32,  -4,  -4, -42, -78,  60,   3,  28,  13, -30,  -2 },
      {  64,  12,  14, -30,  44, -25, -34,  29,  12,  -8,  23,  -8, -64,  38,   3,  -7 },
      {  -6,  45,  33, -39, -53, -19, -51,  51,   1,  17,  19, -16,  40, -28,   6,  -7 },
      {  -8, -17,  40, -37, -16, -27,  21,  15, -35, -40, -57,  32,  19,  56, -13, -23 },
      {   6,  49,   5, -19,   1,  49,   7,  -2, -42, -10, -38,  33, -48, -48,  44, -12 },
      {  31,  26, -26,   4,  44,  28, -27,  -3,  42,   1, -40,  13,  70,  22,  32, -26 },
      { -31,  -3, -50, -12, -23,  45,  -2,  33, -29,  48,   6, -25, -17,  64,  12, -28 },
      {   5,  37,  38,  47, -30,  -5, -40, -67, -16,  29,  11,  28,  -8,  47,   6,   0 },
      {  -8, -35, -32,  35, -16, -52, -23,   9, -16, -31,  16,  13,  -5, -11,  76, -42 },
      {   5,  12,  19,  19,  -2,   2,   3,  -1, -19, -30, -29, -86,   6,  21,  47,  57 },
      {   1,   2,  37, -40,   8,   9,  66, -21,  15,  13,  62,   5,  21,  16,  59,  -9 },
      {  11,  17,  46,  67,   8,  13,  36,  36,   3,   1,   0, -32,  -3,  -9, -19, -71 },
      {   1,   4,   9,  47,   3,   9,  15,  72,   2,   8,  11,  61,   6,  18,  16,  63 },
    },
    {
      { 106, -12, -10,  -1, -17,  49,  -3,  -4, -33,  -4,  27,   0,   1, -18,  -2,   7 },
      { -19, -34,   4,  -2,  84,  29, -33,  -2, -24,  59,  23, -11,  -8, -19,  24,   9 },
      { -60, -16,   7,   9, -27,  56,  -4, -10, -13, -46,  56,  10,  24, -38, -29,  12 },
      {  10, -15, -15,   3,  75,   9,  34, -13, -12, -55, -18,  34, -10,  30, -54, -16 },
      {  -5,  95,   5, -11,  25, -16,  36,   2, -50,  -4,  20,   4, -13, -38,  10,  17 },
      { -17,  18, -72,  -3,  -9,  55,  48, -32,  30,  25,   5,  11, -15,  28,  38,  -2 },
      {  23,  23,  15,  -3,  24, -15, -15, -22,  89,  -4,  57,  -3, -26,  -6, -24,  33 },
      {   2,  60,  40,  11,  14,  65, -44,   7,  20, -12, -27, -19,  18,  28,   4, -44 },
      {   9,  13, -44, -26,  23, -23, -10,  -5,  10,  -2,  12, -16, 109,   1,  -2,   2 },
      {  -8,  27, -57,  24, -19, -12, -60, -16, -25,  36,   3,   8, -22,   7, -68, -13 },
      {  -3,  -3, -40,  11,  10, -16, -44, -22, -15, -72,  -3, -55, -31,   4,  48,  16 },
      {   2,   3,  42,  45,  -7, -12,   2, -79, -30,  12,  10,  13,  25,  53,  13,  35 },
      {   9,  -1,  -8,  74,  11,   2,  19, -28,  27,   6, -51, -12,  15, -76,  -5,  -1 },
      {  -6,  -6,  11, -18,  -4,  13,  42, -11,  -9,  18,  -5,-102,  -7,  14, -50,  11 },
      {   4, -10,  10,  -3,  -2, -28,  14, -44,  -3,   1,  45, -12,  -7, -15,  13,-103 },
      {  -3,  -2,  12, -83,  -8,  10, -22, -68,   1,  -1, -49,  18,  -8, -34,  -7,   9 },
    },
    {
      { -87,  61, -11,   5, -58,  27,   8,  -1,   1, -20,  15,  -2,   5, -11,   2,   2 },
      { -45, -24,  35,  -5,  20, -76,  42,  -3,  40, -47,  -6,  13,   8,   4, -22,   8 },
      {  12, -55,  48, -11, -76,   1,  31,   2, -59,  -2,  14,   8,  -5, -18,   1,   8 },
      { -59, -14,  17,   0,  11, -17, -16,   7, -33,  73, -47,   5, -27,  46,  -3, -13 },
      {  -6,  37,  20, -20,  53,  -9,  48, -31, -41,  18,  57, -15, -35, -12,  33,   5 },
      { -46, -47,  15,   8,  35,   9, -51,  15,  -4,   4,  20, -19,  29, -71,  31,  -1 },
      {  -4,   0, -57,  43,   6, -44, -22,  32, -59, -37,  12,  19, -44,  -9, -16,  16 },
      {   4, -20, -14,  16, -41, -33,  -8,   3,  52,  10,  20, -38, -55,  11,  67, -21 },
      { -18, -35,  24,   1,  19,  63, -17,   4,  24, -26,  38,   5, -62,  36, -36,  25 },
      { -14, -33, -32,   6,   3,   0,  -1, -43, -31, -26,  25, -62,  41,  53, -14, -36 },
      {   1,   0,   8,  58,   3,   1,  27,  58,  10,  35,  56,  19,  49,  37,   9,  14 },
      {  -2,   2,  18, -11,  20,  26,   5,  34, -22, -58, -24,  37,   3,  32,  65, -52 },
      {   4,  17,   9, -67, -15, -36, -65,   4,  -7,  -5,  37,  18,  23,  39,  18,  41 },
      {  19,  44,  63,   9,  -2, -18, -29,  46,  -9, -11,   4, -58, -10,  -5, -39, -40 },
      {  -7, -15, -35, -52,  -4,  -6,  10,  26,  14,  25,  43,  35,  -8, -19, -38, -75 },
      {   7,  12,  34,  53,  -8, -13, -40, -70,   3,   3,  23,  57,   0,  -2,  -5, -40 },
    },
  },
  { // 26
    {
      {-108,  40,   8,   3, -35, -13,  17,   0,  26, -25,   1,   2,   8,  -3,  -5,   0 },
      {  33,  22,  -9,  -1, -87,  54,  14,  -3, -35, -32,  30,  -2,  15, -27,   0,   5 },
      {  40, -27,   9,   1, -26, -53,  32,   7,  43, -68, -26,  17,  19,  20, -37,   2 },
      {  31,  78, -37,  -4,  29,  36,   4, -16,  69, -12,   4,  -7,  -3,  23,   5,  -1 },
      {  -6, -11,  20,  -4,  62,   6,  20, -16,  -1, -48,  68,   6,   9, -55, -14,  33 },
      { -12, -60,  -1,   4, -17,  38, -53,  -3,  65,   6,  13, -27,  52,  -7,   9,   0 },
      {  19,  -1,  48, -19, -31, -30,  46, -14,  44,  44,  34, -18, -25, -10,  54,  13 },
      {  13,  28, -35,   0,   0, -64, -35,  10,  -9, -21,  -8, -40,  25, -58,  46, -21 },
      { -10, -32, -59,  17, -20, -18, -26, -10,   8, -24,  41,  -7, -83,  20,   5,  18 },
      {  12,  21,  20,  25, -15,   1, -37,  41,  32,  24, -25,  32, -37, -64, -36,  38 },
      {   5,  20, -30, -13, -16, -44, -15, -22, -11,  48,  36,   6,  49,  23, -44,  58 },
      {  -5, -10,  -6, -63,  -7,   5,  -4, -80,   2,   2, -45, -21, -26, -37, -35,  -4 },
      {   9,  30,  73, -14,  -2,  -7, -60,   3, -15, -30,  15, -52, -18,  40, -17,   7 },
      {   5,   3,  -1,  55,  -2,  -9,  24,  -7,   7,  29,  28, -48,  -6, -16, -63, -66 },
      {   2,   5,   0, -64,  -6, -12, -28,  22,  11,   7,  47,  59,  -5,  -2, -13, -69 },
      {  -7, -10, -27, -60,   7,  12,  33,  79,   2,   6,  -2, -58,  -8,  -2, -24,  19 },
    },
    {
      {  79, -78,  34, -13, -27,  35, -18,   5, -12,  15,  -7,   3,   3,  -6,   4,  -2 },
      { -35,  16,   4,   0, -81,  65, -21,   6,  39, -40,  15,  -2,  11, -13,   3,  -2 },
      {  68,  45, -75,  18,   1,  30,  -1, -11, -27, -11,  39, -10,   0, -14,   4,   5 },
      {  34, -21,  -7,   7,  -3, -28,  38, -16,  61, -64,  10,   1, -28,  52, -27,   9 },
      {  -2,   3,  10,  -5, -70, -35,  72, -15, -61,  -3,   4,   9,  15, -11, -20,  10 },
      { -15, -21,  41, -13,  43,   3,   1,  -7, -20, -49,  81, -12,   1, -48,  -6,  15 },
      { -23,   2,   8, -12,  -6,  34, -12,  -7, -51,   8,  19,   0, -87,  60,  -1,   5 },
      {  35,  62,  42, -68,   7,   6,  20,  -5,  21,  -4, -16,  39, -27, -23,  31,   1 },
      {   7,  16,   4, -11,  18,  10, -35,  34, -41, -52, -32,  49,  33,  21, -62,   0 },
      { -16, -10,   1,  11,  37,  72,  49, -67,   6,  15, -20,  24,  25,   6, -19,  12 },
      { -11, -27, -27,  42,   1,  -1,   7,  13,   2, -18, -32,  52, -64, -68,   1,  10 },
      {   6,  15,   5,  -9, -11,  -3, -12,  12,  31,  56,  21,  -2, -15, -17, -78,  67 },
      {  -8, -14, -15,   1,   7,  15,  40,  68,  15,  30,  54,  62,  18,  26,  17, -15 },
      {  19,  34,  52,  59,  13,  27,  38,  47,   0,   1, -14, -44, -16,  -6, -23, -34 },
      { -13, -22, -43, -68,   7,  13,  23,  15,   6,   3,  -8, -33, -17, -26, -46, -64 },
      {  -8, -14, -20, -28,  14,  22,  32,  51, -10, -22, -35, -43,   7,   5,  33,  76 },
    },
    {
      { -88, -14,  25,  13,  78,   0, -26,  -7, -17,  18,   9,  -5,   4, -10,   0,   3 },
      {   7,  80, -19, -37,  31, -57,  15,  21, -54,   3,  -2,  12,  10,   7,  -2,  -5 },
      {  36, -25, -14,  -4,  25,  61,  -4,  -9, -74, -32,  41,  17,  28,  -3, -24,   3 },
      { -56,  -1, -36,  -9,  -3,  19,  55,  -4,  15, -66, -28,  31,  -2,  39, -18, -19 },
      { -28,  28,  56,  18, -55, -20, -31,   0,  -9, -51,  17, -11,  56,  -4, -26,   1 },
      {  36, -14,  37,  -4,  35,  -2, -43,  13,  -2, -27, -49,  -8,   4,  77,  24, -22 },
      {  19,  46,  34,   4,  28,  13, -15, -19,  31, -49,  30,  27, -74, -15,  -8,  22 },
      {  24,  14,  14,  80,  20,  -9,  44, -44, -23,  -8, -54, -12,   9, -30,   2,  -4 },
      {  21, -19, -27,  28,  24, -56,  -6, -39,  36,  10,  40,  24,  26,  41, -51,  13 },
      {  21, -45,  25, -33,  32, -30,  25,  52,  21, -30, -23, -10,   9, -51, -47,   8 },
      {   0,  -9,  14,   0, -19,  -6, -29,  -8, -16,  27, -21,  66, -24, -22, -40, -81 },
      {   7,  54,   7,  -7,  18,  65,  -2,  10,  42,  41, -28,  -3,  39,   8, -53,   9 },
      { -13, -16,  28,   3, -30,  -5,  11,   7, -46,  28, -28,  15, -42,  37, -44,  71 },
      {  -5,  10, -41,  39,  -2,   0, -22,  36, -13, -11,   9, -73, -44,  11, -49, -32 },
      {   4,   0,  61,  10,   4,   6,  72,  21,   1,  26,  55, -11,  -7,  34,   1, -43 },
      {   1,   5, -15,  67,   3,   3,  -8,  85,   7,   0,   8,  57,  12,  -1,  24,  17 },
    },
  },
  { // 27
    {
      {  83, -39,  -8,   0, -77,  38,   9,   0,  18, -10,  -3,   0,   0,  -1,   1,   0 },
      {   6,  69, -46,  -1, -34, -55,  55,   0,  37,  -8, -25,   2, -10,  13,  -1,  -2 },
      {  62,  27, -28,   0,  25, -19,  12,  -1, -89,  21,  21,   0,  27, -17,  -8,   1 },
      { -25,  23, -17,   7,  -9,  44,  14, -14, -21, -94,  37,  16,  22,  14, -32,   2 },
      {  48,  15, -23,   2,  68,  28,  -6, -10,  28, -19,  11,   6, -78,  16,  14,   2 },
      { -30, -43, -59,  43,  13,  27,  53, -44,   6,  34,  -1,   7,   0, -35,  -8,  13 },
      { -10,  48,   3,   3, -33,  13, -29,  -5,  20,  26,  71, -29, -21, -59,  -1,  30 },
      {  12, -34, -26,  13,  12, -42, -20,   2,  34,  14,  60, -23,  24,  57, -51,   3 },
      {  15,  43, -21,   5,  26,  51, -30, -11,  39,  19, -31, -20,  77,  13,  17,   4 },
      {  30,   1,  43,  -2,  31, -19,  10,  -9,  37, -15, -21,  32,  19, -58, -71,  19 },
      { -12, -10, -28, -65,  16,  33,  36,  80,   6,   9,   2, -30,  -1, -15, -28,  -3 },
      {  -9,  26,  11, -28, -16,  35,   5, -13,  -2,  59,  18,  79, -10,  37, -32, -22 },
      { -11,   5, -53,  30, -19,   0, -72,  26, -15,  -2, -44,  11, -27, -19, -48, -25 },
      {   0, -24, -45, -50,   6, -30, -29,  -4,  23, -21,  26,  57,  26, -37,  47, -17 },
      {  -8,  -8, -20, -58, -10,  -4, -22, -42, -18,  -1, -30,  -7, -13,  21, -15,  90 },
      {   1,   0,  -2,  49,   2,   1,   3,  70,   2,   1,   6,  52,   9,  13,  24,  74 },
    },
    {
      { 107,  -2, -15,  -4, -52,  29,   8,  -1, -22, -17,  10,   1,   3,  -7,  -3,   3 },
      {  11, -55,   2,  -1,  58,  43, -39,  -7, -43,  40,  36, -10,  -7, -33,  12,  15 },
      { -53, -37,  -7,  10, -35,  61,  17, -13,  -7, -64,  28,  14,  20, -11, -32,   2 },
      {   6, -80,   4,   8, -37,   3, -26,   0,  72,  28, -18, -15,   2,  35,  10,  -3 },
      { -27,  23,  31,  -6, -72, -18, -36,  12,  -6,  24,  31, -32,  12, -56,  25,  26 },
      {   1, -31,  52, -10,  -4, -17, -22,  40, -61, -20, -55,  -5,  46,  14, -10, -27 },
      {  14,  50,  68,   5,  22,  60, -35,  13,  36, -14,  18, -18,   4,  36, -10,   1 },
      {  24, -22,  28,  15,  17, -68, -26, -10,  16, -34,  39,  18,  -8, -13, -65,  31 },
      {  -3,  17, -57,  13,   4, -19, -46, -29, -18, -12,  28, -36,  63,  53,   7,   4 },
      {   8, -15, -12, -66,  31,  -6,  24,  38,  29, -49,   2, -48,  17, -16,  28,  42 },
      {  -3,   3, -24,  30,   1,   3, -54,  -9,  -4, -54, -46, -40, -67, -23,  12, -12 },
      {  15,   2,  22,  59,  22,   4,  27, -49,  22,  -5, -41, -12,  56, -45,  21,  24 },
      {  -1, -17,  42,   0,  -7, -19,  11, -40, -25, -34,  24,  27, -27,  42,  79,  19 },
      {  -3, -13,  15,  51,  -4,  -8,  58,  22, -27,  14,  23, -78, -26,  29, -19,   7 },
      { -10,   8, -12,   5, -12,  19, -10,  11, -21,  17, -44,  20, -11,  30, -20, 104 },
      {  -4,   0,  22, -66,  -9,   4,   8, -85, -11,  15, -22, -40, -12,   5, -39,  -4 },
    },
    {
      { -98,  59,  -3,   3, -37,   1,  17,  -3,  23, -30,   9,   0,   8,  -7,  -3,   1 },
      {  45,  46, -35,  -2, -36,  78,  -6, -14, -23,  -8,  43, -10,   7, -24,  13,   7 },
      {  17,  64, -40,   3,  90, -12,   6, -12,  34, -13,   8,   1, -17,  19,  -6,   3 },
      { -40,   9, -23,   6,  15,  28, -65,  10,  -6,  73, -25, -28,  -5,  13,  36, -22 },
      { -27,  16,  41, -17,  38,   0,  40, -16, -79,  33,  31,   6, -33, -19,  12,  11 },
      {  23,  36,  37, -29, -29,  28,  45, -17,  25,  42, -45,  11, -15,  58, -12, -10 },
      {  -9, -27,  53, -27,  31,  34,   3, -27,  67,  11,  36, -32,  32, -25,  20,  17 },
      {  12,  -7, -19,   0, -33, -50,  23, -17,  13,   5,  47, -54, -37,  34,  57, -17 },
      { -26, -43,  -3,  28,   7,  49, -11,   1,   4, -14,  43,   8, -50,  63, -42,  15 },
      {  18,  27,  56,  33,   3,  12,  -8,  81,  20, -19,   8,  10, -39, -14,  35, -22 },
      {  -8, -12,   3,  14,  28,  27,  35,  11, -32, -34,  -4, -26,  61,  40,  16, -68 },
      {   8,  15,   0,  67,  -7, -18,  30,  29,   7,  61,  30, -24,  42,  -2, -44,  21 },
      {  13,  36,  47, -31,  -6, -26, -62,   1, -26, -22,  16, -52,  12,  28, -46,   1 },
      {   0,   8,   1, -25,  -3, -16, -19,  25,  -5,  10,  40,  66,  50,  52,  43,  36 },
      {  -5,  -6, -11,  -8,   8,  17,  24,  39, -15, -23, -44, -54,   0,  17,  24,  85 },
      {   7,  13,  37,  79,  -1,  -3, -23, -68, -10, -19, -25,  10,   8,  14,  36,  26 },
    },
  },
  { // 28
    {
      {  93, -48,  -2,  -2,  60,  -8, -16,   0, -13,  30,  -8,  -1, -11,  12,   3,  -2 },
      { -39,  -5,  12,   0,  60, -69,  10,   1,  64, -11, -29,   4,  -1,  32, -15,  -5 },
      {  33, -57,  32,  -2, -48, -37,  39,   4,   6, -64,   8,  13,  25, -18, -23,   8 },
      {  50,  47, -37,  -3,  10,  42,  -3, -13,  71, -45,  14,  -4,  26,   6, -17,   5 },
      { -21,  -2,  18,  -6,  71,   7,  17, -15, -15, -25,  66,  -5,   7, -63,  18,  21 },
      {  32,  57,  20, -11,  -9, -27,  68,  -7, -10,   4,  29,   8, -61,  32,  20,   1 },
      {   9, -12,  30, -16, -28,  -9,  -2,  -1,  58,  42,  17, -30,  29, -13,  81, -10 },
      { -22, -53, -15,  22, -11,  18, -21,  -5,  18, -14,  70, -24, -42,  60,  -3,   2 },
      {  12,  16, -69,  33, -13, -69, -35,  21, -11, -26,  11, -11, -17, -32,  36,  -1 },
      {  -6,   1, -18,   5,  21,   6,  31,  -1, -46, -35,  -5, -23,  62,  53,  42, -48 },
      {  -5, -15,  15,  16,  12,  43,  -3,  44,  14, -44, -35,  56, -40,  -3,  60,  -4 },
      { -13, -36, -70,  26,  -1,  13,  74, -21,  17,  39, -10,  27,   6, -13,   8,  21 },
      {  -8, -20, -21, -66,  -5,   3,  -7, -73,  -4, -34, -40, -28, -40, -11,  22,  -1 },
      {   3,  -5,  12,  45,   6,  21,  32,  18,  13,  -4, -25, -77, -35, -38, -21, -51 },
      {  -5, -10, -26, -54,  -2,  -3,   2,  22,  11,  13,  33,  40,  -8, -21, -21, -91 },
      {  -6,  -7, -26, -65,   5,   7,  21,  83,  -3,  -1,  -4, -43,  -1,   7,  -5,  46 },
    },
    {
      {  76, -75,  22,  -3, -49,  44,  -7,  -2,   5,   3,  -6,   2,  -1,  -2,   3,   0 },
      {  57, -19, -12,   3,  67, -45,  15,  -8, -57,  49,  -4,  -3,   3,  -8,   0,   5 },
      {  36,  28, -71,  31, -44, -17,  70, -25,  15, -13,  -6,   0,  -5,   5,  -8,   3 },
      {  41,  30, -47,  14,  32,  45, -51,  11,  -9, -37,  63, -10,  -6,  -1,   5,  -4 },
      { -21,  30, -19,   6, -47,  43, -14,   1, -62,  56,  -9,  -6,  41, -39,  10,   3 },
      {  -9,  -6, -36,  31,  34,  39, -23, -18,  49,  47, -58,  -4, -34,   7,  35,  -4 },
      { -20, -14,  12,   6,  40,  57,  43, -53, -24, -37,  -5,  52,  -2, -31, -15,  15 },
      {  51,  68,  21, -57,   0,   2,  -9,  10,   7, -13, -50,  44,  -6, -12,  21,  -4 },
      {   4,  22,  35, -24,  -4,  13,  39, -19,  17,  35,  54, -40, -55, -33,  37,  23 },
      {   3, -18,  -3,  17,  11, -28,   0,   3,  37, -24,   6,   1,  60, -72,  63,  -4 },
      {  19,  20,  11,  -3,  15,   7, -13, -19,  52,  24,  -2, -24,  45, -20, -76,  52 },
      {  -1,  -4,  -1,  35, -19, -31, -36,  13, -20, -27, -29,  -5, -66, -66, -26,  32 },
      {  20,  34,  60,  65,  10,  15,  14,  -4, -20, -28, -31, -57,  15,  20,  11, -16 },
      {   6,  14,  21,  25, -25, -35, -50, -90,  -1,  11,  24,  34,   1,  21,  16,  -2 },
      {  10,  17,  29,  57,   0,   3,  14,  46,  24,  39,  32,  63,  -9, -15, -25, -40 },
      {   0,   2,   7,  28,  -3,   1,   5,  35,  -7,   1,   6,  33,   9,  40,  39, 100 },
    },
    {
      { -89,  -8,  21,   5,  81,  -5, -20,  -3, -17,  19,   4,  -3,  -5, -13,   2,   2 },
      {  18, -67,  -5,  12,  15,  84, -14, -14, -33, -31,  35,   8,  11,  -2, -18,   1 },
      { -39, -53,  -3,  13, -28,  24,   9, -10,  91,   8, -21,  -7, -35,   5,  10,   4 },
      { -48,  24,  -6,   7,  -9,   4,  35,  -6,  -2, -89, -26,  24,  27,  41, -17, -15 },
      {  18, -10, -61,   0,  44,  13,  77,  -6, -17,  16, -50,   6, -17, -25,  12,   0 },
      {  31, -23,  14,   2,  32,  -2, -31,   6,  -5,   0, -44,  -1,   7,  75,  70, -17 },
      {  24,  44,  51,  32,  13,  28, -14, -26,   2, -39, -29,   6, -62, -42,  18,   9 },
      {  32,  14, -18,  12,  51, -18,  -5, -21,  40,   6,  36,  43, -37,  44, -53, -22 },
      {  16, -34,  22,  59,   0, -38, -11, -44,   0,  12, -50,   4,  51, -17, -43,   8 },
      { -22,  28, -38,  40, -25,  10,  -6, -69,  -9,  25,  34,  34,  16,   3,  57,  -3 },
      { -10, -45,  20, -21, -29, -33,   1,  10, -32,   1,  -8,  82, -34, -22,  10, -42 },
      {  -3,  36,  40, -35,  -6,  65,  11,  -9,  11,  56, -32,  32,  33,  11, -24, -26 },
      {  21,  -8,  33, -18,  34, -12,  33,   9,  47, -21,  32,  42,  51, -29,  39,  40 },
      {   1,  12, -62, -35,  10,   8, -77,   7,  24, -21, -37,  29,  18, -37,  -5,   8 },
      {   7,   8,  -5,  44,  12,   7,   0,  41,  25,  -6,  15, -17,  27, -37,  15, -95 },
      {   8, -13,   9, -67,  10, -19,   5, -79,   8, -21,   6, -44,   3, -16,   5, -46 },
    },
  },
  { // 29
    {
      { -96, -14,  17,   2,  74, -10, -21,  -1,  -9,  22,   5,  -3,   0,  -5,   3,   1 },
      {  37, -51,  -9,   5,  40,  78, -12, -14, -48,  -8,  39,   6,   5, -18, -11,   6 },
      { -46, -72, -13,  16, -44,  38,  16, -14,  60, -26, -12,  10,  -9,  15,  -9,  -6 },
      { -32,  42, -19,  -1,  10,  11,  53,  -4, -38, -78,  -9,  27,  18,  12, -35, -13 },
      {  16,  -1, -75, -11,  49,   3,  61,  -5,  22,  37, -26,   8, -34,   3,  25,  -4 },
      {  13, -56,  -2, -17,   7, -21,  -2,  37, -42,  -6, -67, -19,  46,  40,  17, -19 },
      {  34,   9,  50,  26,  48,  10, -25, -12,  28, -34, -44,  17, -54,  34,  -6, -23 },
      {  -4, -24,  35,  15, -20, -12,  36,  16, -36,  13, -41,  20, -43, -81, -15,  -3 },
      {  23, -39, -12,  17,  15, -71,   1, -32,   5,  13,  26,  47,  16,  12, -64,  -4 },
      {  -9,  16, -26,  78, -20,   2, -20, -68, -34,  16, -32,  -1,  16,   0,  32,  -9 },
      {  18, -16,  19,  15,  31, -27,  29, -26,  29, -49,  -3, -37,  33, -36,  30,  64 },
      {   6,  12,  62,  -1,   5,  31,  51, -13,  20,  46,   0,  43,  60,  11,  18, -16 },
      {   6,  23, -17,  12,  16,  28, -17,   9,  36,  19, -43, -42,  42, -36, -75, -12 },
      {  -2,  -3,  25,  42,  -6,   3,  51,   9, -23,  27,  20, -64, -27,  54, -35,  24 },
      {   7,  -5,  -9,  64,  14, -15,  11,  59,  18, -18,  42,  -9,  13, -20,  27, -65 },
      {  -2,  10, -16,  46,   1,  14, -17,  64,   4,   8, -15,  55,   9,  15,  -2,  76 },
    },
    {
      { 111, -31, -11,  -3,  -4,  37,  -9,  -1, -34,  14,  13,  -1,  -6,  -6,   4,   2 },
      {   1,  23,  -4,   0, -96,  34,  26,  -1,   6, -57,  19,   5,  24, -14, -15,   5 },
      { -49,   6,  10,   3,  13,  53, -20,  -6, -67,  22,  50,  -7,   4, -50,  14,  15 },
      { -25, -96,  34,   8, -48,  -2, -38,  13,   1,  18, -23, -14,  17,   3,  -1,  -8 },
      {  -5, -34,  -6,  11,  50,  52, -21, -22,  66, -41,  40,  -3,  20,  11, -26,  12 },
      {   1, -34,   6,   0,  34, -40,  27,  -1, -60, -60,  -8,  26,  24, -19, -54,  -1 },
      {  13, -24,  54,  -6,  -3, -37,  43,   3,  35,   5,  50,  14, -38, -44,  21,  40 },
      { -20, -15,  16,   7,  -6,  45,  31,  -4, -18,   3,  -6,  34, -84,  53, -31, -10 },
      { -13, -27, -84,  15, -23, -46, -16, -28, -13,  -6,  48, -18, -36,  14,   3,  14 },
      {  14,  35,  28,   7,  -9, -16, -71,  12,   5, -14,  -5, -32, -47, -28, -64,  14 },
      {   5,   8,  39,   3,   9, -12, -24,  29, -26, -64,  27, -28,  -9,  53,  61, -11 },
      {  -6, -17, -26,  -2,  10,  20,  -8,  -9,  12, -48, -53,   5, -47, -64,  49, -26 },
      {   1,  -5, -13,  52,   8,   2,  37,  63,  11,  11,  28, -40,  -5, -21, -19, -68 },
      {   0,  -6,  20, -29,   4,   6,  43, -63,  -8,  -1, -14, -94,  -5,   3, -10,  -9 },
      {   7,   9,  14, 109,   1,   1,  11, -30,  -8,  -2, -30,  -7,   7,   4,  14,  43 },
      {  -9, -12, -31, -19,  10,  20,  24,  71,  -3,  -7, -29, -38,  -4,   7,  -4,  80 },
    },
    {
      { -83,  72, -18,   4, -44,  27,   6,  -4,  15, -25,  15,  -2,   9, -10,   0,   2 },
      { -52, -17,  43,  -4,  29, -77,  44,   2,  32, -27, -15,  16,  -1,  10, -20,   5 },
      { -16, -46,  66, -20, -81,  19,  15,   7, -36,  16, -12,   9,  11, -12,  -1,   1 },
      { -52,   3,  -1,  10,  20,  -6, -47,  25, -17,  69, -60,  -4, -12,  26,   7, -26 },
      {  -2,  41,  10, -24,  24,  -1,  59, -40, -65,  42,  22,   4, -39,  18,  -2,  10 },
      {  15,   2, -54,  48, -45, -41,  18,  31, -50, -25, -21,  27, -17,  19, -31,  -4 },
      { -39, -35, -12,  28,  19, -32, -22,   9, -50,   0,  59, -27,   5, -54,  32,  11 },
      { -21, -50, -19,  20, -27,   7,   2, -35,  30,   8,  37, -45, -45,  66,  -6,  -9 },
      { -18, -11,  30,   4,  35,  49, -38,  28, -20, -40,  23,  41, -48,  13, -47,  18 },
      {  24,  34,  36,  14, -19, -18,   4,  59,  25,   3,  21,  -1, -62,   3,  64,  -8 },
      {   7,  17,  11, -57,  -4, -25, -34,  -7, -38, -60, -17, -57, -14,  12,  -1, -44 },
      {  -8, -13,  10,  36,  34,  45,  51,  23, -19, -26,   0,  -6,  36,  25,  29, -68 },
      { -20, -36, -42, -31,   2,  15,  10, -31,   2, -27, -37,  49, -45, -24,  60,  -7 },
      {  -7,  -4, -17, -48,  -3, -15, -18,  26, -13,  -1,  37,  39,  49,  71,  37,  25 },
      {   5,   5,   0, -12, -12, -23, -21, -20,  14,  25,  48,  55,  -1, -18, -25, -88 },
      { -10, -21, -43, -60,   5,  13,  40,  71,  16,  22,  12, -26, -14, -25, -35,  -9 },
    },
  },
  { // 30
    {
      {  83, -48,  -3,   0, -70,  46,   3,  -2,   9, -11,  -1,   2,   4,  -2,   1,  -1 },
      {  35,  68, -57,   3, -31, -49,  58,  -3,   8, -12, -10,   5,  -2,  10,  -7,  -1 },
      {  49,   2, -21,  -1,  43,   2,   0,  -2, -96,  21,  28,   1,  23, -24,  -9,   4 },
      { -31,  11, -17,  16,   0,  48,  11, -20, -12, -85,  46,  17,  14,  14, -41,   2 },
      { -11, -45, -54,  51,  37,  24,  45, -49,  20,  25, -16,   3, -24, -13,  17,  10 },
      {  49,  30,  -6,  -8,  48,  18, -37,   2,  28, -13,  30,  -4, -79,  26,  12,   2 },
      { -21,  31, -20,  21, -42,  10, -28, -10,  15,  36,  69, -40,   5, -44,  19,  35 },
      {  18, -40, -20,  29,  14, -56, -34,  -9,  23,   1,  28, -32,  42,  61, -29,   6 },
      {  37,  12,  29,   5,  46,  -8,   9,  -3,  47, -42, -13, -15,  49, -62,   8,  33 },
      {   0, -37,  25, -12,  -6, -46,  30,   5,   8,   6,  37,  22, -51, -41, -68,  20 },
      {  10,  18,  52,  73, -24, -31, -24, -53, -34, -21, -21,  12, -25,  -1,  14,   9 },
      { -16, -36, -32,  -4,  -9, -28,  -6,  37, -33, -60, -17, -59, -35, -19,  34,  16 },
      {   3,  22,  32,  30,   5,  39,  37,  29, -10,  21, -25, -69, -10,  28, -48,  33 },
      {   0, -13,  42, -20,   3,  -9,  61,  -6,  -9,  -3,  49,   5,   6,  45,  65,  38 },
      {  -5,   2, -18, -60,  -7,   3, -26, -55, -10,   2, -36,   4,  -2,  15, -18,  82 },
      {   2,   1,  19, -43,   0,  -4,  17, -72,  -1,  -4,   8, -66,  -5, -10,  -7, -63 },
    },
    {
      {-102,  50,   4,   3, -41,  -6,  19,  -1,  17, -30,   4,   4,   8,  -7,  -8,   3 },
      {  26,  32, -29,   0, -63,  81, -16,  -6, -33,   4,  33, -13,   4, -20,  15,   0 },
      {  27, -56,  39,  -3, -72, -23,  39,  11, -13, -42,   8,  20,   9, -15, -24,  17 },
      { -47, -44,  26,   7, -30,   0, -36,  21, -54,  63, -36,  -5,  -9,   4,  17, -18 },
      { -33, -31,  28,   3,  56,  19,  -7, -13, -37, -10,  67,   0,  12, -53,  -5,  31 },
      { -10, -47,  -3,   9, -10,  27, -61,   0,  69, -20, -11, -29,  56,  -9, -11,  -6 },
      {  -7, -12, -67,  46,   7, -15, -40,  23, -42, -48, -10,  16, -24,   5, -48,   1 },
      {  -1,  13,  37,  -9,  24,  69,  11,   7,  -5,   2, -42,  56,  14,  17, -62,   0 },
      { -22, -43,  -6,  -6,  -9,  28,   8, -46,  21,   6,  39,  -9, -58,  72, -18,   9 },
      {  -6, -12, -12, -38,  13,   1,  24, -51, -45, -35, -22, -43,  31,   6, -11, -67 },
      {  10,  25,   6, -54, -18, -32, -47, -45,  -7,  20, -17, -20, -13, -22, -53,  50 },
      {  13,  27,  68,  -6,   4,   5, -40,  36,   1, -45,  12, -35, -55,   4,  -2, -38 },
      { -11, -22, -13, -39,   8,  25, -10, -15,   3, -49, -52,  28, -39, -22,  65,  31 },
      {  -4, -16, -19,   0,  -6,   4,  16, -12,  38,  28,   1,  19, -51, -77, -28, -58 },
      {   4,   1,  14,  65,   7,  22,  40, -16,   1,   1, -48, -69, -21, -21, -10,  38 },
      { -13, -17, -34, -61,  11,  16,  33,  80,   5,   9,   3, -45,  -6,  -1, -22,  19 },
    },
    {
      {-111,   0,  19,   3,   6, -43,  -1,   6,  36,   2, -19,  -3,   0,   6,   0,   2 },
      {  11,  20,  -8,  -2, -91,  -7,  43,   6,  42, -56, -14,  10,   5,   8, -15,   4 },
      { -32, -69,   6,  15, -27,  68, -23, -17,  15, -15,  53,   1,   7, -18, -11,  18 },
      {  31, -63,   7,  -2, -37, -23, -39,  16,  39,  44, -41, -29,  -5,  27,  29, -14 },
      {  -2,  68,  -4,  -9, -28,   6, -30,  -8,  31,  36,  45, -43,   2, -26,  47,  28 },
      {   3,  -1, -67,   4,  37,  21,  32, -41,  67,  19,   6,  18, -30,  35,   7,  -9 },
      {  37, -16,  49,   1,  27, -49,   9,   6,  46,   1,  36,  22, -24, -18, -31,  57 },
      {  14,  27,  32,  10,  40,  28, -49,   3,  37, -63, -13, -37,  16,  47, -13,  -8 },
      {   2, -28, -65, -22,   9, -53, -10, -11, -17, -33,  24, -33,  55,  16,   1,  39 },
      {   1,  19, -18,  46, -30, -38, -59, -45, -14,  11,  20,  18, -20,  11, -57, -30 },
      {  -1,  -7, -39,   4,  10,   4,  -5,  24,   5, -21, -30, -61, -61, -64, -35,   3 },
      {   6, -15,  25,  49,   3, -20,  19, -73, -16, -37, -17, -18, -21, -18,  63,   7 },
      {  11,   5,   0,  49,  13,   8,  19, -15,  30,  31, -32, -10,  83, -47, -30,  -5 },
      {  -9,  12,  -1,  21, -14,  28,   5, -13, -28,  30, -43, -13, -13,  46, -26,  89 },
      {  -2,  -7,  22,  28,  -3,  -8,  60,  12, -14,  21,  49, -71,  -5,  41, -22, -31 },
      {   0,  -3,  30, -85,  -6,   5,   8, -74,   0,  12, -16, -26,   3,  -8, -36, -13 },
    },
  },
  { // 31
    {
      { -90, -26,  17,   3,  77,   6, -23,  -2, -15,  21,  10,  -3,   1,  -8,   2,   3 },
      {  43, -50, -16,   5,  20,  82,  -5, -12, -44, -22,  38,   6,   8, -12, -16,   5 },
      {  34,  74,  12, -14,  46, -28,  -7,  13, -75,   0,  10,   0,  17, -16,  -6,   1 },
      { -46,  19, -23,  -4,   1,  13,  60,   1, -17, -75, -27,  30,  16,  22, -29, -22 },
      { -25,   5,  63,  13, -50,  -5, -54,   2,  -9, -51,  34, -16,  28,   6, -29,  19 },
      {  26, -44,  13,  -4,  17, -11, -31,  26, -23, -11, -67, -25,  33,  61,  18, -27 },
      { -36,  -1, -49, -33, -48,  11,  11,  29, -37,  22,  13, -44,  47, -12,  39,  17 },
      {   4, -50, -32,   7, -11, -70,   1, -10, -22,  26,  20,  38,  35,  -5, -59,  -6 },
      {   8, -28,  20, -60,   8,  -5,  16,  77,  14, -14,  -8, -17, -30, -41, -43,   8 },
      { -11, -12,  30,  51, -34,  11,  13,  -6, -45,  19, -58,  -4, -15, -67,  -4, -26 },
      {  16, -27,  26,  22,  24, -31,  59, -24,   4, -25,  -6, -33,  25, -12,  22,  75 },
      {  -1,  13, -67,  33,  15, -12, -35, -13,   7, -28, -17, -71, -29, -10, -41,  10 },
      {  -8, -29, -13, -20, -16, -39, -17, -12, -46, -41,  11,  22, -75,   4,  51,  10 },
      {   9, -13,   8, -30,  13, -22,   5, -41,  23, -32,  24, -34,  26, -45,  25, -79 },
      {   4,   5, -28,  -9,  11,   3, -51,   5,  29, -30, -41,  54,  40, -55,  25,  33 },
      {  -5,   4,  13, -75, -12,  15,  -8, -77, -12,  20, -41,  -9,  -5,   7, -34,  26 },
    },
    {
      { 110, -27, -13,  -2,  13,  39, -15,  -2, -28,  21,  13,  -3,  -4,  -6,   7,   1 },
      {  19,  23,  -9,   2, -82,  35,  33,  -6,   2, -67,  26,   9,  14,  -8, -27,   8 },
      { -43,  31,   5,  -2,   4,  58, -14,  -8, -40,  30,  59, -17,   3, -47,  35,  12 },
      {  27,  96, -28,  -6,  28,  -6,  49, -12,  19,  13,  -6,  11, -22,  17,  29,  -2 },
      {  -2,  -7,  10,   5,  80,  -1,  27, -10, -16, -45,  36,  32, -10, -21, -58,  25 },
      {   6,  19,  19,  -4, -25, -45,  21,  28, -95,  11, -32,  20, -14, -29, -10,  -8 },
      {  24, -17,  54, -17, -14, -55,  34,   5,  32,   9,  44,  -7,   5, -28,  42,  47 },
      {   1,  -9,  41, -18,  29,  33,  44,   9,  -6, -23, -31,   0,  75,   0,  35, -45 },
      { -20, -40,  28,  10, -14,  48,  53,  -3,  -4,  26,  -6,  42, -61,  42,  15,   7 },
      { -14, -46, -73,  -3,   1, -31,  36, -50, -33, -21,  17, -13,   4,   4,  43, -10 },
      {  -4,  12,   5, -12,   9,   2, -13,  24, -42, -21,   9, -26,  24,  86,  12,  65 },
      {   5,   1,  26,  -7,  14,   7,  -8,  10,  -4, -62, -15, -69, -72, -11,  25, -21 },
      {   5,  11,  38, -28,  -9, -21,   0, -41, -14,  21,  56, -21,  -2,  46, -36, -64 },
      {   3,   0, -11,  64,   5, -17,  -8,  69,   1, -17,  53,  25,  -1,  18,  29, -45 },
      {   2,  -2,   3,  79,   1,   3,  49,  -3,  -1,  32,  -9, -71,  16,  -2, -35,  11 },
      { -10, -16, -42, -65,   3,   9,  38,  74,  13,  24,  17, -30,  -6,  -6, -32,  -9 },
    },
    {
      { -93,  74, -14,   2, -22,  -1,  18,  -6,  20, -26,   8,   2,   5,  -4,  -3,   1 },
      {  33,  31, -40,   9, -44,  85, -33,  -1, -16,   2,  33, -17,   9, -19,  16,   0 },
      { -27, -49,  60, -21, -89,  17,  10,   1, -18,  -7, -11,   4,  16,  -9, -12,  -1 },
      { -43,   9,  13,   4,  14,  16, -43,  26, -22,  88, -46,  -9,   2,   5,  26, -17 },
      {  21,  42,  26, -18,   5,   7,  79, -25, -55,  46,  21,  16, -12,   0,   8,  11 },
      { -14, -17, -62,  37, -30, -45,  -5,   2, -81,  -7,  -6,   4, -26,   1, -16,  -4 },
      { -35, -39,  -3, -19,  22, -18, -10, -50, -12,   9,  58, -53,  12, -41,  43,   1 },
      {  27,  21, -30, -56, -22, -23,   3, -45,   1, -10, -60, -16,  29,   0,  21, -54 },
      { -14, -12,  -4, -34,  -8,  22, -15, -34,   9,   2,  14, -15, -80,  78,  -3,  -8 },
      { -21,  -9,  -9, -45,  43,  36, -30, -37, -33, -14, -19,  46,   9, -27, -54,  28 },
      {   8,  27,  30, -49,   4, -22, -35,  52, -30, -39,  16,  22, -42, -28,  44, -13 },
      {  22,  24,  -6, -20, -42, -48, -35, -19,  32,  41,   0,   2, -16, -19, -12,  76 },
      {  12,  36,  37,  -5,  12, -12, -21,  12, -32, -14,   5, -83,  17,  15, -63,   0 },
      {  -7, -10, -33, -51,  -3, -14,  -1,  42, -10,  14,  43,  17,  65,  60,   3,  15 },
      {  -4,  -5,  -1,   3,  11,  19,  15,   3, -23, -39, -51, -32,   8,  25,  55,  78 },
      {  -9, -23, -44, -45,   2,  17,  48,  55,  20,  15, -16, -41, -34, -41, -26,   6 },
    },
  },
  { // 32
    {
      { -80,  39,   5,   2,  78, -38,  -7,   0, -27,  13,   5,  -1,   4,  -1,  -2,   1 },
      {  -7, -64,  46,  -1,  33,  59, -54,   3, -36, -10,  27,  -1,  16,  -8,  -7,   1 },
      { -66, -30,  33,   5, -22,   3, -20,   0,  83,  -5, -28,  -2, -30,  16,  15,  -4 },
      { -42,  24, -21,   7, -25,  34,  21, -10,  -3, -88,  29,  13,  22,  19, -40,  -2 },
      { -16,  28,  56, -19, -67, -38, -31,  35, -28,   5,  -5, -10,  50, -12,  -8,  -7 },
      { -39, -61, -21,  18, -18,  -1,  56, -18, -30,  26, -48,  15,  46, -19,  -6,  -4 },
      {  34, -22,  29, -17,  37, -38,   7,  15,  -7, -46, -56,  21,   4,  57, -24, -30 },
      {  -7,  47,  23, -22,   0,  66,  17,  11, -27,  -9, -59,  23, -26, -23,  44, -18 },
      { -13,  -4, -66,  -5,  -7,  19, -38,  32, -15,   5, -18, -49,  25,  57,  41, -21 },
      {   9,  27,   2,  72, -14,   1, -52, -67, -13,  13, -24,  25,   2,  17, -10, -31 },
      { -22, -12, -14, -22, -32,   3,  -1,  19, -42,  39,  15,  24, -74,  31, -52, -23 },
      {   5,   4, -45,   2,  16,   7, -44,  46,  30,  -3, -35,  18,  10, -62, -57, -15 },
      {   4,  21,   6, -41,  13,  33,  10,  -9,  42,  54,  23,  47,  56,  35, -13, -28 },
      {   6,  24,  27,  23,   9,  38,  22,   6,   7,  28, -32, -66,   4,  26, -61,  42 },
      {   2,  -2,  15,   5,   7,   2,  28,  -9,   9,  -2,  20, -60,  -5, -28,  -9,-101 },
      {  -4,  -2, -16, -83,  -5,  -7, -27, -78,  -5,  -9, -26, -30,  -2, -15, -19,  11 },
    },
    {
      {-100,  57,   0,   2, -46,   1,  17,  -1,   9, -23,   4,   2,   3,  -3,  -5,   0 },
      {  37,  35, -35,   2, -48,  83, -17, -10, -30,   8,  36, -14,   2, -15,  16,   2 },
      {  30, -47,  25,   0, -72,  -6,  35,   1,  10, -64,  12,  16,  13,  -6, -35,  10 },
      { -34, -58,  37,   3, -50,  12, -40,  24, -37,  49, -33, -12,   3,  -2,  17, -18 },
      { -11,  -4,  61, -18,  24,   0,  52, -11, -43,  20,  57,   3,   1, -42,  30,  26 },
      {  25,  36, -16,  -3, -22, -32,  44,   7, -73,  21, -42,  37, -31,  18, -17, -11 },
      { -30, -23, -15,  22,  43,  28, -33,  18, -53, -32,  15,  32,  -8, -24, -66,   8 },
      {   5,  19,  56, -39,  17,  70,  13,  -3,  32,  15, -41,  41,   5,  19, -32,  -2 },
      { -19, -32,  -2,   2,  -1,  18,  16, -36,   2,  10,  36, -20, -76,  73, -14,   4 },
      { -13, -19, -13, -34,  22,  17,  25, -43, -33, -40, -20, -33,  29,   3,  12, -78 },
      { -19, -39, -48,  48,   4,  23,  55, -24,  14,  42, -10,  34,  48,   4,   5,  16 },
      {   2,  -3,  11,  52,  15,  31,  28,  63,  11, -35,  -8,  17, -39,   0,  56, -35 },
      {  13,  26,  33,  47,   6,  -3,  18,  35,  -9,  20,  27, -51,  47,  39, -48, -27 },
      {  -8,  -7, -14, -36,  12,  11,   2,  41, -32, -33, -10, -11,  39,  63,  33,  63 },
      {   6,   9,  23,  48,   8,  16,   9, -39, -13, -22, -64, -57, -17, -22,  -3,  52 },
      {   7,  18,  38,  45,  -2, -13, -45, -56, -20, -22,  12,  52,  26,  41,  33,  -8 },
    },
    {
      { 109,  -4, -20,  -2, -42,  37,  11,  -2, -18, -16,  13,   2,   1,  -4,  -3,   2 },
      {  11, -62,   7,   5,  61,  41, -45,  -9, -28,  45,  35, -11,  -5, -19,  12,  10 },
      { -51, -24,  -4,  13, -25,  60,  14, -20,  -4, -59,  46,  16,  15, -13, -44,   7 },
      {   6,  68, -11, -12,  68,   3,  36,  -1, -58, -14,  17,  30,  -7, -24, -13,  -3 },
      { -19,  62,   8,  -4, -47,  16, -19,  -2,   0,  34,  51, -33,   0, -40,  43,  30 },
      {   1,  -5,  59, -12, -24, -27, -43,  40, -67, -25, -12, -19,  30, -19, -29, -12 },
      {  32,   2,  44,  16,  15, -54, -13, -18,  49, -12,  45,  24, -20, -24, -42,  40 },
      { -13,  -6, -12,  14, -43,  -2,  -7,  -5, -23,  55, -15,  51, -62, -33, -38, -33 },
      {  -6, -28, -63,  -3, -17, -64,  15, -23, -32,  24,  30,   7,  60,  -6,  -7,  17 },
      {   5,  49, -30,  32,   7,  11, -76, -33,   0,   4,  -2,  -6,  15,  53, -42,  -9 },
      {  -3,  -6, -67, -42,  12, -15, -32,  30,  17, -37,   0, -46, -44, -40, -21,  -2 },
      {   4,  13,   0, -48,   7,  32,   4,  50,  45,  47,  -3,  21,  55, -10, -50,   0 },
      { -11,  -6,   0, -44, -14,   0,   1,  30, -29,   8,  23,  16, -43,  75,  -8,  62 },
      {   1,  -2,   8, -63,  -5, -12, -35, -20,  13, -20,  41,  51,   4,  14,  40, -63 },
      {   3,  -1,  16,   3,   2, -11,  43,   4,   0,  25,  55, -63, -17,  33, -34, -63 },
      {   0,   0, -23,  67,   4,  -8, -11,  87,   8, -11,  37,  31,   3,   9,  25, -16 },
    },
  },
  { // 33
    {
      { 101,   6, -24,  -2, -64,  23,  23,  -1,   3, -21,   3,   4,  -1,   3,  -4,   0 },
      {  23, -57,  -3,   7,  46,  71, -28, -15, -42,   7,  45,   0,   6, -17,  -5,  10 },
      {  37,  83,   4, -15,  52, -22,  10,  12, -56,  19,  14,   3,   2, -23,   7,  10 },
      { -36,  24, -22,   6,  13,  34,  55, -15, -19, -73,   3,  41,   6,   6, -42, -15 },
      { -14,   7, -80,   2,  15,  12,  46, -25,  41,  50,  23,   3, -26, -11,  26,  24 },
      {  25, -47, -33,  -9,  42, -20,  26,  13, -29,  15, -74,  13,   1,  27,   7, -42 },
      { -27, -19, -47,  -2, -48, -40,   1,  -4, -60,  14,  25, -16,  65, -14, -10,   1 },
      {  24, -22,   8,  25,  11, -63,  -8, -32,  12,  17,  30,  53, -26,  -8, -68,  -1 },
      { -10, -31,  36, -48, -17,   9,  55,  57,  -1,  25,   2,  25,  -8, -57, -16,  15 },
      { -17,  21,   5,  45, -38,  30, -13, -23, -31,  34, -32,  11, -39, -53,   2, -54 },
      {   9, -11,  51,  63,   6,  -4,  56, -38,  -2,  10, -18,   5,  46,  -3,  32,  37 },
      {  -2,  26,  16, -24,  -9,  36,  -9,  -5,  21,  58,   9,  53,  57,  46, -11, -34 },
      {  13,  12, -15,  38,  27,  12,   2,  35,  46,   8, -13, -51,  46, -36, -59, -21 },
      {  12,  -9,   9, -46,  17, -20,   3, -53,  31, -29,  16,  -3,  29, -54,  36, -58 },
      {   2,   7, -28, -12,   2,   7, -51,  -8,  13, -14, -59,  52,  28, -44,   1,  59 },
      {  -3,   6,  16, -53,  -5,  15,  12, -67, -12,  22, -36, -51, -10,  10, -54,  26 },
    },
    {
      { 112, -25, -14,  -2,  -6,  44,  -5,  -3, -22,   4,  20,  -1,  -2,  -7,   2,   7 },
      {  -4, -48,  20,   0,  75, -15, -50,   7, -14,  64,  -8, -18,  -5,  -3,  22,   1 },
      { -46,  12,   5,   4,   1,  72, -21, -14, -16,   3,  74, -13,  11, -36,   7,  31 },
      {  11,  64, -11, -17, -41,  -7, -14,  14,  -4,  50, -14, -43,   7,  -3,  71,  -3 },
      {  25,  67,   5, -11,  75, -23,  44,  -8,  17,   2,  28,  18, -14,  -5,   4,  37 },
      { -13,  37, -39,   8,  33,  42,   5, -12, -62,   9, -34,  28, -15,  -3, -10, -63 },
      {   4,   4,  65,  -7, -19, -35,  19,  35, -81,  -6,   5,   8,   5, -46, -10,   6 },
      { -12, -46, -24, -19,  21,  -4,  65, -17, -15, -29,   5, -20,  42, -13,  67, -17 },
      { -16, -26, -58,  16, -31, -33,  28,  -9, -32,  52,  29,  35, -38,  15,   1,  39 },
      {  13,   8, -29,  -2,  -3, -61, -38, -49,   4, -15,  44, -16,  -4, -51, -14, -49 },
      {  -1,  14, -49,  -4,  19, -24, -43,  29, -46, -47,   0, -25,  42,  40, -12,  42 },
      {  -4,  -4,  21,  -4,   3,   1,  -6,  -3, -20, -47,  19, -35, -93,  40,  34, -11 },
      {  -5, -10, -44,  -2,   9,   8,  -8,  52,  23, -26, -37,   0, -43, -79,  11,  27 },
      {   9,   9,   8, 121,   5,  -9,   0,   7,   3, -13,   3,   2,  11,  -3,  31,  -3 },
      {   0,   2,   2,  18,   0,   5,  25, -63, -18,  10, -48, -72, -10, -17, -38,  40 },
      {  -2,  -3, -18,  12,   9,   2,  42,  65,  11,  25,  37, -63,  -2,   9, -46, -39 },
    },
    {
      { -93,  76, -17,   3, -13,  -2,  15,  -5,  20, -24,   9,   0,   5,  -3,  -3,   2 },
      {  25,  14, -25,   5, -76,  83, -18,  -6, -17,  -8,  32,  -8,   9, -18,   6,   6 },
      {  43,  54, -77,  16,  63,   2, -13, -13,  -7,  16,  17, -12, -13,   6,  12,   0 },
      { -47,  -1,   7,   1,   8,  21, -53,  16, -44,  78, -32, -17,  -4,  -3,  29, -19 },
      {   1,  19,  21, -12,   0, -12,  71, -21, -82,  30,  31,  18, -12, -17,   0,  23 },
      {  25,  29, -27,   6, -50, -15,  26,  15, -11,  10, -76,  30, -25,  40, -18, -32 },
      {  15,  26,  39, -56,   4,  33,  21, -37,  53,  31,   0, -15, -33,  38,  38,   1 },
      { -34, -52, -29,  38, -15,  -3,   6, -34,  -5,   4,  35, -20, -61,  57, -11,  -4 },
      {  10,  -6, -19,   5, -41, -57,  23, -13,   7,   7,   5, -63,  28, -22,  62, -33 },
      { -19, -37, -36,  -6,  19,  32,  31, -73,  13,  15, -36,  24,  47, -23, -12, -22 },
      {   4,   0,  -6,  52, -22, -12,  13,  11,  51,  66,   2,  22,  15,  -8,   2,  71 },
      {   2,   2,  28,  65,  32,  51,  56,  41,   4,  -7,  -3, -21,  18,  18,  18, -36 },
      {  12,  24,  24,   3,  -5,  -7, -11, -26, -11,  19,  -1, -63,  51,  40, -74,   8 },
      { -16, -27, -45, -51,   5,  10,  23,  46, -15,  -9,  -2,  -8,  44,  56,  23,  44 },
      {   6,  13,  28,  40,   0,  -6, -31, -51, -30, -40, -25,  24,  25,  41,  56,  34 },
      {   3,   6,   5,  -4,  -9, -18, -17,   5,   8,  28,  64,  62,  43,  39,   4, -58 },
    },
  },
  { // 34
    {
      { -73,  50,  -3,   1,  73, -47,   2,   0, -27,  13,   4,   0,   5,   1,  -4,   0 },
      {  31,  52, -51,   4, -39, -56,  62,  -3,  20,  14, -30,   1,  -5,   1,   1,  -1 },
      { -71,  10,  20,  -2, -35,   4,   2,   5,  77, -48, -13,   4, -25,  25,  -2,  -6 },
      { -13, -41,  41, -11,  -7, -38, -15,  18,  15,  67, -65,  -3,  -7, -18,  35, -12 },
      {  39,  54,  44, -36, -15, -36, -51,  44,  13,   2,  40, -25, -14,   3,  -7,   7 },
      { -28,  11,  37, -11, -59,  -8,  15,  14, -53, -23, -14,  13,  69, -25, -28, -12 },
      { -36,  14, -16,   9, -32,  27,   1, -12,  24,  51,  44, -26,  12, -63,  11,  52 },
      {   9,  54,  20, -40,  20,  66,  29,  18, -10, -10, -35,  21, -21, -31,  47,  -1 },
      {  21,   3,  67,  21,  16, -12,  34, -37,  17,  12,  10,  58, -24, -19, -48,  31 },
      {   4,  44,   2,  54, -24,   3, -60, -57,  -8,   3,  -9,  22,  -8, -16,  26, -49 },
      { -26,  -7, -26, -25, -31,  14,   1,  21, -37,  36,  18,  25, -74,  -4, -46, -40 },
      { -10,   7,  30,  41, -29,   3,  16,  -3, -58,  -3, -11, -41, -46,  45,  26,  51 },
      {  -2,  29,   6, -21,  -6,  41,  -5, -19,  12,  63,  -7,  12,  41,  81, -18,   4 },
      {   9,  20,  -3,  47,  19,  32, -14,  34,  14,   4, -55, -47,  -1, -21, -71,  -2 },
      {  -2,   0, -31, -53,  -7, -12, -52, -42, -14, -21, -49,  13, -12, -10, -25,  63 },
      {  -2,   2, -26,  44,  -5,  -1, -27,  71,   0,   2,  -2,  76,  10,  10,  16,  41 },
    },
    {
      { 102, -59,  -3,  -1,  21,  20, -24,   1, -18,  25,   3,  -6,  -2,   1,   6,   1 },
      {  25,  49, -36,   2, -50,  76,  -1, -14,  -5, -21,  50, -11,   6, -16,   5,  14 },
      {  13, -43,  54,  -8, -68, -10,  57,   1, -16, -31,  10,  28,   1,  -6, -28,  19 },
      { -52, -40,  27,   2, -43,  32, -48,  11, -41,  46,  -2, -26,  10, -21,  27, -17 },
      { -30, -36,  30,  -6,  55,  12,  -4, -26,  24, -13,  78, -13,   4, -16,   2,  46 },
      {  -4, -46, -37,  31, -30, -21, -46,  12,  50, -63,  -2, -24,  28,  -4, -13, -17 },
      { -28, -24, -34,  45,  21,  38, -15,   5, -36,  -6,  -5,  65, -27,  10, -56,   8 },
      { -13, -30, -23, -12,  22,  19,  47, -62, -24, -27,  -2,  -6,  14,  -8,  19, -77 },
      {   8,  11, -26,  21,   8, -60,   2,  33, -66, -21,  39,   3,  -4, -58,  26,   0 },
      {  -8, -22, -44,  24, -33, -34,  22, -31,  -2,  36,  29, -34, -56,  54,  10,  16 },
      {   9,  16,  45, -13,   0,   3, -39,   5,   3, -38,  21,  -2, -88,   7,  -3, -52 },
      {  -1,   7, -19, -72, -13, -33, -47, -55, -31,  -1,  -5,  -2,   8,  -9, -54,  18 },
      {  -9,  -5,  -1, -28,  18,  16,   2,  40, -53, -52,  -8, -38,  16,  74,  11,  22 },
      { -12, -29, -37, -55,  -9,  10,  -2,  10,  22, -17, -22,  49, -36, -16,  67,  31 },
      {   0,   0,   8,  29,  11,  17,  12, -27, -11, -26, -65, -57, -40, -47,  -6,  45 },
      {  10,  18,  32,  46,  -8, -15, -38, -63, -18, -24, -11,  39,  22,  31,  53,  20 },
    },
    {
      {  97,  28, -26,  -8, -49,  22,  37,  -1, -10, -28,   0,  11,   5,  10,  -3,   8 },
      {  46, -58,  -2,  10,  41,  59, -40, -12, -17,  26,  46,  -5,  -9, -19,  -2,  16 },
      { -52,  15, -12,  -6,  24,  58,  52, -19, -20, -49,  38,  31,  17,   2, -22,  18 },
      {   9, -66, -34,   9,  27,  -6,  12,  -2, -19, -36, -57,  22,   0,  21, -23, -60 },
      {  28,  45,  37,   2,  69, -22, -19,  12, -62, -21, -18,   0,  34,  18,  15,  14 },
      {  -8, -14,  46,  44, -60,   6, -56, -11, -14, -56,  13,   1,  34,  -3, -21, -11 },
      { -20, -31, -32, -19, -22,  27,  -3,  12, -13,  -6, -19, -48,  52,  36,  68,  27 },
      { -14,  51, -41,   5,  -7,  33, -24,  -8, -41,   8,  -1, -31, -11, -45,  17, -70 },
      {   7,  25, -61,  45,  26, -30, -30, -56,  43, -13,  27,   7,  17,  34,  13,   3 },
      {   2,  31,  35,  17,  13,  74, -16,   7,  54,   5, -46,   9, -18,  45,  10, -17 },
      { -15,   5,  -9,  24, -33,   7,  -5, -27, -54,  71, -21,  63,   9,  30,  -4,  15 },
      { -14,  12, -50,  31,  -3,   5, -32,  66, -17, -19, -15, -10, -45,  10, -38,  52 },
      {   3,  12, -20, -31,   5,   7, -18,  47,  25,  32,  18,   5,  78,  11, -59, -26 },
      {   6,  -4,   3,  86,  11,   6,  50,  13,  13,  21, -26, -22,  41, -49,  -3,  13 },
      {   0,  -4, -13,   3,   1,  -2, -17,  51,  19, -17,   9,  82,  10, -41,  64,  -6 },
      {   0,   6, -15, -41,   3,  10, -33, -47,  18, -12, -65,  11,  19, -58, -21,  46 },
    },
  },
};

const int8_t g_lfnst8x8[35][3][64][64] = 
{
  { // 0
    {
      {-108,  29,  27,   0,   4,   0,   1,   0,  40, -29,  -6,   5,  -1,   1,   0,   0,  23,   4, -12,  -1,   0,   0,   0,   0,  -3,   3,   2,  -2,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  -6,  75, -25, -15,  -3,  -3,  -1,  -1, -66, -12,  46,  -1,   2,   0,   1,   0,  27, -43,  -5,  14,   0,   2,   0,   1,  13,   9, -14,  -3,   1,   0,   0,   0,   2,  -2,   3,  -1,   0,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
      {  21,  69, -43, -13,  -3,  -2,  -1,  -1,  71, -12,  -1,  -8,  -5,  -1,  -1,  -1, -55,  -2,  26,   6,   1,   0,   1,   0,  -8, -12,   2,   5,   2,   0,   1,   0,  -2,  -2,  -1,   1,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
      {  45,   9,  18, -18,  -5,  -3,  -1,  -1,  12, -78,   6,  35,  -1,   4,   1,   2,  14,  27, -54,  -1,   8,   0,   1,   0, -18,  30,  16, -20,  -1,  -1,   0,   0,  -5,  -6,   7,   5,  -3,   0,   0,   0,  -3,   4,  -1,  -1,   1,  -1,   0,   0,  -1,  -1,   0,   1,   0,   0,   0,   0,  -1,   2,   0,  -1,   0,   0,   0,   0 },
      {   3,   2, -69,  26,   6,   3,   1,   1,  14,   9, -10, -23,   6,   0,   1,   0,  75,  21,   3, -19, -12,  -2,  -3,  -1, -45,  12,  26,   5,  -2,  -2,   1,   0,  -6, -15,   8,   5,   1,   0,   0,   0,  -3,   0,  -1,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,  -2,   0,   0,   1,   0,   0,   0,   0 },
      {  24,  32,  60, -47,  -9,  -5,  -2,  -2,  17,  57, -21,   1,  -9,  -3,  -2,  -1,  36, -23,  -8,  14,   1,  -1,  -1,   0, -46, -11,  15,   8,   4,   2,   1,   1,   0, -11,  -6,   4,   2,   1,   1,   0,  -3,  -2,  -1,   0,   1,   1,   0,   0,   1,  -2,  -2,   1,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 },
      {   4, -40,  -8,   2,   9,   2,   2,   1,  47,   0,  46, -20, -18,  -1,  -3,  -1,  15, -45, -14,  52,  -5,  -1,   0,   1, -14,  30, -45, -10,  19,   0,   1,  -1,  -8,  13,  15, -16,  -4,   2,   0,   0,  -3,  -1,  -2,   5,  -1,  -1,   0,   0,  -2,   3,   1,  -2,   0,   0,   0,   0,  -2,   1,  -1,   0,   0,   0,   0,   0 },
      {  -8, -13,   6, -27,  13,   1,   2,   0, -14, -13,  67, -19, -20,   3,  -2,   1,  -8,  56,  13,  19, -16,  -9,  -2,  -2, -35, -40,   7,  26,  14,   0,   1,   1,  19, -16, -28,   4,   5,   4,   0,   0,   3,   9,  -6,  -6,   1,   0,   1,  -1,   4,  -1,  -1,   1,  -1,   0,   0,   0,   0,   1,  -2,   0,   1,   0,   0,   0 },
      { -21,  -1, -27, -16,  12,   5,   2,   1,  -4,  54,  19,   8, -23,  -6,  -3,  -2, -28,  24, -48,  10,  30,  -7,   3,  -1,   8,   7,  28, -50,  -1,   7,  -2,   1,   1, -23,  31,  19, -19,   0,   0,   0,   3,  -7, -12,   6,   7,  -2,   1,   0,   2,  -1,   3,  -1,  -1,   1,   0,   0,   1,  -3,  -2,   1,   1,   0,   0,   0 },
      {   8,   0,   0, -58,  20,   4,   3,   0,  15,   3,  -2, -21, -14,   6,   0,   1,  32,  32,  11,  -2, -24,  -6,  -2,  -1,  72,  10,  -6,  16,  -8,  -7,  -4,  -1, -49,   6,  15,  12,   7,  -1,   0,   0,  -4, -13,   0,   4,   3,   2,   1,   1,  -7,   0,  -2,   2,   0,   0,   1,   0,  -1,  -2,   0,   1,   1,   0,   0,   0 },
      { -17, -26, -32, -67,  42,  10,   7,   2, -22, -22, -43,  21,  -7,   9,   2,   2, -11, -26,  32,  16,  -9,  -4,   0,   0, -26,  22,  12, -10,   0,  -1,   0,   0,  29,   5,  -4,   0,  -7,  -2,  -1,  -1,   0,   6,   4,   0,  -1,  -1,   0,   0,   4,   0,   0,   2,  -1,  -1,   0,   0,   0,   1,   1,   0,   0,  -1,   0,   0 },
      {   3,   3, -29,  -5,  -2,   6,   0,   2,   7,   8, -16,  53, -21,  -8,   0,  -1,  33,  18, -14,  14,  26, -13,  -2,  -3,  16, -59, -30,  -7,  23,  15,   0,   3,   5,  36, -17, -32,  -5,   3,   2,  -1, -11,   5,  26,  -7,  -6,  -2,   0,   0,  -1,  -2,   2,   4,  -1,  -1,   0,   0,  -3,   0,   3,  -2,   0,   0,   0,   0 },
      {  -1,  25,  12,  12,  -1,  -7,  -2,  -2, -26,   3, -39, -27,   5,  13,   3,   2, -14,  48,   0,  32, -23, -15,   1,  -3, -14,  23, -28, -10,  44,  -8,   0,  -1,   4,  -7,  32, -43,  -9,  14,  -2,   1,   4, -12,  13,  21, -13,  -4,   2,   0,   2,  -4,  -4,  -1,   5,  -1,   0,   0,   2,  -3,   2,   2,  -1,   1,   0,   0 },
      { -10,  -5,  -1,   6, -33,  10,  -1,   3,  -7,   3,   9,  61, -21, -17,   0,  -2,   0,   8,  53,   6,  14, -15,  -5,  -2, -17,  27,  -8,   3,  17,   2,  -2,  -2, -54, -33,   6,   1,  17,   7,   3,   1,  25, -11, -19,  -7,  -1,   5,   1,   1,   1,   7,  -3,  -5,   0,  -1,   1,   0,   5,  -1,  -1,  -3,  -1,   1,   0,   0 },
      { -10, -12,  -8, -20,  -6,  10,   2,   2, -15, -20,  -8, -26,  27,  -1,   2,   1, -14, -14, -21,   3,  14,   9,  -1,   1, -34, -50,   8,   7, -10,  11,   5,   2, -65,  24,  29, -12, -11,  -5,   3,   0,  41, -11,  -4,  12,  -7,  -5,  -2,   0,   3,  11,  -5,   0,   2,  -2,  -1,  -1,   5,   1,   2,   0,  -1,   0,  -1,   0 },
      {   7,  11,   8,  32,  89, -39,  -4,  -7,   1,   5,  -5,  25,  -7,  20,  -6,   0,  -4, -13, -13,  -4, -22,   9,   8,   1,  -1, -18,  -3,  -1,  15,  -3,   0,   2, -33, -25, -10,  -1,   4,  10,   2,   2,  21,  10, -11, -17,  -4,   0,   1,   0,  -1,   5,   8,  -2,  -7,   0,   0,   0,   3,   1,   0,   0,   0,  -2,   0,   0 },
      {  -8,  -5, -22,   3, -30,  16,   1,   3,  -5,   2, -28,  -9,  -8,   6,   4,   1, -14, -12, -49,  43, -13, -12,   7,   0,  17,   6,  30,  34,  13, -27,  -5,  -3, -11, -13, -31,  -5,  43,   8,  -4,   0,  11,  18,  -2, -36,  -5,   9,   1,   0,  -2,   0,  16,   0,  -8,  -2,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -10,   2, -11, -23,  10,  -2,   3,   0,   1,  30,  22,  25,  20, -11,  -7,  -3, -20,  18, -27, -40, -10,  30,   4,   2, -26,  33, -32,  25, -19, -21,  10,  -1,  -2,  20,   2, -25,  35,  -8,  -5,   0,   2, -16,  30, -14, -18,  13,  -1,   1,   4,  -7,   0,  16,  -2,  -5,   2,   0,   1,  -3,   0,  -2,   1,   1,   0,   0 },
      {   2,  -8,  -7, -27,  -7,   4,   5,   1,  11,   1,  -6, -10,  57, -20,  -4,  -1,  10,   6,  -2,  -2,  16,  18,  -9,   0,  16,  -4, -40, -13,   0,  17,   6,   0,  10, -65, -13,  -8, -11,  11,   5,   5,  19,  43,  -9, -10, -21,  -7,   1,  -1, -10,   6,  19,   3,  -1,  -5,  -1,   0,   0,  -1,   0,   4,  -3,   0,   0,   0 },
      {  -1,  -6, -10,   1,  -9,  -8,   4,   0,   3,   0,  -5,  39,  35,  -7, -11,  -2,   4,   2,  -7,  35, -39,  24,   6,  -2,   5, -35, -10,  11,  -9, -25,  12,   3,  21, -15,  39,  19,  14, -10, -15,   2, -25, -22, -34,  17,  19,  11,   2,  -1,  10,  13, -15, -27,   7,   4,   1,   0,  -3,   3,   6,  -5,  -6,   1,   0,   0 },
      {   2,   2,  -3,  -4, -13,  27,  -4,   3,  -2,  -5, -16,  -6, -59,  26,  11,   2,  -1,  -8, -19, -46,  -9,  -6,  13,   3,  -8, -15, -52,  18,   1, -16,   4,   3,  21, -19,  26,  18,  -3, -17,  -7,   0,  31,  18, -18,   5,  12,  -3,  -6,  -1,  -9,  12,   8,  -8,   4,   0,  -1,  -1,   1,  -2,   6,   2,   0,   0,  -1,   0 },
      {  -5,   2,  -3,   2,  10, -10,   2,  -2, -17,  -9, -23, -22,  -8,  -3,   5,   3, -11,   8,  -8,  12,  37,   1, -12,   0, -20,   4, -42,  14,  -9,  20,   7,  -3, -29, -19,  -6,  31,   6, -19,   8,   0, -63,  -5,  12, -25,  29,   8,  -4,   2,  30,  -6,   1, -13, -19,   7,   1,   0,   0,   6,   3,   3,  -2,  -4,   0,   0 },
      {   3,  14,   4,  11,   4,   1,  -3,  -1,  -2,  10, -10,   0,   6,  -3,  -2,   0,   6,  32,   9,  40,  -8,  10,  -7,  -4,  -7,  18, -16, -21, -35,  10,  17,  -1,   5,  40, -23,  15,   1, -41,   4,   1,  52,  10, -15, -16,  34,   2, -13,   0, -22,   1,  21, -13,  -8,  11,   2,  -1,   2,  -9,   1,   6,   0,  -3,   2,   1 },
      {  -8,  -4,  -3,  -3, -35,  -4,   3,   2,  -4,   5,   5,   9,  -6,  34, -12,  -1,  -6,  -2,  10,   0, -47,  22,  20,  -4, -10, -12,  -2, -47,  -5, -17,  14,   7, -37,  -4, -10,  12, -13, -10,  -3,   2, -26,  42,  34,   1,   3, -20,  -3,  -4,   6, -24,  10,  16,  10,   2,  -5,   1,   2,   1, -10,   2,   4,   0,   1,  -1 },
      {  -7, -11, -13, -15, -31,-101,  32,   0,  -3,  -4,  -7,  -9, -23,   0, -17,   5,   3,   1,   3,  -6,  21,  22, -10,  -4,   2,   4,   3,  12,  15,  -7,  -1,  -1,   9,  12,  11,   4,   1,   8,  -6,  -1,  15,  10,  -7,  -1,   4,   2,   1,  -1, -11,  -1,   7,  -1,   1,   5,  -1,   1,  -1,  -3,  -1,   1,   0,   1,   1,   0 },
      {   0,   0,  -2, -16, -12,  -1,   0,   2,   2,   2,   3, -13,  26,  16,  -2,  -3,   2,   1,   3, -29,  -7, -17,  19,   2,  -8,  -3, -16, -50,  36, -12, -18,   8,  -2,  31, -25,  19,  15,  24, -10,  -7,  -3, -27, -33, -25,   1,  14,  14,   1,  20,  27,  17, -19, -27,  -3,  -1,   1,  -4,  -2,  12,  15,  -5,  -7,   0,   0 },
      {   6,  -1,  10,   7,   3,   3,   3,   0,   0, -16,  -3, -20, -29,  -5,  -1,   4,  12,  -2,  21,  11,  15,  24,  -1,  -7,   7, -22,   9, -43, -20, -15,  23,   6,  12, -24,  17, -17,  35, -15, -26,   8,  10, -19,  18, -15, -23,  43,  -4,  -5,  11,  10,  -4,  26, -29, -16,  12,  -1,  -2,   3,  -4,   7,  14,  -5,  -4,   1 },
      {   0,   3,  12,   2,  14,   7,  -2,  -3,  -4,  -2,   5,   0,  11, -45,  13,   1,  -3,   0,  12,  -7,  -3, -23,  -4,   4,  -8,  -9,   7,   0,  15, -11, -13,   1,  -4,  26,  57,   8,  14,   6, -21,  -4, -18,  58,  -4, -22,   7,  22,   3,  -6, -23, -28,  28,   2, -12,   5,   7,   2,   5,  -8, -12,  10,  -2,   2,   1,   1 },
      {   0,   0,   3,  -1,   2,   4, -19,   4,   5,   7,  11,  10,  25,  81, -23, -12,   9,   8,  18,  14,  53, -10,  24,  -8,   1,   5,   5,  25, -17, -22,   4,   8,   5,   4,  27, -14, -10,   5,  -3,  -1,  -5,  13,  -9, -18,   0,   3,   4,   1,  -3,  -7,   7,   2, -12,   1,   3,   1,   1,  -4,  -5,   4,   0,  -6,   0,   1 },
      {   2,   8,   5,   6,  -4,  14,  18,  -5,  -2,   5,   0,   5, -23,   0,  -8,   7,  -1,   8,   4,   2, -18,  28, -23,  -1,   4,   9,  13,  12, -16,  20,  10, -11,   1,  21,  21, -24, -18,   9,  19,   0, -14,  28, -36, -24, -21, -17,  15,   5,  54,  31,  20,  -3,  -8, -25, -10,   1, -20,  -7,  11,  15,  10,   1,  -3,  -1 },
      {  -3,   0,   1,   4,  -5,  14,  25,  -9,  -6,  -4,  -2,   5,  -7,  20, -12,   7,  -5,  -4,  -2,   0, -13,  13, -18,   2, -10,  -8,  -1,   1,  -4,  22,   5, -10,   0,  -5,   4,   2,  -2,  10,  17,  -2, -32, -41,  -9,  -9, -14,  -2,  13,   8, -80,   2,  40,   6, -10, -16,   0,   1,  25, -11, -12,  11,   7,  -6,  -3,   0 },
      {   2,   1,   2,   4,  -3, -13,  73, -21,   5,   4,   8,  10,  16,  16,  23,   3,  -1,  -4,  -2,   2, -12, -68,  -9,  21,  -3,  -7,  -7,  -6, -42,  -5, -16,   1,   2, -14,   7, -13,   3,  -4,  -3,  -1,   6,  -4,  15,  -4,   6,  -5, -15,   2,  13,  14,   4,  10,  -3,   7,  -5,  -5,  -7,  -1,  -1,   4,   8,   0,   0,  -1 },
      {   3,   9,   5,   8,  18,  32,  65, -17,  -3,   2,  -4,  -8,  -4,   0,  -9,  10,   1,   6,   2,  -1,  34,  19,  -9,  -7,  -3,   7,  -3,  -6,  10, -41,  18,  10, -15,  13, -29,   0,   2,   5, -29,   4, -11,  20, -23,  26,  -2,   3,  13, -11, -10,  -7, -17, -17,  24,  -9,  -1,   6,   8,   4,   7, -14,  -7,  10,  -4,  -1 },
      {  -3,  -4,   0, -12,  -2,  -8, -12,   0,   0,   0,   8,  -5,  -7,  -3,  15,  -3,  -3,  -6,   1, -21,   0,  -1,   9,   0,   3,   1,  22, -10,  -1,  -1,  18,  -1, -11, -13, -11, -75,   6, -24,   0,  19, -29,   8, -20,  12,  29,   3, -30,  -2, -15,  18,  18, -35,  -4,  27,   4,  -8,   7,  -5,  16,  10, -23,  -4,   4,   2 },
      {   4,  -1,   3,  -1,  -3,  25,  36,  -7,   7,   1,   7,   4,  14,  20, -19,   9,   2,  -9,  -1, -15,  -9,  16, -29,  -5,  13,  -2,  15,   5,  54,  15,  16, -15,  11,  -4,  16,   0,  -5, -18,  16,  11,  24, -12,  14, -25,  21,   5, -19,   6,  20, -37, -13,  -4, -24,  29,   4,  -1,   5,  15,  -8,   6, -16, -14,   7,   0 },
      {   0,   2,   5,   4,   6,   8,   3,  -4,  -2,   1,   7,   3,   9,  -5, -14,   0,  -1,   0,   5,   4,  10,   3,  -3,  -6,   1,   6,  14,  12,  30, -16,   9,   0,  -2,   0,   9,  30, -12, -16,  -5,   5,  -2,   7,  60,   0,  -6,  -2, -15,  -3, -12,  79,  14, -15,   6,   7,  -1,  -8, -17, -33,   8,  10,   0,   5,   7,   2 },
      {  -2,   2,  -3,  -2,  -6,  -3, -10,   7,  -1,   7,   1,  -1,  -1,  25, -35,  10,  -5,   0,  -7,  -6, -15, -19, -34,   5,  -3,   4,  -3, -12, -14,  36,  -3, -16,  -2,   7,   2,   1,  27,  13,   1,  -8,  -3,  40,   9,  13,  -6,  28,  10, -11, -13,  26, -49, -23,  -6,  -4,  34,   1,  25,   8,  14, -16,  -6, -17,  -3,   7 },
      {   4,   2,   3,   2,   5,   5,   3,  18,  -1,  -7,  -5,  -7,  -9, -33, -80,  19,   3,  -4,  -1,  -5,  -3, -31,  22, -18,   1,  -1,  -2,   4, -20,  -6,  13,   5,   5,  -4,   0,  -9,  -7,  15,  -7,  -4,   6, -11,   9,  -1,  11, -14,   5,   1,  23,   3,  23,  -6,   0,  11, -16,  -2,  48,   8, -16,   8,  -5,   7,   7,  -4 },
      {   3,   2,   1,   0,  -1,  -6,   7,  16,   6,   5,   4,   4,   1,   4,  52, -14,   6,   5,   4,   7,   1,  21,   0,   7,   4,   3,   5,   3,   4, -11, -13,   6,   6,   4,   4,   3, -15,  -6,   2,  -2,   2,   4,  13, -15,  -7,  -5,   2,   3,  17,  21,  -7, -15,  -3, -10,  -4,   3,  97,  12, -14,   3,  -7,  -2, -10,  -1 },
      {  -3,  -4,  -4,  -4,  -8,  -4,  -9, -99,  -1,  -1,  -2,  -5,  -3, -15, -10,  -7,   2,   3,   2,   0,  -3,  -5,  48,  15,   2,   6,   3,   3,   5,  16,  21,   6,   3,   4,   6,   2,  -7,  -6,   2,  10,   1,   4,   0, -23,  -6,  -5,   4,   7,  -9,   9, -24,  -5,  -3,  -8,  -3,   5,  12,  20,   1, -10,   2,  -6,  -7,  -1 },
      {   0,   1,   2,   3,  -4,   9,  18,  12,  -2,   0,   2,   2, -13, -13,   4,   1,  -1,   2,   2,   5, -11,  25,  22,  -2,  -2,   3,   1,   9, -25,  -1, -29,  22,  -4, -12,   0, -14, -45,  29, -15, -20,  10,  -6,  31, -22,  16,  15,  24,  -7, -20,  -9, -15, -32, -26,   6,   8,  11, -24,  14,  32,  15, -20, -21,   1,  -2 },
      {   1,   4,   5,   1,   6,  -1,  22,  44,  -2,  -2,   0, -12,   1,  -7,  -2, -11,   0,   1,   5,  -2,   1,  -5,  26, -21,  -2,  -3,   1, -16,   2,   7,   5,   4,   5,   3,  19,  -3,  24,  -9,  24,   3,  -9,   5,  -8, -52,  -7, -26,  -8,  27, -29,  10, -45,   4,  24,   1, -29,  -6,  -9,  10,  25, -20,  -3,  15,   2,  -7 },
      {   2,  -2,   3,  -1,   5,   8,  -5, -29,   1,  -5,   2,  -5,  -9,  13,   1,  -5,   5,   1,   9,   4,  15,  26,  -8,  15,   0, -10,   2, -17, -13,   7, -26,  -9,   8,  -5,  15,  -3,  60,  22,  15, -21,  12,   0,  28,  11,   9, -27,  19,  15,  11, -13,  20, -39,  17,  14, -25,  10,  -2, -16,   0,   3, -27,  21,   7,  -5 },
      {   2,   3,   2,   3,   3,   4,  37,  20,   0,   1,   0,   5,   2,   5,   6, -13,   1,   2,  -1,   4,   0,   9,  63, -20,   3,   9,   5,  16,  12,  65,  -6,  15,   2,   2,   1,   8,  31, -15, -18,   2,   0,  10,   0,  41, -11,  -3,   0,  -4,  10,  -2,  15,  -4, -19,  -3,   9,   0,  -2,  -7, -10,   3,   6, -14,  -2,   6 },
      {   1,  -4,  -2,  -2,  -1,  -8,   1,   3,   2,  -5,   1,  -1,   0,  -3,   8, -28,  -1,  -9,  -6,  -5,  -7, -24,  -3, -16,   2,  -2,   0,   1,  -4,   5,  45,   6,   1,  -5,  -5,   3,  -7,   6,   3,  12,   7,  -7,   6, -11,  -3,   4,  16,  -1,   8, -32, -24, -28,   7, -22,   4,  22,   6, -80,   4,  30,  -3,   5, -23,   5 },
      {   1,   1,  -2,   1,   0,  12,   8, -10,  -2,  -3,  -8,  -4,  -5,  -4, -16,  -9,   1,   3,  -2,   1,  -1,  27,  10,  16,  -1,  -1,  -5,  -2,   1,   6, -42,  12,  -1,  -4,  -6, -17,  -6,   8,  -1, -15,  -7,   2, -20, -15,  28,  11, -10,  -1,   3,  16, -36,  37,  10,  24,   9, -14,   9, -63, -40,   0,   4,   0,  24,   7 },
      {   0,   2,   1,   1,   2,   4,   8,  -1,  -4,  -4,  -5,  -3,  -6, -12, -21,-104,  -2,  -1,  -3,  -2,  -3,  -3, -22,  -1,   0,   0,   0,   0,   1, -18,   4,  22,   0,   1,  -2,  -1,   4,   6,  28,  -8,   0,   8,   1,  22,  11,   1,   2,  16,  -2,   9,   6,   9, -20,   0,   3,   1,  -3,  25,  -4,   2,   7, -13,   2,   5 },
      {   0,   1,   1,  -2,  -2,   1,  -1,   4,   0,   0,  -1,  -4,  -3,  11,  16, -23,  -2,  -1,  -2,  -5,  -9,  -1, -13, -15,   0,   4,   0,   0,  -6,  29,  27,  17,   1,   3,   5,  -4,  -8,  32, -33,  12,   3, -13,   7, -33,   8,  12,   1, -25,   5,  13,  20, -19,  37,  13,  11,   5, -17,  19, -50, -45, -10,  20,   8,   7 },
      {   0,   2,  -1,   1,   0,  -1,   3,   8,  -2,   0,  -4,   0, -10,  -9,   7,   6,   1,   4,  -1,   4,   3,  -1,  11,   8,  -3,  -2,  -7,  -3, -18, -21,   9, -13,  -1,   9,  -5,  14,  -6,  15,  23,  16,  -4,   4,  -8,   6, -63,  -1, -21,  16,  -5,  -8, -18, -44, -20,  33,   4, -21, -16,   1, -53,  12,  -8, -20,  24,   3 },
      {   1,   0,   1,   5,  -1,   2,   1,  -6,  -1,  -2,  -1,   0,  -8,   4,  -1, -20,   1,   2,   2,   9,  -2,   4,  -2,  16,  -1,   1,  -2,   4, -10,  14,  11,   9,   1,   7,   3,  23,  -2,  31, -18,  16,  -3,  -6,  -7,  -3, -32,   3, -36, -10,   9, -15,  15,  28,  12,  26,   7, -31,  22, -13,  66,   2, -39,   2,  22,   6 },
      {   0,   0,   1,  -1,  -1,  -3,   0,   4,  -1,  -3,  -1,  -3,  -3,   2, -33, -20,   0,   0,   1,  -1,  -3,  -6, -11,  34,   2,  -1,   3,  -1,   8,   7, -41,  20,  -2,  -7,  -2,  -6, -13, -69, -17, -10,  -4, -10,  -7, -19, -48,  13,   6, -10,   5, -19,  11, -32,  31,  -1,   3,  -1,  -3,   2,   3,   0,   3,  12,  -8,   1 },
      {  -1,  -1,  -2,  -1,  -2,  -1,  -7, -34,   1,   1,   0,   2,   0,   4,   7,   0,  -3,  -4,  -5,  -2,  -6,  -5, -21,-101,  -4,  -1,  -6,  -1, -11,  -1, -44,  13,   0,   0,  -1,   1,   1, -14,   1,  23,  -2,   1,  -5,   1, -17,   6,   7,  -5,   0,   1,  -4,   4,   6,  17,  -6,   2,   7,   3,   4,  18,  -5,   0,  16,  -6 },
      {  -2,   0,  -3,   0,  -1,   1, -14,   2,   0,   4,  -1,   5,   1,   8,  -1,   3,  -3,   1,  -4,   0,  -1,   1, -29,  20,   1,   7,   1,   6,   4,  19,  11,  46,  -1,   3,  -2,   1,  14,   8, -48,   7,  -2,   3,  -4,   2,   3, -32, -13,  -6, -14,   4, -31,  -1, -16,   2, -35,  -5,  -3,  18, -13,  59,  17,  12,   7, -32 },
      {   0,   1,   0,   1,   3,   1,   7,   3,   0,   0,  -1,   1,   1,   4,   2, -13,   1,   2,   0,   2,   5,   2,  19,   3,   0,   0,  -1,   0,   1,   9,   5, -45,  -2,   1,  -3,   2,   1,   9,   5,  -4,   2,   2,   2,  -2,  10,  44,   1,   4,   2,   6,  -2,   9,  67,  -2,  -5,   0,  -3,  31,  -7,  74,   0, -10,   3,   1 },
      {   1,   0,   1,   0,   3,  -1,  -1,   1,  -1,  -2,  -1,  -2,  -1,  -9,   1, -18,   1,   0,   2,   0,   6,  -1,  -4,  -5,  -1,  -3,  -1,  -3,  -2, -10,  -3, -47,   0,  -3,   1,  -3,  10, -11, -42,  -4,  -2, -11,  -4, -18,  -3, -76,  20, -20,  -3,   5,  -8,   9,   2,  17,  43,   7,  -2,   5,  -1,  15,  -2, -32,  -6,  29 },
      {   0,   0,   1,  -2,   0,   6,   1,  -3,   0,  -1,   1,  -2,   1,  -1, -12, -10,   1,   0,   2,  -1,   2,  13,   4,  -6,  -2,  -2,  -2,  -3,  -3,  -1, -33, -19,   1,   1,   2,  -2,   0,  29,  -7,  38,   4,  -3,   9,  -6,  18,  -7, -48,  -1,   1, -18,   2, -38,   4, -26,  12, -32,   5,   0,  13,  -4,  69,  11, -17,  10 },
      {   0,   0,   1,   0,   3,   0,   0,  -1,   0,  -1,   1,  -1,   1,  -5,   2,  29,   1,   0,   1,   0,   5,   1,  -3,   3,   1,   0,   2,  -2,   3,  -9,   5,  70,   5,   2,   9,   2,  22,   7,  50,  -2,   0,  -5,   1,  -7,   5, -17, -11, -35,   4,  -1,   7,  -2,  35,  -8,  29,  -3,  -2,   8,  -1,  17,   5, -45, -10,  27 },
      {   0,   1,   0,   2,   1,  -2,   0,   6,  -1,   0,  -2,   0,  -3,  -2, -16,  -3,   0,   2,   0,   3,   0,  -1,   0,  10,  -3,  -1,  -3,   0,  -4,  -3, -31,  -4,   0,   4,   1,   4,   1,  -3,  -3,  57,   1,   0,   2,   1,  12,  -8, -21,   2,  -2,   4,  -3,  11,  -5, -61,  -2,   9, -13,   2, -26,   4, -76,  -8, -15,  -1 },
      {   0,   1,   0,   1,   1,   2,   0,  10,  -1,   0,  -1,   0,  -1,   0,  -1,   1,   0,   1,   1,   1,   1,   3,   0,  27,  -1,   0,  -2,   0,  -3,   1,  -7,   6,   1,   3,   2,   2,   3,   2,   6,  83,   3,   3,   5,   3,   6,   9,  63,   4,   1,   7,   3,  11,   5,  49, -16,  -3,   2,  -4,   8,  -6,  18, -16, -26,  -3 },
      {   0,   0,   1,   0,   1,   0,  -8,  -3,   0,   0,   1,  -1,   0,   0,   0,   4,  -1,  -1,   0,  -1,   0,  -1, -16,  -7,   0,   0,   1,   0,   0,   3,   1,   6,  -2,  -1,  -1,  -1,  -1,   3, -38, -15,   0,   1,   1,  -1,   0,   5, -16,  58,   5,  -2,  10,  -5,  25,  -6, -34,  -1,   0, -15,   0, -25,   6, -87, -11,  -6 },
      {   0,   1,   0,   1,   0,   4,   1,   0,  -1,   0,  -2,   0,  -3,   0,  -7,   8,  -1,   1,  -1,   2,  -1,   7,   0,  -4,  -1,   1,  -2,   1,  -4,   1,  -8,  12,  -1,   3,  -1,   4,  -2,  14,  -3,  -6,  -4,   1,  -6,   1, -12,  -1, -37,  43,   2,   9,   3,  11,   3,  40,  40,  54,  -2,   5,  -2,  10,  -3,  26, -71, -12 },
      {   0,   1,   1,   1,   1,   3,   1,   0,  -1,  -1,  -1,  -1,  -2,  -1,  -6, -13,   1,   1,   1,   1,   1,   4,   4,   1,  -1,  -1,  -2,  -1,  -2,  -1,  -8, -25,   1,   2,   2,   2,   2,   7,  12,  -7,  -4,  -2,  -5,   0, -11,   7, -34, -78,  -4,   3,  -4,   2,  -9,  18, -59,  46,   0,  -7,   1, -13,   6, -25, -27, -10 },
      {   0,   0,   0,   0,  -1,   0,  -1,  -6,   0,   1,   0,   1,   0,   1,   1,  -1,   0,   0,  -1,   0,  -2,   1,  -3, -15,   0,   1,   0,   1,   0,   2,   4,  -1,   0,  -1,  -1,   0,  -2,   0,  -3, -28,   0,   1,   0,   1,   0,   4,  10, -10,  -3,   0,  -6,   4,  -9,  10, -26, -79,  -7,   2, -13,   5, -24,  11, -77,  30 },
      {   1,   0,   1,   0,   1,   0,   4,  -2,   0,  -1,   0,  -1,   0,  -2,   0,  -9,   1,   0,   1,   0,   2,   0,   6,  -4,   0,  -1,   0,  -2,   1,  -3,   2,  -8,   1,   0,   2,  -1,   4,  -1,  11,  -8,   1,  -3,   1,  -3,   2,  -8,   7, -21,   4,  -1,   5,  -1,  10,  -2,  36, -34,  -1,  -8,  -3,  -8,  -5, -18, -16,-110 },
    },
    {
      {  89, -60,  -1,   0,   0,  -1,   0,   0, -61,  25,  13,  -1,   1,   0,   1,   0,   0,  13,  -6,  -2,   0,   0,   0,   0,   0,  -2,  -2,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  61,  32, -51,   2,  -6,   1,  -2,   0,  35, -56,   8,  12,  -1,   2,   0,   1, -55,   8,  31,  -5,   1,  -1,   1,   0,   1,  13,  -6,  -5,   1,  -1,   0,   0,  -5,  -1,   1,   1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
      {  -8, -80,  44,  16,   2,   2,   1,   1,  67,  -7,  -8, -12,  -4,  -1,  -1,   0, -43,  25,   0,  -6,   4,  -1,   1,   0, -11,  14,  -2,   0,   1,  -1,   0,   0,   0,   3,  -5,   0,   0,   0,   0,   0,  -1,   1,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
      { -43, -15, -11,  22,   1,   4,   0,   2, -40, -45,  60,   4,  -1,   1,   1,   0,  -1,  60,   1, -29,   0,  -3,   0,  -2,  31,   0, -32,   3,   3,   0,   0,   0,   1,  -5,   2,   5,  -1,   0,   0,   0,   5,   1,  -3,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,   0 },
      {   4,  38,  53, -44,  -5,  -5,  -1,  -2, -19,  19,  -9, -19,  10,  -1,   2,   0, -66,  18, -15,  16,  10,  -1,   2,   0,  52,  -7, -19,  10,  -6,   0,  -1,   0,   4, -16,   0,   6,  -2,   1,   0,   1,   4,  -2,   2,   1,  -1,   0,   0,   0,  -1,  -3,  -1,   1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
      { -22, -22,  20, -13,   9,  -1,   3,   0, -44, -26, -30,  38,   7,   3,   1,   2, -13, -30,  62,  14, -15,   1,  -2,   0,  11,  48,   4, -36,  -3,  -1,  -1,  -1,  13,   1, -22,   0,   6,   0,   0,   0,   1,   1,   2,   1,  -1,   0,   0,   0,   3,   0,  -2,   0,   0,   0,   0,   0,   1,   1,   0,  -1,   0,   0,   0,   0 },
      { -28, -17, -62,  48,   7,   5,   1,   2,  -7,  59, -13, -12,  -6,  -3,  -1,  -1, -47, -14,  23,   6,   1,   2,   1,   0,  39, -17,  10,   2,  -7,   0,  -2,   0,   8,  -4,   3,  -8,  -1,   1,   0,   0,   1,  -2,   2,   0,   1,   0,   0,   0,   0,   1,   1,  -2,   0,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0 },
      {   3, -15, -18, -20,  17,   0,   3,   0,  13,  -7, -66,  42,  13,  -2,   1,   0,  28,  53,   3,   7, -24,  -5,  -3,  -1,  32, -37,  -4,  -3,  -6,   6,   0,   1, -27, -16,  26,   5,   1,   1,   0,   0,  -2,   8,   6,  -8,  -1,   0,   0,   0,  -5,  -1,   2,   0,   1,   0,   0,   0,  -1,   1,   2,  -2,   0,   0,   0,   0 },
      {  10,  17,   5,  -8,   5,  -5,   0,  -2,  20,  54,   3,  10, -22,  -4,  -4,  -1,  14,  11,  39, -53, -12,   6,  -1,   0, -11,  14, -56, -14,  29,   1,   2,   0,   6, -29, -13,  34,   3,  -3,   0,   0,  -6,  -5,  11,   1,  -5,   1,   0,   0,   0,  -4,   0,   1,   0,   1,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0 },
      {  23,   7,  32,  51, -37,  -2,  -7,   0,  18,  -8,   6,  -9, -10,   6,   0,   1,  33, -27,   6,   1,   0,   4,  -2,   1,  68,   4,  -6, -27,  -4,   0,  -2,   0, -48,  -5,  -4,  -6,  15,   1,   2,   0,  -5,   8,   1,   4,   2,  -3,   0,   0,  -7,   1,  -3,  -2,   1,   0,   1,   0,  -1,   1,   0,   1,   1,   0,   0,   0 },
      {  -5, -31, -25, -63,  45,   6,   7,   1,  25,   3,  47,  -7,  -6,  -8,  -3,  -2,  11, -42,   2,  24,  -6,   1,   0,   1,  34,   6, -27,   7,   1,  -2,   0,  -1, -29,  10,   9,   2,  -3,  -1,   1,   0,  -4,   6,  -3,   0,   1,   1,   0,   0,  -3,   2,   0,   1,  -2,   0,   0,   0,  -1,   1,  -2,   0,   0,   0,   0,   0 },
      {   1,  -6, -32,   2,  -3,   7,   0,   2,   3,  -7, -26, -50,  32,   7,   3,   1,  28,  16, -17,  26,  22, -17,  -1,  -3,   7,  61, -20,   6, -19, -10,   1,  -2,  11, -40, -18,  18,  -4,   7,   1,   2, -11,  -9,  25,   8,  -6,   0,  -1,   0,   0,   1,   4,  -5,  -2,   1,   0,   0,  -3,  -1,   3,   1,  -1,   0,   0,   0 },
      {   4,  14,  -3, -14,   3,   3,  -1,   1,  17,  27,  34, -10,  11, -13,  -1,  -3,   8,  48,  11,  30, -36, -11,  -1,  -2,  -4,   0,  31, -50, -21,  16,   1,   2,   0,   9, -44, -21,  30,   4,   0,   0,   3, -18, -10,  23,   4,  -4,   0,   0,  -2,  -2,   2,   1,  -2,   0,   0,   0,   0,  -4,  -2,   2,   0,   1,   0,   0 },
      { -11,  -9,  -4, -18, -31,  17,   0,   4, -15,  -7, -13, -60,  32,  12,   2,   1,  -2,   0,  52,  -3,   6, -10,  -4,  -1, -27, -37,  -8,   0,  14,   1,   4,   1, -44,  20,   1,  21,  14,  -8,   1,  -2,  25,  13,  -6,   4, -12,  -4,   0,  -1,   1,  -2,  -4,   2,   0,   1,   1,   0,   5,   1,   0,   2,  -2,  -1,   0,   0 },
      {  -8,  12,   0,  21,  27, -17,  -1,  -3, -15,  16,   2,  10,  -2, -11,   3,  -1, -23,   7, -25,   9,  -8,   3,   5,   0, -34,  45,  -1, -15,  -4,   2,   1,  -1, -74, -17,  29,   6,   1,   5,   1,   1,  46,   8,  -5,  -4,  -4,   0,  -2,   0,   7,  -9,  -4,  -2,   0,   2,   0,   1,   6,  -2,   1,   0,   0,   1,   0,   0 },
      {   5,  11,  12,  -7,  10,  -7,   1,  -2,   5,  19,   4,  -8,  -9,  12,  -3,   1,  21,  26,  54, -14,  28, -15,  -8,  -3,   5,  26,  16,  37, -35, -24,   4,  -2,  -9,   5,  17, -44, -28,  18,   5,   1,   1,   5, -34, -14,  25,   6,  -1,   0,   1, -11,  -6,  13,   2,  -2,   0,   0,  -1,  -1,  -2,   0,   0,   0,   0,   0 },
      {  -4,  -2, -20, -26,   5,   3,   3,   1,  -5,  -1, -31,  -7, -23,  14,   3,   2,   0,  13, -28, -47,  28,  22,  -3,   2,  26,  25,  28,  -9,  29, -17,  -8,  -2, -21,  41, -30,   0,   5, -15,   5,  -1,  13, -27, -17,  29,  -1,   0,   2,   0,  -5,  -6,  15,   8,  -8,  -1,  -1,   0,   1,  -2,   2,   0,  -1,   1,   0,   0 },
      {  13,  14,  18,  32,  84, -43,  -6,  -8,  -5, -22, -14, -44,  11,  -5,  10,   2,   7,  -3,  13, -11, -28,  15,   2,   0,   6, -16,   1,  17,   2,   1,  -3,   0,   7,  13, -19,   5,  -1,  -3,   1,   0,  -3,  -1,   5,   8,  -7,   0,   0,   0,  -4,   1,   2,  -3,  -4,   3,   0,   0,   0,   2,   0,   1,   0,   0,  -1,   0 },
      {   7,   8,   4,  22,  -7,   9,  -6,   1,  10,  16,  -2,  24,  56, -23,  -5,  -4,   6,   0,  -4,   9,  -8, -17,   4,   0,  11,  17, -39,  11,   3,  -2,   6,  -1,  10,  70,   3,   6, -25, -11,   0,  -4,  21, -32, -10, -11, -15,  13,   1,   2, -11, -10,   8,   1,   5,   3,  -1,   1,   1,  -2,   3,  -4,  -3,   1,   0,   0 },
      {   0,  -1,  -1,  10,  -4, -22,   8,  -3,  -3,  -5,  -9,   8, -62,  30,   6,   1,   6,  12,   9,  63,  16,  -1, -16,  -3, -13, -18, -44,  11,   6, -13,   4,   3,   1,  15, -19,  -9,   1,   9,   4,   0,  32, -24,  10,  13,  -3,  -1,  -5,   0, -11,  -4,  17, -12,  -4,   3,   0,   1,   1,   4,   1,  -4,   5,  -1,  -1,   0 },
      {   1,   2,  11,   5,  13,   6,  -5,   0,   5,   6,  25,  36,  18,  -6,  -8,  -3,  -1,   2,  15,   4,  36, -10, -10,   0,   1, -21,  41,  18,  -4, -34,   3,   1, -28, -18, -41,  36, -25,  -4,  16,   0,   2, -11,  35,   2, -24,  14,   0,  -1,   2,  19,   2, -25,  -2,   5,  -2,   0,   2,   0,  -8,  -3,   4,   2,   0,   0 },
      {   3,   4,   1,  -2, -11,   7,  -1,   2,   5,   3,  10, -22, -24,   4,   9,   0,  11,  23,  12,  32, -28,  29,  -6,  -3,  15,  23,  39,  -1,  40, -12, -20,  -1,  20,   4,  24,  25, -26, -30,   3,   1,  15,   8,   3, -42, -20,  14,   5,   0, -10,  -4, -21,  -8,  21,   4,   0,   0,  -1,  -3,  -1,   6,   2,  -3,   0,   0 },
      {   8,  -8,   1,  -3,   6,  11,  -1,   1,  13,  -6,   4,   3,   8,   1,  -6,   1,  12, -18,  -3, -16,   4,  -5,   2,   3,  15, -11,  12,   1, -15,   6,   3,   1,  37, -38,  -9,  14,   6,   7,  -4,   0,  90,   9, -30,   1,  -3,  -8,  -3,   0, -38, -14,  12,   4,  -2,   0,   3,   1,  -3,   9,   5,  -2,   0,   0,   0,  -1 },
      {   2,   4,  -2,   5,  21, -34,  12,  -4,   8,  11,   9,   9,  18,  31,  -8,  -2,   3,  10,  -5,  10,  56, -13, -18,   0,   4,  -5,  -2, -50,   5,  -9,   5,   5,  24,  20,  20,  14,  11, -14,  -1,  -3,  -8,  49, -12,   1, -23, -18,   6,  -2,   9, -16, -17,   1,  -8,  12,   4,   0,  -4,  -3,   5,   5,  -2,   0,  -1,   0 },
      {  -8, -15, -15, -18, -42, -94,  33,   0,   5,   0,   7,  -1,  18, -16,  15,  -5,   1,  -5,   2, -13,  -1,  29,  -6,  -2,   2,  -3,   6,  -4, -21,   6,   1,   1,  -2,  -2,  -2,  10, -11,  13,   0,  -1,   4, -17,   3, -10,   2,  13,  -4,   1, -11,   6,   3,  -1,  12,   1,  -3,   0,  -1,   1,  -3,   1,   1,  -1,   0,   0 },
      {  -1,  -1,  -4,   6, -18, -13,   8,   0,   4,   7,   3,  22,  15, -21,   7,  -1,   5,   7,   6,  21,  -5,  23,   6,  -7,   2,  14,  -1,  61,  20,   4, -24,  -4,  -5, -10, -32,  -8,  29, -26,  -4,   6,  -3,  46, -24,  18, -13, -14,  11,  -2,  22, -22,  10,   6, -26,   8,   1,   0,  -7,  -2,   7, -12,  -1,   7,  -1,   1 },
      {   1,   3,   3,  -8,  -6,  -4,   9,  -2,   2,   6,   3,   2, -22,  21,  -7,   2,   2,   3,   8, -20, -13, -24,  13,   2,  11,  23,  19,  34, -36,  31,   8,  -9,   5,  32,  24,  20,  35,   0, -27,  -4,   9,  -2,  34,   6, -31, -26,   0,   5,   4,   5,  -1, -38,  -7,  14,   4,   1,  -3,  -3, -13,  -1,  15,  -1,  -1,   0 },
      {   0,  -1,  11,   6,   4,  -5,   2,  -2,   2,   1,  13,   4,  25,  23,  -7,  -3,   1,  -3,  19,  11,   7,  11,  -9,  -2,  -4,   2,  11,   1,  20,   4, -16,   2,   0, -31,  55,  -2,   8, -26,  -9,   5, -16, -61, -23,  35,  -7, -13,  14,   3, -15,  15,  32,   0, -17,   4,   3,  -3,   4,  13,  -7, -14,  -2,   1,   0,  -1 },
      {   1,   0,   4,  -3,   4,   0, -17,   5,  -3,  -7,  -7,  -9, -25, -87,  26,   9,  11,   8,  19,   9,  53,  -4,  23, -10,   2,  -2,   3, -16,   8,  29,  -9,  -6,   8,   5,  21,  -1,  -9,   1,  -7,   1,   3,   0,   6,  18,  -5,  -5,  -2,   0, -10,   6,   0, -11, -12,   4,   4,   0,   1,   0,  -4,  -1,   4,   6,  -1,  -1 },
      {  -4,  -4,  -2,   2,  -4, -15,  -1,  -3,  -7,  -9,  -5,  -6, -32,  -9, -13,   6,  -6,  -7,  -5,   5, -10, -56,  18,   6,   0,  -5,  11,   2,  -6,   8,  16,  -4, -10, -12, -11,  35,  -9, -10,   4,  -3, -34, -12, -49,  -6, -23,  22,  17,   0,  -7, -42,  16,  10,  14,  26,  -9,  -1,  -3,  12,  14,  -3,   4,  -5,  -5,   0 },
      {   4,  -5,   0,  -4,  -6,  -2, -17,   6,   7,  -7,  -1,  -6,  -5,  -2,   2,   4,   6,  -6,   1,  -4, -12,   8,   0,  -1,  10, -10,  -3, -12, -17, -12,   2,   6,  14,  -6,  17,   7, -18,  -1,  15,  -1,  27, -21, -13,  27,  -9,  14,  -1,  -4,  97,  -6, -11, -11,  -9,   6,  -8,   1, -33,  -6,   6,  -6,  10,   1,   0,   1 },
      {   5,   3,   3,  12,   8,  -1,  47, -12,   3,  -2,   1,   4,   1,   0,   2,  -2,   4,   2,   3,  16,  13, -28,  -2,   9,  12,   2,  15,  14,  39,  28,  11, -14,  -1,   0,  -9,  33,   5,  22, -27,  -7,   8, -22, -16, -13,  32, -14, -14,   9,  34,  26, -22,  24,  -2, -25,   3,   0,  -3, -10,  11,   2, -21,   1,   4,   0 },
      {   0,  -2,  -4,  -7,  -4, -23,  15,  -5,   0,  -2,  -6,  -7,   5,  -7, -28,   7,   1,  -3,  -5, -11,  -7, -50,  10,  12,   5,   3,  -3,  -1,  48,  -2,   3,  -3,   2, -17,  -5, -64,  -3, -11,   7,   6,  21,   7,   7,   1, -25,  14,  18,  -4,   4,  39,  -7, -20,   9,  12,  -8,  -6,  -4,  -8, -17,  10,   6,  -3,   1,   1 },
      {   4,   7,   8,   8,   0,  31,  85, -25,  -2,  -3,  -1, -11,  -5, -22,  15, -10,  -3,  -1,  -1,  -8, -17,  -6, -34,   9,  -1,   2,   3,  -9,  -6, -11,   2,  10,   5,   4,  15,  -1, -19,   5,  15,  -4,  -4,  22,   9,  25,  -6,  13,  -9,  -8,  -4, -13,  30, -12, -12,   3,  -9,   4,  -3,  10,  -4, -15,   8,   4,  -1,   1 },
      {  -3,  -1,  -2,  -2, -15, -25, -36,  16,   0,   4,   2,  -3,  -3,  -7,  -5,   4,  -2,   3,  -2,   7, -31, -24,   1,   9,   6,  10,  15,   3,  27, -41,  32,   3,  10,  11,  21,  25,  -5,  32,  14, -16,  -3,  28,   0,  27,  17, -10, -17,  -5, -12,  18,  23,   7, -34, -13,   6,   1,  10,  -6,  -7, -20,  -6,  16,   1,  -1 },
      {  -1,  -5,   1,  -3,  -2,  -9,  -4,  -5,   0,  -5,   9,   0,   9,  -9, -21,   3,   0,  -3,   5,  -3,   1,  -5,  -7,   9,   4,  -5,  16,   5,  29,  -4,   7,  -2,  -1,  -6,  11, -24,   6,  17,  -4,  -3,   4, -24,  58,   6,  -5, -10, -11,   3,  -3, -82, -19,  26,  -3,  -3,  10,   5, -18,  15,  16,  -3,  -3,   3,   4,  -2 },
      {  -1,  -2,  -2,   0,  -3,   4,   2,  -6,  -3,  -6,  -7,   1,  -8, -35,  10,  -8,  -4,  -7,  -7,  -5, -11,  -8, -48,   8,  -3,  -1,  -8,  11, -13, -43,   5,  16,   1,  -5,   4,   3,  33,  -6,   8,   0,  -5, -27, -21,  -5,  -8, -26,  13,   4, -17,  12, -59,  -3, -11,  19,  26,  -5,  -5, -19,  12,  16,   6,  17,  -7,  -3 },
      {  -4,   2,  -1,   1,  -2,   0,  -4, -26,  -8,  -2,  -4,  -7,  -4, -14, -64,  10,  -3,   8,   3,   8,   5,  25,  -5,  13,  -5,   3,  -2,  -2,  -8,   8,  13,  -5,  -8,   7,  -1,  10,  -5, -13,  -1,   1,  -5,   8,  -5,  -5,  16,  -3,   0,   1, -21,  19,  13, -12,  -4, -14,   5,   1, -85,  -7,  15,   0,   4,   4,   9,  -1 },
      {   3,   0,   0,   2,  -2,  -1,  12,   5,   0,  -6,  -6,  -7, -10, -22, -77,  19,   6,   4,   2,   7,   3,  28, -12,  11,   2,   0,  -3,   3, -18,   7,  22,  -4,   7,   1,   1,   8,  -3, -22,   6,   1,   6,  -1,   4,  10,   6,   2,  10,  -5,  15,  -4, -11,  12,  -3,  -3,  -2,  -2,  73,  -4, -15,  -1,  -9,   2,   1,   0 },
      {  -1,  -1,  -1,   2,  -7,  -1,   9, -19,   0,   1,   2,   9,  -1,  22,   3,  -4,  -1,  -1,   1,   3, -12,  -3,  37,  -5,  -3,  -3,  -2,   5,  -9,   9, -13,  -8,  -3,  -1,   2,   7, -45,  -9,   3,   8,   0,   8,   3,  71,   4,  -2,   3,  -9, -31,   0, -55,   7,   7,  -1,  -6,   1,   1, -34,  20,   9,  -9,   6,  -1,   5 },
      {   1,   2,   2,   4,   7,   3,   6,   8,   0,   2,   0,   9,   7,  -7,  -3,   9,   1,   3,   4,   7,  12,   9,  11,  -6,   1,   6,   1,  19,  10, -12,   9,   8,   9,   3,  19,  10,  60,  13,  29, -14,  -5,  -1, -12,  28,   7,  42, -24, -21,  -9,  -3, -19, -18,  43, -17, -29,  10,  -8,  16,  -6,  24,  -4, -27,   8,   3 },
      {  -1,  -2,  -1,  -2,  -3, -11, -13, -52,   1,   1,   3,   0, -12,  -6,   8,   1,   1,  -1,   2,  -4,   1, -32,   5,   4,   0,   1,   2,   0, -22, -16, -38,  15,   6,  10,  11,  12,  28, -36,  18,  20,  15,   2,  33,  -7,  16,  16,  16, -12,   8,  17,  13,  46,  -1, -16,  -9,  -6,   2,  12,  10, -10, -28,  -1,   5,   1 },
      {   3,   6,   7,   5,   8,   0,  32,  97,  -2,   0,   1,   0,  -9,  -2,  -3,  10,  -2,  -1,   0,  -3,  -3, -10,   0, -22,  -2,   0,   1,   0, -13,  -5,   1,   0,  -2,   3,   5,   0,   4, -23,   3,   4,   6,   1,  21,   2,   1,   6,  17,  -7,  -8,  15,   2,  33,  -4,   1,  -4, -11, -26,  -2,  27,  -4, -14,   3,  -1,   1 },
      {  -1,  -1,  -1,  -1,  -1,  -7, -35,  14,  -1,   1,  -1,   3,  -2,   2,   4, -18,  -3,  -3,  -2,  -4,   0, -17, -81,  11,   2,  11,   4,  16,   5,  65,  -9,  17,  -2,   2,  -3,   3, -18,  -2,  26,  -2,   1,  11,   3,  21,   2,  -1,   1,  -4,  -1,   2,   2,   4,  20,  -5,  -8,  -2,  -3,  -6,  -1,   2,  -6, -14,   4,   5 },
      {  -1,   2,  -4,  -1,  -4,  -4,   1,   6,  -1,   5,  -3,   0,  -1,  -1,  -6,   3,  -1,   4,  -6,  -2,  -7,  -7,   0,  -3,   0,   3,  -4,  -1,  -1,  -3, -10,  -2,  -3,   7, -10,   2, -18,  -5,  -3,   4,   0,   7,  -8,  14,   8, -21,   1,   7,   2,  16, -39, -14,  11,  -6,  16,   1,  -3, 106,  -3, -29,  -1,  -2,   3,  -8 },
      {   0,  -1,   0,   1,   0, -10,   9,  21,  -1,  -3,  -4,  -2,   0,  -2, -21,   6,  -2,  -7,  -5,  -3,  -6, -31,  -2, -27,  -2,  -5,  -4,  -6,   1,  -9, -62,   9,  -1,   1,  -3,  15,  -3,  -8,   1,  16,  -2,  -5,  -8,  -6,  37,  -6, -10,   2,  -1, -33,  -2, -37, -10, -27,  11,  14,  -3, -20, -48,  10,   7,  13,  25,  -3 },
      {   0,   0,   1,  -3,  -2,   0,  -1,  -2,  -1,   0,   1,  -3,  -3,  -5, -31,   8,   0,   0,   1,  -5,  -4,   0,  -7, -24,   0,   0,   1,  -2,   1,  -5, -49,  10,   1,   6,   4,  -3,  -1,  51,  -4,   7,  -5,   2,  -8,  -2, -39,  -2,  -7,   3,   5,  14,  16, -19,   2, -11,   7,  13,  23,  -1,  74,   8, -21,  -2,   4,  -5 },
      {  -1,   0,  -1,   1,  -2,   3,   5,  -9,   2,   5,   3,   5,   3,  10,  20,  97,  -2,  -1,  -3,  -1,  -5,   1, -13,  -1,  -2,   0,  -3,   1,  -8,  13,  -5, -24,  -1,  -2,  -3,  -3,  -6,   6,  29, -10,  -5,  -5,  -8,  -5, -39,  -7, -15,  -8,  -4,   5,  -5,  25,  -8, -19,  12,   6,  -9,  -1, -24,   5,  -3,  13,   7,  -9 },
      {   0,   0,   0,   2,  -1,   3,   0,   0,  -3,  -4,  -6,  -2, -14,  -7, -15, -52,   0,   0,  -1,   2,  -1,   5,   6, -23,  -1,   0,  -3,   1, -11,  -7,   2,  16,  -1,  -3,  -2,  -4,  -4,   9, -17,  -9,  -5,   1, -10,   7, -64,   0, -15,  12,   1,  10,  -4,  49,   7, -37,   6,  12, -14,   0, -48,   3,  -7,   2,  19,   0 },
      {   0,  -1,  -1,  -4,   0,  -6,  -1,  -1,   0,  -1,   0,  -3,  -5,   7,  13, -16,   0,  -2,   0,  -6,   1,  -8,   8,  -8,   1,   1,   0,   0,  -7,  15,  34,   5,  -1, -11,  -3, -20,   0, -55,  22, -19,  -3,  -8,  -5, -18,  -8,   6, -46,   0,  -3,  -5,  -7, -35, -26, -30,   6,  26,  11,  -1,  31, -14, -36,  21,  21,  -2 },
      {   0,   1,  -1,   0,   3,   4,   4,  -1,  -1,  -1,  -2,  -3,   2,  14, -15, -16,   1,   1,  -1,   1,   9,   8,  15, -11,   1,   3,   1,   0,   7,  32, -13,  20,   1,   5,  -2,   5,  12,  33,  42,   3,  -2, -12,  -4, -27,  -5,  24,  10, -29, -13,   5, -34,   7, -50,  14, -17,  -2,  -3,   1,  -7, -49,  25,  18,  -9,  10 },
      {  -1,  -1,  -1,  -2,   2,  -1, -13,  -5,   0,  -1,  -1,  -3,   0,  -4,   2,  13,  -2,  -3,  -4,  -4,  -1,  -5, -36, -20,  -2,  -4,  -3,  -6,  -2, -15,  -1, -47,   0,  -2,   0,  -5,  13,  -5, -53,  10,  -1,   5,  -2,   6,  -3,  36, -10,   7,  -8,   0, -18,  -4,  10,  -5, -21,   8,   1, -20,   2, -71,   6, -10,   8,  19 },
      {   0,   1,   0,   0,   2,   0,   1,  26,   0,   0,  -1,  -1,  -2,   1,   6,  -4,   2,   4,   3,   2,   6,   4,  13, 103,  -4,  -3,  -6,  -3, -12,  -7, -38,   3,  -1,  -1,  -1,  -2,   3,   5, -14, -32,  -3,  -1,  -5,  -1, -14,  10,   9,   0,  -3,  -4,  -7,  -8,  -1, -23,  -3,   0,   1,  -6,   1, -20,  -9,   4,  14,   2 },
      {   0,   0,   0,   0,  -4,   0,  -8,   1,   0,   1,   0,   2,  -2,   6,   1,  -9,  -1,   0,  -1,   1,  -6,  -1, -18,   6,  -1,  -1,  -1,  -1,  -4,   6,  -4, -34,   1,   1,   2,   2,  -3,   1, -24,   5,   1,   4,   0,   4,   0,  58,   4,   1,  -6,   1, -10,   0, -73,  -2,   9,  -4,  -1,  27,  -1,  62,  -6, -18,   1,   3 },
      {   0,   0,   0,   0,   1,   1,  -2,  -3,   1,   2,   1,   2,   3,   8,   4,  30,   0,   0,   0,   1,   3,   2,   0,  -3,   2,   4,   3,   5,   5,  12,  12,  79,  -3,  -1,  -3,  -1,   0,  -3, -46,   1,   1,   8,   1,  13,   6,  52,  -2,   2,   0,   0,  -3,   0,   8,   5,  47,  -6,   0,  -4,  -1,  -9,   1,  24,  -9, -22 },
      {   0,   0,   0,   1,   0,  -5,   0,   2,   0,   0,   1,   1,   2,   0, -12,  -2,   0,   0,   0,   1,   1, -11,   1,   5,  -1,   0,  -1,   0,  -1,   0, -27,  -1,   0,  -2,   0,   0,  -1, -27,  -2, -35,   1,   1,   5,   2,  10,   1, -57,   8,   1,  18,   3,  33,   1,  16,   4,  26,  10,   3,  23,  10,  82,  -5, -11,  -4 },
      {   1,   0,   1,   0,   3,   2,   2,   0,   0,  -1,  -1,   0,   0,   7,  -6, -24,   0,   0,   0,   0,   5,   2,   0,  -3,  -1,  -1,  -1,   0,   0,   9, -12, -52,   5,   2,   7,   2,  18,   8,  36, -15,   3,   7,   3,  10,   7,  34,   5,  45,   4,   2,   7,   0,  28,  16,  45, -10,  -1,  -3,  -4,  -6, -12,  55, -13, -25 },
      {   0,   0,   0,   0,  -1,   3,  -1,  -4,  -1,   0,  -1,  -1,  -2,  -2, -13,   5,  -1,   0,  -1,  -1,  -2,   3,  -1,  -9,  -1,  -1,  -2,  -1,  -3,  -3, -25,   8,  -2,  -2,  -3,  -3,  -5,   2, -12, -51,   1,  -2,   1,  -3,   9,  -6, -35,  -5,  -1,   6,  -3,   5,  -6,  77, -12,   0, -12,   0, -24,  -3, -63, -12,   4,   8 },
      {   0,  -1,   0,  -1,  -1,  -2,  -1, -12,   0,   0,   0,   0,   0,  -1,   1,   4,  -1,  -2,  -1,  -2,  -2,  -4,  -4, -33,   0,  -1,  -1,  -1,  -1,  -1,  -3,   2,  -1,  -4,  -2,  -3,  -3,  -4,  -7, -90,   5,   0,   7,   0,  10,  -1,  62,  -3,   0,  -6,  -1,  -9,  -4, -36,  -2,  14,   1,  -1,   2,   0,   5,  -8, -33,   4 },
      {   0,   0,   0,   0,   0,   0,  -8,   0,   0,   0,   0,   1,   0,   2,  -1,  -4,   0,   0,   0,   0,   1,   0, -14,  -1,   0,   0,   0,   0,   1,  -1,  -1,  -9,  -1,   0,  -2,  -1,  -1,   0, -33,  -2,   0,  -2,  -1,  -1,  -1,  -4,  -3, -53,   3,   1,   6,   0,  17,   0, -46,   4,   0,  19,   1,  28,   2,  92,  -5,   7 },
      {   0,   0,   0,   0,   0,  -2,   1,   1,   0,  -1,  -1,  -1,  -1,  -1,  -6, -16,   0,   0,   0,  -1,   0,  -3,   1,   1,   0,  -1,  -1,  -1,  -2,  -2,  -8, -23,  -1,  -1,  -1,  -2,  -2,  -6,  -4,   1,  -2,  -3,  -5,  -3,  -7,  -3, -27, -80,   5,  -4,   7,  -6,  12, -15,  59, -36,  -2,  -6,  -4,  -9,  -6, -22, -46,  12 },
      {   0,   0,   0,   0,   1,   3,   1,   0,   0,   0,   0,   0,   1,   1,   8, -10,   0,   0,   0,   1,   0,   4,   1,  -1,   1,   0,   1,   0,   2,   1,  10, -16,   0,   1,   0,   2,   0,   8,  -1,  -5,   2,   0,   4,   1,   9,   3,  35, -50,   4,   7,   6,   9,  10,  28,  42,  59,   3,  -2,   5,  -3,   9,  -8,  73, -13 },
      {   0,   0,   0,   0,   0,   0,   0,   9,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  17,   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   1,   1,   1,   1,  32,   0,   0,   1,   0,   1,   0,   2,   2,   0,   3,   0,   2,  -1,   3,   0,  94,  -8,  -1, -14,  -2, -23,  -4, -73,  -5 },
      {   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   1,   1,   9,   0,   0,   1,   0,   1,   1,   5,   2,   0,   1,   1,   1,   1,   2,   1,   9,   1,   0,   1,   1,   2,   2,   8,   1,   0,   1,   1,   2,   1,   4,   4,  19,   3,   2,   5,   1,   8,   3,  30,  12,   1,   8,   1,   9,   2,  16,   7, 119 },
    },
    {
      {-113,  32,  25,   1,   3,   1,   1,   0,  24, -18,   2,   0,  -1,   0,   0,   0,  31,  -7, -11,   1,  -1,   0,  -1,   0,   5,   2,  -3,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -11, -26,  28,   0,   1,   0,   1,   0,-104,  17,  28,   3,   3,   1,   2,   0,  37, -10, -14,   1,  -1,   0,   0,   0,  32,  -1, -13,  -1,  -1,   0,  -1,   0,   5,   3,  -3,  -1,   0,   0,   0,   0,   3,   1,  -1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
      {   9, 100, -35, -25,  -4,  -3,  -2,  -1, -27,  16,  18, -10,   1,  -1,   1,   0, -14, -43,  21,  11,   1,   2,   1,   1,  11, -15,  -3,   7,  -1,   1,   0,   0,   4,  -4,  -3,   2,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
      {  36,  20,  11, -11,  -8,  -2,  -2,  -1,   5, -88,  28,  26,  -1,   4,   1,   2,  27, -20, -35,  16,   3,   2,   0,   1,  -6,  40, -17, -12,   4,  -2,   0,  -1, -12,  10,   5,  -6,   1,   0,   0,   0,  -2,   4,   2,  -1,  -1,   0,   0,   0,  -2,   1,   1,   0,   0,   0,   0,   0,  -1,   2,   0,  -1,   0,   0,   0,   0 },
      {  28,  16,  14, -23,  -1,  -3,   0,  -1,  22,  51, -21, -14,  -4,  -2,  -1,  -1,  86,  -9, -34,   3,  -4,   0,  -2,   0, -31,  -8,  12,   6,   2,   1,   1,   0, -27,  -5,  11,   3,   1,   0,   1,   0,  -6,  -4,   2,   2,   0,   0,   0,   0,  -3,  -2,   1,   1,   0,   0,   0,   0,  -3,  -1,   1,   1,   0,   0,   0,   0 },
      {  11,   2,  93, -36, -14,  -5,  -2,  -2,  22,  22,  25,  -6, -14,  -2,  -2,  -1, -48, -15, -14,  25,   6,   2,   2,   1,  -3, -16, -18,   9,   7,   1,   1,   0,   9,   0, -12,   0,   2,   0,   0,   0,   1,   0,  -3,  -1,   1,   0,   0,   0,   1,   1,  -3,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
      {   4,  44,  26, -14,  -9,  -5,  -2,  -2, -32, -14, -13,  13,   7,   1,   2,   1, -16,  77, -24, -33,   7,  -3,   0,  -1, -22,   1,  40, -12,  -5,   0,   0,   0,   7, -35,   8,  14,  -3,   2,   0,   1,  10,  -5,  -5,   4,   0,   0,   0,   0,   3,  -4,  -3,   2,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
      {  -2,   5,  17,  12,  -3,  -5,  -1,  -1, -16,   5, -85,  37,  20,   0,   2,   1, -23, -54, -27,   5,  21,   5,   3,   2,   9,   8,  23, -23,  -6,   1,  -1,   0,   9,  16,  11, -10,  -7,   0,  -1,   0,   2,   1,   7,  -2,  -3,   0,   0,   0,   1,   1,   2,  -1,  -1,  -1,   0,   0,   1,   1,   3,  -1,  -1,   0,   0,   0 },
      {  15,  23,   1,  31, -18,  -4,  -3,  -1,  28,  20,   0,   1,  -3,  -5,  -1,  -1,   1,  30, -32, -24,  14,   0,   1,   0,  89,  -1, -37, -13,   0,   1,  -1,   0, -13,   1,  10,  -2,  -1,   1,   0,   0, -21,  -5,  12,   2,   0,   0,   1,   0,  -9,  -4,   5,   1,   0,   0,   0,   0,  -2,  -3,   2,   1,   0,   0,   0,   0 },
      {  -2, -13,  -6, -89,  35,   9,   6,   1,   1, -19, -36, -22,   8,  11,   2,   3,   1,  18,   8,  18, -20,  -3,  -2,  -1,  52,  14,   6,   8, -14,  -6,  -2,  -2,  -1,   5,   3,   8,  -2,  -2,   0,   0, -11,   0,   6,   4,   1,  -1,   0,   0,  -4,  -1,   1,   3,  -1,   0,   0,   0,  -1,  -1,   1,   1,   0,   0,   0,   0 },
      { -20,  -7, -33, -19,  12,   6,   4,   1,  -4,  40,  24,  -7, -11,  -4,  -2,  -1, -30,   7, -41,  28,  13,  -3,   2,   0, -22,  57, -14, -39,  13,  -2,   1,  -1,  -3,  -2,  40, -14,  -8,   1,  -1,   0,   7, -21,   5,  10,  -4,   1,   0,   1,   5,  -5,   1,   4,  -1,   0,   0,   0,   2,  -3,  -1,   2,   0,   0,   0,   0 },
      {  -8, -10, -38, -12,   4,   8,   3,   2,  -6, -17,  -5,  12,  -2,   0,   1,   1,  -8,  17, -60,  35,  18,  -7,   1,  -1,  -8, -79, -18,  22,  20,   4,   2,   2,   9,  -3,   3, -11,  -1,   3,  -2,   0,   4,  21,   6, -12,  -4,  -1,  -1,  -1,   0,   6,   6,  -4,  -2,   0,   0,   0,   1,   2,   3,  -2,  -1,   0,   0,   0 },
      {   1,   7,  -2, -15,  -4,   0,   3,   0,   3,  21, -12,  81, -29, -12,  -3,  -2,  22,  29,  41,  34,  -6, -18,  -5,  -4,   2,   6, -17, -24,  23,   2,   1,   0,  21,  -4, -30, -18,  10,   5,   0,   1,  -6,  -2,   5,  -8,   1,   3,   0,   0,  -5,  -4,   3,  -1,   0,   1,   0,   0,  -3,  -2,   1,  -3,   0,   1,   0,   0 },
      { -10,  -9,   4,  -2,  -7,   2,   1,   1,  -3,  -3,   9,  33, -12,  -6,  -2,   0, -14,  -6,  20,  19,   2,  -6,  -1,  -1,  17, -21,  26,   0,  -2,   1,  -1,   0,-100, -17,  35,  10,   6,   3,   2,   1,  11,  -6, -13,  -1,   3,   1,   1,   0,  11,   4, -10,  -4,   1,   0,   0,   0,   6,   2,  -3,  -3,   0,   0,   0,   0 },
      {   2, -20, -17, -10, -16,  11,   3,   3,  22,   5,  29,  18, -19,  -5,  -3,  -1,  11, -30, -17,   2,  15,   1,   0,   1,  30,  -2,  48, -16, -19,   2,  -3,   1,  44, -49,  -2,  39, -11,   0,  -1,   1,   0,   6, -35,   7,  10,  -2,   1,   0,  -7,  13,  -3,  -7,   4,  -1,   0,   0,  -5,   3,   0,  -3,   1,   0,   0,   0 },
      {   0,   3,   0,  27, -24,   7,  -1,   1, -10,  -4, -38, -28, -12,  14,   7,   2,  -4,  -1,  -7,  40, -17, -14,   2,  -1,   7,  37, -20,  39,  23, -21,  -3,  -3,  -3, -64, -13,   8,  26,   4,   0,   2,  18,  -3, -11, -10,   3,   4,  -2,   0,   0,   9,   3, -11,  -1,   2,   0,  -1,   0,   2,   4,  -2,  -1,   0,  -1,   0 },
      {   9,   8,  10,  24,  92, -34,  -6,  -7,  10,   8,  11,  22,  31,  -5,  -8,  -2,  -5,  -5, -14,   8, -34,  15,   2,   2,   3,  -5, -14,  -6, -10,   3,   4,   1,  -1, -40,  -5,   4,  -1,   4,   2,   2,   9,   2, -10,  -4,  -2,   1,   0,   0,   0,   7,   1,  -5,  -3,   1,   0,   0,  -1,   2,   1,  -1,  -1,   0,   0,   0 },
      {  -6, -14,  -7, -36,  -6,   4,   6,   2,  -2,  -5, -16,   8,   7,  -2,   0,   0,   1, -22,  14, -57,  30,  11,  -2,   2, -18,  -4, -66, -20,  13,  14,   3,   2, -15, -48,  -7,  26,  -6,   2,   2,   1,  10,   6,   9,   5,  -9,  -4,  -1,  -1,   5,  11,   3,   3,  -2,  -3,  -1,  -1,   0,   3,   3,   1,  -2,  -1,   0,   0 },
      {  -6,   2, -11,  -6,  16,   8,   0,  -2,  -6,  10,  -2,  29, -65,  22,   4,   0, -19, -13, -32, -48, -42,  11,  15,   4,  -7,   8,   1,  22,   4, -23,   2,  -1,  -8,  23, -11,  27,  26, -16,  -3,  -2, -11,   4,   4, -15,  14,   0,  -2,   0,   4,  -5,   7,  -2,   1,   1,  -2,   0,   2,  -3,   1,  -1,   2,   0,  -1,   0 },
      {  -5,  -4,   3,   8,   0,  -9,   1,  -1, -10,  -7,   1,   5,   1,  -3,  -1,   1,  -5,  -7,   7,  12,   1,   0,  -1,  -1, -16,  -4,   2,  15,   1,  -2,  -1,  -1,   7, -30,  24,  14,  -4,   2,  -2,   1,-106, -16,  38,  11,   5,   3,   3,   0,   6,  -8,  -8,   1,   2,   1,   1,   0,   7,   3,  -7,  -3,   0,   0,   0,   0 },
      {  -9,  -4, -15, -15, -15,  -3,   3,   2,   3,  21,  15,  25,  56, -27,  -8,  -4,  -9,   8, -17,   2,  29,   6,  -5,  -1,  -7,  33,   3,  42, -32,  -4,   1,  -3, -19,  10, -43,  16,  21,  -6,   4,  -1, -13,   8,  -1, -33,   9,   5,   1,   1,  -2,  -4,  16,  -9,  -5,   4,  -1,   1,   2,  -6,   2,  -2,  -5,   2,   0,   0 },
      {   1,   6,   5,   9,  -1,  82, -24,  -1,   0,   6,   6,  -3,  24,  39,  -5,  -7,  -5,  -1,  -6,   9, -21, -13,  16,  -1,  -8, -10,  -2, -44, -26,  -4,  12,   4, -18, -13, -32, -10,   1,  -7,   6,   3, -19,  13,   7,   0,   1,  -9,   0,   0,   0,   1,  13,   2,  -1,  -5,  -1,   0,   5,  -1,   1,   4,  -1,  -3,  -1,   0 },
      {  -3,  -3,  -2,  -2,  15, -31,   9,  -1,  -6,  -5,   3, -36, -24,  -9,   7,   4,  -6,  -7,  -7,  -7,  19,  -5,  -4,   1,   6,   6,  37, -37,  36,  11, -11,   0, -27, -11, -51, -20,  10,  23,   2,   1, -25,  38,   7, -19,  -8,   2,   2,  -2,  -8,  -2,  25,   5, -12,  -1,   0,  -1,   3,  -5,   1,   8,  -3,  -1,   0,   0 },
      {  -4,  -5, -14,  -2, -19, -27,   7,   2,  -5, -10, -12, -14,  -1,  -8,   1,   2,  -3,   3, -18,  14, -18,  22,  -2,   0,  -4, -20,  -9, -32, -11,  13,   6,   1, -12,  11, -52,   9,  18,  -6,   5,   0,  -7, -72, -24,  33,  21,   1,   0,   2,  14,   4,  -9,  -6,   4,   0,  -3,   0,   2,  11,   5,  -5,  -1,   0,  -1,  -1 },
      {  -7,  -3,  -6,  -4, -44, -40,  11,   3,  -5,   4,  -6,  -5,  -5,  -5,  -3,   1,  -4,   9,   2,  20, -47,  40,   1,  -3,  -8,  -1, -14, -26, -42,  10,  15,   0,  -1,  -8,  21,  -1,  -2, -15,   2,   0,   0,  59,  11,  -8,  10, -12,  -4,  -3, -11,   3,  14,  -1,   9,  -1,  -3,   0,  -1,  -9,   0,   5,   3,   1,  -1,   0 },
      {   2,  13,   6,  18,  29,   2,  -5,  -3, -10,   0, -17, -12, -30,  14,  -2,   2,   3,  18,  10,  26,  45,  -4, -10,  -5, -10,  10, -21,  -3, -11,  17,   6,  -2,  -9,  26,  -6,  47, -37, -19,  11,  -3, -14,  25, -32,   2,  32, -15,   3,   0,   1,   5,   2, -28,   6,   2,  -2,   2,   1,  -4,   9,  -3,  -1,   1,  -1,   1 },
      {   3,   4,   1,  -6,  -5,   7,  10,  -2,   3,   3,  -3,  -7,  17, -24,  13,   2,   2,   2,  -2, -12, -19, -23,   4,   5,   2,   1,  -9, -18,  10,  -3,  -9,   2,  10,  12,  16,  -6,  19,   8, -10,  -1, -22,  27, -38, -21,  11,  10,   1,  -1,  82,  28, -27, -22,  -8,   1,  -2,  -1,  -3,   3,   9,   2,  -6,   1,  -1,   0 },
      {  -2,   1,   5,   2,   2,  15,   2,   0,  -6,  -6,  -5,  -7,  14, -58,  20,   1,  -8,  -6,  -3,  -6, -29, -49,   5,   8, -11,  -8, -11, -16,  10,   4, -13,   4,  -2,  12,  16,  29,  14,   4, -12,  -4,  -5,   2, -17,  11,  16,   9,  -2,  -3, -67,  -6,  19,  -9,   9,   8,   2,   0,   0,  -7,   6,   0,  -3,   5,   0,   0 },
      {   5,   4,   3,  10,  13,  27,  15,   1,  -2,  -6,  -7,  -6, -37, -58,  22,   8,  -2,  -4, -11,   1,  -7, -19, -13,   4,   2,   8,   0,  17, -46,  40,  -2,  -5,  -2, -15, -19, -12, -19,  -1,  11,   0,  17,  -1,  39,   3,  -1, -10,  -6,   0,  31, -12, -10,  17,   6,  -4,  -5,   0,   4,   5,  -9,  -1,  10,   3,  -2,   0 },
      {   6,   3,  10,   8,  14,   9,  62, -17,   2,  -6,   3,  -8,  -8,  14,  31,  -3,   6,   1,  13,   6,  27,   8, -12,   4,  -3, -15,  -3, -21, -30, -22,   2,   8,   7,  -8,  18,  -3,  28, -24, -18,   4,   6, -24,  11, -37,  18,  24, -12,   3,  -4,  -5,  24,  -4, -29,   9,  -2,   0,  -2,   1,   1,  13,  -5,  -2,   0,  -1 },
      {   3,  -1,  -1,   9,  -3,  -9,  11,  -1,   7,   2,   6,  -1,  24,   5,   3,  -2,   4,   0,  -1,  20,  -3, -12,   2,   2,   7,  -4,   7, -22,  28, -21,  -8,   4,   3,  11, -13,  69,  10, -16, -10,  -3,  26,  20,  48,  19,  -7,  -5,  -7,  -3,  33, -40, -12,  11,   2,  10,  -1,   2,  10,  -1, -20,  -2,   1,   3,   2,   0 },
      {  -2,   3,  -2,   1,  -7, -17,  57, -18,   3,  11,   9,  14,  11,  32,  33,  -8,  -5,   2,  -6,  -4,  -9, -32,  -2,  13,  -4,   6,   0,  11, -10, -22, -18,   7, -11,   7, -18,  -5, -34,  14,   0,   0,  -3,   6,  12,  32, -14, -20,   0,  -2,  -8,  51,   1,  -7,  21,  -8,  -5,  -3, -10,   5,   7, -12,   7,   2,  -3,   1 },
      {   2,   5,   3,   6,   1,  41,  53,  -8,   1,   5,  -1,   8,  16,   1,  15,   0,  -4,   2,  -4,   1, -16,  54, -37,  -1,   4,   6,   4,  10,  40,  30, -11, -10,  -1,   1,   1,  -3,   3,  -6,   8,   0, -12,  13, -31,   9,  -9, -13,   8,   2, -12, -30, -17,   6,  10,  -9,   3,   4,  10,   2,  -1,  -7,  -3,  -2,  -1,   2 },
      {  -1,  -4,  -5,  -6,   2, -20,  10,   2,   3,   1,  -4,   2,  -1,  15,   7,  -1,   1,  -6,  -4, -11,  15, -27,  16,   2,  -3,  -2, -15,   3, -24, -27,   4,   8,  -3, -14,  -8, -36,  -2,   6,  -9,   3,  -4,  20, -37,  13,  27,   1,  -5,  -2,  -6, -64, -34,  21,  20,  11,  -1,   1,  37,   7, -15,  -8,  -2,   2,  -2,  -1 },
      {   2,   4,   2,  -2,  11,  11,  -6,  12,   2,   2,   1,  -5,  -3, -24,   6,   2,   5,   6,   3,  -1,  15,  26,   9,  -7,   4,   4,   9,  -4,  10, -18,  13,   3,   2,   1,  -4, -14,  16, -28, -10,   7,   5,  13,  22,  16,   6,   5, -13,  -2, -19,  53, -26, -27,  10,   4,   0,  -3,  71,  19, -17, -19,  -8,   4,  -1,  -1 },
      {  -2,   0,   0,   0,  12,   9, -10, -15,  -4,  -1,   1,  -2,  -2, -23,  14,  -8,  -1,   2,   3,   1,  22,  23,  20,  -5,   0,   2,   6,   3,   6, -22,  11,   8,  -8,  -7,  -7, -18,  23, -30, -13,  11,  -5,  20,   5,  37,  10,  -1, -15,  -2,  -3,  -1, -26,   8,  15,   6,  -1,  -2, -81,  -8,  23, -12,   4,  10,   2,   0 },
      {   1,   3,   1,   3,   4,   6,  26,   0,  -4,  -2,  -3,  -6,  -6,   2, -81,  22,  -5,  -5,  -7,  -9,   0, -26, -65,   6,  -2,  -2,  -1,  -7, -10, -16,  19,  -8,   0,   2,   1,  -4,  25,   8,   6,  -5,   2,  12,  12,  15,  11,  11,   1,  -6,  -8,  11, -19,  -2,   4,   6,   8,  -3, -15,   4,   2,  -7,   1,   1,   8,   0 },
      {   3,   2,   6,   3,   4,   5,  14,  34,   1,  -3,  -1,   7,   1, -21, -12,  18,   2,  -1,   4,  -2,   2,   3,  -1, -11,   2,  -2,   1,  13,   8, -32,  14,  -3,   1,  -3,   4, -10,  -8, -16,  -5,   3,  -5,   4, -24,  50,  -1, -13,  -5,  -3,  32,   0,  77,  22,  -9,  -5, -12,  -4,  16, -17,   6,  11,  -2,   9,  -1,  -1 },
      {   0,  -3,  -1,  -4, -10,   3,  11, -41,  -1,  -6,  -1,  -7,   4, -27, -23, -12,   0,  -5,  -1,  -8, -18,  -3,   2,   5,   5,  -3,   8,  -3,  30, -31,  37,   8,  -4,  -7,  -9,  -8, -63, -30,  19,  17,   0, -18,   0, -37,   2,  -2, -10,   7,   3, -13,   2, -18,  19,  17, -13,   1,   1,  -1,   4,   6,  -7,   8,  -1,  -2 },
      {   0,   5,   0,   7,   2,   9,   5,  28,  -5,  -2,  -9,  -5, -12, -24,   0,  10,  -1,   6,  -2,  11,   0,  22,  -2,  -6,  -6,  -4, -12, -11, -17, -52,  -7,   9,  -7,   7, -15,  20, -28,   9, -16,  -4, -10,   0, -20, -21, -49,  36,  20,  -7, -10,   9, -28,  28,   1, -24,  12,  -3,  -6,  -8, -11,  11,  25,  -6,  -2,   0 },
      {   1,   1,   3,  -1,   1,  -1,  12,  90,   3,   3,   8,   2,  10,  13,  16,  35,  -2,  -4,  -2,  -7,  -9,  -3,  -3, -21,   0,  -2,   5,  -6,  10,  -2, -10, -14,  -4,  -7,  -5, -10, -24, -10,   3,  -4,   3,  -9,  18, -23,  19,  -3, -11,   1,  -6, -10, -15, -28,  12,  22,  -4,  -1, -24,   5,  19,  -3, -13,   4,   3,  -3 },
      {   1,  -2,  -2,  -1,   4,  -3,  -2,   4,   1,  -3,   0,   1,   6,   9,   9,  -3,   1,  -4,  -3,  -2,   4,  -6, -13,  -1,   1,  -4,   0,   0,   7,   6,  -1,  -5,   3,  -5,  -2,  -5,  11, -10,  11,   2,   0, -11,   0,  -7,  18, -12,  -8,   5,  -1,  23, -22,  12,  22,   0,  -5,  -1,   7,-101, -33,  36,  12,  11,   1,   4 },
      {  -2,  -4,  -3,  -5,  -5,   2, -46, -17,   1,   1,   1,   4,  -4,  -3,  22,  -2,  -1,   0,  -1,  -2,  -5,  21, -59,  29,  -2,  -5,  -4,  -3,  -8, -32, -60,   7,  -2,  -8,  -4, -15, -19,  -4,   3, -10,   5,   4,   8,  16,  14,  19,   8, -11,   9, -12,  13, -24,   0,  18,  13,  -1,   7, -10,  11,   1, -14,   4,   5,   0 },
      {   2,   1,   1,   1,   0,  14,  27,  -6,  -1,  -4,  -1,   0,  -8,  -6, -28,  -2,   2,   0,   1,   1,   0,  21,  42,  -5,   1,   1,   1,   8,  -9,   8,   2,  12,  -2,  -1,  -7,  -2, -13,  48, -35,  -7,   9,  -1,  17,  23,   2,  20,  14, -12,   5, -19,  -3, -55,  -5,   5,  10,   7,   8, -44,  19,   7, -21,  -2,   2,   6 },
      {   0,   2,   0,   2,  -1,  10,   7,  -9,  -1,   2,  -3,   1,   7,   2, -33,  28,   1,   5,   2,   5,  -2,  16,  27,  28,  -1,   5,  -1,   5,  15,  19, -42,  10,  -5,  -2,  -5,  -3, -23,  -6, -45,  -9,   2,  -3,   0, -15,  49,   5, -10, -13,   1,  22,   9,  41,  15,  10,   0, -12,  -9,  18, -29,   7,   4,  -2,  15,  -5 },
      {  -2,  -1,  -4,   0,  -8, -10,   0,  -5,   1,   4,   1,   8,   0,  12,   2,  26,  -3,  -1,  -5,  -1, -10, -22,   2,  18,   1,   6,   1,  11,  -1,  32,  10,  -1,  -6,  -3, -11,  -4, -16, -55, -19,  13,  -8,   7, -14,  22, -37,  24, -18, -10,  -7,   1, -19,  -4, -52,  26,  26, -11,   5, -12,   7,  22,  -8, -21,   7,  -3 },
      {   1,  -1,   1,   0,   3,  -4,   5,  -5,  -1,  -3,   0,  -6,  -7,  -4, -29,  26,   3,   1,   4,   1,  10,  -7,  19,  23,  -2,  -5,  -2,  -8, -14,  -1, -44,   2,   2,  -6,   4,  -8,  25, -40, -13,  -5,  -4, -11,  -6, -17, -48, -54,  25,  11,  -1, -13,  -2, -24,  18,  -3, -14,   3,   7,  -8,  16, -12,   7,  24,  -6,  -5 },
      {   0,  -1,   0,   3,   1,  -2,  -3,  -7,   2,   1,   5,   2,   2,  -1,   2,  15,   2,   1,   2,   7,   3,   2,   2,  16,   1,   0,   4,   0,   3,  -6,   5,  -5,   2,   1,   1,  11,   1,   6,   3,  -5,   3,  -1,  13,  -8,  17,  -6,  -4,   1,  -4,   7, -22,  47,   7, -21,  -4,  -3,  31,  -6,  95,  26, -25,  -7,  -9,  -3 },
      {  -1,  -1,  -1,  -2,   0,   3,  -9, -21,   3,   4,   3,   3,   6,   1,  29,  81,   2,   3,   4,   2,   8,  11,  12,  52,   1,  -2,   1,  -3,   3, -13,  39, -33,   1,   2,   2,   2,   4,  24,  18, -17,   1,   1,   1,   3,   1,   2,  -3,  -3,  -2,  -7,   3, -19,   3, -13,  -7,   3,  -8,   0, -21,  -2,   8,  -1,  -7,  -2 },
      {   1,   0,   1,   1,   2,   1,  -7,  13,   0,   0,   0,  -3,  -3,  16,   8,  -5,   1,   1,   1,   3,   2,   3, -17,  29,   2,   3,   2,   0,  -3,  26,  27,  29,   1,   4,   3,   9,   7,  14, -22,  10,  -5,  -5,  -8, -14, -32,  28, -35, -17,  11,  -2,  20,  10,  49,  25,  -4, -22,   9, -17,  18, -48,   4,  25,   4,  -1 },
      {   1,   2,   1,   1,   2,   2,   4,  41,  -2,  -1,  -2,   1,  -2,  -7,  -5, -32,   0,   4,   1,   3,   2,   7,  -2,  71,   2,   4,   3,   6,   6,  -5,  18,  63,   0,   1,   1,  -1,   3,   8,  16, -16,  -2,   1,  -1,   3,   6, -32,  -8,  -1,  -6,  -3, -13, -10, -28, -11, -10,  -6,  -8,   8, -11,  24,  -4,  -6,  -1,  -7 },
      {   0,  -1,   0,  -2,   3,  -3,  -6,   2,   1,   2,   3,   4,   5,   7,   2,  -3,  -1,  -2,  -1,  -3,   5,  -8, -13,  -4,   2,   3,   4,   5,   5,  15,   9,  10,  -1,  -2,   0,  -5,  12, -15, -25,   3,   2,   2,   4,   8,  -3,  20,   6, -13,   6, -10,  12, -31,  58, -22,  -5,   8,   1,  30,   6,  70,  42, -24, -22,  -3 },
      {   1,  -1,   1,  -1,   0,  -5,  12,  15,   1,  -2,   2,  -3,   6,   3,  -5, -15,   1,  -2,   2,  -2,   0, -11,  19,  26,   3,   0,   4,  -1,  12,  12,   4,   0,   1,  -8,   1, -10,  -4, -42,  29, -29,   5,   3,   8,   1,  21,  45,  50, -11,   5,   0,  10,   7,   4, -23,  18,  16,   2, -18,  10, -45,  21, -13, -24,  19 },
      {  -1,  -2,  -1,  -3,  -3,  -6, -15,  -2,   1,   0,   2,  -1,   5,  -1,   8,   4,  -1,  -3,  -2,  -5,  -7,  -8, -31,  -8,   2,   1,   3,   0,   7,  -2,  22,   4,  -3,  -3,  -4,  -4, -15,  -3, -69,   8,   2,  -9,   4, -15,  21, -37, -10,  -9,  -1,  -8,  -2,  -9, -26, -58,  27,  15,   0, -12,   2, -33,  16,   4, -16,   7 },
      {   1,   1,   1,   1,   2,   0,   1,  12,  -1,  -2,  -2,  -2,  -4,  -4,  -3, -53,   1,   2,   1,   1,   3,   1,   2,  32,   0,  -3,  -1,  -3,  -3,  -6,  12, -79,  -3,  -3,  -5,  -3,  -6,  -2, -32, -53,  -3,  -1,  -5,  -1, -12,  -7, -20,  22,   2,   1,   2,   0,   9,  15,  13,   4,   0,   6,   1,  10,  -1,  -1,  11,  10 },
      {   0,   0,   0,   1,   2,  -3,  -2,  -1,   0,   0,   0,   0,  -4,  -2,  17,   6,   0,  -1,   0,   0,   2,  -6,  -8,  -7,   1,   1,   1,   0,  -3,   0,  29,  16,   0,  -1,   1,   0,   8, -10, -13, -14,   1,  -5,  -1,  -8, -11, -23,  50, -23,   5,   9,   7,  15,  34,  11,  41,  11, -10,   7, -21,  12, -75, -11,  16,  11 },
      {   0,   0,   0,   1,   1,  -4,   1,   2,  -1,  -1,  -1,  -1,  -4,   0, -12,  -1,   0,   1,   0,   1,   1,  -7,   4,   5,  -1,  -1,  -2,  -1,  -5,   1, -24,  -4,   0,   0,   1,   1,   1, -15,  12,   6,  -3,   1,  -4,   1, -13,  23, -52,   8,   1,  -8,   4, -11,  21, -69, -22,  10,  -9,  -7, -16, -12, -56, -39,  18,  10 },
      {   0,   1,   0,   1,   0,   0,  -1,  17,  -1,   0,  -1,   0,  -2,   2,  -1,  -9,   0,   2,   0,   2,   0,   1,  -4,  39,  -1,   0,  -1,   0,  -2,   3,   2, -25,  -1,   4,  -1,   4,  -2,   6, -21,  82,   4,   6,   6,   6,   9,  14,  41,  58,   1,   1,   2,   3,   7,  -2,  -2, -10,  -5,   0,  -8,   0, -21,   0, -21, -12 },
      {  -1,   0,  -2,   0,  -3,  -2,  -7,  -5,   0,   2,   0,   2,   0,  11,   5,   2,  -2,  -1,  -3,  -1,  -5,  -5, -14, -10,   1,   3,   1,   3,   1,  14,  11,  10,  -3,  -2,  -5,  -2,  -9, -11, -25, -40,   0,   6,   0,   8,  -4,  32,  18,  15,  -6,   4, -10,   7, -15, -10, -64,  29,  -5,   5, -10,   8, -31,  58, -20, -29 },
      {   0,   0,   0,   0,   0,  -3,   7,   3,   0,   0,   0,   1,   1,   5,  -1,  -8,   1,   0,   0,   0,   0,  -3,  13,   7,   0,   0,   0,   1,   1,   5,  -3, -12,   1,   0,   1,   0,   0,  -7,  23,  14,   1,   0,   1,   1,   5,  15,  -5, -16,   1,  -7,   0, -10,   1, -44,  51, -23,   2,  14,   0,  23,  -3,  84,  35, -15 },
      {   0,   0,   0,   0,  -1,   0,   0,  -2,   0,   1,   0,   1,   1,   1,   1,  21,  -1,   0,   0,   0,  -1,   1,   1,  -5,   1,   2,   1,   1,   2,   2,   1,  38,  -1,  -1,  -1,   0,  -1,   2,   4, -19,  -1,   4,  -1,   4,   0,   8, -17,  85,   4,   4,   8,   3,  12,   7,  46,  54,   4,   0,   7,  -1,  16,   2,  28,  -9 },
      {   0,   1,   0,   1,   0,   2,   3,   5,  -1,   0,  -2,   0,  -2,   1, -12,   1,   0,   2,   0,   2,   0,   4,   6,  10,  -1,   0,  -2,   0,  -3,  -1, -15,  -4,   0,   3,   1,   3,   1,   8,  12,  22,  -3,  -1,  -4,   0,  -7,   0, -34, -21,   0,   8,  -1,   9,  -2,  17,  22,  59,  -5,  10,  -8,  14, -15,  31, -70,  49 },
      {   0,   0,   0,   0,  -1,  -1,  -3,   7,   0,   0,   0,   0,   1,   1,   6,  -2,   0,   0,   0,   0,   0,  -1,  -3,  17,   0,   0,   0,   0,   0,   1,   6,  -8,   0,   1,   0,   0,   0,   0,  -8,  32,   0,   0,   1,   0,   1,   2,  20, -23,  -3,   2,  -4,   3,  -7,   3, -39,  56,   6,   4,  11,   3,  20,   8,  77,  50 },
      {   1,   1,   1,   1,   1,   1,   3,   5,   0,  -1,  -1,  -1,  -1,  -2,  -3, -13,   1,   1,   1,   1,   2,   1,   5,  10,   0,  -1,  -1,  -1,  -1,  -3,  -3, -18,   1,   1,   1,   1,   2,   2,   7,  18,  -1,  -3,  -2,  -3,  -2,  -5,  -6, -42,   2,   4,   2,   4,   4,   7,  10,  59,   1,  -8,   1, -10,   3, -18,   4, -96 },
    },
  },
  { // 1
    {
      {-121,   0,  28,   3,   5,   5,   4,   2,   8,   2,  -4,  -1,   3,  -1,  -1,   0,  22,  -2,  -9,  -5,  -3,   2,  -1,   0,  -3,  -1,   3,   1,  -2,   0,   0,   0,  10,   2,  -2,   0,   0,  -2,  -1,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1 },
      {  21, -79,  56,  26, -16,   5,   5,  -1,  49,  23, -34,  -1,   6,  -9,  -4,   1,  -1,  12, -17, -20,   3,   7,   0,  -1, -15,  -2,  12,   8,  -3,   1,   3,  -1,  -1,   6,  -3,  -1,   3,  -2,  -2,   1,  -1,  -1,   0,  -2,  -1,   1,   0,   0,   0,   2,  -1,   0,   1,   0,   0,   0,  -1,   0,   2,   1,   0,   0,   0,   0 },
      {  -2,  68,  -5,   0, -12,  -3,   0,  -3,  94,  -3, -31,  -1,  -3,  -6,  -2,   0, -34,  -9,  16,  -1,   7,   2,   0,   0,  -6,  -5,   7,   6,   3,   0,   2,  -1,  -5,  -4,   1,  -1,  -1,   1,   0,   1,   0,  -3,   1,  -1,   0,   1,   1,   0,  -2,  -2,   0,  -1,   1,   1,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0 },
      {  21,  54,  78, -24, -37,  -6,  -3,  -4, -25, -14,  -1,  21,  11,  -4,   1,   1,  23,  -6, -46,  -8,   9,   7,   3,   1,   5,   9,  10,  -6, -10,   1,   0,   0,  -4, -11,  -7,   7,   9,   0,   0,   1,   2,   2,   0,  -2,   0,   1,   1,   0,   0,  -3,  -1,   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0 },
      {  -8, -17,  -3,  27,  -6,  -8,   2,   1,   2, -95,  -4,  48,  10,   4,   2,   3, -24,  27,  18, -22,   7,   7,  -1,   0,  16,  23,  -2, -21, -11,   1,   1,   0,  -3,   2,  -2,   2,   2,  -2,   0,   0,   4,   6,   3,  -7,   1,   1,  -1,   0,   1,   1,   1,  -1,  -1,   0,   0,   0,   0,   2,  -1,  -2,   0,   0,   0,   0 },
      {   8,  10, -45,  25,  -4,   3,   2,   1,  30, -12,  36,  10, -21,  -5,  -1,   0,  68,  22, -49, -36,   6,   6,  -1,  -1, -34,  -6,  15,   7,   4,   2,   2,   0,  -6,  -4,   7,   3,   3,  -2,  -2,   0,   1,   2,  -7,  -4,   2,  -1,   0,   1,  -2,  -2,   5,   2,   0,   1,   0,   0,  -3,  -1,   0,   0,   1,   0,  -1,   0 },
      {  -2,  25,   9,  89, -16, -32,  -2,  -6, -16,  34,  20, -19,   1,   5,   2,  -4, -11,   7,   6, -33,   0,   9,   5,   2,  47,  -7, -16,   2,  -1,  -1,  -1,  -1,  -5,  -4,  -2, -10,   3,   3,   1,   1,   9,  -7,  -1,  -1,   1,   1,   0,   0,  -1,  -4,   1,  -3,   0,   2,   1,   1,   2,  -1,  -1,   1,   2,  -1,   0,   0 },
      {   2,  23, -19,   4,   3,  -1,   6,   1, -40,  10, -72,   4,  35,  -4,  -2,   3,  18,  33,  30, -40, -15,   5,  -4,  -2, -29, -20,  33,  19, -10,  -2,   4,   0,   8,   0,  -5,  -1,  -2,  -5,  -3,   1,  -1,   2,   3,  -1,  -3,   1,   1,   0,   0,  -3,  -3,   1,   1,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   0,   0 },
      {  10,   5,  20,  28,  17, -24, -17,  -3,  10, -39, -29, -47,   1,  36,  10,  -1,  53,  10,  11,  49, -14, -27,   0,   2,  -7,  11,  -6,  -2,  12,  -8,  -8,   2,  -8, -11,  -2, -16,  -5,  12,   7,  -1,  -9,   3,   6,  15,   3,  -4,  -1,  -1,  -6,   4,  -3,  -6,  -1,  -1,   0,   1,  -2,   2,   1,   1,  -1,  -1,   0,   1 },
      {  -1,  -9, -39,  17,  -4,  -3,  -1,  -1,   1,  14, -51,  30,  20,  -2,   5,  -1,  28, -69, -31,  16,   5,   8,   4,   4,  35,  26,   7, -25, -14,   0,  -5,   0, -13,  12,   8,   3,   3,  -1,   0,   0,   3,  -4,  10,  -1,  -2,   0,  -2,  -1,   0,   0,   6,  -1,  -1,  -2,  -2,   0,  -2,   1,   2,  -2,   0,   0,   0,   0 },
      {   1,  -4,  -4, -44,  10,   8,  -4,   1,  24,  20, -11,   4,  -6,   0,   6,  -3,  35,  57,  11, -15, -13,  -4,   0,  -2,  76,  -1, -21, -13,  -9,   3,  -3,   1, -26, -13,  -7,  18,   9,   1,   5,   0,  -2,  -6,   8,  -2,   1,   3,  -1,  -1,   0,  -5,   5,   4,  -1,  -3,  -1,   1,  -2,  -1,  -1,  -1,   0,  -1,   0,   0 },
      {  10,   3,  22,   4,   1, -16,  -5,  -2,  10,  12,  19,  46,  -8, -19,  -2,  -2,  43, -33,  65,  -7, -34,  11,  -4,   1, -26, -12, -34, -24,  11,  13,   2,   2,   4,  15, -15,   0,   7,  -6,   1,  -1, -14,   0,   9,  -6,  -1,   4,   0,   1,   3,  -1,  -5,   1,   3,   2,   2,   0,  -2,   0,  -2,  -1,   1,   1,   0,  -1 },
      {   5,  12,  17,  12,  92, -22, -25,   0,   1,  16,  -8,  46,  -5, -13,   0,  -2, -22,  15, -37,   2, -19,  -9,  11,   2,  -5,   3,   2,   0,  17,  -5,  -7,  -1,   5,  -7,   5,  -2, -14,   4,   2,   0,   2,  -7,  -2,   0,  -2,   3,   2,   1,  -2,  -1,   0,  -3,  -2,   1,  -1,   1,   1,   1,   1,   0,   1,   1,   0,   0 },
      {  11,   0,   8, -13,  37,  17,  -1,  -2,   8, -32,   2, -39,  -3,  24,   8,   4,   0, -47, -10, -55, -42,  26,  12,   4,  21,  -4,  -8,  29,  -7, -10,   2,   0,  19,  16, -14,  10,  15, -11,  -6,   0,  -7,  12,   1,   1,  -4,  -3,  -3,  -1,   3,   2,   1,   7,  -2,  -2,   0,  -1,  -1,   4,   2,   3,   1,  -2,  -1,  -1 },
      {   1,   9,   5,   6,  -2,  32,  -7,  -5, -15,  12, -24, -14, -70,  -2,  25,   3,  -5,   3,   9, -30,  10,  -3,  -6,   3,  -9,  58,  10, -40,  15,  14,  -9,  -6,  22,  -2,  -5,   1, -12,  -9,   8,   3, -17, -11,   5,  16,   8,  -4,  -1,   1,  -1,  -1,  -3,  -2,   2,  -1,   0,   0,   5,  -4,   1,   1,   2,  -1,  -1,   1 },
      {   0,   2,   2,   1, -11,   5,  -5,  -1, -10, -16, -31,  14, -40,   9,  22,   1,  -3,   3, -23,   3,  20, -10,  -6,   3,  14, -79, -31,   1,  13,  21,   1,   4,   8,  45,  12, -18, -15,  -8,   0,  -1,  -9,  11,   6,   5,   8,  -9,  -6,   0,  -3,  -2,   4,   0,  -1,   0,   0,   0,   0,   1,   2,   0,  -1,  -1,   0,  -2 },
      {  -7,  -8,   9, -15,  14, -41, -14,   6,  -8,  -3,   1, -13, -39,  11,  12,   0,  -5, -14,  18,  -9,   8,  26,  -2,   1, -11, -14,  46,  -9, -19,   5,  -4,  -1, -74,  12,  17,  10,   3,  -2,   9,  -1,  30,  -4,  -7,  -6,   9,   0,  -4,   1,  10,  -3,  -1,   0,  -4,   0,   0,   1,  -2,   0,  -6,  -1,   6,  -1,  -1,   2 },
      {   3,   1,   7,  16,  -4,  38,   9,  -5,  -6,  -7,   1,  30, -41, -25,   7,   5,  11, -10,  20,  25, -15, -21,  -5,  -1,  31,   1,  40,  60,   5, -23,  -6,  -4,  -7,  -1, -21, -26,  16,   9,  -4,   3,  17,  -1, -15, -12,   4,   9,   0,   0,  -5,   1,  -2,  -3,   1,   0,   0,   0,   0,   2,   4,  -3,   1,   3,  -2,  -1 },
      {   1,  -8,  -9,   1,  -4, -30,  -9,   2,   8,   6,   9,   6, -16,  19,   8,  -9,  -4,   2,  -1,  19,  -2,   2,   7,   2,  10, -33,  53, -27, -32,   2,  -8,   9,  58, -19, -54,  20,  14,   5,   4,   1, -16,   7,   7,  -5,   7,   2,  -2,  -2,  -2,  -1,  12,   2,  -6,  -5,   0,   1,  -2,   2,  -7,   0,   3,   0,   1,   0 },
      {  12,   8,  17,   2,  38,  18,  12,  -9,  10,  -5,  25, -12,  28, -19, -10,   3,  23,  -6,  21,  -4,  44,  -3, -24,   1,  24,  -4,  33, -12, -35,  15,   9,  -3,  16,  33,   8, -21, -43,  -4,  10,   2,  -7, -14, -11,  18,  10,  -6,  -2,   2,  -6,   0,  -4,  -1,  -3,   0,   2,   2,  -2,   2,   0,  -1,  -1,   1,   2,   1 },
      {   1,  -9,   4,   8,  19,   6,   2,   0,  -6,  -9, -12,   1,  -7, -12,  -2,   4,  10, -30,  15, -12,  52,   4, -18,  -1,   5, -24, -17,  17,  17,  10,   3,  -1,  -4, -82,  -3,  39, -14,   2,   6,   4,  -1,  24,  -6,  -1,   1,  -4,  -1,  -1,  -3,  11,  -3,  -7,  -4,  -1,   0,  -2,  -2,   4,   2,  -1,   0,   0,   0,  -1 },
      {   0,  10,   6,  30,  12,  75, -13, -12,  -2,   6,  11,   7,  19,  29,   2,  -8, -14,  -2,  -3,   4, -10, -28,   1,   5, -19, -26,   5, -33, -21,   5,   6,   5, -47,  -7, -15,  15,  26,  -8,   5,   0, -15,   2,   7,   1,   2,  -3,  -3,   1,   6,   0,   6,   1,  -8,  -6,   1,   2,   4,   1,   0,   2,  -3,   0,   0,   0 },
      {  -1,   4,  -4,   3,  23,  -8,  -2,  -6,  -5,  -2,  -6, -17,   2, -27,  11,   7,   1,   9,  -8,  10,  50,  18, -14,  -5, -15,  13, -28,  14,  -2,  -5,   9,  -2, -12,  27, -66,  -7,  35,  -9,   3,   1,   6, -16,  40, -16, -13,  17,  -3,  -2,  12,  -6,  10,   4, -14,  -5,   0,   0,  -1,  -1,  -1,   0,   0,   2,   0,   0 },
      {  -6,  -6,  -3,  -4,  -6, -21,   2,  -3,  -5,  -2,   8,  12,  -4,   1, -10,  -2,  -8,  -6,  -2, -11,  10,  -5,   1,   1,   6,   1,  14,  29,   1, -18,  -5,   0, -27,  -2,   8,  -4,   4,   7,  -1,   2,-102, -20,  28,  26,   0,  -2,   7,   2,  11,   4, -13,   0,   4,   1,  -1,  -1,   0,   2,  -6,  -7,   2,   0,   1,   0 },
      {  -2,   0,  -5,  11,  -5,   3,  -2,  -6,  -6, -22,  -2, -36,   0, -87,   1,  20,  -8,   3, -15,  18, -41,  12,  17,  -6,   6, -16,   9, -19,  -8,  36,  -4,  -5, -15, -11,   0,   6,  -4,  -3,  -1,   4, -15,  14,  13,   7,  -1,   7,   1,  -2, -11,   3,  -5,  -4,  11,  -1,  -1,   1,  -2,  -2,  -4,   2,   2,   2,   0,  -1 },
      {  -3,  -3,  -3, -16,   4,   0,  26,   2,   2,   6,   5,   3,  10,   8,  -1,  -8,  -5,  -4,  -5, -19,   2,  16,   0,  -5,   3,  -4,  13, -28,  34,  -7, -15,  11, -26, -13, -37, -70, -10,  23,   6,   4,   0,  50,   8,  10, -15, -17,   1,  -4, -28,   5,   7,   5,   4,  -4,  -4,  -1,   4,  -1,  -2,  -1,  -1,   0,   1,   0 },
      {  -1,   7,   6,   9,   4,   5, -15,  -7,  -1,  14,   4,   3,  -4,  -4,  19, -13,   1,  13,  12,  15,  15,  22,   7,  -6,  -4,  19,  -4,  13, -22, -30,  -1,  10,   0,  24,  25,  28,   1, -18, -10,  -3, -20,  64,   7, -11,   3,   0,  -1,  -4, -62,   7,  18,  -1,   4,  -4,   1,   3, -11, -11,   0,  -1,   3,   4,   2,   1 },
      {   4,   8,   2,   8,  14,   6,  37,  11,  -3,   6, -14,   4, -31,  17, -34,   5,   1,  12,  -7,  27,  -9,  46,  -2, -15,  -3,   6, -21,  -5, -45,   1,  32,   2,   1, -23,   7, -33,   1, -17, -16,  11, -16,  17, -11, -12,  23,   3,   1,  -2,  29,  -1, -12,   3,  -4,  -2,   3,   0,   9,   2,   2,   2,   2,  -4,  -1,   0 },
      {   2,   5,   4,   3,  15, -28,  82,  -9,  -2,   2,  -5,  -8,  -9,  -2,  -6,   3,  -5,  -4,  -8,  -7,  -8, -58, -24,  13,   0,   9,   2, -11,   5,  11,   5,  -2, -10,  19,  -6,  33,  14,  -2, -10,  -8,  -9,  26, -10, -20,   2,   0,   1,   1,  -4, -15,   2,   5,  -2,  11,  -4,   0,  -3,   4,   3,  -1,   5,   4,  -1,   0 },
      {   3,   6,   4,  13,  -1,  15,   6,   5,   1,   3,   8,   4,   5,  24, -23,   2,   4,  11,   0,  24,  -6,  30,  -8, -15,  18,   5,  34,   8,  54,  21,  -2, -10,   3,  20,   1,  32, -24, -24,   7,   2,   4,  23,  46, -17, -22,  -9,  -1,   2,  33,  -6, -18,  -1,   6,   4,   1,  -2,  -1,   2,  -6,  -2,   1,  -4,   1,   1 },
      {  -2,   1,  -8,  -4, -10, -10, -20,  -2,   6,  13,   7,   8,  20,  -1,  44,  10,  -4,   4,  -6,  -2,  -8, -13,  -4,   2,  -5,  21, -10,  24, -16,  22,  -9, -16,  -9,   7, -26,  -6, -25,  -2,  26,   4, -13,  47, -26, -10,  34, -13, -19,   0,  52,  17,  -7, -13,  -1,   4,  -6,  -1,  -5,  -2,   8,  -1,  -4,   1,   1,   0 },
      {   1,   4,  -2,   6,   0,   3,  13,  -1,  -3,  -4,  -8,  -3,   7,  -8,  -6,   1,   0,   9,  -2,  13,   3,  34,  -4, -12,   5,  -9,   1, -20,  40, -31, -14,  16,  -8,  23, -24,  29,   9, -13,  -3,  -2, -13, -14, -72,  30,  17,   5,  12,  -3,  12,  29,   6, -15,   0,  -6,  -1,  -1, -13,   0,  10,  -1,  -4,   2,   4,  -1 },
      {  -3,  -1,  -3,   0,  -5,  -2, -49,  37,  -4,  -2,  -9,  -6,  -7,  -1, -69,   7,   1,  -6,   2,  -8,  11, -21,  14,  -1,   7,  12,   0,   7,  10,  25,  14,  -3,  -9,  18, -26,   2,   1,   0,   8,  -3, -16,  17, -32, -15,  16,  -3,  11,   2, -19, -29,  16,  16,  -1,   5,   2,  -3,  19,   1,  -3,   4,  -4,  -4,   4,   0 },
      {   2,   0,  -1,   2,   0,   1, -13,   3,   1,   2,  -2,   1, -15, -17, -25,  12,   4,   0,   4,  -8,  14, -23,  23,   3,  -1,  -4,  -7, -12, -20, -45,   1,  13,   4,  11,  -4,  11,   0,  16, -14, -10,  25,  40,  14,  52, -18, -15,   5,  -8,  44, -20, -34,  -3,  10,   2,  -2,   3, -16,  -7,   1,  -4,   8,   9,   0,   0 },
      {   1,  -1,   1,  -6,   9,  -2, -24, -45,   0,   1,  -1,  -9,   5,  -5,  -5,   5,   4,   3,  10,  -4,  21,  -2,   6,   7,  10,   0,  16, -16,  20,  19,  -6,  -5,  17,   1,  39, -27,  55,  12, -22,   4, -16,  24, -21, -34, -17,   8,  12,  -1,  32,   8,  -2,   8, -18,  -8,   7,  -2, -21,  -6,   7,   7,  -1,  -3,  -2,  -1 },
      {   4,   5,   4,   5,   7,   2,  -4,  84,   5,   2,   4,  -5,  12, -10,  35,   3,   1,  -1,   1,  -2,   9,  -9, -24, -16,   6,  -4,   9, -13,   5,  -6,  -4,   4,   8,   3,  18,  -5,  37,  -5, -11,  -3,  -7,  11,  -4,  -4,  -9,  12,  -2,  -4,   2,  12, -39,  -7,   2,   6,   7,  -5,  43,   2, -10,   3,  -4,   0,  -4,   0 },
      {  -2,   1,  -4,  -2,   1, -12, -10,  -5,   0,   6,   0,  13,  -6,  11, -18,   0,  -6,   5,  -6,   3, -11,   0, -40,   0,   2,  11,   2,  18, -10,  45,  27, -15,   2,   4,  -3,   0,  43,  -5,   1,   4,  16,   7,   3,  58, -17, -33,   0,   0, -22,  41,  -2, -23,  -1,  -1,   6,  -3, -12,  -2,   4,  -8,   1,   1,   2,   0 },
      {  -2,  -6,  -5,  -1,   4,  -3,   3, -19,   3,   1,   3,   8,  13,  11,  33,   0,   0,   2,   3,  13,  11,  15,  20,   3,  -1,   3,   1,  10,   6,  25,  -4,  -9,  -4, -10, -12,  -2,  15, -29, -22,  12,  -3,  -4, -28,  28,  -3, -10,  -1,   0, -21, -75, -40,  12,  17,  22,  11,   2,   1,   9,   4,  -7,  -4,  -7,  -4,   1 },
      {   3,   3,   7,   0,   6,   0,  26,  21,   1,   1,   3,   5,   4,  -2,   9,  -8,   8,   2,  16,   2,  18, -10,  76,   4,   0,   6,   7,  12,  10,  27,  -1,   2,  -1,   0,   5,  -1,   6, -17, -35,   6,   7,   0,  11,  27,  12, -11,  -6,   3,  17,  21,  48, -13, -26,   4, -10,  -6,  19,  -3, -12,   2,   0,  -3,  -2,   1 },
      {  -1,  -4,  -2, -10,  -2,  -6,  -9,  -4,   1,  -1,   4,  -5,   6,  -2, -23,   1,  -2,  -5,  -3, -12,   1, -17,  -4,  11,   4,  -5,   9, -13,  17, -22,   0,   3,   1,  -9,   1, -18,  12, -47,  -7,  19,  11,   1,  30,  -3,  58,  24, -28,  -5, -11,   5, -14, -40,   3,  37,  16,  -5, -25,  -5,  12,   4, -20,  -7,  13,   2 },
      {   1,  -1,   4,   5,  -2,  -1,  -4,  16,   1,  -2,   3,   0,   8, -16,   9,   1,  -5,  -1,   1,   5, -14,   7, -46,  -7,   1,   1,   5,   1,  15, -13,  -1,   5,   7, -12,  23,  -4,  15,  -5,   5,   2,   1,   7,  10,  32,  27, -15, -19,  -4,  23, -46,  58,  46, -20, -12,   1,   0, -11,  -6,  -1,  -6, -16,   6,   7,   2 },
      {  -1,  -3,  -3,  -6,  -6,  -5,  -9, -49,   4,   2,   0,   1,   1,  -1, -10,  23,   2,   2,  -2,  -1,  -7,   6, -19,   5,  -1,  -2,  -2,  -2,   0, -18, -14, -13,   2,   4,   3,   7,  -4,  -1,  -5,   3,   5,  15,   1,  13,   5,   9,   3,  -8,   0,  -7,   9, -26, -31,   7,   6,   5,  90,  24, -22, -10,  -3,  -2,  -5,   0 },
      {  -2,  -4,  -1,  -4,  -9, -14, -18,  -4,   1,   1,   4,  -1,  -1,  -2,  21,  19,  -1,  -5,  -1,  -2,  -8, -25,  -7, -18,   7,  -2,   9, -10,   5, -24,  61,   8,  -6, -11,  -7, -12, -25, -54,  -8,  24,  -7,  -1, -15,  -5, -50,   0,  15,  -6,   8,   5,  25,   3,  -5,  17,  -4,  -4,  -2,   0,   1,  10,  17,   0, -11,  -6 },
      {   0,   2,   3,   2,   1,   5,   5,  21,   1,   7,   4,   7,   1,  23,  -3,  89,   0,   1,   3,   1,   2,   3,  -5,  -8,  -4,  -5,  -6,  -4, -14,   1, -48, -40,  -3,  -2,  -2,  -2,  -5,   3, -19,   0,  -4,  -4,   2,   1, -13,   9,  13,  -5, -12,  -3,  22,  -8,  -7,  13,   4,  -3, -31,  -6,   0,   8,   7,  -3,   6,  -3 },
      {  -1,   0,   0,  -1,   6,  -1,   3, -14,  -1,   3,   2,   3,   4,   6,  14,  43,   1,  -1,   3,  -1,   5,   3,  21, -40,   2,  -2,   9, -11,  21, -13,  63, -18,   3,   8,   3,  12,  13,  48,   0,   0,  -2,  -7,   4,  -7,  28, -17, -40,   0, -12,  -3,  -3,  -6,  25,  -9, -13,   6,   6,   2,  -6,  10, -11,   1,   4,  -3 },
      {  -1,   1,  -1,   1,  -3,  -7,   1,  -3,   0,   6,   2,   1,   0,   3,  -4,  19,  -1,   1,   2,  -3,   4, -17,  20, -14,  -2,   1,   1,  -2,   2,  -2,  -5,  -7,   1,   2,   3,  -1,   1, -24,  11,  16,   9,   5,   6,   8,  13,  18,  -2,  -8, -11,  33, -22,  61,  -8, -36,   5,  -5,  -7,  66,  11, -43,  -4,   1,  -1,  -4 },
      {   0,  -1,  -2,   1,   2,  -6,  -3, -10,   2,   1,   1,   3,   2,   9,   0,  15,   1,   1,   0,   3,   4,  -1, -10,  48,   3,  -1,   4,   3,   5,   3,   6,  -7,  -6,  -3, -10,   1,  -9, -15, -32,  -6,   5,   3,  -1,  12,   6,  28,  -7,  -7,   4,  26,  -2,  45,  53, -28,  -1,  -2,  25, -52, -17,  15, -16,  -9,   7,   4 },
      {  -1,   0,  -1,  -1,  -4,   6,  -2,   6,   0,   2,   0,   3,   4,  12,  10,  -9,  -3,  -3,  -3,  -4,  -6,   3,  -5,  -3,   1,   6,   3,   6,  10,  22,  31,   9,  -3,   5,  -2,   3, -17,  33, -11,  -9,  -1,  16,  -1,  27,   4,  85,   1, -24,  -1, -14,   0, -15, -30,  -9,  16,   7, -25,   2, -18,  -8,  22, -21,  -7,   5 },
      {   0,  -2,  -2,  -1,  -1,  -6,   4, -19,   1,  -4,  -3,   3,   0,   3,  -8, -12,  -1,  -8,  -4,  -6,  -2, -20,   3, -74,  -1,  -1,  -5,   4,  -4,   4, -26,   0,   0,   2,  -4,   4,  -2,  -8,   4,  28,   7,  -5,  -2,   5,   9,   1,  -8,   3,  -8,   7, -26,  26, -32,  -7,  15,   0,   7, -63, -21,  16,  17,   7,   0,   5 },
      {  -1,   0,  -2,   0,  -6,   0,   8,  11,   0,   1,   1,   4,   3,  -7,   4,  17,   0,   5,  -1,   7,  -5,  11,  15,  57,   3,  -3,   3,  -5,  13, -18,  25, -23,   0,  -2,  -2,   1,  -2,  -8,  25, -16,  -4,  -2,  -1,   0,   5, -22, -10,   9, -17,   0, -33,  11, -70, -19,  21,  -6,  -4, -21,   6,  24,  16,   4,  -9,   0 },
      {   1,   1,   0,   0,   3,   6,   4,   1,  -1,  -1,  -6,  -2,   3,  -2,  -5,  -2,   1,   1,  -1,   0,   3,  -1,   4,  14,   0,  -2,  -6,  -2,   4,  -9,  -2, -16,   7,  -4,   7, -12,  16, -21,  31,   1,  -1,   1, -15,  -5,  -2,  -1, -17,   7,   2,  -5,   2, -11,  19,  -5, -18,  -4, -19,   6, -96, -25,  39,  11,  17,   2 },
      {  -1,   1,  -1,   0,  -4,   4, -19,   4,  -1,   0,  -2,  -5,   1,  -7,   2, -17,  -1,   2,  -3,  -1,  -6,   1, -23,   4,   0,   3,  -2,  -7,   3,  -6,   0,  -8,  -9,   2, -19,   6, -25,  21, -85,   4,  -1,  -9,  -2, -19,   8, -35, -22,  18,   0,  14,  -6,  12, -14,  11,  32,  -2, -14,  23, -25, -17,  16,   1,   3,   4 },
      {   0,  -2,   0,  -4,  -1,  -5,  -6,   1,   2,   1,   4,  -1,   7,   5,   4,  24,  -2,  -4,   0,  -6,  -2,  -9,  -8,  -5,   2,   5,   3,   5,   8,  13,   1,  40,  -1,  -6,  -3,  -9,  -2, -23, -23, -21,   4,  -7,  11, -15,  29, -20,  33,  -5,  -4, -16,  -9, -34,   0, -65,  -8,  38,  -3, -10,   0, -19,  16,  28, -14,  -6 },
      {   0,   2,   1,   3,   0,   3,   1,  -5,   0,   3,  -1,   1,  -3,   6,  -1,  39,   0,   3,   0,   4,  -2,   8,   4,  17,   0,   8,   1,   8,  -2,  17,   0,  82,   0,   1,   3,   3,  -2,  13,  11,   2,  -3,  -9,  -6,  -9, -15,  -6, -30, -24,  -4,  10,  -7,  13, -16,  38,   9, -15,   1, -17, -12, -34, -19,  23,   8,  -6 },
      {  -1,   1,  -1,   1,  -2,  -2,   0,  -6,   0,   4,   0,   6,  -1,  -4,   0,  11,  -1,   1,   0,   0,   0,  -5,  -7,  -2,  -1,   9,  -2,  11,  -4,   5, -13,  46,  -2,   2,   1,   1,   0,  -6,  -1,   1,   3,   0,   2,   5,   1, -12,  -3,  -1,  -2,   1, -10,   5,  -4, -14,  -5,   4, -11,  44, -33,  87,  -9, -43,  13, -11 },
      {   0,   0,  -1,   0,   3,   5,  -2, -10,   1,   1,   0,   2,   9,   1,  10,   8,   0,   1,  -1,   2,   3,   6,  -2,  -3,   1,   0,   1,   1,   8,   2,   8,   9,   4,  -2,   1,  -2,  12,   6,  10, -30,   7,   0,  16,   5,  47,  -8,  51,  -6,   2,  17,   4,  33,  18,  55,  11, -10,   5,   4,  -4,  12,  55,   2, -33,  -6 },
      {   0,  -1,  -1,   0,  -1,  -1,  -2,  -8,   0,   1,  -2,   0,  -5,   1,   8,  -6,  -1,  -2,  -1,  -2,  -2,  -1,   3, -23,   2,  -1,   2,  -1,  -3,   0,  12, -17,  -2,  -8,  -5,  -8,  -7, -15,  -7, -76,   1,   1,  -1,   2,  -1,   5,  26,   9,  -2,  10, -12,  13, -25,  26, -19,  18, -12,  -2, -27,  -6, -62,  11,  28,  -3 },
      {   1,   1,   1,   1,   2,   1,   8,   6,   0,  -1,  -1,  -1,   0,  -5,  13,   2,   1,  -1,  -1,  -1,  -2,  -3,   7,   8,  -2,  -1,  -2,   0,  -1,  -9,  14,  -4,   1,   6,   0,   8,  -1,  16,  15,  43,   5,  -1,   5, -11,  11, -26,  62, -22,   2,   0,   1,  -6,   1,  -6,  40,   0,  -9, -14, -31, -24, -45, -44,   1,  19 },
      {   0,   0,   1,  -1,   2,  -3,   7,  -3,   0,  -1,  -1,   0,  -3,  -4,  -5,   0,   0,  -1,   0,  -1,   3,  -7,  13, -17,   0,   1,   0,   2,  -3,  -2,  -7,  -1,   0,  -5,   1,  -7,   3, -18,  18, -54,  -5,  -3,  -7,  -4, -19, -11, -36, -19,   9,   4,  20,  -6,  23,  -9,  82,  28,   6,   1,   7,  -6,  11, -29,  -4,  10 },
      {   0,   0,  -1,  -1,   1,   0,   6,   1,   1,   1,  -1,   1,  -1,   6,   6,  -1,   0,   0,   0,  -1,   1,   2,   7,   4,   0,   2,   0,   3,   1,   4,   8,   0,  -1,   2,   1,   2,  -2,   8,  10,  22,   1,   8,   2,   8,  -1,  20,  19,  39,   6,  -2,  10,  -4,  15,  -4,  63,   6,   0,  16, -10,  29, -17,  78,  18, -26 },
      {   0,   1,   0,   0,   0,   1,   0,   3,  -1,  -1,  -1,   0,  -1,   0,  -5, -21,   0,   0,   1,  -1,   0,   1,   2,   3,  -1,   0,   0,   1,   0,   3,  -5, -27,  -2,   1,  -1,   1,  -3,   0,  -7,   6,  -3,  -8,  -3,  -6,   0, -11,  -5,-106,  -1,   1,  -2,   2,   0,   1,  -4,   6,  -6,  12,  -8,  28,  -9,  47, -12,  -5 },
      {   0,   1,   0,   1,   0,   1,   0,   7,  -1,   0,  -2,   0,  -1,   0,  -9,   1,   0,   1,   0,   1,   1,   3,  -1,  11,   0,   0,  -1,  -1,   1,  -2,  -6,  -3,   1,   3,   0,   2,   4,   3,   6,  23,  -1,   3,  -3,   4,  -5,   5, -16,  19,  -1,  11,   0,  12,  -3,  24, -12,  88,  -9,  -2, -15,  -2, -20,  -5, -71, -19 },
      {   0,  -1,   0,   0,  -1,   2,   0,   5,   0,  -1,   1,   0,   1,  -1,   6,   1,   1,   0,   0,   1,   0,   2,   2,  11,   0,  -1,   0,   1,   4,   2,   5,   1,   1,   0,   2,   1,   4,   6,   2,  21,   0,  -2,   1,  -1,   1,  -3,   9, -11,   1,   8,  -3,  14,  -6,  20,  -7,  73,  11,   1,  20,   1,  35,  -3,  84,  23 },
      {   1,   0,   0,   0,   1,   0,   1,   2,   0,  -1,   1,  -1,   1,  -2,   1, -10,   0,   0,   0,   1,   0,   0,   0,   1,  -1,   0,   1,  -1,   0,  -3,   0,  -8,  -1,   0,   0,   2,  -1,   3,  -1,   1,   1,  -2,   1,  -2,   2,  -5,   2, -13,   1,  -3,   0,  -1,  -2,  -2,   3,   0,   4, -14,   4, -17,   9, -29,  25,-118 },
    },
    {
      {  85, -68, -12,  13,   4,  -2,  -1,   1, -46,  24,  18, -13,  -1,   1,   1,   0, -18,  19,  -6,  -8,   3,   2,   1,  -1,   7,  -7,  -7,   9,  -1,  -2,   0,   0,   1,   0,   1,   0,  -1,   0,   0,   0,   0,   2,   0,  -2,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0 },
      { -42, -22,  19,   0,  21,  -1,  -2,   2, -20,  77, -18, -34,   3,  -1,   2,   0,  32, -24, -45,  17,   4,   3,   5,  -1,  -6, -20,  11,  21, -15,  -4,   4,   0,   8,   5,   2,  -7,   3,   3,  -3,   0,   5,   1,  -2,   0,   2,  -2,  -1,   1,   3,   1,   2,   0,  -5,  -1,   2,  -1,   0,  -2,  -1,   1,  -2,   1,   0,  -1 },
      {  -7, -48,  87, -22, -35,  14,  -1,  -2,  -1, -30,  16,   8,   8,   5,  -1,  -1,  16,  10, -19,  19,   9, -18,  -6,   3,  -9,  14,  13, -24,   3,   2,  -4,  -3, -13,   3, -11,  -7,   3,   2,   3,  -1,  -8,  -7,  -6,   2,   2,   5,   2,  -2,   5,  -1,  -5,   3,   3,   3,  -1,   0,  -2,   2,   0,   2,   1,  -2,   1,   1 },
      {  74,  45,   4, -38,  -8,  -2,  -1,  -3,  -2,   8, -34,  13,   5,  -4,   0,  -1,  35, -42, -10,  35,  -6,  -3,   0,   1, -21,   1,  26, -11,  -6,   5,   1,   1, -11,   6,   4,  -8,   2,   1,  -1,   0,  -1,  -1,   2,   4,  -1,  -2,   0,   1,  -5,   1,  -1,  -1,   2,   0,   0,   1,  -1,   0,   0,   0,   1,   0,   0,   0 },
      { -17,  39,  -5,   2,  -9,   4,   2,  -3, -95, -10,  30,  15,  10,  -2,   2,  -1,  45,  12,   3,   0,   0, -12,  -5,   1,  32,  -4, -14,  -9,   4,  -1,  -2,  -1, -13,  -9,   9,  -4,  -4,   2,   5,   0,   3,   1,   1,   1,  -2,   5,   2,   1,   1,   0,  -3,   1,   5,  -1,  -2,   0,   2,   0,   2,  -3,   0,   0,  -1,   1 },
      { -17, -41, -22,   2,  29,   3,  -4,   5, -19,  -5,   6,   4,   3,   5,   3,   2,  15, -56,  59,  11, -30,   6,   1,   0,  -5,  27,  14, -34,   2,  -5,  -4,  -4,   2,  21, -36,  -8,  21,  -5,   1,   0,  -2,  -3,  -6,   6,  -2,   2,   3,   0,   6,   4,   2,  -1,  -4,   3,   1,  -1,   0,   1,  -1,   2,   0,  -1,  -1,   0 },
      {   2, -21, -12,  80, -28,  -7,  14,  -3,  10,  -8, -44,  15,   3,  -4,   2,  -2,  57,  14,   1, -17,   8,   0, -11,  -1, -14,  12,  22,  -8,  -1,   6,   3,   2,  -1,   4,  12,  -5,  -9,   0,   4,   0,   3,   5,  16,   9,  -8,  -4,   0,   3, -11,   5,  -3,  -8,   9,  -4,  -3,   1,   1,  -3,   3,  -6,   4,   3,  -2,   1 },
      {   8, -13, -25,  -9,  19,   5,  -5,   3,  32, -30,  57, -12, -24,   6,   0,   3,  61, -20, -32, -20,   2,   9,   3,   1, -22,  -4, -23,  19,  17,  -9,   1,  -1, -27,  19,  19,  -2,  -2,  -3,  -1,   0,   0,   7,  -8,  -4,   3,  -1,   0,   0,   3,  -2,   0,   7,  -5,  -1,   1,  -1,  -1,   2,   0,   1,  -2,   0,   0,  -1 },
      { -10, -20,  -6, -45,  37,   7,   1,   1,  -9,   9, -18,  51,  -6, -23,  -1,   0,  20,  49,  -7, -21, -34,   7,   9,  -3, -40,  -2,   3, -13,  13,  22,   1,   3,  17, -13,   5,  29,  -1,  -6,  -6,   0,  10,  -1,   6,   0,  -4,  -4,  -2,  -1,  -7,   0,   6,  -2,   0,  -1,  -1,   1,   4,   1,  -2,  -4,   2,   0,   1,   1 },
      { -15, -28, -38, -12,  22,  13,   2,   4, -10, -55, -26,   8,  19,   1,   4,   6, -14,   9, -26,  42,  -2, -14,   9,  -2,  22,  -3,  26,  17, -31,   6,   0,  -3,  -2,  15,  37, -32,  -5,  10,  -8,   0,   5,  10,  -8, -12,   8,  -5,  -3,   1,  -1,  -1,  -1,   4,  -4,  -5,   1,  -1,   0,   3,   0,  -1,  -2,   1,   0,  -1 },
      {   2, -14,  -2, -10,  -7,   2,   1,  -1,  36,  29,  -4,   7,   3,  -3,  -4,  -4,  20,  17,  32,   1, -28,  -5,  -9,   1,  56, -58,   3, -20,  -3,  20,  -7,   2, -52,  -1,  15,  -3,  11,   3,   0,   0, -17,   3, -11,  -4,   8,   0,   3,  -1,   5,  -6,  -5,   7,   5,   3,   0,   0,  -3,   3,   2,   2,  -1,  -2,   3,   1 },
      {   3, -14,  38,   2,  18, -23,   4,  -1, -11, -27, -35,  19, -23,   5,  10,   2,  -2, -42,  -8,  -5,  -9,  38,   3,  -1,  46, -10, -32,  16,  22,   6,  -9,   5,  14,   2,  17,  23, -13, -17,  -5,   2, -13,  13,  26, -18, -11,  -6,   1,  -1,  -3,   0,  -7,  -8,   3,   1,   0,   1,  -1,   1,  -1,   2,   0,   0,   1,  -1 },
      { -12,   1,  -6,  16, -36,  -5,  12,  -7, -15,  13,  36,  47, -23, -18,  -7,  -4, -36, -33, -15,  -7,  -1,  16,   1,   7, -19, -31,  38,  -3,  12,  31,  -5,   2,   6,  39,  21,  -9,  -2,  -7, -10,  -3,   6,  -8, -17,  -5,   3,  -6,   2,  -1,  -2,  -3,  -1,  -2,   3,   2,  -2,   1,   0,  -3,   0,   0,   5,   0,   1,   2 },
      {  14,  24,  32,  48,  39, -33, -17,   1,   3, -21,  21, -10,  19, -11,   0,   2,  -5,  12, -26,   7, -44,  14,  22,  -9,  -3, -11,   1, -11, -17,  28,  -1,  -1,  -8,  -5, -14, -15,  35,  -1, -11,   6,  10,  -5,  -5,  22,   0, -13,   5,   1,   4,   1,   8,  -6,  -9,   5,   2,  -1,   1,   5,  -4,  -3,   3,   0,  -3,   0 },
      {   3,  -1, -18,   7,   5,  13,   1,  -3,  26,  33,  27,  27, -26, -21,   3,  -3,   2,   5, -37,  34,  -2,  -8,  10,  -2,  51,  61,  -6, -24,   4,   6,  -9,  -4,  11,  -7,   0,  -5,  -3,  -1,   1,   3, -16, -12,  16,  12, -10,   0,   4,   0, -10,   1,   1, -13,   3,   4,  -2,   1,  -3,  -4,  -1,  -4,   6,   3,  -3,   3 },
      {  -2,   5, -16, -19,  -6,   7,  -1,   4, -12, -12, -30, -26, -22,  17,  11,   3,  -3,  17, -30, -32,  19,  24,   0,  -3,  28,   4,  38, -19,  22,  -4, -23,  -2, -19,  22, -22,  10,  22, -29,   1,   4,   9, -14, -17,  31, -14, -10,  13,  -1,   2, -12,  19,   5, -21,  12,   4,  -2,  -2,   7,   3,  -6,   3,   0,  -2,  -1 },
      {  -7,  -2,  -1,  18,   9,   7,  -6,   1,  -9,  16, -27,  16,   6,   2,   4,  -3, -30,  -2,  -7,  27,   9, -11,   1,  -1, -28,  15, -53, -19,  35,  -5,   3,   3, -66,  19,  22,   1,   7,  -2,   7,  -3,  31,  -1,  -2,   6, -12,   3,  -1,  -2,  18, -10,  -2,   5,  -5,   0,   0,   0,   2,   5,   7,  -4,  -4,  -3,  -1,   1 },
      {  -3, -10, -27,  -3, -28,  30,   4,  -1,   6, -12,   6,  -9,  55,  -8, -17,   0,   2, -17, -19,  19,   5, -13,   5,   2,  -3, -39, -28,  -8,  14,  16,   0,   4,  22,  -2, -15,  27,  21, -28,  -9,   3, -13,   3,  28,  17, -34,  -7,  10,  -1,  -7,  12,  13, -16,  -3,   9,  -2,  -3,   4,  -2,  -4,   3,   0,  -2,  -1,   0 },
      {   4,  -6,   0,  11,  21,  19, -15,   0,   2,  -2,  -9,  50,  17, -10,  -7,  -1,  -8, -39, -24, -46,  40, -15,   6,  10,  12,  -9,   0, -10, -21,  -9,  13,  -4, -16, -41, -22,   5,   7,  20,   6,  -2, -11,  11, -19,   8,  21,  -2,  -1,   4,   4,   8,  11,   7, -15,  -5,   5,  -1,   5,   9,   6,  -9,  -5,   2,  -3,  -2 },
      { -12, -11, -21,  -2, -13, -15,  11,  -1, -13,  -8,   3, -26, -28,  21,   0,   0, -11, -15,  -8,  17,  -3,  15,  -6,   0, -26,   3,  15, -25,   1,  19,  -5,  -1, -22, -80,  29,  17,  10,   6,  -5,   6, -22,  -6,   9,   1,  10, -14,   4,   0,  16,  15, -13, -11,   4,   0,   1,  -2,   4,   7,  -7,   3,   6,  -3,  -1,   1 },
      {   3,   8,  -3,  15, -20,  40, -24,  -1,  -7,  -2, -22,   3, -50,  26,  15,   5,   2,   5,  -9,  -1, -28, -26,  18,   2, -11, -21, -24,  -9,  17,  -1,  -7,  -2,  34,  -1, -13, -36,  30,  19, -12,   2, -24,  11,   7, -12,  15,  -2,  -4,   1,  12, -15,   6,  17,  -7,  -4,   7,  -3,   4,   0,   4,   5,  -5,  -3,   3,   0 },
      {   6,  11,  16,  18,  34,  58, -26,  -6,  -2,  -8,   2, -37,  -3, -40,   9,   8, -11, -10,   7, -10,  -2, -14,   8,   4,  -1, -12,  36, -11,  34,   9, -17,  -6,  -5,  -8,  16,   5, -34,  -6,  14,  -2,   1,   1,  10,  -3,  -8,  20,   0,   0,   0,  14, -13,  -5,  21, -10,  -9,   2,   1,  -4,   2,   6,  -3,   1,   3,   0 },
      {   2,  -4,   5, -14, -27,  18,   6,  -2,   0,   2,  -8, -31,   2,  -9,  -9,   9,   2, -21,   3, -42, -19, -13,  21,  12,  18,  32, -22, -15, -26,  37,  16,  -7,   8,  11,  31,  18,  12,  19, -22,  -5,  48,  -5,  -3,   6,   8, -22,  -7,   3,  -4,   2, -14,   8,   1, -14,   3,   0,  -9,   5,   2,  -1,   2,   4,   3,  -2 },
      {   0,   5,  -7,   1, -14,   1,  -4,   4,  -1,  12,  -7,   0,  57,   8,  -9,  -2,  -5,  -3, -23, -33, -32,  27,  12,  -3,  -5,  34,   2,  -5,   6,  -7, -15,  -7,  -4,   1,  18, -13,  -4, -19,  -3,   7, -45, -14,  -5, -34,   5,  22,  10,  -3,  23, -16, -10,  29,   7,  -1,  -3,  -3,  -3, -10,   7,  11, -13,  -8,   1,   2 },
      {   0, -11, -13,   6, -32,   2, -18,   7,   4, -11, -11,   8,  -4, -19,  14,   5,  10, -10,  -6,  17, -24,  11,  11,  -5,   8,  -9,  -9,  38,   3,  -4, -16,  -5,  -5, -28, -29,  14,  -8,   8,   8,   2,  40, -67, -21, -14,  10,  31,  -4,   3,   4,   4,  14,  -2,   7,   2,  -5,  -2,  -6,  10,   7,  -3,  -2,  -5,   2,   0 },
      {  -3,  -6,  -8, -10, -26, -14,  -9,   8,   7,   1,   5,  -3,   7, -22,  11,   3,  -2,  -3, -12,   2,  -8,  29,   2,  -9,   5,  -3,   0, -29,  11, -23, -10,   5,  10, -29,  -5, -26,   9,  -1,  17,   6,  49,  73, -12, -15,  -9,  14,   0,  -7, -22,   1,   5,  12,  10,  -2,  -3,  -1, -15, -12,  -3,   8,  -1,   2,   0,  -1 },
      {   0,  -3, -10,  -4,  12, -18,   8,  -3,   1,  -9,  -5, -21,   7, -19,  11,   0,   3,  -2, -12,   7,  20,  22, -33,   3,  -3, -20, -32, -57,   2,  22,   3,  12,  32,  10,  -2, -17, -35,  18,  12,  -4, -11, -30,  -7,  19,  29,  10, -12,  -2,  -3, -21,  12,  21,   4, -10,   0,   4,  -6,  10,   9,  -4,  -9,  -2,   5,  -2 },
      {   8,   2,  12,  11,  19,  21, -12,   7,  13,   9,  13,  21,  16,  62, -10,  -8,   6,  -2,   8,   4,   2,   4, -13,  -1,   8, -14,   8, -21,   6, -29,  -6,   9,  37, -19,  42,  -7,  -6, -17,  -8,   4,  42, -32, -15,   6, -11,  -9,  13,   2,  -6,   1,  -7,   2,  -6,   0,  10,  -5, -10,   9,   1,  -1,   0,   0,  -2,  -1 },
      {  -6,  -8,   6, -22,  -8,   2,   1,   2,  -4,   1,  -4,   2,   7, -13,  -7, -11,  -9,  -6,  12, -23,  -7,  17,   4,  -8,  -1,   1,  -8,  23,  20, -14,   1,   6, -16,  -9,  16, -68,   7,  11,   2,  11,  -8, -27,  35,  45,  -1,  -8,   0,   2, -31,  32,  14,   3,   2, -19,   4,   1,  10,  -1, -17,  -5,  14,   6,  -3,   0 },
      {   0,   2, -11,  -8,   3, -20, -14,   8,   4,  -2,   5,  11,   2,   7,   7,   2,  -2,   8, -19, -13,  10,  -1,  -8,   1,   8, -15,  15, -23,  -6, -15,   0,   5,  -1,  22, -18,   1,  13,  26,   2,  -4,  22, -18,  57, -15,  -8,   3, -18,   2,  51,  38, -32,  -5,   4, -19,  -2,   0, -13, -17,  -7,  30,  -2,  -3,   8,  -3 },
      {  -5,  -3,  -8,  -1,  -2, -30, -13,   3,  -4,  -5,  -6,  -9,  -7,  35, -21, -13, -13,  -8, -28, -14, -40, -36,  15,  12,   4,  -8,   8, -25,   6, -21,  32,   6, -11,  13,  -3,  17, -26,  19,  14, -13,  -9,  -3,  -4,   1, -11,  26,  -3,  -6, -44,   8,  18, -12,  24,   7, -11,   2,  11,   0,  -4,  -1,  10,  -8,  -2,   5 },
      {  -4,  -5,  -5, -10,  -8, -44, -57,  30,   1,   4,  -6,   0,   3,  -8,  13,  11,  11,   4,  20,  -2,  41, -19,  43, -17,  -3,  -2,  -9, -10,  11,  20, -12, -11,  15,   7,  20,  -6,  -7,   6,  -8,   6, -17,  -8, -28,  -2, -17,  -5,  22,   8,  -2,  20, -11,  -5,   0,  12,   3,  -8,   3,  -7,   0,   0,   3,  -2,  -7,   1 },
      {   3,   6,   0,   3,  11,   2,  37,   0,   6,   7,   3,   0,  28,  33,  34,  -9,   3,   4,  -7,   4,   8,  -3,  -1,  -5,   8,  10,   3,   6,  48,  29, -21,  -5,   1,   6, -16,   6,  11,  26, -16,  -5,   3,   7, -29, -18,  24,  -9, -15,   4, -25,  42,   6,  -7,   9, -23,  -6,   2,  22,  -3,  -6,  12, -14, -10,   3,   0 },
      {   4,  -5,   1,  -7,  14, -28,  56, -11,   6,  -2,  -4,   0,  -3,  -5,  10,  -1,   1, -13,  -2, -18,  -7, -54,   2,  11,   4,  -4,   0,  11,  17,   8, -13,  -4,  14, -22,  -3, -30, -13,  -1,  -3,   5,  13,   0, -11,  14, -38,   9,  21,  -4,  45,  -8,   8,  14,   7,  13, -18,  -5, -11,  -2,   7,   7,   9,  -7,  -6,   2 },
      {  -1,  -4,  -5,   2,   8, -27, -12, -17,  -2,  -6,  -8,  -6,   0, -17, -72,  10,   6,   4,  -2,  15,   8, -12,  -9,  25,  13,   5,  14,  12,  58,   0,  14,  -7,   7, -13, -12,  -3,  19,   3,  -8,  -1,   6,  -8,  -6, -12,  13, -16,  -1,   3, -12, -17, -25,  19,  -8, -14,  13,  -3, -10,  -7,   5,   4, -12,   9,   9,  -2 },
      {  -1,  -3,   1,   1, -12, -19, -10,   4,   4,   2,  17,  -1,  24, -10,  41, -12,  -1,  -7,   7,   3, -13, -13,   8,   1,   3,   1,  29,   7,  28, -25, -10,   4,   5,  -3,  22,  31,   0,  39, -14, -11,  -7,   8,  27,  16,  -6,  -8, -21,   3,   6, -38,  18,  11, -30,  -9,  18,   7,  -8,  19,   9, -29, -11,  14,   2,  -3 },
      {   2,   2,   1,   2, -11,  -2,  13,   6,  -1,  -4, -11,   0,  -3, -28,   6,   6,  -2,   3,  -8,   6, -30,  -7,  -9,   0,   4,  -2,  -9,  -1,   9, -47,  10,   9,  19,  11,  21,  21,   8, -12,  11,  -6, -20,   5, -48,  40,  30,  -6,   1,  -5,  29,  46, -23, -14,  -9,  -6,   0,  -1,  -5, -15,   0,  -3,  -7,  17,  -3,  -4 },
      {  -1,   1,   3,   1,  11, -13,  -5, -25,  -2,  -2,  -4,  -6,  -2, -25,  26,   3,  -4,  -3,  -2, -15,   6, -28, -45,  19,  -7,   6,  -5,  -4, -12, -11, -21,   9,   2,   3,  29,   1,  56,  -8,   0,  -1,  -9, -21,   1, -35,  -4,  21,   3,  -4, -40,  11,  11,  -9, -10,  23,  10,  -4, -21,   3, -10,  10,   1,  -8,   2,  -1 },
      {  -5,  -6,  -9,  -5,  -1, -10, -39, -26,  -1,   2,   1,  12,   9,  27,  16, -13,  -4,  -8,  -5,  -3,  -6, -19, -14,   1,  -4,   7,  -3,  28,  -8,  27, -32,   7,   0,  -2,  -5,  10,   1, -16,  35,   0,   1,  12,   3,  46,  23, -18,  -3, -11,  -8, -20, -16,   5,  32, -11, -11,  11, -35, -25,   5,  16,  -3,  -3,  10,   1 },
      {  -1,  -1,  -2,  -1,   5,   4,  15, -74,  -1,  -1,  -3,   0,   4,   9,   0,   3,   6,   9,   9,   6,  20,  25,  61,   2,  -4, -11,  -4, -11,   0, -18,   2,  -2,  -1,   4,   1,   2,   7,  16, -14,   3, -10,  -5, -12,   4,  -4,  23,  -7, -13,  -3,  -1,  -9, -18,  -1, -12,  -9,  10, -47,   6,  -1,  -1,  19,  -3,   1,   3 },
      {   3,  -1,   4,   2,  -3,  -4, -27, -15,   3,   3,  -1,   0,  -1,   9, -25,   7,   8,   1,  11,   5,   8,   6, -10,   0,  12,  14,   7,   7,   8,  23,   9,   5,  14,   1,  29,   7,  11,  -1,  11,  -3,   5,  14,  -3,   8,   8,  35, -16, -20,  41,  11,  66,   1, -21, -11, -26,   6,  18,  -7, -35,   4,   5, -16,  11,   5 },
      {  -1,  -3,  -2,   3,  -4, -14,  -1, -38,   1,  -2,  -1,  -2,  -5, -12,  -2, -16,   3,  -2,   2,   3,   1, -18,  22,   7,  -1,  -4,  -5, -16, -19, -26, -42,  18,  -7,   1,  -4,  12, -26, -21, -30,  22,  16,  -3,   9,   1,  23,  -9,   8,  -4,  -5,  -4,   3,  20, -12,  -7,  14,  -3,  60, -21, -24,  21, -11,   4,   6,   1 },
      {  -2,   0,  -3,  -4,  -9,   0, -32, -47,  -1,   4,  -1,  -3,   2,   6,   0,  -4,  -2,   6,  -3,   2, -22,  10, -31,  24,  -3,   8,  -6,   2,  -4,   9,  -8,  -6,  -6,   6, -14,  -8, -41,  10,  -6,   1,  -3,  14,  -7, -18, -23, -33,  22,  13,  14,  48,  23,   3, -20,  -4,  17,  -8, -16,  13,  24, -22,   0,   2, -11,   1 },
      {  -1,   3,   4,   5,   2, -11,  19,  23,  -2,   4,   4,   2, -11,   5, -23,  17,  -3,   1,   3,   7,   5, -22,  20, -17,  -6,  -4,   2,  -7,  -9,  -5, -27,  -7,  -4,  10,   3,  14,   0, -26,   2,  17,   0,   7,  29,  -5,  33,   7,  14,   1, -10,  33,  31,  31,   7, -23, -14,  -9, -44,  35,  18, -20, -21,   6,   7,  -4 },
      {   0,  -1,  -6,  -8,  -6, -11,  -9, -14,   5,   6,   0,  -3,  16,   4,  15,   6,   2,   0,  -4, -12,  -2, -17,   7,  10,   4,   2,   2,  -5,  19,   2,  11,  -6,   6,   3,  -7, -25, -10, -38,  14,  10,  14,   7,  15, -17,  46,  -8, -14,   4,  19,  -6, -17, -43,   6,  26,   3,   3,  10,  54, -29, -29,  13,   4,  -4,  -7 },
      {   0,   2,   2,   5,   1,   5,   3,  30,  -5,  -3,  -4,  -5,   2,  -7,   1, -25,   1,  -2,   2,  -4,  -2, -13,   4,  -2,  -2,   1,  -3,  -9,   7,  -1,   9,   4, -14,  -9,  -7,  -8, -28, -21, -27,   4,  -5,  -6,  -9, -19,  14, -25, -23,  12,   2, -11,  37, -29, -20,  -3,  12,   7, -56, -42, -22,  25,  22,   9,  16,  -3 },
      {   1,   1,   3,  -4,   0,   2,  10,   9,  -1,   3,  -1,   1,   6,   8, -37,  47,  -1,  -2,  -1, -12,  -7,   2,  -6,  -3,  -6,   0,  -8,  -3,  -4,  -5, -61,  -2,  -2,   5,   5,  -8,  -9,  41,   8,   7,   2,   6,   2,   5,   6,   0,  -1,  -9,   1, -16,   9, -65, -10,  20,   5,   3,   6, -13,  19,   0, -12,  14,  13,  -3 },
      {   0,   1,  -3,   4,  -1,   4,  -3,  -3,  -2,   0,  -6,   1,  -1,  -8, -19,  -6,  -3,  -6, -10,  -5,  -8, -12, -31, -60,  -2,   1,  -5,   1,  -2,  -5, -19, -12,   3,  10,  -8,  14,  -4,  23,  -5,  23,  -7,   8, -20,  -1, -19,   2,  13,  12,   3, -13, -17,   4,   4, -32,  13,  14,  -8,  39, -53,   4,  34,  -8,   8,  -6 },
      {  -1,   0,   0,   1,  -2,  -5,  -9,   6,   0,   4,   1,   8,  -2,  -6,   9,  45,   0,   5,   1,   2,   3,   7,  -3,  34,  -1,   3,  -1,   1,  -8, -22,   1,  -6,  -6,  -5,  -1,  -7,   0, -16, -38, -26, -10,   2, -13,  17, -23, -16, -14,  12,   6,  -8,  23,  -8,  58, -24,  10,   8,   2,  31, -15,  22, -24, -18,  22, -12 },
      {   0,  -1,   0,  -1,   5,   7, -10,  -6,  -2,  -3,  -4,  -2,   2, -13, -12, -35,   1,   0,   2,   1,   7,  11, -14, -18,   3,   2,   3,   2,  11, -19,   5, -19,   4,   9,   8,  10,   9,  27, -42,  13,   9,   9,  12,  13,  35,  -7,  19,  13,  14,  -1,  19,  -4,  46,  53, -13, -11,  -6,   3,  21,  17,   5, -23, -17,  14 },
      {   2,   2,  -1,   0,   4,   6,  -2,  28,  -1,  -4,  -4,  -8,  -2,  -9, -21, -56,   4,   5,   1,   1,   9,   8,  14,  38,  -1,   7,  -4,   5,  -5,   3, -36,  64,   4,   4,  -5,  -3,  14,  13,   0, -26,  -3,   5, -10, -12,  -4,  -2,  16, -10,  16,   1,  -5, -10,  23, -11,  -5,  -2,  -5,  19, -11, -29,  12,   2,   5,   6 },
      {  -1,   2,  -2,   4,  -4,   2,  -4,  19,  -2,   3,  -2,   5,  -2,   3,  -5,  -6,   1,   2,   0,   1,  -5,   3,  -5,  18,  -1,   6,   0,  10,   0,  16,  -6,   7,  -4,  -3,  -8,  -5, -25,  -1, -34,   1,  -1,  18,   3,  35,   2,  39, -20, -27, -13,  17, -35,   6, -41,  29,  18, -18, -19,  32,  -5,  26, -12, -36,  18,  10 },
      {   1,   3,   1,   7,   3,   5,   7,  18,   1,   2,   0,   5,  10,   8,  12,  34,   0,   4,  -1,   9,  -1,  12,  -1,  48,  -4,  -6,  -4,   1,   3,  -4, -14, -36,  -2,   5,  -4,  13,  -8,  18,   1, -22,   3,   2,  13,   7,  23,   2,  39,  -7,  -5,  -4,  -2,  26,  -1,   7,   0, -19, -11, -21, -50, -10,  47,   7, -19,  10 },
      {   0,   2,   1,   6,  -4,   6,   7,   0,  -1,   2,   1,   2,   6, -12,  -1,  12,   3,   2,   5,   4,  -3,   3,  22,  -2,   1,   4,   4,   2,   6, -10,   4,  26,   2,   2,   5,   5,  -7,  16,  48, -10,  -4,  -7,   1, -12,   3, -54,   8, -20,  -5,  -1,  14,  14, -18,  19, -13,  -4,  -1,  40,  -1,  68,  10, -16, -11,  -9 },
      {   1,   0,   1,   1,   1,  -4,   9,  -2,   0,  -2,  -3,  -1,  -5,  -2, -17, -49,   2,   2,   2,   3,   0,   8,  15,  16,  -4,  -6,  -8,  -5, -11,  -4, -30, -70,   9,   1,  12,  -3,   5,  -4,  36, -32,  -1,   2,  -4,   6, -12,   0, -30,  40,   1,   9,   1,   5,  -9,  16,   3,   5,   6,  12,  -9,  13, -11,   9,  20,  -4 },
      {   0,   1,   0,   3,   4,   0,  -1,  -1,   0,   1,   1,  -1,   5,   3, -14,  25,   0,  -1,  -1,   2,   2,  -4,   0,  -9,   0,   1,  -1,   0,   0,  -3, -26,  19,   1,  -1,   0,   5,   2,  -9,  -7,  -3,  -5,   1,  -7,  -1,  -6,  -8, -75,   3,   3,  18,   0,  38,  20,  28,  14,   2,   9,  -5,  23, -15,  66,   3, -10,  -1 },
      {   0,   1,   0,   3,   2,   8,   6, -18,  -1,  -1,  -2,   1,  -2,   8,  -2,   9,  -1,  -2,  -1,  -2,   1,   3,   6, -28,   1,   3,   2,   2,   3,  10,  10,  25,  -4,  -1,  -4,   0,  -7,  17,   5, -12,  -1,  -6,  -3,  -1, -12,  -7,  -6,  -1,   4,  12,   3,  25,  23,  56,  15, -22, -18, -11, -54, -18, -50,  39,  21,  -9 },
      {   0,   0,  -1,   1,  -1,   0,  -6,   2,   0,  -1,  -1,  -2,  -2, -10,  -2,   2,   0,  -3,  -1,  -4,  -2,  -4,  -8, -15,  -1,  -5,  -4,  -6,  -4, -19, -10, -21,  -1,  -5,  -5,  -8,  -5,  -3, -15, -19,   0, -11,  -2, -12,  -9, -44, -18, -43,  -9,   2, -15,   5, -13,   6, -62,  21,  -2,  -9, -15, -38, -17, -35,  12,  39 },
      {   0,  -1,  -2,  -3,  -1,  -4,  -1, -10,   0,   0,   0,  -1,   0,   0,  -6,   2,   1,  -2,  -1,  -3,   1,  -7,  -1, -42,   1,   1,  -1,   0,   2,   6,  -5,   1,  -1,  -7,  -3, -10,  -3, -16, -18, -92,   7,   1,  11,  -2,  17,  15,  27, -17,   2,   6,   2,  -2,  -6,   1,  28,  39,   2,   3,  13,  11,   6,  11, -18,   0 },
      {  -1,   0,  -2,   0,  -1,  -2, -15,   5,  -1,   2,  -2,   2,   0,   5,   3, -10,   0,   1,  -1,   0,   1,   2, -16,  10,  -2,   1,  -4,   1,  -2,   3,  -3, -13,  -3,  -1,  -4,  -2,  -1,  -1, -30,   8,  -3,  -1,  -6,  -1,  -9,   8, -11, -31,  -4,  -1,  -2,   1,   0,  -3, -49,   0,   4,  29,   4,  31,   4,  92, -18,   0 },
      {   2,   1,   1,   2,   3,   0,   6,   6,   0,  -2,  -1,  -1,   0,  -3,   3, -16,   2,   0,   1,   0,   5,  -1,   4,   3,   1,  -3,   0,   0,  -1,   1,   8, -33,   4,   4,   2,   2,   5,   0,  18,  23,   1,  -5,  -3,  -7,  -6, -11,  -3, -79,  10,   6,  15,  -3,  28,  -8,  67,   0,   4,   3,   8,  -1,   2,   9,  13,  36 },
      {   0,   0,  -1,   3,   1,   2,   0,   6,   0,   1,   1,   1,   4,   3,   9,   3,   0,   1,   0,   3,   1,   4,   2,  16,   0,   2,   0,   2,   4,  -1,  11,   1,   1,   0,  -1,   2,   2,   8,  -4,  21,   1,   1,   3,  -1,   3,   1,  31,  -2,  -1,  15,  -3,  18,  -2,  29,  -7,  76,   8,   0,  14,  -3,  22,  10,  74, -26 },
      {   0,   1,   1,   1,  -1,   2,   3,  11,  -1,   0,  -2,   0,  -4,   2, -10,   3,   0,   2,   1,   1,   0,   1,   4,  22,  -1,  -1,   0,   0,  -1,   4,  -7,   6,   0,   3,   0,   3,  -1,   5,   4,  29,   1,   0,  -1,   2,  -5,   5, -12,  10,   1,   6,   0,   8,   2,  13,  15,  80, -13,  -1, -23,   4, -34,  -5, -73,  15 },
      {   0,   1,   0,   1,   0,   1,  -1,   4,  -1,   0,  -1,   0,  -3,  -2,  -5, -17,   1,   1,   1,   1,   1,   1,  -2,   2,   0,  -1,  -1,  -2,  -2,   0,  -5, -21,   0,   1,   0,   0,  -2,   3,  -7,   2,   0,  -3,   0,  -1,  -3,  -4,  -7, -46,  -1,   0,   0,   0,   4,   1,  -2,  -4,  -2, -12,  -3, -16,  -5, -21, -27,-108 },
    },
    {
      { 122, -20, -15,  -7,  -4,  -2,  -2,  -1, -21,  -1,   0,   2,   2,   1,   0,   0, -16,  -1,   3,   2,   1,   0,   0,   0,  -7,   3,   3,   1,   0,   0,   0,   0,  -3,   2,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
      {  11,  68, -33, -16,  -2,  -2,  -2,  -2,  69,  -3, -37,  -4,   2,   1,  -1,   0, -39, -34,  13,  18,   4,   2,   2,   1, -20,  -2,  20,   6,  -3,  -1,   0,  -1,   0,   1,   4,  -2,  -2,   0,  -1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0,   0 },
      {  -7,  63,  14, -36, -16,  -6,  -1,  -3, -85,  13,  12, -12,   7,   8,   3,   1, -24, -26,  12,  14,  10,   4,   2,   1,  18,   0,  -4,   7,   0,  -3,  -1,  -1,  20,   0, -10,   2,   1,  -2,  -2,   0,   8,  -1,  -3,   1,   0,  -2,   1,   0,   4,  -4,  -2,   1,   0,   0,   1,   0,   4,   0,  -2,   0,   0,   0,   0,   0 },
      {  21,  44,  12,  -8,  -5,  -5,  -3,  -1,  17,  72,   8, -20,  -6,  -6,  -3,  -3,  61,  15, -34, -14,  -1,   0,  -2,  -1,  -7, -34, -17,   9,   7,   4,   2,   1, -18, -17,   5,   9,   3,   0,   2,   1,  -7,  -4,   4,   2,   0,   1,   1,   0,  -3,  -1,   2,   1,   0,   1,   1,   0,  -2,  -3,   1,   1,   0,   0,   0,   0 },
      { -13, -13, -79, -16,   8,  -1,   8,   1, -37,   2, -61,   7,  25,   7,   6,   1,  53,   0,   7,  10,   5,  -1,  -4,  -1,   5,  -3,  14,   7,  -1,  -2,  -3,  -1,  -8,  -3,  12,   5,  -2,  -1,  -1,   0,   0,   4,   7,  -2,  -3,  -2,   0,   0,   2,   1,   3,  -1,  -1,  -1,   0,   0,   2,   0,   1,  -1,  -1,   0,   0,   0 },
      {   9, -12,  61, -24, -52, -18,   4,  -1,  15, -34, -48, -14,   6,  18,   6,   2,  33, -20, -23,   8,  28,   7,  -4,   0,  15,  26,   5,  12,   7,  -9,  -6,  -1,  -6,  -3,   5,   6,   0,   0,  -1,  -1,   2,   8,   3,  -1,  -2,  -2,  -1,   0,   0,   2,   0,   0,  -1,   0,   0,   0,  -3,   2,   2,  -1,   0,   0,   0,   0 },
      {  15,  63, -11,  36,   6,  -4,  -5,  -2,  -8, -73,  16,  30,  -3,  -5,   0,   3,  34,  10, -19, -18, -12,  -3,  -2,   0,  24,  13, -11, -18,  -6,   3,   1,   1, -18, -10,   4,  -1,   3,   4,   2,   1,  -9,   2,   6,  -1,   1,   3,  -1,  -1,  -6,   1,   2,  -2,   0,   0,  -1,   0,  -4,   1,   3,   1,   1,   0,   1,   0 },
      {   0,   6, -20,  -5, -32, -15,  10,   1,  15,  14, -16, -21,  -1,   9,   4,   0, -28,  65,  33, -32,   4,   7, -11,  -3,  48,  28, -22, -15,  -1,  -2,  -7,  -1,   3, -37, -16,  13,   8,   5,   2,   2,  -4,  -7,   6,  10,   3,  -1,   1,  -1,   1,  -7,   4,   4,  -1,   0,   0,  -1,  -3,  -2,   2,   1,  -1,  -2,   1,   1 },
      {   3,  -6, -13, -79,  12,  29,  -1,   3,  28, -34,  46, -13,  -3,  10,  -6,   0,  34,   8,  28,  28,  -9, -13,  -3,  -1,  11,  -3, -27,  10,   2,  -5,   2,   3,   5,   3, -19,   1,   1,  -2,   1,   2,  -5,   3,  -4,   4,   2,   0,   3,   0,  -8,   1,   3,   5,   0,  -1,   0,  -1,   0,   3,  -1,   0,  -4,  -1,   0,   0 },
      {   0, -18, -31,  16, -13,  -5,   4,   4,  17,  20,  11,  15, -11, -10,   0,  -1,  -8, -61, -22,   8,  16,  11,  -1,   4,  70,  -6, -48,   2,  12,   2,  -2,   2,   9,  19,  -3,  -2,   3,  -5,   0,   2, -17,   2,  13,  -2,  -3,  -2,   0,   0, -10,   5,  11,  -2,  -1,   1,  -1,  -2,  -4,   1,   2,  -1,   2,   2,   0,   0 },
      {   2,   0,   4,  20, -43, -27,  -2,  -3,   3,  23,  14,  33, -18,  -9,  -3,  -3,  41, -29,  77,  11, -17,  -3,  -9,   0, -14,  18,   9, -16,   1,   3,  -1,   0,  -7,   3, -19, -13,  11,   9,   5,  -1,   6,   1, -12,  -7,   7,   8,  -1,  -1,   2,   1,  -6,  -1,   6,   2,  -2,  -1,  -4,  -1,  -1,  -2,   2,   0,   1,   0 },
      { -11,  -3, -36,   6, -50, -12,  13,   0,  -5, -26,  38, -10, -18,  11,   6,   2, -10,   0,   0, -23,  28,  15,  -7,  -1, -48, -22, -14,  30,  38,   1,  -5,  -3, -14,  -3,  19,  26,   8, -10,  -7,  -2,   8,  17,   4,  -4,  -9, -11,   1,   0,  12,   7,  -1,  -5,  -7,   0,   3,   0,   2,   1,  -3,  -3,   0,   2,   0,   0 },
      {  -3,  -9,  -1,  -1,  24,   8,   0,   5,  -4,   4,  21, -49,  -4,  10,  -2,  -1,  -3, -47,  16, -23, -22,  13,  -5,   2,  20,  28,  25,   5, -15,  -8,  -7,  -6, -69, -22,  25,  19,   5,   6,   5,   1,  -9,  -7,   2,   5,  -4,   3,   1,   1,  11,   9,  -5,  -3,  -5,  -4,  -1,  -1,   5,   1,  -4,   4,   0,  -2,   2,   0 },
      {  -2,  -1, -16, -27,  13,  -9,  12,   1,   3,  14,  15,  10,   5,   6,   2,   2,  13, -24, -32, -35,   6,  22,   4,   1, -42,  69,  -7, -44,   5,   5,  -3,  -1,  34,  -1, -17,   9,   2,  -4,   0,   0,  -6, -18,   5,  14,  -3,  -7,  -1,   1,   2,  -2,   7,   5,  -5,  -3,  -1,   2,   2,   0,  -4,   1,  -1,   0,   2,   1 },
      {  -9,   7,  -7, -13, -17,   4,  -5,   1, -10,  15,  -2,   3,  -5, -16,   1,  -6, -17,  29, -15,  32,   8, -12,  12,  -5, -14,  45, -24, -11,  15,  -6,  13,  -5, -66,  48,  19, -30,   4,  -6,   9,  -2, -11,  -5,   8,  -9,   4,   8,  -1,   0,  -3, -12,   8,   2,   4,   8,  -5,  -2,   2,  -7,   3,   3,   0,   1,  -2,  -1 },
      {   5,   8, -24,  40, -17,  12,  -6,  -4,   6,   3,  26, -57, -10,  21,   1,   0,  24,  11,  -7,   6,  18,  -4,   0,   0,  14,  29,  36,  22, -15, -14,   3,   0,  38,  37,   6, -26, -16,  -2,   1,  -1,   9, -11, -18,   4,   9,   0,  -1,   2, -12, -13,   3,   2,  -1,   1,   2,   1,   0,  -4,   1,  -1,   1,   0,  -2,   0 },
      {  -5,  -8, -14, -20, -23, -19,  16,  -2,   8,  11,  44,  41,  12,  -6,  -3,  -5,  -9,  14, -34,  37,  26,   8,  -4,  -2,  16,   0,  48,  -4, -36,  -1,  -7,   0,  -9, -40,  20,  -3,  -8,  15,  -2,   1,  -6,  -4, -11,  -3,  12,   2,   1,   2,   7,   7,  -4,  -1,   6,  -4,  -1,   1,  -4,   2,  -8,  -7,   0,   1,   2,   1 },
      {   0,  -1,   3,  13,  20, -63, -11,  -5,   9,  -6,  18, -21,  79,   6, -17,   2,   2,   1,  14,   6,  20,  20,  -8,  -3, -10, -10, -26,   1, -22,  -3,   8,  -1, -12,  16, -13, -13,  -6,   0,   6,   1,   0,  14,   8,   7,  -4,  -6,   2,   0,   3,  -2,   5,  18,  -9,  -6,   3,  -1,  -7,  -3,  -6,   6,   1,   3,   2,  -1 },
      {  -3,  -5,  -5,  14,  -7, -15,  36,   2, -12, -17,  -9, -41, -30,  -5,  21,   3,  -1,  -3, -13,  40, -18,  19,  -6,  -9, -25, -22,  -8, -13, -19,   3,  -4,   1,  -1, -16, -39, -21,  10,   8,   1,   5, -46, -16,  26,  19,  23,   7,  -7,   0,  -3,   5,   6,  -9,  -2,  -1,  -2,   1,  -2,   7,  -3,  -2,   2,  -2,   1,   2 },
      {  -1,  -5,  -8, -33, -25, -16,  14,  12,   2, -12,  -3, -17, -12, -15,   5,   7,   1,  -1,  -8, -36, -31,  25,   8,   5,  13, -33,   7, -38, -29,  39,  15,  -3,   0,  39,  26, -23,  -9,   7,  -3,  -4,  24,  11, -10,  -3,   4,  -2,  -8,   0,   6, -12, -18,  13,  14,  -1,  -2,   0,  -1,   0,  -2,   6,   4,  -3,   0,  -1 },
      {  -3,   3,  11, -20,   4, -19,   0, -20,  -2,   1,   4,  15,   2,   8,   5, -14,  -3,  11,  19, -33,  -1,   5,  -4,  -9,   7, -12,  38,  10,  13,  -1,  -6,   5,   9,  39,  18,   8,   3,  -2,  -7,   3, -84,  -2,  11,  -7,  -3,   1,   6,   0, -33,   6,  15,  -2,   0,  -1,   3,   0, -13,   3,   0,  -1,   0,  -1,   0,   0 },
      {   1,   4,  10,  15, -27,  62,  13,   3,   0,   8,   3,  12,  28,  53,   0, -17,   0,  -8,  17,  -2,  17,   2, -15,  -5, -11, -22, -11, -48, -22,  -2,   1,  10,  -8,  -1,  21, -10, -16,  -8,  -1,   4,  -6,  -3,  24,  15,  -5,  -7,  -5,   1, -11,   2,  14,   5,   1,  -3,  -2,   0,  16,   5,   5,  -3,  -5,  -5,  -2,   1 },
      {   7,   9,  14,   7,  36, -23,  10,   8,   6,   8,   3,   7, -29,  24,  19,   7,  11,  19,  22,  40,   0,  36,   6, -11,  12,  14,  -7,  -8,  14,  14,  -9, -12,  16,   8,  45,  10, -26, -20, -13,   2,   5,  14,  30,   6, -11, -19,  -5,   4,  23,   2, -15, -18,  -3,   4,   2,   2,  11,   2, -18,  -5,   5,   1,   1,   1 },
      {   0,  -3,   2,  -3,  14,  -3, -17,  10,  -6, -11,  -3, -21, -18, -42, -25,  11,   5,  -8,  16,   3,  64, -18, -26,  -1,   0, -11,  10, -38,  28,  25,   4,   4,  -1, -15,  11,   0, -34,   7,  18,   0, -14, -13,  -3,  25,  -4,   7,   8,  -3, -11, -14, -15,   1,   3,   9,   5,  -1,   1,  -1,   4,   6,  -2,   1,   2,  -2 },
      {  -1,  -5,  -1,   4, -29,  10, -24, -13,  -2,  -5,  -3,  -5,  17, -17, -20,  16,   1,  -6,   5,   4, -23, -19,  12,   1, -13,  12, -32,  21, -28,   9,  16,  -8,  31, -15,  44,  12, -10,  -4,   1,   2, -41, -34,  18,  -6,   3,   9,   3,   2,  41,  -5, -26,   6,   9,   6,  -2,  -2,  -6,  -6,  -2,  -2,   2,   2,   1,   0 },
      {   5,   9,   9,   3,  -7,  23,  26,  48,   5,  -1,   2,  14,  26, -49,  14,  10,   7,   9,  11, -14,  13,  22,  -8, -10,   5,   4,  12,  34, -13,   6,  -5, -10,  -5,  27, -22,  10,   0, -20,   4,  -2,  -1, -29,  11,  12,  -6,   3,   6,   1,   7,  15,  -7,  -4,   4,  -1,  -1,  -6,  47,  13, -13,  -9,  -5,   2,  -3,   0 },
      {   3,   4,   2,  13,  -5,  -9,  57,  18,   4,   2,   7, -20,  39,   0, -12,   5,  -3,  -2,  -9,  24, -30, -40, -17,   0,   5,   3,   2, -29,  34,   5, -13,   3,  11,  13,  16,  34,  13,   5, -11,  -6,  -9,   9, -31, -18,  -2,   9,   5,   0,  -2, -11, -14, -14,  -3,  10,   3,   2,  10,  14,   7,  -2,  -8,  -3,   2,  -1 },
      {  -1,   2,   2,   3,  -7,  19,  14,   0,  -2,   6,   7, -20,  13, -46,   4,  -7,   1,  -3,   8,   2,  -2,  -7,  14,   0,  -9,  26,   3,   7, -16,  19,   3,   3,  14, -24,  10,  -3,  -2,   0,  -5,   0,   7,  63,  46, -21, -15,  -7,  -9,  -3, -52,  14,   2,  -6,   7,   2,  -2,   0, -17,  -7,  -1,  -1,   3,   5,   3,   1 },
      {  -3,  -7,  -1,   0, -12, -14, -14,  -4,  -3,  -9,   2,   1,   5,  -5, -21, -18,  -3,   1,  -4,  23, -35,  26,  -4,   1,  -2,  -5,   5,   5,  10,   6,   0,   1,  -2,  -5,  12,  30,  -2, -15,   0,  -2,  37, -59,  17,  17, -17,   4,  -1,   4, -61, -28,  19,   2,   8,   9,   7,   2,  -9,   4,  -3,  -6,   2,  -1,   1,   2 },
      {  -2,  -4,  -6,   5, -17,  10, -48,  31,   0,   0,   3,  -4,   5,  12, -37,  11,   2,   2, -10,  14, -10,  17,  32, -13,   8,  -4,  29, -15,  21,  15,   4,  -7,   1,  -9, -23,   6,  32,  -2, -17,  -2, -24,  29,  14,  -3,   4, -11,  -3,   2,   5, -31,  -3,  13,  -9,  -7,   5,   4,  39, -16, -14,   8,  -9,  -1,   4,   0 },
      {   1,   1,  -4,   1, -16,  15, -30, -28,   5,  -3,   9, -20,  27, -21,  28,   2,   4,   8,  -1,  16,  -4,  10,  22,   4,  12, -10,  23, -39,  29,   7, -14,   3,   1,  12, -12,   8,   9,  -9,  -4,   1,  11, -21,   8,   1, -21,   8,   0,   1,  29,  59,   6, -18,  -6,  -8,  -2,  -2, -27,  -9,   3,   5,   5,   3,   0,   0 },
      {  -2,  -2,  -6,  -8,  -9, -24, -23,  35,  -1,  -2,   0,  -6, -16,   2,  -4,   0,  -2,   1,  -7,  -7, -11,  -7,  -2,  -3,  -5, -11,   0, -25, -20, -64,  -2,  19,   9,   0,  -4,  10, -33, -11,  18,   1,   0,  10,  12, -36, -26,  32,  16,  -5,  -1,  14,  12, -24,   3,  24,   0,  -9,  23,  -3, -12,   6,   7,   5,  -4,  -4 },
      {  -2,  -4,  -2,   5,  -1, -13, -25,   3,   1,  -1,  -2,  -6,   5,  -3,  49,  -9,   2,  -4,   8,   3,  14, -14,  52,  -3,  -7,  -3, -22, -12, -22,   2,  -8,  -1,  -5,  -5,  10,  36,  -5, -11, -11,   0, -10,   2, -46,  19,  35, -11,  -3,   3, -38,   8,  -6,  -6,  16,   5,  -3,   2,  27,   0,  -2,   0,  -3,   1,  -6,   0 },
      {   1,  -1,   4,  -4,   8, -18,  -3,  36,   1,  -4,   2,  -6,   6,   5,  13,  -2,   1,  -3,   9,  -6,   2, -11,  14,  -9,  -1,  -1,  -8,   8,  27,  11,   3,   0,  17, -26,  41, -50,  18,  36, -12,  -6,   8, -31,   0, -19,   3,  15,  -8,  -2, -10,  31,  42,  15, -17, -10,  -6,  -4,  25,   1,  -2,   6,  -7,  -6,  -1,   1 },
      {  -1,  -1,  -1,  -4,  -5, -21, -13,  13,   0,   3,   3,   4, -15,   6,   4,   9,  -4,   1,  -7,   2,  -3, -62, -21,  10,   2,  -2,  10,  -2, -25,   3,  12,  -1,   2,  29,   4,  30,  23,  25,   6, -14,  15,   2,  40,  47,   0, -27, -15,  -7,   5,  17,  15,  -4, -19, -11,   1,   1,   5,  -9, -12, -11,  -3,   8,   4,   2 },
      {  -2,   4,  -2,   4,  -6,   5, -22,  22,  -5,   4,  -9,   0,  -7,  17, -23,  -3,  -6,   8, -10,  15, -17,  21, -36, -11,  -8,  14,  -9,  12,  -8,  47,   4,  -5,   3,   8,  -5,  23, -23,   3,  38,  -2,  -9,  15, -31, -12,  15,  -7,   9,  -3,  -8,  49,  23, -10,   8, -20, -12,  -1,  -6,  -3,  15,  15, -11, -13,   5,   2 },
      {  -3,  -3,  -6,  -1, -13, -11,  -6, -12,   2,   2,   6,   3,  16,  16,  22,  -8,  -6,  -6, -15,  -8, -36, -30,   0,  -3,   4,   5,   7,  16,  38,  16, -10,   2,  -8, -15, -20, -35, -52,   5,  25,   9,  -7,   5,   6,  24, -19, -24,   4,   3,  -3,   6, -22,   9,  30,   4,   3,  -2,  27,   3, -23, -11,   6,   4,  -4,   1 },
      {   2,   1,   4,   6,  -5,  -5,  34,   4,  -1,  -2,   1,   1,  -8,  -4,  -9,   3,   4,  -3,   8,   3,  -3, -15,  49,   2,   6,  -3,  13,   2,  -3,  14,  32, -12,  -7,  -8, -12,  17, -37, -20,   4,   3,  -8,   4, -14,  12, -26,  -8,   9,   3,  24, -11,  66,  23, -19,  12, -18,  -9, -11,  -3,  14,  -3,   2,   4,  -7,   2 },
      {   0,   2,  -2,  -2,  -3,  -6,   8, -23,  -4,  -2, -13,  -8, -14, -11, -60, -19,   1,   6,   0,  -2,  -2,   2,  11,   5,   2,   4,  -7,  -8,  -7, -13,  21,   9,   0,  -4,   9, -23,  15, -21, -21,   9,  -1,   2, -24,  19,  -9, -30,  18,  12,  -8,  48,  -9, -15, -15,  -4,  10,  -2,  23,  36, -19, -27,   5,   7,  -1,  -4 },
      {   1,  -1,   2,  -2,  -5,  -1, -12, -13,   0,  -3,   0,  -3,  -3, -42,  16, -31,   0,  -3,  -3,   6, -24,   3, -34, -17,   0,   1,   2,  -4,   1, -23, -34,  22,  12,  -4,  26,  -7,   2, -33,   6,  18,   2,  23,  -8,  23,  14, -11,   3,  -3,  23, -16,  26,  37, -20, -16,   2,   4,  19,  -6,  19,  10, -24,   1,   4,  -1 },
      {  -1,  -2,   0,  -1,  -4,   4, -13,   7,  -1,  -1,  -3,  -2,  -5,   0,  28,  -3,  -1,  -2,  -2,   5,   3, -13, -24, -15,  -5,   0,  -9,  -3, -30,  27, -17, -20,  -5,  -5,  -7, -13,  13,  -2, -12,   5, -11, -15, -26, -29, -56, -35,  26,  21,  -7, -18,   0,  -9, -38,   6,  32,   7,  -4, -17,  -4,   4,  19,  17,  -2,  -3 },
      {   0,  -3,  -2,  -4,   5, -13,  -6, -55,   3,  -2,   4,  -8,  -5,   3,  -2, -10,   2,  -2,  -1,  -7,  14, -19, -10,   6,   4,   1,   4,  -1, -14,  42,  -6,  -4,  -5,   3, -17,   8,  -3, -13,  -5,   7,  14,  -9,  27, -46,  17,  14,  -8,   6,  12, -11,  24, -22,   0,  12,  -4,   9,  55,  25,   8,  -1, -17, -10,  -5,  -1 },
      {  -1,  -2,   2,  -9,  -4,  -2, -10,  21,   1,   1,   4,   4,  16,   1, -10,  42,  -2,  -5,   1, -17, -10, -10, -11,   8,   0,  -4,   5,   0,   6,  -1, -21, -22,  -4, -10,   5, -32,  -5, -45, -11,  13,   3,   0,   0,   8,  44, -20, -14,   6,  -3,  -4,  27, -52,   0,  30,   3,   0, -26,  -8,   9, -15,  -9,  15,  10,  -4 },
      {   0,   0,   1,  -3,  -6,  11,  -1, -11,   0,   0,  -2,   5,  10,  -6,  -4, -19,   1,   2,   2,  -3,  -9,  24,  14,  23,  -3,   4, -11,  13,  -1,  16, -16,  22,   0,   6,   4,  -5, -19,  47,   2, -10,  -3,  16, -11,  33,   0,  32,  17, -15,   6, -20,   9, -53, -46,  14,   8,   0,  12, -25,  -7, -19, -10,   6,   4,   7 },
      {   1,   2,   0,   0,   3,  -4,  18,  15,  -2,  -6,  -5,  -2,  -5,   2, -27, -55,   1,  -4,   0,  -9,   8, -21,   4, -22,   2,   0,  -4,   4,  -2,  11, -17,  12,  -7,   7, -16,   5, -22, -14, -40,  11,   6, -16,   2, -10,  16,   8, -27,   7,  12,  13,  -9,   0,   6,  -7,   1,   2, -14, -61, -34,  12,   4,  -3,  16,   8 },
      {   1,  -2,   1,  -3,   6,  -2,   4,   0,   3,  -3,   3,  -1,  10,   7,  18, -35,   1,  -5,  -2,  -2,   4,   2,  -9, -33,   6,  -1,  10,   3,   6,  11,  61,  10,  12, -12,   9,  -6,  46, -22,  43,  20,   2,   4,  -8,  14,  -3,   6,  -9,  -6,   8,  -7,  -8, -32,   3,  39, -13,  -6,   2, -18,  -5,  10,   7,   1,  -7,   2 },
      {  -1,   0,  -4,  -2,  -2,   0, -22,  29,  -1,   1,   2,  -4,   3,  -2,   0, -42,   0,  -4,  -2,  -9,   2,  -9,  16, -59,   0,   2,  -1,   1,   0,   4, -11,  -6,   0,   6,  -7,   3, -10,  21,  -4,  -2,   8,   5,  13,   1,  10,   2,   4,   9,  20, -15,  -3, -12,  -4,   3,   2,  16, -38,  67,  16, -29,   1,  -3,  -4,   0 },
      {  -1,   1,  -2,   0,   4, -10,  -5,   7,  -1,   2,   2,   2,  -7,  14, -12,   2,  -1,   1,  -5,   2,   4, -10,  -1,  -1,  -4,   5,  -1,  10, -17,  20, -37,  -4,   1,   4,   1, -12,  25,  -8, -35,  10,   3,  14,   3,  41, -31,  32,  31, -12,   5,  15,   1,   9,  39,  40,  -6, -13,   2,  -2,  33,  14,   8, -31, -13,   7 },
      {   2,   1,   2,   1,   4,   8,  17, -17,   0,  -4,  -4,  -3,  -3,   2, -22,  10,   1,  -1,   2,  -3,   8,  -5,  26, -24,  -1, -10,  -3,  -7,  -4,   0, -54, -23,   7,  10,  13,   3,  26,  12,  60,  18,  13,   0,   6,  -5,  17,   0,  37,   9,  -3,  -3,  14,  10,   9,   2, -15,  -3,  -2, -12, -42,  -6,  19,  14,   3,  -1 },
      {  -1,   0,  -1,   2,  -1,  12, -11, -17,  -1,   4,  -4,   1,   1,   2,  14,  24,  -3,  -3,  -6,   0,  -6,   3, -29, -29,   1,   5,  -2,  -4,   4,   8,  21, -23,  -3,   0,  -1,  12,  -8,  -5, -25,   4,   3,  20, -14,   7,  14,  52,  -8, -14,  -5,  12,  16,  37, -14,  23,  32, -13,  -4,   8, -49, -33,  14,   7,   2,  -5 },
      {   1,   2,   2,   4,   7,   9, -15,  14,  -1,  -3,  -2, -11,   5,  -9,   9, -23,   2,   3,   2,   6,   5,  11, -25,  42,  -2,   4,  -2,  -3,   3,   8,   0,  28,   2,  -1,   2,   5,  14,  15, -19, -13, -10,   4, -10, -12,   3, -25,   7,   1,  25, -19,  39,   1,  52,  12, -25, -12,  -8,  12, -37, -23,  27,  18,  -5,  -6 },
      {   0,   1,   0,   2,   1,  -4,   0,  26,   0,   2,   1,   2,  -3,  18,  11, -13,  -1,   2,  -1,   2,   2, -10,  14,  42,   1,   4,   8,   3,   0,  30,  -3,  44,  -3,  -2,  -3,  -1,   5, -33,  17,   6,   3,  -5,  20, -12,  32,   7,  38,  -4,  -1,   7, -22,  29, -36,  14,  29, -24, -22,  18,  -9,   6,  -4,   2,  -1, -11 },
      {  -1,   0,  -1,   0,  -1,  -4,  -3,   3,  -1,   4,  -1,   1,  -3,   8,   1,  36,  -2,   1,   1,  -1,   5, -12,   9, -13,  -2,   4,  -4,   8, -11,  21, -16,  58,  -5,  -2,   3,  -7,   9, -30,   8,  16,  -4,  -1, -17,   6, -35,  24, -30, -21,  -4, -17,   3, -14,  -4, -43, -25,  18,  -8,  22, -24,  -9,  11,  -9,  22,  -2 },
      {   0,  -1,  -1,  -1,  -1,  -4,  -6,   9,   0,  -1,   0,  -1,   1,   1,   7, -34,   1,   0,  -1,  -2,   2, -12,  -6,  39,   1,  -5,   0,  -6,   2,  -1,  -2, -60,   7, -10,   3, -14,   6, -40,  14, -46,  -9,  11,  -3,  10,  -8,  34,   5,   9,  -1,  -9,  -4,  -8, -10, -32, -18,  17,  -7,   7, -25,  -5,   3, -28,  -4,  18 },
      {   0,   2,  -1,   2,   5,  -2,  -5,   7,  -2,  -1,  -2,  -3,  -7,   7,  -6, -18,   1,   5,   0,   3,   6,   8,   7,  20,  -2,  -3,  -2,  -2,  -9,   5, -26, -13,   3,   3,  -2,   0,  11,   2,  27,  -2, -11,   0, -12,  -9, -31,   3, -59,   4,   1,  12, -14,  27,  -2,  40,  -7, -17,  -1,   3,  12, -40, -45,  25,  35,   2 },
      {   0,  -1,   1,   0,   8, -10,   2,   6,   0,  -2,   2,  -7,  -3,   8,   8,   2,   0,  -3,   0,  -1,   3,  -2,   1,  -7,   1,   0,   2,  -7,   2,   5,  18,   7,  -2,  -2,   0,  -1,   8, -13,   0,   7,   4,  -3,   7, -13,   0,  13,  41,  -6,   1,   0,  -3,   1,  28, -39,  11,  12,   4, -42,  33, -82, -21,  18,  -6,  14 },
      {  -1,   0,  -1,  -1,  -2,  -2, -10,  12,   0,   1,   1,   1,   2,  -1,   3,  -5,  -1,   3,  -2,   2,  -2,  -1, -10,  27,  -3,   1,   1,  -1,   5,   0,   4, -17,   2,   6,   0,   9,  -4,  15,  -9,  82,   1,   7,   2,   5,   1,  20, -12,  51,  -8,  -6, -11,  -2, -21, -23, -34, -32,   3,   2,  11, -13,  22,  -2, -20,  -7 },
      {  -1,  -1,  -1,  -4,   2,  -2,   0,  -1,   0,   0,   2,   1,   1,   6, -14,  -6,  -1,   1,  -1,   0,  -3,   1,  -4,  -7,   1,   1,   2,   2,   2,  10, -18,   5,   0,  -3,   0,  -1,  -3, -15,   4, -36,  -2,  -1,   2,   0,  14,  10, -25, -16,  -3,   8, -12,  12, -21,   2, -19,   3,  19,  -6,  41,  -8,  75,  53, -33,  -3 },
      {   0,   0,  -1,   1,  -1,   0,  -6,  -8,   0,   1,  -1,   0,  -1,   2,   5,  17,  -1,  -2,  -3,  -1,  -4,   1, -15, -21,   2,  -1,   1,  -1,  -1,   1,   2,  16,  -4,  -1,  -6,  -3,  -6,  -2, -22, -26,   4,  -5,   3,  -5,   3,  -9,  28,  15, -13,   5, -23,  13, -30,  27, -87,   7,  -9, -10, -18, -10, -38,   0, -21,  -2 },
      {   0,   0,   0,   1,   1,  -1,   5,   0,   0,   1,  -1,   3,  -2,   6,   0,   7,  -1,   0,   1,  -1,   1,   0,   1,  -3,   1,   3,  -1,   6,   0,   7,   5,  21,   1,  -1,   2,  -1,   5,  -2,   5, -25,   0,  11,   1,  20,  -4,  34,  -2,  87,   4,   6,  10,  -1,  21,  -3,  21,  31,  -1,  -2,  -1,  20, -18,  40,  -7, -36 },
      {   0,   0,   2,  -1,   1,   1,  -1,   2,   0,  -1,   0,  -1,   3, -10,  -3,  14,   0,   0,   2,   0,   0,   5,   1,   2,   1,   0,   0,  -1,   2,  -5,  -7,  29,  -1,   1,   2,   4,   1,   2,  20, -12,  -3,   2,  -2,  -7,   7, -16, -22,  42,   3,   7,   5,  13, -12,  28,  18,  27,   3, -18,   4, -37,  31, -77, -20,  23 },
      {   0,   0,   0,   0,   1,  -2,   3,  -3,   0,  -1,  -1,  -1,  -2,  -1,  -9,   0,   1,  -1,   0,   0,   1,  -1,   5, -12,  -1,   0,  -2,  -1,  -5,   3, -16,   0,   1,  -2,   1,  -2,   3,  -4,   8, -16,  -2,  -1,  -5,  -2,  -9,  -6, -27, -14,   2,  -9,   2, -13,   5, -20,  19, -50, -11,  -8, -14,  -9, -32, -17, -86, -43 },
      {   0,   1,   0,   2,  -1,   2,   1,  13,  -1,   0,  -1,   0,  -1,   0,  -6,  -3,   0,   2,   1,   2,   0,   3,   1,  25,   0,   1,   0,   0,  -1,   1,  -2,  -8,  -1,   3,   0,   3,   0,   6,   2,  44,  -3,  -2,  -3,  -3,  -4,   2, -10, -31,   1,  10,   1,  12,   0,  16,   6,  83,  -9,   5, -17,  10, -19,  22, -58,  16 },
      {   0,   0,   0,   0,   0,   0,  -2,  -4,  -1,   1,  -1,   0,  -1,   0,  -2,  12,   0,  -1,   0,  -1,   0,  -1,  -2,  -9,   0,   1,  -1,   0,  -2,   2,  -3,  15,  -1,   0,   1,  -1,   2,  -3,  -1,  -7,   0,   2,  -3,  -1,   0,   3,  -7,  22,   0,  -4,   3,  -5,   5,  -8,  12, -36,  -5,  14,  -9,  18, -15,  28, -26, 107 },
    },
  },
  { // 2
    {
      { 119, -19, -21,  -5,  -4,  -1,  -2,   0, -29,  -1,  -6,   2,   1,   2,   0,   0, -20,   2,   0,   2,   2,   1,   1,   0,  -8,   1,   3,   2,   1,   0,   0,   0,  -3,   1,   2,   0,   0,   0,   0,   0,  -1,   1,   1,   0,   0,   0,   0,   0,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  21,  54, -27, -13,  -6,  -1,  -2,  -1,  87,  10, -41, -21,  -5,   2,   0,   0,  -4, -25, -23,  -4,   9,   6,   2,   0, -11, -23,   3,  11,   6,   3,   1,   0,  -1,  -6,   6,   6,   1,   0,  -2,   1,   2,  -1,   2,   1,  -1,  -3,  -1,   1,  -3,   1,   3,  -2,  -1,  -1,   0,   0,  -1,  -1,   1,   0,   0,  -1,   0,   1 },
      {   3,  47,  27, -13,   4,  -2,  -1,  -3, -35,  92,  28, -18,   5,  -6,  -4,  -4,  -6,  -4,  -7, -27,  -3,   0,   2,   2, -27,  -9,  -1,  -5,   9,  10,   4,   4,  -6,  -1,  -3,  12,   5,   3,  -2,  -1,  -5,  -2,   1,   8,   2,  -2,  -1,  -1,   3,  -2,  -7,   2,   1,  -3,   1,   1,   2,  -2,   0,   1,  -1,   1,   0,   1 },
      {  -4,  67,  25, -20, -38, -13,  -2,   2, -34, -31, -16,  -9, -30,  -2,   7,   6, -44,  24,  -8,  35,   0,   2,   0,  -2,  14,  15, -20,   2,  -1, -14,  -4,  -5,  -7,   3,   4, -18,   1,  -6,   9,  -1,   0,  -3,   1,  -7,  -1,  11,   1,   0,  -3,  -7,   7,   3,   2,   8,  -2,  -1,  -3,   3,   0,   0,   3,   2,   1,  -3 },
      { -15,  24, -14,  25,  19,   8,  -2,  -2, -27, -24,   4,   6,  17,  10,  -2,  -2, -64, -48,  17, -15,  19,   5,   3,   1,  -8, -11,  46,   8,  -3,   6,   1,   1,  33,  10,   6,   4,  -7,   0, -10,   0,  25,   8, -11,  -5,  -2, -16,  -2,   2,   6,   9,  -4, -12,  -4,  -6,   4,   1,   6,  -1,  -1,  -2,  -1,  -3,   1,   3 },
      {  23,  -5,  71,  -2, -14,  -6,   0,  -2,  45, -34,  67,  -6, -20,  -4,  -1,  -1, -13, -32,  16, -14, -10,   2,   3,   2, -11,  -5,  -4,  -7,  -3,   7,   3,   1,  -6,   5,  -5,  -2,   4,   5,   0,  -1,  -5,   4,  -5,   0,   6,   2,   0,  -1,   0,   1,  -4,  -1,   3,   0,  -1,   0,   0,   2,  -2,   0,   1,   1,   1,   0 },
      {  -3, -25, -11,   2, -19,   0,  -1,   3,  15,  14,  15, -34, -31,  10,   5,   3, -27,  57, -15, -44,   7,  10,   2,  -6,  24, -19,  -9,  -7,  -2,   0,   1,  -4,  46,   3, -15, -12,   5,   1,  -6,   1,  33,  -4, -24, -16,  -5, -10,  -3,   4,  -4,   6,   4, -13,  -2,   2,   5,   0,   1,  -2,  -3,   1,   6,  -5,   0,   1 },
      {  -9, -40,  31, -13,  -8,   3,   2,  -1, -28,  -8, -15, -24,  -8,  10,   4,   2,   0, -37, -66,   8,  18,  13,   2,   2,   0, -33, -23,  39,  23,   2,  -1,  -2,   2,  -1,  20,  26,   8,  -9,  -2,  -1,   8,   7,  13,   7,  -6,  -6,  -1,   0,   3,   4,  10,  -1,  -7,   2,   1,  -1,   0,   3,   3,  -2,  -3,   1,  -1,  -2 },
      {  21,  47,  16,  32,   4,   4,  -2,  -2, -28, -27, -11,   6,  -2,   7,   2,   2,  65, -17, -10, -34, -11,  -1,   1,  -1,  44, -23, -13, -15, -12,   7,   5,   1,  24, -21, -17,   1,   2,  12,  -2,   2,  11,   3,  -3,   4,   6,  -5,  -1,   1,   0,   6,   0,  -4,  -1,  -2,   2,  -1,   2,  -2,  -1,  -1,   0,   0,  -1,   2 },
      {  -2,   6, -12,  62, -14,  -5,  -8,   1,   5,   8,  20,  62, -19, -10,  -8,  -1, -16,  23, -21,  13, -14, -13,  -1,  -3, -29, -52, -14,   6,  -5,  -5,   9,   2, -14, -17,  19,   3,   4,  -1,   6,   4,   4,   3,  11,   1,  -5,   2,   2,   2,   0,   7,   9,  -9,  -3,   6,   3,  -5,  -2,   5,  -1,  -7,   1,   1,   1,  -2 },
      {  12, -10,  35,  21,  -9, -12,  -4,  -2,  13,  51, -11,  18, -13,  -9,  -4,  -2,  12, -24,  -2,  46,   8,  -6,  -8,  -3,  40,   9,  30,  21, -11, -18,  -9,  -5,  33,   8,   2, -21, -26, -14,  -1,   1,   6,  -9, -18, -24, -15,   5,   6,   4,  -6,  -4,  -9,  -7,   1,  11,   5,   0,  -2,  -3,  -5,  -1,   4,   1,   2,   0 },
      {  -4, -20,  12,  43,   2,   1,  -3,   3,   4,  13, -45,  10, -17,  -5,  -2,  -1, -36, -28, -38, -28, -23,   6,  10,   2,   2,  44, -10, -34,   5,  13,  10,   3, -12,  14, -22, -19,  12,  18,   4,  -2, -17,  -1,   1,   0,  17,  13,  -5,  -3,  -2,  -1,   1,  13,  12,  -4,  -4,  -2,   0,  -6,   1,   7,   2,  -1,  -4,   0 },
      {  11,  15,  24,  17,  46,   3, -14, -11,  19, -11,  15,   6,  52, -13, -20, -13, -22,  35, -38,   2,  35, -14,  -9,   0,  23,  23, -22,   3,  13,  -5,  -4,   2,   0,   2, -14,  13, -14, -10,  -3,   0,  -3, -18,  -7,  19, -10,  -7,   3,   4,  -4,  -1,   5,   4, -12,  -1,   6,   2,   1,   2,   5,  -3, -10,   2,   2,   2 },
      {  -5,  -8, -28, -32,  -1,  22,  10,   0,   3,  22,  23,  20,   2,  15,  -5,  -5, -33, -38,   9,  21,  -4,  -7,  -9,  -1,  49, -20, -42, -13, -23,  -8,   2,   6,   5, -24, -31,   0,  -3,   8,   8,   2,  -5,  -8,   7,  15,  24,  -1,  -1,  -1,  -2,  11,  22,  12,  -5,  -2,  -4,   1,   3,  -1,   5,  -3,  -3,   6,  -5,  -1 },
      {  -3,  14, -57,  -2, -12, -31,  -4,   4,  -4,  -6,  51,  15,  -9, -28,   6,   5,  10, -26, -24, -18,  -9,   6,  12,   7,  22,  31, -17,  26,  25,  14,   2,  -2,  12,  25,   6,  10,  -2, -10,   0,  -6, -13,  -8,  -4, -12, -21,  10,   6,  -3,  -4, -17,  -8,   6,   6,  11,   2,   3,  -4,  -7,   1,   5,   5,   3,  -1,   0 },
      {   0,   0,  -5,  20,   0,  15,  -3,   4,   1,   4,  15, -24, -30,   0,  -9,   0,  -9,  11,   3, -19,   4,   7,  10,   3,  51,  16,  52,  20,   5,  -2,  -8,  -4, -44, -41,  27,  14,  -4,  -3,  -4,   1, -26, -11,  23,   9,  11,  -1,   1,   3,   6,  16,  21,  -3,  -3,  -3,   2,  -4,   0,   2,   1,  -6,  -2,   5,   2,  -3 },
      {   1,   1,   4,  24, -16, -11,  -6,  -8,   4,   8, -10,  19, -22,  11,   2,  -6,   4,   0,  44,  19,  30,  37,  -6,  -1,  -3,  13, -15,   3,  55,  22, -19,  -3,   9, -14, -31,  25,  36, -13, -16,  -6,  -1, -10,  -8,  15,   4, -22,  -8,   0,  -2,   2,   4,   7, -15,  -7,  -7,   4,  -4,  -1,   9,  -4,  -4,  -6,  -3,   2 },
      {  -5,  -4, -15,  13,  -8, -10,   0,   0, -12, -12,  14, -28,  10, -12,  10,  -9,   6, -10, -20,  16,  23,  13,  -6, -15, -23, -25,  21,   4,   6,  10,  -9,  -1, -23, -37, -45, -33, -11,  17,   9,   6,  -9, -34, -45,  -2,   5,  27,  14,   2,   4,   1,  -8,  14,  14,  12,   2,  -3,   2,   2,   8,   9,   0,   0,  -3,   0 },
      {  -1, -10,  17, -13,  20, -29,  -3,  -5,   1,  -2, -33,   5,  -4, -32,   1,   2, -21,   6,  32, -16, -39, -13,  18,   4,   1,  -6, -10,  37,  -7,  15,  13,  -6,   9, -35, -12,  44, -18,  10,   6,  -9,  19, -26, -12,   5,  -7,  22,   4,  -1,  14, -16,   7,   0,   5,  17,  -2,   0,   1,  -3,  10,  -1,  -3,   3,  -4,  -1 },
      {  -2, -10,   5, -23, -15, -40,  -7,  -3,   1,   1,  -6,  12,  23, -39,   3,  -7,  -8,   2,  -2,  11,  16,   8,  13,  -9,  35, -30,  16, -52,  12,  36,   9,  -8,  -1,  -3,  39, -11,  11,  27, -14,  -2,   3,  -8,  10,  16,  -7,   3, -17,  -2,   5,   0, -11,  -1,  -4, -15,  -3,   3,   7,   2,  -6,   5,   1,  -7,   4,   3 },
      {  -4,  -9,  -1, -10,  15, -32, -25,   3,  -7,  -2, -13,  -5, -11, -21, -26,  11,   0, -23,  20, -34,  25,  -4, -13,  11, -10, -11, -11, -35,  17, -43,  -6,  13,  -3, -22,  -2, -22,   0, -32,  10,   5,  -4,   0,  23, -16, -23,   1,  21,   6, -14,   4,  22,  -2, -10,  23,  19,  -2,  -4,   6,   0,   0,   7,   2,   0,  -5 },
      {   3,   2,   7,  12,  29,   7,   5,  10,   4,   8,  -1,   4,  18,  27,  12,   6, -11,  -1,  23, -13,  12,  22,   5, -12,  14,  -7, -49,  16,  12,  12, -19, -11, -20, -23,  43, -28,  -9,   4, -10,   5,  -3,  13,  -7, -52,   0,  11,   3,  -3,  13, -15, -16,   5,  20,   1, -10,   3,  -1,  -6,  -2,  17,   8,  -8,  -1,   4 },
      {   2,   1,  -9,  26,  25, -40,   9,   9,   3,  10,   9, -34,   0, -31,  30,  13,   0,  -4,   3,  26,   5,  10,  20,   2,   3,  13, -19,   8, -30,  -4,  -1,  -3, -19,  -5, -23,  -7,   5,   1,   6,  -2,  22,  54,  10,  10,  -9, -17,  -6,  -2,  12,  32,   5, -16, -14, -10, -13,   2,   0,   4,  -2,   1,   3,  -2,  -6,   2 },
      {   0,   4,  10, -49,   5, -24,  -3,   0,   5,   0,   4,  63,   2,  18,  12,   3,   7,   8, -35, -20,   4,  24,  -2,  -5, -14,  28,  28,   6,  -5,   1, -15,  -5,   3, -28, -11,  -8,  -3,   3,   3,   4,  15,  16,  -5, -17,  11,  -2,   2,   0,  22,  29,  24,  -6,   4,  -6, -10,  -4,  -1,   5,   9,  -9,  -1, -13,  -6,  -1 },
      {   2,   2,   4,   2,  19, -16, -24,  -7,   3,   1,   6,   0,  12,   9, -16,  10,   0,  17,   6,   6,   6,  20,  13,  27,  14, -31,  23,  27,  10, -16,   2,  16,   9,  29, -24, -10,  21,  21,   9,  -9,   6,  16,  33,  -1,  25,  31, -15, -12, -11, -12,  16,  32,  30,   5, -17, -10, -18, -15,  -5,  10,  12,  -5,  -6,  -7 },
      {  -1,   3,   4, -15,  26,   1,  13, -13,  -8,  -5, -15,  18, -36, -18,   1, -14,   3,   8,  11, -17,  15,   5,   8, -13,   3, -33,  -5,  21,   5,  10,   8,  -5, -19,  48, -22, -19, -40,   3, -11, -12, -44,  -2,  -6, -15,   3, -13, -21,   0,  -5,  23,   6,  -1, -24, -10,  -1,   4,  18,   5,   0,  -7,  -3,   4,  -7,   1 },
      {  -2,  -2,  12,  14, -35, -20,  18,  12,   3,   2,  -4,   4,  44,  23,  46,   8, -10,   5,   6, -18,  -8,  23,  13,  -5,  -1,  -8,   5,   3, -17, -19,   6,   2,  11,  -6,  -7,  14, -20,  -9,  -4,  -5, -40, -16,  21,   4, -32,  -9,   6,   7, -39,  12,   7,  23,  -5,  -8,  -5,   1,  -6,  -1,   3,  11,   4, -10, -11,   4 },
      {  -3,  -3,   2,   2, -63,  14,  -7,   5,   2,   4,  -9,   8,  26,  -8, -21,  -8,   0,  -1,  16, -22,  37, -24,  -4,  -8,   8,  13,  -9,  20,  19,  -5,  25,   0, -19,  -5, -23,  13, -22,  24,   7,  -3,  13,  41,   7, -10,  -9,  19,  -9,  -9,   7,  19, -18,  -8,   9,   2,   6,  -9,  11,   0, -14,  -6,   3,   7,  -2,   0 },
      {   2,   0,   0,  10,   1,  -2,  29,   8,   2,   2,   4,   2, -19, -13,  15,   9,  10,  -3,   8, -13,  48, -15,   9,  -7, -26,  22, -14,  16,  -3, -13,   4,  -8,  42, -11,  28, -28, -21,  21,   4,   4,   6, -22,  27,  41,  30,  18,  -7,  -2,   8, -13,  15,  18,  -3, -11,  -7,  -1,   4,   5,  -4,  -6,  -3,  -3,   1,   5 },
      {   0,  -1,   6, -15,  17,   5,  31,  18,   0,  -4,  -5,  13,  -5, -28,  23,   2, -13,   5,  -6,  -8,  -7, -46, -18,   1,  11,  -8,  17,  26,  28, -10,   1,  10,  11, -30,  -9, -16,  47,   1,   4,  12, -20,  24,  -8,  -4,   4, -20,   5,  -6, -31,  -8, -35,  10,  -3, -14,  -4,  -1,  -1,  -2, -12,   7,   2,  -7,   1,   1 },
      {   4,   2,   2,   4,   3,   3,  26,  18,   4,   5,  -3,  12,  -1,   5,  26,  12,   0,  -4,   6, -17,  -4,  -7,  -9,  -5,  21, -15,   8, -15,  28, -39,  -8,  -3, -35,  36,   8,  10,   1, -23,  15,  -8,  38, -12, -30,  31,  -9,  14,   3,  -4,  30,   7,  -2,  25,   7,   9, -25,  -4,  15,  19,  -3,   0,   4,  -9, -15,  -5 },
      {  -4,  -4,  -2,   1, -14, -41, -12, -22,  -3,   3, -10,   5,   8,  18, -10,  -4,   1,  -3,   8, -17,  26, -12,  16,   9,   7,  -1, -12,  18, -39, -17,  12,   0, -12,  13,  24,   5,  33,  -6,   1,   4, -29,   2, -56,  21,  38, -15,  17,   5,   1,   3,  -7,  -8,  11,   3,   8,  -6, -12,   7,  -4,   0,  -3,  -1,   4,  -8 },
      {   0,  -1,   3,  -2,   0,  -4,  17, -11,   1,   6,   3,  12,   8,  23,  20,  12,   3,  -7,   0,  -9,  13,  -2,   3,  14,   6,   7,   7,  13,  -8,   7,  21,   3, -42,   1,  -9, -27,  21,  19,   1, -15,  35, -21,   2,  -4, -16, -10, -21,  -2, -31, -53,  20, -29, -32,  15,  13,   7,   6,  -1,   2, -16,   2,  18,   2,  -5 },
      {   0,  -2,   2, -12,   4,  20,   2,   8,   2,  -2,  -3,  22, -19,   1,  -4,  15,  -4,   8, -10,   0,  23,  35,  40,  25,   2,  -7,   5,  -8, -21,   5,  11,   8,  -4, -16, -22,   1, -16, -34,  -1,   7,  -4,  12,  11,  30,  -9,  -8,  17,  -2,  16, -37, -54,   1,  11,   3,  14,   3,  21,  -4, -19,  10,  10,   7,  -2,  -1 },
      {  -2,  -4,  -5,   2, -17, -22, -29,  -7,  -1,  -4,   3,  -6,   5,   9, -13, -10,   0,  -7,  -6,  -5, -21, -18,   1,   1,  -7,   2,  -4,  17,  -5, -36, -14,   4, -11, -11, -12, -16,   2,   0, -22,  -6,  11, -34,  18, -13,   3, -34, -33,  10,  39, -11, -20,  20, -12, -33, -13,  17,  29, -11, -25,  14,  -5, -10,   9,   6 },
      {  -3,  -1,  -4,  -4,   9,  -6, -12,  -2,  -4,  -2,   2,   6, -19, -10,   6,  12,   1,  -4,  -7,  -1,   9,   9, -11,   7,  -5,   5,  -4, -15,  -1, -24, -18,  -1, -19,  -7,   8,  36, -38,  17,  -2,   6,  28,  -5, -13,  -1,  27, -13, -37,   7, -61,   4, -29,  -1,  24, -15,   0,  13, -23,   8,   6,  13,  18,  11,   7,   6 },
      {  -2,  -3,  -4,   1,   2, -22,  -4, -27,   1,   0,  -3,   2, -14,  15,  14, -32,  -8,   1,   0, -18,  -2,   2, -39, -34,   6,  -4,   6,   4, -20,   4, -49,  -7,  -7,  11,  -3,   8,  -3,  18,  -4,  22,   1,  31,  21,  29, -11,  19,  16,   9,   1, -24, -20,  17, -19,  24,   7,   6,  -2, -19,  -8,  -5,  -7,  17,   2,   9 },
      {  -2,  -3,  -3,  -2, -11,   4,  -8, -37,  -2,   1,   0,  -4,  -1,   8, -14, -11,   1,  -3,   2,  -6, -18, -19,  11,  12,  -1,   0,   1,   7,  10,  30,   8,  -7,  -7, -18,   8, -40, -30, -39, -10,   1,  21,  33, -18,  30, -18, -12,  -8,   1, -23,  -7,  20,  43,  10, -14, -17,   5, -25,  10,   6,  -3,  -9, -10,  -4,   1 },
      {   0,  -1,  -3,   3,   3,  -4,  23, -44,   1,  -1,   3,  -5,  10,   7,  32, -13,  -3,  -2,  -7,   2,  -2, -25,  31,   3,  -1,  -7,   1, -21,  23, -16, -11, -26,   6,  -8, -18,   4,  -1, -29, -23, -11, -10,  -2,  20, -13,  21,  21, -14,  13,  11,   1,  -9, -36,  27,  13,  -4,   3, -22,  -1, -24, -36,   2,  13,   4,  -7 },
      {   0,   1,  -4,  -6,  16,   7, -11, -26,   0,   1,   3,   0,  -5,  13,   3, -36,   1,  -2,   5,   5, -24,  39,  18, -13,   1,   9,  -7,   6,  28, -36,  35,  10,   7, -11,  15, -15,  -1,  20,  35,   4,   7,  -8,  -8,  16, -14,  14,   7, -11,  -6,  20, -26, -14, -31, -16,   6, -16, -24,   1, -27,  -2,  -6,   1,   2,  -1 },
      {   0,  -1,   5,  -3, -19,  23, -21,  16,   2,   3,  -5,   7,  27, -38,  18,  18,   0,   0,   7, -10, -21,  17,  20,  -2,   3,  -9,  -4,  14,  12,  -5, -25, -22, -11,  18,  -3, -21,  -3,  -8,  -9,  11,  24,  -3,   4,   2,  43, -10,  35,  10,  -6,   4,   3,   7, -20, -14,  38,   0, -15, -18,   6, -14, -24,  15,  12,  -9 },
      {   2,   2,   1,   4,  -3,   5,  11, -37,   3,   1,   2,  -1,   0, -12,  20, -44,   5,   3,   3,   6,   3,   1,  27,   2,   2,   8,   2,  -3,   5,  -9,  11,  23,   3,   2,  10,   7,   9,   6,   7,  26,  19,   4,  16, -28,  15, -20,  12,  21,  -5,   2,   2,  29,  14,  18,  15,  -4,  49,  26,  41,  15,   5,  11, -12, -11 },
      {   1,   2,   2, -10,   7,  34, -23,   6,   0,   0,   1,   9,   0, -14,   1,  -4,   3,  -4,   0,  -2,  -1,   2,  21, -19,  -9,   9,  -4,  10, -18, -16, -11, -25,   6,  -7,   6, -15,  32,  11, -46, -23,  -1,  -4,  -7,  23, -29,  -2, -30,  -8,  -7,  30,  -3,  15,   7,  49,   9, -10,  -3,   3,   2,  17,  24,  16,  -7, -13 },
      {   1,  -1,   0,   3, -17,  27,  15, -17,  -2,  -1,  -1,   3,   9, -43, -15, -19,   0,  -4,  -2, -13,   6,  24,  10, -12,  -1,  -6,   4,  -6, -18, -37, -22,  -8,   5,  -5,  -1,  10,  12,  -7,  12,  -5, -13,  19, -16,  -9,  -9,   9, -22, -14,  12, -49,  18, -18, -12, -26, -25,   1, -17,  -6,  38,   9,   0, -14,  -2,   3 },
      {   1,   1,   2,   1,   3,  -9,  33,   8,   1,  -2,   1,   3,  -8,  12,  -2,   3,   2,   5,  -5,  10,   6,  -8,  17,  15,  -3,  -2,   7, -19,  14, -17,  13,  20,   2,  -8,   5,  15,  -8,  17, -21, -11,  -3,  23, -28, -13,  -7,   4,  -4,  11,  29,   2,  20,  37, -22,  -3,  29,  18, -27, -41,   2,  22,  -1,  45,  25,   6 },
      {   1,   2,   3,   1,  -4, -16,  55,  19,  -2,  -5,  -4,   0,  15, -16, -43, -35,   1,   2,   1,  -6, -14,  32, -21, -21,  -1,   3,  -4,  15,  -6,   6,  14,  12,  -8,   3,   2, -14,   6, -31, -13,   6,  20, -15,  11,  11,  18, -11,  -1,  15,  -8,  11,   5, -13,  23,   6,  -5,   2,  -7, -26, -14,   2,  33,  18,  -2,   8 },
      {   0,  -1,  -2,   5, -10,  -4,  13, -10,   2,   1,  -2,  -5,  17,  -9,  -1,  10,  -4,   5,  -2,  -3, -26,  30, -27,  51,   0,   1,  -6,   7,   9, -10, -21,  33,  13, -12,  14, -15,  -9,   1, -26, -10, -15,  18, -19,  15,  14,  27, -25, -10,  -4,   8,  -5, -12, -13,  18,   8,  -2,  44,  23,   1, -17, -16,   5,  11,  13 },
      {   0,   2,  -4,  -1,  11, -20,  13,  38,  -2,   0,   5,  -2,  -8,  24, -32, -28,   3,  -2,  -4,   7,  -4,  -6,  30, -11,  -1,   0,   3,  -4,   0, -15,   4, -38,  -4,  -2,  -9,   8,   9,  -5, -27, -22,  13,  15,   1,  -4,   7,  38,  14,   7, -38,  -7,  -8,   4, -12, -14,  13,   1,  28,  16,  14,   2, -30, -28,  10,  12 },
      {   0,  -1,  -1,   3,  -1,  -9, -14,  18,   0,  -3,   3,  -8,   1,  -1,   3, -13,   2,   2,  -7,   2,   6,  -1,   2,  -6,  -5,   1,   4, -17,  -6,   0,  11,   4,   4, -13,  14,   8, -12, -18,   2, -20,   3,   4,  -6, -21,  31, -17,  -7, -37, -11,  -8, -19,  24, -32,  30, -37, -33,   7, -39,   3, -35, -16,  18, -37,  -7 },
      {   0,   1,   1,  -2,  -7,  15,  27, -15,  -2,  -1,  -4,   6,  -4, -13, -16,  -3,   1,  -1,   0,  -8,  19,  -5,  20,  42,  -4,   3,  -7,   1, -30,  19, -33,  13, -10,   6,   1,   9,  19,   3,  -4,  19,  19, -27,   3, -27, -19,  27,  12,   8, -13,  32, -17,  22, -31,  -6, -30,  -9, -19,   1, -25,  -8,  -6, -11,   6,  -6 },
      {   0,  -1,   0,   2,  -5, -12,   3,  32,   2,   1,  -2,  -3,  13,  14, -24,  -2,  -2,   2,  -2,  -5, -11,  -2,  16,  -4,   2,  -4,   4,   4,   8,   9, -17,  -2,  13,   1,  -7, -10,  -5,   9,  20,  -2, -10,   3,   9,  -2,   8, -14,  -7,  15,  16, -11, -23,   4, -25,  10, -21,   5, -32,  73,  19,  -9,  16,  47,   7, -12 },
      {   0,   1,  -2,   1,   1,  -5,  14,  10,  -1,  -1,   0,  -2,  -9,  -5,   5, -11,  -2,  -1,  -1,   1,   3,   6, -23,   0,   5,   2,  -4,   1,  -9,  -5,  15,  34,  -9,   2,  -5,   0, -10,  12, -46,  -7,   6,  -4,  14,  -5,  -4, -10,  -2,  -5,  23, -13, -19,   6,  12,   1,  33, -20, -51,  24,  17, -15, -29, -42, -21, -34 },
      {   0,   0,  -3,   5,  -1,  -3,  31, -42,   1,  -1,   3,  -3,   3,  -6, -24,  27,  -2,   2,  -5,   3,  -7,  16, -33,  18,  -2,   1,   2,  -5,   3,  -9,   9, -62,   3,  -7,   0,   4,  -6,  25,  -1, -36,   0,   5,   6,  -8,   5, -24,  36,  -3,   2,  -3,  -8,  17,  -7,  -4,   0, -12,   9,   9, -12,  14,  13,   2, -13, -29 },
      {   0,  -1,   0,   0,  -4,  -9,   6, -22,   2,   0,  -1,   3,  -3,  13,   3,   9,  -2,   2,   1,  -8,   4, -16, -10,   7,   4,  -3,   1,   6,  -3,   6,   3,   1,  -4,   4,  -2, -18,   4, -23,   1, -19,  11, -18,  26,   7,   4,  13,   8, -42,   7,  30, -37, -19,  -4,  -8,  -1,  -6, -12, -14,  68,  41,  -8,  18,  16,  20 },
      {   0,   0,  -1,   2,   5,   8,  -7, -21,   1,   1,   2,   0,   3, -16,  12,  25,   0,   0,  -2,  -3,   7,  18, -25, -37,   2,   3,   6,   2, -23,   8,  42,  -3,   0,  -4,  -3,  14,  22, -38,   0,  -6,   5,   1,   5, -26,  20,  25, -20,   2,   4,  -3,  -2,  35, -16,   3,  12,   9,  -9,  38, -24,   5, -12,   7,  19,  34 },
      {  -1,   0,   0,  -4,  -3,  13, -17,   1,  -1,   0,   1,   1,  -1,  -6,   5,   0,   0,   0,  -1,   2,   0,  -1,  -7,  24,   2,   0,   1,   1,   9, -13,  34,  -8,  -5,   3,   0,  -5,  -5,  17, -30,   9,  -2,   4,   2,   4,  14,   1,  26,  46,   9,  -3,  -9, -13, -25,  30, -38,  25, -19,  -5,  20,   3,  20, -25, -25,  54 },
      {   0,   0,  -1,  -1,  -2,  -2,  -8,  -5,   0,   1,   2,   2,   1,   4,   4,  18,   1,  -1,  -1,   2,  -4,  -9,  -2, -19,  -1,   3,   2,  -1,  10,   6,   7,  17,   1,  -1,  -4,   4, -11, -15, -34,  25,   1,   7,  -4,   6,  16,  25,  -1,  24,  -8, -11,  12, -28, -30, -10, -26, -40,   8,   3, -13,  63,  -1,  13, -13, -48 },
      {   0,   0,   1,   0,  -6,  10, -15,  17,   0,   0,  -3,   4,  -2, -10,  31, -57,   2,   1,   0,  -5,   9,  -5, -12,  25,  -1,   0,   0,  -2,  -7,  14, -13,  11,   1,  -1,   3,   2,  -9,  -8,  10, -61,   7,   3,   4,  -3,  19,  -5,  23, -13,  -1,  -6,  19, -11,  -4,   8,  -3,  16, -11,  18, -35,  43,   6,  -6,  14,  -3 },
      {   0,   0,   0,  -1,   3,   5, -20,  -4,   0,  -1,  -1,   0,  -6,  -2,   6,   2,   0,   1,  -1,   0,   6,   0, -10, -10,  -2,   1,  -3,   4, -10,  -3,   1,  27,   3,  -4,   8,  -3,  16,   0, -21, -35,   3,   0,  -4,  -1, -13,  30,  31,  13,  -6,   3,  -4,   0,  13, -59,  -6,  30,  11,   5,   8, -31,  16,  44, -49, -10 },
      {   0,   1,   0,  -2,   3,   5,  -4,  13,   0,   0,  -1,   4,  -6,  -8,   8, -10,   0,   0,   4,  -6,   9,  17, -21,  20,   0,  -2,  -1,   2,  -6,  -7,  21, -20,   4,   0,   2,  -8,  10,   4,  10,  -1,  -2,  -1,   4,   0, -20,  -7,  -9,  41,  -1,   4,  -4,  -6,  36,   6, -36,   4,   0,  -9,   4,   7, -81,  35,  15, -16 },
      {   0,   1,   1,   2,  -5,   2,  16, -13,   0,   0,   1,   0,   6,   2, -22,  23,   1,   2,  -2,   0,  -3,  -3,  17, -22,   0,   2,   4,   1,   5,   7, -18,  19,   4,  -4,   8,  -4,   2,  -3,  35, -25,   1,   4,   3,   4,  13,   3, -18,  21,  -5,  13, -15,  -8, -11,  29,  16,  58,  -4, -13,  -5,  24, -33, -23, -50, -18 },
      {  -1,   0,   0,  -3,  -3,   2,  11, -18,  -1,   0,   0,   2,  -3,  -5, -16,  26,   0,   0,  -2,   4,  -3,  -1,  28, -36,   1,   1,   1,  -3,   8, -10, -18,  38,  -4,   4,  -4,   1,  -5,  24, -22, -17,   6,  -2,   1,   0,  -6, -29,  39, -22, -11,  -9,  14, -10,  24,  12, -26, -17,  -1,  19, -22,   7, -34,  13,   7,  46 },
      {   0,   1,   0,   0,  -1,  -5,   4,   9,   0,  -1,   0,   0,   1,   2,  -2, -14,   0,   1,   1,   1,  -1,  -5,  -2,  29,  -2,  -2,   1,  -1,   0,   2, -11, -28,   1,   3,  -1,   1,   5,  -2,   3,  29,   0,  -6,  -1,  -5,  -4,  -1, -17, -18,   8,  -4,  13,  -6,   6, -13,  33, -21, -11,  17, -20,  27, -20,  28, -75,  47 },
      {   0,   0,   0,   3,  -1,  -4,   0,   3,  -1,   0,   0,  -1,   1,   0,  -2,  -6,   1,   1,  -1,   0,  -1,   6,  -1,   2,  -1,   0,   3,  -1,   2, -11,  18, -12,   1,   2,  -4,   5,  -2,   3, -28,  50,   0,  -3,   0,  -4,   3,   2,  15, -69,  -4,   3,  10,  -4,  -1,  10, -20,  73,   0,  13,  -9,   3, -10,   9,   0, -30 },
    },
    {
      {  78, -67,  14,  -2,   2,  -1,   1,   0, -60,  40,   6,  -7,   2,  -1,   1,   0,   8,   7, -17,   5,   0,   0,   0,   0,  -2,  -4,   4,   2,  -2,   1,   0,   0,   1,   2,  -1,  -1,   0,   0,   0,   0,  -1,  -1,   0,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  57,   3, -38,   5,  -3,   2,  -2,   1,  16, -67,  47,   2,  -1,   0,   1,   0, -36,  43,   7, -25,   8,  -3,   1,  -1,   1,   5, -23,  11,   3,  -2,   1,   0,  -4,   1,   5,   3,  -5,   1,   0,   0,   1,   1,  -3,  -1,   1,   0,   0,   0,  -2,   1,   1,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
      { -55,  -4,   1,  16,   3,   2,   1,   1, -23,  16,  51, -31,   1,  -4,   2,  -2,  -2,  47, -60,  -4,  15,  -4,   2,  -1,  15, -24,  -5,  34, -13,   1,  -1,   1,   4,  -1,  10, -11,  -7,   5,  -1,   1,   3,  -5,  -3,   2,   4,   0,  -1,   1,   1,   1,   2,  -1,  -1,  -1,   0,   0,   1,  -1,  -1,   1,   0,   0,   0,   0 },
      {  -9,  49, -40,  -2,   0,   1,   0,   0, -81,   4,  18,  33,  -8,   3,  -1,   1,  42,  -4,  24, -31,  -8,   4,  -2,   1,   0,  -5, -15,  -6,  19,  -3,   1,   0,   5,   0,   3,  10,  -5,  -4,   1,   0,   1,  -2,  -1,  -4,  -1,   2,   0,   0,   2,  -1,   1,   1,   0,   0,   0,   0,   0,  -1,  -1,  -1,   0,   0,   0,   0 },
      {  35,  48, -52,  16, -10,   0,  -3,   0,  -2,   0, -11, -12,   5,   1,   1,   0,   0, -13, -30,  49,  -6,   0,  -1,   1,  15, -34,  50,  -5, -27,  11,  -4,   2,  -6,   5,   5, -28,  17,   2,  -1,   1,  -1,  -1,   0,   6,   2,  -5,   1,   0,  -2,   2,   1,  -4,   0,   1,   1,   0,   0,  -1,   0,   1,   1,  -1,   0,   0 },
      { -29, -26,  14,  18, -10,   4,  -1,   2, -26, -29,  24,  -3,  16,   1,   2,   0, -10,  29,  27,  22, -32,  -3,  -3,   0,  18,  -2,  18, -65,  -1,  19,  -4,   3,  -7,  10, -25,  -1,  43, -14,   0,   0,   3,   3,  -3,  14, -13,  -9,   6,  -2,   0,   1,  -4,  -2,   1,   4,   0,  -1,   2,   0,   0,   2,  -1,  -1,   0,   1 },
      {   6,  37,  33, -45,   2,  -5,   2,  -3, -28,   1, -44,  18,  23,  -6,   4,  -1, -46,  46,  -4,  24, -21,  -2,   2,  -2,  26,  -8, -26,   6,  -8,  13,  -4,   2,   5, -21,  20,   8,  -4,   2,  -4,   2,   4,   0,   3, -12,   1,   2,   0,   0,   0,  -3,   1,   4,   2,  -1,  -1,   0,   2,   0,   1,  -2,  -2,   0,   1,   0 },
      {  19,  27,   3, -11, -13,   3,  -2,   0,  27,  27,   6, -32,  16,  -9,  -1,  -2,   7,   7, -32, -15, -15,   9,   2,   2,   0, -27, -12, -18,  55,   3,  -5,   2, -12,  14, -18,  48,   5, -39,   9,  -3,   1,  -6,   3,   7, -34,  14,   8,  -3,  -2,   0,   1,   0,   6,   6,  -6,   0,   0,  -3,   1,   2,  -3,  -1,   0,   1 },
      {  13,  45,  25,   1,  -9,  -8,  -4,  -2, -26,   2,   9, -60,  10,  10,   3,   2,  -9, -15,   6,   6,  39, -20,   2,  -2, -22,  46, -21,  -6, -24,  -4,  12,  -2,   5,  -7, -31,  11,  20,  17,  -5,  -2,   2,  -1,   7,  20,  -5, -14,  -5,   2,   3,  -4,  -3,  -2,  -6,   4,   3,   1,   0,   0,   1,   3,   0,  -1,  -1,   0 },
      { -16, -11, -13,  11,  -1,   7,   1,   2, -32, -26,  -4,  -6,  31, -10,   5,   0, -29, -18,   3,  38,   9,   4,   2,   0,   1,  24,  28,  17,  12, -31,  -3,  -3,  -4,  23, -16,  22, -54, -11,  20,  -4,   8,  -7,  10, -15, -19,  39,  -5,  -2,   3,   1,   1,  -7,  11,   0, -10,   3,   2,  -1,   2,  -2,  -2,   1,   1,   1 },
      {   7,   3,  40, -44,   6,  -3,   0,  -1,  -2, -65,  10,   1,  24,  -3,   4,   1,  70,  -2, -25,  17,  -5,  -9,   2,  -2, -29, -13,  11,  11,  -6,   5,   1,   0,  -2,  17,   9,  -4,  -4,  -2,  -3,  -1,  -3,   0,  -7,  -3,  -2,  -1,   1,   2,  -1,   3,   1,   1,   1,   1,   2,   0,  -1,   1,  -2,   0,  -1,   0,   1,  -1 },
      {  -7,  -4,  11, -16,  21,  -9,   3,   0, -22, -19, -11,   3, -17,  18,  -2,   4, -32, -26,  -1, -15,  41,   4,  -2,   3,  14, -13,  20,  25,  19,  -5,  -7,  -1,   1,  10,  21,  -9,  25, -42,  -3,   3,  -7,  20, -21,  24, -30, -22,  32,  -7,   1,   0,   2,  -1, -18,  29,  -3,  -8,   0,   3,  -1,   0,   4,  -1,  -8,   4 },
      {  13,  27,  47, -23,   4,   3,  -4,  -2,  -3,   3,  23,   0, -28, -10,   2,   1, -17,  -3,   4, -22,   4,  30,  -8,   3,  28, -13,  34, -18,  12, -34,  -1,   3, -19,  20, -25, -29, -14,  18,  23,  -5,   4,  -1, -14,  10,  33,  10, -20,  -1,  -3,   1,  -2,   4,  -4, -19,   4,   5,   1,  -1,  -4,   1,   2,   4,   4,  -1 },
      {  13,   5,  26,  39, -22, -10,  -4,   0,   9,  -5, -15, -15,  -3,  20,  -4,   2,  40,  18,  32,  -1,  22, -26, -10,   1,  53, -23, -13,  -1, -20, -24,   9,  -2, -15, -13,  11,  -4, -21,  -6,  14,   0,  -5,  10, -11, -16, -17,  22,  15,  -5,  -6,  -4,  -4,  -1,  13,  19, -11,  -2,  -1,   2,   0,   0,   3,  -5,  -3,   3 },
      {  -4,  -9,  14,  30, -33,  12,  -8,   2, -18, -29, -18, -27,   8,  11,   2,   2, -10, -29,   1,  -7,   8,  18,  -6,   3,  11, -34,  -6,   6,  30,   6,   5,   1,  -5,   5,  33,  26,   8,  15, -21,  -6,  11,   3,  17, -25,  20, -32, -25,  14,  -5,   6,  -8,   7,   0, -28,  27,   4,   3,   4,   2,   2,  -8,   6,   6,  -8 },
      {   9,  22,  41,  27, -35,  -4,   3,  -1,  -6,   6,  36,   3,  -9,   0, -11,  -3, -25, -29,   2, -10, -30,  -5,  16,   3, -29,   8,  11,   6,  -2,  37, -13,  -3,  18,  -3,  25, -22, -11, -19,  -8,  13,  -8,   3, -10, -30,   3,  22,  21,  -5,   2,   2,  -4,   1,  21,   5, -16,  -5,  -3,   1,  -3,  -1,   1,  -6,   1,   5 },
      {   3,  -1, -13, -27,   0,  18,  -4,   3,  12,  15,  30, -25,  40, -18,  -3,  -1,  18, -10,  22,  13,   0,   7,  -2,  -4,  41,  33,   3,   3,  13, -11, -27,   2,  17, -27,  12, -19,  -8, -30,  -5,  12,  -9,   3,  -8, -21,  18, -23,  19,   4,  -7,   9, -16,  18, -13,  -8,  23,  -9,  -2,   0,  -2,   1,  -8,  13,   0,  -4 },
      {  -8, -17, -19, -20, -14,  10,   9,   1,  -8, -18, -15, -28,  11,   7,  -7,   4,   3, -20,  -4, -13,   4,   3,  20,  -4,  10,  -6, -17,  -2,   4,  16, -15,   5,   5, -30,   8, -14,  20,   0,  35,  -2, -14,  25,  -5,  29,  23,  41, -18, -29,   8,  -3,   9,   8,   6, -27, -40,  17,  -1,   2,   1,  -1,  -5, -11,  17,   6 },
      {   7,   8,  16,  38,  -6,   9,  -7,  -2,  -5, -25, -26,  10, -33,   3,   3,   5,  17,  29, -36,  11,   4,  38, -11,  -2, -15,  29,   9,  -9,  20, -21, -26,  12,  35, -45,  -9,  -2,   3, -17,  19,   6, -11,  -5,  19,   5,   4,  -4,   9,  -6,   4,   1,  -3,   1,  -8,   0,   6,  -3,  -2,  -1,   5,  -1,   0,   5,  -3,  -2 },
      {  -5,   3,  -4,  11, -41,  23,  -6,   4,  -2,  16, -29,   3,  41,  -2,  -2,   2,  -7,  37,  17, -18,  11, -19,  -2,  -2, -54, -14,  15,   3,  10, -22,   7,   4,   3,  23,   4, -24,  -3,  -4,  13,  -4, -11,   3, -28,  14,  18,  -6,   8,  -6,   2,   6,  -1,  25, -25,   1,   3,  -6,   1,  -3,   0,  -8,  -4,  12,  -7,   2 },
      {   3,   5,   0,   1,  -9,  -4,   9,  -3,   4,   9, -19,   1,  -4,  29, -18,  -1,  16,  37,  11,  -7,  19,   4,   8,  -2,  11,  35,  21,  23,   7,  24, -31,  -9,  -2,  36,  -6, -10,   7, -27, -24,   6,  12, -13, -13,  -1, -10,  -2, -24,  10,   2, -12,  11, -21,  17, -36,   2,  25,  -2,   1,  -8,   2,  -5, -10,  32,  -8 },
      {  -8, -13, -17, -22, -31,   3,   2,   4,  -9, -18, -32, -18, -12,  12,   8,  -4,   3,  -1, -22, -27,  -3,  11,   3,   7,  21,  17,   1,  -3,   0,  18,   6, -12, -15,  11, -35, -12, -11,   9, -28,  19,   7, -31,   2,  -8,  28,   7,  39,   9,  -5,  13,   4,  19,  15,  19, -11, -32,   2,  -3,   2,   2,   2,  -8, -21,  14 },
      {  -9,  -6, -19, -42,  12, -18,   9,  -1,  -1,   5,   2, -34, -21,  38,  -7,   2, -11,  21,  17,   0,  16,  -4, -23,   6, -36, -23,  25, -31,   2, -14,   6,   8,  29, -11,  13,   0,  -4,  -2,   6,   1,  -5,  -2,  11, -38,  -1,   8,   5,   1,  -4,   6, -20,  -3,  36,  -5,   6,  -3,  -4,   4,  -3,  16,  -4,  -8,   5,  -6 },
      {  -2,   2,  -9,  14,  31, -15,   0,  -5,  -9,  -2, -21, -51, -16, -29,  27,  -1,   8,  15,  21,   2, -25,  43,   1,  -8, -25,  11, -18,   7,  -6,   6, -21,   9, -40,  21,  24, -14,  -1,  -1,   9,   5,  11,  24, -23, -12,   2,   3,   4,  -2,  -4,  -7,  -2,  -1,   5,  14,   5,  -2,   1,   5,  -1,   3,   9,   1,  -8,  -2 },
      {  -1,  -2,   5,  12,  23,  -4,   1,  -2,  -8,  -9, -11,   3,  14, -19, -21,   9, -10, -11,   5, -23,   9, -21,  17,  -4, -19, -36,  -6,   0, -21,  -2, -30,   6, -32, -30, -30, -11,  -5, -23,   9,  31,   9, -25,  19,   8,   5,   4,  20,  11,  -3,   5,  13, -10,  15, -18,  35,  -3,  -2,   8,  -7,  15,  -6,  13,  14, -24 },
      {  -1,   5,   8,  -3,  -5,   5,   4,   1,   5,  15,  26,  21,   5,  32,   6,  -9,   3,   8,  14,  33,  26,  19, -19,  -5, -22,  -4,  -4,   5,   5,  -4, -24,  -3, -58, -10,  12,  13,   5,  -1, -11,  23,  14,   8,  36,   8,  19,  -2,  14, -15, -10,  18,   1,   4,  -9, -15, -33, -15,   4,   0,   1,  -7,  -8, -14,  -4,  24 },
      {   1,  -2,   6, -17, -19,  41, -11,   1,   4,   3,   2, -12,  -1, -26,  23,   1,   6,  11,  42, -20,   9,  26,  15, -16,   0, -26,  30,  26, -21,  15, -17,   3,  28, -11,  -8,  11,   1,  12,  -8,  -1,   1,  -9,  36,   8, -19,  -2,  -7,  -5,   5,   3,   2, -35,  -3,  20, -10,   0,   0,  -3,  -4,  -1,  21,  -9, -13,   7 },
      {  -9, -10, -11, -37, -65,  29,  -8,  -1,  -4,   3,   8,  14, -29,   1,  -9,   9,  -5,  -9, -21,  12,   6,  -2,  -2,  -8,  -6,  16, -19, -16, -15, -13, -12,  13, -41,  -5,  14,  -8,  -1,  -4,  12,  -1,  -1,  12,  -9,  -7, -24,  -1,  -8,  13,   2, -17,  -1, -21,   5,  17,  29,  13,   0,   3,   2,   4,  18,  10,  -1, -15 },
      {   0,  -5,   1,  -3, -11,  -1,  -4,  -1,   3,  -1,  11, -10,  -5,  28,   1,  -2,  -4, -18,  22,  34, -19,  21, -35,   0,  -5, -29, -46,  19,  -2, -12, -14,  18,  36,  13, -15, -16,  -3,  -7,   3,   5,  14, -43, -22,  21,   2,   2,   5,   9,  11, -16,  37,   2,  -3,   1,   4,   0,   0,   1,   1, -14,   2,  -6,   2,   0 },
      {  -5,  -4,  -4,  -7, -18, -36,  19,  -9,  -3,  -2,   7, -13, -20,   8,   7,  -5,   0,  10,  20,  16, -15,   6,  14,  -1,  -7, -15,  19,  16,  10,   8,   4, -21, -20, -36, -22,  14, -26,  -3, -27,  -5, -58,  16,  -9,  17,  -2,   4, -12,  28,  18,  -6,  13,  15, -21,  -3,  14,   3,  -6,  -3,   5, -17,  -5,  15,   1,  -3 },
      {   2,   4,   5,  11,  22,  26, -18,   0,   1,   5,   4,  17,  37,  11,  31,  -3,  -5,  -5,  -9, -11,  23,  19, -23, -14,  -1,   7,   2, -10,   2,  10,   4,  13, -15, -16,   2,  -8,   8,   1, -13, -25, -41,   1, -20,   0,  -2,  14,  -8,  35,  10,  -8,  27,   6,  46,   3,   4, -10,  -8,   9,   0,  12,  -4, -30, -15,  -8 },
      {   2,   2,   2,   1,  33,  56, -19,   6,  -4,  -5,  -4, -21, -18,   0, -25,  -1,   0,   4,   1,   9,   1, -21,  11,   8, -11,  -6,   9, -29,  19,   1,   4, -25,  -7, -30,  15,  -2, -29,   3, -30,   9,  29,  -7, -26,   5,   0,  -3,  -4,  10,   2, -26,  16,   5, -10,   6, -13,  12,   9,  -6,  11, -19,  16, -14,   8,  17 },
      {  -2,  -1,  -3,   4,   3,  -2,   2,   4,  -4,  -6,  -8,  -3,  31,   5, -16, -23,  -8,  -9, -16, -28, -25,  14, -46,  13,  -5,  -5,  -8,   3, -17, -23, -12,  -8,  -7,  -4, -24, -13,  -4,   3, -19,  15, -31,  15,   5, -16, -11, -19,  -1,  -2,  -9,  -5, -25, -19, -11,   9, -18,  36,  -7,  -8,   8,  -3,  30, -13,  31,  23 },
      {   0,   0,  -2,   0,   3,  16, -25,   4,  -1,  -1,  -2,  14, -30, -25,  18,   1,  -3,  -2,  -1,  24,  34,  -6,  11, -18,  -1, -24, -27,  -9, -15,  10, -26,  -8,  20,  33, -33,  18,  -4,  -8, -11,  19, -28,  20, -19, -21,   9,  -6,   4,  -6,   5,  -2, -28,  31,  10,  -8,  -8,  14,   7,  -6,  15,   8, -17,  -5,  14,   6 },
      {   5,   9,   6,   5,  27,  58,  21,   4,   0,   2,  -1,  -8,  -6,  44, -36,   8,   1,   5,   4,  11, -30,  -4,  14,   6,   6,  -1, -12,  25,   5,  -1,   3,  -1,  -2,  28, -19,   0,   7,   8,  11,   4, -27,  33,  24,  -3,  11,   4,   7,   0,  18,   1, -10,  -4,   5,   6,  15, -16,  -6,   5,   6,  10,   8,   7, -22, -19 },
      {   3,   1,   0,   8,   4,  -8,  56, -22,   5,   1,   4,  11,  31,  27,  -1,  -5,   0,  -6, -13,  -9,  26,  25,  15, -22,   1, -12,  -5, -35,  -9,  14, -34, -13,  18,   1,  -2,   8, -22,  19,  -1,  11,  19,   4, -25,  -4,  -1,   4, -16,  -9,  -4, -13,  -7,   0, -14,  18,  12,   1,   3,   1,   9,   3,  19,  19, -30,  -3 },
      {   2,  -1,  -2,  -4, -11,  -7, -12,   6,   5,   1,   3,   5,   9,   0,   7,  -3,   3,   1,  -6, -17,  -9,   1, -24,  15,   2,   1,  23,   8, -24,  -3,   0,   9,  14,  -7, -14,  31,   0,  -2,  -1,  21,  48,  64,  -9,   2,  10,   3,  16,  10,  42, -37,  13,   5,   6, -16,   9, -22,   1,   6,   1, -10,  -8,  -3,   4,  -4 },
      {   0,  -1,   1,   0,  -4,  16,  21,  -1,   0,  -1,   5,  11,   1,   3,   2,  -1,   2,   0,   9,   7,  -5,  21,   5,  14, -10,  -7,   7,  42,  -3,  -4,  20, -11, -15, -39, -33,  -1,  32,  -8,   2, -27,  38, -12, -30, -15, -13,   4,  -6,  -8, -17, -10, -39,  35,  19,   9,  -1,  16,   7,  -1,  16,  16,  -9,   1,   1,  -5 },
      {   0,   1,   2,  -2,  12,  16, -10,  -7,  -2,   0,   1,  -8, -11,  26,  -1,   3,  -3,  -4,  -3, -26, -20,  -9, -21, -10,   6,  14,  27,   4, -46,  15, -16,  30, -12,   3,  15,  48,  -3,  -8,  20,  -2, -13, -27, -11,  -6,   0,   2,  -4,  -9, -12,  21,  12,  36, -20,   4,   3,  21,  22,  -9,  17, -21, -10,   9,  -1,  -8 },
      {   1,   1,   1,   0,  -5,  -9,  14,  17,   3,   4,   8,   5,  11, -12,   7,  -1,   3,   6,   6,   9,  19,  15,  11,  30,   2,   4,   8,  -1,  13,   1,  14, -11,  -3,   6,  10,  12,  15,  17,  12,  26, -17, -13,   8, -12,  21,  -6,  25, -24,  10, -15,  27,  10,   9,  25,   6,  26,  15,   0,  22, -17,  32, -23,  26, -58 },
      {   0,   1,   1,   2,  -8,  20,  56, -29,  -4,  -4,  -2,  -3, -14, -53, -40,   1,   2,   3,   4,   7,  14,   3, -30,  13,   0,   1,   3,   0, -14,  -1,   9,   4,   6,   6,   2,  24,   9,  -5,  -1,  -2, -15,   4, -11,  -5,   8,  -6,   3,   4, -37,  -4,  31,  -1,   5, -11,  -7, -18, -25,  18,  -3,  -3, -11, -17,  -5,  12 },
      {  -1,  -3,  -1,  -1,   5, -12,  -7, -22,  -1,  -3,   1,  -1,   6,  -8, -18,  16,   0,  -1,   2,  -4,   8,  -2,   5,   1,   3,  -1,   6,   1,  -1,   0,  -8,   8, -20,   6,   7,   1,  16,  -4,  14,  -7, -16, -43,  18, -29,   9, -12,  -9, -12,  55, -68, -17,  12, -11,   3, -11, -10, -19,   0,   2,  -3,  -2,  -1, -17,  30 },
      {   0,   1,   1,   0,   4,  19,  -3,  32,  -1,   1,   1,   2,  -2,  27,  44, -14,  -2,  -1,  -4,  -2,  -3,  11,  22,  19,  -2,  -1,  -1,  -8, -20,   5,  18, -16,  -1,  -1,  -2,  13,  -8, -12,  16,  -2, -13, -11, -12, -13,   4,  -9,  15, -24, -32, -33,  13, -19, -19, -19,   5,  -7, -44,  30, -19,   7,  -8,  26,  19,  13 },
      {  -3,  -3,  -2,  -1,  -6, -28, -50,   9,   0,   3,   5,   5,  13,  10, -45,  25,   0,   4,   3,   7, -14,  19,  31,  -8,  -1,   2,  12,  12,  -1, -20, -11,  -1,   6,  -3,   5,  33,   5,  24,   0,  22, -11,   0, -28,  21,   9,  -9,  -5,  -6, -37,  -6,   0,   0,  22,   8,   5,  -8, -13,  12,   6,  25,  13, -16, -13,  16 },
      {   1,   1,   1,  -1,  -2,   8,   9,  15,   2,   1,   0,  -2,   7,  -3, -21,  18,   1,   0,  -2,  -8,  -1,  55,  16,   5,  -1,   0, -12, -23, -47, -25,  26, -31,   2,  11,  31,   1,  -8, -18, -18,   1,  -3, -13,   6,  25, -23,   4,   6,   4,  36,  20, -14,  14,   5, -13,  -3,  -6,  -4,  -7,   5,  -2, -26,  -2,  13,  -7 },
      {   1,   1,   1,   3,  -4,  13,  23, -12,  -1,  -2,  -1,  -2, -11,  -6,  35, -29,  -1,  -3,   0,  -3, -21, -34, -11,  -8,   6,   7,  10,   6,  12, -25, -19,  -1,  -1,   1,  28,  12,   2,  24, -13,  33,  -5, -25, -10,  46, -11,   3,   4,  -8,  13,  -8, -39,   6,  18, -11,   4,  -5, -29,  12,  14,  19,  -4, -11,   4, -15 },
      {  -1,  -1,  -2,  -3,  -3,  -2,  -9, -43,   0,   2,   1,  -2,  -2,  10,  18,  -2,  -3,  -2,  -8, -10, -21, -12,  26, -33,  -1,  -1,   0,  -3,  -8, -51, -11, -37,   7,   6,  -4, -11,  24, -13, -11,  -7,  14,  13,  24, -27,  -4,  -5,  -8,   0, -10,   5,  38,  26,  -1,   5, -10,   9, -17,  28,   5, -19,  -2, -17,  -4, -14 },
      {  -2,  -2,  -2,   1,   0,   6, -24, -76,   1,   1,   1,   2,   5,  12,   8,  26,   0,   1,   1,   6,   4,   7,   2,  26,   0,  -1,  -4,  -2,   1,  23,  13,   8,  -5,  -5,  -9,   5, -26,  -5,   6,  -8,  12,   5, -14,   9,  13, -17,  12, -11,  13,  30, -15, -14,  -3,   6,  -5,   8, -52, -15,   0,   0,   8, -11,  18, -18 },
      {  -2,  -2,   0,  -2,   0,   3, -25, -47,   0,   0,   2,   1,   6,   4,   4, -14,   1,   1,   4,  -5,   5,  12,  -3,  31,  -1,   1,  -3,  -2,  -6,  11,  11,  -8,   9,  13,  12, -18,  -3,  -4,  -1,  15, -10,   8,  36,  23, -15,  25,   0,  13, -51, -49,  -8,  18,  -1,  -9,   9,  -9,  28, -22,  17,   5,  -8,  16,  -2,  -7 },
      {  -1,   0,   0,  -3,   2, -11, -21,  31,  -1,  -1,  -1,   0,  -2,  -7, -29, -43,   2,   3,   4,   0,  18,   6,  -5,  10,   1,  -1,  -2,  -3,   1,  24, -14,  10,   2,   7,   0, -11,   0,   6,   4, -12,  17,   4,  20,   3, -12,  21,  -7,  19,  -3,  12,  -7,  31,  -9,  10,   3,   9, -69,  23,  33, -24,   6,  10, -11, -12 },
      {  -2,  -3,  -1,  -2,  -5,  -3, -24,  -4,  -3,  -4,  -3,  -3,  -5,  -3, -26, -68,   0,  -2,   0,   1,  -3,   2, -12,  -3,  -3,  -2,  -2,  -1,  -8,   4, -13, -37,  -4,  -6,  -1,   9, -10,  -6,  10, -26,  -3,  -3, -11,  -1,   9, -20,  -5, -29,   9,  -5,   6, -14,  -3,   1, -11, -15,  10, -24, -28,  24, -20, -13, -25, -45 },
      {   1,   1,   1,   0,  -1,   5, -10, -20,   2,   2,   2,   2,  -1,  -5,   1, -19,   2,   3,   4,   6,   5,  -1, -14,  29,   1,   4,   1,   1,  -2,   5,  -1, -34,   5,   3,  13,  11,  11,  -7,  11,   7,   3,  -3,   4,   6,  19,  16,  -3,  14,  31,  25,  21,   8,   1,  23,  14,  35,  15,  38,  -5,  57,  11,  31,   1,  40 },
      {   0,   1,   0,  -1,  -7,  -4,  11,   7,   1,   1,   1,  -1,   0,   2,  -3,  11,   1,   1,   0,  -3,  -9,  -1,   0,   1,   1,   0,   1,  -6,   0,  17,  -6,  18,  -2,   3,  -6, -15, -54,  -1,   4, -29,   4,  10,  28,  28, -10, -50,   6, -19,  -4, -18,   0,  31,  17,   9,  -6,  17,  21,  54,   0,  18, -13, -24,   4,   1 },
      {   1,   0,   1,   3,  -2,  -2,  23,   7,   0,   0,   1,   2,  -2,   7,   5, -29,   1,  -1,   2,   1,  -3,  -4,  38,  -6,   0,  -2,   0,  -4, -11, -16,  -4,  25,   5,   7,  10,  15,  -4, -24,  19, -23,  10,   3,  11,  11,  38, -14,   6,  58,  -8, -11,  -3,   3,  -2,  25, -12,  -2,  -8, -40,   7,  26,  10, -22,  32,  12 },
      {   0,   0,   1,  -1,  -2,  -8,   7,  28,   0,   1,   0,  -1,  -2,  -4,  24,  39,  -1,  -2,   1,  -5,  -7, -23, -20,  33,  -2,   0,  -4,   0,  -9,   0, -26, -40,   1,   2,   4,  -2,  -7, -32,   4, -18,   3,   3,   8,  11,  18, -11, -27, -13,  -8,   2,  10,  -9,  15,  -7,  10,  -3, -16, -33,  49,   8,  11,  -9, -36,   5 },
      {   0,   0,   2,  -2,  -1,   2,   3,  -2,   0,   0,   1,  -3,   2,   0,  -6, -12,   1,   2,   3,  -3,  -2,   5,   8, -15,   2,   2,   6,   0,  12,  -5,  11,  25,  -4,  -3,  -2, -15,  -4,  23, -12,   8,   2,   9,   0, -14, -19, -15,  14, -27,   9,   0,  47, -12,   9,   4,  -4,  15, -14, -37,  55,  36, -47,  30,   6,  14 },
      {   0,   0,  -2,  -1,   0,  -2,  -3, -19,   0,   1,  -1,   1,   1,   3,   3, -46,   2,   1,   0,  -1,   4,   2,  36,  31,   0,  -1,  -5,  -1,  -6,  -9,   4,  15,  -1,  -2,  -4,  -3,   1, -10,  19,   1,  -8,  -5, -17,  -5,  -6, -24,  -9,  -2,  12,  24,  -6, -28,  24, -24,  11, -20,  32,  26,  45, -41,   7,  -9, -13,  28 },
      {   0,   0,   0,   0,   0,  -4,  -5,  -6,  -1,  -1,   0,  -1,  -1,  -2,  -9,  -4,  -1,   1,   2,   2,  -1,   5, -26, -78,   1,   2,   4,   6,   2,  20,  42, -29,   0,   1,   3,   4,  -5, -23,  19,  -6,   4,  -3,   5,  14,  26,  -1,  25,  -1,  -3, -12, -14, -20,  10, -15,   9,   4,   3,  13,  40,  -7,  24,  12,   8,   3 },
      {   0,   0,   0,   1,   0,  -5,   0,  -1,  -1,  -1,  -1,   0,  -4,  -4,   0, -14,   1,   1,   0,   4,   2,   0,  13,   4,   0,  -1,  -2,  -6, -10, -21,  22,  50,   0,   2,   0,   7,  -2, -38, -60,  -9,   0,   2,   4,   7,  26,  18,   3, -30,   1,  -7,  -1,   0,   2, -21,  -5,  34,  -1,   1,  -9,  13,  35,  13, -42, -11 },
      {   0,   0,   0,   0,   2,   0,  -2,  -6,   0,   0,   1,  -1,   2,  -3,  -5,  10,   0,   0,   0,  -1,   1,  -2,   2,  13,  -1,  -1,   1,  -2,   0, -15, -42, -13,   2,   5,   7,  13,  11,  31, -14, -60,  -3,  -1,  -3,  11,  12,   2,  57,  12,  -3,  -9,  -7, -17,  29,  -8, -17,  18,   7,   1,  -2, -29,  -6,  46,   4,  -1 },
      {   0,   0,  -1,  -1,   0,   1,   1,   5,   0,   0,   0,   0,   4,   1,   4,   6,   0,   0,  -1,  -2,  -1,   1,  -2,   3,   0,  -1,   1,   1,   2,  -2,  16,   1,   1,  -2,   0,   0,  -1, -16,  10,  42,   1,  -1,  -2,   7,  25, -18, -20,  31,  -7,  -9, -15, -30,  17,  32, -37,  40,  -4,  14,  -7, -26, -54,  19, -36, -19 },
      {   0,   0,   0,  -1,  -1,  -4,  -8,   0,   0,   0,   0,  -2,   1,   0,  -2,  16,   0,   0,   0,  -2,  -1,  -1,  -4,   9,  -1,  -2,  -1,  -3,  -1,  -4, -12,  10,   0,   0,  -1,   1,   3,  -4, -27, -19,  -2,  -6,  -9, -10,  17,  28,  -7,   6,  -2,  -2, -13, -36, -50,  15, -21, -30,  19,  46,  48,  28, -21, -10,  20, -32 },
      {   0,   0,   0,   1,   2,   3,   0,  -2,   0,   0,  -1,   1,   2,   1,  -4, -10,  -1,   0,  -1,  -1,   1,  -1,  -3,  -5,   0,   0,  -1,   0,   1,  -3,  -4,   2,  -1,   3,   1,   2,   2, -11, -25, -13,   0,   5,  10,  15,  31,  25, -19, -44,  -3,  -4,  -7, -11,  19,  56,  56,  -7,  -3,  -4, -11, -34, -36,  -9,  31,  24 },
      {   0,   0,   0,  -1,  -2,  -3,   3,   0,   0,   0,  -1,  -1,   1,   2,   2,   0,   0,   0,   0,  -1,  -1,   2,   9,   1,   0,  -1,   0,  -1,   2,   3,   1,  -7,   1,   1,   0,  -1,   2,  14,  11,  -6,   2,   1,   0,  -5,  -2,  22,  43,  11,  -2,  -4,  -9, -18, -37, -21,  40,  51,  -1,  -3,   1,   2, -33, -65, -37,  16 },
    },
    {
      {-105,  48,   8,   4,   2,   1,   0,   0,  45,   7, -17,  -4,  -3,  -1,  -1,   0,  10, -16,  -7,   4,   1,   1,   0,   0,   4,  -4,   4,   4,  -1,   0,   0,   0,   2,  -3,   1,  -1,  -1,   0,   0,   0,   1,  -1,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -46, -55,  41,   8,   5,   1,   3,   0, -54,  49,  31, -13,  -3,  -4,   0,  -2,  27,  24, -29, -18,   0,   1,   0,   0,   5,  -9, -15,   7,   8,   1,   1,   0,   6,  -1,  -2,   6,  -1,  -2,  -1,   0,   1,  -3,  -1,   1,  -2,   0,   1,   0,   3,   0,  -1,   1,  -1,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
      {  37,   2,  16, -16,   2,  -2,   0,  -1,  18,  65, -50, -23,   2,  -2,  -1,   0,  22, -42, -45,  26,  11,   3,   1,   2,  -6, -19,  22,  26,  -3,  -2,  -1,   0,   0,   1,  10,   0,  -8,  -3,   0,  -1,  -3,  -2,   4,   0,  -2,   1,   1,   0,  -1,   0,   2,  -2,  -1,   1,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0 },
      {   5, -70,  19,  18,   1,   4,   0,   2,  93, -12,  24, -12, -12,  -3,  -4,   0,  -7,  -4,  -7, -20,   1,   2,   1,   1, -17,  -1,   0,   2,   8,   1,   1,   0,  -7,   9,   2,  -2,   1,  -1,   0,  -1,  -4,   4,  -1,   0,   2,  -1,   0,   0,  -2,   4,  -1,   0,   1,  -1,   0,   0,  -2,   0,  -1,   1,   0,   0,   0,   0 },
      {  24,  50, -34, -11,  -3,   2,  -4,   1,  26,  14,  34, -21,  -6,   0,  -4,   2,  28,  38, -34, -45,   6,  -1,   1,   1,   8, -23, -44,  17,  18,   4,   3,   1,  -2,  -6,   8,  21,   8,  -5,  -2,  -3,  -1,  -3,   1,   7,  -2,  -5,  -1,  -1,  -4,  -1,   3,   0,  -2,   0,   1,   0,  -1,  -2,  -2,   0,  -1,   0,   1,   0 },
      {   3,  22,  50,  -3, -10,  -7,   2,  -4, -10, -12,  -2, -29,   2,  -5,   7,  -3, -86,  -3, -19,  -9,  21,  12,   4,  -1, -10,  10, -17,  22,  31,  -5,  -3,  -2,   4,  -5,  -6,  26, -15, -11,  -1,   2,   5,   1,   5,  -3, -19,   0,   3,   4,   9,  -4,  -3,   1,  -3,   3,   3,   2,  -1,  -1,   4,   3,   1,   2,   0,   0 },
      { -14, -20, -65,  48,   6,  -2,   3,   0, -11,   6,  -7,  42,  -6, -11,   3,  -4, -26,  -6, -21,  -1, -10,   6,   4,  -2, -19, -29,   8,  34,  15,   3,  -2,  -2,   3,   0,  28,  18, -27,  -8,  -3,   1,   3,  -2,  10, -19, -20,   6,   3,   4,   9,   0,   5,  -6,  -1,   5,   2,   0,   0,   1,   3,  -4,   1,   2,  -2,  -1 },
      {  10,  11,  23,  13,  -9,  -9,   1,  -4,   3,  30, -22,  28, -19, -12,   2,  -1,   8, -24,  30, -25, -48,  17,   5,   0, -27, -17, -20, -48,  32,  28,   2,  -2,  -5, -11, -14,  29,  24,  -1,  -6,   1,  -2,   4,  10,   4, -13,  -7,  -4,   4,   9,  -4,  -3,   0,  -2,   0,   2,   4,  -5,  -3,   4,   3,   2,   0,   2,   0 },
      {   5,  20,  14,  18, -12,   5,  -3,   1, -28, -44,  38, -19, -22,   5,   0,   8,  41, -61,  -4, -20,   8,  21,   3,   3, -24, -14,  30,  15,  24,  -2,  -4,  -2,  -7,  23,  12, -13,  -4,  -9,  -2,   1,  -4,  11,  -9,  -5,   3,  -1,   2,   2,   4,   6,  -8,  -2,   5,  -1,  -2,   0,  -2,  -3,   0,   3,   1,  -3,   1,   0 },
      {   4,  22, -24,  30,   3,  -1,  -4,   1,   1,  61,  35, -15, -24,  -6,  -3,  -2, -45,  -6,   3, -19,  -4,   9,   7,   1,  -3,  23,  32, -15,  -9,   5,  -1,  -2,   8,  16, -20, -42,   8,   8,  -1,  -1,  -1,  -9, -20,   6,  25,   5,  -2,  -1,   0,   0,   0,   7,   5,  -7,  -5,  -1,   2,   0,  -5,   0,  -1,  -2,  -1,   2 },
      {  21,  27,  50,  56, -21, -20,  -4,  -3,   9, -15,  15,  19, -12, -14,   7,  -8,   7,  22, -28,  20, -23,  -6,   2,  -3,  28, -25,   4,  -3, -34,   6,   3,   2,   0, -23,   6, -16, -17,  26,   6,   1,   3,  -7,  10, -15,   2,  13,  -4,   0,  -2,  -3,   4,  -5,   5,  -1,  -1,   0,   1,   3,  -2,  -3,   3,   3,  -1,   0 },
      {   6,  20,  33, -42,  26,  -9,   0,  -2,  11,  17,  25,  50,  19, -13,  -9,  -6,   1,   7,  12, -16, -17, -27,   1,   1, -31,   3,   6,  12,  -1,  -6,   6,   2,   9,  24,  24, -12, -29, -18,   0,   0,   0,  -5, -30, -30,  -6,  15,  18,   1,  -2,   0,   1,   9,  10,  12,   1,  -7,   2,   4,  -6,  -5,  -2,   1,  -5,  -1 },
      {   9,   0,   8,  35,   1, -12,   3,  -4,   2,  -7, -43,  25, -23,  14,   1, -12,  30,   4,  -5, -36,  33,   5,  -2,  -1,  17,  51, -10,   2,  23, -17,  -9,  -5,  19,  -2, -17,  -4, -15, -20,  -4,  -1,  -2, -31, -19,  -1,   2,   9,  15,   2, -10,  -9,  21,  17,   0,   2,  -2,  -8,   6,  11,   0,  -9,  -6,   0,  -6,   2 },
      {   7,  -1,  -3,  -2,   7,   2,  -1,   5,  12,  26,  39,   4,   0, -16,  -4,  11,   9, -33,  42,  10,  -3,  14,  -1,   2,  66,  22,  -5,  14,   6, -19, -15,  -5,   4, -26,  -9,  16, -32, -19,   0,   6,  -5,   7,  22, -11, -23,   4,   6,   7,   2,   5, -12, -19,  -3,   7,   5,   2,  -7,  -7,   0,   3,   7,   5,   4,  -3 },
      {  -4, -13,   1, -40,  12,   2,   0,   2,  -2, -20,   1,  23, -12,  -5,   3,  -7, -22, -25, -34, -13, -20,  23,  12,  -1,  62, -11,   7,  -2,  12,  40, -12,  -2,  19,   0,  25, -10,  23,  -1, -24,  -1,  -2, -13,  -9,  -3,  20, -14,  -2,   2, -17,  -2,  10,  17,   1,  -7,   5,   0,   4,  10,   0,  -3,  -3,  -1,  -1,   1 },
      {   5,   8,   8,  27,  40, -11, -10,   3,   5,  -5, -17, -27,  21, -20,   8,   7,  12,  49,   9,  10,  -7,  18,  12,  -3,  19,  -5,  41,  -8,  17,  15, -16, -12, -18,  32,   5,   0,  12, -35, -19,  -3,   3,   8, -18,  -4, -28, -23,   6,   7,   5,  -6, -10,  -5,  -7,   7,  12,   6,   2,  -4,  -4,   4,   0,   6,   1,  -2 },
      {  -2,  -3,   9, -15,  -6,   5,  -4,   2,  -1,  12, -31, -12, -36,  17,   6,   4,  -4,  36,  46, -25,  14,  33,  -2,  -4,   4, -23,   9,  13,   6,  -7, -14,  -2, -12, -21,  43, -16, -16,  15,  -2,   6,   3,  18,   5, -33,  26,  15, -12,  -1,   0,   7, -14,   0,  20, -10, -11,   4,   0,   1,  -3,   2,   5,  -8,   0,   3 },
      {   2,   4,   9,  18, -47,  16,   3,  -3,   6,  22,  16,  11,   4,  45,   0, -15,  -9,   1,  21,  16,  30, -17, -29,  -3,   6, -21,   5, -12,   1,  -3,  -6,  13,  28,  12,  32,   3,  30, -19, -17,   1,  -3,  -1,  -3, -14, -13, -38,  14,   8, -12,   1,   0,   0, -11,  10,  24,  -2,   4,   9,   2,  -3,  -3,   7,  -2,  -5 },
      {   6,  11,  18,  22,  56,  -6,  -5,   1,  -4,   5,   0,  24,   0,   5,  -7,  -1,  -2, -12,   3, -21,  23, -14,   3,   3,  -8,  33,  -4,  27, -17,   4,  17,   0,  -3, -11,  29,   8,  28,  23, -10, -10,  -3,  28,  25,   6,  10, -37, -26,  -5,  -6,   6,  -8, -20, -22, -16,   5,   7,   3,   0, -10,  -6,   0,   5,   7,   3 },
      {  -2,   0,  -5, -14, -39,   3,   6,   2,   1,  -1,  -5,  -9, -41,   2,   4,   7,  13,  19,  -9,  20, -33,   6,  13,   2, -13,  62, -10,  26, -14,  31,  12, -10,  27,  29,  13,   4,  -6,   5, -21,  -9,  -2,   6,   3, -15, -15,  -9,  -8,   9,   9,   1, -15, -21,   5,  11,   6,   8,  -5,  -8,   5,   4,  10,   3,   0,  -3 },
      {   0,   9,   7, -19, -16,  17,  -1,   2,   6,   7,  27,  43,   1,   6,   0,  -5,   7,  27, -24,  32,  14,  35,  -1, -12, -28,  26,  29,  -4,  23,   0, -37,  -8, -37, -19,  -4,   3,   5, -10,  -8,  11, -11,  -6,  30,  17,   7,   4,  -6,   2,  -5,   5,  15,  -3, -10, -18,  -6,   2,  -1,   0,  -8,  -8,  -7,  -4,   4,   5 },
      {   5,   3,   0,   7,  21,   8,  -5,   1,   7,  -7,   8, -10,  22, -23,  13,   1,  15,   5,  15,  21,   5,  20,   4, -12, -32, -12, -11,  11,   6,   0, -26,   0,  82, -21, -14,  -3,  -5,  11, -17,  11,   1, -18,  -2,  -9,  23, -10, -16,   7,   6,  -2,  -4,  16,  -7, -23,   8,  11,  -5,  -2,  15,   8,  -5,   2,   6,  -1 },
      {  -1,   0,   6, -12,  27, -11,   3,  -7,  -3,  -5,   5,   1, -70,  -5,   1,   8,  -6,   4,  23,  24,   6,   0,   3,  -6, -15, -29, -20,  28, -26,   3, -11,  -3,  -9, -12, -10, -16,   8, -19,  -2,  11, -12, -13, -12,  40,  -8, -15,  25,   3,   4,  -7,  25,  -4, -18,  21,  10,  -5,   1,   4,   7,  -8,   3,   3,  -8,  -6 },
      {  -3,  -2, -11,   4,   7,   6,  -4,  -2,  -7,  -6,  -2, -33,  15, -14,  -7,  -8,   2,  -9, -13, -14,   2, -15,   9,  -3, -25,  24,  -5, -29, -13,  26, -11,  -1, -11, -61,  25, -26,   0, -13, -19,  14, -28,  -3,  15, -26,  -7,   1,  33,  15, -11,   7,  -1,  -2,  10,  27,  13,  -8,  -1,  10,   1,   1,  11,   3, -12,  -6 },
      {  -3,  -3,  -4, -26,  -4, -27,   0,  -5,   0,  -8,   6, -20, -14,   3,  -2, -17,   4,  11,  10, -32,  -5,  -8,  -3,   4,  -7,  -6,  70,  -7, -11,   0,  16,   6,  32, -19,  -1,  32,  -9,   5,  10, -10,   8, -34,  29,   9, -20,  -1,  -5,  -2,  -1, -10,  23, -21, -13,  -4,   4,  -2,   1,   6, -15, -14,   4,   8,   4,   0 },
      {   4,   4,  16,  24,  -5,  54,  -6,   0,  -1,   3,  -4, -14,  30,  18, -27,   2,  -2,  -1,  30,  -6, -30,  -3,  15,   6,   3,   2,   4,  44, -13,  31,   3,  -7,  -6,  -6,   5,  18,   6, -14, -13,   4,  16, -23,   1,  20,  12,  13,   7,  -3,   1,   8,  34,   2,  11,  -3, -25, -12,  -1,   0,   1, -18, -10, -17,  -9,   3 },
      {  -3,  -3,  -6,   2,   2, -48,   4,   1,   3,  11,  23, -13,  20,  15,  33,  -4,   6, -19,  21,  17,  30,  22,  16, -15,  -8,   6, -22,  -8, -17,  26,   0, -21, -22,  10,  22,  19,  -5,  24, -18, -14,  12, -20, -14,  -7, -18,  12, -16,  -2, -11, -18,  26,  15,  12,   2, -13,   4,   8,  15,   2, -11,   0,  -6,  -7,  10 },
      {  -3,  -7,  -3,  -3, -49,  -9,   0,   2,  -1,   4,  -1,   3,   1, -53,  -6,   0,   7,   1,  21,   3,  17, -40,  34,   3,   5,   3,  12,  21,  24,   8,  16, -10, -19, -26,  14,  -9,   2,  -2,  -4, -11,   3,   6, -18,  11,  12, -10,  -5,  -6,  10, -26,  -6,  33, -24, -19,  11,   5,   7,  -3,  -2,  -2, -16,  13,  10,  -3 },
      {  -1,   3,   3,   7,  -5, -10,  -1,   9,  -3,   2,  -4,  -7,  -3, -34,  -1,  12,  -2,   6,  11,  17,  18,  -5,  25,  -1,   4,  -2, -13, -18,  23,   7,   1,   3,  35,  36,  30,  -3,   7,  -6,  26,   1, -30,  14,  37,  16,   5,  24,  13, -27, -27,  29,  19, -11,  -3,   2, -22, -23,  -4,   7, -16, -14,  -3,  -5,  -8,   0 },
      {   0,   1,  -1,  -1,   0,  26, -19,   5,  -4,   0,  -5, -15, -13, -34, -16,  -6,   1,  -9,  -2,  -3,  -9,  10,  -8,  -3,  -7,  11, -21, -25, -16, -21, -21,   7, -17,  27,  27,   3, -21,   6,   6,  14,  68, -13,  13,  -3,   6, -11,  -1,   6, -23,  -7,  17,  -1, -20,  -9,  29,  -1,  12,  12,   1,  -9,  -4,  16,   9, -10 },
      {  -6,  -3,  -3,   0, -24, -45, -14, -12,  -5,   0, -11, -10,  23,  -2,  -8,  -5,  -4,  -3,  11, -19, -26,   6, -19,  -6,   0,  -4, -13,  23, -15,  -8, -10,   4, -21,  27, -12,  -6, -17,  -6, -15,  10, -51,  -8,  11,  -3,  23, -20,  -3,  25, -13,  28,  19,   2, -11, -28,  17,   4, -18,   5,  -6,  -7,  -9,   8,  16,  -3 },
      {  -1,  -5,   0,  -4, -28, -21, -14,  -1,   2,   1,  -7,  16,  26,   5,  -5,  28,  -2,  -7, -10, -36,   8,  13,  15, -12,   0,   0,  -5,  -3, -43,   4, -43,  -5,  22,  -2,  14, -16,  11, -20,  12,  26,  17,  24, -11,  25, -17,   6,   6, -12,  34, -22,  -5, -22, -11,  -5, -16,   5,  -4, -11,  -7,  -5,   5,   0,  10,   6 },
      {   0,   1,   3,   4,  13,  31, -22,  -2,  -4,  -3,  -1,  -7, -21,   1, -35,   6,  -6,  -4,  -5,   8,   5, -24,  -3,  -9,   4,  -4,  -7, -30,  -9,  -9, -10,  -5,   0,  10,  24,  14, -17, -10,  -7, -10, -55, -24, -13,   1, -20,  11, -34,  -2,  10, -34,   1,   1,   7, -24, -16,  34,  -3,  -1,   5,  -3,   4,  -4,  20,  18 },
      {   0,  -1,  -8,   6,  16, -47,   6, -10,   1,  -3,  10, -22, -19,  18, -21, -13,   2,  -2,  -6,  32, -28, -22, -26,   5,  -1,  18,  -5,  -5,  17,  -3, -12,  13,   8, -17,  25,  12,   7, -28,  -1,  12,  28,  19,  -2,   5,  12,   6,  11,   4,  22,   3,  -7,  20,  17, -33, -17,  -8,   7,   1, -15,  -8, -24, -17,  10,   9 },
      {  -1,   1,   0,   5,  -2,   9, -24,   0,   0,   2,   4,  -7,  -3,  -4,  41,   5,   0,  -1,  -4,  -3, -33,  14,  -5, -19,  -1,  23,  16,   9, -19, -30,  -7,  20,  -1, -12,  25,  47,  20,   3,  32,   7, -32,  16, -30,   3,   9,   3,   4, -18, -14,  -7,   0,  32, -12,  13,   4, -12,   8,   4,  21,   4,  -9,   5, -16,  -5 },
      {   4,   6,   1,   3,   9,  23,  71, -19,   3,   1,   0, -10,  10,   7,  26, -29,   4,   0,   3, -12,  -9,   2, -12,  -8,   9,   2,   0, -15,  -2,   9,   0,  -7,   0,  16,  24, -10, -39,  10,  -3,  11, -17,  15,  16,  32,   8,   0,   5,   4,  34, -15,  -2,  16, -30,  -2,  -5,  -2,   4,  -6,   7,  -9,  -5,  15,  -8,  -4 },
      {   0,  -1,   3,   2,   1, -16, -15,  14,   2,   1,   3,   5,  17,   3, -24,  38,  -1,   5,  -1,  10,   2,  36, -18,  25,   3,  12,  -3, -11,  18,  -8,  37,  14,   0,  -7,  39, -14,  -5,  16,  12,  -5,  -8, -33,   0,  19,  10, -16,   6,   3,  49,  -3,  20,   7,   7,  16,  16,  -4,  -8, -21,  14,   2,   4,  -2,  -7, -19 },
      {   1,   1,   1,  -5,  -3,   5, -10,  -3,   1,   0,   1,  11, -15, -28,  11,  11,   1,   0,  -6, -11,  28, -25,   7, -15,   9,  -4,  14, -17, -26, -14,  -7, -13,  -2,  17,  -7,  37,   3,   3, -36,   7,   4,   6,  17,  -2,  17,  -9,  -1,  27,  51,  36,   8,  15,  32,   4,  -4, -14,   7,   1,  10,   6, -10, -20, -25, -13 },
      {  -1,   0,  -1,   1,   0,   0, -22,   0,  -1,   0,   2,   9,  -5,  15, -39, -12,   0,   7,  -5,   5,  19,  35,  17,  10,  -5,  -8,   6, -19, -29,  21,  32, -14,  13,  -6,  -7,  25, -45,  -7,  -3,  -1,  -1,  35, -10,  22,  -1, -11,  17,   7, -26,  12, -23,  38,  -9,  -1,   2, -11, -14,  -5,  -7,  12, -11,   4,   3,   7 },
      {   0,  -3,  -3,  -7,   2,   3, -63, -22,   2,   0,   1,  -3,   2,  -2, -14, -29,   6,   0,   6,   4,   5,   1, -28,  -7,   7,   2,  -1,  13,  22,  24,  -1,   7,   5,  13, -19, -14,   1,  32,   0,   9,  -7,  32,  14, -12, -12,  14,   6,   6,  30, -19,  16,  12, -23,  15,  -9,  -5,  32,   7,  -1, -15,  17,   7, -21,  16 },
      {  -1,  -2,  -2,   1,   0,   1, -10, -30,   1,  -2,   1,   3,   3,  -3,   0, -36,   2,  -4,   5,  10,   9,  17,  26,   4,   1,  -2, -13,   4,  -9,   1,   5,   0, -11,  13,   6,  20,  13,  -4,  18,  15, -11, -53,  -2,  -6,  30,  -4,  22,  -6,  24,  -7, -57, -24,   0,  -8,   1, -17,  18,   2, -35,   6,  15,   6,  -4,   2 },
      {   0,  -1,   0,  -3,  -6,   5, -14,  55,   0,  -1,  -5,  -4,  -1,   1,  22,  -2,  -3,  -4,  -5,   0, -12,   3,  -9, -15,  -3,  -4,  -9, -11,   0, -12,  16,  -2,   0, -16,   2, -11, -16,  -8, -24, -17,   2, -15, -22,   3,  -7, -12,   2, -10,  17,  49, -17, -13, -35,  -2, -27,  -4,  34,  -3, -11, -34,  -6,  10,  -1,  33 },
      {   0,   1,  -3,  -1,  -9,  -1,  38, -15,  -1,  -2,   0,  -3,   3, -38, -47, -34,   3,   2,  -1,   5,  14,  31, -12,  -9,   3,   3,   1,  -1,  -8,  -4,   7,   7,   0, -10,  -1,  12,  20,   1,  13,   0,  -1,  15, -46, -19,  -5,  -5, -14,  -9,  15,  29,  18, -24,  -5,   5, -10,  12, -18,  27,  19, -19,  10,  -6,  -8,   2 },
      {   1,   0,  -2,  -3,   9,   8,  -9, -17,   2,   1,   5,  -5,  -3,  44,   2,   3,  -1,  -1,  -8,  10, -10,   0,  60,  -1,   2,  -4,  -7, -15,   7, -18,   7, -26,   4,  -6,  -8,  -3,   3,  -7,  18, -14,  -2,  18,  -3, -34,   7,   9,   2,  -3,  31,  16,  33,  -4, -15, -13,  28,   7,   1,  10,  -2, -25, -17,  28,  10, -26 },
      {   2,   2,   2,   4,   3,   3,   9,  56,  -1,   0,  -3,   2,   0,   5,  -9,   1,   1,  -2,   3,   0,  10,  -7, -19, -27,   0,   1,   2,   8,   3,  43, -15,  10,   1,   0,  -8,  18, -10,   8,  24,   2, -12,   9, -20,  17,   7,  22,  -6,   9,  13,   4,  -3,  -9,  26,  -7,  40,   2,  24,  24, -32,   2,   4,  20,   9, -39 },
      {   0,  -1,  -2,  -1,  -2,  19, -32, -41,   2,   4,  -1,   3,   3,  -6,  48,  -7,   1,  -1,  -1,   3,   2,  -8,  -3,  -6,   2,   5,  -4,  -6,  15,  13,  11,  18,   2, -14,   6, -15, -31,  -9,  18, -11,  11,   8, -21,  33,  -9, -36, -13,   4,   1,  15,   6, -16,  37, -12,  -2,   8, -32,  21, -23, -18,  12, -11,  -1,  -6 },
      {  -1,  -1,   0,   3,   5, -11, -10, -23,   0,   0,   0,  -1,  -4,  16, -10,  12,  -1,   0,  -1,   5,   5, -32,   1,  -7,   5,   2,  10,  -7,  25,  14, -27,  -3,  -4,  -2,  12,  18,  -7,  36, -14,  24,  13, -16, -33,  17,  10,  27,  -9,   3, -12,  37, -16, -13, -28,  20,   0,   0, -39, -32,  10,  -6,  10,  30,  -9,  -6 },
      {  -1,  -1,  -1,   2,  -4,  -3,   6,  14,  -1,  -1,   1,  -2,   5,  -5,   2,   9,  -2,   1,  -5,   9,  -4,   2,   2,  26,  -4,  -7,   1, -10,   7, -22,   0,  -3,   5,  -4,  -8,  23, -10,   2, -34,   2,  -9,  31, -21,  -1,  44,  -9,   7,  17, -18, -51,  13, -36,  11,   2,  -9, -15,   4, -15, -18, -45,  30,   5, -18,  -8 },
      {   0,   0,   1,   1,  -5, -18,  -9,   8,   1,   2,   0,   1,  12,  16, -18, -25,   0,   3,   2,  -1, -22,  -4,  32,  -7,   0,  -3,   3,   4,  17, -38,  -2, -13,   7,  -1,  15, -12,  -3,   1, -16,  -4,   3,   6,   7,  52,   5,  10, -13,  16,  -9,   6, -17, -11,   9,  16,   8,  18,  25,  49,  34,  20,  28, -11, -14,  -4 },
      {   0,   0,   0,   2,  -4,   1,   8, -43,   1,   1,   1,   4,   3,  -8,   7,  32,   0,   3,   1,  -2,   7,   6,  -9,  65,  -4,   0,   3,  -3,  -9,  15, -21,  19,   1,  -3,  -3,  12, -10,   0, -12, -16,  -8,   3,  -7,   9,   6,  13,  -1, -12,  -4,  24,  -6,  -2,   3,  14,   4,  20,  60,  20,  10,  -9, -11,  11,  29,   2 },
      {   1,   1,   0,   1,   2,  -6,   7,  15,   2,   3,   3,   1,  -3,  11,   5,   7,   3,   3,  -1,   4,   6, -18,  20,  -9,   4,   4,   9,  -5,  10,  21,  13,   9,  -5,   5,  -6,  13, -18,  -4,  30,  57,   3,   0,   1, -16,  24, -39,  17,  -3,  -2,   7,  26,  -1,   8,   1, -18,  29,  22,  -4,  32,  13,  37,   4,  27,  31 },
      {   0,   2,  -1,  -2,   2,   3,   5,  17,   0,   2,   0,  -1,  -3,   9,  -8,  25,   1,   4,  -2,  -2,  -6, -12,   4,  28,   0,   0,   5,   2,   7,   3, -28, -13,  -4,   5,  -9,   5, -11,  45,   7,  10,  -8,  -8,  -2,  -9, -23, -31,  29, -19,   8,  -8, -26,   8,  -8, -27,  -3, -25, -25,  59,  16, -24,  -2, -21, -11,   1 },
      {   0,   1,   1,   0,   5,   5, -16,  28,   0,   1,   2,   3,  -3,  -6,  31, -21,   1,   1,   2,   0,  14,  -2,  -9,  34,   1,   1,   1,  -1, -11,  12,   8,  12,  -4,   7,   6,   9,  14, -18,  -5, -18,   7,   5,   6,   5,  22,  45,  30,  19,  20,  -5,   1,  -9, -12, -14,  14,  15, -51,  23,  30,   3,  12,  12,  20,  31 },
      {   1,   1,   0,   0,   1,  13,  14,   5,   0,   0,  -2,   0,  -1,  -5,   6,   1,   2,   2,   1,   2,  -2,  11,  24,  19,   2,   2,   0,   7,  -1, -30,  17,  25,   5,   2,   5,  -3,  27,  30, -18,  48,  -4,   1,  -8,   9, -44,  17,   5,  38,  -1,  12,   7,  32,   9,  -5,   0,  13,  -3,   3, -48, -12,  15,   7,  23,  10 },
      {   0,   0,  -1,  -3,   3,  -3,  -9,   7,   1,   1,   1,   1,  -2,   7,   8, -11,   2,   1,   1,  -1,  12,   2,  23,   2,   4,  -1,   0,  -6,   0,  15,  23,  54,  -6,   4,  -6,  -7,  -9,  -8,  -8,  37,  -5,  -1,  -4,  -5, -10,  16, -18,  17,  -8, -12, -16, -29, -18, -12,   0,  -7,   9,  -7,  39, -22, -46, -40,  -2, -35 },
      {   0,   0,   0,   1,   1, -10,   2,  16,  -1,  -1,   1,   0,  -1,   6, -12, -17,   1,   0,   1,  -2,   1,   1,  21,   0,  -2,  -1,  -2,   6,  -8,  -1, -20,  46,  -6,   8, -13,  -3,  -1, -17,  -8,  -8,   4,  -5,  29,  -1,   5, -13, -23, -37,  10,  -2, -22,  41,  29,  31,   0,  11, -21,  -7,  10, -52,  14,  32,  -5,  -3 },
      {   0,   0,   0,   0,  -1,  -1,   0,  16,  -1,   1,   0,  -2,  -3,   9,   2, -25,   1,  -1,   2,   2,  -1,  -7,  23,  38,  -1,   2,  -5,  -5,  -5,   1, -30,  41,  -6,  -1,   0,   1, -16,  -3, -13, -14,  -4,   5, -14,  -7, -10,  -9, -19, -21,  13,  12,  32,   3, -25, -21,  -8, -24,   0, -12, -16,  57,  35, -10, -19, -17 },
      {   0,   0,   0,  -1,  -1,  -1, -11,  21,  -1,  -1,  -2,  -1,   1,  -5,   7, -53,   0,  -1,   3,  -3,  -2,  -5,   2,  27,   1,   0,  -2,   2,   0,  -5, -30, -32,   4,   2,   5,   3,   3,  26,  20, -11,  -7,   2,  -6,   8,  -6, -13,  25,  -4,  -2,  11,   4,   1,  11,  17,   8,  34,   3, -48,   2, -20, -11, -48,  26, -21 },
      {   0,   0,   0,   2,  -1,   0,  -2,  13,   0,   0,   0,  -3,   0,   4,   3,  -1,   0,  -1,  -1,   3,  -5,  -5,   3,  13,   1,  -1,  -1,  -3,   5,  -1,  -7,  -3,   1,   0,  -2,   4, -10,  -3,   4,  18,  -1,   0,  -1,  -9,  28, -11, -36,   4,  15, -14,   7,  -3, -24,  73,  -6,   4, -23,  37, -42,  18, -40, -22,  17,  18 },
      {   0,   0,   0,  -3,   1,   2,  -2,   7,   0,   0,   0,   0,   1,  -2,  -6, -24,   1,   1,   4,  -1,   0,   2,  15,   7,   0,   0,  -1,   2,   7,  -7, -16,  10,   0,  -1,   3,  -5,  -1,  26,   1,  -3,  -9,   6,  -8,  12, -14, -17,   3,   8,   3,  -7,  13, -32,  46,   2,  -2, -40,   3, -12,  21,  21, -54,  52,  10,  40 },
      {   0,   0,   0,   2,  -1,   0,  16,  -5,   0,   1,  -1,   0,   2,   5, -15,  21,   1,   0,   1,  -1,   0,   0,  27, -21,   1,   0,   0,   0,   2,   4, -24,  45,   0,   2,  -1,   5,  -6,  14,  25, -48,   3,   0,   1,  -2,   9,  -6,  25,  48,  -3,  -1,  -3,  -1, -13,   1,  12,  30,   4, -11, -14,  -5, -10, -23, -37,  41 },
      {  -1,   0,   0,  -2,  -1,  -3,  -1,  14,   0,   0,   1,   2,  -1,  -1,   6,  -4,  -1,   0,   0,   0,   3,  -4,  -6,  46,  -3,  -1,  -3,   2,  -6,  11, -11, -29,   2,  -2,   4,  -7,  10, -24,  51,   8,   0,  -8,   2,  -6,  -7,  -8, -42,  58,  -8,   7, -10,  14,  -3,  -6, -28,  -1,   6,  -2,  10, -12,   1,  27, -41,  -8 },
      {   0,   0,   0,  -1,   2,   0,  -7,   3,   0,   0,   1,   1,  -1,  -1,  -3,  -1,   0,   0,   0,   1,   0,   4,   0, -15,  -1,   1,   1,   1,   2,  -4,  -2,  17,  -1,   2,   1,   2,   4,  12,  -7, -25,   5,  -3,   4,  -5,   4, -12,  36,  20,  -2,  -8,  -2,  -1,  -7,   8, -78,  26,  -7,  18,  -2,  18, -12,  32,  22, -60 },
      {   0,   0,   0,  -1,  -2,  -2,  -5,   7,  -1,   0,   0,   0,  -1,  -1,   3,  -4,   0,  -2,   0,   0,  -1,   2,  -7,  19,   0,   0,  -1,  -1,  -1,  -3,  11,  -4,   0,   1,  -2,   1,  -4,   0, -15,  31,   0,   1,   3,   2,  -1,   1,  17, -39,  -3,  -1,   6, -12,  10, -16,   6,  70,  -4,   4, -18,  17, -37,  21, -69,   0 },
    },
  },
  { // 3
    {
      { 107, -43,  -9,  -4,  -2,  -1,  -1,   0, -49,   4,  12,   2,   3,   0,   1,   0, -12,  15,   1,  -2,   0,  -1,   0,   0,  -1,   1,  -5,   0,   0,   0,   0,   0,  -2,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  47,  37, -26,   1,  -8,  -1,  -3,  -1,  76, -56, -12,   1,  -1,   2,  -1,   1, -42,  -9,  24,   0,   5,  -1,   1,   0,  -5,  17,   0,  -4,  -1,  -2,   0,   0,  -9,  -3,  -2,   0,   0,   1,   0,   1,  -4,   0,   2,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   1,  -1,   0,   0,   0,   0 },
      {  17, -66,  46,  -5,   3,  -2,   2,  -1,  65,  38, -36,  -8,  -4,  -3,  -2,  -1,  19, -35,  -7,   5,  -2,   4,   0,   1, -16,   2,  10,  -1,   2,   0,   1,   0, -12,   1,   0,   3,   1,   0,   0,   0, -12,   1,   3,   0,   0,   0,   0,  -1,  -2,   0,  -2,  -1,   1,   0,   0,   1,  -4,   0,   1,   0,   0,   0,   1,   0 },
      {  32,  72,  -3, -31,  -3,  -2,  -2,  -1, -20,  34, -35,  -6,   0,   0,   2,   0,  17, -55, -25,  23,   3,   6,   0,   1, -23, -19,  24,  10,   4,   1,  -1,   1,  11,   6,   3,  -3,   0,  -2,   0,   0,   6,   2,   0,  -1,   0,  -1,  -1,  -1,  -2,   0,   1,   0,   0,  -2,   0,   0,   2,  -2,   0,   2,   0,   0,   0,   0 },
      { -20, -25,  21, -18,   8,   3,   2,   1, -30, -28, -10,  17,   5,   0,   3,   0, -39, -26,  30,  29,  -4,   0,   0,   0, -43,  43,  35, -18,  -5,  -3,  -1,   0,  40,  18, -27, -10,  -1,   0,   3,  -1,  21,   3,  -4,   1,  -1,   2,   0,  -1,   1,   3,   2,  -2,  -1,  -1,  -1,   0,   3,  -1,  -2,   0,   0,   0,   0,   0 },
      {   1,  31,  49,  -2, -12,  -8,  -2,  -2, -11,  54, -20, -10,  -2,  -3,  -1,   0, -65,  17,  16,  -7,  -3,   2,   2,   1,  28,  28, -25, -20,   1,   1,   1,  -1,  -8, -29, -12,  12,   5,   3,   0,   0,  -6, -17,   1,   9,   2,   0,  -1,   0,   6,  -9,   0,   4,  -1,   0,  -1,   0,  -1,  -2,   3,  -2,   0,   0,   0,   0 },
      {  -2,  -2, -10, -29,   6,   6,   0,   1,  32,  31,  46, -18,  -8,  -8,   2,  -3, -37,  29, -40, -15,  -1,   0,   2,  -1, -55, -16, -18,  14,  15,   7,   0,   2,  43,   6,  -7,   9,   0,  -2,   0,  -1,  20,   2,  -3,   1,  -3,   0,   0,  -1,  -5,   4,   7,  -2,  -2,   0,  -1,   0,   5,  -3,  -3,   1,   0,   0,   0,   0 },
      { -15, -26, -59,  24,  -3,   2,   1,   1,   4,  40,   1,  -4,   2,   2,  -2,   1, -58, -20,  -8,  27,  14,   1,   3,  -1,  33, -15,  41,  18, -12,  -3,  -3,  -2,  -3,   7,  19, -24,  -9,   1,   0,  -1,  -5,  -4, -10,  -2,   5,   3,   0,   0,   7,   2,   0,   2,   1,  -1,  -2,  -1,  -1,  -1,   3,  -3,  -1,  -2,   0,   0 },
      {  -3,  -8,  57, -15,  -1,  -3,  -1,   0, -11, -54, -18, -21,   8,   3,   4,   1, -38,   2, -39,   9,  17,  10,   5,   0,  33, -38,  -1,  39,   9,   3,  -2,   0,   2,  13,  22,   2,  -8,  -4,  -2,  -1,   9,  18,   1, -10,  -3,  -1,   0,   0,  -5,   4,  -4,   0,   3,  -2,   1,   1,   2,   3,   2,   0,  -2,   0,   0,  -1 },
      { -10, -17, -40, -48,  15,   8,   5,   2,   4,   3, -73,  21,  20,   5,   3,   0,   4,  48,   8,  27,  -2,  -4,  -6,  -2,   7, -15, -20,   0,   1,  -1,  -3,   0,  13, -21,   2,  19,   1,  -3,   0,   1,  15,   0,  17,   3,  -5,  -3,  -1,   1,  -5,   2,  10,  -2,  -4,  -2,   0,   0,   2,   1,   0,  -1,  -2,   0,   1,   0 },
      { -13, -15, -31,  32,   5,   2,   1,   0, -25,  -9, -32,  -4, -15,   8,   2,   4, -17, -47, -14, -44,  17,  13,   6,   3, -20,  24, -35,  22,  33,  -1,   3,  -4,   2,  -5,   6,  43,  -1,  -7,  -4,   0,  -9,  -3,  26,   4, -12,  -5,  -2,  -1,   1,   2,   7,  -6,  -7,   4,   1,   0,   1,   4,  -2,  -1,   1,   2,   1,  -1 },
      {  -6,   4,  13,   6,   5,   0,   4,   0,  -6,   6,   4,  60,   3,  -6,  -4,  -4, -31,  -5,  27, -16, -40,  -3,  -1,   3, -36, -56,  10, -13,  15,   8,   5,   3, -41,   2,  37,  15,   1,   0,   1,  -2,  -1,  12,   4, -12,  -5,   1,  -4,  -3,  -3,   0,   1,   0,  -3,   0,  -1,   1,   3,   3,  -5,  -7,  -1,   2,   0,   0 },
      { -10, -12, -14, -82,  18,  11,   5,   1,  -1,   0,  44, -14,  12,   3,   0,   0,  -4, -43,  25,  -9,   8,  10,  -4,   2,  32,   5, -12,  -5,  18,  -4,  -2,   2, -36,   2,  -3,  16,  13,  -3,   2,   0,   1,   8,   1,   9,   3,  -2,  -2,  -3,   5,   2,  -3,   6,   4,  -6,  -3,  -1,   0,  -1,   0,   2,  -2,  -5,   1,   0 },
      {   4,   6,  13,  -9,  21,   0,   1,  -4,   5,   4,  29,  37,  16,  -7,  -6,  -3,   5, -11, -10,  18,   0, -14,  -3,  -1,  -2,  42,  -5,  46, -24, -11,  -3,   1,  12, -58,  38,   5, -34,   5,  -3,   3, -13,   9,  16, -24,  -5,   6,   1,   0,  10,  -1,  -3,  -6,  -2,   0,   2,   1,  -3,  -1,   0, -12,  -1,   0,   1,   1 },
      {  -9,  -5,  -2,   8, -12,   2,   5,   2, -14, -18,   6, -48,  38,   3,  -3,   2,  -7,   6, -17,  25,   5, -10,  -1,  -1, -49, -10,  18,  -4, -19,   4,   2,   2, -46, -48,  -8,  14,  19,  11,   4,   1, -14,  -7,  16,  31,   6,  -2,  -2,   1,  16,   8,   6,   7,  -9,  -4,  -3,   0,  -2,   1,   3, -11,  -3,   0,  -1,   0 },
      {  -4,   8,   2,  -9, -20,   3,  -1,   0,  -5,  20, -14, -27,  20,   9,   0,   1,   6,  37,  34, -10,  32,  -1,  -2,  -2, -34,  34,  11,  30,  14, -16,  -6,  -3, -35,  42,  34,  -4, -21, -11,  -2,   0,   0,  18, -14, -23, -10,  -1,   0,   3,   6,  -3, -16,   3,   4,   4,   5,   3,   4,  -4,  -1,   4,  -4,   2,   2,   1 },
      {  -3,  -9,   2,   2, -53,  -8,  -3,   1,  -2,   4,  10, -19, -16,  10,   5,   1,   6, -21,  66,  19,  25,  -5,   3,  -8, -10, -45, -38,  13,  -7,  -1,   4,   1,  39, -26,   6,  -5,  -8,  10,   1,   1,   1,  12,   7,  -6,   4,  -1,  -2,  -2,   0,   3,  -3,  -3,   5,  -4,  -1,  -1,  -5,   7,   6,  -3,  -5,  -3,   0,  -1 },
      {   7,   8,   6,  37,  15,   3, -10,   2,   9,  16,  20,  -1,  40, -11,  -2,  -4,   1, -11,   9,  40,   6, -17,  -3,  -2,  15, -10, -17,  -2,   0, -21,  -2,   0, -10,  26, -32,  36, -13, -20,   9,   3,  37,  13,  23,   4, -43,  -4,   6,   0, -13,  11,   1, -19, -12,  10,   6,   2,  11,  -1,  -6,  -1,   0,   6,   0,   1 },
      {   6,   9,   8,  11,  76,  18,   6,  -6,  -3,   5,  -7, -27, -35,   0,  -3,   2,  -3,  13,  40, -22,  11,   9, -18,   2,  -5, -25,  28,  33, -13, -18,  -6,   3,  -2, -18, -27,  -8,   2,   4,   0,   4,   4,   4,  17,  -1,   7,  -4,   2,  -2,  -5,  -1,  -8, -13,   3,  -6,   4,   2, -11,   3,  -6,   3,   6,  -3,   3,  -2 },
      {   2,   3,  12,  -6, -27,   0,  -4,  -2,   7,   0,  19,  26,  -6,  16,   4,   4,  10,  21,  10,   5,  24,  28,  -3,  -1,  11,  -7,  57,  13,  42,  -4, -13,  -6,  17, -10, -10,  33,  -4, -28,  -6,  -2, -29, -39,  15,  21, -16,  -2,   4,   1,   0,   9,  16,  -4, -10,  -3,  -2,  -1, -10,  -1,  -3, -13,  -2,  -2,  -2,   1 },
      {   3,   2,  -3,  16,  19,   7,  -2,   0,   2,   0,   7, -37,   6,   2,  10,   5,   3,   2,  14,  26, -33,  18,  -2,  -3,   8,  13,  11, -22,  21,  34,  -7,  -5,  30, -15,  49,  16,  39, -19, -21,  -6,  -9,  15,   0, -16, -15, -26,  -5,   2,  -2,   7, -31, -15,   6,   6,   7,  -2,  -1,  -8,   0,  17,   7,  -3,  -3,   1 },
      {   1,   3,   4,   5,  32,   2,   8,   2,  -2,  -5,  -1, -15,  44,   4,  -3,   0,  -1, -11,  20,   8, -26, -13,  -6,   0,  -7, -13, -28,  24,   2,   4,   7,   3,  18,  28,   7,   2, -11,  -9,  -3,  -2, -39, -70, -30,   4,   6,  10,   0,   4,  -7, -14,  16,  16,  10,   4,  -3,  -3,  -9,   5,   9,  -1,   0,  -4,  -4,  -2 },
      {   1,   2,   8,  -3,   1,  -6,   5,  -6,   2,   4,   7,  -1, -28,  -1,   7,   2,   6,   4,  12,  17,  -2,   0,   8,   1,   3,  19,  -6,  26, -22,  29,   9,  -8, -14,  31,  33,  -4,  21,  19, -22,  -5,  31, -16,  19,  32,   3, -24, -15,  -4,  -5, -11,  51, -24, -29,  -7, -10,  -2,   0,  17,  -7, -22,   2,  -5,  -1,   4 },
      {  -2,   4,   6,   4,  39, -24,   0,  -7,  -2,   1,   8,  29, -25,  21,   5,   5,  -6,   7, -12,  45,  53,   9,  -6,  -8, -20,  -4, -36, -21,   6, -17,  -4,  -3, -11,  10,  12, -16,  17,  -7,   2,  -2, -42,  12, -14,  23,  -1, -13,   1,  -3,  -2,   6,  -9,   7, -11,   5,   0,  -3,   4,  -1,   8,  15,   0,   1,  -4,   0 },
      {  -8,  -1, -10, -13, -12,  -9,   6,  -1,  -9,  -5,   4, -20, -42, -18,   8,   3,  -5,   8,  -7,  43, -38,  21,   2,   2,  -8,   9, -13,   2,  10,  -1,   0,  -2, -31,  10, -18,  -4, -34,   5,   7,  -4, -33, -10,  30, -29,  -4,  24,  -2,  -1, -32,  12,   4, -28,  22,   3,   2,  -2,  -7,   6, -10,   6,   9,   1,  -1,  -2 },
      {  -3,  -3,  -6, -22,  12, -58, -10,  -2,  -3,  -7,  -6, -25, -22, -18,   5,   7,   1,   1,   7, -20,  -5, -32,  17,   5,   5,  -6,  22, -16, -20, -17,  19,   0,  12,  -4,  24,  19, -12,  -6,  23,  -2, -12,   2, -20,  17, -39,  12,  18, -10,   8,   4,   9, -13, -12,  37,  -1,  -6,   4,  -1,   1,  -7,  13,   5,  -6,  -1 },
      {  -1,   7,   0,  -8,  -5,  -2,   3,  -6,  -1,  10,  -2,   5,  39,  -5,   1,   0,  -7,   5,   0, -26,   4, -20,   1,   1,   8,   0,   6,  -2, -20,   9,   3,   3,  26,  35, -11,  -7,  18,  16,   0,   2, -67,  41,  46,   2,   8, -10,  -5,  -9, -13,  21,   5, -19,  -5,  -7,  -6,   0, -14,  10, -11,  -7,   0,  -2,   4,  -1 },
      {  -4,  -3,  -7,  -9, -14, -37,  14, -12,   0,   0,   3,  22,  -1,   8,   1,   5,  -5,   0,  -6,   0,  11,  -1,  -9,   3,  -6,  -4,   4,  29, -35,  31,  12,   0, -12,   7, -34,  28,  28,   4, -24,  -7,   1, -11,  -7, -18, -25, -11,   2,   7,   3, -41, -40, -11,  26,   9,   9,  10, -22,   4,   8,  19,   9,  -4,   0,  -1 },
      {  -2,  -2,  -5,   3,  -5, -24,   6,  -1,  -3,   2,  -5,   6,   9, -20,  -9,   0,   2,  -1,  10,   2, -20,  -7,  15,  -4,   2,   9,  -2,  46,  33,  19,  -6, -12, -11, -27, -33, -26,  16, -22,   0,   0,  -6,  36, -46,  14,   0, -17,   7,  -1, -43,  12,  16,  10, -18,   5,   9,   1,  -1,   9,  -4,  -4,   3,  15,   2,   1 },
      {   4,   4,  -4,  10,  14, -20,   7,   1,   3,   2,  -1,  -1, -12, -21,   4,   3,   7,  10,   9,  22, -21,  13,  17,  -5,   6,   2,   2,  20,  18,  10,  20,  -7,   8,  21, -17,  39, -13,  24,   5, -15,  -9,  29,  -3,  12,  24,  14, -17,  -8,  54,   0, -22,  32, -11, -22, -12,   0,   2, -13,  15, -19, -22,   1,  -2,   1 },
      {   0,   1,   2,  -3, -17,  71,  16,   8,  -3,  -3,  -1,  -4, -35,  12,   0,  -6,  -4,   0,  -7,  14, -19, -29, -19,  -1,   4,   2,   1,   4,  -8,  -9,   5,  11,   2,   6,  -5,  14,   1, -12,  -4,   1, -30,  32, -15,  14, -27,  -4,  13,  -8,   3, -44,  11,  11, -11,  17,   2,   3,  -4,  -4,  23, -14,  -1,   3,  -2,   2 },
      {   0,   0,  -3,   1,  -5, -32,  30,  -4,   3,   3,   4, -12,   1,  74, -21,  -3,  -2,  -2,  -8,  -4, -33,  16, -30,  -9,   4,   3,   0,  -7, -17, -29, -14,  -3,   8,   5,   3,  21, -12,   3,   6,  -4,   7,  14, -16,   0,  11,  -5, -23,   0, -20,   8,  -6,   2, -17, -12,  10,  13,  -4,   6, -14, -28,  12,  12,   6,   0 },
      {   2,   2,   1,   4,   2,   2,  -8,   1,   3,   5,  -2,   7,  17,  31,  10,   4,   1,   1,   4,  -6, -20,  51,  14,  -5,  -4,  -2, -16,  17, -16,  10, -27,  -2,  -2,   1, -11, -24,  10, -19,  11,   5,   0,  23,  -3,  16, -18,  33,  14, -10,  49,   1,  21, -21,  21,  29, -22,  -5,   4,   1,   2,  -6,  22, -19, -11,   1 },
      {  -2,  -2,   0,  -3, -21,  20, -32,   9,   0,  -2,  -3,  14,   1, -19,  37,   7,  -2,  -1,  -1,   0, -19,  18, -14,  -3,  -1,   1,  -8,  29, -17, -42,   3,   8,  -2,   6,   4,  12,  31,  16,  31,  -1,  -1, -18, -22,  10,  20, -18,  -4, -13,   0,  20, -27, -22, -27,   4,   9,  -5, -16, -25,  -6,  14,  24,  12,  -5,  -5 },
      {  -1,  -3,  -1, -13,  -7,  -5, -11,  -4,  -2,  -6,  -5,  -3, -15,  -4, -27,  -7,  -4,  -5,  -2,   0, -20, -19, -14,   2,  -1,  -8, -16,   4,   3,   5, -11,  -1,  -3,   9, -12, -34, -11, -31, -22,   8, -10, -18,  14,   1, -11, -29,   7,  27,  66,  20, -20, -12, -30,  -7,  28,   8,   1,   0, -22,  -8,  -1,  18,  11,  -1 },
      {   3,   4,   8,   0,  -2,  19,  26,  13,   2,   5,  -1,  10, -20,  44,  -6,   8,   1,   2,   2,   5,  -3, -47,  16,   1,   6,   8,   3,  24,   3,  17,  27,   0,  -5,   3,  -1,   8,  19,   5,   8,  -2,   7, -13,   2,   9,   3,  21,  14,   5,   1,  64, -19,   3,  19,  12, -11,  -7,  25,  -5, -19,  20,  -1,  -5,  -9,  -1 },
      {   2,   4,   2,   3,   5,  21,  45,   7,  -1,   0,  -5,   2,   4, -18,  35,  -3,   0,  -3,  -5,  -4,  26,  -8,   5,  -9,   0,  -4,  -6, -23,   0,  12,  -1,  -7,  -5,  -7,  -4,  11, -22,  -4, -18, -15,   2,   4, -50,  -4,  -2,   7, -16,   2,  18,  29,   1, -43,   3, -14,  -3,   4, -44,  16, -16, -23,  14,  -4,   4,   0 },
      {  -1,   0,   0,   6,  -7, -12, -24, -12,  -2,   0,  -3,   0, -16, -23, -41,  -3,  -1,   2,  -4,  11,   1, -17, -43,   6,   0,   9,  -2,   6,   3, -11, -20,  22,   1,   8,  11,  26,  37, -15,  12,  15,   7,  11, -18,  -8,  11,  19,   3,   7,  13,  26,  30,  -4,  36, -22,  -2,   0, -28,  21,  13,  -1,  -8, -25,   7,  -4 },
      {   1,  -1,   1,   7,  -3, -20,  30, -11,   2,  -2,   1,  -2,  10,   9,  10,   0,   1,   0,  -4,   7,  -4, -37, -23,  12,   4,   2,   1,   6,  62, -16,   0,   8,   2,  -7,   3, -35,  16,  21,  26,  -3,  12,  -2,  18,   7,  -9,  11, -16, -14,  18, -38,  -2, -24,   3,   0, -17,  -4, -24, -27, -12,   7,  12,  -4,   8,   9 },
      {   0,   0,   0,   1,  -5,   7,  11,  18,  -2,  -3,  -3,   2,  -3, -22, -46,  -8,   0,   0,   4,   3,   6,  14,  14,   1,  -3,   2,  -7,   1, -16,  16,  -8,   7,   6,   6,  21,   5, -15, -25,  14,  11,   2,   2,   3,  59,   9,  21,   9,  -6, -22, -24, -40, -10,   7, -18, -14,   3, -18, -18, -43,   7,   5,   3,  14,   7 },
      {  -2,   0,  -1,  -3,   9,  16, -52, -19,  -2,   1,   0,   0,   4,  33, -31, -18,   0,   3,  -3,   4,  -1,   6,  17,  16,  -2,   3,  -2,  -6,  14,  17,  40,   0,  -5,  -1,  -2, -10, -21,  20,   3, -12,  11,   4,  -2,   4, -12,  -7,   3,  -2,  -8,  15,  -9,   1,  -8,  12,  -2,  -4, -71,   0,  19,  -8,   7,  -2, -10,   6 },
      {   0,  -2,   0,  -3,   4,  18, -30,   4,  -1,  -3,  -2,  -2,   0,   2, -51,  -8,   2,   2,   1,   7,  16,  -2,  23,  -8,   2,  -5,   3,   0,   9,   2,   4,  -8,   2,  -7, -12,  11,  17,  26,  15, -19, -21,  -7, -29, -30,  -8,   4, -36,  -8,   9, -19,   8, -36,  -8,  -5,  -9,  21,  47,   7, -25,  -4,  18,   0,   8,  10 },
      {   3,   3,   1,   1,  -3,   9,  57,  -4,   1,   1,   3,   0,   4, -17, -34,   6,   3,   4,   1,   8,  10,  28,  33,  15,   1,   1,  -4,  -4,   2, -14,   9,   8,   8,   6,   8,  -5,  35,   2,  30,   3,   8, -12,  13, -32,  -1,   9,  30,  -6,  10,   0,  -3,  13, -27,  14,  21,   2, -18,  34,  11, -13,  13,  20,  -6, -12 },
      {   1,   0,  -1,   1,  -3,   4,   5,   0,   1,   2,  -3,   4,   8, -16,   0,   8,   1,   2,   1,  -6,   4,  28, -50,  -1,   2,   6,   0,   3,  12,  19,  32,  34,   5,  -3,   6, -21, -12,  26,  15,  -6,   0,  -4,  -4,  37, -32,   1, -10,  -7,  -5,  15, -23,  -8,  19,   9,  27,  20,  34,  34,  14, -24,  -9,   0,   2,  -8 },
      {   0,   0,  -1,   0,   0,   1, -10, -11,   1,   2,   2,   1,   3,  18,  20, -10,   0,   3,  -6,   1, -21,  -7,  46,   4,  -1,   4,   4,   4,   6, -44,  27,  16,   1,  -6,   1,  -5,   4, -31,  -1,  -5,   1,   3,  -1,  19,  10, -11,  -4,   1,   5, -22, -16, -27,  20, -30,   1,  -4,   5,  64,   2,  13, -30,  -6,   6,  -7 },
      {  -1,  -2,  -3,  -1,  -4,  15,  -1,-101,  -1,  -2,  -3,  -2,  -2,  -7,  10, -22,   1,  -2,   0,   0,  -1, -11,   4, -10,   0,  -1,  -1,  -1,   4,   7, -36,  -9,   0,   4,   3,   8,   0,  13,  -3,  15,   1,  -8,  -6,  10,   8,  22,  10,  -3, -10,   2, -29,  -2, -12,  14,  -8,   2,  16,  10,  -5, -14,  -4,   2, -17,   0 },
      {   0,  -1,  -1,   5,   4,   0, -19, -16,   2,   2,   2,  -3,   2,  12,  25,  22,   1,   0,   0,  13,  12,   3, -18,  31,   2,   1,  -1,  -5,   4,  25,  33,  27,   0,  -3,  -5,  13,  -7, -16,  15,  11,  -3,  12,  -9, -20,  30,  -5,  23,   1,   8, -23,  15,  16,  -1,   6,   9, -12,   5,   9, -73, -19,  11,   2,  -7, -11 },
      {   1,   0,   1,   0,  -6,  -3,   5,  16,   1,  -1,   1,   2,   9, -14, -28, -11,  -1,   1,  -2,   3,   0,  -1, -25, -15,   1,  -2,   5,  -2,  19, -22,  13, -34,   4,   2,   0,  10, -12,  11, -34, -10,   4,  11,   6,  18,  51,  -9,  15,  10,   8, -13,  -4, -19,  12,  51,   2, -25,   8,  26,  -1,   9,  25,  -6, -34,  -5 },
      {  -1,  -1,  -2,   0,  -5,  -4,   4,  -8,  -1,  -3,  -4,  -3,  -2, -18,  -3, -33,   0,   0,  -2,  -3,   3,  13, -14,  -8,  -2,   0,  -7,   3, -11, -29,  40,   2,  -1,  -4,  -3, -18,  30, -39, -30, -30,   3,  -8,   9, -18, -12,   6, -12, -23, -10,  12, -11,  20,   8,   6, -37, -10,   5, -22, -18, -46,  -8,  -8,  -9,  17 },
      {  -1,   0,  -1,  -2,   8,  -5, -25,  45,  -1,   1,  -2,  -1,  -5,   8,   9,  18,  -1,  -2,  -6,   0,  -3, -16,  -2, -29,  -2,   2,  -3,  -1,  11,   2, -34, -21,  -3,   4,  -2,  -6,  15,  27,   3,   8,   2, -12,   0, -12, -18,   5,  38, -16,  -9, -12, -33,   3,  -3, -14,  -7, -18, -12,  30, -17, -51, -24, -13, -15,   1 },
      {   0,   0,  -2,   2,  -3,  -7,  25,  -7,  -1,  -2,  -3,  -2,   3,   5, -15, -25,   0,   1,  -1,  -3,  -4,  19, -12,  -8,  -1,   0,   0,  -2,   0,   9,  19,  13,  -4,  -2,  -7,   5, -19,   9,  14,  29,  -5,  -4, -19, -16,   1, -44,  48, -19,   3,   4,   6, -37, -18, -19, -27, -27,  17, -21,  13,  12, -29, -34,   2,   8 },
      {   0,   0,  -1,   2,   6, -15,  -5,  31,  -1,   0,   0,  -1,  -4,   6,  24, -92,   2,   2,   4,   5,   6,  -4,  14,  -5,   2,   3,   4,   4,  10,  10, -16,  50,  -1,   4,   0,   8,   8,  16,  -8,  17,  -3,  -1,   5,  -1,  10,   6,  12,  12,   0,   0,  -1,   8,   3,   8,  12,   9,   5,   4,   2, -11,  30,  10,  10,   2 },
      {  -1,   0,  -1,  -2,   5,  -4,  -9,   0,   0,   1,   1,   3,   1,   8,  -2,  35,   0,   1,   0,   1,   8, -17,  13,   5,  -2,   0,  -1,  -3,   8,  11,   9,  14,  -4,   4,  -3,   0,   1, -31,  -4,  23,   4,  -3,   7,  -4,  39,   2,   6,  -1, -11,  -4, -17, -36,  11,  -8,   0,   3,  19, -26,  57, -56,  33,  14,   6,   6 },
      {   0,   0,   1,  -2,   2,  -1,  -3,   8,   0,   0,   1,   0,  -5,   3,  12, -22,  -1,  -1,  -1,  -5,   8, -11, -24, -41,   1,   4,   4,   1,  -8,  41,   6, -16,  -5,   0,  -1,  -2,  -6, -48,  64, -11,   4,  -6,  16, -15,  22,   3, -29,  -5,   3,  -8,  -8,  -1, -24,  16,  -9,   0,  -7,  28,  17,  13,  -9,   2,  -9,  12 },
      {   1,   1,   0,   0,  -3,   4,   9,  -2,   0,  -1,  -1,   3,  -7, -10,  -2,  -2,   1,   2,  -2,   4,  -2,  -9,  12,  36,  -2,   0,  -1,   1,  -3,  -2, -30, -10,   1,   3,  -2,   3, -19,   0,  28,   7,   1,  -8,   3,   4,  -6, -72, -21, -25,   8,  11, -10,  32,  42,  10, -28,  -1,   4,  20,  -1, -14,  34, -10,  -1,  13 },
      {   0,  -1,  -1,   0,  -4,  -4,  -5,  16,   0,   1,   0,  -2,   2,  -1,   7,   4,   0,   1,   2,   1,  -1,   0, -22,  72,  -1,  -1,  -2,   2,  -2,   4, -10,   2,   1,  -3,  -4,   0,  -4,  -4, -21,  -4,  -2,   2,  -6,  -4,   1,  32,  -1,   2,  -6,   5, -23,  -1, -52,  -5, -32,  -6,  13,  35,  17,  17,  12, -17,   5,  42 },
      {   0,   0,  -1,  -1,   0,  -4,   5,   9,   0,  -1,  -2,   1,  -2,   3, -12,  -2,   1,   0,   1,  -2,   4,  -8,  14,  36,   0,  -1,  -3,   1,  -4,   8, -41,  40,  -2,   0,  -6,   1,  -4,  -8,   3, -48,   1,  -2,  -2,  -1,  15,  -6, -23, -13,  -6,  -1,  -6, -20,  -8,  27,  26, -21,   1, -22,   2, -12, -44, -24, -34, -33 },
      {   0,   1,   0,   1,   3,   2,  -4,   2,   0,   0,   1,   1,   1,   5,  13, -34,   0,   0,   1,   2,   3,   4, -14,  50,  -1,   0,  -4,   1,  -5,   4,   1, -65,   3,   1,   5,   2,  12,   2,  26, -31,  -3,   1,  -2,   5,  10,   1,  34,  14,  -1,  -2,   4, -19,  23,   1,   8,  15,   8, -12,   7, -19, -27,  23,  38, -14 },
      {   0,   0,   0,   2,  -3,  -1,  -3,  -9,   0,   0,  -1,   0,   2,   0,  -1, -11,   0,  -1,   1,  -4,   0, -10,  -9,  -4,  -1,   2,  -1,   0,  -1,   8,   1, -12,  -2,  -1,   2,  -5,  -8, -10,   4, -46,   1,   2,   7,   5, -13,   7,  23, -19,   0,  -3,   0,  12, -10, -63,  13, -15,  16,   1,  11,  11,  60, -23, -20, -51 },
      {   0,   0,   0,  -1,   1,   1,   5,   5,   0,   0,  -1,   1,  -1,  -1,  -2, -15,   0,   0,   1,  -4,   4,   7,  -5,  21,   0,   0,   1,   3,  -6,   2,   6, -11,   3,  -1,   0,  -2,   7,   3,  16,  30,  -6,   6,  -7,   2, -24,   3, -28,  46,  -2,  -6,   1,  -9,  12, -33,  -8, -48,   5,  -5,   3,  -7,  -4,  45, -68,   1 },
      {   0,   0,   0,   1,   1,  -1,  -7,  -2,   0,   0,   1,  -2,   2,   6,  -8,   6,   1,   0,   1,  -1,   3,  -2, -14,  -7,   0,   1,   1,  -1,  -2,  11, -17,  25,  -2,  -1,   0,   5,  -7,   2, -19, -28,   4,  -4,  -2,  -3,   5,   0,  18, -56,  11,   5,   9,  -7,  14,  -7, -22, -24,  -1,  13,   4,  21,  -8,  88,   8,   7 },
      {   0,   0,   0,   1,   0,   0,   4,  -8,   0,   1,   0,   0,   1,  -4,   3,  14,   0,  -2,   1,  -4,  -3,   1,   4, -15,   0,   1,  -1,   0,   0,  -6, -11,  25,   1,   1,   3,   1,  -2,   6,  18, -60,   0,   3,   4,   4,   6, -17,  41,  48,  -2,  -2,   9,  -1,  10, -14,  -5,  30,   1,  -5,  -2,  -1,   4,   1, -32,  66 },
      {   0,   0,  -1,   0,   0,  -2,  -3,   7,   0,   0,   1,  -1,  -1,  -3,  -1,   9,   0,   0,   0,   0,  -2,   1,  -6,  -3,   1,   1,   0,   0,   5,  -1,  -3,  19,   0,  -2,   0,  -3,  -3,   2,  -1,  -5,   1,   1,   0,  -5,  -1, -12,   7,  39,  -4,  -1,  -7,  -2, -14,  12, -81,  42,  -4,   6,   3,   0,  -3,  15,  -5, -71 },
      {   0,  -1,  -1,   0,   0,  -6,   0,  12,   0,   0,   0,  -2,   1,   2,   4, -10,   0,   0,   0,   0,   0,  -5,  -3,  19,   0,   0,  -1,  -2,   2,  -1,   7, -16,   0,  -1,   2,  -2,  -1,  -9, -10,  27,   0,  -2,   0,  -1,   9,   3,   7, -50,   5,   2,   8,  -7,   2, -14,  10,  81,  -5,  -6,  -3,  12, -13,   4, -67,   3 },
    },
    {
      { -91,  59,  -3,   3,   0,   1,   0,   0,  59, -25, -11,   2,  -2,   0,  -1,   0,  -3, -13,  10,   0,   1,   0,   0,   0,   2,   4,   0,  -2,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  59,   9, -29,   0,  -5,   0,  -2,   0,  50, -70,  16,   5,   0,   1,   0,   0, -51,  18,  25, -11,   4,  -2,   1,  -1,  -2,  16, -13,  -2,   2,  -1,   0,   0,  -6,   0,   0,   3,   0,   0,   0,   0,  -1,   2,  -2,   0,   0,   0,   0,   0,  -3,   0,   1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
      {  34, -21,  13,  -6,   0,  -2,   0,  -1,  56,  21, -53,   2,  -3,   1,  -2,   0,  14, -69,  15,  21,  -2,   4,   0,   1, -30,  12,  33, -13,   0,  -1,   0,   0,  -3,  10,  -9,  -5,   3,  -1,   1,   0,  -6,   2,   1,   2,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,   1,   0,   0,   0,   0,   0 },
      { -38, -83,  55,   7,   7,   1,   3,   1,  32,   3,  14, -16,   1,  -3,   1,  -1, -35,  26,   2,  -7,   4,  -1,   1,  -1,  -5,  30, -14,  -3,   1,  -2,   0,   0,   6,   2, -16,   3,   0,   0,   0,   0,   0,  -1,  -2,   5,  -1,   0,   0,   0,   2,   0,  -1,   2,  -1,   0,   0,   0,   1,   1,   0,   0,  -1,   0,   0,   0 },
      { -12,  12,   9,  -9,   1,  -2,   1,  -1, -57, -10, -12,  27,   2,   4,   0,   2, -12, -20,  71,  -6, -10,   0,  -1,   0,  -2,  54, -10, -40,   7,  -4,   1,  -1,  12,  -3, -31,  10,   6,  -1,   1,   0,   2,  -2,   2,   4,  -3,   0,   0,   0,   4,  -2,  -2,   1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0 },
      {  16,  33,  12, -17,   0,  -5,  -1,  -2,   6,  52, -51, -13,   5,  -3,   1,  -1, -65,  14, -11,  16,   4,   2,   2,   0,  38,   3, -25,   9,  -5,   0,  -2,   0,  14, -35,  -5,  14,  -2,   3,   0,   1,  -2, -12,  12,   4,  -1,   2,   0,   0,   0,  -5,   7,   3,   0,  -1,  -1,   0,   1,   0,   4,  -1,  -1,  -1,   0,   0 },
      {  12,  -2,   7, -17,   4,  -2,   1,   0,  25,  11, -12,  15,  -9,   1,  -3,   0,  54,   7,   8, -49,  -1,   1,   0,   0, -19,   1, -68,  10,  29,  -2,   4,  -1,   1, -29,   9,  44,  -9,  -1,  -1,   0,  -7,  -1,  15,  -6,  -7,   2,   0,   0,  -1,  -2,   3,   2,   1,   0,   0,   0,  -3,   2,   3,  -1,  -1,   0,   0,   0 },
      {   9,  29, -27,  15,  -5,   1,  -3,   0,   9,  37,  42, -43,  -7,  -3,   0,  -2,   1,  -9, -22, -22,  20,   0,   3,  -1, -33,  33,  -2,  -3,   8,  -7,   1,  -1,  13,  14, -55,   3,   7,   0,   2,   0,  -3, -28,  -5,  34,  -1,   1,   0,   0,   5,  -3,  11,   4,  -7,   0,  -1,   0,   0,  -4,   1,   1,  -1,   1,   0,   0 },
      {  10,  -3,   2,   7, -10,   1,  -2,   1,  25,   2,  -2,  -8,   5,  -3,   0,  -1,  37,  34,   5,  13, -31,   0,  -4,   0,  48, -22,   0, -64,  -2,  12,  -2,   2, -28, -10, -37,   1,  40,  -3,   4,   0,  -2,  -6,   3,  24,  -4,  -6,   1,  -1,  -6,   0,   2,  -1,  -2,   2,   1,   0,   0,  -1,   1,   3,   0,   0,   0,   0 },
      {  19,  41,  80, -50,  -4,  -6,  -1,  -3, -11, -37,   0, -24,  20,  -1,   2,   1,  15,  11, -28,  16,   4,  -6,   1,   0, -26,  11,   0,   1,  -4,   1,   2,   0, -13,  14,  -9,   4,   3,  -1,   0,  -1,   6,   3,  -7,   4,  -4,   0,  -1,   0,  -1,  -2,  -4,   1,   0,   2,   0,   0,   1,  -2,  -1,   2,   0,   0,   0,   0 },
      {  -8, -13, -15,  27,  -7,   4,  -2,   2, -15, -27,  -1, -27,  20,   3,   3,   0,  -6, -25, -19,  36,   8,  -3,   1,   0,  -7,   1,  -3,   1,   8,   3,   1,   0,  -9, -12,  -3,  59,   6, -10,  -2,  -2,  13,   4,  58,   1, -41,  -3,  -2,   0,   0,   9,  -2, -34,  -1,   7,   1,   0,  -1,   0,  -1,   0,   4,   0,   0,   0 },
      {  -7,   7,  -4,  -9,  -1,   1,   2,   1,  -2,  29,  -6,  10,   5,  -7,   1,  -2, -42,  20,   7, -16, -28,  11,  -1,   2, -46, -43,  11, -22,  43,  12,   1,   2, -17,  45,   9,  28,  12, -24,   0,  -4,  18,  10, -13,  -4, -22,  -1,   4,   0,   2,  -6,  -7,   2,   3,   6,  -1,  -1,   4,  -3,   1,   2,  -2,   0,  -1,   1 },
      {  11,  17,  37,  11, -11,   1,  -3,   0,   2,  20,  49, -28, -20,   3,  -5,   1, -18, -52,  17, -31,  11,   5,   1,   2,  42,  -9,  -3, -10,  15,  -3,   1,   0, -30,  -7,  19,   0,   5,  -5,   2,  -1,  -5,  33,   2, -18,   2,  -4,   2,  -1,  -5,   7, -24,  -7,   7,   1,   3,  -1,  -2,   2,  -6,   8,   4,   0,   1,  -1 },
      {  -9, -20,  -5, -25,  11,   1,   2,   0, -15, -34, -20,  -1,   7,   3,   3,   1,  -7, -36, -13, -19,  23,  16,   1,   2,  44, -26,  -4,  19,  38, -13,  -4,  -1,  -7,  24, -18,  10,   4, -18,   6,  -2, -14, -12, -18,  41,   1,  -1,   3,  -2,   2,   2,  30,  24, -20,  -6,  -4,  -1,   3,   4,   6,  -9, -10,   3,   0,   0 },
      {  -6,  -3,   1,  -7,  16,  -5,   3,  -1, -10,  -3,  -3,  10, -21,  11,  -2,   2, -18,  -9,  -5, -12,  14,  -4,   0,  -1, -39, -18, -11,  -2, -22,  21,   5,   2, -11, -33,  13, -15,  57,  20,  -9,   2,   0,  29,  12,  44,  16, -39,  -3,  -2,   8,   7,   9,  -7, -32,  -2,   6,  -1,  -1,   3,  -5,  -4,   3,   4,   0,  -1 },
      {  -9, -11, -38, -10,   6,   4,   1,   1,  -6,   9, -28, -53,  30,   3,   4,   1,  12,  13,  24,   9,  30, -17,   3,  -3,  -3,  19, -19,   5, -10,  -8,   5,  -3, -45,   8,  10,   3,   4,   4,   0,   0,   7,  38, -29,  -4,   4,  -7,   1,  -1,   1, -15, -27,  27,   0,   3,   2,   0,   4,  -5,  11,  14,  -9,   0,  -3,   0 },
      {  -1,  -3,   2,  16, -28,   2,  -4,   2,  -6, -18,   0, -39,  30,   0,   4,   0,  -1, -18,   4,   3, -19,  -2,   9,  -1, -14, -22, -14, -13,  10,  32,  -5,   2,  66, -18,  15,   0,  23,  -9, -11,   1, -31,  -2, -16, -13,  14,  -6,   8,   0,   0,   1,  -4,  37,  15,  -8,  -1,  -3,   0,   6,  10,   4, -15,  -6,   0,  -1 },
      {   9,  13,  33,  83, -33,  -1,  -8,   0, -14, -16, -48,  -9, -25,  16,   1,   4,   4,  21,   8, -23,  27,   4,  -3,  -2, -11, -13,   6,   4,   2, -15,   1,  -1, -10,   5,  -2,  -2,  -4,   5,   4,   0,  11,  -5,  -9,   6,  -1,   4,  -3,  -1,   1, -12,   9,   4,  -2,   0,  -2,   1,   1,   1,   7,  -5,   1,  -1,  -1,   1 },
      {  -8,  -5,  -1,   2,  -7,   9,  -1,   1, -11,  -6,   6,  -2,  13, -14,   5,  -3, -17, -16,  -4,   4, -20,   8,   0,   3, -29, -24, -12, -12,  -1, -10,   7,   1, -52, -22, -18,   1, -25,  37,  16,   2,  -1, -17,  22,  -4,  50,  31, -16,   0,   0,  16,  11,  25,  17, -29,  -7,  -2,   4,   4,   6,  -1, -13,  -3,   2,   0 },
      {  -3,  -3, -16,  -4,  -1,   6,   0,   0,  -8, -16, -26,   2, -24,  -1,  -1,   1,  -1,  10, -59, -21,  -7,  36,   1,   6,   4,  49,  27, -21,  39,   5, -11,  -1,  -4, -24,   4,  -4,   9, -12,  -2,   1,   7,  15,   7, -15,  23,  -7,   1,   0,  -1,   6, -33,  10,   6, -13,   3,  -3,  -2, -10,   0,  16,  -4,   0,   2,   1 },
      {  -2,  -3,  -3, -32,  14,   4,   0,   1,   0,  -2,  -9, -37,  -3,  11,   4,   2,   6,  32,  41, -17,  32,  11, -12,  -2, -10, -35,  43,  -4,  13, -31,  -3,   2,  28, -18, -14,  -5,  -8,   1,  10,   0, -15,  -3,  32,  -6,   8,  -4,  -4,   0,  -6,  30, -13, -19,   4,  -9,   8,  -2,  -3,   2, -16,   9,   4,   1,   2,  -3 },
      {   4,  13,  11,  26,  11,  -1,  -2,   0,   3,  13,  18,  22,  25, -11,   1,  -2,   3,  18,   7,  25,   1,   3,  -1,  -4, -15,  15,  19,   9,  19, -20, -20,   2, -21, -35,   9,  -2, -10, -39,  17,   5, -57,  26,  -2,  15, -11,   4,  20,  -6,  13,  15,   9,  20, -22,   3,  -5,  -4,   1,   1,   4, -14, -15,   7,  -2,   1 },
      {  -9, -10, -11, -18,   2, -14,   6,  -1, -12, -12, -12, -12, -29,  15,  -5,   4, -11,  -7, -11, -17,   2, -12,   9,  -1, -15,  -8,   0, -10, -18,  13,  -3,   2, -22, -28,  -9, -19,   2, -20,  13,   6, -31,  -7, -16,  -2, -31,  40,  28,  -3,   1,  -7,  14, -12,  44,  37, -19,  -4,  -3,   9,   1,  15,  19, -19,  -8,   1 },
      {   4,   5,   5,  21,  65, -26,   0,  -6,  -2, -12,   5,  10, -31,  -8,   9,  -1,  -5,  -8, -16,  28,   6,   7, -15,   2,  -4, -26, -21, -18,  -6, -28,  11,   4,  24,   6, -26,  14,  -5,  13,  10,  -3, -13,  14,  -9, -10,  -5,  -4,   0,  -1, -10, -15, -33,  22,   5,  10,   9,   0,   5,  -6,  22,  23,  -6,   4,  -6,  -2 },
      {  -2,   5,  -1,  14,   6, -15,  -1,  -1,  -5,   4, -20,   9,  10,   8,   4,   0,  -6,   8, -13, -15,  -1, -17,   2,   0,   6,  15, -15,   2,  -7,  12,  16,  -2, -19,  54,   0,  -3,  12,  10, -17,  -4, -70, -15,  22, -16,   4, -17,  -1,   7, -11,  48,  -9,  -6,   9,  -4,  11,  -3,   4,   2, -20,  11,  -1,  -1,  -1,  -2 },
      {   7,   9,   9,  25,  70, -22,   7,  -5,  -3,  -6,  -9, -30,  11, -33,   7,  -1,   8,   2,  20, -30, -33,  10,  25,   1,   9,   8,  27,  28,  -4,  38,  -2, -12,  -8,  -8,  -1,   6,  -1,   2, -18,   0,  16,  -3,   4,   6,   0,   7,  -1,   2,  -2,   1,  10,  -2,   5,  -4, -13,  -3,  -2,   4,  -3,  -5,  -3,  -7,   0,   3 },
      {   0,  -2,  -7,  -2, -20,   3,  -3,   0,  -2,  -7, -17, -12,   5, -32,   6,   1,  -2,  -2, -14, -25, -34,  -2,  10,  -1,   6,  -3,  -5,  -4, -13, -10,  14,   4,  26,  16, -14,  -4, -17,  22,  31,  -6, -23,  59,   3,  13,  -4,  26, -14, -12,  30,  -3, -11, -24, -25,   8, -12,   4,  -3, -15, -10, -16,  18,  16,  -1,   3 },
      {   2,   0,   6,   8,   1,  16,  -3,   4,   4,   1,  -2,  41,  56,  -8, -19,  -1,  -8,  -9, -20, -45,  23, -50,   8,   5,   5, -13,  32,  -5, -13,  -4,  22,  -7,   6, -26, -26,  12,  -1,   0,  -7,  -3,  11,   2,  -6,   0,  -6,  -4,   5,   3,  -9,  -8, -21,   8,   8,   4,   6,  -1,  -3,  -6,   4,  22,  -2,   0,   2,   0 },
      {  -4,  -5,  -2,  -4, -22,   6,  -7,   5,  -8,  -8,  -7, -12, -10, -41,  10,   0,  -4,  -9,   7,  -5, -26,   8,   5,   2,  -9,  -9,  -7,  17,  -6,  -9,  -2,  -2, -17, -19, -17, -15, -17,  -8, -12,   9,   2, -31,  -9, -13,  -6, -34,  27,  12, -26,   9, -18,   2, -32,  38,  37, -10,   3,  -9,   6,  -2,  18,  35, -15,  -6 },
      {   4,   3,   1,  20,  35,  15,  -7,   2,  -1,  -2, -14,  -9,  32,  -7, -16,   2,  -2, -13, -12,  -4,  -3, -11,  -8,  -1,  -8,  -1, -45, -23,   9, -22, -20,  15,   8,   7,  15, -48,   1, -29,  27,  11,  34,  -9,  -1,   4,   5,   5,  15,  -8, -11,  20,   2, -24,   1, -10,   3,  -3,  -4,   5, -32,   7,   9,   4,   9,  -1 },
      {  -2,  -1,  -4,  -4,  19,  -7,  -2,  -2,  -4,  -3,   1, -24,   3,   1,  -1,   0,  -5,  -3, -14, -20,  -9, -14, -17,   5,   0,  16,  22, -49,  -8, -23,  24,  10,  -4,  -9,  55,  19, -11,  14,  12, -13, -19, -30, -16,  -6, -12,  -9, -14,   3,  -6, -24,  29,   0, -18,  -2,  20,   8,  -9,  17,   7, -24,   1,  15,   7,  -8 },
      {   2,   0,  -1,  -3,  -3, -16,   5,   6,   2,   1,  -1,  -9, -28,   9,  -8,   3,   2,   3,  -3,   5,  -1, -39,   4,  10,   6,   4,   6, -17,   1,   3,  35,  -4,  13,   2,  15,  11, -16,  -3,   3,   1,  35,   4,   4,  11,   7,  13,  35, -11,  -3,  61,   0,  28, -15,  17, -21, -24,  14, -11,  -1, -23, -25,   3, -19,  12 },
      {   0,  -2,   0,  -1,  11,  -7,  -3,  -3,  -1,  -8,  -3, -13, -15,  30, -35,   0,  -4,  -4,   2,   0, -22, -30, -12,   7,   1, -13,  -4,   5,  -1,   4,  -5,  -4, -12,  -5, -16,  -6, -16, -22, -18,  11,  -6, -20,   2, -13,   2, -11,   5,  14,  73, -14, -20,  -5, -11, -24,   7,   6, -11, -36,  -2, -12, -14,  11,  24,   2 },
      {   2,   0,   1,   4,   0, -31,   2,  -5,   7,   3,   5,   6,  44,  19,  38,  -8,   2,  -5,  -6, -19,   7,  37, -28,   0,   5,   1,   0, -10, -10, -11,  -3,   8,  -1,   2,   6,  -6,   6,  14,  -3,   6,  17,  -9,  14, -13,   7,  -8,  22,   5,  58,   2,   4,  11,   4,  33,   0, -18,  15,  -6,   0,  -6,  20, -14, -29,   0 },
      {   5,   7,   8,   3,  30,  92, -17,  15,   0,  -1,  -2,  -9,  -7,  18, -15,   2,   1,   4,   3,   5, -20,  -3, -11,   9,   3,   6,   2,  15,  14,  -2,  22,   1,   2,   7,   1,  -3,  13,  14,   0,  -6, -11,   2,  11,  -5,   7,  -3,   6,   2,  14,  -5,  16,   5,   8,  27,  -2, -10,  16,  -2,  23,   1,  16,  -9, -23,   0 },
      {  -1,   0,  -5,   3,  -1, -21,  11,  -5,   0,   2,   0,  -8,  15,  57, -11,  -1,  -3,  -2,  -8, -11, -42,   0, -33,   8,  -3,   4,   5,  29,  -4, -15,   0,   9,   0,  -9, -38,   3,  12,  -3,   4,   1,   8,  36, -13, -31,   3,  -2,   0,   4, -44,   5,  26,  -3,  -5,  -3,   9,   2,   8,  10,   0, -23,  -3,   7,   4,   2 },
      {   0,   0,   1,  10,   6,   4,   6,   0,   0,   0,   1,   3,  15,  22,   7,  -8,  -1,  -2,   1,  20,  -2,  16,  -6,  -5,  -3,  -3,  -5,   4,  25,   0,  25, -13,  -4, -27,   2,  11,   0,  17, -17, -23,  -4, -20, -63,  10,   2,  11, -20,  -3,   1,  35, -22, -27,  -9,   7, -22,   8, -34, -21, -17,   2,  29,  -1,  -4,  15 },
      {   2,   0,  -2,   1,  -4,  -2, -12,   4,   1,  -1,  -3,   2, -17, -32,   2,   4,   3,   1,  -3,  11,   0, -12,   2,  -4,   3,   3,   3,  10,   7, -10,  11,   9,   1, -15, -13,  14,  15,   1,   3, -10,  15,  15, -39, -15, -10,  -8, -12,  20,  47,  37,  28,  -3,  26, -10,  38,   6,  28,  37, -24,  12,  -3,  -6,   5, -24 },
      {  -2,  -2,  -1,   2, -11, -40,   3,  -4,   0,   0,   3,  -1,   5,  -9,  -3,   2,   2,   4,   8,  19, -10, -31,   2,  22,   2,   5,   2,  24,  65, -12,  44,   0,  -2,  -5,   3, -39,  21,  15,  10, -13,   1,  -6,  27,   5,   2,   1,   3,  -4,   2, -36,  -1,   1,   7,   6,   5, -11, -16,   9,  -5,  11,  11,  -3,  -4,  -7 },
      {   0,  -1,   3,  -1, -11,  -4,  -4,  -5,   0,   0,   1,   0,  -3,   3,  -7,   7,   1,  -3,  11, -11, -34,   6, -15,  -3,   3,  10,  11,  40, -29, -20,  -1,  24,  -1,   2,  22,  24,  40,  -4,  49,  -5,   4, -34,  -4,  23,  -4,   9,   1, -22,   5,  11, -29,   7,   3,  -6,  -1,   0,   2, -13,   7,  41, -17,   1,   3,  -2 },
      {  -1,   2,  -1,   0,  -3,   3,  -7,  -1,  -2,   1,  -1,   1,  -6, -18,  -7,   1,  -1,   4,   0,   3,   1,   0, -15,   4,   2,   2,   2,   3, -16,  -8,   0,   5,   2,  16,  -6,  28,   5, -10,  -3,  17,   3,  15,  -2,  -4,  33,  -5,  28,   6,   7,   3,  18,  -3,   7,   1,   6, -24, -99,  21,   7,   4,   5, -11,  -6,  -1 },
      {  -3,  -4,  -3,  -5,  -7, -10, -56,  14,  -3,  -3,  -3,  -3,  -7, -41,   0, -17,  -1,  -2,   1,  -2,   0,  -5, -56,  12,  -2,   1,   1,   5,  -5, -11,   9,  -2,  -2,  -4,   1,   5,   5, -12, -19,   3,  -9,   1,  -5,   3,   7, -15,   6,   1, -15, -12,  -1, -11,   3,  -7, -35,   6,  24, -12, -13, -12,   0, -46,  13,  32 },
      {   0,  -1,  -1,   0,   9,   0,  -2, -10,   0,  -1,   1,   5,  -4, -15,  -7,   7,   1,   2,   0,   8,  30,  -6,  28,   2,   0,  -3,   4,   2,   0,   6,   0,   4,  -4,   1, -13,  24,  48,  -6,  27,   9,  -6, -20,  -9, -64,  31,  13,   7, -28,   4, -15,   5,  -5, -26,   2, -18,  -2,  14, -10, -19, -29,  18,  13,   0,   5 },
      {  -2,  -3,  -1,  -8,   9,   1, -85,   3,   1,   3,   2,   5,  -4,  26,  49,   2,  -1,  -2,  -3,  -6,  -7, -10,  35, -11,  -1,  -1,   0,   1,   1,  -9,  -9,   3,  -3,  -4,  -5, -17,  12,   7,   0, -11,  10,   0,   7, -16, -31,   8, -10,  -1,  -2,   6,   2,  16,  -5,  -5,  -1,   1, -31,  -8,  -2,  -8, -22,   8,  -2,   3 },
      {   0,   0,  -1,  -4,  -3,  -9,  12, -11,  -1,   0,  -2,  -1,   6, -28, -32,  -3,  -1,   0,  -1,  -9,   6,   2,  -1, -12,  -1,   0,  -6,  -1,   4, -26, -10,  -8,   0,   3,   1, -27,  31,   8, -11,  -3,   8,  -4,   7, -24, -26,   2, -22,   0,   7,  39,   0,  -1,   2,   7, -14,  24, -14,  -1,  76, -24,  -2, -13,   6,   7 },
      {   2,   3,   3,   0,  -1,  29,  55,  20,  -1,  -2,  -3,   0, -10, -19,  56, -13,  -1,  -4,   0,  -7,   9,  -5, -35,   0,  -1,   0,  -6,   4, -10,  10,  24,   5,  -3,  -6,  -9, -21,  14, -20,   1, -27,   2,  -3,   7, -25, -24,   5,  -3,  -3,   4,  -6,  10,   3,  -7, -13,  -7,  -3, -19, -14, -12,   2, -39,   7,   7,   8 },
      {   1,   1,  -1,   0,   1,  16,  22, -12,  -1,  -2,  -3,   0, -11,   7,  29, -25,  -2,  -2,  -7,  -4, -18, -38,   4, -15,   1,  -2,   4,   6,   6, -25, -31,  -3,  -1,   1,  -3,   7,   6,   6, -10,   8,   0, -13,  -1,  11,   5,  -7,   0,  -8,   9,  -7, -50,   3,   1,   1, -16,   2,   2,  69, -11, -34,   6, -12,  -8,  12 },
      {   0,   0,  -1,  -2,  -1,  -4,  -8,  43,   0,   1,   0,   1,   9,   9, -38,   1,   1,   3,  -1,   0,   0,  54,  22,   8,  -4,  -4,  -9,   0, -27,  -9,  46, -19,  -2,   0,  -8, -10,   0, -20,   0, -21,   1,  -6,   7,  -4, -15,  -3,   2,  -7,   9,  -8, -25,   8,  -7,  -5, -10,  -6,  -3,  52, -15, -15, -16,  -1,   2,   2 },
      {   1,   1,   1,   0,   6, -11, -21,  36,   1,   2,  -1,   4,   7,   6,  16,   1,   1,   1,   2,  -1,  18,  -8, -27,  10,   2,   2,   1,   4,  15,  61,   0,  25,  -3,   1, -10,  -4, -18,   2,  35,   1,   0,  -6, -10,   1,  16,   4,   3, -11,   6,  10, -22, -35,  -7,   6,  11,   9,   2,  33,  49, -10,  -8,  13,  10,  -3 },
      {   1,   2,   2,   0,   3, -15,  16,  98,   0,   0,   2,   0,  -2,   5,  15,  -9,   0,   0,  -1,   1, -10, -12,  21, -18,   1,   2,   5,   1,   4, -31, -11,  -1,   2,   4,  10,  11,  17,   8, -15,  23,   0,   3,   0,   4,  12,   9,  15,   8,  -2,  -2,  10,  -2,   8,  -7,  -2,   8,   5, -18,  11,   6,  24,   9,  22,  -4 },
      {   0,   1,   0,   1,   3,  -1,   2,  27,  -2,  -2,   0,   1,  -4, -10, -34,  13,  -1,  -1,   0,   1,   9,  -1, -36, -22,   2,   2,   6,   5,   9,  31, -25,  23,  -3,  -1,  -6,  -2,  -4,  36,  15,   1,   3,  -3,   8,  -3, -39,  -1,  -9,  -7,  -5,   5, -12,  56,   2,  -1,   2,   3, -21,  -7, -39, -24,  27,  -5,  -7, -11 },
      {   0,  -1,   0,   0,   2,   4,   5,   0,   0,   0,  -1,   5,   6,  -7,   2,  20,  -1,  -1,  -1,   0,   6,  11,  11,  -8,   0,   0,  -3,   5,   2, -10,  11,  16,  -1,  -3,   1,   1,  11,   8,   5,   4,  -6,  -1, -13,  20,   2,  -6,  12,  11, -12, -10, -24, -29,  54,  -7,  21, -24,   1, -35,  -1, -71, -32,  -9, -12, -23 },
      {   0,   1,   0,   1,   0,   6,   7, -12,   1,   1,   2,   1,  -3,   8,  31,  11,   1,   1,   1,   2,  -5,   1,  18,  21,   0,   0,  -1,   3, -10,   3,  27,  -3,  -2,   0,  -5,   1, -17, -26,  29,   2,   6,   2,   6,   6,   1, -48,   3, -24,  -1,  14,   1,  27,  25,  -7,  -4,  33,  -7, -11,  10, -20,  50, -23,  43, -11 },
      {   0,   1,   0,   0,  -2,  15,   5, -14,   0,   0,   0,   1,   4,   0,   3, -12,   1,   2,   1,   1,   2,  18,   8,  41,   0,   0,   2,  -1,   9,  -1,   2,   0,   0,  -1,   0,  -2,   3,  59,  -1,  47,  -4,  -1, -16,  -7, -41,   1,  51,  -6,   1,   5,  -1,  -8, -14, -22,  -7,  -8,   1,   8,  -1,   9, -16,  -8,  41,  -2 },
      {   0,   0,   0,   0,   3,   6, -14, -20,   0,   1,   0,   3,  -1,  -5,   7, -25,   2,   2,   4,   0,  13,   8, -23,  23,   0,   3,  -1,   6, -10,   4,  19,   4,   3,   3,   9,   2,  25, -20, -27,  10,   3,   4,  11,  18,   1,  59,   8,  27,  -4,   5, -17,  23,  16, -10,  13,  17,   4,   5,   1, -20,  33,  54,  17, -10 },
      {   0,   0,   0,   0,   1,   3,   2,   4,   2,   3,   3,   2,   4,  -4,  19, 107,  -1,  -1,  -3,   0,  -4,  -4, -23,  15,  -1,   0,  -1,   0,   1,  -7,  -8, -41,   2,   1,   3,   1,   9,   3,  -9,  17,  -1,  -3,   0,   3,   1,  13,  -7,   8,   4,   1,  -6,   2,   0,  14,  -3,   9,   1,  14,  -5,  -1,  -6,  10,  13,  18 },
      {   0,   0,   0,   1,  -1,  -4,   7,  20,   0,   0,   0,   0,   0,  -3,  -4,  -6,   1,   1,   2,  -1,   4,  -7,  13, 100,  -2,  -2,  -3,  -1,  -9,  -9, -49,  12,   0,   1,   1,   5,   2,   3,  -4, -39,  -2,  -1,  -4,   2,   0,  -6, -21,  -2,   2,   6,   0,   3,  14,  10,  -3,  -8,  -6,  -7,  -5,  -6,  -1,   7, -14,   7 },
      {   0,   0,   0,   2,   0,   3,   6,  -3,   0,   0,   1,   0,   3,   0,  -7,   4,   1,   1,   2,   2,   2,   9,  20,  -2,   1,   0,   3,   0,   7,  -5,  18,  74,  -1,  -1,  -2,   0,  -5,   1, -12,  31,   1,   1,   4,   2,  10, -12, -22,  17,   3,  -1,   3,   6,  12,  50, -14,  31,  -7,   5, -18,   4, -27,  15,  22,  46 },
      {   0,   0,   0,  -2,  -1,  -2,   4,  -4,   0,   1,   1,   0,  -1,  -2,   5,  35,   0,   0,   1,   0,   0,  -1,  11,  -3,   0,  -1,   0,  -1,  -4,  -3,  12,  54,  -3,  -3,  -4,  -3,  -7,  -7, -19,  -7,  -3,   1,  -5,  -2, -19,  -6,  10,   1,   2,  -1,   0,  -3,  -2, -67,  -8, -28,  10,   8,  24,   7,  39,   9, -34,  44 },
      {   0,   0,   0,  -1,  -1,  -3,  -7,  -3,   0,   0,   0,   1,   1,  -5,  -8, -17,   0,   0,  -1,   0,   2,   2, -12, -13,   0,   2,   2,   1,   4,   1,  -6, -37,   0,   1,   1,   1,  -1,  15,  18,   2,   0,   2,  -3,   2,  -6, -42,   9, -22,   8,  -1,  17,  -4,  54,  -7, -27,  -2,   0,   4,  -1,   3,  -5,  77,  -9,  33 },
      {   0,   0,   0,  -1,  -1,   0,  -6,   0,   0,   0,   0,  -1,   1,  -3,   1,   9,  -1,   0,  -1,  -1,  -3,   0, -13,   0,   0,  -1,   0,  -2,   0,  -9,   5,  22,  -3,  -2,  -4,  -2,  -8,  -8, -56,  18,  -1,   2,  -3,   1,   0,   2, -24, -87,   4,   1,   5,  -4,   7,   9, -10, -22,   2,   4,   2,   6,  -5,  15,   8, -50 },
      {   0,   0,   0,   0,   0,  -2,   1,   9,   0,   0,   0,   0,  -1,   2,   7, -14,   0,  -1,   0,  -1,   0,  -4,  -2,  22,   0,   1,   0,   1,  -2,   5,  17, -16,   2,  -3,   2,  -4,   3, -20,  25,  79,  -2,  -4,  -8,  -3, -15,  -9, -62,  25,  -3,   5,  -3,   2,  -4, -11, -17, -26,  -2,  -1,   1,   2,   3,  -2, -36, -14 },
      {   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0,   1,  -1,   1,  -2, -14,   0,   0,   1,   0,   2,   1,   2, -10,   0,   0,   0,   1,  -1,   5,  -3, -15,   1,   0,   2,  -1,   3,  -3,   6,   1,  -1,   0,   0,   3,  -4,  13, -30, -18,   2,   3,   1,   9,   5,  17,  51, -72,   5,  -3,   5,  -7,  13, -11,  51,  57 },
      {   0,   0,   1,   0,   1,  -1,   3,   8,   0,  -1,   0,  -1,   0,   1,   0, -11,   0,   0,   1,   0,   0,   2,   2,  13,   0,   0,   1,   0,   0,   2,   9, -16,   0,   0,   1,  -1,   1,  -2,  -7,  29,  -1,   1,   0,   4,  -3,  16,   5, -50,   3,  -1,   4,  -1,   5, -12,  63,  57,  -6,  -7,  -8,  -5,  -9, -18, -49,  41 },
    },
    {
      { -96,  56,   2,   3,   0,   1,   0,   0,  55, -11, -19,   1,  -2,   0,  -1,   0,   1, -21,   8,   5,   0,   1,   0,   0,   3,   3,   6,  -3,  -1,   0,   0,   0,   0,  -3,  -1,  -1,   1,   0,   0,   0,   1,   1,   1,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -39, -24,  29,   2,   5,   0,   2,   0, -41,  80, -18, -13,  -1,  -2,   0,  -1,  39, -21, -38,  20,  -1,   3,  -1,   1,   2, -18,  23,   7,  -7,   1,  -1,   0,   6,   1,   0,  -9,  -1,   1,   0,   0,  -1,  -3,   2,   2,   2,   0,   0,   0,   2,   0,  -1,  -1,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 },
      {  48,  -5,   2,  -4,  -3,  -2,  -1,  -1,  33,  10, -56,   8,   0,   2,  -1,   1,   1, -64,  31,  33,  -8,   4,  -2,   2, -12,  14,  38, -30,  -4,   0,   0,  -1,  -6,   5, -15,  -6,  11,  -1,   2,   0,  -4,   3,   5,   3,  -1,  -2,   0,  -1,  -1,   1,  -3,  -1,   0,   0,   0,   0,  -1,   1,   2,   1,   0,   0,   0,   0 },
      {  30,  23, -31,   5,  -1,   0,  -1,   0,  39,   3,  -2, -18,  -3,  -1,   0,  -1, -26,   2, -58,  24,  18,   0,   3,   0,   8, -28,  27,  49, -27,   2,  -4,   1,  -1,  -2,  26, -35, -12,   8,  -2,   2,  -1,  -1,  -4,  -2,  14,   1,   0,   0,  -2,   0,   4,  -1,  -1,  -3,  -1,   0,  -1,  -1,  -2,   0,   1,   0,   1,   1 },
      {  15,  84, -40, -10,  -6,  -2,  -2,  -1, -71,  -6, -15,  22,   5,   5,   2,   2,  26,  -7,  10,  12,  -9,   0,  -2,   0,   9, -17,   5,  -7,  -5,   5,   0,   1,   3,  -3,   8,  -2,   4,   1,  -1,   0,   2,  -1,   0,  -2,   1,  -2,   0,   0,   0,  -1,   2,   0,   1,   0,   1,   0,   1,  -1,   0,  -1,   0,   0,   0,   0 },
      {  20,  20,   3, -24,   2,  -2,   1,  -1,  18,  30, -34,   6,   0,  -4,  -2,   0,  10, -19,  -9, -28,  10,   4,   3,   1, -23,   6, -43,  41,  33, -13,   4,  -3,   5, -11,  17,  46, -43,  -6,   2,  -1,   0,  -5,  12, -22, -12,  18,  -1,   1,   1,  -3,   3,   1,   6,   1,  -4,   0,  -1,  -1,   3,  -4,  -1,   0,  -1,   1 },
      {  23,  36,  14,  -8,  -9,  -3,  -3,  -1,  23,  22,   5, -47,   5,   2,   1,  -1,  -1,  14, -32,  -7,   1,   0,   1,   0,   3, -25, -15, -30,  23,  17,   0,   3, -13,   3, -21,  30,  51, -21,  -2,  -2,   1,   2,   1,  32, -27, -20,   7,  -2,  -2,  -3,   0,  -3,  -9,  10,   4,   0,   0,  -1,   1,   5,  -1,   1,  -1,  -1 },
      {  -3,  13,  46, -16,  -2,  -6,   0,  -2, -24,   1, -39, -15,  22,  -2,   6,  -1, -65,  31,   3,  45,   4, -10,   1,  -2,  27,  26, -16,   0, -27,  -2,   1,  -1,   6, -20,  -7,  11,  -9,  10,   1,   1,   8,  -2,  12,  -5,  -9,   7,  -2,   0,   1,  -4,  -1,  -4,   4,   4,  -3,   1,   3,   0,   2,  -1,   1,  -1,  -2,   1 },
      {  13,  21,  29, -17,   3,  -4,  -1,  -2,   2,  21,   0,  -4, -24,   6,  -2,   1,   4,  -9,  12, -36,  18,  -1,   0,  -1, -14,  26, -33,   5, -24,   5,   3,   1,  11, -28,   1, -36,  24,  36, -12,   7,  -5,   7,  -9,  14,  49, -32,  -6,   1,  -2,  -3,  -4,  15, -17, -15,  12,  -1,  -3,   1,  -2,   0,   0,   5,   4,  -2 },
      {   5, -18,  -4,  -5,  -4,   0,   1,   0,  35,  10, -29,  32,   0,  -4,   0,  -2,  30,  44,   4,   9, -49,  -5,  -4,  -1,  15, -45, -26, -24, -17,  24,   6,   4,  -7, -25,  25,   3,  11,  22,  -3,   0,   1,  -1,  11, -21,  14,   2,  -9,   1,  -3,   0,  -2,   2,  12,  -6,   0,   2,   0,   1,   3,   0,   1,  -3,   0,   1 },
      { -13, -30,  -4,  -8,  18,  -3,   1,   0, -10, -33, -33,  15,   7,  11,   2,   3,  10,  -7,   1,  13,  25,  -9,  -3,  -2,  -8, -10,  -9,  30,  -6,   6,   0,   2,  11,  -7,  42,  11,  29, -17, -19,   1,  -3,  20,   3,  10, -28, -51,  17,   3,  -2,   0,   0,  -6, -30,  24,  21,  -5,   0,   4,   2,  -3,   4,   9,  -8,  -3 },
      {  13,  16,  78, -27, -10,  -6,  -2,  -2,   2, -54,   9, -17,  13,   2,   3,   1,  49,   0,  -8,  10,   1,  -7,   0,  -1, -19, -18,  13,   5,  -5,  -3,   2,   0,  -7,  20,  14, -13, -11,  -2,   5,  -1,  -4,   3, -16, -19,   3,  12,   4,  -1,  -2,   2,  -4,   6,  13,   2,  -5,  -2,  -1,   1,  -2,   1,  -2,  -4,  -2,   1 },
      {  -3, -12, -28,  10,  -2,   7,  -2,   2,   0, -29, -11, -35,  22,  -2,   5,   2,  42,   4, -16,  23,  23,  -4,  -3,  -3, -10,  11, -11,  -1, -12, -24,  10,  -1,  -8,  -2, -28,  30,  -6,  39,   5,  -6,  10, -12,  35,  11,  16,  -8, -36,   3,  -2,   3,  -1,  -7, -11, -30,  15,   9,  -1,   0,   2,   4,  -6,  11,   9,  -4 },
      {   7,   8, -15,  17,   7, -13,   1,  -2,  15,  16,  11,   2,  -6,   3,   0,  -1,  38,  18,   6,  11,   9, -37,   2,  -4, -24,  30, -17,   2, -52,   0,  13,  -2,   5,  -8, -20,  -9,  -5,   4,   7,   2, -10,  -2, -10,  15, -36,  22,  21,  -5,   5, -10,   6, -13,   7,  45, -21,  -5,  -1,  -3,   3,  -2,  14, -11, -14,   5 },
      {   9,  13,   7, -22,  19,   2,  -4,   0,  18,  20,   9,  24,   8, -33,   2,  -2,  11,  30,  10,   4,  -5,   3,  -9,   1,  22,  -4,  17,  -3,   4, -49,   4,   0,  -7,  10, -22,  -9, -29,  -3,  12,  -4,  -4, -12, -13,   0,  -3, -23,   9,  10,  -5,   5,  -9,   5, -39,  18,  42,  -7,   2,  -1,   3,  -2,   8,  35, -14, -11 },
      {  -4, -17, -10, -34,  15,   1,   6,   1,  -3, -17, -34,   4,  -4,   7,   0,   2,  15,   7,  -2, -14,  23,  11,   3,   2,  45, -12,  -5,   4,  11, -19,  -7,  -3, -26, -18, -14, -16,   4,  -9,  31,   1, -10,  19, -15,  47,  12,  33,   4, -21,   3,   1,  16,  14, -11,  -1, -37,  -1,   1,  -2,   2,  -5,  -1, -10,   0,  10 },
      {  14,   6,  20,  57, -14,   2,  -8,  -1,   1,  -4,  11, -25, -34,   5,  -1,   3,  19, -21,  20,  -4,  19,  12,  -5,   0,  63,  -8,  -8,  -2,  -1, -20,  -7,   0, -10, -38,  17,   4,  -9,  -4,   7,   3,  -2,   9,  21, -20,  -9,  -2,   8,   1,  -5,   5, -10, -13,   6,  14,  10,  -4,   1,   5,   2,   3,   8,   2,  -8,  -5 },
      {  -1,   4,  25,   0,  18,  -5,   2,  -1, -11, -14,  16,  31, -33, -19,   7,  -1, -28, -30, -16, -11, -13,   9,   6,   5, -17, -37,   1,   0, -13,  -6,  25,  -2, -12,   4,  -5,  15,   5,  42,  15, -14,   8,   3,  26,  24, -13,   9, -25,  -7,   6,   3,  12, -31, -15,  18,  -7,   3,   3,   0,   1,  -4,  25,   1, -13,   3 },
      { -12, -13, -27, -19, -10,   8,  -2,   5, -12, -10,  -1, -40,  12, -12,  15,   2, -15, -24,  -6,  -4, -13,   0,   4,   4, -13, -13, -11, -16,   3,  -2,  24,   1,  -3, -17,  -1,  -1,  -1,  20,   8,  -1, -10,  -4,   2, -18,  24,   5,  35,  -2,   0,  22, -14,  36,   7,  39,   5, -35,   3,   1,  11,   8,   3,   2, -46,  -2 },
      {  -2,   6,   7,  60,  -6,  -3,  -5,  -3,  -8,  16, -34,  -7,   3,  11,   2,   2, -19,  26,  29,  -9,  25, -11,  -8,  -2, -37, -39,  -3,  -3,   1, -24,  14,   4, -25,  33,  13,   3,  -4,  11,  15,  -4,   7,  16, -29,   4,   8,   1,  -5,  -5,  -3,  -3,  -1,  25,  -6,  -2,  -3,  -2,   3,  -1,   5,  -4, -13,  -1,  -5,   1 },
      {   4,   6,  14,  45,  -9,   1,  -3,  -1,  -7, -28, -44,  15, -29,  -5,   8,   1,  10,  31, -42, -11, -20,  40,   6,   2, -22,  28,  14,   1,  18,  10, -16,  -1,  19, -10, -23,  -7,  -3,  -7,  -5,   6,  -5, -20,   8,  10,  10,   1,  15,   1,   4,   4,   4,   8, -10,  13,   0, -11,   1,  -4,   5,  -4,   5,   6, -14,  -1 },
      {  -2,  -7,  -7, -23,   9,  -9,  -2,   0,   0,  -2, -23,  -4, -38,  30,   7,   0,   3,  10,  -1, -17,  24,   8, -14,   3,  34,   2,   6,  -1,  -3,   7,  18,  -6, -14,  47, -16,   6,  -1,  21, -38,  -4,  20, -34, -16, -14,  -1, -12,  -4,  25,  -6,  -5,  11,  -5,  32,  17,   7,  -6,   6,   1,   0,  10,   9, -20, -13,  -1 },
      {  -8,  -5, -13, -34,   2,  -1,   0,  -1,  -5,   2, -11, -29, -32,   9,  -3,   6,  -8,  28,  25,  -4,  15,  21,  -9,  -1, -45, -26,  17, -12,  -3, -23, -25,  13,  19, -25,   1, -18,   4, -12,  26,  10,  -6,  -8,  16,  -8, -10,   8,   0,   1,   3,  -3,   4, -33,  24,   0,  20,   9,  -6,   9,  -7,  16,  16,  -1,  11, -18 },
      {   3,   8,  -5, -12,   2,  17,  -4,  -3,   8,  16,  10, -22,  -4,  -3,  17,  -3,  12,  31,  27,  13,   3,  34,   1,  -4,   0,  22,  32,   7,  21,   3,   4, -12,  24,   7,  29,  10,   7,  24,  -9, -23,  -2,  24,   3,  -9,  -1, -10, -26, -22,   4,   2, -20,  -7, -22,  10, -38,  -2,   3, -10,   4,  -9,  15, -18, -25,  26 },
      {  -5,  -8, -14,  -5, -29,   1,   5,  -2,  -8,  -8, -24, -26, -12, -20,   3,   8,   1,  -9, -15, -18, -24,   2,  28,  -7,  13,  24, -14,  -7,  -7,  10,   2,   6,  -5,  21,  19, -13,  -4,   7,  19, -16,  -6,  25, -44,  -8, -20,   9, -29,   0,   4, -30, -11,  -9,  -5,  11,  22,  28,   0,   3,   3,   4,  19,  28,   7, -19 },
      {  -2,  -9,   4,  -6, -87,   9,   1,   6,   5,   2,  -6,   6,  23,  20,  -1,  -1,  -3,  -2,  15, -24,   6, -16,  -4,   1,  12, -25,  12,   9,  -4,  -3,  15,  -3,  43,  -3, -14,  -5,  -9,   4,  -9,  -8, -13, -24,   4,  17,  -4,  -1,  -2,   1,   2,   5,  23, -12, -20,   6,  -4,   0,  -2,  -1,  -6, -20,  12,   7,  -1,   5 },
      {  -8,  -4,  -4, -20, -55,  15,   0,   3,  -4,   9,  11,  11, -14,   9,   3,  -3,  -1,   6,  -1,  17,   0,  19,  -2,  -6, -23,  14,  -9,   5,  -9,   0, -21,  -4, -79,   2,   2,  -4,  -9,  -5, -13,  24,  26,  17,  17,  -1,   4,  -7,  16,   4,  -9,  -3,  -3, -12, -16,   6,  -6,  -5,   3,  -6,  -1, -12,   9,   8,   0,  13 },
      {   0,  -2,   1,  -6, -26,   1,   6,  -4,   9,   6,  24,  42, -12,  37,  11,   3,  -2,  -9, -27,  45,  23,  27, -21, -11,   1,   5, -32, -19,   5, -19, -10,  15,  31,  12,   3,   9,  12,   5,  26,  -8,  -3,  14, -17,   2,   1,   5, -14,  -6,   5,  -8,  -1,  20,  15,   3,  10,  -7,   0,   3,   9,   7, -11,   2, -22, -17 },
      {  -1,   3,  -2,  -7, -15, -20,   3,   3,   0,   5, -12,  25,   1,  11,  -4,   7,  -3,  17, -32, -36,  20, -30,   5,   8,   6,  28,  30, -23,  -3, -13,  30,   2,  -5,  13,  16,   2,  13,   0,  23,  -5,  -6,  26,  35, -21,   0,  -2,  12, -20,  -9,  22, -33, -17,   7, -11,  -3, -18,  -6,  -2, -18,  18, -10,   5,   6,  -4 },
      {   1,   0,  -1,  -4,  -4, -26,  18,  -1,   4,   6,   2, -16, -32,  -8,   5,   4,   9,   6,  32,  21,  -7,   2,  35,   0,  11,  -3, -12,  24,  -2,  29,   1,  -2,  21,  41,   5,   4,   2,  -3,  22,   9,  35,  -6,  20,  17,   9,  15,  20, -19,  -2,  20,  12,   0, -11, -12,   6, -25,  -3,   7, -15,  -5, -24,  30,  -6, -18 },
      {  -2,  -6,  -9,  -7,   3, -23,  -8,   5,  -2, -12,  -4, -15, -32,   0,  -6,  -6,  -4, -12, -15,   7, -13,  -7, -25,  11,   1, -14, -23, -25,  -9, -23,   0,   1,  35,  16,  -6,   0, -15,  -7,   0,  22,  22,  -6,  -6, -19,   1, -13,  16, -18, -20,  -4, -29,   6, -30,   1, -28,  19,   6, -15,  10, -15,   4,   5,  35,  40 },
      {   5,   2,   3,  10,   8,  45, -14,   5,   4,   2, -10,   3,  40, -21,  -7,   6,   5,  -7,  -5, -16,  -2,   3, -11,  -1,   8,   9, -20,  -2,   3, -15, -11,  24,  17,  26,  11, -25,  18,   2,  20,  14,  59,  -7,   8,   2,  10,  -4,  13, -10,  -2, -20,   9, -28,  13,   6,  -5, -17,  -3,   5, -13,   1,  24, -19, -11,   2 },
      {  -1,   3,  -2,   4,   8,  -9,  -8,  -4,  -2,   3,  -4, -19,  -1,   6,   5,   3,  -3,  -2,  -2,   8, -29,   4, -30,  -6,   5,  17, -30,  -8,   3, -15,  -2,  19, -10,  43,  26, -17,  -7,  -2,   0,   1, -64,   2,  22,  12,   4,  -3,   2,  17,  11,  34,  28, -18,  -8,  -6,   1,  -5, -17,  -4, -17, -22,   6, -16,   0,   3 },
      {   0,   2,   2,   7,  10,  23,  13,  -3,  -2,   1,  -9,  10,  19,  10,  -1,  -5,  -4,  -5,   7,  -5,  13,  15,  27, -17,   2, -25, -31,  10, -12,  13, -29, -22,   3,  30, -35, -21,  -1,  -4,  -9,   5, -39,  -4,  16, -12,  -6,   0, -13, -14, -26,  11, -45, -13,   6,   6,  -8,   5,  -9,   1,  -8,  29,   6,  18, -14,  19 },
      {   1,   4,   7,  -6,   5,  78,  -7,   6,  -4,  -1,  -1,  -1, -38,  19, -41,   2,  -3,   5,   0,  16,  -8, -30,  -2,  13,  -2,  13,   1,  11,   5,   6,  20,   7, -11,   3,  12,   4,  11,  -7,  21,  -7, -14, -39,   1,  -2,  -4,  -1,  -5, -20,  -5,  15,  -1,  14, -10,   1,  -4,  15,   7,   2,  13,   4,  -5,  10,   3,  -5 },
      {   3,   4,   3,   6,  31, -11,  35,  -4,   2,   2,   5, -15,  30,  81,  -6,  -6,  -2,  -4,  -4, -17, -28,   8, -17,   2,  -2,   8,  12,  -7,  -3,   9,  -5,   5,  -8, -11,  12,   3, -16,   4,   8,  -1,  17,  -8,  -3,   1,   7,  12,  -4, -10, -22,  -8,   6, -11, -20,   7,   2,  -3,  -1,  -1,   1,  -8,  18,  36, -14,  -7 },
      {   3,   3,   0,  -7,   9,  38,  -6,  -3,   1,  -4,  -6,  -9, -16,  19, -22,  -5,   1,   5,  -3, -11, -15, -15, -13,  -1,   4, -13,   8,   3, -13,   8,  -8,   6,  24,  10, -32,   7, -22,   1, -17,   8,  14,  77,  13,   4,  -2,  12,   2,  29,  36,   7, -10,   7,   1,  -3,  12,  -8,   0,  -9,   2, -10,  -8,   5,  -9,  -5 },
      {   0,   0,   1,  -5,  -6,  -9, -16,   0,   2,   4,   3,   1,  16,  -4,  -4,  -6,   1,  -2,   7, -34, -14,  27, -11,  -5,  10,   7,  22,  23, -55,  -6, -33,  14,   0,  21,   7,  42,  23,   5,  19,  27, -10,  -9,  13,   2,  -1,  -1,  -1, -14,   5,  13,  -2,  33,   8,  15,   1,  23,  20,   2,  26,   0,  -6,  -6,   5, -11 },
      {   3,   3,   2,  -2,  13,  24,  14,  12,   3,   4,   3,   3,   8, -10,  -4,   2,   1,   3, -10,  -4,  45,  13,  35,   2,  -2,  -3,  -1, -66, -16,  15,  -6, -22,  12,  11,  34,   7, -34,   0, -16,  19,  -5,   0,   8,   6,   6,   5,   9,   0,   0,  12,  26,   7,  -7,   7,   0,  10,  14,   1,  13, -21,  13,   1,  12, -19 },
      {  -1,  -3,  -1,  -8,   0,  11,   0, -21,  -3,  -4,  -4,  -6, -17, -15,  -4,   5,  -3,  -7,  -8, -21,  -2, -10,  -2, -29,  -3,   1,   8, -25, -36,  -9, -30,   6,   1, -13,  20,  12, -13, -17,  -5, -18,  11, -21, -14,   9, -18,  -4, -24,   7, -15,   0,   7,  -1,  -6, -20,  -6, -28, -14,   1, -21,  -3, -47,   4, -42,  34 },
      {   1,   0,  -2,   2,  -6, -25,   0,  -9,   2,   2,  -5,   1,  12,  -5, -32,  -8,   3,  -1,  -3,   0,   8,  -3, -12,   0,  -5,   6, -14,  -4,  13,   1, -16,  -9,  -3,  -9,   7, -27,   5,  -2,  -7, -16,  39,  -7, -11,   4, -12,   9, -23,   5,  25,  83,  -9,  -2,   0,  12,   2,  28,  29,   3,   2,   9,   7,   2, -13,   3 },
      {  -2,  -4,  -1,  -3,  -2, -34, -19, -18,  -1,  -3,   3,   1,   2,   9, -54,  18,  -3,   0,   0,   6,   9, -23,  24,   3,   4,  -6,   2,  -8,  19,   2, -31,  -2,  -7,  10,   6,  -1,   1,  11,   6,  15, -12, -23,  16,  -8,  13,  -2,  -8,  -2,  50, -42, -14,   0, -18,  -1,   0, -16,   1, -22,  15,  -3,  14,  -7, -28,   1 },
      {   2,  -1,   1,  -1,  11,  13, -35,   8,   4,   2,   0,   1,   8,  25,  50,  -4,   0,  -5,   2,  -9,   5,  19,   3,  -1,   0,  -5, -10,  -7, -17,   6,  15,  -1,  -9,  -1,   9, -16,  -3, -14,   1, -22,   9, -29,  10,   0, -12,   9,   9,  -4,  73,  -9, -26,   2, -10,  -4,   0,  -3,  10,  -2, -12,   0, -18,  21,  22,   8 },
      {  -3,  -3,   0,  -5,   9,  -9, -82,  -6,   1,   4,   4,   6,   4,  16,  12,  35,   1,   2,   1,   3,  10,  -8,  16, -14,  -1,   1,   6,  -3,  18,  30,  -3,  32,   3,  -5,  -4,  -8, -15,  11,  11,   1,   3,   9,  12,   8,  -2,   8,  -6,  13, -46,   7,   7,  11,   2,  15,   4,  19,  10,   3,  -8,  -9,   1,  15,  -2,  13 },
      {   0,   0,   2,   4,  -4,   5,  -7,  10,   1,   0,   1,   5,   1,  -4,  -4,   8,   0,   1,   3,  10,   6,   1,  -6,  10,   3,   2,   9,  -1,   1,  12,  -6,  14,   3,  -3,  20,  42,  -8,  21,  -3,  28,  -9,  -8, -46,  37,  21,  17,  16,  17,  15,  21, -29, -28,   6,  -2,   5, -14, -30,   9,  -4,  36,  21,  20,  16,  31 },
      {   0,   1,  -3,   2,   6,  -7, -12,  22,   0,   2,   1,   0,   7,   2, -36,  17,   1,   2,  -3,   5,  10,  33,  13,  16,  -3,   0,  -5,   2, -23,   1,  17,   1,  -3,  -2, -24,  18,  23, -16,   8, -13, -10,  15, -22, -52,  15,   0,   5,  -4,  14,   5,  41, -12,  -9,  -7,   2,  -5,   9,  27, -26, -12,   8,  19,  -8,  37 },
      {  -2,  -3,  -2,  -1,   4,   2, -19,  -1,  -1,  -3,  -3,   2,   4,  -5, -11,  14,  -2,  -5,   0,   0,   9,  19, -12,  -7,   0,  -8,  -4,   3, -14,  22,   6,   4,  -7, -12, -18,  14,  -6,  -1,   3,   2,   2,  -7, -35, -16,  -2,  -7,  -2, -21,  -5,  22, -22, -20, -14,  -6, -13, -12, -28, -51, -25, -31,  -2, -20,   2, -64 },
      {   2,   1,   0,   2,   5,  14,   0, -15,   3,   3,   0,   3,  -8,  28,  46,  20,  -2,  -3,  -7,   2, -23, -36,  52, -15,   1,  -2,   2,   3,   5, -41, -27, -16,   4,  -4,  -8,  17,  19,   5,   0,  15,   5,   2, -25, -23,  14,  -7,   5,   1,   8,  28,  21, -16, -11,  -2,   1,   4,  16, -14,  -5,   6,   8, -20,   6,  -2 },
      {  -1,  -1,  -1,  -2,  -2,  -8, -55,  28,  -1,   1,   0,  -4,   1,  17, -21, -21,   0,   0,  -2,   4, -29,  10,  16,  27,  -5,   2,  -9,   0, -12, -35,   2, -54,   7,   2,  16,   1,   0, -11, -21,  -7,   1,   7,  -6,  30,   2,   8,  -6,  -3, -23,  -8, -10,  -8,   7, -10,  -2, -20,   8,  19,  -4,  11,  -3,  -6, -10, -29 },
      {   1,   0,   2,   1,  -3,   3,  24,  -9,   1,   0,   2,  -2,   6,   3, -14,  38,   2,   0,   6,   2,  -9,  26,  13,  18,  -2,  -1,  -6,   6, -11,  -3,  21,  16,   5,  -2,   5,  -2, -10, -10,   8,  -6,  -9,   0,  -2,  24,  -1,  -5,   0,  21, -15,  -5, -18, -14,   9,  -4,  13, -20,  77, -46,  -5,   9, -12, -19,  11,  16 },
      {   0,   1,   2,   0,  -4,  -5,   7,  -2,   2,   2,   4,  -1,   7,  21, -24,  30,   0,   2,   1,   1, -20,  29,  39,  -7,   1,  -1,  -2,  -1, -10, -14,  25,   7,  -2,  -2,  -1,  -8, -28,  -4,   6, -10,  20,   3,   8,  15, -13, -35,   6,  -6,  10,  11,   4,  23,   4,   8,   5,  20, -61,   4,  11,  25,  -5, -36,  10,  14 },
      {   2,   3,   4,   1,   4,  -8,  13, 103,   0,   2,  -1,   2,  -4,   1,  10,  14,   0,   0,   3,  -3,   3, -12,   7,  -5,  -1,   1,   0,   4,  -1,  -3, -16,  13,   0,   1,   4,  -6,  15,  -9,   4,  -8,   6,  -5,   8,  -4,  -2,  15, -15,  30,  -5,   4,  -6,   6,  -7,  -4,   1,   4, -16, -36,  20,  -4,   0,  -3, -17,   6 },
      {  -1,  -1,  -1,  -1,   2,   5,  -2, -46,   0,   1,   2,   2,   2,  -2,  -6,   2,   0,   2,   2,   0,   9,  12,  13,  -2,  -2,   2,  -2,   5, -12, -21,  21, -10,   5,   3,  10,   0,  44, -25, -10, -12,  10,  -7,  11,  -6,   9,  45,  -9,  42, -16,   6, -12,  13,   0,   5,   5, -15, -35, -23,  26, -29,  23,  17,  20,   7 },
      {  -1,   0,   1,  -1,  -4,   3, -18, -15,   0,   1,   0,  -1,   5,  -4,  18,   7,   1,  -1,   3,  -8,   5,   1, -20,  51,   0,   2,  -6,  -2,  -6,  23, -15, -22,   0,   4,  -3, -13,  -7,   4,  15,  -7,   1,  10,  -4, -16, -16,  -4,  -4, -11,  -6,   9,  42, -15, -12, -13,   7, -21, -16, -36,  57,  41, -12,   8,   3,   8 },
      {   0,   0,  -1,  -1,   2,   8,   8,   6,   0,   0,   1,  -1,   9,   5, -19,  15,   0,   0,  -3,  -2,  -3,  17,   2,   8,   2,   2,   2,   1,  18, -13, -16, -18,   1,  -4,   4, -17,  22,  71,   0,  19,  -4,  -5,  -5, -12, -58,  15,  19,  19,  -4,   0,   0,  13,  -7,   4,  -2, -13,  -5,   6, -20, -17, -38,   7,  25,   3 },
      {   0,   1,  -2,  -3,   2,  -2,   3,   3,   0,   1,  -1,   2,   0,   7,   7,   1,   0,   1,  -2,  -4,   1,  -3,  16,   9,  -1,  -1,  -2,   2,  -1,   8,  -9,  26,  -5,   0, -13,  -7,  -7,   4,   4, -10,  -3,  -4, -13,  -1, -14,  -5,  -4, -11,  -5,  15, -30, -20,   0, -23,  16, -38,   6,  57,  59, -51,   4, -28,   8,   4 },
      {   0,   0,  -2,  -1,   0,   3,   7, -12,   1,   1,   0,   2,   4,   3,  20, -11,   2,   2,   0,   0,   4,  -3,  15,  79,   0,   1,  -1,   0,   0,  -1,  -1,   2,   0,  -2,   1,   0,   3,  -6,  12,  23,   0,  -2,  -2,   6,  11,  -4, -10,   8,   1,  -2, -16, -14,  16,  15,  16,  45, -20,  -5, -19, -39, -32,  -7, -37,  13 },
      {   0,  -1,   0,  -1,   3,  -3, -14,   3,  -1,  -2,  -2,  -1,   4,   7,  -9, -67,   1,   0,   1,   0,   1,  13,  27, -34,   2,   2,   2,   0,  10,  -2,  39,  12,   2,  -8,   1, -11,   1,   2,  23,  52,  -3,   0, -12,  -4, -16, -21, -12,  13,   5,   7,  14, -18,  17, -14,  -6,  -8,   5, -26,  22,  -4, -13,  14, -12,  12 },
      {   0,   1,   1,  -1,   0,  -1,   4,  -6,   1,   1,   1,  -1,   5,  -7,   0,  41,   0,   1,   1,  -2,  -2,   9,  -4,  -5,   0,  -3,  -2,  -1,   0, -17,  23,  22,   2,   1,   0,  -7,   7,  -6, -32,  21,  -2,  -1,   0,  -6,   9,   9,   8,  17,   4,  -5,  -2, -34, -18,  -2, -50,  26,   4,  27,  38,  19, -40,  21, -24, -33 },
      {   0,  -1,  -1,   0,  -4,   0,   0,  -7,   0,   0,   0,   1,   1,   6,  -3, -46,   1,  -1,   1,   1,   8,   5,  13,  11,   0,  -1,  -2,  -4,  -6,  17,  -7,  20,   2,   3,  -1,   4,  -3,  -1,  26, -13,   2,  -4,   1, -17,  10,  15,   1,  50, -12,  -5,  -5,  -2, -63,  14,  -8,  -1,   3,  21, -15,  36, -12, -46,  10, -14 },
      {  -1,  -1,   0,   0,   0,  -2,  -1,  -3,  -1,  -1,   0,   1,  -1,  -2,  -7, -35,   1,   0,   3,   2,   4,   2,  32,  30,   0,  -2,  -2,  -4,  -1, -25, -20,  73,  -3,   0,  -1,   1,   2,  16, -49, -33,  -4,   2,   1,   0, -11,   5,  17, -25,  -4,   3,  11,   9,   9, -10, -12,  -3,  -1, -24,  -6,  10,   8,  19,   3,  -9 },
      {   0,   0,   0,   1,  -1,  -1,   4,  -4,   1,   1,   2,   0,  -3,   0,  20,  18,   1,   0,   1,   0,   0,   1,  -3,  26,   0,   0,   0,   1,  -7,   1,  20,  14,   0,  -4,  -1,  -8,   0, -14, -14,  55,   1,  -3,   1,  -5, -40,  12, -32,  -2,   1,   5,  -3,  43, -21, -41,  -3, -14,   3,  14, -24,  18,  45,  -8, -23,  -6 },
      {   0,  -1,   0,  -1,   0,   2,  -7,  -4,  -1,  -1,  -1,   1,   0,   2,  -3,  -7,   0,   0,  -1,  -1,   3,   9,  -7, -25,   0,   1,   0,  -1,  -1,   3,  16,  -7,   2,   2,   4,   0,   9,  10, -16,  -4,  -1,  -1,  -3,   1,  -8,  43,  50, -26,  -9,  -7,  -4, -13, -23, -41,  53,  42,   3, -13,  10,   4,  -1, -30, -33,  15 },
      {   0,   0,   1,  -1,   1,  -1,  -1,   9,   0,   0,   0,  -2,   2,   0,  -3,   0,   0,   1,   0,   1,  -1,  -1,  -2,  -8,   1,   1,   2,  -2,   9,  -4,  11,  15,  -1,  -3,   0,  -6,   4, -10, -36,  31,   0,   4,   2,  -5,  14,  33, -47, -46,   2,  -2,   9,  -9, -18,  53,  39, -23,  -5,   0,  -2,   9, -36, -17,  25,  10 },
    },
  },
  { // 4
    {
      {-119,  31,  15,   7,   3,   2,   1,   1,  19,   8,   1,  -2,  -1,  -1,   0,   0,  15,  -1,  -4,  -3,  -1,   0,   0,   0,   9,  -2,  -4,  -1,   0,   0,   0,   0,   6,  -3,  -2,   1,   1,   0,   0,   0,   3,  -2,   0,   1,   0,   0,   0,   0,   2,  -1,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      {  24,  56, -22, -12,  -6,  -2,  -2,  -1,  88,  13, -38, -12,  -5,  -2,  -2,   0,   1, -29, -18,   5,   4,   2,   1,   1, -18, -20,   4,  11,   3,   2,   1,   0, -12,  -6,   9,   5,   0,   0,   0,   0,  -4,   0,   6,   1,  -1,  -1,   0,   0,  -2,  -1,   2,   0,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
      { -17, -24, -20,  21,   7,   3,   1,   1,  24, -77, -14,  29,   7,   4,   2,   1, -27, -37,  29,  26,   2,   0,   0,   1,  -1,  18,  40,   5, -10,  -5,  -1,  -1,   4,  20,  16, -12, -12,  -2,  -1,  -1,   0,   8,   0, -12,  -3,   2,   1,   0,  -2,   4,  -2,  -6,   1,   1,   0,   0,  -1,   2,  -2,  -3,   0,   0,   0,   0 },
      {  -3, -75, -19,  18,   6,   2,   0,   3,  49,  -6,  -6,  15,  -2,  -3,  -3,   0,  70,  27,  -6,  -7, -12,  -4,  -4,  -2,   5, -18, -18,  -9,  -1,   4,   2,   0, -13, -14,   2,   3,   5,   3,   1,   0,  -5,   1,   8,   2,   1,  -1,   0,   1,  -2,   1,   3,  -1,  -1,  -1,   0,   0,  -1,   0,   1,   0,   1,   0,   0,   0 },
      { -10, -16, -27,   2,  17,   1,   3,   1, -50,  20, -47,   6,  19,   4,   3,   0,  19, -34, -46,  23,  15,   5,   0,   2,   9, -28,  10,  40,  11,  -2,  -2,   0,  14,  -2,  30,  21,  -9,  -8,  -3,  -1,   4,   2,  17,  -2, -12,  -3,   0,   0,   1,   2,   5,  -8,  -5,   0,   1,   0,  -1,   1,   1,  -5,  -2,   0,   0,   1 },
      {   6,  46, -61,  -7,   3,   0,  -1,   0, -22, -24, -39,  12,  11,   3,   0,   0,  23,  41,  21,  13,  -4,  -5,  -3,  -2,  45,  27, -11, -14, -13,  -5,  -3,  -1,   3, -24, -21,  -4,   2,   4,   2,   1,  -1, -13,   0,   9,   5,   2,   0,   1,  -1,  -5,   5,   6,   0,  -1,  -1,   0,   1,  -1,   4,   1,  -2,  -1,  -1,   0 },
      { -24, -29, -48, -27,  12,   6,   4,   2,   8,   1, -12, -21,  21,   3,   3,   0, -57,  45,   5, -11,  21,  -1,   2,  -2, -45,   8, -20,   7,  21,   1,   1,  -1, -14,   6,   2,  23,  14,  -4,  -2,  -1,  10,  11,   5,  13,  -5, -10,  -2,  -1,   7,   2,   0,   2,  -9,  -4,   0,   1,   4,  -1,   0,   0,  -3,   0,   1,   1 },
      {  -8, -14, -24,  -7,   3,   0,   2,   1,   2,  78,  -9,  15,  -1,  -5,  -3,  -1,   1,  14,  31,  32, -13, -10,  -4,  -1, -24,  11,  42,  -3, -30, -11,  -1,   1,  -5,  12,   1, -33, -23,   1,   3,   1,  -2,  -3, -16, -22,   2,  13,   3,   1,   4,  -1,  -8,  -1,  13,   7,  -1,  -1,   2,  -3,  -3,   3,   5,  -1,  -3,  -1 },
      {   0, -39,  16, -30,   2,   9,   3,   1,  26,  12, -25, -43,   6,   6,   3,   1,  -6, -12, -23,  -3,  22,   6,   3,   1,  49,  51,  11,   7,   5,  -7,  -6,  -1,  30,  16, -26, -13,  -3,  -5,  -2,   0, -11, -22, -20,   9,  11,   3,   4,   0,  -7,  -7,   3,  17,   4,  -2,   0,  -1,  -3,   0,   4,   4,  -4,  -4,   0,   0 },
      {   7,   1,  -1,  56,  16,  -7,  -4,  -2,  24,  24,  11,  51,  10, -16,  -1,  -2, -50,  13, -37,   5,  -3, -17,   2,   2,  20,  15, -36,   5, -17,  -6,   6,   1,  25,   3,  -7,  10, -15,   2,   2,   0,  -1, -17,   5,   4,  -9,   5,   3,   0,  -8,  -9,  11,  -1,  -8,   6,   2,  -1,  -2,  -1,   6,  -5,  -3,   3,  -1,  -1 },
      {  -7, -16,  -6,  17,   2,  -4,  -2,   2, -17,  18, -41,  19,  -5,   0,  -2,   0, -17, -54,  -1, -10, -16,  10,   3,   2, -19,   1, -10, -52,   3,  20,   9,   2, -11, -11, -33, -14,  36,  20,   3,   0,  -2,   3,   3,  26,  30,  -2,  -8,  -2,   5,   6,   8,  17,   4, -13,  -6,   0,   3,   3,   6,   6,  -4,  -7,  -1,   0 },
      {   4,  16,  29,  38,  10, -10,  -5,  -2, -13, -18, -50,  -6,  -4,  -5,   0,   1,  20,  38, -43, -22,  -1,  -2,   4,   1, -42,  41,  16,  -5,   7,   1,   5,  -4, -24,  30,   4, -12,  -5,  -5,  -1,  -2,  -4,   9, -12, -13,  -3,   4,   3,   2,   0,  -1, -11,  -2,   8,   8,   2,   1,   0,  -1,  -2,   7,   7,   2,  -2,  -1 },
      {  -2,   2, -56,   9, -17,   0,   1,   0,  -1,   6,  16,  -7, -33,   3,   3,  -1,  -5,  -3, -17, -50, -24,  17,   7,   3,  20,   0,   6, -18,  12,  26,   4,   0,  39,  34,  35,  -9,   7,   6, -10,  -5,  16,   3, -10, -25,   7,  -3,  -7,  -1,  -2, -10, -17,  -8,  13,   2,  -1,   2,   2,  -2,  -5,   3,  10,  -2,  -1,   3 },
      { -11, -11,  -9,   3, -22,   3,   4,   2,  -9, -22,   3, -23, -26,  -3,   6,   3, -35,  22, -39,   4, -10,   2,  10,   2,   8, -36,  17,  24, -25,  11,   3,   2, -33, -34,   3, -29, -12,  20,   6,   1, -23, -23,   1,  -3,  24,  26,   4,  -1,  -5,   4,  20,  15,  21,   5,  -7,  -3,   1,   5,   7,   6,   5,  -4,  -6,  -2 },
      {   0, -14,  37,   8, -19, -10,   2,  -2,   7,  18, -49,  -3, -36,   5,   3,  -2, -29,  23,  35,   0, -19,  15,   3,   2,  33,  12,  15,   0,   8,  14,  -1,  -1,  -3, -33,  29,  25,   9,   2,  -9,  -3,   7,   5,  36,  -8, -13,  -8,  -6,  -1,   4,   2,  -3, -24,  -9,   3,   1,   0,  -1,   1,  -4,  -9,   5,   4,   0,   1 },
      {  -7,  -8, -10,  -7, -67, -18,   0,   2,  -6,  -6,  -3,  27, -52,   2,   3,   2,  15,  -7,   3,  28,  14,  10,   6,   0, -33,  21, -23,  28,  18,   4,   5,  -3,  15,   5, -28,  21,  -7,  -5,  -3,  -5,   5, -21, -15,   4, -11,   6,   2,   2,   6, -11,   6,   7,   4,  14,   2,   1,   5,   1,  11,   4,   4,   3,  -5,  -3 },
      {   0,   8, -29,  43,   4,   7,  -1,   1,  -2,  20,  37,  -2,  -9,   4,   2,   0,   9, -12,   9, -24,  20,  11,   0,   3,   1,  35,  47,  15,  29,  -2, -10,  -5, -34, -18,   0,  17,  10, -16,  -7,  -3, -38, -24,  11,  25,  -2,  -7,   2,   1, -11,   5,  15,   0, -11,  -5,   1,  -1,  -7,   4,  -1,  -6,  -7,  -1,   3,   1 },
      {  -3,  -1,   4, -36,  12,   1,  11,  -4,  -6,  -5,  -9,   8,   6,   7,   9,  -1,   9, -19,  15, -48, -12, -19,   4,   3, -29,  20, -14, -15, -16, -10,  18,   5,  17,  -9,  28,  20, -30,  11,  10,   1, -31, -34,  28,  -6,  -7,  25,  12,  -1, -15,  -6,  25,  -1,  10,  23,  -2,  -5,  -4,  13,  12,  -9,   6,  -2, -11,  -2 },
      {   5,  10,  19,  -3,  -2,  12,   8,  -3,   5,  -4,   0,  17,   6,  19,  12,  -2,  10,  38,   0,  18,  -3,   4,  -2,  -2, -25, -14,  35,   4,  -6,  18,  -3,   2,  48,  24,  19,   3,  23,  19, -12,  -6,   0, -25,  11,  30,  27, -10, -19,  -7, -12,  -9,  26,  17, -11, -27, -13,   0,  -5,   0,  12, -11, -20, -16,   0,   7 },
      {  -4,  -1, -10, -27,  -8,  -7,   3,   2,  -4,  -1,  13,  -9, -18, -28,   1,   3,   6,  -8, -22,  15, -46, -29,  11,   3,  13,  26,  10,  11, -38,   1,  19,   4, -24,  37,   2,  26,  15,  17,  13,  -3, -12,  28,  10,  35,  -7,  -9,  -5,  -7,  -1,   9,   2,  -7, -30,  -1,   3,  -3,   6,  -3,  -3, -10, -10,  13,   4,  -3 },
      {  -5,   2,   7, -43, -26, -24,  -4,  -1,   3,   7,  11,  50,  26,   1,   3,   1,  -3,   8, -22,  17,  27,  21,   2,  -1,  22,  15, -11, -24,  19,  18,  -1,  -6, -36,  18,  31, -19,   9,  -2,  -8,  -4, -34,  11,  11, -18,   8,  -7,  -2,   1, -16,  10,   4,  -5,   9,  -6,  -3,   1,  -4,  11,   0,  -3,   3,  -6,  -1,   0 },
      {   0,   0, -12,  19, -27,  -5,  -1,   1,   0,   1,  14, -39,   1,   4,  10,   2,   9,   0, -17,  20,  39,   5,   1,  -1, -13,  13,   1, -41, -16,  -6,   2,  -1,  19,  -1, -11, -26, -14,  12,   9,   3,  23,  13,  56,   2, -11,  13,  -2,  -3,  -2,  19,  22, -35, -12,   1,  -4,  -2, -13,   2,  -6, -16,   3,   6,   4,   2 },
      {  -4,  -6,  -2,  -1,  -1, -52, -18,  -3,  -4,  -9,  -4, -20,  18, -56,  -5,   1,  -4,   4,   7,  11,   5, -16,   7,   5,  -4, -27,  26, -28,  16,  15,  15,   1,  16,  -6,  -3,  -1,  17,  -4,   4,  -6, -12, -40, -12,  -3, -23, -16,   7,   0, -24, -26,  -7,  -6, -12,  10,  17,   4,  -7,  -2,   9,   8,   4,  16,   6,  -2 },
      {  -3,  -1,   1,   7, -22,  49,   5,   6,  -6,   0, -11,  -6,  -2,  44,   5,   1, -11,   1,  -8,  11, -15,   6, -15,  -5,   2, -25,  -4, -37, -20, -13, -19,   1, -17,  28,  -7,   8,   4, -26,  -8,   3, -23, -15,  -2,   6, -34, -20,   6,   9, -30, -23,  -2,  -4,  -8,  17,  20,   8,   0,   1,  12,   9,  11,  19,   7,  -2 },
      {   0,  -3,   2, -40,  13,   3,  -4,  -4,   7, -11,  18,  32,   5,  12,  -4,  -3,  -3,   2, -48,  -4, -31,  14,  -4,   1, -13,  17,  39, -18,   2,   4, -11,   0,   2, -43, -18,   1,  -2, -16,  -5,   6,  39, -26,   2,   4, -14,  -6,   1,  12,  28,   0,  -5, -19,  -6,   5,   6,   5,  11,  -6, -12,  -5,   8,  10,   6,   2 },
      {  -3,  -4,  -2,  -9, -22,  -5,  -9,   0,  -6,  -3, -14,  13, -24,  -8, -22,   2,  -4,   0,  -5, -29,  14, -21, -17,   2, -14,   0,  -9,  10, -19, -32, -18,   8,  23, -23,  20, -39,   1, -17,  -3,  18, -25,  10,  -8,   0,  22, -35,  13,   9,   3,   5,  -4,  -3, -36, -27,  21,   5,   0,   3, -10, -15, -16,  12,  24,   0 },
      {   0,   2,  -3,  -4,  35,  23,  15,   2,  -2,  -2,  11, -12, -11,   1,   9,  -2,   4,  -9,  -4,  32, -15,   7,   7,  -6, -24,  29, -28,  21,  -1,  37,   1,  -6,  15, -33,  -6, -19,  13,  22, -17,  -7, -31,   6,  -2, -28,  -1, -17, -19,  -1, -32, -11, -15, -23,  -5,  -4,   5,   9, -24,  -7,  -5,   3,  15,  14,  11,   4 },
      {   1,   5,   0, -15,  31,   3,  20,   0,   1,  -1, -11,  20, -31,  -3,  14,  -2,  -6,  15,   5, -11,  -3,  -1,   1,   1,   9, -33,  11,   5,  35,   6,   3,  -1,  17,  25, -44, -19,   6, -10,   7,   0, -22,  10,  -8,   0, -13,  15,  19,   2,  -3,  48,  24, -27, -12,  18,   3,  -7, -14,  16,  -9, -21,   0,  11,  -2,  -6 },
      {   0,  -1,   1,   0, -15,   2,  25,   1,   2,   1,   1,  18,   1,  18,  32,   5,  -2,   3, -13,  -4,  27, -30,  -2,   8,   2,  -5,  21,  -7,   1, -34,  10,  10,   6, -19,  -4,  27,  34,  21,  23,   0,   2,  11, -14,   3,  30,  16,   8, -13, -23, -12, -50, -21,  10,  11,   2, -11, -22,  -3, -20,   6,  12,   4,  -1,  -3 },
      {   1,  -1,  -1,  13,  -4, -30,  10, -10,   3,   0,  -2,  -9,  24,  18,  48,   4,   3,  -1,   6,  -5,   0,  34,  45,   1,  -5, -11, -10,  -2, -40,   1,   0, -14,   8,  -2,  -3,   6, -11,  -9, -20, -12, -31, -14, -23,  12,   3, -13,  -5,  -3,  32,  22, -26,  -9,  -9,   5,   7,  -2,  21,   7, -21, -17,  -5,   2,   1,  -1 },
      {   1,   0,   3, -12,  31,  -6,  17,   0,   0,   4,  -1,  20, -30,   7,  22,   7,   4,  -2,   8, -21,  41,  13,  26,   2,   9,  -7,   3,   5, -17,   8,  16, -11, -25,   0,   8, -32, -17,   8,   3, -12,  39,  11,  -3,  26,  -9, -14,   2,  -7, -12, -47,  10,  11, -27,   9,   8,  -2,  -2,  -8,  16,   7,   5,  17,   1,  -4 },
      {   4,   4,   5,  11, -19,  37,   5,  14,   4,   5,   3,   1,  26,   7,  -4,   4,   5,   0,   2,  -1,  16, -27,   4,  -2,   4,  11,   6,  11,   4,  27,  23,   0,  11, -12,  18, -21,  39,  19,   9,  -2,  -2,   0, -11,  -9, -16, -10,  19,  -5,  41,  11,  11,   0, -14,  28,  11,  -8,  40,  27,  22,   8,  14,  17, -11, -18 },
      {   0,  -1,   3,   4, -47,  23,  19,  17,  -1,   4, -13,   8,  47, -17,  25,  13,   0,  -6,   8, -32, -22,  -6,   7,  -5,   2,  -3,   4,  19,  17,  12,   0,  -8, -18,   1, -24,  -9, -21,  12,   5,  -5,  21,  -8,  -8, -26,  -7,   8, -12,  -6, -17,  -7,  10,  -5, -24, -24,  -1,   4, -14, -19,  -6, -23, -22,  10,  16,   5 },
      {   3,   3,   2,   0,  18,  22,  44,  16,  -2,  -4,   0,  -7, -32, -33,  20,   0,   1,  -5,  -5,  25,  11, -22,  -4,  -3,   4,   0, -16, -31,  25, -16,  -5,   1, -20,  11,  19,   1,  -8,  -1,  -8,  -2,   5, -42,  -1, -14,  12,   0, -10,  -2,  22,  -9, -11, -11,  -3, -18,  -5,   5,  41,   8,  -6, -23, -16, -10,   2,  11 },
      {  -2,  -3,   0,  -3,   6, -16, -25, -11,  -1,  -1,  -6,   2,  -2,  27, -32, -11,  -2,  -4,   5, -20,  26,  -4, -39,  -6,   7, -10,   0,  19, -18,   9,   7,   2, -19,  22, -21,   2,   3,  48,  -4,  -5,  -5, -33,  -9,  -6,  -9,   4, -34,  -3,   1,  -8, -10, -39, -15,  -4, -17,   9,  22,   9, -14, -17,  12,   1,  -1,  14 },
      {   1,   3,   4,  -9,   1,  44, -31,  31,  -1,  -2,  -4,  17,  -2, -33, -21,  21,   0,   9,   5,  -8,  16,  23,  26,  10,  -1, -12,   9,   2,   1,  -4,  12,  -5,  15,  12, -20,   7, -18,   6,  -4, -12, -25,  -4,  43,   9,  13,   6,  -3,  -5,   8,  -6, -27,   2,   5, -11,   3,  -7,  20, -16, -28,  23,  20,  17,   5, -11 },
      {  -1,  -1,   0,   3,   5,  24, -31, -23,  -1,   0,   1,  -6,  -4,  12, -20, -11,   1,   5,  -1,  21,  -7,  44,  14,  13,  -4,   5,  -8, -14,  20,   2,  25,  19,  -3,  -6,  21,   4, -16,  11,  49,  11,  -1,  -7, -32,   0,   6,  16,  21, -17, -13,  -2, -13,  -4, -28,  -6,   2, -19,   3,   0,   1, -32, -28,   8,   2, -10 },
      {   0,   1,  -1,   4,   7, -15,  -5,   2,   1,   4,   4,  -6,   9,  16,   8,   2,   4,   0,   9,   4, -19,   8,  13,  18,   0,   4,  -6,  26,   5, -19,  26,  10, -11,  17, -14,  -4,   7,  -9,  10,   9,  33, -22,  27, -14,  42, -30,   8,  15, -32,  11,   7, -14,  23, -11,  36,   5,  25,  36,  12,  -4,  18,  10,  20, -10 },
      {   1,   2,   2,   4,   8, -26,  16,  36,   2,   1,   1,  -4,   9, -10,   8,  24,   1,  -1,   6,   3,  -9,  35, -40, -14,  -4,   7,  -9,   9,   2,  11, -54,  -4,   3,  10,  19,   0,   4,  30,  16,  23,   2,  -5,  -1,  26,  12,  27,  21,  10,  -3,  -3,  -2,   2,   0,  18,  17,  -1,  15, -10,  -6,  -4,   8,  34,   8, -14 },
      {   0,   0,   3,   1,  -3,   9,  -9,   1,  -1,   0,  -4,  -2,   6,   6, -11,  14,   1,  -6,  12,   6,  -4, -16,  16,  15,   1,   2, -23,  15,  -6,  15,   8,   9, -14,  17,  21, -25,  32, -34,   0,   8,   9, -34,  -2,  28,   7,  10,  16,   2,  22, -16,  16, -30,  22,   8,  -7,  -2, -38, -35, -40, -21,  14,  -8,   2,   2 },
      {   1,   1,   3,  -2,  -4,  29,  -4,  18,  -2,   1,  -7,   4,  -4, -35, -15,  -1,   2,   0,   2,  -6,  28,   1,  -3,  -6,  -1,   3,   2, -11, -34,  32,  -3,  -8,  -4,  -9,  13,  28, -15,  -7, -15,  -2,  32,  -8, -38,  21,  10, -10,  -2,   7, -32,  45,  -3,  -8,  21,   7,   6,   8, -22,  34,  -2, -15,   3,   3,  -4,   3 },
      {   1,   2,   1,   5,   1, -23,  32,  51,   2,   0,   2,  -3,   0,  30, -13,  23,  -3,   5,  -3,   3,   2,  -6, -25,  11,   2,   0,   6,  -3,   2,  36,  42,  28,   8,  -8,  -5,  -3, -33, -14,  31,   3, -15,   6,   1,  15, -22, -25,   2,   1,   1,   3, -17,  10,  14, -16,  -1,   2,   3,  -2, -14,   1,  -9, -21,   7,  11 },
      {   0,   0,   0,   5, -15,  11,  30, -42,  -1,  -2,  -1,  -1,  13, -15,  -2, -59,   0,   1,   3,   0,   6,   0, -20,  -6,   2,  -2,   5,   3,   9,  25,  21,   7,  -1,   5,  -9,  -2, -25,   8, -15,  17,  -7,   9,  19,  20,  24, -16,  21,  31,   1,  -8, -16,  12,   2,  22,  18,   5,  10, -31, -13, -18,  11,   3, -16,   2 },
      {   0,   0,   1,   0,   4,   3,   4, -32,   0,   1,  -2,   5, -11,   7,  17, -13,   0,  -1,   3,   1,  11,  -8,  -4, -23,   5,   2,  -5,  -4,  -2,  10,   5,  -2,  -5,  12,  15,   9,  -5,  13,  35,   8,   4, -33,   2,  -8, -11,  -7,  -7,  -9,  22,  44,   5,  21,   3, -26,  18,  10, -24, -19, -16,  54,   5,  29,  37,   2 },
      {   1,   2,   0,  -1,  11,  -1,  -5,  17,   0,   0,   4,  -1,  -7,   9,  -4,  24,   1,   3,  -1,   4,  20,   5,   9,  34,   5,  -3,  10, -14, -12,   2,  13,  16,  -6,   6,  -8,  36,   5,  26, -22,  18,  -8,   8, -11, -41,  19,   1,   7,  35,  32,  -1,  24,   9,  -8,  10,  27,  23, -24, -34,  23, -17,  -9,   5,  11,  14 },
      {   0,   1,   1,  -2,   0,   7,  30, -19,   0,  -1,  -2,   7,   0, -25,  -5,  -7,   3,   1,   3, -14,  19,  27,  18,   2,  -3,   1,   4,  11, -39,  24, -20,  19,   2,   6, -18,  13,  31, -18,  29,  41,  -4,   4,   9, -21, -19,  13,  20,   7,  -6, -19,   0, -23,   7, -25,  -9,   4,  21,   8,  10,  18, -12, -30,   6,   8 },
      {   1,   1,   1,   3,  -6,   3,  21,  12,   1,  -1,   0,   0,  10, -19, -13, -26,  -2,   4,  -1,  -2,  -6,  31,  -4, -27,   2,  -3,   6,  -1,   2, -27,  19,   1,   6,  -1,   2,   6,  -5, -10,  26,  -7, -18,  17, -16,  14,  13,  -4, -29, -10,  40, -31,  21, -23,  21,  -6,  -5,  16, -27,  23,  27, -14,  32,  22,  22,  11 },
      {   0,   1,   0,  -2,   3,   5,  -7,  22,  -1,  -2,   0,   0,  -6, -19,  20,  -2,  -1,  -2,  -3,  -4,  -2,  32, -27, -21,  -2,   1,  -2,   6, -16, -27,  14, -15,  -4,   5,  -5,   0,  14,  15,  17,   5,   4, -21,   4, -23,  -9, -53,  17,  -5,   8,  11,   1,   4,  -6,  -5,  -7, -24, -26, -22,   3,  -9,  13, -33, -50, -21 },
      {   0,  -1,   1,  -1,   6,  -6, -22,  17,   0,   1,   0,  -3,  -9,  21,  -3, -10,  -2,  -2,  -3,   3,   5, -22,  17, -68,   1,  -3,   8,  -8,  -2,  13,  11, -53,   5,  -4,  -4,  18,   3,  -4,  15,  -9,  -5,  13,  -6, -10,  10,  10,  27,  23,  -7, -21,  18, -11,  16, -19,  19,   4,  19, -12, -19, -16, -16, -13,  11, -14 },
      {   0,   1,  -1,   2,   1, -18,  50,   8,  -1,  -2,   1,  -4,   3,  10, -53, -13,   1,   0,  -3,   1,   5,   9,   2,  -2,  -2,  -1,   1,  -7, -16,  -3,  10,  -1,  -3,   1,  -3,  14,  10, -21, -29,  -9,   7,  -8,  -3, -27,  15,   8,  -9, -30,  -9,   6,  19,   6, -11,  -1,  -3, -48,   3, -18, -14,   6,  -6,  27,   4, -42 },
      {   0,   0,  -2,  -2,   2,  -5,   1,  11,  -1,  -1,   2,  -2,   1,  12, -11, -10,  -1,  -1,   0,   0,  -6, -12,   2, -11,   2,  -2,  -3,   5,   0,  23, -10, -11, -10,  15,  -9,  -3, -13,  -3, -13,  -1,   2, -19,  28,   1,  17,  14,  25,   6,  33, -10, -54, -11, -22, -10,   2, -15, -52,  36,  42,   9, -22,  -4,   2, -18 },
      {   0,  -1,   0,  -3,   5,   4, -22, -22,   0,   0,   0,   2,  -7,  -3,  20,  18,   0,  -1,  -1,  -6,  12, -19, -20,   1,   0,   0,   3,   1,  -7,  31, -25,   7,   0,   0, -14,   6,  -9, -22,  21,  -5,  -8,   7,  13, -11,  17, -30, -21, -39,  11,  -6,  -4,  -9,  43,  18,   0, -22,  11, -25,  32, -31, -24,  24,   6, -17 },
      {   0,   1,   1,  -4,   3,   0, -10,  32,  -1,  -1,   0,   5,  -2,  -5,   5, -30,   0,   0,   1,  -6,   2,  20, -15, -20,  -1,   2,  -4,  12, -13, -11,  28,   4,  -1,  -2,   4, -15,  29, -24,  -7, -16,   6,  -9,  15,  -3,   4,  27,  -3, -23, -12,  23, -21,   4,  -7,  42,  11,  19,  11, -30,  24,   3, -32, -12,  26,  38 },
      {   0,   0,   1,   1,   0,  -9,  -5,  20,   0,   1,   0,  -1,   1,  13,   0,  -6,   0,  -1,   2,   1,  -1, -17,  26, -11,   1,  -1,   0,  -4,   8,   4,  -4,  -3,   2,  -1,   6,   4, -14, -17,   7,  32,  -3,  -5, -11,   6,  23,  -7,   1,  16, -16,  25,  -6, -25, -22, -11, -55,  -7,  18, -40,  45,  14,  32,  24, -19,  22 },
      {   0,   0,   0,   0,  -4,   0,  36, -16,   0,  -1,   0,   2,   3,   5, -50,  11,  -1,   2,  -1,  -2,  -4,   0,  24,  25,  -2,  -3,  -1,   2,   6, -17, -24, -47,   3,   1,   2,  -8,   1,  31,  22, -33,   0,  -8,   3,  13, -16, -20,  18,  -6,  -4,  20, -15, -10,  25,   3,  14,   9,   2, -27,  28, -11,  -2, -16,   6,  20 },
      {   0,   0,  -1,   2,   4,  -3,   6,  -6,   0,   1,   1,   0,  -4,   4, -14,  15,   1,   1,  -2,   3,   3,  -3,  13, -16,  -1,  -1,   3,   2,  -7,   5,  -3, -19,   0,   4,  -8,   4,  18, -11,  16,  18,   4,  -6,  22, -25,  -2,  -7,  10, -10,  -2,  -3, -19,  65,  -8,  17, -34,  11, -10,  26, -21, -49,  24,  26,   7,  34 },
      {   0,   0,   1,   0,   0,  -2,   7,   9,   0,   0,   0,   4,   0,   4, -19,   2,   0,   1,   0,  -4,   3,  -1,  44, -36,   2,  -1,   2,   3,  11, -19,  -4,  28,   0,   1,   3, -15,   8,  18,   6,  35,   1,  -8,   8,   8, -19, -11, -51,  31,  -5,   6, -15,   4,  22,  25,   2,  13, -10,  -1,  -3,  -5, -44,  11, -21, -29 },
      {   0,   0,  -1,   1,  -1,   7,  16, -45,   1,   0,   2,  -2,   1,  -8,  -2,  65,  -1,   0,   0,   0,  -1,  15, -37, -10,   1,   0,  -1,   2,  -5,  -6,  48, -31,   5,  -4,   6,  -5,  10, -25,  -3,  -6,  -2,  -1,   2,   5,   6,  22, -11,  35,  -3,   5, -13,  -7,  -3, -10, -12,  16,   3,  -3,  16,   6, -15,  19, -16, -12 },
      {   0,   0,   0,  -2,   1,   3,  -9,   9,   0,   0,  -1,   1,  -4,  -4,  18, -22,   0,   1,   0,  -1,   2,   1, -11,  30,   0,   0,   1,   1,   0, -12,  24, -34,   2,  -2,   4,  -7,   8,  17, -27,  41,  -1,   2,  -5,  13, -37,   3,   1,  14,  -2,   2, -11,   9,  32, -12, -32, -29,   3,  -3,  14, -18,  -2,   2,  58, -27 },
      {   0,   0,   0,  -1,   3,  -5,  -2,   8,   0,   0,   1,   1,  -4,   5,   7, -23,   0,  -1,  -1,  -4,   7,  -6,   2,   7,   0,   2,   0,   2, -10,   2,  -7,  28,   1,  -2,   2,  -5,  17, -12,   2, -32,   1,  -1,   4,   1, -26,  34,  -1,   5,  -4,  14, -18,  22,   9, -54,  17,  33,   3, -19,  23, -41,  30,  16, -25, -38 },
      {   0,   0,  -1,   1,  -1,   0,   3, -22,   0,   1,   0,   1,   0,  -1,  -8,  43,   0,   0,  -1,   0,  -3,  -1,  10, -52,   1,  -1,   4,  -2,  11,  -5,  11,  29,  -3,  -1,  -2,  -1, -10,  14, -41,  32,   1,   2,  -5,  13,  -9,   6,  14, -45,  -3,   8,  -8,  -1,   0,  -8,  27,  -6,   3,  -9,  20, -15,  36, -33,  20,   0 },
      {   0,   0,   0,   0,   0,  -1,  -3,  -2,   0,   0,   0,   1,   0,   0,   3,   2,   1,   0,   0,  -1,   0,   7,   1, -20,  -1,   2,  -1,   3,  -3, -11,   7,  32,  -2,  -1,  -1,  -4,   9,  17, -27, -24,   2,   2,   1,   2, -19,  -6,  38,  14,   1,  -4,   2,  -5,  36, -29,  -4, -44,  -1,   6,   0,   2, -27,  49, -18,  56 },
      {   0,   0,   0,   1,   0,   2,   3, -10,   0,   0,  -1,  -1,   0,  -6,   0,  11,   0,  -1,   0,  -1,   2,   5,   3, -23,   0,   1,  -1,  -1,  -2,  -3, -11,  46,  -2,   0,   0,   3,  -2,   9,   5, -58,   4,  -2,  -1,  -5,   6, -17,  23,  39,  -3,   5,   1,   2,  -8,  23, -51,   2,   3,  -6,   2,  -2,   8, -18,  52, -32 },
      {   0,   0,   1,  -1,   0,   4,  -1,  -3,   0,   0,  -1,   1,  -1,  -5,   1,   4,   0,   0,   1,  -1,   4,  -4,   1,  -4,   0,   0,  -2,   1,  -2,   7,  -9,   0,   0,  -1,   4,  -4,   7, -12,  23, -11,  -1,  -1,   0,   0,  -4,  18, -49,  48,   0,   5,  -7,  14, -16,   3,  38, -77,   0,  -2,   9, -13,  28, -24,  11,  26 },
    },
    {
      {  77, -60,   9,  -3,   2,  -1,   1,   0, -66,  42,   4,  -3,   1,   0,   0,   0,  15,   4, -15,   4,  -1,   0,  -1,   0,  -3,  -7,   6,   2,  -1,   1,   0,   0,   2,   2,  -1,  -2,   0,   0,   0,   0,  -1,  -1,   0,   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -56,  -4,  33,  -4,   4,  -1,   2,   0,  -8,  65, -44,   0,  -2,   0,  -1,   0,  42, -52,  -1,  19,  -6,   4,  -1,   1,  -8,   2,  24, -13,   0,   0,   0,   0,   4,   1, -10,  -1,   4,  -1,   1,   0,  -2,  -1,   4,   2,  -1,   0,   0,   0,   2,  -1,  -2,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
      { -53,   7, -10,  15,   1,   3,   0,   1, -29,  19,  42, -22,   2,  -4,   1,  -1,  10,  42, -63,   5,   6,  -1,   1,  -1,  21, -42,  10,  27, -12,   4,  -2,   1,   1,   4,  15, -18,  -1,   2,  -1,   0,   3,  -4,  -6,   3,   4,  -1,   0,   0,   0,   1,   2,  -1,  -2,   0,   0,   0,   1,  -2,  -1,   1,   1,   1,   0,   0 },
      {   0,  63, -44,  -4,  -2,   0,  -1,   0, -66, -10,  12,  28,  -5,   4,  -1,   2,  53, -16,  23, -25,  -4,   1,  -2,   0, -12,  -1, -13,  -2,  13,  -3,   2,   0,   6,   1,   4,   8,  -7,  -1,   0,   0,  -1,  -2,  -1,  -5,   1,   3,   0,   0,   2,   0,   1,   2,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0 },
      { -44, -38,  41, -14,  10,  -1,   4,   0, -21,   8,   8,  15,  -2,   0,   0,   0,  -1,  35,  21, -40,   2,  -3,   1,  -1,  -5,  27, -61,  14,  15,  -5,   3,  -1,  14, -18,   8,  26, -20,   3,  -1,   1,   2,   0,   4, -15,   1,   5,  -2,   1,   2,  -2,  -1,   4,   3,  -2,   0,   0,   0,   1,   0,  -2,  -1,   0,   1,   0 },
      { -31, -30,   2,  25,  -7,   5,  -1,   2, -29, -28,  43, -14,  11,   0,   2,   0,   3,  10,  27,   9, -13,  -2,  -2,   0,   5,  18,  21, -58,   6,   3,   0,   0, -10,  25, -48,   9,  28, -12,   4,  -2,   6,  -4,   0,  20, -18,   0,   2,  -1,   1,   2,  -2,  -7,   2,   4,  -2,   0,   1,  -1,   0,   4,  -1,  -2,   0,   0 },
      {  -1, -34, -28,  27,   3,   4,   0,   2,  36,   2,  45, -21, -17,   1,  -4,   0,  36, -38, -10, -31,  32,  -3,   3,   0, -47,  28,   2,  28,  -1, -12,   4,  -3,   8,   9,  -8,  -9, -13,   9,   0,  -1,  -6,  -2,   0,   5,   8,   0,  -3,   1,   3,   2,  -1,  -2,  -4,  -2,   1,   0,  -3,   0,  -1,   1,   2,   1,   0,   0 },
      {  20,  24,  -1,   2, -14,   2,  -2,   1,  28,  25,   3, -35,  17,  -8,   2,  -2,  17,   9, -34,  -3, -12,   1,   1,   1, -15, -17, -25, -13,  38,   3,   2,   1, -10,   4, -20,  64,  -7, -18,   4,  -2,   7, -17,  25,  -6, -32,  18,  -2,   1,  -2,  -1,   1,  -8,  16,   2,  -5,   1,   0,  -3,   3,   2,  -4,  -4,   1,   1 },
      {  -1,  33,  -6,  22, -11,  -1,  -3,  -1, -37,  16,  -3, -44,  13,   1,   3,   0, -45, -11,  -2,  31,  24, -10,   5,  -3,  -6,  58, -13,  13, -31,  -2,   1,  -2,  21, -26, -17,   7,  -7,  15,  -5,   3,   4,  -6,  20,  -4,   0,   1,  -3,   2,   4,  -2,  -6,  -5,   5,  -2,   1,   0,   0,   0,   2,   4,  -2,  -1,   0,   0 },
      { -11, -23, -66,  45,  -5,   8,   0,   3,   7,  50,  -6,  13, -18,   1,  -4,   0, -36, -12,  30, -16,  -4,   7,  -2,   2,  47, -15,  -8,  -9,  15,  -4,  -1,   0, -10, -11,   9,  10,  -6,  -4,   3,   0,   5,   2,   7, -10,  -2,   6,  -1,   0,  -2,  -2,  -1,  -1,   6,  -3,  -1,   0,   1,   1,   2,   0,  -1,  -1,   1,   0 },
      {  14,  27,   9,  -9,   6,  -6,  -1,  -3,  15,  38,  11, -13, -23,   7,  -3,   1,   5,  11,   2, -34,  15,  -7,  -1,  -2,   9,   0, -33, -20, -10,  11,   3,   1,  -7, -24, -14, -10,  57,   4,  -5,   1,  -8,  14, -12,  47,  -2, -35,   6,  -2,   1,  -8,   2,   0, -27,  14,   5,  -1,  -1,  -1,   1,   0,   8,   2,  -4,   1 },
      {   1,  24,  19, -33,  22,  -2,   1,  -2, -13,  18,  17,  10, -23,  -9,   0,  -1, -50, -16, -13, -28,  17,  23,  -2,   4,  10,   8,  24,  10,  30, -33,   2,  -2, -18,  37, -36,  -3, -25,  -1,  14,  -5,  11, -14, -11,  19,   6,  13,  -9,   1,   1,   0,   6,  -1,  -8,  -6,   1,   2,   2,  -3,  -2,   0,   2,   4,   1,  -1 },
      { -12, -15, -10,   0,  18, -10,   4,  -1, -20, -24, -26,   3,  -6,  20,  -1,   4, -14, -18,  -6,   3,  40, -12,   3,  -1,   0,  -8,   1,  33,   2,   2,   2,  -3,  -3,  15,  17,  26,  12, -30,  -2,  -3,  -3,  16,  -6,  22, -62,  -2,  16,  -5,   6, -10,  13, -24,  -4,  37, -14,   2,   1,   2,   1,  -2,  12, -11,  -5,   3 },
      {  11,  12,  49,  36, -30,   5,  -7,   0,  -7, -16,  -3, -43,   8,   7,  -1,   4,  21, -22,  29, -15,  32,  -7,  -3,  -1,  43, -23, -12,   9,   1, -22,   8,  -2, -35,  18,  14,  -1, -15,   7,   5,  -4,   4,  13, -14, -11,   6,  -2,  -5,   1,  -6,  -3,  -7,  11,  -2,  -3,   6,   1,   2,   2,   1,  -2,  -2,   3,  -1,   0 },
      {  14,  25,  30,  19,  -6, -16,   2,  -2,  14,  30,  30,  14, -21,   9,  -9,  -1,  -2,  12,  20,  -8, -13, -23,   5,  -1,  -5,   6,   9,  -4, -26,  16,  -3,  -1,  13,   9,   4, -28, -15, -20,   2,   4, -11,  -5, -27, -20, -26,  40,  17,  -3,   6, -10,   6, -13,  39,  18, -24,   3,  -2,   1,  -6,   3,   7, -19,   4,   3 },
      {   6,  -3, -13,   0,  25, -15,   1,  -2,  15,   0, -18,  21, -20,   8,   5,   1,  40,  37,   7,  21,  18, -12, -11,  -2,  25,   5,  -5,   1, -42, -28,  14,  -3, -23,   4, -34,   7, -20,  37,   5,  -5,  16, -24,  21,   7,  -9,   8, -18,   7,  -3,  -6,  10, -24,   4,   6,   4,   4,   1,  -1,   0,   2,  10,  -2,  -1,  -2 },
      {  -6,  -6, -31, -14, -31,  16,   0,   3,   4,  21, -34,  -8,  50, -15,  -1,  -1,   6,  29,  16,  -5,  -2, -16,  12,  -5, -14,  18, -14,  21,  -5,   5,  -4,  -2, -20,  35, -14, -18,  -8,  -7,   4,   1,   3,  -4, -44,  28,   6,  17,   3,  -1,   9, -15,  21,  17, -10,  -2, -11,   2,  -1,  -2,  -1,  -8,  -3,   0,   2,   1 },
      {  -3,   7,  13,  45, -20,  -2,  -4,   0, -11,   6, -31,  13, -15,  17,   2,   3, -29,  26,  10, -23,   6,  -6,  -7,   0, -63, -36,  10,  -2,  -2,   6,  16,   1,  13,  35,  -3,   6,  11,  24, -17,  -5,  13, -20,  10,  -2,   7, -20, -10,   8,   0,   7,   2,  -5,  -3,  -4,  11,  -2,   1,  -1,  -2,   0,   0,   1,  -3,  -1 },
      {  -5,  -1,   0, -24,  12,  -4,   2,   0,  -1,  13,  39,   9,  10, -10,   3,  -4, -16, -38,   9,  18, -24, -14,   8,  -2,  -9,  -4, -29,  11, -25,  32,  10,  -4, -50,  34,  22,  12,  11,  22, -27,   2,  32,  -4,   4, -16,  -3, -17,  -8,  10,  -5,  -3,   5, -10,   3,   2,   9,  -1,   5,   4,  -3,  -1,   3,  -3,  -2,   0 },
      {  -7, -13,  -6, -16,  -2,  -8,   9,  -2, -13, -25, -22, -20, -19,  22,  -7,   4,  -6, -22, -16, -24,   6,  -4,  12,   0,   3, -19, -17,  -7,  -8,  29,  -1,   2, -13, -12, -13,  -9,  22,  13,   7,   1,   2,  -8,   7,  17,  22,  38, -18,  -3,   9,  -2,  21,  -2,  44, -37, -23,   9,  -5,   9,  -8,  13, -13, -23,  21,  -4 },
      {  -6,  -1, -25, -40,  37, -28,   6,  -1,   3,  18,   8, -39,   5,  32,   1,   1,  -6,  18,  29,  16,  27, -23, -15,   1, -21, -20,  11, -24,   5, -17,  17,  -1,   4,  11,  12,  10,  -4,   1,   0,  -7, -12,  18, -19, -21,   6,  -2,   3,  -1,  -3,   5, -22,  28,  14,  -8,   0,  -3,   2,   0,   3,   8, -20,  -3,   1,   0 },
      {  -3,   1,   3,  27,  16,  -3,  -5,  -2,  -9,  -5,   3,  -1, -13, -24,  16,  -2, -13,  -4,  -7,  24, -23,  17,  -7,  -2, -44, -15, -22, -18,  -6, -24,   8,  12, -28, -22,  14,   2,  -9,  21,  38, -12, -12,  41, -21,  16,  -2,  27, -21,  -8,  12, -21,   8,   7,  -8,  12, -10,  11,   1,   2,  11, -11,  10,   1,  -4,   1 },
      {   3,   8,  14,   3,   1,   6,   3,  -2,   8,  17,  27,  26,  35,   7, -14,   2,  -2,   5,  27,  12,  18, -17,   5,  -9,  -8, -15,  18,  17,  -1,  -5, -20,   1,   2, -13,   9,  10,   5, -14,  13,  -6,  -7,  23,  23,  22, -12,   0, -34,  -4,  11,  -3,   8, -30, -15, -50,   5,  21,  -8,   6, -15,  -2, -17,   9,  33, -11 },
      {  -7, -10,  -4, -44, -40,  25,  -9,   4,   0,   1,  17,   3,  24, -19,  10,   0, -15, -20,  17, -18,   8, -14,  -1,  -1,   0, -46,   6,  -3, -26, -14,  21,   2,  24, -19, -20,   3,  -8,  28,  11,  -3, -27,  -2,  19,   4,  -5,   9,   6,  -1,   5,  12,  -5, -14,   4,  30,  -7,  -6,   0,  -3,   2,   1,  21,  -7, -14,   5 },
      {   5,   0,   8,  32,  32, -31,  15,  -2,   2,  -8,   7,  21,  24,   4, -11,  -4,  -7, -28, -10,  10, -13, -40,   6,   7,  14, -26, -26,   5,  -4,   6,  16,  -4,  24,   3, -32,  10, -16,  -2,   0,   9, -18, -19,  -4,  20,  15,   7,  18,  -5,   5,  12,  12,  33, -23,  -3,  -6, -10,   3,  -9,  11, -17, -17,  15,  -2,   4 },
      {   5,   1,  -2,  -1,  32,  -5,  -1,  -3,   4,  -2,  -1, -45,  -3, -19,  16,  -1,   7,  -4,  37,   7, -29,  34,  -6,  -4,  18, -16, -17,  22,   5,   7,  -8,   9,  61,  32,   3,  -8,   6,   8,  -1, -12,  -8, -17,  -6,   9,  -7,  -7, -24,   5,  10, -10,  17, -19,   0,  -4,   5,  12,  -3,   2,  -4,  -6,  14,  -7,   2,  -3 },
      {   2,   1,  -3,   1, -15,   2,   2,   0,   4,  -1,  -3,  21,   0,   0,   2,  -1,   1,  -2, -34, -14,  -1, -13,   6,   1,  30,  25,  15, -14,  -2,  18,  29,  -5,  26,  27,  -1,  31,  -1,  29, -19, -23, -29,  40, -23, -10,  -7,  -6, -36,  15,  23, -31, -12,   8,   6,  -7,   5,  13,  -5,  -3,  11,  -1,  -2,   0,   3,  -6 },
      {   0,   4,  -3,  10,  42,  -1,   6,  -3,  -8,  -6, -23, -15,   3, -52,  -5,   3,  -1,  10,   8, -37, -10,  -7,  34,  -5,  -7, -11,  15,  16, -24,  15, -15,  -5, -24,   2, -27,  -1,  -1,  -9,   1,  10, -24,  23,  16, -18,  -8,  -1,   7,   1,   2,   0, -38,  -3,   9,   1,  18,  -1,  -5,   1,   2,  26,  -7,  16,   1,  -8 },
      {   4,   8,  10,  13,  23,  18,  -7,   4,   3,   6,   6,   5,  12,   5,  16, -12,   6,  16,  18,   3,   6,  27,  -8,   2,   3,   4,  22,  26,  11,  15,   3, -19, -20,  -2,  -9,  13,   8,  15, -31,   3, -19,  17,  26,   4,  21,   0,   4, -10,  16,   9,   0,  18,   1,   3, -54,  -4,   4, -12,  11, -16,   2, -49,  -5,  28 },
      {  -6,  -4,  -8, -17, -46, -49,  27,  -7,  -4,   0,   1, -21, -31,  29, -14,  -1,  -1,   9,  10,   7, -37,   8,   7,   8,   0,   5,   2,  23,   3,   6,   0,  -7, -18,  14,  -8,  -2, -12,   1,   3,   4, -28,  18,  30,   1,   1,   4,   0,  -9,  10,   8, -15, -12, -31,   1, -10,   4,  -9, -13,   1, -18,  11,  24,   3,   5 },
      {  -2,   0,  -1,   8,  -5,  -6,  -8,  -2,   0,   2,  -1,  21,  20,  12,   8,  -3,  -6,  -6, -14,  14,  -1,  17, -37,   3,  -1,   6, -39,  11,   7, -27,   2,  11, -23,  19, -11, -30,  19, -15,  -4, -11, -62,  -9,   8,  -7,  -5, -12,   6,  22,  18,  13, -10,  -7,  25,  -3,  17,  -5, -17,   2,  -8,  28, -10, -13,   1,  -6 },
      {  -2,  -3,   2,  -8,  -7,  27,   1,  -2,  -3,  -5,   9,  -6, -33,   8, -12,   5,  -2,   2,  15,  28,  14,   6,   7,  -9, -11, -24,   2,   2,   8,  16, -25, -10, -15, -26, -28,  13, -24, -11, -35,  14, -29, -15, -20,  -8,   5,  -7,   1,  33,  21, -38,  28,  10,   2,   0,  29,  -1,  -5,   8,  15, -12,   4,   7,  -5,  -8 },
      {  -1,   0,  -1,   7,  15, -32,   7,  -1,  -4,  -3,  -4, -21,  40,   9,   5,  -7,  -3,  -1,   0, -38, -32, -10, -18,  10,  -1,  11,  39,   1,  15,  -8,  27,  -6, -22, -44,  10,  -6, -12,   9, -20,  -9,  -2,  -9,  -7,   3,  -3,  -7,   4,  25,   3, -11,  35, -22,   2,   6,  16,  -8,  -7,  13, -15,  -8,  15,  -3,   6,  -6 },
      {   1,   2,  -3,   3,  29,  44,   8,  -2,  -3,  -3,  -9, -10,  23,  40, -41,   1,  -6,  -5, -17,  -9, -13, -11,  -3,   1,  -5,  -4, -20, -28,   7,  -8, -15, -15,   0,  11,  -2, -27, -22,   8,  -8,  17,  -6,   3,   5, -11,   2,  -8,  -9, -10, -10, -11, -20, -24, -11,   8, -10,  28,   4, -16,   8,  -4,  43,  -5,  16,   6 },
      {   0,  -5,  -3,  -4, -12, -19,  -2,   3,   3,  -4,   3,   7,  11,   6,  -7, -10,  -1,  -8,  -2,  -8, -13,  13, -34,  15,  -3, -18,  -2,  27, -16, -24,  -5,   6,  14,  -9, -41,  -3,  21, -13,   1,  14,  51,  22,  -1, -25,  -3, -10,  13,  -2,  17, -55, -21,  15,   4,  -5,  -3,   6,   5,   2,  13,  -3,  -1, -12,  19,   0 },
      {  -1,  -1,  -3,  -8,  -1,   5, -39,  17,  -4,  -6, -10, -13, -33, -43,   5, -12,  -1,   2,  -6,   4,   0, -36, -13,  16,   1,   0,  -9, -25,   3, -14,  19,  -8,   3,  15,  16, -13, -18, -14, -20,  18,  -2,   9,  22,   5,   1, -10,  26,   4,  25,  -8,   8, -15, -16, -20, -10,  -5, -17,  -2, -15, -20, -13, -14,  40,   7 },
      {  -4,  -5,  -5,  -2,  -6, -33, -25,  -2,  -2,  -1,  -3,   3,   9,   0,  42, -10,  -5,  -9, -10, -26,  16,  16, -22, -19,  -1,  -1,   2, -21, -37,  14, -40,   7,  -9,  -1,  -2,  17, -12, -14,  -5,   6, -24, -20, -15, -21,  -4,  -8, -11,  -5, -27,   7,   5,   5, -21,  -6, -11,  32,  11, -15,   1, -21,   6,   3,  26,   9 },
      {  -2,  -4,  -1,   2, -13, -13,  -5,   8,  -3,  -4,  -1,   6, -11, -23,   1, -10,   1,  -1,   8,  24,   0,  -6,  11,  10,  -6, -10,  -7,  22,  34,   3,  28, -25,  -2, -31, -20, -18,   0,   5, -27, -15,   2, -11, -38,   9, -21,  -6, -16,   1, -45,  14, -35,  -9,   4,  -7, -12,  29,   8, -16,   6,  16,   2,  -7,  27,   3 },
      {   0,  -1,  -3,   6, -12, -49,  26, -15,   4,   2,   2,  12,  26, -22,   4,   2,   2,  -1,  -4,   7,  38,  32,  21,  -6,  -2, -13, -16, -35,  15,   4, -24, -14,  11,   5,   3, -14, -29,  17, -23,  21,  18,  24,  -2,  17,   8,  -5,  12,   7,  15,  -3,  -7, -24,  12,  -1,   0, -16,  -8,  -3,  -7,   9,  13, -14,  -9,   4 },
      {   2,   2,  -1,   1,   5,  16, -31,  16,   2,   2,  -4,  -3,  -6,  36,  30, -14,   2,  -2,  -3,   1, -19,  -1, -12,  -1,   1,  -4, -10,  11,  -3,  17,  -9,  -1,  10,   0, -30,   3, -26,  -3,  -9,   6,  16,  62, -16,  13,   5,  10,   9,  20,  -9,  36,   8, -21,   6,  -3,  10, -16,  26,  -4, -29,  13,  -5,  17,  -6,  -5 },
      {   1,   1,   2,   2,   8,  -7, -21,  35,   2,   4,   4,   8,  10,  22,  60,  -4,   1,   2,   3,   9,  18,  -9,  29,  10,   2,  -1,  -2,   2,  35,  24,   7,  -8,  -4,   4,  -4, -19,  10,  10,  21,   4,  -6, -13,  19, -15,  14,  10,  11, -23,   6, -43,  -6,  -3,  -3,  11,   4,   1,  -7,   8,  -1,   9,  23,  30,  10, -17 },
      {  -1,   3,  -1,   4,  -5,   8,  33,  -2,  -2,   3,   1,   3,  -3, -13,  -6,  -8,  -1,   5,  -3,  10,   6,  13,   4,  14,  -2,   2,  -4,   3,  -6, -17,  14, -11,   9,  17, -13,  19,  17,   6,   2,  23, -28,  25,  -3, -22,  17,  -7,  21, -18, -65,  -9,  42,  -5,   4,  -3,  -4,  -1,   0,  39, -24,  -3,  10,  -5,  18, -16 },
      {   2,   2,   4,   2,   0,  16,  30,  14,   1,   1,   2,  -1,   2,  -1,   2, -17,   2,   3,   8,  -2,   6,  32, -27,  39,  -1,   3,   5,  -3, -26,  17,  19,  -8,  -3,   2,  25,  25, -13,  -1,   3,  18,  -3, -17, -10,  35,   6,   4,  27,  -5,  23,   9, -14,  -3,  14,   0,   7,  13,  19, -15,  23,  21,   4,  22,  33, -35 },
      {   1,   0,   1,   3,   4,  19,  17,  -2,  -1,  -3,  -2,  -1,  -3,  10,  10, -21,  -1,  -4,  -1,   1,  -2,   9, -15,   5,   0,  -7,  -2,  -1, -17,  10,   4,  -9,  -1,  -8,  -4,  21, -11,  -8,   2,  -8,   8, -17, -33,  -1,   8,  -4,  -3, -18,   3, -10, -27, -19, -19,  -6, -19, -17, -77,  17, -35,  18,  -8,  13, -14,  33 },
      {   1,   2,   1,  -3,   2,  26,  51, -17,   0,   0,  -1,  -3,   1,  -1,  38, -11,   0,  -2,  -1, -11,  11,   6, -19,   3,  -3,  -3, -10, -13,  -6,  10,  14, -16,   3,   1,   0, -43,  -6, -16, -20, -18,   7,   9,  33,   0, -33,  23, -13,  19, -23, -15,   0,  13,  -6,  -6,  -3, -11,   5,   1,  13, -27, -28,  22,  -6,   4 },
      {  -1,  -1,   3,  -2,  -6,   2,  -3,  10,  -1,  -1,   3,  -1,  -5,  17,  15, -20,  -2,  -4,   0,  -8, -30, -12,  10, -19,   1,   2,   9,   5,   1, -33, -34,   8,   3,   6,  20,   7, -11,  34, -13,  25, -16,  -9,   5,  43, -19, -14,  10, -11, -32, -37,  -9,   2,  22, -10,  -6, -20,   3,  -1,  24,  22, -33,   1,  -3,  14 },
      {   0,   0,   0,   2,  -2,  10,  36, -34,  -1,  -2,  -2,  -1, -10, -20,  22,  15,   0,   0,   2,   6,   7, -44, -38,   6,   1,   5,   6,  11,  27,  23, -15,  42,  -7,  -2,  -7,   6,   0,   8,  19,  -2,  -5,   4, -11,  -8,  25, -15,   9,  -9,  -6, -14,  -1, -34,   8,  -6,   0, -13,  15, -43,   3,   1,   5,  -8,   0,   9 },
      {  -2,  -2,   0,  -2,   0, -17, -32,  10,  -3,  -2,   0,   1,   5, -14, -50, -30,   1,   2,   5,   2,  18,   6, -31,  31,  -3,   1,   0,  -3,   1,  40,   0,   7,  -1,   4,   4,  -5, -17,  -6,  14, -14, -15,  -2,  21,  16,   5,   3, -15,   4, -29, -37,  -5,   4,   2,  -4,  -3, -10,  27,  10, -17,  12,   4,   9, -35,  19 },
      {   2,   0,   0,   1,  -1,   3,  30,  32,   0,  -3,  -4,  -2,  -3,  -3,  -2, -23,   1,  -2,  -4,   0,   2,   6,  19, -11,  -1,  -2,  -3,  -8,  -4, -25,   3,  -9,  -2,  -6,   3,   2,  19,   0,  -7,  -3, -12, -18, -16, -27,  -2,   2,  -4, -11,  36, -10,   1, -22, -16, -10,  -6, -11,  70, -14, -43,   6, -17,   8, -10,  26 },
      {   1,   3,   0,   1,  -6,  -4,  36,  41,  -1,  -1,  -4,  -1,  -9, -13,   3, -58,   0,   0,  -2,   0,   3, -34, -21, -10,   2,   1,   1,   0,  17,   5, -33,   8,  -1,   1,  -5,  -7,  -7,   5,   5,  -3,   8,   5,   3,  -1,   3, -14, -15,  -5,  18,  30,   1,   8,   2,  18,   3,  29,  -4,  49,  17,  -1,   8, -17, -12, -12 },
      {  -1,  -2,  -2,   1,  -6,   6, -14, -79,   1,  -1,   1,   0,  -4,   2,  14,  -1,   0,  -1,  -1,   2, -10, -10,  -6,   0,   1,  -2,   1,   0,   8,   1,  -6, -22,   0,  -2,  -1,  -2, -15,   6,  -4,  -5,   1, -10,   7,   7,   0, -15,   2, -18,  23,  -7,  -8,  16, -10,  21, -20,  21,  33,  43, -31,  28, -14,   2,   7, -22 },
      {   1,   1,   1,   2,   0,   3,  23,  32,   0,   0,   2,   1,  -1,   7,  -1,  33,   0,  -1,   0,   3,  -5,  12,   8, -42,   0,  -1,   0,  -3,  -6,  14,  21,  46,  -2,   0,  -3,   2, -32, -14,   1, -24,   1,  -2,  14,  22,  14, -14,  16,  24, -13, -19, -28,  11, -11,  18, -10,   7,  11,  21, -30,   2,  -4, -27,  30,  -4 },
      {   0,   0,   1,   3,  -1,  -5,   6, -14,   0,  -1,   0,   1,  -6,  -7,   9, -23,   0,  -1,   3,   4,  -2, -23,  -3, -15,   1,   0,   4,   4,  -3, -27,  -9,  -4,   8,   7,  19,  20,  28, -28, -29,   5,   3,  -2,  12,  17,  38,  52, -13,  37, -12, -11, -16,  14, -14,   9,   4,   1,  -3, -17, -15,  27,  45,   0,  -1,  -7 },
      {   0,   0,  -1,   1,   7,   3,  -6,   2,   0,  -1,  -2,   1,   3,  -9, -16,  10,   0,   0,  -1,   2,   9,   2, -31, -20,   3,   3,   4,   8,  18,   6, -10,  19,  -2,  -1,   0,   2,  23,  47, -17,  30,  -4,  -8, -11,   3, -25,  31,  18,   6,   5,   7, -35,   8,  11,  -9, -12,   1,  -3,  17, -37, -49,   8,  35,  -2, -19 },
      {   0,   0,   2,  -1,   1,   2,   2,   4,  -1,  -2,   2,  -2,   1,  -5, -23, -25,   0,  -1,   4,   1,   7,  10,  -1, -13,   1,  -2,   2,  -5,   9,   3,  -6,  25,   1,  -3,  -3,  -9,   0,  26, -13, -15,  10,  10,  13, -19, -12,  -2,  12,   1,   1,  -3,  43,  22,  -9,  30, -22,  23, -29, -55, -22,  38, -24,  25,  12, -19 },
      {   0,   0,   0,   1,   2,   9,   4, -18,   0,   0,   0,   2,  -1,  13,   7, -36,   1,   1,   1,   1,   3,  -1,  46,  62,  -2,  -2,  -1,  -2,  -6,  -8, -30,  25,  -2,  -1,   1,   4,   0,  16,  15, -28,  -1,  -6,  -2,   2, -17,  -9,  10,  28,  -2,   6, -16,  23,   4,  -4,  10, -13,  -3, -10, -30, -31,  13, -24,  20,  -2 },
      {  -1,  -1,  -1,  -1,   3,   1, -16, -25,  -1,  -2,  -2,   0,   2,   5, -10, -56,   0,   0,   0,   0,   3,  10,  22, -38,  -1,  -2,  -3,  -2,  -5,  -6,  27,  23,  -3,  -1,  -4,   0,  -1, -23, -10, -25,   0,  -2,  -4,  -1,  22, -11,  15, -18,  -3,  -6, -11, -34,  17, -13, -16,  -9,   0,  -6,  22, -35,  -2,   7, -18, -47 },
      {   0,   1,   0,   0,   3,   2,  -6, -18,   1,   2,   1,   1,   4,   0,  -4,  -7,   2,   3,   2,   2,   8,   5,  14,   5,   1,   1,   0,  -2,   4,  -4,   6,  44,   1,   3,  -1,  -1,   8,   6, -15, -13,   3,   4,  -2,  -9,   4,   8,  20,  23,   8,  13,  10, -13,  -2,  -7, -19,   7,  10,  42,  41,  20,  15,  46,  29,  62 },
      {   0,   0,   0,   0,  -1,   5,  -1, -27,   0,   0,   0,   1,  -1,   8,  -1, -45,   1,   2,   2,   2,   5,  11,   3, -39,   2,   3,   5,   4,  10,  24,  41,   7,   0,  -1,   0,  -7,  -2,   9,  37,  54,  -1,  -1,   0,  -5, -23,   7,   0,   8,   2,   3,   0,  14, -17, -10,  40,  -2,  -1,  -7,  -2,   9,   6, -29,  16,  25 },
      {   0,   0,   0,   0,  -1,   2,   8,  -3,   0,   0,   0,   2,  -4,  -4,  12,  -1,   0,   0,   0,   2,  -4,  -5,   3, -13,   0,   0,   1,   3,  -5,  -9,  13, -12,  -1,   0,   1,   3,  -6, -15,   2, -11,   5,   7,  13,  21,   3, -28,  23, -14,   5,  -3,   8,  26,  56, -20,  19,  56,  -5, -17, -30, -16,  28,  26, -22,  40 },
      {   0,   0,   0,  -1,   1,   1,  -1,  -2,   1,   1,   1,   0,   5,   7,  14,   6,   1,   1,   1,  -2,   1,   3,  19,  43,   0,   0,   2,  -3,   0,  -8,  19,  55,  -2,  -3,  -1,  -4,  -4, -14, -31,  50,  -2,   1,  -1,   3,   7,  13, -19,  -9,  -2,  -2,  -5, -14, -10,  10,   7,  51,  -4,   4,  -7,  -3, -31, -12, -42,  -2 },
      {   0,   0,   0,  -1,  -2,   2,   3,   3,   0,   0,   0,   0,  -4,  -1,   5,   7,   0,   0,   0,   0,  -2,  -1,  -5,   3,  -1,  -1,  -1,  -3,  -6,   0,   6,  36,  -5,  -4,  -6,  -6, -16,  -5, -42, -11,  -1,   3,   2,   0, -33,   5,  -8, -61,   7,   7,   9,  32, -13, -40,  12, -32,   4,   4,  -2,  26,  49, -10,  -2, -14 },
      {   0,   0,   1,  -1,   0,   2,   2,   3,   0,   0,   0,  -2,  -2,   2,   7,   7,   0,  -1,  -1,  -2,  -4,  -2,  -1,   0,   0,  -1,   0,  -3,  -2,   3,  -4, -12,   2,   1,   3,   1,   1,  19,  14, -18,   0,  -2,  -2, -10, -16,  11,  59,  31,  -5,  -6,   0,  -9, -40, -56, -12,  37,   1,   3,  13,  22,   7, -24, -46, -18 },
      {   0,   0,   0,   0,   0,   1,  -6,  -4,   0,   0,   0,   1,   3,   3, -10,  -4,   0,   0,  -1,   0,   2,   6,  -9,  -4,   1,   0,  -1,   0,   2,   5, -14,  -8,   2,   3,   0,   4,   7,  17, -24, -43,   0,   4,   2,   6,  15,  46,  33, -43,  -5,  -4,  -5, -16, -12,  21,  66,  18,  -1,   3,   4, -13, -27, -21,  16,  21 },
    },
    {
      { 100, -51,  -5,  -4,  -1,  -1,   0,   0, -52,   1,  18,   2,   3,   0,   1,   0,  -5,  21,   0,  -7,  -1,  -1,   0,   0,  -4,   0,  -9,   0,   2,   1,   1,   0,  -1,   3,   1,   4,   0,  -1,   0,   0,  -1,   0,  -1,   0,  -2,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  51,  23, -30,  -1,  -5,   0,  -2,   0,  64, -61, -15,  11,   0,   2,   0,   1, -35, -21,  38,   7,   0,   0,   1,   0,  -9,  19,  12, -14,  -3,  -1,   0,  -1,  -8,   4,  -5,  -5,   5,   1,   1,   0,  -2,   4,  -1,   2,   2,  -2,   0,   0,  -3,   1,   0,   0,  -1,  -1,   1,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
      {  36,  -1,  -1,  -4,   0,  -2,   0,  -1,  15,  52, -52,  -4,  -2,  -1,  -1,   0,  20, -57, -30,  36,   3,   5,   0,   2, -11, -22,  43,  19, -11,  -1,  -2,   0,  -1,   4,  14, -16,  -9,   2,   0,   0,  -4,   0,   3,  -7,   4,   3,   0,   0,  -1,   1,   1,  -1,   2,  -1,  -1,   0,  -1,   0,   1,   0,   1,   0,   1,   0 },
      {  17,  95, -39, -14,  -5,  -3,  -2,  -1, -56,  -2, -15,   9,   8,   3,   2,   1,  -9,  13,  -6,  13,   1,  -2,  -1,  -1,  15, -10, -13,  13,  -6,   0,   0,   1,   6, -13,  10,   9,  -8,   2,  -1,   0,   4,  -5,   5,  -5,  -5,   3,  -1,   1,   1,  -4,   3,  -2,   3,   3,  -1,   0,   1,  -1,   1,  -1,   0,  -1,  -1,   0 },
      { -18,  -6,  12,   3,   0,   0,   0,   0, -59,  -3, -13,  20,   2,   4,   2,   2,   6, -25,  53,  27, -15,  -1,  -3,   0,   9,  28,  36, -47, -21,   0,  -1,   0,   4,   6, -26, -30,  17,  10,   2,   1,   1,   2,  -7,   6,  17,  -3,  -3,  -1,   2,   1,  -4,   4,   1,  -6,   0,   0,   1,   0,  -1,   2,  -1,  -1,   1,   0 },
      { -18, -21, -49,  26,   1,   5,   1,   2,  -7, -38,  35,  28,  -5,   0,   0,   0,  16,  11,  -9,   6,  -8,   0,   0,   0,   7, -28,  34,  27, -20,   1,  -2,   1,   2,  12,  37, -40, -24,   7,   0,   0,  -2,   2, -11, -26,  19,  13,   0,   0,   0,   6,   0,   1,  13,  -5,  -5,  -1,  -1,   1,  -4,  -1,   1,  -5,   0,   1 },
      {   7,   6, -27,   8,  -3,   1,   0,   1,   5,  -4,  -5,  29,  -8,   3,  -2,   0,  85, -17,  20, -25, -25,   2,  -3,   0, -21, -29, -18, -31,  22,  13,   4,   2, -10,  -2,  10,  26,  24,  -5,  -5,  -2,  -4,   7,   9,   2, -11, -14,  -1,   0,  -3,   0,  -2,  -3,  -4,   2,   5,   2,  -1,   1,   1,  -1,   1,   2,   0,  -2 },
      { -21, -39,  -5,  -6,   7,   0,   4,   1, -13, -17, -46,  33,  17,   3,   2,   1,  -3,  -4,  26,  26,   2,  -8,  -2,  -1, -10,  24, -17,  23,   2, -12,   3,  -2,   9,  -2,  19,  42, -38, -11,   1,   0,   5,   0,  18, -26, -38,  13,   5,   2,   2,  -3,   1, -14,   7,  19,  -2,  -1,   2,   1,   4,  -5,   6,   1,  -6,   0 },
      {   5,  23,  50, -17,  -5,  -2,  -3,  -1, -26, -50,  25, -26,   6,   6,   1,   2,  27, -45,   2, -14,  29,   5,   1,   2, -19,   3,  19,  21,  26, -17,  -2,  -2,  -1,  27,   9,   1, -17, -23,   2,  -1,  -5,   8, -15, -11, -10,   3,  11,   0,   0,   5,  -6,   5,   4,   6,   1,  -4,  -1,   3,  -3,   2,  -2,  -1,  -2,   0 },
      { -12,   2, -50,  16,   2,   4,   1,   2, -14,  46,  27,  13, -17,  -2,  -4,  -1, -42, -37,  12, -35,  -1,  13,   5,   3, -31,  23,  13,   6,  38,   3,  -3,  -2,  16,  24,  -9,   8,  -8, -23,  -7,  -1,   1,  -8, -11,   1,  -9,   0,   7,   4,   5,   0,   2,   4,  -2,   2,   2,  -2,   1,  -2,   0,  -2,  -2,   2,   0,  -1 },
      {  16,  19,  35,  -3,  -8,  -4,  -1,  -3,  15,  24,   1,  21, -24,  -4,  -1,  -1,  12,  -2,  23, -34, -21,   4,   3,  -1,  -9,  25, -29,   0,  -5,   2,   3,  -1,  12, -19,   5, -16, -39,  22,   3,   3,  -4, -11,  -5, -41,  32,  35,  -3,   1,   1,  -1,  -2,  12,  35,  -8, -13,  -2,  -1,   1,   0,   1,   2, -14,  -1,   3 },
      {  -1,   0,  41,  11, -12,   0,  -1,   0, -10,  -3,  -7,  11, -13,   5,  -1,   0, -51, -10,   9,  -5, -22,  15,   2,   4,  17, -39,   7, -11,  16,  34,  -4,   1,   4,  -3,  51,   7,  27,  -7, -23,  -3,   8,  17,   9, -33,  -7, -25,  -5,   7,  -5,  -1, -14, -11,  12,  -3,  12,   5,   2,   5,  -2,   4,   5,  -4,   4,  -4 },
      {  -8, -18, -11, -13,  14,   6,   2,   0, -16, -20, -25,  10,  -2,   6,  -1,   2,  -2, -29,  -3, -23,  29,  15,   2,   2,  10,  14, -22,  33,  17,  -3,  -5,  -3,   8, -30,  14,  -2,   7,  19, -10,   2,   3,  10,  19,  15,  49, -22, -19,  -2,   0,   7,   5,  15, -26, -46,   3,   3,  -3,   0,  -1,  -6, -15,   2,  18,   2 },
      {  11,  -1,  -2,  50, -10,   4,  -4,   1,   1,  -5,  10,  -3, -29,  -3,  -3,   1,  15, -25,  -7,  -3,   6,  19,   2,   3,  82,  18,   5,   0,  12, -12, -13,  -3,  -4, -33, -18,   4, -12, -18,   4,   6,  -3,  -7,  23,   0, -15,   7,  12,   3,  -6,  -1,  -1, -19,   0,  12,   2,  -4,  -1,   0,   0,   0,  10,   5,  -3,  -2 },
      {  -6, -18, -28, -67,   0,   8,   3,   3,   9,   5,  -5, -14,  32,   6,  -5,  -2,   0,  -4,   5,  -2,  -8,   3,  -7,   4,  51, -16,   2, -22,  31,   3, -10,   1,  -2,  -1,  -5,   0,   1, -30,   0,   3,  -8,  -9, -16, -17,   9,  16,  15,   4,  -5,   2,  -1,  21,  33, -11,  -9,  -5,  -2,   3,   0,   9, -11, -19,   5,   3 },
      {   7,   7,   6, -17,  14,  -6,   0,  -2,  14,  24, -12,  24,  -9,  -1,  -5,   2,  14,  36,  29, -14,  19,  -3,  -7,   0,  16,  12,   8,  14,  -3, -24, -11,   2,   4,  -6,  22, -34,  -4, -28,  -3,   8,  -7,  10, -35,  -7,  -4, -39,  26,   6,  -5,  -6, -16,   7, -30,  16,  34,  -3,  -2,   0,  -2,   2,   3,  26,   1, -14 },
      {   0,   1,   6,  41, -12,  -1,  -3,  -2,  -9, -22, -66,   1, -13,  10,   4,   2,   6,  47, -25, -12,  -1,  28,   0,   0, -14,  11,  16,  -3,  32,   9, -16,  -1,  -1,  21, -12,  -5,   5, -24,  -6,   5,   3, -14, -22,   5,   9,   9,   6,   3,   4,  -4,  10,  23,  14,  -7,  -9,  -2,   2,  -3,   4,   2, -11, -14,   1,   4 },
      {   4,   3,  29, -13,  19, -12,   4,  -2,   0, -16,   1,  69,   7, -24,  -2,  -1, -15, -21, -36,  -8, -36,  -7,  10,   3,  23, -16,  -7,   5,   3,  10,  11,   1,   5,  13, -15,   0,  -6,  -5,  10,  -5,  -8,  -7, -21,  31,  -3,  14,  11,  -6,   5,   5,  22,  18, -24,  12,  -5,  -6,   1,  -2,   3, -12,  -9,  14,  -6,  -2 },
      {  11,  16,  20,  10,  -5,   5,  -7,   1,  17,  12,  28,  21,  34,  -8,  -9,  -6,  15,  21,   4,  30, -17,  -2, -10,  -1,   5,  13,  25,   3,  21,  -5, -11,  -2,  13,  19,  -9,  20, -21, -13, -18,  -1,  11, -24,  19, -14,   1, -14, -24,  11,  -2,   1,  -1, -10, -15, -51,   8,  14,  -3,  -8,  -7, -14, -21,   5,  31,   7 },
      {   1,  -4,  -8, -25,  -3,  -2,   5,  -1,   4, -18,   7,  -8, -10,   5,  -1,   0,  21,  -9, -30,   8, -12,   1,   7,  -2,   1,  46,  -5,  11, -19,  45,  -4,  -4,  58,  -2,  -1, -10,  20,   2, -39,  -2,  -8, -29, -10,  -4, -12, -24,  -2,  19,   3, -13,   9,  -9,  11,  16,  13,   8,  -4,   0,   4,   3,  13,  -7,  -4,  -5 },
      {  -4,   4,  -7, -32,  15,  -1,   7,  -1,  -6,  -1,  -1,  10,  -8, -17,   0,  -3,   4,   8, -33, -13, -19,  17,  13,   1,  -1,  49,  17,  -3,  -1,  17,  -9, -10, -74,   9,   5,  -2,  -5,   1,  -8,  -3,   4,  33,   8,  -7,  -5,   5, -12,   3,   0,  -6, -25, -25,   3,  -4,  -4,  10,   0,  -4,  -8,   9,  16,   8,  11,   2 },
      {   2,   1, -11,  41,  -2,   3,  -5,   4,  -2,   0, -12, -13,  41, -23,   3,   2,  -1, -19, -27,  -4,   0, -29,   6,   3,  18,  22, -28, -21,   0,  -6,  25,   1,   0,  30,   5,   6,  -4,  20,   0,  -9, -11,  11, -17, -32,  11, -16,  -4,   6, -12,   0, -33,  31,   4,   2,  27,  -3,   3,   1,   6,  22,  -9,  14,  -6,  -9 },
      {  -7, -10,  -9, -35, -36,   5,   0,   1,  -4, -16, -14, -14, -31, -11,  13,   1,  -5,   6, -10,  -6, -21,  29,  -4,   3,  -3, -22,   7, -15,  13, -21,   6,   7,  19,  18, -22,   3, -26,  26,  18,   1,  18, -24,  12, -14,  13,  -1, -10,   2, -10,   0, -12, -12, -19,   5,  33,  -1,   5,   0,   0,  -1,  18,  44,  -2, -15 },
      {   5,   1,   0,  -8, -14,   5,   3,  -1,  10,  16,  -6,   4, -18,  18,   8,  -1,  12,   6,  25,  -6,  20,   6,   3,  -4,  48,   2,  -4,  19,   3,   4,  16,  -8,   7,  66,  10,  11,   5,  31,  -7, -19,   9,  13, -22,  11, -10,   2, -37, -14,   1, -10,   1,  -3, -11,   5, -29,  12,   8,  -6,   0, -10,   8, -10, -11,  18 },
      {   6,   5,  -1,   6,  92, -13,   0,  -2,  -1,  -8, -16, -29,   2, -36,   2,   2,   5,   9,  19, -27,  -7,   1,  17,   5,   8, -18,  26,   0,   6,   3,   8,  -3,  30,   2,  -8,   4,  -5,   8,   2,  -7,  16,  -7,   7,  -2,  -5,   7, -12,  -2,  -6, -10,  -5, -15,   6,  -1,   2,   4,  -3,   1,  -1,  11,   8,  -6,   3,   0 },
      {  -2,  -2,  -2,   3, -14,  -5,   1,   0,  -4,   2,  -3, -40,   5, -14,   7,  -5,   4,   6,  34,  -4, -34,  -4,  16,   5,   7,   3,  22,  27,   1,  28,  23, -15, -38, -16,   7,  13, -19,   9, -30, -11,  -7, -22,   0,   3,   9, -10,  13,  13,  -2,  22,  18,  31, -23,  14,   1, -14,   3,  -5,   5, -29, -14,   7, -22,  -1 },
      {   0,   0,   2,   4,  40,   4,   3,  -5,   0,   0,  16,   9, -26,   0,   8,  -3,  -1, -10,  11,  52,  10,  33,  -5,  -8,  -4, -20, -41,  13,   7,  -3, -19,  -3, -38,  16, -25,   1,  -5,   7, -23,   9,   4, -18, -14,  -6,  10, -13,  -8,  24,  -4, -12,   3,  20,  17,   5,  21,   3,   7,   2,  18,  13,  -2,  -9, -13, -13 },
      {   3,   8,   2,  -4,  12,   8,   3,  -2,   8,   8,  15,   9,   9, -13,  25,   0,   6,  20,  -3,  35,   0,  36,   8,  -5,  -6,  18,   6,  -4,  36,   0,  -7,  -5,  32,   1,  11,   0,  13,   9,   5,  -7, -30,  31,  33,  -2,  19,   2,   7, -17,   1,  37, -15,   9,  -6,  18, -12, -28,   0,  -1, -10,   3,   4,  -1, -41,   1 },
      {  -1,  -4,  -8,   6, -15, -14,   9,  -3,  -1,  -3,   2, -31, -17, -27,  11,  -2,   6,  -6,  17,  34, -32,   8,  28,  -8,  -2,  -1, -30,  15,  -1,  20, -21,  -3,  22,  -7,   4,  -5,  -5, -27,  18,   4,   6,  41, -30,   7,  -8,  19,  -2, -15,   3, -18, -17,  11, -15, -18, -10,  10,  -3,   1,   4,   1, -17,  22,  31,   4 },
      {   6,   3,   6,   4,  11,  63,  -5,  10,   4,  -1,   7,  -3,  21,  13, -16,  -6,   5,  -6,  -8,   0, -21,  12, -17,   2,  -4,  10,   5,   0,  -4,   1, -10,   3,   7, -17,  10,   7,   0,   3,   2,  11,  74,   9, -16,   4,  11,  -1,  11,  -9,   5, -22, -14,  15, -14,  16, -15, -12,  16,  -8,  19,  -6,   9,   2, -20,   7 },
      {   0,   3,   2,   9,  23,  51,  -1,   3,  -6,  -4, -10, -24,   8,  15, -18,  -8,  -9,  -4,   0,  -7, -25,  12, -28,  -6,  -3,  -3, -15, -15,  -2,  -4, -19,   6,   0,   3,  15, -14, -21,  -1,   1,  -1, -71,  -1,  -8,   5,  -8,   5, -24,   5,  -2,  -2,  12,  -7, -17,   5,  -6,  26,  -3,   2,  -3, -20,  13,  19,   1,  12 },
      {  -1,  -1,  -6,  -6,  25, -40,  -6,  -5,  -2,  -3,   8, -18, -39,   3,  -3,   0,  -1,  -5, -18,  11, -13,   9, -30,   0,   5,  26,  -6, -35, -17, -19,   3,  19,  -2,  19,  50,  17,  -1,  -4,  20,   9,   9, -19,   8,  -6,   0, -14,  14, -13,  -6,   1,   7,  17, -14,  -7, -10, -12,  -2,  -8,   8, -24, -26, -15,   1,  16 },
      {  -1,  -2,   0,   1,  -2, -25,   3,  -8,   3,   2,  15,  10,  16,  24,  25,   3,  -2,   1,  -5, -11,  -3,  35,  -8, -17,  -1,   4,  39,   1, -15, -13,  -6,  -1,   3, -42,   7,  43,   2,  12,  13, -25, -17,   0, -30,   5,  -7,   5, -33,  -4,  -2, -18,  -7,  28,   9,   6,  13,  21,  -2,   5,  22,  20,  -9,   9,  -5,  -4 },
      {   2,   5,   3,  -7,  14,  43,   4,   0,  -1,  -3,   0,  -1, -51,  25, -24,   1,   1,  11,  -1,  15, -23, -35,  -2,   8,   5,  13,  19,  29,   6,  -1,  32,  18,   2,   1, -10,  27,  17,  -4,  25,  -8, -14,  13,   1,   0,   3,   3,  -3,  -3,   7,  12,  -4,   9,  20, -14,  19,  -6,   1,  13,  -4,  18, -25,   3,   3, -27 },
      {   2,   2,   5,   3,  -4,  21,  11,   7,   2,   8,   6,  -1,  17, -19,   7,  11,   6,   1,  10,   7,  21,  10,  33,  -4,   3,  19,  -3, -13,  12,  13,  -1,   9, -15,   9,  50,  -8,  13,  -1,  31,  -5,   7, -34,   3,  15,   4,  20,  -8, -13,   8, -23,  48,  -8,  12,  -6,  21,  -6,   4,  22,   6,  -1,   4,  17,   3, -39 },
      {  -2,  -1,   1,  -9,  -7,  33,  15,   9,  -4,  -5,   3,  14, -34, -36, -17,  -5,  -1,   7,  -8,  18,  46,  -8,  22,   8, -11,  -9,  12, -35,   7,  11,  23, -22,   2, -27,   8,   9, -20,  -5,  -8, -19, -12, -23, -12,   0,  -5, -10,   0, -13, -23, -10, -22,  14,  -8,   1, -19,  10,  -1,  -1,  12,   5,  -5,  -5,   7,  30 },
      {  -3,  -4,  -2,  -4, -12, -11, -32,  -2,  -3,  -4,   1,  -2,  -2, -37,  -1,   2,  -1,  -1,   0,  16, -19, -21, -16,  11,  -5,   9,   2,  22,  53, -28,  13,  20, -11, -26,  -1, -26,  31,  10,  14, -12,  -4,  -5, -12, -17,  -4, -10, -22,  -7,  12, -30,  11,  -5,  -2,  11,  -2,  20,   5, -16,  10, -10,  18, -25, -11,  16 },
      {  -1,   1,   2,   8,  13, -18,  24,  -4,  -1,   0,   3,   1,  13,  52,   5,  -7,  -2,  -2,   1,   9, -13,  -4,  -1,   5,  -6,   5,  -8,  20,  24,  14,  20,  -3, -10,  -2,  -8, -22,  15,  -2,  10,  -5,  -3, -33,   7, -12,   4,   2,   9, -35, -64, -20, -15,  -8,  -8,   2, -19, -12,  -7,   6,   4,   8,  12,  28,  11,  11 },
      {   2,   0,  -2,   1,  16, -22,  15,  -4,   4,  -2,   4,  -5,  -5,  45,  -5,   1,   1,   3, -10,  12,  15, -11,  -1,   8,   1,  -3,  -2, -33,  35,  16,  26,  -6,   4, -15,   1, -33, -27,   3, -13, -22,  15,  23,  12,  -7, -14,  19, -11,  18,  48, -14,   7,  10,  -9,   3,   6,   4,  11, -23,   6, -11, -15,  25,  -5, -11 },
      {  -3,  -1,   1,  -8,  -9,  -8, -45,   7,  -1,   2,   6,   3,   2,   1, -33,  23,   1,   9,   5,  -1,   3,   8,  10,  27,  -3,  -1,  12,   1,  -3,  11,  -2,   6,   8,  14,  -1,   3, -13,  12,  -6,  17, -20,  35,  33,  14,   1,  12,   7,  20, -45, -56,  16,  18,  -5,   3,   1, -12,  -3,  -7,  24,  10,  -8,  -2,  -5,  -8 },
      {  -1,  -3,   0,   1,   6, -15, -47,   8,   1,  -1,   2,   1,   9,  11, -36,  18,   1,   0,   9,   9,   8,  22,  14,  18,   2,  -4,  -4,   9,  -6,  21,   7,   4,  -6,   2,   5,  16,   9,   4,  11,  14, -20, -39,  -8,   2,  11,   6,  10, -16,  49,  10, -43,  -1,  -6,   2, -13,   1,  25, -24,  -8,  13,  18,  17,  15,  -4 },
      {  -1,   0,  -4,   1,   3, -18, -28,  10,   0,   3,   0,  -3,   6,   3, -48,   8,   4,   3,  -3,  13,   7,  26,  -1,  13,  -3,   8, -11,  -6,   5,   9,  12,   1,   3, -11,  -3,  -8,  -5,  -1,   6, -20,  31,  16, -28,  -4,  -8,   4, -14,  -1, -39,  63,  24,  -5,   5,   5,   6,  31,  -5,  29,  -6,  -7,   3,   5,  -6,  -7 },
      {  -1,  -2,  -2,  -5,  12,  28, -59, -25,   1,   1,   0,   2,  -5,  27,  34,  20,  -2,  -2, -10,   0,  -2,  -2,  52,  -6,  -2,  -3,  -5, -14,  -9, -14,  -1, -14,  -5,  -5,  -8,   1,   8,  -2,  -6,  -6,   3,  -3, -18, -34,  -7,   0,  -1, -10,  -2,   2,  13, -16, -17,  -4,  -2, -12, -42, -15,  -5, -16, -20,  -1,   0,  -1 },
      {  -2,  -2,  -2,  -4,  -5, -10,   9, -45,  -2,  -3,  -3,   0,  -4,  -8, -47, -12,  -1,  -2,   2, -10,  10, -29,  10, -16,   1,  -3,   6,   3,   0,  11, -57,   0,  -4,   5,  -4,  -3,  13,   3,   6,   9,   6, -16,  17, -20,  -9,   6, -26, -18,   2,   5,  -9,  15,  -7,  -2,  -7,   8, -12,  -2,  10,   5, -18,  22, -53,  -5 },
      {  -1,  -3,  -1,   5, -10,  -9,   8, -20,  -2,  -3,  -4,  -3,  10,  -4, -32, -21,  -1,  -2,   3,   6,  -6,  24,  -4,  -1,  -2,  -4, -11,  -1,  -5,  -7,  31,   4,   1,  -3,   7,   5, -15, -20,   1, -25,  -2,  -1, -21,  26,  19,  -7,  -6,  10,   5, -29, -20, -34,   6, -23,   1, -31, -48, -10, -25,  -2,   3, -14, -41, -17 },
      {   0,  -2,   2,  -6,  -6,  11, -17,   4,   0,   0,  -4,   0,   3,   8,   6,   4,   0,  -4,   3, -11, -23,  14,  22,  -9,  -3,   0, -15,   7, -17, -21,   6,   0,  -2,   4,  10, -40, -10, -28,   1, -35,  10, -19,  39,  27, -29, -10, -17,  -8,  -9,   3, -17,  21,   5,  -1,  10,  12,  36,  -9, -23,  32, -26, -21, -14,   8 },
      {  -1,  -2,   1,  -2,   1,  10, -34,   3,   0,   1,  -2,   4,  -4, -14,  32,  12,   0,  -3,   6,  -3,  20,  -7, -53, -23,   2,   2,   1,   5,  13,  59,  -5,  30,  -7,  -7,  -1, -14, -21,  18,  15,   0,   9, -10,  -3,  17, -28,   4,  -5,   2,  -6,   6, -20,   2,  -1,  -1,   3, -19, -30,   3, -18,  12, -20,   5,   3,   1 },
      {   1,   3,   2,   0,  -2,   9,  32,  36,   0,   1,  -1,   1,  11, -10, -14,  11,   1,   2,   6, -11,   8,  39,  -6,   6,   1,  -4,  -4,  15, -24,  15,  16,  15,  -5,  11, -26, -12,  14,  -9,  22, -12,  -3,   5,  21, -38, -14, -10,   6, -23,  18,   1,  11,  10,  -5,   3,  13,  11, -44, -35,  30,  -7, -16,  -9,   2,   6 },
      {   0,   1,   2,   1,  -4,   5,  -4, -41,   1,   2,   2,  -1,   3,  12,   6, -12,   0,   3,   4,   4, -12,  -9,  15,   4,   4,   3,   0,  -3,  14,   0,  11,   5,   1,  10,  18,  -6, -25,  -1,   9,  20,   9,  -2,  18,  37,   7, -14,   7,  -1,  10,  25,  -7,  17,   9,  12,   8,  35, -52,  -7,  44,  23,  37, -16,  20,   5 },
      {   1,   1,   3,   2,   0,  -1,  27,  36,   0,   0,   1,  -1,  -7,   3,   2,  51,  -1,  -2,   3,   8, -10, -18,  -1,  22,  -1,   2,  -3,   1,   8, -22, -23,  -8,   0,  -1,  23,   8,  19,   9, -13,   0,  11, -16,  -7,  18,   7,  34,  -2,  29, -13,  10, -11,   5,  -5,  11,   3,  19, -41, -39, -32,  16,  -3,   9,  -9,   6 },
      {   0,   0,   0,   0,  -2,   3,  -5, -15,   1,   1,   0,  -1,   2,  -5, -14,   5,   1,   2,   3,   1,  -8,  14, -13,  21,   1,   1,  -4,   7,  -1, -13,   0, -24,  -2,   7,   6,  -7,  17,  15,  -8, -21,   3,  -6,   6,  -2,  -9,  13,  15,  14,  34,  -7,   9,  -4,  -3,   0,  15, -26, -20,  59,  10,  31, -15,  23,   8,  65 },
      {   0,   0,   0,   0,  -3,   1,  -9,  -3,  -1,  -2,  -1,  -4,  -6, -13,  -7,   0,  -2,   0,  -1,  -1,  -9,  -3, -40,  16,  -2,  -1,  -4,  -1,  -1, -15, -20, -54,   2,   4,   3,  -6,   1,  11, -24, -37,  -1,  -2,   4,   3, -14,   2,   3, -39,   4,   2,   2,  11,  -2,   2, -19, -18, -12,  -5,  14,  10,   7,  -8,  31, -65 },
      {   0,   0,   1,  -1,  -2,  12,  -9, -32,  -1,   0,  -1,   3,   1, -18,  11, -49,   2,   0,   6,  -5,  16,  15, -16,   7,  -1,   4,  -3,   5, -13,  -6,  21,  19,   4,   2,  11,  -3,  37, -14, -14, -10,   2,  -6,   4,  -6,   6,  47,   5,  40, -18,  16,  -3, -13, -12,  11, -14,  10,  11, -33,  22,  13, -23,  19,  16, -11 },
      {   0,   0,   1,  -4,   4, -13, -36,  66,  -1,  -1,  -2,   1, -10,   3,   1, -29,  -1,   1,  -1,   0,  -1, -18,   0, -57,   2,   0,   1,   2,   1,  -6,   6, -24,   1,   8,   4,   0,  10, -25,  -2,  -9,   4,  -3,  20,   3,  27,   8,  -1,  11,   9,  -8, -10,   2,   6,   7, -12,  16, -25,  22,  22,  -5,   4,  22, -13,  -3 },
      {   0,   1,  -1,  -1,  -1,  -1,   9,  18,   1,   0,   0,   3,   4,   5,  17,   2,   0,   2,  -3,  -5,  -3,   5,  16,  45,  -1,  -1,  -1,   1,   1, -12,  -4,  26,  -1,  -2,  -2, -14,   2,   9, -10,  13,  -5, -13,   0,   7, -20,   6,  -7,  13,   8,   3, -50,  -9, -12,  -7, -23,  12, -10,  55,  32, -33, -26, -18, -15, -32 },
      {   0,   0,   0,  -1,  -3,  -2,   7,  22,  -1,  -2,  -1,  -1,   4,  10, -29, -44,   0,   1,  -1,   1,  -8,  14,  25, -38,   0,   1,   5,  -1,  21, -14,   0, -12,   2,  -7,   8,  -9,   7,  57,   8,  28,  -5,  -5, -14,  -1, -43,  10,  18,  13,  -4,   5,  -1,   2, -21,  -9,  13, -15,  -5, -10, -10,  15, -11, -32,  16,  -7 },
      {   0,  -1,  -1,   2,   2,   2,  -7,  -3,   0,   0,   3,   2,   1,  -5,   6, -42,   1,   0,   0,   3,   7,  14,   2,  27,   1,   0,   4,   0,  -6,  -8,  11,  11,  -1,  -1,  -5,   5,   7,   4, -13,  21,   8,   2,  14,  -9,  -7,  -7,   0, -15,  10, -26,  11,  47,  13,  20, -16,  28, -28,  18, -73,  -7,   2,  21,   9,  -9 },
      {   0,   0,   2,  -1,  -2,  -7,  -1,  33,   0,  -1,  -1,  -3,  -5,  -1,   5, -15,   0,   0,   1,   0,  -5,  -6,  12,  14,  -1,   0,   1,  -7,  -2, -16,  -1,  47,  -3,  -7,   3,  -1, -14,  -8, -42,  19,  -1,   3, -15,   1, -11, -14, -32, -23,   7,  17,  29,  -7,   2, -13, -26, -38,  -6, -11,  18,  51,   7,  25,  -7,   4 },
      {   0,   0,   0,   1,   1,  -2,  -3,   4,  -1,   0,  -1,   1,  -1,   5, -16, -15,   0,   1,   0,   1,   6,  -8,  16,  -8,   0,  -1,   1,  -1,   3,  -6, -27,  48,  -2,   2,  -2,   6,   0,   8, -38, -38,   1,   4,   3,  16,  15,  27,   0, -34,  -4,  -1, -20,   3,  15,  29,  47, -21,  -5,  -2,   1, -40,   2,  -9,  31,  16 },
      {   0,   0,   1,   1,   0,  -5, -11,  21,   0,   0,  -1,  -1,  -1,  12,  21, -47,   1,  -1,   2,   1,   1, -12,  -5,  47,   0,   0,  -2,  -3,   3,  21, -26, -15,  -1,   3,   2,  10, -18,   7,  12,  -7,  -4,   1,  -9,   3,  25,   1,  22, -29,   2,  -6,  13, -36, -14,   4,  30,  33,   6,  -9,  -2,  27, -30, -18, -26,   7 },
      {   0,   0,   1,   0,   6,   1, -12,  14,   0,   0,   0,   4,  -6,   5,  14, -31,   1,   0,   1,  -1,   5,  -7,  11,  34,  -1,   1,  -2,   4,  -8,   3, -28,  -1,   3,   1,   1,   0,   9,   9,  29, -36,   0,   0,  -2,   5, -29,  -9,  15,  28,  -7,   9,  -3,  27,  27, -50,   1, -16,   0, -21,   8, -25,  44,  21, -13,  12 },
      {   0,  -1,   0,  -1,  -1,   1,  -3,   4,   0,   0,  -1,   2,   6,  -5,  -3,  -9,   0,   0,   1,  -4,  -3,   5,   1,  15,   0,   1,  -1,   5,   7,  -8,   3, -36,   3,  -3,   6, -10,   2,  10,  13,  33,  -1,   2,  -5,  24,  -5, -27, -28,  -4,  -5,   7,  -8, -19,  64,  29,   3, -17,   0, -26,  12, -30, -47,  30,  11,  12 },
      {   0,   0,   1,   2,  -1,  -1,   7,  -1,   0,   0,  -1,  -1,   4,   2, -11,   4,   0,   0,   2,   1,  -4,   0,  19, -12,   1,  -2,   0,   0,  10,  -4, -29,  32,   0,   4,  -4,   4, -14,  25,  20, -49,   2,   0,   1,  -3,  18, -46,  30,  30,   5,  -6,  10, -17,  10,  29, -51,   9,  -1,   0,  -5,  15, -25,  11,  22, -11 },
      {   0,   0,   0,   0,   1,   5,  -3, -10,   0,   0,  -1,  -1,   0, -10,   2,  20,   0,   0,   0,   0,  -3,   9,  -7, -24,   1,   1,   2,   2,   5, -14,  16,  12,   1,   1,   2,  -2,   2,  26, -39,   2,  -3,   5,  -3,  15, -16,   3,  61, -33,  -1,  -6,   0, -14,  32, -41,  -2,  43,   6,  -8,   9,  -6, -12,  31, -27,   0 },
    },
  },
  { // 5
    {
      { 110, -47,  -9,  -4,  -2,  -1,   0,   0, -39,   3,   7,   2,   2,   1,   1,   0, -16,  12,   2,  -1,   0,   0,   0,   0,  -1,   4,  -3,  -1,   0,   0,   0,   0,  -3,   0,   0,   2,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -32, -26,  26,  -1,   4,  -1,   2,   0, -82,  56,   8,  -2,   1,  -1,   0,   0,  51,  10, -21,   1,  -5,   0,  -2,   0,  17, -22,  -2,   1,  -1,   1,   0,   0,  -1,  -2,   7,  -1,   1,   0,   0,   0,   2,   2,  -2,  -2,   0,   0,   0,   0,   3,  -2,  -3,   2,   1,   0,   0,   0,   1,  -2,   1,   0,   0,   0,   0,   0 },
      {  24, -25,   9,  -4,   1,  -1,   0,   0,  53,  36, -29, -10,  -2,  -3,  -2,  -1,  45, -46, -32,  18,   1,   1,   0,   1, -40, -26,  37,  10,  -1,   2,   0,   1,  -5,  21,   6, -12,   0,  -3,   0,   0,  -7,  -2,  -5,   2,   3,  -1,   0,   1,  -1,   2,   2,   3,   0,  -1,   0,   0,  -3,  -1,   0,   1,  -1,   0,   1,   0 },
      { -31, -81,  34,   9,   8,   3,   3,   0,  15,   4,   6,   2,  -2,  -1,   0,  -1, -10,  -3,  27,  -5,  -4,  -1,  -1,   0, -23,  53,  -1, -22,   3,  -2,   1,  -1,  24,   6, -39,   6,   0,   2,   0,  -1,   2, -15,   8,   9,  -3,   3,  -1,   0,   1,   3,   4,  -6,   0,   1,   0,   0,  -1,   2,  -3,   1,   2,   0,  -1,  -1 },
      { -16, -52,  -1,  15,   1,   4,   1,   2,  31, -45,  37,  -2,   5,  -3,   1,   0,  10,  62, -36,  -6,   1,  -6,   0,  -2,   2, -35,  -6,  25,  -7,   4,   0,   2, -13,  11,  26, -11,  -1,  -2,  -2,   1,  -1,   5, -14,  -3,   5,  -2,   2,   1,   0,   3,   3,   7,   0,  -2,   0,  -1,   2,  -1,  -1,  -1,  -2,   0,   0,   1 },
      {  12,  -9, -27,   4,  -1,   3,   1,   0,  27,  -9,   1,  12,  -8,   5,  -1,   0,  77,  -3,  27, -32,  -6,   1,  -4,   0,  13,  -4, -41, -22,  21,  -4,   3,   0, -18, -31,  -6,  35,   2,   1,   3,  -1,  -6,   9,  25,  -3, -11,   2,  -2,  -2,  -5,   0,  -3, -11,   0,   3,   1,   0,  -5,   3,   2,   1,   3,   0,  -1,  -1 },
      {  -2,   8, -46,  19,   5,   4,  -4,   2, -33, -47,  17,  24,   1,  -1,   1,   2,  43, -14,  10,  18, -10,  -5,   1,  -1,  -8,  -3,  32,   0, -12,   1,  -2,   1,  31,  28, -29, -23,   4,  -1,  -2,   1,  -4, -35,  -1,  20,   5,  -2,   1,   2,  -7,  11,  23,  -6,  -4,  -3,  -1,   0,   2,   2,  -8,  -4,   1,   0,   0,   1 },
      {  10,  23,  -8,   1,  -2,   0,  -4,  -1,  20,  42,  13, -23,  -4,  -6,   0,  -2,  -7,  33, -40,  -8,   3,  -3,   3,  -1,  -8,  -9, -29,  19,   1,   4,   0,   1,  46, -20, -32,  13,   1,   5,  -3,   1,  -1, -49,  25,  23,  -2,   0,  -1,   2,  -8,   8,  27, -14,  -4,   0,  -1,   1,   0,   1,  -7,  -4,   5,   0,   0,   0 },
      { -17, -40, -51,  15,  11,   6,   2,   0,  11,  24, -42,  21,   3,   0,  -1,   0, -27,  -3,  14,  28,  -4,  -2,  -3,   0,  48, -18,  13,  10, -13,   0,  -2,   0,  21, -32,  24,  -7, -12,   3,  -2,   1, -14,  -2,  25, -14,   3,   0,   1,   0,  -8,  14,   4, -10,   2,  -2,   1,  -1,  -2,   5,  -5,  -4,   1,   0,   1,   0 },
      {  13,  -1,  68,   3,  -2,  -6,   0,  -2,  -6, -57, -19, -11,   2,   1,   3,   2,  13, -33, -33,  19,   7,   4,   2,   2,  46,   6,  -3,  17,  -7,  -1,  -3,  -1,  12, -26,   0,   7,  -7,   2,  -1,   1, -12,  -8,  27,   1,  -2,   1,   0,   0,  -8,  13,   8, -13,   0,   1,   1,   0,  -3,   6,  -5,  -5,   4,   1,   1,   0 },
      { -11,  -2,  -8,  -5,   2,   1,   3,   1, -30, -25, -29,  13,   8,   6,   2,   1,  -2,   2,  -6,  -6,   8,   0,   1,   1, -74,   1,  -8,  34,  16,   0,   3,   0,  27,  -3,  36,  35, -25,  -3,  -4,  -1,  14,   1,  26, -30, -18,   4,  -3,   1,   2,   2,  -3, -14,   9,   3,   1,   1,   2,   3,   0,   1,   2,  -3,   0,   0 },
      {  -5, -10, -14,  39,  -7,   6,  -2,   2,  -3,  10,  68, -13, -18,  -4,  -3,  -2, -24, -69,  -9, -34,  13,   6,   5,   2,  12, -13,  -4,  17,  21,  -2,   4,  -1,  11,  13,  13,   8, -12,  -4,  -1,  -3,  -3,   6,   6, -15,  -6,   4,  -1,  -1,  -3,  10,  -6,  -7,   7,   0,  -1,  -1,   0,   2,  -5,   2,   1,  -3,   0,   0 },
      { -17, -21, -28, -36,  16,   4,   3,   2, -10, -10,   4, -17,  17,  -1,   2,   0, -21, -25, -14,  32,  12,   2,   3,  -1,   0, -16, -14,   9,   2,  -1,  -2,   1, -43,  13, -17,  35,  24,  -8,   2,  -2,  16,  11,  19,  44, -26, -12,  -2,  -2,  12,  -6,  15, -16, -28,   8,   1,   2,   0,  -1,   5,  -8,   5,   8,  -1,   0 },
      {  -3,   6,  12,  65, -14,   1,  -3,   1, -15,  -3, -18,   5, -12,  -4,   0,  -1, -22,  11, -11, -18, -11,   3,   3,   3, -33, -22,  27, -34,  -1,  11,   7,   3, -40, -24,  -2,   0,  31,  10,   4,  -1, -11,   4,  34,  21,   4,  -6,  -1,  -2,   9,  24,   5, -12,  -7,  -3,   0,   0,   7,  -1, -10,  -1,   4,  -1,  -1,   0 },
      {   7,  11,  38, -27,   9,  -9,  -1,  -2,  14,  10,  50,  30,   0, -13,  -2,   0,  -7,   1,  44,  32, -33, -11,  -5,  -1, -15, -46,   3,  -9, -24,   4,   1,   4,   4,  -5,  10,   0,  -4,   4,   2,   2,   4,  -1,  24, -13, -15,   2,   1,   2,  -4,   6,   1, -31,   1,   8,   1,   1,  -1,   5,  -9,  -8,  13,   1,   1,   0 },
      {   1,  12, -21,   6,  -2,  -2,   1,   0,   4,  12,  35, -33,  10,  -2,   1,   0,  21,  21, -12,  32,  13,  -6,  -1,  -3,  -2,  61,  35,   8, -22, -15,  -4,  -4, -31, -11,   3,  -2, -18,  -1,   2,   2, -11,   9,  31, -20, -13,   8,   0,   1,   7,  16, -22, -26,  10,   5,   0,   1,   1,  -9, -11,   9,   7,   1,   0,   1 },
      { -10, -10,  -8, -59,  -5,   4,   4,   0, -11, -10,  17, -31,   0,   9,   3,   1,   0, -11,  19, -17,  17,   5,   0,  -1, -27,  -8,   5,   3,   3,   3,   0,  -1,   3, -46,   7, -29,  10,  13,  -2,   4, -39,   8,   7,  11,  41,  -4,   2,  -1,  -4,  36,   5,  18,  -2, -17,  -1,  -4,   0,   6, -11,  -1,  -8,  -3,   3,   0 },
      {  -3,   1,  -7,  -9,   0,  -5,   4,   0,   2,   9,  20,  74,  -7, -10,  -7,  -2,  -5, -23, -47,  13, -23,  -3,   1,   3, -14,  42, -34,   2, -15,   5,   2,   1, -27, -20,  15,  -3,  -5,   7,   0,   4, -15, -13,  -4,   4,  18,   4,   1,   0,   7,   2,  12,  21,  -2,  -9,  -4,  -1,   2,  -2,   6, -10, -11,   1,   3,   3 },
      {  -5,  -3, -14, -48,   8,  -2,   6,  -2,   0, -10,  -5,   6, -10,  -7,   7,   2, -12,   7, -44, -16, -40,  22,   2,   6,  18,  -1,  21, -40,  29,  23,   4,   0,   8,  29,  -8,  11,  10,  -7,   0,  -7, -12,  -3,  14, -19,  -7,   8,  -3,  -2, -11,  21, -21, -13,  27,   2,  -5,  -2,   5,  -8, -16,  24,   9, -10,  -2,  -1 },
      {   4,  10,  16,  -7,  20,  -3,   0,  -2,   4,  16,  -9,  27,  -8,  -1,  -1,   0,   4,  17,  13, -37,   7,   2,   1,  -1,  13,  25,  10,  33,  13,  -2,  -5,  -4,   1,  38,  23,  -4,   3, -17,  -7,  -1, -43,  19,   2,  22, -17, -23,   5,   0,  -7,  26,  14, -21, -41,   7,   2,   0,   3,  -1, -27, -19,   6,  14,   2,   2 },
      {   0,   7,   7,  21,  41,   1,  -2,  -3,  -7,   2, -14, -16,  32,  -8,   3,  -4,  -7,  -2,  18,  12, -10, -14,  -1,  -4,  -6, -16, -39,  12,   6,  -3,   5,   0, -36,  31, -24,   6, -25,  -7,   4,   2, -51, -17,   0, -10,   8,  23,   3,   4, -11,  25,   4,  16,  29,  -5,  -5,  -4,   8,   4,  -7,  15, -13, -11,   1,   0 },
      {  -5,  -6,   7, -24, -31,   2,   3,   2,   0,   5,  -4,  17,  -5,   8,   1,  -1,  -6,   6,  14, -29,  22,  13,  -3,   0,  15,  -7,  37,  13,  14, -12,  -3,  -1, -52,   0,  -3,   3, -29,  -6,   2,  -1,  21, -57,   1,  -7,  -4,  25,   3,   4,  -1,   1,  39,  -4,  19,   1,  -4,  -1,   1,  10,   2, -16,   3,  -9,   0,  -1 },
      {  -1,   1,   7,   2, -62,   0,  -6,   0,   4,   5,  13,  31,  30,  16,  -2,   1,  -3,  10,  -3,  35,  36,   5, -15,  -2,   1, -14,   4, -21,  30, -18,  -8,   2,  22,   5, -11,  23,   5, -33,   4,   3, -40,   5,   2,   1, -12,  -7,  11,   1,  10,  12, -13,  19,   0,  -4,   5,  -1,   4,  -5,   8,   5, -15,   1,   0,   1 },
      {  -4,  -4,  -4,   5, -57,  -4,   4,   4,  -2,   4,  -6, -33, -32,  15,  15,   2,   5,   2,  18,  35, -21,  14,   3,  -2,   1,  17, -37,  -3,  -3,  28,   3,  -1,  -8,  43,  36,  -1,  15,   9, -13,  -1,  -3, -12,  13,  -6,   7,  -8,  -7,   0, -19,   8,  20,  -1,   4,  -4,   0,   1,   5,  15,  -5, -19,  -2,  -1,   1,   0 },
      {   6,   8,   7,  12,  40,   4,   2,  -1,   5,   3,  19, -19,  24,  -4,  -4,  -2,   3,   4,   8,  12, -28,  11,   0,   1,   4,  16,  30, -12,  39,  10,  -9,  -1,   1, -19,  35,  22,  13, -29, -16,   2,  18, -23,   7,  -3,  -1, -24,   5,   5, -15,  -1,  18,  34, -10, -21,   7,  -3,  -3,  11,  13, -12, -35,  -1,   7,   1 },
      {   0,  -2,  -2,  -8, -20,   2,  -2,   3,   3,   1,  -3,  -7,  73, -16,  -5,  -5,   2, -15,  -4, -37,   9, -44,   7,   0,   8,   8,   0, -31, -38,  10,  19,   1,   7,  14,  37,  10,  15,  20,  -4,  -2,  -2, -26,  -2,   1,  -9,  -7,  -7,   4,   0,   2,  -1,   0,   1,   6,   3,   0,   2,   5,  -9,  -4,   7,   1,  -2,  -3 },
      {  -6,  -6,  -8,   0,   2, -15,   5,  -2,  -9,  -9,   1, -20, -15,   2,  -2,   2,  -7,  -4,  -6,  -5, -14,   6,   0,   2, -14,  -1,   4, -15,  -7,  -4,  -5,   3,  -3, -23,  13, -10,   0, -21,   0,   2, -62, -26, -20,  -5, -34,  -1,  21,   3, -20, -49,   4,  -3,  -5,  42,  10,   4,  -7,   7,  29,  -3,  23,  10, -10,  -5 },
      {   1,  -2,   0,  -3, -27,  11,  -4,   4,   0,  -2,   0, -12,  -8, -25,  -1,   2,   3,   2,  16,  -4, -47, -14,  19,   2,   9,   1,  29,  35,  -8,  22,   9,  -9,  11, -11, -10,  46,  -7,   3,  -2, -10, -29,  15, -16,  16,  -5,  18, -11,  -3,  47,  -5,   7,  16,   7,  -3, -21,   3,   7, -20,  14, -12,  -5, -12,   1,   6 },
      {  -5,  -3,  -8,  -2, -24, -41,  -5,  -1,  -3,  -1,  -1,  -8,  14,  -9,   6,   0,  -6,   1, -12, -16, -30,  -1,  -9,   6,  -4,  -6,  14,  -1, -17, -23,   5,   2,  -2,   4, -30,  16, -30,  10,  17,   2,   7,  33,   6, -14,   3,   6,   5,  -4, -61,   6,   7,  16, -29, -11,   6,  -1,  -1,   4,   6, -28, -21,  27,   5,   3 },
      {   0,   3,   4,   4,  -4,  27,   5,   2,   0,   7,   5,  17,   0,   2,  -2,   0,   1,   1,   1,  11,  17,  -7,   5,  -3,  -4,   7,  13,  22,  -8,  27,   2,  -6,  -7, -22, -18,  26,  27,   9, -16, -10,   5,  22, -40,  -3, -17, -13,  -5,   6, -66,  21,  12,   5,  23,  19,  -2,   0,  11,  26,  -2,  21,  15, -18, -11,  -6 },
      {   4,   7,   2,   1,  29,  12,  12,  -5,   3,   4,  11,   0,  -6,  39,   4,  -2,  -1,   8, -14,   0,  17,   4, -14,  -3,   2,   2,  14, -34, -16, -15,   4,   5,  20,  19,  16,  18,  -9,  24,  17,  -6, -19,  35,  12,  13,   3,  25, -15, -11,   7, -17,  45,   1,  14,  -5, -22,   0,   5,  35,  22, -16,  -4, -24,  -2,  -1 },
      {  -2,  -2,  -2,   6,   5, -74,  -7,  -5,   3,   5,   8,  17,  23,   0,  17,   1,   2,   3,  11,  -9,  18,  34,  16,  -4,   4,   1,  11,  31,  -4,  22,  -4,  -2,  -2,   8, -10,   3,  32,  19, -15,  -4, -15,   0,  27,  -4,  15,  -7, -16,   2,   1, -35,  -4,   2,  15, -12,  -2,   2, -13,   6,  15,  18,   3, -14,  -1,  -1 },
      {  -1,   2,   7,  -8,   2,  42,  -5,   4,  -4,  -4,   1,   8, -28, -39, -16,  -1,  -2,  10,   4,   8,  27, -15,   9,   0,   2,  -4,  12,  -3,  13,  -6,  13,  -9,  -6,  18,   7,  11, -11,  27,  -3, -10, -20, -19,  32,  10,  21,  -5, -30,  -2, -38, -35, -30,   7, -15, -17,   3,  10,  -3,  -8,  23,  10,  -7,  15,   4,  -1 },
      {  -1,  -2,  -5,  -2,  10, -62,   3,  -5,  -1,  -4,  -3,  -4, -40, -55,   1,   6,   5,   5,   5,  17,  37, -23,  13,   2,   2,   1,   0, -23,   9,  -4,  19,  -1,   8,  -3,  23,  12,  -7,  -8,  13,  -2,  11,  -2, -15,  24, -12,  16,   8,  -2,   1,  26,   4,  17,   7,  12,  -8,  -5,   6,  11, -12,   5, -10, -11,  -6,  -2 },
      {   0,  -2,   2,   1,  13, -19,  12,  -5,   2,  -1,   2,  -1, -26,  49,   0,  -1,  -3,  -4,   7,  -6,  -7,   8, -25,  -1,   3,  -7,   9,   7, -42, -27,  -2,   6,  -1,   5,  17,  49,   4,  10,   1,   4,  -1, -38, -17,  19,   8, -17,  -4,  -4,  -4,  20, -34,  11,  -2,  -5,   4,   0,  21, -26,  -1,  26,  -4,   7,  -1,   4 },
      {   4,   2,   1,   7,   8,   1,  21,   0,   6,   2,   6,   2,  25,  -7,  30,  -8,   6,   0,   4,  11,   7,  39,   7,   1,   1,  -2,  -2,  -8,  22,  12,  12, -11,   4,  -6,   2,  -9, -23,  32,  -7,  -9,   4,  -6,  -3,   0,   4,   6,  -8,  -6,   6,  11,  -6,  14, -24,  33,  -5,   0,  69, -16,  15, -16,  23,  19, -27,   2 },
      {  -1,   0,  -3,  11,   2,  -3,   6, -14,  -2,  -1,   2,  -3,  17,  -6,   6,   2,  -1,   2,  -2,  14,  10,  38,   4,  -6,  -8,  -5,  -3, -26,   3,  15, -10, -11,  -3, -12,   8,  21, -35,   9, -10, -11, -10,  -4, -51,  15,   9,   4, -32,   1,  -7,   2,  -3, -34,  -6, -19,  -3,  10, -50, -15, -28, -18,  11,   4,  24,   6 },
      {   0,  -3,   1,   0,  -7,   8, -21,  13,   0,   0,  -3,  -2,  11, -24, -20,   5,   2,  -4,   6, -11,   1,  37,  17,  -6,  -5,   2, -11,  -8, -31,  -7, -45,   0,   6,  -2,  20,  -4,  -1, -39,   5,  15,   7,  12,   9,  29,   4,  21,   3,  -1, -29, -16,   6,  -9,  17, -14, -24,   1,  48, -23,  -6,  14,   0, -10,  10,   7 },
      {   1,   3,   2,   1,   1,  21, -29,  15,   0,   3,   0,   2,   8, -19,   7,  11,   2,   5,   5,  -2,  -8,  45,  22,   1,  -1,   3,  -4,   4, -18, -24,  -2,   4,   5,  12,  13,   9,   8,   5,  14,   3,   7,  -5,   1,   6,   8,  13,  16,  -8,   6,  45,  -8,  14,  -7,  18,   2, -11, -53,  22,  44,  12,  19,   9, -23,  -8 },
      {   0,  -1,  -4,  -7,  14,   4, -28,  -2,  -1,  -3,  -4,  -3, -24, -17, -12,  -2,   1,   1,  -7,  -9,  25,  18,  13,  -8,  -1, -10,  -3, -22, -41,   7, -15, -10,  -3,   6, -25,  16,  -8,  -6, -38,   3,  -9,  -1,  -4, -45,  -9, -32,  -6,  16,  31,  14,   2,   6,  -2, -10,  24,   9,  25,  25,   8, -29, -13,   3,   7,  -4 },
      {  -2,  -3,  -2,  -1,   4,  -7, -56,   0,  -1,  -1,  -1,  -5,   2,  -8, -16,   4,   0,   0,  -3,  15, -13,  -1,  -4,   2,  -1,   5,   0,   6,  31, -53,   1,   9,  -1,  -6,   7,  13,  40,  24,  19,  -7,  -8,  -9, -24, -28,  14,   0, -19,  -9,   4, -10,  12,  -4,   1,  -6,  -9,   7,  23,   8, -32, -14,  35,   3,   1,   0 },
      {   1,   2,  -3,   4, -11,  -1,  34,  -2,  -2,  -3,  -3,  -7,   6, -50, -23, -25,   1,  -1,   0,  -5,  -9,  16, -78,  -3,   1,   5,   0,  10,   0,  -2, -16,  21,   1,   6,  -2,   9,  -6,  -1,  -3,   4,   1,   5,   1,  10,  28, -17,   3,  11,  12,  -3,  -7,  -4,   2,   6,   5,   5,   2,  34,   3,   2,  24, -19,  -5,  -1 },
      {   0,  -1,  -2,   0,  -8,   2, -13, -18,   1,   2,   2,   3,   4,  18,   0,  -1,   1,   2,  -1,   1, -15,   0,  27,   1,   2,   1,   6,   5,   4,  17,  -1,   2,  -2,  -4,   2,   6, -26,  -9,  14,  -5,   1,  -3,  -4,  29,  10,  11,   7,  -3,  13, -18, -27,  -4,  -6,  -9,  17,   6,  22,  83, -29,  28,  -3,  35,  -1, -11 },
      {   2,   2,   0,   3,   0,  13,  53,  23,   1,   1,  -2,   5,  -5, -17,  61,   2,  -1,  -2,  -3,  -3,  -3, -11,  25, -18,  -3,  -6,   0,   0,  -8, -39, -34, -15,   1,   0,  -7,  10,  20,  -8,  -5,   4,  -3,  -7,   4, -10,   1,  -2,   4, -11,   1, -16, -20,  18,   1, -12, -14,  -9,  -8,  21, -33, -22,  10,  -7,  -3,   3 },
      {   2,   2,   1,   2,   6,  10,  40,  17,   0,  -1,   0,  -2,   1,  -9, -34,   1,   4,   3,   6,   4,  20,  37,   1,  16,   1,   0,   0,  10, -18,  21,  32,  23,   3,   0,  -7,   8,  38,  -3,  30,   0,  -7, -12,  -8, -35,  -8,  35,  12, -19,   4,  -8,  -2,  -5, -22,  -7,  -7, -10,   2,   4, -22,  -5, -29,  20,  13,   6 },
      {   0,   0,  -2,   0,   4, -14,   8,  -2,   0,   0,  -3,   1,   4,   8, -14,  -6,  -1,   0,  -2,  -3,   0, -22,   6, -11,  -1,   1,  -3,   9,  16,  11, -29,   0,   0,   0,  -2, -11,  22,  -2,  19,   0,   1, -14, -15,   6,  -1,  12, -11, -15,  -8,  29, -46, -39,  15,  -9, -20,  15,  13,  20,  56, -51, -11,   2,   9,   4 },
      {  -1,  -1,   0,  -3,  -5,   4, -24,   1,   0,  -2,  -2,  -1,   2,   1,  -5, -18,   0,  -1,   0,  -3,  -9,  25, -18,   0,  -3,  -2,  -7,   6,  -4, -10,   1,   5,  -2,  -7,  -1, -31, -10,  26, -10, -28,   0, -10,   7,  17, -66,  -5, -16, -20,  -2,   0, -31,  28, -10,   3, -24,   2,  -8,  17, -21,  -4, -30, -41,   4,  12 },
      {   0,  -1,   1,  -2, -11,   6,   8, -15,   0,   0,  -2,   0,   1, -24,  48,  -9,   0,  -3,   1,  -5,  -9,   4,  -1,  -3,  -1,   2,  -5,   3,  -5, -28,  20,   9,   0,  -8,   5, -13,   6, -15,  13, -15,  -4,  -3, -21, -13,  -8, -17, -29, -10,   6,   8,  27, -45, -15, -20,   9,   7,  19,   6,  32,  48, -34,  -7,  10,   1 },
      {   0,   0,  -1,   1,  -9,  -3,  37, -68,   0,  -1,  -1,  -3,   5,   6, -46, -19,   1,   2,   2,   0,  -8,   7,  41,   4,  -1,  -3,  -5,   1,  -7, -29,  22, -10,   0,   3,  -5, -10,   2, -30,  -8,  -6,   7,   2,  13,  -7,   3, -16, -21, -10,  -2,   9,   4,  19,   0,  -2, -22,  16,  -7,   1,   3,  -1,  27,  -6, -12,   5 },
      {  -1,  -1,   0,   0,  -6,  -9,  -3,  -9,  -2,  -2,  -1,  -3,  -4,   5, -17, -26,  -2,  -1,  -2,   2,  -1, -21,  -5, -17,   0,  -3,   1,  -9,   7,   5, -48, -19,  -2,   2,   1,   7,   8,  15, -40,   5,   0,  -5,  -5, -24,   8,  39,   0,  -6,   4,   5,  13, -10, -52,  11, -37, -18,  -9,   5,   1,  39, -16,   9, -21,   4 },
      {  -1,  -1,  -1,  -1,   4,  21, -30, -79,   0,   0,   2,   1,  -7, -12,  43, -18,   2,  -2,   3,  -4,  16,  -2, -19,   7,  -1,   3,   3,   4, -13,  33, -21,   9,   0,   2,   2,  11,  12,  -6,  39,   9,   2,  -6,   6, -11,   6,   1,  15, -15,  -5,  -8,  -4,  12,   0,  12,  -6,  -2,  -6, -15, -11, -17,  -7,   1,  -9,   5 },
      {   1,   1,   0,   2,  -1,  -1,  22, -10,   0,   0,   0,  -3,  -1, -10, -10,  37,   0,  -1,   3,  -3,   0,  -4, -16,   5,   0,   0,  -3,   2,   4,  12, -47,  -7,   1,   1,   2, -13,   2,  26,  26,   5,  -2,   0,   3,  -5, -49,  -3, -27,  -4,  11,  14,  22,  31,  13, -33,   6,  14,  -2,  -4,   6,  23,  21,  45,   5, -22 },
      {   0,   0,  -1,   0,  -7,   6,  20, -33,   0,   0,   0,   0,   5,  -7, -13,   9,   1,   3,   1,   4,  -7,   7,  31,  40,   2,  -1,   1,   2,  14, -22, -24,   0,   2,   8,   3,   5,   2,  56,  -4,  31,  -5,   2,  -8,   7,  -8,   1,  41,  24,   0,  -1,  -3, -16,  -6,   2,  30,  -7,  10,  -2,   3,   2, -17, -39,  26,  -3 },
      {   0,   0,   0,  -1,   1,   4,   2,  -7,   0,  -1,  -1,   3,  -5,  -2,   9, -64,   0,   0,   0,  -1,  -4,   9,   6, -34,   1,   2,   3,   4,   4, -18,   6,  24,   7,  -1,   6,  -4,  23,   9, -16,  12,   0,  11,   2,  15, -17,  38, -16,  47,  -1,   5,   2,  11,  28,  11,  25,  12,  -6, -10,  -5,  -7, -12,  39,  19,  -7 },
      {   0,   1,   0,   2,   3,   3,  -1, -14,   1,   2,   2,   1,   2,   2,  18,  44,   0,   0,   0,  -3,   2,  -6, -21,  58,  -2,   0,  -3,   3,  -9, -11,  -5,   3,   1,  -1,   1,  -4,  19, -27, -36, -36,   5,   0,   8,   3,  12,  47, -26,  30,  -1,   7, -12,   9, -14,  22,   3,  24,   2,   7,   1,  -7,  -1,  -1,  20,  -4 },
      {   0,  -1,   0,   0,  -4,   2,   8, -13,   0,   0,   0,   1,   6,  -5, -18,  40,  -2,  -2,  -2,  -4,  -9,  14, -12, -43,  -1,  -2,   0,   5,  -1, -13,   0, -57,   2,   1,   2,  -1,  12,   7,  23,  -2,   2,   0,   4,   3,  15,  -8, -16,  10,   4,  -6,   3,  -9,  18,  53,  11,   9,  -4,   1, -17,  -4, -54,   3, -16,  -6 },
      {   0,   0,   0,   0,  -1,  -2,   6,   9,   0,   0,   0,   0,   2,  -6,  -8, -24,   1,   2,   2,   1,   2,  -2,  10,  51,  -1,  -1,  -1,  -1,   0, -13, -29,  19,  -1,   1,   0,   2,  -5,   8, -20, -42,  -2,   5,   0,   8,  14, -26,   3, -40,   4,  -4,  12, -13,  45,  23,  -2, -18,  -5, -10,  -8,  -5, -32,  37, -19, -22 },
      {   0,   0,   0,  -1,  -1,  -4,   5,  27,   0,   1,   1,   0,  -1,   8,  -2,  -8,   1,   1,   1,   0,   1,  -9,  30,   9,   2,  -1,   3,  -3,   4,  12, -31,  45,  -2,   1,  -4,   5, -15,  -2,  35,  22,   1,  -2,   5, -13,  15, -23, -44,   6,   3,   7,   7,  16, -17,  54,  -3,  32,  -8,   3,  -7,  14,  -7, -16,  22,  23 },
      {   0,   1,   1,   0,   2,  -3,   5,   5,   0,   0,   0,   0,   0,   2, -11, -13,   1,   0,   0,   1,   0,  -7,   7, -27,   0,   1,  -1,  -1,  -2,   5, -22,  -4,   2,   3,   3,   3,   4,   0,   6, -56,   3,   2,   4,  -3,   8,  26,  22, -44,  -2,   5,  10,  10, -11,  -4,  72,  37,   2, -14,  -1,   5,  13, -26,   1,  16 },
      {   0,   0,   0,   0,  -1,   5,   3, -19,   0,   1,  -1,   0,   3,  -7,   0,  53,  -1,  -1,  -1,  -2,  -1,  -3,   3, -33,   2,   1,   3,   0,   6,  -5,   5,  71,  -2,  -1,  -2,   0,  -4,   7, -45,  31,  -1,   4,  -3,   6,  -3,  10,  -4, -29,  -3,  -3,  -2,  -7,   8,  -4,   9,  15,   0,  -3,  -5,  -7, -15,   5, -41,  19 },
      {   0,   0,   0,  -1,   1,   0,   3,  13,   0,   0,   0,  -1,  -2,   4,  -1, -11,   0,   0,   1,   1,   1,   2,   9,  21,   0,   0,  -3,  -2,  -3,   1, -10,   6,   1,  -2,   2,   1,   0,   1,  25, -20,  -3,  -1,  -6,  -3,  -5,  -1,   6,  59,  -1,  -3,  -1, -11,  -7, -33,  -5,  40,  -6,  -7, -10,  -6, -16, -10, -85,   6 },
      {   0,   0,   1,  -1,   2,   2,  -1,   1,   0,   0,   0,   0,   1,  -5,   9,   5,   0,   0,   0,  -1,   4,   3,  -6,  -1,   1,  -1,   0,  -3,   2,  -5,   9, -11,   0,   2,   1,   2,   2,  18, -12, -12,   2,   1,   2,  -2,   2, -17,  55,   3,   4,   2,   9,  -1,  18,   5, -44,  64,  -4,   8,   4,  15,  -1,  33,  29,  57 },
      {   0,   0,   0,   1,  -1,   1,   5, -10,   0,   0,   0,   0,   0,   2,  -2,  19,   0,  -1,  -1,   1,   0,   1,  13, -44,   1,   1,   0,  -2,   0,   2,  -5,  43,  -2,   0,  -1,   3,  -8,   6,   7, -51,   2,  -3,   4,  -8,   8, -13,  30,  27,  -1,   1,  -3,   2, -15,   4, -37,   3,   3, -11,   1,  -1,  -1,  -8,  24, -69 },
      {   0,   0,   0,  -1,  -1,   1,   4, -13,   0,   0,  -1,  -1,  -2,   0,  -9,  18,   0,   0,  -1,  -1,   0,  -3,   7, -15,  -1,   1,  -1,  -1,   1,  -1, -14,  19,   0,   0,   0,  -2,  -1,   4,  13, -48,  -1,   0,   3,   0,  -1,  -8,  -6,  38,   2,   0,   1,   0,   6,   0,   8, -71,   4,   5,   8,   0,   8,  13,   3,  73 },
    },
    {
      {  90, -62,   6,  -3,   1,  -1,   0,   0, -56,  26,  11,  -4,   2,  -1,   1,   0,   4,  12, -12,   0,   0,   0,   0,   0,  -3,  -3,   0,   3,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
      { -56, -16,  35,  -1,   5,  -1,   2,   0, -38,  75, -27,  -6,   0,  -1,   0,  -1,  42, -28, -21,  18,  -5,   2,  -1,   1,   1, -10,  19,  -2,  -4,   1,  -1,   0,   7,  -1,  -3,  -4,   1,   0,   0,   0,   0,  -2,   2,   1,   0,   0,   0,   0,   3,  -1,  -1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0 },
      { -48,   1,  -5,   9,   2,   3,   1,   1, -53, -11,  60, -10,   3,  -2,   1,  -1,  -2,  64, -28, -22,   6,  -4,   1,  -2,  20, -16, -26,  21,   0,   1,   0,   1,   6,  -2,   9,   1,  -6,   1,  -1,   0,   6,  -3,  -4,  -1,   0,   1,   0,   0,   2,  -1,   1,   0,   0,   0,   0,   0,   3,  -1,  -2,   0,   0,   0,   0,   0 },
      {  21,  78, -43, -10,  -5,  -1,  -2,  -1, -70,  -5, -12,  29,   1,   4,   0,   2,  28, -18,  21,   0, -11,   3,  -2,   1,   3,  -9,   4, -10,   5,   2,   0,   0,   3,  -2,   4,   1,   1,  -2,   0,   0,   1,  -2,   0,  -3,   0,   0,   0,   0,   1,  -1,   2,   0,   1,  -1,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0 },
      {  26,  33, -33,   7,  -4,   0,  -2,   0,  31,  19,  -1, -27,  -1,  -3,   0,  -1,  -1,   1, -68,  25,  12,   0,   2,   0,   6, -46,  29,  37, -19,   4,  -3,   2,  -5,   2,  22, -22,  -4,   4,  -1,   0,  -3,  -3,  -2,   1,   5,  -1,   0,   0,  -4,   1,   4,  -2,  -1,  -1,   0,   0,  -2,  -1,   0,   0,   1,   0,   1,   0 },
      { -17, -23, -13,  27,  -5,   5,  -1,   2, -23, -36,  28,  -3,   5,   3,   1,   0, -15, -12,  14,  38,  -7,  -1,  -2,   0,  13,   2,  66, -29, -26,   5,  -4,   1,  -2,  27, -15, -44,  20,   0,   1,  -1,   2,   3, -14,   9,   8,  -5,   1,  -1,   2,   5,  -3,  -3,   0,  -1,   0,   0,   2,   0,  -4,   2,   1,  -1,   0,   0 },
      {  -5, -27, -28,  17,   4,   4,   1,   2,  20, -15,  37,  15, -20,   3,  -4,   1,  77, -30,   3, -41,   7,   2,  -1,   0, -39,  -8,   9,  15,  15,  -7,   3,  -1,   0,  17,  13, -10, -13,  -1,   0,  -1,  -9,   4, -10,  -8,   9,   2,   1,   0,  -1,   4,  -1,   3,   2,  -3,   0,   0,  -4,   2,  -2,   0,   1,   0,   1,   0 },
      {  11,  11,  -9,   6, -12,   3,  -2,   1,  24,  14,  19, -30,   5,  -1,  -1,  -1,  39,  23, -11,  -6, -21,   0,  -2,   0,  10, -21,  -8, -58,  18,  14,   1,   3, -21,   3, -42,  23,  47, -13,   3,  -2,  -2,  -5,   5,  33, -17,  -8,   3,  -2,  -3,  -1,   2,  -2,  -4,   5,   0,  -1,  -1,  -3,   2,   4,  -1,   0,  -1,   0 },
      {   6,  34,   4,   6,  -2,  -1,  -3,   0, -15,  20,  37, -50,  -7,   2,  -1,   0, -24, -36,  -9, -15,  41,  -4,   5,  -1, -34,  50,   3,  10,   2, -18,   4,  -4,  11,  10, -40,   4,  -5,   3,   5,  -1,   2,  -7,  -2,  21,  -3,   2,  -2,  -1,   7,  -1,  -1,  -2,  -6,   2,  -1,   1,   1,  -3,  -1,   3,   1,   1,   0,   0 },
      {  19,  29,  82, -44,  -1,  -9,  -2,  -4,  -1, -40,   5, -12,   9,   1,   3,   1,  38,   7, -10,  -3,   7,  -6,   0,  -1,   0,  -3,   1,   1, -14,  -1,   1,   0,  -3,   2,   1, -31,   7,  10,   0,   1,  -8,   4, -23,   4,  24,  -6,   1,  -1,  -3,   1,  -2,  14,  -3,  -7,   1,  -1,  -2,   1,  -4,   1,  -3,  -1,   3,   1 },
      {   6,   8, -13,  12,   7,  -6,   0,  -2,  16,  51,  11,  19, -34,   1,  -5,  -1, -10,  14,  23, -32,  -4,   1,   0,   0,  23,  13, -14, -12, -27,   7,   1,   1,   6, -30,  -4, -50,  18,  26,  -4,   5,  -9,   0, -16,  11,  44, -15,  -2,   0,  -2,  -3,   0,  14,  -8, -12,   4,  -1,  -3,  -1,  -2,   0,   0,   2,   1,   0 },
      {   2,  -8, -20,  23,  -2,  -3,   0,   0,   3, -21, -15, -11,  14,   4,   3,   0,  52,  19,  -4,  40,  10, -24,  -1,  -5,  -1,  42, -22,   5, -53,  -4,   6,  -2,  12, -36, -18,   5,  -9,  30,  -4,   3,  -2, -10,  26,  -2,  -2,   2,  -7,   3,  -1,  -3,   1, -10,   2,   4,   1,   1,  -4,  -1,   5,   0,   3,   0,  -2,   0 },
      { -11, -19, -13,  -9,  16,  -3,   2,  -1, -18, -26, -24,  -2,  -9,  17,   0,   5,  -5, -24,  -6,  -9,  39,  -5,   2,   0,   0, -11, -15,  20,  -3,  11,   2,   0,  -5, -16,  19,   5,  51,  -2, -12,   0,  -4,  26,   7,  46,  -8, -46,   7,  -4,   4,   2,   9,  -4, -38,   8,   7,  -2,   1,   3,   1,  -4,   1,   6,  -2,  -1 },
      {  -9,  -6, -28, -57,  29,  -1,   8,  -1,   7,  10, -15,  39,   5,  -8,   2,  -2,   1,  44, -15,   6,  -6,  -3,  -3,  -2, -37,  20,  14,   8,  -7,  -8,   0,   0,   4,  34, -29, -12,  -5,  -2,   5,  -2,   0, -16, -20,  33,  -7,   4,   0,  -1,   8,  -6,  15,  -1, -18,   8,  -3,   2,   0,  -6,  -1,  -3,   4,   5,  -3,   1 },
      {  -2,  12,  11,  12, -13,  -5,   1,  -1,  -2,  10,  13,   2,   8,   3,   0,  -2, -16,  13,  -1,   9, -29, -13,   6,  -3, -84,  -1,  -7,  -9, -16,  47,   2,   2,  27,  -3,  29,  -1,  31,   1, -23,   2,  -4,   6,  -1, -18,  -6, -17,   6,   2,   3,   0,  -8,  -1,  10,   7,   2,  -2,  -1,   2,  -1,   2,   2,  -4,  -3,   0 },
      {  14,  19,  32,  31,   9, -16,  -2,  -4,   9,  15,   9,  24, -32,   6,  -3,   1,  10,  16,  17,  -2,  13, -14,  -4,  -3,  15,   3,   9,  10, -22, -10,  -2,  -1,  -7,  11,  11,  -7,  -3, -22,   4,  -1,   0,  14, -16,   3, -59,   1,  18,  -1,   1, -17,  -1, -36,  -3,  45,  -3,   3,   1,   1,  -4,  -2,  22,   0,  -8,   1 },
      {  -9, -14, -31, -60, -21,  18,   6,   5,   0,   8, -10, -50, -11,  33,  -6,   4,   4,   4,  28,  -8,  18, -11,  -9,  -2,   9,  -4,  -5,  -6, -15,   4,   0,   2,  -1,   6,  14, -13,   7,  -1,  -5,   0,   0,   3, -17, -22,  -8,  10,  11,   3,   9, -11,  -8,  -7,  38,  22,  -2,  -2,   2,   2,  -3,  20,  19, -16,  -7,  -2 },
      {   3,   0, -16, -17,  17,  -1,   0,   1,  14,  16,  28,  -5,  52, -19,  -4,  -3,  12, -12,   7,  16,  -5, -12,  10,  -1,  30,  31,   0,   8,  13,   6, -17,  -2,  34, -21,  16,  -8,  -1, -50,  -4,   8, -22,  10, -18,  -9, -11, -13,  32,  -4,  -4,   4, -10,  19, -16,   6,   3,  -8,  -5,  -1,   0,  -5, -13,   7,   3,   2 },
      {   1,  -6,   5, -46,  17,  -4,  10,   0,   7, -10,  48,  26, -29,  -6,  -2,  -1,  -7, -47, -27,   6, -28,   9,   6,   2,  24,  -1,  -5,   0, -15,  28,   1,   1,  20, -21, -10,   9,   5,  13,  -8,   0,   3, -16,  30,   3, -11,  11,  -5,   2,   4,   0,   9, -29,  15,   9,  -7,   2,  -1,   1,  -3,   2,  15,  -7,  -2,   0 },
      {   1,  -5,   2,  26, -16,   0,   2,   0,  -8, -27, -42,   2, -26,  18,  -4,   3,   3,   9, -41, -23,   0,  23,   5,   3,   7,  25,   7,  -2,  22,  19, -20,   1,  47, -16, -14, -18,  10, -31,   1,   7,  -9, -17, -15,   9,   0,  20,  17,  -4,   5,  -6,  13,   6,  24,   4, -19,  -1,  -2,   0,   3,   4,   1, -17,   2,   2 },
      {   4,   5,   6, -14, -15,  21,  -7,   2,   6,   0,   3,  -5, -27,   6,   2,   0,  10,  27,   1,   5,   4,  37, -23,   1,   3,  14,  34,  -2,  22, -20, -12,   1,  33,  -3,   5,   8,  -3,   0, -20,  -1,   7, -15,  29, -22,   4, -49,  -4,  15,  -4,   5, -21, -18, -22,   1,  38,  -3,   1,  -6,  -2,   1,  12,  21,  -7,  -6 },
      {  -5,   1,   1, -14,  10,   9,   0,   0,  -6,  -2,  11,  -5, -12,  -5,  -2,   3,  -5,  17,   7,   9,   4,  17,  -5,  -2, -38, -10,  22, -12,   4, -27, -23,   6, -29, -68,   6, -13,  -3, -19,  36,  10, -27,  24,  29,   8,   4,  23,  12, -16,   7,  14,  -2, -13,   0,   0, -15,   1,  -3,   3,  -4,   1,   4,  -1,   6,   1 },
      {  -3,  -8,   1,   1, -47,  23,  -8,   7,  -4,  -9,  -5,   3,   7, -20,   2,   0, -11, -20,  -6,  -9, -21,  -8,  -2,   4,  -6, -30, -19,  -3, -10, -19,  19,   6,  23, -11, -17, -11, -18,  13,  20,   2, -13, -10, -11,   8,   4,   2,  26,  -5,  -3,   7,  -1,  13, -29,  49,   5, -14,   6,  -4,  10, -10,  23,  26, -35,   0 },
      {   4,  -1,   7,   4,  28, -13,  -1,  -2,   8,   1,   2,  -2,  12,  12,  -3,   2,  -4,   6,  37,  -1,  30, -20,  -8,   0,  -7, -63,  10,  -3,  -4, -17,   9,   3,  70,  -9, -20,   5,   1,  -3,   7,  -2,  -5, -31,   9,   9,  -6,   1,  -3,   0,   0,   2,  14,  -4,  10, -14,  -1,   1,  -4,   3,  -2,   2, -10, -12,  11,  -2 },
      {  -2,   1,  -9,   6,  72, -35,   0,  -6,  -8, -13, -13, -15, -11, -15,  16,  -3,  -6,  -1, -18,  -9,   1,   8,  -1,   7,  -6,  -4,   5, -24,   2,  -9,  19,   0, -12,  -6,  -9,   3,  -2,   8,  -8,   2,  -6,  -7,  -3, -28,   3, -12,  23,   8, -20,   7, -27,  25,  20,  31,  12, -17,   4,   1,  14,  19,   3,  -3, -25,  -1 },
      {   3,  -1, -10,   1,  23,  -3,   0,  -2,   0,  -7, -12, -33,  -4, -34,  16,   1,   6,  -1,   1,  -6, -24,  23,  10,  -5,   3,   0,  -7, -13,  -1,  -7,  -2,   8,  46,   6,  18,  -2,   2,  19,  30, -16,  13,  47, -21,   5, -10,   9, -40, -16,   7, -25, -19, -17, -13,  -5,  -6,  23,   0,  -1,  -1,   0,  14,  10,   6,   0 },
      {   1,   1,  -2,  -6, -14,  12,  -4,   3,   6,   6,  10,  48,   0,  24, -12,   0,  -7,  -5, -46,  11,  32, -27, -20,   3,   3,   6, -11, -40,  -1, -31,  17,   7,  21,   6,   2,  10,  12,   3,  21,  -5,   7,  39,  -3, -16,   1,  -3,  -2, -11,   1,   5, -34,  10,  10,  -8,  -4,   0,   2,  -3,  -2,  18, -12,  -8,   5,   4 },
      {   0,  -4,  -2,   4,  10,  14,  -5,   5,  -5,  -9, -15,   3,  21, -20, -19,   7,  -6, -18, -18, -29,  -7, -21,   0,  -1, -12, -11, -17, -10, -17, -11, -25,   6, -16, -14, -17, -28,  -1, -23,   2,  21,  43, -18,  -7, -14,  -8, -11,   8,  16,   1, -22, -10, -31,   6, -16,  38,   9,  -2,   1, -16,  14,   7,  17,  32, -13 },
      {   1,  -2,  -1,   5,  10,   0,   2,  -2,   1,  -3,  18,  -3, -36,  18,   2,  -2,   2, -13,   8,  69,  -7,  27, -22,  -6, -14, -14, -52,   6,  11, -17, -24,   8,  -2,  -1,  -5,  -5,  -2, -13,   2,   9,  21,  -9, -36,   5,   4,  -4,   5,   8,   7, -21,   9,  25,   1,  -1,  10,   4,   7,   2,  12,  -1, -10,  -1,   0,  -3 },
      {  -8,  -7, -11, -14, -26, -62,  17,  -7,  -7,  -8,  -7, -18, -44,   8, -10,   2,  -6,   0,  -7,   2, -19, -27,   5,   6,   1,  -3,  -3, -10, -11,  -1,   8,   2,   2,   8,  -5,   5, -20, -15,   8,  -2, -39,   4,   1,  -9, -13,   3,   4,   7,  -1,   7,  -6,   1, -18, -21,  21,   5, -11,  -3,  -9,  -9, -29,  35,  25,  -6 },
      {  -4,  -1,   2,   6,  -3,  11,  -4,  -1,  -4,  -1,   6,  13,  23,   0,   3,   5, -10,  -8, -12,  -1,   5,  10, -13,  -4,   0,   4,   3,  -6,   7, -11,   7,   5, -12, -19,   5,   3,   2,   9,  -8, -27, -74, -11, -30,   5, -17,  -7, -36,  29,   3, -31,  18,  -3,  22,   0,  30,  20, -12,  10,  10,  13,  10,   0,  -2, -17 },
      {   0,  -1,  -3,   2,   8, -27,  -7,   2,   2,   2,   3, -16,  22,  25,   7,  -6,  -3,  -8, -16, -12, -17,   0, -42,  11,  -1,  19,  -5, -40,   4, -28,  14,   8,   0,  -1,  51, -13, -10,   6,   0,  -1,  17, -22,   1,   8,   1,  -2,   8,   6,  12,   8,  44, -16, -12,   4,  -8,  -7,   4,   4, -13, -36,   8, -14,   5,   6 },
      {  -3,  -9,  -4, -14, -32, -34,   8,  -3,   3,   0,  12,  10,   4,  -5,   7,  -3,   3,   2,   1,  -9,  20,   4,   8,   4,  -6,  -2,  43,  -7,  -4,   7,  10,  -5,  -6, -53,   4,  23,  -7,   1,  -6,  -7,  54,  -6, -30,   3,  -9,   3,  -8,   2, -12, -40,   2,  25,  -5,   3,  -2,   7,   8,   0,  24,  -7,  -7,   5,  -4,   2 },
      {   6,   6,   8,   8,  35,  45,   6,   7,   0,  -7,  -3, -16, -15,  30, -31,  -7,  -2,   2,  -3,  -9, -42, -36, -16,  11,   4,  -2,  28,  11,  -1,   1,  12,  -8,   0, -19,  -9,  25, -18,   5, -18,   2,  -1,  20, -14,   4,   4,   2,   2,  11,  40, -18,  -4,  17, -15,  -1,  -3,  -3,   7, -16,  11,  -9,  -9,   1,   8,   7 },
      {   2,   3,   7,   0,  17,  40,  -1,   1,  -3,  -1,   0,  -6, -31, -14, -19,  16,   2,   3,   1,   8,   5, -29,  28,  -3,   2,  19,   8, -16,  -1,   5,   9,  17,   0,   6,  48,   9,  12,  -5,  37, -14,   9, -44,   4,   7,  -4,   9,  -4, -19, -21,  -3,  21,  16,   2,  -6,  16,  -9,  -3,  11,  11,  11, -17,  22, -15, -18 },
      {   2,   3,   3,   9,  14, -14,  45, -15,   3,   3,   5,   2,  35,  72, -25,   0,  -1,  -2,  -1,  -3,  -9,   9,   5,  -2,   0,   0,  -2,   5,  14,  12, -15,   4,   0,   0,  -4, -12,  -4,   8,  17,  -1,   6,  -2,   8,   0,   5,  14, -14, -12, -28,  -1, -11,  -4, -11,   4,   2,   8,   1,   3,   4,  14,  24,  44, -25,  -9 },
      {   4,   2,   1,   5, -10, -23,  17,  -3,   6,   4,   2,   5,  11,  -8,   5,  10,   6,   4,   5,  -3, -16,   5,  -4,   1,   5,   8,  10,  27,   3, -11,  11,  16,   8,  -2,  -1,   5,  18,   3,  19,  -4,  22,  29,   7,  18,   7,  10,  12,  22,  54,  16,  18,   2,  30,  -2,  38, -33,   6,  18, -10,  28, -14,   7, -26, -30 },
      {  -2,  -2,   1,  -1,  -6, -10, -25,   0,  -1,  -1,   3,  -1,   9,  15,   9,   1,  -3,  -7,   5, -23, -37,  15, -41,   1,   2,   9,  10,  44, -38, -16,  -7,  28,  -5,   9,  -2,  20,  34,  -6,  35,   0,  -6, -22,   9,  -5,  -6, -11,   4, -25, -19, -16, -18,  17,   2,  -2,  -8,   5,  -7,  -9,  15,  13, -15, -13,  12,   7 },
      {  -1,  -2,  -2,   1,  -5,  -2,  -3,   7,  -2,  -2,  -3,  -4,  -8,   0,   6,   0,   1,   2,   6,  22, -24, -25,  -2,  21,  -3,   8,  -3,  35,  51, -11,  54, -10,  -6, -29,  -8, -31,  19,  -2,   2, -29,  15, -16, -16,   4, -10,  -7,  -9,  -7, -22,  17, -20, -19,   6,  -3,  -8,   6,  -6,  -5, -26,  15,   7,  -7,   9,   9 },
      {  -2,  -1,  -1,  -1,  -9, -27,   7,  -5,  -2,  -3,   2,  -2,   3, -30, -65,   2,   1,   6,   3,  13,   0,  -9,   1,  28,   0,   2,  -1,   6,  27, -18,   7,   0,   2,  18,   5, -11,  14,  10,  -1,  20, -10,   5,  32,   1,  19,  -9,  10, -10,   3, -54,   4,  -1,   6,   1,  -7,  -4,  -8,   0,  29,  -3,  26, -20,  -1,   0 },
      {  -2,   2,  -2,  -5,   0,  13,   0,  -4,  -4,   1,  -1,  -8,  -8,  -5, -22, -10,  -2,   2,  -1,  -3, -15,  -1, -16,   4,  -4,   3,   5,   7, -12,  -9,  -1,   1,  12,  -1, -18,  12,  -6,  -5,  -9,   5,  11,  50, -10,   2,   1,  -4,   1,  14, -80,  15,  49,  -3,  13,  -7,   9,  -5, -15,  32,  -4,  -3,  -1, -10,  -7,   0 },
      {   1,   0,  -1,   2,  -6, -10,  31, -16,   0,  -2,  -4,  -3,  -4, -30, -38,  -1,   2,  -2,   0,  -5,  -1, -12, -18,  -8,   2,   3,   1,   4,  -8, -10, -21,   7,   0,  -5,   6,   8,  10,  -9,  -3,  -5,  -1, -14, -14,   0,  -2,  -8, -27,   4,   4,  58,  -5,  12,   5,   9,  -1,  41,  62,  -8,  10,   0,  15, -18,  -4,   3 },
      {   3,   5,   4,   9,  -6,  26, 101,   5,  -1,  -1,  -1,   0,   0, -21,  28, -25,  -1,  -2,   1,  -4,   5,  14, -16,  11,  -1,  -1,  -4,  -3, -17, -14,  17, -10,  -1,   5,   7,   4,   7,  -7,  -5,   2,  -7,  -9,   1,  -2,   0,  -7,  14, -16,  -4, -16,   0,  -2,   0,  -6,  -6, -12,  -8,   4,  -7,   7,  -7,  -4,  19,   5 },
      {   1,   1,   3,  -4,   2,   8,  -4,  21,   0,   0,   0,  -2,   1,  -3,  10, -11,   1,  -1,   3, -10,   2,   3, -23,   5,  -3,   0, -11,  -1,  -2,   9,  14,  -3,   4,   7,   0, -53,  -6, -23, -12, -21,   4,  17,  55,  11, -28,  17, -13,  25, -14, -22,   1,  23,   4,  -3,  -1,  11,  42, -11,  22,   0, -33,  13, -15,   1 },
      {  -1,  -2,  -1,  -2,  -6,   5,   5, -43,  -2,  -3,  -3,  -5,   4,  -3, -37,   6,  -3,  -4,  -2,  -8, -12,  20,   9, -22,  -3,  -6,  -9,  -9, -17, -19,   0, -17,  -3,  -4,   5, -18, -16, -11, -16, -23,  13,  -9,   5,  11, -25, -15, -16,   1,  16,  10, -16,   5,  -6,  -3, -12,  -9, -56,   1,  -2,  11, -33, -24, -37,  21 },
      {   1,   1,  -2,  -3,   6,   7,  15, -12,   1,   0,  -1,   0,  -3,   3,  11,  14,  -1,   0,  -6,  -3,   6, -10,   9, -11,   0,  -1,   0,   9,   8,  -1,   5,  21,  -7,  -4, -36, -17,  31, -10,  15, -21,   5,   8,   4, -74,  -4, -11, -14, -10,  20,  -5,  34,   1, -14,   3,  -1,   4,  -7,   6,  10, -53,  -5,  -7, -17,  -2 },
      {   1,   1,   1,  -1,   5,   2,  -8,  33,   2,   2,   2,   1,  10,  17, -35,  23,   3,   5,   0,   2,  -2,  61,  41,  26,  -3,  -1,  -7,  -3, -34, -16,  45, -26,   0,   2,  -2,   5,   8, -13,   1,  -6,   0,   0,  -9, -14,  -8,  -3,  11,  -5,   8,  17,  10,   2,  -2,  -2,   2,  13,  21,   9,   8,  -3,  -1,   6,  24,  -7 },
      {   0,   0,   0,  -3,  -9,   5,  11, -16,   2,   4,   4,   2,  -1,  27,  40,  15,  -2,  -1,  -4,  -5, -31, -20,  61, -12,   1,   0,   4,   0,  11, -50, -17, -10,   5,   4,   8, -11,   3,  16,  -8,  22,   0,   6,  17,  12, -14, -14,   3,  10, -12,   0,   5,  13,   7,   9,  12,  13,  13,  -2,  26,   2,  -1, -31,  20,  -4 },
      {   0,   2,   2,   0,   1,  13,   8, -24,   0,   2,   2,   1,   4,  -4,   3,  15,   1,   3,   4,   1,  13,   9, -21,  23,   2,   3,  -1,   0,   8,  31,  16,  24,   1,   3,   1, -15, -11,  12,  10,   6,  16,   8,   5,  11, -15,  11,   8,  15,   4,  35,  -8,  16,   1,  23,  16,  23, -50,  -6,  48, -12,   8,  -3,  48, -18 },
      {   0,   1,   0,  -2,   0,  -4, -16,  16,  -1,   0,  -1,   0,  -5,  -5,   4, -55,  -1,   0,  -2,  -4,  -3, -12,  -5,  -8,   3,   1,   2,   4,   3, -26, -10, -31,   2,   3,   5,  -3,  32,   4, -22,   9,   7,  -8,   5, -12,  -7,  18,  -4,  -7,  22,  17,   5,  16,   2,   5, -19,  27, -29,  52,  21,  11,   8,  43,   3,  -7 },
      {  -1,  -3,  -3,  -2,  -1,  16, -20, -89,   0,  -1,   0,  -1,   3,  -2,  -7, -31,   2,   0,   1,   1,   6,  18,  -2,  -2,   1,   2,   2,   2,   6,   2,  24,  -5,   2,  -2,   0,  -4,  15,  21,  -1,  22,  -9,  -2,  -6, -13, -16,  16,  13,  -3,  -2, -11,   9, -10,  -6,  -2,  -2, -16,  40,  -1, -12,   7, -19,  16,  22,  -4 },
      {   0,   0,   1,   1,   3,   0,   5,  -1,  -1,   0,   1,   1,   0,  -8,   0,  26,  -1,  -2,   1,   1,   3,   2,  -1, -12,  -1,   0,   0,   4,  -8, -33,   6,   2,   4,   1,  13,  11,  45, -11, -33,  15,   2,   1,  -8,  -2,  12,  48,  -8,  35, -18,   5,  -4,  -5,   3,  -4, -16, -14, -21, -64,   0, -18,   7,  27,  -7, -10 },
      {   0,   0,   0,  -1,  -1,  -3,  12,  47,  -1,  -1,  -2,   0,   6,  -6, -30,   2,   0,   1,   1,  -2,   4,  13,   0, -16,   1,   1,   4,   3,  22,  -2, -26,  25,  -1,  -6,  -3, -13,  11,  67,  -7,  10,  -4,  -3, -13, -10, -48,   9,   5,  -2,  -7,   9,   0,   4,  -6,  11,  -1, -27,   2,  -2, -11,  -3, -39,   0,  16,  14 },
      {   0,   1,  -2,  -1,   0,   8,   6, -15,   1,   3,   1,  -1,   0,  -6,  12,  74,   0,   1,  -2,  -1,  -3,  -5, -19,  27,   1,   0,   1,   2,   0,  -4,  -4,   3,   1,   2,  -1,   2,   4,   6, -20,  18,  -5,  -7,  -3,   8,  -4,  10,   1,   9,  -2,  -8, -32,  -2,  -2,  -7, -25,   4,  14,  71,  -8, -20, -12,  15, -11,   9 },
      {   0,  -1,   0,   0,  -1,   0,  -5, -11,   1,   1,   2,   1,   2,  -5,  14,  21,   0,  -2,   1,  -2,   1,  -9, -12,  72,  -2,  -2,  -5,  -3,  -2,  -1, -46, -48,   1,  -1,  -1,  -2,   4,  18,   6, -18,   0,   5,  -9, -14, -12, -11,  -2, -18,   5,  11,  32,  -7,   9,  -3,  -4,  -9,  -2, -35,  16,  24, -14,  19,  -9,   9 },
      {   0,   0,   1,   4,   1,  -4,   3,  -7,   1,   0,   2,   0,  -1,  -8,  -3,  44,  -1,   0,  -2,   3,  -1,   0, -21, -41,   2,   0,   6,  -1,   6,   5,   5, -33,   3,   2,   0,   2,   1,   6,   3,  -7,   6,   7,  24,  -9,   3,  -1,   9,  -9,   2, -15,  31,  49,  -3,  28,   5,  24,  -7,  -7, -50,  26,  17,  12,  28,  19 },
      {   0,  -1,   0,   1,   0,   5,  -2,  -8,   0,  -1,   0,   1,  -1,   3,   0, -28,   2,   2,   3,   2,   7,   4,  30,  62,  -1,  -1,   0,  -1, -10,  -8, -18,  43,   1,   1,   2,   7,  10,   1,  -2, -11,   2,  -4,  12,  -4,  -6,  16,  -3,  34,   4,  -6,  -2,  37, -17,   9,   5,  21, -11,   2, -57,   3,  13, -11,  -6,  23 },
      {   1,   1,   2,   2,  -2,  -3,  -2,  -4,   1,   1,   1,   0,   1,  -2,   5,  11,   1,   1,   2,   1,  -1,  -1,  13,  -5,   0,   0,  -1,  -4,   1, -11,   7,  66,  -2,   0,  -4,   0, -14, -13, -37, -19,   2,   7,   0,  -3,  19, -14,   1, -11,   9,   2,  27, -20,   3,  10, -27,  -3,   5,   1,  27,  47,  -4,  34,  20,  45 },
      {   0,  -1,  -1,   0,  -1,   0,  -5,  -6,   0,  -1,  -1,   0,   0,  -1,  -4,   1,   0,  -1,  -2,  -1,  -5,   0,   0,   9,  -1,  -3,  -1,  -2,  -4, -12, -12,  22,  -2,  -2,  -7,  -2, -16,   2, -40, -29,   3,   5,   6,   8,  -5, -15,   0, -42,  -3,  -4,  -4,  18,  13,  10, -29,  -7,   2,  -7, -26, -10,   0,   3,  12, -89 },
      {   0,   0,   1,   0,  -1,   2,   3,   9,   0,  -1,   1,  -2,  -1,  -4,   9,   8,   0,   0,   1,   0,   0,   2,  -5,   1,   2,   1,   4,  -4,   6,   6,  23,  17,  -2,  -2,   0,  -2, -10,  12,   2,  54,  -2,   2,  -6, -12, -19, -16,  22,   3,   6,   1,  20,  -4, -18, -33, -27,  58,  -9, -20,  -7,  35, -10,  -6, -45, -27 },
      {   0,   0,   0,   1,  -1,  -4,   3,   2,   0,   0,   0,  -1,   0,  -1,  -4,   4,   0,   0,   0,   1,  -2,  -4,  -4,  -7,  -1,  -1,  -1,  -2,   2,   2, -15, -13,   2,   3,   1,   6,   8,   5,   6, -42,   1,  -5,   0, -18,  28,  25,  23,  40,  -9,  -7,  -2, -27, -60,  19, -19,   5,   5,  15,  13,  38, -22, -34,  12, -34 },
      {   0,   0,   0,   0,   2,   4,   3,  -3,   0,   0,   0,   1,   6,   1,  -6,  -9,   0,   1,   1,   0,   1,   3,   6,   0,   0,   0,   2,   2,   6,  -5,  -4,  -2,   1,   3,   3,   9,   5,  12,   1, -13,   2,   2,   1,  14,  35,  -5,  31,  -7,  -4, -11,  -7, -31,  38,  39,  13,  60,  -5, -10, -13, -33, -58,  16, -12,   9 },
      {   0,   0,   0,   1,   1,   3,  -1,  -9,   0,   0,   0,   1,  -1,  -3,   1,   3,   0,  -1,  -1,   0,   0,   1,  -2, -15,   0,  -1,  -1,   0,  -2,  -4,  -8,   6,   0,   1,  -1,   2,   4,  22, -17, -53,   0,   3,  -1,   6, -18,  20,  71, -12,  -1,   2,  -8,   5,   2, -60,  20,  17,   5,  -2,  13, -13,  35,   0, -11,  15 },
      {   0,   0,   0,   0,   0,  -3,  -5,   5,   0,   0,   0,   0,   1,   3,   6,   7,   0,   0,   0,  -1,   0,  -1,  -9,  11,   0,  -1,  -1,  -2,   0,  -1,   7,   4,   0,  -2,  -1,  -3,   4, -11, -44,  18,  -2,   2,  -1,  -2,   7,  50, -18, -63,   1,   1,   3,  -7, -31,  11,  62,  10,  -2,   2,   2,  15, -12, -30, -13,   4 },
    },
    {
      {  88, -52,   0,  -3,   0,  -1,   0,   0, -67,  25,  12,   0,   2,   0,   1,   0,  11,  13, -13,   0,  -2,   0,  -1,   0,  -2,  -9,   2,   3,   0,   1,   0,   0,   1,   4,   2,  -2,   0,   0,   0,   0,  -1,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
      {  53,   7, -20,  -2,  -4,   0,  -2,   0,  33, -70,  20,   4,   3,   1,   1,   0, -57,  42,  20, -13,   2,  -3,   1,  -1,   9,  11, -27,   2,   1,   1,   0,   0,  -5,  -7,   7,   7,  -1,   1,   0,   0,   1,   4,  -1,  -4,   0,   0,   0,   0,  -3,  -1,   0,   2,   2,   0,   0,   0,   0,   2,  -1,   0,  -1,  -1,   0,   0 },
      { -48,  20,  -3,   2,   1,   1,   0,   0, -50,  -4,  37,  -3,   4,  -1,   2,   0,   2,  66, -33, -14,  -1,  -2,   0,  -1,  29, -35, -26,  23,  -1,   3,  -1,   1,   4,  -7,  24,   0,  -6,   1,  -1,   0,   5,  -2,  -5,  -8,   2,   1,   0,   0,   1,  -1,   2,   2,   2,  -1,   0,  -1,   3,  -1,  -1,  -1,  -1,  -1,   0,   0 },
      { -32, -32,  34,  -3,   4,  -1,   2,   0, -36,   6,   4,   5,   2,   2,   0,   0, -10,  12,  46, -20,  -2,  -4,   0,  -1,  -6,  59, -46, -24,  10,  -3,   2,  -1,  15, -19, -28,  31,   1,   1,   0,   1,   4,  -2,  13,   0,  -8,   1,  -1,   0,   5,  -4,  -5,  -1,   2,   1,   0,   0,   1,   0,   2,   1,  -1,   0,   0,   0 },
      { -23, -83,  34,  11,   7,   3,   2,   1,  51,  10,  29, -22,  -4,  -5,  -1,  -2, -25,  25, -30,  -1,   3,   1,   1,   0,   6,  -6,  16,   2,  -3,   1,  -1,   0,  -8,  14, -11, -16,   8,  -2,   1,  -1,  -1,   1,  -9,  14,   2,  -2,   1,   0,  -1,   3,   0,  -1,  -6,   0,   0,   0,   0,  -1,  -2,   1,   1,   2,  -1,   0 },
      {  14, -14,  20, -12,   4,  -3,   1,  -1,  40,  33, -37,   2,  -3,   0,  -2,  -1,   6,   0, -19, -12,   5,   0,   2,   0, -13,  -8, -47,  41,   7,   2,   1,   1,  14, -36,  46,  26, -24,   3,  -5,   1,  -7,   8,  15, -33,  -1,   3,   0,   0,  -2,  -3,  -4,   3,   9,   0,   0,   0,  -3,   2,   4,  -2,  -1,  -2,   1,   0 },
      { -18, -40,  -4,   3,   2,   4,   2,   1,  -5, -47,  20,  33,  -4,   5,  -2,   2,  45, -17,  35, -21,  -6,  -2,  -1,   0, -25,  -2,   6,  11,   1,   1,   1,   0,   8,  12,  39, -29, -13,   2,  -3,   0, -10,  16, -25, -27,  25,   0,   3,  -1,   0,   6,  -7,  18,   5,  -8,   1,  -1,  -2,   4,  -3,  -3,  -6,  -1,   2,   0 },
      { -11,  -2,  16,  -7,   4,  -2,   2,  -1, -31,  13, -30,  11,   1,   2,   1,   1, -72, -11,  13,  27,   8,   2,   1,   1,  27,  20,  25,  19, -25,  -2,  -5,   0,  20, -17,  23, -44, -10,   7,  -1,   2,  -5,  13, -15, -11,  25,  -2,   2,  -2,   5,  -1,  -5,  13,  -4,  -5,   0,  -1,   2,   1,   0,  -2,  -2,   2,   1,   0 },
      {  17,  34,  11, -22,   4,  -2,   0,  -2,  14,  31,  18,  -5, -14,  -3,  -4,  -1,  12,   6,  -4, -47,  10,  -1,   2,  -1, -15,  20, -24,   8,   1,   0,  -1,   0,   7, -12, -13, -38,  17,   4,   1,   2, -16,   6, -39,  40,  33, -14,   4,  -2,   6,  -8,   9,  26, -30,  -5,   0,   0,  -4,  -4,  -1,  -7,  -2,   9,   2,   1 },
      {   5, -15,   5,  -3,  -1,   0,  -1,   0,  14,   2, -30,  17,  -2,   3,  -1,   2,  31,  14,  30,   8, -19,  -2,  -5,   0,  69, -31, -14, -39, -12,   5,   1,   2, -30, -37,   7,  -8,  31,   7,   4,   0, -11,  23,   3,  13,   8, -11,  -1,  -3,  -6,  -2,  -3,   4,  -9,   0,   1,   1,  -3,   4,   3,  -1,   1,   3,   0,  -1 },
      {  18,  31,  69, -13, -16,  -4,  -3,  -2,   6,  -3,  10, -32,   8,   0,   2,  -1,  10,  10,   1,   4, -15,   3,  -3,   0,   7,  -4,   9, -32,   6,   2,   0,   0,  -6,  11,  -9,  -9, -10,   3,   0,   1,   2, -15,  -9, -43,  25,  15,  -1,   3, -10,   8, -22,  23,  39, -17,  -1,  -2,   1,   0,  -2,  14, -15, -11,   3,   0 },
      {   2,   5,  68, -34,  10,  -8,   1,  -2, -19, -57, -30,  -6,  12,   4,   5,   2,  14,  -2, -15,  17,  21,  -3,   1,  -1, -14,  -8,  -2,  21,  -5,  -5,   0,  -1, -13,  13,   7,   9,  11,  -5,   2,   0,   1,  18,  -7,  29, -22, -11,   1,  -2,   4,  -7,  12, -10, -24,  14,   0,   2,   1,   2,   1,  -8,   8,   5,  -3,   0 },
      {   4,  -1,  -9,  -2,   4,  -2,   0,  -1,  13, -16,  -9,   4,   5,   0,   3,   1,  44,  45, -17,  40,  -4, -11,  -3,  -4, -16,  39,   7,   4, -50,  -2,  -4,   2,  35, -30, -19, -28,  -7,  21,   1,   3, -10, -25,  16,   1,   8,   6,  -3,   0,   1,   2,   6, -16,   8,   2,  -2,   0,  -6,  -2,  -4,   2,   8,  -5,   1,   1 },
      {  -2, -12, -13,   4,   0,   6,   0,   1, -11, -36,  12, -36,  11,   1,   2,   0,  12, -42, -31,   1,  29,  10,   5,   2,  32,  14,  -4,  15,  17, -10,  -4,  -2,   1, -37, -14,  11,  -1,   3,  -1,   3,  -9,  -4,  34,   9,  34,  -8,  -3,   0,  -5,  26,  -7,  19, -11, -31,   5,  -4,  -3,  -2,   0,   4, -25,   9,   7,   0 },
      { -13, -18, -28, -28,   5,  -1,   4,   1,   0,  -1, -55,  16,  30,   0,   4,  -1,  -8,  32,  -8,  15, -16,   0,  -3,   1, -22, -13,  -9,  -8,  20,   1,   4,   1,  -5,  28, -19,  15,  -6,   0,   2,   0,  11, -21, -14,  11,  28,   5,  -3,  -1,  -1,   7,  12,  45,  -8, -27,  -1,  -2,   4,  -5,   8,  -6, -31,   8,   6,   1 },
      {  -4,   8,  15,  14,  -3,   4,  -2,  -1,  -8,  -3,  12,   8,  -3,  -4,   1,  -1, -26,   2,  -7,   8, -28,   3,  -1,   0, -63, -34,   5, -28,  -3,  24,   3,   5,  20, -27,  15,   4,  40,  16,  -6,   1,  -5,  15,  37,   5,  17, -27,  -8,  -3,  -1,  21, -15,  -5, -17, -15,   9,   1,   1,  -2,  -4,   1,  -5,  11,   2,  -1 },
      {  -9, -13, -15,  -6,   7,  -2,   2,  -1, -13, -16, -20, -18, -14,  15,   1,   4, -13, -17, -15, -16,  14,   5,   3,   0, -16, -13, -17, -10,  -6,  26,   2,   4,   0, -15,  -2, -12,  39,   5,  -1,  -1, -15,  -5, -14,  14,   1,  21,   0,  -2,  11, -11,  25,   3,  59,   3, -23,   0,  -4,  16,   2,  23,   4, -53,   6,   2 },
      {  -5,   1, -26,  23,  -7,  -2,  -1,   1,  -4,  14,  -2, -58,  31,   2,   3,   0,   8,  17,  16,  26,  29, -22,  -1,  -4, -28,  15, -10,  -4, -20, -10,   4,   0, -29,   2,  19,   4,  12,   1,   6,  -2,  -6,  49, -19,  -8,  -8, -10,   1,  -4,   1, -11, -30,  20,  -1,   5,   2,   1,   2,   0,  11,  12, -11,   1,  -1,  -1 },
      {   7,   3,  18,  97, -25,   4,  -9,   1,  -9, -16, -48,  -7, -30,   5,  -1,   1,   6,  17,  -6, -18,   5,  17,   2,   0,  -3,   2,  -1,   4,  12,  -3,  -5,  -1,   5,   3,  -5,  -6,  -4,   2,  -2,   3,   9, -10,  -9,   5,   8,  -3,   0,   1,  -1,  -8,   8,  12,  -8,   3,   2,   1,   6,  -3,  10,  -9,   1,   4,  -4,   1 },
      {   3,   0,  -8,  -7,   2,  -1,  -1,   0,   6,  -6, -10, -11,   7,   4,  -2,   2,   9,  15,  -7,   3,  -3,   9,  -2,   1,  20,   7,  23, -16,  38,   9, -12,   0,  77,   1,   9,   9,  21, -46,  -8,  -4, -35,  11, -19,  -3, -28, -15,  16,   0,   6,   0, -18,  11,   0,  12,   5,   0,  -3,  -6,   5,   6,  -7,  -2,  -3,  -2 },
      {   1,  -2,   3,  17, -19,  -5,   4,   1,  -1, -12,  24,  11,  19, -13,   2,  -5,  -7, -39, -37,  29, -46,  -9,   7,   2,   6,   6, -33, -13, -12,  21,  12,   1,   4, -11,  -6,   9, -14,   6,   5,  -4, -19,   4, -28,   0,  -6,  20,   4,  -2,  11, -18,   4,  30,  -9,  28, -12,   2,   8,  -9,  30, -18,   5,  11, -15,   3 },
      {  -8,  -9, -23, -36,   0,  12,   0,   0,  -9, -15, -24, -17, -24,  -7,   7,   1,   0,   5, -33, -21, -15,  34,   2,   6,  -6,  21,   7, -19,  16,  -2,  -1,  -1, -19,  -5,  -2, -12,  -1,  19,  -3,   3,  20,   8,  -2, -23,  13, -12,   2,   0, -13, -10, -37,  -4,  -2,  32,  13,   3,   9,  -6,  12,  11,  43,  15, -29,  -3 },
      {  -5,   3,   3,  -1,   3,  -2,   1,   0,  -7,  -1,  -4,   9,  -1,  -8,   0,   0,  -9,  20,   6, -14,   2,   5,  -1,   2, -27, -18,  32,  -3,  16, -30,  -3,  -1, -31, -44, -23,  -3, -34, -22,  26,   3, -58,  15,  10,   5,  -4,  38,  17,  -1,  11,  15,   1,   5,   3,  15,  -9,  -4,  -1,  -5,   2,  -2,   3,   5,  -7,   1 },
      {  -3, -12,  -6, -13, -21,   3,  -4,   1,   0,  -5,  -5, -48,  14,   3,   6,   1,  -5, -14,  42, -12,  13, -10,  -5,  -1,  -1, -59, -11,   0, -14, -19,  11,   3,  47,  -3, -18,  -4, -16,  22,   9,   1,   0, -29,  -2,  -1,   9,   0,   2,  -2,  -5,  -3,   7,  -2, -13,  22,   3,   0,   5,  -4,   6, -12,  25,   9, -15,   1 },
      {  -1,   0,  -2,   5, -14, -17,   8,   3,  -4,  -4, -25,  20,   0,   9,  -5,   0,  -8,  -5, -24, -40,   7, -36,   6,   2,  15,  -2, -13,   1, -32,   8,  28,  -2,   7,  39, -23,   0,  -1,  16,   4, -10, -40,  11,   5,  -7,  -8,  -1,  -5,  -1,  -3,  36, -37, -12,   2,   3,   4,   1,  -3, -11, -16,  27,  -5,  17,   7,  -2 },
      {   6,   6,   7,  21,  79,  -5,  -1,  -4,   3,  -2,  10,  31,  -7, -17,   2,   1,  -2,  -9,  -2,  11,  25,  -6,  -6,  -1,  -2, -23, -11,  -2,  -9, -28,   2,   1,   9,  -7, -19,   5,   4,   4,   1,   4,  12, -16, -10,  -4,  -1, -17,  13,  -1, -15,  -3, -24,  27,  10,  20,  15,  -6,  11,   1,  20,  31,  -8,  -7, -17,  -3 },
      {   4,   7,   6,   1,  -4,  18,   2,   1,   8,  11,  13,  21,   8,  20,   6,  -4,   5,  14,   6,  10,  22,  25, -10,   0,  10,   5,  13,  11,  10,  -4,  13,  -6,   7,  25,  -5,  26,   2,  39, -12,  -9, -21,  25,  19,   4,  24,  -9, -20,  -5,  19,  16,  -2,  17,  -3,  12, -30,   0,  14, -18,  22, -14,  22, -49, -22,  13 },
      {   5,   5,  -3,  12,  65,  -7,   1,  -6,  -2,  -8, -13, -37,   0, -40,   7,  -2,   5,   0,   8, -16, -39,   7,  16,   1,  11,  -1,   3,   3,   0,   5,   3,   0,  21,  19,  -1,   2, -12,  18,  13,  -5,   6,  36,   8,   6,   7,  28, -19, -10,  25,  -6,  -4, -15,  -4, -18, -18,   1,  -8,  -3, -11, -13,   1,   0,   7,   6 },
      {  -1,  -4,  -5, -20, -22, -12,   4,  -1,   2,  -4,  12,  11, -70,  21,   0,   4,   1,   1, -10,  23,  24,  12,  -4,  -3,  -3, -16,  -5, -19, -11, -22,  -3,   3,  25,   3, -11,  16, -12,   8,  24,   1,  16,  33,  -5,   3,  -5,  23, -18,  -4,  12, -30, -15,   6,  -1, -22,  -5,   4,  -3,  -2,   9,   2, -24,  16,  13,   1 },
      {  -5,   1,  -7,  -4,  18, -10,   4,  -4,  -7,   2,   3, -28, -54,   3,   9,  -2,  -3,   1,  21,  42, -14,  27,   0,  -4,  -2,   0, -38,  10,   0,   9, -14,  -4, -18,  33,   3,  -6,  -2,  -2, -15,   3, -56, -21,   5,  -5,  -3, -11,   7,   8, -10,  27,   6,   4,  -6,   6,  13,  -1,  -5,   1,  -5, -10,  -1,   5,   2,   1 },
      {   2,   1,   0,   8, -17,  -9,   5,   0,   4,   6,   1,  12,  16,  -3,   5,  -1,   2,  -5,  -2,  26,   1,  16,  -1,  -1,  -3,  -2, -45,   1,  24, -14, -11,   3,  16,  14,  -2, -37,   3, -18,  29,   9,  33,  28,   9,  20,   8,  23,  24, -18, -22,  43, -11, -12,   3,  10, -22, -14,   9,   1, -25,  13,  13,  -9, -12,   3 },
      {  -1,   0,   1,   8,   2,  33, -14,   3,  -2,  -1,  -4,  15,  25, -16,  -5,  -1,  -8,  -8, -24,  10,  16,   3, -17,   1,  -2,  -4, -31, -38,  18, -42,  -8,   7,   2,  18,   3, -17, -10,   5,   6,   5, -33,   1,  -4, -15,   8, -17, -14,   9,  17, -30,   6, -25,  -2, -19,  22,   5, -25,   7, -15,  -6,  24, -13,  26,  -1 },
      {   0,  -1,   3,   4, -11,  16,   3,   1,   1,   0,   3,  11,  21,   6,   0,  -2,   2,   0,  20,  14, -16,  24,   7,  -4,  -3, -20,  -8,  60,  11,  22, -22,  -5,  -4,  -7, -53,  -9,  10,  -6, -12,  -2,  -9,  12, -10,  -2,   2,   0, -23,   5,   2, -33, -31, -13,  11, -17,   6,  15,  -9,  -3,   5,  35,  11,  14,   4,  -4 },
      {   1,   5,   6,   8,  11,  15,   9,  -3,   0,   7,   6,   4,  13,  15,   3,   5,   0,   5,   3,   7,  16,   9,  20,  -5,   3,  12,   8, -12,  16,   8,   5,   2,   8,   7,  33,  14,   2,  23,  24, -13, -25, -14,  -7,  23,   6,  28, -32, -14, -42,  -5,  15,  -2,  10,  -5,   1,  11,  -6,  34,  -1,  23,  10,  46, -31, -21 },
      {   2,   1,  -2,  11,  37, -30,  25,  -7,   3,   0,   2,  -5,  25,  70, -14,  -5,  -1,  -4,  -6,  -5,  19,  16, -18,  -1,  -4,  -6, -10, -26,   0,  19, -14,  -4,  -9,  -7,   3, -16, -10, -13, -10,   4,   6, -15,  -4, -17,   4,   8,  -7,   1,  23,  -1, -10, -17, -12,  -6, -23,  12,   8, -27,   1, -15,  13,  26,   3,   3 },
      {  -5,  -4,  -4,   5, -16, -90,   9,  -9,   0,   1,   4,   8,  19, -25,  28,  -5,  -2,   0,  -1,  -8,   3,  36,   0,  -4,  -2,   1,  10,  -2,   0, -15,   0,  -5,   0,  -9,   4,  11,   0,  11,  -1,  -6, -10,  -6, -14,   0,   1, -16, -17,  -7, -36, -12,  -1,  -6,  -5,  -4,   0,   9,  -9,   2,  -1,   0,   8, -20,  23,  10 },
      {  -1,  -1,   1,  -2,   4,  23,  13,   7,  -5,  -5,  -4,  -2, -20, -18, -15,  -3,   0,   2,   5,  19,  13, -30,  18,   3,  -5,  -3,  -1,  -5,  32,   8,  31, -12,  -5, -18,  10, -15,  -8,  -2, -12, -18,  -2, -25, -15,  10, -10,   6, -17, -11, -40,  -7, -31, -12, -16,  -7, -34,  19,   4, -41,  -4,  -6,   3, -21,  24,  22 },
      {  -1,   2,   0,   0,  15,  36,  -8,   2,  -4,  -1,  -1,  -7,   4,  30, -15,  -5,  -5,  -6, -10, -17, -27,  16, -31,   2,  -3,   3,   4,   6, -38, -13, -18,   7,   7,  -2, -15,  24,  -2,  -9,  10,   8,  -3,  22, -22,  -7,  -5,  -3,   1,  -9, -71,  12,   8,  -2,  -6,  -4, -14,   7,  -1,   7,   4, -18,  -1, -12,  25,  13 },
      {   1,   0,  -2,  -3,  -4,  18,  14,   1,   1,   2,  -4,  10,   7, -41,   5,  -5,   3,   1,  -2,  -5,  45,  18,  39, -19,  -3, -13,  -5, -30, -27,  31, -33, -19,   6,  -5, -16,   6, -19,  -9, -35,   6,   5,  22, -17, -13,  -5,   6,  -6,  22, -11,  21,  15,  -2,   1,  -5,  -1,   8,  16,   4, -10, -18,  -3,   4, -14,   0 },
      {   2,  -1,  -1,   5,   3,  10, -10,   3,   3,   1,   0,  -2,  -4,  10,  21,   6,   1,  -3,   0,  14,  -9,  14,  -9,  -6,  -1,   0,   2,   6,   7, -16,  21,   7,  -3, -32,  -2,  24,   4,  17,  -1, -17,  16,  -6, -69,   0, -12,  -1, -22,   7,  33,  54,  -1, -10,   6,   1,  17,  -1,   1,  -7, -40,   3,   6,   4,   5,  -5 },
      {   0,   3,  -3,   4,  12, -20,  11,  -4,  -1,   0,   1,  -4,  -8,  25,  -5,  -4,  -1,   5,  -7,  15,  -2, -28,  -6,   4,  -1,  -5,   2,   1,  37,  10,  43,  -6,   6, -10, -36, -16, -18,  10, -30, -12,   8,  43,   3, -17,   5, -10,   6,  21, -26,  -7,  26,   1,   3,   5,  11,  -1, -22,  37, -16, -20,   3,  -4, -13, -13 },
      {   0,  -1,  -1,  -3,  -2, -20,  12,   0,   0,  -2,  -1,  -8,  -5,  -3, -47,   0,   1,  -2,  -1,   2, -19, -21,  -2,  -3,   0,   3,  -3,   2,  10, -21, -15,   7,   4,  -1,   4,  -9,  -5,  -3,   5,  -7,  -4,  15,   1,  11,  -2,  -7, -30,  17,  -2,  14,  24,  -5,  16, -19,  34,  28,  75, -11,  20,  11,  14, -10,  -5,  -4 },
      {   1,   2,   1,   4,   0,  -3,  49,   7,   0,   0,   2,   1,  -5,  -7, -51,   2,   3,   7,   4,  13,  12,   4,  36,  22,   1,  -1,   6,   7,  -4,  12,   9,   3,  -1,   1, -10,  27,  18,  -8,  34,   8,   5,  -3, -15, -21,  26,  -4,  15, -26,  11,  11,   6,   0,  -1,   3,  23, -20, -21,   6,  10, -12,  36,  -4,  27,  -3 },
      {   4,   4,   3,   0,   0,  27,  57,  22,   3,   1,   0,  -3,   8,  -3,  31, -18,   3,   2,   3,  -9,  -1,  32,  -9,   8,   0,   1,  -4,  -3,  -2,  -7,  30,  -1,   3,  -1,   1, -30,  -4,   0,  -1, -19,   4,   0,  16,  -2, -25,  -4,  -8,   0,   9,   4,   8,   7,  -2,  19,   2,  -4,  29,  28,  25,   9, -19,  10,  50,  -9 },
      {   0,   0,   0,   3,  -9,  12,  64, -21,  -2,  -4,  -4,  -5,  -2,   0,  -8, -19,  -3,  -4,  -2,  -7, -16,   7, -12, -23,  -1,  -2,  -2,  -2, -17, -11,   2,   9,  -4,  -3,   6,   1, -18, -21,  10, -16,   3, -10,   0,  17,  -1, -22,  -9, -10,   9,  -8, -15,  -3,   1,  -9,  14,  13, -30,   4, -19,  -3, -24, -31, -59,  15 },
      {   0,  -1,  -2,  -1,   4,   5,   0, -13,   1,   2,   0,   7,   3,  14,  29,  13,  -1,  -2,  -7,  -6,   4,  -9,  33, -14,   1,   2,  -2,  12,  11, -14,  -9,   0,  -4,  -2, -19, -21,  47,   8,  31, -14,  -6,  -4,   9, -48, -23,   5, -10, -30,  -7, -11,  10,   7,   0,  -3,  18,   3,  23, -11, -16, -46, -18,  -1, -15,  17 },
      {  -1,  -1,  -2,  -3,  -1,  -7, -25,  19,  -1,  -3,  -1,   0,  -3, -38, -38,  -6,   3,   2,   2,   5,  23,  28, -52,  18,   0,   2,  -4,   6,  -8,  26,  30,  14,  -3,  -1, -12,  -1,  11, -15,  15, -24,  -8,  -7,   0, -27,   1,  -5,  -4, -19,   6, -10,   6,  -3,  -8,  -6, -22,  11,  26,  19, -26,  -4, -12,  15, -25,   4 },
      {  -1,  -1,   0,  -5,   2,  -3, -37,   1,   0,   2,   2,  -1,  17,  14, -42,  20,   0,   2,   4,  -4,  -2,  48,  37,  22,  -4,  -3, -10,  -7, -13,  -6,  27,  -6,   1,  -1,   3, -26, -14,   0,  -1, -21,   9,   6,   8,   9, -24,  -2, -15,   4,  -6,  10,   8,  -1,   6,  11,  26,  -5, -32, -27,  21,   0, -33,  -9,  -9,   4 },
      {   1,  -1,   1,   4,   5,   7,   3,   9,   2,   0,   3,  -1,   5,  21,  12,  -4,   2,  -1,   1,   5,  -9,  22,  31,   9,   0,   0,   1,  -5, -17,  -5,  32,   1,   2,   2,   6,  20,  -9, -12,  -6,   6, -13,   7,   5,  11,  26,  -3,  32,  10, -10, -46,  10,  -3,   1,   0,  13, -17,  46, -30, -59,  19,  -3,  -4,   2,   2 },
      {   0,   0,   1,   2,   4,   5,   4,  -5,   0,   0,   1,  -1,   1,  -2,  14,   9,   1,  -1,   3,  -1,  16,  -6, -35,   4,   3,   1,   6,  -6,  18,  64, -25,  12,  -1,  -4,  -3,  -5, -39,  30,  37,   1,   1,  11,   1,   9, -15, -12,   0, -18,  -9,  -5,  16,   8,   1,  11,  41, -17,  -1, -34, -10,  14,  -7, -12,  13,  12 },
      {   0,  -1,  -2,  -2,   2,  -1, -31, -22,   2,   1,   1,   0,   2,  20, -10,  16,   1,   1,  -3,   1,  -5,   1,  27, -14,   0,  -4,   0,  -6,  -1,  19,  -2, -19,  -3,  -3,  -4,  -1, -36,  -8,  15,   1,  -1, -13,   3,  17,  -4, -22,  -9, -25,  17,   0, -38,  13, -12,   8,   1,   8,  30,  68, -16,  -9,   8, -23,  21,   7 },
      {  -1,  -2,  -1,  -1,  -6,  -1,  18, -25,  -1,  -3,  -1,   0,  10,  -9, -56,  -5,  -1,  -1,   0,  -1,  -1,  11, -11, -17,   0,  -1,   5,   7,  12, -15, -21,   7,   1,  -1,   4,  -1,  16,  46, -21,  15,  -4,  -6,   5,  11, -19,  16,  -8,  17,  -4, -13, -20,  22,   9,  33, -10, -35,   5,   2, -40, -18, -19,  11,  17,  -8 },
      {   0,  -1,   0,  -1,  -3,   0,   5,   5,  -1,  -1,   0,  -5,  -5,   4, -17,   1,  -1,  -1,  -2,   1, -16,  -4,  20,  -9,  -1,  -2,   2,  -6,   5, -16, -28, -28,  -1,  -3,  -4,   0,  -9,  22,  12, -22,   3,   2,   1, -30, -16, -39,  12, -13,  10,   1,  42,  17, -25,   3, -50,  10, -12,  -7, -18,  56,  -2,   3,   1, -13 },
      {  -1,  -2,  -1,   0,  -4,  13,  -7, -99,   0,   0,   1,   2,  -3, -11,   7,  -1,   1,   1,   2,   2,   7,   3,  -7,  -4,   1,   2,  -1,   1, -10,   4,  21,  10,   1,   3,  -1,   1,   5, -32,   7, -19,   6,   6,  13,  -7,  14,  -1, -12,  10,  -2,   4,  16,  17,   4,  15,  -7,  17,  -3, -21,  -1,   8,  18,  10,  25, -22 },
      {   0,   1,   1,   1,   7,  -8,  -4,  54,  -1,   0,   1,   0,  -3,   4,   2,  10,  -1,  -2,   0,  -3,   1,  -1,   1, -35,   0,   0,   1,   2, -13,   5,  -4,  27,   0,   2,   2,  -3,   2, -27,  13, -23,  11,  -2,  27,   3,  -7,   8, -28,  27,  -6,  -2,  -8,  57,  -4,   7,   4,  11, -21, -11, -27,  -9,  40,  -9,   9, -21 },
      {   0,   0,   0,  -1,   2,   0,  -1,  11,   0,   0,  -1,   0,   3,  -4,  -6,  24,   0,   0,   0,   2,   1,   9,  11, -67,   2,   2,   2,   3,  12,   9,  16,  42,  -3,   2,  -4,   3,  -8,  10,  23,  12,  -4,   1,  -8,  -3,  22, -27,  23,  10,  -1,  -1,  -2, -38,  12,  18, -20,  21,   4,  -9,  22, -14, -32,   0,   9, -34 },
      {   0,   0,   0,  -2,   1,  -2, -19,   8,  -1,   0,  -1,   0,  -1,   4, -13, -53,   0,   1,   0,   2,   2,   0,   3, -50,   3,   2,   4,  -2,   7,  -2,  13,   3,   1,   2,   2,   2,   7,  -6, -21,  22,   6,   6,   6,  -9,  16,  23,  -8, -35,   8,   5,  21,   3,  -8,  33,  28,  27, -11,   1,  -1,  27,  -6,  10,  14,  51 },
      {  -1,  -1,  -1,  -2,   1,  -4,  -7,   1,  -1,   0,  -2,  -3,  -5,   5,   5,  -8,   0,   0,  -1,   1,  -2, -15,  12,  -2,  -2,   0,  -1,  -3,   3,   3, -11,  28,  -2,  -2,  -4,   0,  -3, -30,   3,  -7,  -4,  -2,  -7, -17,  23,  -5, -60,  14,  -5,  -3,   8, -24, -27,  23, -16, -77,   8,   8,  11,  18, -15, -10,  -8,   7 },
      {   0,   0,   0,  -1,   0,   1,  -2,   5,  -1,  -1,   0,  -1,  -3,  -7,  -2,   9,  -2,  -1,  -2,  -1,  -4,   2, -15, -43,  -1,   0,  -1,   1,  -4,  -6,  24, -43,   0,  -1,  -1,  -5,  -8, -17,  -3,  10,  -2,   6,   2,  13, -10, -36, -14, -31,  -2,  -2,   9,   8,  51, -23, -11, -50,   1, -16,   9, -14,  24,  35,  -1,  12 },
      {   0,   1,   1,   1,  -2,  -2,   7,   8,   1,   1,   2,   2,  -1,  -2,   7,  30,   0,   0,   0,   2,   1, -12,  11,  33,  -1,  -2,  -1,  -3,   0, -15, -29,  49,  -2,   2,  -4,   5, -14,  -6, -35, -32,   2,   8,   5,  15,  20, -25,  -3, -32,   5,   0,   3,   2,  41,  29, -14,  17,  -4,   3, -10, -11,  -4,  31,  10,  35 },
      {   1,   1,   1,   2,  -2,   3,  20, -13,   1,   1,   1,  -3,   2,  -5,  -4,  81,   0,   0,   0,  -2,  -2,   9,  -9, -20,   1,   1,   1,  -2,  -2,  -2,  22,  31,  -1,   1,  -1,  -4,   1,   1, -32,  25,  -3,   1,   1, -13, -19,  23,   7,  -8,   0,   1,  10,   9, -35, -25,   2,  -8,   3,  10,   6,  30,  15,  -6,  -6,  31 },
      {  -1,  -1,  -1,  -3,  -1,  -2, -13,  -8,  -1,  -1,  -1,   0,  -3,   5,   1, -49,   1,   0,   1,  -2,   3,   4,  20,   4,   0,  -2,   1,  -4,  -1,  -1,   3,  63,  -3,  -3,  -6,  -8, -15,   9, -28,  11,  -3,   0,  -3,   1, -34,  -6,   9, -48,  -1,   4,  -2,  11,  -3, -38,  -5, -16,   5,  -6,   7,  -2,  17,  -8,  -7, -44 },
      {   0,   0,   0,   0,  -2,   0,   8,  12,   0,  -1,   0,  -3,  -2,  -4,   4,  30,  -1,  -2,  -2,  -1,  -5,  -7, -17,   9,  -1,  -2,  -2,  -2,  -4,  -6,  -5, -33,  -1,   1,  -2,   2,   0, -17, -29,  23,  -1,   3,   2,  -2,  19,  15, -31, -59,   0,  -4,   4,  -6,  -9,  28,  14,  12,   0, -14, -12,  -3, -19, -18,  -1, -68 },
      {   0,   0,   0,   0,   0,   0,   7,   3,   0,   1,   1,   0,  -1,  -2,   8,  -3,   1,   1,   2,   1,  -1,   0,  13,  36,   2,   1,   2,   3,   4,   3,  19,  13,   0,  -1,  -2,  -6,  -3,  -6,  23,  90,  -1,   1,   1,  -4, -16, -31, -39,  18,   2,   6,   8,  17,  13,  10, -22,  27,  -6,  -3, -16,   0,  -6,   7,  -9,  10 },
    },
  },
  { // 6
    {
      {-117,  33,  14,   6,   3,   2,   1,   1,  30,   3,  -3,  -3,  -2,  -1,  -1,   0,  15,  -4,  -3,  -1,  -1,   0,   0,   0,   8,  -2,  -2,   0,   0,   0,   0,   0,   5,  -3,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -30, -34,  21,   4,   4,   1,   2,   0, -98,  13,  30,   5,   4,   1,   2,   0,   7,  38,   4,  -6,  -3,  -2,  -1,  -1,  26,  15, -14,  -8,  -3,  -1,   0,   0,  14,  -3, -11,  -1,   0,   1,   0,   0,   6,  -5,  -4,   2,   1,   0,   0,   0,   5,  -2,   0,   2,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 },
      {  -2,  68,  -8, -11,  -4,  -2,  -2,  -1, -36,  12, -10,  -4,   3,   1,   1,   0, -82, -14,  14,  13,   8,   2,   3,   1, -12,  26,  19,   3,  -3,  -3,  -1,  -1,  20,  16,  -8,  -8,  -6,  -1,   0,   0,  15,  -5, -13,  -1,   0,   1,   0,   0,   7,  -7,  -3,   4,   1,   1,  -1,   0,   5,  -3,   1,   1,   0,   0,   0,   0 },
      { -17, -49,   7,  15,   3,   4,   1,   2,  19, -76,  22,  15,   2,   3,   1,   1, -29, -12,  45,   4,  -1,  -1,   0,  -1,  -7,  37,  25, -16,  -5,  -3,  -1,  -1,   2,  25,  -7, -18,   0,  -1,   0,  -1,   1,  10, -12,  -5,   4,   0,   0,   0,   1,   5,  -6,   1,   1,  -1,   0,   0,   0,   3,  -3,   1,   0,   0,   0,   0 },
      {  14,  13,   2,  -9,  -5,  -1,  -1,  -1,  46,   8,  14, -16,  -4,  -3,  -1,  -1, -15,  55,  14, -23,  -1,  -4,  -1,  -2,  20,  38, -38, -28,  -1,  -2,   0,   0,   2, -19, -52,  -1,  12,   4,   3,   2,  -4, -23, -15,  26,  10,   2,   0,   0,  -1,  -7,   7,  15,  -3,  -1,  -1,   0,  -1,  -3,   4,   2,  -5,   0,   0,   0 },
      {   0,  15, -61,  16,   3,   1,   0,   1, -15, -38, -53,  30,  10,   4,   2,   2,  21,   7,  -1,  29,  -2,  -2,  -2,   0,  55,  17,   4,   1, -17,  -4,  -2,  -1,  11, -20,  -5,  -5,  -7,   3,   1,   1,  -8, -20,  11,   6,   2,   3,   0,   1,  -5,  -4,  16,   2,  -1,  -1,  -2,   0,  -1,   1,   7,  -3,  -2,   0,   0,   0 },
      { -23, -51, -56,   1,  11,   4,   4,   2,  13,  57, -32,  14,   8,  -3,  -1,  -1, -16,  39,   9,  13,  -3,  -6,  -2,  -2, -50,  14,  10,   1,  -6,  -3,   1,   0,  -9,  12,   2,  -1,  -5,   0,   0,   0,   6,   2,  -2,  -4,  -3,   1,  -1,   0,   7,  -4,  -2,  -3,   0,   2,   0,   1,   3,  -5,   1,   0,   1,   1,  -1,   0 },
      {  12,   0,  14,  -5,   0,  -1,  -1,   0,   3,  41, -11,  -1,  -3,  -1,  -1,  -1,  54, -33,   8,   2,  -7,  -1,  -1,   0,   8,  22,  40, -18, -10,  -5,  -1,   0,  27,  29, -23, -43,  -7,  -1,   0,  -1,  -6, -17, -43,  -9,  17,   7,   3,   1,  -4,  -8,  -9,  20,  12,   0,  -1,   0,  -1,  -2,   2,   8,  -2,  -3,  -1,   0 },
      {   6,  55, -25,   7, -10,   0,  -2,   0, -16, -13,  22,  12,  -8,   3,  -1,   0,  37,  39,  47, -13, -17,  -5,  -4,  -3, -33, -15,  10, -28,  -9,   5,   2,   1, -35,  -7,  14, -23,   6,   7,   1,   2,   0,  21,  -2, -21,   9,   1,   0,   0,   3,   3, -13,  -8,  11,  -1,   1,   0,  -1,  -2,  -7,   2,   6,  -1,   0,   0 },
      {  -6,   7,   7, -19,   9,   0,   2,   0, -30, -29, -41, -27,  18,   5,   5,   3,  30, -11, -13,  -8,  21,  -1,   3,   1, -49,  30,   1,   0,  17,  -7,   2,   0, -39,   6, -27,  13,  18,  -4,   2,  -1, -14,  11,  -1,  37,   9,  -6,  -1,  -2,   8,  12,   9,  20, -15,  -9,  -3,  -1,   5,   0,   5,   3, -15,   0,   1,   0 },
      {  -2,  -6,   7,  23,  -9,  -3,   0,   0,  -7,  31,  12,  38, -13,  -1,  -3,   0,  -9, -52,  34,  16, -14,   2,  -2,   1, -15, -15,   0, -30, -10,   4,   1,   2,  -9, -38, -27,   1,  20,  11,   4,   3,  -3, -11,  29,  35,  10,  -3,  -4,  -1,   8,  13,  29,  -1, -19, -10,  -4,  -2,  -1,   3,   3, -15, -10,   3,   3,   1 },
      {  10,  15,  36,  70,  -8,  -6,  -5,  -3,   0,  -2, -21,  47, -25,  -3,  -3,  -1,   0,  11, -45,  -3, -19,   1,   2,   2, -24,  39, -26,  -2,  -2,   4,   3,   0,  -7,  22,  -1,   2,  -5,   0,  -2,  -1,  -2,   5,   0, -10,  -6,   1,  -2,   0,   3, -10,  -4,  -7,   3,   5,   0,   1,   5,  -4,   3,  -1,   4,   1,  -1,   0 },
      {   7,   1,  56, -19,   3,   2,   0,  -2,   8,  23, -24,  -8,   8,   4,   0,  -2,  -1,  17,  26,  31,   8,  -5,  -3,  -2,  23,  43,  23,   7, -19, -17,  -6,  -2, -36, -24,  16,  -4, -15,   1,   5,   1, -23,  10,  33, -17,  -7,   3,   1,  -1,  -3,  17,  -4, -29,   0,   3,  -2,   0,   2,   2, -12,  -8,   9,   3,   0,   0 },
      { -10,  -7, -18,  15,  -4,   4,   3,   0,  -5,   9,  16, -10, -10,  -4,   1,  -1, -36,   0, -43, -17,  -4,   7,   4,   2,  26,  -5,  27,  -5,   0,  10,  -1,  -1, -59,   1,  15, -33,   8,   5,  -2,   0, -50,  -9,  -8,   1,  30,   5,   3,   1,  -8,  17,  14,  22,  15,  -7,  -4,  -2,   5,   7,   1,   8,  -3,  -6,  -2,  -1 },
      {   6,  11,  22,  12,   6,  -3,  -2,   2,   9, -13,  32,  12,  17,  -8,  -1,   1,   6,  42, -16,  45,   3, -11,  -2,  -3, -17, -32,  39,  39, -20, -12,  -3,   0,  -8,   4, -13,  -8, -20,  -4,   5,   0,  12, -26, -18,  25,  -6,   6,   3,  -2,  10,   2,  23,  18, -22,   2,  -2,  -3,   6,  -2,   7, -11, -20,   7,   3,   1 },
      {  -4, -18,  38, -13,  -6,   1,   0,   1,  -7, -23, -52,  -4,  -7,   9,   3,   1, -20,  12,   2,   8, -12,   2,  -1,   0, -27, -40, -12,  -6, -14,   9,   0,   4,   6, -42,  -5, -20,   0,  19,   7,   2,  12, -21, -10, -19,  27,  19,   0,   0,  -4, -14,  -3,  19,  35,   3,  -1,   0,   2,   2,  13,  18,   7, -13,  -4,  -2 },
      {   4,  -6,   9,  27,  -4,   8,  -3,   2,   0,  -4, -29, -40, -30,   5,  -1,   3,  -1,  24,  -3, -36, -19,  15,   2,   3,  10,  -1,  64,  -8,   9,  18,  -3,   0,  17,  -9,  13,  -1,  20,   1,  -7,  -2,  37,  -9,  16,  19,   0, -16,  -7,   0,  15,  -5,   8, -17, -21,  -9,   1,   2,  -1, -10,  -4, -14,   1,   9,   4,   1 },
      {  -6,  -2, -15, -12, -38,   8,   1,   0,  -6,  -8,  27, -14, -36,   9,   0,   1,  15, -10,   8,  -6,  -7,  15,   1,   0, -35,  36,  10,  31,   4,   5,  -5,  -4,   8, -22,  10,  22, -23,  -3,  -5,  -2, -29, -53,  11, -11, -15,  12,   6,   1,  -5, -20,  22,   1,  12,  13,   1,   1,   2,   7,  18,  -2,   9,  -5,  -7,  -2 },
      {  -3,  -9, -11, -23, -61,  11,   6,   0,   3,  -1,   5,   1, -45,  23,   5,   1,   6,   2, -18,  50,   8,  18,  -2,  -3,  15,   4, -11,  14,  13,   2,  -7,  -5, -17,   5, -23, -16,  16,  -1,  -4,   1,  30,  25, -10,  -2,  11,  -3,  -3,   2,  34,  13, -15,  -6,   2,  -4,  -3,   0,   7, -12, -12,  -2,   1,   3,   2,   1 },
      {   1,  -4, -17,  28,  29,   0,   0,   3,  -4,  -1,  13, -20,   4, -16,  -3,   2,   1, -30, -10,  -7,  -2,  -7,   2,   2,   6,   6,  11,   7,  -6, -12,   4,  -3, -41, -42, -38,  11, -16,  -1,   9,   3,  32,  -8, -19,  -7, -22,  15,   7,   3,  26, -10, -33, -16,  13,  23,   6,   1,   9, -13,  -9,  19,  27,   8,  -3,   0 },
      {  -2,   0, -12, -25,  -3,  -3,  -1,   4,   1, -14,  22, -16,  24,  -5,   3,   2,   4,  -4, -57,   3,  -5,  -6,   0,   0, -26,  12,  -2, -41, -31,  -3,  -1,   3,  24,   6,  -7, -33,  -6,  23,  10,   1,  12, -12,  40,  -6,   5,  16,  -6,  -1,   2,  15,  13, -37,   1,   3,  -5,   0,   2,  -6, -16, -14,  14,   9,   2,   1 },
      {  -7,  -5,   4, -29, -22,  -9,   5,   2,  -9,  -6, -19,  -4, -24, -16,   5,   1,  -8,  -2,   5,   2, -34,  -4,   1,   0,  13, -24,   2, -18, -30,  10,   6,  -1, -33,  43, -19,  16,  -9,  11,   7,  -3,  -9,   3,  -8,  22, -34,   6,   2,  -5, -40, -29,  -1, -12, -19,  25,   7,   1, -15,  -9,   6,  -4,  15,  20,  -1,   0 },
      {   1,  -2,   0, -42,  13,  -7,   2,  -4,   5,  -2,   2,  22, -13, -10,  -3,   0,   2, -13,  -5, -32, -38, -12,   6,   4,   8,  26,  -9,  21, -35,  11,  11,   3,  -5, -16,  39,   0, -19,  20,   3,  -2,  18,  23, -16,  12,   9,  11,  -5,  -7,  43,   0,   8,  28, -16,  -4, -14,  -5,   7, -12,   0,  -9, -16,   0,  -2,   3 },
      {   3,   2,   9, -15,   2,  21,  -2,  -1,   7,   5,  15,  21,  33,  26,  -2,  -1,   5,  -5,  -7,   8,   9,  11,  -9,   4,  13,  10,   9, -30,   3,   0,  -8,   2, -36,  19,  25,  18,  22,   0,  -3,  -1,  38, -24,  17,   0,  -1,  -7, -12,  -1,   6, -55,  10,  10,  13,  -7,  -7,   3,  17, -15,  31,  18,   5, -18,  -6,  -1 },
      {  -5,  -6,   5, -27, -15, -32,  -6,  -2,  -1,  -5,  -9,  43,   5,  -8,  -4,   3,  -5,   1, -12, -27,  41,   8,   2,   3,  -5,  -3,   5, -15,  26,  -6, -11,  -2, -10, -33,  12, -38,  -1, -22,  -8,   3,  -3, -16, -25, -10, -19, -16,   9,   6,   2, -23,   1, -12, -28,  10,  15,   4, -15, -11,  -5, -18,   2,  26,  12,   2 },
      {  -1,   5,   5, -36,   0, -19,  -2,   1,   3,  -4,  -3,  45,  11, -16,   6,   2,   0,  11, -11,  -6, -33,   4,   5,   3,   3,   0,  32,  -5,  26,  13,   6,  -2,   0,   6, -17,  47,  20,   0,  -9,  -3,   4, -34,  -7,  -9,   1,   0,  -6,   3,   4,  46, -19, -10,  24,  -4,   6,   3,  11,  23, -23,   3,   5,  -8,   3,   1 },
      {   2,   5,   3, -11,  19,  75,  -1,   3,  -1,  -1,  -3,  21,   3,  65,  -8,  -1,  -7,  -3,  -8, -26, -21,  -4, -15,  -4,  -4, -12,  -3,  12, -10, -22,  -9,   4,   4,  -1, -13,  -2,   5, -12,   1,   3, -18,  -5, -23,   5,   2,  -4,   1,   1,  -2,   9,  -4, -20, -10,  11,   2,   2,   4,   6,  -8, -19,  13,  18,   4,   1 },
      {   2,   1,   4,  -2,  18,  -2,   7,   4,   0,   1, -17,  -4, -11,  -6,   4,   4,   3,  -2,  24, -18,   9,   0,   3,   4,  13, -15, -22,   4,  27,   2,   5,  -4, -20,  39, -13, -25, -16,  -8,  -5,  -3,  20, -16,   7, -16,  -9,  21,   4,   4,  34,  17,  59, -14,  14,  22,  -6,  -4,  27,  20,  10, -29,  10,  -2, -14,  -4 },
      {   2,   6,   1, -19,  64,   5,  15,  -4,  -1,  -5,  13,  -1, -17, -12,  -3,   0,   3,  -1,  -3,  34, -24,   6,   3,   6,  -7,  14,  -9,  27,  18,  42,   2,  -3,  -5,  -6, -14, -28,  24,   9, -29,  -2,  -2,   0,   1, -11,   7, -24, -15,   9, -15, -23,  10, -17,   0,  -7,   8,   8, -29,  -5,   7,  -6,   8,   2,   7,   0 },
      {  -1,  -4,  -1,   1,   2,  -6,  -3,  -9,  -3,  -9,  -6, -26, -11, -15,  -2,  -6,  -3,  -7,   7,  23, -27, -40,   0,   2,   0, -11, -21, -14,   3, -21,   8,   5,   0,  11,  21,  -1,  19, -17,   3,   2, -32, -43,  -6, -11,   6, -27,   0,   5,  36, -22, -25, -13, -15, -13,   0,   5,  44, -10, -13, -10, -17,   6,  10,   0 },
      {  -2,  -2,  -4,  -6,  -6,   8,  -6,   1,  -3,  -1,  -2,  -9,  -6,  -6,  -6,   2,  -5,  -8,  -8,  -8, -37, -15, -10,   2,   1,   6,  -5,  -6,   4, -33,  -9,   5, -19,  -4, -15,  -2, -23, -40,  -1,   7,  39,  -2,  18, -40,  -2, -16,  10,  11, -31,  18,  14,  21,   2, -24,  15,   0, -28,  -5,  10,  -7, -43,  -9,  18,   0 },
      {  -1,  -1,  -1,   1, -28,  15,   0,  15,  -1,   2,   2,   9,   4, -29,  -5,   2,   1,  -4,   0, -11,   9, -63,   2,  -4,   4,  11,  10,  33,  22, -19,  27,  10,   0,  -4,   3,  -1,  39,  20,  26,   5,  18,   5,  -9,  -7,  30,   6,  -9, -10, -28,  -8,  12, -21,   3,   3, -13,  -5,  -7,   0,  21, -11,  15,  10,  -9,  -1 },
      {   1,   3,   0, -22,  28,  -7,  12,  -7,   1,  -5,   5,  16, -51,  -6,  -2,   0,  -1,  11, -17,   5,   3, -19,  -6,   1,  -3,  -4,  22, -17,   2, -37,   1,   3,  21,  -6,   0,  -5,  12, -35,  20,  10, -10,  28,  19,  28, -13,  -9,  27,   2,  13, -10,  11,   0,  21,   6,   4,  -7,   1,  16,  17,  29,  20, -24, -15, -10 },
      {  -3,  -5,  -6,   1, -10, -34, -37, -15,   0,   1,   2, -12,  10,  15, -20,  -6,   0,  -1,  -6,  -6,  -3,   8,  13,   5,   0,  -1,   3,   9, -44,  -4,  -1,   4,  -2,  -8, -19,   6,  18,  -3, -13,   4,  -2,  23, -29, -18,  -7, -27, -14,   8,   1,  -2,  17, -31,  -3, -14,  -1,  20,  21,  41,  41, -11,   4, -24,  -4,   3 },
      {   0,   0,  -4,   3,  20, -15,   8,  -5,  -2,   1,   2,  -2, -31,   3,   7,   3,   5,  -4,   6,   2,  25,  19,  14,  -1,  -5,   8, -19,   6, -30, -26,  -3,   4, -17,  10,  31,   6,   1,  -9,   3,   7,  39, -32, -16,  32,  36,   2,  -3,   1, -35,  -1,  -9, -13,  11,  -9,   2,   2,  -5,  35, -37, -14,   5,  -2,  10,   1 },
      {  -3,  -3,  -1,  -8,   3, -12, -54, -15,  -3,  -4,  -4,  10,  -9,  -3, -59,  -5,  -1,   1,   9, -17,  13,   6, -28,   7,   3, -11,   1,  27,  -1,  -4,   3,   7,  -4,  23, -21,  -1, -12,   4,   6,   8, -11, -11,  25,  11,  18,   2,  -4,   0,   9, -10,  -1,  -7,  17, -21,  -9,   1, -17, -41, -28,   4,  13, -16,   1,   6 },
      {   0,   0,  -1,  -1,   1, -21,  25, -14,   3,   0,   3,  13,  10,  14,  54,  -4,  -2,   2,   4, -24,  -8,  13,  46,  -5,   2, -13,  10,  23,  -5, -21,   5,  -8,   2,   4, -31,   2,  -7, -21,  -4,  -1, -14,   9,  28,  -7,  34,   9,  -5,   3,   0, -29, -14, -17,   2,  -8, -15,   0,   4, -35,   4, -17,  -9, -13,  -8,   4 },
      {   0,   0,   1,   8, -17, -32,  16,   4,   2,   3,  -1,  -5,  28,   3,  13,  12,   0,  -6,   7, -11, -30, -12, -40,   0,   1,   5,  -6,  33,  -5, -11, -58,  -4,   2,   7,   4, -11,  28,   9, -21,   8,   3,   8,   9,  16,  12,   4,  27,  17,   3, -10,  13,   5,   3,  14,  38,   8,   9,   1, -17,  11,  19,  -1,   4,  -7 },
      {   0,  -1,   5,   9, -22,   1,   1,   0,   0,   1,  -3,  -9,  25,   0,  11,   4,  -1,  -2,   5,   6, -11, -12,  20,   0,  -2,   0, -11,  -2, -20,   0,  20,   6,  -1,   3,  15,  -1,  -1, -42, -22,   1,  -2, -12, -10,  20,   0, -38, -21,   0,  29,  23,  19,   8,  30,  13,  -8,  -4, -54, -19,   0,   8,  43,   7,  -4,  -4 },
      {  -2,  -2,  -3,   0,  10, -22, -48, -18,   1,   2,   6, -10,  13,  19, -11,  -2,   1,   2,   2,  23, -38,  31,  26,  16,   0,   9,  -4,  -2,  29,  -8,   8,   5,   4,  -4,  16,  -3,  17, -18,  28,   5,   6,   9,   2,   4,  21,  18,  25,  -2, -15,   5,   1,  14,  -3,  38,  -6, -13,   7,  -6,  12,  -8,   7,  39,  -7,  -7 },
      {  -1,   0,  -2,   8,   8, -21,  16, -41,  -2,  -1,   0, -11,  -3,  25,  -2, -28,  -3,  -1,  -7,   8,   5, -15, -23, -19,   3,   0,   8, -18,   5,  -6,  -5,  -6,  -3,   5,   7,  28,  20,  17,   3,   4,  -6,  -5, -21, -15,   2,  37,  10,  -3,  29,  -5,   6, -13,   6,  12,  -2,  -9, -63,   1,   5, -29, -25, -10,   0,  -2 },
      {  -1,  -1,  -1,   4,  -4, -11,   1, -20,   0,   3,  -2,   1,   8,  31,  16, -10,   0,  -3,   6,  -9, -13,  -5,  16,  -4,  -1,   3, -10,  24,   2,   7,  14, -11,  -7,   5,   1, -34,  23,  13,  43, -12,  11, -25,   2,  -5, -49, -11,  22,  -7,  -8,  11,   5,  -1, -24, -27, -21,  -4, -13,   6, -27,  37,   4, -26, -12,  -6 },
      {   0,   1,  -2,  -1,  19, -30,   0,   3,   1,   1,   3,   2, -30,  26,   8,  23,   0,   1,   5, -10,  17,  -9,  -4,   8,   1,  -4,   1,   6, -17, -24, -10,   4,   3,   9,  -3,   7,  24,  14, -14,  -3,  -3, -12,  10, -25,  -8,   3, -41, -11,  -3,  24,  10,   6,  -1,  16, -16,  -7,   4, -20,  16,  48, -16,  47,  26,   5 },
      {   0,  -1,   0,   2,  -2,   3, -15,  30,   1,   2,   2,  -1,   2,  20,  -6,  31,   0,   0,   0,   4,   9, -17,  27,  19,   1,   4,   3,   2, -16,  20,  25,  10,   1,  -6,   1,  -2,  28, -17,  -4, -17,  -5,   9,  -3,   6, -19,  21,  12, -16,  -1, -15,  16, -10,  31,  26,  35,  -1,   3, -15, -33,  -9, -44, -30,  23,   4 },
      {  -1,  -2,  -3,  -1, -11,  -5,   3, -55,   1,   1,   1,   3,   2,  24,  -3, -35,   1,  -2,  -2,  -8,   6, -38,  -2, -10,   2,   3,   9,   8,   8,  24,  25,   8,  -2,  -3,  -9, -19, -19,   2, -18,  15,  13,   4,  30,  20,  -7, -15, -22,   5, -13,  -9, -16,  20,  23,  29,  14,  -2,  23,  29, -10,  -1, -12,  17,  14,   0 },
      {   0,   1,   1,  -4,  -1,  28, -31, -19,   0,   1,  -1,   7,  -8, -12,  13,  -4,   2,   2,   6,  -5,  14,  -7,  34,  11,   3,  -3,  -5,   3,  -6, -11,  10, -10,  -3,  14,   7,  -6,  -1,  24, -18, -12,   4, -17,  18,   6,  18,  15,  22,  18,  20,  20, -20,  -8, -28,   7,  57,  23, -22,   4,  36,  27,   1,  -8,   7,   1 },
      {  -1,  -1,  -1,  -2,   0,  19, -26, -29,  -2,  -2,  -2,   0,  -8, -21,   9, -27,   0,  -2,   1,   5,  -2, -23,  18,   2,  -4,   0,  -3,  -9,   8,  -1, -56,   0,   3, -12,  11,  -5,   5, -14, -40, -13,   7,   2, -19,  15, -14,  23, -15, -12, -27,  17,   8, -14,  -4,  14, -30,  -2,  15, -29, -11,   9,  -8, -27, -23,   3 },
      {  -1,  -1,  -3,   2,   0, -19, -19,  18,   0,   3,   3,  -4,  -1,  28,   5,  41,  -3,   1,  -3,   6, -16, -31,   0,  36,   2,  -3,   1,  -7,  33, -17, -15, -10,  -5,  -2,   0, -13, -28,  13, -14, -29,  -2,   0,  10,  19,  -3,  10, -26, -17,   7, -15, -31,   8,  -6, -10, -18,   0, -25,  37,   7, -18,  13, -16,  -5,   8 },
      {   0,   0,   0,  -3,   6, -21,  -9,  41,  -1,   1,   0,   1,  -3,  16,  -8,   4,  -1,  -4,   0,  -5,  -1,  -9, -22, -52,   0,   3,  -7,   6,  -5,  27,   4, -30,  -4,   2,  10, -16,  -1, -20,   2,   7,   7, -17,   8,  24,   7,  -4,   3,  18,  -8,  23, -37, -22,   1,  25, -12,   0,   0, -13,  37,   0, -28, -13, -26,   0 },
      {   0,  -1,   0,   2,  14, -19, -18,  -2,   1,   1,   4,  -3, -16,  33,  32,   7,   0,  -2,  -3,  11,   7, -21,  -9,  -9,   0,   2,   4, -17,  17,  -7,  19,  -9,   1,  -7,   8,  18, -26,  36,   4,  -1,   2,   5, -24,   0,   1, -27,   5,  19, -22,  17,  26,  10,   7, -10,  29,  19,  20, -48,  -7, -21,  24, -13, -14,   4 },
      {   0,  -1,   1,  -2,   3,   9, -32,  15,  -1,   0,  -1,   1,  -4, -10,  -1, -16,   0,  -1,  -2,   1,  -2,  -8,  21, -52,  -1,  -3,   3,  -2,  -3, -12, -10, -72,   2,  -1,  -3,   6,  11,   0,   4, -32,  -2,  13,   9,  -8,  -1,   1,  -5,  11,   8, -24,  19,  23,  17,  -5,  -8,  16,  -4,  20, -24,  -1,   8,  24,  22,  11 },
      {   0,   0,   0,   0,   1,  -3,  13, -15,   1,  -1,   1,   1,  -6,   9, -21,  13,   0,   0,  -4,  -1,  22, -24,  -6,  25,  -2,  -1,   3,  -3, -32,  22,   7, -33,   0,   0, -11,  15,  12, -16, -27, -38,   9,  -7,  11, -26,  20, -11,  41, -15, -11,   6,  -9,  21, -16,  -4,  -3, -27,   9,  -3,  -5, -16,  15,  16, -38, -23 },
      {   0,   0,   1,   0,   2,  11, -12,  -9,   1,   2,   0,   3,  -4,   4,  13,  30,   1,   0,   2,  -1,  16, -31,  28,  21,   0,   0,   2,  -2,  -9,  41, -44,   3,  -1,   7,  -2,  12,  -5,   1,  27,  39,  -2,   1,  -1,   3,  -7,  -2,  24,  51,   7,  -2,  -4,  -4,  26, -23, -25,   6,  -9,   3,   5,  -7, -12,  30,   0,  -1 },
      {   0,   0,  -2,  -1,   9,  14, -33,   0,   0,   0,   0,   3,  -8, -24,  37,   6,   0,   0,  -1,  -6,  13,  16,  -9,  -3,   0,   0,   0,  10, -18,  -9, -19,   5,  -1,   5,  -7,  -7,  28,  22,  15,  20,  -2,  -9,  16, -16, -19, -37, -10, -21,   6,   4, -24,  37,   7,  36,   1, -20,  -8,   2,   1, -46,   0, -23,  -7, -13 },
      {   0,   0,   1,  -1,  -2,   3,  -6,  20,   0,   0,   0,  -2,  -1,   5, -11,  11,  -1,   2,  -4,   4,  -2,   1,  14,  -8,   0,  -1,   5,  -6,   1, -22,  28,   8,   0,  -1,  -2,   4,  11,  10, -40,  45,  -2,   3,   6, -16,  -6,  17, -16,  51,  -1, -16,   6,  19, -31,  12, -10,  -2,  -9,  13, -38,  11,  -8,  -6, -49, -18 },
      {  -1,  -1,   0,  -2,   0,   5, -42,  -2,   1,   1,   1,   1,   0,  -4,  49,   8,  -1,  -2,  -1,   0,   0,  -4, -31,  -8,   1,   1,   1,  -3,   3,  23,   5,   0,  -3,  -2,  -6,  11, -27, -11,   4,   4,  -2,   6, -11,  -6,  33, -13,  -3, -21,  10, -21,  22, -34, -15,   0,   7, -37,  -6,  22, -18,  39, -19,   7,  -6, -29 },
      {   0,   1,  -1,  -1,  10, -18, -15,  54,   0,   0,   2,   0, -11,  23,   8, -56,   1,  -1,  -1,   2,   7, -23,  23, -20,   2,   1,   2,  -1,  -2,  20, -20,  56,  -4,   6,  -7,   7,  -2,  -7, -12,   4,   2,  -4,  11, -12,   4,   9,  22, -23,   7,  -2,  -4,  17, -14, -17,   6,  -4,  -4,   7,  -1,  -3,  26,   7,   5,   7 },
      {   0,   0,   0,   1,   3,   3, -14, -21,   0,   0,   1,  -1,  -5,  -7,  32,  21,  -1,   0,  -1,   2,   7,  -3, -41,  -6,   0,   3,   1,   1, -12,  23,  18,  19,   1,  -2,  -1,   0,  18, -45,  13,  -9,   0,   1,   7, -12,   3,  34, -11,   3,  -1,   0,  -1,  21, -21,   0,  -4,  54,  -1,   8,  -6,  -4,  16,   2, -15,  50 },
      {   0,   0,   0,  -1,  -1,   2, -13,  20,   0,   0,   0,   1,   1,  -7,  27, -42,   0,   0,   1,  -4,   1,  15, -29,  36,  -1,   0,  -1,   7,  -6, -21,  19,  17,   0,  -3,   2,  -8,   8,  15, -16, -44,   0,   1,   1,  14, -14, -12,  20,  28,   1,  -5,  -5, -10,  37,  -8, -18,  13,  -8,   4,  -1,  -2, -33,  34, -31,   7 },
      {   0,   1,  -1,   1,  -2,  -3,  17,  13,   0,   0,   0,   1,  -2,   5, -14, -20,   0,   1,  -1,   0,   5,  -7,   6,  28,   0,  -1,   4,  -5,   2,   8, -18,   2,   0,   2,  -6,  14, -15,   6,  31, -21,   2,   0,   2, -15,  35, -40,   2,  20,   3,  -1,  11,   0, -24,  65, -26,  33, -10,   5, -19,  16,  -9, -28,   8,  13 },
      {   0,   0,  -1,  -1,   2,  -2,   8, -14,   0,   0,   1,   2,   0,   0, -21,  22,   0,  -1,   0,  -1,  -2,   2,  21, -38,   2,   0,   0,   1,   3,  -2, -28,  40,  -2,   1,  -3,  -1,  -5,   3,  28, -44,  -2,  -1,   4,  -1,   2,  -9, -29,  -9,   7,  -2,   0,   1,  11,  -4,  33,  29,  -5,   1,  -7,   0, -18,  20, -53, -38 },
      {   0,   0,  -1,   0,   0,  -5,   0,  17,   0,   0,   0,   0,  -4,   4,   7, -40,   0,   1,   1,  -2,   5,  -4, -12,  42,  -1,   1,  -1,   3,  -8,  13,  -3, -24,   2,  -1,   3,  -1,   8, -27,  33,  -6,  -3,   2,  -2,   1,  -4,  30, -55,  31,  -3,   6,  -4,   8,  -9, -11,  25,  -2,   2,  -9,   3, -10,  12,  -7,  10, -58 },
      {   0,   0,   0,   0,   0,  -2,   8,   9,   0,  -1,  -1,  -1,   1,   3,  -8, -20,   1,   0,   0,   1,  -2,  -1,  11,  25,  -1,   0,  -2,   0,  -2,   1, -16, -29,   1,   2,   1,   0,   0,  -3,  22,  25,  -2,  -3,  -2,  -4,   2,  -9, -27, -16,   4,  -2,   6,  -4,   4,  -6,  46, -31,  -4,   3,  -5,   8, -11,  12, -54,  72 },
      {   0,   0,   1,   0,   1,  -1,   2,  10,   0,   0,   0,   1,   1,   3,   1, -11,   0,   0,   0,   0,  -1,  -2,   0,  22,   0,  -1,   0,  -1,   2,  -1,   3, -33,   0,   1,   0,  -1,  -4,   0,  -9,  51,   1,  -1,   1,   1,   0,  -1,  11, -65,   2,   2,   2,  -2,   9,  -3,  -6,  75,  -5,  -1,  -4,   4, -15,  15, -15, -33 },
    },
    {
      {  77, -59,  10,  -3,   2,  -1,   1,   0, -67,  42,   3,  -3,   1,   0,   1,   0,  17,   1, -13,   4,  -1,   0,  -1,   0,  -4,  -6,   6,   1,  -1,   0,   0,   0,   2,   3,  -1,  -2,   0,   0,   0,   0,  -2,  -1,   0,   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  56,   5, -30,   2,  -3,   1,  -2,   0,   6, -63,  43,  -1,   3,   0,   2,   0, -44,  55,  -2, -17,   5,  -3,   1,  -1,  12,  -5, -23,  13,   0,   1,   0,   0,  -5,  -1,  11,   0,  -3,   1,   0,   0,   2,   1,  -4,  -1,   1,   0,   0,   0,  -3,   1,   2,   1,   0,   0,   0,   0,   1,   1,  -2,   0,   0,   0,   0,   0 },
      {  51, -11,  13, -15,   0,  -2,   0,  -1,  35, -19, -39,  20,  -2,   3,  -1,   1, -18, -39,  59,  -6,  -4,   1,   0,   0, -24,  47, -13, -23,  11,  -4,   2,  -2,   1,  -8, -14,  18,  -1,  -1,   0,   0,  -4,   4,   8,  -4,  -3,   1,   0,   0,  -1,  -1,  -3,   1,   2,   0,   0,   0,  -2,   2,   2,   0,  -1,   0,   0,   0 },
      { -13, -68,  47,   4,   4,   0,   2,   0,  57,  19, -10, -26,   3,  -4,   0,  -2, -55,  24, -23,  18,   5,  -1,   2,   0,  16,   1,   4,   5, -11,   3,  -2,   0,  -4,  -3,  -1,  -5,   3,   1,   0,   0,   1,   2,   0,   3,   0,  -1,   0,   0,  -2,   0,  -1,   0,  -1,   1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0 },
      { -42, -23,  30, -15,   9,  -1,   3,  -1, -30,  10,   6,  20,  -4,   0,  -1,   1,   8,  35,  21, -42,   3,  -3,   1,  -2,  -3,  24, -65,  17,  14,  -5,   3,  -1,  17, -25,  14,  25, -19,   4,  -2,   1,   1,   2,   4, -16,   2,   3,  -2,   1,   3,  -3,  -2,   5,   2,  -2,   0,   0,   1,   1,   0,  -2,  -1,   0,   0,   0 },
      {  12,  43,  20, -29,  -2,  -5,   0,  -2, -26,  14, -60,  25,   7,   1,   1,   1, -37,  21,   8,  22, -19,   5,  -2,   1,  54, -33,  -5,  -8,  -2,   7,  -3,   2,  -8, -17,  22,   0,   4,  -2,   0,   0,   4,   7,  -1,  -9,   1,  -1,   1,   0,  -2,  -4,   1,   4,   1,   0,   0,   0,   2,   1,   1,  -3,  -1,   0,   0,   0 },
      {  30,  19,  -6, -14,   8,  -3,   0,  -2,  38,  22, -22,   5, -14,   0,  -2,   1,   9, -23, -33, -16,  23,   1,   4,   0, -25, -13, -12,  64,  -8,  -6,   1,  -1,  12, -16,  44, -17, -28,  13,  -4,   2, -10,   4,  -5, -15,  19,   0,  -1,   1,   0,  -1,   3,   8,  -4,  -4,   1,   0,  -3,   1,   0,  -3,   0,   2,   0,   0 },
      {  -6, -36,  27, -33,  10,   1,   3,   0,  17, -42,  -4,  45,  -2,  -1,  -1,   1,  39,  11,  11, -19, -19,   6,  -3,   1,  -6, -42,  24,  -5,  16,   1,  -1,   0, -16,  37,  19, -23,   2,  -8,   2,  -2,  -6,   9, -29,   1,  10,  -1,   2,  -1,  -2,   3,   5,  10,  -6,  -1,   0,   0,  -1,   2,  -4,  -5,   0,   2,   0,   0 },
      { -21, -14,  -2,   3,   9,  -2,   1,   0, -40, -22,   2,  16, -13,   9,   0,   3, -30, -16,  27,  12,  22,  -5,   1,  -1,   8,  38,  25,  16, -42,  -2,  -2,  -1,  20,  -9,  12, -58,   5,  18,  -5,   3,  -6,  12, -21,   7,  27, -15,   2,  -1,   6,  -1,   1,   6, -13,   0,   3,  -1,   0,   2,  -2,  -3,   4,   1,   0,   0 },
      {   9,  30,  61, -35,  -2,  -7,   0,  -3, -14, -41,  -6, -20,  25,  -2,   5,   0,  20,  24, -35,  33,  -2,  -9,   3,  -2, -45,  23,   3,   3, -19,   8,  -1,   0,  25,  -4, -14,   1,   4,   5,  -4,   2,  -3, -10,   7,   4,  -3,  -2,   0,   1,   3,   2,   0,  -3,   1,   1,   0,   1,  -2,  -1,   0,   2,   0,  -1,   0,   0 },
      {  16,  23,  13, -15,   8,  -8,   0,  -2,  22,  32,  11,  -7, -23,   4,  -4,   0,  19,  17,   1, -33,   8,  -7,  -1,  -1,   7,   0, -28, -21, -12,  10,   2,   2,  -8, -19, -26, -23,  51,   5,  -2,   1, -12,   7, -22,  52,   8, -26,   7,  -3,   3,  -9,  10,   7, -25,  10,   1,  -1,  -2,  -2,   0,  -1,   6,   1,  -2,   1 },
      {   2, -25, -15,  33, -17,   0,  -1,   2,  14, -20, -12, -12,  15,  11,   1,   1,  46,   3,  14,  27, -14, -18,   0,  -2,   3, -15, -22, -14, -26,  29,  -1,   4,   8, -44,  44,   1,  30,   1, -12,   3, -13,  24,  13, -20,  -4, -18,   6,  -1,  -3,   1, -11,   0,   6,   5,   2,  -1,  -3,   5,   3,   1,  -1,  -3,  -1,  -1 },
      { -11, -15, -54, -17,  13,   2,   5,   1,  11,  22, -12,  51,  -4,  -5,  -3,  -2, -19,  36, -17,  11, -33,   7,   0,   1, -42,   5,  16,  -8,   2,  17,  -6,   3,  46, -16,  -8,  -2,  12,  -8,  -3,   1, -12, -13,  14,  11,  -7,  -1,   3,  -1,   8,   3,   6,  -9,  -4,   2,  -2,   1,  -3,  -5,   0,  -1,   2,   0,   1,   0 },
      {  11,  11,   3,   9, -20,  11,  -4,   1,  20,  24,  17,  -2,  11, -16,   1,  -2,  18,  26,  12,   3, -34,   8,  -4,  -1,   8,  14,   4, -29,  -7,  -9,  -1,   0,   5,  -8, -17, -29, -27,  28,   4,   3,  -2, -20,  -7, -32,  56,  10, -12,   4,  -8,   9, -16,  29,   7, -29,   9,  -2,   0,  -5,   1,   0, -13,  12,   2,  -1 },
      {   0,  19,  21,  15, -27,   4,  -1,   2,  -9,  20,  22, -21,  18,  -2,  -5,  -2, -38, -16,   9, -19, -20,   0,   7,   2, -33,  -5,   2,  -8,  32,  26,  -8,   2,  31,  14,  30, -12,  15, -38,  -4,   0, -14,  17, -39, -14,  -3,   0,  18,  -5,   6,  -3,  -8,  24,   5,   5,  -8,  -2,   1,  -1,   1,  -2,  -8,  -2,   1,   1 },
      {   1,   0,  17, -32,  21,   3,   1,   0,  -1, -15,  50,   7, -26, -10,   0,  -1,  -9, -43, -30,  -5, -12,  28,  -1,   4,  27,  -2,  16, -19,  21,   2,  -4,   2,  12, -44,   7,   8,  13,  -2,  -1,  -1, -14,  19,  34, -14,   9, -12,  -1,  -1,  -3,  17, -23,  -5,   1,  -7,   6,   0,  -3,  -1,   0,   8,  -1,   2,   0,  -1 },
      { -13, -21, -32, -19, -13,  18,   0,   3, -12, -20, -25, -21,  47, -17,   7,  -1,   0, -17, -21,   9,  -1,  13,   4,   1,   6,  10, -18,  10,  31,  -6,  -6,   0, -11, -13,  -6,  12,   3,  -5,   8,  -1,  -9,  13,  -7,  35,  28,  -9,  -3,  -3,   4,   6,  11,  36, -39, -17,   3,  -2,   2,  -4,   9, -15, -14,  17,  -1,   1 },
      {  -3,   2,  16,  43, -11,   3,  -7,   1, -17, -15, -32,  -5, -26,  14,   5,   3, -21,   9,  -1, -27,  15,  15, -10,   3, -34, -40,  10, -10,  16,  -6,  14,   1,  14,   7,  -5,  14,  18,  30, -12,  -3,   4, -10,  29,   2,  25, -35, -14,   7,  -3,  19,  -9,   4, -16, -13,  20,  -3,   3,  -4,   1,   3,  -4,  10,  -3,  -3 },
      {   9,   1,  13,  36, -14,   5,  -5,   0,   4, -17, -27,  -3, -10,   5,  -4,   3,  23,  22,  -3, -18,  20,  11,  -2,   0,  31,   4,  28,   4,  21, -28, -14,   0,  43, -31, -20, -12, -18, -32,  26,  -1, -38,   8,  -3,  12, -12,  24,   8,  -8,   9,  -3,   0,  -3,   1,   5, -12,   1,  -7,  -2,   1,   0,   2,  -2,   3,   0 },
      {  -6,  -8, -33, -38,  32, -23,   6,  -3,  -1,  -3, -21, -37, -17,  34,  -1,   4,   6,  16,  -4,   2,  34, -10,  -6,   4,   6,   4,   7, -19,  13,  -3,  11,  -4,  13,  13,  -8,   8,   8,  -8, -10,   0, -11,  -3, -19, -28,   7,  -4,  17,   4,  -8,   5, -25,  37,  29,   0,   2,  -9,   2,   0,   8,  21, -23, -10,  -3,  -1 },
      {   7,  13,   2,   8,  21,  -1,  -9,   1,  10,  18,  13,   4,   6, -14,  11,  -3,  12,  26,  14,  28,   6,   9, -17,   1,   0,  20,  16,   2,  12, -36,   4,  -3,   2,  14,  18,  14, -14, -10,  -9,  -4,   4,  24,  -6, -14, -19, -43,  14,   4,  -6,   3, -29,   1, -47,  29,  22,  -8,   6, -12,   5, -16,  11,  21, -17,   3 },
      {   5,  -5,   3, -16, -25,   6,  -1,   3,  12,   3,   6,   9,  34,   6,  -8,   0,   8, -15,  12, -22,   7, -23,   7,   2,  46,  -4,   5,  27,  -3,  20,   3,  -8,  41,  20, -17,   0,   3, -11, -33,   7,   4, -47,  23,  -4,  -2, -27,  22,   8,  -8,  26,   2,  -6,  -9,   7,   8, -13,   1,  -8,  -8,  -2,  -1,   4,  -5,   2 },
      {  -5,   0,  -5, -37,  -7,   9,   0,  -1,   2,  17,  22, -13,  33,  -4,   0,   1, -12,   2,  44,   2,  23, -27,   2,  -5, -17, -51,  11,  -2, -19, -24,  11,   0,  15,  -5, -11,   9,  -9,  21,  17,  -7, -40,  32,  15,   7,  -5,  10, -18,  -4,  15,   7, -12, -12,  -2,  -5,  -2,   8,  -6,  -3,  -1,   4,   3,   1,   4,  -1 },
      {   1,   2,  14,  33,  19, -31,   8,  -2,  -3,  -8,  16,  35,  -7,  13,  -5,  -6, -12, -21, -16,  21, -14, -32,  -1,   4,   0, -20, -19,  -4, -15,   0,  22,  -7,  -6,  -2, -26,  19, -23,   4,  -5,   3, -38,  -6,  -4,   3,   3,  14,  23,  -1,   9,  10,   4,  39, -16,  16, -14, -11,   0,  -9,  19, -18, -17,   8, -10,  11 },
      {  -1,   8,   5,   3, -26,   6,   9,   2,  -1,  14,   1,  37,  12,  15, -22,   1,   0,  21, -14, -17,  31, -28,  14,   1, -15,  24,  34,   5,  -4,   9,  -8, -11, -54, -16,  -1,   7,   8, -22,  -2,   9,   8,  30,  21,  -3,   4,   1,  12,  -7,  -6,  23, -22,   7,  -1, -21,  -6,  -2,   4,  -5,  -6,   6, -25,   6,  10,   2 },
      {  -8,  -8, -11, -16, -39,  26, -16,   7, -10,  -9, -20,   2,  -3, -25,  14,  -2,  -9,  -5, -25, -30,  -7,   4,   2,   0,  -9,   1,  -3,  -8, -18,  -2,  16,  -2, -26,  -6, -20, -12,  -6,  16,  -1,   5, -18,   3,  -6, -20,  -5,   1,  26,   0,  -6,  -5, -30,   1,  11,  53,  -2,  -9,   3,  -6,  16,  11,  36, -10, -27,   6 },
      {   4,  -1,  -1,  -2, -35,  11,  -3,   1,   8,   0,   2,  29,  -5,  14,  -3,   3,   4,  -4, -37,  -2,  13, -12,   2,  -2,  27,  22,   3, -18,  -4,   9,  17,  -1,  27,  32,   1,  31,   2,  24, -13, -16, -11,  33, -23,  -9,  -7,  -9, -39,  11,  22, -40,   1,   1,   1, -12,  12,  14,  -4,   2,   8,  -5,  -1,   1,   5,  -5 },
      {   3,   8,  -1,  22,  46,  -2,   3,  -1,  -4,  -3, -20,  -9,   5, -37,   3,  -3,   4,   6,   6, -17, -26,  11,  24,  -7,  13,  10,  -6,  37, -14,  33, -15,  -4,  -5,  37, -25,   1,  13,   2,  -7,  -2, -32,  30,  11,  -5,  -2,  -3, -14,  -1,  15,   8, -35,  -4,   8, -11,   2,   7,  -7,  -9,   1,  18,  -8,   3,   5,  -1 },
      {  -4,  -3, -10, -21, -34, -37,  21,  -1,  -2,  -1,  -7, -39, -37,  32, -12,  -4,   4,  11,   6,  -4, -36,   0,  12,   9,   6,  15,   8,   4,  10,  17,  10, -10,  -3,  17,   9,  -4, -19,  13,  -3,   4,  -7,  19,  20,   2,   6,  17,   1, -10,   8,  16,  -3, -20, -29,   3, -18,   3,   2, -18,  -7, -29,  21,  18,   1,  10 },
      {   4,  -3,  -5,   8,  26, -11,  16,  -2,   1,  -9, -14,  -4,  18, -15, -19,   3,  -1, -14, -19, -34, -12, -34,  17,   4,   4,  -9,   3, -23, -19,  -9,  11,   2,  37,  -6,   0,   2,  -4,   8,  32,   1,  61,  19, -13,   1,  -4,  11,   0, -20, -25,  -1, -22,  -2, -15,   3,  -4,   3,  15,  -7,  -6,   5,   3,  21,  -3,   0 },
      {  -1,  -2,  -1,  -5, -32,   9,  -9,  -4,   3,   3,   1,  21, -28,  -7,  12,   3,   2,  -6,  -8,  33,  21,   9,  -5,  -6,   3,   0, -53,  -5, -10, -13,  -9,  11,  19,  43,   8, -16,  19, -12,  32,  -1,   4,   2,  10,  10,  -4,  13,   0, -20, -12,  42, -29,   1,  -3,  -4, -15,   4,   4, -14,  -5,  12, -16,   4,   2,   3 },
      {  -5,  -4,  -6,  -1,  -8, -19,  16, -11,  -7,  -6,  -7,  -1,   3,  10, -30,  11, -10, -12, -17,  -1, -13, -18,  -3,  -2,  -7,  -3, -26,  -9,   2, -21, -14,  12, -19,  -3,  -8, -27,  -9, -23,   7,   1, -30, -18, -10, -12, -17, -14,  -5,  22,  -5,  -7, -13, -27,  -4,  -9,  56,   9, -19,   3, -15,  13,   2,  40,  13, -21 },
      {  -1,  -5,   0, -16, -15,  21,  -4,  -3,  -2,  -7,   5,  -9, -50, -19,  -5,   7,   3,   0,  20,  13,  12,   1,  29, -12,  -8, -27,   5,  11, -21,  21, -32,   0,   6, -10, -26,  15,  -9, -23,   0,  16,  25,  -8,   1, -17,  -5,  -1,   8,  -2,  10, -40,  -8,  22, -18,  -4,  12,   5,   1,  -1,  22, -21, -15,  28,   6,  -5 },
      {  -1,   3,  -2,  -4,  13,  25, -21,   4,  -6,   1, -10, -17, -27, -40,  11,   5,  -1,   5,  -4, -16,   8, -24,  16,  -6,   3,  16,  14, -35, -21,  14,  11,   0,  -4,   6,  38,  22, -15,  -2,   1,  -8, -35, -18,  -8,   4, -10,   7,   4,   4, -16,  32,  34,  -7,  -9,  -4,   4,  -9,  -8,   9, -26, -13, -21,  18,   3,  -2 },
      {  -3,  -5,  -4,  -8, -31, -58,  -1, -10,   2,   1,   3,   0,   5, -16,  40,  -4,  -1,  -3,   3, -27, -14,  26, -40,   0,   1,   3,  17,  20, -42,  -7,   7,  19,   2,  -6,  -8,  19,  17, -10,  15, -10,   9,   9,  -4, -10, -12,   3,   6,   8,   4,   1,  10,  12,   2,  -4,   9, -11,  -8,  12, -14,  -3, -33,  11,   3,  -3 },
      {  -2,  -1,   3,   0, -12,  -4,  19, -13,  -2,   0,   6,   4,  -4,   5,  -7,   5,   1,   1,  12,   6,  -4,  14,  13,  -6,  -1,   3,  16,  27,   0,   7,   1,  -1,  -3, -19,  -3,  19,  14,  12,   2, -21, -24, -22, -32,   3, -20,   0, -37,  -2, -75,  10,  -4,  11,  -1,   8,  -4,  27,  33, -17,  17,   6,   4,   3,   0,   0 },
      {  -1,   1,  -2,   2,  17, -18,  -1, -10,   0,   2,   0,   3,  33,  17,   3,  -3,  -4,  -3, -10, -25,  30,  30,  -7, -11,  -1,   2,   0, -36, -30,  10, -41,   8,  -8,  15,  19,  11,  -1,  -8,   8,  11, -19, -31,  12, -21,  12,  -1,  -8, -23,   1, -12,  -8,   1, -21,   1, -21,  24, -12, -14,  14, -26,  27,   0,   8,   9 },
      {   0,   1,  -2,  -2,  10,  36,  -9,  14,  -3,  -3,  -6,  -8,  -8,  16,  -5, -17,  -4,  -5,  -8,  -2, -15,  -6, -48,  15,  -5,  -7, -14,   6, -20,  -8,  -9,  -3,   5,   7, -35,   4,  -4, -20, -16,  12,  -7,  36,  -6, -28,  11, -10,  10,   0, -43,  -2,  23, -14,  -9, -12, -16,  13,  13,  -3, -25, -19,   3, -12,  32,   6 },
      {   1,   0,   3,  -7,  12,  29, -22,  18,  -2,  -4,  -3, -10,  -6,  11,   2,  -8,  -1,  -3,   1, -15, -25, -28, -34,  10,   2,  10,   9,  15,  -2, -13,  29,   7, -12,  -8,  18, -18,  21,  -7,   6, -20,   5, -36,  12,   2, -29,   7, -16,  -2,  23, -12, -31,  11, -10, -11, -15,  21, -14, -25,  30, -14, -20,   4,  19,   7 },
      {  -2,  -6,  -4,   1,  -5, -33, -27,  10,   0,  -5,  -4,   7,   0, -17,  40, -21,   0,  -6,  -1,   6,  21,  -6,   1,   5,  -4, -14,  -6, -13,  16,  24,  -3, -20,  -2, -28, -10,  -7, -29,  -2, -32,   2,  13,  -6, -40,  10,  -8,  -4, -12,   8,  10,  10, -29, -26,  -6, -16, -11,  12,   0, -36, -19,   9,   2,  -3,  27,  12 },
      {   1,   2,   3,  -2,  16,  16, -40,  17,   3,   5,   4,   3,   6,  62,  43,   3,  -1,  -3,  -1,  -5,  -9, -11,   3, -17,   1,   3,   3,   8,   2,  30, -21,  25,  -3,  -3,  -8,   2, -10,  -3,  28, -14,   2,  11,  -1,   2,   7,  13,  -5,   7,   2,   7,   8,  -5,   4,   6,  21, -10,  24, -16,  -4,  13,   7,  35, -30,  -4 },
      {   3,   0,   1,  -1,   2,  25,  28, -15,   2,  -4,  -3,  -7,  -1,   2, -30,  15,   3,  -3,   0,  -5,  -3,  22, -17,   2,   0,  -3,  -1,  -5, -29,  -5, -10,  10,  -2, -10,   4,  19,  -6,  -7, -10,  -3,  11,   0, -30,   3,  -2,   3,   0,  22,  54,  45,   6,  -6,  11,   4,  10,  -1,  37, -48,   4,  -5,  -2, -15,  -7, -11 },
      {  -1,  -3,  -3,  -3,  -4, -37, -38,  -4,   0,  -1,  -2,  -1,  -3, -27,  -2,   5,   0,   0,  -4,  -7,  -1, -30,   6,   3,   1,   5,   1,  -6,  15, -12, -16,  13,  -1,   8,   4, -22,   1,   0,   5,   9, -10,   8,  32,  -7,  -6, -14,  -3,   3,   6,  -6,  23,  11,  10,  -2,  14,  20,  74, -12,  18, -16,  -1, -28,  11, -12 },
      {   2,   3,   2,   1,   7,   9,  42,  10,   0,   2,  -2,   1,  12,  -5,  22, -15,   1,   1,   0, -10,  27,  23, -13,   5,  -3,  -3, -13, -21,  -3,  17,  17, -15,   6,   5,   1, -43,  -9,  -7, -29,  -5,   0,  17,  28,   9, -29,  32,  -5,  24, -18,  -8,   5,   2,  10,  -8,  13,  -8,  23,   0,  27, -25, -12,  30, -13,  -3 },
      {   2,   2,   2,   3,  -1,  -2,  44,  28,   0,  -1,   0,   4,  -8, -31,   6, -25,   3,   3,   8,  14,  25, -21, -27,  26,   2,   3,   4,   8,  13,  28,  18,  15,  -4,   1,   2,  13,  16, -17,  29, -17, -12, -19,   2, -19,  38,   7,   5, -11,  16, -19,   5,  -9, -12,   2,   5,   7,  24,  -9, -13,  25,  26,  11,  -4,  -9 },
      {   0,   0,  -3,   4,  10,  -1, -11, -29,   1,   1,  -2,   4,   7,  -7,   2,  35,   0,   1,  -4,   8,  20,  -2,  12, -11,  -2,  -3,  -6,   6,  21,  12,  29,   3,  -5, -12, -29, -23,  30,  -7,   2, -31,  12,   4,  -1, -52,  -6,  15, -19,   5,  14,  22,  12,  11, -10,  11,  -8,   5, -12,   2, -32, -44,  21,   3,  -1,  -1 },
      {   0,   2,   0,   0,  -3,   4,  14, -37,  -1,   1,  -1,  -3, -10, -11,  12,   2,  -1,   1,   2,  -3,   2, -36, -40,   9,   2,   2,   5,  -1,  10,  38, -22,  16,   2,   1,  -5,  -3, -34,  -7,  13,  -9,  15,  13,   4,  19,   5, -11, -15,  10, -13,  30,  -6,  11,   0,   7,  14,  12, -37,  19,  37,  -9,  18, -32,  -6, -14 },
      {   1,   2,   0,  -1,  -1, -17,  -3,  79,   0,   1,  -1,   2,  -3,   2, -10, -37,   1,   2,   0,   0,   2,   4,  28,  -8,   0,  -1,   0,   1,  -9, -17,  -6,  11,   3,  -1,  -4,   7,  13,  -8,  -1,  -2,  12,   6, -10,  -7,  -2,  10, -10,  15,   0,  42,   7,  17,   4,   5,  21,  12, -24,  16,  22, -20,  27, -11,  13, -22 },
      {   2,   2,  -1,   0,  -3,  16,  49,   5,   1,  -1,  -4,  -2,  -5,  11,  31, -27,  -1,  -3,  -7,  -7,  -9, -16,   8, -23,   1,  -1,   0,  -6,  10, -14, -37,  11,  -4,  -5,  -7, -13, -10,  23,  -5, -15,   0, -11,   0, -13, -25, -27, -10, -10,  26,  10,  -7,  14, -10,  21, -22,  13,  24,  41, -31,   1, -22,  -5,   6,   7 },
      {  -1,   0,   1,   0,  -3,  -6,  21,  21,  -2,  -2,  -1,   0,   0, -13,  -5,   3,  -1,   0,   0,   1,   7,  -9,   0, -32,   0,   1,   1,  -2,  21,  -1, -20,  39,  -2,   1,  -4, -11,   2,  30,   1, -19,   2,  12,   5, -14,  -4, -12,  18,  -3, -20, -17,  26, -14,   7,   5,  -7, -21, -38, -68,   7,  -7, -29, -13,  -4,   6 },
      {  -2,  -2,  -2,  -5,  -3, -12, -28,  22,  -2,  -3,  -1,  -4,   2,  -4, -64, -19,   1,   1,   3,   1,   8,  22, -13,   5,  -4,  -5,  -7,  -7, -12,  28,  -3,  -1,  -2,  -2,  -5, -13, -30, -12,  12, -47,   3,   1,  13,   5,  -7, -16, -27,   9,  -1,  -8,  -5,  13, -10,  13,  -8, -13,   6,  12, -23,  10, -20, -12, -28,  28 },
      {  -1,  -1,  -1,  -1,  -2,  -7,   1,  -4,  -2,  -3,  -3,  -2, -12, -16,  -7,  -7,  -2,  -3,  -2,  -1,  -5, -30, -24, -49,   0,   0,   5,   4,   2, -23, -23,  10,   3,   0,   7,  11,  23, -15, -39,   8,  -2,  -2,  -1,   8,   8,  35, -18,  15,  -4,  -7, -17,   7,  -5, -31,  -1, -17,   6,   5, -15,  -2,  31,  -4, -45,  36 },
      {   0,   0,   0,   0,  -1,   4,  22,  -2,  -1,   0,   0,  -3,  -5,  14,  35,  -3,  -2,  -2,  -2,  -5, -16,  -5,  31, -12,  -3,  -3,  -5,  -6, -20, -14,  20, -37,  -2,   1,   3,  -5,  -8, -49,  15, -10,  -1,  -3,  13,  -4,  13, -12, -29,  -8,  -7, -14,   5,   3, -19,  -5,  27, -18,  -6, -27, -10,   3,  -6, -59, -12,  -3 },
      {   0,   1,  -1,   0,   0,  -3,   7,  21,   0,   0,  -1,  -1,   2,   0,  -1,  29,  -1,   0,  -1,   1,  -2,  13,  -4, -54,   0,  -1,  -1,  -1,  -3,  13,  49,  35,  -3,  -3,  -9,  -1, -27, -19,  -7,  -6,  -3,  -2,  -9,  13,   7, -26,  23, -37,   1,   1, -17, -23,  10, -30,  -1,  24,  16,  25,  12, -18,  -2,  -9,  -7,  -3 },
      {   1,   1,   2,   2,   2,   1,   7,  23,   1,   1,   3,   0,   3,   1,  -1,  42,   0,   0,   2,   1,  -1,   6, -17, -35,   2,   1,   5,  -2,   2,  15,  10,   8,   3,   4,  11,   3, -11,   3,  -3,  23,   2,   1,  22,  23,   6,  22,  -5,  37,  -8, -15, -18,  35, -15,  41,   2,   5,   3,  -8, -37,  16,   5, -10,  49, -17 },
      {   0,   0,   0,   0,  -2,  -1, -10, -24,   0,   0,  -1,  -3,  -2,  -3,  -5, -27,   1,   1,   1,  -2,  -3,   3,  -2, -35,   1,   1,   1,  -6,  -3,  -9,  20,  -5,   1,   0,   1,  -9,  -8, -19, -30,  -4,  13,  14,  15, -10,   3,   0,  -4, -17,   8,  11,  45,  18,  -9,  17, -10,  47, -15, -15,  15,  63,   5,  28,  -2,  -8 },
      {   0,   0,   1,   2,   1,  11,  -4, -33,  -1,  -1,   1,   2,  -1,   9,   0, -42,   2,   2,   3,   2,   2,  11,  28,  26,   0,   0,   2,  -1,  -8,  -1,  16,  59,  -1,   1,   3,   1, -12,   4, -16, -12,  -1,   0,  12,  19, -19,   6,  25,  -6,  -8, -14, -14,  37,   4, -26,  13,   3,  -5, -17, -40,  -6,  27, -13,   4, -16 },
      {   0,   0,   0,   1,  -2,  -5,   6,  -2,   0,   1,   0,   1,  -2,  -2,  16,  21,   0,   0,  -1,  -1,  -6,  -7,  28,  15,  -2,  -2,  -4,  -5, -13, -25, -10,  11,   1,   5,   4,   6,  -3, -27, -12, -57,   4,   3,  11,  12,  35, -17,  16,  59,  -7,  -9,  -8, -11,  14,   1, -50,  12,   1,   4,  14,   1,   9,  16,   6,  -2 },
      {   0,   0,   1,   0,  -4,  -4,   6,  29,   0,   0,   2,   0,  -4,  -4,   6,  37,  -2,  -2,  -1,   0,  -6,  -2, -19,   7,  -1,  -2,  -1,   1,  -6,   0, -22, -52,   0,   0,   2,   0,  -6,  16,   7, -20,   0,   0,   7,   6, -30, -16,  27, -13,   1,  -2,   8,  36,  15, -37,  -5,  30, -11, -16, -26,   3,  29,   5, -32, -26 },
      {   0,  -1,  -1,   0,   1,   1,  -3, -20,  -1,  -2,  -1,   1,  -2,  -3, -11, -49,  -1,  -1,  -1,   3,   0,  -5, -11, -41,   1,   2,   3,   6,   5,   1,  11, -32,   1,   0,   5,   7,   4,  -6,  29,   9,   0,  -2,   3,  19,  17, -29,  -7,   7,   0,  -6, -11,   4,  57,  17, -11,  -7,   0, -16, -18, -42,   5,  27,   5, -23 },
      {   0,   1,   1,   0,  -1,   1,  14,  10,   0,   1,   1,   0,  -3,  -4,   7,  13,   0,   1,   1,   0,   1,   1,   9,  13,   0,   1,   1,  -1,  -3,   3,  21,  25,  -2,  -1,  -2,  -7, -11, -14,  27,  54,  -1,   3,   1,  -8, -27, -45, -32,  25,   4,   9,  19,  23,  15, -23, -29, -17,  -5,  -9,   5,  15,  27,  17,  -1,  43 },
      {   1,   0,   0,   1,   0,   3,  10,   3,   1,   1,   0,   1,   0,  -2,   7,  13,   1,   1,   1,   1,   0,  -2,   3,  12,   0,   1,   1,   2,  -3,  -6,  -6,  -5,   2,   1,   3,   5,   0,  -5,  -6, -21,   5,   4,   7,  17,  14,   1,   7, -24,   3,   3,   6,  12,  39,  23,  52,  29,  -5,  -7,  -8, -13,   6,  12,  26,  84 },
      {   0,   0,   0,   0,  -1,  -5,   0,  18,   0,   0,   0,   0,  -1,  -5,   5,  24,  -1,   0,   0,  -1,  -3,  -8,  -3,  38,  -1,  -1,   0,   0,  -2, -14,  -7,  25,  -3,  -2,  -2,   0,  -6, -21, -43,  18,  -1,   1,   3,  13,  15,  -4, -51, -47,  -2,  -4,  -4,   3,  26,  33, -14,  -3, -10,  -9, -18, -14, -16,  17, -28, -34 },
      {   0,   0,   0,  -1,  -1,  -1,  -2,  -2,   0,   0,   0,   0,  -3,  -2,  -2,   9,  -1,  -1,  -2,  -1,  -4,  -3, -10, -12,  -1,   0,  -1,  -1,  -6,  -2, -10,  -7,  -2,   0,  -1,  -1,  -6,  -3, -16, -40,  -1,   1,   1,   1, -10,  -6, -10, -42,   4,   4,   5,  13,   5, -10,  -4, -77,  13,   4,  21,  15,  37,  16,  55, -12 },
    },
    {
      {  96, -50,  -1,  -5,   0,  -1,   0,   0, -61,  11,  16,   0,   3,   0,   1,   0,   3,  19,  -5,  -4,  -1,  -1,   0,   0,  -3,  -4,  -5,   2,   1,   1,   0,   0,   0,   3,   1,   1,   0,   0,   0,   0,  -2,  -1,  -2,   0,   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0 },
      {  54,   9, -21,   0,  -5,   0,  -2,   0,  59, -60,  -2,   6,   0,   1,   0,   1, -57,  -2,  35,  -1,   2,   0,   1,   0,  -8,  30,  -3, -12,   0,  -1,   0,  -1,  -5,   1, -10,   3,   2,   0,   0,   0,  -1,   2,   1,   3,  -1,   0,   0,   0,  -2,   0,   0,  -1,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   0,   0 },
      { -37,   4,   4,   2,   0,   1,   0,   0, -38, -28,  35,   2,   4,   0,   1,   0, -27,  68,  15, -22,  -1,  -5,   0,  -2,  28,  20, -52,  -8,   4,   0,   0,   0,   6, -21, -11,  23,   3,   1,  -1,   0,   9,  -3,   9,   3,  -8,  -1,  -1,   0,   3,  -4,   1,  -5,  -1,   2,   1,   0,   3,   0,   0,  -1,   1,  -1,   0,   0 },
      {  27,  95, -38, -11,  -7,  -3,  -3,  -1, -46,  -3, -23,  11,   4,   3,   2,   1,   0, -10,   5,  16,  -3,   2,  -1,   0,  30, -20,   1,   2,  -6,   1,  -1,   0,   1, -18,  13,   0,   0,   2,   0,   0,   1,  -1,   7,  -5,   1,   0,   0,   0,  -1,  -1,   1,  -1,   2,  -1,   0,   0,   1,   0,   1,   0,   0,  -1,   0,  -1 },
      { -22,  -4,   7,   0,   1,   1,   0,   0, -54,  -1,  -1,  10,   2,   2,   0,   1, -12, -36,  44,  14,  -3,   1,  -1,   1,  -2,  49,  34, -36,  -8,  -4,  -1,  -2,  14,  23, -44, -23,  11,   0,   2,   0,   2,  -7, -16,  20,   8,  -1,   0,  -1,   7,   0,   2,   5,  -8,  -2,   0,   0,   3,  -2,  -1,  -3,  -2,   2,   1,   0 },
      { -12,  -9, -28,  11,  -1,   4,  -1,   2, -10, -65,  36,  22,  -2,   5,   0,   2,  56,   4,  28, -15, -11,  -2,  -3,  -1,  -4, -29,  16, -11,   2,   3,   1,   1, -14,  23,  28, -20,   3,  -2,   0,  -1,  -6,  18, -20, -19,   9,  -2,   1,   0,  -6,   6, -10,   9,   7,  -3,   1,   0,   0,   4,  -3,   4,  -4,  -1,   1,  -1 },
      { -15,  -8,  -3,   5,  -1,   1,   0,   0, -10,  -1,  17,  -5,   4,  -1,   1,  -1, -71,  21, -24,  19,   6,   1,   1,   1,  20, -14,  34,  30, -18,   1,  -4,   0,  16,  19,  29, -43, -19,   2,  -1,   0,   1,   0, -29, -24,  20,   7,   3,   0,   0,   2,  -6,  13,  11,  -5,  -1,  -1,   4,   1,  -2,   4,  -6,  -2,   0,   0 },
      { -10, -53, -17,  15,   4,   3,   2,   2,   7,   5, -37,  29,  -2,   3,   0,   1,  -3,  -3,  27,  34, -15,   0,  -3,   1,  59, -10,  11, -16, -21,   2,  -2,   1,  -5, -42,  13,  -1,   4,   7,   1,   1,  -4,   9,  34, -12,  -2,  -1,  -2,  -1,  -4,   6,  -5, -20,   5,   1,   1,   0,  -3,   3,  -1,   2,   9,  -2,  -1,   0 },
      {   5,   2, -51,  14,   2,   5,   0,   2,  30,  40,  23,   5,  -8,  -4,  -2,  -2,   3,  29,  -8, -18, -12,   0,   1,   0,  28, -19,   7, -21,   3,   5,   1,   1,  -7,  -5, -31, -24,  26,   1,   2,   0, -10, -14, -20,  45,  23,  -9,  -1,  -1,   0,   1,  17,  22, -23, -10,  -1,  -1,   0,  -6,   2,  -7, -10,   9,   3,   1 },
      { -18, -15, -80,  16,   5,   6,   1,   3, -11,  29,  -9,  28,  -4,   0,  -1,   1, -19,   0,  10,  -8, -11,  -1,   0,   0, -54,   5,  -6,  11,   8,   6,   2,   1,  24,  20,  12,  23, -13,  -3,  -5,  -2,  15, -10,   7, -14, -21,   4,  -1,   2,   5,  -7,   5, -13,   6,   9,   2,   2,   2,  -1,   3,  -2,   7,  -3,  -2,  -1 },
      {  -9,  -9, -18,  14,  -3,   5,   0,   2, -16, -34,  31, -17,   5,   0,   1,   0, -17, -29, -36,  13,  19,   4,   3,   2,  13, -21,  27,  20,   8,  -7,  -2,  -1, -23,  11,  -8,  18,   8, -10,   3,  -2,   7,   6,  28,  39, -29,  -8,   0,  -3,   4,   9,  15, -36, -34,  11,   2,   1,   1,  -2,  -8, -16,  17,  13,  -3,   1 },
      {   5,   7,  14,  15,  -7,   2,  -2,   1,   1,  35,  48,  -4, -24,  -1,  -6,  -1, -32, -55,  23, -45,  -7,  10,   1,   3,  25, -17,  -7, -18,  30,   7,   1,   2,   8,  -5,  16,   9,  11, -11,  -5,  -3,  -8,  17,   6, -23,  -5,  -8,   1,   2,  -5,   4, -20,  -8,  13,   1,   5,   2,  -1,   2,  -5,  11,   5,  -3,   2,  -2 },
      {  -4,  10,  31,   4,  -2,  -3,  -1,  -1,  -9,  10, -24,   7,  -2,   0,   1,   1, -34,  31,  13,  13, -24,   0,  -1,  -1, -42, -45,   9, -33,  -4,  18,   4,   2, -36,  15,  13,   6,  38,   7,   0,  -2,  20,  33,  -2,  12,  -6, -21,  -7,  -3,   2,  -7, -15,  -4, -15,  -1,   5,   3,   6,  -1,   1,   1,   2,   8,   1,   0 },
      { -16, -25, -13, -36,  17,   1,   3,   1, -17, -15, -41,  -1,  12,   5,   2,   2, -21, -26,  11, -22,  18,   4,   2,   2,   3, -12, -19,  23,  33,  -3,   1,   0,  -4, -14,  19,  16,   8, -16,   0,   1, -15,  22, -14,  18,  16, -12,   6,  -1,   4,  11,   7,  44, -22, -16,  -2,  -5,  -2,   2,   4,  -6, -35,  10,   7,   0 },
      {  -4,   3,   5, -21,   1,  -1,   1,   0,  -2,  -2,  44,  18,   5,  -3,  -5,  -1,  -6, -19,  -1,  11,  -6,  -9,   0,  -1, -49,  -2, -11,   7, -28,  12,   6,   2,  22, -48,  16, -18,  10,  30,  -3,   1, -10,  -7,  35,   9,  39,  -9, -10,  -2,   4,  19,  10,  -7, -14, -25,   0,   0,  -6,   3, -12, -11,  -2,   8,   8,   1 },
      {   4,   5,  18,  88, -20,  -1,  -5,   0, -16, -22, -33,  12, -36,   8,  -2,   2,  -2,   2, -21, -28,   5,  22,   2,   0, -16,  10,   7,   9,  12,   1,  -6,  -1,   8, -26,   3, -10,   1,   1,   0,   2,  -1,  -7,   9,   5,  19,  -1,   1,   1,   3,   1,  13,  10,  -4, -11,  -2,   0,   0,   0,   5,  -6,  -9,   2,   2,   0 },
      {   6,  12,   7,  21,   9,  -7,  -1,  -3,   5,  29,  14,  36,  -7,  -7,  -6,  -1,   6,   0,  25,   0,  16, -14,   3,  -3,   4,  30, -18,  36, -23, -25,   3,  -4, -36,   0,  17, -12, -28,   2,  11,   0,  -5,  44, -17,   7,  -7,   4,   6,  -3,  10, -13, -12,  -4, -38,   9,  -2,   2,   1, -13,   1, -14,   5,  29,  -2,   2 },
      {  -7,   0, -38,  18,  -1,   8,  -1,   0, -11,  13,   2, -61,  15,   2,   2,  -1,   1,   8,   1,   0,  31,  -3,   3,  -1, -22,  24,   9,  -8,  -5, -15,  -1,  -3, -46, -25,  -9,  -6,   6,   4,   8,   3,  -4,  26,  17, -10,  18,   2,  -1,  -2,  -4,  20, -38,   1,  21, -11,   0,  -3,   0,   1,  -5,  29,  -6,  -9,   0,  -2 },
      {  -5, -10,  -8,  13, -12,  -1,   2,  -1,  -8, -10,   6,  25,   6,  -8,  -3,  -2, -14, -45, -43,  15, -23, -12,   5,  -2,   5,  -3, -50, -14, -19,  13,  15,   1, -47,  11, -18,   9,  -2,  22,   7,  -4,   5,  -2, -15,  -1,   8,  27,  -6,  -3,   7,  -7,   2,  23,  23,   1, -11,   0,   4,   0,  15,   8, -11, -15,  -5,   3 },
      {   3,   2,  14,  39,  -7,  -1,  -3,  -1,   4,   9,  15, -14,  20,   2,   0,  -1,  -3,  -3,  52,  27,  27, -19, -10,  -4,  -7, -56,  -8,   6,  -9, -30,   6,   3,  14,   8, -17,  24, -18,   3,  18,   1,   3, -22,   8,   5,  -3,  17,   4,  -6,  -3,   7,  17,  20,   6,  -4,  -3,  -7,   3,   6,   4,  -6, -31,  -7,   5,   0 },
      {  -3,  11,  16, -11,   7,  -2,   3,   1,  -2,  11,  14,  45,  10,  -7,   2,  -1, -10,  20,  -3,   0, -19,   5,   0,   2, -11,  11,  42,  -1,  36,  -5,  -5,  -1, -52, -10,  -3,  15, -22, -25,   1,   2, -32, -25,  21,  -8,  -4,  19,  13,   0,   1,  34,  14,   8,  14,   1,  -5,  -9,   1,   2, -13,  -4, -17, -10,   1,   1 },
      {  -5,  -7,   0, -29,  18,  -8,   1,   1,  -8,  -2,  -9, -18, -37,   5,   2,   2,  -8,  -4,  32, -39,   0,  23,  -2,   3,  -4, -27,   6,  20, -10,  -6,  -5,  -1, -42,  -9, -13, -17, -23,  22,   7,   4,  24, -19,   3,   7,  20,  26,  -3,   1,   0, -21,  30, -13,  14,  -3,  -6,   5,   3,   2,   5,  -7,  35, -13,  -4,   2 },
      {  -7,  -5,  -1, -10, -32,   0,   1,  -2,  -7,  -5, -15,  13,   6,  11,  -2,   1, -12,   1,  -7, -15,  21, -11,  -1,  -4, -22, -27, -13, -18,  -8, -16,   5,   3,  13, -25, -12, -24,   4, -11,  18,   3, -62,  -1, -24,  10, -15,  15,  20,  -5,  -7,   6,  -6, -11,   5,  40,  -2,  -4,  -1,   6,  -1,   9,  38,   6, -23,  -1 },
      {   4,   9,   4,  28,  86, -16,   5,  -2,  -5, -13,  -9, -22,   9, -34,   6,  -1,  -4,  -4,   0, -14, -39, -14,  14,   0,  -3,  -9,   2,  12, -16,  16,   8,  -4,  18,  -2, -17,  -2,  -4,   4,   2,  -2, -33,  14,  -1,   6, -13,   8,   1,   2,  -2,  15,  -1,  -3,   3,  11,  -4,   1,  -1,   2,  -6,  10,   2,  -2,  -3,  -1 },
      {   1,   2,  -1, -11, -21,   6,   1,   0,   1,   2, -27,   6,  -6,   2,  -1,  -3,  -2,  10,  -9, -32,  -5,  -5,  -5,   1,  13,  -6,  -9, -10,  -5, -15,  12,   0,  17,  51, -15,  -2, -18,  24,  17,  -6,  -5,  28,  27,  -2,  13,  24, -17,  -9,   2,  48,  -8, -23,  -6, -23, -21,  -2,  -7,  -8, -37, -12,  -7,   6,  11,  10 },
      {  -3,   1,  -5, -11,  13,  -8,   3,  -1,  -2,  -3,  11,  18, -62,   9,   8,   1,   0,   8,  -1,  49,  10,  39,  -5,  -1,  -6,  -9, -23,  22,   8, -10, -17,  -1,   7,  22, -33,   4,   5,  -1,  -3,   3, -37,  18,  14,   6,   6, -12,  -4,   6, -20,   0, -15,   5,   9, -10,  10,   6,   6,  -2,   1,  30,  -2,  -8,   7,  -3 },
      {   4,   6,   9,  13,  37,   1,   0,  -3,   4,   8,  -5,  41,  31,  13,  -5,  -2,   0,   0,  -4, -14,  52, -11,  -5,  -2,   9,   5,  -9,   7,   4,  -7,   0,   4,  -4,  24,  16, -12,  33,   5,  -8,   1,   4,  -5,   2,  22,  10, -17, -10,  -5,   9,   3,  12, -18,  23, -10,   5,   2,  -1,  15, -11,  18,  21, -53,  -1,   0 },
      {  -5,   0,  -3,   9, -29, -16,  -2,   2,  -5,  10,  -7, -35,  -8,   6,   7,   3,   0,   3,  26,   6,  -3,  -4,  10,  -3,   7,  20, -14,  15, -18,  32,   6,  -4, -25,  33,  36,   7,  15,  16, -24,  -4, -59, -25,  -1,   2,  -7, -13, -15,   5,  -4,  14,  31,  -2,  -1,   4,   6,   6,  -2,  11,  -5, -19,   1,  -4,   1,  -1 },
      {  -2,  -2,  -1,   8,  20, -17,  -3,  -2,  -5,  -5, -13,   4,  20, -23,  -2,   0,  -6,  -7, -10,  -6,  13,   1,  -9,   3,   1,  -7, -23, -29,   7, -23,  -4,   1, -18,  24,   4, -12, -17,  -1,  -5,   3,  -6, -51,  17, -23,  22, -22,   6,  18, -33, -10, -16,  -6, -18, -14,  29,   8,  -3,  -3,  -3,   6,  14,  51,   5, -11 },
      {   0,   2,   7,   4, -10,  34,  -2,   8,  -1,   0,   2,  11,  38,   7,   3,  -3,   0,   5,   6, -11, -10,   4,   4,   1,   5,   2,  31,  27,  14,   2,  16,  -1,   7,   4, -12,  35,   1,  33,   2,  -8, -29,   6,   9,   1,  22,   1, -32,  -4, -30, -33, -13,  13,  -2, -10, -15,  14,  10, -12,  34,  15,  30,  12, -12,   5 },
      {   3,   5,   6,  10,  15,  79,  -1,   8,  -2,   0,  -5, -19,   3,   8, -23,  -2,   0,   0,  11,  32, -17,  -9,  -6,   4,  -4,  -2, -36,   4,  42,  11,  -3,  -1,  -7,   8,  -1, -41,  -4, -19,  -4,  -2,  -6,   0,   6,  -8,  10,   6,   6,   0,  12,   3,   9, -11,  -2,  -3, -11,   2,  -4,  -3,   1, -23,  18,   2,  -9,   4 },
      {   0,   3,  -2,  -6,  28,  12,   8,   3,  -1,   0,   9, -10, -29, -11,   8,  -2,  -2,   5, -15,  10,  24,   7,   3,   0,  -1,  15,   7, -60,   5, -17,   9,   5,  10,   8,  56,   3, -12,  11,  27,   1, -17,   6,  21,  19,  -6,  24, -13, -10,  -5, -24,  18,  13,  -3,   6, -23,   0,  13,   8,  24,  -7,  -8,  -2,  -1,   3 },
      {  -1,   2,   3, -10,   7,  43,  -6,  13,  -7,  -3,  -6,   5, -45,   1, -24,   4,  -4,   0,  -9, -13,  -1, -53,  -4,   8,   0,   8,  14,  12, -19, -19,  30,   6,  -9,  -6,  -5,   8,   9,   7,   2,   0,  11, -10,  -8,  -5, -13, -22,  -2,  -2, -52,   8,   7,   3,  -1,  16,  21,  -6, -14,  20,  -4,   8, -20,  -6,  12,  -4 },
      {  -1,   1,  -2,   2, -16, -23,   4,  -3,   0,   9,   3,   3,  22, -25,  11,   0,   2,   1,   1,   5,  -1,   8,  15,  -3,  -3,   8, -10,   2,  16,   5,  -3, -11,   6,   4, -11, -24,   2,  -9,   0,   2,  22,  45,  15,  -6,   7,   1,  10,   0, -74,   7,  41,   4,   6,   6,  -3,  -4,   4,  35,  11, -21,   5,  -8, -21,  -3 },
      {  -1,  -3,   0,  -4, -21, -10,  11,   1,  -2,  -4,  -7,   8, -25, -65,   6,   0,   5,   5,  17,  16,  25, -27,  46,   3,   4,  -6,   3,  13,  40,  21,  31, -12,   3,  -1,   4,  -8,   1,   8,  -1, -13,  15, -22,   2,  17,   9,   6,  -1,  -8,  14,   3, -28,  -1,   0,   1, -10,  -5,  -2,   0,  -7,  20,  11,   1, -11,   0 },
      {   4,   3,   3,   5,  12,   4,   2,   7,   4,   4,   3,  -3,  -9,  22,  17,  -2,   3,  -2,   3,   8, -13,  16, -13,   6,   5,  10,  -6,   5,  -3,  -6,   9,   4,   4,   9,  18,  21,   5,  20,  11,   9,  27, -17,  -7,   7,  21, -10,  25,  -5,   1,  61, -12,  24, -12,  25,  -4, -21,  32,  11,   5,   9,  30,  12, -47,  -7 },
      {  -1,  -2,  -2,  -6, -17, -10,  -9,   0,  -1,  -4,  -4,  -9,  -7, -22,  -8,   3,  -1,   2,  -3, -12, -19, -35, -20,  11,   2,   1,   5,   3,  -7, -12,  -1,  13,   6,  18,   1,  -3,  -6, -10,  -2,   2, -16,  15,  43,   5,  21, -13,  45,  -3,  20, -21,   0,  11,  -8,  -7,  36,  -7,  27, -29,  24,  -7,   4, -44, -22,   0 },
      {   1,   1,   2,  -2,  -1,  39,  -4,  10,   0,   0,  -2,   9,   3, -38,  13,  -4,   0,  -1,  -2, -10,  29,  27,   8,  -7,  -2, -14,   1, -23, -19,  11, -17, -11,  -1,  -9, -15,   7, -42,  18, -38,   4,  -3,   9, -13,  -7,  -7, -21, -21,   6,  10,  28,  -7,  11, -18,  16,  29,  10,  12,   2,  14, -22,   1, -30,   1,   9 },
      {  -1,   0,   2,   4, -13,  22,  -5,   1,  -3,   0,  -1,  -5,   7, -33,   5,  -5,  -1,  -5,   6,   3, -27,  22,   1,  -1,  -3,   6,  -8,   4, -21, -10, -39,   3,   2,   4,  29,  26,   0, -20,  15,  14,   1, -14, -22,  39,  15,  16,  19, -21, -31,  -5, -23, -10,  -8, -17,   0,  -9, -41,   2, -17,   5,  13, -27,  10,  10 },
      {  -2,  -1,  -6,   4,   6, -33, -12,  -7,  -2,  -2,   3,  -9,  -6, -14,  -4,   0,  -2,   3,  -6,  22,  -8,  -1, -31,   7,  -2,   4,   3,  -6,  39, -35,  17,   8,   0, -13,  -2,  11,  11,  25,   8, -11,  -4,   6, -32, -28,  10, -10, -21, -18,   8,  16,   9,  -2, -11,   5,   8,   7, -49, -15,  14, -30,  25, -25,   3,   9 },
      {  -1,   1,   1,   5,  11, -11,  16,   5,  -1,   1,   3,   3,  -2,  37,  20,  10,  -3,  -1,  -1,  -1,  22, -20, -23,   3,   0,  -3,   2,  -9,  11,  56,  -8,  -7,  -2,   4,  -4,  -9, -28,  22, -12,   6,   4,  -2,  16,  12,  -8,   0,  22, -11, -29,  -8, -23,  15,  -8,  14,  -8, -20, -50, -33,   1,  -8,  -4,  -9,  -6,  13 },
      {  -1,  -2,   1,  -2,  -5,   8,   4,  -5,   0,  -1,  -2,   0,  13,  -3,  41,  -3,  -1,  -6,   5,  -9, -22,  15,  11, -12,  -1,   2, -12,  20,  11, -30,  -1,   7,  -3,   3,   6, -41,  26,  12,   9,  22,   6, -19,  25,  -6, -16,  -4, -31, -23,  -6, -10, -10,  12,  -7,  50,   8, -17,  12, -31, -11,   1, -21, -14,  19,   9 },
      {   0,   2,   3,  -5,  -3,  18, -14,   0,   0,   1,   2,   6,   3, -23,  -9,   3,   0,   6,   3,  -8,  13,  23,  12,  -2,   2,  -1,   4,   3, -24,   1,   4,  -5,   1,  19,  -1,   3,  21,  -4,   5,  12,   4,  13,  25, -11,   4,  -2,  12,  17,  27,  12,  34,  16,  18,  21,  -1, -20, -67, -23,  30,  25,   3,  27,  -5, -21 },
      {   1,   2,   2,   4, -16,  10,  54,   6,   0,   1,   0,  -4,  -1,   3,  60, -18,  -2,  -2,  -1,  -2, -22, -29, -17,  -4,   0,   1,   0,  -1,  11, -21, -14,  13,  -1,  -3,   3,  -9, -21,   6,  -6,  -5,   3,  12,   0,  13, -12, -10,  -9,  22,  22,  -1,  20,   8,  -2, -19,  15,  14, -26,  29,   1,  33,  12,   5,  -2, -16 },
      {   0,   1,   0,  -1,   9,  -4, -36,   6,   2,   3,   3,   2,   0,  44,  25,  16,  -1,   2,  -6,   2, -14,  -5,  68,   0,   1,  -5,  -1,  -7,   4, -18,  12, -11,   2,   8,  -8,   3, -15,  -2,   1,   4,  -5,   0,  10,   1,  18,  -2,   6,  -7,  20, -14, -23,  -4,  -6,  -1,  24, -13, -28,  46,  13, -23,   0,  -7,  -1,  -6 },
      {  -1,  -3,   3,   1,   0,  -4, -52,   7,   1,   1,   1,  -2,  -2,  16,  16,  20,  -1,  -3,   0,   9,  -7, -25,  29,   7,  -3,  -2,  -4, -15,  10, -18, -12,  -8,   4, -10,  16,  -1, -10,   0, -30,   1,  10,   9, -14,  16,  -6,  19,  -9,  25, -20,  20,  29,   3,  -1, -12,   1,  14,  11, -59, -11,  33,  14,  -3,  -1,  -2 },
      {   0,  -1,  -1,   0,  -6,  -4, -45,  32,   1,   1,  -1,   1,   6, -20,  -9,  30,   2,   0,  -1,   3,   1, -14, -25,  10,   3,   5,   3,   6,  26,  19, -44,   7,   1,   1,  -3,  -3,  -3,  44,  26,  26,  -5,  11,   1,   6, -13,   2,   0,   2,  16,  -9,   4,  13,   9,  10,  10, -16,  12,  36, -16,   8,  11,  16,  29, -11 },
      {   0,  -1,   0,   3,  -7,  -5, -15,   0,   1,   0,   0,  -6,  17,   7, -28, -10,   2,   3,   5,   6, -13,  41, -14,   7,   0,   4,  -4,  -2,   5,  10,  60,  23,   3,  -4,  10, -14, -40,  -4,   0, -24,   6,  10,   9,  40,  -7,  -1,  -7,   2,  -2,  -7,   0,  21,   9,  15,  38,   5,  -9,   5, -32,  18,   5,  -5,   9, -12 },
      {   2,   0,  -1,   7,   7,  17,  32,  10,   1,   0,   0,   0,  -2,  -2,  23,   7,   1,  -1,  -5,  10,  14,   3,  19,   1,   1,   0,   4, -10,  -1,   4,  11,   3,   1,  -4,  -7,  33,   9,   7,   6,   5,  -6,  18, -19, -22,  32,  -1,  19,  -5,   2, -29,  28,   1,  16,   8,  25, -23,  12, -27, -65, -22,  13,   8,   7, -16 },
      {  -1,  -2,  -1,  -3,   2,  23, -41, -48,   0,   1,   2,   3,  -2,  -8,  31,  -4,   0,  -2,  -1,  -3,   5,  13,  -6,   0,   0,   1,   1,   1,  -6,   4,   8, -21,  -4,  -4,  -4,  -7,   7,  21,  29, -28,  -4,  -9,   4, -21, -38, -20,  28,   6,   1, -17,   5,  21, -14, -31, -15,  11,  -5,   6, -34,   1,  -1, -12, -29,   3 },
      {   1,   1,   3,   0,  -2,  -5,  -2,  83,   1,   1,   0,   1,   6,  -3,  26,  21,  -1,  -2,  -1,  -1,   0,  17, -10, -20,  -1,   1,  -7,   1,  -9,   9,  39, -20,  -2,  -3,  -2, -10,  -3, -21,  36,  -2,  -1,  -7,  -7, -12, -12, -22,  13,  -3,   0,   5,   8, -17, -10, -28, -11,   1,   5, -12,  15,  11,   1, -14,  14,   8 },
      {   0,   1,   0,   1,  -2, -11,  21,  46,  -1,  -2,   0,  -4,   7,   4, -39, -20,  -1,   1,  -1,   1,  -8,  13,  10,  10,  -2,  -4,   0,  -4,   0, -36,  -8, -33,   1,   2,   2,  -3,  16,  29, -40, -13,  -3,  -2,  12, -10, -14,  13,  22,  24,   3, -19,  -5,  24, -29,   7, -26, -11,  -9,  13, -27,   8,   2, -20,  -6,   4 },
      {   1,   2,   1,   0,   1,   5,  25,  -7,   1,   2,   0,   0,  10,  12, -16,   8,   3,   2,   3,  -1,   7,  14,  26,  28,   0,   3,  -2,  -1,   6,  19, -12,  15,   0,   1,  -1,  -6,  10,  23,  20, -14,  -1,  -3,  -8, -26, -14,  37,  15, -34,  -2,  12,  21, -15, -46, -15,  49,   5,   3,  -1,  19,  38, -12,  -7,   3,   0 },
      {   0,   0,  -1,   2,  -5, -10,  26, -17,  -1,  -1,   1,  -3,   0,  11, -35,  -4,   0,   0,  -1,   2,  -6,  -8,  38,  13,  -1,  -2,  -2,  -6,  -7,   1,  -8,  -3,  -2,   1,  -3,  -2, -23, -12,  29,  36,   2,  -3,   7,  -6, -26, -45, -17, -15,   0,  11,   7,  41,   1, -26, -15, -11,   5, -11,  -3,   0,  47, -11,  23,   2 },
      {  -1,   0,  -2,   2,   1,  -3, -30,  -1,  -1,   0,   0,  -1,  -3,  -4,  23, -37,   0,   0,   1,   3,  11,  -2,  -5,  15,   0,  -1,   0,  -4,  -5,  15,  25,  57,  -3,   1, -12,   4,  -1, -20, -32,  36,  -3,   4,   2, -27,  11,  10,   8,  -7,   2,  -9,  11,   1, -40,   0, -35, -14,   2,  19,  -6,  22,  14,  -6,  22,  17 },
      {   0,   0,  -1,  -3,   6,  -5, -11,  12,  -1,  -1,  -1,   0, -11,   6, -13, -39,   0,   0,  -2,   1,   6, -13,   2, -91,   3,   2,   7,   2,   4,  13,  -4,   1,   2,   1,   0,   3,  16,  -4,   6,  16,   6,   0,  10,  -9,  -8,  33,  -1,  -9,  -1,   5,  -3,  29,  -7, -12,  33,  23,  -4,  -1,  -9,  -7,  20,   5,  -7,   1 },
      {   1,   1,   1,   0,   6,  -2,  -6,  39,   0,   1,   0,   5,  -4,  -5,  17,   1,   1,   0,   0,  -4,  12,  -2,   9,  40,  -2,   1,  -2,   6, -20,  -4, -22,  43,  -1,   0,  -5,  -7,  21, -20,  -9, -49,   5,  -3,  16,  -3, -19,  11,  -8, -30,  -1,   7, -13,  38,  26, -10,   3,  21,  -6,  -4,  -5, -31,  18,  20,  -8,   6 },
      {   0,   0,   0,   1,   0,  -4,  -1,  -5,   0,   0,   1,  -3,  -5,  -2,  -8,  53,  -1,  -1,  -2,  -1,  -2, -10, -21, -18,   0,  -2,  -1,  -6, -12,   4, -13, -22,  -1,   3,  -4,   6, -11, -40, -11, -24,   2,   0,  12, -16,  23,   0, -44, -38,   9, -12,  15,  17, -30,  16,  -7, -11,  -5,  18, -24,  33,   7,  -2, -27,   5 },
      {  -1,  -1,   0,  -2,   0,  -5,  -4,  23,  -1,  -1,  -1,   0,   0,  -4, -18,  -1,  -1,  -1,   2,  -1,  -1,   8,  16, -13,   0,  -2,   0,  -2,   6,  -3, -12,  53,  -3,  -3,  -4,  -5, -14,  14,   2,  23,  -6,  -3, -11,  -1, -13, -21, -24,  -6,  -3, -11,   0, -20,   0, -21, -13, -15, -11, -11, -13,  -1, -34,  -5, -74, -34 },
      {   1,   1,   1,   2,  -1,   5,  20, -10,   1,   2,   2,   2,   1,  -6,   4,  63,   0,   1,   0,   0,   5,   7, -11,  13,   1,   0,   2,   6,  -6,   0,  29,  12,   1,   0,   0,  -2,  23, -11,  -2,  46,   3,  -2,   5,  -3, -15,  41, -13,  48,  -3,   2, -24,  19, -17, -17,   9,  17,  -7,   6,  -5, -23,  11, -15, -11, -12 },
      {   0,   0,   0,  -1,   1,   5,  -3, -11,   0,  -1,   0,   1,  -2,  -7,   7,   5,   0,  -1,  -2,  -2,  -2,   6, -21,  -6,   0,  -1,  -1,   0,  -1, -15,  21, -23,   0,  -4,   1,  -9,   5,   7, -50,   6,  -5,   3,  -7,   8, -32,  17,   1, -46,  -1,   0,  -3,  -3,  15, -46,   9, -60,  11,   1,  18,  -7,  13,  17,   7, -32 },
      {   0,   1,  -1,   1,   0,   6,  15, -14,   0,   1,   1,  -2,   4,  -5, -10,  59,   0,   0,  -1,   1,  -6,  10,  16, -50,   1,   1,   1,  -1,   5, -19,   4,  47,  -1,   2,   0,  -1,  -6,  26, -28, -13,   2,   2,   1,   5,  -6, -19,  46, -14,   1,   4,   5,   1,  12,   1, -22,  17,   0,   4,   4,   7,   3,   9,  19,  32 },
      {   0,   0,   0,   1,   1,  -1,   1,   8,   0,   0,   1,   0,  -1,  -4,  -2,  -5,   1,   0,   0,   2,  -1,   3,  -5,  22,   1,   0,   1,  -1,   1,  -8,  13, -20,   1,   0,   2,   0,   1,   3, -10,  45,  -2,   4,  -2,   6,  -7,  -3,  -3, -15,  10,  -7,  13,  -2,  26, -13,  29,   9,  -7,   7, -11,  10, -12,  28, -37,  89 },
      {   0,   0,  -1,   0,  -2,   0,  -1,   5,   1,   0,   1,   0,   2,   0,   3,   0,   0,   0,  -2,   0,  -2,   1,  -5,  17,   0,  -1,   0,  -2,   1,  -3,   9, -26,   1,   3,  -1,   1,  -4,   4, -16,  41,   1,   1,   4,   1,   5,  -5,  22, -56,   5,   0,   2,   2,   4,  13, -17,  79,  -5,   0,  -9,   3, -11,   2,   1, -50 },
    },
  },
  { // 7
    {
      { 115, -39, -11,  -5,  -3,  -2,  -1,   0, -29,  -5,   6,   1,   1,   0,   1,   0, -19,   2,   3,   1,   1,   1,   1,   0,  -6,   2,   3,   1,   1,   0,   0,   0,  -5,   3,   3,  -1,   0,   0,   0,   0,  -1,   3,   0,  -1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
      { -36, -77,  20,   9,   6,   2,   2,   2, -65,  19,  32,   7,   2,   1,   1,   0,  25,  41,   7,  -9,  -5,  -3,  -2,  -1,  13,   8, -12,  -8,  -3,  -1,  -1,   0,   4,   0,  -6,  -1,   2,  -1,   0,   0,   2,   0,  -4,  -1,  -2,   0,   1,   0,   2,   0,  -2,   1,   2,   1,   0,  -1,   1,   0,  -1,   1,   0,  -1,  -1,  -1 },
      {   5,  82,  -8,   4,  -8,  -5,  -4,   2, -94,   2,  -1,   1,   0,   5,   2,   0,   1,   7,  -4,  -5,   3,   2,   1,   0,   9,   2,  -5,   4,   0,   1,  -2,   1,   7,  -8,  -4,   3,   0,  -4,   0,   1,   6,  -5,  -1,   3,  -4,   2,   3,  -1,   0,  -5,   3,   1,   0,   2,   0,  -2,   1,   0,   1,   1,   0,  -2,  -1,   0 },
      {  15,   5,  22, -22,  -2,  -5,  -2,   0,  10,  98, -11, -22,  -4,  -2,  -1,  -3,  19,   1, -44, -13,   4,   1,   1,   0, -12, -29, -24,  11,   9,   5,   0,   1,   0, -12,  -2,  14,   3,   0,  -1,   0,   0,  -6,   3,   4,  -2,   1,   1,   0,  -1,  -5,   3,   1,   2,   1,   0,   0,  -1,  -3,   2,   2,   0,  -1,   0,   0 },
      { -14, -14,  61,  -7,   5,  -5,   2,   2, -21,  18, -26, -12,  -4,   1,   3,  -1, -68, -43,   5,   8,  14,   8,   4,   3,  -5,  35,  27,   9,   8,   0,  -6,  -1,  27,  24,  -7,  -7,  -2,  -3,  -5,  -1,   4,   2,  -6,  -8,  -4,   0,   2,   1,   3,  -3,  -6,   0,   1,   2,   0,  -2,   0,  -1,  -2,   1,   0,   0,  -1,  -1 },
      { -17,  -7, -81,  23, -15,   3,  -3,   6,   1,  52,  13,  19,  -4,  10,  -5,  -9, -47,  -2,  25, -20,   5,   1,   1,   1, -29,   8,  13,  -7,  -6,   3,  -2,   2,   6,  10,  -6,   2,  10, -10,   2,  -1,   0,  -2,  -6,   0, -10,   1,   8,  -2,   1,  -2,   6,  -1,   2,   5,  -2,  -2,   2,   1,   0,  -1,   0,   0,  -1,   0 },
      {  -8, -20, -43, -42,  16,   3,   6,   2, -18,  10, -75,   7,  20,   5,   6,   4,  34,   0,   8,  52,   7,  -1,  -4,  -3,   0,   7,  25,  12,  -6,  -6,  -6,  -2,  10,  -2,  -3,  -7,  -8,  -1,  -2,   2,  -1, -11,   2,   3,   2,  -1,   1,   1,   4,  -1,   6,   5,  -3,  -3,   0,  -2,   0,  -2,   2,  -1,  -3,  -1,  -1,   0 },
      {  21,  24,  24,  -7,   5,  -7,  -2,  -3,  22,  28,  -2,  12,  -8,  -5,  -5,  -2,  17,  24,  46,   0, -20, -11,   0,   0,   8,  43,   2, -44, -21,  -4,   3,  -3,   7,  -5, -54, -26,   7,  12,   6,   2,  -4, -23, -16,   9,  17,   9,   2,   1,   1,  -7,   5,   6,   5,  -2,  -1,   2,  -1,  -4,   3,   0,  -1,   0,   1,   0 },
      {  -5,  -6, -22, -32,   7,   1,   2,  -3,  12,  -4,   9, -18,   5,   0,   2,   1, -53,  19,   1,   8,  16,   3,   4,   1,  72,   6, -43,   9,   1,  -3,  -5,  -3,  19, -39, -12,  13,  -8,   7,  -5,   0,   8, -23,  12,   6,   0,   3,  -6,   3,   1, -10,  12,  -5,   2,  -3,   3,  -1,   0,   1,   6,  -1,   0,  -2,   1,  -2 },
      {  -1,  -7, -19, -15,  -1,   8,  -3,  -7, -10,  11,  55,   6,  10,  -6,  -1,   4,  27, -89, -14,  16, -10,   2,  -2,  -1,  25,   1,   4, -16,   2,   0,  10,  -3,  15,   5,  -2, -13,  -1,   9,  -4,  -4,   0, -19, -16,   2,  19,  -2,  -9,   2,   4,  -1,   8,  13,  -1, -11,   0,   4,   0,   0,   6,  -1,  -7,   0,   1,   1 },
      {   7,   0,  10,  57,  15,  10,  -1, -12,   1,  24, -18,  42, -13,  -6, -13,  -2,  -8,  -5, -16,  19, -21,  -9,   4,   5,  40, -13,  20,   3, -45,  -6,   8,   0,   1, -16,  31,  -5, -24,  13,  13,   3,   8,  -8,  19, -22,   1,  12,  -5,   3,  -1,   0,   0, -18,  16,   2,   1,   1,  -5,   2,   1,  -4,   1,   2,   3,  -2 },
      {  -2,  19,  13, -48,   9,  -8,  -1,   0,   3,  16,  45,  23,  13, -18,   5,   0, -21,  25,  36,  43, -10, -10, -20,  -5, -21, -36,   2,   3,  -1,  -5,  -7,   5,  11,  11,  35, -13, -23,   3,   3,   3,   9,   9,  -4, -22,  -1,   0,   0,   1,   4,   5, -19,  -2,  -2,   5,   3,  -2,   2,  -5,  -9,   1,   0,   4,   1,   2 },
      {  -8, -10,  16,   2,  11,  -5,   2,   9, -15,  -5, -11,  25,   9,  -1,   4,   6, -19, -37,  21,  16, -12,  -6,  -9,  -5, -39, -22, -31, -28,  -6,  -1,   1,   9, -24, -57, -19,  19,  14,   5,   7,   7,  -7,   5,  37,  28,  -5,  -4,   0,  -4,   4,  15,  13,  -7, -13,  -1,  -2,  -3,  -1,   8,  -3,  -4,   4,  -1,   0,  -1 },
      {   3,   7,  -1,   5, -24,  -6,   3,   7,   6,  22,  36, -31,   5,  -4,   1,  10,  17,  -2,   6,  35,  22,  -4,   0,  -6,  16,  47,  36,  12,  -5, -22,  -4,  -2, -36, -15,  -6,  15,  -9, -11,   4,   3,  -8,  23,  35,   9, -28,  -6,  -2,  -1,  -1,  17, -11, -24,  -8,   9,  -1,   1,   0,   4, -12,  -2,   9,  -1,   0,   1 },
      { -10,  -3,   0,  -2,   9,  16,  -4,   2,  -6, -11,  19, -30,  10,  -8,  -3,   1, -31,  20, -44,  12,   2,  -5,   2,  -3, -27, -16,  33,  11, -33, -10,   4,   4, -36, -13, -24, -34, -18,  -3,  13,   8,   5, -25, -22,  29,  21,  18,   5,   1,   5,   7,  18,  20,   5,  -3,  -4,  -5,  -2,   3,   4,  -5,  -6,  -2,  -1,  -2 },
      {   7,   5,   1,  20,  30,  -5,   1,  -9,   8,  -9,  21,  23,  11,  -9,   0,  -6,   9,  10, -28,  20,   9, -12,   6,   3, -35,  20, -17,  25,  -1, -15,   6,  -6,  49,  28, -27,  26, -12,  -2,   0,  -5,  21, -15,  15,  22, -16,   5,  -4,  -1,   2, -14,  22, -23, -25,  -4,  -1,   3,   4,  -9,   2, -24,   1,   9,   3,   2 },
      {  -5,   7, -17,  -5,  27,  17,  16, -11,  -8,   8, -10, -18,  11, -10, -12,   4, -20,   8, -26,  21, -24, -17,  16,   4,  -2,  29, -30, -36,  -5,   4,  23,  -6, -36,  37,  13,   9,  21,  24,  -2,  -5,  21,  30,   4,   2,  10, -13, -20,   4,   6,   0, -32,   4,  -4, -13,   0,   5,   6,  -5,   1,   4,  -2,   4,   0,  -2 },
      {  -1,   2,   2, -28,  11,  -9,   0,  10,   6,  -1,  -1,  62,  13,  -9,   5,  -5,  -9,  -3, -38, -13, -27,   3, -10, -10,  21,  34,  -8,   7,   8,   1,  -5,  10, -34,   9,   0,  13, -10, -37,  10,  13, -28,   1, -23,   8, -22,   8,  24,  -6,  -8,  -2,  -2,  15,  -1,  20,  -1,  -7,   1,   0,  -9,   4,   6,   0,  -5,   2 },
      {   9,   9,  -2,  50,  69,  18,   2,  17,   5,  13,   0, -23,  35, -15,   3,   3,   2,  -3,  22,  -2,   9, -17, -21,  -4,  10,  -1,  -1,   8,  34, -16, -15,   5,   0, -19,   5,  -5,  -2, -20, -11,   1,  -9,  -1, -21,  -5,   0, -13,  12,   3,  -3,  -1, -13,  25, -13,  -6,   6,  -4,   2,  -1,   1,   9,  -7,  -5,  -4,  -2 },
      {   3,   2,  19,  21, -51,   1,  -7,   2,   8,   8,  -7,  34,  50,  12,  -4,  -2, -11,  19, -11,  18,  31, -13, -12,  -4,  28, -19,  10, -14,   6,   2,  -4,  -1,   3,  17,   1,   1,  29,   5,  -7,  -6,  -4,  10,  -8,  28,   8, -20,  -7,   2,  17,  13,  11,  23, -26, -14,   2,   4,   7,   0,   7,  -7, -21,   0,   5,   3 },
      {   1,   6,   2, -28,  41,  12,   5,  -8,  -3,   7,   7,   9, -11, -13,  -3,  -9,  -1,  11,  11, -43, -11,   9,  -1,   5,  17,   0,  57,   0,  21,  24,   0,  -4,  -6,   2,  20,  13,  12,  11,   0,  -5,   9,  -1,  29,  31,  10,  -5,  -8,  -2,  15,  24,  29,  -8, -16, -17,  -4,   1,   5,  -2,  -7, -25,  -5,   4,   5,   4 },
      {  -1,   4,  10, -19,   5,   9,  -2, -16,   1,   3,   2,  18,  37,  28,  -7,  -5,   8,  -6,  14, -39,  48,  -4,  12,   3, -25,  39, -16,   7, -24, -29,  11,  -7, -16, -22,  42,   4, -19,  11,   5,   1,   4,   1,  -6, -16,  17,   7, -12,   3,  -4,  -5,  10,  10,   6, -12,  -4,   4,  -6,   3,  10, -11,  -6,   2,   1,   0 },
      {  -1,  -3,  -7, -11,  -8, -20,  -6,   2,   0, -10,  -5, -30,  59, -23,   4,   4,  -1,  -2, -15, -37, -26, -39,  -8,   1,  -2,   1,  22, -16, -32,  -1,   9,   8,  46,  -7,   4,  19,  13,   8,  16,   3, -16,  -3,  14, -11,  -9,   5,  11,  -7,   4,  11, -15,  -8,  11,  17,   6,  -3,   5,   4,  -4,   8,   3,   0,   2,  -2 },
      {   1,  -5,   0,   2, -29,  10, -13,   7,   6,  -3,  -4,   6,  13,   0,   9,   3,   0,   7, -18,  -4, -35,  19, -24,  -7, -14,  24,  20, -15,  27,  -3,  -8,   5,   6, -44,  11,   0,  -6,  -6,  -9,  -4,  77,  -6,  -6,  -4,  -3,  -8, -11,  -3,  -9, -29, -13,   0,  -7,  -7,   6,   2, -10, -13,  -5,   3,  -1,  -4,   5,   5 },
      {  -2,  -3,   2,  13, -41,  11,   6,   0,  -3,   4, -11, -20, -19, -44,   8,   8,  -2,   3,  23,   8, -35, -12,   5,  -7, -12,  27, -19,  24,   6,  11,  22,   2,   4,  -6,  39,  17, -25,   4,  -1,   6, -26, -24, -13,  15,   1,  -2,  -7,   0,  10,   7,  19,  28, -22,  -5,   0,   2,   2,   4,  10, -18, -19,  10,  -4,   0 },
      {  -2,   3,  -9,   5,   1,  -8,  -9,   0,   1,  -1,  10,  25, -13, -14,   7,  -1,  -6,  19, -38,  14,   8,  11,   8,   8, -23,  29,   7,   1,  33,   4, -12,  -6,  27, -33,  -5, -11,   4,  26,  -7,  -3, -42,   5,  24, -25,  39,  -3,  -6,   4,   3,  31, -20,  18,  13, -17,  -3,   1,  -1,   8,  -4,  17,   0, -10,   0,   1 },
      {  -2,  -1,  -6,   2,   5, -10, -19,  10,  -1,   4,  -8, -13, -13,   1,  26,  21,  -6,   5, -11,  -4,  10,  21,  -1,  -6,   9,  -8,  -1, -56,  -4, -29,   8,  11,  23,  13,  16,  -5, -34, -14,  16,   2, -17,  25,   6,  18,  12,   0,  -1,  -8, -59,   4,  10,   5, -22, -14,  -3,  -1,  -9,  -8, -11, -22,  -3,   4,   0,   6 },
      {  -2,  -4,  -9,  -3,   9, -59,  18,   6,   0,   1,   2,  10, -21, -14,   1,  12,   1,  -6,  -6, -11,   8, -21,  12,   7,   6,   0,   8,  27, -20,   6,  11,  -2,   5, -33,   2, -32,  12,  27, -14,  -4,   7,  47, -28,  16,  -7, -12,   4,  -2,  -1, -25, -10,   9, -36,   2,   4,  -6,  -8,  -9,  -1, -11,  -1,   6,  -4,  -4 },
      {  -3,  -4,   3,  -8, -11,   4,  -5,  -3,  -4,  -2,  -2,  16,   1, -41, -32,   0,   0,  -2,  15,   3,  12, -32,  32,  18,   3, -16,  20,   3,  32, -12,  23,   0,  -7,  -1, -18,  28,  -7,  -6,   1, -12,  15,  17,   8,   8,   6,  26,   5,  -3, -46, -22,   2,  16,  25,  -5,   1,   0,   4,   4,  21,  19, -11, -27,  -2,  -2 },
      {   2,   6,  11, -11,   5,  45,  17,  19,   1,  -1,  14,   2,  10,  25, -21,  13,  -4,  12,  -6,  13, -13,   3,  19,   1, -12,   1,  16,  -5, -20,  32,  11, -11,  18, -18,  12,  -3,  16,  -9, -38,  -6, -41,  -5,  -5,   4, -25, -24,  -9,   5, -41, -30,   5, -10,  -3,   1,   0,   8, -14,   2,   8,   1,   8,  -5,   4,  -4 },
      {  -4,  -2,  -3,  10,   0, -33,  -4,  -1,  -4,  -1,  -7,   4,  32, -39,  13,  11,  -3,   4,  -1,  -4,  22,  24,  27,   8,  -2, -12,   8, -30,  21,  18,  11,  -8, -34,   2,  -6,  -8, -17,  14, -22, -17, -12, -43, -10, -29, -26,   6, -14,   7,   5,  -6,   0, -25,  -7,   2,   5,   8,  -7, -12,   3, -19,  16,  14,   1,  -2 },
      {   1,   0,  -4,  -6, -26,  55,  36,  10,  -4,   2, -11,  10,  15, -18, -11,   2,  -1,  -8,  -2, -29,  -2, -16,   1,   3,   5,  -8, -15,   9,  16, -12,   2,   1,   6,   1, -29, -45, -36,  -2, -10,   0,  -7,  27,  18,  -5,  13,   7,   4,  -1,  13,   5,  -8, -23, -14,   3,   5,   0,   1, -19, -26, -16,  17,  13,   3,   4 },
      {  -1,   2,   0,  14,  -1, -17, -12,   6,   1,  10,   3,  -5,  47,  25,  28,  -2,  -3,  -2,  17,   1, -43,  15,   5,  -8,   6,   2,  -8,  47,   6,  20,  -7,  -1, -15,  19, -15,  -9,  -4,  34,  -8,   2,  -1,  17,  12,   2,  24,  26, -17, -10, -43,   3,  14,  -4,   0,  -4, -14,  -2,  -5,   8,   3,  -1,  10,  -6,  -7,   5 },
      {  -7,  -4,  -4,  -8,  -5,  -7, -29, -34, -10,  -4,  -7,  -6, -14,   0, -35, -10, -10,  -3,  -2,  -6, -12, -24, -53,   6,  -1,  -2,   2,  18,   3, -24, -12,   1, -21,   1, -17,   7,  12,   6,  -2, -13, -27, -14,  -8, -18,   5, -16, -11,  10, -32,  -3,  -6, -22, -18, -23,  13,  21, -20, -12,  -3, -12,  -8,  20,  18,  12 },
      {  -1,   1,  -1,   6,  -4, -35,  56, -31,   0,   3,   0,  -1,  10, -11, -11,  13,  -2,   3,  -4,  -3,  -1,   7, -22,   5,  -5,   7,  -9,   0,   9,  -8,   5,   3,  -2,  12,  16, -40,   7, -24,  -5,   6,   8, -22,  30,  11,  10, -30,  -3,   0, -23, -10,  25,   0,  18,   6,  16,   3,  15,  41, -12,  13,   1, -13,   7,  -3 },
      {   0,   1,   2,  -3,  -3,  -5,  35, -33,   0,   2,   3,   1,   6,  17,  25,  -5,  -5,   3,   5,  -7, -38,  29,  28,  17,   2, -21,  17,   4, -13, -44,   4, -20,  -1, -11, -29,  27, -10, -25, -10,  -8,  -8,   9,  -5, -26,  -2, -19,  -4,   3,  12,   6, -11,  17, -21, -18,  -1,   3,  24,   0,  18,   0, -18,  -1,  -1,  -1 },
      {   0,   1,   1,  -4,  -3,   3,  30,  46,  -1,  -1,   4,   9, -24,  26,  34,  24,  -4,   0,  -9, -11,   1, -43, -14,   1,   0,  -5,  15,  -8,  13, -32, -21,  -7, -16,  18,   3,  16,  -2,  43,  -2,  -4,  -5, -34,  17,  -8,   2,   7, -10,   7,  -3, -27,   2,  12, -10,  13,  11,  -2,  15,   3,   8,   6,   2,   1,  -5,  -6 },
      {   0,   0,  -2,  -6,   4, -27, -35,  32,   2,   0,  -2,  -3,  -2,  31, -27,   8,   0,   0,  -4,  -8, -27, -28,  15,  11,   0,   3,  -9,  -5,   8,  -4,   4,  -2,   0,  10,  -4, -22, -40, -15, -33,  -7,  11,   6,  17,  -4, -25, -17,  -4,  27,  21,  33,  21,   7,  11, -16,  15,  -3, -15,  11,  12,  -1, -17, -15,   7, -12 },
      {   0,   0,   1,  -4,  -2,   9,  -7,   8,   0,  -2,   0,  -9,  -6, -20,  11, -28,   2,  -2, -12,  11,   1, -14,  -6,  -3,  -2,  11,  -8, -16, -14,  10, -28, -14,  -7,  -9,   6, -13,  21,  -6, -20, -19,  11,  23,   0, -47,  -5,  16,  28,  -4, -18,   4,  49,  -5,  -8,   1,   5,   0,  54,  12,  13, -27, -11,  -2,  -4,   4 },
      {   1,   0,  -2,  -2,   5,   1,   1,  21,  -1,  -2,  -3,   5,  -3, -31,   4, -28,   2,  -4,   1, -18,  20,  18, -31, -30,  -2,  -1,   0,   2, -42,  28,  -6,  -3,  -3,  11, -24,  15, -44,  12,  -9,  -8,   2,  12,  13,   2,  -7, -30, -30,   4,  -2, -11,  -7,  20,   1, -11,   3,  24,  16,   7,   5,  35,  -6, -13,   6,  10 },
      {  -2,  -2,   0,  -1,  -2,  -5, -39,   2,  -2,  -1,   0,  -4,   6, -19,  -3,  -7,   0,  -2,   4,  -5, -15,  10,  31,  -8,  -1,  -2,  -4,   2,  -8, -27, -39,  -8,  -7,  11,  23, -22,  15, -15,  -7,  15, -21,  -2,  36,  19,  16,  -2,   9,   0,  31, -62,  -4,  -7,  -9, -11,  -4,  13,   5, -15,   9,  12,   4, -15,   5,  10 },
      {  -1,  -1,  -2,   9,   1, -14,  20, -37,  -1,  -2,   6, -16,   7,  34, -14, -38,   0,   0,  -3,  12,  -2,  -4, -13,   6,  -4,   3,   8, -26,  19,  35,  -6,  -1,  -1,  -7,   4,   3, -38,  10,   3,  -1, -17,  12,  26,   7,  -1,  25,  23,  13,   7, -29,   2,  25, -10,  29,  16,  -4, -10, -28,  12,   2, -19,  14,   0,   5 },
      {  -1,  -1,   0,   4,   2,   5, -19, -49,   0,   2,  -2,   6,  -9,  11,  56,  23,  -5,  -2,  -1,  -1,   7, -55,   8, -23,  -2,  -1,   4,  -3,   7,  27,  -1, -29,  -5,  -7,  -4,   3, -18, -23, -14,  -1,  -1,   7, -12,  12,  13, -24,  -7, -10,   5,  -4,   4, -15,  29,   3,   5,   7,   7, -13, -12,   1,   9, -13,   2,  13 },
      {   0,   1,   1,  -3,   7,   8, -22,   2,   1,   1,   3,   0,   1,  10,   5,  27,  -4,   6,  -2,   2,   2,   2, -21,   1,   3,   1,   6,   9,  20,   7,  57,  17,  -3,  -2, -16, -22,   7,  -1,  41,   7, -11,  14,  23, -40, -18, -19, -12, -21,  22, -33,  28,  13,  -1,  -2, -26,   1,  16,  -8,  19,   3,  -5,   8,  -1,   6 },
      {   0,   2,  -1,   1,  -3,  -8,  43,   4,  -1,  -3,   2,  -4,  -1,  -7,  14, -21,  -2,   1,  -6,   2,   0, -14,  -6, -39,   1,  -2,   1, -11,   6,  -4,  -7,   2,   1,   4,   0,   1,  11,  -5,   6,  12,  -7,   9,  -5, -20, -17,  20,  -5,  -5,  14,  -9,  15, -15,  12, -46, -29,  -1, -63,   2,   2, -12, -26, -39,  -1,   0 },
      {   0,   0,  -2,  -4,   1,  26, -18, -38,  -1,   2,  -5,   0,  -3,  -8,  14, -18,   0,   2,  -4,  -5,  -5,   0,  -1,   2,   0,  -3,   3, -12,  19, -25, -11,  -4,   9,  -1,   3, -19,  -8,  34,   9,  -4,  -5,  18, -18,  17, -42,   0, -10,   8,   5,   3,  12,  -6, -19,  22,  -5,   2, -17,  60,  29,  24,  28,  -5,  -9,  -5 },
      {   0,   2,  -1,  -2,  -2,  16,  -3, -16,   1,   1,   4,   5,   2, -26,   6,  39,   2,  -1,   1,  -6,   7,   6,   7,   3,   0,   0,  -2,   5, -19,  13, -35,  -9,  -4,  14, -16,  10,  11,  -2,   3,  27,  20,   7,  24, -28,   6, -24,  15,  12, -12,   8,  22,  34,  -9,  17,  15, -16, -56, -12,  17,   2,   1,  21,  -7, -22 },
      {   0,   0,   0,   5,  -1, -19,  14,  13,  -1,  -1,  -1,  -6,   2,  -3, -23,  -4,  -4,   4,  -2,   9, -10,   2, -11, -20,   1,  -3,   9,  -9,   8, -22,  -1,  -1,   7,  -5,   4,  27,  -3,  -7, -16,   4,  -5,  27, -21,  -1,  30,  -5, -17,  18,  18,  -4,  42,   7,  30,  13, -19,  19,  -1, -16,  14,  -3,  67,  26,  -1,  -7 },
      {   0,   1,   2,  -3,   0,   4,  -3,  16,   0,   0,   2,   0,  -3,  -4,  48, -62,   2,   1,   2,   8,   4, -17,  21,  28,   4,   3,   2,   4,  -5,  -7,  25,  47,  -6,   0,   2,  -4,  16, -21, -10, -11,   3,  -4,   9,  -6,  17, -15, -13,   0,  -3,  -4,  15,   0,  -2,   8,  19,   1, -30,   2, -14,  22,   3,  30,  19,  -4 },
      {  -1,  -2,   1,   0,  -3,  -5,  -3,   9,  -1,  -2,  -1,  -1,   2,   7, -10, -40,  -1,  -2,   2,   1,  -3,  -9,  37, -54,   2,  -8,   3,   5,  18,  -6,  16, -40,   2,  -5,  -1, -12,   6,  15,  40,  23,  10, -18,   6,  -7, -14, -35,  20,  24, -27,   7,  -3,  14,  -2,  -9,   5,   0,  11,  -5, -25,  -3,  11,  21,   0,  -3 },
      {   1,   0,  -1,   2,   4,   2,   7,   2,   0,  -1,   1,   0,   5,  -2,   0,  20,   2,  -2,   1,   1,  -5,   0,  25,  10,  -1,  -2,  -2,   0,   6,   2, -44,  16,   5,  -7,  -6,  -2,   0,   7,  19,  -5,  -1,  17, -10,   8, -28,   6, -23,   8,   1, -12,   4,  19,  42,   1, -14,  32,   6,  19, -29, -15, -38,  49,  35,  22 },
      {   0,  -1,  -1,  -2,  -2,  -4,   3, -21,   0,  -1,  -1,   3, -10,   1,  -4, -10,   2,  -2,  -2,  -8,   7, -23,   5, -16,   2,  -2,   0,  -3,  -5,  -3, -12,  30,   4,  -1,   3, -22,   9, -15, -37,  15,  11, -23,  25,   3, -26,  12, -27, -11, -20,  24, -25,  28,  -4,  -2, -55,  -8,  14, -23,  25,   2,  12,  24, -25,  14 },
      {   0,  -1,   0,   0,  -6, -10,  10,  11,   0,  -1,   1,  -5,   4,   6, -19,  -6,   1,  -2,  -3,   3,   1, -22,  19,  30,   1,   2,   2, -12,   6,  27, -34,  12,  -5,  -2,  -1,  13, -33,  -4,  26,  23,   5,  -8, -19, -10,  27, -39,   8, -21,  -4,  -6,   2, -35, -26, -16, -41, -24,   7,  22,  18,  16,   5,  11,   5,  11 },
      {   0,   1,   2,   1,  -5,   2,  17,   1,   0,   0,   2,   2,   4,  -3,  23,  11,   0,  -2,   1,   0,  -1,   4,  -7,  28,  -2,   7,  -6,   5,  -3,   2,  -4,   3,   0,  -1,  14, -13,   2,  12,  20,  -3,   1,   1,  -6,  20, -27,   3,  36,  25, -10,  23,  16, -15,  13, -46,   2,  24,  11, -53,  30,  46,  14,   5,  13,   3 },
      {   0,   0,   2,   0,  -2,  -2,   2,  20,   0,   0,  -1,   0,   4,  -9,  11, -18,  -1,  -2,   1,   1,  -4,   1, -19,  29,   2,   0,  -9,   5,  -4,  -4,  -2, -50,  -2,   1,   5, -12,   0, -10,  19, -62,   1,  -5,  16,  18,   2, -26,  13, -13,  -1,  -1, -19,  11,  29,  13, -36,  -5, -28,   0,  21, -27,   2,  10, -20,  11 },
      {   0,   0,  -1,   0,  -4,   5,  -5,   1,   0,   0,   1,   4,   2,  -7,  21, -14,   3,  -1,  -1,  -3,   3,   7, -19,   7,  -1,   0,   4,   2,   5,   2,   6,  20,  -1,   0,  -8,   1,  -2,   8,  -9,  41,  -8,  11,   0,  17,   1, -19,   4,  41,  -5, -19, -29, -20,  41, -15,   8, -28,  20,  10,  39, -50, -11,  10,  -4, -33 },
      {   0,   0,   2,   2,  -7,   5,  -8, -19,   1,   0,   1,  -4,   2,   5,   1,   1,   2,  -1,   3,   4,  -1,  -7,  -3,  25,   2,  -1,   0,   0, -11,   9,  -7,  27,  -3,   5,  -4,  13,  -8,  13,  -4, -22,  -1,  -5,   6,  -6, -17,  -6,  25,  14,   6, -24,   2,  40,   2, -48, -16, -16,   5,  13, -58, -18,  48, -19, -15, -16 },
      {   0,   0,  -1,   3,  -1,   3,  -7,  -6,  -1,  -2,  -2,  -2,   3, -12,  -3, -14,   0,  -1,  -1,   0,  -3,   9,  -9,  23,  -2,   0,  -2,  -4,   0, -24,  -2, -16,  -2,   1,   3,   3,  -2,  29, -21,  17,  -8,   3, -12,   6, -13, -23, -16,  -1,  -5,   0,  32, -10,  16,  29, -35, -43,   8, -42, -44,  23, -36, -17,  -4, -16 },
      {   0,  -1,  -1,  -1,  -2,  -4,  -5,   5,   0,   0,  -1,   2,   1,  -5,   7,   8,   0,   0,   0,  -3,   3,  -8, -15,   9,   0,   0,   1,  -1,  -3,  17,   6, -41,   3,  -4,   3,  -2,   8, -40,  13,  -3,  -2,   6,   3,   0,   7,  42, -36,  57,   1, -13,   0, -10, -21, -20, -14, -25,   5,  18, -18,  33,  -7,  32, -21, -15 },
      {   0,  -1,  -1,   0,   1,   1,  -4,   7,  -1,   0,  -1,  -1,  -2,  -2,  13,   2,   0,   1,   0,   1,   5,   0, -12,  25,   3,  -3,   0,  -1,  -4,  -1,  29, -15,  -1,   3,  -3,   8,   4,   6, -28,  43,  11,  -3,   3,  -7,   8,   4,  34,  44,   0,   2,  -6,   8, -11,  20, -37,  28, -11,  20, -10,  -8,   1, -16,  20,  63 },
      {   0,   1,   1,   0,   0,   0,  -5,  -3,   0,   1,   1,   1,  -2,   0,   0,  11,  -1,   2,   0,  -1,   2,   8, -10, -27,   1,  -1,   3,  -3,  -2, -12,  21, -17,  -3,   5,   0,   9,  -7,  15, -45,  -5,   0,   1,   6,   7,   0,  11,  46, -38,   8,  -9,  -1,  -4,  14, -30,  -3, -36,  -7,  26,   0,  28,  -7,  57,  29,   2 },
      {   0,   1,   1,  -1,   2,   3,   1,  -8,   1,   0,   0,   4,   2,   1,  -3,  19,   0,   2,  -1,  -3,  -3,   6,  14, -40,  -1,   2,   2,   4,  -2,  -6,   3,  47,  -1,   0,  -2,   5,  -1,  10,  -3, -49,   1,  -1,   0,   0,  19, -20,  29,  47,   2,  -4,   7, -27,   1,  14, -14,  -3,  -1,   3,  -1,  14, -29,   7, -50,  22 },
      {   0,   0,  -1,  -2,   0,  -1,  -5,  -5,   0,   1,  -1,   2,  -1,   3,   6,   3,  -1,   0,   0,  -2,   1,  -5,  -4,  -6,   2,   0,   0,   0,   3,   5,   4,   1,  -2,   0,   0,  -1,  -6,  -3,  -8,  -5,   6,  -4,   5,   1,  13,   3,  26,  -2,  -4,   1, -10,  -6, -16,  16, -49,  62,   2,   7,  -7,   4, -18,   2,  17, -87 },
      {   0,   0,   0,  -1,   0,  -1,  -3,   7,   0,  -1,   0,   0,   2,   3,  -2,  -9,  -1,   2,  -2,   2,   1,  -1,  -5,  24,   0,  -3,   4,  -5,   3,  -3,   3, -11,   1,  -1,  -2,   4,  -4,   5, -12,  37,   1,   3,  -3,   2,  -4,  -3,   6, -31,   4,  -3,   7,  -7,  13, -19,  21,  49,  -6,   9, -16,   6, -14,  25, -89,   3 },
    },
    {
      {  83, -59,   5,  -3,   1,  -1,   0,   0, -65,  33,  10,  -3,   2,  -1,   1,   0,  10,  11, -15,   1,  -1,   0,   0,   0,  -3,  -7,   2,   4,  -1,   1,   0,   0,   2,   2,   1,  -1,  -1,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  55,  14, -33,   1,  -4,   1,  -2,   0,  20, -71,  34,   5,   2,   1,   1,   0, -46,  44,  13, -20,   5,  -3,   1,  -1,   5,   6, -25,   8,   3,  -1,   0,   0,  -5,  -1,   6,   5,  -4,   0,   0,   0,   2,   1,  -3,  -2,   0,   1,   0,   0,  -3,   0,   2,   1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 },
      {  57,  -8,   7,  -8,  -3,  -2,  -1,  -1,  46,  -4, -49,  10,  -1,   3,  -1,   1, -13, -57,  43,  13,  -4,   3,  -1,   1, -23,  30,  17, -27,   5,  -2,   1,  -1,  -5,   2, -16,   6,   7,  -2,   1,   0,  -5,   3,   5,   2,  -4,  -1,   0,   0,  -1,   0,  -2,  -1,   0,   1,   0,   0,  -2,   2,   2,   0,   0,   0,   0,   0 },
      {  10,  68, -31, -13,  -4,  -1,  -1,  -1, -67, -16,  -6,  33,   1,   5,   0,   2,  46, -20,  35, -12, -11,   1,  -2,   0,  -4,   1,  -7, -16,  12,  -1,   1,  -1,   3,  -1,  -1,   9,   0,  -3,   1,   0,   1,  -1,  -1,  -3,  -1,   1,   0,   0,   2,  -1,   1,   2,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0 },
      { -36, -47,  38,  -5,   6,   0,   3,   0, -22,  -6,   8,  13,   1,   1,   0,   1,  -6,  18,  49, -30,  -3,  -2,   0,  -1,  -4,  50, -45, -21,  18,  -5,   3,  -2,   9,  -9, -21,  29,  -3,  -2,   0,   0,   4,   0,   5,  -2,  -7,   3,  -1,   0,   3,  -2,  -3,   1,   2,   1,  -1,   0,   1,   1,   0,   0,   0,  -1,   0,   0 },
      { -17,  -8,  -4,  17,  -4,   2,  -1,   1, -35, -25,  18,  -8,  10,   1,   3,   0, -32,   7,  17,  39, -10,  -1,  -2,   0,  30,  12,  50, -47, -19,   5,  -3,   1,  -4,  23, -34, -27,  31,  -3,   1,  -1,   6,  -1, -11,  21,  -1,  -8,   1,  -1,   1,   1,  -3,  -2,  -4,   2,   1,   0,   3,  -1,  -1,   2,   0,   1,   0,   0 },
      {   9,  36,  27, -20,  -3,  -5,  -1,  -2, -20,  27, -54,   2,  10,   0,   2,   0, -56,  25,   1,  26, -10,   1,  -1,   1,  51,  -8, -24,  -2,  -9,   6,  -3,   1,   0, -35,  13,  12,   1,   3,  -1,   1,   6,   1,  12,  -8,  -3,   0,  -1,   0,  -1,  -5,  -1,   0,   3,   0,   0,   0,   2,   1,   2,  -1,  -1,  -1,   1,   0 },
      {  12,   5,   2,   5, -13,   1,  -2,   1,  28,   8,   5, -19,  10,  -2,   1,  -2,  34,  27, -16,  -2, -25,   0,  -3,   0,   9, -39, -17, -49,  28,  13,   1,   3, -28,   1, -25,  44,  34, -21,   4,  -2,   0,  -5,  14,  19, -31,  -1,   4,  -1,  -5,   0,   3,  -7,   3,   7,  -2,   0,   0,  -1,   3,   2,  -1,  -1,   0,   1 },
      {  -8, -39,  23, -19,   3,   0,   3,   0,   8, -42, -26,  42,   9,   3,   0,   1,  25,  16,  22,   5, -27,   3,  -4,   0,  14, -47,  13,  -6,   3,  10,  -3,   2, -12,  11,  48, -18,   3,  -6,  -2,   0,  -5,  22, -17, -23,   8,  -5,   4,  -1,  -5,   3,  -6,   8,   4,  -1,   2,  -1,  -1,   4,  -2,  -2,  -1,   0,   0,   0 },
      {   8,  27,  73, -27,  -7,  -6,  -2,  -2, -16, -47,  -5, -24,  25,  -1,   6,   0,  10,   9, -32,  26,   6,  -7,   3,  -1, -44,  12,   8,  12,  -4,  -1,   2,  -1,   2,  26,  -9,  12, -11,  -2,   0,  -2,   6,  -7,   0,  -4, -11,   8,  -2,   1,   1,   1,   0,  -2,   4,   2,  -1,   1,   2,  -1,   0,   0,  -2,   0,  -1,   0 },
      {  17,  17,  32, -20,   5,  -6,  -1,  -3,  18,   4,   2,  -1, -17,   5,  -1,   1,  35,  22,   1, -19,   6,  -8,  -2,  -3,   3,  12, -22, -12, -34,   8,   3,   2,   3, -28, -16, -44,  36,  25,  -5,   4, -15,   1, -22,  35,  35, -26,   0,  -1,   0,  -4,   8,  16, -21,  -3,   6,  -1,  -3,  -1,   1,  -2,   1,   3,  -1,   0 },
      {   3,  20,  20, -30,   5,   0,   3,  -2,   3,  34,   8,  16,  -8,  -9,  -2,  -2, -40,   1,  -1, -41,  -8,  13,   1,   3,  -2, -17,  11,  -2,  43,  -6,  -1,   0,  -4,  49, -13, -13,  -7, -23,   9,  -3,  -1,  -5, -49,  16,   7,  11,   6,  -2,   2,  -5,   8,  25,  -9,  -1,  -5,  -1,   3,  -4,   0,  -5,  -7,   4,   0,   2 },
      {  11,  12,   1,   4, -13,   5,  -3,   0,  23,  23,  15,   5,   5, -13,  -1,  -3,  20,  33,   8,  10, -38,   3,  -3,  -1,   2,  14,  10, -27, -10,  -4,  -3,   0,  10,   2, -22, -18, -38,  17,   6,   1,   5, -30,  -2, -47,  29,  32, -12,   4,  -7,   1, -15,   7,  37, -19,  -2,   3,  -1,  -2,  -1,  10,  -4,  -5,   5,  -1 },
      { -12, -20, -39, -28,  15,   3,   6,  -1,   4,  -7, -39,  49,   6,  -6,   2,  -1,   1,  35, -24,  18, -16,   5,   0,  -1, -25,   6,   5,  11,  -5,   1,   0,  -1,  16,   3, -29,   6,  -8,   7,   3,   0,   4, -34,  14,  36,   0,   9,  -7,   0,   7,   0,  34,  -7, -18,  -4,  -6,   1,  -1,  -3,   2, -14,   2,   4,   0,   1 },
      {   6,  -9,  13, -10,   5,   5,  -1,   0,   2, -27,  -2,  -1,  -3,  -8,   1,   1,  26, -24,  -9,  -6,  23,  12,   0,   2,  71,   1,   4,  13,  10, -38,   1,  -5, -45,  -1, -29,   5, -34,  12,  18,  -1,   5,  -6,  11,  16,  14,  21, -12,   1,  -4,   5,  14,   0,  -8, -13,  -2,   1,   0,   0,   0,  -4,  -2,   5,   2,  -1 },
      {   9,   4,  -3, -12,   8, -13,   4,  -2,  17,   7,   0,   4,  -2,   5,  -4,   1,  30,  20,   6,   6,  12, -14,   2,  -2,  36,  34,  14,  17, -10,  -4, -11,  -4,  23,  -2,   0, -15, -12, -43,   4,   2, -17,   0, -26,  -1, -51,  15,  31,  -4,   2, -14,   1, -17,  10,  40, -14,  -4,  -6,   0,  -4,   3,  18,  -9,  -6,   3 },
      {   6,  12,  33, -23,  22,  -5,   2,  -3,   5,   4,  48,  46, -36,  -9,  -3,  -1, -16, -16, -12, -12, -16,  14,  -1,   2,  -1,  -9,  22,  -8,  -8,   5,  -2,   1,   1, -15,   2,  -3,   6,   8,  -1,   1,   7,   2,  41,  -2, -15,  -2,  -7,   0,   2,  10,  -4, -52,   3,   7,   0,   4,  -2,   2, -15,   2,  25,  -4,   2,   0 },
      {   3,  12,  12,  45,  -2,  -8,  -4,  -1,  -2,  19, -22,  14, -27,  10,   2,   2,   0,  36,  33,  -4,  23, -15,  -8,  -5, -23, -20,   0,   6, -31, -23,  15,  -2, -43,  35,  -3,   0, -12,  25,   5,  -5,  19,   6,  -5,   4, -19,   7, -12,   2,   2, -15,   5, -23,  -3,  15,  -1,   5,   4,  -2,  -1,  -4,  12,   0,  -3,   1 },
      {  -7,  -4, -35, -78,  30,  -4,   5,  -2,   7,  22,  17, -18,  21,  -7,   4,  -1,   4,   6,  11,  19,  11,  -9,  -2,  -2,  -3,  13,  -7,  -4, -16,  -6,   6,  -2, -37,  24,   5,  12,   4,  17,  -6,  -2,  18,  23,  -4, -11,  -5, -17, -11,   3,  -2,  -6, -27,  -4,   0,   1,  11,   2,   2,  -1,  -2,  13,   5,   2,  -1,  -2 },
      {  -6,  -2,   0,   3,  -8,  -2,   7,  -1,  -4,   5,  27,  19,  17,  -8,  -3,  -4, -12, -27,  -7,  20, -25, -21,  15,  -3, -19,   3, -27,  -1, -26,  20,  -1,  -1, -46, -23,   4,  -7,  -7,  -1,  16,   5, -14,  29, -16,  18,  -1,  49,   8, -12,  14, -10,  20,  12,   9,   5, -36,   2,   2,   1,   7, -12,  -7, -10,   1,   8 },
      {  -8, -13, -13, -22,  12, -10,   6,  -2, -14, -22, -29, -23, -34,  27,   0,   5,  -9, -12, -15, -28,  21,   2,   3,   3,  -2, -18,  -9, -14,  -7,  20,   4,   2,   4,   3, -13, -10,  21,  12,   5,   0,   7, -16,  -9,  -3,   3,  28,  -5,  -2,  -2, -10,   3,  -5,  59,   1, -22,   5,  -1,  10,   0,  29,  10, -34,   6,   0 },
      {  -4,   4,  -8,  14,  16,   3,   1,   0, -11,  -9, -24,  13, -32,  -8,   2,   0,  -2,  25, -43,  -1,   2,  29,  -4,   3, -25,  34,  17, -20,  22, -20, -10,   2, -37, -32,  -2,  -4,  -1, -13,  16,   3,  -3,  37,  -5, -11,   7,  -2,  14,  -5,  -4,  -3, -35,  16,   0,  11,   8,  -5,   5,  -4,   9,  18,  -1,   6,  -8,   0 },
      {  -5,   0,  -3, -21,   7,  -1,  -1,  -1,  -5,   2,  -1, -36,  14,  -6,  -1,   2,  -7,  22,  39,  -6,  15,   1,   0,  -2, -26, -42,  26,  -7,   7, -32,  -6,   2,  21, -36, -12,  -8, -14,  -2,  30,  -2, -44,  25,  18,  13,   4,  23,  -4, -11,  12,  16,  -5, -10,  -6,  -5, -11,   4,  -8,  -1,  -9,  -5,   5,   2,   3,   1 },
      {   4,   5,   0,   5, -21,  14,   3,  -1,   4,   7, -10, -10,   8,   4, -13,   2,  14,  20,   0, -10,  -3,  17,  14,  -2,   8,  36,  16,  11,  20,  19, -21,   0,   4,  12,  25,  -4,  18,  -3,   6,  -6,  19,  39,   0,  19,   2,  11, -43, -12,  16, -20,   6, -30,  -1, -44, -16,  25,  -8,   2, -14,  -5,   6,  -9,  25,   0 },
      {   7,  10,  10,  16,  60, -28,   7,  -5,   8,   8,  19,  20, -15,  20,  -3,   0,   6,  12,  13,  36,  27,  -5,  -5,  -3,   1,  -1,   7,   8,   6,   1,  -6,   1,   5,   1,   3,  23,  12, -18,  -1,  -5, -10,   0,   5,   4,  -1,   1, -10,  -1,  -5,   9,  10,  27,  14, -37,  -7,   4,   1,   9,   3,  10, -51, -17,  15,  -2 },
      {   8,   2,   1,  24,  66,  -7,   0,  -3,   1, -15,  -3, -27,  -7, -30,   8,  -1,   6, -16,   2,  16, -39,  24,   0,   0,   5, -14, -31,  -3,  -1,  -5,  -6,   5,  35,   8,  -2,   0,  -9,  15,  17,  -3,  18,  12, -26,   5,  -6,   5, -14,  -3,   5, -31,  -5,  -3, -12,   9,   0,  11,   0,  -4,   7,  -5,  21,   9,  -3,  -2 },
      {   0,   4,   2,   8,  34,  -2,   3,   0,  -7,   2,  -9, -33,  10, -16,  -1,   1,   3,   9,  33, -13, -20,  16,  14,   2, -14,   4,   8,  33,   7,  26, -27,  -6, -47, -14,  -2, -26,   3, -25, -15,  18,  27, -31,  17,   2,  10,  -4,  28,   2, -15,  18,  17, -10,   3,  -3,  -2, -10,   3,   8, -12,   0,   5,  -2,   1,  -1 },
      {  -4,  -1,  -7, -16,  -9, -30,   3,  -1,  -3,  -2,   3, -38, -56,  28,  13,  -2,   3,   9,   4,  19, -26,  19, -14,   7,   6,  21,  -2, -12,  13,   7,  13,  -2, -10,  20,  29,   2, -11,  10, -11,  -4, -24,  -5,  14,  -7,   6,   8,   2,  -2,  18,  20,  18,   4, -15,   7, -19,  -7,  -5, -11,   0, -43,  -2,   8,  -7,  12 },
      {   2,   6,  -2,  23,  31, -18,   6,  -3,  -4,  -3, -25,   4,  25,  -5,  -1,  -5,  -4,   1, -18, -37, -22, -24,  14,  -1,  17,  11,  -1,   9, -11,  23,  13,  -8,  -7,  35, -20,   0,  -3,  11, -15,  -1, -57,  18,  19,  -4,   0,  -1,  -3,   3,  19,  29, -31,  -8,  -1,  -5,  -2,   0,  -5, -15, -11,   8,  -2,   4,   4,   3 },
      {  -7,  -3,  -2,  -3,  -8,  -2,  -1,  -2,  -9,  -3,   3,  -4, -11,  -6,  -5,   6,  -9,  -6,   1,  14, -13,   8,  -3,  -5,  -8,   2, -12,   7,   4, -10, -12,   8, -30,  -6,   0,  -9,   0, -14,  -5,  -8, -65, -27, -19,  -9, -17, -16, -24,  27,   1, -32,  22, -16,  13, -14,  44,  24, -17,  12,   1,  11,   8,  10,  20, -22 },
      {  -2,  -4,  -4,  -6,  11, -33,   4,  -2,  -3,  -7,  -6, -16,  -5,  -4, -14,   5,  -3,  -1, -10, -42, -12, -41,  11,   3,   1,   2,  46, -23, -21,  10,  13,   0,  -1, -34,  10,  24, -19,  -1,  -3,  -2,  27,  -4, -18,  -8, -14,   4,  -5,   8,  -4, -23,  14,   5, -10, -11,  17,   6,  -1,   6,   0, -26, -22,  29,  14,  -6 },
      {  -2,  -6,  -7,  -6, -19, -26,  13,  -7,  -1,  -5,  -7,   0, -25,  13, -24,   6,  -1,  -9,  -8,  15,  -2, -21,   5,   1,  -4, -13, -44,  -5,  -9, -10, -21,   5,   9,  17, -18, -31,  -3, -34,  23,  16,  26,   7,   9,  -5,  -3,   8,  14,  -7,  -3,  17, -24, -20, -14, -26,  15,   3,  -7, -12, -23,   4, -18,  36,  19,  -5 },
      {  -2,  -3,   0, -14, -29,  17,   6,  -3,  -3,  -5,  -1,  -5, -54, -39,   1,   6,   5,   6,  22,   6, -10,  11,  37,  -8,   5, -11,  14,  40, -21,  31,  -7,  -4,   9,   6, -32,  28,   5,   4,  -7,   1, -11,  20,   2,  -3,  -9,   2,  -4,  -4, -18,  -7, -14,  16,  -4,   4,   0,   1,   8,  -6,  19,  12, -25,   8,  -4,   1 },
      {  -1,  -4,   0,   2, -20, -58,  22,  -4,   4,  -1,  11,   8,  20,  41,  -3,  -8,  -1,  -1,   7,   4, -16,  22, -20,  12,  -2, -15,  11,  38,  17,   1,   2,  -9,  -3, -22, -46,   2,   1,   6, -24,   1,  13,  17,  -5,  -9,   8,  -8,   1,   8,  14, -30, -16,   4,  -3,  -2,  -3,   4,  -5, -13,  16,   2,  19,  -4,   1,   6 },
      {   2,   1,   4,  -2,  25,  56,  -2,   8,  -1,  -4,   2,   1, -14,  29, -22,  -2,  -1,   1,   0,  20,  27, -24,  16,   3,  -2, -15,  -2, -16,  23,  29,  -9, -23,  -7, -12, -16,  -2, -24,  -4, -34,   8,  -7, -10, -12, -15,   8,  -4,   0,   4,  15, -17, -17,  -9, -14,  -3, -17,  15,   3, -26,  10, -20,  17,  16,  13,  14 },
      {   3,   0,   0,  -5,   3,  25, -14,   1,   4,  -2,  -2,  -2,  -4, -13,  12,   2,   2,   0,  -1, -13,   4,  15, -25,   0,   2,  -8,  -1,  -2, -27,  -5, -14,   7,   2,   8, -16,   0,   2, -18, -12,  13,  15,  19,  16, -14,  -2,  -3,  20,  24,  86, -16,  12,   9,  16,  -6,  14, -18, -12,   2,  -1, -17, -25, -11,   4, -10 },
      {   4,   6,   3,  -2,  15,  49,  26,  15,   1,   2,   1, -12,  -8,  45, -28,   3,  -1,   4,   2,   5, -42, -24,  -7,   9,   2,   7,   5,  28,  13, -12,  34,   7,   4,  -2, -13,  -7,  13,   6,  18, -20,   8,  27,   5,  -2,  -2,   9,   4,  14,  -5,   9,  21,   0,  11,   8,  22, -14,   4,  18,  -5,   8,  -7,  10, -16, -18 },
      {   3,   2,   3,   6,   0, -17,  32, -15,   5,   5,   7,   8,  10,   7,   4,  17,   4,   6,   7,  10,  28,  14,  51, -11,   3,  -1,   2, -14,  19,  29,  -5,  -4,   1,   1,  16,  -4,  -4,  17,  14,  -5,   5, -11,   8,  12,   5,  20,  -5,   8,  39,   1,   0,   9,  12,  13,  28, -21, -11,   6,   8,  15,  12,  40, -35, -25 },
      {   0,   0,   5,  -4,   8,  20,   9,  -8,   0,   0,   4,  -2,  11,  35, -15,   1,  -3,  -5,   3, -24, -16,  13,  -9,  -8,   2,   4,   9,  10, -33, -14, -22,  10,  -9,  11,  20,   4,  20,  -4,  34,   7,  -9, -42,  30,  -2, -13,   1,  -3, -27,   4, -27, -29,  22, -10,   0, -16,  13,  -5, -15,  41,   0,  -7,  20,  -6,   6 },
      {  -1,  -2,   2,  -2,  -5, -12, -17,   6,  -2,  -5,   0,  -3, -16, -35, -19,  -4,   0,   3,   7,  14, -13, -43,  13,  18,   2,   1,   4,  10,  37, -15,  37,  -8,  -3,   1,  17, -21,   4,   5,   2,  -3,   9, -21,  27,  21,   5,  -8,  11,  -9,  26, -31, -23,   6,  -2,  -1,  -6,  -7, -38,  -1,  19,  22,  -6, -26,   9,   3 },
      {   0,  -4,   0,  -2,  -2,  10, -30,   7,   2,  -2,   0,   2,  -4,  13,  47,  -5,   1,  -6,   7,   2, -11,   1, -14,  -5,  -1,   0,   2,  22,   5,  14,  15,   3, -10, -32,   1,  -7,   4,   0,   8, -30,  15, -28, -35,  15, -26,   8, -26,   0,  27,  28, -36,  -6,   1,   0,  -5,   5,  -2, -22, -27,  18, -25,  20,  -2,   3 },
      {  -1,   2,   1,   3,   3,  19,  34,  -7,  -4,   0,  -1,   1,   8,   0,  17, -12,  -4,  -1,  -3,  -5,  11,  26,  -8,   0,  -4,  -3,  -5, -10, -11,   6,  18, -10,  -1,   0,  -5,  -6,  -4,  -3, -18,  -8,   1,   5,  -2,  -5,  -7,   9,  -1,  -6, -34, -16,  -4,  -9, -10,   2, -11, -22, -92,  13, -15,  -6, -27,   9,   0,  10 },
      {   2,   1,   0,   3,  -6,   4,  61, -10,   0,  -4,  -4,  -6,   6, -27, -45,  -4,   2,  -1,   0,  -2,   4,  30,   6,  13,  -3,  -3,  -7, -10, -10, -24,  22, -18,  -2, -12,  10, -16,   6,  -2, -24, -13,   7, -18, -11,  -2, -27,  -2, -21,   5,  22,  29,  -7,   6,  -4,  13,  -6,  16,  28, -11,   0,  -5, -20, -21,  10,   4 },
      {   0,   0,  -2,  -1,   6,  -9,   3,  31,   1,   1,  -2,   8,   9, -29,  17,  -1,   2,   1,  -1,   0,  29,   1,  -5,   9,   1,   3,  -4,   8,  16,  10,  30,   9,  -2,   2, -15, -29,  46,  -5,  14, -21,  -7,  -3,  23, -50,  -4,  15,  -3,  -8,  -5, -31,  13,   0,   1,   1, -15,   4,  28,  -3,  -8, -27,  -5,  37,   3,   5 },
      {   0,   0,   0,   5,   0,   0,   2, -17,  -2,  -2,  -1,   3,  -6, -27, -42,   8,   0,   2,   3,  10,  10,  -3, -28,  15,   1,   4,   8,  16, -13,   7,  13,   8,  -6,  -2,  -1,  47,  29,  -9,  23,  -1,  -3, -15, -36, -18,  31,  -2,  10, -12,  18,  24,   2,  -6,   6, -15, -13,  -5, -30,  -9, -17,  -1,  39,  16, -14,   2 },
      {  -2,  -2,  -3,  -1,   5,  -6, -56, -31,   0,   2,   0,   6,   7,  13, -24,  23,  -2,   0,  -8,   1,  -9,   8,  48, -12,  -1,  -1,  -4,   8,   9, -31, -11, -15,  -2,  -3, -17, -10,  29,  18,  -5,  -4,   9,   4,  -6, -29, -19,  -8, -14,   3,  -1,  20,   5,  15,  10,  17, -13,   3, -27,  16,  -4, -34,  -6,   0,  -2,   5 },
      {  -1,   0,  -2,   0,  -2,   7,  20, -69,  -1,  -2,  -3,  -2,  -7, -21,  12,   1,  -1,  -2,  -1,  -3,   6, -42, -48, -14,   1,   1,   1,   4,  26,  21, -26,  17,  -2,  -4, -11, -24,  -3,   4,   0, -12,   8,   6,  12, -11, -16,   0, -26,   2,  -3,  12,   4,  13,  -5,   2,   3,   7,  -1,  15,  21,  -9,   8,  -6, -21,  15 },
      {  -2,  -1,   0,  -3,   7,  -7, -41,   2,  -1,   1,   1,  -1,  10,  -3, -38,  12,   0,   1,   5,  -3,  13,  22, -22,   7,  -2,   3,  -4,  -2,  -8,  30,  21,  14,   1,   1,   5, -16, -10, -21,  -5, -11,   6,  15,  14,  17,   1,  28, -19,  21, -26,  10,   4,   2,  10,  14,  27,  -1, -14, -57,  33,  -8,  -5,  -5,  -6, -11 },
      {   1,   3,   1,   1,   1,  -3,  23,  51,   0,   2,  -1,   4,  -5,  -2,  26,  -8,   0,   1,   0,  -1,   5, -15,   8,  -9,   1,   1,   1,   5,   0,   5, -10,   7,  -2,   0,  -9,   2,   7,  -6,  17,   3,  15,   1,  -4, -13,   0,  -2, -10, -12,   5,  43,  -4,  21,  -3,  18,  17,  33, -37,   2,  34, -29,  27, -23,  33, -31 },
      {   0,   1,   0,   0,  -2,   6, -23, -20,   1,   3,   1,   1,   1,   2,   7,  -4,   2,   4,   2,   0,   5,  19, -15,  42,  -1,   0,  -4,   0, -13,  17,  36, -36,   5,   4,   0,  -7, -11, -15,   0,  14,   1,   7,   7,  15, -13,   7,   7,   5,   1,  15, -14,  16,   3, -13,  -7,  28,   8,  64,  25,   4,  21,  34,  19, -12 },
      {   0,   0,   0,   2,   3,  13,  12, -49,   2,   2,   2,  -1,   0,  10,  47,  48,   0,  -1,  -2,   0,  -9,  -3,  18, -10,  -1,   0,   1,   1,  -9, -14,  43, -25,   0,  -2,   2,   5,  10, -38,   2,   2,   2,  -1,   4, -10,  23,   7,   3, -16,  -5, -11,   4,  -5,  -9,  -9,  18,   5,  10, -32,  -2,  -7,  14, -27,  11, -14 },
      {   0,   1,   2,   2,   1,  10,   8, -28,   1,   1,   2,   3,   5,  -6,  21, -10,   1,   2,   3,   1,   1,  -9,  -5,  44,   1,   5,   6,   6,  17, -20, -18,   7,   5,   0,  12,   6,  27,  44, -12,  38,  -3,   4, -11,   6,  -7,  20,  22,   8,  -9,   9,  12, -14,  22,  13,   2,  -3,  -2, -39,  16,   3, -16,  28,  43, -16 },
      {   0,  -1,  -1,   1,  -1,   1,  -7, -23,   1,   0,   2,   1,   9,  12, -13,   5,   2,   2,   1,   3,   8,  14,   0,  38,   1,   0,   0,   0,   5,  42,  -1,  13,  -6,   2, -11,   2, -23,  17,  49,  -6,  -6,  -6,   1, -15,  10, -24,   3, -22,   7, -12,   0,  -7, -23,  45,  12,   4,   5,  10, -27,  -3, -31, -30,  31, -13 },
      {   1,   1,   1,   2,  -1,  -1,   9,  27,   1,   0,   2,   1,   4,  -9, -13,  86,  -1,  -1,  -2,   4,  -2,   4, -40, -34,   1,   2,   2,   7,  15,   9,   4,  -7,   0,   0,   1,   1,   0,  25, -10,  36,  -5,  -4,  -3,  17,  -7,  12,   3,  -8,  -2,  -4, -23,   2,   7,  13,   2,  -2,   2,  19, -14, -12, -11,   5,  18,   7 },
      {   0,   0,  -1,   1,   2,   5,  -1, -13,   0,  -2,  -1,   3,  -4,   4,  -3, -32,   2,   1,   2,   6,   4,  -3,  10,   7,   1,  -1,   2,  10,  -1,  -8,   1,  36,   0,   2,   3,  12,  19, -10, -22,  11,  -5, -13,   3,  22,  17,  46,  -8,   6,  -7, -16, -50,  11,   2,  12,  37,  -4,  11,  26, -33, -47,  11, -16,   0,   5 },
      {   0,   0,   1,   1,  -1,   2,  -7,   0,   0,  -1,   2,  -1,  -6,   2,  -7, -11,   0,   0,   1,   0,  -1,  -2, -10,  -8,   0,   0,   6,  -3,  -3,   1,  -6, -15,   2,   2,   5,   3,   6,   3, -12,  15,   6,   4,  24, -11, -13,  25, -11,   2,   2,  -6,  27,  63, -29,  20,   6,  21, -11, -19, -57,  47,  27,   6,  16,   1 },
      {   0,   0,  -1,  -1,   2,  -6,  -9,  -8,   0,   0,  -1,  -4,   3,   2,  -7, -24,   1,   0,   1,   0,  -3,  -1,  -5, -44,   3,   1,   2,   1,   4,  -9,  15,   7,   2,   2,   0,   7,   7, -25, -22,  18,   4,   2,  -5,  -9,  45,  11, -25, -25,  17,   7,  16, -34,  -6,  51,  -1,  11,   4,  22,  25,  29, -25,  25,  17,   9 },
      {   0,   0,   1,   2,  -3,   2,  16, -14,   0,   0,   1,   2,   0,  -4,   4,   2,   0,   1,   2,   3,  -2,   0,   8, -22,   0,   1,   2,   3,  -1,  -3,  10,  21,   1,   2,   4,   4, -15, -14,  17, -20,   8,   6,  18,  27,  23, -40,  10,  49,  -7,  -6,  -7,  18,  46,  27, -34,  28,  -6, -11, -21, -14,   5,  27,  28,  14 },
      {   0,   1,   1,   1,   0,  -4,   4,  17,   1,   2,   1,   1,  -1,  -2,  12,  39,   0,   0,  -1,  -3,  -4, -17,  -2,  53,  -2,  -1,  -3,  -2, -12, -12, -47, -28,   2,   6,   6,  11,  13, -17,  -8, -46,   1,  -2,   5,   2,  33,  20, -18,  27,   0,   0,  -4,  -6, -13,  29,   1,  -2,  -4,  13,   8,  12,  -5,   3,   1,  31 },
      {   1,   1,   1,   1,   3,   5,   1,   0,   2,   2,   1,   0,   4,   7,   7,  36,   2,   2,   1,   0,   2,  11,  26,  23,   1,   0,   0,  -6,  -4,  -4,  17,  90,  -3,  -1,  -5,  -4, -13,   1, -28,  -6,   3,   2,  -2,  -8, -12,   4,  14,  -9,   9,  12,  10,   5, -20, -14,  -3,  16,  -7,  10,  12,  26,  12,   3,   9,  36 },
      {   0,   0,   0,   0,  -1,   0,  -2,   1,   0,  -1,   0,   1,  -2,  -4,  -1,  12,  -1,  -1,  -1,   0,  -4,  -8,  -9,  24,  -1,  -3,  -3,  -2,  -3, -12, -16,  10,  -3,   2,  -5,  -1, -16,  -6, -45, -26,   0,   6,  10,  15,   7, -14,  -4, -68,  -1,  -9,   2,   9,  32,  11, -13,  21,   0,  -5, -20,  -8,  -5,  17, -26, -57 },
      {   0,  -1,  -1,  -1,  -1,  -1,  -7, -13,   0,  -1,  -1,  -1,   2,  -2, -10,  -2,   0,   0,  -1,   0,   3,   8, -10, -37,   1,   1,   0,  -3,   2,   2,  -8,  -1,   1,   2,   4,   5,  10,  24, -14, -59,   2,  -4,   0,  -4,   8,  32,  63,  20,  -1,  -1, -11, -13, -34,  -8, -10,  30,   1,   9,   7,   7,  -6,  -7,  13, -40 },
      {   0,   0,   0,  -1,   2,   2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0, -12,   1,   0,   1,   1,   1,   6,  -1,  -5,   1,   1,   0,   2,   2,   3,   0, -27,   1,   2,   0,   1,   0,  16,   6, -19,   3,   5,   1,   9,  -1,   1,  38, -39,   6,   3,  14,  -2,  33,  -5,  59,  46,  -7,  -8,  -2,  -9,  -8,   4,  -9,  66 },
      {   0,   0,   0,   0,   0,  -2,   3,   7,   0,   0,   0,   0,   2,   0,   9,   5,   0,   1,  -1,  -1,   1,  -5,   3,  17,   1,   1,   2,  -1,   6,  -1,   6,   3,   1,   0,   1,   1,   7,  -1,   8,  47,   1,  -1,  -1,  -7,  10,   3,   3,  35,  -2,  -3,  -1, -16, -13,   5, -12,  73, -13,  -2, -10,   1, -31, -12, -68, -11 },
    },
    {
      {  87, -46,  -2,  -4,   0,  -1,   0,   0, -71,  23,  11,   1,   2,   0,   1,   0,  17,  13, -12,  -1,  -2,   0,  -1,   0,  -3, -13,   3,   3,   0,   1,   0,   0,   1,   6,   4,  -3,   0,  -1,   0,   0,  -2,  -1,  -4,   0,   2,   0,   0,   0,   1,   1,   1,   2,  -1,   0,   0,   0,   0,  -1,   0,  -1,   0,   1,   0,   0 },
      { -67,  -8,  22,   3,   5,   0,   2,   0, -37,  65,  -9,  -5,  -3,  -2,  -1,   0,  58, -21, -29,   5,  -2,   2,  -1,   1,  -3, -21,  17,   6,   0,   1,   0,   0,   6,   7,  -1,  -6,   0,  -1,   0,   0,  -1,  -4,  -3,   4,   1,   0,   0,   0,   4,   0,   2,   1,  -3,   0,   0,   0,   0,  -2,   0,  -1,   0,   1,   0,   0 },
      {  34, -16,   3,  -1,  -1,  -1,   0,   0,  34,  22, -33,  -1,  -5,   1,  -2,   0,   8, -78,  20,  17,   3,   3,   0,   1, -35,  33,  38, -18,   0,  -4,   1,  -1,   0,  14, -29,  -3,   3,   0,   1,   0,  -4,  -4,   7,   8,  -3,   1,  -1,   0,   0,   3,  -2,  -6,   1,   1,   0,   0,  -3,   1,   0,   3,   2,  -1,   0,   0 },
      {  29,  49, -28,  -4,  -3,   0,  -2,   0,  26,  -8, -12,  -1,  -1,   1,   0,   0,  11, -15, -36,  12,   4,   3,  -1,   1,   3, -60,  37,  25,  -5,   3,  -1,   1, -13,  23,  34, -31,  -2,  -4,   0,  -1,  -7,   5, -27,   1,  11,  -1,   2,   0,  -2,   0,   7,  10,  -5,   0,  -1,  -1,   0,  -1,  -1,  -4,   0,   1,   0,   0 },
      {  18,  81,  -5, -16,  -7,  -5,  -2,  -2, -46,   7, -42,   9,   5,   4,   2,   2,  13, -22,  17,  10,   0,   0,  -1,   0,   3,  11, -27,   0,   1,  -1,   0,   0,  13, -34,   8,  26,  -7,   3,  -1,   1,   2,   1,  24, -18,  -5,   1,   0,   0,  -1,  -3,  -7,  -2,   7,   0,   0,   0,  -1,   1,   4,   2,  -3,  -1,   0,  -1 },
      { -17,  32, -21,  -1,  -2,   1,  -1,   0, -53, -37,  22,  10,   5,   3,   2,   2,   6,   1,  28,  11,  -4,  -1,  -2,  -1,  14,  14,  37, -33, -11,  -3,   0,  -1, -10,  29, -43, -26,  18,   0,   4,   0,   4,  -8, -19,  35,   4,  -1,  -1,  -1,   5,   0,  11,  -2, -12,   1,  -1,   0,   2,  -2,  -4,  -3,   4,   1,   0,   1 },
      { -20, -30,  -5,   1,  -2,   2,   0,   1, -17, -39,  -4,  34,   4,   7,   0,   2,  27, -20,  49,  -6, -15,  -2,  -3,   0, -16,  -3,  11,  -3,   3,   1,   1,   0,   5,   9,  46, -16, -19,   1,  -1,   1,  -7,  22, -16, -51,  16,   4,   2,   0,  -4,   4, -18,  21,  18,  -6,   0,   0,   2,   5,   2,   2, -10,  -2,  -1,  -1 },
      {   9,  10, -41,   2,  -2,   4,  -1,   2,  25, -11,  29,  10,  -6,   1,  -2,  -1,  72,   4,   4, -41,  -9,  -3,  -1,  -1, -39, -16, -30, -15,  23,   4,   5,   1, -13,  10, -13,  32,  11,  -3,  -1,  -1,   1,  -8,  10,  11, -15,  -2,  -2,   0,  -2,  -1,   6,  -9,  -3,   2,   0,   1,  -3,  -3,  -1,   0,   4,   2,   0,   0 },
      { -11, -17, -30,  15,   2,   2,   1,   2,  -9, -19,   8,   4,   2,   0,   2,  -1,  -5,  -5, -15,  35,  -6,   2,  -2,   1,  13, -21,  34,   2, -14,   0,  -2,   2,  -9,   3,   7,  22,  -6,   0,   0,  -1,  17,  -5,  53, -21, -37,   7,  -4,   1,  -6,   6,  -9, -57,  26,   8,   2,   1,  -1,   5, -11,  18,  16, -11,   0,  -2 },
      {   2,  35,  40,  -5,  -8,  -2,  -3,  -1, -17,  -3,  52, -26,  -4,  -2,   0,  -1, -25, -16, -20, -27,  10,   4,   4,   1, -37,  12,  15,   8,  28,  -1,   1,  -1,   7,  40,  -2,   9, -16, -13,  -1,  -1,  10,  -8, -10, -28, -12,   7,   1,   2,  -3,   5, -20,  -2,  26,  -1,   2,   0,   4,   3,  -7,  16,  -3,  -7,   0,  -2 },
      { -17,  -8, -68,  -7,   6,   5,   4,   3, -18,  19,  17,  28,  -7,  -3,  -3,   0, -37, -27,  -4, -19,   8,   5,   4,   2, -11,  26,   0,  46,  16,  -7,  -2,  -2,  27,   1,  13,   2, -26,  -7,  -4,   0,   1,  -6,   0,  11,  -2,   2,   4,   1,  14,  -1,  20,  -4, -25,   2,  -1,   0,  -2,  -5,  -1, -21,   6,   7,  -1,   1 },
      {   5,  14, -46,  20,   1,   1,  -1,   1,  14,  57,  34,  -2, -16,  -6,  -5,  -3, -15,   4,  14,  -6, -20,   1,  -3,   2,  21,  -6,   4, -37,  -9,  11,   1,   3,  19, -19, -10, -21,   9,   9,  -1,   0, -13, -15,  -1, -26,  24,   4,  -1,   1,  -8,   7, -25,  15,  29,  -9,   2,  -1,   0,   1,  -1,  24, -10,  -9,   1,   0 },
      {   2,   2,  19,   6,  -9,  -3,  -3,  -1,  -5,  25,  16,  13,  -8,   1,  -1,   0, -29, -40,  24, -29, -27,   8,   1,   2,  42, -27,  -4, -21,  26,  11,   2,   2, -51,   2,  18,  14,  12,  -7,   1,   0,  12,  37,   0,   6, -13,  -5,  -3,  -3,  -1, -12,   8, -14, -18,   3,   3,   2,   2,   5,   1, -19,   8,   4,  -1,   1 },
      {  16,  17,  40, -10, -10,  -2,   0,  -1,  17,   4,  30,  28, -22,  -1,  -3,   0,  26,   9,  12, -22,  -9,  -2,  -2,  -1,   4,  14,  16,  14, -25,  -2,  -3,   0,  26, -29,   9, -43, -17,  17,  -2,   4,  -5,   0,  11,   1,   5,   3,  -4,  -2,   7,  -5,   8, -46, -14,   9,  -1,   4,  -9,   0, -18, -11,  37,  -6,   0,   0 },
      {  -7,   9,   4,  -6,  -1,  -5,   1,  -1,  -4,  28, -16,  23,   6,  -2,  -1,  -1, -38,  36,  19,  18, -28, -12,  -3,  -2, -75, -26,   0, -21, -18,  20,   7,   4,  14,  22,  18,   3,  22,  10,  -1,  -3,  12,   6,  -4,   6, -12, -10,  -4,  -2,   6,  -4,   4, -12,  -9,   3,   0,   2,   4,  -1,  -6, -10,   8,   5,  -1,   2 },
      {   8,   9,  -8, -16,  -8,  -2,   1,  -1,  21,  36,  -5,   9,  20,  -3,   3,  -2,  18,  37,  14,  16, -18, -12,  -2,  -5,  21,  30,   7,   3,  -6, -20,   1,  -3, -31,  36,  -9,   6, -51,  -8,   7,   0,  27,  -9, -17, -28, -19,  26,   4,   1,  -6, -22,   4,   1,  -1,   2,  -1,   1,   2,  -5,   4, -17,  -5,  11,  -3,   0 },
      {   4,   0,  24,  -3, -21,  -3,  -2,   2,   9, -15,  17,  31,  25,  -2,   1,  -4,  -4, -11, -21,  10, -43,  -3,  -2,   2,   8, -21,  16,  -9,  11,  10,   5,   1,  28,  -6, -19,  18, -17,   1,   0,   2,  -4, -28,  32,  -2,  14,  11,  -7,  -1,  -2,  33,  19,  21, -13, -21,  -4,  -1,   3, -13,   2, -30, -38,  21,   6,   3 },
      {  -8, -17,  -9, -95, -10,   1,   6,   2,   4,   0, -28,   7,  15,  12,   0,   1,  -9,   5,  -6, -30, -10,   7,  -5,   3,  18, -14,   6, -11,  24,  11,  -2,   1,  15,   5, -14,  -7,  10,  -9,  -3,   1,  -6, -13,  -6,   1,   0,   7,   2,   1,  -7,   0,   5, -13,  22,   3,   1,   1,  -3,   6, -13,  23,  10, -21,   1,  -1 },
      {  -6,  -1,  -5, -57,   3,   0,   2,   0,   3,   9,  56,   1,  32,  -5,   1,   1,   1, -21,  -6,  34,  14, -27,   1,  -3, -20,   3, -13,   6, -30, -11,   8,   1, -39, -25,   3,  -1,   3,  14,  10,   2,  -9,  32,  11,   8,  11,   0,  -4,  -5,   6,   9, -11,   9,  -2,  -3,  -2,  -2,   1,   4,   5,   6, -11,  -4,   0,   0 },
      {   6,   3,   1, -16,   2,   0,   3,   0,  14,  10,  22, -20,  16,   3,   0,  -1,  16,   5,  24,  25,  17,  -6,  -4,  -4,  30,   4,  10,  -9,   9,  -1,  -6,   1,  56,  19,  21,  20,  22, -31, -11,  -5, -28,  22,  -6,  -3, -32, -31,  13,   4,   3,  12, -19,  -7, -22,   8,   9,   0,  -1, -16,  -4, -12,  -2,  23,   0,  -1 },
      {  -5,  -1, -21, -14, -28,   3,  -1,   1,  -4,  -1,  -8, -56,  -1,  17,   7,   0,   7,  21,  -6, -14, -10,  29,  -3,  -2, -17,  24,  33,  -6,  11,  10,  -6,   1, -17, -15,  12,  -8,   5,  14,  -8,  -1,  25,  21,  27, -15,  30, -13,  -5,  -1,  -7,  10, -15,  -6, -27, -27,  13,   2,   5, -15,  -4, -23,   3,  20,  -3,   4 },
      {  -1, -10,  -7, -10, -23,   3,   1,   1,   0, -18,  14, -25,   0,  -5,   1,   0,   6, -38, -25,  21, -23,  -3,   4,  -2,  15,  12, -40, -15, -23,  26,  10,   3,  30,  20,  17,   0,  12,  28,  -5,  -3,  33, -21, -21, -12,   5,  -1,  -6,  -1,  -1, -38,   8,   1,   1,   6,  -2,   3,   2,  -3,  20, -13,  30,   7,  -7,   2 },
      {   3,   6,  -3,   4,  16,  -4,   1,  -3,   1,  -8, -12,  46,  -1, -17,  -1,  -2,   5,   8, -66, -12, -34,  -4,  10,   3,   5,  55,  12, -10,  -2,  21,   5,   0,  -9,   1,   5,  -2,  22,  -6,  -4,  -2, -11,  34,  -7,  -4, -14, -16,   3,  -1,   1,  -3, -28,   2,  -3,   9,   7,   1,  -1,  -6,   5,  16, -11,   6,   0,  -1 },
      {  -9,  -6, -10,   6, -53,  -3,  -6,   1, -11,  -7,  -9, -18, -15,   6,   4,   1, -11,  -2, -16,   4, -11, -11,   2,  -1, -16,  -5, -11, -18,  -2,  -1,   3,   3, -17, -20, -15, -10, -16, -16,  10,   6, -65,   3, -18, -14, -21,  25,  21,   2,   5,  14,  -4,  -6,   5,  31, -10,  -4,  -2, -15,  -1, -21,  23,  22,  -9,   1 },
      {   2,   5,  -6,  30,  20,   3,  -1,   3,  -2,   1, -11, -36,  83,  -9,  -1,  -4,   0,  -6,   9, -29, -30, -28,   6,  -1,   2,   4,  -1,  21,  -1,  17,  11,   2,   9,   7,  -9, -11,  -5,   2,  -3,  -6, -27,  18,   5,   0,   8,   9,  -8,  -4,   3,  -2,  10, -13,   3,  -2,  -4,   0,   3,   1,  -5,   1,  21, -22,  -1,   3 },
      {  -1,   1,  -4,  -8,  -1, -16,   5,   0,  -4,   3, -10, -18, -30,  11,  -2,  -1,   2,  -6,   8, -22,   3, -16,  10,  -5,  14,  15, -16,  17, -41,  21,  11,   3,  -8,  36,  11,  -5,  19,  17,  -2,  -8, -15,   5,   9,  -4, -12,  13, -11,   2,   6,  24,  26, -19,  22, -17, -16,  -2, -10,  -1, -32,  -8, -53,  21,  15,   3 },
      {   0,   1,   3, -25,  88, -12,  12,  -5,  -5, -15,   0, -27, -25, -32,   2,  -4,  -3,  -7,   8,  -2, -33,  15,   1,   3,  -5, -15,  -2,  -4,  -9,   1,  -4,   2,  -7,  -8, -11,   2, -18,  -3,   3,  -2,  -9, -14,  -2, -12,   1,   5,  14,   1,  -5,   4, -10,  -2, -10,   3,  -3,   1,   3, -11,  -3, -11,  13,  34, -10,   3 },
      {   0,  -3,  -4,   1,   6, -32,   0,  -5,   1,  -5,  16, -26,   1,  15,   6,   1,   7,   1,   5,   4, -24,   3,  -6,  -1, -16,   9,  28,   2,  -1,  19,  -7,   0,   6, -55,  14,  -4,   7, -33,  -1,   2,  17,  -4, -19,   5, -35,   8,   8,   1,   3, -32,  39,  11,  11,  13,   1,  -3,  -9,  20,   5,  -4, -38, -18,   8,  -4 },
      {   1,   7,   8,  -2,  10,  23,   1,   2,   1,   6,  11,   4,  -4,  -8,  -3,  -4,   7,  11,   4,  41,  -7,  22,   2,   1,  -1,  33,  -7,  -3,  28,  30, -17,  -1, -32,  10,  36,   6,  10, -16, -21,  -3, -36, -29,   5,   3,  21,  -1,   1,   8,  17,  17,  36,  -2,  20, -17, -10,   2,   1,  11,  -7, -10,  24, -20,  -5,  -1 },
      {   3,   2,  -1,  -8,   2,  19,  -1,   4,   3,  -3,   2, -12, -33,  -7,   4,  -4,   4,  15,   6,  19, -13,  15,  15,   2,   3,  -7,  14,   5,  22,   9,   6,  -5,  27,  10, -11,  17, -12,  20,  15,  -5, -14,  53,  10,  19,   3,  42, -20, -14,  34, -28,  -7,   6,   8,  -2, -35,   2,  -3, -11,  22,   2,  -9, -22,   1,   6 },
      {   2,   2,   3,  -4,  28,  37, -10,   8,  -1,   0,  -5,  -3,  10,  24, -17,   1,  -6,   2, -15, -33,  30, -28,  -8,   1,   3,   7,  35, -47,  -7, -18,  18,   1,  -2, -16,  40,  16,   1,  16,  12,  -5,  -4, -26,   1,   0, -10,  11, -20,  -2,  20,  -6,   7,   8,   9,   7, -20,   3,   4, -10,  15, -19,   5,  14,  -1,   4 },
      {   1,   0,  -4,   9,  12,   9,  -2,  -3,   0,  -2,  -3,   0,  17,  -1,  -8,  -1,   0,  -9, -20,  17,   7,   4,  -4,   1,  -8,   1, -38, -41,  26, -19, -15,   2,  14,  -4,  14, -44, -11, -31,  20,   8,  28,  23,  -1,  14,   1,  26,   2,  -2,   0,  -1,   0, -24,   8, -27, -16,   6, -17,  -8, -38,   0,  -8,  19,  22,   5 },
      {   1,   4,   5,   6,  19,  23,  10,   0,   2,   5,  11,  23,  18,  60,  12,  -1,  -6,  -4,  -7,  -1,   8,  32, -11, -13,  -8,  -7,  -6,  16, -16,   3, -33,   1,  -1,  -1, -35,  -1,  22,   0,  -9,  -1, -17,  16,  -4, -30,   1,   0,   5,   7, -11, -33,  14, -12,  18, -17, -13,  13,   5, -14,  22,  -8,   1,  28,  -2,   5 },
      {   0,  -2,   5,  -4, -14,  64,   3,  15,  -5,  -5,  -3,  -5,  -3, -23, -25,  -4,  -1,  -1,  18,   5,  -6, -25,  22,   5,  -4, -11,   5,  26,   9,  26,   4,  -6, -16, -19, -13, -14,   0, -12, -21,   0,   6, -23,  -6,  -1, -10, -16,  -5,  13,  10, -29, -24,  -9,  -3,  -5,   7,   7, -28, -15,  -3,  10, -22,  25,  24,   0 },
      {   4,   4,   0,  -1,  -3,  62,   6,  13,   2,   4,  -2,  -3,  -2, -21,  -1,  -1,   6,   0,   1,  -5,  -2,   8,   8,  -4,   1,   2,   3,   5, -15,   5,  -3,  -1,  14, -13, -14,   4,   6,  -4,   7,   3,  39,  32, -18,  -5,  -6,   0,  21,   3, -19,  50,  26,   6,  20,  27, -15, -11,  33,  24,  -6, -14,   0,   3, -17,  -5 },
      {  -1,   3,   1,  -4,   8,  45,  -1,   2,  -4,  -2,   0, -16, -10,  16,  -2,   1,  -5,  -3,  -7,   2, -31,  21, -23,  -7,  -5,   4, -20, -10, -12, -32,  -1,   4,   1,  23,   9, -18, -13,  -1,  12,  12, -26,  -4,  24,  -6,  -6, -18, -17,  -2, -20,  -7,  11,   1, -31,  36,  26,  -3,  -9,  11,   6,   3, -33, -36,  15,  -1 },
      {  -1,   1,   1,  -5,   2,  10, -15,  10,  -4,  -4,  -3,  -4, -32,   1, -18,   3,  -2,   3,  -7,  -1,  -3, -51,  -4,   2,  -2,   0,   3,  -6,  -1,  -8,   9,   8,  10,   8,  -4,  -6,  -3, -22,  -6,  -9, -11,  16,  21,   0,   0,   5,  40,  -1, -32, -16,  20,  -4,  -6, -62,  17,  -1,   7,  28,  34,   0,   8,  -6, -19,  -5 },
      {  -2,  -2,  -3,  -5, -22,  -9,  -9,   5,  -2,  -1,  -2,   1,  18, -73,  -4,  -3,  -1,   3,  -3, -22,  20,  36,   8,  -5,  -5,  -6,  11, -21, -26, -21, -22,   0,   5,   6,   0,  11,  -2,  -2,   8,   0, -21,  10,  14, -12,   7, -21,   1,   8,  13, -34,  26,   1,  16,  -9,  -1,   9, -12,  25,   5,  -7,  -7,  17,  10,  -9 },
      {  -2,  -1,  -5,   2,   5,   4,   2,   0,  -3,  -3,  -1,   0,  -8,   1,  -5,   0,  -2,  -1, -11,   9,   6,   1,  -5, -10,  -2,  -1,   7, -11, -11,   5,   4,   0,   6, -17, -13,  42, -24,   8, -16,  -7, -13,  15, -56,  -3,   1,   0, -44,  -2, -37,  13,  -1, -12, -10, -29,  15,  24, -28,  21, -24, -17,  12, -19,  25,   6 },
      {  -3,   1,   2,   1,  -3,  -5,   2,   2,  -3,   5,   6,   4,   6, -10,   5,   3,   1,   6,  10,   9,   6,   4,  25,   2,  -3,  11,  11,  11,  27,   0,  16, -11,   1,  -2,  19,  -4,  19,  15,  23,  -6, -17, -11,   9,  11,   6,  12, -12, -15, -88, -17,   9,  -4,  -4,  21, -28,  -2,  -3,   9,   2,  15,  -1,  33,   4,  10 },
      {   1,   0,   5,   3, -15,  18,  33,  -3,   0,   0,   2,   2,  12,   0,   3,  -7,   1,  -6,   7, -18,   6,  29,  15,   0,  -1,   1,   2,  -9, -47,  26, -20,  11,  -1,   1,  27,  18, -15, -21,  10,   0,  -7, -16,  -3,  33, -10,  43,  17, -15, -12,  -3, -20,   5, -26, -23,  14,   2,  17, -15,  -1,  31,  -4,  -8, -15,   4 },
      {  -1,   1,  -2,  -3,  17, -10,  -2,   8,   0,   1,   1,   8,   4,  -3,   6,  -1,   1,   3,  -5,  -2,  53,   5,  17,   6,  -2,  -8,  -7, -32,  -2,  59,  16, -10,  -1,   8, -13, -24, -32,  -7, -40,   2,   3,  14,  27, -11,   0,   7,   3,   2, -20,  -6,   5,   6,  -8,  29,  10,  -8,  -3, -10,   9, -29, -10, -19,   2,   1 },
      {  -1,   2,   3,   3,   4,   9,  -7,  -2,   0,   2,   2,   2,   6,  13,   0,   2,  -3,   2,  -2,  -3,   0,   5,  -8,  -2,   2,   0,   7,   0, -14,  11,  10,  -5,   1,   9,  -2,  25,   9,  -5,  16,  -1,  13,   3,   5, -11,  27,   9,  46,  -4,   5,   5,   0,   9, -16,  12, -16, -27, -94,  13, -10,   6,  10,  -1, -12,   4 },
      {  -2,  -2,  -2,  -8,   3,  -4, -68, -17,   1,   3,   0,  -3,   2,  13,  -9,  18,   0,   0,   5,  -9, -17,   7,  52,   1,  -1,  -3,  -6,   3, -25, -11, -33, -30,   2,   6,   5,  -6,   2, -24, -18,   8,   8,  -5,   4,  10, -11,   3, -21,   2,  -6,  27, -14,   7,   0,  -1, -17,   4, -12,   4,  32,  -2,  12,  -3,   7,  13 },
      {   0,   0,  -1,   0,   7,  16, -34, -12,   3,   3,   2,   1,  10,  20,  46,  12,   1,   1,   2,   5,   0,  18,  17, -11,   2,  -2,  -4,  -6,  13,  18,  26,  13,   4,  -2,   2, -11, -34,  28,   8, -10, -10,  -2,  -5,  23, -18, -13,   3,   1,  19,   2,  -2,  -5,  10, -11,  31,   0,  -5,  44,  -3,  19,  -7,  39, -23, -22 },
      {   0,   2,  -2,   1,  -4,   1,  64, -45,   0,   0,   1,  -1,   9,   9, -22, -11,   1,   4,  -2,   1,  -2,  18,  42,  28,  -1,  -4,  -4,  -5,   6, -29,  21, -11,   1,  -1,  -2, -19,  14,  11, -14, -21,  -2,  -4,   4,  -6, -22,   7,   3,   4,   8,  11, -11,   2,   3, -10,   4,   0, -18,  33,  19, -28,   1,  -8,  -5,  -9 },
      {  -1,  -3,   1,   0,   6,  14, -40, -27,   0,  -1,   1,   6,   4, -14,  11,  11,  -1,  -3,  -2,   6,   5,   5,  34,  -3,  -2,  -3,   3,  -1,   9, -27,   6,  10,   0,  -9,  -1,   8,  31,  27, -15,  -9,  -4,  -5, -14, -22,  -2,  25,  30,  -9,  -5, -17,  13,  -8, -10,  -7,  27,  -8,   9, -43, -38, -12, -18, -30, -14,  -7 },
      {   0,   1,   1,   0,  -1,   6,  15, -11,  -1,   2,  -3,   0,   2, -34,  36,   6,  -2,  -1,   4,  -9,   7, -20, -31, -28,   0,   1,   0,   2,  19,   4, -46,   7,  -2,   0,   7, -20,  12,  31,  -6,   9,   5,  -4,  -2,   6, -33,  16, -11, -10, -12,  26,  -1,   0,  10, -10,  11,   0, -41, -19,  42, -10,  -5,  -7, -15,  -3 },
      {   1,   1,   1,  -1,   6,  -2,   1,  24,   1,   1,  -1,   7,  -3,  -7,   4,   6,   2,   3,   3,   2,   7,  11,  10, -39,   0,   1,  -1,  15,  -4,  -4,  39,  21,   2,   0,  -6,  -7,  40, -46,  13,  18,   6,  -4,   9, -17,  11,  42, -40, -17,  14,   5,  -5,   6,   3,   3,  42,   0,  -6,  11,  10, -16,  16,  12,  -1, -11 },
      {  -1,   1,   2,   0,   1, -14, -28,  11,  -1,   0,   3,  -1,  12,   2, -80, -16,   2,   3,   6,   7,   3,  31, -24,  18,  -1,   4,  -3,   6,   7,   6,   2,  30,  -1,  -1,   1,  -4,  -1,  28, -11,  -8,   2,  11, -11,   5, -11,  26,   3,   6,  -4,  20,  14,  -6,   1,  13,  25,   8, -10, -11,  25,  22,  -4,  19,  -7, -10 },
      {   1,   2,   0,   1,   5,  -8,  23,  26,  -1,  -1,   1,  -3,  -7,  22, -11, -18,   0,   0,  -4,   3,  -7, -31,  49, -18,   1,  -2,   6, -12,  11,  -6, -56,  -1,   4,   5,   1,   9,  -6,  24,   0,  12,   0,  13,  13,  -1,   4,   8,  -5,  13,   3,  -6,  20,  32,   0,  18,  37, -10,  -6,   6, -27,  15,  19,  15,   8, -19 },
      {   1,   0,   0,   1,   4, -13,  -4,  84,   1,   0,  -1,   1,   9,   0,   2,  -1,   0,  -2,  -4,   3,   2,  13,   9, -11,   0,  -2,  -8,  -4,  10,  -6, -12, -20,  -3,  -4,  -2, -12,  14,  22,  -1, -21,  -3, -18, -12, -17, -21,  10,  14,  -4,  13,   6, -24, -15, -36, -11, -21,   1,   6,  33,  -2,  -8, -18, -17, -13,  10 },
      {   1,   1,   0,   3,  -1,   0,  14, -18,   1,   0,  -1,  -4,  -1,  -6,  20,  16,   0,   2,  -2,   2,   8, -18,   8,  11,  -2,  -2,   1, -16,   8,  -1,  -4, -17,   2,  -3,   0,   9,   3,  -8, -13,  -3,   3,   3, -19, -32,  33,  22,  -8,  12,  23,  17,  33, -44, -43,  16,   4,   9,   9,   8,  33,  50, -17,  16,   8,   3 },
      {   0,  -1,  -4,  -1,   4,   6, -13, -41,   1,  -1,  -1,   1,   2,  -3, -31,  25,   3,   2,  -2,   2,  16, -12, -18,  16,   3,   1,   1,   8,  10,  41, -37, -11,   2,   5, -12,   2,  13,  11,  48,  22,  -8,  -9,   1, -19,  -3,  14,  -5, -13,   9, -11, -21,  -2, -19,   1,  12, -11,  21,  42, -14, -25,  -5,   1,   4,   1 },
      {   0,  -1,  -1,   0,  -4,   6,  -9, -19,  -1,  -2,  -1,  -1,   0,  -3,   8, -20,   0,  -1,  -3,  -3,  -4, -11, -25, -13,   1,   1,   3,   2,   0, -17, -13,  33,   0,   4,   1,   3,   8,   1, -76,   4,   1,  -1,  12,  11,   9,  30,   7, -22,   3, -14, -22,  12,  -9,  22, -24,   6,  11,  46, -20,   0,  -4,  19,   7,  12 },
      {   0,   0,   1,   0,  -3,   9,   2,   4,   0,   1,   2,   0,  -2,  -9,   6,  45,  -1,  -1,  -1,   0,  -1,   7, -33,  25,   0,   0,  -2,   3,  -7,  -1,  10, -41,  -1,  -1,  -5, -12,   9,   1, -17, -28,   0,  11,  20, -13, -36,  20, -18,  10,   3,  -2,  20,  52, -15, -11,   4,  13,  -3,  -5, -32,  27,  28,  13,  -4,   2 },
      {  -1,  -1,  -2,  -2,  -1,   0,  -7, -28,  -1,  -1,  -2,   1,  -4,  -3, -10, -51,   0,  -1,   1,   2,  12,   3,  13, -37,   1,  -2,   0,  -3,   6,  14,  14,  21,  -4,  -5, -11, -15,  -3,   1,  14,  18,   3,  -3,   4, -36, -35, -22, -12, -12,   6,   8,  22,  22, -38, -21, -39,   5,   2,  -3,  -7,  32,  14,  -9,  -7,   9 },
      {   0,   0,  -1,   3,   0,  -6,  18, -13,   0,  -1,   0,  -3,   4,  -1, -40,  43,  -2,  -1,  -3,   1,  -7,   9,  -4, -77,   1,  -2,   3,  -5,   5,  -3,  22, -44,   1,   0,   6,   5, -10,  14, -22,  40,  -6,   3,  -2,  15,   2,  -2,  19,  13,   0,  -5,  -3,  -8,   6,  -5,  -7,  -8,   9,  -2,  -2,   7,  -8,   2, -11,   2 },
      {   1,   1,   0,   1,   1,  -3,  18,  11,   2,   2,   1,  -2,   1,   4,  -9,  76,   0,   3,   0,   1,   0,   1,  24,   3,   1,   1,   2,  -2,   4,   3,  -4,  60,  -3,   2,  -5,  -2,  -9,   5,   0,   3,   3,  -2,   4,  -9,  -6, -20,  21, -44,   6,   9,   8,   9,  -2,   4, -17,  33,  -4,  -4,   6,   4,   7,  -5,  30,   2 },
      {   0,   1,   0,   0,   5,   2,  -5,  13,   1,   2,   1,  -1,   4,  -2,  26, -12,   1,   3,   3,  -2,   0,  10,  -1,  28,   2,   3,   4,   4,  -1,  -2,  23,  -2,   0,   3,   6,   4,  12,  25,  -1,  73,  -4,   1,  -3,   1, -19,  18,  32,  32,  -2,   7,   0,   8,  -3, -17,   7,   3,   5,   6,  12,   0,  16,  -1,  62,  12 },
      {  -1,  -1,   0,  -1,   0,   0,  -9, -17,  -1,   0,  -1,   0,  -2,   1,   3, -14,  -1,  -1,   1,  -3,   1,  -4,  -5, -42,   1,   1,   1,   5,   1,  11,   1,  -8,  -1,   2,   3,  -2,   6,   1,  26, -52,   6,  -4,  12,   5,   8,  15,  33,  35,  -2,   8, -12,   9,  -4,  17,   2,  68,   6,   2,   6, -10,   5, -11,  40, -20 },
      {   0,   0,  -1,   1,   1,   2,  -3,   4,   0,   0,  -1,  -3,   0,   5,   0, -30,   0,   0,   0,   0,  -1,   0,  -8,   8,   0,  -1,  -1,   0,  -2,  -1,   4, -58,  -1,   2,   0,   2,   0,   3,  -6,  14,  -1,  -2,  -3,  -2,   2,  -9,  19, -91,   7,   6,  10,  -3,  11,   2,  15,  32,  -1,  -6,   2,   2,   1,   9,  13, -31 },
      {   0,   0,   0,   1,  -1,  -1,   3,   2,   0,   0,   0,  -1,  -1,   0,  -1,  -7,   1,   1,   0,   2,   2,  -8,   9,  11,   1,   0,   1,   1,   4,  -1,   2,  -3,   2,   2,   0,   0,   2,   0,   8,  28,   1,   2,   6,   2,   5,  -4,   4,   7,   5,  -2,   6,   4,  12,  12,  23,  71, -12,   6, -12,  -4, -13,  -1, -53,  76 },
      {  -1,   0,  -1,   2,   1,  -2,   0,   6,  -1,   0,   0,  -3,   1,   2,   3,   7,  -1,   0,   0,  -1,  -2,  -6,   1,  22,   0,   0,   0,   1,  -3,   0,  -1,   8,   1,   3,   0,   4,   6,  -5,  -5,  44,   0,  -1,   1,  -1,   6,   4, -13,  26, -12,  -4,  -9,  -1, -14,   0, -34,  44, -10,  -2, -11,  -4, -11,  -8, -43, -83 },
    },
  },
  { // 8
    {
      { -71,  53,  -9,   4,  -1,   1,  -1,   0,  72, -46,   1,  -1,   0,   0,   0,   0, -29,   7,  10,  -3,   2,  -1,   1,   0,   8,   6,  -8,   1,  -1,   0,   0,   0,  -3,  -5,   2,   1,   0,   0,   0,   0,   2,   2,   0,  -1,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
      {  59,  -2, -24,   2,  -4,   1,  -2,   0,  -2, -47,  34,  -4,   5,  -1,   2,   0, -48,  63, -16,  -5,   0,  -1,   0,  -1,  29, -26, -11,  12,  -3,   3,  -1,   1, -11,   2,  16,  -5,   0,   0,   0,   0,   3,   2,  -8,  -1,   1,   0,   0,   0,  -4,   0,   3,   2,  -1,   0,   0,   0,   2,   0,  -1,  -1,   0,   0,   0,   0 },
      {  56, -35,  16, -10,   2,  -3,   1,  -1,  52, -11, -31,   9,  -4,   2,  -1,   1, -43, -24,  39,  -3,   3,   0,   1,   0, -19,  48, -17,  -7,   2,  -2,   1,  -1,   5, -18,  -7,  13,  -3,   2,  -1,   1,  -4,   5,  11,  -5,  -1,   0,  -1,   0,  -1,  -1,  -4,  -1,   1,   0,   0,   0,  -3,   3,   2,   1,   0,   0,   0,   0 },
      { -45, -30,  47, -14,   9,  -3,   3,  -1, -28,  31, -21,  10,  -2,   1,   0,   0, -35,  41,   6,  -6,  -1,   0,  -1,   0,  43,   3, -39,   7,  -2,   1,  -1,   0,   9, -35,  21,   8,  -4,   2,  -1,   1,   0,  13,   3, -13,   3,  -2,   1,  -1,   2,  -5,  -6,   5,   1,   0,   0,   0,   2,   1,   2,   0,  -1,   0,   0,   0 },
      {   6,  41, -30,   1,  -4,   1,  -1,   0, -49, -28,  18,  16,  -1,   4,  -1,   2,  28,  -3,  35, -30,   7,  -6,   2,  -2, -19,  38, -48,  13,   1,  -1,   1,   0,  27, -32,  17,   9,  -9,   4,  -2,   1,  -8,   9,   2, -13,   5,  -2,   1,   0,   5,  -2,  -3,   6,   0,  -1,   0,   0,  -1,   1,   1,  -2,  -1,   0,   0,   0 },
      {  22,  11,   4, -18,   3,  -3,   1,  -1,  34,   6, -38,  25, -10,   4,  -3,   1,  35, -16,   0, -16,   2,  -1,   0,   0,   2, -51,  -9,  28,   1,   3,   0,   1, -18,  -6,  60, -18,  -5,   0,  -1,   0, -14,  27, -26, -14,  13,  -5,   2,  -1,  -1,  -7,   2,  17,  -5,   0,  -1,   0,  -1,   4,   3,  -7,  -1,   1,   0,   0 },
      {  15,  50,  -4, -14,  -2,  -3,  -2,  -2, -38,   1, -40,  25,  -3,   5,  -1,   2, -26, -21,  47,   1,  -4,   2,  -1,   1,  50, -11,  14, -30,   7,  -5,   1,  -1, -32,  25, -19,   4,   7,  -3,   2,  -1,  17,  -3,  -7,   8,  -5,   2,  -1,   0,  -4,  -5,   3,  -1,   0,   0,   0,   0,   5,   0,  -2,   0,   1,  -1,   0,   0 },
      { -23, -30,  33, -20,  13,  -4,   4,  -1, -22, -40,  10,  26,  -6,   4,   0,   1, -12,   7,  21, -19,   5,  -1,   1,  -1, -32,  16,  16,  11,  -2,  -1,   0,   0,   5,  48,  -1, -36,   1,  -2,   0,  -1,   0,   3, -48,  20,  10,  -3,   2,   0,  11,  -8,  20,  11, -14,   2,  -2,   0,   1,   2,  -4, -11,   4,   1,   0,   1 },
      {  -8, -39,  21,  -6,   6,   1,   1,   1,  10, -47,   6,  17,   0,   1,   1,   1,  61,  16,  21, -18,  -8,  -2,  -3,  -1,  16, -19, -21, -17,  13,  -5,   4,  -1, -36,  12, -14,  37,  -7,   2,   1,   0,  15, -13,  29,  -7, -12,   4,  -3,   2, -13,  11,  -1, -16,   9,  -1,   1,  -1,   0,   3,  -3,   4,   2,  -1,  -1,   0 },
      { -16, -30,  -4,  31, -13,   8,  -3,   3, -20, -28,  18, -17,  16,  -2,   3,   0, -17, -36,  26,  23,  -8,   2,  -1,   1,   4,  -7,  28, -17,   4,   2,   0,   1,  -8,   7,  38,  -1,  -4,  -4,   0,  -1,  -3,  39,  -1, -47,   7,  -1,   1,  -1,  -5,  11, -40,  16,  16,  -4,   3,  -1,   6,   0,   7,   9, -11,   1,  -1,   0 },
      {   8,  36,  32, -33,   0,  -5,   1,  -2, -13, -13, -29,   9,  13,  -4,   4,  -1, -12,  27, -24,  20, -20,   8,  -4,   2, -55, -11,   7,  -5,  19,  -1,   2,   0,  19,  19,   4,  26, -20,   1,  -2,   0,  11, -20,  15, -36,   7,   3,  -1,   2,  -9,  16, -18,  17,  12,  -9,   3,  -2,   6,  -7,   4,   2, -13,   4,  -1,   1 },
      { -10,  -7,  -7,  -9,  18,  -4,   5,  -1, -24, -27, -10,  33, -17,   9,  -4,   3, -27, -28, -15,  14,  12,   2,   1,   1, -16, -23,  23,  28, -13,   3,  -1,   0,   9, -20,  26,   3,   6,   1,  -2,   1,   6,   3,  48,  11, -24,  -2,  -4,  -1,   5,  13,  15, -55,   0,   6,   0,   1,  -6,  10, -17,   2,  19,  -6,   2,  -1 },
      {  13,  13,  33, -20,   2,  -4,   1,  -2,  -4, -44, -10,  -5,  10,   0,   3,   1,  32,   8, -16,  38, -18,   6,  -4,   1,  23,  11,  22, -35,  -9,   4,  -3,   1,  21, -45, -11, -17,  28,  -5,   5,   0, -38,  27,  -3,  20,   0,  -6,   3,  -3,  11,  -4,  -2,   3, -11,   5,  -1,   1,  -8,   0,   6,  -3,   3,   1,  -2,   1 },
      {  15,  27,  33, -35,  21, -15,   4,  -5,  12,  17,  42,  -8, -26,   5,  -6,   0,   9,   2,  -1, -19,  13,  -5,   1,  -1,   9,  21,   9,  -3, -20,   1,  -2,   0,  -6,  11,   4, -28,   9,   0,   0,   0,  12,  11,  -3, -25,  -9,  -2,   2,   0, -12,   0, -41, -30,  31,   5,   3,   1,  -2,   1, -18,  39,   9, -10,   3,  -2 },
      {   5,  21,  38,   8, -13,  -2,  -1,  -1, -13,  -2,   4, -46,  22,  -6,   3,  -1,  -6,  -8,  -4,  15,   1,  -4,   3,  -1, -36,   2, -34,   7,  -5,   9,  -1,   2, -53,   2,  17,  21,  20,  -4,   1,   0,  26,  31,  -8,  21, -27,  -2,  -4,  -1,  11, -29,  13, -13,  -7,  10,  -1,   2,   1,   9,   3,  -7,   8,  -4,   0,   0 },
      {  -5,  -4, -25,  15,  -7,   1,   0,   1,   9,  26, -13,  16,  -4,   4,  -3,   1,   0,  54,  36, -13, -11,  -9,  -1,  -2, -39, -26,  13, -38,   7,   5,  -1,   3,  20,   8,  -1,   9,  21, -12,   4,  -3, -21,  37,  -5,   3, -31,   4,   0,  -1,  17, -15, -12, -24,  11,  10,  -2,   3,  -7,   1,  -7,   9,  10,  -6,   3,  -2 },
      {   1,  12,  21, -22,   9,   0,  -1,   0,   1,  15,  54,  10,  -9,  -8,   1,  -3, -26, -26,  -2, -31,   1,   7,  -1,   3,  -7, -31,   1,  -7,  32, -10,   7,  -2, -13,  -4, -22,  28,  -8,   3,   1,  -1, -46,  29,  19,  13,  10,  -3,  -2,  -1,  23,  16,  -6,  12, -24,  -3,  -1,  -1,   0, -13,   9, -15,  -2,   6,  -2,   1 },
      { -15, -18, -51, -25,  39, -14,   9,  -4, -10,  -7, -14,  16, -24,  13,  -2,   1,  -4,   4, -34,  13,  11,   2,   0,   1, -18,  17,  -9,  -8,  -2,  -1,   4,  -1, -45,  -9, -15,  12,  17,   2,   2,   2,   5,  23,  -7,  10,   0,  -3,  -1,  -1,   0, -12, -23,  32,   6,   1,   1,  -1,   6,  -2,  24,  25, -26,  -3,  -1,   0 },
      {   0,   1, -21, -17,  11,  -1,   2,   1,  13,  26,  30,  31,   1, -12,   2,  -3,  10,  17,   7,  38, -43,  19, -10,   3,   2,  39,   8, -13,  11, -11,   0,  -2, -21,   2,  31,  -5, -17,   7,  -2,  -1,  29,  14,   7, -10,   8,  -5,  -4,  -1,   2,  -1,   7, -15, -34,   2,   2,  -1,   2,  -5,  -9, -36,  22,  12,  -2,   2 },
      {   2,  -5, -11, -18,  12,  -6,   4,  -2,   5,  -9,   4,  -3,   3,  -1,  -2,   0,   3, -22, -26,   8,  -6,   4,   1,   1,  36,  10,  -6,  10,  22,  -3,  -2,   0,  41,  36,   2,  36, -27, -12,   0,  -5,  -6,  14, -18, -17, -33,  19,  -6,   3,  28, -54,   9,  -7,  18,   4,  -4,   5, -14,   9,  11,  -3,   5,  -8,   4,  -2 },
      {   6,  11,  12,  30, -12,  12,  -4,   4,  -2,  -5, -32,   7,   2,  -3,   0,   0,  14,  30, -13,  -8,  36,  -9,   5,  -2,  13,  35,  32,  38, -13, -16,   1,  -6, -20,  24, -14,   7, -28,  11,  -6,   1, -14,  36,  24,  -5,   1,  -6,  -4,   0,  26,   2, -26,  -9, -20,   3,   0,  -1,   2, -22,   7,  -7,   2,   5,  -1,   2 },
      { -13, -17, -33, -58, -24,  22,  -3,   3,  -6,  -4, -24, -23,  52, -27,  10,  -4,  -4,  -6,  -7, -23, -18,  11,   1,   0,   2,   7, -12,  15,  13,   1,  -4,   1, -10,   7,  -7, -19,  -1,  -2,   1,   1, -17,  10,   1,  10,   8,  -3,  -1,   0,  -5,  15, -18, -23,   0,   2,   3,   0,  -6, -10, -19,  24,  25,  -2,   2,  -1 },
      {  -2,   4,  12,  22, -28,  15,  -4,   4,  -7,  -2,   3,  34,  -3, -12,  -1,  -1,  -8,   1, -27, -20, -14,  12,   2,   1, -13,  15,  23, -12,  14,  -9,  -2,   0, -44, -38,  -6, -10, -25,   6,   8,   1, -19, -12, -15, -15,  12,  25,  -4,   4,   3, -30,  22,  -6,  35,  -9,  -1,   2, -18,  28,  -7,  17,   9, -12,   5,  -4 },
      {   3,  -5,   6,  12, -17,  17,  -9,   5,   0, -13,  -7,  11, -17, -11,   3,  -1,  -4, -15, -31, -35,   1,   8,  -2,   5,  -2, -16,   2, -24,  10, -11,   7,   1,  30, -18, -15,   2,   6,  16,  -4,   3,  61,  17, -14,   1,   8, -20,   2,  -3, -20, -26, -27,   0, -25,  21,   4,  -1,   9, -15,  14,   0,  35,   2,  -6,   3 },
      {  -5,  -3, -13,  -2,  26, -27,  11,  -7, -12, -12, -24, -66, -33,  23,  -5,   5,  -3,   7,   5,  -9,   6,   6,  -3,   3,  -9,  -8,   2, -23,   8, -15,  10,  -4, -13,  -5,   0,  -9, -20,  22,  -4,   3, -19, -12,   0, -24,  23,   7,   1,   1,   6, -10,   2, -11,  -6,  -6,   2,   2, -14,   0,  -6, -28,  47,   9,  -1,   1 },
      {   1,   1,   1,   5,  -9,  -5,   7,  -5,  -1,  -1, -14, -18, -15,  17, -10,   6,   2,   4, -31, -37,   6,   0,   5,   2,  11,  35,  19, -20,  46,   2,  -2,   0,   5,   2,  33,  18,  12, -29,   6,  -5,  10,  23,  -8,  16, -11,   4,   0,  -3, -20,  43,  22,   0,   0, -19,  -3,  -2,  14,   8, -32,  -9, -17,   6,   4,  -2 },
      {  -2,  -3,  -8, -43,   4,   6,   4,   1,   3,   5,  11, -27,  21,  -7,  -1,  -1,   9,  17,  36,  -9,  31,  -3,   4,  -1,  -4, -12,  59,  16,  -2, -11,  -4,  -1,   7, -39,  -8,  19, -17,   5,   3,   0,  37,  15,   3,   2,   6,   2,  -5,   1,  -8, -20,  22,  12,   2,  -6,  -1,  -1,   2,  24,   8,  -5, -21,  -2,   2,  -2 },
      {   5,   8,   7,  30,  42,   5,  -8,   5,   1,   2,  -4,  -4,   8, -16,  14,  -7,   4,   7,  12,  27,  14,  28, -13,   6,   0,  -3, -11,   9,  35, -45,  16,  -5,  14,   8,   5,   4,  -8,  21,   2,  -2, -10,  12, -11,  26,  14, -17, -11,   0, -22,   7,   7,  -1,  -2,   4,   7,  -4,   5,  28, -18,  47,   8, -23,  -1,  -2 },
      {  -4,   1,  -2, -12,  -1,   9,  -6,   6,  -4,   5,   5,   3,   1,  -8,  11,  -7,  -3,   4,  -1,  -1,  17,  -6,  -6,   1,  -1,  16,  20,  -1, -11,   0,  10,  -4, -12,  -1,  17,  26,   2,  -7, -15,   1, -45, -22, -16, -11, -35, -14,  21,   2, -57,   0,  10,   7,   1,  41,  -4,  -2,  30,  -2,  27, -16,  31, -15, -12,   4 },
      {  -3,  -1,  -5, -20, -20,  30, -24,  14,   0,   4,   0,   0, -14,  -6,  17,  -3,  -1,   3, -13,  -2,  27,  -2,  -6,   2,   3,   9,  -9, -32, -18,   5,   6,   4,   6,  35,  17,   8,  31,  24, -14,  -1,  -8,   4,  31, -13,  26, -37,   1,   2,   9, -10,  18,   9,   4,  -5,   6,  -8, -33,  46,  -7, -11, -12,  -3,  -2,   0 },
      {   0,  -4,   0, -14, -25,  -1,   8,   1,   7,   5,  12,  18,  26,  36, -15,   5,   3,   0,  -3,  30,  49, -26,  11,  -1,   2,  -1,  -2, -22,  20,  -1,  -2,  -2,  -2,  -7,  18,  18,  -1,   9, -11,   1,   9, -20, -19,   8,  21,   0,   8,   0,  44,  15,  -4,   6,   3,  -2,  -5,   1,  17, -14,  -7,  21,  42, -20,  -2,   0 },
      {  -4,  -4,  -4, -13, -37,  -9,   5,  -4,  -2,   1,   5,   9, -15,  12, -10,   3,  -3,  -6,  -8,  34,   7, -28,   7,  -4,  -5, -13, -22,  -5, -24, -10,   6,  -1,   7,   4, -33,  11, -32,  20,  -3,  -1, -13,  40, -27,  -7,   6,   5,  -7,   3, -49,   3,  15, -21,  -7, -10,   9,   0,   9,   1, -35,  -3, -14,  17,   4,  -1 },
      {   0,   5,   1,  20,  19, -38,  25, -15,   1,   6,   2,  17,  57,  20, -23,   6,  -4,  -2, -10, -17, -12, -31,  11,  -2,   2,   3,   7,   3,   4,  19,  -3,  -1,   1,  11, -10,   3,   6,  -6,  -3,   6,  -3,   9,  22,  -7,  22, -14,   4,  -4, -36, -14,  -4,  13, -22,  10,   1,   2, -35,  31,  -1,   2,  19,  -9,   3,  -1 },
      {  -4,  -5,   1, -12, -63, -25,  14,  -5,  -1,  -1,   5, -13, -32,  33, -21,   8,   1,   7,  13,   8, -28,  14,  -2,  -1,   2,   4,   7,  24,  31, -20,   6,  -4,   0,   6,  10, -10,   2,   8,  -7,   0,  -7, -13,  21,  15,   0, -26,   1,   0,  -3, -19,  -4,   0, -11,  30,   1,   0,  -5,   5,  23,  26, -22, -12,   0,  -1 },
      {  -2,   0,   3,  -4, -13,  19,  -1,  -1,  -3,   1,   6,   7, -21,  -6,  -1,   3,   2,   7,  12,  27,  36,  15,  10,  -8,  -3,  -1,  -2,  18,  25,   1, -20,   5,  -3,  -1,   0,   5,  17, -34,  12,  -4,  -4, -13,  -6,   8, -10,  17,  -9,  -5, -34,  -7, -32,   8, -16, -33,  16,   5, -57,   2,  10,  -5,  27,  27,   1,  -3 },
      {   1,   2,   0,  -1, -11,   6,   3,   1,   2,   2,   2,  -1, -22,   6,  -3,   1,   5,   5,   7,   7, -17,  19,   8,  -1,   2,   3,  -3,  26,  -3,  26, -16,   3,   8,  17, -19,   3,  25,   0,  14,   0,  -1,  30,  27,  -5,  27,  25,  -2,  -4,  -8,  -2,  18,  20,  16, -28,  -9,   4,  55,  22,  18,   5,  50, -17,  -6,  -1 },
      {   0,   0,   3,   0,   0,   1,  14, -16,   2,   3,   4,   7,  32,   7, -18,   9,  -1,  -1,  -5, -15,  14,  14,  -2,   0,  -1,   3,  -3, -28, -13, -23,   1,   3,   8,  14,  33, -13,   5,  35,   3,  -3, -13,  -4,  18,  28,  -3,  13, -27,   3, -24, -24,   3, -20,  17, -39,   8,   9,  20, -23,  44,  -2,  -9,  25,   3,  -4 },
      {  -1,  -2,  -2,   1, -23, -37,  -1, -15,   4,   4,   4,  15,  25,  22,  25,  -4,   1,  -1,  -3,  -3,  16,  33, -49,  19,   1,  -1,  -4,  21,  -2, -24,  37, -13,   1,  -6, -22,  -7,  33,   0,  -7,   1,  12,  15,  -9, -17, -16,   4,   5,   5,   4,  18,   2,   7,  32,  -9,  -5,  -6, -16,   0,  -1, -29,  10,   5,  -4,   2 },
      {   0,  -2,   0,   8,  12,   2,   1,  -2,   1,  -2,   2,   4,  11,  10,  -5,   4,   1,  -1,   3,   0,   4,  21,   3,  -4,   0,   0,  10,  19,  19,  10, -19,   3,  -7,  -9, -13,  17,  37,  -5,  -5,  -5, -16, -15,  -9, -32,  13, -52,  -3,   7,   4, -43,  10, -15,   3, -10,  15,  -1,  32, -27, -41,   4,  -9,  39,  -1,   1 },
      {   1,   1,   1,  -4,   8,  26,   1,   8,   0,   0,   1,   3,  11,  10, -12,   5,   1,   2,  -4,  16,  36,  -2,   6,  -3,  -5,  -6, -24, -12,  41, -12,  -1,  -3,  -1,   3, -15, -55,  -4, -28,   9,   9,   2,  11,  33, -17,   4,  15,  -1,  -6, -12, -16,   7,  -2,  15,  45, -21,   1,  19,  -1,  -4, -31, -17,  -1,  11,   0 },
      {   2,   4,   6,   0,  22,  40, -28,  38,   0,   3,   3,  -5,   2,  57,   6, -11,  -1,   1,   5,   4, -29, -25,  -3,   6,   1,   4,   9,  19,  24,  23,  24,  -7,  -1,  -4,  -3,   0,   4,  18, -27,   4,   4,  13,  -5,  13,   9,   6,   5,   4, -18,  -4,   8, -13,  23, -13,  -5, -10, -26, -30,  15,   1,  -1,   2,  -2,   7 },
      {  -2,  -3,   0,  -7,  -5, -18, -47,  15,   1,  -1,   1,   0,   6,  30,  33, -11,  -1,  -3,  -3,  -8,   1, -26, -32,  10,  -2,   0,   3,  -6,  -5,  -4,  15,  -7,  -1,  -3,   9, -11,  -9, -32,  18,   2,  -8, -16,  17,   0,  -6,  21, -21,   1,   0, -32, -17,   3, -45, -23,  20,   2,  33,  17, -15,  14,  -3,  -4,   2,   3 },
      {   1,   1,  -1,  -1,   6,   6,   4,  32,   1,   1,  -2,   1,   8,  13, -16, -15,  -1,  -2,  -6,  -2, -16, -22,  21,  -9,  -1,  -4,  -9,   2,   6, -44, -15,   5,  -2,  -6, -15, -24,  -9,  -7,  -5,  -5,   4,   1,  -4,  -7, -41, -40,  24,   5,  17,  17,  -6,   1,   9, -33,  19,   0,  17,  45,  24,  -8,  13,  24, -12,  -5 },
      {   2,   3,   2,  -2,   7,  51,  18,  10,   1,   3,   2,  -4,  14,  58, -26,  10,  -2,  -2,  -4, -11,  -8,  46,  -9,  -4,  -2,  -2,  -4,  -3, -41, -21, -11,   1,  -1,   3,  -4,  14,   2,  -2,  31, -11,  -2,  -2, -10, -19, -11,  10, -10,   5,   2,  12,   2,  11, -29,  16,   8,   2, -11,   2, -12,   4,  -3, -21,   1,  -6 },
      {  -1,  -1,  -2,   1,  -5, -50, -25,  57,   0,   1,  -1,   5,   9,  13,  15, -36,   0,   0,  -2,  -3,  11,  21,  32,   5,   1,   3,   1,  -2,   2,   5, -35,  20,   2,   3,   4,   8,   4,  27,  39, -18,   0,   1,  -6,   3,   9,  11, -10,   0,  -6,  10,  -3,  -7,   9,  32,  -4,  -1,  -7,  -3,   5,  -8,  -4,  15,  -2,  -1 },
      {  -1,  -1,  -1,   1,   0, -11, -26, -18,  -1,   0,   1,  -2,  -1,  11,  14,  -9,  -1,   0,   0,   1,   9,  11, -20,   3,  -4,  -3,  -8, -15,  -5,   1, -49,  -2,  -3,   1,  -6, -10, -33, -26,  10,  -3,   5,  15,   6,  -9,  -1, -47,  15,  -4, -11,   3,  33,  -7,  -1, -29, -27,  10, -19, -35,  24,  16,   4, -33,  13,  -3 },
      {  -1,  -2,   0,  -3,  -2,  -9,  -6,  34,  -2,  -4,   0,  -1,  -3, -23, -34, -17,  -2,  -2,   0,  -1, -12, -16,   6,   3,  -1,  -1,   3,   6, -16, -57,   5,  -2,   1,   3,  10,  12,  32,  -3,  -5,   4,  -2,  -5,  12,  -1,   2,  13,  23, -20,  -9, -21,  -2,  21,  -1, -23, -41,  13,  -8, -30, -32,  -5,   2, -22,  24,   0 },
      {   0,   1,   1,  -2,   1,   3,  -5,  -3,  -1,   0,   0,  -3,  -8, -12, -10,   9,  -2,  -1,   0,   1, -11, -43, -19,   4,   0,   4,   2,   4, -11, -30,  10,   3,   1,   4,   7,   6,   6, -28,  44,  -9,   6,   8,  12,  -3,  17,  -6, -24,  18,   6,   8,  37,  19,  25,  30,  23, -13, -15, -27,  11,  24,  24,  36, -18,  -4 },
      {  -1,  -2,  -2,  -3,   4,   1, -40, -24,  -1,  -1,  -2,  -1,  -1,  -8, -17,  14,   0,   0,  -3,   2,  -9, -29,   7, -20,   1,   3,   4,  15,  20,  -9,  -4,  -1,  -2,  -7,  -9,  -1,  31,  26,  19,  -5,  -2,  -8, -21, -22,  -5, -12, -57,   2,   1,  19,  -8, -13,  -2,  -3,   8,  16,   1,   2,  15, -31,  -5, -47,  17, -11 },
      {   2,   2,   0,   2,  -3,  -6,  63,  59,  -1,  -1,  -3,  -1,  -5, -20,   9,  -3,   0,  -1,  -1,  -2,  11, -12, -32,   8,   0,   1,  -4,  -2,   7,   4,  -3,   5,  -1,  -1,  -6,  -1,  18, -17, -18,  -4,   0,   1, -13, -27,  18,   7, -23,   3,   1,   8,  16, -33, -23, -11,  14,   0,   0,  -6,  31,   7, -12, -24,   5,  -3 },
      {  -1,  -2,   0,  -1,   2, -22, -35,   7,   1,   1,   1,   0,   7,  -5, -12,  18,   1,   0,  -1,  -4,   5,  24,  25, -33,   0,  -1,  -1,  -8,  -1,  -7,  -3,  17,   0,  -1,  -4, -16,   6,  -4, -64,   3,   4,  13,  12, -20, -20,  29,  -9,   6,   1,   6,  34,  20, -17,   4,  37, -11,  -8, -21,  -1,  23,   3,  -8, -18,  -2 },
      {   0,   1,   1,   0,   2,  15,  13, -41,   0,   1,   0,   0,   0,   9,  45,  -4,  -1,  -1,  -2,  -1,  -4, -30,   6,  13,   1,   2,   1,   7,  19, -27, -28,  31,  -1,   1,  -1,  -6,  28,  35,   8, -20,   0,  -1,   2, -27, -10,  33,  35, -11,   0,  -1,  14,  -8, -29,   1, -14, -11,  -3,  -2,   6,  18, -11,   6, -17,  10 },
      {   0,   0,   0,   2,   6,  -2, -32,  -7,   0,   0,  -1,  -1,   3,   7, -15,  20,  -1,  -1,  -3,  -3,  -7,   1, -37, -19,  -2,  -1,  -1,   6, -11, -28, -40,  14,   0,   3,   1,  25,   6, -27, -30,  11,  -1,   0, -16,  15,  48,  21,  22, -12,   2,   7, -18, -36,  11,  20,  -2,   2,   7,  25,  18, -10, -10,  16,  -3,   3 },
      {  -1,  -1,  -1,  -2,   3,  -5, -31,   9,  -2,  -2,  -3,  -2,   4, -23, -68,   6,   1,   3,   0,   4,  16,   4, -37,  21,   0,   2,  -2,   1,  16,  33,   9, -19,   0,   1,  -4,  -6,   4,  28,  30,  -5,   0,   1,  -5, -18, -13,  13,  43,   2,   3,   5,  11, -13, -24,  -7,   2,  -2,  -1,   7,  12,  18,  -8,  13, -13,   4 },
      {   0,   0,  -1,  -1,   0,  -3, -12, -14,   1,   1,  -1,   1,   4,   2,   5, -15,   1,   3,   0,   1,   2,  18,  58, -14,   0,   0,  -2,  -3, -10, -11,  59,  -9,   1,   2,   1,   9,  11, -35,  25,  23,   1,  -2, -13, -21,  27,  -3,  16, -10,   4,   4,  13, -40, -16, -10, -11,  -7,  -1,   2,  31,  11, -11,   0,  -9,   8 },
      {   0,   0,  -1,   0,   0, -11,   7,  25,   1,   1,   1,   2,   4,  -4,  18,  62,  -2,  -2,  -1,   1,   1,   0,  -7, -47,   0,   0,   0,   2,   4,   0,  15,  27,   0,   1,   2,  -1,  -9,   5,  39, -14,   0,   0,   6,  14,   3,  -9,  37,   1,  -6, -15, -11,  -4,  12, -17,  14, -26,  -1, -14, -10, -21,  -6, -34, -29,   8 },
      {   0,  -1,  -1,   2,   0,   3,   1, -25,  -1,  -2,  -2,   0,  -2,  -4, -25, -47,   1,   1,   2,   1,   1,   0,  13,  45,  -2,  -1,  -3,   0,  -2,  -6,  -9,   8,   0,   0,   1,   0,   0, -18,  -6, -22,  -1,   2,   5,  14,  17,   6,  16,  27,  -3, -12, -12,  -9,  12,  -6,  52, -35,   1, -12,  -7, -28, -10, -45, -24,   1 },
      {   0,   0,   0,   1,  -1,   4,   3, -14,  -1,  -2,  -2,  -1,  -1,   2, -22, -53,   0,   0,   0,   1,   6,   2, -30, -47,   1,   2,   2,   1,   3,  13,  35,  76,  -2,  -3,  -2,  -3,  -9,  -5,  -6, -36,   2,  -1,   2,   0,   2,  -6,  -7,  14,   1,   0,  -3,   2,  -4,  -1, -24,  12,   3,   5,  -1,   4,   4,   5,   5,   3 },
      {   0,   0,   1,   0,   2,   0, -13,   5,   1,   1,   1,   1,   1,   2,   4,  43,   2,   3,   3,   3,   0,   5,  17,  61,   0,   2,   1,   4,  -3,   2,  25,  59,  -2,  -1,  -4,   2,  -5,  -8,  -3, -12,   1,   1,   0,   0,   3,  -6,  20,  -3,   1,   4,   4,   0,   1,   4,  23,  33,   4,   0,   8,   0,   3,   3,  54, -25 },
      {   1,   1,   1,   1,  -1,   4,  11,  -7,   0,   0,   0,   0,  -2,   3,  10, -39,   0,  -1,   0,   0,   1,  -4, -18, -36,   2,   2,   3,   1,   7,   5,  -1,   0,   2,   4,   3,   2,   2,  16,  19,  54,   2,   3,   3,   1,  -1,   6,  32, -35,   3,   0,  10,   5,   6,   8,  59,  21,   0, -13,  -2,  -4,  -2,  -4,  24, -45 },
      {   0,   0,  -1,  -1,  -2,   0,   1,   2,   0,   0,   1,  -1,   0,   0, -14,   1,   0,   0,   1,   1,   1,   2,  -8,  25,   1,   1,   1,   1,   1,   0,  -9,  55,   0,   1,  -1,   1,  -4,   6,  11,  87,  -4,  -1,  -6,  -2, -15,  -9, -24,  -8,   1,   2,   0,   4,   1,  -7,  -8, -36,  -1,   1,  -3,   0,   2,  -4,  -7,  46 },
      {   0,   0,   1,   1,   1,  -3,   0,  -1,   0,   0,   0,   0,  -2,  -2,   9,   1,   0,  -1,   0,   1,  -1,   0,   2, -22,   0,   0,   0,   0,   3,  -5,  -9, -10,   2,   1,   4,   2,  10,   5,   6,  23,   1,   3,   1,   1,   1,  10,  22,  98,  -2,   1,  -3,   0, -10,   0,  -6, -26,   3,   2,   6,   3,   4,   4,  61,   0 },
      {   0,   0,   0,   0,  -1,   1,   3,  -3,   0,   0,   0,   1,   1,  -1,   0,  -6,   0,   0,   0,  -1,   0,   1,   0, -14,   0,   0,   0,  -1,  -1,  -1,   0, -16,   0,   0,   0,   0,   0,   1,   4, -29,   1,   2,   2,   0,   0,   0,   7, -29,   3,   4,   8,   4,   8,   6,  39,   4,   1,   0,   6,   1,   4,   0,  53,  98 },
      {   0,   0,   1,   0,   0,  -1,  -4,  -3,   0,   0,   0,   0,   1,   0,  -7,   1,   0,   0,   0,   0,  -1,   2,  -3,  -3,   0,   0,   0,  -1,  -4,   5,   2,  -3,  -1,  -1,  -2,   0,  -4,   0,  -4, -24,  -2,  -4,  -3,  -2,  -5,  -3, -10, -47,   2,  -1,  -2,   0,  -3,  -2,  -9, -97,   7,   3,   8,   3,   9,   9,  45, -38 },
    },
    {
      { 111, -54,  -5,  -5,  -1,  -2,  -1,   0, -26,   1,   5,   0,   1,   0,   0,   0, -13,   9,   0,   0,   1,   0,   0,   0,  -5,   1,   0,   0,   1,   0,   0,   0,  -4,   2,   0,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
      {  37,  33, -32,  -2,  -5,  -1,  -2,   0,  92, -40, -28,   3,  -3,   0,  -1,   1, -18, -29,  11,   6,   3,   1,   1,   1, -16,   1,  12,   1,   2,  -1,   0,   0, -10,   1,   5,  -1,   1,  -1,   0,   0,  -6,   2,   2,   0,   1,   0,   0,   0,  -4,   1,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 },
      {  22,  29,  -9, -11,   0,  -5,   0,  -1,   3,  76, -47, -11,  -2,  -4,  -1,  -1,  44, -18, -45,  11,   1,   3,   0,   2,   2, -35,  -1,  17,   2,   2,   0,   1,  -2, -19,  10,  10,   0,   0,  -1,   1,  -4,  -6,  11,   3,  -1,   0,   0,   0,  -2,  -1,   3,   0,   0,   1,   0,   0,  -3,  -1,   2,   0,   0,   0,   0,   0 },
      {  18,  61, -30,  -9,  -5,  -6,  -1,  -1, -14,  -9,   6,  -8,   0,  -5,   2,   0, -31,  59,  -7, -24,   3,  -2,   2,   2,  31,   4, -49,  -4,   8,   0,   0,   2,   5, -34, -15,  18,   9,   1,  -1,   2,   4, -15,   5,  16,   3,  -2,  -2,   0,  -1,  -3,   3,   5,   0,  -1,  -1,   0,  -1,   0,  -1,   2,   1,   0,   0,   0 },
      {  -7, -53,  22,   3,  -1,   2,   1,   1,  64, -10,  23, -15,  -6,  -3,  -2,  -1,  51,  37, -22, -30,  -1,  -3,  -1,  -1,  10, -16, -37,  -1,  10,   2,   2,   1, -13, -16,  -2,  13,   8,   1,   0,   1,  -9,  -2,   4,   7,   3,  -2,   0,   0,  -6,   1,   2,   4,   0,  -1,   0,   0,  -3,   1,   1,   1,  -1,   0,   0,   0 },
      {  13,  16,  48, -17,  -7,  -5,  -2,  -3,  29,  59,   5, -20, -10,  -4,  -4,  -1, -11,  -1,  27, -12,  -6,   0,   0,   0, -11,  48,   0, -36,  -1,  -3,   1,  -1,   5,   9, -45, -15,   9,   2,   3,   1,  -2, -11, -21,   7,  10,   2,   1,   0,   1,  -9,  -9,   9,   4,  -2,   0,   1,  -1,  -4,  -1,   5,   1,  -1,   1,   0 },
      {  -4,   6,  24, -13,   4,   1,   0,  -1,  19,  21,  32, -27,   3,  -1,  -1,  -3, -58,  18, -39,  -2,  20,  -3,   4,  -3, -40,  -5,  -6,  52,   4,  -2,   2,  -2,   1,  25,  26,  11, -24,  -4,  -1,  -3,   6,  11,  -3, -14, -17,   3,   2,   0,   5,  -2,  -3,  -7,  -5,   5,   0,   1,   4,  -3,   0,  -3,  -1,   1,   0,   0 },
      {  22,  49, -16,   8, -12,   2,  -5,  -1,  -8,   0,  70,  -2, -15,   2,  -4,  -2,  45,   4,   5, -24, -13,  -1,  -1,  -3, -14, -32,  19,  -6,  -6,   7,   2,   0, -17,  29,  11, -30,   0,   3,   2,  -2,  -5,   6, -24, -16,   6,   4,   2,   0,  -3,  -3,  -6,  -1,   6,   1,   0,   0,   1,  -3,  -3,   2,   3,  -1,   0,   0 },
      {   0,  12,  29, -27,  -7,   0,   0,   2, -29, -26,   5, -25,   1,   7,   3,   3,  -2, -56,   5, -31,  13,  10,   2,   4, -41, -24, -20, -15,  40,   8,   5,   2, -20, -12,  -5,  30,  29,  -4,   0,  -1,  -3,  10,  12,  19,   4, -12,  -4,  -2,   2,  10,   3,   9,  -3,  -7,  -1,  -1,   1,   6,   0,   4,  -3,  -3,   0,   0 },
      {   7,   7,  41, -17,  -7,  -6,   0,   0,  -2, -35,   2, -43,   4,  -6,   6,   1, -13, -14, -42,  21,  17,   4,   4,   5,  62, -17,  31,   5, -11,  -2,  -5,   2,   2,  -6,  -7, -35,   7,   2,   1,   3,  -7, -10, -17,   6,  22,   2,   0,   0,  -8,   4,  -4,  12,  10,  -2,  -2,  -1,  -2,   3,  -3,   7,   3,  -3,  -1,   0 },
      { -22, -32, -55,  35,  -3,   7,   0,   4,   6,  28,   4, -18,  -7,  -1,   1,   1, -45,   4, -19, -10,   0,   8,   2,   3, -17, -34,  23, -18,   7,  11,   0,   3, -13,   2, -22, -21,  31,   2,   1,   1,   1, -12, -13,  30,  22,  -5,  -2,  -1,   3,   4,   9,  15,   1,  -7,  -3,  -1,   2,   3,   1,   7,  -3,  -4,   0,   0 },
      {  -4,  -5, -47, -25,  14,   3,   4,   2, -12, -19, -11, -26,  19,   2,   3,   0,  41,  -2, -27, -11,  31,  -4,   2,  -2, -15,  49,   0,  19,  18, -12,   0,  -4,  14,  41, -26,   1,  -7, -12,  -2,  -2,   1, -14, -30,  13,  -1,   0,   3,   3,   1, -10,  -1,  19,  -1,   1,   2,   1,  -1,  -4,   4,   6,  -3,   0,   0,   0 },
      {  -8,   4,   4, -50,  15,   1,   4,   0,  -6, -12, -26,  -8,  23,   4,   8,  -1,   7,  53,   8,  39, -10,  -9,  -2,  -7, -36, -16,   6, -15, -30,   3,   4,  -1, -58,   8,  -5,  -9,  11,  18,   8,  -1,   0,  17,  -2,  10,   7,   4,  -5,  -4,   4,   2,   3,   2,  -1,  -4,  -5,   1,   4,  -1,   2,   2,   0,   0,   0,   1 },
      {  -3,  -8, -18, -34,  12,   3,   1,   1,  10,  17,  -6, -31,  15,  -3,  -2,  -1,  -6,  13,  29, -12,  -4,   1,  -4,   1,  46, -21,  25, -27,  27,   3,  -5,   0,  10,  33,  22,  20,  20, -27,  -7,  -4,  19,  26,   1, -11, -31, -20,   2,   1,  -5,   5, -22, -21,  -7,   4,   9,   2,  -2,  -1, -12,  -1,   1,   5,   2,  -1 },
      {  10,  16,   8,  42,   0,   1,  -7,  -2,   3,  -2,  33, -13,  19,  -4,  -2,  -5,  12,  -4, -16,  38,   9,  -7,  -1,  -3,   6,  12,  19, -12,  -7, -10,  -3,  -2, -24,  14, -37,  29,   8, -10,   4,  -2,   6,  -8,  37,  35, -37,  -6,   0,   0,   2,  10,  19, -26, -30,   2,   3,   1,   0,  -4,  -1, -14,  -5,   8,   1,   0 },
      {  -9, -13, -35, -15,  -5,   4,   2,   4,  -3,  13,  27, -60,  -3,   2,   4,   2,   7, -12,  37,   3,  26,  -5,   1,   1,  -5,  21,  11,  14, -20, -14,  -2,  -1, -11, -56,   7,  -9, -17,  10,   3,   5, -14,  -1,  32, -14,   7,   9,  -1,   1,  -4,  13,  -5, -21,  12,   5,  -1,  -1,  -7,   4, -11,   1,  13,   0,   1,  -1 },
      {  -2, -10, -12, -55,  19,  13,   5,  -2,  13,  15,  38,  24,   8,   5,  -5,  -4,  -8, -27,  31,  16,  -6,  -6,  -2,  -6,  26, -33, -12,  26,  -7,  -3,   4,  -4,  12,  -1, -17,   5, -11,   1,   6,  -5,  -3, -28, -10,  24,  -3,   8,   3,  -1,   7, -11,  39,  16, -16,  -5,  -5,   0,  11,   0,  17, -10, -15,  -2,   0,   1 },
      {  -1,   7,   9,  22,  11, -11,  -3,   1,  -3,  16,   0,  21,  21, -11,  -1,   2,  -4,  12,  25,   1,  22,  -8,  -1,   1,  12,   9,  14,  23,  28, -17,  -7,  -2, -46, -15,  17,   1,   1, -37,  -2,   1, -56,   9, -23,   5,   0, -11,  21,   4, -19,  12,  -2,  34,  -5,   4,   8,  -5,  -5,   9,   7,  11, -18,  -5,  -2,  -3 },
      {   3,   6,  11,  -9, -10, -10,   5,   1,   4,  -5,  10,   2, -14, -10,  10,   3,  19,  33,  15,  19,   0,   1,   3,   4, -29,   9,  43,  23,   7,  15,  -7,   1,  56, -19,  25,   5,  33,  -1, -22,   2,   0, -12,   7,  29,  20, -36, -10,   0,   6,   7,   8,   9, -19, -16,   3,   1,  -2,   4,  -6,  -9,  -9,  -2,   3,   2 },
      {  -5,  -4,   7,  -8,   2, -21,   6,   3,  -3, -15,  -4,  -3,   0, -29,  13,   0,  -2,  14,   3,  -2, -31,   2,   7,   6, -25, -23,  33,  -5,  22,  -3,   0,   4,   0, -12, -19,  20, -21, -25,  -1,   3,  -2, -69,  -2, -17, -22,  21,  11,  10, -12, -11,   0,  -9,  30,  23,   4,  -2,  -3,   1,  -3,   9,  20,   4,  -5,  -4 },
      {  -2,  -1,  10,  -7, -42,  27,  -2,   2,  -6,  -1, -29,   2, -51,  31,  -5,   5,   2,  23,  14, -28,  15,  24,  -4,   2,  15,   2,  31,  20,  20, -19,  -6,  -3, -33,  16,  -1,  -1, -22,  -7,   8,  -2,  10, -10,  10,  15,   0,  14,  -5,  -3,  13,  -4,  29,  -9,  -9, -12, -10,   1,   6,  -2,   0, -18,   2,   3,   2,   2 },
      {   1,  -2,   0,  19, -58,  29,  -7,   8,   1,   1,   2, -22,   9,  28,   1,   5,   5,   5,  23,  52,   7,   6,  -3,  -5,  -1, -13, -20,  10,   4,   9,  -2,  -2,   4,   2,  -4,  26,  11,  -5,  -4,  -3,  13, -10, -15, -24,  -7,  -8,  -1,   2,  -6, -33, -19,  24,  17,  -3,   5,   4,   9,  -8,  23,  30, -11,  -7,   1,   0 },
      {  -8,  -4, -17, -28, -14,   4,   7,   0,  -2,   5,  17,   6, -23,   8,   8,  -2,  -9, -10, -30,  15, -45,  24,  10,  -3,  17,  47,  -1,   1,  10,  45,   1,  -5, -26,  15,  16,  13,  19,   9, -15,  -7, -44,  -8,  19, -11,  -2,  -3,  -5,   5, -21,   4,  10,  -3,   6,   3,   3,   1,  -3,  -1,   9,   2,  -1,  -2,  -2,   1 },
      {  -5,  -9,  -8,   3, -31, -39,   9,   1,  -9,  -7, -14, -27, -52, -35,  13,   2,   0, -12,  14,   5, -22, -32,  15,   2,   4,  -5, -22,  18, -29,  -2,  14,   2,  -7,   9,   0,   6,   6,  -8,   4,   3,  -4,  21, -28,  21, -20,  -4,  16,  -3,   6, -10,   0,   6, -24,  27,   5,  -5,   0,  -2,  -1, -12,  -5,  14,  -4,  -1 },
      {  -2,  -4,  -8,  -7,   3, -20,   7,   0,   5,   9,  19,  31, -19, -39,   9,   2,   2, -10,   3,  20,   8,   5,   9,  10,  16,   6,  -1,  32,  21, -15,   3,   0, -34,   9, -18,  23,   4,  12,  -4,   3,  42,   7,   1,   4,  30,  -4, -25,  -3,   7,   6, -37,  -5,  20, -26,  -4,   5,   1,  -4, -22,  20,  18,  -3,   8,   5 },
      {   7,   7,  13,  37,  50,  -4,   2,  -5,  -2,  -1, -14, -35,  11,  -3,   4,  -3,   4,  -5,  28, -29, -12,   5,  15,  -5,  14,  10,  11,  26, -15,  27,  12,  -7, -23,   7,  26,  17,   8,  31, -10,  -9,  22, -21,  -6,   6,   3,   1, -19,   0,  -3, -35,  16,   6,   2,  -6,  -6,   6,   8,   1,  20,   4,   0,  -7,  -4,   1 },
      {   3,   4,  -2,  10, -22,  10,   5,   6,   1,   2,  -1, -23,  24, -30,   9,  -2,   3,   5,   3,  -2, -38, -11,   7,   3,   5,   4,  12,  -8,  13, -11,  11,   0,  19,  18,   5,  29, -27,   9,  16,  -2,  -1,  23,  18,  -1,  28,  39,  -1,  -8,   7,  40,  10,  34,  11, -15, -19, -10,  13,  10,  10, -11, -30, -14,  -6,   1 },
      {   0,  -4,  -7, -14,  -4,  15, -12,   3,   1,  -4,  14,   3,   6, -13, -15,  -4,  -1, -10, -29,  -7, -25,   6,  -8,   5,  -2,  25,  26, -22,   1,  -6,  -6,   4, -25, -53,  23,   6, -15, -10,   1,   3,  37,   2, -22,  -4,  -5,  -8,   9,   3,  37, -25, -18,  20, -29,   0,   8,   0,  17, -17,  12,  -7, -21,  12,   1,  -1 },
      {   7,   7,   1,  12,  48,   4,  17,  -6,   2,  -5,   4, -11, -35,  27,  19,  -4,   8,   9,   4,  23, -10,  43,   8,  -6,  -1,  -4,   4,  13,  19,  22,  -8,  -5,  14, -23, -33,   2,  -7, -13,   9,  -1,  -3,  39, -13,   3, -17,  17,  12,  -6,  29,  16, -14,   5,   7,  19, -11,  -8,  20,   1,   2,   9,  -3,  -8, -11,  -2 },
      {  -2,  -4,  -3, -13, -20, -32,  10,   0,   6,   7,  14,  30,   7,  19,   7,   5,   0,   3, -10, -17,  40,  -1,  12,   2,   6,   8,  33, -17, -21,   6,  17,   5,   8,  -6,  20,  29,  18,  13,  24,  -7,   4,  14,   5,  25, -17,  31,   8, -19,   1, -20,  -3,   7,  19,  25, -20, -12,  -8,  -5,  15,  28,   4, -16, -17,  -3 },
      {   1,  -2,   1,   3, -33, -19,   8,   1,   7,   8,   0,   5,  44,  -9,  16,  -2,   7,  -3,  10,   6,  15,   8,  15,  -6,   5,  11, -20,   7,  21,  38,  -7,  -2, -15,  -3,  12, -32,  15, -14, -17,   3,  27,  -5, -14,  -7,  -1,  13,  11,   8,  43,  16,  33, -12,  12,  18, -11,  -6,  15,   9, -13, -29,  18,  -1, -13,   0 },
      {   2,   2,   6,   2, -12,  49,  -3,  12,  -2,  -3,  -7,   7,  30,   4, -19,   4,  -2,  -3,  -2, -20, -34, -30,   4,   3,   2,   2,  18,  42,  -8,   9,   7,   3,   6, -16, -24,  15,  27,  -7,   3,  -3, -25,  19, -24,  14,   0,  10,   0,  -9,  14,  -6, -13, -26,   7,   3, -17,   4, -11, -16, -25,   1,  33,   1,   4,   7 },
      {  -4,  -5,  -4,   2, -23, -22,   3,  -7,  -1,   0,  -2,   1,  30,  -9,  16,  -7,  -1,   4,   9,  -5,  10,  42,  14,  -5,  -2, -15,   7,   0,  13,  23,   0, -11,   5, -11, -23,   1, -33,  28,  -2,  -8, -19,  14, -19,  -4,  -5,   8, -24,  -9, -24, -17, -20,  -6, -41, -12, -10,   5, -34, -27,  -2, -28,  -8,  17,   5,  11 },
      {   2,   3,   0,   8,  15,  16,  25,  -1,   0,   1,  10,  -9,  -5, -12,  18,   2,   0,  -2,  -2,  30,  -8,  -2,   1,   4,  -4,   7, -16, -22,  36, -29, -15,  11,   6,   1,  35, -14,   6, -17,  24,   9, -11,  15,   2,  20,  10,  19, -12, -15,  -5, -60,  15,  -4,   9,  -5, -20,   6, -19,  -4,   4, -21,  30,   5,   7,   4 },
      {   4,   3,   4,   6,  -4,  30,  57,   5,   4,   7,   2,   9,  10,  27,  65,  -1,  -2,  -7,  -4, -10,   1,  -6,  23,  -3,   4,  -2,  -3, -15, -18, -23,  15,  -2,  -6,   3,  15,  10, -11,   5,  -7,  -3, -10, -28, -13,   7,   2, -18, -10,  -3,   9,  15, -26,   1, -13,   7,  14,  -1,  19,  11, -23,  -3,   0,  15,   2,  -1 },
      {   1,   2,   3,  -1,  -5,   8,  25,   4,   1,   4,   5,   2,  14,  17,  11,   4,   0,  -6,   0, -10, -28,  22,  -9,   6,  -6,   6,  -3,  18, -22,  -8, -34,   6,  -6,  -4,  -4,  -2,  -1, -16, -15,  17,  53,  18,  -2,  19,   8,   6,  19,  10, -54,  11,  15,   9,   1,  16,  10, -13, -48,   8,  -2,   9,  -5,  -2,  -5,  -4 },
      {   2,   3,   4,  -2,  10,  53, -19,  18,  -1,   2,  -2,   9, -16, -29,   1,   6,   3,   5,   3,  15,  25,  -9,  17,   4,  -1,   5,   1,  -3,  32,  17,  33,   3,   1,  -6,   4, -10,   9,  34,  15,  -6,  23,  -6,  -9,  -2,   1,   9,  17,  -9, -41,  17,  -3,  -2, -20,  36,  11, -18,  -1,   8,   8,  -8,   7,  18, -17, -10 },
      {  -1,  -2,  -3,   7,   2, -28,  -9, -22,   2,   2,   5,  -9,   0,  43, -12, -16,  -2,  -1,   1,  18, -10,  -8, -60,   9,   4,   3,   3,  -1,  18, -14,   9,  15,  -2,   0,  12,  31,   0,  34,  -4,   0,  -1, -10, -34,   5,  16,   3,  -4,   5,  -4,  20,  -3,  12,   2,  21,   2,  -8,   0,  -1, -12, -34,  32,  11, -16,  -2 },
      {   0,   0,   1,   0,   1,  10, -41,  12,   1,   1,   0,  -2,  13, -21,  -3,  19,   2,   0,  12,   0,   0,  63,   7,  12,   0,  -5, -15,  -1, -31,  -9, -18,   9,   6,  25,  20,  15,  -6,   2,   2,  12, -23,  -9,   7,  35,  19,   7,  15,   5,  26,   4, -18,   6,  -2,  18,  23,  -2,   7,  -7,  -1,   7,  30,  19,  -4, -11 },
      {  -1,  -1,  -2,  -2,  -7,  11, -10,  -2,  -1,   1,   1,   1,  -7, -16, -17,   1,   2,  -2,  -5,   6,   9,   9,   1,   0,  -5,   5,  -3, -18, -16,  12, -21,   2,   2, -13,  12,  17, -20,  -5,  -5, -10,   0,   3, -49,  23, -27,   9, -32,  -6, -24,  11,   4, -16,  11, -11, -10,  25,  19,  72,  -9,  -2,   4, -16,  12,   8 },
      {   0,  -1,   2,   2,  13,  10,  -8, -16,  -2,  -2,  -2,  -5, -16,  15, -23,  -4,   1,  -3,   6,  15,  10, -18,  22,   4,  -2,   0, -16, -13,  -5,  30,  -8,   3,  -8,   8,  27,   2,  -9, -22, -11,   6,   6, -38,  -3,  23,   3,  30,  11, -10,   1,  16, -34,   3,   1, -11, -30, -12, -29, -31, -41,  -8, -32,  -9,   1,   8 },
      {   1,   1,   3,  -2,  -6,  31,  22, -19,  -2,  -2,  -4,   5, -13, -38,   6, -42,  -2,  -1,   3,  -8,  22, -20, -35, -25,   1,  -8,   4,  -9,  -1,  13, -41, -14,  -1,   5,  -2,   3,  -9,  16, -45,   0, -10,   7,  12,  14,  -3,   8,  -7,  15,  15,   1,  -9,  11,   6,  25,  -7,   7,  -7, -15,  25,  15,  12,   1, -15,   7 },
      {   0,  -1,   1,   1, -11, -13,  20,   8,   0,  -2,  -5,  -2,  16,  -4,  -7,   3,   1,  -2,   5,  -7, -14,  -1, -40,  -3,  -2,   4,  -6,  19,  20,  15, -28,   9,   1,   4,   6, -31,   1,  23,  31,   7,  -2, -12,  21,  26, -15,  13,   7, -15, -24, -14, -26,  -8, -18,   6,   3,  -6,  57,  -1, -30,  24, -19,  -3,   3,   1 },
      {   2,   3,   1,   1,  16,  -9,  10,  72,   1,   2,   3,   4, -11,   2,   5,  47,  -1,  -1,  -4,  -1,  19, -13, -18,   9,   1,  -4,   4,  -7,  -9,   6, -37,  -1,  -2,   2, -10,  23,  17,   7, -23,   0,  -7,   2,  -9, -21,  14,  26,  -7,   7,  -1,  -7,  -1,  -7,   6,  11,  -4,   6,  11, -12,   3, -36, -28,  13,   8,   5 },
      {  -2,  -3,  -3,  -4,  -1,  -1, -24, -26,  -1,   0,   0,   4,   1,  -1,   5, -13,   0,  -2,  -5,  -6,  -2,  14,  22, -12,  -6,  -1,  -1,   2, -16, -39,  -8, -25,   2,  -5,  -5,   4,  27, -21, -26,  -4,   3,   3,   5, -13,  17,   2,   2, -13, -44,  -8,  -8,  -5,  -5,  14, -36,  -8,  58, -17,  -2, -25,   3,  -7, -28,  15 },
      {   1,   1,   1,  -1,   1,  -9,  -3,  35,   2,   2,   4,  -1,   4,  15, -14,  22,   0,   1,   1,   4,  -7, -15,   4,  16,   1,  -2,   0, -11,   8,  17,  30, -12,  -1,  -3,  -1, -19, -52, -14, -38,  -9,   5,  19,  21,  35, -10, -28, -16,   1, -17,  -4,   1,  34,  17,  10, -14,   2,  19, -17,  -7,   4,  33,   8,  -7,  12 },
      {   1,   1,   0,   3,   2,   1,  38,  12,  -1,   0,  -2,  -1,  -6, -13, -20, -17,   0,  -1,   0,  -3,  21,   8,  -6, -25,  -1,   3,   3,  -7, -26,  35,  -3,   6,   1,   0,   1,  25,  -6, -24,  40,  22,   4,  -3,  -8, -16,  10, -19,  23,  -8, -23,  17,  12,  16,   1, -37,  -6,  -3,  22, -24,   5, -13,  40,  22,   4,  -1 },
      {  -1,  -1,  -1,  -1,  -3,  -2,  16, -48,   0,   0,   2,   3,   4,  18, -29,   3,   1,   4,   0,   0,   7, -24,  47,  18,   1,  -5,  12,  -1,   9,  17, -22,   0,   1,   1, -13,  12,  -9,   9,  -4,  15,  -2,  20,  12,   6,  25,  12,   7,  18,  -7, -24,  -3,   4,  17,  13,  41,   8,  23,  15,  -3, -17,  -7,  36,  20, -16 },
      {   0,  -2,   1,   0,   1, -17, -35,  14,   1,   2,   0,   0,  -3,  25,  19,   8,  -2,  -4,  -3,   0, -16, -29,  31, -57,   1,   2,   1,  -3,  19, -11, -40, -27,   5,  -8,  14,  -3,   1,  14,  14,  -6,  10,  -1, -12,  13,  -9,  19,  -5,   8,  -6,  29,   5,   4,  -8, -17,  16,  18,   0, -22,  10,  21,  26,  13,   9,  -8 },
      {   0,   0,   2,   2,   0,   4, -12,  19,  -1,  -1,  -1,   2,  -3,  -9,   2, -32,   0,  -1,  -2,   3,   5,  17, -11, -55,   0,   1,   3,  -3,  -3,  -2,  30,  -8,   3,  -6,   3,  10,  -5,  -8,   9,  -7,   0,  14,  -5,  14,  16,  -5, -13,  20,  -6, -22,  23,  -5,  38,  22,  26,   1,   0, -15, -55,  -6, -37,   1,  12, -16 },
      {  -1,  -2,  -1,  -3,   3,  14, -51, -25,   1,   3,   1,   5,  -4,  -3,  71,  11,   0,  -3,  -1,  -9,   9, -24, -15,  26,   3,   2,   6,  -1,  -8,  32, -21,  28,  -2,   0, -10,  10, -17,  -2,  16,  16,   0,  12,   1,  -6,  -5, -18,   6,   7, -10, -13,  16,  10,   9, -10,  -2,  -1,  10, -16, -20,  -8,   1, -22,   2,  -2 },
      {   0,   0,  -1,  -1,   0,  -5,  -6,  25,   0,  -1,   0,   1,  -2,   9,   3, -11,  -2,   0,  -5,   2,  -2, -10,   2, -19,   2,  -5,   4,  -5,   5,   8,  -3, -41,  -3,   5,  -7,  12, -16,  13,  -5,   6,  -2,   1,  11,   6,  19,  -9,  57,   8,   7, -35, -26,  -7, -15, -18,   7, -20, -10,  45,  -8, -24,  20, -48, -10,  -3 },
      {   0,   0,   0,  -4,   0,  -9, -24,  29,   0,   2,   0,   5,   3,  19,  15, -50,   0,   0,   3,  -8,  -3,  -8,   5, -13,   2,   2,   2,  16,  -1,  12,  12,  38,  -3,   4,  -5, -13,   4, -25,  12,  30,   6,  -6,  19,  12,   8,   7, -19,   2,  -3,   2, -41,   1,   5,  16,  -4,  22,  -1,  25,  42, -27,  -6,  10,  14,  21 },
      {   0,   0,   1,   1,   0,  -3,   8,  24,   0,  -1,  -1,  -2,  -3,   5,  -2, -41,   1,   2,   2,   7,   6,   4,  17,  31,  -1,  -1,  -2, -11,   6, -11,  17,  36,   4,  -3,   6,   3,  -2,  10, -33,  13,  -6,  12, -16,   2,   2,  33,  21,  34, -12,   1,  22, -23, -26, -21,  18,  14,  23, -29,  -2,  25,   9, -34,  12,   7 },
      {   0,   0,   0,   0,   4,   0, -12,  23,   0,   1,   1,   1,  -4,   4,  -6, -35,   0,   1,  -1,  -1,   1,   3,  20,   4,   1,  -1,   4,   1,  -4, -10, -22,  18,   0,   5,  -8,   4,  15,  19,  -9,   1,  -2,   1,  17, -33, -23,   5,  -1, -10,  10, -16,  18,  56, -18,   0, -22,  -7, -10,  31, -54,  16,  10,  33,  -2,  11 },
      {   0,   0,   0,   1,   2,   1,  -8,  -8,   0,   0,   1,   1,  -2,  -2,  24, -16,   1,   0,   0,  -2,   5,   5, -32,  28,  -3,   3,  -2,   5, -11,   8,  20, -41,   3,  -5,   5,  -8,  20, -18,  -7, -50,   9,   1,  11,   5,  -9,  41,  30, -12,  -7,   2,  -7,  18,   9, -21,  22,  18,   2,   4,  -1, -10,   9,  26,  45,  13 },
      {   0,   0,   0,   2,   2, -10, -13,   3,   1,   1,   0,  -3,   1,  10,   9, -11,   0,   0,   0,   4,  -1, -11, -10,   5,   1,   1,  -1,  -9,   6,  17,  -3, -13,   0,   4,   2,  17, -23, -21,  -2,  -4,   0,   0,  -8,  -7,  58,  -7,   5, -32,   6,  -5,  25, -39, -28,  27, -17, -13,   1,  14,   3,  53,  -5,  30,  18,  24 },
      {   0,  -1,  -1,  -1,  -1,  -5,  -1,  14,   0,   0,  -1,   0,  -1,   3,  -1, -25,   0,  -1,   1,  -1,  -1, -10,   6,   5,  -1,   0,  -4,   3,  -5,   7, -23,  35,  -2,  -2,  -6,  -4,  -7, -22,  -8, -64,  -1,   1,   2,   2,  11,   6, -23, -37,   0,  -7,  -4,  -9,   0, -20,  36, -37,   1,  -2,   8,  -7,   6,  -4, -36, -45 },
      {   0,   0,   0,  -1,   1,   2,   3, -20,   0,   0,   1,   0,  -3,  -2,  -1,  37,  -1,   0,  -2,   0,   5,  -4, -18, -33,   2,  -1,   4,  -1,  -6,  15,  27,   9,  -1,   0,  -6,   2,   7, -36, -15,  19,   2,   1,  15, -19,  14,  44, -37,  -1,   1,  -7,  -3,  25, -52,  11,  24,   6,   1,  12, -15,   7,  26, -27,   4, -12 },
      {   0,   0,   0,  -1,   0,   1, -11,  -6,   0,   1,   1,   1,   0,   1,  15,  12,  -1,  -1,  -1,  -1,   0,  -2, -17, -18,   0,   1,   2,   1,   1,   1,  25,   4,   4,  -3,   1,  -3,   2,  -3, -29,  33,  -6,   8,  -6,  12, -15,  23,   2,  25,  -2, -12,  -6, -19,  10, -54, -13, -48,   6,  20,   9,  14, -11,  58, -25, -12 },
      {   0,   0,   1,   0,   0,   0,   4,   6,   0,   0,   0,   1,  -1,   4,   0,  -8,   1,   1,   1,   1,   2,  -5,  -3,  29,  -1,   0,  -3,   1,  -6,  11,  -3, -26,   3,   0,   3,   0,   3, -21,  44, -27,   1,   0,   5,  -1,  17,   6, -19,  73,  -2,  -4,  -7,  -2, -21,  -2, -13,  28,  -4,   8, -11,  10,  -2,  13, -58,   0 },
      {   0,   0,   0,  -1,  -1,   0,  -5,  16,   0,   0,   0,   1,   1,   3,   0, -32,   0,   0,   1,   0,   0,  -3,  -2,  41,  -1,   0,  -2,   1,  -3,   2,  -3, -57,   0,   1,   3,  -2,   7,  -4,   7,  45,  -2,  -3,  -7,  -3, -11,   2, -39,  -8,   0,   6,   2,   3,  -5,   8, -10, -25,   3,  -7,   4,   1,   4, -10,  26, -67 },
      {   0,   0,   1,   0,   2,   3,  -2,   5,   0,   0,   0,   0,  -1,   0,  -4,  -8,   0,   0,   0,   0,   2,  -3,   5,  16,  -1,   0,  -1,  -1,  -2,   1,  -7, -29,   1,  -2,   3,   0,   5,  -3,  16,  26,  -3,   0,  -6,  -1, -12,   7, -31, -12,   5,   2,  10,   2,  11,  -5,  60, -42,  -2,  -7,  -2,  -4,   0, -10, -26,  79 },
      {   0,   0,  -1,   0,   1,  -1,  -2,   5,   0,   0,   1,   0,   0,   2,   5,  -5,   0,   0,  -1,   1,  -2,  -2,  -7,   8,   0,   0,   1,   0,   3,   3,  10, -15,  -1,   0,  -2,   1,  -3,   1, -14,  38,  -1,   1,  -1,   0,   1,   5,  16, -58,   2,  -2,   8,  -5,   7, -14,  22,  74,  -6,  -1, -11,   4, -14,  11, -60, -17 },
    },
    {
      { -97,  50,   3,   4,   0,   1,   0,   1,  62, -16, -11,  -2,  -2,  -1,  -1,   0,   0, -16,   4,   2,   1,   1,   0,   0,   0,   2,   3,   0,   0,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  53,  -5, -15,   0,  -3,   0,  -1,   0,  70, -55,  -2,   1,   0,   0,   0,   0, -59,  -2,  28,   0,   3,   0,   1,   0, -13,  27,   0,  -7,   0,  -1,   0,   0,  -5,   4,  -7,   0,   1,   0,   0,   0,  -2,   2,  -1,   2,   0,   0,   0,   0,  -2,   0,  -1,   1,   0,   0,   0,   0,  -1,   1,   1,   0,   0,   0,   0,   0 },
      {  28, -14,   1,  -1,   0,  -1,   0,   0,  39,  24, -31,  -1,  -4,   0,  -2,   0,  56, -67,  -5,  11,  -1,   3,   0,   1, -43, -14,  39,  -1,   1,   0,   1,   0,  -9,  24,   2, -12,   0,  -1,   0,   0,  -9,   3,  -7,   0,   3,   0,   0,   0,  -2,   3,  -1,   2,   1,  -1,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0 },
      { -32, -89,  51,   5,   8,   1,   3,   1,  33,   2,  27, -14,  -1,  -4,  -1,  -2,   9,  26, -16, -16,   2,  -2,   0,  -1, -26,  16, -16,   5,   5,   1,   1,   0,   0,  11,  -7,   4,  -2,   0,  -1,   0,  -1,   2,  -4,   1,  -1,  -1,   0,   0,   1,   1,  -2,   1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0 },
      { -23,  -5,  14,  -1,   1,  -1,   1,   0, -44,  15,  -5,   6,  -1,   2,   0,   1, -27, -35,  39,   7,   3,   1,   1,   0, -23,  65,  21, -27,  -2,  -6,   0,  -2,  25,  18, -46, -10,   5,   1,   1,   0,   8, -13,  -9,  16,   4,   1,   0,   0,   8,  -3,   2,   3,  -5,  -1,  -1,   0,   2,  -2,   0,  -1,   0,   1,   0,   0 },
      { -16,   2, -10,   2,  -1,   3,   0,   0, -27, -49,  43,   5,   3,   1,   2,   0,  -5,   9,  -2,  -1,   1,   0,   0,   0, -17, -24,  42,   6,  -1,  -1,   0,   0, -14,  56,  22, -35,  -1,  -4,   0,  -2,  12,  16, -45,  -9,  11,   0,   1,   0,   3,  -3,  -8,  16,   3,  -1,   0,  -1,   5,   1,  -1,   2,  -5,  -1,   0,   0 },
      {   6,  16, -15,   1,  -3,   1,  -1,   1, -10, -49,  33,   7,  -1,   3,   0,   1,  77,  -1,  32, -31,  -4,  -5,  -1,  -2, -17,  21, -35, -24,  12,  -2,   3,   0, -16,  -6, -20,  21,   9,  -1,   1,   0,  -3,   2,   7,  12,  -7,  -1,  -1,   0,  -2,   1,   5,  -5,  -4,   2,  -1,   0,  -4,   1,  -1,  -2,   3,   1,   0,   0 },
      { -13, -44,  14,  11,   0,   2,   1,   2,  -7, -25,  -9,  15,  -4,   3,  -1,   2,   7, -27,  43,  11,  -7,   3,  -3,   1,  68, -11,  23, -27,  -8,  -1,  -2,   0, -11, -22,  14, -17,  13,   0,   3,   0, -19,  31,  17, -12,   6,  -8,   0,  -3,  -8,  19, -18, -10,   0,  -3,   2,  -1,  -1,   3, -10,   4,   2,   1,   1,   0 },
      {  11,   4,   0,   0,   0,  -1,  -1,   0,  26,  11,  11, -18,   1,  -3,   0,  -1,  16,  22, -19,   5,  -5,   0,  -2,   0,  56, -10,  16, -26,  -8,  -1,  -1,   0, -15,  14, -55, -27,  19,   0,   4,   1,  -7, -32, -29,  44,  12,   1,   2,   1,  -4,  -5,  25,  14, -12,  -2,  -1,  -1,  -2,  -2,   2,  -5,  -4,   1,   0,   0 },
      {  12,  19,  28, -24,   2,  -4,   0,  -3,  11,  30, -23,   0,   0,   1,   0,   0,  12,  17,  32, -15,  -4,  -5,  -1,  -2,  14,  19, -34, -14,   0,   1,   0,   0,  -7,   5,  14, -12,  -1,  -2,  -1,  -1,  -5,  26, -53, -27,  14,   1,   5,   0,  -3, -23, -32,  41,  19,   3,   0,   0,   0,  -2,  15,  19, -13,  -5,  -3,  -1 },
      { -17, -33, -85,  34,   5,   8,   2,   3,  15,  30,  -8,  27, -10,   1,  -4,   0,   6,  29,  28, -19, -12,  -2,  -3,   0, -12,  -8,   1,  -9,   3,   3,   0,   1,  32,  -6,   3, -15,   0,  -1,  -1,   0,  -5, -13,  -9,  -6,  12,  -1,   1,   1,  -1,  -4,   1,  11,   5,  -1,   0,  -1,  -2,  -1,   3,   1,  -2,  -2,   0,   1 },
      { -15, -17, -16,  -4,   2,   7,   1,   0,  -9,   6, -48,   6,  15,   5,   3,   0, -10,  30,  14,  22,  -9,  -3,  -2,   1, -14, -15,   4, -21,   6,  -4,   2,   1, -75,  23, -14,  36,  11,  -2,   4,  -1,  30,  17,   8,  14, -21,   3,  -3,   1,  10, -12,  -4,  -4,  -4,   6,   1,   0,   4,  -5,  10,   3,   0,   2,  -2,  -1 },
      {   0, -10,  12,  -7,  -2,   0,  -1,  -1,  -8, -54, -60,  16,   9,   6,   3,   3,  37,  39, -14,  34,  -3,  -2,  -2,   0,  -1,  26,  18,  12, -24,  -1,  -4,  -1,  31, -11,   4, -11, -16,   7,  -1,   1,  11, -27,  -1, -11,   6,   5,  -2,   1,  -3, -12,  12,   7,   8,  -2,  -1,   0,  -3,   3,   8,  -4,  -4,  -3,   0,   0 },
      { -15, -32, -30, -28,  14,   3,   6,  -1, -10, -16,  -7,   9,   8,   6,   4,   0,  -7, -51,  -6,  -2,  16,  -2,   6,  -1,  29,   8, -31,  42,  10,  -5,  -1,   1, -15,  -3,  12,  15, -14,  -6,  -2,  -1,  -7,  -7, -22,  21,   4,  13,   4,   3,  14, -13,  31,  34,  -9,   1,  -8,  -3,  -4,   9,  24, -13, -16,  -3,  -6,  -2 },
      {  -3,   7,  21, -11,   1,  -3,   1,  -1, -10,  -5, -13,  19,  -5,   3,  -2,   1, -14,  15,   6,  -7,  -5,  -2,   1,  -1, -49, -26,  -1,  -5,  -9,   6,   1,   2,  -9, -49,   5, -24,  27,  11,   1,   3, -33,  24,  -2,  36,  31, -14,   1,  -2,  14,  15,  21,  26, -31, -13,   0,  -1,   5,   5,   4, -21, -13,   7,   1,   0 },
      {   7,   4,  45,   1,  -8,  -2,  -2,  -3,   8,   7,   3,  56, -20,   6,  -6,   1,  -8,  -4,  36, -17, -21,   2,  -4,  -1,   6, -39,  -5,  -8,  11,   4,   1,   1,   8,  14,  13,   6, -12,  -3,  -3,   1,  31, -34,  -9,  -5, -12,  12,   0,   5,   4, -17,  43, -14,   5,   2,  -4,   0,  -4,  16,   1, -27,   7,  -4,  -2,  -1 },
      {  -3,  -5,   9,   3,  -7,   2,  -2,   4, -14, -27, -39, -30,  13,   3,   1,   4, -12,  -6,   1, -27,   7,   6,  -1,   2,  -1, -47,  -7, -17,  38,   6,   5,   1,  53,  29, -15,  31,  21, -15,  -1,  -6, -23,  -7,  -2,   4, -11, -13,   0,  -2, -20,   3,  -6,  18,  -2,   0,   5,   0,   6,   2,   6,  11, -12,  -1,   1,   0 },
      {  -6,  -4,   3,  -3,  -9,   6,  -1,  -1,  -6,  -4,  23,   0,  10,   0,   1,  -2, -11, -12,  -2,  19, -14,   0,  -1,  -1, -18, -19,   1, -21,  -1,   0,   1,   3, -27, -22, -17,  12,  -8,  12,  -1,   3, -13, -46,  17, -25,  28,  24,   0,   8, -13,  19,  -2,  29,  54, -16,  -3,  -4,   4,  16,  14,  31, -25, -23,  -3,  -6 },
      {   6,   2,  11,  80, -32,   4,  -8,   3,  -8, -11, -15,  10, -32,   2,  -3,   2,  -6, -19, -43, -21,   5,  14,   2,   4,   7,  21,  -3, -10,  14,   1,   0,   0, -19, -17,   8,   1,  15,   2,   0,   0,  29,   4,  -8,  -1,  10,  -5,  -3,  -2,   3, -19,  -4,  19,  -2,  -5,   0,   1,   4,  -7,  18,   4, -14,  -1,   2,   0 },
      {  -2, -14, -13, -41,  11,  10,   2,   0,   3, -10,   9, -18, -10,  11,   0,   4,  -1, -22, -10,   4,  -1,   1,   0,   3,  -7, -23, -11, -18, -13,  15,   2,   4,  23, -31,  -1,  -9,  38,  16,  -1,   3,  72,   4,  -6,   8,  13, -17,  -8,  -2,   2, -31, -10, -11,   8,   0,   5,   0, -12,   3,   1,  23,  13,  -5,   1,  -3 },
      {   2,   1,   9,  41,   5,  -1,  -2,   2,  -6,   3, -10, -70,  16,  -2,   4,  -1,   5,   2,  56,  -6,  39,  -7,   4,  -2,  -6, -18,  -2,  24, -13, -14,   0,  -1, -14, -15,   6, -14, -17,  12,   4,   5,  27, -15,  -4,  -1,  13,   3,  -3,   0,  -7,  -3,  18,   1,   1,  -6,  -1,   0,  -1,  10,  -5, -16,   1,   1,   4,   1 },
      {  -3,  -1,  -5, -42,   0,   8,   2,   2,  -4,  -2, -28,  -5, -23,  12,  -1,   3,  -3,  12, -10, -71,  17,  13,   3,   4,   6,  15,  37,   6,  34, -13,   1,  -3, -25, -15,  -2,  -9, -11,  -9,   5,   0,  11, -16,   1,  -7,  20,   8,   0,   0,   5,  25,  -2,  -9,  13,  -7,  -4,  -4,  -2,   1, -30,   3,  14,  -2,   2,   1 },
      {  -8,  -5,  -6,  10,   2, -14,   1,  -2, -18, -13, -30,   0, -16,  -4,   0,   1, -15, -13, -15, -12,  -2,  -1,   4,   0, -12,  -6, -47,  -3, -29,  15,   0,   1, -24,  21,  -6, -35,   2,  15,  -3,   0, -39, -10, -13,   1,   2,   5,  -1,   2, -24, -15,   9, -35,  19,  11,   1,   5, -11,  15, -19,  16,  45,  -7,   1,  -2 },
      {   0,   1,  -8, -10,   9, -13,   2,   4,   2,   6, -11,  -5,   5, -11,   2,  -1,   2,  -4,  -5,   1, -20,  11,  -2,   3,  10,  19, -34, -12,  -3,  15,   1,  -2,  -6,  47,  25,  -6,  13,  23,  -5,  -1,  27,  -9,  25, -14,  35, -23, -16,  -4, -19,  35,  12,   8, -12, -37,  -1,   0,  22,   1, -20, -29, -24,   7,  12,   3 },
      {   7,   7,  10,  20,  53, -18,   8,  -5,   3,  -6,  -8,  45,  -7,  -9,   0,  -3,   3,  -2,   0,  -5,  19,  -3,   2,   0,   4, -24, -14,  36,  -3, -16,   6,  -5,   8,   9, -51,  -2, -17,  11,   7,   3,  27,  22,   6,  14,  13,   3,  -5,  -2,  -7,  23, -21,   5,   1,  -7,   1,   0,   9,  -5, -12,  34, -12,  -5,   2,   1 },
      {   0,   5,   4,   9,  48, -24,   2,  -2,  -2,   7,   5,  12,  19,  -8,   1,  -2,  -2,   3,  -3, -14,   7,  -9,   4,   0,  -4,  23,  27,  -5,  -6,   7,   0,  -2, -13, -16,  37,  12,  28, -18,  -7,  -2,   5, -22, -13,  14, -26, -22,  12,  -1, -58,   4,  11,   1,  -6,  15,  12,  -3,  -5,  34,   0,  21, -16,  -6,  -3,  -5 },
      {  -4,  -4,  -4, -18, -60,  10, -11,   4,   5,  14,   8,  23, -27,  29,  -4,   5,   5,  12,   9,  27,  63,  -3,  -4,  -1,  -2,   1, -12,   8,  -3, -30,  -5,   1,   2,  14,   7,   2,  11,   2,   3,  -2,   0,   1,   2,  17,   0, -10,  -5,  -3, -40,   8,   2,  -1,  -7,  -9,   8,   3,   6,  18,  -8,  19,  -8,  -5,   4,  -4 },
      {  -6, -10,  -8, -16, -14, -15,  -2,   1,  -4,  -7,   0,   3,  -4, -23,   2,  -2,  -7, -14, -10,   3,  -9,  -2,  -3,   1, -10, -15, -12, -20,   0, -25,   5,  -1,  -8, -28, -18, -15, -35,  -5,  14,   3,  16, -14, -17, -24, -14,   1,  14,  -4, -53,   4, -20,  12, -31,  25,  11,   0,  17, -28,  -1, -30,  -1,  35,  -1,   5 },
      {   1,  -3,  -2,   7, -17,  -8,  -5,  -2,   6,   3,  -6,  22,  64,  -5,   2,   0,  -3, -11, -24, -23,  35, -37,   2,  -4,   2,  -4, -10, -48, -25, -15,   9,   2,  12,  -1,  19,   0,  -7,   6,  16,   0,  12,   0, -13,   8, -10,  19,   2,  -2,  34,  18,  -3,  -2,   5,  -5, -12,  -3,  -2,   1, -26,  -1,   0,  -3,   3,   1 },
      {  -1,  -1,   4,  -8, -25,  17, -10,   9,   2,   6,   6,  23,  46,  -6,  -7,   4,  -2,  -6,   7, -42,  -1, -17,   2,   0,   7,  -1,  24,  32, -25,  17,   9,  -6, -11,   2, -19,  11,   2,  29, -14,  -5, -11,  -7,  16, -13,  15, -24, -23,  -1, -23, -41,  -8,  -2, -14, -12,   8,  12,   9, -20,  20,   4,   9,   5,   9,   5 },
      {  -2,  -2, -11, -12,  -5, -23,   8,   4,  -1,   2,  -1, -20, -33,  -2,  -6,   2,  -1,   3,  -9, -28,  -6,  -9,   5,  -4,   3,   1,  21, -17, -38,   2,  10,  -3,  12,   4, -11,  11, -17,  23,   9,  -2,   1,  52,  14,  -3, -22,  33,  -9,  -2, -19, -12,  32,   1,  12,   4, -16,   5,  22,  26,  17,  -9, -34, -13,   3,   3 },
      {   0,   2,  -1,   7, -20,  25,   1,   4,   1,   6,   0,   5,  42, -15,  -5,  -5,   0,   1,  -8,  14, -20,   1,   4,   1,  -2,  10,  -1,  12,  36,   9,  -3,  -3,   8, -11, -22,  -3,   0, -14,   1,  -1,  19,  51, -14,  -1,  19,  18,   6,  -4, -44,  11,  26,   2,  13, -10, -12,  -3, -25,  29, -11, -16,  30, -12,  -2,   2 },
      {  -2,  -6,  -4,   1, -50, -21,  -3,  -2,   0,  -1,  -8,  -8,  -3,   1,   3,   1,   2,  -5,  13,  -9, -35, -11,   5,  -4,   2,   0,  -1,  45, -33,  30,  -1,   0,  -1,  -7, -16,  15,  22, -10, -12,   0,  18,  -6, -33,   0, -19,  -4,  14,   6,  15,  51,  -3,  -1,   2,  15,   6,  -7,   7,  -3, -23,  13, -15,  -2,  -8,  -4 },
      {   5,   6,   8,   7,  25,  86, -17,  12,   0,   1,  -4,  -6,  -7,  32, -18,  11,  -1,  -6,  -7,  -5, -18, -34,  -3,   5,   1,   5, -11,   0, -18,  -7,  16,   7,   1,   8,   4,  -7,  -4,  -2,   6,   1,   3,   0,   2,   1,  -2,  -1,   9,  -2,  -6,  25,   9,   4,  -1,  29,   9,  -5,  39,   6,   3,   2,   4,   6, -14,  -1 },
      {   1,  -2,  -2,  -5,  -7, -12,  16,   8,   4,   0,   1,   8,  22, -14,  -5,   2,   6,  -1,   2,   8,  19,  11,  -1,   0,  -1,   2,  -8,   6,  28,  17,  -4,  -5,  -5, -12, -12, -38,  15, -27,   1,   2,   1,   4,  16, -17, -13, -13,   4,   5,  19, -25,  25,  -2,  20,  21,   3,   5,  81,   7, -18,   7,  -8,  -8,   4,  -1 },
      {  -1,   2,  -1,  -2,  -5,   6,   2,  -6,  -3,   0,  -2,  -1,   8,  17,  -8,   6,  -5,  -2,  -8,  -9, -10, -19, -12,   6,  -3,  -2, -11,   6,  -3, -15,   1,   2,  -9,   8, -24, -35,  -3, -31,  12,   6,   0,   6,  26, -31, -12, -31,   3,   1,  13, -15,  12,  -6,  -1,   5,  19,  -7, -60,  20,  -9,  -1, -56,   8,  18,  -5 },
      {   0,   0,  -5,   9,  19, -18,  -3,  -2,   4,   3,   8,   5,  14,  55,   9,   6,  -1,  -1, -13,   1,  14,  37, -24,   6,   0,  -2,   4,  -5, -18, -14, -11,   2,  -1,  -7, -26,  33,   8,  15,   6,   2,  -9,  12, -32, -44,   5, -31,  -2,   3,   9,  10,  16,   0,  12,   2,  12,   2,  11,  19,   1, -29,  22,  13,  -1,  -4 },
      {   0,  -2,  -3,   1,   8,  15, -11,  14,  -3,  -5,  -1,  -6, -17, -26, -19,   0,  -1,  -4,  -2,  17, -22, -19,  -3,  -1,  -1,   1,   4,   1,  11, -18,   7,  -1,  -2, -20,  -1,  27, -14,   2,   8,  -9, -14,  -5, -46,   0, -12, -14, -30,   1,  11, -21,   0, -15,  -4, -46,   9,  22,  17,  21, -44,  10, -13,   3,  37,  -1 },
      {   0,  -1,   3,   5,  -3, -12,  -8,   1,   2,   3,   4,  -4,  26,  43,  19,   4,  -2,  -4,   5,  -7, -47,  28, -14,   4,  -4,   4,  -4,  20,   2, -41,  -5,   2,   5,  -1,  18,  -8,  19,   6,  42,  -1,  -5,  -3,   4,  22,   0,  29,  -9, -16, -17, -23,   2,  -8,   0,   6, -19,   0,  10, -28, -25,  19, -16,  -4,   6,   5 },
      {  -1,   0,   2,   2,  18,   6,   2,  15,  -2,   0,   2,   7,  -9,  17, -11,   8,  -2,   1,   1,  17,  33, -15,  12,  -6,  -2,   1,   2,  -1,  12,  39,  20,  -5,  -4,  -8,  -3,   7,  24,   6,   2,  -4,  -9,   2, -20, -19,  -3,  13, -13,   0, -19,   5,  23, -13,  24,  11, -25,   2, -21, -68, -20, -16, -29,   7,  15,  13 },
      {   1,   1,   1,  -1,   3,  46,  -4,   1,   0,   1,  -1,   6,  15,  -4,  15,  -6,   1,  -1,   2,  -8,  14,  73,  12,  -3,  -5,  -6,  -7, -14, -25,  32, -32,  -3,  -1,  -5,  -4,  -2, -26,  -5, -22,   6,  -1,   1, -11,   7, -12,  20, -10,   5,  -4,  -1, -12, -12, -24,   7,   3,   0,  -4,   9, -19,   4, -37,  -6,  11,   0 },
      {   0,   0,  -1,   8,   2, -14,  24, -23,   1,  -1,   1,  -4,   1,  66,   3, -12,  -1,  -3,  -6,   3,  -6, -39,  10,  -2,   0,  -2,   0,  -2,  10,  25, -11,   5,  -3,  -6,  -8,  -6, -15, -17, -36,  -2,   0,  -5,   5,  14,  -1,  14,  -5,   7, -15, -20, -45,   3,  -7, -22, -10,   1,  18,  17, -21, -31,  -3,  -7,  -2,  10 },
      {  -1,  -1,   0,  -2,   6, -12,  -6,  19,  -2,  -3,  -1,  -6,   2,   8, -44,   5,  -2,  -4,   0,  -8,  -2,   4, -29,  -1,  -4,  -3,  -4, -10, -29,  -7, -32,   6,   0,   0,   5,   7, -20, -49, -18,   9,   7,  11,   2,  11,   8,  -6,  34,  10, -10,   0,  20,  -7,   5, -42,   7,  -3,   6, -48,  11,  12,  10, -24,   9,   0 },
      {   2,   2,   2,   2,   3,  29,  40, -24,  -1,  -2,  -2,   7, -13, -35,  26,  -3,   0,  -1,  -1,  -9,   9, -14,   6,  11,  -1,  -3,   2,  11, -30, -22, -38,   9,   4,   6,  -4,  24,  44, -24,  15,   1,  -7,  -3,   1, -10,  19,  17,  21,   0,  -7, -19,  -8, -11,  21, -16,  -4, -10,   9,  -3,  -9, -26,  -8,  25,  -6,  -9 },
      {   0,  -1,   0,   1,   0, -14, -32,  28,   0,  -1,   4,   2,  -8,  -4, -24,  -2,   1,   1,   2,   7,   2,   9,  12,  -3,   2,  -2,   7,  -1, -17,  11,  11,   0,   1,   9,   2,  30, -10,  -8,   3,   5,  -2,   4,  24,  17,  34,  -3,  21,  -3,   3, -36,  -9,  39,  16,  34,   6, -20,  -9,  11, -60, -12,   6,  19,  -5,  -8 },
      {  -2,  -3,  -4,  -6,  17,  -7, -88, -10,   0,  -1,  -2,   1,  -4,  -8,  -6,   2,   1,   1,  -2,  -1,   4,  -9,  26,   6,  -1,  -3,  -3,   8,   7, -17, -31,  -8,  -1,   0, -13, -13,  38,   7, -18,  -1,  -1,  -1,   6, -27, -19,  27,  10,   0,   4,  -3,  -8,   2, -12, -12,  -4,  -2,   4,   2,   1, -22,  -1, -27,  -2,  -2 },
      {  -1,  -2,   0,   0,  -1,  11, -23, -46,   0,   2,   2,  -1, -11,   3,  56,   0,  -1,  -2,   1,  -2,  -4, -11,  22,   0,   0,   0,   4,  -9,  -3,  15, -15,  -5,   0,  -2,   1,  -1, -30,  -3,  14,  11,   0,  13,   9,   7, -27, -41,  -6,   5,  -8,  -4,  32,  36,  19, -22,  -1,  -5,   1, -27, -22,  17,  18,   3,  -4,  -3 },
      {   1,   2,   1,   0,   4,   7,  22,  16,  -1,   0,   2,   1,   1,   7, -35,  11,   1,   0,   3,  -3,   6,   9, -12,   3,   1,   2,   0,  10,   3,  27, -25,   1,  -3,   2,  -1, -17,  13,  34,  30,  -7,   4,  -8,  14,  -4, -56,  24,  22,  -3,   1,   3,  -8,  44,   1, -35,   8,   4, -11,   8,  -8,   6,  21,  36,   8,  -5 },
      {   1,   1,   0,   4,  -1,   0,  31,  -5,   1,   0,   1,  -4,  -2,   6,  13,   6,   0,   0,  -2,   5,  -2, -16,   0,   8,   2,   1,   5, -10,  15,  -9,  -2,   2,   1,   2,   0,  12,   2,  36, -29,  -2,  -4,   6,   0, -34,  14,  24,  28,   8,   3, -14,  34,   6, -55,   6,  33, -19,  -2, -15, -33,  32,   2, -25,  -6, -17 },
      {   1,   1,   1,   0,   1, -11,  -1,  56,   2,   2,   0,   1,   5,  24,  22,  -1,   2,   2,  -1,   1,  -8,  -2,  78, -21,   0,  -2,  -3,  -6,   2,  -5,  -2,  -8,   2,   1,  -2,  -4,  -5, -13,   5,   0,   2,   2,  -3,  -9,   9,   4,  16,  11,   4,   0,  10, -23,  -6, -27,   5,  -3,   4,   7,  21,  22,  -1,  45,  -5, -11 },
      {   1,   1,   1,   0,   1,   6,  13,  20,  -1,  -1,  -3,   1,   3,  -3,   6, -31,  -1,  -2,  -3,  -7,  -8,   5,  -4, -36,   1,   0,  -1,   1,  -3, -46,  -4,   5,   4,  -1,   4, -13,  17,   9, -60,  -1,   4,   8,   8,  12, -25,   4, -34,  29,  -3,   7,   7,  25,  33,  13,  22,  13,  -9, -16, -11, -13,   6,  13,   5, -19 },
      {  -1,  -2,  -2,   0,  -3, -10, -24, -37,   0,   0,  -2,  -3,  11,   1, -15,  18,   0,   0,   0,  -1,  -1, -13, -14,   2,   2,   4,   3,  -2,  30,  14, -33,   3,   5,   0,   8,   0, -16,  33,  -7,  10,   0,  12,  -9,  27,  21,  13,   9,  18,  -3,   1,   5, -34,  28,  14,  38,  -2,   9,  -2,  13,   6, -25,  50,   8, -19 },
      {  -1,  -1,  -1,   1,  -2,  -6,   6, -63,   0,   0,   2,  -3,   6,   7, -50,   3,   2,   2,   2,   3,  -1,  22,  42,  15,   0,   1,  -2,  -3, -16, -11,  29,   6,   1,   1,   4,   4,  12, -33,  -7,   2,   2,   0,   5, -14,   7,  25, -33,   0,   2,   5,  17,  15, -13,   0,   9,  16,   1,  -2,   1,  20,  17,  28,   9,   6 },
      {   1,   1,   2,   1,   5, -13, -13,  29,   1,   1,   2,   1,  -9,   0,  34,  12,  -1,  -1,   0,  -1,   5, -28, -29,   2,   0,   1,   2,  -3,  -3,  26, -36,  22,   0,   1,  -1,   5,   2, -36,  12,   5,   4,   0,   8, -11,  19,  34, -52, -10,   3,  11,   8,  18, -21,  21,  17,  23,  -2,   7,   7,  14,  19,  12,  25,   1 },
      {   0,   0,   0,   0,  -2,   5,   6, -12,  -2,  -2,  -2,  -4,  -3,   3, -33, -41,  -2,  -2,  -2,   0,  -4, -14,  -2, -54,   1,  -2,   1,  -4,   6,  -4, -51, -32,   0,   4,   1,  10,   3,  11,  16,  35,   0,  -5,  -1, -16,  12, -17,  -8,   6,   2,   1,   8,   1, -21,  22, -42,   6,  -1,   7,  -3,  21,   1,   7,  -9,  23 },
      {   0,   1,   1,   0,  -5,  -4,  23,  17,   0,   0,   2,   1,   2,   2, -17,   7,   0,   1,   1,   1,  -9,  -9,  46,  68,   0,  -3,   0,  -4,   3, -19, -51,   8,   1,   1,   2,   3, -12,  20, -11, -13,   1,  -2,  -1,  10,  -1, -26,  -4, -13,   3,  11,   0,   9,  12,  35, -16,  13,  -5,  -5,   3,  -3,   2, -16,  31,  11 },
      {  -1,   0,   1,  -2,   0,  -2, -19,   5,  -1,  -1,  -2,   1,   0,   1,  -2, -84,   1,   0,   2,  -2,   6,   0, -13,  33,   0,   2,   2,   1,   3,   9,   5,  70,   0,  -2,  -3,  -4,  -3,   6,  -6, -12,  -1,   3,  -2,  -7,  -3,  -2,   4,  -9,   0,  -3,  13,  -9,  -3,  -9, -15, -12,   3,  -1,   0,  18,  -8,  24, -18,  10 },
      {   0,   0,   0,   2,   1,  -3,  -5,   9,   1,   1,   1,  -2,   1, -10,   4,  55,  -1,   0,   0,   2,  -1,   0, -25,  16,  -2,  -2,  -2,  -3,  -1, -14,  -7,  -7,  -2,  -3,  -3,  -5,  -4,  -4, -52, -10,   0,   1,   0,  -5,  -6,  -4, -15, -10,   1,  -5,   9,   1,  -2, -10, -53, -15,  -3,  10, -12,  23,  -6,  44, -47,  24 },
      {   0,   0,   0,   2,   0,  -5,  11,   2,   0,   0,  -1,   0,   4,  -5, -29,  23,   0,   0,  -1,  -1,  -1,   1,  23, -29,   0,   1,   1,   0,   6,   1, -24,  46,  -2,  -2,  -3,  -2,  -8,   3,  19,   1,  -3,  -2,  -7,  -2,  -7, -12, -48, -35,  -3,   0,  -4,  -6,  -1,  -4,  12, -38,   0, -10,  -2, -13,  -5, -23, -48, -42 },
      {   0,   0,   0,   0,   1,   1,  -3, -11,   0,  -1,   0,  -2,  -1,   6,  -3, -25,   0,  -1,   0,   0,   1,  -3,  -7, -13,  -2,  -3,  -3,  -5,  -5,   5, -17, -46,  -2,  -3,  -1,  -1,  -8,  -9,   1,-100,   4,   0,   3,  -1,  11,   0,   1, -41,   5,   2,  10,  -1,   5,   4,  14,  13,  -2,  -6,  -1,   4,  -4,   6, -15, -15 },
      {   0,   0,   0,   0,   0,  -2,   0,  11,   0,  -1,   0,  -2,   0,   3,  -8, -25,   0,   1,   1,   0,   1,  -6, -13,  56,  -1,  -1,  -2,  -1,  -2,   7,   3, -55,   2,  -1,   3,  -2,   0,  -9,  14,  12,  -2,   0,  -3,   0,  -3,   5, -36,  41,  -4,   1,  -4,  -2,  -9,  -4, -17, -30,   1,  -4,   3,  -1,   2,   3, -15, -63 },
      {   0,   0,   0,   0,   0,  -2,   0,   7,   0,   0,   1,  -1,   2,  -4,  -6,   9,   0,   0,   0,   0,   1,   2,   7,  10,   0,   1,   0,  -1,  -1,   0, -12,  12,   1,   0,   2,   0,   0,   1,  32, -30,   2,   2,   3,   2,   6,  -2, -13,  77,  -3,  -2,  -1,   0,  -3,   0,  25,  33,  -8,  -3, -12,  -5,  -9, -15, -58,  45 },
      {   0,   0,   0,  -1,  -1,  -1,   5,  13,   0,   0,   0,   1,  -1,   0,   1, -21,   0,  -1,   0,   1,   1,  -3,   3,  21,  -1,  -1,  -1,  -1,   0,  -2,  -3, -42,   0,  -1,   3,   1,   4,  -1,  -3,  32,  -4,   1,  -6,   5,  -8,  12, -21, -39,   4,  -1,   5,  -3,  17, -12,  59, -40,   1,  -2,   2,  -4,  -1,   0, -20,  65 },
      {   0,   0,  -1,  -1,   0,   0,   2,   7,   0,   0,  -1,   0,   2,   2,   1,  -7,   0,   0,   0,  -1,   2,  -1,   1,  20,   0,  -1,   1,  -2,   2,   3,   4, -10,   0,   1,   0,   1,  -2,   1,  -7,  45,   0,   0,   0,   1,   1,   2,  14, -41,   2,  -1,   2,  -2,   4,  -7,   8,  88,  -7,  -6, -12,  -4, -13,  -1, -49, -30 },
    },
  },
  { // 9
    {
      { 118, -31, -10,  -6,  -3,  -2,  -1,  -1, -28,  -6,   3,   2,   1,   1,   1,   0, -16,   2,   5,   1,   1,   0,   0,   0,  -9,   4,   3,  -1,   0,   0,   0,   0,  -5,   4,   0,  -1,   1,   0,   0,   0,  -2,   1,  -1,   0,   0,   0,   0,   0,  -1,   1,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
      {  17,  13, -12,   0,  -4,   0,  -1,   0, 102, -33, -14,  -1,  -3,   0,  -1,   0, -26, -36,  14,   4,   3,   1,   1,   0, -32,   9,  18,  -2,   1,  -1,   1,   0, -11,  16,   0,  -5,   0,  -1,   0,   0,  -4,   6,  -4,   0,   1,   0,   0,   0,  -2,   1,  -2,   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 },
      {  19,  -1,  -5,  -1,  -1,   0,   0,   0,  22,  27, -16,  -1,  -3,   0,  -1,   0,  77, -18, -30,   3,  -4,   1,  -1,   1,  -3, -62,   6,  14,   1,   3,   0,   1, -27, -13,  36,   2,   0,   0,   0,   0, -12,  18,  13, -12,  -1,  -1,   0,   0,  -6,   8,  -5,  -6,   2,   0,   0,   0,  -3,   2,  -3,   0,   1,   0,   0,   0 },
      { -25, -67,  25,   5,   8,   3,   2,   0, -20, -29,  29,   1,   4,   1,   0,   0,  25, -16,  17,  -7,  -1,  -2,  -1,   0, -36,   1,  27,  -7,   4,  -1,   1,   0, -16,  49,  10, -14,   3,  -3,   0,  -1,   3,  30, -25,  -9,   1,  -2,   0,  -1,   3,   2, -20,   6,   2,   0,   0,   0,   3,  -1,  -3,   7,  -1,   0,   0,   0 },
      {   0, -55,  11,   6,   5,   3,   1,   2,  39, -43,  20,   1,   1,   0,  -1,   0,  28,  39,  10, -16,  -6,  -4,  -1,  -1,  30,  14, -38, -12,  -1,   0,   1,   0, -13, -40, -13,  18,   7,   4,   1,   1, -17,  -8,  27,  10,  -1,  -1,  -1,   0,  -5,  12,  10, -10,  -2,  -2,   0,   0,  -4,   5,  -4,  -5,   2,   0,   1,   0 },
      {  10,  -9,  11, -10,   2,  -3,   0,  -1,  23,  44,  -6, -15,  -1,  -4,  -1,  -1,  48,   6, -22, -12,   1,  -3,   0,  -1, -21,  28, -12,  -8,   5,  -2,   1,  -1,   6,  25, -52,  -1,   2,   0,   1,   0,  -1, -35, -36,  29,   1,   4,   1,   1,  -1, -24,  16,  26,  -7,   1,  -1,   1,   1,  -5,  18,   0,  -7,   1,   0,   0 },
      {   0,  -1, -25,  16,   0,   3,  -1,   0, -20, -28, -18,  23,  -2,   5,  -1,   2,  40, -56,  28,  13,  -5,   2,  -2,   2,  43,  12,  43, -16,  -5,  -3,  -2,  -1,  20,  -6, -17, -22,   0,  -2,   0,   0, -16, -30,   1,  11,   8,   1,   2,   1,  -9,   7,  24,  -1,  -1,  -2,   0,   0,  -5,   9,   1, -15,   1,  -1,   0,  -1 },
      { -16, -46,   1,   5,   4,   2,   3,   2,  -9,  12, -16,   6,   2,   1,   2,   0, -20, -29, -25,  18,   7,   4,   1,   0, -52,   0,   0,  29,   2,   3,   1,   0,  10,  -7,  -1,  25, -11,   2,  -2,   1,   9, -22,  42,  18, -13,   2,  -3,   0,  14,  14,  43, -21, -10,   0,  -2,   0,   1,   8,  -3, -32,   4,   1,   0,   0 },
      {  15,  25,  81, -24,  -4,  -6,  -1,  -2,  -2, -10,  39, -33,  -1,  -3,   1,  -1,   0, -55, -14, -14,  13,   2,   3,   0,  20,   0,  -7,   7,  10,  -2,   0,  -1,  18,  -1, -10,  10,  -2,  -2,  -1,  -1,  -9,  -3,   9,   1,  -5,   0,   0,   0,  -7,  14,  -3, -13,   0,  -1,   1,   0,  -6,   5, -14,  -3,   6,   0,   1,   0 },
      {  -1, -41,   7,   5,   1,   3,   2,   2,  18, -19,  -9,   0,   3,   0,   1,   0, -17,  -3, -42,  15,   5,   2,   1,   0,  44, -14,  -2,  36,  -4,   1,  -2,   0,  46,  10,  27,  -4, -20,  -2,  -4,   0,   8, -20, -23, -25,  -5,   7,   1,   1, -13, -31,  -1,  21,   9,   5,   0,   1,   3,   3,  29,  16, -11,  -5,  -2,  -1 },
      {   2,  39,   2,   4,  -1,  -1,  -1,   0, -25, -73,   1,   5,   7,   2,   3,   1,   9,   4, -19,   6,   7,   1,   0,  -1, -12, -26, -19,  15,   6,   1,   2,   0, -38,  10,  -1,  17,   2,   0,   3,  -1,   4,   2, -14,  30,  -8,   1,   0,   0,   6, -14,  29,  27, -19,   3,  -1,   0,  13,   3,  31, -11, -19,   2,  -2,   0 },
      {   1,   7,  20,   2,   4,  -4,   2,  -1, -12, -13, -53,   5,   6,  -3,   5,   1,  17,  -1, -30,  30,   9,   2,   0,   1,   5,  73,  -2,  22, -12,  -7,  -4,  -2, -35,   1,  -6,   2, -21,   0,   1,   1, -21,  19,  12, -17,  -6,   9,   2,   0,   5,   6, -24,  -8,   5,   3,  -1,  -1,   1,  -8,  -8,  24,   6,  -2,   0,   0 },
      {  14,  29,  29,  18,  -8,  -4,  -3,   0,  12, -12,  22,  21,  -7,  -2,  -3,  -1,  27,  44,  -1,   8, -15,   1,  -4,  -3, -10,  18,  19,   3, -17,   1,   0,   0,  14,  27,  23, -19, -19,  -6,  -4,   0,  26, -16,   8, -24,  -3,   1,  -5,   0,   6, -20,  23, -32,  12,   8,  -1,   0,  -3,  -8,  -9, -49,  14,   6,   0,   0 },
      {   8,   9,  18,  65,  -8,   0,  -2,   3,  -3,   3,  -5,  49, -18,   1,  -2,   0,   8, -27,  -3, -24, -18,   2,   3,   1, -18,   8, -26, -23,  10,   8,   0,   1,  24,  -4,   6,  26,  19,   5,  -7,  -1,  19,   3,  16,  -3, -14,  -1,  -3,   1,  -1, -21, -20,  -2, -14,   0,   3,   2,  -1,  -5,  16,  36,  -2,  -4,  -1,  -1 },
      {   0,   8, -44, -40,   7,  -1,   0,  -3,  -7, -33,  -3, -22,  15,   2,   2,   2,  37,  -6,  15,   2,  12,  -6,  -1,  -1, -18,   1, -19,  10,   4,  -7,   3,  -1,  43,   4, -12,  28,  -7,  -1,   2,   0,  44,  -1,   7,  -8, -17,   2,  -2,   0,   2, -17, -25, -27,   2,   3,   1,   2, -12, -12, -13,  20,  21,   0,   2,   0 },
      {   1,  11,  19, -24,   5,  -2,   1,  -1,   2,   9,  28,  11,   7,  -2,  -1,   0,  13,   9,  47,  34,  -5, -11,  -5,  -2, -35,   4,  14,  13, -19,  -2,  -2,   0,   5, -32,  12,   1, -17,  15,  -1,   1, -15, -23,  12, -20, -11,  14,   0,   0,  -9,   2,   9,  16, -11,  -5,   2,   1,  12,  28,  36,  42, -19, -17,  -3,  -1 },
      {   2, -13,  48, -19,  -2,   5,   0,  -3,  -1, -10, -54, -13,  -4,   8,  -2,   2,   0,   0,  18,   7,  -2,   4,  -1,   2,  -8, -14,   9, -19,  -9,  -1,   2,   2,   7, -48,  -3, -23,  17,   0,   1,   2,  49,  12,  -1,   1,  33,  -2,  -3,  -2,  31, -20,  -2,   9,   1,   0,  -3,   0,   1, -19,  11, -11, -17,   2,   2,   1 },
      {  -4,   9,  -7,  17,   0,   6,  -1,   1, -17, -31, -12, -32, -10,   5,   2,   0,  -1,   9, -30, -44,  -5,   2,   2,   4, -47,  -1,  10,  -6,  11,   8,  -2,   2,   4, -31,   3, -25,   2,  11,   1,   2, -21, -43, -10, -27,  19,   6,   4,   2, -15,   8,   0,   5,  36,  -6,   0,  -1,  -4,  15,  -9,  12,  22,  -9,   0,  -1 },
      {  -5,  -6,  -8, -15, -34,   0,  -4,   1,  -2,  -6,  32,  34, -20,   8,  -1,   0,  -3, -20, -20,  49,  -3,  15,   3,  -1, -13, -19, -47, -13, -28,   2,   7,   3,  -8,   0, -35, -29,  -7,   2,   3,   2,  -3,  -3,   3,  -5,  34,   6,  -3,  -1,  -2,  -6,  -7,  -1,  32,   2,  -2,   0,  -2, -11, -10,   8,  11,  -2,   0,   1 },
      {  -6,  -1,  -1, -38, -11,  -2,  -1,   4, -10,   4, -28,   7,  -8,  10,  -4,   1,   5, -15,  28, -17,  -4,   7,  -2,   0, -11,  22, -48, -14,   9,   6,   7,   2,  15,  14,  38,  15,   4,  -9,  -3,  -1, -39,  16,   0, -22,   3, -14,  -2,  -3, -23, -16,  11,  12,  26,  10,   1,  -1,   1,  10,  21, -43, -11,   5,  -3,   0 },
      {   1,   4,  29, -13,  -9,  -7,   1,  -4,   2, -14, -41,  30,   3,   2,   4,   0,   0,  40,  -4,  22,   2,  -2,  -1,  -2, -12, -28,   3,  -9,   6,  -1,  -5,  -1,  44,  25, -14,  -1,  18,   5,  -3,  -2, -14,  11,  -6,  19,  -8,  -8,   3,  -1, -34,  38,   2,  11, -13, -12,   3,  -2, -18,  26, -29,  11,  15,   0,   1,  -1 },
      {   6,   6,  11,  30,  63,   1,   5,   2,   3,   6,   1,  20,  36, -11,  -2,   0,  -2, -13,  23,   8,  -8, -16,  -1,  -2,  -9, -12, -23,  26,  -8,   9,   7,  -1,   7, -25, -16,   4, -23,   6,  10,  -1,  -7,  13, -36,  -1,  -4, -12,   5,   0,   1,   1,  -3,  23,  29,  -6,  -4,   0, -12, -12, -13, -24,  29,   8,   0,   3 },
      {  -3,  -2,   5, -31,  53, -17,   3, -12,  -3,  -2, -12,   9,  33, -13,   4,  -1, -17, -10, -20,  -7, -38, -14,   4,   2,   8, -12,  -3, -40, -32,   7,   4,   5, -25,  26,   9,   0,  17,  18,  -5,   1,   1, -26,  14,   0,   9,  13, -10,   3, -17, -24,   1, -29,  -7,   1,   1,   3,  -5,   8,   3,   9,   6,  -6,   2,  -1 },
      {  -5, -12,  16, -16, -28,   6,   3,  -1,  -4,   4, -18,   2, -40,   7,  -1,   1, -13,  -2,  28,   2, -23,   6,  -6,   2,  16, -15,   9,  15,  11,  18,  -5,  -1, -43,   0, -12,  18,  -5,  14,   3,   0,   8, -20, -22,  -5, -44,  -7,   6,   3,  -8, -26,   7,  -9,   6,  -6,   1,   2, -16,  -2,   5,   7,  57,   1,   0,  -1 },
      {  -2,  -2,  14,  30,   4,   5,  -6,   2,  -4,   8,  -5, -10,  21,   6,  -1,   1,  -6,   9,  22,  16,  46,   8,   5,   1,  -6, -15,   9,   3,  16, -27,   8,   0, -22,   2, -12,   7,   5, -39,   3,  -1,  -3, -14,  19,   9,  22,  -7,  -2,   0, -60, -38,  -8, -24,  14,  19,   5,   2, -18,   1,  11,   5,  -1,   4,   0,  -2 },
      {   1,  -1,  -3,  34,  -2,   2,   4,  -1,  -6,  -2, -18, -49, -29, -13,   8,  -2,  -1,  -1,  17,  -2, -23,  -6,   3,  -1,  -2, -19, -19,   3, -37,   2,  -2,   1,  10,  13, -36,  -8, -28,  20,  -1,   1,  -3,  31,   0, -14,  -6,  31,  -2,   2, -28,   9,   0, -23, -19,  16,  -5,   1,   2,  16,  16, -15, -22,   2,  -4,   1 },
      {  -4,  -9,   3,  -3, -25, -41,   6, -19,   0,   0, -11,  30,  19, -24,  15,  -2,  -4,  -4,   9, -21,  42,  -7,   6,   0,  -4, -19, -13, -18,  12, -20, -11,  -3,  -6, -18,   0, -16, -21,   1,  -2,   3, -16, -19, -28, -20, -23,  36,   8,   4,  18,  -6,  -8, -16, -27,  23,  -3,   1,  10,  -7, -17, -13,  -3,   6,  -5,   1 },
      {   2,  -1,   3,  13, -11,  -5, -10,  11,   3,   9, -12, -11,  31,   0,  -5,   2,   4,  -3,  10, -19,  44,  14,  -3,   0,  14,  -3, -17, -33, -15,  15,   2,  -1, -11,  26,   6, -17, -37,  15,   7,  -2,  37,   8,  27,   1, -15,  -4,   0,  -2,   7,  13,  29,  24,  28, -21,  -2,  -1,   6,  26,   9,  17,  21, -17,  -2,   0 },
      {   0,   1,  -7,  -1,  -3,  27,   3,  -2,   0,   8,  11,   8,   3,  21,   2,   2,   1, -13,  -8,  -6,  15, -17,  -2,   4,  -1,  26, -26,  18,  -3, -19,  -9,   1, -11, -15,  24, -31,  18,  15,   3,   1,  26,  20, -29,  -2,  16,  13,   3,   1, -21, -10,  21, -12, -44, -22,  -1,   0, -45,  36,  -7,  -5,  12, -10,   7,   1 },
      {   5,   4,  10,   0,  10,  66,   3,  53,   2,   2, -12,   4,  16,  33, -10,   5,  -2,   0,   4,   7,   5, -14, -14, -11,   5, -15, -17, -10,  10, -15,   4,   0,  -3,  18, -12,  -4,  12,   0,   6,   2, -10, -24,  -8, -30, -12,   8,   2,   2,  20,  10,  -6, -21, -12,   7,   2,   1,  44,   3,  -7,  -2,  -1,   1,  -2,  -2 },
      {  -2,  -2,  -5,   1, -11,  12,   3,  24,  -3,  -4,  13, -19,  15,  -5,  -5,   3,  -3,   1, -23,  18,  18,  -6, -14,  -7, -13,   3,  16, -39, -17,  14,  -7,  -2,  16, -21,   7, -21,  -1,  27,   6,   2, -32,  17,  15,   8, -39, -21,   2,  -2, -17, -31,  -9,  11, -28,  -7,   2,   1, -16, -38,  13, -13,  16,  24,   3,   0 },
      {   3,   3,   0,  12,  28,  -4,  33, -23,   0,  -1,  -2, -20, -18, -15,  21,  -6,   5,   3,  21,  24,  13,   4, -10,   6,   4,  -4, -27,  14,  46,  25, -13,   2,   7,  23,   5, -42,  24,  10, -11,  -1, -18,  -1,  28,  -7,   7,   2,  -4,   2,  17, -23,  22,   0, -10, -12,  -2,   2,  14, -17, -12,  16,  10,  -8,  -2,   0 },
      {   0,   1,  -7, -11,  12,  -8,  15,  -4,   2,   1,  10,  16,   6, -13,   8,  -1,  -1,  -1, -25,  15,   0,  -7, -11,  -1,   3,   7,  12, -19,  51,   8, -12,  -1, -10, -12, -39,  -5,  17,   5,  -2,   1,  33,  19,   1, -47, -21,  -9,  10,   2, -20,  25,   5,  -9,  22,   4,   4,  -1,  -9,  13,  42, -19, -18,   3,  -3,  -2 },
      {   0,  -3,  -2,  26, -16, -17,  -9, -21,   1,   2,   1, -30,   9, -10,   3,  -3,   5,  -5,   2,  47,   6,   2,  -1,   6,   0,  10, -13, -15,  -7,  16,  -4,  -2,  -6,  -5,  17,  20,  24,  12,   5,  -5,  13, -27, -26,  -2,  -1, -30,   0,   2,   1,  -2, -38,  -5,  -4, -23,   5,   4,  36,  45, -13, -28,  -2,   7,   6,  -2 },
      {   0,   1,   3,  12, -36,   9,  10,   8,   0,   3,   0,  -4,  26,  -3,  24,   0,   1,   1,  11,   8,   7, -30,   3,   0,   6,  -1,  11,   7,   2,  -5, -13,   0,  -3,  16,   5,  41,  26,  55, -10,   0,  -3,  -8,   4,  -5,  25,  39, -13,   1,   1,   7,  -2,  14,  30, -11, -14,   1, -25, -32,   2, -15,  -6, -14,  -5,   5 },
      {   0,   1,  -2,  -8,  11, -26,  11,  69,   0,  -1,   0,  13,   0, -15,  -8,  12,   4,   4,   6,  -7,  16,  53,   1, -22,   1,   4,   7,  19,   7,  52,   3, -10,  -3,  -7,  -3,   1,   2,  21, -12,  -2,   4,  -1, -14,   8,  17,  13, -12,   2, -10,  -9, -14, -18, -10,  -2,  -2,   3,  -6,  11, -12,  -5, -19,  -9,   3,   2 },
      {   1,   2,   0,   1,   4,   8, -27, -51,   4,   3,   3,  10,  26,  55,  -1,  -3,   1,   9,   0, -14,  18,  45,  10,  18,   0,  -5,  12,  13, -12,  27,   0,   1,  11,   1, -23,   8,   9,  23,  -4,  -4, -25,  16, -10, -16,   4,  -5,  -7,   4,  12,  -2,   2, -30,  -8, -10,  -6,   9,   8,  -9,  26,  -2,   5,  -4,  -2,   3 },
      {   0,   1,   3, -11,   1,  34,  10, -32,  -1,  -1,  -4,  21,  -6,   6,   4,  -5,  -2,  -2,   3, -19,   2,  -2, -23,   9,  -1,  -3,  -2,   6,   8,  22,  -4,  -1, -14,   5,  -4,  -3, -44,  13,  14,   1,  17, -26,  18,   0,  -2, -12,   1,   1, -35,  -2, -28,  15, -10, -30,  -2,  -3,   0, -19, -33, -13, -53,   5,   8,   6 },
      {  -3,  -4,  -1,   8, -16, -33, -51,  16,  -3,  -2,  -7, -13,  24, -15, -29,  11,  -2,  -4,  10,  -2, -25,   3,  -4,  -1,   0,  -6, -15,  21,   1, -14,   0,  -5,  -8,  11, -14,  -9,  18, -13,   3,  -2,  -2, -13,   6, -46,  -1, -21,   6,   1,  -5,   7,  22,   7, -24, -37,   7,   2, -16, -35, -10,   6, -17,  -7,  11,  -1 },
      {  -1,  -1,   0, -10, -14,  -9,  -2,   6,  -1,  -2,  -6,  23, -12, -31, -10,   3,  -1,   1,  -3, -30,  22, -53, -10,  -2,   2,  -3,  11,  33, -10,  -5,  10,   0,   8,   2, -17, -11,   8,  23,  21,  -2, -10,  18,   7,  13,  20, -32,  -3,  -1,  -1, -20,  -1, -33,  14, -26,   2,   2,  41,  10,  20,   3,   6,   1,   9,  -3 },
      {   0,  -1,  -1,   1, -39,  -1,  40,   3,   3,   3,   2,  -2,  54,   7,  38,  -4,   1,  -7,   1, -18, -37,  14, -23,  -7,  -3,   5,   2,  19, -16,   8, -29,   4,   0, -11,  -9,  -6,   7, -24, -16,  11,   9,  15,  -1,   2,   3, -15,  -4,   4, -21,  -9,  18,   4,   1,  12,   8,   0,  35,   2, -16,  12,   8,  10,   0,  -2 },
      {   1,   2,   1,   7,  -5,   3,  45,   2,  -1,  -3,  -5,  -4,  -3,  -9,  17, -10,  -2,   2,   2,  -1,  11,   4, -26, -11,  -2, -10,  -8, -13, -23,   4,  -9,   4,   1,  11,   4,  15, -15, -28,   7,   6, -18, -17, -20,  -7,  15, -36,   0,   3,  38,  19,  -2, -31,  -3, -24,  10,   0, -51,  -3,  32,  12, -12,  -7,  14,  -2 },
      {  -1,   0,   2,   1, -31,  25,  -6,  -8,  -1,   0,  -4,  -3,  30, -19,   3, -12,   1,   0,   3,   5, -12, -27,  37,  -2,   2,   5,  -3,   6,  18,  53,  65,   4,  -2,   0,   1, -17,  -3,  -4,   6,  -7,  -2,  -4,  -4,  -2,  -3, -13, -17,  -1,  10,   6,   7,  -9,  -8,  33,  -1,   0, -24,   9,  -3,  16, -11,  14,  -3,  -3 },
      {  -1,   0,   2,   4, -16,   5, -23,  -7,   0,  -1,   0,  -6,  26,  -5, -34,  14,  -4,   2,  -5,  11, -28,  -8, -51,  12,  -2,  -4,   3, -10,  34,  19, -25,  -3,   9,  -1,  -3,   4, -21,  -9,  23,   0, -18,  18,   2,  23,  10,  28,  23,   2,  24, -29,  -6, -17,  11,   8,  -6,  -3, -16,  27,  -6,   0,  -7, -27,  -4,   0 },
      {   0,  -1,   1,  10,   5, -43,   9,  -3,   0,   2,  -1,  -6,  -6,  51, -15,   7,  -1,  -3,   5,  -3,  11, -18, -41, -12,   2,   3,  -9,  18, -17,   8,   6,   7,  -3,   6,   9, -18,  31,  22,  31,   7,   8, -15,  -1,  11, -24,  -6,   5,   0,  -1,  15,   3,  -5,  21,  43,   0,  -6, -15,  -5,  -9,  13, -26,  20,  -7,   0 },
      {   0,  -1,   1,  -2,   0,  15, -10, -15,  -1,   0,  -1,   1,  -1, -36, -28,   7,   1,  -3,   4,   1,  11,  11, -44,  13,   0,   8,  -8,  -1,  -1,  19,   0,   7,   0,   2,  19,  11,   0,   2,   0,   9,  15, -10, -14,   1,  39,  14,   2,  -3, -25,  43,   6,  -2, -24,  25,   1,  -4,  18, -30,  26,   8,  36,  30,  -6,   1 },
      {   0,   0,   0,   3,  -5, -29,  25,   5,   2,   2,   2,  -3,   9,  24,  18, -10,  -1,   3,  -2,   3, -23,  17,   1, -20,   3,   1,   6,  -5,  28, -20,  45,  -5,   5,   5,   2,   8, -18,   9,  65,   8,   0,   1,  13,  -9,  27,  11,  13,  -1,  -4,  -1,  -8,   8, -27, -21, -25, -10,   9,   2,  15,  -6,  24,  11,  -3,   4 },
      {   0,  -1,   1,   3,   3,  -3,  12,  -2,   0,   1,   2,  -6, -12,   3,   3,   4,   0,   0,  -8,  15,  11, -10,   4,   6,  -2,   7,   6, -29,  -5,  16,  19,   2,   4, -15,  13,  25,   1, -14,  25, -11,   2,  18, -38,  -3,  -9,   3,  11,  -4, -28,   0,  49, -26,   9, -11,  -9,   5,  27, -36, -27,  17, -17, -48,  -7,   6 },
      {   0,   1,  -3,   3,   0,   5,  24, -19,  -2,  -2,   3,   1,  -6, -14, -43, -24,  -1,   0,  -6,  -3,  18,  10, -28, -17,   4,  -2,  17,  -4, -15,  -6,  28,  32,  -1,  -2, -21,  28,  16,  -2,  -5,  20,  -7,  14,   5, -37,  -1,  15, -30, -10,  10, -26,  16,  29,  -5,   3,   6,  -1,  -4,  24, -31,  -3,   2,   2,  22,   4 },
      {  -1,  -2,   0,   4,  -1, -28,  -9,   7,  -1,  -2,  -2,  -7,   1,  38, -19, -32,  -1,  -1,  -2,  10, -10, -31,  -5, -23,   1,   0,  -4, -11,  34,  -2,  -6,  27,  -3,  -6,   5,  -6, -41,   7, -47,  19,   1,   1, -13,  15,  11, -15, -46,  -2, -10,  10,  -4, -12,  -8,  -5,  -8,  11,  11, -11,   9,   1,   7, -15,  14,  15 },
      {   0,   0,   0,  -1,  -6,  17,  -3, -15,  -1,  -1,  -3,   2,  11, -23, -14, -37,   1,  -1,   4,  -4,  -5,  23,   3, -37,  -2,   4,  -9,  16,  -8,  -5,  10,  17,   2,  -1,  10, -28,  29, -15,   8,  40,   3,  -6,   0,  28, -36,  22,   8,   6, -10,  14, -25, -12,  16, -19, -13,  -9,   5, -14,  21, -16,   0, -40, -12,   0 },
      {   0,   0,   2,  -1,   8,  15, -38,   1,   0,   1,   0,   0,  -5, -15,  18, -33,   1,  -1,   3,   7,  -1,  16,  16, -32,   0,   1,   0, -10,  12,  -7, -30,  -1,   2,   0,  10,  12,  -7,  26,  23,  42,  -2,   8, -13,   2,   8,  -5,  11,  21,   4, -10,  31, -15,  19,   3,   3,  -5,   0,   5, -15,  26, -19,  48,  24,  -5 },
      {  -1,   0,  -2,  -2,   6,  -5, -22,  -5,   0,   0,   1,   2, -14,  16,  14, -16,   0,  -2,  -4,  -4,  19, -36,  -3, -45,   1,  -1,   7,  -1, -17,  40, -11, -29,  -3,   3, -11,  15,   3, -25,  -5,  19,   3,   1,  14, -20,  23,  -6,  23,  38,   0,  -6,  -5,  23, -23,  12,   1,   0,  -1,   2,  -6, -12,  17, -24, -36, -15 },
      {   0,   0,   0,   1,  -5, -11,  30,   9,   0,  -1,   0,  -1,   8,  16, -35, -29,   2,   2,   3,   2,   1,  -9,  39,  23,   0,   0,   0,   3,   0,  15,  -8,  33,   0,   6,   0,   5,  -7,  11,  -7,   6,   3,   0,   6,   2,  14,   9,  67,  -4,  -2,  -1,  -2,  -1,  -4,  -2,  56, -14,   5,  -9,   2,  -5,   0,   1,  13, -24 },
      {   0,   0,   0,   3,  -7,   2,  36, -24,  -1,   0,  -2,  -4,   7,  10, -50,  17,  -1,   0,   0,   2,  -5,  -6,  17, -35,  -2,   0,  -4,  -4,   2,   2,  -1, -65,   0,  -1,   6,  -3,   1, -11, -14, -19,   0,   3,  -2,   6,  -4,  26,  -3,  34,  -5,   6,   6,  -5,  15, -22,   0,  19,   6,  -4,   7,   6,  -2,  30,  27,  -2 },
      {   0,   0,  -1,   0,  -3,  10,  14,  -9,  -1,  -1,  -1,  -1,   6, -16, -24, -34,  -1,  -1,  -1,  -1,  -9,  26,  -9, -17,   0,   0,  -2,   0,   6, -41,   6, -29,  -2,   0,   2,   0,  -9,  47, -13, -33,   1,  -1,  -3,   0,  11, -45,  15,   2,   1,  -3,   7,  -4,  -4,  31,   5,  13,  -4,  -1, -10,   4,  -5, -26, -46,  -5 },
      {   0,   1,  -1,  -3,  -2,  -4,  20,  16,  -1,  -2,   0,   2,   1,  -2, -37, -31,   2,   1,   1,  -1,   4,  -7,  25,  37,  -1,   0,   0,   2,  -8,   9, -36,  15,  -1,   2,  -3,   3,   6, -17,   9,  -1,   1,  -4,   3,  -8,   1, -10,   0,  17,   0,  -2,   2,  -2,  -2,  -8, -77,   0,  -4,   8,  -3,   8,  -5,  26, -39,  18 },
      {  -1,  -1,  -1,  -1,   2,  -1, -18,   4,  -1,   0,  -1,   2,  -1,   4,  12, -63,  -1,  -2,   0,  -3,   3,  -7, -14, -22,   0,   0,   0,   2,  -7,  16, -13, -19,  -1,  -1,  -1,  -1,  10, -17,   2, -62,   0,  -2,   3,  -1,   0,  21,  13, -63,  -1,   0,   0,  -2,   3,  -6,  -1,  -9,   2,   1,   5,   2,   3,  15,  17,  28 },
      {   0,   0,   0,   1,  -2,   6,  14, -14,   0,   0,   0,   0,   3,  -3, -21,  28,   0,  -1,   1,   0,   0,  -2,  39, -18,   0,  -2,  -2,   1,   0,   7, -42, -25,  -1,   5,  -3,   1,  -1,   7,  27,  28,   0,  -6,   3,  -9,   7, -25, -14, -57,  -2,   3,  -3,   5, -13,  35, -15, -30,   3,  -2,   2,  -4,   7, -32,  29,   8 },
      {   0,  -1,  -1,   2,   1,  -1,   0,  -9,   0,   1,   0,  -2,   1,   5,  -6,  16,  -1,  -1,  -1,   2,  -1,  -2,  -1, -23,   0,  -1,   0,  -4,   4,  -1,   0,   7,  -2,  -3,  -1,   3,  -7,  -2, -18,  -6,  -2,   2,  -9,   1,   0,   7, -17, -38,   0,  -5,   6, -12,   5, -30,  -8, -55,   0,   0,  -4,   7,  -8,  25, -48, -75 },
      {   0,   0,   0,  -1,   2,  -5,  -2,  15,   0,   0,   1,   2,  -1,   3,   3, -49,   0,   0,   0,   0,   0,   2, -13,  51,  -1,   0,  -1,   1,  -2,  -2,   1, -44,   1,  -2,   0,  -3,   4,  -2,  16,   2,   1,   2,  -1,   1,  -6,   4, -35,  20,  -4,   5,  -3,   6,  -1,  15,  -8, -15,   2,  -6,   2,  -6,   4, -21,  39, -66 },
      {   0,   1,   0,   2,   0,  -2,  -1,   1,   0,   0,   1,  -1,   0,   2,   9,   4,   0,   0,   0,   1,   3,  -1, -19,  14,   0,  -2,   1,  -2,  -4,  -5,  39, -28,  -1,   0,  -1,   3,   1,   8, -58,  28,  -1,   1,  -2,   0,  -1, -12,  53,  -4,   1,  -6,   3,  -6,   3,   1, -46, -50,   2,   4,  -2,   5,  -4,   1,  30,  21 },
      {   0,   0,   0,   1,   1,  -5,   4,   8,   0,  -1,  -1,  -1,   0,   5,  -4, -17,   0,   0,  -1,   1,  -1,  -5,  -2,  36,   1,  -1,   1,  -1,   1,   1,  10, -55,   1,   0,  -1,   1,   0,  -4,   0,  60,  -2,   2,  -4,   1,  -2,   9, -17, -37,   3,  -3,   3,  -5,   4, -20,  44,  12,  -2,   3,  -2,   2,  -4,  14, -48,  32 },
      {   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,  -1,   1,   5,   4,   0,   0,   0,  -1,   0,  -2,  -5,  -3,   0,   0,   0,   1,  -1,   3,   7,   1,  -1,   2,   0,   0,  -1,  -1, -11,  25,   1,  -2,   3,   0,   1,  -1,  32, -49,   0,   2,   0,   2,   1,   1, -36,  90,  -3,   2,  -2,   0,  -3,   6,  12, -49 },
    },
    {
      { -73,  45,  -3,   3,  -1,   1,   0,   0,  76, -39,  -4,  -1,  -1,   0,   0,   0, -33,   3,  13,  -1,   2,   0,   1,   0,  10,  11,  -9,  -1,  -1,   0,   0,   0,  -2,  -9,   2,   3,   0,   1,   0,   0,   2,   4,   2,  -2,   0,   0,   0,   0,  -2,  -2,  -2,   1,   1,   0,   0,   0,   1,   1,   1,   0,  -1,   0,   0,   0 },
      {  59,  -4, -15,  -1,  -3,   0,  -1,   0,   3, -46,  24,  -1,   4,   0,   2,   0, -58,  60,  -5,  -5,   1,  -1,   0,  -1,  36, -19, -19,   9,  -3,   2,  -1,   1, -14,  -6,  19,  -3,   1,   0,   0,   0,   3,   9,  -8,  -3,   1,  -1,   0,   0,  -3,  -2,   1,   3,  -1,   0,   0,   0,   1,   1,   0,  -1,   0,   0,   0,   0 },
      {  54, -35,  10,  -4,   1,  -2,   0,  -1,  44,  -1, -21,   0,  -3,   1,  -1,   0, -36, -35,  32,   1,   3,   0,   2,   0, -27,  58,  -9, -11,   1,  -2,   0,  -1,  14, -22, -19,  14,  -2,   2,   0,   1,  -8,   2,  21,  -4,  -2,  -1,   0,   0,  -1,   4,  -9,  -4,   3,   0,   0,   0,  -3,   0,   2,   3,  -1,   0,   0,   0 },
      {  41,  17, -26,   1,  -5,   1,  -2,   1,  45, -36,   8,   1,  -1,   0,   0,   0,  23, -26,  -6,   1,   0,   1,   0,   1, -40, -25,  40,   0,   4,   0,   1,   0, -20,  56,  -9, -16,   3,  -3,   1,  -1,   5, -13, -21,  16,  -2,   2,   0,   1,  -2,   0,  18,  -2,  -4,   0,  -1,   0,  -2,   2,  -7,  -5,   4,  -1,   0,   0 },
      { -11, -63,  36,   1,   6,   0,   2,   0,  49,  43, -19, -13,  -3,  -3,  -1,  -1, -26,   1, -32,  19,  -2,   4,   0,   1,  22, -38,  28,   5,  -5,   2,  -2,   1, -18,  12,  12, -17,   6,  -2,   1,  -1,  -3,  11, -12,   4,   2,  -2,   1,   0,  -2,  -2,   0,   3,  -3,   1,   0,   0,   0,   0,   2,  -2,   0,   1,   0,   0 },
      {  27,  23,  -2,  -8,  -3,  -2,  -1,  -1,  22,   2, -33,  11,  -3,   3,  -2,   1,  42, -25,  14,  -6,  -2,  -1,   0,   0,   2, -31, -23,  10,   3,   2,   0,   1, -16, -31,  59,   4,  -2,   0,  -1,   0, -16,  45,  -8, -30,   6,  -3,   2,  -2,  -2,  -9, -20,  19,   2,   0,   0,   1,  -1,   2,  12,  -1,  -4,   0,  -1,   0 },
      { -24, -35,  32, -10,   7,  -2,   2,   0, -19, -27,   8,  11,   1,   3,   0,   1,  -9,  12,  27, -18,   3,  -3,   1,  -1, -41,  23, -10,   9,   2,  -1,   0,   0,  17,  33,  24, -19,  -5,  -2,  -1,  -1,  -4,  23, -57,  -7,   9,   0,   2,   0,  12, -22,   3,  32,  -7,   1,  -1,   1,   2,   3,  11, -14,  -1,   1,  -1,   0 },
      {  -7, -34, -10,  10,   2,   2,   0,   0,  35,  -9,  44, -17,   0,  -4,   0,  -1,  39,  30, -45, -14,   7,  -4,   1,  -2, -46,  -1, -22,  31,  -5,   3,   0,   2,  31, -24,  14,   4,  -8,   4,  -3,   1, -12, -10,  23, -10,   5,  -1,   1,   0,  -5,  16,  -7,  -7,   2,  -2,   1,   0,  -3,  -3,  -4,   7,  -1,   1,   0,   0 },
      { -19, -29, -13,  24,  -2,   5,  -1,   2, -23, -35,  36, -12,   3,  -1,   2,   0,  -8, -31,  12,  14,   2,   2,   0,   1,   4,  -3,  35, -15,   0,  -2,   1,  -1, -20,  16,  12,   4,   1,  -2,   1,  -1,   9,  37,  25, -32,  -3,  -3,  -2,  -1,  -8,  18, -52,  -9,  17,   0,   2,  -1,   6,  -7,   2,  27, -10,   0,  -1,   0 },
      {  -7, -41,  39, -10,   3,   0,   2,   0,  11, -38, -11,  14,   4,   2,  -1,   1,  51,  22,  29, -15, -11,  -1,  -2,   0,  21, -11, -27, -25,  14,  -2,   4,  -1, -40,  10, -18,  32,   0,  -1,   1,  -1,  16, -14,  16,   4, -14,   7,  -3,   1,  -7,   1,  13, -16,   6,   0,  -1,   0,  -3,   9,  -9,   0,   4,  -4,   2,   0 },
      {  12,  10, -13,  10,  -5,   2,  -3,   1,  24,  35,  26, -30,   2,  -6,   1,  -2,  23,  19,  -9,  -7,   0,  -3,   0,  -1,  16,  29, -20, -25,   3,  -4,   1,  -1, -23,  17, -34,  11,   3,  -1,   1,   0,  15,   2, -32,  -9,   1,   1,   0,   1,  -9, -33, -32,  37,  13,   4,   1,   1,   9, -13,  35,  14, -19,   1,  -2,   1 },
      { -10, -34, -43,  23,   3,   6,   1,   3,   6,  -1,  36,   7, -13,   1,  -3,   0,  18, -38,  17, -24,  14,  -2,   4,  -1,  47,  19,  -7,  13,  -7,  -3,  -3,  -1, -19, -26,  -1, -24,   6,   3,   3,   0, -17,  12, -14,  34,   7,  -4,   1,   0,  11,  -9,  30,  -3, -16,   3,  -2,   0,  -6,   9,   0, -18,   8,   0,   0,   0 },
      {  10,   5,  29,  -7,   0,  -2,  -2,   0,   1, -46,  -8,  -9,   7,   1,   4,   1,  37,  15, -12,  36,  -7,   0,  -3,  -1,  26,  14,  34, -28, -21,   4,  -5,   1,  33, -30, -22, -32,  23,  -1,   3,   1, -38,   9, -10,  17,  11,  -8,   3,  -1,   2,   4, -12,  14,  -8,   3,   1,   0,  -1, -10,  12,   2,  -2,   2,  -2,   0 },
      {  10,  27,  36, -39,  12,  -6,   2,  -4,   5,  16,  22,   7, -15,  -1,  -2,   1,   6,   7,  -7, -28,  14,  -3,   2,  -1,   0,  25,   2,  15, -13,  -1,  -2,   0, -23,   4,  -3, -31,   6,  -1,   2,   0,   6,  33,  -6,  15,   0,  -7,  -3,  -1,   7,  -7, -22, -47,   3,  10,   1,   2, -12,   2, -35,  43,  15,  -7,   3,  -1 },
      {   1,  13,  29,  11,  -4,   0,   0,   0, -18, -18,   4, -43,  14,  -3,   4,  -1,  -5, -16, -34,  35,  13,  -2,   2,   1, -30,  19, -12,  10, -14,   5,  -1,   0, -50, -20,   7,  20,  10,   5,   0,   2,  22,  25,  12,  20, -14,  -3,  -2,  -1,  11, -21,  23,  -1, -15,   3,  -2,   1,   1,   8,  15, -30,   6,   1,   0,   0 },
      {  -9,  -4, -32,  23,  -5,   3,   0,   2,   3,  25,  -8,  -2,   0,   0,  -1,   0,  -5,  47,  31,   1,  -8,  -3,  -1,   0, -44, -19,   4, -38,   2,   3,   3,   1, -10,  -6,  -8,  10,  32,  -9,   3,  -1, -34,  42,   4,  27, -17,  -1,  -1,  -3,  22,   0,  -6, -14, -16,   6,  -3,   2,  -5,  -5,  -6,   0,  18,  -4,   1,   0 },
      {  10,  25,  47, -15,  -4,   1,  -4,   2,   9,  20,  66, -13,   3,  -7,   0,  -1,  -8, -17,  16, -18, -11,   2,   0,   0,   8, -18,   4, -20,  22,  -1,  -1,   0,   6,   5,  -1,  17,  -5,   2,   1,  -2, -32,   5,   9,  -9,  -3,   0,   4,  -1,  14,  33,   5,  11,  -7,  -9,  -1,  -1,   2, -14,  -3, -48,   7,   9,  -2,   2 },
      {  14,  13,  26,  68, -21,   1,  -2,   2,   7,  10,  -4, -47,   3,   3,  -2,  -1,  14,   9,  40,   9,  11,  -7,  -1,   1,  17,   4,  -5,   5, -21,   0,   3,   0,  28,  15,  20, -16,  -2,  -2,  -4,  -2,  16, -11, -11, -19,  -6,   8,  -2,   1,  -4,  -4,   5, -19,   1,  -3,   1,  -1, -10,  14, -31,   0,  21,  -7,   4,  -2 },
      {   1,  -6, -26, -26,   3,   3,   2,   2,  17,  13,  18,  30,   3,  -7,  -1,  -3,  18,  18,  -6,  40, -30,   6,  -1,   0,  16,  37,  24, -15,  -5,   5,  -6,  -1,  21,   6,  28,  10, -10,  -5,  -4,  -1,  40,  19,   1, -25, -22,   6,  -1,  -1,  16, -19,   7, -18,  -3,   0,  -3,   0,  -6,   0, -10, -36,  19,   1,   2,   1 },
      {   2,  -4,   5,  -2,  -1,   0,   3,   0,  -1, -19,   6,  -9,  -2,   3,  -5,  -2, -10, -34, -25, -15,   3,   1,   5,   4,  15, -31,  -9,  -4,  18,   3,   4,   0,  53,   2, -12,  36,   9, -15,  -1,  -4,  -5,  11, -12,   9, -39,  16,  -2,   0,  25, -52,  -6, -13,   8,  11,  -6,   3, -19,   4,  10,  23,  17, -14,   6,  -1 },
      {  -3,  -5, -17,  -3,  -4,   2,   0,   2,  -4, -12, -40, -22,  11,  -3,   4,  -1,   6,   6, -27,  -9,  16,   1,   1,   2,  22,  27,  -3,  19,  13, -15,   4,  -4, -15,  33, -11,  14, -31,  -2,  -4,  -1, -34,   9,   2, -14,   7,  18,  -1,   1,  37,  18, -23,   8, -21, -11,  -1,  -1, -15, -29, -21, -13,  38,   5,   3,   1 },
      {  -8,  -9, -17, -32,  -6,   2,   2,  -1,  -4,   0,  11, -25,  21,  -6,  -1,  -1,  -9, -24,  -9,  12, -24,   5,  -1,   0,  -7,  -2, -30, -38,   6,   8,   6,   2,  -7, -14,  12,  -9,  22, -16,  -5,   2,   7, -19, -23, -10,  23,  20,   1,  -1, -30,  13,  -3,  16, -34,  -4,   6,  -1,  -6,  19, -31,  24,  43, -11,   1,   0 },
      {  -8,  -8, -25, -46,   7,  -8,  -2,   0,  -4,   2,  -3, -43,  14,   6,   5,  -1,   0,  -3,  13,  18,   7, -13,   4,  -4,   3,   3, -13,  -2, -21,  10, -10,   4,  -7,  14,  10,  -2,   7,   9,   1,  -1, -44, -10, -13, -10, -30,  -9,   8,   3,   3,   4,  18, -13,  66,  -1,  -2,  -1,  -9,  12, -15,  -7, -22,   0,   1,   0 },
      {  -1,  -5, -14, -34,  40, -10,   0,  -5,  -1,  -4,  -8, -34, -48,  10,   2,   3,   5,   5,  19, -11,  49,   0,  -2,  -1,   2, -19,   9,  -7, -23, -18,   3,   0,  20,  -1,  -5,   9,  20,  18,  -7,   2,  35,   3,   7,  -2, -11,  -3,  -3,   0, -18,   7, -11,  11, -22,  -5,   8,   1,  13, -11,   7, -24,  21,   5,  -3,   0 },
      {   1,  -5,  -9, -30,   2,  -1,   2,  -5,   7,   5,   8, -22,  37,  -6,   1,  -3,   6,  -4,  42,  26,  -9,  -3,  -5,  -1,   7, -23, -24,  31, -13,  14,   5,  -3,  14,  21,  -8,  -9, -10, -11,   2,   3,  17,  -2,  27,  19,  29,  -9,  -8,  -1,  38,  -6, -13,  -9, -37,  19,  -2,   0,  -1, -12,  19,  11, -24,  -3,  -1,   2 },
      {  -2, -10,  -8, -10, -47,  10,  -5,   6,  -1,  -7, -25, -21,  26,   1,   4,   0,  -1,  -8, -15, -47, -12,  -3,   1,   2,   0,  -6,  -6, -18,  13,  13,  -6,   1,  20,  -8,  -6, -18,  10,  13,   7,   1,  44,  25,  -7,  13,   8, -34,   5,  -2,  -3,   9,  -5, -17,  18,   2,  -3,  -1,  11, -27, -19, -29, -10,  20,   1,   4 },
      {  -4,  -2,  -6, -23, -15,   9,   0,   2,  -4,   0,  -1, -35,  -1,  -5,   4,   0,  -1,  17,  16, -22,   6,   6,  -7,   4, -10, -10,  53,  -7,  16,  -8,   1,  -4,  -8, -36, -17,   1, -45,  -3,  10,   0,  -4,   4,   7, -27,  37,  10, -10,   4,  -3, -40,  19,  -5,   1,   8,   0,   1, -19,  32,   2,  -2,  -3,  -7,   5,  -4 },
      {   2,   4,  -2, -10,  -2,   6,   0,   1,   4,   7,  -7, -18,  21,   2,   1,  -1,   9,  11,  -3,  -8,  21,   0,   3,   1,  18,  29,  21,  10,  38,  -8,  -1,  -1,  21,  20,  26,  30,   9, -25,  -2,  -3,  -6,  29,   1,  24, -15,  -3,  -1,  -1, -33,  27,  25,   3, -29,   6,   1,   0,  26,  30,  -1,  21, -34,  -9,  -3,   0 },
      {   5,   6,  -3,  20,  63,  -9,  16,  -2,  -4,  -9, -11, -32, -40, -16,  -6,   3,   1,  -2, -10,   2, -35,  35,  -9,   3,   0,   6, -19, -15,  32,   3,   4,  -1,   4,   3,   4, -15,  -5, -29,  14,  -1,   6,   9,  -1,   4,  22,   1,  -3,   1,  14,   6,   5, -14,  13,   7,  -8,  -1,   2,  -1, -16, -12, -30,   6,   4,  -1 },
      {  -2,  -1,  -2,   4, -30, -45,   5, -32,   1,   0,   2,   8, -31,  40,   1,   9,  -3,  -3, -26,  -1,   8, -28, -14,   2,   7,   7, -14, -37, -11,  14,  16,  -1,   5,  32,   7,   8,   3,   5,  -3,   2, -12,   7,  22, -16,  32, -19,  -9,   3,   9,  -7,  10,  -4, -17,  16,   8,   0, -12,  23,  -3,  -7, -19,   0,   2,  -3 },
      {   2,   4,  -1,  11,  31,  21, -15,  30,   0,   2,   1,  17,  16, -23,  12,  -7,  -2,  -5,  -7,  20,   6,  -1,   1,  -4,   2, -12, -24,  -5, -11, -10, -13,   4,  18,  26, -27,  21, -13,  33,  10,   0, -15,  36,   3,  -1,  34, -37,  -6,   1, -38, -17,   7,  -5,   2,  -1,  11,   2,  -8,  23, -19,  -6,  13,   9,  -2,  -2 },
      {  -3,  -1,  -2,   0,  -5, -15,  16, -16,  -5,  -2,  -3,   0,   2,  -5, -10,   3,  -2,  -3,   3,  -7, -18,  -2, -10,   3,   0,   1,  -4,  26, -12,  19,  14,  -7,  -4,   6, -20,  -8, -12, -36,   6,   1, -29,   2,   1, -14, -21,  -5,   3,  -1, -65, -16, -13, -24, -11,  44,  -7,   3,  25, -15,  11, -34,  21,   2,   1,   3 },
      {  -1,  -4,  -3,   9,  12, -36,   0, -33,   3,   1,  12,   6,  18,  16,  10,   7,   2,  -3,   8,  36,  23, -19, -13,   2,  -5, -11,   0,  -8,  30, -22,   4,  -3,   0, -32,  -2,  17, -43,   5,   6,  -2,   9,   0, -47,  12,   0, -19,  -2,   3,  -2,   7,   7, -14,   4,  11,   2,   3,  11, -38, -22,  15,   4,  10,  -2,   0 },
      {   3,   3,  -2,  15,  46, -22,   8, -18,   4,   4,   0,  15,  51,   5,  -5,   5,   0,   1,  -1, -25,  19, -19,  -3,   1,   2,  -2,  22,  10, -15,  46,   2,  -3,  -2, -11, -22,  13,  23, -33,  -3,   2,  17,   9,   2, -15,   9,   2,   1,   0,   6,   3,  -2,  30,  12, -17,  -4,  -4, -20,  10, -25, -14,  -4,  -2,   7,   1 },
      {   1,  -1,   2,  -8, -21,  34,   7,  26,   1,  -2,   3,  15, -22,  -2, -21,  -5,  -1,  -6,  -8,  17,  28, -14,   5,  -4,  -4, -13, -15, -12, -38,   5,  22,  -4,   3,  -9, -25,   8, -19, -34,   1,   0,   4,  21, -18, -10,   7,  39,  -6,   1,  11,  13,  11,   0,   1,   1, -11,  -2,  37,  -7, -33,  -4, -34,  -2,   8,   3 },
      {   1,   1,   5,   2,  16,  37, -21,  35,  -1,   0,  -2,   4,  -7,  -3,  13,   0,  -2,   0,  -3,  -6,  37, -15,  -9,  -1,   1,  10,   9, -27,   6,  40,  -7,   0,  -9,   0,  39,  17,   2, -28,   3,   2,  -5, -41,  -8,   1,  17, -25,   3,   2,   4,  -4,  -6, -20,  -4,  36,   2,   0, -17, -30,  14,  -2,   5,  -2,  -2,   1 },
      {   1,   2,   0,   3,  11,  -1,  13,  -8,   2,   2,  -2,  16,  50, -21, -18,   2,  -2,  -1, -13,  -9,  44,  21,  -3,  -4,  -3,  -3, -13, -36,  -2, -54,  24,  -2,   1,   2,  12, -33,   7,  -6,  -5,   1,  -5, -14,   5, -12,  -8,  11,  -5,   2,   4,  -6, -13, -31,  -4,   3,   4,   1,  -8,  16,  20, -23, -17,  10,  -1,  -4 },
      {   2,   1,  -1,  11,  35, -13, -12,   5,   0,  -2,  -9,  -9,  18, -19,  20,  -7,  -3,  -7, -10, -38, -32, -29,   5,   1,   1,   1,  10, -24, -35,   3,  -7,   6,  -2,  -4,   7,   0, -21,  34,  -4,  -1,  -6,  -5,  -6, -22, -20,  11,   0,   2,  21,   1,  18, -12, -32,  17,   5,   1,  47,  -2,   3,  18, -11, -13,  -4,   0 },
      {   1,  -1,   1,   0, -18,   9,   2, -47,   2,   1,   1,   8,   1, -41,  18,  14,   3,   0,   2,  -3,  22,  60, -25,   6,   3,   2,   1,   1, -17,  15,   1,  -3,   3,   6,   2,  19,  17,   8,   5,   0,  -4,   1,  -3, -10,  19, -11, -10,   3,  20,   7,  21,   6,  26,  15,  -1,   2,  39,  -4,   1,  11,  24, -25,  -5,  -4 },
      {  -2,  -2,   0,  -1, -18,   3,  19, -26,  -5,  -5,  -3,  -1, -20, -66,  -2,   2,  -2,  -2,   4,  -5, -10,  -2,  -1,   1,   1,   0,  10,   6, -26,  -1,   4,   2,  -4,  -5,   5,  31,   4,   4,   1,  -8,  -3, -13, -22,   2, -19, -19,   3,   1,  -7,   0, -12,   2, -32,  -8,   7,   0, -48,  -4, -21,   4, -39,  33,   3,  -1 },
      {   1,   3,   1,  -2,   8,  -1,  58,  16,   0,   1,   4,  -8,  15,  12, -45,  -7,   0,   1,   4,  -8, -10,   4,   2,   9,   3,   8,  17,  -8, -15,   5,  32,  -7,   1,   7,  32,  25,   1,  29,  -7,  -6, -13,  -8,  10,  26,  28,   0, -10,   0, -13, -28,   8,  -1,  12, -22,  -7,   0,  17, -33,  -5,  13,   9,  11,   6,   1 },
      {   0,   1,   0,   0,   4,  13,  -7, -41,  -1,   2,  -3,  -1,  -3, -22,  17,  12,  -1,  -1,  -1,  -2,   8, -24,   3,   3,   1,   0, -10, -24,  -7,  31, -12,   0,   9,  12,  15, -15, -37,   0,  10,   6,   3,  19,  10,  36,   4,  47, -15,  -6, -28,   8,   5, -12,  13, -40,  -4,   3, -17, -12,  34,   0,   7,  -1,   4,   1 },
      {   1,  -1,  -1,   0,  -2,   9,  25, -20,   0,  -3,  -1,   3, -10, -25, -11, -12,   3,   3,   2,  13,  25, -22,  23,  13,   0,  -4,  -3, -11,  34,  28, -26,  -2,  -2,  -1,   2, -19, -16,   7,  -9,  -1,  -7, -19,  16,  10,  -9, -21,  24,  -5,  17, -26, -34,  10,  -7, -12,  -9,   0,  47,  37, -24,  -9,   8,   5,  -1,   1 },
      {  -2,  -4,   2,  -1,  -5, -13, -37,  -7,   0,  -1,   3,   3,   1,  -1,  27,   7,   0,  -2,   5,   3,   3,   1,  -1, -14,  -1,   2,   2,   9,  11, -14, -21,  12,  -3,   3,  10,   0,  38, -16,   4,   1, -17, -11,  21, -10,  15,  29, -13,  -4, -10, -56,   6,  -8,  -9, -29,  11,   6,  28, -36, -34,  -2, -20,  26,   2,   3 },
      {   1,   2,   1,   3,  13,  50,   9, -24,   0,   2,   3,  -1,   7,  49,   4,   7,  -1,   0,  -1,   3,   7,  44, -13,   3,  -2,  -2,  -9,  -8,   8,  33,  -1,  -6,  -4,   1, -13, -20, -13,  26, -21,   0,  -9,   6,   2, -24, -34,  -9,   7,  -2,  -9, -12,  15,   0, -33, -12,   5,  -2, -14, -19, -12,  12, -23,   6,   5,   3 },
      {   0,   0,   0,   3,  -5, -19,  11,  -7,  -2,  -1,  -2,  -1,  -3,  -9, -29,  -3,   2,   3,   2,   6,  12,  11,  68, -14,   0,  -2,   0,   1,  11,  14, -40,  12,  -1,   2, -15,  10,  16,   7, -27,   9,   1,  10, -16, -32,  29,  12,  25,  -4,  -4,  11,  25, -34,  -9,   7,  -7,   1,  -8, -19,  25,   6,  -1,   2,  -2,   1 },
      {  -2,  -2,  -1,   1, -13, -50,   6,  33,   1,   1,   3,   6,   7,  17,   7, -25,   1,  -1,   0,  -6,  19,  56,  -7,  12,  -3,  -2,  -3, -13, -23,  19, -47,   8,  -1,   0,   7,   9, -24,   1,  37, -16,  -1,  -6,  -5,  17,  -9,  15,   0,  -5,  -7,   3, -21,   2, -15,  -4, -13,   6,  -4,   7,  -5,  -4, -11,   0,  11,   1 },
      {  -1,   0,   2,   0,  -1,  -6,   3,  -4,  -1,   0,   2,   3,   1, -20,  -6,  -9,  -1,   1,   0,   0,  11,  -7,  20,   6,  -1,  -1,  -4, -16,   1, -22,  -8,   0,   2,   8,  10, -26,  -7,   0,  16,   0,   7,   8,  44,  -1, -16,   0,  13,  -8, -12, -11,  32,  54,   6,  47, -29,  -7, -25, -37, -26,  15,   0, -18,   3,   7 },
      {  -1,   0,  -1,   0,  -3, -13,  18,  19,   0,   0,   0,   7,  -3, -21,  35, -21,   1,   1,   3,  12,  13, -13, -20,  25,   1,   2,   0,  14,  35,  20,  19, -12,   0,  -4,  -8,  -8,  29,  49, -24, -14,  -1,   3, -10,  -4,   7,  38, -14,   4, -11,   3, -12, -11,  -1,  16, -14,  -5, -12,  -8,  -7, -20, -22, -37,   9,   4 },
      {   0,   1,   1,   4,   9,  33, -19, -44,   0,   0,  -1,   0,   6,  31,  -5,   0,  -1,  -1,  -2,  -5, -18,   8,  26,  -2,  -1,   1,   9,   3, -19, -25, -14,  10,  -1,  -1,  12,  28,   0,  15,  -6,   6,  -2, -13, -11,  30,  23,  18,   5, -12,  -7,  -2, -39,  -2,   5,  32, -15,  -2,  -4,   9, -24, -31, -18, -34,   6,   5 },
      {   1,   2,   1,  -1,   4,  33,  39, -16,   1,   0,   0,   1,   1,  15,  -4,  -7,  -1,  -2,  -3,   1,  -3, -34, -18,  35,  -1,   1,  -5,   8,   2, -25, -45,  -5,   0,   2,  -6,   2,  30,  -1,  65, -16,   1,  -5,  -7, -30,   7,  -2, -11,  -3,   5,   9,   9, -12,  -9,  -8, -12,  11,   1,  -5,  14,   2,   3,  -4,  12,  -5 },
      {  -1,  -1,  -2,  -1,  -1,  -8, -26,   6,  -1,  -1,  -1,  -1,   1, -10, -45,  35,   1,   2,  -1,   6,   3,   2,  32, -20,   1,   2,   1,   3,  17,  22,   7, -25,  -1,  -3,   1,  -7,   3,  25,  43,  12,  -3,  -8,  -6,  -7, -18, -10, -66,   3,   1,   3,  -8,   5, -10,   1,  13,   5,   2,  -2,  -7,  -6,  -6, -25,  19,  -9 },
      {  -1,  -1,  -1,   0,  -6, -10,   1,  10,  -2,  -2,  -1,  -2,  -4, -13, -27,   5,  -3,  -3,  -2,  -3,  -4,  -8, -41,  -2,  -2,  -3,  -3,  -4,  -6, -20, -40,  -7,  -3,   0,  -2,  -1, -10, -20, -43,  24,  -3,   1,   1,   2,  -1, -29,   3,  -3,  -6,  -7,   2,  -4,  -5, -36,   6,   0,  -2, -20,  -1,  -8, -12, -70,  21,  -5 },
      {   2,   2,   1,   4,  -5,   6,  72,  16,   2,   1,   0,   2,   2,   5,  36,  39,  -1,  -2,   0,  -2,   0,  -4,   4, -45,   0,  -3,   1,   0,   2,  -3, -28,  24,   1,  -1,   0,  -1,  -5,  -3,  -5,  21,   3,   0,   3,   7,  -5,  13, -24,  -8,   3,   2,  -6,  11,   7,  23,  44,  -3,  -2,   1,  -9,  -6,   0,  -4, -17,  -2 },
      {   0,   0,   1,   0,  -1,   2,  18,  -6,   1,   1,   1,   2,  -3,   9,  56, -34,   0,  -1,   0,  -3,  -6,   9,  62,  11,  -1,  -2,  -2,  -5, -12, -12,  16, -14,  -1,  -3,   1,  -4,  -4, -33, -16, -13,  -3,  -5,  -1,  -3,  -3, -38, -44,  16,  -3,   0,   2,   0,   0, -23,  -9,   7,   1,  -7,   4,  -3,   1, -18,  18, -10 },
      {   0,   0,   0,   0,   0,  -6,  11,  17,   1,   2,   2,   1,   1,   1,  37,  58,   0,   0,  -1,   0,  -1,   1,   7, -27,   1,  -1,   1,  -1,  -5,  -6,  19, -40,   1,   2,   2,   6,   2,   7,  19,  17,   0,  -4,  -3,  -5,   4,  -8,  45, -27,  -1,  -4,  -4,  -8, -10, -19, -60,  11,   3,   6,   3,   3,  -3,  -8,  19,   0 },
      {   0,   1,   1,   2,   0,  -3,  -1,   6,   1,   1,   2,   1,   1,   3,   6,  12,   0,   0,   0,  -2,  -4,   5,   9,  52,  -2,  -3,  -1,  -2,  -9,  -8, -28, -82,   2,   1,   4,   5,   3,  -4, -15,  49,  -2,   1,  -1,   7,   2,  14,  -2,  16,   2,  -1,   0,   2,   6,  16,  23, -25,   0,   4,  -4,   3,  -1,  27, -11,   6 },
      {   0,   0,   0,  -2,   1,   0,  -5,  12,   1,   1,   1,   0,   4,   0,  -4,  77,   0,   0,   0,   1,  -2,  -1,   9,  37,  -1,  -1,  -3,  -2,  -5,  -5, -18,  15,  -1,   0,  -1,  -3,  -5,  -5, -28, -72,   1,   2,   4,   5,   2,  -1,  -6,  40,  -2,   0,  -3,   3,   0,   6, -10,  -6,   2,   2,   0,  -1,   1,   4,   9,  13 },
      {   0,   0,   1,   2,   1,   5,   1,  -8,   0,  -1,  -1,  -1,   0,   1,  -8, -20,  -1,  -2,  -1,  -4,  -3,  -6, -29, -57,   0,  -1,   0,  -3,   0,  -3, -35, -28,   0,   0,   2,   0,   3,   4, -34,  -8,   0,  -1,   0,   1,  -1,   4, -43,  26,   1,   8,  -1,   6,   3,  17, -58,  16,   1,  12,   3,   8,   8,  37,  12,   9 },
      {   0,   0,   0,   0,  -1,   0,   5,  -8,   0,  -1,  -1,   0,  -1,   2,  -1, -14,  -1,  -1,  -2,  -1,   0,   0,   3, -46,  -1,  -1,  -1,  -2,  -2,   1,  -1, -38,   1,   0,   2,   1,   1,  -1,  29, -38,   0,   1,   1,   1,   4,   0,  24,  48,  -3,  -9,  -3,  -8,  -5, -18,  10, -48,  -3,  -8,  -2,  -5,  -4, -32, -54,  19 },
      {   0,   0,   1,   0,   0,   4,   4,  -4,   1,   1,   1,   1,   2,   0,   6,  -5,   0,   0,   2,   1,  -1,   0,   0, -15,   0,   1,   1,   2,   2,   1,   4,   7,   1,   2,   4,   0,   2,   6,  14,  18,   2,   4,   5,   8,   2,  11,  40,  79,  -1,   1,  -3,   5,   0,  10,  22,  10,   4,  -3,   1,  -1,   1,   3,  67, -45 },
      {   0,   0,   0,   0,   0,   1,  -4,  -2,   0,  -1,  -1,  -1,   1,   1,  -4,  -4,  -1,  -1,  -1,   0,  -3,  -2,   1, -13,  -1,  -1,  -2,  -1,  -2,   2,  -8, -49,   0,   0,   0,  -2,  -2,  -3, -14, -56,   3,   2,   5,   6,   5,   9,  16, -38,   0,   2,   2,   6,   7,  12,  32,  46,  -2,  -5,  -4,  -3,  -2,  -1, -11, -69 },
      {   0,   0,   0,   1,   1,   0,   0,  -5,   0,   0,   0,   0,  -1,  -3,   4, -11,   0,   0,   0,   0,   0,   0,  -1, -21,   0,   0,   0,  -1,   1,  -4,  -2, -20,   0,   0,   1,   0,   1,  -2,  -1, -34,   0,   0,   0,   1,   3,  -1,   7, -43,   4,   3,   5,   0,   6,   8,  39, -45,   7,   5,  11,   3,   7,   9,  77,  48 },
      {   0,   0,   0,  -1,   0,   0,  -3,   1,   0,   0,   1,   0,   0,   2,   1,   4,   0,   0,   0,   0,   1,  -1,  -5,   0,   0,   0,   0,   1,   0,   0,  -6,  16,   0,   0,  -1,   1,   0,   0,  -5,  -1,  -1,   0,  -3,  -1,  -2,  -2, -15, -19,  -2,  -2,  -2,  -3,  -4,   0, -28, -91,   4,  -1,   3,   2,   2,  12,   3, -78 },
    },
    {
      {  85, -61,   4,  -2,   2,  -1,   1,   0, -61,  29,  15,  -4,   1,  -1,   1,   0,   4,  16, -15,  -2,   1,   0,   0,   0,  -1,  -6,  -2,   6,   0,   0,   0,   0,   1,   1,   2,   0,  -2,   0,   0,   0,  -1,  -1,   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  44,  21, -39,   1,  -3,   0,  -1,   0,  14, -71,  35,  14,  -2,   2,   0,   1, -36,  36,  28, -29,   3,  -2,   1,  -1,   3,  14, -33,   2,  10,  -3,   1,   0,  -3,  -3,   3,  12,  -5,  -1,   0,   0,   2,   2,  -1,  -3,  -2,   1,   0,   0,  -2,   0,   1,   1,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
      {   6,  66, -58,   0,  -1,  -1,  -1,   0, -62,   7,   7,  23,  -5,   4,  -1,   1,  48, -26,   7,  -2,  -9,   4,  -2,   1,  12, -21,  14,  -6,   4,   2,  -1,   1,   0,   0,   7,  -6,   3,  -2,   0,   0,  -1,   0,  -1,  -1,   3,  -1,   1,   0,   0,   1,   1,   0,  -1,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0 },
      { -50,  -4, -13,  18,   5,   5,   0,   3, -14,  10,  52, -19,  -6,  -4,  -1,  -1,   5,  48, -51, -22,  16,  -2,   3,  -1,  16, -29, -18,  41,  -3,  -4,   0,   0,   1,  -7,  21,  -5, -18,   5,   0,   1,   1,  -2,  -3,  -6,   8,   5,  -2,   0,   0,   2,   1,   0,   1,  -3,   0,   1,   0,  -1,  -1,   1,   0,   0,   1,   0 },
      {  36,  21, -34,  22, -12,  -2,  -1,  -1,  32,   2,  -5, -31,   3,   3,   0,   0, -44,  -7, -32,  43,  13,  -3,   1,   1,  29, -20,  39,   9, -32,   5,  -4,   1,  -8,   4,   7, -36,  12,   7,   0,   0,   2,   3,  -4,   8,  10,  -9,   0,   0,  -2,   0,  -1,   0,  -5,  -3,   3,   0,   1,  -1,   0,   1,   0,   2,   1,  -1 },
      { -28, -38, -42,  59,   4,   6,  -3,   4,  -6,  35,  17,  -8, -23,   1,  -1,   2, -26,  -6,  45, -18,  -1,   8,  -2,   1,  28,  10,   4, -35,  10,  -3,  -2,   1,  -1,  -4, -26,   3,  18,  -5,   5,   1,  -2,   4,  -2,  14,  -4,  -5,   1,  -3,   1,   1,  -1,  -4,  -3,   3,   0,   2,   0,  -1,   2,   2,   1,   0,  -1,   0 },
      { -16,  24,  20, -23,  -8,  -1,   0,  -2, -41,  17, -31,  11,  21,  -2,   3,   0, -49,  44,  10,  22, -20,  -3,  -2,  -1,  54,  -3, -22, -13, -12,  10,  -4,   1,  10, -34,   1,  12,   8,   0,  -2,   2,   4,   0,  15,  -6,  -6,  -3,   1,   0,   0,   0,  -3,  -3,   5,   2,   0,   0,   3,  -1,   1,   1,  -1,  -1,   0,   0 },
      {  22,  12, -22,   2,   3,  -3,   4,  -3,  27,  52, -35,   7, -16,  -1,  -3,  -1, -24, -35,   4, -26,  16,  10,   1,   2,   7,   2, -28,  44,  18, -12,   0,  -2,  14, -22,  22,  20, -38,   0,   2,   0,  -4,  -5,  16, -24,  -1,  14,   0,   2,  -1,   0,   1,  -1,   9,  -3,  -3,  -1,  -1,   0,  -1,  -1,   0,  -1,   1,   1 },
      {   0, -30,  -1, -13,  15,  13,  -2,   6,  25,  -6, -18,  44,  -6, -15,  -3,   2,  -1,  12,  14, -19, -33,  17,   0,   3,  18, -58,  21,  -5,  17,  12,  -6,   4,  -7,   1,  41, -21,   7,  -9,   0,   3, -14,  22, -26, -15,  22,  -3,   3,  -7,  -5,   8,  -3,  16,   0,  -8,   2,   0,  -2,   3,   2,   2,  -5,   2,   1,  -2 },
      { -19, -12,  -1, -11,  14,   1,   0,   2, -29, -10,   9,  30, -12,   0,   1,  -1, -28, -11,  20,   3,   8,   0,   0,   0, -13,  48,  23,  18, -31,  -8,   1,   0,  34,  -5,   5, -46, -22,  27,  -3,   2,  -1, -14, -10, -15,  45,   0, -10,   3,   2,   4,   2,  12,   0, -19,   7,   0,  -2,   4,  -4,  -1,  -3,   4,   4,  -3 },
      {  -1,   5, -37, -35,  42,   1,  -1,   2,  17,  26,  13,  21, -11, -21,   4,  -3, -24,  20, -39,  11, -37,  19,   5,   1, -36,  11,  15,  -8,  13,  16,  -9,   2,  24,   3, -27,  12,   7, -12,  -2,   3,   6, -24,   2,  19, -13,   6,   2,  -2,   2,   1,  15,  -9,  -4,   6,  -4,   2,   0,   0,   2,   0,   5,  -1,  -2,   2 },
      {   1,  11,  -5,  35, -62,  11,   5,  -6,  -4,  11, -46,  15,   7,  14,  -8,   1,  -4,  46,  -7, -31,   5,  -4,  -6,   3, -41,  -9,  14,   0,   9,   6,   0,   0,  33,  20,  -8, -10,   4,  -8, -10,  -1,   9, -24,  -8,   6,   4, -10,   4,   7,   1,  -4,   6,   7,  -8,  -5,   6,  -4,   2,   0,   0,  -6,  -3,   4,   1,  -1 },
      {   9,  37,  44, -20,   7,   6,  -8,   4, -16,  13,   8,  -9, -13, -10,   2,   1, -32,  -7,  -1, -37,  24,  17,   2,   1,  24,  -3,  19,   5,  30, -26,   0,   0, -16,  31, -27,  -9,  -6,  -9,  16,  -1,  -2,  -2, -41,  30,   4,   5,  -1,  -8,   1,   0,  10,  14, -16,   5,  -2,   3,   1,   1,   2,  -2,   1,   4,  -4,   0 },
      {  19,  14,  11,  -6,  14,  -3,  -3,   0,  24,  13,  -5,  -1, -27,   4,  -3,   0,  32,  11,  -9, -26,   8,   2,   1,   2,   7,   6, -28,  -7, -10,   2,  -1,   2,   4, -55, -14, -17,  37,  25,  -6,   5, -13,   0,  10,  31,  31, -38,  -4,  -1,  -1,   4,   7,   2, -31, -11,  18,   2,  -3,   3,  -3,  -3,   1,  11,   3,  -5 },
      {  -7, -18,   0, -32, -39, -10,  24, -20,  14,   6,  34,  39,  -2,  19, -15, -15,  -2, -23, -17, -12,   2,   1,  -9,   5,  41,  11,   2,   2,   9,   7, -12,  -4,  -9,  25,  -7, -11,   7, -19, -12,  -5,  15, -15,  36,  -3,  -5, -21,  15,  18,   3,  -5, -12,  -1, -14,  -6,   8, -14,   9,  -6, -13, -11,  -2,  12,   3,   1 },
      {  -3,  21,  11,   1,  19, -25,   8,  -9, -10,  28,  20, -13, -10,   7,   0,  -2, -29, -10,  18,  -1,  -6, -12,   0,  -3, -38,  -4, -30,   6, -10,  31,   5,  -1,  -1,  30,  35,  -4,  29,  -4, -28,   2,   6,  21, -22, -12,  -8, -33,  10,  11,   8, -17, -19,  15,  -9,  13,  16,  -8,   3,  -4,   9,   5,  -6,   6,  -7,  -1 },
      {   4,   3, -17, -19, -25,  34,   6,   4,  16,  26,  20, -31,  66, -34, -12,   1,  12, -11,  14, -10, -27, -17,  19,   0,   9,  18, -23,   1,   3,  19,  -1,  -5,   4,   3, -15, -12,  -2,   2,  -6,   7,  -7,  -5, -26,  -2,  24,   7,   0,  -3,   0,  -2,   2,  20,   2, -14,  -3,   1,  -6,   2,   3,   0,  -9,   1,   4,   1 },
      {   8,  18,  37,  20, -18,  11,  -2,   2,  -9, -14,  35, -20,  -4,  -9,  -6,   1, -18, -28, -16, -18, -13,  19,   7,   4, -12,  -4,  28, -14,  21,  17,  -9,   2,  45, -35,   6,   3,  16, -18,  -4,   5, -25,   6,  25, -23,   1,   0,   7,  -4,  -6,  22, -19,  -9,  19,  -3,   0,  -1,  -3,   2,  -8,  12,  -1,  -5,   3,  -1 },
      { -23, -29, -30, -19,   1,  26,   8,   5, -31, -34, -30,  -4,  26,  -5,  -6,   2, -27, -33, -17,  -8,  28,  -3,  12,   4, -13,  -6,   0,  29,  12,   3,   3,  -2,   3, -13,   3,  17,  23,  -4,  -6,   2, -20,  24,   3,  29, -10, -21,   7,  -5,   7,   7,   6,  -2, -30,   9,   8,  -2,  -3,  -2,   2,  -2,   1,   8,  -6,   1 },
      {  -4,   4,   4,  11, -15,  17, -10,  10, -13,   1,   2,   1,  -4, -13,   4,  -6, -26, -16, -16,   8, -19,  14,  -1,   3, -26, -13, -12, -21,  19,  -5,   3,   4, -49,   9,   6,  26,  -7,  31,  -5,  -5,  32,   1,  26,  -3,  36,  -5, -35,   6,  -3,   0,   1,  -8,  -9, -40,  17,  10,   2,  -1,  -4,   1,  -6,  13,  19,  -7 },
      {   3,   4,  11, -10, -15,  23, -13,  16,  10,  21,  29,  45,  -1,   6,  -9,  -7, -17, -13,   5, -13,  21, -51,   8,   1, -12, -32,  11, -13, -38,   9,  32,  -1,  -9, -11,  -6,  20,   5,  24,  -6,  -6, -22, -10,  -5,   6, -14,  12,  -7,   2,   2,   7,  19, -16,  11,  13,  -6,   5,  -6,   1,   0,  -1,   8, -10,  -2,   1 },
      {   3,  -3,  15,  33,  -5,  21, -18,  16,  -5, -14,  -1,  16, -19,   5,   3,  -2,   2, -23, -25,  21, -11,  -8,  15,   0,  22,  30, -24,  -6,  19,  21,  -3,   1,  -1, -10,  42,  -1,   6, -15,  -9,   1,  21, -35, -30,   8,  -3,   2,  13,   0,   6,  -1,  44,  11,  -5,  11,  -6,  -3,  -8,  18,   4, -22,   5,  -7,  -6,   2 },
      {  -1,  -9,   3, -10,  25, -19,  -5,   6,   1, -14,   2, -35,  12,  15,  13,  -6,   2, -14,  32,  -5,  -4,  -6,  -6,  -2,  18, -47,   0,   9,   6,   8,  10,   2,  48,  25,  -3,  13,   0,  10, -13,  -7,  16, -47,  11,   8,   5, -14,  -9,   9, -13,  14,  33,  -9,  -3, -13,  10,   0,   1,  13, -15, -12,   0,   2,   9,  -6 },
      {  -2,  -6, -22, -43,  -9,  17, -16,  19,   9,  11,   3, -19,  -9,  25,   5,  -3,   6,   8, -15,   8,  38, -14, -15,   3,   5,  14,  11, -47,  21, -21,  14,   8,  30,  -7,  26,  22,   3,  -3,   3, -17,  23,  11, -23, -17,   6, -15,  -3,  -3, -18,  -3, -19,  21,   9,  -2,   8,   0,   1,   7,   3,  13, -10,   2,  -1,  -3 },
      {  -9,  -7,  -8, -10,  -5, -40,  26, -13,  -9,  -5,   2,  -9,  -1,  46, -12,  -5, -20, -12, -12,  10,  -3, -11,  -6,   6, -21, -22, -24, -22,  23,  -1,   4,   3, -13, -23,  -6, -17,   5,  -6,  22,   2, -35, -18, -12,  11,  16,  32,   8,  -4,   8,  -8,  10,  30,  14, -22, -19,   2,  -5,  -1,   3,  -8, -18,  -7,  16,   3 },
      {  -1,   2,   3,  -8, -33,  10,  -5,  -1,   6,  16,  29,   6,   6, -10,   7,   2,  -3,   1,  29,  42,  -7,  19, -18,  -5, -26, -18, -25,   2,   9, -43,  -2,  12,  19,  -6,  28, -27,  -6,  -1,  29,  -8,  -3,   1,  11,  24, -11,  -3,   3, -12,   6,  22,   1,  -9, -32,  16,  -5,   0,  -6,   1,  -8, -12,  16,  13, -16,   5 },
      {   7,   5,   9,  37,  18,  -5,   7, -12,   6,  -6,   9,  27,  15, -10,  -1,  -9,   5,  -1, -16,  32,  -7, -14,   0,   3,  16,  -4,   1,  -5,  15, -18,  13,  -4,  28,   6, -18,  34, -24,  10,   2,  -7, -41,  -2, -10, -11,  -1, -12,  -9,  11,   3,  -9, -12,  33, -30,  -1,  28,  -2,   8, -16,  22,  -7, -19,  27, -12,  -5 },
      {   4,  12,   6,   9,  24,  -6,  -7,  10,   5,  20,  10,   3,  23,  32, -18,  -9,   4,  28,  23,  11,   4, -33,  19,   4,  -5,  24,  39,  17,  30,   5,  -5, -11, -14,  -8,  24,   7,  -9, -24,  -2,   2,  -2,  22,  14,  12,  -4,  -2,   2,  -4,   8,  19,   7, -13, -28, -31,  -6,   4,  -9,   4, -22, -11,  -7,   6,  29,  -1 },
      {  -1,   0,  -5, -18,   2,  45, -30,   0,  -5,  -4, -14, -19, -49,   3,  -6,   8,   2,  13,  11,  25,   8, -18,  25, -10,   0,  -9, -20,  -3,  15,   7,  -8, -16, -10,  27, -11, -20, -13, -12, -16,  10, -44, -27,  24,  -9,   2,  -2,   8,  14,  14,  13, -22,   3,   3,  -2,   6,  -5,   0, -12,  -6,  11,  -4,   2,   4,  -1 },
      {   1,  -1,  -1,  -4, -18,  -4, -12,   3,   4,  -1,  -4,   2, -15,  16,  10,   7,   3,  -5, -18,  -3, -23,   2,  -9,   1,  22,  17, -21,   2,  -2,  13,   6,   2,  40,  41,   8,   2,   5,  13,   0, -17, -32,  44,  -9,  18,  -1,   7, -34,  -8,  26, -11,   4, -27,  -1, -16, -21,  26,   5, -13,  -4,  -5,  12, -15,  27,  -1 },
      {   4,   1,  10,   4,  19,  25,  27, -21,   1,  -5,   5,  11,  -4, -22, -32,  14,   3,  -4,   9,  10,  36,  -5,  24,  -6,   0, -22,  -8, -14,   0,   3, -22, -10,  29,  -3,   8, -10,   6,  -9,  24,   6,  31,  -6,   2,   9, -14,   9, -13,  -6,  12, -57,  -8, -16,   1, -22,   1,  16,  -6,   2,  12,   4,   2,   4,  24, -12 },
      {   2,  -2,   3,   9,  10,  -8, -18,  -4,  -1,  -8,  -8,   7,  -4,   4,  -2,   0,  -4, -14, -13, -10, -22, -42,  29,   3,   2, -16, -11,  -2,  18,  -3,  15, -14,  22,  -5, -19, -27, -11,  20,   1,  -4,  57,  24,  13,  18,   1,  14,   4,  -3,   0,  21, -35,  18, -11,  13, -31,  -7,  13, -12,  -1,   7,  -4,  -2, -12,  22 },
      {   2,   0,   7,  -2, -15,  15, -28,  22,   4,   3,  18,  12, -10,  17,  24,  -6,   4,  -3,  18,  24,  14,  12, -13,  -8,  -9, -19,  -2,  32, -10,  18, -36,   6,  11, -12, -35,  19,   3, -20,  -8,  11,  16,  30,   6,   3,  17,   2,  11,  -6,   0, -21,  20,  29,   3, -12, -15,  -7,  27,  -5,  21,  -8, -27,  -8,   7,  11 },
      {  -1,  -2,  -5,   7,   1, -34, -34,  -4,   3,   5,  -3,  20,  30, -26,  56,   1,  -1,  -6, -12, -18,  35,  -6,  12, -33,   0,   2,  -8, -26, -11,  12, -27,   2,   1,  13,  19,   1,  -1, -13,  18,   7,  -9, -14,  15,  19,   6,  -1,   1,   0,  24,  18, -16,   8,   7,  -9,  -1,  -1, -16,   3,  -2,  11, -16,   5,   6,  -2 },
      {   3,   3,   4,  18,  25,  17, -20,  35,   3,   4,  -2,  27,  34,  32,  -3,   0,  -3,  -6, -12,   2,  -8,   1, -10,  -1,   1,   4, -24,  27,  11, -33,   8,  14,  -1,  18, -17, -26,  47,  -4,  16,   0,   0, -11,  10, -15,  -9,  -4,   2,  -9, -10,  -9, -11,  -4,  25,  -5,  14,   2, -10,   6,  -5,  34,  -4,  -5,   6, -16 },
      {  -6, -10,  -6, -13, -23, -27, -45,  14,  -3,  -6,   0, -13,  -6,   8,  13,  -3,  -5, -11,   0, -17, -36, -22,  17, -15,  -5,  -6,  12,   7,  -2, -15, -14,  -2,  -5, -30,  -7, -13, -18, -12,   6,   5,  -2,  -8, -21, -13, -21,  -2,   3,   8, -10, -51,  -6, -21, -19,   6,  26,  14,   0,  -8,   4,  -2,  14,  29,  14, -14 },
      {  -2,   1,  -1,   6,   0,   2,   1, -25,  -4,   2,  -4,  10, -10,  -5,  -1,  22,  -4,  -8, -13,   5,   2, -19,  -1,  -1,  -5,   3, -26,  -6, -20,   8, -11,  -7,   3,  12, -11,   2,  -6, -23,  27,  12,  -4,  21, -20,  -4,  -4,  -3,   8, -10, -75,  33,  17, -12,   1, -16,  10,   1,  28,  10, -31,   0,  -6,  21,  11, -11 },
      {   1,   0,   1,   3,  -8, -24,  25,  65,   0,  -4,  -6,   4, -10,  -2, -23, -36,   3,   1,  -1,  14,  13,  10,  18,   6,   3,  -9, -19,  -8, -13,  18, -10,   8,  11,  12, -16,  15, -18, -10,   3,   8,   6,   6, -20,   0,  25,  11,  21, -13,   7,  11, -14, -17, -10,  -1,   6, -12, -28, -12, -19,  22,  26,  21,  -6,  -5 },
      {   3,   5,   4,  11,  34,  26,  -8,  -2,  -1,  -3,  -5,  -3,  26,   2,  10,   0,  -7,  -8, -13, -13,   7, -15, -59,  20,  -6,  -8, -12, -31,  -2,   6,   3,  10,  -2,  -7,  -8, -22, -37, -25,  -8,  -3,   0,  10,   2,  -6,  19, -11,  26,  11,  19, -17,   3, -24,  -4,   2,   1, -15,   7, -21, -12, -20,  26,  -6,   8,  11 },
      {  -1,   3,   0,   6,   4,  -4,  18,  28,  -3,   3,  -4,   4,   6, -25,  26, -40,   0,   8,   1,   3,  21,   2,  20, -10,  -4,   3,  -2, -12,  22,   1,   2,   4,  -2,   5,   5, -24,   8,  18, -13, -11, -26,   1,  18, -11, -11,   3,  -4,   6, -48, -39,   1,  10, -12,   7, -22,   2,  28,   8, -29, -17,  12, -17,  11,  28 },
      {   1,   3,  -2,  -3,   9,   3,  -6,  24,  -3,  -3, -10, -19, -12, -16, -30,  -5,  -2,   0,  -8, -26, -20, -13,   1,  -7,  -1,   4,   3, -12, -43, -27,   7,  13,   6,  21,  14,  13,  21, -16,  32,   2,  -3,   2,  35, -22,   4,  15,  16, -16,  26,  -8,  12,  24, -18,  -4,  -7, -10,  23,  16,  -7, -36, -17,  13,  -3,   1 },
      {   0,  -1,   2,   2,  10,  -2, -19,  14,   2,   2,   3,   4,  30,  35, -36,  18,  -1,  -5,  -4, -10,   5,  44,  18,  -9,  -1,   1,  -5, -21, -23,  -9, -10, -19,  -1,   5,  22,   9, -13,   5, -14,  24,  -7, -26,  -4,  10,   5, -10,  -4,  -8, -24,   6, -16,  -7, -13,  10, -21,  10,  34, -36,  24,  -7,   2,  -7,   6,  20 },
      {   2,   0,   3,   3,   0,  11,  21,  22,  -3,  -7,  -6,  -5, -16, -23,   5, -31,   3,  -5,   7,  -7, -12, -19, -37,  16,   1,   7,   1,  16, -15,   9,   2,  18,  -7, -11,  17,  -4,  -8, -19,   0, -13,   1, -33,  -7,  35, -25,  -7, -11, -12, -10,  16, -33,   2,   1,  -5, -10,  20,  20, -19,  21,  17, -36,   6,  31,   0 },
      {  -1,  -2,  -2,  -2, -12, -43,  -8,  26,  -1,  -1,   0,   6,  -1, -46, -37,  20,   0,   1,  -4,  11,   9, -13,  -5,  -2,   1,  -1,  -7,  -9,  18, -32,  11,  -4,   5,   1,  -1, -18, -10, -11, -36,  10,  -3,  18,   7,  13,   1, -17,   9,  14,   4,   5,  31, -18,  23,  -4,  12,  -8,  19,  11,  24,  12, -23, -25,  11,  -4 },
      {   0,   1,   1,  -4,  -1,  -2,   7,  10,   2,   4,   1,  -4,   7,  13,   1,  -3,   1,   5,   5,  -4,   3, -19,   6,  19,   0,   8,   9,  -1,  28,  33, -22,  13,  -3,   7,  15, -10, -11,  29,  38,   7,  -8,  -1,  20,  12,  11,  -2,   0, -21,  11,  -5,  16, -12,  23,  20,  24,  -2,  62,  -2,  20,  13,  12,  15, -33, -22 },
      {   1,  -3,  -3,  -1,  -3, -10,  -7,  -8,   3,  -1,   0,   3,   3, -19,   5,   1,   5,   1,   4,  10,  20,   8,   3,   3,  -1,  -2,  -1,   7,  11,  -4,  13,  -3,  -2, -18,  -6,   0,  36,   2,  -5,  -2,  14, -16, -33, -35, -10,  18,   0,   2,  41,  21,   6,  -5,  -8, -11,   2,   8,  52, -52, -42,   1,   2,   2,   9, -10 },
      {   3,   3,   1,   1,  10,  26,  32,  -6,   2,   1,  -3,  -6,   4,  28,  12, -39,   0,  -3,  -6, -13, -13,   1,  17, -32,  -1,  -4,  -6, -20, -14, -36, -36,  11,   6,  -4,   5, -11,  -8,   2, -33,   7,   2,  21, -13,  10, -16,  20, -12,  30,  15,  16,  20, -17,   9,   1,  13,  -5,  14,   3,  -7,   9, -31,  10, -17,  -8 },
      {   1,   1,   0,   1,   1,   7,  10,  33,   1,   1,   2,   2,   7,  12,   8,  21,   1,   0,  -1,  -1,   0,  38,  10, -10,  -3,  -3,  -4,  -4, -13,  24,  46, -37,  -3, -14,  -6, -16, -26, -21,   9, -26,   3,  11,   7,  21, -21,  -2, -10,  29,  16,   0,  -5,  29,  13,   0,  13,   4,  25,  27, -17,  -2,   2,  18,   6, -25 },
      {   0,  -2,   1,   5,  -1,  -3, -22, -24,   0,  -1,   1,   1,   3,  -3, -23,   0,   1,  -2,   5,   3,  -9,   0,  -7,   6,   1,   1,   4,   7, -11,  -3,  -3,   8,   1,  -4,  17,  23,   1, -19,  -6, -25, -10, -10,   0,  33,  33,  22,  -6,  19,  -3, -33,  -2,   6, -17,  11,  -3, -14,  22,  17, -43,  65,   1, -13, -15,  11 },
      {   1,   2,   4,  -2,   0,   2,  29,  24,   1,   1,   1,  -2,  -7,   4,   8,  53,   0,   0,   1,  -4,  -2, -12, -11, -58,   5,   6,   9,   4,  11,  18,  21,  17,  -2,  -3,   5,  -6,   4, -10,   4, -14,   1,  -5,  11,  -2,   9,  14, -13,  -4, -14, -12,  14, -17, -13,  -8,   4, -18, -13, -45,   1,   9, -23,  13, -29,  27 },
      {   0,  -1,  -1,   3,   3,  -3,  -1,   5,  -2,  -2,  -2,   3,   1,   0, -40,   8,  -3,  -4,  -3,   4,  -4, -11,  -9, -51,  -3,  -4,   2,   8,   6,   7, -36,  38,  -6,  -9, -15,   9, -14,  11,   8, -23,  -1,  -5, -20,   3, -12, -36, -18,   0,  13,  14, -26,  20,  20,  -5,   8,   4,   9,  15, -27, -27,  25, -29,  -3,  10 },
      {   2,   0,   1,   2,  -7,   8,  46,   6,   3,   1,   4,   3,  -6,  10,  41,  24,  -2,  -3,  -3,   4, -18, -17,   7, -19,   1,  -5,   7,   1,   0, -25,  12, -14,  -2,   0,  -2,  20, -10,  -4,   1,  42,   2, -12,  -7,  -7,  21, -30,  16, -28,  22,  -5,  -9,  -4, -16,  13,  -6,   7,  23,  26, -11,  25,  12, -30,  25,   2 },
      {   0,   0,   1,  -1,  -3,   3,  -4,  20,   0,  -1,   2,  -2,  -1,   0,  -6,  23,  -1,  -2,  -3,  -5,  -6,   8, -21,  26,   1,   0,   3,  -9,  -1,   9, -28, -19,   2,   2,   5,  -1, -19,  54,  15,  21,  -2,   1,  -5,  -2, -46, -15,  18,   3,   9, -14,  14,  12, -11,  -6,  -3, -11, -21,  -6, -46,  20, -44,  -6,  -1,  13 },
      {   0,  -1,  -2,  -2,   2, -11,   6,  16,   0,   0,  -1,   1,   6,   5,  -4,  17,  -2,  -3,  -4,  -3,  13, -10, -10,  25,  -3,  -7, -10,  -9,  -1,  11, -33, -13,  -6,  -2, -17, -18,  -1, -26,  -9, -17,   0,  -6,   4, -42,   3,   5, -55,  -9,  10,  15,   6,  -5, -40,  31,  -3,  38,  -2,  32,   5,   9, -16,  -5, -18,  11 },
      {   0,   1,   1,   0,  -1,   2,   8,   3,  -1,   0,  -1,   0,  -2,   0,  13,  -8,  -1,   1,   3,   1, -10,   1,  41,  25,  -6,  -2,  -2,   6,  -9, -15,  -8, -23,  -2,  -2,  -1,   7,   2, -16,  44, -50,  -1,  -4, -11,  -3,  25, -45,   4,  18,   4,  -4,   5, -30,  25,   8,   1,  -7,  -5, -23,  -4, -16, -33, -25,  -7,  23 },
      {   1,   1,   0,   2,   1,  -3,   8,  22,   0,   0,   0,   2,  -2, -10,   5,   9,  -2,  -1,  -1,   3,  -5, -23, -33,   3,  -1,   0,   5,  10,   8, -19, -26, -62,   2,  -2,  10,  15,  25,  -3, -14,  20,  -5,  -6,  -5,  22,  19,  11, -23,   8,  -7, -12, -26,  -2,  30,   1, -15,  -6,  -2,  -4, -21, -36,   9,  18, -32, -19 },
      {  -1,   1,   1,   1,   0,   0,   9,  -1,  -2,  -1,   0,  -4,  -2,  -4,  -3,  -9,  -2,  -1,   0,  -6,  -7, -12,   0,  -7,  -3,   1,   3,  -6,  -8, -12, -15, -14,   0,   2,   7,  -2,  -7, -15,   4,  -7,   6,   9,  27,   0,   0, -20, -11,  -1, -12,   3,  44,  54,   5,  17,  -5,  25, -21, -59,  -6,  26,  32, -19,   5, -37 },
      {   0,   1,   2,   1,   3,  -2,  17,  14,   1,   3,   3,   3,   1,   2,  24,  57,  -1,   0,   2,  -3,  -2,  -7,  27,  44,  -1,  -2,   0,  -3, -12, -23, -32,  30,  -3,   2,   4,   6,   2,   6, -40, -44,  -5,  -5,   4,  11,  -9,  16,  28,  -7,  -3,   2,  -8,  16,   6, -23,   4,   6,   4,  -2,   2,  -6,  22,  13,   2,  -3 },
      {   0,  -1,   0,   3,   1,  -1,  -8, -18,   0,  -1,   1,   5,   3,   2,  -8,  -1,   0,   1,  -1,   5,   4,   9,  12,  -3,   0,  -2,   0,  -2,   4,  -1,  18,  22,  -3,   0,  -7,  -3, -16,  -4, -34,  -3,   0,   1,   6,   6,  14, -23,   2, -56,   7, -21,   5,  -4,  32,  36, -21,  27, -10,  -3, -40, -14, -38,  42,  -2, -15 },
      {   0,   0,  -1,  -2,  -2,   0,   2,   8,   1,   2,   2,   2,   1,   3,   3,  19,   1,   1,   2,   2,   2,   3,  12,  16,   1,   1,   0,   1,   2,   7,   2,  36,   3,   2,   1,   2,  -1,   0,   6,  21,   6,   3,   3,  12,  15,  27, -14,  21, -14,  -1, -24,  -1, -19,  46,  24, -19,  -9, -14, -15, -34, -28, -44,  23, -52 },
      {   0,  -1,  -1,  -2,  -4,  -4,  -8,  -5,   0,   0,   1,   0,  -2,   4,  11, -14,   0,  -3,   0,  -2,  -2,   3,  15,  23,  -1,  -3,  -4,  -5, -12,  -3,  18,   1,   0,  -4,   0,  -9, -15, -28, -13,  19,   0,  -2,   1,  10, -24,  -6, -34, -53,  -1, -11,  -6,   6,  -2, -48,  31, -25,   0,  -9, -11,  -8,   2, -42, -48,   1 },
      {   0,  -1,   0,  -1,  -2,  -2,   4,   4,   1,   0,   1,   1,  -2,   0,   3,  10,   3,   1,   2,   2,   1,  -4,  12,  31,   2,   2,   2,   4,  -3,  -7,   9,  31,   4,   2,   4,   2,   5, -19,  11,  54,   3,   5,   4,  10,  -1,  -6, -36,  35,   8,   1,   7,  20,  30,  12,  21,  15,  -4,  -8, -14,  -2,  16,  28,  12,  64 },
      {   0,   0,   0,   0,   0,   0,  -1,   4,   0,   0,   0,   1,  -1,  -1,   1,  -1,  -1,   0,   1,   1,   1,  -1,  -1,   0,   1,   1,   1,   2,   0,   5,  15,  18,   1,   3,   0,   3,  -3,  -4,   7,  20,   3,   1,   4,   5,   7,  -2,  39,  37,  -6,  -5, -11,  -8,  -9, -28,  -8,  85,   0,   0,  -7,  -7, -12, -22, -55, -13 },
      {   0,  -1,   0,   0,   1,  -2,   0,   4,   0,   0,   0,  -1,   0,  -2,   4,   9,   0,  -1,   0,   0,   1,  -2,  12,  14,  -1,  -1,   0,   0,   0,  -2,   5,  36,  -3,  -4,  -4,  -1,  -4,  -8,   7,  17,  -6,  -5,  -8, -14,  -9, -37, -29,  31,   2,  -4,   3,  -7,  -7, -20, -79, -37,   4,   4,   7,  12,   6,  17, -17, -39 },
    },
  },
  { // 10
    {
      {  75, -53,   4,  -2,   2,  -1,   0,   0, -70,  40,  10,  -3,   1,  -1,   1,   0,  17,   9, -23,   5,  -1,   1,  -1,   0,  -1, -15,   8,   6,  -3,   1,  -1,   1,   2,   4,   3,  -6,   0,   0,   0,   0,  -1,  -2,  -2,   1,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
      {  31,  13, -22,  -4,  -2,   0,  -1,   0,  -1, -55,  39,   5,   3,  -1,   1,   0, -35,  65, -10, -24,   5,  -4,   2,  -2,  11, -11, -36,  32,  -3,   2,  -1,   1,   3,  -9,  28,  -5, -12,   5,  -3,   1,   1,   1,  -7,  -9,  10,  -1,   0,   0,  -2,   1,   2,   5,  -2,  -2,   1,  -1,   0,   0,  -2,  -2,   0,   1,   0,   0 },
      {  37,  -9,  17, -10,  -4,  -3,   0,  -1, -10, -15, -23,  13,   6,   3,   0,   1,  -1,  -9,  56, -17,  -6,  -3,   0,  -1, -22,  45, -47, -18,  23,  -5,   5,  -3,   4,  -9, -10,  48, -20,  -1,  -1,   0,   4,  -4,  16, -23,  -9,  11,  -4,   2,   0,  -2,  -3,   3,  13,  -5,   0,   0,   0,   1,   1,   0,  -4,  -1,   2,   0 },
      {  51,   1, -11,  12,  -2,   0,  -2,  -1,  19, -46,  14, -19,   1,   2,   1,   1, -26,   1,  14,  21,   3,  -1,  -1,   0,   7,   4,  32, -43,  -4,   2,   0,   0, -20,  23, -39,   2,  35, -12,   6,  -4,   2,   3,  -2,  32, -29,   0,   1,  -1,   0,  -4,   4, -15,   3,  11,  -4,   2,   1,   0,   0,   4,   5,  -7,   1,   0 },
      { -49,   0,   2,   8,  -2,  -3,  -1,  -1, -30,  16,  19, -16,  10,   2,   3,  -1,   7,  50, -23,   9,  -7,  -6,  -1,  -2,   9, -17, -24, -16,  11,   6,   3,   1,  -7,  11, -19,  50,   3, -14,   2,  -4,  11, -10,  15,   1, -46,  18,  -3,   3,   4,  -7,   6, -19,  22,   8,  -9,   3,   1,   0,   0,   4,   0, -10,   5,  -1 },
      { -39, -34,  19,  -9,   9,   0,   0,  -2, -30,  -4,  20,   4,  -4,  -5,  -1,   0,   5,  27,  31, -22,  14,  -1,   0,  -3,   3,  33, -21, -17,  -6,  -7,   0,  -2,   0, -11, -24, -20,  31,  10,   3,   2,  -6,  18, -16,  39,  13, -30,   5,  -6,   1,   2,  -4,  10, -38,  13,   5,  -2,   2,  -2,   3,  -5,   8,   7,  -9,   3 },
      { -28, -34,   4,   7,  -9,   0,  -1,   2, -19, -28,  29,  -7,  12,  -1,  -3,  -1,  -3,   9,  20,  10, -15,   1,  -3,   1,  16,  10,  37, -24,   9,   2,  -6,  -2,  -3,  14, -13, -24, -19,   0,   2,   1,   5, -16,   1, -35,  39,  23,  -7,   4,  -2,  12,  -6,  13,  30, -44,   5,  -1,  -4,   4,  -7,   8, -22,   5,  10,  -6 },
      { -27, -32, -15,  -1,   1,  -3,  -2,   2, -41, -43, -17,  18, -18,   0,  -1,   5, -14, -23,  19,  10,   6,  -4,  -5,   0,  -1,  12,  25,  29, -18,   1,  -3,  -1,  12,  10,  32,  -7, -10,  -1,  -9,  -3,  -5,   5,  -6, -15, -36,   6,   6,   0,   4, -12,   9, -25,  13,  37, -14,   3,   3,   0,   5,  -5,  22, -22,  -3,   3 },
      { -14, -38, -10,  -1,  -7,   3,   7,   5,  -6, -28, -23, -16,  12, -11,   3,   2, -37, -19, -30,  13, -10,  -1,   7,   5,  -6,  -9,   4,  16,  19, -12,  -2,  -1,  -4,  13,   5,  43,  -9,  -5,   5,  -2,  10,   6,  23,  14,   2, -21, -11,  -3, -10,   6, -12,   8, -44, -14,  26,  -4,   1,  -2,  -1,  -2, -11,  41, -13,  -1 },
      {   0,  71, -45,   9, -13,   3,  -3,   1, -63, -18,   1,  22,   1,  -2,  -5,   0,  38, -10,  16,  -3, -10,   7,  -2,  -2,  -3, -10,  12, -10,  14,  -7,  -1,  -3,  -7,  13, -10,   2,  -4,   2,   5,  -1,   4,  -3,   0,   4,   5,  -5,  -5,  -2,  -4,   5,  -4,   4, -20,   1,  12,  -3,   2,   0,   2,  -1,   0,  21, -14,   0 },
      {   9,  12,  24,  17,  -4,  10,  -1,  -4,  15,  29,  16,   4,   4,  -9,   0,  -4, -16,  30,  12,  31, -11,  19,   1,  -1,  21,  29,  22,   0, -12, -12,  -1,  -3,   4,   8,  -6,  -3, -39,  18,   2,  -1,   5, -13,   1, -32, -13,  -6,  -4,   3,  -8,   0, -20,  -8, -19,  39,  20,  -4,   7,  -4,  11,  -5,  23,  23, -37,   7 },
      {   2,  27,  33,  -9,  -4,  -9,   1,  -1, -28,   2, -62,  10,   5,   2,   1,   0, -35,  28,  13,  38, -19,  -3,  -3,   2,  43, -14,   2, -19,  -8,  12,  -5,   4,   2, -23,  23,  -5,  22, -15,  -1,   0,  -3,  18,  -8,  -1,   5,  -6,   7,  -4,  -2,  -2,  -5,  17,   1, -10,  -5,   2,  -1,   0,  -2,  -2, -14,   0,   9,  -2 },
      {  -5, -11,   4, -12,  -3,  27,   8,   2,  -5, -23, -16, -16,  10,   8,  11,   9,  -3, -12, -14, -17, -20,  30,  11,  10,  -8, -45, -18, -41,   2,   3,  10,  10,  -9, -14,  -4,  -9,   1,  23,   5,   3,  -1,  -6,   6, -16,  19,  -4,   4,   7,  -2,  12,   0,  18,  19,  32,   5,  -5,   9,  -4,  15,  -5,  23, -29, -39,  23 },
      {  -4,  23, -25,   1,  -3,  -6, -11,   0, -30,  12, -20, -45, -14, -13, -13,  -4, -47, -10, -30,   2,  20,  -9, -12,  -7, -11,  43, -13,   9, -16,   4,  -8,  -7,   4, -21, -35, -15,   9,  26,  -9,   5,   4, -15,  15,  -5,  17,   6,  -7,   1,   3,   3,  -6,  -6,  20,  -2,  -4,  -2,   1,   7,   5,  10,   1,  -5, -12,  -5 },
      {   4, -14,  -8, -13, -17, -28, -33,  -6,  20,  -2, -10,  -1,  -3, -43, -30, -16,  19,  11,   4,   3,  -3, -21, -45, -24,  15, -14,   7,   9,  42,  -5, -19, -13,  -5,   8,   2,  14,  22,   9,  -7, -11,  -5,  -3, -11, -13,   9,  -4,  -6,  -3,   1,   9,  -4,  15,  -1,  11, -10, -11,   3,  -3,   9, -11,   7, -22, -26,  11 },
      {   1, -23, -13, -13, -11,  -9,   1,  -2,  16, -11, -21,   4, -18,   8, -17,  -6,  24,   7,   0, -12, -13, -13,  -3,  -4,   7, -19, -17, -13, -18,  21, -12,  -5,  -5, -24,  -7, -23,   3,  -2,  12,   4,  -5,   1,  -4,  -8, -20,  27,  -6,  -9,   3,  -8,  -3, -22,   7,   6,  12,  13,   0,  25,   1,  26,   2,  62, -30, -35 },
      {   2,   3, -13,  -9, -33, -15,  17,  21,   4,   7,   3, -39, -36, -14,  12,   6,  10,   1,   0, -26, -46, -37,   1,   9,   6,   8,  -4, -27, -38, -38, -10,   1,  10,  10,  16,   1, -12, -26, -12,  -1,  -2,  -2,  -7,  -6,   0, -20,   4,   1,  -7,  -8, -10,   1,  -4,   2,   6,   4,  -3, -13,  -6,  -5,  -1,  -9,   8,  12 },
      {  -3,  12,  22, -45,  -2,  -2,   5,  -3,  -8,  10, -14,   8,  -1,  -7,  -3,  -5, -39,   0, -17, -41, -21,   7,  -6,  -5,  -7,   0,   8,  -1,  29,  -2,   3,  -5, -11,  41, -23, -23, -34,  -9,   7,  -9,  -9,  -2, -35,  14,  -9,  30,   1,   1,   7,   1,  14,   1, -11,   7, -15,  -1,   0,  -3,   8,  -6,  13,  11,  18, -10 },
      {   0, -25, -44,   6, -14,  11,  -6,  -4,  15,  -1, -30,  23,  -7,  -9,   0,  -7,  12,  17,   0,  16, -16,  17,  -6,  -8,  23,  -4, -10,  -5, -13, -10,  13,   0, -13, -22, -30,  -7, -27,  24,   0,   0,  13, -11,  19,  11,   9,   1,   4,   7,   2,  -3,   9,  -9, -20,   1, -10, -15,   3,  -9,   6, -12,  16,  -8,  62,   1 },
      {  -1,   6,  10,   3, -12,   3, -18, -22,  -6, -12,  13, -47,  15,  -1,  -7, -10,  10, -11,  -4,   3,  -1,  10, -13, -19, -21,   0,  -5, -15,   4,  25,  19,  -4,  13,  -2,  23, -10,   5,   6,  -3,  -6, -11,  11,  -4, -29, -22, -15,  31,  -8, -14,  17, -22,  12, -12,  23,   0, -18,  -2,   1,  15,   7,   0,  24,  54, -15 },
      {   3,  18,  62, -25,  23, -20,   1,  -7, -11, -34,  22,  18, -25, -20, -10, -11,   9,  -8, -20,  11, -17,  -3, -15,  -6, -25, -19,  12,   0, -15, -15,  -7,  -1,  -2, -29, -12,   4,  -2,  16,  -2,   3,   2, -16,  29,   2,   4, -10, -12,   3,  -9,  -7,   2, -31,  -4,  -8,  12,  -4,  -3,  -5,  -6,   3,   9,  -2,  16,   2 },
      {   3,   2, -11, -48,   4, -14,  13,  -6,  11, -23, -26,  -6,  19,   6,   6,  -9,  43,  38, -25,  30,  -7,  10,  14,   5, -28,  38,  11,   6,  -4,  13,   5,   9,  40,  15,  -9, -14,  -4, -12,   4,   2,  -1,  -4,   6,   7,  -6,  -5,   3,   2,  17,  17,   3,   7,   0,  -4,   2,   0,  -1,   0,   3,  -3, -11, -11, -14,  15 },
      {   2,   2,   8,  12,  12,   1,   5,  33,   2, -17,   2,  15,  19,  -4,  41,  25,  13,  -4, -24,  15,  -7, -20, -14,  26,  -3,   3,  -2,  11,   5, -17, -10,  -2,   8, -12, -29,  13,  13,  -2, -22,  -3, -12,  -9, -25, -13,  16,  12,  12,   3,  10,  -3, -32,  17,  -5,  23, -34,  -1,  21,  20,   4,  -4,   9,  13,   9, -32 },
      {  -7,  -8, -23,  13, -21, -20,   6, -20,   0,  12,  29,  30,  -2,   0, -10, -15, -29,  -7,  -1,  29, -32,  -6,   7,  -7, -48,  -2,   1,  -3,  -3,  10,   5,   0,  -2, -45, -12,   6,  -6, -27,  12,   4, -33,  -1, -20,   8,   0,  13,  14,  -1,  -8,  -3,   3,  26,   1,   8,  -6,   5,  -1, -18,   0,  -7,  -8,  -2, -12,  14 },
      {  -4,  -6, -42, -40,  44, -15,   7,  -5,   6,  21,   6,  33, -32,   7,  10,   3, -24,   7,  -3,  -6,  12,  16,  -8,   9,   1,  -4,  11, -24,  18, -14,   9,   5,   8,   4,  -1,   7,  24,  -2,   0,  -5,  -1,  -5,  -2, -33,  -8, -25,   7,   0, -10,   3, -33,  -3,  16,   7,  26,  -5,   3,  10,  -6,  10, -10,   4,  16,   2 },
      {   2,   0,  13,  16,   9,  -6,  17, -13,  -6, -19, -18, -14,   1, -13,  -5, -12,  17, -13, -10,  -3,  28, -13,  24, -12,  27,  -5, -13,  24,   1, -12,  10,   5, -16, -10, -34,  -5, -11, -17,  26,  -4, -12, -13, -44, -18, -16,   0,   8,   9,   6, -18, -17,   7,   8,   1,  42,  -8,   8,   7, -12,  -9, -14,  -6,  11,  36 },
      {   1,   2,   2, -14,  14, -37,  11, -11,  -3,  -9,  11, -25, -42,  17,  12, -33,  14, -10,   5,  12,   8,  18,   8,  -2,  34,   6,   7,  11,  -3,  -9,  28,  10, -17,   0,   8,  32, -19,  11,   0,   0,   2,   4,   1,  23,  22,  14,  12,   5,  -4,   7,  11,  39,  12,  21, -21,   4,   3,  -4,   4,  19,   1,  10,  -9, -21 },
      {   5,   3,  -1,   2,  -3,  26,  18, -13,  -1,  -6,  23,  -1, -14,   0, -24,   6,   3, -21, -23, -22, -23,  22,  -1,  -4,  50,  23,  14,  -7,  29,   6,  -5,  -6,  49, -39,   4,  12,   3, -13,  14,  -3,  -4,  19,  10,  10,  -8,  12,  -5,  -1,  27,   1, -17,  -8,  -6,  -7, -10,   3,  13,  10,   3, -10,   3, -10,  -1,   2 },
      {   0,   0,  11,  30, -24, -20,  50, -11,   2,   3,  -3,  17, -17,  20,  -8,   7,  -8,   3,   1,   2, -25,   1,  -8, -12, -15,  -1, -12,  15, -11,  11,   0, -10,  14,  38,   3,   7,  31,  27,   8, -15,  16,  -1,  -5,   7,  16,   3,  -8,  -6,  28,   6,  -1,   7,  16,   6,  20, -42,  18,  21,   4,   8,   8,   3,  13,   7 },
      {  -1,   4,  -2, -43, -31,  24, -27, -13,   1,   6,   8,  -9,   1, -24,   8,  33,   5,   5,  10,   6,   3,   5,  30,  13,  -2,   9,  20,  13,  -4,   8, -13,  11,   0,  -9,  12,  32,  24,  13,  14,  -3,  -9,  -5, -16,  12,  22,  23,  -8,   4,  -7, -32,   6,   1,  20,  12,  26, -15,   2,  -5, -13,   8,  23,  15,  20,   2 },
      {   1,  -2,  -2,  18,  51,  36, -11,  21,   2,  -9, -13,   7, -14,  10,  -4,  24,  10,   9, -22,   0,  -8,  -9, -19, -41,  -3,  21,   3, -17, -11,  -6, -16, -25,   2,  -1,  14,  10, -12,  -4,   9, -22, -15,  -4, -10,  17,  -2,  13,   0,  -7, -20,   4,  21,  28,  14,   5,  14, -25, -14, -23,   3,  19,  -7,   8,  -2,   7 },
      {   0,   2, -16, -10,  25, -47,  28, -12,  -1,   4,  13, -28,  22,   5,   5,  41,   6, -12,   9,  22,  -1,  13, -20,  29,   9,   5, -20,  -4,  10,  -4, -29,  -7, -23, -15,  13, -18,  -9,  10,   8, -15,  -9,  22,  -2,  -2, -12,  21, -24,   6,  -5,  -5,  10,  -8, -10,  -8,   4, -20,  -6, -13,  16, -18,   7,   3,   4,  11 },
      {   0,   3,  -4, -14,  -7, -10,  11,  95,  -1,   8,  14,  -3,   7,  -8,  17, -35,  -6,  -6,   8,   9,  -2,  -1,  15, -29,   1,   5,   4,   3,  20,  13,   2,  16,  -5, -20,   7,  -6,   5,  10,  20,  -3,  -4,  16,   3,  -3,  -8,   3,  -5,  -6,   8,   8,  11,  -9,   4,  -1,  17, -13,   1,   9,  15,   3,  10,  -1,  10,   7 },
      {   0,   0, -12,  -1,  20,  32,  66,  -5,   0,   1,  -9, -13,  29, -10, -20, -45,   7,   9,   3,  -2, -18,   3, -15,   3,  -7,  15,   7,   8,  10,  -5,  -8,  -8, -20,   1,  17,   7,  24,  11,  -4,  10,  -1, -10, -10,  -4,  13,   1,   6,  -9, -29, -30,  -4, -10,   5,   5,  -7,  28,  -9,  -8, -15,   0,  15,  -2,   7, -11 },
      {   0,  -1,  -5,  35,  32,   1, -23, -18,   0,  -2,  -8,   2, -10, -49,   7, -28,   2,   4,  -9,   7, -13,  16,  23,  -1, -22,   7, -19, -19,  -2, -19, -13,  39,   1,  20,  19,  -9,   1, -15,  13,  -1,   3,  41,  -6,  -2,  11,  11, -28,  12,  16,  -7, -10,  -1,  10,  -5,  -3,  -1,   6,  15,   8,  -3,  16,   0,   4, -18 },
      {   0,   0,  10,  29, -16,   5,  27,   9,   2,  -2, -25,  -9, -42,  -8, -15,  20,   7,  41,   5, -16,  25,   0,  13,  26, -25,  -6,  33,   7,  29,  -9,   8,   7,  -7, -11,  -6,   3,  -1,  10,   7,   6, -43,  20,  14,  -5,  -3,   9,  -1,   7, -16,  24,  -9,  -8,   2,  -3,  -8,  -2, -10, -11,  22,   3,  -4,   0,   6,  -3 },
      {  -1,   4,  -8, -37,  -1,  36,  26, -23,  -2,   9,   8,  -4,  20,  -3,  -1, -12,  -6,   2,  22,   8,  21, -24,   8, -26,  -5, -23,  12,   3, -30, -33, -11, -20, -12,  -9,  -8,  11, -15, -12, -10, -16, -16,  31,  15,  -3,  -2,  -5, -10,   1,  26,  18,  -7,   4,   5,  -4, -11, -27,  13,  23,  21,   7,  -6,  -7,  -5, -14 },
      {  -2,  -1,  -2, -12,   5,  15, -12,  19,   0,  -2,   0, -25, -31,  37, -23,  13,   3,  12,   7,  23,  -3,  13,  -5, -25, -29, -10,   2,  -6,   0,   3,   8,  -1,  -9,   1,   5,  12, -11,  28,   1,  -1,   8,  10, -38,  -5,  -3,  -2, -10,  22,  17, -38, -26,  -2, -20, -27, -14,  14,  18,  10, -10, -34, -11,  -8,  -7,  -6 },
      {  -1,  -2,  -3,  15,  -6, -18,  -9,  -4,  -1,  -1, -13, -13,  19,  53,   3, -33,  -6,  10,   5,  -9,  26,  -6, -10,  -2,  -7, -17,  15, -17,  19, -27,  -9,   9,  48, -17,   0,   2,  -9, -16, -14,  -7,  -9, -12, -11,   9,  15,  -3, -29,  15,  -5, -19,  10,  -9,  -1,   3,   8, -18,  -3,  -4, -11,   6,  42,  10,   7,   2 },
      {   1,  -1,   1, -10,   7,  26,  -1,   7,   0,  -2,   8,  11, -18,  12, -30, -37,   2, -10,  -5,  -1,   3, -20,  10,  47,   9,   6,  -6,  -4, -13,  33, -25,  31,  -5,   5,  -7,   7,   5,   3, -24,  -2,  12,  -8, -12,  -9, -17,   2, -14,  13, -17,   9,   0,  15,  -8,  -8, -21, -51,   6, -14,  12, -13,  -8,  13,  -1,  16 },
      {   0,   2,  -6,  -2,  28, -24,  -3,  -5,   0,   3,  -1,   9,  25,  14, -30,  18,  -2,   0,  -9, -11,  -9, -58,  37, -26,  10,   7,  11, -15,  -1,   8,  23,   9,  -7,  10,  11,   5,   5,  13, -11,  25,  -1,  10,  21,   2,  22,  24,  22,  27,  -1,   6, -18, -14,  -5,   9,   2,   3,   5,  -3,  11, -21,  22,   9,   2,  13 },
      {   1,   2,  -2,   2,  17,   7, -16,   0,   0,  -1,   2,  -7, -16,   5,   6, -24,   0,   5,   0,   9,   7, -16,  -7,  40,  -5,  -9,   3, -11,  -2,  11,  -2, -43,   8,   8,  -7,  15, -13,   9,  -5,  23,  -6,  28,  -2,  -8,  17,  14,  12, -59,  32,  -5,  17,   3,   2,  -1,  26,  21,   9,   2,  12, -22,  14,   9,  13,  27 },
      {   0,   0,   1,  -3, -25, -13,  -3,   7,   1,  -6,   3,  28,  25,   0,   4, -15,   2, -10, -24, -12,  11,  10, -14,   1,  17,   1,  -6, -11, -30,  -1,   8,  -3,  21,  24, -17,  12,  -4,  44,  -1,  -5, -38,  50,   0,  -3,  -7,   0,  -6,  11, -16, -31, -19,   1,  17, -11,   2,   6, -16, -31,   3,   6, -19,  -2,  -6,  -3 },
      {   0,   0,  -1,  -5,  -2,   1,   7,  21,   0,  -1,   9,   5,  11, -18, -65,  22,  -2,  -6,  -6,  30,   5, -12,   7,  31,   2,  -2, -16, -13,  14, -19,  40,  -8,  13,   2,   2,  -8, -19,  -1,  -7,  17,  11,   8, -20,   6,  -4, -26, -10,  -1,   9,  -8,  20,  13,  11,   9,   8,  -8,  -6,   4,  -7,  37,  10, -11,   4, -26 },
      {   1,   1, -10,   7,  32,   0,   7,  -2,   0,   8,  -2, -18,  10, -33,   5,  -3, -12,   0,  36,  -6, -11, -16,  -7,  12,  -1, -31,   8,  -1, -11,   5,  -3,  18,  54,  -2,  -3,  11,  -7,  47,  15,  -7, -10, -27, -13,  17, -16,  10,  20,  24,  17,  13,  19,   2, -13,  -4,   5,  11,  -3,   2,  -2,   4, -24,   2,  -1,  -4 },
      {  -1,   0,  -2,   8,   8,  -3, -31,  -9,   0,   1,  -6,  -3,  17,  24,   9, -15,   2,   6,  -6, -18, -38,  -5,  -9,  29,   5,  23,  15,   9,   2,   1,  10,  -8, -37,  -7,  17,  10,  -3,   9,  19,  16, -41, -15,  -8,   8,  -2, -33,   4,  14,  18, -10,   5, -21,   0, -14,  10, -23,  12,  27,  22,  25, -10, -20,  -6, -10 },
      {   1,   0,  -7,   3,  26, -19,   7,   3,   1,   1,  -4, -13,  -3,  -8,   0,  15,   0,   1,  19,  -7, -15,  11,  38, -22,   7, -18,   1,  15,  -1,  48, -26, -14,  17,  12, -31,  10, -15, -13,  -3,  18, -16,  -2,   0, -13,  11, -40, -12, -23, -10, -23,  -7,   0,  -3,  -9, -21, -11,   7, -23,  10,  21,  20, -11,   0, -17 },
      {   0,   0,   0,  -1, -19,   3,  18, -27,   0,  -3,   3,  16, -16,  -5,  39,  -3,   2,  -6, -24,  16,  22, -37,  22,  -9,   5,  13,  -3, -34,  20,  16,  -6, -17,   5,  -6,  16,   5,  -8,  26,  13,  -9,  17,  -7, -35, -10,  -2, -17,  -7,   5, -22,  10,  24, -13, -12, -19,   0,  15,   3,   2,  28,   2,  10, -17,  -4, -23 },
      {   1,   0,  -3,   0,   6,  18,  15, -11,   0,   0,   1,  -7, -12, -25,  42,   6,   1,   0,   7,   3,  -4, -12, -32, -13,   2,  -1,   3,   1,  10,  24,  65,   2,   3,  -4,   0,  -8,   7, -13, -27,  29,  -4,   7,  -3,  -9,   3,   4, -31,  14,  16, -23,  12, -11,  -5,  -7, -11, -28, -12, -13,  -3,   8,  -2,  23, -15,  24 },
      {   1,   2,   1,  -4,   1,   8,   2,   3,   0,   2,   0,   0,  -1,   7,   1,   1,   0,   3,   4,  -4,   8, -20, -44,  15,  -9,  -1,   7,  11,  -7,  22, -10,   3,  -3,  -1, -11,   3,  -8, -23,  56,  -2,  38,  31,   9, -12,  13, -11,  16,  47,   9,   0,  -6,   9,  -1,   3,   2,  19,  20, -30,   6,  31,  23,  -1,   6,   5 },
      {   0,  -1,  -2,  10,   2, -24,   2,  -2,   1,   1, -11,   2,  18, -15,   9,   1,   2,  13,  -8, -35,   2,   2,   9,   3, -17,   5,  41, -14,  -4,  10,  -6, -11, -14, -33,   4,  11, -23,  21,  -5,   7,  49,  31, -17,   4, -15,  -1,   3, -17,   1,   1,  -7,   7,   2,  10, -14, -30,  -2,  12, -50,   8, -12,  -7,  -6,   5 },
      {   0,  -1,   1,   7,  -6, -12,  14,   0,   1,  -1,  -2,   6,  11, -12, -12,  13,   1,   6,  -8, -20,  15,  20, -13, -14,  -6,   6,  21, -12,  -6,   3, -13,  34,  -9, -13,   2,   9,  -4,  -9, -48,  -2,  14,  -5,  -3,  -1,  -2, -11,  17,   1,  18, -34,  32,  14,   7,  -5,  23,  18,  34,  -3,  56,   3, -17,  13,   4,   1 },
      {   1,  -1,   0,  -3,   3,  10, -14,  -9,   1,   2,   0,  -7,   3,  16,   9,   1,   2,  -2,  12,  11, -29, -10,   4,  -5,  -1,  -7,  -7,  26,  23, -27,   2,  12,  12,  -6, -10, -14,  -2,  15, -15,   8,  28,  24, -23,  24,  -4,   9,  -4, -24, -34,  14, -22, -30,  22,  -1,  -1,   7,  36, -22,  26,  35, -23,   2,   2,  24 },
      {   0,   1,   0,   1,  -5,  14,   4, -12,   0,   0,   6,  -2, -14,  16,   6,   7,   1,  -3,  -7,  19, -13, -10,  -1, -11,   5,   4, -12,   2,  32,   3, -38,  38, -11,   7, -17,  -2, -13,  15, -30,  -9, -17,  32,   3, -14,  11, -17,  39,   7,  22,  -2,  24, -15,  -7,  10,   2,   4, -35,  15, -36,  28,   0,   3,  -2,  17 },
      {   0,   1,  -3,   1,   9,   3,  -2,  -5,   0,   2,  -2,  -5,   6,  10,  -4,  10,   2,   1,   2, -10, -20, -13, -19,   1,   5,   6,  11,  17, -16,  26,  31,  37,  -6, -10, -21,  14,  -4,  -3, -15, -41,  -9,  20, -11,  -8,   2, -12, -33, -21, -17,  37,  14,   0,   7,  -1,  33,  22,   8,   4, -18, -28,  18, -15,   2, -30 },
      {   0,   0,   0,  -5,  -4,   3,   6, -10,   0,   0,   2,   3,   6,  17,  -1,  14,  -3,  -2,  -6,  -1,  11, -23, -19, -14,   2,   2,   6,  -8,  -4, -16,  -6,  60,  -2,   4,   4,   2,  -3,  14,  47,  40,  -4, -15,   7, -14,  -4,   2, -25, -58,  14, -12,  -2,  16, -15,  15, -17,  -3,  -9,  -6,   5,   2, -14,  -7,   6,  -5 },
      {   1,   1,   2,   2,   6,   6, -11, -10,   0,   0,   2,  -5,  -5,  17,  12,  -5,   0,   4,   4,  -8, -29, -20,  23,   4,  -2,  -2,  13,  26,  22, -20,   0, -15,   6, -10, -20, -24,  23,  14,   5,  -2,  27,  15,  25, -15, -26, -13,  -8,  14,   2, -34,  21,  42,  16,  11,  11,   0, -32,  -5,  19, -29, -10,   4,   1, -26 },
      {   0,   1,   0,  -2,  -1,   0,   0,   1,   1,   2,   1,  -4,  -1,   1,  -2,   7,   0,  -1,  -2,   1,  -1,  19, -15, -10,   4,   0,  -3,   7, -12, -22,  -4,   3,  19,   7, -16,  12,  11, -12,  17,  55,   1,  23, -11, -18,   2,   0,  33,  10, -46,  11,  50, -12,   4, -15,  -1, -29,  20,  26,  -7, -31,   3,   5, -11, -14 },
      {   0,   1,  -1,   2,   5,  -1,  -2,   0,  -1,   1,  -1,  -4,  -2,  -5,  -4,  -2,  -1,   4,   1,   1,   3, -11,   8,  14,  -3,   0,  -1,  -4,   6,  -5,  31,  11,   9,   2,  -2, -11,   2,  -7,  18, -65,   4,   2,  23, -15,  18,  16,  42, -34, -15, -46,  10, -19, -21, -12, -13, -31,  21,  -2,  -2,   4,  -6, -10, -23,  -8 },
      {   0,   0,   0,   0,   0,  -1,  -1,   2,   1,   0,  -1,  -3,  -2,   5,  -1,  -8,   0,   1,  -4,  -1,   1,  -1,  15,  -2,   0,   3,  -3,  -8,   8,   3,  -9, -18,  -1,   5,   3,  -8,  11,   0,  -5,   6, -16,   6,  14, -22,  13,  14, -33,  19,   4,   2,  32,  -8, -29,  48,  10,   7,  55, -23, -32,   9, -52, -12,  13, -14 },
      {   0,   0,   0,   1,   4,  -6,  -5,   1,   0,   0,  -1,   0,  10,  -1, -13,   6,   0,   0,  -6,  -6,  -6,   5,  17,  -1,   2,   6,   6,   0, -10, -20,  12,  -7,  -3,  -4,  -1,   8,  16,  13,  22, -30,  -8,  -1,  -8, -27,   0, -35, -19,  -1,   7,  12,  35,  -3,  17,  -6, -57,  26,  14,   6,  -4,   2,   6,  49,   9,  46 },
      {   0,   0,   0,  -1,   3,  -4,  -4,  -1,   1,   1,   2,   1,  -2,   3,  -1,   2,   0,   2,   0,  -5,  -8,   4,   3,   0,  -1,  -1,   5,   6, -14,  -1,   6,  -4,   5,   3,  -8,   1,  13,   3, -12,   2,  -2,   8,   7, -19, -17,  35, -19,   4, -33, -10,   6,  38, -49, -15,   2,  22,  -3,  51,   9,  49,  20, -28,   8,  31 },
      {   0,   0,   0,   0,  -1,   0,  -2,  -5,   0,   0,  -2,   1,  -1,   2,   7,  -3,   0,   0,  -3,  -1,  -2,   5,  10,  -5,   1,   5,  -1,   3,   1, -24,   1,   3, -12,  -3,   5,  -9,  23,  16, -14,  10, -11,  -7,   2, -17, -40,  12,  19,  -8,  31,  18,   2,  20,  -8, -48,   9,  -5,  34, -54, -29,  16,  43,   1, -10, -17 },
      {   0,   2,   1,  -1,  -4,   5,   7,  -4,   0,   2,   3,   2,  -5, -11,  10,   3,   1,   2,   2,   7,   6, -21, -13,   1,   1,   2,  -5,  -7,  21,  16,  -7,   5,   0,   9,   9, -26, -28,  13,   2,   4, -24,  -3,  38,  34, -22, -20,  -3,   2, -26, -21,  -8,  33,  30,  -6,  -3,   6,  54,  31, -28, -15,  12,  11,   3,  11 },
    },
    {
      {-107,  51,   2,   5,   1,   2,   0,   1,  42, -16,  -3,   0,  -1,   0,   0,   0,  -5,   0,   4,  -1,   0,   0,   0,   0,   9,   3,  -3,  -1,   0,   0,   0,   0,   2,  -3,  -1,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -53, -24,  24,   4,   6,   1,   2,   0, -97,  34,  18,   1,   4,   0,   2,   0,  13,  21,  -5,  -2,  -2,  -1,  -1,   0,  14,   9,  -5,  -4,  -2,  -1,   0,   0,  15,   4, -10,  -1,  -1,   0,   0,   0,   8,  -4,  -3,   1,   0,   1,   0,   0,   5,  -1,  -2,   1,   0,   0,   0,   0,   3,  -1,  -1,   1,   0,   0,   0,   0 },
      {  -2, -27,   4,   9,   2,   2,   1,   1, -11, -72,  32,   9,   6,   2,   2,   1, -74,  18,  35,  -2,   3,  -1,   1,  -1,   0,  32,   7,  -7,  -2,  -2,  -1,  -1,   5,  21,  -4,  -9,  -1,  -1,   0,   0,   9,   9, -11,  -4,   0,   0,   0,   0,   2,   2,  -4,  -1,   1,   0,   0,   0,   3,   1,  -3,   0,   0,   0,   0,   0 },
      { -27, -47,  11,  11,   8,   3,   2,   2,  -1,  -9, -21,  17,   3,   3,   0,   1,  22, -77,  14,  21,   1,   3,   0,   2, -42,  -2,  45,   3,   0,  -2,   0,   0,   1,  24,  15,  -7,  -4,  -3,   0,   0,  -1,  15,   5,  -9,  -2,  -1,   0,   0,   3,   9,  -4,  -6,   0,   0,   0,   0,   0,   3,  -3,  -1,   1,   0,   0,   0 },
      {  13,  72,   0, -20,  -5,  -4,  -1,  -1, -34,  39,  -6, -11,   0,   0,   1,   0, -56, -27,   9,   5,   9,   3,   3,   1, -51,   8,  15,   9,   8,   0,   1,   0,   0,  17,   2,   4,  -1,  -2,  -1,  -1,  10,   7,  -6,   0,  -3,  -1,  -1,   0,   9,  -1,  -6,   1,  -1,   0,   0,   0,   5,  -2,  -2,   1,   0,   0,   0,   0 },
      {  -6,  11, -38,  12,   4,   3,   0,   1, -36, -40,   0,   9,  12,   3,   3,   1, -15,   5, -39,  18,  10,   2,   1,   1,   9, -70,   4,  35,   3,   3,   0,   1, -22,  -4,  42,  12,  -1,  -3,  -1,   0,   6,  13,  16,  -2,  -6,  -3,  -1,   0,   0,   7,   7,  -7,  -3,  -1,   0,   0,   4,   4,   1,  -3,  -2,   0,   0,   0 },
      {  -8, -26, -23,  -1,  14,   2,   2,   1,  -5,   1, -68,  12,  14,   4,   1,   1, -37, -15, -15,  17,  16,   1,   2,   1,   4,  35, -37,  18,   5,  -2,  -1,   0,  24, -36, -17,  30,  -3,   0,  -2,   1,  -2, -16,  18,  17,  -5,  -1,   0,   1,   4,  -7,  13,   8,  -6,   0,   0,   0,  -1,   0,   9,   1,  -4,   0,   0,   0 },
      {  16,  48, -57,   9,  -3,   1,  -2,   0, -35, -36, -18,  33,   1,   3,   0,   2,  48, -11,  27,  10, -11,  -3,  -2,   0,  23,  27,  12, -14, -10,  -5,  -1,  -1,  25,  12, -14, -12,  -8,  -1,   0,   0,   3, -14, -10,  -4,   1,   4,   1,   0,  -3,  -8,   1,   2,   3,   2,   0,   0,  -2,  -4,   1,   0,   1,   0,   0,   0 },
      {  -6, -19, -34,  10,   6,   1,   0,  -1,  22,  33, -21, -14,   8,  -4,   1,  -2, -18,  33, -43,  26,   2,  -1,  -1,   0,  16,   8,  39,  11, -14,  -4,  -3,  -1,  10,  61,  -6, -25, -11,  -6,  -1,  -2,  13,  -9, -31,  -2,  -3,   2,   1,   1,  -5, -13,  -6,   5,   2,   2,   1,   0,  -2,  -8,   0,   5,   0,   1,  -1,   0 },
      {   8,   7,  23,  10,  -3,  -4,  -1,  -2,   5,  33, -21,  27,  -8,  -3,  -3,   0, -23,  14,  36,  30, -15,  -3,  -3,   0,  48,   7,  18,   6, -26,  -2,  -4,   0,   9, -30,  47, -14, -19,   2,  -1,   0,  -4,  27,  19, -35,  -6,   1,   1,  -1,   2,   4, -12, -17,   1,   4,   1,   0,  -2,   1,  -3,  -6,   3,   2,   0,   0 },
      { -11, -11, -43,   9,  12,   3,   2,   2,   3,  15,  -6,  17,   7,  -2,  -2,   0,  23,  53,  21, -18,   3,  -9,   0,  -3, -57,  33, -20,  10,   4,  -8,   3,  -1, -33,   7,  25,  17,  -8,  -3,   0,   0, -15,  35,   6, -13,  -5,  -3,   0,   0,   2,   9, -19,  -5,  -4,   2,   0,   0,   1,  -1,  -5,   0,   0,   2,   0,   1 },
      {   0,   2,  -8,  40,   1,   0,  -1,   0,  10,  21,  40,  31,  -2,  -7,  -2,  -2,  -8,  16,  11,  37, -10,  -7,  -2,  -1, -29,  -4,  28,   7, -17,  -5,   0,   0, -14, -19, -29,   5,  -9,   4,   2,   1,  -5, -50,  28,  24,  -9,   8,   0,   2,   0,  13,  39,  -7,  -7,   0,  -2,  -1,   5,   8,   6,  -9,  -4,  -1,  -1,   0 },
      {  13,  11,  26,  16,   6,  -5,  -2,  -1,  17,  -4,  39,  17,   5,  -7,  -3,  -2,  20,  -7, -20,  -4,  16,  -9,   2,  -1,  -7,  15, -26,  50,   7, -11,   1,  -2,  63,  20,  15,  24, -31,  -9,  -4,  -1,  27,   1,   0, -12, -24,  -2,   0,   1,   9,  -6,  -3, -11, -11,   3,   2,   1,  -2,  -5,  -2,  -7,  -1,   3,   1,   0 },
      {   3,   2,  25,  29,   5,  -4,  -2,  -1,   3,   6, -14,  63,  -3,  -2,  -4,  -1,  -7,   0,   4,   7, -12,  -4,  -1,   0, -15, -33, -38,  -1,  -3,   3,   3,   2, -15,   4,  -5, -13,   3,   4,   3,   1,   4,   9, -57,   4,  13,   5,   3,   1,   6, -37,  -1,  39,   5,   6,  -1,   1,   7,  -1,  24,   7,  -4,  -1,  -2,   0 },
      {  -4,  -9, -39,  49, -19,   5,  -2,   3,   0,  10,  18,  -9, -29,   1,  -1,   0, -21, -24, -31, -22, -17,   5,   5,   1, -20,  -4, -21, -17, -22,  14,   4,   2,  31, -29,  -8, -25,  -1,  19,   1,   2,   2,   7,   7, -26,  18,   7,   0,  -1,   4,   4, -30, -15,  19,   3,  -1,  -1,  -2, -11,  -9,  10,  11,   0,  -1,   0 },
      { -11, -15, -17, -37, -10,   5,   6,   1,  -2,   1,  -3,  -5, -34,  10,   4,   1,  -4,   5,  16, -40, -14,   9,   5,  -1, -17,  -3,   7,  27, -11,  10,  -1,   0,  23,  -5,  45, -24, -22,   9,  -4,  -1, -11, -33, -14,  -3,   6,  13,   0,   1,  -1, -21,  51,  13,  -1,   4,  -4,   0,  -2,  13,  15, -17,   2,  -3,  -1,  -1 },
      {   0,  -9,   0,  -7, -22,   0,   0,  -1,   7,   3, -34,  21, -18,   1,   0,   0,  -5,  23,  11, -13, -37,   6,  -1,   0,  -8, -31,  -7, -35,  15,  14,   2,   1,  21,  30, -10,  21,  12,  -1,   0,  -2,  56,   3,  19,   5, -17,  -1,  -5,  -1,  27,  15,  15, -25, -11,  -5,  -3,   0,   9,   1, -20, -20,   1,  -2,   1,   0 },
      {   6,  21,  46,  19,   2,  -8,  -1,  -3, -14, -24, -34,  12,  -6,  -8,   2,   1,  -3,   5, -50, -21, -19,  -5,   5,   1, -11,  26, -12, -13, -20,  -1,   5,   1, -24,  32,   6, -12, -21,   8,   5,   0, -30,  -7,  22,  -6,  -6,  14,   0,   0, -27,  16,  23, -15,   1,   6,   0,  -1, -16,   8,  -8,  -6,   8,   3,  -1,  -1 },
      {   6,   6, -15,  43,  35,  18,  -2,   5,   5,  31,  14,  15,  15,  19,  -6,  -1, -13, -23,   5, -46,  21,  12,  -4,   1,  32,   8,  -4, -18,  16,   3,  -6,  -1,   2,  22,  22, -14,  23,  -9,  -7,  -1,  -9,  17,   8,  25,   7, -18,  -2,   0,   1,   7,  33,  -1, -17, -14,  -1,   0,  -4,  13,  -1, -19,  -6,  -6,   2,   1 },
      {  -7, -11, -18,   9, -17,   1,  -2,  -1,  -1,  17,  14,   5, -14,  -8,   1,  -1, -17, -25,  18, -15, -23,   6,   2,   3,  26,  -7,  -8,  -8,  33,  -5,   3,  -2, -24,  27,   5,  49, -17, -17,   1,  -3, -12, -29,  12, -23, -26,   8,   0,   1, -44, -27, -14,   2,  10,   9,   3,   2, -13, -13,  17,  18,   2,   5,   1,   0 },
      {  -1,  -2,  -4, -35,  54,   2,   3,  -4,   7,   8,   1,  29,  50,   1,  -4,  -2, -10,   2,   8, -22,  10,  -2,  -3,   1,  -8, -33,  -5, -16,   8,  -9,   3,   2,  12,  -8, -13, -21, -24,  -1,   8,   3,   0, -32,  -1, -29,   4,  16,   4,   2, -10,  13,  -8, -19,  23,   2,   0,  -2,  -7,  -3, -31,  10,  18,  -1,   0,  -1 },
      {   8,  11,  23,  33,  13,  38,   2,   9,   0, -12, -13, -32,  41,  33,   9,  -2,  10,  23,   0,   1,  -8,  20,  -2,  -3, -25,  -1,  19, -32,  -4,  14,  -7,  -1,  18, -13,  24,  15,  -2,  -4,  -7,   0,  15, -18,  14, -21,  -9,  -1,  -4,   1,  -3, -28,  -5,  17,  10,   0,  -2,   1,  11,  -9,  35,   6,  -7,  -3,  -2,   1 },
      {  -1,  -5, -17,  -3,  24, -13,  -4,  -4,  -2,   1,   6, -37,  17, -24,  -8,  -1,  -2, -25,   8, -10, -11, -31,  -4,   1,   5,  -9, -19, -29, -47, -21,   6,   2, -20,  -4,  -3,  -5, -35,   0,  14,   0,  23,  19,  -1,   6, -36,   9,   4,  -1,  18, -18,  14,  -2, -23,  17,   2,   1,  11,   2,  22, -18, -12,  12,  -2,  -1 },
      {  -3,  -1,   4,  19,  -7, -26, -11,   1,  -7,  -6, -15, -14,   7, -26, -17,   5,  -3,  13,  14,  -1,  10, -30,  -5,   3, -22, -19,   8,  -9,   4,   1,  -1,   8,  30, -14,  11,  -7,  40, -18,  -1,   1, -24, -13, -15,   8, -16, -28,   8,   0, -31, -39,  -2, -30, -17,   7,  12,   3, -30,  -6, -18, -21,   9,   8,   7,   1 },
      {   2,   1,   9,   9,   1,  36,  14,   7,  -3,  -1,  -6,  -3, -20,  51,  12,   1,   4,  -6,   7,  -2,  13,  20,   3,  -6,   4,   4,  -8,  21, -21,  -3,  -8,  -5, -38, -12, -12,  -6,  -8, -10,  -4,  -1,  17, -10, -26,  17, -35,  -2,   0,   1,   5, -24,  -6, -36,  -6,  17,   4,   3, -11, -14, -42,  -3,  27,   9,   2,   2 },
      {  -2,   1,  -6, -43,  22,  16,  11,  -6,   5,   1,  25,  27,  -5,  19,   8,  -1,  -6,  -1, -16,  15, -35,   1,   0,  -4,  -9,  -4,  -2, -16, -39,   8,  -2,   1,  27,   8,   7,   9,  15,   1,  -3,  -1, -19,  26,  22,  35, -14, -18,  -2,  -3, -20,  -9,  -7,  15, -20,   0,   2,   2, -29, -23,   2,  32,   2,   4,   1,   0 },
      {  -3,  -2, -11, -21, -24,  14,   5,   2,   3,   7,  21,  -2,  20,  18,   8,  -3,   2,  -6,  -6,  54,   4,  16,   2,  -5,  -6,   3, -43, -22,   9,  15,  -8,  -3,   6,  12, -10, -14,   5, -12,  -7,  -1, -23,  20,  -9, -31, -10,   6,   0,   2, -14,   0,  24, -22,  12,  13,  -3,   1,  -8,  27,  17, -46,   2,  -5,  -3,   0 },
      {   1,   1,   3,  26, -20,  21,   0,  15,  -2,   5, -17, -38,  11,   8,   8,   2,  -1,   4,  45,   8,  -3, -11,  -1,  -2,  -2, -34, -27,  19, -20, -18,   5,   0,  24,  24, -12,   4, -24,   0,   5,   0, -13,  12,   0,  14,  10,   2,   4,  -2, -22,  36,  12,  31,   5, -10,  -2,  -4, -25,  11, -23,  18,   7,  -3,   0,  -1 },
      {   1,  -1,   0,  -1,  16, -15,   2,   2,   2,   4,   8,  -8,   7,  -8,   6,   2,   7,  -6,  15,  20,  14,  -7,   6,   0,   2,  12, -21,   7, -22,  16,  14,   0, -17,  20,  22,  12,  20,  44,  -1,  -1,  48,  -7,  22,  11,  43,  11, -17,  -2, -23, -32,   7, -12,  32,  -8,  -8,  -1, -23, -11,  -8, -14,  14,  -9,  -2,   0 },
      {   1,   0,  -3,  13,   8, -37,  -2, -21,   2,   2,   1, -10,   7,  -6,   3,  -3,   5,   1,  10,   5,  10,  10,  17,  -5,   4,   5,  -3,  -4,  -1,  47,   4,   0,   9,   5,  16,   4,  17,   0,   3,  -4, -11,   4,  -8,  16, -45,  22,   4,  -2,  22,   5,  18,   0,  27,  25, -10,   0,  10,  29,  -7,  56,  30,  -9,  -9,  -3 },
      {   2,   2,  -1,   3,   7,  19,  13,  21,   2,   1,   8,  -6,  29, -19,  13,   1,  -4, -11, -11,   0, -59, -15,  10,   1,  10,  12,  14,   8,   2,   4,   3,   4,   3,  -8,  14,  41,  10,  -8,  10,  -3, -26,   1, -31,  18,  16,  24,   2,  -5,  28,  -4,  -8,   5,  17,   4, -10,  -1,   7,   5, -36, -38,  14,  -3,  -2,   1 },
      {  -3,  -5,  -2,  -6, -55,   6, -12,  16,   1,   2,  10,  14,  35, -21,  -8,   7,   1,  -3, -14,   2,  13,   6,  -1,   2,  -3,  15,   0, -22,  21,  -4,  -9,   4, -13, -15,  21, -27, -24, -13,  11,   3,  16,   7,  16,  13,  -7,   6,   5,  -1,  24, -17,  22,  29,   2, -14,  -1,  -1, -35, -31, -25,  15,  16,  -3,   7,   0 },
      {   1,   1,  -1,   7,  -3, -13,  30,  15,   1,   2,  -2,   5,  19, -24,  43,   9,   2,   4,  11,  -7,  14,  37,  37,   2,   4,  -9,   9,   1,  10,  39,  27,  -7,   5,  -4, -14,  -7, -21,  37,   8, -14, -11,   6,  -5,   3, -10,   1,  -9,  -9, -12,   4, -12,   4, -43,   6,   2,   0,  -5, -20,   1, -18, -12,  25,   1,   2 },
      {  -2,  -3,  -5,   7,  -4, -23, -16, -70,  -1,  -1,   4, -12,  20,  11,  -6, -25,   1, -10,   0,  12, -27,  23,  -4,  -8,  -2,   9, -17,  -2,  -8,  11,  -2,  -4, -13,  -5,  18,   3, -11,   1, -19,   0,   6, -26, -19, -11,   5, -27,  -9,   2,  -6,  12,  -7,  22, -31, -27,   5,   3,   4,   2, -38,  -9,  -6,   0,   6,   3 },
      {   0,   0,  -2,  -8,  -8,  11,  22,  49,   2,   5,   4,   6,   9, -18,  24,  17,   0, -12,  -7,   5,  -9, -32,   9,   5,   0,  16, -19,  -5, -11,   8,  -1,  -2, -12,  -5,  25, -15,  24,  -8, -15,  -3,  28, -31, -16, -19, -13, -24,   1,   0, -23,  20,   2,  -3, -18, -17,  10,  -1,  26,  27, -11,  35, -16,  -6,   6,   0 },
      {   2,   4,  -1,  14,  35, -27,  50, -17,  -1,  -4,  -1, -13, -33, -11,  50,  -1,   1,   4,  -2,  11, -10,  -1,  20,  -5,  -7,  -2,   0, -12,  26, -29, -19,  -7,   1, -15,  -1, -15,  -6, -39, -11,   1,  15,  11,   4,  -4,  -2,  -1,  -3,   1,   2,  -4,  12,  15,  21, -16,  -5,   0, -28,  -5,  -3,  -7,  -3, -14,   4,   1 },
      {   1,   2,  -3,   2,  31,  16, -26,  12,  -3,  -3,   2,  -9, -10, -19, -24,   2,   3,   3,  -7,   5, -32,  23, -16,  11,  -5,  14,  16,  13,  31,  16,  -2,   7,  -3, -30, -13,   0,  -6,  12,   9,   1,  42,  29, -24,  -5,  -8,  -2,   3,  -3, -39,  15,  17,   8,  -9,   8,   4,  -2, -33,  29,   6,   5,  20,  10,  -2,  -5 },
      {  -2,  -3,  -2,  -2, -44,   1,  21, -41,   3,   8,   0,  22,  27,  15,  21, -13,  -2,  -1, -11, -30,  16, -16,  13,  -6,   0,   4,  26,  -4, -17, -22,  11,  -4,   7, -22,   0,  15,   8, -10,  11,   1,  31,  19, -12,  17,   6,   5,  13,  -8, -46,  12,   2,   1,  16,  23,   1,  -4,   6,  24,   9,  -8,   3,   7,  -8,  -2 },
      {  -1,  -1,  -2,  -5,  10, -27, -12,  32,   1,  -1,   5,   6, -27,   4, -10,  18,   0,   4,  -8,  11,  26,  27,  -2,  11,  -3,  -1,   1, -34, -17,   2,  12,   3,   7,  -2,  15,  19, -31, -19,   0,   2,   1,  -5,   0,  21,  -3, -30,  10,   1, -22,   9, -14,  35,  26,   8,   9,  -1,  35,   8, -21, -33,  24,   1,   0,   0 },
      {   2,   1,   0,   4,   7,  26,  10, -18,   1,  -1,   2,  -7, -19,  -7,  18,  -9,   0,   1,  -3,  25,  11, -12, -13,   1,   0,   1, -11, -25,  32, -38,   5,   4,   8,   8,  29,  -4,  13,  32,   8,   0,  -3, -16,   5,  -3,  16,   6,   9,  -2,  10,   5,   3,  14, -34,  41,  14,  -4,  17,   0,  -7,   8,  44,  33, -14,  -7 },
      {   0,   1,  -2,  -1,   1,  10,  10,   0,  -1,   1,   3,   8, -24,  17,  -8,  -3,  -3,   2, -13,   1,  39, -39,  -7,  -3,   3,  -3,  21, -31,  -6,   8,  14,  -1,   1,  -7,  -2,  37,  -4,  36,  -2, -11,  -7,  -1, -22, -31, -25,  13, -16,   2,  15,   2,   1,  25, -10, -27,  -9,   7, -41,  23, -15,  -9,  -6, -19,   2,   2 },
      {   0,   0,   0,   0,   4, -12,  -4,   4,   0,   1,  -1,   0,   9,  20,  -2,   2,  -1,  -4,   4, -11, -10, -30,   6,  10,   5,   3,  -1,  22,  -2,  20,   8,  -3, -12,   3,  -6, -13,  29,   0, -24,   8,  11, -14,   9, -21, -24, -14,  25,   7,  10,  41,  -2,  41,  10,  33,   1,  -5, -22, -44,  19, -34,  22,  -1,  -9,   3 },
      {   0,   0,   0,  -3, -11,   2,   3,   7,   0,  -1,  -2,   8,  16, -39,  23, -10,  -1,  -1,   0, -21,   2,  14, -23, -14,   2,  -8,  10,   6, -22,  -9, -29, -13,  -4,   3, -22,  15,  -8,  15, -49,  -5,  -7,   7,   5, -11,   6, -43, -26,  14,  17, -11,   3,   0,   3,   7,   1,  11,  -4,  21,  16,   2,  52,  -9,  -8,   6 },
      {   0,   0,   1,  -1,   5,  29,  -1, -12,  -2,  -1,  -3,   0, -11, -29,  -5, -14,   1,  -2,   3,  -2,  12, -16, -12, -17,  -2,  -2,   0,   6,   7,  15, -34,  -7,   5,  -4,  -8, -14,   7,  11, -12,   0,  -3,  19,   8,   2, -30,  28, -14,  -1, -44,  -5,   7,  20,   2,  -7, -24,   6,  53, -36, -27, -15,  -1, -25,  -8,   5 },
      {   1,   1,   3,  -5,   7,  59, -19, -22,  -3,  -3,  -6,   5, -19, -52,   5, -25,   1,   0,   5,   3,  20,   5,  20, -14,   0,  -1,  -9,  -7, -11,  23,  12,  -5,  -3,   3,  16,  -7,  -5, -25,   6,   1,   6, -25,  -2,  23,  -2, -13,  10,  -3,  13,  20, -32,  -6,  15,   7,  -7,  -4, -28,  10,  18, -12, -14,  -7,  -6,   1 },
      {   0,   0,  -1,   3,  -4, -26,  22,  19,   0,   0,   0,   0,   2,  -1,  -1,  -8,  -2,  -3,   1,  -1,  10,  13, -56, -18,  -3,   0,  -5, -15,  -6,  12, -54,  -4,   2,  -1,  23,  15,  -7,   6, -13,   8,   6, -14, -25,  23,  10,  32,  11,   4,  -4,  35, -14, -14,  -5,  20,  -4,   0, -24, -15,  10,   0, -28,  13,  -3,   1 },
      {  -1,  -1,  -1,  -2,   3, -10, -32,  12,   0,   0,   1,   4,  -1, -14,  15,   9,   1,   1,   0,  -6,   8,  55,  11,  -3,  -3,  -1,  -1,   3, -34, -48,  -5,  -7,   2,   1,   1,  12,  48,  -5,  16,   3,   4,  -7,  -1, -28, -10,  31,  15,  -9,   0,   2,  -2,  13, -28,   4, -15,  -8, -12,   1,  -8,  -4,  -4, -34,  -9,  -2 },
      {   1,   1,   1,   2,   5,  18,  24, -17,  -1,   0,   1,  -2,  -9, -24, -16,  11,   2,   0,   2,   0,  21,  10,   9,   9,   2,   1,   3,  -6, -31,  25,   0,  20,  -2,   1,  -8,  35,   7, -34,   2,  17,   0,  11, -17, -37,  32,   4,  10,  13,   2,  11,  47,  -5,  -7,  -4,  20,   6,   8, -38,  -2,  14,  10,  21,  14,  -6 },
      {   1,   1,   0,  -1,   0,   6,  18,  -7,   1,   0,   1,   4,   1,  -5, -16,  13,   2,   1,   3,  -8,   4,   2,   7,   2,   0,  -3,   0,  14, -11,  19, -40,   1,  -3,  11, -15, -10,  15,  -9,  -4,   7,   6,  -3,  40,  -9,  -2,  28,  12,   2,   6, -26, -28,  38,  -2,  -1,  28,  -2,  -6,  52, -20, -15,  -8,  46,  16, -10 },
      {  -1,  -2,  -1,  -3,   5,  -5, -60,  12,   1,   2,   1,   2,  -4,  14,   6,   2,   0,   1,  -1,   3,   5, -23,  62,  -3,  -1,  -4,   4, -16,   2,  -5, -28, -16,   2,  -7,   3,   9,  -9, -14, -45, -12,   3,   4, -10,   4,  21,  24, -33,   0,  -4,   9,   7,  -4, -10,  15,  -7,   6,  -4,  -5,  -3,   7,  -5,  33,  -1,  -2 },
      {   0,  -1,  -1,  -1,   7,  -5, -39,  14,   1,   0,   1,   1,  -7,   7,  53, -36,   0,   0,  -2,  -2,   3, -15, -21, -22,   1,   1,   9,   1,   5,  25,   9, -12,  -2,   0, -12,   9, -11,   2,   3,  18,  -1,  14,  11, -14,  16, -10,  50,  18,   2, -25,  19,   2,   0,  21,  22, -11,   2,   3, -29,   8, -30,  -9,   5,  -3 },
      {  -1,  -1,  -1,  -1,   1,   2, -27,   0,   1,   1,  -1,   3,  -1,  -5,  32, -11,   0,  -1,   0,  -4,  -1,  11, -13, -12,   0,   3,  -3,   5,  -5, -22,   2,  16,   4,  -1,   7,  -2,  24,  22,  18, -11,  -5,   0, -13,   1, -36,   0, -22,  12,   4,  17,  11,   2,  49, -32,  37,  22,   4, -14,   9,  -8,  -8,  48,  34,  -1 },
      {   0,  -1,   0,   1,  -2,  14, -23, -36,   2,   3,   2,   2,   2,  -3,  41,  64,   1,   1,   1,  -2,   3,  -5, -12,  64,   0,   0,  -1,  -1,   1,   4, -24,  38,   0,   2,   0,   4, -10,  19, -12,  14,  -4,   1,  -5,   9,  -6,  -6,   6,  10,   3,  -1,  -5,  -7,   8,  10,  -6,   2,  -3,   2,  -5,  -1, -21, -16, -10,  -1 },
      {   1,   1,   0,   1,  -1,  11,  19,  -2,  -1,  -2,   0,  -1,   2, -17, -29,   4,   0,   1,  -2,   1,  -3,  23, -11,  12,  -1,   0,   2,  -2,   0, -24,  38, -16,   1,  -4,   0,  -3,   9,   9, -50, -28,  -1,   5,   2,  -3,  -7,  -8,  -9,   1,   5,  -6,  18,   7,  30,  44,   3,  -6,  -6,  -3, -32,   9, -57,  11,  -8,  -6 },
      {   0,   0,   0,   0,  -2,   3,  14,   5,  -1,  -1,  -1,  -3,   4,  -3, -23, -19,   1,   2,   2,   4,  -5,  10,  43,   6,   1,  -2,   2,  -7,   9, -20,   0,   8,   3,   1,   6,   0,  -6,  33, -29,  42,   0,   5,  -6,  19,  -7,  16,  29,  46,  -5,   3,  -9,  -8,  -3,   3,  34,  31,   2,   4,   6,   0,   1, -38,  34,  13 },
      {   0,   0,   0,   1,   4,   8,  -5, -21,   0,   1,   0,  -2,   0,  -3, -10,  62,  -1,   0,  -2,   2,  -5,  -6,   7, -30,   0,  -1,   0,  -4,   4,   0, -18, -72,   0,   1,   2,   7,  -5,  22,   9, -29,  -2,   1,  -8,   3,   2,  -9,  36, -13,   2,   3,   6,  -9,   9,  -7,  28, -14,   3,  -9,   8,   0,   7, -13,  24,   1 },
      {   0,   0,   1,   0,   1,   2, -10,  -6,   0,   0,   0,   0,  -1,  -1,  19,   8,   0,   1,   0,   1,   3,  -3, -40,  15,  -1,   1,  -1,   1,  -1,   9,  44,  -6,   1,  -2,   0,  -8,   1, -30, -33, -17,   3,   2,   6,   8,   8,  41,  -6, -13,  -4,   1, -11,   2, -18,   7,  -9,   7,   3,   5,  10,  -1,  22, -11,  76,  19 },
      {   0,   0,   0,   2,   1,   0,  -5,  -6,   1,   1,   0,  -1,   1,   2,   3,  53,  -1,  -2,  -1,  -1,  -5,  -3, -13, -65,   2,   2,   2,   4,   6,   5,  31,   8,   0,  -3,   1,  -8,   2, -20,  11,  18,   5,  -1,   6,   2,   7,  11, -35,  47,  -2,   3,  -6,   8, -11,  18,  13,  39,  -5,   6,  -4,  -2,  -6,  -8, -30,   8 },
      {   0,   0,   0,   0,  -1,   1,  -4,  -1,   0,  -1,  -1,   0,   1,  -4,   7,   0,   0,   1,   0,   1,   0,   3, -15,  21,  -1,   0,  -1,  -2,  -2,   0,  31, -29,   2,  -3,   3,  -3,   2, -10, -38,  -1,  -2,   5,  -4,  12,  -8,  35,  32,  29,  -3,  -4,  -6, -12,  -4, -60,  12,  -6,   2,  -2,  10,  -2,  16,  23, -61, -17 },
      {   0,   0,   0,   0,   0,   2,  -1,  -2,   0,   1,   0,   0,   0,  -3,  -1,  23,  -1,  -1,  -2,   1,  -1,   3,   2, -39,   2,  -1,   3,  -3,   4, -10,  17,   7,   0,   1,   1,   2,   0,  14, -27,  72,  -4,   1,  -6,   1, -13,  -8,   2,  -9,   0,  -3,   1,  -4,   6, -13, -46, -54,   2,   2,   2,   0,   2,  27,  26, -26 },
      {   0,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   1,   1,   1,   1,   0,   0,  -1,   1,  -1,   0,  -1,   2, -27,   1,   0,   1,   0,   2,   3,   1,  52,  -2,  -1,  -2,   0,  -7,  -2, -25, -21,  -1,   2,  -2,   7,  -2,  16, -16, -31,   0,   2,  -5,  -3,  -4,  -2,  67, -50,   0,  -6,   7,  -5,   5, -38, -10, -38 },
      {   0,   0,   0,  -1,   2,   0,  -2,   0,   0,   0,   0,   0,  -1,   0,  -3,  12,   0,  -1,   0,   2,   0,   3,   8, -41,   1,   0,   0,  -1,   3,  -6,   2,  55,  -3,   2,  -4,   1,  -4,   9, -31, -30,   2,   1,   4,  -3,   3,  -3,  60, -39,   1,  -4,   4,  -3,   4, -15, -23,  37,   2,   3,   0,   2,   0,  17, -12,  41 },
      {   0,   0,   0,   0,  -1,  -1,   3,   0,   0,   0,   0,   0,   1,   2,  -1,   0,   0,   0,   0,  -1,  -1,   0,   5, -11,   0,   0,   0,   1,   0,  -2,  -6,  20,  -1,  -1,   0,  -2,  -2,   1,   4, -54,   3,  -1,   3,   2,   4,  -8,  23,  55,  -4,   0,  -6,   0,  -6,   3, -42,  10,   2,  -1,   2,  -2,   4,  -4,  25, -81 },
      {   0,   0,   0,   1,   0,   0,  -1,   2,   0,   0,   0,   0,   1,   1,   4,  -1,  -1,   0,   0,   0,   0,  -2,  -6,   5,   0,  -1,   0,   1,   1,   3,   7, -20,  -1,   1,  -1,  -1,  -3,   0,  -7,  34,  -1,  -2,   1,  -1,   1,  -3,   1, -64,   3,   1,   1,   0,   2,   2,   8,  74,  -1,  -2,  -1,  -1,  -3,  -8,  -4, -69 },
    },
    {
      { -96,  47,   4,   4,   1,   1,   0,   1,  65, -15, -12,  -2,  -2,   0,  -1,   0,  -3, -17,   5,   3,   1,   1,   0,   0,   0,   4,   5,  -2,  -1,  -1,   0,   0,  -1,  -1,  -1,  -1,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0 },
      {  49,  -2, -13,  -1,  -3,   0,  -1,   0,  57, -54,   3,   2,   1,   0,   0,   0, -71,  10,  29,  -3,   4,  -1,   1,   0,  -5,  34,  -9,  -9,   0,  -1,   0,  -1,   0,  -3, -12,   4,   1,   1,   0,   0,   1,   1,   1,   3,  -2,   0,   0,   0,  -2,  -1,   0,   0,  -1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0 },
      {  35, -17,   2,  -3,   0,  -1,   0,  -1,  43,  11, -23,  -1,  -3,   0,  -1,   0,  35, -67,   0,  11,   1,   3,   0,   1, -54,  -1,  43,  -3,   2,  -1,   1,   0,  -8,  35,  -4, -15,   0,  -2,   0,  -1,  -3,   1, -13,   2,   4,   0,   0,   0,  -1,   0,   0,   4,   0,  -1,   0,   0,  -2,   0,   1,   0,  -1,   0,   0,   0 },
      { -26, -87,  47,   5,   8,   2,   3,   1,  39,   8,  21, -16,  -2,  -4,  -1,  -2,   3,  36, -25, -14,   2,  -1,   1,  -1, -26,   8, -19,  11,   5,   1,   1,   0,  -1,   3,   0,   9,  -5,   0,  -1,   0,   1,   0,   1,  -1,  -3,   1,   0,   0,   0,   1,  -1,  -1,   1,   1,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
      { -26,  -7,  10,   2,   2,   0,   1,   0, -41,  17,   3,   3,   1,   1,   0,   0, -10, -17,  27,  -1,   1,  -1,   0,   0, -27,  67,   3, -21,   0,  -4,   0,  -2,  36,   9, -56,   3,   3,   1,   0,   0,   6, -36,   3,  23,  -1,   2,   0,   1,   6,   0,  16,  -4,  -6,   0,  -1,   0,   0,   0,  -2,  -5,   1,   1,   0,   0 },
      { -19,   3,  -6,   1,   0,   1,   0,   1, -31, -58,  44,   7,   6,   1,   2,   1,  17,  20,  21, -18,  -1,  -4,   0,  -1, -24,  -1,   8,  -7,   7,  -1,   2,  -1, -22,  55,   7, -14,   3,  -5,   0,  -2,  15,  21, -45,   0,   1,  -2,   1,   0,   5, -16,  -7,  20,  -2,   2,   0,   0,   4,  -2,   8,  -1,  -5,   0,  -1,   0 },
      { -12,  -3,  17,  -5,   2,   0,   0,  -1, -14,  27, -12, -11,   2,  -2,   0,  -1, -74,   5, -25,  25,   6,   4,   1,   2,  10,  -7,  40,  24, -13,   0,  -3,   0,  24,  28,  15, -38, -12,   0,   0,  -1,  10,   0, -35,  -9,  15,   4,   1,   0,   4, -11,  -1,  18,   3,  -3,  -2,   0,   3,  -2,   6,  -1,  -4,  -1,   1,   1 },
      { -15, -40,  37,   0,   3,  -2,   1,   1, -20, -20,  -8,   6,   5,   2,   0,   1,  -4, -35,  39,  20,  -4,   2,  -1,   0,  40,  16,  22, -28,  -9,  -1,  -1,  -1, -12, -21,   8, -13,  13,   2,   2,   1, -21,  47,  11, -13,   7,  -7,   1,  -2,   1,  17, -36,  -1,   2,  -2,   1,   0,  -2,  -5,  -3,  14,  -1,   1,   1,   0 },
      { -20, -41, -74,  38,   4,  12,   1,   4,  13,  42,  16,  19, -17,  -1,  -5,   0, -19,  -4,  36, -26, -10,   2,  -1,   1,  13, -26,  10, -16,  14,   3,   0,   1,   8,   8,   3,  -7,   8,  -7,  -1,  -1,  -6,   2,  -4,   0,   2,  -3,   2,   0,  -1,   4,   0,   1,  -2,  -1,   0,   0,   1,  -1,  -3,   0,   1,   1,   0,   0 },
      {   9,  -8,  12,  -8,   2,  -1,   0,  -1,  22,  -1,   1,  -6,   2,  -3,   0,  -1,  18,  22,  -9,  13,  -8,   1,  -2,   0,  58, -17,  16, -31, -10,   0,  -2,  -1, -29,   9, -42, -26,  25,   2,   4,   1,  -9, -22, -21,  48,  13,  -5,   1,   1,  -5,  -3,  32,  11, -20,  -4,   0,   0,   1,   4,   1, -16,  -2,   6,   1,   0 },
      {  11,  23,  21,  -5,   0,  -3,   1,  -2,   7,  59, -21,  -7,  -9,  -3,  -1,  -2, -14,   5,  11, -20,   2,   1,   1,   0,   0,  14, -29, -22,  16,  -2,   2,   0, -36,  10, -12,  26,  13,  -7,   3,  -1,  13,  40, -23,   9, -16,  -6,   0,  -3,  14, -34, -27,  18,  -7,  10,   0,   1,   4, -14,  25,   5,  -6,   4,  -3,   1 },
      {  -7, -19, -18, -10,   4,   1,   2,   0,   1,  -8, -64,  39,   5,   7,   1,   2,  19,  46,  19,  34, -27,  -5,  -6,  -1, -29,   2,  -6, -17, -22,   9,   0,   2,  16, -21,  13,  -7,   9,  11,  -2,   1,  16,  -1, -10, -10,  10,  -4,  -3,   0,   0, -25,   4,  17,   4,  -1,   0,   1,  -1,   2,  21,  -5,  -7,  -1,   0,   0 },
      {  14,   8,  -1,  -5,  -1,   2,  -1,   1,  21,   8,  15,  -6,  -5,  -2,  -2,  -1,  35,  -5,   7, -29,   7,  -1,   1,   0,  26,  30, -25,  10,   0,  -5,   0,  -1,  47, -19,   2, -28, -16,   3,  -3,   1, -32,  -7, -36, -20,  34,   6,   3,   3,  -9,  -6, -14,  44,  11, -10,   1,  -1,  -1,   0,  16,  10, -18,  -4,   1,   0 },
      { -11, -31,  -1,  -8,   7,   2,   1,  -1, -14, -25, -28,  16,  -1,   3,   0,   2,  -8, -51,   4,  -4,  10,   2,   2,   1,  31,  -5, -31,  30,  17,  -4,   1,   0,  -9,   6,  14,  29, -18,  -8,  -2,  -2,  -8,  -9, -13,   7, -17,   7,   2,   1,   1, -28,  42,  18,  -4,  11,  -3,   1,   4,  19,  27, -29,  -4,  -2,  -3,   0 },
      {  -7, -17, -55,  10,   5,   8,   1,   2,   0,  -2,  -4,  -8,   3,  -4,   1,  -1,  16,   0, -34,  19,  19,  -3,   1,  -1,  20,  55,   3,  28, -24, -12,  -2,  -3, -47,  -3, -17,  -9, -24,  13,   3,   3,  21,   9,   5,   1,   7,  11,  -5,   2,  20, -11, -17,  -5,   5,  -5,  -4,   0,  -1, -20,   5,  12,   2,  -2,   1,   2 },
      {  -9,  -3, -26, -84,  32,  -5,   7,  -4,   6,  24,  42,   2,  25, -11,   1,  -3, -12, -14,  10,  29, -19, -10,   0,   0, -13,  -2, -29,   1, -17,  11,   5,   0, -10,   8,   5,   1,   4,  10,  -3,   0, -21,   9,   4,   0,   6,  -2,  -2,   1,   1,   9,   5,   3,  -4,  -3,   1,   0,  -1,   2,  -3,  -3,  -1,   0,   1,  -1 },
      {  -2,  12,  15,  25,  -3,   0,  -1,   2, -12,  13,   4,   6,  -9,   1,  -1,   0, -14,   2,  20, -17,   3,   1,   1,  -1, -40,   7,  -3,  21,  -9,   1,  -3,   0, -51, -40,  19, -28,  -6,  14,   2,   2, -25,  11,   3,  16,  36,   4,  -1,  -1,  20,  15,  33,  15, -17, -17,  -6,  -3,  -4,   6,  -6, -34,  -8,   4,   6,   2 },
      {  -9,  -7, -28,   5,   9,   1,   1,   2, -15, -27, -11, -21,  11,  -4,   3,   0, -12, -13, -47,  -4,  11,   3,   3,   0, -27, -12,  -3,   2,  10,   5,   1,   1,  27, -26, -10,   8,  38,  -6,   1,   0, -21,  23,   0,  42,   5, -25,   5,  -3,   0,   9, -20,  26, -34,   0,   7,  -2,   6, -15,  18,  10, -16,  18,  -2,  -1 },
      {  -7,  -3,  -6, -39,  16,  -9,   3,  -4,  -7,  -6, -48,  37, -10,   1,   4,   1,  -7,  17, -23, -42,   1,  15,   1,   1,   2,  17,   4,  -5,  36,  -6,  -2,  -1, -20,  17,   2,  -7,  -3, -17,   0,   0, -26, -11, -12,  -2,   6,   7,   9,   5,   6,  38,  -6,   0,   4,  -4,   1,  -3,  -2,  -5, -41,  19,   7,  -3,   1,   0 },
      {   7,   7,  12,   6,  -2,   2,   3,   0,   9,   2,  41,  63, -19,  -1,  -7,  -2,  -2, -18, -49,   0, -31,   9,  -2,   4,  11,  35,  12, -21,   3,  20,  -3,   2,   6, -12,  33,  -2,  17, -10,  -9,  -1,  13,  -8,   2,  -2,  -6, -14,   4,   0,  24, -11,   1,  -6, -10,   4,   3,  -1,  -4,  -8,  14, -16,  -2,   5,  -2,   0 },
      {   8,   5,  -2,  -8,  15,   5,   0,   1,  12,   7, -10,   1,  -1,  -4,  -2,   2,  23,  23,  22,  -5,  14, -11,   4,  -2,  16,  20,  16,  34, -11,  -6,   2,  -3,  43,  13,  15,  -3,  -6,  -9,   0,  -2,  -5,  45,  -8,  19, -26, -14,   3,  -4,  25,   7,   8, -24, -39,  13,   0,   0,  -9,  -8, -22, -39,  19,  17,  -3,   1 },
      {  -3,   4,  -9,  19, -50,   7,  -4,   5,   2,  12,   3,   5,  24,   8,  -1,  -1,   0,  22, -12,  36,  16, -13,  -6,  -1,  -6,  19,   3, -20,   1,  -3,   1,   2,  -8,  23,  14,  14,  -5, -11,   1,  -3, -76,  -3,  -3,  -7, -30,   6,   6,   1,   1,  17,   4,  11,  10,  16,  -3,   0,  16,  -3,  12, -13,  -5,  -6,  -7,   0 },
      {   8,  14,   7,  43,  72, -18,   3,  -2,  -2,  -6,  -4, -12, -11, -22,   5,  -4,   4,  13,   8,  20, -49,  16,   6,   3,   3,   5,   5,  23,   5,  19,  -7,  -2,  -5,  20, -12,  12,  -8, -10,  -8,  -1, -35,  -7,  11, -10,  -5,   5,   3,   4,   8,   4,  -1,   3,   1,   6,  -2,  -1,  12,  -9,  17,  11, -13,   0,  -3,   0 },
      {   3,   2,   9,  39,  -8,   9,  -6,   7,  -1,   1, -16,  14,   2,  -2,  -1,   2,  -4, -18, -10,  17,  -5, -10,   0,  -3,   6,   2, -72,   3, -19,   9,   6,   1,   6,  51,   9, -16,  18,  21,  -7,   0,  -2,   2,  -1,   4,  20, -20, -14,   4,  -6,   5,  -4,  -6, -16, -16,  12,   3,  -2,   3, -38,  16,  12,   8,   7,  -3 },
      {   3,   2,  17, -12, -11,  27,  -2,   8,   4,  -7,   5,  48, -22,   1,  -7,  -1,  -4,   2,   3,   1,   9, -17,   0,   3,   1, -39,   2,  11,  -9, -11,   9,   0,  17,  16, -36,  18, -30,  16,   9,  -2, -13,  10,  22,  14,  18,  19, -10,  -5,  42,   5,  -7,   0,  -2, -14, -13,  -1,  11, -35,  20,  16, -21,   1,   4,   3 },
      {  -3,  -4,  -1,   1,  -5, -24,   3,  -9,  -4,  -9,   3,   1, -15,   9,   3,   2,  -6, -10,  -4, -12,  -3,   3,  -1,   1, -13, -19, -12,   3, -32,   9,  -4,   4,   2, -22, -16, -24,   9,   6,   4,   2, -42,  -4, -27,   7, -18,  12,   7,  -1,  28, -43,  -1, -29,  14,  33,  -8,   4, -19,  -9,  -6,  27,  49, -12,  -8,   1 },
      {   0,  -1,  -7, -18,   4,  44,  -8,  17,  -5, -12,  -9, -56, -45,   9,  -4,   5,   4,   2,   2,  17,  -1,   7,   2,   1,   0,   3, -15, -26,  27,   8,  -7,   2,   7,  11,  23, -15,  17, -11,  -1,   0,  -2,  -8,  14,  -2,  13,   1,   3,   2,  57,  -4,   6,  -6,  18,  -7,  -4,   2,  -6,  -3,   3, -14,   3, -18,   1,   1 },
      {   0,   2,   3,  -4,  12,  -7,   0,  -7,   2,   6,  18,  22, -43,  10,   2,   2,   5,  10,  18,  49,  40,  22,  -9,  -1,  -9,   0, -14,  11,  26, -34, -17,  -2,  -6,  -4,  14, -26,  -5, -20,  16,   4,  -1, -15,  12,  19,  -9,  -4,  11,  -4, -21, -12, -13,   2, -28,   9,   5,  -2,   1,  16,   5,  32,  -2,  13,  -4,  -2 },
      {  -1,   7,   2, -10,  19,  60, -12,  23, -10,  -3,  -4,   6, -41, -12, -10,   2,  -6,   3, -18, -14,  -4, -20,  13,   2,  -9,   8,   8,  -5, -12,   4,  10,  -2,  -9,  -2, -16, -14,  -9,  11,  -4,  -2, -21,  20,   4,  -5,  -7,  -5,  -4,   2, -66, -12,  -1, -16,   6,  14,   6,   8,  15,  17,  12, -13,  16,   5,  -7,  -1 },
      {   2,   2,   9,  -1,  26,  29,  -1,   7,   1,   6,  -5,   1,  29,   0,  -6,   3,   1,   6,  19, -27,  35, -30,   7,   1,   6,   8,  27,   7, -19,  11,  10,  -4, -14,   5,  38,  13,  23,   5,  -9,  -4, -12, -45,   8,   6,   4, -19,  -5,   3,  16, -16,   2, -10, -15,   2,  10,   1,  -4,  10,  19,  48,  -6,  -1,   0,  -5 },
      {   5,   4,   9,  16,  19,  47,  -9,  25,   2,   1,   8,  11,  31, -14,  -9,   0,  -1,  -7,   2,  30, -13, -27,   7,  -4,  -5,  -9,  -3,   3,  30,  -5,   8,  -1,  -4, -35, -10,   1, -12, -16,   5,  -1,  15,  -5, -35,   9, -26,  15,  11,   4,  10,   9,  -2,  30,   5,   4,  -4,  -3, -31,  -6, -36,  15,  16,  -6,   3,   1 },
      {   4,   6,  -1,  20,  51, -19,  12, -12,   1,   1,   3,  13,  -6, -16,  -1,  -1,   0,   1, -18,   1,  52,  -6,  -3,   3,  -3,  -3,  -5, -46, -20,  -7,   7,   5,  15,   4,  13,   7,  -4,  18,  12,  -2,   7,  23,  -3,  21,  -7,  14,  -7,  -4,   7,  17,  10,  10,  35,  -4, -13,  -1,  -7,  32,   0, -10,  -4, -31,   1,   3 },
      {  -2,  -4,  -1,   0,  11,  -6,  -5, -11,  -5,  -8,  -7, -17, -14, -29,   4,  -1,  -5,  -8,   1,  -4,   5,  -8,   1,   0,  -4, -14,  -8, -43, -22, -16,   7,   0,  -8,  -5,  18,   7, -43,  15,  10,  -1,  -1, -46, -18, -18,  -6,  -4,   1,   1,   7,  12, -32,   1, -33,  11,  11,   1,  -8, -28,   0, -31,   5,  41,  -2,   1 },
      {   2,   1,   3,   4,   3,   8,   2,  -6,   2,  -1,  -7,   5,  59, -33,   4,  -5,   0,  -1,  -2, -12,  -1,   3,  -9,   4,  -2,  -4, -12, -22,  35,   5, -13,   4,   8,   3,  10, -46, -37, -15,   4,   6,  -4,   7,  31,  26,  14,   7,   4,   3,  -2, -36,  -9, -32,   4,  -4,  -6,   1,  20,  -5,  13,  -9,   7,  -1,   0,   1 },
      {   1,   3,   0,   6,  33,  -4, -17,  28,   1,   1,   5,   4,  27,  45,  -5,   0,  -4,   2, -14,  -6,  -5,  14,  -8,  -8,  -1,  -1,   7,  -9,  11, -61,   7,   4,   3,   8,  -4,   7,  23,  11,  19,  -1, -21,   0,  12, -30,   9,  13, -17,  -4,  -6, -31,  -2,  -6, -13, -20,  -4,   6, -28, -11,   2, -24,  10,   2,  16,   5 },
      {   3,  -1,   2,   1,  13,  -8, -19,   8,   4,   1,  -2,  -1,   7,  14,  -5,   4,   0,  -9,  -3, -12, -14, -28,   2,  -4,   0,  -1, -13,   1, -12, -36,   9,   1,   0,  -8,   2, -32,  24,   0,  16,   5,  21,  -6,   5, -15, -18,  13,  13,  -3,  20,  19,   7,   5,  -6,  33,  -6,  -5,  81,   0,  -6,   2, -11,  -7, -19,   3 },
      {  -2,  -1,  -2,   0,   0, -13, -12, -62,  -1,  -2,  -2,  -4, -15,  20,   5,   4,  -3,   0,  -4,  -2, -21, -69,   2,   9,   3,   3,   6,  11,  33,  -8,   4,   1,  -2,   3,   0, -23,   1,  27,  -5,  -5,  -8,  -6,  18,  13, -28, -12, -12,   2,   1, -10, -18,  11,  13, -14,   2,   6, -21,   7,  -1, -10, -13,  -1,  11,  -1 },
      {   1,   0,   6,   0,  19,   3,   3,  -3,   2,   0,   6,  18,   5,   4,   5,   0,   2,   0,  11,   7,  45,  11,  -7,  -4,  -1, -12,   2,  16,   8,  18, -14,   2,  -6,  -7, -19,  -4,  18,   8, -27,   6,  -1, -18,   8, -14,  -8, -40,   0,  14,   0, -12, -16,  24,  34, -10,  17,   2,  31, -43, -17, -41,  25,  -8,  -4,  -2 },
      {  -2,   0,  -1,   0,  -1, -16, -43,  14,  -1,   2,  -5,  15,  11, -17,  -8,   8,  -2,   1,  -1,  -3,  16,  -8,  19,  -4,  -2, -10, -18,   8,  -5,  -4,  -6,  -7,  -6,   7, -28, -30,  11, -23,  -6,   5,   8,   1,   4, -24,  -1, -25,  21,   8,  11,  31,  -2, -22,  13,  18,  14,  -8, -51,   9,  44, -14, -21,  -6,  -6,  -6 },
      {  -1,  -1,   0,  -6,   1, -14, -60,  38,   3,   5,   3,  -1,   2,  30,  18,  12,   0,   0,   5, -10, -23,  21,   6, -14,   3,   3,   5,   2, -13,   5, -13,   1,   2,   6,  23,   9, -32,  22,   3,  11,   3,  -7,   5,  58, -11, -17,  -4,   2,  -2,  10, -23,  18,  24,   1,   4,  -2,   3,   7,   5,   3,  13,  -5,  -4,   0 },
      {  -2,  -4,  -5,  -4, -12, -27, -49,   6,   1,   0,   6,   5,  -9, -46,  24,   3,   3,   5,  -5,  14,  13,  -7,  45, -11,  -1,   0,  17, -11,  26,  -4,   0, -14,   5, -10,  -6,  17,   5,  18, -21,   5,  -8,  12,  -2, -21,  14,   9, -17,  13,  17, -29,  14,   2, -21,  -8,  10,   8,  22,  23, -20,  10,   9,  -3,   8,  -6 },
      {   1,   2,   2,  -3, -12,  -2,  19,  14,   0,   0,  -2,   1,   3, -26,   9,  -9,   1,   4,   3, -10,  -1,  16,  -5,   2,   5,   3,   2,   5,  19,   3,  11,   1,   3,  14,   3, -26,  13,  47,   2,  -8,  10,   0,  37, -11, -21,  17, -12,  -6,   4,  22,  12,  47, -17,  17,  -2,   5, -20,  -3,  32,  11,  50,   3, -21,  -1 },
      {   0,  -1,   1,   3,  10, -11, -25, -17,   0,   3,   2,  -4,  -2,  18,  -4,   9,   2,  -1,   5,  14,  12,  -8,   7,  -4,   1,   8,  -5,  -9,  28,  12,  16,  -3,   3,  -7,  22,   1,   1,  10,  -8,  -7,  10,   8, -27,   9,  19,  12, -17,   5, -30,  25,  38, -29,   1,  10,  -7,   9,   2, -71,  26,  16,  12,  -8,  -3,   5 },
      {  -1,  -2,  -2,  -6,  -1, -36,  28,  65,  -2,  -3,  -3,  -8, -15,  -3,  -7,  -3,  -1,  -2,  -2,  -6,   6, -20, -14, -25,   0,   0,   0, -10,   7,  16,   3,   9,   0,  -6,   2, -27, -16,   9, -10,   1,  -3,   3,   3,  -4, -44, -12, -11,   2,  -5, -14,  23,  -6,  -4, -17,   9,   1,  -3, -22, -11,  14, -43,  -1,  22,   0 },
      {  -1,  -1,   1,  -3,   2,  21, -34, -39,  -1,  -1,  -1,  -5,   2,  -2,  -6,   4,  -1,  -2,  -3, -16,   0,  20,   9,   2,  -3,   0,  -1, -19, -39,  -3, -32,   3,   3,   7,   5,  -8,  -9, -19,  -7,  14,   1,  15,  29, -13, -34,   6,  17,   1, -11, -14,  40,  35, -15, -20,   2,  -7, -12, -31, -15,  23, -10,  -8,   5,   5 },
      {  -2,  -3,   1,  -6, -15, -15,  -6,  12,  -3,  -2,  -4,   3, -14, -57, -16,   5,   1,   0,  12,  -2,  -6,  -8,  17,   0,   1,   6,   1,  23, -14, -14,   7,   2,   0,   8,  24,   6,  43, -17,  22,  -2,  -2, -21,  11,  26,   2,  23,   6, -18, -26, -13, -23,   5,  17,  -8, -26,  -4,  -8, -32, -10, -20,  -6, -24,  11,  17 },
      {   0,   0,   1,  -2,  -9,  14,   0, -14,  -1,  -1,  -1,  -2,   1, -39, -26,  -2,   0,  -1,   4,  -2, -20,  24, -27,   4,  -4,   1,   1,  10,   0, -49, -15,   7,   1, -11,   4,   4,   5,  17,   1,   3,  -5,  -1, -33,   5, -26, -34, -24,   9,  11,  20,  11, -22,  12, -35,  21,  16,  17,   6,  23,  14,  10, -17,  16,  -3 },
      {  -1,   0,  -2,   0,  -2, -29,  -6,   5,  -2,  -3,   1,  -8, -10,  -2, -53,   2,  -2,   3,  -3,   3,  -9, -16, -24,  14,  -1,   0,   7,  -7,  -6, -19,  51,   2,  -2,   1,  -1,   7, -27, -37, -23,  -6,   0,   6,  24,   7,  29, -28,   4,  13,   2, -12,   8,  27,  -6,   8,  18,   3,  -1,   0,  -7,   7,  30, -24, -12,  -2 },
      {   0,  -1,   0,  -1,  -2,   5,   9, -12,   0,  -1,  -1,   1,   9,  20, -44,  -5,   2,   3,   4,   1,   3,  43,  53,  -3,  -2,  -6,  -8,  -1,  -6,   3,  37, -14,  -1,  -5,   0, -22,   0,  -1, -44, -10,   0,  -8,  -1,  15, -30,  32, -28,  17,   4,   4, -27,  -4,  -8,  -5,  -2,   9,   5,   9,   3, -10, -17,  -3,   8,  -5 },
      {   1,   1,   3,   2,  -3,  15,  30,  -4,   0,   1,   1,   2,  -6,  -2,  62, -21,  -3,  -3,   1,  -3, -14,  -8,   5, -39,  -1,  -3,   0,  -1, -13, -44,   1,  11,   1,   1,   8,   2,   1, -16, -43,  15,   2,   1,   5,  15,  11,   4,  -9,  28,  -6,   4,  -3,  -1,  -5,  23,  14,  10, -15, -12,   1,  -4, -12, -37, -16,  -6 },
      {   0,   0,   0,  -1,  -2,  14, -14, -14,   1,   1,   3,   1,   3, -16,  23,  -3,   1,   2,   2,   0,  -1,  36,  -8,  -1,   0,   3,   3,   5,   0,  -5,  55,  -7,   1,  -3,   2,  -9,   3,   4,  38, -25,  -4,   4,  10,   2,  -2, -26, -21,   4,   6, -17,  20,   3,  38,  34,  11,   9, -10, -10, -27,   5, -39,  27, -10, -12 },
      {   0,   1,   0,   2,  -2,  -9,  34,   8,   0,   1,   0,  -2,   1,   5, -18, -15,   1,   2,   1,   3,  -8, -11,  60,   5,   0,   1,   1,  -6,   2, -35, -14,   2,   3,   5,  12,   6,   2,  11,  -2,  12,  -1,   8,   5,  13,  10,  -8,  35,   9,   2,  -5,  24,   1,  45, -10,  15, -15,   0, -11,   3,  10,  13,  55, -12, -16 },
      {  -1,  -1,  -1,  -1,  -1,   3, -10,   3,  -1,  -2,  -1,  -2,   5,  -4, -39, -34,   0,   1,   0,   2,  -1,  12,  -7,  21,   1,   0,   4,   0,  16,  -8, -31,   8,   1,   2,   2,  10,  -3,  57, -23,  -5,  -2,   0,  -5,   3,  12,  -7,  18, -12,  -5, -12,  -3, -10, -12,  37,  -2,  -9, -14,  -4, -27,  -3, -39, -27, -42,  11 },
      {   0,  -1,   0,   0,  -2,  -3,  -1,  16,   1,   1,   0,   1,   1,  10,  43, -37,   2,   0,   3,   2,   0,   3,  35,  69,  -1,  -1,  -6,  -3,  -6,  -7,  13,  21,  -2,  -4,  -1,  -7, -10,  -9,   0, -50,   3,  -4,  -5,  -3, -10, -28,  15, -11,  -2,   4,   2,  -5, -12, -21, -13,   7,   7,  -5,   5,   3,   6, -26,  10,   9 },
      {   0,   1,   0,   2,  -3,   0,  35,   3,   1,   1,   2,   1,   5,   4,   3,  65,  -1,  -1,  -4,   0,  -9,   3,  30,   0,  -2,  -3,  -2,  -6,   2, -10, -13, -64,  -1,  -1,  -2,   0, -15,   7,  20,  -7,  -2,  -2,   1,  -6,   0, -38,  -6,  -1,  -3,  -4,   3,   2, -12,  -6, -23,  10,  -3,  -6,  -5,   0,  -4, -33, -34,   5 },
      {   0,   0,  -1,   1,  -1,   6,   8, -18,   0,  -1,  -1,   0,   2,   7, -10, -34,   1,   1,  -1,   3,  -2,   7,  39, -46,   2,   1,   1,   0,  10,  11,  26,  20,  -2,   1,  -5,  -1, -15,   4,  32,  46,  -3,  -4,  -1, -11,   3, -39,  -2, -47,   2,   1,   3,   1, -10,   6,   1, -16,   4,   3,   5,   2,  10, -28,  16,  27 },
      {  -1,  -1,  -1,  -2,   1,  -6,  -8,   2,  -1,  -2,  -2,   0,   0,   2, -19, -49,   0,  -2,  -1,  -3,   2, -16,   7, -23,  -1,  -1,  -2,  -3,  -7,   1, -45,  -6,  -1,  -2,  -1,  -2,  -1, -35,   5, -28,   2,  -3,   3,  -3,   6, -17, -61,  -2,   3,   7,   2,  12,   9,   7, -41,  33,  -2,   5,  -3,   5,   7,  23, -23,   8 },
      {   0,   0,   0,   1,   1,  -1,   8,   2,   1,   1,   2,   1,   6,  -1,  -2,  53,   1,   1,   1,   1,   0,  -1,  22,  13,   0,   1,   1,   1,   0,  -1, -17,  67,  -2,  -2,  -4,  -1,  -4, -14,  -5,  -8,   0,   0,   1,   0,   6,  13, -35, -54,   4,   1,   0,   4,   3,   8,  45,  22,  -3,  -1,  -6,  -2,  -7,   1, -23,  -5 },
      {   0,   0,   0,  -1,  -1,   2,  14,   6,   1,   0,   1,   0,   2,   2,  -6,   6,   1,   1,   0,  -2,  -2,  -2,   9,  41,   0,   0,   1,  -2,   0,  -8, -29, -16,   1,   3,   1,   1,  -4,   1,   1,  20,   3,   2,   8,   2,  11,  -9, -30,   7,  -1,   6,   4,  15,   3,  66,   4,  -5,   1,  -6,  -3,   0,   0,  -6,  81,  -6 },
      {   0,   0,   0,   0,  -1,  -5,   4,  -1,   0,   0,  -1,   1,  -2,   9, -14, -30,   0,   0,   1,  -2,  -1,  -8,  14,  -3,   0,   1,   1,   3,  -2,  22, -18, -11,   2,   0,   2,   3,   0,  -3,  65, -14,   1,   1,   2,   4,   2,  22,  -3,  34,  -4,  -1,  -6,   1, -15,  -6,  63,  10,   2,  -8,   0, -10,   2, -37,  -9, -43 },
      {   1,   1,   0,   1,   0,   3,   4, -10,   1,   1,   0,   1,   3,   0, -11,  15,   0,   0,  -1,   0,  -1,   4,   5, -52,   0,   0,   1,  -1,   0,  -2, -12,  10,   0,   3,  -2,   3,  -5,  15,  -2, -68,   3,   2,   8,   2,  11,  -2,  52,   3,   3,   1,   0,   4,   1,  18,   8,  46,   5,   5,   1,   2,   4,  -2,  43,  25 },
      {   0,   0,   0,   1,   2,   1,  -3,   0,   0,  -1,  -2,   0,  -1,   0,  10, -21,  -1,  -1,  -2,  -1,  -3,  -4, -14, -10,  -2,  -2,  -3,  -3,  -4,  -9,  -5, -67,   0,  -2,   2,  -4,  -1,  -9, -26, -31,  -1,   1,  -1,   7,   0,  14,  -6, -73,   2,  -1,   3,  -2,  12,  -5,  46, -32,   3,  -4,   2,  -2,   6,  -1,   7,   0 },
      {   0,   0,   1,   1,   0,  -3,   1,   7,   0,   0,   1,   2,   1,   5,   6, -15,   0,   0,   2,   0,   3,  -3,  -3,  29,   0,  -1,  -1,   1,  -2,   6,   7, -30,   0,   1,   2,   0,   4,  -4,   3,  45,   1,   1,   1,   4,  -2,  14,  12,  -1,   4,   3,   2,   0,   8,  -5,  34,  76,  -4,   2,  -1,   1,  -8,  11,  -9,  66 },
      {   0,   0,   0,   1,   2,  -1,  -5,   1,   0,   0,  -1,   0,   0,  -1,  -2, -10,   0,   0,   0,   0,   0,   1,  -6,   0,   0,   1,   0,   0,   0,   0,   8, -10,  -1,   1,   1,   1,  -2,   4, -11,  27,   1,  -3,   0,  -1,   0,  -4,  28, -45,  -1,   2,  -1,   1,  -2,  -2, -30,  63,   0,  -5,  -2,  -2,  -3,  -5,   9, -86 },
    },
  },
  { // 11
    {
      { -97,  45,   2,   5,   1,   2,   0,   1,  62,  -8, -13,  -1,  -3,   0,  -1,   0,  -5, -20,   6,   3,   1,   1,   0,   1,   3,   9,   5,  -4,  -1,  -1,   0,  -1,  -1,  -3,  -7,   0,   2,   0,   0,   0,   1,   0,   3,   3,  -1,   0,  -1,   0,   0,  -2,   1,  -3,  -1,   0,   1,   0,   0,   0,   0,   0,   2,  -1,   0,   0 },
      { -63,   0,  11,   0,   4,   0,   2,   0, -68,  44,  10,  -1,   3,  -2,   1,  -1,  22,  48, -25,  -7,  -2,  -3,  -1,  -1,  19,  -8, -32,   7,   1,   1,   0,   0,   7, -14,  -4,  13,   0,   0,   0,   0,   4,  -6,   1,   7,  -1,  -1,   0,   0,   3,  -1,   2,   7,  -4,  -3,   1,  -1,   1,  -1,   1,   1,  -8,  -1,   2,   0 },
      {   0,  -8,   9,  -4,   2,  -2,   1,   0,  -8, -31,  18,  -6,   5,  -2,   2,   0, -47,  29,  34, -13,   7,  -5,   2,  -2,   0,  67, -25, -26,   4,  -6,   2,  -2,  -1,   4, -55,  12,  11,  -1,   2,   0,   2,  -7,  -8,  31,  -3,  -2,  -1,  -2,   2,  -2,   3,   9,  -9,  -1,   1,   0,   2,   0,   2,   3,  -5,  -3,   1,   0 },
      {  30,  33, -14,  -2,  -2,  -1,  -1,  -1,  30,  -1, -16,  -3,  -1,   0,   0,   0, -19,  27, -15,   0,   1,   0,   0,  -1,   7,  -5, -55,  17,  -1,   3,   1,   1,  -1, -44,   9,  49,  -7,   6,  -2,   0,   9,  -6,  45,  -4, -24,   3,  -3,   0,  -1,  -5,   6, -26,   3,   7,   0,   2,  -1,   0,   2,  -2,  10,  -3,  -1,   0 },
      {  28,  38, -15,  -6,   0,  -3,   0,  -1,  21,  46, -28,  -5,   0,  -4,   0,  -1,   4,  -7, -33,   0,   5,  -2,   1,  -1, -15,  -4,  -8,   6,   6,  -1,   2,   0, -13,   9, -27,   3,  12,  -6,   3,  -1, -12,  -2, -25,  44,  13,  -7,   3,  -4,   0,   6,   8,  44, -27, -12,   2,  -4,  -2,  -3,   1,  -1, -41,   4,   9,   2 },
      {  -1,  29,  -8,   0,  -1,  -1,  -1,   0,  -7, -33,  -2,   5,   4,   2,   2,   2, -37,  27,   8,   2,   1,  -1,   0,   0,  12,  -2, -12,  13,  -2,   2,  -2,   0,   8,  -3,  39, -11, -10,  -1,  -3,  -1,  -7,  10, -21, -42,  21,  -1,   4,   1,  -9,  13, -21,  43,  28, -12,   3,  -8,  -3,   9,  -6,  20, -49, -16,  16,  -2 },
      {  -3,  40, -26,  12,  -4,   1,  -1,  -1, -55,  -4, -26,  24,  -3,   2,   0,   1, -51, -14,  -1,  32,  -3,   4,  -1,   1,   2,  16,  31,  21, -14,   1,  -2,  -1,   8,  12,   9,  -6, -15,   0,  -1,  -1,   7,   4,   6,   7, -14,   1,   0,  -1,   6,  -3,  10, -19, -32,   5,  -2,   3,   2,  -8,   0, -26,   5,  28,  -1,   3 },
      {  18,  54,   2,  -9,  -3,  -3,  -2,  -1, -16,  43,  -8,  -9,  -2,  -1,  -1,   0,  -8,  -4,  -3,  -2,   3,   2,  -1,   0,   6,  21,  10, -17,  -2,  -1,  -1,  -2,   8,   4, -21, -23,   3,   2,   2,   2,   0, -18, -13, -20,  13,   7,   2,   4,  -4, -15, -12, -13,  46,  -3,   0,   6,   5,   7,   7,  29,  48, -36, -19,  -4 },
      {  16,  33, -10,  -5,  -9,   2,  -4,   0,   5, -23,  17,   6,  -4,   4,  -2,   1,  31,  22,  23,  -5, -11,   2,  -4,   0,  61, -18,   2, -22, -15,   3,  -4,   1,  -1, -40,  -8, -39,   9,   8,   2,   2,  -8,  -8,  -7,  12,  29,   2,   2,   4,   2,  -1,  11,   4, -17,  -6, -13,   2,   1,  -8,   3, -20,  10,  36,  -7,   1 },
      {   3,  14,  27, -11,  -4,  -2,  -1,  -1,  26,  16,  38, -16,  -1,  -1,   0,  -1, -28,  30, -20, -29,  -1,  -2,   2,   0, -20,  -6, -11,  14,   8,   6,   1,   2,   7,  34,  19, -10, -13,  -3,  -2,   2,   1,   3, -36, -17, -11,   1,  -5,   6,  -1, -14,  -8,  -1, -11,  21, -17,   0,   4,  -9,   6, -12,  24,  58,  -7,  -1 },
      { -12, -22, -33,  -4,   8,  -1,   2,   1,   7, -24,   2,   0,  14,  -2,   1,   1, -33,  20, -30,  -8,   9,  -3,   3,   0,   0, -25,  -1,  25,  11,   2,   0,  -1, -12,   2,  13, -14,  14,  -6,   1,  -5,  -3,  12,  -7,  28,  29, -12,  18,  -4,   0,  -4,  10,   1, -13, -40,   8,  16,  10,   4,  17,   4,  44, -29, -37,   1 },
      {  -5, -55, -38,  13,   2,   3,   2,   2,  36,  30, -29,  19,  -1,   1,  -3,   1,  -4,  28, -12,  17,  -7,  -1,  -2,   0,  17,  37,   5,  18, -19,   0,  -4,   2,  21,  -9, -11, -21, -31,   5,  -8,   2,  -7, -32, -10, -15,   9,  16,  -2,   4,   1,  -3,   9,   9,  14,   8,  -8,  -2,  -2,  -1,   5,   0,   0,  15,  -1,  -1 },
      {   2, -13,  52, -18,   4,  -3,   1,  -4,   1,  15,  21, -29,   4,  -4,   0,   0, -40, -47, -20,  -7,  11,   3,   3,   3,  30,   6,  11,  21,   6,   4,  -1,   3,  35, -36,   1,  -9, -10,   2,  -7,  -1,   0, -25,  22,   0,  21,   4,   5,  -2,   7,  12,  16,   4,  -5, -20,   2,   1,  -1,  -2,  -3, -13, -15,  -2,   1,   2 },
      {  -8, -11,  -8, -15,  12,   2,   4,   3, -24, -28, -50, -24,  12,   4,   5,   4,   3, -30, -11,  -4,  18,   5,   4,   3,   1,   0, -13,  -3,  16,   0,  -1,   2,   0, -11,  -4,  16,  16,  -1,   0,   6,  -5,  -6,  -2,  -2,   8,  -9, -11,  19,  -3,   0,  -5,  23,  21,   7, -39,  -7,   5,  11,  15,  31,  18,  56,  -9, -14 },
      {   8,  10,  15,   6,  -1,   0,  -1,   3,  -1, -50,  -1,  -2,   3,   0,   1,  -1,  29,  -8, -36,   6,   9,   0,   4,  -1,  50,   1,  -2,  31,   0,  -4,   1,  -1,  19,  31, -11,   9, -19,  -8,  -3,   3,   4, -19, -31,  15, -32,   8,  -8, -10,  -3, -24,  12,   9,   0,  28,  15,  -2,   3,  -1,  13,   6,   3, -27,   0,   2 },
      {  -7, -11, -24, -14,   2,   7,   1,   3,  -1,   4,   8, -30,  -3,   3,  -2,   1, -37,  12, -18,  -4,   4,  11,   0,   3,  15, -28,  23,  -8,  -8,   8,  -6,  10, -30, -10, -10, -20,  18,  10,  -8,  20,  -9,   5,   3,  18,  -8,   8, -36, -14,  -1,  16,   1, -11,   9,  52,  12, -27,  -5,   6, -18,  15,   5, -11,  36,  -3 },
      {   4,  -8,  15, -25,  10, -10,   5,  -2,  10,  17, -24, -26,  16,  -7,   4,   1,  11,  12,   9,  12,  18,  -7,   3,  -2,  41,  29,  13,  13,  -6, -16,  -1,  -4, -17,  -4,   8,  -5, -12,  -9,   4,  -6, -14,  45,   4,  -4, -15, -14,  16, -10, -16,   3, -51, -12, -33,  -6,  24,   4,  -5,  -3, -10,  22,   7,  12,  -2,   1 },
      {   7,  -9,  57,  69,  -4,   4,  -1,   2,   5,  -3, -19,  32, -30,   0,  -2,   0, -22,   0, -33,   6, -29,   8,  -1,   2,  11,  -3,  -9, -15, -13,  11,  -7,   4, -12,  -9,  -6,  -7,  18,  10,  -2,  -1,  -2,  19,   1,   9,   7,  -7,   2,  -2, -10,   0, -20,   4,  -3,   0,   4,  -2,   1,   7,   4,  23,  11,  10,  -1,  -4 },
      {  -4,   3,  24, -21, -14,   2,   6,   3, -15, -20, -34, -13, -10,  12,   5,   2,  16,   0,   1,  -9,  -9,   6,  -4,   1, -41,  16, -17,  12,  -3,  13,  -3,  14, -30, -24,  -3, -16, -30,  12, -22,   5, -16,   9,  -8,  -7,  22,  37,   0, -29,   6,   2,   5,  15,  -1,  14,  34,   0,   4, -11,   6, -29,  17,  -4, -10,  11 },
      {   3,  15,  35,  14,  30, -13,  -6,  -6,   2,  -7,  20,  10,  32, -17,  -6,  -7,   9,  22,  -2,  37,  23, -18,   2,  -2, -20,   4,  21,  21, -14, -23,   3,   1, -48, -13,  -2,  -8, -19,  -4,  10,   9, -22, -10,  13,   6,  -3,   3,  -7,  14,   6,  18,  26,   7,  13,   4, -29,  -6,   5,  14,   6,   6,   2,  -1,  -8,  -6 },
      {  -1,   0,  14, -10, -13,   3,   2,   1,   1, -26, -27, -46, -21,   5,   4,   2,   6,  38, -38,   7,  -5,  14,   2,   2,  -8,   4,  25, -11, -12,   1,  -7, -11, -16,  19, -11, -10,  11,   4,  15, -26,  16, -18,  11,  -9,   1,   1,  26,  21,   2,  -3,   6, -29,  13, -21, -19,  15, -10,  -6, -22, -18, -25,  -1,   8,   2 },
      {  -4, -18,  23, -33, -16,  -3,   3,  -1,   5,  18, -14,  -7, -10,   3,   5,   2, -29,   2,  38,  44,  -8,   7,   3,   0,   9, -45,  -6,  -8, -26,  -3,  -6, -10,  -6,  -4,  -3,   9,   1,  12,  14,  -9,  11,   7, -17,   7, -20,   8,  13,   8,   6, -29,   7,  22,   3,  18, -13,   8,  13,  -5,  35,   0, -11, -14, -24,   4 },
      {  -7,   0, -28,  -9, -45,   0,  -4,   0,  -7,  -7,  53,   2, -28,   8,  -2,   3,   1, -27, -17,  17,  -6,   9,  -3,   1,  -3,   1, -10,  18, -10,   2,   2,  -3, -25,  -3, -26,  10, -23,   1,   7, -18, -27,  -8,   7,   1,  -5,  24,  30,  -1, -16,  14, -19,  -1,   1,  -6,  -3,   8,  -1,  10,  -2,  38,  -5,  20, -10,  -6 },
      {   1,  -1,  13, -23, -49,  12,  -3,   3,   7,  10,   4,  13, -30,  23,  -5,   0,  12,  12,   1,  16,   9,  22,  -9,  -2,  17,  38,   5,  24,  17,   8,  -8,   0,  -5,  17,  23,  13,  22,  -5,  -7,  21,  -8,  28,  10,   9,  11, -16, -37,   7,  11,  18,  14,   5,   5,  -7, -34,  -9,   5,   6,   3,  -9,   7, -17,  -9,   1 },
      {  -6,  -9,  -2, -10, -22, -30,   6,  -9,  -8,  -2, -19,  14, -21, -27,  14,   1,  -9, -14,  -7, -38, -13, -18,  15,  -4,   5,   7,  -8, -12,   2,  -9,  17, -21, -15, -13,  20, -12,   1,  -3,  33,  24, -52, -19, -12, -11, -22, -16,   0,  28, -13,  -1,  11, -10, -16,  12,  -7,   1,  -5,   8,  -1, -16,  -4, -21,  -3,   2 },
      {   2,   2, -12,  41,  16,  13,  -6,   6,  -2,  20,  -2, -28,   3,  11,  -6,   5,  -9,  -3,  16, -34,  -8,  14,  -4,   3,  38,   8,  19,  -9,  27,  11,  -5,   4, -40,  14,  13,  23,  -3, -12, -12, -21, -24,   3,   7, -10, -20,  19,  22,  -7,   9,  12,  19,  12,  16,   8,  -3,   8,  15,   2,  21, -18, -14,   1, -26,   4 },
      {   2,   8,   6,  13,  -2,  14,   4,  12,  -2,  -2,  -3, -26, -32,   2,  -7,  10,  10,  27,  36,  -5, -12,   5,   1,   6, -29,  -2,  39,  34,  14,   6,  -8,  -4,  19, -28,   8,   3,   6,  -8, -10,  -3, -10, -38,  -2,   9, -13, -30,   1,   2, -19, -10,   8,   7, -24,  -1,  21,  -3,   3,  22,  17,  32,   5,   2,   5, -11 },
      {   0,  -3, -11,  15, -29, -43,   2,  -6,   0,  -2,   0, -20, -22, -52,  17,  -3,  -1,   5,   3,   8, -13, -25,   8,  -2,  11,  -9,  21,  -1,  22,   2,  12,  21, -20,  -2, -16,  14, -16, -12, -17,  25,  44,  -3,   2, -11,   3,   2, -27, -10,  15,  -2, -12,  11,   5, -25,   2,  -5,  10, -12,   7,   0,  -2,   8,  -9,  -1 },
      {  -2,   1,  13, -46,  13, -41,  -8, -15,   4,  -2,  -6,  33,   8, -14,   4,   3,  -1,  12,  -4,  16, -23,   1,  -4,   2,   9,  -1,   6, -22,   9,  35,  -7,  16,   2,  29,   9,  10,  18,   1, -37, -22, -24, -22,  14,   2,  -7,   1,   5, -24, -20,   7,  20,  -5,   6, -10,  15,   3,  -3,  26,   5,   6,   0,  16,   6, -11 },
      {   3,   7,  -8, -13,  49,  11,  15,  14,  -1,  -3,   9, -22, -16, -20,   7,   5,   9,  -9,   4, -12, -56, -17,   8,  -3,   3,  14,  -2,  22, -39,  29,   8,  -4,   8,  16,   1,   8,  -6,  40,   3,   5,  -3,  21,   3,  11,   6,  12, -20,   6,   6,  10,  -4,  -4,  -8,  -9, -28,   2,   3,   8,   1,  12,  -6, -13, -14,  -4 },
      {  -1,  -1,   7, -17,   0,  49,  12,  19,   1,  -2,   4,   4,  25,  34,  -1,   6, -11,  -1, -17,  13, -30, -26, -25,  -6,   5,  -6,  11, -23,   1,  -7,  12,  18, -21,   6,  -6,  17,  -6,  -9,   2,  39, -12, -30,   4, -19,  -7, -11, -22,   3, -22, -11, -16,  -1, -10, -31,  11,   2,  -1, -13,  13,  -7,  -7,   2,  -7,   7 },
      {  -2,   0,  -9,  17, -26, -30, -30,   1,  -3,  -4,   2, -25,  25,   0, -38,   9,   7,   3, -14, -11,  21,   5, -20,  14, -15,   8,   1, -27, -21,   6,  -9,   9,  20,   9,  10, -19,  -7,  38,  -3,  11,  -6,   6,  30,  -1,  -8,   0, -20,  -5, -19, -14,  -5,   4, -24,   0,  -6,   2,   8,  -1,  34,  10, -24,  -9, -24,   2 },
      {   1,   2,  -3,  23,  -6,   8,   7,  22,   0,  -1,  -2, -38, -14, -20,   7,   5,  -2,   5,   8,  42,   8, -19,   3,  -1,  -1, -19, -27,  -8,  -2, -14,   2,   7,  41,  25, -19,  -2,   7,  -9,   2,   3, -43,  12,   7, -10,   6, -12,  -1, -12, -23,  37,   9,  -1,  12,   3,  11,  -8,   0,  10,  -5, -34,  15,   5, -14,   4 },
      {   2,   6,   5,   8, -16, -14,  23, -13,   6,  -2,  -5,   4,  32,  29,  46,  -9,   2,  21,  -2,  -6, -13,  22,  19, -22,  -5, -12,  27, -13,  14,   4,   8, -41,  29,   7,  -7,  11, -12,   8,  17,  13,   3,  -2,  22,   6,   4,  29,  -6,   3,  -4,  39,  -6,  14, -14,  10,   7,  -5,   9,  -2,  -1,  14,   5,   1, -20,   1 },
      {   1,   2, -17,   3,  10,  17,  20, -13,   1,  -2,  32,  -7,  -1, -20,  25, -18,  10,  -6, -29,  33,   8,  11,  30,  -5,  -7,  34,   6, -18,  -1,  15,   1, -15,  -9, -17,  22,   5,  30,  15,   4, -12,  -2,  -4,   4, -15,  13, -25,  -4,  -5,   0, -21,  -9,  10,   5,  24,  26,   6,  15, -11,  38, -16,  -8,   7,  -1,   9 },
      {   0,   3,  -4, -18,   4,   3,  -6, -10,   1, -10,   2,   3,  -9,   2, -10, -14,   2,   6, -24,  -5, -18,  -5, -10,   1, -14,   8,  18, -22, -22, -16,   2,  36,  29, -26,   7,   5,   7, -41, -12,   1,   0,   4,  -8,   7, -18, -17,  19,  -5,  36,  30, -14,   5,   1,  31,  -6,   5,  20,  -5, -15,   6, -12,  -7, -47,  -8 },
      {   1,   0,   7, -13, -31,  23,   3,  29,   1,   2,  -8,   9,  -2, -19, -14,   9,  -2,   6,  -7, -18,   2, -53,  -4,  -3,   5,  11,  19, -13, -12, -25,  24,  -9,   5,   9,  32,  12,   6,   7,  28, -23,  15,   3,  20,  18,  16,  21,  15, -25,   9,   4,  20,  13,  13,  11,  14, -15,   4,   6,  15,  15,  11,  16,  17,  -5 },
      {   3,   5,   2,   1,  -4,   9,  58,  44,   2,   3,   4,  19,  -5,  -5,  46,  31,   2,   0,  -3,  -8,  39,  -9,   1,   4,  -4,  -2,  -9, -12,   7,   8, -18,  19,  -1,  -1,   3, -36,  -6,   3, -24,  -1,   3,   4,  11,   2, -26,   2,   4,  -3,  15,  -8,  10, -12,  -4,  -4, -14,   7,   0,   4,   1,  18, -22,  -2, -14,  -7 },
      {   1,   1,   1, -22,  16,   5, -18,  41,   1,  -6,   1,  20,  -1, -30, -30,  23,   0,   3, -14,  13, -14,   5, -16,  20,  -1,  -2,  -2,  -4,  50,  13, -25, -43,  -2,  -6, -25, -12,  -7,   1,  14,  -7,  11,   3,   7, -25,  -4,  -3,   1,  17,  16,  24, -12,   7, -16,  16,   8, -12,  13, -12,   7,  -4,   1,  -8,  -7,  -4 },
      {   1,   0,  -4, -14,  25,  22, -16, -24,   0,   3,  -1,  -3, -38,   7,  11, -15,   0,  -4,  -5,  -5,   2,   8,  11,   0,   6,   2, -21, -20,   1, -31, -11,  -1, -11,  19,   4, -25, -29, -20,   4,  17,  40,   6,  21,  -8,   9,   3,  -3,   7, -22,  10,  23,   1, -20,   3,  16,  -2,   5,  48,  22,   4,  -8,   3,  -8, -17 },
      {  -2,   1,   0,  -8,  22,   1, -12,  22,  -3,  -2,   6,  14, -21, -20,   1,   4,  -4,   9,  -6,  -4,   8,  15,   9,  -9,  -7,  -1,  25, -10,  16, -10,   1,   0,  13, -18,   5,  14,  -9, -27,  -2,   0,  -5,  15,   8,  25,  16,  31, -11,  -8, -57, -25, -14,   3,  15,  14, -25,  12, -46, -20,   2, -10, -11,  -4, -20,  11 },
      {   0,  -1,   2,  -2,   4,  29,  -2, -45,  -1,  -3, -11,   7, -17,   3, -19, -26,  -1,   9,   3,   5,  19, -42, -10, -24,   0,  -8,   6,  13,  45,  32,  33,   1,  11,  -6, -21, -28,  12,  25,   5,   4,  -6,  17,   6,  -5, -20,  13,   3,  -5,   0,  -1,   3, -10,   8,   6,  -1,  18,  -6,  -2,   3,  10, -21,   6, -16,   4 },
      {   0,   2,  -4,   0,  21, -13,  24, -20,  -1,  -6,   4, -13, -27,   1,  -4, -19,  -1,   2,   0,   9, -16,   8, -31, -11,  -3,  -1, -10, -13,  12, -16, -21,  -6,  12,  15,  14, -17, -15, -20,   8,   7, -33,   3,  36,  12,  -7,   7,   7,  -6,  28, -25,   5,  17,   3, -16,  -9, -16,   5, -46,  12,  23,  14,  -1,  31,  11 },
      {   0,   1,  -3,   4,  -9, -11,  52, -16,  -1,  -1,   3,  -3,  11, -15,  -8, -24,  -1,  -6,  -3,  -8,  -9, -16, -63, -14,   1,   2,   6,  13,   3, -25, -49, -13,  -1,  -7,  -5,  16,  19,   7, -14,  -7,  11,   4, -16,  -5,  17,  14,   1,  10, -20,   1,  -5, -12,  -5,  17,  -4,   3,  -6,  30,   8, -15,  -7,  -2,  -4,   1 },
      {   0,  -1,  -6,   4,  20, -49, -16,  23,   2,   2,  13,  -9,  -7,  41,  16,  33,   3,  -5,  -7,  13, -15,  -9, -24, -20,  -2,  16,  -2,  19,   6,   0,   9,   1,   1, -10,   2, -10,  34,  -5,  29,  39,  16,  16,  -2,  -1,  -5,   5,  26,  -6,   2, -13,  16,  -5,  20,   7,  21,  -5,  -7,  10,   7,  -4, -13,  16,  -6,   0 },
      {  -1,   1,  -3,  -4, -17,  -3,   8,  -4,  -2,  -1,   3,  -1,  24,  14,   1,  -2,   0,  -4, -11, -16, -34,  -8,  15,   3,  -6,   2,   5,  10, -12,  -5,   7, -16,   3, -18, -35, -19,  -4, -28, -19, -30,   7,  46,   7, -27, -20, -33, -10, -13, -18,   0,  39,   6,  10,  -4, -15,  -8, -22, -11,  30,   5,   1,  -3,  13,   3 },
      {   0,   0,   0,   5, -16,   5,  31,  12,  -1,   4,   2, -15,  27, -21, -29, -19,   2,  -6,   9,  16, -18,  20,  -4,  -5,   4,  12, -17,  -3,   3,  37,  20,  14, -11,   9,  19, -18,  -1, -18,   8,  15,  26,  -5,  17,   6,  -4,  -4,  22,  22, -17,  -8,  22,  32,  -8,  19,   7,   4, -30,   1, -37,   8,  13,  -3, -15,  -6 },
      {   2,   1,   1,  -5,   5,  -4,  36,   4,   3,  -2,  -3,  10,   4, -10, -15,  -9,   2,  10,  -2,  -2,   0,  44,  -1,  30,   0,  -2,  10,   3,  14,   1,  38,  49,  13,  -5, -17,   9,  -2,  -2,  40,   9,  -5,  16, -11, -15,   9,  15,  18, -11,  -3,   2,  -2, -22, -11,  -5, -15, -15,  19,  20,  35,  -6,  -5,  -6,  26,  -1 },
      {   0,  -2,   2,  -2,   2,  -2,  -1,  36,  -1,  -1,  -4,  -1,  -5, -15, -25, -26,  -2,  -1,  -3,  -6,  24,  27, -11, -69,  -3,  -5,   0, -15, -29,  26,  14, -34,   0,  -9,  -2,   6, -12, -25, -10,  12,  -4,   6, -20,  -3,  -8,  -1, -10, -10,  11,   5,   8, -18,   6, -25,  10,   5,  18,  14,  11,   3,  -7,  10,   8,  -4 },
      {   0,  -2,  -2,  11, -16,  16, -22,  -5,   1,   0,   2, -22,  25,  -6,   7,  10,   2,  -1,  -6,  29, -21, -17,  18,   4,  -7,  10, -10, -15,  17,   3,   1, -13,  11, -16,  18,  -9,  19, -33, -27,   7, -11,  11, -16,   3,  -4,  56,   5,   8,  18, -22,  -7, -15, -27,  -7, -23,  -2,   4,  31,   4,   2,   0, -13,  24, -14 },
      {   0,   1,  -4,  -4,  -6,  24, -26, -11,   1,  -1,   6,   0,   3, -22,  27,   5,   3,   3, -11,  -4,  10,   3, -15, -12,  -3,   5,  21, -17,  -3,  18, -32,  20,  12, -16,  -2,  17, -19,  32,   5,  26,   4,  24, -14,   5, -35,  -6,  40,  16,  -6,  13,  10,  28,   9, -20,  -2,  -9, -21,   7,   8, -14,  28,  -4,  25,  -3 },
      {   0,  -1,  -5,   6,  18, -20,  23,  -6,   0,   0,   7, -10, -24,  32, -14,   3,  -1,  -3, -10,  19,  15, -10, -13,  15,  -1,   8,  -2, -30,   6,   9,  20, -16,  -4, -21,  21,   2, -11,   8,  13, -34,  17, -12, -29,  -1, -30,   5, -22, -27,   0,  18,  -7,  19, -15, -22, -17, -23, -30,  17,  -7, -15,  22,  -9,  -9,  -3 },
      {  -1,  -2,   1,   2,  -2,   5, -31,  21,  -1,   1,  -4,  -3,   7,  -7,  30, -27,  -1,  -4,   4,   3, -11,  13, -11, -48,   1,   1, -10,   8,  18, -20,   6,  21,  -9,   4,   1,  -7,  10,  26,   1, -20,   3, -21,  -1, -23,  10, -13,  14, -37,  -1,  -5,   1,  -7, -33,  13, -33, -35, -12, -14,  -4,   8,  -1, -17, -10, -23 },
      {   0,   1,   2,   2,   5,  -4,  -3,   7,   0,   0,  -1,  -2,  -9,  -3,  -4, -11,   0,   2,   2,   4,  12,   5,   7,   9,   2,   0,  -2,  -4, -11, -14,  -3,  12,   0,   7,   4,  17,  27,   3,  -5,   7,  -5,  -3, -20, -29,   6,  26,  -9,  22,  26,  33,  28, -11, -25,  -3,  24,  26, -63, -32,  29,  33,   3,   3,  -3,   1 },
      {   0,   0,  -3,  -2,   2, -25,  -6,  42,   0,   0,   3,  -3,  -7,  27,   2, -31,  -1,  -1,  -5,   8, -10, -13,   3, -19,   2,   1,   3, -10,  25, -17,  18,  28,   4,  -6,   3,  -6, -17,  21, -18, -28,   1,   8,   5,  10,  -1,  -6,  -7,  45,  -2,   7,  -8,  16,  -4,  14,  -1,  46,  16,  14,   5,  -5,  13,  -9,  34,  19 },
      {   0,  -1,  -2,  -4,  10,  -3,   3,  10,   0,   1,   3,   6, -12,   0,  -9, -19,   1,  -1,  -6,  -1,  10,  -1,  16,  11,  -1,   6,   7,   1,  -7, -20, -10,  -1,   1, -10,  -3,   6,  36,  13,  -7,  10,   6,   9,  -7, -31, -24,  32,   2,  -6, -41,  -7,  35,  21, -15, -20,  -2,  -2,  56, -23, -42,  11,   8,   2,  -2,   9 },
      {   0,  -1,   1,   3,   3,  14,  -4, -27,   1,   0,   0,  -5,   3, -27,  13,  61,  -1,   0,  -3,   7,  -5,  30, -42, -13,   0,   2,  -4,  -6,   2, -21,  36,  -9,   0,  -4,   4,   0,  -2,   4,  -2,  -5,   4,   1,  -7, -11,   3,  -9, -14, -20,  -2,   8,  18,  10, -11,  11, -10,  55,  12,   1, -15,  15,   8,  -2,  13,  19 },
      {   0,   0,   0,  -2,  -7,   7, -10,  13,  -1,   0,  -4,  -1,  16, -10,   1, -26,   1,  -3,   5,   0, -20,   6,  17,   7,  -3,   3, -11,   7,  10,  -4, -38,   6,   0,  -3,   9, -20,   7,  -2,  34,   8,   9, -16,  -6,   9, -23,   0, -11, -43,  10,  14,  -5,  -5,  12, -12, -11,  24, -10,  26,   1,  13,  -5,  13,  -3,  71 },
      {   0,  -2,   0,   3,  -4,   2, -18,  10,   1,   1,   1,  -4,   7,  -1,  33, -29,  -2,   2,  -2,   2,   0,  -2, -29,  25,   0,   0,   7,  -6,   4,  19, -15,   8,   7,  -9,  -5,  11, -16, -11,  47, -24, -20,  22,  -5,  -8,  10,  -1, -41,  -2,  14, -34,  25,   1,   0,  -4,  21,  22,   2,  12, -29,   8,  -6,   6, -11, -36 },
      {   0,   2,  -2,  -5,  -3,   3,  17,   0,   0,  -1,   4,   6,   5,  -6, -31,   8,  -2,   3,  -6,  -3, -18,  28,  31, -18,   0,  -3,   7,   5,  12, -51,  -3,   0,   5,   0,  -9,  -5,   9,  33,  -4,   8, -15,   7,  12,  -3, -24,   4, -12, -11,  31, -23, -11,  22,  15, -18,  17,  16, -22,  24, -13, -21,   0,  11, -11, -43 },
      {   0,   0,  -1,   0,   3,  -4,   6, -14,   0,  -2,   1,   6,  -6,   8, -15,  37,   0,   1,  -3,  -8,  15, -15,  23, -49,   1,  -5,   9,   2, -14,  23, -29,  33,   1,   5, -20,  10,  10, -24,  29, -23,  -4,  -1,  22, -33,  12,  11, -13,  10,  13, -16,  -7,  25, -29,   8,  -3,  -7,  -2,  21,  -6,  -4,  13,  -9,   2,  26 },
      {   0,   0,  -1,  -3,   6,   0,  -9,  14,   1,  -2,   1,   7, -10,  -1,  16, -24,  -2,   7,  -4,  -9,  15,   2, -18,  32,   2,  -9,  14,   1, -16,  -2,  32, -28,   1,   9, -21,  13,   7,   4, -32,  19, -11,   3,  24, -25,   9,  -7,  21,  -5,  29, -32,  -6,  23, -14,   7,  -7, -11, -22,  39, -13,  -2,   7,   2,  -7,  43 },
      {   0,   0,   1,  -3,   1,   9, -15,   2,   0,   0,  -3,   7,   2, -17,  27, -10,   1,  -1,   7, -15,   5,  16, -31,   5,  -3,   1,  -8,  28, -25,   0,  21,   7,  -1,   4,  -3, -31,  40, -22,   7, -24,   8, -15,  23,   5, -37,  29, -22,  22,  -7,  15, -34,  23,   9, -15,  12,  -5,   2,  -3,  22, -24,   8,   1,   3,  -9 },
      {   0,   0,   0,   3,   2,  -7,   4,   8,   0,   0,  -1,   1,  -6,  13,  -3, -13,   0,   0,   3,  -6,  11, -19,   7,  13,   1,  -3,   2,   7, -14,  20, -15, -14,  -4,   6,  -8,   2,   9, -14,   9,  29,   6, -13,  15, -13,   4,  -5,  17, -45,  -4,  13, -18,  20, -10,  16, -31,  60,  -1,  -9,  11, -17,  12, -13,  29, -48 },
    },
    {
      { -75,  53,  -6,   3,  -1,   1,  -1,   0,  70, -42,  -4,   1,  -1,   0,   0,   0, -23,   0,  16,  -4,   2,  -1,   1,   0,   5,  10,  -9,  -1,   0,   0,   0,   0,  -2,  -5,   1,   3,   0,   0,   0,   0,   2,   2,   1,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
      {  46,   6, -23,  -1,  -3,   0,  -1,   0,  -2, -56,  38,   0,   5,  -1,   2,   0, -43,  66, -12, -14,   3,  -3,   1,  -2,  21, -19, -23,  20,  -3,   3,  -1,   1,  -6,  -2,  22,  -7,  -3,   0,   0,   0,   2,   3, -10,  -3,   4,  -1,   1,   0,  -2,   0,   3,   3,  -1,   0,   0,   0,   1,   0,  -2,  -1,   0,   1,   0,   0 },
      { -60,  24, -10,   6,   1,   2,   0,   0, -33,  25,  21,  -6,   0,  -2,   0,  -1,  31,  24, -52,   6,  -2,   1,  -1,   1,  20, -54,  21,  20,  -8,   4,  -2,   2,  -2,  12,  18, -23,   2,  -2,   1,  -1,   1,  -1, -14,   6,   4,  -1,   1,   0,   2,   0,   4,   0,  -4,   1,  -1,   0,   2,  -2,  -2,  -1,   2,   1,   0,   0 },
      {  34,  -7, -19,  10,  -2,   2,  -1,   1,  49, -15,   8, -14,  -2,  -2,   0,  -1, -12, -27, -20,  24,   2,   4,   0,   2,  -2, -21,  66, -14,  -8,   1,  -2,   0, -21,  38, -22, -31,  20,  -5,   4,  -2,  -2,  -3, -13,  28,  -3,  -2,   1,   0,  -2,   2,   7,  -7,  -6,   3,  -1,   0,  -1,   0,  -3,   2,   4,  -1,   0,   0 },
      {  33,  52, -33, -10,  -5,  -1,  -1,  -1,  -5, -22, -14,  29,  -4,   4,  -2,   2,  48, -39,  21, -19,  -2,   1,   0,   0, -27, -15,   4,  17,   7,  -1,   2,   0,   4,   4,  36, -20, -12,   2,  -2,   0, -11,  10, -19, -16,  24,  -4,   2,  -1,  -1,   4,  -1,  19,  -6,  -7,   2,  -2,  -2,   1,  -1,  -5,  -4,   3,   1,   0 },
      {  -6,  43, -32,   9,  -4,   1,  -2,   1, -59, -31,  32,   1,   3,   2,   1,   1,  17,   2,  21,  -2,  -1,  -1,  -1,  -1,  -5,  22,   6, -35,   6,  -3,   2,  -2,  -3,  20, -50,  19,  14,  -5,   3,  -1,  11, -16,   9,  23, -22,   3,  -3,   1,   3,  -1,   6, -15,   3,   5,  -1,   1,   2,  -1,  -1,   4,   1,  -2,   0,   0 },
      { -27, -31,  19, -11,   9,   1,   3,   0, -18, -10,  20,  14,  -5,  -2,  -1,   0,  -9,   3,  22, -28,  11,   0,   2,  -1, -11,  35,   0,   8,  -5,  -6,  -1,  -1,   9,   9, -10, -45,  11,   4,   2,   0, -12,  17, -43,  40,  26, -18,   6,  -3,   8,  -3,   6,  18, -44,   7,  -2,   0,   1,  -4,   0, -13,  10,  12,  -3,   1 },
      { -22, -47,  12,  10,  -3,   3,   0,   1,  12, -33,  47, -18,  11,  -2,   2,  -1,  34,   4,  -2, -13,  -4,  -4,   0,  -1, -37,   2,   9,  -7,  18,  -1,   3,  -1,   4,  31,  -2,   7, -19,  -5,  -2,  -1,   3, -14,   0, -37,  17,  13,  -1,   2,  -4,  16, -14,  20,  25, -25,   6,  -3,  -1,   1,  -4,   8, -26,   0,   5,   0 },
      {   9,  -7, -16,   1,   7,   0,   1,   0,  39,  10,  24,  -6, -13,  -3,  -2,  -1,  42,   5, -33, -29,  19,  -4,   3,  -2, -46,   6, -33,  34,  -3,  -3,   3,  -1,   2,  -6,  -7,  20,   0,   4,   0,   1,  -4,  -4,  12,  29, -29,  -2,  -2,   0,   6,  -9,  16, -29, -13,  24,  -7,   4,  -2,  -1,   1,  -3,  22,  -4,  -4,   0 },
      {   3,  22, -38,  20,  -6,   5,  -1,   2,  -9,  27,  24, -23,   0,  -4,   0,  -2, -42, -21, -29,  13,  13,   0,   5,   0, -19,  45,  13,  30, -13,  -4,  -2,  -1,  39, -19, -10, -16, -24,  10,  -4,   3,  -2, -19,  15, -13,  25,   4,  -1,   1,   1,  11,  -8,  10,  -2, -13,   3,  -2,   1,  -6,   2,   0,  -6,   8,  -2,   2 },
      { -18, -23,  -1,   8,  10,   1,   0,   1, -26, -29,   5,  13, -14,   7,  -1,   3, -18, -22,  16,   9,  15,   0,   0,   0, -12,   8,  21,  20, -25,   3,  -3,   1,  19,   5,  36, -19,   1,   2,  -5,   0,   0,  13,   0, -22, -29,   2,   1,   1,   9, -15,   3, -40,  39,  26,  -7,   4,  -2,  11,  -6,  13,  23, -41,   4,  -2 },
      {   1,  26,  37, -38,   5,  -6,   1,  -3, -14,   0, -30,  14,   9,  -3,   4,  -1, -39,  10, -22,   6,  -4,   3,   1,   1, -48,  -1,   4,  25,  14,  -3,   2,  -1,   1,  57,  -6,  15, -26,  -6,   1,  -2,  16, -18, -19,   1, -14,  17,  -5,   3,   6,  -7,  15,   2,   1,   2,  -5,   1,   3,  -3,  -2,  -7,  -3,   6,  -1,   1 },
      {  19,  36,  36, -17,   8, -12,  -1,  -4,  11,  18,   6, -22,  -9,   5,   1,   2,  14,  18,  -6, -10,  17, -10,   1,  -3,  -3,  15, -10, -11, -21,   9,  -1,   2,  16, -12, -14, -28,  30,  -6,  -3,   0, -16,  -9, -27,   8,   1,   4,   8,  -1,   6,  -7,   9,  10,  44, -10,  -7,  -1, -10,  11, -13,  25, -35, -32,  23,  -5 },
      {   4,   6,  38,  -9,  -4,  -3,   0,  -1, -13, -56, -20, -19,  20,   0,   5,   0,  26,  12, -27,  50,  -7,  -3,   0,  -1, -29,   5,  10,  -4, -23,  12,  -5,   3,  28, -33,   6,  -1,  17,   5,  -3,   3, -11,   6,  26,   8,   3, -15,   3,  -2,  -1,  13,  -8,  -3, -23,   2,   1,  -1,   1,  -3,   4,  -7,   5,  17,  -6,   0 },
      { -17, -29, -57,  20,   8,  -4,   2,   1,  -2,   5, -34,  23,  -4,  15,   0,   2, -13,  20,   4,  16,  -6, -14,  -2,  -2, -35,  -8,  -7,  -5,  -9,  25,   1,   3,  17,  -1,   9,   9,  27, -19,  -5,   1,  -9,   6, -10,  18, -18,   6,   8,  -3,   9,  -1,  14,  13,   5, -17, -10,   0,  -8,   5,  -6,   5, -46,  26,  11,  -4 },
      {   5,  21,   5,  35, -24,   5,  -3,   0,  -1,  27,  14, -46,  15,  -5,   2,  -2, -13,   6,  13,   9,  -7,  -5,   2,  -2, -36,  -1, -12, -22,   5,   9,   0,   1,  -4,  18,  24,  11,  18,  -9,  -1,  -2,  -1,  42, -32, -18, -18, -15,   6,  -1,   2, -16, -30,  17,  -9,  28,  -1,  -1,   7,  -7,  14,  -6,  14,  -3,  -8,   2 },
      {  -3,   4,  18,   0,  11,  -9,  -1,  -2, -22, -18,  14, -30,   0,   6,   2,   2, -15, -54, -12,  -8,  39, -11,   7,   0,  14,  -4, -11,  28,   2,  -1,   6,  -3, -44,  -4,  15,  21,   8,  -9,  -2,  -1,   4,  34,   8,  16, -14,   2,   4,  -3,   3,   6,  -1,   0,  -5, -20,  -6,   1,  -5,   9, -11,   9, -36,  31,  10,  -6 },
      {  -4, -15, -24,  19,  -8,   8,   2,   1,  -6, -20, -45, -33,  22,  -3,   5,   0,  27,   6,  -3,  27,  19,  -5,   3,  -3,  18,  28, -22,  22, -22, -13,   1,  -3, -36,  14,  -7,  -3, -19,  17,   2,   1,  13, -15, -26,   2,   8,   5,  -5,   3,  -1, -16,  22,  27,   4,  14,   0,  -6,   4,   2,  10, -11,  -4, -30,   8,   5 },
      {  11,   9,  17,  23, -10, -12,   2,  -1,  12,   8,   6,   6, -14,  14,  -3,   0,  28,  21,  16,  14,  -3, -16,   0,  -2,  24,  32,   8,  10, -28,  11,  -1,  -4,  -2,  16,   3, -19, -24, -10,  -4,   2,   8,   3, -14, -17, -19,  30,   0,   0,   4, -23,  -3, -26,   3, -17,  -5,  16,  -3,   0,  -6,  -4,   7,  65, -22,  -3 },
      {   2,  -5,  14,  23, -32,   2,  -2,   4,  -3, -16,  -2, -20,  21,   8,  -6,   1,  -5, -30,  -4, -18,  -8,  -5,   7,   2,  24, -15,  -4,   1,  32,  15,  -5,   0,  43, -13,   3,   4,  -5, -33,   8,   0, -31, -17, -17,  16, -10,  31,   1,  -6,  18, -10,  42,   3,   6,  -9,  -2,   4,   0,  -1,   4, -31,  12,  -8, -13,   9 },
      {   4,   8,  34,  69, -26,   8, -10,   4,  -7,   2, -17,   1, -16,   6,  -3,   3, -12,   4,  21, -23,   8,  -1,  -2,  -2, -25, -31, -10,   6, -17,   5,   5,   3,  -8,   8,   0, -12,  11,  24,  -9,   4,  28, -30,  22,  10,  23, -18,  -3,   2,  -6,  12,  20, -15,  -6, -13,   5,  -1,  -1,   6, -15,  -4,  -1,   0,   6,  -3 },
      {  -8,  -3, -16, -33,   0,  -3,   7,   0,  -5,   9,   9, -14,  25,   6,  -6,  -3,  -9,  -6,   3,   4, -15, -13,   7,   2, -32,  -1, -17, -13,  -7,  26,  -7,   0, -40, -18,   8, -19,  18,   0,   5,   8,  23,   4,   5,   7,  30,  18, -12, -10,   7, -18,  17, -18,  19, -44,  25,  12,   1,   7,  -1,  -1,  33,  -7, -30,   6 },
      {  -3,  -1,  -5,  14,   0,  -3,   4,   4, -12, -18, -48, -16, -34,  20,  -3,   5,  -3,  22, -21, -37,  35,   9,  -1,   3, -13,  -6,  19,  -5,  20, -15,  -1,  -5,  -5, -11, -25,  -5,  -3,  -4,  12,   3, -22,  20,  -1,  -1,   9,  18,  -6,  -4,   1,  -8, -35,   2,  16, -10,  15,   5,   5,  -9,  18,  25,  18,   4, -30,   4 },
      {  -2, -11, -18, -37,   8,  -8,   3,  -3,   1,  -5,  -9, -42,  -3,  17,   5,   4,  -1, -18,  16, -11,  20,  -8,  -5,  -1,  17, -26, -18,  -2, -10,   9,  11,   1,  45,  22,  -1,   3,  20,  11, -17,   1,  28, -39, -10,  -9,   0, -23,  13,   4, -25, -10,  -7,   2,   1,  12,   0,   4,   7,   4,  10,  14,  12,  31, -22,  -5 },
      {  -5,  -3, -16, -27, -38,  22,   1,   0,   1,  14, -19, -31,  47, -11,   6,   1,   2,  24,  35, -13,  12,  -1,   3,  -4,  -4,  -6,  15,  14,   5, -19,   3,  -2,   4,   6,  -1, -10, -12,   1,   4,  -4, -26,  12,  12,   6,   0,   3,   3,   4,   6,  25,  -4, -37,  -4,  -5, -18,   7, -14,   6, -36,  19,   7,  11,  22, -13 },
      {   3,   7,   0,  -2,  31, -25,  16,  -7,   6,  14,   5, -12,  -2,  21, -14,   1,   3,  18,  38,  14,  30, -32,   8,   1,   3, -14,  12,  12, -13,  -6,   0,  -8,  16,  20,   4,  12,  -5,  -6,   9,   5, -15,   7,  14,   3, -11,  -2, -26, -10,   6,  29,   1,   1, -15, -15,  45, -14,   6, -15,  -4, -33,  -6, -28,  -9,  21 },
      {   5,   0,  -1,  10, -16,   7,   5,   2,   6,  -1, -18,  18,  27,  -2, -12,   0,  13,   4, -26, -19,   0,   1,  11,   2,  23,  40,   7,  17,  23,  17, -14,   0,  20,  22,  15,   2,  24, -32,  -5,   0,  46,   9,   0,   7,  -6, -28,   1, -12, -14,  -1, -20, -14, -12, -19,  35,  -6,  -1,  -1, -14,  23,  -4, -21,   2,   2 },
      {   8,   7,   1,  26,  73,   6,  -8,  -1,   4,  -1,  -5, -24,  14, -31,  11,  -2,   8,  10,  18,   6,  -4,  23,  -9,   0,  -2,  -9,  11,   0,  23, -25,  -3,   3,  31,  -8,   6,   4, -10,   1,  13,  -9,  30,  18,  -4,  11,  11,   1,  -7,   1,   9, -32,  15,  -8,  -5, -26,  -9,  11,  -6,  16,   2,   6,   2,   2,   4, -12 },
      {   1,   3,   2,  31,  46, -15,   9,  -4,  -4,  -5, -12,   4,  30, -28,   2,  -3,  -6, -18,  -3, -10, -40, -12,  16,  -3,   7,  -7, -33,  17, -14,  22,  -4,  -2,  -7,  22, -18, -13,   0,   1,   1,  -2, -43, -13,  -3,   4,  -2,   6,   1,   1,  -7,  27, -24,  -5,   6,  10,   6,  -1,  -3, -10,  -6,  36,  15,   4, -13,  -2 },
      {   3,  -1,   9,  -2, -23,  26, -10,   6,   7,   1,  11,  39,  37, -20,   2,  -1,   0,  -9,   3,  -1,  21, -22,   6,  -5,   6, -28,   0,  10, -34,  -3,  11,  -7,  25,   7, -26,  14,   2,  28,   0,   5,   2,  23,   6,   9,   1,   5, -20,  -2,  24, -36, -17,  13,   8, -13,   3,   7,   6, -10,  35,  18, -15,  -3, -10,  -2 },
      {  -1,  -5,  -5, -11, -15,  19,   9,   2,  -3, -10,  -3, -18, -39,  -2, -10,   1,  -2, -15, -13,   8, -26,  18,   5,   3,   5,  -5, -32, -14,  -9,  -1, -12,   4,  21,  18,   0,  -3,  -2,  22,  28,   2,  -7,  15, -11,   2,  -2,   9, -59,  12,   3, -15,  -6, -13,  -9,  -4,  34, -21, -20,   6, -22,   4,  -6,  13,  33,  -2 },
      {   2,   3,  -1,   1,   5,  25, -23,  15,   2,   1,   0,   3, -17,  16,  17,  -5,   1,  -1, -18,  25,   2,   3, -32,   6,  15,  15, -28,  -9,   9,  -9,  30,   3,   5,  44,   6,  11,   6,  10, -29,  -1, -37,  15,   9,   0,  19, -21,  20, -20,  27,   5,   8,  -9,   5, -26,   6,  20,  -8,   6, -18,  -4,  11, -13, -24,  -2 },
      {   0,  -3,  -8,   8,  -6, -67,  32, -26,   1,  -5,  -7,  -4,  11,  19,  -3,   5,  -2,  -2, -19, -19, -11, -10,   8,  -1,   9,   2,   7, -17,   1,   8,  14,  -3,   8,  12,  -5,  12,  -8,  28, -21,   1,   9,  17,   8,  -7,  23,  -8,   3,   6,  39, -17,  -7,  -6,  -9,  -6, -27,   6, -19, -12,  -3, -10,  15, -10,  34,  -7 },
      {   3,   4,  -1, -12,  33,  60,  -7,  15,  -3,  -4, -18, -24,  -9,  -8,  -3,   1,   0,   5,  -7, -48, -22, -23,   9,   5,   6,  10,  27,  -6, -18,  38,   6,  -3,  -6,   0,  -2,   3,   1,   8, -21,   7,  10,   7,  11, -15,  -4,   3,   1,  -3,  21,   4,   0,  -2,   2,  10,  -9,  -3,  -3, -20,   2, -37, -11,   3,  11,  15 },
      {  -5,  -3,  -7,  -7, -19, -40, -14, -12,  -4,   0,  -4, -12,  -6, -50,  42,  -7,   0,   6,   9, -23, -13,  36,  -4,   2,   0,   5,  22,  13, -12,   6,   1,  -6,  -8,  -4,  -1,   8,   6,   8, -24,  15, -22,  -3,   3,  -8, -16, -19,  -2, -11,   4, -19,  15,  -3,   5,   0,  45,  -1,  -8,  14,   6, -13, -24,   5, -17,   6 },
      {   1,  -3,  -2, -10, -17,  -6,  -1,  -8,   3,  -1,   1, -18, -33,  -6,  12,   2,   9,   2,  25,  15, -18,  41,   6,  -4,   8,  -2, -14,  35,   3,  25, -22,   4,   8,  18,  -5,  -7,  11,  -5,   4,  -5,  22,  18,  21,  21,   8,  17,   2,   1,  24,  23, -12,   4,   8, -12, -36,  10,  28, -36,  18,   5, -11, -18, -12,   8 },
      {   1,   5,   1,   3,   8,   7,   9, -10,   1,   7,   0,   9,  12,  -4,  -8,  11,   2,   9,  -5,  -9,  30,  19,  18, -12,   8,  24,  13, -10,   3,  16, -23,   7,  -8,  23,  32,  18,  33,  12,   9, -22, -31, -16,  15, -10,  12, -13, -39,  45,  12,   6,  21,  10,   9,  -4, -18,  -3,  12,  20,  18,  -1,   9,  17, -16, -10 },
      {   0,   3,   7,  -3,   0,  23,  20, -11,  -2,   0,  10, -10, -15,  24, -28,   9,   0,   8,  25,  20, -19,  -4,   9,  -4,  -2,  -9,  13,  42,  21,  21, -16,   6, -16,  -2, -22,   9, -10, -16, -19,   0, -22, -13,  14,   0,  11, -27,  15,   2,  13, -39,  -4,   9, -10,   5,  14,  -2, -34,  26,  26,  19,  -1,  -5,   8, -14 },
      {  -2,  -4,  -5,  -6,  -9,   1,   1, -21,  -4,  -6,  -9,   0, -31, -67, -16,  18,  -2,  -2,  -6,   9,  24, -32,  28, -17,  -1,  -6, -16, -24,  -3,  -7,  -5,  18,  -1,   9,   6, -12,  -8, -25,  24, -25,   0, -18,  14, -10,   5,  -9,  32,  -6,  18, -12, -11,  -7, -11,  -9,   2,  17, -12,  -6,  -4, -15,  -4,  -6,  -4,  -1 },
      {   0,   2,   1,   1, -11,   9,  37, -20,  -3,   0,  -1,  -5, -24, -20, -12,   3,   0,   5,  -3,   1, -18, -23,  23,   0,   1,   6,  13,  -2,   5,   9,  22, -11,  13,   5,   1,  17, -12,  26, -11,   0,  -8,  31,  -5,  16,   6,   6,  -4, -14, -61,   8,  21,  -1,   1,  -8, -15,  27,  29,  35,   2,   4,  -8, -15, -14,  -9 },
      {  -2,  -2,  -4,  -5,   0, -23, -54,  29,   0,   2,  -2,   5,  -3,   6,   6,   5,   2,   2,  -1,  -3,   8,   1,  22, -10,   2,  -8,   1,   9, -22,  29, -21,  17,  10,   9, -37,  21,  -8, -12,   8, -17,   4,  36,   1, -15,  20,   8,  12,  12, -26,   3,   3,  12,  -6,  -1,  -9, -14, -17,  39, -20, -17,  23, -10,  -1,  -1 },
      {   0,  -1,   1,  -3,  -4, -16, -21,  51,   0,   0,   1,  -3, -26, -28,   6, -18,   3,   7,  10,  20,  11, -27,  33,  14,   2,   3,  11,  -6,  25,  19,  21, -36,  -1,  -7,  23,  13,  -7,  -1, -18,  23,   5, -16, -14,  16,  12,   6, -14,  15,  21,   9,   2,   0, -10,   2,  -4, -18,  -3,  -5,   6,  35,  16,  -1,   8, -10 },
      {   2,   0,   0,  -2,  -8, -11,   8,  22,   2,  -1,  -1,  -7,   0,   6, -20,  11,   1,   0,  -2,  -6, -35, -16,   5, -10,   2,   2,  10,  10,   4, -24,  22,  15,   3, -13,  -5,  -3,  -1,  19,  18, -29,  24,  13,  -2,  -1,  -9, -14,   4,  21,  44,  17,  18,  12,   8,  25,  10, -12,  -2,  16, -25,  17, -24,   1, -58,   2 },
      {   3,   1,   2,   8,   1,  29,  29, -56,   3,   0,   4,  14,  -5, -16,  22,   1,   3,  -1,   1,  14,  30,  10,  -5,   5,   1, -10,  -2,  -3,   3,  21,  -7, -12,   6, -12, -16,  15,  -3,  -5, -17,  23,  22,  12, -21,  -1,   9,  12,   7,   6,  23,  25,   4,  16,  13,  14,   5, -27, -21,  10, -37,   4,  23,   9, -20,   6 },
      {  -2,  -2,   1,  -2,  -1,   2, -60, -54,   0,   1,   3,  -6,  -3,  17,   5,  29,   0,   0,   2,  -5, -13, -15,  -4,  -3,   2,   9,  18,  -8,  -7,   5,   2,  26,  -3,  -8,  27,  27, -18,  14,   0, -18,   1, -11, -23,  26,   7,  11, -12, -11,  -6,  13,   2,  -7,  -8,   2,   8,   4, -22, -24,  15,  32,   0,  -8,  -5,   3 },
      {   0,   1,  -1,  -1,   4,  -5, -28, -14,   1,   3,  -1,  -2,  13,   4, -39,   9,   1,   5,  -3,  -1,   1,  21,  30,  13,   2,   4, -12, -15,   7, -14,  -2, -30,   6,  23,  16, -24,   2,  13,   7,  41,   5,   5,  32,  13,   1,  14,   8, -36,   1,   5,  -4,   6,  12,  26, -14, -15, -34,  29,  17, -10, -15,   9, -17,   7 },
      {   0,   2,   2,   0, -11,  -4,  19,   0,  -2,   0,  -2,  -4,  -8, -15,  16, -25,  -1,   1,   4, -14,  -9, -22, -52,  42,   3,  10,   8,  12, -19,   2,  20,   3,   1,  11,  17,   4,  17, -17,  30, -16,   5, -10,  23,  12,   2,  24,  -8,   5,  -8,   1, -17,  22, -12,  -3, -20,  -9, -43,  26,  17,   0,  11, -13,  -9,   4 },
      {   2,   4,   4,   1,   2,   3,  41,  40,   2,   4,   5,   1,   1,  13,  44,   7,  -1,  -1,   1,  -8,   7,   6,   9, -42,   3,   2,   8, -20, -11,   5, -31,  25,   2,   9,  23,  16, -14,  13,  13,  10,   2, -11,   9,  12,  10,  22,  13, -49,   8,  -1,   3,   9, -16,  21,   7,   8, -12,   8,   0,  21,  -5,  -2,  -2,   1 },
      {  -1,   0,   2,   0,  -1,   3,  21,  29,  -2,  -1,   4,   3,   1,  -8, -24,  12,  -1,  -2,   2,   6,   3,  35,  12, -21,  -2,  -3,   2,   5, -13, -10,  10, -13,  -2, -11,   0,  20,  12,   8, -15,  -8,   3,  -7, -33,   0, -11,  -5,  -3,   9, -24,  16,  -9,  -6,   4, -26, -16,  23, -72, -29,   6, -16,   1,  -9, -19,  17 },
      {   0,   1,   1,   2,   1,   7,  -4, -12,   0,   2,  -1,   7,  -2,  18,  69,  -3,   0,   0,  -3,  10,  -6, -35,  29, -30,   1,   2,  -9,   8,  44,  -9,   9,  -4,   2,   2,  -5, -36,  13,   9,  -2, -11,   5,   8,  16,   1,  -4,  -3, -15,  23, -22,  -8,  -6, -11,  12,   5,  -6,  -9, -30,  -3,  14, -19,  -9,   1,  -5,  20 },
      {   0,   1,   0,   0,  -1, -15,   2,  19,   0,   1,   0,   3,   5, -31, -26,  34,   3,   5,   6,  13,  17,  -6, -38, -16,   2,   5,   2,  16,  38,  24,  22,  44,   0,   0,  -7, -14,  15,  23, -19, -10,   4,  10,   4,   7,  -1,  18, -15, -27,  -7,   9,  -2,   6,  27,  16,  13, -13,  -3,  -2,   9, -13,  16,  21,  23,   4 },
      {   0,  -2,  -1,   0,   0,   0, -16,  -9,   0,  -2,  -1,   2, -12,  -1,   8,  17,   0,  -1,  -1,   1, -20, -21,   8, -18,   3,   3,  11,  30,   4, -29,  -2,  -5,  -4,  -8, -11,  15,  47,  -4,  20,  38,   4, -18, -13, -28,  -6,  -4, -32, -36,  16,  10, -12,  23,  -4,  -6, -18,  32,   6,  13, -19, -10,  23,   1,   8, -24 },
      {   1,   1,  -1,  -1,  -2,   6,   8,   3,   2,   1,  -3,  -2,  -2,   2,  13,  37,  -1,  -2,  -6,  -9, -16,   5,   1, -22,   0,   1,  -8,  -7, -24, -24,  18,   1,   3,   4,   4,  -6,  -9, -28, -17,  14,  12,   8,  -6,  15, -10,   9,  15,  27,  14,  39,  -7,  -6,   7, -19,  23, -16,   6,  32,  53, -13,  17,  -1,  17, -37 },
      {   0,  -2,  -3,  -2,  -2,  -9,  -7,   7,   0,  -2,  -4,  -3,   3,  -7,  -7,  -4,  -1,  -5,  -5,  -4,   7, -21, -41, -13,   0,  -5,  -6, -16,  13,  14, -47,  12,  -2,  -5,  -1, -16, -38,   3, -24,  31,   3,  -6, -10,  19, -18, -24, -35,  16,   6,  10, -30,  -7,  -4,  -6, -25,  19,  -2,  27,  -4, -13, -20, -18, -26,  -6 },
      {   0,   1,   0,   1,  -4,  -1,   5,   0,  -1,  -1,  -3,   2,  -6,  -4,  -5, -59,   1,   0,  -1,   1,  -3, -21,  15,  28,   3,   1,   1,   7,   6, -25, -42,  41,   3,   2,  -7,  16,  10,  20,   5,   6,   4,  10, -25,  -3,  10, -16,  17, -12,   6,  30,  13, -23,  30,  -3,   4,  -6, -13,  -4,  35, -10,   2,  18,   6,  -9 },
      {   0,  -1,   0,   1,   2,   5,  -7, -10,   1,  -1,  -1,   3,   4,  14,  -1, -48,   2,   1,   4,   3,   8,  25,  32,  -1,  -1,  -3,   1,  -1,   4,  18,  25,   6,  -2, -11,   5,  -6,  -7,  15,  -8, -44,   8, -22,  -7,  23, -18,  13,  -6, -12,  20,   9, -46,  15,   6,   4,  20,  33,   7,  33, -10, -25,   5,  -1,  10, -17 },
      {   0,   0,  -1,  -2,   2,   1,  -7, -14,  -2,  -2,  -3,  -4,  -2,  -2, -19, -48,  -1,   0,  -2,  -4,   8,   6,   1, -30,   0,  -2,  -6, -10,  -2,  -9,  13,  29,   2,   4,  -8, -22,  11, -20, -39,  11,   6,  14,  17, -34, -10,  33, -21,   8,  -2,  10,  35,  12, -42,  10,   5,  16, -14, -10,  11,  28,  -4, -12,  -6, -13 },
      {  -1,  -1,   1,  -2,  -1,   3,  -4,  -6,  -1,  -2,   0,  -4,  -6,  -4, -14, -20,   0,   1,   3,  -5, -15,  -3,  10, -28,   0,   3,  10,   3,  -5, -28,  -4,  29,  -1,   3,  18,   7,   5,  19, -22,  18, -11, -11,  10,  23, -28,  11,  23,  11,  -7, -27, -19,  26,   3, -41, -16, -51,  16,   1, -25,   3,  26,  -1,  -1,  29 },
      {   0,   0,  -1,   2,   3,   0,   2,  -8,   0,  -1,  -1,   1,   3,   1,  -8, -18,   0,  -1,   0,   4,   1,  13,  -3, -31,   1,   0,   1,   6,  -3,   9,  43,  20,  -3,  -4,  -5,  17,   8, -20,  19,  21,  -5, -12, -27, -10,  39,   1,  -8,  -5,   9,  -3, -22, -55, -13,   9, -19, -23,  18,  28,  13,  -5, -25,  -4,  -6,  39 },
      {   1,   0,   0,   1,   1,   1,   7,   9,   1,   1,   1,   2,   3,  -2,   9,  37,   0,  -1,   1,   4,   0,  -3,  20,  54,  -3,  -6,  -7,   0,   1, -31, -26,  13,  -1,  -4,  -4,  -1,  23,   1, -57, -27,  -1, -10, -10, -13,   6,  35, -11,  -6,   9,  -2, -15, -16, -28,   2,  -6,  -2,  12,  25,  -1,   3, -13,  -6,  -3,  29 },
      {   0,   0,  -1,  -1,  -3,   1,  -8,  -7,   0,   0,  -2,  -2,  -4,  -2,  -7,  -8,  -1,  -1,  -4,  -1,   0,  -7, -26, -28,   0,  -1,  -3,  -1,  11,  13, -22, -52,   0,  -2,  -9,  -9,   6,  35,  12, -32,   3,  -1, -13, -27, -28,   2,  20, -18,  17,  22,  13,  -5, -25, -25,   0,  -8,   3,  30,  27,  20,  10,   1,  17,  37 },
      {   0,   0,   0,   1,   4,   3,   0,   0,   1,   1,   2,   3,   6,   5,   8,  18,   1,   1,   2,   4,  13,   9,   8,  29,   1,   2,   0,   3,  22,  30,  14,  30,  -3,  -2,  -4, -10,  10,  47,  33,  36,  -5,  -7, -11, -22, -29,  21,  34,  23,  -1,   0,  -6, -12, -36, -21,   5,  -4,  -3,  -2,   1,  -6, -13, -22, -13, -33 },
      {   0,   0,   0,   0,   1,  -2,  -7,  -3,  -1,  -1,  -1,  -1,  -2,  -5, -15, -12,  -1,  -1,   0,   1,  -6, -14, -28, -32,   1,   2,   5,  10,   6, -12, -28, -38,   3,   4,   9,  18,  30,  15, -12, -10,  -3,  -3,  -2,  10,  33,  47,  25,  22,  -8, -13, -18, -19,  -2,  23,  22,  -5,   2, -11, -13, -24, -19,  -8,  -9, -41 },
      {   0,   1,   1,   1,   1,  -2,  -5,   2,   1,   1,   1,   1,   0,  -2,  -6,   3,   0,   1,   1,   2,  -3,  -4,  -4,   3,   2,   2,   3,   4,   3,  -5,  -3,   8,   5,   6,   7,   8,   7,   5,  11,  35,   4,   5,   6,  12,   8,  12,  23,  53,   4,   6,  -1,  13,   7,  16,  29,  67,   0,  15,  -3,   9,   5,   9,  26,  61 },
    },
    {
      {  93, -45,  -8,  -2,  -1,  -1,  -1,   0, -65,  17,  16,   1,   2,   0,   1,   0,   4,  21,  -8,  -4,  -1,   0,  -1,   0,   8,  -8,  -7,   2,   1,   0,   0,   0,  -2,  -2,   3,   2,   0,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
      { -44, -22,  40,  -2,  -1,   1,   1,  -1, -37,  62, -11, -10,   1,  -1,   0,   0,  56, -18, -36,  12,   0,   0,  -1,   0,  -7, -30,  23,  10,  -3,   2,   0,   1,   1,  10,  10, -11,  -2,   0,   0,   0,   0,   1,  -6,  -2,   3,   0,   0,   0,   1,   0,  -2,   2,   0,  -1,   0,   0,   1,  -1,   1,   0,  -1,   0,   0,   0 },
      {  -9,  79, -50,  -9,  -4,  -1,   0,   0, -55, -18,  13,  24,   1,   4,   0,   0,  40,  -6,   4,  -7,  -6,   0,  -1,   0,  19, -24,   0,  -1,   2,   1,   0,   1, -11,  -6,  18,  -1,   2,  -1,   1,   0,   1,   8,  -3,  -6,   2,  -1,   0,   0,   1,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -53, -19,  -9,  28,   1,   3,   0,   1, -19,   2,  45, -14,  -6,   0,   1,   0, -31,  62, -19, -26,  14,  -5,   1,  -1,  32, -13, -31,  25,   0,  -1,   0,   1,  13, -19,  15,   3, -11,   3,  -1,   0,   4,  -3,   2,  -5,   2,   1,   0,   0,  -3,   2,   2,  -1,   1,  -1,   0,   0,   1,   2,  -2,   0,   0,  -1,   0,   0 },
      {  -6,  14,  53, -40,  -2,  -1,   1,  -2, -16,  -9, -24,  -1,  12,   0,   2,   0,  10,  20,  -9,  28, -13,   0,  -2,   0,  47,  25, -40, -20,  -7,   3,  -2,   0, -13, -42, -16,  24,  11,   3,   1,   0,  -6,  -3,  30,   3,  -8,  -1,  -1,   0,   7,   3,  -6,  -9,   3,   0,   0,   0,   1,  -6,  -1,   5,   2,   0,   0,   0 },
      {  26,  -1,  15, -15,  -3,  -1,   1,   0,  51, -42,  16,  -6,   3,   1,  -1,  -1,  41,  24, -54,  -4,   5,  -1,   0,   0, -17, -28,  -9,  23,   0,   1,   1,   1, -40,   1,  34,   3,  -6,   0,   0,  -1,  -3,  20,   0, -11,   0,   0,  -1,   0,   1,   0, -12,   1,   3,   0,   0,   0,   0,  -3,   1,   4,  -1,   1,   0,   0 },
      {  26,  38,   2, -10,  -6,   1,  -1,   0,  14,  34, -34, -10,   4,  -3,   0,  -1, -62,  -5, -24,  26,  11,   1,   3,  -1,  31, -35,  22,  24, -17,   1,  -1,   1,  12, -19,  34, -18,  -9,   5,  -2,   1,   0,   6,  -1, -12,  10,  -1,  -1,   0, -10,   5,  -2,   1,   2,  -3,   0,  -1,   0,   3,  -1,  -1,  -1,   0,   0,   0 },
      {  -6,  19,  51, -32, -23,   5,  -2,   2, -40, -45,  18,   9,  15,   3,   0,   0, -30,  22, -12,   7,  14, -12,   4,  -1, -30,  20,  30,   6, -11,   0,   3,  -1,  13,  31,  -4, -22,  -2,   4,  -2,  -1,  17,  -6, -29,   5,   7,  -1,  -1,   1,  -3,  -6,  10,   7,  -4,   0,   0,   1,   3,   6,   4,  -7,  -2,  -1,  -1,   0 },
      { -18, -37,  14,  -7,  -1,   8,   1,   2,  -6, -12, -33,  50,   6,  -4,  -2,   0, -15,  23,  47,  -9, -29,   1,  -2,   1,   6, -33,   9, -23,  12,   9,   0,   1, -17,   0,  40,  -3,   8,  -7,  -1,   0,   5,  38, -16, -21,   3,  -4,   3,  -1,  -2,  -1, -18,   8,   4,   0,   0,  -1,   0,   1,   4,   2,  -4,   0,  -1,   0 },
      { -11, -28, -15,  31,   4,   0,  -1,   2, -10, -36,  18, -26,   5,   4,   2,   0,  -4, -24,  -5,  46,   4,  -1,  -2,   0,  52,  12,  29, -15, -29,  -2,  -2,  -1, -40,  14,  -2, -34,  13,   6,   3,   0, -12,  18,  -5,   4,  13,  -4,  -1,  -1,  12,   7, -14,   0,  -3,  -2,   0,   0,   1,  -8,  -1,   7,  -1,   1,   0,   0 },
      { -16, -20, -45, -66,  67,   1,   5,   0,   8,   8,  16,  19,  -1, -12,   1,  -2,  -8,  15, -24,  33, -30,   3,   3,   1,  -5,   6,   4,   3,  -7,  10,  -3,  -1,   8,   9,  -5, -12,   2,   2,  -3,   0,   8, -12,  -5,   5,   2,   0,  -1,   2,   0,  -7,   5,   4,  -2,   0,   0,  -1,  -1,  -1,   2,   1,   1,   1,   1,   0 },
      {   1,  13,  21, -34,  14,  -1,   1,   0,  -4,  28,  45, -27, -12,  -2,   0,  -1, -27, -41,  19, -11,  10,   1,   1,   1,  12,   7, -33,   5,  24,  -6,   1,  -1, -27,  31,   7,  17,  -8,  -9,   2,  -3,   3,  38, -20, -14, -11,   3,   1,   0,   1, -17, -27,   9,   4,   7,   0,   0,   8,  -8,  10,  10,  -1,  -1,  -1,   0 },
      {  13,  26,  10,  10,   3,  -7,  -2,  -2,  23,  24,  -3, -17,  -1,  -5,   2,   0,   6,  47,   6, -11, -15,   1,  -1,  -1,  29, -37,  -2, -28,   0,   5,  -1,   1, -22,  33, -38, -10,  29,  -3,   1,  -1,   5, -10, -38,  34,  -1,  -9,   0,   0,  -4, -23,  18,  14, -16,   6,   1,   1,   1,   0,   6,  -4,  -1,   1,  -1,   0 },
      {   0,  14,  -3,  34,  46, -24,  -6,  -5, -24, -11, -55, -37,  -4,   5,  13,   3,   5,  26,   7,  19,  -2,   5,  -4,   0, -30,  22, -22,  20,  -4, -10,  -2,   0,  -7,   8,   7,  -1, -15,  10,   3,   1,  22,  11, -13, -14,   4,   6,  -5,   0,   6, -28, -15,  18,   6,  -1,  -1,   2,   1,  -7,  15,   8,  -6,   0,   0,   1 },
      {  -1,   8,  -3,   0,  -3,   5,  -6,   1,  -2,   0,  28, -50,  31,  -3,   6,   0,  -1,  11,  30,  31, -20, -27,   2,  -3, -41, -40,  -1, -13, -18,  23,   7,   0,  25,  -4,  -7,   9,  24,  10,  -7,   0,   6,  21,  30,  -6, -12, -13,  -2,  -1, -12,  14, -19, -16,   5,   2,   5,   0,   2,   0,  -9,  10,   6,  -1,   0,   0 },
      { -10,   4, -35,   8, -42,  10,   6,   5,  -1,  40,  -4,  15,  16,  -2,  -5,  -5, -21,  18, -35,   5,   5,  -8,   5,   1, -30,  38,   8, -27, -11,   5,   6,   0, -43,  -7,   5,  25,  16,   1,  -1,  -1,  33,  16,  -1,   8,  -9,  -9,  -2,   0,  14, -13, -17,  -8,  -4,   6,   0,   1,  -5, -12,  10,   5,   1,   4,  -3,   1 },
      { -12, -27, -25, -13, -25,  16,   4,   1,  -9, -24, -31, -13,  39,  10,  -6,  -1,  -9, -26,  -1,  16,  11, -15,   0,   0,  13, -34, -29,  24,   9,  -2,  10,   0, -13,  30,   1,  32,  -1,  -1,  -2,  -2,  17, -24,   5,   4, -22,   9,  -2,   0,  -1, -22,  39,   5,  -3,   3,  -2,   0,   3,  11,  13, -11,  -3,  -4,   1,  -1 },
      {   0,  -8, -23, -28, -27,  32,  -1,   7,  15,  14, -12, -10,  25,   3,  -8,  -1,  29,  14,  20,  -4,  21, -32,  -3,  -1,   6,  22, -21,  10, -17,  -7,  11,  -2,  27, -13,  -6, -26,  -9,  13,   0,  -1, -41,  17, -31,  -3,  25,  -1,   0,  -1,  -1,  -9, -16,  31,  -4,  -2,   1,   0,   5, -13,  19,   4, -12,   6,  -1,   1 },
      { -10,  -5, -15, -21,  17,   5,   5,   1, -19, -18, -37, -20,   0,  10,   0,   4, -17,  10, -22, -30,  25,   7,   1,   3,  -6, -17,  32,   6,  29, -14,   1,  -4,   3,   3, -33,  20,  11, -14,   5,  -2, -36,  31,   7,  30, -12,  -4,   5,  -1,   6,  26, -23,  -8, -13,   8,  -1,  -2,   8, -25,  -8,  12,   3,   7,  -4,   0 },
      {  -3,  -3,   8, -15,  26,   6,   0,   1,  -6,   8,   1,   4, -15,  -2,   0,   2,   1,  -5,  37, -11,  35, -16,   7,  -4, -19, -10,  -1,  22, -30,  -9,   7,  -1, -48, -36,  -2, -12,  -5,  28,   2,   2,  14,  22,  12,  41,   8,  -3, -10,  -2,   2,   4,  29, -11, -24,  -5,  -1,   1,  -5,  15,  -8, -29,   9,   5,   1,   1 },
      {   3, -12, -10, -24,  -3,   6,   8,   1,  11,  -7,  -4, -20, -28,  16,   2,   1,  25,  13,  15,   3,  42,  -8, -12,   3,  27,  -1,  36, -25,   2, -16,  -2,   2,  18, -15,   0,  12, -16,  -1,   4,  -4,  61,  -5, -13,  -8, -16,   5,  -2,   1, -31, -13, -15,  -9,  19,   7,   4,   1, -16,   7,  -6,  17,  11,  -6,   1,  -4 },
      {   0,   4,  -3, -10,  -5, -34,  23, -11,   9,  18,   0,   5,  50, -15,   0,  -6,  10,  25,  29, -19,  -6,   7,   0,  -3,  19,  12,  27,  32, -30, -10,  -1,   0, -18,  33, -19,   9, -37,   0,   1,   3,  -2,  -3,  14, -22,  -8,  18,  -2,   0,  12,  11,  -8, -21,  12,  -4,  -3,   1,  -2,  -7, -23,   0,  10,  -2,   3,   0 },
      {   4,  12,   0,   7,  -1,  80, -28,  10,  -2,   6, -12,  -5, -25, -28,  -1,   9,   3,  21,   8,  16,  -8, -13,  16,  -3,   0,   6,   1,  10,  30,   5,  -4,  -7, -27,  24, -13, -18, -25,  -6,  -2,   1,   5, -14,  17,  -8,  -2,  18,   3,   3,   5,  18,  -5, -13,  18,  -6,  -1,  -1,   7,  -2, -19,   3,   5,  -4,   1,   1 },
      {   3,  -2,   4,   5,  19,  18, -16,   2,  -2,  -8,   2, -29,  47, -25,   1,  -1,   3, -22, -11, -46, -36,  -4,  21,   7,  13,  18,  30,  13,  13,  14,  -8,  -6, -11, -45,  -8,   5,   7,  -7,  -5,   1,  29,   4, -18,  -7,   4,   0,   8,   0, -23,   1,  10,  22,   1,  -7,   1,  -2,  -7,  12,   8,  -1,  -8,   3,   1,   0 },
      {   9,   4,   9,  20,  26,  19,  -5,  -1,   4,  -6, -12,  16,   3, -22, -13,   5,  14,  -8, -11, -12,  11, -24,  20,  -1,  25,  21,  -9,  10, -26,  21,   7,  -5,  40,  26,  26,   8,  12,  -6, -18,   0,   5,  27, -10,  12, -34, -19,   4,   1,  10, -19,  -2, -32, -21,  23,   6,   3,  -5,  -5, -15,  -8,  25,   4,  -3,   0 },
      {  -2,   0, -10, -23,  -4, -14,   7,  -6,  -1,   8, -16, -42,   4,  27, -11,   1,   3,   7,   8, -15,  -3,  -6,  -7,   6,  -1,  26, -15, -11,  20,  13,   2,   1,  -8,  14,  49, -24,  22, -16,  -5,  -4,   6, -19,  -9,   5,   8, -10,  12,   2,  22,  39,   8,   0,  -7,  -7,   4,  -4, -24,  14, -50, -14,   2,   3,   5,  -2 },
      {   2,  10,   7,  21,  39,  -6, -14,   5,   2,  11,   9,  16,  32,   5,  -6,  -3,  -1,   8, -10,  20,  22, -40,   0,  -1,  -1,   6,   5, -45,  13, -24,  18,   4,  -7,   6,  16,  30, -28,   3,  10,  -6, -24,  -1, -14,  -5, -12,  10,   2,  -1, -28,  34,  19,  18,  -4,   3,  -2,   0,  10,  10, -13,  -8, -15,  -1,   1,   1 },
      {   2,   2,   0,   0, -10,  -9, -26,  18,   9,  11,  13,   0,  -6,  56, -10,   0,   6,  17,   9,  22, -37, -17,  -1,  -5,   8,  17,  27,  34,  29, -27,   0,  -1,   4, -13,  -1, -12,   7, -17,  14,   1,  -6,  26,   9,  -8, -34,   6,   5,  -5,  10, -16,  19,  -9, -14,  17,  -4,   0,   1,   6,  16, -23,   7,   2,  -4,   2 },
      {  -3,   1,  -9,  -3, -14,   2,  30, -11,  -5,  -1,   2, -25,  -1, -47,   7,  -1,   8,  18,  11,  27,  13,  41,  -7,   0, -10,  17,   9,  -6,   9,  10, -17,  -1,  -2, -10,  22,   2, -11, -35,  -3,   8, -33,  18,  -4,  15, -17,   0,  17,  -3, -28, -11,  19,  -6,   5,  17,   0,  -3,   9,  28,  16, -32,   0,  -7,  -5,   1 },
      {   4,   1,   2,   8,  -9,   8,   4,  12,   7,  -3,   9,   6,  25, -22,   3,  -4,   8,  -3,  -2,  24,  -5,  32, -14, -11,   6,  -7, -18,  -1,  22, -28,  -6,   7,  30,   2,  -3,  -2, -18,  -7,  19,  -2,  49,  26,  -1,  23,  15,   5,  -6,  -6,  12,  19, -12,   8, -40, -18,  -6,  -2,  -4, -23, -17, -27,  -9,  31,   7,   2 },
      {   3,   6,  10,  -5,  20,  28,  -2,  21,   2,  -1,  11, -24,   1,  -5, -20,  -3,   2,   9,  19, -10,   3,  21,  -3,  -8,   0,  -6,  30, -17, -19, -15,  12,   1,   7,   0,  35,  35,   1,  -7,   4,   0,  -9, -33,  10,  -5,  12,   3,  -4,   2,  59,  -3,   5,  -3, -14,  -3,   2,   0, -10,  -8,  48,   2,  -8,   6,  -3,  -1 },
      {   2,  -1,   3,   6,  -7, -20,  13, -40,   2,  -5,   9,   4,  -6, -10,   3,   5,  -2,  -4,  10,  23,  -2, -43,   9,  14,  -1, -15,   9,   9,  23,  34, -10, -10,   1, -30,  -3,   6, -37, -12,  -5,  -3,   3, -15, -30,  29, -12,   9,  11,   6,  48,   8, -11,   4,   4,  13, -11,  -1, -18, -23,  12,   4,  -7,   2,   0,   0 },
      {   0,   4,   8,  13,  18,  17,  15,  25,  -4,   1,   7,  15,  36,   5, -25,  -6,  -8,  -7,   4,   9,  13,  18, -22,  -5, -20, -14, -19,   8,  -3, -21,   1,   8, -13, -29,  -9, -32,   9, -30,   6,   2,  -6, -22, -22,   2, -19,  10,  24,  -1,   1, -14,  -9,  -7,   5,  20,   4,  -4, -29,  -4, -17,  39,  32,  -6,  -3,  -4 },
      {   1,   2,   5,   3,   5,   0,  -5, -30,   3,   6,   3,  14,  16,  -9,  12,  -3,   3,   4,  13,  -2,  12, -11, -17,  11,   7,  11,   4,  20,  21, -14,  -3,   2, -11,  15,  20, -27,  10,   0,  -3,   1,   8, -34,  40,  19, -20, -41,  -2,   5, -30,  12, -30,   5, -25,   6,  17,  -1,  -1,   0,  41,  21, -16,  18,  -7,  -1 },
      {   1,   3,   6,  10,  -2,  15,  53,   3,  -5,  -3,  -3,  -6, -27,   7, -30,  -6,  -1,  -1,   4,  11, -42, -12, -17,  14,   0,   8,  -2,  28,  -8,  -8,  27,  -3,  -8,   2,  -1,  28,   0,   0,  -4,  -7,   1,  -5, -29,   3,  29,  -6,  -3,   7, -43,  27,   5, -31, -13,   6,   3,   4,   1, -17,  10,   8,  -2,  11,  -1,  -1 },
      {  -4,  -5,  -4,  -8, -11, -16,  -5,  31,  -6, -10, -10, -18,  -2, -13,   6, -10,  -9, -11, -14, -25, -38, -15,   6,  -6,  -1,  -7,  -9, -30,   0, -15,  13,   4,  -2,   3,  14, -24, -40,  25,  16,   0, -15, -18,   2,  31, -11,  13, -14,  -1, -14, -14, -36, -23,  10,  18,  -1,   3, -20,   6,  14, -28,  22,  10,  -2,   2 },
      {  -4,  -8,  -6, -10, -11, -30, -57,  36,   2,   2,  -2,   1,   5,   4,  14,   6,   7,   1,  15,  13,  11,  16,  35, -13,  -1,   0, -10,  14,  -9,   6, -25,  -7,  -7,  -2,  10,  18,  -4, -17,  -7,   3,   4, -21, -33,  13,  23, -12,  15,   5, -19,  23,  -4, -37,  -6,  13,   3,   1,  21, -15,   8,  15,  -2,   4,  -6,   2 },
      {   3,   3,   4,   7,   5, -24,  36,  65,   1,   0,  -1,   7,  -4,   1, -39,  -8,   6,   5,   5,   6,  12,  -5,  34,  -6,   8,   2,   3,  11,  11,  23,  16,  -7,   3,   4,  -4, -10,   2, -10, -25,  -1,  15,   2,  25,  26,  -7,   2,   5,   4,   1,   6, -21,  32,  20, -11,  -2,  -9,  18,  14,   8,   0, -23, -10,   4,   1 },
      {   3,   2,   2,   2,  17,  14,  10, -29,   4,   3,   0,  -3,  26,  38, -28,  11,   0,  -3,  -5, -17,   4,   6,  -7,  12,  -2,  -3,  -8, -24,  -2,  13, -11,  -7,   3,  -3,   7, -29, -14,  -9,   9,   5,  21,  -1,  10,  10,  17,  11,  10,  -7,   0,  -7,  -2, -36,  15,   7, -16,  -1,  70, -15,  18, -17,  -2, -10,   0,  10 },
      {   3,   5,   4,  10,  16,  22,  -7,   4,   3,   6,   4,  15,  24,  35,  18, -12,   0,   2,  -3,   7,   5,   6, -24,   4,   7,   3,   2,  16,  11,  19, -13,   7,   5,  13,  -9,  26,  -5,   0,   0,  -3,  -8,  13,  -5,  33,  37, -24,  -3,  -1,   4,  -8, -24, -22,  39, -13,   0,   2, -41,  40,   7, -18,  -9, -14,   7,  -1 },
      {  -2,  -1,  -3,  -2,  -4, -12, -13, -46,  -1,   0,   0,   3,   0,  -9, -46,  21,   1,   3,  -3,   5,   2,   0,  51,  -5,  -1,  -4, -12,  -6,  -4, -34,   5, -16,   7,  10,  -3,  -5,  19, -16,   9,   6,   2,  15,  10,   9,  22,  34,   0,  -6, -13,   3,  -8, -15,   4,  -9, -12,  -3, -44,  25,  19,  14, -13,  -6,   3,  -2 },
      {  -3,  -2,   0,   1,   1,   4, -15, -18,  -3,  -4,   3,   0,  -5,  18, -33,  13,  -4,   0,   1,   9,   2,  12,  13,  13,  -5, -12,   4,  -3, -13,  -1,   6, -14, -12,  -7, -19,  18, -21, -18, -24,   1,  -8, -22,   5, -23,  12, -29,  14,  15, -10, -25, -39,  29, -16,   3,  19,  -7,  -1,   4, -24, -50,  20,  31,  -9,  -1 },
      {   0,  -2,  -4,  -1,  -1, -26,  -3,   7,   1,  -2,  -1,  -3,  10, -37, -22, -11,   2,   3,  -2,  10,  15, -18,  -7,  10,   1,  -1,   8,  -4,  56,  -9,  17,   4,  -1, -12, -11,  -8,   5,  15, -12,  -2,  -5,  -1,  -6, -25,  36, -21, -16,   6,  22, -30,   3, -42, -19, -13,  15,  11,  22,  27,  -6,  10,  16,  -9,   5,  -4 },
      {   0,  -2,   1,   2,  -8,  11,   2, -22,   2,   1,   9,   0,   4,   0,   0,   9,   2,   0,  10,  10, -23,  13,  20,  -4,   0,   5,  13,  18,   3, -39,   2,   5,   5, -10,  32,  24,  15,  23,   2,  -3,  -9, -18, -16,  48,  -9,  -2, -17,  -3,   6, -24, -21,  19,  18, -11,   2,   5,  43,   7, -43,  17,   7, -19,   6,  -3 },
      {   1,  -2,  -5,  -2,  -1,  -6, -14,  -3,   2,  -2,  -6,   5, -10, -28,   0,  10,   0,  -5,  -7,  -3,  -1,  -6, -27,   7,   0,  -6, -11,  15, -22, -25,  25,   2,   3,  -6, -20,  -1,  25,  -9,  20,  -4,  20,  -4, -25, -12, -16,  -8,  10,  -1,  29,  44, -26,   1,  27,  24,   9,  -1,  25,  38,  14, -33, -12, -19, -13,  -4 },
      {   1,   1,   0,   3,   3,   5,   6, -11,   3,   3,   4,   2,  16,  14,  -1,   5,   4,   3,   2,   9,  10,  39,  22,  -8,   2,   1,  -5,  -3,  27,  -6,  14, -12,   1,  -4,  -2, -12,  10,  47, -46,  -3,   2,   1, -23, -11, -22,  -4, -25,  10,  -1,  33,   6,  -3,  19,   1,  10,   2, -17, -30,  28, -28,  29, -23,  -6,   0 },
      {   2,   1,   3,   5,  12,   4,  34,   4,   1,   1,   1,   6,   2,  -1,  28,   2,   0,  -1,   3,   3,  20, -29,  17, -36,   0,   1,   5,  10,  13,   2, -24,   6,  -5,  -5,   3,   6,  53,  -5,  13,   8,  -8, -30, -11, -25,   5,  30,  -8, -14,  -9,  -8, -31, -13,  -1, -13, -23,   5,   5,  -7,   2, -39,   6,  26,  -2,   6 },
      {  -2,  -1,   1,  -2,   1, -14, -40,  -2,   0,   1,   4,   3,   2, -15, -11,  16,   0,   3,   7,   2,  12,  12, -38,   6,   2,   6,  10,  13,   3,  33,  58,  -7,  -4,  -5,  10,  -1,  19, -22,   8, -33,  -5,  -6,   5,  15,  -3,  28, -25,   6, -21, -17,  -1,  -2,  20, -18, -15,  12,  -6, -34,  -7, -13,  -3,   7,   5,   7 },
      {  -1,  -1,   1,  -3,   3,  -6,  -9,   2,  -2,  -3,   0,  -2,  -7, -21,  -9, -37,  -2,  -2,   1,  -3,   0, -25,   1,  21,  -2,  -4,   0,  -1,  -4, -42, -21,  20,   1,   1,   5,  -2,  16, -12, -13,   9,   2,   1,  11,   3,   7, -22,  27,   8,  -5, -16,  15,  -7,  39, -13,   6,  -9, -13, -61,  -7, -33, -22, -25,   9,   0 },
      {   1,   1,   2,   2,   0,  -3,  13,   4,   3,   2,   3,   4,   7,  -3,  48,  18,   2,  -1,   0,   0,  -3, -13,   0,  -7,   5,   2,   4,  -1,  11, -16,  32, -35,   3,   6,   9,   4,   5, -15, -10,   1,  12,   2,  28,   8,  54,   9,  22,   4,   1,  -7,   5,  20,  -4,  43,  -3, -15,  -2, -11,  -3,  -1,  31, -34, -24,   4 },
      {   1,   1,   1,   1,  -1,   5,  -8,  -1,   1,   0,   1,   2,   0,   3,   8, -65,   4,   6,   7,   7,  10,  17,  27,  73,   1,  -1,  -1,   4,   1,   7,   3,   8,   5,   4,   2,   3,  16,  20,  13, -33,   0,  -3,   2,  -4,   4,  29,  16,   1,   0,   5, -11,   5, -15,  16, -29,  -4,   4,  12,   0,  -1,  16,   6, -15,   7 },
      {  -1,  -1,  -2,  -1,  -4,  -9, -21, -12,  -1,  -1,  -3,  -1,  -8,  -7, -35, -20,   0,   1,  -1,   1,  -2, -13, -25, -23,   3,   6,   4,   7,   0,  21, -17,  44,   0,   8,   2,  16,  -1,  -1,   0,  26,   5,   3,   5,  13,  13,  15,   5, -26,  -9,  20,  -9,  30, -12, -17,   4,  -3,   6,  -3,  19,   0,  63, -29, -15,  12 },
      {   0,  -2,  -1,  -1,   1,   9, -11, -26,  -1,  -3,  -2,  -3,  -6,  11,  13, -56,   0,  -2,  -1,  -4,  -4,   9,   1, -46,   1,  -1,   1,  -2,  -3,   0,  40,  28,  -3,  -9,  -7,  -7,  -9, -19, -51,  17,   3,  -3,  -4,  -1,   0,  13,  -7,  -4,   1,  -4, -12, -12, -18,  23, -13,  -1,   5,   9, -14,   3, -40,  -8,   1,   8 },
      {   1,   0,  -1,   2,   0,   0,   5,  -4,   1,   1,  -1,   2,   0,  10,  37, -12,   0,  -2,  -5,  -2,  -5, -13,   7,  14,  -1,  -2,  -8,  -9, -14, -20,  11,   1,   3,  -1,   1,  -6,  -4, -45, -33,  -9,   1,  11,  -5,  13,  -7,  25,   1,  24,   8,   7,   6,  -7,  11, -61,  34,   6,  18,  15,  21,   7,  33,  27,  -7, -17 },
      {   0,   0,   0,   0,   1,  -2,   5, -15,   0,   0,   0,  -1,   1,  -2,  -6, -11,   2,   1,   2,   2,   3,   5,  27, -32,   3,   3,   1,   1,  11,   5,  24,  37,   1,   1,  -1,  -4,  -3,   3,  28, -18,   5,   7,   6,   1,  15, -32,   8,  -7,   7,   0,  27,   5,  52,  18,   0,  -3,   4,  -6,   0,  10,  14,  58, -41, -16 },
      {   0,   1,   0,   0,  -2,  -6, -11,   4,   0,   1,   2,   1,   1,  -9, -17,  19,  -1,  -1,  -2,  -2,  -3,  -3, -42,  -8,   2,   1,   5,   1,   5, -12, -29,   0,   2,   5,   5,   7,  13,  44, -55, -10,   1,   7,   5,  12,   4,  40,  42,  22,   4,  -4,   2,  -3,  13,  17, -10,  -3,   3,  13,  -2,   6, -12,  43, -16,   0 },
      {   0,   0,   0,   1,   1,   0,  -4,  -2,  -1,   0,  -1,   0,   1,  -4, -18, -32,  -1,  -1,   0,  -2,   1, -10, -10, -11,  -2,   0,  -3,   0,  -5,  -6, -43, -24,   1,   1,   2,  -1,   7, -26,  -2, -13,   2,   6,   8,   2,  15,   3, -70,  44,   2,  12,   8,  21,  15,  45,  -5,  17,   7,   4,  -2,   1,  15,  10,  19, -14 },
      {   0,   0,   0,   1,   1,   3,   9,   1,   0,   0,   0,   1,   4,   4,   2,  29,   0,   0,   1,  -2,   0,   0,  16,   4,  -1,   0,   1,   1,  -1,   8, -17,  60,  -2,  -2,  -3,  -3,  -2,  -6,  -5, -59,   0,   0,   9,  -1,  12,  29, -10,  -6,  -1,  -4,  -6,  -2, -17,  27,  60,  13,  -4,  -7,  -2, -11, -21, -25,  -4, -18 },
      {  -1,  -1,  -1,  -2,   0,   0,   5,   7,  -1,  -1,  -2,  -3,  -1,  -4,   3,  27,  -1,  -1,  -1,  -2,  -6,  -6,   3,   5,  -2,  -2,  -3,  -5,  -8, -23, -11,  23,  -2,  -1,  -4,  -5, -12, -22, -35, -45,   0,  -1,   4,  -5,  -2, -35,  -6, -19,   2,   5,   2,   5,  -3, -10, -69,  46,  -4,   7,   5,   4,  11, -10, -11,  33 },
      {   0,   1,   1,   0,   1,  -1,  -2,   9,   1,   2,   2,   2,   3,  -1,  11,  32,   1,   1,   2,   2,   4,   0,   6,  51,   1,   1,   1,   1,   2,  -5,  13,  43,   0,   1,   1,   2,   6, -10, -24,  54,   0,   1,   0,   1,  10,   7, -32, -27,   5,   5,   7,   4,  18,  31,  -1, -18,   0,   1,   2,   4,   8,  34,  40,  17 },
      {   0,   0,  -1,   0,  -1,   1,   1,  -6,   0,   0,  -1,   0,   0,   2,   7,  11,   1,   1,   0,   0,   1,   5,  11, -15,   1,   3,   2,   1,   2,   5,  19,  34,   0,   2,  -2,   0,   1,   9,  16,   2,   1,   1,   5,  -3,   3,  -6,  43,  61,  -3,   4,  -6,   6,  -8,   5, -28,  20,   5,   2,  14,  -4,  20,  -6,  72, -31 },
      {   0,   0,   0,   0,   0,   2,   3,   7,   1,   0,   1,  -2,   5,   2,   0,  24,   1,  -1,   0,   0,  -1,   0,   1,  15,   0,   1,  -1,   0,   0,   4,  -2,  37,   2,  -1,   0,  -1,  -1,  -4,  10,  34,  -1,   2,  -3,  -1,  -4,   2, -28,  70,  -3,  -4,  -3,  -6, -12, -12, -30, -26,  -4,   1,  -8,   0, -14, -19, -67, -10 },
      {   0,   0,   0,  -1,   1,   1,   1,   5,   1,   1,   2,   1,   3,   3,  -1,  -4,   0,   0,   1,   2,   2,   5,   3,  17,   1,  -1,   0,   0,   4,   4,   6,  -6,   1,   1,   2,   0,   3,   3,   4,  51,   3,   1,   6,   4,   3,  11,  11,  -5,   0,   1,   0,   8,   3,   8,  13, 106,  -5,  -6,  -8,  -4, -12,   2, -27, -20 },
      {   0,   0,   0,   0,   0,   1,   7,  -4,   0,  -1,   0,   1,   0,   1,   1,  -3,   1,   0,   0,   0,   2,   2,   9,  -8,   1,  -1,   1,   1,   2,   1,   2,   6,   2,   0,  -1,   0,   3,   3,  20,   1,   1,   3,   0,   1,   0,   4,   9,  44,   4,   0,   4,   1,   7,   2,  36,  17,  -3,   5,  -2,   4,  -1,   5,   4, 109 },
    },
  },
  { // 12
    {
      { -72,  46,  -4,   3,  -1,   1,   0,   0,  75, -41,  -4,  -1,  -1,   0,   0,   0, -33,   4,  13,  -2,   2,   0,   1,   0,   9,  10, -10,   0,  -1,   0,   0,   0,  -2,  -8,   2,   3,   0,   1,   0,   0,   2,   3,   1,  -2,   0,   0,   0,   0,  -2,  -1,  -1,   0,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
      {  45,   5, -17,  -2,  -3,   0,  -1,   0,  -7, -47,  29,   0,   4,  -1,   2,   0, -43,  69, -16,  -7,  -1,  -1,   0,  -1,  34, -35, -15,  15,  -3,   3,  -1,   1, -13,   1,  26,  -9,   0,  -1,   0,  -1,   3,   7, -15,  -2,   3,  -1,   1,   0,  -3,  -2,   5,   4,  -1,   0,  -1,   0,   2,   1,  -2,  -2,   0,   0,   0,   0 },
      { -63,  25,  -7,   4,   0,   1,   0,   0, -26,  17,  15,  -2,   0,   0,   0,   0,  46,  13, -38,   4,  -4,   1,  -1,   0,  16, -59,  27,   8,  -1,   2,   0,   1, -13,  29,  11, -20,   4,  -3,   1,  -1,   5,  -3, -19,   9,   1,   0,   1,   0,   1,  -2,   9,   1,  -4,   1,  -1,   0,   2,  -1,  -2,  -2,   1,   0,   0,   0 },
      { -42,   2,  20,  -5,   2,  -1,   1,   0, -50,  20, -10,   8,   1,   1,   0,   1,   8,  27,  14, -11,  -1,  -3,   0,  -1,  20,  12, -53,  11,  -1,   3,  -1,   1,  21, -53,  35,  15,  -8,   4,  -2,   2,  -7,  21,   8, -27,   6,  -4,   2,  -1,   2,  -3, -15,  11,   4,   0,   0,   0,   1,   0,   5,   0,  -5,   0,   0,   0 },
      {  31,   6,   1, -10,  -2,  -2,   0,  -1,  44,   3, -31,  14,  -4,   3,  -1,   1,  32, -31,   8, -10,  -1,  -1,   1,   0, -21, -31,  -6,  21,   2,   3,   0,   2,  -8,  -5,  58, -19,  -5,  -2,   0,  -1, -19,  36, -32, -18,  15,  -6,   4,  -1,   2, -10,  -4,  26,  -5,   2,  -2,   0,  -3,   2,   6,  -9,  -5,   2,   0,   0 },
      {   4, -37,  14,   7,   2,   1,   0,   0,  47,  48, -15, -20,  -2,  -3,  -1,  -2, -15,   6, -36,  24,  -3,   3,  -2,   1,  38, -36,   9,  -5,  -5,   2,  -1,   1, -20, -29,   4,  15,   6,   1,   1,   1,   1,   6,  38, -12,  -9,  -1,  -2,   0,  -9,   8, -19, -15,  14,  -1,   3,   0,  -1,   0,   0,  15,  -3,  -2,   0,  -1 },
      { -30, -60,  36,  -6,   7,  -1,   3,   0,   2,  24,  -4,   2,  -1,  -1,  -1,   0, -54,   4,   2,  -2,   4,   1,   2,   1,  10,  16,  15,  10,  -5,   1,  -2,   0,  11,  15,   4, -37,   3,  -3,   1,  -1,  -9,  14, -45,  23,  12,  -3,   3,  -1,  12, -15,  17,  14, -19,   3,  -3,   1,   0,   2,   1, -15,   6,   2,   0,   0 },
      {  19,  30, -14, -10,   1,  -1,   1,  -1,  17,  30, -18,  13,  -8,   0,  -3,   0,  14,   7, -13, -16,   4,   0,   1,  -1,  20,   3, -25,  20,  -6,   0,  -2,   0,   9, -38, -15, -11,   4,   4,   2,   2, -11, -19,   4,  51,  -1,   0,  -1,  -1,  12, -13,  47, -28, -30,   9,  -4,   2, -10,   4, -10, -17,  32,  -7,   2,   0 },
      {  12,  39,  11,  -5,  -3,  -3,   0,   0, -33,   1, -44,   9,   1,   3,   1,   2, -22, -28,  39,  18,  -5,   2,  -1,   1,  61,  -7,  15, -33,   2,  -4,   1,  -1, -42,   6,  -8,   2,  13,  -4,   3,  -1,   8,  22, -12,   9,  -7,   1,  -1,  -1,   1, -14,  -2,   3,  -2,   3,  -1,   1,   3,   2,   6,  -1,   1,   1,  -2,   1 },
      { -12, -51,  37,  -7,   3,  -2,   2,  -1,  15, -33,  -7,  11,   2,   1,  -1,   1,  44,  21,  29, -17,  -4,  -1,  -1,  -2,  -2, -12, -31, -12,  10,  -4,   3,   1, -48,  18,  -3,  27,   0,   2,   0,   0,  20,   2,   4,  10, -16,   4,  -4,   1,   0, -11,  21, -20,  -4,   3,   0,   1,  -2,  10,  -8,  -8,  18,  -6,   1,  -1 },
      {  10,  12,   9,   1,  -8,  -1,  -2,  -1,  22,  28, -15, -15,  11,  -3,   3,  -2,  13,  34,  -6,   9, -15,  -1,  -3,   0,   5,   3, -23, -21,  14,  -1,   3,  -1,   0,  12, -30,  22, -12,  -1,  -1,  -1,   5, -36, -22,  -6,  16,   8,   2,   2,  -7,  -2,   8,  60, -13, -12,  -1,  -2,  11, -14,  29, -34, -24,  18,  -6,   4 },
      {  -5,  13,  11, -13,   0,  -1,   4,  -1, -18,  10, -48,  18,   5,   1,   1,   1, -51,  -4,  -4,  10,  -7,   3,   0,   1, -42, -34,  -3,  22,  11,   1,   0,   1,   9,  26,  25,  24, -19,   3,  -3,   0,  29, -32,  21, -16,  -4,   4,  -4,   3,  -5,  10,  24, -13,  -1,  -4,   1,  -2,   3,   5, -11, -16,  12,  -3,   2,   0 },
      { -11, -26, -18,  18, -13,   6,  -1,   2, -13, -40,   3,  -6,  18,   2,   0,  -1,   2, -28,  -7,  20, -13,   5,   0,   2,  12,  -9,  18,   1,  13,   0,  -1,   0,  -2, -25,   6,   9, -17,   1,   2,   1, -16,   8,  32,   0,  27,   2,  -6,  -1,   4,  22,   1,  15, -52,  -2,   5,  -2,   7, -15,  16, -44,  37,   3,  -7,   4 },
      {  -6, -31, -10,  -5,  11,  -2,   3,  -1,  -4, -44, -38,  22,  -9,   6,   0,   3,  17,  -7,   4,  16,   5,   4,  -1,  -1,  30, -12,  12,   5, -14,  -1,  -2,   1,  33, -16,  -1, -11,  10,   2,  -3,  -1, -19, -37,   5,   7,  -4,  -1,   7,   3,  -6,  16,  33,  10,  27,  -9, -10,  -1,  -5,  10,  -5,   3, -52,  15,   5,   0 },
      { -16, -32, -91,  21,   2,   7,   4,   4,   0,  31,  12,  18, -10,   1,   0,   0, -17, -21,  21, -20,   4,  -2,   4,   0,  11,   0, -23,   8,   8,  -2,   1,  -1, -30,   7,   8,  13,  -6,   2,  -1,   1,  21,  -6,  -6,  -2,   1,   0,  -2,   2,  -5, -10,  11,  10,   5,   1,  -3,   0,   0,   7,   5,  -2, -15,   2,   2,   0 },
      {   6,   2,  19, -28,   0,  -3,   2,   1,  12,  12,  23,  28,  -1,  -1,  -4,  -2,   7, -14,  18, -35, -11,   1,   2,   4,  47,   6,   6,  10,  15,  -5,  -3,  -4,   1,  17,  -2, -22, -32,   2,   2,   2,   6, -38,   5, -27,  38,  -1,  -6,   3, -24,  36, -15,  -1,  -8,   0,   8,  -5,   1,  -4, -25,  16,  12, -11,   3,  -1 },
      {   0,  13,  35, -17,   9,  -6,  -1,  -1, -13,  -7,  30,  -8, -13,   1,   1,   0, -16, -39, -31, -20,  23,  -3,   4,   2,  -9,   3, -17,  33,  -6,  -1,   2,  -2, -54, -14,  -5,  10,   3,   6,   2,   1,   3,   3,  16,  23,   4,  -6,   2,   1,   1,  22,  14,  16,  -5,  -9,  -8,  -1,   9,  -6,   1,  -8, -46,  23,  -1,   3 },
      {   8,   1,  17,  30, -13,   4,  -2,   1,  -2, -16,   5, -48,   8,  -2,   0,   0,  12, -33, -23,  20,   7,   0,   4,   0,  27,  26, -13,  21, -11,   1,  -5,   0,   8,   1,  21,  -4, -12,   7,   0,   0,  45, -23, -16, -21,   6,  -5,  -5,   2, -17, -39,  22,   2,   8,  13,   1,   0,  -4,  29,   0,  -1,   7, -20,   6,  -4 },
      {   8,  10,  23,  39,  -9,  -2,  -9,   1,   8,  19,  46, -13, -11,  10,  -6,   0,  -7, -20,  39, -14,   3,  -2,  -3,  -2,  16, -27,  -3, -16,  -4,   6,   6,   4,  31,   8,  19,  11,   7,  -7,  -9,  -2, -16,  -3,   1, -10, -28,  18,   8,   0,  11,  12,  26, -12,  14, -25,  -4,   0,  -3,  -1, -13, -40,   4,  25,  -3,   4 },
      {  -2,   2,  10,  41, -34,  11,  -5,   4,  -7,   5,  -6, -29,  25, -10,  -6,  -2, -11,   3,  29, -11, -10,   6,   4,  -2, -22, -31, -12, -13,  13,  -5,  -3,   1,  -9, -12,  -9,  -7,   1,  13,  12,  -3, -28,  -8,  -2,  19,  28, -24,  -2,   2,  -2,  20,  12,   6,  -8,  30,  -6,  -1,  -4,  18, -24,  31, -26, -25,  13,  -6 },
      {   2,   6,  -4,  43,  -3,  -4,  -3,   2,  -1,  12, -30, -11, -10,   6,   1,   2,  15,  40,  21,  16,  20, -10,  -1,  -5, -10,  38,  18,  20, -32,  -1,   0,  -1, -39,   2,  15, -33,  -5,   4,  -7,   3,  13,   0,   8,  -7,  12,   2,   3,  -1,  -7,  35,   0,  -9,  -5, -20,   2,  -3,   5,  -9, -28, -13,   3,  23,  -1,   3 },
      {   5,   9,   5,  -9,  -2,  12,   3,  -1,  11,  14,  33,  38,   5, -20,  -4,  -2,   8,  15,   6,  32, -32,  18,  -1,  -1,  -1,  21,  28, -11,   0, -14,  -8,   2,   1,  -6,  17,   6, -14,  23,   2,  -2,  24,  19,  15,  -8,  -2, -37,   6,   2,  16,  -6,  21, -12,  -7,  30, -23,   3,  -2,   7,  -6, -28, -36,   3,  12,   0 },
      {  -7,  -1, -18, -60,  13, -11,   9,  -2,  -3,   8,  24, -33,  16,   2,   5,  -2,  -3,  -5,  18,  35, -27, -10,  -4,   1, -21,   1, -10, -29, -13,  21,  -2,   1, -20, -26,  18, -19,  24, -12,   5,   2,  -1, -12,  -8,  -4,   7,  14,  -3,   1, -22,  14,  20,   6,   6, -10,   6,  -4,  -6,  22, -24,  -9,  11,  -1,   2,   0 },
      {  -3,  -1, -18, -29,  -6,  11,  -1,  -5,  -2,   1, -17, -38,  37, -19,   6,  -2,   3,  -1, -16,  -2,   2,  12,   0,  -3,  21,  26,  -6,  15,  21, -17,   2,   0, -11,  31,   5,  18, -30,   6,  -7,  -4, -48,  12,  -9, -14, -13,   3,   9,   3,  34,   8,  10, -16,  20, -11, -13,   2, -16,  -8, -29,  -6,  -1,  10,   7,   0 },
      {  -1,   4,  14,  38, -18,  -1,  -3,   0,  -3,  -1,   6,  55,   8,   0,   2,   2,  -8,  -2, -37,   5, -37,  -1,  -1,   5,  -9,  25,   5,  -8,  16,  17,   0,  -3, -31, -19,   2,  -2, -10, -27,   4,   3, -27,  -7, -12,  -9,  -1,  39,  -8,   3,   8,  -9,  15,   5,  23, -21,   8,  -3, -14,  18,  -8,  11,   3, -10,   0,  -1 },
      {   1,  -8,  -1,   0, -42,  18,  -6,   4,   4,  -4,  -9,  13,  45, -20,  12,  -5,   4, -17, -13,  -8, -26,   0,  -2,   4,   6,   4, -26,   0,   1,  16,  -2,  -1,  19,   1,   9, -14,  32, -13,  -3,  -3,  46,   6, -12,  24, -13,  -3,  10,  -5,  -5,  17, -16, -15,  -4, -13,  -5,   2,  12, -35, -18,  11,  -7,  33,  -8,   4 },
      {  -4,  -3,   2,  -5,  -8,  29, -10,   3,  -9,  -8,  -3,  -2, -23, -24,   7,   3,  -9,  -5,  -7, -20,  -2,  32,  -6,  -5, -17, -14,   6, -23,   1, -28,   8,   8, -18, -35, -13, -21,  -4,  39, -20,  -1,   1, -19, -17, -15,  15, -12,  29,  -2, -16, -20,  -3,  -8,  32, -16, -10,   5,  -6,  -8,   7,  -8,  32,  28,  -5,   2 },
      {  -3,  -3,  -2, -36, -65,  20, -13,   5,   1,   6,   0,  -6,  18,  12,  -4,   1,   3,  16,  19, -30,  25,  -7,  -5,  -3,  -3,  -4,  49,  16,  -7,   0,   4,  -5,   2, -20,  -8,   7, -10, -14,   4,   6,  15,  11,  18,  -2,   1,  21, -11,  -1,  -3, -25,   8,   5,  -1, -20,  10,   0,  -9,  27,  12,  -4, -17,   3,   0,   0 },
      {  -1,   1,   8,  -4, -13,  16, -10,  13,   3,  10,  20,  29,  41, -15,   7,  -4,   0,  -4,  23,  35,  13,  -8,   3,  -3,   1,  -2, -11,  33, -43,  -2,  -2,   5, -15,  16,   2,   2,   1,  17,  -9,  -1, -38, -22,   2,   6, -19,  -6,  13,   1, -20,   2,   1,   1,  12,   3,   2,   3,  19,  -5,  46,  11,  18, -15, -11,  -2 },
      {   5,   7,   3,  18,  71,   5,  -9,  -2,   3,   3,   5,  -4,  47, -27,   3,  -2,   4,   2,  21,  -3,  -7,  25, -11,  -1,  -1, -10,  14,  21,  12, -15,  12,  -2,  10, -15,  -5,   2, -14,   1, -10,   5,  21,  18,   1,  20,   8,  23,   1,  -7, -10,  -8,   5,   5,  -9, -37,  12,   1,   0,  19,   0,  21,  -8,   4,  -4,  -3 },
      {   0,   2,  -2,   0,   0,  20, -10,   9,   0,   4,  -7,   7, -10, -12,  10,  -1,   2,   7, -25, -19,  25,  15,  -6,   1,   5,  25,  16, -21,  16, -11,   8,  -1,   9,   5,  24,  30,  25, -10, -10,   0, -13,  16, -12,   8, -19,  -9,   8,  -7, -59,  34,  16,  13,  -9,   8,   3,  -1,  35,  33, -13, -10,  10, -19,  -3,  -1 },
      {   0,   2,  -1,   9,  37,  -2,  -2,   0,  -3,  -1, -15, -11,  46,  20, -20, -10,  -3,   3,  -7, -61, -24, -12,  17,   5,  -1,   9,  30,  -6, -16,  32, -20,   0, -13,  -9,   4,   0,   9,  -3,  10,  -4,  -7, -13,   2, -17, -11, -22,   6,   4, -12,  -5,   6, -12,  15,  24,  -8,   3,   3, -11,  22, -22,   0,   5,   1,   4 },
      {   2,   1,   5,  -6,  22,  69, -30,  42,  -2,  -3,  -6, -17, -34, -23,  21, -13,   2,   2,  11,  -6, -12,  11, -10,  10,   1,   3,  -1,   4,  -4,  20,  -2,  -9,  -7,   3,  15, -12,   5, -24,   7,   3,   3, -19,  10,   2,  -8,  17, -17,   4,  32,   8,  -7,   5,  -1,  15,   5,  -6,  -7, -12,  13,  -7,  -9, -22,   5,   0 },
      {  -2,  -1,  -1,  -1, -18, -27,  16, -13,  -3,  -1,  -1, -31, -31,  -5,   4,   4,   2,   6,  20,   0, -46,  16,  -4,   5,  -1,  10,  13,  29,  36, -10,  -1,  -3,  -5, -10,  18, -11, -20, -11,  -1,   3,  -9,  -7,  -5,  23, -24,  18,   3,  -7, -28,   1,   3, -27,   6,  10,  -3,   7,  30, -19,  37,   8, -14,  -1,  -6,   3 },
      {  -3,  -2,  -2,   7,  -6, -48, -12, -30,  -2,   0,  -2,   5,   6, -47,  51,   2,  -3,  -2,   4, -17,   8,  17, -35,  18,  -1,  -3,   4,   7, -24,  -3,  17, -18,  -4,   1,  -2,   1,  17, -15,   4,  10,  -7, -14,   6, -14,  -9,  -3, -11,   0,  -9,  -7, -14,   5, -11,  26,   0,   0, -39,  -1,  -2, -22,  -4,  -1,   6,   2 },
      {  -1,   2,   0,   4,   1,  10,  10,  30,  -4,  -2,  -6,  -1, -22, -18,  21, -17,   0,   0,   0,   3, -28, -11,  17,   6,   1,  -7, -16,  19,  -9,  20, -25,  -2,  10,  25, -35,  -2,  -8,  -3,  14,  -7, -25,  32,  17,  -9,  13,  -5,  -4,   4, -47, -25,  -3, -14, -10,  -7,  -3,   5,  -6,  31, -13, -12,  -5,  39,  -5,   1 },
      {   1,   3,  -2,  -1, -11,  -2,  17, -18,   0,   1,  -8,  -7,  -5,  -4, -16,   7,   0,   0,  -9, -25, -39,  24,   7, -16,   7,  13, -15,  -1, -16,  -2,  18,  10,   2,  36,  10, -25,  38,  26, -25,   4,  -5,   3,  45,   7,   6,  -2,  -5,  -8,  21,   2,  -8,  16,   2, -24,  17,  -3,   1,  37,  21, -11,  -3, -17,  -4,  -5 },
      {   1,  -3,  -6,  -2,  -5, -15,   6,  -1,   2,  -4,  -5,  -2,  -9, -30,   7,   3,   2,  -5,   3, -11, -17,  -4,   1,  -2,   1, -14,   7,  16, -35,  17,   1,  -5,   4, -16, -41,  10,  -7,   9,  -6,   4,  11,  23, -18, -23,   2,  15,   1,  -2,  29,  11,  14,   9,  20,   4,  -2,  -2,  71,  -2, -25, -11,   6, -32,   3,  -2 },
      {  -2,  -2,  -4,  -4,   5,  16, -57, -15,   0,   0,  -2,   0,   3,  61,   2,   6,  -4,  -5, -10,   4, -29,   1, -36,   8,  -1,  -1, -16,  -3, -23,  -7,  45, -12,   2,   9, -11,  -7, -19,  15, -11,  15, -11,  13,  -2,  -3,  -2,   1, -10,  -2, -20, -12,   3, -14, -10,  14,   9,   2,  15,  -9, -16,  -2, -16,   1,   1,  -1 },
      {   0,   0,  -2,   0,  -4, -48, -13,  71,   3,   2,   0,   1,  13,  28,  24, -33,   2,   3,  -3,  -2,   0,  10, -17,  28,   0,   4,  -2,   2,  20, -28,   4, -10,   2,   0,   4, -14,  10,  13,  15,  -8,   7,   3,   9,   8,   6, -16,  -8,   6,  21,   4,   3,   6,  31,  -6,   0,  -6,  15,  14,   7,  -6,  12,   3,  -4,  -1 },
      {  -1,  -2,  -5,  -7,  -4, -30, -57, -10,   0,   1,  -1,  -4, -16, -23,   7,   2,   1,   2,  -3,  -4,   7,  10,   8,   6,   3,   9,   4, -31, -13,  39, -17,   3,   3,  14,  28,  20, -13,  11,  10,  -7,  -6,  -3,   5,   6,  35, -11,   0,   0,  14,  -8,  26, -20, -13, -26,   7,   5,  18,  -2,  19,  36,  -2,  13, -13,  -4 },
      {   0,   0,   2,   0,  -1,  -6, -15,   9,   1,   0,   2,  -7, -17,  13,  22,   4,   1,   3,   9,   3, -32, -35,  -6,   0,   6,   4,  23,  34,  17,  28,  15,   2,  -2,  -5,  -8,  32,  15,   7, -44,   3,   8,  -3, -10,   2,   0, -29,  47,  -9,  13,  12,   0,  19,  -8,  -2, -12,   1, -33,  13,  -3,  21,  16,  -2,  -6,  -2 },
      {   2,   2,   4,   5,   4,  12,  58,  12,   1,   0,   0,  -2,   8,   0,  53, -10,  -2,  -3,  -9, -15,   4, -49,  -5,   5,   1,   0,   7, -37, -14,  -7,  21,  -6,   3,   0,  23,   3, -21,  18, -32,   9,  -6,  -5,  -1,   8,  17,   9,  -8,  -6,   7,  -8,   8, -24, -13,   0,  10,   1,  10,  -4,  10,   9, -19,  -1,  -5,   0 },
      {   2,   2,   4,   4,   7,  31,  17, -50,   3,   4,   5,   2,   7,  19,  34,  11,   3,   5,  10,   8,  18,   3, -30,  13,   1,   2,  -2,  13,  35,  19,  -8, -10,  -2,   6,  -5, -12,  19, -25,   9,   3,  -5,  -1,  24,  -5,  27, -29,  -1,   3,  14, -15,  14,   3,  28,   4,  -8,   3,  37,   8,   3,   3,  17,  18,  -9,   1 },
      {   1,   1,   1,   2,  -5,  12,  39,  -3,   1,   1,   4,  -3,  -8,  37,  16, -17,   0,   1,   4,  -8, -18,  54, -24,  10,  -1,   4,  11,  11, -39, -11, -32,   9,  -2,  -1,   6,  38,  16,  10,  18, -15,  -5,  -6, -16,  -9,  -4,   2,  -2,  11,   4,  -3,  10,  12, -26, -16,   8,  -1,  -7, -19, -13,  30,   6,   1,  -2,   2 },
      {   0,  -1,  -1,   2,   0,  16, -15, -53,   0,  -1,  -4,   0,  -7,   4,  21, -34,   1,  -1,  -2,  -4,  -8, -31,   3,  35,   0,  -3,  -3,  -4,  -4, -10, -23,   8,   5,  -4,  -2,  -8, -24,  33,  34, -21,  14,  10, -10,  22, -23,   2,   2,   4,   2,  32, -15,   7,  14, -21,  -2,   0, -14,  26,  13, -15,   6, -19,  -2,   0 },
      {   0,  -1,  -3,   0,  -2,   3,  20,  -1,  -2,  -4,  -3,  -1,  -5, -22, -60, -36,   0,  -1,   0,   3,   5,  -5, -35,  48,   0,   0,  -1,  -3,  -7,  12,  30, -14,  -6,  -3,   6,   1, -11, -10,   1,  -3,  -8, -23,  -4,  -4, -14, -10,  23,   4,  17, -10, -20,  -4, -25,   1,  -2,   0,  22,  33,  -7,  11,   1,  24,  -7,   1 },
      {  -1,  -2,   0,  -1,   6,  -5, -27,  -8,   1,   0,   2,   1,   2,   2,  27,   2,  -1,  -1,   2,  -4,  -8, -18,  32, -17,  -3,  -2,   9,   1,   9, -37, -20,  19,  -7, -11,  11, -10,   9,  -4,  -1, -12, -12, -35,  11,  -9, -34,  -2,   6,   4,  20, -35, -23,  16, -32,   6,  18,   2,  37,  16, -37,   5,  -7,  15, -16,   2 },
      {   0,   0,   1,   1,   0,  10,  -1,  -9,  -1,  -1,   2,  -2, -10, -25, -24,  -2,   0,   3,   6,   4, -14, -31,  20,  26,   1,   4,  15,  17,  11, -20,  20, -26,   3,  -4,   6,  18,  32,  23,   8,   6,  -9,   2,  -8,  -2,   2, -27, -53,  12,  -9, -13,  20,   9,  16, -14,  39, -16,  -6, -35, -13,   5,   6,   6, -11,   1 },
      {   1,   0,  -1,   1,   2,   2,   5,  13,   1,   0,  -1,   0,   2,  -1,   0,  45,  -1,  -4,  -3,   1,  -5,  -2, -36, -39,   2,  -1,   1,   9,   3,  12,  -6,   4,  -1, -10,  -5,  14,  -5,  14,  20,  -5,   0, -12, -39,   2,  -4, -23, -30,   9,  15,  26, -25, -31,   1,   3,  36,  -5,   0,  38,  14, -15,  -6,  21, -24,   3 },
      {   1,   0,   1,  -2,  -1,   5,  -1, -14,   2,   1,   1,  -1,   0,  14,  10,  -5,   2,   2,   2,   3,  -9,  28,  66,   8,   0,  -4,   0,   0, -28, -17,  23, -29,   3,  -3,   4,  20, -12, -28,  -9,  15,   8,  -2, -15,  16,  15,   6, -12,   7,  16,  24,  -9,   0,  16,  32,  -7,  -5,  -2,  30,  10,   9,  18,  37, -13,   4 },
      {   0,  -1,   0,  -1,  -4,  -4,   0, -10,   0,   0,   2,   6,   4,  15,   9, -50,   2,   2,   6,  15,  25,  39,  26,   2,  -2,  -3,   0,   7,  28,  46, -10, -10,  -5,  -8,  -9,  -2,   7,  23, -40,  20,  -5,  -9, -20,   0,  -3,  -5,  -6,   2,  -1,  -7, -22, -23,  -4,  -7,  32,  -3,   1,  -1, -11, -27, -23, -19,  -3,   3 },
      {   0,   0,   2,   1,   0,  -3,  10,  10,   0,  -1,   2,   0,  -1,   2,   0,  25,   1,   0,   4,  -2,  -3,   8,  -4,  -6,   2,   2,  11,  -5, -13,  31,  -2, -21,   2,   3,  25,  18, -32,  16,  10,   7,  -7, -14,  13,  61,  -4, -11,  -6,  -6, -11, -29, -40,  33,  29, -11, -16,   3,   7,  -8, -38, -15,  12,  -5,  13,  -1 },
      {   0,   0,   0,   0,   0,   4,   2,  12,   1,   1,   1,  -1,   6,  10,   5,  70,   0,   1,  -1,   1,   3,   7,  22,  57,  -3,  -4,  -7,  -4,   2,  -8, -24, -56,  -1,  -1,  -5,  -8,  10,   9,  -2,  -1,  -2,  -4,  -6, -16,  -8,   0,  22,  -4,   3,   1,  -7, -18, -23, -27,  -6,   0,   2,   8,   7,  -7, -12, -19,   9,  -2 },
      {   0,   0,   1,   0,   3,   3, -14,  -5,   0,   0,   0,   3,   1,   1, -13,  -3,   1,   1,   2,   5,  -3, -17, -10,  16,   2,   5,   7,  17,   3, -19, -16,  18,   0,   5,  11,  34,  46,  18,  -4,  -6,  -7, -14, -12,  10,  54,  51,  -6, -20,   2,  -8, -29, -26,   9,  17, -24,  12,   9,   8,  -4, -19,  -7,   5,  19,   0 },
      {  -1,  -1,  -1,   1,   0,   0,  -6,  -4,  -2,  -3,  -1,  -1,  -6,   0,  -7, -24,  -2,  -4,  -2,   1, -13, -12,  -6, -19,  -2,  -4,  -6,  -2, -19, -38, -19,  -8,  -2,   0,  -1,  13,  -3, -58,   1,  40,  -3,  -8, -12,   6,  18, -49,   8, -23,   3,  -4, -11, -21,   4, -44,   2,   8,   4,  -1,   5, -17,  -7, -25,  18, -10 },
      {   0,   0,   1,   1,   2,   2,   4,  -1,   1,   1,   0,   2,   0,   5,  19,  -8,   1,   2,   3,   3,   1,   4,  25, -25,   1,   2,   5,   8,   8,  15,  64,  -3,  -2,  -7,  -1,   1,  16,   6,  38, -27,  -4, -13, -13, -17,  -5, -11, -19,   9,  -2,  -2, -11, -15, -22, -37, -46,  18,  11,   4,   9,   2,   1,  -8,  37,  -9 },
      {   0,   0,  -1,   0,  -1,  -1,   3,   2,   0,   0,   0,  -1,   1,   0,   4,  21,   1,   1,   3,   2,   0,  10,  12,  36,   0,   0,  -1,   3, -11,  -1,  28,  55,  -1,  -3,  -3,   6, -11, -38, -19, -63,   1,   2,   0,  15,  21, -19,   9,  26,  -6,  -9, -10, -14,  14,  -6,  23, -15,  -2, -10,  -7, -21, -13, -23, -10,   0 },
      {   0,   0,   0,   0,   2,   4,  10,   7,   0,   0,   0,   0,   1,   2,   7,   4,   1,   1,   1,   3,   4,   9,  20,  13,   2,   2,   2,   4,   1,   7,  32,  22,   0,   3,   0,   4,   2,   8,  60,  24,   0,  -4,   1,  -7,   6,   2,  31, -69,   3,  -4,   4,  -3,   5,   6,  19,  14,  -9,  -7, -12,  -8,  -9, -12, -50,   9 },
      {   0,   0,   0,   1,   2,   0,  -1,  -6,   0,   0,   0,   0,  -1,  -1,   0, -14,   0,   0,  -2,   0,  -3,  -7,  -3, -39,   1,   2,   0,   1,   0,  -9,  10, -45,   3,   5,   8,   6,  12,   3,  29,  -9,   0,   3,   7,   6,  22,  17,  54,  47,  -5, -11,  -4, -14,   2,   0,   3, -48,   0,  -4,  -2,  -7,  -8, -20, -22,  39 },
      {   0,   0,   0,   0,   0,   0,  -1,  -3,   0,   0,   0,   0,  -1,   1, -10, -15,  -1,   0,  -1,   1,   0,  -1,  -8, -22,  -2,  -1,  -1,  -1,  -3,  -6,  -7, -52,  -2,  -2,   0,  -2,   0,  -8, -18, -62,   2,  -1,   4,  -2,   7,   2, -10, -17,  -1,   1,   2,   2,   5,   4,  -6,  60,  -9,  -6,  -9,  -6,  -6, -15, -55, -27 },
      {   0,   0,   0,   0,  -1,   2,   4,  -3,   0,   0,   1,   3,  -2,   1,   6,  -5,   0,   0,  -1,   1,   0,  -1,   4, -10,   2,   2,   1,   1,   2,   2,   6, -20,   0,   1,   2,   4,   1,   2,  19, -22,   2,   2,   3,   5,   3,   9,  41, -12,   2,   2,   4,   2,   7,  18,  65,  -1,   3,  -6,   5,  -5,   4,  22,  57, -67 },
      {   0,   0,   0,  -1,  -1,   0,  -3,  -3,   0,   0,   1,   2,   0,   1,  -1,  -7,   0,   0,   0,   2,   0,   2,  -2,  -9,  -1,  -2,  -2,   0,  -3,   1,  -4, -20,  -2,  -1,  -3,  -2,  -7,  -4, -14, -51,  -2,  -2,   0,  -2,  -5,  -9, -13, -72,   2,  -1,   5,   2,   1,   2,  12, -33,   2,   4,   5,   8,   4,   7,  36,  68 },
      {   0,   0,   0,   0,   0,   1,  -1,  -4,   0,   0,   1,   0,   1,   2,   0,  -4,  -1,   0,   1,  -1,   1,   3,  -3,  -4,  -1,  -1,   0,   0,   0,   2,  -5,  -1,  -2,  -2,  -1,  -1,  -2,   1,  -5,  -9,  -3,  -3,  -3,  -3,  -5,   1, -14, -33,  -3,  -6,  -4,  -6,  -6,  -3, -38, -88,   0,  -1,   0,  -2,  -2,   1, -31, -67 },
    },
    {
      {-103,  15,  19,   5,   5,   1,   2,   0, -56,  16,  15,   1,   3,   0,   1,   0,   7,  18,  -3,  -3,  -1,  -1,   0,   0,  24,   5,  -9,  -3,  -2,   0,   0,   0,  17,  -4,  -7,  -1,  -1,   0,   0,   0,   7,  -5,  -2,   1,   0,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
      {  48, -58,  15,   1,   4,  -1,   1,   0, -80, -12,  27,   5,   6,   1,   2,   0, -36,  22,  14,  -3,   2,  -1,   1,  -1,   6,  25,  -3,  -7,  -1,  -1,   0,  -1,  13,  10, -10,  -4,  -1,   0,   0,   0,   8,   1,  -5,   0,   0,   0,   0,   0,   3,   0,  -3,   1,   0,   0,   0,   0,   2,   0,  -1,   1,   0,   0,   0,   0 },
      { -44, -15,  10,  11,   2,   3,   1,   1,  39, -70,  19,   8,   3,   1,   1,   1, -47, -24,  35,   1,   4,   0,   2,   0, -24,  29,  19,  -9,   0,  -2,   0,  -1,   2,  28,  -5,  -9,  -1,  -2,   0,  -1,   6,  12, -10,  -3,   0,   0,   0,   0,   2,   3,  -7,   0,   1,   0,   0,   0,   1,   1,  -3,   1,   0,   0,   0,   0 },
      { -10, -70,  16,  13,   4,   4,   1,   2,  12, -39,  10,   9,  -1,   3,   0,   2,  85,  -7,  -7,  -7,  -8,  -1,  -2,   0,  29,  -9,  -2,  -6,  -3,   0,  -1,   0, -12,  -3,   6,   0,   1,   0,   0,   0, -17,   4,   7,  -1,   1,  -1,   0,   0,  -8,   8,   2,  -1,   1,  -1,   0,   0,  -3,   3,  -2,  -1,   0,  -1,   0,   0 },
      {  -8,  -8, -19,   9,   4,   2,   1,   1, -50,   4, -25,  18,   5,   3,   1,   2,   7, -57,  -6,  25,   2,   4,   1,   2, -31, -39,  43,  16,  -1,   1,   0,   0,  -6,  23,  39,  -8,  -7,  -3,  -1,  -1,   9,  26,   7, -18,  -4,  -2,  -1,  -1,   7,  11,  -9,  -8,   1,   0,   0,   0,   2,   0,  -5,   0,   1,   0,   0,   0 },
      {   4,   1,  65, -27,  -8,  -1,   0,  -2,   4,  36,  54, -36, -10,  -1,  -1,  -1,  19, -14,   7, -22,   1,   1,   1,   0, -49, -19,   5,  -3,  11,   1,   2,   1, -23,  19,  -4,   2,   7,  -2,   0,   0,   1,  16, -16,   1,   3,  -2,   0,  -1,   6,   6, -13,   2,   1,  -1,   0,   0,   2,  -1,  -7,   3,  -1,   0,   1,   0 },
      {  15,  34,  24, -15, -11,   1,  -2,   0,  -9, -45,  37, -12,  -4,   4,   1,   1, -25, -28,   8,   5,   4,   6,   1,   1,  55, -51, -11,  15,   4,   4,  -2,   1,   8, -38,  17,  11,   3,   1,  -1,   0,  -1,  10,  19,  -5,  -1,  -2,  -1,  -1,  -4,  15,   0,  -9,   0,  -2,   1,  -1,  -1,   6,  -6,  -2,   1,  -1,   1,   0 },
      { -23, -35,  -2,   3,   9,   1,   3,   1,   5,  -7,  -4,  -2,   9,  -3,   2,  -1, -36,  41, -24,   1,   4,  -3,   0,   0, -42, -22, -29,  24,   6,   2,   2,   2, -49, -30,  22,  32,   2,   2,   1,   0, -14,  11,  34,   7,  -8,  -1,  -1,  -1,   5,  14,  12, -11,  -6,  -1,  -1,  -1,   6,   7,   0,  -9,  -1,   1,  -1,   0 },
      { -10, -23, -14, -61,  10,  15,   9,   4,  -4, -16, -31, -63,  16,  18,   8,   5,   1, -26, -22,  -9,  21,   9,   2,   1,   4,  15,   1,  24,  13,  -5,  -4,  -2,  25,  12, -14,  28,  -1,  -8,  -5,  -1,  10,  -9,   0,  21,  -6,  -5,  -3,   0,   5,   0,   4,   7,  -7,  -2,   0,   0,  -1,  -6,  -1,   4,  -4,   0,   1,   1 },
      {   0,  28, -45, -31,  -5,   9,   3,   3, -20, -25,   3, -29,   2,   9,   5,   2,  17,  51,  32, -15,  -5,  -2,  -2,  -1,  13,   7,  26,  -8,  -4,  -3,  -3,  -1, -51,  13,  23,  -9,   5,  -5,  -1,   0, -17,  22, -14,  -9,   8,  -3,   1,  -1,  -5,   6, -20,   1,   8,  -1,   1,   0,   2,  -2, -13,   6,   3,   0,   1,   0 },
      {   3,  28, -11,   1,  -7,   4,  -1,   0, -38, -23,   6,  11,  -5,   4,   1,   2,  23, -36,  30,  -2,  -8,  -1,  -1,   2, -39,  17,  -1,  -7,   1,  -1,  -1,   0, -25, -22, -30,  22,   8,   3,  -1,   1, -31, -42,  12,  32,   4,   2,  -1,   0, -11,   1,  38,   3,  -4,  -2,  -2,  -1,   1,  14,   9, -13,  -1,  -2,  -1,   0 },
      { -11, -36,  -1, -20, -12,  11,   5,   1,  17,  44,  -7, -12, -24,  12,   1,   1, -10,  -2,  59,  11, -15,   3,  -2,   0,   5,  14,  39,  -1, -13,   0,  -5,  -1,  21, -49,   9,  -4,  -3,   3,  -3,   2, -12, -12,  30, -13,   5,   1,  -1,   1,   5,  20,  -1, -18,   5,  -2,  -1,   0,   0,   3, -10,  -5,   5,  -1,   1,   1 },
      {  -5, -17, -27,  -4, -26,  11,   3,   3,   7, -10,  29,   0, -23,   9,   2,   1,   5,  33,  12,  19, -18,   1,  -4,  -1, -14, -53,  16,  28,  -4,   2,  -3,   1,  16,  13,  -1,   8,  -2,  -2,  -3,   0,  44, -37, -27,   7,  -2,   2,  -2,   0,  -4, -33,  26,   7,   1,   3,   0,   1,  -3,   3,  19,  -8,  -1,  -1,   0,   0 },
      {  10,  22, -14,  18, -21,  14,   1,   4,  -3, -14,  11,   8, -23,  18,   2,   5,  29,  12,  -1, -15, -16,  10,  -1,   2, -22,  29, -21,   1,   2,   3,  -4,   1,  28,   5,   2,  31,   0,  -2,  -8,   0,  48,  25,  34,  -1, -15,  -7,  -7,  -2,  32,  21, -12, -32, -10,  -3,   0,   0,   3, -14, -27,  -9,   6,   4,   3,   1 },
      {  -7, -18, -44,   9,  15,   6,  -2,   4,   0,   7,   0,  -4,  12,   4,  -3,   1,  -7, -15,  19, -38,   5,   6,   1,  -1,  -8, -47, -19, -54,  15,   7,   5,   1,   5, -28, -28, -16,  29,   5,   5,   1,  25,  16,  -3,  15,  16,  -5,   0,  -1,  21,   8, -10,  19,   0,  -6,  -1,  -1,   3,  -7,   1,  16,  -5,  -3,   0,   0 },
      {   8,  13,  -8, -11,  58, -14,  -4,  -5,   9,   0,  30,   0,  66, -25,  -4,  -5,  26,  13,  29,   7,  24, -22,  -1,  -4, -15,  -4,  11,   3,  -7, -12,   5,   1,  22, -10,   8,   0, -23,  -1,   5,   1,  21, -10,  18,  -7, -20,   4,   1,   1,   3,  -1,   8, -14,  -7,   4,   1,   0,   0,   3,  -1,  -8,   2,   2,   0,   0 },
      {   1,   3,  -7,  26,   9,   8,  -2,   1,   2,   4,  17, -22,   3,   6,   1,  -1, -18,   2, -33, -64, -12,  10,   2,   1,   2,  -3,  34, -28,   0,   9,   0,   2,  20,  25,  42,  -6,   2,   0,  -3,  -1, -14, -35,  30,   1,   2,  -2,  -3,   1, -23,   1,  27, -11,   4,  -4,  -1,   0,   1,  11,  -5,  -9,   5,  -3,   0,  -1 },
      {   0,  -3, -16,  18,   0,   8,   3,   3,  11,  22,  27,   1,  10,   3,   2,  -1, -10,   3, -17,  31,  11,   4,  -1,   2,  32, -10,  33,   3,  -2,  -8,  -3,   0, -25,  32, -48,   0,   2,  -7,  -1,   0,  -4,  -4,   6,  36,  -3,   1,  -4,  -1,  13,  53,  12,  -9, -19,  -5,  -3,  -2,  15,   0, -30, -11,  -7,   3,   1,   1 },
      {  -7,  -7, -45,  -5, -13,  -8,  -1,   0,  -2,   0,  36,  -6,  -5, -13,   1,  -2,   3,  -2, -18,  16,   7,  -2,   1,   0, -28,  -3, -31,  16,  12,   4,   2,   1,  39,   3, -20, -11,   1,   4,  -1,   0, -41,  -7,  -5, -31,  10,   2,   1,   0, -41,  27, -24, -12,  23,   1,   3,  -1, -15,   8, -22,  12,  13,   0,   2,  -1 },
      {   4,  10,  25,  -7,  10,  10,  -2,   2,  -6, -33, -14, -12,   8,  10,   1,   3,   2,  13, -44,  12, -19,   8,  -1,   2, -26,  -8,  22, -13, -36,   6,  -4,   3,  -7, -36, -26, -33, -15,  11,   0,   3,  15, -37,  -7, -15,  15,  13,   0,   2,  23,  12, -18,  -9,  22,   0,  -1,   1,  12, -11, -21,  13,  14,  -3,   0,   0 },
      {  -5,  -3,   3, -45, -22, -42,   0,  -4,  -4, -12, -23,  -9, -17, -41,   1,  -1,   3,   8,  -3,   6,  -9, -10,   7,  -3,  -1,  -7, -10, -30,  -3,  11,  12,  -1,   5,  23, -10, -31,   6,  17,   9,   0,  20,  17,  19,  -6,   2,   9,   1,  -1,   1,  20,  41, -20,  -3,   2,  -4,  -1,   4,  27,   5, -33,   3,   0,  -2,   0 },
      {  -1,  -3,  -8,  22, -38, -36,  -8,  -2,  -4, -12,   8, -16, -10, -39,  -3,  -3,  10,   4, -13,  -2,  46,  -6,   3,   0,   0,  17,  39,  11,  50,   2,   2,  -3,  -2, -30,  10,  -6,  19,   1,   2,  -2,   4, -24,   5,  -3,   1,   0,   3,   2,  31,  -9, -16,   5,  -5,   0,   1,  -1,  18, -16,   5,   4,  -7,   2,   0,  -1 },
      {   6,  14,  29,   6, -12,  31,  -5,   5,   0,  -9, -21,   9,  13,  40,  -4,   1,  12,  36,   7,  14,  45,  16,  -7,  -3, -20,  -6,   5,  -2,  32,  -6,  -7,  -1,  26,  -2,   2, -33,  12, -16,  -4,   2,  -2,  10,   6,  -7,  10, -17,   2,   1, -13,  28,  26,  16,   0, -11,   2,   1,  -3,  29,  12,  -1, -13,  -9,   0,  -2 },
      {  -4,  -7,  -5,   0, -19,  21,  -5,   4,  -1,   9,   7, -11,   5,  20,  -6,  -1,  -4, -16,  -1,  15,  35,  11,  -7,  -1,  10,   3, -15, -28,  14,  -9,  -3,   0, -42,   7, -17, -22, -10,  -6,   4,   2,  19, -11,  29, -25, -23,   1,  -1,   2, -17, -40,   6, -43,   8,   9,   3,   1, -28, -21,   7,  -3,  29,   8,   3,   2 },
      {  -2,   0, -26, -17, -11,  10,  -2,  -1,   4,   5,  48,  -6,   0,  14,  -5,  -3,   3, -12, -26,  20,  -1,  15,  -1,  -2,  -4,  26, -10,  -4, -29,  10,   2,   3,   1,  -7,  21, -33, -23,  10,   3,   2, -15,  22,  23,   2,   2,   3,   1,   0,  34,  -3,  21,  34,  -9,  -4,  -5,   0,   9, -13,  45,  10, -23,  -5,  -4,   1 },
      {   1,  -4, -14,  25,   0,  18,  -2,   6,   1,   9,   2, -23,   4,   8,  -6,  -1,   0, -22,   6, -26,   8,  -6,  -3,  -2,  11,  12, -27,  30,  -1,  -5,   0,   0, -15, -10,   2, -17, -14,   0,   2,   0,   6,  -3, -48, -33,   2,  11,   6,   2,  34,   9,  17, -20,   8,   7,  -1,   1,  32,  50,   0, -36,   6,   0,  -3,  -1 },
      {  -2,  -4,  -3, -55,   8,   0,  11,  -6,   9,   7,  12,  61,   3,  14,   2,  -2,  -1,  -6,  -3,   5,  13,  14,  -3,   1,   8,   4, -20, -15,  23,  10,  -6,   1,  -7,  16,  31,   2,  17,  -3,  -8,  -1,  -2, -40,  -4,  -3,  12,  -6,  -5,   1,  23, -10,  -6,  -5,  10,  -7,  -2,   1,  42,   4, -31,  -2,   6,  -4,   0,  -2 },
      {  -3,  -4,   1, -11, -37,  -4,  -7, -12,  -3,  -7, -13,  14,  -5,  10,  -5,  -5,  -4,   6,   6, -30,  23, -10,  -1,  -3, -11, -27,  -8, -13, -20, -24,   6,   6,  18,  -4,   4, -15, -50, -19,  11,   4, -36,  -5, -17,  19, -35,   3,  10,   1,  10,  -1,  -6,   3, -33,  15,   2,   0,   4, -24, -28, -18,  -7,  14,   0,   3 },
      {   3,   5,   1,  -5,  21,   7,  51,  13,   1,  -3,   9,  16,  -5,   0,  46,   6,  -1,   3, -13, -11,  -1, -14,   9,  -3,  -4,  -1,  11,  14,  13,  -9, -12,  -6,  12, -14,  -5, -26,  24,  -6, -14,  -5, -24,  21, -15,   3,  19,   3,  -5,   1,  24, -15,  22, -22, -10,   8,  -7,   2, -36, -31,   4, -31,  14,  13,   3,   3 },
      {   0,   0,   1, -15,   2,  39, -37,  20,  -1,  -4,   2,   1,  -3, -12, -47,  13,  -5,  -2, -15,   9, -25, -62, -36,   7,   0,   0,   3,  -4,  12, -36,  -2,   3,  11,   6,   5,  -2,  30,  -6,  17,  -1, -12,   2,   8,   1,  16,   0,  13,  -3,  13,  -7,  -2,  -7,  -4,  -3,   5,  -3,   2, -11,   3, -11,   0,   2,   3,  -2 },
      {   2,   1,   3,   9, -27,  18,  49,  -2,   1,   2,  -1,  -9,   9,  -9,  50, -10,  -1,  -8,  -3,  10,  -5, -46,   7,  -5,   0,  -2, -13, -28,  -5, -45,  -4,  -3,  -4,   5,   8,  11,  -6, -17,  -3,   3,  17, -16,  18, -18,   4,   5,   1,   3, -13,  14, -13,  17,  10,   6,   4,  -1,  28,   9,  15,  13, -22,   3,   0,  -2 },
      {  -4,  -4, -11, -22,  -2,  -6,  -3, -36,   0,   4,  16,  28, -16,   2,  -9, -16,  -3, -10, -29, -19,  15, -14,  -5,   2,   2,  23,  17,   1, -13, -27,  -4,   4, -17, -30,  13,   0,  -3, -19,   2,   5,  42,   5, -18,  19,   8,  -7,   5,   2, -15,  20,   0,   4,   5,  -3,   5,   0, -54,  34,  -5,   7,   2,  -6,   4,  -1 },
      {   0,   0,   2,  -5,  -1,  -9,  11,  79,   0,   5,   1,  15, -14, -17,   0,  51,  -1,  -4,   6, -20,  20,  -6, -14,   7,   1,   4,  -7,  15,  -5,  -2, -17, -10, -11,   0,   7, -12, -26,   9,  -3,  -7,  18,  -4,   6,   7,  -7,  14,   9,  -1,  -9,  13,  20,  21,   7,   0,   3,  -4,  -8,  -2, -19,  34,   5,  -9,  -1,  -3 },
      {  -1,  -4,   4,  21, -34,   1,  -1,   6,   1,   5,   0, -26,  36,  -2,   5,   0,   3,  -5,   7,  28,  15,  -6,   3,   3,  -4,   8, -19,  -5, -30,   5,  -1,   2,   6,  -3,  35,  -8,  16,  16,  -2,  -1,   7,   6,  -4,  44,  36,   4, -11,  -7,  -5, -29,   1,  11,  10, -10,  -6,  -2, -14,   6, -53, -22,   9,  -3,   3,  -1 },
      {   1,   0,   2,  12,  25,  -9, -13, -40,  -3,  -3,  -9, -15, -29,   3,  -1, -17,   1,  -3,  21,  -7,  -5, -12,  15,   3,   3,  -2, -24,  38,   9, -26,   8,   5,  -9,  16,   3, -36,  -1, -14,  -2,   5,   9, -10,  26,  -3,  16,  -3,  -5,   2,   8,   6,  36,  25,  19,  -4,  -3,  -2,   5, -36, -31,  25,  11,   0,   0,   2 },
      {   0,   0,  -4,   5,  34,  -1,  10,  10,  -3,  -3,   4, -21, -26,  -4,   1,   1,   0,  -7,  -5,  24, -23,  11,  -8,  -3,  -6,  11, -14,  -8,  17,   7,  -8,  -3, -10, -18,  15, -36,   7,  -9,  -2,   1,  16, -14,  -8,  -3, -24, -20,   9,   2, -32,  -9, -14,   6, -52,  -7,  14,   1,   4,   3, -30, -22, -46,   4,   9,   1 },
      {  -3,  -3,   3,  -5, -26,  -9, -13,  -7,   1,   4,  -6,  10,  42,   2,  -8,   0,  -3,  -5,   9, -20, -15,  10, -14,   8,   4,   1, -10,  14,   4,   7, -18,  -2, -15,   8,   1,   4,  -1,  16, -19,  -1,  15, -24,  -3, -35,  30,  15,  -5,   1,  -7,  34,   1,   9, -10,   2,   2,  -1, -32, -44,   8, -29, -45,  11,   7,   1 },
      {   2,   4,   2,  -3,  10,  40,  13, -19,  -2,  -2,  -3,   9, -21,  10,   2, -16,   2,   3,   3, -10,  26, -27,   1,  -9,   1,  -1,   3,  15,   7,  23,  17,  -5,  -1,   9,  -1,  -9,   4,  73,  19,  -8,   5,  -5,  12,   8,  -6,  52,  -3,   0, -10,   0, -18,   3, -19,  11,  -8,   2,  -6,   6,  -7,   5, -15,  -7,  -6,   1 },
      {   0,  -1,   2, -12, -16,  39, -20,  27,   1,   4,   3,  17,  20, -18,  14,  -6,  -1,  -3,  -5,  -4, -22,   3,  58,  -7,   4,  10,   9,   7,  18,  12,  66,  -5,  -6, -13,  -1,   0,  -6, -13,  29,  -7,   9,  -1,  -7, -12,  -7, -16,  -4,  -5,  -8,   5,  10,  16,  -2,  -8, -10,  -3,  -9,  -8, -14,  -4,   4,   8,  -3,   3 },
      {   0,   0,  -4,  -4,  23, -10, -13,  19,   0,   0,   7,   6, -28,  15,  -8,   5,   2,   2,  -8,  -4,  28,  15,  15,   4,  -3,   2,  10,  -2,  -8, -23,  15,   2,   2, -10,  -7,  12, -13,  -9,   2,   4,  14,  20,  10,   7,  45,  20,  -7,  -4, -40,  -8, -13,   5,  21,  25,  -6,  -4,  39, -18,   9, -48, -12,  23,  -3,   2 },
      {  -1,  -3,  -1,  -4, -13, -11,  -6,   2,   1,   2,   1,   9,   7,  13,   6,   7,   1,  -4,  -1, -20,  22, -19,  18,   4,  -3,   3,   0,  25, -50,   0,   7,   3,  -2,  -4,  -7, -13,  39,   7,   5,   0,   5,   1,  -8,  -6,   6, -30,   1,  -5, -28,  -4,   7, -47, -30, -30,   5,   3,  34, -19,  15,  28, -15, -13,   6,   1 },
      {   1,   2,   1,   8,  23, -10, -11,  14,  -1,  -2,  -4, -14, -30,  10,   4,  23,   1,   5,  10,  19,  25,  -7,  36,  21,   1,  -8, -19, -18, -15,  -5,  23,  10,  -1,  22,  21,   9,  -1,  -9,   9,  -6, -11, -38,  -4,   1,  16,  -4,  -3,  -9,  38,  19, -23, -20,  -2,  -7,  -3,  -5, -44,   9,  22,  -2, -23,  -5,   2,  -3 },
      {   1,   2,   2,   3,   2,  -3,  30,  16,  -2,  -2,  -5,   0,  -5,  -6, -11, -28,   0,  -3,   8,  -8,   4,   3, -30, -48,   1,  -3, -10,  18,  -9,  15,  10, -21,  -4,  12,   3, -11, -14,  -7,  34,   5,   7, -19,  12,  22,  19, -32,  22,  10,  -1,  19, -31,   2,   9, -31,   1,   2,  -4, -16,  26, -27,  26,  -5,  -5,  -1 },
      {   0,   0,  -2,  -6,  -5,  -6,  -8,   8,   0,  -1,   1,  13,   3,  -3,   7,  -1,  -1,   2,  -2, -22,  -3,   8,   7, -10,   2, -11,  10,  24, -11, -20,  -5,  -9,  -3,  11, -40,  -5,  -5,  -6,  -6,  -1, -12,   5,  48, -21,  12,  -9,  -3,   5,  22, -49, -25,  14,   1, -19,   2,   3, -15,  40, -22, -16, -34, -32,   4,  -3 },
      {  -1,   0,  -1,   2,  -7, -46,   4,   9,   2,   2,   5,  -1,  22,  51,   9,   7,   0,   1,  -2,   4, -27,  22,   2, -10,  -1,   4,   0,   9,  11, -37,  19,  -7,   3,   1,   7,  -3,  30,   7,  44,   2,   2,  -4,   5,   0,  -3,  35,  43,  -4,   4,   1,  -2,   1, -15,  20,  16,  -7,  -5,   9,  -3,   9,  12,  -2,  -8,  -2 },
      {   1,   1,  -2,   6,   5, -12,  35, -10,   0,  -1,   3,  -6,  -9,  13, -15,   2,  -1,   2,  -7,   9,  10,  -7, -18,  15,   1,   1,  12, -23,  12,  -3,  18,  13,   2,  -8,  -7,  37, -26,  12,  28,  -3,  -4,  15, -19, -35,  12,   0,  13,  -8,  -2, -10,  34,   3,   5, -52,  -5,  -2,  -2, -23, -24,  -9, -12, -39,  -5,  -3 },
      {   0,  -1,   0,  -2,  -7,   8,  10, -25,   2,   2,   0,   4,  12,  -8,  12,  38,   1,   1,   4,  -8, -17,  10,   4,  63,   2,   0,  -3,  16,  24, -14,   1,  30,  -2,   0,  -3, -27, -23,  14,   1,  -3,   6,  -1,  11,  23,   3,  -4,   4, -15, -12,   1, -32,   3, -13, -30,   5,  -9,  12,   5,  25, -21,  33, -13,  -4,  -3 },
      {   1,   0,   2,  -2,   0,  34,  -4,  -2,  -1,  -1,  -4,   2,   0, -46,  12, -23,   2,   2,   7,  -1,   0,  39,   5,   9,  -1,   0,  -6,   3,   3,   4, -17,  13,   5,   3,   6,   0,  -7, -26,  20,   8,  -1,   8,  -3,  21,  23,  34,  45,   4,   1,  -5,   9, -40,   9,   7,  27,  -6,   2, -10,   0,  19, -30, -27,   1, -11 },
      {  -1,  -2,  -2,  -1,   7,   0, -52,   4,   0,   1,   2,  -2,  -9,  -1,  31, -13,  -1,   0,  -3,   9,  -3,   4,  18, -27,  -2,  -3,  -4, -17,  14, -33, -29, -46,   1,  -4,  10,   8, -21,  20, -22, -22,   0,   1, -15,   1,  11,  17,  -2,  -3,  -5,   6,   1,  -5, -26, -38,  -2,   6,   5,  -8,   9,  -2,  27, -27,  -8,   5 },
      {   0,   0,   0,   1,  -8, -13,  12,  -9,   0,   1,  -1,   0,  13,  16, -20,  22,  -1,  -2,   0,  -1, -25, -11,  -2, -20,   0,   1,  -3,   5,  29,  -5,  16, -28,   0,  -8,   4, -22, -19, -14,  -3, -20,   0,   5,  -7,  37, -11,  11, -37, -11,  -6,  -1,  -6, -41,  38,   3, -26,   1,   4,  -5,   8,  10, -42, -30,  -4,   2 },
      {   1,   1,   2,   0,  -4,  -1,  30,  27,  -2,  -3,  -4,  -2,   3, -16, -42, -32,   0,   2,   2,  -1,  -7,  26,  -1, -10,  -1,  -1,  -3,   1,   6, -37,  23,  29,   1,   2,   5,  -1,  -5,   1, -25,  23,  -3,  -4,  -3,  -4,  25,  20, -43,   2,   1,  -2,  -2, -12, -46,  -7, -18,   2,  -3,  12,   5,  28,  14,   1,   0,   7 },
      {   1,   1,   0,   3,  -4,  14,  31, -36,  -1,  -1,   1,  -2,   2, -12, -41,  39,   1,   2,   0,   2,   5,   5,  18, -10,  -1,  -2,  -1,  -3, -10,  11,  10, -50,  -1,   1,  -3,   7,  -6, -20, -13, -42,   0,  -1,   8, -12,  33,   2,  17, -17,   1,  -5,   4,   6, -32,  20,  26,  -2,   4,   6,  -3,  10,  17,   5,  12,   4 },
      {  -1,  -1,  -1,  -1,  -3, -16,   3,  22,   0,   1,   0,   0,   5,  30, -18, -26,   0,  -1,  -1,   1,  -5, -46,  34,   2,   2,   1,   1,   0,   9,  44,  -4,  21,  -3,  -1,  -4,  -5,  -4, -35, -35,   9,   3,   3,   7,   5,  -4,  34,   9,   4,  -4,   3,  -8,  10,  -6,  -3,  36,  -1,   2,  -7,   5, -15,  17, -42,   9,   3 },
      {   1,   1,   0,  -1,   2,  12,  18,   3,  -1,  -1,  -1,   1,  -4, -19, -29, -11,   1,   0,   1,  -2,  10,  27,  17,  -5,  -2,  -1,  -2,   1, -27, -30,  -3,   5,   1,  -1,   3,  -4,  45,   1, -25, -29,  -1,  -5,  -3,  -5, -55,  20,  15, -34,  -2,  11,  -5,  18,  30, -20,  -8, -15,  -2, -14,   5, -26,   0, -14, -10,   1 },
      {   0,   0,  -1,   1,  -2,   5,  13,  -5,   0,   1,   2,   0,   4,   0, -28,  25,   0,   1,  -1,   4,  -5,   2,  46, -21,   1,  -2,   2,  -4,  11, -25,  -3,   5,   1,   2,   4,   3,  -7,  46, -22,  34,  -2,   3,  -6,  14,  -6, -18,  30,  52,   3,   1,   4,  -7,  24, -10,  25,  34,  -1,  -4,   1,  -5, -10,  12,  19,   8 },
      {   0,   0,  -1,   1,   2,   9,  -7, -23,   0,   1,   2,  -1,  -2, -13,  11,  52,   0,  -1,   0,   1,   9,   7, -15, -13,  -1,   1,  -3,   0, -15,  10,  17, -14,   0,   0,   4,   3,  25, -38,   3,  50,   1,  -2,  -1, -10, -12,  45, -16,  43,  -6,   4,  -5,  11, -11, -21, -17,  15,  -1,  -6,   1, -12,   9, -18, -21,   6 },
      {   0,   0,  -1,   0,  -2,   0,  11,  -1,   0,   0,   1,  -1,   5,   3, -29,   3,   1,  -1,   0,   1,  -5, -10,  49,   0,  -1,   0,  -1,   0,   4,   8, -65,   5,  -1,   1,  -1,   1,  -1,  -7,  32,   8,   4,  -2,   3,  -7,   7, -10,  14,  11,  -7,  -1,  -5,   2, -15,  13, -68, -14,  -1,   0,  -3,   0,   3,  -6, -38,  -7 },
      {   0,   0,   0,   1,   1,   6,  -7,  -6,   0,   1,   1,  -1,  -1,  -5,  10,  15,  -1,  -1,  -2,   3,   3,   1, -11, -25,   2,   2,   1,  -3,   0,  -8,  26,  11,  -2,   0,   1,   4,  -1,  22, -35,  28,   1,  -3,  -5,  -2,   5, -39,  13, -11,  -2,   6,   1,   0,  -9,  56,  -3, -36,  -2, -13,  -3,  -9,   7, -59, -24, -32 },
      {   0,   0,   0,  -1,   1,   1,  -5, -12,   0,   0,   0,   2,   0,   3,   4,  27,   0,  -1,  -2,  -2,  -2,  -7,  11, -57,   1,   1,   4,   0,   8,  10, -17,  25,  -1,   0,  -3,   1, -10,  -1,   3,  33,  -1,   2,   0,  -1,   3,  17,  -6, -49,   2,  -6,   2,  -3,   3, -35,  15, -51,  -1,   6,   1,   6,  -5,  40,  16, -28 },
      {   0,   0,   0,   0,   1,   4,  -2,  -4,   0,   1,   0,   1,  -2,  -9,   4,  18,   0,  -1,   0,   0,   2,  10,   9, -30,   0,   2,   0,   0,  -2,  -8, -30,  37,   0,   0,   2,   4,   6,   3,  49, -11,   0,  -3,  -4,  -6,  -2,  -7, -44, -17,  -1,   9,   1,  13,  -2,  29,  19,  29,   3,  -8,   3, -12,  11, -43,  46,  30 },
      {   0,   0,   0,   1,  -1,  -1, -11,  -6,   0,   1,   0,  -1,   1,   1,  15,  17,   0,  -1,   1,   1,   0,  -6, -15, -36,   1,   1,   0,   0,   5,   9,  13,  55,  -3,   1,  -3,   0,  -6,  -9, -29, -28,   4,  -1,   6,   1,  14,   2,  47, -21,  -2,  -3,  -3,  -5,  -4,  -2, -31,  40,  -1,   6,  -3,   7,  -5,  10, -26,  52 },
      {   0,   0,   0,  -1,  -1,  -2,  -1,  -2,   0,   0,   0,  -1,   0,   1,  -1,   9,   0,  -1,   0,   1,  -1,   1,   1, -22,   0,   1,   0,   0,  -3,  -4,  -2,  40,  -1,  -2,  -1,  -1,   2,  -3,   7, -61,   1,   2,   0,   0,   2,   2, -28,  47,  -1,   0,   2,  -1,  -2,  -6,  36,   6,  -1,  -3,  -1,  -2,   0,  11, -52, -56 },
      {   0,   0,   0,   0,   0,  -2,  -5,  -3,   0,   0,  -1,  -1,   0,   3,   7,   3,   0,   0,   1,   1,   0,   2,  -8,  -7,   0,   0,  -1,  -1,  -1,   0,  12,  17,   0,  -1,   1,   1,   1,  -7, -10, -33,   2,   1,   1,   0,  -1,  12,  24,  39,  -4,   1,  -3,   2,  -6,  -2, -56, -14,   5,   0,   4,   0,   8,  -6,  83, -43 },
      {   0,   0,   0,  -1,   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   0,   3,   3,   0,   0,   0,   1,   0,   2,  -3,  -5,   0,   0,   0,  -1,  -1,  -1,   1,   9,   1,   0,   1,   1,   0,   4,   1, -21,  -1,   1,  -2,   1,   1,  -3,  -5,  50,   0,  -2,   1,  -3,   3,  -3,  11, -87,   1,   2,   0,   1,  -2,  -2,   0,  74 },
    },
    {
      {  73, -35,  -1,  -4,   0,  -1,   0,  -1, -77,  27,   9,   3,   2,   1,   1,   0,  31,  12, -18,  -1,  -4,   0,  -1,   0,  -2, -27,   8,   6,   1,   2,   0,   1,  -1,  14,  10,  -9,   0,  -1,   0,  -1,  -3,  -1, -14,   1,   2,   2,   2,   0,   2,   0,   5,   6,  -5,  -3,   0,   0,   0,  -1,   0,  -4,   5,   0,  -1,   0 },
      {   9,   4,  -1,   0,  -1,  -1,   0,  -1, -26,   3,   0,   1,   1,   1,   1,   0, -23,  51,  -7,   1,  -2,   0,  -1,  -1,  32,   7, -51,  11,  -6,   3,   0,   1,  14, -61,  12,  29,  -7,   4,  -2,   3,   7,  -7,  49, -19,  -4,  -1,  -5,   1,  -4,   6,  -5, -22,  12,   4,   0,   0,  -2,   2,  -4,   5,  -5,   1,   1,   0 },
      {  35, -13,   2,  -3,   1,  -2,   0,  -1, -12, -13,   8,  -3,   3,  -1,   1,   0, -34,   8,  18,  -6,   4,  -2,   1,  -1, -19,  66, -11, -14,   3,  -2,   1,  -3,   6,  15, -64,  14,  -1,   2,   5,   0,   8, -32,  -1,  36,  -9,  -2,  -2,   2,   8,  -7,  25,  -5,  -5,   3,  -4,   1,   0,  -2,   1,  -7,   1,   1,   1,   0 },
      {  59,   2,   0,  -1,  -1,  -2,   0,  -1,  19, -56,   5,  -3,   2,   0,   1,   1, -44,  12,  30,  -4,   4,  -1,   1,  -1,   4,   6,  -5,  -6,   0,  -2,  -1,   2,  -8,   4,  29,  -7,   3,   2,  -3,  -3,  -9,  31, -26, -28,   3,   0,  10,  -2,  -9,   2, -31,  31,   8, -13,   4,  -1,   4,   1,   8,  14,  -6,  -1,  -3,   0 },
      { -51,  13,   5,  -1,   2,  -1,   0,   0, -29,  10,   6,  -2,   3,  -2,   1,  -1,  -8,  52,  -7,  -6,   0,  -2,   1,  -1,  17,   4, -29,   3,   0,   0,  -3,   1,   6,  11,  -4,  -1,   0,   7,   0,  -5,  -9,  -3, -55,   8,  -4,   0,  16,   0,   4,  -7,   6,  60,  -6, -24,  -2,  -2,   9,  -5,  20, -10,   0,   5,  -5,   1 },
      { -18,  -6,   2,   0,   1,  -1,   0,   0, -19,  12,   2,  -1,  -1,   0,   1,   0,   2,   3,   3,   1,   1,   0,   1,   1,  -5,  17,   1,  -8,  -1,  -4,   0,   1,   2,  11, -14,   3,   4,   3,  -9,  -2,   3,   4,  -2,  -4, -11,  22,   5,  -3, -15,  -1, -38, -11, -14, -13,  21,  -1,  -2,  25,  -7,  80,  40, -51,  -1, -11 },
      {   9,   3,   0,   0,   0,   0,   0,   0,  15,  -2,  -2,  -1,   0,   0,   1,   0,   2,  -6,   0,   0,   0,   0,   1,   0,  -6,  -5,  -3,   4,  -1,   0,  -7,   1,   6,  -1,   7,  12,  -5,  17,   1,  -9,  -3,  -8,   4,  -9, -23, -12,  30,   0,   3,   3,  13, -16,  22, -54,  -6,   0,   2,  -8,   2, -26,  92,  15, -13,   3 },
      {  50,  41, -13,   0,  -4,   1,  -1,  -1,  54,  25, -24,  -1,  -4,   1,  -1,  -2,  21,   5, -26,  -1,  -2,   2,   1,  -2,   3, -19, -27,   5,  -2,   0,   1,   4,  -9, -19, -12,  15,   4,   7, -11,   1, -10, -14,  -6,  29,  -7,  24,   3,  -3,  -3,  -4,  12,  24, -36,  -7,  10,  -2,   1,   6,  10,   8,  -3, -20,   0,  -3 },
      { -13,  -5,  -4,   2,  -2,   1,   1,   1, -14, -39,  11,   8,   3,   3,   2,   2, -29,  13,   4,   1,  -1,   0,   1,  -2,  20, -20,   8,   7,   0,  -1,   6,   8, -12,  -3,  14, -15,   5, -11, -18,   5,   5,  19,   3,  34,   8,  34, -11,  -8,   7,  -5,  19, -25, -66,  10,  15,   1,  -1,  -3,  -7, -32,  33, -14,   0,   1 },
      { -15, -19,  -3,  11,   2,   3,   0,   1, -24, -22, -16,   8,  -2,   2,   0,   0,   4, -29,  12,  14,   5,   3,   4,   4, -46, -11, -16,   2,   1,  -7, -11,  11,  -1,   8,  17,  55,   3,  26, -24,  -9,   8,   3,  19,  -5, -40,  27,  15,  -3,   2,   6,   4,  14, -18, -18,   6,  -2,   5,   2,  13,  -8, -35,   3,  -1,   1 },
      {  -2, -10, -21,   3,   1,   4,   1,   3,  -3,  -9, -14,  13,   4,   4,   2,   2, -61, -22, -43,  12,  -2,   3,  -4,   4,  -8, -15,  16,  48,   2,  14,  -5,  -8,  24,   2,  18,  -7, -16,   3,  17,   0,  -4, -33,  -4,   4,   4, -19,  -1,   6,   2,   8,  31,  16,   2,  -2,  -6,  -3,   0,  21,   3,  27,  -1, -19,   0,  -5 },
      {  -5, -18,  -8,   3,   4,   2,   1,   1, -25,  -6, -30,   2,   1,   2,   1,   2,   0, -42,   7,  17,   8,   6,   2,  -5,  35,   7,  15, -15,   2,  -1,   9,  -1, -22, -55, -28,  -9,  13,  -3,   6,   4, -24,  20,  16,  32,  12, -14,   6,  -7,   1,  28,  -7,  36,   7, -27,   3,  -6,   4,  -6,  11,   1,   3,  -3,  -5,   2 },
      {  -6,  -3, -11,  -1,   5,   1,   5,   3,  -2, -47,   5,  -3,   9,  -4,   5,  -2,  19,  16,  -9,  -9,   8,  -1,   4,  -2, -10, -43, -28,  -2,   7,   2,   5,  -5, -51,  11, -12,  26,  14,  -7,  28,  -2,  10,  14,   1,  32,  -4, -39,  -7,   5,   3, -17,  15,  -8,  24,  11, -16,   4,   0,  11,   3,  29,   8,  -6,   1,  -4 },
      {   7,  41,  -5,  14,  -8,   2,  -2,   6, -12,  22,   9,  11, -13,  -4,  -3,  -4, -24,  13,  -5,  16,  -8,   3,  -2,  10,  24,   3,  31,   4, -17,   5, -31,  -5, -29,  20,  -7,   0,  -7,  37,  15, -16,  -1,  34,   4,  21, -42, -20,  19,  -4,   2,  12,  -9, -22,   0,  -3,   1,  -1,  -6,  -4, -19,  -5, -23,  -4,  10,   1 },
      {  -2, -42,  20,  -2,   3,   0,   6,   1,  28, -28,  51,  -4,   5,  -4,   0,  -9,  32,   9,  11, -12,   1,   1,   7,   4,  40, -11,  13,  -1,  -5,  -4, -18,   7,  16, -10,  -4, -12,  -3,  26,  -2, -11, -13, -32,   4,   4, -12,  -2,  23,   2,  -2,  20,  25,  -1,  -6, -19,   1,  -2,  -4,  16, -10,  10, -27, -18,   2,  -4 },
      {  -4, -53,  26, -13,   5,  -5,   5,  -8,  31,  39,  31,  -9,   8,  -6,   8,  -2, -23,   5, -25, -17,  -1,  -8,  -2,   5, -38,   5, -19,  18,   7,  11,  -1,  -6, -17,  -7,   3,  10, -11,  -1,  11,  -1, -11,  41,   9,  20,   8, -13,   1, -10,   2,  17, -23,   7, -13,  -8,  12,   0,  -1, -13,  -5, -13,  -1,  -7,   0,   3 },
      {   4,   6,  18,  21,  -8,  -6, -13,  -8,   2,  -2,   6,  40,  -2,   8,  -1,  17,  36,   0,  42,  11, -19,  -9, -32,   9,   0,  26,   1,  19,   0,  22,  -3, -33,   4,  -8,  20,  13, -24, -15,  34,   4,   8,  10,   5,  -3,  -4, -24, -10,   3,  10,  -9,  13,  16, -31,  -1,  -1,  -2,   6,   6,  18,   1,   8, -19,  -2,  -4 },
      {  -5, -27, -30, -32, -14,   4,  -4,  -4,  -1, -20, -39, -21,  -7,  14,   6,  14,  13, -16, -18, -31, -15,   3, -14,  31,  19,  10, -19, -10,   0,  28, -27, -17,  11,  -5,  -7, -10,  -7,  31,  21, -11,   9,   7, -18, -10,  -4,   2,  12,  -3,   3, -27, -17, -15, -12,  13,   9,   4,  -1, -11,   1, -12,  -3,  -8,   3,   2 },
      {  -5,  22,  -8, -12, -15,  -2,   2,   3, -16, -17,  -4,  -4,  -8,   0,   2,   3,  12,  -3,   9, -28, -14,  -8,   3,  -1, -27,   4, -16,  -5,   6,   1,   6,  -4,  -6,  -4,   4,  -6, -16,   8,   3,  18, -23,  -6,  -4,  -7, -14,  -1, -24,   0,  -6,  51,  10,  25,   9,   8,   8, -18, -25,   8, -61, -18,  13, -20,  32,  -3 },
      {   2, -43,  -7,   2,   5,  -6,   1,   0,  23,   3, -15,  -5,  -5, -10,  -9,   1,   3,   8, -15,  22,   2,  -2,  -5, -16,  40,   7,  10,   3,   2, -10,  16,  -7,  14,  16,   3,  21,  13,  -3,   2,  23,   9,   8,  -9,  -1, -61,   0, -33,  10,   1,   4, -12,  21, -11,  24, -17, -13,  -4,  -5, -12,   7,  19,  28,  19,   0 },
      {   0,  -5, -18,  -5, -14,  -7,   0,  12,   4, -17,  -6, -17, -23, -15,  -8,   0,  12,  13, -16,   5, -18,  -3,  -5, -36,   9,   5,  10,   8, -11,   7,  36,   9,  10,  23,  -8,   4, -36, -23,  -7,  11,  19,   9,  10,   5, -10,  -8,   3, -22,  -5,   2,  -2,  -5,  30, -21,  40,   2,  10,  -4,  19, -24, -18, -41, -14,   1 },
      {  -3,   6,  31, -19, -30, -16,  -1,   4, -10,   5,   6, -42, -48, -21,  -6,  13, -23, -12,  -2, -16, -31,  -9, -12,  -9,  -9, -16,   5, -29, -26,  -5,  19, -11, -14, -12,   9,  -5,  10,  -9,  19,   4,  -4, -11,   9,  -9, -14,   0,   7,  18,   6,  -3,  15,   0, -27, -12, -23,   3,   5,   9,  15,  17,  -5,  13,  -7,  -3 },
      {   2,   6,  33,  19,  19,  32,  18,  19,  -5,   4,  25,   2,  10,  30,   5,   8, -15, -17,   0,  -4,   0,  24,   0, -22,  -5, -22,   1, -15,   5,  25,  21, -11, -14, -26, -12,  -8,  -8,  20,  11,  20,  -4, -22, -17, -14, -38,   6,  -8,  -3,  -9, -26, -12,   1,  11,  21,  23,  -2,   0, -10,  18, -15,   5, -16,   4,   2 },
      {   0,   8, -12,   4,  11,  28,  15,  17,   0, -13,  -7,  -9,  10,  28,   0,   3,  15,  15, -13,   3,  10,  32,  -4, -32,  -8,  11,  -3,   5,   8,  22,  34, -17,   7,  15,  -6,  -5,  -5, -10,  24,  -8,   1,   3,   7,  -9,   9,  -2,  24,  12,   1,  25, -15, -14, -44, -34, -14,   6,  -5,  -4, -26,  12, -16,  32,  -3,   2 },
      {   8,  39,  47, -23,  22,  -5,  10,  -8, -16,  -2,  -2, -40,  24,  -9,   9,  -4,   1, -30,  -7, -16,  25,  -2,   1,  -3,  27,   7,   1,   6,  27,   2,   0, -10,  36,  17,  19,  24,  25,   1,   6,  -2,  22,  16,   6,  12,  -1, -15,  -6,  -1,   4,   4,   4,   3,  -4,  -4,   2,  -3,   2,   5,  -1, -10,  -3, -24,   0,  -3 },
      {   1, -14, -20,  -9,  -7,  25,  20,  26,   5,  31, -18,   7, -12,  21,   8, -28,  -7,   1,  41, -18,  -4,   8,  50,  -8,  -3,  -2, -12, -11,  -3, -26,   8,  18,  26,  -2,  17, -10,   0,   2,   9,   5,   3,  24,  -1,   6, -10, -37,   3,   8,  12, -10,  15,  -4, -13,   6, -12,   3,   2,  16,   9,  -4,   0, -22,  -2,  -5 },
      {  -2, -12, -34,   9,  -2,   5, -21,  -6,  10,  25,   8,   8,  -2,  10, -27,  13, -22,  17,  22, -10,  -1,   6, -38, -10,  -3, -29,   6, -35,  25,   4,   2, -30,  17,   8,   5,   9,  57,   1,   9,  -4,  17,  -8,   9,   7,  10,   5,   5,   0,   3,  10,   3,   4,  16, -15,   9,  -4,   0,   0,  -9, -14,  -6, -23,   6,  -3 },
      {  -2,  -5,  10,  -9,   5, -10, -37,  15,  -6,  -1, -13,   1,  -3, -29, -67, -40,   3,  -8,  11,  -7,  20,   5, -12, -47,  -6,  11,  -6,  26,  13,   4,  -1,   1,  -9, -16,   9,   0,  -4,  20,   5,  -7, -23,   1, -18,   3,   6,  -2,  16,   7,  -5, -22,  -1, -13,  -4,  13,  -1,   9, -10,   5,  -6,  -2,   1,  -2,   8,   1 },
      {   1,  17, -29,  16, -13,  -1,  -6,  29,  -8,   0,  44,  17,  -5,  -7,  -6,   7,  -3, -35,   1, -49, -13, -26,  -4, -10,  16, -14,  -9,   4,  27,   2,  -4,  12,  28,  -3, -10,  12, -22,   1, -12,   8,  11,   6,  -1,  21,  -1,  -4,  -5,  -8,   2,   3, -24,   5,  -7,  -1,  -5,   4,   5, -16,   7,  23,   3,  35,  -8,   6 },
      {   4, -10,  38,  19,   3,  18,  15,  81,   7,  -7, -20,   7, -18,  -6, -17,  43,   8,  11, -17,   2,  -6,  -2,  -9,   9,  -6,  15,  -3,   2,  -4,  -4, -12,   4,   4,  -2,   9,  -6,  17,   8, -17, -12,  16,  11,  -3,  12,  20,   5,   1, -10,  21,   1,   9,   5,  18,   3,  -4,  -5,  -3,   3, -11,   6,   5,  -3,   9,  -3 },
      {  -3,   5, -25, -29, -14,  18,   1,  18,   0,   7,  17, -31,   3,  18,  -2,   4,  10,  20,   5, -11,  10,  11,  -1,  -5,  -1,  28,  31,  12,  24,  19, -10,  -5, -28,   9,  18,   8,   1,  25,  -9,   3, -21,   6,  28,  -3,   3,  19, -17,  -1,  -5,  20,  21,  13,   7,  22,  -5,  -2,  19,  13,  45,   7,  14,   9, -12,  -2 },
      {  -2,  -5, -24,  -6,  38, -13,  32,  35,   1,  19, -16, -15,  35, -30,  13,  30, -10,  -1,  29,   3,  -4, -34, -23,  -3,  12, -11,  -1,  -5,   5,  -7,  -6,  -2, -13,   9,  -1,  16, -17, -21,   5,  -9, -41, -14,  -6, -12,  -6,  -4,  12,   5, -18,  -1,   1, -17, -12,   5,   6,   6,  -3,  13,   0, -12,  -6,  -3,  -5,  -5 },
      {  -1,  13, -13,   4,   2, -10,   0,  24,  -4, -24,  10, -20,  19, -11, -14,  -9,  14,  22, -14,  22,   4, -15,   5,   9, -41,  -8,  18,  -3, -16, -31, -19, -11,  35, -19,  -9, -19,  16,  15,  22,   8,  -2,  -9,  16,   5,  -4, -25, -10,  13, -15,   5, -32,   7, -24,   7,  10,  -8,  11, -27,  23, -10,   7,  -9,   1,   7 },
      {   0,   2,  12, -14,   9,  -6,  12, -21,   1,   0, -14, -18, -17,  10,   8,  16,   1,  14,  40,  15,   3,  17, -28, -12, -18, -29,  -3,  10, -23,  25,  -3,  10,  48,  -2,  -5,  -9, -11,  14, -10,  -4, -17,  15,  -1,  45,  -1,  -4,   2, -22, -16,   3,   5,  -2,  14,  21,  11,   0,   0,   2,  -3,  22,   6,  30,  -1,   0 },
      {   0,   1, -11,  -2, -15,  -9,  17,  18,  -1,   2,  34,   9,  -8, -16,  -5, -11,   0, -31,   0,  40, -18,  23,  18,   2,  19,   0, -40,  -3, -30,  17,  12, -12,  14,  29,  -7,  23,  16,  15,  21, -20, -16,   6,   2,  -6,  35,  16,   1,   4, -20,  10,  -4,   8,   2,  25,  -3,   0,   3,   3,   3, -14,  15,  -8,   2,   3 },
      {  -1,   3,  -7,  25,   0,  -4,  13, -22,  -4,  -9,   5,  18, -23,  10,  12,  29,  10,  -7, -20, -25,  21, -12,  -6, -15,  -2,  23, -17,  23, -11, -36,  17, -26,  13,   6,   9, -15,  33,  10,   1,   9, -54,  16,  -6,   7, -16,   9,   2,   7,   2,   0,   4, -27,  14,   4,  -3,  12,   7,   0,  12,  -4, -10,  -9, -16,   0 },
      {   0,   9, -36, -37,  51, -16,  37,  -9,  -1,   3,  27,  15, -11, -32,  -6,  14,   9,  -8,   4,   2, -12,  20,  -3,  -2, -14,  21,   2,  16, -10,   4,   8,   2, -11, -26,  13, -20,  16,  16,  -3,  20,  41,   8,  -7,   3,  -6,   8,   0,   2,   3, -16,   7,   5,   3, -18,  -8,  -6,  -5,   6,  -4,   2,  -6,   1,   7,   1 },
      {  -3,   4,   9, -66, -39, -12, -16,  22,   1,   1,   9,  30,   7,  24,   8,  29,  -1,   1,   5,  24,   7, -11,  20,   7,   1,   2,   9,  14,  16,  -6,  24,   2,  -2,  -4,  11,  11,  21,  -5,   0,  -1, -14, -17,  -3,  14,  -9, -10,   6,  -8, -25, -29,  -7,  -6,   2, -17,  12,   0, -12, -19, -15,   5, -11,   7,   7,   6 },
      {   0,   3,  23, -35,  14,  -2,  -3,  -1,   3, -10, -20,  52,  30,   4,   0, -17,  -2,   7, -21,   4, -44, -33,  -8, -27,  -3,  -4,  -5, -33,   3,  17,   1,   6,  10,  12, -10, -13,   7,  30,  -5,  12,  -6,  16,   6,  -3,  -2,   5, -13,   8,   6,  22,   4, -11,   3,  -1, -12,   7,  19,  15,  12,   2,  -4,   3, -21,  -1 },
      {  -1,  -5,   8,  34,  -5, -21,  33,  -5,   0,  -7, -26,  11, -25, -28,  29, -25,   3,  22,   1, -18, -29,   3,  19,   2,  -4,   9,  30,   1,  43,  28,   9,   9,   2,  -1,  16,  16,  24,  16,  15,   4, -11, -19,   4,  13,  15,   2,   8,  -2, -27, -11, -12,  -2,  -2,   1,   4,  -1, -14, -13,   2,  -1,   0,   5,   0,   0 },
      {   0,  -6,  -7,  20, -30, -19, -12,   8,   3,  11,  -4, -16,  20,  -9,  38,  13,   1,   2,  15,  13,  33, -44,  16, -29,  -5,  -7,   5,  12,  -8,  11,  16,  -8,  -1, -16,  -7, -10,  -4,  42,  25,  11,  39,  12, -12,  -3,  11,  30,  10,   5,   7,   0,   7,   2,   7,   3,  -4,  -3,   1,  19, -14,  -3,  -2,   3, -10,  -1 },
      {   1,   4, -12,   3,  12,  -6, -22,   5,   1,   1,   3,  -2,  11, -23,  -4,   6,   2,   4,   6,   0,  21,  -3,  17,  26,  -5,   3,   3, -18, -18,  52,  31,  29,  11,   1,  20,   3,   6,  -2,   7,   4, -30,  -8,  -3,   5, -13,   5,  -1,   3,  62,   8,  -1, -11,   5,   3,  -6,  -2,   5, -40,   3,  16,  -3, -18,   0,  -3 },
      {   0,   0,  -1,  22,   9, -14, -37,   9,   0,   2,   5,  -6,  -2,   1, -15,  20,   1,  -1,  -9,  -4,   6,   9,   5,  39,  -5,   6,  -8,   1,   6,  27,  19,  40,   7,  -3, -14, -10,  21,  -2,   3,  20,   1,  16, -11, -11, -14, -22,  -2,  -5, -50,  10,   7, -13, -12,  -5,  -3,  -4,   8,  52,   7, -20,  -4,  10, -11,  -5 },
      {  -1,   1, -15,   8, -10,  46,  -5, -13,  -2,  -4,  15,  -2,  20, -18,   8,  -5,   7,  -4, -26,   3,  -2, -28,   7, -20, -10,  28,   8, -36, -27,  25,  -8, -11,  11, -18,  32,   4,  12, -25, -18, -19,   0,   8,  -3,  17, -10, -12,  21,  -9, -24, -18,  14,  10,   1,  13,   8,   0, -30,  15,  -2,   1,   4,  12,   6,  -6 },
      {  -2,  -5,  -7,  22,  -9, -29,   8,   5,  -1,   6,   0, -18,  26,  35,   6, -10,  -1, -11,   2, -11, -12,  -4, -31, -21,   8,  14, -10,   8, -40,   8,   2,  29, -19,  18,  11,  -2,  27,  28, -26,   8,  -2, -12,  11,  -9,  17, -37, -26,  -1,   3, -20,   5,   5, -13, -19,  10,  -7, -22, -13,  -2,   7,  -1,   0,   8,   4 },
      {   1,   4,  -2,  -1,  14,  -4,   5,  10,  -1,  -4,   6, -17,   4,  43,  -9, -36,   4,  -3,   7,  -4, -39, -31, -12,  21,  -2,   5,   9,  39,  -9,   1,  18,  10,  -1, -13,  -9,   5,  30, -27,   9, -11,   7,   7,  -3,   4, -16,  22,  42,  19,  13,   5,  -1,  10,  18,  27,  13,  -7,   8,   1, -15,   0,   3,   4, -16,   1 },
      {   0,  -1,   4,   8,  37, -20, -27,   0,   0,   3, -10,   5,  21,  14, -15,  10,  -3,   2,  -6, -29, -12,   7,  20,  12,   8,   6,   5, -22, -18, -36,  30,  -8,  -3,  16,  25,  -3, -27,  18,  39, -11,   5,   0,  25,  16,   7,  22,   9, -23, -18,  -7,  17,  14,   6,   5,  19,  -7, -14,  -7,   8,  13,   4,  18,  11,   1 },
      {  -1,   1,  -3,   1, -25, -66,  12,  10,   2,   1,   1,  18,  24,   6,  12,   6,   1,   0, -17, -13,   9,  47,  -2, -23,  -9,   8,   7, -22,   3,  -4, -15,   1,   2, -20,   5,   4,   3, -23,  -6, -18,  10,   5, -18, -10, -17, -11,  28, -10,  12,  22,  13,  -4,  -3,  32,  13,   0,  22,   2,   2,   5,   5,   2,   2,   1 },
      {   0,  -4,  -3,  -3,  42, -18, -50,  28,   1,   0,  -1,  -6, -25,  16,  39, -13,   1,  -2,   7,  -7,  -9,   7,  28, -20,   0,   2,   0,   5, -25,  16, -34, -35,  -5,  -4,  13,  10,  10, -18,  -9,  -2,   1, -18, -17,  16,  -1,   4, -27,  -2,  -8,  11, -15,  -9,   4, -13,  -5,   4,   7,   2,  -7,  -5,  -4,   0, -19,   5 },
      {  -1,   1,  -1,   7,  -7, -19,  -2,   9,   0,  -3,   2,  -4,  25,  10,   2,   1,  -1,   6,  -3, -15, -12,  14,  -1,  -3,  -7,  -7,  19,   6,  -6,  11,  -1,  -2,  21,  -8, -22,   2, -12, -16, -10, -14, -16,  33,  14,   1,  -9,  32,  -8,  11, -20, -32,   3,   7,  16, -20, -71,  -4, -15,  -6,   3,  -8,  -4, -31,  26,   8 },
      {  -1,  -3,   7,  -5, -22,   5,  32,  -4,   0,   2, -16,  15,  24,  -3, -36,  -3,  -2,   6,   9, -36,  17,   6,   5,  26,   3,  -6,  24,   4, -48,   6,  18, -27,  -6,  -7,   8,  32, -10,   4, -18,  10,  13, -16, -23,  22,   9,  -1, -13,  16, -15,  16, -22, -10,   6,  -5,  -6,  -4,  11,  12,  -3,  -9,  -3,   2, -25,   3 },
      {   0,   1,  -1,   4,   4,  23, -28, -26,   0,  -3,   3, -18,  19,  -5,   0,  47,   3,   3,   3,   5, -53,  22,  17, -21,  -1,   2,  19,  15,   1, -26,   3,  17,   1, -11,  -5,  37,   4,  19,  -7,  14,  -3,  -3, -24,  -2,  21, -17,  18,  -6,  14,  10, -12,  -6,  -6,  21, -21,   5,  -8,  -6,  -5,  -2,   4, -16, -13,   2 },
      {   0,  -2,   3,  -9,  -2,  18,  -1,  13,   0,   2,  -1,   9,  -5,  -2,  10, -17,  -1,  -1,  -3,  -6,  15,   6, -25,   0,   2,   1,   6,  -3, -18,  -9, -26,  22,   5,  -4,  -5,  27,   6, -13,  33,  47,   5,  21, -25, -16,  10,  19,   4,  -2, -30,  19,  40, -16,   1,  -7,   3,  -9, -22, -47,   7,  17,  -4,  -5, -12,  -9 },
      {   0,   0,  -6,   2,  -1,  10,   2,   1,  -1,  -1,  10,  -4,  -5,   9,  11, -10,   2,  -4, -13,   8,   7,   5, -33,  -3,  -2,  16,   7, -23,  -3, -16,  -3,  36,  -1,  -7,  30,  14, -14,   6,  26,   7, -10, -12,   1,  22,   7,   8,  -6,  -7,  -8, -27, -14,  10,   4,  -3, -20,  11,  68,   9, -46, -10,   1,  -9,  -6, -12 },
      {   0,  -1,   1,  -2,  10,   8,  -3,  -5,   0,   3,  -7,  11,  -8, -18,  -5,  14,  -3,   3,   4, -21,  10,  -3,  14,  -8,   0,  -7,   9,  11, -10,  11,   5,  -3,  -4,   3, -23,  -6,  11,   7, -12, -34,   6,   8,  25, -34,  -9, -15,  -8,  -6, -27, -14,  26,  32, -15,  18,   2,  10,  26, -40, -40,   8,   5, -11, -49,  -2 },
      {   0,  -1,  -4,  -3,  20, -20,   1,  -2,   1,  -2,   9,   2, -34,  31,   1,   5,   1,   2, -13,  17,  21, -15, -17,   5,   0,   5,   7, -29,  15,  -4,  20,  14,   1, -11,  11,  14, -29,  18, -17, -28,  14,   3, -14,   9,   9, -13,  -1,  50, -13,  16,  -1,   5,  -2,  17, -12,  25, -40, -12,   7,  -6,   5, -19, -20,  -1 },
      {  -1,  -3,   6,   5,  -7, -11,   1,  -3,   0,   3,  -5,  -2,   9,   3, -15,  19,  -1,  -6,   4,   7,   2,  10,  13,  -8,   6,   3, -27,  -6,   8,  12, -31,   0, -10,  26,  20, -30,  -9, -13,  -8,  51,   5, -13,  31,  26,  -8,   4,  40,  34, -13,  -6, -14,  14,   6,  14, -14, -21, -10, -13,  -9,  -3,  -1, -10, -36,  -5 },
      {   0,   1,  -4,   7, -14,  -9,  15,   5,  -1,   0,   4, -16,  14,   3, -16, -17,   2,  -3,   1,  20, -10,   6,  -6,  13,  -6,   5, -13,  -5,  19, -10,   1, -10,   0,  -9,  12, -15,   1,   0,  12, -12,  -1,  -6, -13,   4, -20,  10, -28, -62,  10,   2,  -1,  -5,   3,  -4, -25,   7, -40,   2,  -8,   2,  -7,  -2, -67, -11 },
      {   0,  -5,  -1,  18,  -7,   5,  -1,  -2,   0,   7,  -5, -24,  19,  -4,  -6,  -5,   1,  -9,  11,  19, -26,  -5,   4,  12,   0,  10, -25,   1,  22,  -1,  16, -21,  -9,  -7,  23, -23,  -6,   7, -23,  -8,  15,   2, -26,  16,  -9,  -3,  -6,  10, -35,  22,  28, -27,   7,  -8,  -7,  11,  42, -42,   1,  22,  -6,  -7,  23,  15 },
      {   0,   0,  -3,   1,  18, -14,  24, -15,  -1,  -1,   3,  -2, -27,  36, -32,   7,  -1,   1,  -7,   8,  14, -43,  24, -15,   0,  -1,   3, -12,   2,  15,  -9,  -4,  -4,  -1,  -1,  15,  -4,  -2,  -2,   5,  -1,  -6, -10, -13,  15,  -9,  17, -48,  -6,   3,   3, -13, -13,   7, -29, -53,  19,  -7,   3,   3,   1, -11,  14,  14 },
      {   0,   2,  -5,  11,   5, -20,  12,   2,   0,  -2,   8, -17,   0,  15, -26,   0,   0,   4, -10,  17,  -8,  -2,  27,  13,  -2,  -2,  13, -19,   9,   8, -15, -29,  13,  -5,  -6,  22,  -9,   3, -11,  45, -22,  26,  -2, -10,  12,  -1,  -2,   9,  16, -27,  29,   7,   2, -14,  30,  47,   2,  11, -29,  12,  -3,  11,  -6,  19 },
      {   0,   1,  -4,   0,   3,   0, -17,  12,   0,  -2,   7,  -2,  -1, -13,  23, -20,  -1,   1,  -6,   8,  -5,  17, -17,  16,  -2,  -2,   3, -11,  14, -14,  25, -29,   8,  -5,  -5,   8, -16,  17, -30,   3, -17,  18,  -5,  -4,  11, -13,   6,  16,  11, -25,  13, -11,  -1,   0,  10, -82,   5,   4, -12,   6,  -6,   8, -23, -15 },
      {   0,   0,   1,  -4,   1,  11,  -7,   7,   0,   1,  -3,   5,  -3, -13,  19, -11,   0,   2,   1,  -3,   9,   5, -23,  10,   1,   1,  -2,  -4,  -1,  -6,  10,   1,  -2,   5,   5,  -3,   1,   2,  -1,  10,   1,  -6,   6,  -1,  -3,  -4,  14, -13,   2,   2,   0,   6,  -4,  11,  -9,   0,  -8,  13,  -3,   5,   2, -12, -11, 114 },
      {   0,  -1,   0,   7,  -7,   3,  12, -15,   0,   2,   1,  -5,   4,   3, -21,  24,   0,  -2,   0,   1, -10,   3,  19, -21,  -1,   4,  -1,  -2,   9,   3, -25,  32,   4,  -7,   6,   1,   0, -12,  33, -38,  -5,   8, -12,   7,  -4,  16, -42,  42,   3,  -7,   7, -11,   8, -20,  31, -42,   0,   6,  -4,   4,  -6,   7, -21,  35 },
    },
  },
  { // 13
    {
      { 118, -40,  -7,  -5,  -2,  -2,  -1,  -1, -15,  -4,   3,   1,   1,   0,   0,   0, -18,   1,   3,   0,   1,   0,   0,   0, -13,   5,   2,   0,   1,   0,   0,   0,  -4,   4,  -2,   0,   0,   0,   0,   0,  -4,  -1,   1,   1,   0,   0,   0,   0,  -2,   3,   2,  -1,   0,   0,   0,   0,  -1,   2,  -2,  -1,   0,   0,   0,   0 },
      {  -6, -17,   7,   0,   2,   0,   1,   0,-102,  47,   7,   3,   2,   1,   1,   0,  45,  15, -16,   0,  -4,   0,  -1,   0,  21, -18,  -5,   1,  -1,   1,   0,   0,   8,  -6,   1,   1,   0,   0,   0,   0,   1,  -5,   2,   0,   0,   0,   0,   0,   2,   2,   1,  -1,   0,   0,   0,   0,   1,  -1,  -2,   0,   0,   0,   0,   0 },
      { -14,  -1,   5,   0,   1,   0,   0,   0, -30, -22,  16,  -1,   4,   0,   1,   0, -79,  49,  12,  -1,   2,  -1,   1,   0,  40,  39, -26,  -2,  -5,  -1,  -2,  -1,  26, -23, -15,   5,  -1,   1,   0,   1,   3, -15,  10,   4,   0,   1,   0,   0,   0,   6,   4,  -4,  -1,  -1,   0,   0,   1,   3,  -6,  -2,   1,   0,   0,   0 },
      {  22,  18, -16,  -2,  -5,   1,  -2,   0,  34,   3, -11,   0,  -4,   1,  -1,   1,  18,  25, -19,   2,  -5,   1,  -2,   0,  59, -44, -21,   7,  -5,   2,  -2,   1, -15, -58,  26,   6,   3,   2,   1,   1, -23,   5,  31,  -5,   2,  -2,   0,  -1, -11,  19,   1,  -9,   1,  -2,   1,  -1,  -3,   5,  -9,  -1,   1,   0,   1,   0 },
      { -12, -25,   8,   3,   1,   2,   0,   0, -19, -41,  19,   1,   3,   2,   2,   1, -25,  21,   3,   0,   0,   0,  -1,   0,  13, -39,  11,   5,   0,   1,  -1,   1, -39,  24,  45, -10,   4,  -3,   1,  -1,   7,  62, -21, -20,   0,  -5,   0,  -2,   5,  -7, -30,   8,   2,   2,   1,   0,   4,  -8,   5,   8,  -2,   1,  -1,   0 },
      { -26, -68,  19,   8,   6,   3,   3,   1,  -5, -36,   8,   7,   3,   2,   0,   1,  14, -45,  18,   4,   2,   0,   0,   1,  -5,  10,  12,  -5,   0,  -1,   0,  -1,   5, -25,   5,   5,   0,   1,   0,   1, -24,   3,  43,  -2,   2,  -1,   0,   0,  -3,  44,   3, -21,   0,  -5,  -1,  -2,  -2,  11, -22,  -5,   3,   0,   1,   0 },
      {   3, -12,  17,   3,   9,  -3,   2,  -2,  12, -52,  42,  -4,  13,  -8,   1,  -5,  53,  37,   4, -17,   2,  -6,   3,  -3,  19,  -3, -14,  -9,   3,   0,   3,  -1, -32,  20, -29,   4,   1,   1,   1,   0,   6, -32, -15,  23,  -3,   6,   0,   2,   1, -11,  32,   0,  -3,   1,  -2,   0,   3,  10,   7, -14,   0,  -2,  -1,   1 },
      {  -9, -68,  24,   1,   3,   0,   1,   1,  39,  33, -12,  -9,  -5,  -2,   0,   0,  21,  31,  -9,  -6,  -4,  -2,  -1,  -1,   1,  32, -29,   0,  -3,   0,  -1,   0,  13, -23,  -6,   8,  -1,   2,  -1,   1,  -3,  18, -11,  -9,   0,  -3,   1,  -1,   4, -31, -29,  19,   2,   4,   2,   1,  -4, -21,  25,  15,  -4,   1,  -2,   0 },
      {  19,  53,  30, -10,   6,  -6,   1,  -4,  -8,   8,  32, -10,   8,  -9,  -1,  -6,  25,  -6,  40, -19,   7,  -5,   5,  -3,  -2,  42, -14, -20,   4,  -3,   4,  -2, -21, -15, -10,   7,   2,   1,   1,   1, -17,  37,   7, -13,   3,  -2,   0,  -1,  13,  17, -42, -10,   3,  -1,   0,   0,   3, -13, -12,  11,   2,   1,   1,   1 },
      { -11,  -3,  10,  -5,   5,  -2,   1,  -1,   9,  24,  22,  -8,   3,  -6,   1,  -3, -16,  49, -30,  -9,   1,  -1,   1,  -1, -76, -18, -18,  13,   3,   4,   2,   2, -25,  -2,   5,  16,  -2,   3,   0,   1,  17,  -4,  30,   1,  -3,   1,  -2,   0,   8,  26,   4, -28,   1,  -4,  -1,  -1,   0,   0, -28,  -1,   6,   0,   2,   0 },
      {  -1,  14,  -2,   1,   9,  -2,   0,  -2, -28, -30, -21,   6,   8,  -2,   2,  -3,  -5, -24,   8,   1,   9,   0,   6,  -1, -37,  -4, -41,  12,   4,   2,   5,   1, -21, -48,  13,  32,  -1,   5,  -1,   3,   1,  11,  14,   8,  -4,   2,  -1,   1,  17, -33,  23,  19,  -5,   5,  -3,   1,   5,   6,  52, -16,  -8,  -3,  -4,   0 },
      {   4,  -2,  79,  -4,   8,  -7,  -1,  -4,  16,  48,  25, -11,   2,  -6,  -2,  -4, -37, -29,   4,  -4,   3,  -1,   2,   0,  21, -28,   5,   3,   2,   2,   1,   0, -10,   0,   5,  -5,   2,  -1,   0,   0, -14,  -1, -10,   9,   1,   3,   2,   0,   8,  -6,  31,   6,  -6,   1,  -2,  -1,   4,  19,  12, -29,  -2,  -2,  -1,   0 },
      { -15, -17, -32,  -4,  -1,   5,   1,   0,  -6,  26, -35,  -4,   4,   4,   0,   1, -21,   4,   0,   3,   5,   1,   1,   1,   9,  14,   7, -18,   4,  -5,   1,  -1, -77,   6, -38,  -1,   9,   1,   5,  -1, -43,  10,  -8,  34,   0,   3,   1,   2,  11,  20,   6,  17,  -7,  -1,  -3,   0,   8,  -4,  -6,  -3,  -1,   2,  -1,   1 },
      {  -1,  -2, -11,  -2,  -6,   1,  -1,  -1,   4,  13, -22,  21,  -9,   5,  -2,   2,  19,  33,  42,  -3,  -6,  -4,  -4,  -2, -10,  31,  -5,   2,  -6,   0,  -2,   0,   7,   4,  43, -15,  -2,  -3,  -1,   0,  -6,  27, -19,  -9,   2,  -3,   0,  -1,  12,  -2,  35,  -2,  -9,  -2,  -3,   0, -12,  35, -27, -60,  13,  -2,   4,  -1 },
      {   4,  -7,  -2, -15,   5,  -2,   2,   0,   5,   7, -20, -15,  -1,   0,   1,  -1,   2, -13,   0,  13,   0,   1,   2,   1,  38,  14,  14,  -3,  -3,  -3,   1,  -1, -19,  15,  -7,  19,  -4,   2,  -1,   1,  71,  19,  41,  10, -17,   1,  -8,   0,  35, -19,   3, -43,  -1,  -1,  -2,   0,  -7, -22,   1, -25,   8,   2,   2,   1 },
      {   2,   6,  30, -56,  11,  -6,   4,  -5, -11, -32, -43, -38,  11,   1,   4,   0,   5, -14, -54,  -8,  10,   1,   3,   0,   5,  19, -26,  17,   3,   0,   0,   0,   6,  15,  11,   1,   0,  -1,  -1,   0, -14,   5, -20,   4,   3,   1,   0,  -1,  -9,   3,  -2,   4,   2,  -2,  -1,   0, -11,   7, -36,  -4,  14,   0,   4,   0 },
      { -12, -16, -49, -71,   8,   0,   2,  -2,   3,  15,  33, -59,   7,  -1,   1,  -2,  -2,  -2,  34, -21,   5,   1,   4,  -1,  -2, -12,  13,  -5,   4,  -1,   1,   0,  12,  -6,  10,   3,   1,   0,   2,   0,   0,  -3,   7,  -5,   2,  -1,  -2,   0, -19,  -1,   3,   7,   5,  -1,   0,   1,   3,  14,  23,  -7,  -9,  -2,  -2,   0 },
      {   3,  14,  30, -14,   2,  -5,  -2,  -2,  -4,  -7, -16,  -7,   0,  -1,  -1,   0,  12,  37, -13,  40,  -5,   4,  -4,   1,  -4,  39,  55,  12,  -8,  -2,  -4,  -1, -16,   7,  19,  -4,  -7,   0,  -2,   0, -10,  -7,  31, -15,  -3,  -1,   0,   0, -21,  16,  10,   7,  -1,  -1,   0,   0,  23,  16,  46,   9, -24,  -3,  -6,  -1 },
      {   3,   4, -29,  11,  -6,   6,  -4,   0,   8,   0,  45,  -3,   2,  -4,  -1,  -2,   6,  -9, -47,   3,   6,   0,   3,   1,   9,  20,  -7,  23,   1,   0,  -1,   0,  33,  25,   8,  10,  -7,  -1,  -1,   0, -30,  32,  10,  17, -10,   1,  -1,  -1,  46,  26,  11,  18, -15,  -2,  -6,  -2,  21, -17,  17, -25, -12,   0,  -4,   1 },
      {  -1,   4,  18, -19, -69,  11,  -8,   5,  -8, -16,  17, -19, -69,  20,  -3,  12,   4,  -4,  -1,   2, -36,  15,  -7,   5, -10,   3,   3,  -2, -12,  -2,  -6,   0, -10, -20,  -9,   0,   4,   0,   2,   0,  13,   4,  -2,  18,   1,  -7,  -3,  -1,   1,   4,   9,  25,  -2,  -5,   0,  -1,  -7, -18,  -8,  -5,   4,   1,   2,   0 },
      {   0,   1,   6, -34,  22, -10,   3,  -1,   1, -13,  -3,  12,   8,  -4,   0,   1,   3,  17,  12,   9,   7,  -3,  -2,  -3, -11, -15,  44,  15,  -7,   3,  -5,   0,  14, -43,  -2, -36,  -3,   0,   1,   0, -14, -15, -19,  15,   4,   4,   4,   1,  64,  -7,   9,  -5, -11,   2,  -3,   1,  -8, -44, -11,  15,   4,   2,   1,   1 },
      {   6,   4,  19,  14,   6,   6,  -1,   1,   1, -13, -43, -23,  11,  -2,   5,   1,   4,  22,   7, -23,   0,  -1,   0,   2, -10, -34,  12, -45,   8,  -4,   3,   1,  36,  -6, -21,   4,  11,   1,   1,  -1,  23,  15,   0,   7,   5,  -4,  -1,  -2,   8,  40, -11,  29,  -1,  -6,  -3,  -2,  39,  -7,   5, -34, -13,   4,  -4,   1 },
      {   0,  -5,   9, -41,  -1,  -8,   1,  -3,   9,   7, -11,  36,  -5,   7,  -4,   1,   6,   1,  39,  28,  -2,  -1,  -3,  -3,   9, -17, -38,  20,  -1,   2,   2,   1,  12,  29,  -2,  27,  -3,  -1,   1,   1,  18,   8,  -2,  17,  -5,   2,  -1,   1,  12,  30,  13,  16, -16,  -3,  -4,  -1,  33,  12,  -9,  50, -13,  -5,  -2,  -2 },
      {   1,  -1,   3,  20, -12,  12,   0,   4,  -2,   1, -23, -19, -23,   9,   0,   3,  -2,  -2,   1, -79,   1,  -1,   3,   2,   8,  13,   5,   2,  10,   3,  -1,   2,  -4,   8,  31, -10,   1,  -1,  -2,  -1,  -4, -12,   5,  -1,  -3,   4,  -1,   1,  31,  -1,  23, -29, -13,   0,  -1,   0,  22,  16,  10,  50, -16,  -7,  -2,  -2 },
      {  -3,   2,   8, -28, -21, -10,  -1,  -7,  -3,  -9, -15,  22, -15,  -1,  -2,  -3,  -4,  13,   3,  -2,  -8,  -3,   3,  -2,  -8, -24, -16, -27,  -1,   0,   4,   1,  23,  32, -20,  -4,  12,  -2,   0,   0, -53,  17,  21,   1,   8,   2,  -3,   1, -12, -24,   4, -60,   0,  14,  -1,   5,  -4, -23,  25, -14, -17,   3,  -3,   1 },
      {   3,   9,  13,  22,   5,  21,  -2,   0,  -1,   2,  -9, -20,  17,  10,   5,   3,   2,  19,  27,  -9,   5,   2,   3,   2,  -2, -10,  25,  16,   9,  -3,   4,   0,  20,  17,   8,  41,   0,  -5,  -1,  -1, -22,  20,  23,  39, -22,  -5,  -3,  -2, -23, -11,  18,  21, -13,  -7,  -3,   0, -52, -23,  -7,  22,  24,   0,   2,   1 },
      {  -1,  -4,  -1,  25, -20, -19,  -4,  -4,  -4,  -4, -18, -43, -11, -23,   1,  -5,   1,   3,  35,  15,  -7,  -5,   3,   0,   5, -13,  -3,  46,   1,   7,   5,  -3,  -6,  17,   1,  32, -21,   2,  -8,   0, -28, -36,  -6, -30,   8,   7,  -1,   1,  17,  -5, -17,  -5,  13,   6,   1,   2,  31, -24, -13, -20,  12,  -1,   1,   0 },
      {   1,   3,   2, -18,  38,  24,  14,  11,   2,   3,   4,  26,   7,  10,  12,   9,  -7,  -8,  -3, -17, -62,   4,   1,   4,   2,   8,  -5, -31, -25,   2,   1,   1, -11,  17,  40,  21,  -3,   2,  -2,  -1, -13, -37,  12,  -5,  -5,  -4,  -8,  -3,   2,  -7,  -7,  20,   7,  -5,  -3,  -1,  17, -35,  -7, -12,   7,   5,   1,   3 },
      {   4,   4,   7,   8,  -7,  68,   4,  19,   2,   1,   2, -34,  33,  58,  12,   2,   5,   0,   9,  37,   5,  13,  -7,   0,  -5,  -7, -27,  -6, -10,  -3,  -5,   0,  -1,   5, -12, -22,  -3,  -3,  -2,  -1,  -6,  -2,  -1, -31,   6,   6,   3,   2,  19,  -2,  12, -16,  -9,   4,  -2,   1,   2,   0,   2,   6,  -3,  -4,   0,  -1 },
      {   0,   2,  -9,  21,  17, -34,   2,  -6,  -3,  -4,  -3, -40,   4, -34,   9,  -5,  -1,  -4,  -1,  19, -61, -13,  -3,   3,  -5,  -6, -15, -20, -46,   6,   1,   3,   7,   6, -12, -24,  -2,  11,   0,   4,  -2,  24,   8,  -4,   1,   9,   2,   2,  13,   3,  22,   4, -16,   0,  -2,  -3, -21,  21,   3,  29,   8,  -5,   0,  -1 },
      {   3,   4,  -1,   7,  47,  24,  12,  10,  -1,  -2,  -2,  -1, -20,  14,   5,   4,   2,  -1,   6, -13, -37,  -2,   8,   2,  -1,   1,   9,  59, -18,   0,  -3,  -1,   3, -13, -30,   4,  -4, -12,  -2,  -4,   3,  27, -27,  30,  15, -11,  -2,  -3, -21,  11,  -7, -35,  22,   5,   4,   3,  12,  10,   9, -11, -17,   4,  -4,   0 },
      {   1,   1,   2,  -9, -11,  14,   0, -15,   2,  -4,   1,  11,  -5,  13,  12,  -8,   0,   8,  -9,   0,   2,   0,  12,  -2,  -2, -13,  22, -14, -24,  -1,   7,  -2,  11,   0, -16,  19, -28,  -4,  -1,   0, -19,   0,  12,  13, -21,  -6,  -4,   1,  39, -30, -36,   6,  38,  -6,  -1,  -1,   9,  67,  -5,   6,  26, -10,   0,  -1 },
      {   0,  -2,   0,   7, -33,   1,  -7,  10,   0,   4,   7, -10,  39, -14,   7,   4,   3,  -6,   1,  10,   9, -22,   2,   2,  -6,   9,   3,  -5, -32,  -5,   2,   1,   2, -15,  24, -14,  -1,  -3,   4,   1,   3,  11,   3,  50,   3,  -6,  -3,  -2, -30, -37,   6, -11,   2,  11,   0,   2,  58, -11, -35,   7, -10,   4,  -4,   3 },
      {  -1,  -1,   4,  -6, -16,  -2,   2,  -7,  -1,   3, -10,  15,   9,   3,   6,  -4,  -1,  -4,   7, -32,  23, -15,  10,  -2,   0,  -6, -14,  15, -51, -15,   1,  -4,  -8,   5, -11, -33, -58,   9,  -2,   2,   6,   3,   4, -15, -40,   2,   0,   3, -19,  29,  13,  11,  22,  -6,   1,  -1, -10, -28,  20,  -7,   9,   5,  -4,   1 },
      {   0,   0,  -4,  12,  44, -16, -27, -51,  -1,  -1,   5, -10, -30,  37, -20, -19,   2,   1,   4,   3,   8,  41,  -1,   1,  -3,   0,  -9,  -2,   7,  13,  -3,   4,   1,   1,  -5, -29,  -6,   6,   0,   4,  -5,  10,  21,   0, -20,   0,   4,   4, -18, -24,  10,   9,  -2,  -4,   1,   1,  38, -14, -20,   1,  14,  -4,  -2,   1 },
      {   1,   0,   2,   8,  -6,   4,  16, -20,   1,   4,  -4,  -3,  -1,  -4,  18, -19,   3,   2,  12,   6,  17,  22,  21,  -3,   2,   5, -18,   9, -11,  23,  12,   6,   0,  19,  22, -31,  25,   1,   6,   5,   8, -11,  23,  33,  58, -12,  -2,  -3,  14,  12,  -4,  14,  54, -12,  -5,  -4, -19,  -9,  15,   2,  -1,  -3,  -1,   0 },
      {  -2,  -3,   0,  -3, -27, -10, -46, -24,   1,   1,  -3,   9,  56,   8, -34,  -7,   0,  -2,  -2, -19, -28,  57,   3,   3,   1,   3,   9,   4, -21,  23,   9,  -1,  -3,  -6,  -2,  21, -16,  -2,  -3,  -2,   6,  -2, -20,   1,  20,  -8,  -2,   0,   4,  11,  -7, -14, -14,  -2,   0,   5, -14,   3,   6,  -3, -16,  -3,   1,   1 },
      {   0,   2,   4,  -3,  21,  22, -63,  27,  -3,  -4,  -5,  -7,  -9, -16, -62,  26,   0,   0,   7,   0,   8, -30, -37,   0,  -2,  -4, -15,   5, -12,  -1,  -4,  -4,  -2,  15,   5, -18,   4,   4,  11,  -2,  -4, -11,  20,  17,   4,  -5,   1,  -8,  11,  -2, -19,   6,  15,  -7,   5,  -4, -10,   8,   8,  -4,  -1,   1,   6,   0 },
      {  -1,  -4,  -1,  -9,  -7,  43, -34, -31,   0,   1,  -1,   8,   6, -22, -13, -37,  -2,  -1,  -8,   7, -31, -36,  23, -10,   3,   5,   7,  -5,  40, -21,  18,  -1,   2, -11,  -2,   3,  -3,  -1,   5,   3,   6,  14, -12, -18,  11,  10,   0,   2,  -2,   5,  35,  -2,  35,  -2,  -7,   1,  10, -12,   0,  15,  23,  -3,  -4,   2 },
      {   1,   2,   1, -11, -12,  10,  19,  42,   0,  -1,  -3,  14,  22, -37,   0,   7,   0,   0,   3, -17, -18,  37, -24,   5,   3,   0,   2,  20,  35,  40, -13,   4,   1,   3, -30, -23,  -7,  12,  -2,   3,  -4,  22,  26, -18,  -9, -16,   4,   0,  -2, -20,  19,  15,  24, -10,   2,  -1,  17,   3,  -9,   3,  14,  -1,   2,  -3 },
      {   0,   1,  -1,  -3, -11, -26,   6, -18,   1,   0,   3,  12,  25,  44, -11,   9,   0,  -1,  -4, -28,  -9, -37, -10,   3,   2,   0,  14,  31, -20, -22, -18,   2,  -2,   6, -25,   8,  51,   9,  -8,   1,   2,   8,  28, -31,  27,   9,   5,   1,   6, -17,   9,  26,   9,  -3,   4,  -1,   5,   1, -12,  -4,   1,  -6,   2,   0 },
      {   0,   1,  -3,   5,  17, -33, -13,  36,   1,   2,   7,  -4, -13,  20,  -2,  31,   0,   2, -10,   4,  27,   7,  -1,   0,   1,   3,   9, -32, -13,   8,   4,  -9,   2,  -2,  -1,  35, -28, -13,  -3,  -5,  -3,  20, -15, -15,  42,  -4,  -4,   2,   1,   2,  45,  -4,  42,   4,   1,   2,   7, -13,  -3,  16,  17,  -9,   0,  -4 },
      {   0,   0,   0,   9, -11, -43,  12,  -5,   1,   1,   0, -13,  26,  33,  -8,   8,  -1,  -1,   4,  12, -29,  -5, -24,  -3,  -3,   2, -15,  -5,  40, -28, -14,  -1,   2,  -2,  17, -13,   1, -16,   0,  -6,  -4,  -9, -14,  22, -38, -40,   2,  -5,  12,  15,   1, -16,  43,  -4,   8,   0,  -1,  -1,  18,   3,   7,   9,   3,  -1 },
      {   0,   1,  -1,   1,   4,   3,  19, -19,   0,  -2,   0,   1,  -2,  -3,   1, -11,  -2,  -1,  -2,  -7,  -8,  -7, -32,  -8,   1,  -5,  -3,  -2,  23, -38,  -8,  -3,  -5,   5,  -5,  -7, -70, -23,  -3,   2,   3,   1,  26,   6,  54, -24,   2,  -3,   2, -16,  -4,  17, -24,   8,   1,   3, -11,   3, -15,  -2, -29,   0,   3,  -2 },
      {   0,   0,  -2,   1,  -3,  26,  40, -40,  -1,   0,   1,  -3,  -5, -19, -22, -21,  -1,  -4,  -4,  -1,  19,  -4, -71, -15,  -1,   2,   6, -11, -31,  22, -26,   0,   1,  -3,   3,  26,  14,  -3,  -2,   7,  -1,   2, -22,  -2,   2, -18,  -4,  -1,  -4,   9,  11, -15,   1,  -4,  -1,  -1,  10,  -6,  14,   5,  27,   0,   3,   1 },
      {   0,   1,   1,   2,  -5,  -3,  13, -13,   0,   0,   2,  -3,   3,  34,  -6,   7,  -1,  -1,  -1,   5,  -9, -46,  -5,   1,   2,   2,   6,  -8,  26,  65,  -4,   4,   1,  -1,   5,   3, -33,  69,   4,   4,   0,   3,  -7,  15,   5,  18,  -8,  -1,  -3,   2,  -6,  -2,   8,   4,  -6,   2, -11,   0,  -4,  -4, -23,  -3,  -1,   1 },
      {   1,   1,  -1,   3,  -3, -11,  46,  34,  -1,  -2,  -1,  -5,  -3,  20, -65, -36,  -1,   1,  -4,   1,  -4, -10,  16, -27,   1,   1,   4,   1,  -8,  13,  66,   0,   2,  -2,  -1,  -5,   0, -18,  19,  -5,   0,   0,   6,  -5,  -4, -11,   2,   4,  -2,  -4,   0,   7, -13,  -3,   2,   3,   4,   0,  -3,  -1,   3,   2,   4,   3 },
      {   0,   0,   1,   0,  -1,  -7,  11,  22,   1,   1,   1,   3,   2,   5,  -7, -14,   0,   3,   1,  -5,  -4,  34, -24,  -6,   0,  -1,   2,   6,   4, -56,  10,  -6,   5,   2,  10,  -3,  -6,  65,   7,  -2,  -1,   4,  -3,  21,  22,  46,  -5,  -6,  -5,   4, -12,  -9,  -3,   7, -13,   5,  11,   5,  17,   1,  42,   3,  -2,   2 },
      {  -1,  -1,   0,  -5,  -2, -15, -41,  33,   0,   0,  -2,   6,  -1,  22,  39, -46,  -1,  -2,  -1,  -8,  -5, -23,  -4, -44,   1,   1,   2,   4,   3,  27, -19,  -3,  -2,   3,  -1,  -5,   0,  -5, -12,   6,   3,  -4,   6,   7,  16, -30,   6,   6,  -6,   8, -12,  -1, -32,   8,   8,   1,   8,   1,  19,  -4,  45,   5,  -1,   1 },
      {   0,   0,   0,  -5,  -4,   4,  10, -23,   1,   1,   3,   4,   5,   2,  -6,  47,   1,   2,  -1,   0,  -1, -17,  16,  59,   1,   1,  -3,   5,  10,  10,  27,  15,  -1,   0,  -4, -15, -10,  -5,   1,  -3,   3,  -2,   2,   2,  23, -27,  -4,  -7,  -5,   6,  -7,  -9, -35,  -2,   3,  -3,  17,   2,  27,   1,  63,  -1,   5,   3 },
      {   0,   1,   1,   0,  -6,  -1,  20,   4,   0,   0,  -2,   0,   7,   2, -36,   4,   0,   1,   2,  -1, -10,  -2,  24,  -3,  -2,   0,  -4,  -6,   9,  19, -56,  -7,   0,   0,   7,  -8, -12, -52, -23,  -1,   5,   2,   9,  19,   5,  66,  14,  10,  -5,   7, -13,   4,   3,  19,  -7,   1,   2,   0,  10,  -3,  24, -14,  -4,   0 },
      {   0,   0,   0,  -2,  -2,  -5, -12, -10,   1,   0,   0,   4,   3,   9,  35,   6,   0,  -1,   0,  -9,  -6,  -1, -49,   6,   2,   5,   2,   8,   3,  19,  40,  14,   0,  -1,   0,   0,   0, -36,  61,   0,   5,   0,   7,  -1,  -8,  38,  25,  -4,   0,   6,  -4,   5,  10,  48,  -4,  -2,  -3,   0,   1,  -4,   4,   3, -15,  -1 },
      {   0,   0,   0,  -3,   3,  -7,  -6,  28,  -1,  -1,   0,   2,  -4,   8,  -5, -61,   0,   1,   2,   0,   6,  -7, -15,  76,   0,   1,   2,  -5,  -4,  -5,  -5,  64,  -1,  -1,   3,   8,  -5,  -6, -24,  14,  -2,  -1,  -6,   4,  -1,   6,   4,  -1,   1,   2,   0,  -1,   9,  -3,   6,  -7,  -6,  -2,  -3,  -1, -11,  -6,  -1,  -1 },
      {   0,  -1,   0,   1,  -3,  -8,  -1,   5,   0,   0,  -1,   1,   3,   4,  13,  -2,  -1,  -1,  -1,   1,  -4,  -6,  -2,  -4,  -1,  -1,  -1,   1,   2,  -2, -12,   2,   1,  -2,   0,   3,  -9, -12,  51,   1,  -2,   1,  -8,   2,   4,  16,  20, -13,  -8,  -8,  -7, -15, -10, -89,  -6, -14,   6, -10,   7,  -6,  -1, -55,   9,  -9 },
      {   0,   1,   0,   1,   0,   4,   5,  -1,   0,   0,   0,  -3,   1,  -3, -14,  26,   1,   0,   0,   0,   0,   8,  25, -35,   1,   0,   3,  -2,  -3, -13, -19,  67,   1,   4,   0,  10,  -6,  13,  28,  68,  -1,  -1,  -4,  -2,   4, -10,  34,  16,   2,   0,   2,  -3,   0,   8,  12, -12,   2,   2,   0,   0,   4,  25,   1, -10 },
      {   0,   0,   0,  -3,   0,   7,   6,  -1,   0,   0,  -1,   0,  -3,  -4,  -5, -13,   0,   1,   0,   1,   0,   1,  11,  26,   0,   0,   1,  -6,  -5,  -2,  -6, -44,   1,   4,   6,  10,   1,  25,   6, -19,   2,   0,   1,   6,   1, -17,  97,  13,   3,   1,   3,  -2,   1,  17,  39,  -9,  -1,  -3,  -1,  -1,  -1,  -6,   5,  -9 },
      {   0,   0,   1,   2,   0,  -4,   2,   6,   0,  -1,  -1,  -2,  -1,   1, -12, -20,   1,   1,   1,   0,   4,   1,  27,  17,  -1,  -2,   0,  -1,  -6,   1, -51,   4,   0,   2,  -3,   7,  -4,   4,  63, -42,   0,  -3,   2, -12,   4, -18,  -4, -36,   3,  -4,   5,   2,  -3,  10, -42,   1,  -1,   7,  -2,   5,   4,  46,  -9,  19 },
      {   0,   0,   0,   1,   0,  -5,  -2,   5,   0,   0,   0,   0,   0,   8,   3,  -8,   0,   1,   0,  -1,  -1,  -4,  -8,  34,  -1,  -1,   0,   3,   3,   4,   2, -57,   2,   1,   2,   1,   0, -12,  13,  61,  -2,   1,  -1,   0,   5,   8,  -5,  52,   0,  -5,  -2,  -3,   0, -29, -26,  -5,   1,   5,   2,   2,  -3,  53,   5,  17 },
      {   0,   0,   0,   0,  -2,   1,  -2,  -8,   0,   0,   1,   0,   0,   0,  10,   4,   0,   0,   0,   1,   0,  -1,  -7, -11,   0,   1,  -1,  -2,   1,   7,   8,  24,   1,   0,   2,   2,  -7,  -8,  -8, -44,  -1,   5,   2,   3,   9,  35,   6,   4,   3,  -4,   5,  -1,   1, -44,  59,  19,   4,   5,   5,   3,   0,  70,  27,  17 },
      {   0,   0,  -1,  -1,  -1,  -1,   2,   1,   0,   0,  -1,   1,   2,   4,  -1, -13,   0,   2,   0,  -1,  -1,   3,   9,   7,  -1,  -1,   1,   0,   1,   3,  -6, -30,   1,   0,  -2,   0,  -1,  -4,  14,  38,  -2,   2,  -1,  -1,  -1,  12, -31, -67,   2,   2,   6,   2,  -1,  21,  51, -48,   3,  -1,   3,   1,  -1,   4,  50,  -7 },
      {   0,   0,  -1,  -1,  -2,  -2,   0,  -2,   0,   0,   1,   0,   0,   4,   0,   7,   0,   0,   0,   0,  -1,  -2,  -8, -12,   0,   0,   1,  -1,   1,   5,  20,  -2,  -2,  -1,  -1,   0,  -2,  -5, -45,  -2,  -1,   0,  -1,   0,   2,   8,  33, -47,  -2,  -7,   0,  -6,  -2, -23, -35, -64,   0,   1,  -2,   1,   0,  38, -48, -23 },
      {   0,   1,   1,   0,  -1,   1,   4,   1,   0,  -1,  -1,   0,   3,  -1,  -8,  -7,   0,   0,   1,   1,  -1,   1,   4,   8,  -1,  -1,  -2,   0,   2,  -2, -14,  -8,   1,   1,   2,   1,   0,   3,  28, -12,  -2,  -1,  -1,   0,   5,   0, -43,  33,   3,  -2,   0,   0,  -4,  -4,  57, -27,  -4,  -1,  -4,  -2,   6,   2, -79, -41 },
      {   0,   0,   0,   0,   0,  -1,   3,   3,   0,   0,   0,  -1,  -1,   0,  -1,   3,   0,   0,  -1,   1,   0,   1,   0,   0,   0,   0,  -1,  -2,   1,   0,   0,  -9,   0,   1,   0,   0,   0,  -1,  -5,  32,   0,  -2,  -1,  -1,   0,   0,  11, -40,   4,   1,   2,   0,  -1,  -9,  31,  32,  -5,   2,  -4,   1,   1, -14, -61,  86 },
      {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,  -1,  -5,   0,   0,  -1,   0,   1,  -4,   2,   9,   0,   0,   0,  -2,  -2,   4,   1, -14,  -1,   1,   0,   0,   2,  -4,  -1,  28,   0,  -2,   2,   0,  -2,   3,   8, -44,   2,   1,   1,   2,  -1, -11,   1,  85,   0,   1,  -1,   0,   3,  17, -14, -74 },
    },
    {
      {  71, -46,   6,  -4,   1,  -1,   0,   0, -75,  42,  -1,   2,   0,   1,   0,   0,  37, -11,  -7,   1,  -1,   0,  -1,   0, -14,  -3,   7,  -1,   1,   0,   0,   0,   5,   5,  -3,  -1,   0,   0,   0,   0,  -3,  -3,   1,   1,   0,   0,   0,   0,   2,   2,   0,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0 },
      { -75,  17,  12,   1,   2,   0,   1,   0, -11,  40, -17,   3,  -3,   1,  -1,   0,  62, -48,   4,   0,   0,   0,   0,   0, -34,  15,  10,  -4,   2,  -1,   1,   0,  19,   1, -11,   2,  -1,   0,  -1,   0,  -4,  -6,   5,   0,   0,   0,   0,   0,   4,   2,  -2,  -1,   0,   0,   0,   0,  -1,  -2,   0,   1,   0,   0,   0,   0 },
      {  47, -24,   5,  -4,   1,  -1,   0,  -1,  62,  -9, -19,   2,  -4,   0,  -1,   0, -13, -43,  26,  -3,   4,   0,   1,   0, -53,  52,  -4,  -3,   1,  -1,   0,  -1,  23, -13, -10,   6,  -2,   1,  -1,   0, -14,   3,   8,  -3,   1,   0,   0,   0,   4,   1,  -5,   0,   0,   0,   0,   0,  -5,   0,   2,   1,   0,   0,   0,   0 },
      { -29, -37,  39,  -6,   7,  -2,   3,  -1, -17,  46, -30,   4,  -3,   1,  -2,   0, -50,   6,  19,   2,   1,   1,   1,   0,  46,  26, -26,  -3,  -3,  -1,  -1,  -1,  14, -45,   6,   7,   0,   2,   0,   1, -10,  16,   9,  -7,   2,  -1,   1,  -1,   5,  -3,  -8,   3,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0 },
      {   1, -46,  23,  -2,   3,   0,   2,   0,  58,  37, -30,  -5,  -2,  -1,  -1,  -1,   7, -21, -14,  12,  -4,   3,  -1,   1,  19, -48,  29,  -2,   2,  -1,   1,  -1, -46,  32,   3,  -6,   4,  -2,   0,  -1,  15,   2, -10,   3,  -3,   0,   0,   0, -10,  -1,   2,  -3,   1,   1,   1,   0,   3,   1,  -3,   3,   1,   0,   0,   0 },
      { -24, -27,   9,   5,   1,   2,   0,   1, -16,   9,  21, -14,   5,  -4,   3,  -2, -58,  13, -15,  14,  -2,   3,   0,   1, -28,  18,  33,  -6,   2,  -2,   0,  -1,  27,  44, -42,  -2,  -1,  -2,  -2,   0,  23, -45,   5,  12,  -4,   2,  -1,   1,  -7,  14,  10, -10,   2,  -1,   1,  -1,   7,  -3,  -8,   4,   0,   0,   0,   0 },
      {  19,  14, -32,  15,  -7,   3,  -3,   1,  30,  27,   7, -19,   2,  -4,   0,  -1,  18,  -1, -31,  14,  -4,   3,  -1,   2,  42,  -8,   7,  -6,  -2,  -2,  -1,  -1,  27, -41, -29,  10,   0,   2,  -1,   1, -24, -38,  44,   2,   0,   2,   1,   1, -15,  32,  -3, -14,   5,  -3,   2,  -1,   2,  -8,  -5,   9,  -1,   1,   0,   0 },
      {  25,  58,  -9,  -1,  -7,  -4,  -1,  -1, -22,  11, -25,   6,  -1,   4,  -3,   2, -23, -50,  37,   3,   3,   0,   2,   0,  36,  22,   3, -20,   2,  -3,   2,  -2, -39,  19, -25,   8,   0,   1,   1,   1,  34, -17,   4,   5,  -4,   3,  -1,   1, -12,  -5,   6,  -3,   3,  -1,   0,  -1,   4,   4,   0,   0,  -1,   0,   0,   0 },
      {  -5, -32,  28, -12,   8,  -2,   1,  -1,   8, -39, -19,  22,  -1,   3,  -2,   2,  51,  36,  19, -15,  -1,  -4,  -1,  -2,  21,  20, -32, -13,   0,   0,   1,   0,  -5,   2, -26,  12,  -5,   4,   2,   1,  21, -50,   5,   9,   0,   5,  -1,   1, -25,  -2,  25,   0,   2,  -2,  -1,   0,   2,  12,  -2,  -6,   0,   0,   1,   0 },
      {  21,  42,   1, -20,   0,   0,  -2,   0,  20,  48, -31,  15,  -7,  -1,   1,  -1,  -9,  27, -12, -13,  -1,   2,  -2,   1,  -9, -15, -15,  27,   1,   0,  -2,   0,  49,   6,   8, -12,  -3,  -3,  -3,  -1,  10, -28, -39,  10,   1,  -1,   0,   1,  -2, -34,  30,   5,  -5,   4,   1,   2,  -6,   8,   0,  -6,   4,  -2,   1,   0 },
      {  11,  34,  52, -40,   8,  -7,   0,  -3, -12, -22, -35,  33,  -5,   4,  -1,   2, -13,  10,   1,   0,  -3,   3,  -1,   0, -20, -38,  21,  -7,   6,   0,   2,   1,   5,   5,  -1,   7,  -2,  -1,   1,   0, -33,  -2,  32,  -6,   1,   1,   0,  -1,   5,  45, -11,  -9,   2,  -4,  -1,  -1,   8, -11, -15,   4,   0,   1,   0,   0 },
      {  -6, -20,  -7,  -1,   6,  -1,   2,  -1, -10, -34,   6,  20, -10,   5,  -3,   3,  -7, -48,   9,  -6,  10,  -1,   2,  -1,  43, -10,  29,  10,  -8,   1,  -2,   0,  29,   6,  -2, -34,   6,  -5,   1,  -1, -51, -10, -23,  29,   3,   2,   2,   1,  11,  -2,  38,  -4,  -7,   3,  -3,   0, -12,  17,  -6, -10,   6,  -3,   0,  -2 },
      {  12,  19,  32, -10,   5,  -4,  -1,  -2,  11,   1,   7,  -9,   0,  -2,  -2,   0,  19,  27, -20,  17,  -3,   1,   0,  -1,  24,  30,  27, -37,  -3,   1,   1,   0,   1,  15, -46,  -5,   4,   0,   8,  -1, -24,  22, -16,  13,  -3,   6,  -1,  -1,  41, -38, -23,  16,  -2,   1,  -5,   1, -15, -18,  23,  -2,  -2,   0,   1,   0 },
      {   2,  -5,  11,  11, -14,   6,  -3,   3, -12, -42, -31, -13,  20,  -3,   5,  -1,   1, -25, -10,  40, -18,   7,  -4,   2,  12, -24,  13, -14,  11,  -1,  -1,  -1,  53,  -4,  13,  17,  -4,  -4,  -3,   0,  25,  -4,  -7, -34,  -1,  -1,   1,   0, -26, -27, -16,  24,   9,   2,   4,   1,   2,   2,  30,   0,  -7,   1,  -3,   0 },
      {  -1, -14, -32,  -7,   0,   8,  -1,   3,   9,   5,  -1,  28, -11,   1,   3,   0,  14,  11,  39, -30,   1,  -1,  -1,   2,  34,   3,  12,  11,   6,  -8,  -1,  -4,  40,  40,  -2,  -6,  -5,  -4,  -7,  -1,  18,  12,  -8, -24,  -1,  -4,  -1,   0,  13,  16, -46,  -9,   5,   2,   5,   0,   9, -40, -13,  26,   0,   2,  -2,   1 },
      {  10,  16,  34,  33, -28,  -2,   0,   1,   7,  14,   9, -40,  23,  -2,  -3,  -1,  12,  15,  35,   8,  -8,  -8,   2,  -4,   9,  15,  -9, -14,  -3,   7,   0,   1,  11,  29,  23, -22,   0,  -5,  -1,   0, -31,   1, -28, -10,  10,  -2,   4,   0, -22,  40,   7,  16,  -2,  -1,  -2,  -3,  31,   5, -18, -17,  -3,   1,  -2,   0 },
      { -11, -15, -55,  39, -14,   6,   1,   2,  -9,   3, -62,   4,  10,   3,  -2,   2,  -9,  32,  29,   6, -12,  -2,   0,  -1, -22, -20,  -1, -27,  10,   2,   3,   0, -12,   8, -15,  13,   0,  -1,   3,   0, -35,  -6,   2,  11,   1,   5,  -1,   1,  30,  -3,  11,  10,  -2,  -1,  -3,  -1,  -4,  -2,   9, -10,   0,   0,   1,   0 },
      {  -2,  -2,   9,  -9,   2,   3,  -2,   0,   9,  21,  56,   8,  -6,  -4,  -2,  -2, -12, -17,  29, -26,  -2,   1,  -2,   0,  -1, -43, -28, -12,  21,  -5,   4,   0,   2,  20, -17,  34, -15,   2,   0,   0, -29, -20,  11, -20,   3,   5,  -1,   1,   2, -12,  -4,  32,   5,  -3,   2,   0, -10,  12,  38, -15, -11,   1,  -1,  -1 },
      {  -5,  -3, -33, -17,  30, -15,  10,  -5,  -2,   6, -11,  29, -23,  11, -10,   1,  -7,   3, -35,   7,  12,  -5,   1,  -1,  -1,  29,   7,  -9,  -8,   4,   1,   2, -15,  10,  14, -16,  -7,   3,   2,   2, -20, -12, -11, -12,  14,  -5,   7,   1, -38,  18, -14,  49,   0,   0,  -2,  -2,  36,  -7,  35, -15, -16,   3,  -5,   0 },
      {  -2,   1, -20, -26,   8,   1,   0,  -1,   7,  19,  24,  25,  -5,  -1,   2,  -2,   7,  27,  34,  20, -22,   6,  -5,   2,  -3,   9,  54, -36,  10,  -8,  -1,  -1,   8, -18,  12,   9,  -2,  -5,   1,  -4,   4,  28,   4, -16,  -8,   1,  -3,  -1, -28,  -9,  22, -13,   6,   0,   2,   1,  -1,  48, -16, -13,   3,  -4,   0,  -1 },
      {  -8,  -8, -20, -64, -17,  19,  -5,   3,  -4,  -4,  -3,  -9,  40, -28,  20,  -9,  -1, -12, -17,  10, -40,  24,  -7,   6,   8,  22, -19,  12,  31, -13,  -2,  -2, -21,   5,  -2,  13, -14,  -4,  -4,  -3, -25,  -6, -27,  -4,   8,   0,  -2,   2,   6,  10,   9,   4,  -7,   3,   5,   0,   2, -11, -17,   9,   5,   2,  -1,   1 },
      {   5,   2,   5,  26,  -9,   0,   7,   0,   9,   3,   4,  36,  -4,   0,  -8,   2,   2,  -8, -40,   8, -13,   1,   3,   1,  19,  25,  -9,  -5,  16,   3,   1,   2,  11,  32,  33,  29, -19,  -2,  -3,   2,  21,   6,  26, -13,   6,  -3,  -2,   0,  51,  19,  22,   1, -13,  -8,  -5,  -4,  -4,  13, -11, -43,   5,   0,   1,  -1 },
      {  -1,   6,   8,  22, -19,  14,  -4,   4,  -5,   2, -19,   9,  14, -12,   6,  -4,  -2,  14, -17, -27,   6,   4,   1,   1,   5,  29,  18,  35,  -2,  -4,  -2,  -2, -15,  30,   5,  -6,  -7,  -2,  -2,   0, -33,  -1,  24, -24,   0,  -2,  -1,   0, -35, -10, -20,  -8,  12,  -1,   6,  -1, -56,  41,  13,  16,   0,  -1,   0,  -1 },
      {   2,   1,  21,  30, -46,  38, -15,  15,   2,   2,  12,  50,   7, -20,   8,  -5,  -1,   2,   2, -22,  -7,  13,  -5,   1,  -7,   1,  35,  18,   1,  -9,  -4,  -2, -16, -41, -17, -11,   4,  -4,  -2,  -1,  13,  -3, -10,   8,   7,   1,   0,   0,  -1,   3,   5,  19, -11,   1,   2,   1,  30, -23,  20, -15,  -1,   3,  -2,   0 },
      {   1,   0,  -6, -30,  11,   6,  -4,   2,  -1,  -5, -17, -54,   2,  -4,  16,  -2,   9,  12,  23, -10,  25,   2,  -4,   1,   9,   2,  35,  47, -14, -10,  -3,  -4,   1,   0,  -7,   2, -15,  10,  -6,   1,  14,   2,  30,  -9,  10,  -5,  -1,  -1,  30,   7,  10,  18, -15,   0,   1,   0,  22,  20,  24, -39,  -5,   1,  -1,   1 },
      {   0,   2,   4,  -5,   9,  -7,   5,  -8,   3,   6,  15,   9,   6,  -1,  -4,  -1,   4,   9,  37,  46, -21,   1,  -4,  -1,  -3,   6,  13,  13, -12,   3,  -2,   2, -16, -11,  39, -27,   3,  -3,  -3,   1,   9, -49,   6,  -4,   6,  -3,   5,   1,  18,  12,  -5,   5,  -7,   0,   0,  -2, -58, -35,  25,  -4,   3,   2,   0,   1 },
      {   4,   7,  16,  40,  33, -13,  -7,   3,  -5,  -7,  -1, -16, -50,  16,   8,  -5,   3,   6,   5,   9, -14,  20, -12,   7,   2,  10,  14,  25,  44, -36,   2,  -8, -11,  -4,   6,  25, -34,  10,  -8,  -2, -19,  -3, -17,  -2,   7,  -6,  -4,   4,  -8,  -9,  19, -14,  -9,   8,   9,   0,   9, -20,   1,  23,   1,   0,  -1,   0 },
      {   2,   2,   1,   0,  -4,  17,  -4,   7,   3,   3,   7,  24, -12,  -4,  11,  -5,   7,   4,  14,  61,  -3,   7, -11,   2,   2,   5, -29,  37, -24,   0,  -5,  -2,   1,  27,  -3,   8,  27, -15,  -5,  -2, -16,  11,  19,  16, -32,   1,  -3,   4,   8,  -5,   6, -15,  13,   7,   4,   2,  39,  19,  28,  23, -17,  -5,  -4,  -2 },
      {  -4,  -7,   1, -21, -61,  -2,   1,  -2,  -1,  -2,   5,  -7, -29,  30,  -9,   7,  -1,   1,  -4,   4,  14, -20,   6,  -1,  -2,   7,  15,  -7,  -7,   6,  -3,  -1, -16,  -8,  21,   0, -12,   3,  -4,   2, -17, -42,   2, -36,   9,   3,   3,   2,  30, -38,  -2,  -8,  18,   3,   1,   3,  28,   6, -21,  25,  -8,  -4,   0,  -1 },
      {  -4,  -4,  -2, -26, -50,   1,  11,   5,  -2,  -1,  -5, -24, -37,  24,  -6,   0,   1,   5,  -5, -18,   1,  -5,   5,   0,   4,   9,   8,  -8,  14,   4,   0,  -1,   6,   4,  14,  30,  21, -21, -10,  -1,   5,  17,  -9,  30, -28, -13,   4,   5, -18,  31,  19, -10,   5,   5,   0,  -3, -28, -22,  44, -10,  -5,   0,  -4,  -2 },
      {  -1,  -1,  -2,  -8,  -2,  49, -39,  21,  -3,   0,  -3,  -4, -38, -24,  24,  -3,  -1,   0,  -1,  27,  50,   0, -10,  -3,  -2,  -2, -15, -24, -17,  -6,   4,   3,   1,  14,  10,   4,  -5,  12,  10,  -5, -10,   2,   4, -11,   5,  10,  -7,  -5, -22,  -7,   6,  -4, -12,  -1,  -1,   4, -20, -31, -23, -28,  17,   2,   7,   3 },
      {  -3,  -3,  -3,  -4, -25, -30,  11, -17,   2,   4,   8,  21,  34,  25, -21,   4,   1,   3,   6,  34,  23, -31,   7,  -2,   0,  -6,   2,  39,  -6,  -1,  -2,  -2,   1,   5, -33,  25, -27,  12,  -4,   0,  -9,  30,   3,  -7,  10,  -4,   2,  -2, -31, -13,  20,   2, -15,   3,  -3,   2, -10, -34, -14, -14,  10,   4,   3,   3 },
      {   2,   3,   4,   3,  36,  22,  -9,   5,   2,   5,   2,   1,  54,  -8,  -7,   1,   0,   0,   1, -20,  28, -22,  14,  -7,   5,  11,  27, -10,  -3,  17,  -3,   3,   0,   1,  44,  41, -11,  -5,   4,  -2, -12, -14,  -1,  13, -23,  12,  -3,  -1,   1, -13,  23, -17,   9,  -4,  -4,   3,  18, -28,  16,  22, -16,   0,   2,   2 },
      {   2,   2,   2,  11,  17,   7,   6,  -7,   0,  -1,  -1,   5,  -4,   3,  11,   1,   1,   1,  -3,  12,  31,  -8,  -5,  -3,   0,   0,  10,  15, -13,   2,  -2,  -2,  -9, -20, -22,  44,  22, -28, -14,   0,   3, -12, -56, -35, -22,   5,   2,   5,  20,  36,  -4,  30,  16,  -7,   8,  -1, -22,  21, -24,   4,   2,   0,  -3,  -3 },
      {  -1,  -1,   0,   7,   8, -23,  20, -23,  -3,  -4,  -2,  -9, -12, -26,   6,   3,  -1,  -3,   1, -21, -29,  17,  -5,   3,   1,   5,   8,  15, -18,  11,  -5,   1,  -6,   5,   6,  16,  34, -26,  -7,   1, -19, -13,  18, -15, -35,   3,   2,  -2, -16, -38,  -7,   0,  11,   5,   0,   3,  16, -35, -27, -48,  21,   3,   1,  -1 },
      {   2,   3,   2,  -1,  21,  29,   9,  61,  -1,  -1,   0,  -6, -15,  25,   3, -20,  -1,  -2,   4,   4, -36, -38,  27,  -1,   1,   0,   3,  15,   7,  51, -21,   3,  -4,   0, -12,   7,  -2, -32,   1,  -2,  -1,   0,  10,   4,  23,  -7,   3,   2,  -8,  -9,  -9,  13, -16,   6,  -2,  -2,  -1,   2, -14,   0,  17,  -3,   1,  -2 },
      {   0,   0,   0,  -3,   6,  31,   1,   0,   0,   0,   0,  -1,  12,  39, -22,  14,   0,   0,   0,  15,  10,  -2,   8,  -4,  -1,   1,  -5,  18,  46, -20,  -6,  -2,  -3,  -8,   2, -18, -12,  -5,  18,  -4,  -3, -14, -12,  19, -34,  20, -10,  -2,  -2,  -2, -48, -21,  24,  -9,   1,   6,   6,  10,  -6, -52,   0,   2,   9,   0 },
      {   0,   1,  -1,  -2,  -3,  20,   7, -35,  -3,  -3,  -5, -12, -34, -46, -36,  24,   1,   2,   5,   8, -11,  -1,  27, -10,   0,   0,   2,  18,  14,  24, -18,   7,   2,  -2, -10, -11, -23, -15,  26,  -1,   8,  17,   4,  -5,   5,  39,  -5,  -5,  -2,   8,  24,  16,  22, -28,  -5,   5,  -6,  -8,  -3,  17, -25,   7,   3,  -1 },
      {  -1,   0,  -1,   0,  11,  -7, -31,  28,   0,   1,  -3,  -2,   8,  35,  13, -13,  -3,  -4,  -9, -14, -30,  -9, -13,   6,  -2,   1,  -9,   0, -19,  -3,   6,  -1,  -4,  -2, -18, -33, -11,   8,   1,  -3,   2,  16,   2, -43, -27,  21,   4,   4,   6,   8,  42,  -8,  47, -10,   3,  -4,  -9, -24,   6, -15, -28,  -2,  -2,   0 },
      {   2,   3,   3,  -2,  12,  60,  19, -16,   0,   1,   2,  -9,   9,  51, -37,  18,   0,   1,  -2,  -7, -10,  33,  -1,  -5,   0,   3,  -2,  -3, -13, -28,  11,  -5,   2,   9,   2,  -2,  33,   8,  -9,   1,  -1,   6,  22, -10,   1,  -6,   2,  -2,  -4,  -8,  24,  36, -16,  -6,   0,   4,  -4, -10, -15,  22,  10,  -1,   0,   0 },
      {  -2,  -4,  -2,  -1, -13, -40, -44,  13,   2,   3,   3,   2,   4,  20,  21,  -8,   1,   2,   5,   0,   9,  16,  -3,   1,   3,   6,   8,  11,  21,  -3, -11,  12,   5,   3,  18,   7,  18,  -5,  29, -12,   6,  -1,  25,  36, -11,  17,  -9,  -6,  -2,  -5,  -2,  63,   7, -17,  -7,   1,  -8,  -1, -32,  18,  -8,   4,   6,   3 },
      {   0,   0,  -1,  -3,  -6,  -9,  10,  53,  -1,  -2,   0,  -1,  -9, -21, -42,  -3,   0,  -1,   4,   1, -24,  18,  21,  -9,   0,  -2,   7,   9, -51, -23,   6,   3,   2,   1,   7,  30, -17,  49,  -1,  -3,  -3,  -4, -18,  24,  -6,  -5,  -3,  -4,   1,   5, -16,   9,  30, -10,  -5,   1,  -4,   5, -11,  -6,   0,   3,   7,   2 },
      {  -1,  -2,  -2,   1,  -3,   1, -36, -36,   1,   2,   1,   0,  -2,  41,  15,   8,  -2,  -3,   0,  -8, -29,  31,  -1,  -3,   1,   1,   4,   6, -46,  29,  -8,   8,   1,   3,   0,  30,  -8,  -1,  24,  -6,  -2,  -2, -23,   1,  29,  25, -26,  -7,  -5,   9, -17, -28, -21, -14,  -1,   9,   3,   7,  18,  -7,   7,   3,  11,   3 },
      {   0,   0,   2,   2,   2,  28, -28, -52,   0,  -1,  -1,  -1, -12,  -1,  19,   1,  -3,  -4,  -3,  -4, -38, -64, -14,   1,   1,  -1,   3,   0, -18, -12,  16,  -6,   1,   0,   5,   5, -27,  14, -18,   5,   1,   2,   4,  35, -18, -25,   6,   5,  -1,   0, -14,  25,  12,   6,   2,  -7,  -3,   7, -16,   0,  -3,   3,  -7,   1 },
      {   1,   1,   2,   3,  -8,  10,  68,   7,   1,   1,   2,   1,   3,  14,  34,  -7,   0,  -1,   3,  -2,  14,  26, -38,  13,  -1,  -4,  -2, -11, -25,   8,  10,   0,   1,  -3,   5,  -6, -59, -32,   4,  -4,   2,   5,   7,  25,  -2,   8, -13,  10,  -4,  -6,  -3,  11,  13, -18,   6,  -5,  -5,  -5, -11,   4, -11,  12,  -4,   0 },
      {  -1,  -1,  -1,  -2,  -7, -12,  -2,   9,  -1,  -1,  -1,  -3,  -3,  -5, -13,   0,  -1,  -1,  -3,  -6,  -8, -12,   0,  -4,  -1,  -1,  -3,  -7, -15, -17,  -1,  -1,  -1,  -2,   2, -13, -35, -12,  13,   0,   1,  -2,   2, -15, -63,  41,  -3,   1,   1,  12,   0,  14, -50,  15,  -5,   3,   8,  10,  17,  23,  56, -14,   4,   2 },
      {  -1,  -2,  -1,  -1,   3,  -1, -49,  -8,  -1,  -1,  -1,  -1,   4,   2, -44, -12,   2,   2,   2,   1,  14,  46,   6,  10,  -2,  -4,  -5,  -8, -13,  16, -15,  -3,  -4,  -2,  -6,  -4, -41, -47,  -9,   6,   1,   0,   2,   4, -13, -47,  32,   6,   3,  -2,  -3,   1,   4,  26,  -2,  -6,  -3,   3, -10,  -4,  -8,  -3,  -7,  -6 },
      {   1,   1,   1,   4,   0,   5,  20, -14,   1,   2,   1,   5,   1,  11,  51,  10,   0,   0,  -1,   0,   1,  14,  54, -22,  -2,  -4,  -5, -12,  -7, -18, -76,  20,  -1,   2,  -3,  -6, -12,  25,   8,  -6,  -1,   0,   0,  -6, -20, -26,   8,   2,   2,   3,   6,   3,   4,  16,  -2,   1,  -3,  -3,   0,   0,  -3,  -2,  -3,  -3 },
      {   0,   0,   1,   1,  -1,   9,   2, -15,   1,   1,   3,   3,  -2,   6,  12, -17,   2,   3,   5,   6,  11,  30,   8,  10,   2,   2,   4,   6,  24,  70,  21,  -6,   0,  -1,  -1,  -7,  -9,  55, -25,   7,   0,   3,  -1,   3, -20,  -1,   4,   2,   2,   6,   4,  14,  31,  -1,  -8,   4,   5,   0,   7,   6,  48, -17,   5,  -2 },
      {   0,   0,   0,   1,  -2,  12,  15, -27,  -2,  -3,  -3,  -2,  -4,   0, -26,-103,   0,  -1,   1,  -3,   5, -10,  -8,  34,   0,  -1,   0,  -2,  -2, -11, -25,  18,   1,   1,   1,   2,  11,  15,  27, -11,   0,  -3,  -4,  -6,  -3,   0,  -4,   2,   3,   1,   0,  -4,  -6, -10,  -3,   6,   1,   1,  -1,  -1,  -6,   2,   1,   0 },
      {   1,   1,   1,   1,   0,   8,  12,   8,   0,   1,   0,  -2,   1,  -3, -14,  32,  -2,  -2,   1,  -6,  -5,  -8, -72,   1,   1,   2,   5,   8,   6,  21, -23,  20,   0,   0,  -3,  10,   2,  24,  54, -12,   2,  -2,  -5,  -6,  -3, -20,  12, -14,   2,   8,   2,   2,   8,  45, -19,   6,  -4,  -1,  -3,   2,  -5, -29,   3,  -1 },
      {   0,   1,   1,   1,   3,  -1,  -1,  -8,   1,   1,   1,   0,   6,   1,   5,  -5,   1,   1,  -1,  -1,   5,  -8,   7,  14,  -1,   0,  -2,  -3,  -9, -37, -14, -17,   1,   4,   6,   5,   0, -32,   9,  11,   5,   5,   6,  12,  48,  18,   7,  -9,   3,   6,   8,   4,  64,  20, -14,   8,   6,  -5,  16,   2,  57, -33,   8,  -2 },
      {   0,  -1,  -1,  -3,   2,  -6, -12,   2,   0,   0,  -1,  -2,   3,  -3, -19,  -2,  -1,  -1,  -1,  -1,   3,  -8, -34, -55,   1,   1,   0,  -2,   6,   1, -23,  31,  -1,   1,  -2,  -7,   0,  -6,  -2, -10,  -2,   1,   3,  -3,   6, -33, -34,  28,   4,  -3,   7,  -2,  20, -38,  38, -15,   9,  -1,  13,   4,  47,  21, -14,  -2 },
      {   0,   0,   1,  -1,   1,  -2,  -8,   9,   0,   0,   1,  -1,   2,  -2, -13,  32,   0,   1,   0,  -2,   2,  -8, -23,  71,  -2,   0,  -2,  -8,  -3,   6, -66, -34,  -1,   3,   4,   2,   5,  19, -24,   8,  -3,   1,  -2,   7,  -1,  14,  10,   6,  -2,  -1,  -3,   1,  -8, -33,  23, -13,  -1,   2,   2,   1,  10,  15,  -9,  -2 },
      {   0,   0,   1,   0,   1,  -1,   0,   1,   0,   0,   1,   1,   2,  -5,   1,  21,   1,   1,   1,   1,   2,  -5,  36,  56,   0,   0,  -1,   1,  -3,   2,  30, -12,  -1,  -2,  -2,  -5,   3,  -3,  47, -22,  -1,  -3,   0, -11,  -8, -41, -63,  13,   4,   4,  10,   6,   8,  19,  24,  -4,   1,  -3,   3,   1,  10,  10,  -9,  -5 },
      {   0,   0,   1,   0,   1,   5,  -2,  -4,   0,   0,   0,  -1,   1,   3,  -4,  11,   1,   1,   1,   0,   2,   4,  10,  23,   1,   1,   2,   3,   0,  -3,  12,  64,   1,   0,   2,   5,   3,   6, -16, -40,   3,   3,   2,   6,  15,  43,  26,  48,   0,  -1,  -5,  -2,   6,  39,   9, -35,   0,   2,  -2,  -4,   5,  13, -30,  14 },
      {   0,   0,   0,   0,   0,   3,   1,   5,   0,   1,   1,   0,  -1,   3,  15,  14,   0,   1,   3,   3,  -2,  -4,   8,  26,   1,   1,   1,   6,   1,  -9,  25,  39,   0,   1,  -3,   3,  -1,  -7,  34,  14,  -2,  -2,  -6,  -8,  -7, -28,  61, -33,   2,  -3,   4, -11,   3, -46, -19,  -3,   8,   4,  12,   8,  28,  44,   1,   9 },
      {   0,   0,   0,  -2,  -2,   3,   2,  -1,   1,   1,   1,   1,   0,   4,   4, -12,   0,   0,   1,   1,  -1,   8,   4, -24,   0,  -1,   2,   5,   0,   5,   3, -41,   3,   2,   4,   8,   7,  13,  53,  67,   0,   1,  -2,   2,   1,  15,  19,  29,   0,  -2,   0,  -2,   0,  16,  18, -45,  -1,  -1,  -6,  -2,   1,  15, -41,  15 },
      {   0,  -1,   0,  -1,  -1,   0,   0,  -1,   0,   0,   0,   0,   1,  -4,   7,   8,   0,   0,   1,   1,  -1,  -1,   7,   7,   0,  -1,   1,   5,  -3,  -7,  11,   5,   1,  -1,  -1,   3,   1,  -1,  18,  -2,  -1,  -1,  -3,  -2,   0, -14,  19,  60,  -5,  -6,  -6,  -4, -14, -51, -13, -26,   2,  -7,  -1,  -2,  -6, -77,  18, -37 },
      {   0,  -1,   0,  -1,   0,   1,  -1,   4,   0,   0,   0,  -1,   1,   2,  -3,   3,   1,   0,   1,   1,   1,   1,   3,  21,   1,   0,   1,   0,   0,  -3,   2,  65,   0,   0,  -1,  -1,  -3,  -2, -16,  72,  -1,   1,  -1,   5,  -3,   5, -30, -41,  -2,   0,  -2,   4,  -3,   2,  37,  -5,  -2,  -4,  -4,  -3, -10, -42, -17,  -7 },
      {   0,   0,   0,  -1,   1,   2,  -3,   3,   0,   0,   1,   0,   0,  -1,  -2,   6,   0,   0,   0,   0,   0,   0,  -9,  14,   0,  -1,   1,   1,   3,  -4, -11,  20,   1,  -1,  -2,  -3,   3,  -4, -14,  51,  -1,   1,  -2,  -7,  -2,  -7, -38,  46,  -2,   1,   2,   1,   4,  11, -67, -11,   0,   3,   1,   4,  -3,  30,  50,  29 },
      {   0,   0,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,  -2,   0,  -2,  -1,   0,   0,   0,   1,   1,   2,   3,  -1,  -1,  -1,  -1,   2,   1,  -1,  -5, -11,  -1,  -2,  -3,  -2,  -1,  -6, -10, -27,  -2,  -3,  -3,  -4,  -4, -11, -23, -35,  -1,  -3,  -2,  -4,  -2, -23, -41, -53,   0,  -1,   1,   0,   0, -36, -41,  73 },
      {   0,   0,  -1,  -1,   1,   1,  -3,   0,   0,   1,   0,  -1,   1,   2,   0,  -2,   0,   1,   1,   1,   1,   1,  -4,  -3,  -1,  -1,   0,   2,   1,  -3,  -2,   0,   0,  -1,   0,   0,   0,   3,  -8,  -2,  -1,  -2,  -1,  -3,   1,   8, -23, -24,  -2,  -2,  -2,  -4,   1,   5, -56, -38,   0,   0,   0,   2,   5,  24, -42, -90 },
      {   0,   0,   0,   1,   1,  -3,  -4,   5,   0,   0,   0,   0,   0,   0,   1,   6,   0,   0,   0,   0,  -1,  -1,  -1,   5,   0,   0,   0,   1,   0,  -2,   0,   5,   0,   0,   1,   0,  -1,   0,  -1,  13,   0,   0,   0,   1,   1,  -3,  -5,  35,  -2,   0,  -3,  -3,  -3, -15, -33,  84,  -4,   0,  -3,  -3,   1,  -7, -78,  11 },
    },
    {
      {  86, -53,  -3,   0,   0,   0,   0,   0, -67,  31,  13,  -2,   2,   0,   1,   0,   2,  14, -11,  -1,  -1,   0,   0,   0,   7,  -8,  -2,   4,   0,   1,   0,   0,   0,  -1,   3,   0,  -1,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0 },
      { -40, -36,  37,  -1,   5,  -1,   2,   0,  -6,  62, -27,  -6,  -3,   0,  -1,  -1,  58, -34, -23,  15,  -4,   3,  -1,   1, -26, -13,  29,  -2,   0,   0,   0,   0,  -1,  16,  -3,  -9,   2,  -1,   1,   0,   2,  -3,  -5,   3,   1,   0,   0,   0,   2,  -1,  -1,   1,   0,   0,   0,   0,  -1,  -1,   1,   1,   0,   0,   0,   0 },
      {  -8,  71, -41,  -3,  -6,   2,  -2,  -1, -67,  -7,  15,  14,   3,   1,   0,   1,  61, -23,   0,  -6,  -1,  -1,  -1,   0,   8, -12,   4,   0,   1,   0,   0,   0,  -6,   1,   6,  -1,  -2,   1,   0,   0,  -2,   4,   0,  -3,   1,   0,   0,   0,   1,   1,  -1,   0,   0,  -1,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0 },
      {  50,  21,   5, -16,  -5,  -1,  -1,  -2,   8,   1, -42,  12,   5,   0,  -1,   1,  -4, -63,  42,  14,  -4,   1,   0,   1, -29,  50,   7, -24,   1,  -2,   1,  -1,   8,   1, -28,   8,   3,   2,  -1,   0,  -1,  -9,  12,   4,  -2,   0,  -1,   0,   0,   0,   2,  -4,  -1,   0,   0,   0,  -2,   1,   1,  -1,   1,   0,   0,   0 },
      { -12,  28,  40, -25,  -6,   2,  -1,  -3, -27,  25, -54,  13,  10,  -4,   1,   2, -32,  10,  -3,  26,  -8,   1,  -1,   1,  66, -10,  -2,  -7, -12,   4,  -3,  -1,   7, -38,  13,  -3,   3,   2,  -2,   1, -11,  10,  10, -10,   5,  -3,   1,   0,  -1,   2,  -6,   1,   0,  -1,   1,  -1,   0,   0,   1,   2,  -2,   1,   0,   0 },
      {  42,  22,  37, -49,   4,  -2,   1,  -3,  27, -43, -10,  21,   5,  -2,  -1,   2,  32,  24, -51,  14,  -4,  -3,  -2,  -1, -35, -14,  11,  10,  -7,   4,   1,   0, -10,   1,  22,  -7,  -4,   2,  -2,   1,  -3,   7,  -3, -12,   6,   0,   1,   0,  -3,   1,  -4,   4,   1,  -1,   0,   0,  -1,  -2,   1,   1,  -2,   1,   0,   1 },
      { -28,  -8,  24, -14,   5,   5,   2,  -1, -27,   5,   3,  16,   3,  -4,  -3,   0,  20,  33,  -2, -27,  -1,  -4,   3,   0, -32,  53, -45,   9,  13,  -4,   1,  -1,  36, -33,  -5,  33, -17,   3,  -3,  -1,   6, -17,  26, -19,  -5,   5,  -1,   2,  -3,   7,  -1,  -8,   6,  -3,   0,   0,  -1,   0,  -4,   1,   2,   0,   1,  -1 },
      { -22, -21,  40, -19,   6, -12,   3,   2, -37, -52,  24,  -4,  10,   7,   4,  -2,  10,  13,   8,  19, -11,  -1,  -1,   0,  22,  28,  11, -40,   1,   2,   0,   0, -28,  19, -37,   1,  25,  -6,   2,   0,   8,   0, -12,  31, -13,  -3,   0,   0,  11,  -4,   2,  -3,  -7,   5,  -1,   0,   1,  -3,   0,   0,   3,   0,  -1,   1 },
      {   7, -25,  10,  -9,   8,  -3,   3,   0,  24,  -1,  -9,  14,  -1,   2,  -2,  -1,  36, -17,  42, -28,  -8,   1,  -1,   1,  26, -14, -44,  -4,  21,  -5,   0,   1, -61,  -5,  23,  27,  -5,  -2,   4,  -1,  -6,  38,   5, -17,  -5,   3,  -1,   0,   5,   0, -13,   2,   5,  -1,  -2,   1,   2,  -6,  -2,   4,  -2,   0,   0,  -1 },
      {  11,   2,  10,  29, -20,   4,  -2,  -1,  16,  -6,  22, -58,  20,  -4,   3,  -1,  25, -27, -34,  26,  21,  -7,   1,  -2,  25,  39, -21,  20, -31,   1,  -1,  -2, -24, -31,   1,  -3,   0,  11,  -1,   4, -13,  -3,  25,   0,   5,  -4,  -1,   0,   1,  10,  -4,  -8,   1,   0,   0,  -1,  -1,  -2,  -3,   4,   1,  -1,   0,   1 },
      {  -3,  18,  29,  18, -27,   1,  -4,   1, -28,  -1, -29, -20,  22,   2,   2,   1, -23,   4,   9, -13,  14,  -8,   2,  -1, -20,   9, -19,  31,  -9,   3,   2,   0, -34,  40,  17,  -2, -10,   5,  -5,  -1,  59,  -5, -26, -16,   3,  -3,  -2,   2,  -3, -27,   5,   9,   5,   0,   1,   0,   5,   4,  10,  -1,  -4,   0,  -1,   0 },
      {  10,  -4,   8,  53, -36,  -1,  -6,   2,   2, -25, -47,  -9,  14,  15,   4,   2,  34,  42,  26, -25,  12, -12,  -5,  -2, -11, -28,  28, -11, -12,  -3,   5,   0,  13, -23, -25,   7,  12,   3,   2,   2, -12,  11,  12,   9,  -3,  -2,  -2,  -1,  -2,   7,   1,  -5,  -2,   2,   0,  -1,  -5,   1,   0,   0,   1,  -1,   0,   1 },
      { -18, -16, -17,   6, -15,  14,  -2,   3, -20, -14,   9,  -9,  29, -14,   3,  -1, -23,  -9,  12,  42, -33,  -4,   5,   0, -51, -16,   0, -17,  -7,  27,  -1,   2,   0, -12,  35,   8,  20,  -8,  -8,  -2,   1,  35,  25, -19, -14, -11,   4,   1,   4,   9, -27, -11,   5,   4,   4,   0,   0,   1,  -6,   9,   0,   1,  -2,  -2 },
      {  -3,   8, -10,  16,  21,  -8,   0,  -1, -10,  12,  -7,   9, -37,  17,  -4,   2, -18,  10, -23, -11,  22,   1,  -1,   2, -19,  42,  35,  -7,  -2, -18,   4,  -4, -29, -32,   5,  -3,   3,   6,   3,   1,  10,  53, -12,  -2,   0,  -2,  -5,   0,  37,  -7, -29,   1,  -3,   0,  -2,   1, -10, -20,  11,   7,   3,  -2,   1,   1 },
      {  10,  10, -39,  10,   1,   7,  -7,   3,  27,  34, -15,  11,  -3,  -2,   0,  -3,  23,  42, -18,  35, -28,   2,  -5,   0,   6,  10, -20, -36,   3,   3,   0,   2, -28, -10,  -6,  13,  21,  -1,  -2,  -2,  49, -27,   5,  11, -11,  -5,   1,   1,  -9, -21,   8,  -8,   1,   5,   2,   0,  -6,   1,   7,   1,   0,   0,  -1,   0 },
      {   6,   2,   1,  11,  26, -19,  -2,  -4,   3,  -3,  -6, -23,  -4,  11,   8,   0,  28,  18,  35,  29, -22,   3,  -8,  -3,  -4,  17, -37,   8,  -7,   0,  -4,   4,  41,  -2,  23, -38,   7,   4,   0,   0, -22,  13, -36,   4,  13,  -8,   3,  -2,  19, -34,  -7,  27,  -9,   3,   1,  -2,  -9,  -3,  17,  -5,  -5,   0,  -1,   1 },
      { -11, -43, -24,   1, -42,  30,  -1,   3,   5, -23,  -6,  52,  17, -15,  -3,   0,  13, -17,  -9, -14,   3,  -4,   4,   2,  24,  37,  19,  -6, -14,   9,  -2,  -2,   8, -16,  28, -20,   1,  -3,  -3,   1,   5,  -9, -31,  -9,  16,  -3,   3,   0,  -6, -13,   1,  21,  -1,  -4,   2,   0,  -5,   9,  14,  -3,  -8,   3,   1,  -2 },
      {   4,  20,   3, -30, -36,  24,  -4,   4,  19,  48,  29,   2,  32, -18,  -7,  -3,   0,  25,   8, -17,  12, -27,  13,  -4,  -5,   5,  -7, -10, -17,  16,   3,  -5,   3,  24, -29,   4,  10,   2,  -7,   2, -27,  24, -14,   7,  -2,  -3,   2,   2,  31, -12, -13,   8,  -9,   1,  -3,   0,   6, -26,   4,   2,  -1,   4,  -2,   2 },
      {  15,  17,  31,  31, -29,   9, -12,   1,   6,  -3,  16, -13, -13,  -4,   7,   2,   0, -16, -29,  -1,  -2,  15,   2,  -1,  15,  -3,   0, -36,  46,  -4,  -2,   1,  32,  10,  16,  36, -10, -23,   4,  -6,  -8,   4, -18, -10, -32,  13,   2,  -1,   5, -29, -21,   4,  12,   8,  -2,   4, -11,   8,  19,   1,   0,  -4,   0,  -4 },
      {  -1,   1, -13,  21,  -5,   0, -12,   9,   0,  -2, -18,   4,  15,  11,   6,  -8,   1,  22, -10,  16, -25,   9,  -9,   1,  12,  29,   8,   0,  27, -18,  -6,   3,  -6,  50,  -8, -13, -16,  -5,   6,   0, -22,  -8,  10, -31,  23,   9,  -1,   1,  -1,  36, -31,  13,  14, -12,  -2,  -4,  26, -31, -19,  23, -19,   6,   4,   1 },
      {  -2,   8,   6,  19, -10,  17,  -9,  -2, -10,   3, -14,  11,   4, -19,   8,   2, -14,   8,  -8, -11, -29,  22,   5,  -4, -21,   8,  -6,  17,  19,   0, -14,   3, -42,  -4,  15, -18,  -3, -16,  10,   2, -49, -25,   9,  30,  -7,  11,   5,  -2,  44,   2,  34,  -8, -13,   5,  -7,   0, -14,  10,   2, -26,   8,  -1,   1,   0 },
      { -11, -20, -13,  -4, -13,  24, -23,   0, -12, -19, -19,  25, -33,  23,   8,   5,  -5,  -9, -13,  39,  47, -25,  -6,  -6,  -6, -22, -41,  -8,  -5, -12,   8,   0,   3,  25,  -3,   9, -12,  20,   0,   1, -17,  -8,  25,  16,   4,  -2,  -6,   1,  29, -20,  -2,  -2,   0,  -2,   0,   0,   4,  -9,   9,   3,  -3,  -1,   1,  -1 },
      {  -5, -17, -37, -17,  15, -23,  22,   1,  -1, -19, -32, -12,  34,  -7,  -2,  -4,   0, -10, -27, -12, -16,  19,  -2,   3,  14,  -7,   4,  33,  -8,   7,  -2,  -2,  10,  10, -36,  27,  -3, -10,   0,  -1,  -3,   0,  16, -15, -14,   4,  -2,   3,  34, -40, -11,  -1,   6,   7,  -4,   3, -22,  -9,  24,   9,   1,  -4,  -1,   3 },
      {   4,   1,   6,   0, -29, -53,  18,   6,   9,   6,   5,  19,   0,  33, -10,  -6,   1,  -8,   2, -14,   2, -10,   4,  -3,  10,   0,  -7,  -1,  -3,  11,   3,   1,  14,   9,  37, -17,  16,  -7, -11,  -1,  17, -36,   6,  10, -14, -12,   6,   2,  40,  24, -21, -30,  -2,  12,   1,  -3, -24, -21, -21,  13,  11,  -1,  -2,   3 },
      {  10,   2,  20,  10,  20,  48, -23,   0,   6, -11,   7,   6, -30, -40,  15,   8,  15,   2,  22,  -7,  -9,  10,  14,  -9,  20,  -3,  13,  24,  -8,  13, -16,   1,  15,  13, -11, -27,  -2,  -3,   0,   3,  38,  -2,  30,  -2,   1,  -3,  -3,  -5,  22,  -4, -31, -25,  -6,  -3,   6,   1, -18, -10,  -4,  16,  12,   2,   0,  -1 },
      {  -5,  -8, -16, -41, -17,   2,  -5,   5,   2,   2,   3, -33,  24,  -8,  26,  -8,   5,  13,  26,  19,  18,   2, -10, -13,  -3,  -7,  23, -12,  22, -25, -11,   6,  -3, -23,  13,   3, -50,  13,  20,   0,  20, -14, -19,  11,   8,  21, -15,  -2,  23,  12, -11,  -5,  12, -14,  -3,   3, -31,   2,  -6,  -4,   6,  -4,   7,  -1 },
      {   0,  -3, -18, -10,  41,  16,  -9,   4,  -5,  -9, -35, -39, -10,  -7,  -8,   8,   6,   9,  -2, -12,  17, -48,  43,  -6,   8,  13,  11, -19,   8,  34,  -3, -12,  -6,  12,  24,   9,  -1, -10, -22,   2, -20, -22, -20, -10, -19,   1,  10,   8,   1,  17,   3,  -1,   8,   7,  -3,  -3,  -5,   4,  -3,   4,   4,  -5,  -2,   0 },
      {  -1,   2,   7,   9,  -3, -19,  18,  -9,   4,  12,  24,  17, -14,   5,  -9,   6,   7,  25,  31,  26,   8,  -4,  -7,   2,   7,  16,  35,  -3, -30,   9,   4,  -5, -16,  25,   9,   7, -25,   1,  -8,   3, -22,  -8,  30, -46,  -5,  11,  -1,   1,   5, -26,  10,  -7,  21,  -1,  -1,   2, -16,  42,   7, -15,   4, -11,   3,   0 },
      {  -2,   2,  13,  -2,  -9, -17, -20,  18,  -4,  -1,  20,  -3, -20,  11,   6,  -2, -10, -10,   8, -11, -21,  -8,  23,  -9, -17, -25,   3,  -8,   1,  15, -10,  -4, -28, -40, -27,  -7,  -2,   8, -12,   6, -10, -49,   1, -22,  25,  -6,   1,   5,   1, -15, -22,  30,  14,  -7,   6,  -4,  -1, -12,  34,  21, -22,   4,  -1,   3 },
      {   4,  -3,   5,   4,  -2,   3, -25,  29,   7,  -2,  12,   9,  12,  40, -15, -12,   5,  -8,   3,  21, -28, -38,  39,   2,  13,   1,  21,  32,  26, -10, -11,  -1,  13, -15,  -4,  -1, -13,   5,  -4,   0,  21,  23,  10,  -6, -16,  13,   2,   4,  16, -16,  43,  -2,   1,  -4,  -6,  -3,  19, -15,  -8, -19,   2,   0,   4,   4 },
      {   2,   5,  11,  14,  21,  33,   9, -10,  -1,   0,   8,   3,  26,  17, -27,  -2,  -9,  -8,  -8,  -8, -26, -33,  -9,  24,   2,  -8,  -6,  -7,  12, -24,  39, -13,  -2,   9,  -4, -11,   4,  24,  -3, -11,  -3,  -3,   3,   0,  14,  -1, -15,  10,  -6,  10,  -5,   2,   1, -12,   6,   0, -72,   7,  11,   3,  -3,   4,   9,  -6 },
      {  -4,  -1, -10, -31, -13,  -1, -39,  30,  -1,   8,  -5, -12,  -4,  31,  25, -13,  -3,   2, -10, -25,  -1,  -1,  -3, -14,  -1,  14,  -4,   6,   9,  -9,  -9,  12,  -4,   9,  -6, -23,  31,  -7,   7,   0, -14,  17,  13,  -6,   2, -10,   3,  -2, -25, -26, -10, -11,   6,   8,   4,  -3, -26,  56, -28,  17,   9,  -5,   3,  -2 },
      {   0,   2,   8,  22,  34, -17, -42,   5,   2,   7,   9,  40,  34,  -4,  41, -11,  -5,  -4,   3,   6,   5,  -7,   0,  -7,  -2,   3,   6,  14, -31,  -5,  -9,  13,  -4,   3,   9,  26, -10,  15,  -2,   0, -19,  -7, -25,  17, -27,   2,  -2,  -4, -31,   1,  -4, -18,  12,   8,   1,   2, -18, -25,  11,  17,   9,  -6,  -3,  -1 },
      {   6,   4,   6,  15,  29,  -2, -21,  12,   5,   0,   5,   2,  15,   3,  29, -10,  -2, -12, -16, -23,  -9,   0, -16,  -4,   5,  -6,   2, -29, -25,  14,   2,  10,  12,   4,   7,  32,   6,  16,  -3,  -3,  16,  -4,   1, -13,  28, -12, -11,   5,  54,   9,  10,  24,  -5, -21,   9,  -5,  15,  29, -31, -28,  -9,   1,   0,  -4 },
      {   0,  -2,   1,   7, -15,  11,   4, -14,  -3,  -5,  -2,   2, -25,   5,  -5,  14,  -2,  -5,  -4,  13, -20,   3, -10,   2,  -4,  -5, -13,   5, -16,  23,   2,   2,  -3, -13, -26,  13,  -8, -11,  -1,   2,  -8,   6, -52, -19,  15,  16,  -7,   2,  10,  -8,  31, -22,  23,  -7,   0,   0, -18, -14, -60,  39,   9, -15,   7,  -2 },
      {   2,   2,   5, -16,   8,  40,  -5,  17,   0,   1,   7, -11,  -5,  18,  -9,  -8,   2,   1,  20,   0,   8,   8, -28,   9,  -1,  -1,  12,  21, -26, -31,  16,   6,   2,  -8,  19,  14,  53, -28,  18, -11,   0, -30,  -5, -14, -26,  23,  -4,   1,  18,  10,   6,  14,  19,   6, -13,   3,  17, -10,  13,  14, -16, -16,   4,  -2 },
      {   2,   4,   4,  15,   3,  -1,  21,  73,  -3,  -2,  -5,  16,  -6, -16, -33, -24,   3,   3,   1,  27,  19,  25,  32, -27,  -4,  -6, -18,   9, -12,  -5,   0,   9,  -1,   4, -16,  15,   2, -12,  -6,  -1,   0,   8, -19,   2,   6,  11,   1,  -1,  -1,  27, -12,  17,  -7,  -8,  -1,   0, -29,  19,  -4, -10,   7,   5,  -2,   1 },
      {  -2,  -2,  -5, -11, -17,  -4, -24, -20,   3,   5,   2,   1, -18,  25,  25,  18,   2,   1,  -1,  11,  -5,  -3, -12,   4,   5,   6,  -7,  23,  -1,  32, -14,  -6,   3,  11, -18,  19,   5, -27,   5,   8,  15,  22,  -7,  -6,   6,  -3,  17,  -5,   0,  59,  10,  11,  -1,   2,  -6,  -7, -37,   5,  41, -27, -19,  -7,   5,   3 },
      {   3,   5,   4,  19,  23,  -7,  16, -26,   3,   5,   4,  26,  35,   1,  -1,  10,  -1,  -4,  -6,   5,  28, -20,  14,  -4,  -5,  -8, -17,  -6,  14,  13, -26,   0,   1, -11, -15, -32,  11, -19,  24,  14,  14,   5,   5,  12,  -8,  33,   1,  -9,  14,   3, -11,  14,  32, -13, -21,  -1,   9,  38,  -3,  26, -21, -13,  12,   5 },
      {   0,   2,   6,   3,  12,   3,  11,   4,  -3,  -1,   6,  -5,  11,   0,   3, -14,  -3,   0,   2,  -4,  34,   7,   3,   2,  -4,  -1,   5, -28,   7,   3, -40,  17,   3,   6,  18,   4,  18, -21,  17,   3,   1,   3,  27, -24,  34,  -7,  12,  -9, -12, -21,  47,   3, -20,  -7, -10,  -3, -38, -47,  -2,  13, -12,   9,   4,   6 },
      {  -2,  -5,  -5,  -9, -12, -22, -34, -25,   1,  -1,  -1,  -7,   2, -11, -14,  29,   2,   3,  -1,  -5,  -7,  37,  57,  -9,   0,  -2,  -6, -16, -22, -46,   0, -20,   9,  14,  11,   4,  23,  23,  -8,  10,   3,   9,   0,  -8,   1,  14, -10,   2,  10,  14,  24, -12,   1, -22,   3,   5,  -4,   8,  14,  21, -14,   0,   3,  -9 },
      {  -2,  -4,  -3,   2,   0, -22, -29, -19,   0,   0,   3,   9,   8, -27,   6,  23,   1,   0,   5,  17,  26,  -2,  -6,  -6,  -1,  -4,   2,  14,  22, -14,  26, -17,  -7, -12, -14,  -9,  16, -14, -14,  -6,   5, -11,   6, -28, -30,  -2,   5,   8,   0,  -4,  -1,  57, -24,  12,   5,   0, -22, -12, -46, -13,  11,   9,  -8,  -2 },
      {   0,   2,   1,   0,   4,   5, -11, -17,   1,   4,   5,   1,   9,  28, -20,  17,   1,   5,  10,  10,  -5,   1,  13,  -1,   0,   5,  18,  22,  -5,   6,   6, -11,  -3,   0,  20,  47, -13, -36,   5,   7,  -2, -10,   8,  42,  36,   9,  -1,  -8,  -6, -18, -25,  15, -49,  -3,   6,   0,   4,  18, -14,  32,  -3,  11,  -1,  -1 },
      {   2,   3,   2,   3,  -7,   2,  25, -14,   2,   1,   3,   8,  -7, -21,  15, -21,   3,   6,   8,  20,  10, -13,   5,  16,   3,   3,  14,  31,  38, -16, -22,  14,   0,   0,  -3,  25,  23,  30, -14,  -8,  -7,  -2, -15,  -4,   1, -46,   2,   9,  21,  13,  16,   0,  -8,  11,  16,  -7,   3,  32,   7,  44, -18,   1, -14,  -3 },
      {   1,   1,   3,   2,   9, -17, -25,  35,   2,   2,   3,   3,  15, -23,   3,   1,   3,   1,   4,   2,  18,   7, -27,  -3,   2,   2,   0,  -5,  22,  18,  56, -22,   3,  -7,   3, -18, -15, -29, -25,  -2,   5,  14,   7,   6,  22,   3,   1,  10,  12,   2,  33, -29,   0,   9,   5,  -6,   9,   7,  25,  34, -25, -10,  -6,  -1 },
      {   0,   1,  -1,  -2, -19, -29,  15,  13,  -2,  -1,  -6,  -4, -35, -36,  29, -19,   0,   0,  -3,  -5, -19, -35,  -1,   9,   3,  10,   3,  10,  -7,  16,  10,  -6,   4,  11,  17,   7,  14,   7,  12, -15,   0,  12,  23,  30, -10,  38, -32,   6, -13,  -3,  11,  29,   0, -33,  -6,  12,  -9,  -8,  -3,  15, -14,   6,  16,  -9 },
      {  -1,  -2,  -2,   2,  -1, -15, -17,  -6,   0,  -2,  -3,   2,  -3,  -8,   8, -32,   1,  -1,  -2,   6,   6, -20,  14,  57,  -4,  -8, -14, -10, -13,  -9,  -4,  -1,  -3,  -5, -15, -13,  -5, -26,  28, -24,  -1,  -7, -19, -38,   4,  23,  -6,   1,   6,   7, -10, -31, -49,   1,   6,  16,  13,  17,  14,  -1,  20,  18,  -1,  -8 },
      {   0,  -1,   0,   5,   2,  -3,  24,  10,   1,   1,   2,   4,   3,  -5,  15,  45,  -4,  -6,  -7,   0, -16, -39, -11, -65,   2,   1,   0,   0,   9, -15,   7,  17,   1,   2,  -1,   3,  17,   1,  20,  13,   0,  -3,  -5, -29,  13,  14, -29, -20,   0,   8,   8, -12, -31,  19,   7,  12,  10,   2,  23,   6,  24,  -1,  -7,   3 },
      {   0,   0,   0,  -1,  -4,  10,  12,  11,   0,   0,  -1,  -4,  -6,   4,  15,  -6,  -2,  -3,  -5, -11, -21,  -3,  -5,  15,  -1,  -4,  -7, -23, -28, -21,  -7,  -4,   2,   5,   1, -10, -38,  10, -16,  16,  11,   8,  14,   1, -19,  14,  34,  -9,   6,  16,  17,  42, -27,  34, -18, -14,  -4,  10,   3,  45,  22, -28, -10,  16 },
      {  -1,  -2,   0,   1,  -2, -11,   1, -18,  -3,  -4,  -2,  -3, -13, -33,   2, -25,  -2,  -5,  -2,  -2, -17, -39,   4, -11,  -1,  -4,  -1,   0, -15, -51,  12,  11,   2,  -1,  -2,  10,   0, -33, -17,  23,   4,   2,  -3,   7,  25,   2,  35, -22,   1, -10, -12, -19,  18,  22, -11,  -8,  -6,   0, -16, -27, -42,   2,  -4,   3 },
      {   1,   1,   1,   2,   1,  11,  27,   3,   1,   1,   1,   4,   5,  12,  49,  -4,   0,   0,   2,   6,   9,   9,   4,  15,  -1,  -3,  -1,   7,  17,  -8,  -3, -35,  -1,  -1,   2,   6,  34,   3, -53,  32,  -2,  -4,  -6,  -4,  22,  41,  10,  14,  -7,  -9, -13, -16,  -9, -26,  -2, -16,   6,  -5,   3, -17,  30, -15,   1,   9 },
      {   0,  -1,  -2,  -3,  -1,  10,  -3, -47,   0,   0,  -2,  -4,  -1,  19,  -3, -53,   0,   0,  -1,  -2,   3,  11,   4, -38,   2,   2,   1,  -2,   3,  13,  28,  56,  -3,  -3,  -5, -11,  -5, -12, -39,  -2,   1,   3,   4,  -6, -18,  18,  -1, -15,   2,   6,   9,   8, -10, -33,  20,   7,   1,   0,  14,  11,  10,   8,  -7,   1 },
      {   1,   1,   0,  -2,   4,  10,  18,  24,   1,   2,   0,  -3,   0,  14,  11,  28,  -1,  -2,  -2,  -4,  -1,  -4, -22,   6,  -2,  -1,  -3,  -3, -14, -11, -21,  -9,  -2,  -3,  -2, -10, -18, -13, -17,  26,  -4,   0,  -7,  -3, -49,  -1,   6, -10,   1,  -3,   4, -13, -17, -36,  35,  -2,  -5,   5,  -7,   6, -45,  57, -10, -12 },
      {   1,   2,   1,   0,   1,  16,  28, -11,   2,   3,   1,   5,   2,  22,  39, -16,   2,   1,   0,   0,   1,  14,  40,  -9,  -1,  -1,  -6,  -7, -13,  -8,   7, -13,  -1,  -1,  -2, -11, -17, -32,   0,  -1,   1,   5,   5,   1,  -4, -14, -60,  36,  -6,   3,   5,   7,   1,  46,  22,  -1,  -2,  -7,  -9,  -9, -28, -10,   6,  -7 },
      {   0,   0,   0,   0,   0,   1,  12,   5,   1,   1,   1,   1,   4,   5,  13,  46,   1,   0,  -1,  -2,   1,  -2,   6,  -2,   0,   0,   0,  -1,  -6, -12, -11,  45,  -3,  -3,  -5,  -4,  -6,  -4, -32, -78,   1,   4,   4,   7,   9,  19,  31,  40,  -2,  -3,  -8,  -7, -11,  -3, -10,   4,   0,   0,  -1,  -4, -18, -12,   8, -15 },
      {  -1,  -2,   1,   1,   0,  -7, -20, -11,  -2,  -3,  -1,  -1,  -2, -12, -43, -25,  -2,  -4,  -2,  -4,  -4, -14, -44, -31,  -1,  -3,  -2,  -3,  -5,  -9, -54, -20,   1,   0,   0,  -1,  -2,  -7, -18,   6,   0,   2,  -3,  -1,  -9,   4, -23,  49,  -5,   1,  -6,  -7, -27,  -5,   4, -23,   3,   1,   5,  -2,  -5, -30,  24,  -9 },
      {   0,   0,  -1,  -1,  -1,  -8,   0,   4,   1,   1,   0,   0,   2, -10,  -4,  17,   0,   0,  -2,  -1,  -2, -19,  -1,  18,  -1,   0,  -4,  -3,  -8, -27, -15,  10,   1,   3,   2,  -1,  -4, -50,  -4,  -2,   1,   6,  10,  16,  15, -24,  -7,   9,   4,   4,  10,  20,  46, -26,  49, -11,   5,  -4,  15,   5,  58,  -5, -22, -12 },
      {   1,   1,   2,   4,   3,   5,  14,  -7,   1,   0,   1,   3,   0,   8,  22, -17,   0,  -1,   0,   2,  -5,  -3,   3, -38,   0,  -2,   0,   1,  -7, -12,  13, -42,   0,  -2,  -2,   1,  -9, -18,  14, -10,   1,  -2,  -3,  -5,  -4, -45,  39,  19,   2,   0,   7,   0,   2, -30, -38,  35,   7,   3,   6,  13,  24,  18,  36, -32 },
      {   0,   0,   0,   1,  -1,   2,   1, -12,   0,  -1,   0,   0,  -5,   0,   4,  -8,   0,  -1,   0,   0,  -4,   0,   2, -23,   1,   1,   2,   0,  -7,   3,  14,  -5,   2,   1,   3,   3,   2,   7,  42,  -6,   1,   4,  -1,   0,   1,  21,  25,  62,   0,  -1,   4,  -6,  11,   2,   1, -54,  -2,   4,   3,   6,   8,  50, -36,  40 },
      {   0,   0,  -1,  -1,  -2,  -4,   3,   7,   0,   1,   0,   0,  -1, -10,   8,  18,   0,   1,   1,   3,   0,  -1,  16,  17,   0,   0,   0,   1,   1,  -5,  32,  39,  -1,  -1,  -1,   0,  -6, -11,  17,  23,  -4,  -5,  -5,  -7, -18, -34,  -6,   4,  -3,  -5,  -5,  -6, -22, -34, -11, -58,   4,  -1,   3,  -1,  -1, -24,  60,  26 },
      {   0,   0,   0,   1,  -1,   3,   4,  -2,   0,   0,   0,   1,  -1,   2,   5,  -3,  -1,  -1,  -2,  -1,  -5,  -4,   2, -12,  -1,   0,  -3,  -1,  -4, -10,  -6, -34,  -1,  -1,  -1,  -3,  -6, -14,  -1, -46,  -4,  -1,  -3,  -3, -10, -25, -17, -27,  -5,  -3,  -2,  -9, -10, -42, -13,   5,  -2,   1,  -5,  -4, -15, -37, -47,  68 },
      {   0,   1,   0,   1,   1,   1,  10,   0,   0,   0,  -1,   1,   2,   1,   4, -11,   0,   1,  -1,   0,   1,   0,  11, -14,   0,   1,   1,   1,   2,   8,  13, -18,  -1,   1,  -1,  -1,   2,  15,  45,  -4,  -1,  -2,   2,  -3,  -5,  12,  50,  -6,  -2,  -5,  -3,  -1, -14, -11,  72,  -6,   0, -10,  -5,  -7, -18, -58,  -5, -30 },
      {   0,   0,   0,   0,  -2,  -1,  -1,   3,   0,   0,   0,   0,  -1,   2,  -1,   5,   0,   0,   0,   0,  -2,  -2,   4,  10,   1,   0,   1,   1,   3,   4,  10,  28,   0,   0,  -2,  -3,  -4,  -3,  14,  42,   0,  -2,   1,  -2,  -4, -12,  -9,  36,  -6,  -3,  -6,  -4, -15, -26, -46,  12,  -5,  -3,  -3,  -4, -10, -30, -72, -46 },
      {   0,   0,   1,   0,   0,  -1,  -5,   5,   0,   0,   0,  -1,   0,  -5,  -4,   6,   0,   0,   0,   0,   0,  -2,  -3,  10,   0,   1,   0,   0,   2,  -3,   2,  20,   1,   2,   3,   2,   1,  -1,   4,  36,   0,   1,   0,   4,   1,  -2,  18,  49,   0,   2,  -1,   2,  -1,  -2,  26,  81,  -1,   2,  -2,  -1,  -6, -10,  16,  63 },
    },
  },
  { // 14
    {
      {-108,  40,   5,   5,   1,   2,   1,   1,  52,  -8,  -7,  -2,  -2,  -1,  -1,   0,  12, -11,   1,   0,   0,   0,   0,   0,  -3,   1,   2,   0,   0,   0,   0,   0,   2,   1,  -1,  -1,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  43,   7, -12,  -1,  -3,   0,  -1,   0,  89, -38,  -6,  -2,  -1,   0,  -1,   0, -63,  -8,  15,   1,   3,   1,   1,   0, -20,  16,   3,  -1,   0,   0,   0,   0,  -3,   5,  -3,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,  -2,   2,   0,   0,   0,   0,   0,   0 },
      { -29,   5,   0,   2,   0,   1,   0,   0, -33, -19,  13,   0,   4,   0,   1,   1, -78,  49,   7,   1,   1,   0,   1,   0,  62,  16, -19,   0,  -4,   0,  -2,   0,  19, -22,  -3,   2,  -1,   1,   0,   0,   1,  -2,   4,   0,   0,   0,   0,   0,   2,   0,  -1,  -1,   0,   0,   0,   0,   3,  -2,  -2,   1,   0,   0,   0,   0 },
      {  20,  19,  -9,   0,  -4,   1,   0,   0,  34,  -2,  -4,  -1,  -2,   0,  -1,  -1,  25,  21, -15,  -2,  -3,   0,  -2,   0,  64, -58,  -9,   2,  -1,   1,   0,   1, -55, -26,  27,   0,   6,   0,   1,   0, -18,  24,   5,  -4,   0,   0,   1,   0,  -1,   2,  -4,   1,   0,  -1,  -1,   1,  -2,   2,   1,   0,  -1,   0,  -1,   0 },
      { -26,-100,  36,   3,  11,   2,   4,   1,  27, -12,  17,  -7,   2,  -3,   1,  -1,   3,  41, -15,  -4,  -2,  -2,   0,  -1, -17,  -8,  -8,   8,   0,   2,   0,   0, -15,   5,  11,   1,   0,   0,   0,  -1,   0,  12,  -1,  -3,   0,   0,   0,   0,   2,   4,  -4,   0,   0,   0,   0,  -1,   2,   1,  -1,   1,   0,   0,   0,   0 },
      {  12, -13,  11,  -4,   0,  -1,   1,   0,  38,  53, -18,  -5,  -5,  -2,  -2,  -1,  17,   4, -11,  -1,  -2,   0,  -1,   0,  15,  20, -20,   3,  -4,   2,  -1,   0,  59, -57, -11,   5,  -1,   1,  -1,   1, -39, -25,  23,   1,   3,   1,   1,   1,  -3,  12,   3,  -3,   0,   0,   0,   0,  -4,   0,  -1,   0,   0,   0,  -1,   0 },
      {  -2, -22, -12,   3,   1,   3,   1,   1, -11, -80,   7,  10,   6,   4,   1,   3,  41, -35,  28,  -2,   2,  -1,   0,  -1,  26,  38,  -2, -10,  -3,  -4,  -1,  -1,  -1, -20,  -9,   3,   0,   1,   1,   0, -38,   7,  18,   0,   2,   0,   0,  -1,  -5,  23,  -3,  -5,  -1,  -1,   0,  -1,   2,  -1,  -7,   2,   1,   1,   0,   0 },
      {  17,  22, -12,  -2,  -2,  -1,  -1,   0,  16, -34,  23,  -4,   3,  -1,   1,  -1,  51,  63, -10, -11,  -4,  -4,  -3,  -2,   1,  14, -25,   0,  -1,  -1,   0,   0,  16,  11, -21,   6,  -3,   1,  -1,   1,  39, -44, -12,   6,  -3,   2,  -1,   1, -22, -21,  18,   0,   2,   1,   1,   1,  -5,   6,   3,  -2,   1,  -1,   0,   0 },
      {   3, -30,   8,  -2,   0,   3,   4,   4,  19,  21, -31,   2,   1,   4,  -5,   1,  13, -19,  16,   9,   1,   1,  -4,   1,  66,  32,  23, -15,  -6,  -7,  -1,  -1,   2,  31, -19, -15,   5,   0,  -1,   3,  28,   3, -32,   2,  -6,   4,   1,   0,  11, -45,  -9,   6,  -3,   0,  -3,   7, -21,  -9,   6,   4,   0,   5,  -7,  -2 },
      {   9,  27,  19, -11,  -3,  -1,   1,   1,  -1,   8,   5,  -8,   2,   0,  -2,   0,  16,  30,  16, -11,   0,  -4,  -2,  -2, -29,  51, -23,  -4,  -2,  -2,   1,   0, -23, -30,   3,   6,   6,   3,   0,   3,   6,  52,  -1,  -6,  -6,   0,   0,   0,  50, -18, -36,  -1,  -3,   0,  -3,   5, -11, -23,   1,   8,   1,   4,  -5,  -2 },
      {  14,  27,  96, -29,   5, -14,  -3,  -7,  -1, -13,  25, -17,   1,  -5,   3,  -2,  -5, -30, -37,   8,   3,   5,   4,   1,  18,  17,  -7,   5,  -3,   0,   0,  -1, -16,  10,  -7,   0,  -6,   0,   1,  -2,  -6, -13,  -3,   6,   3,   0,   0,   1, -10,   5,   9,   2,   1,   1,   1,  -4,   6,   6,   2,  -4,   0,  -2,   4,   1 },
      {  -9, -10,  -5,   5,  -1,   3,   0,   1,  -3,  35,   1,   0,  -4,   2,  -1,  -4, -10,  -4,  36,  -9,   3,  -2,   1,   0,  -5,  24, -18, -17,   2,  -5,   2,  -2, -79, -16, -35,  17,   4,   3,   2,   2,   2, -35,  17,  22,  -5,   4,  -1,   2,  -8,   4,  34,  -7,  -1,  -1,  -1,   0,   1,  24,   3, -12,   1,  -2,   0,  -1 },
      {   3,   5, -15,   7,  -3,   4,  -2,   1,  17,  36,  84, -13,  -6,  -8,  -1,  -3,   3,  -1,  25, -26,   1,  -4,   2,  -1,  19,   3,  -2,  -9,   2,  -2,   3,  -1,  10,  46, -14,  -4,  -2,  -2,   0,  -2, -33,  12, -14,   2,   1,   1,   1,   0,   0,  21, -15,   0,  -1,   0,   0,  -1,  18, -15, -12,   4,   0,   0,   1,   1 },
      {   0,  -9,  15,   7,  -3,   5,   1,   2,   0,  -4,  38,   2,   2,   1,  -3,  -2,  -3, -45,  37, -26,   7,  -2,  -1,   1,   0, -36, -43,   3,   7,   3,   4,   0,  27, -28,  18,  16,   4,   0,  -3,   1,  28,  11,  -3,  -6,  -9,   1,  -1,   0,  -3, -41,  10,   3,   0,  -1,  -1,   4, -32,  26,  19,  -6,  -1,   0,  -5,  -3 },
      {  -4, -15, -47,   2,  15,  -6,  -4,  -4,   5,   6,  12,  -8,  -6,  -8,   6,   2,  -4, -40, -58,   4,   5,   5,   9,  -1,  16,  26, -42,  25,   6,   2,  -3,   0, -14,   3,   4,  26, -21,   0,   4,  -5,  35,   3,  16,  -2,   1,  -5,  -5,  -2,  10,  -3,  -7,  -7,   2,   6,   4, -12,  11, -30,   6,   3,   8,  -2,  16,   8 },
      {   5,   6,  19,  91, -33,   5,  -9,   1,  -5,  -7, -16,  22, -11,   4,  -2,   0,  -5,   1, -37, -40,  12,   2,   3,   2,  -2,  11, -10,  -1,   9,  -6,   2,  -2,  -3,  16,  -4,   5,   1,  -2,   1,  -1, -31,   1,  -5,   6,  -6,   2,   0,   1,   9, -19,   7,   3,  -2,   2,  -1,   2,  -3,   0,  11,  -5,   1,   1,  -1,  -1 },
      { -10,  -2, -29, -46,   8,   3,   7,  -3, -11,  -8, -13, -23,  10,   0,   2,  -1,  -7,   3, -16,  -5,   4,  -1,   0,   0, -12,   1, -29,  11,   2,   2,   3,   1,  -5,  19,  -3,  -4,   2,  -3,   1,  -1, -72,   1, -31,  14,   4,   3,   1,   0,   4, -42,  17,  24,  -3,   3,  -1,   2, -17,   8,  31,  -7,  -2,  -1,   0,   0 },
      {  -3,  -1, -18,  22,  -1,   2,  -3,   1,   1,   3,  59,  14,  -4,  -6,  -1,  -4,  -3,  -6, -29,  38, -15,   7,  -4,   0,  -9,  18,  44,  -5, -19,   3,  -2,   1, -15, -48,   8, -36,   8,   0,   2,   1,  -4, -13, -17,   0,  15,  -1,   3,   1,  -2, -20,   2,  13,  -1,  -1,   1,   1, -24,   4,  26,  -1,  -2,   1,  -1,   0 },
      {   0,   6,   0,   4,  18,  -8,  -1,  -7,   0,  20, -11,  33,  -7,  -3,   0,  -2,   2,   8,  10,  -2,  -4,   0,   2,   0,  -1,  44, -20,   8,  -7,   2,  -1,   1,  -7,  12,  45,  -9,  -8,  -2,   2,  -2,  -3,  30, -12, -28,   7,  -5,  -1,  -3, -77,   0,  -6,   4,   9,   2,   5,  -4,   5,  32,   3,   1,   1,  -2,   6,   2 },
      {  -3,  -1, -11,   4, -34,  19,   2,  10,  -6,   1,  -2, -86,  15,   3,   1,   0,  -2,  -4, -12, -28,  14,  -4,   1,   1,   2,  21,  32,  22,   1,  -6,   0,  -2,  -8,  -7,  23,   4,   5,   0,  -5,   4,   6,  -6,   9, -13, -11,   4,   4,   0, -29,   4, -11,  -8,   3,  -4,  -3,  10, -25,  17, -18,   9,   0,   5, -12,  -5 },
      {   0,  -1,   6,  12,  28, -11,  -3,  -5,  -5,  -6, -11, -25,  -4,  -6,   5,   2,  -6,  -3,   3, -24,  -6,   4,   3,  -1, -10, -23,  11, -22,   7,  -6,   1,   1,  -4, -30, -52,   7,  -4,   1,   7,  -1,   2,  18, -13,   7,   7,  -1,  -1,  -3, -60, -13, -25,  14,   1,   6,   4,  -5,   5, -47,  22,  21,   4,   3,   8,   4 },
      {  -4,   1,   1, -15, -26,  12,   3,   5,  -1,   4,  11,  31,  17,   1,  -7,  -1,  -7,   4, -21,  36,   2,  -3,  -7,   1,  -8,  -9, -14, -28,  -1,  -8,   5,   2,  -2,  26, -20,  19,   7,   3,  -1,   3, -15,   8,  41,  -9, -22,   4,  -1,  -2, -31,  -8, -17, -26,   2,  -1,  -2,   9, -62, -16, -12,  15,   7,   7,  -7,  -4 },
      {  -4,  -9,   3,  -6, -70,  18,  -1,  13,  -2,  -2,  -5,  -4,  -7,   5,  -4,   4,  -1, -14,   5,  30,   1,  -2,  -5,  -1,  -8,  -6, -46,  -9, -12,  10,   0,   2,  -7,  -9,   1, -23,  16,   8,  -4,   4,   5, -19, -21,  -9,   9,   1,   4,   1, -16,  -7, -25,  20,   4,  -1,  -1,   7,  53, -17,  -6,  15,  -9,   3, -11,  -4 },
      {   6,   8,  11,  52,  21,  -4, -12,  -2,  -1,   2,  -6, -40,  15,  -6,   7,   0,   7,   6,  40,  66,  -7,  -3,   1,  -4,   0,  -8, -24,  44,  -8,   7,  -6,   2,  -4,  19,  -9,  -9, -18,   2,   1,   0, -14,  -6,  14, -14,   4,  -2,  -2,   1,   4,  -6,  -2,  -1,   4,   1,   3,  -4, -15,  -8,  10,   3,   1,  -2,   7,   2 },
      {  -1,   1,   8,  -7,  20,  -2,   7,  -3,  -3,  -2,  -3,  24,   3,   3,  -1,   2,  -7,  -4,  17, -39,  -7,   0,  -3,   1,  -1,  -2,   0,  15,   2,   7,   0,   2, -21,   4,  35,  -9,  -6,   2,  -4,   1, -12, -75,   0, -30,   4,   1,   2,   1,  10,   0, -34,   7,  10,   1,   0,   0, -26, -42,   0,  24,   1,   2,  -2,   0 },
      {  -4,  -5,   4, -22, -50, -17,  -4,  -5,  -1,  -2,   5,  10, -71,   4,   4,   4,   6,  13,  24,  13,  23,   3,  14,  -5,   4,   1,  28,  35,  26,  -6,  -7,  -4,  -1,  -4,   1,  23, -31,  -7,   4,  -6,  -7,   1,   2,   1,   1,  -1,  -2,   0,  -7,  -9,   3,  -4,   8,   7,   3, -11, -11, -16,  22,  -5,   7,  -6,  15,   7 },
      {   1,   2,   4,  -4,   1,  12,   4,   0,   5,   8,  11,  30,  67,  -5,  -1,  -4,   3,   3,   6,   6,  50, -10,   1,  -1,  -1,  -2,  29,  35,  -9, -20,  -2,  -3,  -2, -15,  -1,  29, -19,   2,   1,   4,  -9,  -2,  -2,  -1, -20,   8,  -1,   1,  -9, -31,   2,  -1,  -1,   3,  -3,   5,  45, -12,   0,  -3,   0,   3,  -5,   0 },
      {   2,   3,   2,   4,  25,  13,   3,   8,  -1,  -4,   8, -17, -31,  14,  -7,   2,   0,   2,   1,  15, -45,   9,  -5,   6,   1,   6,  18, -22,  19,  -3,   4,  -1,   4,   9,  27,  34,  15,  -5,  -4,   1, -17, -13,  37,  -4, -25,   3,  -1,   2,  -2, -53,  -9,  -7,   2,   4,  -2,   9,  48,   5,   4,   2,   0,   2, -10,  -4 },
      {   1,   0,   4,   0, -18,  15, -10,  -4,   3,   4,  -4, -13,  35,  -8,   8, -10,   4,   8,   8,   6,   9,   0,   4,   1,   5,   6,   5, -54,  11,   1,   1,   0,   6,  13,  33,  16, -20,  -4,   3,  -8,  10,  -8,  19,   6,  10,  -7,  -3,  -1,   5,  27, -13,  19,   1,   1,   3,  -9,   1,  -3,  78, -14,  -9,  -6,  13,   5 },
      {   2,   3,  10,  -9,  -3,  86, -21,   9,   0,   0,  -4,  14,  -4,   4,  -5,   3,  -2,  -3,   1, -10, -52, -17,   2,   4,   3,   3,   8,  36, -26,  12,   3,   0,  -1,   1, -13,  29,   1,  -4,   3,  -5,   3,  11, -15,  -1,   7,  -6,  -3,   0,  -4,  18,  15,   3,  -1,  -7,   0,  -4,  -7, -15,  14,  -6,   0,  -1,   5,   0 },
      {  -3,  -5,  -5,   0, -35, -56,  -6,  -2,   2,   1,  -2,   6,  36, -21,  11,   0,  -1,  -4,   2,  -4, -62,  27,   2,   2,   3,   0,   9,  10,  -4,  15,  -6,   0,  -2,  -3, -16,  48,   0,  -2,  -3,  -2,  -7,  -3, -28, -19,  -9,  -2,   1,   2,   8,   5, -16,  -2,  10,   3,   3,  -2,  -1,  25,   0,   2,   0,  -3,   0,   0 },
      {   1,   1,   1,  11,  20,  19, -19,   7,  -2,  -3,  -4,  -8, -12, -11,  -4,   4,  -1,  -1,  -4,  28,  20, -28,  12,  -2,  -4,  -3,  -5, -41,  11,   4,  -1,   0,  -1, -15,  19,  41, -30,  -2,   5,  -6, -11, -16, -64,  11,  -4,  -1,  -1,   1,   8,   4, -15, -14,  11,   5,   2,  -4, -12,  15, -39,  11,   6,  -4,   7,   3 },
      {  -2,  -2,   4,   5, -16, -24,  -9,  -7,   0,   3,   0, -10,  15,  10,  -4,   5,   1,   0,  15,   1, -16,  18,   3,   3,  -4,   5,  -3, -15, -35,  -1,   0,   3,   5,   0,  50,  10,   2,  -1,   6,  -3,  -2,   5,  -6,  44,   3,  -3,  -4,  -5, -13,  -9,  52,  -8, -16,  -1,   0,  -3,  -8, -60, -26,  -3,  11,   5,   9,   3 },
      {   1,   2,   1,   9,  12,  -3,  13,  -2,   2,   3,   5,   7,   8,  41,  -7,  -1,   1,   0,   1,  24,   4,  13,  -5,   2,   3,   9,  -8,  28,  58,  -7,   1,  -1,   1,  -3,   9,  21,  50, -17,  -5,   0,   6,  -4, -30,  30, -18,  -8,   5,  -2, -12,  37,  -9,  18, -21,  -1,  -8,   9, -10, -11,  11,   3, -12,   7, -16,  -4 },
      {   1,   1,   3,   4,  -7,   3,  32, -10,  -2,  -2,  -4,   3,  -2, -91,   6,  -1,   1,  -1,   4,   4,  -3, -40,  -1,   3,   4,   8,   7,   2,  37,  35,   3,   0,   0,   0,   8,   2,  29,  -2, -11,  -2,   1,   3,   7,   2,   1,  -4,   0,   0,  -5,  -5,  25,  11,  -9,   3,   0,   1,   0, -18,  -8,   5,  -8,   4,  -3,  -2 },
      {   0,   1,  -2,   6,   8, -15,  18,  -7,  -2,  -3,  -3,   2, -26, -14,   3,   1,   0,  -1,   1,  -1,  19, -19,  -3,   4,   0,  -5,   2,  16, -70,  13,  -3,   6,   3,  12,   6,  29,  28,  13,  -6,   1,   1, -10,   6,  53,   3,  -1,  -1,  -2,  -5,   4, -49,  -8, -12,  -2,   0,   6,  -1,  18,  21,  -3,   1,   2,  -9,  -2 },
      {   2,   3,   2,  11,  -4,   7,  92, -19,   0,   0,   0,  -8, -10,  14,  24,  -5,  -2,  -4,  -3,   8, -13,  -9, -27,   8,  -2,  -2,  -6,  -9, -16, -34,   4,   4,   1,  -3,   2,  11, -22,   7,   4,   6,   1,   3, -24, -24,  -7,  17,  -2,   1,   4,  18,  19, -15,   9,   0,  -8,   3,   2,  -9,  15,   0,   5,   1,  -2,  -1 },
      {   0,   0,  -1,   3,   4, -21,  -8,  -1,   3,   3,   6,  -3,  -3,  44,  16,   7,   1,   1,  -6,   3,  24, -37,   6,   0,   3,   4,  12, -13, -16,  28,  -3,   8,   0,   7,  -7,  38,  30,  19, -10,   0,  -1,   7,   7, -43,  29, -16,  -7,   1,   7,  -2,  37,  27,   8, -11,   6,  -2,   0,  -3,   3,  38, -15,  -6,  -3,  -5 },
      {  -2,  -3,  -3,  -5,  -2, -37, -34,  15,  -1,  -2,  -2,  -1,  -1,   3, -23,  -2,  -3,  -6,  -5,  -2, -27, -81,   6,  -4,   0,   0,  -4,   6,   4, -42,   8,  -4,   1,  -1,   7, -18,   2,  19,  -1,   4,   1,  -1,  -7,  -2, -24,  17,   9,   4,  -2,  13,   8, -22,  -4,   8,  -8,   7,   6,  -9,  28,  -3,  -2,   2, -10,  -4 },
      {   0,   0,   2,   0, -11,   4,  21,  -2,   3,   4,   3,   9,  20,  37,  65,  -2,  -1,  -2,  -1,  -4, -27, -42,  36,  -3,   1,   0,   3,   2,  20,  13, -18,   1,  -1,  -3,   1, -22, -14,  10,  -5,  -4,   0,   0,  14,  38,  -1, -18,  -8,   2,  -6, -19, -22,   7,   3,   5,   2, -12,  -4,   6, -23, -13,   2, -10,  14,   6 },
      {  -1,  -3,  -1,   2, -14,  -1,   0, -68,   0,   0,   2,  -4,   9,  10, -51,   4,  -1,  -1,   0,   1, -14, -27, -37,  11,   1,  -1,   4,   5,  11, -23,  17,  10,   0,   0,  -3,   7, -10,   1,  21, -10,   3,  -2,  14,  12,  28, -11,  -7,  -9,   4, -12, -16,  36,  -2,  -4,   8,  -8,  -2,  16, -25,   1,   2,   0,  18,   7 },
      {  -1,  -3,  -1,  -4,  -4,  14, -21, -70,   0,   0,  -2,  -5,   3,  -1,  26,  -8,   1,   1,   3,   1,   8,   2,  32,  14,  -1,  -2,  -3,  -4,  -2, -15,   7,   1,   1,  -3,  -2,  -9,  49, -24,   2,   0,  -3, -10, -14, -26,   4, -10,  -3,  -2,   4,  -8,  -3, -59,  -7,   8,   8,  -7,   9,  -1,  10,  -2,  13,   0,   8,   5 },
      {   0,   0,   0,   0,  -4,  -4,  17,  -8,   0,   1,   2,  -4,  13,  29, -44,  13,  -1,   0,  -2,  -6,   4, -10, -25,   1,   2,   3,   6,   0,  10,  80,  -2,   4,  -3,  -4,  -6, -18, -15,   5,  -2,  -1,  -3,   3, -10,   6, -24, -18,  -5,   0,  -4,  -1,   2, -52,   8,  10,   6,  -3,   9,  -2,  22,   0,  13,  -4,   7,   3 },
      {  -2,  -2,  -5,   0,   4,  -3, -18, -50,  -1,  -2,   0,  -3, -15,   9,   5, -25,   0,  -1,  -4,   3,   3,  -3,  15,   9,  -1,   0,   0,  -9, -16,  35,  -1,  -2,  -2,  -3,  -5,   2, -25,   1, -12,  10,   2,   0,   0, -21, -52,  20,   8,   5,  -6,  10,  -1,  40, -19,  -6, -13,  18,  -4, -15, -12, -45,  -9,   6, -27, -11 },
      {   0,   0,  -1,   0,  -2,   8,   6, -27,   1,   1,   1,   4,   1,  -5, -12,  19,   2,   2,   1,   1,  -6,  12,  46,  14,   0,   0,  -2,   3, -10,  -1,   7,  -3,   2,   0,   2, -15, -16,  -7,  -6,  -3,   4,   6,   8,  27, -39,  18,   3,   0,   8,  10,  15,  19,  20,   5,  -6,   2,   3,  18,  13,  84,   1, -11,  -5,  -2 },
      {  -1,  -2,   0,  -5,  -1,   6, -41, -14,   1,   2,   1,   3,  -7,  -8,  57,   6,  -1,  -2,   0,  -5,   4,  -1, -83,   3,   0,   2,   0,   2,   6,   0, -34,   9,   1,   0,   5,  -7,  -5,   5,  17,   0,  -1,   1,  -1,  12, -28,   1,  -4,  -7,   2,   4,   8,  -1, -10,   6,   4,   6,   3,   7,   5,  34,   5,   3,  -4,  -2 },
      {   0,   0,  -1,  -1,   1,   9, -10, -19,  -1,  -2,  -2,   2,   0,  -1,  -2, -63,   0,   0,   1,   0,  -6,  21,   4, -13,   2,   2,   3,   7,  26,   6,   3,  18,   3,   2,   5,   5,  -5,  76, -15,  17,  -1,   5,  -5,  19,  30,  24,  -3,   5,  -2,  -3,   2, -23,  14, -21, -12,   8,   2,   0,   3,  17,  -3,   2,  -9,  -5 },
      {   0,   1,   0,   0,   3,   5,  -4, -12,   2,   3,   4,   0,   3,  -5,   7,  92,   1,   3,   0,   2,   2,  14,  13,  31,  -1,   0,   1,  -1,  15,  -2,  -1, -11,   0,   1,   2,   1,  -2,  64,   4,   2,  -1,  -2,  -4,  -7,   9,  16, -11, -11,  -3,   2,  -4,  -5, -17, -11,  -7,   1,  -2,   1,   5, -22,  -4,  11,  -9,  -5 },
      {   0,   0,   0,   0,   0,   4,   8, -15,   0,   0,   2,   0,   1,   0, -27,   4,  -1,  -1,  -2,  -4,  -7,  -5,   9,  -2,  -3,  -3,  -4,  -7,   2, -11,-102, -13,  -2,   2,  -1,   5, -13, -19, -27,  10,  -1,   2,   2,   0,  25,   2,  28,   1,  -1,  -5,   2, -14, -19, -15,   3,   2,   0,   1,   2,   7, -25,   5,  -8,  -6 },
      {   0,   0,  -1,   1,   3,  -2, -10,  -4,   0,   1,   1,   0,   5,   4,   3,   8,   0,   0,   0,   1,   5,  -5, -15,  36,   0,   2,  -1,   2,   3,   9, -15,  -6,   2,   0,   2,   2,  23, -20,   1,  -8,   3,   7,   4,  15,  17,  54,  10,   7,  -2,   0,  -1,   1,  87,  11, -13,  -1,  -3, -14,  -4, -26,  -4, -28,  -7,  -7 },
      {   1,   1,   1,   1,   1,   8,   7,  -7,   0,  -1,   0,  -4,   0,  -6, -18,   0,   0,  -1,   0,   1,   3,   5,   6, -50,   0,  -1,   1,  -5,  -6, -11, -44, -16,   3,   1,   3,  -1,  40,  38,  10,   0,  -2,  -3,  -3, -11, -41,  -5, -14,  -5,   3,   7,   2,  29,  27,   3,   5, -18,  -1,  -1,  -1, -14,  49, -12,  29,   9 },
      {   0,   0,   0,  -1,   0,  -1,   1,  25,   0,  -1,   0,  -1,  -2,  -2, -10, -29,   2,   2,   3,   1,   6,   1,   4,  92,  -1,  -1,  -1,   0,   1, -11, -20,  30,   0,  -1,  -4,  -4,   3,  19, -25, -24,  -1,  -1,  -3,  -8, -11, -22,   5, -10,   2,   3,   2,  12, -10,   6,   4, -14,   1,   2,  -1,  -1,  43,   3,   8,   4 },
      {   1,   1,   2,   0,  -3,   1,   5,  10,   0,   1,   1,   1,   4,   8,  -1,  -1,   0,   1,   1,   1,  -2,  -7,  17,  -9,   1,   1,   2,  -1,   7,  22, -12,  43,   1,   3,   1,   6,   5, -11,  51,  20,   3,  -1,   5,  -9,  22,  60,  23, -21,   2,   1,  -2,   6, -39,  36, -11,  -5,   2,   9,   1,   7,  39,   1,  -4,   8 },
      {   0,   1,   1,   0,   0,   7,  -6,  -9,   0,   1,   0,   0,   3,   7,   4,  12,   0,   0,   0,   1,  -2,   3, -17, -18,   1,   0,   1,  -1,  -1,  -4,  17, -20,  -1,   0,   1,   5,  -5,   3, -85,   6,   2,   3,   1,   2,  20,  20,   6,  14,   0,   3,   1,   8, -16,  69,  19,  -5,  -2,   3,  -3,   1,  28,  -2,  -2,   3 },
      {   0,   0,   0,   1,  -2,   1,   5,   6,  -1,  -1,  -2,  -2,   0,   5,  -3, -19,   0,   0,   1,   0,   2,   4,  -9,   8,  -1,   0,   1,   1,  -5,   6,   4, -36,   0,  -2,   0,  -1,  12,  16,   2, -51,  -1,  -2,  -2, -10, -13,  43,   2, -11,  -3,  -6,  -6,  -9, -43,  -1,   1, -22,  -3,   0,  -8,   8, -44, -55,  32,  19 },
      {   0,   0,   0,   0,   0,  -4,   1,  -2,   0,   0,   0,   2,   3,   3,   8, -12,   0,   0,   0,   2,  -1,  -7,  -4,  12,   0,  -1,  -2,   1,   4,  13,   7, -85,   2,   2,   2,   3,  -6,   0,  25, -11,   2,   0,   2,   2,  23,  -7,  32,  22,  -1,  -1,   0,   3, -11, -26, -15,  16,   4,   3,   4,  12,  60,   2, -21,  -2 },
      {   0,  -1,   0,   0,  -1,  -7,  -4,   5,   0,   1,   0,   1,   3,   2,   3,  11,  -1,   0,  -1,   1,  -1,  -9,  -5,  -8,   0,   0,   0,   0,   7,   3,   4,  17,  -1,  -3,   1,  -4,  -2, -33, -40,   0,   0,  -1,  -3,   0,   3,  44, -48,  -3,  -2,   0,  -6,   1, -15, -76,   9,  11,   5,   2,   6,   2,  42, -10,  19,  -6 },
      {   0,   0,   0,   1,   1,   2,   0,  -6,   0,   1,   0,   1,   1,  -3,   9,  20,   0,  -1,   0,   1,   1,   3,  -5, -22,   1,   1,   1,  -1,   2,  -6,  20,  36,  -1,   0,  -1,   2,  -2,   7, -17, -10,   1,   1,   0,   2,  -9, -16,  91, -15,  -1,  -3,   1,  -2,   5, -36,  22,  -7,   1,  -6,   1,  -7,  13, -45, -10,  -3 },
      {   0,   0,   0,   0,  -1,   1,   1,   7,   0,   0,   0,   0,   0,   3,   1,  -5,   0,   0,   0,   1,   1,  -2,  -4,  29,   0,   0,   0,   1,   1,   4,  10, -19,   2,   1,   3,   2,   7,   3,  11,  80,   0,   1,  -2,   5, -15,  17,  21,  33,   1,   1,  -3,   5,  -4, -15,  58, -30,  -2,   0,  -2,   3, -13,   9,  39,  -1 },
      {   0,   0,   0,   0,  -1,   0,  -4,   0,   0,  -1,   0,  -1,  -1,  -7,  -5,   5,   0,   0,  -1,   0,   0,   1,  -7,  10,   0,   0,   0,  -1,   0,  -7,   0,   2,   0,   1,   0,   1,  -3,  -5,  -4,  61,  -1,  -1,  -2,  -2,   4, -30, -19,  12,  -3,  -2,  -2,   0, -16,   6, -64, -13,  -1,  -3,  -1,   1,   6, -79,  -5,  -4 },
      {   0,   0,  -1,  -1,   0,   2,   3,  -4,   0,   0,   0,   0,   0,  -2,  -3,  11,   0,   0,   0,  -1,   2,   5,   2,  -5,   1,   1,   0,   0,   1,  -4,  -9,  34,   1,   0,   1,  -1,   3,   6,  17, -36,   1,   2,   1,  -2,   0,   2,  -3, 110,  -3,   1,  -1,   0, -15,   6,   6,   7,  -1,  -2,  -1,   0,   3, -12,   3, -24 },
      {   0,   0,   0,   1,  -1,   0,  -5,  -2,   0,   0,   0,   0,   0,   3,   5,   9,   0,   0,   1,   1,   3,  -2, -11,  -5,   0,   0,   1,  -2,  -4,   5,   7,   8,   0,   0,   0,   1,   7,  -5, -22,  -3,   0,   0,   1,  -2,  -8,   7,  32,  23,  -1,  -4,  -2,   0,   4, -18, -76, -39,   1,   7,   0,   4,  -6,  53,  38,  40 },
      {   0,   0,   1,   0,   2,   0,  -1,   1,   0,   0,   1,   1,  -1,   0,   2,   4,   0,   0,   0,  -1,   1,   1,   3,   8,   0,   0,   0,   1,   0,   1,  -2,   3,   0,   0,  -1,   1,   3,   6,  -3,  11,   1,  -1,   0,  -3,   2,  -6,  17,   1,   1,   4,   1,   2,   2,  11,  -7, 102,   0,   0,   0,  -2,  -4,  -9,  64,  31 },
      {   0,   0,  -1,  -1,   0,  -1,   0,  -1,   0,   0,   0,   0,   1,   1,   2,  -3,   0,   0,   0,   0,   1,   0,  -1,  -3,   0,   0,   0,   2,   0,   2,   5,  -6,   0,   0,   0,   0,   0,  -1,  -2,  -6,   1,  -2,   0,   0,   2,  -1,  14, -22,   0,  -1,  -3,  -1,  -3,  10, -24,  -3,   2,   2,   1,   2,   0,  11,  53,-109 },
    },
    {
      { -70,  55,  -9,   2,  -2,   1,  -1,   0,  71, -48,   0,   2,   0,   0,   0,   0, -25,   5,  13,  -5,   2,  -1,   1,   0,   4,   8,  -9,   0,   1,   0,   0,   0,  -2,  -4,   2,   2,  -1,   0,   0,   0,   2,   1,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
      {  56,  12, -39,   6,  -3,   1,  -2,   0, -11, -49,  44,  -3,   1,   0,   2,   0, -36,  56, -13, -13,   6,  -2,   1,  -1,  28, -20, -16,  17,  -4,   1,  -1,   1,  -8,  -3,  17,  -6,  -2,   2,   0,   0,   2,   4,  -6,  -2,   4,  -1,   0,   0,  -3,   0,   2,   2,  -1,   0,   1,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
      {   2, -59,  46,  -4,   7,  -2,   1,   0,  48,  23, -25, -12,  -1,  -3,   1,  -1, -66,  30, -11,  18,  -3,   3,  -1,   0,  32, -12,   2,  -1,  -5,   2,  -2,   1,  -1,  -9,   6,  -4,   3,  -1,   1,   0,  -4,   6,  -1,   0,   1,  -1,   0,   0,   1,  -1,  -2,   1,  -1,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
      {  72,  -4,  -2, -14,  -3,  -2,   0,  -1,  38, -26, -23,  15,  -1,   2,  -1,   1, -22, -32,  41,  -8,  -1,   0,   1,   0, -37,  47, -13, -10,   8,  -3,   2,  -1,   9, -10, -12,  16,  -4,   1,  -1,   1,  -4,   0,  10,  -5,  -2,   2,  -1,   0,  -2,   2,  -2,   0,   3,  -1,   0,   0,  -2,   1,   0,   0,  -1,   0,   0,   0 },
      {  14, -24, -30,  34,  -2,   4,  -2,   2,  58,   7,  37, -40,  -5,  -3,   0,  -1,  16, -17, -42,  12,  15,  -4,   2,   0, -42,  -2,  33,  12, -14,   2,   0,   0,   3,  20, -11, -18,   5,   3,  -1,   0,  -2, -12,  -1,  12,   1,  -3,   1,   1,  -2,   4,   3,  -5,  -3,   2,   1,  -1,  -2,  -1,  -2,   2,   2,   0,  -1,   0 },
      {   0,  12, -50,  44,  -5,   5,  -4,   2,  -5,  49,   4, -23, -12,   2,  -1,   0, -33, -45,  33,  -3,  18,  -2,   2,   0,  58,  12, -10,  -4,  -4,  -4,   0,  -1,  -7, -10,  -7,   4,  -1,   2,   1,   1,  -7,   2,   6,   2,  -1,   0,  -1,   1,   5,  -4,  -1,  -3,   1,   0,   0,   0,  -1,  -1,   2,   1,   1,   0,   0,   0 },
      {  36,  34,  -2, -14,  -6,  -1,  -2,  -1,  24,  -4, -34,  12,   5,   0,   0,   1,  13, -20,   1,  16, -13,   5,  -2,   2,  31, -49,  38, -23,   7,   1,  -1,   0, -51,  40,  -7,  -6,  12,  -8,   4,  -1,   9,   2, -15,  12,  -8,   3,   0,  -1,  -2,  -8,   9,  -1,  -2,   3,  -2,   1,   0,   2,  -1,  -2,   2,  -1,   1,   0 },
      {  11, -15,  -5,   0,  13,   2,   0,  -1,  48,  25,  -7,  15, -23,  -1,  -2,   1,  60,   7,   9, -40,   9,   2,  -1,   0,  13, -24, -43,  27,  10,  -8,   2,  -1, -16, -13,  37,   1, -19,   6,   0,   0, -13,  12,  -8, -14,  13,   0,  -1,   1,   2,  -6,  -1,  10,  -2,  -2,   1,   0,  -6,   1,   3,  -3,  -1,   2,   0,   0 },
      {  23,  49, -12, -17,  13,  -3,  -1,  -3,  14,  24, -20,  10, -12,  -1,   0,   2,  11,  14, -31,  22,  -6,   7,  -2,   1,  16,  -7,  20,  -5,  -3,   1,  -5,   1,  69, -41,  -5, -12,   4,  -2,   2,   1, -43,  -2,  11,   6,   3,  -2,   3,  -1,   8,   4,  -4,  -3,  -2,   1,  -1,   0,  -6,  -6,   3,   2,   1,   0,   0,   1 },
      {  12,   3,  21,  26, -34,   4,  -5,   2,   8, -20,  -7, -46,  40,  -2,   3,  -1,  47,  11,  -6,  32, -13, -12,   3,  -3,  24,  18, -23, -20, -13,  14,  -4,   1, -22, -40,   0,  19,  13,  -4,  -1,   2,  -7,  15,  20,  -8, -11,   1,   2,  -1,  -4,   1, -11,  -2,   7,   0,  -1,   0,  -2,   0,   3,   4,  -3,  -1,   1,   0 },
      {   5,  27,   6,  27, -42,   5,  -4,   1, -14,  14, -45, -11,  23,   6,  -2,   1, -26,  -4,   8,  -6,  13, -15,   5,  -1, -47, -24,   2,  32, -22,  11,   2,   0,   5,  -5,  41, -25,   5,   3,  -6,   1,  -3,  28, -22,  -6,  11,  -8,   4,  -1,   4,  -7,  -2,  13,  -7,   2,   0,  -1,   5,   0,   2,  -5,   0,   1,  -2,   1 },
      {   0, -12,   6,  37, -31,   9,  -3,   3,   4,  -7, -20,  -5,  17,   0,  -1,   0,  13,  33,  38, -32,  -2,  -7,   1,  -1,   8, -32,  -1,  -1,  11,   1,   2,  -2,  47,  40, -38,  11,  -9,  -4,  -1,  -1, -21, -39,  12,   3,  -3,   9,  -1,   1,  -5,  12,   9,  -7,   5,  -4,   0,   0,  -1,  -1,  -5,   1,  -1,   0,   1,   0 },
      {  -4, -23,  30,   1,   7,  -4,   2,  -2, -11, -54,   8,   2, -12,   9,   2,   2,  14, -34,  18,  -9,  28,  -7,   0,   0,  31,   2,  19,  13, -27,  -1,   3,  -2,   8,   3,  -8, -30,  13,   7,  -4,   0, -53,  27, -12,  21,   8, -12,   3,   0,  26,  -3,   1,   1, -15,   6,  -1,  -1,  -5,  -1,   1,  -4,   6,   0,  -2,   1 },
      { -20, -40, -53,  22,   0,   4,   2,   3,  -4, -13,  -9,  29,  -4,   1,  -3,   1,  -4,  13,  13,   5, -24,   7,  -4,   1, -27, -26,  27, -32,  19,   1,  -2,   2, -13, -26,   6,  15,  11, -13,   7,  -2, -33,  42,  11,  -7,  -9,   5,   0,  -3,  17,  10, -16,  -2,   4,   0,  -4,   3,   3,  -6,   3,   2,  -2,  -1,   2,   0 },
      {   0,  -7, -18,  32,  23, -20,   3,  -1,  -3,  -7, -46,  10, -34,  25,  -2,   2,  20,  46,  14,   7,  21, -10, -11,   1,  -2,  22,  10,  -9, -28,  -7,   9,  -2,  -8, -19, -14, -13,  18,  15,  -4,  -1,  42, -11,  -4,  23,  -1, -15,   0,   1, -11, -13,  14, -10, -11,   9,   2,  -1,  -1,   9,  -2,  -3,   7,  -2,  -2,   1 },
      {   0, -12, -12,   7,  14,   1,  -1,   1,   4,  -7,   9, -17,   3,  -3,   5,  -1,  12,   2,  28,  37, -28,   7,  -2,  -2,   6,  12, -21, -24,  15,   9,  -9,   4,  48,  31,  33, -15,   6, -20,   4,  -2,  15,  11, -52,   3,  -5,   4,   4,  -4,  11, -38,  11,  16,  -3,   4,  -5,   4, -13,   8,  10, -10,   1,  -1,   1,  -1 },
      { -10, -34, -34, -13, -34,  31,  -3,   5,   5, -14, -17,  45,  31, -20,  -4,   0,   9, -21, -18,  -1, -14,   3,   9,  -1,  31,  21,  10,  17, -15,   8,  -9,   0,  13, -18,  13, -23,   3,  -2,  -1,   3,  16, -24, -10,   3,  11,  -3,   2,  -2, -37,  12,  17,   3,  -7,   0,   1,   0,   7,   8,  -4,  -6,   2,   2,  -1,   0 },
      {  -2,  -9,  -6,   8,  34, -25,   3,  -1, -13, -34, -22, -25, -14,  19,   8,   0,  -4, -35,   6,  39,  -7,   7, -10,   3,  -4, -40, -20,  10,   5,  -1,   3,   4,   7,  -4,  31,  -3, -12,   6,  -2,  -2,  -9, -33,  19, -12,  11,   0,  -1,   1, -41,  35,   3,   0,   7,  -5,   3,  -2,  10,   4,  -4,   7,  -5,   1,  -2,   0 },
      {  -6,   2, -23, -29,  13, -12,  12,  -2,   1,  24,  -1,  -9,  26,  -6,  -1,  -4,  -4,  18,  17,   7, -23,   0,  -2,   0, -19,  10, -42,  11, -17,  15,  -3,   2, -32,  10,  -6, -17,  29,  -5,  -3,   2, -47, -23,   0,  38,  -4,  -5,   6,  -2,  16,  18,  29, -11, -16,   9,  -6,   0,   9,   0, -16,  -7,   9,  -4,   0,   1 },
      {   2,   6,   1,  35,  11,  26, -17,   2, -14, -17, -40,  -3,  -7, -33,  19,   1,  -2,   9, -27,  11,  -6,  34,  -6,  -2,   1,  36,  -3,  16,  21, -23, -10,   2, -23,  20,   9, -11, -28,   2,  12,  -3, -13, -17,   9,   4,  16,  10, -11,   1,  44,   6,  -3,  -2,  -9, -10,   3,   2, -14,  -6,   1,  -3,   3,   5,   1,  -3 },
      {   6,  16,  33,  36, -11,  -7,  -4,   1,   0,  11,  23,   6, -16,   4,  -4,   0,  -1,   9,  -3, -14, -10,   6,  -6,   3,  -1,  26,  17, -32,  31, -16,   2,   0, -12, -14,  19,  -9,  -5,  -5,   6,  -3, -30,   0, -37,  13,   6,   4,   1,   0, -42,  29,  26,   8,  -9,   0,  -2,   2,  39,   8, -18, -12,   1,   2,   0,   0 },
      {   8,   7,  17,  22,  -5, -26,  14,  -4,  11,  10,  27,  21,   5,   5, -12,   0,   7,   3,  33,   4, -42,  11,  -4,   4,  10,   6,  14,  16,  -4,   0,  -4,   0,   1,   1,  11, -47,  14,  -2,   1,   0,  33,  -9,  19,  -7,   8,  -9,   4,  -3,  25,  39, -45,  -4,  -5,  -1,   1,  -1,  -2, -28,  -2,   7,  -2,   6,  -2,   1 },
      {   4,   6,  22,  -1,  14,  44, -31,   8,  -2,   0,  21,  -8, -14, -51,  24,  -1,  -1,  -3,  41,  -1,   6,  21,   3, -10, -19, -37,   3,  -4,  -5,   2, -13,  11,   0, -43,  -3, -10,  16,  -3,  12,   0,  23,  -5,   7,  26,  -2,  -4,  -2,  -3, -10,   0,   9, -17,  -7,   5,   1,   1, -10,  10,  -3,   2,  10,  -2,  -1,  -1 },
      {   2,  -1,  16,  34, -12, -32,  20,  -6,  -2, -11,   6,  17,   8,  15, -17,   1, -13, -30, -31, -20, -27,  14,  -3,   8,  -5, -21, -18,  12,  20,  -4,   0,   0,   1, -37,  -7,   4,   8,  -8,   8,  -1,   9, -20,   1,  24, -17,   9,  -2,  -2,  19, -33,  27, -11,  -1,   4,  -7,   5, -38,  22,   4,  -9,   9,  -6,   3,  -1 },
      {   1,   2,  13,   9, -20,  24, -19,   8,   1,   2,   9,  39, -37,  16,  -2,   3,  -4,  -4, -14,  19,  29, -35,  14,  -6,  -2,   1, -22, -29,  -6,  27,  -4,   4,  -3,  15,  40,   6,  16, -15,  -9,   2,  -6, -35,   4,   4, -22,   2,   9,  -2,  25,  21,  -5, -16,  14,   5,  -6,  -1, -20,  -5,  -5,   7,   2,  -9,   3,   2 },
      {   3,  -3,  -3,   1,  17,   3,   2,  -2,   0, -11, -26, -18,   0,  -2,  -2,   4,   3, -14, -26, -30,  10,   5,   6,  -3,  10,  -1, -21,  12,  17,   4, -10,   2,  24,   4,   0,   8,  18, -26,   8,   1,  33,  31, -11,  29, -35,   9,   6,  -6,  10,  31,  -6, -24,   1,  10, -13,   4,  48, -33, -15,  11,   3,  -7,   4,   1 },
      {  -3, -10, -11, -26, -27,   2,  -6,   6,   2,  -5,   1, -18,  13,  17,   1,  -5,   0,  -6,   0,   6,  12, -14,  -5,   3,  -2, -20,  -7, -35,  25, -26,  22,  -3,   5, -28,   0,   4, -20,  29,  -4,  -2,  21, -22, -24,  25,  26, -14,  -6,   5,  41,  17,   3,   9, -30,  -4,   9,  -5,   4, -37,   6, -11,   6,   8,  -6,   2 },
      {   6,   3,   9,  26,   5,  -2,   2,   2,   8,   2,  15,  39,   4, -18,   0,  -6,   2,  -9,  -3,  38,  -4,  -1,  -6,   2,   0, -17, -16,   6, -27,   2,   4,  -1,  14,   7, -11,  31,  -9,  20,  -6,   0,  11,  14,  10, -20,  13,  -8,  -7,   3,  36,  -8,  28,  -5, -12,  -3,   4,   0,  62,  16, -29, -16,   1,   0,  -1,  -1 },
      {  -1,  -2,  -4, -19,   0,  -7,  14,  -6,   2,   3,   3, -17,  15,   9, -14,   8,   8,  10,  27,  13,  12, -18,  15,  -6,   7,   9,  42,  20,  11,  -5,  -8,  -6,   2, -12,  30,   2, -38,  -1,   7,   4,   9,  10,   1,   3,  -8,  25, -12,  -1,  24,  18,  31, -41,  15, -14,  -4,   4, -16,  29, -40,   8,   4,  -5,  10,  -5 },
      {   4,   5,   4,  23,  61,   3,  19, -21,  -1,  -4,  -4,   0,  33, -29, -20,   9,  -2,  -4,   0, -20,  10, -25,  18,   0,   0,  -3,  21, -15, -15,  31,  -8, -10,  -8, -18,   3,  23,  -5, -13,  -9,  12,   2, -31, -29, -14,  14,   6,  12,  -4,  12,  12,   0,  25,  -7,  -5,  -5,   1,  -7, -17,  11,  -7,  -8,   6,   1,   2 },
      {  -1,  -4,  -4, -16,  -6, -13, -25,  16,   0,  -5,  -6, -29, -16,  12,  38, -15,   0,  -4,   0, -43, -27,  11, -19,   8,   8,   9,  35,  -1, -10,  23,   1,   8,   1, -11,  24,  13,  18, -12, -13,  -4,   4, -30,   9, -31,  -2, -10,  13,   4,  34, -13,  16,  14,   5,   8,  -2,  -9,  21,  11, -19,  -5,  -2,  -5,  -4,   4 },
      {   1,   3,   1,  15,  32, -26, -53,  36,   4,   8,   2,  23,  39,  17,  41, -31,  -3,  -4,  -9,  -7, -10, -36,   4,  -5,   3,   3,  -1,   3,  13,  12, -10,   9,   3,   8,  -8,   1, -19,  10,   1,  -2,   3,  26,   1,  11,   6, -13,   1,  -2, -14,  12,   0, -11, -12,   7,   0,  -1, -29,   0,  -7,   2,   6,   1,  -1,   2 },
      {   1,   3,  -2,   3,  -1, -12,  17, -20,   2,   5,   1,   2,  26, -27,  11,   1,  -1,  -2,  -9, -18,  20,  34, -33,  10,   4,   3,  -2, -31, -23, -20,  27,   0,  17,  23,  35,  27,  21,  12, -10,  -9,   5,  16,   4,  -2,  -6, -20,  -3,   5,  -7,  33,  -2, -17, -11,  10,   6,  -5, -40,  18, -11,  -3,   9,  -2,  -7,   1 },
      {   2,   3,  -1,  -1,  15,  31,   8,   3,  -4,  -4, -10, -22, -14,   2, -30,  10,  -4,  -5, -16, -24, -30, -25,  32,  -6,   2,  -3, -13, -30,  21,   8,   6,  -6,   8,  14,  -1, -27,  10,  16, -11,  10,  11,  21,  38,   0,  21, -17,   4,  -1,   5,   8,  -3,   9, -21,  -3,   8,  -6,  -3,  50, -21, -11,   4,   8,  -4,   2 },
      {   1,  -2,   0,   2,  25,   2,  -7,   2,   3,   2,   3,  11,  48,  -1,   1,  -3,   1,   0,   2,  -8,  26, -10,  -4,  -8,  -1,  -1,   4, -16,  13, -35,  15,   1,   1,  -4,  25, -35,   7,  14,   1,  -7,  -9, -28,  19,  10, -33,   4,  -5,   5,   9, -33, -32,   2,  26,   7,   1,   2,  42,  28,   9,   1, -10,  -7,   1,   0 },
      {  -2,  -3,  -8,  -9, -16, -35,   0,  -4,   1,   1,  -5,   2,  -2, -13,  15, -12,   2,   1,  -5,   2,  43,   2, -14,   2,  -1,  -7, -18, -14,   5,  14, -23,  11,  -1,  -8, -30, -24,   7, -24,  31,  -3,  12,   1, -13, -11,  14,  20,  -3, -10,  17,  31, -19,  36,   2, -16,  -6,   7,  -2,  51, -15, -10, -13,   4,   7,  -3 },
      {  -2,  -4,  -4,   0,  -9, -31,   0, -15,  -2,  -3,  -6,  -1,  -6, -22,  11,  11,  -4,  -7, -11, -25, -13,   2,  14, -18,  -4,  -4, -16, -48, -30,  13, -15,  11,   0,   1,   1, -28, -19,  21,   8,   0,  -6,   0, -10,  -3,  21,   4, -16,   3,  -3, -28,   2, -32,  27, -17,  15,   2,   5, -27, -26,  40,  -9,   5,   1,  -6 },
      {   2,   3,   1,   3,  16,  38, -14,   5,   0,   0,  -2,   0,   0,  15,   2,   4,  -3,  -6,  -5,   0, -30, -14, -25,  17,  -1,  -3, -12,  14, -12, -18,  36, -12,  -1, -12, -28,  -6,  24,  12,  -8,  -5,  -1,   2, -53, -13,  -6,  13, -11,  12,  10,  21,  -8,  -3,  35, -20,   9,  -5, -14,  16, -23,  28, -22,  -1,   3,  -6 },
      {   1,   1,   5,   4,  15,  33,  -2,  -9,   2,   2,   7,   7,  29,  41,  -8,  -1,   0,   0,   0,   3,  31,  22, -31,  12,  -4,  -2,  -8, -14,  11,   5,   1,  -9,  -2,  -2,  -4, -27,   5, -40,   9,   6,   4,   7,  28, -23,  22,  -2,   7,  -3,  -5, -22,  31,   7, -12,  -5,   1,  -4,  -9, -33, -39,  19,   0,   1,   4,   0 },
      {   0,   0,  -1,   1,  -9,   2,  22, -15,  -2,  -2,  -2,   1, -12, -35,  17, -21,  -1,  -1,   1,   6,   0, -25, -43,  42,   1,   1,  -1,   2,  26,  44,  25, -29,  -4,  -5,  -3, -16, -37,  -9, -26,   9,   0,   3,   2,  21,   4,  -3,  17,   2,  -6, -15, -22, -18,   3,  12,  -5,  -6,   7,   7,  -5,  10,   9, -10,  -1,   8 },
      {   2,   3,   2,   0,   3,  -3,  35,  91,  -3,  -5,  -6,  -6,  -4, -19, -30, -53,   1,   1,   0,  -1,  13,  20,  16,  11,  -2,  -2,  -3, -13, -12, -10,   4,  -5,  -1,  -2,   0, -10,  -2,  -1, -12,   1,  -2,  -2,  -7, -15,  -4,  11,   5,   1,  -1,  -1,   1,  -4,   5,   2,  -8,   1,  -8, -11, -10,   6,   5,   0,   5,   1 },
      {  -3,  -4,  -5,   2,  -8, -27, -21, -13,  -2,  -3,  -4,   2,   0, -28,  -6,  17,   1,   3,   1,  18,  15,  11,  28, -20,   0,  -2,  -2,  13,  41, -16,   4,  -6,  -4,  -7, -20,  -9,  12,  11, -34,  14,  -5,  -5, -19, -35, -20, -29,  35,   0,   1,  -3,   3,   9,   4,  27,  -3, -12, -12,  -7, -40,  15,   9,  -7,  -8,  10 },
      {   1,   2,   2,   3,   2,  18,  11,   7,   2,   3,   2,   2,  15,  40,  -6,   1,   1,   3,   1,   0,   9,  52,   6, -11,  -1,   0,   0,  -2,   5,  24, -27,  11,  -4,  -2,  -8,  17, -19,   5, -12,   9,  -2,  -6, -28,  13,  30, -36,  24,   1,  10,   6, -23, -30,  10,  13,  -1,  -6,  13,  38,  19,  29, -10,  -7,  -9,   4 },
      {  -1,  -1,  -1,  -2,  -2,   2,   2,   4,  -2,  -3,  -3,  -4,  -4,  -3, -22,  -5,  -1,  -2,   0,   3,  -6, -22, -12,  -1,   1,   1,   6,  15,   0, -11, -13,  13,   0,   1,  10,  36,  11, -14,  32, -17, -12, -23, -15,   7,  11, -10, -15,  -3, -12, -30, -55,  -4, -42,  11,   9,   7,   0,   5, -53,  23,  19,   9,  -3,  -1 },
      {   1,   1,   2,   4,  -2, -10,  25,  13,   1,   1,   0,   6,   1, -17,  13,  -4,   3,   3,   4,   7,  19,  -9,   9,   2,   5,   8,   9,  19,  48,  20,  -1,   8,   3,   3,  12,  16,  52,   6, -11, -17,   0,   3,  10,  24,  23, -24, -23,  14,   1,  -8,  19,   6,  16, -34,  24,  -6,   0,  -7,   4,  23, -36,  20,  -1,  -8 },
      {  -2,  -3,  -2,  -6,  -2, -34, -61,  21,  -1,  -1,   1,  -2,   3, -18, -43,  28,   1,   1,   0,   0,   4,   9,  -3,   4,   1,   3,  -1,   1,  -2,   6,  41, -28,   1,   4,   6,   5,  12, -32,   0,  10,   0,   0,   4,   1,  34,  17,  -2,  -7,   3,  -3,   2, -26, -15,  -6,   3,   2,  11,  14,  29,  18,  -7,  -1,   3,  -2 },
      {   1,   1,   1,  -1,   1,   6,  21,  -7,  -1,  -1,  -2,  -5,  -1, -10,   8, -10,  -1,  -1,  -2,  -2,   0, -35,   0,  11,   0,   2,  -2,  -1,  16, -15,  -1,  10,  -2,   0,  -7, -15,  15,  -5,  43, -21,  -1,  -4,  -8, -53,  17, -19,  -6,   3,   7,  10,  28, -38,  -6,  44,  -7,   5,   7,   5,  42,  22,   7, -13,  -8,  -3 },
      {   1,   2,   1,   1,   4,   3,  18,  36,   1,   1,   0,   0,   3,  -6, -10,  49,  -1,  -2,  -2,   0,   1,  -5, -46, -59,   3,   3,   2,   3,   9,  35,  32,  46,  -3,  -4, -10, -10, -19,   1,  -5, -29,   1,   2,  -2,  -8, -13,  -1,  -5,   0,   1,   5,   4,   6,  -3,   5,   0,   5,  -2,  -3,   2,  -6,  13,   0,  -1,  -3 },
      {   1,   1,   0,   3,   2,   2,  28,   8,   2,   3,   2,   5,   5,  18,  59,  15,   0,   0,  -2,   1,   6,  10,  30, -23,  -1,  -2,  -1,   0,  10,   6,  20, -28,  -1,  -5,  -5,  -3,  23, -13, -24,  22,  -2,  -8, -12, -20,  17,  31, -36,   2,  -6,  -6, -23, -24, -22,  -5,   3,  12,   6,   3,  -1, -16,  37,  -8,   4,  -5 },
      {  -1,  -1,  -2,  -3,   6,   1, -29, -12,  -1,  -2,  -2,  -6,   5,   6, -36, -14,   1,   2,   1,  -2,  13,  25,  -1,  28,   0,   3,   1,   0,  20,  52, -12,  -7,  -1,  -1,  -3, -10,  11,  44,  10, -21,  -5,  -4, -12, -25, -23,  -7, -34,  11,  -3,   1,  -8, -18,   3, -12,  19,  13,   0,  -9,   0, -34,  31,  -9,   8,  -9 },
      {   0,  -1,   0,  -1,   0,   3,  -7, -30,  -1,  -2,  -2,  -2,  -3,   0, -16, -48,  -1,  -2,  -1,  -1,   2, -11, -21, -24,   0,   0,  -1,   2,   7, -17,  -8,  46,  -2,  -1,  -3,  -5,  14, -11, -44,   7,  -1,   1,  -5, -17,   2,  25,  29,  -2,  -2,  -5,  -2, -40, -17, -39,  11,  -1,   4,   4,  16, -12,  10,  32,  -6,   0 },
      {   0,   0,  -1,   1,   2,  -3,  -1,  -9,   0,   0,  -1,   2,  -1,   4,   3,  -9,   1,   1,   1,   3,   1,  10,  40,   8,  -1,  -2,  -4,   0, -13,   7,  27,   8,  -3,  -7,  -7,  -9, -31, -55,   3, -29,  -1,   1,  -4,   0, -28, -14, -17,  46,   2,   5,  -1,  -2,  -8,   8,  44, -30,   1,   7,  -1,  -9,  -8,  28, -30,   7 },
      {   0,   0,   0,  -3,   2,  -4,   6,  30,   0,  -1,  -1,  -3,   2,  -5,   6,  44,  -1,  -2,  -3,  -2,   3, -17, -20,  19,  -2,  -3,  -5,  -1,   0, -22, -32, -16,  -2,   0,  -3,  -1,   3, -22,  11,  14,  -4,  -3,  -7,  -9,   5, -37,  19,  -4,  -3,  -7, -10, -37,  31, -19,  16, -11,  -2,  -7,  -6, -62,   0,  23,  -6,   4 },
      {   0,   0,   1,   0,  -3,  -3,   6,  18,   1,   1,   2,   2,  -1,   0,   3,  16,  -1,  -1,   0,  -3,   0,  -5, -29,  -1,  -1,  -1,  -1,  -3,   0,   2, -43, -34,   1,   1,   3,  -5,  -3,  15,  -2,  32,  -2,  -1,  -1, -11, -47,  11,  -6,  -2,   5,   7,  15,  13, -39, -10,  45, -12,   5,  11,  22,  45,   9,  18, -27,   0 },
      {   0,   0,   0,   0,   1,   2,   6,  10,   1,   2,   2,   3,   5,   7,  23,  37,   1,   2,   1,   2,   6,  10,  29,  56,   0,   0,  -1,  -2,   1,  11,  17,  26,  -1,   0,   0,  -3,   1,  29,   8, -31,  -2,  -1,   0,  -8, -13,  30,  41, -29,  -2,  -9,  -3, -14, -30, -14, -17,   6,   1,   2,  -4,   9, -13,  34, -13,  11 },
      {   0,   0,   0,   2,   3,  -4,  -2,   6,   1,   1,   1,   2,   7,  -2,  -6,  24,   1,   1,   1,   1,   3,  -8,   0,  30,   0,   1,  -1,  -1,   0, -13, -29,  26,   1,   2,   0,   4,  11,  -1, -21, -24,   4,   7,   4,  11,  32,  40,   7,  32,   2,   3,   1,  18,  24,   5,  14, -42,   3,   1,   5,  11,  59, -34, -11,  15 },
      {   0,   0,  -1,  -2,  -1,  -1,  -4,   3,   0,   0,   0,  -2,  -1,  -1,   0,  19,  -1,  -2,   0,  -2,  -1,  -4,  -9,   6,  -2,  -5,  -4,  -7,  -3, -12, -43, -17,  -1,  -4,  -4,  -8,  -7,  -8, -52, -47,   2,  -1,   1,  -8,  -5,   1, -12,  37,  -3,  -1,  -4, -12, -39,  11, -24,   7,  -3,   2,   1,  -3, -45, -11,  44, -15 },
      {   0,   0,   0,   0,   1,  -2,  -2,   0,   1,   1,   0,   0,   1,  -2,   5,   8,   0,   1,   0,  -1,  -3,  -4,  14,  29,   0,  -1,  -1,  -3,  -8, -16,   5,  17,  -2,  -1,  -3,  -2, -16, -30, -27,  -9,   0,  -1,   1,   0, -13, -48,  -5, -29,   4,   5,   5,  18,   6, -34,   6,  43,   7,   7,  11,  27,  47,  -6,  21, -47 },
      {   0,   0,   0,   1,  -1,  -2,   3,  -2,   0,   0,   0,   2,   0,  -1,   7,  -6,   0,   0,  -1,   0,  -2,   1,   3, -28,   0,   0,  -1,   0,  -3,   4,   9, -39,   1,   2,   3,   2,   3,   7,  38, -23,   1,   0,   0,  -1,  -7, -16,  45,  54,  -2,  -2,  -1,   2,  -7, -54, -40, -17,   4,   4,   7,  14,  35,   9,  16,  -4 },
      {   0,   0,   0,   2,   0,  -2,   9,   5,   0,   0,   0,   0,   0,  -4,   6,   4,   0,   0,   1,   0,   0,  -1,   8,  12,   0,   2,   2,   1,   2,   1,  16,  37,   0,   1,   1,  -1,  -2,   2,  24,  39,  -2,  -2,  -3,  -6, -13,  -7,  20,   9,  -3,  -3,  -6,  -7, -32, -24,  35, -36,   0,  -3,   1,  -7, -25, -66,  46, -14 },
      {   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,  -1,  -1,  -3,  -3,  -1,   0,   0,   0,  -1,  -2,  -2,   1,  -1,   0,   0,  -1,  -2,  -3,  -6,  -4,   1,  -1,  -1,  -2,  -5,  -5, -11, -10, -14,  -2,  -2,  -4,  -7, -10, -26, -39, -31,  -2,  -2,   1,  -4, -12, -49, -31, -16,   1,   2,   5,   5,  -5, -38, -23,  81 },
      {   0,  -1,   0,   0,   0,  -2,  -5,   3,   0,   0,   0,   0,   1,  -1,  -3,  13,   0,   1,   1,   0,   2,   1,  -4,  26,   0,   1,   0,  -1,   2,   2,   3,  41,   0,   2,   1,  -2,   0,   2,  -1,  61,  -1,  -1,  -2,  -4,  -8, -12, -39,  42,   0,   0,   1,   3,   0,  10, -47,  -9,   2,   3,   4,  13,  13,  48,  32,  21 },
      {   0,   0,   0,  -1,   1,  -1,  -2,   6,   0,   0,   0,   0,   0,  -2,  -2,   8,   0,   1,   0,   0,   1,  -4,  -2,  11,   0,   1,   1,   0,   1,  -1,  -2,  16,   0,   1,   1,  -1,   1,  -1,   1,  27,   1,   2,   3,   1,   2,   2,  10,  57,  -2,  -1,  -1,   0,  -8, -17, -14,  68,  -4,  -4,  -7,  -5, -17, -37, -69,  -8 },
      {   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,   0,   0,   1,   5,   1,   0,   0,   0,   0,  -2,   0,   1,  -2,   0,   0,   0,   1,  -2,  -3,   2,  -5,   0,   0,   0,   2,   1,  -1,   5,  -6,   2,   2,   2,   2,   4,   2,  25,  19,   3,   4,   5,   5,   6,   8,  51,  64,   0,   2,   2,   6,   7,   7,  50,  76 },
    },
    {
      {  97, -33,  -7,  -2,  -3,   1,  -2,   0,  55, -28,  -5,   0,  -3,   0,   0,   0, -21,  -5,   5,   2,   2,  -1,   1,   0, -31,  10,   3,  -1,   2,   0,   0,   0, -11,   8,   2,  -1,   1,   0,  -1,   0,  -4,   3,   0,  -1,   0,   0,   1,   0,  -2,   1,   0,   1,   0,   0,   0,   0,  -1,   2,  -1,   0,   0,   0,   0,   0 },
      {   4, -67,  43,  -3,   6,   2,   0,   1, -59, -30,  40,  -1,   2,   1,   3,   0, -27,  30,   2,   1,   0,  -4,   1,  -1,   9,  33, -21,  -8,   0,  -2,  -2,   0,  17,   9, -12,  -2,   2,  -1,  -2,   1,   6,  -1,  -5,   1,   1,   1,   1,   0,   3,  -3,  -2,   3,   1,   0,   0,   1,   1,   0,  -2,   1,   0,   0,  -1,   0 },
      { -63, -30,  32,  -1,   3,   3,   0,   1,  70, -51,  15,  -1,  -3,   1,   0,   0,  39, -13,  -2,   0,  -1,  -2,   0,   0, -16,  14,  -5,  -2,   2,  -1,   0,   0, -13,  12,  -4,  -1,   1,   0,   0,   0,  -8,   7,  -1,   0,   0,   0,   0,   0,  -1,   3,  -2,   1,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   0 },
      {  37, -36,  10,  -3,   1,  -1,   1,  -1, -16,  16,   5,  -4,  -1,   0,  -1,   0,  91,   9, -20,  -3,  -5,   0,  -2,   1,  39, -29, -11,   7,  -1,   2,   1,   0, -27, -17,   6,   5,   0,   2,   1,   1, -18,   4,   8,   0,   0,   0,   0,   0,  -7,   6,   0,  -2,   0,   0,   0,  -1,  -2,   1,   0,  -1,   0,   0,   0,   0 },
      {  -6,   3,  31, -13,   1,   8,  -7,  -2,   2,  52, -11,  -7, -11,  -1,   6,  -1,   6,  41, -36,  12,   7, -10,   7,  -2, -78,  10, -34,   4,  12,   1,  -5,   2, -14, -14,   5,   0,   5,   3,  -5,   1,  -2,  -5,  11,  -4,   1,   0,   7,  -2,  13,  -4,   3,   6,   2,  -3,   1,  -1,   9,  -2,  -5,   4,   1,   0,   4,   1 },
      {  -3, -21,  -9,  15,  -3,   1,   0,   2, -52,  -2,  -4,  16,  -2,   4,   0,   2,  28, -50,  21,   4,  -2,   1,   0,   0, -64, -13,  37,  -5,   5,   0,   2,  -1, -26,  38,  10,  -7,   3,  -3,   0,   0,  17,  19,  -8,  -4,  -1,  -2,  -1,   0,   9,   0,  -8,   0,   0,   0,   0,   0,   2,  -2,   0,   0,   0,   0,   0,   0 },
      {   7,  17,  21, -58,  17, -13,   6,  -5,  -4,   7,  36, -60,  29,  -7,  -5,   0,  -4, -19,  19, -26,   9,  10,  -2,   6, -15,  -5,  12,  36,  -4,   5,  12,   0, -18,   6, -18,  29, -12,   3,   9,  -4,   3,  -4,   0,  13, -11,  -2,  -5,   0,  -5,   8,   2,  -6,  -8,   2,   1,  -3,  -1,   0,   6,  -5,  -3,  -1,   2,   0 },
      {  22,  22,  32,  -9, -21,   3,  -2,   6,  16,  31,  27,  -1, -21,   7,  -3,  -1,  35, -27,   3,   9, -13,   1,  -4,  -4,   1,  13,  -3, -21,  -1,  -6,  -2,  -2,  61,  23, -27,  -9,  10,  -9,  -3,   3,  36, -18, -27,   2,   5,   5,  -3,   2,  -6, -20,  -6,   6,   3,   3,   1,   5,  -8,  -6,   3,   2,   2,   0,  -8,  -2 },
      {  11,  43,  46,   2, -22,  11,  -7,  -3,  -2, -13,  50,   7, -31,   5,  -1,  -5, -25,  -7,  11,  12, -10,   1,   2,  -5,  11, -43, -13, -12,   2,  -1,  -5,   4, -42,  -6,  19, -19,  11,   0,  -4,   1, -16,  29,  -8, -10,  10,  -1,   5,  -1,   9,   5,  -4,   3,   5,  -3,   1,  -1,   5,   6,  -9,   5,   1,   1,   2,   1 },
      {  -2,   7, -13, -14,   6, -19,  -3,   2,  13,  13,   4,   1,  14,   0,  -6,   2,   1,  52, -12,   4, -15,  -8,  -6,   6,  18,  21,  19, -19, -24,  -6,   3,  -6, -55,  39,   3,  -4,   2,  -9,   5,   2,  39,  22, -31,   0,   4,   9, -13,   3,  11, -23, -20,   1,   0,   7,  -3,   9, -13, -12,   6,  -2,   1,  -1, -15,  -5 },
      {   3,  19, -33,  36, -15,   4,   8,  -4,  -1,   2,  30,  -8, -13,  -4,   0,  -3,  27,  34,  57, -50,   9,  -1,  -1,  -2, -13,  32, -15, -24,  20,  -3,   1,   2, -17,  -3, -27,   1,   4,   6,   1,  -2, -16,  -6,  -2,  11,  -2,  -3,   6,  -1,   7,   3,  10,   4,  -2,  -3,   1,  -5,   8,   2,  -3,   1,  -2,   1,   7,   2 },
      { -17, -57,  15,   6, -12,   7,   4,   4,  29,  53,  -1,  -9, -11,   3,  -4,  -2, -25,  20,  38, -22,  -9,   8,  -7,  -3,   5, -45,  45,  -9,   4,   3,   2,   0,  13, -22,   4,  -2,   4,  -1,   5,   0,   9,   4,  -6,   2,   0,  -2,   0,   0,  10,   7, -10,  -2,   0,   0,   1,  -1,   0,  -3,  -2,   1,  -1,   0,   0,  -1 },
      {   5,  15,  32, -10,  35,  -3,  -4,  -3,  -3,  11,  -6,  18,  22,  -9,   4,  -6,   9,  14,  22,   6,  -6,   5,  -6,  -8,  -6,  23,  47, -24, -23,   6, -12,   1,   1,   8,  21, -26,  -8,   5,  -2,  -1, -56, -30, -10,  -5,  12,   2,   9,   1, -37,  -3,  15,   5,   7,  -4,   3,  -1,   0,  19,  -1,   3,  -1,  -1,   4,   2 },
      {  -4, -14,   1, -28, -18,  13, -14,   0,  -2,  36, -23, -14, -23,  15,  -4,   8, -11, -42,   7,   6,  13, -10,  12,   9,  29,  59,   2, -10,  10,  -9,   1,  -5, -20,   3, -11,   0,   9,  -8,   2,  -2, -48,  24,   4,  -7,   4,   3,   1,  -2,  15,   8, -23,   5,  -3,   2,  -1,   3,   2, -14,   0,   4,   1,   0,  -3,  -1 },
      {  10,   5,  42,  48,  23, -32,  10,  -4,   9,  22, -10,   8,  29, -35,   4,  -4, -11, -29, -28, -49,  28,   3,   2,   0,  13,   2, -13, -24,   8,   6,   0,   2,   2,  18,   5,   2,  -9,   8,   1,  -1,   1,  24,  -1,   2, -10,   0,   2,   0,  18,  -1,  -6,  -1,  -6,   0,  -1,  -2,   6,  -8,  -2,   3,  -2,   2,   4,   1 },
      {   2,  -1, -28, -23,  57,  -7,  21,   1,   5,   3,  13, -10,  24, -10,   8,  -8,  13,  -3,  32,  31,  -7,  -1,  -5, -10,  -6,  -7, -28, -21,  -5,  -7, -11,   6,  30, -19,   5, -19,   7,  -2, -10,   3,  -4,  46,  -4,  -8,   5,   1,   4,  -3,  35,  11, -15,  10,   2,  -3,  -3,   3,   0, -10, -14,  10,   1,   3,  -3,  -1 },
      {   0,   6,   2, -37,  -2,  18,   3,   8, -13, -36, -43, -41,  -8,   7,  17,   2,   3,  13, -15, -34,  20,  -2,   6,  -8,   1, -31,   4, -51,  34,  -5,  -8,   3,   9,  11,   8,  -6,  16,  -8,  -3,   5,  -1,   3, -17,  11,   2,  -3,   8,   1, -12, -18,  -3,  22,   5,   0,  -1,   0,  -5,   0,   1,  11,   0,   2,   0,   0 },
      {  11,  23,  26,  -1, -16,  -8,   6,   3, -10, -33,  -8,   4,   4,   3,   5,   4,  15,   3,  -5,  -8,  -7,  11,  -2,  -1, -11,  31,  42, -20, -10,   3,   2,  -3,   3, -73,  14,  16,  -1,  -3,   3,   7,  19,  -9,  11,   1,  -1,   4, -10,   4,  38,  13, -29,  -6,  -4,   4,  -1,   4,   0, -21,   1,   0,   4,  -1,  -9,  -2 },
      {   7,  13,   0,   3, -19,  14, -11,  -6,   2,  -8,   4, -23, -10,  -8,  10,   0,  18,  33,  -2,  -9,  11, -11,  12,   1,   8,  12,  30,  42,  -2,  10,  -4,   0,  42,  41,  39,  -9, -18,  12,  -2,  -5,  -2,  28,   6, -26,  -3,  -6,   7,  -3,  19,  14, -19,  -9,   6,  -3,   1,  -8,  28,  -3, -20,   3,   1,   0,  13,   6 },
      {  -2,   3, -23,  13,  -6,  -3,  13,  12,  -2,   4,  33,  -8,   3,  15,  -8,   1,  -8,  10, -53, -10, -26,  28, -14,  -1, -21,  -5,  12, -22,  -1,  -1,  12,  -1,  20,  17,  -9,  11,  11, -17,   7,   5, -41,  18,   2,   6,   4,  -1, -10,   3, -27,  51, -20,  -6,  -3,   5,   0,   5,  -5, -11,   0,  -2,  -1,   0, -17,  -6 },
      {  -7, -14, -37,   4,  -1,   1, -36,   0,   4,   8,  56, -17,  11,  -9,   0,  11,  -3, -21, -21, -12,  29, -38,  24,  13,  -7,  -5,   5, -22, -12,  -3,   0,  -6,   2, -24,  39, -12,  -6,   6,  -5,   1,   0, -29,   2, -10,   4,  16,  -2,   1, -12, -23, -16,  10,   6,   4,   0,   7,  -7,  -3,  -4,   5,   4,  -3,  -5,   1 },
      {  -1,  -9, -13,  -3,  -3,   0,  11,   3,   1,  -3,   1, -18,   3,  -6,  -1,  -4,  -4, -14, -27,   2, -18,   2,  -6,  -3,   9,  -8,   7,  -7, -15,  -8,  -8,   3, -13,  12, -17, -19,   7,   7,  -1,  -4,  -8, -51, -21,  14,  17,   4,   1,   1,  57,  11,  17,   2,   1,  -6,  -1,   0,  68,  10, -22,   0,  -4,  -1,   1,   1 },
      {   2,   7,   7,   2, -17, -13,  -6,   4,   1,  -7,   4,  26,  26, -23,  14,   3,   3,  20,   5,  42,  48, -37,  11,   0,  -1, -28,  25,   5,  28,  -9,   4,  -6,   4,   1, -52,  16,   1,  -5,   3,   4, -22,  -4, -21,   6,  -6,   6,  -3,   2,   1,  14, -22,   8,  -2,   9,  -3,   2,  12, -13,  -2,  -1,  -4,   0,  -7,   2 },
      {  -6,  -9,  -9,   6, -27, -50,  53, -11,  -5,   0,   1, -29, -33, -51,  38,  -9,  -5,  -6,  -8,  14, -17,   0,   6,  -2,  -3,  10,   8,  18,   8,   5, -10,   3,  -4,  -9,   4, -11,   9,   7,  -8,   5, -17,  -3, -17,  -8,   6,   2,  -7,   6, -11, -19,  -1,   1,   8,  -1,  -3,   2, -27,   5,   3,   3,   5,   3,  -1,  -1 },
      {   0,   1,   3,  -5,   8, -17,   6,  10,   5,  10,   2,  -7,  -2, -11,  17,   3,  -1,  -3,  14,  21,   2,  -1,   8,   1,  16,  15,   1, -24,  37,   0,  -1,  -4, -18,  11,  19,   0,  22, -10,  -9,   9,  47, -19,  44,  -8, -10,  -1,  -6,   2, -37,  48,  -4,  -7,   2,   1,   2,   1,  30,  20, -26,  -1,   1,   1,  -6,  -4 },
      {  -3,  -2,  -2, -29,  -7, -31, -32,  -9,  -5,  -7, -21,  22, -20, -17,   6,   4,   0,  -2,   8, -33, -31, -26,  23,  15, -11, -12, -19, -16, -38,  26,   5,  -5,   9,  -9, -30,   9, -12,  19,  -2,  -1,   5,  19, -12,  -8,  13,   7,  -7,  -1, -20,  25, -14,  -3,  14,   0,  -2,  -2,  20,  19, -13,  -4,   5,  -8,   2,   0 },
      {  -3,  -5, -10, -46, -24, -16,  22, -16,   0,   3,  15,  36, -11,  -6, -23, -11,   4,   3, -16,  -6,  39,  -5, -15,   7,  -2,   8,  13, -19,   9, -22,  11,   5,  13,  -5,  17, -14,  -9,   8,  14, -16,   7,  24,   1,   9,   1,   3,  -3,  -6,  14,  14,  47, -13, -14,  -4,  -8,   1,  -9,  34,   0, -21,  -2,   2,   2,   1 },
      {  -2,  -1, -10, -36,   8, -24,  -6,   5,   4,  -1,  19,  44, -24, -14,  13,  11,  -1,   5,   6,  -6, -18,  19,   4,   5,   4, -13,   0, -11,  19,  25,  -7,  -1,   8,  29,  16,  29,   3,  -1,   2,  -1, -22, -20,  42,  13, -18,  -3,   4,   0,  26, -34,  -5,   3,  -8,   3,   2,  -2,   9, -36,   9,   3,  -3,   0,   1,   1 },
      {  -5,  -8,  -6, -18, -54,   7,  20,   4,   2,   4,  -3,   6,  36,   4,  11,  -8,   1,  -1,  11,   7,  38,  48, -12,  -8,  -2,  -4, -29, -18, -27,  33,  -3,   6,  -9,   6,  -3, -22, -26,  11,  -2,   5,   4, -10,   8, -25,   9,   1,  -4,  -1, -13,  -2, -18,  -8,  12,  -3,   3,  -1,  11, -21,   4,   8,   2,  -3,  -1,   2 },
      {   0,   1,   8,  -2, -32,  17,  26,  18,   0,   6,  -9,   7,  27,   1,  22,  -6,  -2,  -6,  19, -13, -17, -45, -18, -24,  -4,  -4, -23,  -1, -28, -33,  -7,  -4,   5,  16,  32,  44,   1, -14,  -4,  12, -17,  -2,  14,  18,   1,  16, -13,   6,   5,  -7,  -2, -10,   9,  14,  -6,   8,  -3,  17, -16, -14,   1,   0, -14,  -3 },
      {   1,   3,  13,  13, -21, -26, -22,   3,  -5, -10, -16, -41,  11, -14, -30,  16,   2,   9,  13,  18, -16, -17,  -3,  10,  -5, -16,   2, -24, -14, -19,  10,  -3,   8,  15, -18, -32,  -1,   5,  -1,  -3, -11,   3,  56,  -9,  -2,  12,  -8,  -3,  12,  -3,  34, -20,  -3,   3,  -5,   3, -17, -26,   8,  -2,   4,   4,  -1,  -4 },
      {  -3,  -3,   1,   4, -18, -48, -25,  30,  -4,  -5, -11,  -9,   1, -11, -54,  18,   4,   3,  15,   9,   9,  22, -29,  -2,  -6,   6, -18,  16,   7,  -6,  -2,  -2,   8, -13,  33,   8,  14, -15,   5,  -5,  -9,   9, -28,  19,   1,  -2,  10,  -2, -20, -19, -12,  15,  -4,   4,   2,   4,  33,   4, -21,   2,  -4,   4,  -6,  -2 },
      {  -1,  -2,  -7,   8, -21,  20, -30,  23,   2,   7,  12, -16,  23, -23,  10,  -3,   1,  -5, -12,  27, -11,   5,   5, -20,   5,  15,  17, -19,   3,  53, -17,  -7,   0,  -7,  -7,  19,   5,  15,  -9,   4,   5,  38,   6,  44,  -1, -19,   3,   2,   6, -11,  31,   5,   9, -10,   2,  -2,  -3,  25,   5, -12,  -2,  -5,   6,   6 },
      {   0,  -1,   2,  13,  12,   4,  27, -25,   0,   1,   2, -13, -33,  18,  -2, -11,  -1,  -1,  -3,  21,  15,  -5,   0,   4,   0,   0,  15, -18, -26,  -9,  -4,   4,  -2,  -7,  -4,  29, -37,  15,  -8,  -6,   6,  24,  10,  29,  -3,   9,  -5,  -2, -37, -21,  19,  13,  -3,  -3, -10,   7,  41, -41, -23,  11,   3,   4,  -3,   0 },
      {   1,   0,  -1, -11,  -4,  19,  14,  15,   1,   0,   8,  15,  24, -13,  10,   6,  -1,  -4,  -6, -22, -22, -14,   2,  -4,  -2,   7,   8,  14,  15,  -8,  10, -12,   1, -22,  -6, -18,  36,  -2,   4,  -1,   2,  30,  -1, -35,   3,   5,  -7,   0, -29, -30,  27, -14,  -7,   5,   1,   2,  56, -26,  34,  -2,  -9,   1,  -5,   1 },
      {   0,   2,  -2, -13,   6,  22,  14,  25,  -2,  -2,   1,  14,  -8, -27, -26,  12,  -4,  -6, -10, -17,   3, -32, -48,  -3,   2,   4,   6,  14,   3,  14, -36,  -5, -10,   1, -10,  -6,  16,  31,   1,  -4,  12,  -9,   4,   1,  20,  -8,  10,  -3, -12,  20,  -1,   2,  12, -11,  -2,  -2, -30, -45, -36,   8,   1,   2,  11,   4 },
      {   1,   3,  -2,   5,  25,   0,  11,  59,   0,   0,   6,  -8, -37,   1,   3,  43,   0,   2,  -5,   2,  28,   7, -18,  -7,  -1,  -4,   9,  -8, -12,   3, -23,  -5,  -3,  -1, -20,  20, -25,  -1, -13,   4,  -3,   6,   2, -42,  -7,  19,  -2,   2,  10, -10,   7, -23,   6,  13,   4,   1,   7,  36,  17,   0,  10,  -1,  -7,  -8 },
      {   0,   0,   0,  -5,  -5,  25,  14,  26,  -1,   0,  -4,  -1, -11, -34, -11,  15,  -1,  -2,   3,  11, -26, -16,   5,  -9,   1,   4, -14, -11,   3,  -5,  24,   5,  -4,  -1,  19, -22, -48,  20,  13,   1,   3,  -9, -21,  23, -52,   7,  11,  -2,  -7,  29, -13,  14, -19,   4,   3,  -9,   5,  -9,  39,   7,  -8,   6,   4,  -1 },
      {  -2,  -4,  -1,  -1,  -1,  -5, -28,   7,   1,   2,  -2,   2, -14, -20,  23,   3,   2,   1,   9,  -4,  21,  19,  27, -21,   1,   0,  -7,   7, -28, -22, -23,  -1,  -1,   5,  20,  18,  11, -24, -15,  -3,  10,  -4, -25,   6,  27,   3,   5,  -3,   2,  46,  43,  -5, -14,  -5,   8,   3,  -7, -42,  31,  10,  -2,   8,  -6,  -1 },
      {  -1,  -1,  -2,   2,  -8, -31, -10, -28,   1,   1,   4,  -7,   5,  39,  11,  -7,  -2,  -3,  -1,  -2,  -4, -31, -50,  16,   1,   2,   0,   3,  12,  13, -54,   7,   2,  -1,   6,  -8, -10,   0,  -8,  23,   3,   4,   1,   3, -11,   6,  15,   7,   2,  19,   2,  16,  10,  11,   7,  -8,  15,   0,  54,   2, -13,  -2,  -2,  -6 },
      {   3,   5,   9,   4,   1, -12,  35,  28,   2,   2,   1,   7,  -2,  25,   9,  41,   1,   5,   5,  -8,  20, -12,  20,  19,   0,  -2,   2,  21, -26,  17,  20, -12,   4,   5,   3, -30,  17,  -1, -10,  -7,   2,   5,  25,  38,  29, -16, -13,  -4,   5,   7,  -2,  50,   2,  -3, -13,   0,   1,  10,  20,  10, -12,  -7,  -8,  -1 },
      {   1,   1,   4,   7,  15,   1,  15, -13,  -1,  -3,  -7, -12, -11,  11, -34, -14,   1,   3,  12,  15,  11, -18,   1, -15,   1,  -2, -12, -12,  18,  48,  29, -18,   6,  11,  27,  25,  23,  38, -11,  -8,  -7, -15, -22,  -9,  24,  16, -29,   2,   9,  12,   1, -24,   6,   9, -16,   8,   3,  -8,  32, -11,  -1,  -7, -13,   3 },
      {   0,   0,  -2,   4,   4,  30,   1, -27,  -2,  -4,  -2,  -6, -13, -56, -22, -14,   1,   3,   0,   8,  10,   4, -21,  12,  -1,  -3,   3,  -9, -24, -14,  -2,  20,   2,   6,  -6,  22,  19, -13,  11,  -8,   7,  -1,  31,  -2,  33, -10,   8, -12,  -4,  -7, -38,  14,  -4,  -5,   6,  -3,  14,  27,  44,   6, -11,   3,  -2,   1 },
      {   1,   2,   5,   8,  -9,  14,   8,  -5,   2,   2,  -2,  -5,  23,   0,  22,  47,   1,   2,   8,  16, -17,  13,  -2,  78,  -1,  -3,  -9, -17,  15,  -9, -19,  10,   4,   1,  16,  27,   0,  18,   4, -32,  -3,  -1, -26,  -6,  11,   6,  -6, -16,  -1,   3,   6, -10, -11, -21,  -4,  -5,  -4,  -1,  -3, -11,  -6, -10,  10,   3 },
      {  -1,   0,  -1,  -4, -16,   4,   2,  -6,   0,   1,   1,  11,  20,  -4, -29,  -3,   0,   2,  -3,  -7, -18,  22,  17,  -8,   3,   2,  10,  10,  16, -14, -12,   3,  -1,   4,  -5,  23,  -1,  26, -32,   2,   2,   4,   7, -25,  -6,  49, -10,   2,   6,   8,  15,  54,   3,   3, -15,   4,  -9,  18,  -2,  56,  -4,  -8,  -5,   1 },
      {   0,   1,   0,   2,  -1,  14,   0, -17,  -2,  -3,  -5,  -3,  -7,  -1,   6,   4,   1,   3,   2,  -3,  15,   9,   0,   2,  -3,  -4,  -2,   6, -32,  23, -29,  -8,   2,   1,  -6, -20,  47,  12,  -9,  -4,  -3,   2,   1,  13, -42,  41,   0,  -4,  -6,   0, -12,  -7, -69,  23,   3,   8,  -4,  10, -16, -16,  -9,   7,  -8,   2 },
      {   1,   1,   3,   0,  11,  27, -23, -15,  -2,  -1,  -4,   3, -10, -31,  15,   3,   0,   1,   6,  -9,   3,  24, -28,  21,  -2,  -1,  -7,  17,  21,   4,  -5, -22,   0,  -2,   1, -27, -25, -22,   4,   0,  -3,   3,   2,  20,   4,  23, -47,  20,   3,   5,   7,  14,  27,  18, -26,  30,   5,  -2,  -2, -18,  25,   2, -30,   7 },
      {   0,   0,  -2,   2,   7, -16, -26,  25,   0,   0,   1,  -2,  -5,  15,  36, -19,   0,  -1,  -2,   1,  -3,  15, -17, -29,   1,   2,   1,   8,  17, -33,   5,   0,  -1,   1,  -3, -10,  -8,  50,  19, -23,   4,   6,  15,  20,  38,  41, -17, -16,  -7,  -6, -26,   1, -14,  -6,  14,  -9,   0,   1,  10, -15, -11, -10,  19,   5 },
      {   0,   0,  -1,  -9,   2,   4,   4,   0,   0,   0,   1,  11,   4,  -2, -26,   3,   1,   2,  -3,  -8,   2,   7,  22,   7,  -1,  -1,   3,   7,   9, -24, -23,  23,  -1,  -1,  -3, -18,   2,   7, -72,  15,  -4,   2,  -1,  40,  -4,  -6, -33,   2,  -7,  -7, -27, -48,  17, -11,   8, -11,   6,   1,  15,  -3,   9,   4,  17,  -6 },
      {   0,   0,   0,   1,   5,   8,  -2,  14,   0,   1,  -1,   1,  -4,  -6,   4,  -3,   1,  -1,   1,   2,   6,   5,   1,  25,   2,   5,   3,   5,  -2,  15,  31,  71,   0,   0,   3,   2,  13,   1,   5,  50,   2,   4,  -4,   4,  17,  37,   9,   0,   2,   1,  18,   9,  16,  45,   3, -30,  -3, -12, -13, -21,  -6,   6,  10, -12 },
      {   0,   0,   0,   1,   2,   1,   9, -23,   1,   1,   2,   3,   2,   1,   1,  26,   0,   0,  -1,   4,   9,  15,  10,  -5,   0,   0,   0,   4, -11, -14,   0, -24,   2,   2,   1,  -1,  33,  18,  38,   0,   3,   6,   2,  29, -19,  34,  44,   0,   0,   4,  -3, -24,  60,  -1,   5, -10,   3,   2,  12,  20,  37, -15,  -2,  -6 },
      {   0,   0,  -1,  -1,   2,   2, -18,   1,   0,   0,  -1,   2,  -7,  -2,  24,  -4,   0,  -1,   1,   3,  10, -11, -38,  14,  -1,   3,  -3,   1,  -7,  10,  32,   8,   3,  -1,   4,  -2,  12,  -4,   9,   9,  -5,   2,  -7,  13, -11,  -4, -39,  13,   3,  -3,  11, -29, -15, -17,  19,  -6,  -1,   9,  -8,  86,  -6, -12,   5, -17 },
      {   1,   1,   2,   2,   2,  -7,  -3,  47,  -1,  -1,  -1,  -1,  -2,   9,   0, -57,   1,   1,   3,  -3,   6,  -8,  12,  26,   0,   0,   1,   5, -11,   3, -19,  41,   0,   2,   1,  -1,  20,   5,  21, -25,  -4,  -5,   1,  -3, -32,  -4, -10, -21,   0,   4,   0,   4,  11, -22, -31,  17,   4,  -1,  15,   4,  44,  -2, -16,   5 },
      {   1,   1,   0,   1,  -2,   0,  10,   5,  -1,  -1,   2,   0,   5,   2,  -4, -14,   0,  -1,  -4,   0,  -9, -11,  10,  -2,  -1,   2,   1,  -1,  11,  24, -15,  -4,  -3,  -6,  -2,  -8, -16, -47,   1, -63,   4,  -1,   2,  15,  14,  21,   3, -33,  -2,  -1,  -1, -23,  12,  55,  -8, -27,  -3,   2,  -4,  32, -14,  16,  15,   6 },
      {   1,   1,   1,   1,   2,   2,  12,  13,  -1,  -3,  -2,  -3,  -7,  -7,  -9, -26,   0,   1,   0,  -2,   4,  -6,  -4,  15,   0,  -1,   1,   2,  -2,  14,  12, -17,   0,  -1,   0,  -3,  -5, -46, -22,  -3,  -1,   2,   1,   1, -10,  46,  13,  -2,   1,   1,   3,   6,  10, -45,  29,   6,  -3, -10,  -2, -22, -22, -69,  18,   0 },
      {   1,   1,   1,   0,   1,   1,  12,   7,   0,   0,  -1,   1,   3,   4, -19,   5,   1,   1,   4,  -2,  -6,  -7,  37,   0,  -1,  -1,  -3,   7,  18,  12, -34,  25,   2,   2,  -1,  -9, -22,  -7,  47,  22,   2,   4,   7,  15,  36,  19,  13,  43,  -3,  -4,  -5, -32, -24, -14,  -5,  39,   0,   8,   0,  20, -20, -10, -16,  11 },
      {   0,   0,   1,  -1,  -1,   1,   3,  26,  -1,  -1,  -2,   1,  -1,   2,   3, -38,   1,   2,   1,   2,   2,   7,  -1,  47,  -1,  -1,  -1,  -2,  -6, -11,  15, -30,   1,   1,   1,   5,   5,  21, -21, -13,   1,   2,   2,   4,   3,  -4,  32,  53,   4,   4,   0,  -2,  11,  29,  36,  33,  -2,   2,   5,   7, -11,  36,  16,  22 },
      {   0,   0,   0,   0,  -5,  -1,  10,  -6,   1,   1,   0,   2,  11,   3,  -4,   9,  -1,   0,   0,  -1,  -9, -14,  -6,  -3,   2,   0,   2,   2,  13,  16,  10,   7,   0,   2,   1,   1, -12, -12, -11,  -2,   1,   1,   2,   4,  28,  10,  10,  -8,   0,   1,   2,  13, -40,   3,  51,  10,   5,   2,   5,   4,  95,   2,   9,   5 },
      {   0,   0,   0,  -1,  -1,  -3,  -9,  -9,   0,   0,   0,   1,   0,   2,  17,  18,  -1,  -1,  -2,  -6,  -4,  -5, -20, -10,   2,   2,   4,   6,   6,   5,  31,  35,  -3,  -1,  -5,  -4,  -4,   4, -36, -10,   2,   0,   1,   6,  -3,   7,  38, -36,   1,  -1,   0, -19,  11, -14,   3,  37,  -4,   2,  -4,   8, -18,   9, -56,  52 },
      {   0,   1,   1,  -1,   0,   7,  -3,  11,  -1,  -1,  -2,   2,   2,  -6,   1, -28,   1,   0,   3,  -4,  -2,   3,   5,  22,  -1,  -1,   0,   6,  11,   0,   0, -28,   1,   0,   0,  -5, -16,  14, -16,   5,  -1,   0,   3,   7,  20,  -1,  34,  -9,   1,  -7,   0, -11, -18,  -8,  -2, -39,  -3,   2,  -7,   3,   2,   1, -72, -64 },
      {   0,   0,  -1,   0,   0,  -3,   3,   8,   0,   0,   1,   0,   0,   4,  -3,  -9,   0,   0,  -1,   0,   0,  -1,  -2,  14,   0,  -1,   0,  -2,  -1,   8,   4, -35,   1,   1,   0,   4,  -2, -19,   8,  44,   0,  -1,  -1,   1,   7,  32,  -3, -19,   1,  -2,  -3,  -4,  -6, -50, -21, -35,   0,   5,   3,   4,   4,  66,  11,  44 },
      {   0,  -1,  -2,  -1,  -1,  -2,  -7,   6,   0,   1,   2,  -1,   0,  -3,  11,  -6,   0,  -1,  -2,   2,  -2,   5, -17,   6,   0,   1,   0,  -2,  -1, -11,  14, -11,  -1,   0,   0,   0,  -2,  -2, -24,  16,   3,  -1,   0,   4,  14, -11,  39,  13,  -4,   0,  -7, -16, -31,  22, -79,  -8,   2,   1,   7,  12,  22, -47,  28,  25 },
      {   0,   0,   1,   1,   3,   1,   7,   7,  -1,  -1,  -2,  -2,  -5,  -3, -12, -15,   0,   1,   1,   2,   1,   6,  13,  17,  -1,  -1,  -2,  -2,  -1, -15, -19, -24,   1,   1,   3,  -2,  -3,  14,  17,  37,  -2,   1,  -2,   1,  -1, -26, -23, -39,   2,  -1,   8,  -2,  -2,  38,  37, -19,  -2,   0,  -5,  10,  -3, -44, -27,  59 },
      {   0,   1,   0,   1,   0,   2,   1,   3,   0,  -1,   0,   1,   1,  -2,   2,  -7,   0,   1,   0,  -2,  -2,   4,   3,  10,   0,  -1,   1,   3,   5,   2,   1, -21,   1,   1,   1,   0,  -6,   5,   4,  38,  -1,  -2,  -3,   5,   6,  -3,   8, -71,   2,   1,   1,  -7,   2,  12,  -7,  72,   1,   4,   2,   8, -10,   9,  36, -45 },
    },
  },
  { // 15
    {
      {-115,  34,   7,   5,   2,   2,   1,   1,  41,  -7,  -3,  -2,  -1,  -1,   0,  -1,   4,  -4,  -1,  -1,   1,   0,   0,   0,   5,  -1,  -1,   0,   0,  -1,   0,   0,   8,  -3,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -34,  -1,   6,   0,   0,   0,   1,   0,-100,  26,   7,   5,   2,   2,   1,   1,  64,   3,  -9,  -2,  -2,  -1,  -1,  -1,  10,  -8,   0,   0,   0,  -1,   0,   0,   4,  -2,   1,  -1,  -1,   0,   0,   0,  -1,  -1,   1,   1,  -1,   0,   0,   0,   5,   2,  -2,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
      {  19,  -5,  -4,  -2,   0,   0,  -1,   0,  56,   9, -11,  -2,  -4,  -1,  -1,   0,  99, -21,  -7,  -2,  -4,  -1,  -2,   0, -25, -17,   3,  -1,   1,   0,   1,   0, -34,   2,   3,   1,   2,   1,   1,   0, -12,   4,   2,  -1,   1,   0,   0,   0,  -1,   1,   1,   0,   0,   0,   0,   0,  -3,  -1,   0,   0,   0,   0,   0,   0 },
      {  -5,   5,   3,  -2,  -3,   0,   0,   0, -12,   2,   3,   5,   1,  -1,   0,  -1,  -4, -27,   4,  -2,   4,   1,   0,   1,-108,  23,  16,   3,   3,   0,   0,  -1,  40,  24,  -3,  -3,  -4,  -2,  -1,   0,  22, -10,  -5,   3,  -2,   0,   0,   0,   2,  -2,  -4,   1,   0,   1,   1,   1,   0,  -1,   0,   1,   0,  -1,   0,   0 },
      {  16,   1,   4,  -2,   3,   0,   0,  -2,  18,  19, -12,  -2,  -1,   0,   1,  -1,  21,  -7,  -5,  -4,   1,   2,   0,   0,  22,  18, -14,   1,  -1,   0,  -2,   0,  98, -33, -16,   0,  -3,   0,  -1,   0, -42, -26,  10,   1,   2,   1,   1,   0, -15,   1,   2,   0,   0,   1,   1,   1,  -8,   3,   2,   1,   0,  -1,   0,   0 },
      {  26, 100, -33,  -5, -13,   1,  -5,   0,  -4,  -9,   3,   4,   2,   1,  -1,  -1,  12,   8,  -9,  -1,   2,   3,   0,   0,  14, -14,   8,   1,   1,  -1,  -1,   0,   7,   9,   3,  -3,  -4,  -2,  -1,   0,  40, -26,  -7,   4,  -3,   0,   0,   0, -30,  -9,   3,   0,   2,   1,   2,   0,   1,   4,   0,   2,   0,  -1,   0,   1 },
      {  -9, -51,   8,   0,   6,   2,   2,   2,  19,  17,   2,  -3,  -4,  -2,   0,   0,  17,  26,  -2,   7,  -9,  -4,  -4,   1,  24,  17,  -7,  -1,  -4,  -1,  -2,  -1,  17,  33, -12,   4,  -1,   0,  -2,   0,  80, -28, -14,  -4,  -1,   0,  -4,   1, -27, -28,  14,   1,   3,  -1,  -2,  -2, -13,   2,   5,  -3,   0,   2,   1,  -1 },
      {  12,  32,   2,  -3,   3,   0,  -1,  -2,  10,  51, -20,  -8,  -2,   2,   0,   0,   1,  16,   0,  -3,   0,   1,   0,  -1,   6,  32, -16,  -4,   0,   1,   0,   1,   6,   1, -10,   3,   2,   0,   0,   0,  21,  37,  -9,  -5,   0,  -1,  -1,  -1,  80, -26, -18,   0,  -4,   0,  -2,   0,  -8, -31,   5,   1,   1,   1,   1,   0 },
      {   1,  -4, -23,   1,   2,   2,  -1,  -1,  -2, -85,  37,   2,   3,   2,   5,   1,  30,  45,  -6,  -1,  -7,  -2,  -3,  -3,  -7,  51, -17,  -3,  -2,  -2,  -3,   0,   5, -12, -11,   5,   0,   3,   2,   1,  -4,   2,  -2,   2,   0,  -1,  -2,   0,  27,  -6,   2,   2,  -1,   0,  -2,   0,   3,  -6,   2,  -2,  -1,   1,  -1,  -1 },
      {   0,   5,  19,  -3,  -3,  -5,  -1,  -1,  14,  44,  27,  -2,  -1,  -7,  -2,  -2,  -6,  86, -30,  -2,  -9,   0,  -3,   0, -35, -14,  -5,   2,   3,   1,   0,   1,  -9, -24,  10,  -1,  -1,   0,   0,   0,  -7, -23,  12,   2,   2,   2,   0,   1,  -6,  31,   0,  -4,  -1,   0,   1,   0,  24,  -1, -11,   0,   0,  -1,   0,   1 },
      { -15, -29,-101,  -7, -12,   2,   3,   4,  -4,  11, -45,  -3, -10,  -1,   1,  -4,  -9,  16, -25,  -2,   8,  -4,  -2,   1, -11, -12,   4,  10,   1,  -6,  -7,  -4,   6, -10,  16,   6,  -2,   2,   1,   3,  -4,   3,   8,  12,  -4,  -3,  -2,   0,   6, -10,   5,   8,   5,   0,  -1,   0,   4,   4,   6,   2,  -3,  -1,  -1,  -1 },
      {   5, -12,  31,   3,   6,  -3,   3,  -3,   1, -41, -16,  11,  10,   8,   2,  -1,   0,  -1, -47,   0,  11,  11,   1,   3,  -5, -67,  -1,   5,   8,   2,   0,   3,  27, -35,  21,  -7,  -2,  -2,   2,  -1,  37,   5,   0,   3,  -3,   1,   4,   0,  30, -15, -21,  -3,  -4,   2,   4,   4, -20,  -6,   3,   5,   2,  -3,   0,   2 },
      {  -3,  -6,  -7, -24, -25, -11,  -5,   4,   8,  22,  74,  -3, -17, -16,  -8,  -4,  -6, -25, -17,   4,  -3,  -8,  -4,   3,   8,  15,  37,   8,  -1, -10,  -5,  -6,  -5, -32,  31,  -4,  -7,  -1,   1,   1,  -3,  -2,  -5,  13,  -8,  -2,  -3,   2,   6, -29,  -3,  11,   5,   1,   1,  -1, -41,   0,  15,   3,  -1,  -1,   1,  -1 },
      {   3, -12, -14,  -3, -11,  -5,  -3,   3,  14,  -4,  27,   7,  -4,  -1,  -4,   0,   9,  -2,  -2,   6,  -5,   0,   1,   0,  29, -16,  47,  -1,   2,  -4,   3,  -1,  40,  51,  21, -13,  -5,  -3,   0,  -2,   4,  17,  -7,  -3,  -4,  -2,  -1,  -1,  25,  42, -20,  -5,  -2,  -2,   0,   0,  49, -20, -22,   3,  -2,  -1,  -2,   0 },
      {   1,   1,  23, -46,   2,  -9,   3,  -2,  -3,  -7, -32, -11, -10,  -5,   3,  -2,   0, -33, -36,   8,   6,  -1,  -4,   6,  12,  51,   0,   5,  -2,  -7, -10,  -5, -16, -31,  18,   9,  -3,   2,   0,   3,  29, -15,  15,  14,  -3,  -1,  -1,   1,   2,  22,   2,   7,   2,  -1,  -1,  -1,  53,   8, -13,   1,  -5,  -1,  -1,  -2 },
      {   1,  -4,  -9,  63, -18,   8,  -4,   4,   2,   7, -16,  48,   0,   6,  -7,   4,   6,  -1,  36,  22,  -8,   5,   2,   2,   2,  28,  26,  -9,  -4,   1,   5,   2,  -5, -53,  22, -15,   2,  -6,   0,  -2,  29, -10,   6,  -7,   4,   1,   4,  -1,   4,  27,  -9, -11,  -3,   0,   1,   0,  -5,   3,  -4,  -1,   3,   0,   1,   3 },
      {   0,  -3, -32,  11,  25,  11,   0,   4,   2,  11,  49, -36,   0,   1,  -6,   2,  -4, -37,  17,  11,  -6,   0,   0,  -1,  -5, -32, -31,  -4,   1,  10,   5,   4,   5, -39, -34,  16,   6,   6,  -1,   1,  35,  -3,  15,  -9,   6,  -1,  -2,  -1,  16,  16,  17,  -3,  -3,  -2,  -2,  -3,  31,  -8,  -6,  -6,  -1,   3,   2,   0 },
      {  -1,   4,   4, -53,  13,  -9,   4,  -1,  -2,  -8, -13, -24,  -7,  -4,  -9,   1,  11,  30,  57,  11,  -8,  -9,   2,  -4,  -8,  -9,  21,   0,  -4,   3,   8,   2,  21, -29,  18,   1,   4,   1,   0,  -1,  19,  61,  -2,  -9,   1,  -3,   2,  -2, -22,   7,   0,   0,   2,  -1,   0,  -2,  -9,  27,  -1,  -4,  -1,   0,   1,   0 },
      {  -4,  -6,  -5, -56,  15,  -1,  -1,  -4,   0,  -6, -13,  15,   8,   8,   3,   0,  -1,   8,  47,  -4,  -1,   6,   0,  -5,   3, -18,  34,  -1,   1,   0,   0,   0, -13,  -7, -23,  -7,  -3,   1,   0,   1,  -6, -73,   3,   6,  -1,   2,   0,   2,  38,   3,   0,  -3,  -5,  -2,   0,   1, -12, -24,   2,   2,   2,   0,   0,  -1 },
      {   1,   8,   8,  20,  -3, -11,  -4,   4,  -9, -15, -24, -56, -28,  -8, -22,  10,  -2,   4, -27,  53, -32, -12,   3,  -2,  -2,   3,  36,  -9,   1,  10,  21,   4,   0,   9,  -6,   0,  12,   7,   0,  -3, -10, -15,  19, -22,   9,   0,   1,  -3,   7,   6,  22,  -5,   0,  -6,  -2,  -4, -24, -23,   9,  -4,   1,   4,   2,   1 },
      {  -2,   4, -23, -28,  55,  15,   2,  -4,   3,  12,  11,  24,  34,  25,  12,   1,  -5, -10, -37,  22, -18,  18,   3,  -1,   2,  19,  -7,  -7,  -5,   7,   1,   5,  -3,  12,  11,  -4,   1,  -5,  -2,  -2,   2,  11,   3, -23,   7,   3,  -2,   1,  -7,  44,  -4, -17,  -9,  -4,  -3,   0, -48,   5,  -1,  -4,   5,   2,   0,  -1 },
      {   3,   5,   1,   4,  40,  -5,   3,  -2,   2,   7,   6,  31,  -3,  14, -11,   3,  -2,   1,  -5,  41, -16,   6,   1,  -1,  -2,  -7,  10,  -7,  -4,   6,   7,   5,   4,   3,  16,  -6,   2,  -2,  -1,  -3, -20,  -9, -23,  -8,   2,  -2,   3,   0,  24, -54,  30,  10,   0,   2,   1,   0,  44,  54,   4, -11,  -1,  -1,   0,  -1 },
      {  -3,   3,   3, -32, -28,  -8,  -7,   4,  -1,  -4,   1,  68, -40,  -9,  -4,   2,  -3,  -6,   4,  41, -20, -12,  -8,   3,  -6, -16, -45,   7,  -7,   0,  -4,  -2,   6,   5,  -4,  12,   7,   3,  -2,   0,  -2,  15,  16, -10,   7,  -2,  -4,   0, -11,  -1,  16,   3,   3,   0,  -4,  -4,   2, -44,   2,  -3,  -1,   6,   1,  -1 },
      {  -2,  -5,  -6,  -1,  32,  10,   3,   2,  -2,   1,   1,  -8,  16,  11,  -7,   2,   0,   2,   0,   5,   6,  12,   6,  -4,  -5,  11,  12,  -8,   3,   7,   7,   2,  -6, -20,  18, -15,   0,  -6,  -4,  -2, -10,   5, -11, -18,   0,  -1,   3,  -2, -48, -45, -27,   6,  -4,   2,   4,   3,  31, -77,  -4,  17,  -1,   1,   1,   3 },
      {  -1,  -1,  -4,   0, -12, -21,   0,  -6,  -1,  -2,   0, -34,  19,  -5,   2,   1,   2,  -7,  30,   8, -11,   0,  -1,   6,  -1,   0, -58,  16,  -3,   9,   4,   2,   5,  13,  81, -14,  12,  -7,  -4,  -3,  -5, -33, -21,  -4,   4,   2,   2,   1,  15,   6,   5,  -7,  -3,   2,  -2,  -2,  -7,  -7,  -9,  -1,   0,  -1,  -1,   2 },
      {  -1,  -5,   5, -17, -46,  61,  -6,  11,   1,   3,   4, -23,  13,  26,  -1,  -5,   4,   7,   4,  69,  33,  26,  -3,   3,  -1,   0, -12,   6,  11,  -2,  -4,   1,  -3,   4, -14, -10, -10,  -8,  -1,  -1,  -7,  -1,  -7,   8,  -4,  -1,   4,   2,   1,  -2, -22,   4,   1,   1,   5,   1,   0,  16,  -3,   6,  -1,  -4,   1,   1 },
      {   6,   9,  11,  24,  45,  24,  -7,  14,  -1,  -2,  -2,  -1,  16, -26,  13, -15,   4,   8,  15,  23,  12, -15, -18,  10,   2,  -4,  15,  65, -17, -16, -32, -10,   4,   9,  16,  23, -13,  -5,  -5,   4,  -2,   4,  26,  23,  -1,  -4, -10,   5,   4,   3,  12,  16,   7,   2,  -8,  -7,  -4, -15,  13,  -3,  -9,   4,   3,  -3 },
      {   2,   1,  10,  -5, -13,  70, -15,  17,  -2,  -4,  -7,  -9, -26,  31,  15,  10,  -5,  -4,  -4, -43, -67,   4,  -3,  -2,  -2,  -5,   7,  16, -16,   0,  -9,   1,   3,  -9,  15,   4,   1,  -3,   1,  -3,  -2,  -1, -25,  -9,   6,  -2,  -6,   0,   3,   0,  11,   6,   1,  -5,  -7,  -3,   8,  -2,   7,  -4,   2,   6,   0,  -5 },
      {   2,   2,   6,  -3,   7,  26,   7,  11,   1,   0,  10,   3, -25,  34, -16,   1,   2,   4,  10, -25,  30,   9,  11, -10,   3,   6,  -4, -19,  15,   4,  13,   2,   4,  21,  40,   8,  -3,   2,  -2,   0,   4,  -4,  85, -10,  -6,  -8,   1,  -4,   7, -17,  19,  -4,  -2,  -3,   3,   0,  -4,   4,  -4,   1,   1,  -1,   2,   3 },
      {   2,   2,   0,   4,  27,   7,  19,   8,  -1,  -1,   1,   5, -59,  23, -33,   3,   2,   3,  -5,  -7,  49, -12,   9,  -7,   2,   7,  -6,  38,  14,  16,  20,   4,  -3,  -5,   0,  -2, -10,  -3,  -5,  -1,  -1,  -4, -56,   5,  -6,  -2,  11,  -1,   1,  29,  16,  -2,   2,  -1,   4,  -2, -13, -10,  -3,   2,   2,  -1,   1,   5 },
      {  -2,  -3,   6,  -8, -40,  -2,  -2,  -1,   3,   3,  -1,  10,  67,  11, -11,   5,   1,  -1, -10, -11,  15,  -4,  14, -12,   2,   6,  20,  35,  11,  20,  16,   5,   3, -10,  -4,  15,   7,   6,  -3,   2,   1,  15,  -6, -18,  -2,  -1,   0,  -2,   3,  -3,  70,  -9,  -4,  -5,  -4,  -3,   7, -16,  -3,  -9,  -1,   3,  -1,   0 },
      {   2,   2,   7,   4,   5,  10,   5,   9,   0,   2,  -2,  -5,   3,  11,  31,  -2,   0,   3,   2,   9,  26, -14,  -8,  10,  -3,  -8,   5, -79,  -9, -16, -12,  -7,   7,  -4,  17,   4,  -2,  -1,   5,   2,  -1,   8, -27,  40,   6,  -2,  -6,   6,   1,  15,  53,  14,   0,  -4,  -9,  -1,  -7, -26,   5,  -3,  -4,   6,  -2,  -3 },
      {   0,   2,   0,  12, -11, -40,  40, -31,   1,   4,   8, -13, -26,  56,  63,   0,   1,   2,   6,  13,  -1,  34,   3,   0,  -1,   0,  13,  30, -12, -10, -15,  -4,  -1,  -3,  -6,   8,  -5,  -1,   8,   1,   0,   6,   7,  -2,   6,  -1,  -7,   1,  -5,  -8,  10, -14,  -1,   1,  -4,   4,   0,  -7, -11,   3,   7,   0,  -4,  -3 },
      {   0,   1,   1,   4,  -1,   2,  -1, -18,   0,   0,   1,   5,  -1,   0,   6,   6,   0,   1,   6,   6,  -2,   6,  -9,   0,   2,   0,  10, -14,  64,   7,  -8,  -5,   3,   1,  18,  94,  14,   0,  -4,  -1,  -4, -10, -19, -16,   1,  -3,  -3,  -3,  -3,   5, -24,   4,  -3,  -1,  -4,  -2,   1,   3,  19,   3,  -3,   2,  -3,  -3 },
      {   2,   3,  -1,  -5,  -5,  39,  54,  -2,   0,  -1,   2,   5,  -9, -55,  50,   4,   0,  -1,  -4,   3,  24, -32,  10, -10,   0,  -2,   6,  -5,  -7,  12,   6,  -3,   0,  -4,   7,  -8,   4,  14,   1,  -1,  -2, -10,  -6, -50,   4,   2,  -9,  -1,   6, -10,  -5, -19,   0,  -4,  -5,  -2,   1,   7, -17,  -8,   4,   5,  -1,  -3 },
      {   2,   3,   2,   2,  -8,   4,  68,   4,   2,   2,   1,   5,  28,  30, -19,   4,  -2,  -2,   1,  10, -31, -50,  23,  -8,  -1,  -2,  -8,  -2, -22,  -6,  26,  -2,   0,   3,   0,  24, -17,  11,   7,   3,   0,  -6,   6,  34, -19,  -8,  11,  -4,  -2,   3, -26,  16,   3,  -3,   5,   0,   1,  -2,  10,   5,   1,  -3,  -5,   3 },
      {  -1,  -2,  -1,   2,   1,   2, -51, -13,  -1,   0,  -1,  -1,  -4,   8,  -2,  -7,  -1,   0,  -3,   5,  14, -16,  17, -10,  -3,  -2,   4,  -6, -57,   1,   9,   5,   0,  -2,   5,  51, -13,  12,   5,   2,  -5,  -7, -10,   2,  -5,  -2,   1,  -5,   0, -16, -13, -22,   2,  -6,   2,   2, -12,   3, -70,   0,   5,   0,   2,   1 },
      {   1,   1,  -1,   2,   5,  15,  40,  -3,  -2,  -2,  -4,   3,  -1, -26, -20, -10,  -1,  -2,   2,  -1, -34,  29, -21,   7,   1,   2,   2,  -1,  39,   5, -10,   0,   0,   0,  -6,  -1,   3, -10,  -4,  -2,  -3,   9,   1,  38, -10,   3,   8,   1,   0, -13,  19, -11,   1,   5,   0,  -5, -17,  -6, -78,   3,  -2,  -2,   5,   3 },
      {   0,  -1,   0,  -2, -12,  -3,  34,   9,  -1,  -3,  -1,   0,   0, -28, -25,  -4,   2,   2,   0, -10,   7,  66,  -6,  16,   0,  -2,   4, -14, -59,  33,   1,   8,   0,   4,   4,  32, -30, -18, -12,  -2,  -2,  -1,  -4, -11,  16,  -6,   5,   3,   4,   8,   3,  14,   8,   6,   1,  -5,   1,  -2,  22,   4,   0,   3,   7,   7 },
      {   0,  -2,   0,   3,   2,  28, -12, -95,  -1,   0,  -1,   1,  -9, -23,  -5, -24,   0,  -1,   2,   4, -11,  10,  27, -12,   1,   0,  -4,  -3,   2,   6,   5,   2,   0,   2,   7, -12, -30,  14,   0,   5,   2,   5,   2,  18, -26,  -2,   5,   0,  -1,   6,  16, -11, -10,  -3,   2,   3,   6,  -7,  30,   6,  -1,   0,  -1,  -3 },
      {   0,   1,  -1,   0,  -6,  14,  17, -59,   0,   0,   3,  -1,   6,  25, -40,  -9,  -1,   0,  -4,  -9,  17, -25, -53,  21,  -1,  -2,   8,  -3, -25,  -6, -22, -12,   1,   1,  -3,  -7,  49,  -6,   5,  -4,   0,  -4,   2, -23,  22,  -8,  -1,  -2,   2,   1,  -5,  15,  16,   3,  -4,  -4,  -3,   0,  -9,  -5,  -8,   3,   6,  -4 },
      {   0,   1,   1,   3,  -2, -12,   2,   2,  -1,  -2,  -6,  -6,   7,  10, -18,  -2,  -2,  -2,   3,  -2,  -6, -12, -34,   0,  -2,  -4,  -9, -11,  15, -40, -20,  -6,  -3,  -5,   0,  -4, -86,   4,   2,   4,  -2,  -2,  -6, -58, -24,  -6,   1,   1,   3,   2,  11,   4,   4,   2,  -2,  -2,  -4,  -1, -12,   1,   0,   2,  -2,  -7 },
      {   1,   1,   2,   5,   0,  -6,  -2,   0,   0,  -1,  -1,   2,  -1,   0,   9, -54,   1,   1,   1,   1,  -6,   8,  46, -21,   0,  -1,  -2,  -1,   8,  -9,   1,   1,  -1,  -2,  -4,  -4,  17,  -9, -11,   7,   4,  -4,  -3, -28,   0, -12,  -7,   2,   6,  12,   5,  82,  21,  -4,  -1,  -3, -11,   3, -25,  25, -11,   3,   2,   0 },
      {  -1,   0,  -1,   3,  -2,  -9, -14,  -4,   1,   1,   0,   1,  -1,  16,  22,   7,   0,   0,   1,   2,  -3, -25, -21,  14,   2,   1,   4,  -4,  10, 101, -19,   5,  -1,   0,  -2, -15, -27,  13,  -4,  -5,   0,   1,   9,   8, -19, -11,  -7,  -1,  -1,   0,  -9,  32,   6,   1, -11,  -5,  -4,   2, -11,  -6,  -4,   4,   1,  -2 },
      {  -1,  -2,  -2,  -4,  -3,   2,  -2, -12,   0,  -2,   1,  -2,   1,   5,   1,  -4,   1,   0,  -1,  -2,  -4, -37,  47,  47,   1,   0,   2,  -2,  18,   1, -16,  19,  -2,   0,  -6,   1, -29, -63, -24, -12,   0,   0,   7,   1,  53,   9,   7,   2,   3,  -8,   1, -19,  16,   9,  -1,  -4,   2,   1,  -3,   1,   2,  -2,   7,   7 },
      {   0,   0,   2,  -1,  -3,  -5,   4,  24,  -1,  -2,  -1,  -2,  -1,  13,  -3, -95,  -1,  -3,  -1,   0,  -8, -11, -18, -31,   0,   1,   2,  -8,   4,  23, -10,  -3,   0,  -1,   2,   2,   0,   5,  12,   8,  -1,   0,  -5,   2,  15,   4,   5,   5,  -1,  -4,  -3, -51,  25,   6,   0,   1,   8,  -2,  19,  -4,   3,  -8,   3,   3 },
      {  -1,  -1,  -1,   0,   0,   4,  -6, -14,   0,   1,   0,   3,   3,  -6,  10,   2,  -1,  -2,  -1,  -2,   1,   0, -36, -33,   2,   2,  -1,   9,  13,   0,  53,  -2,   0,   2,   5,  -2, -35,   2,  27,   2,   4,   2,   4,  17,  82,  -1,  -8,   1,  -1,  -3,  -5,  28, -13, -13,   1,   9,  -2,   2,  -9,  -1,  -3,   2,  -2,   3 },
      {   0,  -1,  -1,   0,  -3,  -1,  -5,  -5,   0,   0,  -1,  -2,  -4,  -3,  23, -19,  -1,  -2,  -3,  -2,   1, -11, -34, -11,   0,   0,   0,   0,  -8,  -3,  44,  15,  -1,  -2,  -5,   8,   8, -84,  -6,  -3,  -1,   0,   6,  -4, -48,  -5,  10,   8,  -1,  -1,   0,   9, -24,  -1,   1,   3,   2,  -3,   3, -35,   7,  -4,  -3,   1 },
      {   0,   1,   1,   0,   3,   2,  -1,  -4,   0,   0,  -1,   0,   2,   1,   0,  -5,   1,   1,   1,  -2,   2,  20,  18,  24,   0,   0,   0,   1,  11, -11,  17,  -8,   0,  -1,   3,  -3,  -6,  23,   7,   2,  -1,  -1,  -1,   2,  -1,   4,   1,  -5,  -3,   0, -13,  14,  52,  -8,  -6,   2,  -1, -13,  -3,-101,  12,  -6,   1,   0 },
      {   0,   0,  -1,  -2,   0,   4,  -5,   7,   0,   0,   1,   3,  -2,   4,  18, -41,   1,   2,   2,  -5,  -3,  -1, -13,  84,   1,   3,   3,   7,   3,  -9,  50,  15,   0,   0,  -2,   5,   2,  44,  -6, -12,  -1,   1,  -1,  -9,  -4,  -1,  20,   5,  -1,  -1,   4,   3, -29,  10,   4,   0,   0,   2,   0,  20,  -6,  -2,  -1,   0 },
      {   0,   0,  -1,   1,  -8,  -3,  11,   4,  -1,  -1,  -1,  -3,  -5,   6, -17, -20,   0,   0,   0,  -1,   3,  -4,  15, -12,   0,  -1,   3,  -1,  -4,   0, -40,  20,   0,   1,   2,   1,  -3,  15, -11,  -1,   0,  -1,  -1,  -2,  14,  33,  -2,   5,  -5,  -3,  -8,  17, -88,   1,   0,  -2,  -2,  -2,  -4, -50, -22,   4,   1,   2 },
      {   0,   0,  -1,   0,   1,   5,   0,  -7,   1,   1,   1,   0,  -1,   3,   5,   9,   0,   0,   2,   1,   1,  -4, -13,  -6,   1,   1,   2,   0,  -3,   1,   5,  33,   2,   2,   4,   7,   0,   4,  16,  15,   1,   3,   5,  -5, -11, 106,  10,   6,   2,   2,   6,  18,  29,  28,   2,  -5,   0,  -2,   1,  16,   8, -13,   0,  11 },
      {   0,   0,   0,   1,  -2,  -2,   1,  -4,   0,  -2,  -1,   0,   1,  11,   0,  -1,   0,   0,  -2,  -1,  -1,   0, -12,  -8,  -1,  -2,  -1,  -4,  -3,  -3,  30, -36,   0,  -3,  -1,  -4,  -5,   9, -96,  22,  -2,   2,  -1,   9,   1,  25, -49,   6,   1,  -1,  -7, -10,   5,   7,  -7, -10,   0,   4,  -1,  -1,  -9,  10,  13,  -7 },
      {   0,   0,   0,  -1,   1,   2,  -3,   4,   0,  -1,   0,   0,  -1,   1,  -9,  -4,   0,   1,   0,   0,  -2,   2,  23,  19,  -1,  -2,  -2,   0,  -1,  18,  11, -98,   0,   0,  -2,   2,  -4, -23,  45,   8,   0,   1,   1,  -9,  -4,  28,  -1,  15,   0,  -1,   2,   1, -22,  22, -12,  -2,  -1,   2,  -4,   2,  -7,   2,  -3, -15 },
      {   0,   1,   2,   0,   0,   0,   4,   1,  -1,   0,  -1,   1,  -1,   2, -13,   0,   0,   0,   1,  -1,   2,   4,   9,  18,   0,   0,   0,   0,   5,  -2,  12,  44,   0,  -1,   1,  -3,   2, -10,  40,  62,  -2,  -2,  -3,   2,  -8, -17, -63,  15,  -2,  -2,  -8, -13,   0,  21, -36, -10,   0,   2,   0,   0, -26,  27,   0,  -5 },
      {   0,   1,   0,   1,   0,  -2,   3,   2,   0,  -1,   1,   0,   0,  -1, -10,  -7,   1,   1,   2,   0,   1,  -1,  -1,  20,   0,  -1,   0,   2,   2,   5, -10,  -2,   1,   1,   0,  -1,   3,  -4,  -1,  48,  -1,   2,   0,  -1,   2,  14,   0,  24,  -1,  -4,  -3,   6, -13, -85,  -8,  -8,   4,   3,   0,  12,  69,  -2,  -2,  -3 },
      {   0,   1,   0,   1,   0,  -2,   5,   1,  -1,   0,  -1,   2,   0,   1, -11, -12,   0,   0,   2,   0,   2,   4,   2,   9,  -1,   0,  -3,   1,  -1,  -1,   6,   4,   0,  -2,  -1,   1,  -4,  -8,  20, -75,  -1,   0,  -2,   2, -15,  28, -67, -16,  -1,  -1,  -1,  -6,   6, -55, -11,   8,   1,   1,   3,   9, -22, -10,  -4,   3 },
      {   0,   0,   0,   1,   0,  -1,   0,   1,   1,   1,   1,   2,  -2,   1,  12,   9,   0,   1,   1,  -1,  -4,   2,  -3,   8,   0,   0,  -1,   0,  -2,   6,   0,  -7,   0,   1,  -1,  -1,  -1,  -8,   0,  44,   0,   0,  -1,  -7,  -2,   7,  30,  -9,   1,  -2,   1, -10,  12, -55,  27,   4,  -3,  -1,   2,   0, -90, -23,  14,   9 },
      {   0,   0,  -1,   1,  -1,  -3,   0,  -2,   0,   0,   1,   0,   0,   2,  -2,   7,   0,   0,  -1,  -2,   2,   2,   2, -12,   0,   0,   0,   1,   0,  -3,   3,   4,   0,   0,   0,   0,   1,   8,  -8, -31,   0,   0,   3,   1,  -3,   2,  33,  96,  -1,  -3,  -8,  -6,  16, -16, -40,   5,   1,   2,  -1,   0, -26,  42,  -4,   8 },
      {   0,   0,   0,   0,   1,   2,   1,  -1,   0,   0,  -1,  -1,   1,   0,   0,   1,   0,  -1,   2,   1,  -5,  -6,   0,   9,   1,   1,   0,   1,   3,   5,  -7,  -4,   0,   1,   2,   5,   1,   2,   9,   3,   0,   1,  -2,  -5,  -4,  -5, -37,  46,   2,   4,   8,   3,   2,   9,  86,  51,   0,  -2,  -2,  -6,  12,  17,  22,  35 },
      {   0,   0,  -1,  -1,  -1,  -3,   0,   2,   0,   0,   0,  -1,   2,   2,   5,  -8,   0,   0,  -1,   0,   2,   3,  -2,   1,   0,   0,   0,   2,   2,  -1,   2,  -3,   0,   0,   0,  -3,  -1,  -6,   7,  -3,  -1,  -1,   0,   5,  -5,  18,  18, -46,   1,  -1,  -3,  -2,   5, -15,  24, -20,   1,   4,   2,   0,   0, 109,  12,  -4 },
      {   0,   0,   0,   2,   0,  -2,   0,   1,   0,   0,  -1,   1,   1,   1,   0,   2,   0,   0,   0,  -1,   1,   2,   3,  -4,   0,   0,  -1,  -1,   0,   2,   4,  15,   0,   0,   2,   1,   0,   0,   8, -13,   1,   1,   0,   3,   4,   0,  -6,  32,   0,   0,  -2,  -1,   5,  -1,  49, -50,   0,  -3,  -1,  -3,  -3, -12,  21, -96 },
      {   0,  -1,   0,  -1,   1,  -1,  -5,  -5,   0,   0,  -1,   0,   1,   3,   8,   2,   0,  -1,  -2,  -2,   0,   0,   0,  -4,   0,   0,   0,  -2,   2,  -6,   2, -10,   0,   0,  -2,  -2,  -1,   5,  11, -10,  -1,   0,  -2,   0,  -2, -11,  -9,  14,   0,  -1,  -2,  -1,  -6,   2,  15, -99,   2,   1,   2,  -3,   3, -12,  17,  70 },
      {   0,   0,  -1,  -1,  -1,  -2,   2,   1,   0,   0,   0,   0,   2,   1,   4,   3,   0,   0,   0,   1,   0,  -2,   2,  -5,   0,   0,  -1,   0,   2,  -5,   2,   1,   0,   1,   1,   1,  -1,   2,   9,  -6,  -1,  -1,  -2,   1,  -5,  -2,   7,  -7,  -1,   1,   0,   3,  -7,   0, -34,  18,   3,   1,   3,   3,  10,  -8, 120,  -1 },
    },
    {
      { -68,  53, -13,   7,  -2,   2,  -1,   1,  71, -48,   8,  -4,   0,  -1,   0,   0, -30,  11,   6,  -2,   2,  -1,   1,   0,   8,   2,  -5,   2,  -1,   0,   0,   0,  -3,  -1,   2,   0,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0,  -2,  -1,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      {  67,  11, -39,   9,  -8,   2,  -3,   0, -18, -44,  40,  -8,   6,  -1,   2,   0, -46,  47, -12,  -2,   3,  -2,   1,  -1,  34, -16,  -7,   7,  -4,   2,  -1,   0, -10,  -3,   9,  -4,   1,  -1,   0,   0,   2,   1,  -4,   1,   0,   0,   0,   0,  -4,   1,   2,   0,   1,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0 },
      {  29, -53,  40, -15,   7,  -4,   2,  -1,  57,  13, -38,  10,  -7,   1,  -2,  -1, -73,   9,  13,   3,   2,   2,   0,   1,   8,   6,  -1,  -4,   1,  -1,   0,   0,   6,  -3,  -3,   3,  -2,   1,   0,   0,  -4,   3,   2,  -1,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
      {  59,  24, -32,   5,  -7,   0,  -3,   0,  29, -22,   1,   1,  -2,   1,  -1,   1,   5, -53,  36,  -6,   3,  -1,   2,  -1, -54,  50,  -7,  -3,   3,  -1,   1,  -1,  11,  -3, -12,   4,  -1,   1,  -1,   1,  -4,  -4,   7,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,  -1,   0,   0,   0,  -5,   0,   1,   0,   1,   0,   0,   0 },
      {  20, -31,  -2,   9,  -2,   4,  -2,   2,  57,  -8,  19, -19,   5,  -6,   0,  -2,  37,  18, -42,  15,  -9,   2,  -1,   0, -50, -38,  29,   0,   1,   4,   0,   1, -28,  30,   1, -11,   7,  -4,   2,  -1,  10,  -4,  -8,   6,  -4,   1,   0,   0,  -3,   0,   2,  -1,   0,   1,   0,   0,  -2,  -1,  -1,   0,   0,   0,   0,   0 },
      {  28,  36,  20, -33,   9,  -8,  -1,  -2,   6, -18, -39,  34,  -9,   4,   1,   2,  28,  -6,  25,  -2, -12,   5,  -3,   2,  28, -21,  -5, -13,  10,  -3,   2,   0, -73,  10,   8,   2,   5,  -2,   2,   0,  23,  10,  -5,  -1,  -4,   0,   0,   0,  -6, -11,   0,   3,  -1,   2,   0,   0,  -1,   2,   0,  -1,   0,   0,   0,   0 },
      {  -3, -12,  33, -40,  28, -17,   9,  -5, -15, -58,   7,  34, -17,  12,  -2,   3,  23,  48,  -9, -17,   5,  -4,  -1,  -1, -44,  20, -17,  10,   0,  -1,   1,   0,  28, -14,   8,   2,  -7,   4,  -2,   1,  -5,   5,  -3,  -3,   4,  -2,   0,   0,   4,   1,   0,   1,   0,   0,   0,   0,   0,   2,  -1,  -1,   1,  -1,   0,   0 },
      {  11, -22,  -4,   5,   4,   2,   1,  -1,  40,   2,  10, -14,   4,  -2,  -4,   0,  50,  -3,  -9,   4,  -6,  -3,  -1,  -2,  44,  15, -33,   8,  -4,   2,  -2,  -1,  -9, -63,  25,   0,   1,   2,  -1,   1, -48,  22,   9,  -7,   3,  -1,   3,   0,   6,   6,  -8,   3,   0,   0,   1,   0,  -8,   0,   2,   1,   0,   1,   0,   0 },
      { -17, -39, -13,  29, -15,  13,  -1,   5,  -1,   7,   6,  -2,  -1,  -1,   1,  -1,   0,   6,  37, -51,  24,  -8,   2,  -1, -21,  -6, -49,  34,  -6,   3,   0,   2, -33,  13,  31,   9, -11,   4,  -1,  -2,  33,  15, -12, -15,   5,  -3,  -1,  -1,  -1, -10,  -1,   6,   2,  -1,  -2,   0,   1,   4,   1,  -1,  -1,   0,   0,   0 },
      { -12, -16,  42,  12, -16,   8,  -8,   1, -24, -29,  13, -34,  32, -16,  10,  -1, -12,  -1,  -6,  38, -23,   8,  -1,  -1,  -9,  56,  -5, -15,   2,   1,  -3,   2, -40, -19,  -3,   2,   9,  -3,   3,   1,  36,  13,   2,   1,  -7,   2,  -1,   0,   5,  -9,  -7,   0,   0,   1,   1,   0,  -5,   0,   2,  -1,   0,   1,  -1,   0 },
      {  -9, -49,   4,   6,   6,  -1,   3,   0,   3, -46,  20,   1,  -3,   3,   1,   1,  24, -21,  22, -10,   7,   1,  -1,   0,  47,  10,   0,  -8,   1,  -4,   0,  -1,  -2,  41, -42,  10,  -5,   0,  -1,  -1,  -7, -42,  16,   4,  -2,   3,  -1,   1, -26,  22,  10,  -7,   3,  -2,   1,  -1,   9,   5,  -6,   1,  -1,   1,  -1,   0 },
      {  17,  17,  31,  22, -58,  24, -18,   9,   3, -21, -50, -10,  39, -15,  16,  -5,  27,  29,  13, -13,   1,  -1,  -3,  -1,   2, -21,   0,  -1,  -7,   4,  -3,   4,  41,   0, -11,  17,  -6,  -1,   3,  -4,  -8,  -2,   7, -10,   2,  -1,  -4,  -1,  -7,  -1,   0,   2,   5,  -1,  -1,   0,   7,   1,   2,   0,  -3,   0,  -1,  -1 },
      { -10, -14,  12,  10, -10,   4,  -3,   2, -24, -41,   1,   4,   1,   2,   4,   1, -22, -37,   9,  -7,  14,  -1,  -1,   1, -15, -30,  33,  -2,  -1,  -3,   1,   1, -24,  10,   7,   1,  -1,   1,   2,  -2, -64,  50,   3,  -1,   6,  -2,   0,  -1,  37,  10, -10,   0,  -2,  -1,  -2,   0,   3,  -3,  -1,   1,  -1,   0,   0,   0 },
      { -14, -18, -45,  10,   4,   8,  -1,   6,  -2,  14,  -8,  20,  -5,  -6,   0,  -3,   4,  55,  36,   3, -22,   6,  -5,   0, -17,   2,   3, -39,  23,  -1,   1,   1, -19, -15, -35,  22,   6,  -7,   6,   1, -10,  -2,  31,   0,  -9,   4,  -1,   1,  37,   8, -14,  -7,  -1,   1,  -1,   0, -10, -14,  -2,   2,   2,   1,   0,   0 },
      {   2,  13,   1,  33,   0, -16,  10,  -7,  -9,   4, -39, -15, -17,  30, -11,   7,  -4,  19, -26,  -6,  44, -21,   6,  -5, -14,  21,  23,   9, -23,   2,  -2,  -1, -42, -25, -13,   7,  -1,  11,  -2,   3,  -4, -13,  26,   2,   3,  -7,   0,   0, -28,  38,   0, -12,   5,  -2,   1,  -1,  13,  -1,  -6,   6,  -1,   0,  -1,   0 },
      {  11,   4,   9,  12,  -5, -10,   3,   0,  14,   3,   3,  -4, -20,  12,  -4,   1,  20,   8, -16, -17,  37, -14,   6,  -1,  34,  29,  22,   2, -18,   2,  -3,   0,  11,  22, -35,   2,  -4,   1,   2,   1,  26,  28,  -8,   4,   0,  -2,  -2,  -2,  60, -42, -11,  -1,  -3,   1,  -3,   1, -25,  -9,  10,   1,   2,   0,   1,  -1 },
      {  10,  36,  45, -28,  -6,   6,  -5,   1,   7,  35,  41, -13,  11, -16,   3,  -6,  -8,  13,  -1, -22,   1,  -1,   1,  -1,  -9,   7, -30,  24,  -6,   1,   0,   0, -28,  26, -22,   5,  -6,   1,   0,   1, -33, -27,  23,   0,   4,   3,   1,   0,  31,  23,  -6,  -3,  -2,  -3,  -1,   0,   9, -11,  -7,   1,  -1,   1,   1,   0 },
      {  -3,   0,   2,   9,   8, -13,   9,  -3,  -8,  -1, -19, -26,   4,  13,  -9,   4,   5,  22,  46,  25,  -3, -13,   8,  -9,  -6,   1, -18,  14, -27,  20,  -9,   5,  -3,  17,  -5, -51,  35, -11,   3,   2, -34, -10, -22,  38, -11,   1,   5,  -2,   4, -19,  24,  -6,  -8,   5,  -4,   1, -14,  13,  -4,  -6,   6,  -3,   2,  -1 },
      {  -6,  -7,  -4, -14, -25,  46, -22,  12,   2,  10,  13,  48,   2, -25,   3,  -2,  -2,   5, -12, -32,  12,  -9,   4,  -5,   2,  32,  35,  -7,  -5,   8,  -9,  -2, -21, -23,  -5, -14,   9,  -6,   2,   0, -17, -12, -32,  25,   3,   2,   4,   1, -14, -13,  34,   0,  -8,   5,  -1,   1,   4,  18, -10,  -9,   3,   0,   0,   0 },
      {   6,  16,  36,  25,   3, -19,   8,  -9,   3,  15,  41, -25, -19,  12, -10,   5,  -1,  10,  25, -16,   3,  -2,  -1,   1, -10, -18,   7, -37,  24, -13,  10,  -6,  -1, -20, -18,  33, -10,   4,  -3,   1, -16,  16, -17,   1,  -5,   2,  -1,   2, -37, -35,  20,  -1,   6,   0,   2,   0,  -8,  31,  -1,  -2,   1,  -1,   0,  -1 },
      {   2,   1,  28,  21,   0,  14, -13,  10,  -5, -16,  13,  12, -27,  -2,  -1,  -1,  -7, -26,   4,  -7,   1,  -1,   5,  -1,  -6, -46,  -1,  18,   0,  10,  -4,   3,   5, -51, -10, -18,  16,  -4,   3,   4,  27, -40,  31,  10, -10,   3,   3,  -2,  16,   0,  -5,  -8,  -2,   3,  -2,   1, -50,  -3,   2,   1,   6,  -1,   2,   0 },
      {   9,   6,  27,  27,  -2,   0,  -7,   4,  10,   8,  28,  12, -23,   5,  -5,   2,  11,   8,  35,  -4,  -6,   2,   3,  -3,   9,   2,  20, -10,   0,   2,   0,   2,  19,  -8,  12, -33,  23,  -3,   1,   3,  26,  24,  -8,   9,  -8,  -3,   4,  -3,   3,  36, -20,   2,  -9,   3,   0,  -1,  66, -31, -17,   1,   0,   1,  -1,   2 },
      {  -5,  -9, -14, -65,   1,   6,   2,   6,   3,   4,  20, -19,  19,  -4,   4,  -7,   2,  -6,  26,  11,  25, -28,  11,  -5,  -5, -26,  12,  -5, -44,  27, -13,   4,   4, -24, -28,  16,  13,   0,   1,   1,  23,  30,  20,  -6,   1,  -9,   0,  -4, -20,  -1,  -7,  -2,   1,   2,   1,   1,  12,   5,   6,   1,  -3,   0,  -2,   0 },
      {   0,   2,  11,   8, -30,  11,   2,   4,   0,   2,  13,  31, -15,  14, -20,   8,  -2,  -5, -14,  35,  19, -41,  24,  -9,  -2,   1, -32, -54,  -5,  34,  -4,   1,   4,  26,  36,  23,  14, -12,  -6,   4,  -9, -16,   3, -11,  -7,  -3,   4,  -2,   4,   3, -14,   3,   8,   0,  -2,   0,  -5,   4,   3,   2,  -3,   0,   1,   0 },
      {  -6,   0,   1,   4, -30, -12,  -1,   3,  -2,   9,  11,  29, -13,   5,   4,   1,  -4,   8,  -9,  -7, -13,  19,  -1,   2,  -2,  13,  -4,  17,  -6,  -3,  -3,   4,  -3,  20, -22, -31,  22,  -9,   4,  -1, -20,  29,  30, -14,  -9,   0,   1,  -3, -60, -20, -43,  17,   7,   4,  -3,   2, -26, -14,  25,   5,  -4,   1,   1,   2 },
      {   0,   3,   8, -10,  10,  13,  -1,  -9,   1,   4,   9, -12,  26,   1, -18,   9,   5,  16,  35,  -7,  -1, -24,   9, -10,   7,  11,  45,  -5,   7,  -9,   6,  -6,   1,   8,  28, -11, -17,  18, -17,   5,   1, -12, -23, -13,  11,  -3,   4,   2, -13,  35, -24,  12,  -6,  -6,   8,   0, -61, -23,  23,   1,  -1,   3,   0,   0 },
      {  -4,  -8,  -4,  -2, -37, -36,  33, -30,   5,   1,   6,  28,  58,  25, -38,  23,  -5, -13, -12, -27, -31, -13,  15,  -1,   0,  -9,  -1,   8,   4,   0,   8,  -9,  -1, -15, -16,   1,  10,   1,  -9,   5,  12,   3,  -4,  13, -17,   7,   3,   1,  14,   6,  13,  -8,   7,  -2,  -1,   1,   3,  -7,  -9,   2,  -2,  -2,   3,   0 },
      {  -3,  -2,  -3, -23, -46, -31,   4,   5,   5,   9,  11,  -5, -11,  13,  29, -14,   4,   4,   6,  -5,  18,  30, -25,  16,   3,   8,   0, -21,  -8, -14,   7,   5,   7,   6,  30,  -4,  10,  -1,   8,  -7,  12,  13,   7,  23,  -4,  -6, -13,  -1,   4,  42,  18, -33,   1,   0,  -4,  -3, -39,  20, -19,   9,   6,  -6,   0,  -1 },
      {   4,   7,   3,   3,  41,  24,  -8,  -4,  -4,  -4, -17, -19,   5, -21,  -1,   2,  -3,  -7, -26, -42,  -2,  -7,  18,  -7,   1,   1, -26, -18,  18,  14, -10,   4,  10,  15,  -8,  14,  26, -16,   4,  -2,   3,  39,  -8,  27, -33,   6,   1,   0, -20,  27,   5, -26,   3,  -2,  -2,   2, -22, -24,  -6,  14,   6,  -1,   2,   0 },
      {   7,  11,  11,  45,  43,  -9,   2,  -5,   6,  12,  15,  53,  32, -24,   7,  -8,   2,   4,   9,  23,  -7,   1,  -9,  -1,   2,   4,   9,  17, -35,  19,  -7,   5,   4,  13,   0,  11,  -9,  -1,   2,  -2,   6,  25,  16, -19,   7,   0,  -7,   2,  -2,  16,  18, -16,   5,  -7,   3,  -1, -19,  35, -16,   4,  -2,   0,   1,   1 },
      {   0,  -1,   2,  16,   6, -41,  20,  -6,  -1,  -2,   3,  12,  15, -27,  21, -16,  -2,  -7,  -3,  -7,  21,  24, -16,   8,  -5,  -9, -10, -26, -29,  19,  -7,   8,  -7, -15,  -6,  19,   7, -10,  10,  -3,  -8, -26, -42,   3,  17, -10,   4,  -1,  -7,  -3,  -2,  19, -19,   9,  -8,   1,  -5, -59,  14,  -1,   3,   6,  -4,   2 },
      {  -1,   0,  -1,   3, -16,  18,  -8,   1,   1,   2,   2,  23,  -6,  -6, -10,   3,   2,   0,  -3,  50,  23,  -4,   7,  -1,  -6, -15, -40,   7,   7, -31,  16,  -7,  -1,  -3, -41, -10, -30,  37, -13,   6,   4,  28, -22,  28,  16,  -6,   1,   4, -12,  16,  13, -11, -15,  -9,   5,  -3, -11, -27,  -4,  -1,   3,   5,  -2,   1 },
      {  -1,  -1,   2,  -3,  -1,  -3, -34,  33,   1,   4,   9,   8,   2,  57,  10, -17,  -2,  -2,   7,  18, -16, -15, -15,  -7,   1,   5,  10,  31,   2,   2, -19,  13,  -3,  -3,   1,  29, -11, -10,  13, -11,  -8,  -2, -25, -12, -25,  13,   1,  -3,  -5,  -6,  23, -23,  30,   0,  -3,   5,  -4, -46,  -2,  13,  -8,  -1,   2,   2 },
      {   1,   1,   1, -10, -18,  21,  45, -31,  -2,  -3,   1,  -3, -24, -44, -28,  26,   5,   5,  16,  24,  23,  28,  -4,  -8,   2,   3,  18,  28,   2,   2,  -8,   0,   1,  -2,  16,  16, -11, -23,  11,   0,  -4,  -3,   0,  -6, -20,  19,  -4,  -1,  -6, -15,   2, -27,  24,  -6,  -3,   4,   0, -23, -16,  22,  -7,   2,   4,   1 },
      {   1,   5,   1,  -3, -17,  13,  27,   9,  -2,  -1,  -4, -11, -30,  14, -21,   0,  -1,  -1,  -5,   2, -25,  17,   2, -10,   1,   0,  -5,  10,   5,  25, -29,  16,   0,  -2, -30,   7, -11, -29,  19,  -6,   9,  17, -30, -25,  22,   6,   8,  -7,   6,  49,  17,  34, -14,  -8, -11,   2, -14,  23,   6, -23,  -1,   5,   1,  -1 },
      {  -2,  -3,  -3,  -2,  -7, -51,  -8,  13,   1,   1,  -2,  -1,   7, -28,  26,  -7,   1,   0,   2,   1,  15, -22,  49, -21,   1,  -1,   3,  -4,  45,  12, -33,  17,  -3,  -5,   0, -33, -45, -10,  14,   1,   1,   1,   7,  -1,   8,  17,  -1,  -1,  -5,  -3,  -4,  -1,   7,  -3,  -8,   3,  -1,   1, -26,   4,  -3,   0,   1,  -4 },
      {   0,   0,  -3,   0,   7, -10, -21, -17,   3,   4,   2,  12,  15,  -4,  15,  15,   1,   0,   1,  11,  36,   7, -14, -16,  -2,  -6,  -8,   8,   8, -19,  -5,   6,  -1, -10, -13,   5,   7,  -7,  11,  -4,   1,  -9, -35,  16, -46,  23,  -3,   0,  12,  27, -39,  15,  29,  -3,  -1,   7,  10,  41,  42, -12, -13,  -5,   0,   0 },
      {  -4,  -6,  -8,  -4, -16, -27, -49, -30,  -2,  -2,  -7,  -8, -43, -27,  20,  24,  -1,   1,   0,  -4, -40,  -1,  31,  -1,  -1,   0,   1,   6, -37,   6,  11, -10,  -2,  -1, -11,  17,  10,  15, -26,  14,  -4,  -3, -30, -22,   2,   3,   6,   0,   6,  10,  -3,  -8,  15,  -4,   7,  -2,  -5,   5, -14,  16,  -5,  -2,  -1,  -4 },
      {   0,   0,   1,   5, -13, -20,  20,  13,  -1,  -1,   0,   5, -21, -29,   7, -26,   2,   3,   5,   9, -21, -18,  26,   1,   5,   6,  16,  33,   8,  -1,  -1,   2,   2,   6,  18,  41,  -1,   8, -22,  -1,  -1,   5,  16,  38, -22,  -9,   5,  -1,  -2,   7,  13,  -8, -18,  11,  -8,   2,   9,  -1,  52, -38,  15,  -4,  -2,   1 },
      {   0,   1,  -1,   0,  -5,  -2,  30,  27,  -3,  -3,  -7,  -2,  -6, -15,  -1, -43,  -2,  -3,  -2,   1,  -8, -36, -30,  33,   2,   2,   0,  16,  18,  29,  28, -24,  -2,  -6, -18,  -4,   0,  -9, -17,  10,  -2,  -3, -36,   2,   0,   2,   7,   3,  -5,   4, -48, -18,  15,  -1,  -2,  -5,  -1,  19, -24,  17,   0,   2,   2,   3 },
      {  -2,  -2,  -2,  -5,  -1, -20, -31,  -1,  -1,   0,  -2,  -2,   9, -10, -10,   3,   2,   4,   2,  17,  36,  -7,  -6,  10,   3,   3,   6,  25,  30,  -8,  28, -20,   3,   3,  -1,  18,  38, -32, -11,   9,   1,   6,   8, -30, -16,  10,  12,  -4,  -3,   6,  20,  36, -32,  15,  -5,   0, -12,  -8, -38, -24,  13,  -4,   3,   2 },
      {  -1,  -1,   0,   1,  -6,  10, -14, -74,   2,   3,   4,   1, -10,  11,  37,  16,  -1,  -1,   2,   4,  -8, -39, -40,  24,   1,   0,  -1,   6,  20,  38,  -2, -12,   0,   1,  -1,  -8, -26, -19,  17,  -2,   3,   8,  14,  15,  11,  -1,  -7,   4,  -4,   2,   9,   0, -10,   9,   1,   0,   1, -11,  16, -11,   2,  -1,  -4,   1 },
      {   0,   0,   1,   1,   1,  -3,  14,  11,  -2,  -2,  -2,  -3, -12, -21,   2, -16,   0,   0,  -1,  -6,  -2, -33,   3,  32,  -2,   0,  -4, -19, -23, -24,  22, -15,  -2,  -4,   0, -28,   3,  -7,   9, -16,   3,   6,  11, -17, -12,  49, -21,  11,   4,   8,  42,  25,  29, -37,   1,  -1,   1, -17,  23,  -7, -20,   9,   3,   1 },
      {  -1,  -1,  -1,  -3,   3, -20, -39,  34,  -1,  -1,  -2,  -6,   5,  -4, -39,  22,   1,   2,   1,  -5,   8,  36, -13,   0,   0,   1,   0,  -8,  -6,  46,  21, -34,  -1,   0,   5,  -8, -36, -29, -18,  20,   3,   8,  13,  24,  14,  11,   5,  -2,  -1,   2,   7, -13,  16,  -7,   1,  -4,   5,  -4,  30, -12,  -5,   7,  -1,   4 },
      {  -1,  -2,  -1,  -1,  -3, -15, -29,  -3,  -1,  -2,  -1,   0,  -3,  -6, -41,  -7,   0,  -1,   1,   0, -13, -17,  -2,   4,   0,   0,   2,  11, -12, -20,  -5,  11,   3,   2,  10,  30,  10, -16,  41, -17,   1,  -4,   0,  41,  38,  23, -41,   6,  -9,  -7, -34,   2, -28, -26,  18,   4,  -1,   6, -14,   4,   3,  15,   6,   1 },
      {  -1,  -1,  -1,  -3,  -8, -25, -15,   9,  -2,  -2,  -2,  -4,  -7, -37, -38,   7,  -3,  -3,  -5,  -4,  -1, -28, -47, -27,   0,   0,  -5,  -4,  21,   1,   3,  25,   1,   1,  -8, -19,  19,  21,  18, -19,   1,   0,  -2, -38,   3, -28,  -3,   5,   5,   4,  21, -33,  -6,  18,  11,   4,   7,   3,  31,  11,  -1,   1,  -8,  -2 },
      {   0,   0,   2,   1,   2,   1,  -2, -10,   1,   2,   4,   5,   8,  10,  10, -13,   3,   4,   6,   9,  18,  17,  34,  22,   1,   3,   4,  18,  43,  12,  -5,  -6,   0,  -2,  -3,   6,  47, -10, -11,   3,   1,  -1,  -3,  -1,  44, -15,  -9,   6,   7,   3,  20,  -6,   7, -20,  14,  -4,   6,   7,  38,  45, -35,  15,  -3,  -3 },
      {   0,   0,  -2,  -2,  -2,  -9,  -2,   5,  -1,   0,  -1,  -3,  -7, -11,  -6,   8,   0,   0,   0,  -1,   0,   4, -30,   2,   2,   3,   5,   6,  14,  35, -20,  15,   2,   4,   5,  11,   8,  53, -14, -12,   0,   4,  13,  31, -20, -20,   5,   3,   1,  -4,   6,  51,  21, -30,  13,   1, -10,  -6, -32,  -8, -39,  24,   0,  -2 },
      {   1,   1,   1,   1,   0,   0,  31,  37,   2,   3,   2,   2,  -2,   3,  51,  52,  -2,  -3,  -2,  -1,  -8, -17, -18, -57,   1,   0,   1,   8,   3,   2,  35,   6,   0,  -1,  -1,  10,  12, -16, -10,  21,   0,   2,   0,  10,  13,  11, -29,   3,  -1,  -1,   4,   7, -15, -13,  21,  -3,  -1,  -3,   0,  11,  -3,  10,  -2,  -3 },
      {  -1,  -1,   0,  -3,  -5,   5,   0, -14,   0,   0,   0,   0,   0,  16,  -1, -25,   0,   0,   0,  -1,  11,  27,  13, -12,   0,   1,   1,   0,   9,  44,  12,  13,  -2,  -3,  -4,  -8,   4,  25, -19, -19,  -4,  -2,  -7, -23, -17,  32, -43,  49,  -1,  -4,  -7, -16, -15, -13,  37, -13,   4,  -5,  -1, -28,  29,  -7,  -4, -10 },
      {   1,   2,   1,   0,   5,  -1,   6,  37,   1,   2,   3,   0,   4,  -3,   6,  62,   1,   3,   3,   6,  -1,   2,  29,  58,   1,   2,   2,   9,   3,  17,   8, -16,  -1,  -2,  -3,   4,  -6,  18,  40, -34,  -4,  -4, -11,  -5, -17, -24, -27,  14,   0,   6,  -5,   2,  -7,  25,  -2,   1,  -1,   2,   1,   3,   7,  -9,   2,  -7 },
      {  -1,  -1,  -1,  -1,  -1,   3,  -7,  -9,  -1,  -1,  -1,  -1,  -4,   1, -13, -27,   0,   0,   1,   2,  -5,   4,  10, -20,   1,   1,   2,   2, -11,  10,  47,  21,  -1,  -3,  -1,   0, -31, -16,  -1, -23,   1,   2,   6,  19, -24,  -8,  -6,  24,   2,   6,  11,  44, -14,  37,  -8,  -3,   0,   8,   3,  65, -15, -13,  -5,  -2 },
      {   0,   0,   0,   0,  -5,  -6,   0,  13,   0,   0,   0,   0,  -1,  -1,  -2,  13,  -1,   1,   0,   0,   1,  -2, -24,  17,  -2,  -1,  -2,  -3,   0,   5, -54, -18,   0,   3,   3,   3,  -3,  27, -24,  17,   0,   1,   4,  -3, -14,  52,   6,   0,   0,   1,   6,  11, -48,  -3,  -2,   7,   7,   9,  16,  58,  22, -17,  10,   6 },
      {   0,   0,   0,   0,  -2,  -3,   0,   4,   0,   1,  -1,   0,   1,   3,   3,  16,   0,   0,  -1,  -3,   2,  -5,   1,  28,  -2,  -2,  -3,  -3, -11, -20, -16,  26,  -2,  -3,  -2,  -7, -20, -42, -36, -22,  -1,  -2,  -2,  -7, -29, -13,  25,  18,   1,  -4,  -7, -19, -35, -10,  37, -14,   0,   7,  -2,  -1, -19,  53, -30,  17 },
      {  -1,  -1,   0,  -1,  -5,  -3,   0,  -3,  -1,  -1,  -1,   0,  -2,   1,  -1, -25,   0,  -1,   0,  -1,  -3,   4,   9, -15,  -1,  -1,   1,   1,  -3,   1,  -6, -45,   1,   2,   2,  -2, -11,   6,  55,  56,  -4,  -3,  -3, -11, -42, -35, -15,  -5,   2,   1,   7,   8, -15, -20,  27, -25,   1,   4,   5,  11,   9,  30, -14,   8 },
      {   0,   0,  -1,  -2,   2,   3,  -3,  -6,   0,   1,   0,   0,   3,   8,   6,  -1,   0,   1,  -1,  -2,   2,   8,  11,   8,   2,   3,   4,  -1,   3,  25,  48,  30,   0,  -1,   3,  -1,  -4,  28,  24,   4,  -3,  -3,  -2, -11, -16,  25,  19, -49,  -4,  -6,  -4, -26, -46, -27, -32,  21,   3,   6,   7,  -2, -29,  25,  14,   5 },
      {   0,  -1,  -1,  -2,  -1,  -3,  -3,   0,   1,   1,   2,   2,   4,  -1,   4,  10,   1,   1,   2,   3,   3,  -1,   2,   4,   1,   2,   4,   5,   6,   8,  22,  21,   0,   0,   1,   3,   3,   0,  13, -23,   1,   0,   3,   7,   6,  -7,  45,  -9,   1,  -2,   3,  15,  24, -53,   1, -16,   4,   2,   7,  23,  82,  -2, -27,  12 },
      {   0,   0,   0,   0,   1,   2,  -3,   0,   0,   0,   0,   1,   1,  -4, -12,   0,   0,   0,  -1,   0,   2,  -2,  -2,  37,   0,   1,   0,  -1,   1,  -1,   2,  55,  -3,  -3,  -4,  -7, -13, -17, -24,  44,  -3,  -1,  -3,  -1, -13, -20, -45, -49,   1,   2,   7,  13,  11,  -6,  31,  20,   3,  -3,   3,   4,  19, -21,   9, -32 },
      {   0,  -1,   0,   0,   0,  -2,  -3,  -5,   0,   0,   1,   1,  -1,   2,   5,  -7,  -1,  -1,   0,   1,  -1,   2,  -3, -23,  -1,  -1,  -1,  -2,  -2,  -2,  -7, -44,  -2,  -4,  -3,  -7,  -7, -19, -25, -46,   0,  -4,  -2,  -7, -15, -39, -18,  -7,   0,  -2,  -2,  -2, -14, -41, -20,  57,   4,   1,   0,   8,  16,  14,  28, -39 },
      {   0,   1,   0,   2,   0,   1,   8,  -1,   0,   0,  -1,   1,  -1,   2,   2, -11,   0,   1,  -1,   1,   0,   5,   7, -10,   0,   0,  -2,   0,   1,   7,   1, -31,   0,   2,   0,  -1,   2,  11,   2, -34,   0,   2,   1,   4,   7,  29,   0, -62,   2,   3,   1,   5,   8,  38,  52,  35,  -4,   0,  -5,   2,   3,  13, -58,  -3 },
      {   0,   0,   0,   1,  -2,  -3,   5,   5,   0,   0,   0,   2,  -2,  -3,   3,  -2,   0,   0,   0,  -1,  -1,   1,   0,  -1,   0,   0,   0,   0,  -4,   0,   4, -10,   0,   0,   1,   2,  -5, -10,  21,  -7,   0,   0,   1,   3,  -3,  -7,  47,   4,  -2,  -4,  -3,  -7, -23, -33,  50,  -2,  -1,  -4,   0,  -6, -38, -81,   2, -31 },
      {   0,   0,   0,   0,   1,   0,  -3,  -1,   0,   0,   0,  -1,  -1,  -3,  -4,   0,   0,   0,   0,   0,  -1,  -1,   0,  -8,   0,   0,   0,  -1,   4,   2,  -6,  -9,  -2,  -3,  -4,  -4,  -1,  -6, -22, -20,  -2,  -5,  -4,  -5,  -8, -19, -46, -32,   0,  -4,  -3,  -2, -12, -29, -27, -22,  -1,  -2,  -5,  -3, -18, -50, -35,  72 },
      {   0,   0,   0,  -1,   1,   2,  -1,   3,   0,   0,   0,  -1,   1,   2,  -4,   3,   0,   0,   0,   0,  -1,  -1,  -3,   8,   0,   0,   1,   1,  -1,  -1,   0,   8,   1,   2,   2,   2,   4,   2,   6,  37,   0,   0,   1,  -1,   1,   0,  -3,  44,  -1,  -1,  -2,  -2,  -3, -15, -37,  57,  -2,   0,  -4,  -3,  -8, -10, -85, -21 },
      {   0,   0,   0,   0,  -3,  -1,   3,   0,   0,   0,  -1,  -1,  -1,   0,   5,  -3,   0,   0,   0,  -1,   1,   3,   7,   4,   0,   0,  -1,  -2,   0,   3,   8,   6,   1,   1,  -1,  -1,  -2,  -1,   9,  13,   1,   1,   0,   0,   2, -10,  12,  27,   2,   2,   5,   3,   9,   2,  37,  76,   0,   1,  -1,  -2,   8,  -6,  34,  79 },
    },
    {
      { 105, -31,  -9,  -3,  -3,  -1,  -2,  -1,  51, -18,  -6,  -2,  -2,  -1,   0,   0,  -3, -11,   1,   0,   1,   0,   1,   0, -30,   7,   2,   1,   1,   1,   1,   0, -11,   4,   1,   1,   1,   1,   0,   0,  -7,   2,   1,   0,   1,   0,   0,   0,  -3,   2,  -1,   1,   0,   0,   0,   0,  -4,   0,   0,   0,   0,   0,   0,   0 },
      {  23, -58,  40, -10,   7,   0,   1,   0, -63, -19,  39,  -7,   6,   0,   2,   0, -47,  21,  14,  -6,   2,   0,   1,  -1,  19,  25, -13,  -3,   1,  -1,  -1,   0,  17,   7, -13,   1,  -1,   0,  -1,   0,   9,  -2,  -7,   4,  -2,  -1,  -1,   0,   1,   0,  -1,   1,  -1,   0,   0,   0,   2,  -1,   0,   0,  -1,   0,   0,   0 },
      {  46,  38, -39,   8,  -7,  -1,  -1,  -2, -56,  63, -34,  13,  -3,  -3,   2,  -1, -35,  18, -14,   9,   2,   0,   2,  -1,  -7,   0,  -2,   2,   0,   0,   0,  -1,  21, -15,   2,   0,  -1,   1,   0,   0,   3, -11,   5,  -2,   2,   2,   0,   1,  -2,  -2,   2,   0,   1,   0,   0,   0,   1,  -3,   2,   0,   0,   0,  -1,   1 },
      { -17,  40, -15,  -2,  -2,  -1,  -3,   4,  19, -31,  12,  -3,   0,   6,  -2,   2, -89, -14,  27,  -2,   3,   5,  -2,   1, -34, -11,  24,  -2,   4,   0,  -1,   1,  -5,  26,   6,  -5,   0,  -3,  -2,   0,  27,  12, -10,   2,  -4,  -2,  -2,  -2,   9,  -4,   0,   0,   0,   1,  -1,  -1,   6,   2,  -3,  -2,   1,   0,   1,   0 },
      {  21,  -7,  11, -10,  -2,  -2,  -1,   2, -51,   8,   3,  -1,   0,  10,  -1,   3,  29, -23,   9,  -1,  -5,   7,  -3,   1,  -1, -62,  32,  -2,   6,   5,  -4,   3, -59,  21,  27,  -9,   2,   1,  -2,  -1,  24,  28, -14,   0,  -4,  -4,  -5,   0,   7,  -1,  -5,   2,   1,  -1,  -1,   0,  -3,  -3,   2,   1,   0,  -1,   0,   1 },
      {  24,  30, -38,  10,  -7,   8,  -1,   4,   4, -41,   7,   3,   2,  10,  -3,   3,  -8,   7,  12,  -8,   5,   2,  -3,   3,  99, -16,  -7,  -8,   1,  -2,  -4,   2, -13,  -3,  -3,  -7,  -4,  -1,   3,  -1, -17,  -3,  -2,   6,  -2,  -1,   0,   0,  -8,  -1,   9,  -1,   1,  -1,  -1,   1,  -5,   3,   0,  -2,   0,   1,   1,   0 },
      { -12, -31, -37,  47, -19,  18,  -3,   2, -36, -43, -23,  47, -17,  13,  -3,   3,  11, -37,  13,  19,  -9,  -1,  -2,   0, -18,  25,  16,  -7,  -1,  -7,   8,  -3,  16,   8,   5,  -9,   6,  -1,  -1,   0, -17,  10,   8, -13,   4,  -2,   2,   1,  -2,  11,   0,  -3,   0,  -3,   0,  -1,  -1,   1,  -1,   0,   1,  -1,   0,   0 },
      {  18,  38,   2, -22,   9,  -7,   2,  -3, -24, -10,  34, -24,  16,  -9,   2,  -2,  25, -77,  12,  -8,  17,  -3,   4,   2,  -9,  -9,   0,   4,   0,  -1,   1,  -1,  55,  -5, -19,   9,  -5,  -2,   2,   0, -12,  -6,   0,   3,   4,   4,   4,   0,  -2,   9,  -2,   0,  -1,   0,   0,  -1,  -1,   4,  -4,   1,  -1,   0,   0,   0 },
      {   9,  26, -13,  12, -24,  13,  -5,   3,  -7,   9,  17,  -9, -12,   8,  -6,   3,  39,  22,  14, -32,  12,  -4,  -4,   2, -14,  43,  -4, -26,  16,  -8,   0,   0,  -9,  62, -32,  -7,   4,  -8,   6,  -3,  26, -18, -23,  11,   0,  -3,   1,   0,   2, -20,   1,   2,   3,   0,   0,   1,  -3,  -3,   1,   1,  -1,  -1,   1,   0 },
      {   0,  24, -18,  -6,   5,  11,  -3,   9, -27, -20,  33, -13,   5,  15,  -8,   0,   5,  23,  10, -16,   0, -10,  -4,   1, -39,  14, -27,  13,  -5,  -9,  12,   0, -55, -32,   5,  28,  12,   3,  -5,   3, -36,   2,  33,   3,   0,   1,   0,   0,  10,  20,  14,  -7,  -7,  -3,  -1,  -3,   7,  14, -12,  -4,   2,  -1,   1,  -2 },
      {  -3,  -4,  28,   2, -34,  20,  -7,  -2,  15,  50,   4,  -8, -27,   9,  -6,   4, -21, -18,  41, -27,   4,   0,   2,   3,  12,   7,  20, -30,  16,   1,   2,   0,  -6, -16,   1, -11,  14,  -1,   1,  -2, -57,  17,  -7,  -1,   8, -10,   4,   2,  -5,  25, -14,   5,   1,  -5,   3,   1,   7,  -8,   2,   7,  -4,   0,   1,   0 },
      { -13, -44, -57, -13,  32,  11,  -6,   8,   4,  15, -21, -34,  26,   3,  -8,   5,   9,  11,  53, -28,  -2, -10,  -8,   0,   2,  -7,  34,   0,  -3,  -8,   2,  -1,  16, -18,  -3,  10,  -3,  -3,  -2,   2,  17, -13,  -3,  10, -10,  -2,   3,   1,  15,   2,   3,   3,  -7,  -1,   4,  -3,  -7,   5,  -2,   1,   0,   1,   2,  -1 },
      {  -5,   4, -16, -26,  38, -12,  17, -13,  -9,   2, -29, -10,  29, -24,  12,  -6,  -6,  -8, -10,  16,  14, -10,   8,   1,  10,  45,   2,  -3,  -2,  -1,  -1,  -1, -29,  35, -10, -14,  -9,  -3,  14,  -1, -37,  41, -14,  -3,   4,   3,   6,  -1,  28,  13, -23,   3,  -1,   1,  -1,  -1,   7, -11,  -5,   6,  -2,   3,  -2,   1 },
      {  -2,   7,  12, -55,  -3,  32, -14,   6,  -6,   1, -30, -21, -11,  37, -12,   5,  -6, -26,   1,  29, -35,   1,  -6,   0,  11,  33,   3,  27, -14,  -6,   7,  -1, -13,  16,   3,   5,   5,   3, -12,   3,  -5, -20,  -1,  -1,  -6,  -9,  -3,   5, -29, -22,  21,   9,  -2,  -4,   5,   0, -19,   2,  17,   3,   2,  -2,   4,  -2 },
      {  15,  24,   9,  -9,   6, -15,   6,  -2,  11,  11,  27,  22, -27,   9,  -9,  -5,  20,  18,  42,  23, -41,  -7,  -1,  -2,  17,  25,  13,  16, -29,  -7,   4,  -2,  22,  -9,  10,   3,   3,   2,  -4,  -2,  24,  39,  10, -13,   0,   1,   1,  -1,  46,  -2, -15, -11,  -2,   1,  -1,   1,  -7,  -6,  -8,  -6,   6,  -3,   1,   0 },
      {  -6, -20, -16,  18, -11,  32, -16,   4,  10,  22,   2, -12,  -9,  13,  -7,   2, -17, -43, -27,  -6,   1,   3,   6,   2,  16,  -4, -53,  16,   9,  -3,   8,  -2, -12,  -7, -25,  25,   8,  -2,   1,   3,  34,  28,  -3,   8,   2,   0,   0,   1,  49,   2, -10,   2,  -7,  -5,   0,   0, -13, -11,   0,   2,   0,   1,   0,   0 },
      {  -1,   4,  19,  15,  18, -24,  12,  -8,   2,  14, -14,  43,   7, -22,  11,  -5,  -6, -33,  57,   4,  10, -16,   5,  -9,   2,  13, -25,  -5,   4,   2,  -2,   1, -42, -23, -19,   6,  -8,  10,   5,  -1,  13, -28,   1,  13,  -9,   9,  -4,   1,  -7, -27,  12,   0,  -2,   4,  -4,   2, -25,   3,  16,  -2,   2,   2,  -2,   0 },
      {  -5, -19, -19,  -3,  21, -17,   8,   5,  10,  33,  19,  21, -16,   7,  -9,  -5,  -5,   1,  15,  22, -14,  -7,  -7,  -1,   1, -36, -24,  19,  -2, -14,   2,   1,   6,  55, -33,  14,   6, -10,   5,  -6, -22,  11,   3,  13,  -5,  -1,   1,  -4, -32,  21,  18,  -9,   2,   3,  -3,  -1,  18,  31, -12, -13,   5,  -2,  -1,  -2 },
      {  -1,  10,  -9,  14,  22,  29, -13,  -4,   2,   1,  24,   5,  23,   1,   1,  -2,   4,  23,   9,  10,  -1,  -9,   0,  -5, -17,  -9, -16,   5,   6,   2,   7,  -2,   9, -11,  -4,  -4,  -7,   8,  -1,   1,  -2,  44, -36, -22,   5,  -4,  -4,   6, -52,   3, -16,  10,   3,  -9,   3,   3, -59, -18,  19,  16,  -1,   2,   0,   3 },
      {   8,   7,  23,  52,  44,   8, -38,  13,   8,  23,  23,  13,  47,  14, -37,  10,  -5, -14, -16,  -2,  -9,   4,  -6,   7,   7,   9,  18,   3, -19,  -9,   5,  -1,  -5,  15,  17,  -9,  -4, -10,  -7,   6, -18, -19,  -1,  -4,  -2,   1,   3,   3,  14, -17,  10,   5, -10,   2,   5,  -1,  15,  -2,  -6,   4,  -4,   4,   1,   0 },
      {   3,   7,  14,  10,  13, -27,  -2,   4,  -1,   2, -32,  -3,  10,   4,  -3,   4,  -1, -13,  -4, -59,  -3,   3,   7,   3,  14,  26, -11, -12,  -6,   0,   8,  -2,   8,  16,  29,  23,  15,   7, -15,  -3,  30,  44,  30,  -6,   0,  -7,  -9,  -2, -40,  19,  12, -13,  -2,  -3,  -1,  -1,  -5,  18,  -1,  -8,   0,  -5,   0,   0 },
      {  -9, -19, -24,  19, -27, -11,  29, -13,   1,  20,  32, -26,  -7, -12,  12, -10, -12, -12, -10, -21,   1,  -1,  10,   3,   2,   3,  12,  24, -40,  11,   0,   3,  -7,  26,  25,  -3, -20,   7,   5,   0, -24,  -5,  25,  -9,   8,  10,   2,   0,   0, -38,   1,  -9,   9,   4,  -4,   2, -47,   3,   4,  -4,   5,   2,  -2,  -1 },
      {  -1,  -3, -18,  20, -20,   8, -15,  -3,   4,  13,  24, -23,  10, -31,  21,  -4,   4, -11,  18,  25,  -8,  -3,  18, -11,  13,  25,   0,  29,  -9,  18,  -6,  -4, -21,   0,  14,   3,  -6,  13, -12,   4,  33, -13,  -4, -12,   5,  -2,  -5,   5, -36,  30, -14,   8,   5,  -3,   3,   3,  56,  -7,  -6,  12,  -6,   1,  -3,   3 },
      {   4,   7,  28,  24, -16,   8,  -6,   3,   0,   0, -10, -19,   4, -19,  -2,  -1,  -4,   7, -12,  30,  27, -26,  -7,   2,  12,   6,  52,  16,  10, -26,   7,  -9, -12,  -3, -29,  25,   2,  -8,   6,  -1,  14,  -1,   6,  -8,   3,   5,   6,  -1,   2,  30,   2, -20,   4,  -6,   3,  -6, -38,  44, -12,  -7,   7,  -2,   5,  -1 },
      {  -1,  -6,   3, -30, -11,  34,   8,   5,  10,  13,   6,  27,  19,  21,  -9,  -1,   3,   1,  13,  11,  62, -31,   1,  -2,   5,   0, -20,   6,  20, -14,   2,   8,  12,  19,  49, -14, -21,   9,   5,  -2,  12,  -7,  24, -23, -12,  10,  -5,   3,   7,   9,   3, -18,   0,  -3,  -4,   2,   3,  -4, -11,  -2,   0,   1,  -2,  -1 },
      {   5,   9,  26,  22, -17,   4,  -7,   4,  -6, -18, -39,  -6,  -1,  -4,   4,   1,   4,  18,  25, -18,  14, -23,  10,  -5,  -3, -22,  -3,  57,  -7, -11,   6,   0,  15,  19,  -2,  44,  -7,   0,   0,  -4, -26,  13, -12,   7, -11,   6,  -3,  -1,   3, -31, -13,   8,  -6,   7,   2,   3,  25, -32,   1,   6,  -3,   3,   0,   1 },
      {  -4, -10, -21, -10,   0, -30, -27,  17,  -1,   2,  22,  -9, -20, -32,  -9,   8,   0,  -7, -11,  17,  10, -28,   2,  -4,   6,  11,   4,  -2,  46, -30,   7,  -6,   1, -13,  28,  -3,  40, -18,  -2,  -1,  -5,  22,   3,   3,  -5,   2,   1,  -4, -14, -55,  -1,  10, -10,  10,   3,  -1,   7,   6,   6,   0,  -3,   1,   0,   2 },
      {   3,   0,   9,  -1, -11,  22,   7, -15,  -1,  -5, -13, -19,  15, -14,  -9,  -4,  -4,  -1,   8,   0, -15, -21,  -3,   1,  -4, -24, -17, -33, -38, -32,  33, -11,  -6,  -9, -30, -31,   1, -18,  20,  -3,  15,  14,  30, -38,   9,   3,  15,  -2, -14, -28,   4,  -8,   8,   5,   6,   1,  29,  -3,  -5,   9,   3,   3,  -2,  -1 },
      {  -2,  -6,  -5,  -8, -21, -16, -18,   1,  -3,   2,  -6,   1,   1,  11, -38,  12,   5,  -1,   9,   3,  39,  31, -29,  11,  -5,  16, -10,  14, -26,  21, -19,   5,  -5, -26,  -6,  -1, -32,  -4,  13,  -4,   2,  30, -30,  -5,   5,   1,  10,  -5, -11, -26,  -9,  -6,  17,   5,   3,   4,  20,  46,  -1, -11,   4,  -4,  -2,  -3 },
      {   4,   6,   5,   5,  32,  32,  25,  -6,  -1,  -3,  -1,   8, -35,   9,  16, -16,  -3,  -7, -14, -33, -16, -26,  19,  -6,   1,   3,  -3,  10,  -2, -17,   3,   4,  -1, -13,  24,  -8,  -8,  -5,   9,   7,   5, -12, -42,  -3,  -2,  18,  12,   1,  -5,  -6, -28,  16,  -3,   6,   1,  -3,  12,  58, -20,   5,   4,   5,  -2,  -2 },
      {  -5,  -6, -10, -23,   9, -28, -15,   8,  -2,  -4,   0,  21, -25,  -1, -38,   6,  -8,  -8, -12, -18,   7,  -8, -11,   6,   5,  -2,   6,   2, -15,  -6,   8,  -3, -16,   1,  -8,  18,   2,   0,  10,  -5,   8, -34,   0, -18,  18,   9,   7,  -2, -22,  13, -52,  -5,   8,  -2,   4,  -2, -17, -43, -45,  19,   5,  -1,   4,   1 },
      {   0,  -5,  -9,  -5,   6,  13,  45, -15,   7,  12,  22,  23,  16,  40,  21, -20,   0,  -6, -18,   1,  20,  18,  -4,   2,   7,  14,  42,  -5,   4, -16,   4,   0,  -8, -20, -22,  42,  10,  -6,  -9,   0,   8,  15,  -2,   4,  11,   3,  -5,   0, -23, -30,   4,  -8,  -2,   2,  -4,   1,  28, -25,  -3,  -5,  -3,  -2,  -3,   1 },
      {   1,   0,   2,  18,  26, -25,  11,  -8,  -1,  -2,  -8, -22, -14,  34,  -2,  -4,   1,   1,  11, -11,  18,   9,   8,  -6,   0,   0,  -4,  51,  28,  -3, -12,   6,  -4,  -6, -16, -43,  33,  -4,   3,   0,   4, -12,   1, -47,  39, -12,   1,   3,   7,   3,  13, -13,  13,  -5,  -5,   4,   4,  -7,  15,   0,   0,  -6,  -3,   2 },
      {   3,   4,  10,  13,   8,  19,   7,  -8,  -1,   0, -13, -18,  14,  -4,  20, -13,   2,   3,  29,  24,  -3,  54, -30,  -1,   1,  -3, -24,  -8,  18,   4, -14,  -2,   5,  14,  23,  14,  29, -26,   2,   3,  -4, -10,  23,   0,  10,   1,   6,  -6,  -4, -21, -40,  -3,  -4,  -5,   5,  -5, -16,  15, -42,   5,   0,   3,   3,   1 },
      {   1,   2,   2,   9,  35,  19, -19,  -7,  -4,   0,   0, -20, -40, -14,   2,   1,   0,  -5,  -1,  -3, -16,  21,  -4,  -5,  -1,   6,   8,  -6,  40,   5,  -4,  -7,  -3,  -4,  -1,  10, -54,   4,  19,  -4,  -6,   7,  12, -18, -31,  16,  -3,  -1,  -8, -10,  17, -47,  16,   3,  -9,   4,   9, -24,  -8, -15,  15,   0,  -5,   5 },
      {   0,   2,   2,  15,  -8, -21,  10, -34,  -2,  -1,   2, -14,   0,  21, -11,  -9,   1,  -3,   0,   7,  -2,   5,   0,   1,  -4,  10,  -8,  10,  -2, -39,  -6,   9,  -3, -12,  27,  -7, -20, -54,  34, -11,  13,  -2,   0,  37, -25, -19,  30, -11, -14,  31,   2,   9,  -3,  11,   5,   1,  -8, -24,  16,   5,  -5,   6,   0,  -2 },
      {  -2,   0,  -5, -24,  13,  23, -46, -55,   0,  -6,   7,  31,  -9, -22,  -9, -24,   1,  10,   1, -14,  13,  16,  33,   1,   4,  -7,  16,   7,  -3,  22,   2,  -1,   0,  12,  -2,  22,  13, -16,  19,  -4,   2,  -7,  25,  -6,  15, -12,  16,   1,   7,  -2,  -5,   4,  -6,  -9,  12,   4,   2,  18,  37,  -3, -12,  -2,  -3,  -4 },
      {   1,   2,   6,   4,  20,  19, -11,  27,  -2,   1,  -2,  -2, -25,  11,   9,   6,   3,   4,   0,   4,  29,  -1,  29,  -8,  -2,   3,  15,   9, -20,  24, -28,  13,  -1,  -6, -26, -36,   1,  -5,   4,  13,  10,  25,  45,  35,  -7,  14,  12, -11, -19,  -6,   7,  24, -20,  15,   3,  -8,  -9, -11, -32,   8,  -3,   4,  -3,  -3 },
      {   3,   4,   2,  19,  20,  13,  35, -24,  -1,  -3,   2, -14, -19,  -7, -30,   3,  -2,  -3,  -6,   8,   1, -43, -59,  28,   2,   1,   3,  -3,  10,  41, -13,  15,   3,   5,   3,  10,  13,  28,  11,   0,   7,   2,  21,   5,  -9, -16,  11,   3,  -5,  -2,  -7,  17, -10,  -4,  11,   5,  13,  -4,  20,   9,  -3,  -2,   0,  -2 },
      {  -1,  -3,  -2,  17,   4, -35, -22, -36,  -1,  -3,  -6, -32,  -8,  51, -13, -34,   1,   4,   3,  30,  12, -12,  28, -23,   0,  -7,  -4, -37,   6,   3,  20,   2,   4,   4,  -1,  10, -18,  35, -13,   5,  -5,  -1,   8,   6,  -2,   8, -14,   7,   3, -10, -16,  14,  -2,  -6,  -2,   4,   3,  14,  -8,  17,  10,  -6,   4,   0 },
      {   2,   2,   1,   7,  18,   6,   5,  47,   0,  -1,   1,  -8,   1,  20,  15,  27,  -1,  -1,  -1,  14,   0, -24,  25, -11,  -2,  -3,  -3, -22,  -9,   9, -13,  -1,  -1,   1,   6,  18, -11, -17,  15, -23,  -1,  -9,  18,  12,  25, -33,   1,  -8,  -1,  -8, -46, -16,  35,  -7,   1,   4,   7,   5,  40, -21,   8, -13,   1,   1 },
      {   2,   1,   3,   6,  28,   1,   9,  13,   1,   1,   1, -11, -42,  -9,  14,  24,   2,   4,   9,  14,  39,  32, -17,  10,   3,  -1,  -3,  -5, -22, -12,  59, -33,   1,   6,  20,   7, -11,   5,   6,  -4,   5,  -4,  -4,   8,  28,   9,  -6,   0,   4,  11,  17,  30,  -7,  -8,  -9,  -2,   2,  -4,  19,   6, -19,   3,  -7,   0 },
      {   0,  -1,   3,   4,   9,  18,  -5, -24,  -2,  -4,  -9, -13,  -8, -15, -19, -13,   1,  -1,   5,  13,   9, -25,  10,   1,  -3,  -1,  -8, -11, -23,  12,  -2,   6,   3,  -5,  23,  -4,  25,  -7,  -8,   4,   1,   1, -36,  44,  44,   0,  -6,   0,  -5,  -3,  28, -41,  12, -11, -10,  -1,  -2, -23, -29, -40,   4,  -5,   1,   3 },
      {   0,   0,   1,  -4,   4,   2, -16,   4,  -1,  -2,  -1,   3, -13, -10,   8, -26,  -2,  -3,  -2,  -9,  -1, -13, -40, -31,  -1,  -2,  -3, -12,  -8, -36, -59,   1,   0,   5,   4,  13, -29, -14, -21,  24,   0,   8,  14,  -2,  44,  11, -11,  19,   8,   5,  17,  27,  23,  -5,   1,   5,   2,  -4,  13,  16,  -9,   1,   4,   5 },
      {  -1,   0,  -2,   4,   4,  -4,  -8,  10,  -1,   0,  -1, -15, -14,   9,  10,  -6,  -1,  -2,  -1,   9,  17,   2,  -8, -19,   0,  -4,  -7, -24, -36,  -2, -26,   6,  -3,   6,   3,  14,  27,  -2,  -1,  12,  -2,  -9, -20, -35, -23,  30,  17,   8,  -3,  14, -12, -30, -57,  30,  -1,  -3,   0,  -9,  36, -25,  -7,   6,  -3,   2 },
      {  -1,   0,   2,   0,  -7,   4,   4,  -1,  -1,  -2,  -4,  -5,  14,   1, -23,  14,   1,   5,   8,   7, -24,  10,  18,  44,  -1,   0,   2,   1,  23, -19,   0,  35,   1,   2,  -2,   2,  -7,  16,  17,  18,   5,   1,  15,  15,  38,  57, -15,   5,  -8,   7, -21,   2,   2,  32, -23,  -3,   1,   4,  23, -14, -19,  17, -14,   3 },
      {   0,  -2,  -2,   2,   9,  -6,  -9, -23,   0,   1,   0,  -2, -18,  -2,  27,  31,   1,   2,   5,   8,  10,  -4,  19,  71,  -3,  -5,  -9,  -9,  -7, -28, -20,  24,  -3,  -1,  -5,   4, -22,  -5, -55,   8,  -2,  -3,  -2, -16,  -1, -30,   7,  -2,   1,  -8,  -2,  -8, -13, -11,  17,  -9,  -1,   0,  -9,   1,  -2,   2,   8, -11 },
      {   0,  -2,   1,   1,  -7, -11,  14,  24,   1,   0,  -1,  -4,   3,   7,  -5,   3,   1,   2,   8,  14, -13,  -9,  31,  12,   1,   1,  -3, -21,  12,  32, -12,  11,   2,   3,  13,  42,  -9, -22,  37,   0,   2,   2, -18, -37,  25,   9,  27,  -4,   5,  -3,  50,  15, -12,  -8,   9, -13,  -3,  -2, -20,  31,  -6,  -1,  -2, -11 },
      {  -1,  -2,   0,  -2,  -3,  -9, -47, -12,   2,   4,   3,   5,  11,  25,  63,   6,  -1,  -2,  -2,  -5,  -8, -16, -29,  23,  -1,   1,   1,   3,  -6, -14, -10,  -7,   2,   0,   6,  -4,  14,  44,  49, -19,   1,   2,  -3,   7,  11,  14,  23, -13,   0,  -2,   9,   8,   7,  -6,   3,  -6,   2,  -5, -11, -15,  12, -13,  -6,  -3 },
      {   0,   0,   2,  -1,  -3, -11, -16,  34,   0,   1,  -2,   1,  -1,  -2,  18, -54,   0,  -1,  -1,  -3,  -1,  -5, -22,   5,   2,   3,   2,   4,   8,  13,  47,  50,  -1,  -1,  -4,  -5, -23,  -8,   8,  36,   1,  -1,   0,  -7,   8, -39,   8,   7,  -2,  -2,  -7,   4,  -9,  -1,   2, -10,  -1,  -5,  -5, -38, -16,  16,  -7,  -4 },
      {   1,   1,   1,   2,   3,   4,  14,  12,   0,   0,   1,   0,  -1, -20, -15,   3,   1,   1,   1,   2,   7,  29,  13, -15,   1,  -1,  -1,  -1,  -7, -30,   4,  44,   1,   4,   4,   2,  16,  29,   7,  37,  -1,   2,  -2,   8, -15,  15,  33,  -3,  -1,   0,   0,   4,  20, -47,  20, -17,   7,  -8,  18,   3,  40, -38,   8,  -8 },
      {  -1,   1,   0,  -1,  -9,  -5,  16,  -7,   0,  -1,   0,   2,   9,   2, -17,  17,  -1,   1,   0,  -3, -15,   1,   7, -25,  -1,  -3,  -1,   6,  24,  -9,  -9, -47,  -2,   3,  -3,  -8, -30,  25,  -8,  10,   1,   2,   7,   6,  26,  -6,  47,  10,  -7,  -7, -10,   3, -41, -23,  20,  -8,   0,  -3, -10, -48, -32,  -7,   4, -14 },
      {   0,   0,   0,  -2,  -1,  -2,  -2,  20,   0,  -1,   0,   2,  -3,  -3,   4, -16,  -1,   0,  -3,  -1,  -5,   7,  -2,   4,  -2,  -1,   1,   5,  -5, -15,  -4,   6,   2,  -2,   1, -10,  -5,  24,  21,  -2,   0,   3,  -2,  30,  12, -16,  -2,   6,  -1, -10,  -3, -67, -31, -19,   4,   9,   5,  13,   9,  68, -42,  -2,  11,   7 },
      {   0,   0,   0,  -4,  -1,  -7,  -4, -12,   1,   2,   1,   8,   4,  15,  18,  26,   0,   2,   2,  -1,   1, -11, -11,  -9,   2,  -1,  -2,   3,  10,  25,  26, -18,   1,   5,   4,   3,   3, -10, -21,  55,   0,   3,   0,  18,   7,  12,  42,  12,   3,   3,   1, -31,  35,  44,  35,  -9,   0,   1,   9,  28,   9,  13,  -5,   4 },
      {  -1,   0,  -1,  -1,  -1,  -8,   0, -27,   1,   1,   3,   2,  -1,   7,  12,  61,   0,   0,  -3,   0,  -1,  -5,   0, -27,  -1,   0,   4,  -1,  -9,  -1,  -2,  17,   2,  -1,  -3,   7,   7, -11,  45,  45,  -1,   1,   1,  -4,  -9, -19, -42,  42,  -2,  -5,  -2,   1, -11, -28,  -8,  14,   1,   5, -17,  -8,  -3,  24,   0,   1 },
      {   0,   0,  -1,   0,   3,   7,  -1, -22,   1,   1,   2,  -1,  -2,  -4,   3,  34,  -1,  -2,  -4,   0,  -3,   2, -10, -52,   1,   1,   1,  -3,   1,  -5,  13,  55,  -1,   0,  -2,   7,  -3,  15, -12, -27,  -1,  -1,  -3, -11,   8,  -3,  -1, -57,   5,   0,   8,   1,   9,  24,  17, -11,  -4,   4, -11,  -2, -48,  -3,  13,  12 },
      {   0,   1,   1,   0,   5,   7,   5,  10,  -1,  -2,  -1,  -2,  -2,  -9, -14,  -5,   1,   1,   1,   7,   8,  18,  19,  -3,  -1,  -2,   0,  -8,  -7, -26, -17,  -1,   2,   2,   3,  15,  17,  48,   4,  -5,   0,  -1,  -3,  -5,  -8, -45,  39,  19,   4,  -1,  11,  10,  29,  29, -14,  32,  -6,   1, -16,  -2,  -4,  50, -25,  21 },
      {  -1,  -1,  -1,   3,   3,  -8,  -2,   5,   0,   0,  -1,  -4,  -4,  10,   4,  -8,   0,   0,   1,   3,   1,  -7,   4,  16,  -1,   0,  -1,  -3,  -8,   9,  -1,  -2,  -1,  -2,  -4,   0,  17, -13,  -8,  -2,   0,   1,   0,  -4, -43,  34,  16,  27,   1,  -2,  -2,   2,  52, -37,  -1,   8,  -3,  -1,  -5, -13, -81, -10,   2,   4 },
      {   0,   0,   0,   0,  -1,   4,  -4, -10,   0,   0,   0,   1,   1,  -1,  -5,   5,   0,   0,   0,   1,  -1,  -5,  -2,  -5,   0,  -2,   0,  -1,  -3,  -3,  -9, -17,   1,   1,   0,   6,   2,   5,   1,  35,  -2,  -2,  -2,  -5, -11, -29,   5, -23,   0,  -4,  -6,  -2,   7,  13, -77, -48,  -2,   1,  -1,  15, -20, -41, -37, -26 },
      {   0,   0,   0,   0,   0,  -4,   1,   1,   0,   0,   0,  -1,  -2,   7,  -7, -10,   1,   0,   0,  -1,   1,  -4,   1,   0,   0,   0,   0,   0,  -4,   0, -12, -20,   1,   0,  -1,  -1,   7,   7,   1,   9,   0,   0,  -1,  -1,  -7,   8, -12, -48,  -1,  -2,  -3,  -8,  14, -42,   7, -56,   6,  -1,  16,   0,   1,  86,   2,  -6 },
      {   1,   0,  -1,   4,   5,   5,   6,   5,   0,   1,   1,  -3,  -4,  -5,  -7, -16,   1,   0,   1,   2,   5,  10,  12,  11,  -1,   1,  -3,  -2,  -4, -12, -11, -21,   1,  -1,   3,   1,   8,  21,  29,  16,  -1,   0,  -4,  -5,  -8, -19, -41,   7,   2,   2,   3,   3,  21,  47,  44, -19,  -2,  -5,  -5, -11, -19, -13,  48, -54 },
      {   0,   0,   0,  -1,  -2,   0,   4,   1,   0,   0,   0,   1,   0,   3,  -5,  -8,   0,   1,   2,   0,   2,  -2,  10,  20,  -1,  -1,  -3,  -1,  -3,  -1,  -8, -25,   0,   0,   1,   2,  -4,   2,  17,  41,  -1,  -3,  -1,  -7,  -6, -12, -13, -33,   0,   0,   2,  10,  -3,   5,  11,  -5,  -1,   1,  -2, -12,  -7, -26,  23, 100 },
      {   0,   0,   0,   1,  -1,  -2,   4,   4,  -1,  -1,  -2,  -3,   0,   1,  -4,  -6,   0,   0,   0,   1,  -1,  -2,   2,   3,  -1,   0,   0,  -3,  -2,  -7,  -7, -10,   0,   0,   0,   0,   2,   0,   3,  18,  -2,  -3,  -4,   0,  -5,   2, -30, -36,   2,   4,   0,  -4,  -2,  -8,  54,  36,  -6,  -1,  -4,  -2,  -9, -14, -92, -16 },
      {   0,   0,   0,  -1,   0,  -1,  -2,   3,   0,   1,   0,   0,   2,   2,   4,   0,   0,   0,   0,   0,  -2,  -4,  -1,   5,   0,   1,   1,  -1,   2,   4,   2,  -9,   0,   1,   1,   3,   2,  -2,  -2,  38,   0,  -1,  -1,  -4,   0,   7,  14, -59,  -1,   3,   0,   1,   0,  -9, -32,  81,   1,  -1,   4,  -1,  -1,   9,  44, -35 },
    },
  },
  { // 16
    {
      { 111, -30,  -9,  -5,  -3,  -2,  -1,  -1, -52,  10,   5,   2,   2,   1,   1,   0, -10,   5,   0,   0,   0,   0,   0,   0,   4,  -1,  -1,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,  -4,   1,   0,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
      {  43,   4,  -7,  -2,  -3,  -1,  -1,   0,  95, -25,  -6,  -4,  -3,  -1,  -1,   0, -64,  -1,   9,   2,   3,   1,   1,   1, -20,  11,   1,   0,   1,   0,   0,   0,   4,   2,  -4,   0,   0,   0,   0,   0,  -1,  -2,   0,   1,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
      {  24, -15,  -1,  -1,   0,   0,   0,   0,  35,   3,  -6,  -1,  -1,  -1,  -1,   0,  86, -26,  -7,  -2,  -2,  -1,  -1,   0, -73,  -3,  10,   1,   3,   1,   1,   0, -21,  12,   1,   1,   1,   0,   0,   0,  -2,   2,  -1,   0,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
      {  -9,  18,   1,   2,  -3,  -2,   0,   1, -36,   5,   0,  -1,   0,   0,   0,   0, -22, -23,  10,   0,   3,   0,   2,   0, -66,  43,   5,  -2,   3,  -1,   0,  -1,  77,   6, -27,   3,  -2,   1,  -1,   0,  14, -21,   3,   4,   0,   0,  -1,  -1,   1,   0,   5,  -3,   0,   0,  -1,   0,   1,  -1,   1,  -1,   0,   0,   0,   0 },
      { -10,  12,   1,   2,  -1,   0,  -1,   0, -31, -20,   8,   1,   3,   2,   2,   1, -38,   5,   1,   0,   1,   1,   1,   0, -44,  -8,   5,   4,   2,   1,   1,   1, -68,  34,   9,   0,   1,   0,   1,   0,  67,  12, -12,  -1,  -3,   0,   0,   0,  19, -13,  -2,   1,   0,   0,   0,   0,   3,  -2,   0,  -1,   0,   0,   0,   0 },
      { -29,-101,  20,  -2,  13,   2,   5,   1,   3,  19,   4,  -1,   0,  -2,  -2,  -1, -23,  49,  -7,  -2,  -5,   0,   0,   0, -32,  -5,   5,   1,   1,   0,   1,   0,  10,  -7,   4,   3,   1,   1,  -1,   0,  -7,   7,   2,  -3,   1,   0,   0,   0,   3,   7,  -5,   0,   0,   0,   0,   0,   2,  -2,  -3,   1,  -1,   0,   0,   0 },
      {  11,  -7,  11,  -8,   0,  -3,   2,   0,  26,  39,  -7,  -2,  -6,   2,  -1,  -2,  21,  18,  -4,   3,  -1,   0,  -4,  -2,  28, -19,  -5,   0,   2,  -3,   1,   2,  32,  24,  -7,  -1,  -5,  -1,   0,   0,  85, -30, -11,  -1,  -3,  -1,   1,   0, -32, -21,   7,   1,   1,   0,   1,   0, -11,   5,   4,   0,   1,   0,   0,   0 },
      {  -1, -13, -27,   7,  -2,   6,  -2,   2,  -8, -91,  25,   3,   9,  -2,   3,   1,  28,  42,  12,  -8,  -4,  -3,  -1,   0,   3,  20,  -6,  -5,  -3,   0,  -1,  -1,   6, -11, -11,   4,   2,   3,  -1,   0,  16, -16,   1,   1,   1,   1,  -2,   0, -45,   3,   7,  -2,   2,   0,   1,   0,  -4,   9,  -1,  -1,   0,  -1,   0,   0 },
      {   1, -35,  18,   4,   1,  -6,   4,   2,   9, -10,   1,  -7,  -5,  -1,   0,   1,  17, -23,  27,  -6,   3,  -1,  -2,  -1,  38,  67, -17,  -8,   1,  -4,  -2,  -2, -16,   4, -37,   6,  -1,   3,   1,   1,  21,   1,   4,   6,  -1,   0,  -2,  -1,  62, -13,  -2,  -6,  -2,   0,  -1,   0, -17, -10,   5,  -1,   2,   1,   0,   0 },
      {  10,  44, -28,  -2,  -3,  -1,  -2,   0,  10,  26,   5,   1,  -2,  -3,  -1,  -1,  18,  84,   1, -11,  -4,  -5,  -1,  -2, -16,  11, -14,   2,   1,  -1,   0,   0,   9, -30,   1,   6,   1,   1,   0,   1,   9,  10,   6,  -3,   1,   0,   0,  -1,  53,   2, -12,  -2,  -2,   1,  -1,   0,   5, -18,  -2,   2,   0,   0,   0,   0 },
      { -14, -23, -97,  35, -10,  16,  -6,   3,   2,  26,   8,  -8,   7,  -9,  -3,  -1,  -6, -19,  32,  -9,  -1,  -5,   6,   1,   4, -17,  12,   2,  -3,   4,   1,  -2,   4,  13,  -3,   1,   5,   3,  -2,  -2,  -2,   4,  -9,   2,  -1,   1,  -3,   1,   6, -16,   3,   2,  -1,  -1,  -1,   1, -30,   0,   5,   0,   1,   0,   1,   0 },
      {   4,  -7,   7,  -9,   0,   2,  -1,  -1,   2, -49,   9,  10,   5,  12,   3,   3,  10,  -8,  -9,  10,  -2,   5,   1,  -2,   8, -37,   3,   5,   1,   2,   1,   4,  59,  28,  35, -13,  -4,  -7,   0,  -1,   5,  41, -22,  -5,  -4,  -2,   4,  -1,  49, -17, -16,   7,  -1,   1,   0,  -2, -17, -15,   4,   1,  -1,  -1,   1,   0 },
      {  -1, -10, -22,  -1,  -3,   2,   2,   3,   6,  -1,  14,   2,  -5,   5,   0,  -1,   7,  -2,  12,  -6,   6,  -2,   1,  -3,  16,  -2,   2,   2,   6,  -3,   0,   0,   9,  53,  -4,   3,  -4,  -2,  -1,   1,   1, -14,   5,  -3,  -2,   1,   4,   1,  15,  37,  -6,  -1,   0,   0,  -1,  -2,  97,  -8, -22,   2,  -3,   1,  -2,   1 },
      {   0,   6,  24, -15, -14,  -6,   7,   1,   5,  10,  79,   4,  -7,   2,  -1,  -4,  -2,  -8,  50,  -3,   3,   1,  -9,  -3, -14, -49, -41, -18,   3,  -6,   0,   2,  -1,  -2, -29,   9,  -1,   1,   1,   0,  -9,   5,  12,   8,   1,   0,   0,  -1,  -2,   6,   1,  -5,   1,   1,   0,  -1, -12,  -1,   2,  -1,   2,   0,   0,  -1 },
      {  -1, -16, -14,  17,  -7,   3,  -1,   3,   3,  -2,   4,   9,  -2,   6,  -1,   2,  -2, -43, -10,   4,   1,   1,   5,   0,  -2,  -4, -35,   7,  -1,   5,   1,   2,   8, -74,  33,   6,   1,  -1,  -1,   1,  51,  16,  28, -13,  -2,  -5,   2,   0,   1,  26, -12,  -3,   1,  -1,   1,  -2,  22,  -3, -12,   4,   1,   0,   0,   0 },
      {   2,   0,   9,   8, -16,  -4,   3,   0,  -7, -17, -57,   3, -12,  -4,  -4,   2,   2,   5,  35,  11,  -7,  -4,   0,   0,   3, -44,  47, -27,  12,  -5,   3,  -3,   4, -25, -53,  11,   4,   7,  -2,   0,  17,  31,   6,   6,   5,  -2,  -6,  -4,   5,  18,  -1, -12,   0,  -1,  -1,   2,  11,  -4,   1,  -1,  -1,  -1,   0,   0 },
      {   6,   7,  16,  48, -26,  -1,  -1,  -4,   9,   9,  62, -20,   3, -19,  -3,  -3,   0,  -2, -54,   7, -14,  -3,  10,   5,  11,  14,  51,  -3,   0,   1,   2,  -7,   4,   0, -21,   6,  10,   8,  -6,  -2,   9,  29,  -2,  -3,   3,  -1,  -7,  -1,  -4,   5,  -7,  -4,   1,   0,  -2,   1,   1,  -3,  -7,   2,   0,  -1,   0,   1 },
      {  -8,  -2, -31, -89,   0, -11,   6,  -7,   3,  13,  20,  27,  -5,  -1,  -6,  -5,   1,  -9,   1,  18,   2,   1,   2,   0,   7,  33,  35, -20,   4,  -7,   2,  -1,  -4,  -4,   5,  10,   1,   0,  -6,   1,   9,  37,  -3,  -9,   2,   0,  -1,   0, -22,   5, -14,   0,   2,   0,  -2,   1,  -5,  -8,  -3,   5,   0,   1,   0,   0 },
      {   5,   9,  29,  30,  14,   7,  -5,   0,   2,  17,   2,  -1,   1,   0,   3,   0,   8,   7,  68, -19,   2,  -9,   1,  -3,   0,  31,  10,  15,  -6,  -1,  -1,   2,   5,  -4,  32, -11,  -3,  -1,  -1,  -1,   1,  54, -30, -10,   2,   0,   2,   1, -34, -18, -11,   9,  -2,   2,   3,   0,  22,  11,  -1,   1,  -3,  -1,   0,   1 },
      {   3,   3,   0, -35,  55,  -7,   0,  -5,  -2,  -9,  16, -74,  20, -21,   5,  -3,   1, -13,  14, -22,   0,  -5,   5,   4,   4, -20,  33,  35,  -1,   0,  -1,  -2,   4, -28,  -1,   5,   0,   5,  -2,   0,  10, -19,  -5,   2,   0,   4,  -2,   1,  16,   1,   2,   2,  -4,   1,   1,   2,   7,   0,  -4,   1,  -3,   0,  -1,  -1 },
      {  -3,  -4, -18, -14,  -7,  -7,  11,  -2,  -2,   0, -11, -46,   4,  -5,   5,  -1,   1,   1, -22,  44, -13,   0, -10,   2,  -5,   1, -51,   3,  -4,   8,   1,   2,   3,  -2, -31,  -1,   5,   5,   4,  -2,  -6,  45, -18,  12,   2,  -4,  -4,  -2,  -9, -41,  17,   3,  -1,   3,   0,   0,  37,  27,   7,  -5,  -1,  -2,   0,   1 },
      {  -2,  -6,   6,  -5, -48,   0,  -9,   2,   1,   1,  16,  -6,  -7,   9,   4,   1,   5,   7,  31,  73,   7,   1,   1,  -5,  -2,   5,  31,  17, -10,   5,   1,   6,  -6, -27,  21, -32,  -3,  -3,   2,   1,  -1, -43, -22,   3,   2,   3,   6,   3,  21,  -2,  11,   7,  -1,  -1,   0,  -1,   8,  16,  -1,  -4,  -3,  -2,   0,   1 },
      {  -3,   1,   4,  -8, -45,   0,  -6,   1,  -6,   0, -19, -70,  -5,   3,   4,   2,   3,   9,  12,  10,   6,   1,  -2,  -4,   1,  12, -18,  11,   4,   5,   0,   0,   1,  32,  19,   1,  -2,  -1,   0,   0,   1,  16,  24, -10,  -2,  -3,   3,  -1, -21,  49, -20,  -7,   2,  -2,   2,  -2, -34, -31,  -4,   6,   3,   0,   0,  -1 },
      {  -4,  -8,   4, -23, -70, -10,   0,  -2,  -2,  -7,  -6, -12, -32,  -4,   0,   0,  -2,  -3, -12, -71,  36, -10,   9,  -2,   0,  -6,   7,   5,  -1,   2,  -1,   0,   3, -12,  10,  12,  -1,   1,   0,   3,  -4,  -1,  -8,   0,  -1,   7,  -2,   2,   6, -33,  13,   6,  -1,   2,  -1,   1,  12,  25,  -6,  -4,   5,   0,  -3,   0 },
      {   1,   4,  -8,  11,  19, -44,  26, -16,  -2,  -4,   0,   9, -42, -24,   5,  -5,   2,   6,   6,  14,  -5,  -9,  -3,  -1,   2,   1,   0,  22,   8,   2,   3,   0,   6,  21,  15,   9,  -3,   8,  -2,   0,   2,   9,  24, -12,   3,   2,  -5,   0,  14,  25,   6,  -9,  -2,   1,  -2,   0, -23,  76, -21,  -2,   3,   0,  -2,   1 },
      {  -1,  -2,  -9,   9,  17, -16,   0,   2,  -5,  -8,   6,  -1, -94,   6,  -9,   3,  -2,  -4,  -4,   6, -19,  -2,  -2,  -2,  -2,  -2,  -2,  34,  13,  -5,   2,   0,  -5,  -8, -10,  -8,   0,   4,  -2,   1,  -7,  -9, -10,  -7,   9,   2,   1,  -1, -15, -23,  -9,   2,  -1,   2,   1,   1,   4, -56,   6,   7,   2,   2,  -1,   0 },
      {   1,   2,  -9,   0,  20,  22,  -5,  11,   0,   4,   8, -24, -39,  38,  -3,   8,  -1,  -1, -15, -20,   2,  12, -12,  -4,  -1,   7,   0, -41,  10,  -7,  -1,   6,  -1, -10,   0, -30,  -6,  -7,   9,  -3,   8,  10, -49,  29,   1,  -1,   6,   0,   7,  52,  13,   7,  -2,  -2,   1,  -2, -10,  30,  10,  -8,  -1,   2,   2,   0 },
      {  -1,  -3,  -9,  24,   4, -79,  42, -17,   0,   0,  -1,  -5,  17, -12,  15,  -7,   1,   1,  -2,   1,  47,   8, -11,  -3,  -2,   0,   0, -33,  -4,  -2,   1,   3,  -1,  -8,  16, -13,  -4,   1,   8,   2,   0,  -5, -25,  19,   0,   0,  -2,   2,  -8,   5,  -1,   1,  -6,   4,  -3,   0,   4, -36,   9,  -1,   2,   3,   0,   1 },
      {  -1,  -1,  -1,  -8,  -2,  10,   1,   6,   2,   2,  -2,  31,  10, -14,   5,   0,   1,   1, -11,  -5,  43,  -3,  -8,   3,   0,  -3,  -7,  69, -30,   2,  -8,   1,   0,   0, -43, -35,  -5,   7,   4,   0,   7,  26, -13,   9,   7,   9,  -2,   8,  -4,  36,  24,  10, -12,  -3,  -1,   0, -12, -13,  -9,  -6,  -1,   1,   0,   2 },
      {  -3,  -3,   3, -11, -25, -40, -22, -15,   1,   2,  -7,  17,  18, -16,   2,  -4,  -3,  -8,   5, -21, -74,  -5,  17,   6,   1,   0, -20,  18,   0,   9,  -4,  -6,  -1,  -3,  -1,   7,  19,   6,  -1,  -2,   0, -11, -57,   3,   2,  -3,  -1,  -9,   3,  34,   1,  -2,   5,   4,  -1,  -1,  -2,  -4,   9,  -5,  -2,   0,   2,  -1 },
      {   2,   2,   6,  -5,  -2,  51,  26,   7,  -3,  -5,  -7,   3, -20, -83,  10, -12,  -1,  -6,  -2,   9,   8, -26, -10,   3,  -2,  -1, -15, -26, -15,   1,   6,   1,   1,   0,  13,   7,   0,   8,  -2,  -1,   1, -15, -30,  -7,  19,   6, -12,  -2,   4,  12, -28,   8,  -3,   0,   0,   4,   2,  -3,  14,   5,   0,  -1,  -2,  -2 },
      {   0,   0,  -2,  -8,  -7,  -3,  12,  -4,   0,   0,  -1,  -5,  -7, -14,   5,   3,  -1,  -2,   7, -23, -48,  28, -25,  10,  -2,   1,   9, -24, -53,   2,  -5,  -5,   2,   5,  10, -63,  13,   6,  10,  -1,   7,   3,  43,   6,  20,   7, -10,  -2,   3, -11,  13,  13,  -3,  -5,   1,   1,   6,  -7, -18,   0,   1,   1,  -1,  -3 },
      {   1,   3,  -1,   2, -15,  24,  80,  -5,   1,   1,   0,   1,   7,  39,  12,  -4,  -2,  -2,   2,  -7, -24,  24, -44,  -3,   1,   3,  15,  30, -18,   0,   4,   7,  -1,  -5,   2,  45,  -1,  -5,   5,  -3,  -2,  -6, -15,  -7,   4,  -6,  -3,   1,   2,   5,   1, -10,   1,   1,  -2,  -1,  -2,  -2,  11,   0,   2,  -2,   1,   3 },
      {  -1,  -1,  -7,   0, -10,  -8,  22,   3,   0,  -1,  -2,  -1,  18,  20,   2, -10,  -2,  -2,  -7, -12,  -4,  -9,   7,  -3,  -2,  -3,  -4,  12,  22, -31,   7,   5,  -6,  -5, -30, -49, -12,   5,   0,   2,  -2, -11,   4, -27,  14,  -4,   4,  -8,  -3,  -4, -77,  30,   8,   5,  -1,   1,   0,  27,  17,   8,  -2,   1,  -6,   1 },
      {   1,   2,   3,  -2,   3,  19,  38, -40,   1,   0,  -1,  -4,   3,  33,   2, -17,  -1,  -1,   0,   2, -12, -77,  43,  -6,  -1,   0,  -7, -12,  -5, -10,  16,  -6,   0,   1,   5, -18,   2,  16, -15,   4,   0,   0,   4,   2,  -2,   4,  -1,   3,   1,   2,  33,   7,  -8,   2,  -8,  -3,  -6, -17, -22,  -4,   1,  -4,  -3,   2 },
      {   1,   0,   1,  -4, -11,  15,  33,  19,   1,   1,   2,   6,   9, -34,  21,   1,   0,   1,   4,  -7, -16,  14,  -1,  -1,   2,   6,   2,  10,  90,   1,  -4,   7,  -1,  -5,   9, -31,   6, -19,   0,   0,   1,   7,  13,   7, -24,  -1,   2,  -1,  -1,  -1,  38,   8,   1,  -2,  -5,  -3,   1, -12,   3,  -5,  -1,   4,  -2,   0 },
      {  -1,  -1,  -2,   0,  -5, -19,  -9,  38,   0,  -1,   5,  -1,  10,  -2, -26,   3,  -2,  -2,  -3,  -2, -10, -34, -26,   0,   1,   2,   6,   5,  -3, -43,  19,   0,   1,   4,  19,   2, -39,  16,  -2,  -1,   2,   8,  17, -11,  27,   0,  -5, -12,   4,   8,  40,  -7,   2,  -4,   1,   2,   9,  -2,  63,  -7,  -6,  -3,   1,  -3 },
      {   0,  -1,  -1,  -4,   3,  -3,  -7,  23,   2,   3,   4,   7, -12,  15,  77,  -1,   0,   1,  -1,   0,  -5, -26,  15,   4,   2,   0,  10,   8, -12,  18, -16,   6,   2,   5,   6,   7,  22,   7,   0,   0,   3,   3,  28,  48,  -8,  -3,  -8,   6,  -4,   2, -23,  -2, -14,   2,  -5,   0,   7,   6,  51, -12,   0,   1,  -6,  -1 },
      {  -1,  -2,  -1,   4,  -1,  18, -14, -81,  -1,  -1,   1,  -1, -11,  -9,  12, -28,   3,   3,   3,   1,   6,  46,  23,  10,   0,   0,  -2,   3,  -2, -19,  -5,  -5,   0,   0,  -1,   2, -24,  -3,  -6,   2,   3,   5,  12, -12,  -5,   1,  -1,   3,   2,   8,  14,  14,   5,  -7,  -9,   0,   9,   1,  55, -10,   0,  -3,  -2,  -3 },
      {   0,   1,  -2,   2,   2,  -8,   2,  44,   0,   2,  -3,  -4,  -3,   2,  64,  -2,   0,   0,  -2,   3, -10,  27,  44,  -5,  -2,  -2,  -5, -13, -17, -31,  -5,   0,  -2,  -3,  -9,  16, -39,  -5, -12,   0,  -3,   3, -14, -39,  14,  16,   6,   5,   1,  -1,  21,  11,   7,  -3,  -3,  -3,  -6,  -5, -29,   7,  -1,  -5,  -4,  -2 },
      {  -2,  -3,  -3,   1,  -1,   4, -38, -39,   1,   2,   2,   3,   6,   8,  55,  -3,  -3,  -6,  -6,  -2,   1, -31, -67,  -3,   0,  -1,   3,  -7,  28,  -3, -18,  -4,   0,  -4,  -4,  -3, -10,   7,  25,   0,  -1,  -2,  -6, -30,  27,   3,   1,  -1,   1,  -6,   4, -19,  -5,   2,   8,  -3,  -3,  -4, -23,  -5,  -1,  -5,   2,   0 },
      {   0,   1,  -1,  -3,   4,  -6,   9,   4,   1,   1,   5,   1,  -4,   0,   1,   6,  -1,   0,   0,  -8,  -6, -18,   2, -11,   1,  -1,   6, -11, -11,  90,  -3,   0,  -3,  -2, -16, -18, -50, -14,   1,   3,   0,   1,   1, -39, -21, -13,   9,   2,   1,   5,   8,  -4,  24,   4,  -2,   1,   3,   0,  26,  15,   3,  -4,   4,   3 },
      {  -1,  -1,  -1,   2,  -7,   2,  -5, -11,  -1,   0,  -2,   4,   2, -12,  -5,  -3,  -1,  -2,  -2,   2, -24, -10, -16,  -6,   0,   1,   4,   8,   0,  -9,  -6,  -2,  -2,  -3,   4,  16, -81,   2,   6,   6,   0,   3,   5,  69, -25,  12,   6,   2,  -4,  -5, -22,  26,  12,  -3,  -1,  -3,  -3,  -3, -30,   1,   0,   2,  -2,   1 },
      {   0,   1,   3,   2,   0,  -6,  -2,  -6,   1,   1,   1,   0,   1,   9,  -2,  43,   1,   2,   4,   6,   3,  -4,  -9,  33,   1,   1,   1,  -8,  13,  22,  -2,   2,   1,   3,  -2,  30,   7,  15,   0,  -6,   2,  -2,   7, -12,  18,  -2,  -5,  -2,   2,   8,  12,  98, -26,   3,  -3,  -2,  -3,  -3,   2,   6,  -8,   6,  -2,   5 },
      {   0,  -1,   0,   5,   3,  -9, -15,  18,  -2,  -4,  -4,  -3,  -4,  -2,  -7, -81,  -1,  -2,  -2,   1,  -5, -11, -24, -45,   0,   0,   6,  -5, -13,  -5,  -8,  12,   1,  -1,  -5,  13,  28, -25,  10,   6,   1,   3,   6, -12, -25,  -6,  19,   8,   2,   5,  12,  50,   1,   2,  -4,   3,  -2,  -1,   1,   3,   0,  -1,  -7,   1 },
      {   0,  -1,  -2,   0,  -2,  -4,   9,  -9,   0,   0,   1,  -3,  -6, -12,   0,  48,  -1,   0,  -2,   4,  -3, -29,  -6,  24,  -2,  -2,   2,  -3, -26, -47, -44,  -4,   0,  -2,  -6,  -1,  13, -37,  22,  -6,  -1,   1,  -1, -21, -63,  -1,  22,  -3,   0,   3,   2,   0,  11,   1,   0,  -2,   1,   2,  10,   9,   6,   0,   1,   3 },
      {  -1,  -1,  -1,  -1,   0,   0, -10, -15,   1,   1,   2,   2,  -1, -11,  22,  22,   0,   0,   0,  -4,   1,  -4,   5, -22,   1,   1,   2,   8, -11, -12,  67,  -6,  -1,   0,  -6,   1,   1, -63,  11,  -4,   0,   1,   4,  17,  19, -55,  -4, -19,   0,   0,   6,   6,  -4,  11,   3,   1,  -1,   1,  -4,  29,  -3,  13,   4,   1 },
      {   0,   0,   2,  -1,   0,   1, -13,   3,   0,   1,   0,   2,  -2,   4,  21,   2,   0,   0,   0,  -1,   3,  11, -21,   2,   0,   0,  -2,  -1,  -3,  -9,  55,  -6,   1,  -1,   0,  -6,  10,  59,  -3,  14,  -3,  -1,  -8, -11, -68,   5, -32,   4,   2,   2,   5,   4,  12, -10,   2,   2,   1,   2,   1,  44,   4,  11,   1,   3 },
      {   0,   0,   1,  -2,   0,   5,   5,   9,   0,   0,   0,   2,  -4,   3,   0, -27,   1,   2,   2,   0,  11,   2,  10,  27,   0,   0,   2,   3,   1,  -5, -44,  -6,   0,   2,   7,  -1,  -3,  38,  15, -15,  -1,  -2,  -3,  10,   9, -83,   1, -34,   2,  -3,  11,   1,  33,  -8,  -4,   2,  -2,  -1, -12,  26,  -8,   9,   6, -11 },
      {   0,   1,   0,   0,   4,   3,   7, -10,   1,   2,   1,   1,   3,  -3,   2,  43,   0,  -1,  -1,   2,   5,   3,   4, -78,   0,  -1,   0,   5,   2,  -3, -23, -28,   1,   1,   2,   5,  22,  18,   3,  29,   2,   2,   2,   8,  19,  12,  -6,  -7,   1,   2,   5,  17,  55, -14,  -5,   1,   0,   0,   8,   5,  -5, -10,   2, -24 },
      {   0,   1,   0,  -3,   2,   4,   2,   8,   0,   1,   2,   1,  -2,  -7,   0,  10,   0,   0,   0,   1,  -2,   9,  12, -27,   0,   0,   2,  -1,  -5,   1,  20, -28,   0,  -1,   0,   1, -11,  29,  50, -10,  -2,  -2,   0, -19, -27, -30,   5, -18,  -4,  -2,  -9,   5, -34,   4, -13,  -1,   0,  -4,  -6, -78,  18,   2,   4, -13 },
      {   0,   0,   0,  -1,  -4,   4,   2, -10,  -1,  -1,   1,   1,   4, -11,  -5,  17,  -1,  -1,  -3,  -4, -11,   9,  12, -45,  -1,  -1,   1,  -4,  -1,  -4, -19,  36,   0,   1,  -1,   0,  -7,  35,  12,   5,  -1,   0,   1,   0,   4,  -9,  48,   4,  -2,  -3,   4, -17, -65,  -6,  -5,  -3,   0,   5,   9,  48,  -9,  14,  -6,  22 },
      {   0,   0,   1,   0,   2,   3,   2,  -1,   0,   0,   1,   1,   0,   0,   0,  -8,   1,   1,   1,   0,   2,  -2,  22,  21,   1,   2,   2,   3,   3,   7,  30,  24,   2,   2,   0,   8,  11,   9,  86, -14,   2,   4,   5,   6,  19,  39,  31,   4,   0,   0,   3,  -1,  46,   0, -17,  -5,   0,  -2,   3,   8, -10,   9,   4,  14 },
      {   0,   0,   0,   1,   2,  -2,  -6,  -8,   0,   0,   0,  -1,  -1, -10,   8,  20,  -1,  -1,  -1,  -4,  -1,  -4, -11,   0,   1,   1,   2,  -3,  -8,  -5,  19,  76,  -1,   0,  -5,   0,   7,  15, -39,   5,  -1,   0,   0,  -4,   0, -31,  31,  18,   1,  -2,  -1,   4,  39, -28,   0,   1,  -1,  -5,  -8, -47,  13,  15,  -3,   6 },
      {   0,   0,   0,   0,   2,   3,  -6,  -7,   0,   1,   1,   0,   2,   4,  -5,   4,   0,   0,  -1,   2,   0,   1,  10, -21,   1,   1,   2,   1,  -2,  -2, -20,  69,  -1,   0,   0,   1,  -7,  -4,  11, -16,  -1,   0,   1,  -5,  -7,  18, -67, -54,   0,   2,   4,   1,   0,  38,  -9,  12,   1,  -3,  -4,   0,  22,   4,   6,   1 },
      {   0,   0,  -1,   0,  -2,  -2,   5,  -6,   0,   1,   0,   2,  -5,  -2,   5,   1,   0,   0,  -2,  -2,   3,   1,   2, -16,   0,   1,  -1,  -1,  -4,   3,   6,   7,  -1,   0,  -2,   0,   0,   5,  -9, -52,   0,  -1,   5,  -1, -18,  10,   3, -21,   0,  -1,  -2,   1,   3,  -6,  54,   7,  -3,  -2,  -2, -16, -93, -11,  13,   1 },
      {   0,   0,   1,  -2,  -1,  -3,  -1,   1,   0,  -1,  -1,  -2,   1,   3,  -9,   3,   0,   0,   1,   0,  -5,  -6,   7, -20,   0,  -1,  -3,  -1,   4,  -6, -18,  -5,   0,   0,   1,   0,  -6, -11,  17, -36,  -1,  -2,  -2,  -7,   5, -17, -59,  68,   0,  -2,  -4,  -1,   1, -39, -19,  -1,   1,   0,   1,  -4, -13,  40,   7,  38 },
      {   0,   0,   1,   1,   1,  -5,   1,  -1,   0,   0,   0,   0,   2,   5,   3,  -7,   0,   0,   0,   1,   2,  -7,  -2,   2,   0,   0,  -1,   2,   1,  10,   5,  -2,  -1,  -1,  -2,   2,   2, -16,  -4, -26,   0,  -1,  -3,   1,   0,  24,   4, -50,  -1,  -3,  -5,   2, -16,-101, -11,  12,   3,   2,   0,  11,  28,   2,  11, -18 },
      {   0,   0,  -1,   0,   0,  -2,   1,   3,   0,   0,   1,  -1,  -3,   0,  -6,  -3,   0,   0,   0,  -1,  -3,  -3,   8,  17,   0,   0,   0,   0,  -2,   0,  -2,  29,   0,   1,  -3,   2,  -4, -14,  28,  82,  -1,  -1,  -3,  -6,  -4, -14, -32,   0,  -2,  -1,  -4,  -4, -18, -35,  -4, -28,  -3,  -1,  -5,  -7, -48, -31,  10, -15 },
      {   0,  -1,   0,   0,  -1,  -2,  -3,  -2,   0,   0,   0,   1,   1,  -1,  -4,   7,   0,   0,   0,  -2,  -3,  -1,  -2,  -1,   0,   1,   0,   1,   3,  -2,   8,  22,  -1,   0,  -1,  -2,  -2,   3,   6, -45,   0,  -1,   0,   2,  -3,  -9,  -3,  47,   0,   0,  -2,  -1,  -9,   6, -32,   6,   0,   0,  -2,  19,  -1, -53,  -4, -84 },
      {   0,   0,   0,   0,   2,   0,  -7,   2,   0,   0,  -1,  -1,  -2,  -3,   2,   4,   0,   0,   0,  -3,   6,   3, -16,  -4,  -1,   0,  -1,  -1,  -4,   0,   5, -10,  -1,  -1,   1,  -3,   5,   1, -27, -19,   0,  -2,   0,   1,  -5,   0,   6, -26,  -2,  -2,  -4,  -4,   7,   6, -89, -34,  -1,  -1,  -5,  -4, -29, -39,  -3,  47 },
      {   0,   0,   0,  -2,   1,  -1,   1,   0,   0,   1,   0,   2,   0,  -4,  -1,   2,   0,   0,   0,  -1,   0,   3,   9,  -6,   0,   1,   1,   1,   4,  -4,   2,   3,   0,   1,   2,   0,   2,   7,  15, -12,   0,  -1,   2,   7,   3, -17,  -9,  11,   1,   2,   1,   8,   4, -15,  46,   1,   1,  -2,   2,   6,  41, -89,   1,  55 },
      {   0,   0,  -1,  -1,   1,   0,   1,   0,   0,   0,  -1,   1,   0,  -4,  -1,  -5,   0,   0,   0,  -1,   1,  -1,   1,  -7,   0,   0,   3,   0,  -1,   0,  -1,   3,  -1,  -1,  -2,   0,   2,   4,  -7, -21,   0,  -1,   2,   0,   1,   5,  -1,   0,   0,  -1,  -1,   4,   1,   5,  25,-112,   2,   2,   6,   6,  22,  14,  38, -17 },
      {   0,   0,  -1,   0,   2,   0,  -1,  -1,   0,   0,   1,   1,   2,   1,  -5,   3,   0,   0,   0,   1,   1,   0,   0,  -1,   0,   0,   1,  -1,  -2,   5,   1,   0,   0,  -1,   0,   1,  -1,  -2,   8, -11,  -1,   1,  -4,  -2,  -1,  -2, -12, -11,   0,   0,   3,  -5,   4, -12,  15, -35,  -2,  -2,   0,   0,  -3,   5,-119,  -5 },
    },
    {
      {  33,  37,   2,  -8, -24,  -5,   2,   4,  -9,  -5, -28,   1,  13,   9,  -2,  -2, -10, -53,  36,  22,  12,   2,  -1,   2,  55,  35, -16, -17, -14, -11,  -3,  -3, -30,  13, -15,  -5,   4,   3,   5,   1,  13, -14,   8,   6,   1,   1,   0,   0,   4,  -7,   2,   3,  -3,   2,  -2,   1,   0,   3,   3,  -3,   3,  -1,   1,   0 },
      {  76, -49,   3,  11,  -1,  15,   2,   8,  14,   9,  -9,  14, -30,   6,  -4,   3, -12,  13, -20, -24,  31,   5,  -8,   1,  19,   4,  -3,  15,   6, -34,   6,  -6,  -8, -22,   2,  -9, -17,  16,   5,   2,  -4,  15,  -4,  -3,   2,  -3,  -5,   0, -14, -18,   8,   4,   2,   4,   2,  -1,  -9,   2,   3,  -1,   2,  -1,  -2,   0 },
      {  38,  45, -47, -24,  -1, -36,  -5, -19,  -3, -18,  14,  -5,  12,  -5,  20,  -6,  13,  -4,   0, -21,  10,   3, -11,   3,   8, -36,  34,  38,  -6,  16,   8,  13,  11,  -6,  -8,  -6, -14,  14,  -1,   4, -24,  11,   7,   2,   8,  -2,  -2,  -2,  10,  -3,  -7,   1,   0,  -4,  -2,  -2, -12,  11,  -6,  -1,  -1,  -1,  -3,   0 },
      {   6,  15,  33, -13,  10,  11, -13,  12, -45,   4, -31,  41, -15,   7,  -2,  -2, -17, -13,  35, -40,  -4,  -4,   7,  -6, -23,  -4, -10,  18,  10,   2,  15,  12,  40,  24,   6,  -4,   8,  -6,  -7,  -4,  -5,  18, -16,  -1,  -6,  -8, -10, -10,  28, -15,  -3,   3,  -6,  10,  -1,   1,  -5,  -4,   4,  -9,   4,  -3,   0,   0 },
      { -37,  17,  16, -27,  12,   7,  -4,  -8,  54,  10, -13,  32,  -5, -14, -12, -38, -15, -14, -28,   1,  28,  -8, -19,  -6,  15, -14,   1,   5, -31,   2,  28,   1, -20,  -4,  21,  -5,  11,   9, -12,  -2,  12,  25,   5,  -7,  -5,  -6,  -2,   1,   9, -11, -12,   8,   3,   5,   7,   0, -13,  -4,   3,  -2,  -1,   1,  -1,   2 },
      {   9,  -2,  24,   4, -12,   2,   6,  -9, -51,   2,  -8,  14,   0,  -2,  -4,  -3,  82,  -1, -19,  -8,  10,   3,  -6,  -4, -11, -22,  13,   2, -24,  -7,  12,  -9, -43,  12,   1,   6,  16,  -3,  -7,   2,  25,   3,   3,  -7,  -9,   3,   5,   2, -16,   1,   0,   3,  -3,   0,   4,   0,   6,   0,  -6,   7,   4,  -3,  -3,  -1 },
      {  41,   0,  17,  -4,  -4,  -4,  -3,   3,  72, -19, -26,  -2,   6,   2,  -4,  -5,  19, -12,  21,   5, -10,  -5,   9,   1, -67,  -4,  17,  -7,  -4,   5, -12,   2,   1,  28, -19,   3,   1, -12,   5,   3,  19,  -5,   0,   0,   3,  -3,   4,   2,  -5,  -2,   2,   1,  -4,   3,   1,   2,  -2,   6,  -6,   2,   2,  -2,   2,   1 },
      { -31,  11,  21,  42,  -2,  -9, -11,  -4,  45,  10, -17, -11, -13,  -1,  10,  20,  40,  -1,  12, -45,  -1,  15,   0,   2,  44,   0,  -6,  36,   5, -11,   0,  -5,  13,   6, -17,   1,   3, -14,  -2,   0, -23, -22,  -2,   4,  -6,   4,   1,  -1,  -2,  -3,   9,   5,  -5,   8,  -8,  -2,   9,  -8,  -2,  -4,   4,   2,  -4,  -1 },
      {   7, -19, -19, -14,  28,  -1,   1,  -5,  -1,  51,  -4,   5,   8, -12, -20, -10,  41, -30,   5,  11,   4, -22, -25,  16, -17,  23, -26,   3,   1,  25,   6,  -4,  24,   0,   1, -13, -17,  24, -11,  -9, -25, -27,  16,   9,   2,  -2, -10,   0, -16,  -1,   2,  -5,   0,  14,  -3,  -6,   1,  -2,  17, -17,  -8,   9,   3,   9 },
      {  -4,  -1,  21, -29, -31,   9,  -7,  18,   5, -14,  -6,  -7, -16,   8,   8, -11,  39,  -3,  -2,  37,  24, -15,  22, -15,  14, -10,  -1, -26,  -3, -19,   1,  18,  48, -18,  13,  -5,  10,  -2,  14,   4, -48,   9,  -6,   4,  14, -11,   1,  -1,  -9,   9,   0,   2,  12,  -9,   1,   6,   4,  11,  -5,   8,   5, -11,  -7,  -3 },
      { -14,  20, -23,   2, -13, -12,  -3,   7,  -1,  18,  25, -25, -39,  -1,   5,  -4,  13, -31, -24,  -9,  29,  26,  -3,   3, -21,  22,   3, -10,   8, -34,  -6,  10,  33,  17,  -4,   0,  -8,  13,   8,  -3,  53,   6, -14,   4,  -6,   3,   3,   6,  17,  -8,   4,  -8,  20,   5,  -8,  -2, -11, -13,   5,   2,   0,  -6,  -2,  -1 },
      { -11, -31,   8,  -4, -16, -29,  -5,  14,  -4,   5, -12, -21, -17,  -3,   2,  -2,   0, -22,   0, -15,  16,  -5,  -2,  10, -12,  -4,  10, -18,  21,  16, -19,   3, -53,   6,  -3,   9,  -4,   1,  -2,  -6, -47,  25,   9,   9,   4,   1,   4,   8,  53,  12,   1, -12,  12,  13, -11,  -3, -13, -19,   5,   1,  -2,  -1,  -4,   1 },
      {  11, -30,  23,   4, -18,   5,  20, -22,  -5,  27, -10, -16,  27,  -5,   5,   8, -13,  -2, -17,  17,  -1,  -2, -35,  -7,  18, -20,  38, -13,  -9,   5,   2,  -7,  41,  13, -26, -13,  19, -10,  -7,   5,   7, -18,   7, -12,  10,  19, -11,  -6,  41, -19,  -4,   3,  -3,  -1,   6,   3,  -9, -17, -10,   8,  -3,  -3,  -1,  -2 },
      { -25, -10,  14, -19, -33,  13,  46,   2,   1, -23,  24,  11,   7, -12,  -3,  32, -15, -39,  13, -27,  13,   9, -17,   7, -30,  17,  13,  13, -11, -12,   6, -11,  -1, -26, -13, -14,   0,  21,  -3,   8, -20, -10,  -1, -12, -17,   8,   0,  -7, -17,  16,   3,  -1, -10, -10,   8,   1,  -4,  15, -13,  -5,  14,   1,   3,   3 },
      {   1,   0,  11,  19,   1, -44, -11, -15, -20, -16, -24,   6,  -6,  15,   7, -30, -38,  -5, -27,  -2,  26,  10,  -7,  -3, -20,  12,  10,   0, -23,  -5,   6, -14,  20,   9,   4,  20,   2, -15,  -1,  -3, -22, -33,  23,   1,  -5,   2,  16,  13, -39,  26,   7,   6,   2,  18,  -7,  -1,   6, -15,   5,   9,   0,  -2,  -7,  -4 },
      { -29, -28,  -2, -38,  -8, -31,   5,  12,  -5, -17, -20,  17,  -7,  -2, -18,  -7,   2,  13,   7, -16,  -9,   0,  -1,  15,  32, -13,  19, -21,  19,  11, -16,   9,   9,  11, -24,  15, -19,   6,  -1,   0,  23,  -8,  -3,  -4,   0,   2,  -3,   5, -38, -31,  18, -16,   5,  17,  -1,  -2, -27,  29,   2,   4,  -3,  -2,   1,   9 },
      { -34, -24, -23,   8,  35,  19, -16,  -5,  -9,  -4, -27,  -7,  -5,  43,  18,   4,  -3, -32,  19,  28,  -4, -13,  -4,   3,  -7, -19,  36,  27,  -2, -37,   3,  -2, -15,   7, -12, -25, -13,   8,  18,  -9,  -1,  10,  -9,   9,  12,   4,  -6,  -3, -19,  -3,  -7,   9,   0,  -4,   0,  -1,  -9,  -9,  -4,  10,  -6,  -3,   7,  -5 },
      {  21,   0,   4, -12,  45,   1,   6,  -9,   8,  13,  34,  26, -33, -11,   4,   5,   0, -27,   4,  -1, -21,  15, -16,   6,  25,   2,  30, -32,  18,   1,   0,   0,   3,  19, -12,   6,  25, -23,  -5, -11, -14,  23, -13,  -4, -10,   9,   9,   2, -20,  35,  -8,  -1,   4,  -1,  -2,  -2,  23,   3, -17,  11,   2,   1, -11,   3 },
      {  -6,   3,   1,  -7,  28, -25,   4,  -5,   0, -19,   5,  46,   1,  -5, -18,  13,  17,  29, -17,  28,  -9,  15,  15, -15,  -5,  41,   4,   0, -12, -29,  18,  -6,  -1,   0, -49,  -5, -20,   0,  13,   2, -20,   3,  12,  13,  -9,   3,  -3,   3,  38, -12,   2,  -5,   1,  -7,   2,   3, -10, -12,  -1,  -6,   1,   0,   3,  -6 },
      {  -2,  14,  -2, -24,  -8,  12,   2,  24,  -1,  18, -37,   3,   8,   9, -19, -20,   6,  25,  -4, -11,  -6,  19,  -4,   9,   8,  11,  32,   7,   4,  -3, -33, -12,   8, -17,  -4,  -6, -22,  26,  12,   5,  12,   4,   0, -18,  -1,  -9,   3,   0,  14,  47, -12,   9,  -9,   9,   4,  -8,  45, -21, -13,   2,  -6,   8,  17,   8 },
      {   1,  12,  14,   8,  18, -17,  33, -10,   7, -19,  -8,   3,  20,  23,   7,  -2,  18,   4,  14,  12,  13,  16, -40,   0,  -1,  -6, -18, -15,  37, -12,   5,  -8,   3,   0,  38,  14, -43, -10,   4, -13,  -5,   0, -24, -40, -12,  18,   1,  -1,  12,   3,  -6,  16,  10,  -7,  10,   9, -16,   2,  11,   4, -14,   3,   1,  -2 },
      {  15,  14,  23,  -6,  26,  11,  -5,  10,  -5, -10,  10, -17,   0,   6, -22, -12,   8, -18, -11,  -5, -29,  16,  23,  13,   7,  -3,  10,   3,  -3,   3, -17, -10,   4, -41,   8, -12,  22,  -3,  -1,   0,  -3, -22,  -4,  -3, -13,  -6,  10,   9,   0,   5,   0,   5,   2,  12,   1,   0, -75, -27,   2,   6, -10,   6,  17,   2 },
      {   4,  10,  42, -18,  10,  -8,   8,   2, -11,  19,   7, -42,  -1, -25,  -9, -16, -15,   8, -10,  10,   4,   7,  11, -10,   0,  12,   3,  37,   3,  -6,   6,  -3, -18,  38,  -7, -20, -22,  -7,  -4,  -6, -16,  -9, -36,  -6,  16, -16,  -6,  13, -16,  -4,  -9, -20,  23, -12,  11,  17,   4,  22, -20,  -5,  -5,  11,  18,   0 },
      {  -2, -15,  42,   4,   9, -31,   0,   4,   0, -12,  34,   9,   7,  12,  13,  -2,   5,  11,  12,   4,  20, -20,  -9,  10,  10,  12,   4,  19,  18,  13, -13,  -5,   6,   4,  10, -35,  16,  18,  35,  -4,  32,   2,  24,  27,   4, -13,  10,  -1,   8,  12, -30,  22,  -3,   8, -20,   5,  -1,  21,   6,   0,   0,  21, -13,   3 },
      {   7, -26,  -2,  11,   5,  -7,  28,  11,   7, -42,  21,  -2,  -7,  24, -17, -10,   7,  -5,   5,   1,   8, -16,  16,   1,  24,  -7,  -2,  12, -36,  29,   9,  11,  13,  23,   5,  -7, -12,  15, -22, -17,  14,  -2, -35,   8,  -4,   5, -13,  -1,  11,  28,  14, -25,  -7,   2,   5,  10,  14, -38,  11,  -7,   3, -12,   6,  -3 },
      {  -3, -14,  -4,   7,   1,   2,  12, -29,  -8,   2, -28, -21, -13, -31,  12,   5,   8,  14,  26,  12,   9, -20,  -5,  11,   8,  45,   9,  19, -18,  18,  -2,  10,  17, -12,  -5,  32,  -6, -15,  -3,   7,  15,  48,  -3,  -7,  -9, -16,  13,   8, -10,  20, -15,   7,  -8, -12,   7,   5, -36,  -8,  -9,  -3,   5,  -5,  -5,  -2 },
      {  -3, -12, -24,  32, -20,  22,  -5,  13,   3,   7,   5,  26,  17,  18,  -3, -38,  10,  -2,  -6, -23,  -6,   6,   6,   2,   3,  39,  23,   2,  -8,   7,   2,  14,   1,   8,   5,   5,   7, -13,  -5, -17, -10,  -9,  -8,   2,  18,  19,  -8,   7,  17,  15, -25,  -4,  13, -10,  10,  13, -22,  56,   4,  17, -12,  -2,   8,  -9 },
      {  -1, -26, -10,   9,  10,  16, -33,  21,  -6, -44, -18,  16,  13, -59,  13,  10,   5, -35, -24,   8,  13,  19,   9, -22,  10,  -4,  -2,   7,  24,  18, -10,  -2,   9,  10,   8,  -7, -20,  -2, -13,  10,  12, -16,   5,  -7,  -1,   1,   1,   5,   3,   5, -22,  11,   4, -23,  -3,  -1,   4,  -8,  -3,  -3,   2,  10,   0,  10 },
      {  -1, -45, -15, -34,   4, -32,   6,   8,   7,  25,  11,  -3,  20,  17,  13,  -5,   5, -20,  20,  -7, -11,  42,  36, -16,   5,   1, -18,  17, -23,  -3,  11, -19,   8,  -1,  25,  16,   7, -19,  11,  -4,   7,  12,  19, -24,   5,  -4,  12,  10,   7, -14,  -4,   2, -10, -13,  -9,   9,  11,  -5, -15,   9,  11,   2,  11,   2 },
      {  -3,   8,  -4,  51, -15, -24,   4,  14,  -3,  29,   1,  29,  11,  -4, -20,  -7,  -6,  -7,  20,  34,  27,  26,  11,  -4,  -9, -26,   3,   0,  -4,  26, -19,  -4,   6, -29, -15, -25,  -4, -10,  -2,  -1,  -4,  29, -30,   7, -18,   1,  10,   7, -14, -15,   3,  -7,   5,   8,   5,   0,   1,   3, -16,   1,   9,  12,   2,  13 },
      {  -3,  -5,   4,  19,  41, -36,   3,   0,  -6,  -5, -17, -11, -16, -22,   1,  -8,   0,   1,  27, -24,   0,   3,  -4,  -3,  -7,  -6,  -5, -37, -30, -13,  -1,  10,  -3, -40,  -3, -17,   8,  18, -11,  17,  13, -18,  -9, -10,  47,  -5,  -1, -23,  11,   3, -13,  -5,  15, -22,   9,   6,  12,  14,  -3,  -2,   3, -11,  -1,  -7 },
      {   1,  16,  16,   6,  18, -10,  -7,   9,   3,   1, -20,  -9,  -3,  40, -12,  53,   2,  -8, -34,   5,   4,   8,   6,   2,   2,  27,  25,  -6,  -6,  34,  -3, -12,   8,   5,  34,  -6,  -4,  15, -26,  16,  -3,  24,  10,  -3,  10,   5,   6,  -7,  -4, -17,   6, -22, -10, -19,  -8, -23,   2,  27,  -6,  -2,  -2,  -6,   4, -10 },
      {   0, -14,   3,  -7,   7,  -8, -11, -41,  -6, -26, -13, -19,  12,  -8, -11, -32,   3, -18, -23,   4, -11,   0,  10,  28,   0,   8, -18,  17,   9, -13, -23, -11,   9, -18,  -4, -22,  14, -15,  -6, -15,   4,  25, -12,  -3, -20,  19,  -7, -17,  20,   0,  37,   9,  -5,   7,   4, -15,  26,  39,   4,  -9,  11,  -8,  -5,  -6 },
      {   5,   8,  23,  -3,   6,  -9,   2,  -7,   4,  41,  -7,  20,   8,  -9,  40,  12,  -8,  -9, -13,   3,  -1, -25,  46,   8,   3, -19,   9,  -1,  -1, -12, -12,  11,  -3, -12,  -1,  12, -31,  17,  -4, -27,  13, -20,  -3,  -3, -15,  14,   4,   7,   9,  26,   9, -19, -18, -23,  11,   2, -11,  12,  26,  11,  25,   3, -20,  -5 },
      {   0,  -6, -18,   3,  14,   8,   5,   6,   0,   5,  -9,   0,  -8,  -5, -22,  17,  -6,  10,  10,  11,  20,   5,   0,  18,   5, -31, -17,   2, -15, -12,  10, -17,   7,  32,   8,  13,   5,   9,   0,  15,   1, -17,   3,  18,  -9, -21,  16,   2,  32,  40,  38,  18,   5,  -9,   1, -13, -23,  39, -32,  18,  -5,  11,   4,  17 },
      {   3,  -6,  14,   5, -10,  -8, -59,  -2,  -2,  -1,  19, -11,   9, -11, -30,  31,  -5,   0,   5,   1,  -9,   5, -11,   5,   0,  -2,   1, -11, -48,  -3,   0,  17,   1,   7,  -1,  17, -13,  20,  28, -36,  -8,  10, -16, -37,  -4,  10, -23,  -8,  -4,   5,  -9,  23,   2,  11, -14,  -8,  -4,   6,   5,   2,   0,  18,  -6,   6 },
      {   0,   6,  19,  27,  15,  28,   7,   4,  -7, -14,  19, -20, -28,   5,  -3, -26,   0, -15,  14,   8,  -3,   0,   9,  -1,   3,  -8,   6,   1,   0,  17,  14,   4,   7, -11, -18,  17, -34,  16,   0, -11,   0,   0,  60, -28, -18, -14,   5,   7,  16, -25,  -1, -10,   2,  17,  14,   2,  20,  19, -18,  24,   3,  -8,  -3,   4 },
      {  -7,  16,  -7,  -3,  -4,  14,  -2,  -5,  -4,  -4,  19,  22, -30,  -8,   5,  -5,   3,  11,   8,  -4,  15, -18,   2,  16,  -2,   5,  -3,  -2, -15,  13, -40, -50,   0,  17,   3, -25, -13, -21,   7, -16,  -7,  -2,   6, -34,  50,  10,  -2,  22,   2, -14,  22,  16, -14, -16,  -1,  -2, -11, -13,  -1,   3,  15, -13,  -2,   4 },
      {  -3,   4,  13,   8,  29,  22,  26,   8,  -4,  -2,   4,  -7,  27,   6,  29, -13,   1,   2,  -9, -11,  20,  37,  -1,  -3,  -1,  -4, -11, -14, -17,   2, -31,  36,   2,  20, -23,   7,   8,   7,  10,  11, -11,  30,  21,   2,  13,  -7, -30,   6, -14,   5,  29,   8, -13,  -9,  -5,  -5,  -9,   3,  26, -20,  -5,  21,  13,  11 },
      {  -1,  -2,  -4,   2,  14,  13,   3, -17,  -1,  -4,   3,  -8,  11, -17,  15, -16,   4,  11,  15,  -8,   7,   1,  12, -19,   0,  29,  43, -19,  -5,   1,  -4,  -3,  -6,   7,  36,   9,   3,  33,   5,  -4, -16, -15, -32,  22, -15,   3,  22,  -5,   8, -41,  27,  20, -12,  13,  -7,  10,  11,  -6,  -6,  10,  14,  10,  -5,  36 },
      {  -3,  -6,  14,  -2,  -1,  14,   0, -53,   0,   6,  19,  22,  21,  18, -26,  10,  -2, -26, -10, -10,   6,   2,  18,  12,   2,  -3,  -5,   7,   3,   6, -24,  21,  -2,  -4,  -7,  25, -28,  -5,   4,  38,   1,  -5, -18,   8,  31, -33,  18,  -2,   3,  -5,  -4,  17,  24,   3,  -6,   3,  13, -11,   3,  19, -15, -18, -15,  -2 },
      {   0,  -1, -21, -13,  16,   3,  20,  10,   0,   4, -15,  -8,  -2,   0,   1,  19,  -3,  17,  -4,  -5,  16,  -6,   7,  -2,  -3,   6, -11,   1, -17,   8, -21,  34,  -9,  11, -10, -42,  16,  -7, -12,  -2,  -9, -15,  -5, -34, -46,  -6,  14,  19,   1, -11, -16,  24,  10,   7, -27,   9,   8,   8,  16,  37,  -9, -26, -14, -17 },
      {  -3,  -3,  -2,   0,  -6,  11,   3, -20,   0,   8,   2,  -9,  -7, -17, -38,   4,  -2,  10,  19,   8,  10,  41,  12,   1,   3, -13,  26, -11,   8, -10,  11,  19,  -4,  13,  28, -24, -23, -17,  -7,  -7,  -1,  -6,  27,  16,   9,   1,   7,  -8,  -2,  15, -10,  10, -34,  18,  -6,  14,  -9,   9,  26, -32,  25, -30,   3, -24 },
      {  -2, -13,   9, -28,  -7,  12, -34, -13,   2,   4,  10,   2, -28,  33,  35, -10,   3,  16,  11,  20,   9,  42, -24, -21,  -5,  10, -14,   9,  -3,  33,  -9,  12, -14, -12, -23,  -2,  -5,   5, -28,  -4,   6, -29, -20,   2,  -2,  -3,  12, -19,   9,   8,   5,   6, -11,   0,  16, -18, -12,   7, -10, -10,   9,   1, -15,  -8 },
      {  -2,  -1,  -2,  -7,   9,   5, -20,  19,   0,   4,  20, -18,  45,  10,   2,  -5,   3,   6,   6, -34,  28, -25,  13,  -2,   4,   9,   7, -34,   2,  -7,  22, -20,   4,   5, -23, -20, -28, -30, -27,  -5,   4,  16,   3,   4, -11, -16,  19, -25, -10,   5,   4,  17,   9,   5,  20, -24,  -7, -12,  -7,   4,  11,  14,  -2, -21 },
      {   2,   7,  10, -20,  14,  15,  -6,  13,   0,  -4,   1,  -2,  11, -13,  -4,  17,  -1,  14,  23,  14,  44,  15,  14,  22,  -6,   5,  11,  26,   8,   0,  13,  -2,  -2, -22,  -9,  22,  18, -11, -25, -23,  11,   0,  11,   5,  27,  46,   3, -17,  -1,   2,  14, -11,  23,  14, -16,  12,  12,  -7,  17,  23, -31,  -2,  -1,  13 },
      {  -1,  -2,   4,   4, -16, -10, -18, -24,   5,   5,   1,   4,   1,  -2,  16,   7,   1,   6,  16,   0, -18,  12,   8, -21,   0,  13,   5,  -6,   9,   5,  20, -16,  -2,   3,  10, -40,  -4,   9, -13,  26,   7,  15,  13, -12,  -5, -16, -45,   5,  -6,  29,  27,  -7,  26,  16,  14,   8,  -9,  -2,  28,  37,  -5, -17,  -1,  37 },
      {   2,  -3,  11,  -7,  -6,  -2, -18, -23,   1,   0,   9,   3,  -6,   9,   4,   0,  -1,  10,  18, -10,   8,   1,   0,  17,   0,  -9,  -2, -16, -15, -17, -33,  -9,   5,  -1,  10,   4, -11,   2, -38, -10,  -3,  -1,  20,  38, -40,   9, -30,  21,  -3, -13, -27,   8,  16, -41,   0,  11,   8,  -6, -20, -12, -24, -12,  33,   9 },
      {   2,  -2,  10,  -1, -11,   6,  15,  18,  -1,  -7,  -8,   6,  11,  -9,  16,   6,   0,   3,   2,   2, -43,  16, -21,  17,  -1,   3,  -2,   6, -26, -13,   5,  25,   5,  -5,  22, -21, -22,  -5, -20, -24,   4,  11,  11,  27,  21,   2,  25,  40,   1,  -5,  15,   8,  39,  12,   1, -29,   4,  -7, -12,  -5,   3,   4, -36,   2 },
      {  -1,   6,  -3,   6,  -2, -20,   8,   6,   2,   0,  -3,   3,  11,  -2,  -5,   8,   0,   0, -22,   3,  -1,   6,  -1,  12,  -4,  10, -22,   5,  20, -10, -13,  25,  -3,  11,   3,  -1,  11,  13, -19, -43,  -4,   5,  10,  11,  23, -18, -18, -32, -11,  -4, -11,  16, -34,   7,  31,  10, -11, -17, -42,  41,  13, -25,  -2,  21 },
      {  -1,  -3,  -9,   4, -10,   5,  -1, -27,   2,  -7, -15,   9,   0,  -1,  23,  22,   1,   3,  -7,  -3,   3,  18,  24,  25,   2,  -6,  -5, -19,   8,  15,  14, -17,   6,  12, -10,  -9,  25,  40,   6, -35,  -3,   7,  -1, -13,   7, -22,  30,  12,  -3,  -3,  -2,   6,  24,  16,  39,  27,   5,  -2,   5, -25, -21,  11,  22, -28 },
      {  -1,  -1,   7,   4,  -8,   6, -29,  -6,   0,   0,   0,   2,  17,   1, -19, -18,   0,   5,   8, -10,  -6,   7, -32,   5,   0,   2,  -7,   0,  10,  -1,   5,  17,   3,   7,  -3, -21,   3,  16,  14,  -9,  -2,  10,  14,  -6,  -8,  -3,  48,  -8,  -7,   3,  34, -39,  -6, -62, -18,  32,  -5, -19,   2,  19,  -1,  -5,  -2,  -4 },
      {  -1,  -2,   5,   7,   9,  -7, -20,  12,   2,   4,   9,   8,  23,   6,   7,  -2,   1,   1,   3,   8,  17,  -8,  -3,  27,  -2,  11,  20,  -1,  26,   2,  29,  22,  -8, -19,   0,   8,  16,   1,  -4,   3,  10, -17, -10, -25,   2, -24, -32,  48,   5,   3,  31,  12,  -3,  -5, -20,   3,   5,  -4, -35, -31,  17, -32,   5, -12 },
      {   0,  -6,   2,  15, -18, -13,   3,   6,   2,   4,  16,  16,  -8, -12,  11, -25,   0,   7,  15,  24, -12,   2,  24,  42,   1,   6,  11,  11,   5, -22,  -1,  13,  -3,  28,  15,  -4,   6,  21, -29,  37, -13,  -2,   7, -37,  -6,   4,   0, -39,  -6, -10,  -2,   6,  -3,  -9,  -5, -30,  -5, -17,  -3,  -6, -23,  21, -11, -20 },
      {   1,   4,   3,   2,  -1,  13,  20, -22,   2,   5, -14,   1,   7, -10,   4, -15,  -1,   6,   3, -11,   2,  11,  23, -23,   0,  -2, -13, -11,   9,  22,   9, -16,  -2,   5,  -3,  -6,  -8,  27,  29, -19,   2,   3,   1,  10,  -1,  24, -22, -18, -14,  12,  -2,  -5,  22,  -2, -48, -32, -18,  -8, -44,   2, -10, -25, -16, -31 },
      {   0,   1,   0,  -4,  -9, -10,  10,  -5,  -2,   9, -10,  -2, -30,  12,  -7,   3,  -2,  -5, -14,  -7,  -5, -10,  19,   7,  -1,   6,   3,  -3,   5,  32,  26,  41,  -9,  -7,   0, -19, -14, -22,  24,   5,   7,  -8,  14,   1,  -2,  38,   1, -18,  -4,   2,  22,  50,  20, -27,  15,  11,   5, -12, -15,  -9,   2,  24,  25,  21 },
      {  -1,  -1, -12,   0,  -1,   2,   5, -40,   2,  -9,  -4,   1,  -1,  17, -32,   4,   2,   6,  13,  -4,  16, -15,  12, -38,   0,   2,  -1,   3,  23, -21, -13,  25,   1,   1,   2,  -5,  24,   9, -28, -15,   3,  -5,  16, -25,  15,   6,   3,  29,   1,   7, -10, -19,  12,   0,  -4, -27,  -4,  -6, -11,  -4,  31,  52,   2,  12 },
      {  -1,  -2,  -4, -10,  12,  13, -14, -14,  -1,  -8,  -8,  -5,   8,   4,   4,   3,   2,  10,   2, -12,  23,  16,   9,  45,  -1,   3, -13, -14,  -9,   4,  10,  11,  -1,   6,  13, -20,  -1, -15,  38,  28,  -5, -10,  -3,   1, -10,  25, -12,  18,   2,  -5, -26, -48, -16,  13,  36,  -4,   1, -11, -25,   5,   2,  11, -35,  16 },
      {   1,   2,  -1,   2,  13,   4, -10,  14,   0,   7,   6,   2,  16,  13,   2,   0,  -1,   5,  -5,   3,  -6, -11,  -1,  -6,   4,  -2, -10,  11,  -5,  -8, -25,  -4,   1,  18,  -2,   8,   3,  21,  -8,  27,  -4,  10,  14, -15,  -6,  38,  19, -19,  -5,  11,  -6,   2,  43,  24,   7,  41, -14,   2, -19, -38,  45, -32, -11,  18 },
      {   2,  -1,  -3,   0,   2,   6,  11,   4,   1,  -3,  -7,  -1,   1,   4,  -7,  -3,  -1,   9,   2,   6,   5,  18,  -4,  30,   0,  -1,  13,   9,   3,  13,   5, -11,  -2,   6,   1,  14,  18,  -6,  20, -19,  -6, -14,   0,   4, -12, -30, -14, -23,   0, -13, -19,  -5,  47, -34,   8, -37,   3, -10,  36,   5,  59, -14,   4,  14 },
      {  -2,   2,   0,  -6,   4,  -4,  -2,  12,  -1,   2,  -2, -17,  -2,   2, -29,  -9,   0,  -1,   9,  -1,  -4,  13,   1,  -3,  -3,  -2,  -1,   6,  14,  10,   3, -18,   3,   3,  -9,  15,  10,  31, -10,  19,  -4,  10,  -3,  20,   2,  25, -24,  33,  -7,  -6,   3,  38,  -3, -24,  29,  21,   3,  -7,  14,  26,  19,  18, -46, -47 },
      {   1,  -2,   3,  -4,   7,  -2,  -9,   0,   2,   1,   7,  17,   2, -20,  12,  10,   3,   0,   9,   4,   5,  -6, -21,  -6,   2,   5,   1,  10,  -8,   2, -26,  19,  -1,   0,  27,   3,  -2,  -6,  18,  13,   2,  -4,   7,  11,  -5,   9,  -9,  -6,   2,  -3,  25, -17,  17,  27,  20,  -5,   4,  -6, -11,  36,  25,   9,  56, -63 },
      {   2,  -3,  14,   3,  -1,  -6,  -8,  -1,   0,   1,   9,  -6,   7,  -5,   3,  -9,  -2,   4,   5,  -1,  -1,  -1,   1,   2,   1,  -4, -17,  -1,  10,  -1,  21,   7,   4,  10, -17,   1,   3,  27,   2,  17,  -5,  11, -13,   2,  11,  40,  36,  35,  -3,   2, -20,  15, -20,  -1,  12, -52,   0,   2,  18,  17,   6, -42,  47,  18 },
      {   0,   4,   1,  -1,  -7,  11,   1, -10,   0,  -8,  -9,  -2,  -5,   7,  19,  -5,   1,   4, -13,   5, -12,   9,  -2,  39,  -3,  -4, -17, -17,  -4,   3,  24, -10,   3,  -3, -13,  -5, -18,   4, -26,  13,  -2,   8, -14,   6,  12,   1,  -4,   1,   2,   0, -20,   0,  -8,  15, -60,  31,  -1,   4,  -4,  28,  48,  36,  26,   0 },
    },
    {
      { -66,  78, -56,  35, -21,  11,  -7,   4,  -5,  11,  -4,   0,   1,  -2,  -1,  -1,   6, -17,  15, -10,   6,  -1,   2,   1,   4,  -5,   4,  -2,   1,  -1,   0,   0,   2,  -4,   3,  -1,   2,  -1,   1,   0,   1,  -2,   0,  -1,   1,  -1,   1,  -1,   2,  -1,   1,   0,   0,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   0,   0 },
      {  48,  -5, -44,   0,   8,  -3,  -4,  10, -44,  55,  11, -19,   8, -10, -14,  -3, -47,  -1,  14, -24,   6,  18,  -2,  21,   9,  12,  -1, -14,   6, -19,  -2,  -4,   2,  -5,   8,   7,  16,  -8,  11,  -1,   8,   2, -14,   3,   4, -11,  11,  -6,   4,   0,  -1,   9,  -2,  -4,   5,  -5,  -3,  -4,  -5,   8,  -5,   5,   0,   0 },
      {  82,  22, -30,  28, -31,  12,  -6,  -5,   9,   0, -42,  33, -15,  11,   8,   3,   9,  -6,   2,  10,   3, -17,   4, -16, -22, -19,  14,   4,   0,  13,   3,   4,  -4,   2,  -2,  -7, -10,   6,  -7,   0,  -6,  -3,  13,  -3,  -3,   8,  -8,   4,  -6,  -1,   2,  -7,   2,   3,  -3,   4,   2,   3,   5,  -6,   4,  -4,   0,   1 },
      {  26,   6, -28,   6,  -4,   5,  -6,   6,  84, -48,  36, -33,  12,  -9,  -7,  -2, -12, -11,  12, -17,   3,   9,   0,  10, -14,  21,  -8,  -4,   2,  -9,   0,  -3, -17,   4,   5,   2,  10,  -5,   7,  -1,  -7,   8,  -8,   2,   2,  -5,   5,  -2,   0,   1,  -1,   5,  -2,  -2,   3,  -3,  -4,   0,  -3,   5,  -3,   2,   1,   0 },
      { -23,   7,   1,   0,   1,  -2,   1,   0,  46,  18, -36,  20, -17,   8,  -8,   1, -71,  58, -35,  20, -11,   8,  -3,   3, -10,   2,   8, -11,   9,  -8,   3,  -2,   0, -11,  13,  -8,   7,  -4,   3,   0,   0,  -2,   3,  -2,   2,  -1,   1,   0,   4,  -3,   3,   0,   0,  -1,   0,  -1,   0,  -2,   0,   1,  -1,   1,   0,   0 },
      {  -5,  -8, -14,  -8,  19, -11,   5,   3,  25,  55, -26,   1,  -2,  -7, -11,  -2,  72,  -4, -30,   3,  -7,  14,  -6,  13, -41,  19,  -9,  -3,   4, -10,  -2,  -2, -37,   8,  10,   2,  13,  -5,   6,  -1,   6,   0,  -5,  -2,   6,  -8,   8,  -4,   5,  -1,   2,   4,   0,  -3,   3,  -3,  -4,  -4,  -2,   4,  -3,   3,   0,   0 },
      {   9,  22, -25,   9,   7, -11,   7,  -4, -33, -24,  13, -10,  25, -15,  16,  -2,  -5,  37, -40,  28,  -5,  -4,  -3,  -3, -29,  49, -43,  28, -12,  11,  -2,   2,  15,  26, -20,   5,  -6,   4,  -5,  -2,  13,  -6,   0,   0,  -2,   1,  -2,   1,  -9,  -4,   8,  -7,   4,   1,  -1,   2,   0,  -2,   5,  -5,   2,   0,  -2,   2 },
      {  21,  15, -15,  -1,   0,  -3,   0,   1,  15,   2,  -3,  -8,   2,  -2,  -1,   0,  35,  32, -33,   5,  -9,   4,  -3,   3, 103,  -7, -20,  -1,  -4,  -4,  -1,  -1,  -1, -11,   0,   1,   2,  -2,   2,   0, -19,   0,   3,  -2,   2,  -2,   1,   0,  -9,  -1,   4,   1,   1,   0,   0,  -1,  -9,  -1,   2,   1,  -1,   1,   0,   0 },
      {  19,  63,  12, -52,  33, -31,  19, -14,   4,  10,   8, -31,  23, -14,  13,  -1,  -8,  -3,   0,  14, -14,   7,  -8,  -3, -16, -45,  22,   8,  -6,  13,  -1,   3, -12, -18,   5,   2,  -9,   6,  -4,   1,  -6,  -3,   5,   2,  -2,   4,  -5,   2,  -1,   0,  -1,   0,   0,   3,  -2,   2,   1,   0,   1,   0,   1,  -1,   0,   1 },
      {   2,  -2,   2,  16, -24,  20, -13,   4, -43, -30,  37,  -9,  -2,   7,  -4,   2,   9,  34, -25,   8, -12,   7,  -4,  -1, -17, -19,  15, -11,   9,  -5,   2,  -2, -68, -27,  35, -14,  10,  -5,   5,   1,  -7,   4,   9,   0,   0,   1,  -1,   1,  15,   2,  -3,   4,  -3,   1,   0,  -1,  10,   1,  -4,   3,  -2,   0,   1,  -1 },
      {  11,   9,   0,  -8,  -5,  13, -12,   7,   3,  -4,  23,  -8, -11,  12, -10,   3,  35,  13, -19,   4,  -5,   7,   0,   2, -23,   1,  14, -18,  18,  -9,   4,  -3,  88, -46,  21, -15,   6,  -4,   3,   1,  22,  -7, -10,   8,  -5,   2,   1,   1,  -5,  -3,  -5,   6,  -2,   0,   1,   0,   1,  -5,  -2,   4,  -2,   1,   2,  -1 },
      {  -5, -21,  12,  24, -29,  20,  -9,   0,  22,  42,   4, -36,  32, -29,  21, -10,   5,  42,   0, -28,  30, -25,  10,  -9, -10, -31,  -7,  21, -16,  17,  -4,   6,   4, -17,  -7,  13, -15,   8,  -5,   1,  13,   6,  -9,   5,  -5,   7,  -2,   2,  12,  -2,  -7,  -1,   0,   2,  -2,   2,  -1,  -3,   2,  -2,   2,  -2,  -1,   1 },
      {   7,  10,  -3,  -6,   7,  -8,   5,   1,   5, -18,  -5,  10,  -4,   2,   1,   2,   4,   7,   3,   0,   5,  -8,   2,  -3,  23,  -6,  12, -11,   8,  -3,   1,  -1, -21,  17,  -3,   2,  -2,   3,  -3,  -1, 111,   5, -18,   5,  -8,   1,  -5,   0,  22, -11, -11,   4,  -3,   2,  -3,   1,  -5,  -5,   0,   0,   1,   0,  -1,  -1 },
      { -14, -37, -36,  25,  11, -17,  14,  -3,  -4, -20,  -2,  -1,  21, -19,   6,  -3,   4,   5, -20,   8,  -5,  12,  -7,   8,  -8, -58,  38, -21,  14, -10,   2,   2,  12,  33, -14,   4,   3,  -4,   1,  -1,  -3,   2, -16,  10,  -6,   0,   2,  -3, -49,   2,  11,   3,   0,  -2,   3,  -5, -15,   6,   0,   0,   1,   4,  -2,   0 },
      {  -3, -26, -21,  25,   1,  -4,   6,   0,   7,  -4,  -4,   6,   7, -11,   0,  -3,  -9, -35,  -5,  45, -30,  30, -19,  10,   6, -24, -14,  33, -26,  25, -14,   7,  16, -25,  12,  -4,   5,  -2,   1,  -2,   9,  24,  -8,  -4,   5,  -4,   5,  -1,  52,  16, -14,  -4,   3,  -7,   3,  -1,   8,  -3,  -2,  -1,   0,  -3,   3,  -1 },
      {   5,  17,  30,  -7, -34,  38, -27,  21,  10,  31,  34, -14, -19,  12, -12,   5, -11, -25, -24,  28, -22,  17,  -6,   3,   0, -27, -12,  16,   0,  -5,   3,  -3,  -1,  43, -25,  13,   3,  -3,   5,  -2,  17,  -6,  -3,  -6,   8,  -7,   6,   1, -27,  -4,   8,  -2,   5,  -4,   4,  -3,   6,  -1,   0,  -1,   0,   1,   1,   1 },
      {  -6, -15,  -6,   3,   2,  -1,   2,  -2,   4,   5,  -2,   2,   1,   5,  -4,   1,  -5,  -1,  27,  -1, -10,   6,  -9,  -1,   2,   5, -19,  27, -26,  22, -13,   3, -16, -38,  33, -17,   8,  -4,   0,   1,  35, -27,  15,  -8,   8,  -1,  -1,   0, -83,   1,   1,   1,   3,   2,   3,   1, -12,   9,   3,  -1,   1,  -5,   3,  -2 },
      {  -3,  -2,   3,  -8,  11,  13, -22,  18,  -8, -20, -38, -14,  29,  -5,  -5,   1, -19, -56, -60,   1,  38, -20,  16,  -4,   4,  -6, -11,  15,  13, -21,   4,  -3,  -5, -23,  21,  11,  -3,  -3,   1,   2,   5,  -9,   9,   1, -11,   2,   7,  -1,  -6,  -6,   1,  10, -11,   5,   2,  -2,  -6,   0,   7,   5,  -3,   3,   1,   0 },
      {   1,  -2,  13,   7, -19,   9,   3,  -7,  -3,   8,  -2, -21,  10, -13,  17,  -8,  -4, -13,  -1,   8, -22,   3,  -2,  -3,   4,  20,  24, -16,  -6,   6,  -3,   2,   7,  20,  20, -22,   9,  -3,  -3,  -1,  -1, -17,  22, -11,   7,   5,  -6,   4,  23, -31,   4,  -4,  -2,   2,   0,  -2, -90,  -4,  15,  -5,   4,  -3,   2,  -3 },
      {   4,  21,  24,   8,  -3,  -8,  -8,   3,  -9,  -4,  -3,  21,  -6,   2,  -8,   6,  -2,   0,   2,   0,  -1,  -2,   2,  10, -10,   3, -13,   7,   0,  -1,  -1,   1, -11, -32, -20,   8,   6,   0,   5,   0,  -9,  72, -53,  17,  -9,  -2,   2,  -4, -24, -10,  -3,   1,   6,  -1,   2,  -1, -55,  -1,   6,  -5,   3,   1,  -1,   1 },
      { -11, -18, -46, -47,  19,  31, -27,  22,  -8, -14, -20, -22,  -7,  40, -28,  17,   7,  27,  20,   2,  -8,  -1,   2, -13,  -9, -15,   4,  18, -24,   5,   1,  -1,   0,  12,  -5,  19, -18,  -4,  -6,  -2,  -7,   8,   5,   7,  -8,   5,   3,   3,  11,  -1,  -7,   0,  -2,  -1,   1,   2, -29, -10,  13,  -5,   6,  -6,   2,  -1 },
      {  -2,  -9, -17,  11,  16, -22,  11,   0,  11,  15,  53,  35, -22,   3,  -9,   3,  -4,  -4,   2,  20,  17, -20,   5,  -8,  -6, -22, -25,  13,  13, -20,  12,   0, -11, -25, -20,  18, -10,   4,  -1,   2,  -7, -51,   4,   5, -16,   9,   3,   2,  18, -10,  -4,   5, -11,   9,  -5,   3, -37,  -3,   4,  -4,   3,   2,   1,   0 },
      {  -9, -17, -31, -33,  -7,   9,   6,   6,   8,  27,  22,  -5,  -3, -10,  25,  -8,  -5, -13,   9,  48,   9, -32,  16, -28,  10,  12,  -6,  -8,  19,  -2,  13,  -2,  -6,   9,  23, -43,   8,   2,  -9,  -1,  -6,  38, -15,  11, -18,  17, -17,   5, -13,  -2,   3,  -6,  -2,   3,  -5,   5,  16,   6,   7, -10,   5,  -1,  -3,   4 },
      {   1,  -5,   4,  20,  12,  12, -25,  14,   6,  10,  -2,  -2,  22,  -7,   6,   2,   2,  -4,  13,  35, -28,   2,   0,  -2,  13,  45,  60,   4,  -6,  -4,  -1,  -3, -13, -27, -16,  36, -21,   9,   1,   3,  -5, -22, -20,  24, -17,   8,  -3,   0,  -8, -25,   4,  -1,   0,   6,  -2,   2,  21,   7,  -4,  -4,  -1,   3,  -3,  -1 },
      {   0,  -7,   5,  21,  -4, -16,  20, -13,  -2,  -1, -20, -33,   2,  26,  -1,  12,   2,  -1,  16, -12, -22,  14,  -8, -10,  -2, -14, -23,  28,  14, -31,  36, -23,   3,  -6,  -3,  -5,  15, -19,  16,  -3,   4,   9,  11,  14, -16,  15, -17,   6,   7, -58,  26,  -8,  -3,   7,  -6,   4,  19,   5,   4,  -9,   4,   0,  -7,  -1 },
      {  -7, -16, -24,  -5, -16,  -8,  32, -31,   1,   7,  17, -22, -44,  28,  12,   1, -11, -24, -45, -29,  -8,   2,   4,  -3,   1,  14,  16,  -7,  -7,  20,  -6,   4,  -6, -18,   1,  28, -34,  29, -17,   9,   5,  18,   4,   3,   1,  -4,  -5,   5,  -8, -13,  10,  -8,  17,  -9,   0,   0,   2, -12,  -2,   1,   4,  -3,  -2,  -1 },
      {  -5,  -9, -15, -48, -29,  16,  -8,   7,  -4, -20, -18,  18,  -3, -33,  22, -25,   2,  -1,  12,   1,  -1,  17,  -5,  12,  -6, -12, -29, -29,  12,  16, -13,   8,  -7, -28, -26,  11,   9,   5,   9,  -5,   1, -21,  -9,   1,   7,  -2,   9,  -1,   6, -45,  34, -12,  12, -11,   3,  -3,   8,  -5,   0,   2,  -4,   2,  -2,   2 },
      {  -3,  -1,  -3, -22, -46,   2,  25, -14,  -7, -10, -26, -15, -26, -15,  30, -12,   1,   0,   3,   3, -16,  11,   3,   4,  -1,   8,  -2,  17,   0, -25,  15,  -8,  -2,   0,   1,  19,  -2, -17,  15, -11,  -4, -23, -22,  22, -12,   5,   0,  -1,  -7,  39, -45,  32, -15,  18, -11,   4, -10,  -1,  -8,   4,   2,   2,  -5,   1 },
      {   3,   9,   8,  -1,   1,  -8,   2,   3,  -1,  -4,   7,  17, -25, -11, -12,  -4,   0,   1, -11, -13,  19,   6,  11,   7,   2,  -7,   3,  11, -51,  15,  -7,   6,   5,  34,  56,  -3,  -2, -13,   8,  -8, -13, -20, -40,  43, -23,  14,  11,   2,   7, -30,  16,  -5,  -1,  -5,   5,  -4,  10,   5,   4,   2,   1,   2,   2,   1 },
      {   0,   3,   5,   8,  12,  22,  -5, -11,   4,  10,  -1,   3,  28,  25,  -2, -12,  -5, -12, -10, -10,   4,  38, -27,  -2,  -2,   1, -34, -54, -10,  18,  16,  -3,  -9,  -3, -15, -19, -25,   5,  -4,  -6,   2, -10,  -1,  34, -30,  33, -17,  13,   0,  22, -14,  -1,  -2,  -3,  -1,   7,  -6,  23,   1,  -3,   6,  -4,   3,   4 },
      {  -1,   1, -10, -19,   4,  30,   8, -29,   3,   3,  11,  25,  18,  23,  12, -17,   1,   4,  -8, -11,   4,  -4, -12,   3,   1,   1,  11,  33,  13,  -9, -16,   7,   5,   6,   4,   9,  10, -10,  -6,   6,   0,   7,  -4,  -3,   8, -13,  -3,   8,   9, -17,  -5,   7,  -1,  -2,  10,  -1,  -9,  84, -31,  11,  -8,  -3,   2,  -4 },
      {   1,   1,   0,   3,  14,  30,   7, -37,   2,   3,   5,  16,  20,  24,   8, -18,  -1,  -1,   1,  -2,   4,   2, -19,  10,   3,  -2,   1,  15,  21,   6, -28,  11,  -2,  11,  -2, -18,  -2,  -4,  -1,   2, -11,   0, -12,   1,  -4,  -1,   6,   0,  -9, -36, -29,  31, -18,  16,   4,   1,  13, -70,  14,   9,  -9,  -7,   2,  -5 },
      {   0,   2,   3, -12, -19,   7,  11,  -8,   4,   4,  14,  38,  26, -13,  -3, -10,   5,   7,  13,  13,  22,  29, -15,   2,   1,   6,   3,   0,  -3, -21,  21, -17,   8,   7,  40,  48, -35,  -2,  12, -11,   8,  36,  43,  -6,  -7,   0,   9,  -2,  -8,   5,  17,  -8,  -6,   4,  -3,   1,  -5, -17,   8,  -3,  -2,   2,  -1,   0 },
      {   2,   3,  -2,  10,  45,  16, -15,  35,   2,   1,  11,  26, -15, -14,  50, -28,  -4,  -3, -10, -41, -52,   0,  28, -22,  -4, -10, -12,  -2, -12,  -8,  -1,  -2,   4,   2,   6,   9,   1, -10,   8,  -6,   2,  18,  23,  -1,  -5,   3,  -6,  -2,  -3,  -2,  -5,  16,  -4,   8,   0,   1,   2,  -6,  -6,  -5,   7,   4,  -1,   3 },
      {   5,   7,  12,  27,  40,  39,  22, -26,  -4,  -6, -12, -28, -18,  14,   9, -40,   3,   5,  19,  33,  12,   2,  31, -11,  -1,  -6, -13, -13,   6, -14,   8,   7,  -3,  -4,  15,  21,  -3, -12,   3,   1,   2,  -8, -22,  -5,  19,  -9,  19,   8,  -3,  14,  10, -31,  13, -26,   0,  -1,  -7,  -6,   2,  10,  -1,   0,   2,   5 },
      {  -2,  -5,  -9, -11,  -9,  -7, -25, -60,   2,   3,  10,  11,  18,   5, -21, -15,  -1,  -8, -14,   4, -10, -20,  38,  -2,  -2,  -2,   9, -24, -55,  -3,  20, -10,   1, -11, -26,  -2,  20, -32,  18, -10,   6,   5,   5,  -6,  17,  -5,   4,  -4,   0,  -1,   4,   2, -13,  16, -11,   9,   9,  -1,   8,  -5,   3,   1,  -1,   1 },
      {   0,   1,   2,  16,   9,  -3,  -1,   0,  -4,  -1,  -8, -25, -32, -20, -16,   7,   2,   4,  10,  30,   9,   0,   9,  10,  -4,  -4, -19, -30,  -6,  16, -12,  13,   4,   5,   0,   8, -30,  -1,   2,  -2,  -1,  23,  24, -15,   1,   8,   4,  -5,  -9, -41, -24,  32, -33,  25, -11,   2,   4,  29, -20,  32, -15,  13,  -2,   6 },
      {  -2,  -3,  -4,  -4,  -8, -13,  -8,   4,  -1,  -1,  -2,  14,  27,  -3, -12,   7,  -2,  -2,   0,  -8, -25,  -6,   4, -18,  -1,  -7, -18, -13,   3,  -1,  40, -19,   2,  10,  30,   8, -24,  22,  -9,   2, -10, -12, -39, -30,  36,  -7,   1,  14,  -6, -31, -55,  -8,  23, -10,  -1,   5,   4,   0, -12,   0,   1,  -2,   0,   2 },
      {  -1,  -1,   3,   0, -15, -23, -37, -32,  -1,   0,  -7,  -4,   7,  -3, -19, -20,   0,   2,  15,  -1, -44, -13,  17,  16,  -2,  -6, -26,  -2,  22, -24, -20,  19,   3,  10,  25,  17,  -8,  19, -46,  31,   3,   0,   1,  17,  -8,  17,  -3,   2,   3,  16,  21,  11, -14,  -3,  16, -15,   2,   3,  12,  -3,  -1,  -3,   4,  -4 },
      {  -2,  -2,  -2,   4,   0, -45, -55, -11,  -1,   2,   1, -12, -17,   7,   0, -42,   1,   2,  -2,   2,  22,  24,  -7, -34,   4,   5,   8,  11,  40,  37,  -3,   4,   0,   8,   5,  15,   9, -15,  21, -11,  -2,   3,  13,  12,  -3, -11,  10,  -4,  -2,  -5, -25,  -7,   7,  -6,  -1,  11,  -2,   6,   2, -17,  16, -12,   1,   0 },
      {   1,   2,   3,  -2, -25, -27,  25,  40,   1,   1,   3,   3,  18,  32,   6, -23,   3,   3,  -1,   6,  19,  47,  33, -29,  -2,   3,   8,  -5, -21, -18, -16,  15,  -3,  -7, -15,  -4,   1, -20, -35,  25,  -1,   2,  -9, -21,   8,   8,   8,   0,  -1, -11,  -8,  12, -22,  -6,  11,  -5,   8,   7,  24,   1,   3,  -7,   0,   0 },
      {   1,   2,   3,   2, -10,  -6,  24,  17,   2,   4,   2,  -3,  12,  12,   2,  16,  -1,  -4,  -1,   7, -11, -46, -24,   7,  -3,  -2, -13, -37,  -3,  -4, -30,  29,   3,   0,   9,  40,  18, -53,   0,   2,  -1,   4,  19,  21,  -2,  -3,   4,   7,  -3,  -9, -32, -18,  24,  -3,   5,   6,   9,   1,   9, -19,   8, -13,   1,  -7 },
      {   1,   2,  -2,  -2,  11,  24,  14,  -6,  -3,  -4,   0,  -2, -25, -50, -38, -18,  -1,  -1, -10, -21,  -7,   3, -25,  -4,   1,   6,  12,  13,   5, -21,  -1,   8,  -2,  -7, -14, -13, -13,  -3,  -7,  -1,   1,   6,  -3, -19,   4,  24,  19,   7,  -4,  -6, -20, -25, -16,   5,  -6,   5,  13,  21,  48, -30,  22, -16,  16,  -4 },
      {  -1,   0,  -2,  -5,  -5,  -8,  -6,   1,  -1,  -3,  -2,   3,  -6, -19, -16, -24,   1,   1,  -2,   2,   6,  -1,   1,   0,   0,   2,  12,  16, -19, -31,  10,  12,  -2,  -3, -22, -23,   1,  -3,  -4,  14,   3,   9,  33,  21, -11,  24,  -6,  14,  -7, -10, -29, -29,  29, -25,  24, -12,  -5, -25, -48,  42,  -9,  -7,  15,  -4 },
      {   2,   3,   3,   3,   8,  21,  22,  -3,  -1,  -3,   0,   3,  -2, -19, -15,   8,   2,   3,   2,  11,  13,  11,  24,   3,   0,  -1,   1,  -2,  -1,  17,  17, -24,   2,  -1,  -4,  17,  43,  32, -24,  -1,  -2,   3,  28,  45,  22, -11, -12,  -7,  -2, -14, -19,  42,   4, -23,  20, -11,   4,   3,   7, -25,  22, -19,   9,   5 },
      {   0,   1,   0,  -3,  -3,  15,  24,  -2,   0,   2,   2,   9,  12, -17, -40,   9,  -1,  -2,  -6, -11, -14,  12,  28, -24,   2,   5,   3,   6,  30,  47,  -4,  -1,   1,   3,   4,  32,  34, -22,  -3,  14,  -2,  -2, -10, -23, -22,  41, -20,  11,  -1,  11,   8,  -9,  -3,  11, -15,  12,  -5, -12, -20,  27,   2,   1,  -2,   3 },
      {  -1,  -2,  -3,  -2,  13,   6, -30, -18,   1,   1,   4,   5, -12, -11,  35,  42,   2,   2,   1,   1,  22,  37,  30,  53,   1,  -1,   1,   8,   9,  -4,   7,   5,  -2,   0,   2,  -1,  -6, -36, -31,  15,   0,  -3,   3, -11,   2,   5, -30,   7,  -2,  -6, -19, -22,  24,  11,  -8,   7,   2,   0,  13,  -7,  13,   0,  -5,   7 },
      {   0,   0,   0,  -3,  -3,  10,   8, -15,   2,   3,   0,   6,  15, -16,  -9,  47,  -1,  -2,   1,  -6, -11,  21,  26, -39,   2,   3,  -5,   4,  19,  -6,  -8,  40,  -2,  -1,  -3, -26, -35, -16,  14,  10,   4,  -1,   6,  35,  19, -16,  44,  -3,   2,   1,  11,   7,   4,   0,  -7,   0,  -4,  -4, -30, -18,  19,   2,  -4,  11 },
      {   0,  -1,  -2,   0,   5,  -1, -10,   9,  -1,  -2,   0,   2,  -9,  -7,   4,  -4,  -1,  -2,  -3,  -3,  13,   4, -35, -20,   0,   0,   1,   1,  -8,  -3,  10,  13,   0,  -1,   2,  10,  16,  -1, -15,  31,   1,   4,  -3,  39,  63,  -4, -26,  46,  -1,   5,  11,  -8, -27,  22, -32,  16,  -3, -12,   1,   5, -23,  25,   1,   1 },
      {  -1,   0,   1,   0,   0,   4,  -4, -11,   0,   1,   0,   5,  -6, -20,   8,  18,  -1,   0,   2,  -6, -11,  -4, -11, -20,   0,   2,   0,  -2,  -1,  18,  28,  -9,  -2,  -2,  -6,  -3,   0, -23, -11,  34,   0,   0,   6,   1, -45, -55,  18,  22,  -1,   0,  -6,   7, -13, -40,  10,  -1,   3,   6,  41,  35, -27,   7,  -6,   1 },
      {  -1,  -2,  -2,  -2,  -9,  -6,  19,  24,   1,   2,   4,   7,  16,  17, -13, -19,   0,   1,   0,  -7, -19,  -9,  40,  42,   2,   1,   6,  18,  28,  25,  13,  12,  -2,   0,  -5, -14, -17,   6,  20, -23,   3,   2,  18,  35,  22,   2,   9,   9,   1,   2, -13, -19,   2,   4,  -4,   8,   0,  12,  44,  26, -21,  20,  -2,  19 },
      {   0,   1,   0,  -3,  -2,   2,   4,   9,   0,  -1,  -2,  -2,   4,  14,   9,  -6,   0,   1,   0,  -2,  -1,  -2,   7,  36,   1,   1,   3,   0,  -4,  16,  36,  34,   0,   1,   2,   4,  24,  22,  -1,  15,   0,   1,   5, -12, -28, -11,  28,  18,  -7,  -6, -13, -31, -57,  -7,  -8,  -4,   0,  -8, -39, -45,   7,   9,   1,   0 },
      {   1,   1,   0,   1,   2,  -1,   8,  19,   0,   0,  -1,  -2,  -3,  -3,  -9, -26,   0,  -1,  -1,   2,   0, -19,  -6,  24,   2,   4,   3,   5,  13,  34,  36, -19,  -3,  -5,  -5, -11, -41, -56, -17,  12,   1,   0,  -6,   1,  18,  -3,  -1,   8,   3,   4,  26,  27,   1,   7,  14, -27,  -2,  -7, -23, -15,  12, -16,  10, -39 },
      {   0,  -1,  -4,  -1,   5,  -2,  -8, -10,   1,   2,   2,   2,  -4,   5,  33,  34,   0,   0,  -2,  -1,   2,   1,  -3, -11,   2,   2,  -1,  -5,   4,  14,  23,   7,   0,   1,   1,  14,  11,   7,  29,  17,   2,   3,  -5,   4,  29,  45,  22, -25,   0,   5,  -4, -19, -24,  15,  47, -31,   1,   5,  18,  19, -24, -31,  17, -25 },
      {   0,   0,   1,   0,  -2,   8,   7,   3,   1,   0,  -1,   0,   1,  -4,  -8,  -6,  -1,  -2,  -2,  -5,  -7,   5,   1, -12,   0,  -4,  -7,  -3,   4,   3,  -6, -27,  -1,  -1,   1,  -5,  -2,  -6, -45, -12,  -2,  -4,   0,  27,  -3, -41,  -8, -68,  -1,  -3, -10, -56, -26,  33,  -2,   5,  -3,  -3, -14,  12,  -8,  -4,   1, -14 },
      {   0,   0,   0,  -2,   2,   9,   8,   7,  -1,  -2,  -1,   0,   2,  -2, -17, -21,   0,   0,   1,   3,   3,   6,   4,   1,  -1,  -2,  -3,  -1,  -8,  -3,  23,  45,   2,   4,   5,  -1,  -1,  20,  42,  58,  -1,  -3,  -1,  -3, -15, -30, -40, -32,   3,  -1,  -3,   7,  33,  33, -12,  -9,   1,   6,  15,  -5,   0, -15, -11, -17 },
      {   0,  -1,   0,   0,   1,   0,  -4,   2,   0,   0,   2,   0,  -1,   4,   4,  -4,  -1,  -2,  -1,  -5,  -2, -10, -31,   2,   2,   5,   6,   5,  -2,   2,  47,  49,  -1,  -3,   3,   7,   5, -14, -28, -19,   0,   0,  -6,  -8,   9,   8,  -4, -44,   2,  -2,  16,  19,  -2, -12,   2,  16,   0,  -4,   0,  27,  23, -12,  10,  65 },
      {   1,   1,   0,  -2,   3,   7,   2,  -9,   0,  -2,   0,  -1,  -6,  -2,  -8,  -9,   0,   0,  -1,  -3,   4,  -1, -11, -12,   1,   1,   4,   4,  -1,   9,   8,  -2,   0,  -2,  -2,  -1,   2, -15, -13,  12,   2,  -1,   5,  -4,   4,  28,   6, -32,   1,   0, -14,   3,  20,  -3,  26, -26,  -1,  -1,   0, -42, -39,  71, -50,  19 },
      {   0,   0,   0,   0,  -1,   0,   1,   5,   0,  -1,  -1,   0,   1,   9,  10, -11,   0,   1,   2,   3,   1,  -9, -11,  12,   0,  -2,   0,  -3,  -4,   7,  -6, -48,   1,   4,   0,  -5,  10,   7,   0,  60,   2,   2,   5,   7,  -3,   9,  55,   2,   2,   1,   0,  -5,  21,  43,  -3,   6,   0,   4,  -2,  17,  42,   7,   2,  43 },
      {  -1,  -1,  -2,  -2,  -4,  -2,   4,  -1,   0,  -1,  -1,   1,   1,  -3,  -6,   1,   0,   0,  -1,   0,  -1,  -6,   3,  -2,  -1,  -1,  -1,  -1,   4,   1, -26, -21,   0,  -1,   0,   1, -24, -14,  39,  20,  -1,  -4,  -7,  -3,   7, -16, -54,   3,   0,   0,  -5, -18, -39, -17,  30, -32,   0,  -5,  -5, -14,  13,   2,  25,  63 },
      {   0,   1,  -1,  -2,   0,   0,  -3,   4,   1,   0,   3,   4,   2,   1,   5,  -2,  -1,  -1,  -3,  -1,  -2,  -9,  -2,  16,   0,  -1,   1,   4,   0,  -1, -12, -22,   1,   2,   0,  -5, -11,  -6,  16,  37,   0,   0,   4,   9,  18,  29,   0, -41,  -2,  -2,  -5,   2, -21, -66, -54,  22,  -3,   1,  -5,   0,  -6, -29, -37,  -7 },
      {   0,   0,   0,  -2,  -1,   1,  -3,  -5,   0,   1,   1,   3,   5,   6,   5,   5,   0,   0,   1,   1,   0,  -3,  -9,  -4,   0,   1,  -1,   1,   1,   1,   0,  10,   0,  -1,   1,   3,   4,   3,   6,  -6,  -1,   0,   1,   2,   7,   3, -10, -11,  -1,  -3,  -8,  -6, -13, -20, -17, -43,   5,   1,  14,  36,  77,  61,  -4, -40 },
      {   0,   0,   1,   1,   3,  -1,  -4,   1,   0,   1,   2,   2,   0,  -1,  -5,  -2,   0,  -1,  -1,  -2,   0,   0,  -4,  -4,   0,   1,   1,  -1,  -1,  -1,  -1,   2,   1,  -3,  -1,   1,   6,   2, -11, -23,   0,   0,   1,   2,   0,  -6,   4,  27,   0,   1,  -2, -10,  -3,  18, -22, -72,   1,   0,   1,  14,   3, -49, -74,  25 },
      {   0,   0,  -1,  -1,  -1,  -2,  -2,  -2,   0,   1,   0,   1,   3,   7,   6,   1,   0,   0,   0,  -2,  -1,   2,   1,  -5,   0,   0,   0,   1,   1,   3,   0,  -4,   0,   1,  -1,   3,   6,  -5, -10,  -1,   0,   1,   2,   0, -10,   5,  12, -14,  -1,  -3,  -5,   5,  22,  -4, -70, -60,   1,   3,   3, -12, -35,   5,  71,  13 },
    },
  },
  { // 17
    {
      { 117, -23, -10,  -6,  -4,  -2,  -1,  -1, -39,   8,   3,  -1,   1,   1,   1,   0, -17,   2,   1,   1,   1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0,  -3,   1,   0,   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1 },
      {  29,  -1,  -1,   1,  -1,   0,  -1,   0, 105, -20,  -9,   2,  -4,  -2,  -2,   0, -53,   4,   5,   1,   3,   1,   1,   0, -33,   4,   2,   1,   0,   2,   1,   0,  -1,   4,  -3,   0,   1,   0,   0,   0,   1,   0,  -1,  -1,  -1,   0,   0,   0,  -3,  -1,   1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,  -1,   0,   0 },
      {  21,  -6, -12,  -3,   0,   2,   2,   1,  34,  12,   5,  -9,  -3,   0,   2,  -1,  97, -26,  -5,  -3,  -6,  -5,  -1,   1, -43,   0,   5,   1,   6,  -1,  -2,   1, -45,   6,   5,   1,   1,   1,  -1,  -1,  -6,   4,   2,   0,   0,  -1,  -1,   0,   4,   1,  -2,   0,   1,   1,  -1,   0,  -3,   0,  -1,  -1,  -1,   1,   1,   0 },
      {   9,   1,  -5,  -4,  -3,   1,   1,   1,  34,  -7,  -2,  -2,  -1,  -3,   0,   0,  10,   4,  -1,   1,   1,   0,   0,   0, 101, -12,  -6,  -5,  -3,  -2,  -1,  -1, -44, -15,   6,   3,   2,   1,   0,   1, -48,   6,   4,   1,   1,   1,   0,   1,   0,   4,   0,   0,   0,   0,  -1,   0,   2,   0,   0,  -1,   0,   0,   0,   0 },
      {  -6, -73,   1,  16,   9,  -2,   3,   2,  22,  45,  -7, -40,  -1,   1,   3,   0,  16,   6,   4,  -4,  -4,  -2,   0,   2,   9,  -2,   0,   5,   5,  -5,  -4,   0,  67,  -8,   8,  -1, -11,   0,  -1,  -1, -29,  -4,   4,   4,   3,   0,  -2,   1, -15,   5,  -3,   1,   3,   2,  -2,   1,   5,  -2,   0,   0,  -2,   2,   0,   0 },
      {  20,  75, -17, -22,  -9,   0,  -1,  -1,  10,  -4,  10,   2,  -2,   0,   2,  -1,  21, -42,   4,  10,   0,  -4,  -1,   2,   9,  10,   0,  -3,   3,  -2,  -2,   1,  72,  -6,  -5,  -2,  -1,   0,  -2,  -1, -30,  -8,   4,   0,   1,  -1,  -1,   0, -22,   1,   1,   1,   2,   1,   0,   0,  -7,   1,   0,  -1,  -1,   1,   0,   0 },
      { -13,  17, -30, -17,  -1,   1,   4,   2,  -5,  53,  18, -55,   0,   2,   8,  -3, -45, -46,  15,  12,  -9,  -9,   1,   6,  -4,   9,   5,   0,  15, -12, -11,  -1, -31,  -4,  24,   4,  -9,   2,  -3,  -3,   2,   3,   6,   6,   3,  -4,  -5,   1,  33,   2, -11,   0,   7,   4,  -5,   1,   0,   0,  -2,  -3,  -4,   5,   1,   0 },
      {  15,  15,  12,  -5,   1,  -2,  -1,  -2,  19,   0, -13, -38,   4,   3,   2,  -1,  21,   8,   2,   5,  -1,  -5,  -2,   2,  42, -13,  -2,   1,   1,  -6,  -2,   3,  12,  11,  18,   2, -14,  -1,   0,  -2,  99,  -9, -10,   3,   0,   1,  -1,   0, -16,  -9,  -5,   2,   2,   1,  -2,   2,  -4,   1,   5,   2,  -2,   1,  -1,  -1 },
      {  -6, -22, -28,  29,  21,   8,  -4,   2,   0,   6,  19,  28, -16,   8,  -1,  -3, -12, -52,  23,  -8, -16, -13,   9,   4,  19, -19,  -3,  18,  14,   6,  -4,   7,  -6,  46, -34, -13,  11,   1,  -4,  -5,  16,  19,  -3, -12,  -8,  -5,   4,  -5, -43,  -7,  10,   3,   1,  -2,   4,   1, -16,   3,  -4,  -1,   4,  -2,   0,   1 },
      {  -3,  27,   7,  15,  -3,  -3,  -1,  -2, -12,  19, -16, -23,   4,  -4,   3,   2, -13,  13,  -2,  -4,   6,   3,  -2,  -1, -23, -29,   5,   1,  -2,  -3,   4,  -2, -32, -12,  16,   5,  -5,   0,   2,   1, -18, -16,   2,   4,   3,   3,  -1,   2, -96,   9,   5,   0,   3,   0,   1,   0,  27,   8,   1,   1,  -1,   1,  -1,   0 },
      {   1, -18,  82, -32,  16,  -8,   3,  -4,  -7, -41,   9, -24,   5,   5,   1,   0,  -5, -46,  -1,  21,  -4,  -3,  -3,   4,   0,   8,  -1,   4,   2,  -3,  -4,   4,  -5,  35,   4,  -6, -12,   0,   0,  -4, -19,   5,  -5,   6,   1,   0,   1,   0,  -7,   5,  -6,   4,   2,   1,  -2,   1,  26,   2,  -1,   1,  -2,  -1,  -1,  -2 },
      { -15, -42, -34, -59,  -9,   7,  14,   4,   0, -23,  38,  11,   4,  -5,   3,   1,  -3,   0, -14,  17,   2,   0,  -4,   1,   3,  47,  -4, -18,  -1,  -3,  -2,  -4,  -8, -21,  10,   9,   6,   2,  -2,   1,  17, -20,   2,   3,   3,   0,  -5,   1, -48,   4,   1,  -2,   2,   3,   0,  -1, -19,  10,  -2,  -3,  -2,   2,   2,   1 },
      {  -3,  -1, -59,  17,   7,   6,  -3,   1,  -9, -66,   8, -35,   6,   8,   6,   0,   9,  19,  34,  -5,  -3, -10,   1,   2,  -4,   8,  -6,   5,   5,  -7,  -4,   4,   6,  14,   4,  -5, -11,   0,  -2,  -3,  -5,   6,   1,   4,   0,   1,   1,  -1,   7,   9,  -2,   4,   1,  -1,  -2,   1,  62,  -5,  -2,   2,  -3,   0,  -1,  -1 },
      {  -5, -12, -19, -13, -15,   4,   4,   2,  -2,  18, -90,  29,  10,   7,   0,   1,   0, -30,   1,  11,  15,   2,   0,  -2,  10,  37,  21, -17,  -6,  -6,   5,  -3,  -4,  23,   0, -11,   1,  -2,   1,   1,   6, -11,   6,   1,  -2,  -1,   0,   0,  -1,  -5,   3,   3,  -1,  -3,   0,   0,  38,  -1,   0,  -2,  -1,   0,  -1,   1 },
      {  -2,  -7,  -5,  37,  -6,   6,   1,   1, -15, -50, -43, -34,  10,   3,   9,   2,  -4, -25,   1, -14,   9,   0,   2,   1,  -5,  12,  15,  -3,  -2,  -7,  -1,  -3,  -4, -11,  21,   4,  -9,  -1,  -1,   0, -14,   7,   0,   5,   2,   1,  -3,   1,  -7,  -9,  -4,  -1,   2,   1,  -2,   0, -80,  -1,   8,  -1,   1,   0,   0,   0 },
      {   8,  19,  16,  56,  22,  -8,  -4,  -2,   4,  12,  17,  -9,  -2,   0,   0,   0,  -1,   5, -31, -18, -22,   3,  -1,   1,  12,  91, -13,   8,   5,   3,  -2,  -2,  -9,   7,  -6,   0,   4,  -1,   2,   1,   2, -22,   8,  -4,  -3,  -3,   1,   1,  -7,  -4,   3,  -3,   1,   0,   0,  -1,  11,   3,  -2,   2,   0,   1,   0,   1 },
      {   0,  14,  -4, -39,  52, -11,   0,  -5,  -4,   9, -32, -13,  -3,  17,  -3,  -1,   5,  51,  33,  33, -35, -11,  -5,   1,  -6,  14,  -8,  14,  10,  -5,  -1,   5,  -9,  19, -21, -22,   1,   1,   3,  -3, -18,  -8,   8,   2,  -4,  -4,   4,  -2,  -6,  -9,   6,   3,  -1,  -3,   0,   3, -37,  -2,   3,   2,   0,   1,  -2,  -1 },
      {   2,  -7,  21,   5,  15, -21,  -6,  -5,   0,  -1,  -7,   8,  -4,  -5,  -4,   3,   7, -28,  83, -12,   2,  14,   1,  -2,  -1,  15, -13,   3, -14,   3,   3,  -3,  -7, -69, -23,  12,   5,   2,   2,   3,  17, -14,   0,   5,   2,   3,   0,   1,   0,  10,   7,  -4,  -3,  -2,  -1,  -2,   8,   6,  -3,   0,   0,   0,   1,   1 },
      {  -3,   3,   7,   4, -59,  -9,  -2,   2,   1,  18,  24, -14,   9, -11,  -4,   1,   6,  21,  51,  -4,  40,  12,   0,  -4,   2,  20, -12,  -9, -22,   3,   2,  -2,  -7,  60, -16,   2, -13,  -4,   1,   2, -14, -17, -13,   9,   6,   7,   1,   1,  -3, -14,   3,   1,  -2,   1,   0,  -1, -20,  -4,   3,   1,   0,  -3,   1,   0 },
      {   1,   6,  26,  -8, -30,  36,   4,   9,  -1,   6,   4,  -1, -38,  14,   7,  -1,   2,  26,  23, -16,  -3, -35,   3,   1,   3,  31,  61,   1,  25,   0, -12,   2,   0, -16,  -8,   1,  -1,   5,  -5,  -3,   1,  50, -11,  -7,   0,  -2,  -3,  -5, -13,  -1,  -9,   5,   2,   6,   2,   2,   6,   1,  -5,  -2,  -1,  -1,   3,   1 },
      {   0,  -7,   8,  38, -52,  13, -13,   0,  -1,  -5,  -1,  -2, -14,  10,   2,  -1,   2,  -2,  -2,  88, -10, -12,  -6,   3,  -4,  -2, -31,   0,  14,  -2,  -2,   3,  -3, -24, -12, -32,   8,   2,   6,   0,   6,  -2,  16,  -2,   1,  -3,   1,  -3,  -1,   6,   1,   1,   1,  -3,   0,   0,   0,  -1,   0,   0,   0,   1,  -2,  -1 },
      {  -1,   2,   8, -24, -24,   9,   6,  -1,  -2,   4, -25,   0,  -4,  -6,  15,  -2,  -2,   2,   1, -43,  -3, -14,   3,   5,  -6,   9, -89,   9,  10,   2,   2,   6,   2,   0,  14,   2,   5,  -1,  -1,  -5,   2,  46,  31,  -6,  -4,  -8,   0,  -2,  -9,  -5,  -2,   0,   7,  -2,  -1,   0,   1,  -4,  -5,  -1,   3,   4,   0,   1 },
      {   2,  -2,   9,  -1,  -1,  53,  11,   2,  -4,  -8, -11,   6, -56, -19,   9,   3,  -2,  -1,  12, -17, -12, -19,  -7,   4,   1, -15, -11,   0,  18,   6,   0,   1,   0,  16,  13,  20,   6,  -1,  -3,  -2,  -5, -79,   7,   0,   7,   2,  -1,  -4,  15,  16,  -6,  -3,   5,  -1,  -2,   0,  -3,   8,   3,  -2,  -1,   3,  -1,   0 },
      {  -4,  -7,  -3, -21, -55, -46, -16,  -5,  -1, -13,  -8, -21,   8,  21,   9,  -1,  -3,  -2, -19, -33, -35,   6,   7,   5,   4, -12,  22,  41,  14,  -4,  -8,   0,  -2,  -3, -41,  -1,  21,   5,   1,  -1,   2, -33,  16, -13,  -4,  -6,   0,  -2,   5,   4,   7,   2,   6,   0,   0,   1,   1,   4,  -6,  -3,   1,   2,  -1,   0 },
      {   2,   0,  18,  13,  -8,  16,   8,  -3,   6,   4,  17,  31,  76,  21,   7, -13,   0,   6,  26, -18, -29, -28,   5,   3,   0,  -9,   9, -34,  15, -23,  -3,   6,   2,   1,  27, -19,   2,   0,  -3,  -7,  -1, -21,  23,   0, -10, -12,  -2,  -1,   2,   1,  -7,   5,  11,  -4,  -7,   5,  -4,   8,  -3,  -4,  -4,   6,  -4,   0 },
      {  -2,  -1,  -2,  13,  -7, -82,   6, -10,   0,  -5,   2,  34, -45,   8,   8,   8,   2,   7,  15,   8,  -8,  -4,   2,   3,   0,   1,   3, -17,  25, -10, -10,  -2,   4,  16,  57,  12,  -7,   4,  -4,  -1,  -1,   5,   4,   0,   2,   3,  -3,   0,  -2,  14, -13,  -4,   2,   7,  -4,   1,  -4,  -5,  -1,  -2,  -1,   2,   1,   0 },
      {   0,   1,  -2,  -1, -17,  19, -21,  -5,   0,   4,  -9,   0,  14, -14, -14,  -2,   0,   1,  11,  10, -64,  28,   0,   3,   1,   8,   4,  -1, -31,  15,   6,  -4,   4,  20,  19,  29,  19,  -5,   1,   3,   2,  24, -21,  -9,   7,   1,   0,   3,  -4,  67,  -4,  -9,   1,  -3,   0,  -3,  -7, -15,   3,   3,   0,   0,  -1,  -1 },
      {  -1,   2,  -3,  -4,   3,  11, -24,   6,   2,   7,   0,  13,  -3,  80, -16,   2,  -1,  -1,  -8, -10,  33,  -5,  10,   1,   0,   7, -22,  28,  -6, -11,  -6,   5,  -3,  -4,  11,  -8, -28,  -1,   1,  -2,  -2, -10, -32,  -1,   7,   1,   5,  -2,   3,  50,  -5,   6,  -2,   0,   0,   0,  -9,  31,   3,   3,  -4,  -7,   3,   0 },
      {  -2,  -6,  -1,  -6, -10,  14, -44,   5,   0,  -2,   5,  10, -13,  22, -14,   2,   0,  -5,   9,  -1, -36,  17,   4,  -1,  -2,   0,  -3,  19, -23,   2,   3,   3,   2,  -8,  46,   0,   5,  -6,   3,  -1, -10,  -7, -22,  -2,   7,   0,   1,   2,  -3, -85,  -5,   1,   0,   1,   3,  -1,  12,   5,   8,   2,   1,  -1,  -1,   1 },
      {   1,   1,  -1,   6,  -5,  -2,  46,   6,   1,   2,  -3,  22,  30, -34,  11,  -7,   3,   3,  11,  22,  11, -11,  -6,   5,   2,   8,   5,  92,  -2,   3,  -8,   4,   1,  -2,  27,   7,   1,   3,  -6,  -5,  -2,  -1, -17, -13,  -5,   2,   1,   0,   2,  -1, -13,  -1,   5,   0,   1,   2,   1,  16,   2,  -5,   1,   2,   1,   1 },
      {   2,   4,  -2,  10, -10,  -7,  59,  25,  -1,  -3, -10, -12,  -2,  18, -20,   7,   1,   2,  -1,   5, -29,  14,  -3,  -5,  -2,  -8, -14, -28,   3,   8,  -7,  -7,   4,   5, -18,  20,  11,  -4,  -5,   1,  -3,  12, -22,  10,   2,  -2,  -2,   5,   5, -15,  -1,  -7,   1,   4,  -4,  -3,   5,  81,  -4,  -6,  -4,  -2,   5,   3 },
      {   1,   1,   1,   5,  -6,  -2,  51,   5,   0,   3,  -2,  -1,   8,  43,   3,   9,  -3,  -6,  -5,   1, -12, -17, -11,   2,  -2,  -3, -19,  -8,   0,  -3,   1,  -3,  -3, -10, -15,  28,   7,   9,  -1,   0,  -4, -13, -46,  -7,  -3,   7,   3,  -1,  -6, -13,  -6,  -1,  -5,   2,   3,   3,   5, -80,  13,   1,  -1,  -5,   1,  -2 },
      {  -3,  -4,  -7,  -2,  -5, -13, -16, -57,  -1,   4,  -6,  -3,   6, -30,  18,   4,  -3,  -3,  -7, -13,  -8, -44, -23,  22,  -1,   3, -11, -20,   1,  14,   7,   6,  -1,  -8,  -5, -26,   2,  13,   4,  -3,  -6,  -1, -65, -10,   2,  13,  -1,  -3,   5,   0,   2,  16, -10,   2,   6,   3,   1,  28,  16,   9,  -4,  -2,  -5,  -3 },
      {  -1,  -1,   1,  -1,   6,  -5, -52,  27,  -1,   0, -13,   6,  27, -22,  -2,  -7,   1,   5,  -4,  19,   8, -20,  36,  -5,  -1,   2, -13,  -7,  53,   4, -19,   4,   0,   0, -18,  63, -10,   7,  -8,  -1,  -5,  -3, -25,   5, -10,  -2,   6,   1,  -1,  -9,  -7, -12,   5,   3,  -4,   1,   1,   4,   7,  -4,  -2,  -2,   0,  -1 },
      {   0,   0,   0,  -7,  -9, -10,  10,  40,  -3,  -4,  -3,  -7,  -6, -33, -62, -10,  -1,  -3,   0, -18,  -8,   8,  23, -19,   0,   3,  -5,  -4,  32,  11,   8,   3,  -3, -11,  14, -62,  -5,  -5,  -4,  -2,  -1, -11, -32,  -1, -13,  -5,  10,   4,  -2,  11,  -2,  14,   8,  -4,  -6,  -1,  -2, -21,  -1,  -1,   0,  -2,   1,   1 },
      {  -1,  -3,  -3,   1,  -7,  -5,  -1, -29,  -2,  -2,   0,  -5,   7,   7, -83,   0,  -2,  -6,   1,   0,   5, -44, -42, -11,   2,   0,   7,   8,   4,  27,  14,   6,   0,   2,   7,  32, -15,   4,   6,   3,   0,   2,  34, -10,  -5, -10,   6,   2,   0,  -5,  17, -11,  -8,   4,   4,   0,   0,   1,   1,   9,   0,   1,   3,   1 },
      {   0,   0,  -1,   3,  -1, -19, -10,  50,  -1,   0,  -2,   8, -11, -21,  -5,  -3,  -3,  -3, -12,  -4, -22, -62,  -3,  -5,  -2,  -2,  -2,   2, -65, -39,  -7,   8,  -2,  -2, -13,   0, -15,  10,  -2,  -3,   1,   1,   3,  36,  16,   6,  -2,  -6,   0,   5,   2,   3,   6,   1,   0,   0,  -1,   0,  -3,  -5,  -1,   3,   4,   1 },
      {   1,   1,   4,  -2,  11, -15,  -9,  41,   0,   1,   0, -22,   3,  14,  17, -16,   1,   0,   6,   5,  25, -33,  11, -12,   0,   0,   2, -14, -23,  38,   0,   2,   3,   5,  24,  -5,  65,  11,  -1,  -2,  -1,  -8,   4, -51,  -1,  -1,   4,   2,   3,   5,  11,  -7, -23,   3,   4,   0,   0,   7,  -9,  10,   5,  -5,   1,   2 },
      {   0,   0,   2,   1,   0,   6,  19, -58,  -2,  -2,  -1,  -9, -13,  -5, -31, -24,   2,   3,  -4,   9,   7, -12,  80,   1,  -2,   1,  -2,  -2, -24, -42, -10,   0,   0,  -1,  -1,   8,  20,   0, -21,   2,  -2,  -2,  -1, -17,   4,  -1,   9,   2,   0,  -6,   7,  -8,   3,  -6,  -7,   3,   1,   1,  -9,  -1,   2,  -4,   1,   0 },
      {   1,   0,   5,   0,   2,   7, -12,  10,   0,   1,   4,  -5,   2,  -7, -17,  29,   1,  -2,   4,  -3,  12,   7, -29, -10,   1,   2,  -3,   9,  27, -80,  22, -17,   4,   9,   9,   6,  51,  15,   7,   9,   0,   9,  -9,  -1, -16,  30, -11,  -1,   3,   4,  26,  -1, -13,  -8,   5,   4,   0,   7,  10,   6,   0,   2,   0,   1 },
      {   0,  -1,   0,  -1,   2,  -2,   3,  21,  -1,  -1,   0,   3,   1, -12,  -6,  12,   0,  -1,   1,   1,  -7,  10,  -6,  12,  -2,  -1,  -2, -13,  -3, -29,   1,   8,  -3,  -4, -15,   3, -43, -11,  -6,  -3,  -2,   0,  12, -95,  33,  13,  -5,  -9,   0,  -4, -11,   4,  11,   3,   0,   5,   0,   6,  22,  21,  -5,   0,  -3,   0 },
      {   0,   0,  -3,   3,   7,  -1, -11,  -7,  -1,  -1,  -4,  -1,   5,  -2, -26,  12,   0,   1,   1,  -3,  11,   0, -10,  10,   0,  -2,  -4,   0,  -4,  -4,  -8, -14,   0,   3, -17,  -8,  32,  11,  -1,  -1,   1,  -3,   9,   4,  12,   4,  -8,  -1,  -8,  -3,-109,   3,  -9,  15,   1,  -1,  -1,  -1, -17,  -1,  -4,   3,   7,   4 },
      {   1,   2,   2,  -1,   1,  11,   0,   2,   0,   0,  -2,   3,  12,  -5, -16,  62,   1,   1,   1,  -3,  -3,   5,  13,  31,  -1,  -2,  -2,  -1,  -4,  19, -72, -20,   0,   1,  10, -13,  -7,  46,  -8,   1,   1,  -3,  12,  -1,   6,  23,   3,  -1,   1,  -3,  29,   6, -15,   0,   1,  -3,  -1,  -7, -15,  -2,  -9,  -6,   7,  -3 },
      {   0,   0,  -1,  -1,  -4,   3,   2,   4,  -1,   1,   1,  -3, -14,   1,   2, -30,   0,  -1,   0,   1,  -6,  16, -10,   3,  -2,   0,   0, -10, -21,  -7, -10,  12,   0,  -2,  -1,  11, -22,  49,   6,  -1,  -3,  -3,   8, -14, -89,  15,  17,   3,   0,   0, -21,  23,  -2, -17,   6,   1,  -1,   9,   4,  13,  22, -13,  -5,  -4 },
      {   0,   0,   1,   2,   3,   0,  -6,   3,  -1,  -2,  -6,   1,  -4,  -4, -14, -57,   0,   3,   0,  -5,  15,  20, -51,  35,  -2,  -1,  -3,  -9,   4, -24, -62,  19,   0,  -2,   0,   2,  20, -18,   1, -19,   0,   0, -16,  -3,  17, -22,  17,  -3,  -1,  -1,  15,  -5,   3,  -4,  -4,   0,  -1,  -8, -21, -10,  -1,   3,   0,  -5 },
      {   1,   0,   0,  -1,   0,   5,   0,  -5,   0,  -2,  -2,   4,   0,   1,   7, -40,   0,   0,   1,  -4,  -6,  13,  -9, -32,   1,  -2,  -4,   2,   8,  -4,  11, -13,   1,   5,  -1, -15, -19,  85,   5,  11,   1,   1,  -7,  -8,  36,  -7, -17,   3,  -1,  -3,   2, -51,  -1,   4,  -4,  -6,   1,   1,  -7, -15, -13,   1,  10,  -5 },
      {   0,  -1,   0,  -1,  -2,   2,   2, -10,   0,   0,   0,   5,   0,   0,   6, -30,  -1,  -4,   0,  -6,  -8,  -3,  -7, -59,  -1,  -1,  -4,   4,   3,   2, -54, -25,   0,   0,  -2,  -6,   4, -23,  -9,  29,   2,   3,  14,   8,   8,  24,   4,  13,  -2,   3,  -6,  18, -23,  10,  12,  -4,   3,   1,  60,  19, -14,  -9,   0,  -4 },
      {   0,   1,   0,   4,   0,  -2,   6,   8,  -1,   1,   0,   1,   4,   4,  -9, -35,   0,   1,  -1,   3,  -2,  12,   9,  26,   0,   3,   0,   2,  14,  10,  32,  17,   0,   0,   3,  19,   7,  30, -19,  -1,   1,   2,  16,  16,  42,  26,  -6, -15,   1,  -1,   8,  80, -23,  -1,   7,   3,  -1,   0,   2,  -8, -14, -11,  -3,  -6 },
      {   0,   1,   0,   1,   0,   3,  -6,  -8,   0,  -1,  -5,   1,   4,  -1,   9, -10,  -1,  -1,  -1,   4, -10,  -6, -13, -37,   0,  -3,  -7,   1,   2,  -4,  -3, -24,  -1,   0,   1,   6, -21, -26, -15,  21,  -1,  -5, -14, -18,  -9,  28, -21,  -5,   1,  -4,   8,  26,  -7,  16,   2,  -8,  -5,  -4, -93,  -1,   6,  -6,  10,  -1 },
      {   0,   0,   0,  -2,  -5,   1,   2,   7,   0,   0,  -2,   3,   3,   4,  -2, -22,  -1,  -1,  -1,  -3,  -8,   2,   6,  36,   0,   1,   1,   2,   8,   5,  11,  27,  -1,  -2,  -1, -14,  -9, -20, -28, -12,   0,  -2,   9,  16, -13,  67,  -8,   8,   0,   0,   1, -60, -37,  28,  12,   3,   0,  -1,  -2,  21,  -4,  -3,  10,  13 },
      {  -1,   0,  -1,   1,  -3,   2,   2, -10,   0,   2,  -1,  -1,   2,  -3,   9,  32,  -1,  -2,  -2,   0,   0,  14,  -8, -48,   0,   2,   2,  -2,   2,  -5, -12,  86,  -1,  -2,   0,   2,   8,  17, -35,   0,  -1,  -1,  -5,   8,  14, -13,  15,   2,   0,   1,  -5,   4, -10, -11,  -2,   5,  -2,  -1, -17,  34,   2,  20,  -1,  22 },
      {   0,  -1,  -2,  -1,   0,   0,   7,  -1,   0,   0,   1,   3,   3,   3,   3, -11,   0,  -2,  -1,   0,   0,  -5,   9,   0,   1,   0,   3,   1,  11,   5,  -1, -11,   1,   1,   1,   5,   5,   7,  63,  -8,   2,   3,   2,  11,  31,  43,  43,  22,   1,  -3,  -1,   4,  39, -15, -10,  -3,  -2,   0, -20,  56,  17,   6,  -3,  27 },
      {   0,   0,  -2,  -2,   0,  -1,   4,  -7,   0,   3,   1,  -4,   4,  -1,  -1,  10,  -1,  -2,  -1,  -3,  -2,   0,   3, -30,   0,   3,   3,  -3,   3,   4,  -7,  34,   0,  -2,   4,  -1,   3, -13,  39, -13,   1,   1,   5,  -4,   9,  44,  23, -50,  -1,   0, -11, -18,  -4, -26,   5,   7,   0,   2,  -3, -56,   3, -37,  -9, -35 },
      {   0,   0,   1,   1,   5,  -3, -10,  -5,   0,   0,  -1,  -2,   2,   9,  -2,   1,   0,   2,   0,   2,   8,   1, -20, -11,   0,   1,   0,  -3,  -5,  14,   5, -17,   0,  -1,  -1,  -5,  16,   8, -71, -23,   1,   2,   3,   4,   1,  31,  13,  -8,   2,   0,   4,  -4,  85,   1,  14,   5,   1,   0,  11, -10,  12, -12,   4,  -1 },
      {   0,  -1,  -1,  -2,  -5,   0,   0,   5,   1,   2,   2,  -1,   1,  -4,  10,   6,   0,  -2,   0,  -1,  -8,  -2,  -7,   0,   0,   0,   2,   2,   7, -11,  18, -43,  -1,  -3,   2,   0, -15,  -2, -38, -25,   1,   0,   1,   6,  17, -25,  68,  -1,  -1,  -1,  -8, -13, -47, -29,  13,  -4,   0,   4, -12,  22,  19, -24, -16, -24 },
      {   0,  -1,   0,  -2,   0,  -4,   3,   8,   0,   0,   1,   0,   1,   3,  -3,  -6,   0,   0,   1,  -2,   3,  -5,  -2,  27,   0,   0,   1,   1,   3,   8,  -2,   4,   1,   0,   3,  -3,   6,  -6, -14,  81,   0,   1,  -2,  11,   1,  -3, -17, -40,  -1,  -4,  -7, -19,  19, -53,  -1,  -6,   0,   1,  -7,  41, -11, -10, -11, -28 },
      {   0,   0,  -1,   2,  -1,   0,   4,  -3,   0,   1,  -1,   3,   0,  -4,  -3,   1,   0,   0,  -1,   1,   1,   7,   8, -15,   0,   0,   0,   2,  -1,  -1,  -6,  -1,   0,  -2,   0,   3,   6,   3,  15, -61,  -1,  -2,   0,  12,  -5, -17, -47, -29,   1,   0,   5,   1,   8,  19,  -1,   3,  -1,  -1,  -3,  64, -12, -34,  12, -50 },
      {   0,   0,   0,   0,  -1,  -5,  -1,  -2,   0,   0,   1,  -1,  -1,   3,  -6,  -7,   0,   0,  -2,   0,   0,  -4,   3,  -8,   0,   1,   0,  -2,   2,   9, -10, -14,   0,  -1,  -1,   3,  -8,  -8,  -4, -43,   0,   0,  -2,  -1,   5,  10, -54,  10,   0,   0,  -5,  -2, -14, -89,  30,  -1,   0,  -2,   1,  -6,   6,  23,   2,  32 },
      {   0,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,  -5,  -4,   7,   5,   0,   0,   0,   0,   1,  -2,  -2,   2,   0,  -1,   0,   0,  -4,  -5,   9,   7,   0,   0,   0,   3,   2,  -4,   6,  -3,   0,   0,   2,  -5, -14,  -7,  21,  39,   0,  -1,  -3,  -1,  15, -17,  32,  -4,  -4,  -2, -15,  -3,-101, -38,  13,  10 },
      {   0,   0,   0,   1,  -1,  -1,  -3,  -2,   0,   0,   1,   1,  -4,   2,   1,  -7,   0,   0,  -3,   1,   2,   3,  -4, -10,   0,   0,   0,   2,  -8,   9,   4, -24,   0,   0,  -2,   2,   0,  -2,  -9, -21,   0,   3,  -2,  -1, -19,  14,  20, -56,  -2,  -2,   1,  -1, -11,  -8, -68,  15,  -1,   0,   2,   6, -56,  45,  -8,  22 },
      {   0,   0,  -1,   0,   0,   0,  -5,   0,   0,   1,   1,   3,  -3,   3,   1,   4,   0,   0,  -2,  -1,   1,  -6, -11,   1,   0,   1,   2,   4,  -1,   4,   3,  11,   0,   0,  -5,   0,   4,  -5, -15,  -5,   1,   2,   3,  -4,   4,  18, -16,  67,  -2,  -2,  -4,   3,  -4, -28, -86,  -6,   0,  -1,   2,  -9,   4, -22,   2, -41 },
      {   0,   0,   1,   0,   1,  -1,   0,  -3,   0,   1,   0,  -1,   3,   0,   5,   2,   0,   1,   0,   2,  -1,  -2,   0,   4,   1,  -1,   2,  -2,   3,  -4,   6,   7,   0,   0,   1,   0,   2,  -4,  -1,  -4,   0,  -1,   1,  -4,  -3,   1,  15, -20,   0,  -1,   0,   5,  -2,  -8,  -7, -95,   1,  -1,  13,  -1,   5,  10,  77,  -1 },
      {   0,   0,   0,   0,   0,   0,   2,   2,   0,   0,   0,   1,  -3,   5,  -5,   0,   0,   0,   0,  -1,   3,  -1,   6,  -3,   0,   0,   1,   0,  -2,   1,   0,   5,   0,   0,  -2,   1,   1,   2,   1, -11,   0,   1,  -2,  -3,  -4,  17,   4,  19,   0,  -1,  -1,   0,   7,   9,  28, -46,  -1,   1,  -8,  -2, -20,  69, -57, -62 },
      {   0,   0,   0,  -1,   0,   0,   1,  -1,   0,   0,   1,  -1,   1,  -1,  -3,   1,   0,   0,   0,   0,   0,  -2,  -3,  -1,   0,   0,   0,   2,   1,  -4,  -4,   2,   0,   0,   1,   0,   1,   3,  -4,  -8,   0,   1,   1,   1,   0,  -4, -18, -12,   0,   0,   1,  -4,   0,  11, -18, -66,   0,  -2,   2,  -1,   0, -50, -76,  51 },
    },
    {
      { -86,  18,  19,   8,   2,   0,   1,   0, -78,  14,  22,   7,   3,  -1,   0,   0, -17,   5,   8,   0,   0,   0,   0,  -1,  18,   1,  -5,  -1,  -2,   0,  -1,   0,  20,  -3,  -7,  -1,  -1,   1,   0,   0,  12,  -3,  -3,  -1,   0,   0,   0,   0,   5,  -1,  -1,   0,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0 },
      { -66,  26,   8,  -4,  -2,  -1,   3,   3,  48,  11, -13, -11,  -4,   0,   1,   1,  85,  -5, -18,  -7,  -1,   0,  -3,   0,  -7,  -9,   3,  -2,   2,   0,   0,   0, -26,   0,   3,   1,   3,   0,   1,   0,   0,  -1,   1,   3,   0,   0,  -1,   0,  -1,  -1,   1,   1,   0,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0,   0 },
      {  23,  61, -11, -27, -10,  -1,   0,  -1,  -7,  77, -25, -38,  -4,   3,  -1,  -2, -18,  35, -20, -12,   2,   1,  -1,   0,   3,  -8,  -4,   8,   2,   0,   0,   0,   4, -18,   1,   9,   2,   0,   1,   0,  -4, -15,   5,   7,   1,   0,   1,   0,  -3,  -5,   1,   1,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 },
      { -10,   8,  23,  20, -13, -12,  -2,   1,  10,  11,  16,   3,  -8,  -9,  -5,  -2, -33,  10,   5,  -4,   0,  -2,   0,  -1, -95,  20,  -1,   5,   7,   4,   0,   0, -44,  12,  -7,   2,   4,   4,   2,  -1,  30,  -4,  -6,   0,   0,   0,   1,   0,  18,  -5,  -3,   0,  -1,   0,   0,   0,   7,  -4,   0,   1,   0,   0,   0,   0 },
      {  47, -16,  51,  15, -29, -21,   1,   3, -38,  24,  41,  12, -27, -12,   2,   1,  56,  15,  -6,  -9, -11,  -2,   0,   0,  13,  -1, -17, -15,   4,   5,   3,   0, -15,  -2, -13,  -3,  11,   6,   1,  -1,  -8,  -3,  -4,   2,   5,   2,   0,   1,  -5,  -1,  -2,   1,   2,   1,   0,   0,   2,  -2,   0,   1,   0,   0,   0,   0 },
      { -21,   9,  38,   7, -16,  -7,  -3,  -2,  74,  -2,  39,   1, -20, -11,  -5,  -5, -42,  22,   9,  -9,  -8,  -2,   2,  -2,  34,  11, -26,  -4,   1,   4,   0,   1,  33,  -9, -19,   3,   3,   4,   1,   0, -16,  -7,  -4,   1,   3,   2,   2,   0,  -9,  -1,  -2,   2,   1,   0,   1,   1,  -4,  -2,  -1,   1,   1,   0,   0,   0 },
      { -12, -53,  -4,   2,  36,  17, -15, -14,  -4,   0, -18,  -8,  20,   6, -11, -10,   3,  59, -41, -13,  -1,  -4,  11,   2,   3,  20, -57,   3,  12,   5,  14,  -1, -16,   4,  -8,  20,  -6,  -2,   2,  -2,   1,   0,   2,   1,  -7,   2,   4,   0,  -5,   3,  -1,   3,   1,  -1,   2,   0,   3,  -5,   5,   1,  -1,   1,  -1,   1 },
      { -11, -39,  48, -38, -29,   1,  25,  18,  -7, -22, -16, -50, -10,  26,  23,   8, -12,  11, -30,  -6,  12,  15,  -8,  -1, -12,   9,  34,  -3,   3,  -8,  -8,   3,  17,  17,   9,  -4,  11,  -6,  -4,   0,  -9,   1,   9,  14,   4,  -2,  -6,   0, -12,  -5,   3,   0,   2,   0,  -2,   0,   4,   2,  -2,   0,   1,  -1,   0,  -1 },
      {   9,  -7,  -7, -18,  14,   7,  -7,  -4,  -1,   7,  16,  -8,   3,  -6,  -4,  -2,  36, -11,   8,  -6,  -4,  -4,   3,  -1, -52,   9, -13,   8,   4,   4,   1,   0,  92,  -6, -14,   9,  -5,   1,   0,  -1,  32, -12,  -6,   0,  -3,   1,   0,  -1, -17,   0,   1,   2,   1,  -1,   1,   1, -16,   0,   2,   1,   1,   0,   0,   0 },
      {  -3,  -2,   0,  67, -12, -29,   5,   9,   5,  17, -50,  42,   8, -11,   3,   4,   0,  17, -37,   9,   8,   1,  -8,   0,  -8,  19,  20, -18,  -4,  -2,  -3,   2,  41,   6,   6, -22,  -1,   3,   1,   1, -14,  -9,  12,  -4,   0,   3,  -1,   1, -10,  -6,   6,  -5,  -2,   2,   0,  -1,   7,  -1,   1,  -2,   0,  -1,   0,   0 },
      {  12,  15, -12,   6, -18,   1,   7,   6,   5, -17,  -9, -11,   0,   5,   2,   4,  11,  20,   6, -13,  -2,   1,  -6,  -1,  43,  49,   7,  -7, -11,  -3, -10,  -2,  -5,  28, -11, -10,  -1,   1,  -1,  -1,  85, -12, -10,  -1,   0,  -2,  -1,   1,  20, -20,   0,   3,  -1,   1,   0,   0, -13,  -5,   1,   2,   1,   0,   0,   0 },
      { -13, -26, -34, -12,  13,   9,   4,   4,   2,  24,   5,   2, -18,   0,   6,   5,   8,  46,  56, -22, -34,  -6,   6,   0, -11,  28,  40, -36, -21,   1,   5,  -1,  -1,   5,   4, -16,   1,   3,  -2,   0, -31,   1,  -9,   1,  10,   2,  -3,   0,  -4,   0,  -7,   5,   5,  -1,  -2,   0,   3,   2,  -3,   3,   3,  -1,  -1,   1 },
      {  13,  39,  17,   6,  13,   4, -16,  -3, -10, -24,  19, -19,  45,   9, -24,  -7,  15,  -4,   7,  -9,  21,  -4, -12,  -7,  -5,  60,   6,  12, -18, -17,  -9,   1,  -4,  18, -12,  -2, -16,  -7,   6,   3, -51, -19,   3,   1,  -1,   0,   6,   2, -14,  -1,   1,   0,  -2,   2,   1,   1,  14,  -4,  -3,   0,  -1,   1,   1,  -1 },
      {  -7,  17, -51,   6, -46,   7,  15,   0, -16, -27,  -9,  -5, -46,  -2,  17,   3,   2, -22, -13, -17, -18,   5,  16,   1, -10,  37, -44,   6,  18,   9,  -4,  -3,   6,  11, -11,  13,  14,   1,  -7,  -3, -33,   6,  10,   3,   8,  -1,  -2,  -1,   6,   4,   4,   2,   0,  -1,  -1,   0,  12,  -4,   1,  -1,   0,   0,  -1,   1 },
      {   0, -32,   3,  11, -13,   8,   6,   2,   9,  22,  -1,  -7,  10,  10,  -6,  -5,  12,  -4,  13,  -4,   9,   1,  -8,  -5,   7,   5,   7,  29,  -3,  -3, -13,   1,  20, -21,  -4,  18,  -4,   0,   0,   1, -14,  -7,   4,   0,  -4,  -3,   3,   2,  99,   1,  -7,  -3,  -6,   1,   1,   1,  29,  -7,  -7,  -3,  -1,  -1,   0,  -1 },
      {  11,  39,  20, -28,  33,  17,  -2,  -6,   0, -12,   6,  23,  -5,   9,  13,   9,  -1,   8, -30,   3,  -1,   5,   4,  11,  -9,   7,  -6, -52,  11,   2,  13,   1,  16,  18,   4, -24,   3, -10,   0,   5, -14,  34,  -6,  -4,  -6,  -3,  -3,   0,  59,   7,  -3,   0,  -4,  -3,   1,  -3, -21,  -2,   3,  -2,   0,  -1,  -1,   0 },
      {   6,  12,   9,  36, -20,  74,   3, -23,   1,   1,   6,  22, -18,  58,   8, -16,   6,  22,   7,   9,  -4,  12,   0,  -2,  -7, -11,  14,  14,  11, -15,  -4,   4,   0,  -7,   5,  10,  -5, -24,  -1,  10,   1, -12,   0,  -7,  -7, -11,   3,   3, -15,   7,   5,  -5,  -1,  -3,   2,   0, -22,   1,   2,  -5,   0,   0,   1,   0 },
      {  -3,   6,  -3, -41, -18, -11,   9,   6,  -3, -12,  -1,  30,  -8, -17,   0,   4,  11,  44,   1,  77,   3, -23,  -7,  -1,  -3,  25,   3,  44,  -2,  -9,  -2,  -4,  -2, -10,   8,  12,   0,   4,  -2,  -5,  -9,  12,  -3, -10,   3,   7,   0,  -1,   2,  -5,  -5,  -9,   3,   3,   1,   1, -22,  -5,   2,  -4,   1,   1,   1,   0 },
      {   0,  -9,  11, -10,   4,  17,   1, -11,  10,  45,  -5,   5, -10,   5,   1,  -8,  -1, -24,   4,  32,  -8,  -5,   7,  -1,  11,   5,  -7,  14,   1,  -8,   6,  -1,   8,  76, -13, -10,  -9,  -8,  -1,   1,   7,  28, -12, -15,  -1,   1,   1,   0, -15,   6,  -8,  -3,   2,   0,   0,   1,  57,  -5,  -3,   0,  -1,   0,  -1,   2 },
      {   7,  28,  -6,   3,   2,  -1,   1,   0,  -2, -49,  13,  -2,  -4,  -1,   3,   5,  10,  47,  -3,   1,  -8,  -5,  -2,   6,  -6, -31,   8,  -3,  -2,   6,   7,   0,  14, -16,  13,  -1,  -1,   0,  -1,   0,  20,  -2,   4,  -1,  -3,   1,  -1,   1,  -3,  17,  -2,  -1,   0,   1,   0,  -1,  89,   5,  -8,  -2,  -3,   0,  -1,   0 },
      {  -1,  -3, -14, -24,  -8,   7,  23,  21,   6,  10,   0,  37,  30,  29,   9,   0,   3,   5,  21,   6,  51,  19, -18, -13,  -4,  -1, -23, -26,  26,  11, -18,  -4,  -6, -14, -38, -12,  31,  11,  -9,  -6,   0,  -7, -19,   8,  12,  -3,  -4,   2, -12, -13,   7,  10,  -3,  -4,   2,   3,  23,   6,   2,   5,  -4,  -3,   1,   1 },
      {  -1,  -5,  33, -32,  -3,  18, -19, -11,  -3,  -1, -43,  24, -20,   9, -20,  -2,  -6, -36,  -4,  11, -23,  -4,   9,   7,   1,  36,  -5, -21, -23,  -1,  18,   2,  -9, -51,   0,  -4,   2,   4,   9,   2,  13, -36,  -6,  -1,   9,   4,   5,   1,  -4,   5,  -6,   2,   5,  -1,  -1,  -1,  27,   5,  -2,   0,   2,   2,   0,  -1 },
      {  -1,   2,  -3,  12, -17,   6, -18,   6,  -7, -22, -27, -28,  -1,   8, -16,  10,   0,  11,   0,  16, -18, -16,  -5,   4, -10, -46,  -8,   4, -23, -11,  13,   0,   3,   9, -68, -32,  -2,   8,   6,   1, -16, -10, -38,  -6,  17,   8,   2,   1,  14, -15,  -3,   8,   8,   4,   1,  -2, -13,  -8,   9,   5,   1,   0,   0,   0 },
      {  -5,  -1,   2, -10, -38,  -6, -58, -13,  -6,  -8, -26,   0, -13,  -8, -56, -16,   3,  11,  24, -16,  27,   3, -22,  -5,   3,  -4,  14, -15,  38,   9,   2,   3,   9,  12,  12,  10,  13,  -2,   9,   3,   1,  35,  -1,   0,  -6,  -2,  10,  -1,  -2,  18, -11,  -1,  -3,  -3,   6,   0,  -5,  -2,   0,   0,  -2,  -1,   3,   1 },
      {   7,  22,  40,   2,  16, -26,  21, -13,  -5, -19, -53,  13,  13,   6,  28, -20,   5,  10,  44, -20, -19,  23,  24, -16,   0,  -7,  -5,  20,   0,  22,   5,  -7,   5,  11,  -9,  38,   1,  11,  -7,  -3,  -6,  14, -14,   2,   1,   2,  -2,   0,   1,   0, -17,   0,   8,  -1,  -3,   2,  -7,  -5,  -1,   1,   1,  -2,  -1,   0 },
      {   1,  -1,  12,   5,  22,  11,  11, -19,  -3,  -4, -21, -15, -18, -48,  29,   4,   0,  -1,  23,   6,  19, -39,   3,   8,   7,   6,  16,  -9,  57, -19,  -7,   5,  -7,  -5, -34,  -2,  -2, -19,  -2,  15,  -2, -29,   6,   7, -23,   4,   2,  -1,   2,  15,  21,   0, -13,   2,   3,  -2,   3,   3,   8,  -4,  -4,   1,   0,   0 },
      {  -2,   0,  -7, -17,  14, -16, -17, -15,  -2, -10,   7,  14, -15,   7, -14, -21,   0,   2,   1,  21, -18,  17,  -6, -11,   1, -23,   2, -12,  16,   8,  -7,   1,   2,  41,  16,  12,  10,   0,  -1,   1, -11, -68,  28,  21,  -4,   3,  -1,   2,  14, -52,  10,  10,   3,  -1,   0,   5,   2, -19,   3,   2,  -1,  -2,   0,  -2 },
      {  -1,  -1, -13,  19,  10,  -8,  -8, -35,   0,  -1,  17, -29,   1,  -2,  16, -26,   1,  -6, -23,  21,  10,  16,  20,  -5,   7,  17,  26,   5,  32,  16,   7,  -4,   2, -16,  18, -16,  26,  10,  -4,   1,  -4, -11, -61,  -9,  20,   8,  -4,   2,   0,  -8, -43,   1,  10,  -2,   0,   5,   3,   0,  -1,   2,  -2,   1,  -2,   3 },
      {   2,   4,  -4,  -1,   7,  20, -16,  45,   1,   2,  10,  33,   1, -14,  -8,  30,  -5,  -4, -32, -18,  -8,  -2,  13,   9,   8,   0,  42,  14,  11,  -4,  11,  -5,  -1,  15, -21,  59,  10,  -6,   1,  -8,  -3,  -8, -35,  33,   5,  -7,  -2,  -3,  -2,   1, -11,  14,   1,  -4,   1,  -2,   4,   4,   1,   2,   1,  -1,   0,   0 },
      {  -4,  -4,  -7, -16, -57,   9,  12, -20,   2,  -1,   5,  13,  34, -21,  22, -16,   0,   1,   1, -12,  34, -23,  32,   8,  -4, -15,  -4, -27, -18,  -4,  41,   5,   2,  14,  19,  14, -39,  -6,  10,   1,  -4, -20, -14,  -3, -10,   0,   5,   3,   3, -11, -10,  12,   4,   3,   3,  -2,  -2,  -5,   0,   2,  -3,   1,   1,   1 },
      {   2,   4,  13,  22,   9,   6, -20,  34,   2,   3,  -8, -19,  -8,  25,   4,  54,   5,  -1,  37,  13,  28,  -8,  28,  29,   3,   7, -20,   8,  22,   6,  40,  -9,   5,  -2,  33, -10,   4,   4,  -1, -17,  -3,  -4,  10,  -1,  -3,   1,  -3,  -6,  -1, -30,  -8,   1,   0,   0,  -3,  -4,   2, -21,   5,   4,  -1,   3,  -3,   1 },
      {   1,  -1,   8, -17,   7,  18,  20,   6,  -3,  -4, -10,  21, -13, -17, -13,  23,  -3,   4,  -4, -45,  10, -10, -27,  11,   0, -13, -13,  30,   7,  17, -23,  -3,   0,  23,  29, -22,  -2,   8,  -4,  -2,  -9, -41, -25, -52,   4,  13,   6,  -1,   1,   7, -20, -19,   5,   0,  -1,  -1,  -6,  11,  -7,   1,   5,   0,   0,  -3 },
      {  -2,  -1,  -8,   1, -10, -16,  -6,  12,   2,   5,   5, -15,   9,  10,   4,  15,   2,   5,   7,  34,   0,  37,   2,   9,   2,   3,  -2, -10,  -3,  34,  24,  -1,  -2,  23, -13,   7,  -2,   2,  10,   1,   0, -44,   6,  -4, -14,  -3,   3,   3,   8,  75,  -1, -16,  -4,  -3,  -4,  -4, -14,  34,  -4,  -8,   2,  -1,  -4,   0 },
      {   2,   3,   4,  15,  -5,  -2,  18,  -1,   0,   2,  -7, -12,  28,   7,  -4,   2,  -2,  -6,  14,  14, -26, -40, -28,  15,  -3,  -6, -32, -29,  -1, -40, -13,   9,   6,  13,  34,  24,  46, -14,  -4,  -3,   3,  -9, -12,  27,  21,   1,  -1,  -2,   2,  20, -29,  -5,   8,   0,  -1,  -1,  -6,   9,  -6,  -5,  -4,   2,   0,   0 },
      {  -4,  -3, -13, -12,  14, -59, -10, -24,   2,   0,   5,  20, -32,  40,   2,   4,   1,   5,  -1, -26,  34,  -3,  17,  10,   0,  -3,  -1,  19, -11, -57,  18,   6,   0,   4, -10,  -1,   6, -24, -12,   7,   3, -10,  -8,  -2,  13,  -8,  -6,   7,   4,  10,   1, -10,  -5,   4,  -7,  -2,  -1,  20,   0,  -7,   1,   6,  -1,  -1 },
      {   0,   0,  -8,  10,  16, -15,  48, -24,  -2,   0,   2,  -8, -33,  28, -22,   9,  -2,  -8,  -4,   7,  15, -14, -50,  24,   1,   5,   7, -18,  -5,  24,  17,  -1,  -1,  -4, -12,  31, -32,  21,  14,  -8,  -5,   8, -18,  16, -29,  -2,  12,   9,  -9, -13, -19,  -5,  -5,   2,   6,  -2,   1, -12,   4,  -2,  -2,  -1,   3,   0 },
      {  -1,  -3,  -1,   2,  -1, -16, -13,  23,  -1,   0,  -8,  -5, -13,  -9,   8,  15,   2,   3,   9,  14,  -9,  38, -11, -12,  -4,   1, -27, -12,   8, -16, -37,   1,  -2, -12,  19,  -2, -58, -49,   1,  15,  -4,   0, -43,  10, -19, -13,   9,  -4,  -2, -11, -14,  20,   9,   1,   2,   1,   5,  13,   0,   1,  -1,   2,   4,   2 },
      {  -2,  -3,  -8, -10, -11, -18,   7,  12,   2,   1,   8,  -1,  28,   8,  -3,   6,  -1,  -5,  -3,  -5, -34,  23,   1,  17,   0,   0,  11, -19,  28, -22,   7,   6,  -2, -14, -19,  19,   8, -25,   7,   3,   1,   0,   2, -58, -18,  11,   9,   2,  -8, -11,  -6, -45,  -8,  15,   3,  -1,   4, -50,  -5,  -8,   2,   7,   1,   1 },
      {   0,  -1,  -3,  -1,   8,   3,   2,   6,   0,  -1,   0,   7, -16,  -4,   6,  -1,  -1,   0,  -4,  -2,  24,   6,   0, -18,  -1,  -6,  -2,  -2, -12, -11, -12, -14,  -2,   0,  -4, -12, -10,   0,   0,  -3,   6, -13,  14,  19,  13,   2,  -5,  -6,  -5,  53, -35,  17,  12,   2,  -7,   1,  -6, -90, -10,  10,  12,   1,   0,   1 },
      {   1,   4,  -3,   6,  11,   7, -16,  35,  -1,  -3,   1,  -9, -22,  15,  13, -18,  -1,  -2,  -5,   5,  16, -26,  12, -52,  -1,  -4,  -1, -30,   6, -20, -10, -15,  -1,   2,  -7,  23, -21,  29,  14, -14,   2,  -1,  17, -18,  -2,  32,   4,  -1,   2,  -8, -29, -33,  14,  16,  -2,   8,  -2,  32, -20,  -6,   9,   2,   1,   1 },
      {   0,   1,   2,  -4,  -7, -23,  12,  14,   1,   1,   4,   1,  10,  39, -20,  -9,  -3,  -4,  -6,  -4, -28, -36,  18, -26,   2,   3,   7,  -1,  30,  -7,  -4, -30,   1,   0,  21, -11, -15,  22, -12, -16,  -2,  -7, -27, -12, -23,   3,   5,  -5,   3,  31,  40,  35,  -6,  -1,   4,  -2,  -1, -10,  29,  12, -10,   0,   4,   1 },
      {   0,  -1,   4,   8,   9,   6,  -8,   6,  -1,  -1,  -3, -16, -18, -12,  -3,  -1,   0,  -1,   1,  10,  25, -11,  -2, -25,  -1,  -1,  -3, -24, -39,  26,  -5, -16,   1,   2,  10,  34,  12, -27, -37,   2,  -2,  -4, -26, -24,  21, -21, -15,   9,   2,   1,  36, -34, -10,   1,   0,   3,   4, -21,  38, -11, -12,   0,   6,  -3 },
      {  -2,  -3,  -1,  -6,  -9,  -2, -41,  -7,   0,   1,  -2,   5,   3,  11,  36,   6,  -1,  -1,  -5,  -4,  -6, -18,   0,   5,  -2,  -2,  -4,   4, -28,   8, -28,  -5,  -1,   0,  -3,  -9,  23,  15, -27,  -7,  -4,  -3,  -6,   8, -66,  -9,   2,  12,  -8,  -3, -32,  -4, -55,  -2,  18,   6,   0,  -6,  -8,  -1,  -8,   2,   1,   0 },
      {  -1,  -1,   0,   0,  -1,  -3, -35,  -7,   1,   3,   2,   3,   1,  16,  33,   6,   0,   0,   1,   2,   6,   1,  -5,  39,   1,   3,   1,  -6,  -2,   2, -39,  50,  -1,   4,   8,  24, -19,  43,   4,  13,   0,   4, -17,  -1,  18,  43,  -2,  -9,   1,  11,  34,  -1,  15,  12, -10,  -4,   4, -16,  17,   0,   2,  -3,  -4,   0 },
      {   1,   1,  -4,   1,  13,   0,   2,  43,  -2,  -3,  -1,   1, -16,  -3,  21, -37,  -1,   0,   4,  -7,   3,  -3, -19, -23,   1,   3, -10,  17, -10,  -2,  35,  41,   2,   2,  19, -11,  27,  -4,  59,  17,  -1,  -2, -25,   6, -13,  -4,  24,  10,   0,   3,  18,  -1, -23,   1,   5,   5,   1, -15,   8,   4,   2,   2,  -6,  -1 },
      {   0,  -1,   1,   1,  -4,  18,   6, -26,   0,   0,  -1,   1,  -1, -24,  -4,  17,   1,   3,   5,  -4,   3,  47,  -7, -10,  -2,  -2,  -7,   1, -11, -42,  10, -43,   2,   2,  13,  -5,  10,  20,  17, -29,   0,   2, -21,  30, -13,  45,   2, -15,   1,   1,  25, -22, -24,  17,   0,   0,   2,  -1,  12, -20,  -7,   1,  -3,  -2 },
      {   1,   3,   0,   6,   7,   8,  31,  12,  -2,  -3,  -1,  -9, -14, -14, -38,  -5,   2,   5,   1,  13,  17,  26,  36,   6,  -1,  -4,   1, -13, -20, -24, -13,  31,   1,   2,   0,  15,  19,  13, -22,  18,   0,   1,  11, -25, -18,  24,  23,  -7,   0,   1, -24,  32, -24,  -8,   0,  -3,   0,  11,  31,  34,  -2,  -5,  -7, -11 },
      {  -1,  -2,   0,   1,   1,  -9, -14, -28,   2,   2,   3,   0,   1,   8,  30,  27,  -1,   0,  -3,   1,   5,  -8, -11,  -7,  -1,   0,   7,  -3, -15,  11, -17, -32,   1,   0,  -6,  19,  23, -31,  53, -21,   1,   3,  16, -31,  14,   5,  42, -17,   0,  -2,  -3,  33,   5,   1,   7,  -6,   3,   8,  20,  32,  -6,  -8,  -4,   6 },
      {  -1,  -1,  -2,  -5,  -5,  -6,  17, -11,   1,   1,   2,   6,   7,  15, -28,  26,   1,   2,  -2,  -3,   3, -22,  53, -17,   0,  -4,   3,   3,  -3,  36, -34,  16,   0,   2,  -2, -19,   9, -25,  26,  23,   1,   4,  -5,  40,  -9,  16,  28, -10,  -2,  -2,  -6, -36,   8,  16, -10,   3,   0,  -5,  12, -26,  -3,   4,  -1,   2 },
      {   0,  -1,   3,   1,   6,  -2,   7,  -3,  -1,  -2,  -4,  -5,  -5,   1,  -8,   7,   1,   3,   0,   8,   0,   2,  24,   5,  -1,  -4,  -1,  -9,   6,  -2, -17,   0,   1,   5,   5,   7, -18,  17,  14,   1,   1,  -2,  -7,  -5,  52, -15,  27,   5,  -3,   1,  16,   2, -59, -19,  33,   2,  -6,  -8, -60,  -9, -31,   4,  15,   1 },
      {   1,   1,   0,   4,   5,   0,   3,  10,  -1,   0,  -2,  -2,  -8,   5,  -7, -26,   0,  -1,   0,   3,   9,  -6,  -2,  14,   0,   2,   1,   1, -11,  26,   6,   3,   0,   0,  -1,   1,  13, -41, -31, -23,  -1,   2, -11,   2, -10,  41,   6, -27,   1,   1,  22,  22,  -3,  53,   5, -28,  -6,   2, -57,   9,  10,  11,  -6,   8 },
      {  -1,  -1,   0,   0,  -5,   1,  -1, -11,   1,   3,   3,   4,   9,   6,   5,  39,  -2,  -4,  -1,  -3, -11,  -2, -23, -40,   1,   3,  -1,   6,   8,  10,  28,   0,   0,   0,   4,   6,  -4,  -4, -16,  62,   3,   3,  11,   0,  19,  44, -18,  36,  -1,   0, -11,  23, -24,  30,  12,   7,  -1,   2,   4,   7, -12,  14,   4,  20 },
      {   0,   0,   1,  -5,   1,   0,  -2,  18,   1,   0,   1,   6,   0,   3,   7, -30,   0,  -1,   1,  -7,  -8,   0, -11,  16,   0,   0,   4,   7,  10,  10,   3,  10,  -1,   0,  -6, -13, -24, -14,  -4, -30,   2,  -4,  20,   7,  34,   5,  -5, -29,  -1,   3, -38,   3, -37,  30,   9,  -7,   0,   8,  57, -33, -38,   5,   3,   4 },
      {   0,  -1,   0,   1,   0,   0,   3, -14,   0,   0,  -2,  -3,  -2,   1,  -7,  20,   0,   1,  -1,   3,   1,  -3,  -1, -28,   0,   0,   0,  -5,  -4,   1,  -3,  25,   0,   0,  -2,  11,  10,   6,  27,  -4,  -3,   0,  -5, -28, -17, -14, -51, -40,  -3,  -5,  -3,  37, -15,  -2, -53,  -7,  -1,  -3, -20, -42, -19,  17,   4, -16 },
      {   0,  -1,  -1,   0,   3,   0,  -5,   7,   0,   2,   2,  -2,   1,  -1,  10,  -3,   0,  -1,   0,   0,   3,   3,  -6,   5,  -1,   0,   0,  -1,  -3,  -6,   7, -28,   0,   2,   2,   3,  12,  23, -18,  44,   1,   2,   3, -10,   0, -18,  51,  -7,  -2,  -5,  -2,  28,  20,  20,  11, -30,   1,  -3,   1, -70,  13, -12,  -6, -35 },
      {   0,   0,   0,   1,  -1,  -5,   0,  20,   0,   0,   0,   0,  -2,   8,  -3, -28,   0,  -1,   1,  -1,   3, -11,   3,  34,  -1,   0,  -2,   1,  -4,  11, -10, -45,   0,  -2,  -1,   2,  -3, -27,  19,  27,   1,   3,   0,  -1,   1,  50, -25,  20,  -2,  -4,  -3,   9, -21, -46, -32,   7,  -2,  -2,  -5, -19,  -3, -41, -15, -12 },
      {   0,   0,  -3,   0,  -3,  -5,  -4,  -7,   1,   1,   3,   2,   3,   6,   9,  14,  -1,  -3,  -2,  -4,  -5,  -6, -17, -35,   1,   3,   1,   4,   1,   1,  26,  25,  -2,  -2,  -3,  -3,  -7,  -9, -29,   4,   1,  -2,  -1,   7,  -1,  25,  10, -63,   0,  -3,  -3, -20,   3, -60,  25, -24,   1,  -5,  -6,   3,  -8, -45,  -2,  -6 },
      {  -1,  -1,  -1,  -3,  -3,   1,  -4,  -1,   0,   0,   1,   3,   4,   2,   3,  -5,   0,   1,  -1,   2,  -6,  -4,  -4,   0,   0,  -1,  -2,   0,   9,   0,   0, -23,   1,   1,   2,   0, -11,  15,   3,  34,  -1,   0,  -5,  13,  12, -29,  -1, -22,  -1,  -5,  -5, -32, -29,  29, -45, -49,   1,   2,  -3,  58,  -8,  -8, -28, -30 },
      {   0,   0,  -1,   0,   4,   1,  -1,  11,   0,   0,   1,   2,  -7,  -2,   2, -11,   0,   0,   0,  -4,   6,   6,   1,  13,   0,   2,   1,   4, -10,  -5,   0, -14,   1,   1,  -3,  -2,  12,   6,   5,  22,   0,   0,   4,   1, -23,   6,  -5,   0,   1,   3,   1,  -8,  49,  -1,  11, -12,  -3,  -2, -11,  14,-100,  13,  28,   3 },
      {   0,   0,   0,  -1,  -1,   3,  -5,   0,   0,   0,   1,   1,   2,  -4,   8,   2,   0,  -1,  -2,  -2,  -3,   0, -13, -11,   0,   0,   4,   0,   4,   1,   9,   9,   0,  -2,  -3,  -2,  -7,  -3, -33, -23,   1,   3,   2,   4,   5,  11,  59,  27,   0,  -2,   2,   1,   1, -33, -66,   0,  -2,  -4,  -4,  -9, -27,  41, -37,  16 },
      {   0,   0,   0,   0,   0,  -2,   2,   7,   0,  -1,   0,   0,  -1,   2,   0, -15,   0,   1,   1,   0,   4,  -1,   0,  14,  -1,  -2,  -1,  -2,  -6,  -1,  -6, -26,   1,   2,   1,   4,   3,  11,   8,  35,  -1,  -2,   0,  -1,  -2,  -3,  -7, -40,   0,  -4,  -1,  -5,  -7, -32,   8, -29,   3,   2,  11,  -5,  29,  59,   3,  74 },
      {   0,   0,  -1,   1,  -2,  -2,  -2,  -1,   0,   1,   1,   1,   3,   4,   6,  -3,   0,  -1,   0,   0,  -1,  -2,  -3,   0,   0,   1,   1,   1,   2,   9,   0,  -4,   0,   0,  -2,  -1,  -3, -12,   3,  -3,   0,   3,   2,   7,   6,  24,  -6,   3,   1,   0,  -2,  -4,  -1, -29,  19, -11,   2,   2,   9,  11,  18,  75,  27, -84 },
      {   0,   0,   0,   1,  -2,  -1,  -1,   3,   0,   0,   0,   1,   1,   3,   5,  -4,   0,   0,  -1,  -1,   0,  -3, -10,  15,   1,   1,   1,   0,  -1,   3,  14, -17,   0,   0,  -1,   1,  -1,   2, -15,  27,   0,   0,   0,   1,   2,  -5,  22, -57,   1,   1,   5,  -2,  -4,  11, -21,  97,  -1,   2,  -7,  13,   0,   9,  17, -10 },
      {   0,   0,   0,   0,   1,   2,  -3,   0,   0,   0,   1,   0,   2,  -2,   6,  -2,   0,   0,   0,   1,  -2,   2,  -2,   0,   0,   0,  -1,   2,   2,  -1,   5,   2,   0,   1,   1,  -3,   3,   0,  -6,  -5,   0,  -1,   1,   4,   2,   5,  22,  11,   0,  -1,  -3,  -1, -11,   1, -47, -25,   1,   2,   4,   4,  18, -23, 108,  13 },
    },
    {
      { -72,  76, -46,  29, -16,   9,  -4,   2,  26, -30,  18, -11,   5,  -2,   0,   0,   1,  -6,   6,  -5,   3,  -2,   1,  -1,  -1,   3,  -2,   2,  -1,   0,   0,   0,   2,  -2,   2,  -1,   1,  -1,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -78, -32,  60, -29,  25, -11,   8,  -3,  53,  -6, -14,   5,  -6,   2,  -2,   0,   8,  17, -17,   8,  -6,   3,  -2,   1, -10,   0,   4,  -3,   2,  -1,   1,   0,   3,  -1,   0,   1,  -1,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   2,   1,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  52, -11,  -6,  -2,  -1,  -2,   0,   0,  76, -59,  30, -20,   8,  -4,   1,  -1, -44,  21,  -7,   6,  -1,   1,   0,   0,  -7,  10,  -7,   6,  -4,   2,  -1,   1,  -8,   2,   1,  -1,   1,   0,   0,   0,  -3,   2,  -1,   2,  -1,   0,   0,   0,  -4,   2,  -1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 },
      { -26,   0,  13,  -5,   6,  -2,   0,  -1, -61, -10,  37, -16,  14,  -4,   4,  -1, -67,  55, -26,  14,  -4,   0,   0,  -1,  28,  -9,   3,   1,  -4,   3,  -3,   0,  13,  -5,   3,  -4,   2,  -2,   0,   0,  13,  -1,  -3,   1,  -2,   0,   0,   0,   5,  -2,  -1,   0,   0,   0,   0,   0,   5,  -1,  -1,   0,   0,   0,   0,   0 },
      {  -2,  15,  -7,   1,  -3,   1,  -1,   1,  28,  43, -52,  28, -19,   8,  -5,   1, -86,  -9,  27,  -7,   8,  -1,   2,   0,   2,  -3,   7,  -5,   4,  -3,   2,  -1,  31,  -4,  -4,   1,  -2,   0,  -1,   0,   0,  -3,   2,  -2,   1,  -1,   0,   0,   3,  -2,   1,   0,   0,   0,   0,   0,   1,  -1,   1,  -1,   0,   0,   0,   0 },
      { -13,  -9,  21, -20,  17, -10,   6,  -2,  -9, -25,  19,  -6,   4,  -1,   2,   0, -15, -57,  51, -21,  13,  -2,   1,   2,  69,  18, -28,  11, -10,   2,  -3,  -1,  -8,  20, -16,   9,  -5,   1,  -1,   0,  -6,   3,   0,  -2,   0,   0,   0,   0,  -7,   1,   0,   0,   0,   0,   0,   0,  -4,   1,  -1,   1,   0,   0,   0,   0 },
      {  -3,  -7,  -5,  19, -17,  18, -15,   5,  22,  13, -19,   9,  -2,   0,   1,   0,   1,  17, -12,   5,  -1,  -4,  -1,  -1,  67, -48,  26, -14,   4,   0,  -3,  -1, -73,   8,   7,  -2,   6,  -4,   5,   0, -14,   5,   2,   0,  -2,   1,   0,   0,  -2,   1,   0,  -2,   1,   0,   1,   0,  -1,   1,  -1,   1,   0,   0,   0,   0 },
      { -13,  -8,  13,   0,   1,   7,  -9,   2, -19,  -1,  11,  -7,  11,  -5,   5,  -1, -33, -21,  29, -11,   9,  -6,   1,  -1, -69,  -5,  19,  -3,   4,   0,  -1,  -2, -72,  31,  -4,   1,   3,  -3,   4,   0,  29,   3,  -4,   2,  -5,   3,  -2,   0,  13,  -3,  -1,   0,  -1,   0,   0,   0,   7,   1,  -3,   1,   0,  -1,   0,   0 },
      {  20,  62,  -4, -49,  47, -42,  24, -14, -12, -20, -13,  30, -24,  18,  -9,   5,   0,   6, -11,  13, -12,  11,  -8,   4,   2,   0,   6,  -8,   9,  -6,   4,  -2, -36,  -5,   7,   0,   0,   1,   1,   0,   1,   2,   0,  -1,  -1,   0,   0,   0,   1,  -2,   2,   1,  -1,   1,  -1,   0,   1,   0,   0,   0,   0,  -1,   0,   0 },
      {  -7,  -8,  -8,  12,  -8,   1,   6,   0,   5,  27, -17,  -2,  -3,  -2,  -1,  -2,  -7,   8, -10,  13, -11,  11,  -4,   3,   9,  63, -60,  27, -17,   6,  -1,   2, -45, -32,  32, -16,  11,  -1,  -1,   0,  27,  -8,   5,  -4,   4,  -3,   2,   0,  14,  -2,  -3,   2,   0,   0,   0,   0,  -2,  -5,   4,  -1,   1,   0,   0,   0 },
      { -11, -12,  -3,   9,  -4,   5,  -4,   3, -21, -20,   1,  20, -16,  15,  -7,   5, -13,  15,   3, -18,  16, -12,   5,  -3, -22,  23, -11,   2,   2,  -2,   1,  -2, -18, -14,   7,   2,  -2,   3,   1,   0,-103,   7,  15,  -5,   6,  -3,   3,  -1,   8,   9,  -1,  -1,   0,   0,   0,   0,   9,   3,  -3,   1,   0,   0,   0,   0 },
      {   0,  18,   0,  -8,   8, -10,   6,  -7,   7,  43,   4, -41,  37, -28,  15,  -9, -12, -26, -11,  44, -37,  27, -15,   6,  -8,   2,   0,   3,  -3,   4,  -4,   1,  -7,  20, -17,   2,   3,  -2,   3,  -1, -55, -12,  10,   0,   0,   1,   0,   0, -16,  -6,   9,  -3,   2,   0,   0,   0,   2,  -2,   1,   0,   0,   0,   0,   0 },
      {  -3,   2,  20,   4, -26,  30, -27,  13, -21, -46, -34,  32,  -3,  -2,   6,  -1, -10, -19, -22,  29, -19,  12,  -7,   3,  -2,  30,   1, -11,  11,  -6,   1,  -3,   1,  13,  -3,   2,   1,  -4,   5,  -3,  10,  -3,   1,  -3,  -1,   0,   0,   1, -67,  -2,  12,  -4,   3,  -1,   2,  -1,  -4,   7,   2,  -1,   1,   0,   0,  -1 },
      { -13, -38, -26,  15,   2, -12,  18,  -5,  -9, -24,   4,  -4,  -4,   7,  -6,   3, -15, -52,   3,  28, -25,  24, -12,   9,  -1, -18,  20, -21,  16,  -9,   9,  -4,   7, -39,  38, -19,  14,  -6,   2,  -2,  -3,  29, -16,   8,  -4,   0,  -1,   1,  19,  14, -12,   5,  -4,   1,  -1,   0, -17,  -4,   3,   0,   0,   0,   1,   0 },
      { -17, -41, -51,  31,  13, -24,  26, -15,  -1,   7,  12,  10, -31,  29, -21,  11,  -3,  15,  -9,   9,  -8,  10,  -8,   4,  -1,   7, -10,   2,   2,  -5,   6,  -4,  -5,  55, -32,  12,  -8,   5,  -3,   0,   7,  -9,   4,   0,   2,  -1,   1,   0, -21, -13,  13,  -7,   4,  -1,   0,   1, -19,  -3,   3,   0,   1,  -1,   1,   0 },
      {  -5,  -7,  -6,   3,  14, -16,  13,  -7,   5,  19,  16, -14,  -3,   8,  -8,   3,  -3,  -4,   7, -16,  16, -11,   5,  -2,  -9, -22,  -1,  11,  -9,   5,   0,   0,  -7, -31,  28, -16,   5,  -1,  -2,   2,   3,   0, -10,   5,  -2,  -1,   2,  -1, -89,   7,   8,  -1,   3,   0,   0,   0,  51,   8,  -6,   3,  -2,   1,  -1,   1 },
      {  -3,   4,  -2,  -8,   6,  -5,   5,  -3,  -4,   5,  -8,  -4,  13, -17,  12,  -7,  -2,  20,  14, -20,  12,  -7,   2,   0, -11,  -9,  -1,   9,  -8,   9,  -5,   1,  -8, -19,   6,   1,   0,   2,  -3,   1,  -8,  21, -14,   4,  -2,  -1,   0,   1, -38,  -8,   5,   0,  -1,   0,   0,   0,-104,  -7,  15,  -3,   3,  -2,   3,  -1 },
      {  -9, -23, -41, -13,  10,  -8,  12,   1,  -5, -19, -38,   3,  31, -41,  32, -18,  10,  43,  42, -17, -10,  18, -16,   9,   4,  11,  -5, -24,  21,  -9,   6,   1,  -2,  -9, -10,  16,  -8,   5,   1,  -3,   5,  11,  -2,   4,  -5,   4,  -5,   3,  -4,   2,   1,  -2,   2,  -1,   0,   0,  33,   2,   0,   1,  -2,   1,  -1,   0 },
      {   4,   9,  19,   9,  -4,   2,  -3,   0,  10,  31,  38, -16,  -4,   9, -12,   6,   1,   6,   6,  -9,   1,  -6,   2,  -2,  13,  51,  20, -57,  49, -36,  19, -11, -10, -26,  -5,  16, -13,   9,  -1,   3,   3,  -7,  -8,  11,  -9,   5,  -3,   2, -10,  -8,   4,  -3,   2,  -2,   0,   0, -11,   1,  -2,   1,  -1,   1,   0,   0 },
      {   3,  11,   6,  -6,  -2,  13, -17,   7,   8,  22,  22,  19, -13,   6,  -1,   0,   9,  20,  20,   9, -14,   7, -14,   7,   3,  20,   0,  11,  -7,   7,  -4,  -8,   9,  33,   6, -19,  13, -13,   7,  -1,  -9,  81, -51,  13, -11,   3,  -1,   0,  -1,  -7,   5,  -3,  -2,   1,   0,   0,  16, -11,   2,   0,   0,  -2,   2,   0 },
      {  -1,   0,   5, -16,  -2,  20, -29,  13,   4,   7,  40,  24, -17,  11,  -8,   2,   7,  22,  56,  10, -34,  29, -25,  11,  -8, -34, -22,   9,   8, -12,   7,  -7,   3, -15,   5,  -8,   8, -11,  10,  -4,   1, -35,  30, -17,   8,   1,  -2,   1,  -6,   1,  -4,   3,  -1,  -1,   2,  -3, -13,   8,  -5,  -2,   1,  -1,   0,   0 },
      {   8,  18,  46,  39, -35,  -9,  31, -20,  -6, -11, -14, -12,  -4,   6, -13,   7,  -3,   5,   7,  -8, -20,  23,  -8,   6,  -9, -37, -39,  12,   8, -11,  15,   1,  -8, -20, -26,  30, -19,  16, -12,   3,   1,  30, -10,   6,   4,  -2,   1,   0,   2, -13,   5,  -4,   5,  -1,   0,   1,  11,  -4,   3,  -1,   1,   2,   0,  -1 },
      {  -7, -13, -33, -34,  17,  21, -31,  20,   5,  13,  22,  16,  -6,  -8,  10,  -6,  -9, -25, -34,  12,   4,  -8,  10,  -8,  -6, -21, -29,   7,   7,  -8,   1,   1,  -8, -29, -16,  38, -37,  23,  -9,   3,   6,  33,  -3,  -6,   7,  -5,   0,  -1,   5, -18,  17,  -6,   5,  -3,   3,  -1,   5,  12,  -4,   0,   1,   0,  -2,   0 },
      {  -4,   4, -11, -44, -34,  10,  12,  28,  -3,   8,  -3, -33, -34,   5,   2,  -9,  -4,   2, -15,  -9,   8,  15,  10,   6,  -4, -11, -30, -19,   9,  -5,  16,  14,  -4,  25,  -3,   2,   5,   5,  -6,  -7,  -2,  -1, -10,  15,   0,  -2,  -1,   3, -12,  58, -39,  13,   4,  -6,   0,   1,  -4, -20,  11,   3,   1,   4,   0,   0 },
      {  -4,  -7, -15, -39, -14,  14,  -8,   9,  -8, -15, -35, -53,   2,  36, -34,  19,   6,  11,  29,  28,  -3, -15,  21, -16,   0,  -3,  -1,   3,  -1,  -4,   6,   3,  -1,  -8,  -5, -12,   8,   0,   0,  -1,  -6,  -8, -14,  22,  -7,   1,   1,  -2,   6, -49,  23,  -7,   7,   0,  -1,   1,   1,   5,   0,   0,  -1,   1,   0,  -1 },
      {  -3,  -4,  -6, -13, -24,   4,   9,  -6,   0,   2,   8,   5,  -5,  10, -14,   7,   4,   8,  25,  31, -10,   0,  -1,   1,   2,  24,  52,  10, -38,  31, -21,   3, -11, -38, -39,  33, -13,  11,  -5,   5,   8,   6,   1, -13,  13, -10,   6,  -3,  -6,  37, -11,   2,   0,  -4,   0,   0,   5, -19,   8,  -1,   1,  -1,   2,   1 },
      {   5,   7,   6,  41,  72,  16, -38,  24,  -6,  -7, -28, -45, -11,  13,  -7,   6,   2,   7,   9,   3, -13,  -1,  -6,  -2,   1,   2,   4,  19,   9,  -9,   1,  -9,   1,   3,   3,   8, -21,  13,  -6,   4,   3,  10,   4, -10,   4,  -2,   1,   0,  -5,  29, -18,   5,  -6,  -4,   2,  -4,  -1, -24,   7,   0,  -1,   0,   1,   1 },
      {   1,   2,   5,  17,  11,  -2,   7, -12,   1,   0,   6,  36,  21, -10,   3,  -2,   4,   5,  25,  63,  11, -41,  32, -25,  -4, -14, -35, -10,  19,   1,  -1,  -2,   0,   3,  -8, -15,  -6,  13, -10,  10,  -3, -15, -14,  27, -19,   9,  -3,   0,   0,  32,  -6,  -3,  -1,   1,  -4,   0,   0, -19,  10,  -5,   2,  -4,   1,  -1 },
      {  -1,   2,   0, -14, -22, -11,   8, -13,   2,   5,  -1,  -6,  25,   4, -11,  10,   3,   7,  13,  19,   0, -20,  16,  -8,   1,   2,  -1, -10,  11, -12,  10,  -8,   6,  31,  44,   0, -16,  13, -16,   7,   7,  43,  51, -53,  28, -17,   4,  -3,  -9,  -3, -15,   6,  -3,  -2,   2,  -2,  -2,  -2,  -4,   3,  -4,   3,  -1,  -1 },
      {  -2,   1,  -8,  -4,   4,  -2, -13, -18,   2,   8, -11,  -1,  44,  24, -22,  27,  -2,   3, -12, -11,  -6,  -5,   1,  -6,  -1,   2, -14,   7,   3, -15,   2,  -7,  -1,  -7, -36,   1,  21, -23,  22,  -6,   7,  26,  11,   4, -13,   9,  -7,   1,  -1,  51,  -9,   4,  -5,   5,  -1,  -1, -11,  61, -21,  -1,  -1,  -2,  -1,   0 },
      {  -1,  -3,  -7, -14,   5,  16,  -3,  -1,   3,   0,   1,  16,  51,  14, -37,  27,  -3,  -8, -19, -28,   1,   2, -19,   9,  -3,  -9, -19, -20,   6,  15,  -8, -14,   1,  -4,   0,  12,  11, -17,  12,  -1,   0,  -6,   1, -11,  -4,   0,   4,  -1,  -2,  -3,  12,  -2,  -5,   2,  -3,   0,   8, -75,  27,  -6,   1,  -4,   5,  -1 },
      {   3,   2,   7,  18,  16,  -8,   6,  21,  -1,  -1,   1, -14, -28,  -5,  11, -11,   2,  -1,   3,  15,  14,  -9,  15,  -2,  -5,  -4, -21, -46,   3,  24, -20,  18,  -1, -20, -43, -14,  38, -29,  24, -12,   7,  33,  40, -18,   9,  -4,   3,   1,  -5, -21,  -3,   1,  -1,   1,  -3,   3,   3, -10,  -1,   6,   0,   0,   1,  -2 },
      {   2,   4,   6,  13,  10,  -6,   4,  33,   1,   3,   3,   7,  17,  21, -14,   0,  -1,   2,   8, -13, -34,   9,  10,  -8,  -3,  -1, -15, -51, -27,  42, -29,  32,   4,  14,  38,  26, -20,  12,  -1,   0,   0,  -4,  -4,  18,  -9,  -3,   5,  -2,   3,   5,   1, -16,  15,  -9,   3,  -1,  -5,  33,  -3,  -2,   0,   1,  -2,   3 },
      {   1,   3,   7,   1,  -8, -12,  -8,   1,   0,   0,  -5, -15,  -1,  13,   3,  -4,   4,   3,   5,  35,  69,   9, -46,  31,  -2,  -3,  -6,  -2,   2,   6,  -5,  -7,   5,   2,  22,  18, -10,  -3,  14, -10,   4,   5,  14,   5, -13,   9,  -7,   5,   8,  24,  36, -40,  20, -13,   8,  -4,  -5,  -1, -14,   7,  -4,   2,   1,   2 },
      {   2,   3,  -2,  -6,  23,  77,  35, -56,  -2,  -1,  -2, -14, -25, -23,  -9,  13,   0,   0,   2,   6,   3, -11, -12,   9,  -1,  -1, -11, -14,  15,  28, -14,  -5,   3,   6,  15,  16,  -2, -10,   5,  -2,  -1,  -5, -16, -11,  10,  -6,   2,   1,   1,   2,  -4,   1,  -6,   3,  -7,   3,  -2,  24,  -9,   7,  -3,  -3,   2,   1 },
      {   1,   1,   0,   8,   5, -10,  -4,  27,   0,   1,  -2,   9,  22,  -7, -25,  17,   3,   3,   0,  24,  48,  16, -27,  24,  -2,  -1,  -8, -17, -13,   7,   8,  -2,  -3,   0, -11,  -7, -15,  10,  -7,  -3,  -5, -10, -32,  -5,  19,  -9,   6,  -1,  -6, -30, -48,  44, -25,  12,  -3,   2,   1,  24,  -6,   4,  -1,  -1,   1,   1 },
      {  -3,  -5,  -3,  -1, -23, -49, -44, -24,  -1,  -1,  -4, -13, -20,  -2,  12,   9,   0,   0,   3,   5, -12, -21,  -3,  -5,   0,  -2, -11,  10,  37,  19, -36,   2,   3,   3,  19,  47,   0, -19,  27,  -9,  -3, -11, -23, -11,  -9,   2,   3,  -3,  -3,  -8, -20,  17, -17,  17,  -5,   0,   2,  -3,  -1,  -4,   4,  -3,   3,  -2 },
      {   0,   1,   2,   4,  -1, -13,   5,  21,   0,  -1,  -2,  -3,   9,   6,   1, -15,   0,   1,   2,  -2,  -8, -17,   1,   0,   0,  -3,  -7,  -6,  17,  -5,   3,  -1,  -2,   0, -10, -12,  14,  -4,   3,  -7,  -1, -18, -60, -49,  53, -37,  23,  -6,   9,  27,  40, -21,   4,  -1,  -2,   3,  -4,  11,  -3,   4,  -5,   1,  -2,  -1 },
      {   0,  -1,   0,  -2,   0,  13, -18, -49,   2,   1,  -2,  -7,  15,  73,  48, -46,   0,   0,  -1,  -3,   5,   7, -14,   3,  -2,  -2, -10, -26, -16,  -5, -11,   2,   0,  -2,  -8, -19, -14,  13,  -1,   7,   0,  -1,  -3,   2,   4, -17,   5,  -1,  -1,  -5,  -5,  17, -15,   7,  -1,  -1,   2,  -5,   8,  -9,   5,  -6,  -2,  -1 },
      {   2,   2,   2,   8,  12,  22,  40,  22,   2,   5,   6,  10,  21,  31,  14, -19,   2,   3,   5,  17,  21,  22,  26, -16,   2,   3,  13,  33,  13, -26,   4,  25,   1,   3,  17,  45,   8, -32,  31, -26,  -3,  -1,  -7,  -8,   2,  -3,  -4,   1,  -4, -12, -17,   3,   5,   0,  -2,   5,   3,   2,  20, -18,   6,  -2,   0,  -3 },
      {   1,   1,   0,   4,   3,  -2,   6,  11,   0,  -1,  -4,  -4,  -7,   5,  20,  -7,  -1,  -1,   0,   8, -19, -45,  -1,  17,  -3,  -5,  -8, -23, -60, -12,  48, -48,   2,   0,  10,  43,  21, -18,   3,   1,  -2,  -6, -10,  -7,  -3,  15, -18,  12,   0,  -5,  -8,  -3,  -9,   6,   0,   1,   1,  -1,  -5,   4,   0,  -8,   6,  -4 },
      {   0,   0,   2,   1,   0,  -7, -12,  -7,  -1,   0,   2, -10, -16, -26, -17,   2,   1,   2,   4,   3,  11,   6,  -2,  -5,   0,  -1,  -1, -16, -20, -14,   8,   1,   1,   4,   8,   0,  -7,  -8,   7,   2,   2,   5,  12,   6,   6,  -8,   8,  -3,   6,  16,  38,  17, -21,  11,  -4,   2,   9,  29,  77, -50,  29, -15,   1,  -1 },
      {  -1,  -1,  -1,   3,   9,  -5, -35, -39,  -1,  -1,  -1,   2,  -6, -19,   3,  29,   2,   4,   4,   4,  20,  52,  34, -32,  -3,  -1,  -6, -26, -41, -29,   8,  17,   0,   1,   7,  11,   4, -21,  -1,   5,  -2,  -4, -20,  -7,  22,   4,  -1,  -7,   0,   3,  -2, -14,  14,  -6,   6,   0,  -3, -20, -21,   9,  -8,   9,  -9,  -1 },
      {   0,  -1,  -2,  -2,  -3,  -1,   2, -11,   0,  -1,  -1,  -2,  -3,  -8,  -3,  11,  -1,  -3,  -3,  -2, -12, -35, -21,  13,  -2,  -2,  -3,  -5, -17, -29, -12,  24,  -1,  -5, -10, -32, -35,  -2,  34, -17,  -2,  -3, -16, -30,  -8,  18, -11,   6,   0, -11, -39, -21,  49, -30,  16,  -4,   2,   2,  26, -15,   7,   2,  -5,   1 },
      {   0,   0,   1,   0,  -7,   5,  25,  22,   1,   1,  -1,  -5,  -6,  10,  32,  42,   1,   2,   2,   2,  -3,   8,   9,  -2,  -1,  -2,  -3,  -7, -10, -13, -24,   9,  -1,   0,   1, -16, -32,   4,   5,  -5,  -2,  -2, -11, -40, -24,  37, -30,  15,   2,  12,  44,  36, -32,  15,  -7,   2,  -2,  -7, -14,  10,   1,   2,   4,   6 },
      {   0,   1,   1,   0,   0,   6,  12,   7,   0,   1,   1,   3,  -1,  -9,  10,  23,  -1,  -1,   2,   2, -10, -35, -36,  -1,  -1,  -1,   0,   0, -19, -50, -21,  33,   2,   4,   8,   8,   3,  10,  25, -11,   3,   5,  16,  52,  34, -21,  24, -15,  -2,   3,   8,   7, -18,  17,  -9,   5,  -3, -18, -27,   2,  -3,   7,  -6,   4 },
      {   0,   0,   0,  -1,   2,   0,   2,   8,   3,   3,   3,   1,   1,  16,  67,  81,   0,   0,   2,   0,   3,   5,   6,  13,   1,   0,   2,   3,  15,  24,  13, -14,   1,  -1,  -3,  -4,  14,  10,  -7,   7,   1,  -2,   4,  14,  14, -21,   7,  -2,   0,  -6, -15, -26,  12,  -2,   3,  -3,   3,  11,  27, -16,   4,   0,  -3,   9 },
      {   0,   0,  -1,  -1,   4,   6,   0,   1,   0,   0,   1,  -1,  -1,   3,  -1,   3,   0,  -1,   0,   3,   8,  -5, -27,  -7,  -1,  -1,  -2,  -3,  -2, -30, -33,  28,   2,   2,   1,  23,  73,  38, -47,  26,  -2,  -1,  -9, -30, -21,  18,  -8,  -8,   0,  -5,  -7,  -7,  -2,  -7,   6,  -6,   3,   9,  11, -10,   8,  -4,  -1,   4 },
      {   0,   0,   1,   4,   8,  -1, -22, -22,   0,   1,   3,   7,   7,   2,  -7,   1,   1,   2,   3,   3,  -7, -11,  37,  74,   1,   2,   3,   5,   3,   4,  23,  47,   0,   1,   3,  10,  20,  20,   0, -25,   0,   1,   1,   5,   7,   3, -11,  20,   1,   2,  20,  25,   3,  -3,   6,  -5,   1,  -2,  19,  20, -18,  19, -14,  21 },
      {   0,   0,  -2,  -1,   1,   1,   2,   5,   0,  -1,  -2,  -2,   2,   0,  -1,   0,   1,   1,   1,   1,  -3,  -2,  33,  66,  -1,  -1,  -2,   0,   2, -24, -34,  -2,  -2,  -2,  -2,  -2, -17, -48, -47,  12,   0,  -1,   0,   5,  -5,  -2,  17, -20,  -1,   0,  -4, -12,   0,   3,   6,   2,   0,  -5, -28, -31,  29, -20,  13,  -9 },
      {   0,   0,   0,   1,   2,   4,   0,  -3,   0,   0,  -1,  -1,  -4, -12,  -5,   3,   1,   1,   1,   0,   2,  16,  28,  17,   0,  -2,  -1,  -2, -13,  -7,   3, -20,   2,   1,   1,  -3,  13,  59,  43, -27,  -1,  -2,  -2, -21, -54, -33,  34, -13,   0,  -1,  -3, -10,  -3,   7, -13,  14,  -1,  -6, -25, -15,  22, -11,   0,  -9 },
      {   0,   0,   0,   1,   2,  -1,  -3,  -4,   0,   0,   0,   0,   2,   0,  -1,   3,  -1,  -1,  -1,  -2,   0,  -3,  -4,  -4,   2,   3,   3,   0,  11,  27,  25,  20,   0,   1,   1,  -1,  12,  24,  14,  -3,   0,   2,  -3,   6,  36,  15, -42,  -1,  -1,   2,  10,  21,   6, -24,  19,  -8,   0, -13, -45, -55,  33, -34,  16, -29 },
      {   0,   0,   0,   4,   2,   0,   3,   1,   0,   0,   0,   2,  -3,  -2,  12,  11,   0,   0,   2,   4,   0, -11, -17, -10,   0,  -1,  -2,  -1,  -2,   4,  20,   9,   0,   0,   0,   9,   0, -29,  -8,  16,  -1,  -2,   0,  -6, -45, -45,  21, -27,   3,   2,  21,  64,  32, -41,  27, -14,  -2,  -2,  -9,   2, -10,   7, -10,  -9 },
      {   0,   1,   1,   1,  -1,  -1,   7,   3,   0,   0,  -1,   0,   1,  -2,  -2,  -2,   1,   2,   0,   0,   2,  11,  24,   9,  -2,  -3,  -4,  -2,  -1, -21, -59, -62,   2,   3,   3,   5,  15,  20,  21,   3,   1,  -1,  -1,  21,  31,   3,  -5,  19,   1,  -2,   6,  33,  15, -48,  19, -11,   2,   7,  13,  11, -11,   1,   5, -11 },
      {   0,   1,   1,   1,   2,   2,  -1,   0,   1,   1,   1,   0,   2,   6,   9,   0,   0,   0,   1,   1,   1,   1,   0,   4,   0,   1,   1,  -2,   1,   5,   5,  -7,   0,   1,   3,   1,   3,  13,  -3, -15,   2,   3,   3,  -2,  16,  56,  32, -57,   2,   5,   6,  27,  57,  41, -13,   8,   0,   3,  14,  18, -13, -24,  12, -26 },
      {   0,  -1,  -1,  -2,  -2,   1,   5,   1,   0,   0,  -1,  -1,   2,   4,   4,  -5,   0,   0,   0,   1,   1,   1,   6,   7,   1,   0,  -1,   4,   5,   8,  20,  10,   1,   1,   0,   2,  -2,   1,  35,  51,   1,   0,  -2,  -5,   9,  52,  60, -16,  -1,  -2,  -1,  -4, -34, -44,   7,  -6,   0,   1,  -4,   7,  20,   1,  -1,  38 },
      {   0,   0,   0,   0,   1,   1,   3,   7,  -1,  -1,  -1,  -1,   0,  -1,  -6,  -8,   0,   0,   0,   1,   3,   3,   6,  13,   0,   0,  -1,   1,   1,  -2,  -5,  -5,   1,   2,   2,   2,   4,   2,  33,  77,   0,   0,  -2,  -3,   5, -14, -24,  12,   0,   2,   1,  15,  50,  49, -22,  -1,  -2,  -3, -11, -17,  16,  27, -15,  27 },
      {   0,   0,   1,   1,   2,   0,  -1,  -3,   0,   0,   0,   0,   1,   2,   1,  -4,   0,   0,   0,   0,   0,  -3,  -8, -19,   0,   0,  -1,  -4,   0,   8,  -3, -25,  -1,  -1,  -1,   2,   5,  -4, -31, -56,   0,   1,   0,   0,   7,  20,  24,  17,   0,   0,   3,  20,  24,   0,  -9,  -5,  -1,  -5, -20, -32,  37,  36, -27,  61 },
      {   0,   0,   1,   1,   1,  -2,  -5,  -3,   0,   0,   0,   2,   4,   0,   2,   6,   0,   0,   1,   0,   0,   0,  -6,  -9,   0,   0,   2,   3,   1,  -6,   3,  13,   1,   1,   2,   2,   2,  -5,  -6,   0,   0,   0,   2,   4,   1, -17,   9,  43,   1,   1,   1,   9,  23,  -2, -13,   4,   2,   0,   5,  66,  73, -49,  26,   9 },
      {   0,   0,   0,   0,   1,   0,  -4,   0,   0,   0,   0,   0,   0,   1,   4,   6,   0,   0,   0,   1,   1,  -1,  -6,  -1,   0,   0,   1,   1,   0,   0,   4,  13,   0,   0,   1,   3,   1,  -4, -10,  13,   0,   2,   1,  -3,  -4,  15,  52,  79,   0,   1,   2,   8,  14,  -8, -38,  19,  -1,  -1, -10, -33, -38,  -6,   5, -40 },
      {   0,   0,   0,  -1,   0,   4,   9,   2,   0,   0,   0,  -1,  -1,  -2,  -2,  -1,   0,   1,   1,   0,   0,   2,   3,  -6,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,  -1,  -1,   0,   4,   6,  -2,   0,   0,  -1,  -1,  -3,   6,  32,  33,   1,   1,  -2,  -6,  -2,  54,  78, -68,   0,  -3,  -2,   0,   0,  -6,  -8, -14 },
      {   0,   0,   1,   0,  -1,  -1,  -1,  -1,   0,  -1,  -1,  -1,  -2,  -3,  -1,  -2,   0,   0,  -1,  -1,   1,   3,   3,  -6,   0,   0,  -1,  -2,  -3,  -6,   0,   2,   0,   0,  -1,  -1,   0,   2,   2,  -4,   0,   0,   1,  -1,  -5, -13,   0,   1,   0,  -1,  -1,   4,  11,   5,   4, -11,   0,  -1,  -6, -21, -51, -40,  77,  72 },
      {   0,   0,   0,   0,   2,   1,  -2,  -2,   0,   0,   1,   1,   1,   2,   0,   1,   0,   0,   1,   1,  -1,  -1,  -3,  -3,   0,   0,   0,   0,   2,   2,   0,   0,   0,   0,   1,   1,   3,   3,   0,   5,   0,   0,   1,   0,   1,   3,   7,  11,   1,   2,   1,   4,   2,  15,  70,  94,   1,   1,   1,   0,   9,  32,  32,   0 },
      {   0,   0,   0,  -1,  -2,  -1,   2,   4,   0,   0,  -1,  -2,  -1,  -1,  -3,  -1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   1,   2,   0,  -5,  -5,   0,   0,   0,   0,  -1,  -2,   0,   5,   0,   0,   0,   1,   0,  -1,  -5,   6,   0,   0,  -1,   2,   1,   4,  22,  39,  -1,  -5,  -3,  -3, -25, -78, -81,  30 },
    },
  },
  { // 18
    {
      {-117,  20,  15,   6,   5,   2,   2,   1,  22,  -3,  -2,  -1,  -1,   0,   0,   0,  38,  -5,  -4,  -2,  -2,  -1,  -1,   0,   1,  -1,  -1,   0,   0,   0,   0,   0,   5,  -2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  -5,   2,   2,  -1,   0,   0,   0,   0,-111,  16,  12,   6,   5,   2,   2,   1,  47,  -6,  -5,  -2,  -2,  -1,  -1,   0,  34,  -3,  -3,  -2,  -1,  -1,  -1,   0,   2,  -2,  -1,   0,   0,   0,   0,   0,   6,  -1,  -1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      {  39,  -6,  -1,  -3,  -1,  -1,  -1,  -1,  30,  -3,  -4,  -2,  -1,   0,   0,   0, 105, -12, -12,  -5,  -5,  -2,  -2,   0, -39,   5,   3,   2,   1,   1,   1,   0, -29,   1,   2,   1,   1,   1,   1,   0,  -6,   1,   1,   1,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,  -3,   1,   0,   0,   0,   0,   0,   0 },
      {  14,  11,  -2,  -1,  -2,  -1,  -1,   0,  37,  -4,  -1,  -1,  -1,  -1,  -1,   0,  15,  -9,  -2,  -1,   0,   0,   0,   0, 110, -12, -13,  -4,  -4,  -1,  -1,   0, -27,   4,   3,   1,   1,   0,   0,   0, -31,   2,   3,   1,   1,   0,   1,   0, -10,   1,   1,   0,   0,   0,   0,   0,  -7,   1,   1,   0,   0,   0,   0,   0 },
      { -17,-116,   1,   2,  14,   7,   6,   2,   5,  12,  -1,   0,  -2,  -1,  -1,   0,   8,  41,   6,   0,  -4,  -3,  -2,  -1,  12,  -1,  -2,  -1,   0,   0,   0,   0,  -3,   7,  -2,   0,  -1,   0,   0,   0,  -6,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0 },
      {  16,  -3,   4,  -2,   1,   0,   0,   0,  13,  -2,  -5,  -1,  -1,   1,   0,   0,  23,  -5,  -2,  -2,   0,  -1,   0,   0,   9,  -1,  -2,   0,   0,   0,   0,   0, 117, -13, -14,  -3,  -4,  -1,  -1,   0, -30,   4,   3,   1,   1,   0,   0,   0, -14,   0,   1,   1,   1,   0,   0,   0,  -7,   1,   1,   0,   0,   0,   0,   0 },
      {  -8,   5,  -1,  -1,  -1,  -1,   0,  -1, -21, -11,   1,   2,   2,   1,   1,   0, -11,   6,   3,   0,   0,   0,   1,   0, -25,   6,  -3,   1,   0,   0,   0,   0, -19,   3,   1,   0,   1,   0,   0,   0,-117,  10,  11,   2,   4,   1,   2,   0,  19,  -2,  -1,  -1,  -1,   0,   0,   0,  10,   0,   1,   0,   0,   0,   0,   0 },
      {   3,   9,  18,  -1,   1,  -4,   0,  -1,  12, 115,  -4,  -4, -12,  -8,  -5,  -3,  -4,  -7,  -3,  -1,  -1,   1,   1,   0,  -9, -43,  -2,  -1,   3,   3,   2,   1,  -3,  -8,  -2,   1,   1,   1,   0,   1, -12,  -6,   1,   1,   1,   1,   1,   0,  10,  -2,   0,  -1,   0,   1,   0,   0,   0,  -1,   1,   0,   1,   0,   0,   0 },
      { -10,  13,-114,   7, -12,  12,   0,   5,   3,  17,   6,  -3,  -1,  -3,  -1,  -1,  12,  28,  34,   0,   1,  -4,  -1,  -3,  -2,  -6,   0,   1,   0,   1,   0,   0,   4, -15,   4,  -1,   2,   0,   0,   0,  -1,  -1,   1,   1,   1,   1,   0,   0,  -8,  -1,   3,  -1,   1,   0,   0,   0,  -1,  -1,   0,   0,   1,   0,   0,   0 },
      { -10, -34, -32, -16,  -3,   8,   3,   3,   0,   5,  -8,   0,   2,   4,   0,   2,  -8,-106,  11,  10,  14,   3,   2,   3,  -1,   5,   2,   0,  -1,  -1,   1,   0,   6,  35,   9,   2,  -3,  -3,  -1,  -2,  -1,  -1,   3,  -1,  -1,  -2,  -1,  -1,  22,   5,  -3,   2,  -1,  -2,   0,  -1,  -1,   2,   0,   0,  -1,   0,   0,  -1 },
      {   7,   6,  -3,   3,   0,   2,  -1,   0,   7,  -7,   3,  -1,   0,   1,   0,   0,  10,  18,  -1,  -5,  -2,  -2,   0,   0,  14,   5,  -3,  -3,  -1,  -1,   0,   0,  13, -11,   0,  -1,   1,   0,   0,   0,  13,  -4,  -1,   0,   0,   0,   0,   0, 122,  -9, -10,  -1,  -3,  -1,  -2,   0,  -3,   1,   0,   1,   0,   0,   0,   0 },
      {  -3,  -2,  -3, -50,  -5,  -4,   4,  -1, -11, -27, -73,  -3,   2,  10,   2,   3,   2,   8,   4,  16,   5,   2,  -2,   0,  -2, -72,  26,   9,   6,   1,   1,   1,  -5, -19,  -1,   8,   1,   1,   0,   0,   3,  15,   7,   0,  -2,  -2,  -1,  -1,   6,   7,   0,   2,   0,  -1,   0,   0, -16,   5,   3,  -1,   0,  -1,   0,  -1 },
      {   3,   1, -10,  67,  10,   7,  -4,   0,  -7,  12, -86,  -4,  -8,   4,   1,   1,  -3,  -4, -26, -20,  -8,   0,   3,  -1,  10,  36,  29,   3,   3,  -2,  -2,  -2,   0,  -2,   7,  -3,   3,  -1,   1,   1,  -1,  -4,   5,  -1,   1,   1,   0,   1,  -1,  -3,   3,  -4,   0,   1,   0,   1,  19,  -5,   0,   0,   1,   0,  -1,   1 },
      {  -5,   7,   1, -73,  -8, -13,   4,  -2,   0,  28, -28,  -3,  -2,   3,  -1,   1,   1,  17,   2,  20,   5,   3,  -1,   1,   8,  81,  11,  -4,  -4,  -5,  -1,  -1,  -2,  -4,   2,   3,  -1,   0,   0,   0,  -4, -24,   0,   0,   1,   0,   0,   0,  -3,  -5,  -1,   3,   0,   0,   0,   0, -29,  -5,   3,   1,   1,   0,   1,   0 },
      {   1,   3,   6,  38,   7,   4,  -3,  -1,  -7,  -2,  -3,   2,  -2,  -1,   0,   0,  -3,  -4,  23,  -8,  -3,  -3,   1,  -1,  -8,  -1,  -5,   2,   0,   1,   0,   0,  -5,  10,  -5,  -3,   0,   0,   0,   0,  -5,   1,   0,   0,   0,   0,   0,   0,   0,   0,  -2,  -2,  -1,   0,   0,   0,-117,   6,  10,   1,   3,   0,   1,   0 },
      {  -6,  -6, -32,  -9,  -6,  -5,   1,   1,   0,  -1,  16,   3,   2,  -2,  -1,  -2, -11,   5,-112,   5,  -4,  10,   2,   2,  -3, -11,  -9,   0,   2,   2,   0,   1,   5,   7,  29,   3,   5,  -1,   0,   0,  -3,   0,   1,   1,   1,   0,   0,   0,   0,  -1,   6,  -3,   0,   0,   0,   0, -28,   1,   3,   0,   1,  -1,   0,   0 },
      {  -3, -18,  -2,   6,  11,  -4,   3,  -2,   1,  -4,  12,   5,   2,  -4,  -3,  -3,  -8, -36,  -7,   2,  -5,   5,   1,  -1,   1,  14,  -6,   1,   0,   1,  -2,  -1, -15,-114, -14,   8,   8,   6,   2,   4,  -3,   1,  -2,   2,   1,   2,   1,   1,  -2,  15,   4,  -3,   1,   1,   0,   0,  -7,   3,   2,   0,   1,  -1,  -1,   1 },
      {  -1,  -3,  11,  -9, -21,  33,  -4,   6,  -1,   0,  25, -71,   0,   9,   6,   9,  -3,   0,   0, -50,  17, -15,  -1,   4,   5,   0,  55,  10,  -1,  -3,   4,  -1,   1, -14,  37,   7,  -5,  -2,   1,  -5,  -6,  -5,  -9,   4,  -4,  -5,  -4,  -2,  -2,   3,  -5,  18,  -1,  -6,   1,  -2,  -7,   4,  -6,   2,  -4,   2,   2,  -2 },
      {   4,  13, -14, -21, 102,   2,  18,  -2,  -4,  -2,   0, -50,  -1,  -9,   1,  -3,  -3,  -4,  -3, -10, -36,  -2,  -4,  -1,   0,   1, -14,  13,   4,   4,   0,   0,  -1,  11, -11,   4,   0,   2,  -1,   1,   1,   4,   2,   5,   0,   2,   1,   0,   0,  -6,   2,  -2,  -2,   2,   0,   0,   2,   0,   1,   2,   1,   0,   0,   0 },
      {   3,   7,  -3,  -5,  49,  15,   6,   0,   8,   5,  21,  74,  17,  14,  -2,   2,   0,   0,   1,  -2,  -5, -11,  -4,   1,   4,  -5,  74, -24,  -6, -10,   1,  -2,   2,  -1,  18,   0,  -7,  -1,   0,  -2,  -6,  -5, -12,  -7,  -5,  -2,  -1,  -1,   0,  -1,  -3,   5,  -1,  -2,   0,   0,  -3,   0,  -5,  -1,  -1,   1,   0,  -1 },
      {  -2,  -4,  -6,  17, -11, -42,  -5,  -7,   0,  -2,  24, -29, -12, -26,  -3, -11,   2,  -2,  -4,  49, -13,  22,   5,  -3,   9,   1,  70,  25,  10,   3,  -7,  -1,  -1,  11, -38,  -6,   5,   4,  -1,   6,  -3,  13, -21,  -1,   2,   6,   2,   3,   5,  -3,   1, -19,   2,   6,   0,   2,  -2,  -4,  -2,   0,   4,  -1,  -1,   2 },
      {   3,   1,   7,  22,  12,  14,   2,   3,  -2, -12,   0, -16,  -1,   5,   0,   2,   5,   3,   6,  62,   5,   4,  -6,   1,  -2, -17,  -5,   6,   3,   0,   1,   1,   4,  -8,  36, -11,  -2,  -4,   0,  -2,  -8, -94,   2,   6,   3,   3,   0,   1,  -2,   6,  -6,  -3,  -1,  -2,   0,   0,   3,   8,   0,  -1,  -1,  -1,   1,  -1 },
      {  -4,  -7, -10, -12,  -9, -48,  -4, -10,  -1, -13,   0,  15, -10, -10,  -2,  -4,  -4,  -6,   0, -71, -14,   9,   7,   1,  -1, -16,   8,  -3,   4,   3,  -1,   0,  -2,   9, -32,  20,   6,   6,   0,   3,  -5, -73,  -3,   6,   3,   7,   2,   2,  -1,  -3,   1,  -3,   1,   3,   0,   1,   1,   7,   3,   0,   2,  -1,  -1,   1 },
      {  -1,  -2,   4,  -3,   7, -39,  -3, -11,  -3,  -7,   3,  14, -75, -15,  -9,   0,   2,  -3,  22,  -6, -15,   6,   2,   1,   2,   3,  -4,   5,  19,   6,   3,   0,   8,  -7,  79,   4,   7,   1,  -1,   0,   2,  21,  -1,   0,   4,   2,   0,   1,  -1,  -3,  -9,  -6,  -1,   3,   0,   1,   1,  -3,  -1,   0,   3,   0,   0,   0 },
      {   2,   3,   1, -10, -12,  66,  18,  18,  -4,  -9,   4,  14, -84,  -4, -10,   3,  -3,  -3, -15,   6, -13, -17,  -9,  -3,   1,   1,   6,  -2,  18,   3,   6,   0,  -5,   3, -46,   1,   3,   0,   2,  -1,  -1, -10,  -4,   0,   6,   0,   0,   0,   0,  -6,   4,   3,   3,  -2,   0,  -1,   0,   3,   0,  -1,   0,   0,   1,   0 },
      {   0,   0,  -6,   6,  20, -21,  -6,  -2,  -3,  -6,  -3, -27, -27,  -3,   0,   2,   2,   3,  -6,   4,  53,  11,   9,   1,  -4,  -5,   2, -97,  -1,  -5,   7,   2,  -3,  -2, -14, -19, -19,  -1,  -2,   0,  -1,   4,  -9,  11,   4,   2,   0,   0,   0,  10,   6,   5,  -4,   0,  -1,   0,  -1,  -2,   0,   6,   0,   0,   0,  -1 },
      {   1,   7,  -4,   2,  20,  -6,  -3,  -1,   1,   5,   4,  11, -17,  21,   2,   7,   3,  12,  -5,  -7,  45,  10,   7,   2,   2,   7,  -5,  44,  11,   2,   1,   0,   0,  10, -10,  12, -11,  -3,  -2,  -2,  -2,   0, -16, -11,  -3,  -6,  -2,  -1,   7,  96,   5,  -4,  -6,  -7,  -1,  -4,   1,   0,   1,  -3,  -1,   0,   1,  -1 },
      {   0,  -2,  -2,   4,  24,  -9,   3,  -1,   1,  -1,   5,  11, -13,  14,  -3,   4,   3,   1,  -5,  -4,  77,  13,   8,   0,   2,   0,  -3,  48,   9,   2,   1,   0,  -4, -10, -10,   2, -17,  -5,  -3,  -2,   1,  -1,   8,  -3,  -1,  -5,  -2,  -2,  -4, -76,   1,   9,  -1,   3,   1,   2,   0,   3,  -2,  -2,  -2,   0,   0,  -1 },
      {   3,   4,   5,   4,  -4,  12,  62,  12,  -2,  -4, -13,   6,  16, -68,  -7, -17,   2,   2,   1,  -4,  32,  -7,  -7, -10,  -1,   0,  -5,  -1, -19,  14,  -1,   3,   3,   3,  14,  38,   2,   4,  -7,   2,  -4,   1, -50,   6,   7,  10,   4,   4,   0,   1, -10, -17,  -1,   3,  -1,   1,   1,  -4,   3,   0,   5,   1,  -1,   1 },
      {   1,   2,   1,  15,  -5,  -7,  30,   6,   1,   3,   9, -11, -16,  55,   2,   6,   1,   3,   4,  12,  -9,  34, -10,  -1,  -2,  -2,   4, -15, -41, -19,  -7,   0,   2,   3,  -2,  82,   3,  -1,  -5,  -1,   3,   5,  22,   5,  10,  -2,   1,  -1,  -1,  -6,  -1, -11,   0,  -4,   1,   1,  -1, -10,  -1,   3,   2,  -1,   0,   1 },
      {  -1,  -2,  -3,  -3,  -2,  -1, -14,  -1,   0,   1,  -9,  -3,  -4,  50,   5,   7,  -1,  -4,   1,   0, -19,   6,   2,   2,  -3,  -5, -19,   3,  -7, -10,  -4,  -1,  -1,  -2,   1, -11,  -2,  -1,   1,   1,  -8,   1,-108,   2,   2,   2,   2,   0,  -2, -20,  -3,   1,   2,   1,   0,   0,   2,  -4,   8,   3,   2,  -1,  -1,   0 },
      {   1,   2,  -1,   7, -14, -29,  94,  10,   2,   4,   5,  -1,   3,  33,  25,   7,  -1,  -2,   1,  -9,  -5,  -1, -17,  -5,   1,   2,   2,   0,  29,  -9,   0,  -3,  -1,  -1,   1, -50,   6,   1,  -1,  -1,   1,   2,   7,   4,  -2,  -2,  -6,  -1,   0,   4,   4,   8,   1,   2,  -2,   0,   0,  17,  -1,  -3,  -1,  -3,   0,   1 },
      {   0,   0,   2,   5,   1,   0,  -7,  -4,   2,   2,   3,  -3,  22,  13,  10,  -2,   0,   0,   3,  11,   9, -27,   0,  -3,   3,   5,  -4, -19,  99,  -2,   4,  -1,   3,   4,   0,  51,  27,   8,   0,   1,  -1,   0,  -8,   5, -11,  -2,  -2,  -1,   0,  -9,   9, -12,  -5,   1,  -1,   0,  -1, -19,  -1,   1,  -2,   0,   0,   0 },
      {  -1,   0,  -6,   7,   1, -25,  -8, -25,   0,   4,  -1,  -1,  -9,   3,   8,   7,  -2,   0,  -9,  16,   5, -73, -15,   0,   0,   5,  -1,   3, -15, -13,  -3,   0,  -1,   3,  -8,  23, -11,  15,   3,  -1,   1,   8,  -3,  20,   6,   0,  -2,   0,  -2,   1, -28,   9,   3,   1,   3,   0,   5,  77,   1,  -9,  -3,  -2,   0,  -2 },
      {   0,  -2,  -2,   3,   0, -25,   6, -11,   0,  -4,   2,  -4, -13,   2,  14,   3,  -2,  -5,  -6,   7,   7, -71, -12,  -4,  -2,  -8,  -1,  10, -31, -17,  -4,  -2,   0,  -2,  -3,  -7,  -2,  12,   4,   0,  -1,  -9,   4, -16,   4,  -3,   1,  -1,   3,   0,  23,   8,   0,   1,   0,   0,  -4, -84,   0,   9,   3,   4,   0,  -1 },
      {   0,   0,   5,   0,   1,  -5,  -1,   1,  -1,  -1,   0,  -6,   1,   1, -39,  -4,   1,   1,   7,   1,   6,  -8, -11,  -4,   0,   1,   1,  -2, -12,  -8,   8,   3,   3,   4,   9,   2,  33,  -1,   5,   0,  -1,  -2,  -7, -31,  -2,  -4,   4,   2,   8,  -5, 100,   0,  -7,  -5,  -3,  -1,   1,  41,   2,  -1,   0,  -1,   0,  -1 },
      {   0,   0,   0,   4,   2, -32, -11,  95,  -1,  -1,  -1,  -2,   8,   5, -61,  13,  -1,  -1,  -2,   1,  -7, -15, -21, -11,   0,   1,   0,   3,   7,   2,  14,   5,  -1,  -1,  -2,   2, -19,   6,   2,  -4,  -1,   1,   1,   9,  -7,  -5,   0,  -3,  -1,   2, -18,  13,   1,   0,   2,  -3,  -1,  -6,  -2,   1,  -2,   1,   2,  -1 },
      {   1,   0,   4,  -5,   7,  -5, -19,  72,   1,   2,   1,   4, -10, -12,  80,   4,   1,   0,   2,   1,   3,  -7,  36, -10,  -1,  -1,   1,   0, -10, -15, -26,  -3,   1,  -1,   5,   2,  17,   1,  -5,  -2,   0,  -1,   2,  -7,   7,   5,  -2,   3,   0,  -4,  17, -21,  -5,   4,  -2,   1,   0,  18,   4,   0,   1,  -1,  -3,  -1 },
      {   0,   0,   3,  -4,   1,   9,  -7,  -1,   0,   0,   2,  10,   1,  10, -13, -21,   0,   0,   3,  -4,   7,  -2,   4, -15,   0,   1,   4,  13,  -3, -14, -16,  -3,   1,   3,   6, -13,   7,   2,  -3,   5,   4,   5,   5, 111,  14,  11,   4,   2,   3,   7,  32, -12,   0,   7,   1,   2,   0, -11,   4,  -6,   1,  -3,  -1,   4 },
      {   0,   0,   0,  -6,   7,   6,  -1,   0,  -1,  -1,   1,  -7,   0,  17, -36, -24,   1,   0,  -2,  -7,  23,  -8,  15, -18,  -1,  -4,   0,  -1,  -9, -43,  -7,   0,   2,   3,  -3, -22,  78,   5,   2,   7,   1,   0,   6, -22,  13,  10,   7,   3,  -5,   3, -50, -25,  -9,   5,   0,   4,   0,   0,   1,   1,   4,  -1,   0,   3 },
      {   0,  -1,   2,  -5,  -8,   7,   9,  -4,   0,  -1,   1,  -7,   3,  12, -22, -39,   0,  -1,   0,  -6, -11, -15,  34, -22,   0,  -1,  -2,  -1,  17, -35, -14,  -6,  -2,  -4,   8,  -2, -76,  -1, -14,  12,   1,  -2,   8, -27,   4,   9,   9,   7,   0,  -1,   8, -49,   9,   9,   0,   1,   1,   8,  14,   0,   6,  -1,  -1,   2 },
      {   0,  -1,   1,  -3,   1,   0,  -3,   3,   1,   1,   3,  -3,   0,  40,   4, -27,  -1,  -1,   0,  -2,   7, -19,  -1, -24,   2,   3,   8,  -8, -13, 100,   4,   7,   1,   1,   2,  -8,   6,  23,   8,  11,   1,  -1,   5,  -6,  11,  -6,   3,   4,  -1,  -2,  -3, -36,  -8,   6,   0,   1,   1,   2,   7,   2,   4,  -5,  -3,   3 },
      {   0,  -1,   0,  -4,   3,   4, -24,   2,   0,   0,   0,   0,  -2,  -6,  21, -20,  -2,  -2,  -1, -12,  13,  21,-103, -17,  -1,  -1,  -1,  -2,  10, -16, -36,  -1,   0,   0,   2,  -7, -10,   2,  20,   7,   0,   1,   1, -13,   9,   5,   3,   1,   0,   0,  -2, -18,   5,   1,   5,   3,   1,   1,  15,   0,  -2,  -2,   3,   1 },
      {   0,   0,   0,   3,   0,  -4,  -2,   6,  -1,  -1,  -5,   1,  -8,   5,  17, -88,   0,  -1,   0,   5,  -4,   5,   2, -28,   0,   1,  -4,   1,  -4,  -6,  11,  13,   0,   1,   0,   9,   0,  -5,   7,   1,  -3,  -2,  -5,  -1, -35, -13,  -5,   2,   1,   4,  -2,  52,   2,  -3,   0,   0,  -4,   3, -51,   0,   3,   0,   2,  -1 },
      {   0,   0,   0,   3,   1,  -4,  -1,   6,   0,   0,   3,   1,   2,   0,  11, -38,   0,   1,   1,   4,  -2,   1,   5,   3,   1,   1,   7,   3,   3,  -5,  40,  11,   0,   1,   1,   6,   4, -18,   3,  -4,   2,   2,   3,  -1,  34,  -6,  -6,  -3,   0,   2,  -3,  37,   1,  -4,   0,  -2,   8,  -2,  98,   1,  -8,  -4,  -3,  -1 },
      {   0,   0,   0,  -1,   1,  -4,  -5,   3,   0,   0,  -2,   1,  10,   1,   6,  -4,  -1,  -1,   0,  -2,  -3,  -4,  -6,   4,   1,   0,   0,   2,  11,  -1,  29,   2,  -1,  -2,  -1,   0,  -5, -34,  -4,  -1,   2,   2,  -4,  -7, 102,  -6,   1,  -1,   0,   2,   4,  -3,  11,   3,   1,   0,  -5,   0, -56,  -4,  -7,   2,   0,   0 },
      {   0,  -1,  -1,  -1,  -1,  -9,   0,   0,   0,   0,   1,  -1,  -5,   0,   3,  -3,  -1,  -1,  -1,  -3,  -1, -20, -13,   9,   0,   0,   1,  -3,  -9,   4,  12,   9,  -1,  -5,  -6,  -1,   8,-102,  -7,  -2,  -2,  -1,  -1,  10, -38, -32,  -4,  -4,  -1,  -2,  -5, -43, -11,   7,   3,   1,   1,  -1,   3,  -8,   4,   1,  -1,  -1 },
      {   0,   1,   0,   2,   0,  -3,   6,   8,  -1,   0,  -1,  -2,   0,   7, -20, -28,   0,   1,   1,   2,   0, -12,   6,  60,  -1,  -1,   1,  -5,   9,  22, -87,   5,  -1,   0,  -1,   1,   5, -20, -27, -17,   0,   0,   1,  -8,  17,   2,  12,   2,   1,   3,   0,  31,  -1,   9,  -2,  -1,   1,   0,   4,   0,  -4,  -1,   2,  -1 },
      {   0,   0,   0,   1,   0,  -1,  -2, -13,   0,   0,   1,  -1,   2,  -4,  -8,  34,  -1,  -1,  -2,   3,  -4,   1,   7, -91,   0,   0,   1,  -3,   5,  16, -35, -25,   0,   0,   0,   7,   4, -37, -27,  11,   0,  -1,   0,  -8,   9,   5,   3,   4,   1,   4,  -1,  43,   0,   7,   2,   6,   1,   0,  10,   4,   3,  -2,   2,  -1 },
      {   0,   0,   0,   0,   2,   1,   3,  -2,   0,   0,   0,  -1,   0, -11,  -6,   3,   0,   0,   1,  -1,  -1,   7,  12, -10,  -1,  -1,  -1,  -5,   1,  -6, -24,  -4,   1,   1,   1,  -3,   4,  23,  12,   2,  -1,  -4,  -5,   4,  11,-107,  -1,  -8,   1,  -2,  -2,  -6,  36, -16,  -4,   2,   1,  -2,   8, -32,   4,   7,   1,  -9 },
      {   0,   0,   0,  -1,  -2,   2,  -9,  -1,   0,   0,  -1,   1,  -1,  -2,   5,  -9,   0,  -1,  -1,  -1,  -3,   4, -23,   2,   0,   1,  -1,   2,  -3,  -4,  20, -11,  -1,  -1,  -2,  -7,   3,  28,-107,  -7,   0,  -1,  -1,   1,  -1, -25, -13,  -3,  -1,  -2,   0,  -8, -36,  -8,  17,   2,   0,   2,  -1,   7,  -2,   2,  -1,  -2 },
      {   0,   0,  -1,   0,  -5,   1,   6,  -6,   0,   0,  -1,   3,   1,  -7,  -3,   8,   0,  -1,   0,   1,  -9,   3,   8, -11,   0,   0,   0,   3,   5, -10, -19,  41,   0,  -1,  -1,   0, -15,   3,  32,   2,   1,   0,  -1,   5,  16, -30,   5,   0,  -3,  -3,  -3,   2, -96, -14,  -5,   6,   0,   3,  -2,  39,  -7,   9,   0,   9 },
      {   0,   0,   0,  -1,   2,   2,  -2,  -4,   0,   0,   1,   2,   1,   1,   3,  16,   0,   0,   0,  -1,   2,  -1,  -1,  -9,   1,   2,   2,   2,   2,  -3,  -5, 102,   0,   0,   1,  -2,   5,   1, -28,  21,   0,   0,   1,   2,  -3,   2,  -4,  -2,   1,   2,   2,  -2,  52,   0,   1,   0,   1,   3,   0,  38,   5,   8,   0,  10 },
      {   0,   0,   0,   0,   1,  -1,   1,   1,   0,   0,   0,   4,   0,  -1,  -2,  -7,   0,   0,   0,   1,   3,  -1,   0,   9,   0,  -1,  -1,   3,   1,   1,   3, -50,   0,   1,   0,   0,   5,  -4,  10, -12,   0,   0,   1,   7,  -1, -23,  -4,  10,   0,   2,  -1,  -5,  29,   4,   6,   0,   2,  12,   0, 107,   9,   6,  -1,  11 },
      {   0,   0,   1,   0,   0,   4,   0,   4,   0,   0,   0,   0,   0,   0,  -5,   2,   0,   0,   1,   1,   1,   0,  -1,  20,   0,  -1,  -1,  -1,  -3,  -3,  -4, -11,   1,   2,   2,   4,  -1,   2,   3,  82,  -1,  -2,   0,  -2,   4,  -9, -70,   9,   1,   2,   2,  12, -12,  54,  -6, -12,   1,   1,   1,  -3,  14,   4,   5,  12 },
      {   0,   0,  -1,   0,   0,  -3,   0,  -4,   0,   0,  -1,   0,   0,  -2, -11,  -6,   0,   0,   0,   0,  -2,  -2,   6, -15,   0,   0,  -1,  -1,  -2,   3, -17,   4,   0,  -1,  -3,  -1,   0,  -3,   7, -49,  -2,   0,  -3,  -1,   5,  16,-100, -15,  -1,  -3,   1, -16,   7, -36, -11,   1,  -1,  -2,  -1,   0, -27,  -1,  10,   3 },
      {   0,   0,   0,   0,  -1,   5,   0,  -5,   0,   0,   0,  -1,   0,   0,   0,   4,   0,   0,   0,   1,  -1,   6,   0, -13,   0,   0,   0,   0,  -3,  -4,   6,  14,   0,   0,   0,   3,  -1,  12,   5, -58,  -1,  -1,  -1,  -3,  -4, -14,   0,  -4,   1,   5,   3,  -1,  -1, 103,  12,   6,  -1,   0,   1,  -3, -36,   8,   3,   2 },
      {   0,   0,  -1,   1,   0,  -2,   3,   1,   0,   0,  -1,   1,  -4,  -3,  -1,  -3,   0,   0,   0,   1,   0,   0,   5,   1,   0,   0,   0,   0,  -4,   0,  -3,  -9,   0,   1,   1,   0,   2,  -2,  -2,  45,   0,   0,  -2,   0,  -8,  -3,  16, -31,   0,   0,   0,  -5,   7,  -9,   3,  -5,  -3,  -2,  -4,  14,-106, -32,   3, -13 },
      {   0,   0,   0,   0,  -1,   2,   0,  -1,   0,   0,   0,   0,   2,  -2,  -1,  -1,   0,   0,   0,   0,  -1,   0,   1,  -1,   0,   0,   1,   0,   1,  -2,  -3,   2,   0,   0,   1,   0,  -1,   2,   3,  -2,   0,  -1,   0,  -1,   5,   3,  -3, -99,   0,   1,   1,   0,  -5,  16,   7,  -5,   1,   2,  -2,  13,  47, -50,  -5, -36 },
      {   0,   0,   0,   0,   0,   0,   4,   1,   0,   0,   0,   0,   1,  -1,  -1,   2,   0,   0,   0,   0,   0,  -1,   9,   3,   0,   0,   0,  -1,   2,   1,  -5,   6,   0,   0,   0,   0,   1,  -1,  15,   4,   0,   0,  -1,  -2,   1,   1, -13,  20,   2,   1,   3,   1,   0, -10, 122,  11,   0,  -2,   0,  -5,   3, -16,   4,  -8 },
      {   0,   0,   0,   0,   0,  -1,   1,   2,   0,   0,   0,   0,  -1,   2,   2,  -6,   0,   0,   0,   1,   0,  -1,   4,  -1,   0,   0,   0,  -1,   0,   4,  -2, -12,   0,   0,   0,   0,   1,  -2,   2,   9,   0,  -1,  -1,  -1,   0,   9,  11, -66,   0,   0,   1,   0,   1,  -7,  26, -12,   0,   4,   3,  -8,  -5,  84,  19,  56 },
      {   0,   0,   0,   0,   1,   0,  -2,  -2,   0,   0,   0,  -1,   0,  -3,   1,   4,   0,   0,   0,   0,   0,   0,   0,  -5,   0,   0,   0,  -1,  -1,  -3,  -1,   2,   0,   0,   0,   0,   0,   0,  -2, -15,   0,   0,   0,  -2,   0, -10,   7,   6,   0,  -2,  -1,   0,  -1,  -2,   4, -77,   0,  -2,  -1,  -7,   4, -59,   7,  79 },
      {   0,   0,   0,   0,   0,  -1,   1,  -4,   0,   0,   0,   1,   0,   1,   1,  -2,   0,   0,   0,   0,  -1,   2,   0,  -8,   0,   0,   0,   0,   0,   1,  -3,   4,   0,   0,  -1,  -2,   2,   0,   1,  -3,   0,   1,   1,   3,   3,   4,   1,  12,  -1,  -3,  -1,  -5,  -3,   1,   7, -97,   1,   1,   2,   7,  -1,  37,  14, -70 },
      {   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,  -1,   0,   2,   2,   0,   0,   0,   0,   0,   1,   5,   1,   0,   0,  -1,   0,   1,   0,  -2,   1,   0,   0,   0,   2,   1,  -1,   8,   6,   0,  -1,   1,  -2,  -1,   1,  -8,  16,   2,   0,   3,   2,   7, -15, 125,  -6 },
    },
    {
      { -90,  30,  19,  -2,   3,   1,   1,   0, -72,  19,  18,  -2,   3,   0,   1,   0,  -1,  -3,   2,   0,   0,   0,   0,   0,  25,  -8,  -6,   2,  -1,   0,   0,   0,  16,  -4,  -4,   1,  -1,   0,   0,   0,   6,  -1,  -1,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
      { -17, -77,  53,   2,   6,   0,   2,   0,  -3, -66,  37,   5,   3,   1,   1,   0,  14,  -3,  -8,   4,  -1,   1,  -1,   0,   9,  21, -18,   0,  -1,   0,   0,   0,   2,  14,  -8,  -2,   0,  -1,   0,   0,   0,   6,  -2,  -1,   0,   0,   0,   0,  -1,   3,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 },
      {  68, -28,  12, -10,   0,  -2,   0,  -1, -78,  24,  19,  -4,   2,   0,   1,   0, -55,  17,   9,  -1,   2,   0,   1,   0,   0,  -1,  -2,   1,   1,   0,   0,   0,  11,  -3,  -4,   1,   0,   0,   0,   0,   7,  -2,  -2,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
      { -39,   8,  19,  -5,  -1,   1,   0,   0,  40,  -8,  15,  -9,  -3,  -1,  -1,   0, -89,  27,  17,  -1,   3,   0,   1,   0, -56,  11,   7,   2,   3,   0,   1,   0,   5,  -3,  -3,   2,   0,   0,   0,   0,  11,  -3,  -4,   1,   0,   0,   0,   0,   5,  -1,  -2,   0,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
      { -21, -29, -54,  59,  -2,   7,  -2,   2, -27, -28, -49,  51,   3,   5,   0,   2, -26,   4,  -7,   1,   6,  -1,   1,   0,  -5,  16,  14, -20,   2,  -1,   0,   0,   8,   8,  11, -12,  -1,  -1,   0,   0,   6,   2,   5,  -4,  -1,   0,   0,   0,   2,   0,   2,  -2,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0,   0 },
      {  17,  35, -19,  14, -15,   4,  -4,   1,  -6, -49,  39,   8,  -5,   3,   0,   0, -20, -78,  49,   5,   2,   3,   1,   1,   4, -22,   2,   6,   0,   2,  -1,   1,   5,  10, -12,   1,   1,  -1,   0,   0,   2,  10,  -7,  -2,   1,  -1,   0,   0,   1,   4,  -2,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 },
      {   2,  -1,  -1,   0,   1,  -1,   0,   1,  40,  -4, -10,   0,   1,  -1,   0,   0, -45,  13, -10,   4,   1,   0,   1,   0,  95, -19, -19,   1,  -4,   0,  -1,   0,  48,  -8,  -8,   0,  -3,   0,  -1,   0,  -7,   2,   2,  -1,   0,   0,   0,   0,  -9,   2,   3,  -1,   1,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0,   0 },
      {  -9, -14, -41, -45,  56, -10,   8,  -1, -10, -29, -22, -51,  53,  -7,   7,   0,  -6, -11,  14, -22,   6,   2,   0,   1,   2,   8,  18,   5, -18,   5,  -3,   0,   5,   7,   7,  11, -13,   2,  -1,   0,   3,   2,   1,   7,  -5,   0,   0,   0,   1,   0,   0,   3,  -3,   0,   0,   0,   0,   1,   0,   2,  -1,   0,   0,   0 },
      {  13,  48, -27,   6, -17,   6,  -4,   1,  -9, -28,  29, -16,   0,   2,   2,   0,   3,  37,  -7, -24,   6,  -4,   2,  -1,  16,  78, -37, -12,   3,  -5,   1,  -1,  -8,  21,  -7,  -2,   2,  -2,   1,  -1,  -8, -10,   6,   3,   0,   0,   0,   0,  -3,  -9,   4,   2,   0,   0,   0,   0,   0,  -3,   0,   1,   0,   0,   0,   0 },
      { -10, -30,  -7,  18, -20,  13,  -1,   2,   9,  18,  21, -17, -19,   7,  -1,   0,  13,  27,  68, -50,  -6,  -3,  -1,  -1,  29,  -2,  50, -26,  -4,  -1,  -2,   0, -17,  -2,   7,   3,   0,   0,   0,   1,  -9,  -1,  -9,   8,   2,  -1,   1,   0,  -1,   0,  -7,   4,   1,  -1,   1,   0,   0,  -1,  -3,   1,   1,   0,   0,   0 },
      {   7,  -4,  -6,   7,  -7,   3,  -1,   1,   4,   5,   6,  -6,  -4,   1,  -1,   1,  36,   4,  11, -17,   1,  -2,  -1,   0, -35,  10,   1,  -5,   2,   0,   0,   0, 107, -12, -18,   2,  -4,   0,  -2,   0,  31,  -5,  -7,   3,  -2,   0,   0,   0,  -7,   0,   1,   1,   0,   0,   1,   0,  -7,   0,   1,   0,   0,   0,   0,   0 },
      { -11, -21, -28, -40, -43,  42,  -1,   6, -10, -13, -27, -36, -49,  39,   0,   5,  -5,  -7, -12,  10, -21,   8,   2,   1,   3,  -1,  -8,  27,   6, -11,   3,  -2,  -5,   6,  -1,  13,  13, -11,   1,  -1,  33,   0,  -2,   2,   6,  -5,  -1,   0,   8,   1,   1,   0,   2,  -2,  -1,   0,   0,   0,   2,   0,   1,  -1,   0,   0 },
      {   8,   8,  10,   8,  12, -11,   0,  -1,  11,   4,   3,  10,  11,  -8,  -1,  -1,   3,  -2,  -2,  -7,   6,  -1,  -1,   0,  23,   4,  -1, -18,   1,   2,   0,   0, -27,   0,  -5,  -6,  -2,   2,   1,   0, 110, -13, -17,   2,  -6,   1,  -2,   0,  32,  -3,  -6,   3,  -2,   0,  -1,   0,  -2,   0,   0,   0,   0,   0,   0,   0 },
      {  11,  30,  -2,  -4,  -4,   0,   0,  -1, -14, -63,  18,   6,  -2,   3,   2,   0,  12,  52, -17,  14, -16,   2,  -3,   0,   0, -32,  38,   7,  -9,   2,  -1,   0,   2, -57,  33,   4,  -2,   2,  -1,   1,   6,  -7,  -1,   3,   0,   1,   0,   0,   2,   5,  -7,   0,   1,   0,   0,   0,   2,   4,  -3,   0,   1,  -1,   0,   0 },
      {  -7, -16, -38,   9,  26,  -9,   4,  -1,   7,  19,  28,  19,   1,  -9,  -1,  -2,  11,  27,  42,  56, -38,   1,  -5,  -1,   4,  18, -14,  54, -26,   1,  -2,  -1,  -5,  -1, -15,   7,   2,  -1,   1,   0,   9,  -6,  -5, -11,   6,   0,   0,   0,  -4,  -2,   1,  -7,   3,   1,   0,   1,  -3,  -2,   0,  -3,   1,   0,   0,   0 },
      {   7,  22,  33, -40,  12, -12,   6,  -4,  -3,  -2, -23,  26, -24,   8,  -1,   1,  -4, -14,  12,  28, -37,  12,  -3,   2,  14,  47,  51, -27, -16,  -2,  -1,  -1,  19,  32,  20, -19,   0,  -3,  -1,  -1,   1,   0,   0,  -2,   4,  -1,   0,   0, -11,  -5,  -2,   3,   2,   0,   0,   1,  -5,  -2,  -1,   1,   1,   0,   1,   0 },
      {   2,   1,  -6, -10,  -4, -11,   9,  -2,   2,  -4,  -3,   3,  -7, -10,   8,  -1,   6,  11,   3,   7,   0,  -5,   1,  -1,   3,  -4,   3,  -2,   5,  -1,  -1,   0,  18,  17, -11,  -5,   4,   1,  -2,   1, -29,  10,  -9,  -3,   3,   1,  -1,   0, 110,  -8, -14,  -1,  -3,   1,  -2,   0,  34,  -1,  -5,   1,  -1,   1,  -1,   0 },
      {  -4,   3,  -9, -18, -24, -47,  36,  -6,  -9, -19, -17,   0, -22, -46,  33,  -3,   5,  28,   8,   0,   8, -22,   6,   0,  -3, -22,   4,   1,  25,   0,  -6,   2,  -6,  32, -29,   0,  15,   6,  -7,   1,  11,  32, -15,  -5,   4,   5,  -3,   0, -30,   7,   5,  -3,   1,   3,  -1,   0, -14,  -1,   6,  -1,   0,   2,  -1,   0 },
      {  -7, -19,  -7,  -2, -27, -35,  27,  -5,   2,   7,   1, -10, -22, -38,  24,  -4,  -8, -39,   1,   3,   6, -15,   6,   0,   7,  36,  -3,   9,   5,   2,  -7,   0,   2, -51,  39,   7,   2,   9,  -6,   2,  -2, -47,  21,   7,  -1,   8,  -3,   2,   6,  -9,  -1,   2,   0,   3,  -2,   0,   3,   3,  -5,   0,   1,   1,  -1,   0 },
      {  -8, -16, -43,  -4,   3,  -1,   6,  -1,   9,  20,  63, -21,  -5,  -6,   3,  -3,  -4,  -4, -38,  20,  -7,   7,   0,   0,  -3, -18,   8, -35,   9,   3,   1,   1,   9,  27,  41, -45,   4,   1,  -1,   0,   8,  19,  13, -12,  -3,   0,  -2,   0,  -2,   5,   3,   3,  -2,   0,   0,   0,  -5,  -2,  -2,   4,   0,   0,   0,   0 },
      {  -2,  -3, -14, -11,  31, -22,  21, -12,  -3,  -6,  -2,  13, -31,  17,   7,  -2,  -4,  -3,   2, -32, -49,  39,  -6,   4, -11, -17, -43, -44,  -5,  19,   0,   3, -10, -31, -30, -10,  10,   2,   1,   2, -10, -15,  -5,   7,   6,  -4,   0,   0,   3,  -2,   4,   6,   1,  -3,  -2,   0, -11,   3,   4,   2,   1,  -2,   0,   0 },
      {   5,   8,  17,  52, -17, -11,  12,  -3,  -2,   3,  -6, -32,  -3, -13,  21,  -3,  -8,  -7, -32, -27, -45,  15,   7,   3,  -5,  -4,   8,  25, -62,  19,  -3,   3,   3,  28,   7,  14, -20,   2,  -2,  -1,   6,  15,   1,   1,   5,  -4,  -2,  -1,   5,   1,   0,  -2,   7,  -2,  -1,   0,   5,  -2,   0,   0,   4,   0,  -1,   1 },
      {   1,  -1,  -3,  -5,   3,  -1,   3,   2,   2,  -1,   0,   4,  -4,  -1,   2,  -1,   3,   1,   3,  -2,  -4,   3,   0,  -1,  -1,  -6,  -5,  -6,   1,   3,   0,   0,   1,   2,  -5,  -5,   2,   1,  -1,   0,  12,  -3,   5,  -2,   0,   0,   0,   1, -36,  -6,   4,   1,   0,   1,   0,   0, 120,  -7, -11,   1,  -3,   0,  -1,   0 },
      {  -5,  -7, -10,  -8, -17, -36, -48,  29,  -3,   2,  -6, -12, -11, -29, -51,  27,  -4,  -8,  -9,  -1,  -8,   7, -20,   8,   2,  24,   0,  -6,  -6,  19,  -1,  -4,  -5, -31,  -6,  -2,   1,  13,   9,  -6,  -1,  43, -25,  -5,   2,   2,   7,  -4,   1,  32, -10,  -5,   2,  -1,   4,  -3,   6,   3,   0,  -1,   1,  -1,   3,  -1 },
      {   0,   0,  -8,   8,  11,  23,  32, -19,   5,  15,  14,   2,   9,  16,  31, -18,  -5, -17, -29,  11,   5,  -3,  13,  -3,   6,  34,  24,  -9,   4, -16,  -2,   3,  -5, -36, -23,  17,   7,  -6,  -5,   4,  -8,  34, -58,   8,   8,  -5,  -3,   1,  -6,  30, -18,  -6,   5,  -3,  -1,   1,   9,   3,   1,  -3,   0,  -1,  -1,   1 },
      {  -6,  -9, -26, -24,  -7,  -7, -12,   9,   3,   4,  29,  35, -13, -11, -14,   6,  -5,  -5, -39,  -5,  12,  -1,  -3,   0,   0, -11,  27,  -6, -16,   7,   0,  -2,  -1,  26, -25,  53, -25,   6,   2,  -4,  -7, -37, -21,  38,  -2,   1,   3,  -1,  -6, -29,   1,  10,   4,   0,   1,   1,  -2,  -5,   1,  -3,   2,   0,   0,   0 },
      {  -2,   0,  -6, -39,  -3,  12,  11,  -4,   4,   7,  10,  54, -14,   7,  10,  -5,   2,   3,  12, -32,  17, -12,  16,  -5,  -4,   4, -28,   7, -31,   9,   1,   1,   1,  -7,  34,  18, -39,   9,  -4,   1,  10,  48,  24, -10, -13,  -1,  -3,   0,   7,  27,   6,  -6,   2,  -2,  -2,   1,   0,   4,  -2,  -2,   3,   0,  -1,   1 },
      {   2,   4,   1,  17,  25, -33,  16, -17,   2,   4,  10,   5, -19,  25, -15,   0,   0,   2,  -5, -19,  17,  48, -32,   7,   7,   9,  12,  39,  50,   7, -15,   0,   8,  16,  29,  38,  19, -10,  -1,  -2,   6,  17,  12,   9,  -3, -10,   4,  -1,   2,   6,  -1,  -3,  -5,  -1,   3,   0,   4,   0,  -3,  -3,  -1,   0,   1,   0 },
      {  -3,  -3,   0, -10, -48,   8,  22,  -8,   1,   2,  -6,  -1,  30,  13,  10,  -8,   5,   9,  12,  29,  35,  53, -27,   2,   0,   5,   0,  -7, -14,  60, -27,   6,  -3,  -2, -12, -17, -15,  22,  -5,   1,  -2,  -5,  -6,  -2,  -7,  -3,   5,   1,  -3,  -5,  -4,   5,   0,  -5,   4,   0,  -2,  -1,  -1,   2,  -1,  -3,   0,   0 },
      {   1,   3,   3,  17,   4,  -5,  -1, -14,  -3,  -2, -13, -23, -13,   2,  -5, -10,   3,   9,  18,  39,  -7, -11,  -7,  -4,  -4,  -9, -30, -28,   9,  -6,   4,   3,   2,  21,  21,  20, -40,   4,   2,   1,  -2, -10,  -8,  55, -26,   2,   2,  -2,  -3,  45, -31,  13,  -1,  -5,   3,  -2,   5,  24, -12,  -2,   5,  -3,   1,   0 },
      {   0,   1,  -1, -12,  -6,   5,   7,  29,   2,   3,  -2,  23,  22,  -2,  12,  26,   0,   2,   4, -23,  -4,   2,   2,   8,  -3,  -9, -27,  19,  -3,   4,  -4,  -2,   7,  18,  43, -16,  31, -10,  -5,  -4,  -5, -19, -46,   6,  27,  -4,  -1,  -1, -11,   0, -61,  16,  13,  -1,   1,  -1,  -3,   8, -18,   0,   1,   0,   0,  -1 },
      {  -4,  -6,  -6, -14, -16,  -9, -34, -75,  -2,  -3,  -4,   6,   5, -12, -31, -68,  -2,  -1,  -4, -18,  -2,  -2,  -4, -18,  -1,  -4,  -3,  10, -11,  -2,   8,   9,   3,   7,   6, -11,  15,  -4,   5,  13,  -1, -12, -11, -14,  17,   1,   3,   8,  -4,   5, -21,  -5,   7,   0,   2,   3,   0,   0,  -6,  -2,   0,   0,   1,   2 },
      {   0,  -3,   3,   4,  -7,   4,  -2, -11,   0,  -1,  -6,  -2,  10,  -6,  -8, -11,  -1,  -2,  11,  13, -14, -10, -11,   4,  -3,   0, -22, -13,   2,  -8, -11,   7,   1, -17,  27,  12,   2,  -8,  -7,   3,   3,  53, -14,  37,  -1,  -6,   1,  -3,  -2, -71,   5,  29,   0,   3,   1,  -1,  -4, -48,   6,  11,  -3,   4,   2,   2 },
      {  -3,  -4,  -1, -12, -44, -12,  -2,   8,   4,   6,   5,  17,  61, -12,   6,  -4,   1,   1,   5,   4, -33,  28,  10,  -2,   0,  -1,   1, -20,   7, -21,  19,  -3,   2,   0,  -3,  32,  30, -35,   9,  -2,   4,   5,  34,  22,   2, -15,  -3,   0,   6,  27,  18,   1,   2,  -3,  -3,   1,   2,  10,   4,  -5,  -1,   1,  -2,   0 },
      {   2,   3,   2,  -4,   3,  40,  13, -10,   0,   1,   2,   8,   8, -25, -11,   5,  -4,  -4,  -7,  -9, -23, -35, -63,  33,  -2,  -3,  -1, -11,  -2,  -6, -60,  24,   0,   5,  -3,  12,  16,   3, -19,   3,   2,  -8,  15,  -3,   9,   6,   3,  -6,   4,  20,   9, -10,   1,   3,   7,  -6,   2,  11,   6,  -4,  -4,   3,   5,  -2 },
      {  -2,  -3,  -4,  -2, -23, -44,  37, -11,   1,   1,   3,  -2,   9,  37, -24,  20,   0,   0,  -2,  -4,  18,  -9, -31,  28,  -3,  -5,  -2,   4, -32, -61,   9,  12,  -3,  -5,  -7, -24, -28, -28,  14,   2,  -2,  -5,  -8, -11,  -8,  -4,   4,  -2,  -2,  -4,  -2,   2,   2,   3,  -3,  -1,   2,   0,   1,   0,   2,   2,  -1,  -1 },
      {  -1,  -1,  -4,  -6, -12,   7,   0,  -1,   1,   3,   2,  15,  16,  -1,   2,  -5,  -4,  -5, -11, -22, -39,   3,  -2,  -3,   3,   6,  13,  31,  26, -11,  -4,  -1,  -3,  -5, -23, -23, -10,  19,  -3,   2,  -3,   8,   7,  -1, -69,  27,  -2,   1,  -4,  -8, -24,  37, -33,   6,   3,  -4,  -3,  14, -25,  13,  -3,  -4,   4,  -1 },
      {   1,   1,  -2,   4,  13,  12,  12,   0,  -1,   0,   0,  -6, -15, -18,  -6,   3,   0,   0,  -5,   8,  21,   0,  -9,  -7,   0,   3,  10,  -9, -14,   0,   5,  -5,  -3, -11, -32,   5,   6, -18,  12,  -1,   8,  24,  55,   0,  25, -24,   5,  -1,  -5, -23, -47,  24,  18,  -8,   3,   2,  -6,  32, -51,  10,   8,  -2,   2,  -1 },
      {   0,   1,   1,  -1,  12,   8,  -7,  -4,  -1,  -1,   1,   8, -20, -16,   6,   0,   1,   1,  -1,  -6,  27,   1,   3,   9,   1,   3,   8,  28, -27,   7,  -1,   8,  -1,  -2, -10, -42,  25, -23,   8,   2,  -1,  -5,   6,  52,  -2, -21,   5,  -2,   5,  40,  22,  51, -20, -10,   2,  -5,   3, -22,  22,  16,  -6,   0,  -1,  -1 },
      {   1,   1,   3,  -3,   5,  39,  13,  -6,  -2,  -3,  -4,   1, -12, -59,   7,   2,   0,   0,   3,  -6,  -2,  27, -26,   3,   1,   0,  -4,   2,  13,   0,  25, -13,   0,  -2,  10,  -5, -27, -44,  46, -10,  -4,  -6, -23, -18, -30, -17,  20,   0,  -1, -11,   0, -22,  -4,  -4,   2,   3,   0,  -4,  11, -11,   3,   3,  -3,   1 },
      {   0,   0,  -1,   0,   2,   0, -10,   2,   0,  -2,   0,  -1,  -1,  -4,  12,   5,   1,   0,  -1,   2,  -2,  11,   2,  11,  -1,  -2,   2,  -8,  10, -11,  -8,   5,  -1,   1, -13,  15, -16,  -3,  -6,   1,   0, -15,  23, -26,  -1,   3,  -5,   0,  -1,  34, -42,   0,  13,  -3,  -1,  -1,  -6, -97, -10,  21,   7,   4,   1,   0 },
      {   2,   2,   2,   2,  -4,  12,  57,  -4,   0,  -1,   0,   2,  12,   6, -39,   0,  -2,  -3,  -3,  -6, -12, -28, -23, -30,   1,  -1,   2,   4,   3,  16,  38, -48,   2,   3,   2,   1,  14,  40,  23, -25,   1,  -2,   5,   8,  15,   9,   6,  -5,   2,  16,   4,  11,   8,  -8,  -1,   2,   0, -25,   8,  11,   0,  -4,   0,   4 },
      {   0,   0,   0,   0,   2, -13,  -4,  41,   0,  -1,  -1,  -1,  -6,   5,   8, -31,  -1,  -2,  -2,  -4,   3,   2, -29, -77,  -3,  -4,  -4,  -2, -13, -22, -48, -41,  -1,  -2,  -2,  -9,  -5, -30, -21,  -7,  -1,  -3,  -5,   0,  -8,  -8,   2,   5,   1,   4,   2,  -5,  -3,   3,   5,   4,  -1,  -5,   1,  -2,   2,   3,   3,   2 },
      {  -1,  -1,  -1,  -4,  -3,  -1,  -7,   0,   1,   1,   2,   6,  12,  -3,   1,   0,  -1,  -2,  -2, -12, -19,  -1,   2,   2,   2,   3,   7,  24,  42,  -6, -11,  -1,  -3,  -4,  -8, -38, -47,  11,  -9,   9,   0,   0,   4,  25,  20, -28,  -5,   2,   4,  -1,  15,   0,  69, -33,  -2,   1,   2,   3,   5, -13,  27,  -9,  -1,   0 },
      {  -1,  -1,  -1,  -1,  -6, -21,  -6,  -3,   1,   1,   2,   0,   9,  28,   9,  -1,  -1,  -2,  -3,   1, -17, -51,  10,   4,   1,   1,   1,  -5,  14,  39, -16,   1,   1,   1,   2,  14,  -3, -18,  19,  -3,  -2,  -9,  -8, -30, -16, -64,  37,  -5,   0,   0,   4,   8, -17, -31,  18,  -4,   0,   0,  -6,  22,  -8,  -9,   1,  -3 },
      {   1,   2,   2,   4,  -1,  13,  26,   3,  -2,  -2,  -3,  -3,  -2, -26, -46,   4,   1,   2,   2,   4,  -2,  19,  21,  -7,   0,   0,  -2,  -3,   9,  -7,   5,   4,  -1,   2,   7,  15, -24, -14, -51,  20,  -3, -10, -20, -35,  15, -17, -27,  14,   1,   8,  21,  37, -12, -13,  -7,   0,   0,  13,  -3,  41, -14,  -5,   4,  -3 },
      {   1,   1,   0,   0,   1,   3,  35,  18,  -1,  -2,   1,  -2,  -1,  -2, -47,  -2,   0,   0,  -2,  -2,  -2, -11,  28, -11,   1,   1,   4,   3,  -6,  23,   4,  29,  -1,   0,  -8, -13,  19, -16, -27,  32,  -1,  -1,  12,  22, -37, -21,  -1,  11,  -2,  -5, -28, -43,  -4,  -3,   4,   1,  -1, -17,  11, -41,   9,   3,   4,  -1 },
      {  -1,  -1,  -2,   0,  -3,  -1,  -7, -40,   1,   1,   3,  -1,   3,   0,  -3,  33,   0,   0,   0,   1,  -5,  -1,  19,   2,   0,   0,   2,  -1,   4,   4, -20, -44,  -2,  -3,  -7,  -2, -15, -13, -20, -22,   1,   3,  20,  -5,   6,  -6,  -7,  -5,  -4,  -7, -44,  13,  -8,  13,   1,   3,   7,  14,  75, -18, -19,   4,  -3,  -2 },
      {   1,   1,   1,   1,   3,  -2,   8,  47,   0,   0,   1,   0,  -2,   0,   3, -42,   0,  -1,   0,   0,  -1,   8,  -8,  -8,   1,   2,   1,   0,   7, -15,  11,  48,   1,   1,   2,  10,  -9,  22,  25,  17,   1,   2,  12, -19,  19,   8,  26,  -5,  -1,  -4, -27,  30,   0,  -9,   8,  -8,   5,  12,  64,  -2,  -6,  -9,  -1,  -1 },
      {   0,   0,   1,   1,   2,   5,  -1, -15,   0,  -1,  -1,  -2,  -4, -16,  -2,  14,   2,   2,   3,   3,  11,  27,  21,   0,  -2,  -4,  -5,  -6, -20, -48, -13, -12,   1,   3,   5,   5,  26,  42, -17, -13,   0,  -2,  -3,  -4, -18,  -8,  36, -14,   0,  -7,   3,  -8,  -4, -60,  36,  -3,   0,  -2,  -1,  -9,   9, -30,  14,  -4 },
      {   0,   1,   0,   0,  -1,   4,  21,  -3,   0,  -1,  -1,   0,   3,  -4, -28,  -1,   1,   1,   0,  -1,  -5,  12,  52,  -8,   0,  -1,   0,   4,  10,  -1, -39,  10,  -1,   0,  -1,  -9, -17, -15,  19, -16,   2,  -2,  -1,  10,  20,  26,  55, -36,   3,   6,  11,  -3,  10,  41,  17, -21,  -1,  -2, -19,  10, -12,  19,   7,  -6 },
      {   0,   0,  -1,  -3,  -5,   1,  -3,   8,   0,   0,   0,   4,   2,  -4,   9, -10,  -1,  -1,  -1,  -5,  -8,   7, -10,   6,   1,   1,   1,   8,  10, -17,  15,  -1,  -2,  -2,  -4, -16, -30,  27,  -9,  -3,   2,   1,   7,  25,  43, -36,  -6,  -4,  -2,  -2,  -9, -37, -55,  -3,  13,  -4,  -5,  -8, -19,  -8, -62,  15,   7,  -6 },
      {   0,   0,   0,   0,   2,  -2,   3,  23,  -1,  -1,  -1,  -1,  -4,   1, -10, -57,   1,   1,   1,   0,   5,  -2,   5,  44,   0,   0,   0,   0,  -2,   6,  11, -10,  -1,  -2,  -2,  -3,   2, -11, -28, -67,  -2,  -3,  -2,  -7, -13, -10, -26, -50,   0,  -3,  -2,   3,   6,  -3, -12, -24,   0,   2,  -3, -14,   8,  -6,   3,  -7 },
      {   0,   1,   1,   0,   1,   2,   7,  10,  -1,  -1,  -1,   0,  -2,  -5, -20, -15,   1,   2,   1,   1,   0,   7,  27,  16,  -1,  -2,  -1,   0,  -3, -16, -45,   4,   1,   2,   1,  -1,   8,  23,  39, -30,  -1,  -1,   0,   8, -11, -18, -12,  29,  -2,  -2,  -8, -16,   0, -10, -53,  49,   0,   6,   5,  27, -12,   7, -23,  24 },
      {   0,   1,   1,   2,   0,   3,   2,   0,  -1,  -1,  -2,  -4,   1,  -2,  -6,   0,   0,   0,   1,   2,  -2,   3,   8,   4,  -1,  -1,  -1,  -5,   3,  -9, -14,  -7,   1,   1,   2,   8,  -4,  22,  11,   2,  -1,  -3,  -6, -23,  13, -27,   6,  14,   1,   9,  13,  51, -27,  15, -14,   3,  -1, -12, -22, -83,  31,  29, -11,   1 },
      {   0,   0,   0,   1,   0,  -3,   5,  16,   0,   0,   0,   0,   1,   4,  -5, -26,   0,   0,   0,   0,  -2,  -7,  10,  43,   0,   0,   0,   0,   6,  14,  -5, -30,  -1,  -2,  -3,  -3, -14, -33,  -4,   1,   1,   4,   4,   6,  25,  48,  18,  46,   1,   2,   2,   4, -19, -21,  28,  38,  -1,  -5,  -7, -18,   1, -34,  20,  18 },
      {   0,   0,   0,   0,   2,   2,  -4,  10,   0,   0,  -1,   0,  -3,   0,   7, -17,   0,   0,   1,   0,   3,   5,  -8,  28,   0,   0,  -1,   1,  -5,  -8,  26, -19,  -1,   0,  -1,   0,   9,  20, -40,  -1,   0,  -1,   0,  -1, -21, -27,  47,  33,   2,   4,   1,   3,  34,  42,  27,  18,   0,   8,   6,  20,  -4,  48,  -5,  19 },
      {   0,   0,   0,   0,   1,   1,   4,  10,   0,   0,   0,   1,  -1,  -1,  -5, -20,   1,   1,   1,   1,   4,   5,  13,  32,  -1,  -2,  -2,  -2,  -5,  -8, -22, -50,   1,   3,   2,   4,   7,  15,  38,  45,  -1,   0,   0,   3,  -5, -15, -28,  21,   0,  -3,  -3,  -9,   3,  18,  10, -47,   1,   4,   9,  17,  13,   1,  32, -44 },
      {   0,   0,   0,   2,   3,  -2,   2,   0,   0,  -1,  -1,  -3,  -5,   2,  -1,   3,   0,   0,   0,   1,   2,  -2,   3,  -8,   0,   0,   0,  -2,  -4,   3,  -5,  14,   1,   0,   1,   5,   8,  -3,  21, -15,  -1,  -1,  -4, -11, -24,   7, -30,  16,   3,   3,   8,  24,  51, -14,  25,   7,  -2,  -4,  -9, -33, -79,  17,  37, -10 },
      {   0,   0,   0,   1,   2,  -1,  -2,   5,   0,   0,   0,  -1,  -3,   1,   0,  -8,   0,   0,   0,   1,   2,  -1,  -3,  16,   0,   0,   0,  -2,  -4,   2,   1, -34,   0,   1,   0,   2,   4,  -6, -14,  44,   1,   2,   2,  -2,  -9,  19,  38, -13,  -1,   0,   0,   4,  17, -27, -53, -27,  -3,  -3, -10, -16, -50,  -4, -55,  -9 },
      {   0,   0,   0,   0,   0,   2,  -1,  -2,   0,   0,   0,   0,   0,  -4,   4,   3,   0,   0,   0,  -1,   0,   0,  -5,  -5,   0,   0,   0,   0,  -1,  -5,   5,  10,   0,   0,   0,   0,   0,   8, -10, -15,   0,   0,  -1,   0,  -3, -25,  22,  20,   1,   4,   3,   4,  12,  58, -26,  13,  -2,  -8,  -5, -10, -16, -88,  10, -39 },
      {   0,   0,   0,   0,   0,   1,   1,   3,   0,  -1,  -1,  -1,  -1,  -3,  -2,  -2,   0,   0,   0,   0,   1,   3,   3,   8,   0,  -1,  -1,  -1,  -1,  -4,  -6, -17,   0,   2,   1,   2,   3,   8,  14,  40,  -1,  -3,  -2,  -5,  -6, -19, -22, -49,   0,   3,   2,   5,   8,  31,  21,   9,   0,  -2,  -3,  -7, -15, -39,   0,  85 },
      {   0,   0,   0,   0,   0,   0,  -3,   0,   0,   0,   0,   0,   0,   1,   6,   0,   0,   0,   0,   0,   0,  -1,  -5,   2,   0,   0,   0,   1,   0,   2,   7,  -9,   0,   1,   0,   0,   1,   0,  -9,  25,   0,  -1,   0,  -1,   0,   0,  25, -40,   0,   1,  -1,   2,  -3,  -8, -57,  42,   2,   2,   4,   2,   4,  22,  88,  -5 },
      {   0,   0,   0,   1,   0,   0,   2,   5,   0,   0,  -1,   0,   0,   0,  -4,  -8,   0,   0,   0,   0,  -1,   0,   1,   7,   0,   0,   0,   0,   0,   0,  -2,  -9,   0,   0,   0,   1,   1,   1,   5,  24,  -1,  -1,  -2,  -1,  -3,   1, -16, -47,   1,   2,   2,   4,   2,   6,  36,  78,   0,  -1,  -1,   2,   2,   6, -41, -63 },
    },
    {
      { -67,  64, -20,   4,  -2,   2,  -1,   1,  62, -53,   9,   3,  -2,   0,   0,   0, -18,   8,   9,  -7,   2,  -1,   1,   0,   3,   3,  -7,   2,   1,  -1,   0,   0,  -2,  -1,   2,   1,  -1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -50,   1,  43, -24,   8,  -3,   3,  -1,  11,  37, -60,  24,  -3,   0,  -2,   1,  31, -50,  31,   4, -10,   5,  -2,   1, -16,  18,   3, -17,  10,  -2,   1,   0,   5,  -2,  -8,   9,   0,  -2,   1,   0,  -2,   1,   4,  -2,  -2,   2,   0,   0,   2,  -2,  -1,   1,   0,   0,  -1,   0,  -1,   1,   1,  -1,   0,   0,   0,   0 },
      {  -6,  43, -53,  27,  -8,   3,  -2,   1, -45,   4,  33, -11,  -2,   3,  -1,   1,  66, -45,   5,  -7,   5,  -2,   0,   0, -28,  15,   4,  -1,   1,  -2,   2,  -1,   5,   2,  -7,   1,   0,   0,   0,   0,  -1,  -1,   1,   1,  -1,   0,   0,   0,   2,  -1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  73,  14, -11, -29,  11,  -5,   1,  -2,  17, -39, -12,  39, -16,   4,  -2,   2,  -7, -13,  41, -30,   4,   3,  -1,   0, -28,  34, -25,   3,  12,  -8,   2,  -1,   7, -10,   2,  10, -10,   3,   0,   0,  -4,   3,   2,  -5,   3,   1,  -1,   0,   0,  -2,   0,   2,   0,  -1,   0,   0,  -2,   2,   0,  -1,   0,   0,   0,   0 },
      {   1,  -1,  34, -41,  20,  -5,   2,  -2,   9, -36,  22,   7,   0,  -4,   2,   0,  41,  26, -59,  24, -14,   3,  -1,  -1, -54,  10,  27,  -7,  -2,   4,  -3,   1,  23,  -3, -13,  -2,   4,  -2,   0,   1,  -6,  -1,   3,   2,  -2,   0,   1,   0,   1,   1,  -1,  -1,   0,   1,   0,   0,  -2,   0,  -1,   1,   0,   0,   0,   0 },
      {  15, -42,  -3,  41, -20,   4,  -1,   1,  67,   8,   8, -51,  18,  -2,  -1,  -1,  -4, -28,   3,  19,   3,  -7,   3,   0, -40,  38,  -6,  -2,  -8,   5,   0,  -1,   7,  -7,  -5,   3,   4,   0,  -2,   1,  -3,   0,   5,  -1,  -2,   0,   1,   0,  -1,   0,  -1,  -1,   2,   0,   0,   0,  -2,   1,   0,   1,  -1,   0,   0,   0 },
      {  49,  41, -11,  -1, -19,   9,  -6,   1,  24,  -3, -27,  -1,  18,  -8,   1,   0,   0, -19,  -5,  34, -28,   6,   0,   0,  16, -31,  46, -38,   8,   6,  -5,   2, -28,  27, -21,   2,  16, -14,   6,  -1,   5,  -6,   2,   7, -11,   4,   1,  -1,  -5,   2,   1,  -3,   3,   1,  -2,   1,   1,  -1,   0,   1,   0,   0,   1,   0 },
      {  -3,  36,  11,   1, -24,   9,  -3,   0, -51,  -1,  -9, -14,  35, -11,   2,   1, -57,   7,   6,  28, -23,   1,   5,  -2, -16,  32,  -7, -17,   3,   9,  -6,   1,  46, -34,  -2,  11,   2,  -6,   0,   2,  -4,   7,   7,  -8,  -3,   3,   1,  -1,   4,  -2,  -4,   2,   3,  -2,   0,   0,   0,   0,   1,   0,  -2,   1,   1,   0 },
      {   2,  11, -44,  -1,  37, -24,   9,  -3,  16,  44,  -3,   9, -49,  23,  -6,   0, -42, -11, -20,  16,  19,   3,  -7,   6, -14,  -9,  35, -12,  -2,  -6,   6,  -1,  38, -17,  -1,  -6,   4,   4,  -3,  -1, -17,   5,   2,   4,   3,  -5,   2,   0,   8,  -1,  -1,  -1,  -3,   3,  -1,  -1,  -4,  -1,   2,   1,   1,  -1,   0,   0 },
      {   9,  17,   3,  25, -47,  28, -12,   4,  14,  20, -41,  20,   8,  -4,  -4,   3,  18,  37,  -4, -39,  21, -18,   9,  -5, -35, -37,  18,  23, -13,   5,   1,   0,  31, -11,   3, -12,  -1,   3,  -2,   2, -17,   7,  -1,  -2,   6,  -4,   1,   0,   1,   1,  -2,   2,  -1,   0,   1,   0,  -1,  -1,   0,  -1,   0,   1,   0,   0 },
      { -25, -37, -10,  19,  -6,  10,  -3,   3, -19, -17,  12,  16,  -7,  -1,   1,   0, -29,  15,  22, -27,   3,   5,  -2,   0, -37,  -8,  29, -28,  29, -15,   4,  -1,   3,  40, -49,  28, -10,  -5,   7,  -4,  16, -20,  14,   1, -11,  11,  -6,   1,   3,   5,  -1,  -7,   6,  -3,  -1,   1,   2,  -2,   0,   1,  -1,  -1,   1,  -1 },
      {   3,  33,  22, -13,   5,  -4,  -2,  -1, -14,  20,  -2,  -9,  -3,   2,   1,   0, -38, -15, -11,  -3,  23, -12,   5,  -2, -50,   7,  -2,  34, -29,   8,   1,  -1, -42,  30,  -1, -20,   7,   8,  -7,   3,  50, -25,  -4,   9,   1,  -7,   1,   1,  -6,   1,   6,  -5,  -2,   2,   2,  -1,   4,   0,  -3,   1,   2,  -1,  -1,   1 },
      {  -8,  -5, -42,  -7,   6,  10,  -1,   1,  13,  38,  -2,  12,  11, -27,   9,  -3,   2,  31,  -6,   8, -45,  36, -10,   1, -39,  -6, -22,   2,  35, -12,  -6,   6, -28, -12,  39,  -3, -14,  -5,  10,  -4,  20,  -5,  -9,  -7,   8,   6,  -5,   0,  -3,   2,   4,   4,  -2,  -5,   2,   1,   3,  -2,   0,  -2,   0,   3,   0,  -1 },
      {  -4,  -6,  -3,  38,  -5, -27,  14,  -4, -19, -26, -38,   4, -19,  47, -20,   6,  12,  36,   3,  19,   3, -22,   1,   3, -18,  18,   6, -32,   4,   4,   6,  -5, -46, -28,  22,   0,  14,  -4,  -4,   2,  16,  12, -10,   2,  -6,  -2,   6,  -2,   1,  -4,   2,  -1,   1,   3,  -4,   0,   3,   2,   0,  -1,   0,  -3,   1,   0 },
      {  14,  35,  25,  11,  -6,   3, -11,   4,  14,  36,   8,  -9, -20,   6,   2,  -2,  11,  22, -35, -12,  18,  -6,  -2,  -1,  24,  27, -10, -16,  18, -21,   8,  -3, -15, -20, -22,  46, -35,  15,   3,  -2,  17, -16,  26, -24,   4,   8,  -9,   5, -13,   4,  -7,   0,  11,  -8,   5,   0,   2,  -1,   0,   2,  -3,   2,   1,  -2 },
      {  -9, -11,   5,  -4, -17,  43, -29,   9,  -7,  -6,  18,  28, -14, -24,  15,  -1, -23, -29, -25,  -8,  26,   2,   5,  -6,  -3,  22,  27, -18,   0,  -4,  -4,   2, -51, -26,  14,  -1,   1,   0,   6,  -1, -31,  33, -15,   3,   4,  -2,  -1,   0,  21, -10,   0,   5,  -6,   2,   0,   0,   1,   2,   0,  -3,   1,   1,  -1,   0 },
      {  -1,  15,  -3, -23,  10, -19,  20, -10,  -1,  23,   8, -24,  21,   5, -11,   3,  -3,  25,  19,  -4,  -1, -11,   7,   0, -21, -11, -13,   6,  -6,  11,   0,  -3, -43,  15, -20,  24,  -3,  -3,  -3,   3, -71,  10,  17,  -7,  -4,   5,   2,   0,  43, -13,  -6,   2,   3,  -1,  -3,   1,  -9,  -1,   5,  -1,  -2,  -1,   2,   0 },
      { -10, -19, -52, -28,  -1,  22,  -5,   4,   5,  14, -16,  31,   6, -13,   1,   1,  11,  26,  -5,   9,   9, -17,  14,  -5,  23,  45,  -6,  -3, -38,  32, -12,  -1,  -3,  -2, -22,   3,  27, -11,  -7,   7,  10, -31,  28,   1, -12,  -1,   7,  -3,   0,   6,   0,  -7,   6,   2,  -3,  -1,  -6,  -1,   0,   3,   0,  -3,   1,   0 },
      {  -3,  -8, -17,  16,   5,  -1,  -8,   2, -17, -37, -55,  -8,   3,  -8,  24,  -5,  -2,  -8, -20,  27,  10,  26, -25,   8,   1,   7,   3,  26, -12, -24,  15,   0,  -8,  17,   4,  -4, -21,  28,  -7,  -6, -25, -23,  34, -18,  23,  -9,  -9,   8,  16,   9, -11,   5,  -4,  -7,  10,  -4,  -9,  -2,   7,   1,  -2,   4,  -3,   0 },
      {   4,  14,  23,  43,  -5,  -4,  -1,  -1,   0,  12,   4,  25, -22,   5,  -7,   3,  -7,  10, -20, -11, -32,  21,  -6,  -1,   8,  41, -22,  11,  10, -11,  -5,   3,   1,  31, -10, -40,  24, -16,  13,  -3, -46, -20,   1,  28, -14,   4,   2,  -2,  -5,   1,   6, -14,   2,   5,  -3,   3,   4,  -2,  -3,   1,   4,  -3,   2,   0 },
      {  -8, -13, -12, -15, -37,   0,  32, -14,  -1,  -5,   7,  30,  15,  22, -48,  18, -15, -35, -37,   8,  -5, -18,  28,  -2,  -2,  -8,  -8,   6,  19,   4,  -1,  -8,  -3,  15,  22,   7, -25,   3,  -6,   7, -21, -17,   9, -16,  11,   5,   1,  -2, -32,  18,  -2,   2,   6,  -9,   1,   1,   6,  -3,  -2,   1,  -4,   4,   2,  -2 },
      {   5,   0,   3,   6, -22,  -3,  16,  -8,   8,   1,   4,  17,   4,   5, -18,   8,   3,  -8, -21,   1,  -3,   5,   5,   2,  14,   5,  -7,   4,  11,  -6,   1,  -3,  16,  12,   5,  -4,  -4,  -2,  -1,   1,  43,  -3,  -4,  -1,   0,   3,  -1,  -1,  93, -25,   0,   1,  -3,  -2,  -3,   2, -43,  12,   5,  -4,   3,  -1,   2,  -1 },
      {   2,   1,  14,  10,   6,  19, -45,  28,   0,  -3,  14,   9, -39,  17,  18, -12,   2,  -5,  13,  39,  -4, -31,  15,  -8, -14, -33, -42, -18,  10,  27, -12,   5,   9,  17,  24,   3,  -1,  -9,  -1,   2, -10, -26,  16, -15,   5,  -2,   5,  -2,  20,  11, -11,   4,   3,  -2,  -1,  -2,  -1,   2,   1,   0,  -3,   0,   0,   1 },
      {   1,  -3,  16,  27,   6, -17,  12,  -9,   1,  -7,  20,  22,  -6,  -5,  -2,   1,  -5, -21,   4, -19, -26,  24,  -4,   1,  -5, -26,  12,  23, -22,  17, -14,   6, -19, -52, -10,   2,  26, -16,   7,   3,  11,  -8,  46, -28,  -4,   2,   3,  -4,   2,  22, -32,   5,  10,  -4,   0,   0, -12,   1,   5,   3,  -7,   2,   1,   0 },
      {   0,  -3, -13,  26,  36,  -9,  -6,   1,  -5, -14, -37,  -2,  -3, -10,   8,  -2,  -6, -18, -43, -16,  -8,  -6,  20, -13,  -3,  -8, -22,  21,  -2,  30, -30,  11,   4,  13, -15,  33, -15, -21,  16,   2,   6,  36,  -9, -10, -12,  17,  -2,  -6,  -9, -31,  22,  -6,  10,  -3,  -6,   8,   5,  10,  -4,  -3,   0,  -3,   5,  -1 },
      {  -3,   2,   2,  -7,  -2,   0,  -2,  -2,  -6,  -3,  -4, -14,   0,  -2,   8,  -2,  -3,   0,  -2,  -7,  22,   5,  -5,   3,  -6,  -6,  -4, -10,   9,   2,  -2,   2,  -4,  -3,  18,   4,   4, -14,   8,  -4, -12, -13,  -1,  14, -15,  11,   0,  -2, -40,   3,   7,  -7,   5,   0,  -4,   1,-107,  14,   4,   1,   3,  -4,   4,  -2 },
      {  -1,   1,   3,  22,   4,  -1,  -6,   1,   3,  12,  20,  35,   0,  -1,  -3,  -2,  -1,  12,  16,  21, -35, -10,   1,  -3,   0,  11,   6,   1, -24,   5,   7,  -5,  -3,  24, -26,  -3,  -9,  27, -19,   6,   3,  35, -10, -31,  35, -26,   3,   5, -30, -28,   3,  21, -13,   0,  11,  -6, -38,  13,   2, -11,   4,   6,  -5,   2 },
      {  -3,  -4,  -3, -13, -48,  -9,   6,   0,   2,   3,   8,  22,  -8,  12,  16, -14,  -1,  -7, -14,  21,  18,  21, -49,  29, -10, -22, -42,   0, -19,  -4,  30,  -6,   2,   3, -13,  26,  28, -13,  -4,  -6,   2,  20,   4,  -5, -20,   5,   0,   2, -12, -25,  17,  -4,   7,   6,  -5,  -1,  15,   6,  -1,  -2,   1,  -7,   1,   2 },
      {  -5,  -9, -16, -31,  -8,  24, -29,  27,  -3,  -5, -12, -26, -16,  37, -17,   1,  -5,  -8,  -9,  -3, -41, -14,  14,  -9,   0,  -9, -12,  34,  20, -32,  14,  -7, -10, -24, -41, -18,  14,  15,  -2,   1,   7,  24,  12,  14, -15,  -1,  -4,   4,  -2,  -1,  -8,  -4,   7,   2,   3,  -1, -14,   3,   4,   0,   2,  -2,   1,   0 },
      {  -1,  -3,  -6,   0,   5, -40, -21,  46,   4,   6,   2,  18,  51,  27,  23, -49,  -7, -11, -22, -40, -19, -23, -12,  21,   4,  10,  11,  -8,   5,   9,  12,  -8,   0,   3,  13,   5,  -3,   0, -10,   1,  -2,  -7, -10,   3,  -4,  -7,   9,   3,   7,  16,  -9,   5,  -2,   8,  -3,  -5,  -4,  -7,   2,  -2,   0,  -2,   0,   5 },
      {   2,  -2,   5,  19,  16,  -3,   1,   1,   2,  -3,   9,  29,  21, -13,   0,  -4,   0,  -8,   9,  22,  -7, -13,   7,  -4,  -2, -23,  -3,   5,  -7,  -8,   9,  -2,  -7, -44, -14,   7, -22,  36, -14,   2, -11, -36,  -7,  34,  -9,   1,  -6,   9,   7, -13,  48, -42,  13,  -2,   2,   0,   0,  18, -23,  10,   5,  -6,   3,  -2 },
      {  -4,  -6,  -3, -25, -47, -23,  -1,  10,  -2,  -3,   0, -26, -36,  18,  15, -13,  -1,   1,  10, -16, -23,  15,   2,   0,   5,  10,  30,  11,  -3,  20, -30,  14,  -2,  -9,   3,   5, -10,  -6,  13,  -2,  -9, -34,  -4, -14,  31,  -7,   2,  -1,   0, -32,  33,  -2,  -7,  -1,   2,  -1,   6,  26, -23,   0,   4,   2,   0,  -1 },
      {  -1,   0,   0,  -3, -17, -25,  -7,  17,   3,   7,  12,  24,   7,  14,   8, -11,   5,  11,  25,  43,  16,  -9,   1,  -6,   3,  15,  24,  41,  -8, -10, -21,  14,   0,   7,   0,  -8, -27,  -9,  37, -18,   3,  26,  -9,   0, -13,  36, -26,   2,   1,  16, -10, -12,  24, -20,   0,  11,  -4, -14,   3,   4,  -9,   2,   9,  -8 },
      {  -2,  -2,  -5,  -2, -32, -28,  11, -12,  -2,  -2,  -7,  -1, -27, -20,  20,  -5,   0,  -2,  -9,   3,  -7,   5,   9,  -5,  -3,  -7, -27,  -8,  -2,  16, -19,  10,  -5,  -8, -36,  13, -11,  20,  -6,   6,   1,  11, -37,  28,  18, -19,   8,  -3,  10,  57,  -5, -13, -12,   6,   1,  -3, -13, -36,  20,  -1,   5,   4,  -6,   5 },
      {   2,   4,   7,   5,  12,  50,   6,   1,  -2,  -2,   0,  -9, -22,  13, -18,   0,   0,   2,   8,  -1, -33,   4, -27,  33,   4,  12,  24,  29, -22,  22,  18, -24,   2,   8,  18,  42, -15, -15, -12,   6,  -3,   0, -30,  -3,  -2,   5,   8,  -1,  10,  28,  12, -22,  14,  -4,  -6,  -1,  -8, -19,   0,   7,   0,  -7,   7,   1 },
      {  -1,  -1,  -1,  -5, -19, -18,   8, -21,  -1,   1,   0,  -7, -28, -32,  34,  -8,   0,   2,   2, -11, -19, -19,  12,   6,   2,   8,  13,  -4,  16,  12,  11, -17,   3,  19,  28, -18,  -7,  24, -35,  16,   5,  32,  33,  -1, -26,   3,   5,   1,  -2,  -8, -13, -32,  39,  -6,  -1,  -2,   2,   9,  -6,  18, -11,  -5,   5,   0 },
      {   2,   3,   4,   8,   1,   7,  32,  22,  -2,  -4,  -2,   4, -11, -31,  -5, -35,   3,   3,   7,  25,  20, -14,   4,  32,   4,   3,  10,  37,  51,  12,  -5, -25,  -5, -15, -20,  -9, -13, -20,  -8,  16,   1, -10,   7,  10,   1,  -1,  22, -12, -10, -21,  -9,  17, -21,  17, -17,   0,   4,   0,  18, -17,   9,  -5,  -2,   6 },
      {   1,   3,   6,   3,  -3,  -8, -16,  29,   1,   4,   6,   6,  13,  17, -14,   7,   4,   7,   8,  12,  19,  56,  24, -27,   1,   4,  12,  17,   9, -10, -17,  10,   2,  13,  21,  32,   4,   5, -27,  12,   4,  19,  26,   6, -12, -35,  32, -14,  -3,  -1,  11, -26,  -5,  23, -11,  -5,  -1, -16,   1,   7,   6, -10,  -4,   5 },
      {   3,   4,   3,  -1,   5,  23,  50,  57,  -3,  -2,  -4,  -9,  -7, -10, -36, -44,   0,   1,  -2,  -9,   7,  17,   3,  14,  -4,  -4, -16, -33, -18,   6, -15,  15,   0,   6,  -2, -21,  -2,  15,  11, -16,   3,  22,  17,   1,  15,  -8, -12,   8,   2,   5,   1, -22,  12, -15,   8,  -1,  -2,   6,  -9,  16,  -8,   3,   3,  -5 },
      {   0,  -1,   0,   1,  -9, -20,   7,  40,  -2,  -5,  -3,  -4, -24, -32, -16,   0,   0,  -1,   1,  -2, -15,  11,  28, -20,  -2,  -2,   4,  -2, -35, -19,  18,  -2,   4,   1,  19,  28,  10,  15, -18,  -2,  -7, -28, -28,   9, -27,  28,  -7,   5,  -6, -23, -25,  26,  11,  -8,   2,   3,   4,  -2,  33, -26,   3,   3,   1,  -3 },
      {   0,   2,  -1,   1,   0, -28, -13,  28,  -2,   0,  -5,  -4,  -5, -32, -14,   9,   1,   4,   0,   6,  15,   8,  15,  -9,   1,   3,  -2,   6,  13,  -9,  29, -27,  -1,   2, -14,  -5,  27, -26, -14,  16,  -3,   9, -20, -37,  15,  18,   3,  -4,   9,  47,  25,   2,  -4, -14,   2,  -3,  -3,  32, -43,  14,  -5,   4,   4,  -5 },
      {  -1,  -3,   1,  -3,   0, -17, -40,   1,   0,  -3,   1,  -5,   4,  -2, -22,   7,   1,  -1,   1,  -7,  10,  26,   8,  21,  -1,  -5,  -5, -17,  -2,  21,   5, -30,  -2, -11, -12, -33, -12, -10,  -7,  18,   1,  -7,   3,  -9,  20,  14,  -3,  -3,  -2, -34,  22, -10,  18, -26,   6,   1,  -4, -64,  28,  10, -12,   8,   5,  -4 },
      {  -1,  -1,  -4,  -3,  -3,  -4, -24,   3,   0,   2,  -2,  -1,   4,  10, -18,  14,   0,   3,   0,   1,   6,  31, -21,   5,  -2,  -2,  -9, -15, -20,  19, -10, -14,  -3,  -1, -15, -19, -55,  -2,  -3,  25,   0,   2, -12,  28,  -9,   1,  23, -17,  -2,   6, -43,  12,  20,  -1, -10,   5,   9,  57, -10,  -7,  -2,  -7,   6,  -1 },
      {   1,   1,   0,   2,   8,  22,  22,  -6,   1,   1,  -1,   3,  11,  28,  23, -15,   0,   0,  -2,   3,  17,  15,  -9, -23,   1,   2,   0,  10,  29,   3, -17,  18,  -1,  -4, -17,  -5,  27,   9, -32,  15,  -6, -18, -47, -23,  -3,  -6,  18,  -1,   0, -11, -13,   1,  42, -35,  12,  -5,   1,   8,  14,   9, -34,  15,  -2,  -1 },
      {  -1,  -3,  -4,  -3,  -2, -20, -16,  15,  -1,  -2,  -3,  -1,  -3, -37, -33,  27,  -1,  -3,  -2,   2,  -1, -32, -35,   4,   1,   0,   0,   8,  11,  -6,   3,  19,   0,   0,   0,  10,   7,   0,  27, -23,  -2, -12, -19, -11,   9, -38,  25,  -5,  -2, -23, -29, -20,  24,  24,  -6,   0,  -3,  -8,  -4,  36, -27,   3,   0,   4 },
      {   0,   1,   1,   3,   1,  -9, -20,  -3,   0,   1,   1,   4,   2,  -1, -16,  10,   1,   3,   0,   3,  10,  17,   9,  19,   2,   5,   4,  18,  30,  50,   3, -13,   1,   3,   5,  25,  23,   8,  -2, -28,   1,   6,   3,  39,   4, -16, -38,  37,  -2,  -2, -19,   2,  -5, -21,  41, -23,   4,  22,  -2,  -1,  -5,  13, -18,   0 },
      {  -1,   0,  -2,   1,   3,   1, -24, -44,  -1,  -1,  -3,   1,  -1,   1, -10, -48,   2,   3,   1,   5,   1,   8,  54,  51,  -2,  -3,  -5,  -1, -20, -47, -21,  -9,   2,   3,   2,  16,  14,  -8,  10,   2,  -2,  -5, -16,   0,   5,  -8,  10,  -1,  -2,  -6, -24,  -6,   1,   1,  -1,   0,   1,  10, -15,  21,  -9,   7,  -3,   2 },
      {  -1,   0,   1,  -2,  -1,  -8, -31, -18,  -1,  -1,   0,  -5,  -4, -18, -44, -15,   0,   1,   2,  -2,   3, -12, -25,  29,   1,   3,   5,   2,   9,  -1, -28,  42,   0,   4,   3,  -5,   8,  31, -14,  -6,  -1,   6,  10, -15, -31,   6,  22,  -5,   6,  25,  40,  18,   9, -18,   3,  -4,   3,  10,  15, -39,   7,   8,  -2,   0 },
      {   0,  -1,  -2,  -1,  -1,   3,  -7, -18,   0,  -1,  -2,  -2,  -1,   4, -12, -22,   1,   1,   1,   2,   7,  26,  19,  -9,   1,   1,   0,  -1,   9,  34,  23,  10,  -2,  -5, -12, -24, -16,  13,   0, -31,  -4, -10, -26, -52, -34,   0, -10,  38,   0,   2, -12,  -9,  -1,  42, -14,  -7,   0,  -6,  -5,   3,  23, -32,   6,   3 },
      {  -1,  -1,  -1,   2,  -5,   0,   0, -10,  -1,  -2,  -3,   0, -10,  -6, -18, -29,   0,  -1,  -1,   3,  -4,  -4, -20, -28,   1,   0,   0,   3,  -5,   0,   5,  27,  -1,  -2,  -1,   1, -23, -23, -26,  17,   1,   1,   5,  19, -26,  10,   3,  10,   1, -13,  -9,  38, -19,  -8,  23, -16, -10, -40, -60,  30,  -5,  15, -12,  14 },
      {  -1,  -1,  -1,  -3,  -1,  -4,  -2, -12,  -1,  -1,   0,  -2,  -1,  -7, -11, -21,   0,  -1,   1,  -1,  -1, -12,  -6, -29,   0,   0,   1,  -3, -11, -21,  13,   5,  -1,  -3,   2,  -8, -24, -64,  -5,  10,   1,   2,   4,  -5,   6, -19,   1,  29,   1,   5,   7, -41,   9,   5,  30, -27,   6,  23,  44, -24,   2,  13, -26,  18 },
      {   0,   0,   1,   1,   1,   3,  19,  25,   1,   1,   1,   1,  -1,   6,  31,  47,   1,   0,   0,   0,  -2,   4,  28,  40,  -1,  -2,  -3,  -4, -11, -11, -12,  -9,  -2,  -3,  -7, -15, -25, -10,   4, -29,  -1,   0,  -2, -10, -36, -12,  27,  18,   3,   6,  23,  27,  15,   5,   7, -31,  -3,  -5,  -8,   3, -16,  24, -29,  25 },
      {   0,   1,   2,   2,   0,   2,  -4, -15,   0,   1,   1,   1,  -2,   2,   2, -17,   0,   1,   1,   1,  -4,   9,  27,  -8,   2,   4,   3,   5,   0,  16,  59,  35,  -1,  -1,  -3,   0, -10, -15,  -5, -43,   4,   5,  15,  31,  29,   0,  24,  -8,   1,   7,  20,  28,  19,  -6, -37,  15,   1,  10,  13,   4, -37,   8,   6,  -2 },
      {   0,  -1,  -1,   0,  -2,  -1,   3, -12,  -1,  -1,   0,  -1,  -5,  -4,  -6, -31,   0,   0,   1,   2,   0,   4,  -5, -37,   0,  -1,   1,   1,  -2,   6,   5, -56,   3,   2,   6,   5,  -1,  16,  57,  14,  -1,  -3,   3,  -1, -33, -52,  -2,  -2,   5,   9,  26,  31,  18,  -4,   0,   9,  -1,  -4,   1,  -7, -19,  17,   2, -11 },
      {   0,   0,  -1,  -1,   2,  -1,   5,   0,   0,  -1,  -2,  -2,   2,  -3,  10,   4,   0,  -1,  -1,   1,   3,  -3,  10,   5,   0,  -1,  -2,   0,   3,  -4,  14,   4,  -1,  -3,  -3,  -2,  -1, -23,  -4, -25,  -1,  -2,  -2,   0,   4, -39,  15, -11,   1,  -6, -20, -12,  33, -36,  11,  19,  -2, -25, -46, -56,  52,  -1,   9, -26 },
      {   0,   0,   1,   1,   0,  -3,  -1,   0,   0,   0,   0,   0,   0,  -6,  -3,   5,   0,   0,   0,   0,   1,  -6,  -9,   0,  -1,  -1,  -1,   0,   0, -21, -39, -10,   0,   0,   0,   2,   2, -24, -34,  -4,   1,   2,   4,  15,  21, -19, -25,  46,   2,   8,  15,  42,  42,  21, -10,  -6,   0,  -5,   1,   3,  -7, -47,  41, -16 },
      {   0,   0,   0,   0,   3,   3,   1,  -4,   0,   0,  -1,   0,   4,   1,  -4,  -8,   0,   0,   0,   2,   5,   3,  -1, -11,   0,   0,   1,   1,   4,   3,   5,  -5,   2,   3,   5,   6,  19,  23,  33,  19,   2,   4,   4,  11,  39,  36,  15,   4,   2,  -1,  -3,   7,  54,  21, -12, -45,  -2, -12, -23, -26,  20,  -9, -25,  48 },
      {   0,   0,   1,   1,   1,   0,   4,   7,   0,   0,   0,   1,   0,  -1,   6,  10,   1,   2,   1,   2,   2,   2,  21,  31,   1,   2,   3,   3,   4,   5,  35,  53,   0,   0,  -1,   0,  -3,  -6,  21,  60,  -2,  -3,  -5,  -2, -11, -31, -43, -32,   1,   3,   8,  17,  19,   7,  25, -12,   0,   3,   9,  10,   7, -25,   8,   2 },
      {   0,   0,   0,   0,   1,   2,  -2,   3,   0,   0,  -1,  -2,  -1,   2,  -2,   4,   0,   1,   0,   0,   2,   5,  -1,  10,   0,   0,  -1,  -1,   1,   2,  -4,   8,   0,   0,  -1,  -2,   0,  -2, -18,  17,  -1,  -2,  -2,  -4,  -2,  -6, -48,   1,   0,   0,  -7, -21,   1,  41, -31,  13,  -4, -10, -30, -57, -53,  52,   7,   8 },
      {   0,   0,   0,   0,   0,   1,  -3,  -3,   0,   0,   0,   0,   0,   1,  -6, -11,   0,   0,   0,  -1,   0,   0,  -6,  -7,   0,   0,   0,  -1,   0,   1, -13, -19,  -1,  -2,  -2,  -4,  -2,  -7, -32, -39,  -1,  -2,  -3,  -4,   4,   1, -38, -64,   2,   5,   9,  15,  33,  35,  23,  31,  -1,   5,   6,  19,  21,   6, -46,  16 },
      {   0,   0,   0,   0,   0,  -2,  -3,   4,   0,   0,   1,   0,   0,  -2,  -3,   5,   0,   0,   0,   0,   1,  -2,  -5,   2,   0,   1,   1,   1,   2,  -2,  -4,   7,   0,   0,   1,   1,   3,  -5,  -6,  14,  -1,  -1,  -2,  -2,  -3, -24, -28,  28,  -2,   0,  -2,   6,   5, -39, -69,  23,   2,   8,  10,  30,  49,  31, -11,  48 },
      {   0,   0,   0,   1,   1,   3,   2,   4,   0,   0,   0,   1,   2,   4,   3,   4,   0,   0,   1,   1,   2,   3,   1,   9,   0,   1,   1,   2,   3,   5,   5,  13,   0,   1,   2,   3,   3,   3,   6,  29,   1,   3,   3,   4,   9,  18,  21,  40,   2,   3,   4,  10,  24,  43,  13,  32,   0,   2,   7,  19,  34,  54, -20, -63 },
      {   0,   0,   0,   1,   1,   0,  -1,  -4,   0,   0,   0,   1,   0,   1,   0,  -5,   0,   0,  -1,   0,  -1,  -1,  -3,  -8,   0,   0,   0,   0,  -1,  -1,  -2,  -4,   0,  -1,  -1,  -1,  -3,  -2,  -7, -17,   0,  -1,  -1,   0,  -3,  -1, -16, -37,  -1,  -3,  -1,   0,   3,   6, -26, -85,   3,   4,   9,  16,  29,  46,  40, -37 },
      {   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,   1,   1,   1,   2,   3,   0,   0,   1,   1,   2,   3,   4,   4,  -1,   0,   0,   0,   1,   2,   3,   3,  -3,   0,   0,   1,   1,   2,   2,   3,  -8,   1,   1,   2,   2,   2,   5,  17,   1,   1,   2,   2,   3,   4,  18,  48,  25,   2,   5,   7,   5,  17,  36,  84,  61 },
    },
  },
  { // 19
    {
      {-115,  21,  10,   6,   4,   3,   2,   1,  44,  -9,  -4,  -2,  -1,  -1,  -1,   0,  19,  -1,  -1,  -1,  -1,   0,   0,   0,  -6,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -24,   6,   1,   1,   0,   0,   0,   0, -93,  10,   9,   3,   4,   1,   2,   1,  78, -12,  -6,  -2,  -2,  -1,  -1,  -1,  10,   5,  -1,   0,  -1,   0,  -1,   0, -21,   0,   2,   0,   1,   0,   1,   0,   8,  -1,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
      {  33,  -8,  -3,  -7,   1,   1,  -1,   0,  49, -10,  -7,   2,  -2,  -1,   0,   0,  75,   5,  -5,   0,  -6,  -2,  -1,   0, -78,  10,   5,   1,   5,   0,   1,   0, -23,  -3,   4,  -2,   1,   0,   0,   0,   1,  -3,   3,   0,   0,   0,   0,   0,   4,  -1,  -1,   0,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
      { -13,   9,  -1,  -2,   1,   1,  -1,   0, -44,   4,   1,   3,   1,   0,   0,   0, -25,   7,   0,   0,  -1,   0,   0,   0, -81,   0,   8,   1,   4,   0,   1,   0,  80, -13,  -4,  -3,  -2,  -1,  -1,   0,   6,   7,  -1,   1,  -1,   0,   0,   0, -12,   1,   0,   0,   0,   0,   0,   0,   3,  -1,   0,   0,   0,   0,   0,   0 },
      {  18,  -8,  -2,   2,  -3,   0,   0,   0,  25,  -9,   0,  -2,   1,   0,   0,   0,  57, -15,  -3,  -3,   1,  -1,   0,   0,  49,  -7,  -5,   0,  -3,   0,  -1,   0,  86,   3, -11,   0,  -3,  -1,  -1,   0, -32,  15,   1,   0,   0,   0,   0,   0, -23,  -2,   2,   0,   1,   0,   0,   0,   3,   0,   0,   0,  -1,   0,   0,   0 },
      {   5,  -6,   2,   8,  -1,  -3,   1,   1,  15,  -5,   7,  -6,   1,   1,   0,   0,   8, -19,  -2,  -4,   5,   0,   1,  -1,  13,   5,  -2,   1,  -4,   1,  -1,   0,  13, -13,  -1,   2,   0,   0,   0,   0, 115,   1, -13,  -1,  -3,  -1,  -1,   0, -31,  16,   3,   1,   1,   0,   0,   0, -17,  -2,   1,  -1,   1,   0,   0,   0 },
      {   1, -25,   4,   1,   4,  -1,   1,   0,   3,  -5,   1,   1,  -2,   0,   0,   0,   6,  -2,   3,  -2,   0,   0,   0,   0,   6, -16,   0,   0,   1,   0,   1,   0,  30,  -1,  -2,   0,  -2,   1,  -1,   0,  23, -12,  -1,   0,   0,   0,   0,   0, 115,   0, -10,  -1,  -3,  -1,  -2,   0, -15,  13,   0,   1,   1,   0,   0,   0 },
      {  22, 108, -24,   8, -11,  -2,  -5,  -1,   3, -39,  14,  -3,   5,   1,   1,   0,  -1, -31,   3,  -3,   4,   0,   1,   0,   1,  14,  -4,   0,  -2,   1,  -1,   0,   2,  -3,   0,   0,   0,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,  23,  -3,   0,   0,  -1,   0,   0,   0,   7,   3,  -2,   0,  -1,   0,   0,   0 },
      {  -2, -15,  -1,   5,  -1,   0,   1,   1,   4,  11,   5,  -3,   0,   0,   0,   0,  -1, -23,   2,  -2,   3,   0,   1,  -1,  -1,   1,  -4,   2,  -3,   0,   0,   0,   0,  -5,  -2,   2,  -1,   0,   0,   0,  16,  -2,  -3,  -1,   0,  -1,   0,   0,  10, -13,   1,  -1,   0,   0,   0,   0, 121,  -2,  -9,   0,  -2,   0,  -1,   0 },
      {   4,  28,  -6, -34,   7,   9,  -3,  -2,  -5, -20, -36,  22,  -2,  -2,   2,   3,  12,  84,   3,  12, -20,  -4,  -3,   0,  22, -36,   7, -10,  16,  -3,   3,  -1,   6, -14,  10, -13,   2,  -2,   1,   0,  25,  -9,  15,   2,   0,   2,   0,   0,  -5,  -6,  -3,   2,  -1,   1,   0,   0,  24,   1,   0,   3,  -2,  -1,   0,   0 },
      {  -4, -34,  14,  15,  -1,  -5,   3,   1, -14,-101,  26, -13,  12,   3,   2,   2,  -6,  29, -19,  -3,   5,   0,  -1,   0,   3,  37,   1,   4,  -9,   1,  -1,   0,  -4, -18,  -4,   3,   2,   1,   0,   0,  -7,   8,  -6,  -2,  -1,  -1,   0,   0,  -1,  -2,   1,   0,   1,   0,   0,   0,   9,   0,   0,  -1,   0,   0,   0,   0 },
      {  12,  19, 100, -19,  14,  -7,   3,  -4,  -3,   3, -51,   5,  -7,   3,  -1,   1,  -3, -21, -14,   1,  -4,   2,   0,   1,   6,  38,  10,  -1,   1,  -2,  -1,  -1,   3, -14,  -2,  -1,   0,  -1,   2,   0,  -2,  -2,   2,  -1,   0,   1,  -1,   0,   5,   1,  -3,   0,   0,   1,   0,   0,   2,   0,   2,   1,   0,   0,   0,   0 },
      { -11, -23, -49, -46,   6,  14,  -3,  -1,  -6, -23, -38,  23,  -9,  -4,   3,   1,  -6, -35,  31,  14, -12,   0,   3,   3,   6,  60,  -5,  -8,  11,  -4,   0,  -1,   9,  14,  16, -10,   2,  -3,   0,   0,   1, -31,  15,   6,   0,   3,  -1,   2,  -3,   9,  -4,   1,  -3,  -2,  -1,  -1,  -2,   5,   1,   2,   0,   0,  -1,   0 },
      {   1,  10, -21,  18, -14,   1,  -1,   0,  10,  55,  12, -11,   2,  -2,   0,  -2,   5,  52, -16,   5,  -2,   2,   0,   1,  15,  84,   5,   2,  -6,  -2,  -2,   0,  10, -33,  -1,   0,   4,  -2,   2,   0,   1,   6,   0,  -3,  -1,   0,  -1,   1,  15,  -3,   3,  -1,   3,  -1,  -1,   0,   1,   3,   2,   0,  -1,   0,   0,   0 },
      {   2,   1,  -7,  61, -17,  10,  -4,   3,  -9,  -8, -67, -22,  -4,   4,   1,   1,   2,  19,  11, -13,   4,  -2,   0,  -1,  -5,  12,   7,   7,  -5,   1,  -1,   0,   2,  72, -18,   6,  -4,  -1,  -2,   0,  13,  15,   5,  -1,   1,   0,  -1,   0,   2, -12,   6,  -5,   3,  -1,   0,  -1,   3,   3,   3,   0,  -1,   1,   0,   0 },
      {  -1,  -8, -19,  31,   1,   8,  -4,   3,  -5,  -7, -48,  -3,  -8,  -2,   0,  -1,   1, -16,  46, -11,   5,  -3,   1,  -1,   2, -14,   9,   0,   3,   1,   1,   1, -14, -84,  -4,   0,  -3,   2,   1,   1, -12,  42, -12,   4,   0,   3,  -2,   0,   5,  13,   4,  -2,   1,  -2,  -1,   0,  -4,  -7,   1,   1,   1,   0,   0,   0 },
      {  -4,  -3, -16, -48,  -1,  -2,   2,  -4,   1,  -1,  -1,   8,   2,   3,   2,   1,  -5, -12, -29,  23, -10,   5,   1,   4,   0,   5,  19,  -9,   5,  -4,  -2,  -1,  -9,  17, -26,   4,  -4,  -1,   1,   0,  12,  96,   1,  -5,  -1,  -1,  -3,   0,  13, -26,  13,  -2,   3,  -2,   0,   0,  -2,  -2,   1,   2,  -2,   1,   1,   0 },
      {  -2,  -6, -29,  19, -13,  10,  -6,   4,  -1,  -2, -35,  -8,   5,   2,   0,   1,  -7, -25, -94,  14,  -8,   7,  -1,   2,   3, -19,  36,  -7,   6,  -4,   2,  -1,   0, -23,  13,  -1,   5,  -3,   1,  -1,  -5, -33,   0,  -1,   0,   1,   0,   0,  -2,  -5,   2,  -2,   1,   0,   1,   1,  -3,  -2,   3,   0,   0,   1,  -1,   0 },
      {   2,   7,  -8,   0,  30,  -4,   1,   2,   0,   0,   3,  10, -21,   6,   0,   1,   1,  11, -18, -18,  -2,   3,   0,   0,   2,   4,  26,  -5,   7,  -2,   0,  -1,   1,  25,  22,  11,  -6,   1,  -2,   2,  -9,  22, -34,   9,  -1,   2,   0,   2,   6,  99,  -2,  -1,  -5,  -3,  -2,  -1,  17,   2,   2,   0,   3,   0,   0,   0 },
      {   5,   2,  12,  31,  36,  24,   2,  -4,   1,  -2,  32, -26, -21, -16,   8,  -5,   2,  -8,  31,  45, -35,   7,   3,   6,   5,   1,  67, -13,  23, -12,  -2,  -2,   1,   5,   7,  -2,   2,  -5,   5,   1,   2,  -4,   9,   6,   1,   5,  -4,   2,  -6, -16,  -2,  -3,   0,  -4,  -2,  -1,  -1,  20,  -1,   5,  -1,   1,   0,  -1 },
      {  -1,   0, -12, -13,  14, -18,  10,   0,   3,   1,   5,  36,   0,  14,  -8,   3,   0,   3,  10, -64,  21,  -7,   3,  -4,   5,   6,  72,   1,  -2,   0,   3,   0,  -2,  -1, -41,  14,  -5,   9,  -2,   2,   1, -24, -11,  -7,  -1,  -2,   3,  -1,  -6, -29,   1,  -2,  -1,   1,   0,   0,  -3,   5,  -6,  -3,   1,  -1,   1,   0 },
      {   3,   0,  -3,  36,  50,  13,  10,   1,   2,   0,   6,  48, -14,   6,  -5,   2,  -5,  -5, -20, -40, -19,  -7,  -2,  -1,   1,   9, -30,  -7,  10,   2,   4,   0,   3,   3,  57,  -7,   3,   1,   0,   0,   5,  26,   8,   2,  -2,  -1,   2,   0,  -2, -42,  -8,   1,  -3,   1,  -1,   2,  -5,   3,  -2,  -1,   0,  -2,   0,   1 },
      {  -1,  -6,  23,  24, -78,  21, -17,   5,   5,  -1,  17,  74,  22,   1,   8,  -2,  -1,  -5,  13,   7, -12,   3,   1,   1,   0,   0,  22, -19,   6,  -7,  -2,   0,   0,   5,  14,  -1,   5,  -3,   3,  -2,   0,   9,   6,  -2,   0,   1,  -2,   1,   1,  16,  -1,  -1,   3,   0,  -1,  -1,   2,   3,   5,   2,  -1,   0,  -1,   0 },
      {   0,  -1,   5, -17, -10,   8,  -9,   3,   0,   0,   2, -29,  17, -15,   4,  -4,   2,   2,  13, -17,   7,  -5,   1,  -1,   3,   4,  28,  -8,   6,   1,   0,   1,   6,  11,  52,   8,  -3,   0,   1,  -4,   1,   2, -17,   3,  -3,   1,   1,  -2,  10, -17,   5,  -2,   1,   0,   1,   0,  -3, -96,   3,  -1,  -1,  -2,   0,  -2 },
      {  -3,  -2,  -2, -11, -24, -53,  -3, -10,  -1,   0, -13, -15,   0,  17,   2,   3,   3,   5,  14,  12,  24,  12,   5,  -1,   0,  -4,  14,  -9, -10,  -5,  -3,   0,   4,   2,  66,  22,   4,   1,   0,  -3,  -2,   7, -39,   9,  -1,  -1,  -1,   1,  -6, -27,  -4,  -3,  -2,  -3,   1,   1,  -1,  55,  -3,  -1,   2,   1,   0,   0 },
      {   0,   0,   9, -32, -29,  64,  -7,  14,  -2,   1,   6, -40,  -1, -29,   3,  -5,  -4,  -2,  -7, -55,  -2, -18,  -2,  -4,   2,  -2,   7,  23,  12,   9,   1,   3,  -2,  -2,  17,  -6,  -1,  -1,  -1,  -3,   0,  12,   4,   5,  -2,   1,   1,   1,  -4,   0, -11,   4,  -3,   1,  -1,  -1,   0,  51,  -1,   2,  -2,   0,  -1,   2 },
      {   1,   2,   4,  -3, -19,  31,  -9,   5,  -3,  -7,   8,  17, -96, -13,  -8,   1,   3,   8, -10,  18,  39,   2,   5,   1,  -1,   1, -13, -12,  -2,   2,   2,   0,   0,  -2, -10,  24, -17,   0,  -3,   0,   0,  -9, -30,   9,   7,   0,  -1,   0,   0, -19,   3,  -5,   1,  -1,   0,  -1,  -2, -13,  -1,   0,   4,   1,  -2,   0 },
      {  -2,  -2,  -2,   1, -23, -35,   5,  -7,  -1,  -3,   7,   1, -49,  12,  -9,   4,   0,  -4,  -1,   0,   8,  -2,   2,   0,   3,  -3,  26,  69,  -6,   3,   1,   2,   1,  -1,  23, -29,  15,  -5,   0,   3,   3,  13,  62,   6,   1,  -4,   0,   1,   1,   9,  -7,   6,   1,   0,   1,  -1,   1, -15,  -3,  -1,   3,   0,   1,   0 },
      {  -2,  -1,   0,  -4,  -6,  -5, -42,  -5,   1,  -1,   2,   4, -13,  29,  18,   3,  -2,   0,   3,  -6, -66,  -1,   2,   5,   1,   0,  -9,  58,  10,   0,  -6,   0,   0,   1, -11, -12,  -1,  -2,   5,  -1,  -3, -11, -62,  -6,   1,   1,  -3,   3,   1, -14,  20,  -2,   6,  -5,   1,   0,  -2,  -6,   2,   6,   2,   0,  -2,   0 },
      {  -2,  -4,   1,  -2, -24, -34, -22, -13,   0,   0,   8, -31, -33,  25,  18,   4,  -3,  -2,  -3, -42, -40,  -1,   1,   2,  -1,   0, -15, -60,  29, -19,  -2,  -3,  -1,  -7, -11,  21,  -7,   3,   2,   3,   1,  -2,  38,   9,   8,   1,  -1,   0,  -1,   6,   0,   2,  -1,  -1,  -2,   0,   0,  -4,  -4,   6,  -2,   1,  -1,  -2 },
      {   0,  -3,   5,   2,  14,   9, -57,  -1,   1,   3,   6,  -7,   4,  36,   9,   2,   1,   0,  -3,   1,  43, -10,   8,  -6,   0,   3,   8, -29,  -4,   4,   2,  -2,  -2,   5,  -2, -85, -13,   2,   3,   3,   0,   4,  -4,  23,  -1,   7,   1,   2,   0,  -2,   7,  11,  -2,  -2,   6,   3,   0,   5,  -6,   3,  -1,  -2,  -2,   2 },
      {   0,   0,   0,   3, -30, -12,  73,   1,  -1,  -1,   1, -11, -11,   6, -33,   2,   0,  -3,   2,  -2, -13, -11, -15,   1,   1,   3,  -4, -18,  43,  -5,  19,   1,  -1,   5,  -8, -58,   5,  -3,  -1,  -1,  -2,  -2, -38,   5, -11,  -3,   5,   0,   1,  -3,  -8,  -4,   2,   4,  -4,  -3,   1,  -4,  10,  -6,   0,  -4,   3,   1 },
      {  -1,  -2,  -2,  11,   3, -46, -35,   0,   0,  -1,  -2,  12,   8, -74,   7,  -3,   0,   0,  -8,  -1,  13,  14,  -1,   5,   3,   5,  -5,  13,  63,  22,   4,   5,  -2,   3, -10,  -9, -33,   3,   1,  -2,   2,   3,  -4,   4,  -5,   1,   2,  -1,  -3,  -3, -23,  -1,   1,   7,   1,   0,   1,   3,  -2,  -1,  -2,   2,   1,   1 },
      {   0,   0,  -1,  -3,  -1, -24, -10,  12,  -2,  -3,  -2,   3, -19, -65, -10,  -5,  -1,   0,   0, -21, -31,  21,   0,   0,  -2,  -4,   4, -27, -67,  -2,  -7,   1,   0,   4,  -5, -36,  40,  -7,  -1,  -4,   0,   0, -11,   4,   5,   6,  -1,   1,   2,  -1,  18,   1,   1,   1,  -4,  -1,   1,   2,   8,  -1,   4,   4,   2,   0 },
      {   0,  -1,   3,   1,   1,  -5,  -1, -11,  -1,  -2,   1,  -1,  -9, -12,  -8,  -6,   0,  -1,   2,   1,  13, -22,   4,  -2,   0,   0,   0,  -8,  21,  11,  -2,   2,   2,   0,  18,  -8,  -2,  14,  -2,   1,  -2,  -5,   9, -78,   0,  -2,   5,  -1,   5,   9,  85,   1,   5,  -1,  -1,   0,   0,  12,  -6,  12,  -4,   0,   1,   2 },
      {   1,   1,   1,  -2,   2,  15,  23, -28,   1,   1,   2,   4,   4,  -3,  44,   0,   2,   3,   0, -22,  28,  75, -20,  20,   0,  -3,  -8,  15,  17, -40,  -6,  -7,   0,  -1,   0, -12,   3, -12,  13,  -3,   0,  -4,   4,  16,  -3,   6,  -7,   1,   2,  -1,  38, -12,  -3, -12,  -5,  -1,   1,   1,  13,   5,   2,   2,  -3,  -2 },
      {   0,   0,   0,   2,   7,  -8, -11,   2,   1,   2,   0,   4,  -6, -12,  42,   1,  -1,  -2,  -4,   6,  24, -57, -19,  -8,   0,   0,  -4,   2,  22, -34,   0,  -1,   2,   3, -10,   9,  78,  19,   9,   1,  -1,   2,  -8,   4, -22,  10, -11,  -2,   0,  -2,  -6,   0,  -1,   1,   2,  -3,   2,  -1,  23, -10,   3,   2,   1,  -1 },
      {  -1,  -1,  -4,  -1,   0,   7,  -4, -19,   0,   0,  -1,  -7, -15,  11,  27,   3,   1,  -1,   3,  -4,   4,  25, -13,   9,  -1,  -2,   4, -11, -14,  15,  -2,  -1,   0,   0,   1, -16,   1,   0,   7,  -2,  -2,   4,  -7, -83, -10,  -1,  -2,  -1,  -4,   1, -67,   3,  -1,   1,   0,   1,   3,  -3,  35,  10,  -3,  -3,  -1,  -1 },
      {   0,   0,   2,  -2,   7,   6, -42,  27,  -1,  -2,   2,  -7,   5,  11, -77,  -2,   1,   1,   5,  -4,   8,  41,  -1,   2,   0,   1,  -6,   6,  23, -52,  20,   0,   1,   2,  -2,  10,  32,   4,  -3,   2,   0,   4,   2, -22, -13,   4,   7,  -3,   0,   0, -10,   4,  -1,   3,   3,  -3,   1,   0,   0,   1,  -3,  -1,   2,  -1 },
      {  -1,  -2,  -1,   3,  -2,  10,  -6, -92,   0,   0,   0,   1,   9, -18, -32,  16,   0,   1,  -5,  -6,   1, -23,  58,  18,  -1,  -2,  -4,   4,  -6, -22,  -4,  -7,  -1,   0,  -2,  -2,   2,   1,  -4,   1,   0,   0,  -4,   7,   3,  10,   1,   0,  -1,   3, -11, -20,   7,  -2,  -1,   1,   1,   0,  18,   3,  -4,   3,   2,  -2 },
      {   0,   1,   2,  -2,   3,   3,   6,  22,   0,   0,   0,  -2,  -2,  15,  10, -10,   2,   3,   0,  -5,   8,  26,  73,  -5,   0,  -1,  -4,  -8,  31,  47, -30,   2,   0,   0,  -6,   1,  52, -40,  -6,   0,  -1,   0,  -1,  -1,  -7,  -8,  -3,   0,   1,   0,   1, -20,   8,   0,  -2,   1,   0,  -2,  -5,  -8,  -2,   0,   0,   0 },
      {   1,   1,  -1,   1,  -3,  -4,  22,  25,   0,   0,  -5,   1,  -1, -12,  25, -19,   0,   2,  -4,   0,   1,  -4,  51, -14,  -1,  -1,  -4,   7,  -1, -65, -30,  -9,   0,   1,   2, -17, -20,  22,   3,   1,   0,   2,  -9, -24,  27,   6,   4,   0,  -2,   0, -21,  29,   0,  -3,  -5,   2,  -3,  -4, -44,  -8,   0,   2,   2,  -8 },
      {   0,   1,   0,  -2,   5, -11,   1,  42,   0,   0,   3,   0,   2,   4,   7, -29,   0,   0,   0,   0,   0, -11,  30,   0,   0,  -1,  -1,   6,  -5, -21,  15,  -3,  -1,  -2,   5,  -4, -31,   8, -19,   1,   1,  -2,  11,   5,  16,  -9,   4,   4,  -1,  -2,   9, -38,   3,  -2,  -1,  -1,   5,   6,  90,   4,   0,  -2,  -1,   1 },
      {   0,   1,   2,  -1,  -6,   8,   2, -16,   0,   0,  -1,  -1,  -1,   4,   8, -33,   0,   1,   0,   3,  -8,  28,   9,  -8,   1,   2,   0,   0,   1,  27,  30,  -4,   1,   1,  -1,  -5,  22,  94, -30,   3,   0,   1,   1,   9,   5, -13,  -7,   6,  -1,   1,  -6, -25, -11, -19,   6,   1,  -1,  -3, -21,  -3,   5,  -2,   0,   3 },
      {   0,   0,  -1,  -3,   1,  -2,  -4,  15,   0,   0,   1,   0,   1,   2,  -7,  14,   0,   0,   0,   0,   4, -12, -23,  10,   0,   0,  -1,   0,  -1,  -5, -28,   5,   0,   0,   2, -11,   1,   5,  37,  -2,   2,   1,   2,  -9,  34,  12,   7,  -7,  -2,   3, -12, -99,  -2,   1,   0,   2,  -3,  -1, -31,  13,  -4,   3,   4,   2 },
      {   0,   1,   0,  -1,  -2,   0,   6,  17,   0,   0,   0,   0,  -5,   1, -12,  12,   0,   1,   0,   0,  -6,   5,  11,  15,  -1,  -2,  -1,  -2,  -9,   3, -54,   6,   0,   0,   2,   0, -23,  46,  29,  -2,  -1,  -4,   9,  11, -88,  12,   1,   2,   1,   0,   4,  -2,  25,  -6,  -2,   0,   1,   1,  10,   5,   4,  -1,  -1,   2 },
      {   0,   0,   2,   0,  -1,  -3,   3,  18,   0,   2,   1,  -1,  -1,  -7,  29,  48,   0,   0,   1,   1,  -3,  -3,  25,   1,   0,   0,  -2,   1, -14, -18,  64,  -6,  -1,  -1,   2,  -3, -11, -20, -35,  -5,  -1,   0,   3, -12, -49,   9, -11,   9,   0,   2,   0, -32,   5,   2,   5,  -6,  -3,  -1, -34,  17,   6,  10,   1,   9 },
      {   0,   0,   1,   1,  -1,   5,   3,  21,   1,   2,   2,  -3,   4,  -4,   0, 101,   0,  -1,   3,   2,  -1,  14,  12,  -1,   0,   1,   2,   0,  10,  12, -10,  -9,   1,  -1,   2,  -2,   6,  37,   0,   8,   1,   1,   0,  -1,  38,   0,  -5,  -3,   0,  -1,   9,  17, -20,   0,   0,  -5,   1,   0,  26, -22,   4,   4,  -1,   3 },
      {   0,   0,   0,   0,  -3,  -4,  10,   1,   0,   0,   1,   1,   1,   1,  -7,  -6,   0,   0,  -3,   0,  -1,   6,  17, -49,   0,   0,  -3,   3,   2,  14,  24,  18,   0,   2,  -3,   7,   0,   1,  43,  10,   1,  -1,   0,   2,   6,  79, -20,   8,  -1,  -3,   0,  11, -22, -27,  -8,  -1,   0,  -3,  11,  42,  -5,   4,   2,  14 },
      {   0,   0,   0,  -1,   1,   5,  -5, -22,   0,   0,   0,   0,  -2,  -6,   1,  11,  -1,  -1,   1,   3,  -7,  13,  -6,-105,   0,   1,   0,   1,  -1, -13, -10,   5,   0,  -1,   1,  -3, -10,  -8,   7,  20,  -1,   0,   0,  -1, -20, -40,  13, -10,   1,   0,   2, -23,   1,  10,   5,   2,  -1,   1,   7, -19,   1,   2,  -1,  -9 },
      {   0,   0,  -1,  -1,   2,   1,   0,   4,   0,  -1,  -2,   0,   0,  -3,   4,   0,   1,   1,   0,  -2,   1,   0,  26,  14,   0,   0,  -1,  -2,  -6,   2,  50,   7,   2,   2,   1,   2,   2,   9,  88,  -6,  -1,  -1,  -1,   4,   7, -59, -20,  -7,   1,   3,   3,  10,  22,  13,  -2,  -5,   0,   1,  -2,   4,  -4,   3,   6,  -1 },
      {   0,   0,   0,  -3,  -2,   0,  -3,   1,   0,  -1,   0,  -1,  -2,   1,   2, -11,   0,   0,   1,  -1,  -4,   3,  -1,   5,   0,  -1,  -1,   0,  -4,  11,  26, -46,   1,   1,   4,   2,  -9,  -5,  20,  37,   0,   0,   4, -10,  -3,  48,  -3, -19,   1,   1,   4, -11,  29,  -5,   5,   0,   0,   4,  -3, -81,   7, -19,  -2, -23 },
      {   0,   0,   0,   2,  -6,  -2,  -1,   3,   0,   0,  -1,   0,   1,  -4,  -3, -10,   0,   1,  -1,   2,  -3,  -8,   9,  13,   0,   0,  -1,   1,  -3,   4,  -4, -58,   0,   0,  -2,   1,  -1,  -8,  19,  26,   0,   1,  -2,   0, -24, -18,  11,  -8,  -2,  -2,   5,  -3, -94,   6,   1,  -1,   0,  -2,  -1,  19,  23, -11,  -3, -16 },
      {   0,   0,  -1,   0,   0,  -1,   5,   5,   0,   0,   0,   1,   1,  -6,   6,   9,   0,  -1,   0,   0,   4,   2,  -6,  -3,  -1,   0,   0,   2,   1,   6,   1, -52,   1,   1,   1,   3,  14,   5,  -4,  54,   0,   0,  -4,  10,  11,  -5,  48,  -5,   1,   0,  -5,  11,  55,  -5,   1,  -2,   0,   0,   0,  62, -17,  -7, -10,   7 },
      {   0,   0,   0,   1,  -1,   2,   0,  -3,   0,   1,  -1,   0,  -3,   4,  18,   3,   0,   0,   0,   5,  -4,   5,  13,   9,   1,   1,   1,   1,  -4,   2,  23,  40,   0,   0,  -2,   3,   6,   8,  13, -19,   1,   3,   1,   3,  -6,  18, 103, -22,   0,   0,   4,   2, -17,  17, -17,  10,   0,   1,   1, -16,  -1,  -4,  -8,   0 },
      {   0,   1,   1,  -1,  -1,   0,  -3,   2,   0,   0,  -1,   0,   3,  -4,   1,  -1,   0,   0,  -1,   0,   1,  -6,   0,  23,   0,  -1,  -1,  -3,  -2,  -7,  -1,  56,   0,   0,   2,  -2,  -3, -11,  -4,  82,   0,  -1,  -2,  -6,  -5, -21,   3,  -4,  -1,  -2,  -4,   0, -11, -57, -13, -15,  -1,  -2,  -4, -16,  14,  17,   5,  25 },
      {   0,   0,   1,  -1,   1,  -1,  -4,  -1,   0,  -1,  -1,   0,   2,  -1,  -7,   2,   0,   0,   0,  -1,   3,  -5,  -5, -10,   0,  -1,  -1,  -2,   4,   3,   1, -34,   0,   0,  -2,   2,   1,  -9,  18, -33,   1,   1,   3,  -6,   4, -10,  34,  78,   0,  -1,  -4,   0,  10, -50, -14,  -2,   1,  -1,   0, -19,  42,  26,   1,  17 },
      {   0,   0,  -1,  -1,   0,   5,   2,  -6,   0,   0,   0,  -2,   3,   5,   2,  -9,   0,   0,   1,  -1,   0,   7,   1,   1,   0,   0,   1,  -1,   3,   0,  -4,  15,   0,   0,   2,   1,   0,   7,   1,  28,   0,   1,   0,  -2,  10,  16,  -4,  17,   1,  -2,   1,   1,  14,  75,   9,  -8,   2,   0,   2,   9,  87,  16,   3,  20 },
      {   0,   0,   0,   0,   1,   0,  -1,   2,   0,   0,   0,   0,   4,  -1,  -4,   8,   0,   0,  -1,   0,   2,  -4,   0,  -5,   0,   1,  -1,  -2,   4,   3,   2,   7,   0,   0,  -2,  -1,   1,  -4,  -7, -28,   0,   0,   2,   0,   6,  -5,  -3, -61,   0,   0,  -3,   3,  20, -49,  -4,   6,   2,   2,   1,  26,  73, -16,  -1, -49 },
      {   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   1,  -5,   2,  11,   0,   0,  -1,   0,   2,   0,   2,   3,   0,  -1,  -1,  -1,  -1,  -3,   2,  33,   0,   1,  -1,   1,   4,   0,   1,  19,   0,   0,   0,   0,   3,  -2,   0,  66,   0,   0,   1,  -4,   0,   4,  -1,   2,   0,   1,  -3,   8,   1, -83, -15, -56 },
      {   0,   0,   0,  -1,   1,   0,  -8,  -1,   0,   1,   0,  -1,   1,   3,  -1,   1,   0,   0,  -1,   2,   3,  -1,  -4,  -6,   0,   0,   0,   2,  -1,   2,   1, -12,   0,   0,   1,  -1,  -1,   4, -10,   1,   0,   0,   0,   2,  -2,  -3, -15,  -7,  -2,  -1,  -1,  -3,   3,  16,-121,  -5,   0,   0,  -3,   0,   1, -18,  15,  10 },
      {   0,   0,   0,  -1,  -2,   0,   1,   2,   0,   0,   0,   0,   0,  -3,  -3,   4,   0,   0,   1,   0,   0,  -1,   0,   3,   0,   0,   0,   1,   2,  -5,   4,   0,   0,   1,  -1,   1,   2,  -2,   2,  10,   0,   0,  -1,  -1,  -1,  -6, -12,   1,   0,   1,   0,   0,   1,  -3,  -4, 120,   0,  -1,   0,  -2,   9,  -8, -27,  28 },
      {   0,   0,   0,  -1,   0,   1,   1,   3,   0,   0,   0,   0,   1,   3,  -3,  -1,   0,   0,   0,   0,   2,  -1,  -3,   6,   0,   0,   0,   1,   0,   7,   5,   8,   0,   1,   1,  -2,   1,   5,  -3,  18,   0,   0,  -1,   1,  -2,   5,  -3,  18,   0,  -1,   0,  -1,   0,  12, -22,  18,   0,   0,   1,   5, -14,  83, -22, -83 },
      {   0,  -1,   1,   0,  -1,   3,  -1,   1,   0,   0,   0,   0,  -1,   1,   5,   3,   0,   0,   0,   0,   0,   3,  -2,   2,   0,   0,   0,   0,   0,   2,   1,   0,   0,   0,   0,   2,   0,   0,  -2,   9,   0,  -1,   0,   3,  -1,   2,  11,   9,   0,   0,   2,   3,   1,   1,   9,  33,   1,   0,   3,   4,  -4,   2, 120, -20 },
    },
    {
      { 114, -19,  -6,  -6,  -5,  -3,  -1,  -1,  18, -16,   3,   0,   1,   0,   0,   0, -43,  -4,   5,   0,   0,   1,   1,   0,   3,  -2,   0,   0,  -1,   1,   0,   0, -14,   2,   4,   0,   0,   1,   1,   0,  -8,   5,   2,   0,   1,   1,   1,   0,  -9,   2,   3,   1,   1,   0,   0,   0,  -3,   1,   2,   0,   0,   1,   0,   0 },
      {  19,   1,  -2,  -5,  -2,   4,   0,   0, -44,   0,  10,   1,   4,   2,   0,   0,  30,  -3, -15,   5,  -2,  -2,  -1,   0, 104,   8, -26,   6, -11,  -2,  -3,  -1,  18,  13,  -3,   3,  -1,  -2,  -1,   0,  -9,   5,  -6,   1,   3,  -1,   0,   0,  15,   2,  -7,  -3,   0,   0,   0,   0,  -5,   2,   3,  -1,   0,   0,   0,   0 },
      { -15,  -2, -11,   3,  -2,   1,   0,   1, -76,   9,  12,  -1,   4,   0,   2,   0, -85,  11,  29,  -1,   6,  -1,   2,   0,   0, -11,  10,   1,   1,   1,   0,   0,  31,  -7,  -4,  -5,  -2,   0,   0,   0,  22,  -5,   1,  -2,  -2,   0,   0,   0,   1,  -4,   1,   1,   1,   1,   0,   0,   3,   0,  -1,   0,  -1,   0,   0,   0 },
      { -12, -12,  -1,  -9,   2,   9,   1,   0,  68,  -5, -11,  -2,  -3,   2,  -3,  -1, -35,  44, -13,   8,   0,  -6,  -2,  -1,  19,  16,  -6,   5,  -4,  -5,   0,  -2,  70,  -1,  -3,   3,  -3,  -4,  -2,  -1,   1,  -9,  -5,   7,   2,   1,  -1,   0,  40,  -1, -15,  -1,  -2,   1,  -2,   0,   4,   3,   1,   0,  -1,   0,  -1,   0 },
      {  28, -49,  19, -14,   7,   1,   1,   0, -42,  28,  -8,   6,  -5,   2,  -3,   0,  40,  50, -27,   6,  -6,  -2,  -4,  -1, -34,  12,   6,  -2,   6,  -6,   3,  -2,   1, -20,  10,   1,   3,  -4,   0,  -1,  38, -13,  -7,   8,  -3,   3,  -2,   0,  32,  -5, -10,   0,  -2,   2,  -2,   0,  10,  -2,  -2,  -1,  -2,   0,  -1,   0 },
      {  34,  73, -39,  -6, -13,   3,  -6,  -2, -14,  22,  -9,   1,  -3,   0,  -4,  -3,  27, -25,  -8,  -5,   4,  -4,  -1,  -2, -32,  13,   3,   5,   2,  -6,   2,  -3,  60, -10,  -7,  15,   0,  -6,  -2,  -1,   3,  -4,  -4,  11,  -2,   4,  -2,   0,   1, -12,   4,  -1,   1,   3,  -1,   0,  14,  -5,   0,   0,  -2,   1,   0,   0 },
      {   3, -62,   8,   5,   2,   3,   1,   0,  -3, -23,   7,  -9,   7,  -1,   5,   2,  35, -25,  22,  -1,  10,  -1,   4,   2, -13, -16,  16,   8,  -5,   7,  -1,   5,  67,  20, -28,  -5, -10,   5,   0,   1,   6,  10,  -4,  -7,   1,  -4,   1,   0, -46,  11,   7,   0,   2,  -4,   1,  -1,  -3,   1,  -4,   2,  -2,  -2,   0,   1 },
      {  -6,  11,  49, -48,  -1,   3,  -1,  -4, -27, -41,  -2,   1,   3,   8,   1,   0, -17, -27, -48,  39,   0,   2,   1,   3, -28,   9,   1,   6, -10,   1,   0,   2,   9,  29,  10,  -6,  -6,   0,   0,   1, -38,  -1,   7,   5,   7,   0,   2,   0,  12,   8,  -9,   2,   0,  -3,   0,  -1,  13,   6,   3,   1,  -1,  -2,  -1,  -1 },
      {   0,  -8,   8,   0,  10,  -7,   2,   1, -10,  39, -19,   7,  -7,   0,  -3,   0,  -2,  38, -16,  -8,  -5,   2,   0,  -2,   7, -22,  16,  -9,  12,  -3,   1,  -2,  17, -23,  14,   6,   3,   1,  -4,   0, -86,   3,  14,   3,  -1,   3,   0,   1, -46, -14,  10,   1,   2,   2,   1,   1,   1,  -6,  -2,  -1,   1,   1,   0,   0 },
      {  -2,  27,  35, -20,   7, -15,   3,  -1,   9, -41,  27,   7,   0,  -3,  -4,  -2,  -1,  11, -23,  -3, -10,   3,  -3,  -1,  18,  -6,   1, -17,  16,  -5,   5,  -4,  16, -52,  22,   2,  11,  -5,   0,  -2,  48,  12,   5,   7,  -8,   7,  -5,  -1, -48, -12,  18,   4,   5,   7,   0,   0, -15,  -3,   3,  -2,   1,   3,   1,   0 },
      {  -8, -13, -34,   8,  -3,   9,   2,   0, -17, -54,  39,  -7,   9,  -2,   2,  -4,   6,  32,  10, -22,   0,  -8,  -3,  -2, -19,  66, -24,   7,  -6,  -1,   0,  -6,  -8, -15,  -2,  19,   2,  -4,  -2,  -1, -35,  10,  -2,   9,  -1,   3,  -2,   0,  -9,  -7,   2,  -2,   1,   3,   0,   0,  18,  -9,   0,   1,   0,   1,  -2,  -1 },
      {  -1,  -2,  18,  10,  -1,  -8,  -3,  -1,   5,  49,  -1,  -1,  -4,  -7,   0,   3, -21,   0, -28,  10,  -2,  11,  -1,   1,  -6,  61, -46,  -2,  -4,   4,  -3,   1,   0,  37, -23,  -6,   1,  -3,   0,   0,  20, -12,  -5,  -7,  -3,  -1,  -1,   2, -54,  -4,   7,  -7,   2,   1,   2,   0, -19, -11,   3,  -2,   4,  -1,  -1,   0 },
      {  -7,   1, -47, -25, -12,  14,   0,   0,   0, -20, -48,  12,  -5,  19,  -2,   0,   3,  26,   1,  37,   2,  -3,  -2,   1,  22,  15,  40,   2,  -4,  -8,   0,   0, -19,  21,   5,  -9,  -6,  -2,   3,   1,  25, -14,  10,   5,   6,   1,   1,   0, -46,  -2,  -1,   4,   0,   1,   0,   1,  35,   1,  -5,   2,  -3,   0,   2,   0 },
      {   0,   7, -27,   4,  -8,   6,  -2,   2, -13, -21, -18,   7,  -3,   9,  -1,   3,   7,  19,  -4,  11,  -1,  -2,   0,   0, -15,   6,   9,   2,  -6,  -2,  -2,   0,   6,   7,  21,  -9,  -5,   1,   0,   1,  -6,  -1,  -1,  -2,   1,  -1,   0,   1,   2,  17,  -9,   4,  -1,  -3,   0,  -1,-110,   4,  12,  -1,   4,  -2,   3,   0 },
      {   4,  25, -13,   9, -14,   4,  -9,   1,  -3,  -6,  13, -16,  -1,  -1,   8,   2,   4,  56, -30,   6,  -3,   9,   3,   1, -16, -54, -35,  33, -13,  16,  -5,   5,  -4,  16, -21,  -2, -13,   6,   1,   3,  11,  43,   5, -21,   4,  -5,   1,   2, -12,  23,  -1,  -2,   1,  -8,   3,  -1,  15,  11,   3,   3,  -1,  -3,  -2,   0 },
      {  -7, -23,  -7,  51, -15,  11, -13,  -1, -10, -19, -57,  23,  -2,   3,  -5,   3, -18, -31, -52, -35,   2,   0,  11,   2,   2,   5, -12,   0,   7,   4,   0,   1,   5, -20,  18,  25,  -1,   4,  -5,   1,  11,  18,  12,   1,  -3,   2,  -2,   0,   1,  21,   5,  -7,   1,  -3,   2,   0,  13,   8,   2,   1,   3,   0,   0,   1 },
      {   8,  18,  17,  37,   2, -26,   6,   5,  -6, -12,  -9,  31, -12,   1,   3,   5,   8,  19,  22,  -2,  -7,  13,  -7,   0,  -4,  21,   9, -46,   7,  -3,   0,   3,  17,  36,  22, -48,   8,  -1,   8,  -1,  -3,  51, -12, -15,  -4,   1,   2,  -1,  12,  -4,  -7,   9,  -4,   4,  -2,   0,  19,   9,  -8,  -1,   1,  -1,   1,  -1 },
      {  -9, -28, -62, -16,  53, -17,  18,  -2,   3,   4,   9,  15,  -7,   0,   0,  -1,  -6, -29, -21,  31, -41,  10, -17,  -1, -10, -20, -32, -19,  -9,   2,  -4,  -5,   5,  -5,   9,  -8,   4,  -8,   0,   1,   1,  23,   2,   3,  -2,  -3,  -6,   2,   6, -15,   6,   0,   1,   1,   2,   2,   1,  -7,  -1,  -2,   1,   1,   2,   0 },
      {  -5, -10,  -9,   3, -33,  21,   3,   5,   9,  34,  30,  -4,   6,   7,   3,  -2,  -5, -18,  -8,  27,  24, -21,  -4,  -1,   0,   8,   0,  -6, -21,  -5,  -3,  -3,   0, -41,  37, -28,   1,   2,   6,   1,  -5,  25, -36,   7,  11,   0,   2,  -2, -18,  46, -30,  12,  -4,   0,  -6,  -2,  11,  -4,  -1,   2,  -2,   0,   2,   0 },
      {  -5,   3,  13, -43,  25,  -1,  13,  -3,  -1,  16, -26,  -5,  -7,  10,  -2,  -6, -10,   0,  12, -28,   2, -16,  -5,  -2,   3,  14,  19,  10, -10, -12,  -1,  -8, -11,  10, -18,  44, -11,  -1,  -1,  -2,   8,  76, -32,  17,  -1,  -1,  -1,  -4,  -2,   9,  -4,   2,   3,   0,  -6,   2, -12,  -9,  -1,   2,  -1,   1,   0,  -2 },
      {   5,  11,   9, -24,  36,  -3,   3,  -4,  -3,   9, -26, -19,  13,  -9,   5,   3,   8,   2,  42, -13, -21,  -2,   9,   2,  -2,  20, -30,  10,  -2,  24,  -2,   6,  10, -11,  42,  -5,  -2,  10, -10,   0,   5,  -7,  46,  -7,  -3, -10,   0,   2, -11,  47, -23,  -1,  -2,  -8,   2,   0,  11,  24,  -3,   0,   5,  -3,  -2,   2 },
      {  -5, -11, -10, -10, -15,  11,   0,   2,   5,  10,  22, -15,   2,  -6,  -6,  -7,   1,  -2,  -9, -34,  20, -24,   3,  -5,  -3, -20, -10,   0,   7,  -4,   1,  -2,   4,  58,  72,   9, -10,  -1,  -7,  -1,  14,   5,  18,  16,  -3,   5,  -3,   0,  -8, -42, -13,   4,   6,   5,   1,  -1,   1, -11,   5,   0,  -2,   3,   1,  -1 },
      {   6,  14,  -5,  40,  50,  -6,   9,  -3,  -1,  -2,  17, -46,  21, -11,  10,   2,  -6,   3, -39, -12, -23,   5,  -4,   7,   6,  10,  60,  12, -10,   8, -11,  -1,   0,  15,  15,   3,   0,  -1,   0,   0,   4, -17, -29,  -3,   1, -10,   3,   2,  -7,  11, -15,  -4,   0,  -3,   0,   2,   5,   2,  -5,  -1,  -1,   1,  -2,   0 },
      {   5,  10,   4,  12,  21,  47,   2,  -5,   2,   7,  33,  44,  13,   9, -13,  -6,  -1,  11,  -4,  -6, -15, -27,   3,   5,  -4, -13,   5, -49,  -5,  -8,  13,   0,  -2,  25, -25,  21,   3,   2,  -7,  -6,   2, -12,  22,  35,  10,  -3,  -8,  -4,  -7,  37,  -2,  -5,  -5,   2,  -2,  -3,  -2,  22,  -7,   0,  -4,   0,   3,   3 },
      {   8,  15,  24,  25,  30,  14,  25,   5,  -4, -19, -30,  14,   2,  20,   6,   0,   1,   4,  12,  -6,  12, -31, -18,  -7,  -7, -19, -34,   0, -52,  -8, -18,   1,  -1, -14,  -1,  -5, -22,   6,   3,   6,   2, -26, -22,  -2,  20,  -3,   5,   0, -15, -36, -24,   9,   0,   0,  -3,   1,   3, -19,   1,  -1,   0,  -3,   1,  -3 },
      {  -1,  -3, -10, -14, -13, -24,  25,  -5,   1,   9,  14,  -9, -24,  -8,  10,  -3,   0,  -1, -22, -25,  20,  -6, -19,  -3,   4,  15,  21, -35, -38,  12, -10,  -7,  -1, -11, -12,  -6, -58,  19,   9,   6,   4,  -5,  40, -22,  10,   7,  -5,   2,  11,   5,  22,  14,  13,  -4,  -8,  -3,  -3,  21,  -7,   8,  -2,  -4,   0,  -3 },
      {   4,   1,  11,  18,  13,  50, -13,   7,   0,   8,  -9, -43,  25,  15,   4,   0,   0,  -5,  -3,  26,   0, -27,  -1,   2,   1,  15,   6,  -3,   4, -13,   1,   1,  -1, -13,  -8, -24,   5,  -7,   0,  -4,   4,  39,  51, -14,   6, -11,   0,   0,  14, -29,  33,   1,   3,   3,  -4,  -1, -10, -15, -18,   5,  -1,   3,  -2,  -2 },
      {   0,   0,   6,   1,   1,   0, -23,   3,   8,  16,  17,  63,  38, -14,  -8,  -4,   2,   2,   8,  13, -16,  10,  26,  -4,   1,  14,  26,  34, -45,  30,  -1,   5,   3,  -4,  22,  12, -14,  17,  -8,   2,   4,  14,   5, -17,  13,   0,   5,  -1,   6, -22,  22,  -9,   3,  -5,   7,   1,  -1, -10,  11,  -3,  -4,   1,   2,   4 },
      {  -3,  -2, -13, -24,  -7, -24,  -3,   6,  -1,  -3, -28, -20,  82, -18,  -1,  -6,   0,  11, -10, -21,  23,   9,  11,  -6,  -3, -11, -13, -42, -23,   0,   8,   8,  -2,   3,  -7, -17,  22,  -6,   1,   0,   0,  -4, -17,  14,  17,   3,   2,  -2,   3,   5,  25,  -2, -10,   0,  -2,  -2,  -1,  -2,   5,  -2,  -7,  -1,  -1,   3 },
      {  -3,  -8,  12,  13, -35,  -6,  -7,  -4,  -2,   6,  -7, -26,  -4,   1, -17,  -7,  -1,  -5,  14,   0, -61,   1, -24,   7,  -3,  -1,  -1,  14, -41, -23,  13,  -3,   1,   0,   6,   4,  10, -14,  16,  -7,  -1,   2,  -2,  25,  17,  17,  -5,  -2,  -6, -16,  11,  12,   3,  12,  -5,   5,   0,  58,  15,  -6,  -6,   4,   1,   2 },
      {   1,   1,  10,  28, -25, -29,  15, -13,  -1,  -4,  -7,  -6,  19, -26,   4,  -8,   2,   3,  11,  37,  -1,  19, -20,   3,   1,  -7,   5, -10, -15,   2,   9,  -8,   1,   1,  -1,  40, -17, -26,   4,  -4,   2,   8,  37,  26,  -9,  -2, -10,  -3,   2,  12, -34,  -6,   2,  18,   0,  -2,  -3, -50, -19,  -1,   6,   2,  -3,  -2 },
      {   0,  -1,  -1,  13,  15, -41,  31,  -4,   3,   8,  17,  15,  11,  55,  -7,   4,   0,   0,  -3,  -6,  32,   9,   6,   2,  -2,   1,   0,  44, -11, -51,  -8,  -3,   2,   8,   9,  -4,  22, -19, -10,   2,   3,  -4,  30,   4,   6,   8,   6,  -4,   4,  24,  24,   3,  -7,   9,  -4,  -4,   2,   4,  -6,  -6,   4,  -3,   2,  -2 },
      {   0,   1,   5,  -8, -12,  10,  -8,   4,  -1,  -4,   0, -11, -26,  18, -13,   1,  -1,   3,   4, -29, -43,   6,  -6,  -2,   1,  -5,   4,  -6, -20,  -3,   4,  -3,   5,  10,   5, -20,  10,   9,   5,   4,   1, -12,   5,   2,   8,  -1,   2,  -2,   7,  36,  29,   6,  -5, -11,  -2,  -2,   9, -83,  31,   4,  -2,   0,   1,   0 },
      {  -1,  -3,  -3, -23, -30,  19,  39,  -3,   2,   3,   1,  24,  10,  -9,  26,  -4,  -3,  -7, -15, -38, -26,   0,  -7,   2,  -1,   0,  13,  12,  -2,  14, -17,  -7,   1,  -2, -20, -14,  48, -27,   4,  -8,   2,   2,  20, -33,  17, -15,   9,  -3,  -5, -13, -39,   6, -12,  14,  -2,  -1,  -7,  -3, -17,   5,   4,  -1,  -4,  -3 },
      {   1,   0,   5,   2, -23,  16,  37,  10,   2,   6,   4, -16,  10,  58,  19,  -2,   2,   2,  10,  15,  -3,  36, -20,   0,  -2,   2,  -1, -30,   3,  24, -12,  16,   6,   2,  19,  45,  15,  23, -13,   3,   2,   0, -11, -15,   3, -12,  -4,   9,   3,  -9,  14, -32,   3, -15,   7,   6,   2,  10,  13,   5,   9,  -1,   4,  -2 },
      {   1,   2,   1,   2,   2,  16,  40, -40,  -1,   0,  -7,  14,  30, -30, -18,  -5,   0,   1,   2,   5,   4, -19, -50,  20,   0,   0,  -4,  18,  35, -14,  16,   5,   1,   5,   7,  -2,  -9,  16,  27,   3,  -2,  -1,  -7, -34,   1,  25,  -2,   3,   2,  14,  17, -16,  12,   5,  -1,   0,   4,  -4,  33,  -3, -11,   4,   3,  -5 },
      {   1,   1,   3,  13, -14, -20,  38,  16,  -1,  -1,  -7,   2,  -5, -26,  31,  -3,   0,   3,  -3,  20,  -7, -51,   1, -14,   1,   2,   3,  16,   9,  23, -24,  -6,   1,   6,  -9,  -6,  30,  42, -21,   0,   2,   4,   7,  48, -14,   6,   9,   3,   5,   2,  26,  13,  -6,  -8,   2,   2,   1,   1,  15,   0,  11,   1,  -4,   9 },
      {  -1,   0,  -2,   0,  -2, -28,  -3,  14,  -3,  -7,  -1,  -7, -30, -18, -39,  16,  -1,   4,   1,  14,  10, -42,  -6,   7,  -1,  -1,  -2,  -4, -11,  11,  16,   7,   3,   5,  12,  25,  38,   5,  -2, -13,   1,  -6,  -8, -31,  46, -23,   1, -11,   1,   7,  14, -31,   4,   0,  -3,  -5,  -2,  -3, -42,  11,  -7,   9,   2,   2 },
      {   1,   2,   3,   2,  10,  21,  13,  -9,   0,   0,  -2,   7, -10,   1, -28,   9,   0,   2,  -1, -12,  33,  36, -41,   8,   2,   2,   0,  24,  -1,  56,   9,  -3,  -1,  -1,   4, -23,   5, -14,  17,  -7,   0,   3,   1,  51,   7, -32,  -6,  -7,   1,  -5,  20,  13,  -4, -17,  -9,   3,  -4,   6, -16,  -7,  -9,  11,  -1,   7 },
      {  -1,   0,  -6,  -2, -13, -21,   8, -20,   1,   0,   4,  -8,  16,  29, -53,   7,  -1,  -3, -10, -11, -10, -12,   3,  17,   1,   3,   7,   2,  -9,   4,  28,  15,  -4,  -9, -24, -28,   4,  47, -23,   5,  -1,   9,  11,   8, -31,  -1,  -9,  -2,  -3, -21, -46, -18,  10, -22,   7,  -1,  -2,  -5,  -6,  -5,   7,  -1,   0,   1 },
      {   0,   0,   2,  -9,  -6,   2, -22,  62,   1,   0,   1,  13,  25,  -1, -22,   8,  -3,  -3,  -1,  -9, -11,  -5, -63, -30,  -1,  -2,   1,   8,   0, -11, -25,  -2,   0,   6,   0,   1,  -6,  -4,  -8,   9,  -2,  -8,  -3, -11, -49,   3,  12,   6,   0,  15,   4,  -1,  16,  -1,   7,  -1,   0,   4, -32,   5,   5,   4,  -1,   1 },
      {   1,   3,   5,  -4,  -9,  25,  39,  -4,  -1,  -3,  -6,  -3, -15, -23,  -5,  10,   0,   2,   5,  -3,  -4,   2,  27,  -8,  -1,  -5,   1,   1, -22,  -4,   3, -15,  -3,  -2,   7, -28, -17, -41, -38,  -2,  -3,   0, -17,   5, -26,   6, -10,   0,   2,   6,  23, -65,  22,   3,  -4,  -4,   2,   6,  -9, -17,  18,  -5,  -8,  -5 },
      {   0,  -1,   0,   1,  16,   2, -26,  -4,  -1,  -1,   3,  -6, -21, -19, -19,  -5,   1,  -1,   1,   3,  35,   2, -29, -19,   0,   2,  17, -11, -22,   0, -31,   6,  -1,  -4, -11,   0,  34, -15, -19,  17,   3,  12,  25,  -3,   6,   1,  -9,   5,  -2,  -2, -21, -26, -15,   4,  25,   0,   5,   9,  65, -16,   4,  -6,  -1,  -5 },
      {   0,   0,   1,   7, -10, -13,  20,  16,  -1,   0,   1,  -5,  11,   8, -46, -38,   1,  -1,   1,   3, -19,  -3,  14, -25,   0,   4,  -2,  -1,  45,   7, -25,  -3,  -2,  -5,  -9,  -8, -37, -19, -22,  10,   1,   3,  -2,   6,  49, -30,  27,   0,   0,   3,  -7,   7,  18,   4,   7,  -1,   0,   2,  17, -13, -10,  -3,   2,  -8 },
      {  -2,  -1,   1,  -5, -10,  -6, -35, -62,   1,  -1,   0,   8,  -2,  -4,  24, -39,  -1,   3,   4,  -7,  -8,  -4, -20,  15,   0,  -3,  -4,   0,   0, -24, -50,  21,   0,  -1,   8, -10,   3,  10, -12,  14,   0,  -3,  -9,  17,   3, -28,   0,  -5,   2,   2,  15, -19,  16, -18,   9,  -3,  -4,  -3, -38,  -2,   1,  -3,   5, -11 },
      {  -1,  -1,  -1,  -2,  -8, -15, -10, -30,   1,   2,   1,  -2,   9,  30, -19,  21,   1,   3,   1,   5,  -4, -40,  14,  24,  -1,  -4,  -1, -16,  -6,  36, -34,  -9,   1,   2,   1,  11,  -4, -33,  30,  -7,  -3,  -5,  -5, -11, -47, -23,  30,  -7,   1,   2,  13,   6,  -6,  26,  -7,   3,   4,   4,  21, -10,  10,  17,  -5,  -3 },
      {  -1,   0,  -2,  -6,   4,   7,  -3,  19,  -1,   0,   2,  -5,  -3,   0, -10,  21,   1,  -1,   0,  -5,  -2,  18,  10,  26,  -1,  -1,  -1,  -1,   9,   4, -51,   2,   1,  -3,  -3,  -8,  -8,  28,  36, -11,   2,   7,  10,  27,  27,  56,  29, -11,   0,  -2, -12, -45,  -5,  22,  -3,   8,  -4,  -7, -18, -14,   1,   6,   9,   3 },
      {   0,   0,   0,   0,  -6,   7,  16,   7,   0,  -1,   2,   1,   3, -29, -11,  21,   1,   0,   1,   5,   1,  20,  17,  -3,  -2,   0,  -3,  -6,  -7, -52, -12, -40,   2,  -2,   3,  10,  -1,  23,  34,  -5,   2,   1,   9,   2, -13, -56,  26,  -9,  -3,  -8,  -5, -11, -15, -48,   4,   9,   2,   8,  13,   6,  -9,  11,   1,   2 },
      {   1,   1,   1,   0,  -3,  20,  25,  -7,  -1,  -1,  -1,  -8,  -8, -24, -37,   5,   0,   1,   2,   6,  -1,  17,   9,  32,   0,   1,  -3,  -6,  -9,  -8, -21,  32,  -1,  -2,   5,  12,  -5,  -5, -38,  17,   0,   2,  -8,  -7, -17,  16,  19,  -6,   0,  -3,  14,  37, -52,   7,  29, -13,  -1,   0, -17,  31, -13, -14,   4,   8 },
      {   0,   0,  -2,   1,   6,   2,  15,  18,  -1,   0,   2,  -1, -17,  -4,   3, -55,  -1,   0,  -5,   0,   2,  -6,  13, -18,   1,   2,   2,   3, -12,   1,  30,  62,   1,  -2,   2,   3,   6, -10,  52,  30,   0,   1,   8,   5, -26,  -4,  39,  -2,  -1,   0,   1, -21, -11,  -7,  -8,  -5,  -1,   1,  -5, -14,   2, -10,  -1,   4 },
      {   0,   0,   1,   7,   9,  12,   5, -22,  -1,  -1,  -3,  -4,  -4,  -2, -25, -20,   1,   2,   3,   9,  11,  16,  20, -25,   0,  -2,  -2,  -6, -12,  10, -20, -17,   1,   4,  -1,   7,  37,  -3,  11,  -5,   0,  -2,   2,  -4, -13,  34,  18,  -4,   2,   2,  -7,  19,  69, -17, -18,   2,  -1,   1,  -1,  53,  -9, -12,  -2,   6 },
      {   0,   1,   0,   0,  -5,  -9,   9,  18,   1,   2,   3,   6,   5,  -6,   6,  38,   0,   0,   0,  -1,  -8, -12,  10,   5,   0,   3,   3,   3,   9,  -9, -30,  68,  -2,  -4,  -3,  -1,  -9, -30,  10, -12,  -2,   1,   6,   9,  -2,  -8, -49, -21,   0,  -2,  -6,   0,  22, -29, -21,   6,   2,  -4,  21,  17, -32,  -1,  18,   5 },
      {   0,   0,   4,   4,  -4, -10,  -1, -22,   0,  -1,   0,  -2,   3,  17,  -8, -14,  -1,  -2,   1,   0,  -9, -15,   3, -46,   0,   0,   0,   5,   8,  14, -11, -23,   0,   1,   1,  -7, -10, -12,  12,   4,  -1,   0,   1,  -3,  -5,  20, -26,  31,  -3,  -1,  -1, -32, -60, -17,  -8,  21,  -4,   0, -16,  30, -43,   6,  11,  22 },
      {   0,  -1,  -1,   0,  -5,  11,   5, -31,   0,  -1,   2,  -6,   4,  -6, -15,  46,   0,   0,   0,   6,  -9,   4,  10, -55,   0,   0,   1,  -4,   5,   0, -11,   9,   1,   4,   2,  17,  10,   9,  12,  50,   1,   2,  -2,  -8,  12,  13, -13,  -7,   1,   0,   3,  27,  -5, -10, -23, -10,   0,  -1, -23, -54,  22,  -4,  -4,  -8 },
      {   1,   0,   0,   1,   3,   8,   7,  12,  -2,  -3,  -5,  -6,  -3,  -4, -19, -52,   0,   0,  -1,   6,   6,   9,  10,  -8,  -1,  -2,  -2,   2,  -7,   1, -20,  -8,   0,   3,   1,   3,   7,  24,   0, -50,  -2,  -5,  -3, -22, -24,   1, -44, -48,   0,   4,   4,  15, -19,  11, -17,  13,   1,  -1,  -6, -39,  -4,   1,  20, -25 },
      {   1,   1,   4,   0,  -1,  -3,   6,  23,   0,   1,  -2,   0,   4,   1,  -7,  -8,   1,   2,   2,   0,   0, -13,  16,  36,   0,   1,   0,   3,  -1,  13,  -5, -32,   1,   1,   0,   2,  11, -13,  33,  40,  -1,   0,  -1,   5,  -2,   5, -36,  27,   2,   0,   4,   5,  10, -20,  38,  -9,   1,   0, -10,  -5,   1, -36,  43, -47 },
      {   0,   0,   1,   1,  -1,   0,  -1,   8,   0,  -1,   0,   0,  -4,  -4,   0, -18,   0,   1,   2,   3,   4,  -5,  -1,  37,   0,   0,   0,  -9,  -4,   0,  -6,  -3,   0,   1,  -2,   9,   6, -25, -21,  -9,   1,   1,   4, -16,   2,  29,  16,  40,   0,   1,  -5,  18,   6, -61, -18,  14,   1,  -5,  -6, -53,  -6,  40,   7,  31 },
      {   0,   0,   0,  -3,   2,   6,  10,   1,  -1,  -1,  -1,   0,  -1,  -9, -20, -12,   1,   1,  -1,  -3,   3,   8,  18,  -7,  -1,  -1,   0,   4,  -4, -13, -14,  21,   1,   1,  -2,  -5,   0,  11,  20,  -1,  -1,  -1,  -4,   1,   5, -16, -37,  49,   3,   5,   3,   9,   7,  29,  15,   9,   3,   0,  -2,  23,  60,  61,  14,  21 },
      {   0,   1,   1,  -1,  -1,   5,  11, -12,   0,   0,  -3,  -5,  -7,   4,  12,  16,  -1,   0,  -2,  -3,   2,   4,   9, -25,   0,   2,  -2,  -3, -10,  -1,   6,  10,  -1,  -1,  -1,  -1,   8,  11,   5, -19,   0,   0,   0,   1, -12,  -7,   9,  17,   3,   7,   1,   8,  31,  29,  68,  -6,  -2,  -3, -11, -36, -68,  21,  20,  11 },
      {   0,   0,   1,   1,   1,   1,   1,  16,   0,  -1,  -4,   0,  -3,   0,   7, -13,   0,   0,   2,   1,   0,   9,   6,  24,  -1,   0,  -3,   1,  -2,   1,  -4, -20,   0,   1,  -2,  -2,   9,  14, -13,  69,  -1,  -1,   1,  -2,  -9, -10, -16, -19,  -1,  -1,  -5,  -9,   8,  31, -41, -31,  -2,   4,   6,   5, -45,  54,   1,   5 },
      {   0,   0,   1,   1,   1,  -2,   1,  -4,   0,   1,   0,  -3,   3,   8,  13,  -4,   0,   0,  -1,  -1,  -1,  -9,  -4,  -8,   0,   1,   0,  -1,   5,   8,  14, -14,   0,   1,  -1,  -4,  -7, -19,   2,  23,  -1,  -1,  -1,  -5,   7,  19,   2, -79,   1,   0,  -1,  -1,   2, -24,  49,  14,   1,   2,   0,  10,  30,  32,  37,  36 },
      {   0,  -1,  -1,   1,   1,   1,   0,  -4,   0,   0,   2,   1,  -3,  -6,   5,  13,   0,   0,  -1,   0,   2,   1,   0,  -7,   0,   0,  -1,   0,  -5,   1,  16,  14,   0,   0,  -2,  -2,   2,  -2, -20,  14,   1,   0,   4,   5,   0,   6,  26,   1,   0,   4,   0,   3,  -2,   3,  -8,  78,  -1,   0,  -3,  14,  -7,  37,  19, -78 },
      {   0,   1,   0,  -2,  -2,   2,   4,  10,   0,  -1,   1,   1,  -1,  -1,  -6,  -5,   0,   0,   0,  -1,   3,  -2,   6,  20,   0,  -1,  -1,  -1,   0,  -5,  -6,  -5,   0,   1,   2,   1,   4,   4,   7,  32,  -1,  -1,  -1,  -4,   0, -10, -24,  -7,   2,   4,   4,   5,  14,  18,  14,  83,  -1,  -4,  -7, -13, -11, -31, -56,  41 },
      {   0,   0,   0,   1,   1,  -2,   4,   6,   0,   1,   0,   3,   1,   8,  -2,  -8,   0,  -1,   1,   2,  -3,  -4,   7,  -2,   1,   0,   1,   0,   3,   3,  -7,   5,   0,   0,   1,   0,  -2,  -7,  16,  -7,   0,   0,  -1,   1,   5,  22, -18, -12,   0,   1,   0,   1,  -6, -30,  38, -25,  -2,   1,  -4,   7,  -6,  42, -89, -43 },
    },
    {
      { -56,  67, -43,  25, -11,   5,  -3,   1,  42, -53,  33, -17,   7,  -2,   1,   0,  -9,   5,   0,  -3,   3,  -3,   2,  -1,  -1,   5,  -4,   4,  -3,   2,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,  -1,  -1,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -59,  11,  -4,  -2,   3,  -4,   0,  -1, -25,  25, -20,  20, -11,   7,  -4,   2,  72, -51,  30, -18,   9,  -2,   1,   0, -24,  20, -11,   1,   2,   0,   1,  -1, -13,   2,  -1,   4,  -1,   1,  -1,   0,   5,  -5,   0,  -4,   0,  -1,   0,  -1,   7,   3,  -2,   2,   0,   1,   0,   0,   0,   0,   2,   2,   1,   1,   1,   0 },
      {  68,  24, -54,  33, -19,  12,  -8,   3, -61,  -6,  18, -10,  14,  -8,   5,  -2,   0, -27,  22, -14,   4,   0,   0,   0,  -3,  18, -11,   3,   0,   0,   0,   0, -12,   0,  -1,   6,  -1,   1,   0,   0,   4,   0,   2,   0,   0,   0,   0,   1,  -5,   1,   0,  -2,   0,  -1,   1,  -1,  -3,  -3,  -1,   0,   0,  -1,   0,   0 },
      { -29,   3,   5,  21, -13,  20, -10,   8, -44,  11,  11, -20,  18,  -9,   6,  -3,  34,  22, -45,  19, -14,   9,  -4,   1,  42, -41,  28, -16,   6,   4,  -2,   3, -40,  11,   0,  -1,   7,  -4,   5,  -2,  18,  -3,   6,   5,  -2,   3,  -2,   1,   1,   7,  -8,  -2,   1,  -5,   1,  -2,   1,  -8,  -3,   3,  -1,  -2,   1,  -1 },
      { -41,   1, -22, -16,  34, -25,  10,  -9, -67, -13,  -6,  35, -16,   9,  -4,   2, -54,  30,   6,   0,  -5,   0,   3,   0,  21,  17,  -8,  -5,   0,  -3,   1,  -2, -18, -16,  10,   3,   0,   1,  -1,   1, -11,   2,  -8,  -5,   1,  -2,   2,  -2,   1,  -1,   6,   0,  -1,   5,  -2,   1,   0,   5,   2,   0,   2,   2,   0,   0 },
      {  36, -18, -24, -15,  37, -27,  15, -12,  33, -44,  11,  14, -13,  14, -10,   4,  49,   3,  -7,  -1,  -3,   0,  -4,   3,  24,   2,  -5,  -1,  -2,   2,   0,   1, -65,  -1,   9,   1,   4,   3,  -1,   1,  -1,  -1,  -8,   0,  -2,   0,   0,   0,  10,   5,  -4,   2,  -1,   2,  -1,   1,  -3,   1,   2,   2,   3,   0,   1,   0 },
      {  -5, -30,  -2,  -6,   4,  -9,   0,  -5, -31, -24,  41, -18,  11,  -3,   0,   3,  40,  30, -31,  19,  -1,  -1,   4,  -3, -50,  17, -18,  20, -18,  11,  -9,   1,  39, -29,  22, -21,   7,  -4,  -2,   2,   5,   8, -11,  -4,   1,  -5,   3,  -4,   5,   3,   1,   3,  -2,   4,  -1,   1,  -6,   3,   5,   1,   2,   2,   1,   1 },
      {  18,  43,   9, -12,  31,  -3,   2,   5,  -6, -19, -20,  19,  -5,  -2,   8,  -7,   6,  -5, -24,  22, -34,  28, -19,   8, -48,  12,  30, -35,  34, -20,  10,  -2,  10, -15,  -6,  11,  -6,  -1,   3,  -2,  22,   8,  10,   6,  -1,   5,  -3,   6, -16,   5,  -4,  -5,   0,  -9,   3,  -2,  -1, -11,  -9,  -2,  -4,  -4,   0,   0 },
      {   7,  21,   9, -21,  27, -12,  10,  -5, -15, -15,  12,  -7,   1,   2,  -1,   1,  26, -34,   4,  14, -12,   8,  -8,   4,  76,  -4,  -7,   1,  -5,   0,   1,   0,  71,   7,  -8,  -6,  -3,  -2,   0,   0, -15,   4,   5,  -2,   3,   2,  -1,   3,  -9,  -4,  -1,  -2,   0,  -3,   0,   0,  -2,   0,  -1,  -2,  -1,   0,   0,   0 },
      { -11, -12,  28,  10,  -3,  13,  -5,   6,   8,   0,  -2, -22,  26, -26,  19, -10,   0,  -7,  -1,  16, -30,  18, -13,   5,  12,  61, -27,   4,   4, -11,   3,   0, -33, -14,   8,   6,  -2,   3,   3,  -1, -63,   7,  10,   2,   3,   2,   0,   4, -18, -12,   3,  -3,  -1,  -3,   2,  -2,  -1,  -4,  -5,  -5,  -2,  -1,  -3,   0 },
      { -16, -41,  23,   6,  21,   0,   2,   0,  -8, -37,  19, -14,  15,  -3,   6,  -4, -15, -11,  14, -19,  -2,   5,  -1,   0,  -1,  22, -11,   6,  -2,   2,   1,   1,  -4,  35, -34,  16,  -6,   3,   3,  -2,  64, -24,  18,  -1,   0,   4,  -5,   4, -39,   6,   0,  -6,   3,  -8,   1,   0,  -6,  -9,  -1,   1,   0,  -2,   1,  -3 },
      {   4,   9,  25,  17, -13,  17, -12,  13,  10,  -3, -18,  12, -15,   6,   1,  -2,  -4,  -3,  -9, -17,  14,  -2,  -3,  -1,  39,  26, -25,  12,   8,  -5,   1,  -6,   0, -60,  41, -14,   3,   3,  -2,   2,  57,  12,   3,   5,  -2,   0,   2,   3,   9,  14, -17,   4,  -3,  -3,   3,  -3, -15, -10,  -5,  -1,  -4,  -5,   0,   0 },
      {  -1, -21, -19,  32,  -3,  12, -15,   9,   4, -15, -25,  29, -12,   2,   4,  -3,  32,  41,   4, -32,  15, -12,   9,  -5,  10,  -1,  20, -12,   6,  -2,  -1,  -3,  23,  -9,  -9,   6,  -4,   1,  -2,   3, -27,  20,  -2,   3,   3,  -2,   3,   1, -70,  -3,  12,  -6,   0,  -3,   2,   0, -19,  -6,   0,  -1,  -2,   0,   0,   0 },
      { -18, -57, -56,  28,  10,  -1,  -1,  -9,  22,   4, -18,   1,  28, -20,   8,  -3,  -9, -19,  24,  16, -32,  16, -11,   6,   8, -10,  14, -24,   9,  -2,  -5,   3,  23, -27,   2,   3,   4,  -3,   6,  -1,  13,   5,  -7,  -4,  -5,   4,  -4,   0,  30,  11,  -9,   4,   0,   0,   1,   1,  -9,   3,   5,   0,   0,   1,  -1,   1 },
      {   2, -13,  20,  22,  -4,  14,   1,   7, -13, -35,  -1,  -1,  -5,  10,  -6,   0,  13,  23,  -2, -28,   5, -17,   6,  -3,   2,  12,   2, -18,  22, -12,  12,  -4,  20,   1, -18,  38, -26,  18,  -5,   0, -21, -11,  21,   1,   4,   1,  -3,   3,  73,  -4, -12,  -1,   3,  -6,   1,  -2,  23,   2,  -8,   3,  -3,  -3,  -2,  -2 },
      {   5,   1,  -3, -31,   2,  32, -51,  28,  -2, -39, -25,  40,   6, -23,  29, -14,   0,  -3,   5,   8,   5,   9,   8,  -4,  -1,  -2,  -1,   4, -22,  28, -21,  10,   0,  14,  -1, -22,  14, -21,   8,  -3,  -8, -23,  13, -15,   9,  -6,   3,  -4,  17,  13,  -5,   2,  -4,   1,  -1,   0,  17,  -2,   2,   6,   0,   4,   3,   0 },
      {  -6,  -4,  37,   5,  15,   3,   2,  -1, -18, -42,   1, -11,   9,  -6,   6,  -4, -17, -41,  12, -20,  21,  -3,  -6,   4, -17, -42,  25,   7,  -6,   5,  -8,   1, -15, -15,   0,  -8,   5,   3,  -4,   3, -26,  39, -16,  15,  -7,   2,   0,   4,  14,  -1,  -4,  -1,  -6,  -5,   2,   1, -52, -11,   1,  -6,  -2,  -6,   2,   0 },
      {  -2,  -7,  11, -14, -13,   3,   0,  -4,  -4, -23,  -2,  -4,  15, -13,   8,  -2,   4, -13,  16,   5,   9, -11,   4,  -2,   6,  -5,  14,  -8,   0,  -8,  -1,   2,  -9, -21,   0,   5,  -6,   5,  -6,   0,  22,  43, -38,   4,  -5,   1,   2,  -1, -13, -29,  28,  -7,   0,   6,   0,   2,  84,  14,  -5,  -2,  -1,   3,  -1,   1 },
      {   6,  22,   3, -29, -16,   2,   4,  -3,   0,   2, -22,   1,  24, -25,  14,  -8,  18,  41,  20,  -1, -15,   0,   1,   2,  10,  14,  -9,   3, -12,   1,   4,   5,   0,  14, -19,  23,  -7,   4,   1,  -2,  30,  17, -28,   7,  -4,   1,   1,  -2,  28, -23,  32,  -5,   4,   6,  -4,   1, -57,  22,  12,  -5,   7,   3,  -2,   0 },
      {  -5, -21,   1, -26, -51,  -9, -24,   8,   0, -10,  48,  21, -23,  -3,   6,   1,  -4, -20,  -1,  27,  22,   8,   5,  -4,  10,  15,  12, -29,  42, -31,   1, -13,  -2,   9,  -3,  -4, -17,  -1,  -9,   2,  -3,  -3, -17, -15,  -5,  -4,   8,  -7,  -1,  10,   5,   8,  -3,  11,   3,   2, -28,   6,  10,  -1,   0,   9,   3,   5 },
      {   0,   4,  15,  -1, -29, -20,   8, -10,  -8, -19,  -2,   5,   6,  -7,  12,  -8,  12,  29,  44, -24, -13,  35, -34,  19,  -6, -44, -24,  17,  20, -30,  20, -12,   2,  -1,  18, -19,  15,  -2,   1,  -1,  -6, -28,   6,   2, -15,   8,  -1,   2,  -9,  13,  -6,  13,  -4,   9,   1,   0,  16,   6,   4,  -3,   3,  -1,  -2,   1 },
      {   4,  12,  16, -25, -28,  -3,  22, -17,  -5, -16, -14, -14,  34,   6, -20,  13,   3,   3,  27,  18,  -4, -29,  30, -13,   1,  -2,  14, -31,  -7,  20,  -6,   5,  -7, -45,  -2,  22,   3,  -1,   3,   0,  -7, -30,  15, -19,   0,   1,  -5,  -4, -25,  32, -21,  13,   3,   3,  -2,   3,  -7,   5,  11,   1,   3,   6,  -3,  -2 },
      {  -3,   4, -16, -38,  32,  12, -12,  15,   8,  33,  39,  -2,   8, -28,  23, -22,   2,   9,   7, -17,  12,  -8, -13,   8,  -3, -31, -18,   3,   0,   0,   3,   1,  -6, -32,  -5,  34, -38,  26, -14,   6,   0,   7,  11, -13,  22, -12,   5,   2, -10,  14, -17,   4,  -3,  -1,   4,  -4,   3, -14,  -2,  -4,  -3,  -3,  -2,   2 },
      {   1,   1,  -3,  25, -12, -30,   6, -17,  -4, -16, -45,  -5,  -1, -12,  12,   1,   0, -11, -27,  29,  23,   6, -17,   4,   0,  -9,   2,  29, -14, -14, -11,  -9,  -4,  -4, -15,   4, -27,  22, -31,  16,   6, -11,  23, -40,  20, -28,  20,  -6,  -3,  -8,  11,   8, -11,   9,   5,   0,  -4,  17,  -5,  -6,  -9,   6,   0,   5 },
      {  -5,  -9, -13, -16, -24,  16,  10,  -8,  -2, -16, -24,  21,  22,   3, -12,  -2,  -8, -23, -49,  -7,  -3,  -2,  -5,  13,  -4,  -8, -34,  -1, -11,   2,  23,   2,  11,  36,  34,  39,  -2,  25, -10,  -7,  -7,  15, -22,   9, -10,   4,  -7,   1, -13,  31, -11,   5,   2,   1,   0,  -2,   4,  -6,  18,  -2,  14,  -2,   0,  -3 },
      {  -4, -11, -12, -13,  -1,  37,  -4,  20,  -7, -18, -17, -22, -22,  27, -16,  10,   1,   5,  32,  36, -30, -18,   2,  -5,  -8, -21, -25,   0,  13,  -1,  13, -10,  -6,  20,  18, -16, -22,   8,  -2,   1,   8,  27,  16,  -5,  16, -17,   0,  -1, -10, -41, -26,  20,   6,   1,  -1,  -1, -11,   0,  -1,   4,  -5,  -3,  -3,   0 },
      {   1,   8,   0,   0,   7, -15,   4,   3,  -1,   6, -14, -26,   4,  -2,   7,  -2,   3,  29,  25,   8,  20,   8,  -2,  -7,   4,  32,  21,   5,   0,  -7,  -3,  -6,   1,  36,  -7, -12,   2,  -4, -10,   9,   2,  49,  -8,  -8,  -2,  -8,   9,   0,  10,  51, -38,  20, -13,   4,   0,  -1,  17, -36,  23, -13,   1,   1,   2,   3 },
      {   4,   7,  23,  30, -11, -23,  -5, -11,   4,   3,  17,  33,  -4,  -6, -15,  15,  -1,   2,   3,   7, -16,  24,  12,  -9,  -3, -23, -51, -13,   8,  27, -19,  13,  -4,   0, -22,  25,   6, -24,   9,  -5,   3,  50,  18, -32,   9,   5,   5,  -7,   3,   1,   2,  -2,   1,   6,  -2,   0,   3, -14,   6,  11,   1,   4,   4,   3 },
      {  -4, -10, -16, -27, -23,  35,   6,  -7,  -1,  -2,   2,   3,   1,  32, -27,  10,  -1,   3,  -6,   1,  14,  23, -35,  25,   3,   8,   7,  15, -21, -15,  22,   1,  -4, -31, -55,  -3,  27,  -3,   9, -12,  -2,  11,  13, -12,  17,  -5,  -9,   6,   7,  -7,  13,  -3,  -7,  -2,  -1,   1,   1, -33,  10,   0,   8,  -2,   2,  -3 },
      {  -3,  -5, -25, -28,  10, -10, -10,  11,  -1,  -2, -29, -32,  15,   4,   2,   2,   1,   1, -12,  11,  36,  -2,  22, -17,  -1, -13, -33,  15,  39, -20,  -4,  -8,  -2,  -7, -10,   4,   5, -21,   6,   0,  -5,  13,  34,   6, -44,  22,  -7,   7,   2,  12,  21, -28,   6, -10,   1,  -1,  -3,   0,  -7,   8, -12,   0,   5,  -2 },
      {   3,   7,   6,   7,  29,  -8, -27,   0,   5,   6,   0,  -1,  45,   6,  -7,  -1,  -1,  -5, -11, -21,   0,  -6,  38, -15,   0,  -6,  -1, -13,   9, -37,  35,  -5,   0,   4,  16, -16,  21,   1,  -8,  -2,   3,  -7,   0, -39,  23,  -1,  -7,   2,   4, -30,  17,  20, -25,  15,  -7,  -3,  -3, -20,  28,  -8,  14,   4,   2,  -2 },
      {  -3,  -5, -12, -28, -17, -20,   3,  -8,  -1,  -1,  -6, -25,   2,   4,   7, -13,  -3,  -2, -16, -44,  14,  17, -15,  18,   3,  15,  25,  -7,  17,  24, -22,   7,   5,   7,  30,  24,   6, -26,  23, -11,   2,  11,  20, -23,  -4,  -9,   6,  -8,  -4, -47, -26,  21,  -3,  14,  -1,   2,  -3,   3,  -6,  17,   0,  -5,   1,  -4 },
      {   2,   2,   4,  14,  16,  -4, -38,  17,  -4,  -1,  -9, -35, -41,  22,  13, -14,  -1,  -4,  -6,  10,  -1,   5,   1,   9,  -1,  -6, -16,  -4, -10, -11,  -4,  21,   1, -16, -27,  35,  26,   4,  13, -11,  -3,  -7, -24,   3, -14,   9,   2,   2,  -9,   5, -14,  31, -14,  14,  -8,  -2,   3,  41,  29, -15,  18,   4,  -7,   2 },
      {   0,   1,  -9,  -9, -17, -19,  -7,  -4,  -1,  -2,  -5,  17,  -7, -23,   0,   7,   0,  -7, -19,  -8, -35, -13,  22, -14,   2,  -1,   8,  49,   8,  -5, -16,  -8,  -1, -15, -34,  -4, -19,   5,   2,   5,  -1, -12, -13,  13, -24,  24, -12,   3,  -2, -32, -37,  20,  12,   3,   5,   1,   6, -41,  25, -16,  -2,  -4,  -1,   1 },
      {   1,   0,   4,   5,   1,  -9, -15,   7,  -1,  -6, -15, -31, -21,  10,  -5,   5,   1,   4,  17,  12,   9,  30,  -5,  -5,   1,  -7,   1,  -3,   1,  18,  -7,  14,   3,  -5,  30,  22, -22,   5,  -3,  -5,  -3, -34, -35, -13,   8,   0,   1,   0,   3,  -8,  26, -37,  20,  -7,   7,  -1,   1, -62,  29,   8,   6,   2,  -2,   0 },
      {  -1,  -2,   0,  -7, -10, -14, -15,  35,   2,   0,   4,  -3,  15,  40, -11,  -6,  -2,  -6,   2, -22, -44,  28,  25, -10,   3,   0,  31,  51,   1,  -3,   8,  -1,  -3,  -5,   3,   9, -17,   0,  -5,  -8,  -3,  17,  -8, -23,  14,  -6,   4,  -6,   0,  32,   7, -13,  15, -14,   1,   3,   0,  31, -16,  16,   0,   2,  -3,   4 },
      {   2,   2,   5,   7,  19,  34,  -2, -31,  -1,  -1,   7,   3, -22, -38, -10,  30,   2,   4,   7,  10,  18,  10,   2,   4,   0,   3,  25,  28,  -9,  -9,  24, -13,  -2,   5,  24,  17, -19, -23,  17, -17,   0,   6,   6, -38,  -5,  37, -22,  21,  -2,   3,  -7,   2,  13, -10,   9,  -1,  -1,  28,  16,  -1,   2,  11,  -2,   4 },
      {   0,  -1,   4,  16,  -8, -26,   9,  36,   2,   4,  14,  31,  18,  24,  10, -42,   3,   5,  17,  41,  18, -26, -10,  14,  -2,   1,  12,   9, -27, -12,   4,  11,   2,   6,  18,  20,  15,  12,   5, -15,   3,   5,  15, -23, -23,  29, -11,  14,   2, -23, -14,  -9,  13, -12,   0,  -1,  -4, -15, -11, -10,  -1,  -3,  -4,   3 },
      {   2,   9,  10,  12,  14, -14, -51, -17,   1,   6,  -4,  10,  45,  27, -19,  13,  -1,   8,   3,  16,  18, -14, -26,  38,   1,   7,  -6,   7,  23,  29,   2, -18,  -1,   1, -14, -10, -32,   3,  18,  -2,  -3,  -7, -31,   8,  -6,  -1, -14,   9,  -3,   0, -13,   7,  -6,   7,   8,  -4,   0,  10,  -7,  26, -11,  -1,   9,  -2 },
      {   2,   3,   5,   8,  22,  48,  20,  -2,   1,   3,  10,  18,  17,   9, -24,  -1,   2,   3,   9,  13,  -2,  -3,  -7,  -3,   2,   0,   7,  37,  36, -11, -40,  12,   4,   5,   9,  29,  11, -14, -12,   3,  -1, -17, -17, -13, -30, -17,  36, -12,  -2,  -1,   8,   5, -28,  18, -15,  11,   0,  -6,  -5, -15,  -1,  -9,   2,  -5 },
      {   1,  -1,   2,   4,  -2,  -8,   4,  -1,  -2,  -5,  -4, -12, -30, -31,  -3, -32,   0,  -8,  -4,   3, -24, -56,   1,  30,   1,   1,   6,  20,  28,  18,  33, -12,   2,   1,  -5,   0,  26,  -1,  23,  -9,   0,   4,  -3, -23,   5, -16,   9,   3,   2,  24,  22, -14,   2,  31,  -7,   1,   0, -19,  -1,  23,  -8,   6,  -7,  -7 },
      {   1,   2,  -1,   3,   0,   7,  42,  19,  -2,  -2, -11,  -1,   9,   1,   9,  -8,  -1,  -1, -15, -11,  11, -14, -22, -27,  -2,  -4, -20,  -4,  17, -26, -23,  33,  -1,   3, -14, -31, -23,  -3,  47, -17,  -3,  -5, -40, -30,  34,  17,  -9,  10,  -2,   9, -11,   1,   6,  -4,  -5,   1,  -2,   8,  -1,  12,  -5,   3,  -7,   0 },
      {   1,  -2,   0,  12,   0,   1,  29,  27,   0,  -1,   1,   5,   4,  11,  -3, -34,   0,  -6,  -7,   2,   4,   4,   4,  18,   1,  -1,  -1,  -4,   1,  28,  15, -20,  -4,  -9,  -1, -13, -54, -46,  12,  -2,   0,  -1,  17,  14,  -3,  -1,  -2,  -9,   1,   4,  27,  12,  -9,   5,  -6,   8,   7,   8,  46, -38,  18, -14,  11,  -7 },
      {   1,   4,   2,   2,   1,  18,  10, -39,   1,   5,   7,   8,  -1,  27,  45, -16,   0,   2,   0,   1,  -1,  -2,  11,  11,   2,   3,   2,  12,  42,  24, -15,  21,   0,   1,  -3, -20,  14,  35, -17, -11,  -1,   8,  13, -15,  -6, -21, -33,   0,   1,  -1,   7,  -5,  17, -32,  13,  -6,   3,  13,  40,  -6,  10,   5,  -3,  -2 },
      {   1,   2,   1,   6,  12,   7,  29,  41,   0,   2,  -3,   0,   4, -27, -27, -11,  -1,   2,  -5,  -3,  20,  29,  10,   9,   0,   4,  -8, -12,  10,  24,  21, -19,  -1,  -7, -24, -17,  21,  11, -40,   2,  -3,  -8, -25, -23, -21,  -4,  12,  -2,  -4, -20, -34, -10,  26,  -5,  -3,  -1,   1,  21,  15,  27,  -2,   6,  10,  -6 },
      {  -1,  -2,  -1,   0,   1, -16,  -5,  28,  -1,  -2,   3,   8,   6, -31, -38,  16,   0,   0,   7,   9,  17,  13, -18, -30,   2,   4,  18,  18,  22,  19,  16,  48,   1,   0,   9,  -6,   1,  47,  25, -10,   1,   5,  31,  27,   0, -13, -10,  -3,  -2,  -1,  11,   8,  -7,  21,  -3, -11,   2,  12,  15,  -3,  21,  -1,  -1,   4 },
      {  -2,  -4,  -5,   2,  -5,  -2,  19,  -4,   0,  -3,  -7,   0,  -5,  -4,  11, -15,   2,   2,   5,  21,  22,  24,  38,   8,   2,  -2,  -4,  20,  28,  24,  22,  12,   2,  -2,  -5,  13,   5, -17, -19, -10,  -1,  -9, -21,  21,  38,  12, -27,  14,  -1,  -4,  -4,  30, -25,  -5,  -4,  -7,  -4, -14, -56,  -6,  20,  11,  -7,   2 },
      {   0,   0,   2,   6,   2,  -7,   7,  -5,   1,  -1,   3,  13,  12,  -2,   7,  -2,   2,  -2,   2,  14,  15,   3,  20,  11,   1,  -4,   1,  16,   0, -30,  12,  11,   0,  -2,   4,  20,   7, -28,  14, -17,   2,  -3,   3,  49,  33, -39,  14, -11,  -1, -11, -33, -29,   4,  -5,   0,   8,   2,   8,  40,  17, -38,  27, -11,   8 },
      {   1,  -1,   1,   4,   1,  -1,  16,  18,   2,   2,   6,   9,   5,  19,  51,  29,  -1,  -5,  -2,   7,   1,  -7, -37, -61,   1,  -1,   4,  11,   5,  28,  31, -36,   1,  -4,  -2,  29,  21, -15,   7,  25,  -2,  -4, -11,  -1,  -3, -13, -14,   4,   0,  -6,   2,   7, -12,   6,   5,  -5,   0,   1,  12,  15,  -5,   1,   7,   1 },
      {  -1,  -3,  -3,  -8,  -3, -10,  -9,  16,  -2,  -2,  -1,  -5,   3,  -8, -32,  -1,  -1,   0,  -2,  -9,  -6, -24, -30,   6,  -1,   3,  -2,   1,  21,  13,  -3,  14,   0,   5,   4,   7,  24, -25, -26,  25,   2,  -3,  -3,  -1,  27,  25,   8,  -5,   1,   0,  17,  20,   1, -46,  38, -21,   1,   9,  16, -29, -36,  14, -21,  22 },
      {   1,   1,   0,   1,   2,   5,  23,  27,   1,   2,   3,   2,   2,   5,   8,  21,   0,  -1,  -1,  -7,  -6,  11,  18,  22,  -1,  -1,   1,  -7, -10,   3, -27, -18,   0,   0,   4,  -1,  -6,  18,   5, -20,   0,   0,   8,   7, -19,  -8,   8,  55,   2,   4,  33,  41, -11,   8,   5, -13,   1, -13,  11,  27, -25,  37, -20,  37 },
      {   0,   0,  -2,  -5, -10,  -5,   1,   4,   0,   1,   1,  -2, -10,  -8, -26, -24,   1,   3,   5,   2,  -6,  12,   7, -12,   0,   0,  -1,  -1,  -2,   1, -34, -33,   1,   4,   6,   9,  14,  15,   1,  -6,   1,   0,  -3,   1,  23,  -5, -38,  43,   0,  -1,  -5, -28, -49,   7,  29,   4,   2,  15,  14,  -8, -28, -30,  16, -33 },
      {   1,   3,   3,   5,   7,  -7,  -5,   5,   0,   0,   0,   0,  -1, -17, -20,  -7,   1,   1,  -1,   0,   7,   2,  -2, -12,   0,  -4,  -8,  -3,  -9, -16,  -7, -24,   1,   0,  -1,  10,   7,  -6,  25, -25,  -2,  -3, -14,  -5, -21, -66, -32, -22,   2,   4,  13,  38,  33, -36,  22,   6,   1, -10, -25, -22,   2,  12,   2, -20 },
      {   0,   0,  -1,  -1,   1,  -9,   4,  23,   0,  -1,  -1,   1,   6,  -7,  15,  43,   1,   0,   2,   7,  10,  -3,   9,  33,  -1,  -2,   3,   8,  18,  -1, -26, -36,   2,   2,   7,  12,  25,  24,  19,  -1,   0,   0,   2,   7,  44,  20,  17,  -7,   1,   3,  10,  12,  31, -19, -41,   4,   2,   3,   9, -18,   7, -23,  21, -26 },
      {   1,   1,   3,   5,   5,   5,   3,   0,   0,   1,   2,  -1,  -5,  -1,   0, -14,   0,   0,   3,   6,   0,  -3,  -9,  -5,   1,   3,   6,   5,  -9, -18,  -9,  19,   0,  -1,   3,   6,  -3, -10, -39,  -9,  -1,   2,   9,  15,   8,  23, -34, -30,   0,   9,  25,  46,  21,  20,  -4,   6,   0,  -5,  15,  59, -32, -22,  26, -26 },
      {  -1,  -1,  -1,  -1,   0,  -9,  -9,   3,   0,   0,  -1,   0,   7,   2,   5,  19,   0,  -1,  -2,  -5,  -4, -17, -18,  -9,   0,   1,  -1,  -3,   7,  -3,  -3,  16,   0,  -1,  -2,  -1,   3, -35, -44, -42,   0,  -1,   2,   7,  10, -21,  10,  62,   1,   4,  -5,  -5,  48,  36,  10,  -7,  -1,   5,  -4, -25,  19, -16,  10,  -8 },
      {   0,   0,  -1,   0,   2,  10,  -9, -38,   0,   0,   2,   5,   5,  12, -10, -50,   0,  -1,   0,   9,  13,  23,   6, -39,   0,  -1,   1,   1,   0,  12,  11, -18,   0,  -1,  -2,   0,   5,  17,  24,   7,   0,   1,   2,  12,  11,  16,  57,  26,   2,  -1,   9,  28,  37,   1,   6,  -4,   0,   0,   3,   2, -14, -11, -23, -21 },
      {   0,   1,   3,   2,   0,   4,   1,  -2,   0,  -1,   1,   3,   1,  -3, -17, -22,  -1,  -1,   1,   3,   2,   0,   2,  10,   0,   0,   3,   5,  -1, -22, -24,  -2,   0,   1,   4,   9,   4, -18,  10,  79,   0,   1,   2,   6,   9, -21, -29,  28,   0,   0,  -1,  -2,  29,   5, -18, -38,   1,   4,   7,  24,  43,   1,  -1,  16 },
      {   0,   0,   1,  -1,   2,   5, -11,  -6,  -1,   0,  -1,  -3,   1,   7, -12, -17,   0,   0,  -4,  -8,   1,  10, -15, -13,   0,   1,  -1,  -3,   5,  14,  -5,  -8,   0,   1,   3,   1,   4,   8,  -9,  11,   0,   2,   2,  -2,   8,   6, -39,   6,   0,   2,  -1,  -5,  25,  28, -58,  50,   0,   1,  -5, -38, -38,  45,   3,  26 },
      {  -1,  -1,  -1,  -2, -11, -12,  -9, -11,   0,   0,   0,   0,  -8, -11,  -4,  -5,   0,   1,   1,   0,  -6,  -7,  -2,  -7,   0,   0,  -1,   0,  -2,   2,  22,  30,   0,  -1,   1,   0,  -7,  -7,   0,   4,  -1,  -3,  -5,  -6, -15, -22,  16,  45,  -1,  -3,   3,  15, -21, -63, -51,  18,   2,  -1,   5,  14, -16, -24,  48,  10 },
      {   0,   1,   0,   0,   2,  -3,  -7, -11,   0,   0,  -3,  -3,   0,  -8, -16, -26,  -1,   0,   1,   2,  -2, -16, -24, -27,   0,   0,   1,   3,   4,  -3, -25, -45,   0,  -1,   0,   1,   4,  -1, -20, -51,   0,   1,   7,  15,  16,  13,   6, -15,  -1,  -2,  -2,  -7, -15, -24, -18,   2,   1,   1,   9,  28,  52,  11,   0,  47 },
      {   1,   1,   0,   1,  -1,  -5,  -3,  10,   0,   1,   1,   1,   0,  -1,  -1,   4,  -1,   0,  -1,  -1,  -4, -10, -12,   3,   0,   0,   1,   3,   3,  -3,  -7,   9,   0,   0,   3,  -1,  -4,  -3,  -8,  21,   1,   5,  10,   8,   2,  11,  12,  23,   0,  -2,   1,   5, -10, -17,  20,  66,   1,  -1,   5,  21,  39,  58,  -7, -60 },
      {   0,   1,   1,   1,   3,   1,   2,   5,   0,   0,  -1,   0,   2,  -5,   2,   6,   1,   1,  -1,   4,   8,   2,  10,   9,   0,   1,  -4,  -1,   6,  -2,   4,   2,   1,   2,  -2,   2,  11,   2,   7,  16,   0,  -2,  -2,  -1,   6,  -6,  -6,   4,   1,   1,   3,   9,  13,   4,  36,  80,   1,   1,   1,  11,  11, -60, -23,  54 },
      {   0,   0,   0,   0,   3,   9,   8,  -4,   0,  -1,  -2,  -4,   0,   7,   2, -12,   0,   1,   1,   0,  -3,   0,   8,   2,   0,  -1,   0,   3,   2,  -3,   1,   4,   1,   0,   2,   4,   8,  11,  20,   7,   0,  -2,   0,   4,   9,  11,  18,  -4,   1,  -1,  -1,   3,  11,  15,  50,   2,   0,  -2,  -6,  -9,   5,  31,  98,  32 },
    },
  },
  { // 20
    {
      {  42, -76,  50, -29,  16,  -3,   3,  -1, -51,  38, -16,   6,   6,  -6,   4,  -1,   3,   8, -12,  16, -13,   7,  -3,   0,   1,   3,   2,  -5,   1,   0,  -1,   1,   4,   1,  -4,   2,   0,   0,   0,   0,  -4,   1,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,  -2,  -1,   0,   0,   0,   0,   0,   0 },
      { -84,  15,  10,   2,   4,  -2,   3,  -1, -34,  37, -23,  14,  -2,   3,   0,   0,  67,  -6,  -9,   8, -12,   5,  -5,   1,  -4,  17,  -2,  -5,   6,  -5,   1,   0, -20,  -3,   9,  -2,  -1,   0,   1,   0,  -1,  -4,   3,  -1,   2,  -1,   0,   0,  -5,   1,   1,  -1,   0,   0,   0,   0,  -5,  -1,   3,   0,   0,   0,   0,   0 },
      {   4, -20,  17,  -9,   6,  -6,   4,  -2,  46, -37,   5,  -4,  -2,   4,  -2,   2,  59,   5, -22,   3,  -9,   3,  -3,   0,  75,   0, -20,   5,  -5,  -3,   0,  -1,  16,  17, -10,  -3,  -1,   0,  -2,   0, -30,  11,   4,  -3,   2,   0,   0,   0,  -9,   1,   0,   0,   2,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0 },
      {  37,  -7,   2,  -4,  -1,  -1,  -2,   1,  37, -15,  -3,   3,  -9,   4,  -2,   0,  44,  33, -23,   0,   1,  -5,   1,  -1, -56,   3,  20,  -2,   5,   0,   1,   0, -76,  -1,  12,  -1,   5,   1,   1,   0,  -3,   9,  -2,   2,   2,  -1,   1,   0,  -3,   1,   1,   1,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 },
      { -58, -20,  54, -18,  11,  -8,   4,  -1,  43,  -1, -19,   8,  -2,   2,   1,  -2, -58,  16,  10,  -8,   7,   1,  -2,   2,  11, -31,  11,   5,  -1,  -1,   0,  -1, -37,  19,   1,   2,  -1,   3,  -1,   2,  -1,   4,  -4,   2,   0,  -1,   0,  -1,   5,   2,  -3,   2,  -1,   0,  -1,   0,   4,   0,  -2,   0,   0,   0,   0,   0 },
      {  31,   8, -17,   9,  -7,   1,  -1,   0,   5,  40, -25,   7,  -6,   3,  -3,   1, -13, -47,  33, -13,   5,  -1,   1,   0,  45,  24, -19,  -2,   1,  -5,  -1,  -1, -62,   6,  19,   1,   0,   4,   0,   0, -37, -13,   9,  -1,   4,   0,   1,   0,   7,   6,  -3,   3,  -1,   0,   0,   0,  -3,   0,   1,   1,   0,   0,   0,   0 },
      {  14,   5,  -3,  -7,   2,   5,  -4,   2, -23,  -6,  -7,  24, -10,   1,   4,  -2,  38, -30,  40, -33,  16,  -5,  -1,   2,  10, -70,  32,  -3,   0,   2,   1,   0,  -7,  21, -31,   9,   1,   1,  -3,   0,  23,  27, -16,   4,  -5,   1,  -1,   0,   6,  -5,  -2,   4,  -3,   1,  -1,   1,   4,   5,  -6,  -1,   2,   0,   0,   0 },
      {  11, -13,   8,  13, -10,   0,   0,   0,  60,  37, -43,  12, -11,   6,  -5,   1,  13, -26,  14,  -6,  -1,   5,  -5,   3, -35,  20, -22,  14,  -3,  -2,   3,  -1,  38,  32, -22,  -2,  -3,  -2,   2,   0,  39,  -9,   3,  -2,  -4,   0,   0,  -1, -20,   6,   5,  -6,   4,  -1,   0,   0,  -1,  -4,   3,   0,   0,   0,   1,   0 },
      { -11, -21,  29,  -8,   8, -11,   9,  -5, -10, -42,  41, -31,  24,  -8,   5,   0,  14, -48,  28, -10,   7,   0,   1,   0, -15,  36, -30,  21, -11,   4,  -5,   1, -27,   2,   0,   8,  -6,   3,   0,   0,  40,  -9,  -6,   4,  -5,   2,  -2,   1,  15,   8,  -8,   1,  -2,   1,  -1,   0,   6,  -1,  -2,   1,   0,   0,   0,   0 },
      {  -1,   0,  12, -36,  35, -16,   8,  -2,  21, -11,  -1,  32, -26,   8,  -4,  -1,   9, -22,  30, -22,   7,   0,  -6,   2, -31,   9,   0, -10,   6,  -7,   5,  -3,  36, -57,  27,   2,  -5,   5,  -3,   1, -43,   5,  -2,   0,   6,   0,   0,   1,  18,  -6,  -6,   6,  -5,   1,  -1,   0,  -2,   5,  -1,   0,   0,   0,   0,   0 },
      { -14, -11,   5,  12, -18,  18, -11,   3,  -8,  -5,  10, -15,  24, -18,  12,  -3,   5,  -5,  18, -10,   5,  -2,   3,  -2, -46,  -4,  -5,  11,  -7,   7,  -3,   2,  11,  47, -24,   0,   5,  -4,   0,  -1, -84,   6,  23,  -4,   2,   0,   0,   0,   0,  -3,  -3,   1,   2,  -1,   0,   1,   6,   4,  -2,   0,   0,   0,  -1,   0 },
      {   7,  30,  -8, -25,  38, -34,  21, -10, -30, -17,   9,  24, -37,  31, -19,   7, -17,  12,  -7,   1,  -4,   6,  -8,   4,  -8,  26, -16,   2,   4,  -8,   5,  -3, -10,  55, -33,  14,  -7,   3,  -2,   0,  -5,  -3,  12,  -3,   2,  -4,   3,   0, -17,  -2,   7,  -3,   1,   1,   0,   0,  15,  -1,  -3,   1,  -1,  -1,   0,   0 },
      {   5,   5,   7,   3,   8, -28,  20, -11,  14,   4,  16, -35,   8,  11, -11,   7,   6, -42, -10,  28, -15,   7,   0,   1, -10, -38,  36, -37,  26, -15,   8,  -1,   7,  21,  16,  -2,  -8,   2,   0,  -1, -10, -53,  12,   8,  -4,   3,   0,   0,   1,   6,   6, -13,   4,  -1,   0,   0, -14, -14,   7,   4,  -2,   1,   0,   0 },
      {   5,  14,  19,   0,   0,  -5,   7,  -2,  -6,  -2,   2, -14,  15,  -4,   3,  -1,  -9, -21,   6,  -4,  -1,   7,  -4,   1,  -2,   1,   8,  -1,   5,  -5,   0,  -1,  -3, -13,  23, -10,   0,   3,  -2,   3,  -8,  41, -16,   7,   0,   1,  -1,   0,-105,  -6,  16,   1,   5,   0,   0,   0,  14,  -7,  -5,   4,  -2,   0,   0,   0 },
      {  27,  63,  30, -24,   6, -21,   6,  -6,  -5,   5, -30,  -8,  35, -21,  20,  -9,  11,  19,   1,  -1, -13,  21, -14,   9,  -3, -31, -28,  35, -14,   7,  -6,   2,   4,   0,  18, -14,   1,  -2,   3,  -1,   2, -23,   8,  -3,   3,  -1,   1,   0,  12,   1, -13,   4,  -2,   2,  -1,  -1,  17,   9,  -3,  -2,   0,  -1,   1,   0 },
      {  -5,   3, -29, -18,  24,   4,   5,  -2,  26,  21,  16,  -2,  13, -25,  10,  -5,  -8, -23,   3,  43, -54,  34, -22,   6,  -4,   4,   3, -14,   1,   3,  -5,   5, -10,  -1, -25,   8,  19,  -9,   3,  -2,   1,  44, -16,   2,  -5,   2,  -2,   1,  21, -13,   3,  11,  -6,   2,  -1,   1,  19,  17,  -9,  -2,   0,  -1,   0,  -1 },
      {   0,   8,   8, -12,   3,   4,  -7,   5,  14,  24,   5, -16,   5,  -9,   3,  -2,  11,   8,  -1,  -4,  22, -26,  18, -11,  10,  17,  24, -24,   6,   9,  -9,   4,   3, -12, -18,  30, -17,   7,  -1,   0,   4, -25,  11,  -8,   3,   0,   0,   1,  -4, -13,  -6,   5,  -3,   2,  -1,   1,  90,   6, -18,   6,  -6,   1,  -2,   1 },
      { -13, -38, -42,  -4,  24,   4,   4,   1,   3,  -1,  20,  14,  -9,  -6,   6,  -4,   4,  12,  23,  -6,  -9,  11, -15,   5,   0, -31, -13,  13, -20,  13,   0,   0, -13, -20,  -3, -24,  30, -14,   5,  -2,   8, -50,  39, -17,   5,  -2,   2,   1, -39,   3,   2,   1,  -2,   1,   1,  -1,  16,  -5,   0,   1,  -1,   1,  -1,   0 },
      {  -2, -20,   4,  35, -24,  -5,  10,  -7, -14, -16, -12,   2, -18,  30, -20,  14,  -2, -10,  -6,   5,  -9,  10,  -6,   3,  -6, -10, -11,  11,   5, -17,  12,  -8,   5,   8,  35, -42,  23, -10,   2,  -1,  -2,  14,  -8,  14,  -9,   1,   0,   0,  26, -16,  -3,  -5,   0,   2,  -2,   1,  69,  11, -13,   3,  -1,  -1,  -1,   0 },
      {  -2,  -2,  -9,   8,  24, -29,  29, -17,  10,  64,  38, -42,  -6,  14, -20,  10,   9,  20,  -3, -20,  21, -22,   9,  -3,  -1, -18, -16,  23, -24,  12,   1,  -1,   3,  -4,  -2,  -7,   7,   3,  -2,   0,  -9,  25, -20,  16,  -7,  -1,   5,   1,   9,  10,   0,  -2,   0,   4,  -2,   2,  -6,   3,   2,   0,  -3,   0,   1,  -1 },
      {   3,   8,   4,  -5,  13,   9, -18,  13,   1,   6,  22,  13,  -8, -14,  14, -12,   4,  20,  20,  -6,  -2,   2,  -9,   0,   6,  18,  13, -23,  10,  -4,   3,  -1,  12,  28,  24, -21,  16,  -9,   5,  -2,   2,   1, -17,  22, -19,  11,  -2,   3,   0,  84, -27,  -1,   1,  -1,   0,   0,  16, -19,   4,   4,  -3,   0,  -2,   1 },
      {   6,   8,  20,   3,  -7,   4,  -3,   0,   6,  16,  31,  -4,   8,   1,  -1,  -3,  10,  46,  58, -19, -20,   1, -13,  -1,  13,  25,  12, -10,  23, -23,   8,  -5,   4,  16,   9, -15,  10,  -5,   2,  -3,  11, -10,   3,   2,  -8,   3,  -4,   0,   7, -60,  27,  -4,   1,   3,  -1,   2, -16,  -4,   0,  -2,   0,   1,   1,   0 },
      {   0,  16,  12, -42,   5,  22, -31,  21,   4,  12,  24,  12, -20,  -7,   1,  -8,  -4, -36, -42,   0,  21, -22,  13,  -3,   0,   3,   3,   5, -19,  18,  -7,   2,   3,  26,  21, -30,  33, -19,   8,  -3,   5, -13,  -2,  15,  -8,   5,   0,  -1,   0, -34,  17,   1,  -6,   3,  -2,   3, -10,  15,  -1,  -5,   1,  -1,   2,   0 },
      {  -9, -25, -52, -27,  19, -20,  16, -12,  -3, -22, -56, -17,  31, -12,   1,   4,   2,  14,  14,  -6,   9, -17,  20,  -6,   2,  17,  17,  -3,  -1,   8,  -9,   8,   9,  26,  25,  -2,   6,  -6,   2,   0,   1,  -7, -19,  33, -13,  12,  -2,  -3,  -4, -13,   1,   5,  -9,   7,  -3,   1,  -6,  15,  -6,   0,   0,  -1,   2,   0 },
      {  -2,  -8, -12,   9,  12,   2, -10,   6,   0,   0,   2,  16,   2, -18,  21, -14,  -1,   1, -19, -10,   8,   9, -16,  10,  -4, -10, -29,   1,  -3, -11,  11,  -2,  -8,  -5, -13,  10, -23,  10,  -2,   2, -17, -27, -58,  37, -14,   8,   0,  -2,   3, -27,  19, -22,   8,  -3,   4,  -2,  17, -57,  14,   1,  -2,   0,  -2,   1 },
      {   0,   4,  10,  -8, -21,  15, -15,  11,  -1,  -2,   8,   3, -13,  20, -29,  11,   4,  13,  29,  39, -17,  -4,  13, -10,  -4, -14, -23,  12,   1,  -3,  -3,   0,  -3, -12, -14,  26,  -1,  -5,   3,  -2, -14, -37, -45,  30, -15,  12,  -2,   4, -25,   4, -24,   6,  -2,   3,  -1,  -1, -10,  46, -15,  -2,  -1,  -2,   2,  -3 },
      {  -1, -15, -12,  22,   5,   0,   8,   0,   3,  13,  26,  21,  12, -12,  18, -13,  -4,  -2, -33, -34,  10,  17, -32,  15,   0,   1,   5,   0,   7, -24,  10,  -9,   4,  19,  24,  14, -30,  17,  -9,   5,   6, -14,  -3,   5,   3,  -3,  -1,   3, -10,  -6, -24,  19, -10,   5,  -2,   1,  -8,  65, -18,   0,   0,  -1,  -1,   0 },
      {  -5,  -9, -22, -36,  -6,  12, -19,  10,   2,   8,   9,  -5,   5,  -3,  -6,   2,   1,   6,   1,   3,  -9, -17,  23, -12, -10, -35, -64, -16,  34, -27,   8,  -4,   0,  17,  37,  18, -24,  10,  -5,   0,  15,  27,  27,  -3,   8,   3,  -2,  -1,   0,   1,   2,  10,  -1,   1,   0,  -2,   7, -13,   3,  -3,   1,   1,   0,  -1 },
      {   3,  -4,   4,  39,  18, -18,   2,   0,   4,   1,  18,  39,  -5,  -5,  11, -11,   7,   4,  25,  51,  12,  -6,  13,  -6,   0,  -6,   2,   6, -15,  36, -34,  16,   7,  27,  47,  17, -19,   7,   2,  -3,   2,   9,  -9, -12,  23, -13,   4,  -2,  -1, -19,  -8, -10,   5,  -3,   4,   0,   2,  -7,   8,  -2,  -2,   1,   0,   1 },
      {   5,   8,  20,  40,  35,  -5,  10,  -7,  -1,  -9, -11,  10,  11, -10,  16, -10,  -3,   1,  -3,   6,  24, -23,   8, -10,  -4,  -8, -46, -56,  29,   3, -10,   6,  -8, -14, -30, -27,  18, -16,   7,  -2,   3,   4,  11,  12, -14,   7,  -2,   7,  -3,  -5,  -1, -11,  -1,   7,  -3,   4, -10,  29,  -3,   1,  -4,   1,  -1,   1 },
      {  -2,  -4,  -5,  -4,  21,  12, -20,  26,  -5,  -5, -15, -46, -28,   9,  18, -12,   3,   6,  14,   2,   8,  10, -13,  10,  -2,  -4,  -7,  -4,   6,   7,   0,   2,   0,   5,   9,  -9, -16,   6,  -6,   1,  -2,  -4, -26, -26,  23, -16,   1,   1,   9,  21,  58, -32,  16, -10,   2,  -2,   3,  42,  -8,  15,  -5,   4,  -3,   3 },
      {   1,   2,   7,  16,  16,  17, -13,  11,  -3,  -3, -15, -38, -35,  22,  14,   1,   2,   6,   2,  -2,   7,  33, -24,   4,   0,  -4,  -4, -14, -21,  25, -10,   5,   2,  -1,  18,  34, -11,  -7,   0,  -1,   3,  11,  40,  32, -42,  15, -15,  10,   0, -18, -14,  19,  -3,   4,   0,   2, -10, -17, -16,  13,   0,  -5,  -2,  -1 },
      {  -1,  -3,  -8,   5,  24, -16, -26,  17,  -2,   2, -10, -35, -28,  -8,  19, -17,   1,  -3,   8,  32,  26,  -4,   1,   2,   3,   4,  12,  34,  30, -31,  19,  -7,  -2,   0, -17, -29,  -6,   0,   7,  -3,   0,   4,  -5, -12,  14,  -3,   1,   2,  -6, -22, -54,  23,  -7,   5,  -2,   1,  -5, -10,  12, -16,   3,   0,  -2,  -1 },
      {  -1,  -4, -11, -12, -10, -26, -10,  10,   5,   7,  11,  25,  21,  20,  21, -11,   1,  -3,   5,  29,  21,  -2,   0,  -9,   4,   2,  16,  39,   9, -15,  -2,   2,  -4, -16, -11, -14, -28,  10,   9,   1,  -2,  13,  38,  21, -48,  15, -18,   4,   5,  14,  30, -29,  20,  -3,  -1,   2,   3,  16,  -9,  -4,   6,  -7,   1,  -3 },
      {  -4,  -2,  -9, -43, -55,   4,  35, -20,   2,   5,   9,  -2, -10,   3,  17,   4,   4,   8,  12,  26,  51,  27, -22,   9,  -1,   3, -13, -39, -12,  11,  -3,   0,  -2,   3,  -8, -32, -18,   0,  -6,  -1,   2,   3,  -6,   3,   8, -17,   4,  -2,  -1,  -8, -19,  -1,   7,   0,  -3,   5,  -3,  -1,   2,   6,  -4,  -1,   0,   0 },
      {  -4,  -6,  -1,  -4, -38, -43,   4, -11,  -3,   1,   4,  -8, -29, -15,  26, -23,   0,   0,  -6, -16,   5,   6,   3,  16,   3,  -1,  -5,   4,  23,  -4,  -4,   9,   2,  -4,   8,  55,  39, -41,  33, -16,  -2,   4,   0,  -1,   6,   7,  -3,  -1,  -4,   1,   0, -24,  15,  -2,  -1,   0,   3,  12,  16, -15,   5,  -3,   1,  -3 },
      {   0,  -3,  -5,  -5,   2,  11,   5,  -3,   3,   6,   6,   7,  22,  30,  -4,   3,   0,  -2, -11,  -6,  21,  14, -16,  -3,   1,  -1,  -5,  26,  68,  26, -32,  11,   2,   2,   2,   2,   3, -28,  13,  -3,  -6,  -7, -24, -32, -12,  -6,  -5,   2,   3,   5,   7,  24, -17,  11,  -4,   3,  -7, -25, -35,  21,  -3,   2,  -1,   3 },
      {  -3,  -6,  -2,  -2, -22, -21,   1,  11,  -2,   0,   5,  -6, -29,  -7,   4, -12,  -4,  -2,  -5, -32, -45,   5,  18,   0,   0,   0,  -2,   1,  35,  47, -33,  27,  -2,  -3, -12, -37, -26,  34,  -7,  -1,  -1,  -4,   2,  20, -11,  -8,   1,   0,  -3,  -5,  -9,  12,   0,  -8,   3,  -3,   5,  10,  26,  -7,  12,  -2,  -2,   2 },
      {  -1,   0,   1,   2, -18, -16,  23,   3,  -1,   0,   0,   1, -30, -19,  36, -15,  -1,  -1,  -2,   5, -20, -21,  18, -13,  -3,   0,  -4,   4,  -9, -16,  -4,  -3,   2,   1,  -4,  -5,  21,  32, -10,   8,   1,  -5,  -3,  -3,   8,  12,  -7,   5,   2,   1,   2,   8,   8,   2,  -1,  -1, -17, -20, -82,  30,  -7,   1,  -3,   0 },
      {   2,   6,   2,  -7,  19,  64,  41, -35,   1,   0,  -2,  -4, -22, -23,  12,  15,   0,  -2,  -3,   3,   1,  -4,   0, -18,   3,   1,   8,  36,  35,  -2,  -9,  -3,   1,   1,   8,  11,   9,   6, -26,   8,   2,  -1,  10,  13,  -7,  -1,   1,   4,   0,  -4, -16, -37,  28, -15,   7,   3,   2,   7,  13,  -4,  -1,  10,  -8,   3 },
      {  -1,  -1,  -8,  -8,   9, -10,  -8,  22,   1,   4,   1,   2,  13,  10,  -2,  14,  -3,  -2,  -8, -28, -19,  17,  18, -28,  -1,   1,   2, -12, -16,   1,   4,   5,  -5,   2,   1, -13, -24, -33,  14,  -4,   1,   6, -12, -33, -12,   5,  -6,   3,  -3, -19, -52, -47,  41, -16,   6,   1,  -7,   0, -16,  19, -13,   6,  -9,   5 },
      {   0,   0,   0,   0,  -6,  -3,  18,  -1,  -3,  -3,  -3,   0,  -9, -18,  -9,  -8,   0,   1,   1,   9,   4, -17,  -7,  11,  -1,  -2,  -6,  -6,  -2,  24,  28, -16,  -1,   7,  10,   1,   4,  19, -11,   8,  -4,  -2, -19, -66, -53,  47, -39,  14,  -2,  -5,   3,  14,   4,  -4,   5,  -8,   4,   6,  24,  -5,   1,  -5,   5,   0 },
      {  -1,   0,   2,   6,   3, -23, -33,  -9,   0,   1,   4,   6,  -6, -55, -42,  24,   4,   5,   2,   3,  30,  56,  29,   3,   3,   3,  -2,   3,  22,  23,  32,  -3,   0,  -1,   0,   2,  13,  14, -15,  12,   2,   1,   9,   7,  -6,  -6,   8,  -6,   1,   0,   7, -11,   2,  -3,   1,  -2,  -5,   2, -25,  12,   5,   0,   3,   3 },
      {   0,   1,   5,   5,   2,  13,  -4, -45,  -1,  -3,   2,   1, -28, -36,   8,  35,  -1,   0,   3,   0, -16, -18,   7,  12,   0,  -2,   3,   5,  -9,  -1,  10,  -1,  -3,   0,   2, -16, -57, -29,  57, -21,  -2,   0,  -1,  -6,  -9,   0,   3,   9,  -2,   7,  17,   5, -19,  17,  -1,   2,  -2,   1, -16,   0,  11,  -1,  -1,  -1 },
      {   1,   0,   2,   7,   6,  -1,   0, -15,   2,   4,   6,   2,  -9,  10,  36,  21,   1,   3,  -2,  -8,   8,  49,  48, -39,  -2,  -1,   1,  -4, -13, -28, -26,  -6,   0,   2,  -3,  -6,  -3, -23, -23,  14,  -1,  -9, -14,  -9,  11,  20, -33,  -4,   1,   7,  22,  33,  -4,  -3,   3,  -1,   5,   4,  17, -21,  14,  -8,   6,  -6 },
      {  -1,  -1,  -2,  -5,  10,   6, -45, -60,   3,   4,   3,  -1,  15,  43,  31,   5,   0,  -1,   0,  -2,  -9,  -6,  22,  55,  -1,   2,  -2, -11,   1,  12,   5,  -2,   0,   1,   1,  -2,  15,  22, -14,  -1,   1,  -2,  -7, -10,  -7,  -5,   4,   0,  -4,  -4, -17,  -9,  19, -11,   7,   0,  -1,  -2, -20, -21,  20, -11,   4,  -3 },
      {  -2,  -2,  -3,  -5,  -8, -21, -47, -33,  -1,  -1,  -3,  -5,  -9, -12,  -1,  32,  -2,  -1,  -3,  -3,   3,  -7, -42, -31,  -2,  -3,  -1,  -3,  -9, -29, -49,  15,   1,   0,   0,   1,  18,  34,  -7,  -2,   0,   1,  -4, -19, -15,   7,   4,   5,  -3,  -2,  -6, -22, -23,  17,   2,   4,  -1,   4,  19,  21, -19,   9,  -3,  -2 },
      {   0,   0,  -1,  -1,   6,  14,   0,  -5,  -1,   2,   1,  -5,   3,  -6, -43, -11,   0,   2,  -1,  -4,  12,   9, -12,  27,  -1,  -1,   2,  -2,   1, -10, -39,  11,  -1,   0,  -5, -14,  -6,  16,  32,   6,   2,   3,   6,  11,  50,  65, -38,   2,   3,  -1,  -7, -12,  17, -12,   6,  -9,   0,  -5, -15, -10,  10,  -5,   8,  -6 },
      {   1,   0,   0,   2,   7,   9,  24,  40,  -1,  -2,  -3,  -5,  -1, -11,  -5,  24,   1,   3,   0,  -3,  12,  14,   8,  13,  -2,  -2,  -3, -10,  -6, -27, -38,  -6,   2,   2,   4,   4,  13,  32,  30, -14,  -2,  -2,  -5, -23, -38, -18,  41,  -9,  -1,   0,   1,  -3,   9,  -6,  -7,  -2,   1,  -5, -15, -45,  34, -13,   6,  -7 },
      {   0,  -1,  -1,   1,  -4, -17,  -1,  38,   2,   3,   2,   4,   4,   0,  32,  73,  -1,   0,   1,   0,  -6, -17, -17,  18,   1,   1,   0,   1,  17,  32,  -6, -56,   0,  -1,  -3,   3,   2,  -3, -14,   5,   1,   3,   3,  13,  27,  20, -13,  -1,  -1,  -1,  -5, -18,  -5,  20, -14,   1,  -1,  -1,   0, -10,   2,  -5,   0,  -2 },
      {  -1,  -1,   0,   3,   0, -16, -18, -12,  -2,  -3,  -2,   1,  -1, -12,  -5,  15,  -1,  -2,   0,   3,   2, -22, -38, -20,   0,  -2,  -1,  -2,   3,   6,  -5,  -3,   0,  -1,   1,  -1,  -4, -18,   0,  21,   1,  -3,  -1,   5,   1,   8,  31, -15,   4,   1,  12,  56,  82, -24,  -3,   5,   0,   3,  -2,  -2,   9,   1,  -1,  -2 },
      {  -1,  -2,  -2,   0,  -6, -11, -10, -10,  -2,  -1,  -3,  -8, -10, -16, -26, -33,  -1,  -1,   1,  -1,  -7, -16, -22,  -7,  -1,   0,   0,  -5,  -4,   1, -24, -22,  -2,  -2,   2,  -5, -22, -45, -52,  23,  -2,   0,   1,  -7, -10, -13,  -4,  -6,  -1,   4,  -4, -27, -23,  -4,  -4,   1,  -2, -10, -33, -50,  19, -11,   4, -11 },
      {   0,   0,   0,   1,  -1,   0,   4,   7,   0,   1,   3,   2,  -6,  -9,  -5,  -6,   1,   2,   3,   5,  -2,  -3,  30,  64,   0,   0,   3,   6,  -3, -26, -39, -15,  -2,  -3,  -3,  -2, -12, -28, -23,  27,   0,   3,   6,  -6, -20,  20,  33, -28,   1,  -3,  -1,   5,  -1,   3,  -5,   2,   0,   4,  20,  51, -20,   8,  -3,   5 },
      {   0,   1,   2,   2,   1,  -6,   5,  22,   0,   1,   1,   3,  -1, -16,  -3,  40,  -1,   1,  -1,   3,   5, -19,  -7,  38,   0,   0,  -1,   2,  -3, -24,   3,  76,  -1,  -1,   1,   0,   2, -19, -41, -21,   0,  -2,   0,   1,  -4, -25, -44,   9,   0,   2,   1,  10,  13,  -4,   8,  -1,   0,  -3,  -7,  -7,   2,   3,  -1,   9 },
      {   1,   1,   2,   0,  -1,   8,   6,   4,   0,   1,   3,  -2,  -4,  10,  16,   9,   0,   0,   1,   1,  -4,  -3,  -8,  -8,   0,   0,  -2,  -1,   1,   5,  28,  57,   0,   1,   2,   2,  -1,   2,  15,  77,   0,   0,  -1,  -9,   3,  18,  13, -45,  -1,   0,  -6, -18, -13,  32, -23,  -1,   1,   1,   0,  -8,  19, -14,  11,   2 },
      {   0,   1,   2,   1,  -3,  -7,   2,   4,   1,   2,   2,   0,  -1,   5,  16,  19,   0,   1,   1,   4,   5,  -5, -11,  -5,   1,   0,  -2,   1,   5,   4,  13,  14,   0,   0,   0,  -1,  -2,  -6,  -1, -10,   1,   2,   3,   1,   0,  34,  29, -27,  -1,  -2,   3,  -1, -50, -88,  38,  -7,  -1,   2, -10,  -9, -12,   5,  -2,  12 },
      {   0,   0,  -1,   0,   2,  -3,  -3,  -3,   0,   0,  -1,  -2,   0,  -5,  -2,   6,   0,   1,   1,   1,  -2,  -9,  -1,   7,   0,  -2,  -3,  -3,  -1,  -5, -13, -26,   1,   1,  -1,   4,   9,   8,  42,  51,  -1,  -1,  -1,   8, -12, -60, -61, -24,  -1,  -2,  -5,   6,  -7, -44,  -4,  -8,   0,   1,   4,   9, -10,  15,   6,   4 },
      {  -1,  -1,  -1,  -1,  -5,  -4,  -2,   2,   0,  -1,   0,   2,   1,   2,   1,   7,   0,   1,   1,  -1,   0,   7,   9,  13,   0,   0,   1,   0,  -1,  -1,  -4,  10,   1,   2,   1,   0,  -1,  -1,  11,  49,   0,   0,   1,   1,   0,   5,  22,  76,   1,   1,   8,   5,  -1,  -6, -14,   1,  -1,  -3,  -5, -45, -54,  23, -24,  25 },
      {   0,   0,  -1,  -3,  -1,  -4,   1,   5,   1,   2,   1,   1,   5,  -1,   2,   5,   0,   0,   1,   2,   2,  -9,  -4,  -3,   0,   0,   0,   0,   0,  -7,  -1,   1,   1,   1,   0,   3,   5, -11,   0,  29,   1,   0,  -2,   2,   8,  -9,  14,  70,   0,  -5,  -4,  -6, -13, -24,  11, -18,   2,   0,  13,  47,  65, -42,   6,  -4 },
      {   0,  -1,   0,   1,   3,   6,   1,  -2,  -1,  -1,  -1,   0,  -3,  -6,   0,   5,   0,  -1,  -1,  -3,  -4,   0,   3,   1,   0,   0,   1,   4,   5,   0,   2,   8,  -1,   0,   1,  -2,  -3,   1,  -1,  -6,   0,  -1,   0,  -1,   1,  -5,   2,   8,   0,   0,   0,   0,   1, -24, -34,  21,   0,  -1,   2,   2, -59, -72,  50, -51 },
      {   0,   0,   0,   1,   1,   0,  -3,  -6,   0,   0,   1,   1,   0,   1,  -1,  -4,   0,   1,   0,   1,  -2,   1,   0,  -2,   0,  -1,  -2,   0,  -4,  -1,   2,   5,  -1,   0,   1,   2,  -1,   4,  -8, -21,   0,  -1,  -2,  -3,   3,  12,  -5,   4,  -2,   0,   0,  -3, -10, -38,-100,  39,   1,  -2,   3,  10,  31,  25,  -6,  26 },
      {   0,  -1,  -1,  -1,   0,   2,  -2,  -5,   0,   0,   0,  -1,  -2,  -3,  -5,  -3,   0,   0,   0,   0,  -1,  -1,  -2,  -5,   0,  -1,   0,   1,  -1,  -4,  -7, -10,   0,  -1,   0,   1,   1,   2,   2,  -7,   0,   0,   0,   1,   1,  -4,  -7, -13,   0,   1,   1,   1,   5,   8,   7,  11,  -1,   1,   1,  -3, -12, -72,  -6, 101 },
      {   0,   0,   0,  -1,  -1,  -3,   2,   8,   0,  -1,  -1,  -1,   1,  -2,  -4,   0,   0,   1,   0,   0,  -2,   0,   2,   6,   0,  -1,  -1,   0,   0,   0,   2,  -1,   0,   0,   1,   2,   2,   4,   7,  17,   0,  -2,  -2,  -2,   2,   0,  -6,   5,   0,   2,   1,   0,  -4,   2,  44, 116,   1,   0,   2,   1,  13,   2, -10, -11 },
      {   0,   0,   0,   0,   0,   0,   0,  -2,   0,   0,  -1,   1,   2,  -1,  -4,   1,   0,   0,   0,   1,   2,   0,   2,   3,   0,   0,  -1,   0,   0,   0,   2,   4,   1,   1,   1,   0,   3,   3,   3,   0,   0,  -1,  -1,  -1,   1,  -1,  -9, -15,   0,  -1,   3,   2,  -4,  -7, -14,  -7,  -2,   1,   2,   2,  -4, -42,-112, -37 },
    },
    {
      {  80,  -1, -22, -12,  -3,  -3,   1,   1, -31,  -9,   9,   5,  -4,   1,  -1,   0, -63,  38,   0,  -6,   0,   0,   0,   1, -25,   7,   3,  -4,   1,  -1,   2,   0,  35, -20,   0,  -2,   0,   0,  -1,   0,   3, -10,   4,   4,  -1,   0,   0,   0,  -9,   7,  -2,   0,   0,   0,   0,   0,  -6,   2,   2,   0,   1,   0,   0,   0 },
      {  24,  29, -34,  -4,   5,  -2,  -2,  -1,  67, -37,   6,   0,  -8,   2,   0,   1,  25,  27, -12,  -6,  -2,  -1,   0,  -1, -39,  12, -11,   3,   4,  -1,   1,   0, -47,  14,  23,  -6,   1,   0,   1,   0,   2, -21,   8,   5,   1,   0,   0,   0,  19,   8,  -9,   0,  -1,   0,  -1,   1, -16,   9,   6,  -1,   0,  -1,   0,   0 },
      { -39, -16,  24,   9,   1,   0,   1,   1,  45,  19, -30,  -3,  -2,   0,  -2,   0, -38,   5,  17,   2,  -2,  -1,   1,   0, -80, -12,  28,   1,   9,   2,   3,   0,  24, -17,  -3,  -1,  -1,   0,  -1,   0,   7,   0,  -6,   0,  -2,   1,   0,   0,  11,  -3,  -2,   0,  -3,   0,   0,   0,  12,  -4,  -7,   3,   0,   0,   0,   0 },
      { -28,  38, -21,  -6, -26,   9,  -1,   6,   7,  17,  -6,   2,  12,  -7,   1,  -4, -64, -12,  19,  11,   4,   0,   1,   0,  40,  -3,  -5, -25,   5,  -4,   2,   1, -26, -22,  28,   0,  -3,   4,  -1,   0, -19, -26,  -4,  12,   2,   2,   0,   1,  37,  22, -16,   0,  -1,  -2,  -2,   0,   2,   3,   0,   0,  -1,   0,   0,  -1 },
      {  70, -20,  29,   1, -14,   9,  -2,   3,  -1,  35, -42,   3,  13,  -3,   0,  -1,  14, -22,  12,  17,   3,  -5,   2,  -3,  -4, -38,  19,  -8,  10,   3,   0,   0, -48,  18,  12,   4,  -4,   1,  -1,   0, -13,   1,  -3,  -1,  -1,   2,  -1,   0,  16, -10,  -3,   4,  -1,  -2,   1,   0,  13,  -4,  -8,   2,   1,   1,   0,   1 },
      {   7,  23,  -7,  31,  27, -19,  -5, -13, -49,  -2,  17, -14, -12,  10,   1,   6,  -9, -65,  11,  -3,  -1,  10,  -2,   5, -40,  26,  -7,  24,   9,  -5,   3,  -4, -20,   3,  11, -14,   2,  -1,   1,   1,  21,   2,  -6,   2,   4,  -3,   0,  -1,  35,   0,  -8,  -6,  -3,   0,  -1,   0,   7,  -3,  -2,   3,  -1,   0,   0,  -1 },
      {  18,  21,  10,  13,   9, -21,  -9,  -9,  42, -18,   5, -21, -15,  10,   1,  -1, -27, -28,  -8,  11,  16,   6,   0,   6,  25, -35,  -7,  17,   3,   4,   0,  -1,  50,  50,  -4, -18,  -9,  -5,   0,   2, -33,  21,  -4,   2,   3,  -2,   4,   1,  11,  -6,  -2,   0,  -1,   2,   0,   0, -16,   0,   4,   3,   0,   1,  -1,   0 },
      {  24, -59,   2,  30,  15,   3,   2,  -5,  55,  14,   8, -12,  -4,  -1,   4,   2, -26, -20,   3, -19, -10,   8,  -2,   1,  37,  52, -11,  13,  -4,  -8,   2,  -2, -16, -28,  -5,   3,   5,   1,   0,   0,   8,  16,  -5,  -6,   2,  -1,  -3,  -1,  -7,  11,   2,  -5,   0,   0,   2,   1,   3,  -3,   2,  -2,  -4,   0,   0,   0 },
      {  21,  22, -31, -37, -23,  -1, -15,   4,  30,  10,  18,  10,  17,  -9,  -3,  -3,  26, -52,  22,  25,  -1,   4,   3,   3,  -5,   9,  25, -24,  -9,   4,   0,   0,  27, -14, -34,   0,  11,  -1,   2,   2,  35,   4, -26,   3,  -2,  -6,   6,  -1, -11,  17,   6,   0,  -1,   0,  -1,  -3,   1,   1,   2,   1,   0,   0,   0,   0 },
      { -11,  19,   6, -28,   5,  -5,   3,  -4,   6,  14,  10,   9,   1,   5,  -5,   1, -54,  -5,   5,   1,   5,   1,   0,   5,  -6,   1,  -7,   5, -13,   2,   2,   0, -52,  44, -21,   6,  12,  -1,   0,  -2,  28,  10,  13,  -9,   2,  -6,   2,   0, -53, -32,  22,   1,   2,  -1,   3,  -1, -14, -13,   1,  -2,   1,   1,   0,  -1 },
      {   3,  17, -17,  -3,   1,   3,   4,   1,   1,   4,  -2,   7,   5,  -4,  -4,   1,   9, -13,  27,  -2, -13,   1,  -2,  -2, -20,  44,  -1,  -6,   4,  -5,   1,  -1,   5,   2,  -9,   2,   8,   2,   0,   1, -95,   2,  18,   0,   3,   1,   2,   1, -27, -24,  10,   0,  -2,   0,   1,   0,  30, -11,  -8,   2,  -2,   0,   0,  -1 },
      {  -5, -40, -23, -30,  15,  10,  23, -11,  -3,  12,  -2,  16,   9,  -3, -10,   2,   0,   0,  38,  -9, -22, -11,  -3,   7,   1,   1, -27,   0, -10,   4,   5,  -1,  18,  39, -27,   2,  14,   2,  -4,  -1,   4, -19,  23,  -5,  -2,   4,  -2,   2,  63, -16,   4,   3,  -3,  -1,  -1,   0, -19,  18,   5,  -2,  -1,   1,  -2,   0 },
      { -11, -32, -27,  -6, -10, -31, -12,  -4,  -9,  65,  36, -27, -15,   6,  -5,  -5,  11,  14, -11,   7,  21,  13,   1,   7, -19, -10,  12,   3, -12,   0,   2,  -1, -16,   2,  -2, -13,  -1,  -7,   5,   4, -34,  -6,  -9,  13,   4,  -3,  10,   0,  -4,  22,  -4,  -1,  -1,   2,  -2,   1, -44,  -6,  13,  -1,   0,  -1,   0,  -1 },
      {   6,  32, -45,  28,   9,  28,  17,  -3,   7,  60, -18,  -1,  -1,  -6,   3,   2,  14,   5,  12, -15, -12,  -7,  -3,  -5,  -4, -21, -38,   9,   6,  -5,   1,   1,  24,   2,  31,   6,  -1,   3,  -5,  -2,  22,  36,  13,   0,   0,   5,  -8,  -2, -23,   6,   0,   0,  -1,  -1,   2,   0,  -3,  -9,   0,   1,  -1,  -1,  -1,   0 },
      { -11, -34, -61,  25,  -4,   9, -10,  12,  -5, -38,  26,   0,   5,  -8,   5,   4, -12,   3,  25,  -8,  -6,   3,   5,  -6,  -1, -59,  12,  13,   4,   4,  -1,   2, -24,   4, -15,  -1,   0,  -1,   6,   1,  -5,  17, -18,  -2,  -1,   1,   0,  -2,  -6,   2,  12,  -2,  -2,   1,  -1,   0,  39,  -7, -12,   5,  -1,   0,   0,   0 },
      {  -1,  -4, -11,  28,   9, -24, -14,  12,  -3,  -6, -36, -27,  -9,  10,   5,  -3,  10,   9,  57,  -3,   3,   1,  10, -10,  19,  13,  17,  -3,  13,   3, -11,   4,  12,  28,   5,   5,   2,  -3,   1,  -3,  21, -64,  -1,  13,   7,   3,   6,   0, -33,   3,  -2,  -8,   0,   1,  -1,  -1, -13, -14,   0,   3,   3,   1,  -1,   1 },
      {  -8, -22,   6,  -1,  -4,  25,   5,   3, -15, -40, -23,  18,  15,  -7,   2,   0,  -6, -28,  13,   6, -25,  -3,  -3,   2, -18,  -1,  -8,  -9,  12,   0,   3,  -2,  -6,   6,  13,  -7,  -5,   4,  -2,   2, -18,  25,  -5,   4,  -3,  -1,  -3,   1, -26,  49,  -7,   3,   0,  -2,   2,  -2, -71, -14,  13,  -3,  -3,  -1,  -1,  -1 },
      {  -9, -39, -21, -31,  26,   8,  10,  -4,   1,   3,   2,   5,  -5,  -5,  -1,   0,  -8, -24, -44,  34,  -9,  -5,  -5,   4,  -4,   5,  -5, -16,  10,  -2,   3,  -2,  15,  29,  52, -19,  -4,   2,   2,   2,   7, -29,  -4,  15,   1,   3,   2,   1, -28,   2, -14,   3,  -3,   2,   0,   0,  40,   0,  -8,  -2,   0,  -3,   1,   1 },
      {   7,  -9,  35,  -9,  -5,   1,   9,   3,  12,  -9,  52, -17,   6,  -2,   2,  -4,   8, -20,  36,   9,  11,  -9,   0,   0, -15, -36, -60,   2,   3,   2,  -1,   3,   3, -40,  16,  10, -11,   5,  -3,  -2,  -2, -35,  16,   0,  -2,   4,   1,   1, -21, -16,  -2,  -1,   4,  -3,   0,   1,  -4,  -4,   3,  -1,  -1,   1,   0,  -1 },
      {  -2, -10,  -1,  37, -68,  22,   2,   5,  -7,  -7,   8, -51,  49,  -5,   3,  -7,  -5,   2, -18,  16,   7,   0,  -6,  -1, -19,  27, -15, -10, -15,  -4,   6,   3,   9,  36,   0,   5,   2,  -2,  -2,  -1,  11,   1,   7,   5,  -4,   5,  -4,   1,   3,  -9,   8,   3,   6,  -1,   4,   1,   5,  12,   1,  -4,   0,   0,  -2,   3 },
      {   1,   7, -11,  12,  21,   5,  29, -21,   0, -10,  -1, -34,  11,  -1,  -9,  12,   3,  15,   9,  53, -27, -11, -17,   9,  16,  -4,  29,  -2,  -7,   9,  12,  -4,  -7, -35,   6, -36,  20,  -1,   1,  -2,   3,   9,  -6,   9,  -7,  -2,  -4,   1,   0, -50,   2,  11,  -2,  -1,   2,  -4, -29,  -5,  -9,  -2,   1,   2,  -3,  -2 },
      {  -1,  -1,  22, -16,   5, -29,   4, -15,  -3,  -5,  -3, -12,  -3,  13,  -5,   2,   4,  38,  47,  30,   4,   5, -10,   3,   0,  14, -32, -18,  -3,   7,   6,   2, -11,  13,  -2, -28,  -2,  -6,   2,   8,   8,  47,  -3,  21,   3,  -1,   5,   1,   0,  45,  -4,   4,   3,  -5,  -4,  -3,  41,   8,   2,   2,   2,   3,  -3,   1 },
      {  -4, -20,  -8, -22, -31, -20,   4, -10,   2, -16,  12,  13,  -4,   6,  -3,   3,   6,  -1,  25, -26,  35,  -3,   5,   1,   5,  17,  23, -14,   7,  -4,  -2,   5,  10,   3,  59,  10,  -8,  -1,  -4,   0,  12,  40,   8,  11,   0,   0,   1,  -4,  16, -40,  -3,  -3,  -1,  -1,  -2,   5, -12, -40, -10,   6,   5,   0,   0,   1 },
      {   3,  11,  21, -14, -20,   8,  25, -25,   3,   2,  17, -11,  12, -11, -10,   9,   4,  -7,  13, -64, -18,   5,  -4,   7,   7, -12,  36,  30, -18,  -1,  21,  -6,   1,  11,  12, -37,  17,  -3,  -8,   2,   0, -15,   6,  17, -14,  -1,  -8,  -1, -19,  20, -28,  12,  -1,  -3,   3,   0,  16,  14,   3, -11,   1,   1,  -4,   1 },
      {  -5, -14, -29, -14, -18, -26,  -4,   1,  -4, -15, -54, -15, -16,  29,  -1,  -4,  -1, -29,  -8, -17,  21,  -2,   4,  -1,  -3, -13, -17,  -9, -17,  -9,   2,   3,  -7, -32,  -7, -16,   2,  -2,  -4,  -1,  -1,  12,  28,  10,  -2,  -2,  -1,   1, -19, -20,  -5,   6,   0,   2,   3,   1,  -2,  60,   7,  -7,  -1,   0,  -1,   3 },
      {  -2,  -1,   0,  -7,  -1,  19,  -2,  26,  -3,   7,   9,  18,  -5,  -4,   9, -16,  -2,   8,  30,  21,  27, -27,  -4,  -3,   1,  24,  12,  56,  -4,  13,   0,   3,   0,   9,  24,  -4, -19,  15,   0,  -5,  -5,   9, -40, -19,  -4,   9,  -2,   5, -10, -10,  -8,   2,  -1,   3,   0,   1,  -8,  61,  -2,   0,   2,  -3,  -5,  -4 },
      {   1,   3,   7,  29,  25, -15,  26, -16,   2,   3,  33,   6,   6,   6, -12,   8,  -3,  -3,   2,  -8,  -4, -15,   0,  -3,  -2,  -6,  26, -53,  19,  -8,  -2,   6,  -3,  14,  -1,  52,  -4,  10,  -3,  -2,  -4,  12,  -2,  28,  -9,   5,  -2,  -1, -12,   3, -10,  -8,   8,  -1,   0,   3, -10,  57, -11,  -1,   1,  -2,   4,  -1 },
      {   1,  -3,  12,  30,  -9,  34, -21,  -6,   0,   5,  11,  41, -28,   8,  27,  12,   3,   4,  10, -14,   8,  13, -29,   2,  -4,  -2,  -9, -42, -11, -12,   0,  -4,   3,   7,  -1, -44,   5,   0,   5,  -4,   1, -12, -28,  22,  32, -11,  -3,  -1,  -1, -28,  13,   8,   9,  -6,  -3,  -3,  -3,  11,  22, -10,  -8,   4,  -3,   4 },
      {   2,   1, -10,   8,   5,  24,  39, -24,   1,  -7, -13,  -1, -10,  19, -23,  11,   2,  -7,  -5,   1,  59, -49,  -9,  14,  -2,   3,   4,   4, -20, -19,  13,  -1,  -4,  -2, -28,  -1, -25,  31,  -1,  -5,  -3, -16, -10,  11,   0,   9,  -2,   1,   1,  23,  14,  -1,   6,  -6,   0,   2,   7, -32,  -1,   0,   2,   0,   1,   1 },
      {   1,   4,   1, -38,  21,   9,  -4, -11,  -6,  -5, -22, -51,   5, -13,   6,  10,  -2,   1,  -4, -35,  -6,   3,  -5,  -6,  -6,  -1, -20,  -9, -13,  20,   2,   4,  -3,  -2,   9,  30,   2,  -2,  -2,  -2, -11,   4, -75,  16,   9,   5,  -3,  -6,   4, -14,  18, -18,   7,  -4,  -2,   4,   1,  -2,  14,  -3,  -4,   1,   0,   4 },
      {   0,   2,   3, -20,  28,  56, -25,   8,  -3,   0,  16, -52, -25,   2,  26,  -3,  -1,  -3,  11, -12,  13, -16,   9,  -3,   0,   4,  30, -32,   4,  16,  -6,  -4,  -3,   3,   1,  -9, -16,  -2,  -9,   4,   2,  13,  44, -26,   2,  -6,   9,  -4,   7,  12,   4,  16,   0,  -2,   6,   0,   2,   7,   3,   8,  -6,  -1,   0,   1 },
      {  -3,   1,   0, -21, -23,  12, -17,   1,   0,   6,   6, -17, -17,  31, -18,  -3,   2,   7,   0,  -5, -32,  -6,   1,   8,   7,   4,  -5,  13,  74, -51,  25,  -4,  -2,   2, -17,  -3, -25,   7,  -7,   5,   6,   1, -19,  10, -12,   1,  -1,   0,   1, -14,   8,  21,   5,   5,   1,  -8,  -1,   1,   4,   3,  -2,  -6,   8,  -7 },
      {   0,  -1,  -2,  17,  -7,  17, -18, -13,  -2,   1,  -2,   8, -22, -19, -26,   0,  -2,   3,  -3,   0,   9, -21,  56,   6,  -2,  11, -23, -13,   5,  43,   0,  -8,  -4,   2, -23,  -4,  -2, -18,  -4,  13,   1,   4, -18,  18, -30,  -2,   3,   0,  -5, -22, -57,  17,  -9,   7,   6,  -1,   2, -11,   5, -13,   4,   3,   0,  -5 },
      {  -1,   1,  -2,  -2,  41,  -4, -51, -14,   3,   6,   1,  11,  82,  24,  12,   0,   0,   9,   1, -10,   5,  -9,  27,   0,   0,  -1,   0,   3, -19, -24,   3,  -6,   2,  -4,   2, -18, -32,  12,  -2,   4,  -1,   2,   3,  15,  -4,   0,   0,   0,   0, -11, -10,  -1,   1,  -6,  -1,  -3,  -2,  -3,  -1,  -8, -10,  -3,   4,  -1 },
      {   0,   0,   1,   6,  -1,  -1,   1,  16,   1,  -2,   9,   1,  -1,  72,  -8,  -6,  -1,   1,   3,   3, -17, -20,   0,  -8,  -2,  -7,   2, -16, -29, -10,  17,   2,   3,   5,  16,  18,  40,  -5, -16,   0,  -6,   1, -23, -49,   4,   4,   3,   3,   0,   2, -35,   7,  -6,   6,   9,   0,  10, -16,  31,  -3,  -8,  -6,   0,   3 },
      {   0,  -3,   4, -11,   8,  17, -24,   1,  -2,  -6,  -4, -13, -19, -26,  12,  11,   1,   5,   7,  13,  18,  24,  -9, -15,  -2,  -3,  -7,   5, -18, -56,   9,   7,   2,   5, -11,  19,  41,  35,  16,  -2,   0,   8,   1,   7,  -5,  15,  -2,  -2,   0,  -3, -51,   5,  -6, -17,   0,   1,  -9,  -4, -37,   0,   8,  -7,   9,   3 },
      {  -2,  -3,  -3,   0, -19,  -5, -30, -53,   0,   4,  -4,   5,  13,  15,  11,  23,   0,   3,  -5, -12,   5, -25, -29,  26,   1,   0, -15,  -1,  26,  29, -12,   4,   1,  -3,   0,  -2,  22,  -1,   2,  -1,  -3, -11,  -9, -34,  10,  -4,  12, -10,  -9,  11,  -5,   4, -11,   4,   6,   7,  -8,  10, -57,  28,  -6,  -7,   5,   1 },
      {   2,   3,  -1,  -3,   5, -17,  29,  51,   1,   6,   3,  -7,   7,  10,  21, -37,  -1,  -3,  -8, -27, -13,  -8, -35, -10,  -1,   4, -11, -21,  -4,  17,  -5,  17,  -4,   5, -25, -27, -30,  16,   0,   2,   2,   3, -14,   7,  -5, -11,  -4,  -1,   3, -10, -33,   7,  -8,   9,   3,   2,  -8, -18, -45,   7,   4,  -7,   4,   2 },
      {  -2,  -2, -11,  -6,   1,  -7,  28, -32,   0,   0,  -4,  -4,  20, -15,  30,   2,  -2,   2,   1,   1,  21,  16, -23, -10,  -1,   1,   5,  -3,  37, -10, -40,  15,  -4,   0, -19,  -8, -13,   2,   7,   0,   0,   0,  -5, -37, -15,   5,   7,   1,  -6, -12, -43,  -3,  10,  -6,  -4,  -2,   8,  -4,  59,   2,  -5,  -1,  -2,   4 },
      {  -1,  -1,  -9,  -5,  -7,  17,  -5, -21,   1,  -2,  -3,   6, -11,  34,  21,   8,  -1,  -2,   6,  18, -16,  29, -37,  10,   1,   1,  10,  26, -25,  18,  -7,  -7,   0,  -2,  -6,  51, -44, -29,  -8,  12,   0,  -5,  11,  26,   3,  -8,  -8,   5,  -2,  -1, -24,  26,  13,  -7,  -1,   6,   3,  -7, -10, -19,   3,  11,  -7,   1 },
      {   0,  -1,   7,  -9, -12,  12, -10,   6,   2,   5,   8,   1,  -5,  29,  23,   7,  -4,  -5, -10,   5, -19, -46,   3, -22,   1,   1,   4,  28,   6,  12, -48,  28,  -1,  -2,  -6,  -3,  37,  12,  14,  -4,  -1,   9,  15,  57,  -6,   2,   2, -12,   6,   2,  -1, -24,  20,   3, -11,   1,   3,  -8,  16,  10,   5,   0,   4,   3 },
      {   1,   2,   4,  -2, -17,  -2,  31, -26,  -1,  -1,   2,  -5, -17,   2,   9,   7,  -2,   1,   4,   6, -29,   4,  54, -17,  -3,   0,  -2,  -1, -32, -32, -56,   9,   0,   3,   5,  -9, -25,   9,  -9,  18,  -1,  -3, -19, -16,  23,  -8,  16, -10,   1,   2,  10,  22,  -9,   4,   1,  -7,  -3,   7, -24,   5, -12,  -9,   9,  -9 },
      {  -1,  -2,  -3,   1, -19, -26, -15,  -9,  -1,   0,  -1,  -6, -13, -26,  13,   3,   1,   3,  -2,   7, -31, -25,   0,  17,   0,  -3,   7,   3,  -6,  16,  23, -16,   0,   2,  -4,  12, -18,  68, -16,  -1,   4,   6,  14,  -3,  47, -25,  -6,  10,  -2,  -7, -23, -24,  -3,  -5,   5,   1,   8,  -3,  24,  -3, -15,  10,   5,  -5 },
      {  -1,  -2,  -3,  -7,  -8,  17,  16,  -2,   1,   4,   2,   3,  -2,  28,  -6,  30,   2,   0,  -2,   8,   7,  29,  22, -50,   1,   2,  -1,   0,  13,  19,  26,  -3,   2,   2,  -6, -18,  -9,   2,  -8,  16,  -1,  -5,   0, -17,  -2, -21, -16,   3,  -2,  -1, -11, -62,  37, -28,  -2,   0,  -7,   5, -26,   4, -11,   9,   2,  -1 },
      {   0,   2,   2,   0, -13, -10,   0, -15,   0,   2,   0,   5, -20,   5,  60,  12,   0,   1,   3,   6, -14,  -6,   3,   2,   0,   0,  -2, -11, -16,  -7,  16, -28,  -2,   4,   0,  -1, -14,   3,  -4,  -7,  -2,  -4,  -3,  -2, -69,  22,  -5,   2,   0,   0,  18, -40, -38,  16,  15,  -7,   1,  -3,   2,  -1,  16, -13,  -5,  -3 },
      {   0,   1,   1,  -4,  -8, -13, -13,   9,   1,   2,   2,  -4, -10, -18, -10,  29,  -2,  -4,   3,  -2, -25, -35,  -3,   6,  -1,   1,   3, -10, -18, -18,   4,   4,  -1,   2,  -3, -11, -33, -42,  41, -19,   2,   3,   4, -24,  21,  49, -21, -10,  -2,  -3, -19, -18,  39,  12, -12,   6,  -2,  -1,  -6,  -2,  11,   7,   4,   0 },
      {  -1,   0,  -2,   8,  -3, -22, -24, -17,   1,   4,   0,   5, -10,  -7, -12,   6,  -3,  -6,  -3,  -5, -16, -11, -15, -52,   0,   3,  -1,  -7, -12,  20,  15,  33,   3,   1,  12,  -1, -11,  35,  24,  17,  -1,  -6,   0, -18, -41,   4,   0,   9,   3,   3,  26,  50,  15, -28,  -9,   6,  -1,  -3,   8, -10,  10,  15,  -2,  -2 },
      {   0,  -1,  -1,   4,   3,   9, -16,  -4,  -1,  -1,  -4,  -2,  -8,   0, -24, -10,   1,   2,   7,   2,  -8,  19,  -6,  39,  -2,  -3,  -1,  -6, -14,  -6, -21,  33,   0,   3,  10,  -6,  -4,  24, -24, -35,  -3,   0,  -2, -18, -38, -33,  14, -16,   2,   2,   2, -25,  46,  29,  -4,  -3,   4,  -1,  -9, -33,  23,   5,  12,  -6 },
      {   1,   2,   2,   1,  -6,  19,  -6,   0,   1,  -1,   2,  -4,  -5,  25, -40, -21,   0,   0,   0,  -9, -16,  28,  -6,  32,   1,  -2,   4,  -7,  -6,  19, -31, -11,  -1,  -2,   4,  -7, -11,  31,  34,  17,   4,   3,   3,  -3,  -3,  51,  18,  22,  -3,  -8,  -1, -32, -25, -36, -17,  -7,   1,   3,   1,   3,   1,  -3,  -7,   6 },
      {   0,   0,   3,   4,   0,   6,  -8,  16,   0,   0,  -3,   6,  -1,  -9, -33,  40,  -1,   0,   2,   3,  -5,  18, -22, -10,   1,   2,  -1,   4, -23,   5,   5,   9,  -1,   0,   2,   1, -23,   6, -12,   0,   3,  -2,   7,  14, -18,   1,  32, -19,  -1,  -7,  -8,   0,  -4,  30,  17,   2,   0,   3,  22,  71, -31, -25,  16,  14 },
      {   0,   1,  -1,   1,   1,   5,  -4, -31,   0,  -2,   3,   9,  -6,   2,   8, -40,   0,  -1,  -2,  10,  -2,  22,  14, -12,   1,   0,   6,   1,   5,  15,  26,  52,  -3,  -4,  -9,  -7,  -2,   7, -42, -14,   1,  -1,   5,   6,  22,  64, -17, -22,  -2,   0,   4,  -3,  -3,  17,   8,  25,  -1,  -4,  -1,  -1,  -8, -11,  -2,  12 },
      {   0,   0,  -1,  -5,  -2, -10,   5,  31,   2,   5,   5,  -3,   6,  12,   3,  61,   0,   0,  -5,   0,   3,  15,  13,   5,  -1,   3,  -3,  -2,  13,  12, -25, -12,   0,   2,  -5,  -9,   0,  21, -32, -59,   0,   3,  -4,   3,  -3,  29,  -5,  19,   1,   4,   0,  29, -10, -14,   9,  26,   0,  -3,  -6, -15,  -7,  32,  -9,   6 },
      {   0,  -1,   0,  -1,  -1,  -5, -13,   8,   1,   0,   3,  -1,   3, -10, -28,  20,  -2,   0,   0,   1,   4, -14, -36, -38,   1,  -1,   1,   3,   9,  -2, -23, -25,  -1,  -1,  -3,  -6,  -3,   0, -32,  35,   2,   3,   5,  -3,  13,  15,  19, -33,   2,   0,   3, -16, -21,  13,  15,   5,  -1,  -1,  -3, -67,  11, -25,  -4,   3 },
      {   1,   0,   2,  -5,  -2,  10,  12,   1,   1,   1,   1,  -4,   8,  21,   2,  29,   1,   0,   4,   1,   7,  24,  16,  -1,   1,   1,   1,   1,  11,  24,  17,  24,   3,   0,   1,   6,  -7,  20,  67, -21,   0,   0,   6,  -1,  18, -16, -17, -15,   1,   2,   0,   9, -31,  38,   3,  -6,   2,   1,   7, -34,   3, -41,   9, -15 },
      {   0,   1,  -3,   2,   1, -12,  13,  14,   1,   0,  -4,   4,   2,   3,  29,   4,   0,  -1,  -1,  -3,   7,  12,  13,  23,  -1,   0,  -2,   0,   5,  20,   8, -39,   1,   0,   2,   3,  19,  28,   4,  31,   1,   0,   0,   0,  -8,  38,  14, -23,   2,   0,   4,  27,  69,  18,  -8, -22,   0,  -3,  -7,  -8, -26, -24,  -2,  -5 },
      {  -1,  -2,   0,  -4,   2,   3, -11, -14,   0,   0,   3,  -4,   3,   6, -21,  -9,  -1,  -1,   3,  -1,   6,  13,  -8, -23,  -1,  -1,   0,  -6,  -3,   6, -29, -41,  -1,   0,   0,  -3,   5,  22, -13,   5,  -1,  -1,  -2,  -3,   6,   4, -74,  17,  -2,   2,   0,  11,   6,  47,  -9, -12,   0,   0,   5,  33,  41,   2,   5,  -7 },
      {   0,  -2,  -1,   4,   2,  -7,   1,  23,   1,   2,   0,   1,   2,   1,   6,  28,   0,  -1,  -2,   3,   0,   1,   8,  38,   0,   1,  -1,  -2,  -1,   3,   1,  23,   0,   0,  -1,  -1,  -3,   0, -26,  26,  -2,  -4,  -1,  -5,   0,  -6, -38, -53,  -3,  -1,   1,   8, -10, -55, -18,  -4,   2,   4,  15,  19,  46, -30,  -1,  -7 },
      {   0,   1,   1,  -1,  -1,  -6,  -6,   2,  -1,  -1,   0,  -2,  -9,  -1, -12, -20,   1,   3,   6,   5,   3,  -3,  -9,  10,   0,  -1,   2,  -2,  -3,  -5, -22, -13,   1,   4,   4,   0,   3,  -2,  16,  -1,   0,  -2,   0,  -2, -27,  -9, -60, -41,   1,   2,   6,  -3,  -8, -17,  15,  30,   2,  -2,   0, -11, -77,   9,  23,  -7 },
      {   0,   0,   1,   0,  -1,   6,   7,  -3,   0,   0,   0,   0,   5,   9,   0, -10,   0,   1,  -2,   1,  -2,   7,   4, -10,   0,   0,  -2,  -1,   1,  -4,  13, -25,   0,   1,  -2,   2,  -3,  15,  16,  -5,   0,   0,  -1,  -3,   0,  -7,  24, -70,  -3,  -2,  -6,   0, -17,  21, -46,  28,  -2,   4,   1,  14,  12,  58, -41,  -1 },
      {  -1,   1,  -1,  -1,   1,   0,   5,   0,   0,   1,   2,   2,   2,   5,   2,  -6,   0,   1,   2,   1,  -4,   4,  14,  10,   0,   0,  -1,   3,   5,  -2,  -3,  -8,   0,  -1,   3,  -1,  -2,  11,  27,  15,  -2,  -1,  -4,  -2,   5,  -7,  -3, -18,   2,   0,   3,   3,  13,  -2,  76,  13,  -1,   3,   3,   2,  34,  21,   5,  81 },
      {  -1,  -2,  -1,   2,  -5,  -3,   1, -17,   0,  -1,   1,   2,  -3,   2,  -6, -30,  -1,  -1,   0,   1,  -8,  -2,   1, -29,   0,   1,  -1,   3,  -2,  11,   1, -47,   0,  -1,  -1,   1,  -6,  -5,   1, -71,  -1,   2,   0,   2,   4,  -8,  16, -26,   1,   0,  -3,  11,  11, -41, -13, -11,   1,  -1,   1,   4,  13, -48,  24,  21 },
      {   0,   0,   0,   0,   2,   2,   0,  -3,   1,  -1,  -2,   2,   2,   3,  -2, -12,   1,   0,   0,  -1,  -3,   5,   0,  -3,   0,  -2,  -1,  -2,  -2,  -2,  -5, -13,   1,  -1,  -1,  -3,   2,   2,  13,  -5,   1,   0,  -1,   5,   6,   2,  16,   0,   2,  -1,  -1,  -2,  21, -14,  60,  56,   2,   0,   6,  17,  25, -17, -24, -80 },
      {   1,   2,   2,   2,   0,   2,   3,   2,   0,   0,  -2,  -2,   0,  -2,   7,   4,   0,   1,   4,   3,  -4,  -2,   8,   9,   0,  -1,   0,  -1,   0,  -5,   2, -11,   0,   0,  -1,  -4,   3,   1,   0,  21,   1,   0,  -1,  -3,  -6, -14,   3,  33,  -2,  -2,   3,   0,   7,  12, -47,  93,   0,   0,   2,  -4,   7, -39,  18,  33 },
      {   0,   0,   1,  -2,   0,  -4,  -9,  -3,   0,   0,   1,  -2,  -6,  -5, -11,   0,   0,   1,  -1,   1,  -5,  -6,   3,   1,   0,  -1,  -1,  -3,   0,  -9,  -9,   7,  -1,   0,  -1,   2,  -4,   3,   1, -12,   0,   0,   0,  -1,  -6, -17, -20,   9,   1,   0,  -2,  -1,  10,   0,   1,  -6,   0,  -3,  -8,   2, -23, -42,-107,  28 },
    },
    {
      { 109, -10,  -9,  -7,  -5,  -5,  -2,  -1, -13,  -2,   4,   2,   0,   1,   1,   0,   5,  -4,  -6,  -2,   0,   2,   0,   0, -59,  17,   7,   4,   3,   0,   0,   0,  -2,  -8,  -1,   1,   2,   0,   0,   0,   1,   2,   1,  -1,  -1,   0,   0,   0,   6,  -5,   0,   0,   0,   0,   0,   0,   3,   1,  -1,   0,   0,   0,   0,   0 },
      {  54,  -9,   1,  -2,  -1,  -1,  -1,  -1,  44, -17, -10,  -4,  -1,   0,   0,   0, -32,  -2,   9,   3,   3,   0,   1,   0,  77, -24,  -6,  -4,  -3,  -1,  -1,  -1,  56,  -3,  -9,  -3,  -3,  -1,  -1,   0,  -6,   7,   0,   2,   1,   0,   0,   0,  -7,   1,   1,  -1,   0,   0,   0,   0,   4,  -3,   0,   0,   0,   0,   0,   0 },
      {  17,  -2,  -1,  -1,   0,   1,   0,   0,  62,  -7,  -8,  -3,  -3,  -1,  -2,  -1,  76, -12,  -9,  -3,  -1,  -1,  -1,   0,  26,  15,  -7,  -2,  -5,   0,  -1,   0, -58,  24,   7,   4,   1,   0,   1,   0, -29,   4,   4,   1,   1,   0,   0,   0, -10,   2,   0,   0,   0,   0,   0,   0, -13,   2,   2,   0,   1,   0,   0,   0 },
      {  17,  -5,  -3,   1,  -1,  -2,  -1,   0, -93,   5,  13,   3,   3,   1,   2,   1,  45, -27,  -6,  -4,   0,   1,   0,   0,  58,   8, -11,  -2,  -1,  -1,  -2,   0,   3,   6,  -4,   2,   1,   0,   0,   0,  -1,  12,  -1,  -1,  -1,   0,   0,   0, -21,  -1,   7,   1,   1,   0,   1,   0,   7,  -4,  -1,   0,   0,   0,   0,   0 },
      {  20, -11,  -5,   1,  -2,   0,   0,   0, -10,  21,   2,  -2,  -2,   0,   0,   0, -72,   2,   8,   1,   4,   3,   1,   0,  34,   4,  -6,  -1,   0,   0,  -1,   0, -89,  11,  11,   6,   4,   1,   1,   0,  18, -15,  -2,   0,  -1,   0,   0,   0,  -7,   3,   2,   0,   1,   0,   0,   0,   3,  -4,   0,   0,   0,   0,   0,   0 },
      {  -5,   3,  -8,  -7,  -6,   1,   0,   0, -12,  15,   4,   1,   3,   0,   1,   0, -29,  13,   3,   5,   2,   1,   0,   1, -13,   1,   5,   0,  -1,   1,   0,   0,   3,  19,  -2,   1,  -3,   0,  -2,   0,-111,  18,  10,   3,   3,   1,   1,   1, -32, -17,   7,   0,   3,   0,   1,   0,  10,  -8,  -2,   1,   0,   0,   0,   0 },
      {  -8, -52,   1,  21,  14,   5,   1,   0,  17,   7,  11,  -3,  -2,  -2,   0,   0,  -3,  -2,  -3,  -8,  -5,   2,  -1,   0, -22,   8, -11,   2,   7,   2,   0,   0,   9,  -4,  -3,   2,   3,   0,   1,  -1,  28,  28,  -7,  -4,  -3,  -1,   0,   0, -94,  18,  20,   2,   3,   0,   1,   0,  -5, -16,   3,  -2,  -1,   0,   0,   1 },
      {   5, -76,   3,  10,  16,  10,   4,  -1, -18,  16,  -2,   2,   3,  -1,  -1,   0,  16,  39,  15,  -2,  -3,  -6,  -2,  -1,   1, -50,   3,   1,   8,   4,   2,   0,  -1,   7,   6,   0,   0,  -2,  -1,  -1, -18, -19,   1,   5,   3,   1,   1,   1,  28,  21, -13,  -2,  -3,   0,   0,   0, -47,  15,   7,   0,   1,  -1,   0,   0 },
      {  10,  19,  -1,  -3,   1,  -1,   3,   2,   5,  56,  -6,  -6,  -4,  -2,  -3,  -1,  30,  77,  21,   0, -12,  -8,  -4,  -2,  14,   8,  -2,  -5,  -1,  -1,   1,   0,   3, -35,   9,  -9,  -1,   1,   1,   1,  10, -25,  -6,   9,   3,   2,  -1,   0, -19, -19,   1,   2,   1,   1,   0,   1,  41, -13,  -8,   4,   0,   1,  -1,   1 },
      {  14,  69,  -5, -15, -10,   2,  -1,  -2, -11,  14, -17,  -2,   6,  -1,   0,   0,  -1,  14,  12,  10,   3,  -6,   0,   1,  -2, -24,  12,  -7,  -7,   1,   3,   0,  -2,   0,   5,  -4,  -4,   0,  -1,   0,   9,  17,  -3,   5,   2,   2,  -1,   0, -34,  33,  -4,  -1,   1,   0,   0,   0, -79,   1,  13,   1,   4,   0,   2,   0 },
      {   3,   8,   7,  14,   6,  -4,   1,   3,  10,  70,  17,  -9, -19,  -4,  -3,  -2, -10,  -4, -11, -23,  -3,   6,   0,  -2,   3,  47, -35,   6,   5,   1,  -3,   2,  33,  56,  -5,   3,   0,  -2,  -1,   0,   6,  -2,   8,  -8,  -2,  -1,   0,  -1,  22,   5,   3,  -1,  -3,   0,  -1,   0, -25,  17,   5,  -1,  -1,   0,   0,   0 },
      {   0,  -2,   3,  15,   8,   1,  -2,  -1, -16, -70, -10,  -1,   7,   5,   3,   3,  -9,  63,  -4,  -6,  -7,   0,  -3,   0,  12,  70,   7,   0,  -2,  -4,  -2,  -1,   5,  22,  -5,  -4,   2,   1,   1,  -1,  -1, -17,   4,  -2,   0,   1,   1,   0,  -6,   5,  -3,   0,  -1,  -1,   0,   0, -23,  -3,   4,  -1,   0,   0,   0,   0 },
      {  13,  33,  70,  39,  23,  -8,  -1,   0,  -4, -26,  28,  -7, -11,   0,   3,  -1,   1,   3,  13, -13,  -6,   4,  -1,  -3, -12, -40, -50,   6,  13,   2,   0,   2, -12,   3, -13,  -5,   5,  -1,   2,   0,  -9, -25,  13,  -3,  -2,   2,   1,  -1, -11, -16,   4,   0,  -2,   1,   0,   0,   0,  -8,   0,  -1,  -2,   1,   0,  -1 },
      {   5,   9,  -6,   0,  -5,   8,  -2,  -1,  -9, -10, -37,  -9,   1,   4,  -1,   2,   6,  24,  -2,   2,   1,  -6,  -2,   2, -14, -41,   1,  -3,  -2,   6,   1,   0,   0,  87,   0,  -1,  -7,  -3,  -3,  -1,  26,  25,  -2,   7,   2,  -1,  -1,   2, -10,   1, -11,  -1,   0,   0,   1,   0,  49,  14,  -6,  -2,  -2,  -1,   0,   0 },
      {  -3,  14, -90,  20,  -8,  -4,   1,   5,   6, -18,  43,  -7,  -7,  -5,   5,  -2,   7,  -3,  45, -23,  -8,   4,  -2,  -4,  -4, -18,  -8,   5,  11,   0,  -1,   1,   1,  16, -19,   8,  -3,   0,  -2,   0,   5, -28,   0,   0,   1,   2,   0,  -1,  -3, -13,  10,   1,   1,   1,   0,   0,  -7, -11,   0,   1,   1,   1,   0,   0 },
      {  -2,  -7,   7,   9,   7,  -2,   3,   0,  -8,   1, -70, -14, -11,   1,   1,   3,   1, -44,  67, -13,   7,  -1,   0,  -1, -10,  21,  -5,   0,  -5,   2,  -3,  -1,  13,   3,  37,  -2,   0,  -6,   1,  -3,  -6, -38,   7,  12,   3,   2,   0,   0, -15,  -7,  -8,  -3,  -1,  -1,   0,  -1,  -2,  -4,   3,   0,  -1,   0,   0,   1 },
      {  -3, -20,  19, -64, -24,  -4,   5,   0,   4,  -4,  34,   6,  11,  -4,  -5,  -1,   4,  -4,   6,  41,   7,  -4,   1,   2,  -5,   0, -11,  -1, -11,  -2,   6,  -1,  11,  37,  15, -20,  -8,  -1,  -1,  -2,  16, -37,  -1,   7,   5,   3,   0,   1, -13, -36,   3,   5,   5,   1,  -1,   0, -19, -38,   1,   8,   5,   0,   1,   0 },
      {  -3,  -5, -17, -24,  -2,   3,   7,   3,  -4, -15,  -4, -11,  -5,  -1,  -2,  -1,   0,  11,  28,  32,   6,  -1,   0,  -3,  -9,  13, -88, -11,  -8,   1,   1,   2,  -6, -17,  -1, -16,  -8,  -1,  -2,   1,  -8,  32,  -5,   1,   1,   2,  -1,   0,  20,  48,   6,  -3,  -4,  -3,  -2,   0,  14,   2,  -1,   7,   2,   2,   1,   0 },
      {  -5, -15,  25, -14, -21, -19,   4,   2,   0,   1,  -4, -18,  -4,   1,   1,   2,   4,  14,  55, -19,  16,   7,   2,  -4,   3,   9,  13,  14,   9,   1,  -5,  -3, -13,  -4, -39,  13,  -3,   3,  -2,   0,  13,  66,   2,   5,  -9,  -4,  -1,   1,  18, -48,   7,   0,   1,   1,  -1,  -1, -19,  -7,   1,   1,   0,   2,   1,  -1 },
      {   3,   5,  33,  -9,  -9,   4,  -3,  -3,   6,   0,  52,  -3,  12,  -2,  -1,  -1,   5,  -7,  55,   6,   7,  -4,  -1,   2,   3,  16,  43,  14,   2,   1,   0,  -1,   2,  13,  -4,   7,  -4,  -1,  -1,   0,  -6, -10, -16,   2,  -2,  -1,   1,   0,  -7,  61, -11,  -4,   0,  -1,  -2,  -2,  31,  33,   3,  -3,  -1,  -1,   0,   1 },
      {  -6,  -5,   1, -41, -43, -20,  -1,   3,  -4, -10,  -8, -48, -16,  -2,   2,   2,   2,  11, -32, -31,  24,  15,   2,   0,   0, -10, -15,  23,  22,  10,  -7,  -2,   3, -15,  -1,  41,   1,  -6,  -6,   0,  -2, -31,  25,  15,  -7,  -7,   1,   3, -18,  18,   4,  -1,   0,  -1,   0,   1,   4,  24,   4,  -4,  -2,  -1,   0,  -2 },
      {   2,   5,   7,   7,   1,   3,  -1,   1,   0,   1,   5, -15,  -3,  -8,   0,   2,   3,   6,  -3, -25,  10,   3,   1,   0,   1,  -5,  18,   6,   6,   4,  -1,  -1,   4,  10,  37,  13,   7,  -4,  -1,   0,  -3,  16,  12,  -2,  -2,  -5,   2,   0,  30,  41,  20,  -4,  -3,  -4,   0,  -2,  10, -96,  11,   3,   0,   2,   1,   0 },
      {   1,   4,  -3,   7,   0,  13,   1,   4,   1, -15,  37,  -8,   3, -13,  -7,  -1,  -2,  11,   7, -10,  11,   1,   3,   1,   4,  -1,  -3, -16,   0,   7,   3,   1,   5,  -6,  94,   5,   7,  -9,  -1,  -5,   7,  34,  21,   4,   0,   0,   2,   0,   0, -30,   5,   0,  -2,   0,   0,   1,  -3,  41,   1,   1,  -2,   0,  -1,   0 },
      {   0,   2,  -3, -57,  72,  -9,  16,  -3,  -4,  -8,  10, -48, -16,   3,   4,  -1,  -7,  -7,  -3, -32, -48, -14,   0,  -2,   0,  -3,  11,  -6,  -8,  -6,  -2,   5,  -3,   1,   3,  -2,  -4,   8,   1,   3,  -3,   8, -17,  -4,   3,  10,  -1,   0,  -1,  -2, -11,   3,  -5,   2,   0,   2,  -1,   1, -12,   5,   4,   1,  -1,  -2 },
      {   1,   6, -11,  15,  30,  11,   5,  -1,  -2,   2,   1, -27, -33,  10,   9,   5,   4,   2,  -1,  79,   5,  10,   0,  -2,   6,   3,   9,  67,  25,  -2,  -7,  -5,   3,   0,  12,  25,   5,   1,   1,  -1,   2,   6,   5,  -4,  -2,   2,  -1,   0,   0,  -9,  -8,   0,  -2,   1,   0,   1,  -5,  -5, -11,  -5,  -1,  -2,   0,   0 },
      {  -2,  -1,   5,   4, -30, -37,   4,   5,   0,  -9,   2,  38, -42,   9,  -1,   3,  -2,   5,  -4, -13, -22,  10,   1,   0,   0,  -3, -13,  11,  10,  -1,  -5,   1,   3,   3,  31,  19,   1,   3,  -1,   0,  -8,   8, -77,  10,  -5,   5,  -4,   0,  -2,   2, -38,   0,   5,   3,   1,   2,  -5,  -9, -15,   0,   4,   0,   0,  -1 },
      {  -2,   4,  -5, -39,  28, -29,   3,  -2,   3,  -1,  -8,  83,  -9,  11,   4,  -4,   1,   7,  15, -12,  -8,   9,   1,  -6,   3,  -4,  -2,  25,  24,   0,  -3,   4,   0,   4,   7,   4,  15,   6,   1,   1,   7,   5,  56, -12,  -5,  -1,  -2,  -2,  -5,  13,  13,   1,  -3,  -1,   1,  -1,   7,   7,  -2,  -1,   0,   0,   0,  -1 },
      {  -1,  -8,   6,  19, -12, -50,   8,   0,  -2,   0,   4, -13, -47,  10,  -2,  -1,   3,   2,   5,  36, -17,  17,  -3,  -3,  -3,  -4,  23, -66, -19, -19,   0,   1,   1,   5,   0,  20, -15,   9,  -7,   1,   2,  -4,  33, -13,   0,   9,  -3,  -2,   1,  12,  21,  -3,  -2,   1,  -1,   1,  -1,   4, -15,   5,   7,   3,   2,   1 },
      {  -3,  -8,  -2,  11, -42,  -8,  -7,   3,  -1,  -1,   4, -20, -26,   1,   5,  -7,  -3,  -5,  -5, -10, -37,  -3,   2,   1,   4,  -2,  12,  36,  -1,  -5,  -3,   2,  -6,  -5,   7, -87,  -5,   1,   9,  -1,   0,  11,  36,  -4,  -2,   1,  -2,  -2,  -2,  11, -23,   7,   1,   5,   2,   0,   0,   0,  -8,   1,   3,   0,   0,   1 },
      {  -2,   1,  -5,   6,   6, -66,   5, -10,   1,   2, -14, -21,  45, -11,  -5,   1,   2,   2,  -3,  11,   4,  13,  -1,   0,   2,  -6,   4,  12,  35,  -5,  11,  -1,   1,   7,  17, -33,  14,   9,   3,  -2,  -5,  -6, -37, -33, -11,  -3,   2,   3,   5,   0,  47,  -4,  -3,  -3,  -3,  -2,   1,  15,  12,   5,  -1,   4,  -1,   3 },
      {  -2,  -4,  -4,  16,  -2, -48, -30, -13,   3,   6,   5, -10,  55,   4,  -9,  -3,  -3,  -3,   1,  12, -41, -15,  -1,   0,  -1,   1, -15,  21, -37,  -2,  10,   3,   3,  -1,   3,  36,  23,   0,   4,   2,   3,   9,  24,  39,  12,   1,  -1,   0,   2,   2, -27,   4,   6,   4,   3,   0,   0,  -7,   3,  -1,  -1,   2,   0,   0 },
      {  -1,   1,  -7,   2,  21, -22,  -7,  -4,   4,   6,   9,  -7,  24,  11,  18,  -1,   1,  -5,  -4,   7,  30,  15,   0,  -8,   2,  10,   1, -44,  64,  10,  -2,  -1,  -1,   0,  -2, -19, -15,  -9,   1,   3,   1,   5,  14,  28, -18,  -2,  -2,   2,  -5,   3, -62,   6,  -2,   3,   2,   1,  -1, -15, -20,  -8,  -2,  -1,   0,  -2 },
      {  -1,   1,   9, -10, -17,  29, -47,  -5,   0,  -3,  -8,   0,  -8,   4,  -2,  -2,  -2,  -8,   5,  15, -62, -17,   4,  -3,   2,   6,   2, -22,  61,  21,   1,  -1,  -1,   2,   0,   8,   4,  -6,   0,   4,  -1,   1,  -7,  31,  -4,  -7,   0,  -4,  10,  -1,  45,   8,   6,   1,   5,   0,  -3,   4,   2,  -3,  -6,   1,  -1,  -1 },
      {  -2,  -1,  -7,  -6,  26, -21, -38,  -7,  -3,  -5,   4,   2, -46, -46, -30,  -7,   1,   3,  -2,   7,  31,  15,   4,   5,   0,   0,  11,  -8, -14,  18,   3,   4,  -1,   2, -16, -25,  48, -24,   3,   1,   0,   1, -10,  34,  23,   0,   1,   0,   0,   0,   0,   6,  -2,  -1,   5,   1,   4,   3,  25, -10,  -3,  -1,   0,  -4 },
      {   0,   0,  -2,  11,  21, -24, -21,  -6,   1,   0,   5,   9,  -8,  18,  14,   4,   0,   1,  -9,  -8,  31, -27,  15,  -8,   0,   0,   4,  22, -25,  22,  -6,  -3,  -2,   0,   7, -15, -68,   4,   0,   1,   1,  -3,  -8,  49,  -7,  -3,  -8,   3,   6,   3,  43,   6,   4,  -3,   2,   4,  -5,   6, -26,   1,   2,   1,   0,   1 },
      {   1,   0,   3,   4,  -1, -11,  11,  15,  -2,  -3,  -8,   7,   7, -87,   6,  -6,  -1,  -1,   0,  10, -20, -10,  -4,   2,   1,   4,   1,   3,  -2,  66, -12,  -4,  -1,  -2,  -5,  14, -29,  10, -10,  -3,   1,  -1,   9, -25, -10,  -6,   6,   4,  -2,   0, -14,   7,  -2,  -5,  -2,  -1,  -1,  -2, -19,   4,   5,   0,   0,   2 },
      {   1,   2,   0,   9,  -7,  -3,  51, -19,  -1,  -4,  -1,   6, -14, -41,  -8,  -1,  -3,  -5,  -6,  -1,  12, -61,  -9,   8,   1,   5,  -1,   1,  30, -43,  37,  -6,   0,   2,  -5,   4,   7,  30,   8,  -2,   0,   3,   9,  38,   6,  11,  -6,   2,   2,   4,   4,   7,  11,   5,  -4,   1,   1,   2,  -3,   1,   1,   0,   0,   2 },
      {   2,   2,   4,   4,   4,   7,  53,  32,   1,   1,  -1,   7,  16, -17,  32,   6,  -1,  -2,  -6,   6, -25,  56, -13,  -3,  -1,  -1,   6,   3, -11,  -8, -19,  -8,  -1,   3,  -3, -10,   7, -22,  -7,  -2,   1,   1,  -3,  64, -18,   9,  -4,   1,   4,   3,  20,   5,  -1,  -4,   2,  -1,   2,   6,  17,   6,   6,   4,   2,   1 },
      {  -1,  -1,  -2,  -2,   7,  -3, -16, -33,   0,  -1,   0,   6,  -8, -17,   3, -10,   0,   0,  -3,   2, -13,  16, -19,   5,   1,   1,  -4,   5,   9, -23,   4,  -6,  -3,  -5,   6,   6, -67,  -6,   3,   6,   1,   4,   3,  -4,   2,   5,   1,   2,  -1, -10, -16, -27,  13,  -4,  -4,   1,  11,   0,  82,  -2,  -3,  -2,   0,   1 },
      {   0,   2,   2,  -2,   3,  14,  -8, -13,  -2,  -2,  -6,   2,  11, -12, -56, -11,   0,  -1,   3, -10,  -4,  48, -44,   6,   1,   0,   0,  14,   6, -21,  33,   7,  -2,  -1,  -1,   5, -32, -28,   9,  -5,   0,   1,   1,   1,  17,   5,   5,   7,   2,   6,   8,  21,   4,   2,  -3,   3,  -7,  -1, -56,   6,   2,   2,  -2,   0 },
      {  -1,  -2,   4,  -6,   1,   4, -49,   8,   0,   2,  -5,   1,   6, -47,  57,  -8,   1,   4,   2,  -3,   8,  -2,  18,  -7,  -2,  -4,  -4,   2,  -2, -69, -22,  -1,   1,   3,   4,   7,   4, -18,   0,  12,   0,   0,  -5, -15,  -1,  19,  -3,  -3,   0,   2,   3,   6,  14,   2,   6,   0,  -3,   2, -30,   3,   2,   8,  -1,   2 },
      {  -1,   0,   0,   0,  -9,  -4,  24, -71,   1,   1,   4,   2,  -6,   6,  51,   8,  -1,   0,   0,  -2,  -1, -13, -31,   6,   1,   2,  -1,  -4,  -6,  28,   3,  -9,  -1,  -1,  -2,   0,  17, -61,  12,   6,   0,   2,  -2, -13,  21, -13, -16,   3,   2,  -1,   7,  -8,   7,   7,  -1,   0,  -2,   1, -15,   3,   5,   0,   0,   5 },
      {   0,   1,  -1,   3,   3,  -8,  23,  13,   0,  -1,  -2,   4,   3,  -5, -38,   6,   0,   0,  -2,   4, -13, -21,  67,  -1,   0,   1,   2,   4,   7, -16,   6,  -9,  -3,  -2,  -7,   1, -14, -76, -21,  -3,   0,  -1,   4,  -5, -12, -16,  -3,   3,  -2,   0,  -2, -35,  -1,   6,  -3,  -5,   0,  -3,  -6,  -7,  -2,   2,   4,  -6 },
      {   0,   0,   2,   4,  -1, -20,   9,  48,   1,   0,   1,   1,   6,  11,  -8,  10,  -2,  -3,  -3,   3,   5, -36, -40,  -6,   0,   1,   2,   1,  21,  -8, -50,  10,  -1,  -3,   1,   0, -14, -29,  -5,  -3,   2,   4,  -1,  -7,  70,   0,  12,  -5,   0,   1,   2,  24,   1,   1,   1,  -5,   1,   1,  15,  -1,  -6,   4,   0,  -1 },
      {   0,   0,   4,   1,  -2,   5,  11,  -8,   1,   1,   0,  -1,   8,  -2,  16,   4,   1,   0,   0,  -6,  -5,  35,  46,  -7,   1,   3,   0,   1,  12,  14,  16, -12,   0,  -1,   1,  -5,  -4,  32, -24,   7,   4,   3,   1,   6,  86,  16,   0,   4,   1,   3,   2, -33,  25,   1,   1,   2,  -1,   0, -14, -10,  -2,  -1,  -3,  -3 },
      {   0,   1,  -2,   3,  -3,   7,  11, -65,   0,   0,   1,   1,   8,  -5, -32,  -9,   1,   1,   1,  -1,  -3,   6,  18,  17,  -2,  -2,  -3,  -3,   2, -15, -86, -14,   0,   0,   1,  -4,   3,  20, -28, -16,   0,  -1,   1,   8,  -3,  -5,  10,   4,   0,   1,   1,  19,  -8,  -2,   3,   5,  -1,  -1,  -8,   4,  -3,  -4,  -1,  -1 },
      {   0,   0,  -1,  -4,   0,   1, -13,   9,   0,   1,   4,   2,   3,  -8, -14,  21,  -1,  -2,  -2,  -1,   1,  -9, -45,   4,  -1,  -1,  -1,   0,   1,  -3, -31,  -5,   0,   0,  -1,  -9,  10,  13,  10,   4,   1,   0,   3,  13, -10,   8,   6,   0,  -1,  -2,   6,-101,  -7,   6,  -1,  -2,  -5,  -1, -28, -16,   5,   1,   5,  -6 },
      {   0,   0,   1,   1,  -6,   2,  19,  -2,  -2,  -2,  -2,  -1,   3,  -1,  -8, -66,   0,   1,  -2,   3,   0,   2,  21, -49,  -1,  -1,   3,  -3,   0,   4, -23,   2,   2,   3,   0,   8,  -2,  -1,  84,  14,   0,  -1,  -4,   4,  11,  -1,   6,  -2,  -1,   1,   3, -12, -14,  -3,  -8,   6,   0,   0,  -2,   2,  -8,   2,   0,   3 },
      {   0,  -1,   1,   6,  -4,   3,   4, -12,   0,  -2,  -2,  -1,   1,   4,   1, -49,  -1,  -2,  -4,   2,   6,  -9, -14, -34,   1,   3,   5,   8,   3,   6,   6,  63,  -1,   0,   0,  -2,   6,  -6, -66,  29,   1,   0,  -3,   6,  -5,  25,   0,   4,  -2,  -3,   1, -19, -21,   2,  15,  -4,   0,   0,   1,  33,  -4,   0,   1,   3 },
      {   0,   0,   0,   2,  -5,   1,  -4,  -6,   0,   1,   0,   1,   3,   4, -20,  35,   0,  -1,  -2,  -3,   4,  -6,   4, -28,   0,   1,   2,   2,   4,  19,  -5, -40,   0,  -1,  -2,   1,   2, -13,  10,  17,   0,   1,   4,  -9, -10,  97,  -5,  17,   1,   2,   0,  13,  -6,  15,   9,   1,   3,   0,   9,  20,   9, -11,   1,   8 },
      {   0,   0,  -1,   2,  -2,  -2,  -8,   5,   0,   1,   0,   4,  -1,   1,  15,   4,   0,  -1,   0,   0,   0,   2,  20,  67,   1,   1,   2,   1,  12,   1,   5,  15,  -1,   0,  -5,   5,  -6,  -4,  25, -21,   1,   1,   1,   9,  18,  16,  11,   7,  -4,  -2,  -2, -14, -61, -10,  -7,  -1,   0,   1,  -1,  68,  -9,   4,   3,   4 },
      {   0,   0,   0,  -1,   0,   2, -14,   0,   0,   0,   2,   3,  -4,  -4,   2,   5,  -2,  -3,  -2,  -2,  -3,  -3, -16, -56,   0,   0,   1,   1,  -2, -14,  16, -66,  -1,   0,   0,  -2,   3,   7, -27, -14,   0,   0,  -2,   2,  14, -40, -13,  16,  -5,  -1,  -7,  -2, -54, -19,  -7,  12,   0,   2,   1,  30,  -2,  14,   0,   1 },
      {   1,   0,   0,   3,   6,   5,  -2,  -2,   0,  -1,   0,  -1,  -2,  -7, -17,  42,   0,  -1,  -1,   0,   3,   9,   7, -20,   0,  -1,   2,  -1,   4,   0, -13,  18,   1,   2,  -1,   0,   3,  10,  16,  34,   0,  -1,   4,   3,  -2, -41,  -9,  -5,  -1,   2,  -7,  -3,  47,  30,  -6,  -6,   1,   5,   3,  78,  12,   9,   5,  13 },
      {  -1,  -1,   0,  -1,  -9,   4,   5, -22,   0,   1,   1,   2,   3, -18,   0,  66,   0,   1,   1,   2,  -7,   9,  13, -27,   0,   0,   0,   1,  -2, -10,   1,  51,  -1,   0,   2,   1, -13,   6,  15,  29,   0,   0,  -1,   2,   9,  -8,  -6, -15,  -2,  -2,   1,  13, -62,   1,  -3,  -2,   0,  -2,   2, -39,  -5,  -3,   0, -10 },
      {   0,   0,   0,   2,   1,  -1,   2,  13,   0,  -1,   0,   0,   3,   4, -10, -22,   0,   0,   2,  -1,   0,   1,  -4,  46,  -1,  -1,  -2,   1,   1,  -1,  -5, -34,   1,   1,   5,   0,  -1,   6,  -6,  96,   0,  -1,   0,  -1,   4, -12, -41,   6,  -1,  -2,   1,   9, -20,  -1,  -6, -18,  -1,   0,  -2, -10,   3,  -3,   9,  -8 },
      {   0,   0,   1,   0,   0,   4,  -5,   6,   0,   0,  -1,  -2,   7,  -4,  -9,  -9,   0,   1,   0,  -1,   1,   3,   1,   0,   0,   0,   0,   0,   6,   3, -12,  19,   0,   0,  -1,   3,  -1,   5,   4, -44,  -1,  -1,   0,  -5,   7,  10,-112, -10,   1,   0,   0,  -4,  -2,  14,  -3,   9,   1,  -1,   5,  -1,  12,   4,  10,  -2 },
      {   0,   0,   0,  -1,   1,   1,  -2,   6,   0,   0,  -2,   2,  -1,  -1,  10, -12,   0,   0,   2,   0,   0,   3,   0,   4,   0,   1,  -1,  -1,  -3,  -2,   3,  -1,   0,   0,   2,   1,  -4,   6,  -2, -10,   0,   0,  -3,   0,   2, -13,  14,  58,   1,   2,   6,   5, -20, 106,   2,   5,   1,  -1,   7, -14,   1,   3,   7, -14 },
      {   0,   0,   1,   4,   0,   3,  -5,  -3,   0,   0,  -2,   0,   3,   4,   0, -12,   0,   0,  -1,  -2,   3,  -2,   5,   1,   0,   1,  -1,   3,   8,   1,   2,   0,   0,   0,  -1,   2,   0,  -2,  -2,  -2,   1,   1,  -1,   3,   7,  -1,  19, -26,  -1,  -2,   0,   2, -15,   5, -15,   1,   5,   2,   6,  -4, 120,   5,  -1,  -4 },
      {   0,   0,  -1,  -1,   1,  -2,   2,  -3,   0,   0,  -1,   0,   2,  -3,   2, -11,   0,   1,  -1,  -1,   3,   4,  -4,  -1,   0,   0,  -1,   2,  -3,   2,   8, -28,   0,   0,   1,   1,  -6,   1,  -8,  -7,   0,   0,  -3,   1,   0,   3,   8,-102,   0,   1,   2,  -1, -15,  51,  11, -20,   0,   0,  -3,  11, -26, -19,  -6, -11 },
      {   0,   0,   0,  -2,  -1,   3,   5,  -5,   0,  -1,   0,   1,  -1,   3,  -6,   0,   0,   0,  -1,   0,   2,   1,   2,   8,   0,   0,   0,   1,   0,   6,   2,  -5,   0,   0,   0,   1,  -4,   3,   3,   4,   0,  -2,   0,  -2,  -1,   7,   4, -15,   0,   0,  -4,   1,  -4,   6,  45,  -4,   0,   1,   3, -14,  -3, 110,  13,  34 },
      {   0,   1,   1,  -1,   1,  -2,   3,   5,   0,  -1,   1,  -1,   1,  -2,  -3,   1,   0,   0,   2,   0,  -1,   1,   2,   3,   0,   0,  -2,  -1,   1,  -7,   2,  -3,   0,   0,   0,   2,  -5,   2,  12,  -3,   0,   0,  -1,  -2,   6, -20,  -5,  11,   1,   0,   1,  -5,  -9,  -5, 103,  -5,   0,   1,   0,   0,  19, -51,  -3,  41 },
      {   0,   0,   0,   3,  -2,   1,   5,  -3,   0,   0,   0,   0,   1,   0,   0,   3,   0,   0,   1,   1,   1,  -2,  -1,  -1,   0,   0,   1,   0,   0,   1,   1,  -1,   1,   1,   1,   1,  -1,   1,  11,   0,   0,   0,   1,  -3,  -2,   0,  -5,   8,   0,   1,   0,   1,   7, -12,  49,  -1,   0,   1,   2,  17,   6,  21, -39,-106 },
      {   0,   0,   1,   1,  -1,  -1,   3,   7,   0,   0,  -1,   0,   3,   0,  -2,   2,   0,   0,   3,   1,   1,  -2,   0,  13,   0,   0,   0,   1,   1,   1,   0,   0,   0,   0,  -1,   0,   1,  -5,  -4,  20,   0,   0,   1,  -3,  -1,  -2,   1, -18,   1,   1,   1,  -3,  -1,   9,   4, 119,  -1,  -1,   3,   1,  -4,  -1, -31,   9 },
      {   0,   0,   0,  -1,   0,   0,  -1,   3,   0,   0,   0,   1,  -2,   0,  -2,  -1,   0,   0,   0,   1,   0,   0,   1,   1,   0,   0,   0,  -1,  -1,  -1,  -4,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   0,   0,   1,   3,  -1,   0, -11,   9,   0,   0,   1,  -4,  -5,   9, -16, -32,  -2,  -2,   1,  -5,   1,   9,-116,  35 },
    },
  },
  { // 21
    {
      {  75, -75,  43, -25,  14,  -8,   4,  -2, -28,  29, -19,   9,  -4,   2,  -1,   1,   1,   9,  -8,   4,  -2,   1,   0,   0,  11,  -3,   3,  -3,   1,  -1,  -1,   0,  -2,   1,   0,   1,  -1,   0,   0,   0,  -1,   0,   1,   1,   0,   0,   0,   0,   0,   1,  -2,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -76, -24,  62, -32,  20,  -7,   7,  -1,  50,  19, -30,  16, -13,   4,  -3,   1,   0,   9, -12,   7,  -2,   3,  -1,   1,  -4,  -2,   2,  -3,   3,  -1,   2,   0,  -7,  -3,   1,   0,   0,   1,   0,   1,  -2,   2,   1,  -1,   0,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,  -1 },
      {  32, -21,  -4,   1,   3,  -3,   2,  -2,  72, -52,  17,  -8,   2,  -2,   0,  -1, -69,  33,   7,  -2,   2,  -1,   2,   0,   2, -12,   1,   5,  -1,   2,  -1,   0,   4,  -8,  -6,   5,  -3,   2,   0,   1,  10,   1, -11,  -1,   1,  -1,   0,  -1,   6,   6,   2,  -1,  -1,   0,  -1,  -1,  -4,  -1,   0,   1,   1,   0,   0,   0 },
      {  30,  -8,   3, -13,   2,   0,   1,  -1,  55, -28,  19,  -9,   3,  -5,   0,  -2,  80, -29,  -6,  -1,  -2,   0,  -2,  -1, -32,  37, -12,  10,  -4,   0,   0,  -1, -25,   1,   1,   2,   2,  -1,   0,   0,  16,  -4,  -2,  -1,   0,   1,  -1,   0,  -5,   3,   0,   1,   0,   1,   1,  -1,  -2,   0,  -1,  -1,   0,   0,   0,   0 },
      {  16,   6, -28,  16, -12,   7,  -3,   2,  20,  50, -31,   9, -10,   1,  -1,  -1,   2,  29,  -5,  -7,   5,  -5,   2,  -2, -81, -18,  31,  -6,  10,  -5,   5,   0,   2, -38,   9,  -3,   2,   3,  -1,   0,   8,   1,  -4,  -1,  -1,   1,  -1,   0,  -2,  10,   0,   0,  -1,   0,   1,   0,  -5,   0,   1,   0,   1,   1,  -1,   0 },
      { -22, -15,  20,   4,  -2,   0,  -1,  -1, -44, -43,  42, -16,   9,   0,   4,   2,  14,  54, -59,  24,  -2,   0,   0,   0, -33,  -5,  16,  -5,   3,   3,   1,  -1,  -7, -12,  10,  -4,  -5,  -1,   2,   0,  13,   1,  -7,   1,  -1,  -2,   1,   1,   7,  -2,  -4,   0,   2,   1,  -2,  -1,   6,   2,  -1,   1,   0,  -2,   0,   0 },
      {   7,  18, -17,   1,  -4,   1,   3,  -4,   0,   5, -26,  12,  -6,   5,  -3,   0,  -5,  32, -14,   1,   2,  -4,   0,   0,  40,  19,  -9,  -7,   2,   0,   0,   0, -95, -14,  31,  -4,   9,   5,   2,   1,  10, -16,  -9,   5,  -3,   3,  -2,  -3,  20,   8,  -4,  -1,  -3,  -4,   1,  -2,  -3,  -5,   1,  -1,   2,   0,  -1,   2 },
      {   2,  19,  25, -25,  18, -17,  10,  -5, -33, -14,   2,   3,  -5,   7,  -1,   4, -52, -28,  32,  -4,   3,   6,  -3,   3, -65,  40, -18,   8,   0,   0,   2,  -1, -24,   7,  10,  -7,   2,  -2,   3,  -1,  16, -15,   8,   0,   2,   1,  -1,  -1,  13, -11,  -3,   2,  -1,   1,   0,   0,   8,   5,  -3,  -1,   0,   0,   0,  -1 },
      { -28, -44,  -4,  17, -11,   9,  -5,   0, -15,  -1,   6,   1,  10,  -7,   5,  -6,   6, -25,  34, -25,  14, -13,   2,   1,   1, -29,   5,   2, -10,   6,  -3,   6, -19,  -7, -15,  12,   3,   4,   2,  -2,  72,  12, -29,   1,   0,  -6,   0,  -7,  34,  17,  -8,  -7,  -3,  -4,   1,   0,   6,   7,  -1,  -3,   2,   3,  -4,   3 },
      {  24,  51,  15, -26,  17, -20,   8,  -9,  -9, -43, -27,  43, -24,  14,  -8,   4,  25,   0,   7,   0,  -5,   2,  -1,   2,   7, -46,  25,  -7,   6,   2,   3,  -3,   3, -13, -18,   9, -11,  -3,   2,   0,  19,  21, -16,  -4,   5,  -2,   0,   3,  10, -16,  11,  -2,   3,   3,  -8,   0,  10,   9,  -1,   1,  -1,  -1,   0,  -1 },
      {   6,  -4,   7,  20, -15,   1,  -4,   2,  15,   9,   7, -13,   1,  -1,   6,   3,  -5, -16,  -1,  23,   3,   2,   0,   1, -22, -47,  27, -10,   2,   8,  -2,   4, -52,  69, -18,   0,  -9,  -5,   4,  -2, -30,   1,  13,  -1,   5,  -8,   3,   0,  12, -29,  -4,  -1,   3,   1,  -3,   1,  21,   4,  -3,   1,  -1,  -1,  -2,   0 },
      {   5,  13,   4, -15,   3,   2,   2,   6,   3,  27,  40, -17, -10,   2,   1,   3, -25, -46, -17,  31, -12,  20,  -2,   1,  15, -13,  31,  -1,  17, -13,   6,  -9, -22, -23,   8,  -3,  -3,  -2,   3,   4,  34,  14, -20,  -5,  -3,   1,   0,   7, -55,  12,   9,  11,   3,   9,  -2,  -2, -12,  -8,   3,   2,  -2,  -3,   3,  -4 },
      {   6,  26,  -9, -28,  35, -22,   6,  -7,   6,  35,  15, -35,  19, -16,   1,  -4,   9,  28,   6,  16, -41,  15, -12,  10,  -3, -33, -35,  14, -11,  -2,  -3,   6,   5,  16,  -8, -15,  25,  -2,  -2,  -4,  22,  15,   5,   8,  -3,  -3,  -4,  -6,  30,  10,  -6,  -9,  -1,  -7,   4,   1, -12,  -9,   7,  -2,   3,   7,  -1,   3 },
      {   9,   8,   9,   0,  -3,   2,   0,  -3,  16,   4,   4,  -9,  10,  -6,   3,  -1,  -4, -40, -21,  16,  -9,   6,  -1,   3,  14, -11,  21, -20,   7,   0,  -2,  -1,  27, -30,  31, -10,   3,  -2,  -3,   0,   2, -55,   8,   8,  -6,   9,  -1,   5,  64, -11, -24,   9,  -3,   0,  -1,  -1,  14,  27,  -3,  -1,  -3,  -2,   1,  -2 },
      {  14,  22,   8,  15, -19,  11, -12,   0,  10,  10, -29,  25,   2,  -1,   3,   0, -18,  -8, -61,  15,   7,  -5,  -1,   4,   0,  20, -23, -11, -17,  12,  -5,   5,  25,  49,  -1,  -1,  -5,  -5,   1,  -5,  61,   7,   2,   3,  -1,  -7,   2,  -4,  -6,   5,   1,  -6,   6,  -5,  -1,   2,   4,   4,   3,  -4,   3,   3,  -2,   2 },
      { -12, -25, -54, -13,  48, -24,  14, -12,   2,   2,  27,  21, -38,  24, -14,   5,   5,   4, -13, -26,  14,   1,   2,   0,   0,  -8,  10, -23,   8,  -6,   4,   1,   9,  41,  19,  -7,  -2,   8,  -5,   0,  15, -28,  11,  10,  -4,   2,   1,  -2, -13,   4, -13,   6,  -5,  -1,   5,   1,   1,   7,  -5,  -2,   0,  -1,   1,  -3 },
      {   8,  21,   6, -24,  -1,  -6,   4,  -4,   7,  40,  51,  10, -16,  -3,  -4,  -7,  -8,  -3, -34, -22,  46, -21,   8,  -6,   3,   3, -20,  15, -17,   4,  -5,   5,  -5, -21, -26,  32, -13,   7,   6,   0, -23,  16, -15,  -6,   5,  -1,  -3,  -3,  22,  -4, -10, -10,  -2,  -5,   0,   2,  27,   8,  -7,  -1,   4,  -3,  -4,   3 },
      {  -5, -25, -17,  48,  -3, -16,   9,   0,   5,  11,   6,  13, -44,  28, -17,  10,   1,   1,   1,  35, -29,  28,  -8,   2,   3,  19, -26,  31,  -1,  -7,   4,  -8,   4, -17, -10, -11,   3, -17,  10,  -2,   4,  17, -16, -11,   3,  -1,   0,   6,  20, -32,  12,  -1,   3,   4,  -7,  -1,  30,  13,  -6,   6,  -3,  -3,  -2,  -1 },
      {   0,  -9,  -8,   8,  14, -14,   8,  -6,   2, -14, -10,  -1,   7,   4,   4,   1,  -6, -32, -11,  11,   5, -13,  -4,   7, -10,   2, -10, -35,   4,   1,   1,  10,   4,   0,  44, -18,   5,   9,  -1,  -4, -37,  71, -40,  25,  -9,  -6,   0,  -7,  14,  20, -10, -10,  -2,  -5,   4,   1,   6,  -5,   5,  -6,   2,   6,  -5,   3 },
      {  -2,  -5,  -3,   6,  -3,   5,  -5,   2,  -9,  -7,   8,  28, -16,   9,  -9,   5,  -5, -20, -17,  16,  -7,   7,   3,   0, -13,  -5, -11,   8,   9,  -6,   6,  -5,  -6,  11, -28,  14,  -6,   1,  -1,   3, -24, -14, -13, -11,   1,   5,   1,   6,  38,  39,   4,  -5,   2,   2,  -4,  -1, -86,  -2,  20,   2,   4,  -3,   3,  -3 },
      {   8,   7,  37,  37,   8,  13, -21,   3,   3,  -2,  24,  29,  -7,   1,  -5,   0,   5,  -7,  -1, -32,  10,   7,   4,  -1,  -7, -52, -45,  16, -16,   9,   0,  -2,  -6, -11,  43, -32,  12, -10,   0,  -2,  -4, -15,   6,   8,  -5,   5,   2,   6, -21,  -6,   4,  -2,   6,   1,  -2,   1,  -8, -16,   3,   1,  -4,   0,   3,  -1 },
      {  -8, -25, -39, -38, -18,  15,  -1,   8,  -6, -21, -11,  27,   3, -10,   4, -10,  -9, -25, -20,  28,  -5,  -3,  -2,   5, -10, -29, -38, -10,  -2,   4,   2,  12, -12, -33, -21,  -7,  26,   7,  -4,  -1,  -6,   3,  45, -11,   4,  -3,  -4, -13, -16, -11,  -2,   2, -11,  -7,  10,   5,   9,   0, -16,   0,   3,   8,  -4,   6 },
      {  -5,  -7, -11,  26,  30, -34,  22, -24,  -4,  -4, -30, -15,  32, -10,  -3,   4,  -4, -21, -26,  -4,  -3,   9,  -5,  11,  -9, -12, -17,  -1, -11,  19,  -7,   5,  -7, -19, -26,  26, -27,   5,   1,  -1,  -9, -43, -28,  14, -10,   6,   0,   0, -23,  -3,  19, -12,   9,  -1,  -5,  -5,  11, -27,  -4,   5,   3,  -2,   1,   0 },
      {  -1,  -3,  -4,   0,  25,  31, -26,   8,   2,  10,  26,  47,  23, -34,   3,  -9,  -5,   7,   2, -22, -35,  29,   2,   3,  -4,  25,  11, -50,  -2,  -8,   4,  -2,  -6,   1, -19,  10,  12, -10,   4,  -2,  -5,   6, -19,   9, -13,  11,   6,   3,   9, -29,  21,  -4,   3,  -3,  -1,  -1,   7, -10,   2,   7,  -4,  -1,   1,   0 },
      {   0,   3,  -6, -11,  15,  31, -19,  17,  -5, -12, -13,  14,  -8,  -8,  11,  -8,  -3,  -1,  -5,   2, -27,   8,  -6,  -3,  -7,  -8,   1,  38,  -3, -20,   3,  -6,   0,  17,  -1,   7,   2,  -5,  -1,   3, -20, -23, -31,  -3,   3,  -3,  -1,   5, -11,  62, -24,  -2,  -2,   3,   4,   0,  59,   9,   5, -10,  -3,   2,   0,  -3 },
      {  -8, -13, -33, -43,   1,  13, -12,   2,  -3,   0, -13,   3,   6, -13,   7, -12,  -2,  -4,  -6,  10,  12,  -5,   5,  -3,  -3,  -5,  14,  49, -24,  20,  -8,  -7,   9,  14,  43,   3,  -2, -17,   6,   1,  -3,  -2, -40,  -3,  16,   2,   7,  12,   9, -49,  16,   3,   3,   9,  -7,  -3, -20, -17,   4,  -1,  -1,  -5,   2,  -1 },
      {   1,   5,   8, -14, -31,  12,   2,  -1,   1,   6,  17,   8, -28,   8,  -4,  -5,  11,  17,  41,  50,   6,  -6,  -7,   1,  -4,  -3, -24, -37,   9,   8, -11,   8,  16,   4,   8,  19, -10,   7,   3,  -6,   4, -35, -33,  25,   6,  -7,   6, -10,  -9,  -2, -15, -15,   0, -10,   4,   0,   5, -42,  -8,   6,   5,   0,  -4,   7 },
      {   2,   1,   5,  -7,  -4,  23,  -7,  14,  -5, -14, -21, -46, -38,  13, -16,   1,  -4, -14, -26, -32,   4,  12,  10, -11,  -1,  -4,   0,   0,  16, -28,   4,   0,   5,   4, -15,   3,  23,   5,  -4,   3,   6,   6,   2,  12, -11,  17,  -2,   0,  24, -14,   7,   1, -14,  -5,   5,   4,  12, -67,   9,  13,   4,  -3,  -2,  -1 },
      {   0,  -2,  -4, -16, -16,  28,  11,   6,  -2,  -1, -13, -26,   4,   2,  -5,   1,   1,   8,  -3, -20,   7,  37, -17,   0,  -5, -18, -46, -11,  34, -11,  16, -16,  -3,   7,   9,  24, -27,   9,  -2,   4,   3,   7, -19,  20,  -4,  18,   3,   5, -11, -22,  -3,   2,   3,  -3,   3,  -4,  -9,  61,   7,  -2,  -5, -12,   5,  -5 },
      {  -2,  -2,   7, -10, -47,  -7,  17, -19,  -1,  -2,   9,   0,  -7,  30, -20,  -4,  -6,  -5,  -9, -32, -43,   4, -29,  21,  -3,  -2,  26,   4, -43,  11,  -6,  11,  -1,   5,  13,  26,  39,   3,   1,  -8, -11,  -5,  -4,  10,  -7,  11,   5,  -4, -15,   1,   3, -21,  -2, -24,   8,   3,  -6,  16,   6,  -2,   5,   3,  -4,  -1 },
      {   3,   6,   9,  21,  46,  14, -38,  14,  -4,  -5, -12, -30, -35,  10,  15, -25,  -2,  -4,  -3,  10, -12, -24,  19,  -5,   0,   4,   2,  -3, -18,   0, -13,  19,  -5, -11, -12,  17,  16,  10,  -1,  -3,  -2,  -6, -12,  20,   9, -13,   6, -14, -17, -24, -11,  -3,  -5, -11,   9,   2, -28,  39, -26,  -3,  10,  10,  -8,   3 },
      {   1,   1,   3,  20,   4, -10,  14, -10,  -3,  -9,  -8,   0, -35, -53,  24, -23,   1,   1,  -2,  -6,  19,  53, -38,  10,   2,   3,   9,  -2,  -5,   7, -12,   7,   1,  -2,  14,  29,   2, -16,  19, -11,  -1,  14,  18, -14,  23,  -1,  -3,  -6,   0,   6, -36,   7, -16,   9,   2,  -1, -13, -14, -11,  -4,   1,  -1,   2,   1 },
      {  -1,  -2,   4,  15,  -4, -30, -14,   5,   1,   1,   7,  16,   8, -11,   7, -23,  -2,  -6,  -5,  -5, -33, -20,  -4,  -3,   0,  -2,  -4,  20,  40, -30,   5,  -2,   1,   6,  34,  43, -15,  44, -33,  14,   7,   4,  15, -22,  18,  -8,   5,  -5,   4, -20,   3, -19,  12, -13,  10,   0,   4, -16,   3, -10,   6,   8,  -2,   1 },
      {   2,   5,  10,  12, -14, -20,  27,  -5,   0,  -1,  -1,  22,  17, -29,  -4,  -7,   1,   4,  -5,   5,  19, -12,  -5, -16,   2,   1,   7,  25,  28, -29,  17,   4,   5,  15, -22, -26,  50,  10,   0,   2,  -1, -14, -40,  12, -16,  20,  -4, -11, -15, -22, -21,   9, -28,  -2,   7,   6, -12,   8, -22,  -6,   8,   5,  -1,   3 },
      {   2,   3,   7,   3,  11,  30,  29, -21,  -2,   1,   2, -15, -18, -18,   9,  -9,   2,  -1,  -1,   3,  17,   1,  -7,  10,   1,  -2,   0, -12,   7,  -1,   8,  -2,   6,   9,  11, -20,  20,  17,  -7,  -1,  -4, -16, -13, -45,  21, -14,   0, -15,   8,  16,  72, -32,  -6, -11,   2,  -2,   7,  17, -14,  13,  -6,   7,  -6,   4 },
      {   4,   4,   4,  18,  35,  64,  52, -25,   0,   1,   0,   7,  23,   9, -34,  16,  -1,  -2,   2,  16,  -3, -27,  -8,  -1,  -1,  -3,   2,   7, -12, -19,  -1, -12,  -1,  -1,   8,  30,   7,  -7,  -9,   0,   5,  12,  14,  -9,  11,  -5,  14,  -2,  -1, -12, -32,  10,   2,   2,  -3,   2,  -5, -12,  -2,   6,  -1,  -5,   4,  -4 },
      {  -2,  -4,  -4,   1,   3, -19,  -4,   3,  -1,  -5,  -5,  -7,  -6, -31, -24,  11,   0,  -2,   2,  13,  22,  -6,  17, -18,   0,   0,   0, -12,  -2, -28,  -6,   2,  -1,  -8,   0,  10,  30, -34,   2,  -7,   2,  -9,   8,  11,  22,  -8,   5,   5, -12,  -7,  -1, -48,  17,  -3,  -2,   1,   7,  17,  71, -16, -13,   0,  -4,  -7 },
      {   0,  -1,   0,   5,   5,  24,  13, -15,  -2,  -1,   6,  11, -33, -38,   5,  -4,  -1,   0,   6,  17,   4,  -7, -11,  19,   3,   8,  18,  35,   0,  23,   4,   3,  -3,  -4,  -9, -13, -17,  35, -22,   8,   3,   8,  25,  61, -33,   9,  -5,   9,   4,  -1,  15,  -4,  11, -11,   7,  -7,   4,   4,  29,   5,  -2,   6,  -3, -10 },
      {   2,   3,   0,   0,  10,   0, -30,   4,   0,  -2,  -5,   1,  24,  14, -43,   4,   3,   1,   7,  19,  33,  61,   0, -14,   2,   5,  17,   8, -29,   9,  18,   6,   1,  -3,   0, -11,  -6,  45, -22,   5,   0,   3,  -3, -21,  -5,  -6,  17, -14,  -1,  -2, -22, -21,  -9, -20,  17,  -5,   1,  -2,   1,   5,  -3,   6,  -7,  -5 },
      {   1,   0,   1,   2,   6,  13,  -4, -19,   2,   2,   2,  -1,   2,  33,  40, -12,   0,   1,  -5, -12,   5,  -5, -30,  35,   3,   6,  -2,  -8,   4,   3,   8, -10,   0,  -9, -26, -32,   1,  16,   1,  13,  -2, -10, -18, -21,  21, -19,  -7,   2,  -5, -27, -26,  11,  -3,  -7,   5,  -7,  -1, -13,  58, -33,  -2,   6,  -7,   5 },
      {  -1,   0,   1,  -5, -20,   4,   5, -58,  -3,   1,  -2,  -8, -21, -28, -17,  -4,  -2,  -1,  -2, -16, -36,  10,  66, -18,  -2,   2,  -8,  -4,  13,  35,  -4,   1,  -1,   2,  -2, -12,  -4,   3, -14,  12,  -3,   3, -13, -19,   0, -17,  18, -11,  -6,  -1, -24,  17,  -1,  13,   4,  -6,   2,   4,  10,  -7,   5,   1,  -5,   5 },
      {   0,   0,   0,  -3, -17,  -7,  35,  30,   0,   0,   0,   3,  -8, -15,  19,  29,   0,   0,  -5, -13, -12,  15,  13, -11,   0,  -3,  -1, -10, -55, -38,  16, -18,   0,  -3,  -2, -12, -16,  22, -15,  14,  -2,  -7,  -5,  30,  25, -42,   8, -12,   6,  -6,   6,  11,   4,  20,  -8,   2,  -9,  -3, -12, -20,   4,  15,  -6,   0 },
      {   1,   2,  -1,   1,   6,   5,  45,  65,   0,   0,   1,   5,  -2,   0,   3,  -1,  -2,  -2,  -7, -18, -21,  -5, -12, -34,   0,   3,   4,   2,  18,  40, -41,  12,  -1,  -2,  -5, -21, -10,  10,   0,  -7,   1,  -2,  -8, -16,  -5,  -8,  12,  -4,  -3,  -8, -19, -25,  -5,   1,  10,   4,  -5,   2,  19,  28, -20,   0,  -8,   4 },
      {  -2,  -3,  -4, -14,  -8,  13, -17, -31,   2,  -1,  -2,   1,   6,   6,   0,  14,  -2,  -4,  -9, -19,   0,   9, -12,  13,   4,   6,  10,  20,  32, -22, -13,   4,  -1,  -5,  -8, -24, -11, -11,  19, -13,  -2,  -1,   4,  19,  -5, -47,   2,  -9,   0,  -6, -22, -51,  38,   8,  -8,   4, -12,  -5, -38,  16, -15,  13,   1,   2 },
      {   0,   1,  -1,   1,  -2,   2,  21, -19,   1,   0,   2,  12,  13,  35,  50, -29,   3,   2,   3,  10,  20,  37,  20, -28,  -2,  -2,   1,   8,  -2, -27, -46,  11,   0,   0,   3,  11,  13, -19, -25,  20,  -1,   2,   7,   9, -32,  -7,  16, -11,   3,   5,  14,  10,  18, -14,   2,   0,   2,   3,  11,   6,   7,  -6,  -8,  15 },
      {   0,   2,   2,   4,  -2,  -8,  -9,  -2,   1,   5,   4,   8,  18,  18, -12,  -9,   0,   1,  -6,  -2,   8,  11,   3,  19,   0,   3,  -1,   8,  24,  17, -13,   6,   0,   0,  -5,  16,  28,  11, -13,   3,   0,   1,   1,  43,  56, -18,  22,  -1,   3,  12,  16,  29, -29,  29,  -8,  -1,   7,   3,   8,  33, -43,  16,   3,  -7 },
      {   1,   1,   0,   4,   5,   5,  17,   0,   1,   3,   3,   1,   8,  24,  16, -43,  -1,  -3,  -2,   1,  -5,   5,   5, -37,   4,   3,   6,   1,   6,  39,  70, -15,   1,  -2,  -4,   3,   6, -22,   6,  21,   3,   2,  15,  27,  19,   0,   1,   1,   3,   7,  -3, -41,   2,   9,   3,   8,   0, -10, -12, -16,   4, -16,   3,  -1 },
      {  -1,  -1,   0,   1,  -5,   6,  10, -24,   0,   0,   0,   4,   9,   6,  -2, -40,   0,  -1,   2,   3, -15,   0,  18, -21,  -1,  -2,   0,   1, -20, -40,  -5,  36,   0,   1,   2, -24, -38,  33,  53, -19,   3,   2,   7,   6,  28,  27, -10,  15,  -2,   0,   2,  -4, -17,  -4,  -7,  18,  -1,   1,  14,  15, -13,  -9,   1,  -8 },
      {  -1,  -1,  -2,   0,  -5, -11,   6,   0,   0,  -1,  -5,  -3, -10,  -3,  28,  -3,   0,   2,   1,   4,   3,  -5,  25,  24,  -1,   0,  -3, -12, -22,  -5,  -3, -31,  -1,   4, -10,  -8,  19,  14, -25,  14,   3,   6,  -7, -12,  17,  47,  -3,  24,  -3,   2, -31, -18,  42,  -3,   9,  -8,  -2,  -2, -20,  36, -44, -11,  15, -13 },
      {  -1,  -1,   0,   0,   1,   3, -18, -22,   1,   2,   3,   4,   2,   1,   4,  27,  -2,  -3,  -2,   0,  -7,  -2, -19, -22,  -1,   0,   3,  -1,   1,   1, -56, -35,  -3,  -4,  -8, -19, -12,  -1, -18,  17,   2,   7,   6,  17,  32,  28,   0,  -7,   1,   2,   9, -31, -46,   3,  18,  -6,   0,   0, -16, -28,  17, -32,   7, -12 },
      {  -2,  -2,  -1,  -3,  -7, -15, -11,  -6,  -2,  -1,  -4,  -8,  -7, -12, -35, -62,  -3,  -2,  -3,  -4,   1, -24, -46, -17,  -1,   0,  -3, -13, -17, -18,  -2, -36,  -5,  -2, -11, -27, -21, -25, -24,   1,  -2,  -1,  -3,   0,  -5, -11,  35,  -4,   1,   5,   2,  21,  12,  -1,  -1,  -7,   0,   3,   1,  29,   6,   1,  -5,  -3 },
      {   0,   0,  -1,   0,   1,  -8, -14, -13,   0,   0,  -3,   0,  -3,  -1,  23,   8,   0,  -1,   2,   6,  -2, -10,  -4, -14,  -2,  -4,  -7,  -5, -12,  -1,   4, -43,   3,   3,   9,  19,  20,  32,  40,  22,   1,  -1,   7,  -8, -43, -29,   6,   0,  -4,  -4,  -8, -17, -38,  18, -31, -12,   3,   6,  18,  44,   3, -11,  16,  -9 },
      {   0,   0,   1,  -1,  -3,   2,  -5, -13,   0,   1,   1,   0,   0,  -1,   6,  12,  -1,   1,   1,  -3,  -5,  -2,  -7,  -6,   0,   1,  -2,  -3, -11,  -9,  17,  30,  -2,   1,   5,   1, -23, -39, -43,  34,   1,  -1,  -8,  -3,   4,  -9, -60,  17,  -5,  -8, -10, -12, -39,  -2,   8,   6,  -4,   5,   8,  50,   2,  27,  -6,   8 },
      {  -1,  -2,  -3,  -1,   5,   0, -10, -15,   0,   0,   0,  -2,  -3,  -8,  10,  35,  -1,  -3,  -2,  -4,  -1,   2, -26, -48,   0,   1,  -1,  -3,   8,  17,   1,   0,   1,  -1,  -3,  -2,  16,  15,  16,  20,   0,   1,  -2,   6,  35,   9,  14,  34,  -1,   6,  -2,  23,  34, -49, -27,  19,   3,   1,  -2,  18,  27,  29,  -5,  15 },
      {   0,  -1,   0,   0,  -1,   8,   1,  -6,  -1,   0,  -1,  -6,  -3,   6, -21, -25,  -1,   2,   2,   4,   4,   5, -16, -21,  -2,   1,  -4, -14,  -5,   5, -27,   5,  -1,   3,  -4,   4,  18,  26, -20,   3,   0,   0,  -2,  -8, -19,  -3, -30,  57,   1,  -2,   2, -12,   9,  54, -31,   9,   0,   1, -13, -42,  -9,  31,   9,   9 },
      {   0,   0,   2,   3,  -1,  -3,  17,  18,  -1,  -2,  -1,   4,  11,  -1, -22, -28,   1,   0,   0,  -2,   5,  14,  36,  29,  -1,   1,   4,   1,  10,   3, -22, -47,   1,   2,   0,  -3,   4,  -1,  15,  -8,  -1,   1,   2,  11,  11, -37, -33,  47,  -2,   3,  -3,  -1, -13, -45,  16,  -7,  -1,   0,  -7,   5,  36,   4,   2,   4 },
      {   0,   1,   1,   1,   0,  -4,  -5, -21,   1,   0,   0,  -1,  -2,   3,  22,  15,  -1,   0,  -1,   4,  -2,  -6,  -2, -27,   0,  -2,  -3,   2,  -5,  -1,  13, -32,  -1,  -2,   0,  14,  -8,  -7,   2, -50,   0,   0,   2,  -2, -20,  -7,  16,  30,   2,  -1,   8,   7, -24, -13,  57,  16,  -4,  -4,  -3, -12, -52,  32, -17,  13 },
      {   0,   0,   0,  -1,   0,   0,   7,  10,   0,   1,   1,  -2,  -1,   0,   6,  -3,   1,   2,   1,   0,   2,  -1,  17,  33,   1,   0,  -1,   2,   5,  -6,   2,  -1,  -1,  -1,   1,  -4, -18, -18,  -3,  15,   2,   2,   9,  -4,  -8,  29,  57,  20,  -4,   3,  -9, -28, -39, -15, -53,  -6,  -2,  -2, -10, -14,   0,  42, -41,   3 },
      {   0,   1,   1,   1,  -1,   0,   2,   3,  -1,   0,  -1,  -1,  -1,  -6, -21,  -7,   0,   0,  -1,   1,  -3,  -5,  -6, -11,  -2,  -2,  -1,  -3,   0,   1, -12,   4,   2,   3,   3,   4,  -2,  -9,  36,  68,   0,   1,   0,   0, -11, -11, -22, -14,   3,   2,   7,  15,   3, -38,   4, -28,  -5,  -6, -16, -30, -63,   7, -13, -26 },
      {   0,   0,  -1,  -1,  -3,  -3,   1,  -1,   0,   1,  -1,  -3,  -4,  -6,  11,   2,   0,   2,   0,  -2,   0,  -1,  -3,   3,   0,   2,  -6,  -3,   2,   2,   5,  26,  -2,  -3,  -4,  -3,   2,  -4, -31, -11,   0,   0,  -7,  -7, -10, -36,  18,  31,   0,  -1,  -1,   6, -17, -34, -25,  47,  -1,  -4,  -9, -12, -13, -44,  25, -65 },
      {  -1,  -1,  -1,   0,   3,   0,  -5,   0,  -1,  -1,   0,  -1,  -3,  -4,  -8,  -6,  -1,  -2,  -1,   1,  -4,   3,   5,  -1,  -1,   2,   6,   5,  -2,   6,  -2, -32,  -1,   0,   1,   1,  -6,   3, -19, -19,  -2,   0,   1,   6,  -1,   4, -35, -40,  -3,   1,  -5,   4,  -8, -30, -57,  40,  -3,   0,   5,  -8, -38, -21,   3,  59 },
      {   1,   1,   1,   2,   2,  -3,  -5,   2,  -1,  -2,  -2,  -1,   1,  -8, -12,   4,   1,   2,   2,   2,   3,  -1,  -4,  27,   0,   0,  -1,  -2,  -2,  -2,  -5,  -2,   1,   1,   3,   2,   3,   6,  18,  48,   1,  -2,   0,  -1,  -6,  -5,  14,  17,   0,   3,   2,   0,   7,  25,  32,  75,  -2,   4,   0,   6,   2, -36, -52,  26 },
      {  -1,  -1,   0,  -1,  -1,   1,  -3,  -5,   0,   1,   1,   3,   6,   5,   7,  -1,   0,   1,   1,   0,   0,   4,   0, -14,   0,  -1,  -2,   0,   4,   7,  -3, -22,   1,   2,   2,   6,   8,   6,  -7, -30,  -1,  -1,  -1,  -1,   3,   7, -38, -19,  -2,   0,  -3,   7,  13,  10, -20,   2,  -1,  -2,   1,   9,   9,   1, -88, -62 },
      {   0,   0,   0,  -1,  -1,  -2,  -3,  -3,   0,   0,  -1,   1,   0,  -5,   1,   9,  -1,  -1,  -2,  -3,  -4,  -4, -10,  -8,   0,  -1,  -2,   1,  -1,  -8,  13,  26,   0,   1,  -2,  -2,  10,   8, -12, -14,   1,   1,   3,   4,   5, -18,   4,  41,  -2,   3,  -2,   2,  -5,  -4, -11, -66,  -2,  -1, -10,   2, -13, -65, -43,  42 },
    },
    {
      {-117,  26,  10,   4,   3,   2,   2,   1,   2, -10,   1,   0,   0,   0,   0,   0,  40, -10,  -5,  -1,  -1,   0,   0,   0,  -2,   2,   1,  -1,   0,   0,   0,   0,  -2,   0,   2,   0,   0,   0,   0,   0,   2,  -2,   0,   1,   0,   0,   0,   0,   5,  -1,  -1,  -1,   0,   0,   0,   0,   3,   0,  -1,   0,   0,   0,   0,   0 },
      {  -2,   5,  -5,  -4,  -3,   0,   0,   0,-122,  22,  11,   6,   4,   2,   2,   0,   0, -16,   5,   2,   2,   0,   0,   0,   8,  -7,   1,   0,   1,   0,   0,   0,  14,  -1,   0,   0,  -1,  -1,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   6,  -3,  -1,   0,   0,   0,   0,   0 },
      {  36,  -4,   2,  -1,   0,   0,   0,  -1,  -2, -10,  -3,   0,   0,   0,   0,   0,  97, -28, -10,  -6,  -2,  -1,   0,   0,  47,   7, -14,  -1,  -4,  -1,  -2,   0, -34,  25,   8,   0,   1,   0,   0,  -1,  -8,   0,   4,   2,   1,   0,   0,   0,   2,   4,  -5,  -2,   0,   0,   0,   0,  -9,   4,   2,  -1,   0,   0,   0,   0 },
      {  18,   1,  -4,  -1,  -3,   0,   0,   0, -13,   2,   5,   2,   2,   0,   0,   0,  47,  20,  -7,  -3,  -4,  -1,  -1,  -1,-102,  25,  15,   4,   4,   1,   1,   0, -27, -26,   9,   3,   2,   0,   1,   0,  13, -11,  -4,   1,   1,   0,   0,   0,   3,  -1,  -1,   0,   0,   0,   0,   0,   6,  -1,  -2,   0,   0,   0,   0,   0 },
      {  11,   1,  -5,  -6,  -3,  -3,  -1,   0,   5, -25,  -7,  -5,   1,   2,   2,   0,  43,  10,  -5,   1,  -2,  -4,  -2,  -1, -14, -17,   6,   1,   3,   3,   1,   0, 103,  -8, -20,  -2,  -7,  -1,  -3,   0, -13,  37,  -3,  -1,  -1,   0,   1,   0, -12,   2,   3,   0,   0,   0,   0,   0,  -9,   5,   1,   1,   0,   0,   0,   0 },
      { -23,-108,  11,  14,  13,   4,   2,   1,  -6,  21,   6,   0,   1,   0,  -1,  -1,  13,  43, -10,  -7,  -8,  -4,   0,  -1,  14,  15,   4,  -1,  -1,  -1,   0,  -1,   5,  -5,  -1,   4,   4,   3,  -2,   0, -12,  -5,   2,  -4,  -1,   2,   0,   0,  -2,  -3,   4,   2,   0,   0,   0,   0,   3,  -5,  -2,   1,   0,   0,   0,   1 },
      {  -1,  16,  -3,   1,   0,  -3,   1,   1,  -5,  -3,  10,  -2,   1,  -1,   1,   0,  -5,  18,   3,   2,  -2,   0,   0,   0, -19,   4,  -1,   4,   3,   2,   1,   0, -12,  46,  -1,   0,  -1,  -1,   0,   0,-102,   5,  21,   2,   6,   0,   2,  -1, -15, -38,   4,   2,   3,   0,   0,   0,  16, -10,  -5,   0,  -2,   0,   0,   0 },
      {   7,   4,  32,  16,   9,  -1,  -1,  -1, -24, -88,  -7,   8,   9,   5,   2,   1, -13,  35, -16, -23, -12,  -3,   2,   1,   7, -15,  12,  -3,  -2,  -2,  -1,   0, -10,   1,  -2,   9,   4,  -1,  -1,  -1,   2, -12,   4,  -2,   3,  -1,   2,   0,  50, -11, -11,   0,  -5,   1,   0,  -1,  -7,  19,  -4,   1,   0,  -1,   1,   0 },
      {  -5, -45,  -6, -11,   0,   2,   1,   0,   8,  -7,  -4,   2,  -4,   0,   0,   1,  -8, -72,   7,  14,  10,   5,   1,   0, -37, -18,  -5,   6,   1,   2,   0,   0,  21,  40,  12,  -7,  -8,  -4,  -2,  -2,  -5, -24,   6,   3,   4,   1,  -2,  -1,  54,  -2, -20,   0,   1,  -1,   0,   0,   9,  16,  -2,  -5,  -3,   1,   1,  -1 },
      {   3,   2,  49,  18,   9,  -3,   3,   0,   2,  42, -18,  -9,  -2,  -3,  -3,  -2,  11,  10,  10, -23,  -8,   3,   3,   2, -20, -95,  -6,   5,   2,  -3,   0,   5, -15,   9,  -9,   4,   0,  -2,  -3,  -1,   2,   5,   2,   7,   4,   2,   0,  -2,  -8,   2,  -2,  -1,  -3,   1,   1,   1, -18,   4,   2,   0,   2,  -2,   1,   0 },
      {   0,  25, -26, -20, -11,  -4,   0,  -1,   5,  45, -10,  -5,  -4,   3,   2,  -3,  16,  54,  15,  14,   5,  -5,  -5,  -4,  18,   1,  -1,  -3,   5,   6,   1,   1,  17,  -4,   0,  -6,  -2,  -5,  -2,   1, -17, -50,   0,   0,   0,   0,   3,   2,  61,   9, -19,  -2,  -6,   0,  -1,  -1, -22,  22,   3,   1,   0,  -2,   0,   0 },
      { -12, -17, -56,   0,  -2,   0,   2,   7,  -9, -39,   4,  -2,   6,   7,   3,   1,  -1,  17,  63, -10,  -8,  -9,  -1,  -2, -10, -18, -18,  14,   8,   4,  -1,  -1, -16,  22,   8,  -3,  -4,  -1,   1,   0,   0,  -1,  -7,  10,   0,  -1,   3,   1, -27,  44,  -5,  -4,   3,  -3,   0,  -1, -40,  -8,  14,  -3,   0,   1,   1,  -1 },
      {  -1,   1, -60,  46,  19,  18,  -1,   4,  -1,  10, -68,  -4, -10,  -1,  -2,   2,   4, -10,  15, -26,  -7,  -4,  -1,  -4,   6,  -6,  25,  -4,  -3,  -6,  -1,  -3,  -8, -18,   3,  10,   3,   2,   0,   0,  -4,  26,   5,  -6,   1,  -1,  -1,   1,  20, -22,   3,  -5,  -2,  -2,  -1,   0,  31,   6,  -7,   2,  -1,   1,  -1,   1 },
      {  -2,   7,  30,  12,   2,   2,  -1,  -2, -12,  15, -67,   7,  -6,   6,   3,   1, -14,   7, -18, -10,  -2,   1,  -1,  -1, -16,  52,  -3,  -7,  -7,  -4,  -3,   0,  12,  56,  10,  -5,  -1,  -7,   0,  -1,  10,   9,   9,   2,   2,   0,  -2,  -1, -10,  24,   1,  -4,  -1,  -2,   1,  -1, -42,   4,  16,  -2,   3,  -1,   0,   0 },
      {   4,  12, -10,  76,  25,  15,  -1,  -2,   4,  16,  61,   2, -13,  -6,  -4,  -5,  -5, -18, -17,  -7,  -8,  -2,   0,  -1,  -7,  15, -13,   1,   1,  -2,   1,   2,   7,  -6,  -8,   1,   7,   6,   0,  -1, -15,  13,  -5, -12,   0,   2,  -2,   1,  22,  12,  -4,  -2,   0,   0,  -1,   0, -47,  15,  10,   1,   1,  -3,  -1,   1 },
      {  -3,   3, -42,  -9,  -2,   5,   4,   0,   2,   9,  30,   0,   0,  -4,  -2,   1,   2,  38, -36,   0,   1,   3,   2,   2,  -6, -17, -20,  -9,  -5,  -4,  -1,   3,  -5,  64,  -8,  -6,  -4,  -2,  -1,  -2,  54,  34,  18,   1,   2,  -4,  -1,  -1,  16, -21,  12,   5,   2,   2,  -1,   1,  18,   2,  -6,  -1,   2,   1,  -1,  -1 },
      {  -9, -15, -18, -35,  -8,   0,  -2,  -5,  -5,  -5, -14,  20, -15,   0,   1,  -2,  -5,  -7, -36,  40,  15,   5,  -5,  -1,  -3, -24,  35, -14,   5,   3,   3,  -1, -34, -19, -13,  -5,  -1,   2,   0,   2, -23,  43,  -1, -20,  -4,  -4,  -3,   2,  16,   7,  17,  -4,   0,  -1,  -2,   1, -54,  -3,   9,   3,   2,  -1,   1,   0 },
      {   1,  -4,  30,  11, -13,  -4,  -3,  -2,  -2,  -1,  -3,  14, -11,   5,   1,  -3,   8,   6,  66,  26,  15,   4,  -4,  -1,  -1,  24, -14,  26,   1,   7,   3,   0, -15,   3, -62, -20,  -1,   0,   3,   1,  15,  34,  13, -23,  -1,  -1,  -3,  -2,  20, -20,   8,   6,   1,  -2,  -1,   0,   6,  18,   0,  -1,   1,  -1,   0,   0 },
      {  -1,  -4, -16,  27, -38,  -3, -10,   0,   2,  -7, -18,  78, -10,   2,  -2,   0,   3,  -2, -16,  -1,   6,  -1,   3,   4,  -4, -16, -47, -17,  14,   1,  -1,   6,   8, -16,  -6, -14,   9,  -2,   0,   2,   1, -40,  22,  -6,  -4,   5,   0,  -1, -20, -20,  13,   3,  -1,   5,  -2,   1,  -7, -10,  -6,   2,   2,  -3,   1,   2 },
      {  -1,  -2,   0,  12,  11,   4,   8,   0, -10, -13, -25, -30, -13,  -2,   1,   4,  -5,  16, -30,  48,   7,   4,  -7,  -2,  -5,  -8, -74,  17,  -4,   1,   3,   0, -10, -26,  17, -19, -11,   0,   1,  -2, -14,  19,  -3,   4,   4,  -1,  -2,   1,  -2,  25, -23,  -8,   1,  -7,   1,   0,  28,  18,   6,  -7,  -3,   0,   0,  -3 },
      {  -5,  -2,  -4, -33, -22, -10,   1,   3,   5,  16,   2,  28,  38,   2,   1,   1,  -8,  -9, -11, -58, -11,  -7,  11,   5,  -6,  10, -36,   2, -12,  -1,  -6,   1,  -8, -13, -24,  26,   7,   4,   1,   0, -25,  34, -12,   1,  -6,  -3,   2,  -3,  22,  38,   0,  -1,  -2,  -2,   1,  -2,  18,  28,   3,  -1,   0,  -1,   1,   1 },
      {  -5,  -9, -15, -21,  -3, -15,   0,  -4,  -5,  -5,  -9, -63,   8,  -4,   5,   2,  -6, -16,  -6, -28,   3,   6,   4,   9,  -6,  10, -32,   0, -11,  -5,   0,   2,  -3, -25, -15,  -1,   2,  -3,  -3,  -2,   6, -13,  35,   1,  -5,   8,   1,  -4, -10, -52,  11,  15,  -6,   6,   2,   3, -49,  22,  -5,   2,   7,  -2,   2,   0 },
      {  -2,  -3,  11, -16,   0,   8,   6,  -3,   2,   5,  16,  32, -21,  -5,  -1,  -1,  -1,   7,  29,  -5,   5,  -3,  -6,  -3,   3,   4,   3, -26,   6, -10,  -4,   2,   2, -11,  62,   3,  -1,  -2,   0,   0,   3,  35,  16,  20,   3,   0,  -1,   1, -16, -28, -29,  -8,   5,   1,   1,   3, -10,  64,  -4, -13,  -2,  -3,  -2,   0 },
      {  -5,  -8,  -9,  37, -76, -12,  -5,   5,  -3,  -3,   1, -24,  -5,  -8,   1,   3,  -2,   2, -21,   2,  27,  10,   3,   5,   4,  -3,  34,   9,   2,   6,   3,   4,  -9,  13, -19,  11, -12, -10,   1,  -2,  -2,  -8, -27,  17,   8,   5,   2,  -5, -24,   6, -18,   4,   1,   1,   1,   0,   8,  46,   0, -11,   2,  -2,  -1,  -3 },
      {  -1,   2, -17, -17,  63,  -1,   3,  -1,  -1,   0,  -6,  26,  -1,   9,  -1,   0,  -1,  -4, -17,  17, -25,   4,  -3,   3,  -1,  -2,  14,  13,  -3,  -5,  -1,   3,  -4,  11, -50,   2,   2,   6,   2,   3,   2, -33, -12,   4,  -3,   1,  -1,   0, -40,   0, -10,   6,  -1,  -1,  -1,   0,   6,  59,  -4, -11,   3,  -3,   1,  -1 },
      {   2,  -2,   4,   6,  21,  -3,   2,  -9,  -3, -10,   3, -14, -10, -11,   9,   2,   4,   2,  21,   0,  40,  15,   1,  -1,  -9,   0, -13, -96, -13,  -8,   2,   5,  -3,   4, -24,   5,  -8,  -7,  -1,  -1,  -9, -11, -14,  11,  -1,  -2,  -1,  -2,   0,  19,  28, -10,  -9,   1,  -1,   0,  16,   9,   6,   7,   5,   3,  -1,   2 },
      {   1,   0,   9, -22, -23,  74,   5,   6,  -3,  -6,   0, -15, -59,  -8,   0,   3,  -2,  -3,  -1,  -1, -14, -39, -12,   3,  -4,   1, -15,   2,   4,  -1,  -8,  -3,   1,   7, -21,  43,  21,   9,   3,   0,   0, -11,  -8,   1,  -7,   3,   7,   4,  -3,  -2,   4, -17,  -6,   2,  -4,   1,   1,  -1,  -4,   8,   3,  -3,   0,   2 },
      {   0,   0,  -5,  -2,  12,  22,   9,  -5,   1,   1,  -3,  24,  12,   4, -12,  -5,   2,   7,   2,   1,  45,  22,   2,  -3,  -1,   1, -10,  16, -60,  -8,  -1,   5,   2,  -6, -13,  41, -23,  -5,  -3,  -5,   2,   1, -20,  17,  17,  -4,   0,   2,  -2, -27, -45,  -3,   6,  -4,   3,   2, -21, -35,  -4,  -7,   3,   4,  -1,  -4 },
      {   0,   3,  -4, -13,   8,   0,  -7, -10,  -2,  -5,  11,   0, -52, -12,  -6,   0,   1,   6,  -6, -39,  31,  30,  13,   0,   2,  -3,  18,  18, -13,   4,   3,  -2,   3,  -8,  -6, -10,   4,   4,   3,  -2,  -3,  -9,  66, -27,   6,   4, -15,  -3,  -8,  41, -20,   0,   7,  -3,   0,   1,  17,  -4,  13, -11,  -1,   1,  -1,  -2 },
      {  -1,  -2,   0, -10,  10,  -3,   1,   7,  -1,   3,  -1,  -3, -24,  15,   4,  -6,  -4,  -5, -13, -35,   1, -15,  -2,  -8,   1,   9,  -4, -22,  54,  24,  10,  -2,  -8,  -5, -32, -39, -20,  -6,   5,   1,   1,  12, -19,  31,   1, -12,  -2,  -3,   2, -15, -57,   2,   7,   2,  -1,  -2,  -3, -29,  -3,  -8,  -9,  -5,   0,  -2 },
      {   2,   1,   5,   3,   1,  66,  11,   7,   2,   1,   0,  -3,  63,  -6,  12,  -7,   1,   0,  -6,  -8,  44, -31, -14,  -1,  -1,  -7,  13,  -1,   0,  12,  -5,  -6,  -1,  -1,  -1, -43,  -1,   7,   7,  -2,  -1,  -9,  18, -14, -19,  -9,   1,   3, -12,   0,  -2,  -1,   3,  -6,  -2,   0,   2,  14,  19,  -8,  -1,   4,  -2,  -1 },
      {   0,   0,   2,  13, -14,   6, -14,   5,   2,   4,   2,  -9,  26,  29,  -6,   0,  -1,  -6,   4,  35, -35, -21,   6,   0,  -2,   0,  10, -39, -14,  11,  -2,   0,  -4,  -7, -18,  27, -15,   7,  -3,  -3,   3,   6,  66,  12,   6,   0,   1,  -1,  -4,  25, -34,  12,   5,   5,   3,  -2,   0,  -7, -30,  -5,   2,  -7,   0,   1 },
      {  -1,   1,  -1,   2,  19, -33, -13,  -3,  -1,   0,  -7,  -1,  14, -48,   2,  -2,  -1,   2,  -1,   5,  32, -32,   5,  -9,  -3,   0, -10,  -4,  26,  43,   6, -11,   0,  13,   6,  43,  10,   8,  -3,   5,   8,   2,  -5, -52,  -1,  12,   2,   3, -12,  -3, -28, -11, -11,   8,   2,  -2,  -4,   1, -23,   6,  -5, -10,   4,   0 },
      {   2,   1,   3,   7,  18, -12,  29,  25,   2,   1,   2,   9,   1,   4,  12,   6,   0,  -2,  -7,   1,  34,   0, -33,   1,   5,   5,   1,  27,  48,  -2, -13,   4,   0,  -7,  -2,  46, -36, -16,   3,  -3,   3,  -1,  37,  40,  -9, -12,  12,  -4,   8,  11,  30,   1,  -9,  -5,  -1,   0,   1,   2,   4,  12,   6,   0,   3,   0 },
      {   0,   0,   0,  -2,  -3,  10,  -2,   7,   2,   2,   6, -11,   0,  90,  -7,  -9,   0,  -1,  -3, -10,   7,  23,  -1, -21,  -1,  -4,  -8,  -7,  15,  17,   1,  -5,   0,   3,  12,  20, -13, -19,  -7,   0,   2,  -5,  -9, -56,  13, -10,   5,   3, -12, -10,   6, -23,  -5,  -1,   4,  -2,   5,  13,  12,   8, -10,  -7,   3,   3 },
      {   1,  -1,  -2,   4, -15, -25,  83,   3,   1,  -1,   1,   5,  -4, -20,   1,  -1,  -1,  -6,   0,   1, -35,  -6, -28,  -1,  -1,  -2,   5, -18, -27,  28,  -3, -15,   1,  -3,  -8,   1,  -4,  -7,   7,  -6,   4,  -6,   9, -20,  11,  -9,   6,   2,  -3, -12, -13, -13,   3,  -1,  -1,  -2,   5,  -5,  50,   1,  -5,   1,   0,  -3 },
      {   1,   2,   1,   2,  -2,  -2,  55,  -5,  -1,  -1,   0,   3, -19,  16,   1,   2,  -1,   0,   0, -16,  12, -16, -11,   7,  -1,   0,   2,   3, -22,   7,  -3,  -3,  -2,   3,  10, -25, -19,  -1,  -4,   1,  -6,  -1, -14, -19,  -3,   1,   3,   3,  -2,  20,  13,  30,  -6,   2,  -1,   0,  -9,   4, -89,   2,   6,   1,   3,  -1 },
      {   0,   0,   1,  -4,  -7,  34,  20,  12,  -1,  -2,  -5,  -4,  25, -33, -14, -18,   0,   2,   5,   4, -22,  81, -15,  -5,   2,   5,  -3,  -7,  35,   6,   6,   6,   2,   0,  -6,   4,  30, -25,   3,  -1,   0,   2,   3,  -5,  -3,  -4,  -5,  -5,  -1,   8, -10, -14,   2,  -4,  -2,  -5,  -8,  -3, -38,  -5,   0,   2,   3,  -2 },
      {  -1,  -1,  -4,   4,  -6, -29,  18,   3,   1,   1,   2,  -7,   5,  28,  25,  -3,   0,  -1,  -3,   5,  23,  -8, -31,  -9,  -3,  -3,   4,  -2,   6, -65, -22,   3,   1,   5, -11,  -3,  65,  12,   3,   4,   0,   6,   6,  -4, -11,  11,   1,   8,  -3,   3, -36, -19, -16,   5,  -1,   2,   3, -12,  -7,   2,   2,   7,   2,  -1 },
      {   0,   1,   2,   1,  -1,  -2, -25,  40,  -1,   0,   1,   3, -13,   3,  19,  17,   0,   2,   1,  -3,  -8,  26, -38,  -7,  -2,  -5,  -3, -10, -20,  17,   1,  -3,  -1,   5,  10,   9,  -3,   4,  -2,  -4,  -2,   3, -11, -14, -72,   8,   1,   6,  -1,  -4, -20,  56,  -1, -14,  -2,  -2,   4,  -2,  14,  11,   9,   5,   2,  -1 },
      {   0,   0,   0,   4,  -6,  -6,  12, -51,  -2,  -1,  -4,  -6,   7,  17, -68,  20,   0,   2,   2,  -4,   7,  13, -22,  19,  -1,   1,  -1,  -2,  16,  11,  12,  -4,   1,   6,   1,   6,  -4,  61,   4,   3,   0,  -1,   0,  11, -44,   5,  -3,   7,   3,  -3,   1, -18,   3,  -1,   0,   1,   1,   0,   4,  -7,   7,  -1,  -4,  -7 },
      {   1,   2,   2,  -2,   8,  -5,  16,  69,  -1,  -1,  -4,  -3,  -2,  13, -55,  14,   2,   1,   1,   7,  25,  -8,  24,  -2,  -1,   0,  -2,  -8,   0,  12,   6, -12,   3,  -1,   3,   2,  49,  16,  15, -10,  -1,   1,  -4,  12,  27,  -7,  -4,  -8,   4,   0,  12,  35,   7,   9,  -6, -11,  -4,   4,   8, -17,  -2,  -2,   0,  -2 },
      {   0,  -1,  -2,   6,   4,  -1,   3, -50,   1,   1,  -2,  -1,  -4,  24,  47, -16,   1,   0,  -1,   4,  10,   5,  -3,  -1,   0,  -4,  -2,   7,  -8,  58,   8,   1,   2,   1,   7,  17,  55,  -1,  10,  -4,   1,   2,   0,  33,   2, -31, -21,  -2,   3,  -1,  12,  22,  14,  -3,  -6,  -9,   1,   2,   0,  -9,   2,   3,  -4,   2 },
      {   1,   2,  -1,   3,  -1,  15,  43, -25,   0,   0,  -4,   1,   2,   4,  12,  15,   3,   2,   4,  13,   3,   5,  77,  21,   0,   2,   0,  -3,  33, -27,   8,  -3,   0,   1,   2,  16,   1,  -7, -17,  -8,   0,   3,  -2, -15, -26,  -6,   9,  -5,  -3,  -1, -21,  41,  -4,  -8,   2,  -1,   2,  -1,  20,  21,   6,   4,   0,   2 },
      {   0,  -1,  -1,   4,   4,  -8,  -2,  13,   0,   0,   4,   2,  -6,   2, -17,  17,  -1,  -1,  -2,   4,   5,   1,  33,  -6,  -1,   1,   8,  11, -27,  35, -24,  -3,  -2,  -1,  -3, -22,  20, -30, -27, -11,  -1,   4,   9,  28, -46, -19,  21,   5,  -4,  -1,  -5, -50, -27,  15,   7,  -1,   0,   4, -11,  37,  -7,  -1,   2,   8 },
      {  -1,  -1,  -3,  -4,   3,   7,  -5, -30,  -1,  -4,   0,   2,   2,   1, -46, -46,  -1,  -2,   0,  -1,   4, -11, -25,   1,   0,   0,   2,   1,   1,   7,  -5,  14,  -3,  -1,   2,  -6,  14, -39,  -6,   1,  -1,   8,   1,  11,  15,  39,   8,   8,  -2,   7,  -6,  47, -43,   5,   5,   9,  10,   1,  20,  35,  -3, -15,   2,   5 },
      {   0,   0,   0,   0,  -6,   2,  13,  20,  -1,  -2,  -4,   1,  -5,  -2,  22, -64,   0,   1,  -1,   2,  -2,  24,  29, -29,  -1,   1,  -1,   2,  -1,  13, -23,   4,   1,   3,   3,  -3, -17,  78,   5,   8,   2,   0,   0,  14,   2,  12,  10,   0,   0,  -2,   3,   0, -29,   6,  -1,   3,   0,   3,   1,  19,  -4,  -3,  -7,   6 },
      {   0,   0,  -1,  -1,   6,  12,  -7,  -7,   1,   3,   1,   2,   5,   6,  21,  53,   1,   2,   2,  -1,   1,  26,  -8,  47,   1,   0,   2,  -5,   9,  28, -67,   0,   0,  -2,   1,  -9,   3,  16,  12,  -4,   0,   6,  -7,  -4,  39,  33,  26,  -6,   3,   0,  -8,   7,  -7,  16,   1,   6,   0,  -4,   3,   8,  -1,  -4,   0,   0 },
      {   1,   1,  -1,   0,   3,   9,  13,  21,   1,   1,   3,   7,   4,  14,  29,  14,   0,   0,   1,   1,   1,   6,   2,  36,   2,   2,   3,   4,  -7,  12,  70,   3,  -1,   1,   3,   1,   2,   2, -16,  13,   1,   6,   3,  12,  -3,  48, -20,   4,   3,   1,   4, -18, -49,  26,   0,   7,   0,  -5,   4, -23, -17, -14,  -1,  -5 },
      {   1,   2,  -1,   1,   6,  -8,   7,   9,  -1,  -2,  -2,  -4,  -7,   7,  -4, -43,   0,   1,   1,   3,   1, -11,  10,  27,   1,   2,   1,  -4,   4,  13, -38,  34,  -1,  -2,   2,   5,   1, -15, -38, -23,  -1,   2,  -5,  -8, -28,  31,   8, -12,   6,  -3,   9, -10,  20,  -7,  -1,  -5,   4,  -9,   6, -70,  18,   5,  -2,  -4 },
      {   0,  -2,   0,   2,   1,  -6, -16,   2,   0,   1,   2,  -1,  -5,   2,  -3, -30,   1,   1,   2,   2,  -4,   1,   9,  48,  -1,  -2,  -1,  -1,  -4,  -7,  -2, -37,   3,   0,   5,   1,  -7, -16,  70,  -7,  -1,   1,  -1,  -9, -12, -36,  20,   4,  -1,  -2,  -5,  -7, -49,  -7,  -6,   5,   1,  -5,  -3, -25,  27,  -2,  -3, -13 },
      {   0,   1,   0,   1,  -2,   6,   0, -20,   1,   1,   3,   4,   1,  -2,   3,  12,   0,   0,  -2,   1,  -2,   7,   5, -52,   0,  -2,  -3,   1,   3,  -5, -18, -89,  -1,  -2,  -6,   1,  -1, -11, -19,   1,  -1,   0,   0,  16,  -6,  30,   8,  -6,   3,   0,  13,  10, -15,  -2,  11,   0,  -2,  -3,   0, -43,  -6, -11,   3, -12 },
      {   0,  -1,   0,  -2,  -2,  -1,  -2,  -5,   1,   0,   2,   1,  -1,  -5,   2,  26,  -1,  -3,  -1,  -3,  -2,  -1,  -2, -24,   0,   1,   2,  -2,  -1,   8,  22,  37,   0,   0,   0,  -6,  14,  14, -14,  -1,   2,   1,   5,  -1,  19, -15,  59,  -5,   0,  -2,   2,   3, -40, -75,  17,  12,  -1,  -4,  -2, -27,   4,  -4,  -2,   7 },
      {   0,   1,   0,  -1,   4,  -4,  -4,  11,   0,   1,   0,   4,   0,   9,   9, -24,   0,   0,  -2,  -2,   6,   2,  -4,  28,   0,  -1,   0,  -1,  -1,   4,  35, -44,  -1,   1,  -2,  -6,  12,   5,  -7, -33,   3,   1,   2,  11,  10,  37,  37,   4,   1,   1,  -1, -19,  51, -38,   3,  15,   1,   6,  -7,  41,  26,  -4,   1,   3 },
      {   0,  -1,   1,   0,   3,   0,  14,  -7,   0,   0,   0,  -1,   2,   8,   1,  17,   0,   0,  -3,   1,   0,  -3,  26, -45,   1,  -1,   1,   2,  -4,  10,   7,  31,   0,   0,   0,   2,  -1, -23,  77,  16,   0,   3,   2,   4, -21,  56,  19,   8,   4,  -1,   4, -10,  22,   4, -15,   6,   0,  -1,  -2,  -4,  18,   2,  -7,   1 },
      {   0,   0,   1,   2,   0,  -1,   2,  10,   0,   0,  -1,  -3,   0,  -1,  -3,  -3,   0,  -1,  -1,  -2,   3,   5,   6,  11,  -1,   0,  -1,  -1,   1,   8,  -5, -10,   0,   0,   0,  -1,   5,  -6, -29,  91,   0,   2,   4,   3,   9, -12,  17,  58,   2,   1,  -2,   4,  14,   2, -15,   7,   3,   0,   6,  -8,  44,  -1, -14,   4 },
      {   0,  -1,  -1,  -2,   0,   4,   2,   1,   1,   0,   2,   2,   3,   1,   2,  12,   0,  -2,   1,  -1,  -3,  -3,  -7, -19,   1,   0,   1,   4,   1,   3,   5,   4,   0,  -1,   5,  -8,  -2,   3, -12,  -7,  -2,   1,  -5,  -4,  10,  -2, -24, -52,  -1,  -2,  -7,  -8, -17,  16,  13, -21,   5,  -2,   4,  12,  97, -25, -21,   1 },
      {   0,   0,   2,  -1,  -2,   0,  -4,  -1,   0,   1,   1,   2,   3,   0,  -3,  10,  -1,  -1,  -1,  -2,  -5,  -1,   1, -17,   1,   1,  -2,   1,  11,  10,   4,  -3,  -1,  -1,  -2,   1,  -5,   3, -11, -50,  -1,   1,   1,  -1,  17,   5, -19,  49,  -2,   0,  -2,   1, -25,  10, -23,  25,   0,   3,   0,  -5,  32,  82,  -9,  12 },
      {   0,   0,   0,   0,   3,   3,  -6, -10,   0,   0,  -2,   1,  -3,  -2,   1,  -9,   0,  -1,   0,   0,   2,   4, -15,  -6,   0,   0,  -1,   3,  -3,  -5,  30,   0,  -1,  -1,   1,   1,   3,   4,  -9,   6,   0,   2,   2,  -6,  -2, -16,  77, -39,   0,   6,  -2,  11,  13,  72,  -4,   0,   2,   1,   4, -10,  -4,  38,  -4,   4 },
      {   0,   0,   0,   0,   1,  -1,  -2,   2,   0,   1,   3,   1,  -1,   8,  -8,  -4,   0,   0,  -2,   1,  -1,  -8,   2,  11,   0,  -1,   0,  -2,  -3,   9,  -6, -12,  -1,   0,   1,   1,   0,  -6,  -3,  47,   2,   1,   2,   3,   2,  20, -14, -60,  -2,  -4,  -4,  -9, -14, -42, -32, -27,  -3,   3,  -2,  18, -10,  69,   1, -21 },
      {   0,   0,   1,   0,   0,  -1,  -4,  -4,   0,   1,   0,  -1,   3,  -1,   2,   5,   0,  -1,   1,  -1,  -1,   2,   0, -10,   0,   1,   1,   1,  -2,  -5,   0,   6,   0,  -1,   1,  -1,   0,   1, -18, -21,   0,   0,   1,  -2,   3,  -8,  13,   0,  -1,   1,   0,   0,  -3,   0,-101,  10,   1,  -1,   0,   5,  -3, -39, -17, -56 },
      {   0,   0,   0,   0,   0,  -3,   1,  -2,   0,   0,   0,  -2,   1,  -1,  -3,   2,   0,   1,   0,   1,   2,   1,  -5,   7,   0,  -1,  -1,   2,  -2,  -6,  -2, -14,   0,   1,  -1,   1,  -4,  -4,   1,  -1,   0,   0,  -1,  -1,  -5,   2,  -3, -16,  -1,   0,  -4,   0,  -3,  -9, -45,  -5,  -2,  -2,  -2,  -9, -16, -15, -49, 103 },
      {   0,  -1,   0,  -2,   0,   1,  -4,  -5,   0,   1,   0,   1,  -1,  -3,   4,   1,   0,  -1,   0,  -2,   0,   2,  -2,   2,   0,   0,   0,   1,   0,  -6,   7,   3,   0,  -1,  -1,  -2,  -5,   1,   1, -17,   1,   0,   3,   1,   7,  10,  20,  41,  -2,  -1,   1,   0,  -6,  -1,   3,-113,  -1,   1,  -1,   2,  -6,   3, -24,  -9 },
      {   0,   0,   0,   0,   1,  -2,   0,   5,   0,  -1,  -1,   1,   0,   1,   0,  -1,   0,   0,  -1,   1,   0,  -2,  -1,   2,   0,  -1,  -2,  -1,   3,   1,  -4,   0,   0,   0,  -2,   4,   4,  -6,   1,   0,  -1,  -2,   0,  -3,  -3,  -1,  -5,  -3,   0,   2,   2,   3,   6,  -2,  37,  26,  -3,   1,  -5,   7, -21,   9,-110, -37 },
    },
    {
      {  27, -23,  -3,  -5,  20,   3,   2,   2, -29, -22,  12,  16,   5,   0,   2,  -1,  97, -26,   0,  16, -21,  -3,  -5,  -2,  -6,  12,   1,   5,  -3,   1,  -1,   0, -28,   6,  -9,  -6,   2,   1,   2,   0,   6,  14,  -5,  -4,  -1,  -1,   0,   0, -23,   3,   3,  -3,   4,   0,   1,   0,   5,   0,  -1,  -2,   0,   0,   0,   0 },
      {  86, -39,  -2, -22, -22, -27,  21,  35,  -3,   2, -27,   4,  -5,  -5,  20,  18, -15,   5,  -1,  -4,  13,   3,  -1,  -9,   6,  15,  -2,  -8,  -3,  -2,  -7,  -5,  26, -13,  -7,   3,   3,   3,  -6,  -3, -12,   0,   3,  -2,   0,  -6,  -3,   1,  -1, -13,  -1,   2,  -8,  -2,  -3,   5,   4,  -7,   5,   2,   2,  -1,  -1,   0 },
      {  25,  31, -27,   5,  -2,  -6,  13,  13,  29,  17,   1,   7,  -5,   1,   2,   9, -18,  -3,  12,   2,  -5,   2,   0,  -3, -45,   0,  -1,  -1,   0,   2,  -5,  -2, -91,  -5,  -1,   9,   9,   0,  -3,   2, -15,   0,  -3,  -5,   2,  -5,  -1,   0,  11,   6,   1,  -8,  -3,  -1,  -1,   2,  29,  -1,  -1,  -5,  -1,   0,  -1,  -1 },
      {  49,  15,   4,  20,  12,  10, -24, -30, -58,  17,  30,  -6,  -1,   1, -10, -13, -23,  -7,  -5, -19,  -6,  -1,   6,   5, -57,   6,  -1,   2,  12,   2,   6,   4,  18, -14,  -2,   0,  -3,  -1,   2,   2,  36,  -8,  -2,   5,  -1,   6,  -1,  -1,  10,  -2,  -1,   4,   5,   2,   0,  -1,  -2,   5,  -3,  -1,  -3,   1,   2,  -1 },
      {  28,  -7,  22,  22,  15,   5, -26, -28,  63,  29,   3,  -7, -10,   0,  -5, -19,  32,  20,  -7, -20,  -8,  -5,   2,  10, -17,  15, -13, -12,  13,  11,   2,   0,  14,  15,   0,   1,  -4,   8,   1,   0, -49, -14,   9,  13,  16,   4,  -5,  -3,  -7,  -6,  -1,  -3,   0,   3,   0,   6, -23,  -4,   8,  -4,   2,   5,   4,   1 },
      {  10, -16,  33,  -6, -13,   4, -12, -18, -23,  12,  -7, -14,   4,   3,  -3, -10, -11,   4, -28,  -3,  17,   1,   1,   4,  49,  45,  -4,   8,   0,  -5,  -3,   3, -64,   1,  31,  -3,  -3,  -5,   3,   0,  14, -36,   1,   6,   1,   7,   1,   0,   0,  -7,  -3,   9,   2,   2,   2,   0, -21,  -7,  -4,   3,   4,   1,   0,  -2 },
      { -28,   1,  15,   2, -36, -11,  -3,  18,  19, -28, -10, -11,  -2, -10,   1,  16,  26,   9, -14, -14,  13,  10,   8,  -7, -61,  59,  12,   4,  -9, -17,   3,   0,  12,  -3,  17,   1,   3,  -5,   5,   3,  14,  -2,   7,  -9, -17,   4,   9,   4,  21,  -8,  -4,  14,   1,   1,   4, -12,   0,  13, -12,  13,   1,  -2,  -4,   2 },
      {   4, -34,   3,  -4,  -7,  12,  -7,  -8,  50,  17, -22,  -1,  -3,  10, -11,   6,   1,  -1,  -2,   7,   5,   4,   0,   1, -10, -16,   4,   6,   7,   6,   2,   1,  -7, -43,  -1,   3, -12,   1,  -5,  -1,  73,  12, -12,  10,  -6,   7,   2,  -2, -31,  36,   9, -10,   0,  -1,   3,   0, -14,   9,   8,  -3,   1,  -3,  -3,  -1 },
      {   6, -24,  37, -16,   9,   9,   9,  -4,  15, -16,  -6,  12, -10,  11, -17,   3, -20,   7, -28,  26, -13, -15, -14,  -1, -25, -30,  28,  12,   8,  -3,  14,   8,  -7,  44,  -6, -25,   2,  11,  -1,  -8,  15,  -8,  14,  11,   3,   6,  10,  -4,  -4, -41, -13,  -2,   9,  -6,   9,   2,  27, -14, -22,   5,  -9,  -5,  -8,  -2 },
      {  50,  48, -13,  -4, -15,  16,   3,  -9,  29, -28,  27, -15,  11,  -7, -11,  -4,  -2,  15, -14,  25,  13,   4,  -5,   1,  22,  -4,   7,  30, -13, -10,   9,   0,  10,  32,  19,  -5,   8,  -7,   4,   1,   5,  25, -12, -16, -14,   0,   2,   1, -12,  22,  -3,  15,   6,   0,   4, -10, -18,  22,  -9,   1,   3,   3,   0,   7 },
      {   8, -41,   5,  12,   6,   8, -13,   0,  33, -54,  38,   9, -11,   2,  -5,  -7, -24, -50,  25,   0, -12,  -7,  15,  -1,   8,   0, -26,   6, -12,   8,  -3,   8,  -3,  -9,   5,  11,  -5,   2,  14,   4,   1, -18, -19, -10,  -9,  -5,  -5,   0,  37, -11,  -8,   7,   3,   7,  -4,  -9,  -1,   1,   4,  -4,   7,   8,   4,  -3 },
      { -15, -31,  15, -10,   4,  13,  -7,  -2, -18,  -8,   4,  15,   0,   2,   2,  -2, -37,  -3,  -8,  10,  15,   8,   5,   5, -38,  11, -29,  14, -12,   6, -11,  -9,  -6,  15,  15,   9,  -3, -10,   9,   4, -35,  35,  11, -23,  -4, -10,  -1,  -3, -61,  19,  20,  -2,  11,  -1,  -3,  -3, -13, -14,   7,  -2,  10,   5,   3,   1 },
      {   6, -24, -21,  28,  -1,   1, -25,  -5,  -6,  -7,   4, -15,   5,  -3, -16,  19,  -3,  37,  50, -22,  -8,  -1,  15,  -3,  13,  -7,  35, -10,   7,   2,  -4,  17, -15,  -4,  14, -13, -22,   4,   8,   9, -12,  21,  -6,  -4, -27,   5,  23,   9, -27, -26,   2,  19,  -7,  -8,  11, -12,   6,  -9, -12,  28,   2,   2,  -3,  -2 },
      {  -2, -16,  -3, -23, -17,   6,  16,  -1, -18,   6,  -7,  13,  -3, -10,  -9,   1,  15,  19,  21,  12,   2,   1,  -5, -13, -30, -50,   5,   8,  -5,   8,   7,   3, -10,   3,  12,  -3,   6,   4,   0,  -4, -17, -40,  -7, -14,  -6,  15,   2,   4,  26,  24, -11,  28,  -4,   0,  -2,  -1, -62,  -7,  10,  11,   9,   3,   0,   5 },
      {  -7, -20, -42, -10, -13,   8,   9,  -5,  -5,  -6,  10, -31,  21,  28, -15, -19,   6,   6, -13,  29,  20, -10,   0,   6, -17,   5,  14,  21,  21,  -6,  -7,  14,  11, -21,  26,  20,  14,  -2,  -5,  -7, -26, -10, -44,  28,  12,  -8,   6,  -1,   0, -26,  12, -26,  -2,   1,   6,   6,   3,  -8,   6, -15,  -4,  -1,   0,  -4 },
      { -11,  -6, -39,  -6,   7,   1,  19,  10,  13,  25,  36,  -5, -10,   4,  -3,  -4,   1,  11,  11,  -3,   4, -10, -17,   7,   1,  30, -12,  -1,  -8,  15,  17,  -4,   8,  18,   0,  -8,  28,  13,   9,  -8,  43, -12,  11, -38,   8, -16, -15,   1, -15, -25,   1,  12, -16,  -7,   8,   7, -11, -46,  12,   9,  -4,   6,   6,   4 },
      {  -1,  29,  -1, -11,  23, -26,   7,  10,   9, -18,   0,  27,  15,  -9,   2,  -3, -20, -10,  31,   6, -17, -32,  11, -17,  -5,  33,  25,  -3,   7, -20,  -4,   4,   3,   2,  -9,  -3,   2,   0,  -7,   0,   7, -27,  -7,  35,   2,   1,   2,  -3, -42,  -3,  -9,  -6,  24,   5,  -4,  -5, -42,  -5,  10,  -5,  12, -14,  -3,  -8 },
      {   1,   9,  12, -12,  22,  -8,  -9,  -4,   4, -44,   6,   0,  27, -11,  28, -19,   6,  67,   0,   2,  24,   1,  -9,   8, -10, -15, -21,  -9, -19,   6,  -5, -12, -10, -36, -18,  -9, -15, -16,  15,   7,  13, -11, -11,  -2,  17,  -9, -12,  -1,  -3, -14, -14,   3,  10,   2, -13,   5,  13, -17,  -3,  -8,   0,   9,   7,   1 },
      {   0, -28, -37,   2,  27,   5, -21,  -4,   3,  11, -20,  14,  40,   9,  -1,  -3, -18,  10,   3,  35, -21,  39, -16,   8,  -4,  28,   8, -13,  -8,  -9,  -4, -10,   4,  23, -13, -17, -10, -10,   4,  12,   3, -17,   2, -24,  10,   3,   8,   7,  19,  -2,  -1,  -8,  -2,  -8,  -4,  -4,  -9,  47,  -2,  -6,  -2,  -6,   8,   5 },
      {   0, -17, -10, -41,  20,  -5,  13,   4,   2,  12,  20,   7,  15,  17, -14,   0,   3,  25,  -5, -37,   9, -35,  10,  -8,   2,  11,  -8,   1,   7,   2,  -3,  11,  -5,  32,  -1,   5, -33, -20,  -2,  -4,   9,  34,   4,   4, -21,   1,  10,   3,  42,  20,  -6, -29,  13,   9, -18,   5,  -6,  -1,   3,   2,   5,  -6,   1, -18 },
      {   8,  23,  -9, -23, -16,  12,  -9,   5,   6,  17, -12,   8,  11,  11, -26,  -6,  17, -22,  30,  -3,  21,  15,  -6,  -5,   2,  -8, -24,  31,  -5,  -7,  12,  -4,  14,  -6,   3,  10, -43, -18,  29,  11,  -3, -40,  35,  -4, -11,   5,  11,   2, -14, -19,   3, -18,  16,  -1,   4,  -9,  23, -15,  -9,   2,   5, -12,  -7,   4 },
      {   4,  15,  -2, -28,   9,  16,  -5, -12,  12,  18, -22,   9,  23, -11,  20, -27,   4, -29,  -5,  11,  -4, -17,  24,   2,  -6, -15,   8, -22, -16,  -4, -11,  -4,  -1, -21,  43,  -7,   5,  -4,  11,  -2,   6,  40,  21,  -3,  -5,  -6,  -3,   8,  15, -50,   7,  23, -10,  -1,  -7,   5, -17,   1,   3,   6,   6,  -9,  10, -13 },
      {  -9, -11,  20, -33,  14, -16,  10,   0,   6,  36,  41, -11,  -3,  -7, -18,  -4,  -5,   0, -16,  -7, -21,  13, -14, -14,   6,   3,  28,  22, -22, -24,  24, -10,  -4, -43, -33,  11,   0,  -1,  13,   6, -25,  10, -13, -10, -12,  -3, -10,  12,  -7, -16,   0,  19,   7,   8,  -4, -15,  -1,  20,  -4,   0,  -6,   6,  -4,  -3 },
      {  -2, -11, -18,  19, -24,  15,   2, -22,   2,  15, -19,  41,   3, -23, -12,  -2,   3,  -9,  -1, -24,  31,   3,   6,   5,   2,  -1,   6,   1, -26, -16,   2,   2,  -2,  28, -36,  26,   4, -20,  -2,  -4,   9,  15, -43,   8,  27,  11,   2,  -2,  -2, -22,  -4,  26,  15,  19,  -9,   2,   0, -12, -15, -11,   5,  -2,   0,   4 },
      {   2,   2,  -2,  29,   2,   1,  24,  -6, -11, -25, -22,  13, -15,  32, -16, -12,  -4,   5,   7,  -2,  19, -11, -14,  15,   1,  11,  23,   9,   0,  19,  40, -11, -13, -19, -13,  13, -22,  -3,   4, -14,  -8,  25,  33,  -4,  -4, -12, -15, -18,   8, -20, -17,  -7, -15,  12,  17,  -2, -30,  21,  19, -19,   3,  -2,   1,  21 },
      {   0,  14,  12,   2,  -8, -29,  -7,  14,   4, -14,  10, -30,  23,  20,   2,   5,   1, -19, -13, -16,   8,  37,  -2,  -1,  -8, -24, -19,  -4,   5,   6, -12,  12, -16,  22, -41,   3,  -6,  -5,  -5,  -9,  12,   8,  -1,   9,  -3,   0,  22,   9,  -6, -35,  30,  -4, -27, -13,  -6,  -9, -44,  -4,  18,   0,  -5,  -6,  -1,  -6 },
      {   3,  27,  -6, -16,  -2,  31, -16,   1,  -1, -15, -13,  37, -43,  33,  -4,  29,   0,  18, -22, -12, -28,   8,  -4,  -6,   9,   9,  -2,   1,   7,   6, -13,  -4,   4, -18,   2, -11,   7, -16,   2,  11,  -5,  -4, -30, -21, -14,  25,   0,  -6,   1, -27,  22, -13,   9,   5, -17,   6, -11,  -8,  27,   6, -18,   2,   8,  -4 },
      {   4,  -1, -11,  16,  41, -54,  -2,  -2,   4, -11,  -8,  29, -15,   8, -21, -14,   4,  -2, -23, -18,  17,  23,  -4,  -7,   3, -19,  27,  -3,   5, -17,  -5, -20,   4,   5,  40,  14,   5,  -3,  26,  -4,   8, -10,  -2,  -3,   3, -13,  -9,   4,   8,  15,  37,  -3,  -5,   0,  -3, -12,   0, -16, -16,   8,   7,  16,  -3,   0 },
      {   4,  17,  34,  29,  20,   6,  23,  22,   1,  18,   8,  15,  15,  -8, -20,  15,  -1,  -9,   2,  29,  11,  11, -36,  -1,  -2,  16,  -6, -18, -19,   3,  -1,  28,   9, -18,  22,  -5, -39,   9, -14,   0,  -6,  19, -33, -11,  13,  13,  10,   4,  15,  -1,  -5, -16,  -2,  -9,  12,   4, -13, -32,  -8,  12,   2,   1,  -7,  -6 },
      {   7,   5,  16, -15,   7,  13, -11, -15,   0, -18, -46, -37,  11,   4, -11, -20,  -7, -12,  37, -10, -18,  -2, -21,   4,  -1,  18,  20,  25,   4, -14,  -3, -10,   0,   1, -23,  -6,  19,  -1,   7,   0,  -4,  21,  -3, -18,  19,   1,  -5,   1,  26,  25,  18,  -2,   0,   2,  -4,  -1,  -2, -46,   3,   6, -12,  11, -19, -13 },
      {  -3,   3, -15,  -1,  17, -30,   7, -11,  -1,  -3, -33, -23,   8,   6, -13,   4,  -4, -17,  -9, -12,   1, -30,   0,  -4,  -9,  -2, -33,   4, -23,  14,  33,   7,   1,   4,  10, -41,  -5,  -6,  -9,  -5,  -7,  -8, -28,  -6,  -4,  40, -20,  16, -22,  -3,  -3,   2, -25,  29, -13, -13,  14,  17,  -2,  12,   2,  17,  -1,   2 },
      {  -3,   8,  20,  -7, -47, -47, -13,  -5,  -4,  17,   4,  29,  -3,  31,  -7, -26,   2,   9,  17,  33,  -8, -14,  16,  19,   2,   6, -15,  -6,  26,  -8,  -3,   0,   0,   5, -11, -19,  -7,  13,  30,  -3,   4,  22, -24,  -7,   7,   1,   0,  21,   9,   4,  -2,   8,   7,  -5,  15, -11,  11,   0,  17,   2,   9,  17,  12,  19 },
      {  -3,   8, -18,  -4,  14, -21, -28, -13,  -2,  14, -13, -20, -27,  17,  24,  27,   3,  -9, -13,  25,   1,   5,  24,   4,  -3,  -2,   4,  20, -13,   4, -21,  -6,   2,   2, -17,   8, -12,   7,   0,  12,   0,   8,  -9,  -7, -16, -12, -14,  -8,   5,  -6, -52, -19,   1,   4,   7,  -3, -27, -28, -43,   0,  21,   9,  -4,   2 },
      {  -1,   5,  19,  -5,  22,  -8,  16,  -2,  -5,   9, -22,  -3,  -6,  -9, -22,   9,   4,  15,  20, -14, -10, -17, -25,  12,   2,  -2, -38,  21,  13,  12, -29,   0,   3,  -4,  12,  17,  37,   9,   8,   7,  11,   6, -16, -13,   0, -21,  36, -17,   0, -26,  -3,   1,   3,  -8,   0, -13,  -4,  38, -20, -14,  27,  -3, -15,  12 },
      {   2,  -3,   2,  24,  -3,  -9,  12, -10,  -5,   6, -11,   5,  -3,  18,  18, -10,   3,  34,   9,   8, -32,   4,  10,   1,   0,  -4, -21,  34, -56, -30,  -3,  24,   1,  12,   9,  22,  -7,  34, -18,   8,  12,  -2,  10,  15, -13,  -4, -19,  -3,   8,  -7,  31,  -2,   3,  15,   8,  -3,   6,   0,  19,   0,   0, -12,  -5, -20 },
      {  -3,   0,  -1, -14,  14,  -2,  20, -44,   6,   1,  17,  24, -23,  20,  18,   2,   1,  -9,  21,   1,  45,  18, -15,   6,  -3,  17,   5,  21,  -2, -10, -31,  24,   5,  -9, -10, -44,   6,   0,  -4,   5,  -9,   0,  10,  22,  -8,   3,  10,  -5,   3,  10,  -2,  22, -38, -12, -12,   6,  12,   8,   3,   3,   2,  -6, -13, -11 },
      {  -2,  -1, -27,  14,  -8,  -6,  17,  -4,   3, -12,   3,  19,  15,   0,  -4,  -9,  -5,  10, -26,   9, -22,  -4,  15, -20,   8,  11, -23,  21,  29,  20,  -5,  16,   0, -28, -16,  10,  -4,  15,   8,  -7,  -6,  12,  34,  -5,  25,  41,   5,  -2,   3,   3,  10,  22,   3, -11, -14,  -7, -12,  -9, -47,   6,  -8,   2, -22, -15 },
      {   4,   5,  12,  12,   7,   6, -40,   7,   2,   5,  -5,  21,  34,  17, -22,  26,   4,   3, -17,  12,  26, -38,  12,  -5,   4,  -3,  10,   0, -13,   3,   0,  37,   4, -17, -29,   8,  39, -10, -12, -13,  -9, -18,  18, -25, -13, -22,  -7,  10,   5,  11,  12,   4, -10,  -7,  10,  -6,  12,  -2,   9,   5,  23,   5,  -1, -13 },
      {   0,   1, -25,   0,  29,  22,   2,  13,  -2,   1, -21, -18, -30, -30,  -1,  -8,  -3,   1, -29,  19,   7, -21,  -5,  -5,  -3,   6, -21, -12,  14, -31, -11,  19,  -3,   1, -26,  -6, -30,   4,  25,  -8,  -3,  -4,  -3,  16, -23, -12,   9,  -8,   9,   5,  19,  33,   4,  -2,  24, -26,   0,  -8,  24,   1,   0,   9,  -4,  23 },
      {   2,  -5,  -4,  22,   6,  21,  14,  30,   3,   6,  -9,  14,  16,  33,   6, -26,   0,  -4, -13, -35, -16,  14,   3,  12,   1,  -7, -13,  17,  -5, -22,   1,   3,   1, -12,   9, -27,  21, -13, -21,  -3,  -5,   2,   0,  18, -13, -17,  10,  26,  -3,  10, -40,   7,  14, -26,  -6, -33,  -6, -19,  -8,   1, -10,  11,  13,  26 },
      {   1,  10, -10, -35,   5,  -9,  -8,   1,   1, -17,   1,  -8, -16,  22, -20, -12,  -1,  -4, -14, -15, -16,   0,   3,  30,  -1,   4,  18, -15, -26,  18, -22,   6,   5,  -4,   6,  20, -29,  14, -39,  -5,  -2, -11,   5, -17,  31,   5,  31,  17, -10,  15,  -6,  39,  -5,   5,  16,  -3,  15,  -7,   5, -20,  17, -21,   3,  11 },
      {   2,   2,  16,  -9,  16,  -8,  17, -18,   3,   2,  -7,  -1,   1,  -1, -11, -22,   1,  12,  11,  25, -13,  31,  36,  -7,   2,  11,  -9, -24,  12,   5,  22,   6,   5,   5,  -1,  33,  16, -39, -22, -11,   2,  -1, -12,  -8, -38,  17,   1, -38,  -2,   2, -12,   7, -25, -13,  -4, -21,  21, -23,   4,  -4,   0, -15,  -2,   6 },
      {   3,   3,  -1, -19, -10,  24,  -8,   3,  -6,  -6,   5,  -4, -18,  24, -24,   8,  -3,  12,  15,  14, -12,  10, -17, -15,   0,  11, -15, -44, -29,  17,   7,  -3,  -1,   2,  -4,   3,  19,  -7,  21, -13,   0,   9,   7,  58,   6, -14, -20,  23,   3,   8,  15,  -1, -11,  10,  -6, -16,  -7,   0, -40,   9,  18,  -7,  -4,   8 },
      {   3,   2,   2,  -2,   2,   3,  22, -27,  -3, -11, -14, -15, -14,  26,  -4,   5,   2,  -2,   0, -10,  12,   4,  15, -46,  -1,  -6, -14, -36,   6, -30,  22,  30,  -2,   1,   1,   2,   4,  14,  17,  35,   0,   1,  -4, -30,  27, -24,   4,  14, -11,   5, -18,   0,  13, -16,  13,   0,  -6,   9,  -5, -20, -18, -14,  13, -28 },
      {   1,   7,   9,  -2,  22,  19,   8,  30,  -3,   1, -13,   1, -15,  12,  10, -41,   0,   1,   4,  -3,  24,   0,  21,  19,  -1,   7,  11,   1,  -6,  27,  -6,  -2,   0,  13, -16,  28,   3,  19,  21,  -2,   1, -11, -24,  -3, -13,  16, -12,  30, -11,   1, -10,  -2,   2, -28,   8,  13,   2,  30,  -3,  41, -15,  -3, -25, -28 },
      {  -1,   4,  -5,   1, -13,  -1,  -8,   2,  -3,   1,  -4,   8,  15,   0, -31,  14,   1,   0,  -9,  -9, -10,   1,  24,  11,  -1,   2, -12,  -3, -16,   1,   4, -30,   1,  -3,   4, -32, -19,  23,  12, -39,  -9,  -7, -23,   3,  -9, -41,  -6, -36,  -2,   8,  -6,  16,  -7, -26, -10,  13,  -2,  -1,  -3, -23, -26, -19, -35, -30 },
      {   0,  -6, -14,  13, -20, -14,   1,  -9,  -1,  -9,   1,  -9,  -8, -13, -30, -32,  -1,  -6, -14,   2,  13,  -9, -26,  21,   1,  -5,   1, -32, -14, -12, -13,  -7,  -4, -11, -17, -14,  20,  29,  -5,  24,  -6,   2,  12, -11, -46,  26,  -3,  -3,   0,   0,  -2, -25,  25, -14, -29,   5,  -4, -12,  -3,  -6,  17, -36,   4,  -3 },
      {   0,  -4, -18, -12,  10, -17, -29,  -7,  -2, -13,   2,  10,  -9, -19, -31,   5,  -1,   3,  -3,   5,   8,  23,  -4,  -1,   4,  -3, -34,   4,   5, -20,  10, -12,  -6, -11,   5,   3,  22,  -1, -27,  -8,  -1,  21,  15,  18,  20,   3,   7,   5,   6,  -7, -36,  -5,  15,  10,  36,  13,  -2,   4,  32,  51, -12, -10, -18, -14 },
      {  -1,   3,   3,  -5,  -6,  14,   7, -20,   4, -11,   1,  18,  41,  12,  11,  17,  -3,  -4, -14, -19, -12, -11, -34,  -1,  -3,  -2, -10,  -5,  34, -18, -13, -30,   1,   6,  -2,  31, -13,  25,   4,  18,  -3,  -5, -16,   1, -22,   0, -33,   9,  -7,   4, -15,  16, -41,   7,  24,  -7,   6,  -9,  -2,  20,   9, -26,   6,   2 },
      {   1,   0,  -3,   1,   2, -14,   8, -25,   3,  -4,  -3,   1,   3,  12,  11,  10,  -3,  -7,  -1,   4,  -7, -29,  -3,   8,  -1,   5, -12,  -1, -25,  -3,  26, -28,   1,  -8, -11,  12,   0, -16,  -6,  11,  -6,  -7,  -4,  14, -17,  -1,  46,  -4,   6,   9,  27,   7,   4, -40,  25,  48,  -7,  -3, -25,  23, -17,  13,  30,  14 },
      {   2,   0,   4, -25,  12,   7,   3, -12,   1,  -6,   3,  14,  -4, -24,   5,  23,   2,   4,  17,  -1,  22,   2,  25,  47,   2,   2,  -7,  -2,  16, -11,  19,  -1,  -4,  -6,  -1,  -5,   2,  25, -32,  -8,  -4,  -3,   0,  -9,  -2,   2, -18,  27,   9,  -7,  34, -24,   6,  -3,  17, -39, -13,   1, -31, -17, -35,  -8,   7,  12 },
      {   0,   1,  -3,  -6,   0,   3,  -7,   1,   2,   1,   2,  16,  14,  -3,  28, -18,  -4,  -1, -15, -12,   1,  14,  14, -13,   3,   3,   5,   0,   0,  33,  35,   8,   0,  -3,  -3, -24,  13,  17,   6,  38,  -4,   2, -17,   1, -10, -19,  31,  -2,   9,   7,  17, -11,   1,  33,  27,   5,  -1, -14,   0,   0,  15, -28, -51,  30 },
      {   2,   0,   4,   1, -18,  18,  23,  -6,   1,  -5,  12,   3,  -4,   0,  -8, -25,  -1,   3,  -2,  -7, -13,  -9,  22,   1,  -1,  -7,   0, -24,  -9, -45, -30,  -4,  -1,  -8, -13, -10,  -8, -13, -10, -18,   1, -14,   6, -30,   8, -10,  11,  -8, -14,  -3,   3, -41, -33,  28,  12,  15, -10,  12, -21,  26,   6,  28, -28,  16 },
      {  -1,  -6,   6,  18,   6,   2, -21,  37,  -1,  10,  11,  -3, -22,  11,   7, -23,   2,   9,  12,   7,  14, -19,  -3, -27,   1,  -6, -10,  12,   4, -16,  -7, -40,  -1,   0,  -1, -22,   4, -29, -18,   1,   0,   5,  -6,  -8,   3,  25,   2,   2,   5,  -2,  13,  11, -19,   9,  39,  -4,  -5,   1, -23, -33,   3, -41,   8, -29 },
      {  -3,   7,   0, -18,   5,  14, -19,  23,  -1,  -1,  -4,   9,   7,  11, -19, -20,   1,   0,  11,   0,  20,  17,   1, -24,   0,   6,  -7,  -7,  15,  -5,  10,  -9,  -2,   5,  -4,   0,  -5,  48, -42,  23,   4,   5,   0,   9, -20,  -1,  -1, -22,  -1,  -9,   6,  19,  -2,  13,  -8,  23,   2,   5, -15, -16,   8,  65,  12,  -6 },
      {   2,   0,  -3,   0,  -1,  -3,  -7,   8,  -2,   1,   2,  12,   3,  -6,   6,   9,  -2,   3,  -3,   5,  -7,  -3, -12,  32,   1,  -2,  -2, -15,  -9, -11,   9,   5,   2,   0,   7,  20,  19,   1,  37,   6,   1,  -1,  13,  12, -22,  22,  43,  17,  -8,  10, -26,  -5, -29,  33,   7,  -2,  -4, -12,   5, -44, -33,  30, -14, -36 },
      {  -2,  -2, -12,  -5, -19,  -6, -19, -14,   2,   2,   6,   4, -10,  16,  47, -11,   1,  -8,   4,  -4,  -1, -13, -49,   2,  -1,  -6,  -1, -25,   5,  -5,  24,   3,  -1,   3,  10,  11, -11, -14, -27, -23,  -2,  -3,  -5, -12,  -4,  -6,   3, -15,  -4,  -5,   6,   3,  28, -21,  -1, -35,  -1,   9,  -5,   6,  12,  22, -46, -29 },
      {   0,  -1,  -1,  -9,   8,  14, -22,  -3,  -3,  -4,   6,   4,  -2,  -6,  11,   7,  -2,   7,   1, -10,   7,  -8,  10, -31,   2,   2,   3,   7, -21, -24,  17, -17,  -3,   1,   6,  11,   4,  36,  14, -22,   1,   9,  -3,   3,  21,  41,  15,   3,   3,  -2,  -9, -30, -30, -53, -10, -10,   6,  -2,  22, -19,  28,   4, -11,  33 },
      {  -1,  -2,  -1, -19,   4, -11, -32,  27,   1,   1,   1,   2,   4,  14,  14, -10,  -2,  -5,   9, -19,   1,   2, -19,  13,  -2,  -7,  -2,  -4,  -3, -22,   6,  53,  -4,  -2,   9,   6,  10,   8,  22, -19,  -4,   7,  -5,  -6,   5,  18, -17, -53,  -6,   9,  -5,   5,  -2,   7,   7,  25,  -7,  10, -27,   7, -27, -25,  16,  31 },
      {   2,  -5,   6,   5,  10,  -2,  15,   0,   2,   9,  13, -15,  23,  36,  -7,  25,   4,  -5,   7,   4,  19, -13,  14,  -3,  -1,  -6,   3, -24, -18,   1, -34, -25,   0,   4,   5,   3,  12,  -4,   3,  26,   3,   1,  12,   5,   6,  33, -17, -23,   5,  -3,   3,  18,  33,   2,  17, -16,  -1,   2,  15,  -7, -32,  14, -43,  33 },
      {   1,  -2,   1,   3,  -1,   2, -11,  28,  -4,   2,   7,  -1, -14,   4,  -8,   6,   2,   2,  -3,  16,   2, -24,  11,  36,  -1,  -5,  -2,   2,   7, -25,  29,  -3,  -3,   2,   2,   8,  -9, -12,  -8,  63,   1,   2,   2,   8,  34, -13,   2, -17,   1,   3,   3,  10, -34, -10, -51, -12,  -7,   0,   2,  24,   1, -11, -17,  -2 },
      {  -2,  -1,  -8, -17, -14,  -5,   4,   5,  -2,  -7,  -2,   8,   0, -27, -16,  -6,  -2,   0,  -8, -23, -21,   5,  -1,  16,  -1,  -2,   4,   3,  -7,  10,  -6,  26,  -1,   4,   2, -13,  -2, -31,   2,  40,   1,   3,   1,  14,   4,   9, -43,  -9,   0,  -3,   3, -11, -14, -47,  51,   8,  -5,   3,   1, -26,  20,  27, -11, -14 },
      {   2,   1,  -3,  -8, -11,   4,   4,  -2,  -1,   3,   1,   7,   6,  -5,   0, -13,   0,  -3,  -7,  -8,   7,  -6,   1,  -7,   1,   1,  12,   0, -23,  27, -11, -10,   0,  -2,  -9,  -9,   4,  19,   9,  12,  -3,  -2,  -3,   7,  19,  -1,  23, -53,   1,   7,  -3, -22,  -9,   5,  20, -65,   2,   2,   3,  30, -12,   7,  47, -28 },
      {  -2,   0,   0,  -1, -13, -17, -26, -11,  -2,   1,  -5,  14,  -9, -21, -10, -26,  -1,   1,   1,   8,   0, -23, -21, -43,  -1,   3,   2,   9, -13,  24, -24,   9,   1,   6,   5,  15, -11,  -6, -22,  14,   3,   1,   0,  -2,  -9, -25,  -5,  21,  -3,  -2,  -9, -10, -25, -12, -14,  -8,  -2,   3,  -1,  -7, -67,  -1, -11,  30 },
    },
  },
  { // 22
    {
      {  60, -67,  34, -15,   7,  -4,   2,  -1, -54,  54, -21,   6,  -1,   0,   0,   0,  17, -11,  -3,   5,  -5,   3,  -1,   1,  -3,   0,   4,  -3,   2,  -1,   0,   0,   2,   0,  -2,   1,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,   1,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -53, -20,  58, -32,  19,  -9,   5,  -1,  52,  16, -54,  26, -12,   3,  -3,   1, -17, -10,  22,  -2,  -4,   4,  -2,   1, -13,  15,  -3,  -8,   7,  -4,   2,  -1,   7,  -5,  -2,   5,  -3,   1,   0,  -1,  -5,   6,   0,  -1,   0,   0,  -1,   1,   3,   0,  -3,   2,  -1,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0 },
      {  54, -39,  13, -15,  12,  -5,   4,  -3,  57, -35,  14,  -8,   1,   0,   0,   0, -48,  49, -33,  15,  -4,   3,  -1,   1,  14, -13,   1,   5,  -2,   0,  -1,   1,  -3,   1,   3,  -3,   2,  -1,   0,   0,   0,   3,  -7,   1,   1,  -1,   1,   0,  -4,   3,   0,   0,   0,   0,   0,   0,   0,   1,   0,   1,   1,   0,  -1,   0 },
      {  -6, -12, -23,  36, -15,   7,  -2,   3,  11,  54, -14, -22,   5,  -4,  -2,  -2, -70, -19,  33,   0,  13,  -6,   6,  -1,  47, -20, -17,   7, -10,   6,  -4,   0, -12,  12,   2,  -3,   4,  -1,  -1,   2,   6, -13,   7,   0,   0,  -1,   1,   0,   1,  -4,   1,  -1,  -1,   2,  -1,   1,  -2,   0,   1,   0,   1,  -1,   0,   0 },
      {  73,  22, -31,  -1,  -6,   6,  -3,  -3,  47,  -5, -25,  13,  -6,   4,  -1,   1,  10, -46,  34, -19,  10,  -3,   2,   0, -30,  28, -19,   7,   5,  -4,   1,   0,   0,  -2,  -2,   9,  -5,   0,  -1,   0,  -6,   6,  -7,  -4,   5,  -1,   1,   0,  -1,   2,  -3,   3,   0,  -1,   1,  -1,   1,   4,   3,   0,   2,  -1,  -2,   1 },
      {  -2,  -8,  -4,  15,  -9,  -3,  -3,   5, -30,   0,  17,  -7,   4,  -3,  -2,   2, -60,  32,  29, -11,   1,  -3,   2,  -2, -72,  32,   5,  -5,   2,   0,   4,  -1,  43, -25,  -9,   3,  -6,   6,   0,  -2,  -3,   4,  10,  -2,  -4,   1,  -3,   2,   5,  -3,   1,   0,   0,   0,   0,   0,   1,  -3,  -5,   0,  -3,   1,   2,  -1 },
      {  18,  44,   6, -48,  22,  -9,   3,  -4, -50, -21,  -7,  49, -24,  12,  -2,   3, -44,   9,  20, -23,   7,   0,   0,   0,  38,  -7, -13,   7,   3,  -6,   2,  -1,  -8,  -8,  16,   0,  -4,   3,  -1,   0,   3,  -2,  -5,  -1,   3,  -1,   0,   0,   1,  -1,   1,   1,  -1,   0,   0,   0,   4,  -2,   0,   0,   0,   0,   0,   0 },
      { -15, -45,  14,  -6,  17, -13,   8,  -1,  -7, -36,  51, -17,   1,  -4,   0,   0,   8, -35,  25, -13,  10,   0,  -2,   2,  27,  45, -55,  23, -14,   3,   0,  -1,   9,  12,   2,  -2,  -5,   5,  -1,  -1,  -7,   1,   6,  -1,   3,   0,  -2,   1,   2,  -4,   2,   1,  -1,   1,   1,   0,  -3,   3,  -8,   0,  -2,   0,   1,  -1 },
      {  11,  17,  24,  -9, -23,  17, -13,   1, -12, -26,  -7, -17,  38, -24,  14,  -2, -16, -27,  15,  21, -33,  17,  -6,   1,  34,  -4,  11, -28,  21,  -7,  -1,   3,  38,  14, -35,  22,  -9,  -5,   5,  -6, -14,  32,  -8,  -5,  -1,   4,  -4,   2, -12,  12,  -4,   2,   3,  -4,   2,  -2,   7,  -2,   2,   1,  -3,   2,  -2,   0 },
      {  -3, -17, -19,   8,  15, -12,   9,  -2,   9,  -4,  -2,  14, -23,  16,  -9,   3,  19,   3,  13, -30,  21, -11,   2,   0,   9, -61,  11,  17,  -9,   3,   2,   0,  72,  13, -30,  -3,  -5,   3,  -3,   1,  -3,  29,   3,   0,   2,  -3,   1,  -1, -18,  -2,   9,  -2,   0,   1,   0,   0,  -5,   3,  -3,   0,   1,  -1,   1,   0 },
      {   0,  25,   4,  -7,   4,   1,  10,  -9,  -1,  20, -16, -23,   8,   9,   3,  -3,  -3, -13, -39,  29,   1,  -5,   5,   2,  12,  16, -29,  27, -15,  10, -11,   2,  53, -59,  29,  -8,  14,  -4,  -4,   6,  13,  18, -24,  -2,   7,  -6,   8,  -6, -12,   0,   4,  -2,  -1,   1,  -2,   1,  -7,   0,  13,   0,   6,  -3,  -2,   2 },
      {   5, -11,  18,  32, -44,  21, -15,   7,  -3, -21, -42,  10,  14,  -4,   2,   1,  28,  54,  18, -35,  10, -17,   6,  -6,  34,  24, -14,   1,  -7,   6,   0,  -1,   0, -20,  19,  -7,   2,   2,  -1,   2, -19,  10,  14,  -3,  -2,   0,   0,  -1, -17,  -5,   9,   0,   0,   1,   0,   1,  -3, -10,   1,   0,   0,   1,   1,   0 },
      {   8,  12,  53,  -4, -18,  12,  -8,  -1,  -7, -26,  -7, -13,  28,  -7,  11,  -4,  -4, -16,  10, -13,  16,  -9,   5,  -1, -35, -40,  -9,  38, -22,  17,  -7,   3, -14,   9, -13,  -7,   9,  -3,  -1,   3,  57, -24,   1,  -5,  -2,  -3,   2,  -3, -14,   6,   4,  -1,  -1,   0,  -1,   1,   1,  -2,   3,   2,   3,  -1,   1,   0 },
      {   2, -25, -13,  21,   2, -18,  13,  -3,   1, -29,  -6,   7,  -5,  10, -12,   4,   8, -13,  20,   8,  -6,  -4,   0,   1,  23,   2,  18, -41,  14,  -3,   3,  -3,  -1, -55,  -5,  19,  -6,   7,  -3,   3,  72, -15,   3,  -1,  -9,   1,   0,  -1,   9,  17,   3, -11,   2,   2,  -5,   3, -18,   8,   0,   1,  -1,   0,   5,  -2 },
      { -19, -35, -14,   6, -10,  36, -13,   1,  -9, -16,  -3,   9,   2,  -5,  10,   1,  -7, -11,  -8, -22,  11,   6,  -1,  -1,  -7, -26,   1,  22,   7, -10,   1,   3, -13, -29,  12,  22, -11,   1,   2,  -2, -14,  15, -61,   1,  15,  -4,   3,   1,  44,  20, -26,   6,  -2,  -5,   4,  -5,  24,  14,   5,  -4,   2,   0,  -7,   2 },
      {  -7,  -3, -26, -27,  12,  -2,   3,   1,   8,  32,  12, -15,  10, -17,   7,  -1,  20,  41,  41,  -9, -35,  19, -11,   3,  -5,  -4, -24,  18,  14,  -9,   2,   1, -22,  -7,  -2,  -2,  -6,   3,   5,  -4,  36,  32, -27,   1,   2,   2,  -3,   2, -39,  18,  12,   3,   0,  -4,   3,  -2,  11, -12,  -4,   1,  -1,   2,  -3,   0 },
      { -13, -15, -14,  12, -10,  22, -10,   1,  -7,   2,  -6,  34, -13,   6,  -3,   3, -20, -11, -53, -12,  17,  -1,   5,  -2,   0,  42,  -1,  -4,   4, -15,   4,   0,   8,  41,  -9,   3,  -7,   0,   0,  -3,  40,   4, -12,  -8,   3,   1,  -2,   2, -51,  29,   4,   5,   3,  -5,   6,  -3,  12, -11,   5,   3,  -1,   1,  -2,   1 },
      {  -2, -10, -24,  -6,  -7,  20, -13,   7,   1,  -6, -17,  27,   2, -15,   9,  -5,  10,  13,  -1,  19, -34,  18,  -9,   2,  -3, -15, -48,   4,  21,  -9,   1,   2,  44,  22,  18, -23,   3,  -6,   2,  -2,  22, -47,  27,  -3,   1,   4,  -3,   1,  33,  -1, -14,   7,  -4,  -2,   1,   0,   7,  12,  -2,  -1,   1,   0,  -1,   0 },
      {   5,  23,   6,   4,  15, -10,  11, -12,  -3,  16, -18, -28,   2,  11,   1,   1,  10,  39, -13, -10,   8,  -6,  -2,   1,  14,  23, -18,  17,  -1,  -5,   2,  -2,   4,  16, -54,  54, -22,  12,  -8,   4,   9, -24,  -1, -23,  23,  -9,   6,   1,  35,  10, -13,  -1,   0,  -2,   1,  -2,  -2,  14,   5,   2,   2,   1,  -3,   1 },
      {   9,  17,  29, -18, -18,  21, -24,  15,  30,  46,  61,  16,  -7, -26,  12,  -5,   9,  -1,  -8, -39,  13,   4,  -3,  -3,  21,   6,  21, -10,  -4,  -6,   3,   0,  23, -26,   4,   4, -10,   8,   0,  -2,  17, -15,   9,  -3,  -1,   2,  -5,   4,  15,   0,  -2,   6,  -2,  -1,   1,  -2,  -4,   2, -12,   1,  -2,   3,   1,  -1 },
      {  10,  18,  44,  44,  -1, -14,  -8,   1,   2,   3,  15, -20, -30,  32,  -7,   5,   4,  -3,   5,   9,   1,  -2,  -7,   4,  -6, -24, -29,  -9,  41, -34,  23,  -9,   4,   0,  24,  -7, -20,  18,  -9,   1, -12, -12, -11,   8,   2,  -4,  -4,   3, -14,  26, -10,   0,   0,  -5,   7,  -4,  19, -24,  -1,   3,  -2,   5,  -3,   1 },
      {   3,   9,   8,  -4,  -9,  -8,   0,   4,  -1,   8,   5,  -1,   2,   2,  -5,   5,   2,  14,  13,  -5,   4, -11,   5,  -2,  -2,  10,  -3, -14,  -7,  13,  -5,   1,   9,  38,  12, -27,  10,  -4,   2,   0,   5,  18, -31,  25, -14,   2,   0,   0,  35,  62, -12, -13,   1,   2,  -5,  -1, -63,   9,  27,  -6,   2,   0,   6,  -1 },
      {   7,   8,   4,   5,  17, -17,   1,   0,   0,  -5, -22, -39,   2,   5,  -1,   4,   4,   2,   7, -26,  -2,   0,   0,   0,  16,  28,  41,   1,  -6,  -4,   5,   0,  20,  23,  18, -36,   9, -10,   7,  -6,  30,   0, -23,  23, -10,   6,  -7,   2,  25, -14, -10,   9,  -2,   1,   4,  -1,  60,  11, -20,   2,  -5,   0,  -3,  -1 },
      {   4,  11,  14,  23,  -6,  -3, -11,   8,  -3,   3,   0,  10, -12,  -3,  -2,  -2,  -5,  -1, -22,   5,   1,   1,  -3,   1,  -7,  -5, -34, -12,   4,  -4,   8,   0, -19,   9,  -9, -13, -13,   7,  -2,   0,  34,  78,  10,  -8,  -4,  -4,  -4,   1,  39, -44,  13,  11,  -2,   1,   0,   1,  -9,  22, -17,  -2,   0,   1,   0,  -1 },
      {   5,   7,  12,  51,  11, -21,  18, -14,   0,  -2,  -6,  26,  10, -17,   9,  -2,  -8,  -9, -11, -25, -32,  56, -35,  20,   4,   8,  22,  41, -10, -13,   9,  -2,  -3,  -6,  -2, -12,   5,   8,   1,  -3,  -9,  -8,  -5,  -1,   6,  -2,  -3,   3,  -2,   4,  13, -11,   0,   0,   4,  -3, -31,  10,  -4,  -2,  -3,   2,   1,   1 },
      {  -3,  -4, -11, -11,  23,  -5,  10, -15,  -3,  -4, -15, -28,  36, -23,  15,  -9,  -4,  -5, -14, -34,  24,  -1,   0,   2, -10, -13,   2, -26,  22, -19,   5,   2,   9,  19,  55,  19, -38,   7,  -6,   1,   6,  -4,   1,   1,  13,  -1,  -2,   2,  -6, -27,  19,   3,  -4,   1,  -2,   3, -35,  -9,  12,   2,   4,   1,   1,   0 },
      {   1,   3,  13,   8, -34,  -8,  35, -22,   7,  12,  22,  33,  24,  11, -32,  17,   3,   7,   1,  -5, -11,   3,  12,  -2,  -4,  -4, -20, -16,   7,  13, -15,   4,  10,  18,   7,  26,   1, -13,  10,   1,   0, -16, -45,  20,  -1,   2,   6,  -6,   1, -36,  25, -30,  14,   1,  -8,   6,  21,  12,  -3,  -1,  -2,  -1,   4,  -1 },
      {  -1,   4,  -6,  -7,  -9, -34,  32, -17,   2,  10,   1,  10,  51,  -7, -26,   9,  -4,  -7, -25, -32,   3,   1,   8,  -3,  -4, -15, -22, -16,  -6,  -5,   7,  -1,  -8, -18,  -3, -13,   2,   3,  -3,  -1, -10,  13,  38,  -4, -15,   4,  -6,   4,  -4,  38, -40,  25,  -1,  -2,  -1,  -5,  27,  13,  -4, -10,  -3,   3,  -1,  -1 },
      {   6,   9,   5,  35,  56,  -9, -21,  12,   0,   1,  -5,  14,  -4, -33,  28, -18,   3,   4,  -8,  -7,   9,   3,  -2,   2,  -6,  -7, -29, -33, -22,  37, -22,  14,   0,  -7,  -1,  21,  -3, -17,  17,  -8,  -6,   1,  -1,  16, -15,  10,  -3,   0,  -2,  20,   5, -26,   5,  -2,  -4,   4,  30, -12,   3,   4,  -6,   0,   1,  -2 },
      {  -1,  -6,  -6,  -7, -10, -12,  12, -10,   4,   1,   2,   4,   9,   3,   4,  -2,  -2,  -7, -10, -13,   5,  -4,  -7,   6,   3,   1,  -2,   2,  19,  -2,   1,  -3,   0,  -5, -20, -20,  20, -15,  12,  -5,   1,   5,   2, -29,   5,   2,  -5,   0,  46,   7,  39,  -6,  -8,   5,  -3,   3,  14, -86,  23,  13,   0,   2,  -2,   1 },
      {   4,   2,  15,  27,  19,  15,  13, -15,   2,   2,  22,  10,  -8,  13,  -5,   4,   3,   7,  19,  -4, -22,   3,   4,   3,   0,   2,  15,  -7,  -2,  19, -17,   7,   4,  15,  31,  19,  17, -30,   9,  -2,  15,  16,   9, -42,  26, -10,   6,  -1,  -6, -12, -42,  36, -18,  10,  -1,  -1,   1, -10,  38, -15,  11,  -3,  -3,   2 },
      {   1,  -1,  -6,  -1, -19, -36,   0,  23,   4,   2,   2,  32,  27,  -1,  11, -24,   5,   4,  17,  41,  15,  -5, -17,  12,   1,   2,  16,  -1, -39,  13,   9,  -2,   7,  17,  25,   2, -26,  40, -21,   6,   7,   8,  -9, -16,  30, -22,   9,   2,   7,   7,  -8,  -3,  -1,   5,   7,  -7,  23, -18,  -5,  -4, -10,   3,  -1,   2 },
      {  -3,  -3,  -4, -17, -49, -22,  14, -11,  -3,  -2,  -5, -18, -34,   2,  25, -14,   1,   7,   0,   3,  41,  27, -52,  31,  -4,  -4, -10, -11,  -7,  -2,   5,  -5,   1,   2,   8,  25,  21, -30,  20, -13,   5,   5,   5,   5,  -7,  13, -15,   4,  -9,  -5, -22,  -4,   1,   5,   1,   0,   1,   8,   4,   0,   0,  -4,   1,  -1 },
      {   0,   1,   0, -10, -15,   8,   1,  -6,  -1,  -3,  -8, -22, -25,   6, -20,  27,  -4,  -8,  -6,  -7, -26,   9,   3,  -5,  -6,  -8,   6,   2, -25,  20,  -6,   6,   1,   8,  39,  20, -20, -14,  12,  -6,  -3,  11,  27, -31,  22,  -1,  -4,   0,  12,  44,  35, -14,   4,  -3,  -4,  -1,   4,   7, -47,  21,  -9,   6,   2,   0 },
      {   4,   7,   6,   5,  24,  51,  23, -34,  -2,  -1,   0,   1,  -1, -38, -27,  29,   7,  11,  19,  28,  42,  23, -12,  -4,  -1,  -1,  -4, -22, -30, -18,  34, -14,   4,   2,  -2,  -9,  20,   3, -20,  10,   1,   2,  -4,  -5,  -2,  -1,   6,   2,   5,   7,  10,  -1,  -3,  -1,   3,  -2,   9,  -6,  -4,   4,   2,   1,   1,  -1 },
      {   2,   0,  -1,  -6,  -7,  16,   7,   6,   1,  -3,  -5,  -6, -14,  -8, -29,  13,   1,  -3,  -5,   2,  -9,  26,   5,  -3,   3,   3,   7,  32, -12,   4,  -7,   1,   1,   2,   8,  25, -35,  12,   0,   1,  14,  17,  33,  51, -24,  -4,   3,  -5,  12, -12, -32, -17,  26,  -4,   0,   2,   3, -49,  25, -15,   1,   3,   3,   0 },
      {   2,   4,   2,  12,   3, -19,  -1, -10,   0,   1,  -4,  14,   8, -20,  -9,  23,   6,   7,   8,  35,  39,   6,  15, -20,   5,   3,   7,  25,  18,  -3, -32,  19,   2,  -5,  -3,  -8, -27,  -9,  40, -23,   1,  -6, -16, -11,   1,  28, -24,  11,  -1,  -3, -19,  27,  -2,  -6,   5,   0, -22, -18, -37,   6,  -8,   5,   3,  -4 },
      {   0,   0,   5,   0,  -5,   6,  20, -28,   1,   2,   7,   1, -14, -26,  16,   3,   0,   1,   7,  12,  19, -16,  -2,   1,   2,   7,  16,  17,  42,   8, -23,   5,  -5,  -2,   3, -32, -22,  15,   9,  -4,   0,  12,  27, -13,  18,  -2,   3,  -3,  -2,  12,   8, -45,  13,   0,  -1,   4,  29,  41,  43,  -6,  -9,   4,  -3,  -2 },
      {   0,   1,  -2,  -1, -19, -11,  21,  -7,  -4,  -7, -11, -11, -41, -59,  -2,   9,  -4,  -7, -14, -16, -20, -16,  -2,  -3,  -2,  -2,  -8, -15,  15,  42,   0,  -5,  -1,  -1, -15, -16,   0,  18, -18,   7,  -1, -10, -22,  11,  29, -18,  11,  -1, -11,  -9, -28,   9, -16,  25,  -6,   2,  -9, -21, -25,   1,   0,   2,   4,   2 },
      {   1,   0,   1,  -6, -11, -12,   2,  42,  -3,  -6,  -6, -20, -16,   2, -18,   5,   2,   1,  -1, -10,  12,  61,  20, -25,  -2,  -5, -10, -29, -23, -20, -24,  26,  -2,  -7, -15, -26,  -9,  10,  11, -13,  -4, -13,  -9, -11,  26, -16,   9,   1,  -5,  -4,  16,   6, -10,   1,  -3,   5,   5,   9,  35, -19,   3,   0,   3,   2 },
      {  -1,  -1,  -2,  -5,  10,  28, -14, -23,   2,   4,   3,   1,  23,  42,  -3,   7,  -1,   0,   0,  -5,  -1,  -1, -32,  18,   0,   2,  -7, -22, -14,  16,   5,   7,  -5, -14, -23, -40, -41, -21,  24,  -5,   1,  -4,   4,  28,  41, -25,  -3,   0,  -6,  -9, -20,  -8,   4,  12,   1,  -3, -11,  -5,  -4,  24, -16,  -3,  -1,   0 },
      {  -1,  -3,  -2,  -1,   7, -15, -22,   6,   1,   1,   1,   0,  13,  -1,   4,   8,   1,   1,   1,  -7,  12,  16,  14, -25,  -1,   0,   0,   1,  15, -32, -11,   3,   0,   5,   6,   6,  48,  -8,  -9,   4,   6,  14,   8, -11,  19, -24,  24, -11,  -3,  -8, -42, -53,  22,  13,  -2,   6, -10,  -8, -24,  35, -23,   4,  -6,   4 },
      {   3,   5,   4,   0,  16,  13,  19,  76,  -2,  -1,  -2,  -2,  16, -15, -56, -14,   0,  -1,  -5,   0,  12, -14, -20,  16,   2,   0,   2,  16,  26,  12,  27, -17,   0,  -2,   3,   5,   4, -24,  15, -13,   0,   1,   0,  -8,  18,  12, -11,  13,  -5,   0,  -2, -20,  -2,  20,  -4,  -2,  -2,   9,   3,   1,  -2,  -4,   6,   5 },
      {   1,   2,   4,   4,   2,  13,  15,   0,  -1,   1,   4,   5,   9,  10,   1, -35,   1,   4,   6,   8,  26,  22,   5,  12,   1,   8,   6,   9,  33,  -2, -25,  10,   0,   3,  -1,   6,   1, -14,  -3,   7,   0,   8,  18,  34,   2, -35,  38, -15,   5,  26,  19,  18, -26,  20, -28,   6,   6,  -1, -41, -15,  32,  -8,   3,   4 },
      {   3,   4,   5,   5,   2,  35,  71,  17,   0,  -2,  -1,   1,  -9,   5,  19, -34,  -1,   0,   2,  -3, -16, -31,  -1,   0,  -2,  -2,  -3, -10, -21, -39, -26,  29,  -1,   1,  -1,  -3,  12,  17,  -1, -27,   1,   3,   2,  10,  10,   4, -16,   7,   4,   1,  -9,   0,  15, -18,   3,   4,  -7,  -9, -18,  10, -15,   5,  -4,  -5 },
      {  -1,   0,   2,   4,  -8, -21, -16, -14,   0,   0,   2,  10,  -4, -20,   2,  -2,   0,   0,   1,  11, -12, -21,  30,  -8,  -1,  -1,   2,   9, -23, -41,  15, -12,  -2,  -4,  -3,  14,   6, -30,   1,  -3,   5,   7,  16,  48,  60,  19, -26,   2,   2,   7,  11,  -8, -36,  13,   4,  10,   5,   3,  12, -11,   7,  -5,   0,  -1 },
      {   0,   0,  -2,  -4,   8,   0, -27, -11,   0,  -1,  -3,  -5,   6,  -5, -12,  23,  -1,  -3,  -4,  -2,  -2, -26, -14,  27,  -1,  -3,  -3,  -6,  -4, -14,  -1,   5,   2,   1,   2,   8,  33,  38,  50, -33,   2,  -3,   2,  -1,   8, -29,   8,   9,   0,  -1,   7,  -9,  19,   4,   2,  -6,  -4,  -5, -15, -64,  18,  12,  -6,  -7 },
      {   0,   1,  -1,   3,   4,   5,  21,  30,   2,   3,   0,   6,  13,  16,  44,  68,  -2,  -3,  -5,   3,   0, -10, -42, -48,   0,   0,  -1,   5,   2,  -2, -10, -11,   0,   2,   3,  13,   2,   8,   3,  -1,   1,   3,   7,  14, -13, -16,  -9,   8,  -1,   0,   6,  -2, -39,  24, -12,  -1,   6,   2,   4,  -2,   0,  -4,   7,  -1 },
      {   1,   0,  -1,   0,   2,  12,  25,  11,   1,   1,  -1,  -2,   2,  17,  39,   1,   0,   0,  -1,  -4,   0,  17,  34,   3,  -1,  -1,  -1,  -9,  -7,   2,  10, -53,  -1,   0,  -4, -18, -19, -10,  18,  25,   0,   1,  -5, -24,   8,  25,  -1,  -9,  -3,  -2, -22, -36,   6,  -6,  -3,  11,  -4, -12, -29, -48,  22,  -4,  10, -10 },
      {   0,   0,  -1,  -3,  -1,   7, -13, -12,   0,   1,  -1,  -2,  -1,   1, -15,  -4,   0,   0,  -2,  -1, -11,  -4,   4,  31,  -1,  -1,  -1,   2, -18, -24, -10,   3,   0,   3,   2,  -2, -12,  38,  13, -17,  -2,   2,  -8, -21, -26,  26,   8,  -7,  -4,  -3, -19, -21, -44,  36, -44,  27,   7,   0,   5,  31,  10, -27,  26, -18 },
      {   1,   1,  -1,  -1,  -1,   8,  16,  -2,   1,   1,  -1,   1,  -3,   4,  25,  -2,   3,   2,   0,   6,   5,  20,  64,  24,   1,  -1,  -2,   5,   6,  16,  35, -11,   1,   0,   2,  11,   7,  12,  26, -22,   0,  -1,   2,   7, -19, -38, -15,  22,   0,   1,   6,   7, -15,  30,   7, -14,  -2,   5,   8,  12, -52,   6,  -2,   7 },
      {   0,   0,   1,   4,  -5, -15,   8,   2,   0,   0,   2,   4, -10, -24,  11,  -7,   0,  -1,   2,   7, -10, -26,  -5, -26,   1,   0,   2,   8, -19, -28,  44,   6,   0,  -1,  -1,  -5, -30, -43,  29,  -7,  -3,  -7, -10, -15, -25, -36,  32, -19,  -3,   2,   0,  15,  33,   9, -33,  24,   0,   3,   8,   1,   2,   3,   0,   2 },
      {   0,   0,   0,  -2,  -4,   5,   4, -21,   1,   1,   1,  -1,   1,  15,  -2, -35,   0,  -1,  -1,  -1,  -2,  10,  -4, -60,   2,   2,   1,   3,   7,  30,  55,  43,   1,   1,   3,   2,   6,  20,  19, -18,   2,   3,   3,   0,   0,  15,   4,  18,  -1,  -3,  -9, -30, -27,  -8,   8,  -1,  -3,  -3,  -7,  -6,  28, -12, -14,   7 },
      {   0,   0,   2,   2,   0,  -3,   9,  11,   0,  -1,   1,   2,   0,  -7,  -1,  -1,  -1,  -1,   0,  -1,   1,   1,  -1,  -4,   0,  -1,   0,  -4,  -1,   7,   1, -46,   2,   3,   3,   6,  12,  31,  60,   7,   2,   3,   8,  19,  15,   7,  37, -32,   1,   0,   1,  21, -14, -46,  13,  11,  -1,   6,  12,  47,   8,   5, -17,   0 },
      {  -1,  -1,   0,   0,  -4,  -8,   8,   9,  -1,  -1,   1,   3,  -2, -13,   3,  13,   1,   1,   2,   2,  -4,  -6,  17,  22,   0,   1,   1,  -3,  -8, -12, -11,   3,  -1,  -2,  -1,  -6, -13, -15,   4,   8,  -1,  -1,   0,  -6, -18, -37, -11,  45,  -2,  -5,  -8, -21,  -5,  -6,  27, -24,   1,   3,   1,  39,  80,  -3, -21,   5 },
      {   0,   0,   0,   0,  -1,  -2,   8,  17,   1,   1,   2,   2,   3,   2,  20,  42,   0,   1,   1,   0,   3,  -4,  13,  40,   0,   0,  -1,  -1,   0,   2,  20,  38,  -2,  -2,  -2,  -3,  -9, -10, -21, -32,   1,   0,   5,   2,  -1,  40,  44, -34,   0,  -1,  -1,  -5, -11,   7,  23,   4,  -4,  -2,  -3,  -4,   5,   6, -32,  43 },
      {   0,   0,   1,   3,   2,  -3,   4,   8,   1,   1,   2,   3,   4,   2,  22,  20,   0,   0,   1,   1,   4,  10,  14,   3,   1,   0,   0,   0,   9,  14,  22,  14,   0,   0,   1,   3,  15,  19,  -1, -12,   0,   2,   6,  10,  30,  13, -12, -10,   4,   5,   6,  26,  63,  15, -11,  -8,   5,   2,   5,  22,  33, -59,  36, -15 },
      {   0,   0,  -1,   0,   0,  -4,  -6,   2,   1,   1,   0,   2,   4,   5,   7,  13,   0,   0,   0,   2,   4,   7,   7,   8,   2,   1,   1,   3,   6,  18,   9, -21,   0,  -2,   0,  -3,   2, -14, -40, -65,   1,   1,   0,   4,  10, -12,  14,  34,  -2,  -4,  -4,  -3,  -7, -59, -41,  29,  -1,  -1,   2,   3,   4,  12,   2, -34 },
      {   0,   0,  -1,  -1,   0,  -2,   2,  16,   0,   0,   0,   1,   1,  -4,   4,  21,   0,   0,   0,   2,   6,   0,   7,  29,   2,   3,   3,   3,   9,   5,  27,  56,   1,   1,  -1,   3,  11,   4,  11,  53,  -1,   1,   0,   0,   6, -20, -42, -30,  -3,  -3,  -8,  -7, -20, -45, -31,  16,   1,  -1,  -1,   0,  -1,  20,   8, -16 },
      {   0,   1,   0,  -1,  -2,  -3,   3,   4,   0,   0,   0,  -1,  -1,  -9,  -7,   2,   0,   0,  -1,  -2,  -1,  -7,  -1,   2,   0,  -1,   0,   1,   4,  -3,   1,   8,   0,  -1,   0,   1,  -8, -18,  -5, -14,  -1,  -2,   0,  -6, -14, -25,  11, -35,   1,   0,   0, -11, -26, -16,  61,  -9,  -3,  -3,  -5, -12, -19, -58,   6, -67 },
      {   0,   0,  -1,   0,   3,   3,  -3,  -4,   0,   1,   1,   1,   4,   3,  -8,  -9,   0,   1,   1,   1,   2,   7,  -1, -10,   0,   1,   1,  -2,   1,  11,  -5, -23,   0,   1,   1,   0,   0,  10, -16, -38,  -2,  -2,  -4,  -7,  -9, -29, -65, -59,   0,   3,  -1,  -2,   5,  12,  17,  50,   0,   4,  -3,   3,  25,   5, -21,  26 },
      {   0,   0,   0,   0,   0,  -2,  -3,  -7,   0,   0,  -1,  -1,  -2,  -7,  -3,  -4,   1,   1,   0,   0,  -2,  -8,   0,  -3,   0,   0,   0,  -2,  -2,  -8,  -5, -12,   0,   0,  -1,  -3,  -3,  -6,   0, -22,   1,   0,  -1,  -5,  -4, -17,  -9, -25,  -1,  -6,  -5, -13, -18, -44, -23, -61,   0,  -3,  -2,  -6,  -6, -30,  47,  67 },
      {   0,  -1,  -1,   0,   1,   1,   2,  -1,   0,   0,  -1,   0,   1,   4,   3,  -8,   0,   0,   0,   0,   0,   2,   4,  -9,   0,   0,  -1,   1,  -1,   2,   8,   0,  -1,  -1,  -1,   0,  -1,  -3,  -6, -25,  -1,  -1,  -1,   1,   0,  10,   1, -43,  -1,  -4,  -1,  -3,  -2,  22,   3, -56,   2,   3,   5,  13,  26,  77,  33, -44 },
      {   0,   0,  -1,  -1,   1,   2,   0,  -2,   1,   1,   0,   0,   3,   3,   3,   0,   0,  -1,  -1,  -1,  -1,  -4,  -2,   3,   0,   0,   0,   2,   3,   1,   3,  10,  -1,  -1,   0,   2,   4,  -3,   3,   2,  -1,   2,   1,   0,  -1,  -8,  18,  21,   2,   2,  -1,   2,  -4,  -5,  55,  54,   4,   2,   0,   3,   6,  25,  90,  25 },
    },
    {
      {  99, -15, -11,  -3,  -3,  -2,  -2,  -1, -72,  17,   6,   2,   3,   1,   1,   0,  -5, -16,   4,   0,   1,   0,   0,   0,  16,  -2,  -5,   0,  -1,   0,   0,   0,   0,   6,  -1,  -1,   0,   0,   0,   0,  -2,   1,   2,   0,   0,   0,   0,   0,  -1,  -2,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0 },
      { -46,  33,  -5,   3,  -2,   0,   0,   0, -71,  -6,  14,  -1,   4,   1,   1,   0,  73, -27, -10,   1,  -3,   0,  -1,   0,   3,  32,  -4,  -2,  -1,   0,  -1,  -1, -16,  -9,  12,   1,   2,   0,   1,   0,   9,  -8,  -7,   2,   0,   1,   0,   0,  -1,   5,  -1,  -2,   0,   0,   0,   0,   1,  -3,   3,   0,   0,   0,   0,   0 },
      {  53, -17,  -1,  -4,  -2,  -1,   0,   0,  42, -26,  -1,  -2,   0,   1,   0,   0,  59,  21, -17,   1,  -5,   0,  -2,   0, -59,  37,  12,  -2,   3,   0,   1,  -1, -17, -27,   9,   2,   1,   1,   0,   1,  13,   2, -12,   1,  -1,   0,   0,   0,  -6,   9,   2,  -2,   0,   0,   0,   0,  -4,  -2,   2,   0,   0,   0,   0,   0 },
      {  -8,  28,  -2,   2,   0,  -2,  -1,  -1, -41,  -2,   4,   3,   2,   0,   1,   1, -14,  36,  -5,  -5,   1,  -1,   0,  -1, -58, -33,  31,   0,   4,   0,   3,   0,  61, -42, -20,   5,  -3,   1,  -1,   0,   4,  27,  -8,  -4,   0,  -2,   0,   0,   0,  -3,   7,  -2,   0,   0,   0,   0,   2,   0,  -2,   2,  -1,   0,   0,   0 },
      { -22, -50,   7,   1,   4,   3,   1,   1, -13,  64,  -7,  -1,  -4,   1,  -3,  -1, -26,   0,  11,  -2,   2,   0,   1,   1, -11,  10,   4,  -4,   1,   0,  -1,  -1, -25, -46,  17,   5,   2,   0,   2,   1,  51, -12, -28,   1,  -2,   0,  -2,   0,  -8,  31,   4,  -5,  -1,  -2,   0,  -1,   0, -12,   6,   2,   1,   0,   0,   0 },
      { -16, -74,  37,  -8,  10,   1,   4,   2,  -6,  26, -22,   7,  -5,   1,  -1,   1,  59,  -2,   6,  -4,  -1,  -3,  -1,  -1,   0, -23,  -1,   5,   1,   2,   0,   1,  34,  21, -22,  -2,  -3,   1,  -2,   0, -27,  21,  14,  -5,   1,  -1,   1,   0,  -6, -12,   5,   4,   1,   1,   0,   0,   3,  -1,  -4,   1,   0,   0,   0,   0 },
      {   5,   2,  -9,   8,  -7,   3,  -1,   1,  27,  -5,  18,  -8,   4,  -2,   0,   0,  -6, -82,  13,  11,  -1,   4,   1,   2,  -5,  21, -27,   0,   0,  -1,  -1,  -1,  56, -13,  -2,  -3,  -4,  -1,   0,   0,  17,  36, -12,   2,  -1,  -1,  -1,   0, -34,   7,  16,   0,   4,   0,   2,   0,   7, -16,  -3,   2,   0,   1,   0,   0 },
      {  -7, -44,  45,  -2,   4,  -1,   1,  -1, -24, -87,   8,  11,   6,   2,   2,   3, -24,   6, -20,   1,  -3,   3,   0,   0,  28,   6,  -4,  -3,   0,  -2,   0,   0,  14, -12,   3,   2,  -1,   1,  -1,   0,  17,  -3, -21,   7,  -1,   2,  -1,   0,   3,  26,  -8,  -9,   0,  -1,   0,  -1,  -4,   3,   8,   0,  -1,   0,   1,   0 },
      { -12, -30,   2,   4,   7,  -1,   1,  -3, -10,  11,  25,  -4,  -1,  -2,   2,   1, -32, -13, -24,   7,   2,   4,   0,   1, -32,  53,   9, -11,   0,  -4,   1,  -1,  23,  -6,  33,  -3,   0,  -1,   0,  -2, -49, -33,  14,  15,   2,   3,   1,   1,  17, -42, -16,   2,   1,   3,   0,   1,  14,  15, -11,  -5,  -3,  -1,   0,   0 },
      {  -1,  -1,  28,   6,   1,  -2,   1,  -2, -10, -14,   8,   4,   2,   0,   2,   2,  -8, -31,  22,  -5,   2,   2,   0,   1, -69, -34, -31,  15,   1,   4,   2,   1, -15,  32,  -2, -24,  -1,  -6,   3,  -2,  42, -29,  32,   5,  -1,   1,  -1,   0,  17,  -4, -22,  11,  -1,   2,   0,   0,  -7,   5,   2,  -3,   1,  -1,   0,   0 },
      { -12, -19,   1,   9,  -4,   0,   1,   0, -15, -16,  24,   3,   2,   0,   0,   1, -21,  10, -13,  -1,  -4,   4,   1,   1,  -8,  13,   4,   4,   0,  -1,   1,   0, -54,   1, -20,   1,   3,  -2,   2,   1,  15,  50,  28, -20,   2,  -4,  -1,  -3, -27, -49,  37,  16,   1,   5,   1,   0,  14, -25, -26,   8,   2,   1,   1,   1 },
      { -12, -41, -68,  11,  -2,   7,   2,   3,   2,  -6,  73, -25,   0,  -5,   0,  -1,  13,  16,  -4,  25,  -4,  -1,  -1,   0,  -1, -34, -14,  -3,   6,   3,   1,   0,   0,   9,  -6,  -4,  -3,   2,  -1,   1,  -2,   4, -10,   5,  -2,   1,   0,   0,  16,  15,  -5,  -2,  -1,  -1,  -1,   1, -18,  11,   6,   0,   2,   0,   0,   0 },
      {  -4,   6,  -1,  10,  -4,   1,  -1,   2, -15,  11, -13, -11,  -6,   0,   2,   3, -11,  55, -15,  19, -10,  -2,  -4,  -2, -11,  35, -40,  -3,   9,  -3,   1,  -1,  25,  48,  16, -31,  -1,  -4,   2,  -1,   5,   6,  17,  17,  -5,   2,   0,   0, -41,  30,   8,  -2,   9,  -4,   1,   0,  -9, -29,  16,  -1,   4,   4,   0,   0 },
      {   6,  23,  40, -53,  18,  -7,   6,   0,   8,  28,  32,  14, -21,   4,  -2,   1,  -4, -16, -65,  40,   2,   0,   0,   0,  -4, -17,  -3, -23,  12,  -2,   1,  -1, -12,   7,  -3,  -2, -10,   6,  -2,   1,   1,  17,  -4,   0,  -5,  -1,   2,  -1,   6,  13,   6,   2,   0,  -4,  -2,   1, -17,   8,   1,   0,   4,   1,   0,   0 },
      {  -1,  -5, -14,  49,  -6,   1,  -6,  -2,   5,  18, -13,  10,  13,  -5,   3,  -1,  -3, -13, -71, -48,   5,  10,   3,   4,   3,   9,   8,  11, -18,  -2,   2,   0,   7,  16, -10,   2,   6, -11,   3,  -2,  21,  27,  12,  -6,   5,  -1,  -5,  -2,  12,  14, -18,   3,  -3,   2,   0,  -2, -25,  32,   9, -13,  -1,  -3,   1,   1 },
      {  -9,  -5,  -7,   6,   2,   2,  -3,   1, -15,  -9, -30,   1,   2,  -3,   5,  -2, -23, -26,   8,  14,  -1,   5,   0,  -1, -39,   8,  -9,  20,   1,   1,  -1,   0, -39,   4, -26,   5,   6,   6,   0,   1, -68,  24, -27,  -5,  -3,   1,   4,   0,  11,  34,  12, -16,  -1,  -7,   0,  -3, -30,  -5,  22,   4,  -1,   1,  -1,   0 },
      {  11,  14,  59,  39,  -7,  -8,  -4,  -6,   9,  16,  51, -38,  14,  -7,   2,   0,   7,   4,   8, -26, -12,   4,   2,   0,  11, -13,  12,  17,  -3,  -3,   1,   0,  -3, -13,  20,   7,   3,  -4,   3,  -1, -27,  -4,   7,   5,   8,  -2,  -1,   0,   2,   9,  11,  -6,   1,   2,  -1,   0, -31, -40,  11,   9,  -1,   3,   4,   2 },
      {   5,   5,  15,  32,  -4,   2,   1,  -2,   1,   4, -26, -32,   4,  -5,   6,   2,   5,   5,  -7,  40, -16,   2,  -6,  -2,  19,  -8, -33,  13,  26,  -8,   3,  -1, -11, -59,  -9, -40,  -1,   7,   2,   3,  -4,  18,  19,  18, -19,  -1,   0,   0,  16, -13, -17,  14,   7,  -5,   1,  -1,   7,  30,  -7, -14,   4,  -1,  -2,   0 },
      {   3,  -2,  -5,  45,   0,   1,  -4,   1,  -1,  -8, -33,  -8,   6,   5,  -1,   0,   5, -13, -36,  18,   2,   5,   2,   2, -12, -54,  -4, -31,   8,  -1,   4,   1, -12,  15,  43,  17, -15,   0,  -5,   0,  -4,  -8, -32,  11,  11,   1,   2,   1, -12, -19,  12, -22,   0,   0,   2,   0,  36, -23,  -5,   7,  -5,   0,  -3,   1 },
      {   0,  -5,  -3,  13,  22,  -7,   2,  -1,  -6, -10, -30, -56,  -7,  10,   4,   3,  -4, -15,   9,  31,   0, -12,  -4,  -1,   2,  17,  51, -29,  11,  -3,  -4,  -2,  11,  23,  -5,   5,  -9,   5,  -3,   0,  29,  -5,  14, -12,   1,  -4,   0,  -1,  27,  -5,  22,  10,  -4,   5,  -2,   1, -48,   0,  -7,  12,   6,   1,   3,   0 },
      {   2,   8,  33,  14, -19,   2,  -2,   3,  -1,  13,  19, -42,   2,   3,  -4,  -2,  -4,   9,  10,  17,  -7,  -2,   3,   2,  -8,  19,  16,   9,  -4,   1,  -2,   0,  -8,  38, -41,  19,   3,   3,  -3,  -1,  10,   4, -44,  -2,   9,  -1,   2,   0, -20,  -1, -20, -15,   4,   3,   0,   0,  43,  46,  -2, -16,  -7,  -4,   0,  -4 },
      {  -4,  -5, -12, -14,  15,   3,   0,  -3,  -3, -11,  -3, -15,  -3,  -3,   2,   2,  -3, -20,  -4,  -3,   5,   2,  -1,  -2,   0, -22,  55,  10,   3,   0,  -2,  -1, -16,  -6,   6, -11,  -2,   6,   3,   0, -10,  -4,  21,  31, -13,   4,  -1,   1, -79,   3, -38,  12,  12,  -5,   5,   2, -11,   2,  21, -22,  -1,  -3,  -1,   0 },
      {   4,   1,  -1,  53,  -4,  -4,  -7,   1,   6,   3,  12,  48, -33,  14,   1,   3,   4,  -8, -13,  33,  -1,  -5,   2,  -3,   6,   4,  23,   9,   7,  -3,  -6,  -2,   6,  -7, -52,  -4,   5,   6,   6,   1,  -4, -52,  12,  -6,  -6,   2,   6,   0, -10,  20,   5,   9,  -1,   0,  -1,   1,  11, -25, -13,   6,   7,   3,  -1,   1 },
      {  -1,   2,  13,  28, -34, -11,  -3,  -7,   0,   5,   7,  51,   9,  12, -10,   4,   3,  10,  38,  24,   3,   4,   4,  -1,  -2,   5,  -3, -52,  -1,  -1,   1,   2,  -6,   0,  14,  24, -17,  -3,  -4,  -1,  -3,  24,  -1,  16,   7,  -5,  -3,  -1, -15, -12, -20,   7,   8,   7,   2,   3, -48,  25,  -3, -14,   1,  -3,   4,  -2 },
      {   5,   9,   2,  31,  89, -18,   1,  -6,  -2,  -5,  11,   0, -54,   7,   6,   2,   1,   9,  19, -20,  13, -15,  -2,   0,   0,   4, -26,  -4,  -4,  11,  -1,   1,  -5,  -7,   8,  10,  -8,  -7,   2,  -1,  -5,  15,  -5,  -3,  11,   6,  -2,   1, -15,   4,  -3,  -5,  -8,   2,   3,  -1,   3,  25,  -7,  -2,   0,  -3,  -1,   1 },
      {  -2,  -1,  -8, -15,   2, -23,   3,  -3,  -3,   0, -16, -15,   4,  16,  -7,   0,  -1,   7, -20,  19, -22,  10,   5,   5,   0,  -6, -36,  45,  -9,  -6,   1,   0,  10, -10,  -5,  52,  14,  -4,  -2,  -1,   4, -27, -13, -15,  18,   4,   1,   1, -27, -26, -19,   7,  -8,  10,   4,   0, -44,  -9, -32,  -7,   6,  -6,   2,   1 },
      {   0,  -1, -11,   0,  18, -33,  15,  -8,   3,   4,   2,  24,  -7,  19, -10,   4,  -2,  -4,   7,   4, -58,   2,   1,   1,   5,   8,  18,  26,  34, -17,   1,  -4,   2,  11,   3,   8,  -2,   4,  -3,   0,  17,  23,   1,  15,   9,  -3,   1,  -3,  29, -23, -18, -25,  16,   3,   6,   0,  13, -22,  50, -10, -11,   2,  -2,   3 },
      {   1,   3,   0,   6,  53, -24,   8,  -6,   5,   8,   7,  18,  84, -16,  10,  -8,   1,   3,   4,   5, -21,  12,  -3,   1,   0,   0,  -2, -24,  -7, -14,   6,   0,   1,   3, -19, -18,  14,   8,  -4,   3,   1, -23, -12, -14, -21,   7,   1,   1, -19,  -6,  19,  -4,  -4,  -7,   1,   1,  -3,  14,   5,   3,   0,   3,  -3,  -1 },
      {  -1,  -3,  -5,  -8,   2, -31,   3,  -6,   0,  -2,  -3, -27,  19,  17,  -7,   4,  -1,   0,  -4,  10,  19,  11,   1,   2,  -2,   4, -18, -41, -23,   5,   3,   2,  -5, -20, -29,  30, -11, -18,   1,  -2,  -9,   1,  50, -16,  23,  -4,  -6,  -3,   7,  28, -31,  -1,   0,   6,   4,   3,  34, -26,  28,  -6,  -3,  -4,   4,  -1 },
      {   5,   4,   9,  10,  21,  52,   1,   5,   1,   1,  -1,  13,  -3, -45,   3, -10,   3,   6,   2,  50,  17,   5, -13,  -3,   4,   3,   5,  16, -46,  16,   2,  -1,   6,  -3,   6,  13,  20, -13,  -1,  -7,  15,   7,   2, -11,   9,   6,   3,   1,   8, -30, -13,  -6,   1,  -6,   2,   3,  -7, -16,  44,  -3,  -6,   0,   0,  -3 },
      {  -1,  -3,   3,   1, -24,  -2,  -6,  -5,   0,  -4,   3,  -6, -17,  19,  -5,  -1,   0,  -2,  -6, -11,  19,  -7,   6,  -2,   2,  -4, -17,  -9,  31,  -5,   3,   2,   4, -15,  -4, -13,   5,   1,   1,   1,   0, -27,   3, -47,   8,  -3,   4,   3, -27, -33,  31, -17, -15,   1,   0,   4, -15,  30,  70,   3,  -6,  -8,  -3,  -1 },
      {   0,  -1,   4,   7,  -4,   5, -12,   7,  -1,  -2,   4,   3,   0,  41,  -5,  -3,  -2,  -3,   7,  17, -41, -14,   7,   3,  -4, -10,  13,   2, -22,  -1,  -1,  -2,   3,   5,  47,   4,  41,   0,  -5,   1, -10,  18,  12, -45, -20,  14,   3,   3,  -1,  25,  -4,  37, -27, -12,   6,  -2,  22,  16,   7,   0,  18,  -7,  -1,   0 },
      {  -2,  -4,  -2,   0, -14, -81, -13, -14,   2,   0,   0,   4,  -1,   2,  15,  -3,   5,   4,   4,  30,  54,  18,   3,   1,   4,  -1,  18,  35, -25,  10,   4,  -2,   2,   5,  22, -30,  13,  -9,   2,  -1,   5,  11, -13,   6,   2,   7,  -5,  -3,   7,  -3,  13,   0,  -3,   1,   0,  -2,   6,   5,   3,   3,  -2,   1,  -2,   2 },
      {  -1,   0,   1,  -3,  10,  18, -23,  13,   3,   4,   4,  14,  46,  38,  20,  -5,  -2,   0,  -5,  17,   9, -61,  -5,   2,   3,   4,   6,  27,  11,  26, -19,  -1,  -2,  -5,   7, -16, -27,  -6,  16,  -2,  -1,   7,  10,  -2,  40, -12,   0,   1,  -4,   4, -10, -22, -11,  14,  -2,   5,  -4,   3, -10,  12, -18, -12,   0,   2 },
      {  -1,  -1,   0,   5,  -3, -10,  16,   6,  -2,  -3,   0,  11,   6, -63,  -5,  -8,   0,  -1,   6,  16,   8,  10,   1,   0,  -1,   1,  11,  21,  56,  -6,   3,  -2,   2,   5,  29,  27,  -8,   3,  -4,   1,  -1,   3,  35, -42,  11,  -2,   1,   2,  -9,  33,  -2, -19,  -5,   8,  -5,  -3,  11,  20, -20,  -3,  -3,   3,   3,   0 },
      {  -1,  -1,  -2, -10, -12,   0,  -8,  -8,   0,   0,  -4,   2, -16,  -9,  23,  -2,   0,  -3,   2,   8, -55,  30, -15,   8,   0,  -6,   6,   2, -32,  10,   2,  -2,  -2,  -8,  -1,   8, -17, -30,   7,   1,   1, -11,  29,  34,  21,   4,  -7,  -4, -12,  12,  41, -22, -16,   6,  -4,   1,   4,  41,   4,  33,  -8,  -6,  -3,   0 },
      {  -1,  -2,   0,  -7,  -8, -15, -54,  25,   1,   2,   0,  -3, -14, -22,  56, -16,   1,  -1,   2,  -7,  -6, -14,  19,   0,   0,   0,  -6, -28,  16, -12,  -8,   4,  -1,  -1, -11,  10,  46,  13,  -2,   0,   7,  10,   8,  11, -10,  27,   3,  -1,   0, -11, -29, -30, -17,   4,  -1,  -1,  -5, -12,  -4,  19,  -4,   1,  -8,  -3 },
      {   1,   2,   0,  -5,  17,  28, -27,   5,   0,   0,   1,   0,  -6,  21, -23,   1,   3,   2,   9,   6,  -1,  67,  30,   8,   4,   2,  14,  -7,  10, -19,  18,   1,   0,   1,  -1, -48,   5, -27,   0,   5,  -3,   5, -10, -23,  28,  -4,  -8,   1,  -1,   2, -24, -20, -14,  13,   1,  -5,  -9, -10, -22,   3, -16,  -5,   6,   0 },
      {   1,   0,  -1,   0,   2,  10,  32,  -3,   0,  -2,   0,  -2,   1,  15, -29,  11,   2,  -3,   1,   8,  17, -28,  31, -14,  -3,  -7,  -1,   1, -23, -45,   1,   1,  -2,  -5,   3,   0,  56,   8,  -8, -10,   3,   2,  28,  28,  10,  -5,   8,   2,  -1,  11,  24, -45,  12,  10,  -4,   1,  -2,  15, -16,  -8, -19,   7,   4,  -5 },
      {   0,  -1,  -2,  -1, -19, -16,  49, -26,   0,  -2,  -5,  -1, -14, -29,   8,  22,  -2,  -7,  -4,  -4, -29, -33, -15,  11,   0,  -5,   5, -28,  -5,  32,   2, -11,  -2,  -3,  -5, -41,  33, -18,  11,   1,  -8,   0, -20, -24,  29,   4, -13,   1,  -7,  -2, -22,  -6,  -6,  15,  -7,  -1,  -7, -12, -14,  -4,  -7,   0,   2,  -2 },
      {   1,   3,   3,   9,  -6,  -7,  54,  48,   0,   2,   1,  -1,   4,  14, -24, -12,  -2,   0,  -2,  -2,  11,   4,   3, -13,   2,   7,   3,   5,   2,  28,  20,  -3,  -1,  -1,  -5,  -3,  -9,  -4, -23,  11,   1,   6,   4,  12, -27,  28,   3,  -1,  -7,  -8, -24, -16, -33,   0,  11,  -9,  -6,   1,   2,  59,  -2,  -4, -11,   1 },
      {   1,   0,   1,   0,  -3, -14,  18,  88,  -1,  -3,   1,   1,  -8,   5,  19, -33,  -1,  -2,   2,  -5,  -6,   5, -29, -12,  -1,  -5,   1,  -8,  -9, -21,  -1,  -4,  -1,  -4,   1, -13,  -1,  -6,  -3,  -3,  -3,  -6,  -5,  -5,  36,   7,  -8,   6,   1,   3,  17,  16,  17,   9,   6,  -5,  -4,  -1,   1, -45,  -7,   7,   2,   4 },
      {   0,  -1,  -1,   0,   2, -17, -17,  21,  -2,  -3,  -4,  -3,  -7, -20, -25, -23,  -1,  -1,  -5,   1, -27, -14,  70,   6,   0,   0,  -1,  -3, -22,  54, -10,  14,   0,  -6,   1, -10, -18,  -1, -15,   0,  -1,  -9,   5, -11, -27, -21,   1,  -2,   0,  -2,  11, -10,  26,  18,  -3,  -4,  -4,   0,   1, -22, -16,   6,   6,   1 },
      {   1,   1,   2,  -4,   8,   7, -15,   1,   0,  -1,  -1,   2,  15,  -6, -21,   0,   0,  -2,   0,   1,  10, -13,  15,   5,   2,   2,   3,  -1,  44,  32,  32,  -5,   0,   2, -11,  17,  35, -48,  -1,  10,  -1,  -3,  -3,  42,  14,  21, -16,  -3,   7,   3,  29,  27, -23,  10,  14,  -9,   5,  -1,   9, -31,  18,  -8,  -3,   5 },
      {  -1,  -1,   0,  -1,   3,  -9, -37,  20,   0,   0,   2,  -2,   0,  13, -35,  25,  -2,  -3,  -1,  -4,  -2,   9, -67, -18,   0,  -2,   0,  -5,   5,  39,  16,  -9,   1,   0,   2,   7,  30,  -5,  15,   0,   1,  -2,  10, -10, -30, -19,  -3,  -1,  -1,  -2,   0, -40,  27,  -9,  -6,   1,  -5,   4, -15,  -6,  -6,  14,   3,   1 },
      {   1,   0,   2,   1,  -4,  20,  38, -19,   2,   1,   4,   0,   1,  38,  62, -17,   1,   2,   4,   1,  11,  23,  10,  12,   1,   1,   2,   2,  12,  25,  -8, -13,   2,   1,   3,  12,   4, -31, -11,   6,   2,  -2,  10, -12, -33,   3,  -1,   5,   4,   0,   7, -36,  17, -18,  11,   3,  -4,  -5, -14, -38,   3,   5, -17,   7 },
      {   1,   2,   1,   3,   8,  -1,  23,  19,   0,   0,  -2,   0,   6,   7,  -4,  -3,   1,   3,   0,   0,   7,  38,  13,   6,   0,  -2,  -3,  -7,  20,  26, -73,  -2,   3,   1,   2,   8,  35,   5,  51, -10,   0,   5,  -3,  18,   6, -24,  13,   9,   0,  -3,   2,  16,  -4,   3, -29,   0,   3,   5,  10,  16,   7,  -9,  12,  -1 },
      {   0,   0,  -1,  -1,   4,  -4,   3,  13,   0,   0,  -1,   1,   1,  -3,  19,   4,   0,  -2,  -1,   3,   0, -18,  19, -30,   0,  -2,   1,   4,   8, -26,  15,  13,  -2,   0,   1,   6,   7, -67,   4,   1,  -4,  -2, -13,  -8,  -5, -42, -17,  -1,  -5,  -3, -13,  16,  47, -14, -14,   3,   3,  10,   5,  46,  23,  12,   1,  -5 },
      {   0,   1,   0,  -3,  -1,  -2,  -5,  24,   1,   2,   0,  -1,   6, -11,   3,  82,   1,   1,   1,   2,  -1,  -1,  21, -28,   0,  -1,   1,   5,  -8, -13, -39, -15,  -1,  -1,   1,   0, -29, -29,  15,   2,   0,   0,  -3, -11, -16,  51,   2,  -2,   3,  -5,   7, -11,  -5,   4,   9,  -1,   0,   0,   2, -18,   8,   2,   5,  -5 },
      {  -1,  -2,  -1,  -1,  -3,  -8,   0,  -6,   0,  -1,   1,  -1,   0,  -8, -12, -19,  -1,  -3,   0,  -4, -10,   2,   5, -51,   0,   2,   5,  -4,  -6,  11,  -7,  27,   0,  -1,  -2, -15,  -6,   4,   0,  -4,   3,   2,   3,   2,  30,  -7,  25, -10,   1,  -5,  -2, -32, -20, -47,   7,  13,  -2,  -4, -10, -21,  71, -21, -11,  -4 },
      {   0,   1,   0,   2,   5,   5,   4,  22,   1,   2,   1,   3,   6,   7,  36,  61,   0,   1,   1,   0,   0,  19,  23,  -1,   1,   2,  -1,   4,   1,  32,  35,  -6,   1,  -1,   4,   3,  11,  44, -35,  12,   2,  -2,   5,   5,  36, -28,  12, -10,  -2,   3,   3,  14,   3, -12, -22,  -3,   2,   6,   8,   1,   8,  22,  -9,  -7 },
      {   0,   0,  -1,   0,   0,   4,   6, -19,   1,   2,   0,   0,  -2,   4,  22,  -3,   0,  -1,  -2,   0,  -5,  12,  -6, -84,   0,   1,  -1,   0,   6,   9,   9,  34,   1,   4,   3,   6,   5,   0,  -2,  15,   0,   1,  -6,   1, -15,  -9,  12,  -8,   2,   6,   2,  20, -21,  35,  -7,  15,   3,  -7,  -1, -13, -51, -22,  17,  -7 },
      {   0,  -1,   0,   1,  -1,  -6,   6,  13,   0,  -2,   0,   0,  -5,  -8,   7,   0,   0,   1,   2,   4,   0, -13,  10,   4,   0,   1,   0,   0, -16, -22,  26, -32,   2,   4,   1,   6, -15,   0,  50,  41,   0,   1,   0,   6, -25, -54,  -9,   5,  -1,  -7,  -3, -12, -57, -13, -14, -22,   0,  -9,   1, -26,  -2,   3,  -1,  -8 },
      {   0,   0,   0,   0,  -3,  -2,   2,  -5,   0,  -1,  -1,  -3,   1,   6,  -8, -14,   0,   0,   1,   1,  -5,  -3,  22, -11,   1,   2,   0,  -2,   4,   9,  28,   6,   1,   0,  -1,   6,  -7,   4,  53,   0,   3,   1,   2,  -5,  18,  49,   6,   4,   1,   0,   0,  12,  15, -64, -34,  14,  -4,   3,  -6,   1, -44,  29,  -8,  10 },
      {   1,   1,   0,   1,   0,   7,  13,   6,   1,   1,   1,   2,   1,   9,  10,  13,   1,   0,   1,   5,   2,   3,   2,  11,   1,   1,   3,   6,  -1,   3,  13,  74,   1,   2,   1,   3,  -2,  26,  25, -17,  -1,   0,   2,  -2, -10,  10, -80,   1,   0,  -3,   2, -17,  -8,  10,  -6,  -8,   2,  -1,   3,  -6,  23,   8,  20,   4 },
      {   0,   0,   1,   0,   1,   7,  -6, -17,   0,   0,   0,   0,   0,   1,  -5, -17,   0,   1,  -2,   1,   0,   5,  10, -39,  -2,  -1,  -4,  -1,  -4,   8, -23, -42,   2,   1,   2,   1,   7,  31,   1,  65,   0,  -2,   4,  -4,  20,  12, -58,   3,   1,   0,   1,   0,  30,   0,  21, -23,   1,   6,   2,  15,   8,  -5,  -3,  15 },
      {  -1,  -1,   0,   0,  -2,  -5,  -6,  -7,   0,   0,   1,   0,   0,  -5,  -8,   4,  -1,   0,  -1,   1,  -2,  -3,  -8, -12,  -1,   1,  -1,  -1,   5,  -1, -42,  12,  -1,  -1,  -5,  -2,   3, -16, -46, -10,   1,   0,   0,   8,   6, -24, -22,  26,  -3,  -3,  -4,   6, -43, -41,  15,  -8,  -1,   0,  -1,  -4, -29,  61,  10,  33 },
      {   0,   0,  -1,  -1,   0,   5,  -3, -15,   0,   0,   0,  -1,   3,  10,  -4, -16,   0,  -1,   0,   0,   1,  -1, -12,  -3,   0,   0,   1,   2,  -5,  -3,  -3,  19,  -1,  -2,   1,  -3,  -7, -10,  -4,  41,   0,   4,   1,   0,   3,  24,  35,  21,  -3,  -2,  -6,  -7,   0,  47, -51, -26,   3,   1,   7,  -5,  36,  63,  -1,   2 },
      {   0,   0,  -1,   0,   0,   0,  -5,  -9,   0,   0,  -1,   1,   1,   5,   4, -12,   0,  -1,  -1,  -1,  -4,   3,   9, -30,   1,   0,  -1,   2,   1,   8,  19, -51,  -2,   0,  -3,   0, -10,   9, -10, -60,   1,   1,   0,   2,  -3,   7, -22,  70,   0,  -2,   0,  -3,  -6,   9, -27,   0,   2,  -3,   2,  -1,  16,  -2,  23, -32 },
      {   0,   0,   0,   2,   0,  -5,  -3,   4,   0,   0,   2,   2,  -1,  -2,  -5,  11,   0,   0,   0,   1,   1,  -6, -12,  18,   0,   2,   1,   0,   0,  -6, -13,  26,   0,   0,  -3,   0,   1, -13, -38,  31,   1,  -1,   4,   2,  -1,   8,  -5,  36,   1,  -1,   4,  -2,  10, -36, -54, -40,  -2,  -7,  -2,  -7, -17, -66,   3, -12 },
      {   0,   0,   0,   2,   2,   3,   0,   0,   0,   0,  -1,   0,   0,   6,  -6, -17,   0,   0,   0,  -1,   1,  -3,  -4,   0,  -1,  -1,  -2,   3,   3,  -5, -16, -31,  -1,  -2,   0,  -1,  -3,  -8, -30,  -8,  -2,  -1,  -1,   1,  -6,  10, -31, -84,  -3,  -3,  -2,  -4, -15,  -7, -61,  21,   1,  -3,   2,  -9,   3,   7,  26, -17 },
      {   0,   0,  -1,  -2,  -1,   3,   1, -10,   1,   0,   1,   0,   1,   4,  10,  -4,   1,   0,   0,   1,  -1,  -1,   1, -15,   1,  -1,   0,  -4,   2,   3,   8,  -5,   0,  -1,   0,   0,  -2,   1,  10, -40,   0,  -3,   0,  -4,   3,   2,  22, -38,  -3,  -2,  -1,   0,   7,  -7,  12,-104,   0,   3,   0,   3,  -3,  -1,  29,  16 },
      {   0,   0,   0,   0,   0,   0,   1,   6,   0,   0,  -1,   1,  -1,  -3,   8,   6,   0,   0,   1,   0,   1,  -3,   5,   2,   0,   0,   0,  -1,  -5,  -4,  18, -14,  -1,   1,  -2,  -1,   0,   2,  -2,   3,   1,   1,  -1,   1,  -5,   0,  13,  10,  -1,  -2,  -2,  -8,   1,   5, -19,  27,   2,   1,   2,   1,  16, -25,  53, 102 },
      {   0,   0,  -1,   0,  -1,  -2,   2,   2,   0,   1,  -1,   1,  -2,   3,   8,  -5,   0,  -1,  -1,  -2,   1,  -2,  -1,  17,   0,   0,   0,   2,   3,   6,   5,   5,   1,   0,   0,  -1,   3,  -1,   0,  34,   1,  -1,   2,   3,   5,   7,  19,   7,   0,  -2,   1,  -5,   6, -21,  35,  12,   1,   3,  -1,   6,   2,  13,  98, -49 },
    },
    {
      {-105,  36,   8,   3,   2,   2,   1,   1, -29,  -9,  11,  -2,   2,   1,   1,   0,  32, -32,   1,   2,   0,   0,  -1,   0,  24,  -4, -14,   1,  -1,   1,  -1,   0,   6,   8,  -3,  -4,   0,   0,  -1,   0,   0,   2,   2,  -2,  -1,   1,   0,   0,   2,   0,   2,   0,  -1,  -1,   0,   0,   1,   0,   0,   1,   0,   0,   0,   0 },
      {  31,  14, -10,  -1,  -2,  -1,   0,  -1, -85,  40,  10,   2,   2,  -2,   2,  -1, -34, -23,  28,  -2,   4,  -1,   3,  -1,  30, -41,   5,   9,  -1,   0,   0,   1,  20,  -4, -18,   6,   1,  -1,   1,   0,   5,   7,  -6,  -4,   3,  -1,   0,   0,   2,  -1,   0,  -1,  -1,   3,  -1,   0,   2,  -1,  -1,  -1,  -2,   0,   1,   0 },
      {  35, -57,  20,  -3,   5,  -1,   1,   0, -45, -25,  14,   3,   4,   3,   1,   1,  54, -28, -26,   7,  -3,   2,  -2,   2,  31,  31, -29,  -6,  -4,   0,  -2,   0,  -9,  31,   3,  -9,  -2,  -1,  -1,  -1, -10,   4,  10,   0,  -2,   0,   0,   0,  -5,   0,   2,   2,  -1,  -2,   1,   0,  -2,   1,   0,   1,   2,  -1,   0,   0 },
      {  41,  62, -42,  -3,  -7,  -1,  -2,   0,  33,  12,   5, -11,  -2,  -3,   0,  -1,  49, -51,  26,   0,  -3,   0,  -1,   0,  -1,   4, -11,  15,  -3,   2,  -1,   0, -27,  21, -10,   1,   5,  -2,   2,   0,  -9,   3,   5,  -7,   3,  -1,   0,   0,  -3,  -3,   3,   0,  -1,   2,  -1,   0,  -2,  -1,   1,   1,  -1,   0,   1,   0 },
      {   3,  24, -20,   0,  -3,   2,  -2,   2, -54,  24,   4,  -3,   2,   3,   0,   1,  28,  13, -14,  -3,  -2,  -1,  -2,   1, -49,  45,  17, -15,   2,  -1,   0,  -2, -18, -38,  42,  -3,  -1,   2,  -1,   0,  18, -36,   0,  16,  -7,   4,   0,   2,   3,  -3, -13,   5,   3,  -5,   2,   0,   4,   1,  -5,  -3,   5,  -1,  -1,   1 },
      {  -5,  -9,   6,   0,   2,  -2,   4,  -2,   8,  55, -35,   8,  -4,  -7,   0,  -1,  58,  49,  -3, -18,   0,  -7,   4,  -5,  53, -19,  35,  -3,   1,  -4,   2,  -1, -16,  -1,   2,  16,  -1,  -1,   2,   1, -18,  -1,  -3,   5,   7,  -3,   0,  -1,  -2,  -7,  -4,   1,   1,   8,  -2,   0,  -2,  -3,   0,  -4,  -5,   1,   2,  -1 },
      { -11,  -1, -37,  12,  -9,  10, -11,   6,  -2,  33, -43, -21,  -4,  13,  -4,   5, -41,   1, -10, -11,   9,  -1,  -6,   5,  20,  19, -30, -29,   0,   3,  -5,   2,  -5,  41,  28, -26,  -8,   8, -10,  -1, -23,  -4,  32,   0, -11,   7,   3,   0,  -4,  -5,  20,   0,  -1, -17,   6,   0,   2,  -6,   5,  11,   9,  -1,  -5,   2 },
      {   9,  38,  50, -34,   7, -10,   5,  -5,   7,   0,  21,  10,   1,  -4,   2,  -3, -39,   6, -11,  22,  -6,   4,   3,  -1,  41,   9,  -6,   7,   2,  -5,   3,  -2, -46,   9,  22,  -7,  11,  -5,   6,  -1, -10, -39,  16,   2,   2,   1,  -2,   1,  15, -21, -20,  13,  -3,  10,  -3,   1,   3,  -1, -11,  -4,  -2,   0,   3,  -1 },
      {  24,  42,  11, -22,  -3,  -1,  -6,   2,  15,  -8,   6,  -7,   2,  10,  -2,   3,   9,  -5, -45,   6,   0,   1,  -5,   5,  37,  -7,  -5, -26,   6,  -1,  -4,   0,  54, -40,  22, -12, -12,  11,  -9,   2,  -9,  22, -16,  18, -18,   7,  -1,   1, -25,  23,  -2,  -1,   9, -13,   5,  -2,  -3,  -2,   9,  -2,   9,   0,  -2,   1 },
      {   9,  43, -13,   0,   1,  -5,   6,  -5, -24, -17,   1,  23,  -4,  -9,   3,  -2,   8,  59, -29,  15,  -4,  -1,   6,  -6, -18,  18,  -7,  10,   6,  -7,   6,  -2,  35,  29, -24, -10,   9, -12,   7,  -1, -15,  29,  27, -22,  13,  -5,  -1,  -2,   1, -26,  18,   5,  -8,  16,  -4,   2,  -3,  -5,  -5,   4, -10,   0,   3,  -1 },
      {  -2,  -2,  58, -26,   0,   0,  -7,   1,   0,  56,  -2, -11,  -6,   7,  -4,   2,  15,  -8, -20,  20,   3,   3,  -5,   4, -56, -31, -22,   2,  11,   7,  -2,   3, -19,   1, -15, -19,   2,   8,  -5,   0,  -9,  21,   4, -16,  -8,   4,   3,  -1,  -2,   3,  24,  -7,  -3, -11,   3,  -1,   2,  -9,  12,   9,   3,  -1,  -2,   0 },
      {   7,  10,   5,  48, -20,   2,  -2,   0,   3,  -3, -39,  69,  -8,  -5,  -1,  -1,   1, -26, -31,  12,  22,  -8,  11,  -4,  -2, -25,  -5, -19,  25,  -5,   7,  -2, -14,   3,  -3, -23,   3,  -1,   3,   1,  37,  -5,  -2,  -2,  -7,   0,  -3,   0,  10,   7, -17,   6,  -4,   7,  -1,  -1,  -2,  12,  -2,  -9,  -2,  -3,   3,  -1 },
      {   3,  17,  -6,  -8,  -1,   4,  -9,   3, -28, -57, -19,   3,  10,  15,  -4,   5,  -7,   2, -16, -18,  -1,   2,  -7,   5,  -9, -30,  38, -14,  -2,   3,  -5,   2, -61,   1,  -1,  10,  -8,  14,  -7,   1, -38,  28, -15,   2,  -3,   0,   6,  -2,   8,  -4,  14, -13,   8, -13,   3,  -1,  10,  -6,  10,   3,   3,   1,  -3,   2 },
      {  -3,   6, -21, -29,  11,   0,   2,   0,  -9,  -1, -30, -11,  -3,   3,  -2,   4,  -8,  23, -15,   9, -18,   6,  -4,   0,  24,  24, -28,  20,  -3,  -5,   0,   1, -31,  -8, -19,  -1,  17,  -3,   3,   0,  50,   6, -35, -24,   5,   3,   0,  -1,   8,  50,   9, -20, -10,   0,   1,  -2,  -8,   4,  27,   1,  -6,  -2,   0,  -1 },
      {  -4,  11,  30,  16,  -9,  -6,   4,  -3,  -1,  26, -10,  26,  -2,  -9,   3,  -2, -26, -28,  -7,  -4,   7,   1,   5,  -1,  -6,  62,  16,   1, -15,  -3,   1,  -3,  -1,  17,  -2,  41,  -4,  -7,   5,   0, -19,  25, -30,  11,  22, -12,   3,  -2, -35,  11,   8, -20,  18,   6,  -1,   0,  -5, -14,  10, -12,  -6,   6,   1,   1 },
      {  -1,  -4,  -2, -60,  34,  -9,  13,  -7,   0,  -4, -44,  -9,   1,  -6,   4,  -2,   1, -32,  -6,   5,  -4,   4,   8,  -4, -17,  -9,  37, -19,  -7,  -7,   5,  -1,  25,  38,  24,   5, -13, -14,   5,  -2,  33,  13,  10,  14,   8,  -9,  -4,  -2,  19,  -6,  -5,  12,   0,  14,  -7,   0,  -3,   1,  -3,  -3,  -8,   3,   3,  -1 },
      {  14,  30,  45,   7, -14,   1,  -9,   1,  -5, -10,   7, -15,   8,  -4,   0,   0,   9,  29,  21, -52,  24,  -6,   1,   0,  -7,   4, -10, -21, -11,  16,  -6,   4,  10,  43, -17,   6, -29,   7,  -4,  -2,  28, -16,   0,  14,  -6,  -5,   3,  -2,  26,  26,  -8,  -4,   1, -11,   3,  -1, -14,  17,  14,   3,   4,  -2,  -2,   0 },
      {  -1,   2,  -9,  -2,  11,  -4,  -1,   0,  -6,   2, -32,   5,  -6,  10,  -1,   1,   2,   0, -12,  23, -19,   8,  -5,   4, -20, -12, -24,  20, -13,   4,  -3,   3,  23,  12, -17,  13,  -2,  -1,  -3,  -2, -57, -35, -17,  19,   9,   2,   6,   2,  22,  12, -49, -17,   9, -12,   4,   1,  -9,  39,   4, -19,   8,  -3,  -3,  -1 },
      { -10, -10, -42,   0,  24,  -5,   6,   0,  12,  32,  71,  15,   1,  -5,  -4,  -4,  -8,  11, -31,  17,   6,  11,   3,   0,   0,  -7,   4, -34,  -4,   4,   1,   1, -21,  28, -10,   5, -23,   3,   0,  -1,   6,  23, -20,  17,  -8,  -9,   3,  -4,  23,   6,  -9,   2,   9,  -5,   0,  -1,  -5,  13,   3, -11,   5,   0,  -1,   0 },
      {   2,   7,   9,  23,  41, -19,   3,  -2,   6,  15, -15,   4,  48,  12,  -6,   0,  -7, -12, -21, -40, -26,  28,  -7,  12,   2,  16,   7,  11, -35,   4,  -1,  -3,  -2, -17, -37, -15,   3,   0,   6,  -1,   8,  13,  15, -12, -26,   1,  -2,   0,   0, -12, -13,  28,  -1,  -9,   3,  -3,  13,   7, -16,  -2,  15,  -7,   2,  -2 },
      {  -4,   2,   4,  -8,  -3,   8,  -7,   4,  -7,  -4,   5,  27, -22,  25,  -9,   5,   2,  20,  41,  16, -36,   2,  -4,   3,  -6,  -3,  19,  10,  -3, -10,   3,  -3, -10,  29,   2, -24,   5,  -2,  -3,  -3,  10,  16,  12,  22, -25,   5,  -6,   2, -53,  25, -22,  27,  12, -17,   7,  -3, -22,  -5,   5, -12,  25,  -5,   0,   0 },
      {  -4,  -5,  -2, -17, -61,  33, -10,  10,   1,   4,  -2, -30, -28,  12,   1,   3,  -5,  -4, -37,  -7,  -2,  -9,   1,   2,   5,   6,   4,  25,  -6,  -2,  -3,   1,  -5,   4, -24,  24,  -1,  -3,   3,  -3,  27,  22,   9,  10,  -4,  -5,   0,  -1,   3, -23, -22,  18,  16,  -4,   1,   0,  30,  21, -28, -17,  10,   0,  -2,   0 },
      {  -3,   1, -10,   6,  15, -14,  10,  -9,  -7,  -7,  -1, -24,  27, -25,  13,  -7,  -5,  10, -26, -19,  22, -11,  11,  -4, -13, -36, -36,  23,   2,   4,   2,   4, -24,   1,  20,  14,  -7, -13,   1,  -1,   5,   4,  24,  19,  13,  -7,  -7,  -1, -60,  19, -13,   4,   9,  19,  -3,   1, -15,   5, -10, -22,  -3,   5,   4,  -1 },
      {  -8,  -9, -10, -49,   5,  -5,   8,   3,  -2,   1,  -9,  15,  17, -19,   5,   0,  -3,   1,  28, -16,  46, -21,   7,  -6,  -2,  31,  -9,  -7,  44,   0,  -1,   0,  -4,  -9, -33, -23,   1,  19,  -5,   6, -23,  15, -14,   1, -24,  14,  -1,   0, -10,  -5, -29,   6,  -3,   3,   5,  -2,  21,   9,  -7, -18,  -3,  -5,   3,  -2 },
      {   6,   7,  16,  44,   2,   1,  -1,   1,  -2,   1,  -8, -43,  33,  -9,   3,  -1,   4,  14,  17,  32, -13,   3,  -3,   2,   1,   2,  -2,  10,  19, -10,   6,   1,   0,  24,  27, -23,  12,   1,  -4,   5,   8,  39, -37,  25,   1,  13,  -4,  -1,  19,  -8, -26,  -7,   3,  10,   5,  -3,  38, -11,  -4,  -8,  -7,   2,   3,   0 },
      {   4,   6,  -3,  -8,  27,  -5,  -8,   5,   2,  -6, -14,  22,  -1,  16,  -6,   4,   9,  14,  14,  17,  -3,   4,  -8,   2,   3, -10, -37, -16,  -4,  14,  -4,   4,   1,   2, -20,  22, -25,   7,  -7,  -4,  33, -26, -15,  22,   9,  -8,   7,  -1, -29, -43,  13, -18,  22, -13,   6,   3,  48, -31, -13,  -2,   4,   6,  -8,   4 },
      {   3,   7,   8,  17,  33,   7, -13,   0,  -1,  -1,   7, -36,  -8,  27,  -3,   6,  -3,  -5, -14,   9,   3, -11,   8,  -5,   1,   2,  33, -29,  28,  -9,   0,  -2,   1,  22, -24,  10,  17,   7, -10,   1,   9, -26, -14, -27,   9,  25,  -2,   5, -44, -10,   0, -18, -26,   4,   6,   2, -10,  36, -24,  -3,  -3,  -9,   0,  -2 },
      {   2,   2,   4,   0,  11, -10,  12,  -5,   0,   2,  15,  10, -50,   1,   2,   2,  -1,  -2,   3, -51, -30,   0,  12,  -3,  -9, -10, -19, -25,  -8, -24,  10,   1,   0,   0,  15,  -9,  15, -22,  -4,   0, -13,   7,   3,  -6,  12,   9,  -7,   2,  -3,  23,  -9, -11, -13,  11,  -1,  -2,  67,  10,   5,  -7, -11,  -2,   0,  -3 },
      {   2,   7,  11,  21,   9, -40,  27, -11,  -7,  -7, -21, -22, -45, -18,  13,  -2,  -2,   3,  23,  30, -10,  -2,  -7,  -3,   7,  14, -11, -25,   8,  19,  -3,   0, -18, -29, -12,  15, -25,   3,  -1,   0,  -3,  30,  15,  12,  -6, -12,   2,  -1,   8,   8,  13,  19,  -4,  -6,  -4,   3,  -9,  33, -31,  10,  10,  -3,  -3,  -1 },
      {   4,   5,   9,   1,  37,  62, -23,   4,   1,   3,  -4,   9,  10,  42,  -5,  -5,   0,   1,  18,  -4,  -1, -22,  32, -14,   0,   8, -33,  -8,   3, -22,  19,   3,  -8, -24,   7,  17, -10, -15,  -5,   7,  -9,  28,   6,   8,   8,  -8,  -6,   0,  24,   0,   2,   9,   7,  12,  -5,   1, -24,   7, -18,   5,  -4,   5,   3,   2 },
      {   2,  -1,   3,   9,  22, -26,  -9,   7,   3,   2,   8,   2,  -3,   2,   4,   0,  -2,  -7,  -2,  -7, -18, -39,   6,   9,   7,   8,  11,  23,  33, -14, -11,   0,   3,  -6,  -1,   6,  14,  17, -18,   2,  16,   6,  46,  12,  -5,  -2,   0,  -1,  20, -15,  18, -32,  36, -11,   4,   0,  -1,  20,  44, -35,   6,  13,  -9,   4 },
      {  -2,  -4,  -3,   0, -17, -25,  13,  -5,  -3,  -3,  -3, -23, -17,   7,  -8,   1,  -4,  -6,  -9, -32, -24,  16,   6,   1,  -5,  -4, -22, -26,  30, -25,  16,  -4,  -4,  -4, -27,  -4,  19, -13,   4,   1,  -1,  -8, -27,  28,  -1,   1, -11,   0,   6, -39,  -7,   9,  12,  11,   0,  -2, -52, -38,  12,  -3,   1,   2,   4,   0 },
      {  -2,  -2,   0, -13,  -6, -28, -12,  -2,   1,   4,   4,  29,   2,  14, -34,  11,   2,   5,  11, -25,  11,  32, -32,   5,  -1,  -4, -21,  17,  13,   2,  -3,  -2,   5,   7,  45,  24,  24,  14, -10,   2,  13,  30, -11,  -2,  10,  21,   8,  -1,   4, -21,   7,  -9,  -7,  -1,  11,   0, -20,  40, -31,   5,  -1,  -2,   2,   3 },
      {  -3,  -4,   3, -13, -32,   6,  -3,   5,   1,   2,  -1,   0,  30,  -3,  10,  -1,   1,   4,  21,   8, -23,   8,  -2,   3,   2,  -1,  -5, -45, -28,  11,  -3,   3,  -8, -17,   0, -39,  -4, -14,   7,  -2,   6,  12,   3, -18,  26, -18,   0,   4, -20, -31,  -7, -29,  14,   9, -11,   8, -13,  47,  16, -25, -20,   6,  -3,  -3 },
      {   0,   1,   0,  -4,   3,  15,  18,   0,   2,   3,   0,  17,   0, -38,  34,  -2,  -3,  -1,  -6, -12, -34, -53,   8,  14,  -6,  -6, -12,   6, -29, -18, -32,  15,   0,   8,  11,  -4,  -3,  23, -30,   0,   5,   0, -36,  -3,  -3,  22,   5,  -8,  -4, -26,   9,  11,  -6,  -8,   6,  -7, -19,   8, -14,  21,   6, -10,   2,  -2 },
      {  -4,  -4,  -4, -13, -34, -36, -26,  29,   2,   3,   0,   7,  49, -11,  -6,   6,   2,   2,   4,  20, -21, -19,  13,  -4,  -4,  -4,  -3, -36,  -5, -26,   7,  -4,   0,   3, -17,  32,  25, -10, -13,  -1,  -5,  -8,  19,   9,   4,  16,  -8,   5,   9,  36,  14,   4,   2,   1,   0,  -3,  11,  -6, -30,  12,  12,   3,   0,  -1 },
      {   0,   1,   1,  -2, -19,  27,  52, -27,   2,   2,  -1,   9,  30,  16,  17, -16,   1,   1,   3,   9,  -3,  27, -12,   0,  -1,  -1,  -2,   0,  14, -24,   4,   5,  -1,   2,  -9,  28,  -1, -11,  -8,  12,   1, -11,   9,  36, -24,  16,  -7,   4,  -9,  -5,  36,   6, -25,  -2,   2,  -5,  15,  33,  33,   4,  -9,  -9,   2,  -3 },
      {  -2,  -3,  -2,  -4,  -4,  -3, -41,   3,  -1,   1,   0,   2,   5, -28, -15,   1,  -4,  -5,  -7, -13, -45,  13,  34, -19,   0,   2,   2,  24,  34,   9,  19,  -5,   1,   7,   4, -11, -29,  19,  16,   0,  -7, -15, -17,  18, -20, -31,  11,  -2, -14,   0,  25,   6,  -7,   1, -15,   1,   9,  29,  -3,  23, -28,  -4,   0,  -1 },
      {   1,   1,   2,   4,  19,  31, -30,  14,  -2,  -3,  -3,  -2, -12, -53, -42,  11,  -4,  -5,  -3,   6,  -4,  -5, -37,   1,   0,   2,  -1, -17, -13,  16,   1,  -1,  -4, -12, -18,  12,  13,   2,   0,  -5,   1,   7,  20,  33,  16,  28,  -1,  -4,  -5,  -9,  -8,  21, -23,  17,  16,  -7,  -1,   2,  37,   1, -17,  -4,   7,  -3 },
      {   2,   2,   1,  -2,   6,  47,  24, -14,   1,   4,   3,   8,   9, -16,  12,   3,  -2,  -3,  -3, -15, -34,  13, -35,  14,  -1,  -1,   3, -16,  43,  23, -21,  -5,   3,   6,  -8,   0,  25,  17,   4, -10,  -1,  -5,  18,   3,   9, -14,  13,  -3,   5,  31,  -4, -19,  28,   6,  -7,   4,  -4, -16, -42,   4,  -8,  18,   3,   1 },
      {   4,   4,   6,   1,  18,  24,   4,  21,  -2,  -5,  -4,  -7,  -7, -48,   0,   7,   3,   3,   4,  -3,  21,  13,  -9, -17,   0,  -7,  -2,  -3,  -7, -20,  23, -16,   3,   5,   7,  -4,  32, -33,  13,  -3,  -4, -11, -23,  -9, -22, -10,   5,  13,   1,  -7,  36,   5,  27, -33, -10,   3,   1,  20,  -9, -24,  39,  11, -12,   4 },
      {   0,  -1,   0,  -3,  -4,  10, -28, -33,  -1,  -3,  -4,   2,  -5, -22,  21, -33,   2,   1,   4,   3, -16,  27,  41, -13,   1,  -1,  -6, -15,  -6,  35,   7,  12,   2,  -1,  10,  30,  20,   7,  -1,   4,   4,  11,  13, -14, -36,  25,   6,  -1,   6,  -9, -12, -25, -20, -14,  11,   9,  -8, -23,  -1, -34,  10, -17,  -3,   3 },
      {   1,   2,   1,   1,   7,  -1,  18,  83,   1,   2,   2,   6,  -2,  21,  48,  25,  -1,   0,  -2,  -7,  -5,   9,  13, -11,  -1,  -2,  -4,   2,   7,  50, -12, -17,   1,   3,   9,  11,  21,  -5,   5,  -9,  -1,   6,  -4,  16,  -8,  -3,  -7,   4,   2,  -5,   4,   6, -25,  10,   2,   1,  -3,  -1,  11,  -5,  -6,  -8,   4,   0 },
      {  -1,  -2,  -3,  -3,   4,  -7, -39, -38,   0,   0,   0,   2,  -5,  17,  24, -13,   1,   3,   1,  -1,  16,  -1,  11,  31,   1,   2,   3,  -3,  -3,  34, -15,  17,   2,   2,  -6, -16,  49,  -6, -14,  -3,  -5,  -7, -23,  27,  21, -15, -13,  -5,   3,   6,  30,  44,   4,  12, -11,  -2,  11,  15,   5,  -6,   7,  21,   2,   1 },
      {   0,   1,   0,  -2,   1,   6,  -1,  14,   0,  -1,  -1,   0, -19, -12,   7,  12,   2,   0,  -1,   3,  17,  42,  21,   5,   3,   4,  13,  19, -28, -13,   9,   2,  -5,  -6, -18, -36, -12,  11, -44,   2,   1,  -7,  11,  32,  -3,  25, -26, -11,   6,  10,  25, -36,  10,  20,  13,  -5, -11,  -2, -35, -12, -20,  10,   8,   3 },
      {  -1,   0,   0,  -1,  -7,   1,   5, -13,   0,   1,   0,   3,   4,   6, -22,  30,  -1,   0,   1,   0, -17, -37, -18,  40,  -2,  -1,  -4,   2,  -1,  26,  19,  -5,   3,   6,   9,   6, -27, -14,  17,   3,  -8,  -8, -15, -11, -35,   6, -34,   7,   9,   2,  29,  -8, -19,  33,  -2,  -5,  -2,  -6, -17, -52,   5,   5,  11,  -6 },
      {   0,   0,   0,   1,   4,   4, -30,   7,   0,   1,   0,   4,  -1, -21,  29,  31,   1,   2,   4,   2,   1,  32,  -2,  53,   2,   3,   2,   1,  34, -14,  -8,  24,  -2,  -3,   3,  20, -28, -50, -27,  -2,  -1,   3,   5, -27,  -6,  -4, -16,  -8,  -4,  -6, -23,   1,   0,  -8,  -7, -15,  -1,   0,  12,  20,   5,   2, -10,  -5 },
      {   1,   2,   2,   4,  -3,   9,  37,  -6,  -1,  -2,  -3,  -7,   4,  -6, -30,  51,   3,   4,   5,  11,   5,  15,  51,  27,  -2,  -2,  -8, -11, -10,   2,  20,  -3,   3,   1,   7,  21,  20,  50, -23, -16,   0,  -1,   4, -22,   1, -12,   4,  -3,  -5,  -1,  -9,  25,  12,  -7, -17,  -2,   5,   2,  18,  -5,  -9,  12,   3,  -6 },
      {   1,   1,   4,   6,   0,  -1,  -9,  26,  -2,  -3,  -4,  -6,  -5,   4,  -9, -26,   1,   2,   6,   5,  -4,  26, -12, -22,  -4,  -2,  -8,  -9,   3, -31, -55,  22,   3,   1,   5,  12, -14,  17, -10,   0,  -6,  -9,  -7, -28, -18,  -8,  15, -16,   1,   7,  12,  35,   2,  32,  -8, -16,   1,  -6,   2, -50, -14,  21,   5,  -7 },
      {   0,  -1,   0,   2,   4,  -6, -21, -21,   1,   0,   0,  -2,  -3,   5,  14,  30,   1,   2,   3,   5,  16,  -8,  -7,  12,  -1,  -2,  -5,  -2, -20, -38, -29, -26,  -2,  -4,  -4,   2,  17,  -3,  14, -48,   1,   8,   5,  26, -39, -27,  28,   8,   0,  -3,  -8, -22, -37,  -4, -15,  19,  -6,  -7,  -8,  -2, -29,   8,  -8,  -2 },
      {  -1,   0,  -1,   2,   3,  -6, -15,  -7,  -1,   2,   0,   2,  -9,   3,  29,   1,  -1,   1,  -4,   1,   0,  -5, -34, -27,  -3,  -2,  -5,  -2, -17,   6,  24, -11,  -1,   8,  -1,  15,   6,  23, -12,  31,   1,   1,  -2, -14, -33,  14, -35,  29,  -5,  -5, -21,   3,  18,   7, -35,   1,  -1,   2,  10,  25, -27,  54, -10,  -8 },
      {  -1,  -2,  -1,  -1,  -3,   1, -24, -29,   2,   2,   2,   2,   6,   1,  53,  40,   0,  -2,   2,  -2,  -8,  14, -16, -26,  -1,  -1,  -4,   0,   7, -16,  17, -46,   0,  -3,   0,   0, -28,  25,   5, -16,   0,  -2,   0, -12,  37,  23,   9,  -7,   7,   4,  10,  27,   6,   6,  19,   2,   2,  -3,   9, -26,  17, -13,  11,   4 },
      {   0,   0,   0,   1,   5,  -4,  -3,  17,   0,  -1,  -1,   0,  -8,   6,  21,  -4,   2,   2,   4,   2,   8,   4,   9,  40,   0,  -2,  -2,  -4,  -6, -11,  22,  32,   0,   1,  -1,   2,   0,  15,  71,  -4,   1,  -1,   4,   5, -11,  52,  41, -12,  -3,   3,   7,   0,  29,  23,  -7,  -2,   0,  10,   2,  14,  -6,  11,   0,   4 },
      {   0,   1,   1,   0,   2,  -3,  -5,  14,   0,   0,   1,   0,  -1,  -1,  22,   3,   0,   0,   0,  -1,   1,  -4, -21,  23,   0,  -1,   1,  -3,  -4, -18,  39,  28,   2,   1,  -1,  -1,   9,  22,   7,  51,   2,  -1,   5,   4,  18, -37,  13,   4,   1,   6,   1,  13, -29, -42,  24, -15,  -7, -12, -16, -32, -39,  -6, -13,   6 },
      {   0,   0,  -1,   2,  -3,  -6,   6, -23,   1,   0,   2,  -2,   3,   6,  -4,  53,   0,  -1,  -2,   3,  -2, -12,   7, -36,   0,  -1,   3,  -2,  -6,  20, -11,  17,   0,   0,  -1,  -2,  12, -43, -11,  35,  -1,  -5,  -1,  -4, -28,   1,  44, -23,   0,   2,   1,   6,  26,   8,  45,  -2,  -1,   3,   1,   9, -25,  12,  12,  12 },
      {   0,   0,   0,   0,   2,  -3,  -1,   8,  -1,  -2,  -2,  -3,  -3,   2,   3, -46,   1,   2,   2,   4,   4,  -7,  -6,  27,  -1,  -4,  -3,  -5,  -2,   3,  18, -43,   0,   3,   5,  11,   4, -10, -11, -38,  -3,  -5,  -7, -22, -21,  -5, -16, -15,  -1,   2,   1,  16,  35, -11,  48,   1,   2,   8,   1,   7, -53, -12,  21,  18 },
      {   0,  -1,   0,   0,   5,   3, -13,  -7,  -1,  -2,  -1,  -4,  -6,  -4,   8,   9,   1,   2,   4,   6,  10,   6,   0,   3,  -2,  -3,  -6, -10, -12, -14, -19, -15,   1,   1,   3,   8,  21,  23,  28,  33,   0,  -1,   6,   2, -15, -37, -26,  -9,  -3,  -6,  -2, -24,  18,  20,   5, -56,   4,   7,   4,  19,  20, -43,  44, -14 },
      {   0,   0,   0,   2,   3,   2,  -4,  -3,   0,  -1,   0,  -1,  -4,  -5,  -8,   8,   0,   2,   0,   1,   3,  10,  19,  18,  -1,  -2,  -2,   1,   3,  -2, -43, -15,  -1,   1,  -1,   0,  -8,  -8,  11,  33,   1,   1,  -1,   1,   9,  21,   0,  85,   5,   5,   9,  16,  27,  -9,  -1,  28,  -1,  -5,  -3,  -7, -30, -33,   6,  -8 },
      {   0,   0,   0,   3,   3,  -4,  -3,  -1,   1,   1,   1,   0,   1,   5,   5,  20,  -1,  -1,  -1,   0,   3,  -6, -23, -29,   1,   2,   1,   1,   0,  -1,   9,  62,   2,   0,   1,   3,   8,   1,   2, -33,  -1,  -1,   0,   0,  -4,   5, -32, -17,   3,   2,   6,   3,  22, -24, -34,  35,   1,  -4,  -2,  -8, -25, -50,  36,  -9 },
      {   0,   0,   0,   1,   0,  -2,   2,  -7,   0,   0,   0,   0,   2,   6,  -9,   3,   0,   1,   1,   1,   1,  -6,  10,  -3,   0,  -1,  -2,   1,   1,  13, -21, -37,  -1,  -1,  -1,  -3,  -1, -20,  13,  14,   0,   0,   1,   6,  11,  37,  -5, -46,   1,   0,   3,   4,   4, -45, -53, -51,  -4,  -5, -10, -15, -35,  -9, -22,  -8 },
      {   0,   0,  -1,   0,   0,  -3,  -4,   6,   0,  -1,   0,  -1,  -3,   7,   5, -18,  -1,   0,  -1,  -1,   2,  -6, -17,  11,   1,   1,   2,   3,  -2,  11,  37, -14,   0,   0,  -1,  -3,   1, -12, -51,   2,   2,  -1,   1,   2,  -7,  -2,  67,  13,   0,   1,   2,  -3,   9,  20, -35, -13,  -1,  -6,   1,  -5,   3, -31,  22, -54 },
      {   0,   0,   1,   0,   1,   2,  -2,   3,   0,   0,  -1,  -2,   0,  -1,   0,   3,   1,   1,   0,   1,   2,   0, -12,  12,  -2,  -1,  -2,  -3,  -6,  -4,   2, -15,   1,   2,   0,   3,   9,   8, -16,  31,  -1,  -3,  -1,  -2, -12, -10,   1, -30,  -2,   5,  -1,   5,  12,  42, -26,  38,   1,   0,   4,   2,   2, -54, -59,  53 },
      {   0,   0,   0,   0,   0,  -1,  -2,   9,   0,   0,   0,  -1,  -2,  -3,  -2,  -7,   0,   0,   0,   0,  -2,   1,   1,  17,   0,   0,  -1,  -1,   2,  -5,  -4, -19,   0,   0,  -2,   1,  -7,  -5,  -8,  33,   2,   0,   2,   1,   3,   4,  15, -17,  -3,  -2,  -5,  -3, -20, -19, -35,  33,   3,   0,   4,  -3,  11,  24,  82,  56 },
      {   0,   0,   0,  -1,   1,   2,  -3,   6,  -1,  -1,  -1,   0,  -1,  -2,  -1,  -3,   1,   0,   1,   0,   2,   3,   4,  12,  -1,  -1,  -2,  -1,   0,  -7,  -9, -22,   0,   1,   1,   3,   2,   3,  13,  35,   0,  -1,  -1,   1,   1,  -3, -13, -51,  -1,   1,   2,   1,  -1,  -7,  18,  67,   1,   1,   1,   3,   5,  12,  -1, -79 },
    },
  },
  { // 23
    {
      { 103, -34, -14,  -3,  -2,  -1,  -1,   0, -52,  33,   3,   0,   0,   0,   0,   0, -18,  -9,   9,   2,   2,   0,   0,   0,  12,  -3,  -5,   1,   0,   0,   0,   0,  -4,   4,   0,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
      {  24, -44,  18,   0,   3,   0,   1,   0,  79,   7, -32,   1,  -6,   0,  -2,   0, -50,  46,   8,  -7,   2,  -1,   0,  -1, -18, -22,  17,   4,   0,   1,   0,   0,   4,   0, -12,   3,   0,   1,   0,   0,  -2,   3,   1,  -3,  -1,  -1,   0,   0,   0,   1,   1,   2,   0,   0,   0,   0,  -1,   0,   0,   1,   1,   0,   0,   0 },
      { -46,   1,   6,   8,   0,   2,   0,   1, -23,  40, -13,  -4,   1,  -1,  -1,   0, -51, -18,  41,  -2,   2,   0,   2,   0,  51, -52, -20,  14,   0,   1,   0,   1,  10,  25, -17, -10,   0,  -2,  -1,  -1,   1,  -1,  10,  -1,  -1,   0,   0,   0,   0,  -1,  -3,   0,   0,   1,   0,   0,   1,  -1,   2,   0,  -1,   0,   0,   0 },
      { -32, -76,  41,   3,   8,   1,   2,   4, -31,  34, -36,   8,   0,   2,   1,   0,  58,  12,  -3, -17,  -2,  -2,  -1,  -2,   0,   8,  10,   3,  -5,  -1,  -1,   1, -13,  -3,   4,   4,   3,  -3,   0,   0,   1,  -4,  -1,   0,   4,   2,  -1,   0,  -1,   4,  -4,  -5,  -1,   1,   1,   0,   2,   0,   2,  -1,  -1,   0,   0,   0 },
      {   4, -24,  -7,   2,   2,   2,   1,   2,  55,  15,   1, -15,  -4,  -3,   0,  -2,   9, -47,  11,  20,  -2,   2,  -1,   1,  14,  40, -52,  -3,   4,   1,  -1,   0, -33,  35,  33, -20,  -3,  -2,   1,  -1,  -5, -23,  11,  15,  -2,   0,  -1,   1,  -2,   5, -10,  -1,   5,   0,   1,   0,  -2,  -1,   2,  -3,   0,   1,   0,   0 },
      { -11,   2,  -9,   7,   1,   3,   1,  -2,  10,  65,  14, -22,  -6,  -5,  -1,  -3,   6, -42,  19,  15,   3,   2,  -1,   1, -73, -12,   4,   1,   8,   4,   2,   0,  22, -34,  -5,   1,  -3,   8,   2,   2,  19,   9, -18,  -3,  -9,  -3,   1,   0,   4,   3,  11,   7,   0,  -3,  -2,  -1,   2,  -4,  -7,   5,   2,   1,   0,   0 },
      {  22,  -3,   7,  -1,  -4,   2,   0,  -4,  16, -26,   6,  -1,   2,  -2,   0,   1,  56, -19,  15, -10,  -4,  -1,  -3,   1, -16, -36, -15,  34,  -2,   5,   0,   1,  13,  42, -55, -15,   7,   4,   2,   0, -14,  23,  28, -17, -11,  -2,   1,  -2,   3, -12,   5,  19,   0,  -2,  -2,   0,  -4,   4,  -5,  -1,   2,   0,   0,   0 },
      {   4,  37, -55,  21,  -8,   2,  -3,   4,   2,  31, -32, -20,  12,  -2,   2,   1,  27,  49,  41, -26, -18,  -4,  -2,  -1,  -9,   7,  10,  27,   1,  -6,   1,  -1, -25,  -2,  12, -10,  12,  -5,   0,  -1, -11, -20,  18,   4,   4,   4,  -1,   1,  -5,  -8,  -5,  -6,   1,   3,   2,   1,   0,  -1,   6,  -1,  -3,   0,   0,   0 },
      { -17, -56, -58,  44,   7,   7,   3,   3,   3, -19,  44, -34,   1,  -1,   0,   0,   8,  23,   1,  17, -10,  -2,  -2,   0,  26,  -7, -12,  -3,   8,  -2,   2,  -1,  15, -15,  -7,   5,  -5,   3,  -2,   2,  -9,  27, -18,  -3,   2,  -2,   1,  -1,  -3,   5,  13,  -8,  -1,   0,   0,   0,  -3,   1,   0,   4,  -1,  -1,   0,   0 },
      {   5,  11, -14,  -3,   2,   2,   2,  -2,  28,  21,   9,   4,  -9,  -1,  -3,  -2,  15, -24,  -8, -15,  13,   2,   2,  -1,  43,  -8,  32,   0, -21,   4,  -5,   1, -64,  -7, -24,  42,  12,   5,   4,   1,  17,   6, -26, -35,   1,   1,   2,   1,   6,   8,  20,   9, -13,  -3,  -3,  -1,   0,  -3,  -8,   7,   7,  -2,   1,   0 },
      {   7,  -7,  12, -10,   2,  -1,   0,   6,  -4, -46,  -4,   6,   9,   1,   5,  -1,   1,  -6,  48,  -3, -11,   5,  -3,  -3,  -1, -14, -19,  34,   5,   1,  -2,   3, -24, -45,  -2, -16,  11,   4,   4,   1,  46,  -7, -33,   9, -14,   0,  -3,   3,  -5,  45, -21, -11,   6,  -6,   1,  -2,   0,   0,  11,  -7,   1,   2,  -1,   0 },
      { -12, -16,   7,  -5,   4,   3,   3,  -5, -14,  -1,  62,  -3, -11,   0,  -4,  -1, -10,  27,  -2,   8,   4,  -1,  -1,  -1, -23, -25,   8,  10,   4,   5,   1,  -1, -40,   6,   5,  -1,   5,  10,   5,   1,  -1, -57,  17,   3, -13,   1,   3,   2,  35, -18, -28,  30,  -3,  -5,  -3,   0,   0,  12, -15,  -4,   6,  -1,   1,  -1 },
      {  -1, -14,  19,  46, -16,   7,  -5,  -7,  -6, -24,   2, -20,  25,  -4,   1,   1, -32, -34,  15, -38, -11,   9,   1,   7, -10,  45,  15,  29, -29, -12,   2,  -4,  17,   0,  -2,  24,  27,  -3,  -4,   0,  -7, -18,  -4,  -8,   2,  14,   3,   0,   8, -22,   2,  11, -10,  -1,   1,   1,   1,   5, -11,   6,  -1,  -3,   0,   0 },
      {  22,  17,  35,  10,  -8,  -4,  -3, -10,  22,   4,  15, -25,   3,  -1,  -3,   3,  35,  -9,  20,  -6,  -4,  -6,   1,   3,  29, -38,  19,  -9,   1,   2,   2,  -1,  26, -10,  23,  32, -24,   8,  -8,   1, -30, -19, -28,  44,  12,  -1,   5,  -2,   3,  -4, -22, -12,   9,   4,   0,   1,  -7,   8,   0,   2,  -7,   0,   1,   0 },
      { -10, -22, -56, -44,  18,   9,   5,  17,   5, -12,  -5,  28, -12,  -2,  10,  -6,   2, -24,  26, -19,  20,   3,  -3,  -9,  -6,   4,  23,  -7, -26,  14,  -9,   3,  35,   5,   1,  21,  -4, -18,   7,  -3, -18, -13,   2,   2,  14,   2,  -7,   5,  16,  10, -38,  -2,   0,   1,   4,  -1,   0,  11,   0, -16,   5,   0,   0,   1 },
      {  -2,   6,  17, -14,   7,  -5,  -2,   4, -11,  -4,   6,  -4,  -8,   3,   0,   1,  -8,   8,  20,  35,  -5,  -8,  -2,   0, -29,   2, -20,  13,   7,   2,  -1,   1, -18,  17, -12,  24,  12, -15,   0,  -3, -49,  -8, -27, -30,  41,  13,   1,   2,  11,  11,  14, -48, -29,   8,   2,   0,  -4,   2,  22,  16,  -7,  -9,  -2,  -1 },
      {  -3,  -2,  19, -63,  14,  -3,   5,   1,   7,  12,  24, -13,  -8,  -3,  -1,   0,   0,  11,  35,  13, -18,  -4,  -5,   0,  33,  39,  10,  22,   9,  -8,  -3,  -4,  13, -40,  -2,  -7,  13,   7,   1,   1, -27,  19,  15,  -7, -13,   5,   1,   0, -14, -29,  13,  17,   1,  -3,   2,   0,   6,  -7, -17,   5,   2,   2,  -2,   1 },
      {   1,  -4,  -9,   4,  31,  -7,   2,  -5, -17, -38, -55, -38, -15,  20,   3,   4,  -3, -19,  27,  30, -13, -12,   0,  -4,  -4,   0,  28, -26,  26,   7,  -5,   0, -15,   9,   1,   7, -25,  17,   3,   1,   8,  -6,   2,  -2,  -5,  -9,   4,  -1,  22, -19,  12,  19,  -9,   0,  -6,   1,   3,   3, -23,  14,   7,  -2,   2,  -2 },
      {  -4,  -8,   6, -15,   2,  -1,   2,  -6,  -6, -17,  28,   3,   3,  -6,  -1,  -1,   1,  13,  45, -32,  17,   1,  -1,   3, -23, -13, -11, -26, -24,  10,   5,   2, -15,  25,  35,  16, -18, -13,  -4,  -2,  26,  15,   7,  11,  15,  -1,  -4,  -3, -34, -16,  41,   3,  -2,   9,   1,   1,   6, -33, -10,  23,  -2,   3,   2,   2 },
      {  -8,  -6,   1,  15, -19,   4,   0,   8,  -6,  -4,   4,  20,   8,  -2,   3,   0, -17, -16,   9,  11,   1,   1,   1,  -1, -16,  10,  14,  10,   6,  11,  -3,   2, -49, -14, -31,   9, -31,  -2,   8,   2, -26,  33,   7,  28,  17, -30,  -2,  -3, -22, -31, -20, -15,  39,  14,  -1,   2,  14,   5,  -7, -27, -15,   5,   3,  -1 },
      {   1,   4,  10, -20,  27, -15,   2,   6,   0,   8,  14, -64,   7,   8,  10,  -2,  -4,   4, -23,  10, -28,   9,  -2,  -1,  -2,   2,  22,  14, -22,  -3,  -1,   2,   2,  23, -14,   9,  -2, -31,   1,  -3,  46,   8,  19,   4,  30,  -2, -14,   1, -20,  10, -24,  -9,   4,  18,   6,   0,  -3,  10,   9, -19,  -4,  -4,   4,   1 },
      {   6,  -1,  -1,  27,  -4,  -6,  -7,  -5,  11,   0,   1,  35, -22,   7, -11,   7,  14,   0,  13,  27,  12, -27,   7,   6,  13,   4,  -9,  -3,  17,   2,   5,  -2,  21, -22, -19,   7,  10, -18,  -2,  -2,  41, -34,  16,  -9,  33,  16,  -9,   1, -36, -30, -10,   2, -22,  24,   5,   3,  13,  15, -18,   0,  -8, -14,   2,  -1 },
      {  -8,  -6,  -9,  -9, -16,  11,   1, -12,  -6,   4,  12,  15, -18,   0, -13,   6,  -4,   5,   9,  15, -14, -22,   5,   6,  -3,  26,  39,  18,  17, -11,  -6,  -6,  14,  52, -20,  12,   1,  30,  -1,  -1,  13, -27, -24,  17, -14, -11,  16,  -1, -40,  25,  10,   1,  19,  -8,  -7,  -1, -16, -19,  18,   0,  -1,   5,   1,  -2 },
      {   5,  19,  23,  23,  -5,   3,  -5,   4,  -5,  12,   3, -10, -28,  -3,  11,  -6,   4,  39,  26,   3,  27,  -1,  -9,  -9,   9,  46, -31,  -3, -17,  33,  -3,   2,  16,   7, -27,  19, -26,  -5,  16,   1,  12,   9,  -8,  -1,   0, -21,  -4,   2,  30,  22,  -6,  21,  10,  -3,  -5,  -4, -14,   4, -10,  -9,  15,   5,   7,   0 },
      {  -3, -11,   1, -27, -80,  29,  -6,  14,   2,  -6,  -6, -24,  43, -32,   7,  -5,   7,   4,   8,  32,  20,   1,  -7,  -1,  10,   0,  24, -17,   2,  13,  -3,  -2,   4,   4,  -3, -17,  -4,  -8,   7,   0,  17,  -9,   3, -12,   3,  -2,  -5,   4,  15,  -2,   8,  -5, -12,   7,  -3,  -1,   4,  11,   2,   8,  -5,  -3,   1,   0 },
      {   2,   0,  15,  20,  15, -13,  -3,   5,   1,  -4,  22, -20,  -2,   3,   5,  -4,  -5, -16,  12, -33,  37, -22,   7,  -2,   2,  11,  33, -23,  22,  18, -10,   4,  -6, -18, -14, -37, -14,   2,   7,   1, -22, -14,  37, -17,  -7,   1,  -9,   4, -21,  29,   4, -10, -14,   4,   1,  -3, -15,  -3,  41,   1,  -5,  -6,  -1,   0 },
      {   0,   5,   6,  17,  -3,  12,  -7,   8, -10, -10, -17, -22, -20, -13,  12,  -3,   1,   2,  -5,  37,  35,   8, -11,  -6,  -6, -16,  10,   8, -18,  11,   4,  -1,  -8,   2,  18,   5,  24, -13,  -2,   1, -30,   6,  -5,  -2, -15,  29,  -1,   2, -56,  14, -17,  26,  -3, -23,  14,  -4,  13, -21, -16, -23,  25,   3,  -5,   1 },
      {   2,   2,  -2, -12,  12,  27,  -7,   6,   0,  -3,  -3, -27, -19,  -6,  12,  -3,  -2,   2, -26, -32,  17, -25,  -5,  -1,   3,  -1, -15,  26,   8,  25, -14,  -4,   9,   5,  -1,  -6, -10,  16,  14,  -1,   7, -17, -22,  -3,  -5,  -9,  10,   4,  -5, -19,   6, -23,   0,  -5,  -5,  -2,  81,   0,   3,   3, -13,   1,  -5,  -1 },
      {   2,   4,   2,   7,  22, -37,  14,   0,   4,   7,   7,   8,  50,   0,  -3,  -7,   8,  13,  18,  13,  -3,  46,   6,  -6,   1,   9,  -1, -41,  -7,  -9,   3,   9,   9,  10, -33,   0,   1,   8,   0,   1,  -9, -32, -11, -14, -15,  -7,   5,   3, -16,  10,   0,  -6,   6, -11,  -7,   0,  53,  -2, -11, -11,   1,   7,   0,  -1 },
      {   7,   4,   9,  25,  19,  13, -15,  30,   7,   3,   9,  28,   9,  -4,  27, -14,   4,  -7,  -8,  17, -27,   6,  -7,  -8,  10,  -8,  27,  23,   7,  10, -14,  -1,  11,   0,  12, -17,  -2, -23,  24,   0,  -5, -17,   6,  -1,  15, -18, -17,   9,  19,  15,  15,  -1,  10,  14,  -1,  -4,   4, -60, -26,  23,   3,   8,   6,   5 },
      {  -7, -11, -16, -20, -47, -54,   9,   1,  -3,  -6, -13, -18,  -9,  41, -17,   5,  -2,  -6, -23,  -7,   9,   7,  12,  -1,  -1,  -8, -20,  30, -14,   5,   2,   1,   5, -23,  -5,   9, -21,  -5,   8,   0, -20, -31,   9,   2,   8, -21,  -5,   1, -13,  19,  17,  17,   0,  10,  -7,  -4,   0, -23, -12,  14,   6,  -3,   7,   0 },
      {  -1,   0,  -9, -15,  39,  34,  10, -29,  -4,  -4, -20, -12,  32, -39,  -2,   9,  -3,   0, -21,   0,  13,   0, -13,  13,   1,  -2, -25,   0,  -5,  20,  18,  -7,  -2, -23, -33,   8,  -7,  -3,  -1,   7,  -8, -40,  -3,  10,   4,  -3,  -1,  -5, -15,  -9,  13,   2,  12,   7,  -2,   2, -45, -20,  -6,   2, -10,   3,   5,   1 },
      {   3,  -3,  -1,   8,   1, -19,  19, -19,   1,  -8,  -3,  -4, -43, -13, -17,  10,   6,  -3,   8,  19,  33,  31,  -8,  -1,   9,   4,  26,  15, -18, -10,  34,  -3,   3,  -2,  -4, -25,  10, -22, -12,   8,  13,  -6,   1,  10,  -1,   3,  -2, -10,  37, -22,  18, -26,  26,  -5,   1,   2,  12, -22,  31,  -3, -16,  16,  -8,   4 },
      {   2,   1,   5,   0, -14,  55,  -8,  -1,   0,  -2,   0,  -1, -29,  14,  -2,  11,   0,  -4,   5,   5, -49,   1,   6,   6,  -2,  -4,  -9, -35, -34, -13,   8,   0,  -6, -24, -25,  12,  -9, -18, -11,  -5, -10, -14,  36,  13,  -7,   2,   4,  -7,   2,  27,  10,   9,   1, -12,   3,   1,  31,  -4,  26,  11,   0,  -1,  -6,   0 },
      {  -2,   2,  -5,   2,  11,   1,   0, -10,   0,   8,   3,  18,   4,   7,  -7,   5,  -3,   3,  -4,   5,  -6,  27, -14,   7,   2,  12,  19,  23, -14,  30,  20,  -6,  -7,  11,  -9, -31, -34, -12,   3,  12,  -7,   9, -23,  34, -15,  15, -15,  -3, -13,   7,  -3,  11, -37,   1,  14,  -5,   8,  38,  -3,  51,   0, -28,  -3,  -1 },
      {   2,   2,   6,   6,  -2,  11,  10,  17,  -3,  -6,  -2,   1, -39, -28,   3,  -1,   0,  -4,   1, -10, -22,  50,  -5, -10,   1, -11,   2,   3,   8,  -1,   5,   3,   6,  -3,  24,  -3,  -9,  -7,  18,  -2,  -6, -20,  -2, -34,  -1, -33,   3,  12, -34,  -7,  25,  -4,   7,   3, -25,   1, -19,  53,   0,  -7,   1,   7,   7,  -9 },
      {   1,   1,   0,   9,  -6,  -8,  56, -15,  -2,  -2,  -3,   7,   2, -48, -27,   9,  -3,  -3,  -2,  -1, -14,   3, -27,  -5,   1,   2,   0,  14,  29,  12,   6,  -7,   3,  -1,  18,  38, -14,  15,   8,   0,   4,  18,  37, -11,   8,   5,   1,   3,   7,  28, -24,   9, -15,   5,  -2,  -2,  27, -18,  12,   8,   0, -12,   2,  -2 },
      {   2,   2,   4,   8,  12,  -3,  40, -22,   1,   1,   1,  10,  26,   7, -12,  13,   0,  -1,   5,  18,  -5, -62,   1,  11,   0,  -6,   8,   4, -36,  10,   1,  -9,   2,   0,  23, -19,  14, -23,  17,  -1, -11,   3, -13,  -7,   6, -26, -13,   5,   3,   9,  25,  23,  15,   2, -17,  -9,  15,  32,  13, -18,  14,  11,   1,   1 },
      {  -1,   1,   1,  -9, -13,   2,  12,  -2,  -3,  -1,   3, -11, -23, -20,  -2,   3,  -4,  -3,  -1, -35,  -9, -10,   0,  -1,  -4,   4,   1, -20,  37, -23,  23,  -3,  -7,  12, -18, -21,  11, -23, -20,  19, -15,  -2, -27,  11,  13,  13, -34,  -9,  10,  31,  12,  11,   4,  28,   9, -11,  15,   4, -47, -24,   4,   1,  13,   8 },
      {   3,   3,   6,   2, -16,  25,  47,   8,   2,   2,   4,  -2,  -1,  47, -13,   3,   1,   0,   3, -14, -25,  11, -11,  -8,   2,   0,  -2, -17,   3,  25,  10, -10,   4,   4,  -1, -18, -10,  -3,  21,   8,   4,  -4, -30, -24,  -5,   6,  -6,   2,  -5, -36, -16,  -5, -18,  -6,  18,  -1, -17, -39,  11, -37,   5, -16,  -4,   4 },
      {  -5,  -8,  -7,  -4, -16, -49, -30, -20,   1,   0,   0,   9,  -4, -19,  50,  -3,  -2,  -5,  -3,  -9, -35, -12, -29,  15,   1,  -4,  -4, -14,  20,  29,  15, -22,  -1,   0,  -6,  19,   5, -22,  14,  12,   3,  -8,  -2,   5, -14,   5,  -9,  -1,   0, -17,  11,   2,   4, -26,  10,  -3,   1,   1,  26, -12,  20,   3, -21,   4 },
      {  -1,  -3,  -1,  -6,   9,  17, -67, -33,   1,   0,   1,   4,  16,  21, -34,   9,   3,   1,   8,   5,   3,  24,   7,  15,   2,   1,   6,  14,  18,  17,  24, -12,   4,   4,  25,  13, -12,   2,   8,   7,  -4,   7,   4, -27,  -1,  -2, -14,  -2,   9,  17,   9,  16,   1,  19,   2,  -7,  14,  -3,  12, -43, -19,   0,  14,   1 },
      {   2,   1,   2,   0,   7,  13,   7,  17,   4,   3,   5,   8,  24,   6,  20,   5,   1,  -1,  -1,   0,  15,  -6,   2, -15,   2,   4,   2,  32,  21, -33,  18,   8,   1,  -1,   5,   5, -61, -16, -38,   5,   7, -12,  -1, -19,  24,   7,  12, -21,   2,  -2,  17,  18,   2, -29,  12,  10,   5,   7,  13, -15,  36,   3, -23,   3 },
      {   0,   1,   2,   1,  -3,  15,  14,   2,   2,   4,   7,   1,  12,  54,  13,  15,  -1,  -2,   1,  -2,  17,  -3, -53,  -9,   0,   0,   4,   2,  14,  10,  21, -18,   2,   3,   5,  21,  16,  -3,  -1,  12,   4,  -2,  31,  -1, -19,  11,   4,  -2,  -4,  22,  12, -42,  22,  -9,  -6,   7,  -5,  24, -38,  10, -11,  22,  -3,  -3 },
      {  -1,  -1,  -2,   1,  -8,  -7,   1, -39,   1,   2,   2,   8, -12,  -1,  46,   7,   0,   1,   3,   4,  -4,   7, -34,  14,   0,  -2,   5,  -1, -35, -17, -12, -10,  -1,  -5,   3, -34, -19,  47,   4,  -7,  -3,  -1,   3,  -8,  41,   6,  23,  -2,   1,  18,   6,   2, -19,   6,   4,  15,   0,  -4, -21, -29, -23,  -8,  -3,  -2 },
      {   0,  -1,   1,   1,   1, -12, -16,  12,   0,  -1,   3,   0,  -1, -23, -53,  13,  -2,  -3,  -2,  -4, -21,  -7,  11,   1,   1,   1,   9,   2, -12,  43, -10,   8,   0,  -3,  -2, -16,  -6,  -1,  -7, -17,   4,  -4,  14,  13,   6,  10,  32,  -6,   9,  10,  23, -39,  -9, -23,  11,  15,  -7,   3, -40, -31,  34,  -2, -19,  -1 },
      {  -1,  -1,  -1,  -3,  -7,  14,   3, -67,   1,   2,   3,  -4,  -4,   1,  23,  22,   2,   2,   3,  -5,   5,   6,  50,   9,   1,   3,   5,   2,  10,  -8,  -2,   0,   0,   1,   6,  -7,  -5, -27,   7,   5,   0,  10,   4, -22,  -7, -25,   1,   9,  -4,   3, -36, -34,  -1, -16, -11,   7,   0, -11, -27,  21,  32,  -6,  -6,  -8 },
      {  -1,  -1,   1,  -4,   4,   0, -33,  15,   0,   0,   2,  -7,   3,   1, -32,  20,  -1,  -3,   0, -13,   5,  -6, -40, -11,   0,   1,   0, -17,   1, -21,  40, -18,  -2,  -1, -10, -21,  16,   9,  18,  -2,  -1,   8, -10,   0,  45, -27,  12,  12,  -7,  -4, -32,  12,  -4, -18, -30,   2,   3, -11,   5,  26,  30,   9,  -9, -13 },
      {  -2,  -1,  -1,   0,  -5, -12, -10,  -4,   1,   3,   3,   0,   1,  -5,   5,  63,  -2,  -1,  -2,  -4, -15,  -3,  -4, -46,  -1,   2,   0,  -7,   0,  34, -21,  17,   0,   2,   4, -12,   2, -19, -29,  -4,  -3,   1, -19, -28,  -2,  13,  22, -20,  -5,   1, -18,  27,  24,  14,  -1,  18,   0,  -4,   3,  20, -34,  17,  -1,  -1 },
      {   0,  -1,   0,  -2,   1,  16,  14, -20,   1,   2,   1,   0,   8,  20,  15,  17,   1,  -1,  -1,  -4,  15,  37,  10, -24,   2,   1,   1,   3,  36,  17, -29,   3,   0,  -2,  -7,   5,  41,  -8,  10, -14,   1,   4,   4,  54,  34,  -8,  14,   8,   6,   2,  30,  23, -25,   0,  -5,   9,   6,   6,  20, -15,  18,   6,  -4,  -4 },
      {   0,   0,   0,   1,  -2,   4,   7,  -8,  -2,  -1,  -2,  -6,  -3,   9, -26, -14,   1,   2,   2,  -7,   6,  22, -14,  55,   0,  -2,  -1,  -8,  19,  -3, -50, -20,  -1,  -1,  -3,  -9,  -5, -16,  15, -23,   0,  -2,  -9, -18,  27,  24,   2,  -5,  -1,   3,  -6,  16,  49, -13,  27,  -4,   1,  13,   0,  26,   5,  12, -17,  20 },
      {   0,   0,   0,  -2,  -5,   5,   9, -14,  -1,  -1,  -1,  -1,  -1,   8,   6, -65,  -1,  -1,  -1,  -2, -12,  -5,  27, -13,   3,   2,   4,  -1,   6,  40,  51,  26,   2,   0,   7,  -6,  24,  17,  -8,   4,  -1,   2,   3, -10,  35,  -8,  17,  -7,  -4,  12,  -5,  15,  23, -15,  16,  19,   1,   9,  -4,  18, -18,  -1, -15,  12 },
      {   0,   0,   1,   0,   3,  -1,  -9,  -4,   0,  -2,   0,  -3,   3, -12,  -8,  19,   1,   0,   1,   3,   5,  -4,  21, -62,  -2,  -4,  -5,  -3, -10, -24,  -7, -33,   1,   1,   1,   5,   2,  16,  49,  27,   2,  -4,   7,   2,   7,   2, -11,  -1,   0,   3,  14,  -4,  24, -14,  47,  -3,  -3,  12,  -4,   7,  -8, -37, -14,  29 },
      {   0,  -2,  -1,   1,  -4,   6,  -5, -19,   1,   0,   0,   3,  -1,   5,   2,   2,   0,  -2,  -2,   3,  -7,   8, -20, -17,   1,  -1,   0,  -1, -12,  17, -10,  23,   3,  -4,   2,  -5,   9,  46, -28,   1,   1,  -7,   6, -10,  11,  -2, -34, -20,  -4, -12,  -1, -21,  19,  25,  23, -28,   5,   7,  16,  13,  66,   9,  34,   9 },
      {   1,   2,   2,   0,   6,  -2,   8,  39,   0,   0,   0,  -1,   0,  -2,  18,  35,   2,   2,   5,   3,  11,  10,  13,  48,  -1,  -1,   1,  -3,   2,  14,  12, -30,   2,   2,   8,   4,  28,  14, -25,   6,  -1,   0,   3,  -5,  -4, -56,  12, -37,   2,   7,  -4,   3,  -2,  17,  30,  12,   3,   0,   1,   2,   4, -34, -11,  11 },
      {   0,   1,   0,   1,  -1,  -1,   9,  22,   0,   1,  -1,   1,  -1,  -3,   6,  18,   0,   2,   0,   2,  -8,   8,  45,  11,  -1,   1,  -2,   2,  -7,  24,  -4, -45,   0,   0,  -1,  -2,  -6,  30, -17,  42,   4,   2,   8,   0,  29,  21, -28,  14,  -2,   0,  -5,   1, -13, -30, -16, -18,  -2,   2,  -1,   4, -11,  58,   4,  -2 },
      {   0,   0,   0,   1,  -1,   1,  -7, -18,  -1,  -2,  -1,   0,  -2,  -1, -26, -43,   0,   0,   1,   1,  -2,  -1, -22, -15,  -2,  -2,  -3,  -2,  -5,  -9, -43, -14,   1,  -1,   1,  -1,   3, -18, -13,  67,  -1,  -2,   0,  -1,  -6, -34,   9, -36,  -6,   0,  -9,   0, -29,  13,  10,  25,   0,  -1,   2,  -1,   1,  29, -17, -11 },
      {   0,   0,   0,  -1,   2,  -1,   2,   3,   0,  -1,   0,   1,   1,   3,   7, -24,   1,   0,   1,   1,   0,   3,  12, -12,   0,  -1,   0,  -2,  -4,  11,   1, -61,  -2,  -2,  -2,  -7,  -4,  -8, -19,  17,   0,   2,   0,  -3,  -2,  32,  51,  23,   4,   1,   3,   2,  32,  41, -24,   2,   4,  -2,  10,   1,  31, -34,  21, -26 },
      {   0,   0,   0,  -1,   0,   3,   0,  12,   0,   0,   0,  -1,  -1,   4,   1,  21,   0,   2,   0,   1,   1,   5,  -1,  30,   0,   0,  -1,  -1,  -2,   3,  -7,  58,   3,   2,   3,   3,   6,  15,  24,  67,   1,   1,   3,   3,  10,  28, -10,   5,   3,   1,   6,   4,  26,  14, -34,  19,   1,  -2,   5,  -2,  14, -30, -37, -14 },
      {   0,   0,   0,   0,   0,   0,  -3,  -6,   1,   0,   1,   1,   0,   3,  -7,  -9,   1,   0,   1,   0,   0,   3, -15,  -8,   1,  -1,   1,  -1,   3,   9, -28,  -2,   1,  -1,   3,  -3,  10,  -9, -40,   6,   1,  -2,   3,  -2,  17, -17, -27, -11,   1,  -1,   5,   1,  14, -60, -23, -33,  -2,  -4,  -7,  -5, -26, -62,  16, -37 },
      {   0,   0,   0,   1,   0,   1,  -4,  -6,   0,   0,  -1,   0,  -1,   2,  -7,   6,   0,   0,   0,   0,   1,   0, -17,   1,  -1,   1,  -1,   0,  -1,   2,  -8,  12,  -1,   0,  -1,   1,   0,  -2, -48,  18,   1,  -1,   0,  -3,  -3, -27,   9,  96,   1,  -1,   2,   3,  -1,   7,  28, -19,  -2,   0,  -2,  -1,  -7,  -3, -31,  29 },
      {   0,  -1,  -1,  -1,   0,  -1,  -1,   1,   1,   0,   1,   1,   1,   1,   4,   5,   0,  -1,   0,  -1,   0,  -1,  -3,  11,   1,   0,   2,  -1,   4,  -2,   1,  22,   0,  -2,   2,  -3,   1, -17,  14,  36,   4,   3,   5,   3,   7,   1,  62, -20,  -2,  -1,  -5,  -3, -11, -14, -14, -61,  -1,   1,  -2,  -1,  -4,  -1,  41,  60 },
      {   0,   0,   0,   1,   0,   4,  -3,  -7,   0,   0,   0,   1,  -2,   3,  -1, -13,   0,   1,   0,   2,  -1,   7,   0, -18,   0,   0,  -1,   2,  -3,   7,  -7, -23,  -1,   0,   0,   1,  -1,   8, -19, -23,  -1,   1,   0,   0,  -6,   4, -23, -22,  -4,   0,  -6,  -4,  -4,  18, -67,  -7,  -3,  -3,  -4,  -5,   3, -14, -56,  69 },
      {   0,   0,   0,   0,   0,   2,   3,  -3,   0,   0,   0,   0,   1,   0,   4,  -1,   0,   0,   1,   0,   0,   3,   8,  -8,   0,   0,   0,  -1,  -1,  -1,   8,   3,   1,   0,   1,   1,   1,   4,  10, -11,   1,   0,   2,  -2,   0,  -1,  18, -20,   0,   2,  -1,   1,   0,  27,  15, -84,  -5,   0,  -7,   1,  -6,  14, -63, -53 },
    },
    {
      {-116,  18,   7,   2,   2,   1,   1,   0,   9, -18,   4,  -2,   2,  -1,   1,   0,  36, -16,  -9,   1,  -2,   0,   0,   0,  11,  15,  -6,  -2,  -1,   0,   0,   0,  -1,   1,   3,  -1,   1,  -1,   0,   0,   2,   4,  -3,  -2,   0,  -1,   0,   0,   0,   4,   5,  -2,   0,  -1,   0,   0,   0,  -1,   0,   0,   1,   0,   0,   0 },
      {  -2,  -9,   6,   1,   1,   0,   0,   0,  96, -30, -15,  -1,  -3,   0,  -1,   0, -10,  51, -15,   0,  -3,   0,  -2,   0, -45,  17,  14,  -3,   3,  -1,   1,  -1,  -3, -11,   8,   0,   0,   0,   0,   0,  -4,  -5,   2,   2,   0,   0,   0,   0,  -1,  -3,  -2,   1,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0 },
      { -37,  25,   4,   2,   1,   1,   1,   0,   2,  21,  -7,   0,  -3,   0,  -1,   0, -69,  26,  26,  -4,   4,  -1,   1,   0,   4, -68,  25,   4,   3,   2,   1,   1,  35, -22, -22,   8,  -3,   2,  -1,   1,   4,   7,  -9,   0,   1,   0,   0,   0,   5,   0,  -5,  -1,   0,   0,   0,   0,   1,   2,   0,   0,  -1,   0,   0,   0 },
      {  12,  69,  -5,   0,  -2,   0,  -1,   0, -38, -14,  21,   4,   2,   2,   0,   1,  33,  12, -11,  -3,   0,   0,  -1,   1, -57,   6,  31,  -3,   2,  -2,   1,  -1,  11, -50,  10,   7,   0,   2,   0,   1,  13, -14,  -7,  10,  -3,   3,  -1,   1,   4,  -8, -15,   6,   1,   1,   0,   0,   4,   5,  -4,  -1,  -1,   0,   0,   0 },
      {  19,  62,   5,   2,  -2,   0,  -1,   0,  24, -43,   7,   7,  -1,   2,  -1,   1,  16, -14,   6,   1,   1,   0,  -1,   2,   6, -37, -23,  13,  -2,   4,   0,   1,   8,  54, -32, -13,  -2,  -2,  -1,  -1, -22,   5,  40,  -2,  -1,   0,  -1,   0,   4, -22,  -9,  11,  -2,   2,  -1,   1,  -1,   5,  -1,   1,  -1,   0,   0,   0 },
      { -10,  42, -10,   3,  -2,   1,  -1,   0,  14,  52, -11,  -8,  -3,  -2,  -1,  -2, -29,  -5,  24,  -1,   4,   0,   1,   0,   4,  22, -12,  -4,  -1,  -1,   0,   0, -36,  17,  47, -12,   3,  -3,   1,  -2,   4, -61,  25,  16,  -2,   4,   0,   2,   7, -11, -20,  10,   0,   1,   0,   1,   2,  -2,  -4,  -1,   1,   0,   0,   0 },
      {  17,  39, -30,  -2,  -5,  -1,  -2,   0,  48,   8,  18, -12,  -2,  -4,   0,  -2, -13, -63,  14,  17,   2,   4,   1,   1,  -1,   7, -21,   4,   0,   0,   1,   1,  -3, -24,  -5,  11,   3,   0,   1,   1,   9,  21, -47,  -9,   4,  -1,   1,  -1,  -9,  31,  14, -23,   1,  -3,   0,  -2,   1,  -4,  12,  -1,  -3,   1,  -1,   0 },
      {   0,  32, -31,  -7,  -5,  -1,  -1,   0,  -1,  53, -17, -10,  -1,  -2,   0,  -2,  11,  30, -10, -12,  -1,  -3,   0,  -3,  -4,  23,  17,  -9,   1,  -2,   1,  -1,  -3,  14, -13,  -2,   2,  -1,   2,  -1, -18,  48,  20, -33,   3,  -6,   2,  -3,  -4, -19,  59,   0,  -3,   0,  -1,   0,   2, -20,   1,  13,   0,   2,   0,   1 },
      {  -8,   7, -12,  -1,   0,   0,  -1,  -1, -28, -41,  16,   8,   3,   1,   2,   0, -75,  -4, -36,  18,   1,   4,   0,   1, -14,  38, -32, -21,   7,  -4,   3,  -2,  25,   7,  16, -27,   2,  -5,   0,  -1,  13,  10,  16, -12,  -3,  -1,  -1,  -1,   6,  -6,  18,   2,  -3,   0,  -1,   0,   4,  -4,   0,   6,   0,   1,   0,   0 },
      {  13,  30, -16,  -8,  -3,  -3,  -1,  -1,   6, -26,   0,   4,   1,   2,   0,   1,  -7,  22, -34, -14,  -1,  -1,   1,  -1,  83,  13,  34, -26,  -5,  -6,  -2,  -3,   6,   0,  28,  25,  -9,   4,  -4,   2, -22,   2, -18,  22,   2,   1,   2,   0,  -5,   6, -12,  -8,   7,  -1,   1,   0,  -5,   4,   4,  -8,   0,   1,   0,   0 },
      {  -2, -26, -25,   0,   1,   1,   2,   1,  24,   6,  51, -13,  -2,  -4,  -1,  -2,  -3, -35,   8,   4,   2,   2,   0,   1,  13,   6,  31,  -4,  -5,   0,  -2,  -1,  17, -23,   2,  12,  -2,   3,  -1,   1,   1,  10,  31,   5, -10,   0,  -2,   0,  -5, -45,  -2,  49,  -1,   5,  -1,   2,  -3,   8, -48,   5,   9,   0,   2,   0 },
      {   6,  -4, -14,  12,   0,   3,   0,   0,  12,  24, -36,  -9,   3,  -2,   2,  -1,  29, -10,  -4, -10,   1,   1,   0,   0,  -4,  -9, -21, -22,   5,  -2,   2,   0,  81,   0,  31, -33, -12,  -1,  -5,   0,  11,  13,  13,  31, -19,   4,  -4,   1, -20,  19, -11,  11,   5,  -5,   2,  -1,  -6,   1,   9,  -3,   4,   0,   0,   0 },
      { -13, -16,-106,   7,  -1,   4,   0,   2, -12, -25, -14,  -1,   2,   0,   1,   0,   1,  15,  23, -14,   0,  -3,   1,  -2,  -7,   4,   5,  16,   0,   4,   1,   1, -11,   7, -20,  -8,   3,   0,   3,   1,   4,  -6,   6,  -2,  -2,  -2,  -1,   0,   5,  11, -24,  -4,   3,  -1,   1,   0,  -3,  20,  12, -19,  -1,  -2,   0,  -1 },
      {  10,  11,   9,   4,   2,  -2,  -1,  -1,   7, -29, -43,   5,  11,   4,   2,   1,   0, -12,   8, -33,   3,  -3,   1,   1,  30,  -5,  14,  -3,  -5,   1,  -2,   1, -25, -28,  -8, -21,   5,   2,   2,   0,  69,  -5,  15, -24, -18,   2,  -6,   1,  -1,  18,  23,  31, -15,   0,  -4,   0,  -1,  -7,  -7,  23,   4,  -2,   1,   1 },
      {  -2,   5,  -1,   6,  -2,   2,  -1,   1,  10,  41, -18,  38, -10,   3,  -2,   1,  -4, -16, -79,  10,   9,   5,   4,   2,  -1,  -6,  -3,   0,  -1,  -5,   0,   0, -17,  -5, -36,  -2,  10,   2,   1,   0,  16,   8,   4, -10,  -8,  -1,   0,  -1,   2,   1, -35,  12,   5,  -1,   3,   0,  -1,  30,  -7, -35,   5,  -4,   1,  -2 },
      {  -2,  -6,  -8,  66,   0,  13,  -2,   3,  -6, -15, -42,  14,  10,   5,   2,   2,  -9, -38,  -5, -17,  18,   2,   2,   2, -27, -17,   8, -35,  -7,   2,   0,  -1, -19,   7,  15,  26, -21,   2,  -4,   1, -26,   9, -17,  12,   9,  -2,   3,  -1,  11, -29,  11,  -5,   1,  -1,   1,   0,  15,  -7, -15,  11,  -3,   1,  -1,   0 },
      {   4,  -6, -25,  20,  12,   3,   2,   3,   8,   6,  26, -14,  -5,  -3,   0,   1,  27,  28, -24,  34,   7,   1,  -1,   0,  18, -50, -29, -32,  22,   0,   4,   0, -11, -17,  17, -19,  -2,   6,  -3,   1,   1, -25, -10, -25,  -5,   6,   2,   1,  36,  -2,  17,  -6, -23,   2,  -7,   0,  -1,  11,  -6,  26,  -5,  -1,   0,  -1 },
      {   3,  13,   2,  64, -11,   8,  -3,   4,  -5,  -2, -11,  13,  -8,   1,   0,   0,   3,  36,  18,  39, -13,  -5,  -5,   0,  26,  23, -34,  38,  14,  -4,   1,  -2,   1, -22, -10,   5,  13,   1,   0,   0,   2,  11, -13,  11,   1,   3,  -1,   0, -32, -10,  -2,  16,  10,   2,   3,   0,   0, -15, -30,  -1,  12,   2,   2,   0 },
      {   6,   1,  11,  -5,  -7,  -7,   2,  -2,   2,  -3, -44, -16,   6,  -4,   2,  -4,   2, -20,   7,   0,  -1,  -3,  -1,   0,  17,  34,   4,  18,  17,  -1,   2,  -1,  42, -27, -17,  -6,  19,   1,   6,   2, -32, -27,  -3, -35,  10,   5,   1,   4,  40, -28,  -9, -25, -18,   9,  -4,   2,  11,  30, -13,   2, -10,  -4,   0,  -1 },
      {   3,  -2,  16,  68,   1,  10,   3,   3,  10,  12,  53, -12,  19,  -6,   0,  -1,  -4,   7,  -2, -52,   8,   2,   0,   1,  11,  16,   4,   2,  -8,  -3,  -2,  -2,   8,  -6, -15, -15,  10,   2,   0,   2,   8,  -4,  18, -13,  -6,   4,  -2,   2,   0,  -1,  -6, -23,   2,   1,  -1,   0,  13,   5,  43, -14, -19,   1,  -4,   0 },
      {   1,   1,  -8,  19,   6,   7,   0,   2,   3,   9,   8,  19, -18,   1,  -4,   2,   5,   7,  -5,  -2,   1,   8,  -4,   3,  -6,  15,   3,   5, -40,  -1,  -4,   1,  44,  41,  -6,  53,  10, -14,   2,  -6,  29, -31, -13, -18,  26,   4,   1,   4,  21,  23,   7,  16, -17,   1,  -7,   1, -22,   1,  -6,  29,   8,  -1,   4,  -2 },
      {   2,   9,  10,  10,  11,  -9,   1,   0,  -9,  -2, -10, -93,  19,   5,   7,  -3,  -8,  12, -17,  -9, -25,   7,  -3,   2,  -7,  -2, -19,   6, -20,   2,  -2,   1, -13,  13,   1,  17, -14,  -8,   0,  -2,  18,  24, -10,   4,   2,  -8,   3,  -2,  14,   3, -12,   8,   1,   0,   0,   0, -12,  32, -24, -16,  13,  -3,   3,   0 },
      {   3,  -2,  -4, -15,  13,   2,   3,   3,   5,  -2,  -6,  -3, -16,   0,  -2,   3,  13,   9,  10,  16,   8,  -2,  -2,  -1,  11,  -1,  -7, -21,  -3,   5,   2,   1,   9,   8,  -2,  13,  -9, -11,  -1,  -4,  67,   5,   2,   5,  36,  -4,  -1,   0, -10, -50,  -5, -36,  16,  18,  -1,   5,  47,  -3,  -1, -21, -25,   9,  -4,   0 },
      {   1,   7,  14,   1,  27,  -2,   1,  -1,  -5,  -3,  -1,  -8,   3,  -6,  -4,   5,   2,  12,  33,  15,  54,  -5,   0,  -3,  -1,  24,  -4, -60,   8,  25,   1,   3,   1,   2, -29,   6, -18,   7,   4,   1, -12, -10,  -5, -29,  -3,   7,   2,   2, -29,  14,  -7,  16,  -9,  -4,   3,  -2, -20,  13,  -7, -35,  13,  -5,   1,   0 },
      {   1,   1,  15,  21, -40,  -3,  -4,  -3,   0,   2,   2,   1, -37,   2,  -5,  -1,   1,  -3,  23,  34,  -2, -19,   0,  -2,   3,  21,  31, -11,  28,   4,   1,   1,  -7,  11,  -4,  -3, -32,   0,   0,   1,  20,  29,  15,  20,  -6, -16,  -5,  -6,  54,  15,  -1,   3,   4,  -2,   1,  -1, -11,  30,  26,   7,   1,  -3,  -1,  -1 },
      {  -2,   0,   1, -18, -72,  -7,  -6,  -1,   5,   5,  16,  27,  33, -22,   8,  -7,   5,  21,   9, -38,  37,   5,   2,  -3,  -1,   0, -36,   0,  -6,  14,   0,   4,   0,  -1,   7,  -1, -15,  -6,   4,   0,   9,  11, -17,   4,   3,  -8,   2,  -1,  18,  -2,   0,  -9,   1,   6,   1,   1,  -1,  13, -39,  -1,  11,  -1,   4,  -1 },
      {  -1,   3,  -3, -16,  36,   5,   6,   1,   3,  13,   1,  16,   4,  17,  -7,   9,  -3,  12,   4, -10,  -9,  13, -10,   7,   8,  24, -35,  -4, -28,  22,   0,   3,  -5, -30, -36,   3, -42,  -3,   2,   3, -16,   6,   4,  35,  -8, -13,  -4,  -2,  15,   1, -11,  12,  -9,  -4,  -2,  -2,  29,  12,  11,  47, -14,  -8,  -7,  -1 },
      {   3,   7,   3,  -1,  62,   0,   8,  -6,   6,   9,   8,  35,  45, -10,  -1,   0,   2,   5,  23,  -3,  -8,   7,   0,  -1,   0,  12,  13,   1,  26, -17,  -2,  -5,   1,  28,  10, -19,  19,   9,  -6,   0,  15,  32, -32,   8,  -9,   5,   3,   0,  44,  -4, -11,   2,   8,   1,   1,   0,  -7,   3, -21,  -6,   4,   4,   3,   0 },
      {   3,   0,   4,  -7,   1,  23,   4,   3,  -1, -10,  -4,  -7, -39, -14,  -6,  -2,   4,   8,  11,  -5,  19,   9,   2,  -3,   4,  -2, -16,   3, -30,  -8,   5,  -1,   3,  -7,  17,  13,  24,  -6,  -3,  -1, -13,  14,  -4, -11, -35,   4,   0,   0,  44,   8,   2,  29,   4, -26,   1,  -8,  55, -21,   6, -41,  15,   7,  -3,   4 },
      {  -3,   2,   3,  -4,  -9,  18,  -3,   5,  -5,   1,   3, -10, -22, -14,  -6,   1,  -5,   5,   2,  -8,  13,   5,   0,   0,  -7,   4,  14,   5,  17,  -4,   3,   2,  -7,  31,  13,  -5,  20,   7,  -9,   3,   3,  23, -26,  -3, -15,  28,  -1,   4, -42,   0, -23,   4, -44,  -3,   5,   0,  41,  44,  -4,  43,  -9, -19,  -2,  -4 },
      {   0,   2,   6,  -1,  23,   8,   4,   1,  -3,  -4,  -1,  23, -44,   2,  -3,   1,  -1,   1,  20, -26,  -4,   2,   0,  -3,   0,  -7, -25,  -4, -17,  -8,   8,   1, -14, -31,  36,   0,  17, -24,  -2,  -2, -11,  38,  22, -10,  28,  10,  -7,   0,  -2,  -6,  -2, -13,  -9,  19,   2,   2, -48,  38,   1,  -7,  -4,   0,   4,  -3 },
      {   0,   0,   3,   1, -30,   4, -10,   5,   3,   4,   6,  26,  16,  47,  -3,   4,   3,   3,  26,   9, -71,  17,  -1,   6,  -1,  -2,   3, -37, -28, -12,  -1,  -1,   3,   6,   4, -14,  -8,   6,  -2,   0,  -4,  -6, -12, -28, -10,   7,   4,   1, -10,  -4,   8,   2, -10,  -6,   0,  -1,  17,  25,  -4, -28,   4,  -8,  -1,   0 },
      {  -1,  -1,   1,   3, -10,   1,  -5,  -3,   0,   0, -12,   4,  66, -24,   4,  -5,   3,   2,   4,  53,  16,  15,  -8,   2,   5,  -8,  13,   5, -33,   9,   6,   1,  -6, -11,  22,   9,  18, -25,   4,  -5,  -6,  20,  23,  -4,  23,   6,  -6,   4,  -1,   8,  -7,  25,  -8,   6,  -1,   0,  11,  15,  38,   2, -18,  -3,   0,  -1 },
      {  -1,   0,   2, -11, -13,  44,  -6,  11,  -2,  -2,  -9,  -9,  -4,  -6,  -4,   0,   0,   2,   1,  12,  12,  44,  -4,   3,   4,   5,   3,   6, -25,   7,   5,   1,  -1,   2,  -8, -12,  -1,  17,  -4,   5,  13,  -3, -29,   5, -38,   5,  10,   2,   3, -54,  -1,  -6,   5,  -8,   2,   3, -58,  -8,  28,   7, -13,   4,  -2,   3 },
      {   0,   1,  -2,  20,  15, -74,  10, -15,  -1,  -2,   2,  12, -20,   0, -10,  -3,   1,   0,  -1,  10,  14,  39, -14,   2,   1,  -3,  21,  30, -25,  33,  10,  -4,   2,   7,  22, -20, -19,  -9,  17,   3,   1,   0,  -6,  -7, -28, -14,   7,   4, -10,  -6,  12, -23, -12,  -8,   0,   3,   6,   9,  -9,  -9,   6,   0,  -1,   1 },
      {  -3,  -4,  -5,   0, -18, -74,  -4, -16,   0,   0,  -3,  -4,  -2,  16,  -9,  -2,   1,   3,   7, -11,  -2,  11,   2,   5,  -3,   2, -26, -26,  15, -17,  -3,   3,  -3,  -4, -12,  30,  30,  13, -13,   2,   5,  13,  -4,   7,  -5,  32,   1,   0,  11, -23, -24,  23, -12,  -2,  10,   0, -12, -26,  29,   8, -16,  -5,  -3,   3 },
      {   0,   0,  -3,  -6,  10,  10,  -4,   7,   2,   2,   1,  14,  17,  24,  11,   2,  -2,  -1,  -7,  -2,   1, -44, -13,   1,  -1,  -3, -16,  34,  17,  33, -26,  -1,   6,   0,  30,  41, -21,  26,   5,  -3,  11,   6,   8,  -5, -26,   6,   6,  -2,   2, -16,  14,   1, -48,   0,   7,   0,  -6,   3,  20, -31,  -8, -10,   0,   1 },
      {   0,   1,   0,  -2,   6, -14,  31,  -2,  -2,  -4,  -7,  13,   6, -58,   5,  -2,  -1,   1,   6,  -1, -17,   6,  11,   0,  -3,  -3, -17,  -2,  12, -42,  -5,  -3,   0,  -3,  -4,  39, -12,   9, -12,  -4,   4, -20,  11,   3, -30, -13,  10,  -2, -19, -10,  36,   0,  21, -15, -10,   4,   2,  48,  17,   1,   4,   7,  -4,  -3 },
      {   0,  -2,   0,  -3,  12,  -6, -38,   4,   3,   1,   4,  -2,  15,  60,  20,  -4,   3,   3,   4,   7,  39, -11,  40, -12,   2,   1,  -1,  11,   1, -35,  22,   0,   3,  -1,  -1,  10,   7, -39,  -2,   2,   4,  -4,   1,   7, -22,  -7,  -3,   1,  -4, -14,  10, -15,   8, -19,  -1,   0,   3,  21,  11,  14,  30,   3,  -6,  -1 },
      {   0,   0,  -2,   2,  -3,  11,  13,  -3,   0,   0,  -1,   0,  10, -25,  -3,   0,   0,  -2,   0,  13, -16,   9,  10,   0,   0,   4,   7, -11,  -7, -10,  23,   0,  -5,  -5, -20,  15,  -8, -25, -20,   5,   7,   2,  30,  22, -28,  30, -17,  -4,   6,   9,  -6, -52, -52,  11,  17,   1,  -6, -32, -19,  -9,  29, -16,   7,  -3 },
      {   0,   0,   1,  -2,  -4,  19,  -1,   3,  -1,  -1,  -1,  -5,   7,  34, -38,   4,   2,   1,   2,  -6,  14,  73,   3,   4,   1,   0,  -2,  10,  49,  -7,  15,  -2,   1,   1,   5,  28,   1,   7,   2,  -3,  -1,  -2,  24,   8,   2, -17,   0,   1,  -1,  28,  17,   3,  -9,  14,  -6,   0,   3,   3, -14, -17, -33,   9,  10,  -1 },
      {   0,   0,   0,   2,  10,  -9, -62,  -3,   0,  -1,   2,   1, -10, -30,  11,   1,  -1,  -1,  -4, -10, -12,   5,   8,  -1,   1,   0,   1,  -2,   4,  12,  12,   0,   0,   2, -19, -17,  14,  -9,  -3,   3,  -3, -18, -22,  50,  21,   1, -13,  -2,   5,  -8,  48,  25, -33,   6,   4,  -1,   2,  15,   6, -34,  -9,   0,   0,   0 },
      {   0,   1,   0,   0, -13,   6,  79, -17,   0,   1,   1,  -1,   1,  21,   4,  -7,   0,   1,  -3,  -5,   4, -17,   3,   2,   0,   0,   2,  -3,   6,  -7,  12,   0,  -1,   7,  -6, -13,   9, -44,   5,   6,  -4,  -6, -34,  15,  -2, -31, -28,   7,  -6, -16,   9,  26, -30,  11, -16,  -6,  -7,  -4,  11, -16, -14,   1,   3,  -1 },
      {  -1,   1,   0,   2,   0, -14,  13,  11,  -2,   0,  -2,  -5,  -4,  14,   0,  -1,   0,   5,   4,  19,  30, -17,  29,   3,  -3,   5,   4,   9, -49, -37, -30,  12,   0,   1,  -5, -19,   0,  47, -30,  -2,  -5,   1,   3,  19,  -6,   6,  -6,  -2,   4,  18,  17,  -8, -10,  33,   0,  -1,  -1,   8, -23,  -8, -40,  11,   9,   0 },
      {   0,   0,   0,   2,   8,   0, -24, -15,   0,   1,   1,   5,  -8, -23,  22, -16,   1,   3,   0,  -3, -17,   6,  61,  -5,   0,   3,  -2,  14,   7,  -8,  18,  17,   3,   4,  12,   5, -54, -13,   1,   9,  -4,   3, -12, -42, -10,  -4, -13,   1,  -2,   1, -27,  21,  -2,  21,  -4,   3,  -1, -23,   2,  -2, -38,  -4,   7,   0 },
      {   1,   0,   0,  -2,   0,  -1,  32,  17,   2,   0,   2,   0,  -4,  11,  40,  -6,   0,  -2,  -2,  -4,  14,  34,  -7,  -9,   4,   3,   5,  23,  10, -24,   8,  11,   0,  -4,   1, -17, -41,  17, -23,   2,   3,  -3,  -2,  -2,  55,  36,   9, -10,   6,  -4,  10,  17, -10, -25,  19,   4,   5,  -3,  15, -12,  27, -20,  -6,  -1 },
      {   0,  -1,  -1,  -2,   0,  -3,  -2,   1,   0,  -2,   0,  -2,  13,  -7,-100,  10,   0,  -2,  -6,  -4,   4, -26,  19,  -2,   0,  -2,  -5,  17,  -9, -16,  24,  -8,   2,  -1,   7, -10, -26,  12,   2,   4,   2,  -3, -13,  -7,  20,   9,  -9,   2,   0, -15,   6,  11,   0,   4,   6,  -6,   0,   2,  16,  -3,  28, -13,  10,  -2 },
      {   1,   1,   1,   4,   0,   7,   7, -22,   1,  -1,   0,   0,   1,  -7,  -5,   3,   2,   1,   4,   4,  -6,   6,  53, -12,   2,   0,  -1,  -8,  -7,   6,  -9,  11,   1,  -7,   3,  -6,  21,  27,   7, -21,   5,   9,  10,   9,  26, -57,  46,   3,  -3,  -5, -17,  -7, -40, -36,  -2,  16,  -3, -10,   1,   0,  13, -12, -16,  -3 },
      {   0,  -1,  -1,   0,   0,  23,   4,-101,   0,   0,   1,  -1,   0,  12,   0, -20,   1,   2,  -3,   1,   4, -11,  -3,  26,   1,  -1,  -1,   5,  -7,   8,  25,  18,   0,   1,  -1,   3,   2,  15,   1,   6,   3,  -4,   6,   6,   2,  31,   8,   2,   1,   4,  17,  -5,  19, -15,  16,   4,   0,  14, -10,  -1, -12,  -3,  -7,   6 },
      {   1,   2,   2,   0,   0,  -8,  21,  42,   0,   0,   1,   2,   4,   3,   7, -18,  -1,  -1,  -1,   0, -10, -20,  15, -33,  -1,   0,  -3,  -5, -10,  22,  73, -13,   1,   5,  -1,  10,  17,  25,  21,  28,   5,   0,   8,   9,  -2,   9,   5,   6,   3,  10,   8,   0,  14, -25,  14,   7,  -5,   5, -16,   1, -36, -15,  -8,   5 },
      {   1,   1,   0,   2,   2,   0,  22,  15,   0,  -1,   0,   2,   1,   7,  -6, -29,   0,   1,  -3,   0, -12,  17,  54,   4,   1,   0,   1,  -2,  10,  65, -24,  35,   2,   1,   3,   5,  24,   4, -25,   2,   0,  -4,  -1,   3,   0,  14, -28,  -7,   2,  -5,  12,  -2,  26,  14,  11, -12,   8,  14,  12,   8,  31,   4,  14,   5 },
      {   1,   0,   0,   0,   7,   4,  -3,   0,   0,  -3,   0,   0,  -6,  -1,   7, -59,   2,   2,   4,   2,   6,  12,  -2,  11,   0,  -4,  -3,  -9,  -4, -25, -21, -28,   1,   0,  -2,   2,   4,  15,  58,  -6,   0,  -2,  -1,   6,  -2, -17, -13,  -2,   0,  -3,   4,  -2,   5,  26,  31, -12,   5,  -3,   7,   3,  11, -61,  18,  11 },
      {  -1,  -1,   0,  -2,  -3,  -2,   2,  -8,   1,   2,   0,  -1,  -1,  -1,  35,  54,   1,   1,   3,   0,   3,   8,  15,  46,  -1,   0,  -3,   1,  -3,   3,  34, -27,   1,   3,   8,   8,   8,  50,  12,  15,   2,   6,   4,  -4,   8, -17, -44,  18,   0,  -4,  -7,  -3,  -1,  17, -14, -12,   4,   1,  12,   3,  31,   7,   7,  -5 },
      {   0,   0,   1,  -1,   6,   0, -16,   2,  -1,   0,  -1,   3,   2,   0,  -3, -67,  -1,  -1,   0,  -2,  -2,  -3, -40, -11,  -1,   1,  -3,  -5,  -3, -15,  25,  42,  -1,   1,   1,   5,  -1,  33, -11,  20,   1,   3,   4,  -4,   5, -30, -22,   4,  -6,  -3, -14,  -4, -23,   2, -27,  -5,   3,  -2,   7,   2,  24,  43,   1,  -5 },
      {   0,   0,   0,   0,  -3,   6,   6, -35,   0,   1,   0,  -2,  -1,   6,   8,  14,  -1,  -1,  -2,   2,  -3,  13,   4, -94,  -2,  -1,  -4,  -2,  -4,   4, -15, -26,  -1,   0,  -5,   2,  -2,  14,  20,  18,  -1,   1,  -1,   7,  -2,  20,   7,  23,   1,   1,   2,   4,   0,  37, -13,   7,   9,   3,  12,   4,  20,  20,  15,  -3 },
      {   0,   0,   0,   0,   2,   0, -10, -22,   1,   0,  -1,   0,   1,   1,   1,  13,   1,  -1,   1,   0,   3,   1, -22, -33,   0,  -1,  -2,   2,  -4,   6,   2, -18,   4,   1,   7,   3,  11,   3, -70,  28,   2,  -8,  -1, -17,  -1, -50, -21, -31,  -3,  -2,  -3,   0,  11,  -2,  30, -15,  -1,  -1,  -1,   2,  -9, -45,   8,  -6 },
      {   0,   0,   0,   2,   2,  -2,  -2,   4,   0,   0,   0,  -1,   0,  -2,   1,  39,   0,   0,  -1,   1,   2,   1, -13, -10,   0,  -1,   1,  -1,   0, -11,  -1,  87,   1,   1,   4,   3,  10,   0,  28,  27,   1,   1,   1,   7,  -8,  -1,   3,   6,   4,  -1,   9,  -4,  15,  11, -26,   3,   1,  -5,   0,  -6,   2, -57,   6, -22 },
      {   0,   0,   0,  -1,   0,   2,  -5,  -1,  -1,   0,  -2,   0,   1,  -4, -11, -10,   0,   2,   1,   5,   1,   7,  14,  10,  -1,  -1,  -1,   1,   2,  -4, -51, -21,  -1,  -1,   1,   0,  -4,  -9,   9,  83,   4,   4,   9,   2,  16,  10, -25,  25,   0,   1,   0,  -8,  -5, -48, -20, -12,  -7,  -2, -12,  -5, -11,  -1, -23,  -9 },
      {   0,   0,   0,   0,   0,   1,   5,   7,   0,   0,  -1,   0,   1,  -1,   2,  10,   1,   0,   1,   0,   0,  -3,   3,  25,   0,   0,  -1,  -1,   1,  -5,   4,   8,   2,   1,   1,  -1,   4, -12,   8,  70,   0,  -3,  -1,  -1,  -1, -13,  61, -28,  -1,   4,  -1,  10, -10,  42,  39,  -4,   6,   3,  10,   3,  15,  27,  -5,  30 },
      {   0,   0,  -1,   1,  -1,   1,   3,  -4,   1,   0,   0,  -1,   5,  -5,  -2,  30,   1,   0,  -1,   2,  -3,   5,  -1, -26,   1,  -2,   0,  -1,  -2, -15, -13,  31,   2,   1,   2,   2,  -1,   5,  46, -11,   0,  -3,  -1,  -5,   3,  -3, -50, -44,  -2,  -4,  -4,  -5, -10, -32,  53, -20,  -4,   5,  -7,   6, -11,  37,  -4,  25 },
      {   0,   0,  -1,  -1,   1,   0,  -3,   7,   0,   1,   1,   0,   0,  -2,   0,   9,   0,   0,  -1,   0,  -2,  -2, -11,  -1,   0,   0,  -1,   0,   5,  -6,  -5,  21,  -1,  -1,  -1,   1,  -4,  -7, -19,  -5,  -1,   0,   1,  -4,   0, -18,  -7, 100,   3,   1,   2,   0,   2,  -3,  51,  10,   0,   2,   0,   2,  -1,   1,   5,  45 },
      {   0,   0,   0,   0,   1,  -2,  -1,  -1,   1,   0,   1,   1,  -1,  -1,   8,   2,   0,  -1,   1,   0,  -2,  -5,   1,   4,   1,   1,   1,  -1,   0,   0,   3,   1,  -1,  -3,  -3,  -3,  -2,  -5,   2,   3,   4,   1,   5,   4,   6,   2,  30,  10,  -2,  -6,  -2,  -4,  -5, -32,  -1, -62,   1,   2,   1,   3, -12,  11,  98, -16 },
      {   0,   0,   1,   0,   0,   1,   2,   3,   0,   0,   0,  -2,  -1,   2,  -1,  -3,   0,   1,   1,  -1,   0,  -1,  -2,  11,   0,  -1,  -1,  -3,  -1,   0,  -6,  -1,   0,   0,   0,  -1,   2,  -2,   9,  10,  -2,  -2,  -3,  -3,  -3,  -8,  -7,   7,   2,   1,   2,   5,   2,  -3,  52,  62,   4,  -2,   4,   2,   0,  22,  23, -91 },
      {   0,   0,   0,   0,   1,   1,  -2,   0,   0,   0,   0,   0,   0,   1,  -1,   3,   0,   0,   1,   0,   2,  -2,  -2,   6,  -1,   1,  -2,  -2,  -1,   1,  -6,  -4,  -1,  -1,  -1,  -1,   2,   0,  -2,  15,  -2,   1,   0,   1,   3,   0, -15, -25,  -1,   0,  -2,  -1,  -1, -21, -27,  83,   3,   1,   4,   0,   0,  -7,  61,  58 },
    },
    {
      { 105, -51,  -4,  -2,  -1,  -1,   0,  -1,  30,  -6,  -9,   2,  -1,   0,  -1,   0, -17,  23,  -8,   1,  -1,   0,   0,   0, -18,  14,   1,  -2,   1,   0,   0,   0, -10,   4,   4,  -1,   0,   0,   0,   0,  -5,   1,   2,   0,   0,   0,   0,   0,  -3,   1,   1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,   0,   0 },
      { -43, -16,  23,  -1,   3,   0,   1,   0,  88, -56,   1,   1,  -2,   1,  -1,   0,  30,   6, -21,   7,  -4,   2,  -2,   0, -17,  26, -10,   0,   0,   0,   0,   0, -11,  11,   1,  -2,   1,  -1,   1,   0,  -7,   3,   2,  -1,   1,   0,   0,   0,  -2,   1,   1,  -1,   0,   0,   0,   0,  -2,   1,   1,   0,   0,   0,   0,   0 },
      {   2,  64, -55,  10,  -8,   2,  -2,   1,  54,  41, -43,  -4,  -3,  -2,  -1,  -1,  -2,  21,   7, -16,   5,  -3,   1,  -1, -22,   3,  23,  -9,   2,  -1,   0,   0, -10,  -9,  15,  -1,   0,   1,   0,   0,  -3,  -7,   7,   1,   0,   1,   0,   0,  -2,  -4,   3,   1,   0,   0,   0,   0,  -1,  -2,   2,   0,   0,   0,   0,   0 },
      { -37, -26,  24,   1,   4,   2,   1,   0,   7,  -1,  -6,   2,  -1,   1,   0,   0, -82,  54,  -2,  -2,   2,   0,   2,  -1, -27, -18,  24,  -5,   4,  -2,   2,   0,  23, -33,   4,   4,  -1,   1,   0,   0,  18,  -8,  -9,   4,  -1,   1,  -1,   0,   9,   1,  -6,   0,   0,   0,   0,   0,   2,   1,  -2,   0,   0,   0,   0,   0 },
      {  19,  19,   9, -22,   4,  -4,   1,  -2,  50, -10,  32, -28,   4,  -4,   0,  -2, -24, -41,  53,  -4,  -2,   2,   0,   1,  13, -54,  10,  21,  -9,   5,  -3,   2,  12,  -5, -25,  20,  -5,   3,  -1,   1,  -4,  12, -17,   5,   2,  -1,   0,   0,  -5,   6,  -4,  -1,   2,   0,   1,   0,  -3,   3,  -2,   0,   0,   0,   0,   0 },
      {  10,  29,  -4, -15,   2,  -3,   0,  -1, -22, -23,  48, -19,   7,  -2,   2,  -1,   0, -17,  17,   7,  -5,   4,  -3,   1, -78,  37,   0,  10,   0,   2,   1,   0, -21, -26,  18,   0,   5,  -1,   2,   0,  19, -30,  -1,   5,  -2,   2,   0,   1,  12,  -6,  -9,   5,  -2,   0,   0,   0,   4,   2,  -6,   2,   0,  -1,   0,   0 },
      {   4,  14,  57, -63,  20, -11,   5,  -4,   8,  56,  14, -38,   1,  -3,  -1,  -1,  21,  38, -18,   4, -12,   3,  -4,   1,  17,   8,   4,   3,  -5,  -1,  -1,   0, -13,  -1,  14,   3,  -1,  -1,   1,  -1,  -9,  -9,   6,   4,   0,   0,   0,   0,  -1,  -8,   2,   3,   0,   0,   0,   0,   1,  -7,   1,   3,   0,   0,   0,   0 },
      {   2, -33,   8,   8,   1,   0,   1,   1,  -6,  -3, -10,   3,   4,  -1,   1,   0,  69,   0,  13, -30,   8,  -5,   1,  -2,  -7, -24,  57, -19,  -1,   3,  -2,   1,   7, -48,  20,  18, -14,   8,  -3,   3,   7,  -7, -22,  20,  -7,   2,  -1,   0,   0,  11, -15,   5,   1,  -2,   1,   0,  -4,   7,  -4,   0,   1,  -1,   0,   0 },
      {  18,  34,  -2,  -9,  -5,   0,  -2,   0,  -5,  -5,   2,  11,  -7,   3,  -2,   1,  29,  22, -45,  35, -11,   3,  -2,  -1, -29, -16, -13,  18,   8,  -3,   4,   0,  56, -30, -32,   1,   5,  -1,  -1,   1,   7,  40, -27,  -9,   2,  -1,   0,  -1,  -8,  19,   4, -12,   3,  -2,   1,  -1,  -3,   6,   5,  -5,   1,   0,   0,   0 },
      {  17,  32, -22, -16,  10,  -4,   1,  -1,  -5, -56,  20, -24,  21,  -3,   4,   0, -13,  18, -18, -21,   9,  -1,   1,   0,  50,  20,   9, -34,  -1,   0,  -1,   0,  45,  -7,  32, -13, -13,   4,  -4,   2,   4,   1,  12,  12, -14,   4,  -2,   1,  -5,   1,   0,  12,  -4,  -1,   1,  -1,  -4,   1,  -2,   3,   1,  -1,   1,   0 },
      { -14, -17, -24, -10,  15,   0,   6,  -1, -23, -12,  -9, -40,  25,  -3,   7,  -1,  13,  33,  17, -46,   0,   1,  -2,   1, -38,   6,   9,  12, -22,   8,  -4,   2,  14,  39, -36,  15, -11,  -3,   2,  -2, -23,  33,  10,  -9,   6,  -7,   3,  -2,  -2, -13,  19,  -6,   3,  -1,  -1,   1,   1,  -7,   8,   0,  -1,   0,  -2,   0 },
      {  -7, -22, -17, -34,  48, -18,  12,  -4,   5,   9, -53, -20,  33,  -8,   6,  -1,  -8, -50, -23,   0,   5,   8,  -1,   5, -11,  -2, -22,  -6,  -3,   6,  -3,   2,   6, -10,   2, -15,  -2,   2,  -1,   1,  43,  -5,  -6,  -9,  -6,   1,  -4,   0,  11,  21,   0,  -6,  -5,   1,  -2,   0,   0,   7,   6,  -1,  -4,   0,   0,   0 },
      {   7,  21,  35, -21,   6,  -8,   2,  -4, -11, -16, -15,  25,   1,  -1,   1,   0, -11, -31, -21,   2,  23,  -6,   6,   0, -40, -38,  29, -38,  22,  -1,   1,   3,  -9,  18,  12, -21,  -7,   8,  -6,   2, -43,  24,  25,  -3, -10,   4,  -1,  -1,  -5, -18,  13,   7,  -4,   1,   1,  -1,   7, -10,   0,   5,  -1,   0,   0,   1 },
      {   3,  25,  -7,  -9,  14,  -8,   0,  -1, -18, -35, -19,  10,  15,  -2,   3,   1, -19,  12, -30,   4,  -2,   4,  -1,   1,  21,   0,  15,   9, -14,   7,  -4,   1, -73, -10, -12,  39,  -8,   4,   3,   1, -17,   8, -44,  11,  10,  -6,   4,  -2,  -1,  12, -11, -15,  10,  -2,   1,  -1,   4,   1,   3, -10,   2,   0,  -1,   0 },
      { -12, -18,  -8,  18,  -4,   2,  -3,   2,   9,  35,  39, -18,   6, -11,   1,  -3, -17, -22,  -8,  -3,  19,  -6,   4,  -1,  -1,  36,  -1, -27,  10,  -1,  -2,   0, -13, -25,  13,  -1,  -8,  11,  -3,   2,  -8,  53, -30,  -4,  -9,   2,   2,  -2, -39,  16,  31, -14,   3,  -4,   4,  -2,   7, -21,  11,   0,  -1,   2,  -1,   1 },
      {  16,  28,  43,  20, -18,  -7,  -5,   0,   4,  -2,   8,   4,  16, -16,   3,  -3,   8,  16,   4, -44,  42, -20,   6,  -5,   3,   0, -12, -19,  17,  -2,  -3,   3, -17,  14, -32, -12,   2,   3,  -2,   0,  51,  -2, -11, -24,   0,   1,  -2,  -2,  29,  15,   5, -19,  -5,   3,  -3,   0,  -3,   8,  14,  -5,  -5,   2,  -2,   0 },
      {  -3, -14, -31,  23,  23, -17,   8,  -4,   5,  -4,  31, -13,  36, -25,   8,  -6,  15,  26, -11,  15,  12,  -2,  -1,   0,  11, -46, -11,  14,  21,  -3,   2,   2, -24, -32,  -1, -27,  26,   4,  -1,   2, -22, -15,  13, -28,   3,  10,  -4,   2,  11, -26,   7,  -7,  -9,   7,  -3,   2,  16,  -6,  -5,   2,  -7,   3,  -1,   0 },
      {   4,  13,  24,  34,  24, -43,  12,  -8,   5,  18,  -1,  37,  37, -38,   9,  -7,  -4,  -1,   8,  -1,  16,   4,  -6,   2,  -8,  24, -15,  13, -15,  19,  -8,   3,  32,   4,  -4,  12, -11,   6,  -1,   0, -31, -20,  -2,  21,   1,   4,   1,   2, -29,  -1, -25,   3,   7,   2,   2,   0, -12,   6,  -5, -10,   5,  -1,   1,  -1 },
      {   2,   6,   3,  -7,  38, -32,  12,  -9,  -3, -12,  -4,  42,  -4,  -4,  -2,  -1,   6,  14,  50,  22, -29,   8,  -6,  -2,  15,  23,  43,   0,   4,  -8,   2,  -4,  -7,   0,  -4, -13,  19,  -4,  -1,   3,  34,  23,   7, -23,   3,   2,  -3,   1, -13,   1,  34,  -4,  -5,   5,   0,   1,  -6, -15,   7,   9,  -2,   3,   0,   0 },
      {  -9,  -5, -18, -12,  12,  -3,   5,  -2,  -7,   1,  12,  -6,  -6,  -3,   2,   1,   3,  26,  26,  14,  -5,  -9,   3,  -2, -11, -15, -10, -13,  27, -10,   5,  -1, -14,  35,  15, -30,   5,   8,  -8,   5, -21,   3, -14,  13, -16,  15,  -5,   1,  -6,  64, -34,  -7,  -3,  -5,   5,  -4, -34,  26,  23, -15,   6,  -5,   3,  -1 },
      {  -5, -10, -28, -20,  24,  -4,   3,  -1,  11,  18,  21,  17,  -8,   1,  -4,  -1,   2,  -1,  -1,  18,  16, -11,   3,  -4,   4,  23,  15, -27,  30,  -8,  -3,  -1,  13,  27, -34,  18,  -4,  14,  -9,   6,   3,   7, -31,  15,  -7,   8,   2,  -1,  43, -37, -26,   9,   0,   1,   3,   0,  28,  13, -30,  -5,   1,  -1,  -1,   2 },
      {   1,   2,  -8,  35,  13, -12,   0,   1,   7,  16,  37,   7,   4,   4,  -9,   2,   1,   7, -31,   8, -20,  32, -11,   4,  -9, -32,   8,  -8, -34,  23,  -3,   4, -19,  29,  12,   0, -36,   4,   2,  -2,  40,  25,  15,  11, -20,  -6,   3,  -5,  18,   2,   2,  25,  -6,  -5,   3,  -2, -18,   9,  -7,   9,   5,  -5,   4,  -1 },
      {   3,   0,   6,  24,  -4, -17,  15,  -8,   4,   8,  13, -14,  -4,  -6,  10,  -5,  -7, -23, -18,   0, -24,  12,   3,   2,   2,  11,  38, -11, -15, -14,   4,  -2,  22,  13,  24,   3,  14, -24,   3,  -3, -30,  -3, -12, -22,  30, -11,  -4,   1,  58,  10,  -1, -39,   7,   5,  -9,   1,  -4,  15,  24, -13,  -5,   5,  -3,   0 },
      {   0,   0,  -1,  -2,  35,  32, -26,   8,   4,  11,  16,  14,  20,  39, -24,   9,  -5, -10, -10,  -9,  15,  10,   4,  -1,   4,  16,  14,  -5,   2,  -2,   8,  -5,   0, -19, -30,   8,   1,   2,   7,   0, -24, -22,   9, -31,  -2,   0,   3,   2,  -1,   0,  10,   2, -17,   5,  -1,   1, -67,  23,   0,   4,  -3,  -2,   5,  -2 },
      {   8,  12,  28,  29,  13,  19, -25,   4,  -2,  -3, -27, -14,  28,  24, -23,   4,   2,   6,  31,   8, -17,  28, -16,   0,   4,  12, -13,   4,   0,  11,   4,  -4,   6,  -6,   9, -24,  -2,  17,   6,   1, -25,  13, -26,   5, -33,  15,  12,   0,  36,   3,   2,  -4, -21, -11,   9,  -2,  39,  -7,   1,   9,  -5, -11,   3,   1 },
      {   0,  -2,   4,   5,  -7,  27, -26,  11,  -1,  -1,  -9, -13,  44, -18,  -6,  -1,   3,  -1,  -3,  52,  34, -32,   6, -10,   0,   1,  34,  41,   5, -10,  -3,  -3,  10,  35,  33,  22,   1,   4,  -5,  -1,  20,   4,   1,   9,   5,   7,   1,   0,   7,   3,  14,  -1,   3,   2,   1,   0, -16, -11,   2,   8,   0,   2,   0,   2 },
      {   0,   4,  16,  20, -14,  -3,  15,  -5,  -3,   3, -15, -28,   9,  -5,  15,  -3,  -9, -12, -12, -12, -15,  -7,   5,   2,   6,  21,  25,  27,   8, -22,   4,  -1, -14, -10, -13, -13,  29,  -3,  -3,   6,   0,  44,  17,  -8, -12,  12, -13,   4,  -7,   4, -33,  38,  -9,   8,  -2,   2,  -6,  42, -42,   3,   5,  -6,   4,  -2 },
      {  -7, -10, -13, -37, -40,  -7,   7,   0,   1,   1,  -2,  24,  -8, -32,   7,  -5,  -3,   3,  12,  11,  17,  -9, -11,   7,   9,  17,  -2,  17, -22,  30, -21,   6,  -4, -30,  -1,  -4, -21,  22,  -1,  -2, -22,  17,  21, -17, -22,   3,   2,  -4,  48,  12,  16,   6, -14,  -5,   1,  -2, -12,  23,  11,   8,  -2,  -5,   3,   0 },
      {  -2,  -2,  -6, -24,   7,  30, -18,   8,   6,   8,  20,  35,  26,   3, -13,   0,   2,   3,   4,  -7,  16,   6,   7,  -4,   1,   0,  13,   5, -29,   7,  13,  -3,  -1,   1,  10, -17,   0, -25,  11,  -3,   6,  14,  25, -17,  11, -16,  -1,   3, -19,  19, -28, -12,  12,  -2,  -2,   1,  62,  33,   1, -29,   2,   0,  -4,  -2 },
      {   5,   7,   6,  12,  43,  16,  12, -12,   3,   2,   1,  15,  -3,  10,  21,  -8,   2,   6,  15,  -5,   1, -36,  34,  -4,  -2,  -1, -37,   6,  -2, -33,  20,   2,  -5, -22,  13,  21, -25, -22,  -4,   4, -11,  21,  19,  20,   1, -22, -12,   2,  35,  18,  18,  14,  11,  -8, -11,   0,   3,  -1,  -5,  21,  10,   3,  -5,   1 },
      {   4,   5,   3,  20,  23,  18,  11,  -6,   1,   0,   6, -10, -24,   5,  26,  -7,  -3,  -7, -20,   6, -19, -28,  20,  -4,   1,   5,  28,  15,  -9,   4, -13,   6,   4,   9, -25,   7, -16,  38, -18,   8,   0, -40,  19, -13, -26,  18,   0,  -1, -19,  22,  16,  -7, -19, -12,   8,  -3,  40,  11,  24,  -6,  -5,  -9,  -1,   1 },
      {   2,   2,  11,  17,  10,   0,   0,  -2,  -5,  -7, -17, -32, -10,   5,   4,  -1,   3,   4,  22,  34,  -2, -12,   6, -10,   4,   3,  -6, -15,  11,  15,  -2,  -4,  -7, -17,   2, -12, -38,  17,   7,  -2,  10,  15,  21,  -7,  10, -17,  16,  -8,  -8, -44,  -5, -33,  31,  -8,   7,   1, -18,  39,  -1, -43,  11,  -3,  -3,   1 },
      {   0,   0,   2, -10, -27,  45,  34, -24,   4,   5,   1,  11,  40,   8,  50, -32,   3,   2,   5,  19, -10,  20,  19,  -5,   2,   1,  10, -18, -17,   5,  10,   4,  -3,   2, -21, -30,   1,  -5,   2,  12,   4, -11, -20,   5, -13,  -9,   1,   9,   0, -14,  -4,   6,  10, -16,   0,   4, -20, -26,   2,  -7,  11,  -3,  -4,   2 },
      {   1,   1,  -2,  -5,  15,  12,  25, -23,   4,   2,  10,  24,  -9,  33,   0, -14,  -2,  -5, -11, -41,  15,   8, -11,   1,   2,  -1,  -5,  52,  17,  -4, -19,   2,   4,   1,  35,   2,  24,  26,  -8,   2,  11,  23,   3,   8,  -9,  11,  16,  -4,  10, -26, -12, -13,  11, -11,  13,   0, -12, -11,  27, -28,   5,  -6,   1,   6 },
      {   2,   6,   4,   7,  14,  17,  24, -30,  -3,  -3,  -3,  -9, -23,  27,  18, -17,   4,  12,  10,  14,  38,   5, -14,   9,  -2,  -2,  -6,  -6,  -3,  39, -22,  11,   4,  12,  27,   9,   0,   2,   0,  -1,  -2,   2, -18, -39,  33,  -5,   4,   5, -10,  21,  -4,  -2, -15,  14,  -4,   5,   6,   7, -46,  28,  -6,  -1,   5,  -3 },
      {  -2,  -3,  -2,   6, -20, -20,  29, -27,  -1,   0,  -2, -17,  11,  26, -17,  -8,   2,   1,   1,  20,  37,  41, -26,   1,   5,   6,  12,   1,  39,  -8,  17,   0,  -2,  -4, -22,  28,  -4, -15,   9,   4,  -1, -12,  35,  24,  -8, -15,  -3,   5,   1,  28,  23,   7,   4, -12,  -1,   1,  19,  11,  15,  -9,  10,  -3,   0,   0 },
      {   3,   3,   6,   9,  16,  44,  -7,  -3,   1,   3,   1,  -1,  11, -14,   4,   6,   0,  -1,   3,   8, -19, -10, -42,  36,   0,   2,  -1, -28,  11,  12, -52,  20,  -3, -13, -18,   6,  25,  -6, -22,  -7,   1,  10,  27,  28,  27,  -2, -10, -10,   9,   9,  -6,  13,  13,  11,   1,  -6,   1,  -4,  21,  -8,  -7,  15,   0,  -1 },
      {  -1,   0,   1,  -9, -21,  10,  13,  -6,   0,   3,  -1,  16,  28,   4,   7,  -7,   0,   5,   8,  -3, -23,  13,  -1,   3,  -6, -10, -24, -15,   2, -24,   0,   7,  -2,   7,  22,  34,  -5,   1, -19,  12,   6,   2, -10,  11,   0,  26, -12,   5, -19, -27,  30, -22, -29,  24,  -5,   0,   5,  61,   7,   6, -31,   1,   4,  -2 },
      {   2,   2,   1,   8,   8,   0,  19,  23,  -1,  -1,  -3,  -9, -17,  18,  12,  21,   2,   3,   5,  17,  32,   8,  10,   8,   4,  10,   8,  17,  -7,  12,  11,   8,  -4,  -7,  -7, -16, -27, -24,  -5,   7,   3,  15,  -2,  11,  -3,   4, -35,   8,  -3, -23, -35,   2, -32,  38, -17,  -2,  -1, -17,  47,   0, -34,  11,   1,  -7 },
      {   2,   2,   3,   5,  25,  24,  -9, -10,  -3,  -5,  -5,  -7, -34, -46,   8, -19,   0,  -2,  -3,  -3,  14,   4,  -8, -17,   1,  -3,   1,  -2,   2,  21,  15, -14,   0,   4,   4,   1,  30, -27,  37, -10,   1,   1, -25,  16,  -7, -12,  15,   7,  -6, -20,  21,  33, -23, -13,   0,   6,  -1,  35,  34,  -2,   1,  -7,  -2,   3 },
      {  -4,  -5,  -5,  -3,  -6, -28, -63, -38,   0,   1,   0,  -4,  -2,  20,  -3, -54,  -1,  -1,  -3,  -2,  -6,   6,  31, -32,   1,   4,   2,   4,   1,  25,  -7,  11,  -2,  -5,  -9, -14,   9,  11, -18,  17,   1,   7,  10,  28,  26,  -4, -11,   3,   4,   2,  -5,   3,   0,  13,  -4,   4,   0,   4,  14,  15, -15,   7,   0,   1 },
      {  -1,   0,  -2,  -3,  -5, -11,   8,  31,   3,   5,   8,   9,  17,  22,  23,  17,  -3,  -2,  -5, -11, -10,  -8,  -4,   9,   2,   3,  13,  25,  35,   9, -20,   0,   0,   7,  -1, -37, -12,   8,  11, -13,  -3,  -9, -18,  22,  12, -50,  16,  -5,  -8,  -2,  12,  -1,  15, -12,  -9,   8,   7,  32,   9,  50, -15,   2, -13,   3 },
      {   2,   2,   4,   0,   4,   2,   1,  60,   1,   1,   0,   2,   0,  -1,  44,  15,   2,   3,   6,  -1,   2,  46,  40, -21,   1,   1,  -7,  -9,  21,  37,   8, -10,   2,  -2,   7,  31,  33,  21,  -8,  -1,   2,  10,  22,   5,  16,  12,  -1,  -3,  11,  13,   7,   2,   2,  -2,   5,  -3,   1,  -2,  -5,  -9,  10,  -3,   4,   2 },
      {   0,  -1,  -2,  -2,  -6,   9,   9, -15,   1,   2,   3,   6,  15,  -8,   2, -15,  -2,  -4,  -3,  -3, -33, -15,   0, -11,   1,   0,  -2,  -8,  32,  11,  11, -26,   1,  -1,  -1,  28, -17,  15,  55, -27,   4,  11,  27,  -7,  -3,   0,  30,   2,   0,  13, -35, -10, -22,  24,  -3,  21,  -1, -25,   5,   7, -36,   8,  -3,   7 },
      {  -1,  -1,  -1,   2, -13, -36,  10,   9,   2,   2,   0,  -3,  23,  49,   2,   7,   2,   3,   7,  20,   5, -38,  12,  -5,  -3,  -5, -10, -26, -31,  -8, -22,  -2,   1,  -4,   4,  14,  28,   0,   7, -17,  -1,   5,  -7, -14,  15,  -1,  27,  -8,  -1,  -4,  -5,  48, -20, -10,   8,   5,   8,   4,  34, -17,   4, -14,  -1,   3 },
      {   1,   2,   0,   3,   2,  16,  42,  -9,  -1,  -1,   0,  -5,  -3, -14, -39,  17,   0,  -1,  -2,   5,  15,   2,   8, -24,   0,   2,   1, -11, -33,  -4,  25, -23,  -3,  -7, -14, -14,  16,  27, -19,  -5,   3,   5,  13,  23,  36,  22,  17,  -7,   0,  -1, -14, -12,  18, -13,  18,   0,  -1,   9,  13,  52, -17,  -7,   2,   7 },
      {   0,   0,   1,   0,  -4,   8,  -3,  -4,  -1,  -2,  -5,  -1,   7, -23, -18,  -2,   0,   0,   4,  -1, -17,  19,   6,  -3,   1,   2,  -7,  -3,  30,  -2,   1,  13,   3,   2,  13,  28, -19,   6, -28,  25,   1,   1,  -5, -58,   7, -22,  -6,   7,  -8, -14, -37,  31,  12, -29,  -9,  -6,   5,  -2,  41,  30,  11, -14,  -4, -11 },
      {   1,   2,   2,   2,  -3,  10,  37,   6,  -1,   0,   1,   0,   9,   4, -47,   4,  -2,  -1,  -3,  -7, -18, -22,   3, -29,  -1,  -1,  -3,   4,   5,  39,   0, -13,   1,   4,   5,  -8,   3,  23,  -3,  14,  -5,  -4, -15,   0,   3,  -5, -27,  30,   6,   6,  25,  27,  15,  20, -44,  15,   2,   3,  -4, -30,   9,  31, -12,  -2 },
      {  -1,  -2,  -2,  -1,  -6,  -7, -19, -37,   2,   3,   4,   2,   7,   7,  13,  43,   1,   0,  -1,  -2,  -6, -21,   9,  63,   1,   0,   3,  -1,   3,  33,  49,   9,   2,   5,   4,   5,  14,  24,  26,  15,   6,   5,   0,   8,  17,  17, -16,  20,   3,  -7,   6,  -2,   7, -14, -10,   3,   1,   3,   8,   2,   8, -23,  -4,  -6 },
      {   0,  -1,  -1,  -3,   0,  -2, -10,   7,   2,   3,   3,   5,   6,  16,  37,  -5,  -3,  -5,  -5,  -7,  -5, -16, -61, -37,   0,  -2,  -1,   2,   5,   3,  10, -47,  -2,  -3,  -2,  -6, -26,  -7,   8,  12,  -1,  -1,  -3,  -9,  26,  36, -27,   3,   3,  -1,   2,  11,  18,  13,  12, -19,   6,   0,  14,  11,  25, -19,  16,  -5 },
      {   0,  -1,  -1,  -2,  -7,   6,  22,  -9,   1,   1,   2,   3,  13,  -6, -27,   2,  -1,  -3,  -4,  -8, -20, -18,  25,  -4,   3,   2,   4,  10,  26,  33,   3,  -1,   0,  -1,   0,  -9, -11, -29, -15, -10,   0,  -2,  -5,  13,  23, -19,  -7, -33,   1,  -1,   9,  -7, -43,   1,  27, -37,  -1,  -8, -11, -15,  14, -40,  35, -19 },
      {   0,   0,   0,  -1,  -2,   6,  -7, -28,  -1,  -1,  -2,   1,   0,  -7,   4,   7,   2,   3,   3,  -2,   5,  30,  31,  22,   1,   1,   2,  11,   8, -17,   4, -15,   0,   2,   2, -12, -26,  19,  -1, -58,  -4,  -8, -19, -14,  15,  20,  -3, -35,   2,   4,  15,  43,  28,  34,   4,  -8,   9,   8,  12, -12,   3,  23,  11,   3 },
      {   0,   0,   1,  -1,  -1,   3, -14, -15,   0,   1,   0,   2,  -2, -13,  19,  27,   0,   0,   1,  -3,   5,  13, -21,  -7,  -1,   0,  -3,   6,  -5, -25,  11, -33,   0,   0,   2,  -9, -13,  29, -30,  16,  -1,   1,  -1,   4,  36, -39,   2,   6,   3,   8,   5,  20, -41, -25,   3,  11,  -5, -11, -17, -41, -41,   0, -12,   3 },
      {   1,   1,   1,  -1,   4,   5,  10,  41,  -1,  -1,  -1,   1,  -4,  -9, -12, -40,   0,  -1,  -1,  -4,  -1,   6,  -9,  13,   1,   3,   3,   5,   2, -26,  20,  59,   1,  -2,  -2, -17, -24,  26,  29,  19,   0,   0,  -2,   7,  51,  17,  22,  13,   4,   8,   9,  18, -11,  10,   5,   9,   2,  -3,   1, -13,   9, -16,  10,   1 },
      {  -1,  -1,  -1,  -1,  -1,   5,  -6, -17,   0,   1,   1,   1,   0,  -7,  -1,  25,   0,  -1,  -1,  -1,   0,  17,  12, -12,   0,   0,   2,   2,   4, -13, -37,  -5,  -3,  -4,  -5,  -9, -17,  -7,  -4, -23,   2,   5,   4,  12,  21,  21,   2,  17,  -2,  -4,  -5, -19, -39,   1, -34,  44,   2,  -3,   4,   8,  65, -23, -29,  10 },
      {   0,   1,   1,   0,   2,  -4,   2,  28,  -1,  -1,  -1,   0,   0,   8,  -6, -58,   0,   0,   1,   2,   1, -10, -20,  31,   0,   0,   0,  -4,  -5,   5,  34, -18,  -1,   1,   1,   6,   6,  19, -27, -59,   2,   2,   1,  -1,   0,  -7, -41, -14,  -5,  -8,  -9, -12, -18, -30, -23,   5,  -3,   0,  -1,   3,   9,  -3, -17,   7 },
      {  -1,  -1,   0,   0,  -2,  -3,   2,   4,   0,   0,   0,   1,   3,   4,  -1,   0,  -2,  -2,  -3,  -5,  -9, -10,  -4,   6,   3,   3,   3,   8,  18,  32,  23,  11,  -2,  -2,   1,  -1, -19, -50, -37,   4,  -2,  -2,  -4,  -9,   6,  38,  22,  -8,   2,   4,   6,  13,   6, -14,  32,  36,  -4,   2,  -5,  -8, -15,  13, -37,  53 },
      {   1,   1,   0,   1,   3,  -2,   2,  11,  -1,  -1,   0,  -1,  -4,  -3, -14, -17,   1,   2,   0,   1,   2,   6,  25,  25,  -1,  -2,  -1,  -1,   4,  -3, -38, -33,  -1,  -1,  -2,  -2,  -9, -15,   3, -13,   0,   0,  -1,   0,   1,   8, -22,  61,   3,   2,   8,   4,  28, -13,  17,  26,  -1,  -3,  -3,  -7, -37, -49,  33,   9 },
      {   0,   0,   1,   1,  -1,  -4,   0,  12,   0,  -1,  -1,  -2,  -1,   1,  -6, -25,   1,   1,  -1,   0,   2,   0,  15,  49,   0,  -1,   1,   3,   2,   3,  -5, -57,  -2,  -1,  -3,  -4,  -7,  -9, -15,  28,   2,   2,   3,   5,   8,   7,  48,  31,   0,   0,  -1,  -1, -17,   7,   9, -29,   2,   4,   5,   6,  27,  42, -15, -33 },
      {   0,   0,   0,   0,   1,  -1,   4,  -1,   0,   0,   1,   1,  -1,   2,  -5,  -4,   0,   0,  -1,   0,   1,   7,  11,  -9,   0,   0,   2,   3,   6,  -1, -25,   1,   0,   0,  -1,  -3, -10, -17,  30,   2,   0,   1,   1,   2,  15,  48,  -6, -24,  -1,   2,   3,   6,   6, -53, -38, -37,  -5,  -2, -12,  -8, -38,  -6, -45, -37 },
      {   1,   1,   1,   1,  -1,   2,   6,   2,  -1,  -1,  -1,  -1,   0,  -3, -14,  -8,   1,   2,   2,   2,   3,   7,  15,  17,  -1,  -2,  -3,  -3,  -4,  -8, -23, -22,   1,   2,   2,   2,   2,  14,  38,  36,   0,  -1,   0,  -2,   6, -15, -41, -22,  -2,  -1,  -3,   0, -11,  13,  26, -32,  -1,   1,   1,   4,  10,  -6, -40,  72 },
      {   0,   0,   0,  -1,  -3,  -3,   1,  -6,   0,   1,   0,   1,   4,   3,   1,  13,   0,  -1,  -1,  -1,  -4,  -1,   5, -17,   0,   1,   0,   1,   4,  -2,  -9,  23,   0,   0,   1,   0,  -7,   3,  23, -27,   1,   3,  -1,  -2,  14,   8, -35,  41,   0,  -6,  -5,  -5, -22, -48,  49, -11,   0,   4,   0,   5,  11,  69,  18,   1 },
      {   0,   0,   0,   0,   0,  -1,   1,   4,   0,   0,   0,   0,   1,   1,  -1,  -4,   0,   1,   1,   1,   1,  -1,   6,  17,   0,   0,   0,   0,   0,   0, -10, -22,   0,   1,   0,  -2,  -4,  -7,  12,  39,   0,   0,   0,   2,   7,  13,  -1, -48,  -4,  -2,  -4,  -3,  -8, -37, -39,  36,   3,   1,   3,   2,   4,  31,  73,  25 },
      {   0,   1,   1,   1,   0,   1,   5,   5,   0,   0,  -1,   0,  -1,  -1,  -6,  -7,   0,   0,   0,   0,   1,   0,   4,   6,  -1,  -1,  -2,  -1,  -2,  -1,  -5, -11,   2,   2,   2,   2,   3,   4,  15,  21,  -2,  -2,  -3,  -1,  -2,  -4, -33, -39,   2,   0,   3,   1,   2,   6,  52,  70,   0,   1,   2,   3,   5,   3, -27, -65 },
    },
  },
  { // 24
    {
      {-110,  43,  12,   3,   2,   1,   1,   0,  -1, -21,  10,   0,   2,   0,   1,   0,  33, -21,  -5,   2,   0,   1,   0,   0,   9,  -1,  -7,   1,  -1,   0,   0,   0,   5,   2,  -3,  -2,   0,   0,   0,   0,   2,   2,   0,  -1,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      {   9,  23, -11,  -2,  -2,   0,  -1,   0, -89,  46,  17,  -5,   2,  -1,   1,  -1,  -9, -40,  30,  -1,   2,   0,   1,   0,  34, -33,  -6,   9,  -3,   2,  -1,   1,   8,   2, -13,   4,   0,   0,   0,   0,   4,   1,  -4,  -1,   1,   0,   0,   0,   2,   1,  -2,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
      {  32, -24,   5,  -1,   1,  -1,   0,   0, -33, -42,  26,   7,   3,   1,   1,   1,  60, -43, -33,  16,  -3,   3,  -1,   1,   5,  44, -38,  -4,   1,  -1,   0,   0, -14,  22,   4, -13,   2,  -2,   1,  -1,  -3,   6,   7,  -4,  -1,   0,   0,   0,  -2,   0,   4,   0,  -1,   0,   0,   0,  -1,   1,   0,   0,  -1,   0,   0,   0 },
      {  22,  68, -39,  -6,  -6,  -1,  -2,  -1, -14,  14,   8, -11,   2,  -2,   1,  -1,  42,  13,  -6, -12,  -3,  -1,  -1,  -1, -35,  35,  34, -13,  -1,  -2,   0,  -1,  -2, -36,  35,   8,  -3,   1,   0,   1,   4, -13,  -3,  14,  -3,   1,   0,   1,   0,  -5,  -3,   5,   1,   0,   1,   0,  -1,  -1,  -2,   2,   1,   0,   0,   0 },
      {  20,  38, -32,  -1,  -5,   0,  -1,   0,  70,  23,  -1, -20,  -2,  -5,   0,  -2,  12, -47,  37,   2,  -3,   2,  -1,   1,  15,  -3, -29,  18,  -6,   3,  -2,   1, -25,  35,  -7,  -6,   5,  -4,   2,  -1,  -9,  -1,  13,  -6,   1,  -1,   0,   0,  -5,  -1,   4,   2,  -2,   1,   0,   0,  -2,  -2,   1,   2,  -1,   0,   0,   0 },
      { -23, -12,  -3,   6,   2,   3,   1,   1, -12,  49, -34,   3,  -1,   0,   0,   0, -21,   8, -15,  -3,   8,   0,   2,   0,  22,  37, -25, -30,   5,  -4,   1,  -1, -13,  27,  53, -33,  -4,  -4,   0,  -1,   9, -32,  33,  13, -11,   2,  -2,   1,   3,  -6,  -1,  13,  -2,  -1,   0,   0,   1,  -5,   0,   5,   1,   0,   0,   0 },
      {  12,  49, -13,  -8,  -2,  -2,  -1,  -2,   3, -48,  34,   9,   0,   1,   0,   0, -80, -15, -21,  26,   8,   5,   2,   2,  -3,   2, -25, -20,   9,   0,   3,   0,  19,  -4,  11, -17,  -5,  -1,  -1,   0,  10,  -4,   2,   3,  -6,   0,  -2,   0,   4,   2,  -2,   2,   0,   0,   0,  -1,   1,   2,   0,   0,   0,   0,   0,   0 },
      {  -8, -30,  40,  -6,   2,   1,   1,   0,  20,  19,  40, -17,  -5,  -3,  -2,  -2, -11, -41,  25,  18,  -6,   5,  -1,   1,  -5,  19,   0,   2,  -6,  -1,   1,   0, -12, -35,  37,  22,  -7,   1,  -1,   1,  12, -29, -38,  34,   2,   1,   1,   1,  -3,  13, -25,  -5,  11,  -2,   1,  -1,   1,   0,  -4,  -6,   3,   1,   0,   0 },
      {  -2,  14,  44, -22,  -1,  -4,   0,  -1, -11,  29,  22,  -6,  -9,  -1,  -1,   0,  -1,  23,   8,  29, -12,  -2,  -4,   0, -65,   3, -23,  37,  17,  -3,   2,  -1, -13,   4, -21, -28,  28,  -2,   4,  -1,  10, -16,  25, -21,  -2,   6,  -1,   1,  11, -15,   2,   9,  -9,   2,   0,   0,   1,  -1,  -3,   2,   0,  -1,   1,   0 },
      {  23,  29,  58, -36,  -3,  -5,  -1,  -3,   4, -20,  21,  -7,  -4,   2,  -1,   1,  15,  30,   2, -14,  -6,  -3,  -1,  -1,  50, -20,  23,  -2, -10,   1,  -3,   1,  -9,  28,  10,   3, -15,   0,  -2,   0, -12, -12,  27,  27,  -8,  -2,  -2,   0,   0, -12, -18,  28,   4,  -2,   0,   0,  -5,   3,  -9,   2,   7,  -1,   1,  -1 },
      {  13,  30,  35, -16,  -5,  -5,   0,  -2,  19,  50, -12,  26, -15,  -1,  -3,   0,  10,  -2, -56,  37,  11,  -5,   3,  -1,  30,  -1,  -4, -18,  18,  -5,   1,  -1,   0, -13,  -8,  12,   6,   7,  -2,   2, -15,  22, -32, -14,  12,   2,   2,  -1,  -8,   9,  18, -24,   1,   2,   0,  -1,  -1,  -5,   9,  -2,  -5,   2,   0,   1 },
      {  15,   4,  31,   5,   0,  -2,  -4,  -1,  10,   0, -40,  18,  10,  -1,   1,  -1,  24, -47,   3,  -7,  21,   1,   3,   0, -41, -48,   0, -24,  11,  11,   4,   3,  33, -13,   9, -34, -22,   6,  -2,   2,  11,   7,  14,  18, -26,   0,  -3,   0,  -2,  10, -11,  15,  -1,  -4,   2,  -1,  -2,   4,  -2,   2,   3,  -1,   0,  -1 },
      {   5,  22,  29,  15,  -5,  -2,  -3,  -1, -14, -21, -35,  43,  10,   2,   0,   1, -22, -30,   3, -25,  20,  -1,   3,   1, -15,  10,  25,   3, -10,   7,  -3,   1, -69,   7,  -9,  20,  10,  -2,   5,  -1,   0, -29,  -4, -17,  16,  -1,   2,   1,  11, -10,  -7, -14,  -2,   5,  -1,   1,   6,  -1,   0,  -6,  -5,   1,   0,   0 },
      {  -5, -14,  15, -26,  10,   1,   2,  -1,  10,  13,  24, -26,  10,  -3,   1,  -1, -16, -30,  -8, -28,  10,   7,   3,   1,   1,  19,  30, -28, -20,   5,  -1,   1,  -4, -17,  -8,   5, -10,  -1,   4,   1,  19,  26,  18, -40, -14,  -5,  -2,   0,   9, -15,  49,  23, -28,   0,  -4,   1,   1,  -4, -12,  30,  -5,  -4,   2,  -1 },
      {   3,  15,  27, -37,   5,   0,   2,  -2, -12,  -1, -39, -26,   8,   5,   4,   1, -16,  -8,   5, -26, -18,   9,   0,   3, -28,  34, -18,   9, -31,   0,   1,   0,  44,  33,   1,  20, -11, -17,   1,  -4, -13,  16, -11,  16,  21,  -8,   1,   0,  -6,   9,   8, -27,  13,   0,  -2,   0,   1,  -9,  20,  -1, -10,   5,  -2,   1 },
      {   8,   6,   3, -36,   8,  -1,   3,  -1,   0, -28, -43, -22,   9,   2,   4,   2,   8,   2,  12,  -3, -16,  -4,  -1,   0,  39,   4,  -5,  17,  11, -12,   0,  -3,   1, -33,   4, -23,  33,   5,  -1,   2,  44, -27, -18, -23, -15,  18,  -4,   5,   6,  35,  -1,  -7, -24,   1,   0,  -1,   0,   6,  15,   9,  -8,  -4,   1,  -1 },
      {  14,  18,  37,  72, -33,  -3,  -9,   2,   8,   3,   1,  19,   4,  -5,   0,  -3,  -2,   3,  17, -18, -10,   2,  -1,   0,  25,  38, -11,  24,  -9,  -9,   0,  -1,  42, -25,  -3,  -2,  17,  -9,   2,  -1,  14,   9,  11,  -5,   0,   1,  -5,   0, -13,  -7,  11,  15,   1,   2,   1,   0,   1,  -5, -12,   8,   5,   1,   2,   1 },
      {  -8,  -5, -20, -34,  -6,  10,   1,   1,   0,   9, -23,  27, -25,   8,  -2,   3, -14, -21, -24,   9, -19,   1,   1,   0,   1,  22,  16,  29, -18,  -7,  -3,  -2,  -5,  -7, -26,  12,   8,  -4,   3,   0,  27,  25,  11,  16, -17,  -6,  -5,  -2, -17,   9, -23,  48,  23, -11,   4,  -3,  -5,   8, -19, -22,  30,  -2,   1,   0 },
      {  -3, -16, -11, -27,  -2,  -1,   3,  -1,   3, -12, -13,  30, -20,  -2,  -1,   0,   2, -30,  24,  30, -10, -14,   1,  -2,   5,   3,  35,   2,  16, -18,  -3,  -4,  43,   0,  26,  11,  28,  -3,  -5,  -1, -34, -18,  23,  -7,  24,  17,  -2,   1,   6, -40,   7,  -3,  -9,  17,   1,   3,   4,  -6, -15,  10, -12,   2,   2,   1 },
      {   3,  -7, -12,  -5,  -4,   4,   2,  -2,  17,  20,  23,   9,   7,  -4,   3,  -3,  11,  -5, -48, -32,  17,   8,   1,   2,   0, -17, -11,   5, -30,   5,   0,   1,  38,   5, -13,  17,  13, -20,   5,  -4,  20, -52,   8, -17,  21,   4,  -4,   4,  24,  -2, -34,  -3, -12,  12,  -5,   2,  -6,  28,   0, -10,  -9,  -1,   1,  -3 },
      {  -4,  -7,  -6, -40, -24,   6,   3,   1,   6,   4,  29,  52, -12, -14,   1,  -4,   9,  18,  47, -27,  44, -17,   2,  -3,   2,  29,  -5, -19,  15,  14,  -7,   1,  13,  11, -16, -12,  -2,   8,   4,  -1,  26,   6,  -6,   0,  -3,   0,   0,   1, -12,  15,   1, -13,   9,   0,   3,   0,   3, -10,  19, -11,  -9,   3,  -2,   2 },
      {   2,   1,   0,   9, -46,  16,  -6,   6,  -6, -11, -12,  -2, -57,   7,   2,   3,  11,   8,   9,  27,  -8, -16,   2,  -3, -11, -18,  -9, -17, -15,  -1,   0,   2,  -2,  21,  11,  30, -22, -20,   0,  -2,  39, -11, -11,  -3,  14, -17,  -1,  -1,  27,  19,  24,  10, -15,  -5, -10,  -1,  -3,   2,  -2,  33,   9,  -2,   0,  -4 },
      {   4,   7,  -5,  -3,  71, -18,  -2,  -6,   4,   4,   5,  58,   2,  16, -11,   2,  11,  16,  23,  16, -26,  28,   0,   0,   4,  -8, -20,  -3, -32,  16,  11,   0,  -1, -16,   6,   7, -19, -21,   8,  -1,  12, -11,   6,  -4,   5, -14,  -1,   1, -14,   7,  15,   2,  -7,   1,  -4,   1,   5,  -8,   0,  17,  -3,  -3,   0,   0 },
      {  -3,  -8,  -5, -18, -38,  27,  -4,   4,   2,  -3,   9,   6, -12,  14,   3,   0,   0,  -3, -18, -24, -20,   9,   5,   1,   4, -33, -14,  17, -15, -10,   8,   1, -25, -39,  22, -23,   1, -21,   2,   1,   5,  12,  26,  13,   8, -14,  -6,  -2, -23, -22,  14, -27,  19,   8,  -2,   3,  27, -17,  19,  -2, -33,   8,  -5,   1 },
      {   8,   3,   3,   7,  26,  -4,   1,   0,   6,  -2, -13, -23,   8,   1,   2,   1,   9,   0,   7,  21,   4,  -7,   0,  -1,  14,   8,  10, -15,  24,  -1,   1,   1,   4,   3, -21,  -6, -13,  -9,   5,  -1,  64,   9,   4,   8,  28, -21,  -2,  -1,  24, -48,  -6, -15,  36,  15,  -8,   6,  27,   0, -14, -25,  -6,  17,  -4,   1 },
      {   1,   0,   3,  24,  17, -16,  -1,  -5,   4,   7,  29,   9,  -3,  -3,  -5,   3,  -3, -13, -13,  -6, -26,  -4,  -1,   0, -10, -13,  55,   4,   2, -28,   0,  -3,   2,  49,  24, -17,  24,  -9,  -8,  -1,  35,  16,   6,  -2,   4,   9,  -8,  -2,  -5,  38,   4, -11,   8,   5,   2,  -2,  -7, -16,  35,  -5,  -5,   5,   1,   1 },
      {   1,   3,   2,  10,   6,  10,  -6,   3,   4,  10,   6,  11, -38,  27, -14,   8,  -4,  -8,   6, -16, -39,   6,   1,   2,   5,  25,  14,  -9,  11,  -7,  -1,   2,   4,  -1, -26, -28, -22,  11,   3,   5, -21,   7,  -7,   7, -21, -10,   3,  -4,  44,  -3, -27, -16, -19, -11,   4,   1,  20,  40,  14,   8, -33, -14,   0,  -3 },
      {   4,   6,  11,  13,  42,  13,   4,  -5,  -3,  -6,   2, -22, -60,  21,   6,   5,  -1,  -5,   8, -37,   5, -27,  15,  -2,  -5,  -3, -22, -36,  15, -18,   5,   4,  -4, -20, -10,   7,  19, -17,  -8,   4, -14,  -6,  12, -14,  14,   7, -13,   0, -26,   6,  -2,   7,   8,   8,  -2,  -2, -21,   0,   3, -26,  18,   6,   0,   2 },
      {   2,  -2,  -1,  -9,  -5,  16,  -2,  -2,   5,   1,   5,  14,  19,   3,   1,  -2,   5,   0,  10,  -8,  -6,  14,   7,   1,   5,  -4,  -1,  -8,  -1, -23,  15,  -2,   5,  -9,   6, -18,  17, -17,   9,   4, -28,  -9, -16, -15,  -7, -26,  -1,   4,  67,   7,  22,  11,  40, -28, -15,  -1, -12, -18,   9, -16,  41,  20, -10,   3 },
      {   3,   2,  -2,  -3,  13, -25,   6,   1,   4,  -1,   3,   9, -17, -13,   4,  -1,   2,   2, -10, -34,  -7,  -7,   1,   4,  -8, -15, -28,   0,  -1, -16, -11,   0,   2,  13,  13,  21,  21,  22, -20,  -1,   8,   6, -24,   8, -19,  24,  11,  -5,  18, -16,   6,  10,  -2, -16,   9,  -3,  78, -21, -16,   4,  16,  -4,  -7,   3 },
      {   5,   3,   5,  14,  13,  38,  -9,   4,  -1,  -6,  -2, -19, -28,  19,  -6,   2,   4,  -2,  -4,  27,  33,  14,  -7,   0,   6,   6,  26,  17, -23,  29,   5,  -6,  20,   7,  -1,   1,  -2,  15,  14,  -4,  -4, -29,  13, -20, -30,   3,  19,   5, -14,   0,   6,  -8,   1, -21,   5,   3,  37, -28,  28, -25,   7,   5,  -7,   5 },
      {   0,   3,   6,  -2,  -7,  23,  -4,   0,  -3,   0,  -2,  -5,   8,  -3,  -4,  -3,   5,  14,  22,  22,  18,  22,  -2,   0,  -2,   0,   4, -29, -34,  -2,  20,   2,  -8,   7,  20, -16,  25, -36,  -1,   5,  -9,  45, -12, -21,  14,  22, -17,   7,   2,  13, -34,   8, -15,  13,   5,  -5,  38,  25, -19, -19,   1,  -4,   9,  -3 },
      {  -3,  -4,  -1,   2,  -4, -68,  15,  -5,  -4,  -5,  -7, -12, -28, -50,   8,  -3,  -3,  -3,   0,  11,  13, -12, -14,   5,   6,   7,   4,   7, -15,  13,  -4,   2,  -4, -29,  -2,  -3, -10, -17,   5,  -1, -21,   5,  39,  -5,   4, -17,  -5,  -2,  19,  28,  -6,  -2,   9,   1,  -9,  -1,  23,  12,  27,  -8,   1,   8,  -7,  -1 },
      {   3,   6,   3,   3,  43,  47, -12,   1,   1,   0,   3,  11,   2, -10,  11,  -7,  -3,   3,  -5,  -2,  30, -49,   3,  -4,   3,   2, -10,  51,  -5,   3, -20,   2,  -2,  -9,  36,   7, -10,   6,  -8,  -1,   1,  36,  12,   5,   5,  -1,  -4,  -2,  36,   8,  -4,  -3,   0,   4,  -4,  -1,  -8,  14,   5,  16,  -7,   2,   1,  -3 },
      {   0,   0,   0,   3,   5,  24,  10,  -4,   2,   5,   5,   3,  39,  -9,  25, -10,  -1,  -1,  -2,  19,  -7, -46,  29,  -2,  -4,   7,  11, -11,  -4, -33,   2,   9,   1,   3, -36,  -3, -12, -20,  -9,  11, -14, -32,   4,  15,  -1, -14, -18,   3, -29,  15,   8,   7,  -6, -13, -10,  -5,  46,  11,  12,  19,   5,  -7,  -8,  -6 },
      {  -3,  -4,  -3, -10, -11, -21, -36,   4,   3,   3,   5,  -6,  19,  57, -38,   2,   0,  -4,  -3, -12,  16,   0,   5, -10,   1,   5,   0,  18,  47,   4,   5,  -4,   3,  -3,   8,  17, -14,  -9,   6,  -1,  -6,   1,  25,  -6,  25,  -8,  -6,  -3,   3,  37,   0,   1,  -2,  15,   4,  -5,  35,   2,  -4,   8,  30,  -9,   7,   0 },
      {   0,   0,   6,  -1, -14,   6,   0,  -1,   2,   2,   1,   7,  29,  10,   2,  -8,   2,   3,  15,  17, -30,  -4,  11,  -4,  -2,  -3, -16, -30, -15, -11, -10,   2,  -3, -12,   6,  40,  -3,  35, -25, -10,  14,  11,  48, -16,  -4,  21,  24, -12,  10,  -1,   0, -16,   0, -10,  12,   7,  -7,  29,  33, -34,  15,  -3,  -9,   6 },
      {  -1,  -2,  -5,   1,   5,   0, -44,  23,   0,  -1,  -4,  -5,  -1,  -8, -16,   0,   2,   2,  -1,  12,  20,  20,   4,  -4,   2,   7,  -4,  -9, -16, -13,   3,  11,  -6, -11, -38,   9,   5,  -7, -35,   5,  -2,  -8,   9,  41,  -9,  39, -29, -13,  29,   0,  10,  11,   4,  22,  15, -17, -10, -32,  25,   9, -17,   7,  14,   1 },
      {   0,  -1,  -2,   2,   0, -28,  31, -14,   0,  -2,   0,  -2,  -8,  30,  17,  -2,   3,   2,   5,   6,  50,  27,  28, -12,   1,   4,   1,  23,   3, -32,  31,  -2,  -2,   3,   6,  12,   7, -19, -10,  13,  12,  15,   1,  17, -11,   1,   2,   4,   0, -31,  -3, -10, -10, -32,   5,   4, -15,  19,  28,  19,   6, -24, -11,   2 },
      {   3,   2,   1,   8,  -5,  26,  76, -20,   3,   3,   1,   3,   8,  19,  12,   7,   0,  -3,  -3,  -1,  -8,   8, -32,  18,   3,   2,   0,  -6,  14,  25, -13,   0,  -2, -11, -12,   2,  -1,  -5,   4,  -4,   0,   7,  12,  18,  23,  18, -15,  -4,  23,   7,  -8,   8, -10,  16,  15,  -5,   7, -35,  22,  13,  21, -21,  12,   2 },
      {   1,   3,   0,   5,  -6,  -8,  41, -13,   0,   1,  -1,  -1,   4,  34,   1,   0,   3,   5,   5,   9,  16,   6,  16,   5,  -4,   2,  -6,   6,  -6, -29,  -5,   9,   2,   3,   2,  20,  -8,  18, -15,   3,  -5,  -2,  10, -14, -29, -16,  -5,   2,  10,  33,  -6,  11,   7,  21, -14,   0,   2, -21, -36, -15, -57,  39,  10,   0 },
      {  -2,  -3,  -4,  -2, -12, -36, -15,   1,   3,   3,   2,   0,   3,  36,  33, -12,   0,   0,  -2,   8, -13, -40,  36,  -4,   2,   3,   2,   5, -29,  17, -12,  14,   2,  -3,   6, -26, -24,  -5,   2,   5,  -2,   7, -21, -23,   1,   6,  -8,  -1,  11, -21, -25,   2, -22,  10,  15,  -7, -11, -46,   7, -22,  12,  -9,   3,   5 },
      {   0,   0,   1,   1,   3,  15, -18,  -6,   0,   0,   1,   2,  -9, -28,  61, -26,  -1,   0,  -3,  -6, -22,  33,  38, -13,   2,   4,   3,   2,  39,  28,  35,  -2,   0,   5,  -1,  25,   2,  -1,  13,  15,   1,   5,  22,   7,  -4,   2,   0,  12,  13,  23,  -7,  -9, -12,   2,   4,   8,   3, -19, -19,  -6,  -8,  -6,   0,   6 },
      {   1,   2,   1,  -3,  -3,   5,  20,  31,   0,   1,   1,  10,  -6, -13,  11,  22,   0,   0,   1,  -7,  -8,  16,  -3,   6,   0,   3,   4,  25,  11,  -4,  15,  -1,   2,   8,  12,   0, -49,   2, -25,  21,   1, -10, -10, -22, -23,  19, -37,   4,  -6, -13,  22, -18,  -4,  39, -14, -22,   3,  25,   2, -25,  36,  10,   3, -10 },
      {  -1,  -1,  -1,   1,   0, -15, -12,  10,   2,   4,   5,   1,  -6,  44,  61,   2,   0,   1,   2,   4,   8,  12,  -2,  13,   0,   2,   9,  -6, -21,  26, -19,  -4,  -1,   1,  -5, -22,  30,  30,  -7,  -9,  -4,  -2,   1,  22,  13,  14,  -6,   2,   0,  13,  23,  -8,  27,  13, -23,   2,   7,  36, -10,  24,  12,  17,  -8,  -5 },
      {   1,   1,   2,   1,   3,  -6,  -2,  72,   2,   2,   0,   3,  16,  10,  27,  41,   0,   0,   3,   3,  -3,  -2, -11,   5,  -1,  -3,  -7, -10,  24,   6, -17, -11,   0,   1,   9,  31,  21, -21,  21,  -8,  -2,  12,  -3,  -2, -16, -21,  -8,  13,  -3, -10, -16,  29, -16, -10, -12,  -1,   2,  -4,  32,  -8, -23,   8,  -4,  -3 },
      {  -2,  -2,  -3,   0,  -4,   5,  -8, -27,  -1,   1,  -2,   0,   0,  -5, -18,  -5,  -1,  -1,  -1,  -1,  -9,   0,   6,  -3,  -1,   4,  -5,  -8,   0,  10,  20,  -4,  -1,   0,  -4,  10,  10,  23, -10,   6,   1,  14,  -2,  -9, -11,  -7,  -3,  23,  -9, -19, -48,   3,  -7,  26, -50,   4,   3, -16,  34,  45,  23,  44, -17, -15 },
      {  -3,  -3,  -2,  -1,  -4,  -4, -37, -70,   0,   1,  -2,   3,   5,  14,  24, -17,   0,   1,   2,   1,  -1,  -8, -48,  25,  -1,  -3,  -6,  -1,  15,  -3, -10, -13,  -1,   6,   5,  23,  -3, -24,  14,  11,  -1,  -2, -12,   0, -34,   3, -19,   5,  -1, -15,  16,  13,  -1,   9,   3, -10,   1,  16,  18, -21, -16,   1,   3,  -3 },
      {   0,   0,   1,  -4,   3,   7, -19,   3,  -1,   0,  -2,  -1,  -6,  -6,  28, -12,   1,   1,   3,   2,   6,  40, -26,   9,   1,   2,   4,  -2,  13, -25, -52,  10,   2,  -3,   0,  -4, -16,  -8, -47,   1,   1,  -3,   7, -37,  10, -25,  -6, -11, -15,   1, -36, -16,  10, -28,   6,  -9,  -2, -19,  -9,  24,  10,  -8,   3,  -7 },
      {   0,   0,  -1,  -2,   2,  -3,  -8,  20,   2,   3,   3,   3,   2,   7,  20,   7,   0,  -1,   1,   0,   7, -23, -44,  13,   1,   2,  -2,  -7, -13, -44,  33, -22,  -4,  -4, -15,   3, -19, -10,  31, -18,   0,   5,  16,   1, -23,  27,  23,   3,   4,  10, -21, -38,   2,  13,   0,  14,  -9, -22, -30,  20,  15,  -4, -20,   6 },
      {   0,   1,   1,   1,   4,   5,   3,   7,  -2,  -2,  -3,   0,  -8,  -8,   8, -39,   1,   2,   1,   0,   6,  32,  20,  -6,  -2,  -2,   0,   5,   9, -41, -39,  19,  -3,  -1,  -1,  -7, -20,  -5,  46, -34,  -4,   1, -13,   4,   3,  13,  38,   2,  -4,  -2,   5,  24,  -8,  44,  -5,   6,   7,  10,  26,  -1,  12,  11,   9,   5 },
      {   0,   1,   0,  -1,   1,  -1,  16,  19,   0,   1,   2,   4,   1,  16,  -1, -52,  -3,  -3,  -4,  -5,  -6,  -8, -36, -49,   1,   3,   5,   3,   8,  19,   0, -11,  -3,   1,  -6,  -7,   0, -36, -36,   0,  -1,   0,  -6,  14,  -3,   5,  17,  -3,   0,   1,   8,   0, -31, -11,  12,   9,  -2,   7,  -6,  -9,   3,  56, -21,   3 },
      {  -1,  -1,  -1,  -2,  -1,   6, -14, -28,   1,   2,   2,   2,   2,  -9,   7,  59,   1,   1,   1,   0,   7,   8,  21,  41,   2,   3,   5,   8,   5,  -2,  -9,  -3,  -1,  -4,  -5, -12,   1, -20, -28, -16,   0,   3,   1,  25,   7, -16,  36, -14,   2,   5,   7,  -5, -53,   8,   6,  14,  -1,  -4,  -9, -19,  17,  30, -17,   4 },
      {   1,   0,   2,   0,   1,   3,   5,  -2,   0,   0,   0,  -2,   0,  -5,  11,   4,   1,   0,   2,   2,   1,  -6,  11, -22,  -1,  -1,  -1,  -6,  -7,   7,   9, -61,   1,   2,   0,   0,  15,  -5, -24, -16,  -4,  -6,  -9, -10, -26, -45,  21, -22,   0,  -3,  -5,   2,  22,  51,  41, -15,   4,   8,   9,  19,   6, -15,  25,  14 },
      {  -1,  -1,  -1,  -1,  -2,   1,  -4, -15,   0,   1,   0,   1,   2,  -3,  22,  30,  -1,  -1,   0,  -4,  -2,   2,  -2, -13,   1,   3,  -1,   9,  10,  -2,  35, -14,  -1,  -1,   3,  -6, -37,   9,  -4, -26,  -3,  -5, -13, -29,  36,  38,  -1, -26,  -5,   1,  -9,  46,  24, -24,  17,   6,   7,   7,  18,  15, -25,  31,  -6,  17 },
      {   0,   0,   0,  -1,   0,   0,   9,  13,  -1,   0,   0,  -1,   3,   5, -12, -15,   0,   1,  -1,   1,  -7, -12,  20,  53,   2,   2,   5,   1,  18,  34,   9,  51,   1,   5,   5,  11,   9, -44,  -9,  -8,   0,  -1,  -3, -17, -34,  19,  23, -24,  -3,  -3,  -6,  -7,  37,   4,  14,  -5,  -3,  14,  -1,  28,  -3,  10,  -8,   8 },
      {   0,   0,   0,   2,  -4,   1,   0, -23,   1,   2,   1,  -1,   4,  -1,   0,  43,  -1,  -1,  -1,   2,  -5,   1,   7, -61,   0,   2,   1,  -1,   4,  15, -36,   6,  -1,  -2,  -4,   1,  -5, -39, -14,  -2,  -4,  -1,  -1,  -6, -25,  34,  18,  13,   4,   4,   3,  -5,  28,   9, -34,  33,   2,  -6,  -7,   0,  -6, -30,  -6, -13 },
      {   0,   0,   0,  -2,   1,   1,  -8,   2,   1,   1,   0,   3,   2,   4,  22,  18,  -1,  -2,   0,  -2,   3,  -4, -42, -25,   0,   2,  -1,   1,  -4,  -9,  32,  68,   4,   2,   4,  -2,  10,  17,  -4,  48,   0,  -3,  -3,   0,  10, -26,  42, -24,  -2,   3,   1,  18,  -2,  27,   4,   6,   3,   0,  11,  -3,   1, -12,   7,  -6 },
      {   0,   1,   1,   0,   1,   2,  -2,  10,  -1,  -1,  -2,   1,  -1,   3,  -1, -31,   0,   0,   1,  -1,   3,   1, -10,  21,   1,   2,   3,   5,   5,   3,  29,  -3,   0,   1,   3,  -4,  -1,   5, -40, -50,  -4,  -1, -11,  -1,   5, -14,  -7, -19,   2,   3,   6,  23,  -5,  17, -44,  28,   0, -10,   5, -24, -14, -55, -36, -11 },
      {   0,  -1,   0,  -1,   0,  -1,   0, -13,   1,   0,   1,   1,   1,  -1,  10,   8,   0,  -1,   0,  -1,  -2,  -4,  -4, -31,   2,   0,   2,  -1,   3,   3,  19,  31,   1,   0,   2,   5,  11,   0,  12, -69,   1,   2,   4,   7, -17,  -8, -18, -38,  -2,  -8,  -3, -25, -16, -21, -19, -33,   1,  -1,  -7,  -5,  11,  13,  49, -30 },
      {   0,   1,   1,   1,   0,   5,   2,  -9,  -1,  -1,  -1,  -3,   1,  -4,  -6,   9,   1,   2,   0,   3,   2,  11,   3, -27,  -1,  -1,  -1,  -3,  -2,   0, -20,  14,   2,   4,   4,   6,  11,   7,  31,   2,  -1,   0,  -2,   0, -13, -11, -47, -58,  -2,   1,  -6,  -9,  -8,  15,   8,  -3,  -1,   2,  -8,   6,   5,   0, -71,  43 },
      {  -1,   0,  -1,  -1,  -1,  -1,  -4,  -2,   0,   0,   0,   1,  -1,  -1,   1,   7,  -1,   0,  -2,   0,  -1,  -3, -13,  -5,   0,   0,   0,   2,  -1,  -2,   7,  38,  -1,   1,   0,   2,   1,   7, -17, -36,  -2,  -2,  -4,   0, -10, -19, -34,  55,  -1,  -3,   2,  -7,  -1,   3,  24,  53,   3,   8,   3,   6,  11,   4,  24,  64 },
      {   0,   0,   0,   1,   1,   1,   0,  -8,   0,   0,   0,  -1,  -1,   0,   1,  12,   0,   0,   0,   0,  -1,   2,  -2, -13,   1,   0,   1,   0,   1,  -1,   6,  25,   0,  -1,   0,   0,   0,   6,  -5, -35,   0,   0,   2,  -2,   5,  -8,   4,  60,   4,   3,   5,   5,   8,  12,  46, -53,  -3,   0,  -5,   1,  -7,  -6, -59, -39 },
      {   0,   0,  -1,   0,  -1,   2,  -1,  -3,   1,   0,   1,   1,   1,   1,   0,   2,   0,  -1,  -1,  -1,  -1,  -2,  -7,  -5,   0,   1,   2,   0,   1,   3,   2,   1,  -1,  -2,  -3,  -4,  -1,  -6, -13,  -3,   3,   2,   4,   5,   7,   6,  29,  18,  -1,   0,  -2,   1,  -5,  -5, -49, -75,  -1,   0,  -4,  -5,  -8, -15,   1,  79 },
    },
    {
      {-109,  53,   0,   5,   1,   2,   0,   0,  15, -17,   8,  -3,   2,  -1,   1,   0,  28, -16,  -1,   1,  -1,   0,   0,   0,   1,   3,  -4,   0,   0,   0,   0,   0,   3,  -3,   2,  -1,   0,   0,   0,   0,   3,  -2,   0,   1,   0,   0,   0,   0,   2,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      {   9,  28, -22,   4,  -5,   1,  -1,   0, -95,  34,  14,  -2,   4,  -1,   2,   0,  21, -48,  22,  -5,   4,   0,   1,   0,  30, -14,  -8,   4,  -2,   1,  -1,   0,   6,   3,  -6,  -1,   0,   0,   0,   0,   4,   0,  -1,  -1,  -1,   0,   0,   0,   2,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,  -1,   0,   0,   0 },
      { -33, -77,  60, -10,  13,  -3,   5,  -1, -34,   7,  -8,   8,  -2,   3,  -1,   1,  44,   5, -33,  10,  -8,   2,  -2,   1,  24,   7, -10,  -6,   0,  -1,  -1,   0,  -3,  14,   5,  -4,  -1,  -1,   0,  -1,  -3,   0,   2,   1,   0,   0,   0,   0,   1,   2,  -1,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -40, -10,  13,   1,   4,   1,   1,   0, -17,  37, -20,   3,  -2,   2,  -1,   1, -72,  40,  10,  -7,   4,  -2,   1,  -1,  23, -53,  24,  -3,   1,   1,   0,   0,  29, -18, -14,   7,  -3,   1,  -1,   0,   5,   7,  -5,  -4,   1,  -1,   0,   0,   6,   0,   3,  -1,  -2,   0,   0,   0,   2,  -1,   1,   2,  -1,   0,   0,   0 },
      {   0, -24,  18,  -1,   2,  -1,   0,   0,  19, -56,  43, -12,   9,  -3,   3,  -1, -31, -39,  29,   5,   4,   2,   0,   1,  34, -27, -34,  23,  -7,   5,  -1,   1,  15,  26, -37,  -2,   2,   0,   0,   0,  -4,  21,   1,  -9,  -1,  -1,  -1,   0,  -4,   5,   4,   0,  -1,   0,   0,   0,  -2,   2,  -1,   0,   0,   0,   0,   0 },
      {  -9, -18, -14,  24,  -6,   6,  -2,   2,  47,  57, -49,   0,  -4,  -1,  -1,  -1,   9, -28,  36, -18,   5,  -3,   1,   0,  24,  -6, -16,  10,  -8,   3,  -3,   1, -35,  45,  -5,  -2,   2,  -4,   2,  -2,  -6,  -9,  17,  -4,   3,   0,   1,   0,  -1,  -3,   2,   2,  -1,   1,   0,   1,  -1,  -3,   1,   1,   1,   0,   0,   0 },
      {   2,  27,  -6,  -5,   0,  -3,   0,  -1,  -6, -10,   4,   0,  -2,   2,  -1,   0, -49,  12, -37,  22,  -5,   4,   0,   1,  65,  26, -23, -20,   0,  -4,  -1,  -1, -21,  32,  42, -14,  -2,  -3,   1,  -2, -11, -27,  18,  14,   0,   0,   1,   0,   1,  -8, -12,   6,   2,   1,   0,   0,   1,   0,  -2,  -3,   2,  -1,   1,   0 },
      {  19,  28, -19,   4,  -7,   1,  -3,   0,   8, -20, -15,  11,  -3,   2,  -1,   1,  51,  45, -28,  -7,  -2,  -5,  -1,  -1,  44, -38,  19,  -3,  -3,   1,  -1,   1,  15,  27, -26,   6,  -4,   0,  -1,   0, -39,  36,   6,  -9,   4,  -2,   2,  -1,  -2,  -4,  13,   0,   0,  -1,   0,   0,  -7,   2,   3,   1,   1,  -1,   0,   0 },
      {  21,  47,  61, -72,  23, -20,   8,  -6,  21,  23,  -6,  12, -12,   2,  -2,  -1,   5, -26,   2,  16,  -7,   3,   0,   0,  -1, -29,  11,   1,   5,   0,   2,   1, -14,   3, -10,   4,   1,   2,  -1,   1,   0,  -8,   1,  -1,  -1,   3,   0,   1,   1,  -2,  -6,   1,  -2,   2,   0,   1,   0,   2,   1,  -2,   0,  -1,   0,   0 },
      {   5, -15,  -1,   2,   3,  -1,   0,  -1,  17, -18,  17, -13,   5,  -4,   2,  -1,  18, -10,  21, -27,  14,  -6,   3,  -2,  52, -14,  48, -18,  -6,   4,  -5,   1, -33, -36,  26,  35, -14,   8,  -1,   1,   4, -23, -37,  24,   1,  -1,   1,   2,   1,  15, -16,  -6,   4,   0,   0,   0,  -3,   5,  -1,  -4,  -2,   0,   1,   0 },
      {  -7,  -5,   2,  -9,   9,  -3,   1,  -3, -23, -22,  23, -12,  11,  -6,   3,  -1,   7,  38,  32, -25,   2,  -5,   0,  -2, -27, -24,  17,  14,  -4,   5,  -1,   1, -28,  26,  -3,  -5,   6,  -3,   3,   0, -14, -48,  49,   0,  -4,   4,   0,   2,  23, -33, -20,  22,  -7,   3,  -1,   1,  -2,   7, -13,   0,   3,  -2,   1,   0 },
      {   3, -13, -13, -11,  15,  -7,   5,  -3,  28,  41,  34, -64,  22, -13,   3,  -3,  18,  -6,   1,  14, -15,   5,  -2,   2,   7, -12,   3, -13,   4,  -2,   0,   1,  45, -14,  37, -19,  -1,  -1,  -2,   0, -23,  14,  18,   9,  -4,   3,   0,   0,  -4, -23,   9,  13,  -2,   2,   0,   1,  -1,  -3, -12,   6,   0,   1,   0,   1 },
      {   5, -10, -16,  -3,   7,  -3,   4,   1,   3, -28, -45,  22,  -2,   7,  -1,   2,   6, -30,   0,  -8,   6,   6,  -2,   1, -12, -30,  -6, -21,   5,   2,   0,   2,  63,  13,  21, -17, -16,  -2,  -4,  -3,   2, -26,   7,  31,  -6,  -1,  -3,   0,  15,  14, -45,   5,   4,   1,   1,  -1,  -7,  19,   3, -10,  -1,  -1,   2,  -1 },
      {  -4,  -7,  -3, -26,  14,  -6,   7,  -1,  -5, -17, -16,  15,  -3,   3,   0,   0, -18, -33,  -2, -50,  24,  -6,   5,   0,   6,  42,  34, -35,  -6,  -1,  -2,   0,  16,   1,  -3,  13, -10,   1,  -2,   0,  -3,  26,  33, -18,   0,  -3,   0,  -2,  -6, -34,  31,  18,  -7,   4,  -1,   1,   7, -10, -16,  11,   0,   2,   0,   0 },
      { -12, -21, -34,   7,   9,  -1,   7,  -1,  -3,   0,  10,  13, -22,  14,  -6,   2, -16, -34, -15,  50, -28,  13,  -6,   5, -23, -16,  30, -13,   7,  -7,   0,   0, -24,  15, -12,  32, -10,   2,  -4,   0, -40,   5, -16,   6,  13,  -5,   4,  -2,  31, -29,  -1,  -4,   5,   2,   1,   0,   4,   5, -14,   2,   1,   0,   0,   0 },
      {  -9, -16, -60, -45,  52, -21,  20,  -4,  -6, -19, -31, -21,  13,   1,   0,   0,   3,   9,  -5,   3, -26,  13,  -8,   5,  13,  -4,  -7,  17,  -3,  -3,   2,   0, -21,  -9, -11,  -8,  12,   0,   1,   0,  35,  -1,  -4, -20,  -6,   2,  -2,   1,   2,  23,  11,   0,  -9,  -1,  -1,  -1,   3,   0,   6,   6,   2,  -3,  -1,   0 },
      {   8,   1,   8,  31, -14,   5,  -6,   3,   6, -10,  -1,  -5,  16, -10,   2,  -1,   0, -29, -48,  -2,  13,  -2,   4,   1,  -9, -55,   2, -30,  14,   5,  -1,   3, -35, -14,   7, -33,   6,   4,   2,   0,  39,   3,  12, -26, -12,   2,  -3,   0,  24,  -4,  18,  14, -15,  -1,  -4,   0,   6,   2,  -9,  18,  -1,  -2,   1,  -1 },
      {  -2, -14, -25,  -9,  -2,   7,  -4,   3,  20,  17,  20,  46, -19,   3,  -5,  -1,  11,  -3, -15,  11,  18,  -7,   3,  -3,  28,   3, -15,  27,   7,  -2,   6,  -4,   6, -55, -22,  -6,  17,   2,   4,   2,   2, -12,  18, -16, -10,   6,  -4,   3,  -8, -28, -33,  25,   1,   2,   1,   1,  12,  21, -27, -19,   5,   0,   0,   0 },
      {  -3,   1,  -9, -20,   1,   9,   0,   1,   3,  25,  10, -26,  35, -15,   7,  -6, -10,  -7, -61, -27,  32,  -6,   2,   0, -16,   4, -17,  14, -30,  16,  -8,   4,   2,  17, -24,  37, -21,  -1,   2,  -1,  -1,  -4,  -8,   1,  14,  -9,   2,  -2,  16,   6, -21,  -9,   8,   0,  -2,   1,  -5,  30,   0, -17,   1,   2,   0,  -1 },
      {   7,  12,  27,  20,  -5,   5,  -4,  -1, -12, -26, -59, -21,  34,  -4,   5,   0,  -8, -18,  -5,   5, -24,  11,  -8,   5,   4,  -2,   0,  44, -29,   6,  -4,   1,  -6, -40,  15,   8,  10,  -8,   3,  -2, -31,   5,   2,  -1,  10,  -2,   1,   1, -11, -29,  -7,   4,   3,   2,   1,   2,  11,   1, -23,  -2,   6,   2,   0,   0 },
      {  -2,  -5,  -2, -16,  -3,   1,   3,   0,   2,  -2,  -5,   9,  10,  -9,   3,  -2,  -3,  -7,   4, -16,  21, -11,   3,  -1,   4,   6, -22,  -1,  21,  -9,   5,   0,   3, -27,   9, -15,  21,   4,  -1,   4, -66, -24, -13, -33,   7,  11,  -1,   2,  50,   8,  14, -33, -17,   0,   0,  -1,  -7,  -6,  35,  14, -15,   0,  -3,   1 },
      {  10,   6,  11,   5, -23,  16,  -4,   6,  13,   4,   3,  -9, -18,   6,  -1,   0,   9,  -2,  -3,   5, -36,   8,  -9,   5,  22,  23,  19,  13, -24,   0,  -6,   0,  42,   7,  -9,  24,  12, -11,   3,  -3,  37, -35,  -1, -22,   8,   1,   0,   0,  56,   3,  11,  11, -25,   1,  -2,  -2,  -6,   4,  -6,  28,   2,  -6,   1,  -2 },
      { -11, -19, -19, -52, -57,  41, -16,  14,  -4, -11,  -8, -10,  -4,  -6,   6,  -2,  -2, -14,  -9,  26,  15, -16,  10,  -7,  -4, -14,  38,  24,  -5,  -5,  -2,   0,   0,  13,  20,  -6,  20, -11,  -1,  -1,   7,  -4,  13,  -1,   6,   5,  -3,   1, -29,   2,   6,  -9,   2,   2,   3,   1,  -4, -17,  22,  -7,  -5,   3,  -2,   1 },
      {  -1,  -3,  -9,   2,  12,  -4,   8,  -5,   3,   4,  25,  32, -29,   8,  -7,   0,  -1,  -4, -19, -25,  -5,  13,  -8,   8, -15, -34, -26,  19, -44,  14,  -6,   3,   1, -11,  42,  26,  -4, -17,   9,  -4, -14,  -9,  15,   2,  10,  -2,   0,   1, -17,  24,  21,  14,  -4,   0,   1,  -1,   2, -44,   7,  18,   7,  -2,   0,   2 },
      {  -3,   1,   7,  23,  15,  -9,   0,  -7,   1,  12,  17,  10,  10,   0,  -2,   1, -15, -20, -18, -14, -17,  11,  -3,   6,  -2,  16,  55,  16, -14,  -8,   3,  -3,  -3,   9,  -6, -27,  33,  -1,   2,   0, -26,  12,   8, -10, -23,  12,  -6,   3, -11,  45, -24,  24,  -8,  -4,   1,  -4, -34,  16,  24, -12,  11,  -5,   1,  -2 },
      {   3,   0,  -2,  27,  22, -27,   9, -10,   4,   1,   0,  28,  38, -41,  13,  -8,   4,  -5,  -8,  32,  15,  -7,  -1,  -1,   2,  -1,  18,  -3,   9,   5,  -4,   0,  24,  13, -15,  18,  13,   0,   4,   0,  11, -51,  -2, -14,  18,   3,   2,   2, -38, -15,  17, -17,   3,   5,  -3,   4, -19, -26,   1,   2,  -4,   4,   0,  -1 },
      {  -5,  -2,  -5, -22,  -8,  16, -14,   2,  -4,   1, -12,  21,  21, -30,   8,  -6,  -1,   9,  20,  31,  12,  -9,  -4,  -2,  -3,   8, -15, -10, -11,  20, -12,   4,  -8,  -7,   1,   4, -17,  10,   1,   1, -14,  11, -12,  13, -29,   2,   1,  -1,  17,  20,  16,  41,  -6, -11,  -1,  -3, -56,  -3, -33,  23,  27,  -5,   4,  -1 },
      {   0,  -2,  -4, -12, -11,  -9,  10,  -5,   7,   7,  13,  46,  39, -33,   9,  -6,   2,   7,   8,  -9,  -5,  15, -14,   8,   4,   3,  12,  13, -19,  -3,  -1,  -1,   1,  19,  19, -21,  15,  -4,   0,  -1,  18,  32, -25,  20,  -7,   7,  -2,   0,  43, -19,  -6,  -2,  18,  -6,   5,  -1,  43, -23,   4, -14,  -2,   7,  -4,   3 },
      {   6,   4,   4,  15,  43,  31, -27,   4,   4,  -1,  -1,   1,  17,  18, -12,   4,   3,  -5,  -3,  22,   8, -13,   2,  -5,   9,   7,  10,  25,   6,   2,   3,  -3,  -1, -16,  -7,  10, -34,  15,  -2,   5,  14,   6,  40,  16, -16,  -7,   1,  -5,  29,  -9,   1,   4,  22, -13,   5,  -3,  -5, -33,  55,   0,  -8,   8,  -3,   5 },
      {  -1,  -1,   1,   4,  -1, -12,   6,  -1,   0,  -1,   4,  15,  -2, -17,   8,  -5,   4,   6,  27,  34,   4,   2,  -5,   0,   3,   3,   4, -25, -32,  31, -11,   2, -10, -19,  16,  -2,  -9, -33,  13, -10,   8,  27,  42, -24,  30, -22,   5,  -4,  11,  11,  -5, -23,  -1,  14,  -8,   4,  -7,  45,  24, -13, -12,   0,  -2,  -2 },
      {   3,   5,   5,  14,  39,  -6,  -3,  -9,  -4,  -6,  -6, -11, -30,  -4,   4,  -4,   0,   1,   2,  11,  36, -45,  14,  -9,  -3,  -2,   6,  34,  17, -13,  -8,   1,   7,  22,  47,  13,  24,  -2,   1,   0,  12,  28, -12, -21,  23,   9,  -2,  -1,  26,   9,  -5,   4,   7,   1,  -3,   2,  -7,   3, -28, -14,  20,   2,  -2,   0 },
      {   3,   5,   2,  -1,  48,  55, -35,  21,   1,   2,   6,  11,  19,  11,   0,  -2,   2,   5,  11,  12,  -1,  -8,  14,  -6,  -6,  -1,   5, -26, -31,   5,  -1,   2,   5,  16,   1, -16,   0, -10,   5,   0, -14, -13, -26, -24,  -1,   6,  -2,   3, -20,  12,  -3,  -1, -23,  -2,   0,  -2,  51,  11, -14,   7,  -3,  -6,   0,  -1 },
      {  -1,   0,   1,   6, -20, -27,  15, -11,  -3,  -2,  -9, -18, -11, -21,  -3,  -1,   1,   0,   1,  28,  29,  -7, -19,  11,  -1,  10,  14,  10,  11,  18,  -7,   1,   5,   3,  -7,  16, -20,  22,   4,   2, -18,  -5,  19,  -6, -21,   8,   1,   0,  -1,  41,  -7,  23, -15,  -7,  -2,  -4,  74,  -2,   6,  14,   5,  -7,  -1,  -1 },
      {  -1,  -1,  -2,   0,  -2, -28,  -4,  -4,   1,   1,   2,  -4,  -4,  43, -25,  12,   3,   4,  -1,  14,  51,  24, -12,   4,   1,   4,  18,  32, -25,  30,   3,  -2,   4,  11,  12, -39, -21,   7,  13,  -4,   2,  -9, -34,  -5, -15, -13,   1,   6,   0, -27,  18,  -7, -18,  -2,  -3,   1, -16,   9,  -4,   3, -20,  -1,   2,  -4 },
      {   0,   0,   2,  -5, -18,  -2,  -3, -10,   1,   1,   0,  -1,  46,  31, -34,  14,   5,   8,  14,  13,  23,  30, -12,  10,  -2,  -5, -10, -21,   9, -36,  32, -13,  -1,   5,  15,  38,   1, -17,  -1,  -2,   5,  10,   3, -43,  11,   6, -10,  -1,  -1,  10, -28,  21,  -6,  -6,   4,  -2,  -8, -18,  -2, -10,   5,   5,   4,   4 },
      {   2,   1,   1,   1,   2,  -3,  -1,   6,  -1,  -4,  -7, -30, -23,   2,   3,   6,   3,   0,  -1,   8,  20,  10,   3,   3,   2,   4,   3, -32, -20,  17,  -2,   1,   1, -12, -36, -24,  22, -28,  13,  -9, -10, -18,  -7,   0,   6, -13,  -1,   2,  22,  11,   2,  13,  31,   0,   0,  -1,   4, -52,  -7, -50,  29,  16,  -2,   4 },
      {   0,   0,   1,  -5, -12,   2,  16,   6,   5,   5,  11,  19,  39,  43,  -9,   7,   0,  -2,   2,  -3, -13, -12,   2,  -2,   3,   0,  13,  14,  16,   1,  -8,  -1,   3,  -2,   5, -18,  -7,  24,  -8,  -4,  -6, -17,  24,   0,  22, -30,  17,  -7,   8,  42,  36, -18,  33,   0,  -3,   5,  12,  11, -40,  -9,  24,   5,  -3,   3 },
      {   2,   1,   3,   7,   2,  16,  34, -32,   0,  -3,  -4,  -2,  -3,   0,  -5,   4,   2,   0,   4,  13,  19, -11,   2,   5,  -2,  -7,  -6,  -7, -35, -45,  21,  -6,   1,  -4,  -2,   8,   7, -19, -10,   7,   0, -18,   2,  24, -26,  18,  -9,   5,   0,  -8,  54,   5,  13, -22,  -4,  -2,   8,  45,  16, -16,  14,  -2, -10,   3 },
      {   1,   2,  -2,   5,  -1,  -7,  34, -12,   2,   4,   0,   0,  17,  31,  15,  -5,   1,   2,  -4,  16,  20, -28,  29,  -8,   2,   6,   3,   0, -21, -17,   7,   9,  -7,  -8, -24, -12,  11, -28, -12,   1,  -1,  13,   9,  42,  15,  -2, -15,   1,   9,  15, -24,  -1, -21,  27,  -1,   2,   0, -26, -18,  44, -29, -10,   1,   0 },
      {  -2,  -4,  -4,  -2, -18, -31,  12,  -8,   1,   0,   0,   1,   9,  22,   0,   4,   0,  -1,   4,   5, -23, -51,  30,  -9,  -4,  -3, -10, -16, -38,   2, -13,   8,   2,  -1,   4,  -8,  -5,  36, -16,   0,   3,  -3, -21, -29,   8,  -4,  17,  -4,  -6, -25, -21,  25,   1, -16,   0,  -2,   1,  -7,  39,   2,  31,  -4,  -6,  -2 },
      {  -3,  -4,  -4,  -2,  -3, -57, -58,  48,   1,   1,  -2,   3,  16,  10, -17,  11,   0,  -1,  -1,   3,   3,  -2,  21, -15,  -2,  -2,  -8,  -2,   1,  -1,  -1,   0,   1,  -2,  -2,  24,  34, -11,   0,   0,  -1,   1,   7,  37,   1,  11,  -7,   2,   8,   3,  29,  10, -11,  16,  -3,  -6,  14,  19,  13,  -4,  12,  -9,   5,  -3 },
      {   0,  -1,   0,   3,  -7, -26,  -5,  22,   1,   1,   1,   5,  -5, -12,  41,  -7,   1,   0,  -1,   0, -12,  10,  45, -12,   1,   0,   5,  21,   0, -33,  15, -10,  -1,   2,   9,  -4, -57, -31, -16,  11,  -6,  -9,  -4, -29, -30, -32, -18,  12,   5,   9,   5,   1,  -6,   3,  -6,  -7,   2,  -7,  -9, -17,  -2,  14,   4,  -3 },
      {   3,   4,   5,   7,   5,  21,  51,  40,   1,   2,   2,   4,  12,  33,  40, -19,   2,   4,   4,   6,  16,  26,  -9,  -7,   0,  -1,  -4,   3,  23,  13, -32,   8,   0,   1,   8,  16,   9,   4, -12,  -8,   2,   6,  -2,   4,  -8,   4,  -7,  12,  -2, -14,   7,   9, -34,  15,   4,  -6,  -1,   0,  24, -27,  31, -18,   7,  -4 },
      {  -1,  -1,  -1,  -5,   2,   8,  -7, -22,   1,   1,   1,   3,   6,   2,  10,  -9,  -1,  -2,  -1,  -9,   3,   2,   0,   3,   2,   3,   4,  11,  41,   7,  16,   1,   0,   2,  -3,  -9,  -1, -44,  34,  -7,  -3,  -3, -23,  11,  28, -48,   0,   2,  -7, -11,   1,  36,  10, -13, -19,   0,  13,  17,  31,  37,  36,  -8,  -4,  -4 },
      {   1,   1,   1,   2,   4,   3,  10,  38,   0,   0,   2,   0,  -9, -25, -10,  15,   0,  -1,  -1,  -1,   6,  -5, -28,   9,   2,   2,   7,   9, -12,  -6,  42, -15,  -4,  -3,  -7, -29, -37,  11, -25,   1,   2,  -2,  -4,   3,  52,  30, -31,  -3,  -2,  -2,   0,   2,   2,  30,   6,  -8,  -4,   2,   8,  14,  40,  -7,   6,  -3 },
      {   1,   2,   2,   3,   7,   5,  15,  22,  -2,  -2,  -3,  -4, -13, -21,  -7,  -5,   3,   4,   3,   6,  25,  61,  41, -30,   1,   1,   3,   1, -13,   3,  15,   2,   1,   1,   3,   1,  29,  28, -37,  12,   0,   1,   7,  -1,  15, -24,  14, -12,   0,   2, -10,   2,  20, -35,   5,   1,   2,   8,   0,  34,   3,  -1, -14,   6 },
      {  -2,  -4,  -2,  -3,  -4, -13, -47, -38,   1,   1,   0,  -1,  -4,  32,  49, -48,   2,   2,   1,   2,  10,  28,   2,  18,   1,   1,   1,  -4,  -7,   0,  -4,   9,   0,  -3,  -3,  -4,  -9,   8, -15,   2,   3,  -2,  10,  -9,  10,  43, -16,  -1,   1,   0,  -1, -18,  30,  -2,   0,   4,   2,   1,  -3,  14,  21, -25,  -5,   0 },
      {   0,   1,   1,   0,   3,   0, -20,  -9,  -1,  -1,  -1,   1,  -5,  -8,   2, -15,   1,   3,   3,   1,  10,  13, -16,  15,  -1,   0,  -2,   5,  -1, -47, -42,  28,  -3,  -5,  -7, -16,  -9,  13, -28, -14,  -2,  -7, -10,   6,  41, -16,  12, -19,  -2,   8,  16,  50, -14,  12,  11,   2,   1,   8,  10, -18, -37,  25,   4,   7 },
      {   0,   1,   1,   0,   6, -15,  -4,  48,   1,   2,   3,  -3,  -6,  11,  38,  21,   0,   1,   0,   1,  17, -18, -49,  29,  -2,  -4,  -1,  -6, -12, -35, -19, -17,  -1,   0,  -2,   2,  14, -11,   0, -22,   0,   0,   2,  -6, -22, -16,  16,  -8,  -8,  -6, -17, -16,  26, -26,   7,  -4,  -4,   6,  -3,  40,   2,  13, -10,   5 },
      {   0,  -1,  -1,   0,   1,  -4,  -8, -16,   1,   0,   3,   0,   0, -15, -44,  -1,   0,  -2,  -3,  -3,   6,   6,  -1, -12,   2,   2,   6,   6,   6, -20, -41,  20,  -2,  -5,  -8, -12, -19, -42, -25,  -4,   2,   1,   7,   4, -12,   0,  23, -17,  -3,  -7, -12, -43, -32, -11,  26,  -2,   1,  -8,  -4,   5,  52, -15,  -2,   9 },
      {   0,   0,  -2,  -2,   2,   7, -11, -38,   2,   2,   2,   3,   2,   4,  27,  70,   1,   2,  -1,  -1,   0,   3,   9,  18,   0,   0,  -2,   1,  13,  39, -16, -20,   0,   3,   4,   6,  15, -12, -53,  14,   0,  -2,  -4,  -2,  -3,  -8, -25, -10,  -2,  -2,   4,   3,  -3,  13,  23, -23,   4,   9,   4,   5,   3,  24,   0,   8 },
      {   0,  -1,  -1,   0,   1,  -6, -17, -11,   1,   0,   0,   1,  -3,   1,  31, -13,   0,  -3,  -2,   0,  -1, -15, -16,   6,   0,  -2,   0,  -2,  -6,   4,  50, -22,   0,  -1,   0,   1,   8,  17,   2,  -1,   3,   3,  10,  19,  19, -32,  16,  -9,   3,   3,   5,  -9, -56, -41,  47,  -8,  -2,  -7,  -9, -29,   2, -21,  -1,  11 },
      {   0,   0,   0,  -1,   3,  11, -13, -16,   0,   0,   0,   1,  -5, -10,  10,  14,   3,   2,   1,   3,  10,  16,  21,  42,   0,  -2,  -2,   0, -12, -23,  18,   3,   1,   1,   1,   5,  10,  38,  14, -11,   3,   5,   7,  11, -15, -21,  36,  16,   2,  -2,  -7, -21, -27,  54, -30,  16,   1,  -2,   5,   4,  37,  26,  -6,   0 },
      {   0,   1,   0,   0,   2,   0,   1,  15,  -1,   0,   1,   0,   0,  -3, -31, -32,   0,   2,   2,   0,  -4, -19,   9,  68,   1,   2,   0,  -1,   5,  19,  15,  47,   0,   1,   3,   7,  14,  -2, -39, -17,   0,   0,   2,  -1, -11, -30, -45,  16,  -2,  -4,  -2, -11,   0, -21,  -8,  -5,  -3,  -1,  -3,  -3,  -3,   2,  -1, -14 },
      {   0,   0,   1,  -1,   1,   3,   7,   7,   1,   0,   1,   2,   1, -12,  10,  53,   1,   1,   2,   1,   0,  13,  49,  38,   1,   2,   3,   6,   7, -10, -15,   3,   0,  -3,  -2,  -3, -15,   5,  27, -48,   0,   1,   3,   3,  17,  21,  -6, -17,   3,  -1,   3,   1,   0, -32,  -7,   9,  -3,  -5,  -3, -11, -13, -55,   8,  -3 },
      {   0,  -1,  -1,  -2,   0,   2, -14, -16,   1,   1,   1,   0,  -3,  -7,   2,  25,   0,  -1,   1,   0,   2,   5, -28, -26,  -1,  -1,  -2,  -7, -13, -28, -32,  -2,   0,  -3,  -3,  -5,   8,  27,   4,  29,   1,   3,  10,   6,  16, -34, -46,  40,   0,  -2,  -6, -15, -11, -23, -40,  -9,   1, -11,  -4,  -6,   6, -25,  14, -33 },
      {   0,   0,   1,   0,   2,   5,  -2, -12,  -1,  -2,  -2,  -3,  -3, -10, -14, -18,   1,   0,   1,   0,   4,  -3,  -3,  -8,  -2,  -3,  -3,  -2,  -8, -10,   1, -27,   2,   2,   4,   4,  14,  13,  -5, -28,  -2,  -4,  -7, -12, -27, -42, -10, -27,   1,  -3,   3,   1,  31,  52,  28, -21,   3,  -1,   5,   3,  -5, -57,  39,  -7 },
      {   0,   0,  -1,  -1,  -1,  -1,   4, -10,   0,   0,   0,   0,   0,   0, -10, -20,   0,   0,   0,   1,  -3,   6,  22, -15,  -1,  -1,   0,  -2,  -3,  -4, -19, -30,  -1,   0,  -4,  -1,  -1,  19,  28, -58,   0,   2,   5,   6,   9,   7, -42,  29,   0,   0,  -1, -10,  -7, -15,  31, -21,   3,   9,   1,  17,  13,  56,  23,  18 },
      {   0,   0,   1,   1,  -2,  -2,   6,   7,   0,   1,   0,   1,   0,   1,   9,  17,   1,   1,   2,   3,   3,   6,   6,   8,   0,   1,   1,  -1,  -4, -13,   4,  69,   2,   2,   3,   1,   0,  12,  55,  47,   2,  -1,   3,  -5,   0,  -4, -15,  -9,  -1,  -2,   0,  -7,  11,  10,  47, -45,  -2,   8,  -1,   6,   6,   4,   9,  19 },
      {   0,   0,   1,   0,   0,   0,   5,  10,  -1,  -1,  -2,  -2,   1,  -4, -20, -17,   1,   1,   1,   0,   3,  -1,  22,  50,  -1,  -1,  -1,  -1,   2,  -2, -31, -51,   0,   0,  -3,  -6,  -3,  -8,   4,  51,   2,   2,   0,   2,  16,  -1,  25,  59,   1,   1,   0,   9,  10,   2,  25,  -7,  -3,   2,   1,   3,  -8, -22,   1,  19 },
      {   0,   1,   1,   0,  -2,  -1,   1,   5,   0,  -1,  -1,   0,   0,  -1, -10, -15,   0,   0,   1,   0,  -3,  -1,   0,   7,  -1,  -1,  -1,  -2,  -2,  -2, -18, -25,   1,   2,   0,   0,   1,  11,  15,  13,  -1,  -1,  -3,   2,  -3, -14, -45, -45,  -3,  -3,  -3,  -3, -14,   9, -33,  -5,  -4,  -3,  -2,  -7,   2, -11, -60,  71 },
      {   0,   0,   0,  -1,   0,   2,   6,   3,   0,  -1,   0,   0,   1,   5,  -4, -23,   1,   1,   1,   0,   3,   3,  14,  24,   1,  -2,   1,   1,  -1,   0,  -9, -32,  -1,  -2,  -1,  -4,   3,  -2,  10,  24,   1,   0,   3,   8,   7,  12,  13, -60,  -1,  -3,  -1,  -7, -12, -14, -33, -62,  -1,  -3,  -3,  -2,   3,  25,  23, -52 },
      {   0,   0,   0,   0,   0,   2,  -3,  -6,   0,   0,   0,   1,   2,  -3,  -3,   2,   0,   0,   0,  -1,  -1,   3,  -1, -11,   0,   0,  -1,   1,   4,  -3,   2,   2,   0,   0,  -1,   1,  -6,   2,   0, -31,   0,   0,   1,  -2,   2,   3,  12,  27,   0,   2,   1,   5,   4,  17,  15, -60,  -4,  -3,  -5,  -3,  -9, -12, -90, -42 },
      {   0,   0,   0,   0,   0,   0,  -3,  -4,   0,   0,   0,   1,  -1,  -1,   4,   3,  -1,  -1,  -1,  -1,  -1,  -3,  -9, -10,   0,   1,   1,   2,   1,   5,  15,  17,  -1,  -1,   0,  -1,  -1,  -6, -22, -22,   1,   1,   1,  -1,   3,   8,  34,  29,  -1,  -1,  -1,  -1,   3,  -7, -50, -68,  -1,  -2,  -3,  -3,  -9, -13,  33,  63 },
    },
    {
      { -93,   7,  20,   5,   4,   2,   1,   1,  76, -16, -19,  -3,  -3,  -1,  -1,   0, -11,  22,   3,  -3,  -1,  -1,  -1,   0,  -7,  -9,   7,   2,   1,   0,   0,   0,   3,  -2,  -4,   1,   0,   0,   0,   0,  -1,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  45, -56,  12,   4,   1,   1,   1,   0,  46,  47, -31,  -7,  -4,  -2,  -2,  -1, -61,   7,  30,   3,   2,   0,   1,   0,  11, -22, -11,   5,   2,   2,   0,   0,  -5,  13,  -3,  -6,  -1,  -1,   0,   0,  -1,  -2,   5,   0,   0,   0,  -1,   0,  -2,   1,   0,   0,   0,   0,   0,   0,   0,  -1,   1,   1,   0,   0,   0,   0 },
      {  58,  12,  -2, -13,  -4,  -3,  -2,  -2,  31, -55,  -2,  17,   2,   2,   1,   2,  -7,  65, -22, -11,  -3,  -2,  -1,  -1, -37,  -8,  34,   4,   1,  -1,   1,   0,   8, -15, -17,   6,   3,   2,   0,   1,  -5,  11,   2,  -7,   0,  -1,   0,   0,   0,  -4,   1,   2,  -1,   0,   0,   0,  -3,   2,   0,   0,   0,   0,   0,   0 },
      {   8,  66, -23, -14,   1,  -6,   0,  -1,  10, -33,  24,   8,  -6,   3,  -1,   0, -73, -13,   4,  11,   9,   1,   3,   1,  53,  -4, -17, -10,  -3,   0,  -1,   0,  -8,  15,   9,   0,  -3,  -1,   0,  -1,   0,  -7,   2,   3,   2,   0,   0,   0,  -1,   0,  -2,   0,   0,   1,   0,   0,   1,   0,   0,  -1,   0,   0,   0,   0 },
      {  14, -30, -33,   5,   9,   4,   3,   1,  54,   2,  37, -12, -14,  -3,  -4,  -1,   8, -16, -39,  25,   6,   3,   0,   1,  -9,  58,  -6, -31,  -2,  -2,   0,  -1, -16, -15,  38,  12,  -5,  -1,  -1,  -1,   0,  -6, -17,   4,   5,   1,   1,   1,  -3,   4,   5,  -4,  -1,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0 },
      { -16,   8, -52,  13,  12,   1,   2,   1,  -8,  37,  52, -23, -14,  -3,  -2,  -2, -21,   9,   8,  22,   4,  -4,   0,  -1, -36, -47,  15,  -2,  13,   6,   2,   2,  39, -23, -21,   4,  -9,   2,   0,   1,  -5,  17,  -8,  -4,   4,  -3,  -1,   0,   4,  -5,   5,  -2,   0,   2,   0,   0,   0,   2,  -2,   0,  -1,  -1,   0,   0 },
      { -18, -28,  19, -18,   6,   4,   2,  -1, -23,  15,   7,  32,  -7,  -7,   0,   0, -29,  -5, -32, -17,  21,   5,   3,   1,  36,  -2,  48, -15, -23,  -1,  -2,  -1, -10, -32, -20,  42,  11,   1,   0,   1,  11,  14, -24, -27,   3,   3,   1,   0,   0,   4,  15,   3,  -8,  -3,  -1,   0,   2,  -1,  -6,   2,   4,   0,   1,   0 },
      { -11, -60, -26,   7,  24,   0,   6,   3,  -5, -42,  31,  -2, -16,   7,  -2,   1,  23,  21, -14,   5,  -1,  -9,  -1,  -1,  48, -25, -12,   8,   1,   5,  -4,   0,  -9,  46, -25,  -9,  -2,   0,   3,  -1, -13,   9,  23,  -8,   0,  -4,   0,   1,  -1,  -7,   0,   7,   2,   2,  -1,   0,  -2,   1,  -1,  -2,   1,   1,   1,   0 },
      {  -4,   1,  -4,  32,   7, -10,  -2,   1, -23,  -5,  -5, -50,   1,  18,   3,   2,  -5,  53,  27,   9, -22, -17,  -3,  -3,  28,   9,  15,   8,  17,  -1,  -4,  -1, -50, -34,  17,  10,  -2,   9,   3,   1,  22, -14, -26,   2,   3,  -2,   2,   1,   1,  10,   6,  -8,  -3,   0,  -1,   0,   4,  -5,   0,   6,   1,  -1,  -1,   0 },
      {  -3, -25,  53,  29, -10, -17,  -2,   0, -24, -39,  -1, -29,  19,  18,  -2,   1, -37, -16, -31,  31, -13,   2,   6,   2,  -7,   2,  -8, -22,  15,  -4,  -1,   1,  47,  -2,   5,  -2, -19,   2,  -3,  -1,  -9,  17, -11,   9,   8,  -4,   1,  -1,   3,  -1,   3,  -9,   0,   4,  -1,   1,  -1,   5,  -2,   3,  -3,  -1,   0,   0 },
      { -18, -14,  12, -28,   4,   2,   7,   2, -29,  17,  10,  27, -11,  -2,  -5,   0, -13,  53, -12,   1,  11,   0,   0,  -2,  -9, -11,   1, -28,  -1,   4,   4,   3,   8,  15,  45,   0, -15,  -7,  -4,  -3, -12, -52,   5,  37,  10,   4,   1,   1,  13,   4, -34, -11,   4,   2,   1,   1,  -1,   5,   9,  -5,  -3,  -1,  -1,   0 },
      {  18,  39,  12,  29,  -4, -14,  -7,  -4,  17,  30, -22, -24, -10,  14,   1,   1,  33,   0, -29,  11,  14,  -9,   4,   0,  18, -29,  30, -30,  -9,   7,  -5,   1,  -5,  45,   1,  31, -15,  -7,   0,  -2, -29,  -7,   0, -10,  22,   3,   3,   1,   6,  -3,  10,  -4, -14,   3,  -1,   0,  -5,  -2,   1,   8,   6,  -4,   1,  -1 },
      {  14,  -2,  42, -44,  12,  -2,   4,  -2,  17, -21,  26,   8, -15,   0,  -4,   1,  31, -23,  39,  27,   9, -10,  -4,  -2,  -4, -39, -19,  -6,  11,  20,   2,   3, -19, -13,  -7,  19, -21,  -3,   1,   2,  11,   4, -38,   6,  17,  -2,  -3,  -1,  -1,  18,  -2, -20,  -2,   4,   2,   0,   2,   0,   5,   2,  -2,  -3,  -1,   1 },
      {   8,  -4,  21,  23, -15, -12,   2,  -1,  19,  -4,  50, -21,  19,   6,  -6,  -3,  -2, -31, -14, -24,  -2,   9,   7,   1,  -4, -24,  34,  37, -18,  -8,  -5,   2, -13,   3,  -8,  -5,  20,   6,   1,  -2,   9, -46,  -4,  14, -15,   4,   2,   3,   7,  17, -39,   2,   9,  -5,  -1,   0,  -3,  10,  10, -15,   0,   1,   0,   0 },
      {   1, -14, -47,  25, -18,  14,   3,   3,   2, -24, -29,  19,  18, -13,   0,  -1,   8,  -7,  19, -18,  -2,   8,  -1,   0,  17, -12,  -4,   6,  -9,  -7,   2,   0,  36,   1,  10,  22,   7,   4,  -6,   0, -27, -11, -53,   7,  13,   8,   6,   3,   7,  32,  -1, -38,  -3,  -7,   0,  -2,  -1,  -2,  20,   4,  -6,   2,  -1,   0 },
      {  -8,  -7, -17, -16, -76,  -5,  14,   7,   7,  10,  10,  10,  63,  -3, -18,  -6,  11,  15,  -6,  26, -10,  16,  12,  -2,   9, -27,  -6, -28,   1,  -8,   2,   7, -27,   0,   1,  -3,  -9,   2,  -5,  -3,  16,  15,   5,  -2,  -4,  -4,   2,  -1,  -4,  -4,   9,   8,   4,   2,  -3,   2,   3,   1,  -5,  -1,   3,   0,   1,   0 },
      {   6,  -1,   2, -31,  -3, -11,   7,   3,  16,  11,   6, -10,  11,   5,  -7,  -2,  18,  -4,  30,  -6, -13,  -6,   1,   0,  48,  28,  44,   0,  16,  -1,  -7,  -2,  31, -24,  -3,   2, -18,  18,   4,   3, -46,  14,  11,  34,  -4, -11,  -2,   0,   0, -25, -15,   3,  21,  10,  -1,   0,  -1,   8,  -6, -11,  -8,  -1,   2,   1 },
      {  -3,  12,  20, -23,   4,  -2,   2,  -3,   2,  30,  26,  11,  10,   0,  -3,  -2,  -6,  15, -24,   6, -42,   0,   3,   0,   1,  17, -24,  30,  32, -18,  -3,  -1,  -9,  13, -16,  -4,  20,  19,   2,   1, -44,  -8, -15, -32,   3,  -7,   9,   2,  27,   8,  19, -29, -13,  -4,  -6,   1,  -7,  -2,  17,  18,  -8,   1,  -2,  -1 },
      {   0,   5, -20,  15,   8, -27,   8,   1,   3,  11, -39,  43, -21,  35, -12,  -3,   5, -12, -34,  31,   4, -18,  17,   1,   6,  -7,   1,  -2,  25,   1,  -9,   2, -13, -27, -29, -29,  -3,  12,   6,   0,  11,  -3,  -3,  10,  -9, -11,  -2,  -1,   7,  23, -23, -12,  23,   4,  -3,  -2,  -1,   7,  15, -22,  -7,   2,   3,   1 },
      {  -8, -14, -17, -16,  -7,  -7,   7,   2, -12, -25, -21, -11,   4,   7,  -5,  -3, -11, -18,  36,  41,   9,  -2,   3,  -2, -26,  20,  15, -17, -10,   6,  -2,   1, -15,  -2, -29,  18,  16,  -6,   1,  -2, -30, -40,  17, -31,   3,   7,   4,   2,  43,  -8, -13,  14, -17,  -5,  -7,   1, -10,  18,  -5,  -6,  11,   1,   0,  -2 },
      {  -4, -12, -20, -29,  17, -56,   2,   6,   4,  -4,  -1,   5,   4,  58,  -7,  -8,   1, -16,  12, -40, -18,  -9,  24,   5, -15, -17,   0,  -8,  -7, -18, -16,   5,  13,   5,  36,   4,  -3,   8,  -2,  -8,  14,  -4,   2, -10,  -4,  -1,   3,   3,   8,   2,  25,  10, -15,   0,  -4,   1,  -1,  -9,  -7,  24,   3,  -1,   1,   0 },
      { -11,  -8, -22, -13,   4,  15,  -3,   1,  -9,  -7,  -9,  -4,  22,   0,  10,  -3, -26, -27, -10, -27, -18, -15,  -7,   5, -31,   8,  30,   0,  19,  20,  -7,  -6, -39,  32, -16,   7, -34,  -3,  15,   5,   5,   2,   2,  13,  30, -18,  -9,  -2, -14, -16,  -3, -33,   2,  19,   6,   0,  19,   2,  13,   2, -20, -11,  -3,   1 },
      {   6,  -2,   9,  50,   4,  20,  -9,   3,   7,  -9,  11,  44,   1, -16,  12,  -2,   2, -18,   6, -24, -14,  -5,  -6,   2,   2, -29,  -8, -19,  28,   0,  -2,  -3, -20, -35,  20,  -4, -20,  12,   5,   3, -20, -17,  14,  -4,  -1, -11,  -2,   2,  34, -32,   7,  15, -12,   7,  -5,  -1,  -2,   8, -27,  14,   2,   0,   4,  -2 },
      {   5,   0,  -9, -35,  14,  32,   6,  -5,   3,  -9, -23, -28,   1, -35,   4,   5,   7, -16, -20,  10, -16,   1,  -6,  -3,  17, -19,  33, -15,  28,  -4,   2,  -3,  22,   1,   5, -33, -12,  12,   2,   5,  30, -32,   3,  -9, -27,  -6,  -1,   3,  12,  32,  18,  14, -14,  -5,  -8,  -4,  -1,   0,   5,  23,   1,   2,   3,  -2 },
      {  -1,  -2,   5,   8, -12, -19, -10,   3,   2,   4,  19,  -7,  21,   2,  12,  -8,  -1,  11,   8, -32,  40, -12,  -7,   6,   6,  19,  -9, -22, -10,  37,   4,  -1,   0, -18, -21, -32, -20, -19,  11,   8, -15,   3,  19,  -2,  -6,  -6, -20,  -4,  24,  46,  18,  -5,  -9,   4,   5,  -4, -24, -14,  33,  18,  -2,  -3,   0,   4 },
      {  -7,  -8,   5,   3, -34,  -2,  -3,   0, -10, -11,   5,  15, -27,   3,  -3,   1,  -8,  -7,  11,  24,  39,  -7,  -1,  -1, -16,   1,  36,  31,   3,   0,  -4,  -6, -15,  21,  22, -32,  -2,  15,   8,   1, -29,  -8, -22,  10, -15, -10,   5,   4, -14,  -5,  48,   4,  18,  -3,  -8,  -5,   0, -34, -12,  13,   4,   9,   4,   2 },
      {   0,  -4,  -1,   4, -30, -14, -25,   2,  10,   8,  13,  25,   6,  12,  20,   1,   8,   9,   5,  14,   5, -30, -13,   5,  16,  20,   1,  18,  22,  20, -10, -11,  42,   7,  -7,   0,   4,  -9,  20,   7,  31, -39,  -8, -25,  18,   6, -13,   2, -26,  -9,  -4,   8, -29,   9,   6,  -2,  27,  18, -16,   6,   2, -13,  -1,   0 },
      {   0,  -5,  -9, -23,  -5, -14, -19,  -5,   3,  -3, -16, -27,  -2,   5,  19,   3,   7,  -2, -23,   3,  38,  21,  -6,   3,  13, -22,  -8,  22, -16,  -5,  11,  -3,  -8, -29,  14, -34,   8,  -9,  -8,   4,  -4,   4,  -6,  -3,   5,   6,   1,  -3,  25, -49,  -2, -33, -12,   3,   1,   6,  23,  30,   5,  17, -29, -10,  -7,  -2 },
      {   6,   5,  12,  24,  32, -13,  39,   0,   1,  -4,   8,   5,  18, -22, -41,  -3,   6,   5,  21,  -7,  30,  13,  27,  -9,   0,   9,   9, -25,  15,  -4,  -2,  11,   5,  -1, -10, -19,  12,   7,  -8,  -8,   4, -22,  12, -21,   6,   7,   9,   7, -33, -17,  16, -27,  -8,   3,  -6,   2,  38,  -4,  14,   2, -22,   2,  -1,   0 },
      {   2,   3,   3,   2, -44,  26,  52,   9,  -1,  -2,  -1, -16, -48,  11, -55,   3,  -3,  -1,  -7, -30,   1, -18,   8, -11,   4,  12,  -9,  20,   3,  13,   2,   8,   9,  -2,  -1,   2, -19, -14,  -2,  -4,  13,   8,  -4,  -5,   9,  -8,   0,  -1,  30,  -5,  -1,   4, -15,   4,  -1,  -2,   6,  19,  -2,   8,   0,  -9,   1,  -1 },
      {  -1,   0,  -4, -16,  13,   3,  15,  -8,  -3,   0, -11, -22,  36,  -4,  -5,   0,  -3,   1, -19, -25,  38, -33,   1,  -5,  -4, -12, -21,   4,  33,  24,  -9,  -5,  -3,   1,  27,   1,   8,  12,  13,   5, -19,   5, -33, -32, -11,   5,  -2,   5, -11, -13, -11,  20,  10,  -4,  -2,  -4, -18,   9,  -6, -46,  18,   9,   4,   5 },
      {   1,   1,  -7,  -1,  26,  -8,  29,  -3,   1,   0, -11,   8,  28,   0, -25,  -1,   0,  -7, -18,  19, -19,  10,  -4,  -4,   8,  -9,   6,  39,  -9,  21,  21,   1,   0, -23,   6, -11,   0, -43,  -1,   6, -10, -19,   3,  13,  34,  17, -21,  -7, -12,  -8,  13,   5, -11,  15,  16,   1, -35, -32, -23,   9,  15,  -7,   1,   5 },
      {   0,   3,  -6,  -5,  11, -18, -11,  -6,   0,   5,  -6, -10,   4,  -2,  -5,   5,   1,  12,  -9, -18,   9,  53,  10,   1,  -2,  12, -25,   8,  -2,   2,   6,   3,  -1,   5, -38,   5, -45,   4,   2,   2,  -8, -29, -30,  10,  -1, -24,   2,  -3,   8,  14,  14,  22,  13,   2,  -5,  -4,  35,   7, -48, -11,  13,   3,   7,   1 },
      {  -2,  -1,  -2,  11,   7, -23, -11, -14,  -1,   2,   4,  14,   0, -24,  -5,   3,   1,   6,  17,   2,  19,  29,  15,  -6,   1,  19,  19,  12,  21, -17,  -3,   6,   0,  28,   3, -11, -22,  13,  -8,  -1,  37,  16, -21,  -5,  17,  -3,  -1,  -3,  16, -15, -11,  -1, -10,  11,  -1,   5, -72,  29,   6,   5,   5,  -6,   4,  -3 },
      {   5,  10,   7,  10,  32,   4,  22,  28,   3,   9,   3,  15,  31,  -3,  -3, -23,   0,  10,   5,  19, -11, -30, -20,   5,  -2,   8,   8,  -2, -51,  20,  -8,  -5,   7,  24,   2, -20,   7,   4,   7,  -1,  18,  12, -32,  15, -14, -12,   0,   0,  35, -10,  19,   1,  12,  -3,  -8,  -2,  11,  35, -14,   3,  -6,   6,   4,  -1 },
      {  -4,  -1,  -2,   5,   5,  -9,   1, -14,  -2,   3,   5,  15,   0,  17,  -8,  13,  -5,  -5,   5,  -5, -15,   3, -16, -11,  -3,   4,  18, -10,  13,  14,  33,   4, -17,  10,  -1, -27, -15,  -8, -25,  10, -25,  21, -20,   8, -24,  40,  13, -12, -23,  10, -10,  14, -33, -38,  14,   8,  19,  38,   4,  10,  23,  -8, -18,  -6 },
      {  -2,   1,  -1,  -1,   0, -29,   4,  46,  -7,  -4, -11,  -4, -14, -41,  -3, -52,  -3,   0,  -9,  10,   1,  18,   2,  18,  -3,  -5,  -1,  22,   2,  14, -19,   9, -11,  -6,  18,   8,  -5,  17,  22, -14, -20,  22,  11,   0,  -3,   0, -10,  18, -20,  22, -27,   8, -33,  -2,  -2,  -9,  12,  12,  -2,  16,   8, -11,   5,  -3 },
      {  -2,   2,   0,  -2,   0,  28,  15, -21,  -3,   2,  -2,  -5,   3,  21,  -1,  23,  -1,   4,  -2,   2,   4,  -4,   4,   4,  -5,  -5, -18, -11, -39, -12, -27,  -7,  -5, -14, -17, -10,  -4,  29,  22,  -2, -19, -13,  -5,  11, -10, -18,   0,   8, -52,   2,  -6,  -4, -29,   8,  -4,  -6, -38,  28, -15,  26, -12, -21,   9,  -5 },
      {  -3,  -3,  -3,   6,  -6, -50,  41, -62,  -2,  -1,  -1,   5,  -6, -37,   8,  41,  -2,  -1,   3,   5, -10,  -4, -21,  13,  -6,  -5,   1,   3,  -6,  -4,  -8, -20,  -4,   5,  18,  13,  11,  -3,  16,  10,  -5,  15,   9,  -1,  -4,  -8, -24,  -5,  14,  18,  -1,   8,   6,  14,  -1,  -8,  22,  12,   0,   2, -12,  -2,   3,   4 },
      {   1,   1,   1,   7,   0,  -8,   3,  15,  -1,  -1,  -2,   6,  -9,  23, -16,   0,   1,   0,   5,   4, -18,  47, -41,  -9,   2,  -5,   1,  -3,  -6,  28,  26, -10,   7,   0,  15,  16, -14,   0,  24,  13,   2, -14,  -5, -38, -27, -31,  -6,   2, -19, -22,  -3,  10,  18,  -7,  -1,  -7, -16,   3,  43,   0, -26,   7,   1,   4 },
      {   0,   1,   1,   5,  19,  35, -29, -26,   1,   2,   2,  12,  24,  13, -20,  24,   1,   5,   5,  23,   5,  12,  44,  -3,   2,   3,   7,  29, -14,  15, -34,   5,   3,  -3,  23,   5, -25,  -8,  16,  -7,  -3,  18,   4, -10,  -4,   0,  -4,  10,  20,   1, -12,  23,  -6,   0,   0, -10,  32,  -7,  39,  10,  11,  -7,   2,   2 },
      {  -1,   0,   5,  -2,  -6,  13, -20,  -1,  -3,  -3,   4,  -4, -23,  -2,  -4, -21,   1,   2,  11,  -6, -30,  13,  38,  23,   1,  -8,  12, -18,  -4,  -9,  -7, -14,  -6,  -1,  13, -30,  15, -44,   2,   5, -19,  -1, -12, -18,  12,   2, -26,  -2, -15,  16,  27,   6,  14,  19,   6,  -9,  -5,  44,   0, -38, -15,  -3,   5,   9 },
      {   2,   2,   4,   0,   8,  21,  20,  12,   2,   2,   5,   1,   7,  32,   0,  -4,   2,   1,   9,   2,  19,  21, -43,   5,   0,  -1,  11,  -3,  -3, -54,   7, -18,   0,   1,  10, -25,  -5,  25,  18,   7,   0,  13,   7, -15,  46,  -8, -18,   1,  11,  21,  -7,   3, -18,   5,  14,  -1,  19,  -6,   5, -28,  31,  -2,  -4,   7 },
      {   0,   2,   0,   3,   4,  18, -12,  -9,   0,   1,   2,   1,   7,  24, -25,  -8,   2,   3,   7,   2,  11,  21,  -2,  11,  -1,   3,  10,  -8,  19,  19,   4, -16,   0,  18,   9,   4,  22,   6,   3,  10,   3,  21,  -1,  -8, -29,  13,  -5,  -4,  21,  24, -28, -16, -13,  15,  20,  -3,  -2, -25, -59,   3, -53, -15,   3,   4 },
      {  -1,   0,  -1,   2,  11,   5, -14,  18,  -1,   0,   0,   5,  19,   6,  -6,  12,  -3,  -4,  -1,   9,  24, -21, -14, -33,  -3,  -1,   3,  28,   8, -21,  24,  25,  -3,  -3,  20,  25,  -5,  -9, -33,  -3, -12,   2,  28, -10,   5, -24,   2,  -9, -13,  43,  13,  12,  15,  25, -12,   6,   3,  42,  -8,   8, -29,  -7,  13,  -8 },
      {   0,  -1,   0,   0,  -5,   2,   1, -25,   1,   1,   1,   4,   0,   4,  -1,  -4,   0,   0,  -1,   4,  -7,  11, -20,  21,   4,  -1,   4,  19,  -9,  36, -22,  -8,   4, -11,  14,  -1, -26,  16, -12, -29,   8,  -2,  35,  -8,  20,   9,  42,  10,   0,  18,  29, -21, -11, -30,  -9,  19, -12,  12, -13, -46, -17,  16, -10,  -9 },
      {  -1,  -2,  -2,  -4,  -3, -12, -37,  -3,   0,   0,   1,  -2,  -6, -11, -24,  28,  -1,  -3,   3,  -7, -18,  -5,   3, -37,   2,  -2,   9, -14, -16,  25,  20,  14,   2,   0,  11, -25,  28,   2,  15,  -6,   0,  13,   4,  -8,  13, -48,   1,  12,  -2,   5, -18, -38, -17,   2, -20,  -8,   4,  -5, -24, -11,  36,  22,   8,  -9 },
      {   0,  -1,  -1,   2,   0,   3,  -3, -36,  -1,  -1,  -2,   0,   1,  -5,  -7, -39,   1,   2,   1,   6,  -1, -28,  22,  47,   2,   3,  -1,   7, -22,   3,  47,  -9,   3,  -1,   4,  -1, -13,  29, -47,  12,  -1,  -6,  -3, -18,   2, -43,  -4, -14,  -8,  -6, -19,   9, -14,   8,  11,   4,   5, -14,  10,   2,   4,   9,  -2,  -3 },
      {   2,   2,   3,   4,   2,  17,  46, -20,   3,   2,   5,   4,   4,  25,  56, -26,   2,   2,   7,   2,   8,  19,  22,  -1,   4,   4,  10,   6,  15,   5,   7,   6,   4,   3,  11,   0,  -1, -38,   8,  -1,   2,   9,   7, -19,  -7, -22,  10,  12,  -6,   9, -21, -25,  23, -13, -30,   6,   0,   8, -21,  27,  24,   5,  -3,  -7 },
      {   0,   0,   1,   0,  -6,  -7,  14,   4,   1,   2,   2,   1,  -7, -10,  36,   4,  -1,   0,   2,   1, -19, -12,   6, -37,   0,   4,   5,  -3, -27, -10,  -9,  38,  -1,   0,   1, -30, -34,  -2,  -5,  -7, -10,   3, -23, -36,  12,  12,  -5,  -6,  -1,   1, -29,  20, -12,  -4,  -2,   6,   2, -35, -10, -19, -50,   9,  11,  -1 },
      {   0,   0,   1,   2,  -2,   7,  11, -14,   0,  -1,  -1,   2,  -4,   6,  23, -22,   0,   1,  -1,   7,   2,   7,  27, -38,   0,   0,  -2,   9,  -2,   7,  11,  37,   3,   4,   6,  23,   6,  28,  22,  26,   5,  -5,  17,  22, -22,  -5, -42, -23,   7,  -4,  33, -20, -29, -14,  29,   5,   0,  12,   2, -31,   0,   2, -11,   8 },
      {   0,   0,  -1,   0,  -1,   9,   9, -20,   1,   1,  -1,   2,  -2,  20,  18,  -5,   0,   0,   0,  -3,  -1,  21, -26,  14,   1,   2,   3,  -8,  18,  22,   0,  48,  -1,   4,  -5,  -4,  34,   0, -19, -47,  -2,   3, -20,  24,  -4, -10,   1,  21,   5,  -7,  10,  25, -27,  51,  -6,   0,   7,  -4,  21, -11,   2,   0,  18,   1 },
      {   0,   0,  -1,  -1,   3,   2,  -7,  -8,   1,   2,   1,   0,   9,   7, -10,  -2,   0,   0,  -1,   0,  11, -12,   5,  19,   0,   1,  -1,   4,  15, -25,  33,  -2,  -4,   3, -10,  11,   0, -23,  25,  -4,   4,   6,  -2,  30,  11,  -2, -21,  34,   1,  -6,   1,  15, -31, -49, -34, -22,  -5,   2,   1,  -3, -38,  46,  22,   2 },
      {   1,   2,   1,   0,   0,   1,  21,  49,   1,   3,   2,   2,   1,   6,  23,  49,   0,   2,   1,   5,   1,  13,  14,  23,  -1,   2,   3,   9,   3,  12, -21, -20,  -2,   2,   3,   6,  -8,  -3, -49,  30,   1,   2,   4,   0, -20,  -4, -13,  -8,  -3,  -3,  -2, -17, -34,  16,  -9, -20,  -2,  -1,   1, -16,   8,  40,  16,   7 },
      {   0,   0,   0,  -1,  -3,  11,  -3,   2,   0,   0,   0,  -4,  -3,   9,  -4,  18,   0,   0,  -1,  -3,  -3,   9,  -2,  29,  -1,   1,   1,  -5,   2,  27,  -1,  38,   0,   1,   2,  -6,  11,  42, -12,  -9,   1,  -3,   6, -14,  35,  28, -52, -19,  -3,   2,  -5,  -2,  43, -29, -10,  -6,   0,  15, -10,  33,   3,  14,  20,  14 },
      {   1,   0,   1,   1,   4,  -2,   4,  23,   1,   1,   1,   2,   4,  -7,   8,  40,   1,   1,   3,   4,   1, -15,   9,  38,   2,   2,   5,   4,  -2, -16,  35,  18,   1,   2,   2,   4, -20,  -9,  13, -61,   0,  -4,   0, -17, -32,  -9, -18,  17,   4,  -5,   4, -24,   5, -12,  29,  -7,  -6,  -4, -10, -18,   7, -36, -20,   6 },
      {  -1,  -1,  -1,  -1,  -1,   0, -13, -28,  -1,   0,  -1,   0,   1,  -1, -21, -30,  -1,   0,  -1,   1,   1,  -9, -38, -11,   0,   1,  -1,   2,  -3, -26, -37,  17,  -1,  -3,  -6,   3, -25, -24, -19, -14,  -1,  -6,   4, -14, -30,   6, -15,  11,  -6,  -7,  -7, -31,   5,   6,  35, -28,   0,   8,  -9,  13,  17,  46,   8,  29 },
      {   0,  -1,   1,   0,  -2,  -2,   0, -11,   0,   1,   2,   1,  -4,   3,   8, -11,  -1,   0,   0,   0,  -9,  -5,   5, -12,  -1,   3,  -2,   5, -16,   3,  40,  -7,   1,   0,   0,   2, -23,  26,   2,  16,   6,   1,  12,  -8, -10,  54,  12,  37,   3,  -1,   8, -19,   8,  34, -34, -45,   2,   1,   1, -13,  13, -27,  38,  18 },
      {   0,   0,   0,   0,   2,   3, -10,  -4,  -1,  -1,   0,   0,   0,   0, -22,  -2,   0,   1,   1,   0,   1,  -5, -20,  40,   0,   0,   0,   2,  -4, -20,  -7,  40,   2,   1,   2,   3, -15, -16,  26,  25,  -1,  -2,  -1,  -7, -27,  19, -10, -38,  -3,  -4,   0, -20,  -6,  19, -48,  29,   1, -10,   1, -15,  13,  -1,   6, -51 },
      {   0,  -1,   1,   0,   2,  -3,  -2,  -8,   0,   0,   1,   1,   2,   0,  -3, -13,   0,   0,   1,   0,   1,   3, -19,  -9,   0,   0,   1,  -2,   7,  -2, -30,   1,   1,   0,   0,   0,   7, -14, -44,   7,   2,  -2,  -1,   4,   1, -27, -22,  -5,   3, -10,   4,  -5,  -3, -50, -18, -44,  -1, -16,   1, -19,   1, -69,  23, -23 },
      {   0,   0,   0,   1,   2,  -2,   1,   3,   0,  -1,   0,   1,   2,  -5,   0,   9,   1,   0,   2,   1,   2,  -3,  -3,  33,   1,  -1,   3,  -1,   2,  -3,  -2,  52,   1,  -1,   5,  -4,   9, -11,  21,  39,   2,   2,   3,   2,  12, -17,  63, -27,   0,   2,  -1,   5,  -4, -11,   9, -36,  -3,   2,  -3,  -2,  -4, -12,   0,  58 },
      {   0,   0,   0,  -1,   1,   0,   3,   0,   1,  -1,   1,   0,   1,   0,   5,  -5,   1,  -1,   1,   0,   2,   0,   6,  -8,   1,   0,   2,   1,   4,  -2,  10, -32,   0,  -1,   1,   1,   0,  -5,  19, -46,  -1,   1,   2,   2,  -7,   1,  17, -69,  -1,  -3,  -1,  -8, -10,  -9,  12,   9,   4,   1,   8,   1,   8,   1,  82,  12 },
      {   0,   0,   0,   1,   1,   1,  -8,  -2,   0,   0,  -1,   1,   0,  -2,  -8,  -5,   0,   0,   0,   1,  -1,  -2,  -7,  -9,   0,   0,  -1,   1,  -2,  -3,  -6, -15,   0,   1,  -1,   2,  -3,  -3, -12, -15,  -1,   0,  -3,   2, -12,  -5, -24, -13,  -4,  -1,  -5,  -3, -14,  -1, -63,  42,  -1,  -6,   0,  -5,   2, -19, -27,  85 },
      {   0,   0,   0,   0,   1,   3,   1,  -2,   1,   0,   1,   0,   2,   3,   4,  -8,   1,   0,   0,   0,   2,   2,   2, -11,   1,   0,   0,   1,   1,   1,   6, -13,   0,  -1,  -1,   1,   1,   7,   6, -38,   0,   0,   0,   1,   5,   9,   0, -58,  -3,   1,  -3,   6,  -5,  17, -29, -74,  -3,   1,  -1,   2,  -7,  18, -61, -17 },
    },
  },
  { // 25
    {
      { 107, -17, -12,  -5,  -3,  -2,  -1,  -1, -59,  23,   4,   2,   1,   1,   0,   0, -13, -12,   5,   1,   1,   0,   1,   0,  10,   1,  -4,   0,   0,   0,   0,   0,  -4,   5,   2,  -1,   0,   0,   0,   0,  -1,  -1,   2,   1,   0,   0,   0,   0,  -1,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -35,  21,   0,   3,   0,   0,   0,   0, -82,  -8,  17,   3,   5,   1,   2,   0,  73, -32, -14,   0,  -3,   0,  -1,   0,   0,  27,  -6,  -2,  -1,  -1,  -1,   0,  -4,  -7,   8,  -2,   1,   0,   0,   0,   7,  -2,  -1,   2,   0,   0,   0,   0,   1,  -1,  -2,   0,   1,   0,   0,   0,   2,   1,  -1,  -1,   0,   0,   0,   0 },
      {   0, -73,  44,  -7,  10,  -2,   3,   0,   4,  43, -45,  14,  -8,   3,  -2,   1,  49,  19,   0, -13,   1,  -4,   0,  -1, -37,   3,  17,   0,   0,   1,   0,   0,  -4, -11,  -1,   4,   0,   0,   0,   0,   2,   2,  -2,  -1,   1,  -1,   0,   0,  -1,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  42,  29, -32,  11,  -6,   3,  -4,   0,  26, -37,  12,  -8,   3,   0,   1,   0,  36,  40, -12,   0,  -6,  -3,  -2,  -2, -70,  29,  17,   1,   5,  -1,   1,  -1,   6, -31,   8,  -1,   2,   1,  -1,   1,   0,   8,  -4,   2,   2,  -2,   0,  -1,  -5,   0,   3,   0,   0,   1,   1,   0,  -1,  -1,  -1,   0,   0,   0,   0,   1 },
      {  17, -19,  -3,  -4,  -3,  -1,   1,   0,  48,  -8,   0,  -5,   0,  -2,  -1,  -1,  22, -24, -22,   6,  -2,   5,  -1,   3,  33,  64, -23,  -9,  -4,  -3,  -1,  -1, -60,  24,  28,  -1,   3,  -1,   2,  -1,  -2, -23,  -3,   6,  -4,   4,   0,   1,  -1,   8,  -6,  -5,   2,  -3,  -1,   0,  -2,   1,   6,  -1,  -1,   0,   0,  -1 },
      { -12,  25, -22,   3,  -3,   2,  -1,   0,  14,  74,  -5,  -7,  -3,  -4,  -2,  -1, -24, -54,  28,   6,   6,   1,   2,   1, -43,  27, -13,  -4,   5,  -1,   0,   0,   2, -30,  21,  -8,   5,   0,   1,   0,  15,   3, -14,   2,  -4,   1,   0,  -1, -11,   6,   8,   2,   0,  -1,   1,   0,   5,  -5,  -4,   5,  -1,   1,   0,   1 },
      { -12,  20,  10,  -8,  -1,  -2,   0,   0, -26,  16,  -2,   6,  -2,   2,   1,   0, -33,  50, -19,  -3,   2,  -1,   0,   0,  16,   2,  38, -12,   3,  -2,   0,   0, -50, -47,  25,  15,   1,   4,   3,   1,  31, -37, -24,   5,  -2,   2,  -1,   2,   9,  13,  -7,  -6,  -1,  -1,  -1,  -1,   2,  -1,   4,  -1,  -1,   0,   0,   0 },
      { -14,  12, -31,  27,  -5,   5,  -3,   2, -14,  37, -15, -17,   4,  -1,   1,   0,   4,  44,  14,  -7,  -3,  -6,  -2,  -4, -10,  -2,  -4,   3,   2,  -2,   2,  -1,  -7,  47,  32, -21,   2,  -3,  -1,  -1, -32, -36,  43,  17,   5,  -2,   1,   1,  12, -27, -20,  12,   0,   4,   1,   1,   0,   4,  -7,  -2,   2,  -1,   0,   0 },
      {   6, -15,   8,   9,   0,   5,  -1,   1,   6, -19,  22, -14,   0,  -3,   2,  -1,  11, -21,  37,  13,  -6,  -2,  -1,   0, -29, -35, -19,  31,  -4,   4,  -1,   2,  -3,   0,  -4,   5,   1,  -1,  -1,   1,  34, -64,  -7,  14,   5,   1,   0,   2,  22,  24, -39, -16,   4,   2,  -1,   0,  -6,   6,  23, -17,  -1,   0,  -1,   0 },
      { -19, -73, -35,  50,  -6,  13,  -1,   5,  -9,   0,  41, -40,  17,  -7,   3,  -2, -10,  14,   2,   7,  -7,  -1,  -2,  -2,  19,  10,   0,  -2,  -1,  -3,   1,  -2,  17, -25,  16,   1,   0,   4,  -2,   1,   3,  14, -19,   6,   2,  -3,   1,  -1,  -7,   2,  13,  -4,   0,   1,   0,   0,   4,  -6,  -7,   4,  -2,   1,   0,   1 },
      {  -8, -18,  29, -13,  11,  -2,  -1,  -2, -20, -26,  -4,  24,  -7,   2,   0,   1, -47,   1, -22,  17,   0,   0,   0,   0, -23,  47, -26, -15,  11,  -4,  -1,  -1,  51, -12,  29, -10,  -3,  -1,  -1,   1, -27, -17,   3,  16,   2,   3,  -1,   0,  12,  -6, -27,  -7,   7,   1,   1,  -1,   3,  11,   9, -11,  -3,   0,  -1,  -1 },
      {  -7,   5, -15,  27,  -6,   9,  -5,   3, -23, -23, -63,   9,   8,   1,   4,   2, -15,  25,  20, -17,   0,  -4,  -2,  -3,   1,  24, -39,  13,  -8,   3,  -1,   0,  -1,  16,  -9, -17,  10,  -7,   3,  -1,  28,  11, -11, -14,   4,  -2,  -2,   0, -16,  49,  15, -17,  -6,  -2,   1,  -3,  -7, -14,  24,   6,   0,  -1,   0,   0 },
      {  -9,  -7,  -6,  -7,  -4,   1,   1,  -1,  -5,  36,  39,   3,   1,  -8,   1,  -1, -10,  22, -65,  31,  -2,   7,   1,   4, -32,  -8, -32,  -4,  10,   2,   1,   2, -15,  30, -35, -14,   4,  -6,   3,  -1,  25,  12,  14, -24,  -5,   3,  -3,   1,   4,  13,   4, -10,   3,  -1,  -2,   1,  -6,  -1,  23,  -8,  -3,   1,   1,  -2 },
      {  -1, -12,  31,  34, -13,   9,  -6,   2,  -8, -27,   1,   5,  16,  -1,   3,   0, -32, -21,  -4, -20,   7,   1,   2,  -1, -25,  15,   3,  21, -11,   4,   0,   0, -37, -14, -17, -11,  14,  -6,   3,   1,  31,  -2,  44, -18,   0,  -3,  -4,   0,  -6, -45,   2,  29,   0,   9,   1,   2,  16,  -5, -27,  -6,   0,  -1,  -1,   0 },
      {  -3,   0, -25,  23, -29,  16,  -5,   0,   2,  14, -20,   5,  19, -15,   7,  -2,  -6, -23, -26, -26,   6,   6,   1,   4,   1,  19,  37,   6, -16,   4,  -6,   1,  10,  -5, -42,  32,  12,  -6,   5,   1, -36, -11, -10, -21,   4,  10,   0,   1,  28,   2, -30, -16,   1,  -3,   0,   1,  -6,  26,  26, -12,  -5,  -2,  -1,  -2 },
      {  16,  24,  54,  57, -24,   2,  -5,   0,   7,   4, -13, -20,  27,  -2,   1,   0,   4, -20, -42,  -5,   5,   4,   0,  -1,  -8, -36,   1, -27,   1,   1,   5,   0,   8,  13,  34, -12, -11,   1,  -4,  -1,  -4,   6, -19,  21,   0,  -4,   0,   0,   0,  18,   5,  -8,  -1,   0,   0,  -1,  -5,  -2,   0,   6,  -1,   0,   1,   1 },
      {  -4,   8,  48, -18,   1,  -9,   4,  -2, -17,  -1,  19, -54,  21,  -5,   3,  -1, -18,  11,  34,   4, -26,   5,  -4,   3, -23,  28,  11,  17,   5,   2,  -2,  -1, -25,  20, -16,   3,  11,   3,   4,   0, -42,  10, -21, -10,  -5,  10,   1,   0,  12,   3,  17, -11, -11,  -7,  -1,  -1, -12,  -7,   8,  22,  -2,   0,   1,  -1 },
      {  12,  27,  36,  30,  -3,   5,  -9,  -1,  14,  29,   5,  -1,  -7,  -4,   4,   1,  20,  26,  12,  25, -14, -10,  -4,  -4,  43,  15, -28,  33,  17,  -2,  -1,   1,   5, -43, -26, -16,   8,   0,   1,   5, -11,  17,   6,   2,   9,  -6,  -3,  -3, -12,  -7, -24,  -7,  14,   7,   3,   1,  17,  20,   8, -21,  -5,  -1,  -1,   0 },
      {   6,  11,   8,  14,  27,  -9,  -1,   0,   4,   8, -18, -31, -14,  12,  -1,   0,  -2,  -9, -19,  31, -29, -10,   0,  -4,  14,  29,  48,  10,   0, -12,  -4,  -4,  43,  28,  -9,  19,  10,   3,  -2,   1,  41, -28,  13,  -8,  17,  -6,   3,   2, -15,  -8,  16,  -2,  -2,  10,  -1,   2,   0, -23,   5,  -8,   0,   0,   0,   0 },
      {   5,   7,  24, -23, -20,  10,  -4,   0,  12,  19,  44,   3,  11, -19,   6,  -4,   9,  18,   6, -61,  15,   5,   1,   4,  16,  19,   3,   7, -30,   7,  -4,   1,  41,   3,  15,  -3,   5, -12,   5,   0,  19, -12,  25,   3, -15,  11,  -8,   3, -16,  14,   3,   5,   5,  -4,  -1,   0,   1, -26,  24,   7,  -2,   3,  -1,  -1 },
      {   4,   7,  13,  38,  14,   5,  -5,   4,   5,   8,  31,  44, -23,  10,  -5,  -1,   4,   6,   9,   4,  21, -14,   5,  -6,  -1,   3, -20,  -3,  -9,   5,   0,   0,   1,  -1,   6,  31,  10,  -1,   2,  -1, -28, -26,  -8, -38,  37, -11,   4,   3,  26,   3,  35,  -4, -23,  11,  -1,  -2, -13, -27,  -9,  17,   4,   2,   0,   2 },
      {  -7,   1,   4,   6, -13,   4,  -1,  -1,  -9,   0,  14,  15, -13,   1,  -1,   4, -10,   5,   9,  -1,  14,  -7,   0,  -1, -19,   4,   5,   1, -12,   7,  -3,   1, -20,  33,  -3,  46, -22,  -1,  -6,   2,  -6,  16, -18,  31,  29,  -6,   7,  -4, -63, -18, -18, -33,  27,   7,   4,   3,  26, -15,   8, -13,  -9,   2,  -2,  -1 },
      {   0,  -6,   5, -13, -54,  37, -13,   3,   1,  -9,  -9,  20,  17, -34,  12,   0,  11,   3,  28,  58,   6,   7,  -6,  -1,   5,  -2,  18, -38,  28,   6,  -4,  -3,   7,   3,  -4,  -5,  10,  -9,   5,   1,   2, -22,   8, -14,  -2,  16,  -5,   0, -17,  -4,   9,   1,  -7,   3,   1,  -2,  15, -11,   0,  20,  -9,   5,   2,  -1 },
      {  -1,  -2,   5,  28,  39,  -1,  -7,  -1,   0,  -2,  31,  19,  -4,   3,   2,   2,  -6,  -7,  22, -13,   6, -19,   1,  -9,  -3,  11,  38, -38,  30,   3,  -8,  -4, -11,  18, -16, -22,   5,   4,   1,   2,  -3,   7,  11,  -3,  -4,   1,  -5,  -3, -32,  46, -29,  24, -15,  -1,   4,  -4, -13,  29,   7,  -3,   7,  -8,   3,   1 },
      {   3,   7,  11,  -4,   0,   3,   0,   0,  -1,   0,  -7, -45,  -6,  -7,   5,  -1,   3,  13,  25,   1,  26,  -4,   1,  -5,   2,  16, -21, -59, -14,  15,  -1,   3,   3,  -3,  -7,  34, -36, -12,  -4,   1,  31,  21,  20, -11,  22,  -8,  -2,  -1,  31, -11,  -5,   5,  -2,   9,  -3,   2, -18,  28,   5, -16,  -2,  -2,   1,   0 },
      {   1,  -3,  -6,  -2,  28,  18,  -9,   2,  -1, -11,  -5, -20, -29, -11,  12,  -1,  -1, -20,  -5, -12,  14, -23,  -4,  -1,   0, -15,  18,  -2,  22,   4,  -1,   0, -18, -16,  14, -25,  -8,  -6,   6,   3, -15,  15,  32,  -5,   9,  -3,  -6,  -2,  22,  -6,  12, -31,   8,   7,  -1,   1,  34, -28,  60,  11, -21,   6,  -4,  -1 },
      {   6,   5,  -4,  -4,  47, -10,   5,   0,   5,   1,  -4, -13,  33,  -9,   8,   0,   7,   6,  -6, -24,   1,   9,   2,   3,  -1, -11, -26, -20,  10,   8,   1,   0,  10,   2,   0,   4,  15,  -8,   7,  -1,   5, -26, -22, -33,   5,  21,   0,   7, -26, -24,  -6, -26, -26,   2,  -2,  -1,  57,  24, -14,   9, -14,  -4,  -1,  -3 },
      {   1,   1,  -7,  15,  24, -24,   8,  -4,   6,   7,   2,  37,  60,   2,  -1,  -1,   2,  -1,  18,  -5, -40,  28,   2,   0,   3,   4,  -3, -30,   8, -10,  10,   0,  -5, -13, -22,   4, -27,   9,  -7,   4,   5,  -4,  21,  20,  -2,  -9,   4,  -2,  13, -17,   5, -19,  22,   4,  -1,   2, -18, -31,  22,   9,  -8,   5,  -3,   3 },
      {   2,   4,   2,  -9,  18,  37, -13,   7,   4,   4,   6,  30,  15, -29,  20,  -6,  -1,   0,   4,  -9, -31, -18, -15,  -5,   2,  -1,   8,  18,  13,   1, -11,  -4,   0,  27,  36,   7,  25,  -5,   6,   1,  33,  41, -24,   7,  19,   0,  -5,  -1,  32, -24,  -9, -10,   0,   8,   5,  -3, -24,  19, -10,   0,  -7,   1,   2,   1 },
      {   1,   3,  -1,   4,  30,  19, -13,   1,  -4,  -2, -14, -14,  -9, -24,  18,  -4,  -1,  -4, -13,   6, -34,  10,  -4,   6,  -4, -21, -25,  -5,  -5,  -9,  -7,   2, -10, -20,  14,  55,  10, -25,   3,   0, -22,   6,  21,  -7, -11,  22, -11,   1, -26,  16, -22,  39,   9, -10,   3,  -3, -14, -30,   4,   5,  20,   2,   2,   2 },
      {   2,   1,   3,  32, -22, -16,  -3,   1,   0,   0,   5,  12, -53,   2,  -3,  -1,   0,   6,  10, -22, -27,  14,  -2,   8,  -2,  11,  -7,  -7,  35,  -9,  -3,   2,   6,  21, -11,  23, -15,   0,  -4,   2,  22,   5, -15,   8, -35,  25,  -6,  -1,  35,  -9,   0,  13,   3, -21,  -2,   0,  41,   5,   3,  27,   1,   3,  -1,   0 },
      {  -1,  -3,  -4,   5, -10, -30,   9, -10,  -1,  -4,  -6, -14, -12, -14,   3,   3,   1,  -3, -11, -18,  31, -14,   2,   8,   6,   0, -10,  12,  56,  12,  -6,   3,   5, -10,  -5,   5,   4,   4,   2,   1,   8,  -9,   3,  -3,  -1,  22,  -2,   4, -22, -23, -21, -24,   7,  14,  -2,   3, -66, -13, -14,  35, -12,   9,   4,  -1 },
      {   3,   6,   0,  -5,  32,  44, -14,  21,   0,   5,  -6,  -8,   4, -19,  18, -13,  -1,   2, -15,  -5,  19, -17,  -7,   2,   0,  14,  -7,   5,   6,  24,  -4,   2,   3,   4, -45,  -7, -36,  16,   9,  -2,  -4, -22, -16,  35, -17, -17,   1,   1,  14,  12,   3,   4,  13,  -5,  -5,  -2,   8, -24, -46,  11,   5,   5,   2,  -1 },
      {  -3,  -6,  -4, -26, -37,  10, -11,  11,   3,   4,   4,  -6,   6,  33,  -1,   3,  -5, -10, -16, -27, -53, -22,  -1,  -5,   0,  -2, -12,   2,  24,   1,  -3,  -3,   0, -12,  -4,  11, -27,  -3,   6,  -2,  -3,  -8,  24,   3,  40, -22,  -1,   0,  -8,  26,  -4,  -3, -25,  23,   2,  -2,  10,  14, -13,  32,  -4, -16,   3,  -2 },
      {   0,   0,   4, -10, -12,  11,  12,  -7,  -2,  -4,   1,  -9,   2, -26,  -3,   2,   0,  -2,   5, -32,  -2,   0,   5,   2,  -2,  -2,  -6,  -8,  48, -44,  11,  -9,  -1,   7,   6,   9,   6,  26, -17,   2,  -8,  -9,   2, -33,   7, -18,  11,  -9,   4,  21,  17,  -2,  34,   2,   1,   4,   8, -11, -23, -60,   3,  13,  -1,   3 },
      {  -1,  -3,  -5,  -6,  -9, -63,  -5,  -1,   3,   1,  -1,  -1,  26,  -8,  10,  -6,   1,   1,   0,  14,  21, -26,  -6,   2,   2,   4,   5,  -2,   2,   5, -10,  -3,   1,   5,  -2,  11,  33, -13,   8,   2,   3,  23,  11,  15,  21,   0, -11,   0,  34,  33, -29,   3,   7,   8,   4,  -3,  39, -35, -27,   2,  12,   6,   1,   4 },
      {   1,  -1,   0,  -8, -12,  11, -18,  12,   3,   1,  -1,  -5,  24,  49,   5,  -2,   3,   2,   1,  -5,   2, -60,  40, -16,   0,  -5, -17, -21,   8, -11, -28,   4,   1,   1,  -7,   4,  18,  14,   0,   0,   2,  -3, -25,  -4, -24,   5,   4,  -2,   1, -32,  -7,  18,   5, -16,   2,  -1,  -3, -22,  28, -12,  14,  -1,  -1,   1 },
      {  -1,  -1,  -2,   0,   3, -18,   4,  -7,   0,   1,  -5,   9,  36, -20,  -5,  -9,  -1,   1,  -4,  13,  21,  -3,   2,   0,   1,   5,   2,  46,  32,  25, -11,   6,   3,   8,  25,  26, -48,  21,   8,   1,   0,   1,  -6, -18,  -3, -17,  -1,   5,  -4,   1,  11,  50,  -8,   2,  -8,  -3,  14,  13,  37,  -2,   2,  -8,   2,   3 },
      {  -3,  -5,  -4,   0, -10, -40, -51,  20,   0,  -1,   1,   2, -16,   7,  40, -22,   0,  -3,   4, -10, -20,  16,  -2,   6,  -1,  -4,   4,  -7,   8,  40,  -3,  -4,  -1,  -1,   4, -14,  -8, -22,  29,   4, -10, -12, -13,  -5,   0,  -4, -19,  11,  -9,  -8,  28,  -4,   5,  -1,   2,  -8, -15,   0,  -3, -52,  -4,   5,   1,   1 },
      {   1,   2,   2,   1,   1,  17, -14,  26,   1,   2,   1,  -2,   3,  37,  -7, -13,   3,   2,   8,   8,   7,  36,   6,  -8,   3,   5,  14,  11,   6,   5,  15,   9,   4,   7,  22,  -4, -29,   6,   4,  -3,   4,  24,   1, -52, -18,  -2,   7,   6,   7,   6, -59, -23, -11,  -2,   1,   2,   0, -38, -20,  -5,  -3,  -1,  -1,  -4 },
      {   1,   2,  -2,   6,  -6,  12,  56, -20,   0,   0,  -1,   2,  -4,  10, -32,  19,  -1,  -2,   1,  -3, -30, -24, -27,   1,   2,  -2,  -2, -16,   0,  69,   1, -11,   2,  -1,   5,   5,  13,  -3,  23,   9,  -5,   1,  -4,  -9, -26,  -7,  -4,   0,   3,   2,  -6,   6,   6,   9,  -2,  -1,   2, -20,  -5, -25, -25,   7,  10,  -2 },
      {   0,   1,   3,   7,   0,   0,   2,  19,  -1,   0,   1,   9,  10,  11,   8, -21,  -2,  -1,   5,  13,   5, -21,  -5,  12,  -2,  -1,   0,   7,   4,   7,  -5,   5,  -1,  -5,  15,  40,  14,  25,   7,  -9,   0,   7,  53,  12, -47,   1,  13,  16,  -5,  15,  26, -52, -11, -23,  -9,   4,   0,  31, -12,  -9,   7,  -2,  -3,   0 },
      {   0,   0,   1,   2, -13, -24,  14,  14,  -1,  -3,  -2,   3, -23, -60,  -4, -18,  -1,  -3,  -1,  -4, -28, -17,   0,  -7,  -1,  -3,  -2, -17, -24,   3, -15,  14,  -1,  -6,  -1, -22,  -6,  39,   7,  -8,   1,   1,  -4, -15,   8, -25,  22,  14,  -9, -17, -22, -11, -22, -20, -15,  -9,  -1, -16,   5,   1,  34, -26,  -5, -10 },
      {   0,   1,  -1,  -6,  -2,   9,   0,  24,   0,   0,  -3,  -6,  -4,   1, -16, -13,   1,   0,  -4, -10,  18,  31, -13, -10,   3,   1,  -9,  -4,  29,   3,  33,  -3,  -1,  -2, -14,  17,  37,  21,  -4,  16,   2,   8,  -1,  46,   8,  -5,  -1,  12,   1, -10,  -2,  15, -63,   8,   4,  -4,  -3, -23,  21, -27,  -7, -17,  -8,  -4 },
      {   0,   0,   0,  -1,   3,  -7, -20,  15,  -2,  -2,  -3,  -5, -11,  -3,  -6,   2,   0,   2,   0,  -1,  -4,  31,  17, -14,  -3,  -5,  -6,  -9, -15,   0, -45,  16,   0,   1,   1,   6,  28,  47,  17, -16,  -3,   3,   4,   3,  -3, -17,  18,   7,   0,  13,  -5,  19,  18,  29, -16,  11,   2,   9,   0,  14, -62,  27,   7,   0 },
      {   2,   2,   2,   3,  -6,  -5,  52,  50,   1,   0,   1,   3,   0,  24,  39, -34,   0,   0,   2,   1,  10,  -9, -62,  16,  -2,  -2,  -3,  -5,  -8, -36,   1,   1,   1,   1,  -2,  -5,  -5,   9,  16,  -5,  -1,  -4,  -9, -11,   8,   7, -13,  -4,  -8,  -4,  -3,  19,  14,   8,  -3,  -5,  -3,   4,  12,   6, -14,   5,   1,   1 },
      {   1,   0,   2,  -2,   3,   6,   2,  23,   0,  -3,   1,  -7,   1,   7, -31,  -4,   1,   0,   1,  -4,   9,  20,   4,  -1,   1,  -1,   1,  -6,  20,   5,  19, -13,   4,   2,   0,   6,  30, -29,  39,  -6,   4,  -4,  -4,  -1,   7, -59,   0,  18,  -5,  -9,  -5,   0,  39, -22, -18,  -6,   1,  19,  14,  30,  39,  19,  -1,   8 },
      {  -2,  -2,  -1,  -5,  -2,  -6, -47, -37,   1,   2,   0,  -3,  -1,  13,  29,  25,   0,   0,  -2,  -1,   3,  -6, -53,  17,   0,  -2,  -8,  -8, -13,  10,  33, -28,   3,   2,   2,   7,  11,  49,  -6,  21,   4,  -3,  -1, -16,  -7, -23,  18, -11,  -4,  -6, -13,   7,   5, -14,   8,   7,   0,   2,   8,  20,   4,   5, -11,   2 },
      {  -1,   0,   0,  -4,   2, -16, -24,  62,  -1,  -1,  -2,  -1,   5, -18, -49, -15,  -2,  -2,  -5,  -1, -10, -23, -14, -12,   1,   0,   0,   0,  -7,  27,  27,   1,  -1,  -2,   2,  -1,   3,  -2, -55,   8,  -1,  -6,   0, -13,   0,  22,  13, -18,  -4,   2,   4,   6,  21,   1,  23,   0,  -1,  12,   5,   9,  -7,  22,  -4,  10 },
      {  -1,   0,   0,   0,   0,  -5,  -2,   8,  -2,  -1,  -2,  -1,  -6, -15,  -6,  -1,   1,   2,   1,   5,   1,  -6,  22,   3,  -3,  -1,  -2,  -1, -16, -20,   7, -18,   0,   5,   1,  10,   1,  11,  -1,  40,  -4,  -3,  -3,  -9, -42, -34, -61,  -6,  -3,  -4,  -4, -15,   2,  45,  21, -39,   0,   5,  -2,  12,  -7, -27,  12,  18 },
      {   0,   0,   0,  -1,  -5,  -7,  35,   8,   1,   0,   1,   3,  -8,  -5,  69,   0,  -1,   0,  -3,   3,  -6,  -2,  41, -36,   0,   2,   2,   6,   1,  20,  38,  -2,   0,   2,   3,   5,  12, -22, -35,  14,  -1,   0,  -2,  -1, -14, -30,  27,   3,  -2,   7,  -5,  11,   2,  -3,  -7,  11,   2,   4,  -1,  14, -21,   0, -17,  -5 },
      {   0,   0,   1,   3,   0,   6,  -7, -18,  -2,  -2,  -2,   2,  -1,  -3, -12, -52,   0,   3,   5,   4,  -9, -23,  40,  77,   0,   2,   1,  -6,  -8,   0,  47,  -6,   0,   3,  -1,  -3,   4,   4,   2, -23,   1,   4,  -6,   1,  10,  -1,  -5,  11,  -1,   2,  -9,  15,   0,   1, -13,  14,  -1,  -2,   1,   3, -31,  -3,   1,  -5 },
      {  -1,  -1,  -2,  -1,  -3,   0,  -6, -19,   0,   0,   0,   2,   3,  -2,  -3,   0,  -1,  -3,  -1,   1,  -7, -19,   0, -46,  -1,   1,   1,   0,  -9, -10,  50,  58,   0,   3,  -1,   2,  -6,  22,  23,  19,   0,   1,   3,   0,  13,  19, -46,  17,  -4,   3,   1,   0,   1, -19, -30,  -7,   0,   6,  -6,   0,  -6,  37,   8,  -1 },
      {  -1,  -1,  -1,  -3,   0,   3, -11, -44,  -1,  -1,   1,  -2,  -1,  10,  -9, -62,  -1,   0,   0,  -4,   2,   4, -34, -13,   0,  -1,   0,  -7,   7,   0, -15,  62,   0,   1,  -1,   1,  15, -26, -24,   5,  -2,  -7,  -7,  -3, -20, -31,   8, -10,  -2,  -2,   4,  -3,  10,  21,  11,  18,   2,   0,   0,   6,   5, -21,   1, -10 },
      {   0,   0,   0,   1,   2, -14, -27,  14,   0,   0,   2,   3,   5, -13, -21,  31,  -1,  -1,   2,   2,   5, -26, -27, -12,  -1,   0,   0,   2,  -2, -36,  13, -11,  -1,   1,  -2,   6, -10, -41,  10,  -4,  -3,  -1,  -8,  -2, -26,  -9,  13,  17,   0,   0,   1,   5,   1,  -1, -48,  17,  -7,  -6,  -6,  -7, -40, -41,  -6, -35 },
      {   1,   0,   0,   1,   2,  -3,   2,   3,   0,  -1,   0,  -2,  -2,  -3,  -2,  -3,   1,   0,  -1,   2,   5,   5,  11, -28,   1,  -1,  -1,   0,   3,   7,  23,   6,   3,   0,   2,   6,   8,  16,  45, -23,   3,  -4,   6,  -1,  13,  14,  -3, -37,   4,   3,   7,   2,  31, -32,  48,  15,  -3,   0,  -5,   7, -22, -69,   0,  10 },
      {   0,   0,   0,   0,  -3,  -8,  -1, -19,   0,   0,  -1,   0,  -1, -10, -12, -30,   0,   0,  -3,   2,   1, -10,  -6, -23,   0,   0,  -3,   5,  -4, -25, -13, -32,  -1,  -3,  -5,   4, -15, -23,   9, -12,   1,  -3,   3,   1,  -7, -31,   0, -21,  -6,  -6,  -9,  -3, -40, -51,  33, -14,  -6,  -3,  -5,  -2, -42,  46,   1,  14 },
      {   0,  -1,   0,   0,   0,  -5,  -2,  12,   1,   1,   2,   1,   0,  -6,  13,  48,   0,   0,   1,  -1,  -2,  -6, -10,  15,   1,   3,   2,  -1,   5,   2,  18,  39,   1,   0,  -2,   5,   5,  -3, -20, -87,  -2,  -2,  -6,   1, -17, -30, -29,  -6,  -1,  -8,  -3,  -2, -17,  10,  23,   1,   1,  -4,   7,  -6,   5,  12,  11,   5 },
      {  -1,  -1,   0,   0,  -3,  -2,  -5, -18,  -1,  -1,   0,   1,  -4,  -7,   0, -26,  -1,  -1,   0,  -1,  -3,  -9,  -4, -45,  -1,   0,  -3,   0,  -7,  -9,  28, -42,  -2,   1,  -4,   1, -10,   2,  26, -39,  -1,   2,  -7,   2, -22,  24,  23,  28,   1,  -1,   0,  -1,  -1,  60,  13,  22,   5,   4,   8,   0,  27,  14,   9,  17 },
      {   0,   0,   0,  -1,  -1,   6,  -1, -18,   0,   1,   0,  -2,   1,  11,   0,  -5,   0,   0,   0,  -1,   3,   3, -13, -12,   0,   0,   0,   0,   3,   5,  -5,  -9,  -1,  -2,  -3,  -2,   4,   5, -36, -24,   0,   0,   2,  -3,  13,   7,   0,  82,   3,   1,   2,   6,  26, -21,  17, -63,   1,  -2,  -3,   2, -21, -25,   3,   6 },
      {   0,   0,   0,   0,   1,   0,  -3,  -6,   0,   0,  -1,   0,   0,   8,   5, -23,   0,   0,  -1,  -1,   2,  12,  -1, -24,   0,   0,  -1,   2,   2,  15,   0, -38,   0,  -3,   1,   3,  10,  13, -30, -27,   0,  -1,   4,   0,   7,   3, -25, -41,   0,  -4,   0,  -5,   6, -10, -46, -13,   3,  -1,   3,   3,   8,  -9,  67, -43 },
      {   0,   1,  -1,   0,   0,   1,   7,   4,   0,  -1,  -1,   0,   1,   0,  -6,  -6,   0,   0,   0,   1,  -1,   4,  -1, -12,  -1,  -2,  -1,   0,  -2,   6, -21, -27,  -1,   0,   1,  -2,  -1,  14, -22,   0,  -2,  -1,  -2,  -4,   7,  -9, -63,  37,  -1,   1,   0,  -2,   5, -15,  21,  75,  -1,   1,  -2,   4,   5,   2, -32, -35 },
      {   0,   0,   0,  -2,  -1,   0,  -1,   5,   0,   0,   0,   1,  -1,  -2,   4,   5,   0,   0,   0,  -2,  -1,   0,  -7,   3,   0,   0,   0,  -1,   0,  -1,  -7,   1,   0,  -1,  -2,  -5,   0,  -5, -18,  14,  -1,   0,  -1,  -1,   5,  -6,  -3,  21,  -1,  -3,  -4,  -3,  -7, -20, -20,  51,   1,  -5,   0,  -3,  -7, -19,  62,  86 },
      {   0,   0,   0,   2,   1,  -3,   2,  10,   0,  -1,   0,   1,   1,  -5,   4,  12,   0,   0,   1,   0,   2,  -4,   7,  14,   0,   0,   1,   1,  -2, -11,   8,  14,   0,   0,   0,   0,  -3,  -4,  15,  30,   1,   0,   1,   2,   0,  -6,  26,  29,   2,   0,   3,   3,  -1,  -3,  57,  13,   2,   3,   4,   2,   3,  14,  76, -56 },
    },
    {
      {-108,  39,  12,   4,   3,   1,   1,   1,  39, -30,   1,   0,   0,   0,   0,   0,  19,   1,  -7,  -1,  -1,   0,   0,   0,  -3,   2,   2,  -1,   0,   0,   0,   0,   4,  -4,   1,   0,  -1,   0,   0,   0,   1,  -1,  -1,   1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -25,  27,  -7,   0,  -2,   0,  -1,   0, -91,  13,  25,  -1,   4,   0,   2,   0,  27, -59,   8,   6,   1,   1,   0,   1,  27,   0, -25,   3,  -1,   1,   0,   1,   0,  12,  -2,  -6,   1,  -1,   0,   0,   4,  -1,   5,  -1,  -2,   1,   0,   0,   1,  -1,   0,   3,   0,   0,   0,   0,   2,   1,  -1,   0,   0,   0,   0,   0 },
      { -25,   2,   7,   3,   0,   0,   0,   0,   5,  42, -18,  -3,  -2,  -2,  -1,  -1, -64,  -2,  47,  -5,   6,  -1,   1,   0,  32, -66,   1,  17,  -1,   4,   0,   2,  15,  10, -30,   2,   0,   0,   0,   0,   3,   5,  -2,  -8,   2,   0,   0,   0,   2,   2,   1,  -2,  -3,   1,   0,   0,   1,  -1,   1,   1,  -1,  -1,   0,   0 },
      {  -4,  26,  -5,  -5,  -2,  -1,   0,  -1, -56, -11,  11,   8,   1,   2,   0,   1,  -8,  42, -15, -13,   1,  -3,   1,  -1, -28, -21,  62,  -3,  -3,  -1,   0,   0,  24, -47, -15,  31,  -1,   3,   0,   2,   2,  14, -23,  -7,   8,  -1,   2,   0,   3,  -3,   4,  -5,  -1,   2,  -1,   1,   2,   2,  -2,   1,  -1,  -1,   0,   0 },
      {  43,  92, -38,  -8,  -8,  -3,  -4,  -2,  37, -13,  26,  -8,   1,  -3,  -1,  -1,  -6, -23,  16,  12,  -2,   2,   0,   0, -21, -12,  -9,  14,   5,   2,   1,   0,   4,  -1, -16,  -4,   4,   1,   1,   0,  -4,   6,   1,  -6,   1,   0,   0,   0,   1,  -2,   2,  -3,  -2,   0,   0,   0,  -2,   2,   1,   0,   0,   0,   0,   0 },
      {   8,   8,   3,   6,  -2,   0,  -1,   1,  28,  67, -15,  -6,   0,  -3,  -2,  -2,  79,  -9,  13, -13,  -2,  -2,  -1,  -1,  19, -12,  20,   2,  -6,   2,  -3,  -1, -17, -22,  -3,  25,  -4,   1,  -1,   0,   2,   1, -32,   1,  11,  -3,   0,   0,  -2,   9,   5, -19,  -2,   1,   0,   1,  -2,  -3,   3,   1,  -3,  -2,   0,   0 },
      {  14, -16,  11,  -4,   1,  -1,   0,  -1,   2, -50,  -3,  18,   1,   4,   1,   2,  44,   7, -13, -12,   1,  -2,   0,   0,  26, -30,   2,  20,  -9,   5,  -3,   1,  -1,  37, -62,   4,   8,  -2,   2,  -1, -18,  34,  16, -33,   4,  -2,   1,  -1,  -2,  -5,  16,   3,  -8,   1,  -1,   1,  -5,   1,   1,   2,   0,  -1,   1,  -1 },
      {   8, -19,  13,   2,   3,  -1,   1,   0,  22, -23,  41,   1,  -2,   1,  -2,   0, -23, -39, -30,  25,   6,   5,   2,   2,  44,  -4, -10, -28,   0,  -1,  -1,  -1,  29, -20,   1,  25, -16,   2,  -4,   1,   2,  22, -50,   5,  14,  -4,   1,  -1,  -3,  18,   7, -25,  -1,   1,   0,   0,  -3,   0,   4,   1,  -5,  -1,   0,   0 },
      { -11, -45,  14,   8,   1,   1,   1,   1,   4,   0,  26, -10,  -1,  -2,   1,  -2,   9, -33,  29,  18,  -8,   4,  -3,   1, -74,  -3, -16,  36,   8,   1,   3,   2,   3, -32, -30,  -7,  19,   2,   4,   1,  17,  11, -17, -19,  -8,   4,   0,   0,   1,  10,   2,  -2,  -4,  -2,   0,  -1,   3,   4,   0,   1,  -1,   0,  -1,   0 },
      {  15,  11,  39,  -3,  -2,  -3,  -2,  -2,  -3,   0, -12,  10,   4,   2,   1,   1,  31,  11, -11,   4,  -3,   2,  -2,   0, -13, -46, -21,  -7,  18,   1,   3,   0,  55,  -8,   7, -43, -24,   3,  -6,   1,   8,   9,   8,  43, -23,  -6,  -3,  -2,  -2,  15, -25,   2,  20,  -9,   1,  -1,  -3,   4,   9, -13,   3,   4,  -3,   0 },
      { -15,  -7, -44,  12,   7,   4,   1,   2,   5,  49,  -3,  -8,  -3,  -3,  -2,  -1,   2,  10, -39,  17,   6,   0,   2,   0, -13,   6, -14, -31,  13,  -1,   2,  -2,  33,   0, -11, -25,  -6,   1,  -1,   1,  -1,  45,  20, -28,  -8,  -7,  -1,  -2,   3, -16,  45,   6,  -7,  -2,  -3,   0,   1, -11,  -1,  12,   1,   1,  -2,   0 },
      {   8,  23,  76, -35,   0,  -8,  -1,  -3,   9,  36,  28,  17, -13,  -5,  -4,  -2, -13,  -6, -40,   8,  15,   3,   4,   1,  -1,   2,   8,  -7,  -1,   0,  -1,   0, -25,  -9, -10,  -5,  16,   2,   2,   0,  16, -11,  18, -26, -11,   6,  -2,   2,   5, -13,   5,  25, -14,  -1,  -1,  -1,   2,   0, -12,   3,   6,  -2,   1,   0 },
      {   7,  26,  23,  19,  -1,  -4,  -3,  -1, -18, -10, -58,  21,  11,  11,   3,   3, -11,  11, -12,   9,  -2,  -1,   1,   1,  -6,   7, -37,  -1,  21,   0,   3,  -2, -18, -10,  -1, -19,  11,  -1,   2,  -1,  15, -10, -25, -29,  18,   5,   1,   1,   8,  21,  20, -48, -22,   9,   1,   1,   4,  -4,  21,  13, -19,  -7,   0,   0 },
      {  -4,   2,  31,  -2,  -3,  -3,  -2,  -2,  -8,  18,  21, -17,  -4,  -2,  -3,   0,   3,  19,   2,  34,  -8,  -3,  -1,  -1, -35,  -7, -10,   1,  11,   5,   1,   3,   6,  48,   7,  23, -16,  -5,   0,  -2, -53,  -4,  -1,  14,  45,  -6,   6,   0,   8, -33,   0, -31,   3,  19,   0,   4,  -3,  -2,  -3,   7, -16,  -1,   3,  -1 },
      {   5,   8,  25,  -4, -10,  -5,  -1,  -1,   1,  -8, -55,   2,   7,   0,   3,   1, -15, -65,   4, -27,  11,   3,   1,   2, -39,  10,  28, -21, -20,   6,  -2,   2,   4,  20,   8,  18, -23,  -3,  -1,   0,  -6,  36,   1,  11,   0, -10,   1,  -2,  -9,  -6,  20,   8,  10,  -2,  -1,  -2,   1,  -8,  -4,   9,   5,   2,   0,   1 },
      {  -3, -18, -36,   4,  11,  -1,   3,   2,   5,  -8,  -4,  22,  -4,   2,   0,   0,  10, -29, -22,   7,  34,  -3,   5,  -2, -27, -57,  14, -37,  10,  13,  -1,   4, -17,  10,   8,   2, -11,   2,  -2,  -1,   7, -43,   9, -11,   9,   8,   0,   3,  22,   0, -25,  -2, -28,   9,  -1,   2,  -3,  23,  -8,   1,  -5, -12,   3,  -1 },
      { -11,   1, -20,  16, -12,   3,  -1,   3,   1,  25,   3,  50,  -2,  -4,  -4,  -1, -19,  -6, -43,   0,   6,   5,   2,   1, -13,   5,  -1,  21,  -4,  -4,  -2,   0, -34,  14, -25,   2,  10,   3,   4,  -1,  -4,  17,  -8,  18,   3, -17,   4,  -3,  -7,   8, -28, -19,  49,  -5,   2,  -1,   5,   2,  15, -44,  -2,  15,  -2,   2 },
      {   4,   1,   2,  13,  -4,   0,  -1,   1,  10,  17,   3,  20,  10,  -5,  -1,  -1,   0, -10,  -8, -40,  10,  10,   1,   1,  -3,  34,  -3,  14, -29,  -3,  -3,  -3,  77,  10,  -7,   5,  26, -21,   3,  -5,   8, -26,  14, -14,  24,  12,  -9,   5,   5,  -3, -24, -12, -17,  17,   2,   2,  -5,  12,   1, -11, -13,  -8,   6,  -1 },
      {  -9,  -7, -22, -96,  12,   4,   5,   0,  -3,   4, -30,  -2, -29,   3,  -1,   3,   5,  -6,  -5,  31,   0,  -8,  -2,  -1,  -2,   8,  12,  17,   0,  -4,  -3,   0,  22,  17,   6,   2,  19,   1,  -1,   0,  -3,  -1, -13,  -6,  -1,  11,   1,   2,  -7,  25,  -7, -11,  -1,  -6,   0,  -2,   3,  -1,  20,  -7,   0,   0,  -2,   0 },
      {   3,  -2,   2,  21, -15,  -1,   0,  -1,   3,  -4, -40,   7,   1,   2,  -1,   1,   0, -28,  -8,  39, -10,  -3,  -1,  -1,  23,   8,  22,  18,  33, -15,   1,  -1,  16, -36,  -2,  -6,  24,  23,  -5,   5, -50,  -5,   3,  -5,  -7,  13,   8,   0,  -2, -40, -17,  13,  -1,   3,   3,   1,   1,  16, -28, -11,  -1,  -2,   2,   0 },
      {  -1, -10,  -7, -39,  10,  -4,   3,   0,   7,  -9,   7,  25,  18,  -1,   0,   0,   4,  -5,  28,  -7,   7,  -2,   3,   0,  11,   6,   6,  -7,  23,   0,  -2,  -2,  -6, -23,  -6, -20,  -9,   4,   1,   0,  39,  11,  21,  13,  22, -16,  -3,  -4,  -5, -47,   8, -24,  21,  38,  -1,   8,  11,  -8, -17,  -2, -46,  11,   7,   2 },
      {  -5,  -7,   0, -19,  -4,   1,   4,   2,   3,   4,  28,  24,  27, -15,   1,  -3,  -3,  -4,  17, -44,  24,   6,   2,   1,  -4,   0,   2, -14,   7,  -2,   1,   2,  -3, -10,  11, -27,  12,  23,  -2,   5, -72,  -6, -11,  -7, -31,  11,  10,   1,   3,   8,  11, -19,  -5, -11,  -1,  -1,   4, -13,  28,  14, -17,   7,  -5,   0 },
      {   0,  -4,   0, -12,  12,  -3,   2,   0,   1,   1,  -2, -78,  35,   2,   6,  -1, -11, -17, -51, -34, -31,   9,   2,   5,   2, -20,  11,  19,   0, -18,  -4,  -1, -15,  -2,   0, -16,   5,   5,  -3,   0,   1,   2,   8,  -1,  -1,  -2,  -2,   0,  12,  -3, -13, -19,  -1,   9,  -1,   2,   8,  14,  10,  -8, -22,  -3,   0,  -1 },
      {   1,  -6,  -2,  17, -46,   4,  -4,   3,   2, -10,  10,  -3,  -9, -16,   2,  -3,   6,  -3, -16,   1,  20,  -4,   1,   1,  -1, -37,  15,  26,   5, -17,   3,  -1,   3,  18,  53,  -4,  42,  -6,  -7,   0,  24,   6,   0,   1,   8,  24,  -4,   2, -35,  -5,  21,  -2,   6,   8,   4,   1,   4, -39,   1,  14,  -8,  12,   3,   4 },
      {   0,   0,  -9, -18,  31,   5,   0,   0,   1,  -2, -11,  26,  -1,  13,  -1,   1,   0,  -2, -12, -12, -28,  23,  -2,   5,  -7, -36, -51,   8, -42, -11,  11,  -1,  -6, -37,  28,  23,  15, -29,   0,  -2, -24,   1,   0,   3,  10,   3,  -3,   1,  10, -22,  14,  12,  10,   0,   0,  -2,  -1, -10, -22,   5,   7,  13,   0,   1 },
      {   2,   5,   6,  -6,   1,   9,   0,  -1,  -6,  -7, -17, -43,  17, -28,   9,   2,   5,  17,   7,   9,  69,   0,  -7,  -2,   0,   2, -29, -16, -17,  30,   7,   4,  -7, -17,  -1,   5,  11, -10,   3,   3,  -9,  18,  -9, -19,   6,  15,  -3,   3, -15,  -4, -18,   4,  11,   5,  -1,  -2,   5,   9, -15, -43,  -2,  21,   4,   0 },
      {  -8,  -9, -11, -32, -94,  24,  -3,   8,   0,   3,  -5,   4,  12, -12,   1,   0,  -2,   9, -10,  -4,  -3,   8,  -5,   2,   2,   1, -26,   9, -12,   8,   3,  -1,  -7, -13,  -9,   4, -25,  -4,   3,   1,   2,   5,  -6,  11,  -1,  -9,  -1,   0,  13,  -8,  -7,   5, -18,  -1,   1,   1,   5,  27,  -9,  25,   0, -23,   0,  -4 },
      {   4,  -3,   3,  -1, -17,   6,   1,   2,   1,  -9,  -8, -12,   6, -18,   2,   1,   5,  -5,  -1,  12,  11,   1,  -2,  -3,   6,   1,  10,   2,  -2,  -9,   2,  -3,  15,   4,  -7,  -2,   7,  -4,   3,  -1,   8, -40, -10, -13,  -2, -32,   4,   1,  77,  -1,  25,   1,  40, -12, -21,   0,  -9, -18,  -1,   4,   7,  47,  -5,   1 },
      {   1,   4,   0,  -7,  10,  -6,   5,   2,   1,   3,   2,  15,  27, -14,  -3,  -2,   3,  13,  11, -14, -12,  10,  -3,  -3,  -6,   1, -10, -11,  20, -24,   0,  -1,   3,  45,  17,  11,  22,  24, -21,  -1,  27,  31, -48,  -4, -14,  21,  12,  -6,  33, -17,  -7,   5,  -8, -14,   9,   3,   9,  24, -38, -14,  10,  -4,   1,   2 },
      {   3,   1,   6,  10,  -7,   7,  -6,   2,   0,  -7,   4, -10, -85,   4,   4,  -2,   0,  -6,   8, -35,   5, -27,  -4,  -3,   4,   2, -12,  -6,  13,  -5,   1,  -2,  -4, -13,  12,   2,   6,  -4,  -7,  -2,  -5,  29,  14,  -1,   5,  10,  -3,  -4,  43, -10,  -6, -11,   1,  -3,   3,   1,  37,  23,  18, -10, -10,   5,  -5,   0 },
      {  -5,  -3,  -2,  -7, -23, -57,   9,  -3,  -2,   3,   1,   1,   1,  -5, -12,  -1,   0,   2,  14,  -1,  -6,  11,   1,  -4,   0,   0,   1, -19,  17, -27,  10,   4, -12, -10,   9,   0,  14, -15, -10,   3, -14,  28,  22,  -7,  21,   0, -23,  -1,  17,  46,  -7,  11,   6,  31,  -2,  -6, -54,  15,  -2,  -6,   4,   6,  15,   1 },
      {  -2,   1,  -3,  -9, -15,  13,  -3,   3,  -3,  -1,  -6, -17, -27,  17,  -2,   2,  -2,   0,  -9, -37,  -5, -19,  20,   0, -11,  -1, -36, -22,  27, -17,  -9,   0,   7,   1, -30,  30,   3,  31,  -8,   0,   3, -34, -10,   2, -11,   1,  17,  -2, -34,  -3,  10,   7,   9,   6,   1,   3, -45, -13, -19, -11, -13,  17,  13,   1 },
      {  -2,  -5,  -4,  -5, -26, -57,   7,   4,   1,  -4,  -1,  -9,   2,  11,  -8,  -3,   5,   2,   9,  -1, -13,  18,   8,  -3,  -2,  -6,  -5, -39,   1, -23,  13,   3,   6,   2, -21,   2,  10, -17,  -1,   0,   0, -28,  -4, -11,   7,  -9,  -2,   2, -32, -15,   3,  -5,   7, -19,  -6,  -2,  66,   1,   9, -15,  21,   3, -14,   0 },
      {  -2,  -1, -10,  -2,  17, -61,   6,  -1,  -1,   0,  -4,   0, -16, -23,  -6,   0,  -3,   1, -22, -28,  24,   2, -12,   2,   4,   4, -17,  52,  32,  32, -10,   4,   8,  -1,  12,   6, -21,  15,  17,  -1,  -5,  -7, -17,   2,   4, -22,  -2,   6,  -7,   4,   5,   8,   0,   1, -15,  -2,  12,   2, -28,  25,  10,   0,   2,  -7 },
      {   0,   3,  -1,   9,  -8, -10,   3,   5,  -4,  -1,  -2,  -8, -45,  20, -13,   3,   2,   6,   5,   7,  -6,  70, -11,  -2,   2,   1,  16,  -6, -31,  15,  18,   2,  -2,  15,   4, -26,  -6,   8,  11,   5,   2,   1, -24, -14, -34,  -5,   6,   4,  -4,  -8,  -9,  -9,  -2,   9,  -7,   0, -14,   1, -18,  -2, -44,   9,  16,  -4 },
      {   0,   1,  -3,  -1,  14,  12,  18,  -5,  -4,  -4,  -4,  -5, -23, -34,   2,   3,  -3,  -4,  -5, -14,   7, -14,   7,   1,  -1,   5,   9,  -7,   4, -25,  21,  -7, -11,   4, -28, -46,   2, -45,  -3,   4,  -8,  -4, -47,  18,   9,  -4, -35,   8,  -8, -30, -15,  -1,  -5,  -2,   4,  -5, -25, -11,  -8,  21,  20, -14,  -1,  -1 },
      {   1,   2,  -3,   8,  14, -10,  13,  -4,  -3,  -2, -13,  -7,  -9, -49, -15,   2,   2,   3,  -8,  16,  12,  27,  -7,  -2,  -1,   6,  -9,  -5, -15, -14,   7,   5,  -2,  -7, -39,  26,  12,  19, -32,   0,   5,  -6,  23,  47,  -9,  26,   9, -13,  13,  14,  -1,   1,  -3,   8,  18,  -2,  21, -18,   6,  36, -21,  -6,   4,   6 },
      {  -3,  -6,  -6,  -5, -12, -40,   0,  -4,   0,   0,   0, -10,   6,  50,  -3,  -3,  -1,   0,  -4,  -1,  28,   4,   1,  -5,   3,   6,  -5,  15,  -3,  18,   4,   1,  -3,  -8,  -6,   1,  -3,   0, -10,  -1,  11,   6,  -2,  18,  -1,  24,   7, -10,  29, -41,  -8, -12, -13, -26,  30,   7, -35, -40,  24, -11,  16, -11, -16,  11 },
      {   0,   2,  -2,   0,  -2,   8,  -7,  12,   1,   4,   3,  -6,   7,  57,  -2,  -4,   2,   5,  -2,   2,  29,  -8,  30,  -2,  -3,   5,   1,  10,  -8, -18, -17,   3,  -1,   5, -11, -18,  -1,   2,  -9,  -2, -18,   2, -18,  16,   5,  -3, -23,   1,  21,  33,   0,  22,  -1,  30,   8, -12,  44, -28, -38,   2,  -5, -10,  25,   7 },
      {   0,  -2,  -3,   3, -14, -33,  19,  -5,   0,  -2,  -2,   4,  -4, -31,   2,  -9,  -1,  -1,  -1,   8, -41, -53,  -8,   6,  -3,  -7,  -6, -19, -43,  22, -38,  -1,   1,  -4,   1, -17,   1,  19,  14, -13,  -2,   1, -14, -11,  -9,   4,   3,   6,  12,  -6, -15,   7,  -6,  15,   6,  -3,   3, -39,  -3, -19, -14,  -9,  17,   5 },
      {  -2,   1,  -2,   5,   5, -31, -19,   2,   1,   6,   1,   2,   5,  24,  17,  -8,   0,   6,  -3,  14,  17, -39,   2,   1,  -1,   1,  -2,   2, -44, -15, -11,  10,  -2,   3,  -6, -15,  12,   8, -16,   0,   5,   2, -15,  17,  -4,   8,   7,   1, -14, -14,   3,   3,   7,   0, -16,   6, -11,  58,  10,  45, -15,  37,  -7, -12 },
      {  -1,  -1,  -3,  -3,  -8,   6, -11,   5,   2,   6,   3,   2,   9,  17,  -3,   3,  -4,  -6, -10, -11, -31,  -4, -46,   7,   1,   4,  10, -17,  26,  52,  11,  -8,  -2,   7, -20,  -4,  12, -18,   7,  14,  -2,  -3,  -8,  20,   7,  45, -16,   2,   2,  12,   0,   7,  -8,  15,   7,  -9,  11,  -5, -12,  12,   6,  56,  -7,  -4 },
      {   2,   2,   1,   2,   3,  12,   7,   8,   1,   1,  -5,  -4,   8,   9, -28,   1,   2,   1,   2,  -1,  -2,   3,  30, -11,   2,   2, -10, -12,  14,   0,  -1,  13,   1,   0,   4,  16,  13,   5,  27,  -6,   7,  28,  -1, -19,   4, -12,  11,  17,   2,  -9, -72,  15,  -1,   2, -20,   6,   6, -28,  23,  50,   2,  24,  -4,  -6 },
      {  -2,  -5,  -4,  -5,   8, -12,-111,  21,  -2,  -5,  -4,  -2,  -2, -30, -17,  -8,   0,  -1,   1,   1, -10,   6,  30,  -5,   2,  -1,   2,  -6,   0,  10,   2,   3,  -1,  -1,  -3,  -3,   7,   1,   9,   2,  -1,   0,  -4,   6,  -4,   2,  -7,   9,   7,  -6,   3,  -3,  -2,   4,   6,  -8,  -5,  -5,   4,  -7,   5, -11,   4,  -2 },
      {  -1,  -1,   0,  -4,  -2,   6, -17,   4,  -1,   0,   4,  -3, -12,  -3,  -3,   1,  -2,  -2,   2,  -8,  -2,   6, -19,  -6,  -1,   2,   8,  -5, -17, -11, -17,  10,  -3, -10,  -3, -36, -12,  11, -24, -17,  -3,   4,  18, -17,  36,   4,  39, -13,  -9,  23, -22, -38,   3, -19,   5,  22,  -4, -16, -53,  15,  34,  11, -19,  -2 },
      {  -1,  -1,  -1,  -1,  -1, -11,  -9, -23,   2,   3,   1,  -1,  -1,   7,  72, -18,   0,   1,  -1,   2,   1,  -7,   2,   0,   2,   1,   1, -15,   6,   8,  30,  -7,   1,   4,  -7,  13,  39, -14,  31,  13,  -2,   9,  12,  21, -17, -20,  24,  16,   8,  18, -11, -15,  -1,  -8, -11,  17,  -1, -15, -32,  11, -12, -26,  -9,  -4 },
      {  -1,   0,  -1,   1,  -1, -13,  -2,  15,   1,   3,   1,   8,   7, -10,  12,   6,  -1,   0,  -3,   7,  -8, -31,  23,   4,   0,   6,   1,  20,  -1, -14,  40,   8,   2,   3,   5,  23, -49, -24, -22,  16,   0,  -1,  14, -31, -40,  15, -21, -10,  -2,   7, -20, -27,  -9, -15,  12,  -1,   6, -15, -24,   1, -21,  30,   4,  -1 },
      {   0,   0,  -1,  -1,   1,  -3, -15,  13,   1,   3,  -1,   6,  -7,  -5,  83,   2,   1,   2,   2,   1,  -4,  38, -17,   4,  -1,   1,  -4,  -2,  14, -11, -44,   7,  -1,   0,  -3,   6, -21,   3, -29,   2,   0,  -2,  -9, -22,   8,  11, -21,  -3,  -2, -12, -20,  22,  12,   8,  -4,  -7,   0, -18,  25,   8,   1,   1,   1,  -7 },
      {  -1,  -2,  -4,   0,  -5,   1,   0, -62,   1,   1,   0,   6,  -5, -11,  13,  -7,   1,  -1,  -3,  -5, -14,  10,  53,   0,   1,   4,   4,  12,  -3,  15,  28,  17,   0,   0,   1, -19, -22,   7, -20,  11,   0,   0, -15,   1,  33,  16,  32,   1,   4,   7,   8,  36, -16,  13,  -7,  11,   8,  -3,  11, -19,  -7,  21, -15,  -9 },
      {   0,   1,   2,   1,  -4,   1,  30,  24,  -1,  -2,  -1,  -1,  -7, -10,  14,   9,   0,   2,   2,   5, -11,  21,  68,   0,   1,   1,   5,   0,   3,  34, -48,  15,  -2,  -3,   0,   0,  22, -12,  -7,  -6,  -1,   6,  13,  26, -12,   2, -15,   7,   1,  -2,  14, -34, -22, -10,  -9,  -8,  -7,  14, -15,  -6,  17,  23,  -6,   0 },
      {   2,   3,   3,   0,   6,  -5,  20,  86,   1,   0,   1,   0,   1,  -6,  17,  -5,  -1,  -3,  -4,  -3, -11,  -9,  10, -50,   1,   0,   3,   3,  -1,  21,  35,   6,   2,  -1,   0, -12,   3,  -1,  -6,   4,  -1,  -3,  -8,   0,  24,   2,  30, -11,  -1,   7,   8,  28,  11,   1,   9,  12,  -3,   4,   8,  -2, -11, -14,  -2,   9 },
      {   1,   1,   0,  -1,   2,   3,  11,  -3,  -1,  -2,  -3,  -1,  -3,  -5,   3, -83,   1,   1,   0,   1,   2,  13,  13,   6,   2,   0,   2,   1,   8, -10, -12,   7,   2,   0,  -2,   0, -20, -14,  40, -15,   0,  -5,   0, -11,  -7,  27,  12,   9,  -1,   1,   6,  -5,  25,  10,  50,   2,   0,  19,   1,  14,  11,  10,   9,  27 },
      {   0,  -1,   0,  -1,  -2,   4,   1, -36,   1,   0,   3,  -2,  -1,   6,   3,   9,   0,  -1,  -2,  -5, -11,  -3,  17, -37,   1,  -2,  -2,  -7,  -9,  41,  10,  -1,   5,   1,   7,  -2,  19,  23, -22,  30,   2,  -7,   1, -25,   0,  -9, -28, -27,  -4,  -5,  -7, -17,  49,   1,  29, -17,   3,  13,  -2,  32,  20, -19,  22,  12 },
      {   1,   1,   1,  -1,   1,  12,   6,   3,  -1,  -2,  -2,   2,   0,   5,  -8, -62,  -1,  -1,  -2,  -2,  -4,  -4, -20, -18,   1,   2,  -2,   1,  -5,   0,  16,  64,   1,  -1,  -3,   3,   0,  24, -13,  17,   5,  -1,   6,   8,   4,  10, -31,   8,   2,  -8,   5, -16,  -5,  -5, -57, -12,  -4, -13,  -2, -13,  12, -13,   6, -17 },
      {   0,   0,   1,  -1,   1,   0,   7,  11,   0,   0,   1,   2,  -6,  -4,  22,  12,   1,   0,   1,   0,   5,   4,   1, -22,  -2,  -2,  -3,   6, -21, -26,  24, -34,   1,   1,   1,  -1, -14,  52,  29,  23,   6,   3,   8,   7,  -1,   0,  -1,  26,   1,  -8,   1, -22, -25,  44,   8, -24,   2,  -2,   7,  -8,  50,  24,  -6,  -2 },
      {   0,   0,  -1,   1,  -2,  -3,  -2, -17,   1,   1,   2,   0,   2,   1,  -9,  28,  -1,  -1,  -3,  -1,   2,  -3,  -4, -59,   0,   1,  -1,   4,   5, -15, -32,  15,  -1,   0,  -4,   3, -16, -44,  20,  18,   0,  -1,  -4,   1, -32,  40,  37,  -9,   5,  -3,   8, -10,  18,  33, -29,  -1,   2,   0,  -2,  -2,  10, -18,  -4, -12 },
      {  -1,  -2,  -1,  -1,   0,   1,  -5, -20,  -1,   0,   0,   2,   5,  -2,   3, -20,  -2,  -3,  -1,   1,   0,   4, -12, -63,  -1,   0,   1,   4,   6,  -4, -14, -34,  -4,  -4,  -1,   0,  -3, -13, -15,  -9,  -2,   4,   6,  10,   9, -16,   4,  23,  -5,  -2,   0,   1, -31, -40,   5,   8,   5,  -1,   6,  -5,  -8,  27,  69, -11 },
      {   0,   0,   0,   0,   1,   4,  -2,   4,  -1,  -2,  -1,  -1,  -2,   1,  -9, -38,   0,   0,  -1,   0,   1,  -3, -13,   5,   0,   0,   0,   3,   5,  14,  -4,  -7,   1,   1,   6,   6,  15, -28, -55,   6,  -3,  -1,  -7,  -5, -34, -53,  30, -19,  -6,  -5,  -8,  -4, -23,  55,  11,  -1,   2,  -4,  12,  -5,  22,   9, -12,  -8 },
      {   0,   1,   1,   1,   0,  -3,   7,  12,  -2,  -2,  -1,  -1,  -1,   3, -18, -33,   0,   1,   2,   1,   1,   2,  18,  15,  -2,   1,  -2,   1,  -2,  11, -11, -81,  -2,   0,  -2,   0,  -9,  -1, -29,  33,   1,   1,   6,  -9,   6,  30,   5,  28,   2,   2,  -4,  -3,  27,  -8, -38,  -3,  -3,  -4,  -6,   5,  -7, -22, -17, -19 },
      {   0,   1,   0,   2,   1,   2,   0,  -3,  -1,  -2,  -1,   1,  -1,  -3, -12,  -6,   0,   1,   1,   1,   1,  -1, -11, -22,  -2,  -2,  -3,  -1,  -2, -14, -34,  17,   1,   2,   3,   0,   2,  -4,   5,  71,   1,   1,   1,  -1,  15, -26,  -5,  24,  -3,   6,  -1,  13, -15, -21,  39, -25,  -5,  -1, -12,  -5, -35,   7, -52,   9 },
      {  -1,   0,   0,   0,  -2,   2,  -5, -15,  -2,   0,  -2,   1,   0,  -2,  -1, -15,  -2,   0,  -1,   2,  -3,   3,   0, -53,  -2,   0,  -3,   1,  -2,   2,  -2, -26,  -1,   3,   0,   8,  -2,  12,   0, -50,  -7,  -1,  -5,  -1, -13,  -2, -50, -13,  -5,   6,   0,   6,  -7,  12, -12,  36,  -1,  -1,  -9,  -4, -16,   5, -64,  25 },
      {   0,   0,   0,   0,   1,  -1,  -3,   1,  -1,  -1,  -2,   0,  -1,  -5,   4, -23,   1,   1,   1,   1,   4,   3,   9,   3,  -1,  -1,  -1,   0,  -2, -11,  -9, -27,   0,   0,   1,  -3,   2,   1,  36,  22,   0,   2,  -1,   6,  14,  -1,   2,-102,  -3,   0,  -9,   4, -23, -14, -24, -28,  -2,  -4,  -1,  -1,  -6,  11,   5,   3 },
      {   0,   0,  -1,   0,   0,  -1,  -7,  -6,   1,   1,   1,   1,   0,   1,   6,   5,   0,   0,  -1,  -2,  -2,  -2,  -6,   1,   1,   1,   1,   3,   3,   3,   6,  -1,  -1,  -1,  -1,  -2,  -4,  -8, -19, -10,   1,   3,   1,  -1,   1,   0,  23,  19,  -4,  -3,  -8,  -2,  -7,  -3, -38, -51,  -2,   4,  -1,   4,  -4,   1,   4, 102 },
      {   0,   0,   0,   0,  -1,   0,   3,  -6,   1,   0,   1,  -1,   3,   4,   6,  -4,   0,  -1,   0,  -1,   0,  -3,   3, -15,   1,   0,   2,  -1,   5,   2,  12,  -2,   0,  -1,   1,   1,   1,  -3,  -4, -58,   2,   1,   3,  -4,   3,   1,  14,   5,   0,   0,  -1,  -1,   8,  -6,  15, -91,  -5,  -1,  -8,   2, -17,  -5, -24, -51 },
    },
    {
      {-103,  52,  -2,   3,   1,   1,   0,   1, -43,   6,   6,   0,   4,  -1,   1,   0,  15, -16,   2,  -1,   2,  -1,   1,  -1,  13,  -9,  -1,  -1,   1,  -1,   0,   0,  15,  -6,  -3,  -1,   0,   0,   0,   0,   6,   4,  -4,  -1,   0,   0,   0,   0,   3,   1,   3,  -2,  -1,   0,   0,   0,   3,  -2,   2,   1,  -1,   0,   0,   0 },
      {  35,   2,  -8,  -2,   0,  -1,  -1,   1, -79,  52, -10,   2,   3,  -1,   1,   0, -55,  16,   8,  -4,   8,  -4,   3,  -1,  -2, -30,  18,  -6,   4,  -1,   2,   0,  21, -23,  -2,   1,  -1,   0,   1,  -1,  15,   0,  -5,  -3,   1,   0,  -1,   1,   1,   3,   3,  -2,  -1,   0,   0,   0,   4,  -4,  -2,   1,   0,   0,   0,   0 },
      {  -6,  56, -32,   5,  -6,  -2,   0,  -1,  43,  25, -22,  -5,  -1,  -4,  -1,   0,  -6,  47,  -5,  -7,   0,  -1,  -1,  -1, -58,  22,  25, -10,   9,  -4,   2,  -1,  -4, -34,  25,  -1,   4,   0,  -1,   2,  -5, -12,  -3,   4,   1,   3,   0,   1,   1,  -7,   2,  -5,   0,   3,   2,   0,   1,  -3,   3,   0,  -2,   1,   0,   0 },
      { -46, -38,  34,  -6,   7,   2,   0,   2,  43, -18, -10,  10,  -4,   3,   0,   0, -61,  40, -10,   9,  -1,   0,   2,  -1,  -9, -31,  11,   3,   0,   1,   2,   0,  25,  -8, -27,   5,  -2,  -1,   2,  -2,  16,  17,  -9, -11,   3,  -4,   0,  -1,   3,  11,   7,  -2,   0,  -2,  -1,   0,   1,   0,  -1,   2,   2,   0,  -1,   0 },
      {  -4,  21, -21,  -2,   0,   2,  -1,   1,  24,  37, -36,   1,   3,  -1,   0,  -2, -27,  11,  -2,  -9,  11,  -5,   3,   0,  45,   0, -12, -13,  -4,   4,  -2,   0,  -5,  68,  -7, -18,  -3,  -4,   1,  -2, -15,  12,  41, -11,   0,  -4,  -1,  -1,  -1, -17,  25,   9,  -3,  -1,  -2,   0,   4, -16,   7,   7,  -2,   1,  -1,   0 },
      { -21, -43,  24,   3,   3,   4,   1,   3, -34,   3, -28,  16,   1,   2,   1,  -2,  31,  50, -37,  -9,   4,  -5,  -1,   1,   3,  23,  19, -18,  -1,   1,   1,   1, -18,  13,  29,  -8,  -3,  -5,   4,  -2,  20, -38,  29,   1,  -2,  -1,  -4,   3,  12,  -8, -18,  12,   2,   0,  -3,   0,  -4,   4, -13,  -4,   7,  -3,   0,  -1 },
      {   6,  -7,   1,  -7,   4,  -2,  -1,   2, -35,  -7,  -5,   9,   1,  -1,   2,   2,  18,  16, -42,  15,  -5,   1,   3,  -4, -39,  45, -21,  -2,   7,  -5,   5,  -2,   6,  12, -27,  -7,   7,  -5,   4,  -1, -12,  55, -11, -25,  11,  -4,   2,  -2,  -7,   7,  46, -19,  -3,   0,   0,   0,   6, -13,  17,   8,  -7,   2,  -1,   0 },
      {  12,  -6, -29,  21,  -6,   4,   3,   0,   7, -11, -37,  14,   4,  -1,   3,  -2,  51,  14,  -5, -22,   9,  -5,  -1,   2,  25, -39,  35, -18,   1,   1,  -4,   6,  32,  -3,  -2,   0, -13,   3,  -5,   4,  -4,  44, -20,   8, -17,   3,  -4,  -1,  -7,  34,  -6,  -4,   1,  -6,   1,  -3,  -9,  12,  19,  -8,   6,  -7,   2,   0 },
      {   8,  18,  -5,   9,  -5,  -4,   7,  -4, -23, -68,  33, -10,  13,  -6,   4,   0, -36,  31,  -8,  -9,   5,   4,  -9,   5,   6,   5,  -7, -16,   7,  -4,  -4,   4,  37,  14,  37, -18,   1,  -4,  -9,   7, -37,   5,  16,  18, -12,  11,  -5,   0,   3, -14,  -9,   3,  -2,   3,   3,  -3,  -3,   7,  10,  -3,  -3,  -3,   2,   0 },
      {  11,  12,  27, -25,   0,  -5,   1,   0,  -1,   5,  17, -14,  -5,   1,  -2,   1,  42,  27,   1,  -7,  -8,   1,  -4,  -3,   0, -39,  28,  11,  -1,   0,   1,  -1,  18,  -7, -31,  24,  -1,   2,  -2,   0, -42,   3,  25, -11,  17,  -3,   1,   2,   1, -54,  12,   4,  -1,  11,   0,   2,   1,  -5, -31,   6,  -5,   5,   0,   0 },
      {  -6,  -2, -36,  43, -19,   4,   0,   1, -15, -10, -20,  41, -22,  11,  -2,   1, -11,   2,   1,  14, -22,  12,  -2,  -6, -21, -17, -14,  23,   5,  -4,   1,   3, -17,  19, -32,   2,  14,  -5,   1,   3, -28,  -7,  -2,  -9,   7,   0,   5,  -2,  33, -22, -26,  -6,   2,   2,   0,   1, -11,  37, -12, -20,   2,  -3,   2,   0 },
      {  22,  35,   0,  -5,  -4,  -1,  -4,   1,  13, -13, -19,  19,  -5,   2,  -1,   4,   3,  -5, -61,  37,   1,  -9,   6,   2,  36, -13, -34,   6,  12, -11,  10,  -6,  32, -29,   0,  -8,   9,  -2,   2,  -4,  25, -19,   9, -11,  16,   0,  -5,   5, -22,  -1, -14,  -6,   8,   7,   1,   2,  -2,  -3, -25,  -7,   4,   4,   1,  -2 },
      {  12,  20,  -3,  -2,   2,   7,  -8,   3,  -5, -33,  -4,  11,   7,  -1,   1,  -2,   9, -22,  15,   4,   7, -13,   6,   4, -59, -24,  13,   5, -11,   9,   2,   2,  11,  22, -12, -26,  -6,  -5,   3,  -5,  53,   9,  45, -10, -17,  -3,  -9,   0,   1,   0,   2,  38,  -4,  -4,  -5,  -1,   1,  -8, -13,  10,  11,  -5,  -1,   0 },
      {   1,  -6,  -1, -13,  12, -10,   9,  -2,  -1, -19, -35, -22,  32, -12,   5,   0,   5, -18,  -1, -40,  38,  -3,  -4,  -1, -25, -19, -27, -17,   5,  12,  -3,  -3,  11,   3, -23,  -6,   0,   0,  -2,  -1,  -2, -41,  -3, -14,  22,  -6,   0,   5,  19,  -3,  -6, -45,   0,  12,  -1,  -2,  12,   3,  17, -15, -21,  11,  -1,  -2 },
      { -20, -56, -46,  29,  -3,   3,   0,   1,   6,   1, -11,  -7,  -2,  -1,  -2,   3,  -2, -33,   1, -11,   2,   0,   0,  -1,  -1,   8, -11, -16,   7,  -7,   0,  -2,  27, -44,  17,  -8,   1,   6,  -7,   2,  -6,   7,  17,  -5,  -2,   3,   3,  -1,  -7, -42,  27,  17,  -7,   5,   4,   2,   7, -14, -28,  19,  -4,   6,  -1,   1 },
      {  -9, -19,  30,  -6,  -5,  -2,   6,  -3,   4,  44, -13,  -8, -11,   0,   2,   2,  21,  -1,  20,  20, -13,   4,  -4,   4, -33, -21, -40,  20,  12, -10,   5,   2,  43,  11,  27, -35,  18,  -8,  -6,   3, -30,  -4,  16,  15, -11,  15,  -6,   2, -18,  18, -15,   1,   1,  -1,   3,  -3,  -8,   1,  19,  -6,   3,  -3,   1,  -1 },
      {  -5,   1,  19,   5,  -1,  -2,   1,   2, -11, -27, -32,  32,   0,  -4,   3,   5, -12, -22,   4,  18,  15, -19,   5,   1, -14, -31,  24,  -4,  16,  -9,   2,  -2, -47,   5,  27,   4,   4,   4,  -6,   0, -38,  -3, -11,  15,  15,   5,  -3,   2, -43,  -3,  11,  -8,  14,   0,   5,   1,   7, -42,  -3,   9,   4,   7,   0,   2 },
      {  -5,   0,   7, -17,  11,   0,   0,   0,   3,  20, -11,   2,  11,   1,   0,   0, -16, -35, -24,   3,  17,  -1,   5,   0, -25,  32,  13, -25,  -9,   7,  -2,  -2,  30,  25, -10,  26, -27,   4,   0,  -2, -19,  21, -12,  34,   3,  -7,   5,  -1, -17, -14, -41,  15,  19,  -3,   3,   2,   0,  19, -32, -26,  13,  -3,   2,   0 },
      {   6,   7,  35,  14,   0,  -9,   2,  -3,  16,  28,  36,  26,   0,  -3,  -1,  -4,  -1, -30, -36,   5,  18,   7,   3,  -2,  -7, -25,  10, -52,  25,   2,  -2,   4, -19,  -8,   9, -33,  -1,   9,  -5,   7,  -9,   9,  -2, -15, -19,   4,   2,  -1,  38,  -8,  12,   0, -16,  -1,  -2,   1,  -3,  23,   7,   0,  -8,  -4,   0,   0 },
      {   3,   3,  -4, -62,  18,  -4,   3,  -3,  -5, -21, -45,   6, -25,  19,  -1,   0,  -7, -32,   4,  19, -39,  17,   1,  -6,   8,  11,  34,  -9,  -6,  -8,   1,  -2,  18,  -5,  20,  -1,   7,  -6,  -5,   3,  -8,  -9,  -1,  -3,  -2,   5,  -1,  -1,  36,  -3,  12,   1, -17,   1,  -4,  -1,   0,  17,  19,  11, -14,   0,  -2,  -1 },
      {  13,  12,  43,  40, -33,   3,   0,   0,   7,   0,  -3,  47, -16,   3,  -1,  -2,   4,  -9,  34, -13,  12,  -2,   0,  -6,   6,  36,  12,  -6,   1,  -1,  -1,  -4,  51,   9,   0,  21,   6, -10,   4,  -1,  14, -18,  -4,   9,  15,  -8,   3,   4,  12, -10,  18, -11,   4,   3,  -5,   3,   8, -13,  12,  13, -11,   9,  -2,   1 },
      {  -1,  -8, -27,   1, -14,   6,  -4,   1,  -1,   3,  18,  -7, -19,   6,  -9,   0,   9,   7, -31,  32,  -6,   2,  -4,   1, -16, -32, -16, -13, -18,  19,  -2,  -3,  11,  16,   2,  18, -43,   4,   5,  -5,   0, -31, -14,  39,   3, -24,   8,  -4,  22,  -4,  10,   3,   8, -16,  -4,   2,  14, -30,  33,  27,  -1,  -2,  -7,   4 },
      {   6,  21,  40,   1, -26,   5,  -3,  -3,  -9, -11, -39,  -8, -23,  21,  -7,   6, -10,  -1, -10, -52, -14,  20,  -5,  -1,  -1,   1, -40,  -2, -12,   8,   0,   0, -22, -30, -17,  -8, -26,  16,   0,   1,  -9,   8,   0,  18, -30,  -7,  11,  -5, -10,  -2,   3,  21,  -4, -11,   5,  -2,  -1,   3,  -2,  15,   4,  -2,   0,   0 },
      {   3,   5,  -4,  -6,  49, -36,  11,  -8,  -1,  -4, -11,  36,  22, -21,   2,  -2,  11,  25,  30,  23,  13,  -6,   6,  -7,  11,  14, -24,   0,  23,   7,  -1,   0,   4,  -9, -32,  -7,  -8,  19,  -3,   6,  -9, -17, -25,  21, -28,   5,   3,   3,   9, -10,   0,  30, -19, -11,   0,   1,   0,  -6,  -7,  22,   5,  -5,  -4,  -2 },
      {  -4,  -9,  -7, -12,  -9,  12,   5,  -1,   3,   6,   7,  26,  11,  -4,   2,  -1,  -5, -10, -25,   1,  32,   1,  -8,   9,   7,  13,  22,  43, -14,  11, -11,   1,  -3, -32, -23,   1,  19,  -1,   0,   3, -17,  -4,  31,  10, -22,   3,  -7,  -1, -12, -11, -11,  19,  -1,  -7,  -6,  -2,  38,   3,  58, -17, -12,  -2,  -6,   0 },
      {   7,   9,  25,  43,  25, -22,  15,  -4,   2,   6, -26, -15,  47, -20,  15, -10,  -4, -13, -22,  11, -14,  16,  -8,   5,  -1,  -9,  -4,  38, -31,   3,  -8,   6,  -3,  -7,  30,  33,  -3, -12,  -3,   1,   1,  18,   9,   9,   7,  -2,  -5,  -2,  35,  -1,  19,   8,   0,   2,  -1,  -3,  -9,   1,   0,   3,   6,  -2,   2,   3 },
      {  -3,   0, -12,   9,  25, -11,   1,  -1,   0,   3,  18,  20,  -3,   3,  -4,   2,  -8, -12, -19, -26,   0,  18,   6,  -5,  -9,   1,  10,  -6, -21,   4,  11,  -2,  12,  -1, -14,  20,   1, -11,   7,  -1, -30, -40,  20, -20,   4,   0,  -2,   6, -35,  39,  11,   9, -19,  -1,  -3,  -1, -62,   1,   9,  24,  11,  -5,   1,   0 },
      {  -1,  -2,   1, -10, -38,  19,  -4,   3,  -1,  -1, -10, -31,  -3, -10,   5,   3,  -2,  -4, -18,  -1,  39, -37,  18,  -4,  -4,  -1,  -6,  13,  22, -29,   1,   2,  -1,  13,  -5,  34,  25, -13,  -6,   7, -11,  -6, -19,   6,  -3,  16,  -9,   4,  27,  18,   1,  36, -33,   6,  -4,  -2,  -6,  12,  -9,  34,   3, -10,   4,  -2 },
      {   4,   5,  12,  19,   0,  44, -33,  18,   0,  -2, -12, -13,  29,  10,  -6,   3,   1,   1,   9,  28,   2,   8,  -3,   8,   4,  10,  11,  -4, -27,  30,   3,  -2,  13,  -4,   3, -20,  -4,   8,  17,  -7, -38, -18, -22, -45, -11,   6,   2,   1,   4,  13, -10,  -3, -16, -17,  -5,  -3,  30,  -5, -21,  14,  11, -16,  -3,  -1 },
      {   0,   1,  11,  24,  10,  24, -18,   2,  -5,  -4, -14, -28,  11,  29, -11,   5,   0,   3,  14,  42,  15, -19,  14,  -4,   3,   9,   8,   2,  24,   8,  -1,  -3,  -5, -10,  -8,   4, -30,  24,   0,   1,   4,   3,  27,  12,   1,  -8,  12,  -7, -14, -12,   6,  -9, -10,  20,   2,   9, -46,  31,  31,  -1, -30,  10,   1,  -1 },
      {  -1,  -2,  -6, -33,   4,  21, -11,   9,   2,  -2,   6,  29,   4,   6,  -2,  -1,   2,   5,  19,  -6,  32, -17,  10,  -9,   3,   1, -26,   9,   1,  -3,   0,  -2,  -1, -29,  25,   7, -11,  -2,  15, -11, -25,  19,  25,  -9,   5, -29,  13,  -5,  49,  20,   6,  -2,  38, -15,   0,   5, -27,  -9,  -8, -11,  25,  10,  -3,   4 },
      {   3,   2,   4,  18,  44,   7, -27,   9,   5,   3,   9,   3,  12,  10,  -9,   2,  -1,  -3,  -7, -33, -35, -34,  33, -18,   1,  -5,   0,   2,   8, -31,  22,  -7,  -1,  -4,  -2, -12,   2, -23,  17,  -7,  -8,  10,   9,  20,  20, -16,   1,   1,   3,  -1, -30,  -3,   1,  -3,   0,   7,  36,  10,  16,  32, -10,  -9,  -8,   2 },
      {   2,   0,  10,  21,  -9, -17,  11, -10,  -2,  -8, -11, -47, -15, -22,  13, -12,  -4, -13,  -2,  11, -14, -15,   7,  -7,   8,  15,  28,  -9,  22, -13,   7,  -4,   4, -12, -32, -26,   5, -11,  10,   2, -11, -16,   1,  -8,  -2, -27,   0,   4,   9,  -6,   2,  10,  27, -35,  -7,   2, -17, -20,  19, -24,  42, -11, -11,   3 },
      {   0,  -3,  -2, -11,  -1,  19, -22,   3,   3,   6,   3,   5,  28,   4,  -8,   7,  -2,  -8, -25, -26,  -6,  11,  -7,   2,   1,   2,  20,  40,  11,  -3,  -6,   1,  11,  19,   0, -32,  20,  24,  -5,   1,  14,  -1, -35,   9, -12,   8,  13,  -7,  14, -31, -14, -10,  20,  -5,   6,   3, -51, -32,   3,  25,  -3,  11,  -5,   4 },
      {  -5,  -5,  -2, -12, -40, -25,   8,  -8,  -1,   3,  -1,  11,  -5, -55,  17, -13,  -6,  -4,   0,   3, -14, -35,   1,  -4,   3,   9,   8,  17, -20,  19,  -5,   1,  -1,  -8,   7, -29, -36,  11,  15,  -8, -11,  11,   1, -14,  10, -15,  -3,   3,  -4,   4, -26, -14,  -5,  19, -15,   2, -13,  16,  -6,  37, -24,   4,   0,  -6 },
      {  -1,  -2,  -3, -13, -21,  16,  14,  -3,   1,  -2,  -3,   3,  33,  11,  12, -11,   4,   8,  11,   8,   7,  34,   2, -10,   5,  -9,  -7, -21,  -5,  24,  17, -10, -12, -11,  -1, -23,  27, -45,  33,   3,   4,  12, -22,  26,  34, -24,  -6,  11, -13, -14,   2,  33,  -3,   9, -21,   8, -21,   8,   4,   2, -13,  -2,  -6,  -3 },
      {   0,   0,  -5,   9,   4, -25,  31, -11,  -1,  -1,  -1, -15,  -9,  34, -11,   9,  -1,  -4,  -5,   4,  27,  35, -27,  13,   2,   7,  21,  23,  48,  12,  -1,  -1,   1,   5,   0,  -9,  -2,   4,  17,  -4,  -4,  10,  21,   1,  15, -18,  13,   5,   9,  34, -20,  -5,  11,   0,   1,   4,  23,   1, -22,  45,  -2,   2,  -2,   0 },
      {   2,   4,   7,  13,  25,  -9, -17,   0,  -2,   0,  -6,  -9, -35,  -8, -36,  18,   0,   2,   5,   5,  13, -14, -33,  14,   3,   6,  -7, -23, -23,  15, -20,   7,  -5, -14, -10,  -7,  16, -13, -14,  -3,  -4,  21,  -5, -10,  34,  12, -14,  -7,  19,  -1, -39,  31,   2,  26,   1, -10, -16, -31,  21,   5,  -4,  11,   8,  -2 },
      {  -2,  -3,  -2, -10, -32,  -7, -30,  25,   1,   2,  -1,   7,  47, -26, -31,  16,  -2,  -2,   4,   7, -38,   3, -30,   9,   2,   4,   5, -25,  17,  -8, -13,   3,  -3,  -6, -26,   2,  15,   7, -16,  -2,  -5, -12,  26,  29,  10,  -4,  -4,  -9,   4,  33,  20,   5,  22,   2,  10, -12,   5,  16,  -9,   6,   1,   6,   2,  -4 },
      {   1,   0,  -2,   7,  -7,  21,  23,  -4,   2,   4,   6,   1,  14,  25,  39, -22,  -1,   0,  -2,   6, -20,   8,  22, -11,   2,   9,   5, -13,  -3, -10,  11,   2,  -4,  -8, -30,  -6, -13,   6, -15,   7, -12,   0,  17,  14, -13,  19, -23,   1,  26,  31, -21,  -1,   2,  36,  -2,  -4,   0, -54, -14,  -8, -20,  26,  10,  -2 },
      {   1,   1,  -1,  12,  12, -14, -16,  -1,  -2,  -3,   3,  -9, -22,  -2,  -6,   4,   1,   2, -13, -13,   2, -12,  19, -10,   0,  -3,  13,  24,  -9,  22,  -2, -14,   9,  10,   8, -28,  17,  30,   9, -11, -13, -14, -17,  20,  12, -13,   9,  -2,   6,  31,  37,  37,  -2,  13,  -9,   9,  13,   9, -22, -41, -24,  38,   4,   1 },
      {   3,   5,   2,   1,  12,  34,  26,  -8,   0,   1,   2,   3,   3,   0,  40, -22,   1,   4,   0,  -6, -20,  -3,   1,   0,  -3,  -5, -10, -14,  14,   5, -41,   5,   4,   4,   5,  10,  12,  29, -25, -11,   2,  -9,  -8, -31,  28,   1,  22, -12, -15,   3,  -4,  39,  37,  -4,   0,   6,  10,  24,  19,  24,   2,  11,  -7,   7 },
      {  -2,  -3,  -1,  -2, -33, -52,  12,   5,   1,   1,   2,   3,  38,  41, -23,  25,   3,   4,   6,   5,   0,   0,  21,  -4,  -2,  -2,  -1,  -1, -18, -39,   8, -10,   3,   5,  12,   1, -14,  15, -14,  -3,  -2,  -2, -15, -43,   2, -17,   8, -12,  -3,  -5, -17,  31,   1,   8, -10,   1,   5,  -8,  20, -11,   1,   9,  -6,   0 },
      {   1,   2,   1,   2,   9,  14,  39,  -9,   1,   6,   6,   3,   2,  34,  -4,  19,   0,   2,   1,   4,  13, -37,  -8,  17,   2,   6,   6,   8, -42, -14, -11,   4,  -2,  -5, -13, -39,  -4, -15, -44,  14,  -4, -10, -12,  21,  13, -15, -28,   1,   1,   0,  24, -19,   7, -16,   1,  -6,  -5,  27, -20,  17,  11,   1,  -3,   4 },
      {   0,   0,  -1,  -5,   5,  20,   6, -13,   2,   4,   4,  15,  12,  14, -15, -10,  -3,  -7,  -4, -19, -26, -23, -29,   7,  -1,   2,  -2,  20,  45,   5, -32,   6,   5,   3,  10,   0, -35, -13,   3,   8,  -4,  -9, -14, -12,  10, -31, -18,  -1,  -3,   2,  10,  16, -31,  20, -12, -14,  -2, -21,  -3, -33,  -2, -39,  11, -12 },
      {   1,   3,   1,   4,  19,  18,  -6,  24,  -1,   0,  -2,  -5, -11, -27,  -7,  11,   0,   1,   0,   2,   6,  24, -12,   5,  -2,  -2,  -2,   6,  12, -30,   3,  12,  -2,   6,   3,   5, -25, -29,  -7,  28,  -2,  -6, -13, -19, -36, -23, -20,  18, -10,  -2,   1,  11,  31,   2, -47,  -4,   3,  11,  -5,   5, -27,  45,  -8,  -1 },
      {   0,   1,   1,   1,  -3,  -1, -10,  -7,  -2,  -1,  -1,  -3,   0,  -7,  -4, -13,   1,   3,  -1,  11,   6,   9,   3, -47,  -1,  -2,   2,  -4, -10,  -3, -40,  -5,  -1,  -1,   3,   1,  21, -22, -41,  12,   2,   6,  11,  -2, -12, -48,  32,  -2,  -9,  10, -22, -10, -33, -35,  22,   1,  -3, -18, -17,  -2, -31,   7,   4,  -3 },
      {   2,   2,   2,   4,   8,  32,  46, -30,   0,   0,  -1,   2,   7, -20, -22,   9,   1,  -1,   0,   6,  -7, -20, -35,  17,  -2,  -5,   2, -13, -15, -35,  21, -30,  -1,   4,  -5,   4,   9,  -4,  38, -27,  -1,  -3,  -3,  -3, -27,  -3,  38, -18,   2,   4,   0,   5,  -4,   2,  19,   0,  -4,  -4,  -5,  -7, -25,  11,   9,   0 },
      {   0,  -1,   1,  -1,   1,  15, -12, -17,   0,  -2,   0,   2,  -4, -24, -12, -22,   2,   1,   2,   1,  25,  35,  -9, -43,   3,   2,   8,  19,  -1, -23,  33, -29,  -1,   1,  -1, -22, -42,  -5, -29,   2,   2,  -6,  -4,  -5,  32,  21,  -4, -24,   1,   3,   9,  17,  13,   6,  27, -23,   9,  11,   4,   1,   2, -19,  10,   0 },
      {   2,   2,   2,   1,   5,  10,  31,  76,   1,   0,   2,   2, -12, -22,  35,  58,   1,   0,   2,  -4,   0,  14,   9,  10,   2,   3,   4,   6,   5,  -4,   1,  -6,   1,  -1,  -2, -10, -15,   2,  -3,  -7,   3,   1,   1,   3,  20,  -6,   8,  -2,   7,   0,  -4,  10, -16, -11,  26,   5,  -8, -10,   2, -12, -20, -27,   3,  -3 },
      {   0,   1,   1,   1,  -2,  27,  25, -18,  -1,   1,   1,   2,   5, -20, -32,  20,  -2,   0,   0,  -6, -15,   2,   2, -13,  -1,   1,  -5,  -2,   9,  22,  31, -37,  -2,   2,   8,  16,   8,  43, -32,  11,   3,  16,  14,  -2,  18,   1, -22,  39,   7,   3, -13,  -3, -17, -28, -35,  14,  -2,  -7,   3,  -6,   5,   2, -23,  -5 },
      {   1,   2,   0,   4,   3,  -8,  20,  48,   0,   0,   1,  -6,  -2,  16,  -4,  -6,  -1,  -3,  -4,   2,   2, -15, -32, -67,  -1,  -3,  -1,   0,   1,   6, -32, -46,  -1,   1,   0,  -1,   4,  -7,  11, -30,   1,   1,  -2,   6, -22,  23, -11,  -7,  -5,  -5,   2, -12,   3,  17, -29,  -4,  -1,   5,   1,   6,  22,   4,  -8,   0 },
      {   0,   1,  -1,  -1,   0,   0, -15,  12,   0,  -1,   1,   3,   1,   4,   9, -33,   0,   1,   2,   0,  10,   9, -35,  13,  -1,   0,   0,   1, -10, -52, -10,  24,  -2,  -3,  -7, -14, -21,  13,   5, -20,   0,   4,   5,  15,  36,  24,  21,  14,   0,   4,  -2,   0, -42, -28, -31,  33,  -3,  -5,  -2, -13,   6,  17, -27,  -4 },
      {   1,   1,   2,  -1,   4,  21,  32,   7,  -1,  -2,  -2,  -3,  -2, -20, -40,   5,   1,   0,  -1,   4,   4,  16,  29, -29,   0,  -1,   3,  -1, -18, -16, -17,  51,   3,  -1,  -2, -12,   9,  24,  18,  27,   3,  -2,   5,  12,   7,  -6,  32,   1,  -3,   1,   4, -10,   0,  37, -22,   2,  10,  -4,   3,   3,  31, -18,  22, -16 },
      {   0,   0,   0,   4,   7,  -1, -18, -16,  -1,  -1,  -1,  -2, -11,  -5,  15,   1,   2,   3,   5,   3,  13,  26,  18,  24,  -1,  -2,  -2,  -7,  -9, -28, -17, -24,  -1,  -1,  -2,   2,  20,  25,  -9, -45,  -2,   2,   3,  11,  -7, -26, -23, -25,   0,  -2,  -1, -13,  10,   3, -45, -17,  -5,  -2,  -6,   6, -14, -41,   9, -44 },
      {   0,  -1,   0,  -1,   0,  12, -14, -26,   1,   2,   2,   4,   2, -17,  20,  38,   1,   1,   0,   0,  -4,  17,  16,  19,   2,   0,   5,   4,  10,   4, -16, -27,  -2,   0,  -2,  -7, -13, -28, -20, -17,   0,   1,   0,   7,  -4, -10,  27,  -2,   0,  -1,  -1, -10, -33,  42,  -1,  -4,   9,   4,   7,   2,  59,  37, -23,   0 },
      {   0,   0,   1,  -1,  -5,   3,  18,  -8,  -1,  -2,  -3,  -1,   1, -14, -36,   3,   1,   1,   1,  -2,  -3,  11,  37,   7,   2,   2,   0,   3,  14,  14, -17,  17,   0,  -4,  -3,  -1,  -9, -30, -11, -32,  -4,  -3,   1,   1,  -1,  26, -12, -54,  -1,   5,   0,  -6,   1, -13, -24,  31,  -3,  -6, -13,  -6, -19, -10, -20,  57 },
      {  -1,  -1,   0,   0,   1,  -4, -22, -26,   0,   0,  -2,   0,  -1,   0,  15,  25,   1,   1,   2,   4,   5,   3, -13, -10,  -1,  -2,  -1,  -5, -17, -19, -28, -24,  -1,   0,  -5,  -3,  -2,   3,   7,  28,  -1,  -3,   3,  -1,  -9,   7,   9,  38,   5,  11,   7,   4,  27,  32,   3,  48,   0,  -5,  -6,  -3, -18, -52, -28,  35 },
      {   0,   0,   0,  -1,  -3,   5,   8,   8,  -1,  -1,  -1,  -1,   0,  -7, -30, -25,   1,   2,   2,   2,   1,  22,  29,  17,  -1,   0,   0,   1,   0, -19, -38, -19,  -1,   0,  -2,  -5,  -7,   8,  14, -36,  -1,   1,  -6,  -2,  -1,  -9, -39,  59,   1,  -7,  -1,   1,  -5,  -1,  41,   7,   0,   6,   9,   4,   8,  15,  35,  25 },
      {   0,   0,   0,   1,  -5,  -2,   2, -27,   1,   1,   2,  -1,   5,   5,  -5,  31,   1,   1,  -1,   0,  -5,   0,  29,  -6,  -1,  -2,   0,   1,   2,  -1, -47, -11,  -1,  -3,  -3,  -3, -21, -20,  35,  17,   1,   0,   6,  -1,  16,  53,   7,  13,   3,   1,   4,   3,  15, -40,  -2, -35,  -4,  -2,  -5,  -4, -17,  12,  -9, -45 },
      {   0,  -1,  -1,   1,  -2,  -6,   5,  10,   0,  -1,   0,  -1,   3,   5, -17, -14,  -1,  -1,  -2,  -2,  -3,  -9,   7,  11,   1,   2,   3,   7,   5,  21,  16,   4,  -2,  -2,  -2,  -3,  -9, -37, -45, -35,  -3,  -3,  -6,  -4, -14,  21,  40,  37,   4,   1,   8,   5,  25,   8,  -4,  43,   0,  -2,   1,  -1,  -7, -16,  21, -54 },
      {   0,  -1,  -1,  -1,  -2,   3, -14, -35,   2,   2,   1,   0,   9,   3,  27,  54,   0,  -1,   0,   2,  -2,   0, -23, -40,   1,   0,   1,  -2,   5,  -1,   7,  34,   0,   0,  -1,   1,  -6,   1,   3, -33,   1,  -2,   0,  -1,   2,   2, -15,   7,  -1,  -5,   0,   2,  -2, -21, -17,  22,   5,   3,   7,   0,   4,  10,  70,  15 },
      {   0,  -1,   0,  -2,   1,  -7,  -5,  15,   0,   0,   0,   2,   1,   2,   4, -16,   1,   1,   1,  -2,  -3,  -5,   6,  33,   0,  -1,  -1,   0,   0,   1,  -6, -53,  -1,  -3,  -3,  -4,  -3,  -5,  -5,  45,   2,   2,   6,   4,  10,  12,  34, -10,   0,   0,  -3,  -5,  -4, -13, -42,   1,   1,  -3,   2,   4,   7,  -1,  71,  33 },
      {   0,   0,   1,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,  -5,  -2,  -1,   0,   0,   0,  -2,  -1,   6,  -3,   3,  -1,  -2,   0,   0,   0,  -2, -12, -19,   1,   3,   2,   2,  -2,   0,  16,  35,  -1,  -3,  -1,  -6,   0,  -4, -39, -47,  -1,   0,  -2,  -2,  -4,  -5,  25,  80,   4,   2,   4,   5,  16,  19,  11, -50 },
    },
  },
  { // 26
    {
      {-113,  36,   9,   4,   2,   2,   1,   0, -17, -15,  12,   0,   2,   0,   1,   0,  24, -22,   0,   1,   0,   0,   0,   0,  13,  -6,  -4,   2,  -2,   0,  -1,   0,   8,  -3,  -4,   2,  -1,   0,   0,   0,   2,   2,  -4,   0,   0,  -1,   0,   0,   3,  -1,  -1,  -1,   1,   0,   0,   0,   1,   0,  -2,   0,   0,   0,   0,   0 },
      {  16,  11,  -3,   0,  -3,   1,  -1,   0, -91,  44,  14,   2,   1,  -1,   1,   0, -38, -27,  25,   1,   4,   0,   1,   0,  24, -36,   7,   7,  -1,   1,  -1,   0,  10, -16,  -6,   9,   0,   2,   0,   0,   6,  -3, -11,   4,  -2,   1,   0,   0,   2,  -1,  -4,   0,   2,   1,   0,   0,   1,  -1,  -5,   1,   1,   0,   1,   0 },
      {  17, -25,   6,   6,  -1,   2,   0,  -1,  -9, -21,   2,   4,   2,   1,   0,   1,  27, -29,  -8,   8,  -1,   0,  -4,   3,  43, -18, -41,  15,  -7,   5,  -1,  -1,   3,  54, -44,   2,  -3,  -1,   0,  -2, -24,  43,  -4, -17,  10,  -5,   1,   0,  -5,   5,  15, -21,   4,  -3,   0,   0,   3,  -2,  14, -11,  -3,   2,  -2,   0 },
      {  22,   4, -11,  -5,   0,  -3,   0,   1, -45,  -5,   9,  -2,   4,   3,   2,   0,  59, -35, -13,  -2,   0,   4,   3,  -1,  14,  56, -17, -14,   0,  -2,   0,   1, -26,  15,  35, -18,   3,  -3,   0,   0,   8, -29,  26,   7,  -7,   2,  -1,   0,   2,  -6,  -6,  21,  -4,   2,   0,   0,  -2,   5,  -8,   9,   1,  -2,   1,   0 },
      { -12,  31,  -4, -11,   3,  -3,   0,   1,  22,  40, -26,  -9,  -2,   1,  -1,   0, -31,  13,  22, -13,  -1,  -5,   2,  -4,  25, -17, -22,  -8,   4,  -1,  -1,   2,   1,  62,  10, -35,   0,  -6,   0,   0,  -3, -22,  53, -10,  -3,   8,   0,  -1,   4, -14,  -2,   5, -11,   2,   0,   1,  -1,  -1,   5,  -1,   2,  -3,   1,  -1 },
      {  28,  91, -31,  -6, -10,  -1,  -3,   0,   0,  12,  11,  -8,   1,  -2,   0,   0,  29,  -1,   3,   0,  -5,  -4,  -4,   0, -43,   3,   1,   7,   2,   2,   0,  -1,   1,  -4, -18,  -1,   3,  -1,   2,  -1,  -9,  21,  -5, -23,   8,   1,   2,  -1,   4,  -2,  21, -26,  -4,  -1,   0,   1,   3,  -6,  19,  -4,  -5,   1,  -1,   0 },
      { -16,  -8,   2,  12, -10,   8,  -2,  -2, -48,   8,  -9,  16,   5,   0,   2,   1, -11,  61, -38,  -1,  -1,  -1,  -5,   5, -26,  26,  -3,  11,  -5,   4,   2,  -2, -17,  21,  -3,   3,  -6,  -4,   6,  -4, -32,   2,  11,  13,  15,  -5,   1,   2,  -2, -31, -18, -31,  20,   1,   2,   1,   5,  -1,  -2, -31,   8,   5,  -1,  -1 },
      {  -3,  13,   4,   9, -16,   6,   2,   0,  32,  10,  18,   1,  -2,  -4,   1,   0, -42, -40,  16,  36,   2,   3,  -8,   7,   3,  28, -30,  32,  -5,   2,  -4,  -4, -56,   5,  11,  23,   7,   1,   6,  -5, -12, -26, -28,   8,  11,  -1,   3,   3,   5, -11,  -3, -16,  14,   2,   1,   1,   2, -10,  -9,  -3,   1,   6,   0,   1 },
      {  22,  32,   0,  -3,  -2,  -4,  -4,  -3,  15, -32,  15,   4,   6,   2,   0,   1,  19,  -4,  16,  -6,  -1,   3,   2,   2,   7, -34,   7,  13,  -3,   9,   4,   2,  12,   9,  -7,  10, -12,   0,  -2,   0, -20, -20,   4,  48,   3,   0,  -1,   2,  -9,  -2, -70,  13,  16,  -2,   4,   1,  -6,  26, -13, -30,  10,  -2,   1,  -2 },
      {   5,  18,  -1,   8,  -6,   3,  -1,  -2,  18,  24,  18,  -1,  -7,  -6,  -4,  -2,  10,  14,  -8,   1,   1,   0,  -4,   3,  47,  11,  38,  -1, -13,  -4,  -4,  -2,  -4,  -4,   5,  32,  -8,   8,   0,   0, -31,  41,  17,  13, -12,  -9,  -1,   0, -11, -39,  28,  39,   8,   1,   0,   0,  -6, -18, -35,   9,   7,   5,   2,   1 },
      {   7,  13,   0,  -1,  -3,   0,   0,   0, -14, -64,  25,  12,   6,   3,   1,   2, -52, -13, -20,  20,   4,   9,   1,   2, -28,  -2, -15,   4,   8,   1,   2,   0,  12,   1, -21, -20,   5,  -1,   1,  -1,   3,   9,  36, -15, -10,  -1,  -2,  -1,  11, -29,  12,  34,  -7,   5,   1,   0,  -3,   5, -27,  18,   2,  -2,   2,   0 },
      {   9,  37, -15,   5,  -2,  -1,  -3,   0,   1, -22, -15,   6,   1,   4,   0,   1, -43,  -1, -45,   6,   5,   5,   1,   2,  48,  25, -14, -38,   0,  -3,   3,  -2,  30,  -8,  37,  -4, -11,   1,  -3,  -1,  -7,  15, -25,  14,  -3,   2,  -1,   1,  -9,  31,  -4, -14,   1,  -7,   1,  -2,   0,   2,  19,  -8,   0,   1,  -1,   0 },
      {  14,  14,  70, -24, -11,   0,   9,   0, -16, -26,   3,  -5,  -5,   4,   5,   5,  15,  28,  16,  30,  -7, -14, -12,  -1,  -1, -20,  -1,   5,  13,  -6,  -9,  -4,  17,  17,  44,   9, -10,   5,  -4,  -5, -20, -17, -16,   5,   7,  13,   8,   1,  16,  -1,  24,  -7, -12,  -1,   1,   2,  -1, -21,   3,  23,   1,   4,   0,   2 },
      {  11,  16,  60, -14,  18, -15,  -7,  -2,   8,   5,  55,   3,   1,  -6,  -7,  -3,  -6,  20, -20, -18, -10,   2,   1,  -5,  36,  12,   6,  -3,  -6,   0,   1,   2, -22, -20, -21,  -9,  22,  -1,   5,   6,  25,  -9,  13, -33,   0,   1,  -4,   1,  -8,   0, -16, -19,  -2,   6,  -4,  -1,   7,  11,  13, -18, -13,  -3,   0,  -2 },
      {   3,  12, -25,  -1,   4,   0,   3,   1, -14, -57, -35,   6,   8,   6,   3,   3,  -9,  13,  15, -34,   9,   0,   4,  -2,  32, -16,  41,  -2,  -6,   2,  -1,   1, -60,   4,  -9,  18,  -1,  -1,  -1,   1,   1, -24,  -3, -12,  10,  -4,   3,  -1,  15,  -4,  17, -11,  -6,   4,  -2,   1,   8,  -3,  10,  12, -10,   2,  -2,   1 },
      {  -4, -19,  -5,  -4,  10,   0,  -3,   0,   0, -19,  48, -38,  10,  -8,  -1,  -4, -22, -23,  31, -42,  -9,  -3,  -2,   0, -20,  19,  17, -19, -26,   0,  -2,   2,  24,  23,  27,  14, -16,  -6,  -3,   1,  -5,  11,  16,  16,   9, -10,  -1,   0,   0, -15,   9, -29,  12,   1,  -5,  -1,  13, -11,  24,  -4, -14,   7,  -3,   0 },
      {  -5,  -3, -37,   8, -10,   0,  -3,   1, -10, -14,  29, -25,   0,   1,   0,   0,  -1,  64,  43,   7, -16,   8,   0,   0,  28,  32, -30,  38,  -7,   6,   4,  -1,  17,  -6, -14,  -5,  11,  -3,   0,  -3,   7,  -6,  -8,   5,   0,   2,  -3,   1,  -2,  24,   1,  18,  -3,  -5,   2,   0,  -6,   2,   9,  23,  -1,  -3,   2,   0 },
      {  -4,  11,  32, -37,  -8,  -1,   4,  -3, -11,  -1, -15, -13,  -1,   7,   5,   1, -15,  -1,  11, -14,  -2,   7,   7,   5, -18,  13,   0,  13,  -5,   8,  -2,   2, -45,   7, -11, -21, -29,   9,  -1,   1,  17,  45,   1,  35, -34, -11,  -3,  -1, -19,  35,   4,  24,  10, -17,   7,  -2,   9, -10,  14,  -4,  17,  -4,   0,  -1 },
      { -10, -23, -47, -52,  -6,   6,  13,   5,   1,  -8,  35, -20,  -9,   1,   4,   3,   1,   6,  -6,  29,   6, -18,  -8,  -2,  12, -17,  18,  -6,  31,  -8, -11,  -2, -18,  -1,  15, -19,  -3,  18,  -2,  -1,  -5,   5,  -4, -17, -19,  24,   8,  -1,  -5,  -5,  -4, -11, -11,  -4,   9,   2,   1, -10,  -4, -46,  15,  -4,   0,   1 },
      {   6,   9,   7,   8,  -4,   8,   5,   3,  -4, -18, -30,  21, -15,  -2,  -3,   0,  -4, -10,  23,   7, -18, -25, -12,  -6,   6,  36,  28,  26,  -1, -19, -14,  -4,  30,  22,  -1,  15,  18,  -7,   2,  -3,  60,   9,   3, -10,   1,  11,  -1,  -3, -15,  -5,  -3,  -3,   8,   7,   0,   0,  -7, -16, -13, -41,  17,   2,   1,  -1 },
      {  -4,   3,  -7, -53,  10, -10,   9,   0,  -2,  16,  -2, -28,  -9,   8,   2,   5,  -8,   4, -31,  12, -10,   6,   2,   0,  -4,  14,   9, -11,  -9,   1,  -1,  -2,   9,  36, -23,  45, -11,  -2,  -6,  -2,  25,  -5, -19,  -5,  32, -12,   6,  -1,  42,   7,  -8,  24,   6,  -1,  -3,   1,  -2,  33, -12,  10,  -3,   1,   0,   1 },
      {   0,   0,  21,  -7,   6,   1,  -2,  -1,  -7,  -5, -29, -38,  13,   4,   0,   4,  -8, -10,  25, -29,  -3,   6,  -1,  -1,  -2,  34,  -7, -11,  -7,  10,  -1,   3,  10, -15, -15,   2,   0,   9,   3,   2, -48,  -3, -22, -39, -20,  18,   3,   2,   3,  -4,  -6,  -6, -25, -10,   1,  -1, -26,  19, -38, -20,  26,  -5,   6,  -3 },
      {  -3,   3,  -2,  45,  11,   2,  -4,   2,  -7,   2,  37,  22, -20,  -2, -10,   1,  -5,  12,   1,  -1,  10,  -5,  -3,  -4, -25, -12,   4, -30, -20,  -4,  -1,   5, -23,  41,   8,  28,  -1,  -4,  -3,   1,  -4,  -7,  -8, -21, -14,  10,   1,  -3,  -3,  46,   3,  28, -22, -18,   0,  -7, -28,  -2,   5, -15,  22,  -5,   3,  -1 },
      {  -7,  -8,   1, -27,  -7,   5,  -4,   1,   1,   9,   4,  52, -10,  10,  -4,   1,  -9, -17,  13,  12,   6,   8,   8,   2,  -1,  38,  49,  14,   9,   2,  -1,   0,  20,  13, -20,  -4,  15,   5,  -1,  -1, -49,  -8,   6,  -1,  -9,   7,   4,  -1,   0,  12, -14,   6, -17,  -4,   6,  -1,   4,   9,  43,  19, -20, -10,   1,  -2 },
      {  -4,   3, -10, -39,  50, -15,  -4,  -2,  -7,  -4,  -5,   2,   5,  -4,   0,  -6,  -7,  -2, -12,  -8,  -6,  -3,   5,  -7, -11,  -7, -10,   8,   1, -19,   3,   5,  -3,   6,   2,  12,  35, -13,  -1,   6, -19,  -7,  -6,   5,  26,   5,  -3,  -3, -71,  10,   1,   0,  12,   5,  -2,  -2, -24, -31, -15,  27,   8,  -1,   0,   1 },
      {   1,   5,   7,   2,  30, -15,  -1,  -3,   0,   2,   6,  18,  27,  -4,   5,  -3,   4,  15,  42,  22,  53,  -1,  14,  -7,   1,  24,  -8, -48,  36,  14,   7,   0,   3,   8, -30,  -1, -12,  -1,  -9,   2,  12,  -4, -21,  14,  16,  -5,   4,  -1,  -9, -20,  14,   8,   7,  -2,   2,   2,  19,  -9,  -4, -17,   4,   8,  -5,   2 },
      {   1,   2,   7,  21,  33,  -8,  -2,  -3, -11,  -7, -26, -44,  -9,   6,  -4,   2,  -3,  -4,   4,  40, -17,   4,   0,   4,  -5,  -3,   3,   1,  -2, -10,  14,   2,  -7, -17,   7,  14,  25,  -2, -11,  -2, -23,   3,  20,  -8,   7,  -4,  -1,   0,  -2,   1,  -4,  41,  -1, -11,  -2,  -2,  34,  -6,  34, -43, -29,   6,  -6,   0 },
      {  -3,   1,   4,  29,   2,  -7,  -4,   5,  -4,   0,   4,  -9,   1,  15,  -3,   1,  -6,   1,  21,   9, -20, -14,  -8,  -6,  -3,   5,  20, -14,  37,  -9,  -2,  -2, -27,  15,  20, -22,   5,  11,   0,   0,  -2,  56,  -7,  -7,  39,  11,   4,  -6,  -7,   8, -33,  -6,   0,  17,   3,   1,  -3,  38, -22,  31, -19, -10,   5,   0 },
      {  -2,  -2,  -9, -28,  18, -15,  -3,  -3,  -2,  -7,  -4,  44, -47,  -6, -10,   2,   3,   7,  37,   8, -34,   5,   3,  -3,   1,  -3, -24, -31, -18, -17,  10,   3,  -9, -21,  10,  -7,  11, -12,  -2,   5, -16,  18,   7,   4,  -7, -21, -11,  -1,  35,  -1,   8, -16,  25,  -1, -13,  -8,  -3,  17, -13,  -9,  13,  13,  -1,  -2 },
      {   1,   3,  -9,   2,  20,  13,  -9,   1,   0,   0,  -1,  12, -14,  25, -12,   3,   3,   5,   1,  12, -42,   4,  -2,   0,  -3,  -3, -10, -16, -24,   7,  -1,   7,   0,  10, -10,  10, -27,  12,  -1,   3,  15, -10, -23,  -1, -30,   9,   5,  -4, -33, -35, -16, -14, -34, -16,   6,  -2,  55,  -6, -13,  23, -10, -20,   3,  -4 },
      {  -4,  -4,   5, -21, -64,   5,  -4,   4,  -2,  -6,  -4,   2, -40,  13,  -3,   7,   1,   1,  15, -10,  31,  -3,   3,   8,  -3,  -2, -20, -21, -15,  17,   8,   1,   1, -17,   9,  22,  -5,  -7,   3,  -9,   2,   1,  20, -29,  39,   5,   4,   1, -49,   1,  -2,  13,   2,   1,   0,   5,  18,  13,  -2,  -5,  -9,  -2,  -2,  -2 },
      {  -2,  -1,  -7, -25,  31,   5,   2,  -6,   1,   3,  -7,  33,  48,  -2, -10,   0,   3,   9,  13,   7,  22,  22, -24,   7,  -3,  -5, -18,  20, -30,  27,  -8,  -2,  -4,  -9,  39,  24,  -4,   1,   1, -12,  13,  35,  16,  -9,   0,  15,   3,   5,   7,   2,  -8,   1, -30,  16,   7,   3, -11,  12,  -1, -14, -27,  -1,   7,   0 },
      {   1,   3,  -2,  -6,  -1,  -4,   0,   5,  -1,  -3,  -3,   6, -14, -31,   1,  -5,  -1,   5,  -6, -44,  31, -32, -12,  -7,  -7,   1, -27,   5,   9,  -9, -20,  -1,   3,  10,   8,  11,  43,  26,  -1,   0,  -5,  17, -23,  -5, -24,   9,   7,   6,  25,  -6, -31,  22,  11, -15,  -3,  -4,  42, -13,  -5,  14, -18,   9,  -1,   5 },
      {   0,   0,  -3,  -2,  -8,  -8,  -1,  -1,   3,   2,   3,   7,  39,   5,   8,   1,   5,   3,   4,  18,   7,  18,   5,   3,  -1,  -8,  29,  11, -31,  -4,  15,  -4,  11,  20,  24, -29,   7,   4,  14,   4,  -4,  -8, -12, -43,  -5, -24,   4,   9, -10,  25,   0,   2,  40, -22, -16,  -2,  42,   7, -31,   3,  12,  19,  -6,  -4 },
      {  -1,   2,  -2,   5,   2, -26,   6, -12,  -8,  -8, -13, -18, -30, -32,  -1,  -1,   1,   4,   3,  27,  14,  15,  -7,  -6,  -4,  -1,   9,  -8,  -2,  28, -11, -10,   2,  15,   0,   3,  -3,  29,  12,   2,  11,  -7,  -3,  -9, -46, -20,   8,  13, -29, -15,  -7, -17,  24,   6, -14,   3, -43,  17,   9,  12, -40,  11,   4,   0 },
      {  -4,  -5,  -5,   1, -41, -69,  14,   0,  -1,   1,   1,  16,  18, -27,   1,  -1,  -3,  -2,  -3, -15, -43,  29,  11, -12,  -5,  -3,  -1,  -4,  15,  10,   3,  -8,   0,  11,   4,   4,   7,   4,   2,  15,  -5,  -2, -15,   9,   8,  -3,  -8,   8,  -4,   5,  10,  14, -33,  16,   1,   2,  11, -21, -15, -27, -18,  -7,   4,  -4 },
      {   0,  -2,  -1,   4,   5, -33, -11,  -4,  -2,  -4,  -7, -19, -20, -48,   0, -11,   1,  -4,  -5,  25,  26,  -2,  -7,   0,   4,   2,  19,   9, -12,  24,   9,   6,  -2,   0,   2,   3,  12, -38,   9,   0,   3,  14,  16,  16,   3,  17,  -7,   7,   5,   3,  -8, -27, -34, -12,  -1,   4,  29,  14,   0,  23,  50,  -8,  -7,  -6 },
      {   0,   1,   7, -13, -30,  48, -26,   9,  -2,  -2,  -3, -17,  10,  -8, -10,  -6,   4,   6,   3,  12,   6,  34,  -2,   4,  -3, -10,  11, -32, -22, -11,  17,   6,  -7,  20,  -9, -16,  57, -19,  -7,   9,   9,   8, -27,  27, -11,  12, -16,   4,   2, -12,   2,  -1, -12,  20,   2,  -6,  -9,  -1,  -4,  -5, -16,  10,  10,   0 },
      {   2,   4,   1,   6,  30,   6,  10,  -5,   0,   0,  10,  -7, -54,  41, -10,   9,   1,   0,  -4, -24,  30,  35,   0,  10,   0,   2,   6,  29,  25,  10,  27,  -4,   9,  13,  13, -10,   7,   3,  30,   2,   1,   2, -10,  13,   3, -19,   0,  15,   7,  11,  11, -10, -24,  19,  -7,   9,  13, -10, -27, -17,  -8,   8,   2,  -1 },
      {   0,  -3,  -1,  -3,   3,  -2, -39,   7,   0,  -2,  -5,  -1,  -8,   4, -33,  -3,   2,  -6,  -1,   0, -18,  21, -31,  -3,   3,   0,   9, -17,   3,  56,  -2,  -6,   1,  -5,  -8, -15,  -2,  12,  12, -11,   4,  -1,   4, -11,  24,  13,  13,  -5,  23,   7, -23,   9,  25, -13,  15,   3, -12, -62,   1,   7,  11,   9,   1,   3 },
      {  -2,  -2,   0,  -3, -13,  -4, -19,   1,  -1,   1,  -1,  -3,  28,   8, -41,   9,   1,   0,   2,  -4, -11,  11, -24,  22,  -1,  -3,  -8, -10,  53, -14,  23,  -3,   0,   4,  -4,  49,  21,  13,  15, -14,   5,  -8,  32,  10, -32,   0,   5,  -2,  -7,  17,  18, -21,  16, -13,   0,   2,  12,  23,  -2,   7,  14,  -4,   1,   2 },
      {  -1,  -3,   0,  -2,  -4, -13, -12,   0,   2,   0,   3,   8,   5,  25,   5,   1,   0,  -3,  -1, -14,  -7,   5,  10,  -9,   1,  -9,  -3,  12,   8,  12,  14,  -3,  -2,   3,  27,  -5,  30,  -7,  11,   1,  15,  14, -27,   0,   8, -18,  25,   4,  -4, -54,   7,  15,  -8, -58,  -3,   7, -28,  22,  37,  -4,  23,  -8, -15,   1 },
      {   1,   0,   3,   9,  -8, -24,  25,   1,   2,   0,   0,  -8,  -7,  29,  11,  -7,   4,   3,   4,   0,  19,  61,  26, -12,   0,  -2,  -3,   2,   5, -27,  -8,   6,   1,   1,   8,  22,   1, -21, -28,  -1,   6,  15,   7, -22, -21,  24,   5, -14,  12, -13, -28, -17,  28,   4,  29,  -7,   0, -23,   9,  13,  13, -14,  11,   8 },
      {   1,   3,   1,   9,  19,  11,  18,   9,   0,   2,   4,   7, -17, -20,  14,  12,  -1,  -2,  -5, -17, -16,  32, -10,   8,  -2,  -5,   4,   5,   0,  32, -22,  -9,   3,   5,   5, -12,  15,  26, -25,  -8,  -1, -12,  -7,  10,  14,  44,  -2, -13, -14,  -3,  42,   9,  29,   6,  23,  -3,  21,  43,   7,  -6,  20,   7,  10,   4 },
      {  -1,  -2,  -2,   0,  -3, -36,  19,  -3,   0,   2,   0,   0,  -1,  12, -23,   6,   0,  -2,   0, -10,  13,   2, -76,  13,  -1,   1,   2,  -1, -14, -39,  10,  -8,  -2,  -1, -13, -24, -24, -30,  19,  -6,  -4,  -7, -26,   1,  -3,  24, -12,  -6,  -7, -19,   1,  22,   7,   8,  -3,  -4,  -2,   9,  19,   3,  -2,  14,  -1,   1 },
      {  -2,  -2,  -1,   1,  -4, -43, -39, -12,  -1,  -1,  -2,  -4,  -4,  12, -37,  15,   1,   0,  -5,   5,  26, -18,  27,  22,  -2,   1,   3,  18, -32, -22,  14,  14,   2,   3,   2, -14,   2,  58, -34,  -7,   7,  -3,  -3,  10,  12,  11, -12, -18,   7,  -7,  11, -11,  -6,   1,  11, -10,  -9,   1,  -4,  -4,   2,  -1,  11,  -4 },
      {  -1,   0,   1,  -4,   4,  -5, -51,  44,  -2,  -2,  -2,  -6,  -3,   0, -40,  -6,   0,   2,  -4,   0,  11,   2,   8, -37,  -1,  -3,  -3,  20,  12,  -2, -47,  -9,   1,   0,  12,  -3, -17, -34,  -5,   6, -10,  -6, -13,  -6, -11, -30,  13, -12,  -6,  -1,  15,   9,   4,  28,   7, -10,  12,  21,   2, -12,   2,  -7,  12,  -2 },
      {  -1,  -1,  -3,   5, -11, -28,  11,  -2,   2,   4,   2,  -5,   9,  71,  15,  -6,  -1,  -1,  -3,  10,   4, -21, -23, -23,  -2,  -1,  -1, -14, -19,  29, -44,   9,   0,  -5, -11,   6,  38,   7,  -8,  19,  -3,  -1,  16,  24, -10,   0,   3,  11,   0,  10,   7, -11,  -5,  14, -17,  -6,   5,   4,  -1,   2,  13,  25,  -3,  -7 },
      {  -1,  -2,  -2,  -2,   4,  -5, -13,   4,   1,   2,   3,   2,   5,  15,   1,  26,   0,  -1,  -4,  -3, -11, -32,  34,   6,  -1,  -2,   0,   7,  14,  37,  33, -28,  -3,   7,  11,  14,  -9, -25,  -7,  -2,   1,   7, -14, -18, -26,  30, -51,  18,   1, -11,  -9,  10,  15,  30, -12, -13,   0,  -2,  16,   8,   8,  31,   3,  -6 },
      {  -1,  -1,  -2,  -2,  -7,   9,  -4, -54,   0,   1,   2,   2,  -4,  -4,  -7, -12,  -1,  -1,  -2,  -9,  -4,   7, -27,  27,   2,   1,   5,   9,  32,  17, -20,  29,   1,   4,   9,  -4,  -2, -24, -65,   2,   1,  -4,  -9, -17,   0, -37, -28, -13,  -7,  -2,  -7,   0, -11, -15,  -3, -13,   1,   0,  -7,   1,  -2,  20,   0, -10 },
      {  -1,  -1,   0,  -1,  -1,  -3, -41,  25,   1,   2,   1,   6,  -8,  -2,  32,  -9,  -2,  -1,  -3,  -4,  -2,  25,  -7, -26,   0,   4,  -1,  14,  12, -22,  11,   9,  -3,  -2,  -6,   7, -27,   6, -22,   0,   5,   1,  12,   5,   0,  37,  12,   5,  -3,   6,   1,  -7, -14, -24, -26,   2,  -6,   6, -13,  -1, -19,  62, -37,  -2 },
      {  -1,  -2,  -2,  -3,  -1,   2, -39, -57,  -1,  -3,  -2,   2,  -4,   8,  20, -62,   0,   0,   0,   0,  -2,  -2,  -1, -12,  -1,  -2,  -4,   6,   6,  -1,   9,  23,   1,   4,   8,  10, -13,   9,  37,  38,  -1,  -2,  -5,  -6,  -1,  29,  -8,  -1,   7,  -1,  16,  13,  22,  18,   7,   0,   7,  12,   8,  -7,   0, -19,  10,   1 },
      {  -1,  -1,   0,   2,  -1, -17,  -8, -25,   1,   2,   4,   1,  -3,   6,  11,   0,  -1,  -1,  -4,  -5,  -9, -30,  14,  17,   0,  -1,   4,  -1,  -9,  16,  15,   3,   0,   2,  -3,  -8,  11, -45,  -5, -33,  -3,  -1,  -6,   9, -25,  12,  60, -30,   1,  12,  12,  -3,   7,   7,  35,  25,   0,   5, -23,  -7, -30,  14,  -5,  21 },
      {   0,   0,  -2,   2,   0,   0,  -2, -12,   0,  -1,   0,   0,   4,  -7,  -9,   0,   1,   3,   2,  -1, -13,  16,  30,  40,   0,  -2,  -3,  -8,   5, -13, -37,  41,  -1,   0,  -1,  -2,  -2,   9,  25, -35,   0,   5, -10,  -7,   2,   3,  27,  10,  -4,  -8, -13,  14, -10,  42, -39,  16,   5,  -3,  24,   4,  33,  32,  -6,   3 },
      {   0,   0,   0,   0,  -6,  21,  36, -40,  -2,  -1,   0,  -1,   4,  -9, -64, -12,  -2,  -2,  -2,  -1,  -3,  -2,  13, -54,  -1,   2,  -2,   5,  -5,  -4,  23, -20,  -3,  -4,  -2,  -3,  -6,   9, -24,  19,  -1,   1,   2,   1,   6,   8,  31,  18,  -5,   0,  -4,   2,  -6,   9,   1,  12,   6,   7,   7,   7,  17,  33,  -3,  17 },
      {  -1,  -2,  -2,   1,   1,   1, -36, -54,   1,   1,   2,  -1,  -2,   4,  23,  19,   1,  -1,   3,  -2,   0,  13,   3,  -5,  -2,  -3,   0, -10,  -3, -29, -36, -77,  -2,  -2,  -8,  -2,   2,  -2,   4, -33,   1,  -1,   5,   4,   8,  -2,  -5,  32,  -2,   3,  -2,  -2, -12, -10,   8,   0,   5,  -1,  -2,   1,   5,   0,  14,   1 },
      {   0,   0,   0,   1,  -1,   2,   8, -19,   0,   0,   1,  -2,   1,   3, -42,  25,   1,   1,  -2,   1,   5,   4,  27,   2,   0,   0,   3,   3,   1,   6, -41,  17,  -1,   1,  -2,  -7,  -2, -28,  10,   9,  -1,   1,  -1,  11,   4,  44, -16,  19,   8,   4,  14,  -3,  22, -36, -30,   6,  -4,   0, -19,  -6, -33, -32, -39,  -3 },
      {   1,   1,   2,  -1,   0,  -3,  20,  17,  -1,  -2,  -1,  -2,  -2,  13, -19, -88,   0,   1,   0,  -1,   0,  -1,  16,  13,  -1,  -2,   1,  -2,  -3,   6,  -5, -43,  -1,   2,   0,  -3,   4,   6,  -2, -47,  -3,  -2,  -4,   2,  -2,   3, -28, -26,  -1,   0,   2,  -3,  -5,  -9, -18,   4,  -2,   4,  -7,  -2,  -5,  -2, -25,  -7 },
      {   0,  -1,   0,  -2,  -4,   8,   1, -28,   1,   1,   1,  -2,   3,  -5,   9,  28,   0,  -1,  -3,  -2,   5,   8,  -5, -41,   1,   0,  -1,   3,   1,   8,  16,  -6,   0,   0,   1,   1,   5,   3,   8, -26,  -2,   3,  -2,   6,  -3,   5,  15, -72,   3,   0,  -2,   1,   0,  12, -54, -33,   5,  -6,   2,  -2,   3, -19,   1, -39 },
      {   0,   0,   0,   0,  -1,   2,  17,  -5,   0,  -1,  -1,  -3,   2,  -4, -12,   2,   1,   0,   0,   4,  -1,   0,  26,   8,   1,  -1,  -1,  -1,   5,  -2, -22,   1,   1,  -2,  -1,   4,  -1,  -5,  55,  16,   3,   0,   8,   4,  11,   2,  -9, -33,  -3,  -1,  -3,  -4, -16, -31,  39, -57,  -2,   5,  -3,   3, -11,  58,  17,  -6 },
      {   0,   0,  -1,   0,   1,  -3,   9,  10,   1,   0,   1,   2,   0,   2,  -5, -19,   0,  -2,  -3,  -2,  -2,  -7,   5,  -5,   1,   0,   2,   3,   5,   1,   6,  17,  -2,  -3,  -4,  -5,  -5, -15,  -8, -25,   1,  -1,   2,   7,   6,  16,  15,  32,   1,   0,   8,  -1,   8, -31, -27,   6,   0,  -1,  -4,  -3, -13,   6,  96, -37 },
      {   0,   0,   1,  -1,  -3,   0,   1,  -6,   0,   1,   0,   0,   1,   3,   3,  12,  -1,  -1,  -2,   2,  -1,   0,  -3, -56,  -1,   0,   3,   0,  -1,   2,  -1,  41,   0,   2,   1,   0,   1,   6,  18, -55,  -3,  -2,  -3,   0,   0, -10, -55,  -9,  -3,   0,   2,  -2,  -3, -12,   9,  27,   3,   5,  -1,  -2,  -2,   7,  14,  59 },
      {   0,   0,   1,   0,   1,   2,   5,  -5,   0,   0,   1,  -1,   0,  -1,  -4,   3,   0,   0,   0,  -2,  -1,  -3,  -1, -27,  -1,  -1,  -1,  -4,  -1,  -7,   2,  24,   0,   1,   1,   1,   3,   5,   8, -32,  -1,  -1,  -4,  -2,  -4,  -8, -11,  23,   1,   0,   4,   3,   6,  10,  56,  20,  -1,   2,   1,   4,  -1,   7, -34, -90 },
      {   0,   0,   0,   0,  -1,  -2,   3,   0,   0,   0,   0,   0,  -2,   0,   1, -14,   0,  -1,  -1,  -2,  -3,  -3,  -5,  -8,   1,   0,   4,   2,   1,   6,  15,  20,   0,   0,  -1,  -5,  -2,   0,  -7, -42,   1,  -2,  -1,   2,   3,  -1,  22,  52,  -1,   0,   7,   0,   3,  12,   6, -93,  -1,   1,  -5,  -3,  -2, -21, -21,  23 },
    },
    {
      {-109,  40,  10,   2,   3,   1,   1,   0,  25, -27,   8,  -1,   0,   0,   0,   0,  34, -12,  -8,   3,  -2,   0,  -1,   0,   2,   6,  -5,   0,   1,  -1,   0,   0,   2,   0,   0,  -1,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {   7,  56, -32,  -1,  -5,   0,  -2,   0, -71,  26,  32, -13,   6,  -2,   3,  -1,  15, -55,  26,   6,  -2,   2,   0,   1,  20, -15, -15,  13,  -2,   1,  -1,   0,   4,   3,  -9,   0,   3,  -1,   1,   0,   3,   1,  -2,  -2,   0,   0,  -1,   0,   1,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -16, -66,  50, -10,  11,  -2,   3,   0, -68,   0,   3,  16,  -2,   5,  -1,   2,  42, -10, -32,  10,  -1,   0,   0,   0,  20,  15, -15, -10,   4,  -2,   0,   0,   0,  12,   1,  -7,  -1,   1,   0,   0,   1,   2,   2,  -2,  -2,   0,   0,   0,   0,   1,   1,   0,  -1,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
      { -47, -19,  19,   5,   4,   1,   2,   0, -16,  51, -29,   4,  -5,   0,  -1,   0, -52,  17,  31, -15,   7,  -3,   2,  -1,  36, -54,  17,   5,  -3,   3,  -1,   1,  16,  -1, -18,  10,  -3,   1,  -1,   0,   4,   2,  -6,  -2,   3,  -1,   1,   0,   2,   2,  -1,  -3,   0,   0,   0,   0,   3,  -1,   1,   0,   0,   0,   0,   0 },
      {   2, -20,  19,  -4,   3,  -1,   0,   0,  24, -44,  39,  -5,   0,  -1,   0,   0, -59, -31,   8,  32,  -4,   5,  -1,   2,  20, -11, -55,  20,   5,  -1,   2,   0,  12,  26, -22, -18,   9,  -3,   0,  -1,   2,  12,   4, -11,  -1,   2,  -2,   0,   0,   4,   2,  -2,  -2,   2,   0,  -1,  -1,   2,   1,  -1,  -1,   0,   1,   0 },
      {  -3,  23,  -9,  -3,  -1,  -1,   1,  -2, -46, -50,  33,   4,   4,   3,   2,   1, -39,  45, -39,   4,   3,  -1,   2,  -1,   5,  -7,  26, -24,   7,  -3,   1,  -1,  39, -34,   5,  11, -12,   6,  -3,   1,   1,  14, -16,   8,   0,  -2,   1,  -1,   1,   1,   0,  -5,   3,   0,   0,   0,   1,   3,   0,   0,  -2,   0,  -1,   0 },
      { -16,   5, -49,  39, -10,   6,  -1,   1, -12,  13, -32,  18,   0,   1,   1,  -1, -29,  10, -23,   2,  10,  -3,   3,   0,  27,  31, -33, -32,  15,  -4,   2,  -1,   1,  28,  31, -40,  -2,   1,  -1,  -1,   3, -12,  31,   0, -12,   2,  -1,   1,  -1,  -6,   7,   7,  -3,  -1,  -1,   0,   2,  -5,   3,   3,  -1,  -1,   0,   0 },
      {  15,  51,  45, -61,  14, -13,   5,  -6,  15,  37,  -1,  11, -20,   7,  -6,   2, -14,   2, -31,  36,  -1,  -2,   1,  -1,  14, -11,   1, -19,  18,  -4,   2,   0, -13,   8,  20, -13,  -5,   4,  -3,   2,  -7, -13,  15,  11,  -7,   2,  -1,  -1,   0,  -9,  -7,  10,   1,   0,   1,   0,   0,  -2,  -3,   0,   2,  -1,   0,   0 },
      {  24,  21,  -7,  -6,  -5,   1,  -1,   0,   6, -45, -28,  23,   0,   5,   0,   3,  28,   6, -15, -25,   8,  -4,   2,  -1,  57, -31,  18,  -2, -13,   4,  -5,   2, -13,  49, -35,  13,  -7,  -2,   0,  -1, -19,  20,  17, -11,   7,  -7,   1,  -2,  -3,  -7,   7,   1,   1,   1,  -1,   0,  -3,   1,   2,  -1,   0,   0,   0,   0 },
      {   2,  20,   5,   6, -10,   2,  -4,   0, -17,  10, -21,  23,  -3,   1,  -1,   1,  16,  38, -15,  15,  -8,  -1,   0,  -1, -56, -26, -21,  35,  16,  -3,   6,  -1,  23,  -1, -42, -34,  27,  -4,   2,   0,   0,  16,  23, -33,  -4,   5,  -4,   1,   5,  -3,  11,  -1,  -9,   3,   0,  -1,  -2,   1,   8,   0,  -3,  -3,   2,   0 },
      {   2,  11,  16, -11,   1,   1,  -2,   1, -11, -29, -49,  51,  -5,   9,   0,   4, -25, -65,   9, -20,  30,  -4,   7,  -1, -40,   5,   7, -19,  -2,   4,  -2,   4,  10, -13,  -3,   0, -10,   5,  -2,   1,  21,   4,  -9,   1, -11,   0,   2,   0,   3,  13,  -7,   4,  -1,  -6,   2,   0,   2,   6,   0,  -4,   4,  -3,  -1,   0 },
      {  -6, -32, -52, -22,  36,  -6,  12,  -4,  14,   5, -10, -19,  -6,   5,  -1,   1,  21, -23,  -5,  29, -12,   4,  -2,  -1, -19, -37,  13, -26,  16,  -3,   5,  -1,  43,  16,  -1, -12, -28,   3,  -4,   3, -12,  34,   6,  19,  -8,  -8,  -1,  -2,   0,   2,  -1,  -1,  10,  -1,  -1,   0,  -3,   3,   2,  -2,   1,   2,  -1,  -1 },
      {  -6,  -3,   8, -31,  12,   0,   2,  -2, -25, -34,   4, -22,  15,  -1,   5,   0,  -5,  22,  45, -31,  -8,   1,  -3,  -1, -19,  -5,  12,  11,  -8,   3,  -2,   1,   6,  25,  17, -28,  -3,  -2,  -1,  -1,   4, -29,  51,  21, -20,   6,  -3,   1,   3, -17, -22,  36,  -2,  -2,   2,   0,  -2,   3, -13,   2,   5,  -2,   1,   0 },
      {  19,   8,  42,  33, -41,   8,  -6,   2,  24,  12,  20, -14,  15, -15,   6,  -5,  29,  -5,   2, -23,  30,  -9,   2,  -1,  12, -24, -15, -26,  14,  11,  -3,   4,  44, -14,   2, -23, -19,  17,  -5,   3,  -1,  12,   6,  18, -27,   1,   1,   0, -11,   6,  -3,  16,  -1,  -8,   2,  -1,  -5,   0,   0,   4,   4,  -2,   0,   0 },
      {  -2,  12,  18, -24,   1,  -1,   6,  -3,  -2,  19, -18, -50,  25,  -2,   6,   0, -20, -11, -28, -33,  -4,  12,   0,   5, -10,  37,  -7,  -3, -39,   4,  -1,  -1,  30,  15,  -2,  11,  -7, -16,   5,  -3,  -1,  35,  14, -17,   7, -10,  -2,  -1,   5,  -5,  40,  -2,  -7,   1,  -4,  -1,  -1, -10,   7,  17,  -3,   0,  -2,   1 },
      {  -9, -18, -16,  -2,  12,   5,  -2,   2,   9,  13,  49,   6,   1,  -6,  -2,  -2,  -8, -16, -16,  -9,  32,  -3,   4,  -1, -11, -15,  36, -29,   4,   6,   0,   3, -37,  -7, -18,  -6,   9,  10,   1,   2,  16,   5,  26, -47, -15,   8,  -1,   3,  11, -16,  26,  27, -29,   6,  -3,   1,   0,   1, -17,  14,  -3,  -3,   4,  -1 },
      {   9,   0, -15, -34,  32,  -4,   2,   1,  12,  -7, -16,   4,  -8,  11,  -5,   3,  18,  11,  28, -14, -11,  -2,   0,   0,  50,  27, -12,   6,  10, -10,   0,  -2,  24, -49,  -2, -19,  23,  13,  -2,   3,  18,  12, -12, -25, -26,  18,   0,   3, -14,  18,  16,  13, -24,  -5,   6,  -1,  -2,  -1,   0,   9,   1,  -8,   3,   1 },
      {   4,   1,   5,   9,  -8,   3,   1,  -1,  -4, -20, -39, -45,  45,  -9,  11,   0,  -1, -15, -34,  16, -22,  19,  -8,   6,  10, -42,  15,   1,   3,  -6,   4,  -2, -34, -12,  13, -23,  16,  -3,   0,   1,  34, -14, -16,  -3, -18,  12,  -6,   3,   8,  25,  -7,  -2, -13,  -2,   1,  -1,   2,   6,   6,   1,  -2,  -3,   2,   0 },
      {   1,  12,  13,   1,  26, -22,  11,  -8, -14, -15,  -8, -16,  13,  -7,   5,  -2,   4,  30,  42,   3,  15,  -6,  -1,  -4, -17,  -6, -30, -32,  28,  19,  -4,   6, -41,   9,  -4,  -8, -36,  17,   2,  -1,   1,   7, -23, -10,   2, -12,  11,  -3,   0,   5,  37, -28,  -4,  -5,  -2,   1,   6, -16,  12,  14,  -8,   5,  -1,   1 },
      {  -2,  -3, -24, -22,  34, -12,  10,  -3,   9,  16,  21,  19,  21, -11,   2,  -3,   3,  13, -35, -41,  28,  15,  -4,   4,  -3, -18, -45,   7, -31,  25,   1,   1,  -5, -10, -12,  -7, -12,  -8,  10,  -1,   2, -25,  -8,   5,  -6,  -4,   0,   4,   0,  23, -32, -10,   2,  -7,   1,   0,   2,   8,  19, -20,   0,  -5,  -1,  -1 },
      {   7,   5,  17,  36,  39, -43,  15, -13,   3,  -4,  -5,  20,  36, -13,  -1,  -2,  -1, -19,   7,   0,  -2,  18, -11,   4,   6,  -4,  13,  -4,   0,  -1,   5,  -4,  19, -19,  17,  -8,  19,  -6,   2,   2, -64, -12,   8, -11,  14,  17,  -6,   3,   8, -24,  -1, -11, -15,  15,   0,   1,   2,  -2,  -8,   7, -12,   0,   2,   0 },
      {  -2,   6,   7,  24,  14, -15,   2,  -3,  -1,  11,  10,  32,  18, -10,  -7,  -1, -10,   2,   1,   1, -35,  28, -12,   5,   4,  13,  14,  18, -15,   1,   5,  -4, -32, -13, -11, -18, -16,   6,   4,   5,   5,  48,   7,  23, -35, -25,   2,  -5, -13,   1,  11,  27,  37, -23,  -2,  -4,   4,  -9,  -7,  -6,  19,   8,  -8,   2 },
      {  10,  11,   9,  21,  26, -26,  10,  -8,   9,   7,   6,  16,  15, -23,   5,  -6,  12,   9,   8,  27,  -2,   3,  -5,  -2,   7,  16,  20, -10,  -6,  10,  -4,   0,  37,  30, -15,  12,  -2, -13,   4,  -6,  79, -23,   3,  -7,  20,  -4,  -3,   2, -11,   7,  -3,   5,   4,  11,  -9,   2, -13,   5,  -4,   0,   0,   7,   2,  -3 },
      {  -9,  -7, -14, -32, -17,  10,   1,   3,  -5,   3, -22,  -3,  45, -50,  20, -10,  -5,   6,   4,  43,  42, -18,  -9,  -1,   5,  28,   8,  11,   5,  19, -11,   4,  -2,  -4, -22,   6,   9,  12,   2,  -2, -23,   9,  -8,  -1,  -7,   9,   7,  -2, -20, -16, -13,  10,  11,  -4,   7,   1,  -4,   2, -13, -18,  10,   1,   3,   2 },
      {  -4,  -1,   3, -21, -13,   6,  -5,   8,   3,  11,  28,  17,  20, -15,  -2,  -2,  -5,  -3,  12, -25,  -8,   8,   2,  -2,  -4,  19,  42, -11,  21, -24,   8,  -8,   6,  42,  -3, -33,  24,   7,  -8,   7,  -8,   2, -17,  -1, -11,  11,  -4,  -1,  -3,  29,   0, -42,   7,  -1,   3,  -2,   2, -10,  43, -14, -18,   6,  -4,   3 },
      {  -2,  -4,  -4, -21, -26,  26, -16,   7,   0,  -4,   6,  39,   0, -16,   2,  -5,  10,  12,  25,  28,  -5,  27, -23,   4,  -2, -23,  -3, -18, -44,  29,   4,  -1,   7,   4,  47,  -4,   2, -30,  16,  -2,  -7,  11, -10,  -8,   7,   3,  -4,   8,   0,  20,  20,   5, -19,  -2,  -3,  -3,   3, -11,   1,  24,   2,  -3,   3,  -2 },
      {  -1, -11,  -7, -25, -19, -11,   9,  -3,   6,  -8,   1,  12,   9, -20,   9,  -6,   7,  -8,   3, -15,  14, -22,   6,  -2,  13, -17,  -8,  13,  13, -18,   7,  -2,  -8, -39,  24, -17,   7, -25,   0,   3,  26,  11,  29,   8,  29, -27, -18,  -1,   6, -19,  20, -19,  37,  25, -22,   2,  13,  -4,   1,  -1, -19,  29, -11,  -3 },
      {   0,   4,   4,  20,  20, -26,  16, -13,  -9,  -7,   2, -19, -40, -22,   6,   2,  -2,   2,  -6,   3,  38, -37,  -5,   1,  -6,  -6,  14,  29,  -9,  -4,  -7,   2,  -8,  13,  37,   3,  17, -14,  -7,   0,  -9,  29,  10,  -5,  -5,  13, -12,   0, -15,  51,  -2,  14,  -2,  -5,  -1,  -7,   1, -12,  20,  -4,  19,   3,  -4,  -2 },
      {   2,   5,   2,   2,  28,  20, -13,   9,   1,   4,   4,  10,  32,   2,   2,   6,   4,   8,  -1, -12,   4,  -4,  24,  -6, -15, -25, -30, -18,  -4, -45,  14,   5,  -2,  18,  22,  29,  33, -10, -31,   1,  10,  12, -21,   3,   4,  22,   3,  -6, -32, -23,  12,  12,  12, -12,  11,  -2,  -6, -14, -20, -10,  20,   9,  -6,   4 },
      {  -5,  -9, -12, -15, -46, -46,  45, -16,  -2,  -5,  -2,   1, -23, -26,  16,  -1,  -1,  -2,   1, -11, -18,  18,  -1,   6,  -6,  -6, -12, -29,   6, -23,  10,   2,   3,   0, -24,  19,  20, -11,  -4,   2, -13, -33, -10,  -4, -12,  -4,  -1,   4,   0,   7,  15,  21,   7, -25,  -7,  -2,  -8,  -7,  -5,  11,  24,   2,  -8,   1 },
      {   4,   1,  -3,  -1,  15,   5, -13,  10,   5,   1,   1,  10,   9, -16,   1,  -1,   5,   3,  -4,   5,   6, -20,   0,   0,   7,   0,  -2,  23,  -5, -12, -10,   1,  17,  12,  16,   4,   0,  22, -13,  -1,  -2, -16, -19,  -7, -19, -11,  15,  -9,  93,   2,  15,   4,  13, -25, -12,  -1, -10,   7,  -4,  20,  15,  20, -15,   2 },
      {  -1,   0,  -5,  -7, -20, -25,  23,  -9,   1,   6,   2,   8,  -7,  -6,   1,  -2,   1,   8,   0, -22,  -5,  16, -28,  11,  -2,   1, -15,   0,  18,  10, -22,   5,  -8,  19,  19,   9,  14,  13, -10,  -6,  27,  57, -19,  11,   6,  31,   0,  -6,  35, -25, -31,   2, -12,  18,  20,   0,  14,  11, -17,  -3, -12, -12,  13,   3 },
      {  -6,  -9, -13, -11, -18, -18,   6,  -9,   1,   2,   4,  20,  21,  10, -21,   2,  -4,  -8, -21,  -7,   1,   1,  -4,   0,  -6, -12,   3,  52,  15,   4, -17,   2,   5,  10,  20,  15, -19,  36,  -9,  -5,   5, -19,   1,  15,  -4,   7,  10,  -9, -37,   8,  28,  -6, -10,  11,   6,   0, -33, -21,  -8,  45, -11,  -4,   9,   2 },
      {  -3,   1,  -1,  -4, -16,  -6,  12, -12,  -1,   8,   3,   9,  41,  21, -18,   6,   0,  13,  18,  15,   7,  -7,  17,  -2,   1,  13,  -9, -11,  -3, -32,  20,   4, -16,  -5,  -8,  14, -23, -13, -21,  15, -10,  21,  17,   2,   8,  -2, -25,   1,  22,  44, -17,  12, -14,  21,  -7, -11, -37,  37,  -1,  10,   3, -10,  10, -10 },
      {   0,   4,   4,   0,  22,  17, -36,  23,  -6,  -1,  -7,  -3, -21, -49,  11,   3,  -8,  -5,  -7, -22, -29,  12, -21,   1,  -4,  -1, -15, -20,  28,  14, -15,  -8,  -8, -10,  -9,  20,  12,  12,  -5,  -4, -12,   3,  21,   9,  11,   3,  -8,  -2,   1,  29,   1,   6,  12,  25,  -9,  -4, -51,   6,  -7,   1,  -9,  16,  13,  -4 },
      {   1,  -1,  -2,  -2,  -7,  -9, -19,  14,   0,  -3,  -5,  -6,  10,  -8, -45,  13,   4,   2,   4,  12,  12,  30, -20,   0,   0,   1,  -8,  -4,  10,  -8,   6,   0,   8,  -5,  -3,  43,  -3,  19, -21,  -3,   3,  -9,  43,  -9,  -5,   9,   7, -10,   4,  27,  -1,   2,   6,  -1,   0,  -2,  71, -13,   8,  -9,  -5,  14,  -2,  -4 },
      {   1,  -1,  -1,   2,   6,  18,  12,  -8,   4,   4,   5,  22,  11, -32,  33, -15,  -3,  -7,  -5,  -1, -57, -49,  27,  -1,  -2,  -2,  -5,  -4,  -9,  -3, -24,  16,  -2,  -8,  -6,  16, -23,   5, -14,   0,  -1,  -2,  20,   3,  -1,  12,  -7,  -4,   0,   6,  -4,  -3, -21,  -3,  12,  -2,  43,  -5,  18,  16,  -3, -23,   3,   0 },
      {   1,   3,   1,  -3,   4,  19,  11,  -9,   2,   6,   4,   4,  27,  32,   4, -16,  -3,  -1,  -5, -16, -10, -17, -16,  10,  -4,   1,  -1,  -3,  40,   2, -20,   4,   6,  15,   3,   2,  13,  -3,  40,  -8, -14,   2, -15, -18,  17, -24,  -2,  20,  -4,  39, -17,  23,  11,   7, -24,   5,  29, -20, -40,   9,  15,  11,  -2,  -6 },
      {   3,   3,   7,   6,   4,  46,  47, -29,   1,   0,   2,   5,  -3,   8,  29, -14,   2,  -1,   8,   5,   4,  33, -10,   6,   3,  -2,   1,   9,  13, -11,   2,  10,   2,  -7,  25,  29,  -8,  11,   4,  -7,   3,   4,  36, -33, -15,  -1,  28,   0,  -2,   1,  -7, -25,   9, -17,   0,  15, -27,   9,  11, -20,  -8,   4,  -9,   8 },
      {  -5,  -6,  -3,  -7, -21, -51, -28,  14,   0,   1,   4,  -2,  16,  17, -12,  -5,  -1,  -2,   5,   2, -28, -20,  22,  -7,  -3, -10,  -2, -17, -15,  13, -31,  17,   1,  -5,  25,  -2,  10,  17,  13, -20,  -2,  10,  17, -35,  11,  -5,  29,  -7,  -6,   8,  -5,  -4,   6, -13,   2,  13, -22,  17,   7, -42,  10,   1, -10,   7 },
      {  -1,  -3,  -2,  -4,   1, -25, -36,  22,   2,  -1,   0,   4,   6,  -6,  28, -17,   0,  -2,  -2,  -6,   7,  18,  18, -21,   6,   3,   2,  22,  50,  -2,  32,  -9,   1,   3,  11,  28, -17, -29,  34,  12,   7,  12,  12,   4,  -2,  -5,  -2,  15,  13,  -5,  -6,  -5, -23, -21,  13,   8,  -4, -15,   9,   2,  22, -35,  -1,   7 },
      {   0,  -3,  -2,   2,  -1,   3,   9,   0,   3,  -2,   2,   3,  15,  37,   8,   3,   2,  -1,   5,  13,   5,  -4,  -7,   6,   3,   0,  -1, -13,  -2,  12, -26,   2,  -2, -15, -16,  12,  17, -10,  -2,  -6,   5,  -4,  13,  21,  -3,   9, -15,   0,  32,  -4,   4,  16,   9,  11,  15,  -8, -26, -80,  41, -21,   4,  -2,  13,   0 },
      {   1,   0,   0,  -1,  -5,  -2,  -2,  -6,  -2,  -4,  -6,  -6, -15, -26,  -4,  -6,   2,   1,  -1,  -1,  -5,  19,  36, -16,   1,   1,   0,   6, -21, -11,  22,  14,   4,   1,   9, -17, -26,  26, -14,  15,  -1,   4, -14, -42,   4, -10,   1,   6,   7,   3, -38,   3,  12,  23, -11,   4,  -6, -61, -26,  -7,  -7,  -3,  31,   1 },
      {   2,   1,   3,   0,  -3,  16,  29,  14,   0,  -2,   2,   1,   9, -25, -46,  41,  -2,  -7,  -1, -11, -17, -42, -40,   5,   1,  -5,  -2,  -3,   6,  -6,  27, -37,   1,  -3,   8,  -3, -24,  -2,  20,  -5,   6,  -4,  -1, -24,  17,   3,   2,   3,   3, -12,  -1,  -3,   4, -10,  16,   0, -14, -11,   3, -17,  25, -26,   9,   8 },
      {   0,   2,   2,   3,  -1,  -5,  25,  44,  -1,   0,  -2,   5,   4,   2,   7,  26,   0,   2,   2,   4,  -7,   0,  15,  -3,  -3,  -4,  -6,  10,   4,   8, -13, -10,   4,  12,  22,  28,  -3,  13,  12,  14,  -3,   0, -16, -14, -42, -16, -29,  33, -11, -25,   3,  18, -10,  14, -39, -16,   4,  16,  28, -20, -21,  14,  15, -21 },
      {   0,   0,   0,   2,   6,   5,  -4, -19,  -3,  -3,  -4,  -3,   2, -22, -39,  -2,  -1,   0,   2,   5,   3,  -9,   8,  -4,  -2,  -2,  -6, -13,  -7, -22,   6,   8,  -6,  -3,  -6,  10,  33,   8,  30, -16,   3,  18,   1,  23,  -1, -22,  35,  19,  -3, -16, -35,  13, -16, -12, -23,  24, -10,  -9,  39,  40, -28,  -6,  -6,  -9 },
      {  -2,  -3,  -2,  -3,  -5,  -1, -45, -68,  -1,  -2,  -4,  -1,  -2,  -4, -26, -31,  -1,  -2,  -2,  -7, -13, -19, -22,  21,  -1,  -1,  -5,   0,   6,   0,  10,  15,   3,   6,   3,  17, -19, -12,  14,  17,   2,  -7, -13, -16, -32,  23, -27,   1,  -2, -22,   5,   1,   3,   5, -16, -13,  -7,   1,  12, -23,  -2,  14,  -5,  -8 },
      {  -1,  -1,   0,  -1,   0, -15, -31,  19,   1,   3,   2,   2,  -3,   7,  43,  16,   0,   2,   3,   3,  12,   5, -43,  49,  -2,  -1,   0,   1,  -8, -60, -22,   4,  -2,  -3,  -4, -12, -35,   2,  -5, -10,  -1,  -3,  -7, -22,  13,  -7,   4,  -4, -15, -11, -20,  15,  -3,   1,  -3,   4,  -1,  -4,  21,  14,  -6,  -5,  -5,  -3 },
      {  -1,  -2,  -2,  -3,  -5,   4,   2,   0,   0,   1,   0,  -2,   5,  10,   0,  12,  -3,  -4,  -2,  -4, -10,   6,  -2, -20,   2,   5,   2,   0,  14,  39,  -2, -12,  -2,  -5,   0,  -4,  -7,  -1, -38,  16,   3,   9,  -1, -26,  25,  14, -20,  -5, -11, -16, -43,   2,  -5, -28, -18,  -9,  -9, -11,  21,  34,  19,  30, -56,  -9 },
      {  -1,   1,   0,   1,  -4, -19,  -3,  21,   1,   3,   1,   0,  -1,  19,  22,   3,   0,   3,   1,   7,  -3, -33, -18,  -2,  -2,   0,  -6, -14, -28,   0,  25, -21,  -4,   2, -11,  15,   5,   3,  16,  -5,   5,   8,  19,   3, -47,  30,  -4,   8,  -1,   8, -21, -48,  14,  -1, -10,  13,  -7, -19, -41,  23, -21,   9, -16,  -7 },
      {  -2,  -2,  -2,  -3,  -4,  -6, -21, -47,   2,   3,   2,   5,   6,  10,  44,  37,   0,   0,   1,   1,  -4,  -9,   9,  16,  -1,  -1,  -3,   1,  -5,  21,  44, -46,   2,   6,   7,   7,  15,  45,  -4,   3,   1,   2,   1,   6,  22, -10,   6,   9,   4,  10,  10,  27,  -3,  -3,   6,  10,  10,   3,   8, -13, -19,   5,   5,   4 },
      {   0,   1,   1,   2,  -1,  -5,  10,  17,  -1,  -1,  -2,   0,   0,  14,   9, -61,  -3,  -3,  -4,  -2,  -4, -13, -46, -68,  -3,  -3,  -6,  -1, -15, -17,  14, -20,  -1,   1,  -1,   0,  -8,  13,  -4,  20,  -3,  -3,  -9,  -3,  12,   4,  12,  -2,  -2,   2,   6,  36,   0,   7,  12,   7,   1,   7,  19,  -7, -22,  14,   3,   8 },
      {  -1,  -1,  -1,   0,  -3,   1, -13, -35,   1,   1,   2,   0,   1,  -2,   6,  25,   1,   1,   3,   0,   4,  14,  19, -49,   0,  -1,  -2,  -3,   3, -16, -40, -56,  -2,  -3,  -1,   0, -14, -28,  14, -40,  -5,   1,  -6, -16, -22,  16,   8, -13,   1,  -2,  -9,   1,  23,  14,   3,  -9,  10,  10,   8,  19,   9,   9,  23,   6 },
      {   0,   1,   0,   0,  -4,   4,  13,  11,   0,   1,  -1,   4,  -2,  -8,  -3,   7,   0,   1,  -2,  -2,  -5,   6,  23,   9,   1,   4,   2,  13,   5,  -6,  12,   1,  -3,  -1,  -9,  -9, -36, -13,  14, -31,  -5,   4, -10,  19,  10,  58,  22,   7,   9,  10,  15,  26, -27,  20, -24,  35,  -1, -11,  -3, -28,  -7,  35, -37,   3 },
      {   0,   0,   0,   2,  -1,  -4,   0,   4,   1,   2,   1,   1,   4,   0,  -1,  21,  -1,  -1,  -2,  -2,  -8, -18,  -4,   6,   1,   3,   2,   3,  20,  23,  12,  12,   2,   3,   4,   9,   0, -59, -29,   1,  -7,  -4, -15, -17, -30, -32,  23, -39,  -2,   7, -10,  24,  -3,   2,  20,  15,  -4,  -7,  -8,  -2, -57,   4,  -4,  20 },
      {   1,   1,   1,   1,   4,   0,   9,  22,  -1,  -2,  -3,  -4,  -1,   2, -11, -28,   2,   2,   2,   3,   4,   5,  31,  33,   0,  -1,   0,  -2,   4,  10,  20,  -1,   0,   2,   4,   5,  -4,  12,  27, -25,  -2,  -2,  -5, -19,  10,  29, -30,  -4,   2,   2,  -3,  27,  57,  -9,  29, -18,  -1,  14,  18,  30, -31,  -7,  -6,  27 },
      {   0,  -1,  -1,   0,   0,  -3,  -1,   4,   2,   2,   4,   1,   4,  13,  19,  28,  -2,  -3,  -3,  -6,  -6,  -8, -16, -16,   4,   4,   5,   2,   6,  19,  39,  60,   0,  -2,   0,  -3,   2,  -7,   3,  -3,  -1,  -1,  -2,  -4,   0,  29,  29,  -2,   0,  -5,   2,   3,  35,  16, -11,  11,   8,  16,  28,  25,  30,   9,  45, -23 },
      {   0,   0,   1,  -2,   0,  -1,   4,  12,  -1,  -1,  -1,   2,  -1,   6,  -1, -49,   0,   0,   2,  -2,  -3,  -5,   3,  29,   0,   0,   2,   0,   3,  13,  34, -30,  -1,  -1,   3,   3,   6,   4, -31, -61,  -2,   1,  -4,  -3, -17, -29,   3, -23,  -6,  -9, -11,  -9, -23,  29,  -8,  -1,   2,   0,  14,   2,  40,  13,  15, -12 },
      {   0,   0,   0,   0,  -2,   0,   7,   3,   1,   0,   0,   1,  -2,  -4,   9,  10,   0,  -1,   0,   0,  -2,  -2,   7,  -3,   1,   0,   0,  -1,  -2, -12,  15,  12,   1,  -1,   0,  -1,  -1,   8,  48,   4,   3,   2,   5,   4,   7,  -6,  19, -45,   0,   1, -11,  -1, -31,  -5,  30, -74,  -3,  -8,  -7,  -3,  -6,  46,  -2, -28 },
      {   0,   0,   0,  -1,   0,  -3,  -2,   0,   0,   0,   1,   0,   1,   3,  -3, -14,  -1,  -2,  -2,  -3,   0,  -4,  -3,   0,   2,   2,   5,   6,   8,   2,   0,  -6,  -1,  -3,  -2,  -9, -10, -18, -29, -36,  -1,   3,   3,   2,  19,  14,   5,  28,   3,   2,   6,   1,   2, -70, -15, -21,  -5,  -5, -16, -13, -38,  -1,  41, -51 },
      {   0,   0,   0,   0,   1,   4,  -2,  -7,   1,   1,   1,   1,   0,  -2,  10,  19,   0,   0,   0,  -1,   2,   6, -18, -43,   1,   1,   0,   1,  -1,  -6,  16,  34,   1,   2,   3,   0,   6,  22,   9, -57,  -2,  -2,  -7,   4,   1, -10, -52, -42,   1,   0,   6,   4,  -5,  -6, -30,  -1,  -1,  -5,  -7, -15, -13, -41, -26,   0 },
      {   0,   0,  -1,   0,  -2,  -1,  -3,  -2,   1,   2,   0,   2,   0,   0,  -2,  12,  -1,  -1,  -1,   1,  -2,  -4,  -9,  -9,   1,   2,   1,   1,  -2,  -7,   3,  26,   0,  -1,  -4,   0,  -5,   4,   1, -24,   0,   3,   4,   5,   0, -14, -40,  31,  -2,  -3,  -9, -11, -28, -27,  22,  16,  -1,   3,  -2,   4,   5,  62,  36,  65 },
      {   0,   0,   0,   0,   0,  -1,  -6,  -8,   0,   0,   0,  -1,  -2,  -6,   0,  10,  -1,  -1,  -1,  -1,  -2,   0,  -8, -17,   1,   1,   1,   3,   1,  -1,   7,  20,   1,   1,   1,  -1,  -4,   0, -17, -34,  -2,  -4,  -5,  -7, -12, -19,   5,  70,   3,   1,   5,   1,   9,  32,  61, -18,  -2,   1,  -3,  -1,  -1,  -2, -46, -31 },
      {   0,   1,   1,   0,   0,   3,   4,   1,  -1,  -1,  -1,  -1,  -1,  -4,  -6,   0,   0,   1,   1,   1,   1,   5,   9,   1,  -1,  -1,  -1,   0,  -1,  -4, -10,  -4,   1,   2,   3,   1,  -1,   5,  17,   9,  -2,  -2,  -4,  -7,  -1,  -7, -41, -28,   1,   0,   1,   2,   1, -11,  47,  75,  -2,   2,   1,   2,   3,  21,   8, -68 },
    },
    {
      {-101,  14,  21,   4,   5,   2,   2,   1,  58, -30, -16,   1,   0,   2,   0,   1,   5,  22,  -4,  -6,  -2,  -2,  -1,   0,  -1,  -7,  11,   3,  -1,  -1,  -1,   0,   6,   0,  -8,   3,   2,   1,   0,   0,  -2,   3,   0,  -2,   1,   1,   1,   0,   1,   1,  -3,  -1,   0,   0,   0,   0,  -1,   1,   1,  -2,   0,   0,   0,   0 },
      { -34,  50, -10,  -4,  -2,   0,  -1,  -1, -54, -31,  38,   6,   4,   1,   2,   0,  54, -40, -31,  11,   2,   2,   0,   1,   2,  33, -12, -16,  -1,  -1,   0,  -1,   3,  -4,  14,   0,  -3,  -2,  -1,   0,   1,   1,  -3,   4,   1,   0,   0,   0,   0,  -1,   0,  -1,   0,   0,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0 },
      {  34, -17,   2,  -4,  -1,  -2,   0,  -1,  27, -28,  -2,  12,  -5,   0,  -3,   1,  25,  50, -52,  -8,   3,   0,   1,  -2, -39,  42,  37, -24,  -1,   0,   2,  -1, -12, -27,  24,  19,  -5,  -3,  -1,   1,  -1,  -5,  -7,   2,   7,   0,  -2,  -1,   0,  -4,   6,   0,  -2,   3,   0,   0,   0,  -1,  -3,   5,  -1,  -1,   0,   0 },
      {  21,  35, -10, -14,   2,   0,   0,   0, -35, -49,   2,  20,   5,  11,   5,   2,  -8,  39,   0, -19,   0,  -6,  -1,   0,  17, -42,  39,  14, -12,  -6,  -6,  -1,  22,   0, -47,  17,  10,   8,   0,  -1,  -9,  21,  -6, -18,   2,   2,   6,   2,   1,   0,  -7,  -4,   1,  -2,   0,   0,  -4,   6,   5, -10,   0,   1,   0,   0 },
      { -20, -64,  21,  19,   3,   6,   1,   2, -50,  16, -33,   2,  11,   3,   5,   1,  67,  14,   9, -27, -13,  -4,  -3,  -1,  -6,   2,   7,  21,  -6,  -6,  -1,  -1,  -1,   3, -15,  -5,  17,   0,   0,   0,  -2,   1,   3,  -7,  -4,   5,   2,   1,   1,   3,  -4,   0,  -1,  -3,   1,   0,  -1,   2,   4,  -4,   1,   0,   1,   0 },
      {   1, -55,  22,   2,   8,   3,   3,  -1, -10, -54,  -5,  17,   4,   8,   3,   3, -35,  -5, -32,   6,   6,  -1,   0,   2,  54,  16, -27, -30, -10,  -1,  -2,  -1,   8,  40,   6,  -6,  -7,  -8,  -1,  -2,  -5,   0,   9,   4,   5,   2,   0,   2,  -1,   5,  -7,  -4,  -1,   0,  -1,  -2,  -2,   4,   6,  -5,   2,   0,   1,   0 },
      { -32, -20, -32,  12,  10,   6,   3,   2, -23,  44,   9, -15,  -5,   6,   0,  -1, -25,   1,  -1,   5,   6,  -5,   2,   0,  33,   7,  39, -32, -18,  -4,  -6,  -2,  13, -22,  12,  52,  -2,  -2,  -3,  -2,  -5,  -2, -35,  -3,  26,   9,   6,   3,   2,   0,  -2, -15,  -9,   5,   1,   1,  -2,   4,  -1,  -2,   1,  -3,   1,   0 },
      {   2, -39, -46,  20,  17,   1,   2,   4,  19, -20,  40, -33, -13,  11,   1,  -1,  26,  11,  -3,  44, -17,  -9,  -2,  -2, -12, -20, -19,   1,  28,  -4,  -4,   0,  13,   4, -20, -11, -13,  20,   1,   0, -13,  31, -17,  -7,   1, -12,   5,   1,   0,   1,   6, -12,   8,   3,  -3,   1,  -4,   5,   4,  -5,  -8,   4,   1,  -1 },
      {  -5,  15,   5,  10,  -9,   2,  -3,  -1, -30,  -2, -28,  -8,  18,   4,   4,   2, -11,  35,  16,   1, -15,  -7,  -2,   0, -14, -13,  -5, -23,  23,  -9,   3,  -1,  36,  -7,  40, -21, -34,   6,  -5,   1,  -6,  -3,  -7,  59,   1, -16,   3,  -3,  -1,   0, -20, -15,  31,   2,  -2,   1,  -1,   0,   5,  -8,  -6,   9,   0,   0 },
      {  21,  24, -28,  10,  -4,   2,  -2,   4,  30, -17, -23, -11,   3,   2,   2,   0,  33,   5,  30, -23, -19,  -4,  -3,   0,  53,  24, -10,  18,  -9, -12,  -5,  -2, -40,  23,  18,  14,   6,   2,   2,  -1, -15,  -9, -29,  11,  18,  -9,   1,   1, -11,   9,   0, -21,  13,   6,   0,  -2,   3,  -4,  13,  -9, -13,   3,   4,   2 },
      {   6,  -7,  20,   3, -10,   6,   1,  -2,   0, -29,  -9,  -7,  24,   3,   7,   3, -30, -24,  21,  21, -22,   3,  -4,   3, -14,  46,  -2,  25,   5, -21,  -3,  -5,  20, -47,  -4,  -5,  35,  19,  -1,   0,   0,   2, -38,  -8, -18,  17,  16,   5,   2,  16,  -5, -21,  -8, -18,  -2,   1,  -1,   2,   7,  -1,   6,  -1,  -4,  -1 },
      {  15,  -5,  71,  -3,  -7, -10,  -8,   0,   8,  -2,   1,  -2,  -2,  -4,  -1,  -2,  13, -39,  -5,   1,   2,   9,   2,   1,  -4, -27,  -2,   3,  -4,   5,  -1,   4,  -1, -13, -11,  38, -29,   0,   1,   2,  -4,  36, -32,   9,  29, -22,  -7,  -4,  -2,   0,  11, -23,  16,  21,  -1,   0,   1,  -1,  -6,   5, -18,   1,   5,   1 },
      {  15,  21,  -1,  21,  -7,   1,  -1,  -6,   8,  -5, -35, -16,  12,   9,   7,   5,  14,  -4,  15,  30,   3, -12,  -7,   2, -32,  11, -11, -15, -12,  12,   1,  -1,  38,  36, -12,  16,   3, -27,   0,  -5, -20,  -7,  20, -13,  42,  23,   3,   3,   9, -11,  -6, -21, -36,  11,   8,   2,  -9,   6,   5,  11,   6, -18,   2,   0 },
      {   9,   4, -29,   1,  -1,   7,   2,   4,  30,  27, -34,  19,   2,  -1,   4,  -1,  10, -42, -35, -25,   9,   9,   1,   3,  -6,   8,  11, -15, -20, -20,  -2,  -4,  46,  15, -10, -12,   6,  11, -10,   2, -19,  37,   1,   3, -23,  -2,   5,  -1,  -6,   4,   4,   1,  26, -13,  -3,  -3,  -1,   1,  13, -19,  -4,  17,  -3,   1 },
      {  -7,  24,  -4,  27,  -8,  -5,  -3,  -1, -22,  14, -50,   1,   1,   6,   3,   3, -17,  17, -31,   6,   5,  -5,   2,  -4,  11,   6, -19, -18,  31,   8,   0,  -2, -27, -12, -15, -22,  -8,  13,  10,   5,  12,  20, -12, -39,  -4, -15,  -6,   1,   2,   8,  47, -21, -15,   5,  -7,  -3,   8,  -8,  12,  15, -12,  -2,   2,  -1 },
      { -10,  -8, -18,  12,  -9,  12,   3,   3,  -1,   4,  24,  55,   0, -16,  -3,  -1, -10,   8,  -9, -17,  19,  15,   1,   1, -36, -26, -37,   8,  -2,  -2,  16,   3,  -1,   5,   3,  -7,  23, -23,  -2,   1,  -1, -26, -41,  -2,  24,   7,  -5,  -1,  -4,  18,  -6, -35,   0,   6,   3,   0,   9,   0,  23, -13, -20,   0,   3,   3 },
      {   7,   6,  26, -61,  19,   4,   4,   0,   8,  25, -14,  27, -33,   8,   6,   1,  32,  12,   8,  44,  22, -19,  -5,  -2,  21,  -2,   1, -19,  20,   9,  -8,  -3,   4,   1,  -4, -15,   3,  19,   1,   0, -11, -18, -23,  -3,  -7,  15,   7,   8,  -8,  12, -17, -15,  -6, -10,  -1,  -2,  -2,   4,  17, -12,   3,   1,   0,   0 },
      {   9,  22,  42,  47,  -2,  -7,   0,  -1,   6,  33,  30, -15,  10,  12,  -1,  -2,  -6,   9, -49,   6, -19,  -2,   7,   1,  14,  -3,  15,  11,  20, -20,  -6,  -4,  -5,  20,  -6, -12,  26,  20,  -2,   0, -27, -28,   3,  -2,   3,  10,   5,   4,   0, -14, -27,   0,   4,   8,   1,   1,  -2,   7,   4, -22, -10,   2,   3,   0 },
      { -16,  -7,  -9, -41, -29,  14,   2,  -1, -10,  21,   6,  27,   8, -13,  -5,   3, -16,  19,   0,  29, -17,   3,  -3,  -1,  -2,  46,   1,  28,   3, -14,  -3,  -6,  -2,  14, -25,   5,   6,   1,   8,  -4, -11,  14,  27,   4,  28, -28,   3,  -1,   7, -27,   9, -12,  27,  26, -13,   2,  -3,  -1,  -9,  -3, -17,   7,  12,  -1 },
      {  -1,  -6, -23,  15,  27,  -5,  -1,   5,  -1, -15, -28,  19, -18,  23,   3,  -1, -13, -40,  -6,  -2,   3, -26,   4,   0, -11,  11,  23,  18,  26,  20, -10,  -4, -14, -21, -19,  -9,  -8,  -3,  17,   2,  29,  -9,  11,  13,  20,  -4, -16,   0, -15,  -7, -45,  -1,   7,  29,  10,   0,   6,  14,  -3, -25, -13, -10,   7,   4 },
      {  -1, -17, -14, -21, -32,   0,  11,  -2,  12,  -7,  12,  -1,  42, -11, -10,   3,   8, -25,   3, -27,   0,  11,   3,  -5,  10,  -6,  37,  -7,  27,  22,  -6,  -2,   9,   5, -13, -24, -29,  26,   6,   9, -23, -35,  -2,   5,  -6,  -3,  -3,   5,  17,   1,   5, -20, -33,  15,  -6,   1,   3,  10,   8,  23, -18, -14,   2,  -3 },
      {  -9,   0,   2, -40,  42, -12,   2,   5, -20,  -7,  -7, -34, -33,   9,  -2,   1, -23, -14, -16, -25, -22,  -7,   7,  -4, -37,  -1,  12,  23,   4, -18, -14,   1,   3,  45,  26,   5,   7,  12,  -4,  -6,  -2, -14,  -6,  -6,   4,  -3,   2,   0,   8,   9,  20, -20,  -4,   5,  -4,  -2,   4, -13,  17,  15, -12,   2,   0,   2 },
      {   1,  -5, -13,  -7,  -2,  -9,  -6,   4,   7,  -3, -14,  37,   5,   7,  -6,  -5,  10,  -7,  -6,   5, -14,   7,  18,   0,  18, -29, -10,   8,  29, -35,  -4,   6,   8,  -8,  41,   7,  18,  23, -27,  -7,  14,  19,  10,  -3,  12,  33,  13,  -6,  -7, -24,   4,   0, -19,  19,  17,  15, -13,  -5, -36,  26,  -9, -17,  -2,   2 },
      {  -3,  -4,  -9,  31, -14,   4,  -1,   1,  -8, -16, -22,  31,  -7,  -4,  -6,   2,  -2, -24,   5,  34,  17,  -5,   5,  -3, -22, -18,  23,  -1,   7, -11,  -6,   2, -35,  26,   3,  27,  -3,  21,   1,  -7, -18, -29,  -6,   2,  -8, -23,  27,   3,  25,   4,  -1,   7,  16, -26, -24,   4, -10,  -3, -17,  15,  24,  20,  -8,  -7 },
      {  -2,  -1,  -6, -30, -19,  12,  13,  -4,   5,  10,  -3, -17,  47,  17,  -3,   6,   2,  -4, -20,  14, -25,  -1,   8,   4, -18, -24,  -1, -27, -21, -20,  -9,   3, -44,  14, -12,   5,  -1,   1,  -2, -12,  52,  12,   4,   8,   2,   1,  18,   2, -10,  21, -23, -12,  -4,  -8,   3,   4,  -3,   7,  19, -14,  10, -14,  -2,  -1 },
      {   1,   3,   2,  -6,  60,  -9,  -8,   1,   3,   3,   4,   8,  40,  21,  15,  -7,  -1,   4,  22, -20,  17,  15,  10,  13, -16,   7, -30, -30, -20, -10, -12,   6, -17, -20,  -4,  -2,  -4,  24,   4,  -7, -21,  -9,   9, -23,   1, -13,  10,  10,  22, -20, -10,  -4,   7,  15, -18,   1, -23,   9, -12,  -6, -15,   8,  11,  -4 },
      {  -2,  -5,   3,  -8,   0,  20,   7,   3, -10, -22, -13, -40,   3, -31,   7,   5,  -3, -11,  12,  14,  45,  15, -10,  -6,  -5, -12,  31, -18,   6, -20,  21,   0, -29,   1,   9, -30,  24, -12, -16,   7, -11,   9,  -2, -10,   8,  13,   4,  -7, -14, -44,   4,  -5,  14,   4,   8,   8,  11,   1,   1, -21, -24,   4,   3,   3 },
      {  -5,  -2, -13, -30, -18,  -1,   4,   7,  -8,  -5, -34, -13, -10,  14,  -5,  -2,  -4,  -6, -21,   9, -41,  30,   7,   2,  -5, -18, -24,  -4,  -3,  -1,  15,  -2,  -9, -16,  -4,   7,  -6, -14,   4,  10, -46, -21, -28, -14,  -1,   0, -25,  -4,  12,  -7,  -9,  42,  -1,   3,  12, -11,   4,  20, -19, -21,   7,  -7,   8,   6 },
      {  -1,  -3, -15,  -9, -19, -16, -14,   4,  -3,  -2, -20, -36,  -1, -14,   7,  -6,   5,  10, -37,  15,  49,  17, -11,   0,  12, -13, -20,  43, -21,  14,  -7,  -1,   6, -17,  16,  14,   8,  18,  10,  -2,   7,  -4,  18,   8, -12,   3,   0,   9,  27,  21, -30, -21,   5,  12,  -4,  -2, -21,   0,  -3,   3, -13,   0,   7,   2 },
      {   9,   1,   3,   4,  -7,  -1,  16,  -3,   8,  -9,  -6, -10,   0,  -1, -10,   1,  15,  -4,   8,  13,  -5,   7,   9,  -8,  10, -10,  10, -17,  -6, -17,  -5,   0,  35,  -4,   0, -13,  21,  -5,  -6,  -6,  50, -40,  -5, -20,  17,  -9, -19,  11,   2,  23,  21,  21,  15,  36, -34, -10,  -8,  -2, -12,  -5,  -4,  39,  27,  -4 },
      {  -1,  -1,   4,  -2, -28,  14,   6,   7,  -2,  -3,   2,  -6,   2, -10,  -1,   1,  -2,   5,   7, -20,  13,  -9,   0,   1, -14,   3, -18,  -3,  15,  20, -18,  -8,  20,  24,  29,  27,   2,  31,  36,  -8,  13,   8, -18, -34, -10, -19,  18,  27, -42, -25, -14,  23, -10,  -3,  -8,  -1,  13,   6, -12, -27,  -4, -14,  15,  -1 },
      {  -8,  -3,  -3,   5, -18, -59,  -4,   7,  -9,   0,   0,   5,   2,  41, -19, -11,  -4,  13,  16,  18,  15,  41,  15,  -5, -12,  19,  10,   7, -26,  21,  -5,   7,  -4,  26,  -9, -12, -14,  18, -18,   0,   9,   2, -20,   4,  -7,  25,  -7,   1, -33, -13,   2,  -1,   5,   3,   5,  -1,  13,   1,  -3,  -2, -11,  21, -15,  -4 },
      {   7,  11,   4,  16,  27,  60,   0,  -4,   0,   2,  -9,   4, -29,  -9,  -1,   5,   3,   6,  -8,  10, -21,  52,  -6,  -3,   7,   4,   2,  14, -30,   5,  11,   1,   7,   4, -10,  -9, -30,  15,  -6,  -5,  28, -10,  -8,   8,  -9,  -2,  20,   5,   1, -21, -10,  -8, -20,  -3, -13,   6,  16,  23,  -7,  17, -18, -10,   2, -11 },
      {  -3,   0,   8,  11, -37,  38, -15,  14,  -3,  -1,  14,   5, -31,  29,  21,  -7,  -8,   1,  18,  -5,   9, -18,  -4,   4, -15,   7,  10, -12, -28, -13, -16,  -3, -16,  11,   5, -25,  -7,  21,  -5,   1,  -7,  15, -12,   4, -10,  13, -15,  -2,  19,  21,   3,  17,  -6,  29,  19, -10, -37,  -9,  -4,   4, -16, -19,  25,  13 },
      {   0,   1,  -5,  -8,  26,  19,  -4,  -5,   3,   6,  -2,   1,  47,  16,  20,   2,  -2,   0, -27,  16,  12, -16, -15,   9,   2, -10,  11,  36, -16,  16,  -1,  -8,   3,  12,   8, -25,  -3, -18,  19,   4,  -4,  -9, -41,  13,   4,  -3,  -4,   1, -34, -13,  24,   9,   6,   1,   3,  -2, -25, -19, -30,  22,  23,  -1,   5,   3 },
      {   1,   5,   1,  12,  31, -21,  25,  -3,  -6,   1,  -6,   7,   8, -28, -38,  11,   1,   6,   9,  18,  11,  29,  16,  -6,   1,  13,  14,   2,   4, -18,  15,  10,   5,  12,  -6,   2,  -4,   0,  11,   0,  -7,   8,  -5,   7, -14, -23,   5,  -1, -11,  33,  -5,  16,  -6,   8,  23,  -6, -23, -25,   3,  -9,  -2, -53,  33,  19 },
      {   1,  -5,   2,   2, -19, -14,  12,   5,   5,  -2,   4,   3, -17,  43,  -8,   8,   7,  -4,   2,   8,  15, -14,  21,   1,   9, -15,   4,  12, -36, -37,  -1,   2,  11, -29,  18, -26,   2, -20,  26,  -9, -10, -12,  12,  -4,   3, -43,   8,  11,  -4, -26,   7, -10,  -9, -14, -11,  -1,  25,   1,  21,  12,  -4, -33,  -6,   2 },
      {   0,   1,   2,  -2,  -6, -23,   1,  -1,   2,   6,   4,   5,  -5,  -9,   6,   3,   0,   1,   2, -10,   1,  12,  -3,  -6,   6,  11,  13,  19,  15, -10,   6,   4,   2,  13,  21, -24, -26, -23,   3,  -1,  25,  21, -29, -30,  20,  -8,   1,  -3,  33, -19, -19,  10,  -6, -12, -11,   7, -60,  33,  26, -10,  20,   1, -15, -10 },
      {   1,   3,  -1,  -8,  -8,   3,  -4,   9,   0,   2,  -3, -24,  -7, -10,   0, -12,   0,   9,  -9, -24,  42, -19,  22,   3,  -1,  22, -25,   7,  15, -38, -11,  21,   9,   1, -46,  -4, -19,   5, -27, -15,   7, -20, -22,  19,  21,  -3,  17, -12,  -6,   7,   4,  16,  -2,  -4,   7,  13,  12,  22, -26,   7,  20, -13,  -7,  -2 },
      {   2,   6,   5,   3,   3,  34,  70,   5,  -5,  -1,  -4,  -7,   0,   9, -62,  -1,  -4,   1,  -4,  -9,  10, -21,  20, -10,  -2,   1, -14,  14,  -6,  19,  -9,   6, -10, -15,   7,  -4,  -4,   1, -14,  -9, -22,  22,   3,   5,   6,  17,  -7,  13,  -2,  -4,  -8,  -4,  -5,   6, -11,  -7, -22,  -2,   1,   5,   6,  26, -12, -12 },
      {  -1,  -1,  -4,  -1, -13, -36,  28, -14,  -1,  -1,  -1,  -5, -32,   6,   1,  18,  -2,  -5,  -5, -23, -13,  -8, -37,  -3,  -1,  -3, -15, -24,  -7,  15,  19,  -9,  -5,  -6, -27,  -2,  32,   6,   4,   3,   9,  -9, -11,  39,  -4,  -4,  32,   4,  -6, -27,  10,  -6,  -3,  -8,  -8,  15, -43,   1, -16,   9, -13,  -8,  20,  -5 },
      {  -3,  -3,  -2,  -3, -14,  19, -29,  -7,  -3,  -4,   1, -12, -21,   8,  15,  -1,  -2,  -5,   0, -13,  -6,  34,  27, -11,   0,  -1,   1,  -7,  19,  -3,   4,  12,  -2, -17, -11,   3,  -9, -10,   1,  -2, -20, -14,  26, -12,   8,   0,   5,   9, -64,  11, -10, -20,   5,  -3, -18,   0, -50, -28,  12,   7,  19,   3,  -6,  -8 },
      {   4,   0,   4,  -3,  -4,  24,  12,  17,   4,  -1,   0,  -6,   6,  31,  -8, -18,   6,   0,   1,  -3,   3,   0,  12,  -6,   7,   7,  -5,  -1,  13,  10,  23,   5,  19,   7, -11,  22,  -6,  -5,   4,   9,  30, -29,  -2, -17, -20,  -9,  -1, -17,  30, -25,  10, -20,  23, -18,  33,  12, -22, -40, -10, -24, -19,  -6, -30,  25 },
      {   1,   0,  -1,   6,   2, -10,   3,  24,  -3,  -5,  -7,   2,  -6, -61,   5, -30,   1,   2,  -1,  13, -22, -10,  29,  23,  -1,  -6,  -3, -18, -24,   5, -41,  -2,   1,  -2, -12,  -6,  16,   8,  12,  -2,   7, -12,  -8,  11,  -9,  -5, -29,   7, -11, -29,   2,  -5,  10,   1,  13, -14, -24,   7,  15,  19,  -5, -13, -36,   3 },
      {   0,  -3,  -1,   6,  -3, -12,  20, -11,   0,  -4,  -2,   8,  -4, -23,  20,  20,  -1,  -5,  -5,   5, -12,  -1, -17,  -3,  -1,  -8,  -5,  17, -24,   5, -31,  -2,  -1, -22,   5, -12, -42,  25, -15, -24,  -6, -21,  13, -10,  28,  16,  18,  -2,  -1,  -3,  31,   8,  -2,  -3,  19,  17,   0, -34,  13, -51,   9,   0,  -2,   5 },
      {   2,   3,   3,   4,   1,  -8,  59,  46,   1,   3,   5,   4,  -3,   6,  40, -18,   0,   2,   3,  -7,   5,  31, -45, -15,  -2,   2,  -1,  -6,  12, -28,  -5, -19,  -2,   3,  -5,   2,  -9,   7,  12,  11,   2,   3,   9,  -7,  12,  -2, -18,   8,  -7,   9, -20, -10,   3,   2,   3,  -9,  15, -15, -14,  24,  27,  -1, -12,   3 },
      {   2,   1,   2,  -3,   0,   3, -15,  15,   1,  -1,   1,  -6, -12,   1,  -9, -26,   3,   0,   1, -11,  -6,  30,  27,  11,   5,  -1,  12, -12,   8,  24, -10,  -9,  -2,   3,  -1, -33,  23, -10,  35, -22,  -1,  22,  -5,  22,  28,   6,  24,  30,  35,   8,   9, -16,  -6,  -6,   3,  10,  16, -21, -30, -24,  32,   2,   6,   1 },
      {   0,  -1,  -4,   2,  -6,  -1,  14, -56,   3,   1,   2,  -1,  -8,  -1,  23,  39,   3,   3,   1,  -8,   7,  12,  28, -10,   0,   2,  -7,  -2,  11,  -5, -38,   1,  -2,   6,  -9,   5,  -1, -14,  -7,  -2,   8,   6, -19,   2, -25,  10, -20,   2,  22, -36, -21, -20,  11,   6,   4, -29,  16, -25,  -5,  12,  26,   1,  22,  14 },
      {   1,  -1,  -2,  -2,  -2,  -4, -16,  46,   2,  -1,  -3,   0,   8,  -6,  -9, -20,  -1,  -3,  -8,   1, -12, -12, -10, -16,   4,   0,  -3,  -3,  -2,  17,   2,  28,   3,  -8,  -8,  -4,  -4,  -7, -13, -21,   1,  -1, -13, -10,   1,   7,  22, -13,   7, -28, -28,  18,  10,  -3,  -8,   3,  14, -39,  41,  33,   8,   9,  55,   8 },
      {  -1,   1,  -2,  -2,  -8, -14,  32, -21,  -1,   3,  -1,  -2,  -6, -15,  44, -15,  -1,   2,  -2,   4, -15, -21,  34,  18,   1,   6,   8,   1, -13,  15,  24,  14,   1,   6,   4, -16,  -8, -14, -11,   0,  -9,   9,  -8, -26,   0,  -7,  14,   4,   0,  26, -36,   7,   3,  -5,  -9,  29,  11, -27, -31,  15, -40,   3,  -7, -15 },
      {   0,  -1,  -1,  -3,   2,   5, -26, -29,   0,   1,  -2,  -5,   0,  -1, -36,  24,   1,  -1,  -1,  -8,   1,   3, -26, -28,   0,   1,  -7, -16,  -8, -10, -12,  -8,   5,  -2,  -9, -16,   8,  18,  33,   3,  -1,   2, -14, -12,  34,   7,  -1,  -3,   8,   8, -29,  31,  22,  -1,  10,   9,   7, -32,  -1,  31, -17,  -5, -47,  -6 },
      {  -1,  -1,  -3,  -3,   1,  -8, -21,  -1,   1,   0,  -1,   3,   1, -16,  -5, -10,  -1,  -2,  -3,   0,  -7, -27,  -7, -26,  -2,   1,  -2,   0, -21, -18,  40,  31,  -2,   4,  -4,  -6,  -4,  29, -10,  23,   6,   6,  -6,  11,   4, -10, -33,  20,   4, -18, -12, -16, -42,   1, -28, -25,   3, -35, -19, -33,  20,  -2, -11,  -9 },
      {   0,   0,  -1,   0,   6,  11,  15,  -1,   1,   0,  -2,   1,  11,  14,  45,   1,   0,   0,  -1,   8,  14,  18,  30, -26,   0,   2,  -4,   9,  16,   7,   1,  16,   0,   0,  -4,  20,  21,  12,  10,   7,   3,  10,   0,  44,  24,  -2,  -6,  17,  12,  -4,  17,  52, -24, -16,  -2,  -5,  -3,  -8,  41,   1, -28,   9,  -9,  -4 },
      {   1,   4,   2,   1,   4,  -3,  -6,  36,   1,   3,   2,   3,   7,  -6,  15,  33,  -2,   1,   0,   3, -10,  -6,   5, -83,  -1,   1,   1,  -5,  -8,  15, -27,  32,   0,   1,   4,  -4,   9, -27,   1, -10,  -9,   2,  -2,  -1, -25,  -2,   9,   3,   1,  12,   8, -12,   8,  15,  -2,  16,   0,  23, -28, -10,  -9, -12, -16,  -4 },
      {   0,   1,  -1,   0,   0,  -3, -12,  25,   0,   1,  -3,   1,   2,   2, -14,  15,   1,   1,  -1,   3,   4,   6, -24,  33,   0,   0,   1,   2,  20, -27, -38,  -8,  -3,   0, -12,  11, -18, -48,  -2, -10,   0,   1,  -6,  13, -24,  13,  -8,  17,   4,  -2,  -4,   9, -30, -12, -27,   0,  -6, -35, -11, -17, -44,   4,  -6, -32 },
      {   0,  -1,   0,   1,   0,  -7,   3, -14,   0,  -1,   1,   3,   1, -11,  -4, -25,  -1,  -2,   0,   0,  -4, -15,   8, -25,  -2,  -1,   0,   3,  -4, -31,   8,  11,  -4,   0,  -5,   3, -34,  -6,  54,  11,   3,  -4,   9,  -1,   3,  58,  -3,  23,  -2,   0,  14, -10,  20, -39,   4, -11,  -4,  20, -11,   6, -24,  10,  35,   7 },
      {  -1,  -1,  -1,   0,  -3,   5,   7, -39,   0,   0,  -2,   3,   0,  12,   1, -40,  -2,  -3,  -1,   0,  -2,  10, -48,   9,   2,   0,   9,  -2,  15,   7,  -5,  65,   0,   1,   5,   1,  11,  -6, -11, -54,  -5,   0,   1,  -2,  -5,  -7, -11,  27,   0,   3,   9,   1,  18,   1,   3, -21,   2,  10,   0,  -1,  -6, -23, -14,   0 },
      {   0,   1,   1,   0,  -2,   2,   6,  12,   0,   1,   2,  -1,  -1,   2,  -2,  19,   0,   2,   0,   0,  -3,  10,  -4,  17,   2,   0,   5,  -2,   0,  15, -30,  36,   3,  -1,   8, -14,  12,   9,   2,  12,   2,   5,   9,  -3,  33,  -9,  -4, -56,  -7,   9,  -8,  -8, -14, -53, -21, -36,  -5,   6, -28,  -3, -20,   6,   8,  39 },
      {   0,   0,   0,   2,   0,   2,  10, -27,  -1,  -2,  -3,   1,  -3,   3,  -3, -54,   0,   1,   0,   2,   0,   5,  -7, -11,  -2,  -3,  -4,   1,  -2,  -1, -33,   5,   0,   0,   3,   3,   1, -15,   8,  56,  -2,   3,  -3,   6, -11,  15,  30, -35,   2,  -2,   7,   9,   0,  30, -50,  16,   7,  -7,  12, -16,   9, -25, -10,  12 },
      {   0,   0,   0,  -1,   1,  -6,  -7,  12,   1,   0,   1,   0,   3,  -3,   3,   8,   0,  -2,  -1,  -3,  -2,  -8,   5,  -3,   0,   0,  -1,   4,   4,  10,  18, -34,  -3,  -2,  -3,  -3, -12, -12, -33,  -2,   2,   0,   0,  -4,   9,  16,  41,  40,   3,   3,   3,  12,  18,   6, -29, -62,   0,   0,  -3,   4, -13, -22, -27,  51 },
      {   0,   0,   0,   1,  -2,   2,   8,  28,   1,   1,   1,  -1,  -4,   2,   5,  53,   1,   1,   0,  -1,   0,   1,   6,  50,   1,   2,   1,   1,   2,   4,  21,  52,   1,   4,   0,   5,  -2,  15,  26,  18,   3,  -1,   0,   1,  -1,  39,  -5,   9,   9,  -1,  11,   2,  35,  23, -28,  14,   7,  -1,  10,   2,  18, -24, -13,  -6 },
      {   1,  -1,   0,   0,   0,   0,   3,   6,   1,  -1,   1,  -1,   3,   5,   3,  -7,   1,  -1,   0,  -1,   2,   0,   9,  -9,   0,  -1,   1,   1,   1,  14,  -2, -12,   3,   0,   5,  -2,   5,   6, -12,  12,   5,  -4,   3,  -7,  14,   6,  11, -24,   8,   1,   9,  -2,  30, -14,   7, -53,   1,  -5,  -4,  -6,  -1, -40,  12, -91 },
      {   0,   1,   0,   1,   2,   4,  -6,   0,   0,   0,   0,   0,   5,   6,  -8,  -2,  -1,   0,   0,   1,   1,   2,   0,  -8,  -1,   1,  -1,   2,   1,  14, -18, -20,   0,   1,   1,   0,   7,   8, -44,  19,   2,   1,   3,  -3,  18,  -1, -49,  37,   4,   1,   3,  -2,  27, -44, -28,  51,   0,   3,  -1,   3,   0, -44,  23,  10 },
      {   0,   0,   0,   0,   1,  -2,   5,  -1,   0,   0,   0,  -1,   3,  -1,   9,  -5,   1,   0,   2,   0,   4,  -2,  12, -13,   0,  -1,   0,  -2,   1,   0,  27, -36,  -1,   3,  -2,   5,   0,  12,  20, -75,  -4,   4,  -6,   8,  -9,  23, -32, -52,  -1,  -2,   2,   1,   0,   1, -51,   7,   0,  -3,   0,  -7,  -3, -22,  -7,   7 },
    },
  },
  { // 27
    {
      {-105,  34,  16,   3,   3,   1,   1,   0,  49, -34,  -4,   1,  -1,   0,   0,   0,  12,   7,  -9,  -1,   0,  -1,   0,   0,  -2,   1,   3,  -2,   0,   0,   0,   0,   3,  -1,  -1,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  18, -49,  21,   3,   2,   0,   1,   0,  76,  12, -42,   0,  -5,   0,  -2,   0, -47,  49,  13, -11,   0,  -2,   0,  -1, -11, -17,  16,   4,   1,   1,   0,   0,  -3,   0,  -5,   2,   1,   0,   0,   0,  -1,  -2,   2,   0,   0,   0,   0,   0,  -1,   1,   0,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0 },
      { -51, -63,  32,  18,   5,   3,   2,   2, -32,  57, -24, -11,   3,  -1,   1,  -1,  17, -31,  28,  -3,  -2,   0,   0,   0,  24, -13, -19,   9,  -2,   2,  -1,   0,  -1,  13,  -4,  -3,   0,  -2,   1,   0,   2,  -2,  -1,   0,   1,  -1,  -1,   0,   0,   3,  -1,  -2,   0,   0,   0,   0,   2,  -1,   0,   0,  -1,   0,   0,   0 },
      {  -4,  48, -27,   1,  -3,   0,  -2,   0,  19,  25,  15, -14,  -2,  -4,   0,  -2, -70, -33,  43,  16,   3,   2,   1,   1,  37, -34, -27,  10,   4,   2,   1,   1,   0,  14,  -5,  -9,   1,   0,   0,   0,   3,  -4,   2,   1,  -2,   1,  -1,   0,  -1,   1,  -2,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
      {  10, -44,   0,   6,   3,   0,   2,   1,  42, -27,  24,  -6,  -3,   1,  -1,   0,  -6, -34, -26,  33,  -3,   5,   0,   3,  19,  51, -52, -16,   6,  -3,   0,  -1, -21,  23,  28, -18,  -2,   0,   0,   0,   0,  -6,   4,   5,  -3,   0,   0,   0,  -3,   3,   1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0 },
      {   3,  11, -40,  18,  -2,   4,  -1,   0,  40,  48,   2, -39,   2,  -7,   0,  -2,  42, -17,  27,  10, -14,   0,  -2,   0, -63,  21,  -3,  13,   9,  -3,   2,  -1,   3, -26,  13,  -7,   5,   3,   0,   0,   0,  -3,  -6,   4,  -4,   2,   1,   0,  -1,   0,   1,  -1,   0,  -1,   0,   0,  -1,  -1,   0,   2,   0,   0,  -1,   0 },
      {  17, -12,  13,   0,   0,  -2,   0,  -1,   6, -51,   8,  -4,   6,   4,   2,   1,  28,  -6,  29,   5, -18,   1,  -4,   0, -24, -35, -29,  53,   7,   2,   1,   1,  10,  29, -50, -28,  16,   0,   2,   0,  -4,  17,  19, -15,  -9,   2,  -1,   0,   1,   0,   0,   6,  -3,  -1,   1,   0,  -3,   4,   1,   0,   1,  -1,   0,   0 },
      {  13,  33, -17,  10, -11,   1,  -3,  -1,   1,  -4, -61,  -2,  14,   0,   3,   2,  31,  16,  15, -36,  -5,   0,   0,  -2,  32,  19,  -8,  12, -14,  -2,   0,  -1, -36,  42,  28, -20,   4,  -8,   0,  -1,  -3, -26,  29,  11,  -4,   2,   1,   1,  -7,  -4,  -4,   7,  -1,   1,   1,   1,   1,  -3,   1,   4,   0,  -1,   0,   0 },
      { -22, -37, -72,  43,   3,   7,   2,   3, -20, -25,  27, -34,  10,   1,   2,   0,  -4,  56,   4,   0, -13,  -3,  -2,  -2,  25, -14,   5,  -1,   5,  -2,   0,   0,   1,   1,   4,   0,  -5,   3,  -2,   0,  -3,   5,  -3,   5,   0,  -3,   2,   0,   1,  -2,   0,  -3,   1,   0,  -1,   0,   1,   1,   0,   1,  -1,   0,   0,   0 },
      { -12,   6,   5,   6,  -2,   1,   0,   0, -32,   5, -25,   4,   9,   3,   0,   2, -31,  18, -20,  12,  -2,   3,   2,  -1, -44,  -9, -27, -23,  27,  -2,   8,   0,  42,   6,  25, -58, -20,   6,  -4,   2,   4,   0,  29,  22, -26,  -6,  -1,  -2,   3,  -4,  -4,  10,   6,  -6,   1,   0,   1,  -1,   2,   3,   1,   1,  -2,   1 },
      {  -1,  -4, -45,  -7,   6,   2,   2,   2,  24,  35,  -6,  19, -17,  -2,  -1,  -2,  25, -10, -41,  -2,  27,  -4,   4,  -2,   7, -23,   1, -33,  -2,   7,   0,   2,  12,  36, -32,  -5,  -8,  -7,   3,   0,  -9,  31,  34, -32,   3,  -7,  -3,  -1,   2, -10,  24,  10,  -7,   4,  -4,   1,  -3,   0,   1,   3,  -1,   0,   0,   0 },
      {  18,   9,  34, -13,   4,  -6,   1,  -2,  19,   9,  31, -30, -11,   2,  -1,   0,  43,  16,   6,  19,  -5, -11,  -2,  -2,  20, -45,   9, -21,  12,   9,  -1,   2,  14,  25,  24,  11, -37,   3,  -5,   0, -20, -12,   7,  43,   7,  -7,   2,  -2,  -1,  -8, -22,   6,  11,   2,   0,   1,  -4,   3,  -7,  -3,   4,   1,  -1,   0 },
      {  15,  -1,   5,  49, -13,   2,  -3,   0,   9, -34, -28, -31,  29,   1,   6,   1, -11, -55,  -8, -27,   1,   9,   3,   2,  -3,  -6,  19, -17, -24,   5,  -1,   0,  44,   6,   1,  19, -25,  -8,  -2,  -1, -11,  22, -13,  11,  10,  -9,   2,   0,  -1,   4,   5,  -5,   0,   1,  -2,   1,  -2,   4,  -2,   6,   0,  -1,  -2,  -1 },
      { -14, -24, -17, -49,  10,   6,   3,   0,   1,  -2,  29,  -1,  -1,  -9,  -1,  -1, -17, -13,  23, -35,  -5,   9,  -2,   3,  -6,  30,  25,  13, -35,  -6,  -4,  -3,  36,  12,  13,   6,   5,  -7,  -2,  -2, -15,  -4,  39,  14, -15,   5,  -3,   0,   1, -22, -16,  38,   6,   0,   2,   0,  -5,   5, -14,  -3,   7,   1,   2,   1 },
      {  -8,  -8, -27, -17,  21,   3,   1,   2, -16, -23, -25,  19, -30,   9,   0,   2,  -6,   0,  40,  13,  16, -16,  -4,  -2, -39,   7, -24, -15,  -5,  20,  -1,   2,   5,  26,   3,  29, -22, -23,   7,  -4,  -8, -17,  -9,   9,  47,  -7,  -7,   1,   6,   5, -11, -26,   2,  16,  -3,  -1,  -1,   2,   9,  -6,  -5,  -1,   4,  -2 },
      {  -7,  13,  23,  31, -15,  -2,  -8,  -7, -13,  25,  25, -16,   5,  -4,  -2,   4, -24,  21, -38,  19,  -4,  12,  -2,   0, -35,   2,  -4,  31, -34,   1,   4,  -3, -12,  47,   7,  19,  10, -30,   3,  -4, -17, -13,   4, -14,  27,   5,   1,   2,  -8, -14,   5,   7, -14,   8,   5,   0,   2,  -3, -13,   8,   8,  -5,   4,   1 },
      {   4,  17,  22, -11,  17, -12,   3,   4, -11,   3, -20, -55, -13,  16,   4,  -1, -11,  15,   8,  18, -35, -16,  -2,  -1,  21,  50,   6, -11,  25,  -9,  -7,  -2,  20,   1, -24,   8,   6,   5,   5,  -2, -15,  19,  25, -23,  17,  -7,  -7,   0,  18, -21,  23,  -7,  -5,  11,  -9,  -1,  -5,  -2,   6,   2,  -6,   3,   3,  -1 },
      {  -5,  -6, -13, -51,  36,  -1,   3, -12,  -8, -11, -26, -30,   3,  10,  -7,   9,  -2,  -8,  -9,  22, -20,   6,   8,  -3,  -5, -23,  12,  10, -14,  -3,   4,   0, -15,  11,  26,   2, -19,  -3,  -7,   1,  16,  13, -30, -10, -15,   0,  14,   1, -23,  18,  35,  16, -26, -10,   7,   2,   7,  -8, -13,  34,   9,  -6,  -6,   2 },
      {   1,   2, -13, -21, -28,  12,  -3,  -2,   7,  10, -21,  29,  12, -15,  -3,   1,  14,  19,  -3,  26, -12,  15,   0,  -1,  25,   7, -24,  19,  -4, -22,   6,  -2,  56,   4,  14,   9,  19,   2, -14,   1, -53,  16, -38,   3,   5,  14,  11,   0,  -2,   5,   2, -18,  -1,  -8,   4,   2,  -5,   0,   2,   2,   0,   2,  -1,   0 },
      {  -4,  -4, -25,   9,  15,  -6,   3,   0,  -2,   2, -29,  10, -22,  26,  -6,   3,  -9, -31, -40,  30, -30,  -9,   9,   1,   0,  -4,  40,  30,  15, -16,  -7,  -2,  -6,  15, -19,   6,  12,  16,   0,  -1,  -8, -19,  -1,  23,  -4,  -3,   9,  -1,  -6,  -3, -41,   2,  26,  -6,   6,   2,   2,   9,  -2, -29,   5,  -2,  -1,   0 },
      {   5,   0,   5,  38,  25,  -7,  -4,  -4,  -1, -14,  -4,  10, -24,   8,   0,   2,   7,  -8,  20,  19,  25, -13,  -1,  -1,   7,  -6,  20, -15,  10,   1,   3,  -3,  24,  -5,  20,  -5,  28,   3,  -7,   5, -20, -39,   4, -31, -11,  40,   3,   6,  14,   2,  -3,  24, -47,  -8,  10,   1, -13,   5, -13,  15,  16, -10,  -3,   2 },
      {   1,  -2,  -1, -26,  22,   1,  10,   5,   4,   6,  14, -32,  14,  -2,   4,  -5,  -2,  -4, -29, -38,  -4,   5,   1,   4,  -7, -17,  -1, -10,   0, -13,   1,   1,  18,  22,  -1, -10,  30,   3,  -4,  -2,   4, -46,  -5, -10,   4,  19,  -3,   3,  22,  32,  -9, -46,  -9,  -5,  -4,   0,   4,  -1,  42,  -2, -23,   1,  -5,  -1 },
      {  -7,  -3,   0,  -3,  13,   1,   4,  -3,  -4,  -1,   8,  14,  38,   1,  -9,   0, -12, -12,  11, -13,  -7,   9,   8,   2, -21,  -7,   6, -13,  19,  -2,  -2,   4, -48,   6,  -2,   1, -16,  30,   1,   0, -77,  -5,  -4,  -9, -10, -15,  17,  -1,  31, -25,   8, -10,  -5,  -7,  -6,   5,  -8,   2,  -5,  12,  -7,   1,  -4,   1 },
      {   8,  15,   6,  28,  80, -17,   0,  -7,   7,  15,  12,  18, -20,  20,  -2,   3,   0,  10, -14, -46, -11,  -1,   3,   2,   9,   6, -36,  27,  -8,   6,   0,  -1,  12, -14,  10,  -5,  -8,  -5,   4,   2, -18,  26, -15,  10,  -1,  -5,  -5,   0,  -9,   2,  -2,  -3,   6,  -2,   6,  -3,   2,  -1,  -3,  -1,   5,  -1,   0,   4 },
      {   2,   0,  -6, -19,  22,   0,   4,   0,  -2,  -6, -30, -35,  23, -14,   5,   1,   3,   2, -24,  23,  23,  27, -15,   2,   9, -17, -26,  -1, -26,  22,  17,   1, -15, -51, -13,   0,   5, -22,   6,   3, -19,  -4,  12,  -4,   5,  10, -14,   2,  18, -14, -37,  12,  -1,   4,   0,  -4,  -9,  17, -10, -21,   6,  -1,   7,  -4 },
      {   0,  10,   8,   4,  11,  13,  -4,  -3,  -7,   5,   1, -24, -11, -16,   7,   2,  -4,  17,  12,   9,  41,   7,  -9,  -1,  -3,  38,  11,  -7, -14,  27,   5,   1,  -2,  30, -39,  -9, -13,  13,   3,   2, -20,   5, -23,   3, -39,  -3,  18,   1, -22,  32, -10,   1,  12, -39,   3,   1,  -3,  -5,  16, -18,  13,   4, -10,  -1 },
      {   1,  -2,  -1,  -1,  34, -12,   4, -11,   6,   4,   6,  15,  59, -16,  -8,   2,   4,   1,   8,  20,  21,   8,   4,  -6,  11,  18,  31,  11,  23,  16,   1,  -3,  19,   7, -13, -26,  -6,  -2,   7,   4,   8, -19,  -4,   0,  19,  -1,  -9,   8, -50, -17,   1, -15,  -5,  33,   1,  -2,  14,  -1,  -9,   1, -23,  -4,  10,   0 },
      {   3,   5,   8,  11,  33,   9,   4,   5,  -4,  -7,  -5,  -2,   9, -18,   5,  -4,   1,   2,  18,   3,   0,  17,   3,  -2, -15, -11,  -9, -41,   3, -40,  12,   7, -13,  18,   3,  26,  36,   1, -38,  -3,  18,  24,   5,   1,   4,  28,  10, -13, -27, -33,   1,  -3,  22, -14,   7,   9,  14,  -2, -10, -28,  -1,  14,  -8,   2 },
      {  -5,  -8,   0, -14, -16,  -7,  -8,   0,  -8, -12,  -5, -30, -30, -15,  16,  -6,  -9, -14, -13, -19,  47, -30, -10,   2,  -5, -11,  14,  29,  29,   7,  -8,  -5, -15,   0,  39, -23,  20,   3,   4,   5, -28,  36,  -9,   1,  10,  17,  -6,   0, -20, -21,   5, -15,   7,  11,   8,   1,   4,  -2,   0, -12, -11,   4,  -2,   5 },
      {  -4,  -2,  -8, -13, -20, -55,  10,  -7,  -2,   3,  -4, -16,  27,  21, -13,  -3,  -1,   9,  -6, -22,  21, -17,  25,   0,  -9,   1, -38,  -8,  17,  -2, -31,   4,   9,   6,  -7,  32,   9,  27,   1,  -9,   3, -11, -12, -12,  -3,   2,   2,   5, -26,  13,  -9,  21,   0,   5,  -1,  -5, -24,   5, -23, -18,  24,  -5,   9,  -1 },
      {   3,   1,   3,   9,  22,   1,   6,  -5,  -4,  -8,  -5,  -1,  -8, -46,   0,  -2,   1,  -3,  -2,   6,  11,   7,   0,  -3,   0,  -9,   8,  11,  11, -24,  -1,   2,   3,   3,  14,   0,  17,  -5,   3,  -2,  -6, -18,  -2,  -7,   4, -39, -13,  11,  14,  33,  28,  19,  39,  11, -45, -11, -27, -15, -15,  -2,  15,  37,  20, -12 },
      {   5,   4,   2,  -4,  13,  70, -14,  13,   3,   0,  -7,  -1,  13, -23,  27,  -2,  -2,  -1, -19, -10,  -4, -35,  -9,   0,   1,  -1, -25,  16,  21,  20, -27,  -2,  15,   1,  -1,  28,  -9,  21,  13,  -9,  19, -24,  -8, -14,  -7,  -5,  16,   7,   2, -28,  -2,  10,  -6,  -6,  -6,   6,  16,   1, -20,   0,   7,   3,  -4,  -1 },
      {   2,  -3,   1,   3, -13, -10,  11,   1,  -1, -10,  -3, -14, -39,  14, -18,  10,   6,   0,   0,  -1,  23,  22,  28,  -3,   4,  -6, -16,  10, -21, -12,   6,   7,  14, -25,  -7,   0,  -7,   7,   1,  -8, -10, -43,   6, -11, -10, -24,  22,  11, -27, -40,  19, -11,   9, -12, -22,  12,  40,   0,  -1,   8, -25,  13, -13,  -2 },
      {  -1,   1,   5,   1,  -7,  44,  -4,   5,  -4,  -4,   3,  -4,  -6,   3,  -6,   0,  -4,  -5,  -1,  -6, -22, -16,  23,   7,  -4,  -8,  -7, -20,   2,  -7,   2,  -1, -11, -25, -12,   4,   3,  -8,  -4,   0, -47, -10,  30,  -1,   1,   8, -26,   0, -58,  46,   6,   9,   0,  18,  14, -18,  17, -15,   0,  -3,  -2,  -9,  13,   4 },
      {   0,  -2,  -4,   5,   5, -41,   4,  -5,   0,  -2,  -5,   7,  14, -31, -15,  -5,  -1,  -2,  -4,  14,   5, -17, -28,  14,   0,   1,  -3,  14,   3,  14,  -4,  -6,   4,  -7,   8,  34,   0,  12,  -2, -11,  -7,   8,  42,   5,  -5,   7,  15,  -8,  19,  31,   1,  -3,   4,  -8,  10,   6,  67, -31,   5,  11,   2,   7,  -8,   4 },
      {   1,   0,   3,   0,   1,  -6,  -9,  14,   2,   0,   3,  10,  19,  35,  14, -12,   4,   1,  14,  17, -19,   1,  -8,   2,  -3, -15,   9, -20, -19,  19,  -8, -10,   6,   8,  39,   1,  28,  -4,  56,   0,  -6,  15,  -6, -35, -18, -18,  -8,  29, -13,  -6,  -9,   5,  13, -12, -16, -11,  20, -13,  20, -30,  -4,   9,  -1,   0 },
      {  -2,  -6,  -3,  -7,   3, -10, -16,   5,   4,   1,   5,   5,  28,  28,  29,  -7,   0,  -3,  -3,  10,   8, -41,   4,   9,   3,  -2,   0,   3, -15,   6,  12,  -6,   3, -17, -12,  -9,  14, -27,  -5,   6,  -5, -24,  13,  21,  12,  -3,   6, -15,  -4,  -8,  37, -17,  27, -30,  11,  17, -31, -32, -10,   5,  20,  14, -33,  15 },
      {   1,   2,   1,   7,   4,  23,  24,  11,  -3,  -3,  -4,  -1,  -7,  15, -33,  -3,   0,  -2,  -4,   5,   8,  36, -19,   0,   1,  -6,  -2,  15,   6,  26,  -1,  -2,   5,  -3,  13,  28,   5,  26,  12, -11,  10,   3,  26,   7, -10,  14,  15,  -3, -12,  -6,   9, -13,   6,  10,  12,   0, -60, -50,  13,   4, -20,  -5,   8,  12 },
      {  -1,   0,   4,   3,   8,   8,  -8, -10,   0,   3,   5,   4,  15,  -1,   8,   8,  -3,   0,   3,  11,  11, -21,  16, -16,   2,   1,   3,  17,  -5, -22,  -2,  15,  -4, -15,  21,   6, -13,  -1,  -7, -16,   3,  24,  41, -20,  20, -12,  38,  11, -24,  20, -34,  -6, -14, -14, -26,  25, -29,  24,  37,   9,  -5,   2, -15, -21 },
      {  -4,  -6,  -6,  -5, -11, -30, -16,  -5,  -2,  -2,  -1,   6, -18, -29,  24,   4,  -4,  -8,  -5, -23, -37,  35, -33,  -4,   1,  -3,  -3,  -7,  46,  15,  31, -13,   1,  -3,  -4,  21, -10, -17,  16,  14,  -4, -18,   8, -11,   2,  -3,   0,   6, -39,  -3,   0,  -4, -11, -24,  -4,   8, -16,  12,   2,   2,  13,   0, -20,  -5 },
      {  -3,  -2,  -3,   4,   9, -28, -58,  -1,  -3,  -1,  -6,   6, -10, -24,  30,   2,  -1,   1,  -2,  17, -14,   2,   3,   7,  -5,  -8,  -6, -10, -33,  13, -45,  -6,  -5,  -5,  -9,   6,  11,  26, -12,  -9,  -3,  -6,   4,  26, -23,   9,  -6,  -3, -13, -13,  15,  -3,  -5,  15,  -5, -10, -19,  14,  27,  19, -28,   6,  20,  -7 },
      {  -1,  -3,  -2,  -8,   3,   4, -72,  13,   2,   2,   3,  -5,   2,  41,  25, -15,   4,   3,   3,  -4,  24,  50,   1,   0,   2,   4,   4,  13,  20, -23,   5,   6,   3,   1,   2,   2, -23,  10,  -6,  -8,   2,  -1,  10, -10,  14,   4,   9,  -4,  15,  25,  -2,   3,  -1,  10,   5,   7,  16, -21, -24, -16,  -4,   6,  12,  -4 },
      {   0,   1,  -1,   0,   3,  -5,  32,  45,  -1,   1,  -1,   4,  -5,   4,  28, -41,  -2,   0,  -3,   2,   5,  16, -13,   1,  -3,   1,   3,  -3,  -4, -13, -16,  24,  -6,  -2, -15, -13,  10,   9,  10, -18,  -6,   9,   0,  38,  17,   6,  -4,  28, -14,  -4,  -7,  -8, -27, -13, -27,  -4,  10, -18, -11,  33,  32, -15,   4, -24 },
      {  -1,   2,   1,  -2,   0,  15, -17,  -8,  -1,   3,   2,  -4,   0,  24, -38,   5,  -1,   1,   0,  -6,   1,  -6, -45,  17,  -1,  -2,  -1, -16, -12,  12,  17, -19,   0,   1,  -3, -12,  33,  28,  14,  -7,   1,   4,   8,  20,  15,   0,  22,   5, -10,  20,  18, -16,   5,   7, -15,  16,   2,  55, -36,  10,   5,  13,   5, -15 },
      {   0,  -2,   2,  -3,   3,   0, -31,  -4,  -2,  -5,  -1,  -5, -12, -39, -35,  -8,   1,   0,   2,  -5, -17,  16,  59, -13,   3,   0,   5, -13,   4,  30, -17,   8,   1,  -4,   1, -20,  24,  -9,  26, -20,   7,  10,   6,  10,  25,  -6,  24,  16,  15,   2,  -2, -15,  -3, -16,  17,   8,  -1,  -9, -17, -11,  20, -19, -10,  15 },
      {   1,   2,   2,   4,   1, -11,  33,   2,  -1,  -1,  -1,   5, -11, -23,  29, -26,   1,   3,   4,   5, -17,   4, -14,  32,   1,  -5,   3,  -3, -30, -14, -15, -12,   2,  -7,   7, -29, -33,  27,   6, -21,  -1, -12,   2, -30,  25,  -1,  21, -10,  -6,  14,  15,  -1,   4,   5,  28,  13, -13,  11, -20, -40, -13, -15,  -5,  15 },
      {   1,   3,   3,   3,   9,   0,   7,  39,  -2,   0,  -3,  -1,   6,  -3,  10, -32,   2,   5,   2,  12,  24,  10,  27,   1,  -1,   2,  -1,  16,  14, -12,  15,  -6,   1,   7,  10,  33,   8,  -4,  23,   7,   3,   2,  15,  12, -20, -17, -14,  -1,   6,  14,  21,  -2,  14,  -3,  25,  -7,  -1,  63,   4,  20, -14, -25, -10,  33 },
      {   0,   1,  -1,   0,  -2,  -1,   7, -43,  -1,   1,  -2,  -3,  -8,  -4,  36,  -4,   0,   3,   0,  -1,   2, -10,  10,   3,  -3,   2,  -9, -16,  -5, -13,  24,  -9,   0,   6, -12,  10,  25,   6,  23,  19,   2,  11,  -3,  15, -13, -19,  11,  17,   7,  12, -32,   4,  -7,  17,  -3,  23,  -7, -30, -40,  23, -52, -20,  -9,   5 },
      {   0,   1,   0,   0,   0, -19, -26,  64,  -2,  -1,  -2,   0,  -3, -19, -36, -31,  -2,  -1,  -5,  -2, -17, -36,   0,  -5,  -1,   3,  -1,   2, -19,   8,  43,  22,   1,   6,   1,  -5, -13,  10, -16,  11,   2,  -1, -12, -22, -15,  -3,  -8,  -1,  -7,  -1, -23,  -5,  -5,  10,  -8,  11, -15, -12, -23,   0, -23,  13,  -4,  -5 },
      {  -1,   2,   0,   2,  -5,  -2,  30,  11,   0,   4,   3,   7,   0,  14,  23,  -9,  -3,   1,  -2,   0, -15,  11,  18, -21,   0,   4,   4,   7,  12,  58, -16,  20,  -1,   1,   9,   5,  12, -26, -33,  10,  -7,   3,   5,  10, -10,  16,  18,  -5,   7,  24,  25,  11,  -6,   2, -22,  26,   6,  17, -20, -30, -36,  14,   2, -22 },
      {   0,  -1,   1,  -2,   2,  -7,   7,  29,   2,   1,   3,   0,   1,  -3,  29,  45,  -1,  -3,   0,  -2,   1,  -3,  -2, -67,  -1,  -2,  -2,  -2, -17,  10,  31, -10,   3,  -1,   6,  -4,  20,  45,  -7,  11,  -2,  -3,   6,   3,  19, -29,  10, -17,   2,   3,  10,   1,  -3,  10,  13,  -1,  10,  -7,   6, -11,  20, -28,  21,  -1 },
      {   0,  -1,   0,  -1,   2,   0,  -6,   3,  -1,  -2,  -2,  -3,   2,   4,  -8, -37,   0,  -1,  -3,  -8,   0,   0,   0,  12,   1,   0,  -5,  -6,  13,  13,  21, -17,   0,  -3,  -8,  -8,  15,  -2,  -7, -10,  -6,  -7, -27,   5,  34,  -2,  31,  -9,  -1,  -7,   3,  73,  22,  11,  -1,  18,   7,  -6,  52,  17, -10, -16,   3,  -6 },
      {  -1,  -2,  -2,  -1,  -4,   7,   3, -45,   0,  -1,  -1,   1,  -5,   6,  20, -43,   1,   0,   0,   2,  -5,   4,  26,  31,   2,   2,   4,   8,  -6,  37,  39,  24,   5,   4,  15,   7,   8,  34, -30,   9,   8,   0,  11, -17,   0, -17, -13,  -9,  -2, -12,  -5, -25,   5,   0,   1, -14,   5,   6,   7,   8,  34,   1,  24, -11 },
      {   0,  -1,   0,   1,   3,   2,  -8, -16,   0,  -1,   0,   2,   5,  -7, -14, -19,   0,  -1,  -2,   2,   5, -17,   7,   5,  -1,  -2,  -4,  12,  -4, -21,  30, -12,  -2,  -1,  -4,  14, -10,   7,  12,  13,   3,  -1,   5,  41,  -8,  -1,   0,  28,   7,  -9,  40,   6, -55,   1,  -9,  -4,  -3,  -6,  21, -66,   0,  -7,  -1, -19 },
      {   1,   1,   2,  -1,   2,   2,   2,  35,   0,   0,   1,  -2,  -1,   0,  -2,  24,   2,   3,   4,   0,   9,  12,   9,  60,   1,  -1,   3,  -2,  11,   6, -15, -41,   0,   1,   6,   0,  22, -19, -33,   2,   1,   6,  -2,   1, -14, -57,   8, -20,  -4,  -1, -17, -11, -34,  19,   2,  12,  -2,  -9,   1,  -9,  14,  -7, -10, -13 },
      {   0,   2,   0,   2,   2,   3,   2,  16,  -1,   1,  -1,   1,   3,  -5,   0,  34,   1,   4,   0,   4,   2,  13,  34,  35,  -2,   1,   0,   9,  -5,  12,   0, -19,  -1,   5,   5,   1, -10,  45,   9,  40,   5,   8,   4,   0,  38,  32, -44,   6,   4,   6,   1,  18,  12, -27, -24,  16,   1,  -4,  -1,   5, -10,  -1, -29, -16 },
      {   0,   0,   0,   1,   1,   6,  -7, -14,  -1,  -1,  -2,  -1,   0,   2, -13,  -5,   0,   0,  -2,   2,   2,  10, -31,  -3,   0,  -1,  -1,  -1,  -2,  -2, -27,  67,   4,  -1,   3,  -1,  22,  16,  -1,  23,  -3,  -4,  -3,   9,  21, -44, -30, -22,  -8,   5,   1,  23, -29, -10,  14,  18,   1,   2,  10,  -9, -11,  11, -32,  30 },
      {   0,   0,   0,   1,  -2,   4,  13,  -1,  -1,   0,  -1,   0,  -3,   3,   0, -36,  -1,   0,  -1,   0,  -1,   9,  16, -35,  -2,   1,  -3,  -1,  -6,   2, -12, -58,  -4,  -1,  -6,  -4,  -4,  13,  -3,  24,   1,   4,   3,   4,  15,  -4,   1,  -2,  -4,   3, -19,   1, -29,   0,  14,   1,   2,   2,   3,  11,   7,  79,   0,  27 },
      {  -1,   0,  -2,   0,  -4,  -5,   7,  11,   1,   3,   2,   3,   2,   2,  21,  41,   0,   0,  -1,  -2,  -3,  -4,  11,  28,   1,   3,   3,   1,  10,  13,  39,  19,  -2,   3,  -3,   1,   4,   7,  -4, -66,   1,   6,  -7,  13,   9,   2,  -5,  30,   2,   2,  -1,  16, -20,  -7,   6,  -9,  -3,   4,   3,  -2, -13,  43,  12,  46 },
      {   0,   0,   0,  -1,   0,   0,  -3,  -4,   1,   0,   0,   0,   3,   3, -11, -10,   0,  -1,   0,  -1,   5,  -8,  -8,  -5,   1,   0,   1,   3,   6,   1,  -4, -23,   1,   0,   3,  -2,   5,  -6, -15, -14,   3,   1,   3,   0,  23, -23,  -9, -23,  -2,  -2,  -5,   5, -12, -79,  -2, -49,  -5,  -3, -21,   0, -48, -16,  32, -14 },
      {   0,   0,   0,   1,   0,  -2,   7,   2,   0,  -1,   0,   0,   0,  -6,   4,   5,   1,   1,   1,   1,   1,  -3,  21,   7,   1,  -1,   1,  -1,   3, -10,  11,  14,   2,   2,   1,   2,   1,  -5,  62, -18,  -2,  -1,  -6,   7, -10,  10,  -9, -76,  -4,  -3,   0,  -1, -16,  -7,  12,  36,   1,  -1,   3,   3,   2,  26,  34, -37 },
      {   0,   1,   1,   2,   0,   5,   0, -15,   0,   0,   0,   1,  -1,   4,  -3, -14,   0,   0,   0,   2,   2,   5,  -2, -33,   1,   0,   2,   2,   0,   6,   4, -25,  -1,   3,   3,   2,   3,  10, -16, -68,  -3,  -1,  -6,  -4, -11, -11, -58,   7,   4,   1,   2,   1,   6,   7,  22,  21,  -2,  -1,  -3,   0,  -4,  -4, -50, -43 },
      {   0,  -1,   0,   0,   0,   2,  -5, -10,   0,  -1,   0,   1,   0,   1,  -6, -13,   0,  -1,   0,   2,  -1,  -1,  -3, -11,   0,  -1,   2,   3,  -2,   1,  -1, -24,  -1,   1,   2,   0,  -5,  -1, -11, -21,  -2,   2,  -2,   0, -11,   8, -29, -25,  -5,   0,  -3,  -9,  -8,  -9, -68,  47,   4,  -3,   1,  -9,  12, -33,  21,  65 },
      {   0,   0,   1,   0,   0,  -1,   1,   3,   0,  -1,   0,  -2,   1,   0,  -7,   4,   0,  -1,   0,  -1,   1,  -4,  -9,   4,   0,  -1,  -1,  -1,   2,  -5, -12,  -1,  -1,  -2,  -1,   2,  -3,  -8, -15,  17,   0,  -3,   2,  -5,  -1, -13, -20,  55,   1,  -4,   4,  -4,  11, -28,  45,  68,  -1,  -1,   8,   5,   0,  -1,  62,  -9 },
    },
    {
      { 110, -31, -13,  -4,  -3,  -1,  -1,  -1, -39,  30,   0,   0,  -1,   0,   0,   0, -24,  -2,   8,   0,   1,   0,   1,   0,   4,  -2,  -2,   1,   0,   0,   0,   0,  -5,   2,   1,  -1,   0,   0,   0,   0,   1,  -1,   2,   0,  -1,   0,   0,   0,  -2,   0,   1,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0 },
      {  13, -25,   7,  -1,   2,   0,   1,   0,  87,   2, -26,  -1,  -5,  -1,  -2,   0, -56,  53,  10,  -3,   0,  -1,   0,   0, -21, -20,  17,   5,   2,   1,   1,   0,   2,  -3,  -7,   2,   1,   0,   0,   0,  -4,  -1,   2,  -1,   1,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,  -2,   0,   1,  -1,  -1,   0,   0,   0 },
      {  35,  -2,  -9,  -3,  -2,  -1,   0,  -1,  25, -35,   7,   1,   0,   1,   0,   0,  55,  28, -41,   2,  -5,   0,  -2,   0, -42,  63,  16, -15,   0,  -3,   0,  -1, -13, -21,  28,   6,   0,   1,   1,   0,  -3,  -3,  -5,   4,   2,   0,   1,   1,  -3,  -3,   3,  -1,   0,   1,   0,   1,  -2,  -1,  -1,   2,   0,   0,   0,   0 },
      {  -1, -41,  11,   5,   2,   1,   1,   1,  37,   3, -13,  -3,  -1,  -2,   0,   0,   8, -30,   1,   9,  -2,   3,   0,   1,  26,  33, -58,  -2,   0,   0,   0,   0, -27,  54,  25, -31,   1,  -6,   1,  -2,   0, -20,  29,   9,  -6,   1,  -1,   0,  -3,   2,  -5,   6,   1,   0,   0,   0,  -1,  -2,   3,  -1,   0,   0,   0,   0 },
      {  26,   2,  10, -10,   1,  -2,   0,  -1,   4, -65,   9,  16,   2,   3,   1,   2,  23,   6, -16,  -4,  -1,   0,  -1,   0,   6, -28, -10,  19,  -3,   3,  -1,   1,   8,  39, -59, -10,   4,  -1,   1,   0, -15,  36,  28, -28,  -4,  -4,   0,  -1,   3, -10,  13,  10,  -3,   0,  -1,   0,  -4,   5,  -1,   1,   2,  -1,   0,   0 },
      {  27, 100, -29, -12,  -9,  -3,  -4,  -2,  24,  -2,  11,  -2,   0,  -3,   1,   0, -20, -18,  16,   2,   4,   2,   0,   0, -20,  -1, -11,   6,   1,   3,   1,   1,  -7,  14,  11, -16,   1,  -3,   1,   0,   1, -23,  25,   9,  -5,   1,  -1,   1,   5,  -8, -15,  10,   2,   1,   0,   0,  -3,   4,   0,  -5,   0,   2,   1,   0 },
      {   7,   0,  -6,   2,   0,   0,   0,   0,  20, -37,  38,  -4,  -1,   0,  -1,   1, -56, -45, -27,  34,   5,   5,   2,   2,  13,  14, -24, -26,   5,  -2,   1,  -1,  24, -11,   7,  14,  -9,   2,  -2,   1,  -4,  27, -38,  -3,  12,  -3,   1,   0,  -4,  10,  18, -21,  -2,   1,  -1,   0,   4,  -4,   2,   7,  -3,   0,   0,   0 },
      {  -8,  11,  12,  -8,   1,  -1,   1,  -1, -34, -29,  -3,  14,  -1,   5,   0,   2, -39,  50, -37,  -8,   5,  -4,   1,  -1,  51,   7,  19, -32,  -2,  -3,   0,  -1,   7,  22,  33,   0, -13,  -1,  -2,  -1, -11, -20,  25,  20,   4,  -1,  -1,   0,   3, -17, -13,  13,   4,   4,  -1,   1,  -2,  -1,  -6,  -1,   4,   0,   1,   0 },
      {  13,  -5,  40, -12,   2,  -2,  -1,  -1,  -9, -49,  -2,  20,   3,   2,   3,   1,   2,  -6,  12,  -3,  -1,   2,  -1,   0,  -5, -20,  -9,  22,  -1,   4,   1,   2, -12,  -9,  -2,   2,   2,   0,   1,   1,  10, -39, -34,  30,   3,   3,  -1,   2,  -2,  32, -54, -27,  14,  -2,   2,  -1,  -1,  15,  20, -27,  -9,  -1,   0,  -1 },
      {  -7, -13,  14,   5,  -4,   0,   0,   1, -25,  -2,  -2,  -8,   4,   1,   1,   1,  -5,  13,  -2,  18,  -8,   0,  -1,   1, -62, -13, -32,  -4,  24,   0,   4,  -1,  62,  -3,  16, -45, -17,   1,  -3,   0,   3,  19,  21,  32, -15,  -5,  -3,  -1,  -8,  -5,  -7,   7,  14,   0,   2,   0,   2,  -5,  -2,  -3,   3,   2,   0,   1 },
      { -11, -17,  31, -19,   8,   1,   3,  -2,  -6,   5,  40,  16, -10,  -3,   0,  -2, -32, -17, -23,   8,  11,   6,  -1,   2, -23,  16,   8,   5,  -7,   1,   1,   0, -23, -31, -18,  -9,  21,   0,   5,   2,  38, -18,  25, -25, -25,   5,  -5,   1,   4,  -3, -11,  48,  -6,  -6,  -2,  -2,  -3,  11, -21, -11,  14,  -1,   1,  -1 },
      {  -5,  22,   8,   8,  -3,   0,  -2,   0, -31, -17, -46,   9,   7,   3,   3,   1, -28,  29,   0,  -6,  -6,   0,   0,  -1,  -9,  25, -33,   8,  10,  -3,   1,   0, -25, -11,  -2, -27,  13,  -2,   4,  -1,  34,  -1,  -9, -36,  -7,  -1,  -1,   0,   8,  25,  39, -22, -21,  -4,  -2,  -2,   6, -11,  22,  24,  -7,  -1,  -3,   1 },
      {   5,  21,  64, -21,  -1,  -7,   2,  -1,   1,  49, -12,  26,  -9,  -2,  -4,  -2,   7, -10, -42,  11,  18,   2,   0,   0, -32, -12,  -6, -21,  -2,   6,   3,   4,  -5,  26,  -6,  15, -12,  -4,   2,  -3, -28,  -7,  -6, -12,  25,  -4,   4,   1,  16,  -9,   9, -25,  -5,   3,  -5,   1,   2,   2,   8,   8,  -7,  -2,  -1,  -1 },
      { -23, -23, -54,  14,   9,   5,   1,   1, -24,   0,  41,  -8,  -4,  -2,  -1,   0, -16,  30,  -9,  19,   1,  -1,   1,   0, -39,  -6, -16,   6,   5,   5,   1,   0, -33,  31, -16,   7,   4,  -6,   3,  -2, -11, -24,   1, -10,  24,  -5,   4,   0,  22, -17, -11, -29,   1,   7,   1,   1,   2,  12,   8, -12, -15,   0,   0,   0 },
      {   7,  -4,  43,  -5,   4,  -5,  -1,  -2,  -5, -27,  33, -32,  -2,   2,   2,   0,   8,   9,  56,  14, -14,  -4,  -2,  -2,  -9, -17,   1,  -3,   7,   3,  -1,   1, -13,   4,  26,  16, -16,  -1,  -2,   0,   2, -23,   0,   0,  11,   1,   0,   0,  18, -21,  25,   7, -33,   8,  -3,   1,  -5,  -4, -21,  47,   4,  -6,   2,  -1 },
      {   7,  15,   5,  -9,   2,  -4,   0,   0,  17,  28,  43,  -8, -11,   1,  -2,  -1,  29,  44, -19,  35,   1,  -5,  -1,  -1,  43, -34, -26,  -2,  30,   3,   2,   0, -27, -27,   0, -27, -10,  12,   2,   3,   8,   9,  -8,   3, -20,  -9,   0,   0,  -7,  19,   3,  -4,   6,  -5,   0,  -1,   0,  -7,  12,  -3,  -3,   4,  -2,   2 },
      {   0,  20,  38,  22,  -9,  -1,  -3,  -2,  -9,  11,   5,  -7,   1,   4,   1,   0, -16,  21,  17,  25, -22,  -1,  -1,   2,   9,  50, -11,  44,  10, -11,  -2,  -3, -13,   1, -21,  13,  18,   5,   2,  -1, -30,  21, -12,   8,  17,  -6,   3,  -1, -30, -29,  -8,   0,  24,   2,   4,   2,   2, -12, -23, -22,  13,   4,   2,   0 },
      { -13, -13, -12, -98,  -5,   7,   5,   0,  -3,   2,  12,  20,  -9,  -8,  -1,   0,  -4,   5,  34,   7,   6,  -5,  -4,  -1,   4,  31,  13,  14,   3,   0,  -3,  -3,  13,   3,  -1, -10,   6,   7,   1,   2, -14,   4,   8,   4,  -7,  -3,   2,  -1,  -8,  -3,  13, -21,   9,   2,   1,   1,  14, -15,  16,   8, -24,   6,  -2,   1 },
      {  -4,  -4, -29, -27,  -5,   3,   2,   1,  -8, -15, -56,  17, -17,   8,  -1,   3,   8,  -4,   9,  60,  15, -14,  -4,  -2,  -5, -12, -10, -18,  22,   9,   0,   1, -21,  -6, -10,  15, -15,  -1,   1,  -1,   2,  -2, -11,   1,  16,  -2,  -1,  -3,  -3,   4,  -8,  22,   6,   0,  -1,  -2,  -9,   2, -36,   1,  43,  -3,   5,  -3 },
      {  -3,  -1,   6,  -1,   1,   0,   0,   2,  -2,  -9,  -4,   4,  -2,   5,   0,   1,  -9, -16,   1, -22,  -1,   1,   4,  -2, -15, -14,  10, -11,  -4,   1,  -3,   2, -61,   0,   4,   5, -39,   0,  -4,  -1,  42,  46,  14,  31,   6, -26,  -2,  -6, -15, -25,  12,  -7,  30,  15,   5,   3,  14, -17,  -4,  -9, -25,   9,   2,   4 },
      {   1,  -8,   3,  11,  -4,  -2,  -2,   1,   0, -14, -26, -11,  19,  -3,   3,   1,  -6, -31,  -7,  10,   1,  -1,   0,   2,  11,  -3,  21,   2,  38,   1,   3,  -3, -20, -46,  -2, -32,  -1,  20,   4,   3, -49, -10,  18,  -7, -20,  -7,   7,   0,  38, -38,  -4, -17,   0,   2,  -4,   5,  15,  18,  -2,  -1, -23,   4,  -6,   0 },
      {   0,   6,   8, -36,  18,  -4,   3,  -1,  -8,   2, -23, -89, -12,   9,   5,  -2,  -7,  -8, -31,  18, -37,  -8,  -4,   1,   0,   3,  22,  10,  -2, -10,  -2,   2,  -3,  18,  -4,   4,   1,  -3,  -1,   1,  15,  15,   4,  -2,   1,   3,  -4,  -4,  10,  17, -15,   1,  -6,   1,   5,  -3,  -6,  16,  16, -13,  -3,  -2,   1,  -1 },
      {  -8,  -2,  -1,  -5,  -1,   1,   1,   2, -13,  -7,   8, -10,   6,  -1,   1,  -3, -14,  -3,   0, -17,   3,  11,  -2,   2, -26, -12,   9,  -1, -13,  -2,   1,   3, -40,  21,  19,  -6,  -4,  19,  -4,   0, -61,  13, -26,   5, -42,  -1,   8,   0,   2,  36,  13,  22,  12, -32,   1,  -7,  -8, -27,   5,  -1,  30,   9,  -3,   3 },
      {   2,   7,  13,  -4,  21,  -1,   5,  -6,  -4,  -1,  -1, -30,   8,  -7,   2,  -2,   1,  19,  41, -16,  33,   3,  -3,   2,   2,  30, -26, -64, -18,  19,   8,   7,  -1, -16, -35,  -3, -22,  -5,   2,   5, -16,  11,   1,  -9,  -2,  -2,  -6,  -3,   2,   3,  -7,   6,   2,  -2,  -1,  -4, -11,  27,   8, -33,  -9,   2,  -1,  -2 },
      {   2,   3,   8,  34,  10,   4,  -1,   0,  -3,  -4,   5,   1, -35,  15,  -3,   6,  -1, -12,  15,  20,  17, -20,   7,  -1,  -7,  15,  52, -21,  34,  11,  -8,  -5,   2,  33, -21, -18, -11,  17,   3,  -2,   6, -17,   1,  -8, -15,   2,   6,   6, -47,  11,  -5,  -5, -19,  -5,   2,   3,  19, -16,  26,   0,  -5,  -9,  -4,  -3 },
      {   2,   2,  -1, -11,  33,  -5,   4,  -2,  -4, -11, -12, -14,   3,   0,   4,   1,   5,   2,  -5,   5,  22,   9,  -1,   3,  -5, -34, -16,  -7, -14,  -2,   7,   2,  -9,  -9,  34,  -2,  41, -12,  -6,  -2,  -9,   5,  -5, -21,   1,  40,   2,   5, -54, -37,  -2,  -6, -18,  -7,   8,   1,  33, -14,  -8, -21, -13,  -3,  -5,   0 },
      {   4,   7,  -3,   9,  87, -11,  -6,  -4,   0,   0,  -5,  34, -46,  12,  -4,   3,   2,  -2,  15,   3, -33,   5,   4,   2,   1,   5, -12,  -1,   9, -12,  -2,  -2,   9,  -8,  11,  23,   4,   0,  -7,  -7,  -8,  10,  13,   2, -11,  -5,   5,   0,  35,   3,   2,   6,   9, -12,   6,   0,   6,  -4,  10,  -6,  -5,  12,  -2,   4 },
      {   5,   0,  13,  15, -28,   6,   0,   5,  -2, -12,  -5, -18, -20, -23,   3,  -2,   9,   4,  11,  21,  41,  -4,  -3,  -5,   3,  -4,   9,   4,   1,   1,  -5,  -5,  18,  15,  24,  19,  20,   2,   0,  -1,  23,  -4,   2, -25,  -9,  -9,  10,   0,  43,   3,  18,   3,  29,  -4,  -2,   1,   5, -28,  -4, -46, -16,  39,   0,  10 },
      {  -3,  -3,   1,  11, -31,  -8,   1,   1,  -6,  -7,   1, -13, -45,  -3,  -3,   0,  -6,   0,  -3,  -6,  26,  -5,  -2,   4,  -1,   2, -18,  24,  -3,  18,   4,   2,  -8, -33,   9,  35,  -9, -18,  -3, -10, -27,  17,  58,   6,  14,   6,   3,  -1,  -8,  35, -15,  -2, -11,  -4,   4,   1,  25,   2,  20,  17,   2, -12,  -4,  -2 },
      {   1,   3,   2,  11, -30,  40,  -2,   5,  -1,   2,  -5,  18, -29,   1,   6,  -6,   2,   9,  10,  31, -14,   1, -14,   1,   3,   3,   3,  -4, -47,  -3,   6,   7,  -5,   8,   7, -12,  -4,  -6,   1,   5,   6,  30, -27,   9, -24,   9,  -2,  -1,  23, -32, -21,  13, -19, -26,   1,  -5,  25,  31,  23,  11, -13, -11,  -9,  -1 },
      {   3,  -1,  -6,  -3,   5,   9,   4,   2,   2,  -6,  -9, -24, -24,  12,  -6,   5,   3,  -2, -19,  -7,  -7,   1,   8,   3,   3,  -4, -18,  12, -39,   0,   4,   1,  15, -18, -30,  -4, -29,  -3,  10,  -2,  -6, -59, -13,   5,  -4, -29,   0,   6, -16, -18,  26,  12,  24, -11, -23,  -2,  41, -18,  -2,   9,   4,  17,  -5,  -4 },
      {   1,  -1,   4,  16,  15,  -7,   5,  -3,  -1,  -3,   1,   8,  32,  -1,  -6,  -1,   2,  -4,  10,  56,  -3,  11,  -4,   4,   2,  -4,  34,  -5, -43,  -1,  21,   1,  -8, -13,  -7, -26,  10, -40,  -6,   3, -10,  -4,  26,  10,  27,   7, -16,  -5,   3,  21,  14,   3,  24,   5,  -2,   0, -13, -27,  27,   9,  -5,   3,  -2,  -1 },
      {  -1,  -3,   0,   0, -12,  21,  -1,   1,  -2,  -6,   2,   2, -57,  13,   3,  -4,  -2, -11,   1, -25,  -6, -25,  -5,   0,  -2, -10,  -4,  -3,   0, -21,   0,   5,  -4, -17,  12, -33,   8, -14,   9,  11, -23,  -1, -13, -16,  22,  -2, -33,   3,  -7, -13,   3,   7,  -3,  33,  -5, -17, -57, -17,  14, -15,  -5,   4,  12,  -2 },
      {  -3,  -4,   1,  -4, -50, -13,  -7,  -5,   0,  -1,   2,   8,  -6,   9,  -7,   3,   2,  -2,  12,   2, -56,   6,  19,   8,  -7, -13,  -9, -53,   5, -47,  10,   5,  -6,  -4, -15,  17,  23,  13, -22,  -6,  -5, -16,  20,  -5,   3,  11,  12,  -6, -12,  13,  12,   4,   4,  -9,  -4,   8,  11,   5,   1, -20, -15,   6,  -4,  -4 },
      {   5,   2,   7,  -1,   8,  27,   2,  -2,   5,  -1,  10, -11,  -6, -25,  13,  -2,   1,  -4,  -3, -27,   6, -45,  -3,   3,   1,   1,   0, -21,  19, -11,   3,  -3,  -5,  -1, -21, -17,  31, -10, -15,   0,   9,   2,  -3,  24,  23,  33,  -4, -15,  26,  -5,   2, -12,  32, -13,  13,  -1,  34,  -8,   4,  22,  47,   0,  -7,   5 },
      {   0,   2,   1,   4,   6,  24,  -6,   5,  -4,  -1,  -9, -11, -18,  -1,   8,   8,  -1,   0, -10,   9,  29,  14,  12,  -3,   0,   2,  -7,   9, -11, -16,   5,  -2,  -1,  -7, -20,  11,   1,  19, -39,  -3,   6, -15,   8,  25, -20,  24,  29, -15, -13, -15,  14, -16,  12, -15,  23,  18, -65,   5,  -2,  27, -20,  -3, -11,   9 },
      {   2,  -1,   2,  -3,   2,  47,   4,   2,   4,   3,   2,   7,  43,  27,   9,  -3,  -1,  -4,  -4,   5, -17, -47,  -7,   1,  -3,  -2, -18,   0, -11,   8, -17,  -4,   4, -15,  -2,  25, -26,  17,  -2,  -7,  -1,  -6,  18, -17,  -1,   7,  12,  -6,   6,  -7, -18,   6, -26,  -2,  20,   9,  -7, -49,  23, -22,   3,   5,  -7,   7 },
      {  -1,  -2,   0,  -2,   7,  -4,   4,  -4,   0,  -4,   3,  -3,   3,  -5,   4,  -3,   0,  -3,  10,  -2,   5,  -7,  -3,  -1,   0,  -6,   2,  -4,  -9, -16,  -5,   7,   0,  -6,  22, -10,   6,  25,  26, -22,  12,  14,  11, -24,  32, -24,  40,  18, -15, -15, -17,  -5,  22, -31, -45,  27, -24,  17,  31,  13,  27,   5, -21, -18 },
      {   3,   5,   3,   4,  15,  71,  -1,  -4,  -4,  -2,  -6,  -1,  -5, -56,  -1,  -8,  -1,  -1,  -6,  -1, -27,  22, -26,   9,   0,  -7,   2, -11,  10,  23,  14,  -2,  -2,   1,  -1,   5,  13,  16,  21,  -8,  -7,  -2,  11,   3,  13, -11,   4,   7, -18,  28,  -1,   5,  -2,   6, -13,   7,  -1,  -4, -28,   8, -28,   1,  10,  -6 },
      {   0,   3,   5,   5,  10,  26,   0,  -1,  -1,   1,  -1,   4,  26,   1,  18,   5,  -1,   0,   1,  14,  10, -39,  21,  -5,  -7,  -8,  -5,   2,   3, -17, -19,  -9,  -2,   5,  17,  34,  15,  -5,   1,  11, -10,  -4,  11,  -8, -28, -22, -32,  -4, -28,   7,  19,   7,  29,  13,  -9, -25,   4,  54,  20,  17,  -8,   9,  11,  -2 },
      {   1,   3,  -3,   8,   7, -18,  15,  -8,  -1,   0,  -5,  -7,   3,  15, -25,  -8,   0,   4,  -1,   5,  25, -20, -10,  11,  -5,   2, -12,  -5, -21, -29, -13,  13,   0,   3, -11,  -1,  18,  55,  12,  -5,  -1,   9,  -6,   5,   3,  14,  -2,   5,  16,  13, -30,  21,  17,  30, -11,  -7,  14, -22, -13,  37, -39, -14,  21,  -3 },
      {   2,   3,   2,   0,  14,  26,  10,   0,   1,   1,   5,   6,   2, -14, -12,   1,   1,   3,   5,   2,  19,   6,  12,   5,   0,   3,  10,  25,  -8, -56,   1,   9,  -1,   0,  14, -28, -43,   8, -43, -12,  -6,   2,  -2, -21,  24,   1,  13, -14,   0,  24,   5,   2, -19,  25,   3,  10,  25,  26, -15, -21,  13,   7,   4,  -2 },
      {   1,  -2,  -2,   0,   0,   0,   8,   9,   0,  -2,  -6,  -5,   6, -46,  -4,  -2,   2,   1,   1,   5,   5,  46,  -3,  -8,   1,   1,  -5,  -9,  16, -42,  -5,   7,   0,  -3, -15,  24,  -9,  -1, -10,   7,   7,  -3,   7,  -6, -21, -13, -23,   7,  -3, -27, -34, -10, -14,  10,  -7, -29,  -1, -35,  38,  15,  29,   3,  16,  -5 },
      {   2,   3,   6,  -4,  -2,  45,  -5,   5,   2,   2,   5, -10,  -2,  67, -15,   9,   3,   2,   8,  -2,  11,  39,  11,  -6,   1,  -1,   4, -15,  -7,   0,  -3, -12,   0,  -2,   2,  -3,  27, -10,  10,  -8,   1,   9,  15,   1, -21, -11,  11,  14,  14,   8,  -7, -40,   8,  30, -10,   6,   9,  -3, -14,   0,  34, -10,  15,  -5 },
      {   0,   0,  -1,   0,   7, -13, -24,   4,   0,   1,  -2,   0,  -4, -26,  31,  -7,   0,   2,  -3,   7,   4, -33,  36, -10,  -1,   1,   0, -10, -27, -14, -26,   4,  -2,  -3, -12, -12,  21, -25,  -7,  -4,  -7,   9,   1,  23, -17, -39,   9,  11,   3,  14, -20, -21, -41,   3, -17,  13,  -3, -21, -37,   2,  -7,  10, -10, -16 },
      {  -1,  -1,   2,  -1,  -2,  -7, -18, -13,  -3,  -4,  -5,  -1,  -8, -37, -39,  20,   0,   1,  -1,   8, -16,  -4,  41,  -1,   0,  -4,  -3,  -2, -29,  44, -27, -16,  -2,  -7,   6, -11,  15,  39, -13, -16,   0,   1,  -1,   7,   4,  -2,   8,   4,   2, -13,  15,   3,  -4,  33,  18,  -8,  -1,   8,  27, -13,  27, -12,  16,   4 },
      {   0,  -1,   0, -10,  -1,  24, -19,  -7,   1,   0,  -2,  -8,  12,  13,   6,  14,   2,   2,  -1,  -7,  18,  25,  62,  -7,   1,   2,  -4,   4,  33,   1,  -2,  -5,  -1,   1,  -1,  -2,  11, -18,  -8,  -3,  -4,   6, -21,   3,  36, -14,  -1,  10,  14,  -1, -12,  57,   0, -20,   1,   9,   7, -14,  21,   8, -29, -18,  -1,   1 },
      {   0,   0,   2,   0,  -1,  -7, -35,  43,   0,   1,   2,   0,  -7,  -7,  42, -10,   0,   2,   3,   3, -15,  23,  -1, -26,   0,  -2,   1, -11, -11,   0, -54,   0,   3,  -5,   3, -13, -21,  -3,  17,  -3,  -7,  -7,   1, -37,   8,  21,  10, -14,  -5,   9, -12,   0,  38,   8,  24,  -1,  10,  -1,  -9,  13,  -5, -20,   5,  15 },
      {   2,   4,   2,   2,  -3,   4,  59,   5,  -1,  -1,   0,   0,  -4, -18, -41,  11,  -1,   0,   2,  -4, -23, -20,  40, -10,   1,   3,   0,  -2,   2,  14,   7,   0,   0,   2,   8,  -1, -16, -37, -10,   7,  -1,   5,  -7, -38, -25,  10,  11,   4,  -1,   0, -28, -12,  30,  -5,  -8,   6, -10,  -9, -23,  15, -12, -19,  -9,   6 },
      {   0,   2,   0,   3,  -2,   1,  57,   7,  -1,   0,  -1,   2,   0,   1,  -2,   6,  -3,  -1,  -3,   0,  -3,  11, -25, -32,  -2,  -1,  -4,  -4,  12,   0, -61,   0,  -6,  -5, -12,  -1,  12, -27,   1,  -1,  -7,   0,  -7,  34,   3,  17,  37,  12,   0,   1,  20,  25, -16,  20, -14,  11,   1,  12,  19,  -5,  -6,  25,   5,   0 },
      {   1,   1,  -1,   2,  -2,  -5,  51, -34,   1,  -2,  -2,   1,  -7,  -8,  35,  -5,   3,  -1,   3,   2,   4,   7,  -3,  11,   0,  -6,   0,  -4,  -8, -22,  -1, -15,   4,  -2,   4,  -1,  11,   4,  24,   1,  -1,  -8,   2,   5,  -1, -45,   9, -11,   3,   0,  15,  -1,  -1,   0,  40,   6,   2,  -5,   1, -14,   6, -66,  -7,  23 },
      {   2,   2,   2,   1,   3,  -3,  26,  69,   0,  -1,   1,  -1,   1,  12, -20,  -3,   0,  -2,   1,  -4,   0,  -3, -12, -39,   0,  -1,  -1,  -8,  -7,   9,  14,  -7,   2,  -3,   3,  -9,  28,  26, -29,   7,  -1,  -1,   4,  -2,  27, -49, -16,  -9,  -6,   3,  -9,   1, -13, -26,  27,  -6,   3,   7,  -5,   7,  -2,  11,   2,  15 },
      {   1,   1,   1,   1,   5,  11,  11,  23,   0,  -1,   2,   1,   2,  -5, -15,   9,   0,   0,   1,   6,   5,  -1,  29, -16,   1,   0,   2,   9,  -4, -19, -21,  15,  -1,  -4,  -2,   1, -13,  20,  39,   8,   4,   2,  21,  25,  14,  34, -30,  19,  10,   4,  29, -21, -11, -49, -20, -24,  -7,  -2, -16, -30,  -2, -31,   6,  -5 },
      {   0,   0,  -1,   3,  -5,   1,  36, -27,   2,   3,   1,   3,  -7,  15,  53,   0,   0,   2,   0,   0,  -9,  18,   3,  12,   2,   2,   3,   1,  -4,  37, -25,  10,   2,  -1,   4,  -7,   1,  25, -62,  -5,   4,   0,   6, -11,  12,   3, -26,   5,   4,   0,  -5, -13,   9, -17, -26, -20,   0,  -7,  -9,   1,  -2,   1,  16, -27 },
      {   1,   1,   0,  -1,   5,  -4,   5,  49,   0,  -1,  -1,   2,   1,  -5,  25, -29,  -1,   0,  -1,   2,  -2, -11,   7, -13,   1,   3,  -3,   7,   7,  13,  41,  41,   2,   3,   0,  13,   6,   5, -14, -19,   1,  -3,  -6,   9, -19,   6,   7,   6,   7,  -5,  21,  14,   9,  38, -26,  23,   4,  -5,  19, -19,   5, -46,   7, -26 },
      {   0,   0,   1,   0,   1,  -5, -19,  -3,   1,   3,   2,  -1,   1,  -8,   4,  82,  -1,   0,  -1,  -3,   1, -12, -43, -38,  -1,  -1,  -4,   1,  -5, -12,  10, -40,  -2,   0,   2,   2,  -5,   0, -17,  11,  -2,   1,  -2,  -5,  -4,  -3,   0,  -2,   6,   3,  -1,   5,   6,  -1, -19,   9,   3,  -9,  -2,  -9,  -8, -35,   2, -36 },
      {   0,   2,   0,  -1,  -2,  -9,  24,  47,   0,   2,   0,   2,  -2,  -5,  32,  22,   0,   3,   2,   1,  -2,   1,  14,  53,  -1,   0,  -2,  -2,  -5,  -7,  15, -63,  -4,   2,  -5,   0, -10,   7,  10, -26,   0,   6,  -3,   4,   1,  18, -14,  46,   4,   2,  -3,   4,  -4,  13,   2,  10,  -2,   4,   1,   4,   0,  19,  -4,  10 },
      {  -1,  -1,  -1,   3,   0,  -2,  18, -24,   0,   2,   0,   4,  -7,   0,  41,  14,  -1,   0,  -1,   2,  -6,   6,  35, -59,  -1,   2,  -1,   2, -11,   4,  46,   2,   0,   1,   0,   1,  -6,  31,  29,  38,  -1,   6,   6,   0,   7,  22,   7,  -2,   0,   7,  -9,  -1,  -8,  22,   3,   6,   8,   0,   2,   9,  10,  37,   2,  12 },
      {   0,   1,   0,   1,   3,   5,   3,  22,  -1,   1,  -1,   0,  -2,  -1,   0,  37,   1,   2,   0,   1,  -1,  14,   2,  39,  -2,  -1,  -2,  -1,   1,   9, -22,  34,   0,   3,   4,   6,  11,  15,   0,  56,   0,   2,   2,   2,   4,  -5, -16, -22,   2,   5,  -2,   4,  -2,  25, -17,  17,   3,  -4,   1,  -2,   8,  -3, -72,  11 },
      {   0,  -1,   0,   0,  -4,  -1,  10,  -4,   1,   2,   2,   0,   1,  -1,   8,  58,   0,   0,   0,   1,  -1,   1,  -2,  -3,   1,   1,   4,  -2,   2,   2,   0,  55,   1,   0,  -1,  -1,   5,  -5,  21, -73,  -3,  -1,  -5,  -1, -10,  -4, -35, -25,  -1,  -2,  -5,  -5,  -4,  -5,  13,  25,  -2,   2,  -2,   6,   0,  19,  14,  19 },
      {  -1,   0,   0,  -1,  -3,   4, -11, -25,  -1,   0,  -1,  -2,   2,  -1,  -8, -11,  -1,   0,  -1,   1,  -1,  -1, -18, -33,   0,   1,   0,  -1,   0,  -5,  -6,  19,   0,   1,  -2,   3,  -5,   5, -21,  -9,  -2,  -2,   1,  -5,  -3,  -1, -35,  88,   1,   4,   5,   0,  13,   9,  30,  21,   2,   6,  -6,   2,   2,  -4, -46,  13 },
      {   0,   1,   0,   1,  -1,  -1,  10, -10,  -2,  -1,  -2,  -1,  -5,  -1, -15, -41,   0,   1,  -1,   0,   3,  13,   9, -23,  -2,  -2,  -3,   1,   0,  -1, -15, -47,  -1,   2,   0,   4,  -2,  10,  11, -20,  -3,   0,  -5,   1,  -6,  12, -54, -47,   2,  -1,   0,   1,  -2,   2,  -8,  52,   0,  -2,   2,  -1,   2,   5, -29, -37 },
      {  -1,  -1,  -1,  -1,   2,  -2, -13,  -9,   0,   1,  -1,   0,  -1,   0,   2, -11,  -1,  -1,   0,   0,  -1,  -3,  -8, -15,  -1,   1,  -1,   0,   0,   2,   1, -21,  -2,   0,  -3,   1,   0,   2, -22, -11,  -2,   2,   0,   3,  -1,   4, -11,  -9,  -4,  -1,  -4,  -2,  -7,   1, -67,   8,   0,  -1,   1,  -6,   1, -18,   3,  97 },
      {   0,   0,   0,   0,   0,  -3,  -4,   1,  -1,  -1,  -1,   0,  -3,  -5,  -4,   0,   0,   0,  -1,   0,   0,  -4,  -2,  20,  -1,  -1,  -2,   0,  -3,  -1, -12,  11,  -1,  -1,   2,   1,  -2,  -2,  -4,  53,  -2,  -1,   2,  -4,  -2,  -3, -16,  17,  -1,  -3,   1,  -5,   2, -19,   7,  81,   4,   2,   8,   1,   8,   1,  70,   6 },
    },
    {
      { 106, -53,  -5,  -2,  -1,  -1,   0,   0,   6,  16, -14,   2,  -2,   1,  -1,   0, -28,  26,   0,  -4,   1,  -1,   0,   0, -10,   1,   8,  -2,   1,   0,   0,   0,  -5,   0,   1,   1,   0,   1,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -2,   0,   1,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
      {  22,  32, -24,   0,  -3,   0,  -1,   0, -90,  56,   9,  -6,   3,  -2,   2,  -1,  -2, -32,  27,  -5,   3,  -1,   1,   0,  22, -22,   0,   6,  -3,   1,  -1,   1,   6,  -3,  -6,   2,   1,   0,   0,   0,   5,  -1,  -2,  -1,   1,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0 },
      {  16, -48,  29,  -2,   3,   0,   1,   0, -43, -40,  32,   9,   1,   3,   1,   1,  40, -42, -23,  26,  -5,   4,  -2,   2,   9,  27, -45,   5,   3,  -1,   0,   0,  -7,  21,  -5, -13,   5,  -3,   1,  -1,  -1,   5,   2,  -5,   0,   1,  -1,   0,  -1,   4,   0,  -2,  -1,   0,   0,   0,   0,   2,   0,  -1,  -1,   0,   0,   0 },
      {  46,  64, -51,  -1,  -9,   0,  -3,   0,  32, -12,  19, -13,   1,  -4,   0,  -1,  40, -27,   1,   7,  -4,   1,  -1,   1, -33,  37, -11,   3,   0,  -1,   1,   0, -16,   0,  16,  -6,   4,  -2,   1,  -1,  -4,  -8,   6,   3,  -1,   1,  -1,   0,  -3,  -3,   0,   3,   0,   0,   0,   0,  -3,   0,  -1,   1,   1,   0,   0,   0 },
      { -11,  -2, -14,  14,   0,   2,   1,   0, -47,   9, -37,  19,   4,   4,   1,   1,  38,  40, -52, -16,  10,  -5,   2,  -2, -14,  38,  23, -40,   8,  -4,   2,  -2,   1, -15,  35, -10,  -8,   2,  -1,   1,   2, -10,  10,   4,  -4,   0,   0,   1,   3,  -6,   4,   3,  -2,   0,   0,   0,   1,  -1,   1,   2,  -1,   0,  -1,   0 },
      {  -4, -36,  -2,   6,   1,   1,   1,   1,  25,  26, -36, -12,   5,  -3,   2,  -1,  70, -24,  34, -42,   3,  -2,  -1,   0,  30,  -2,  13,   4, -22,   8,  -7,   2, -25,  26,   2,  17, -13,  -1,   1,  -1, -12,  -3,   4,  12,   1,  -5,   2,   0,  -5,   3,  -5,   4,   5,  -2,   0,   0,  -2,  -1,   1,   0,   2,   0,  -1,   0 },
      { -16,  -9, -40,  38,  -8,   7,  -2,   2,  14,  19, -46,  21,   0,  -2,   0,  -1, -34, -15,   4,   7,  11,  -2,   2,  -1,  12,  14, -52,  -1,  15,  -3,   2,  -1,  -3,  38,  -4, -42,  15,  -3,   2,  -1,   4,  -1,  28, -21,  -4,   4,  -2,   0,  -2,  -2,  11,  -2,  -5,   2,   0,   0,   1,  -4,   5,   0,  -3,   0,   0,   0 },
      { -15, -11,  37, -38,  12,  -1,   3,  -2,  11,  60,  10, -39,  -1,  -1,  -1,  -2, -14,   0,  10,  16, -14,   6,  -2,   3, -15,  32, -27,  -4,   1,  -4,   2,  -1, -10,  -2,  48, -20,  -1,  -3,  -1,  -1,  13, -32,   9,  21, -10,   6,  -3,   1,  -1,   3,  -8,   7,   0,   0,   1,  -1,   2,  -6,   0,   4,  -1,   0,   1,   0 },
      {  10,  30,  45, -22,  -6,  -4,  -2,  -2,   2,  32, -21,  22, -13,   0,  -3,   0,  43,  35,  -9,  15,  -8,  -5,  -1,  -2, -21, -20,  -6,  35,  14,  -5,   2,   0, -21,  12, -39, -13,  29,  -2,   1,  -1,  -7,   1,  18, -31,   6,   5,  -2,   1,   4,  -8,  10,  -2,  -8,   6,  -2,   0,  -1,  -1,   3,   2,  -3,   1,   3,  -1 },
      { -25, -38, -23,  30,  -7,   6,   1,   3, -23,  -4,   7, -13,   2,  -3,   3,   0,   2,  15,  40,  -9, -12,   0,  -5,   1, -65,  30,   5,  44, -14,  -2,  -1,  -1,  -2, -20, -16,  12,  24, -12,   4,  -3,  14, -13,  -6, -15,  16,   4,  -3,   2,   6,  -1,  -4,  -5,  -1,   6,  -2,   0,   4,   1,  -2,  -3,  -1,   1,   1,  -1 },
      {  -8, -21,  -5,  25, -10,   4,  -1,   0,  28,  58,   9,  19, -23,  -3,  -5,  -1,  13, -36, -30,  40,  13,  -4,   2,   0, -12,   1,   4, -13,  14,  -1,   1,  -2,   2, -37,  -6,  23,   0,   4,  -1,   1,  -1,  24, -42,  -7,  13,  -3,   3,  -2,  -6,  11,  10, -19,   7,  -3,   1,   1,   1,  -1,   2,  -5,   1,   2,  -1,   1 },
      {  -7,   6, -22,  -7,  -2,   1,   2,   2,  -2,  14,  27, -16, -20,   2,  -1,   1, -13,  42, -19,   8, -21,   0,  -1,  -1,  49,  42,  -5,   5, -25,  -9,   0,  -2, -18,  32,  -5,  38,   5, -22,   3,  -2, -16,  12,  -4, -10,  38, -13,  -1,  -1,  -1, -11,  19, -12,   9,   7,  -6,   1,   0,  -7,   1,   2,   0,   4,  -1,   0 },
      {   5,   3,   1,  34, -26,   0,  -3,   3,   1,  -9,  12,  29, -19,  -3,  -2,   1,  14,  33,  31,  21, -12, -16,   0,  -6,  45,  -2,  -4,  20,  31, -15,   1,  -3, -10, -16,  24, -10,  18,  10,  -4,   3,  -2, -34, -17,  34,  -3,  11,   1,   1, -13,   6, -36,  -4,  18,  -1,   6,   0,  -2,   0,  -2, -14,   5,  -1,  -1,   2 },
      {   4,  19,  32,   7,  -7,  -4,  -3,   0,  -6,  -8, -14,  39,   1,   1,   0,   0, -37, -35, -16, -15,  22,   3,   6,   2,  -8,  10,  11,   1, -22,   6,  -3,   1, -66,   4,  10,  30,   6, -14,   8,  -3,  14, -40,  -5,  -3,  20,  -7,  -1,   4,   9,   3, -18,  -3,   4,   4,  -5,   1,   2,   6,  -1, -10,   4,   0,   0,  -1 },
      {  -1,  10,   9,  13, -17,   7,  -4,   2, -13,   4, -13,   4,  -6,   5,   1,   2,  -3,  14,   1,  15,  -7,   5,  -1,   1, -45, -14, -29,   1,  -6,  18,   4,   2,   8,  34,  -9,   6, -47,  -5,   9,  -3, -23,   9,  -1,  48,   7, -38,  13,  -5,  -2,  -9, -12, -10,  43, -18,  -2,   4,   2,  -4,   8, -15,   7,   6,  -9,   3 },
      {   3,   2,  31,  39, -34,  -3,  -3,  -1,  11,  24,  65,  15,  -7, -22,   3,  -4,   0,  10,   6, -37,  38,  -6,  -3,   1,  -6,  15,   5, -17,  -7,  14,  -5,   1,  19,  14,  -8,  -8,  -6,   1,   2,  -1,  -4,   4,  31,  -5,  -8,   2,  -2,   2,  -4, -21,   3,  25,  -6,   4,  -1,   1,  -4,  -1, -11,   6,   5,   0,   1,  -2 },
      {  16,  29,  46,  30,  -7, -10,  -3,  -4,  -1,  -3, -24, -19,  42,  -7,   5,  -4,  -5,   8,  14, -27,   0,  12,  -6,   0,  17,  41, -24,   4,   2,   8,  -2,  -2,   9, -23,  -7,  -6,   0,  11,   3,   0,   2,  25, -39,  -4,   1,   1,   8,  -2,  -7,  11,  16, -36,   3,  -3,   3,   1,   8,  -8,  15,   4, -16,   2,  -3,   2 },
      {   9,  10,  19, -20, -17,   6,   0,  -1,   4,  -1, -33,  29, -24,   4,   1,   1,  -3, -27,   2,  -6, -16,  -4,   3,   1,   5,  33,  15,  19, -25, -21,   3,  -1,  73,   1,   5,   7,  15, -26,  -2,  -1,  17,  14,  11,  11,  11,   0, -12,   2, -20,   9, -13,  16,  12,   4,  -4,  -2,  -7,  -2,  -4,  -4,  11,   1,   1,  -2 },
      {  -9,  -2,   1, -23, -13,  11,  -2,   5, -10,  -2,  -2,  20, -24,   6,  -2,   4, -18,  -9,  26,   2, -17,  -3,   4,  -4,   0,  43,  30,  -1,   6,  -8,  -1,  -1, -41,  -3, -22, -12,  -5,  32,   2,   4,  -6,  33,  -5,   2, -51, -10,  18,  -7,  -4,  -6,  20,  20,   8, -34,   5,   0,   7,  -8,  -6,   5,  11,  -4,  -7,   4 },
      {   3,   2,  20,  -1, -16,   6,  -1,   2,  -7, -21,  -9,  -1, -32,   2,   3,   3,   2, -10,  52,  13,  -1, -32,   6,  -5, -16,  -5,  14, -41,  26, -13,  -2,   1,   1,  21,  28,   5, -11,  -6,  -7,   2,  -1,   1,  -4,  -8,  31,  -4,  -8,   1,  15,  -7,  34, -25, -15,  21, -12,   2,   6, -12,   2,  31, -18,   9,   0,  -2 },
      {   7,   2,   9,  32, -40,  24,  -4,   3,   4,  -6,  -3, -25, -24,  29,   1,   5,   7,   0, -17,  -6, -32,  17,   6,   4,  15, -29, -13,  10, -23,  -2,  17,  -3, -14, -36,  15, -10, -15, -19,  18,   1,  33,   0,  17, -10, -13, -26,  -3,   7,   9,   7,  22,  14, -11, -11, -14,   5,   5,   2,  -5,  23,   1,  -3,  -8,   1 },
      {   0,   2,  20,  43,  -5,  -3,  -7,  -2,  -6,   2,  -8, -24,  20,   8,  -3,   1, -14, -24, -18,  13, -31,  14,  -1,   2, -11, -12,  33,  14,   1, -25,   5,  -4,  -8,  17,  31,   4,  23,   1, -13,   1, -60,  21,  13,   7,   0,  26,   2,  -4,  -4, -25,   0,  10,  -5,   4,  15,  -1,  -6,  -4, -11,   1,   4,  -5,   4,   3 },
      {   2,  -5,   2,  10,  29, -29,   6,  -4,   9,   3,   4,   7,  35, -13,   2,   0,   1, -14,   4,  15, -16, -10,   2,   0,  14,   6,  22,   8,  20, -12,  -2,  -2,  -6, -26,  -7, -18,  -8,  -3,  10,   3,  24,   8,  33,   4,  17, -44,  -7,   4,   4, -43,   8,   5,  39,  17, -25,  11,   8,  -3, -15,  -8,  -9,  23,  -8,  -1 },
      {  -7,  -8,  -8, -24, -35,   6,   2,   4,  -5,  -2,  11, -24,   1, -14,  13,  -1,  -8,  -4, -25, -45,   9, -30,  17,   0, -12, -17,  -3,   8,  22, -32,   2,   4, -30,   2,   6, -24,  12, -12, -12,   8,   5,  37,  -4,  10,   7,  -3, -23,   1, -31,  22, -10, -13,  24,  13,  -7,  -4,  -3,   0,  14, -15,  -7,   8,  -1,  -3 },
      {  -2, -10,  -5, -28, -54,  14,   0,   7,   1,  -8, -19, -30,   0, -19,  16,  -1,  -5, -15, -23,   4,  27, -16,  -9,   1,  14,   7,   0,  30,  15,   9, -15,  -3,  20, -14,  -4,  17,   5,  21,   1,  -5, -16, -33,  11,   2,  -4,   0,  19,  -7,  32, -41,   8, -14,   4,  -9,   1,   6,  10,   3,  -3,  -1, -13,   8,  -8,   7 },
      {  -4, -10, -11, -23,   7,   4,   0,   2,   6,   4,  25,  52,   1,   2,  -2,   1,   0,  -6,   6, -33, -15,  25,   1,   3,  -4, -14,  -6,   6, -23,  -8,  16,  -4,  11, -19,  26, -21,  12,  -5,   9,   3, -48, -18,   0,  -8,  -5,  -8,   8,   9,   0, -10,  18, -37,   5, -12,  -9,  10,   5, -17,  26,   4, -28,  13, -17,   6 },
      {  -4,  -2,   0, -17,   1,   1,  -5,   3,  -9, -14, -18,   6,   5, -10,  -2,   3,  -1,   8,  21,  15,  37,  -4,  -2,  -1,   3,  -3, -35,   5,  -6,  18,   9,   0, -16, -50,  13,  19,  -4, -14,   3,   2, -38,  13,  17,   4,  11,   1, -16,   1, -49, -10,  27,  26,   0,   8,   2,  -7, -15, -15, -22,   9,  22,  -1,   6,  -4 },
      {  -2,  -1,  -8,   2, -14, -47,  18,  -2,  -4,  -3,  -4,  20,   4, -68,  14,  -1,   1,   8,  -7,  16, -12,  -1, -22,  10,  -5, -18,  -3,  22, -26,   3, -23,   3,  -4,  14,  38,   6,   0,   6, -11,  -8,  18,  17,  -6,   3, -16,  -1,   3,  -5,  16,  25,  24,   7,  -1, -12,  -8,  -6,  -5,  -6,   3,  21,   7,   5,   3,  -5 },
      {  -5,  -5,  -8, -12, -25,  25, -22,   6,   0,   3,   7,  21,  19,  13, -20,   2,  -2,   0,  -5,  18, -10,  17, -22,   5,   3,   2,  25,  21,  -9,  39, -20,  -5,  -9,   6,   7, -14, -24,  17,   3,  -9,  23,  13,  12,  18,   9,  13, -10, -10, -27,  -3,   3, -15, -10,  44,  15, -12,   3,  -6,  12,   7, -36,  -2,  26,  -7 },
      {   2,   0,   4,  -2,   2,  41, -16,   5,   0,  -4,   7,  -7,  35,  -2,  11, -11,  10,  11,   9,  35,  42,   3,  10,  -9,   3, -10,  31,   6, -13,   2,  16,   0,   4,  30,  29,  -9,  24, -12,  14,  -3,  29,   6,  -9, -19,   1,  -4,  22,  10, -20,   4,  13,  -2,  13, -31,   0,  17,   4, -20,   5, -13,  12,   9, -18,  10 },
      {   4,   0,  -2,   1,  17,  -1, -12,   5,   4,  -3,   2,  11, -12,   6,  -6,   1,   6,   1,  -5, -24,  -1,   8,   2,  -6,  -6, -18, -30,  -5,  -2, -14,   1,  -1,  -4,   1,   4,  29,  16,   2, -26,  -1,  46,  11,  -2,  25,  -2,  33,   1, -19,   1, -46,  20,   3,  13,  -7,  30,  -4,  46, -33,  -4,  -6,   1,   4,  -2,  11 },
      {   3,   6,   6,  24,  46,  -8,   0,  -6,  -3,  -3,   5, -12, -36,  18,   7,   5,  -6,  -7, -17, -11,   8, -39,   9,   4,  -2,   3,   8,  44,   8,   7, -23,   3,   7,   3,  13,  12, -22,  27,  -2,  -8,   1, -16,  28, -10,  -4,  -3,  16,  -4, -44,  21,  16, -13,  -2, -12,   2,   6,  -1, -14,  18,   4,  -9,   5, -11,   4 },
      {   2,   3,   2,   1,  13, -19,   7, -10,  -4, -10,  -9, -20, -54,  -9,   3,  -3,   0,  -2,   9,   5,  22,  12, -34,   5,   7,   2,  14,  -4,  -7,  35,   7,  -8, -10, -14,  16, -26,  16, -20,  24,  -2,  -8,  39,  19,  -4,   7,   9,   2,   3,  15,  -4, -28, -13,   5,   0,  12,   4,  37,   8,  13, -28,  -7,  -3,  -7,   6 },
      {  -2,  -4,  -1,   2, -15, -32,   8,   6,  -4,  -7,  -8, -23, -20, -20,   6,  15,  -1, -10, -10,   2, -21,  22, -10,   3,  -3,  -8,  18, -33,  -3,  19,   2,  -3,   0,   8, -33, -21,  19,  -7,   7,   0,  12, -40, -22,   2,   6,   5,   5,   3, -62, -26,   8,  -2,   1,  -8,   0,   2, -20, -24,   7,   5,   0,   1, -12,   1 },
      {   3,   2,   7,  13,   3,   2, -31,  22,  -3,  -5,  -3, -20,  -7, -15, -35,  15,   2,   1,   4,  18,  17,  12,  -3,   1,   3,   3,  10, -16, -21, -18,   6,   3,   2,  -6, -21,   6,  12,   5, -26,   2, -12, -22,  23,   8, -24,  -1, -22,  -7,  11,  33,  24,   6,  24,   7, -20, -20,  15,  -1,  37, -25, -10,  30,   8, -14 },
      {   3,   4,   3,  11,  11,  19,  20,   1,   3,   3,  -3,   5,   8, -17,  44,  -4,   0,   3,  -3,  28,   1, -33,  21,   4,   4,   4,   6,   2, -44,  -7,   4,  17,  -6,  -7, -23, -18, -24, -24, -18,  19, -12,  -8, -11,  13, -18,  14, -25,   2, -11, -20, -14,  -4, -24,   0,  -1,  -6,  42,   3,   9,  12, -20,  -8,  -9,  -2 },
      {   4,   7,   4,   9,  31,  48,  -4,  -1,  -2,  -3,  -7,  -6, -28,  -9,   4,   4,  -1,   2, -10, -14,   4,   3, -20,   6,  -2,  -6, -10,   7, -12,   6, -29,   4,   1,   8,  10, -18,  17,  21,   4,   0,  -7,   3, -46,  -3, -10,  -8,  -7,  13,   6, -21, -10,  17,  14,  18, -38,  -7, -10, -11, -34,   6,  -6,  33,  16, -19 },
      {   1,   3,   6,  -1,  -3,  13,   0,   6,  -2,  -1,  -4,   0,  -6, -13,   5,   8,   2,   7,  15,   8, -17,  -5,  12,   4,  -2,  -6, -19, -28, -36, -10, -14,   6,  -5, -13,  12,   8,  -1,  22, -22, -12,  13,  14,  18, -27,   1,  11,  26, -13, -20, -21,  -8, -14,   4,  -1,   5,  16, -35,  57,  -7, -39, -24,   9,  -8,   7 },
      {   2,   1,   5,   6,  17,  22,  -3,  -4,   1,  -2,   3,   1,  -8,  -3,  34, -19,  -2,  -4,  -1,  -2, -11, -18,   5,  -6,   5,   1,   6,  -4,  -7,  20,  -8,  -6,  -6, -15,  -9, -19,   7, -17,  -1,  -1,   3,  -2,   3,  15,   6,   5, -13, -21,  41,  -1,  25,   1,  16,   2,  34, -15, -67, -26,  29, -16,  10,  -6,   9,  19 },
      {   1,  -1,   2,  -4,   5,  11, -48,  37,   2,   2,   2,   1,  13, -29, -11,   5,  -2,  -5,  -8, -10, -40, -46,   5,  -9,   0,  -5,  -6, -20, -10,  18, -21, -14,   4, -11,  -2, -17,   8,   3,  19, -21, -15,  -4,   2, -13,  28,  -5,  13,  -5,   2,  16, -13,   9,  -2, -10,  10,  -1,  37,  -6,  -8,   4,  26, -12,   0,   2 },
      {   1,   2,   1,  -1, -21,  18,  50, -24,   2,  -1,   4,   3,  17,  25,  22, -18,   0,   1,   6,   6, -15,   5, -24,  27,  -1,  -6,  -3, -30,   1,  -5, -38,  14,  -1,  -9,  -7,   9,  10,   7, -10, -15, -12, -10,  20, -20,   7,  -4,   8, -10,  -9,  31,  -2,  12,  17,   5,   3,  -1,  30, -38,  -6, -21,   5,   9,   6,   4 },
      {  -1,  -2,  -2,  -4, -23, -46,   3,   2,   1,  -1,  -1,  -5,   0,  32, -11,   9,   3,   3,   4,   6,   5, -16,  39,  -9,  -1,  -3,  -2,   9, -29,   7,  -6,  -7,   9,   7,  21, -17, -21,  18,   9,  11,  -4,  -3, -29, -44,  -9,  -2, -19,   5,  10, -19,  -2,   7,   8,  10,   8, -24,   4, -29,  -5, -44,   7,   2,  16,   2 },
      {   2,   1,   3,   3,   9,  16,  21,   1,  -1,  -3,  -1,  -3, -17, -40,  24, -14,   2,  -2,  -2,  -5,  -6,  37,  41, -30,  -1,   0,   7,   4,   8,   1,  27, -16,   6,   6,   0,   1,  -5,  17,  13,  -2,  -3, -15,  11,  -2,   0, -11,   7, -10, -10,  20,  21, -18,  14,  18,  13, -18,  23,   9, -37, -12,  -3, -18,  43,   2 },
      {   0,   0,  -3,  -2,  -2, -11,   3, -16,   0,  -1,   1,  -5,  -5,  16, -21,   8,   1,   1,  -3,   1,  30,  -1,  -4,   7,   3,   1,   9,   9, -34, -16,   8,  -3,  -4,  -5, -13, -52, -10,   5, -34,  -9,  -4, -11,  -8,   5,  36,  10,  16, -31,  10,  17,   7,  -9,  30,  -5,  18,   6,   1,   9, -42,  38,   6,   5,  -5,  13 },
      {   3,   3,   5,   0,   8,  19,  18,  19,   3,   2,   6,   6,   7,   1,  -3,  38,   0,  -1,   3, -11,  -4,  -2, -32,  39,   2,   0,   9,  10,  25,   3,  -6, -12,   5,  11,  13,  -7, -15, -27,  -5,   9,   1, -14, -14, -18,  -9,  -6, -28, -16,  -9, -11,  32, -17,   5, -12,  16, -28,  12,  43,  -7, -21,  36,  -4,  -4,  11 },
      {   1,  -1,   1,   0,  -1,  26,  -4, -50,  -2,  -5,  -4,  -3,  -3, -43, -41, -23,  -2,  -5,  -3,  -6, -15, -11, -29,  -6,   3,   1,   0,   5,   5, -17,  37,   0,  -3,  -3,   0,   2, -46,  -1, -12,  13,   5,  -9,  -1, -38,   3,   6,   3,  -5,  -5, -13, -20,  -2,  -1,  -6,   7,   0,  -6, -26,  14,  -9,  13,   2,   4, -11 },
      {   1,   1,   1,  -1,   0,   3,  -3,  51,   1,   1,   0,   2,  -8,  -3,  20,   0,   0,  -1,   1,  -2,  -5,  43,   7, -17,   5,   8,   6,  11,  17,  20, -10,  30,  -2,  -7,   4,  -8, -26,   5, -54,  16,   2,   8,  16, -20,  23,  12, -10,   5,  10,   2, -22, -18,  -4, -14, -18,   7,  -7, -25,  -7,  -3,  30,   4, -19, -14 },
      {   1,  -1,   2,   1,   3,   6,  23,   0,   1,  -2,   2,  -4,  -2,   5,  -7,   5,   2,  -1,   5,   2,  20,  14,   5, -19,   1,  -4,   0,   2,  -8, -31, -27,   0,  -1, -10,   5, -16,   6,   9,  -7, -32,   2,   4,  -3,   5,  21, -28,  18,  -2,  -5, -33,  -6,   0,  -9, -10,  -4, -20,   5,  11,  57,  14,  31, -33,  32, -36 },
      {  -3,  -4,  -4,  -4,  -8, -15, -45, -51,   0,   0,   1,   3,  -2,  18,  19, -55,  -1,   0,   1,  -2, -14,   3,   5,  -9,   0,  -1,   3,  -3,   1,  26, -13,  19,   3,   3,   5,  -5,  13,   3, -16,  11,  -3, -10,  -4,   1,  -1,   2, -15,   0, -12,  -8,  25,  -9,  11,  -9, -13,  -9,  18,  44,   3,   1,  29,  10,  -2, -22 },
      {  -1,  -1,  -1,  -2,   4, -17, -51,  15,  -1,  -2,  -2,  -1, -11,  -4,  17, -30,   2,   3,   4,   1,  15,  25, -13,  14,   1,   0,   1,   4,  13, -45, -32,  17,   1,  -2,  -1,  -4, -22, -42,   6,  -7,  -5,  -6, -13, -15, -14, -29,  17, -25,  -7,   0,  -8,   9, -15,   3,  19,  -8,  -3,  -9, -16,  -9, -30, -10,   8,  10 },
      {   1,   1,   1,   1,   1,  14,  11, -20,  -2,  -3,  -3,  -2,  -6, -26, -46,   1,   1,   1,   0,  -3,  -3,  14,  52,  33,   3,   4,   2,  10,  15,  29, -19,  38,   3,   0,   7,  -8,  15,  -6, -23,  13,  -5,   1, -14,   9,  19, -26,   3,   4,   5,  -1,  11,  33, -14,   0,  10,  10,   9,   3,  11, -11, -17, -16, -12,  27 },
      {   1,   2,   2,   3,   0,   2,  36,  27,  -1,   0,  -1,  -2,   4,   9, -33, -41,  -1,  -2,  -2,   0,  -2,  -8, -12, -60,  -3,  -2,  -2,  -1,   3,  19, -20, -26,   1,   2,   7,   7,  12, -23, -18,  17,  -4,  -1,  -4, -15, -10, -17,  -9, -16, -12,  -1, -12,  12,  13, -20,   9, -21,  -1,  16,   0,  23, -30,  15, -10,  15 },
      {  -1,  -1,  -1,  -1,   1,   3, -20,  -9,   0,   1,  -2,   0,  -1,  -6,  30,  21,   2,   3,   4,   5,   1,  21,  16,  23,   0,   1,  -4,  -6,   9,   2,   7,  -4,   2,   2,   7,  18,   1,  24,   8,   2,  -1,   2,   2, -39,   3,  -2, -26,  -9,  -3,  -5, -36,  21,  48,   1,  21, -18,   4,   9,  33,  56,  -9,   3,  -5,  21 },
      {   1,   0,   1,   3,   2,   1,   4,  12,  -1,  -2,  -1,  -7,   0,  -3, -36, -22,   1,   0,  -1,   6,   4, -12,   4,  25,  -1,  -2,   1, -11, -15,  -6, -13,  46,   1,  -2,  -2,   8,  19,   6,  26,  14,   7,   3,  24,   1, -22,  -2, -17,  14,   0,  -9, -14, -60,  20, -16,   1, -23, -15, -13, -24,  23,   9, -32,  13,   7 },
      {   1,   1,   1,   0,   1,   1,  19,  33,  -2,  -3,  -3,   0,  -7,  -4, -13, -43,  -1,  -1,  -2,  -1,   1,  10,  -3,  -4,  -1,  -1,  -1,   6,  -3, -23,  -9,  24,  -1,   0,  -3,  -8, -27,  11,  38,  -7,  -3,  -6, -13,  -1,   3,  41,   3,  15,  -1,  -3,  21,  19,  20,  43,  -4,  19,   6,  21,  25,   4,  28,   5, -22,  32 },
      {   1,   0,   1,   1,   2, -10,  12,  41,   1,   1,   1,   2,   0,  14,   7, -27,  -1,  -3,  -2,  -6,  -5, -24, -15,  30,   2,   1,   1,   4,   4,  21,  65,  42,   3,   4,  10,   0,  11,   9,  -4, -11,   1,  -1, -13,  -8,   8,  -7,  18, -33,  -1,  -6,   4,  24,  11,  -3,  -3,  -5,   0,   6,   8,  -1, -20,  -6,  21, -35 },
      {   0,  -1,  -1,   0,  -1,   0,   9,  -8,   0,  -1,   1,   0,   2,  -2, -16,   1,   1,   1,   2,   1,  -4,   2,  49,  15,   4,   3,   5,   2,  17,  19,  -5,  -5,  -3,  -4,  -1,   3,  -9, -48,   9, -55,  -2,   6,  -1,  -4, -30,  33,  20, -33,   2,  -6, -15, -17,   4,   4, -23, -13,  -2,  -5,  -1,  14,  13,  40,   1, -15 },
      {   0,  -1,   0,   0,  -1,   2,  -1, -15,   2,   2,   3,   2,   6,   1,   3,  49,  -1,  -2,  -1,  -2,  -5,  -3,  -6, -41,   0,  -1,   0,   0,   8,   4, -21,  47,   1,  -4,   3,  -1,  -8, -32,  24,  41,  -2,  -4,   1,  -9,   2,  35,  34,  -3,   6,   8,  14,  14,  36,  -7,  -1,   0,   2,   6,   2,   2, -20, -15,  17, -25 },
      {   0,   1,   1,   0,   2,   2,  -1,   0,   0,  -1,  -1,   0,  -1,  -2,  -5, -17,   0,   1,   3,   0,   0,   6,   5,  14,   1,   1,   2,   1,  10,  -5, -11, -26,  -2,   1,  -4,  -3, -20,  -2,   3, -39,  -1,  -3,  -3,  -3,  20,  38, -38,  23,   4,   0,  16,  17,  26, -52, -26,   5,  -4,   1, -10, -14, -43, -44,   0, -19 },
      {   0,   0,   0,  -1,   1,  -1,  -6,   0,   0,   0,   0,  -1,  -1,   4,  -1,   4,   0,   0,  -1,   0,   0, -10,  -7,  -5,  -1,  -1,  -1,  -1,  -3,  17,   1, -16,   2,   2,   4,   6,   7, -20, -42, -16,  -4,  -5,  -9, -17, -45, -15,   5,   5,   1,  -3,   2,  -9,  37,  48,  -9,  51,   3,  -7,   3,   7,   8, -61,  -2,  -2 },
      {   0,   0,   1,   0,   1,   6,   2, -12,   0,   0,   0,   0,   2,  -3,  -3,   6,   1,   1,   0,   1,   4,   8,   1, -16,  -2,  -2,  -1,   0,  -5,  -6,  -3,   4,   3,   4,   6,   2,  10,  21,  29,   4,   0,   1,  -1,   3,   3,  -3, -25, -79,   0,   3,  -2,  -1, -11,  15, -28,  -8,  -2,  -9,  -6, -11,   0, -35, -65, -19 },
      {  -1,  -1,   0,   1,  -2,  -2,  -4, -15,   1,   2,   2,   1,   2,   1,   4,  25,  -2,  -2,  -2,  -2,  -3,  -3, -21, -47,   1,   2,   2,   1,   3,   3,  18,  56,   0,   1,   2,  -1,   1,   3,  -4, -64,  -3,  -1,  -6,  -2,  -5, -10, -36,  -2,  -3,  -2,  -1,  -1, -11, -11, -12,   3,   2,   3,  -5,  -6,  -3,   5,  16,  55 },
      {   0,   1,   0,   1,   0,   1,  10,   4,  -1,  -1,  -1,  -3,  -2,  -5,  -8,   3,   0,   0,  -1,   1,   0,   5,   8,  -9,   0,   0,   1,   0,   0,  -4, -11,  16,   1,   2,   0,   1,   2,   2,  18, -17,  -4,  -5,  -4,  -3, -10, -14, -44,   0,   3,   3,   3,  -5,   0,  -6,  70,  57,   0,   4,  -7,   4,   0,  29,  -6, -59 },
      {   0,   0,   0,   0,   0,   3,   0,  -6,   1,   0,   0,   1,   0,  -2,   3,   4,   0,  -1,  -1,   0,  -2,  -2,  -4, -11,   1,   0,   1,   1,   2,   0,  11,  14,  -1,   0,  -1,   1,  -1,  -1, -13, -36,   2,   0,   0,  -1,  -4,  -8,  21,  52,   5,   4,   6,   3,   9,  25,  32, -65,  -3,   0, -14,  -3, -11,  -7, -64, -27 },
    },
  },
  { // 28
    {
      {-110,  44,   6,   4,   2,   1,   1,   0, -12, -17,  11,  -1,   2,   0,   1,   0,  29, -25,   2,   0,   0,   0,   0,   0,  13,  -7,  -6,   3,  -1,   1,   0,   0,   8,   1,  -7,   0,   0,   0,   0,   0,   2,   3,  -2,  -2,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -12, -26,   5,   1,   3,   1,   1,   0,  85, -56, -15,   2,  -1,   2,   0,   0,  33,  16, -36,   2,  -4,   1,  -1,   0, -13,  35, -13,  -8,  -1,  -1,   0,   0, -10,  16,   4,  -8,   0,  -2,   1,   0,  -6,   6,   9,  -4,  -1,  -1,   1,   0,  -2,   0,   5,   1,   0,   0,   0,   0,  -1,   0,   1,   0,   1,   1,   0,   0 },
      {  31, -32,   2,   1,   1,   1,   0,   0, -44, -32,  17,   3,   5,   2,   2,   1,  36, -65,  -9,  11,   0,   4,   0,   1,  30,   6, -48,   8,  -3,   3,  -1,   1,  -5,  34, -21,  -9,   1,  -2,   1,  -1,  -3,  15,   2,  -9,   2,  -1,   0,  -1,  -1,   4,   6,  -6,   2,   1,  -1,   0,   1,   1,   3,  -3,   1,   1,   0,   0 },
      {  19,  51, -26,  -3,  -5,  -1,  -2,  -1, -28,  22,   2,  -6,   1,  -2,   0,  -1,  48,   2, -18,  -4,  -4,  -1,  -2,   0, -16,  60,  -5, -16,   1,  -5,   0,  -1, -13,   5,  46, -21,   2,  -3,   1,  -1,   6, -21,  30,   7,  -8,   1,  -1,   2,  -1,  -8,   0,  16,  -6,  -1,   0,   1,  -2,  -2,  -3,   7,   0,  -1,   1,   0 },
      {   0,  19, -32,   3,  -3,   2,   0,   1,  30,  46, -42,  -6,  -4,  -1,  -1,  -1,   0,   8,  11, -14,   1,  -3,   0,  -2,  30, -28, -26,   9,  -4,   1,  -1,   1,  -5,  45, -33, -25,   8,  -5,   1,   0,   1,  11,  39, -34,  -3,  -2,   0,   0,   4,  -5,  16,   9, -14,   2,   0,   1,  -2,  -2,  -1,   7,   2,   0,   1,   0 },
      { -43, -66,  19,   7,   6,   2,   3,   1, -35,  20, -28,   8,  -1,   2,   0,   1, -40,  17, -23,   0,   5,   0,   1,   0,  19,  24, -18, -25,   5,  -3,   1,  -2,   7,  15,  32, -24,  -4,  -3,  -1,  -1,  12, -12,  21,   7,  -7,   0,  -2,   0,   3,  -3,   3,   8,   0,  -1,  -1,   0,   4,  -4,   0,   2,   0,   0,   0,   0 },
      {  -6,  -8,  27,  -6,   1,  -2,   0,   0,  52,  36,  28, -15,  -4,  -7,  -2,  -2, -15, -50,  52,  15,  -1,   2,  -1,   1,   4,  11, -23,  13,  -2,   2,  -1,   0, -22,  16,  33,  -2,  -1,   0,  -1,  -1,   2, -28,   3,  28,  -3,   4,  -1,   2,  -8,   4, -18,  10,   4,   0,   1,   1,  -3,  -1,  -6,   1,  -1,  -1,   1,  -1 },
      {   2, -28, -18,   4,  -3,   3,   0,   2,   8,  22, -32,  -9,  -3,   0,   0,   0,  59,  -8,   7, -30,   1,  -5,   0,  -1,  58, -11,  38, -16, -13,  -1,  -3,   0,  12,  -7,  20,  27, -19,   4,  -3,   1,  -6,   8, -18,  33,  -2,  -2,   2,  -1,  -8,   8, -11,   2,  14,  -3,   2,   0,  -1,  -1,  -1,  -5,   7,   2,  -1,   1 },
      { -10, -40,  51, -14,   5,  -4,   2,  -1, -19,  25,  18,  -7,  -4,  -3,  -2,   0,  58,  35,  24, -10, -11,  -3,  -4,  -1, -28,   2,  20,  32,  -7,   2,  -2,   0, -17,   3, -25,   4,  14,  -1,   2,  -1,   4,  -3,  19, -21,   1,   1,   1,   1,   2,  -6,   1,  13, -16,   2,   1,   2,  -4,   4,  -8,   9,  -3,  -3,   2,  -1 },
      { -14, -14, -16,  20,   0,   4,   0,   2,   5,  53, -22,   8,  -1,   0,  -1,  -1,  12, -15, -14,  18,   5,   1,   1,   0, -19,  32, -13,   5,  11,  -1,   2,  -1, -34, -22,  -9,  14,  18,   2,   3,  -1,  -6,   3, -47, -26,  25,   1,   4,   0,   0,   5,  16, -50,   4,   5,  -1,   2,   8,  -6,  15, -11, -14,   3,  -2,   1 },
      {   1,  10,   3,  13,  -4,   1,  -2,   0, -18, -16, -41,  26,   2,   4,   1,   2,  14,  28,  17,   1,   0,  -3,   0,  -2, -22, -20, -22,  30,   7,   3,   3,   0, -13,  30,   1, -12,   7,  -2,   1,  -3, -11, -27,   1,  37,  12,   4,   3,   0,  -8,   3, -49, -21,  35,   2,   4,   2,   1,  -4,  11, -39,   4,   8,  -3,   1 },
      {   6,  -9,   3,  -4,   1,   1,   1,   1,  15,  24,  -8,   0,  -4,   1,  -2,   0,  30, -12, -24,  25,  -2,   1,  -1,   0, -21, -36, -31,  -3,  18,  -3,   3,   0,  46, -49,  -4, -40,   7,   2,  -2,   2,  38,   0,  -9,   1, -25,   8,  -5,   2,  -1,  22, -25,  15,  -5,  -3,   3,  -2,  -1,  10,  -6, -10,  10,  -1,   1,  -1 },
      {  17,  38,  69, -37,   5,  -9,   0,  -4,   8,  18,   9,   1,  -2,  -1,   0,  -1,   6,  30, -33,  15,  -3,   0,  -2,  -1,  55,  -1, -16, -10,   8,   0,  -1,  -1,  14,  11,   5,  -1,   0,   1,  -3,   1,  -8,  -4, -15,   1,  13,   3,   3,   1,   1,   0,   1, -30,  -1,   5,   1,   2,   2,  -4,  13,  -3, -16,   2,  -1,   1 },
      {   6,  -2,  13,  -3,   1,   0,  -2,   0,  -5, -23, -30, -23,  14,   2,   3,   0,  15, -10,  36, -25,   8,   0,   3,   0, -28, -29,  -8, -25,  11,   6,   5,   3,  10,   3,  30, -23, -16,   0,  -4,   2,  -5, -16,  -9,  -2,   1,   5,   0,   2,  17,   8,  24, -38, -31,   4,  -6,   2,   7,   3,  22,  37, -39,  -6,  -3,  -2 },
      {  -9, -12, -29, -17,   6,   2,   3,   1,  -2,  18,  30, -51,   6,  -4,   3,  -1,   2,  30,  12,   6, -19,   7,  -7,   1, -27,  15, -31,   1, -13,   4,  -1,   0,  54,  24,  -3,   3, -21,  -5,  -3,  -1, -16,  32,  -9,  21,   9,  -9,   5,  -4,  -2,  -1,  11, -18,  24,  -2,   3,  -1,   1, -13,   9,  -7,   6,   8,  -4,   3 },
      {   0, -10, -36,  15,  -3,   2,   2,   2,   1, -27,  24, -19,   3,  -5,   3,   0,  -2,  37,  30,  -8,  -9,  -3,  -5,   0,  51,  25, -18,  23,   0,  -5,  -4,  -2,  13, -20,   2,  -6,  29,  -1,  -1,   1,  30, -41, -14, -22,   6,  16,  -1,   4,   3,  12, -20, -14, -21,   9,   2,   1,   0,   8,  -1,  -1, -13,  -1,   1,  -1 },
      {  -2,   4,   1, -59,  28,  -6,   7,  -3,  -5,  -7, -25, -47,  23,   4,   2,   0, -18, -17, -19, -43,  -8,  11,   1,   2,  -6,   8,  -3,   3, -26,   6,  -5,   2, -37, -15, -18,   0,  -4,   1,   4,   1,  32, -11,  -6,  -5, -12,  -4,  -1,   0,   6,   9, -18,  -2,   9,  -1,   2,  -2,  11,  -2,   5, -26,   7,   6,  -1,   1 },
      {  -7, -23, -37, -16,  22,  -1,   5,  -2,   6,   7,  48, -11,  10,  -2,   1,  -2,  13,   3, -29,  12,  24,   0,   5,  -2,   1, -49,  16, -25,  19,   5,   2,   1, -23,   9,   7, -11,   4,   8,   0,   0, -18, -31,  18,  -4,  11,  14,   0,   2,   9, -40, -10, -14,   4,  14,  -3,   6,   0,  -1,   0, -20, -11,   8,  -2,   0 },
      {   3,  -3,   5, -23,  10,  -7,   3,   1,  10,   2, -13,  25,   9,  -8,   0,   2,  -6, -37,  12, -25,  16,  -1,   3,   2,  -8,  34,  32,  10,   9,  -7,  -2,  -5,  66,  -3, -14,  -5,  28,   2,  -5,  -3,  -1,  -4,  31,  -8,  16,  15,  -4,   1, -11, -26,  -3, -15,  14,  21,   1,   3,   0,  -3,   7, -20, -22,  10,   1,   5 },
      {  -1,  -2, -12, -24,  12,  -6,   4,  -3,  -4,  -4, -34,  -1, -18,   7,  -2,   0,   1, -14, -17,  40, -25,   2,  -3,  -1, -12, -13,   0,  14, -24,  -5,  -1,   1,  27,  19,  17,  50,   6, -20,   3,  -4,   7, -40,  -7,  -7,  33,   2,   2,   2,  32, -11,  -2,   8, -16,   8,  -7,   2, -10,  11, -23,  21,  12,  -3,   3,  -2 },
      {  10,   8,  28,  59, -12,   3,  -2,   4,  10,  11,  14,  -4,  26,  -5,   9,  -3,   6,   0, -14, -38,  26,   5,   7,  -6,  -9,  -2, -25, -17, -19,  18,  -5,   3,  18,   1, -16,  25, -17,  -7,   3,  -2,  34, -32,  -1,  -1,  17, -11,   2,   0,  22, -28,   8,   2,  10,   6,  -8,   3,   6,  -7, -11,  -8,  20,   7,  -3,   0 },
      {  -1,   4,   7,   4,  24,  -7,   0,  -1,  -5,   4,   0,  -8,  14,  -4,   0,   2,  -2,   2,   4, -12,  33,  -3,   3,   0, -14,   8,   0, -32,   5,   5,  -3,  -2,  15,  31,  -1,  12,  17,  11,  -6,   0, -41, -14, -28, -29,  -7,  30,  -3,   4, -16,  44, -13,  17, -27,  -3,   6,  -9, -31,   3,  -3, -17,  41,  -3,   8,   0 },
      {  -6,  -6, -16,  -3, -17,   3,  -1,  -2,   6,  21,  30,  36,  -6,  -3,  -3,  -2,  -7,  -5, -34, -31,  -4,   8,  -3,   2, -15,   1,   8,   5, -31,   6,  -5,   3,   9,  32, -13,   2, -17,  -5,   5,   1,   4,  -6,  -5,   9,  -9, -16,   5,   4,   2,  33, -41, -15, -23, -10,   2,   3,  -5,  32,  15,  14, -42, -18,   1,  -6 },
      {  -1,  -1,   6,  20,  -7,   3,  -5,   1,  -2,   6, -16, -32,  24,   4,   2,   4,  -9,  -8, -32, -12, -13,  12,  -1,   4,   7,   2, -20,  40,   2,   5,   2,  -3,   8, -35, -10,   4,  -1,  13,   1,   0, -71, -19,  10,  15, -12,   6,   5,   0, -22, -27, -11,   7, -16,   1,   6,   0,  -4,  10, -10,  21, -12, -17,   3,  -1 },
      {   1,  -2,  16,  22, -47,   8,  -8,   1,  -1,  -2,  12, -39, -28,  -1,  -5,   1,  -9, -16, -29, -27, -24,  -8,  -2,   4,  -8, -25,  14,   9, -12, -31,   5,   1,  -2,   7,  33,  -4,  35, -22,  -2,   0,  -6,  10,  15, -16,  15,  18,  -3,  -1,  -1,  26,   4,  -9,   7,  15,   1,  -1,  -2, -11,  11, -23,   5,  15,   1,   2 },
      {   4,   6,   5,  39,  47, -20,   2, -10,   1,   2,  -3, -17,  59,  -8,  -3,  -9,  -2,  -2, -13,  24,   0,   7, -15,   4,   6,  -8,  29,  29,   4,   6,   5,   4,   2,  14,  27,  -4,   4,  -4,  20,  -5,  24,  28,  18,  -3,  15,  -7,  19,   5, -12,  26,  -3,  -5,   1,  -3,  -5,  18,  -5,  13,   0,   2,  -9,  -1,  -6,   0 },
      {   3,   7,  11,  29,   2,   3,  -4,   4,  -1,   2, -10, -33,  -3,  16,  -2,  -3,  -6,  -7,  -1,  14, -23,  -9,  -2,  -8,  -3,  18,  27, -16,  21, -12, -16,   1,  -3,  33, -25, -15,  -2,  15, -26,  10,  14,  24, -25,  13,  -5,  22, -17,  -5,  34, -31, -38,   0, -15,  12,  13,  -9,  11,  18, -18,  -4, -14,  -5,   9,   0 },
      {   1,  -4,  -2,  -1,  -9,  -1,   2,  -2,   5,   5,   6,   2,  13,  -8,  -2,  -9,   1,  -7,  -9,   2,  -5,   0,   1,  -1,   7,   7,  29,  30,  17,  -3,   0,  -5,  12,  13,  -8, -18,  13,  22,   8,  -3, -14, -31,  -9,   6, -29,  -3,   7,  -3,  65,  16,  20,  -1,  19, -31,  -4,  -5,  24, -14,  37,  12,  35,  -9,  -8,   6 },
      {  -2,  -1,  -9,  -2,  12, -15,  -2,   6,   0,   4,   4,   5,  19, -14,   7,   9,   2,   7, -21,  -5,  10, -14,  12,   0,  -4,  -9,  -6,  26,   7, -16, -14,  -2, -10,   9,  -6,  15,   7,  -7, -49,   7,  25,  -7,  -4,  34,   2,  21, -52, -14, -21,  10,  33,   8,  11,  13,   3, -30, -17,   1,  26,  27,  -4,  -6,  11,  -3 },
      {  -2,   0,   1,  -8, -34,   9,  -5,   5,  -4,  -5, -10, -48, -26,   3,   0,  12,  -2,  -3,  -8,  21,  39, -19,  14,   0,   0,  13,  15,   3,  40,  19,  10,  -8,   3,   8, -26,  15, -18,  17,  14, -12,  22, -18,  23,   8,  20, -13,   7,   0, -23,  21,   1,  -3,   5,  -4,  -8,  10,  -8,  30,  13,   2,   1, -15,  -5,  -2 },
      {   4,   2,   8,  28,  33,   1,   5,   6,   1,   0,  -1, -13, -14,  12,   0,   7,   2,  -5,   9,  20, -28,   8,   6,   0,   7,   6,  18, -22, -11,  -9,   0,   0,   4, -13, -26, -16, -12, -32,  -7,  12, -20, -34,  14, -10,  -2, -49, -36,   9,  -4,  10,   0,  -1,   9, -11, -20, -16, -26,   1,  30, -19, -12,   8,   1,  -8 },
      {  -3,  -5,  -2, -17, -27, -29,   6,  -6,   0,   1,  -3,  -1,  25, -39,   2,  -9,  -1,   3,   7,  -8,  17, -23,  -6,   3,   5,   7, -15,  -6,  22, -15,  -5,  13, -10, -21,  -2,  -4,   2, -23,  -3,  11, -26,  21,   7,   3,  24, -29, -15,   6,  50,  15, -21,   4,  13,  11, -21,  -2, -21,  28, -24,   5,   2,  10,   4,  -8 },
      {   2,   3,  -1,  18,  52, -29,  -3,  -7,   1,   2,  11,   7, -44,  18, -16,   3,  -1,  -2,   1, -32, -18,  -7,  -3,   9,   0,   1, -25,  -6,  22,  -4,   7,  -3,  -4, -20,  -1,  22, -15,  27,  -8, -13,  -6,   5,  37,   3,  19,  14,   8, -18,  28,  26,  -4, -10,  13,   4,  16,  -1,  10,  16,   2,   7,   5,  -8,   2,  11 },
      {   0,   0,   1,  10, -16, -17,   1,  -4,  -4,  -6, -13,  -9, -10, -30, -10,  -4,   1,   2,  -1,  29,  22,   4, -24,   4,   0,  -2,   0,   9, -26,  30, -29,   0,   2,  -5,  10,  18, -18,  30, -17,  -5,  -4,  -6,  17, -18, -29,   1,  -2, -17,  21,  12,   8,   7,   2,  -1,  25, -18,  -9, -14,  -1, -31, -45,  29,  15,  11 },
      {  -4,  -5,  -2,  -9, -39,  -4,  -6,   2,   2,   4,   2,  18,  39,  17,  -5,   7,  -1,  -3,   8,   3, -53,  34,  -7,   5,   3,  -1,   9, -19,  13,  12,  -5,  -5,  -7,  -3,  -8,  -8, -18,  23,  -8,  -7,  11, -13,  11,  -2,  11,   8,  16, -19,  -1,   7,  16, -16,   4,  23,  13,  -2, -55,  14,  -1,  -5,  22,   6,   1,   7 },
      {   2,   3,   3,   1,  24,  29,  -3,   4,  -1,  -1,   1,  -5, -20,  16,  11,   2,  -3,   1,  -5, -19,  16,  -6,  15, -13,   0,   0,  -9,  22,   4, -14,   3,  -4,   0,   4,  12,  -3,  15,  14,  10,   0,   9,  10, -20,   5, -23,  -6,  21,   2,  25, -16,  13,   1,  22, -14,  -7,  13, -83,  10,  -9,   1, -24,  14,  -1,   3 },
      {  -1,  -2,  -4,   8,  -5, -60,  21, -12,  -2,  -4,  -2,  -3, -18, -34,   2,   1,  -2,  -6,  -9, -19, -34, -12,  -2,   1,  -8,  -8,  -1,  -7,  31, -18, -13,  -1,   3,  19,  -7,   2,   3,  29,   5,   1,  11, -19, -24,  -4, -11, -28,  17,  13, -36, -13,   2,   5,   0, -10, -14,  11,  -5, -28, -15,  10,  -9,  15,  -6,  -3 },
      {  -1,   0,   0,   8,   0, -39,  24,  -9,  -2,  -1,  -2,  -1, -38, -14,  28,  -9,  -4,  -4,  -2,  -1,   3,  -5,  -5, -11,   3,   6,   9,  17, -24,  46, -11,   5,  -3, -11,  10, -30, -17,  -8,  16,   2,   1,  16,  -5,   1,   4,  28,  -9,   8,   2, -26,  -6, -17, -10,  20, -10,  -1, -28,   7,  32,   1,  35, -21,  -6,  -8 },
      {  -1,   1,  -3,  -2,  -2, -11, -19,   8,  -1,   2,   3,   0, -12,  33, -35,  12,   2,   5,  -4, -11,  13,  44, -10,  10,   2,   2,   3,  19,  10,  25,   2,   7,   4,  10,  10,  -8,   1,  25,  22,   6,  -1,   5, -15,  -4,  -1,  -4, -39,  23,   4,  13,   2,   8,  16,  37, -46, -14,   9, -12, -22,  11,  -5,  14,   7, -20 },
      {   0,   2,   5,   6,  -8, -20,  12,   5,  -1,   1,  -4,  -1,   8, -19,  -4,  19,   3,   8,  11,  14,  -9,   9,  31,   4,  -1,  -1,  -8, -14, -34, -29,  25,  -3,   0,   5,   4,  16,   6,  24,   5,  -7,  -1,   3,   5, -10, -38,  28,  -3,   5,  -3, -12,  12, -11,  25,  -2, -35,   0,  17,  63,   2, -16,  -4,  -7,  -8, -13 },
      {   0,   2,  -1,   7,   2,  -8,   9, -10,  -3,  -1,  -6,  -9, -16, -25,   5,  -5,   1,   3,   1,   5,   5,  49, -19,  12,   2,   5,   2, -27,   8,   7,  30,   3,  -8,   0, -37,   7,  28, -40,   9,   3,   0,  -6,  -4,  39, -13,  31,  23,  -2,  12,  14,  10,  24,  10,  23,   1,  19,  -6, -12,   6,  10, -25,  -3,  -5,   9 },
      {  -1,  -2,   0,  -3, -13, -17,  -4, -17,   0,  -4,  -3,  -8,  -5,   5, -11,  -7,   3,   2,   6,  -9,  12,  56, -15,  -8,   0,  -9,  -9, -13,   8,   6, -11,  -8,   1,   0,  18,  10,  43,   4, -12,  -5,   8,  15,  -1,  -9,  22, -20, -20,   1,  -3, -28, -10,   5,  11, -58,  17, -10,  -7,  18,  18,  -2,   0, -36,   4,   5 },
      {   1,   1,   1,   2,  11, -13, -18,   7,  -1,  -3,   1,  -5, -17, -18,  -9,  16,   0,  -1,  -4,  -3,  14,   4,   7,   3,   2,  -1,  18,   4, -27,  11,   2, -19,   0,   5,  -3, -53,   3, -16, -12,  -4,  -8, -19, -27,  -8,  16, -14,  11, -26, -13,   4,  15,  -9,  35,   0,  33,  -5,  12,  32, -46,  24,   8,   8,   0,  17 },
      {  -1,  -2,  -3,   0, -18, -38,  31,  -5,   3,   3,   3,   0,   9,  36,  41,   4,   2,   1,   0,   4,   8,  15,  53,  -8,   2,   3,   4,  13,  -4,   0,  51, -19,   1,   2,   4,  -2, -14,   1,  -5,   5,  -2,   5,   2,  -9,  10,   3, -22,   1,  14,  16,  -8,   6,  -5, -13,  23,   5,  -1, -28, -21,  -4, -18,   6,  -1,  17 },
      {  -2,  -4,  -3,  -1,  -7, -40, -28, -30,   1,   0,   3,   0,   8,  26, -42,   3,   1,   1,   4,  -2,  12, -19,   4,  -1,   1,   2,  -4,   8,   9, -21,  11,  -5,   3,   4,   0,   6, -11, -36,  47,  -3,   4,  -3, -15,  10, -25, -10,   4,  -8,   1, -22,  15,   3, -21,  16,  29, -24,  -4,  13,  17, -24,   0,  -1,  16,  15 },
      {   2,   3,   1,  -1,   8,  28,  44,   0,   0,   1,   1,   6, -11, -11,  32,  -9,   1,   1,  -2, -18, -14,  25,  -1,   1,   0,  -3,   1,  21,  49,  13, -12,   5,   3,  12,  22,  12, -13, -21,   7,  15,  -4,  -4,  -9, -10, -16, -15,  -7,  -4,  -4,   4,  28,   5,  -4,  17,  12, -15,  23,  41, -16, -34,   0,  -4,  11,  12 },
      {  -1,  -1,  -2,  -2,  -4, -30,  -9,  23,   1,   1,   2,   1,  -7,  25,  -8,  28,  -1,  -1,  -1,  -2,  22,  20, -16,  10,   2,   5,   4,  23,  14,  -8,  -4,  17,   0,  -1,   5,   2, -20, -32, -60,  21,   0,   0,   0,  -2, -15,   5,  31,   2,   3,  -7,   6,  -8, -22, -21, -13,  29,   3,   9,   1, -18,  14,  20, -27, -18 },
      {   1,   1,   2,   0,  12,   8, -13,  55,  -1,  -2,  -2,   0,  -7, -56,  -4,  29,   0,  -1,  -2,  -2,   4,  33,   6,  -6,   1,  -2,   1,   9,  12, -18,  35, -26,  -3,   2,   9,   7, -14,   3,   9, -17,   7,  14,   2,  -6,  -1, -21,   1,  -9,  10, -14, -27,  -9, -29,  15,  12, -22,  -8, -18,  16,   5,  17,   5,   6,  -3 },
      {   1,   0,   1,  -1,   2,  21,  -6, -29,   0,  -1,  -2,   4,  -3, -33, -29,  -2,  -1,  -4,  -5, -10, -12, -13, -23,  24,   2,   4,   5,  10,   7,  11,  58,  16,   2,   4,  -4,  -6, -27,  -1,  -4,  17,   8,  -7,  -2,  -2,   7,  26, -31,   0,  -1,  -9,  12, -19,  -6, -43,  -6,  -5, -18, -11, -31, -20,   0, -15,  -2,  -5 },
      {   1,   1,   3,   4,  -2,  17,  33, -53,  -1,  -2,  -1,  -5, -10,   6,  -6, -32,   1,   1,   2,   8,  36,  29, -15,   0,   0,   3,   5,   7, -14, -52,   9,   2,  -1,   0,  -6,  -2, -14,  18, -19,   6,   4,  -5,  11,  -1, -13, -19,   4,  -6, -11,  -1, -12, -32,  -4,  28,   6, -11,  -4,  -2,  -4,  24,  27,  11,  -3,  -4 },
      {   2,   2,   2,   2,  -7,  -7,  54,  37,   1,   1,   2,  -3,   8,  21,  -7,  32,   2,   0,   1,   3,  16, -11, -32,  27,  -1,  -2,   3,  -4,  -2, -24, -19,   6,   0,   1,  -2,  -3,  -8, -16,  21, -12,   3,  -4,   3, -10,  -8,  10,   6, -14,   3,   1,  -5, -16,  26,  -6,   3, -14, -10, -26, -21,   5,  -7, -55,  18,  -9 },
      {  -1,  -1,  -1,  -3,   5,  -2, -42, -12,  -1,  -3,  -2,  -1,  -4, -20,  -3, -21,   2,   2,   2,   4,  -1,  18,  41, -18,   0,   1,   2,   8,   2, -17, -20,   4,   3,  -1,   5,  -3, -22,  -2, -16,  19,   0,  -4,   1, -16,  -1,  -3,  17,   1,   5,  11,   6,   3,  14,  25,  -5,  18,  -3, -26, -13, -25,  -8, -74,   7, -18 },
      {  -1,  -1,  -1,  -1,   1, -11,  -9,  12,   0,   0,   2,  -1,  -3,   4,   6, -26,  -1,  -2,  -2,   0,  -1,   7,  -4, -57,   0,   1,   5,  -4,   5,   5, -14, -24,  -2,  -3, -12,   5,  13, -19,   8, -31,  -1,  -1,  23,  30, -20,  11,  -4,   2,  -3,  16,  16, -48, -18, -25, -15, -16,   1,  -5, -45,  -7,   6,   9,  16, -27 },
      {   1,   2,   0,   4,   3,   6,  39,   0,   0,  -1,   0,   0,  -5,  -7, -37,  16,   1,   2,   0,   0,   5,   4,  35,  22,   1,  -1,   0,   3, -22,   1, -13,   6,  -1,  -5,  -5, -22,  25,  25,  -7,  -8,  -1,   6,  -1,  49,  29, -22,   8,  11,  13,   6,  30,   2, -47,   3,   2,  10,  -1,  -1,  -8, -33,   5,  -8,  14,  -6 },
      {  -1,  -1,   0,  -2,   2, -15, -36,  34,   0,   0,  -1,   0,  -8,   1,  42,  -7,   0,   1,   0,   7,   9,   9, -19,  13,  -1,  -2,   1,  -4, -14, -29,  -6,  49,   0,   1,  -8,   0,   8,  27,  32,  31,   1,  -6,  11,  19,  -1,  -8, -26,  16,  -2,   5,  16, -21, -13,  -4,  24,  13,  -2,   6, -14,  -7,  10,  -6, -10,  34 },
      {   0,   1,   1,   1,  -2,   3,  19,  13,  -1,  -2,  -1,  -2,   4,  -4, -45,   4,  -1,   0,  -1,   2,  -5, -15,   8, -14,  -2,  -2,  -4,  -6,   2,  29,  -3, -28,  -1,   2,   4,  21,  19, -10, -11,  39,   2,   5,  14,  -2, -36, -13, -20,  34,   1,   2,  -8, -30,   9,  19,  20,  30,  -8,  -5,  -9,  14,  12, -15, -27,  40 },
      {   0,   1,   0,  -2,  -2,   1,  22,  26,  -2,  -1,  -1,   3,   1,  -5, -32, -24,  -1,   0,   0,  -4,   5,  11, -25, -48,  -1,   0,  -4,   8,  -4, -37, -10, -25,  -3,  -4,  -5, -14, -39,   6,  16,  -1,  -7,  -8, -13,  -4,  31,  27, -22,   8,  -3,   4,  13,  31,   2,  -8,   2,  40,   2,  12,  13,  -9,  -9,   8,  -7,  14 },
      {   0,   0,   1,   3,   1,  11,   4, -17,  -1,   0,  -1,  -1,  -2, -17, -19,  24,   1,   1,   0,   4,   0,  28,  53,   5,   1,   3,   3,   2,  13,  -9, -51,  28,  -1,  -4,  -5,   2, -19, -20,  30,  15,   1,   4,  12,  -5,   7,  37,   2,   5,  -2,   2, -17, -13,   4, -33,   0, -13,  -2,  -8,  -2,  24,   3,  31,   1,  21 },
      {   0,   0,   0,  -1,  -3, -10,  18,  51,  -2,  -2,  -1,  -4,  -1,   6, -35, -69,   0,   0,  -2,   0,  -2,  -3,  26,  -7,   1,   2,   1,   0,   2,  14,  27,  56,   2,   1,   1,   1,   5,  -2, -11,   8,   2,  -4,  -1,  -7,   1,   0,  24,  -9, -12,  -6,  -6,   3,  10,  -4,   4, -26,  -4,   4,   8,  11,  -2,   5,  16,   2 },
      {   0,   0,   0,   0,   0,   1,  10, -10,   1,   0,   1,   0,   3,   0,   1,  47,  -1,  -3,  -1,   1,  -2,  -5, -18, -63,  -1,   0,  -1,  -4,  -6,   4,  19,  16,   1,  -2,  -1,  -1,   8,  16,   6,  64,   1,  -2,   8,   4,  16,   0,  30,  16,   2,   6,   6,   6,  10,   0,  21, -27,   7,   2,   7,   4,  -2,   2,  23, -30 },
      {   0,   1,   0,   0,   1,   1,  10,  -5,   0,   1,   0,  -1,   1,   0, -11,  13,   0,   0,   0,  -1,   0,  -1,   5, -33,   0,   1,  -1,  -1,   1,   5,  -3,  28,   4,   2,   4,   4,  16,   5,  17,   5,  -4,   1,  -2,   3,  -3, -11, -20, -85,  -2,   4,  -5,   5, -14,   8, -10,   0,  -1,  -8,  -7,  -6,  -7,  -7, -71,  -8 },
      {   1,   1,   1,   0,   1,  -4,  -1,  17,  -1,  -1,  -2,   0,  -1,   3,   3, -35,   1,   1,   1,   0,   3,   0,   6,  44,  -2,  -2,  -3,  -3,  -2,  -3, -14, -59,   0,   0,  -2,   2,   1,   5,  23,  57,   3,   1,   5,   6,  13,  11,  12,  -5,   1,  -1,   5,   4,  -3,  -9,  12, -30,  -3,   3,  -2,  -3,  -5,   7, -39, -45 },
      {  -1,  -1,  -1,  -1,  -1,  -4,  -4,   2,   0,   1,   1,   1,  -1,   4,   5,  -2,   0,   0,   0,   0,   5,  -2, -10,  -4,  -1,  -1,  -1,   1,  -8,  -2,   1, -30,  -2,  -1,   0,  -5,   1,   5,  -8,  37,   2,   3,   3,   7,  10,   1,  25, -30,  -5,   0,  -5,  -1, -15,  -7, -71, -25,   3,  -3,  -2,  -8,  -1, -10,  23,  73 },
      {   0,   0,   0,   1,   0,  -2,   0,   3,   0,  -1,   0,  -1,  -1,  -1,   0,  -4,   0,   0,   1,   1,  -1,   1,   7,  13,   0,   0,  -1,  -3,   1,   7,   0,  -9,   2,   1,   2,   6,   8,   0,  15,  33,   0,   0,   2,  -2, -12,  -5, -20, -56,  -3,  -4,  -7, -12,   0,  -7,   0,  65,   3,   1,  11,   9,  13,  16,  70, -28 },
    },
    {
      { 100, -57,   2,  -5,   0,  -1,  -1,   0,  -3,  20, -15,   3,  -2,   1,   0,   0, -41,  23,   8,  -4,   2,  -1,   1,   0,  -5,  -6,   9,   0,   0,   1,   0,   0,   0,  -2,  -2,   5,  -1,   1,   1,   0,   0,   0,  -3,   0,   0,   0,   0,   0,  -4,   3,   0,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0 },
      {   9,  32, -31,  -2,  -1,   2,  -4,   1,  88, -31,   3, -16,   2,  -1,   2,  -1, -28,  11,  14,  -4,   0,   0,   0,   1, -53,  16,  14,   3,   1,   1,   1,  -1,   7, -18,  -1,  12,  -4,   3,   2,   0,   3,  -3, -14,   2,  -1,   1,   0,   0,  -8,   7,   2,  -5,  -3,  -1,  -1,   0,  -3,   3,   2,   1,  -1,   0,   0,   0 },
      {   4, -46,  38,  -1,   2,  -3,   3,  -1,  36, -48,   2,  20,  -6,   1,  -3,   2,  39,  25, -52,  12,  -2,   0,  -1,   0,  -8,  41, -13, -12,   4,  -4,   0,  -2, -21,  17,  15, -12,   5,  -1,  -3,   0,  -7,  -1,  14,  -1,   1,   1,   1,   0,   3,  -8,   0,   3,   2,   1,   1,   0,  -2,  -2,  -2,  -2,   2,   0,   0,   0 },
      { -59, -33,  41,  -9,   8,   1,   2,   0,  25,  33, -42,   4,  -7,   3,  -2,   0, -20,  52,  -4, -18,   2,  -3,   0,  -1,   2, -28,  34,  -6,  -1,   0,   1,   0,  16, -18,  -2,  13,  -7,   1,   1,   1,   5,  -2,  -8,   3,   1,  -1,   0,   0,  -2,   6,  -1,  -1,  -3,  -1,   0,   0,   1,   0,   2,   3,  -1,  -1,   0,  -1 },
      {  25,  48, -27,   4, -12,   0,   0,   0,   5,  43, -29,  11,  -9,  -1,  -5,   2,  49,  40,  -8, -16,  -2,  -4,   2,  -4,   1,   5,  32,  -3,  -3,  -2,  -2,  -1, -37,  -7,  26,  -2,   7,  -1,  -5,   1, -12, -14,  24,   9,   2,   0,   3,   0,  10, -23,  -5,  12,   7,   1,   3,   2,   0,  -4,  -5,  -3,   4,  -1,   0,   2 },
      {  13,  11,   4, -29,  10,  -2,  -4,   0, -47, -10,   4,   5,   0,   8,   3,  -1,  26,  -9, -37,  -4,   5,  -2,  -2,   3, -35,  21,  38, -35,   8,   0,   0,   1,  23, -39,  10,  29, -25,  11,   6,   0,   8,   5, -35,   9,  -5,  -2,   1,   0, -16,  20,   3,  -8,  -7,  -4,  -1,  -2,   0,   4,   5,   7,  -5,  -2,   1,  -2 },
      {  -1,  20, -25,  27, -13,   3,  -1,   1, -24, -43,  35,   5,   8,  -4,  -1,   2, -42,  62, -41,   9,  -1,   0,   2,  -2,  32, -28,   8, -10,   7,  -3,   0,   0,   6, -23,   4,  -6,   7,   2,  -3,   1,  -9,  11,  -3,   0,   4,   0,   0,   1,  13,  -7,   0,  -3,   6,   2,   1,   0,   5,   0,   0,  -6,   1,   1,  -1,   0 },
      { -21, -24, -47,  46,  -3,   7,  -2,   4, -15,  -2, -43,   3,  17,   2,   4,  -2, -25,  -9,  -1, -24,  14,   4,   0,   1,  -3,  39, -11, -40,   3,   0,   2,   0,  13,  10,  38, -22, -12,  -1,   2,  -2,  16, -27,  10,   7,  -7,   1,  -3,   2,   1,   7,  -1,  -1,  -7,   0,  -2,  -1,   1,   1,   3,   4,  -4,  -1,  -1,  -1 },
      {   2, -28, -32,  36,  -9,   9,  -5,   3,   4, -30, -13,   2,   8,   0,   3,   0,  47,  -2,   6, -49,  15,  -4,   1,   2,   8, -21,  20,   5, -18,  14,  -6,   3,   4,  11, -46,  26,  -9,   1,   3,  -2, -19,  36,  -6, -17,   7,  -5,   1,  -2,  -6,   1,  18,  -3,  -6,   2,   0,  -1,   0,   2,   1,   5,  -2,  -2,   0,  -1 },
      { -11, -41, -13,  13,   7,   4,   3,  -1,  -3,   6,  40, -53,  14,  -7,   4,  -3,  25,  16,  26,  -6, -27,   9,  -5,   2,  14,  17,   2,  18, -19,  -2,   0,   0, -19, -25,  32,  24,  -1,  -7,   1,   2,  -3, -15, -26,  36,  11,   0,   0,   0,   4,   6, -23,  -6,  15,   2,   1,  -1,  -1,   3,   1,  -8,   2,   3,   0,   0 },
      {   9,  18,   4, -48,  22,  -3,   2,  -4,  13, -24,  -5, -28,   3,  10,   1,   2, -21,  -6,  -2, -28,  -9,   8,   0,   2,  62,  12,  14, -19, -27,   1,   0,   0,   5,  26,  21,  16, -28,  -8,   3,   0, -12,  10,  24,  11,  -7,  -8,  -2,  -1, -15,  -3,   9,  24, -11,  -2,  -3,  -1,  -3,  -3,  -7,  16,   3,  -3,   3,  -2 },
      {  -1,  -8, -26,   7,   5,   3,  -1,  -1,  32,  45,  -9, -23,  -3,   3,  -3,  -2,   1, -23, -39,  42,  -8,   2,   1,   2,  36,   5,  -9, -20,  11,  -8,   2,   0, -10, -11,   5,   9,   4,   4,  -1,   2, -28,  21, -28, -25,  16,   2,   5,  -1,   0,   2,  33, -30, -15,   3,  -1,   3,   5,  -3,   9,  15, -13,  -2,  -1,   1 },
      {   0,  -5, -10, -11,  10,   3,   1,  -1, -17,  10,  -5, -48,  15,   3,   3,  -2,  20,  35,  -6,  -1, -28,  11,  -5,   2, -31,   8, -24,  20,  -1,  -7,   7,   0,  38, -18,  -5, -36,  11,  -7,   3,   3,   2,  25,  30, -23, -21,   4,  -6,   0,  -1,   0,  29,  23, -23,  -3,  -2,  -2,  -3,  -1,  -3,  22,  -1,  -6,   1,  -1 },
      {  11,  -5,  -1,   8,  -1,   1,  -1,   0,  28,  12,  18, -11,   4,  -7,   1,  -1,  31,  -2,   2,  11,  18,  -7,  -3,   0,   0, -57, -18, -33,  29,  14,  -2,   4,  25,  12,  20, -29, -39,  19,  -6,   1,  -9,  14,   4,  41, -28,  -1,   6,  -5,  -9,  11, -16,  10,   5,  -9,   6,   0,  -7,   1,   5,   0,   4,  -1,   0,   0 },
      { -10,  -7,  10,  18, -35,  16,  -6,   2,   2,  43,  58, -21, -16,  -6,   6,  -5, -22,  -6, -33, -12,  21,  -8,   4,   2, -28,  13,   9, -16, -14,   7,  -5,   2, -10,  22, -12,  19, -22, -10,   7,  -3,  15,   0,  21,  -7,   4, -18,  -3,   0,   0, -18,  11,  28,  -3,   0,  -6,  -1,  -1,  -3, -13,   9,   9,   2,   1,  -3 },
      { -19, -34, -55, -22,  49, -19,  11,  -2,  -5,  -6,  19,  11, -22,   8, -10,   4,  -3,  -8,   5,  20, -27,   1,  -3,   0, -21, -24,  37, -20,  20, -10,  -2,   1, -21,  19, -11,   1,   3,   7,  -5,  -1,   9,  -6,  22,  -8,   1,   3,  -4,  -1,  15, -13,  -6,  15,  -8,  10,   0,   0,   0,   7, -12,  -2,   2,  -2,   4,   1 },
      {   1,  -8,  17,  -5,   1,   5,   0,   1,   7, -12,   2, -11,  46, -16,   3,  -2,  -5, -28,  -3, -12,  24,   3,   0,   2,   5, -24,  21,  -1,  -4,  13,  -5,   2, -63, -44,  12, -25,   9,  14,   4,   5,  23,  -2,   2, -24, -29,   9,   3,   2,  15,   3,   7,  13, -22,  -4,  -2,   0,   0,   7,  -7,  14,  -7,  -5,   2,  -2 },
      {  -4, -14,   3,  16, -11,  -8,   4,   1,  11,   6,   3,  24,   1, -18,   0,   0,  -6, -30,  15,  19,   2,  -3,   0,  -1,  14,  31,  36,   1,  -9,  -8,  -8,  -1,  41, -38,  12,  -3,  21,   6, -13,   1, -24,  41,  11,   6, -17,   9,   0,  -3,   3, -22,  -6,  36,  16,  -5,  11,  -2,   4,  -5, -28, -16,  17,   5,   0,   6 },
      {   6,   7,  33,  17,   1, -17,   2,  -3,  -3,  -6,  31,  -1,   3,  -4,  -4,   1,  10,  15,  54,  -8,  12,  -5,  -1,  -5,   2,  10,  12, -35,  24,  -5,  -1,  -1,  19,   5,  12,   4,   5,   9, -11,   1, -28, -26,   1, -19,  19,  16,  -1,   4,  19, -23,  22, -17, -33,  14,  -4,   5,   2,   6, -10,  34, -22,  -6,   2,   0 },
      {   1,   6,  23,  11,  -4,  -5,   5,  -3, -15, -16, -33, -32,  26,   1,   6,  -2, -15, -24, -15,  12, -14,  15,   0,   2, -46, -29,  19,  -1, -17,  -1,   5,   3,  -3,  24,  21,  -1,  -8, -14,  -2,   0, -35,  20,   3,  19,  23, -11,   5,   1,  34, -35,  -5, -16,  11,   7,  -2,   5,   9,  -4,  -2,  -4,  -7,   7,  -5,   2 },
      {  17,   5,   9,  27, -10,  -8,  -3,  -1,  20,   0,   6,   9,  -9,   4,  -7,   4,  16, -27, -28,  -8, -25,  20,  -4,  -1,   8, -39,  13,  -1, -27, -19,   8,  -3,  36,   2,  20,   9,  30, -32,  -3,   1,  32, -27,   1,  -5,  12,  18, -16,   4,   5,  25, -14,  13, -23,  14,   2,  -5, -10,  10,   0,  11,  -1,  -8,  11,  -3 },
      {   5,   9,  -1,   1,  55, -38,  11,  -3,  13,  29,  21,  12,  47,  -6, -10,   1,   0,   4, -26, -26,  23,  34,  -4,  -2,   0,   9, -18,  -8, -16,  25,   4,  -2,  21,  -6, -25,  -2,   5,  -4,   5,  -3,  -1,  -9,   1,  -4,   6,   2,   3,   3,  -1, -24, -16, -13,  12,   4,   1,   7,  -1,   3,  -3, -27,  -4,   3,  -4,   3 },
      {   7,  19,  21,  37,   7, -27,   4,  -6,   0,   6,  -4,  -9,  25, -13,   0,  -3,   1,  15,  13,  18, -33,   5, -14,  -2,  10,  22,  33, -16,  12,  -4,  -6,  -5, -13,  39, -20,  -8,  -5,   4,  -2,   2,  58,  36, -20,   6,  -8,  -3,   1,  -6,  -7,  12,  12,  -1,  10,  -4,   9,  -3,  -8, -13,  11,   2,   3,   4,   3,   4 },
      {   0,   6,  10,  29,  20, -12,   4, -10,   2,   5,  10,  47, -20,  21, -10,   4,  -8,   0,  18, -11, -32,  25,   2,   0, -13,  -5, -34, -12, -23,   3,  18,  -1, -27, -31,  12,   3, -31, -10,   6,   3, -13,  24,  -6,   6,  -5, -30,   4,   1,  17,  15,  21,  14,  -2, -16,  -6,  -2,  10,  -3,  -5,  23,   8,  -5,  -1,  -2 },
      {   0,  -4, -15, -22,   3,  -9,  10,  -1,   2,  -3,  -3,  16,   4, -37,   7,   2,  12,  14,  18,  34,  26, -31,  -5,   2,   1,  -5,  -9,  -5, -31,  30, -12,   4,  11,  -9,  28,   0, -14, -33,  10,  -4,  35,  25,  11, -11,  38, -20, -19,   8,  13,   6,   6,  -8,   3,  26, -15,  -2,   2,  -2,   3,   7,  -6,   6,   5,  -4 },
      {  -5,  -2,  -4, -26,   0,   7,   7,  -6,   1,  12,  28,   7, -18,   1,  -8,  -5,  -8,  -1,   3, -52,  14,  -1,   0,   1, -13,   4,   4, -12, -10, -23,  -4,   4,  -6,  19,  28, -27,  34,  -7, -14,   2,  -3,  48,  -9,  -2,  -4,  22,  -5,  -9,  20,  22,   2, -24,  10,   1,  12,  -5,  14, -15,  37, -12, -13,   3,   0,   6 },
      {  10,  -2,   8,   0,  -5,   5,  10,  -2,  12,  -8,  -4, -22, -22,  28,   6,   2,  10, -18,   3, -25,  -4,  -2,  17,  -4,  20,   0,  -2, -17,  30, -28,  10,   0,   9, -41, -20, -17,  -6,  -3,   1,   8,  43,   7,  18,   8,  15, -29, -10,   5,  20, -30,   5, -15,  32,  15, -24,  12,  -3,   5,   0, -14, -15,  11,  -7,   1 },
      {   1,   7,  10,  26,  26, -37,  10,  -4,  -8,  -7, -16, -27, -21,   7,  -9,   6,  -9,  -4, -11, -19,  23, -28,  -7,   1, -10, -10, -31,  15,  26,  -7, -16,   7, -19, -13,  24,  29,   4,   5, -18,  -3,  10,  20,   3,   2,  20,  25, -10, -12, -31,  -3,   5,  14,  -3,  19,  13,  -5,   3, -15, -26,  -4,  16,  -3,  13,   3 },
      {   4,   0, -10, -16,   1,   7,   4,   3,   4,  -2, -12,   0,  -5,   2,   2,  -1,   4,   5,  -7,  -4,   5,  -4,  -2,  -8,   5,   0, -30,   3, -12,   3, -23,   2,  18,  12,   1,  28,   2,  20, -21,  -7,  36,  -1, -34,   8, -14,  14,  24, -15,  73, -26,  11,  10,  -2, -24,  14,   8,  -6,  10, -19,  15,  10,  -3,  -6,   8 },
      {   7,  11,  13,  18,  18, -19,   8, -12,   0,   2,  -9, -24, -21, -21,   9,  -3,  -6,  -6, -11,  -9, -10, -40,   3,   9,   5,  19,   3,  -6,   1,  12, -19,   6,   1,  -9, -27, -10,  -7,   4,   6,  -3, -30,  -9,  10,  -9,  -1, -11,  -9,   0,  51,  40, -30,  12,  -7,   1,   0, -10,  -2,  31,  29,  -9,  14, -10,   6,  -6 },
      {  -3,  -3,  -3,   7,  -3,  -1,  -8,   0,   3,   8,   9,  22,   1,   5, -11,   4, -10,  -8, -15,  -8, -16, -10, -17,   6,   5,  13,   9,  52,   6,   8, -15,  -3,   7,  -4,   9,  -3, -31,  24,   2,  -5,  12,  18,  26,  19,   1,   4,   8, -11,  -3, -21, -21, -23, -11,  16,  14,   4,   3,  24,  16,  30, -52, -32,  10,   0 },
      {  -6,  -6, -11, -30, -50, -36,   6,   6,   2,   5,  -8,  11,  25, -44,   4,   4,  -4,   1,  -5, -18, -17,  19, -28,  13,  -3,  -1, -25, -18,  27, -26,   2,   1, -12,  -3, -20,  24,   3,  -5,  -5,  -3, -12,   3,  10,  21,  -1,   0, -16,  -5,   6,  20,   6,   2,  13,   8,  -7,  -6,   9,  22, -21,   9, -12,   3,  10,  -7 },
      {   0,   4,   2,  -2, -16,   8,   0,   9,  -3,  -4, -13, -22, -31,  26, -16,   5,   1,  10,  13,  22,  23,  34, -26,  -2,   4,  13,  -5, -29, -15,  27,  11,   2, -14,  -9, -32, -14,   0, -20,   0,   2,   6,  19,  -9,  15,   1,  25, -13,  -2,   3,  -6, -34,   7, -23,  11,  23,  -8,  31,  15, -15,  -8, -18, -15,  11,  -1 },
      {  -1,   3,   3,  -7,  14,  30, -13,   9,  -1,   5,  -2,  17,   9, -15,  23,   6,  -7,  -4,  -3, -18, -40, -34,  20,  -2,  -6,  -2, -19, -29,  -6,   9, -21,  -8, -11,  -6, -18, -21,  19, -22,  -7,   3,  -7,   6, -33,  29,   9,   9, -14,   0, -26, -39,  -1,  13, -14,  23,   9,  -6, -24,  13,   1,   8,  -1,  -7,   3,  -2 },
      {   6,   5,  12,   6,  30,  49, -43,  13,   3,   0,   0,   2,  23, -11,  -9, -12,   2,  -2,   4,   9, -16, -13,   0,  -1,  -3,  -3,  -9, -24,  -2,  -4,   6,   0,   1, -11, -11,  33,   3,   6, -14,  -7,  10,  -4,  51,   5,  12,  19,   2, -14,   8,  21,   8,  -6,  17,  -2,  18,  -1,  45,  -8,   4,  -9,  -3,   3,  -1,  10 },
      {   1,  -6,  -3,  -1,  -8, -31,  20,   3,   2,  -8,   2,  -7, -27, -21,  10,  -2,   1, -11, -11,  -8, -14,  -9,  -1,   5,   3,  -7,  11,   5, -17,  29, -11,  -2,   8, -10, -12, -28,   5,  -1,  -3,  -3,   5, -40,  -4,  24,  -4,   0,   6,  -6, -14, -11,  42,  -9,  11, -14,  16,  -2,  64, -28,  11,  14,  -1,  -2,  -6,  10 },
      {  -6,  -6,  -3,  -4, -14, -17,  -2,   4,  -6,  -3,  -1,  -6,   6,  17, -34,  25,  -5,  -4,   1,  17,  14,   2,  18, -16,  -5,   3,   4,  -4, -22,  13,  -1, -10,  -6, -15, -18,   6,  -6,  -3, -31,   5,  -9, -14,  23,  26,  -7,  18, -21, -12,  11,   9,  34,  -7,  19,   7,  14, -17, -61, -14,  32,   7,  -2,  -3,  10,  11 },
      {  -2,  -4,   3, -13, -32,  -3,  -8,   8,   0,  -4,   0,  -3,  -6,   6, -18,   0,  -3,  -7,   7, -17, -41,  15, -23,  10,   1,  -4,  -9, -21,  18,  51, -21,  -9,   5, -14,  15,   4,   8,  20,  20, -22,   3,   4,  11, -27,  23,  -7,  32,  -3,   1,  -3,  -7,   2, -11,  13,  14,   9, -24, -29,  16, -28,  19,   3,  -1,   9 },
      {  -2,   0,   2,   8,   1, -15,  10, -10,  -8,  -8,  -5, -11, -33, -36,  30,  -4,  -2,   3,   6,   7,   1,   6,  11,   0,  -5,  -7,  -8, -23, -26,  -8,  21,   0,  -9,  -9,   2,  16,  11,  -3,  -4,  -8,   9,  10,  28,  -8,  -5,  12,  39,  -7, -29,  -9,  -8, -19,  -1, -34,  15,  24, -31,  34,  12,  -4, -20,  -6, -24,  12 },
      {  -2,  -5,  -5, -10, -31, -23,  21,  -3,   3,   4,   2,   1,  37,  30, -13,  10,   4,   5,   6,   8, -11, -23,  36,  -6,  -4,  -1, -10, -17, -27, -30, -18,  13,   0,  -5,  -9,   0, -13,   8, -38,   6,   7,   1,   5, -18,  11,  -9,  23,  -6, -18,  -2, -34,  -3, -21,  -5,   2,   5,  31,  -6,  13,  -4,  18, -19,  -5, -16 },
      {   1,   4,   6,   5,   5,  -7, -34,  36,  -4,  -1,  -3, -14, -26, -28, -37,  27,   0,   3,   0,  -5, -16,   7,   9,   2,  -4,  -2,  -2,  -3, -31,   3, -10,   0,   2,   9,  12, -14, -11,  34, -30,  -2,   2,  -1, -16, -25, -20,  -7,  -3,   3, -11,  -3,   7,  -9,  16,  16, -33,  -4,   7,  25, -19, -23, -15,  15,   4, -26 },
      {   3,   4,   2,   9,  17,  30,  25,  -9,  -2,  -3,  -6,   3, -15, -20,  14,  -9,  -1,   0,  -1,   6,   6,  29, -39,  12,  -2,  -1,  -5,  -6, -12, -20, -42,  25,   0, -12, -18,   4,  -1,   9, -32,   6,  -5, -11,   8,  -1,  -5,  -8,   6,   0,  -6,  -6, -21,   0,  -2,   5, -23,  -6, -18, -53,  15,  21, -14,  12,   9, -24 },
      {   1,   5,   4,   5,  -1,  29,  46, -41,  -1,   3,   1,  -1,  11,  28, -23,   3,  -1,   2,   0,   5,  -1, -20, -25,  17,  -5,  -4,   1, -11, -24,   1, -21,  -8,   2,   5,  -2, -11,  25,  22,   6, -12,  -3,  -8,   4,  15,   8, -15,  17,  -4, -15,  25,  28,  -3,  13,   7,  -1,   4,  -5,  38, -35, -11, -16,  10,   9,   9 },
      {   2,   1,   2,  -2,  -6,   2,   8,   4,   3,  -1,   4,   8,   3,   7,   9, -30,   2,   0,   3,  -3, -12, -16, -44,  18,   3,  -6,   0,   2, -16, -24,   8,   1,  -1,  -7,   4, -15, -55, -14,  -8,   8,  -3, -18,  -9, -40,  -5,  22, -21, -26,  -1, -23,  10, -14,   7,  -1,  29, -25, -12,  14,  -1, -28,  13,  15,  -5,  15 },
      {   1,   4,   2,   0,  -6,  10,  31,   1,   1,   6,   3,   8,  -3,   7,  65,  -6,   0,   4,  -2,   2, -14,  38,  47,  -1,  -1,   3,   1,   8,   0,  26,  -6,   6,  -1,   6,   6,   3, -18,  37, -25,   2,   2,   5,   8,  -7,   8,  21, -24,   7,   4,  23,   9,  -3,  -3,  29,   7,  -3,   4,   5, -15, -19,  -5,   2,  11,  14 },
      {   1,   0,  -4,  -8,   2,  16,   7, -12,  -2,  -2,  -3,  -6,  -4, -39, -25,   2,  -1,  -4,  -8, -14,  -2,  -7,  15,   0,   1,   6,   5,   5,  -1,  -4,  48, -15,   3,   0, -13, -29, -18,  13, -19,  12,  -1,  -4, -15,  -9,  38, -12,  12,  -9,   6,  20, -24,   2,   2,   0,  26,   6, -17, -35, -44,  19,  -7, -11,   7,  19 },
      {   0,   0,  -1,   3,  -2, -26, -13,  17,   2,   0,   0,   0,   2,  17,  13, -11,   2,   5,   2,   6,  -5, -28,  18,   5,  -1,  -3,  -9,  -7, -19, -19,  -8,   7,  -1,   0,  -6,   5,  -1,  -4,  32, -23,  -7,  -6,  -6,  -2, -27,  -5,   4,   1,  11,  10,  -7,  16,  -8,   5,  14,   6,  -9, -44, -21, -23, -72,  19,  17,  16 },
      {  -3,  -2,  -2,  -5,   3, -20, -48,  16,   0,   1,   4,  -1,  -2,  24,  17, -30,  -1,   2,   0,   0,  10,  18,   2,  -6,   2,   2,  12,   3,   1,   1, -42,  33,  -3,  -3,  -3, -32,   8, -20, -15, -11,  -5,  -2,   3,   1,  25, -19,  13, -27,  -6,  26,  -3, -13,  21,  -4,  14,  10, -15,  18, -31,  30,  11,   5,  -9,   5 },
      {   2,   4,   5,   4,  12,  16,  34,  62,  -1,   1,   3,   2,  -1, -17,  -8,  20,  -1,  -1,   1,   1,  13,  14,  -1, -18,   1,   4,   6,  12,   5, -36, -11,   0,  -3,  -3,  -6,  -9, -21, -21,  14, -17,  -4,  -1,   0,  17,   2,  -2,  43,   0,   5,  15,  20,   5, -22,  38,  13,  19,   1,  10,   9, -13,  19,   2,  -4,  17 },
      {   0,   1,   0,   2,   1,  10,  14,  -6,   0,   1,   2,  -1,   2,   2,  -1,  18,  -2,  -3,  -4,  -7, -12,  -8,  -1, -21,   0,   3,   5,   9,  32,  33, -14, -23,   2,   1,   7,  17,  -2, -59, -46,  23,  -3,   0,   8, -14, -36, -27,   9,  -2,   1,  12,  -8, -23, -16,  -8,  13,  -1,   4,   0, -22, -12, -21,  26, -13,  -2 },
      {   0,   2,   1,   3,  -1,   1,  15,  27,   2,   5,   5,   1,   0,  15,   9,  22,  -2,  -1,  -4,  -4, -16, -26, -10,  22,   0,   3,   6,  -7, -11,  11,   4,   4,  -2,   0,  -8,  -5,  -4,  -8,  18,  -5,  -1,   9,  15,   8, -20,  45,  -8,  11,   9,  13, -12, -51,   9,  -7, -30,  24, -13, -26, -42,  15,  27, -19, -20, -15 },
      {   0,   1,  -1,   1,   4,  -8,   5,  47,   2,   4,   4,   5,   7,  21,  17,  14,  -2,  -2,  -4,  -2,   4, -16, -28,  22,   1,   2,   3,   8,  -4,   9,  25, -12,  -2,   0,  -1,  -8,  12,  24, -31,   0,   1,   6,   3,  18,  40,  -9, -18,   3,   2,   8,  -1,  24, -29, -34,  -2,   9,  -6,   5,  14,   1, -15,  47, -30, -10 },
      {   0,   0,   1,  -1,   0,   3,  12,  24,   0,  -2,   0,   2,  -1,  -8, -12, -42,  -2,  -3,  -1,  -4,  -8,  -5, -11, -46,  -1,  -1,   2,   2,   2,   3,  -9,  23,   0,   1,   4,  -3,   1,   2,  -2,  21,   0,   1,   7,  15,  21,  -8, -12,  19,   0,   3,  23,  -7, -19, -46, -11,  -5, -14,  -3, -15, -45, -13, -50,  17, -18 },
      {  -1,  -2,  -4,  -2,  -6,  -9, -28, -46,   0,   0,   0,   7,  -6, -13, -19, -28,   0,   0,  -4,  -1,  -1,   0,  15,  16,   4,   7,   3,  13,  -4,  -3,   6,  23,  -2,  -5, -12,  -8, -15,  -3,  -5,   5,   7,   4,   8,  38,  -1,  22,  17,  12,   8,   0,  17, -17, -56,  18,  -8,  12,  -3,  -3,  -8, -24,  15,  45, -10, -19 },
      {  -2,  -2,  -4,  -3,  -5, -13, -16, -16,   0,   1,   0,   1,  -1,  13,  26, -11,  -3,  -3,  -4,  -2,  -2, -14, -34, -61,  -2,   0,  -2,  -1, -13,   2,   5, -39,  -2,  -1,  -5,  -8, -13,   5,  -5,  30,  -7,  -2,  -4,  -5,   9,  35,  35,   9,   2,  15,   2,  26,  23,  29, -24,  24,   5,  -3,   5,   2, -10,   5,  -8, -15 },
      {   0,   2,   1,   1,  -3,   7,  15,   0,  -1,  -1,  -1,   3,   6,   1, -28,  -9,  -1,   0,   0,   1,  -8,  -1,  11,  35,   1,   3,   3,   4,  18,  21,   3,  52,   1,   5,   2,   4, -10, -27, -15,  15,  -2,  -6, -16, -11,   5,  28,   4,  -1,  -2,   6,  10,  46,  28,   7, -17,  39,  -1,   6,  15,  -4, -30, -18, -35,   9 },
      {  -1,  -2,   0,  -2,   4,   1, -38, -32,   1,   1,   3,   0,   1,  -3,  28,  66,   2,   1,   4,   3,  11,  12, -15,  27,  -1,  -1,   4,   4,   1, -21, -25, -15,   0,  -1,   1, -11, -13,  -8,  -3,  10,   0,  -7,  -4,   2,  17,  -6,  -8,  -7,   7,   5,  20,   6, -13, -18,   5,  25,   1, -12,  -4, -40,  -4, -42,  -6,   4 },
      {   0,   0,   0,   0,  -4,   9,   9, -13,   1,   1,   0,   0,   1, -10, -16,  41,   0,  -2,  -2,  -4, -13,   9,  15, -39,  -1,   0,   0,   1,   7,  -4, -35,  20,   0,  -3,  -1, -10, -22,  -1,  38,   3,  -1,   3,   4, -10,  30,  40, -10,   2,  -5,  -1, -12,   5,   4, -46,  -4, -15,   5,   6,  -4,  12, -11,  48,  -2,  13 },
      {   0,   1,   0,   1,  -1,  -5,  -5,   8,   0,   0,   0,   1,  -2,  -2,   2,  17,  -1,   1,  -1,   1,  -2,  -5, -14, -11,  -2,  -1,  -7,  -7, -14,   1,   1,  40,   3,   5,   2,  12,  20,  29,  29,  83,   0,  -2,   6,   5, -12, -25,  -7, -15,   0,   1,  -3, -10, -12,  28,  16,  -7,  -1,  -3,  -4,   3,   5,   3, -35,  29 },
      {   0,   0,   0,   0,  -3,  -2,  -5,  -4,  -2,  -2,  -4,  -4,  -4,  -7, -23, -27,   0,   1,  -2,   0,   1,   7,  -4,  -6,  -3,  -4,  -6,  -2,  -8, -16, -36, -33,  -4,  -1,  -3,   2,  -8,  11,  -9, -11,  -2,   3,  -1,  -2,  -8,  -7, -43,  66,  -5,   3, -10,  -3,  -3,   2,  14,  31,   3,  -1,   0,  11,   5,  -4, -43,  44 },
      {   0,   0,  -1,   0,   4,  -6,  -2,  19,   0,   0,   0,  -1,  -2,   5,  -3, -11,   1,   2,   2,   1,   7,   7,  21,  52,  -1,  -2,  -2,  -2,   4,  -6, -24, -32,   2,   1,   1,   0,  -3,  -7,  13,  52,   5,   3,   7,   0,  16,  16,  40,  46,   5,   3,   6,   4,  18,  -8,  23, -44,  -4,   4,  -7,   5,  -8,  -1,   1, -24 },
      {   1,   2,   1,   2,   1,   1,   9,  -2,  -1,  -1,  -2,   0,  -3,  -6, -10, -21,   2,   2,   1,   4,   4,  12,  19,  -1,   0,  -1,   1,   1,   3,  -4, -16, -41,  -1,   0,  -1,   0,   1,   2,  15, -14,   0,   2,   0,   8,   7,   4,  -8, -60,   1,   6,  -3,   7, -15,   6, -33, -41,  -1, -10, -10,  -9, -14, -18, -70,  -9 },
      {   0,   0,   1,  -1,   1,   5,   6, -13,   1,   0,   1,  -1,   2,  -1,   4,  17,   0,  -2,   0,  -2,  -1,  -1,  -4, -23,   1,   0,   1,   0,   2,  -4,  11,  26,   0,   0,   2,   1,   0,   8,   9, -31,   1,  -4,  -2,  -5,  -1,  -7, -11,  23,   4,   3,  11,   4,  17,  17,  63,  -9,  -5,  -5,  -6,   0,  -7,  -5, -47, -74 },
      {   0,  -1,  -1,  -2,   0,  -3, -14,  -9,   2,   3,   2,   0,   3,   6,  14,  20,   0,  -1,  -3,  -3,  -1,  -9, -19,  -8,   2,   2,   2,   2,   2,   2,  18,  32,  -1,   0,  -2,  -4,   1,  -4, -25, -32,   3,   2,  -1,   4,   2,   0,  32,  44,   0,   1,   0,   1,  -2,   7, -27, -70,  -1,  -2,  -6,  -4,  -5, -17, -24,  47 },
    },
    {
      {  92,  -8, -21,  -4,  -4,  -1,  -1,   0, -74,  19,  22,   0,   2,   0,   1,   0,  10, -27,  -3,   6,   1,   1,   0,   1,   4,  11,  -6,  -4,   0,   0,   0,   0,  -1,  -1,   5,   1,  -1,   0,   0,   0,   0,   0,  -2,   1,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0,   0,   1,  -1,   0,   0,   0,   0,   0 },
      {  40, -56,   2,   7,   1,   2,   0,   0,  40,  55, -33, -10,  -4,  -2,  -1,   0, -55,  16,  35,  -1,   1,   0,   0,  -1,   8, -22,  -9,  11,   1,   1,   0,   1,  -7,  12,  -4,  -8,   3,  -1,   0,   0,   1,  -5,   6,   1,  -3,   0,   0,   0,  -2,   2,  -1,   1,   2,  -2,   0,   0,   1,  -2,   2,  -1,   1,   1,   0,   0 },
      {  54,  18,   2, -17,  -4,  -3,   0,   0,  16, -50,  -2,  19,   3,   4,  -2,   0, -18,  62, -15, -20,  -2,  -3,   1,   0, -17, -26,  49,   9,  -3,   1,  -1,  -1,  11, -13, -25,  16,   3,   3,   1,   2,  -9,   6,   0, -11,   1,   1,   0,   0,   2,  -4,   1,   4,  -6,   1,   0,   1,  -3,   2,  -2,   1,   1,  -1,   0,   0 },
      {  -9, -66,   4,  15,   6,   3,   1,   1,  -8,  27, -14,  -8,  -2,   0,   1,   1,  71,  34, -15, -13,  -7,  -3,  -2,  -1, -46,  16,  25,   5,  -1,  -3,   0,   1,   0, -23,  -3,   8,   6,   1,   0,   0,   1,   9,  -5,  -4,  -1,   0,   1,   0,   1,   0,   5,   0,   1,  -2,   0,   0,  -1,  -1,   0,   0,   1,   1,   0,   0 },
      { -26, -10,   6,  -2,  -3,   2,   4,   3, -39,   1,  12,   1,   7,   1,  -3,  -4,  20, -18,  41,  -9,  -8,  -3,   2,   2,   9, -75,   6,  39,   8,   8,  -4,  -2,  21,  11, -52,  -7,   0,   9,   4,   4,  -5,   7,   4, -12,  -6,  -2,  -2,   0,   4,   3,  -5,   5,  -9,   2,   0,   1,  -3,   2,  -2,   2,   2,  -1,  -2,  -1 },
      {  13,  25,  46,   5,  -6,  -5,  -6,  -4, -28, -19, -63,  10,  13,   6,   8,   5,  20,  21,  14, -21, -12,  -4,  -5,  -3,  -5,  17, -31,  23,   3,  -3,   2,   2, -12,  25,   5, -39,  10,  -4,   0,  -2,   7,  -7,  30,  13, -14,  -2,   0,   0,  -2,   2, -11,   3,  16,  -4,   0,  -1,   2,  -2,   9,  -8,  -1,   3,   1,   1 },
      { -16, -35,  20,  -9,   1,   3,   2,   1, -25, -11, -16,  28,   2,  -2,  -1,   0, -10,  -6, -14, -35,   7,   5,   5,   1,  74,  16,  25,   4, -28,  -8,  -5,  -1, -39,   0,  -4,  31,  20,  -1,  -2,  -1,  10,  -8, -15, -16,   1,   7,   2,   1,  -5,  10,   0,   0,  -5,  -3,   0,  -1,   4,  -1,   2,  -2,   1,   0,  -1,   0 },
      { -19,  64, -22,  13,  -2,  -2,  -4,  -2, -32,  57, -25, -13,   2,   1,   3,   1, -15,  29,  30,  -9,  -5,  -7,  -1,  -3,  -2,   0,  28,   4,   2,  -5,   0,   1, -14, -40,   7,  17,  12,   1,  -1,  -1,  12,  -7, -19,  -7,   6,   5,   3,   2,   2,   0,   9,  -9,  -2,  -2,   0,   0,   5,  -4,   2,   4,  -1,  -2,  -1,   0 },
      {  24,   8,  59,  -9, -18,   0,   2,   1,  31,  -6, -13,   3,  15,  -5,  -3,  -3,   8, -58,  23,   1,  -8,  10,   1,   3, -17,  10,  11,  20,   9,  -3,  -1,  -4,   3, -49,   2,   8,  -2,  14,   2,   5,   7,   3, -43,   5,  -2,   0,   1,   2,  -3,  12,   3, -18,  -8,   0,   0,   1,  -3,  -1,  -1,  10,  -3,   0,  -2,  -2 },
      {   5,  20, -11, -13,   2,   2,  -3,  -2,  22,  24,  11,  19,  -8,  -7,   0,   1,  -1, -25, -18,  -3,  15,  10,  -1,  -1, -20,  19,  -7,  17, -12,  -5,   2,   3, -25, -11, -44, -19,  35,   1,   2,  -1,  22,  29,  30, -46, -31,   1,   1,   1,  -2,  -3,  17,  33,   1, -14,  -3,  -1,   0,  -1,  -5,  -7,  13,   6,   0,  -1 },
      { -17,  -7,  49,   8,   6,  -2,  -4,  -3, -40,  17, -26,  11,  -8,   2,   5,   4, -48, -11, -39,   4,  21,  -1,   1,  -1, -24,  -2,   3, -33,  -4,   2,   4,   3,  50,  -4, -14,   4,  -9,  -9,  -4,  -1,  -7,  32,   3, -12,   8,  -3,  -1,  -1,   8, -10,  14,  -4,   2,   2,  -1,  -1,   0,   1,  -1,   0,  -2,  -1,   2,   0 },
      {   1, -14,  -6,  43,   6,  -6,  -2,  -2, -11, -38,  -2, -39,   6,  14,   5,   3, -17,  20,  -1,  31, -29,  -9,  -3,   1,  24,  20, -27,  12,  26,  -4,  -2,  -1, -16, -20, -26, -10,   2,  10,   5,   0,   1,  37, -22, -26,  -1, -12,   4,   1,  -1,   2,  36, -11,  -3,  -5,  -6,  -2,   6,   2,   1,   9,  -1,  -1,   1,  -1 },
      {  21,  24,  -6,  20,   2,  -7,  -3,  -2,  27,   6, -34, -18,  -1,   9,   2,   1,  37, -22,   8, -10,  -3,  -6,  -1,   0,   8,   2,  -4, -10, -26,   3,  -1,   3,  11,  48, -14,  40,   4, -21,  -4,  -4, -30,  13,   2, -29,  33,  13,   1,  -2,   3,  -7,  24,  -7, -24,   3,   3,   0,  -1,   1,   0,  15,   4,  -9,   0,   2 },
      {  -8,  -9, -34,  22, -17,   8,   2,   6,   0,  -8, -10, -16,  38,  -5,  -2,  -6, -22, -24,  -1, -36, -29,  12,   2,   4,   6,  35,  31,  14,  -2, -23,  -3,  -8,  50,  -3,  10, -16,  12,  17,  -1,   2, -22,  -8,  24,   4, -21,   5,   2,   2,  13, -15,  -4,  17,  -9,  -1,  -2,   2,  -7,   6,  -7,   7,  -2,  -2,   0,   0 },
      {  -7, -11,  -7,  10,   7,   6,  -4,   0, -12, -20,  26,  15,   3,  -9,   5,   2, -23,  11,  21, -10,   3,  16,  -6,  -2, -42,   6, -43,  25, -33, -16,   9,   6,  14,   4,   6,  14,  48, -17, -14,  -8,  -9, -10, -28,  -8,   8,  33,  10,   0,   5,  11,  -8, -27,  -2, -14,   7,   1,   8,   0,  14,  -5,  -7,  -3,  -4,   2 },
      {  -1,  -8,  26,   7,  18,  -5,   3,   4, -14, -17,   4, -17, -13,   7,  -2,  -3, -10,   0,  13,  19,   7, -13,   1,   0, -18,   3,  10, -13,   8,  13,  -1,  -3, -11,  -1,  18,  -7,   5,  -3,   3,   3,  12, -57,  16, -33, -11,  24,   0,   6,  22,  10,   3,  32, -56,  -7,  -2,   1, -10,   2, -10,  29,  21, -11,  -4,  -5 },
      {   7,  12,  48,  46,   6, -15,  -4,  -5,   5,  13,  48, -49,  14,   9,   3,   1,  -7,   0, -21, -11, -18,   6,   3,   3,  -1, -16,  14,  14, -25,  -8,  -1,   3, -15,   6,   2,  16,   9, -14,  -4,  -3,  14,  -4,  14,  16,   6,   3,  -6,  -1, -17, -14, -16,  22,  18,   4,   5,  -2,   5,   3, -11, -21,   8,   6,   4,   4 },
      {   1,  10,   9,   9, -17,  -7,   5,   1,   6,  35,  10,  21,  22,   6, -10,  -5,  -1,   4, -64,   7, -20,   2,  10,   6,  -4, -15,  -2,  16,  32, -12,  -9,  -6, -16,  31,  -6, -16,  14,  23,   7,  -1, -17, -34, -13, -18,   8,  -1,   9,   3,  14,  11,  -2, -36, -15,   3,  -3,   3,  -1,  -2,  19,   8, -13,  -4,  -3,  -3 },
      {   3,  -4, -39,  49,  -7,   2,  -1,   1,   1, -25, -31,   8,  14,   3,   1,  -1,  -7, -29, -24, -27,  13,   5,   9,   3, -27, -43,   1, -14,  -7,   3,   0,   1, -19,   0,  11, -10, -16, -11,  -1,   2,  54,  -1, -14,   6,  -3,  -1, -11,  -2,  -8,  26,   6,  -1,  -3,   2,   4,  -2,   1,  -7,   3,   3,   4,   0,  -1,   2 },
      {   3,   3, -15,  24,  25,   5,  -5,  -1,  -1, -14, -30, -11, -38,   4,   8,   4,  -5, -28,  -9,  34,  16, -17,  -5,  -4,  -6,  15,  38,  17,  11,   7,   1,  -1,   1,   8, -30,   8,  15,   7,   1,   2, -19, -22, -10,   8,  -6,   3,  10,   2,  -3,  11, -47,  -8,  22, -13,   2,   0,   2,   5,  14, -37,   7,   6,  -7,  -2 },
      { -13, -11, -11, -30,  10, -18,   4,   2, -10,  -3,   4,   9,  -6,  29,  -6,  -5, -29, -17,   9, -15, -34, -26,   7,   7, -38,  35,  11,  31,   5,  15, -16,  -9, -14,  18, -14,  11, -13, -18,  20,   0,  18,  -6,   9,   3,  36, -16, -21,  -2, -12,  -1,   9,  -6,   4,  29,   6,  -6,   8,  -9,  -5,   1,  -6, -12,   7,   6 },
      {   7,  10,   7, -16,  58,  15,   6,  -5,   4,   3,  12, -30, -46,  -7,   6,   5,   1,  -6, -15, -46,  -3, -12,  -4,  -3,   7,  -8,   8,  24,  -3,  -3,  -9,   1,  19,  14,  26, -19,  -2,  11,   4,  -3,   8,   8, -12,   4, -20,  -9,   6,   3,   3,  18,  29, -17,   5, -15, -12,  -1,   1, -10,  17,  24,  -9,   9,  -1,  -3 },
      {  -6,   1,  -2, -13, -20,  -7,   4,  -4,  -7,   4, -10,  10,  -2,  12,  -7,   0,  -6,   5,  -4,  16, -35,  -1,   8,   4, -14,  -7, -12,   2, -10,  -7,  -8,   1,  12,  26,  19,  38,  -3,  17,   7,  -7, -10,   1, -30,   3, -24, -22,  14,   5,  -5,  35,  10,  42,  17, -29, -17,  -7,  -7, -16, -14,  -6,  47,  18,   3,  -5 },
      {   5,   5,   3,  51,  10,  -4,  -8,   2,  11,   9,  19,  51, -23,  23,   7,  -3,   1,  -1,   6, -37,  -2, -27,  -2,   2,   9,   4, -28,   3,   5,  16,  -9,   4,  -3, -35,   3,   5, -14,  -4,  16,   4, -39,   1,   4,   2,  -3, -11, -13,   0,  35,   7, -14,  13,   6,   4,  -2,  -7, -17,   4,  -3,   1,  10,  -3,   3,   2 },
      {   3,   0,  -8, -12,   9, -22,  -8,   2,  10,  11, -12,  15,  -7,  30,   5,  -2,  12,   4, -27,  13, -14, -22,  -3,   6,  31, -12, -14,  22,  15,   3,   1, -12,  46, -21,  21, -12,  25, -15,  -4,   7,  23,  10, -11,   4,   5,  42, -11,  -7, -34,   1,  -6,  14, -17,   1,  30,   4,   9,   4, -10,  -9,  16,  -8,  -8,   8 },
      {   4,   0,   4,  24,  21,  15,  16,   6,  -1, -11,  -7,  34, -28, -11, -17,   0,   4,  -8,  15,   8, -11,  -2,  18,  -4,  -6, -14,  -2,   1,  24, -35, -17,   5,  -1,   7,  28,  19,  16,  32,  -5, -15,  21,   3,  11, -14,  19,  -3,  18,  10, -14, -53,   5,   4,   1,  13, -10,  11,   3,   3, -28, -12, -13,  10,   6,  -4 },
      {  -2,  -6,  -7, -30,  47, -36,  -2,  -9,   2,  -3,  -7, -11,  39,  41,  10,   3,  -2, -16,   5, -24, -11, -11,  11,   4, -13, -18, -15, -43,   6,  -7,  -3,   6, -20, -17,  -3,  10,   2,  25,  -9,  -3, -18,   2,   8,  -6, -15,   6,  23,  -1,  -4, -11, -13,  -7,   2, -17,  -1,  11,  10,  10,  10, -16,  -5,   2, -13,  -3 },
      {  -3,  -1,  -9, -29,   7,  20,   5,   2,  -7,  -5, -26, -23, -11, -17,   3,   1,  -8, -10, -30,   8, -34,  15, -16, -10,  -4, -25, -16,  21,  -4, -12,   5,   7, -29, -22,  19,  16,   3, -12,   4,  -2,  -6,  30,  16,   6,  18,   2, -11,  -2,  43, -10, -34,   0, -14,   6,   6,  -2, -32,  13,   0,   4,  17,  -9,   1,  -4 },
      {  -3,   0,   8,  13, -15,  11, -17,   2,  -1,   5,  19,  24, -14,   7,  15,   1,  -8,  -4,  13,   4, -28, -20, -17,  -3,  -8,   9,  27, -14,  12,  -7,   8,   1, -23,  28,  -6,   4,  -9,  10, -32,   0,  11,  42,  13,  34, -14,  29,  26, -18,  -3,   9,  -6,  -8, -15, -22,  20,  15, -10,  -8,   4,  30,  -3, -18, -14,  -3 },
      {  -7,  -8,   0,  -7, -35,  -9, -12,   0,  -6,  -9,   3, -15,   6,  -1,  19,  -1,  -8,  -9,   0, -14,  43, -26, -13,   0, -15, -11,  15,  21,  20,  22, -12,  -3, -28,  20,  47, -10,  13,   5,  15,   7, -32,  25, -21, -17,  -6,  13,  -5,  -3,   2, -36,  14,   1,   8,  -9,   8,   0,   0,  -1, -14, -16,   7,   6,  -7,   5 },
      {   4,  -4,   5,  -5, -20,  -9,  -6,   1,   7,  -6,  10, -14,   7,  -7,  10, -10,   6,  -1,  13,   3,  -1, -35,  -9,  -2,   8,  -3,   4, -32,  15, -12, -23,   4,  21,   2,   6,  18,  40,  10,  15,  -8,  34,  12,  25, -16,   4,   5, -15,   8,  38,  35, -10, -14,  19,  14,   0, -13, -14, -35,  16, -17, -13,  10,  10,  12 },
      {   0,  -3,   1,  -3,  18,  -6,  10,  -4,  -2,  -6,  -4,   9,  -1,   6, -12,   3,   2, -12,   6,  15, -11,   5,   4,   4,   1, -14,  22,  -7,   4, -30,  19,  -5, -18,  -7,   0, -29,  13, -24, -36,   8, -33, -16,  -3,   8,  10,  15, -11, -24,  25,  12,  49,   8,  36,  17,  18,   1, -16, -28, -18,  -6,   4,  19,   4,  18 },
      {  -3,  -9,  -7,  -6, -48, -42, -12,  -7,   2,  -6,  -3,  -8, -39,  43,   4,  14,   4,  -7,  -3,  13,  15,  -8,  13,   0,   3, -12,   4,  18, -31, -32,  -2,   4,   4, -18,   7,  -6,   5,  -5, -14, -12,   8, -12,  18,   8,   2, -18,  16,   3,  20,  -9,   7,  -7,  10,  -4, -21,  11,   7,  24,  11,  27, -17,  -1,  -2, -12 },
      {   1,  -5,   1,  -6, -21,  42, -19,  -8,   3,  -4,  -1,   1,  -6, -21,  29,  11,  -1, -13, -12, -10, -34, -29, -30,  11,  -1, -17, -10, -14,  11,  15,  22, -11,   0, -20,  -4,  16,   9, -15,   0,   3,  -7, -36,  23,  -6,   8,  -8,   5,  -4, -11,  -3,  23,  -2,  13, -12, -11,  11,  47,  11,  -7,  -1,  -4,  19, -12,  -8 },
      {   2,  -1,  -1,   1,  31,   1, -30,  -5,   6,   5,   1,  30,  29,   4,  40,   9,   5,  -2,   8,  33, -23,  -2, -17, -12,   3,  -6,  24,  -7, -22,  -4,  -6,  -2,   3,  21,   8, -14,   8, -19,  14,  12,  19,  -1, -35,  -3, -17, -14, -18,  12,  25, -28,  -1,   9,   9,  -1, -18, -20,  -2,  27,   1,  12, -18,  -3,  12,  -1 },
      {   2,  -2,  -3,  -4,   2, -11,   2,  -4,   6,   3,   8,   1,  14,   1,  26,   4,   2,   4,  -1, -16,  10,   3, -12, -14,  10,  17,  -6,  17,  16,   5,  14,  -1,  18,   6, -16,   2, -31,  -7, -11,  -4,  43, -20, -12,  -9,  31,  -3,  33, -10,  36, -23,   7,  -6,   8, -21,  -2,  30, -37, -10,  -8, -21,  27,  21, -18,  -4 },
      {  -1,  -1,   2,  -6, -12,  -6,  -3, -29,   2,   1,  12,  -2, -12, -20,   5,  34,   0,  -1,  14, -27,  -3,  -1,  19, -15,  -2,   8,   7, -17,  41, -40,  -6,  14,   1,  12,  -3, -15,  -7, -15, -21,   5,  -1,   8, -18, -14,  17,   0, -18,   0,  -1,  15, -13,  10,  -8,  12,  -1, -15,  23,  44,  22, -11,  41,  -4,   5,  -6 },
      {  -4,  -4,  -5,  10,  19, -36, -17, -17,  -2,  -2,   4,  18,  16, -20,  28,  11,  -3,  -4,   4,  11,  12,  14, -18,   0,   1,  -5,  15,  12,   6, -11,  11,   0,  -2, -10,  15,  -5,  12,  -4,  14,  -2, -20,  10,  34,   7,  28, -15,   1,  21, -44,  22,   0, -18, -16,   3, -32,   5, -20, -28,  -2,  17,  26,  12,   0, -21 },
      {  -2,   0,  -6,  -5,   3, -19, -31, -30,  -3,   0,  -9, -11,  -8, -26,  29,  10,  -2,   2, -12, -11,  15,   6, -10,   7,  -5,   2, -24,   8,  17,  -3, -10, -13,   0,   4, -18,  31,   4,  28, -18,   2,   0, -22, -10,  10, -20,   1,   4, -11, -11,  11,   8,  22,   9,  39,  16,   0, -34,  15, -28,  14, -30, -22,  15,   6 },
      {   0,  -2,  -2,   8,   8, -53,  10,  -4,   0,  -1,  -1,  10,  -3, -20, -27,  -6,   0,  -4,   2,  -1, -10,  24, -17,  -1,   4,   1,  10,  18,  33,  30,  45,   8,   4,   8,  25,  28,  -8, -16, -15,   6,   8,  14,  10, -20, -18,   0,   3,  -6,  22,  12,  -7,  10,  10,  -8, -10,  -1,  31,   8,   5,  -2, -34,   0,   7,   0 },
      {   2,   1,  -3,   3,  20, -13,  16, -11,   2,   1,  -5,   6,  31, -32,  -9, -10,   1,   1,  -8,  14,  27, -28,   0,   0,   2,   2,  -3,  17,  -6,   5, -30,   6,  -1,  -2, -11,   5,   8,  14,  -4, -16,   4,   5,   9,  42,  14,   6,  11,  -2,  36,   5,  10,  -1,  11,  -6,   4,  -4,  52,  15, -22,  31,  33, -10,   0,   9 },
      {   0,  -2,   2,   3,   7,   2, -47, -18,  -1,  -3,  -1,   5, -13,   3,  15,  -3,   3,   3,   8,  11,   0,  57,  23,  19,   1,  -1,   4,   7,   6, -11, -25, -21,   2,  -2,   9,  14, -20,  13,  26,   7,   1,  -2,  21,  -8,   3,  -3, -16, -14,  18,  -6,   7,  -6,   1, -23,  40,   0,  29, -18,  18,   1,   9,   5, -21,  24 },
      {   1,   2,   1,  -5,   8,  25, -22,  23,   2,   3,   1,  -5,   6,  46,  10, -13,   1,   3,  -1,  -5,  26,  13, -45,   0,  -1,   3,  -1,  12,   2, -32,  26,  12,   1,   8,   4,  -2, -14,  29, -18,  -6,   1,   4,  -7, -14,  11, -13,  -5,  -3,  12,  19, -18,   8, -12,  26,  -6,  -5,  43, -19, -28,  -3,   6,  -4,  20,  -6 },
      {   0,   1,   2,  -2,   7,  32, -15, -19,   1,   3,   2,   2,  23,   2,   6,  12,   0,   2,   1,  16,   1,  -7,  32,  10,   1,   0,   9,  23,  -8,   9, -26,   5,   0,  -7,  22, -10, -29, -17,  -4, -15,  -8,  17,  12, -23,  17,  26,  34,   7,   6,  46,   2,  21,   8,   9, -18,  20,  -1,  10,  -7, -28, -34, -15,  -4, -13 },
      {   0,   4,   3,  -1,   6, -25,   6,  42,  -2,   2,  -3,  -2,  -8, -36,  24, -56,   0,   6,   3,   0,   2, -21,   5,  43,  -4,   3,  -4,  -3, -12, -39,   0, -23,  -6,   6,  -2,  -8, -15,  -5,  -3,  15,  -4,   9,  -9,  -6,  10, -13, -12,  -6,  -1,  16,  -7,  13, -10, -10,   3,  18,   0,  26,  -2, -11, -16,  18, -21, -12 },
      {   0,  -1,  -4,  -9,  -5,  -9, -34,  10,   1,  -1,  -4, -12, -16,  -7,   5, -15,   2,   1,  -8, -15,  -9,  13,  40,   2,   3,   3,  -6,  -9,  24,   5, -15,  14,  -4,   0, -23,  -7,   7, -37, -10,  -1,  -4,   1, -18,  21,  -9,  19,   2,  22,  14, -14, -19,  -5, -14, -11, -32, -12,  17, -42, -47,   5, -12,  12,   5, -17 },
      {   0,   0,  -2,   4, -12, -29,  36, -31,   0,   1,  -1,   3, -22, -16,  16,  25,   0,   0,   0,  -3, -23,   6, -28, -19,   0,  -2,   2, -15, -33,  -5, -11,   8,  -2,  -1,  -7, -29, -12,  21,  31,   1,  -6,  -1, -11,   1,   8,   5,  -1,  -3,  10,   9,  -3,   0, -29, -10,  18,   9,  22, -19, -38, -33, -25,  -9,  -3,  -3 },
      {  -2,   0,   0,  -3,   0,  -2, -26,  11,  -2,   2,   0,   0,  -5,  13, -27,   2,  -1,   0,   1,  -7,   3,  15, -18,   7,  -2,   2,   2,  -5,   9,   0,   8,  -2,  -5,   5, -11,  -1,  -1,  25,  27, -10,   5,   0,   9,  30,  28,  43, -17,  14,   5,  15,  32,  -2,  -5, -36, -16, -18, -19,  50, -11, -20, -11,  26,  27,  -1 },
      {   1,   2,   2,   2,   3,   6, -17, -23,   0,   1,   3,  -3,  -8,  10, -24,   0,   2,   5,  10,  -6,  -6,  12,  11,  34,   2,   4,   8, -18, -11,  36,  15, -25,   3,  12,  10, -18,  44,   8, -10,  10,  11,  16,  -4,   2,   3, -40,   8,   1,  19,  15,  -8, -28,  -9,  18,   5,  20,  -7,  28, -39, -14,  20,  -2, -17, -10 },
      {   0,  -2,   0,   0,  -5, -22, -20,  30,   0,  -3,  -2,   3, -21, -15,   8, -28,   0,  -4,   1,  -5, -26,  30, -27,   5,   1,  -5,   7, -16, -12,  33, -30,  12,  -1,  -7,   1, -26,   0,  12, -10, -30,  -3,  -5,  -8,  -6,   7,  18,  33,   0,  -1,  -8,  14,   6,  15,  41,  -8,  -3,   8,   4,  32, -12,  29, -13,   6,  -7 },
      {  -1,   0,  -3,   2,   0,  -3, -12, -30,  -1,   1,  -2,   4,   5,  -9, -20,  17,  -2,   2,  -3,   8,  -3, -22, -10,  27,  -2,   4,  -6,  13, -23,  16, -16, -15,  -2,  -4,   0,  -1, -17,  19, -45,  -1,   2,   6,  -4,   1,  -1,   5, -14, -10,   7, -19, -10,   6, -36,  14,  -3, -17,   2, -20,  30,  -7,  -6,  65,  15, -17 },
      {  -1,  -3,  -1,  -4,  -2,   0, -19, -10,   1,  -1,   0,   3,  -1, -15,  26, -28,   0,  -3,  -3,   5, -10, -21,  44,  -6,   3,   0,   6,  10, -12,  15,  47,  48,   1,  -4,   6,  -9,  -1,  24, -16, -40,  -2,  -7,   3,   0,   7, -22, -10,   6,  -4,  -1,  11,  -5, -23, -13,  21, -18,  -6,   2,   1, -18,  -1,  -7,  17,  33 },
      {  -1,   2,   0,   1,   0,  -6,   9,  56,   0,   3,   1,   3,   5,   4,  23,  34,  -2,   2,   0,   6,  -3,  14,   5, -29,  -1,   3,   3,   7,   3,  17, -38,  15,   1,   0,   7,  12,  -2, -10, -26,  -6,  -4,  -2,  18,  -6, -24, -15,  -8,  -6,  -6,  23,   5, -31, -14,   6,  22,   4,   0,  28, -28,  -3,  -1,  45,   1,  15 },
      {   0,   1,   0,   1,  -7,   7,  52, -21,   1,   2,   1,   4,  -8,  18,  46, -15,   2,   2,   3,   2,   3,  13,   3,  35,   0,  -1,   2,   4,   6,  -4,   9, -37,  -4,   1,   1,  10, -15, -13,  -7,  -8,   3,   5,  14,   4, -28,  21,  17,  30,   7,  -4,   8, -29,  -6,  12,  -8, -37,   3,   9,  -5, -16,  17,  -2,  26,  11 },
      {   0,   0,   1,   0,   3, -12, -32,  31,   1,   0,   2,   0,   0, -21, -14,  20,   1,   0,  -2,   1,  -2, -24,  26, -36,  -1,  -2,  -7,   8, -22, -13,  29, -36,  -2,   0,   2,   9, -15,  12,  15,  42,   1,   2,  10,   9, -19,  10,  26,  -3,  11,  -1,  18, -18, -16,  27, -14,  -9,  11,  -8,   4, -33,  18, -16,   8,   3 },
      {   0,  -1,   1,  -1,   1,   3,  -8,  -1,   0,  -2,   1,   0,   3,   4, -15, -15,   1,  -4,   0,  -5,  10,   2, -25, -23,   0,  -7,   1,  -2,  17,  -8, -27, -34,  -2,  -7,   2,  -2,  17, -35, -17, -10,  -7,   0,  -1,  20,  11, -50,  37,   8,  -2,   3,  12,  21, -38, -28,  17, -17,  -2,  -4,  16, -27, -17, -16,  23,  16 },
      {  -1,  -2,  -2,  -1,  -3,  -6, -18, -10,  -1,  -1,   0,  -3,  -3, -12, -23, -11,   0,  -1,   0,  -5, -10, -15,  -8, -12,   0,  -2,   1,  -8, -16,  -7,   0, -26,  -1,  -6,   2, -22, -23,  -5,   3,  -5,  -2,  -1,  -1, -45,  11,   7,  31,  -1, -19,   4, -40, -20,  29, -16,  12, -24,  -7,   8, -32,  36,   5,  19,  26,  38 },
      {  -1,  -1,   0,   1,  -3,   9,   0, -46,  -2,   0,  -1,   1,  -2,  15,   0, -68,  -2,   1,   0,   3,   1,   5,  -1, -47,  -2,   3,   0,   2,   1,  -5,  -6,  24,   0,   2,   0,   7,   3, -12,  23,  28,  -2,   1,   4,   9, -15,   8,  15, -22,  -6,   7,   8,  -6, -15,  30,  -4,  16,   4,  20,   3,   1,   0,  43,  -1,  -2 },
      {   0,  -1,   0,   0,   2,  -2, -18, -16,   0,  -1,   1,   3,   4,  -7, -31, -21,   1,   0,   0,   6,  -2, -13, -28,   9,   3,  -2,   1,   7,  -6, -20, -26,  24,   5,  -4,   4,   8, -26, -15, -20,  -5,   4, -16,  14, -16, -34,  -3, -28, -10, -10,  -7,  15, -45,   5, -19,  -2, -12, -13,   2, -20, -30,  15, -39,  -9, -31 },
      {   0,   0,   0,   1,  -3,   3,  16,  -2,   0,   0,   0,  -1,  -4,   6,  25, -24,   0,   0,  -1,   0,  -5,  13,  20, -38,  -1,  -1,  -3,   1,  -7,  28, -16, -26,   2,  -4,   6,  -4,  10,  33, -44,  42,   0,  -1,  11, -10,  33,   2, -25, -21,  -4,   5,  -1,   1,  18, -32, -21, -14,   2,  -2,  -2,   1,  -1, -21,  21, -26 },
      {   1,   0,   0,  -1,   2,   2,   5,   9,   1,  -1,  -1,   0,   1,  -3,  10,  15,   2,   0,   3,   1,   4,   1,  10,  40,   2,  -1,   3,  -2,   5,  -1,  -5,  22,   4,  -2,   3,  -2,  12, -12,  28,   7,   4,  -2,   1,   2,  -2,  -7,  21, -88,  -3,  -4,  -7,  -8, -12, -12, -17,   3,  -1,  -1,   0,  -2,   5,  -6,  59,  -3 },
      {   0,   0,  -1,  -2,  -3,   2,   2,   2,   0,   0,  -2,   0,  -3,   2,  -2,   6,  -1,   0,  -2,   0,  -1,   3, -15,  32,   0,   1,  -2,   5,  -3,   0, -14,  52,   2,  -2,   1,  -1,   1,   8, -15,  72,   5,  -2,   7,   3,  18,  -5,  28,  14,  -4,   0,   3,   4,  -5,  -2, -18, -39,  -7,   1,  -2,  -4,  -5,  -4, -32,  47 },
      {   0,   1,   0,   0,  -2,   0,  -3,  -8,  -1,   1,  -1,   0,  -5,  -2,  -3,  -3,  -1,   1,  -1,   0,  -3,   1, -11,  -2,  -1,   1,  -2,  -1,  -2,   0, -13, -15,  -3,   1,  -3,   0,  -4,   1, -17,  -9,  -2,   3,  -4,   3, -10,   6, -34,  -1,  -3,  -2,  -6,  -5, -12,  -3, -61,  61,   3,  -1,   5, -10,   8, -12,   9,  77 },
      {   0,  -1,  -1,  -1,  -1,  -1,  -5,   1,   0,  -1,  -1,   0,  -1,  -3,  -2,   0,   0,  -1,  -1,   0,  -3,  -4,  -1,  16,   0,   0,   0,   0,  -4,   1,  -6,  24,  -1,   0,  -1,   0,  -4,  -2,  -5,  45,  -1,  -2,   0,  -3,  -3,  -1,   3,  56,   5,  -3,   5,  -4,   9, -10,  40,  57,   5,  -4,  10,   3,  10,  -1,  69, -13 },
    },
  },
  { // 29
    {
      {-106,  20,  22,   3,   5,   1,   2,   0,  50, -33,  -9,   4,  -1,   0,   0,   0,  13,  12, -11,  -4,  -1,   0,   0,   0,  -2,   0,   4,  -1,  -1,   0,   0,   0,   3,   1,  -2,   1,   0,   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  14, -56,  14,   7,   2,   2,   1,   1,  67,  33, -43,  -8,  -4,  -1,  -2,   0, -44,  40,  28, -11,  -3,  -1,   0,  -1,  -4, -21,  11,  12,   0,   0,   0,   0,  -5,   2,  -4,   0,   4,   0,   0,   0,  -1,  -4,   2,   0,  -1,   1,   0,   0,  -2,   1,   0,   1,   0,  -1,   0,   0,  -1,  -1,   0,   1,   0,   0,   0,   0 },
      {  48,  18,   4, -16,  -2,  -3,   0,  -2,  24, -57,  -4,  24,  -3,   4,   0,   1,   0,  52, -48, -14,   6,  -1,   0,   0, -27,  20,  41, -19,  -5,  -1,   1,  -1,   2, -16,   6,  15,  -3,  -1,   0,   0,  -5,   3,  -1,   0,   3,  -1,   0,   0,   0,  -2,   1,   1,  -1,   0,   0,   0,  -2,   1,   0,   0,   1,   0,   0,   0 },
      {  14,  74, -26, -17,  -3,  -4,  -1,  -1,  43,  -5,  28,  -3, -10,  -2,  -2,  -1, -68, -15,  16,  24,  10,  -1,   1,   0,  14, -26, -12,  -3,   7,   5,   0,   1,   2,   8,  -4,  -4,  -5,   1,   0,   0,   0,  -1,   2,   0,   0,  -1,   0,   0,  -1,  -2,  -1,   2,   0,   0,   0,   0,  -1,  -1,  -1,   0,   1,   0,   0,   0 },
      {   2,  44,  11,  -9, -11,  -4,  -2,  -1, -33,  -6, -23,  10,  17,   4,   2,   0,  11,  22,  33, -39, -17,  -1,  -1,   0,   0, -43,  30,  51,  -5,  -4,  -4,  -2,   3,  -2, -45,  -3,  22,   6,   4,   1,   1,   3,  10, -17,  -9,   2,   0,   1,   2,  -2,  -3,   5,  -4,  -2,   0,   0,   0,   2,   1,   0,   0,  -2,   0,   0 },
      {  -6,  26, -51,   9,   2,   2,   0,   1,  23,  46,  18, -37,  -6,  -6,  -1,  -2,  42,  15,  22,   6, -11,  -5,  -3,  -1, -42,  11,  33,   5,   7,  -3,  -1,  -1,   2, -41,  11,  20,   2,   7,  -2,   0,  -3,   0, -19,   8,   4,   1,   4,   0,  -1,  -3,   0,  -3,   4,   0,   1,   2,  -1,  -1,  -2,   0,   0,   1,   0,   0 },
      { -29,   1, -11,  -4,   8,   3,   2,   2, -36,  37,  -4,  11,  -6,  -1,   0,   0, -36,   6, -14, -27,  29,   0,   5,  -1,  37, -18,  47, -38, -31,   8,   0,   2,   5, -13,   7,  44, -15, -11,  -2,   0,   4,   1, -15,   0,  16,  -1,  -3,   0,   0,  -1,   2,  -1,  -3,   3,   0,  -1,   3,  -1,  -2,   0,   1,  -1,   0,  -1 },
      {  17,  34,  22,  -5, -17,  -1,  -3,  -1,  30,  16, -34,   8,  18,  -6,   0,  -1,   8, -48,   9, -37,  -4,  15,   0,   3,  16,  49,   1,  10, -34, -12,  -1,  -3, -24,   2,  34,  16,  19,  -9,  -4,  -2,   2, -20,  -6,  10,   3,  11,  -1,   1,  -7,   3,  -5,  -1,  -1,  -1,   4,   0,   1,  -2,  -1,   0,   0,  -1,   0,   1 },
      {  -5,  34,  10,   7, -10,  -1,  -1,   0, -29,  13, -50,  -7,  13,   1,   2,   2,  -8,  27,   8,  -5,   0,  -1,  -1,  -3, -20,  -8, -11, -29,  21,   3,   7,   2,  19,  -1,  47, -39, -36,   3,  -3,   2,   4, -11,  15,  44, -16, -12,  -3,   0,   1,   2, -10,   4,  14,  -4,  -3,   0,   0,  -3,   3,   3,   1,   2,   0,  -2 },
      {  -7, -12, -74,  26,   8,   4,   2,   2, -10, -48,  -7, -31,  13,   5,   5,   2, -15,  26,  16, -25, -23,  -1,  -3,   1,  44,  25,  -2,   8,  -8, -10,  -4,  -3,  -8,  19,  17,  -3,   6,  -1,   0,  -2,  -1,  -9,  12,   5,  -1,   3,   0,   1,  -1,  -1,  -2,   4,  -1,   0,   0,  -1,  -1,   0,   1,   0,   0,  -1,   1,   0 },
      {   5,  -1, -41,   0,   8,   5,  -1,   3,  20,   5,  -9,  14,  -8,  -5,   3,  -2,  33, -12,   3, -26,  38,   8,  -2,   0, -32, -19, -15, -13, -38,  28,  10,   2,   6,  33, -16, -13, -16, -41,   9,   3,   0,   2,  37,  -9,   3,  -3, -24,   0,   1,  -2,   0,  19,  -4,   5,   2,  -8,  -4,   0,   2,  -1,   6,  -1,   2,   1 },
      { -16,   0, -26,  -3,  -5,   2,   2,   3,   2,  43,   4,  27,   2,   0,  -4,  -3, -22,  -2, -49, -13,  -1,   0,   4,   3,  -1,  29,  10,   2,  17, -15,  -7,  -5, -10, -15,  -8, -41,  24,  21,   7,   3,  13,   2,  32, -26, -45,   1,  -1,   1,   2,  -5,   8,  29,  -5, -13,  -3,  -2,   1,   1,  -3,   0,   6,  -3,   0,   0 },
      {   0,   2,   8, -52,  16,   1,   1,   0,   1,  17,  22,  20, -44,  -3,  -1,   0,  34,  40,  17,  10,   8, -21,  -6,  -3,  42,   5,   3,  12,   6,   9,  -7,  -2, -33,  18,  20,  -4,   4,  -1,   3,  -2,   2, -29,  17,  21,  -5,   5,  -3,   1,  -7,   2, -21,  14,  11,  -2,   6,  -1,   0,  -3,   0,  -8,   5,   2,   0,   2 },
      {  16,   9,   2,  14,   1, -11,   1,  -3,  23,   2, -25, -17,   2,  21,  -1,   1,  33, -25,  -9,  -4,  -4, -22,  10,   2,  38,  -7,  20, -20,  27,  21, -22,   1, -27,  10, -19,   7, -42,  30,  28,  -4,  -2,   9,   5, -12,  -4, -40,   8,  11,   0,  -2,  10,   3,  -6,   2, -21,   0,  -2,   4,  -1,   7,   2,  -6,   3,  -7 },
      {  -3,  19, -24,   1,   5,   3,  -2,   0, -15,   5, -62,  24, -24,   1,   1,   2,   4,  17,   8,  30,  27, -18,   0,  -3,  -4,  22, -26,  -9,  20,   7,   0,   1,  -6,   5,  -4,   6,  24,   3,   0,  -6, -11,  -9, -17, -42,  22,  20,  13,   4,   3,   3,  20, -29, -26,   0,   2,   4,   2,  -3,   9,   7,  -5,  -1,  -5,   0 },
      {  -7, -25, -27, -45,  -1,  15,   5,   1,   1, -30, -14,  36,   2, -14,   1,   1,  -5, -25,  42, -12,  22,   6,   0,  -1,  -3,  13,  -3,   2,   4,  15,   3,   2,   1, -46, -11,  -6, -14,  20,   9,   2,  11,  18, -25,  15, -24, -20,   9,   5,  -9,   8,  -1,  -5,  25,  -8,  -8,   2,   4,   0,   3,  -8,  -6,  10,  -4,  -2 },
      {  -5,   0,  20, -27,  24, -10,   4,   1, -17, -15, -17, -41, -24,  20,   5,   0, -24, -14,  17,  14, -18, -26,  -6,   5, -15,  30,  12,   5,  -6,  13,  -4,  -7,   3, -30,   1,  -2,  -3, -32,  17,   7,  20,   6,  10, -17,  -2,   0, -40,   6,  -3,   6,   9,  30, -18,   9,   3, -20,   1,  -1,  -8,  10,  17,  -9,   7,   3 },
      {   0, -11, -12, -43,  14,   2,   3,  -1,   5,  -2, -19,  -6, -22,   6,   1,   1,   4, -36,  -2,  -5, -17, -17,   4,   3,  -8,  -4,  26,  -7,   8, -22, -17,  -1,  68,  38,   8,  12,   7,  23, -17, -11, -25,   8,  17,  11,  -2,  18,  22,  -3,   5,  -5,  -5,   7,  -6,  -8,   4,  10,  -6,   4,  -1,   3,   3,  -2,  -3,   0 },
      { -12,   0, -17, -20, -18,  -3,   2,   3,  -6,  21, -11,  19,   2,   6,  -3,   0, -17,   4, -22,  12, -29,  -4,   6,   3, -12,  26,  -1,  31,   3,  -9,  -7,  -7,   3,  21, -16,   6,  -8, -18,  23,   2, -12,   9,   8,  16,  42, -41, -20,   5,   1, -16,  -7, -26,  37,  34, -11,  -7,  -1,  -1,   6, -16, -17,   9,   8,   0 },
      {  -4,  -6,  -9, -73,  -6,   0,  13,  -5,   4,  10,   7, -18,  41,  -6,  -8,   2,  11,  18,  -5,  -6, -25,  28,   4,   0,   3, -18, -20, -19,  -7, -13,   9,   5, -12,   4,  15,  -6,  -8,   0,  -4,   2,   3,  -4,  -7, -22,   3,  -9,  -4,   7,  11,   3,  26, -15, -34,   9, -16,  -4,   3,   4,   4,  30, -10, -13,   4,  -7 },
      {  10,   6, -11,  -8, -12,   5,   0,   4,  14,   3, -34,  -9,  -3,  -9,   5,  -1,  26,  -1, -24,  25, -10,   1,  -9,  -5,  46, -21,  -1,  -6,   0, -20,  14,  -2,   2, -19, -11, -12,   4, -20, -31,  13,  15,  56,  -4,  10,  -2,  19, -21, -25, -14,  -1,  16,   5,  22,  10,  26,  -4,   2,  -3,  -1,  -8,   0,   8,   1,  12 },
      {   4,  12,  27,   3, -14,   6,   0,  -1,   0,  13,  17, -12,   7, -25,  -2,  -2,   3,  33,  35,   1,  16,  13,  -7,  -6,  25,  56,  -9, -15, -11,   7,   8,   3,  41,   5, -23,   6,  -6,   4,   1,  -1, -26,  30,  22, -13,  -4, -12,   9,  -1,   3, -23,  26,  15,   7,  -2, -11,   4,  -8,  -5,  -9,   7,   3,   8,   2,  -1 },
      {   4,  15,   6,  -7,  59,   4,   4,  -9,  -9,   4, -19, -30, -19,   3,  -4,   3,  -5,   6, -18,   5,  -7,  25,   9,  -4, -14,  -3, -18,   2, -31, -17,  14,  11, -22,  10, -11,  13,   9,   5,  -6, -11,  10,  10,   0,  -5, -13, -23,  28,  10,   2,  10,   6,   5,  33, -27, -26,   8,   3, -12,   5, -20,   6,  31,  -4,  -8 },
      {   7,   9,  11,   2,  74,  -8,  -5,  -2,   9,  11,  16,   4,   3,  13,   8,  -3,  -1,  -1,   1, -51,  -4,  -9,   1,   3,   7,  12, -13,  -5,  20,   8,   5,   0,   8, -19,  -9, -26,  -8,   3,  -8,   3,   1,  10,   8,   5,  15,  14,  -2, -12,  -1, -16,   5, -36,  -4,  25,  18,  10,   5,  -1,   6,   1, -36,   0,  -3,   7 },
      {  -9,  -8,   2, -22, -17,  -6,   2,  -1, -12,  -9,  -7, -41,  16, -11,  -1,   4, -13, -11,  -1,  17,  32,   9,  -8,  -8, -20,  15,  38,   2,   5,  21,   9,   1, -42,  17, -17, -11,  -6,  10, -19,  -2,  19,  -6,  33,   5,  13,  15,  26, -19, -13, -26,  -3,  -8,   1,  11,  17,  29,   4,   3,   0,  -6, -20,   1, -12,   3 },
      {   5,   0,   2,  -8,  13,  14,   3,   1,   4,   2, -11, -35,  10, -12,   4,  -1,   4,  -6, -33,  14,  24,  15, -21,  -4,  26,  -3,  20,  27,  -5,  29,  22, -13,  30,  -8,   0, -25,  10,  -4,  25,  20, -43, -29, -20,  -7, -22,  -1, -10,  15,   9,   6, -24, -20,  -2,  -6,  -4, -10,  -2,   2,  11, -10, -15,   5,  -1,   0 },
      {   4,  -3,   7,   7,   2, -10,   7,   0,   0, -13,   5, -16,  -2, -31, -12,   0,   8,  -5,  10,  -7,  41,  -7,  19,  -4,  -1, -15,  16,  -8,  14, -31, -25,   9,  11,   6,  26, -14,  35,  11,   0,  -4,  23,  -3,   1, -18,   4,  -3, -24,  24, -13,  -7,   8,   0,  17,  41, -30, -20,   2, -14,  -2,  -8, -21,  24,  22,  -3 },
      {   0,   4,   7,  -4, -29,  40,  -9,   7,  -3,   0,   5, -29, -39, -32,  23,  -1,  -8,   3, -27, -40,   4, -24, -32,   5,  -7, -10, -26,  24,  -3,  16, -16,  -6,  -3,   2,  20,   7,   1,   9,  18, -14,  22,  34,  -1,   5,   5,  -2,  14,   6,   8,   1,  10, -14,  -6,  -4,  -4,   6,   5,  -1,   5,  16,  -7,  -7,  -2,  -3 },
      {   0,   2,   5,  10,  13,  46,   4,   0,  -6,  -9,  -3,  12, -17, -39,  14,   3,  -8, -18,  15,   0, -21,  -2, -17,   3,  -8,  -3,  24, -25,  19, -20,  13,   8, -42,   0,  -9,  -4,   1,  -9, -23,  11, -42,  -4,   9,   3,  -3,  -6, -17,  -6,  50,  -8,  -1,   3,  -2,   8, -15, -11,  -6,   8,  -7,   7,  -4,   1,   9,  -4 },
      {  -5,   2,   1, -11,   7, -13,   1,   5,  -7,   4,   6, -10,  21,   0,   5, -10,  -8,   7,   1, -13,  14,  -9,  -3,   4, -20,   7,   2,   1,   4,  10, -36, -13, -22,  44,   2,  -1,  -7,  13,  -1,   8, -25,  12, -46,  -4, -24,   6, -16, -27,  12,  11,  32,   1,  24,  -8,  35, -27, -10, -16,   8,  -4,  14,  13,   8,  26 },
      {  -1,  -2,  -1, -10,  16,   0,  -2,   0,   0,   0, -11,   3,  49,  -8,  16,  -7,   0,   2,  -5,  13,  -5, -13, -21,   6,  -3, -22, -31, -17,  -3,  14, -26,  -8, -11, -28,  17,  32,  19,  14,  25, -14, -20,  11,  26,  -9,   8,  16,   3,   8,   7, -39, -29,  21,   7,   3,   9,  -1, -10,   2, -16, -29,   4,   9,   6,   8 },
      {  -4,  -4,  -7,  -4, -17, -58,  10, -15,  -4,  -3,  -6, -17, -23, -11, -42,  -3,  -4,  -3, -11, -25,  15,   8, -11,   5,   1,   0, -25,   8,  16,   2,  19,   7,  -9, -24,  -8,  22,   0,  21, -13,  -6, -34,   6,  27,  11,   8,   4,  -8,  -8,  23,  39, -10,   2,   0,   6,   5, -15,   7,  -1,  -2,  -5,   1,  -5,  18,   0 },
      {  -4,  -4,   2, -10, -27,  -4, -22,  13,  -3,  -4,   1, -23, -30,  40,  23,   3,  -1,   4,   0, -10,  23,   5,  51,   4,   0,  -5, -17,   1, -20, -25, -19,  12, -18, -26,   0, -17,  -3,  -8,  -9,  -3, -46,   6,  -7,   0, -12,   2,  -1,   2,   1, -38,  -9,  -2,  -1,  -2,  -2,  -1, -17,  17,   9,  -2, -13,  -6,  -8,  -3 },
      {  -3,  -2,  -4,  11,  24, -22,  -4,   0,  -7,  -5,  -6,  24, -21,  -3, -16,   1,  -2,  -6,   3,  11,   2,  47, -14, -12,  -3, -11,  17,  41,   5,   1,  12,  -4, -12,   9,  50,   0, -17,   7,  18,   8, -15,  38,  -6,  -7,  -1,  -1,   4,   3, -12, -39,  15,  11, -12,   3,   1,   1, -16,  -6, -21,  20,  -2, -18,   4,  -1 },
      {  -3,  -9,  -1,  -7, -16, -40,  -9,   2,   4,  -3,  13,   4,   9,  13,  22,  -6,   0, -10,   7, -15,   1, -31, -13,   0,   4, -11,  12,  -3,  19,  -2,  33,  -4,  -4,  -2,  29,  -7,  15, -30,   8,  19, -12,   8,   8, -18,  14,  -4,   4,   1,  19,  -4,  23,  -9,  16, -29, -21,  27,   5, -15,   6,  -3,  15,  40, -35,  -8 },
      {  -3,   3,   2,   0,  -2, -11,   4,  -8,  -1,   8,   7,   2,   7,  12,   1,  -3,  -4,   6,  13,   0,  -2,  -9, -21,   0,  -9,  15,  11, -22,  -2,  -6,   6,  19, -16,  29,  -3,  -4,   9,  -1,   2,  10,  -5,  60,  -9,   1, -17,  17, -13,  24, -25,   6, -55, -25, -20, -16, -16, -18,  17,  -3,  23, -12,  -7, -13,   1, -26 },
      {  -2,   0,   3,   2, -13,  25,   9,  -7,  -5,  -4,   4,  -3, -16,  12, -18,  12,  -3,  -4,   3, -13, -21,  -7,  -3,  15, -11, -15,  -6, -24,   0,   9,  26, -21,  -9,   3,   1,  10,  16,  11,  12,  12, -39, -14,  10,   0, -13,  -1,  -6,   4, -75,  -6,  32,  -3,   8,   4,   6,  -4,  31,  -2, -18,  -6, -13,   8,  -2,  10 },
      {   1,  -2,   2,   6, -19,   0, -14,  10,   2,  -4,   3,   6, -23,  42, -10, -19,   4,  -1,  20,  17, -17,  23, -26,   5,   1,  -5,  15, -27, -34,   3,  -2, -10,   3,  -2,  11,  -7,  12,  18,  23, -10,  16,   2,  19, -12, -20,  -1,   6,  -6,  44,   9,   2, -20,   7,  14,  15,  -3,   9, -21,   9,   0, -41,  15,  11,  20 },
      {   1,   2,   4,  -3,   4,  61,   3,  14,   1,  -1,   7,  -3,  27,  32, -33, -18,  -1,  -1,   2,   2,  24, -23,  26, -12,   0,   2,   6,  28,   7, -12,  19,  24,  -5,  -5,  27,   6, -17,   3,  -1,  -2, -15,  21,  25, -21,  12,  -3,   3,  -2,   6,  30,  -9,  -2,   0,   5,  14,   0,   5, -26,  -6,  -3,  12,  -2,  -1,  11 },
      {   2,   2,   2,  -6, -18,  13,  24,  14,   0,  -1,  -1, -13, -21,  14, -16,   6,   1,  -1,   1, -33, -12,  22,   2, -37,  -2,   5, -17, -17,  43,   1,  22,  -1,   8,  14,  -4,  22,  -4,  -5,   7,  25,  28, -10, -26, -31,  -2,   2,  -5,  -6,  12, -29, -31,  16,   5,  -3,  13,   2, -19,   0,  -1, -31,  -3,   6,   0,  11 },
      {  -3,  -1,   1,  -8,   4,   2, -56, -10,  -3,   1,   0,  -8,   0, -12,  31, -11,  -1,   3,   4,  -3,   3,  26,  13,  16,  -2,   1,   4, -19,  24, -17,  14, -14,  -4,   8,  -5,   7,  13,   0,  30,  10,  -2,   8,   5,  -4,   7,  -6,  -1,   7, -19,  52,  -7,   9,   4,  14,  13,  26, -48,  -9,  -7,  -9,  -5, -25,  -2,   5 },
      {   0,   2,   1,   6,  -5,  -5,  68,   2,  -1,   3,   3,   4,  -1,  15,   4,  10,  -2,  -2,  -2,   0,   1, -26, -24,  -5,  -3,  -3,  -4,  -4, -24,  -3,   0,   3,  -6,  -8,  -5, -13,   6,   1,  -6,  -1,  -8,  -1,  -4,   8,  -3,  -3,   9,   3, -14,  14,   3,  -2,  -1,  12,  -6,  19, -80, -30,  -9,   8, -15,  -6,   2,  -5 },
      {  -3,  -2,  -6,  -6,  -6,  12, -51,  -6,   1,   5,   0,   5,   8,  37,   7,  17,  -3,   0,  -8,  -2,   1,   1, -31,  -1,  -1,   5,  -7,  17,   8,  21,   3,  13,   3,   4,  -3,  16, -16,  12, -38,   5,   3, -11,  -7,   0,  -9,  24, -11,  -4,  -8,   4,   7,  17,  -6,  21, -34,  -8, -20,  -8, -15,   2, -11,  43,  23, -32 },
      {  -1,   0,   1,   2, -10,  13,   2, -54,  -1,   2,   4,   2,  -5,  -9, -32,  26,  -2,  -1,   2,  -5, -27,  10,  31,   9,   5,  -1,  14,  -4,   4,  46, -18,   3,  -2,  -3,  17, -27,   5, -28,   2, -28,  -8,  22,  -6,  -9,   5,  19,  16,   9,  10,  12,   6,   3,  -5,  10,   6,   4, -21,  12,  16, -28,   4,  24,  -5,  -2 },
      {  -1,  -2,   1,   4,   8,   6, -16, -25,   1,  -1,   4,   2,  10,   7,  -8,  35,  -1,  -4,   3,   7,   8, -17,  11, -41,  -2,  -3,  -2,   1,  -3,   4,  16, -20,   0,  -3,  10,  29,  27,  12,  -1,  18,  12,   5,  29,  13, -15, -17, -11, -11,   2,  -9,   9, -18,  10, -11,   3, -21, -31,  29,  53,  31,   0,  -7,   7,   9 },
      {   0,   0,   1,   4,  -5,  -1,  29, -63,   1,   2,   2,   5,  -3,   5,  50,  13,   1,   3,   5,   5,  12,   2, -11,  17,  -1,   2,  -1,  18,   1, -24,  -2,   1,   8,  -1,  21,  15, -35,  -5, -10,   1,   1,  -6,   9, -46, -16, -12,   2, -25,  -6,  11, -10,   3,  12,   5,   7,   2,  20,  15,   5, -12, -23,  -7,   7,   6 },
      {   1,   2,  -3,  -1,   0,   5,   6, -36,   1,   2,  -4,  -3,   7,  23,  -8, -10,   3,   7,   0,   2,  20, -25,  17,  24,  -2,   3,  -6,   3,  -9,  12,  28,  -5,   0,  20,   2,   2,  34,  11,  -1,  24,  15,  16, -26,  29, -10, -10,  -4,   6,  39, -12,   8,  14,  -1,  14,  -8,  16,   9,  28, -51,  -9, -13, -23,  -1,   0 },
      {  -1,  -4,  -3,  -6,  -6,  11,  -4, -50,   3,   0,   4,  -4,  -1,  26,  18,  19,   1,  -4,   0,  -9,  17,  12, -23, -36,   2,  -1,   4,  -8,   0, -28,   3,  -4,  -2,  -7, -19, -30,   6,  11,  22, -12,  -2,  -6,  -9,  25,  41,   4,  10,  10,   6, -12,  10,   4, -10,  -9,   6, -23,  15, -34,  -9,  -7,  34,  -5,  17,  10 },
      {  -1,  -4,  -4,   0,  -8,   0,  -1, -35,   0,  -4,  -4,   0,  -4, -16,  16, -46,   0,  -3,  -8,   4, -14,   0,  36,   8,   2,  -1,   0,   6,  13,  26,   8,  22,   4,  -2,  -6,  22,   5,  -9, -11,  -3,  11, -12,  12,  12, -20,   1,  -9,  -1,  -3, -29,   2, -19,  -7, -26,   2, -22, -11, -68,  -3,   8, -15,  -8,   7,   0 },
      {   1,   3,   1,   2,   2,   8,  41,  20,   1,   3,   1,   1,   5,  26,  34,  28,   1,   2,  -1,   4,  15,  44,  15,   1,   1,   2,   0,   7,  39,  15, -25,  -7,  -2,   1,  -4,  30,  19, -22,   6,   1,  -4,  13,  21,  34, -13,  20,  -6,  19,  14,  21,  15,   2,  -2,   7, -10,   9,  23, -11,   2,  14,  -5,   6,  -7,   2 },
      {  -1,  -1,  -1,   0,  -2, -18,  -1,  19,   1,   2,   2,   5,   0,  -6,  44,  -4,  -3,  -2,  -4,   0, -22, -11,  15, -59,  -2,   0,   3,  -2, -10,  36,  22,  41,   1,   5,  12,  -9,  23,   7,  -7, -15, -16,   5, -11,  -5,  -1, -15,  18,  -4,  -6,  18,   5,  24,   3,  26,  -1,   2,  22,   6,   3,  -8,  -8, -17,  29,  15 },
      {   0,   1,   2,  -1,  -2,   5,  23,  12,   0,   1,   2,   0,  -2,  17,  24, -41,   0,   0,   3,  -2,   6,  17,  -6,  30,   3,   1,  11,   4,  15,   3,  31,  17,   0,   4,   4,  -4,   4,  12,   3, -30,  13,   0,   8,   8,  17,   4,  -1,   4, -14,  -1,  16, -16,  -8, -17,  -4, -35, -30,  57,   2, -11,  18,  25,  22,  22 },
      {   0,   0,   0,  -4,   1,  -4, -14,   2,   1,   0,   1,  -1,  11,   2, -21,  -2,   0,  -2,   0,  -2,   7, -14, -13, -35,   0,   0,   4,  -1,  17, -16, -16,  -5,  -2,   1,  -6,   3,  -6, -55,  10, -29,  -2,  -3,   3,   2, -17,  10,  15,   3,  -2,  10,  15, -15,  15, -44,  -6,  -6,  -5,  23, -43,  -7, -44, -18,  31,   1 },
      {   0,  -1,  -1,  -2,   2,   9,   1, -11,   0,  -2,  -1,  -5,   2,  24, -14, -36,   0,  -2,   0,  -8,  11,  13, -24,  14,   3,  -3,   0, -11,  24,  13,  -8, -18,   0,  -3,   0,  -6,  34, -31, -23, -27,  -9,   1, -16, -11,   1, -31,  21, -17, -12, -11, -27,  16,  30,  17,  -4,   8,   8,  -1,  18,  49,  19, -14,  13, -11 },
      {   1,  -1,   2,  -2,   1,  -4,  -9,  33,   3,   0,   4,   0,   5,   5,   6,  47,   3,   0,   5,  -1,   6,  -1,  -4,  34,   3,  -2,   5,   2,   7,  10,  22, -29,   5,   1,  12,  -1,   7,  -2, -37, -34,   7,   2,   2, -11,  -4, -30,   2,   4,  10, -11,  11, -12,  -4,  -7,   2, -20,  -3, -27,   4, -49,  10, -44,  12,  -9 },
      {   1,   0,   3,   2,   2,  11,   8,  -4,  -1,  -3,   1,   1,   1,   9, -11, -42,   0,  -2,   2,   2,   3,   4, -14,  -6,   0,  -4,  -1,   3,   3,  -3, -10,  -1,   0,  -5,   3,   3,  -3,   3,   7,  13,  -5,  -8,   0,  13,  16,  19,  -3, -11,   1,  -1,  44,  35,   5,  -3, -21,   2,   2,   4,  57, -45, -22, -40,  -9, -38 },
      {   0,   0,   0,   0,   2,   3,  -9,  -9,  -1,  -1,  -1,  -3,  -4,   7,  -9,  -4,   0,   0,   0,  -1,   4,  16, -31, -16,   1,  -3,  -1,  -2,   6,  -1, -33,  53,   7,  -2,   6,   8,  24,  -8,  -2,  -1,  -1,  -9,  10,   7, -24, -43, -21, -24,  -8,  -1,  15, -34, -21,  14,  13,  10,  -6,  11, -19, -30,  27,  -5, -39,  -8 },
      {   0,  -1,  -1,  -1,   1,   0, -10,   1,   1,   1,  -2,  -2,   8,   0, -15,  39,   0,  -1,  -4,   1,   3,  -6, -17,  35,   2,   1,   2,   0,   9,  -8,  21,  67,   3,   5,  -7,   5,   3,   0,  31, -33,   1,  -9, -17,   4,  -1,  16, -21,  -4,   2, -15,   4,  24,  24,  -9,  -4,  13,   0,  -8,  24,  22, -18, -17, -12,  31 },
      {  -1,  -1,  -1,   1,   0,  -9,  -7,  -5,   0,   0,   0,   2,   4,  -6,  -5, -17,  -1,  -1,  -3,  -1,   5, -19, -12,  -3,  -1,   1,  -5,   8,   5, -29,  23,  -8,  -4,   3, -10,  20, -22, -29,   5, -14,  -1,  -5,  -6,  16, -51,   6,  35,  41,   4, -15,  13,   0, -24,  55,   9,   3,   1,  -2,  27, -12,  19,  10,   0,  21 },
      {   0,   0,   0,   2,   1,   2,  -1,  -2,  -1,   0,   0,   3,   4,  -3,  -4,  -5,   0,   1,   0,   3,   9,  -3,  -5,  -3,   0,   1,  -1,   6,   6, -10,  -2, -13,  -1,   2,   0,   2,   8,   0, -17, -20,  -6,   8, -12,  17,  13, -50, -25,  10,  -6,   2,  -5,  39, -57, -29,  29,  23,   5,  -3,  18,  -4, -44,  40,   3,  12 },
      {   0,   0,   0,   2,  -3,  -1,  -6, -10,   0,   1,   1,   2,  -1,   7,   3, -12,  -1,  -1,  -2,  -2,  -5,   6, -18, -15,  -1,   1,   0,   0,   0,  11,  -8,   3,  -1,   0,  -1,  -1,  -8,  13, -46,  -1,   0,  -5,   4, -15,  13,  -2, -12,  80,   3,   5,   6,   8,  34,  -5,  19, -19,  -5,  11,  -7,  10, -14, -19, -52,  20 },
      {   1,   0,   1,   1,  -1,   2,  10,  -4,   0,   0,   0,  -1,   0,  -1,   9,  -7,   0,   0,   0,   1,   0,  -1,  13, -25,   0,   0,   1,  -1,  -2,  -5,  27, -20,  -2,   7,  -3,   4,  -7,  19,  16, -64,  -2,  -3,  -6,  -5, -16,  16, -50,   3,   8,  -4,   4, -10,  17,   9,  29,  23,   3,   5,  -7,   9,  10,   6,  -9, -62 },
      {   0,  -1,   0,   0,  -2,   2,   1,  -3,   0,  -2,   0,  -1,  -3,   3,  -4,  -1,   0,  -2,   1,  -1,  -2,   3,  -6,  11,   0,  -1,   1,   0,  -3,   5, -14,  20,   1,  -3,   2,  -2,  -3,   2, -12,  30,   0,  -6,   6,  -6,  -1,   0, -17,  42,   2, -10,  13, -20,  14, -18,  26,  62,   4,   4,  15,  -4,  26,  -6,  74,  -3 },
      {  -1,   0,  -1,   0,  -1,   2,  -3,  -9,  -1,   0,   0,   1,  -3,   1,  -5, -15,  -1,   1,   0,   0,  -2,   4,   0, -19,  -1,   1,   0,   3,  -5,   6,  -5, -31,  -2,   0,   0,  -2, -10,  11, -17, -26,  -4,   4,  -3,  -2,  -7,  15, -42, -12,  -5,   0,  -8, -15,  -7,  -2, -66,  42,   1,  -4,   3, -13,  15, -14,  -5,  64 },
    },
    {
      {-107,  32,  15,   3,   3,   1,   1,   0,  27, -42,   6,   3,   2,   0,   0,   0,  28,  -9, -12,   3,  -1,   2,  -1,   1,   2,   5,  -6,   2,   0,   0,   0,   0,   2,   4,  -6,  -2,   2,  -1,   1,   0,  -1,   1,   2,  -4,   1,   1,  -1,   0,   3,   0,  -1,  -1,  -1,   1,   0,   0,   0,   1,   1,  -1,  -1,   1,   0,   0 },
      { -16,  25, -16,   1,   2,   3,   0,   0, -96,  29,  18,   2,   4,   1,   1,  -1,  30, -56,   3,   2,   5,   4,   0,   1,  12,   5, -16,  -3,  -1,   4,   0,   0,  -8,   1,   1,  -2,  -1,   1,   1,   0,   9,  -1,  -4,  -6,  -1,   2,   1,   0,  -2,   1,   5,  -2,  -4,  -1,   1,   0,   5,   0,  -1,   1,   0,  -1,  -1,   1 },
      {   0, -17,  17,   0,   1,  -3,   1,   0,  14,  -2,   2,  -2,   1,   0,   1,   0, -30, -37,  41,   8,   0,   2,  -1,   1,  39, -49, -36,  34,  -1,   3,  -1,   2,   0,  40, -49, -11,  11,  -4,   2,   0,  -5,   9,  18, -20,   1,  -1,  -3,   1,   5,   2,   1,   2,  -5,   3,  -3,  -1,  -4,   2,   3,   1,   2,  -1,   0,  -2 },
      { -40,  32,  -5,  -1,   0,   2,   0,   0, -11,  54, -17,  -4,  -3,  -1,  -1,   0, -42,  30,  34, -19,   5,  -5,   3,  -2,   1, -49,  43,   3,  -3,   3,   0,   1,  13, -24,  -8,  22,  -6,   5,  -2,   0,   4,   0, -18,   4,   4,  -1,   4,  -1,   0,  -1,   0,  -5,   2,  -2,   1,   1,   4,  -2,  -2,   0,  -2,   1,  -1,   1 },
      {  22,   0,   3,  -2,   8,   1,  -2,  -1, -13, -46,  11,  21,   3,   6,   0,   1,   6, -12, -19,   9,   4,   3,  -2,   1, -18, -18,  34,  11,   3,   5,   0,  -1,  -7, -34, -50,  42,  12,   3,   4,   2, -11,  14, -34, -38,  24,   5,   1,   1,   9,  -2,   7, -25, -15,   9,   1,   1,   3,   7,   0,   0,  -9,   1,   0,   0 },
      {   0,   9,   4,   8,  -1,   0,  -3,  -1,  27,   2,  15,  -6,   3,  -2,   1,  -1, -44, -50,  11,  35,  -2,  -1,   1,   2,  10,  13, -27, -10,   9,  -3,   0,  -1,   3, -24,  37,  16,  -5,   0,  -1,   0,  -5, -17, -43,  29,  25,   2,   0,   0,   5,   2, -16, -39,  11,  10,   1,   1,   5,  -2,   4,  -5, -19,   6,   0,   0 },
      {  19,  11,  16,  -8, -15, -10,   1,   0,   2, -11,  45, -28,  -1,  -7,   2,   0,  56,  13,  33, -12, -17, -10,   0,  -2,  29,  -8,  15,  23, -13,  -7,  -4,   2,  25,   7,   0,  23,  -8,  -2,  -7,  -1, -16,  23, -17,  41,   8, -12,   0,  -3,  -8,   0,  -8,  -6,  28,  -1,  -2,   0, -10,  -4,   3,  -4,   1,   4,   0,  -1 },
      {  35,  97, -27, -14,  -6,  -5,  -3,  -1,  24, -11,  26,  -4,   1,  -1,   1,  -1, -13, -13,   8,   7,   0,   2,  -1,   0, -31,  -4,  -3,   6,   5,   0,   0,   0,   5,  -3,  -6, -18,   4,  -1,   2,   1,   2,   0,  24,  -6, -11,   1,  -4,   0,   1,  -7,   1,  20,  -5,  -2,   0,   0,  -4,   2,  -5,   3,   7,  -3,   1,   0 },
      {   3,   0,  22,  -3,   1,  -2,  -2,  -1, -41, -60,  14,  17,   2,   5,   1,   2, -44,   5, -15,  -2,   0,  -1,   0,   1,  39, -23,  21, -19,  -2,  -4,   1,  -2,  17, -16,  16,  -9,  -9,   0,  -2,   3,   3, -15,  19,  26, -18,   2,  -7,   2,  14,   1, -33,  29,   5,  -2,   3,  -2,  -2,   7,  -5,  -6,  10,  -1,   2,  -1 },
      {  -8,   1, -13,   1,  -5,  -4,   4,   1,   6,  12,  44, -24, -11,  -5,  -2,   1, -36,  14, -27,  12,  -1,  -2,   3,   0,  49,  32,   4, -32,  -9, -10,  -2,   0,  10,  12,  -2,  22,  -4,  -1,  -3,  -1,  -2,  32,  -4, -30,   5,  -9,   3,  -1,   2,  -9,  46,  12, -15,  -3,  -8,   0,  -6,  -2, -10,  21,  12,  -6,   0,  -2 },
      {  15,  37, -15,   2,  -1,  -3,  -2,  -1,  11,  -2, -39,  18,   6,   3,  -2,   1,  11,   4,  -7, -24,   8,  -1,   0,   0,  65,  14,  19,   0, -13,   1,  -3,  -1, -49,  28,  10,  23,  13,  -5,   2,  -2, -15, -28,  13,  -5,  27,   6,   0,   2,   1, -10, -12, -15,   1,  15,   0,   3,  -2,   0,   2,  -9,  -7,   3,   0,   0 },
      {   1, -20,  10,  -9,   6,   3,  -1,   0,  33,   9,   8, -17,  -3,  -2,   0,  -1,  19, -48,  29, -13,   3,   5,  -3,   2,   9,   7,  33,  -3, -19,   6,  -1,   0, -20, -32,  13,  25,   1,   1,   3,   2,  14, -27,  -2, -12, -23,  13,  -5,   5,  -2,  19,  -9,  44, -32, -11,   3,  -2,  -2,   8, -14,  13,  24, -14,   3,  -2 },
      {  -1,  -1, -27,  -9,   8,   6,   1,   1,  14,  -4, -25,  22,  -4,   3,  -2,   0, -16, -41, -20, -14,   9,  10,   1,   1,  -6,   6,   8,  -8, -20,   5,   0,  -2,  -4,  15, -20,  22, -26,  -1,   4,   2,   8,  42, -12,  23, -16, -24,   4,  -3, -22,  18,   3,  15,  47, -23,   3,  -4,   6,  -6,   4, -30,  24,  12,  -5,   5 },
      {  13,   9, -11,   0,  -5,   3,   0,   1,   9, -19, -36, -12,  12,   4,   3,   0,  19, -17,  -3, -24,   3,   6,   2,  -1,   5, -38,  -7, -44,   0,  10,   4,   3,  35,  19,  15,  -6, -44,   7,  -4,   2, -15,   9, -13,   4,  -4,   1,   8,   0,   4,  12,  15, -22, -39,  -4,   5,   2,   0,  12,  10,  35, -13, -23,   0,  -2 },
      {  -2, -14, -44,  28,   4,  -1,   0,   2,  24,  30,  17,   5,  -4,  -6,  -5,  -1,  36,  -5, -37,  37,   8,  -4,   1,   1,  19, -55,   2, -20,  17,   1,   3,   0,  -1, -23,  -3,  -8,  -1,   1,  -2,   2, -24,   0,  14,   4,   2,   1,  -4,  -2,  17, -17, -20,  14,  10,   4,   3,   0,  -4,   7, -11, -18,   8,   4,   0,   1 },
      {   7,  13, -19,  22, -11,   0,   0,   1,   5, -13,  -9,  -3,   6,  -1,   3,  -1,   7,  24,   0,  -5,  -7,  -6,   4,  -2,  43,  -2, -24,  15,  11,  -7,   0,   2, -27, -44,  -8, -35,  15,  14,  -2,   2,  28,  12, -42,  -7, -39,  -4,   3,   0,  -4,  37,   7,  -8,   1, -25,  -2,  -4,   6,  -1,  18,   2,  -1,  -5,  -2,  -1 },
      {  15,  27,  90, -28,  -1,  -9,  -1,  -2,  11,  44, -11,  26,  -9,   3,  -2,  -1,  14,  -9, -44,  13,   6,  -3,   3,   0,  15,  -5,  -3,  -8,   9,   0,   0,   0,   6,  -9,  -6,  -8,   3,   6,   0,   1,  -3,  13,  -5,   0,  -9,   2,   0,  -1,  -7,   5,   2,   2,  -1,  -8,   4,  -1,   0,  -3,   3,   0,   0,  -6,   1,   2 },
      {   3,  16,  -1,  23, -10,   3,  -4,  -1, -14,  -2, -27,   5, -24,   1,   3,   2,   4,  19,  -5,  35, -12, -16,   3,  -2, -11,  -2, -28,  11, -20,  -7,   4,   1,  22,  20,   4,  36,  12, -25,   0,  -6, -15, -22, -22,  -5,  17,   0,   2,   0,  19,  30, -14,  18, -11, -36,   2,  -1,  -3,   4,   3,   3,  49, -14,  -4,  -3 },
      {   1,  -1,  -9,  12,  -9,   1,  -4,  -1,   0,   0, -55,   1,   3,  -1,   1,   0,   4, -28,  14,  -1,  -9, -10,  -2,   1,  19,  36,  16,  14,  11, -21,  -3,  -2,  63, -36,  -3, -12,  31,   0,  -6,  -1, -20,  25,  17,  -4,   1,   7, -16,  -2,  19, -17,  11,   9,   8,  10,  -5,  -2,  -4,  -3,  -7,   3,  -1,   5,   0,   0 },
      {   2,   6,  23,  14,  -5,  -2,  -1,  -1, -10, -12, -31, -19,   4,   8,  -1,   4,   2,  -3,  16,  29, -31,  -7,  -2,   4,  -8,  -9,   0,   1,  19,  -1,   2,   1, -58, -18,   6,   8, -32,   8,   2,  -1,  -3,  17,  16,  16,   9, -36,   7,  -1,  23, -22,  31,  13,  16,   6, -17,   4,   0,   6, -18,  28,  17,   0,  -8,  -5 },
      {   6,   3,  10, -23,  28, -10,   3,  -2,  -4, -18, -11, -28,   8,  -2,   2,  -3,  11,   4,  20,   6,  31,  -8,  -7,  -2,  -1, -24,  -3, -36,  10,   0,  -2,   0,  -2,   1,  28,  -4,  19, -12,  -6,   0,   1,  -1,  -8, -28,   5,  14, -11,   2,  12,  15,  30,   1,  19,   2,  -8,  -7,  -2, -35,  12, -23,  26,  53,  -6,   7 },
      {   6,   2,   5,   6,  -7,  -1,  -2,  -1,   5,   4,  -7, -27,  -5,  -3,   8,  -2,   5,  -8, -30, -22,  -7, -23,  11,  -4,  -2, -14, -14,  -3, -34,  -3,  -1,   2,  12, -18, -10,   1,  -6, -41,   7,  -1,  72,  -5,  13,  -7,  32, -27, -18,   0,  18,   7,  -7,   1,   7,  24, -13,  -3,  -1,   6,   1,  -3, -19,   5,   7,  -4 },
      {   8,   6,  17,  64, -14,   1,  -2,   4,   6,  -8,   8,  54,   3,  -5,  -3,   1,  10,   7,  37,   1,  34, -12,   4,  -8,   0,   2,  -5, -22,   6,  14,  -6,   1,   7,   5,   4,  -1, -10,  -5,  -3,  -2,  33,  15,  -9,   0,  16,  11,  -4,  -2, -30, -13,  13,  14,  -2,  14,   3,  -5,  -1,  -8, -30,   0,  10,  -4,  14,   4 },
      {  -4,  -3, -25, -51,  22,  -8,   1,  -2,  -4,  -7, -20,  -6,  -9,  17,  -4,   1,   6,  13,   4,  43,  -7, -10,  -6,  -3,  18,   1,   4,   9,  34,  10,  -6,  -3,   3,   2,  -4,   3,  -7, -13,  -2,   2,  35,  18,   0,  19,  21,  21, -22,  -3, -44,   7,  -8,   2, -16,  21,  14,  -7,   6,   4,  -6,   6,  13, -25,   9,   1 },
      {   5,   4,   9,  22, -10,   9,  -1,   4,  -3, -10, -14, -46,  22,  -2,   4,   2,  -6,  -8, -31,  -7,  -7,   4,   4,   3,  -2, -19, -10,  19,   9,  -9,   9,   3,   9,   3,  11,  26,  18,  37, -13, -12,  20,   2,  12,  -3,   9,  30,  42,  -4, -42,  -8,  -2,  10,  -2,  -6,  13,  27,   3, -10, -27, -17,   1,   7,  -7,   1 },
      {   2,  -6, -16, -40,  -7,  25,   0,   1,   7, -10,  -1,  45, -46,  -7,  -3,   2,  15,  -1,  12,  15, -15,  -1,  11,  -2,   7, -21,  -5,   2, -20, -19,   3,   4,   7,   7,  21,   9,  13,  19, -14,  -9,  33, -11,  -3,   0, -11,   2,  26,  -9,  19, -21,  24,   3,   7, -11,   1,  11,  14,  -9, -17,  16, -28,  16,  -6,   3 },
      {  -4,  -7,  -3,  15, -15, -11,   5,   2,   2,   3,  14,  20,  14,  -9,  -6,   0,  -9, -21, -21,  -8,  -9,  -1,  -4,   2, -18, -29,  34,  23,   1, -37,  -6,  -1, -19,  21,  41, -10,  31, -14, -23,  -3,   9,  25,   5,  12,  11,  16, -21,  -7,   2,  21,  19, -11,   2,   7,  -1,  -3,  -5,  -4,  29,  27,  18, -21,  -4,   0 },
      {   1,  -3,  -6,  -3,   1, -15,   5,  -2,   5,   3,   8,  16,   7, -14,  -6,   7,   6,  -6,   4,   5,  18, -11,   3,   6,   5,  10,  21,   0,  47,  -6,   6,  -3,  26,  12,  -8,   6,  -2,  34,  16,  -5,  24, -33,  18,  -6,  11, -36,  28,  16,   2,  43,   2, -11,  14,   8, -34,  21,  -9,  16,  11,   8,  16,  -3, -20,  -9 },
      {   4,   8,   9,  36,   5,   1,  -6,  -4,   0,   2,  -5, -28, -36,  19,  -3,  -4,   2, -11,   6,  25, -20, -17,  -2,  -2,  -6,  13,  47, -23,  14,  -1, -15,   5,  -6,  45, -17,  -9,   0,  20,  -7,   3,  26,  11,  -7,  -2, -18,  17,   1,  -6,  28,  10, -23,  -5,  -4,  -7,   7,  -5,   0,  -1,  13, -17, -26,  11,  12,   3 },
      {   0,   0,   4,   2,   4,   1,  -6,   4,   4,   6,  12,  12,  -4,   1,  -6,  -5,   2,  10,  22, -10,  -3,   9, -21,  10,   3,  14, -20, -22,  -3,   3,  24,   2,  -7, -28, -11,  18,  -9,   1, -22,  -5,  12,  30,  50,   5,  20,  25,  20, -17,  31,  24, -10, -14,  -3,   0,  38,  20,  11,   3,  34,  -6,  12,  -4, -20,   9 },
      {   1,  -4,  -4,  -8,  39, -11,   0,  -5,   6,   2,   8,   2,  35,   3,  11, -15,  -1,  -3,  -5, -25,  16, -44,  14, -14,  -2,   6, -12,   9,  19,   6, -26,  -1,   2,   9,  -2,  10,   6,  21,  -1, -12,  19,  -1, -12,  10,   5,   0,   1, -14,  40, -47,  -6,   2,   3, -38,  15,  -7,  11,  -2,   7,   8,  17, -25,   0,   7 },
      {  -6,  -7,  -1, -19, -41,   0,   2,   0,  -5, -10,  -3, -15, -29, -26,  12,   6,  -5, -12,   6, -16,  18, -49,  25,   5, -12,   2,  -3, -38,  18, -22,   6,   0, -23,  -8, -33, -10,  13,  -3, -32,  -5, -26, -18,   2,   9,  13,  -3,  -4, -12, -23,   5,   3,   3,  10,  -7,  13,  -1,   6,   0,  -4,  -5,  -4, -22,   5,   2 },
      {  -2,   2,  -6, -13, -14,   3,   0,   0,   0,  10,  -4,   8,  31,  -7,   3,  -4,   0,   9, -10,  -8, -10,  22,  -9,   2,  -5,   6, -22,  -8,  13,  -7,  -5,  -3,  -1,  12, -29,  22,  -1,  24, -37,   6,   9, -17, -23,  21,  -9,  17, -16, -12,  33,  22,   5,  22,   7,  36,   8, -17, -42,   6, -21,  14,   3,  22,  25,   1 },
      {  -4,   0,  -4, -17,   8, -17,   9,  -4,  -7,  -2,  -8,   4,  -7, -16,  -1,  -3,  -3,   5,   9,  -3,  23, -21,  -5,  -3, -11,  -3, -19,  -2,  -9,  -8,  -3,  17, -17,  -1,  30,  12,  13,  18,  17, -10, -12,  57,   0, -12,  -1,  -1,  13,   9,  17,  23, -34,  13,  -7,   8, -25,   9, -35,  11, -23, -15, -30, -22,  21, -14 },
      {   1,   2,  -5,  19,  36,  -9,  -7,  -5,   1,   0,  13,   8, -54,  41, -18,   4,  -5,  -6, -13, -46, -21, -23,   8,   7,  -2,  -8, -25,  12,  19,  10, -17,   2,   2, -14,   4,  12,  -2,  17,  10,  -6, -24,  -9,  11,  14,   7,  11, -10,   1, -10,  24,  20,   9,  -1,  15,  -9,  -6,  -3, -20,  -3,  14,  -2,  14,   8,  -1 },
      {  -2,  -3,  -1, -12, -31,   6,  11,   9,   3,   4,   2,   5, -17,  35,   0,  11,   2,  -3,   6, -28, -15,  19,   2,  -8,  -4,  -1, -15, -26,  22, -22,  -4,  14,  -4,  -1,  -4,   1,  10,   3,   2,  12,  17,  14, -16,  -5,   6,  10,  -4,  21,  -5, -44, -34, -10, -15,  -3, -35,   4, -20,  11,  16,  -7,  40,  14, -23, -16 },
      {   5,   8,   7,  15,  76,  -3,   3,  -5,   0,   2,  -1,   6,  10, -11,   2,   9,   5,   3,  16,  -2, -38,  20,  10,  -1,  -3,  -3,  -6, -36,  -7, -49,  10,  -5,   1,   0, -12,   2,  22, -11, -16,   5,   1,  -8,  -2,  -2,  -1, -12,  -2,   3, -27,  -4,   1,   0,  12,  -8,  -3,   5, -12,  26,  -6,  -4, -12, -19,   1,   0 },
      {  -4,  -5,  -6, -12, -12,  -1, -10,  -7,   3,   6,   8,  27,  34,  -8,  18, -19,  -1,  -4,   5,  -6, -65, -32,   3,  -7,   1,  -1,   5, -13,   9, -18,  -5,   5,  -1,   3,   5, -10, -16,  -2,  25,   0, -11,   6,  -2, -27,  11, -19,  15,   4, -11,  16, -21,   7, -22,   8,  27,  -2,   8, -36,  -9,  -6,   9,  27,  10,   6 },
      {  -2,  -2,  -1,   4,   0,  37, -13,  12,  -1,   0,   1,   5,   5,  23,  39,  -2,   0,  -1,   3,  18,  13, -12,  15,  -8,   2,  -2,   3,   8,   9,  -2,   9,   0,  -1, -10,   6,  23, -10,  -6,   3,   4,   3,  -6,  23,  -5, -14,  -9, -12,   1, -17,  -6,   9,  -4,   2, -19,  10,   7, -77, -26,  38,   4, -21,  -5,   6,   4 },
      {  -4,  -7,  -7,  -4, -29, -79,  17,  -7,   0,  -3,  -3,   8,   5,  24, -20,   1,   0,  -5,   5,   1, -25,  15, -10,   2,  -2,  -5,   0, -13,   0,  18,  10,  -9,   4,  -6,  -1,   9,   8,  -9,   2,  -4,  12, -18,   9,  -4,  -5,   0,  -3,  -1,   4,  -6,  20, -15,   8, -36,   7,   5, -26, -22, -18, -24, -15, -24,  13,   5 },
      {   4,   3,   5,   5,  -4,  27,   0,   0,  -1,  -6,  -2,  -9, -27, -41,   7, -14,   3,   2,  -3, -15,  -8,  31, -46,   5,   1,  -8,  -9,  -7,  28,  -7,  -4,  -2,   4,   2,   4,  20,   5,  14,  26,   3,   8,  -7,   2, -12,  -6, -15, -18,   7,   7, -12,  -3,  -5,  16,  11,   3, -21,  14, -42,   7, -12,   7, -50,  17,   5 },
      {   1,   1,   2,   1,  15,  -9,   2,   2,  -1,  -3,  -2,  -3, -41, -25,  21,  -3,  -5,  -8,  -6, -22,  -2,   0, -13,  -5,   3,   3,  11,  23,  15,  11,  15, -12,  -1,   4,  13, -33, -12, -20,   6,  12,   3,   0, -27,  -7,   6,  -1,  25, -10,  12, -10,   9, -10,   1,  14,  41,  15, -49,  26, -24, -11,  25,   3, -17,  19 },
      {   0,   1,  -4,  11,  15, -46,  14,  -2,  -4,  -3, -13,   2, -19, -57,   5,  -4,  -3,  -3,  -8,  14,   4,  -5, -21,  -5,  -2,   4,  -3,   6, -23,  -4, -29,  13,  -1, -13, -18,   5, -19,  10, -18,  -7,   5,  -7,   9,  11, -13,   4,   6,   7, -22, -22, -17,  -9, -22,   1,  -6,  -3, -16, -26,  27,  28,   6,  31,  -9, -16 },
      {   1,   0,  -4,   5,  13,  -8,  11,  -9,   1,   0,   0,  -7,  -5,  15,   3, -11,   2,  -1,  -3,  -3,  31,  30,  22,  -7,   2,   6,   9,   4,  14, -19,  43,   6,  -5,   5, -19, -11,  13, -22,   3,  11,   1,   9, -11,  27,  13, -13,  17,  12,  17,   6, -24,  12, -12,  -5,   6,  10,   6, -63, -23,  38,  -9,   1,  -4,  -1 },
      {  -1,   0,   0,   6, -10, -16, -11, -15,   0,   1,   1,   0,  -8,  -5,   6, -11,   0,   1,  -2, -13,  18,  -1, -27,  -5,   2,  -2,  -8,  14,  44, -28,  23,  -6,  -2,   7,  13,  35,  -7, -22,   2,   8,   3,   7,  12,  -7, -15, -11, -23,   8,  -9,  -3,  -4,  -1,  -8, -35,  17, -12,  23,  43,  -7,  14, -24,  45,  18,   0 },
      {   0,   0,  -1,  -4,   8,  31, -12,   6,   2,   3,   6,   6,  11, -17,  11,  -4,  -4,  -3,  -7,  -9, -20, -20, -49,  14,   3,   1,   9, -12,   6,  38,   7,  10,  -3,   2, -14, -14,  12, -22,  -6,  -4,   0,   2,   2,  16,  12,  25, -10,   2,  -2,  10,  -3, -15,   8, -44, -43,  22, -10,  -8, -35,  20,  -7,   8,  -7, -10 },
      {   1,   1,   2,  -4,   9,  13,  -4,   4,   1,   1,   3,   0,   4,  15, -39,   1,   1,   0,  -1,  -3,  15,  -9, -17,   5,   3,   6,   2,  27,   5, -49, -10,  19,   2,  -4,   1, -11, -63, -16, -43,   4,  -1, -14,  -3, -30,  -2,   6,   2, -10,   7,   2,  -4, -13,   2,  -8, -13,  10,  -3, -21, -21, -24,   4, -16,   7, -14 },
      {  -2,  -3,   0,  -7,  -6,  12, -48,  -1,  -2,  -3,  -1, -12,   0,   9, -51, -15,   1,   0,   2,   5,  19,   8, -22, -15,  -1,  -3,   3, -14, -23, -25, -24,   4,  -3,  -7, -12, -22,  20,  16,  34, -17,   3,  -4,   6,  18,  22, -15,   9,   5,  -8,   1,   4, -18,  -1, -18,  28,  -5, -26,  10,  -8,  18,  15,  11,  22,   5 },
      {  -1,   0,  -2,   2,  -4, -17,   2,  12,   1,   3,   0,   0,   7,  32,  31, -17,   2,   2,  -1,   7,  14, -31, -35,  11,  -4,  -2,  -6, -11, -18, -39,  -4,   3,   0,   3, -16,  12,   1, -18,  40,  10,  -1, -12,  -2,  12, -42,  31,  23,   1,  -1,  -2,  10, -12,  21,  34,  -4,   9,  12,   3, -10,  14,   3, -14, -16,  10 },
      {   1,   1,   1,   0,   4,   9,   2,   4,  -1,   0,   0,   2,   7, -36, -22,  29,   0,   2,  -2,   0,  -4, -15,  28,   6,   0,   2,  -1,   6,  -1, -24,  27, -10,  -6,   4, -13,   8, -14,   5,  45,   3,   7,   3,   5,  39, -12,  28, -31,  14,  19,  -2,  28, -15, -41,  -5,  -5,   9,  -4,  -9,  -7, -38,   6,   8,  -6,   1 },
      {   2,   1,   1,   0,   2,   4,  54,  19,   1,   1,   1,  -2,  -9,   4,  40,   3,   1,   1,   2,   3,   1,  23,  22,  -6,   3,   2,   7,   3,  -9, -16,  -6,  12,   0,  -9,  -2, -14, -19,  19,  20, -28,  -1,  -2,  22, -11,  18,  22, -14, -12,   3,  16,  -6, -34,  19, -23,  -1, -15,   6,  10, -23,  15,  21,  11,  40,  13 },
      {  -2,  -3,  -1,  -5,   0, -29, -66,  71,  -1,   0,  -1,  -2,  -3, -19,  11,  26,   0,   0,   0,  -4,  -5,  22,  25, -27,   0,   1,   0,   0,  17,   0, -17,  -8,  -2,   0,   0,   5, -10, -17,   7,   8,  -2,   4,   1,  -9,   1,  22,  13,  -2,  10,   6,  -6,   7,  17,  -5, -13,   5,  13,  -6,   1,   9,  -8,  -4,  10,   0 },
      {   1,   3,   1,   2,   7,  -1,  10,  66,   0,   1,   1,  -1,   0,  13,  -1,  12,   0,   2,  -1,   2,   6, -15, -10,  10,  -3,   0,  -3,  -1, -11,   7,  -7,   1,   2,   6,   7,   9,  24,  19, -27,  13,   1,   5,   9,   1, -21, -50, -34,  14,   1,   8, -12, -31, -21,   7,  27,   8,   7,  -8, -35,  -1,   9,   5, -19,   8 },
      {   2,   3,   3,   1,  -5,  25,  64,  27,  -1,  -2,  -2,  -3,  10, -11, -47,  17,  -1,  -1,   0,  -3,  -5,  -4,  -2, -11,   0,   1,  -3,  -5,  27,   4, -43,  -3,   0,  -3,  -4,  12,   7, -43,  16,   1,  -3,   0,   1,   3,  -6,   3,  27,   2,   2,   5,  -4,  18,   5,  -6,  14,   4,  -8,   2,  11,  13, -22,   6,  -4,  25 },
      {   1,   1,   1,   0,   5,  -6,  11,  43,  -1,  -3,  -3,  -2,   2,  -7, -46, -27,  -2,  -4,  -2,   0, -11, -27,  -2,  -7,   0,   2,   1,   1,  -9,   8,  75,  14,   3,   5,   2,  -3,  -7,  20,   9,   0,   0,  -6, -13, -20,   0,  12, -11, -13, -11, -15, -13,  11,  24,   7,  -7, -19,  -7,   5,  21,  23,   1,   1,   5,  13 },
      {   0,   0,   1,  -1,  -2,   9, -16, -39,   0,   0,   0,  -5,   2, -12, -26,  26,   1,   1,   0,  -2,  -4, -10,  36,  22,  -1,   0,  -3,  -6,   1,  17,   2, -25,   1,   1,   5,  16,   3, -15,   1,   1,   5,  11,  12, -25, -43,  11,   2,  -1,  -1,  -5, -33, -45,  30,  12,   9,  -5,  -8, -10, -20,  35,  24,   0,   6,  -6 },
      {   0,   0,   1,   0,   6,  12,   4,   6,   0,  -1,   1,   1,  -2,  -1,  -3, -34,  -2,  -6,  -3,   0,  -7,  -7,   9, -56,   0,   1,   4,   1,   6,  17,  15,   9,   0,  -5,  -1,  10,  18, -24, -23,   1,  -3,  -1,  20,  21, -28,  -8,  41,   6,   2,   2,  18, -32, -20,  14, -19, -25,   6,   6,  -3, -23,  23,  -3,  38, -22 },
      {   0,   0,  -2,  -1,   6,   2,   8,   5,   3,   4,   3,   1,   3,  16,  10,  78,   0,  -4,  -2,  -3,   5, -17, -45,  -7,   1,   0,   2,   0,  -7,  -4,  27, -38,   1,   1,  -1, -10,   4,   2,  -2, -28,   0,   1, -11,   3,   2, -13,  20,  -8,   3, -11,  -4,   9,  -7,   7,   4,   0,   6, -10,  19,   3,   6,   6,  43, -23 },
      {   1,   1,   1,   0,   3,  -3, -11,  23,   0,  -1,  -1,   1,  -1,   0,   0, -22,   1,   0,   2,  -3,   1,  15,  15,  60,   2,   0,   4,  -3,  20,   2,   1,  21,   1,   4,  -2,   3,   5, -29,   5, -84,   2,   0, -13,   7, -19, -15,   3, -24,   0, -11,   0,   3, -16,  11,   1,   4,  -4,   4,   6, -10,   3,   6,   5, -15 },
      {  -1,  -1,  -1,   1,  -2,  -2,  -5, -24,   0,   0,   1,  -1,   2,  -3,   0,  35,  -1,  -1,  -3,  -1,  -4,   3,  -4, -47,   0,   2,   1,  -2,   5,  11,   6,  83,   2,   3,  -1,   5,   5,  -4,   5, -19,  -4,  -7,  -7,  -3, -18,  -6, -27,  -4,   1,   0,   5,   1,  14,  16,  19,  48,   9,   3,   3,   5,   5,  -3,   3,   9 },
      {   0,  -1,  -1,  -1,   3,  -5, -14,  -6,   0,   0,   0,   1,   5,  -2,   5,  39,   0,   0,  -2,   0,   3,  -1,  -3,  40,   1,   0,   2,   4,   1,   1,   4,  57,   3,  -2,   3,   2,  10,  -3,   2,  38,   5,   0,   8,   8,   6, -11,  31, -30,  -3,  -3,   0,  -6, -17,  -8, -16, -55,  -5,   0,   0,  -4,   2,   1,   8,  46 },
      {   0,   1,   2,   2,  -3,   1,   0,  -8,  -1,  -1,   0,  -1,  -1,   2,  -3,   1,   0,   0,   1,   0,   3,   8,  -7, -43,  -1,  -1,  -2,  -5,   3,  -8,  -2, -25,  -2,   0,   0,   6,   3,   3,  16, -10,  -4,   0,   5,  -1,  -8, -13, -15, -96,  -9,   9, -14,  -6, -17,   2, -24,  -8,  -5,  -8, -15,  -2,  -6,   4, -43,   1 },
      {   0,   1,   1,   1,   1,   2,   3,   4,   1,   1,   2,   0,   5,   2,   4,  21,   0,  -1,   0,   0,  -2,  -2,  -4, -14,   0,   1,   4,  -1,   4,   0,   5,  25,   1,   2,   1,  -1,  -3,  -1,   9, -19,   2,  -3,   4,  -4,   7,  13,   5,  20,   4,   9,   8,   1,  17,  -5,  35, -68,  -9,   0, -17,  -1, -17,  -6, -53, -64 },
      {  -1,  -1,  -1,   0,  -2,   0, -13, -12,   1,   1,   2,  -1,   0,   4,   4,   7,  -1,   0,  -2,   1,  -1,  -2,  -8, -27,   0,   1,   0,   3,  -2,  -5,   8, -17,  -3,  -1,  -4,   1, -11,  -1, -23, -60,  -1,   4,   1,  -2,  -6,   3,   1,  47,   1,   3,  -8,  -2,  -6,   5, -21, -29,  -4,  -2,  -8,  -2,  -6,  -6, -29,  77 },
    },
    {
      {-105,  58,   2,   1,   0,   1,   0,   0,  -2, -15,  15,  -4,   2,  -1,   1,   0,  29, -22,   1,   2,  -1,   0,   0,   0,   7,   0,  -6,   2,   0,   0,   0,   0,   5,  -1,  -1,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   1,   2,  -1,   0,   0,   0,   0,  -1,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -17, -33,  27,  -3,   4,   0,   1,   0,  80, -65,  -5,   9,  -4,   3,  -2,   1,   9,  30, -34,   6,  -2,   1,  -1,   1, -23,  25,  -2,  -7,   3,  -1,   1,  -1,  -4,   4,   5,  -3,  -1,   0,   0,   0,  -5,   2,   2,   0,   0,   0,   0,   0,  -1,   1,   0,   0,  -1,   1,  -1,   0,  -2,   1,   1,   0,   0,   0,   0,   0 },
      {  -3, -57,  44,  -3,   3,   1,   1,   0, -63, -26,  31,   9,   1,   4,   0,   1,  28, -35, -24,  19,  -2,   4,  -1,   1,  18,  18, -36,   0,   3,   0,   0,   0,  -2,  16,  -5, -11,   3,  -1,   2,   0,   2,   5,   0,  -7,   0,   0,   0,   1,   2,   3,   1,  -2,  -3,   1,  -1,   0,   1,   2,   0,   1,  -2,   0,   0,  -1 },
      {  57,  53, -36,  -7,  -4,  -3,  -1,  -1,  20, -34,  42, -15,   4,  -2,   0,   0,  31, -40,   1,  21, -12,   7,  -3,   2, -18,  28, -26,  13,  -1,  -2,   2,  -1, -11,   4,   0,  -4,   7,  -2,   1,  -1,  -4,   2,   0,  -2,   1,   1,   0,   0,  -3,   0,   2,   0,   0,   0,   1,   0,  -1,   1,  -1,   0,   1,   0,   1,   1 },
      {   5,  20, -28,   6,   1,   0,  -1,   1, -42,   5, -17,  14,   4,   2,   0,   1,  28,  30, -59,  -7,   8,  -3,   1,  -2, -16,  47,  19, -43,   8,  -5,   2,  -3,   8, -21,  42, -10,  -9,   3,  -1,   0,   3, -11,   7,   8,  -6,   1,   0,   1,   3,  -6,   3,   2,  -1,  -1,  -2,   1,   0,   0,   1,   2,  -1,   0,  -1,  -1 },
      {  -3, -27,   9,   7,  -2,   1,   1,  -1,  35,  57, -35, -14,  -2,  -6,   3,  -4,  70, -32,  25, -10,  -1,  -5,   0,  -1,   4,  30,  -6,   1,  -2,   1,  -5,   2, -26,  10,  24,  -1,   0,   1,  -3,   0,  -4, -12,   2,  14,  -1,   3,   1,  -2,  -8,   3,  -3,   1,   4,   0,   3,  -1,  -3,  -3,  -1,   1,   3,   0,   1,   1 },
      {  16,  21,  63, -55,   3,  -6,   2,  -3,  -9,  11,  32, -11, -17,   8,  -4,   1,  27,  24,   1,   5, -19,  -2,  -1,  -2, -12,  -6,  40,  12,  -7,  -6,  -1,  -1,   3, -21,   7,  31,  -3,  -1,  -2,   1,  -6, -11, -15,  25,   5,  -3,   1,   0,   1,  -2, -19,   6,   9,  -2,   1,   0,  -2,   1,  -7,  -2,   6,  -2,   2,   0 },
      { -16,   1,  20,  -6,  -8,   6,  -3,   0,  13,  45,  19,  -4, -16,  -3,   1,  -2, -56, -12,  -2,  51,  -3,  -6,   3,   0, -37,  27, -30,  -4,  31, -10,   2,  -1,  -1, -24,  26, -24,  15,   2,  -5,   3,  10, -16,  -5,   1,  -1,  10,  -4,  -1,  -3,   1,  -1,  -2,  -5,   3,   3,  -2,   2,  -2,  -2,   1,  -1,  -2,   1,   1 },
      { -21, -45, -31,  33,  -1,   7,   1,   3, -14, -16,  42, -30,  16,  -6,   2,   0, -13,   0,  33, -20, -11,   9,  -5,   3, -31,  30,  21,  12, -25,   5,   0,  -2,  -3, -33,  19,  28,  -6,  -6,   7,  -3,   6,  -8, -24,  15,   7,  -5,   0,   2,   2,   1,  -6,  -4,   5,  -2,  -1,   1,   0,   1,  -4,   0,   2,  -2,  -2,  -1 },
      {   3,   3,   5, -24,  10,  -6,   1,   0,  20,  -9,  19, -24,  10,  -2,  -2,   0, -38, -30,  -2, -13,  12,   8,   0,   1,  57,   3,  -4, -47, -12,   9,   1,   1,   6,  11,  42,  -1, -39,   0,  -2,   2,   3,  -5,   6,  36, -15,  -8,  -3,   3,  -5,   5,  -7,  18,   4,  -6,  -4,  -2,   0,   0,  -3,   4,   7,  -1,   0,  -2 },
      {  12,   0,  10,  37, -13,   6,  -4,   1,  13,  -4,   5,  33,  -7,  -3,   2,  -3,  13, -47,  -9,  19,  34, -15,   8,  -2, -17, -35,  25, -35,  22,  14,  -6,   4,  14, -27,  -6,  18, -24,  15,  -3,   2,  -8,  15, -39,  11,   8,  -4,   7,  -3,  -2,  10,  -8, -19,   3,   1,   4,   0,   0,   6,  -2,  -4,   0,  -4,   2,   3 },
      {  -4,  -7,  -7,  20, -12,   5,  -3,   4,  16,  38,  76,  12, -20, -11,  -4,  -1,  12,  26, -15,  -7,  31, -10,  -5,   1,  -1, -11,   7, -12,  -4,  17,  -2,   1, -21,  28, -10,  -1,  -8,  -5,  11,  -3, -18,  -5,  35, -15,   0,  -9,   0,   4,   6, -23,   8,  18,  -6,   1,  -8,   5,  -4,  -2,  -6,   4,   4,   1,  -4,  -3 },
      {  -1,  10,  -3,  11, -15,   1,  -1,  -1,   6,   2,   1,  40, -16,   1,  -1,   0, -25,  -2, -15,  10,  -6,  -6,   1,  -1,  64,  39,  15,  14, -10, -14,  -5,  -4, -49,  -8,   5,  40,  27,   0,   2,  -3,  -5,   3, -23,  -9,  24,   4,   6,  -1,  -9,  -5,   6, -14,   9,   4,   3,   0,   0,  -4,  -3,  -8,   3,   3,   2,   2 },
      {   9,  24,  37,  33, -27, -10,  -2,   0, -15, -22, -12,  48,   9, -10,   0,  -1, -15,  -6,  27, -21,  13,   1,  -7,   2, -25,   6,   0,  15, -12,   4,  -1,  -1, -22,  11,  14, -23,   3, -12,  12,  -1,  11, -25,  23,  25, -14, -10,  -7,   9,  -9,   1, -31,  22,  22,  -4,  -7,  -1,  -4,   0,  -4, -14,  13,   3,  -1,  -4 },
      {   0, -16, -28,  27, -16,   8,  -1,   3,   3, -12,  17,   7, -22,   2,  -2,   0,  26,  33,  13,  31, -26, -12,   0,  -2,  43, -29,  -4,  20,  27, -21,   3,  -2,  24, -14,  25, -22,  11,   7, -12,   4,  -4,   1,   4,  32, -21,  11,  -7,  -2, -12,  10, -19,  16,  14, -10,   8,  -5,  -2,   0,  -3,  -6,  14,  -5,   2,   1 },
      {  16,  22,  46,  37, -25,   9, -15,   4,   3,  -5,  11, -39,  44, -12,   9,  -7,   6,  28,   4, -21,   7,   4,   1,   1,  28,   2,  -2,  -2,  17,  12,  -7,   2, -15, -24,   4, -21,  12,  23,  -6,   3,   7,   3, -16, -16, -12,  19,   3,  -2,   3,  -2,  14,  -8, -25,  10,   6,  -3,   6,  -1,   1,  16, -15,  -4,   4,   0 },
      {  -4, -14, -15, -31,  25, -22,   9,  -8,  -6, -21,  14,  19,  -7, -13,  -1,   0,  -3, -19,  27, -23,  19, -20,  -6,   3,  -6,   6,  37,  -8,  31, -10, -18,   5, -21,  -5,  -1,  -6,  22,  37, -30,   2,  12,  -5,  18, -14, -20,  41,  -3, -14,   0,   1,   0,  17, -17,   6,  15,  -7,  -1,  -2,  -3,   0,  -4,   4,   9,   2 },
      {   3,  -8,   5,  -8, -22,  19,  -9,   5,  16,  11,  -2,  13,  -3,  -8,  -2,   0,   3, -36, -20,  -8,  -7,  14,  -5,   7,   8,   1,  31,  13, -32,  -8,  10,  -5,  28, -46,  -3, -12,  22, -23,   8,   5,  15,  27,  18, -29, -25,   0, -18,  11,   1,   6,  14,  40, -15,  -8, -12,  -1,   2,   5,  -9,  16,  10,  -9,   1,  -7 },
      {  -3,  -5,  -8,  14,  46, -23,  16,  -9,   9,  16,   6,  -3,  34,   5,  -1,  -1,   4,  -6, -45,  13, -14,  32,  -1,  -1,  -1, -39,   0,   4,  -5,   0,   9,   1, -51, -25, -14,  -6,   8,   7,   8,   5,  28, -18,  -5,   7, -20,  -2,   2,   1,   4,   0, -18,   6,  11, -10,   3,  -1,  -2,   3,  -8, -17,   9,   0,   2,   3 },
      {  -1,  -2,  -6,  -3,  -4,  31, -11,   5, -13, -28, -18, -20, -24,  27,   0,   2,   3,  -6,  19,  27, -12, -10,  11,  -6,  -6, -29,  27, -20,   7,  -4,   3,   4, -59,  11,  27, -11,  -5, -17,  17,  11,   4, -14,  -4, -20,   3, -12,  -6,  16,  12,  -6,   3,   6, -29,   7, -22,  -1,   4,   6, -12,  20,   3, -12,  -2, -10 },
      {  -1,  11,  18,  33,  12, -13,   7, -13,  -9,   7,  -4,  -2, -11,  13,  -7,  -1,  -9,   7,  20,  17, -36,  13,  -9,   3, -13,  27,  15, -26,  -1,  -8,   4,  -1, -19,  15, -24,   0, -24,   8, -11,  -1, -14,  57,   6,   8, -23,  -7,   7, -20,   0,   5,  26,  22,   5, -24,  20,  -6,  -1,  -3,  -6,  13,  20, -16,   5,   7 },
      {  -1,   2,   1,  18, -37,  15,  -1,   0,  -6,  -1,  -6, -36, -21,   5,  12,  -3, -15, -19, -51, -27,  -7,  -1,  10,   0, -23, -22,  -4,   8, -15, -18, -10,   8,  -6,  39,   8,  26,  18,   3, -32,   0,  -3,  -6,  -3,  15,   2,  32,  -4, -18,   4,   7,  -8,  16,  -4,  -9,  21,  -9,  -8,  -1,  -3,  -4,  11,  -2,  10,   4 },
      {   7,   6,  15,  38,  36, -31,  17, -11,   5,   3,   6,  -5,  15, -14,   5,  -2,   0,  -6,   5,  16, -13,   3,  -2,   1,   6,  -1,  27, -22,  -6, -27,  -9,   5,  33,  26,  12,  -5,  29, -38, -13,  12, -17, -10,  -1, -10,  31,   8, -31,   9,  11,  -9,  -5,  -4,  -8,  28, -14, -12,  -9,   1,  -4,  -1,   1,   5,   9,  -8 },
      {   7,   9,   3,  20,  16,  -6,   5,  -2,   6,   4,   9,   5, -34,  25,   2,   6,   8,  14,   8, -18,   0, -15,  25,  -5,  10,  20, -22, -13,  -1,   4,  -5,   4,  17,   0, -31,  16,  -8,  -7,  -1,   2,  69, -33, -16, -10,  -1,  -4, -12,   5,  -5,  31,  -1,  14, -22,  -4,  -5, -10,  -6,  11,   3,   8,  13, -17,   9,  -4 },
      {   4,  -1,   8,  24,  10,   7, -14,   9,   3,  -6,  -9, -32,  -2,  17,   4,  -2,  -9, -25, -13,  13, -10, -26,  18,  -7,  13,  20,  40,  22,  18,   3,  -3,  -8,  20,  -1, -25,  -5, -11,  20,   7, -10,  20, -20,  47,   6,  10, -17,  18, -10,   4, -34,  -2,   4,  23,  13,   5,  12,  11,  11, -15, -10,  -8,  13,  -8,  11 },
      {   4,   2,   8,  16,   3,  -2,  -5,  -2,  11,   5,  12,  35, -14,  28, -25,   5,  -1, -15,   9, -27, -38,  39,  -4,  -3,   0, -12,  -1,  -5,  -3, -13,  25, -11,  21,   9,  27,   3,   3,  22,   3,  -9,  -5, -31,   7, -10,  -2,  12,  33,  -3,  25, -10,   5, -22, -22, -11,   8,  18,  13,  -3,   4,  19, -26,  -3, -10,   7 },
      {   2,   1,   1,  -1,   3,   5,  -7,   0,  -3, -10, -16,  13,  17, -38,  13,  -6,   8,  12,  13,  57,  13,   5, -13,   2,   1,  -9,  -6,  -1, -34,  26,  -2,  -5,  11,   8,  14,  33,  -5,   1, -11, -15,  28, -27,   8,  -1,   5,  24,  13, -20,  16, -21,  23,  16, -12, -13,  21,   2,   5,   4, -15,  19,  19,  -4,  -5,  11 },
      {   1,   6,   9,  15,  40,   1,  -5,  -5,  -7,  -3,  -6,   1,   6,  17, -11,  12,  -8,  -4,  -8,  -3,   7, -18,  19,   6,  -3,  -5, -17,  27,  -8,  19,  -2,  -3,   1, -26,  17,  19,  -7,   8,   8,  -7, -61, -19,  28,  -7,   3,   6,  -3,  -1, -46,  18,   9,  10, -24,   2,   9,  -9, -18,  16,  -4,  14,  10, -24,  10,  -2 },
      {  -2,   2,  -9,  -2,  11,  16, -13,   5,   1,   9,  -5,   1,   6,  -4,  -1,  -3,  -9,  -1, -10,  18,  -2,   6, -21,   2,   7,  29,  26,  16,  -2,  28, -25,   1,  17,  38, -19, -15,   6,  23,  18,   2, -26, -15, -39,  -3, -31,   4,  16,  16,  18,  13, -47,  10, -22, -13,  -7,  -5,   0,   9,   2,  -2,   5, -19,   6, -12 },
      {   1,   0,   1,  11, -19, -56,  41, -14,  -3,  -5, -13, -17, -35, -29,   9,   3,  -4,  -8, -18,  -4, -20, -14, -36,  13,   2, -10,  -5,  23,  16,  20, -24,   5,   1, -11,  19,   7, -15,  17,  19,  -9,   7,   3,   3,  -6,   6, -17,   9,  11,   8,   1,   7,   2, -15,  -4, -25,  10,   1, -10,   4,  19,  -1,  -5,  -2, -11 },
      {   0,  -6,  -3,  -5, -16,  -2,  15,   1,   1,  -6, -12,   2,  10,  -2,  14,   2,  -1,  -5,   9, -11,  -2,   1,  13, -13,   6,   8, -29, -18,  21, -24,   5,   6,   2, -27, -36,  20,   5,   3, -15,   7, -38, -30,  -9,   8, -15, -12,   0,  10,  33, -53,  -1,  23,  -5, -17, -21,   5, -28,   4, -16,   9,  17,  -9,  -3,  -8 },
      {  -5,  -7,  -9,  -6, -44, -32,  -5,   4,  -2,   0,  -7, -11,   4,  -7, -38,  12,   1,   0,  -2,  13,   2,  10, -14,  15,   7,   8,   9, -11,   7,   7,  34, -16, -11,  -8, -35,   4, -15, -10,  -2,   4, -17, -28,  21,  10,  11,  12, -28, -11,  17,  29, -16,  -9,  -6,  19,  14, -19,  27,  23,   0,   8,  -5, -16,   9,   3 },
      {   4,   4,   1,   5,  23,  22,  11,  -3,   1,  -4,  -4,  10, -37,  12,  22, -14,  -1,  -2,  -1, -15,  -9,  -4, -18,   6,   7,   2, -19,   0, -20,  32, -20,   5,  -7, -32,  -9, -19,  -9, -24,  -3,   5, -11,   7,   4,  20,  -1,  12, -20, -18,  29, -35, -22, -10, -12,  22,  27, -20,  32, -18,   1,  -4, -16,  -8,   8,   9 },
      {   4,   4,   9,   9,   9,  -1,  11, -18,  -5, -13,  -7, -35, -37,  -5,   5,  -6,   3,   3,  12,   1,  31,   8,  -2,   5,   7,   3,  -3, -30,  -5,   7,  19,  -3,   6, -17,  -3,  -1,  40,  -9,  19,   2, -24, -15,   4, -29,  -2,  -1,  26,   0,   6,  28, -11,   8,  16, -30,  13,  16,  20, -14,  -4, -28,  16,  20, -17,  21 },
      {  -3,  -5,  -5, -12, -27,  -1,  21, -14,  -2,  -3,   6,   3, -11, -16,  41,   2,  -1,  -2,   3,   6, -30,  13,  17,   2,   4,  -1,  24, -25, -22,  23,   0,  -5,  -2,  -1, -20, -35,   6,  -4,  -1,   0,  -4, -26,   3,   5,  -1,  -2,   9,  -7, -45,  -7,  10, -24,  -8,  -6,  15,   5, -45,  21,  20,  -7, -18, -11,  -3,   8 },
      {   0,   1,   2,  12,  -1,  -6, -26,   8,  -4,  -4, -11, -14, -31,  22, -28,  20,   3,   0,   5,  15,  10,  41, -28,  -3,   0,   0,  -1,  -5,  -6,  27,  -5,  -6,   3,  -9,  -7,   7,  -2,  15, -13,   1,   2,  -8,  -9,   0, -11,  18, -15,   5, -23, -22,  14,  16,   7,  18, -19, -14, -37, -21,   6, -16, -14,  47,   9, -20 },
      {   2,   2,   5,   5,  18,  -9, -24,  24,  -1,  -2,   3,   3, -11, -29,  10,  10,  -2,  -7,  -9,  -3, -14, -26,  16,   6,  -2,  -3,  10,  13, -13,  -5,   7,  -6,  -7,   4,  -8, -11, -13,  -2, -22,  13, -19, -35, -19,   3, -32,  -7, -23,   3,   2,   8,  42, -19,  13, -28, -13, -16,  36, -23,  30,   5,   5,  20, -20,  -7 },
      {   0,   0,   0,  -2, -10,  14,  -2, -17,   3,   3,   5,   9,  16,  12,   7, -22,  -1,   1,   5,  10, -25, -20,   6,   5,   3,  -1,  -4,  -6, -18,   8, -14,  16,  -3, -14,   8,   5, -14,  19,  -4,  -8,  -9,  -8,  26, -22,  -4,  -9,  -5, -11,  61,  47,   1,  -5,  15,   9,  -7, -15, -42,  -4,  17, -13,   1,  34,  13,  -2 },
      {  -2,  -6,  -4, -11,  -9, -30, -15,   7,   2,  -2,   2,   1,   2,  15, -15,   3,   0,  -5,  -1, -15,   3,   3,  20, -12,   0,  -7,   4,   6,  16,  22, -20, -28,   7,   6,  19, -22,  22,  -6,  38, -13,  -2,  -4, -31,  -3,   9, -18, -13, -15,  25, -17,  32,  12,  16,  -1,  33, -30, -17,   1,   5,   0,  32,  -8,  12,  30 },
      {  -2,  -5,  -6,  -6, -31,   2,  45, -30,   1,   2,   3,   8,  25,  35,  -1,  -9,   3,   2,   3,  17,   2,  10,  27,   5,  -4,   1,   7,  -1, -13,   3, -17,  10,  13,  10,  -3,   5,  -1,  24,   5,   8,  -7, -19,  -4,  -6, -13, -12,  -8,   7, -27,  -3,  15,  12,  -2,  15,  -4, -16,  50, -46, -18,  -3,   0,   7,  14,   3 },
      {   2,   3,   5,   7,  -1,  -3,  41, -12,   1,  -3,   0, -11,  -3,  27,   2,   1,  -1,  -2,   1,  18,  56,  -1,   1, -12,   7,  12,  20,  41,  -6, -16,  34,  -4,   1,   6,  10,  -4, -18, -11,   3,  13,   6,   3,   1,   8, -25,  14,  -9,   3,  26,  -8,  11, -21, -10, -12,   9,   3, -32, -13,  35,  -2,  -9, -20,  -1,   5 },
      {   2,   1,   1,  -1,  10,  52,   5, -39,   1,  -2,   2,   1,  11,   0, -14, -10,  -1,  -7,  -9, -14,   0,  -6, -56,  28,   2,  -4,   7,  -2,  36,  -6,  -6, -16,   2,   6, -17,   4,  12, -19,   7,   0,   7, -29,  -1,  22,   4, -13,  -7,  -7, -10,  13,  29, -11,  15,  -8,   1,  -2, -18, -20,  -8,  27,   5,  -4,   1,   4 },
      {   2,   4,   3,   2,   1,  10,  23, -10,   0,   1,  -1,  -7, -17,  -3, -23,  27,   1,   6,   2,  -1,  19, -10,  -8,  25,  -1,  -3, -10,   1, -37, -31, -24,  -4,   1,   0,   8, -19,  -2,  17,  -6,  -9,   0,  -4, -16,  -9, -28, -12,   3,  -6,   7, -11,  15, -25,  23,   9,  -5, -13,   5,  65, -39,  -3,  -9,   5,   6,   3 },
      {   0,   2,   0,   3,  16,  16,  -5, -23,  -5,  -3,  -4,   3, -28, -49,  20, -20,  -1,   4,   0,  -4,   3,  34,  35, -12,  -2,   4,  -2,  12,  17,   3,  30, -12,   0,  12,   4,  20,   3,  24,  25,  15,  -2,   7,   2,  17, -13, -12, -18,  13,  11,  11,  11,  10,   7,  23,   1, -12,   0,  11, -20,  11, -33,   7,  30,   3 },
      {  -2,   0,  -5,  -3,  -1, -13,   2,  -7,   0,   3,   3,  11,  -2,  29,  26, -14,  -1,   3,  -4,  -8,   9,  25,   3,   8,   5,   8,   8,  31,  19,  34,   7,   9,  -1,   4,   8, -21, -20, -16, -44,  29, -10,  -5, -19, -10,  21,   2,  -4, -10,   7,  19,   9,  12,  12,  -5,  -3,  21,   0,  27, -43,  21,   7,  24, -26,   5 },
      {   3,   5,   8,   2,   3,  34,  46,  39,   2,   4,   2,   6,   9,   4,  31,  47,   0,  -1,   5,  -4, -11,  27,  -3,  22,   0,   2,   4,  -2,  37,  18,  -1, -12,   1,   8,  10,  25,   2,   0,   6, -13,  -1,   5,  12,  -5,  -5,  11, -14,  -1,   8,   6,  10,   5,   2,   8, -11,  -8,  17,  26,  26, -26,  21,   8,  -6,   4 },
      {   0,   1,   1,   0,  -4,  -4,  -3, -16,   2,   4,  -1,   4,  11,  -1,   7,  15,  -1,  -1,  -2,  -9, -28, -47,  -5,  28,  -2,   0,  -7,  -4,   2,   5,  44, -13,   5,  12,   5,   9, -26,  -2,  35,  18,   4,   5, -23, -26, -19,  36, -17,   6,  -9, -20, -15,   4,  -4,   9,  15,   4, -18, -11, -32, -19,   1,  22, -16,  16 },
      {   0,  -2,   3,  -3,  -6,   2, -18, -20,   1,  -1,   2,  -3,  -4,  -3,  14, -30,   1,  -1,   2,  -7,  -1,  24, -23, -18,   2,   0,   0,   5,  19, -14,  -4,   0,   7,   0,  15,  14, -34,   2,  -5,  14,   5,  -8,  14, -55,  -1,  -8, -12,  -2, -23, -25, -18, -27,  13, -29,   9, -40,  11,  18,  10, -20,  24, -12,  15, -14 },
      {  -1,  -2,   0,   0,  -1,   3, -11, -58,   1,   2,   1,   0,   3,  13, -16,   0,   0,  -1,   1,  -2,  -4, -19,  15,  14,  -2,   1,  -2,  -3,  -5,  41,  24,   3,   1,   3,   6,   0,  34,  21, -13,   1,   9,  21,   6,  26,  17,  10, -12,  22,  10, -26,   4,  -8,  -4, -25, -25,  -9,  18,  33,  27,  -8,  19,   1,  -9, -19 },
      {   2,   3,   2,  -1,   8,  32,  33,  -3,  -1,  -2,   0,  -1,  -7, -37, -33,  22,  -1,  -2,  -4, -11,  -6,   1,  16, -18,   3,   3,   6,  11,  -6,   7,  33,  -1,  -2,  -4,  -2, -40, -24,  14, -22, -16,   1,  -3, -16, -17,  41,   3,  19, -15,   2,  -4, -14,   8,   4,   5,  -9, -15,  -7, -27,  -6,   3,  34,   1,  19, -12 },
      {   1,   2,   2,  -3,   0,  -2,  24,  43,   0,   0,   1,   0,  -3,  10,  -3, -44,  -2,  -2,   0,  -7,  -9, -14, -20, -43,  -3,   0,   0,  -9,  -8,  20,  14,  -8,   4,   2,  -6,  -4,  14,   9,  11,   1,  -2,   1,  12,  18,  -6,  32,   0,  27,  -1,  13,  30, -36,   4,  -1, -11,  -6,  -3,   5, -44, -13,  26,  -6,   9, -18 },
      {   0,   0,   1,  -2,  -4,   3, -12,  -1,  -2,  -3,  -2,   0,  -3, -23, -38,  21,   0,   0,   0,  -2, -10,  16,  27,  21,   3,   5,   6,   8,  20,  22, -19,  59,   5,   1,   6,   8,  -3, -28,  -4,  22,   7,  -1,  15,   2, -17,   0,   5,   6,  11,  -3,  17, -31,  -8, -11,   7,  17, -11, -17, -27, -40,   6, -23,  12,   5 },
      {   1,   2,   2,   4,  -2,  10,  33,   7,   0,   0,  -1,   0,  12,  -8, -40,  -8,  -2,  -2,  -2,   1, -16, -29,  -8, -43,  -1,  -1,  -3,  -1,  10,  32,   1, -16,  -2,   5,  12,  18,  10, -29, -33,   8,   1,   0,  -8, -17, -18, -27,  -2,   4,  -7,  -3,  -5,  23,  -1, -21,  14,   1,   4,  19,   9,  -9, -50,  13,  11,  15 },
      {  -1,  -1,  -1,   1,  -4,   3,   0, -31,  -1,   0,   0,   2,  -2, -12, -17, -27,  -1,  -1,  -3,  -5,  -5,   5,   4, -23,   0,   1,   1,  10,  10, -20, -16,   2,  -3,  -6,  -9,  -4, -28, -34,  15,  -8,  -8,  -7, -11,  16, -15,  23,  33,   3,  -5,   1,  22,  14, -34,  36,   1,  10,  16,  31,  27, -21,  29,  38, -11,   3 },
      {   1,   3,   3,   2,   4,  15,  11,   7,  -3,  -2,  -3,  -2,  -8, -13, -33, -29,  -1,   1,   0,  -1,  -8,  11,   0,  -2,  -2,  -1,  -6,  -4, -12,  -1, -20,   0,  -1,  -1,  -1,   1,  -8,  38,  -2,  30,   0,  -5,  -3, -14,  10,   7, -41,  15,  -2,  -3,   0,  12,  23,  27,   2,  34,  -8,   3,  26,  -2,  13, -35, -51,  32 },
      {   0,   0,  -1,   1,  -4,   5,   9,  -9,   0,   2,   1,  -1,   3,  -7,  -9,  36,  -2,   0,  -2,   3,  -1,   7,  15, -39,   0,   4,   0,  -3,   0,   4, -32,   2,  -1,  -3,  -7,   7,   3,  -5,  20,  26,   3,  12,  22,   5,  -8,  22,  -5,   8,  -3,   3, -30, -33,   6, -18,   1, -15,  -4,  -2,  -3,  63,  15,  40, -18,  23 },
      {   0,   1,   0,   1,   4,  -1,  20,  32,  -2,  -2,  -2,  -2,  -6,  -1, -24, -55,   0,   0,   0,   2,   2,   6,  20,  47,   1,   3,   5,   0,  10,   0,  15,  45,   0,   2,  -2,   2,   1,  -9,  14, -41,   1,   0,  -8,  -2,  -4,  11,  -5,   0, -15, -17, -13,   1,   7, -19,  -6, -27,  -8,  11,  11,  32,  -9,  23,  -5,  15 },
      {   0,   0,   0,   0,  -1,  -4,  -4, -24,   0,   0,   0,  -4,  -2,   1,  -1,  -6,   0,   0,   0,   1,  -3,   6,  31,  -3,   1,  -1,   1,  -3,  10,  12, -24, -41,  -4,  -1,  -1,   3, -14, -30,  -9, -55,  -4,   2,  -2, -18, -22,  45, -10,   2,   5,   2,  -9,  -7,  41,  -3, -31,  29,   5, -13,  -7,   7, -13, -27,   7, -12 },
      {   0,  -1,  -2,   0,   0,  -8,   0,   0,   0,  -1,   1,  -1,  -2,   3,   1,  -1,  -1,  -2,  -2,  -3,  -3,  -9,  -1,  -4,  -2,  -2,  -2,  -5,   2,  23,  24,  19,   5,   5,   8,  16,  25, -14,  -3,  25,   3,  -1,  -3, -12, -41,  -6,  41, -36,  -5,  -5,  -5,  -5,  39,  57,  -9, -15,   0,   1,  18,  27,  15, -24, -15, -26 },
      {   0,   0,   0,   0,   1,   3,   3,   0,   1,   0,   2,   3,   5,   4,   3,  14,   0,  -1,   1,   0,  -4,  -8, -10, -16,   0,   0,  -1,  -4,  -2,  -6,  -7,  38,   1,  -4,  -5, -10, -12,  -1,  42,  32,  -6, -12,  -7,   1,  31,  48,  10, -15,   3,   1,  25,  35,  41, -27, -12,  -2,  12,  17,  15,   0, -36, -13,  23, -12 },
      {   0,   0,   0,   1,   0,   0,  11,  16,   0,  -1,  -1,   0,   0,  -2, -10, -14,   1,   1,   2,   2,   0,   0,  22,  48,  -2,  -2,  -4,  -2,   0,  -2, -21, -49,  -3,  -4,  -2,   0,  -9,  -9,  -9,  46,   5,   4,   9,  -1,  14,  11,  45,  48,   9,  -4,   0,   8,  13,  -3,  31, -15,  -7,  -5,  10,   7,  -1,   4,  10, -32 },
      {  -1,  -2,  -1,   0,   0,  -4, -14,   0,   0,   0,   1,   0,  -2,   8,  16,  10,  -1,  -1,  -1,  -1,  -2,  -7, -21,   3,  -1,  -1,   0,  -1,  -5,   4,  15,  19,  -2,  -1,  -2,  -3,  -4,  -1, -25,  -5,  -2,  -5,  -2,  -7,  -6,   3,  16,  58,  -3,  -5,   3,  -1,  21,  19,  -4,  17,   8,   3,  20,  19,  19,  -2,  61,  64 },
      {  -1,  -1,   0,   0,  -1,  -1,  -8, -10,   0,   1,   0,   1,   1,  -3,   7,   7,   0,   0,  -1,  -1,   1,   3,  -4,  -7,  -1,   0,   0,   4,   3,  -2,  -1,  -1,  -1,  -2,  -1,   1,  -6,  -7, -15, -10,   1,   4,   6,   5,   8,  10,  32,  26,   5,  11,   7,  12,  -1,   5, -30, -74,  -8, -10, -17, -19, -26, -29, -51,  44 },
      {   0,   0,   0,   0,   0,   3,   2, -13,   1,   0,   0,  -1,   0,   1,   8,  21,   0,  -2,  -1,   0,  -2,   1,  -7, -20,   1,   0,   4,  -2,   5,   0,   5,  37,  -1,   1,  -2,  -2,  -2,   0,  -2, -43,  -2,  -6,  -5,  -7,   0,  -4, -13,  54,   6,   6,   8,   9,  21,  12,  69,  -4,   2,   1,  -3,  11,  -7,  -1, -31, -45 },
    },
  },
  { // 30
    {
      {-104,  29,  16,   3,   3,   1,   1,   0,  57, -29, -10,   1,  -1,   0,   0,   0,   1,  11,  -2,  -2,  -1,   0,   0,   0,   1,  -2,   2,   0,   0,   0,   0,   0,   1,  -1,  -2,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  33, -48,  14,   3,   2,   0,   1,   0,  77,  18, -35,  -6,  -5,  -1,  -1,   0, -58,  33,  20,  -3,   2,  -1,   0,  -1,  -1, -17,   5,   4,   1,   2,   1,   0,  -5,   4,  -2,   0,   0,   0,   0,   0,  -2,  -2,   1,   0,   0,   1,   0,   0,  -1,   1,   0,   0,   0,  -1,   0,   0,  -1,   0,   1,   0,   1,   0,   0,   0 },
      { -32, -81,  52,   5,   5,   2,   2,   1, -25,  29, -35,   8,   0,   0,   0,   0,  52,  -2, -11,  -6,   0,  -2,  -2,   0, -15,  10,   4,  -2,  -1,   1,   1,   0,   3,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   1,   0,   2,   1,   1,  -1,  -1,   0,   0,   1,   0,   0,   0,   1,   0,   0,   0,  -1 },
      { -34,  -4,  -7,  14,   1,   4,   0,   1, -13,  62,   3, -19,  -2,  -5,  -1,  -2, -31, -60,  43,  15,   3,   2,   1,   1,  37, -17, -39,   8,   1,   2,   0,   1,  -5,  17,  -1,  -9,   1,  -2,   0,   0,   6,  -5,   0,  -1,   0,   1,   0,   0,   0,   2,   2,  -2,  -1,   0,   0,   0,   2,  -2,  -1,   1,  -1,   1,   0,   0 },
      {   5, -37,   9,   2,   6,  -2,   1,   0,  12, -48,  19,   6,  -4,   4,   0,   1, -21, -23, -47,  26,   5,   3,   2,   2,  44,  41, -39, -31,   2,  -1,   0,   0, -19,  25,  28,  -9,  -8,  -2,  -1,  -1,   2,  -4,   2,   3,   1,   0,   0,   0,  -3,   3,   3,  -2,  -3,   0,  -1,  -1,   1,   1,   0,   2,   0,  -1,   0,   0 },
      {   7, -10, -39,  26,   0,   2,   0,   1,  51,  19,  20, -42,  -3,  -3,   0,  -2,  40, -27,   4,  26,  -7,   1,  -2,   1, -54,  40, -10,  -5,   7,  -1,   3,  -1,   5, -17,  19,   1,  -4,  -1,  -1,   0,  -7,   4, -12,   5,   5,   1,   0,   0,   1,  -3,   4,  -4,  -1,   1,   1,   0,  -2,   2,  -1,   2,   0,  -1,   0,   0 },
      { -16,   8, -11,  -4,   5,   1,   1,   1,  -7,  53,   1,   6,  -9,  -4,  -3,  -2, -26,  11, -29,  -8,  13,   0,   4,   0,  25,  30,  48, -38, -12,  -4,  -3,  -1,  -9, -40,  36,  39,  -5,   1,  -1,   0,   9,  -9, -26,   7,   8,   1,   0,   1,  -6,   2,   2,  -1,   1,  -1,   0,   0,   4,  -3,  -4,  -2,   4,   0,   0,   0 },
      {  26,  48,  28, -32,  -3,  -4,  -2,  -2,  21,   9, -58,  27,  -1,   2,   1,   1,  22, -33,  20, -19,  -1,   0,  -1,   0,  -5,  37, -13,  -1,  -5,  -3,   0,   0, -24,  14,  32,  -7,  -1,  -3,   0,  -1,   6, -19,   8,  21,  -3,   0,  -1,   1,  -5,   1, -10,   2,   9,  -1,   0,   1,  -1,  -4,   2,  -3,   1,   1,   0,   0 },
      {  13,   5,  43, -26,   5,  -4,   1,  -2,  20,  -3,  19,  21,  -9,   0,  -1,   0, -16, -59, -10,   9,   8,   7,   3,   3, -14,  12,  15,  12, -12,  -1,  -2,  -1,   5, -24, -47,  28,  22,   1,   3,   0,  -5,  22, -17, -43,   6,   3,   2,   1,   6,  -3,  17,   0, -16,  -1,  -2,   0,  -2,   4,  -6,   5,   0,  -2,   0,  -1 },
      {  -5,  14,  37, -22,   7,  -7,   2,  -2, -15,  13,  11,  12, -10,   5,  -2,   0, -36,   1, -10,  34,   5,  -3,  -1,   0, -49, -14, -19, -28,  24,   3,   4,   1,  58,  -8,  21, -16, -28,   0,  -3,   0, -10,   6,   4,  30,   2,  -4,  -2,  -3,   6,  -5, -12,  -6,  10,   2,   1,   1,  -1,   0,   2,  -2,  -2,   1,   0,   1 },
      {   8,  -1,  50,  14,  -8,  -7,  -2,  -3, -16, -39,   8, -39,  17,   4,   4,   2,  -4,   3,  47,  16, -25,   1,  -5,   1,  15,   0,   3,  17,   1,  -4,   0,  -1, -10, -34,  22,  25,  -6,   5,  -3,   1,  11,   0, -47,  26,  12,   2,   3,  -1, -11,  10,  -6, -19,   7,   0,   1,   2,   5,   0,  -1,   1,  -2,   1,  -1,  -1 },
      {  10,   4,  24, -33,  17,  -7,   3,  -2,  27,  24,  54,  -8, -25,   2,  -6,  -1,  40,   8, -10,  18,   2,  -4,  -1,  -1,   9, -45,  11,   1,  -8,  10,  -1,   2, -28,  40,   2,  22,  -8,  -9,   2,  -2,   5, -26,  -2,  15,  16,   2,   0,   1,  -9,   6, -18,  -2,   4,   5,   2,   1,  -1,  -3,   1,  -5,   2,   1,   1,   0 },
      {  17,  30,  18,  56, -11,  -6,  -6,  -2,   1,  -3, -42, -31,  14,   8,   7,   3,   9, -19, -26,  10,  16,  -3,   2,   1,   7, -30,  21, -40,  -4,  11,  -1,   1,  11,  34, -11,  18, -19,  -6,  -4,  -1, -20,   3,  -2,  -8,  20,   2,   6,   0,   4, -11,  10, -13, -15,   2,   0,   4,  -4,   4,   1,   7,  -3,  -2,  -2,  -2 },
      {   6,  25,  40,  19, -25,  -2,  -4,  -1,   3,  32,   8, -27,  17,   1,   1,  -1,  -5,  16, -45,  26, -21,   0,   2,   3,   0,   5,   4,  14,  21, -14,  -1,  -2, -34, -15,  -9, -40,  22,  14,   1,   1,  27,   3,  21,  -7, -30,  -2,   0,  -1,  -9,  -3,   6,  20,   7,  -5,  -2,  -1,   3,  -3,  -6,  -3,   4,   4,   2,   0 },
      {  -4,   9, -13,   2,  15,  -7,   1,   0, -15,  -1, -23,  25, -21,  11,   0,   2,   1,  26,  10,  35,  20, -11,   0,  -4, -27, -11, -31, -16,  28,  13,   6,   2, -48, -10,  -3,  -3,   1,  -2,   4,   0,  37,  -7, -30, -34,  15,   6,   4,   1,  -5,   4,  20, -25, -26,   3,   0,   2,   8,   0,   6,  10,  -6,  -6,  -3,  -1 },
      {   4,   1,  -4, -31,  40, -18,   6,  -5,  -3, -10, -21, -43, -33,  23,  -4,   2,  17,   2,  25,  20,  -4, -21,   0,   0,  31,   6,  12, -24,  14,  10,  -3,   1,   3, -23, -12,  -2, -24,   8,   3,   2,  -2,  26,  30,  -9, -20,  -8,  -6,  -1,  -1,  -4,  11,  37,  -7,  -3,  -2,  -5,  -1,   4, -11,   3,  17,  -3,   0,   1 },
      {   3,  17,   6, -21,  -3,   3,   0,   1,  -2,  23, -11, -11,  -6,  -5,   1,   0,  12,  44,  -6,  26, -24,  -1,  -3,   0,  33,  28, -28,  24,   9, -13,   1,  -4,  30,   7,   6,  12,  28,  -3,  -2,  -2, -54, -12,  -2, -24,  27,  11,   6,   1,  16,  -7,  14, -20, -16,   5,   1,   3,  -8,  -1,   1,  11,  -5,  -1,   0,  -1 },
      {  13,   2,  -5,   4, -24,  11,  -3,   4,  31,  13,   4,  46,  15,  -8,  -1,  -3,  41,   1,   0,   4,   7,  -1,   2,  -2,  46, -27, -22, -14,  23,   4,  -3,   1,   9, -38,  -7,  -8, -20,  18,   0,   2, -11,  43, -19,  13,  -6, -14,   4,  -3, -10,  -5,   9, -12,  18,   5,   3,   3,  -3,   0,  -3,  -3,  -9,   1,   0,   3 },
      {   6,   8,  26,  -3, -16,  12,  -5,   2,   5,   4,  26, -11,  23, -25,   7,  -5,  11,  23,  20, -24,  27,  14,  -7,   0,  -4,  -7, -36, -34, -37,  20,   7,   6,   9, -14,  24,  -2,  -1, -19,   0,  -1,   4,  -3,   6, -24, -19,  13,  -6,   3,   6,  13,  18,  29, -33, -15,  -2,  -5,  -3,  -3,  -2,  21,  22,  -9,  -4,  -2 },
      {   4,   1, -19, -17,  16,   0,   0,   1,   5,  -4, -32, -14, -13,  -1,   5,   3,   7, -14, -38,  -1, -23,  10,  -9,   3,   2, -41, -16,  15, -14, -28,   5,  -4,  35,   4,  27,   8,  21, -10, -15,  -2,  35,  25, -26,  -4,  -7,  17,  -2,  -3, -23,  34,  16,   7,   4, -14,   3,  -1,  12, -11,   7,   3,  14,   5,  -1,   1 },
      { -10, -11, -14, -65, -14,  12,   5,  -2,  -2,   1,  -2, -23,  51, -13,  10,   2,  -6,  -5,   3,  -2, -28,  18,   1,  -4, -10,  -7,  19, -26,   8,  -7,  -5,   0, -15,  24,   3,  -7, -15,  12,  -8,   0,  15,  15,  -6,  -1,   4,  -3,  12,  -4,   2, -32,  23, -13, -16,  14,   6,  10,   0,   4,  -5,  22, -26,  -5,   0,   0 },
      {  -5, -10, -18, -16, -40,  26,  -4,   1,  -4, -13, -21,  16,   8, -21,   0,   0,   4,  -7,  15,  71,  -4,   5,  -1,  -2,  -1,  -9,  25, -23,  -4,  -5,   6,   1,  -3,   2, -14,  11,  17,   0,  -9,   1, -12, -27,  -3,   5, -18,   2,   4,   4,  12,  27, -19,  20,   7, -24,  -8,  -7, -10,   1,   3, -11,  27,   0,  -3,  -3 },
      {  -5,   5,   1,   0, -20,   9,  -4,   4, -10,   7,  -6,  -1,  -9,  -7,   9,  -2,  -8,  24,   4,  21,   0,   4, -11,   1, -11,  22, -21,   2, -38,   4,  12,   0, -18,  18, -18,  32,  -6, -36,  -3,   0,  15,  60,  16,  12,  25, -12, -10,  -9, -15, -28,   6,   7,  36,   1,   4,   1,   4,  -3,  -6, -18,  11,   9,  -3,   3 },
      {  -4,   2,   7,  17,  -2,  -1,   0,  -4,  -5,   3,  17,  19, -19,   7,  -7,   1,  -6,   5,  15, -17, -16,  -6,   2,  -4, -17,  -1,   3,  -1,  22, -26,  -7,   2, -22,  42,  25,  22,  -9,  33,  -9,  -3, -27,  25, -20,  -6, -22,  -9,  24,  -3,  23,  -2,  37,  19,  12, -32, -13,   2,  -4, -16,  -8,  12,  28,   9,  -3,   0 },
      {   4,  -1,   8,  13,   0,   8,  -8,   2,   4,  -7,  11,  10, -28, -11,  13,   5,  10,   5,  28,  13,  31,  -3,  -4, -12,  13,  12,  25,  -8,   4, -15,  -5,  -2,  30,  20,  13, -18,  34,  11, -27,  -1,  29,   4,  -4,   0,  -9,  39,  21,  -8, -24, -50,  -4,  11,  -9,  -1,  17,  18,   3,   9,  -9,  -6, -15,   6,  -5,   0 },
      {   0,   1,   1,   0, -36, -12,   2,  -3,   5,   8,  10,   7,   3,  13, -17,   5,   3,  10,  13,  -3,  -8, -15,  18,   3,   6,  35,   6,  -3,   9,  12, -11,  -4,  45,  32, -31,   8, -16,   6,  10,  -4,  36, -19,  -3,   3,   3, -18,   0,   5, -45,  32,  30,  -3,  -6,   3,  -9,  -4,  23,  -7,   1,  12,  12,  -8,   2,  -2 },
      {  -3,  -5,  -8,  14,   4, -51,   8,  -8,   6,   9,   7,  29,  35,  50, -12,   5,   3,   4,  10,   2, -40,  11,   9,   5,  -3, -11, -22, -30, -15, -25,   4,   0,   3,  -9,  -3,  26,  21,   1,  -9,  -3,   6,  -7,  16,  -5,   0,  12,   7,   0,  -8, -22, -26,  12, -11,   0,   3,   7,  -2,  19,  -9,  -7,  10, -12,  -6,  -4 },
      {  -5, -10,  -3,  -7, -69,  -3,  -8,   6,   0,  -5,   4,   4, -30,  31,   1,  -3,  -3, -14,  -2,  -6,  -6, -36,   1,  -1,   0, -15,  15,  15,  11,  -6,  -4,  -2,  -7,  -5,  34,  -4,   1, -21,   6,   7,  -1,  17,   7,  -1,  16,  12, -34,  -5,  14,  -8,  -6,  11, -28,  21,  16, -14,  -8,   6,   3,  21,   8, -25,   3,   4 },
      {   6,   8,   7,  31,  42,  38,  -9,   9,   2,   5,   4,  34,  -2, -21,  11,  -2,  -2,   0,   5,   9, -62,  19,  -8,  -9,   7,   9,  20,   2,  11,   9,  -4,  -1,   6,   9,  -2, -12, -22, -21,  12,   2,  15,  10,  -3,   1,  15,  -3, -14,   4,   1,  -4, -15,  17, -29,   2,   8,  -6,   0,   2,   4,  19,  16, -26,   0,   0 },
      {   0,   0,   4,  11,  11, -19,  11, -12,  -1,  -3,   8,  17,   3, -12, -36,  -7,   2,  -3,  20,  13,  -3,  13,  -4,  13,   1,  -2,  31, -14,  -6, -16,  32,  -1, -14,  -3,   7, -26,  13, -20, -21,   8, -27,  12,  30,  11,   8,  21, -20, -13,  -8,  29,  39, -21,   4,  16,  11,  -7,  31, -24, -10,  14, -12,   4,   5,   2 },
      {  -1,   2,  -3,  -5,  10, -53,   9,   3,   4,  11,   5,   6,  30,  15,   7,  -9,   4,   9,   6,  16,  18,   0, -14,  -9,  13,  19,  24,  10, -13,  12,  -2,   2,  10,  24,  -1, -36,   1, -19,   7,   9,  12,  24, -42,  -3, -11,   0, -13,  -5,  36,   5,  -7,  -5,   8, -11,   3,  -3, -28, -10,  35,   2,  13,   4,  -5,   6 },
      {  -3,   4,   0,  -8, -10,   7,  -4,  -2,  -1,  10,   5, -12, -17,  13,   7,   5,  -5,   4,  -9, -20,   1,   4, -14,   4,   2,  18,  -7,  -5,   4,  -1,  19,   2, -15,  -1, -38,  -4, -19,   5, -35,   1, -15,  18,   0,  24, -12,  28,  25, -21,  -4,  29, -28, -21, -20, -29,  27,  20,  12,  16,  30,  12,   7, -27, -25,  -6 },
      {  -4,  -1,  -7,  -3,  -4,  -4,  -4,  -8,  -5,   1,  -9,   5, -14,  24, -12,  15,  -2,   2, -10,  18,   7,  37,  21,   0,  -3,  -7,   5,  32, -42,  17,   4,  -5,  -7,  -4,   7, -32, -18,  -7,   8,   3, -36, -11, -33,  17, -18, -14,   7,   6, -34, -31,  19,  10, -14, -12,  -6,  -1,  14,  26, -16,  11,   6, -13,  -2,  -8 },
      {  -4,  -4,  -6,  -7,   7, -21,  -3,  -3,  -1,   0,  -3,   7,  43, -20,   6,  -9,  -5,  -7,  -6,  12,  21, -17, -24,   7,  -4,  -2,   9,  33,  29,  29,  -9,   0,  -4,   5,  29,  18,  -3,  11,  17,  -5, -22,  17,  18, -14,   8,  15,   6,   5, -34,  -6, -17,   8, -12, -14,   4,   7,  49,  -4,   1, -12,  28, -11, -14,  -1 },
      {   3,   5,   2,   6, -18,  -4,  18,  -3,   0,   1,  -5,  -3, -20,  -5, -38,  -5,   5,   7,   5,  14,   9,  37, -14,  -4,   5,   4,   6,  10, -18, -12,  14,  16,   8,   9,  20,   7, -16,  29,  22, -10,  21,  31,  15, -21,  -7, -13,  21,  16,  23,  14, -40,  -5, -16,  11, -25,   5,  20,   5,   8,   9, -32, -11,  14, -13 },
      {  -3,   4,   3,  -1,  12,  23, -17,   3,  -5,   5,   4,   0,   1,  21,  11,  -8,  -6,   6,   9,   5, -15, -32,  -2,  10, -13,   3,   1, -28, -32,  14, -29,  -3,  -9,  17,   6,  -9,  35,  20,  21,  -5, -30,  32, -12,  -4, -28,   2,  -5,  13, -37,  25, -19, -14,  -5,  10,   3, -16,  -6,  -6,   8,  -5, -31,  -4,  16,   3 },
      {  -2,   4,   1,   0,  -2, -18,  16,  -7,  -7,   1,  -9, -10,   0, -33, -26,   8,  -2,   9,  -3,  -9,  14,  30,  13,  -5,  -7,   8, -16,  -3,  27,  -9, -18,   1,  -9,  16, -16,  11,  -5,  11, -15,   9, -16,   9, -36,   0,  -2,  20, -30, -10, -16,  10, -19,  39,  -3,  30,  25, -30, -14,   9, -10, -18,  -3, -19,  26,  11 },
      {  -2,  -1,  -5,   6,  -2,   1, -12,   2,  -1,   2,  -4,  14,   8,  28,   6,   0,  -3,   1,  -5,  18,  -1,  15, -26,   1,  -2,   4,   1,  25,  -7,  39,  10,  -9,   5,   1,   8,  24, -20,  17,   3,   2,   8, -11,  13,  -2, -15,  19,  18, -23,   4,  25,  14,  21,   2,  13,  43,  19, -37, -22, -38,   9, -29,   9,   6,  16 },
      {  -1,   1,  -2,   3,  13,   5,  22, -23,  -2,   0,  -2,   3,  15, -15, -19,  11,  -3,  -1,  -5,  10,  17, -25,  25,  -3,  -3,  -1,   0,  19, -17,  -2, -16,  11,  -6,  -8,   4, -10,  12,  15, -19, -17,   8,  21,  17,  32,  35,  -3,  23,   9, -11,  13,   3,   2, -23,   5, -14,  15, -59, -15,  -7,  14,  20, -23,  -3, -13 },
      {  -5,  -4,   1,  -3,  -9, -10, -22, -12,  -3,   0,   8,  -2,  -7,   7,  27,  13,  -2,  -3,   9,  -9,   5,  37,  -5, -11,   0,   1,  12,  -4,  32, -21,  26,   2,  -4,  -2,  10,  -2, -11, -22,   9,  11, -16,   4,  12, -28,   7,  -9,  -1,  15, -56,  13,  -7,   1,  -2,  -1, -23,  10, -51, -16,  20, -10,  -7,  13,  -7, -18 },
      {  -2,  -3,  -4,   4, -18, -37,  19, -22,  -3,  -3,  -3,   1, -33, -46,  -6,   7,  -4,  -6,  -6, -15, -35,  -4, -45,  15,   1,  -6,  -2, -18,   4,  49,  16, -18,   0,  -1,  -4,   0,  31,  10,  17,   2,   1,  -3,  -6,  17,  -9,  -6,   5,   6,  -6, -16,   6,  -8,   2,  -3,  -7,   4, -23,  15,  -2,  11,   2,   1,  -5,  -8 },
      {   2,   4,   3,  -4,   5,  46,  24, -23,   2,   4,   4,  -4,  22,  23, -43,  12,  -1,   0,  -1, -12,  -2, -29,  -1,   8,   5,   6,   5,   2,  20,  19,  47,  -6,   6,  15,  10,  -5,   1, -25,  -5,  16,   6,  12, -21, -18,  -6,  10,  -1,   6,   2,   3, -11,  -3,  -2,  -9, -11,   2, -10,  31, -30, -28,   2,  17,  -1,  -8 },
      {  -1,   1,  -1,   2,   9, -21, -37, -17,   2,   7,   4,  12,   6, -23,  38,   5,   1,   6,   3,   9,   7, -17,  -1,  25,   0,   4,   4,  12, -10, -22,   2,  -7,  -2,   9,   7,   2, -17,  11, -12,   6,   7,   7,   7,   1,   1,  -4, -11,  -4,  11,  33,  10,  -1,  -4,  15, -15,  -3,  -3,  81, -19,   7,  -7,  13,   6, -11 },
      {  -2,  -1,   0,  -3, -17, -18,  -7,  -3,  -1,   0,   1,  -2, -16,  -7,   7,  -4,  -1,  -1,   4, -11, -24,   4,  22,   4,  -1,  -1,   5, -16,  -2,  23, -29,   6,  -4,  -5,   5, -38,   0, -26,  -9, -14, -10,   8, -15, -22,  41,  -1,  41,  -2,  15,  31,  -9,  30,   7,   1,   3,  33,  21,  -2, -23, -36, -12,  10,  -6,   0 },
      {   0,   0,   4,   4,   1,   8, -13, -21,   0,  -1,   2,   1,   1,  -7, -33,  36,   1,  -2,   5,   6, -11,  12,  36,  -9,   0,  -4,   5,   4,  -2,  26, -24, -14,   0,  -6,  24,  13,   0,   0,   3,  -3,   6,   5,  26, -20,  -7,   8,   4, -18,   4,  -1,  19,  -9,  19,  -3,  25,   2,  -7,  31,  69, -11,   1,  10,  -2,  11 },
      {   1,  -1,   0,  -4,   5,  20, -31, -48,   3,   2,   3,  -7,   3,  23,   6,  39,   3,   3,   5,  -6,  22,  23, -27,  12,   2,  -1,   0, -13,   2, -11,   7, -25,   4,   4,  -3, -13,  29,  26,  23,   7,   5,   2,  -8,   8,  30, -29,  -4,  -5,  22,   5,   0,   8,   5,  19,  16,   2,  18, -16,  -4, -10,  27, -24,  -6,   5 },
      {   2,   3,   5,   7,   0,  10,  67,  -1,  -2,  -2,   1,   2,  -3, -10,  17, -15,  -1,  -1,   5,   5,   2, -20, -11,   7,  -3,  -4,   2,  -3, -20, -35,   4,  -9,   1,   0,  10, -11, -15,   8,  19,  11,  -9, -16,  -3, -32,   2, -29,   4, -21, -22,   7,   5,   9,  19,  11,  33,  23,  -7,  18,  11,  -4,   0, -24, -29,  18 },
      {   0,   0,   2,  -6,   5,   0, -22,  50,  -3,  -4,  -2, -11,  -4,   5, -27, -49,   0,   1,   4,  -6,  11,  32,  14,  16,  -1,  -2,   6,  -8,  18,  -3,   6,  -3,   0,   7,   5,  -3,  24,   5,  35, -14,   6,  10,   7,  13,  10,  -4,   7,   5,  -3,  10,  13, -22,   0,  -6,  20,   5, -26,  35, -18, -13,  28, -10, -20,  18 },
      {   0,   0,   2,  -3,   0,   6, -14,  43,   0,   0,   2,   1,   7,  13, -28,   4,  -4,  -5,  -1,  -3,  -8, -12, -60, -17,  -1,  -1,   0,  -3, -18,   0,  10,  15,  -3,  -6,  -3, -19, -12,  31, -23, -12,  -7, -14,  -6,  -5,  33,   7,   7,  -6, -13,  -3,  23,  23,   1,  23, -16,  -9,  -5,  17,  32, -24,  14,   9,  18, -11 },
      {   0,  -1,   1,  -1,   7, -12, -44,   6,   0,  -2,  -1,  -2,   0, -18,  -9, -21,   0,  -1,  -2,   1,  -9, -21,  26,  14,   1,  -1,  -1,   6,  -6,  14,  43,   3,   7,   6,   9,   8,   2,   7, -31,  13,   4,  -7,  -9, -16, -20, -25,   0,  -8,   5, -12,   4,   5,  12,  15,  -8,  12,  -4, -22,  11, -25, -16, -67,  -6,  -9 },
      {  -1,  -2,  -3,  -3,  -7,   2, -25, -41,   0,  -2,  -3,   0,  -8, -11, -23, -35,   1,   1,   0,   1, -11,  -5,  -4,  51,   1,   0,  -2,   1, -22, -18, -14,   2,   3,  -5,  -7, -17, -42,  22,   6,   0,  10, -10,  -4, -23,  13,  23, -10,  -8,  -6, -24, -12, -18,  -8, -26,  14, -17, -18, -24,  -2, -14,  13,  22,   0,  -3 },
      {   1,   1,   1,   3,  -1,  18,  31,  -6,   0,   1,   2,   5,  -2,  25,  16, -20,   2,   3,   4,   4,   9,  27, -21,  65,   3,   0,   4,   7,  19,   8, -36, -12,   1,   0,   3,  11,   1, -23, -22, -14,   3,  -1,   4,  13, -16,   9,   5,   8,  10,  -3,  26,   7,  -4,  18, -27,  -4,  -4,   2,  23, -27,  -9, -23,   2, -20 },
      {   0,   0,  -1,   1,   3,   8,   4,  13,  -1,  -1,  -2,   0,   2,   7,   0, -21,   0,   1,   0,   0,   5,   6,  27,  44,  -1,   1,  -2,   0,   7,  20,  25, -12,  -1,   2,   1,   2,  18,  18, -24,   6,  -3,   4, -11,   0,  18, -26,  -2,  -7,  -9,  -5, -27,  36,  -4,  21,   5,   8,   8,  -4,  30,  51,   6,  53,  -7,  -2 },
      {   0,   0,  -1,   0,   4,   3, -25,  15,  -1,  -2,  -2,  -3,   2,  -4, -42,  21,  -1,   0,  -1,   1,   7,  -2, -29,  12,  -2,   0,  -1,  -3,   5,  -8, -49,  12,   0,  -1,  -1,   0,   1, -31, -11,   9,  -3,  -6,  -6, -21, -22, -11,  -2, -28,  -8,  -2, -23, -15,  29,  22, -16,  26,  -7,  -1, -24,  42,  12,  -5, -25, -16 },
      {   0,   0,  -1,   0,   4,  -4, -17,  17,  -1,  -1,  -3,   1,   7, -15, -28,   9,   0,   1,  -2,   4,   5, -24,   1,  10,  -2,  -1,  -9,  -2,  -3, -33,  -1, -43,  -5,  -1, -16,   8, -10, -15,  48,  25,  -4,   2, -10,  30,  -4,  25,   6,  15,   9,  -1,  12,  37, -23,   2,  -7,  35,   3, -11,  21,  -7, -30,  10, -20, -12 },
      {  -1,  -2,  -2,   1,   3,   7, -18, -57,  -1,  -2,  -3,  -1,   3,   4,  -4, -50,  -2,  -2,  -5,  -2,   2, -10,  -7, -22,   1,   1,  -4,  -1,  10, -10,  11,  39,   0,  -3, -11,   6,   6, -28,  25, -26,  -5,  -8, -12,  13, -15,  13,  31,  -2,  -3,  -4,   8,  19,  15,  34,   1,   8,   6,  14,  20,  22,   7,   0,  17,  13 },
      {   0,  -1,   1,  -1,   3,  -6,  -9,   6,   0,   0,   2,  -1,   8,  -1, -12,   5,   1,  -1,   2,  -1,  12, -11, -14,  12,   1,  -1,   2,  -1,   8, -13,   5, -52,  -2,  -5,   0,  -6,  -5, -26, -11, -28,   1,  -3,   2, -12,  18, -11,  37,  17,  -8,  -4, -10,  12,  27, -39,   7, -38,  -6,   6,   0,  34, -23, -27,  35,  25 },
      {   0,   0,   0,   1,   0,  -8,  -4,   2,   0,  -1,   0,   1,  -1, -12,  -8,  -4,   0,  -1,   0,   1,   1,  -8, -11, -11,   0,  -1,   0,   4,   0, -14,  -4, -28,  -1,  -1,   4,   6,  -1, -27, -16, -11,  -2,   2,   3,   6, -31, -55,  27, -24,  -2,  -2,  -7, -21, -60,  20,  23,  -3,   1, -10,   0, -31,  20,  30,  27,   4 },
      {   0,   0,   0,   0,   0,  -5,  -2,  11,   0,   1,   1,   2,   2,  -8,  -1,  47,   0,   0,   1,   2,   0, -10,  -2,  37,   2,   2,   4,   1,   5, -11,  20,  44,   0,   2,  -2,   1,  -3,  -8,   8, -79,  -4,  -2, -12,   1, -13, -11, -30,  19,   3,  -5,   3,   6,  -7,  -8,  28,  23,  -2,  -4,   2,  -4, -12,   5,  -3,   9 },
      {   0,   0,   0,  -1,  -3,  -3,   6,  19,   0,   0,   0,   0,  -4,  -8,  19,  27,  -1,   0,   1,  -1,  -6,  -4,  21,  12,   0,   1,   2,  -1, -10,  -1,  22, -27,  -2,   0,  -2,  -8, -22,   3,  11, -17,  -1,   5,  -4, -14, -19,  43,  36,  21,   5,  -3, -10, -20,  19,  62,  -8, -20,   0, -15,  -6,   1,  45,   1,  -8,   4 },
      {   0,   0,   1,   2,  -1,  -3,  -9, -12,   0,   1,   0,   1,  -1,  -1, -13,  -3,  -1,   0,  -2,   2,  -1,  -1, -15, -11,  -1,   1,  -1,   5,  -2,  -1, -11,  15,   0,   0,   1,   4,  -2,  -5, -31,  15,   2,   0,   3,   2,  -5, -18,  13,  68,   4,   3,   6,   6,  -6,  15,  28, -42,   1,   2,   0,  -4, -16,  13, -72,  16 },
      {   0,   0,   1,   0,  -1,   2,   0,  -9,  -1,   0,  -1,  -2,  -1,   4,  -5, -29,  -1,   1,   0,   0,   1,   3,  -3, -31,  -2,   0,  -2,  -1,  -4,   3,  -8, -46,  -3,   1,  -2,  -1,  -8,   8, -24, -24,  -2,  -1,  -1, -12,  -7,   1, -38,  54,   4,   2,   0,  -9,  18,   5,  29,  49,   3,   4,  -3,   6,  16,   1,  19, -44 },
      {   0,   0,   0,   0,  -1,  -2,   0,  10,   0,  -1,   0,   0,  -2,  -2,   1,  15,  -1,   0,   0,  -1,   0,  -1,  -5,  30,   0,   0,   0,   0,  -1,  -3,  -6,  35,   0,  -2,   0,   0,  -1,  -6,  -6,  64,   1,  -1,   6,   0,  -2,  -3,  17,  50,  -1,  -3,   4, -10,   2,  -3,  20,  36,   1,   3,   7,   2,  21,  -4,  58,  33 },
      {   1,   0,   0,   0,   3,  -3,   4,  12,   1,   0,   0,   0,   4,  -2,   5,   7,   0,   0,   0,   0,   3,  -5,   7,   9,   0,   0,   0,  -1,   5,  -4,   3,  27,   3,   0,   2,   1,   7,  -8,  24,  16,   3,   0,   1,   2,   4,  -7,  28,  -5,   5,   0,   7,   4,  14,  -6,  53, -35,   1,   0,   0,   4,  -1,  -1,  13, -94 },
    },
    {
      { 109, -46,  -8,  -3,  -1,  -1,  -1,   0,  -1,  20, -14,   2,  -1,   0,  -1,   0, -29,  25,   1,  -4,   2,  -1,   0,   0,  -9,   1,  11,  -3,   0,   0,   0,   0,  -4,  -2,   3,   2,  -1,   0,   0,   0,  -2,  -1,   0,   2,   0,   0,   0,   0,  -2,   0,   0,   1,   1,   0,   0,   0,  -1,  -1,   1,   0,   1,   0,   0,   0 },
      {  17,  38, -20,  -3,  -4,  -1,  -1,   0, -83,  48,  15,  -6,   4,  -2,   1,  -1,  -4, -37,  34,  -6,   2,   0,   1,   0,  26, -31,   3,  11,  -5,   3,  -1,   1,   7,  -3, -12,   8,   0,  -1,   0,   0,   4,   0,  -6,  -1,   2,  -1,   0,   0,   2,   0,  -1,  -1,   1,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0 },
      {  24, -35,  24,  -6,   3,  -2,   1,  -1, -40, -41,  38,   7,   3,   2,   1,   1,  29, -44, -25,  28,  -5,   6,  -2,   2,   2,  28, -50,   9,   5,  -1,   0,   0,  -6,  20, -11, -19,  11,  -3,   2,  -1,   1,   6,   6, -16,   0,   3,  -1,   0,   0,   3,   4,  -3,  -6,   3,   0,   0,   1,   2,   0,   1,  -4,   0,   1,   0 },
      {  38,  65, -57,   6,  -7,   1,  -4,   0,  20, -11,  -1,  -4,   4,  -2,   1,  -1,  46, -18,  -2, -10,   1,   0,  -1,   0, -21,  46,  -2,  -4,  -3,  -2,   0,  -1, -22,   8,  25,  -5,  -1,  -2,   1,  -1,  -1, -11,  13,   9,  -2,   0,  -1,   0,  -1,  -6,   1,   5,   4,  -1,   0,   1,  -1,  -1,  -1,   0,   4,  -2,   0,   1 },
      {  14,   0,   3,  -6,   0,   0,   0,   0,  66,  -1,  20, -24,  -4,  -2,  -1,  -1,  -2, -33,  55,   5, -13,   4,  -3,   1,   5, -25, -16,  45, -14,   6,  -1,   2, -19,  23, -36,  18,   7,  -6,   3,  -1,  -9,   9,  -7,  -8,  13,  -4,   0,  -1,  -3,   2,   1,  -7,   3,   2,   0,   0,  -2,   0,   0,  -1,   0,   1,   1,   0 },
      {  15,   9,  31, -27,   3,  -6,   1,  -2, -13, -37,  50, -13,   1,   0,   0,   0,  29,  18,  -3,  -9, -12,   2,  -2,   1, -11,   5,  50,   6, -18,   1,  -4,   0,   8, -33,  -1,  47, -12,   1,  -1,   1,  -6,   9, -31,  23,  10,  -8,   2,  -1,  -3,   1, -11,  -1,  15,  -3,   0,   0,  -1,   2,  -4,  -6,   4,   2,  -1,   0 },
      {  -7,  -6,  16,  -4,  -5,   5,   0,  -1,  16,  51,  49, -34, -14,  -5,   1,  -3, -21,  -8,  12,  34, -14,   3,   1,   1,  -9,  21, -29, -16,   6,  -8,   7,   0,  10, -12,  53, -10,  -7,  -1,  -4,   3,   1, -15,  -7,  38,  -9,   3,  -3,   0,  -3,   0, -11,   9,  11,  -8,  -1,  -3,   1,  -4,  -4,   0,   9,   2,  -1,   0 },
      { -13, -60,   8,   7,   6,   2,   2,   2,  -1,  33, -30, -16,   5,  -2,   2,  -1,  68, -29,  13, -33,   3,   0,   0,   0,  22,   8,  15, -17, -19,   8,  -4,   2, -17,  15,  15,  11, -21,   1,   0,   0,  -6,   1,   2,  19,  -7,  -5,   0,  -1,  -5,   3,  -2,   9,   4,  -5,   0,  -1,  -2,   0,  -2,   2,   4,   0,  -1,   0 },
      {   0,  16,  45, -21,  -8,  -1,  -1,  -1,  -7,  55,  -8,  18, -18,  -2,  -2,  -1,  46,  44,  -4,  13, -12,  -8,  -1,  -2, -24,   9,   0,  36,   9, -10,   1,  -1, -23,   8, -12,  -4,  32,  -4,   3,  -1,  -5, -15,  18, -17,  14,   5,   0,   1,   4, -14,   2,  -5,   4,   7,  -2,   2,   0,  -3,   0,  -4,  -2,   5,   1,   1 },
      {  15,  24,  42, -41,   8,  -9,   1,  -3,   3, -21,  -3, -18,  15,   4,   2,   0,   3,   7,  18, -29,  -2,   4,  -1,   1,   1, -26,  -1,   2,   1,  11,  -2,   1,   8,   8,   3, -46, -13,  11,   1,   0,   7,  -8,  39,  11, -47,   7,  -1,   0,  -3,   7, -13,  38, -18, -10,   5,  -1,  -3,   3,  -1,  10,   6, -11,   1,  -1 },
      {  22,  31,  28,  -3, -13,   2,  -4,  -1,  24,   4,   3,  42, -25,  -1,  -4,   1,   7,  -6, -31,  17,  15, -10,   4,  -3,  67, -23,  -3, -25,  13,  -7,   0,   1,  -7,  29,   3,  10, -16,  13,  -7,   2, -18,  18,  -7,  22,  -2,  -4,   5,  -3,  -5,  -2,   4,  -1,  19, -11,   1,  -1,  -1,  -4,   4,  -7,  12,  -1,  -1,   1 },
      {  10,   4, -13, -17,   9,  -4,   4,  -2,  28,  32,  17, -26,   4,   0,  -2,  -2,  30, -15, -35,  17,  -1,   3,   2,   1,   3, -23,  15, -29,  13,   6,  -2,   1,  17, -39, -12,  -9,   8,  13,  -2,   3, -15,  21, -11, -44, -11,  14,   2,   0,  -5,   3,  20,  -1, -41,   7,   3,  -1,  -2,   4,  -1,  20, -19,  -2,   3,  -1 },
      {  11,  -6,   2,  38, -17,   2,  -4,   0,   4, -14, -20,  26,   0,  -2,   2,   0,  40, -15,  24,  25,   2, -10,   1,  -1, -37, -44,  -2,  11,  36,  -1,   1,   2,  29, -32,   3, -24,   2,  13,  -5,   2,   1,  -1, -18,  26,  -9,   7,   2,  -2, -10,  12, -26, -10,  14,  -2,   4,   0,  -5,   4,   2, -14,  -2,   6,  -1,   3 },
      {  -2,  -5, -39,   5,  -3,   9,   3,   0, -15,  -9,   2, -37, -13,   6,   4,   2,  28,  57,  -7,  12, -40,   2,   0,  -1,   7, -27, -28,   4,  -9,  -5,  10,   2,  33,  27,  -9,  -9, -17, -16,  -1,   2, -23,  31,  12,  12,   1, -16,  -3,  -1,  -2,  -1,  13,  -1,  11, -10,  -5,  -3,   0,  -4,  11,   4,   7,   2,  -5,  -1 },
      {   6,  11,   4, -36,  17, -17,   5,  -1,   6,   0, -60,  -8,   8,  10,  -2,   2,  -4, -25, -24,  25, -36,  13,  -4,   3,  32,  13,  -1,  30, -17,  -8,   2,  -7,  24, -27,  15,  -2,  13, -27,   6,  -1,  14,  -2,  -9,   9,  12,  -8, -11,   3, -15,  24, -13,  -2,  12,   7,  -1,  -1,  -1,  -1,   6, -10,   0,   4,  -2,   2 },
      {  -3, -20, -29,  15,  -8,   9,  -1,   4,   8,  -9,  31,   8,  -9,  -9,  -1,   0,  12,  28,  16, -14,   1, -12,   0,  -2,  67,  21,   9,  32,   9, -20,  -4,  -4,   7, -26,   2, -10,  41,   7,  -6,   5,  14,  -6,   5,  -5, -20,  26, -13,   3, -11,  11, -13,  21,  -6,   4,   3,  -6,  -2,   2,  -3,  -6,   8,  -5,   6,  -2 },
      {  12,  15,  29,  14,   8,  -6,  -2,  -2,  -2, -17, -11, -32,  27,  -2,   1,   0,  10,  24,  27, -11,  11,  -4,  -5,  -2,  34,   6, -18, -12,  21,   3,  -2,   0,  16,  10,  28,  -2,   0,   5,  -7,  -1,  -5, -21, -13,   4,  36,  17,   1,   1,   2, -12,   5, -47, -17,  32,   3,   5,  -4,  -2,   3,  13, -38,   5,   8,   0 },
      {  13,  11,  34,  43, -48,  12,  -8,   1,   7,  -4,  -1,  -3, -19,   1,  11,  -1,   2, -17,  -9, -20,  -2,  -5,  14,   2,  -6,  -9,   1, -13, -35, -14,  12,   5,  36,  -4,   6,  14,  11, -44,  -3,   6,  11,  -8,  18, -16,  18,   0, -23,   9,   3,   2,   7,  22, -20,   7, -14,  -6,  -5,   7,  -7,  15,  10,  -2,   1,  -6 },
      {   5,  16,  27,  38,  -6,  -8,  -5,   0,   9,  27,   6,  -6,  46, -14,   2,  -5,  -5,   7, -14, -36,  12,  13,  -4,   3,   0,  33, -22,  15,  -8,  11,  -5,  -1,  16, -20, -26, -17,  -1,  10,   6,  -1,   1,  43,  -9,  -1, -16, -17,   4,  -3, -12,  12,  16,  -6,  30, -11,  -2,  -1,   5, -14,  15, -15,   1,  14,  -6,   3 },
      {   1,  12,  16,  14, -24,  19,  -7,   2, -12,  -2, -29,  -4, -25,  27,  -6,   2,  -3,  12,  19,  12, -19,   1,  -4,  -2,   3,  36, -17,   9,  -2,  14,  -1,  -3, -10,  -9,  -9,  17, -20,  20,   8,  -3,  -9,   6, -38,  -7, -19,  12,  19,  -1,  -8,  26,   2,  -2, -41, -27,  20,   0,   0,   7,   4,  21,  12, -38,   1,   0 },
      {   6,   5,  -5,  38,  32, -16,   3,  -2,   8,  12,  16,   8,  22,  -4,  -6,   1,  15,  25, -25,  25,  16,  22, -11,   1,  12, -21, -18,  10, -22,  42,  -1,  -1,   1,   0,  -6,  18, -19,  -7,  21,  -9,  21, -46,   3,  -3,   7,  -5,  12,   5,   1,   8, -28,  14, -15,  -5,   1,   6,  -8,  15,  -9,  -5,  16, -11,  -1,   0 },
      {  -2,  -8, -20, -56, -12,  11,   1,   0,  13,  15,   8,  29, -15,  -5,  -2,  -2,  10,   2,   1, -32,  32,  -5,   5,   0, -13,  14, -34,  -3, -12,   6,  -5,   3,  65,  15, -15,  12,  -6,   4,   0,  -6,  19, -13, -10,   5,   7,   8,  10,  -1,   2,  17, -14, -19,  -1,  -1,   5,   4,  -8,  10,  11,  -3,  -8,  -5,   1,   1 },
      {  -3,  -1,   2,  15,  12, -26,   2,  -4,   1,  10,   7,  32,   6,  -3, -20,   0,  -4,  -6,  26,  30, -11,   4, -23,   2,  -6,  33,  39,   5,   5,  12, -15,  -6,  49,  29,   1,   2, -12,   3,  -3,  -7, -13,  34,  34,   7,  15,   8,   0,  -9,  -9,  -5,   4,  20, -12,  23,  16,  -1,  -5,  -3,  -7,  15,   0, -10,   8,   2 },
      {   2,   7,   5,  -3,  42, -22,   4,  -3,  -6,  -3, -11,  -8,  14, -21,   4,  -2,   6,   7,  15,   8,  31, -17,  -3,   0, -17,  -3,  -9, -20,   3, -16,   2,   4,   3,  21,   6,  15,  31, -21, -22,   5, -15,  12, -29,  -9,   3,  30, -24,   4,   1,  27,   5,  11,  -8, -20,  -3, -15,   8, -11,  19, -11,  50, -13,  -8,  -2 },
      {   6,   1,   8,   2,   5, -11,   9,   0,   2,  -7, -13, -39, -29,  -8,   5,   2,   8,  -2,  15,  24,  15, -34,   1,  -4,  11,  22,  16, -12,  32,   5, -13,  -1,  20,   0, -27,   0, -12,   7,   4,  -7,  58, -13,   7, -18,   1, -28,  12,   0,  16,  -5,  25,   7,  19, -16, -16,  10,  -3,   7,  -3,  -3,  19,  12, -16,   5 },
      {  -1,   0,   0,  11, -23,   7,  -4,   0,  -5,   0, -14, -47,  12, -14,   1,   0, -13, -17, -45,   0,   1, -10,  -2,   0,  -7,  -5,  13,  28,   2, -11, -23,   0,  20,  33,  -1,  24,  23,  38, -18, -12, -25, -21,   8,   5,  -6,  26,  28, -14,   8,  -3, -18,   9,   5,  -8,  18,   1,  -5,   3,  -9, -11,   5,  11,   3,   5 },
      {  -3,   4,  -3,  -1,  -9,  -1,  -4,   7,  -3,  11,  -2, -28,   0,   8,   1,   2,  -4,   5, -25,   1,  17, -26,   4,  -1,  -8,   7,  -6,   0,  22,   4,  -5,   3, -27,   5, -40,  -2, -25, -14, -13,  10,  25,  26,  -3,  32,  14,   9, -39,  -5, -22,   6, -34,  11,  -8,  41,  12, -14,  -7,  19,  -8,  -9,  -7, -19,  19,  -9 },
      {  -2,   2,  -4,   4, -36, -32,  19,  -5,   1,   5,  10,   8,   6, -29,  15,  -4,  -1,  10,   3,   4, -42,  14,  -2,   7,   0,   1,  25, -31, -12,  -1, -22,   5, -14,  42,  -7, -27,  16,   1,  -3,  -1,  24,  -6, -37,  -5,  -8, -11,   4,   1,   6,  34, -23,  -9, -10,   0,  -8,  -5,  -2,  12,  13,  -1, -18,  10,   1,  -4 },
      {  -8,  -7,  -8, -24, -52, -13,   5,  -2,  -8, -10,  -4,  -4,  11, -31,  14,  -3,  -4,   9,   9,  25,  35, -30,  -6,   6,   5,  -2,  -7,   8,  -8,  34, -21,   6, -17, -20,  20,   2,  -8,  -7,  13,  -5, -35,   0,  22, -12,  15, -16,   3,   3, -38,  23,   3,   7,  -4,   3, -11,   1,   1,  -7,   7,   7,   4,  12, -12,  -2 },
      {  -2,  -5,   1,   0, -12,  29, -19,   7,   2,   4,  22, -15,  16,  34, -26,   0,   0,  -7, -11,   2,  17,   3,   5,  -8,  -2,  -6,  45,  14,  20,   1,  19, -13,   2,  30,  17, -22,  -2, -13,  25,  -1,  12,  -5,   9, -11,  13,  -6,   0,   9, -12,  35,  -2, -28,  -3,  -6,   0,   7,  13, -13,  40, -13,  13, -20, -15,   1 },
      {   2,  10,   6,   8,  21,  21, -10,  -2,  -8,  -5,  -5, -16, -39,  24,  -6,   0,  -2,   7,  17,  11,  22,  18,   2,  -3,  -2,  13,   7, -33, -22,  20,   7,  -1,   3,   9, -13,  -8,  22,  -6,   7,   1, -21,   0,   1, -17, -11,   6,   9,   1, -33,  22, -19,  12,  17,   9,   4,  -4, -23,  -4, -10, -34, -25,  46,  15,   4 },
      {   0,  -2,  -1, -17, -30,  31, -19,   8,   0,  -6, -11,  -2,  27, -10, -10,  -1,  11,  14,  10,  39,  24,  27, -14,  -4,   0, -14,   1,   0, -32,   6,   4,   0,  -8,   6,  19,   6,   3,   0, -12,  -4,  56,  29, -10,  -3, -11,  24,   0, -11,  16, -16,  24,   7,   7,   5,  19, -10,   7, -11, -11,  -4,  -2,  10,  17,   4 },
      {   3,   4,   3,  18,  26, -14,   2,  -2,  -2,  -3,   8,  -3, -60, -15,  -9,   5,  -2,  -5, -16, -24,  -5, -17, -23,   7, -10, -21,  -6,  18, -21,  18, -24,   1, -15,  -9,  33,  -5,  -5,   6, -15, -14,  29,  19,  17,  -1,   8,  16,   5, -20,  -4,  28,  13, -18,   1,   3,  16,  -6,  21,  -8,  25,   4, -14,   2,   4,   4 },
      {  -1,   4,   1,   4,  21,   5,  -4,  -2,  -3,   4,   4,  12,  25,   2,  -7,   4,   0,   7,   2,   5,  -5, -19,  29,   4, -11,  -8,  -9,  14,   2, -43,  12,   8,  -8,  16,  22,  32, -14,  22,  -7,   4,  24,  11,   9,  -9, -12, -10,  14,   9, -32,  48,  18,   8,   0,   6, -19,  11, -31,  18, -19,  24, -19,  18,   1,  -7 },
      {   0,   0,  -2,  -5, -12,   1, -21,   4,   1,   0,  -1,   0,   0,  20, -47,   9,   3,   3,   0, -14,  -3,   7, -29,   5,   0,  -4, -17,  -1,  25,   7, -21,  -8,  -1,  -4,   7,   9,  -2,  13,   3,  -6, -16,  -4, -13,  -9,   4, -18, -21,   4,  56,  38,  -3,  26,   8,  28, -27, -13,  11,  -9, -14,  13,  13,  29,  11, -13 },
      {   4,   5,   4,   5,  22,  27,  16,  -4,   2,   3,  -4,  -1,  -8, -10,  32, -14,   0,   1,  -4,  10,  27, -25,   8,  -5,   3,   8,  12,  33, -12,   3,  -2,   5,   8,  14,  25, -16,  -5, -12,  21,   6, -11,  39, -21,  -7, -10, -14,   7,  18,  47,   0, -36,   7, -11,   0, -14,  13,  24,   9, -20,   6, -26,  -1,   0,   0 },
      {   0,   1,   2,   4,  12,  11,  -4,  18,  -3,  -1,  -2,   0, -14, -32,   0,  14,  -2,  -5,  -8,  16,  14,   3,   1,  -5,  -4,   5,  18,  27, -17,   6,  -4,  -6,  10,   8,  -8, -10, -16,   7,  -2,   8, -19, -36, -18,  15, -37, -11, -40,  10,   5,  -2,  41, -11,   1,  10, -21, -21, -34,  -7,  26,   6, -17, -14,  27, -13 },
      {  -3,  -3,  -1,  -2, -20, -45,   1,  -9,  -5,  -3,  -6, -18, -28, -34, -21,  12,   0,   0,  -2, -12,  29,  39, -21,  -2,   0,  -2,   2,  11,  16, -17,  50, -18,  -5,   0,  10,  -2,   1,  -8,  18,   3,  -8,  20,   2,  -1,  -3,   1,   8,  10,   9,   8,   2, -11,   7,  -7,   6,  13, -21,  36, -24,  -4,  -2, -18,   8,   4 },
      {  -2,  -4,  -6,   4, -20, -51,  36, -22,   1,   0,   4,   0,  -9,  44,  14, -19,   3,   3,  -4,  -1,  18,   9,  30,  -2,  -2,  -1,  -1,  24,  12,   3,  11,   7,   5,  12,   8,  15, -17,  -3,   5,  20,   5, -14, -13,  -2, -21,   7,  -8,   3,   4,   5,  22,  12,  -4,  11,  25, -12,  17, -43,  -2,  -8, -10,  19,  10,  15 },
      {   0,  -1,   0,   2,  -2, -16, -23,  -1,   3,   2,   3,   2,  14,  -5, -18,  -2,   1,   2,   9,  21,   0, -29,   8,   1,   9,  14,   6,  -6, -31, -19,  29,   8,  -5, -19, -23,  -8, -35, -10, -25,  19, -18, -17,  28,  -9,  -9,  14,  -2, -20,  45,  26, -15, -23,   5, -13,  27, -10,  11,  14,  10,  -4, -15,  17, -14,   6 },
      {  -3,  -3,  -4,  -3,   0, -23, -25,  14,   1,   4,   9,   6,   5,  36,  11,  -3,  -3,   1,  -4, -11,   3, -21,  10, -14,   1,  -5,   2,  16,  -5,  16, -22, -10,  -5,   3,  24, -28,   5, -19,  -4,  -2,  -3,  12, -26,  -9,  13,  -4,   2, -21,  16, -14,  10,   1,   4, -15,  23,  -7, -73,  24,  -5,   9,  17,  15,  -8,   7 },
      {   2,   1,   5,   0,  -4,  22,  40, -21,   0,   0,   1,  -2,   8,   8,  16, -14,   1,   1,  10,   2,  -7,   7,  -1,  13,   1,  -3,  -5, -11,   0,  -9, -17,  13,  -2, -12,  18,   7,  -8,  22,  -1,   3,  -5,   6,  25, -35,  10, -17, -14,  -9,  25,  18,  -1, -10,  22,  12,  12, -11, -34,   8,   4, -44,  12, -47,  30,  11 },
      {   1,   2,   1,   0,   3,  41,  40, -21,   0,   4,   3,   6,   5,   1,  34,  -5,  -3,  -3,  -7,  -2,   3,  17, -38,  19,   3,   4,  11,   5,  25,  14,  16,   2,  -9,  -2, -17,   3,   3, -25,  -7,   4,  -1,  -6,  10,  31,   4,  24,  -1, -14,  15,  32,  21,  -3,   1,  -3,   8,   7, -24,   8,   3,  34,   6,  30, -19,   7 },
      {  -1,   1,   0,   3,  -9, -13,  14,  -3,   0,   3,  -2,  -3,  -6,  40,  -1,  20,   1,   2,   4,   9,  24,  32, -14,  19,   2,   7,   3,  16,  -8, -10, -21,  29,   0,  -1,  11,   4, -11,   9, -40,   5,   2,  13,  10,  -8, -15,   7, -11,   6,  -6, -26, -17, -21, -15,  -8, -43,   2,   1,  34,  23, -11,   7,  17, -26, -25 },
      {   0,   0,   1,  -3,  -1, -10, -14,  28,   0,   0,  -2,   5,   1,  -6,  38,   2,  -3,  -3,   4,   0, -25,   6,  27, -33,   1,   4,  -4,  -3,  18,  44,  27, -13,   0,  -1,  11,  34,   9,  22,  11,  -2,   4,  19,  26,   8,  -5,  19,  -7,   8,  24,   8,   5,  -1, -16,  12, -14,  -4,   2,  27,  33, -28,   7,  22,   4,  -7 },
      {   1,   2,   1,  -5,  -1,  10,   2,  -5,   1,   3,   2,   8,   5, -10,   3,  17,  -4,  -3,  -5, -14, -30, -29, -21,  29,  -1,   1,   6,   9,   4,  28,  26,   0,  11,  18,  14, -10, -32,   9,  11,  17,   1,  -8, -21, -39,  -4,  31, -24,  10, -25, -14,   5, -16,  15,  -3,   0, -13,  15,  15, -40, -20,  24,  20,   1,  -7 },
      {  -1,  -2,  -1,   0,  -5,  -2, -14, -26,  -1,  -1,  -2,   0,  -1, -15,  -9, -10,  -3,  -5,  -4,  -5, -15, -21, -36,  -1,   0,   0,  -7,  -2,   7, -14,  11,  11,  -5,  -3,   0,  18, -21, -13,  23,   1,  13,  11,  -2, -10, -17,  18,  13,  13,   8, -23, -28,   7, -19,   5, -14,  13, -52, -53,  35, -19,  -7,  10,  -7,  -9 },
      {  -1,  -1,  -3,  -2,   5,   6, -16, -36,  -3,  -4,  -5,  -4, -16, -22,  -1, -25,  -1,  -1,  -4,  -1,   0,  31,  34, -26,   2,   7,   8,  13,  -6,  17,  20,  24,  -2,  -3,  -4, -27, -13,  39, -27,  19,   5,   2,  -6, -11,  42,   7,  -1,   2,  -1,   3, -19,   5,  11,   7, -14,  -3, -12, -22,   2,  33,  30,   3,  -3, -10 },
      {   1,   1,   1,  -4,   4,   4, -15,  59,   1,   1,   2,   4,  -5, -19,  28,   5,   1,   0,   0,  -9,  -4,  34, -13, -15,   1,   3,   2,   9,  38,  -1, -10,  18,   3,   3,   7,  11, -25, -31, -42,  16,  -2,  -6,   5, -31,   0, -17,   8,  -7,  -7,  14, -15,   3,  -1, -16,  16,   3,   0, -31, -20,  -8, -12,  -5,  -1,  -9 },
      {  -1,  -1,  -2,   0,  -6, -13,   7,  30,   1,   2,   4,   0,  -2,  15,  24,  46,   2,   2,   0,   3,   3,  -8,   6,  21,   0,   1,   4,  -1, -26,  -4,  18,  -1,  -3,  -6, -11, -28,  -8,  16, -10, -20,  -5,   1,  -8,  13,  23,  28,  18,   0,  17,  20,  15,  13,  -2,  24, -25,  43, -10, -36, -19, -23,   6, -20,  -9,   2 },
      {   3,   5,   6,   3,  -3,  18,  78,  36,  -1,  -2,  -1,   1,   3, -18, -46,  26,   0,  -2,   2,  -5,  -1,  -3,  17, -29,  -1,   0,  -5,   2,  -6,  -7,   9, -22,  -3,  -4,   4,  -7, -13,   3,  -4,  -1,  -1,   6,  -8, -10,  12,  -4,   6,  -1,  -2, -13, -16,  12, -14,   6,  18, -13, -10,  -2,  18,   0,  12,  27, -14,   2 },
      {   0,   1,   1,   0,   3,   6,  -7,  14,  -3,  -3,  -5,  -3,  -6, -20, -37,  13,  -1,   1,  -1,  -1,  -6,   6,  39,  54,   2,   2,   0,   1,  13,  33,  -4,  50,   1,   4,   7,  11,  25,  -7,  11,  33,   3,   9,   0,  12, -10,  -4,  -1,   0,  14,  -4,   3,   2,  -5,  12,  11,   1, -15,  -2,  -1,  -5,  -8, -13, -40,  29 },
      {   0,   0,   1,  -1,   8,  11, -25, -16,  -1,  -1,  -1,   2,  -9, -21,  21,   5,   1,   3,   5,  -3,   4,  26,  18,  14,   1,   1,   0,   8,  10, -37, -15,  -2,  -1,  -1,   2,  -9, -35, -14,  22, -22, -10,  -9, -19, -11,  -4,   2,  22, -33,  -5,  -4,  12,  20, -14,  52,  10, -14,  16,  25,  21, -18,  23,  14, -20,  11 },
      {  -1,   0,   0,  -1,  -4,  10,  -2, -29,   0,   0,   1,   5,   2,  -3, -18, -18,  -2,   0,  -2,  -5, -10,  -7, -19, -44,  -3,  -2,  -5,   4,  -3,  15,  -3, -12,   0,   5,   5,  -1,  -3,  -5, -45,  19,  -5, -10, -17,   0,  -4,  -7, -12,   7,  -6,   3,  28,  27,  -8,  15,  -6,  44,   6,  22,  -3, -39, -22, -14, -38,  29 },
      {  -1,  -2,  -1,  -3,   0,   7,   0, -31,   1,   1,   2,   4,  10,  -4, -17,  23,   0,  -3,   0,  -6,  -3, -15,  10,  11,   4,   5,   6,  10,  15,  18,  23,  30,   0,  -4,   3, -11,   6, -25, -27,  -8,  -2,  -2,  -8,  14,  17, -24,  39,  -9,   4,  -3,  20,  18, -27, -21,  -2,  -6,   8,  10,  11, -29,  -2,   9,  58, -31 },
      {   0,   1,   1,   0,   0,   1,  12,  13,  -2,  -2,   0,   0,  -4,  -9, -30, -53,  -2,   0,  -2,  -2,  -9,   6,  23, -20,   1,   2,   4,   2,   1,  -2, -32,  20,  -4,  -7,  -7, -10, -14, -28,  11, -24,  -5,   4,  11,  18, -13,  32,   7,   7,   4,  11,  19, -39, -16,  -1, -22,  12,  -5,  -1, -41, -23,  24,   9,   9, -13 },
      {   0,   0,   0,  -1,   4,  -5,  -1,  38,  -1,  -1,  -1,   0,  -4,   7,   6, -32,  -1,  -2,  -1,  -1,   4,  -4, -40,   0,  -2,  -3,  -1,  -2, -16, -24,  -2,  36,   1,  -2,  -4, -15,   2,  25,  23,  51,   0,  -6,  -3,  15,  27,  -5,  23,  39,  -2,   2,  23,  16,  -8,  11,  26,   0,   5,  20,  14, -10,  13,  12,   7, -12 },
      {   0,  -1,   0,   0,   2,   7, -13, -36,   1,   2,   2,   0,   0,  -9,  10,  58,   1,   1,   1,   2,  -2,  11,  18, -11,  -1,  -1,  -1,  -4,  -1,  -6, -50, -21,  -3,  -5,  -5,   0,  -6,  -9,   3,  29,   2,   5,  15,   7,   2,   4,  -3,  58,   9,   6,   2, -19,  -7,  -6,  36,  19,  -1,  -8,  -8,  -2,  16,  16,   9,  -2 },
      {   1,   0,   0,   0,   2,  -1,   2, -16,   1,   1,   1,   0,   0,   5,  10,  20,   0,  -2,  -2,  -1,   3, -12, -19, -33,  -2,  -1,   0,  -3, -13,   4,  11,   0,   4,   2,   1,   2,  25,   5,   1,  28,   5,   2,   8,  14, -23, -32,  28, -28,  -1,   7,   2, -34, -20,  35,  -4, -44,  -4, -18, -36,  -9,  20, -13, -33, -27 },
      {   0,   0,  -1,  -2,   5,   6,  -9,  -1,  -1,  -1,  -1,   2,  -7, -10,   4, -31,   2,   2,   2,   1,   6,   9,   6,  47,   0,  -1,   0,   5,  -2,  -9,   5, -50,   0,   1,  -1,  -3,  -2,  28, -12,  -7,   6,   4,   6,  16,  25, -29, -27,   2,   2,   0,   3,  -2, -53, -28,  11, -13, -10, -17, -18, -33,  -7,  19, -22, -11 },
      {   0,   0,   0,  -1,   1,   1,  -1,  -8,   0,   0,   0,   1,   3,  -2,   1,  31,   0,   0,   0,  -1,   2,   9,  -6, -34,  -1,   1,   0,   2,   3, -18, -17,  29,   2,   1,  -1,  -3,  -6,  14,  48,  22,   2,   1,   3,   9,  31,  15, -31, -50,  -2,   7,   5,  12, -20, -46, -24,  -8,   3,  -3, -13, -15, -23,   8,   1,  22 },
      {   0,   0,   0,  -1,   2,   0,  -7,   5,   0,   1,   2,   1,  -2,  -1,   6,   9,  -1,  -1,  -2,  -1,   1,  -6, -17, -11,   1,   2,   3,   1,  -7,  -7,  22,  40,   1,   2,   0,  -1,   8,  14, -10, -38,   0,   1,   0,   4, -16, -40, -30,   6,  -7,  -2, -12, -26, -32,  12,  16,  24,   1,  -5,  -4,   7,  34,  26,  23,  61 },
      {   0,   0,   1,   1,   0,   0,   2, -10,   0,  -1,   0,   0,  -1,   0,   2,   9,   0,   0,   1,   1,  -1,   1,   1, -19,   1,   0,   1,  -2,   1,  -2,   9,  43,   2,   2,   3,   1,   6,  10,  15, -57,  -3,  -3,  -5,  -2,   2,  -4, -34,  15,   5,   3,  10,  12,  14,  -3,  49,   5,  -2,   2,  -6,  -5, -20,  -4, -40, -62 },
      {  -1,  -1,  -1,  -1,  -1,   1,  -4,  -4,   1,   2,   2,   1,   2,   5,   4,   7,  -1,  -1,  -2,  -3,  -2,   1,  -8, -10,   1,   1,   2,   3,   6,  -2,   2,  12,  -3,  -2,  -4,  -6,  -9,  -3, -14, -36,   1,   2,   3,   6,  16,  13,  22,  58,   3,   4,   5,   7,   2, -10, -17, -78,  -5,   2,  -9,  -7, -21, -10, -19,  47 },
    },
    {
      {-106,  32,  16,   3,   3,   1,   1,   0,  40, -37,   0,   2,   1,   0,   0,   0,  22,   3, -13,  -1,  -1,   0,   0,   0,  -2,   5,   0,  -2,   0,   0,   0,   0,   4,   0,   0,  -1,  -1,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,  -1,   0,   0,   0 },
      {  -7,  35, -14,  -1,  -2,   0,  -1,   0, -75,  -6,  40,  -2,   5,  -1,   2,   0,  46, -59, -11,  15,  -1,   3,   0,   1,  11,  23, -28,  -6,   1,   0,   0,   0,   0,   4,   8,  -7,  -2,   0,   0,   0,   3,   2,  -1,   2,  -2,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   1,   1,  -1,   0,   0,   0,   0,   0 },
      { -39,   2,   7,   7,   0,   2,   0,   1,  -8,  59, -13, -14,   0,  -3,   0,  -1, -23, -35,  60,  -2,   0,  -1,   0,   0,  29, -46, -24,  29,   0,   4,  -1,   2,  -2,  18, -21, -12,   7,  -1,   1,   0,   5,  -5,   7,  -6,  -4,   2,  -1,   0,   0,   4,  -3,   3,  -2,  -1,   0,  -1,   2,  -1,   0,   0,   1,  -1,   0,   0 },
      { -25, -97,  34,  17,   8,   4,   3,   2, -24,   9, -27,   4,   2,   3,   1,   1,  38,   0, -14, -11,   0,  -1,   0,   0,  14,  24, -14,  -7,  -4,   0,  -1,   0,  -7,  14,  14,  -7,  -3,  -3,   1,  -1,   2,  -4,   4,   5,  -1,  -1,  -1,   0,  -2,   3,  -1,  -1,   2,   0,   0,   0,   2,  -1,   0,   0,  -1,   1,   0,   0 },
      {  19, -11,   0,   1,   0,   0,   0,   0,  57, -20,  10,  -8,  -2,  -2,  -1,   0, -32, -34,  -4,  36,  -2,   4,  -1,   2,   1,  25, -64,  -8,  14,   0,   2,   0,  -9,  31,  22, -37,  -5,   0,   0,  -1,   1,  -5,  15,  11, -11,  -2,   0,   0,  -3,   4,   0,   2,   4,  -3,  -1,   0,  -1,  -2,   3,   0,   1,   1,  -1,   0 },
      {  18,  -5,  29, -13,   0,  -2,   0,  -1, -18, -62,   1,  28,   3,   4,   1,   2,   5,   9,  -8,  -6,   0,   1,   0,   0,   3, -29, -15,  34,  -3,   4,   0,   1,  -2,  32, -59, -19,  19,  -2,   4,   0,  -1,  12,  26, -34,  -8,   2,  -1,   0,   3,   1,   2,  10,  -9,  -2,  -1,  -1,  -1,   4,   3,  -2,   3,  -2,   0,  -1 },
      {  13,  -4, -15,   9,  -2,   2,  -1,   1,  29,  18,   2, -24,   2,  -4,   1,  -2,  69,   7,  29,   6, -22,  -2,  -5,   0, -61,  13,  -8,  44,  15,  -4,   1,  -1,  -9, -17, -13, -19,  22,   7,   2,   0,  -1,   1,   0,  -9, -11,   5,   2,   0,  -2,   0,   2,   0,   0,  -3,   2,   1,  -1,   1,   1,   0,  -2,   0,  -1,   1 },
      {  -2, -36,  -3,   7,   7,   1,   2,   1,  28,   0,  48, -10,  -9,  -2,  -3,  -1,  -5, -33,  -8,  27,   4,   3,   1,   1,  11,   2,   5,   5, -11,   3,  -3,   1, -14,  -8, -33,  51,  11,  -2,   2,  -1,  -8,  18, -40, -37,  30,   3,   5,   0,   2,   0,  17, -21, -16,   7,   0,   2,  -1,   5,  -5,   3,  -5,  -4,   2,   0 },
      {  28,  31,  29, -28,  -7,  -4,  -2,  -1,  31,  -2, -31,  16,  -1,   0,   0,   0,  36, -27,  20, -26,   0,   0,  -1,   0,  32,  25,   8,  15, -31,   1,  -3,   0, -33,  27,  12,  33,   7, -13,   0,  -3,  -7, -17,  -4,  13,  22,   4,  -1,   0,  -8,   3, -12,  -7,   7,   6,   3,   0,  -2,  -5,   3,  -6,   0,   1,   2,   0 },
      { -11,   6, -52,  10,   3,   6,   0,   2,   3,  19, -19,   6,  -1,  -3,   0,   0, -15,   1, -19, -21,  14,   2,   2,   0,  12,  43,   5,  -5, -20,  -6,   0,  -2, -26,   4,   0,  -8,  37,  -9,   3,  -2,  14, -15,  29, -53, -10,  18,  -3,   3,   3,  -4,  18,  25, -31,  -5,   0,  -1,   1,   0,  -3,   6,   9,  -6,  -1,  -1 },
      {   3,  19,  69, -27,  -1,  -6,  -3,  -2,   3,  53,  -6,  33, -11,  -2,  -4,  -1,  -3, -18, -37,  18,  25,   2,   3,   1, -43,  -2,  -4, -17,   9,   7,   3,   1,  -1, -19,  -2, -10,   7,  -1,   4,   0,  11,  -4,   0, -22,  -6,   7,  -3,   2,   2,   4,   7,   3, -14,   0,   0,  -2,   2,  -2,   1,   4,   0,  -4,   1,   0 },
      {   6, -12, -25,  -8,   3,   4,   3,   2,  33,  20,  14,   3, -13,  -4,  -3,  -1,  32, -32, -19, -18,  20,   3,   2,   1,  -3, -26,  22, -23, -22,  10,  -2,   3,  16,  17, -21,   3, -30,  -3,   1,   1,  -4,  35,  30,  12, -24, -20,  -5,  -2,  -5,  -3,   2,  42,  21, -10,  -6,  -3,  -5,   3,  -8,  -9,  19,   8,   1,  -1 },
      {  -5, -16,  26, -19,   1,   1,   0,   0,  -1, -12,  39,   1,  -2,  -5,   0,  -1, -26, -12,  11, -11, -10,   8,  -1,   2,  15,  22,  20,  30, -16, -15,  -2,  -3,  -8, -43,   9,  -1,  27,  13,  -4,   2,  17,  -4, -13,  15, -47,   9,   2,   1,  -9,  19, -18,  35,  17, -27,   2,  -4,   3,  -2,  -3, -13,  22,   5,  -4,   2 },
      {  -2,  -2,  -2, -41,  16,  -6,   5,  -3,   3,  13,  35,  -8, -26,   2,  -6,   2,  32,  50,   7,  38,  -1, -14,  -1,  -5,  40, -20,   4, -16,  24,   9,  -1,   2, -34,  14,  10,  -1,  -8,  10,   1,   0,   1, -29,  11,  -3,  -5,   7,   5,   3,   1,   4,  -8,  17, -21,  -6,   0,   2,  -4,   1,  -1,   7,  15, -12,  -3,  -2 },
      {   5,  16,  26,  51, -18,  -3,  -1,   0,   7,  32,  46, -13,  14,  -7,  -1,  -2,  -3,  28, -43, -12,  -5,  11,  -1,   0,  21,   0,   2,  25,  -3,  -6,  -1,  -2, -24,  13,  -5,  -7,   9,  -1,   6,   0,   1, -10,  26,   9,   9, -16,   0,   0,  -9, -28,  -6,  -2,  27,  18,  -4,   3,   2,  -1,  -9, -10, -18,  12,   4,   1 },
      {  18,  17,  15,  34, -12,  -7,  -4,  -1,  12,  -7, -36, -10,  19,   7,   4,   0,  17, -12,  -7,  -1,  -2,  -9,   3,  -1,  50,  -2,  10, -15,  37,   2,  -7,   1, -26, -31, -13, -15, -16,  36,   6,   5,   4,  28, -22, -15, -29, -15,   9,   2,  -7,   9,  24,   0,  -7, -12,  -8,   0,   1,   1,  -1,  11,   3,  -4,  -4,  -3 },
      { -14,  -3, -36, -32,   3,   6,   2,   1,  -5,  20, -19,  52, -23,  -1,  -4,   0,  -4,  12, -12,  25,  -3,  -6,   1,  -1,  16,  21,  -8,  17,  19, -10,  -1,  -3, -15,  -7, -25,  -7,  12,  10,   4,   1,   1,  22,  -9,   7,   2, -22,   5,  -4,  -2,  -3,   2, -23,  51,   0,  -6,   1,   4, -10,  11, -29, -13,  24,   0,   2 },
      {  -2, -19, -19, -11,   6,   4,   0,   2,  10, -19,  21,  27,  17,  -6,   1,  -1,   2, -24,  17, -33,  34,  -2,   6,  -1,  -8, -19,  14,  -9,  17,  20,  -5,   2, -32, -26,   2, -25,  -8,  16,  10,   5,  25, -30,   5,   3,  -2,  -2,  -9,   5,  -6,  -5, -22, -16,  -8,  36,  -2,   1,   1,   8,  -5,  -1, -46,   3,  13,  -3 },
      {   7,  -2, -19,  29,  -5,   3,  -3,   1,  15,   6,  -3,  28,  -3,  -5,   1,   0,  18, -15, -34,   3,  13,  -1,   4,  -2,  15, -35,   8,   5,  13, -19,  -2,  -1,  34,  17,  12,  -4,  36,  18, -18,  -1, -25, -26, -21,   7, -12,  38,  17,   0,  -3,  14, -33, -15, -10, -18,  16,   6,  -3,   4,  11, -16,   0,  -8,  -9,   4 },
      {  -1,   6, -11,  51, -10,   2,  -4,   2,  -8,  -9,   0,  27, -15,   0,   2,   1,   2,  17,   5,   5,  25,   1,   0,   0, -25, -20, -29,  -5, -20,  24,   8,   3, -47,  -3,  -3,  15, -14, -23,  12,  -1,  19, -17, -28,   7,  -1, -12,  -8,   3,  -6,  32, -10,   4,  10, -24,  -3,  -7,   7,  -5,  10, -11,  35,  -2,  -8,  -1 },
      {   3,   2,  -3, -26,  -5,  11,   2,   2,  11,  13,  26,  12,  20, -17,   4,  -4,  16,  35,   5, -41,  15,  28,  -4,   2,  21, -11, -48,  -9, -30,  -2,  11,   1,  13, -12,  13, -16,   0,  -9,  -7,   4, -13,  19, -26,  -6, -17,   4,   2,  -4,   9,   6,  34, -25,  -3, -15,   1,  -1,  -2,  -6,  12,  21, -15,   9,  -5,   3 },
      {   2,   1,  16,  28, -24,   5,  -3,   2,   1,  -4,  18,  36, -10, -17,   1,  -2,   4,  14,  55,  10,  43, -15,  -5,  -4,   7,  40,   1, -11,  14,  19,  -7,  -1,  39,   2,  -1,   8,   3,  14,   7,  -1, -31,  -5,   8, -12,  -2,  -2,  10,   8,   1, -25,   4,  14,   1,  -3, -11,   3,  -8,  -3, -18,  -2,   7,  15,   1,  -3 },
      {  10,   2,   1,  15,   1,  -1,  -4,   3,  11,  -6,  -8,   9, -11,   1,   3,  -1,  18,   9,   5,  25,   4,  -9,  -4,   0,  33, -12, -17,  -2,  -5,  -3,   8,  -1,  23, -45,   7, -10,  18, -40, -10,   5,  33,  31,  10,   6,  22,  28, -39,  -8, -21,  -3,  -5,  19,  -1,  33,  21,  -9,  -4,   0, -14, -10,   0,  -4,  13,   7 },
      {   7,   6,  12,   6,  45,   3,  -1,  -1,  -6, -19, -13, -38, -10,  -8,   7,   0,   6,  -4,   6,   9,  30,  -5,  -2,  -3,  -6, -25,   4, -20,  -4, -11,   5,   3, -12,   6,  20,   4,  31,   2, -14,   1,   1,  -9,  -4, -21,  -6,   9,   6,   3,  14,   2,  22,   4,  33,   1, -18,  -1,  -4, -16,  -9, -20, -13,  63,   8,   0 },
      {  -3,   7,   6,   4,  33,   7,  -5,  -1, -10,   3,  -6, -16,   9,   4,   8,  -1, -10,   8,   7,   6,  23,  10,  -9,   3,  -7,  19,  -4, -15, -13,  23,   9,   2, -35,  -6, -27,  -5, -10,   5, -16,   2, -33,  31,   8,  15, -11,  37,  18, -19, -13, -26, -24,  -7,  -2, -12,  45,  13,  -3,   6,  -1, -25, -14, -17, -11,  13 },
      {   3,   3,  -2,  -8,  15,   1,   0,   0,   2,   4, -12, -35,  -5,  10,   3,   1,   1,   2, -21,   4,  15,   0,  -3,   2,  20,  14,  -6,  15, -13,  32,   3,  -2,  24, -27, -20, -17,  -4, -20,  36,  10, -48, -29, -11, -10,  -9, -19, -28,  18,  17,  14, -31,  -3,  -2,   5, -19, -21,  -7,   9,  17, -18,  -5,  -1,  10,  -4 },
      {  -6, -13,  -6, -13, -59,  18,  -1,   6,  -4, -10, -22,   5,  25, -16,   4,   1,   3, -10,  -1,  44,  -8,  29,  -4,  -2,  -2, -26,   9,  -8, -17, -11,  21,  -2, -31, -22,   5,   1,   6, -12,  -7,   7, -32, -15,   7,   4,  -6,  -4,  -1,  -2,  29, -31,  -2,  14,   3,  -3,  -5,   2, -14,  13, -19,   7,   0,   7,   2,  -2 },
      {   5,   3,   2,  19,  64, -16,   6,  -5,   8,   5,   7,  46,  20,  29,  -6,   0,   2,  -7,  11, -12, -40,   8,  13,   4,   1, -11, -23, -12,   2, -23,   3,   4,  -6, -31,   5,  11,  -3,   0,   1,  -1, -32, -10,  16,  -5,  10, -10,  -1,   4,  19, -24,  -7,  12,  -4,  -4, -13,   2,   1,   4, -10,   0,  18,   2,  -3,  -4 },
      {  -7,  -5,   1,   0, -18, -15,  -4,   3,  -3,   2,  18,   3,   8,  -8,  -3, -10, -10,  -6,   1, -24,   5, -41, -10,   7,  -7,   0,  10,   1,  31,  -8,  -7,  -2, -30,   5,  15,  -6,   5,  -5, -10,   7, -45,  27,  12,   0,  13,  14, -37, -13,  43,  31,   7,   1,   9,  15,  17, -20, -12, -13,  20,   1,  23,   3,   3,  12 },
      {   0,  -4,  -1,  -8,   6, -24,   6,  -3,   8,   8,   8,   1,  83,  -8,   1,  -4,   6,   4,  12,  21,   4,  10,   1,  -1,   5,  15,  25, -27,   8,   3,   6,   1,  15,  14, -29, -14,   5, -22,  -8,   3,  26, -18,  -5,  -6,  12,   2,  -9,  -2,  -3,  10,  -2,  -8,  13,  -6,   2,   3,   7,  -7,  24, -19,  28,  12, -18,   1 },
      {  -2,   8,   0,  11,  13,  14,  -8,   2,  -4,  10,   2,  21,  18,  12,  -5,   0,  -2,  15,  10,  22, -21,  17,  10,  -3,  -9,  10,  12, -11, -20,  -8, -18,   2,  -4,  40,  -5,  -7,  -4,  15,  -3,   5,  -9,   5, -39,  -1, -31,   2, -22,   0, -10,  24,   2,  22,  -5,  30,   6, -28, -30,  -7, -18,   2, -22,  -3,  47,   4 },
      {  -6,  -3, -11, -23,   1,  -3,   5,  -7,  -6,   0, -12,  -8,  51, -19,  -2,  -5,  -3,   9, -15,   2,  22,  -2,  -3,   1,  -9, -17, -30,  23,  12,   9, -20,   3,  -3,   0,  37,  33,   9,  25,  26, -10, -11,  28,  -1, -13,   8,  -3,   6,  17, -37,  -6, -14,  20,   0,  11, -12,  -5,   5, -10, -14, -21,  18, -15,   9,  -4 },
      {   0,   1,   0,  -4, -31,  -7,   7,  -7,   0,   3,  -4, -16,  -1,   1, -16,   0,   3,  11,   3, -15, -11,   7,   4,  12,  -2,   7, -41, -42,   2, -14,   6,   3,   7, -20, -39,  37,   8,  28, -18, -11,   1, -16,  30,   0,   3,   2,  16,  -4,   5,  51, -18,   9,   4,  16,   1,   4,  -2,   7,  -5, -20, -16,   6,  15,  -2 },
      {  -4,  -3,  -1,   1,   1,  -6,   3, -16,  -6,  -6,   2,   1,  -9,  -7, -27,  10,  -2,  -5,   7,   5,   5,  12,  22,  14,  -4,  -4,  17,  -2,   4,  -4,  28,  -3,  -7,  -4,  -2, -12,  10, -12, -16,   7, -50, -12,  14,   9,  -1,  -7, -13,   3, -69,  22,  36,  -6,   1,   7,  -7,  -4,  44,  -9,  -3,   9,  -2, -10,  -1,  -7 },
      {   4,   5,   8,  18,  23,  40,  -3,  -2,   1,   1,   4,  12,   3, -33,  -1,  -9,   2,   1,  13,  24,  -5,  17, -23,   5,   6,   1,  26,  -1, -13,  -6,  20,   5,  -1,  -1,  21,  -4,   3,   2,  13,  -8,   3,  37,  37,  -7,   6, -12,  18,  15,  19,  45, -24, -16,  -8,   6, -24,   9,   7,  -3,  23,   5, -13, -21,  -2,  -8 },
      {   1,   0, -10,   4,  20,  -5,  -4,  -9,   2,   2,  -9,  14,  17,  -4, -18,  -1,  -1,  -3, -19,  12,   2,  12, -22,  13,   9,   4,   2,  44,   6,  50,  15,  -8,  12,  -8,   0,  13, -17,  11,  -1,   1,   3, -21,  11,  21,   4,  14,  15, -16,  22,  33,  31,  12,   1,   8,  25,  13,  -7, -10, -33,  26,   1,  25,   1,   6 },
      {  -3,  -4,  -6,   6, -13, -81,   9,  -9,  -2,  -3,  -1,  -1, -11,  37, -35,  -3,   4,   4,  12,  15,  34,  37,  12,  -7,   0,   2,  11,  16, -18, -13,   7,  11,  -2,   3,  23,  -8,  -3,   7,   8,   2,   4,  25,   7,  -9,  -3,  -2,   8,   2,  18,   7, -10,  -7,  -1,   0,  -2,   5, -11,   3,   5,  -3, -15,  -4,  -1,   2 },
      {  -3,  -5,  -2,  -3, -28,   5, -18,   2,   3,   4,  13,   8,  11,  60,  11,  -7,  -1,  -2,   3,   3, -36, -25,  -8,  13,  -2, -12,   7, -18, -11,  36,   8, -10,   0,   2,  20,   4,  15, -11,  33,   2,  -6,  15,  17, -16, -11,  13,  11,  12, -24,   9,  -6, -13,  -5, -17,   8,  10,   2, -18,   3,  -5, -19,  38,  -4,   1 },
      {   2,   1,  -1,   6,   3, -32,  20,  -2,   0,  -4,  -7,   3, -16, -41, -19, -11,   2,  -2,   3,  -3, -30,  20, -40,   9,   0,  -6,   0, -27,   4,  21,  -9, -11,   1,  12,  -3,  14,  21,   9,  29,   4,   7,   3,  -5,   4, -25,   7, -25,   7,  14, -24,  -8, -13,  -1,  -7,  17, -26,  50,   1,  -5,   4,  -5,   4,   9,  12 },
      {  -4,   2,  -5,  -2,   2, -22,  -6,  -8,  -3,   9,  -2,   4,   0, -14,   6,   9,  -7,   5,  -8,  -6,  -5,  -5, -17,   2,  -3,  10,  -3,   4,   7,   7,   4,   7, -12,   9, -10,  -5,   1, -10, -24,   1,  -4,  26, -12,  20,   5,  25,  -5,  14, -18,   6, -28,   2, -24,  -3, -42,   1,  -6,  53,  -6,  28,   5,  50,   2, -42 },
      {   3,   3,   1,   0,  -5,  28,  -1,   1,   4,   3,   0,  -1,  19,  33, -19,  -1,   5,   7,   0,  22,  23, -27,  10,   9,   2,   2,  -3,  14, -31, -10, -33,  11,   2,  -3,   2,   0, -14,  18, -20, -12,   3,   0,   7,   1,  -7,   7,  -4, -11,  23,  -5,  -5,  -7,  -2,   5,   3, -20,  76,  22,   8,  12,   4,  17,   7,  -2 },
      {   2,   2,   3,  -6,  -7,  31, -15,   0,   2,   2,   3,  -4,  -6,  19, -14,  -1,   2,   2,   4, -23,   0,  36,  -2,  -3,   3,   8,   2,   9,  44,  -7,  50,   0,   4,  20,  15,   4, -27,  -4,  -3,  24,   7,   8, -21, -37,   8,   3, -15,  -6,  11,   1, -20,  13,  -2,  -1,  10,   1,  31,  14, -23, -32,   6,  13,  -3,  -2 },
      {   0,  -1,   1,   3,  -6,  26,  10,  22,  -1,  -4,   0,  -4, -11,  16, -27,   5,  -1,  -6,  -2,  -4,  -1,  50,  -5, -28,   2,   0,   2,   3,  16,  11, -16,  13,  -3, -13,  -9,   3,  -7,  14,  -8, -13,   0,   0,  16,  10,  11,  25,  -3,   8, -13, -18,  -3, -10, -17,   3,  -7, -24, -17, -47,  43,   7,  30,  29,  16,  -8 },
      {   2,   3,   1,  -1,  -6,  16,  65,  -7,   1,   1,   1,  -5,   6,  11, -30,  -2,  -1,  -1,  -5,  -9,   2, -28,  10,   3,   0,   4,   2,  -5,  -2,  11,  17,  13,  -5,   6, -11, -20,  19, -14,  -8,   1,   5,  15, -18,  24,   5,   3,   8,  15,  24,  -2, -11,  -8, -16, -12, -22,   9,   2, -43, -48, -10,   1, -13,  10,  -9 },
      {  -1,  -2,  -3,  -2, -10, -24,   2,   6,   1,   0,   0,   6, -11, -14,  20,  -8,  -1,  -3,  -3,  -1, -17,   3, -47,   8,   1,  -3,   5,   4, -14, -13, -33,  15,   0, -11,   1, -33, -45, -14, -19, -20, -17, -13, -14, -35,  14,  10,  20, -13, -10,   9,  -9,  13,  -1,   3,  -6,  14,  21, -39, -16, -22,  -7,   1,  10, -16 },
      {  -1,  -4,  -6,  -3,   5, -20, -47,  18,   2,   1,  -1,  -1,   7,   5,  36, -30,   1,   2,  -6,   6,  16,  -9,  -4,  -6,   1,   6,  -6,   9,   6, -19,  32,  23,   3,   3, -18,  29,  -3,   6, -11,  17,   7,  -5,  16,  19, -19,   1, -21,   3,   4,  -7,  -9, -14, -10, -19, -19, -12,  15, -50, -16,  19, -12, -14,  15, -11 },
      {  -1,  -4,  -1,  -5,  -4,  -6, -72,  -7,   1,  -1,   3,  -4,  -6,  -2,  -8,   6,   2,   2,   1,  -9, -12,  20,  25,  11,   2,  -3,   3, -12,  -3,  36, -20,  -9,  -2,   0,   1, -23,  26,   4,  -8, -13,   4,   9,  -8,  21,  18,   8,  16,   6,  31,  -6,  19,  -1,  -1,   6, -25,   3,  15,  -1, -11, -34,  18, -24,  14, -15 },
      {  -1,   0,  -2,  -6,   2,   8,  -3,   4,   0,   1,   1,  -6,  -6,  37,  25,  -7,   1,   2,   3,  -5,  27,  25, -36,   4,  -3,   0,   0, -18,   2, -35,  -8, -18,  -5,   5, -21, -20,  27,  11,  38,  -1,  -8,   2, -16,  27,  25,  -4,   9,  18,   0,  13,  -3,  14,   6,  28,  11,   9,  15,  -5, -21,  33,  23,   1, -22,   2 },
      {   0,   0,   0,  -3,   1,  -2, -22,  -6,  -2,  -3,  -4,  -6,  -7, -40, -16, -41,   0,   1,  -3,   0, -11, -11,  56, -11,   2,   2,   1,  -4, -17, -11,   8,  21,   0,  -5,  -8, -25, -11,  10,  28, -13,  -4,   3,  -6,  -8,  25,  -5,  10,  13,  -6,   4, -25,  -1,  -8,  -9,  38,  -2,   0,  -9, -16,  22,  13,  37,   0,  14 },
      {   0,   1,   0,   1,  -1,   1,  26,  -2,   0,   1,   0,   1,  -3,  -1,  51,  -3,   1,   3,   3,   3,   6,  38,  50,   8,   2,   4,   6,   7,  39,   2, -42,  -1,  -4,   1,   1,  18,  -3, -46,   8, -17,  -7,   9,  14,   7, -16,  11,   2,   6,  13,  20,   0,  -5,  -3, -12,   6,  11,  18,   3, -14,   4, -10,   9,  16,   7 },
      {  -1,  -3,  -1,  -2,   1,  13, -24, -67,   1,  -1,   2,  -4,   6,   7, -23,  16,   1,  -3,   2,  -5,  14,  -3, -30,  -2,   2,  -1,   3,  -4,  14, -26, -10,  -2,   0,  -4,  -5,  -1,  -4, -26,  16, -19,   5,  -6,  11,  -7,  -4,   0,   6, -19,  -9,  -5,  -9, -32,  13, -34,   3, -12, -19,   6, -19,  15,  13,  -6,  46,   9 },
      {   0,   0,   1,  -2,   4, -11,   6,  55,   1,   1,   4,   3,   9,   5,  -3, -11,  -2,  -3,  -1,  -6,  -1, -13, -13, -63,   0,   0,   2,  -1,   6,   1,  21,   3,  -1,  -2,   0,  -6,   6, -17,  15, -11,  -8,  -2,   1,  10,  10,   5,  37,  -3,  -5,  12,  21,   7,   7, -12,  -2,  -5,  18,  36,   0,  -7,   0,  -9,  52,  17 },
      {  -1,  -3,  -2,   4,  -8,  -8,  19, -68,   3,   2,   4,   6,  -3,  -3,  51,  22,   0,  -2,  -1,   2,  -4,  -6,   7, -24,   0,   0,   0,   5, -12,   0,  34,  -7,  -1,   0,  -1,  -5,  -8,  24,  10,   8,   1,   3,  -6,  14,  14,  10,  13,   7,  12,  -8,  13,  24, -18,   9,   6, -11,  21, -12,  34, -16,  -1,  10,  27,   1 },
      {   0,  -1,  -3,   2,  -1,   2, -17, -24,   0,  -1,  -2,  -1,   3,  -3, -27,  -1,  -1,  -1,  -3,  -4,  -7,  -7,   6, -39,  -1,   1,  -9,  -7,  15,  12, -12,   8,   1,   0,  -5,  14,  10, -41,  -4,  24,  -6,   4, -17,  -7, -23,   5,  36, -13,   8,  -6, -24,  35,  22,  33, -11,   7,  17, -23,  23,  38, -20,  -8, -22,   2 },
      {   0,   1,   1,   2,  -2,   8,  -8,  -2,  -1,   0,   1,   2,   6, -10, -25,  10,   0,   0,   0,   3,  -2,  -7,   3,  -7,  -2,   1,  -2,   2,  -1, -31, -17,   7,  -3,  -1, -12,   5, -32,  -3,  49,  20,   2,  -6,   6,  18,  10,  64, -13,  23,   5,   8,  31,  12,   4, -20, -22,  -4,  -3,   3,  -4, -23, -32,  -4, -31,  -1 },
      {   0,   1,   0,  -1,   3, -10, -17,  22,   1,   2,   1,   0,   7, -16,  -6,  76,   0,   1,   0,   2,   2, -10,   3, -38,  -1,   1,  -4,   2,  -9, -16,  -9, -45,  -2,  -3,  -1,   4,  -9,  -7,  -6,   8,  -2,   8,  11,  -2, -13, -21,   0,  25,   8,  11, -14, -19, -23,   3,  28,   5,   8, -22, -23,  -7,   3,  15,  -8,  -9 },
      {   1,  -1,   0,   2,   4,  -4,  -3,  15,   0,  -2,  -2,  -2,   6,  -1, -22, -14,   2,   0,   0,   0,  12,  -9,   5,  48,   1,  -1,  -2,   5,  11, -15,  18, -62,  -3,   1,  -7,   4, -15, -15,  -8, -25,   5,   1, -14,   0,   4,   7,  12,  39,   6, -18,  -5,  30,   9, -14,  13,   5,   0, -11,  35,  15, -19,  -5,  32, -19 },
      {   0,   0,   1,   1,   1,   5, -13,  -8,   0,   0,   0,   0,   5,   3, -15,   8,  -1,   0,  -1,   1,   4,  -2, -15, -19,   1,   0,   0,   0,  10,  -3,   4,  38,  -1,   4,  -1,   8,  16,  -5,   9, -64,  -7,   0, -14,  18,  -8, -26, -43,  -3,   6,  -3,  12,  26, -25,  -9,  20,  50,   5,  -2,  18, -12, -26,   9,  -7, -12 },
      {   0,   1,   2,   1,   3,   6,   6,  -4,  -2,  -1,  -2,  -3,   1,  -2,  -8, -27,   0,   1,   1,   4,   4,   7,   4, -22,  -3,  -2,  -3,  -3,  -7,   1, -32, -44,   0,  -2,   1,  -3,  11,  16,   2,  52,  -2,   2,   5,   4,  29, -14, -27, -44,  -5,   8,  -7,  12, -13, -38,  -7,  43,   5,  -5,   8,   3, -11,   5,  22, -19 },
      {   0,   0,   1,   1,   0,   5,  -2, -24,  -1,  -2,  -1,   2,  -3,   4, -11, -63,  -1,  -1,   0,   1,   3,  -3, -18, -16,  -1,  -2,  -2,   7,   3, -14, -17, -10,  -2,  -2,   1,   8,  -4, -24, -17,  20,   1,  -4,   7,  17, -24, -25,  21,  31,   0,  -1,  26, -14, -36,  20,  25, -14,  -9,  12,   0, -40,   7,  10, -12, -25 },
      {   0,   0,   0,   0,   2,  -4,  -4,  17,  -1,  -1,  -1,  -1,   1, -10,  -7,   5,   1,   1,   1,   1,   7,  -6,   8,  32,   0,   0,   1,   2,   4, -22,  14, -18,  -1,   3,  -3,   4,  -2, -15,  37, -11,   1,   3,  -6,  20, -18,  -7,  25, -72,  -3,  -7,  -2,   0, -49,  22, -27, -11,  -1, -10,   8, -20,   6,  23,  -4,  16 },
      {   0,   0,  -1,  -1,   2,   0,  -8,   0,   0,   0,   0,   2,   6,   3,  -1,  13,   0,   0,   1,  -1,   4,   8, -13,  13,   0,   0,   2,   1,  13,   5, -16,  14,  -1,  -2,   1,  -5,  16,  -7, -23,  10,   1,   4,  -1,  11,  41, -37,  16,   7,  -3,   6,  -5,  33, -24, -44,   3, -59,  -1,  -2,  -1,  -1, -42,   7, -19,  36 },
      {   0,   0,  -1,   0,  -1,   0,   2, -17,   0,   0,  -1,   1,   0,   6,   4, -32,   0,   0,  -1,   1,  -3,   2,   3, -42,   0,   1,   0,   2,  -4,   8,   8, -60,  -3,   0,  -3,  -1, -11,   9, -22, -56,  -2,   1,   0,  -6,  -2,  11, -23,  -4,  -3,   5,  -8,  -4,   0,   0, -37, -16,   0,  -2, -19,   0, -10,  -2, -35,  37 },
      {   1,   0,   1,   1,   2,   0,  11,   2,   1,   0,   2,   0,   2,   2,   8,   0,   1,  -1,   2,   0,   1,   2,  11, -15,   1,  -1,   3,  -4,   3,   4,   9,  -8,   2,   1,   4,  -1,  10,   2,  18, -25,   0,   0,   3,  -6,   8,   2,   5, -39,   4,   4,   2,   3,  19, -11,  25, -53,  -4,  -2, -11,   4, -15, -12, -22, -89 },
    },
  },
  { // 31
    {
      { 102, -19, -25,  -4,  -3,  -1,  -2,   0, -56,  31,  14,  -3,   0,  -1,   0,   0,  -9, -18,   8,   6,   2,   1,   0,   0,   3,   3,  -7,  -1,   0,   1,   0,   0,  -2,   0,   3,   0,   0,   0,   0,   0,   1,   0,  -1,   0,   1,   0,   0,   0,  -1,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  22, -71,  21,  10,   2,   2,   1,   1,  55,  33, -46,  -8,  -3,   0,  -1,   0, -36,  34,  28, -10,  -5,  -2,   0,  -1,  -4, -19,   4,  14,   1,   0,   0,   0,  -4,   7,  -7,  -3,   3,   0,   0,   0,  -1,  -2,   3,   0,  -1,   0,   0,   0,  -1,   2,  -1,  -1,   1,   0,   0,   0,  -1,  -1,   1,   0,   0,   1,   0,   0 },
      {  49,   5,  12, -17,  -4,  -3,  -1,  -2,  12, -62, -11,  31,   3,   5,   0,   2,   5,  54, -46, -23,   2,   0,   1,  -1, -19,  14,  41,  -8,  -7,  -4,  -1,  -1,  -1, -10,  -4,  14,   2,   1,   0,   0,  -5,   4,   1,  -4,   2,  -1,   1,   0,   0,  -2,   0,   1,   1,   0,   0,   0,  -2,   1,   1,  -1,   0,   0,   0,   0 },
      { -28, -66,  21,  22,   2,   4,   1,   2, -59,  10, -18,  -3,  10,   2,   3,   0,  69,  13, -17, -21,  -8,  -2,  -1,   0,  -5,  18,   9,  -2,  -4,  -3,  -2,   0,   2,  -3,   7,   2,   0,   0,  -1,  -1,   0,   1,  -1,   4,  -1,   1,   1,   0,   1,   1,   0,  -1,   0,  -1,   1,   0,   0,   1,   0,   0,   0,   0,  -1,   0 },
      {  -4,  47,  -3,  -8,  -7,  -2,  -2,   0, -36,  11, -28,   4,  13,  -1,   2,   0,  12,  21,  49, -40, -10,  -2,  -2,  -2,  -1, -45,  31,  47,  -7,   0,  -3,   0,   3,  -4, -35,   0,  15,   3,   2,   0,   2,  -1,   7,  -7,  -5,   2,   0,   0,   2,  -1,  -5,   4,   1,   0,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0 },
      {   3,   0,  49,  -4, -11,  -2,  -1,  -1, -29, -40, -38,  31,  19,   4,   3,   2, -37, -27,   0, -15,  -1,   7,   2,   2,  45,  13, -39,   7,  -2,  -7,  -1,  -1,  -7,  28,   7, -31,   7,   3,   0,  -1,   6,  -7,  14,   3, -15,   1,   0,   1,  -2,   7,  -3,   1,   2,  -4,   0,  -1,   3,  -1,   3,   0,  -1,  -1,   0,   1 },
      { -20,   3,  -7,  -6,   7,   3,   1,   1, -15,  35, -16,  12,  -3,  -3,  -1,  -1, -40, -10, -16, -31,  25,   5,   5,   1,  46,   3,  41, -34, -38,   0,  -2,   0,  -2,  -7,  20,  50,  -2, -12,  -5,  -2,   4,  -8, -18,  -2,  18,   9,   1,   0,  -2,   1,   1,  -4,  -8,   0,   3,   1,   3,  -1,  -1,   1,   2,  -2,   0,  -1 },
      { -17, -46,   7, -20,  12,   5,   4,   1, -17, -17,  56,  27, -22,  -3,  -3,   0, -16,   2, -13,  15,  17,  -2,   2,   0,  17, -41,  -3,  15,  -9,   6,  -1,   2,   3,  15, -54,   9,  17,  -5,   3,   0,   0,  12,   3, -34,   2,   4,   0,   0,   3,   1,   3,   6,  -9,  -2,   0,  -1,   0,   4,  -1,  -4,   4,  -1,  -1,   0 },
      {  18,  -2,  24,  -4,  -7,  -2,  -2,   0,  38,  -5,  11,  -8,   7,  -7,  -1,  -2,  32, -50,  13, -11, -13,  15,  -3,   3,   2,  40,   2,  41, -32, -11,  -1,  -2, -28,  -4, -14,  34,  39,  -8,  -2,  -2,  -2,  -3, -18, -21,  11,  16,   3,   1,  -5,   3,   2,  -5,  -6,  -2,   3,   2,   0,   1,  -2,  -1,   1,   2,  -2,  -1 },
      { -13, -22, -90,  12,  18,   9,   4,   2,  -2, -41,  -7, -15,   3,   5,   0,   2, -14,  18,  17, -10, -19,  -2,  -1,   0,  32,  33,  -1,  15,  -2, -12,  -5,  -3, -16,   3,   4,  -7,  18,   6,   0,  -3,   7,  -5,   4, -13, -11,   5,   3,   2,  -1,   3,   5,   6,  -7,  -4,   0,   0,   2,  -1,   1,   0,   1,  -3,   0,   0 },
      {   6,  -5,  21,   7,   4,  -7,  -1,  -2, -14, -45,  12, -38,   0,  14,   4,   2,  -3,  14,  36,  18, -25, -15,  -4,  -1,  27, -13,  -1,   2,   4,   8,  -5,   2,   2,  17,   5,  40, -30, -12,  -1,  -2,  -9,   4, -24,  25,  49, -11,  -2,  -3,   0,  -2,  -4, -28,   6,  19,   0,   2,  -1,   1,  -1,   4,  -7,   0,   1,   0 },
      {  11,   3, -29,  -6,  16,   2,  -3,   2,  17, -10, -30,  18, -13,   0,   3,  -1,  44,  -8,  10, -10,  35,  -2,  -8,  -1,  -8, -18, -28, -15, -30,  32,  12,   2,   0,  49,   2,  -2, -12, -40,   1,  -1,  -4, -12,  25,  -1,  24,  15, -13,  -3,  -4,   1,  -7,  -2, -14,   7,  11,   1,  -1,  -1,   5,   0,   0,  -7,  -1,   1 },
      {   2,  -4, -11,  60,   2,  -8,  -3,   0,  -2, -23, -15, -28,  31,  13,   2,   1, -28, -43, -24, -15,  -7,  16,   6,   4, -37, -20,   8,  -8, -15,  -5,   1,   2,  42,   6, -26,   6,  -6,  -8,  -8,  -2,  -4,  25,   5, -15,   4,   2,   0,  -3,   7,  -1,   7,   3, -11,   0,   2,   1,  -4,   7,   0,   3,   2,  -5,  -1,  -1 },
      {  -2, -24, -19, -34, -10,   8,   5,   1,  20,  -8,  28,   8,  16, -12,  -3,  -2,   8, -27,   7, -40,  -9,  18,   2,   3,   5, -11,   9,   5, -26,  -7,  -2,   1,  17,   8,   4,  -3, -28,   5,  -3,   1,  -1,  22,  11,  51, -17, -30,  -4,  -4,  -3,   1, -10,  12,  47,   2,  -7,  -5,  -2,  -1,  -1, -10,  -5,  15,   4,   0 },
      {  14,   2,   8,   6,  25, -11,  -2,  -2,  16,  -9, -20, -21,  -5,  13,   0,   1,  25, -31,  -2,  -3,  22, -21,  -1,   1,  41,  -4,  27, -15,  21,  35, -16,  -1, -30, -18, -31,  -5, -33,  24,  23,   4,   9,  18,   2, -14, -24, -24,  -4,   5,  -5,   8,  11,  20,  -3,  -1,  -6,  -5,  -1,   2,  -4,   2,   5,  -2,   2,   0 },
      {  12,  21,  -5,  56, -18,  -7,  -3,  -2,  18,  29,  22,   4,   4,   0,  -5,  -3,  18,  23, -44, -14, -13,  -3,  12,   1,  43, -18, -10,  10,   4, -17, -12,   1, -19,  46,  -7,   1,   7,   9,   4,  -7, -16, -10,  14,  -1,  10, -17,   8,   2,  -1,  -9,   1,  -6,  18,   5, -10,   0,  -2,  -5,   2,  -5,  -2,   8,   1,   0 },
      {  12,   4,  25,  17,  10,  -4,  -4,  -3,   5, -15,  43, -41,   8,  -9,   5,   2,   7,  26,  17, -26,  12,  17,  -3,   1,  13, -25, -11, -21, -24,   5,  10,   7,  -4,  -5,  27,  -7,   1,   1, -15,   4,  -2, -18,  -2,   6, -39,  28,   9,   2,  -4,   8,  -8,  27, -24, -34,   4,   1,   1,   1,  -1,   8,  20,  -5,  -8,  -3 },
      {  -4,  -9, -13,  39, -27,  -2,   0,   3,   3, -11,  19,  43,  40, -17,  -7,  -3,  -3,   8,  22, -14,  29,  22,   2,  -3, -14, -11, -20,  -4,   6,  22,   6,   3, -38, -35,   0,  -2, -16,   1,  12,   4,  22,  -7, -34,  -3,  11, -13,  -5,   3,  -1,  10,   5, -21,  -3,  16,  -2,  -2,   3,   0,  -1,   3, -11,  -4,   5,   1 },
      {   8,  -7, -18, -13, -22,  16,   2,   0,  15, -12, -26,  24,   6, -12,  -4,   2,  33, -17,   8,  18,  -1,   4,   7,  -5,  22, -37,  -3, -21,  20, -20,  -5,   7,  30,   2,   7,  14,   6,  41, -20,  -7, -25,   5, -36,  -5, -10,  14,  35,   2,   2,   4,   5, -20, -15, -22,  -4,   9,  -4,   6,   3,   8,  -2,  -4,  -4,  -4 },
      {  -4, -12,   0, -28,  14, -18,   6,  -2,   8,   5,  27,  -5,  35,  13,  -6,  -2,  -5, -18,  -7, -31,  -3, -11,  11,   5, -14,  -1,  22,   7,  31,   9, -26,  -3,   9,  28,  15, -12,  -1,  26,  15,  -6,  -3, -46,   6,  -2,   7,  12,   0,  13,   0,  -1, -20, -17, -37,  25,   7,   2,  -2,   3,   4,  10, -19, -25,   6,   2 },
      {  -2,   6,  -3, -16,  49,   4,   0,  -4, -10,   0, -15,  -5, -14,  15,   0,   5,  -6,  -6, -25, -12, -20,   7,   9,   4, -39, -30, -24,   9, -10,  -7,  -7,   3, -38,  17,  12,  13,  -7,   7,   6, -11,  24, -14, -36,  -1,  -7, -22,  17,   7,   4,  28,  13, -15,  11, -23, -20,  -2,   5,  -8,  13,   5,   8,  15,  -2,  -6 },
      {  -3,   1,  -6, -49, -29,  -6,   9,   2,   5,  17,  -9, -25,  39,  15, -10,   1,  10,  18, -23,   9, -37,  27,  10,  -1,  17, -15, -20, -16, -10, -11,  13,   5, -24,  -6, -13,  -5, -17, -17,  -7,   1,  21,  26,   8, -18,   6,  -2, -10,  -8,  -4,  -1,  18,   5, -25,  10,  11,   0,   0,   5,  -1,  14,   0, -26,   4,   1 },
      {   9,  -2,  -2,  17,  31,   1,  -9,   1,  10,  -5,   0,  15, -21,  11,   6,   0,  10, -18, -23,  -3, -18,  -7,   4,   3,  21, -38,   8,  17,  15, -21,   8,  -1,  -7, -37,  16, -21,  13, -14, -25,  10,  33,  11,  -6,  28,   7,  32, -20, -15, -20,   7, -21,  -9,   2,  19,  30,   2,   4,   1,   0,  -8, -15, -12,   2,   6 },
      {  -7, -11,  -8,  11, -70,   0,   1,   5,  -6, -15,  -5,   3, -23,  19,  -9,   4,  -4, -21,   2,  29, -10, -15,  14,  -4, -18, -18,  38,   1,  -7,  -2,  -5,   0, -31,  20,  11,   9,  -4,  -7,   6,  -2,  23, -28,   6,  16, -10,   0,  -5,   4,  -7,  14, -21,  26,  -6, -15,   5,  -5,   3,   5,   1,  -3,  25, -12,  -1,  -2 },
      {   3,  -2,  10,   1,  11,  -5,   6,  -4,  -1, -14,   1,  -2,  -3, -17, -10,   4,   8,  -8,  29,   0,   8,  -4,  17,  -6, -13, -16,   5, -37,   6, -41, -16,  14, -15,   3,  19,   9,  21,   9, -23, -15,  30,  -1,  36, -21,  12,  -3,  -5,   7,   2, -15,  22,  24,   7,  38, -29,  -4,  -2,   0, -22,   6, -11,  24,  21,  -3 },
      {   1,   4,   4,  14,   9,   0,  -7,   0,   6,  10,  14,  51, -25,  33,  -1,  -5,   2,   1,  29, -30, -47, -18,  20,   7,  -7,  13, -30, -16,  -2,  -6, -16,   2,  13, -13,   6,  23, -18,   1,  16,  -4, -11,  20,   4,  -8,  -3, -10,  -3,   8,  -2,  -9,   0,  23, -23,  -7,   8,  -7,  -2,   0,  -9,   6,  20, -28,   2,   4 },
      {   5,  -2,  -2,  -7, -28,  16,  -2,   4,   6,  -7,  -4, -18, -17, -16,  14,  -3,   6,  -3, -17, -19,  -4, -19,  -4,   6,  14, -19, -12,  14,   1,   1,  -2, -10,  39, -15,  29,  -8,  14, -21,  28,  17,  -5, -16, -27, -24,  -2, -17, -45,  17,  10,  25,  25,   8,  -6,  18, -22, -25,   0,  -3,  -1,  14,   4,   7,  17,   3 },
      {   2,  -4,  -3,   4,  33, -39,   1,  -2,   4,   2,  -5,  25,  32,  29, -14,   0,  12,   1,  10,  45, -17,  29,  11,   1,  17,  -8,  14, -10, -16,  -7,  17,  -2,  14, -18,  -3,  -1,  11, -17,  10,  10, -25, -42,   1,  -1,  -9, -13, -19,  12,  15,  12, -18,  10,  12,   1, -18, -12,  -9,   5,   3,  -7,   9,  16,   4,  -3 },
      {   5,   6,  10,  14,  16,  61,   5,  -1,  -5,  -4,  -8,  -6, -30, -23, -17,  10,   0,  -2,  -5,  12, -15,  53,   4, -13,   4,   1,  10,  12, -13,   4,  15,   2,   3,  -5,  -6,  -4, -20,  10,  16,  -6,  -5, -23,   8,   2,  -3, -19,  11,  15,  -3, -17, -18,   4, -19,  19,  -7,   1,   0,   8, -12,   2, -21, -32,  26,  -1 },
      {  -4,  -4,  -2,  -4,  -6,  21,  -4,   3,  -4,  -5,   2,  -3, -13,  -4,  21,  -1,  -9, -18,   3, -24, -18, -11,   0,   6,  -9,  -6,   6, -27,  15, -11,  21,   5, -51, -12, -16,  -9,  -4,  -9, -21,  22, -62, -16,  13,  -4,  -1,  -1, -15, -23,  46,   0,  -5, -14,  -1,   1,  -3,   0, -12,   6,  12,   9, -10,   5,  -3,  -2 },
      {   2,  -2,   5,   5,   8,   8, -15,  -5,   7,   4,  16,  23,   1,   1,  20,   7,   4,  -8,  14,   0, -17,   6,   5,  -5,   8,  10,  43,   6,  10,  -5,  24,  -3,  11,  24,   6, -37, -24, -28, -10,  12,  22, -12, -12, -12,   9, -11,   6, -10, -16, -24,  45,  -8, -12, -19, -20,  19,  15,   6,  -2,  30,   8,  12, -25, -12 },
      {   2,  -6,  -3,  -6,  -9,  21,  -2,   3,   7,  -3,  -1,  -6,  18,  10,  32,   0,   0,  -8, -13,  -2, -11, -27,  -7,   3,   7, -17, -20,  -6, -10,  14,   3,  -4,   4, -43,   2,   9,  14,  -7,  23, -12,  21, -17,  35,  -2,  16, -10,  23,  20,  -1, -36, -30, -12,  -3, -16, -30,  27,   8,  16,   4,  -8,  -4,  16, -26, -15 },
      {  -3,  -3,  -1,   1,  15,   4,  -4,   6,  -3,  -1,  -2,  18,  29, -20,   4,  -5,  -6,  -9, -10,  16, -22,   6, -51,  -7,  -4, -25,  12,  21,  16,  22,   8, -19, -22,  15,  51,  23,   0,   7,  17,  10, -13,  25,   9, -13,  -1,   8,   1,  -5,  14, -21,   5,  28,   0,  -1,  12,   3, -17, -10, -27,  -1,  13,   4,  -1,   8 },
      {   2,   6,   5,   3,   9,  57, -21,   8,   2,   6,   4,   5,  39,  17,  39,   2,   0,   9,   7,  27,  -6, -22,   3,  19,  -8,   0,   6, -25, -31,   4, -30, -11,  -7,  19, -15,  -7,  15,   7,  -1,   4,   0,  -6, -25,   5, -21,   7, -15,  -8, -24, -10,   9,   4,   9,  12,  13, -14,  -3,  -3,  -4, -13, -13,   2,  18,   8 },
      {  -6,  -7,  -6, -10,  -8, -25, -23,  -1,  -7,  -8,  -8, -12, -17,  30,  12,   0,  -3,  -1,  -2,   2,  37,  21,  37,   6,  -6, -14,  -5,  29, -15,  -9, -19,   1, -18, -12,  31, -16,  -4,  -1,   8,   1, -41,  12, -16,  -9,  -3,  -4,   0,   3, -18, -59,   8,   1,   5,   2,   1,   1,   0,  14, -10,  -3, -11,  -4,  -1,   4 },
      {  -1,  -2,   1,   2, -21,  -5, -35,  10,   3,   2,   9,   1, -25,  61,   5, -14,   3,   3,   6,  -3,   2,  20, -29,   5,   5,   0,   3,  -2, -19,  15,  17,  -9,  11,   0,   9,   0,   6,  25,   2,  -6,  19,   0,  14,  -6, -14,   3,  21, -13,  28,  21,  14, -12,  -1,  18,  -1,  18, -19, -16,  -4,  19, -40,  16,  14,   2 },
      {   4,   5,   7,   6,   7,  14,  22,  13,   5,   4,   7,   3,   7,   9,  10, -18,   5,   4,  12,  13,  20,  -7,  -9,  -7,   5,   5,  21,   7, -19, -37, -14,  17,  -6,  13,  23, -27, -12, -12,  11,  -6,  13,  48,  -8, -19,  -1,  -6,  -4,  22,  44,   3, -41, -19, -12,  -7, -13,  -9, -10,  -5,  29,   5,   0, -12,  -2,  -4 },
      {  -2,   1,   2,   3,  -1,  -4,   6,  -3,  -5,  -3,   5,  -5,  -7,   7, -25,  -7,  -3,  -4,  12,  -8, -13,   0,  -7,  -4,  -8,  -8,  13, -33,   1,   5,  38, -11, -11,  16,   9,  -1,  37,   8,  30,  22, -18,  36,   9,  -3, -11,  -2,  -9,  25, -52,   2,  -4, -34,   1,  -6, -11, -15,  35,  -3,  14,  -2, -21,   0,   0,  -6 },
      {  -6,  -1,  -8,  -2,  -5, -31, -42, -11,  -2,   8,  -4,   9,  -7, -17,  31,  20,  -5,   5,  -8,   2,  -8,   7, -37,   4,  -2,  10,  -6,   5,   1,  -1,  -8,  11,  -6,   9, -12,   8, -14,   8, -23,   8,   1,  14,  -7,   6,  -6,   7,  -9,   8, -25,   4, -34,   5, -23,   8, -40, -16,  -5,  29,   8,  12,   0,   6,  33, -38 },
      {  -2,   0,   1,   6,   5,   7,   6,  -5,  -2,   1,   3,  11,   9,  -5, -22,   5,  -4,  -1,  -6,  20,   0, -21,  -6,  18,   0,  -7,   8,  22, -38,  10, -24,  -2, -12, -26,  10, -21,  -4,   5, -10, -29, -37,  11,  33,  30,  13,   9,   9,  10,  -5,  32,  35, -12, -19,   5, -10,  -7,  20,   2,  -6,  39,   8,   1,   0, -12 },
      {   0,  -2,  -1,   3,   0, -49,   3, -16,  -1,  -4,  -3,   2, -21, -48,  38,   6,   2,  -1,   5,  10, -19,   5, -11,  16,  -2,  -9,   7, -24, -32,  -9, -14, -22,  -2,  -1,  -7, -14,  17,  12,  26,  -1,  13,  -6,   3,  14,   0, -21,   6,   1,  15,  -9,  19, -17,   1,  -5,  18,   6, -14,  -4,   9,   3, -11, -40,  -8,  18 },
      {   3,   2,   1,   4,  -2,   1,   2,  18,   2,   0,  -2,   2, -14,  -5, -38,   9,   3,   2,   0,  -1, -19,  22,  -7,   7,   3,  -1,  -3, -11,  -1,  28, -41, -15,   3,   3,  -2,   1,   6,   4,  -1,   2,  20,   0,  -7,   4, -13,  19, -21, -15,  35, -41,  -3, -11,   0,   1,   8, -25,  29,  36,  32,  23,   1,  33, -15,   5 },
      {   0,  -2,   1,  -3,   7,  10, -60,  12,   0,  -3,  -1,  -3,  10, -18,  11, -35,   3,   4,   4,  12,  18,  19,  12,  27,   0,   0,   0,   9,  20, -29,  -6,  -8,   5,  -1,   5,  33, -14,   4,   7,   5,   4,  -6,  36, -12,  -5, -15,  -8, -11,   2,  24,  -7,  -9,   6,  -7,  13,  -6,  49,  -1,  12,  10,   7, -16, -15,   9 },
      {   3,   1,  -1,   5,   2, -17,  48,  27,   2,   0,  -3,   5,   0,  16,   8,  -3,   1,  -2,  -6,   4,  13, -26, -36,   7,  -3,  -1, -11,  16, -21, -20,  23,  14,   5,   1,  -4,  17, -16,  26, -23,  21,   5, -25,  -1,  -2, -19, -16,  -8, -20,   1, -28,   4,  -7,  -3,  -1,  -4, -16,  47,  -3, -24,   9,  -9, -20,   8,   3 },
      {   0,   0,  -2,   1,   9,   9, -18, -13,   0,   0,  -3,  -1,  12,  -4, -11,  -2,   0,   3,  -6,   8,  17, -22,  43,   2,   1,   6,  -5,  19,  -5,  13,  37,   6,   0,  10,  -1,  24,  23,  21,  -8,  33,  24,  13,  10,  39, -12, -12, -16,  10,  23, -21,   8, -17, -20,  -6,  -9, -15, -42,   7,  -3,  30,  12,  -8,  -1,   5 },
      {   0,  -2,   1,   1,  -2,  16,  -2, -60,   5,   3,   8,   1,   9,  28,   9,  63,   3,  -3,   4,  -4,  22,   2, -24, -39,   0,  -3,  -4,   5,   5, -25,   3, -13,   3,  -8,   7,  13,  -1,   6,  10, -10,  -1,  -4,  17,  -6,  -1,  -9,  -6, -12,   1,   0,   8, -10,   8,  -1,  20, -21,   6,  10,  29,   6,  16,  -5,   3,  17 },
      {   0,   1,   2,   4,   0,  -8,  -3, -32,   0,   1,   2,   4,   6, -22, -16,   7,  -1,   3,   1,  10, -18, -37,  27,   4,   0,   2,   3,  13, -22,  30,  11,  17,   7,   3,  31,   6, -10,  -9, -14,  -4,   1,   6,   2, -42, -31, -13,   8, -33,   5,  14, -10,  -3,  15,   6,   4,  21,   8,  35,  15, -17, -35,  -8,  10,  -1 },
      {   0,   4,  -1,   7,  -6,  10,  53, -32,  -1,   4,  -2,   6,  -9,  12,  27,  16,   2,   7,   4,   6,   6,  32,  10,  51,  -1,   2,  -1,   8,  17,   4, -19, -16,  -1,   8,   6,  24,  -6,  -7, -10,  24,   1,  11,  13,  -2, -20,   5, -15,  10,  -9,  21,  -6,  -2,  -7,  12,  -9,  12, -11,   2, -15,   5, -20,  10, -42,  -7 },
      {   1,   2,  -2,  -3,   2,   5, -18, -29,   2,   4,  -2,  -4,   3,  10, -22,   5,   3,   5,   1,  -7,  12, -12, -10,  18,  -2,   1,  -4, -10,  -6, -21,  13, -28,  -5,  14,  -8, -20,  10,   8,  14,  13,   3,  10, -28,  16,  22,  -1,   6,   8,  39,  13,  -9,  24,  -6,   8,  -4,  21,  35,  47, -48, -19,   1,  -8, -11,  -6 },
      {   0,   3,   3,   3,   6,   5,  -2,  34,  -4,   0,   0,  -3,  -2,  -4, -33,  -2,  -2,   2,   2,  -1,   8, -14, -18, -16,  -6,  -1, -10,  -1, -20, -35, -10, -24,   0,  -3,   0,  -1, -14,  18,  11,  21,  -3, -12,   2, -13,   1, -12, -16, -21, -39,  13,  -3,  -1,   2,  10,   3,  25, -53,  35,  10,  13,   2,   1, -39, -12 },
      {   2,   3,   3,   4,   7,  -8,  27,  48,   2,   3,   4,   4,   5,   1,  35,  11,   1,   3,   2,   1,  10,   1,  29, -31,   0,   2,  -3,   6,  14,  -2,   5, -45,  -3,   5,   1,  18,   7, -18,  11,   4,  -4,  12,   7,  20, -11,   6,  21,   4,  10,  17,  15,  15,   2,   1,   6,  21,  15,  39,  18,  14, -10,  -1,  41, -11 },
      {  -1,   0,  -2,   0,   0,  -6, -25,   8,   0,   1,  -3,   2,   4,  -7,  -7,  31,  -2,   0,  -6,   7,  -5,  -9,   2, -51,  -2,   1,  -7,   1,  -7,  -4, -24,  26,  -1,   3, -10,  15,  -1, -16,   8,  22,  -7,   3,  -7,  15, -31,  13, -12,  24,  17,   1,   5,   9,   1,  21,  -7,  31,  24, -41,  -7,   8, -31, -13, -39, -11 },
      {   0,  -1,   1,  -1,   2,   3,  -7,   6,   0,  -2,   1,  -3,   4,   8, -35,  14,   2,  -2,   2,  -6,  15,  -4, -26,  30,   2,  -3,   1,  -8,  25,  -3, -19, -15,   1,  -1,  -1,   7,  30, -57, -14, -17,  -2,   1,  -8,   9, -33, -37,  10, -14,  -3, -17,  -9, -18,   4,  -5, -11,  24, -12, -14, -11,  17,  11, -33,  23, -10 },
      {  -1,  -2,  -3,   2,  -3,  -2,  10, -49,  -2,  -3,  -5,   0,   2, -11, -14, -38,   0,  -1,  -4,   3,   0, -17,   4,   2,  -2,  -1,  -8,   1,  -7, -27,   1, -29,  -2,  -4, -18,   6, -31, -11,  28,   9,   2,  -9,  -7,  -2, -14,  32,   1,   0,  -4, -21, -15,   3,   5,  -9,  14,  20,   1, -27,   2,  42, -10,  30,  36, -11 },
      {   0,   1,  -2,   1,   3,   3,  15, -24,  -1,   1,  -3,   1,  -3,   7,  26, -50,  -2,   1,  -3,   1,  -1,  15,  -8, -34,  -1,   1,   0,  -6,   4,  12,  -7,  44,   0,   5,  -2,   0,  35,   1,  11, -20,  -8,   3, -12,  15,  -9, -26, -11, -24, -11,   7, -12,   7,  -2,  24,  10,  23,  11,  29, -11,  40,  22,   7, -11,  11 },
      {  -1,   0,   1,   1,   0,   1,  -3,  -6,  -1,   0,   1,   3,   2,   4,  -5, -29,  -2,  -1,  -1,  -1,   1,  -9, -29,  14,   0,  -1,   0,   4,  16, -14,  21,  18,  -6,   3,  -8,  16, -13, -23,  -3, -35,   0, -12,  -6,  13, -24,  19,   4,  53,  -1, -10,  36,  15,  17,   6,   6, -19, -18,  42,  22,  -5, -30, -16, -19,  12 },
      {   0,   1,  -1,  -2,  -2,  -6,  -5,   8,   1,   2,   1,  -2,  -3,  -6,   7,   1,  -1,   1,  -1,  -3,  -7,  -1,   6, -28,   1,   2,   2,  -4,   0,  -4, -13,  -2,   0,   1,  -5,  -1, -16, -25,  12,  -6,   5,  -1,  14,  -9, -25,  21,  11,   0,  -1,  14,  -9, -56,   8, -15,  -4, -36, -13,  25, -76,   1,   0,  11,  -1,  26 },
      {  -1,   0,  -2,  -1,   0,  -5,  -6,  20,   0,   1,  -1,  -1,   1,  -8,   9,  35,  -1,   1,  -1,  -2,  -2, -11,  14,  17,  -1,   2,  -1,  -5,  -3,   7,  37, -18,  -1,   4,  -4,  -7,   0,   8, -11, -36,  -4,   5, -19, -22,   3,  -2,  11,  19,  -6,   0, -42,  19,  30,  25,  12, -13,   3, -13,  -9,  61,   4,   1, -25,  23 },
      {   0,  -1,   0,  -1,  -1,  -1,   0, -21,   0,  -1,  -1,  -1,  -1,   3,   0, -42,   0,  -1,  -1,   0,  -1,  -5,   1, -38,  -2,  -1,  -4,   1,  -6,  11, -11, -61,  -2,  -4,  -2,  -3,   0, -11, -46,  12,  -2,  -1,  -5,   0,  -5, -17,  33,  -2,  -4,  -1,  -8,  11, -16,  27, -30, -34,   4, -14,  13, -11,  -2, -12, -25,  17 },
      {   0,   0,   1,   0,  -2,  -5,  -3,   3,   0,  -1,   1,   0,  -6,  -7,   7,  -3,   1,   0,   1,  -1,  -3,  -2,   6,  -8,   1,  -1,   1,   2,  -6,  -4,  22, -27,  -1,   4,   1,   3, -11,  11, -17, -47,  -3,   1,  -3,  -2, -34, -19, -33,  23,   8, -10,   5, -32, -30,  29,  34,  20,   6,  -9,  -7, -37,  31,  30,  -5, -34 },
      {   1,  -1,   1,   0,   1,   5,   4, -10,   1,  -1,   1,  -2,   0,   7,   6, -15,   1,   0,   1,  -3,   0,   6,   4, -13,   1,  -2,   0,  -5,   1,  14,  -7,  -5,   3,  -1,   5,  -4,  10,   7,  -7,   5,   6,  -7,   3, -15,  10,   5,   2,   4,  11,  -2,   5,  -4,  48,  -4,  26, -31,  -5,   2, -15,  15,   2, -37,  -9, -93 },
      {   0,   0,   0,  -1,  -2,   4,  -1,  -4,   0,   0,   1,   0,  -4,   7,   0,  -5,  -1,  -1,   0,  -2,  -7,   4, -10,   1,  -1,   0,   0,  -2,  -9,  13, -14,   5,  -1,  -2,   2,  -7,  -8,   9, -33,  48,   2,  -5,   9, -20,   0,   8,   8,  57,   0,   0,   4, -24,  37,  19,   9,  41,   5,  14,   6,   5,  50, -11,  24,  27 },
      {   0,   1,   0,   1,   0,   0,  -3,  -3,   0,   1,   1,   1,   3,   3,  -3,  -6,   0,   1,   0,   0,   0,   2,  -5, -22,  -1,   2,  -1,   0,   2,  10, -16, -16,  -1,   0,   0,  -1,   6,   7, -43, -11,  -1,  -1,   3,  -6,  24, -34, -46,  32,  -2,   6,  -9,  11,  -4, -66,  21,  14,  -1,   4, -15,  15, -47,   6,  11,  15 },
      {   0,   0,   1,   1,   0,   2,   4,  -4,   0,   1,   1,   0,  -1,   5,   4,  -9,   0,   1,   1,  -1,   0,   6,   7, -10,   0,   1,   0,   0,   1,  10,   8, -24,  -2,   3,  -2,   4,  -6,  21,   1, -39,   0,   3,  -2,   0,  -1,  44, -53, -20,   4,  -3,   5,  -9,  29, -10, -64,  27,   2,  -3,   1,  -4,  13, -43,   0,  12 },
    },
    {
      { 109, -39, -11,  -3,  -2,  -1,  -1,   0, -23,  35,  -8,   0,  -2,   0,  -1,   0, -28,  10,  11,  -2,   2,  -1,   1,   0,  -1,  -7,   5,   2,   0,   0,   0,   0,  -4,  -1,   0,   2,   0,   0,   0,   0,  -1,  -1,   0,   1,   1,   0,   0,   0,  -2,   0,   0,   0,   1,   0,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0 },
      {  -2,  29, -13,   0,  -2,   0,  -1,   0, -87,  22,  32,  -5,   5,  -2,   2,  -1,  25, -62,  15,  10,   0,   2,   0,   1,  18,  -2, -29,   8,   0,   1,   0,   1,   3,   7,  -5,  -9,   2,  -1,   1,   0,   5,   1,   2,  -1,  -3,   0,  -1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   2,   0,   0,  -1,   0,   0,   0,   0 },
      {  37,  -8,  -5,  -1,  -1,  -1,   0,   0,  -6, -57,  21,   8,   3,   3,   0,   1,  43,  -5, -58,  15,  -3,   3,  -1,   1, -20,  59, -13, -25,   5,  -5,   2,  -2,  -3,  -2,  26, -10,  -4,   1,  -1,   0,  -2,   3,   3,   5,  -4,  -1,   0,   0,  -1,   0,   2,   1,   1,  -1,   0,   0,  -1,   2,   0,   1,   1,   0,   0,   0 },
      { -11, -66,  34,   5,   3,   2,   2,   1,  19,  -7, -14,   6,   0,   0,   0,   0,  -6, -25,   8,  17,  -2,   3,   0,   1,   5,   0, -63,  19,  10,   1,   2,   0,  -8,  32, -12, -46,  12,  -2,   2,  -1,   6,  -3,  22,  -9, -17,   3,  -2,   1,  -3,   6,  -1,   7,  -3,  -4,   0,  -1,   0,   1,   3,  -1,   1,  -1,  -1,   0 },
      {  36,  58, -31,  -3,  -6,  -1,  -2,  -1,  62, -19,  34, -12,   1,  -4,   0,  -2, -17, -25,  23,  23,  -6,   4,  -2,   2, -18,  -3, -26,  29,   5,   0,   1,   0, -12,   5, -18,  -6,  17,   0,   2,   0,  -3,  -1,   0, -16,   3,   5,   0,   0,  -2,  -2,   2,  -3,  -6,   3,   1,   0,  -3,   1,   0,   1,  -2,  -2,   1,   0 },
      {  10, -17,  29,  -6,   3,  -3,   1,  -1, -22, -43,  14,  21,  -1,   5,  -1,   1,  25,  11, -10,  -5,   0,  -2,   0,  -1,   5, -23,  21,  33, -10,   2,  -3,   1,  -9,   5, -61,  31,  26,  -4,   3,  -1,  -8,  10,  -9, -45,  23,   8,   1,   1,   4,  -3,   5, -10, -18,   8,   0,   1,  -1,   5,  -1,   0,  -3,  -4,   2,  -1 },
      {  -3,  27,   9, -17,   3,  -2,   1,  -2, -34, -33,  19,  22,  -4,   3,   1,   2, -75,  30, -32,  12,  11,   0,   3,  -1,   3, -26,  -7, -15,  15,   0,   3,   0,  22,  -7,  -9, -32,   0,   7,  -1,   2,  10,   3,  17,  -7, -26,   3,  -1,   0,   3,   1,  -2,  15,  -5,  -8,   1,  -2,   1,   2,   1,  -3,   5,  -3,   0,   0 },
      {  -8, -28,  16,   5,   5,  -2,   2,   0,  12,  -7,  39,  -6,  -6,   1,  -3,   0, -46, -31, -13,  38,   3,   1,   2,   1,  31,  -4, -12, -25,   1,   1,  -2,   1,   3,   5,  12,  43, -20,  -4,  -3,  -1,  -8,   7, -34,  14,  41,  -8,   2,  -3,   0,  -3,   3, -27,   7,  16,  -1,   0,   0,   1,  -5,   0,  -9,   4,   2,   0 },
      { -21,  -9, -15,  23,   1,   1,   0,   2, -10,  48,  -1,   3,   1,   0,  -3,  -1, -14,  13, -21,  30,   4,  -2,   1,   0, -32,  27,  -7,  -4,  25,  -8,   4,  -3,  -7, -35,   6,  -7,  37,   1,   2,   0,   3, -18, -14, -46,  15,  23,  -2,   3,   4,  -7,  14, -16, -37,  11,   4,   2,   2,   0,  -1,   8,  -9, -12,   2,   0 },
      {  -5, -17,  29, -16,   4,   1,   1,  -3, -11,   4,  34, -15,  -5,  -5,   4,   1,  17,  22,  21,  26, -17,   0,  -5,  -1, -67, -15,  -6,  30,  20,   5,   4,   0,  26, -38,  -5,  -5, -10,   7,   0,   3,   2,  10, -16,  33,  -3, -11,  -1,  -3,  -3,   0,  -8,  -9,  28,   2,  -2,  -2,  -3,  -1,  -1,  -8,  -3,   8,   1,  -1 },
      {  -2, -26,  12, -25,  10,   0,   2,   0,  22,  16,  45, -14, -12,  -2,   0,  -1,   3, -30,   0,  -3,  11,   5,   1,   1,   7,   5,  28, -14, -23,   9,  -2,   1,   0, -30,  -3,  11,   1,  12,   1,   3,   6,  -2, -14, -12, -48,  16,   5,   3,  -5,  15,  -4,  37, -28, -34,   8,  -1,   1,   4,  -5,   4,  25, -19,  -4,  -1 },
      {   2, -20, -44,  35,   0,   4,   1,   2,   3, -25, -16, -22,  20,   5,   1,   1,  19, -10,   6,   4,   4,   0,   1,  -1,  -6, -47,  -1, -34,  25,  12,   1,   3,  20, -13, -20, -17, -37,  19,  -2,   4, -10,  31, -16, -10, -12,  -5,   6,  -1,   4,   3,  14, -12, -25,  -4,   4,   2,   0,   7,   1,  14,  -8, -18,   1,   1 },
      {  -2,  -9,  -7,  46, -15,   5,  -4,   4, -11, -38, -19,  22,  21,  -3,  -1,  -3, -29, -24,  17, -14,  -7,   4,  -3,   5,   7,  14,   8,  27,  -2, -20,   2,  -3, -11, -41,   9,   2,  27,   8,  -7,   2,   2,  -6, -42,   9, -20,   2,   7,  -2,  -1,  15,  -4,  -8,  23, -26,  -4,  -2,   8,  -1,   9,  -7,  11,  19, -13,   0 },
      {   3,  -5,  15,   1,  -3,  -4,   0,   0, -17, -41,  10, -35,  13,   4,   1,   1,  -5,  18,  46, -13, -16,   2,  -3,  -1,  11,   2,   4,   3,   4,   3,  -2,   1,   5,   3,  41,  -5,   6,  -2,  -3,   0,   2, -37,   1,  12,  10,  35,  -9,   7,  -3,   2, -28,  -2, -44,  23,  14,  -1,   0,   2,  -7,  12, -13, -28,  12,   0 },
      { -26, -41, -74,   2,   5,   7,   1,   0, -17, -18,  25, -29, -11,  -6,   6,   2, -25,  19,   8, -14,  -5,   3,  -1,   0, -18,  23,  -3,  16, -31,   6,  -2,  -1,   8,  17,  -9,  14,   5, -21,   6,  -4,  11,   8,  20,  -7,   6, -10,  -8,  -1,   4,   0,   7,  18,  11,   3,  -7,  -3,  -2,   0,  -4,  -7,   9,   7,   2,  -1 },
      {   2,   9,  17,   1,  -3,   2,  -1,   0,  -6,  -9, -32,  22,  -3,   2,   6,   1, -22, -45,   2, -24,  27,   4,   4,   1, -75,  -1, -10, -16, -29,  24,   0,   3,  15,   6,  -1,  20, -16, -23,   6,  -3,  13,  -7,  -2,  -1,  16,   3,  -6,   1,   0,   9, -13,  -2, -16,   2,   6,  -2,   0,   1,   2,  -4,   5, -18,   2,   2 },
      {  -1, -17,  -9, -18,   8,   7,   2,  -2,  15,  -1,  14,  16,  21, -10,   2,  -2,  -4, -38,  -8, -37,  19,   4,   3,   1,   4,  -3,  27,   1,  15,  -4,  -5,   1,   2, -32,  -2, -20,  10,  17,  -3,   3,  22,   0,  24,   1,  -1,  -1, -15,   1,  -6, -13,  -2,   1,  14,  56, -10,   0, -11,   4, -13,  -8, -45,  16,  25,  -1 },
      {   6,  -6, -14,  41,  -9,   0,  -4,   1,  16,  10,  13,  28, -11,  -5,  -2,  -5,  23,  -2, -16,  16,  10,  -6,  -3,  -2,   9, -33,   4,   0,  -8, -19,   0,  -1,  63,   7,   9,   3,  30, -22, -15,  -1,  -9,   0,  12,  17,   9,  34, -18,  -1,  -4,   0, -30,  22,   6,   6,  16,  -8,  -2,  -1,  -3, -17,  16,   3,   5,   3 },
      {  -4,  -6, -42, -68,  29,   6,   5,  -5,   0,  -8, -38,  15, -25,  -1,   8,   4,   8,  -2,  -7,   7,   5,  -3,   2,  -1,  19,  -1, -22,  18,   4,   9,   2,  -1,  24,  -9,   4,   8,  14,  10,  -1,   3,  -6,  -7, -27,  10,  -4,  22,  10,   1, -21,  14, -13, -31,  -2,  -7,  12,   0,  -6,   1,  12, -16, -12, -11,   1,  -1 },
      {  -8, -10, -28,  19,  -5,   5,   1,   1,  -3,  -1,  23,  46, -13,   5,  -3,   2,  12,  14,   5,  18,  24,  -7,   8,  -3,  -6, -19,   7,   0,  16,  21,  -1,   1, -65,   8,  -4,   0, -20,  11,  12,  -1,  11, -37,   0,  20,  -1, -10,  10,   2, -14,  -2, -37,   3,   5,  -1,  10,   3,  -1,   5,  -4, -21,   3, -22,   4,   2 },
      {   3,  -4,  -3,  -7,  23,   0,   1,  -2,  -1, -17,   2,  -8,  -7,  -4,   0,   7,  15,  -4,  13,   9,  27, -15,   9,   2, -11, -36,   1, -20,   8,   3,  11,   1,   3,   0,  25,   8,  15,  -4,  -4,  -1,   3, -54,  11, -17,   7,   0,   7,  10,   2,   5,  22,   1,   3, -12, -44,  10,  -6,  11,   2,  10,   9,  56, -28, -13 },
      {   2,   2,  -4,  -4,   9,  -4,   4,  -3,  -3,  -9, -38, -29, -17,  17,  -3,   4,  -2, -25, -32,  28, -46, -15,   6,   3,   5, -28,  10,  16,   6, -15,   3,   0, -27, -35,   0,   7,  -7, -26,   3,   3,  29,  -7,  13,   6,  11, -14, -26,   1,  -6,  11,  -7,  29,  -3,  15,  -4, -11,   0,   5,  -6,   1,  22, -11,   6,  -5 },
      {  -9, -10, -10, -24,   7, -17,   4,   2,  -9, -11, -16,  13, -15,   8, -13,  -3, -12, -31,  28,  33, -18,  -6,  -7,   0, -28,   8,  54, -17,  24, -28,  -7,  -2,  -9,  37,   5,  -6,  14,  17, -12,  -3, -32,  16,  10,  -3,  -4,   9,   3,  -3,  27, -15,  -9,   7,   7,  -4,  10,   0,  10,  -8,  -6,   0,   1,   5,   1,   7 },
      {   6,   7,   7,   7,  -4,  -4,  -6,  -1,   1, -15, -17,  21, -38,  -3,   0,   0,   9,  11,  57,  33,  27, -17,  -7,  -7,  26,  41,   4, -18,   9,  23,  -3,  -1,  17, -26, -14,  -5,  -7, -12,  10,   0,  34,  21,  -1, -12,   5,  -7, -17,   1,   1,  13,  11,  16,   1,  16,  -6,  -5,   3,  -4,  -8,   3,   6,   8,  12,   1 },
      {   0,  -4,  -2,   8, -16,   5,  -1,   2,   4,   2,  -5,  19,   2,   5,  -3,   9,  -7, -15,   3,  14, -29,  -3,  14,   1,  10,  25,  20,  20,  -1,  -2,  10,  -6,  51,  -4, -18,   6, -15,  14,  18,  -2,   1, -33,  32,   9,   6, -28,  32,  10, -16, -24,   0,  -5, -15, -13, -19,  31, -12,  22, -12,   9,  -6, -26, -19,  -7 },
      {   5,  14,  15,  38,  30,  16,  -2,  -4,  -9,  -3, -25, -40, -11,   0,   7,   5,  -7,  -1, -16,  20,  15,  10,  -1,  -5,  -4,  12,   1,  10,   0,  21,  -5,   0,   6,  16, -11,  27,   9,  43,   4,  -4, -16, -12,   0,   0, -24,   7,  21,   4,  -6,   4,  -7,  23,  11,  11,  -3,  -1, -20,  -3, -24, -25, -12,  12,  40,  -1 },
      {   1,   1,   7,   8,  30, -17,   4,  -5,  -1,  -4,  -3, -30,  23,  -2,  12,  -2,  -9,  -9,  -3,  -8,  27, -33,  -2,  -1,   6,  21,   9,   6,  32,  20,  -3,   0,  -7, -11, -18,   2,  -2,  -9,  -7,   5,   3,  16,  28,  12,  20,  28,  -1, -12, -27, -36,  -8,  -7,   6, -18,  26,   7,   2,   4,   7, -21,  21,   3, -44,   6 },
      {   5,   3,   2,   2,  30, -13,   3,   3,   4,   2,  -3, -23,  26,   1,   3,  -3,   6,   4,  -2,  30,   0,  28,  -1,  -8,   5,   2,  25, -24, -39,  -5,   5,  -4,   6,   9, -29, -37,  13, -22,   8,  -4,  43, -20, -18,   2,   1,  -6,   5,   2,  21,   4, -27, -25,  13,  -6,   0,  19,  14,  -4,  10, -22, -18,  -1,  -5,  11 },
      {   2,   1,   4,  13, -24,  -8,  -1,  -4,   4,   6,  15, -26, -46,   7,  -8,   0,  -2,  -8, -18, -36,  -5, -31,   2,   3,  -4,   8,  14,   2,  38,  -8,  -7,   5,  15,  30, -10,  -6,   1,  22,  -1,  -5,  39,  -1,  -9,  -7,   6,   0,   8,   5,   1,  48, -20, -22,   0,  -4,  -8,   2,  20,   3,  19, -10,  -7,   3,  -6,  -8 },
      {  -2,  -3,  -8, -15,  -1, -30,  12,  -3,   4,   8,  -2,   1,  47,   0,  -7,  -5,  -2,   2,  -4,  -5,   5, -13,  -6,  -1,   7,  21,  -9,  -2,  25,  13, -20,   3,  20,  -2, -28,   5,  -8,   1,  14,   2, -38, -36, -16,  -4,   2, -24, -29,   8,  31,  21, -30,  13,   0,   9, -17, -23,  -5,   1,   9,  -5,  30,   3,  11, -10 },
      {  -4,  -6,  -1, -26, -84,  19,  -7,  11,  -3,  -4, -18, -27,   8, -23,  21,  -1,   0,  -5, -20,  28,  27,   2,  -8, -11,   3,  -7,  19,  14,   3,  25, -19,  -7,  -3,  14,   1,  -8,   7,  -5,  -5,   1,  13,  -5, -20,   7,  -4,   9,  -1,  -3,   8, -24,  -1,   5,  -1,   0,  -2,   3, -12,   6,  -2,   7,   5,  10,  -5,  -1 },
      {   0,   0,   5,   9, -14,  12,  -8,   3,  -2,  -2,   2, -19, -50,  16,   4,   8,  -6,  -8,  -5,  -9,   7,   0,   3,   8,   7,   6,   6,  -6, -18,  18,  20,   1, -12, -30, -17, -32,  -2, -26,   9,  11, -71,  -6,  10,   6,  -5,   2, -18,   8,  10,  -6, -22, -18,  -1,   0, -11,   2,  13,   4,   4,  -7, -26,  10,  -9,  -7 },
      {   6,  -2,   2,   0,   5,  11,   9,   3,   6,  -8,  -1,  -5, -25,   2,  22, -20,   6,  -3,   0, -20,   3,  -8,  -9, -11,   9, -11, -26,   3,  11, -13, -14,   3,   8, -23,  21,  14,  -1,   7,  20,  -6,  12,  -5,   3, -25,   2, -18,  14,   0,  70, -44, -22,   3,  10, -12,   8,   9,  -2, -24,  -9, -13,   0, -15,  -9,  21 },
      {  -1,  -2,   8, -11, -27,  -3,  11,  -1,  -4,  -9,  -3, -23,  14, -40,   1,   1,   2,   2,  12,   7,  24, -16,  21,   1,  -1,  -7,  -3, -22,   1, -36,  22,   4,  -3, -10,   7,   8,  23,  -6,  12,  -3, -21,   7,  21, -17,   6, -27,   8,  17, -30,  20,   5,  -4,  18,  -7,  -5,   6,   7, -16,   0, -40,   2, -42,  17,  11 },
      {   2,   1,   4,  16,  30,   5,   4,   0,   1,  -1,   9,   6,  -1, -11,  18,  -4,  -3, -11, -12,   8,  14,  -9,  -1,  -2,   4,  -7,  42,  47,   6,  23,  -5, -13,  12,  22,  44, -28,  -9, -18,  39,   2, -14,   3, -13, -13,  -2, -22, -12,   6,  -4,  16,  27,  -8,   0,   3,   6,  -6,  -7, -23,  19,   8,   2, -18,  -2,  13 },
      {   2,   2,   4,   8,  14,   9,   4,   2,  -1,  -4,   6,  -2, -25, -11,  27,   6,  -1,  -3,   7, -20, -11, -25,  24, -14,  -4,   7,   0, -23,   5, -26, -14,  -1,   2,  15, -32, -16,   9, -16, -23,  13,   1, -36, -37,  19, -11,  -9, -15, -10, -12, -39,  35,   0,   4,  -2,   4, -16, -27,  -4,  10,   5,   7, -18,  23,   1 },
      {   0,   3,   2,   1,  10,  -1,   4,   4,   0,   5,   0,   6,  -2,   3,  28,   2,  -1,   4,   5, -11, -16,  22,   1,  -8,  -2,   0,  -2, -11,  18,  -1,  -8,   7,  -1,  17,  17,  -6,  10, -12,  13,  10,   2,  38, -36, -23,  -1, -13,  -4,  31, -42, -33, -43,  16, -12,   2, -42,   7,   3,  12, -13, -15,   8,   8,  -9, -26 },
      {   1,   3,   0,   9,  25, -10,  -2,  -3,  -3,   0,  -4,   4,  -7, -65,   9,  -5,  -3,   2,  -8,  10,  19, -30, -31,   2,   0,   4,   1,  22, -21, -25,  -1,  14, -13,  -3,   1,  -5, -18,   1, -28,  -2,   0,  16,   7,  13,  -1,   1,   7,  -1,  21,   0, -23, -12, -20,  -7, -33,   3,   0,  26,   2,  25,  -4, -14,  15, -26 },
      {   1,   4,   3,   1,   7,  37,   5,  -6,  -1,   4,   0,   3,  -1, -21,  20,   3,  -1,   2,   6,  -4,  -2,  38,   8,  -1,   4,  10,   8,  -7,  34,  -8,  40,   0,  -7,  10,  -8,  13,   6, -11,  -5,  19,   2,  11,   4,   9,   3,   1, -13,  -8,  33,  37,  -6,   5,  -5,   7,   3,  -1, -54,  27,  -2,  -7,   2, -11, -38,   9 },
      {   1,  -1,  -4,   8,  -3, -64,   9, -15,   0,  -4,  -6,   1, -25, -47,  -9,   9,   3,  -3, -10, -13,  -5,  46,   0,  11,   3,  -8,  -4,   2,  20,  31,  18, -10,   1,   3,  -3,  12,  21,   3,   3,  10,   6, -11,  -5,  22,  -6,  -5,  -9,   3,   9, -18,  21,   5,   0,  -1,   0,  -8,  35,  -4, -20,   5,  -6, -13,  -9,   3 },
      {  -1,   0,   1,   6,  -4, -11,  -2,  -2,  -1,   1,   3,  12,  -5, -29,  17,  -2,  -4,  -3,   0,  -3, -52, -19, -36,   0,  -2, -12,  -3, -37, -18,  38, -38, -11, -11, -13,  -3, -10,  29,  15,  11, -17, -10,   4,  13,   5,  11,   3,   9,  -3,  -9,  22,   2,  -5,   0,   5,  -7,   8, -42, -18,   4,  -8,  -1,  -1, -19,   7 },
      {   3,   4,   3,   2,  13,  56,   3,  16,   1,   0,  -2,   5,  10, -25,   4,   2,   2,   4,   1,  15, -13,  -8,  26,   6,  -2,   3,  -3,  -7,  -8,   7, -32,   5,   3,   5,  -8,  18,   7,  25,  -2,   0,   2,  -3,  10,   9, -12,  -2, -31,  16,   3,  -1,   0,   1,  -5,  14,  -7, -29,  59,  -6,   9,   1, -14,  -9, -43,   2 },
      {  -1,   2,  -1,   4,  -2, -43,   3,   3,  -2,   0,  -4,  -5,  -9,  10, -11,  14,   1,   7,  -1,   3,  22,  -8,  57,  -4,  -4,   1, -10,  17, -25, -17, -40,   6,  -3,  -3,   9, -15,  -3,  17, -10, -10,  -9,  15, -19,   4,   1,   0, -10,   5,  10,  13,   3,  16,   1,  21,  -4,  22, -28,  15, -23,  -3, -13, -10, -43,  18 },
      {   0,   1,   3,   3,  -2,   1,   0,  -7,  -4,  -5,  -5, -11, -23,  -9, -36,  -8,   1,   0,   4,   3,  13,  41, -42,   7,   0,   0,   2,   2,  -3, -24, -25,   1,   8,  -4,   2,  -2, -25,   8, -13,  -6,  -4, -23,   2, -36,  -3, -17, -14, -34, -38, -14, -11,  -3,   7,   8,  -2, -29, -17,   6,  23,  -9,  -3,  -8, -29,  15 },
      {  -1,  -2,   3,  10,  -8, -22,  12, -11,   1,   0,   2,  -2,  -4,  16,   4,   6,   1,  -1,   7,  13,   4,  22,  31, -18,   2,  -1,   4,   0,  -5,   6, -15,   4,  -5,  -9,  13,  21,  17,  10,   5,  -7,  12,  20,  34,  11, -14,  -3,  -7,  19,  13, -17,   0, -26, -12, -14,  -7, -30, -27,  13,  72,  -4,  -2,   2,  19, -20 },
      {   2,   2,   4,  -5,  17,  28, -22,  10,   1,  -2,   1,  -8,  -7, -10, -40,  17,   1,  -2,   2, -16,   5,  14,   7,  -8,   3,  -4,   0,  -6,   8,  13, -24, -17,   0,  -9,   8, -26,  27,  -2,  10, -25,  -8,  10,  -7,   9,  32, -14,  24,  -4,  14,  -5,   6,   3,   9,   1,  11,  -1,  15,  61,   2,  -3,  37, -13,  19,  -4 },
      {  -1,  -2,  -4,  -2,   9,  13,  -4,   8,  -2,  -2,  -1,  -6,  -9,  15,  -9,  -8,  -1,  -3,  -5,  -5,  28,  35, -15, -23,  -1,   2,  -1,   5,  16, -15, -11,   7,  -5,  -8, -21,  -4,  14,   3, -12, -14,  -6,   2,   3,  39,  24,  -1,  11,  22,  -6,   9,   4,  18, -12,  10, -26,  31,   1, -63,  21,  29,  17, -16,  -8,  -5 },
      {   0,   1,   0,  -2,   0,   7, -11, -17,  -1,  -1,  -1,  -8,   5,  26, -44, -14,   0,   2,  -2,  -2,  21, -33, -35,  16,  -4,  -4,  -4,  -6, -15,  -7,  24, -13,  -2,  14,  -2,   2,  33,  14,  31,  14,  15,  13, -15,  29, -11, -15, -17,  30,   4, -21,   6,  -8,  -8,  -1,  -7, -28, -26,   3, -19,   4,  -9, -16, -21,  -5 },
      {   0,  -2,   0,  -2,   0,   6, -44,  -5,   3,   1,   3,   5,   7,   5, -21,  18,   1,  -2,   0,  -1, -12,   0,  14,   2,   5,   2,   8,  11,  -6,  29,  13,  15,   2,   2,  14, -14,  -9,  13, -49,  16,   8,   3,   2, -14,  10,   7, -10,  -8,  19, -10,   8, -21,  -3,  -7, -41, -12, -10, -36, -14, -47,  26, -22,   4,  -4 },
      {   1,   1,   0,   2,  -6,  17,  64, -31,  -1,  -2,  -2,  -2,  -6, -19, -19, -14,   0,  -1,   0,  -1,  -5,  -2,  33,   5,   2,   4,   7,   0, -20,  14,  26,  17,   3,  -4,  -2, -32,  -1,  20,  -4, -11,   2,   5, -11,  -7,  29,  -2,  24,  -4,  -3, -10, -20,   4,  -3,   5,  16, -31, -11, -26,  -9,  18,  20,  19,  -6, -18 },
      {  -1,  -1,  -1,  -1,  -2,  -9, -10, -29,   2,   2,   1,  -1,   1,  40,  63, -20,   2,   1,   2,   5,  18,  11,  -3,   0,   0,  -1,   5,  -2,  -7,  -2, -11,  -1,   1,  -2,   4,  -8,  29,  14, -19,   3,  -5,  -1,  15,   8,  16, -30,  14, -34,  -7,  19,   0, -12,  -1,  -5, -11, -39,  21,  14, -32,  17,   6, -17,  -5,   0 },
      {  -2,  -4,  -1,  -4,   2, -15, -69,   6,   0,  -2,  -2,  -3,  -5, -22,  22,  -3,   0,  -2,  -1,  -2,  -2,  12,  18, -15,  -1,   0,   0,  -4,  -9, -29,  18, -12,  -2,  -3, -11, -10, -25,  34,  43,  -6,   1,  -6,   1,  -8,  24,  25,  -7,  20,  -6,  -2,  -9,  -2,  11,  -2,  17, -30,  -7, -11,  -3,  17,  16,  14,  -5,  -9 },
      {   2,   2,   1,   1,  -5,  -7,  53,  53,   1,   1,   1,   4,  -7,  16,  24,  20,   0,   0,   2,  -3,   5,  12, -37,  -1,   0,  -1,   2,   7, -14, -14,   5,   1,  -1,  -3,  -1, -16, -17,  32,   0,   6,   3,  -6,   4,   7,  28,  18, -13,  30,   6,   4,  33,  -6,  -2,  -6,  -1, -14,   4,  18, -12, -35,  18,  -3,   1,  -5 },
      {   0,   1,  -1,  -1,   9, -15,  -3,  89,  -1,   0,   1,  -5,  -3,  -1,  -8,  -1,  -1,   0,  -3,  -2,   2, -10,  11,  -8,   0,   0,  -3,   0,  -1,  20,  30,  -3,   3,   1,   3,   6,  22,   9, -20,   5,   4,   7,   4,  -4,  -1, -31,   2, -17, -12,  -1, -31,  -4,  -3,  -4,   0, -19, -13,  -6,   2,  34,  -3,   6,   4,  32 },
      {  -1,  -1,   0,  -1,  -5,  -6,  -1, -23,   0,   1,   0,   2,   1,   6,  14,  -6,  -2,  -1,   0,   0,  -7, -19,  -8, -21,  -1,   1,  -5,   0, -17,   5,  47,  -8,  -1,   3,   2,  -3,  -2,  29,   0,  -3,  -7,   2, -18,   3,   5,   8,   9, -13,   1, -13,   4,  33,  -6,  39, -29,   9,  22,  23,  52, -11,  10,  -3,  -5,  50 },
      {   0,   1,   0,   1,  -2,   2,  15,   3,   0,   1,   2,   1,  -3,  -5,  -9,   2,   0,  -1,  -2,  -5,   0,   2,   8, -28,   0,   1,   1,   2,   6, -17,  -6, -18,  -3,   0,  -7,   6, -11, -25,  26, -26,  -1,   6,  10,  17, -39,  18,  17, -15,  -3,   1,  -1, -48, -31,  25, -32, -21,   3,  -8, -34, -11,  34,  19,   0,  36 },
      {   0,   1,   0,   4,   0,  10,  16, -21,  -1,   0,   0,  -3,  -5,   3,  -4,  -7,  -1,   0,  -2,  -1,  -3,   7,   0,  -9,  -1,  -1,   1,  -5,  -6,  12,  -5,  13,  -2,   2,   1,   3,   5,  21,  -1,  17,   2,  -2,  11, -16,  -6,  19, -46,  19, -11,  -1, -13, -38,  46,  -9, -11,  36,  -3,  17, -13,  47,  34, -16,  18,  34 },
      {  -1,   0,   0,  -2,   1,  -3, -11,  17,  -3,  -3,  -3,   0, -10,  -7,  -7,-105,  -1,   2,   0,   1,  -7,   9,  16,   7,  -1,   0,   2,   2,  -2,   5,  -2,  22,  -1,   1,   0,  -4,  -1,   6,  -1,   5,  -4,  -4,   2,  12,   0,  -9, -14,   4,  -4,   1,  20,  -8, -25,   3,   7,  22,   0,  17,  -1, -29,  13,  15,   6,  -4 },
      {   0,   1,   0,   1,   0,  -1,  12,  -3,   0,  -1,   1,   1,   3,   0, -18, -20,  -2,  -2,  -1,  -3,  -6,  -8,  -3, -99,  -2,  -1,  -3,   3,  -2,   9,  11, -37,  -2,   0,   6,   4,  -4,  -1, -13,  -1,  -6,  -2,   1, -11,   1,   2, -17,  -9,  -2,  10,   5,   5,  13, -14,   1,  -1,   7,   4, -15,  -4, -26, -12, -12, -37 },
      {   0,  -1,   0,   1,  -1, -11,  -7,  15,   0,   0,   0,   0,   1,   8,  -1, -20,  -1,  -1,  -1,   1,  -7,  -2,  -8, -10,   2,   2,   4,  -2,   6,  14,   4,  52,   1,  -1,  -1,   2,  -4, -31,  -8, -10,  -2,  -8,  -2, -11, -21,  26,  48,  36,   4,   2,  13,   1,  46,  32,  -4, -33,  -3,  14,   1,  12,  -2, -30, -20,  -8 },
      {   1,   1,   1,   2,   0,  -6,   1,  15,   0,  -2,   0,  -3,  -6,   4,   9,   7,   1,   1,   4,   2,   6,  -3,  -2,  12,  -1,  -3,  -3,  -5, -14,  -3,   9, -27,   2,   2,   5,   7,   7,  13,   0,  24,  -1,   0,   3, -10, -32, -12,   6, -22,   6,  -4,  -4, -22,  11,  61,  23,  30,   5, -10,  12,   1,  44, -10, -16, -60 },
      {  -1,  -1,  -1,   1,  -1,  -1, -13, -19,   0,   0,   0,   1,  -2, -10,  -5,  22,  -2,  -2,  -2,   0,  -3,  -2, -16, -41,  -1,   1,  -1,   3,  -9,  -5, -15,  27,   1,  -3,   2,  -2,   2,  -1, -16,  63,  -2,  -4,   8,   0,  -1, -36,  22,  38,  -3,   8,  -2,  -8, -26,  13,  42,  -1,   6,   1,  -3,  -5,  11,  32,  -3,  26 },
      {   0,   0,   0,   0,   2,   1,   5,  -6,   0,   0,   0,  -1,  -2,   0,   3, -26,   0,   2,   1,  -1,   0,   3,  10,  28,   0,   2,  -3,   2,   2,   3, -11, -88,  -2,  -4,  -5,  -1, -18, -11, -43,  15,   1,   1,   4, -11,   3,   5,  25,  53,   2,   4,  -9,   1,   8,   4,  -5, -13,  -4,  -1,   4,   2,   1,  -1,   4,  18 },
      {   0,   0,   0,   1,  -1,   0,  11,  13,  -1,  -2,  -1,  -1,   1,  -2, -18, -11,   0,   0,   1,   1,   1,   2,  13,   1,  -2,  -1,  -2,  -3,  -4,  -6, -26,  -4,  -1,  -1,  -1,  -3,  -1,  -2,  40,  86,   2,   5,   5,   6,  19,  34,  28, -37,   3,   6,   0,  13,  10,  -8, -34,  -9,   2,  -2,   9,   5,  -8, -10,   7,  -1 },
    },
    {
      {-102,  63,  -5,   4,   0,   1,   0,   0,  -3,  -8,  13,  -6,   2,  -2,   1,   0,  26, -24,   4,   1,  -1,   1,   0,   0,  13,  -4,  -6,   2,   0,   0,   0,   0,   7,  -2,  -1,  -1,   0,   0,   0,   0,   2,   0,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      { -34, -40,  34,   1,   5,   0,   3,  -1,  69, -66,  -1,  13,  -4,   2,   0,   1,   8,  24, -34,  11,  -2,   0,  -2,   1, -19,  26,  -2,  -9,   2,  -2,   0,   0,  -4,   6,   2,  -4,   0,   0,   0,   0,  -5,   4,   1,  -2,   0,   0,   1,   0,  -3,   2,   2,  -1,  -1,   0,   1,   0,  -2,   1,   1,  -1,   0,   0,   0,   0 },
      {   6, -33,  46, -18,   7,  -4,   1,  -1, -65, -18,  40,   6,  -1,   2,  -1,   1,  33, -39, -22,  25,  -6,   5,  -1,   1,  15,  13, -42,  10,   4,   0,   1,  -1,  -2,  15, -10,  -8,   5,  -2,   2,  -1,   2,   4,   1,  -6,   1,   0,  -1,   1,   2,   1,  -2,  -1,  -1,   0,  -1,   0,   2,   2,  -1,   0,   0,  -1,   0,   0 },
      {  50,  40, -44,   1,  -8,   1,  -2,  -1,  42, -32,  38, -20,   2,  -2,   0,  -1,  34, -41,  13,  12, -13,   6,  -3,   2, -20,  26, -21,  16,  -4,  -2,   1,  -1, -12,   7,   0,   2,   4,  -4,   1,   0,  -8,   6,  -1,   1,   3,  -1,   0,   0,  -5,   1,   3,  -2,   3,   1,   0,   0,  -3,   1,   0,   0,   1,   2,  -1,   0 },
      { -14, -24,  35, -24,  11,  -3,   6,  -1,  33,  10,  32, -35,  -1,  -2,   0,  -1, -32, -13,  57,   3, -18,   5,  -3,   3,   6, -32,   7,  40, -22,   3,  -4,   2,  -1,   6, -34,  24,   6,  -6,   1,  -1,  -2,   6, -15,  -3,  12,  -3,   0,  -1,  -1,   4,   0,  -5,   1,   3,  -1,  -1,  -1,   2,  -2,   0,   0,   2,   0,  -1 },
      {   6,   0,   1,  -8,   8,   0,  11,  -6, -42, -47,  22,   0,  14,   0,  11,  -2,  23,   7,  15, -26,  -6,  -5,  -4,   3, -14,   7,  79, -23, -36,   3, -11,   5,   2,  -8,   0,  20, -14,   3,  -5,   2,  -9,   8, -10,   8,   1,   1,   4,  -3, -11,   0,   8,  -3,  -3,   4,   5,  -2,  -3,   2,   0,   0,  -1,   1,   0,   1 },
      {  -9, -46,  -9,  19,   5,   5,   6,  -2,  24,  31, -43,   6,   3,  -3,   4,  -2,  61, -48,  23, -15,   6,  -5,  -2,   1,  25,  -3,  11, -14, -18,   4,  -7,   4, -20,  31,   1,  -6,  -6,  -1,  -1,   1, -15,   6,   8,  -1,  -2,   2,   1,   0, -13,   5,   9,  -3,  -5,   2,   1,   0,  -3,  -1,   3,   1,  -1,   0,   1,   0 },
      {  11,  23,  35, -62,  16, -10,   3,  -2,  21,  50,  -1, -31,  -7,   3,  -4,   0,  47,  24, -29,   3, -15,   5,   0,  -2,   1,  14,  12,  -6,  -3,  -3,   2,  -2,  -6, -13,  26,  15, -10,   2,  -1,   0,  -4,  -7,   0,  21,  -3,  -2,   0,   1,  -2,  -5,  -6,   9,   7,  -3,   1,   0,  -1,  -3,  -2,   1,   4,   0,   0,   0 },
      { -10, -26, -11,   7,  12,   2,   1,  -1,   9,   9,  36, -38,  12,  -4,   0,  -1, -46, -33,   1,   1,  12,   9,  -1,   2,  21,  26,  -8, -50,  -5,   5,  -1,   2,  10, -12,  47,  -7, -29,  11,  -5,   3,   5,  -3,  -9,  25, -21,   3,   2,  -1,  -3,   8,  -1,  10,  -4,  -8,   2,  -2,   0,   1,  -4,   6,   2,  -2,   2,  -1 },
      {  29,  29,  17, -18,  -3,  -4,  -4,   0,  22, -38,  -9,  22,  -3,   6,  -5,   4,  -3, -18, -19, -10,  17,   3,   3,  -3,  68, -60,   0, -15,   7,   9,   1,   0,  12,   0,  -8,   9, -21,  11,  -4,   3,  -5,  17,  -9,  13, -10,  -3,   1,  -1,  -3,  10, -11,   7,   4,  -6,   2,  -1,  -3,   1,  -1,  -2,   3,  -4,   1,   0 },
      {   8,  27,  37, -34,  21, -12,   5,  -5,   3, -10, -22,   8,  18,  -3,   5,   0, -18, -31,  29,  -4,  14,  -3,   1,   1, -45,  -9,   6, -11,  18,   6,  -4,   4, -19,   7,   3, -49,   5,  17,  -5,   3,   9, -10,  26, -11, -30,  20,  -3,   1,  -9,   3,   6,   8, -24,   4,   5,  -2,   2,  -1,   2,   7,  -8,  -6,   3,   0 },
      {  -4, -16, -34, -17,  30,  -2,   0,   1, -12, -30, -19, -52,  38,   3,   2,   3,  15,  38,  -1, -41, -11,  12,  -2,   2,  20,  -6, -32,  25,  -9,   8,   3,  -3,  13,   8,  -9, -21,  -2,   2,   4,  -2,   5,   0,  22,  -5, -22,   1,  -1,   1,   8,   0,  -4,  15, -11,  -7,   2,   0,   1,   3,   1,   6,  -1,  -7,   1,  -1 },
      {  -2,   6, -37, -22,  40, -10,  17,  -8,   3,   8,   6,   3,   3,   1,   2,  -1, -23,   4, -36,  50, -15,   1,  -3,   4,  34,  -2,  21, -13, -15, -10,  -7,   4, -13,  13, -12,  -8,  21,  -2,  -7,   1, -11,   7,   7, -40,  11,  14,  -1,  -2, -13,  -2,  33, -14, -25,  14,   2,  -1,  -3,   0,   5,   4,  -9,   1,   3,   1 },
      {  10,  29,  36,  17, -26,   6,   1,  -3,   2,   2,  25,  -1,  -1,  -8,   8,  -2, -14,  37,  19, -31,  11, -13,   0,   0,  50,  37,  -9,  -3, -18,  -4,  -7,   3, -31,  42,  12, -13,  13,  -6,  -1,  -1, -18, -14,  25,  -9,   8,   8,  -1,   1, -11,  -6,   5,  -1,  -4,   9,  -1,   0,  -3,  -4,   2,   2,  -4,   5,   2,   0 },
      {   3, -17,   2,  28, -16,   4,  -1,  -1,  10,  12,  48, -19,   0, -11,   4,  -2,  39,  29,   9,  18,  11, -21,  -2,   0,   3, -42,  15,  -1,  38, -11,  -6,   2,  25, -23,   5, -34,  15,  17, -15,   7, -15,   7,   7,  -5, -16,  23,  -9,   0, -11,   1,   4,   6, -14,  11,   4,  -2,  -8,   4,  -1,   5,  -6,   3,   2,   0 },
      {   2,  -1,   1,   8,  13,  -9, -10,   2,  12,  18,  27,  -2,  30, -17,  -8,  -1,   8,  -5, -29, -31,  63,   4,  -3,  -3, -12,   2,  -4,  -5,   6,  32,  -7,   1, -24, -13, -28,  37,  -1,  16,   8,  -8,   5,  -7, -12, -36,  12,   7,   6,   1,  19, -14,  -5,  -1, -22,   7,  -1,   4,   1,   5,  -6,   4,  -8,  -5,   4,  -1 },
      {  -3,   8,  -3,  21,  28, -20,   3,  -6,  -9,   2,  11,  -3,  14, -10,  -1,  -1,   3,  30,  10,  35,   1,   3, -14,  -2, -24, -32, -14, -17,  12,  25,  -5,   3, -28,  49,   1,  -4, -44,   8,   6,  -5, -28,   6,   4,  27,  11, -25,  15,  -8,   7, -13,   0, -13,  21,  -6,   2,   2,   2,  -6,   1,  -4,   3,   5,  -3,   3 },
      {   5,  -4,   1,   5,  31, -21,   0,  -4,  -2, -24, -20,   5,  30, -16,  -4,   0,   8,  -2,  27,  12,   0,  -7,  -4,  -3,  24,  -1,  -2,  11,  24, -10,   1,  -3,  -5, -19,  45,  15,  16,  -2,  -8,  -1,   9, -40, -14,  15,  40,  17, -10,   2,  10,  -9, -25, -24,   2,  34,  -7,   3,   0,   2,  -8,   0, -20,  15,  -1,   0 },
      {  13,  21,  36,  24, -10, -10,   9,  -7,   2,  10, -28, -26,  42, -17,  13,  -5,  -4, -10, -19,  -1,   4,   3, -10,   7, -12,  10, -15,  -4, -16,   0,  -5,   4,  61,  -5,  -7,  -8,  -5, -11,  -4,   0, -13,  32, -10,  -5,  21,  -4,  -2,  -5, -13,   8,  27, -30,   7,  12,  -5,   0,  -5,  -3,  11,   1, -11,  14,  -3,   0 },
      {  -3,  -8, -14, -40,  -9,  19,  -2,   1,  -3, -10,   2, -25, -33,  19,  -2,   1,   3,  -1,  12, -23,  14, -16,  13,  -5, -14, -10, -10, -50,  33,  -9,   6,  -1,  27,  36,  -7,  10,  -4, -14,  -5,  -2,   4,  -2,  10, -10,  34,  -8,  -9,   4,  15,  -9,  16, -13,  -3,  27, -21,   7,   0,   0,   0,  10, -15,  15,  -6,  -1 },
      {  -2,  -2, -11, -18,  33, -23,   9,  -7,   4,   7,   4,  26,  -4, -20,  -1,  -2,   0,   8,  29, -11,  11, -22,  -6,  -1,  22,  52,   0,   1,  41, -10, -13,  -1,  16,  -9, -36,   0,   4,  25,  -6,   1,   8,  36,  -4,  10, -15, -10,   7, -10,  -4,  -4,   9,   4,  33,  -8,  -1,   1,   2, -10,   4, -10,   4,  17,  -8,   4 },
      { -13, -11, -11, -18, -20,   9,   5,   1, -11,  -5,  -4, -24,  21,  -9,   9,   2, -21, -18, -37, -18,  21, -19,   9,   3, -11, -17,   9,  31,  21, -26,  -3,  -1, -51,   4,  18,   7,  15,   0, -16,   9, -24,  30,   3,  20,   0,  -5, -17,  -7, -18,   6,   6,   1,  33,   7,  -5, -11,   2,  -8,   4,  -8,  -2,  17,  -1,  -1 },
      {   3,  -5,  -2,   6,  18, -22,   6,  -1,  16,  12,  33,  27,  11, -17,   7,  -4,  -1, -14, -20, -22,  -4,  12,   1,   3,  -3, -21,  20,   0, -22,  -9,  11,  -3,  16,   3,   4, -30,   9, -31,  12,   3,  24, -14,  41,   0,   2, -26, -15,   9,  19, -15,  -6,  17,  33,   5, -29,   6,   1,   1,  -7,  -6,   3,  30, -16,  -2 },
      {   6,   7,  20,  31,  -2,  -6,  -1,   1,  -2,  -3,  -7, -30,   4,  18, -10,   1,   5,   1,   2, -13, -41,  30,   0,  -3,  13,  -1,  11, -35,  27,  -8,  17,  -3, -51, -23, -22, -20,  15,   4,   7,  11,  19,  17, -31, -14,  -7, -15,  -3,   4,  21,  -2,  10, -13,   8,  -3, -21,   5,   9,  -1,   1,   4, -10,  11,  -8,  -5 },
      {   2,  -4,   3,  -8, -36,  25,  -4,   5,   1, -12, -16, -25,  11,  -5,  11,  -3,   3,  -4,   0,  42,   9, -22,   0,  -1,  21,   8,  18,  15,  -1,  24, -23,   4, -12, -19, -12,  -6, -26,  19,   9, -12,  50, -18,  28,   1,   2, -27,  15,  -3,  -3, -27,  23,   3,   9,   8,  -9,   8,   1,   5,  -4,   1,  -5,  14, -11,   4 },
      {   7,  10,  17,  38,  17, -12, -16,   2,  -1,  -3, -16, -35,  -8,   5,  -7,   7, -11, -26, -22,   7, -15, -17,   4,   0,   2,  14,  38,  13,  23, -18,  -9,  -2,  22,  41, -13,  22,   7,  13, -15,  -1,  -7,  -5,  19,   5,  -3,  13,   1,  -4,  34, -24, -16,  30,  -8,  -6,   3,  -4,  -3,   7, -16,   4,  16,  -5,   6,  -3 },
      {  -5,  -5, -10, -18, -26,  23, -14,   5,   9,  19,  16,  29,  34,  17, -22,  -1,  -5,  -2, -14, -14, -33,  28,  -9,   1,  -1,   6,  12,   7,   3,   8, -15,  -2,  13,  32, -10, -22,   0,  37, -14,  -6,  11,   2, -13,  23,  -1,  13,   4, -15,  10, -13, -19, -20,   5,  31,  14,  -6,  -8,   4,   1,  -2, -31,   3,  15,  -1 },
      {   3,   3,  -7,  -5, -13,  -7,  18,  -1,   3,   2,  -1, -11,  11, -31,  25,  -4,   8,  11, -11,  10,   7, -22,   8,   6,  -4,  -8,  -3, -22, -11, -13, -10,  10,  -3,  39, -25,  -6,  18, -11, -14,   6,  52, -32, -51,  14, -10,  11,  -9,  -1,   4,  25, -18,   0,   4, -17,   2,  -9,  -1,   1,  16, -14,  17,  -5,   0,  -2 },
      {  -6,  -9, -12, -24, -49, -42,  20,   4,  -3,  -8, -13,  -2, -19, -51,   9,   3,  -3,  -1,   7,   5, -10,  11, -27,   7,   3,   1,   2,  -5,  -4,  23, -14,  -1,   0,   1,  19,  -9,   8,  -1,   5,   1, -15,  32, -10, -13,   2,   5,  -5,   7,  42, -22, -13,  16,  -9,   4,  -3,  -4,  10,  -6, -15,  18,  -3,  -7,   8,  -7 },
      {   0,  -4,   0, -18,  -4,  32,   4,   1,   6,   1,   1,   1,  29, -12,  38, -17,  -6, -12,  -1,   1, -10, -14,  20,  -6,   8,   4,  11, -11,  19,  -8,  13,  -3,  10, -10, -12, -18,  13,  -8,  31,   1, -51, -21,  -8,  -4,   3, -15,  26,  13,  17, -30, -23, -18,   0, -24,  12,  16,  10,  -5,   5, -15,   4, -24,  -8,  13 },
      {   0,  -3,   0,  -7,  -7,   3,  -3,   1,  -3,  -8,  -3,  -7, -15,  -4, -17,   5,   2,  -5,  -2,  -5,  11,  26, -27,   6,  -1, -10,   8,   7,   1,   3, -13,  -5,  14,  11,  23,   7,  31,  30,  -2, -16, -20, -41,  -6, -19, -20,   6,  36,  -7,  18,  23,  27, -19,  19, -34,  -9,   5,   0, -11,  17,  -4,   6,  38, -27,   9 },
      {  -2,  -5,  -5,  12, -15, -40,  46, -20,   0,   1,   4,  -4, -23,  -1,  23,  -4,  -6, -11, -19, -33, -33,  10,  -1,  11,  -4, -10, -17,  -6,  19,   1, -16,   6,  -4, -18, -21,  19,  -4,  19,  -9,  -5,  -1, -39,  30,  19,   9,   0,  13,  -5, -31,  -6,   8, -15,  -3,   5,  14,   6,  -5,   5,   0,  -4,  -9, -19,   5,   6 },
      {   2,  -2,   5,   1,  -3,   9,   4,   4,   5,   0,   4,  26,  28,   0,  11,  -4,   6,   5,  12,  -3, -10,   5,  15,  -5,  -4, -17, -28,  -8, -12, -35,   3,   7,  -3,  -4,  13,  31,   3,  -1, -34,   2,  12,   3,  -6,   6, -16,   7,   5, -17,  24, -55,  55,  13,  -9, -13,  10,  -5,   3, -20,  -4,  22,  12,   0,  -1,   7 },
      {   0,   3,   0,  -3,   6,  26,   0,   0,  -6,  -4, -17,  -5,  -6,   1,   7,   6,  -2,   3,   6,  18,  10,  23, -11,   6,  -4,   4, -21, -31, -22,  15, -15,   7,  -7, -26, -41,   1,  23,  -5, -25,   6, -40, -19,   4,  23, -10,  14, -16,  -4,  -1, -20, -17,  27,  15,  17, -13, -11, -26,  18, -19, -14,   5,  33,  -2,  -7 },
      {  -6,  -5,  -3, -13, -38, -19,   1,   2,  -3,   2,  -2,   9,  54,  -9, -28,   4,   1,   6,  17,  21,  -4,  29,   1,  -1,   3,   7,  10, -18,  30, -15,  22, -12,   6,  15,  -9,  36,   5, -10,  12,  -1,  -6, -10,  19,  -6,  -6, -11,   0,  16, -33,  20,   4,  24,  -5,  -7, -17,   3, -21,  19,  -4,   4,  18,  -6,   1,  -9 },
      {  -1,   2,  -2,  -1,  -2,  -6,  14,   2,   1,   6,   2,  18,   9,   9,   4,   2,  -2,   1,  -4,  -4, -20, -31,  29,   8,  -2,   8, -10,  19,  -3, -11, -26,  18,  -1,  -5,  -1,   0, -40,  21,  -5,   9, -33, -22, -20, -13, -21, -21, -14,  17,  26,  26,   2,   0, -17,   6, -48,  -5, -12,  18,   1,  27,  -7,  13,  -9, -19 },
      {   0,   1,  -2,   5,   0,  -6,  45, -14,   0,   1,  -1,   0,   0,  58,  25, -17,   1,   2,   1,   4,  32,  31,  -2, -10,   3,   8,  12,  27,  21,  22, -15,  -5,  12,  21,  14,  -8,   8,  -9,  13,  15,   5,  -2, -38,   3,  -4,  -8, -12,  17, -16, -27,   3,  20,  -6,   0,  -8,   3, -15, -19,  -8,   9,   6,   6,  -5,   5 },
      {   3,   5,   3,   3,   2,  35,  40, -31,  -1,   2,  -3,  -1,  -2,  -2,  17, -22,   2,   6,   5,  10,  15,  21, -29,   9,  -2,  -1,  -2,   4,   2, -46,  11,  18,  -7,   3,   7,  17, -22,   2,  -7,  -4,  17,  25,  21, -21, -14,   6,  23,  -7,  25,  18, -36,  -8,  -1,   1,  -7,  15, -14,  13, -21, -10, -17,   3,   6,   3 },
      {   2,   5,   6,   4,  23,  31, -21,  -5,  -1,   2,   2,  11,   2, -17,  26, -15,  -3,  -1,  -4, -16, -36, -12, -17,   7,  -5, -14, -12, -10,  20,  19, -41,  -3,  -4,   3,  27,  13,  31,  -5,  17, -24,  13,  17, -11,  -7,  10, -28,  14,   1, -11,  18,  10,  32,  -8,   5,  -4,   2, -38,   7,  -9,   3,   8,  -1,  15,   2 },
      {  -3,  -3,  -2,  -6,  -8, -36,  -1,   0,  -2,  -2,  -3,  -9,  -5,  12, -10,  15,   0,   0,   6,  -3,  15,  -6,  39, -17,  -3,  -3,  -4,  -2, -21, -16,   3,   0,  -4,  -3,  14, -28,   7,  13,  -9, -18,   2,  12, -16, -18,  15, -15,  27, -18,   3, -18,  -7,  -4,  13,  -9,  12,   8, -69,  39,  -9, -16,   4,  -6,   0,   9 },
      {  -1,   1,  -2,   2,  -2,  -5,  -4,   5,   0,   4,  -1,   3,   8,   8,  15, -12,  -1,   4,  -1,   1, -21,  -7,  29,  -8,  -8,  -6, -12, -10, -14,  20,  -3,   2,   3,  12,  13,  15,  14,  31, -20,   0,   9,  30,   5, -16,   6,  26,  10,  15, -36,  -7, -35,   0,  -3, -12, -17,  28,  38,  12, -25,  -1,   9,  22, -51,   1 },
      {  -1,  -2,  -1,  -7, -14,  -9,  10, -11,   5,   5,   6,   6,  26,  49, -12,  -7,  -1,  -2,   4,  -2,  -3, -25, -30,  19,   4,   2,  -2, -10,   9,  11, -26,   4, -10, -17,   1,  -6,   4, -28, -11,  -8,   0,  12,  21,   4,  25,   9,  -9, -22,  34,  13, -12,  -4, -11, -15,  19,  -9,   1,  17,  27,   2,  38,  -1, -38,   4 },
      {  -2,   0,  -1,  -3,   7,  11, -18,   5,  -2,   3,   1,   7, -16, -31,  19,   3,  -5,  -2,  -9,  -9,  -3,  10, -19,   0,   4,   8,  11,  21,  19,  18,  30,  -5,  -7,   9,  -5,  -2, -16, -24, -14,  21,   5,   3,  -7,  26,  -7,  19,  -6,  -2,  -5, -10,  15, -22, -23, -17,  -1,  21, -40,  40,   5,  29,  -9,   1, -47,   4 },
      {   1,   2,   1,  -1,   7,  12,   7,  16,   1,   0,   1,   0,  -1,  -2, -30,  49,  -1,  -2,  -3, -12,   5,  -8, -44,  23,  -3,  -5,  -4,   1,   3, -34, -29,  17,   3,   0,   2, -16,  -8, -17,  -4,  -1,   2,  -6, -22,  -9,   5, -16,  16,   6, -42, -43, -16,  -2,  -9, -16, -18,   9,  20,  11,  -7,   1,   1, -15,  -4, -17 },
      {   1,   0,   3,   3,   6,  -2,  -4,  44,   0,   0,   3,  -1,  -5,  -1,  30,  12,   2,   2,   8,   6,  -2,  32,  12, -18,   2,   2,   4,  -1,  12,   5, -28,  14,   3,   0,  15,  22,  -4,  -4, -16,   3,   7,  14,  24, -19, -15,  -6, -18,   8,   2, -16, -19, -20,  -3,  -5, -11, -16, -12,   3,  67, -35,  -6,  -8,   5, -15 },
      {   0,   0,  -2,   0,   0, -10, -20, -18,  -1,  -2,   0,  -2,  -6,  -2, -13, -19,   2,   3,   0,  10,  26,  25,  23,  -4,   1,   2,   5,   4,  -9, -14, -26,  34,  -2,  -4,  -9,  -8,  36,  -6, -12, -13,  -2,  15,  11,  56,  11, -23,  19,  16,  11,  13,   3,   1, -30, -10, -20,  32,  16,  -2,  16,  17, -20, -18,   1,  -9 },
      {   2,   3,   2,  -3,   7,   9,   5,  24,   3,   2,   1,   4,   2,   1,  20,   9,   6,   5,   7,   2,  12,  30,  22,  19,   5,   6,   9,   5,  18,  13, -15,  10,   5,   4,   8,  -7,   8,  -8, -23,  15,  -1,  -6,   5, -10,  15, -20,  -5, -15,   8,  24,  19, -14,  24, -17,  20, -18,  38,  52, -48,  12,   4, -15,  18,  -3 },
      {  -1,  -2,  -2,  -2,   0,   3,  -7,  19,  -1,  -1,   0,   0,  -1, -11,  12, -13,   0,  -3,  -2,  -2,  -5,  22,   2, -26,   1,   5,   8,   4,  10,  -8,   3,  13,  -2,  -9, -13, -20, -25,  -6, -31,  -4,  -5, -14,  -2, -23,  43,  -5,  16, -39, -19,  16, -28,  30,   1, -15,   4,  -7,  -8, -45,   9,  45,   4,  15,  -2,  -9 },
      {   4,   6,   5,  10,  11,  25,  66,  33,   1,   2,   2,   3,   0,  -9,  -4,  32,   0,   0,  -1,   1,   5,  -3,   9, -25,  -2,   0,   0,   0,  -5,   1,  12, -47,  -2,  -1,  -4,  -3,  13,  12,   7, -39,   0,  12,   2,  27,  13,   1,  14,  -1,  10,  11,   1,  17, -13,  10,  -2,  -6,  14,  22,  23,  21,  -2,   5,  -7,  -5 },
      {   1,   1,   2,  -1,   7,  35,  25, -42,  -1,  -4,  -1,   0,  -4, -27, -44, -22,  -1,  -3,  -1,  -8,  -8,   1,   5, -22,  -1,   0,   1,   7,   1,  27,   8,   1,   4,  -2,  10, -11,   0,  22, -37,  19,   3,  -2,   4, -11,  27,  -7, -19,  19,   1, -14,   7,  -8,   8, -19, -18,  -4,  -5,   6,  14, -11,  37, -22,  16, -19 },
      {   1,   3,   2,   3,   5,   9,  28,  60,  -1,   0,   2,  -1,   6,  -2, -23,   9,  -1,   1,   0,   1,  -2,  -7,   2,  26,   0,   2,   1,  -2,  -1,  14,   0,  20,  -1,  -1,  -1,   0,   8,   5, -16,  14,   2,   3,  11,   4,   7,  -1, -11,  16,  14,  20,   4,  -5,   7,   3,  -1,  37, -35, -50, -28, -12,   0, -30, -15,  29 },
      {   0,   1,   1,   1,   0,   2,  -8,   2,  -2,  -2,  -3,  -3,  -8, -21, -16, -24,  -1,   0,  -1,  -5,  -1,  29,   5, -32,   0,   0,   0,   4,   3, -29, -47, -22,  -4,   0,  -5, -10, -30, -14,   3, -19,  -9,  -7, -14,  -9,  -6,  12, -25,  -6,  -1,   5,  11,   9,   4,  32,   0,  22,  18,  14,  -5, -17,  23, -20, -27,  49 },
      {   1,   1,   3,   6,   2,  18,  13, -38,  -1,  -1,  -1,  -2,  -5, -23, -25,  25,   3,   2,   3,   3,  -4,  17,  52,  52,   2,   2,   4,   2,  13,  17,   4,  22,   4,   4,   8,   7,   0, -13,  12, -12,   0,   3,  -6, -12,  -8,  -8,  -7, -25,  -5, -10, -13,  20,  -4,  10,  13, -24,   1, -11,  18,  -1, -21,   1, -37,  21 },
      {   0,   0,  -1,  -2,  -3,  -6,  13,  22,   0,   0,   1,   1,  -2,  -3, -11, -22,  -1,  -1,  -5, -10,  -8,  -3,  12,   0,   4,   3,   6,  13,  20,  23,  22,  34,   0,   0,   3,  -9, -11, -37,  -8, -35,  -4,   2, -12,   9, -16,  17,  32, -24,  -2,  -3,  16,  -5, -29,  18, -21,   2,  11,   1, -16, -47,  28,  25,  18, -22 },
      {   2,   3,   2,   4,   3,  -2,  22,  29,  -2,  -4,  -3,  -6,  -5,  -7, -46, -56,  -2,  -2,  -4,   0,  -2, -20,  -4, -32,  -3,  -4,  -5,  -7,  -4,   2, -11,  19,   4,   3,   3,  14,  18, -17,  15,  22,   0,  -3,   4,  -4, -29, -25,   3,  -5, -11, -10, -13,  -9,   3,   0,  36, -24,  -2,  17,  -4,  22, -28,  12, -15,  -5 },
      {   0,   0,  -1,  -1,   7,   5,   0,  23,  -2,  -3,  -1,  -1,  -7,  -6,  -4, -27,   0,   1,  -1,  -4,   5,  18,  14,  31,   0,  -3,  -1,   0,   2, -23,  -8,  29,  -4,  -2,  -4,  -9, -12,  22,  53, -13,   7,   8,   2,  11,  29,  52,   3,  33,  -2, -12,  10,  -4,  21,  -7,   3, -27, -15,   9,   6,  22,  27,   5,  -2, -16 },
      {   0,  -1,  -1,   1,   1,   5,   0, -19,   1,   0,   3,  -2,   2,  -2,  -3,  19,   1,   0,  -2,   3,   1,  -1,  20, -20,   0,  -1,   4,  -4,   2,   9, -29, -22,  -1,  -4,  -6,   4,   7, -29,   7, -29,   6,   6,  15,   4, -32,  18,  -2, -10,  -4,  -1, -17, -59,  13,  -9, -23, -18, -19, -33, -35,  36,  17, -15,   4, -16 },
      {   0,   1,   1,  -1,   3,  -5, -10,  15,  -1,  -1,  -2,  -2,  -9,  -1,  11, -39,   1,   2,   3,   1,  12,   7, -15,  29,   2,   1,   1,   0, -10, -17,  22, -28,   2,   3,   0,  -4,   9,  34,  -6, -28,  -2,  -2,  -2,   7,   3, -36, -28, -26,  -8,  -8,  -3, -14, -29,   1, -22, -41,  -3, -10, -19, -17,   4, -28, -35, -19 },
      {   0,   1,   1,   1,   2,  -5,   3,  12,  -1,  -2,  -2,  -1,  -2,   8,  -7, -39,   0,   0,   2,   5,   1, -11,   4,  45,   1,   0,   3,  -2,  -1,  11,   8, -22,  -2,   0,   0,   3,  -3, -12, -21, -45,  -6,  -4,  -7, -12, -15,   9, -27,  -8, -10, -18,  -9,   3,  30, -24, -13,  42,   4,  15,  26,  27, -21,   0,  40,  13 },
      {   0,   0,  -1,   1,   2,  -2,  -2,  -2,   1,   1,   1,   0,   3,  12,  13,   4,  -1,  -1,  -1,   3,   1, -10, -22, -24,   1,   0,   1,  -1,  -4,  14,  27,  48,   5,   3,   6,   5,  19,   8,   6,  -9,  -4,  -4,  -5,  -5, -12,  12,  10, -26,  -2,   1,   6,  14,  45,  33, -39,  -9,   8,  15,  16,  23,   1, -58, -16,   9 },
      {   0,   0,  -1,   0,   2,  -2,  -5,   4,  -1,   0,  -1,  -2,  -6,  -5,   6,   6,  -1,   0,  -1,  -2,   6,   7,  -4,  16,   1,   1,   1,   1,  -3,  -3,  20,  13,   0,   1,  -1,   0,  11,  15, -13,   0,   2,   3,   1,  -1, -19, -39,   6,  17,  -7,  -6, -23, -29, -16,  44,  14,   1,   3,   4,  16,  38,  65,  24,  19,  42 },
      {   0,   0,   1,   2,   3,  -4,  -9,  -4,   0,   0,   0,  -2,  -1,   3,   4, -12,   0,   0,   0,   0,   4,   2,   3,  20,   1,   1,   1,   2,   2,   1,   0, -27,  -3,  -1,  -2,  -4,  -9, -21, -43, -11,  -1,  -4,  -8, -14, -14,  -3,  43,  52,   6,   4,   5,   9,  29,  42,  31,  -4,  -1,  -8,   7,   3,  -1, -22, -30, -47 },
      {   1,   0,   0,   0,   2,   1,   2,   9,  -1,  -2,  -1,  -1,  -4,  -4,  -6, -14,   1,   1,   2,   2,   6,   4,  18,  30,  -2,  -3,  -4,  -1,  -6,  -1, -18, -39,   1,   0,   4,   3,   6,   2,  21,  68,   2,   1,   4,   3,  -2,   9,  32, -55,   1,  -3,  -5,   0,   0,  25, -23,  28,  -1,   1,  18,   5,  21, -11,   7, -25 },
      {  -1,  -1,   0,   0,  -1,  -6,  -4,   7,   0,   1,   1,   1,   0,   3,  -7, -15,   0,   0,   0,  -1,  -1,  -6,   1,  10,   0,   0,  -1,   1,   0,   5,  -4, -24,  -3,  -3,  -4,  -4,  -6, -17, -22,  21,   4,   5,   4,   6,  10,  13,  47,  24,  -3,  -2,  -4,   4,  -6, -13, -45, -60,   1,   4,   2,  -4, -18,  -5,  13,  66 },
    },
  },
  { // 32
    {
      {-103,  17,  24,   3,   4,   1,   2,   0,  57, -32, -14,   4,  -1,   0,   0,   0,   7,  15,  -9,  -5,  -1,   0,   0,   0,  -2,  -1,   5,  -2,  -1,   0,   0,   0,   2,   0,  -1,   2,  -1,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      {  25, -50,  11,   5,   1,   2,   1,   1,  62,  34, -43,  -6,  -4,  -2,  -1,   0, -51,  36,  31, -13,  -1,  -2,   0,  -1,   0, -25,  12,  13,  -3,   1,   0,   0,  -5,   5,  -7,   4,   4,  -1,   0,   0,  -1,  -4,   1,  -1,   2,   1,   0,   0,  -1,   1,  -1,   0,   0,   0,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0 },
      {  26,  88, -34, -19,  -4,  -4,  -2,  -1,  42, -20,  23,  -3,  -5,  -1,  -1,   0, -53,  -5,  17,  13,   1,   1,   1,   0,   6, -14,  -4,   4,   3,  -1,   0,   0,  -3,  -1,  -2,   2,   0,   0,   1,   0,   0,  -2,  -1,   0,   3,   0,   0,   0,  -1,  -1,  -1,   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   0,   1,   0 },
      {  45,  -2,  10, -15,  -2,  -3,   0,  -1,  14, -55,  -2,  26,  -1,   5,  -2,   2,  10,  54, -53, -15,   5,  -1,   2,  -1, -23,  18,  44, -20,  -6,  -2,  -1,   0,   3, -16,   1,  20,  -3,   2,  -1,   0,  -6,   6,  -3,  -2,   5,  -1,   2,  -1,   0,  -3,   2,   0,   0,   1,   0,   1,  -2,   2,  -1,  -1,   0,   0,   0,   0 },
      {   7, -35,  15,  -3,   5,   0,   1,   0,  18, -24,   5,  13,  -6,   3,  -2,   1, -45, -23, -34,  30,  13,   2,   3,   1,  37,  29, -42, -44,  11,   2,   1,   0,  -7,  22,  34, -19, -22,   1,  -1,   0,   2,  -5,   4,  16,  -4,  -6,  -1,   0,  -3,   3,  -1,  -1,   4,  -1,  -2,  -1,   1,   0,   0,   0,  -1,   1,   0,  -1 },
      {   8, -19, -39,  13,   6,   1,   2,   1,  57,  20,  26, -38,  -9,  -2,  -3,  -2,  36, -11,  -1,  31, -14,  -1,  -3,   1, -50,  41,  -7,  -6,  12,  -4,   2,  -1,   3, -23,  24,   4,  -2,   1,  -1,   1,  -6,   2, -13,   6,   6,   1,   0,   1,  -1,  -1,   4,  -5,   1,   3,   1,   0,  -3,   0,  -2,   2,  -1,   0,   1,   0 },
      { -31, -17, -64,  22,  11,   5,   2,   3, -21,  10,  22, -20,  -5,  -1,   1,  -1, -27,  21,  -6, -14,   6,  -4,   1,   0,  47,   5,  36, -25, -24,   2,  -4,  -1,  -3,  -6,  13,  43,  -9,  -9,   0,  -1,   0,  -1, -11,   2,  19,  -3,  -1,   1,   0,  -5,   3,  -1,   0,   5,  -2,   1,   1,   1,  -5,   0,   2,   1,   0,  -1 },
      {  -6, -37, -28,  21,  13,   1,   4,   0,  -7, -46,  46, -14,  -2,   6,  -2,   2, -12,  34,  -7,  23, -13,  -2,   0,  -1,  -5, -36, -20,  23,  23,  -3,   1,   1,   6,  12, -38, -29,  15,  11,  -1,   1,   0,  12,  12, -22, -17,   6,   3,  -2,   4,   1,   3,   6, -10,  -6,   3,   0,  -1,   3,   1,  -1,   1,  -5,  -1,   1 },
      { -14,  17, -12,  12,   0,  -1,  -1,   0, -22,  33, -19,  -1,   3,   3,   1,   0, -24,  19, -17,   9,  12,  -5,   4,  -1, -41, -32,  13, -40,  30,   6,   3,   2,  47, -19,  10, -30, -38,  14,  -2,   2,   1,  14,   8,  20, -19, -16,   2,  -2,   5,  -3,   3,   3,  10,  -5,  -4,   1,  -1,   0,   2,   3,  -1,   2,  -2,  -1 },
      {  10,  12, -44,  28, -11,   6,  -2,   1,  -5, -37, -49, -15,  25,  -1,   6,   1,  21,  28,  23, -33, -20,   3,  -4,  -1,  11,  21, -21,   8,   1,  -7,   2,  -1,  -7,  15,  33, -33,  -4,   0,  -2,   0,   2, -11,  19,  27, -15,  -4,  -3,   1,  -4,   1,  -8,   4,  13,  -3,  -1,  -1,   0,  -2,   4,  -1,  -1,   4,  -1,   0 },
      {  -6,  -9,  -6,  24, -10,   3,  -1,   1,  -6, -17, -15,   2,  29,  -4,   4,  -1, -49, -45,  -5, -23,   0,  21,   4,   5, -19,  36,  10,   6, -16, -16,   3,  -3,  18, -42, -10,  -6,  29,   8,  -5,   3,  10,  18, -18, -23, -30,  13,   6,   0,  -1,   5,  15,   5, -10, -19,   3,   1,   4,   0,   0,   2,   2,  -2,  -8,   3 },
      {   3, -10,  33,   7,  -4,  -2,  -1,  -2, -22, -47,   0, -26,  11,   2,   4,   3, -21,  -4,  43,  14, -23,  -1,  -4,   0, -15,  -2,  -7,   9,   1,   2,   1,   0,  44,  -3,   6,  29, -21, -10,  -3,  -1, -19,   8, -28,  25,  40,  -8,  -1,  -1,   6,  -4,  -2, -29,   6,  16,  -2,   2,  -2,   2,  -1,   6, -10,   1,   2,  -1 },
      {   5, -16, -36,  -6,   3,   8,   0,   3,  20, -17, -11,  17,  -3,  -4,   4,  -2,  15, -44,   2, -20,  28,   5,  -2,   3, -25, -13,  -4, -12, -36,  22,   2,   1,  40,  46, -26,   6, -17, -26,   7,  -2, -16,  15,  29,  -8,  11,  -7,  -7,   1,   4,  -7,   3,  12,  -3,   2,  -2,   0,  -6,   3,   0,   0,   5,  -2,  -1,   0 },
      {  12,  -8,  25,  -2,  -7,  -2,   1,  -1,  23,  -3,  39, -36,  19,  -1,  -1,  -1,  12, -25,  -8, -30, -21,  12,   5,   4,  24, -21,  25,   6, -14,  -8,  -7,   3,  -5,  -2, -10,  -1, -28,  20,  -2,   2,  -1,  21,   6,  33, -33, -33,   7,  -4,  -3,   3,  -2,  22,  37, -13, -14,   0,  -2,   2,  -4,  -4,   4,  15,  -1,  -2 },
      { -14, -23, -22, -83,  32,   6,   7,   1, -13, -16,  -3,  18, -32,   1,   2,   2,  -2,   4,  46,  -1,  -2, -10,  -5,  -1,  -1,  20,   4,   7,   4,   3,  -4,  -1,   7, -23,   6,  -3,  -9,  13,   2,   1,   6,   9,  -7,   9, -27,  -7,   5,   1,  -4,   4,  -1,  12,  12, -18,  -4,  -1,   2,  -3,   0,  -2,   4,   5,  -6,   0 },
      {  13,   7,  -7,  12,  16, -10,   0,  -2,   7, -20, -55, -25, -10,  26,  -2,   5,  24, -29,   6,  22,  -4, -24,  10,  -4,  32,  -5,  18, -20,  31,   6, -13,   6, -17,  -4, -30,  11, -18,  26,   2,  -1,   2,  24, -10, -28,   0,  -9,  12,  -1,  -2,  -1,  23,  -4, -18,  -4,  -5,   4,  -1,   3,  -2,   9,   1,  -6,  -1,  -2 },
      {  -6,   5,  -2, -26,   3,  -8,   6,  -2,  -5,  19, -17, -19,  -7,  14,  -8,   4, -14,  -4, -28,  15, -41,  -3,  10,   0,  16,  32,  24,  19,   8, -37,  -5,  -3,  42,  29,  -7,  -3,  28,   6, -12,  -3, -35,  -3,  40,  -1,   6,   6,   2,  -3,   7, -21,  -8,   8,   6,  12,   0,   2,  -5,   3,  -5,  -7,  -3,   3,   6,   1 },
      {  -8,  -7, -17,  13, -21,   3,   1,   3,  -2,   6,  -5,  45,   0,  -8,  -4,  -2, -12,  -2,  -5,  11,  -8,  -5,   1,   0,  -5,  23,  -4,  27,  16,  -4,  -6,  -2, -22, -15, -29,  -4,  -7,  -2,  20,  -2,  17,  24,   9,  -4,  24, -51, -12,   4,   0, -10,  11,  -9,  42,  33, -26,  -2,  -1,  -1,  -3,  -8, -15,  21,  14,  -6 },
      {   4,  -1,  -7,  -3, -46,  22,  -5,   3,  19,  12,   9,  32,  16, -29,   3,  -4,  28,  18,  13,   9,  16,   4,  -7,  -3,  50,   0, -15,  -3,  14,  -4,   3,  -3,  41, -12,  -1,   1,   6,  18, -11,   5, -40,  29, -19,  -4,  -7,  11,  12,  -3,  -2, -12,  15, -10,  -2,  -9,   3,   4,  -5,   3,  -3,   8,  -2,   0,  -2,   0 },
      {   3,  16,  25,  -4,  -8,  10,  -6,   1, -10,   6,  -8, -44, -12, -11,  13,   0,  -3,  37,   4,  29,   3,   6, -16,   2,   6,  28, -20, -10, -33,  13,  18,  -5,   1,  -7, -14,  -2,  -4, -39,   6,   6,  11,  29,  20, -21,  -5, -12,  -9,   1,   1, -11,  32,  28,   3, -18, -13,   0,   3,  -2,  -9,   8,  24,  13,  -9,  -5 },
      {   5,  13,  18,  54,  25,  -7,  -6,   0,   0,   3,  15,  17, -31,  12,   1,  -1,  -6,   6,  22,  -5,  22, -18,  -1,   1,   3,  40,   5,  11,   2,  27,  -7,  -2,  10,  -3, -17,   2,  -3,  13,  19,  -3, -40, -25,  10,   5, -15, -12,  20,   6,  11, -11, -26,   4,  10, -28,  -4,  10,  -4,   5,   1, -15,   9,   4, -17,   0 },
      {   2,  -7,  14,   1,  -9,  -1,  -1,   3,   4, -12,  14, -14,  15, -10,   3,  -4,   2,  -7,  27,   6,  28, -10, -13,  -2,  16,  17,  39,  -8,  18,  24,  -5,  -4,   8, -27,  -5, -25,  -2,  -1,   8,   7,  13,  -2,  35,   5,  -1,  22, -32,   3, -16, -11, -22,  27, -16,  46,  18, -14,  -2,   4, -13,  -7, -13, -15,  25,   3 },
      {   5,   9,  12,   2,  54, -21,   3,  -3,   8,  17,  21,  11,  14,  24,  -6,  -4,   2,  15,  11, -56, -13,  10,   9,  -1,   8,  11, -42, -13,   4, -10,  -6,   7,   8, -16,  10,  -3, -14,   3,   0,  -3,  -8,  29,  12, -25,  11,  -1,  -9,  -1,   9, -19,  19,   3, -19,  23,  -2,  -4,  -6,  -2,  -7,  16, -11,  -6,  13,  -4 },
      { -11,  -5,  -8, -37, -18,  -4,   7,  -2,  -6,   9,   4, -16,  59,  -4,  -4,  -1, -11,  12,   3,   2,  -6,  25,  -1,   1, -13,  23,   2, -13,   4,  30,   2,  -1, -32,  11, -27,  -6, -24,  16,  20,   0, -30, -27,   4, -14,  -1, -19,  17,   4,  18, -14,  -6, -18, -23,  -7,  -3,   9,   0,   7,   6,   6, -11, -20,  -9,  -1 },
      {  -3,   1,  -7,   5,  -4, -11,   0,  -1,   1,   8,  -5,  20,   4,  28,  -4,  -1,  -1,   3,   0,  14, -23,  12,   6,   1,   0,   4, -11,  -2, -11,   9,  13,   1, -19, -11, -24,  13, -15, -12, -12,   4,   8,  42,  11,  36,   5,  18,   6, -23, -23, -25, -21,  -4,  10, -14,  55,  10,   7,  -3,  -4, -23,   1, -30, -19,  25 },
      {   2,  -4,  -2,  10,  18, -12,   9,  -7,   1,  -9,  -7,  25,  24,   1, -12,   4,  13,  -2,  34,  37,  -7,  32,  16,  -4,  -1, -22,   1, -48, -14, -27,  23,  10, -11, -24,  -8,  14,  21,  -8, -22,   6, -16, -26,  14,  -4,  -6,  -2,  -3,   2,   8,  -7, -16,  20,  14,   1, -24,  -1,  -4,   8,  -8, -13,   5,  24,   9,  -9 },
      {  -7,  -3,   7,   9, -47, -14,   0,  -1,  -5,   1,  22,  15,   0,  13, -22,   1,  -5,  11,  38,  -1,  -4, -11,  22,  -8, -25,  19,   6, -19,   5, -14, -21,  11,   2,  44,  13,  23,  -5,  24,  -8, -18,  33,   1,  12, -10,  -5,   1,  20,   1, -23,  -5,  17,  27, -12,  -6,  -3,   4,   2, -11, -15,  10,  15,  -4,   5,  -2 },
      {   0,   3,   7,   2,  41,  16,  -4,  -4,  -5,  -5,   1,  -8,  15, -46,  22,   3,  -6,  -4,   5,   1,  29,   2, -16,  -4, -17,   1,  17,   2,  17,  -7,  -4,   1, -26,  29,  14,   0,  11,  16, -26,  -4,   8,  22,   8,   2,  11,  16,  26, -12,  -5, -51,  12, -10,  17,  -2,   2,  22,  -2,   2,  -8,  -6, -16,  21, -12,   1 },
      {   4,   2,  -2,  -3,  17, -46,  14, -10,   9,   8,   9,  -6,  17,   4, -19,   2,  12,  13,   8,   2,  28,   5,   8,  -1,  18,  22,  14,  -4,  -8,   1,  19,   2,  32,  15, -30, -29,  15,  -9, -10,   5,  36, -10,  -8,  16,  10,  -4,   5,  -2, -26,  24,  16, -39,  10,  -3,  -7,  14,  16,  -9,  15,   6, -16,  12, -18,   1 },
      {  -6,  -6,  -3, -10, -33, -22,   5,  -7,  -8,  -9,  -4, -31, -24,  -7, -25,  -1,  -7,  -1,   0, -27,  40, -23,   7,   6, -13,   0, -29, -11,   2,  -5,   6,  10, -24, -24, -13,  14,  12,  14, -29,  -5, -32,  10,  35,  11,   7,  10,   4, -16,  24,  24,   6, -10,  14,  10,   6,   4,  16,   0,   0,  -4, -10,  13,   4,   5 },
      {   3,  13,   6,  12,  26,  37,  -5,   0,  -3,   9,   2,  12,   0, -28,  -2,   0,  -4,   7,   8,  -1, -32,   5, -21,   8,  -6,  24,   4, -28,   7,   8,   6,  -7,  -3,  29, -36,  17,  -1,  18, -13,   3,  -4,   4,   1,  12, -15,   9,  -8, -16,  10,  59,  13,  -2,   7,  12,  19, -20,   1, -16,  10,   7,  -5,   1,  27,   9 },
      {  -2,   5,  -5,  -1,  13, -23,   2,  -1,  -8,   2, -19, -17,   8,  11, -13,   1,   0,  16, -11,  22,  42,  39,   0,  -8,  -5,   3, -15,  26, -22,   6, -20,   4,   3,  16,   5,  11,  -5,  25,  10, -11, -19,  11, -37,  -4, -19,   3,   3,   4, -19,   2,  -2,  17,  20,  15,  10, -28,  -4, -11,  -5, -25,  -5,   5,  44,  10 },
      {   6,   3,   1,  -7,  11,  24,   8,  -4,   4,   1, -11, -15,  13, -25,  -1,   3,   4,   5, -18,  -6,  -7,  -5,   3,   8,   6,  -6, -42,  -7,   4,  16, -12,   1,  33, -23, -12,  43,  15,  22,  16, -19,  38, -14,  25,  -5,   3,  -7,  26,  15, -32,   3, -32,   0,  -1,   1, -11,  13,  16,   4,  -5, -25,  -8,  -6,   1,  -7 },
      {   1,   3,   3,   6, -12,  40,  -7,   6,  -3,  -2,   0, -11, -58,   4,  -8,  13,   1,   2,   4, -14,   4,  47,  12, -12,   0,   9,   3,  -4, -12, -20,  21,  14,  -4,  -1, -22, -23, -23,   6, -20,  -1,   6,  -7, -10,   3,  -1,  -2,  16,   0, -19, -19, -15, -13, -22,  10, -15,   2,   8,  19,   3,   3, -29, -15,  17, -24 },
      {   0,  -1,  -2,  11,   7,  -8,  23, -25,  -3,  -5,  -4,  13, -12, -10, -42,   7,  -2, -11,  -4,   3, -10,  28, -25,   8,   2, -12,  16,  29,  12,  26,  48, -12,  -2,  -1,  34,  13,  -5,  11,  14,   9, -14,   6,  26,   1,  -1,  -5,  10,   6, -15,   9,  23,   2, -14,  -2,  -3,   6,  32,  -7, -33,  16,   5,  -6,   1,  -5 },
      {   0,  -1,  -4,   3,  -7, -36,   2,   5,   1,   3,  -3,   2,   0,  -9,   6,  -9,   0,   2,  -1,   3,   9,  21, -43,  -6,   6,  11,   3,  15,  20, -24,  14,   7,   5,   7,   4,  21, -33,   3,  -9,   4,  24,  29,  11, -23,   4,  13,   3,  -3,  26,  22, -58,   1,  -8, -11, -11,   3, -19,   0,  27,   2,  12,   5,  -4, -24 },
      {   2,   0,   3,   2,  -4, -14,  21,  -5,  -1,  -7,  -2, -11, -23, -43,   4,   4,   4,  -2,   3, -22,   6,  25,  16,  -9,   4,   1,  -5,  -9,  39, -32,  11,  -8,   8,   3,  -8,   8, -14, -12,  13,  15,   2, -28, -12, -23,   7, -23, -14,  -1,   1,  -7,  10,  17,  17,  -3,  24,  -2, -12, -14,  -5, -20,   8, -27, -18,  50 },
      {   2,   2,  10,   2,  -7,  21,  15,  10,   0,  -4,   6,  -9, -11,  23,  -5, -22,   5,   1,  19,   7,   2,   6,  -8,   1,   0, -11,  15, -29, -20, -13,   8,  -8,  -2,   9,  30,  -9,  28,   9,  42,  -5,   3,  33,  17, -28,  -9, -14,   8,  21,  18,   6, -15, -32,  15,  -1,  10,   7,   1, -24,  15, -12, -38,  -7,  -3,  25 },
      {  -5,  -1,   1,  -2,  -4, -12,  -2,  -1,  -4,   2,   5,  -5,   0,  -3,  20,   0,  -7,  -2,   1, -16,  -9, -13,  -5,   8, -12,   0,   1, -20,  25, -11,   2,   0, -14,  14,  -5,   5,   3, -17,   8,  15, -44,  19,  -8, -15,  -7,   8, -25,  19, -73,  -1, -18,  -9,   1, -10, -19, -18,  39, -13,  18, -12,   1,  11,  -5, -18 },
      {  -1,   0,  -2,   4, -13, -51,  22,  -4,  -2,  -1,  -6,   1, -34, -35,  -6,  12,  -1,  -2,  -1, -10, -32,   9, -28,  -1,  -2,  -7,  -6, -32, -11,  27, -25, -17,  -4,   6,  -9,  -8,  37,   7,  12,  -4,   6,  15, -20,  26, -12,   2,  -2,  11,   9, -27,  -1,  -2, -11,   6,  -8,  -4, -22,  22,   4,   2,  -6,  -3,   4,  -8 },
      {   2,   7,   5,   9,  17,  16,  55, -11,  -3,   1,  -2,   5,  12, -42, -30,  12,  -3,   4,   2,  16,   4, -32,  19,  -7,  -2,   4,   6,  10, -32, -25, -22,  21, -11,  -4,   5, -23, -17,  -7,  -5, -12, -16,  28,  -2,   5,   8,  -6,  -2,  14,  -4,  17, -16,  -1, -17,  -2,  -9,  -4,  -2,   6,   6,  -2,  17, -34,  -8,   0 },
      {  -1,   4,   1,  -5,  -3,  27,  15, -17,  -2,   3,   1, -11,  11,  16, -30,  13,  -5,   0,  -3, -18,   7, -20,  29, -22,  -3,   9,  -5,   0,  25,  31,  45,   2,  -2,  19,   3,  13,  22, -10,  -3,  33,   8,  22, -24,   3, -11,   3, -10,   7,   2,  -3, -16,   5,   3,   0, -11,   4, -54,   5,  -2, -20,   1,   9,   3, -11 },
      {  -2,  -1,  -1,   0,  -5,   2,  30,  11,  -3,  -2,   0,  -1,  -4,  15,  22,   2,  -5,  -5,  -6,  -3,   0,  -2, -34, -10,  -2,  -6,  -5,  -3, -10,  -4,   4,   2,  -7, -20, -10, -10,   2,   7, -10, -10, -18, -17,  10,   7,  14,  -4,  21,  -3, -41,   8,   8,   2,  -4,   7,  -1,  17, -66, -63,  -8,  14,   5,   1,   1,  -7 },
      {  -4,  -3,  -2,   0,   6, -22, -54, -29,  -4,  -2,   0,   3,  -8, -26,  28,   0,  -1,   4,   5,  14,   0,  17,  51,  24,   1,  -1,   7,  11,  -4,   5,   5,   7,  -6,  -4,  16,   8,   6,   4,   8,   1,  -9,  18,  28,  11,  -3,  -9,   3,  15, -17,  27,  11,  -8, -14,   1,  -5,   6, -44,   4,  19,   5, -11, -22,   0,  -2 },
      {   0,   1,  -1,   5, -11,  -7,  20, -50,   2,   5,   2,   7,  -1,  14,  32,  32,   1,   4,   3,   7,  -3, -29, -20,  31,  -3,   1,  -7,   3, -30,  -6,   4, -15,   5,   7,   8,   5, -20,  14, -23,  25,   4, -13,  -6, -34, -24,  -9,  -6, -21,  -8,   3,  -4,   1,   8,  15,   1,   7,  -5,  25, -12,  -8, -39,  -7,  -5,   8 },
      {   1,  -2,   2,   4,  -8,   4,  10, -18,   1,  -3,   3,   1,  -7, -11, -23,  19,   1,  -5,   0,   1, -25,  14,  16, -12,   7,  -3,  10,   5,   4,  46, -24,  12,   6, -15,  19,  -5, -17, -24, -19, -24,  -2,  -9,  12, -39,  -6,  15,   3, -26,  -6, -11,   5,  -3,  18,  -6,   9,  -5,  -2, -21,  58,  -3, -14,  28,   2,   3 },
      {   1,  -1,   1,   6,   3,   0,  22,  13,   1,  -2,   0,   1,   2,   1,  11,   3,   0,  -2,  -4,  11,  16, -11,   6, -21,  -2,   0,  -8,  16,  -5, -10,  11,   3,   1,  -2,  -3,  42,   3,  11,  -4,  18,  11, -12,  26,  21, -39, -13, -23,   4,   0, -40,   7, -26,  -8,  -6,  -1, -37,   1,  -8,  36,  51,  -5,  -9,   8,  11 },
      {  -2,  -2,   0,  -3,   6,  14, -44, -13,  -2,   0,   2,  -4,   8,   8, -52,  41,  -1,  -2,   3,  -2,  10,  -2, -33, -17,  -3,  -5,  -5,  -9,  -8, -30, -11, -37,  -3,  -9,  -4,  -2,  -5,  10,  25,   1,  -3,  -5,  11,  -4,  13,   5,  -2,   4, -26,   0,   2,  -5,  14,   0,  12,  -9, -25,  31,  21,  17,  16,   7, -20,  11 },
      {  -1,  -3,  -1,  -2,  -7,   9,  20, -63,   4,   3,   7,   1,   3,  18,  47,  24,   1,  -3,   4,  -8,   9,   5,   1, -23,   2,   3,   3,  -6,  10,  -7,  12, -14,  -2,  -4,  -6, -10,  11,   0,  17, -18,   4,  10,   3,  18,  28,   2,  28,  12,  16,   1,   1,   9,  -5, -11,   5, -13,  21,   2,  26,  -1,  30,   6,  39,   9 },
      {   0,  -4,  -3,  -2,   1, -17, -51, -21,   2,  -2,  -1,   0,   0, -27, -13,  -2,   0,  -5,  -6,  -2, -20, -39,  -9, -22,   2,  -2,  -1,  -5, -32,   1,  36,  25,   6,  -3,  -4, -19,  -6,  28,  -2,   3,   7,  -8, -12, -14,   9,  -9,  -1,   0,  11, -28, -16,  -7,  -5,  -8,   0, -19,  11, -47, -10, -13,   5,  -5,  18,   7 },
      {  -1,   0,   3,   0,   2, -19,  11,  49,  -1,  -1,   3,   2,   4, -11,   9, -21,  -4,  -3,   2,  -3, -14, -28,   0, -15,  -1,  -2,   3,   1, -11,   3,  44,  -5,   0,  -2,  11,  -6,  -9,   5,   1,  -4,  -5,  -7,   4, -16,   7,  -2,  31,  -9, -26,   7,  21,   3,   8,  -5,  10,   4,  -9,  58,  28, -16,   5,  -2,  41,  11 },
      {   0,   0,   0,   0,   2, -13, -19,  25,   2,   3,   3,   4,   8,   0,  16,  56,  -1,  -2,  -1,   2,   0,  -3, -11, -71,  -1,  -1,   0,   2,   5,   1, -11,  10,   1,   0,   9,  11,  10, -28,   5,   9,  -9,   7,   8,   5, -19,  -9,  15,  -8,  11,  26,  10,   7,   0,   6, -12,  18,  20,  -8, -12, -14, -30, -29,  -2,  -6 },
      {   0,  -1,  -3,  -1,   0,  -8,  -2, -22,   0,  -2,  -3,  -1,   0, -14,  -6, -18,  -1,  -1,  -4,  -2,  -4, -17,  12,  -5,   0,   1,  -2,  -3,   6, -16,   1,  -4,  -3,  -1, -15,   4, -22, -46,  33, -17,  -1,  -7,  -6,  -6, -36,  40,  25,   9,   0,  -4, -14, -18,  31,  -2,   7,  18,  -9,   1, -37,  43, -14,   3,  32,  -5 },
      {   0,   0,  -1,   4,  -1,  -2,  11,   0,  -2,  -1,  -4,   3,  -3,  -9,   4, -13,  -1,   1,  -5,   7,  -8,  -4,  21, -21,  -3,  -1,  -9,   4, -20,  -4,   6, -10,  -1,  -1,  -9,   2, -18,  34,  29,  37,  -3,  -4,  -8,  10,  22,  41,  -7,  22,   6,   2,  10,  51,   7,   8,   7,  32,  14, -13,  30,  15, -34,  18, -22, -16 },
      {   1,   0,   1,   1,  -6,  -1,  12,   7,   2,   2,   4,   4,   0,   2,   4,  20,   1,  -1,   1,  -1,  -6,   1,  -5,   7,   2,   0,   2,   5,  -1, -13,  -3,  12,   1,  -4,  -6,   5, -32,   4,  16,  -3,  10, -10,  22,  10,  -1,  44, -15,  31,  12,  -2,  42, -33, -10, -31,  -6, -30, -20,   0, -19, -60, -14,  17,   0, -10 },
      {   1,  -1,   2,   1,   3,  -2,   3,  -4,   0,  -4,   0,  -3,   1,  -6,  -9, -13,   3,  -1,   6,   1,   9,  -5,  14,   5,  -1,  -4,   0,  -3,  -3, -15,  -8, -52,   2,  -1,   3,   2,   2,  -5,  12,  17,   3,  -6,   3,  -2,  -7, -11,   9, -41,   9, -11,  15, -12,   5,   1,  20,  -5,   3, -23,  13, -28,  14, -30,  12, -80 },
      {   1,   0,   0,   0,   1,  -8,   0,  46,   1,   1,   1,   1,   2,  -7,   0,  55,   2,   1,   1,   0,   1,  -3,  33,  35,   1,   0,   0,   2,  -4,   7,   4, -61,  -3,  -2,  -1,  -8, -11,   2, -16,  12,   4,  11,   1,  -6,  11,  10,  13,  14,  11,  -3, -25,  -3,  -5, -11, -15,   2,   6, -30,  -9,  13,   4,  12,  26,  23 },
      {   0,   1,   0,  -1,  -2,  -1,  12,   1,  -1,   0,   0,  -2, -10,  -9,  22,  -3,  -1,   1,   0,  -1,  -8,  17,  27, -40,  -3,   1,   0,  -4, -16,  12, -23,   5,   2,  -1,   7, -11,   0,  27,  17,  41,   1,   4,  17, -17,  13,   2, -12, -35, -12,  18, -20, -40,  -1,  -5,  13,  -5,  -2,  15, -40,  12,  30,  35,  -1,   1 },
      {   0,   0,   0,   1,   0,   3,  -8, -20,  -1,  -1,  -1,   2,  -1,   4, -13, -54,  -1,  -1,  -2,   1,  -4,   0,  -8, -23,  -1,  -1,  -2,   4,  -1,   3, -30, -36,  -2,  -5,  -2,   0, -10, -20, -27,  45,   2,   0,   9,   7,   2,   9,  37,   4,  -4,   3,   4,   0, -18,   1, -48, -22,  -1,  -4,   3, -18,  -9,  -7,   1,  38 },
      {   0,   0,  -1,   0,   1,   4,   0,   2,  -1,  -1,   0,  -1,   3,   2, -16,  13,   0,   0,   0,   0,   6,   0, -15,  38,   0,  -1,   0,  -2,   7, -11, -18,  44,   2,  -2,   1,  -2,  13,  -8,  38,  43,   3,  -2,   1,  13,  22,   4,   2, -50,  -3,  -6,   1,  14,  -4, -42, -21,  13,  -4,  -3,  12,   7, -15, -10,  41,  12 },
      {   0,   0,   0,   0,   0,  -3,   6,   3,   0,   0,  -1,  -2,   3,   2,   7,  -3,   0,   0,  -1,  -1,   1,   0,   8,  -6,   0,   0,   2,  -3,   6,  14,   8,  -7,   0,   0,   2,  -7,  13,   6,  -7, -17,   1,   2,   2,  -8,  38,   5,   9,  -9,  -2,  10,  -9,  22,  35, -39,  -6, -72,  -9,   9, -23,  28, -42, -28, -22, -20 },
      {   0,   0,   0,  -1,  -2,   1,   0,   8,   0,   0,   2,  -1,  -4,   8,   6,  15,   0,   0,   1,  -3,  -4,   7,  -5,  19,   0,   0,   1,  -5,  -3,  16,  -4,  35,   2,  -1,   2,  -5,   3,   8,  -7,  40,   5,  -4,   4, -14,   0,  30,  30,  30,   8,  -5,  10, -14,  56,  36, -11, -10,   6,   0,  10,   7,  45, -42,  -9,  -8 },
      {   0,   0,   0,   0,  -1,  -3,   2,   8,   0,   0,  -1,   1,   1,  -8,  -1,  13,   1,   0,   0,   0,  -1,  -9,   4,  21,   1,   0,   0,   0,   2, -12,   6,  23,   1,   0,   0,   0,   0, -23,  23,  22,   0,   3,  -5,   7, -10, -17,  63,  -4,   2,   1,  -6,  13, -32,  34,  35, -47,  -1,  -3,   1,  -1,  -9,  45, -29,  -6 },
      {   0,   0,   0,   0,  -1,   0,   3,   5,   0,   1,   0,   0,  -2,   1,  10,   4,   1,   1,   1,   1,   0,  -1,  12,   3,   0,   0,   0,  -1,  -4,  -3,  28,  -9,  -1,   3,  -1,   2,  -4,  12,  39, -37,  -2,   5,  -6,   8, -13,  36,  -1, -62,   0,  -3,   0,  -1,  -3,  36, -58, -12,   0,  -5,   5,  -9,  14,  -7, -32,  22 },
    },
    {
      {-104,  50,   6,   2,   2,   1,   1,   0,   2, -28,  18,  -2,   2,   0,   1,   0,  30, -23,  -8,   6,  -1,   1,  -1,   1,   6,   5, -11,   0,   1,   0,   0,   0,   4,   1,  -1,  -4,   1,   0,   0,   0,   2,   1,   0,  -1,  -1,   1,   0,   0,   2,   0,   0,   0,  -1,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      {   9,  36, -23,   0,  -4,   0,  -1,   0, -68,  56,  13, -16,   4,  -3,   1,  -1,   7, -42,  48,  -9,  -2,   1,  -1,   0,  21, -26,  -2,  22, -11,   3,  -1,   1,   0,   1, -13,   9,   3,  -4,   2,   0,   3,  -2,  -4,  -1,   4,  -1,  -1,   0,   1,  -1,  -1,   0,   2,   1,  -1,   0,   2,  -1,  -1,   0,   1,   0,   0,   0 },
      { -38,  18,  -9,   4,  -1,   1,   0,   1,  34,  34, -43,  -1,  -5,   0,  -1,   0, -29,  47,  22, -38,   9,  -5,   0,  -1,  10, -29,  52,  -6, -15,   4,  -2,   1,   3,  -8,   1,  28, -11,  -3,   0,   0,   0,  -4,  -3,   8,  10,  -5,   1,  -1,   0,  -3,  -2,   1,   5,   1,  -3,   1,   0,  -2,   0,   0,   1,   2,   0,   0 },
      {  -7,   1, -10,   5,   1,   3,   1,   0, -69,  10, -15,  11,   7,   3,   2,   1,  -1,  14, -47,  -8,  16,  -4,   5,  -2,   8,   9,  17, -58,  14,   1,   0,   0,  20, -19,  33, -14, -28,  15,  -4,   2,   3,   4,   0,  17, -20,  -5,   5,  -1,   4,  -1,   3,   4,   1, -10,   2,   1,   2,  -1,   1,   2,   1,  -2,  -3,   1 },
      { -42, -75,  60,   4,   6,   2,   3,   1, -31,  15, -16,  16,  -3,   2,  -1,   1, -18,  -2,   7,   5,   4,  -3,   2,  -1,  26, -32, -17,  11,   7,   3,  -2,   1,   4,  13, -23,  -6,   5,   3,   0,  -1,   3,   3,  -1,  -9,   0,   0,   0,   0,   1,   5,  -1,  -3,  -3,   1,   1,  -1,   1,   1,   1,  -2,  -1,   0,   1,  -1 },
      { -14,  13, -24,  18,   0,   3,  -1,   1,  23,  38, -28,  -1,   5,  -4,   1,  -2, -36,   0,  19,  15,   5,   1,   0,   0, -16,   1, -38,   1,  33,  -4,   5,  -2,   2,  -8,  17, -60,  22,  13,  -3,   2,   9, -15,  15,  -9, -34,  23,   0,   1,  -2,   3,  -2,   6, -17,  -8,  11,  -2,   2,  -3,   3,   1,  -2,  -7,   0,   2 },
      {  -3, -30, -31,  30,  -3,   6,  -1,   1,   5,  -2, -53,   7,   7,   1,   1,  -1,  53, -23,   3, -45,  15,   6,  -4,   3,  12,  29, -15,  -3, -30,  11,   3,   1, -17,  34,   5, -10,  -3, -17,   9,  -1,   2,  -6,  23,  -7,  -2,  -4,  -3,   2,  -4,   3,   3,   7,  -4,  -2,  -7,   2,   0,   0,   1,   2,   2,  -1,  -3,  -2 },
      {  -7, -31,   3,  -8,   9,   5,   2,   1,  35,  11,  10, -48,  12,  -2,   4,   0,   2, -36,  26,   2, -15,   9,   3,   0,  -4,  -6, -13, -18,  -8,  10,  -1,   2,   8,  -9,  20,   2, -48,  15,   1,   0,  -6,  13, -20,  43, -26, -24,  14,   0,  -1,   7,  -1,   2,  18, -28,   1,   3,   1,   1,   1,   0,   8,   1, -10,   4 },
      {  -6, -15,  24, -22,   9,   1,   1,  -1, -25,  -3,  -5, -22,  19,   0,   4,   0,  43,  30,  16, -28, -25,  10,  -3,   0, -43,  15,  28,  36,  -9,  -3,   2,  -3,   5, -38,   5, -11,  25,   8,   0,   2,   4, -12,   0,   3, -30,  19,  -1,   4,  -1,   3, -13,  16, -12, -10,  11,  -4,  -1,   2,   0,  -2,   3,  -7,   0,   1 },
      { -22, -26,   0,  12,  -1,  -1,   1,   0,  -1,  54,   9, -21,  -7,  -9,   0,  -2,   8,   6,  -5,  27, -17,  -6,   0,   0, -19,  45,  -7,  -5,  -1, -18,   2,  -3,   7, -12,  40,  17,  11, -27,  -1,  -3,   8, -14,   0,   7,  45,  -6, -15,   5,   4,  -7,   5, -10,   9,  33, -15,   1,  -2,   0,  -7,   4,  -8,  15,   9,  -6 },
      {  15,   8,  21, -16,   1,  -1,  -4,   0,  32,  45, -18,  12, -12,  -3,  -2,  -2,  74, -12, -15,  22,  17, -20,   4,  -6,  17, -17,  15, -11,  33,   0, -11,   3,  -7,  -4,  -2,   5,   7,  26, -13,   0, -16,  10, -16,  -1,   0,  21,   6,  -4,  -6,   3,  -2,  -8,  -6,   4,  12,  -2,  -2,  -1,   2,  -3,  -3,  -3,   6,   1 },
      {  -1,  -8, -27,  58, -15,   3,  -2,   1, -16, -30, -23,  11,  14,  -7,   5,   1,  24,  19,  28,  15, -11, -14,   0,  -3, -30, -20,  -3,  15,  38,  -8,  -1,   0,  14, -16, -18,  -1,  -6,  13,  -5,   1, -10,  16, -23,  11,  16, -18,   9,  -1,   5,  -2,   6, -30,  23,   1,  -9,   6,  -2,   2,   5,  -4,  -8,  15,  -3,  -1 },
      {   9,  36,  41, -24, -15,  -1,  -1,  -1,  -6,  21, -24,  17, -22,   3,   0,   2,  13,  27,  -6,   9, -20,  -2,   3,   1, -11,   5, -22,  20,  -2,  -8,   9,  -1,  -9,  21,  -4, -24, -16, -12,  13,   0,   2,  -4,  25,  14,  -7, -45,   7,   2,   6,  -8,  10,   2,  36, -24, -23,  11,   3,  -4,   5,   1,   8,  15, -21,   0 },
      {  13,  22,  26, -29,  27,  -5,   2,  -5,  -4, -16, -35, -30,  36,   4,   2,   1,   3,  -1,  13,  -9,  -1,   1,   1,  -3, -19, -19, -21, -23,  12,  20,  -3,   1,  13,   6,  -5, -14, -28,  -3,   4,  -3,   6,   1,  15,  -6,  24,  -7, -15,   6,   4,  -1,   9, -10, -21,  48,  -6,  -6,  -7,   4,   1,  13, -25,   1,  29, -10 },
      {   5,  11,  40,  14,   1,  -5,  -5,   0,   1,  -1, -11,  48,  -5,  -6,   0,  -1, -15, -52,  18,  -5,  42,  -4,  -1,   1, -61,   9,  11, -11,  -9,  19,  -4,   4,  -8, -22,  10,  16,  -2,  -4,   4,  -3,  14, -22, -12,   4,   6,   1,   1,   3,   1,   4, -13,   0,   3,  -5,  -5,   5,   1,   2,  -4,  -5,   4,   5,  -9,   1 },
      {  12,   6,  25,  19, -25,  -7,  -4,  -2,   1, -28,  17,  12, -20,  -2,  -6,  -1,  11,  11,  46, -10,   0, -15, -12,   4,  38,  -5,   4,  -4,  13, -10,  -4,   2,  12,   9,  41, -17,  -9,  -7, -16,   4,  10, -16,  18,  44,  -1,  18, -23,  -8, -12,  10, -23,  21,   7,  20,   1, -11,  -2,   1,  -7,   7,   5,   2,  12,  -8 },
      {  -6,  -5,   7,  30, -22,   3,   0,   1,  -2,  22,  55, -11,  -4,  -3,  -4,  -3,  16,  41,  13,   0,  26,  -6,  -9,   2, -27,   5,  -2, -10,   7,  38,  -9,  -1, -22,  24, -15,  -2, -34,  16,  21,  -8,  -3, -10,  20, -14,  -1,  -8,  16,  -3,   6, -14,   3,   8, -23,   2,  11,   3,  -1,   0,  -7,   8,  -2, -21,   7,   5 },
      {  -2, -14,  -6,  -7,  -2,  -2,   4,  -1,  15,   6,  32,   4,  24, -23,  -3,   0,  -1, -20,  -6, -28,  16,  -4,   4,   1,  10,  -3,  32,   3,  22, -11, -10,   0, -15, -17,  -9, -27,  14,  27,   1,   1,  10,   1,  24, -17, -18, -32,  -6,   8,   8,  -5,  13,  17,  28,  22, -42,   0,   0,   7, -10,   3, -13,  39,   7, -14 },
      {  -3,  12,  12,   1, -11,   6,  -1,   3,  -5,   9, -16,  19,  24,   4, -10,  -3, -24,  -3,  -1, -11, -20,  16,  -9,  -3,  16,  57,  -3,  23,  10,   7, -19,  -9, -35,   8,  -7,  11,  -8,  39,   1,  -8,   0,  15, -19,  23,  -5,  -5,   7, -16,  -4,   1,  -2,  -6,  21,  22,  28, -18,  -1,  -8,   3, -11, -13, -10,  35,   4 },
      { -18, -24, -48, -38,   8,   0,   1,   4,  -9,  -1,  -5,  21, -45,  27, -18,   4,   3, -14,   4,  17, -20,  -6,   0,  -2, -44, -19,  17,  -2,  11,  -4,  -4,   2, -20,  14, -13,  -2, -20,   2,  -4,   0,   6,   5,  18,  13,  -5,  -7, -19, -12,   1,  10, -12,  22,   6,  19,   4, -18,   2,   3,  -1,  -2,   7,  -9,  20,  -7 },
      {  -8,  -7,  -7, -33,  38, -10,   0,   0, -17, -21,  -5,  -9, -11,  -3,   3,   4,  -6,   9,  49,   0,   7, -24,   1,  -6,   7,  32,   9, -20,  25, -11,  -9,   4, -35,  30,  16,  10,  10,  18, -16,  -1,  -4, -21,  -3, -11,   6,  11,  22,  -3,   5, -16,  17, -21,  -3, -19,  -6,  17,   1,  -4,   6,   7,  -5,  13, -25,   5 },
      {  -4,  -1, -14, -19,  33, -10,  -2,  -1,  -1,  15,  21,  10,  10, -23,  -6,  -3,   4,  27,  -4, -18,  42,   4, -24,   2, -23,   1, -30,   6, -10,  14,  -8,   1,  40,  34,  -6,   2,  14,  -1, -19,  -5,  -6,  21,  -4,  20,   0,  19,  -8, -14,  -6,  14,  -6,  -3,  36,  -3,   4, -13,   0,  -7,  16, -17,   6,  28,   2,  -5 },
      {  -1,   7,  25,  11, -44,  21,  -7,   4,  -5,   2,  -5, -30, -19,  -4,   6,   1, -16, -19,  -2, -27,  -8, -15,  14,   1, -36,   8,  10, -17,  -7, -32,  -7,   6,  13,  35,  -8,   3,  12,   5, -37,  -1, -27,  37,   6,  -9,  -9,  17,  -2, -10,  10,  -1,  23,   2,  -8,  -4,   2, -13,   3,   3,  -1,  15,  -2,   9,   7,  -9 },
      {  -1,   3,   6,  20,  10,  -7,   2,   1,  -8,  -4,  -6, -19,  43, -11,   1,  -1,  -5,  -8, -32,  19,  -2,  13,   1,  -2, -16, -47,  18,  12,   0,  -6,   0,  -3, -50,  37,  32,   0,  14,  -4,  -9,  -5, -25, -18,  18,  24,  11,  11,   3,  -5,   3, -19, -16,   7,  14,   0,  10,  -4,   2,  -1,  -8,  -9,  13,   3,  -5,   7 },
      {  -1,  -7,   3,  -4,   8,  -2,   2,  -3,  10,  12,  26,  49,  12,  17,  -2,  -7,  10,  -3,  23, -12, -18,  50,  -6,  -2,  -5,  -8,  15, -23, -10, -16,  33,  -5,   4,   4,   6, -28,  -3, -17,  -3,  15, -21,  18,  -3,   3,   2,   9,  -2, -10,  10, -38,  16, -25,   0,   7,  16,   8,   1, -12,  -1,  12, -30,  -2,  -7,  13 },
      {   2,  10,   7,  21,  17,   5,  -9,  -3,  -7,   1,  -5,   2, -15,  15, -16,   5,   4,  16,   8,  22,  -5,  43, -23,   0,   1,  -7,  -7, -13, -16,   6,  16,  -2, -20,   5,  13,  14,   5,  15,  -2,   1, -21,  -6, -31, -16, -25,   4,  10,   5,  -3,  23,   7,  12, -12,  -5, -42, -14,  -6,   7,  -7,  19,   4,  32,  27, -43 },
      {   0,   6,  11,  16,  12, -25,   6,   2,  -2,   7,   8, -16,   3,  -6,   4,   3,   2,  13,  -2, -13,   4, -15,  -5,   7,   0,   4,  -8, -22,   4,  -8,   8,   6, -64, -21, -24, -17, -11, -35, -14,  25,  12,  23, -29,  -3,  -6,   0, -50,  -8,  -7,  18,   3,  -8,   4,  -6,   3, -22,  10,  -7,  14,  -5,  11,  -4, -15,  -2 },
      {  -6,  -9, -15, -38, -55,  28,   1,   4,  -2,  -6,   8,   0,  -8,  -4,   5,   4,   3,  10, -16, -35,  13,   9,   5,   2,  -6, -31, -43,   1,  -1,  -2,   8,  -1, -32, -15,  17,   7,   2,   6, -10,  -1,  11, -34, -25,  10,   6,   9,   7,  -4,  -2,   5,  -3, -31,   1,   6,  -3,   7,   0,   5,   7,   4, -24,   9,   2,  -3 },
      {  -4,  -8,  -8, -24, -22,  20, -13,   5,  -6, -16, -19,   1,   4, -30,  19, -11,   0,   7,  18,  50,  33,  -7, -15,  -8,  11,   2,   8,  15, -36,  24,  -8,  -2, -13, -29,  14, -13,  -9, -14,  -7,   0,  -9,  23,   5,  -7,  -6,   4, -10, -14,  20, -25,  31,  12,   7,   1,  13, -15,   9,  -9,  -6,  14,   2,  11,   8,   0 },
      {   5,   4,  10,  11,  14,  33,   8,   4,  -4, -12,   2, -27, -40,  19,  13,  -9,   5,   1,  10,  -9,  16,  14,  18, -15,   1,  -9,   7, -22,   7,  11,   4,  -8,  -3,   2,  12, -18,  40,  -5,  32, -12,  -1,   0, -17,  -6,   4, -20,   0,  16,   6,  -1,   6,   7,  31,  10,  16, -16,   4, -25,  14, -33,   2,   4,  26,  31 },
      {  -7,  -2,  -6, -22, -13,  -5,  11,  -3,  -3,   1,   8,  33,  26, -31,   4,   2,  -3,   8,  20,  -2,  -5, -21,  24,   4,  -5,   0,  -9,  -6,  -2, -15,  -4,   2,  -2,   7,  28,   6,   2,  -7,  26,   0, -63,   8,  -8, -13,  -4, -24, -10,  36, -15,  31,  -2,   2, -16,  -1,  -7, -11,  -5,  -8,  11,  -8,  13, -24,  15,   3 },
      {  -4,  -3,   2, -11, -15,   0, -12,  -6,  -7,  -4,  -5, -24, -15,  13, -26,   1, -12, -13,  -8,  -3,  13,  -1, -42,  19,  -1,   7,  30,  24,  19,  27,  12, -10,  -2,   2,  25, -26,  -3,  -5,  27,   2, -25,  35,   5,  -8,  15,   7,   0,   6, -21,  30, -15, -33,   0,   3,  -6,  17,  17,  -1,  -1,   0, -16,   3, -25,  -4 },
      {  -1,   4,   3,  13,  38,  41, -22,   3,  -3,   3,  12,   3,  -5,  35,   7,   6,  -6,   7,  -2,  -7,  23, -19,  44,  -9,   2,   8, -16,  26,  -9,  -4,  12,  11, -15, -28,  -2,   4, -18,   3, -12,  10, -35,  10,  30,  17,   0,  10,  -2,   6, -38,  -8,   0,   0,   1,   9,   2,  11,  -7,   3,  -8,  -5, -14,   9,  -4, -14 },
      {  -1,   1,  -4, -11,  -8,  36, -17,   3,   1,   5,   4,   1,  28,   6,  18,  -8,   2,  11,   8,  28,  16,   2,  29, -11,   1,  16,  31,  -1, -14,  -9,  10,  15,   3,  42, -14, -30,  -6,  -7,  -4,  15,  33,   0, -32,   6,   0,   1,   2,  15,  13,  21, -42, -13,  -4,   4, -13,  13,  15,  12,  -6, -10,  -2,  -5,   3, -16 },
      {  -4,  -4,  -6, -13, -21, -10,  16,   5,  -1,   0, -10, -10,  -5,  15,  19,  -2,  -7,  -5,  -2,  -4,  11,   8,  -5,   0,   6,  21,  -3,  -5,  27,  28,  -2,  -2,  10,   1, -32,  -3,  13, -10,   8,  17, -49, -18, -23,  28,   1,   2, -22,  11,  10, -54, -28,  18,  -7,   1, -18,  -8, -16,   8,  -7,  -5,   7,   6, -14, -21 },
      {  -3,  -2,   2, -15, -24,  15,   6,  -6,  -5,  -3, -11, -11,  20, -15, -14,  13,   2,   2,  17,  18,  22,  26,  -7,   1,  -2,  10,   1, -26,  29, -13,  53, -11,  -4,  -2, -26,  32,  11, -11,  -1,  15,  13,   9,  25,   5,   0,   4,   7,   2, -41,  10,  13,  22,  16,  -3,   7,  17, -16,  10, -16, -12,  14,  -8,   7,  12 },
      {   2,   0,   2,   0,   3, -11,  10, -14,  -1,  -4,  -9, -18,  -1,  21, -26,  -4,   2,   0,   0,   2,  46,   1,  -7,  11,   2,   7,  -9,  26,  -1, -51,  19,   9,  13,   0,   3,  11, -29,  15,   0,  21, -20, -24,  -8, -23, -13,  -8, -13,  -8,  43,   1, -11,  -5,  14,   7,  -1, -15,  10,  -7, -24,  -1,  -3, -14,   7,  31 },
      {   0,   3,  -1,   2,  13, -15,  40, -21,  -5,  -1,  -8,  -2, -34, -34,  42, -12,  -2,   4,   0,  -5,   3,  11, -13,   9,  -1,   0, -12,  -3, -12, -18,   3,  16, -19, -13,  -1,  -9,  -5,   4, -10, -11,  -5,  38,   6,  -1,   7,   1,  33,   1,   2,  -5, -48,   8,  -3,  11,  -6,  31,  -2,   8, -23,  -8, -13,  -6,  19,   1 },
      {  -1,   1,  -1,  10,   9, -19,   8,   3,  -5,  -2,  -5, -10, -34, -48,   8,  23,  -3,  -4, -11,  14,   0,   5,  15,  12,  -2,   5,  24,  11, -25,  13,  12,   2,  10,  16, -18, -21,  -4,  15,  -3,  21, -11, -41,  -9,  22, -17,  -3, -13,   4, -29,  11,  30, -23,   3,  12,   6,  -1, -16,  13,  -2,   2, -34, -10,   7,  15 },
      {   0,  -1,   2,  -2, -11,  11, -20,   6,  -1,  -1,  -3,  -6,   7,  -4, -24,   6,   2,   4,  13,  15, -17,  11,  -9,  10,   6,   6,  -2, -45, -29,  -4, -29,  17,   2, -25, -32, -19,   5,   6, -17, -15, -34, -31,  25, -21,  14,  -7,   6, -22, -12,  12, -37, -13,  15,   1,   3, -10,  10,  -1,  -1, -26,  -3,  10, -13,  12 },
      {   0,  -5,   1,  -2,  -4, -22,  14, -25,   3,   0,   5,  -4,   0,   1,  10,  -4,  -1, -11,   0,  -5,   6,  -9,  17,  -5,   6,   8,  14,   4,   4,   2,  37, -12,   4, -12,  -8, -12, -10, -10,   3,  -6, -10, -23,  22,   0,  16,   4,  28, -20,   9,   2,   3, -29,   5, -26,  18, -21,  27,   3,  30, -10,  21,   6,  43, -53 },
      {  -4,  -3,  -3,  -6, -29, -61,  16,  -4,  -3,  -2,  -3,  -9,   6,  40, -18,   4,   3,   8,   9,   9,  27,  21,  42, -13,  -3,   0,   0,   7, -21,   0, -21,  14,   1,  -2,  14, -25,   8,   8, -13, -12,   6,  18, -19,  -6,  17,  -6,  15, -10, -31,  -2,  19,  13,   4,   8,  -3,   7, -15,   9,  -1,  -9,   6,  -3,  -8,  -3 },
      {   0,  -1,   2,  -1,   1,  29,  32,  -4,   0,   3,   3,   0,  10,  21,  42, -13,  -1,  -7,  -4,   7,   8,   4, -39,  13,  -6,  -2,  17,  11,   7, -26, -22,   8,  10,   9,   5,  -6, -21,  -7,  -8,  -3,   6, -22,  -9, -26,   2, -12,  -2, -20, -53,  -5,   4,   2,   4,  -9,  -4, -12, -33, -28,  24,   8,  -1,  -7,  -2, -11 },
      {   1,  -1,   2,   2,  -2,   1,   1,   5,  -1,  -5,  -5,  -5, -10, -22, -22,  14,   3,   0,   2,   6,   9,  30,   5,   6,  -2,  -2,  -1,   5,   5, -12, -14,  25,   5,   0,   3,  17,  -3,   9,  -6,  -1,  19,  10,  11,   8, -10,  -2,  -8,  29, -26, -45,  -3, -13, -16,   0, -24, -14,  60, -38,  21, -18,  -5, -23,  -1, -19 },
      {  -2,  -1,  -2,   3, -22, -21,  30, -23,   0,   1,  -2,   1,  14,  36,  24, -18,   1,   3,   4,  24, -10, -24,   2,  20,  -4,  -3,  -2, -18, -24,   9,  -2,   3,  -4,   5,  -7,  28,   8,  16,  23,  10,   9,   2,  25,   7, -36,  29, -13,  25,  -3,   8,  -1, -16,  11,   7,   3,   5,  13, -20,   4,  -9, -26,  38,  -2,   7 },
      {  -1,  -2,  -3,  -4,   0, -10, -12,  -1,   0,   1,   1,   5,   2,   9,  45, -17,   0,   1,  -4,  -6,  -3,  29,  12,   6,   0,   4,   0,   5,  26,  21,  -5,  23,   5,  -2, -10,   6,   8,  -2, -16,  10, -17, -16,   4,  11,  15,  -6,   4, -18,  -2,  43,  20,  14,   0,   8,  -9,  10,  54, -32, -26,  32,  14,  -5,  -2,  19 },
      {  -3,  -5,  -3,  -2,  -3, -26, -49,   2,  -1,  -1,   0,  -9,   5,   5,  -9,  -9,  -4,  -6,  -9,  -1,  10, -33, -27, -19,  -4,  -3,  -1,  -8, -16, -33,   5,  -9,  -4,  -4, -28, -14,   6, -14,  18, -14,  -8, -11, -19,  29,  -7,   3,  24,   7, -15, -10,   1,  16,  -6,   7, -12,  32,   0, -41,  14,  20,  -7,  -3,  32,  10 },
      {  -2,  -3,   0,   0, -14, -29, -32,  25,  -1,   0,  -1,  -3,  -9, -13,  25,  -9,  -3,  -4,  -3,  -5, -16,  10,   4, -51,   0,   3,   0,   9,   6,  16,  32,  -2,   5,   5,  14,  10, -31,  22, -15,  22,   2,   1,   9, -42,   4,   9,  -9,   7, -11,  -5, -31,   6,   3,   2,  -7,  -8, -15, -19,  23, -16,   3,  25,   0,  23 },
      {   0,   0,   2,   6,  -4,   1,  20,  13,   0,   1,   2,  -4,  -1,   7,  -2,   2,  -1,  -1,   2,  14,  22,  29,  -5, -27,   6,   8,  10,   8,   7,  -4, -13,   6,  -7, -10,  -7,  13,   7, -15, -20, -10, -11,   3,  32,  27, -10, -10,  -6,   4,  39,  34,  -4,  -5, -20,   1, -12,  19, -20,   7,  70,   9, -18,  -7,  -2,  14 },
      {   1,   1,   1,   2,  -2,  25,  39, -20,   2,   3,   3,   7,   9,   7, -25,  39,  -2,  -3,  -4,  -1, -17, -25, -10,  12,   1,   3,   3,   7, -12,  28,  23,   9,   0,  -1,   0, -20, -11,  12, -39,   9,  -5,  -5, -19,  -2,  26,  11,   4,  11,  10,  -3,  10,  32, -10,   4, -14,  27,  12, -25,  21, -14,  17,  16,  13,  26 },
      {   0,  -1,   2,  -3,   0, -15, -23,  49,   3,   5,   4,   4,   7,  29,  35,  43,   0,   0,   3,  -1,   3,  -5, -22,  42,   1,  -2,   0,   0,   2,   1,  -3,  23,   0,  -2,  12,   1,   7,  -7,  -2,   9,   2,   8,   8,  -4,  -8,   0,   9,   4,   3, -16,   7, -20,   8,  -5, -16,   4,   5,  40,   8, -15,  19,  -8,  44,  24 },
      {   3,   4,   1,   2,   5,  -1,  46,  64,   1,   0,  -1,   1,   7,   5,  -2,  22,   3,   2,   3,   5,  10, -14,   3, -19,  -1,   2,  -3,  -3, -11,  -9,   9, -44,   5,   2,  -3,   0,   7,   7,   5,  -4,  -2,   3,  -3,  -2,   7, -14,  -1, -12, -21,   2, -26,  11,  -3, -12,  10,  -7,  44,   8,  -6,  41, -21,  15,   1,   6 },
      {   2,   1,   3,   2,  -1,  -6,   7,  48,   0,  -3,   0,  -4,   0,  -2,   4,  17,   1,  -1,   3,   1,  -2,  -4,   0,   9,  -1,  -5,  -3,  -6,  -7,   1,   3,  -4,   3,  -1,   8,   5,   5,   9,   6,  20,  -4,  -2,   3, -10,  -6,   1,  -3, -23,  29,  14,   9,   7,  17,  18,  11,  44, -16, -49, -10, -32,  -5, -28, -11, -59 },
      {   0,  -1,  -1,  -2,   2,   1,   4,  10,   0,  -1,   1,  -2,   2,   7, -11, -22,  -1,  -1,  -1,   1,   3, -12, -10,   4,  -1,  -2,  -3,  -9, -20,  13,  20,  26,  -3,  -8,  -1,  -8,  35,  43,  -9,  29,  13,  13,  20,  18,  35, -16,  -2,  17,   5,   0, -12, -21,   7, -28, -11, -28, -33, -22,  -7,  39,   8, -30,   8,  -3 },
      {   0,  -1,  -1,   2,  -1,  -7, -14, -16,   0,  -1,  -1,  -2,   6,   1, -16,  -4,   0,   0,  -3,   3,  -2, -19,   5,  14,   0,  -3,  -1,   3,  12,  24,   6,  25,   4,   7,  18,  18,  26, -33, -22,  33,   3,   3,  -7, -10, -22, -53,  18, -27,  -6,  -9, -21,  19, -15, -13,  23,  -3,   2,  -3,   6,  -4, -52,  10,  15,  -4 },
      {   0,  -1,  -3,  -3,  -3,  -4,  11,  27,  -1,  -1,  -1,   0,   0, -10, -17,  27,   0,  -1,  -4,  -2,   1,  14,  21,  14,   4,   5,   8,   9,  21,  -1, -11,  36,  -2,  -2,  -9,  -6, -22, -37,  14, -17,   6,   2,   2,   1,   2,  25,  21,  23,  11,  10,  -9,  -7,  -2, -21,  18, -22, -33, -42, -28,  18,  -6,  40,  13,   4 },
      {   0,   0,   1,   0,   0,   4,   9,   2,   1,   1,   1,   1,   7,  -3,   5,  13,   0,  -1,   1,   3,  -4, -20,  12,  -3,   2,   3,   3,   0,  -1,  13,  46,  24,   2,   3,   3,   7,  11,  -2,  17, -50,   2,   9,  -4,   8, -24,   7,   9, -58,   0,  -4,  -6, -11,  -9,  26, -47, -32,  -9, -14,   0,  -3,  -1, -16, -22,  16 },
      {   0,   0,  -1,  -1,  -3,  -3,  12, -13,   2,   2,   4,   2,   3,  16,   3,  19,   0,  -1,  -1,  -3,  -1, -10, -20, -52,   0,   0,   2,   1,   0,  18, -15, -30,   2,   0,   1,   0,   6, -26, -45, -10,   0,  -3,  10,   0, -26, -13,   0,  19,  -3,   5,   3, -41,   0, -10, -18,  -4,  -5, -16, -55, -21,   0, -21,  15,   3 },
      {   1,   1,   3,   2,   2,  11,  29,  13,  -2,  -2,  -1,  -3,  -1, -11, -36, -36,  -3,  -2,  -4,  -4, -14,  -7,  10, -27,  -2,   0,  -1,  -1,   4,  19, -17,  33,   3,  -1,   4,   3,  -2, -23,  12,  24,  -8,  -6,  -4, -19, -27,  25,   2, -16, -15, -13,  -4, -20,  38,  11, -13,  35,   6,  17,   8,  29,  25, -11,  24,   4 },
      {   1,   2,   1,  -2,   1,   4,  20,  38,  -1,   0,   0,   3,  -3,  -1, -12, -47,  -1,  -1,  -2,  -5,  -1,  -7, -23, -12,  -1,   0,   0,   4,  -5, -14,   6,  17,   0,  -1,  -3,  -6,  -2,  -1,   4,  26,   1,  -3,   3,  24,   0, -15,  30,  -2,  -6,   7,  26, -10, -52,  21,   6, -15,   5,   8, -14, -45,  23,  32, -10,   5 },
      {   0,   0,   1,  -1,   1,  -5,   1,  31,  -1,  -2,  -2,   0,   2,   1, -14, -60,   0,   1,   1,   1,  -3,  -1,  11,  30,   2,   2,   4,   4,   8,   9,  20,  -2,  -1,   1,  -1,  -1, -10,   1, -25, -54,  -3,  -6,  -5,  -3,   7,   3, -41,   5,  -3,  -2,  10,   6,  -3, -40, -14,  24,   0,   4, -11, -23, -25,  -5,  25,   4 },
      {   0,   1,   0,   1,  -3,  -1,   6,   1,   1,   2,   2,   3,   6,   5,   8,  13,  -2,  -1,  -2,  -3,  -8, -10, -34, -50,   0,  -1,  -2,  -2,   2,  -6,   8,  59,   0,   0,   0,  -2,  -3,   1,  19, -10,  -6,  -8,  -4,   0,  18,  18,  -1,  16,   4,  11,  23,  31,  23, -21,   8, -14,  20,  28,  -8, -15, -45, -18,  -8, -16 },
      {   0,   0,   0,  -1,  -2,   4,  -1, -18,   0,   0,   1,  -1,   0,  -3,   3,  27,  -1,   0,  -1,  -1,   3,   7,  -7, -26,   0,   0,   1,   0,  -3,  -9,  -1,  32,   2,   4,   5,   1,  12,  25,  27, -14,   2,   4,   2,   7,   1, -26, -58, -32,  -6,  -6,  -5, -21, -32, -22,  24,  40,  -2,  -4, -15,   2,  25,  32,  20,  -1 },
      {   0,   0,  -1,   1,  -1,  -7,  -1,   9,   0,   0,   0,   0,  -2,   6,   2,  -4,   0,   0,  -1,  -1,   2,  -5, -16,   5,   3,   2,   2,   0,  -1,   6,  33,  25,   1,   2,  -2,   2,   5,   2, -32, -45,  -4,  -9,  -8, -11, -26, -30,   0,  44,  -3,  -5,  -5,  -9,  14,  43,  50,   6,   4,   7,   7,  28,  27,   3, -23, -18 },
    },
    {
      { -87,   5,  31,   4,   2,  -3,   2,   0,  54, -43, -29,   0,   1,   2,   1,   1,   4,  36,  -2, -12,  -6,  -5,  -2,   0,  -9,  -8,  14,   5,   7,   1,  -1,   0,   6,  -1,  -7,  -1,  -3,   6,   0,   0,  -2,   4,   2,   3,  -2,  -3,   0,  -1,  -1,   2,   1,   0,   1,   2,  -1,  -1,   0,  -1,   0,   1,  -1,   2,   2,  -1 },
      { -46,  44, -18,  -4,   5,   2,   1,   0, -27, -25,  46,   7,  -1,   2,   0,  -1,  41, -36, -48,  19,   6,   3,   2,   2,   0,  32,  -8, -31,  -5,  -1,   0,  -1,   2,  -5,  18,   7, -10,  -5,   0,   0,   0,   0,  -4,   2,   5,  -3,  -1,   1,   3,  -3,   1,   0,   0,   0,  -1,   0,   0,   2,  -1,   0,   2,  -1,  -2,   1 },
      { -35, -19,   2,  12,   1,   0,   0,   1,  20,  32,   5, -29,  -4,  -4,  -1,  -1, -18, -39,  32,  44,   0,  -2,  -4,   0,   1,  -4, -61,  -8,  31,   8,   4,   2,   6,   9,  19, -33, -20,   8,   0,   0,   0,   0,  -2,  21, -10,  -6,   1,  -1,  -2,   6,   2,  -5,   6,  -4,  -1,  -1,   2,  -3,   0,   3,  -2,   3,  -2,  -1 },
      { -41,  40,   0,  -5,  -1,   4,  -1,   0, -26,  35,   4,  -1,   2,  -5,   1,   0,  22, -24,  39,  -3,  -9,   4,   0,  -2,  19, -22,   1,  53,  -6, -10,   0,   0, -16,   0, -32,   0,  42,   0,  -3,  -1,   6,  -8,   4, -26,  -4,  19,   5,   0,   3,  -3,  -5,   3,  -9,  -5,   5,   5,  -1,   0,   1,  -4,   3,  -3,  -3,   1 },
      {  24,   0,  16,  -6,  -9,  -2,  -1,   0, -28, -58,   2,  15,  18,   1,   3,   1,  18,  -4,  -1, -11, -14,   5,  -2,   2, -18,   9, -29,  30,  21,  -7,   1,  -2,   7,  -1, -12, -53,  10,  25,   2,   4,   2,   9,  13,   6, -48,  -7,   6,   0,  -3,   8,  -1,  11,  13, -19,  -6,   1,   1,   0,   3,  -2,   0,   9,  -3,  -2 },
      { -13, -46,  -2,   6,  10,  13,   0,   1,  30,  28, -13,  27,  -2,  -1,  -2,   0,  -1,  -4, -38,  -5,  11,  16,   7,  -1,   9,  43, -14,  -4, -24, -12,   7,  -2, -26,  -6,  13, -20,  35,  -9,   0,   1,  12, -14,  12, -30, -27,  16,   7,   3,   3,  -2,  -2,  17,  -9, -25,  -1,   7,  -1,   3,   3,  -5,   9,  -3, -14,   2 },
      {  19,  51,  15, -17,   4,   0,  -4,  -4,  32, -36,  16,  25,  -7,  -4,   0,   5, -63,  -9,  -1,  21,  35,  15,   0,  -7,  27, -16, -20,  -8,  -1,  13,  12,   6, -19,   0, -13,  -7,   3,  -6,  -4,   5,   7,   4,   7, -17,   1,  15,   2,  -7,  -1,  -2,  -5,   4, -17,  -3,   7,   7,  -3,   1,  -4,  -3,   1,  -6,  -1,  -2 },
      { -23, -39,  34,  12,   3,  -1,  -1,   1, -52, -21, -20,  19,  21,   5,   2,   1, -14, -29,   6, -37,  14,  11,   4,   1,  43, -10, -22,  -9, -31,  15,   0,   2,   0,  25,  -3,   8, -20, -16,   6,   0,   5,   4,  -4,   6,  14,  -4,  -7,  -1,   0,   5,  -4,  -9,  -3,   8,   7,  -3,   1,  -1,   0,  -1,  -5, -10,   2,   4 },
      {  17, -37,  28,   7,   4, -10,  -2,  -1,  12, -37,  22, -15,  -8,   2,   0,   1,  29,  -3,   8,  30, -10,  -7, -11,   1, -19,  18, -23,  21,   8,   3,   0,  -2,  -1,  10,  -2,  19,  29, -13,  -6,  -1,  -4,  -3,  -9, -18,  34,  42, -11,  -4,  -7,   0,  -3, -14, -33,  21,  23,  -5,   0,   0,  -3,   4,  -4, -16,   6,   4 },
      {  11,  32,  12,  -1, -18,  -1,  -1,  -3,  11,  -8, -26, -19,  26,  -4,   3,   3, -13, -11,  13, -17, -32,  12,   0,  -1,  32,  46,   5,   6,   0, -31,  -2,   3, -18,  -7,  53,  10,  10,  17, -21,  -1,  -1, -12, -16,  33,   5,  17,  14,  -7,  -6,  -3, -10, -16,   1,   2,   7,   7,   0,  -2,  -1,   4, -10,  -7,  -1,  -2 },
      {  18,  18,   2,   4, -13,   3,   1,   2,  52,  10,   0,  28,  27,   5,  -6,  -6,  18, -29,  -9,  -8, -20,   8,  10,   3, -16,   4, -22,  28, -21,  -4,  -3,  -3,  -6,  26, -16,  -7,  -7, -27,  20,   2,  -5,  -5,  11,   4,  31, -38, -16,  10,   7,  -8,  -3, -17,  24,  23, -18,  -1,  -3,   2,   7,  -1, -10,   2,   0,   1 },
      {   4, -28, -15, -25,   8,   3,   1,   1,  42, -23,  42, -18, -12, -16,   1,  -2,  15, -36,  27, -21,   8,  10,  -8,   4,  15,  -6,  15,   9, -37,   4,  -3,   3,  -3,   7,  -9,  31,  -4,  13, -10,   2, -13,  15, -23,  17, -26,  -2,   9,  -3,  -4,  12,   0,  12,  16, -25,  -6,  -3,  -2,  -1,  -2,  -2,   9,  15,  -2,  -3 },
      {   9,  29,  -8, -18,  11,   3,  -1,   3,   7,   3, -23,  -3, -17,   5,   6,  -1,  31,   8,  19, -23,  10,  -3,  -7,   4, -29, -19, -16, -15, -31,  16,   9,  -1,  19,  30,  32, -19,  -7, -34, -12,   1,   5, -21,  23,  26, -10,  32, -13,  -6,  -4,   7, -22,  16, -11, -12,  21,  -4,  -1,   0,   7,  -2,  12, -11,  -9,   5 },
      {   8,  12,  81,  -9,  -2,  -5,  -1,  -2, -11,  21,  -6,  11, -15,  -6,   0,   1,  -3, -11, -17,  22,   3,  -2,  -4,   3, -25, -14,  -3, -11, -14, -23,   3,   2,  26,   7,   6,  24,  16,  -6, -20,  -9, -12,  -4, -15,   2,   0,   0,  14,   5,   5,   3,   8,  -5,  19, -27, -23,  -2,   3,   0,   4,   3,  13,  31,  -9, -14 },
      {  17,   5,  47, -14,   6,   0,   1,  -4,  17,  26,   2,   8,  16,   6,  -8,   1,  51,  -4,  -2,  -3,  12,  -3,   4,  -3,  12,  -9,  12, -18,  21,  33, -13,   2, -31,  -9,  -9,  -4, -33,  33,  19,  -2,  -7,  -8,  -8,   2, -13,  -4,  22,   4,   1,  -9, -13,  -8, -27, -11,  10,  11,  -2,   5,   2,   4,  -5, -30, -12,   2 },
      {  10,   0, -18,  21,   6,   6,  -1,  -1,  -8, -26, -17, -14, -27,  11,   7,   8,  26,  34,   5,  34, -13, -11,   4,  -5,  38,  -6, -18,   1,  -6,  -3,   5,   6, -41,  36,  -6,   8,  -9, -12,  -4,  -7,   0,  -9,   5,  -1,  14, -20,  27,  -1,   9, -11,   0, -10,  11, -32, -21,  18,  -3,   4,   6,  -2,  16,  11, -24,  -9 },
      {   3,  15, -31,  48,  -7,  -1,  -5,   2,   7,  -9, -26,   3,  23,   9,   7,  -3,  -7, -17,   5,  -9,   9,   8,   0,   2, -31,  -3, -11,  -3, -10,  14,  10,  -2,  29, -13,  -9,  12,  -1,   3,   7,   3, -24,  11, -40, -22,  -2,  10,  22,   6,   1,  -3,  15, -30, -22, -39,   6,  13,   4,  -3,   6,   7,   3, -12, -28,   2 },
      {   0,   0,  18,  17,  -6,   7,  -9,  -4,  -7,   5,   7,   3,   9, -25,   8,   1,  25,  24,  29,   5,  24,  22, -14,  -8,  -7,  15,  -4,   3, -12,  12,  38,   5, -24, -63,   0,  -9, -10, -36,  -7,  21,  14,  22, -10,  18,   9,  -8, -18, -17,   3,  -7,   2,  -5,  17,  -2,  -1,   8,  -2,   4,  -5,  -4,  -4,  16,  -6,  -3 },
      {   4,  16,  15, -19,   3, -15,   1,  -1,  -7,  -7, -33, -29, -23,  27,   5,   4, -14, -24,  20,   1,  -9, -16,  -3,   3, -20,  49,  -2,  10, -15,  25, -10,  -2, -21, -17,  -5,  18, -18, -10,  26,  -3,  17,   3,   2, -25,  -9, -24, -15,   7,  -3,   9,  22,  30, -12,  -3, -13, -15,   4,  -4,  -2,  12,  19,  -1,   1,   1 },
      {   0,  -3,  18,  25,   0,   2,   1,  -1,   1,   3,  27, -41,  18,  29,  -3,   5, -19,  -5, -29,   5, -50,  23,  19,   1, -16, -33,  10,  -1, -20, -20,   8,   6, -15, -10, -16,  -3, -21, -17, -12,   1,  20,  14,   9,   9,  -9,   2,  -3, -14,   3,  -2,  -9,  21,  -7, -13,  19,   6,  -1,   4,  -3,  -5,  16, -25,  -8,   9 },
      {  16,  12,   0,  33,   5,   6,  -3,  -4,  12,  -5, -31, -41,  16, -18,   1,   5,  32, -20, -10, -10,  20,  19,  -3,  -5,   6, -18,  -3, -32,   0, -11,   3,  10, -10,  10,  -7,  -7,   7,  11, -23,  -6,   1,  14,  18, -28,  -5,   7,  -2,  -2,   6,   8,  23,  17, -13,  26, -23,  -8,  -7,   2,  -9,  12,  -7,  21,  40, -14 },
      {   0, -11, -14, -53,  -9,   1,   3,   2,   1,   5, -50,   8, -21,  -4,   9,   5,  13, -28, -24,  15, -10,   2,  -5,   0,   8,   1,   3,   2,  11, -15,  10,   5,   6,  -5, -29,  -5, -13,  -4, -26,   8, -12,  51,   9,   7,  13,   8,  -2, -26,  -5, -11,  20, -10,   6,  -2,  25,  15,   0,  -3,   2,   2,  -6, -13, -12,   8 },
      {   8,  10,  19,  21,  -4, -12,  -6,  -5,  11,  10,  16, -29,  16,  -1,   7, -10,   0,   5,  -7,  -3,   1, -34, -15,   3,  52,  27,  16,  -1,  -1,  11,  -4,  -8,  43,   5, -15, -18,  10, -20,  12,  17, -37,  17,  34,   3,  -2,   4, -20,   4,  -2, -14,   6,  13,  -5, -15,   6,   5,   1,  -4,  -4,  -4,   6,   8, -23,  -4 },
      { -14,  -5,  -1, -37,  -8,   5,   4,  -6, -13,  10,  26, -22,  25,  -8, -12,   6, -15,  30,  15, -16,  10,  26,  -1,  -7, -30,  29, -20,   0, -16,   2,   4,  11,   7,  21,  -9,   6, -10,   7, -11, -14, -13,  11,  39,  -6,  15,  -4,  35,   0,   3, -26,  12,  -3, -21,   0, -16,  24,   0,   0,  -4,  16, -11,  -5,  -7, -19 },
      {   2,  15,   7,  48, -25,   2,  -5,   2,   1,  14,  -3,  22, -29,  -5, -15,   1,   4,  12,   4,   9,   9,  15,  -6,  -7,  -7,  28, -12,   7, -12,  10, -11,  -2,   4,  22, -14,  25,  -4,  26,   3, -20,  -4,  30,  12,  17, -10,   1,  20, -15, -11,  -8,  -7,  26,  23,   6,  33,  -7,  -1, -13, -13, -24,   7, -13,  26,  20 },
      {   2,  -6,   1,  26,  18, -17,   4,  -3,   6,  -9,   5,   3,  -7,  -1,  15,   1,  -7, -38,  10, -29,  18, -39,   4,   3, -24,   3,  17, -10,  21,  -2, -17,   3,  -8,  -8,  -5,   1,   8,   2, -13, -19,  48,   7,  33,   8,  23,   7,  21,  -6,  -9, -19, -16,  -2,  16,   1,  -1,  32,  -1,   6,  -2, -13,  -5,  15, -29,  -6 },
      {  -2,  -4,  -3,   7,  65,  11, -13,  -4, -10,  -3, -22,  -2,   8,  22,  -3,   0, -13, -21,  -7,  28,   5,  20, -12,  -7, -15,  17,  32,  25,  -2,  21,  19,  -2,   0,   4, -12,   1,  -8,   7,  -2,  13, -36, -19,  12,  22,  -2,   9,   9,  -7,  14,  -4, -33,   2,   6,   8,  -7,   9,  -4,   7,  -2, -13, -12,  13,  18, -15 },
      {  -4,   7,  10,  10, -16,   8, -11,   8,  -8,   9,  12, -36, -33, -23,  23,   2, -18,  -2, -46, -36,  17,  -3, -23,   6, -19,  -4,   5,  28,  12,  22,  15, -11, -40,  31,   6,  -9,   4,   1,   9,  17, -10,  -9,  -7,   8,  -4,  -2,  -3, -11,   1,   7,  14, -25,   6,  -3,   7,   3,   4,  -7,   3,  13, -18,  -1,  -3,   9 },
      {  -3, -14,   2,   2, -62,  13,   5,   5,  -6, -21, -14, -14,  -7, -36,  -4,   2,  -1, -27,  12,  28,  21,  15,  12,  -3,  -9,   5,  47,  -5,   6,  -8,  16,  11,  11,   7,  -8, -10,  -1, -13,   8,  -3,  -2, -35,  15,  -1,   3, -16,   2,  17,  -2,  -3, -28,   5, -12,  -7,  -9,  -1,  -1,   2,  -1, -11,   8, -23,  -6,   6 },
      {  -4,   0, -11, -21, -12, -32,  10,  -4,  -5,   2, -15, -30,  57,   0, -12,  -7,  -2,  13, -23,  23,  38, -13,   9,   7, -10,  -2, -10,   2, -13,  14, -21,  14, -27,   0, -15,   6,  18,   3,   0, -15,  -2,  -3,  -6,  14,   6,  10,  -9,  -1,  -9,  27,  -8,  -5,  22,  -9,  18,  -6,   9, -12,  16, -31,  10,  17,  -3,   8 },
      {   8,   9,  14, -10,  53,  15,  19,  -1,  -2, -10,  -6, -35,  -8, -30,   6,   3,   1,   8,   6,   2,   5,   9,   9,  -1,   1,   2,  -8, -12, -18, -19,  -5,  -4,  21,  -6,  -1, -11,  17,   3,  23,  -9,   7,  -5, -11, -14,  -1, -24,   9,  13,   9, -20,   9, -12,  25,   6,  15,   1,  -4,  -8,  -2, -26, -26, -25,  -9,  51 },
      {   5,   4,  -1,   3,   2,  -1,  13,  -9,   6,   5,  -2, -23, -41,   8, -16,   8,   6,   4, -30, -18,  14,   6,  14, -12,  21,   4, -18,  33, -11,   4,  -2,  17,  30, -36, -11, -15,  -4,  -5,   0, -26,  14,  11,  -5,  18,  10,  -1,  14,  24, -18,  22, -28, -22, -17,   0, -17,  -6,   6,   4,  21,  -4,  -6, -10,  16, -29 },
      {  -1,   8,   1,   6,   8,   3,   2,   3,   0,   8,   7,   7,   1, -14,   2,   2,  -9,   8,  14,   7, -22,  -4, -25,  17,  -2,  35,  20, -36, -18,   2,  -7,  -2,  -2,  26, -56, -35,  -5,   0, -23,  -9,  29, -11, -26,   2,  -7,   6,  -2, -12, -10,  12,   1, -28,  -1,  -1,   3, -26,  13,  19,  23,   6,  -9,  -5,   5, -18 },
      {   3,   0,  -4, -14,  -5,  35,  -4,  11,   3,   2, -13,  -8,  31, -30,  22,  -2,   0,   3, -24,  18, -17,   5, -49,   1,   9, -10,  -8,  16,   0,  24,  -8, -28,  25,   2,   8,  22,  -9,   3,  24,  -2,  53,   9,  -3,  -6,   6,   8,  15,   9, -10,  -6, -17,  12,  -2,  -6,  -2,  10,  -4,  14,  -7,  -5,  13,  12,   8,  -9 },
      {   3,   4,  -1,  -4,  17, -60,  -6,   2,   5,  11,   1,  -4,   1,  -5, -11,  -2,  12,   6,  -2, -13,  -3,  39,  12,  -6,  11,  14,   2,   0,  37,   1,  43,   0,  26,  29,  -9,  22,  -4,  -3,   2,  24,  35, -16,  -6, -12,  -3, -10,  -1,  -9, -22,  25,   4,   1,   4,  -6,  -4,  17,   4,  -9,  -8, -14,   6,  -1,  -5,   3 },
      {   5,   0,   1,   2,  -3,  25,   8,   2,   5,  -2,   2,   0,   2, -11, -14,   6,  -1,  -1,  -5, -14,  -2, -14,  -5,  -5,   9,  11, -17,   6,   8,   5,  -4,  -1,  21, -25, -24,   7, -16,   2, -19,  -9,  -3, -53,   9,   0,  17,  -1,  14, -35,  52,  49,  17,  12,   9,   4,  15,  20,  -3,  -3,  -5,  14,  20,   1,  -4,  17 },
      {   1,  -5,   3,  -4, -21, -17, -12,   0,   4,  -4,   4,  -3,  -4,  42,  -9,  -5,   8,  -4,   2,  24,  12,  10,  -3,   2,  11, -11,  31,  10, -37,  21,   2, -16,  15,   2,  19, -44,  11,   3,   6, -14,  23, -12, -15,   3,   9,   3,  31,  -9,  -5,  -2,  45,   5,  13,  17,   8,  30,  -3,  -8, -13,  13, -19,   7,   6,   2 },
      {   2,   2,   3,   6,  16,   4,   9,   2,   2,   3,   5,   0,  13,  -6,  -4,   5,   1,   4,  -3,   3,  32,  11,  21,  -3,   4,  14,  23,  47,  20,   2, -22,  13,  12,  26,  21, -18, -36,  -9, -32, -11,  18,  15, -23, -18,   2,  12, -11,  -2,  21, -29,   7,   0,   9,  -5,   2, -27,  -9,   3,   9,  20,  35,   3,  -7,  14 },
      {   0,  -1,  -7,  10,  24, -34,   7, -11,  -4,  -7,  -5,  15,  -1, -50, -52,   9,  -5, -12,  -5,   2, -24,  12, -26, -15,  -2, -17,  12,   5,  -4,   5,  -6, -15, -17, -12,  13, -12,   4,  -5,  16, -19, -10,  -1,  22,  23,  17,   0,   7,   5,  -7,  -7,  30,   0, -17, -16,   1, -17,  22,  -4,   8,  16,  23,   1,  -4,   7 },
      {   1,   5,   6,   5,  21,  58,  13, -10,  -2,   2,   6,  13,  17,  -7,  -4,   4,  -1,  -1,   5,   5,  -1, -31,  22, -13,  -8,   7,  16,   8,   6,  -3,  19,  20,  -7,  20,   7,  -5,  -2,  -7,   3,  10,   4,  14,  -7,  13,  13,   3,   2,  17, -46,  39,  38,  11,  -7,  -8,   0,  17,   8, -30,  -9, -12,   8, -13,   9, -11 },
      {   2,   3,   0,  -3,  -7,  41,   8,   8,  -1,  -7, -10, -15,  -9,  15, -56,  10,   2,  -3,  -2, -11,  -3,  26,  18,   1,   3,  -1,  -5, -11,  26,  21, -12, -14,   1,   1, -12,  12,  14, -14,   8,  -4,   4,  -9, -10,   5,  -8,  18, -15,  -5, -12, -19,   9,   4,  19,   7,   9, -20,   9,   7, -28,  -8, -19,  23, -46, -29 },
      {   0,   1,   4,  -7,  -4,   7,   5,   0,  -1,   1,   5,  -4,  -3,  15,  -7, -14,  -2,  -4,  14,  -6,  -7,  17,  -3, -28,  -4,   8,  18, -22,  19,   3,   9,  -8,  -6,  16,   7,  -7,  27,  -9,  34, -12,  11,  45,   5,   5,  11,   6,   7,  11,  56,  36,  -8, -18,   4, -11,  14,  -9, -44, -11,  23,   1,  10,   6,   1,  -9 },
      {   1,   3,  -2,  17,   6, -11,  15,   0,  -4,   1,  -6,   9,  -3, -16, -17,   0,   0,   6,  -5,  27,   7,  -1, -19,  12,  -2,  -4,   4,  16, -26,  -6, -17,  15,   3,   6,   9,  16,  -5,  24, -10,  18,  19,  12,  13,   6, -22,  -3, -30,   6,  14,  30,  -3,   0,  -8,  17, -15,  11, -30, -20,  -6,  21, -35, -24, -53, -16 },
      {  -2,   3,  -1,  -1,   2, -26,  14,  -8,  -5,   3,  -5,   1,  -7, -32,  10,  10,  -6,   5,  -4,  -9, -10,   1,  13,   9,  -8,   3,  -6,  -2,   3,   3, -22, -15,  -9,  10,  -5, -25,   0, -20,   4,  -5, -24,   9, -54,   8,   8,   5,   3,   5,  17,   3,  -5,  43,  -7,  19,  -7,  38, -25,  18, -25, -25,  10, -13,  -1, -33 },
      {   0,   1,  -2,   0,   8,   7, -50,   5,   0,   2,  -4,   2,  -5,  -1,  24, -11,   1,   6,  -2,   5,  16,  23,  30,   9,  -1,   7,  -6,  -1,   3, -16, -24,  -4,   2,   0, -31,   7,  12,  10,  13,   2,  15,  -7,   2,  61,   9,   8, -12,  22,   0,   2,  25,   4, -35,  -4,  -9,   0, -15,  17,  -8,  18,  10,  14,  -9,  28 },
      {   0,  -2,  -4,   0,  -7,  -4,  48,  -5,   1,   2,  -2,  -2,   4,  28, -11,  14,   0,  -4,  -8, -10,  17, -19, -38, -35,  -3,  -1, -11,  -3,  -4, -14,  22,  13,   2,   6, -19,  -2,  18,  21,   0,  32,  17, -18, -22,  33,   8,  -8,  -7,   3,   7, -40,   8,   8, -14,  -4,  -7, -14, -26,   2, -16,  -3,  20,  -5,   8,   6 },
      {   0,  -3,  -4,   0,  -5,  -8,   2, -18,   2,   1,   0,  10,   8, -19,  26,  23,  -2,  -3,  -9,   6, -34, -10,  27,  -3,   4,  -3,   0,   1,   5,  62,  18,  40,  10,  -3,   1,  19,  23, -14, -17, -21,   7, -13,   6,  15, -26,  -4,   5,   7,  12, -20,  24,   2,  -1,   0,   5, -17, -31,   0,  11,  -2, -24,   1,   7,   0 },
      {   0,   2,   2,  -1,  -9,   7,  -4,  -9,  -2,  -1,  -5,  -5,  -3, -10, -51,   3,  -3,   0,   1,   4,  -8, -30,  25,  -7,  -3,   5,   0,  -4, -10,  20,   1,  -2,   5,  15,   5,   4,  13,  -2,  -4,  35,  14,  26, -14,   3, -15,   5,  -3, -10,  18,  -4, -19,  -2, -14,   3, -25,  39,  25,  33,  21,  11, -11,  15,  20,  44 },
      {  -2,   1,   0,   1,  -7,   1,  12,  15,  -1,   5,   6,   6,  -1,  19,   0,  35,  -2,   2,   1,  -7, -13,  22, -38,  -1,   3,   3,   5,  -8,  16,   9, -13,  16,  -7,  -4,  -8,  -8, -12, -19, -12, -17, -20,   1,  -5,  -8,   7,  24,   0,  28, -29,  17, -12,  10,   4,   2, -32,  16, -24, -13,   8,  -2, -13,  24, -11,  61 },
      {   2,   2,   5,  10,   1,  -4,  83,  30,  -2,  -1,   0,   4,  -8,  -4,  19, -17,   0,   1,   5,   9,  -3,  24,  26,  29,   2,   2,   4,  -1,  -2,   9, -17,  -9,  -3,  -7,  -4,   6,  -3,  -5,  16,  23,  -4,  -5,  22,   9,  -9,  17,   1,  -3, -10,  -3,  10, -21,   0,  -7,  10,  21,  -4,   2,   3,  22,   8,  21,  27,  -4 },
      {  -2,  -2,   1,   3,  -4,   1, -13, -19,  -3,  -3,   1,  -5,  -2, -12,  -9,  14,  -2,  -2,   2,   1,   2,  -4,  15, -31,  -4,  -1,  -4,   1,  -8, -17, -15, -21,   3,  -4,  -2,   3, -22,   8,  15,  23,  -5, -21,   2, -17,   7,   9,  -5, -18, -39,  -6,   9,  12,  12, -11,  20,   4, -60,  14,  60,   3, -16,  12,   4, -14 },
      {   1,  -2,   1,   0,   6,  16, -12,  -8,   0,  -3,   1,  -1,  -5,   4,   3,  -5,   1,  -2,   0,  -2,  -2,  25, -18,  -4,  -1,  -4,   2, -10,  11,   6, -38,  32,   9,   1,   2,   0,  28,  -9,  10,  -9,   1,  -8, -15,   5, -12, -30, -28, -42,  -9, -23,  -8,  -4, -13,   9,  -3,  40,  11, -49,  33,   5,  27,  -1,   7, -13 },
      {   0,   0,   0,  -5,  -4,  -7, -22,  23,  -2,  -2,  -3,  -2, -14, -21, -14, -41,  -2,   0,  -6,  -7, -28,  -9,  10, -10,  -3,   3,  -6,  -5, -18,   6,  -5,  43,  -3,  -6, -10, -19, -26,  17,   8,  23,   9, -11,  -9, -15,  21,  33,   0,  17,  15, -21, -10,   1,   4, -16,   3,  -1,  -6, -48, -20,  -9,  -9,  27,  10,  -5 },
      {   1,   0,   2,   5,  -8,  -3,  12,  12,   3,  -1,   4,   5,  -5,   7,  27,  10,   3,   0,   4,   2,   0,  19, -10, -40,   0,  -1,   2,   3,   2,  -5, -12,  11,   0,   5,   4,   7, -12,  11,  -1,  10,   3,  -5,   1,   5,   4,  13, -16,  32,  32,   0,  28,  27,   0,   3,  15,  12,  62,  16,  44, -28, -25,   4, -11, -14 },
      {   0,  -3,  -3,  -2,  -2,  -8, -15,  -2,   2,  -2,  -1,  -3,   7,  -2,   0,  37,   1,  -1,   1,   3,   8, -13,   9, -10,   1,   1,  -5,  -7,   2, -10,  -5, -43,  -4,  -6, -22,  12,  -8, -28,  -3,  17,   9,  -5,   2,  12, -34,  28,  18,  33,  15, -16,  -2, -15,  24,  21,  -9,   6,  10, -58,   7,  20,  12, -22,  14, -11 },
      {   1,   2,   3,   1,   9,   1,  -9,  78,  -2,  -5,  -1,  -4,   6,   0, -27, -47,  -3,   0,  -2,  -3,   7, -16, -10,  -1,   4,   0,  -5,   4,  11,   0,  19,  -7,  -8,  -6,  -3,  11,   7, -17, -28, -22,  -5,  -2,  -7,   8, -15, -17,   4,  20,  -8, -13,  14,  17,  -2,  10,   4,  14, -21,   9,  33,   2,  -2, -21,   6,  -6 },
      {   0,   1,  -4,   1,   2,  -7,  15, -28,   0,   0,  -3,   2,   7,  -4,  13, -44,  -1,   1,  -6,   5,   5,  -8,  -7, -47,   0,   2,   0,   3,   7, -23,   2,  16,  -3,   9, -14,   9, -11, -34,  23, -31,  -3,  -5, -16,   4, -28,   8,  10,  -7, -11,   3, -16,  14,   6,   6,  12,   3,  13,   9, -20,  53, -29,  10,  -5,   9 },
      {   0,   0,   1,   2,   3,   5,   2, -19,  -3,  -1,  -3,  -6,  -6,  -3,  10, -23,  -2,   2,  -2,  -3,  -8,  15,   5,   1,   0,  -3,  -5, -14,  -3,  29,  10,  13,  -1,  -6,  -4,  -7,  25,  26, -11,  29,  -9,   4,  -2,  -5,  26,   4,  17,  27, -18,   3, -15,  24,  44,  33,   1,  -6,   9,  11,  21,  44,  38, -21, -27,   2 },
      {   0,  -1,  -1,   2,  -7,  10,  13, -63,   0,  -3,  -1,   1,  -5,   8, -13, -47,   0,   0,   0,  -5,   6,   8, -18,  58,   3,  -1,  -3,   5,  19,  -8,  10, -14,  -4,  -1,  -5,  10, -13, -12,  -9,   4,   8,  -5,  -5,   7,   5,   9,  13,  11,  14, -13,  13,  24, -10,  -8, -11,  -7,  -2, -21,  32, -21, -21,  -3,   6,   8 },
      {  -1,  -1,  -1,   0,   0,  -4,  -1,   7,  -1,   0,   0,   1,   3,   5,   8, -22,   0,   0,   1,   2,   1,  11,  -1, -32,  -3,  -3,   1,  -2,  11,  19, -57, -11,   1,  -5,   4,  -1,  18, -28, -31,  40,   3,   1,   7,   2,  15, -13,  25, -19,  -4,  13,   2,   7,   7, -34, -29, -28,   1,  -3, -16,  -7, -27, -36,  14,   3 },
      {   2,   1,   0,   2,   4,  -2,  10,  36,   2,   2,   2,   2,   9,  -5,  13,  20,   3,   1,  -1,   3,   5,  -8,  20,  15,   4,  -1,   0,   0,   1,  -6,  25,  -3,   4,   1,  -3,  -3,   7,  11,  35, -11,   1,  10, -12,  13,  25,  19,  13, -51,   5,  -5, -13,  43,  -4,  -4, -37, -30,  14, -26,  27,  18, -24,  -7, -12,   1 },
      {   0,   0,   1,  -1,  -3,   4,  20,  -6,   0,   1,   0,   1,  -2,  -6,  18, -19,   0,   1,  -2,   2, -10,   2,  23, -35,  -1,   2,   0,  -1, -13,  19,  13, -61,  -3,   1,  -4,  -7,   0,  20, -43, -13,  -2,   2,  -7,  -2,   6, -26, -11,   1,  -7,   0, -19,   2, -37,  13,   4,  -6,   3, -38,   2,  -8,   0,  35, -23,  13 },
      {   1,   1,  -1,   1,   4,  -2,  -4,  13,   1,   1,  -1,   2,   2,   0,  -7,  27,   2,   0,  -1,   1,   6,  -2,  -6,  23,  -1,   0,   1,   3,   4, -14, -12,  20,   1,   2,   3,  -3,   6, -24,  -3,  39,   2,   6,  -8,   0,   1, -42,  52,  13,   1,   8, -16,  19, -30,  15,  54, -19,   4, -10,   6,  14, -11,  38,  -7,  -4 },
      {   0,   0,  -2,   0,   4,   0,  -1,  -4,   0,   0,  -1,   0,   3,   3,  -7,  23,   1,   1,   2,  -1,   6,   4,   2,  16,   0,   2,  -1,  -4,   8,   4,   5, -14,   0,   3,  -4,  -1,  13,  13, -17, -17,   5,  -1,  -6,  -2,  30, -20, -41,  28,   5,  -5, -20,  24,  17, -59,  28,  21,   5,  -9, -11,  46, -31,  -8,  31,  -9 },
    },
  },
  { // 33
    {
      { 106,  -7, -26,  -7,  -4,  -2,  -2,   0, -56,  23,  21,   0,   0,   0,   0,   0,  -5, -14,   2,   5,   1,   1,   0,   0,  -1,   3,  -2,   0,   1,   0,   0,   0,  -1,  -2,   1,   1,   1,   0,   0,   0,   0,   1,  -1,  -1,   0,   0,   0,   0,  -1,  -1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0 },
      { -18,  80,  -7, -18,  -4,  -4,   0,   0, -39, -50,  41,  18,   2,   1,   0,   0,  20, -26, -38,  10,   8,   3,   0,   0,   4,  15,   1, -14,  -4,   0,   1,   0,   4,  -3,   4,   1,  -3,  -1,   0,   1,   1,   4,  -1,  -2,   0,   0,  -1,   0,   0,  -1,   1,   0,   0,   1,   0,  -1,   1,   0,  -1,   0,   0,   0,   1,   0 },
      { -35, -53,  -7,  25,   7,   4,   0,  -1, -68,   2,   5, -14,   7,   2,   2,   1,  74,   9,  -9, -14, -14,  -2,   0,   2, -13,  16,   7,   3,   1,  -3,  -1,  -1,  10,  -6,  -1,   4,   1,   2,  -1,  -1,  -3,   2,  -1,   1,   1,  -1,   1,   1,   6,  -1,  -1,   0,   0,  -1,   0,   1,  -1,   2,   1,   0,  -1,   0,  -1,  -1 },
      { -38,  22, -41,  14,   7,   3,  -1,  -1,  -4,  56,  37, -38,  -8,  -4,   0,  -1,  -5, -47,  42,  27,  -5,  -1,   0,   3,   9,  -9, -35,   6,  11,   3,   0,   0,   0,   0,   2,  -5,  -3,   1,   0,  -1,   3,  -3,  -4,   4,   1,  -1,   1,   1,   2,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      {  12,  46, -39,  -6,  -2,   0,  -5,  -5,  15, -14,   5, -14,   0,   1,   1,   1,  31,  52,  40, -25, -21,  -7,   0,   3, -33, -28,  19,  41,  11,  -2,  -2,  -2,   4, -11, -27,   7,  13,   6,   1,  -1,  -6,  -4,   3,  -3,  -1,   1,   4,   2,   6,  -1,  -6,   2,   0,  -2,   1,   2,  -3,   1,   3,  -1,  -1,  -1,  -2,   0 },
      { -10,  38,  50,  -4, -19,  -2,  -2,   0, -50,  29, -51,  15,  24,  -3,   3,  -1,  -5, -24,  37, -36,  -5,   6,  -1,   0,  24, -16,   4,  24, -13,  -3,   0,   1,  -1,   3,  -9,  -2,  10,  -1,  -1,   0,   4,  -3,   3,   0,  -5,   2,  -1,  -1,   0,   0,  -2,   0,   2,  -1,   1,   0,   2,  -2,   1,   0,   0,   1,   1,   0 },
      {  15,  -4,  60,   0,  -3,  -7,  -1,   0,  -4, -28,   7,  -5,   1,   6,   1,   2,  19,  -2,  14,  46, -13,  -5,  -5,  -2, -34, -24, -52,  16,  34,   7,   4,   0,   9,  13, -13, -44, -11,   4,   3,   2,  -2,   4,   8,   6,  -7,  -6,  -2,   0,   1,   2,  -3,  -3,   3,   1,  -1,   0,  -1,   0,   1,   2,   0,   1,   1,  -1 },
      {   5,  20,  27, -16,  -5,  -3,   2,   5,  23,  70,  -2,   9, -12,  -9,  -3,  -4,  45,  -9, -29,  20,  27,   0,  -5,  -7, -48, -14,  34, -16,  -8,  11,   5,   3,   7, -19,  -6,   3,  -5,  -5,   2,   4,  -5,   8,   1,  -8,   1,   3,  -4,  -2,  -2,  -4,   4,   1,  -2,   4,   2,  -1,   0,  -2,  -3,   1,   2,  -1,   1,   1 },
      {  27,  18,   7,  28, -20,  -4,  -2,   0,  41, -11, -28, -35,  24,   1,   0,   0,  34, -38,   1,   1, -29,   8,  -2,   1,  -8,  66,  -2,  13,   2, -15,   0,  -1, -13, -13,  25,   5,  12,   4,  -1,   0,  -4,   0,  -8,   2,   1,   3,   3,   0,  -5,  -3,   4,  -4,  -1,   0,   0,   2,   0,   0,  -2,   1,  -1,   0,   0,   0 },
      {   3,  29, -38,  46,  -1,  -4,  -1,   3, -15,  -2, -62, -12,   9,   7,   3,   0,  -2,  20,  -6,   8,  15, -12,  -2,  -5, -10, -17,  -1, -44,  14,  14,   5,   3,  13,  11,  28, -26, -38,  -5,   1,   3,   0,   1,   9,  17,   4,  -9,  -6,  -2,  -2,   1,   5,  -7,   1,   5,  -1,  -1,   0,  -3,   0,   4,   1,  -1,   1,   1 },
      {   3, -18, -37, -32, -10,  17,  -2,  -4,  21, -15, -16,  38,   7, -20,   1,   0,  29, -41,  19, -34,  21,  19,  -2,   6, -31,  -5, -41,  -7, -25,   7,   9,   1,  18,  31,   2,  -5,  -4, -17,  -6,  -2,  -6,   0,  13,  10,   1,   1,  -2,  -1,   6,   2,  -5,   2,   4,  -3,   0,   2,  -4,   2,   5,  -2,  -3,   1,  -2,  -1 },
      {   0,  -1,   2, -76,  15,   1,   3,   3, -10,  -8, -29, -21, -33,   2,   2,   1,  27,  15,  36,  32,   0, -21, -12,  -5,  26,  21,   7,  -3,   0,  15,  -2,  -1, -11,   9,  29,  17, -18,  -9,   4,   1,   1,  -6,  -8,  13,  18,   2,  -3,   0,  -6,  -1,  -1,  -5,  -3,   5,   4,   0,   1,  -3,  -2,   3,   1,  -3,  -1,   1 },
      { -12,  -9, -42, -21, -15,   4,   6,  11,  -3,   9, -29,  34,  -2,  -9,  -5,  -2,   8,   0, -16,  28,   5,  -2,  -7, -10,  19,  17,   9,  30,  25, -11,   1,  -3, -23, -12, -12, -58,  24,  22,   9,   8,  12,   9,  12, -20, -35,  -4,  -2,   1, -12,   1,  10,   6,   6,  -3,  -4,  -3,   5,  -3,  -2,  -1,  -1,   3,   1,  -1 },
      {  -4, -14, -12, -11,  10, -13,  11,  13,  -7, -28, -15, -50,   8,   9,  -3,  -2, -21, -38,  -6,  23,  -8,  12,  -8, -11, -16, -21,  49,  23, -34,  -4,   7,   7,  31,  23, -25,  -8,   2, -15,  -1,   5,  -2,  23,  16, -11,  -1,   5, -11,  -8, -10,  -2,   9,   5,  -3,   6,   5,  -2,   1,  -2,  -6,   1,   4,  -1,   3,   2 },
      { -18,  -2,  -3, -44, -17,   4,   0,  -5, -20,  14, -10,  -4,  10,  -3,  -3,   1, -49,   7,  -6,  -2, -24,   6,   9,   7, -51,  29,   0,  -7,  27, -25,  -3,  -4,  49, -20,  22,   4,  -3,  20,  -6,  -6,  -9,  -1,  -4,  14,  -8,  -8,  12,   3,  13,  -7,  -3,   2,   8,  -7,  -6,   3,  -5,   3,   3,  -1,  -2,   3,  -2,  -3 },
      {  -8,  -1,  -4,  32, -12,   7,  -5,   0, -20, -18, -18,  28, -21, -12,   4,  -1, -20,   6,  33,  43,  29,  -2, -10,  -4, -30,  21,  -6,   2,  -6,   4,  16,  -1,   7, -21,  -9,  24,  35, -22,  -8,   1,  -3,  -4, -18, -30,  21,  34,   1,  -3,   1,  -3,   1, -11, -20,  -6,  11,   5,   2,   1,  -1,  -3,   4,  -3,  -6,   0 },
      {   4,  14,  -5,   5,  23,  -2,  -3,  -2,   0,  18, -27,   3, -47,  22,   1,   1,   6,  -6, -49,  -2, -20, -19,  10,   5,   6, -11, -24,  23,  12, -30,  -7,  -2,  18,  47,   8,  12,  37,  -6,  -9,  -9, -14, -18,  14,  -9,  16,  18,   7,   3,   9,   3,  -7,  -5,  -7,  -2,   2,   4,  -4,  -2,   5,  -2,   2,   0,  -3,  -1 },
      {  -5,   2,  -1,  -5, -45,  -3,   9,   7,  -1,  17,  24,  -3,  34, -17, -11,  -2,   8,  48,  -6,  27, -11,  35,   1, -12,  14, -11,  -5,   7, -28, -26,  18,   9,   1,  34,  36, -12,  14, -21, -18,   3,  -3, -13,   1,  11,  12,  17,  -1,  -6, -13,  -3,   0, -12,   1,   5,   6,   3,   1,  -7,  -3,   3,   0,   0,   1,   1 },
      {   1, -15, -13,  -6, -58, -10,   0,  -3,  10, -10, -12,  14,  23,  11, -11,   0,  19, -22,   2,  30,  -8, -14,  20,   4,  28, -39,  11, -20,  29,  -7, -19,   3,  12,   4, -23,  42,  -5,  21,  -1,  -8, -17,   6, -21,  -9,  13,  -8,  13,   5,  10,   1,   1,  -9,  -4,  -1,  -7,   3,  -4,   5,   2,   1,   0,  -2,  -1,  -2 },
      {  -2,  -2,   3,  23, -27,  -4,  -3,   2,   6,   8,  25,  35, -11,   5,   1,  -6,  -6,  -5,  17,   0,   1, -31,  -5,   4,   4,  37,  37,  28,   9,  16, -15, -10,  32,  29,  -2, -13, -14,  -9,  20,   3, -13, -10,  21,  25,  24, -14, -22,   0,   1,  -5, -17,  -5,  16,  31,  10,  -8,  -4,  -4,  -2,  -6,  -6,  -4,  11,   9 },
      {   6,   0,  13,  10,   3,  10,  -4,  -3,   3, -10,  11, -22,  -3, -33,   4,   0,   6,  -5,   8,  -3,  45,  -3, -13,   0,  13,   0,  30,   4,   9,  14,  -4,  -3,  27,  34,  21, -11,  11,  39,  -2,   0, -27, -31, -14,   1, -39,  16,  31,   7,   4,   3,  -7,   1,  -8, -33, -10,   8,  -4,   0,   1,   2,   3,   7,  -9,  -7 },
      {   8,   5,   4,   6,  56,  -5,  14,   1,  11,   4,   6,  39,  13,  18,  -3,  -2,  16,   0,  41,  17, -27,   3,  10,  -3,  19,   9,  17, -44, -12, -33,   1,  13,  29,  -7,  -7, -21,   8,  12, -25,  -1, -24,   8,   8,  -1, -11,  13,   2,  -8,   1,  -7,  -9,   2,   0,  -1,   1,   4,  -5,  -2,  -1,  -1,  -2,   3,   3,   2 },
      {  -5,  -9,  -6,  -1,  33, -16,  10,   7,  -2,  -8,   7,  31,  27,   5,  -7,  -4,  -5, -25,   5,  -6, -10, -10,   3,  -2, -34, -31,  22,  22,  14,   0, -15,  -1, -24,   4,  54,  -3,   2,  13,   8,   4,  11, -22, -39,  -1,   9,  11,  -6,   4, -13,  19,  13, -24, -17,  15,   5, -11,   7,  -9,  -6,  11,  -1,  -8,   4,   7 },
      {   3,   0,   8,  21, -19,  54,   6,   4,  -3, -14,   1,   8, -59, -25,   0,   9,   1, -16,  10,   5, -25,  -2,   7,  -6,  -9, -24,  25,  -4, -13, -31,   3,   9, -13,  -5,  15,   0,  -4,   3, -21,  -5,  19,  12,  -9,  30,  -5, -18,   8,  -3,  -8,  11,   7,   0,  26,  -2, -18,  -5,   9,   2,  -5,   0,   1,  12,   9,  -3 },
      {  -6,   6,  -1,   1,  -4,  -7,   2,  -5, -10,   7,  -3,   3,   2,  27, -12,   2,  -6,  12,  14,  21,  -4,   1,  22,   8, -34,  25,   3, -18, -44,   0, -19,  -1, -37,  44, -23,  -2,   3,   8,  21, -14,  -3, -32,  -4, -15, -10, -34,  13,  16,   8,  16,  12,   3,   6, -12, -30,  -8,   1,   0,   2,   2,   9,   9,   7,  -5 },
      {  -3,  -7,   4, -13, -34,  -3,  -5,   4,  -1,  -7,  12, -25,   3,  14,  16,  -5,  -2,  -6,  11, -15,  14, -51,  10,  13, -17,  -6,   6, -40,  -6, -14, -28,   1, -23,  -1,   6, -24,  32,   8, -22, -10,  26,  -4,  32,   4,  -7,  38,  -9,  -5,  -1,   0, -14,  11, -11,   7,  22,  -1,   4,  -2,   1,   0,  -4,  -7,   1,   8 },
      {   2,   5,   1,  -6,  11,  55,  17,  -1,   0,   4,  -3,  -6,  34, -43,  -2,   1,  -7,   1, -16,   8, -44, -28, -30,   4,   3,  -1,   0, -26,   4,  27,  11, -14,  -2,   5, -19,   5,  24,  -7,  10,  11, -16, -21,   9,  -8,   4,   1, -17,   2,  23,  27,   3,  -6,  -8,  10,  -1, -12,   3,  -5,  -1,   4,   0,  -1,  11,   6 },
      {   5,   2,   6,   2,   1, -25,  23,   5,  -1,  -7,  -1, -17, -14, -36, -33,  -2,   8,  10,  12, -14,  30,   2,  13,  -5,   8,   3, -20,  -4,   4, -20, -22,   7,  28, -10,  10,   5,  15,   7,  -2,  -7,  -5,   4,   2, -27,   3, -12, -28,  11,   6,  21,  30,  18,   2,  36, -16, -33,   5,  -5,  -8,   6,   4,  12,  35,  11 },
      {  -4,  -2,   0,  -3,  16, -30,   2, -29,  -8,  -4,  -8,  -7,  -2, -34, -30,  11,  -5, -12,  -2,   2,  11,  24,  22,   4,  -8,   8,  22,  -6,  38,   2,  21,  14, -39,   7, -30,  -1,   6,   1, -31,  -2,  -8, -25,  25,  36,  22,   4,   3,  -8,   6,   6,  -8, -12,   5,  14,   1,  10,   7,  14,   1, -10, -15,  -1,   0,  -3 },
      {  -1,   5,   1,   8,  21,   5,  35,  27,  -4,   2,  -2,   4,  26, -25,   9, -11,  -4,   5,   1,  10,  16,  -9, -29, -13, -15,   3,  -8,   8,  12, -17, -22,  12, -24,  25,   3,  38,  -5,  17,   2, -22,  17,  33,  27,  -6,   8,  -7,  17,   3,  -3, -35, -33,  -1,  18,   6, -12,  10, -10, -10,   2,  -8,  -4,   8,   7,  -8 },
      {  -2,  -3,  -2,   6, -19, -34,  27,  20,  -2,  -3,   5,   9, -23,  26, -45, -23,  -2,  -4,   7,  -8, -14,  -4, -32,  -8,  -3,  -4,  -6,  -5,  -6,  21,  36,   6,  -7, -12,  18,  24,  10,  16,   6,   5, -10,   6,  32,  16, -27,  -1,   4,  -5,  18,  34, -14,  -2,   0, -19,   7,   3,   2, -21, -11,   2,   9,   3,  -5,   0 },
      {   0,  -5,  -5,  -3,   0,   6,  18,  18,   0,  -9,  -8, -11,   5,  16,  21,  -5,   0,  -8, -11,  13,  28,   4,  -3,  -2,   6, -16,  -4,  18, -24, -25,  -7,   2, -10, -50,   2, -13,  -3,  -9,   0,   0, -34, -45,  -3,  16,   6, -11, -13,   9,  35,   5, -46, -12,  13,   9, -14,  -8, -18,  -2,  19,   0,  -6,   7,   5,  -3 },
      {   4,   2,  -5,   1,  16, -25,  -5, -18,   6,   3,  -9,   2,  17, -14,  -3,  -6,   6,   7, -13,  21,   8, -25,  -9,  26,   8,  -1, -14,  10, -38,  10, -11, -10,  35, -17, -14,   1,  28,  13,  -4, -12,  45,   6, -14,  45,   5,  -8,  16,   7, -12,  22,  -2, -26,  24,  -3,  -3,   8,  23,  -1,  -5,  -7,  -7,  10,  -8,  -2 },
      {  -3,  -6,  -3,   5,   3, -35, -45, -25,  -2,  -6,  -5,  10,  -1, -16,  26,  15,   1,  -5,   3,  21, -22,  15, -17,   4,  -2, -24,   6, -13,  -5,   9,  -2, -21,  -7,  -4,  33,  23,  29,   0,  15,  10,  -5, -10,  19,  -3, -23,  -9, -18,   3,  -6, -18, -12,  39,  25,  12,  -6,  -9, -16,  12,  -1, -18,   1,  14,  11,  -1 },
      {   4,   4,   2,   5,  13,  51,  -2,  -5,  -1,  -2, -11,  -2,  13,  25, -39,  -1,   5,   4,  -3,  24,  16,  28,  35,   9,   1,  -2,  -1,  14,   3,  12,  -8,  -3,   8,  -1,   7,  23,  -4,  13,  10,  -8,  25, -11,   9,   8, -23,  12,  -2,  -1,   2,   5, -23,  30,   8,   9,  40, -12,   5,  -1,  -4, -21,  -2, -19,  13,  27 },
      {  -3,   3,  -4,   3,  -5, -10,  -4,  13,  -7,   1,  -9,  -1, -23, -18,   0,   1,  -7,   6,  -9,  -3, -16,  22, -19,   0,  -6,  -1, -16,  -5, -27,  12, -41, -17, -20,   0, -16, -12,  -5,  37,   1, -15, -53,  21, -24,  17,   2,  15,  12, -17,  -5,  -6,   7, -10,   8,  12,  43,  -4,  -3,  -1,  -1,  -3, -10, -18,  11,  25 },
      {   1,  -2,   0,   2,  -6,   4,  16,  20,   0,  -4,   0,   1, -11,  -4, -31,  17,   1,  -5,   0,  -8, -20,   8,   1, -15,   2,  -9,  -6, -10, -11,  21, -15,  -4,  15, -23,   3, -11,   6,   4,  25, -14,  -1, -31,  25,   0,  15,   4,  12,  25, -43, -41, -16,  -3, -32,   2, -20,   1,  36,  32,   0,  -3,  -5,  -9,  -5, -19 },
      {  -3,  -1,   3,   0,   9,  22, -31, -19,  -3,  -1,   3,   2,  17,  -6, -23,  -9,  -5,   2,  -1,  11,  11, -32,  33,  16,   1,  -7,  -2,  27, -24,  13,   3,   9, -16, -18,  33,  -4,   3,   3,  -3,  -8, -49,  42,  42,  -2,  10,  -7,  13,   9,   5,  -5,  20, -10, -10,   1, -13,   5, -10,   5,  -8,   5,  -1,   4,  -8,  -8 },
      {   2,   1,   1,   3,  -7,  -3,   9, -53,   2,   1,  -1,   1, -16,  -8, -22,  12,   2,  -2,   2,   2, -12,  22, -21,   7,   6,  -2,  13,  -3, -13,   0, -12, -12,   3,  -1,  12, -27, -10,   9,  10, -12,  37,  22,   4, -24,  17,   7,  13,  19,  35,   7, -30, -12, -27,   8,  -3,  -1, -45,  -8,  20,  16,  -6, -16,   0,  -5 },
      {   0,   0,   4,  -6,   5,  20, -62,  29,   0,  -2,   3, -11,  -4,  20,   3, -35,   2,   1,   9,  -7,  12,  38, -15,   0,   2,  -1,   7, -22,  16, -12,  12,  -3,   3,   7,  -6,   0,  28,  19,  19,   1,   7,   9,   9,   1,  11, -12,   0,  15,  -3,  10,  -4, -42,  -5,  27, -13,  -3,   9, -16,   0,  11, -25,   1,  11,  -7 },
      {   0,  -1,  -3,   2,   2, -10, -45,  36,   3,   2,  -2,  13,   3, -21,   6, -17,   0,   0,  -5,  13, -21,  10, -13,  11,   5,   5,   4,  16,  -4,   5, -24,  19,   4, -10,  -4,   0, -45,  -1, -24, -33,  17, -24,  21,  -6,  10,  24,  10,  -9,  27,  26,   7,  29, -19,  -8,  -6,  -2,  14,  -2,  -4,   8,  16,   9,   7, -16 },
      {  -3,   0,  -4,  -2,  -5,  14, -10, -34,   0,   8,   3,   2,  13,  41,  -6,  43,  -3,   3,  -3,  -1,  17,  -5, -51,  -9,  -4,   6,  -5,  12,   0,  -9,   4,   0,  -4,   6,  -4,  10, -13,  15, -33,   0, -17,  14,  -2,   4,  -2,  14,  -4,  -6, -22,  21, -13,   3, -10,  11, -18, -26,  31,  19,   4,   5,   5,  10,  35, -16 },
      {   0,   2,   0,   0,  -2, -13,   5,  -6,  -2,   1,  -2,  -3,  -9, -22,   8, -33,  -1,   4,   0,   2, -12,   5,  18,  42,  -1,   2,   6,  -2,   5,  -1,  -3,   5,  -6,  11,   8,   0, -14, -21,   0,   5, -14,  29, -10, -25, -25,   1, -19,   8,  -8,  12, -54, -21,   0, -17,  -2, -24,  54,  18,  26,  -3,   1,  -8,   0,   3 },
      {   2,  -1,   3,   2,  -1,   4,  30, -21,   2,  -2,   3,   4,   0,   7,  36,  12,   3,   0,   9,   7,  14,  29, -14,  19,   3,  -7,   8,  -1,  11, -30, -19, -33,  -2,  -7,  10,   7,  -6,  -8,  27,  -1,  -8,  -2,  44,   0,   9,  -4,  13,   2,  19,  18,  34,  -3,  -3,  -7,  10,  17,  49,  -3, -23,   5,  -3,  -7, -19,  12 },
      {   2,   0,   2,  -2,   2,   0, -11,  12,   2,  -2,   1,  -5,   4,  10, -41,  -6,   1,  -2,   2,  -6,   5, -15, -24,   8,   1,  -7,   6,  -4,  -4,  -9,  15, -19,  -7,  10,  13, -24,   3,  -4,  -3,   1,   3,   5, -32, -22,  23,  -4,   6, -11,  57, -32,   2,  22,  26,   4,   4,  19,  39,  30,  -6, -25, -17,  13,  -4,   0 },
      {   1,  -2,  -1,   0,   1, -14, -23,  -1,   3,   0,   2,   2,   0, -16,  15,  36,   0,  -1,   0,   2,  -4, -21,  21, -54,   0,   4,  -7,   6, -20,  -6,  20,  18,   7,   1,   3,  13, -11,  23,   6,  12,  15,  -9,  18, -17, -16, -13,  12, -10,  35, -12,  -4, -33,   2,   4,  17,  -8,  33,   1,  21,  18,  -7, -18,   7,  31 },
      {   0,  -2,  -1,  -1, -10,   5,  14, -31,  -1,  -2,  -2,  -3, -10,   4,  20, -25,  -1,  -1,  -4,  -3,  -8,  -4, -10,  26,  -1,  -3,  -7,   4, -14,  11,  22,  36,   7,  -7,   5,  -2, -18,  31, -19,  -9,  -2, -34,   4, -50,   3, -18,   6, -16, -26,  -4,  -2,  -3,   9,  30,   4,  24,   4, -17, -29, -20, -31,  -7,   5,   1 },
      {   1,   0,  -2,   0,  -5,   2,  30,  -2,   2,   2,   0,   4,  -5,  24,  37, -11,   0,   2,  -2,   4,  -4,  25,  10,   3,   4,   1,   6,   7,   8,  59,  -4,  22,   3,   4,  13,  -3,  36,  -2, -27, -22,   7,  14,  -5,  19,  -9,  -7,   4,  -2,  41, -28,  19,  -2, -15,  12, -14,  -9,  13,  14,   9,  12,  -1,   8,  28,   3 },
      {   0,   1,   0,   4,   0,  -2,   4, -13,  -1,   0,  -4,   5,  -3, -10,  12, -37,   1,   2,  -3,   7,   4,  -7,  -5,  15,  -1,  -1,  -7,   6, -17, -31,  30,   2,  -5,   9, -12,   7,  -8,  48,  11,  28,   6,  -2,  -5,  35,  13,  -3, -26,  10,  12, -30,   7,  26, -46,  -4,  -3, -20,   2,  19, -13,   6,  19, -26,   1,  -3 },
      {   1,   2,   1,   4,   1,   0,   3,  29,  -1,   1,  -2,   3,  -5,   0, -20,  28,   1,   4,   2,   8,   3,   2,  -5,  48,  -1,   2,  -4,   6,  -7,   0, -17, -25,   1,   5,  -2,  20,  -8,   0, -34,  30,  11, -10,   9,  -5, -41, -11, -21, -26,   8, -24,   9, -26, -10,  22,  -7,   5, -14,   8,  -5,  35, -21, -11, -22, -21 },
      {   0,   4,   0,   0,  -1,   1,   1,  14,   0,   4,  -1,   0,  -2,  -3,  17,  26,  -1,   7,   0,   1,  -2,  10,  18, -25,  -2,   6,  -5,   1,  -7,  13, -25,  27,   2,   9,   7,  -2,  11,  17,   6,  28,  -1,  25, -10,   2,   6,  -3, -26,   0,  20,  44, -23,   9, -16,  11,  -7,  26,   4,  18, -35, -38, -16, -14, -26, -38 },
      {  -1,  -2,  -4,  -3,  -1,  -2, -12, -12,   0,  -2,  -4,  -2,  -2,  -4,  -6,  16,  -1,  -2,  -4,  -4,  -3,   2,  11,  -2,   0,  -1,  -2,  -2,   2,   6,  -9, -10,  -4,  -2,  -9,  -8,   2, -13, -13,   7,  -6,   3, -15,   4,  -2,  -2,   3,  12,  14, -35, -33,  -2,  -8,  -4,  -9, -11,  24, -74, -69,   9,  24,   7,   7,  -2 },
      {  -1,   0,  -2,   2,  -2,   7,  -1, -42,  -1,  -2,  -2,   0,  -4,   3, -16, -47,  -3,   1,  -3,   1,  -6,  -2,  -6, -33,  -3,   1,  -8,   0,  -4,  -4, -40,  25,  -1,  -2,  -9,  10, -12, -19,  23,  35,  -1,  -8,  -3,  13, -28,  46,   2,  -5,  10, -14,  14,  -3,  19,  17, -15,  11,  11,   1,   1,  21, -21,  13,  -4, -18 },
      {   0,  -1,   0,  -1,   2,   7, -16, -17,   0,  -1,   1,  -1,   2,  -2, -18,   9,   0,   0,   0,  -1,   6,  -1, -36,   7,   2,  -2,   2,  -4,   9, -19, -22,  61,   1,  -2,   1,  -2,  17, -23,  22, -39,  -4,  -7,  -8,  10, -19, -39, -29, -21,   3, -13,   2, -17, -19, -12,  27,   8,  -3,  13, -10,   2,  26,  -3, -17,   7 },
      {   0,  -1,   0,  -1,  -2,   7,   3,  -1,   1,  -1,   0,  -2,  -2,   4,  -6,  20,   2,   0,   0,   0,  -2,  14,   5,  22,   0,  -2,   0,  -1,  -2,  12, -13,  21,   2,  -2,   4,  -1,   0,  37,  18,  21,   8,  -2,  14,  -6,  38,  21, -42, -33,   2, -16,  10, -17,  13, -51,  -9, -16,   4, -28,  23, -14,  -6,  42,   0,   3 },
      {   0,   0,  -1,  -3,  -3,  -7,   1,  11,   1,   2,   1,   0,  -2,   0,   6,  14,   1,   2,   0,  -3,  -4,  -5,  -4,  15,   2,   0,   0,   0,   0, -10,  12,   8,  -2,   1,  -6,   2, -21, -14,  17,  -9,   8,   1,   4,   3, -27,  37,   5,  26,  14,  -7,  15, -45,  -6, -12, -11, -24, -28,  31, -48, -48, -22,  12,  27,  22 },
      {  -1,   0,  -1,  -2,   1,   0,  -3,  13,  -1,   1,  -2,  -1,   1,   4,   0,   9,  -1,   1,  -1,  -3,   1,   7,  -6,  26,  -1,   1,  -2,  -2,   5,   7,  -9,  35,   1,  -2,   1,  -7,  12,   1,   9,  45,   3,   2,   3,  -7,  26,   6,  37,  -9,  -6,  10, -23,  10,  25,   2, -21,   2, -12,  37, -29,  44,  27,  -7,   3,  50 },
      {  -1,   0,   0,   3,  -1,  -4,   1,  12,  -1,   1,  -1,   3,   0,   1,   3,  38,  -2,   0,  -2,   4,  -1,  -5,  -8,  30,   0,   1,   2,   0,   0,  -2,  25,  41,  -3,   5,  -4,   2,   0,   0,  21, -23, -11,   9, -15,  13,  -6,  29, -10,  48,   4,  -3,   9,  35,  18,  14,  -8,  -2,  12, -29,  17,  27, -41, -17, -26,   4 },
      {   0,   1,  -2,  -1,   0,   1,   7,   1,   0,   1,  -2,   0,  -3,  -5,   8,  10,   0,   1,  -1,  -1,   0,  -2,  26,   6,   0,   0,  -1,   0,  -4,  -2,  27, -17,  -2,   2,  -3,   1,  -7,  16,  36, -33,   0,   5,  -9,  -1,  -7,  22, -24, -13,  -1,   5, -20,  -5,   0,  16,  17,  50,   0,  21, -18,  47,  16,  48,  31, -17 },
      {   0,  -1,  -1,   0,   0,  -8,  -1,   5,   0,  -1,   0,   0,   1,  -7,  -3,  10,   0,   1,   0,   2,   3, -11,   7,  17,   1,  -2,  -1,   0,   1, -12,   3,   3,   2,  -4,   2,  -1,   8, -24,  36,   7,  -2,   7,  -9,  16,  -7,  -7,  31, -60,  -3,   4,  -6,  21, -39,  -6, -28,  17,   2, -24,  15, -13, -35, -22,  51,   0 },
      {   0,  -1,   0,  -2,  -1,   1,   5,   6,   1,  -1,   1,  -2,  -2,   0,   4,  14,   1,  -1,   2,  -4,   0,   1,   4,  27,   1,  -1,   0,  -5,   0,  -3,  12,  17,  -1,  -2,   0,  -5,   0,   4,  26,  15,   0,  -1,   0,  -1,  -6,  18,  36, -10,   9,  -7,  12, -16,  16,  49,  14, -18,   5, -13,  23, -33,  66,  -5,  -3, -48 },
      {  -1,   0,  -1,   1,  -3,   2,  -2,   5,  -1,   1,  -1,   2,  -3,  -1,  -3,   5,  -1,   0,  -1,   1,  -5,   1,  -6,   8,  -2,   2,  -1,   2,  -4,  -3,  -8,   6,  -3,   2,  -3,   3,  -9,   0, -10,  30,  -1,   4,  -3,   9, -12, -10,   9,  28,  -5,   1,  -3,   6, -50,  33,   3,  36,   7, -13,  29, -26,   9,  64, -14,  50 },
      {   0,   0,   1,   0,   0,   3,  -6,  -2,   0,   0,   1,  -1,  -1,   3,  -6,  -1,  -1,   0,   0,  -2,   1,   3,  -4,  -4,   0,   0,   0,   1,  -1,   4,  -1, -13,   0,  -1,   0,   1,  -4,   9,  -7, -18,  -1,   1,   0,   1, -10,  24, -38,  -2,   2,  -4,   7, -15,  16,   7, -61,  53,   4, -10,  13, -22,  43, -49,   4,  35 },
      {   0,  -1,   0,   1,   1,  -2,  -3,   6,   0,  -1,   0,   1,   0,  -1,  -5,   6,  -1,  -1,   0,   1,   0,  -1,  -8,  14,  -1,  -2,   1,   1,   0,  -1, -14,  16,   0,  -3,   2,   0,   1,  -6, -14,  37,   0,  -4,   5,  -3,   4, -13,  -6,  54,   1,  -4,  13,  -5,  12, -32,  38,  46,   2,  -4,  11,   2,   5, -22,  68, -25 },
    },
    {
      {-107,  37,  15,   2,   3,   1,   1,   0,  32, -38,   4,   2,   1,   1,   0,   0,  23,  -2, -14,   1,  -1,   0,  -1,   0,   0,   6,   0,  -4,   0,  -1,   0,   0,   4,  -1,   0,   0,  -1,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0 },
      {   5, -41,  19,   1,   3,   0,   1,   0,  81,  -9, -41,   6,  -5,   1,  -2,   0, -37,  59,  -3, -14,   1,  -3,   0,  -1, -13,  -8,  25,  -1,  -2,   0,   0,   0,  -3,  -1,  -2,   6,   0,   0,   0,   0,  -3,  -1,   2,   0,   2,   0,   0,   0,  -1,   0,  -1,   1,   0,   1,   0,   0,  -1,  -1,   1,   0,   0,   0,   0,   0 },
      {  31, -25,   6,  -3,   2,  -1,   0,   0,  -8, -58,  15,  18,   0,   4,   0,   2,  32,  13, -69,  12,   3,   2,   0,   1, -19,  49,  -2, -33,   6,  -2,   1,  -1,   0,  -4,  22,  -5,  -9,   1,  -1,   0,  -2,   4,   1,   4,  -2,  -3,   0,   0,  -1,   0,   3,   1,   1,   0,  -1,   0,  -1,   2,   0,   0,   0,   0,   0,   0 },
      {  39,  86, -47, -15,  -5,  -4,  -2,  -2,  41, -19,  29,  -9,  -4,  -2,  -1,   0, -16,  11,   1,   8,  -1,   0,   0,   0, -22,   1,  20,  -1,   2,  -1,   0,   0,   0, -17,   2,  11,   1,   1,   0,   0,  -4,   1,  -4,   0,   4,   0,   1,   0,   0,  -4,   0,   0,   0,   2,   1,   1,  -2,   1,  -1,   0,   0,   0,   0,   1 },
      {  -8,  16,  -2,  -6,   1,   0,   0,   0, -53,   5, -19,  16,   2,   3,   2,   1,  19,  36, -13, -46,  10,  -4,   2,  -1,   7, -15,  69, -12, -25,   2,  -3,   1,   8, -20,  -6,  42,  -5,  -5,  -1,   0,  -1,   2, -10,  -2,  12,   0,   1,   0,   2,  -4,  -2,   0,  -1,   2,   0,   1,   1,   1,  -1,   0,   1,  -1,   0,   0 },
      { -21,   3, -26,  14,   2,   2,   1,   1,   2,  49, -12, -15,  -1,  -1,  -1,  -1, -32,  -3, -14,   8,  14,  -2,   4,  -1,  11,  12,   2, -62,  11,   4,   1,   1,   9, -13,  55,  -4, -41,   5,  -2,   1,   5,  -6,  -6,  34,  -4, -13,   1,  -1,   1,   0,   2,   1,   9,  -1,  -3,   0,   1,  -3,  -2,   3,   0,   2,   0,   0 },
      {  -1,   9,  26, -16,   2,  -4,   0,  -1, -23, -32,  27,  31,  -8,   3,  -2,   1, -76,   2, -23,  18,  23,   0,   4,  -1,  37, -41,  -9, -21,   7,   8,   0,   2,  13,  13, -23,  -1, -11,   2,   0,   1,   0,   8,   2, -11,   1,  -3,   1,  -1,   4,   0,   1,   0,  -4,   0,   0,   0,   1,   2,   0,  -1,   0,  -1,   0,   0 },
      {   0,  29,  -2,   5,  -7,   1,  -2,  -1, -30,  -4, -34,  14,  14,   2,   2,   1,   3,  42,   0, -12,  -4,  -3,   0,  -1, -24, -18,  -8,   7,  38,  -4,   3,  -1,  13,  -9,   0, -66,   3,  23,  -1,   3,   6,  -2,  28,   6, -45,   1,   4,   1,   2,  -1,  -4,  15,   3, -14,   1,   1,   0,   0,   4,   0,   1,   0,  -3,   1 },
      {  21,  25, -14,  -2,  -5,   1,  -2,   0,  17, -28, -50,  15,  13,   1,   4,   0,   9, -32,   2, -46,  13,   6,   3,   1,  42,   5, -17, -11, -31,   7,  -2,   3, -13,  49,   7,  -8, -15, -15,   2,  -2,  -2,  -3,  30,  11,  -1,  -9,  -5,   0,  -5,   2,  -1,  10,   7,   2,  -3,  -1,  -2,  -1,   3,   0,   1,   2,   0,   0 },
      {  12,  -7,  29,  -7,  -5,  -2,   2,  -1,  -2, -32,  17, -28,   8,   3,   2,   1,  20,   6,  32,   0, -23,   1,  -2,   0, -10, -32,   1,   4,   2,  10,  -1,   3,  15,   3,  -5,   6, -56,   6,   4,   1,  -8,  20, -11,  50,   1, -40,   5,  -3,   1,   2,  -7,  -6,  35,  -3, -10,   2,  -2,   2,  -1,  -1,   0,   8,  -2,  -1 },
      {  18,  40,  85, -41,  -9,  -8,  -2,  -3,  13,  45,  -8,  28, -14,  -1,  -3,  -1,  19,  -8, -10,  12,   9,  -2,   0,   0,  -5,   8,   0,  -1,   4,   1,   0,   0, -15,  -3,  16,   1,   3,  -2,   0,   0,   3, -17,   1,  15,   5,   3,  -3,   1,  -4,   0, -12,   0,   8,   5,   3,  -1,  -1,  -4,   0,  -4,  -1,   2,   3,   1 },
      {  -4,  18,   5,  18,  -8,  -4,  -1,  -2, -31, -12, -21, -12,   1,   5,   2,   2, -19,  39,   8,  20, -24, -12,  -1,  -1, -10,  14, -24,  11,   6, -10,   2,  -1,   8,  22,  19,   2,   6, -36,  -1,  -3,   5, -14,  14,   7,  59,   0, -24,   0,   5,  -7,  -2, -22,   1,  44,  -2,  -4,   0,  -3,   0,   3, -12,   1,  14,  -1 },
      {  -2,  -1,  20, -14,  -2,   2,   0,   1, -16, -32,  13, -16,  15,  -2,   4,   0, -45,   0,  23, -23, -24,  13,  -3,   3,  23,  40,   7,  22, -20, -21,   1,  -4,   0, -18,  41,   2,  24,   6, -10,   1,   7, -19, -16,  27, -24,  24,   8,  -2,  -4,   8, -19,   5,   5, -19,  12,   1,   4,  -3,   0,  -5,   4,  -2,  -6,   3 },
      {  -9, -11, -13, -35,   7,  -2,   4,  -1,   2,  24,  51, -24, -19,  -2,  -5,   0,  15,  49, -10,  -2, -15,  -3,   0,  -1,  38,   6,   6,   6,  -9,  -6,  -4,  -1, -13,  33,   3,  -8,   0,  -5,   3,  -2,   2, -10,  48,   1, -10, -10,  -6,   2,   0, -11,   0,  37,   5,   3,  -7,  -2,  -5,  -1,  -8,   2,  10,   4,   4,  -2 },
      {   5, -10,   2,  -4,  16,  -4,   0,   0,   4, -15,  15, -10,  -4,   3,   1,  -1,  24,   2,  26,   0,  28,  -6,  -6,  -2, -17, -35,  -1, -24,  -4,  24,   2,   2,  26,  12,  22,  -4,  -6, -15, -13,   1,  -5, -14,  12,  16,   2,  56, -12,  -6,  -6,   4, -27,  12, -37,   6,  41,  -4,  -3,   0,  -2,  -1,   6, -26,   5,  10 },
      { -15, -11,  -3, -15,  -1,   5,   2,   1,  -8,  10,  16,  17,  -4,  -7,   2,  -1, -31, -22,  -4, -23,  10,  15,   1,   4, -77,   8,  -5,   5, -43,   4,   9,   0,  20,  -3,  -8,  -6,  -3, -34,   7,   0,  17,   3,  28,   4,  -4, -16, -17,   1,   6,   0,  -4,  24,  17,   2,  -8,  -5,   0,   0,  -3,  -8,   8,   8,   1,  -1 },
      {  -3, -15, -28,  27, -24,   3,  -1,   3,  13,   3,  23,  62,  -7, -18,  -5,  -3,  12,  -3,  -2,   3,  11,  -5,   1,  -2,  21, -14,  11,  21,  15, -19,  -2,  -2,   9,  -8,   5,  -3,  19,   1, -11,   2,  -3,  -1,   0,  40,   8,   5, -13,  -5, -14,   0, -31,   0,  42,  19,   7,  -7,  -2,  -1,  -2, -27,  -4,  12,  10,   4 },
      {   8,  -3,  19,  -7,  -1,  -6,   1,  -1,  14,   2,   1, -27,  31,  -2,  -6,   1,  -2, -31, -14, -17, -16,   4,   3,   2,  18,  -3,  26,  -6,  15, -20, -10,   0,  31, -30,  -6, -19,   5,   6, -11,   5,  12,  32,  14,  -7,  -2,   6, -37,  -3,  -4,   3,  18,  18,  -8,  51,  -1, -19,  -3,   2,  -9,  12, -14,  -1,  30,   0 },
      {   9,  16,  18,  33, -18,   5,  -4,  -1,  10,  26,   6, -13,   6,  -8,   7,  -2,   6,  14, -31,  -3, -12,  18,  -2,   2,   7,  -6, -19,  -1, -25, -23,  13,  -1,  61,  23,  -2,   8,  15, -13, -20,   1, -33,  33,  -1,   2,  -1,  17,  15,  -9,   1, -12,  13,  -5,   6, -29,   9,   7,  -5,  -2,   3,   2,  11,   5, -16,   3 },
      {   7,  12,  34,  79, -23,  -4,  -7,  -1,   5,   4,  45,  -4,  33, -11,   0,  -2,  -3,  17,   9, -25,  25,   6,   0,   0,  -1,   8,  -1,  -5,  -6,  23,  -3,   2, -36,  -6,   0,  -3,  -8,   2,  13,   0,   7, -13,   5, -11,  -4,  -7,  -3,   5,   3,  -5,   2,  11, -11,   7,  -4,  -2,  -2,  -1,  -8,   5,   3,  -2,   4,  -1 },
      {  -8,  -3, -15, -35,   7,   0,   5,   1,   3,  23,  16,  10,  41,   3,   0,  -7,  -3,  26, -14, -26,   6,  25,  -4,   3,   5,  17, -28,  12,  -6,  26,   2,  -3, -11,  -5, -18, -17, -15,   4,  14,   2,  -1,  26, -16,   9,   9,   4,   2, -10,  -8,  -2,   1, -39,   3,  13,  39,   3,   3, -12,  15,  -8, -30, -21,   1,  20 },
      {  -6,  -9,   3, -17,   1,  -1,   2,  -1,  -2,   0,  14,   5,  27, -18,  -2,   0,  -8, -14,  19, -16,  19,   0,   0,  -1, -25,  17,  23, -10,  23, -10, -16,   3,  19,  45,   7,   5,   9,  41, -11,  -8, -34, -20,  17,  -7,  10,   3,  30,   1,  14, -35,   0, -20,   1,  14, -16,  16,  -6,   4,  -7,   4, -28,  15,   3,  -6 },
      {   3,  -5,  -6, -19,   9,   4,   2,   0,   7,  -2,  13, -11,  23, -16,   4,   3,  16,  13, -17, -18,  11,  24,   2,   1,   2, -53, -28,  -9, -15, -18,  13,   6,  -5, -14,  24, -13,  19, -13, -12,   3,   7, -37, -25,  -9,   2,  -3,  -5,  14,   0,  19,  -5, -22,  -3,   3, -43,   1,   1,  10,  11,   8, -15,  27,  -2, -23 },
      {  -2,   0,  -2,  17,   8,  -4,   5,  -1,  -3,   8,  -4, -19,  20,  22,  -3,   2, -13, -27, -56,   5, -39,  14,  12,   2, -23, -40,  25,  11,   1, -16,   2,   0, -45,  26,  -3,   4,  -1,   8,  -6,  -2,   4, -20,  11,  19,   4,   6,  10,  -8,  -2, -10, -24,  -1,   0,  -2,  20,   7,   0,   6,  -2, -12,  -3, -18,  -4,   7 },
      {   4,  -2,  -2,   6,  -6,   2,   0,   2,  10,   7,  -2,  -1,   0,  27,  -1,   2,   1,  -9, -18,  -2, -13,   6,   4,   1,  21,  14,   2,  20,   2,  27,   2,  -6,  58, -21, -14,  -4, -12,  -1,  40,   2, -16, -52,   6,  -5,   4, -19,   2,  23,  -6,  -9, -40,  -2, -20,   1, -13,   8,  -3,  17,   0,  -8,  -7, -18,  -6,  -8 },
      {   5,  14,   5,  20,  88, -19,   5,  -9,  -6,   4, -13, -16,  13,   1,  -5,   0,  -3,   8,  -8,  12,  22,   7, -13,  -1,   2,  14,  -6,   6,  -4,  11,  -5,   2,   4,   3, -15,   7,  11,   9,   1,  -1, -25,   1,  -6,   2,  -9,   3,  -8,   2,   0,   9, -21,  13,  29,   5,  -7, -18,  -3,  -8,   2, -30,  10,  24,  21,  -2 },
      {  -1,  -3,   2,   1,  29,  -3,   2,   2,   1,  -4,   1,  27,  11,  15,   8,   1,   2,  -4,  27,  11, -16,  22,   1,  -1,  11,  -2,   1, -16,  -1, -12,  32,   2, -27, -43,   3, -11,  -6, -29,  -9,  16, -42,  11,  25,  -3,   9,   1,  -7,  -4,  -5, -56,   2,   5,  -4,  -5,  -5,  15,  -1,  10, -17,   7,  -8,   1, -23,  -7 },
      {   5,  -1,   8,   2,   5,   2,  -2,   5,   2,  -8,  -3, -45,  -9, -16,  21,   0,   0,  -9, -13,  -2,  36, -26,  -6,  -1,  12,   5,  17,   3,  10,  14,  11,  -3,   3, -14, -18, -20,  14, -18,  -5,   5,  48,   3,  20,   8,  18,  10,  11,  -5, -15, -23,  -6, -16,  23,  -7,   2,  34,   3,   1,  -1, -20, -18,  20, -37,   2 },
      {  -2,  -4,  -6, -14, -31,  25,   2,   2,  -4,  -8, -27,  -8,  41, -24,   4,  -3,   7,   8,   3,  67,  17,  42, -11,  -8,   9,  -4,  15,   2, -26,  19,  11,  -2,   2,  -1,  -2,   7,   5,   3,   5,  -1,  15, -11,  -2,  -3,  -8,  -1,  10,  -3,  17, -16,   2,  22,   7,   7,   6,   0, -12,  -5, -20,  -9,   2,   4,  18,   9 },
      {   9,   5,   2,  17,  42, -12,   8,  -3,  11,   4,  10,  38,   8,  19,  -6,  -2,  13,   6,  22,   5, -26,  21,   8,  -5,   5,   0,  -4, -15,  -8, -19,  -1,   3,  28,   9,   3,   9,  -4,  12,   0, -13,  79,  -2,   4,  -7,   7,  -4,  29,   0,   2,   0,   7,  -1,  -3,   2,   1,  20,  -5,  -4, -11,   4,  -7,  -3,  -2,   8 },
      {  -1,  -2,  -4, -12,  -9,  -7,   5,   1,  -1,  -3, -13,  -9,   3,  -5,   3,  -1,  -1,   4, -15, -23,   8, -16,   2,  -1,  -3, -16, -65,   9,   8,  -6, -28,   3,  -4, -39,  10,  45,   0,  25,  -2, -19,   7,  11,  11,  -2,  13,   5,  17,   0,   7, -35, -15,  29,   1,   6,   4,   5,  -4,  13, -28, -10,  15,  -3,   6,   2 },
      {   2,   7,   7,  14,  -3,  57,  -7,   5,  -6,  -4,  -2, -17, -63,  23,  -1,   5,  -1,   5,   5, -14,   6,  38,   3,  -8,  -3,   4, -11, -13, -11,  -5,  13,   4,  -9,  -5, -14,  -3,  -2,  19, -14,  -5,  -2,   3,   1,   1, -13,   7,  17,  -2, -18,  -9,  -9,  -6,   3,  22,   1,  -3,   1,   5,  -4,  -7, -38,   4,  38,  -3 },
      {  -5,   2,  -5,  -9, -23, -11,   2,  -1,  -2,   9,  -5,  -1,  25,  -3,  -4,   6,  -3,  11,  12,   7,  -8,   7,  -7,   2,  -5,  28,  -8, -35,  12, -12,   0,  12,  -6,  12, -54,  14, -11,  -3, -34,   5,  15,   4,  -5,   8,   0,  15,  -6,  -4, -44,   9, -40,   6, -14,  -7, -20,  -5,   3,  23,   6,  -8,  16,  -1,  -4, -37 },
      {  -4,   0,   0,   1,   5,   2,  -9,  -4,  -6,  -1,  -1,   7,   8,  -2, -15,   8,  -6,  -1,  -6,  12,   3,   8, -17,   3,  -7, -13,   9,  28,  -4,  25,  -2,  -8,   2,   5,  39,   8,   4,  11,  17, -13,  -2,  29,  20,   2,   7,   4,  15,  18, -88,   2,  17,  -1, -11,   0, -17,   5,  14,  -7,  -4,  10,   4,   3,   3, -21 },
      {  -1,  -2,   5,  -8, -14,  -1,  15,   1,  -2,  -7,   2, -16,  15,  21,  -1,   1,   5,   3,  19,  -2,  13,   1,  48,  -4,  -6,  -6,  -5, -36,  12, -29,   5,   1,   4,   2,  -4,   8,  26,  -7,  26,   0, -13, -19,  -2, -19,   3, -18,  -2,  16, -47,   2,  18,   7,  26,  -4,  15,  -5,   5, -17,   4, -32,  11,  -9,  -1,  38 },
      {  -2,  -3,  -5,   2,  -6, -46,  -8,  -6,  -3,  -5,  -6,  -3, -17, -13, -27,  10,   0,  -2,  -6,  -9,   0,  29,  -5,   3,   1,   2, -10,   8,  12,  17,  21,  -9,   7, -16, -23,  19,  -9,   8, -24,   3,  -5, -54,  18,  17,   1,   8,   1, -26, -13,  12,  54,   1,  14,  -1,  16,  -2,   1,   6,  -6,  18,  -3,  12,   0,  10 },
      {  -2,  -4,  -1,   2, -27, -50,  16, -16,  -3,  -4,  -4, -14, -16,   0, -40,  -8,   2,  -3,   9,   2,   3,  24, -21,   3,   3,  -3,   6, -17,  -8, -10,  17,  -3,  -8, -15,  13,  -5,   8,   5,  16, -11, -10,  32,  35, -19,   9,  -3,  20,   9,  26,  36, -35, -15,   0,   1,  -2,  14,   5,  -3,   7, -23, -11,  -1,  -1,   2 },
      {  -2,  -2,  -4,   1, -15, -59,   9,   2,  -3,  -3,  -5,  -7, -32,   3,   7,   1,   1,   0,  -1, -11,   0,  37,  -7, -13,  -3,  -4,   6,   0,   6,  10, -18,  -7,   1,  25,  11, -20,   6,  -6,  12,   7,  15,   9, -54,  -6,  -6,   2, -16,  10,  -6, -59,  -3,  13,   1,   0,  -4, -15,  -1,   2,   2,  -6,  -2,   2,   0,  -6 },
      {  -2,  -3,  -4, -10, -15,  13, -31,   4,   2,   4,   5,  -1,  17,  76, -13,  -9,   0,  -1,   0, -13,  15,  11, -19,   2,  -2,   1,  -1,  11,  36,  10,  21,  -4,   1,  16,  16,  18,  -2,  -7,   0,  20,   8,   9, -11, -16,   2,  16,  -9,  -7,  14,  15,  -1,  14,  15,   1,   0,   6, -17,  -4, -22,  -5,   8,  38, -15,  -6 },
      {  -3,  -4,  -4,   1, -18, -24, -11,   6,   2,   2,   3,   1,  10,  59,  -5,   5,   2,   4,   4,  24,   5, -26,  19,   3,   0,  -2,   0,  -5, -58,  12, -44,   7,  -1, -19,  -7, -28,   7,  14, -21, -23,  -8,  -6,   6,  10,  -1,  13,   6, -11,   1,  -9,  13,  -8,   7,   3,  -5,   0,   2,  -5,  18,  -1,  -8,  24,   2, -12 },
      {  -2,   1,  -2,   1,   3, -18,  16, -10,  -4,   1,  -4,  -6,  -4,   8,   1,  11,  -1,   6,  -3,  11,  54,   8,  48,  -7,   0,  13,   5,  51,  -5, -33,  13,  23,   0,   3,  12,  -7, -34,   1, -20,   9,  -2,  12,  -8,   0,   2,  -3,   4,   1,  -5,  14,   0,   7, -12,   7,  -8,  15, -36,  19,  -4,   8,  -2, -24,  -5,  -1 },
      {   4,   0,   1,   0,  -1,  11,  16,   0,   5,  -2,   1,  -5,   9,  15, -32,  -6,   6,  -1,  -2,   4,  29,   0,  -5,  -7,   4,   2,   3,  22,   5, -25,  11,   9,   8,   0,   3,   9, -23,  -4,  -9,  -1,  15,  -7,  12,  -6,  -7,   0,  -7,  -2,  13, -33,  -5, -23,   0, -16,  -2, -33,  84, -11,  11,   8,  16,  -4,   9,  -2 },
      {   0,  -2,   0,  -5,   6,   5, -16,  -1,   2,   0,   2,  -3,   5,   6, -15,  12,   0,  -2,   3,  -8,   5,  17,  47, -24,   3,   5,   3,  -4,  23,  32,  15,  -9,  -3,  -1,  -5,   4,  44, -12,   4, -14,   7,  24,   9,  49,  14,  -2,  17,  17,  30,  -5,  13,  -3, -14,  -8, -13,  -9,   1,  27,  27,   1,  25,  -6,  18, -12 },
      {   1,   4,  -2,   2,   0,  -6,   7, -11,  -1,   2,  -5,   0,   0,  -2, -39,   7,   1,   8,  -1,   5,  -3, -31,  22,  29,  -3,   1, -10,  -7, -23,  14,  24,   1,   2,  18,  -5,   7,  24,  11,   6,  24,   2,  11, -19,  25, -28,  -1, -19,  15,   7,  -3,   5,  -1,  -8,  10,  -8,  10,  25,  12, -62,  -7, -34,  -7, -15,  -2 },
      {  -1,   0,   2,   4,   2, -14, -37,  19,  -1,   0,  -2,  -1,   2,  -2,  18,  -3,   2,   4,   6,  17,   4,   3,   6,  10,  -2,  -4,  -5,   0,  -3,  -8,  -5,  -7,   1,   4,  15,  47,  -1,   9,  12,  15,  14,   8,  33,  -8, -46,  -2, -32,   6,   0, -10,  -2, -29,   4,  -2,   4, -20, -16,  40,  39,  -7, -29,  -2, -11,   9 },
      {   1,   3,   4,  -2,  13,  22, -49,  13,  -2,  -1,  -3,  -3,  -1, -23, -56,  14,   1,   4,   5,   0,  -2,  14,   4,   7,   1,   1,   5,  -2,  -4, -10, -34,  -5,   1,   2,   7, -28, -12,  -5,  -2, -15,  -4,  -1, -18, -23,  22,  -7,  -3, -11,  -1,   2,  -2,  13,   9,   2,  17,   4,  13,  52,  -2, -18,  10,  -3, -30,  12 },
      {  -1,  -2,   0,  -5,  -7,  -8,   1, -33,   3,   3,   5,   6,   8,  -5,   9,  12,  -1,  -3,   1,  -8, -22, -41,   6,  -9,   2,   0,   3,   1, -12,  16,  49,  -6,   2,   4,  18, -14, -19,   2,  -5,   2,   6,  -1,  -9, -36,  -1,   8,  21, -18,  -8, -13,   5, -12,   9,  -4,   9, -16,  -5,  45,   1, -34,  13,  17,  43,   0 },
      {   2,   2,   3,   3,   0,  28,  44, -22,  -2,  -3,   0,  -3,   3,   6, -61,   9,  -2,  -3,   0,  -4,   7, -17, -31,   0,   0,  -2,  -3,  -5,  -5, -10,  -1, -13,  -2,  -4,  -5,   1,   5,  -2,   8,  -6,  10,  -6,  -1,  11,  -7,   1, -13,   4,  -6, -29,   7, -16,   0,  -7,   0, -20, -67,   1,  22,   9,   8,  -1, -14,   1 },
      {  -2,  -1,  -3,  -2,   3, -14, -18,  -3,   0,   2,   0,  -3,   5,  20,  16, -16,  -1,   1,   0,   3,  26, -15, -45,  -3,  -2,   0,   0,  -9, -14, -36,  13, -14,  -4,   4, -16, -13,  20,   1,  27,  -4,  -3,   0, -11,  25,   7, -13,  16,   6, -10,   3,  -6,   6,   6,   0,   6,   4,   8,  48, -10,  54,   6,  -7,  23,  31 },
      {   1,   5,   2,   5,  -4,  12,  77,  10,   0,   4,   1,   5,  -2,  13,  17,   6,  -1,   3,   3,   2,  -4,  19,   1,  17,   1,   2,   5,   3,  11,  33, -28,   5,  -1,   3,  13,   3,  11,  -7, -18,   3,  -4,   7,   8,  -8,   0,   1,  -5,  -3,  -2,  18,  -1,  -2,  10,   1,  -2,   2,  18,  56,  -2,  15,   2,  23,  -9,  32 },
      {   0,   0,   1,   2,  -1,   7,   7, -61,   1,   0,   2,   1,   4,  12,  17,  -8,   2,   2,   4,   3,  13,   7, -13,  16,   1,  -2,   2,   2,  17, -16, -33, -28,   0,   1,   1,   9,   0, -42,   6, -22,  -4,  -6,  -3,   9, -24,  -6,  10, -14,   7,   7,  29,   9,   0,   3,  10,  14,  14,  16,  -1, -29, -23, -23,   3, -43 },
      {   2,   4,   4,   2,   2,  -6,  32,  83,   0,   1,   1,   2,   7,   4, -11,  -2,   0,   1,   1,  -1,   3, -18, -18, -36,  -2,  -1,  -4,   2,  -6, -11,  18, -32,   0,   5,   2,   4,  -5,  -1,   7,  -1,  -7,  -3,  -8,   4,  -8,   2,   6,  -8,   4,  10,  24,  19,   0,  15,   4,  21,   6,  13,  13, -26, -10, -15,   4, -34 },
      {   1,   1,   1,  -1,   1,  10,  11,  -7,   0,  -1,   0,   0, -12,   3,  29, -22,   2,   2,   5,  -1,   6,  20,   0,  20,   0,  -1,   3,  -3, -17, -17,   6, -19,  -2,   0,  -6, -16,  -9,  56,  25,  10,   3,  -3,  11,  15,  34,  22, -25,  13,   9,   6,  27, -20,  -6,   4,   2, -19,   2,  16, -16, -29,  39,   1, -24, -26 },
      {   1,   0,   2,   1,   2,   5,   0, -23,   0,  -1,   1,  -1,   0,  12,   5,  -7,   1,   1,   3,   1,   7,  -2, -13,  -3,  -1,  -2,  -5,  -2, -10,  19,  19,  -5,   7,   2,  16,   2,  35,  41, -37,  23,   4,   0,  -4,  -9,  26, -54, -14, -35,  -4,  -3,  -8,  31,   1,   2,   4,  -2,  10,  -1,  35,   1,  -9, -35, -20, -10 },
      {   1,   0,   1,  -3,   3,  -9,  -5,  33,   1,   0,   2,  -5,  -3,   9,   8,  19,   1,   1,   2,  -5,  12,   7, -12,  14,   3,  -2,   6,  -3,   5,   7,  -9,   4,   3,  -1,   8,  -9,  23, -13, -19, -12,   4,  -1,  11,  -8, -10, -37,   2, -30,  -3,   3,  -5, -53,   5, -13,   1, -23, -22,  -4, -57,  -7,  27, -37,  15, -16 },
      {   0,  -1,  -1,  -1,  -4,  -1,  -8,   2,   3,   4,   4,   5,   3,  12,  16,  76,   0,  -1,   1,  -2,   4,   0, -45,  19,   1,   1,   0,   4,   2, -11,   4,  18,  -1,   0,  -2,   0,  -4,   4, -13,   1,   0,   0,  -1,  17,   7,  -2,  -1,  25,  12,  -2,  23,  10, -26,   0, -24,   4,   7,  -6,  13, -45,  -1, -10,   1,  40 },
      {  -1,   0,   1,   0,   1,  -1,  -5,   2,  -2,  -1,   0,   1,  -1,   1, -10, -75,  -3,  -1,  -3,  -2,  -6,  -4, -19, -29,   0,   1,   2,   7,   2,   3, -10,  48,   0,  -2,   1,  -3,   3, -10, -23,   4,  -1,  -4,   2,  16,   2, -22,   2,  15,  -9,   2,  24,  -4, -39, -15,  -3,  -9,  -9,  14, -11, -41, -10,  10,   4,  21 },
      {  -1,   0,  -1,   0,  -1,   4,  -5, -31,   0,   2,   1,   0,  -1,   2,   3,  24,  -1,  -1,  -2,  -1,  -6,   2, -10, -93,  -3,  -1,  -5,  -5, -12,   3, -24,  -2,   0,  -1,  -1,   4,   2,   5,   3,  31,  -5,  -1,   9,  -3, -19,   9, -13,  -2,   0,  10,   1, -26,  13,  21, -11,  20,  11,   3, -18,   2,  21, -12, -19,   9 },
      {   0,  -1,   0,   0,   0,   1,  -6,   4,  -1,  -2,  -1,  -1,  -7,   2, -10, -29,   0,  -1,  -1,  -2,  -5,   0,  -9,  38,   1,  -2,   0,  -3,  -1,  12,   5,  33,   0,   0,   2,   1,   4,  -7,  -5,  17,   0,  -5,  -7,  -4, -27,  15,  21,   5,   2, -15,  11, -11,  28,  55, -21,  36,   2,   2,  28,  -1,  50, -34,   3,   2 },
      {   0,   0,  -1,   0,   1,  -3,  -1,  -2,  -1,  -1,  -2,   0,  -4,  -4,   0,   0,   1,   1,   0,   0,   6,   2,  10,  11,  -1,   0,  -2,   4,  -4, -10,  -1, -51,  -1,   0,  -2,   2,  -2,  -4, -19,  11,  -2,   2,  -4,   8, -23, -41,   9,  -4,  -7,  -7, -15, -15, -55,  38,  24,  21,   4, -13,   5,  -7,  35,  51,  -3,  14 },
      {   0,   0,   0,   0,   0,   0,   8,   3,   0,  -1,  -1,  -1,  -3,  -4,   0,  26,   0,  -1,   0,   1,   2,   2,  -2,   1,   0,  -2,  -1,  -3,   0, -13,  -3,  68,   3,  -1,   3,   3,  15,  -1,  56,  25,   0,   0,   1,  13,  -3, -10,  19, -49,  -2,   0,   2,   5, -20,  15,  30, -11,   4,   5,   9,   1,   1,  30,  -9, -29 },
      {   1,   0,   1,   1,   1,   2,   0,   4,   0,  -2,   0,  -1,   0,  -2, -18,   0,   0,   0,   1,   0,   3,   4,   5,  17,  -1,  -2,  -3,  -1,  -2,  -6, -35, -15,   0,  -3,   2,  -3,  -1,   0,   8,  76,   5,   0,   9,   1,  22,   4,  -9,  -9,   1,  -4,   3,  -4,  -2, -46,  -7,  40,  -4,  -2,   7,  -5,   1,  -1,  60,  -3 },
      {   0,   0,   0,   1,   0,  -3,  -2,   3,   0,   0,   0,   1,   2,   0,   1,   9,   0,  -1,  -1,   1,  -1,  -5,  -5,   2,  -1,  -1,  -1,  -1,  -4,   1, -16,  -7,   0,  -4,  -1,  -4,  -2, -13, -24,  55,   3,  -2,   7,  -6,   8,  -7,  57,  59,   4,   7,   3,   7,  17,  15,  34, -50,  -2,   1,  -3,   7,  -9,  -5, -15, -22 },
      {   0,   0,   0,   0,   1,   0,  -2,   4,   0,   0,   0,   0,  -1,   0,   0, -15,   1,   1,   0,   1,   1,  -2,   9,   5,  -1,  -1,  -2,   1,   0,  -2,  -8, -33,  -1,  -2,  -1,  -2,  -5, -11,  15,  24,   0,   4,   1,   8,   3,  21,  40, -56,  -5,   4,  -4,   8, -14,  12, -63, -47,  -1,  -7,  -3, -11, -11, -12, -19,  34 },
    },
    {
      {-102,  63,  -3,   2,   0,   0,   0,   0,   2, -19,  16,  -4,   2,  -1,   1,   0,  26, -22,   1,   3,  -1,   1,   0,   0,   6,  -1,  -5,   1,   0,   0,   0,   0,   5,  -1,  -1,  -1,   0,   0,   0,   0,   1,   0,  -1,   0,   0,   0,   0,   0,   2,  -1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0 },
      {  21,  40, -38,   2,  -4,   1,  -1,   0, -69,  64,   4, -16,   5,  -3,   1,  -2,   0, -29,  40, -11,   2,  -1,   0,   0,  16, -24,   7,   9,  -5,   3,  -1,   1,   4,  -6,  -3,   6,   0,   1,   0,   0,   3,  -2,  -3,   2,   0,   1,   0,   0,   1,  -1,   0,   0,   0,   0,   1,   0,   1,  -1,   0,   0,   0,   0,   0,   0 },
      { -27, -45,  76, -21,   6,  -4,   2,  -1, -73,   7,  21,  15,  -5,   1,   2,   0,  -2, -10, -14,  15,  -1,   0,   0,   1,  18,  -7, -14,   4,   1,   1,  -1,   0,   7,   3,  -6,   0,   1,  -1,   0,  -1,   6,   1,  -4,  -1,   2,   1,  -2,   0,   1,   2,  -2,  -1,   0,   1,   0,  -1,   2,   0,   0,   0,   0,   1,   1,  -1 },
      { -49, -20,   9,  16,  -1,   3,   0,   1,   6,  43, -71,  15,   3,  -1,   1,  -1, -22,  43,   4, -34,  14,  -6,   2,  -2,  11, -22,  34, -15,  -4,   3,  -2,   0,   7,  -7,   5,   3,  -6,   2,   0,   0,   3,  -4,   2,   2,  -2,   0,   1,   0,   2,  -1,   0,   1,   0,  -2,   0,   1,   1,  -2,   1,   0,   0,   0,  -1,   0 },
      { -13,   3,  24, -28,  14,  -5,   4,  -3,  42,  18,  29, -35,  -5,   0,  -4,   1, -59,  22,  46,  16, -22,   3,  -3,   1,  -3, -32,  -6,  43,   2,  -2,   2,   0,   1,  -7, -18,  12,   9,  -1,   0,   1,   1,  -2,  -6,  -2,   5,   1,  -1,  -1,  -2,   1,  -3,  -1,   2,   3,   1,  -2,  -2,  -1,   0,  -2,   0,   1,   1,  -1 },
      { -12, -49,  10,  14,  -1,   3,  -1,   2,  37,  30, -24, -15,   5,  -4,  -1,  -2,  57, -53,  41,  -7,  -7,   3,  -4,   2,   6,   7, -24,  23, -11,   2,  -2,   0, -18,  25,  -9,  -5,   7,  -4,   2,  -1,  -5,   1,   4,  -3,   0,   2,  -1,   1,  -2,   4,  -1,  -1,  -1,   0,   0,   1,   0,   0,  -1,   1,   0,  -1,   0,   0 },
      { -18, -19, -37,  49,  -8,   5,  -1,   2,  -7,  11,  -6,  22,  -6,   0,  -5,   1, -40,  -6,  -4,  47,   4,  -8,   1,  -2,   3,  -5, -50,   6,  42, -13,   4,  -1,   4,   1,  -1, -39,  21,   5,  -4,   2,   6,  -4,   8, -16,  -6,   5,  -2,  -1,   1,   1,   4,  -6,   0,   2,  -1,  -1,   0,  -1,   3,  -4,  -1,   2,  -1,  -1 },
      {   5,  31,  20, -43,  14,  -7,   2,  -2,  15,  67, -22,  -8, -17,   6,  -2,  -1,  18,   5, -35,  37, -17,   0,   1,   0, -23,  32,  -9, -10,  13, -13,   5,  -1, -10,  -6,  30, -19,   8,  -2,  -1,   0,   2, -14,  12,   2,  -3,   5,  -2,   1,  -1,  -6,   2,   4,  -2,   0,   1,   1,   2,  -5,   0,   2,   0,  -1,   0,   1 },
      { -28, -44, -34,   8,  17,   3,   4,   1, -25,   4,  24, -69,  27,   0,  -1,   1,  -2,  27,  -1, -12, -20,  11,  -5,   1, -24,  40,  -5, -10,  -9,   0,   3,  -1,  11, -13,  25,  -8, -13,   3,   0,   1,   5,  -5,   3,  12, -12,  -5,   4,  -1,   4,  -3,   3,   4,   1,  -6,   0,   1,   0,   0,  -2,   3,   1,  -1,  -2,   0 },
      {  -4, -12,  -3, -12,  10,  -1,  -1,   2,  32,  17,  19,  -9,  -2,  -2,   3,  -2, -42, -56,  -9,  12,  27,   4,   6,   0,  34,  -6,   0, -61,  -6,  15,  -5,   2,   5,   4,  22,   4, -37,  10,   1,  -2,  -2,  11, -14,  18,  -8,  -5,   7,   1,  -1,   5,   1,   2,  -3,  -9,   2,   4,   3,   2,  -4,   5,   0,  -4,  -2,   2 },
      {  -6,   7,  28,  30, -51,  15,  -7,   1,   0,  16,  21,  21, -14, -18,  14,  -5, -30, -17,  23, -17,  17,  -8,   4,   4, -63,  40,  -1,   1,  -9,   7,  -1,  -2,  -8,  -7,  24,  18,  -1, -11,   5,  -5,   6,  -9,  -6,  13,  13,   2,  -3,   2,  -4,  -3,  -2,   1,   3,   3,   2,  -2,   1,  -1,  -1,   2,   0,   3,   1,   0 },
      {  -7, -26, -17,  26, -24,   8,  -3,   3,  13,  25,  43,   1, -42,   4,  -6,   0,  37,  19, -10,  34, -16, -17,   7,  -4,  24,  -9,  44,  -7,   5, -20,  -1,   0,  12, -20,   5,  32,   9,  -3,  -8,   1,  -8,   8, -20,   4,  20,   5,  -1,  -1,  -2,   2,  -3,  -4,   3,   4,   1,   2,   1,   1,  -6,  -1,   2,  -1,   1,   1 },
      { -11,  -2, -28,  -6,   4,   6,   3,   0,  -6,   6,  -5,   8, -21,  16,   1,   0, -36,  -5, -29, -11, -22,  15,   5,   1,  34,  35, -12,  13, -45,  -8,   6,  -3, -27,  35,  -8,  33,  13, -35,  11,  -4,   3,  -3,   9, -17,  32, -15,  -5,   4,   1,  -6,  12,  -3,   3,   6,  -9,   0,   2,  -1,   1,   0,   2,   3,   1,  -3 },
      {  12,  24,  33,  26, -45,  13, -17,   8,  10,  14,  30, -18,   8,   6, -10,  -2,   2,  38,   1, -13,   1,   8,  -3,  -6,  51,  12,  -9,  -8, -17,  -1,  -3,  -2, -10,  29,  -4, -41,   1,   8,   3,  -4,  -4, -11,  22,  -1, -23,   1,  12,  -3,   7,   0,  -3,   7,  -6, -15,   6,   6,   6,  -4,  -3,   4,   0,  -7,  -3,   6 },
      {  -1,   2,  -1,  11,  17, -26,   9,  -6,  14,  34,  32,   0,  47, -41,   1,  -6,   8,  13, -38,  -9,  52,  -7,  -9,   5,   3,  10,  -7,  10,   4,  17, -15,   4,  -6,   1, -23,  23,  15,   8,  -7,   1,  -7,   1,  -3, -29,  15,  14,  -5,  -2,  -2,  -7,  10,  -7,  -8,  17,   1,  -4,  -6,   0,   2,  -4,  -6,   5,   5,  -2 },
      {  12,  12,  22,  33, -21,  -5,  -2,   1,   9,   3, -27, -12,  39,  -6,   3,   0,  -8, -25, -34,  18, -26,  38,  -8,   1,   8,  -2,  -2,  22, -30,  -8,  15,  -7,  46, -44,  -3,   5,   7, -13,   4,   2,  -2,  14, -24,   5,   0,  -2,   3,   3,   2,   3,   0,  -9,   5, -11,   0,   6,   1,   2,  -2,   0,   0,  -1,  -5,   2 },
      {   6,   7,   0,  -3, -22,  19,   8,  -8,   4,   2,  -7, -20, -11, -16,  20,  -3,  17,  24, -18, -18,  -1, -15,   4,   9,  18, -28, -62,  10,  19,  14,   3,   5,  19,   8,  27,  16, -31, -14,   1,   3,  -3,   9,   0,  31,  19, -19, -15,   3, -22,   1,   7,  -4,  20,  17, -10, -11, -11,   0,  11,   0,  -1,  14,   3,  -9 },
      {  -8,  -8, -18, -10,  -6,  23, -10,   4,   2,  27,  17,  22, -13,  15, -10,  -1,  14,  14, -20,   8,   7,  29,  -7,  -2, -46, -25, -13,   3, -15,  31,   9,  -4,   4,   8, -45,  -3, -31,   3,  25,  -5, -16,  17,  -4,  -7, -13, -32,  24,  -2,   9,  -5,   5,  -5,  11, -26,   1,   8,   1,  -1,   5,  -5,   3,  -3, -10,   4 },
      {   5,  15,  22,  57,  24, -41,  14, -11,  -9,   1,   0,  -2,  14,   3,  -8,   3,   0,   8,   8,  25, -30,  20, -10,  -1, -18, -19,  14, -22,   9,   4,   4,   1, -35,  39,  10,   7, -28,  -7,  -1,   0, -19,   9,   4,  16,  20, -15,  -4,  -2,   5, -13,   6,  -5,  10,  12,  -6,  -4,   0,  -7,  -2,   6,  -1,   5,   6,  -3 },
      {   2,   3, -10,   3,  33, -21,   8,  -3,   5,   4,   9,  58,   2, -15,   3,  -4,  10,  25,  33,  11,  -8,   7, -16,   0,  34,  37, -19,  20, -12,  22,  -6,  -7,   1, -24,  21,   7, -17,  11,  14,  -5,  14, -17, -19,  25, -11, -10,  15,  -1,  -3,   2, -15,  -1,  15, -15,   5,   3,   0,  -1,   2, -10,   6,  -1,  -7,   5 },
      {   1,  -9,  -5,   2,  17, -11,   1,  -3,  15,  11,  36,  30,   0,  10, -21,   2,   9,  -5,  -1, -29,   4,  29,  -7,   2, -10, -36,  -3, -12, -20,  -6,  23,  -2,  28, -11,  24, -26,  10, -37,   5,   9,  15, -27,  30,  -5,   1,   3, -32,  12,   0,   4, -17,  21,  -6,  22, -14,  -9,  -3,   1,  -8,   2,   2,   5,   7,  -8 },
      {  -7, -12, -11,  -8, -15,  18,  -2,   2,   0,  -1,  12,  12,  23, -18,   9,   5,  -6, -17, -33,  -5, -17,  15,   6,   7,   5, -34,  25,  36,   7,  -1,  -1,  -2, -65, -13,  35,   2,   7,  20,   0,   2,  -6, -19,  10,  13, -31,  -7,   7,  -2, -19, -10,   8,  14,  12,  -8,  -2,  -5, -11,   6,   1,  -3,   1,   5,  -4,  -5 },
      {   3,   0,  -1,  10,   4,  11, -16,   9,  -7, -17, -16, -27,  -2,  -6,  13,   2,  10,  11,  -1,  51,  28, -11,   3,  -6,  -5, -30,  -3,   4, -34,  43,   1,  -7, -16,  -1,  19,   8,  -2, -17,  19, -11,  29, -46,   3,  -7,  21,  13,   7,   9,  10,   1,  -9,   1, -25,  -2,   5,  11,   8,   8,  -5,   4,   3, -10,  -3,   4 },
      {  -7,  -4, -17, -18,  -7,  11,  -6,  -4,  -7,  -1,  -3,  15,   5,  -9,   2,  -9,  -4,  13,   1,  12,  -5,  26, -27,   6,  -9, -11,  -5,   1, -18,   9,  -8,  -5,   9,  39,  29,  -1,  20,  -4, -14,  -9, -33,  32, -15,  17,   1,  52,  -3, -13, -11,  16, -18,   3, -14,   4,  42, -16,   6, -12,   5,  -1,   3, -16,  17,   7 },
      {   4,  10,  16,  27,  22,  26, -22,  -5,  -1,   3,  -4, -14, -11,  59, -46,   9,  -1,  -6, -12, -19,  12,  13, -17,  -2,   5,  11, -15,   9,  23,  18,   0,  -4, -21, -28,   4,  28,  -7,  19,   5,  -3,   3,  -2, -11,  -6,   0,  12,   0,  -8, -19,  17, -12,   2,  -8,  15,  15, -20,  -7,   1,  -1,  -7,   3,  -6,  22,  -7 },
      {  -1,  -2,  -7,   7,  -8, -22,   0,   1,   1,   2,  -6,   0,  -2,   7, -27,   4,  -8, -15, -23,  15,   4, -31,   1,   3, -11,   2,  10,  39, -23, -11, -33,  14,  34,  26,  -8,  15, -33,  19, -22,  -9,  10, -19,  41,  24, -15,  -2,  -3, -16,  10,  11, -18,  18,  18,   1,   5, -11,   0,   2,  -6,  -8,  16,  -6,  11,  -2 },
      {  -4,  -4,  -4, -23, -33,   0, -10,  -1,  -5,  -8, -11,  -2,   3, -22, -28,   3,  -2,   1,  18,  29,   3,   2, -52,  14,  12,  15,  21,  -1,  -1,  22, -14,  -6,  -5, -21,  10, -14, -13,   0,  22,  -8,  -6,  31,  26, -28,   7, -27, -16,   2,  -4,  14,  11,   2,   4,  26, -26, -13,  -1,  -2,  -7,   7,  -3,  10,  16, -15 },
      {  -1,   2,   5,  -8,  13,  10, -34,  16,  -6,  -4,  -2,   9,  24,  -1, -32,  18,   1,   8,  23,  17,  29, -10,  23, -10,  -1,   4,  -8,   3, -16, -32,  31,   0,  10,   3,  27,  19,  -6,  -6, -11,   7, -57,  -7,  17,  -5,   4,  -3,  10,   0, -20,   9,  22, -21,   8, -12, -10,  13, -16,  12,  -3,  -3,   0,   7, -13,   0 },
      {   2,  -1,   5,  13,   6,  -9,  16,  -1,   1,  -4,   1,   3, -31,   7,  19,   5,  -2,  -9,   4,  -8, -16,  12,   6,   2,   6,   3,   2,  11,   5,  34, -12,  -3,  37, -10,  17,   9,   0,  20,   6,  -7, -34,  -4,  41, -22,  -1,  12,  17,  -2,  -6, -41,  45,  18, -31,   1,  13,   6, -24,   3,   4,  15,  -5,  -6,   0,   6 },
      {   2,   7,   7,  26,  33, -20,   2, -13,  -3,   2,   7, -14, -36, -13,  12,  -2,  -4,  -3, -16, -17,   7, -38,   4,  13,  -6,  -7,   8,  24, -23,   1,  12,  -4, -11,  -7,  20, -29,   0, -10,  21,  -3,  -6,  47,   7, -19, -14,  -1,   5,   7, -24,  44,  -1,  -3,   1, -22,   5,   5,  12,   0,   0,  -4,   8,  -3, -10,   6 },
      {  -4,  -3,  -6,  -7, -37, -54,  26,  -6,  -7,  -6, -10, -14, -36, -23, -25,  34,   3,   7,   1,  -5,  10,  29, -13,  10,  -2,  -6, -20, -18,  -2,  -6,  25,  -2, -14,  -1, -12,  24,   3,   7, -12,   5,  13, -19,  -7,  -1, -21,  14,   7,  -2, -15,  24,  10,  -3, -11, -21,   1,  -1, -18,  16,   2,  -2,   7,  -6,  -6,  -1 },
      {  -2,   2,  -1,   6,   6,   7,  -5,  21,  -2,   4,  -2,   7,   5,  -7,  21,   7,   1,  11,  10,  22,  -9,  -2,  36, -10,   2,  13,  -5, -13, -28,  11,  11,   7, -21, -28, -45,  -8, -23, -13, -18,  16,   2,  19,  12,  14, -18,  14, -40,  -8, -33,   2,  18,  30,  -9,  16,  13, -23, -12,   5,   3,   0,  12, -13,  17,  -5 },
      {  -3,  -5,  -3,  -4, -18, -15,  36, -19,  -4,  -7,  -6, -10,  -6,  30,  14, -24,   0,   3,   6,   9,  41,   3,  -8,  -9,   7,   0, -16,  12,  -9, -15,   2,  16, -26, -35,   8,   3,  -9, -17,  -8,  17, -47,   8,   5,  -6,  -4,  -4,  -9,   0,  30, -27, -32,  20,   4, -12,   0,  -1,   7, -20, -11,  -2,  14,  -8,  -2,  -1 },
      {   2,   3,   1,   2,  -1,  -1,  -1,   9,   1,   2,   3,  15,   1,   3,  -2,  -2,   4,   9,   3,  -9, -32, -10,  12,  10, -13, -22, -38, -18, -43, -15, -40,  22, -13, -21,  17,   4,  -8,  31, -26, -13,  -7,  -4, -23, -35,  -5,  -2,   3,  -4,  20,  16,  11, -11, -22,   4, -24,  -7,  21,  -5,   0,  17,  -9,   2,   9, -16 },
      {  -7,  -9, -11, -17, -30, -50,  -1,  10,   0,   3,   4,   2,  11,  33,  -9,   8,  -3,  -4,  -7, -13,  -8, -11,  33, -13,   0,  -1,   0,   6,  -6,  30, -11, -10, -11, -18,   0, -38,   4,   9,  11,  -8, -22,  -5,  -2,  20,  38,  -1,  -5,  -2,  -7,   8,  -7, -36,   7,  23,  15,   9,  -2,  18,   9, -16, -16, -10,   3,  15 },
      {   1,  -4,   0,  -8, -11,   0,  11,  -9,   5,  -2,   6,   9,  13,  11, -11,  -9,   8,   2,   9,   2,  10,   5,   4,  -4,   1, -10,  -5,   6,  -8, -21,  23,   0, -11, -12,  19,   9, -25,   4, -12,  -3,  62,  35,  33, -12,   9,  15,  12, -21,   6, -19,  32, -28,  10,   7,  36,   6,  20, -20,   5,   0, -14,  -2,  -2,  23 },
      {  -2,   0,  -1,  -4, -17,  16,  39, -42,   1,   5,   3,  12,  36,  20,  21, -28,   1,   3,   7,  13, -19,  -5,  13,   3,  -2,   4,   7, -13,   9,  -7,   8,   5,   7,  10,  -4,  18, -18,   5,  20,  -1,  -1, -18,  19, -29,  -2,  -3,   8,  12, -29,  50, -19,  -7, -18,   1,   3,   9, -22,  10,  -3,  -2,  -2, -17,   5,   5 },
      {   3,   4,   9,   9,  17,  31,  28,  -3,  -2,  -4,   0, -10, -17,  -3,   7,  -2,  -1,  -5,   4,   6,   8,  15, -25,   9,  -3,  -4,   7, -25, -26,   4, -28,  25,  -2, -16, -10, -21,  22,  -3, -31,  -5,  -3, -20,  18,   0,  18,  -5,  11, -28, -27,   3,  -2, -25,  42,   0,   5,   5, -23,   0,  18, -33,   0,  15, -12,   8 },
      {   0,  -1,   1,  -1,   2,  -3, -16,  24,   1,   1,   1,   4,  16, -19,  -1,  14,  -3,  -4,  -3, -14, -33, -33, -25,  18,  -1, -13, -27, -35,  13, -17,  -4, -11, -14, -27, -17,  26,   7, -17,  22, -23, -12,   5,  37,   3,  14,   1,  12,  -9,   9, -10, -29,  14,   0,  -9,  27,  14,  -2,   5, -13,  -9,  14, -17,  -8,  18 },
      {  -1,  -1,  -3,   1,   5,  15,   0,  -1,   1,   1,   3,   4, -13, -11,  12,  13,  -6,  -7, -10,  -8, -13,  -2, -30,  23,   6,   9,   9,  14,   9,   6,  22,   2,  -3,  -8, -21,  -6, -43,   1, -14,  26, -25, -31,  15,   7,   2,  33, -17,   1,  -3,   1, -12, -44, -14,   3,  -2,  10,  48,  -1,  -8,  14, -29,  -1, -17,  -8 },
      {  -2,   2,  -2,   0,  -3,  -8,  -8,  32,   0,   7,   2,   6,   4,  21,  17,  15,  -1,   6,  -1,   2,   3,   9, -16,   4,  -3,   4,   0,  14,   0,  -4, -10,  20,  -5,  10,  10,  -5, -19,   1, -23,   8,  26,  39,  -2, -10,  12,  12,  10,   5, -22, -18, -44,   8, -22,   2, -21,  33, -41,  30,  -9,   8, -14,   7, -31, -14 },
      {  -1,  -1,  -2,   0,  -2, -13,  -2,  18,  -2,  -3,  -3,  -5,  -1,  -1,  -2, -12,   3,   4,   4,   4,  11,  35,  24,  -7,  -2,   0,  -5,  -4,  15,   5, -16,  10,   6,  -4,  -5,  18,  21,   8,  -6, -16,   1,   6,  27,   8,  10, -11,  16,  14, -59,   6, -16,  20,   5,  -4, -16,  18,  70, -23,  -9,   6,  -5,   0,  -2,  -8 },
      {   2,   1,   3,   1,   5,   5,   4,  31,  -1,  -3,  -4,  -9, -21, -43,   9,  -6,  -1,  -4,  -2,  -7,  -8,  24,  27, -33,   3,   4,   3,   1,   3,  -2,  31,  -9,   4,   3,   9,  -1, -21,  30,   1,  -2,  -5,   2,  12, -39,  11,   1,  13,  -5,  22,   0, -51,  -9,  15,  13,   4, -19,  -6,   1,   9, -26,   0,   1,  30, -11 },
      {  -1,   0,  -2,  -1,   4,  14, -16, -74,  -3,  -2,  -7,   0,  -9, -28, -32, -11,  -2,   2,  -1,  -1, -12,  -1,  25,  -2,   1,   2,  -9,  -7, -14,  18,  11,  11,  -5,  -7, -11,  -7,  14,  15, -12,   5,   0,  17,   7,  17,  10,  11,  17,  -3,   7, -22, -15,   9,   1,  11,  -8,  13,   0,  36, -45,   3,  -6,   7,  -3,  -4 },
      {   3,   6,   3,   1,   4,  35,  76,  25,   1,   3,   0,   5,  12,  -1, -21,  57,   0,   2,   1,   2,  10,  -9,  13,  18,   1,   4,   3,  10,  -2,   3,  10,  -1,   3,   8,   7, -15,   3,   2,   6,  -9,  -4,   8, -19,   3,  13, -11,   9,  -6,   5,   0,   0,  19,  -2,  10,   4,   2,  21,  16, -30,   9,   3,  -7,  11,  12 },
      {   0,   1,   0,  -3,  -1,  -9, -24,  13,   0,   1,  -1,   1,  -1,   2,  17,  -4,   1,   4,   3,  -4,  -3,  29,  18,  23,   3,   4,   2,   5,  28,  22, -14,  25,   1,   3,  12,   3,  -4, -14, -16,  27,   1,  -6,  -1, -30,  14,  -9, -13, -20,  25,  37,  14,   9,  32, -14,  26,  -7,   2,  19, -50,  13,  12,   0, -13,  25 },
      {   0,   2,  -1,   6,  11,   1,   0,   9,  -2,   1,  -1,  -4, -14,  13,  37,  -9,   3,   8,   1,   6,  25,  24,  -4,  14,  -2,  -3,  -8,  -1, -13, -34,  -6, -36,  -6,  -4, -13,   7,   7,  36,  23, -28,  -4,   4,  16,  28, -11,  17, -15,  12,  17,  19,   8,  -7,  24,  24, -10,  11, -13,  20, -31,  21, -23,  18,  13,  14 },
      {   0,   0,   0,   2,  12,   9,  -2,  -4,  -1,  -1,   1,   1, -12, -27,  -7,  -9,   0,   0,  -2,  -6,  15,  29,   5, -17,   3,   3,   5,  19,  22, -28, -25,  20,  -9, -15,  -5, -12, -15, -43,  -7, -25,  -4,   1,   7,  28,   7, -23,  22, -12,  20,  25,   9,  -7, -42,   8,   5,   8,  -7,  22,  20,  15, -16, -30,  17,   3 },
      {  -2,  -4,  -3,  -4,  -2, -13, -24, -20,  -1,  -4,  -3,  -4,  -6,  -6,  11,  10,   1,   0,   1,  -2,  10,   6,  -2,  20,   0,  -3,   3,   4,  -5, -44, -11,   6,   6,   4,  -6, -27, -34,  -5,  26, -16,  -1, -27, -38, -28,  10,   1,  14, -10, -47, -39,   1,   7,  12,  13,  18,  -9,  -1,  -5, -24,   2,   6, -15,  21,   3 },
      {  -1,  -2,  -4,  -3,   1,   1,   6,   7,  -2,  -2,   0,   2,  -1, -10, -11, -18,  -4,  -6, -10,  -9,  -6, -15, -20, -40,  -2,  -3,  -2,   6,  -5,   1,  19, -18,  -6,  -6, -12, -18,  -7,  17, -20,   0,  -1, -16,  -9,   6,  42,  -2,  10, -14, -13,  23,   5,  27,   4, -13,   5,  10, -29, -46, -18,  59,  -5,  10,  -4,   5 },
      {   0,   1,   1,   4,   3,   4,   3,  19,  -5,  -5,  -7,  -6, -13, -21, -38, -19,  -2,   0,  -3,   4,  -4,   4,  17,  24,  -2,   0,  -2,   0,  -8,  -2,  -9,  32,  -2,   2,   0,   7,   0,   7,  27,  28,  -4,  -2, -11,   3, -19,  -5, -14,  26,   0,  -3, -12,   5,  -5,  33,  12,  19, -24, -42,   4,  -4, -37,  -6, -18,  49 },
      {  -1,  -3,  -3,  -1,  -1,  -3, -27, -30,   2,   2,   3,   2,   6,   8,  25,  51,   2,   1,   4,   7,  10,  10,  23,  52,   5,   7,   6,   1,   0,   3,  23,  16,   0,  -7,  -5,  -3,   9,  15,  -4, -12,   0,   3,  12,  10,  22, -10,  13, -11,  22,  16, -20,  15,  -8,  -1,  -4,  -4, -12, -46,  35,  10,  -5,  12,   6,  -4 },
      {   0,  -1,  -2,  -1,  -1, -14,   3,  30,   1,   0,   0,   1,   0,  11,  10, -25,   0,  -1,  -2,   1,   1, -17,  -6,  16,   1,   0,   3,  -3,  -4,  14,  34,  47,   4,   1,  -3,   3,  28, -10,  16,   0,   4,  -1,   5,  21, -24, -11,  36, -24,  -6,  -7,  -2, -39,  -8,   7,  -1, -25, -13,  -8, -37,  20,  -4,   3,  30, -27 },
      {   1,   0,   0,  -1,   1,   6,  15,  12,  -1,  -2,   0,  -1,   1,  -5, -30, -28,   0,   1,   1,  -2,  -2,  12,  35,  34,   5,   4,   7,   5,  15,   9,  -4,  -2,  -4,  -6,  -5,  -9, -18, -14,   8, -55,   1,   2,  -1,   0,   8,  48, -15,   4,   7,  -6,   2, -18,   9, -19, -21,  -4, -21, -21,  -5,   6,  36,  -2, -28, -25 },
      {   0,   1,   1,   1,   7,   9,   1,  -2,   0,   0,   1,   1,  -1,  -7,  -8,   6,   1,   1,   2,   1,   6,   5,   1,  13,  -1,  -2,   0,  -2,  -5,  -7, -16,   1,   0,  -1,   2,  -3,  15,  25,  -3,  25,  11,   8,  17,   7,  25, -19, -25,  50,  -8, -21, -12, -35,   7, -50,  12, -22, -13,   9,   6,  37,   9, -44,  26,  -2 },
      {   0,   0,   1,   0,   0,  -2,  -4,  19,  -1,  -1,  -1,   1,   0,   3,  -2, -47,  -1,   0,   0,  -2,   2,   0, -22,  35,   0,   0,   2,   4,  -4, -12,  40,  26,  -2,   2,  -2, -12,  -3,  40,  -3,   2,  -1,  -5,  -7,  14,  42,  -5,  -2,  10,   6,   4,  29,  35, -15,  -3,  -3,  12,  13,  29,  17, -35,  18, -15,   6,  -1 },
      {   0,   1,   0,   3,   0,  -6,   8,   6,  -1,  -2,  -3,  -3,   1,   4, -19, -32,   1,   1,   3,   9,   3,  -9,  14,  34,   2,   2,   2,  -4,  -3,   0,   6, -25,   2,   0,  -2,   3,   1, -13,  -4,  -9,  -4,  -6,  -5, -13,  -4, -13,  -3, -23,  -8, -12, -21,  23,   7, -22,  26, -30,   9,  40,  36,  32, -55,  30, -10,  -8 },
      {  -1,  -2,  -2,  -1,  -3,  -1,   1,  -1,   0,   1,   1,   2,   5,   2,  -2,  14,  -1,  -1,  -2,   0,  -7,  -3, -10, -29,   2,   2,   0,  -2,   2,   0,   2,  61,  -3,  -5,  -4,   0,  -7,  -5,  32,  -6,  -3,   0,   2,  -3,   3,  45,  14,  26,   4,   3,   2,  11,  31,  -6,  -6,  -8,  12,  27,  37,  37,  13,  36,  15,  28 },
      {   0,   1,   1,   2,   3,   2,  -4,  -9,   0,   1,   0,  -1,  -2,  -3,   7,   4,   0,  -1,  -2,   0,   4,   4,  -7, -23,   0,   0,   1,   1,  -1,  -1,  20,  22,   5,   7,   7,   8,  14,  29,  16, -31,  -1,   1,  -3,  -2,  -6, -22, -49, -53, -10, -11, -15, -13, -13,   8, -20, -18,   9,  12,   8,  23,  16, -28, -44,  34 },
      {  -1,  -1,  -2,  -1,  -1,  -1,  -5,  -4,   0,   0,   0,   1,   1,   3,   4,   6,  -1,  -1,  -1,  -1,  -5,  -6,  -9, -10,   2,   2,   3,   4,   6,  11,  17,  19,   0,   1,   0,  -1,  -7,  -8, -33, -59,  -2,  -2,  -3,  -4,  -5, -13, -29,  32,  -5,  -7,   0,  -3, -19, -39,   1, -12,  -7,  -8, -30, -34, -35,  28,  26,  38 },
      {   1,   1,   1,   2,   2,   0,   0,  -4,   0,   0,   1,   0,   1,   0,   0,  11,   0,   0,   0,  -1,  -2,  -7,  -4, -12,   0,  -1,  -2,  -2,   2,  10,  10,  24,   2,   4,   7,   7,   7,  -3,  12, -16,   1,   0,  -6,  -9,  -9,  13, -30, -17,   9,   2,  10,  18,  45, -17,   1,  50,  -5,  -3,   5,  -3, -51, -50,  21, -46 },
      {   1,   1,   0,   0,   0,   3,   7,  -1,   0,   0,   0,   1,   3,  -4,  -1,  13,   1,   1,   1,   2,   2,   9,   4, -23,  -1,  -2,  -1,  -2,  -2, -21, -23,  18,   1,   0,  -1,  -2,  -2,  14,  53,  20,   3,   5,  10,  17,  30,  12, -11, -16,   3,   9,  11,   8, -26, -35,  13, -37,  -2, -10, -23, -33, -30,  25, -29, -34 },
      {   0,   0,  -1,   0,   0,   1,   2,  -2,  -1,  -1,  -1,   0,  -3,  -5,  -5,   2,   0,   0,   0,   1,   1,   5,   2,  -5,   1,   1,   1,   1,  -2,  -5,   0,  15,   0,   2,   1,   0,   3,  12,   5, -13,  -3,  -4,  -4,  -2, -10, -28, -36,   6,  -1,  -1,  -6, -15, -18,   9,  67,  51,   3,   4,   4,  14,  37,  50,   3, -38 },
      {   0,   0,   0,   0,   0,   1,   2,   3,  -1,  -1,   0,   0,   0,  -3,  -8,  -9,   0,   0,  -1,  -1,  -2,   3,  10,  13,  -1,  -2,  -1,   1,   4,   1, -13, -13,  -1,   1,   3,   2,  -3,  -3,  19,  30,   4,   2,   1,   3,   5,  12, -20, -56,  -9,  -8,  -8,  -8, -20, -38, -37,  40,  -1,  -2,   5,  -4,  11,  27,  61,  31 },
    },
  },
  { // 34
    {
      {  95, -52, -25,  18,   2,  -6,   3,  -4, -24,  18,  31,   6,  -9,  -7,   5,   2, -14, -13,   1,   2,  22,   4,  -5,   0,   2,   6,  -1, -14,   4,   4,   2,   0,  -3,  10,   0, -11,  -5,  -1,   2,   2,  -2,  -4,   5,  -4,  -2,   0,   0,  -1,  -3,   6,  -3,   1,   0,  -1,   0,   2,   1,  -3,  -1,   2,   0,   2,  -1,  -1 },
      { -48, -34, -38,  19,  15,  11,  -7, -12,  34,  38,  43, -36, -18,  -6,   8,  13,  11,  -1, -17,  35,   8,  -6,  -6,  -5, -10, -14, -13,  14,   5,   5,   3,  -3,   3,   5,  15, -27, -11,  -3,   1,   7, -10,   7,   6,   0,  -2,   5,  -6,  -7,   5,  -7,  -4,   3,   3,  -1,   7,   4,   2,   3,  -1,  -1,   0,  -3,  -4,   0 },
      {   0,  42, -17, -28,  -3,  -7, -20,  -7, -44, -41,  41,  14,   3,   3,   5,   3,  18,  -5,  -3,  59,   7,   3,  15,   1,   1,   1,   0, -13, -13,  -4,  -3,   1,  12,   7, -12, -36,   1,  -1, -12,  -1,   0,   1,  -3,  -4,  -2,  10,   5,  -4,  19,   9,  -7,  -9,  -5,   1,  10,   6,  -7,  -2,   4,   5,   1,  -4,  -5,  -1 },
      {  29,   0,  -5, -30,  -5,   8, -23,   0, -12,  60, -45, -20,   6,   6,  -3,  -1,   5,  37,  31,  29, -23,  -9,  12,  -1, -21, -26,  17,  26, -20,  -4,  -5,   0,  11,  -7,  -7,  -7,   8,  -1,  -9,  -2,  -4,   6,  -8,   4,   4,   8,   3,  -2,  17,  -5,  -3, -10,  -3,   2,  10,   3,  -9,   1,   4,   1,  -2,  -8,  -5,   1 },
      { -20, -64,  24,  -1,   2,   7,  -7,  -4, -46, -12, -21,  13,   8,   2,   6,   4,  60,  28, -40, -10,   5,   0,  -2,  -1, -19,  20,  15, -13, -15,   2,  -2,   0,   9,  -5,   6,  -4, -10,  -3,   1,   1,  -4,  10,   4,   1,  -1,  -3,  -2,   0,  11,  -3,  -3,   2,   2,   0,  -2,  -1,  -3,   0,   1,  -1,  -2,   2,   1,  -1 },
      {   3,  50, -26,  11,   6,  10,   6,  -3, -66,  35,  18, -34,  -2,  -3,   0,   3,  30, -30, -18, -35,  -8,  -2, -10,  -3,   0,   1,   4,  21,  -1,   2,   4,  -1,   1, -17,  12,  26,  -1,  -2,   6,   1,  -2,   5,   3,   3,   2,  -5,  -8,   0,  -6, -12,   7,   9,   4,   1,  -4,  -6,   8,   7,  -3,  -4,  -1,   2,   5,   0 },
      { -30, -19,  38,   4,  -7,  -8,   2,   8, -38,  45, -21,   4,   9,   5,  -4, -12, -13, -61,  28,  24,   6,  -2,   9,   8,  25,   9, -37, -25,   5,   6,  -1,  -3,   0,   9,  16, -12, -12,   0,  -3,   1,   9,  -2, -10,   5,   4,   1,   2,  -3,   0,   5,  -1,  -5,  -5,  -2,   6,   6,  -1,  -3,   4,   3,   0,  -1,  -6,  -2 },
      { -29, -11, -15,  24,   0, -12,  10,   3, -28,   1,  14,  14,   0, -10,   4,   1,  23,  23,  66, -27,   4,   5,  -7,  -5,   0, -47,   0,  -1,  16,   8,  -4,   3,   7,  20, -45, -11,   5,   4,  10,  -2,   5, -16,  15,   1,  -8,  -6,   5,   8,  -5,  17,  -4,  -1,   5,   5, -12,  -8,  -4, -10,  -3,   1,  -2,   3,   6,   3 },
      { -28,   2, -32, -15,   5,  -2,  -4,  -2, -14,  28,  -8,  16,   0,   0,   6,   6, -51,  12, -36, -18,  31,  10,   4,  -6,  29, -25,  43, -38, -31,   2,  -5,   6,   8,  24,   0,  17,  -5,  -5,  -5,  -7,   4, -10,   2,  -4,   6,   0,   8,   6,  10,  -3,  -5,  -6,  -5,  -2,  -3,   3, -11,  -1,   5,   4,  -1,  -2,  -5,  -4 },
      {  10,  20,  -1,  45,   0,   8,   3,  -5, -24, -32, -45, -26,  15,   1,   4,   5, -29,  29, -10,  -2,  13,  -7, -10,  -1,  13, -18, -16,  12,  14,   9,   9,  -1,  21,  18,  26, -45, -28,  -4,  -1,   4, -11,  -1,  18,  13,  -4,   1, -12,  -8,  -5, -11,  -4,   5,   1,   1,  13,   9,   9,   7,   1,   1,  -2,  -6, -12,  -6 },
      { -10, -40, -13, -15,  -1,   0,   1,  -2, -15, -30,   6, -20,  18,  17,  -8,   7, -12, -13,  13,  -5, -21,   4,  -1,  -2,  54,  12,   7,  66, -30, -12,  -3,  -1, -11,  29,  -2,  10,  17, -15,  -1,  -2,   2,  -3,   2, -15,  -2,  13,  -9,  -4,   0,   4,  -1,   2,  -8,   3,   9,  -2,   7,   7,  -1,   1,   6,  -2,   0,   2 },
      {  -9,  -4, -87,  -6, -11,  -6,  11,  11,  14, -12, -41,   2,  24,   9,   0, -11,  21, -10,  34,  -2,  16,   6,   1,   5,  -7,  40,   0, -22,  -3,   5,   1,  -4,  -7, -18,  15,  -3, -13, -11,   0,   5,   2,   1, -10,   5,  -3,  -4,  -6,  -4,   3,   3,   4,   3,  -3,   0,  -1,   2,   6,   0,   4,   3,   6,   4,   0,  -1 },
      {  13,   0,   2, -27,  23,   1,  -6,  -1,  22, -16, -12, -11,  -1,  -9,  -4,  10,  19, -44,   1, -18,  -5,  22,   0,  -6, -36, -25, -22,  -8, -29,  13,   1,   8,  38,  48,   6,  14, -13, -25,  -5,  -9, -10,  -5,  19,  27,   9,   5,  -9,   2,   4,  13,  -2,  -3,   2,  12,   4,  -9,   3,   1,  -1,  -3,  -4,  -7,   3,   1 },
      {   5,  21,  17,  28, -34,  -1,   1,  -9,  30,  22, -12,   7,  12, -24,  11,  -1,  37, -33,  -2,  -5,  24,   2, -13,   4,  12,  23,  55,  17, -11,   0,   7,   3,   6,  21, -16, -29,  15,   6,  -6,  -2, -18, -17,   9,  -8,   0,  12,   2,   1,   4,   3, -22,  -3,  -3,  -1,  11,  12,   1,  -2,   1,   8,  -4,  -4, -11,  -7 },
      {   0,   8,  11, -39,  34,   8, -21, -16,   5,  11,   3,  30,  -1, -12,  27,  19,  -2,   0,  24, -44,  23,  27,  -4,  -9,  11,   3, -15,  17,  -6,  -3,  -2,  -1, -14, -24,  22, -34,  -8,   1,  -1,  11,  -2,  28,  12, -11, -31,   9,   4,  -9,   3,  -9, -10,  17,   4, -16,  10,   5,  -1,  -8,   3,  -1,  -2,   0, -10,   6 },
      {  -5, -12,   0, -61, -41,  26,  30,  -4,  -7,  -3,   2,   9,  14, -17, -12,   6, -15, -12,  -3,  13,  23, -29, -22,  -4,  -6, -14,   4,  10,  10,  28,  23,   5,  -8,  -8, -22,  -5,  -8,  13,  11,  -2, -15,   8,  18,  11,   1, -26, -20,   1, -13, -16,   3,  16,  12,  10,   5,  -4,  16,  10,  -3,  -5, -10,  -4,   2,  -3 },
      { -11,  -5, -19,  -6,   8,   8,  13,   3, -11,  18,  -3,  47,  12,  -8,   0,  -6, -22,   2, -18,   4, -11,  15,  -7,  -1, -24,  26, -19,  30,  31, -40,  10,   6,  62,   2,  -2,   0,  21,   5, -10,  -7,  -6, -26,   0,  -8, -13,   8,   1,   4, -17,  -3,  -9,  -8,   2,   6,  -3,  -2,  11,   1,  -4,   7,   4,   1,  -1,   2 },
      {  -9,  -9,  18,   7,  20, -22,  13,  15,  -2,  -4,  34, -30,  21,  18, -13, -18, -35,  -4,  20, -18,  -3,  13,   3,   4, -26,  22,  37,   0, -20,  -4,  13,   4,  27, -39, -20, -16, -24,  -1,  -4,  -5,  -8,  12,   3,  22,  -1, -18,  -3,  -3,  15,   1, -19,  -7,   2,  12,  13,   8,   6,  -2,   5,   1,  -5, -10,  -5,  -7 },
      { -10, -12,  -8,  38, -36,  10, -26, -19, -11, -19, -12,   9,  15, -17,   0,  19, -34, -36,  -3,  -4,  14,   7,  23,  -3, -50, -23,   0,  18, -10,  -4,  -8,  -3, -10, -21,  -1,   6,  26,   7,  -2,   1,   4,  27,  -3, -13,  -5,  -1,  -1,   0,  28,  12,  10,  -4,  -4,  -9,  -2, -12,  -4,   3,   1, -10,   0,   3,  12,   8 },
      {  -2,  -9,  -8,  -5,  26,  21,   5,  12,  -8, -21, -24,  -1, -51,  29,  22, -13,  -4, -39,   1,   1,  -2,  -8, -13,  -7, -16,  -7,  27,  18,  29,  -5,  -1,   0,  -4,  16, -25,  -5, -11,   9,   4,   1, -10,   9, -18,   1,  -8, -13,  25,   4,   8, -22,   3,  -8,  -2, -25, -10,  17, -22,  14,  24,   8,   8,   0, -14,  -7 },
      {   6,  -2, -12,  -1,  29, -13, -10,  -1,   8,  10, -25,  -2,  25,   4, -10,   9,   3, -24, -39,   8, -39,  15,  22,  -2,  16,  -7,  11,  -8,  15,  -2, -11,   1, -23, -10, -37, -41,   0,  12,   8,  -7,  22,  -4,  31,  11, -31, -21,  -8,  10, -11,   8,   6,   7,   7,  12, -14, -18,   7,   5,  -1,   0,   0,   0,  14,   6 },
      {  -7,   4,  -6,  14,  34,  -7,  -3,   0,  -8,   3, -16,  37, -51,   9, -27,   0,  -4,  -2,  -1,  -3,  18, -33,   9,   7,  -8,  10,   7,  12, -10,  17,   7,  -4, -15, -16, -15,  -5,  18, -11,  -4,  -3,   8,   0,  22,  13,  24,  27, -22, -18, -12,  18,   9,  -2, -12,  21,  38,   1,  22,   2,  -2,  -7,   4, -15,  -3,   5 },
      {   5,  -5, -18,  13, -15,  -2,  10,  23,   8,   2,  11,  27,  27,  11, -10, -19,  12,  -7, -22, -19, -39,   4, -20,  -3,   1, -30, -29,  -9, -17,  10,  14,   9,  -8,  -6, -15,  -7,  12,  10,  -7,  -5,  12,  23,  -3,   3,   9,  20,  12, -15,   1, -27,  -6,   1, -12, -15,  29,  30, -19,  -2,  -3,  -9, -11, -18, -32, -12 },
      {   2,   7,  -2,  -7, -22, -16,   8,  -2,   6,  14,  19,  31,   6,  41, -21, -17,   8,   4, -15, -28,  11, -21,  28,  10, -20, -15,  -6,  34,  -7, -13, -13,  -4, -23,  27,  28, -30, -29,   3,   4,   2,   7, -21, -18,  18,   3, -18,   9,   3,  16,   8,   1,  -4,  17,  -2,  -8,   0,  13,   1,  14,   0,   5,   8,   6,   0 },
      {   5,  -4,   4,  -3,  24, -41,  -1,   4,   7,  -8,  -5,   5,  27,  23,   8,  -6,  13, -20,  -1,  18,  26, -18,   0,  19,  -5, -40,  22,  12,   4,   9, -19, -15,  33, -14,  10,  15, -10,  10,   7,  13,  15,   4,  -7, -44,   6,  11,  -4,  -7, -30, -21,  -4,  24,  -4,  -1,  -5,   4,  11,   0, -19, -10,  -9,   9,   3,   4 },
      {  -6,  -7,   6, -28,  14, -20,  20,   9, -13,  -4,   0, -33,  20, -19,  -5,   6,  -4,   2, -10, -11,   7,  13,  -3, -14, -34, -10,  21,  -6,  43,   1,  16,   2, -36,  14,  17, -13,  24,   6, -13,   7,  14, -27, -29,   5,   0,  38,  -3,  -7,  -4,  24,   3,   5, -25, -10,  15,  10,   3,  -5,   2,   1,   1,   1,  -6,   0 },
      {  -4,   1,  -4, -15, -16,  11,  -9, -10,   0,   3,  12, -15,  -5,   8, -14, -10, -14,   3,   4, -25, -24, -37,  -6,  28, -12,  27,  11, -20,   3,  31, -23, -26,  36,  25,  11, -16,  16,  20,  20,   4,  15,  11,  24,  -8, -33,  17,  10,  -1,  -2,  16,  20,   8, -11, -24,   6,  -5, -12,  -4,  -4,  -1,   5,  -4,  -3,  10 },
      {  -2, -12,  -7,   5,   5,   8, -18,   1,  -1, -13,  13,  30,  -1, -41,   9,  11,  -3, -11,  17,  -7, -46, -15,  16,   0,  18, -10,  37,   0,   8,  25,  -9,   2,  14, -14,  40, -13,  -8,   5, -13,   1,  -9, -15, -39,  28,  14,  -2,  -8,  -2, -15, -10,  17,  -3,  15,  15,  -2,   1, -10,  16,   0,  -1, -14,  -4,   2,  -8 },
      {  -2,   0,   0, -14, -10,  -4,  -9,  -4,   2,   5,  -1, -15,   1,  -3,  47,  -6, -12,   2, -17, -28,   3,  -2,   5,  17,  -6,  16, -29,  10,  -4,  15, -14, -13,   5,  -8, -45, -32,  -1, -23,   4,  15,   7, -13, -21, -16,  47,   6, -19,   3,  -3,  -3,  28,  -7, -10,  19, -19,   4, -15,  -5,  -7,  15,  -8,   7,  -3, -15 },
      {   6,  -9,  -7,  -8,   9,  11, -15,  -3,  11,  -8,  -2, -22,  12,  -5,  -4,  -9,  17,  -4, -15, -22,  37, -27,  18,   5,  36, -14, -23,   7,  14,  -1,  -9,  10,  25, -21, -18,  12,   5,  46, -23, -13, -12,   6, -27,  27,  -9,  -2,  19,  -2,   2,  20,  -4, -30, -15,   3,  15,  -8,   7, -12,   3,  16,   1,  -6,   3,  11 },
      {   2,  -1,  -2,  -8, -16,   4, -10, -15,   5,   4,   3,   9,   4,  19,   8, -12,   3,  -7,  -2, -15, -26,  -5,   0,   2,  14, -13,  19, -14,  58, -19,  17,   6,   2,   7,   3,  12, -33, -21, -13,  14, -14,  31,  15, -31,  14,   9, -20, -18,  36,  19,   2, -13,   0,  26,  20,  -5,   4, -12,  -8,  11,   0,  -4,   8,  18 },
      {  -2,   0,   2,   7, -12,  46, -24,  -8,  -2,  -1,  10,  -7,  16,  10,  12,  10,  -3,   2,  11,  -2,   8,   5,  20,  -7, -15,  17,  11, -11,  11, -23, -28,   1, -11,  22, -25,  13, -27,  -2,  -2, -11,  39,   1, -10,  18,  11,  19,  13, -12, -42, -24, -11,  17,  15,  10,  20,   6,  22,  -6,   0, -10,  -6, -16,  -8,  19 },
      {   3,   2,  11,   9,   7,  44,   0,   3,   7,   5,   9,   6,  11, -11, -23, -16,   8,   6,   2,   5,   6,  20,  13,  15,   4, -14,   5,   3, -11,  11,  17,  10,  11, -12, -16,   4, -32, -17, -19,  -5,  13, -17,  -6, -22, -17,  -1, -11,  -8,  -1,  19,  48,  20, -20, -27,   5,  13,   9,  29,  13,  18,  31,  23,  -4,  -7 },
      {   2, -10,   0, -12, -25,  -4,  26,   4,   3,  -8,  -5, -23, -37,  -6, -18,  -1,   9,  -8,   6, -19,   3,  -2,  41,  -4,  15, -17,  -1, -26,  -2, -55,  -8,  -1,  21, -17,   0, -26,  20,  -5, -15,  -3, -12,   1,   9,  -5,   3,  20,   1,   5,   6, -19,  14,  33,  11,   1, -10,  10,  12,  16, -10, -20,  -1,  -3,  -5, -13 },
      {   2,   4,   4,   8,  41,  16,   3, -12,   2,   2,   9,   1,  46,   3,  -9,   8,  -1,  -2,  16, -10,   6, -53,   0,   1, -14,   9, -10, -24,  -4, -13,  -3,  15, -22,  14,  -3,  -3,  16,  -3, -11, -10, -28,  -8,   7, -28,   1,  14,   1,   8,  12, -33,   2,  -6,   0,  21, -10, -13, -17,  35,  21,   7,   7,  -5,   6, -10 },
      {  -1,  -8,   2,  -4, -10,  22, -17, -16,   1,  -9,   4,  -4,   3,  36,  34, -10,  -1, -12,   8,   0,  -4, -21, -22,   8,  -5, -23,   9, -12, -17, -25,  28,   9,  -9, -30,  22,  -3,  13,  -3, -18,   7,  20, -31,  37,  28,  -2,  11,  -4,   9, -24,  11,  -3, -20,   0,  -9, -16,   2,  -5,  -9, -22,  18,   9,  17,   0, -13 },
      {   2,  -1,   4,   2, -19,   0,  -3,   1,   1,   0,  -2,   3,  -3,  25,   7,  24,  -5,  -5,  -9,  -4, -13,  -7,  -2, -24,   3,  -8, -11, -11,  -3,  30,  -5, -11,   6, -30, -19,  11,  -8, -24,  21,  19, -27, -41, -22,   0, -38,  40,  15, -14,  16,  10, -31,   9,  13,   4,  -9,   0,  25,  22,   6,   8,  10, -13,  -3,  10 },
      {   3,   4,   6,   9,   7,  31,   4,  19,   3,   3,   1,   5,  -3,  24,   3, -18,   1,   0,   0,   2,  10,  14, -27,  -1,   6, -10,   9, -14, -20,  -6,  27, -11,   7,  17,   1, -29,  28,   2,  11,  10,   4,  21, -55,   9, -18,   6, -16,   5,   7, -11,  12,  14,  -3,  29,   2, -33,   8,  -8, -21,  -6,   5,  -2,  39,  18 },
      {   0,  -1,  -4,   0,  19,  35,  -5,  18,  -4,  -8,  -3,  -9,  -1, -13,  13, -30,   0,  -1,  -3,  -3,  -7, -10,  34,  20,  -6,  -9,  -4,  -9, -12,   0,  16,  -6, -12,   6,   9,   4,  25,  -5,  15,  24, -10, -23,  -3, -38, -12, -42,  -2,   3,  -2,  22, -21,   0,  19,   7,  11,  21,  21, -21,  -1,  -6, -31, -20, -27,   5 },
      {  -1,  -4,   4,  -2, -22, -16,  -8,  16,   1,   1,  10,  -4,  -5,   9,  19, -23,   6,   8,  -3,  -2,  21,  26,  30,  25,   0,  -8,  -1,   5,  23,  22,  -2, -14,  -3,   3,  10,  16,  31, -24, -13,  -2,   4,   0,  19,  17, -35,  -7, -18,  -1, -14, -37,  -6,  -8,   1,  25,  25,  -7, -27,  31,  25,  19,  11,  -8,   2,   2 },
      {  -3,   1,   1,   4,  -3, -17, -20,  -1,  -2,   7,   6,   9,  14,  20,  21,  11,  -2,   4,   3,   0, -10,   1,  26, -13,  -3,   2,   0, -11,  -4, -18,  13,  -8,  -3,  15,  -9,   9,   6,  19,  23,  16, -68,   0,   0,  22,  13, -20, -22,  -8, -31,   2,  18,  26, -33, -28,  18,   2,   0,  -5,  -1,  -1,  24,   5,  -4,   6 },
      {   2,   0,   0,   0,  14,  12,  56, -27,  -2,  -2,  -2,   9,  12, -26,  18, -11,   0,   0,  -4,  11,  -6,   6,   4,  42,   3, -17,   5,   0,  -8, -22, -18, -35, -11, -12,  -1,   3, -12, -16,  36,  -1, -12,  14,  -2,  23,   9,  24,   8,  23,  -1,   8,  -8, -21,  -2, -16,  16,  -8,  -2,  12,  -5,  10,  12, -10,  -3,  23 },
      {   0,  -3,   0,  -1,   4, -13,   2, -73,   1,   0,  -4,  -3,  -4,  20, -18,  19,   6,   2,   9,  -6,  -3,  24,  15,   9,   2,   5,   0,  -9,   5,  19,  46,  -9,   6,   7,  -7,   1,  11,   5,   2,   9,  10,  -7, -17,  -8,   2, -16,  22,   3,  -3, -33,  19, -34,   4,  -8,   6,   9,  24,  10, -18,  -3,  -7, -15,  -3, -11 },
      {   2,   2,   0,  -2,   6,   3, -16,  19,   4,   4,   6,   4,  17, -17,  -1,  -9,   1,   4,  -3,  -3,   3, -25,  19, -10,  12,   3,   5,   2,  17,   8,  27, -13,  12,  10, -10,  -4,  16, -49,   0,  24,  13,  36,   4,  23,  12,   2,  19,   5,  -3,  -3, -27, -16, -14, -30, -27,  -7,  29,  12,  10, -32,   5,  31,   5, -10 },
      {  -2,  -3,  -1,  -3,  -1,  13, -18,  18,  -3,  -2,   6,   5,  -5,  -4,   5, -32,  -3,   3,   7,   1,  -3,  25,   1,  -1,  -5,   0,   2,   3,   8,  16, -26,   0, -17,   2,  14,  -1,  -7,  25,   7, -13, -17, -10,  25,  -1,  11,  11,  12,  17,  25, -25,  -3,  -8, -53,   8, -18,  -6,  53,  17, -37,   9,  -6, -12,   3, -16 },
      {   8,   2,   5,   0,   1,  15,   7,  -1,  10,   2,   6,   7,   6,  15,   5,  31,  12,   0,   7,  -1,   7,  -5,   3, -31,  10,  -2,   0,  -2,  11, -11,   3, -17,  24,  -9,  16,   2,  13,  10,  37,   2,  49,  -1,   1,   4,  22, -27, -20,  20,  25,  18, -12,  14, -38,  17,   0,  13, -10,  29,  19,  14,   4, -25, -20,  -6 },
      {  -3,  -2,  -4,  -6, -20, -15, -22,   3,  -3,  -2,  -2,  -7, -22,  -6, -11,  19,   0,   0,  -1, -13, -17,  -3, -29,  24,  -9, -13,  -5, -15, -10, -17,   3,  -1,  -6,  -5,   6, -19,  -9, -13,  -7, -29,  -5,  15, -22, -32,   0, -18,   2,  17, -47,  10, -31, -10, -31,  -2,   8,  -8,  27,  17,  35,  21,  -1,  -8,  14,   0 },
      {   0,   3,   3,   7,  -1, -19,  20, -11,   0,   4,   1,   5,   4,  -5,  27,  -2,  -1,   1,  -2,   6,  -4, -27,  11, -25,  -9,  -1,   4,   4,  -6,   7,  -8,  19,  -2,  10,   5,   6,   7, -28,  -2, -30,   2,  41, -15,   4, -39,  -6,   6,  15,  -9,  10,  25,   6,  -2,  24, -10,  22,  28, -24, -22,  30,  25, -10, -28, -23 },
      {   1,   0,  -2,   4,   6,  23,  28, -42,  -1,  -2,  -2,   4,  -6,  12, -11,  -6,  -3,  -2,   0,   1, -16,   3,  19,   7,   0,  -3,   0,  -1,   6,  26, -16,   3,   3,  -5,   0,  -1,  12,  -8, -29,  -5,  -3,  -5,  -7,   5,   9,  -7, -16,  -3,   8, -17, -49,  42, -31,  -7,  -9, -12,  -6, -51,  16,  15,   1,  18,  12, -13 },
      {  -1,  -1,  -3,  -2,   3, -31,  -8, -16,  -1,  -3,  -5,  -3, -13, -37,  28, -14,  -5,   1,   1,  -4, -16, -26, -16,  13,  -4,   3, -10,  -8, -15, -18,   4,   7,  -2,  15, -14,  12,  -5,  19, -11,  15,  32, -15, -10,  12, -12, -23, -14, -20,  37, -21, -22,  22,  -1,  -2,  24,   5,  23,  26, -13,   2,  28,  22,   4,  15 },
      {  -1,  -2,  -4,  -3,  -6,  -6, -12,  -2,  -1,  -6,  -4,  -5,  -9, -26,  -9, -27,  -1,  -4,  -2,  -5,   0, -15,  34, -28,  -2,   4,  -3,  11,  -7,   1,  39,  22,   2,  -7,   4,  -5, -30,  -2,  46,  -7,  13, -13,   3, -18,  -3,  20,  14,  21, -13, -19, -17,   2, -14,   5,  13,   0, -38, -19, -13,   8,  21, -11,  37,   4 },
      {   0,  -1,   2,   1,  -1,  -3,  10,  11,   1,   0,   2,  -4,   1,  19,  21,  10,   0,   1,   1,  -1,  -3,   1,  10,   6,   2,   4,   2,  -2,   5,  25,   5,  54,   9,  -9,  11, -18,  25, -13,  10, -38,   1,  -6,   5,  -6,  30,  -7,  13,  12,  12,  -8,   6,   6,  10, -22,   8,  -2,  14,   8,  -1,   5,  -1,  19,  -8,  74 },
      {   1,  -2,   2,  -4,   4,   8, -13,  27,   0,  -5,   1,  -3,   5, -15, -35,  21,   1,  -7,   0,  -7,   2,   3,  -6,  -1,  -1,  -9,   4,  -5,   3,  -4,  -2, -44,  -1,  -3,   5,  -9,  14,  -5,   8,  -8,  -2,  -6,   7,  -4,  22, -14,   6, -30,   7, -30,  -3, -16,  21,   4, -12,  17,   6, -34, -13,  46,  54,   8, -10,  18 },
      {   1,  -3,   1,  -1,   3,  -8, -28,  -7,   2,  -2,   1,   2,  10,  -3,   2,  -7,   2,  -4,  -2,   1,   8, -12, -29,   6,   7,  -3,   0,   4,  19, -12,   0, -30,  -1, -14,  -2,   3,   0, -45,  -2, -48,  -6, -36,  -4,  10,  -7, -23,  16,  14,  25,  -8,  23,  24, -12, -13,  22,  -8,  -7, -26,  -3, -40,  -5, -26,   1,  -4 },
      {   1,   3,   2,  -4,   3,  13,   0,   8,   1,   2,   1,   0,  -2,  10, -10,  21,   2,   0,   1,  -6,   6,  -1,   5,  14,   1,   0,   1,  -7,  14,  -4, -28,  30,   1,   1,  -4,  -3,   6, -15,  33, -27,  -1,  -3, -19,  -2, -18,  -8, -50, -21,   4,  -2, -23, -38,   1, -27,  14,  19, -11,   7, -43, -19,   1,  16,  13, -35 },
      {  -1,   3,  -1,   0,   0,  10,  24,  20,  -1,   5,  -1,   2,  -1,   8,  30,  33,  -1,   3,  -1,  -3,  -5,  -2,  23,  24,  -3,   6,  -3,   7,  -1,  14,  28, -24,  -6,   7,   2,  -1, -11,  14, -34, -30,  -2,   5,   6, -10,   5, -11,  32, -25,  -3,  25, -20,  12,  -9,  14, -13,   9, -21,  20, -37, -26,  23, -16,  24,  -4 },
      {   2,   1,   2,   4,  -3,  -1,  24,  35,   3,   0,   3,   2,  -3,   1,  34,  35,   3,   0,   3,   1, -10, -12,  21,  -4,   0,   1,   0,  -2, -10,   2,  19, -20,   4,   3,  -2,   6, -17,  21,  -8, -12,  15, -10,  10, -13, -12,  19, -26,  15,  18, -26,  17, -28,   4, -25,  21, -26,  30, -36,  33,  -2, -10,   2,  12, -17 },
      {  -1,  -1,  -4,   0,   8,  -9, -27,  -4,  -1,   0,  -1,   1,   3,  -1,   2,  -8,  -2,   5,   1,   3,   4,   3,   6,  35,   4,   2,  -3,   3,   9,  -5,  38,  -6,  -3,   9,  -4,  13,   6,  15,  26, -44,   0,  15,  -5,  18,   7,  32, -19, -12,  17,   9,  -6,  35,  29, -16, -36,  18,   8,   5,  15,  34, -29, -11,   9, -16 },
      {   1,  -2,   0,  -4,   5,   5,   3, -15,   1,   0,  -1,  -4,   8,   7,  21, -35,  -1,   1,  -1,   0,  -5,  11,  -8, -38,   3,  -4,  -2,  -8,   2,  15, -21, -14,   6,  -5,   7,  -6,  23,  11,   3, -18,  -1,   4,   8, -15,  30,  -4,  11,  -3,  -5,  21,  -2,   1,  44, -11,  34,  14,  20,  14,  22,  -4,  28,   2,  48, -34 },
      {  -1,  -2,  -2,  -3,  -2,   2, -17, -10,  -1,  -2,  -1,  -4,  -2,   3,   0, -10,   0,   0,   0,  -3,  -2,  13,  -2,  17,   2,  -3,   0,  -1,   6,  13,  13,   3,  -1,  -2,   2,   0,  -6,  26,   7, -19,  -4,   2, -10, -10,   7,  19, -23,  33,  -1,   5, -17,  -4,  18,  25, -12, -18,   1,  -7,  14, -52,  58,  14, -56,  -8 },
      {  -2,   1,  -4,   3,   4,   3, -13,   8,  -1,   1,  -1,   2,   0,  10, -15,  16,  -4,   0,   0,  -4,  -1,  -2,   2,  -7,  -2,   6,  -6,   0,   4,   7,   3, -33,  -5,   5,  -9,   7,  -7,   3, -22,  -1, -10,   5,  -5,   2,  -4,   8,  -4,  76,   6,   4, -12,  14,  15,   7,  33,  36,  -8,  10, -31,  18, -21,  50,   3,   5 },
      {   0,   1,   1,   2,   2,  -7,  -2,   0,   0,   1,   1,   1,  -1,   1,  -6, -19,   0,   2,   1,  -3,   5,  -6,   4, -28,  -1,   6,  -1,   1,  -1,   8,  -3, -11,  -2,   4,  -6,  -3,  -7,  -3, -41,   1,   5,   3,  -9,  -9,   7,  -4, -44,  21,  -7,   3, -15,  -3,  20, -54, -20, -26,  -8,  16, -31,  14,  10, -59, -11,  21 },
      {   0,  -1,   0,   0,   1,  -1,  -5,   8,  -1,   1,   1,   2,   6,  -1,  -2,  19,  -1,   1,  -1,   1,   1,   2, -11,  35,   3,   0,  -1,   2,   6,   4, -11,  31,   2,   0,  -2,   4,   4,   0, -11,  38,  -1,   2,   5,  11,   0,   5, -10,  37,   8,  -7,   6,   2,   0,  -1, -16,  60,  11, -20,  18, -18,  32, -46,  43,   9 },
      {   0,  -1,  -2,   0,   1,  -1,  -7,  21,  -1,   0,  -1,   0,   1,   3,  -5,  24,  -1,   0,  -2,   2,   3,   3,   0,  32,  -1,   1,   2,   4,  12,   0,  -1,  33,   2,   0,  -1,   2,   5,  -7,   8,  31,   5,   4,  -2,   3,  17,  -4,  35,  17,   8,   5,  -3,  30,  22, -13,  33, -54,   0,   0, -20,  30,  16, -15, -12, -49 },
    },
    {
      { 100, -59,  -3,   2,  -3,   1,  -1,   0,  -4,  30, -24,   5,  -1,  -1,   0,   0, -27,  17,   8,  -6,   2,  -1,   0,   0,  -2,  -5,   9,  -2,   0,  -2,   0,   0,  -3,   0,   4,   3,  -3,   0,   0,   0,  -3,   0,   1,   2,   2,  -2,   0,   0,  -1,  -1,   1,  -1,   2,   0,  -1,   0,  -1,   0,   0,   0,  -1,   1,  -1,   0 },
      { -21, -32,  18,   4,   2,   1,   1,   0,  55, -46, -28,  23,  -3,   2,   0,   2,  -7,  44, -51,  -3,   9,  -4,   2,  -1, -12,  12,  22, -40,   6,  -1,   2,   0,   5,  -9,  15,   4, -20,   1,  -1,   2,  -6,   5,  -1,   6,   4, -10,   0,   1,   2,  -3,   4,  -2,   3,   1,  -6,   1,  -1,   1,   1,   1,  -1,   2,  -1,  -2 },
      {  42, -10,   0,  -2,   2,   1,   1,   0,  19, -58,  47,  -5,   2,   6,  -1,   0,   6, -41, -23,  41, -13,   9,   0,   2,  -9,  20, -48,   1,   8,   3,  -1,   0, -10,  16,  -1, -20, -10,   7,   3,  -2,  -4,   4,   2,   5, -13,  -8,   9,  -2,  -2,   1,   0,   1,   3,  -8,   0,   3,  -1,   4,   0,   0,  -1,   1,  -1,   0 },
      {  11,  25, -21,  11,  -5,   2,  -2,   0, -72,  16,   7,  22,   3,  -5,   2,   1,   3, -12, -43,  16,  22,  -6,   3,  -1,   2,   7,   3, -56,  27,  -4,   3,   0,  16, -22,  25, -11, -22,   7,  -3,   4,   0,   5,  -4,   7,   1, -10,  -3,   2,   6,  -6,   8,  -3,   4,   3,  -7,   0,   1,   0,   0,   1,  -2,   2,   0,  -1 },
      { -24, -41,  46,   5,  -5,   0,  -3,   0, -24,   9, -10,  34,  -9,  -7,   1,   1,  -1,   8,  -8,  34,   0,  -8,  -2,  -1,  -3,  13, -24,  26,  38, -21,   6,  -4,  11,  -5,  -2, -27,  50,   5, -10,   4,   3,  -1,  -2, -17,  -3,  32,  -7,   0,   0,  -1,  -3,  -5, -11,   6,  12,  -4,   2,  -4,   0,  -3,  -2,  -4,   3,   3 },
      { -22, -54,  43, -13,  16,  -3,   3,   1, -51, -11,  -3,   4,   3,  14,  -1,   0,  18, -17,  -8, -28,  11,   5,   4,   0,  27, -20,  -7, -10, -27,  27,  -8,   4, -11,  26, -23,   0, -29,   4,  11,  -4,   6,   0,   6,   0, -12, -18,  11,   1,   0,   7,  -3,   6,   0, -12,  -4,   4,   0,   3,   0,   2,   2,  -1,  -4,  -2 },
      { -26, -13, -17,  17,   6,   5,   2,   0,  29,  37, -21, -22,  12,   3,   0,  -2, -19,  -5,  22,   9,   3,   5,   2,  -1,   0, -21, -23, -16,  33,  10,   0,   1,  -1,  -5,   6, -43, -23,  40,   1,   1,  -2,  -1,  -2,  12, -46,  -6,  26,  -4,   0,  -1,   1,   0,   4, -29,   6,   9,   1,  -1,   0,   2,   0,  -2, -10,   5 },
      { -29, -37,   4,  22,  -9,   6,   0,   1,  13,  31,  -8,  -5, -10,  -9,   3,   1,  -1, -38,  22,  28, -18,  -5,  -1,   1,  -9,   7, -25,  -6,  -9, -11,   7,   0,  11,   0,  22,  15, -27, -37,   4,   0,  -2,   4,   2,  20,  38, -33, -24,   3,   4,  -4,   6,  -6,  19,  22, -20,  -7,   1,   1,   0,   0,  -1,   8,   5,  -7 },
      {  -5,  10,  31, -13,   8,  -9,   3,  -2,   6,   9,  56,   2, -16,   2,  -8,   0, -65,  23,   6,  38,  12, -10,   3,  -3,  16, -47,   6,  -7,  15,   9,  -2,   1,   2,  -5, -24,  21, -15,   4,   3,  -1,   1,   2, -13,  -4,  17,  -8,   1,  -1,   1,  -1,   1,  -5,   0,   8,  -3,   0,   0,   0,  -2,  -1,   0,   1,   3,   1 },
      {   0, -15,  27, -20,  -6,   1,   1,  -1, -15,  13,  36, -26,  -1,  -6,   6,   0,  40,  26,   5,   7, -35,   8,  -2,   1, -22,  21,  33,  16,   8, -16,  12,  -4,   9, -31,   7,  19,  -6,  18,   2,   5,  -9,   8, -25,  22, -10, -19,  26,   1,   2,  -9,   2,  -9,  22, -26,  -7,  16,   0,  -2,  -3,   1,   1,   6, -17,   0 },
      {  -8,  15,  -9,  19, -21,  15,  -1,   2, -32, -41, -36,  30,   2,   3,  11,   2, -60,  -8,   9,  -1, -27,  14,   1,   0,  -4,   1,   4,  33, -12, -15,   9,  -1,  20,   6,  -8,   9,  -4,   6,   3,  -3,  -5,   8,  -5,   5,  -3, -22,  28,  -3,   5, -10,  -4, -16,  16, -20,  -7,  15,   3,   1,   1,  -5,  -4,   4, -10,  -4 },
      {   4, -15,   9,  20,  -8,   8,  -5,   0,  17, -19,   5,  17,  11, -17,   3,   0,  15, -57,  25,  -5,  21,  -3,  -7,   0,  22, -19,  34, -14,   8,  -4,  -7,   2,  14,  -9,   5,  36,  -2,  24, -29,   2, -18,  20, -22,  11,   5,  36,   8, -16,   2, -10,  -2, -11,  -4,   0,  28,  -2,  -2,   2,  -4,  -5,  -6,  -3,  -2,   9 },
      {  16,  15,   8,  32, -14, -10,  -1,  -1,   8, -14, -17,  23,  17,  -8,   0,   2,  49,   5,  17,   6,   7, -22,   2,  -3,  -7, -34,   5,  14,  39,   3,  -1,   6, -19,  -7, -45,  -9,  -7,   9,  12,   5,  -2,  -1,   4, -21,  14, -34,  -3,  11,   4,  -6,  10,  -8,   3,  11, -34,   8,  -3,   5,   0,   5,  -7,   7,  -7, -13 },
      {  10,  33,  25, -27,  -2,   5,  -1,   1,  30,  48, -19,  50, -34,   1,   0,  -4,  13, -24, -18,  15,  20,   1,   2,  -1,   4,  22,  -8,   1, -13,   9,   3,   0, -16,  22,   2,  20,   0,   5,  12,  -8,  -3,  -6,   0,  -6,   4,  -9,  34,  -1,  -3,  -8,  -2,  -8,   7, -23,   2,  20,   0,  -4,  -2,  -5,   2,   2, -13,   4 },
      { -23, -46, -68,  26,   5,   3,   5,   2,   2,  11,  34,   5,  -6,  -8,  -1,  -1,  -4,  -2, -22,   3,  16,   5,   7,   1,   5,  10,   9,   0, -10,  -7,   0,   1, -20,   0, -19,  23,  31,   4,   9,  -3,   5,  -9, -13, -36,   8,  -2,  22,  11,   3,  -3,   7,  -7,  -9, -13, -16,  17,   0,   2,  -3,  -2,  -1,   3, -14,  -7 },
      {  -5,  -6, -31,  23,   5,   3,  -1,   0, -20,  -2,  -9,   2, -26,  27,  -6,  -2,  26,  37,   7,  30,  -9, -11,  -1,  -5, -29,   1,  -4,   8,  11,  25, -10,  -1,  -4,  39, -12,  21, -26,   9,   3, -11,  -8,   8,   2,  17,  20,  12,  13, -22,   1,  -3,  -4,  -9,  -1,   7,  46,  -7,   0,  -3,   2,  -4,   2, -16,  13,  21 },
      {   2,   2, -18, -17,   4,   1,   4,  -1,   4,   2, -23, -19,  -5,   6,  -2,   3,  17,  12, -28,  22, -36,  -3,   8,  -2,  88, -17,   2,   6,  17, -37,  -5,  -1,  -5,  30,   5,   8,   7,   7, -18,   2,  -8,   8,   0,  10,   4,  -6,  -1,  -5,  -5,   3,   1,  -4,  11,   4,  -7,  -2,  -2,   1,   1,  -2,   0,   4,   2,  -7 },
      {   7,   4,  -6,  24,   2,   5,  -3,   2,   9,  11,   3,  20,   7,   3,   6,  -7,  22,  13, -32,  10,  -7,  25,  -4,   3,   5, -54, -25,   7, -37,   4,  28,  -5,  17, -33, -19,  -6,   5, -40,   5,  12,  -9,  20, -13,  13, -11,  10, -14,  -2,  -6,   6,   5,   4,  10, -25,  21,   0,  -2,  -3,   4,  -6,  10,  -4, -14,  17 },
      {  -5, -10, -20, -50,  18,   3,   9,  -2,   6,  14, -23,   1,  -3,  15,  -8,   2,  -3, -42, -37,   8, -23,  10,  11,  -1, -40, -20,  23,  18,   7,  -3, -13,   2,  18, -21, -34,   3,  -6,  20,   4,  -3,  -1,  20,  -6, -13,  -9,  -8, -14,   0,   1,   1,   1,   3,  -9,  22,   0, -16,   3,   6,  -1,  -7, -10,  -6,  22,  -1 },
      {  -4, -10, -19,   7, -24,   4,  -2,   4,   6,   0,  32,  53, -38,   2,  -8,  -3,   3,   1,  14, -13,  -4, -20,  14,  -1,  14,   3,  23,  16,  -4,  -9, -18,   2, -15, -10,   3, -20,  -5,   9,   6,  -1,   7,  -3,   9,  28, -38, -15, -21,   1,  -7,  15, -15,  28,  29,   1,   2, -27,   6,   1,   2,  -9,  12,   9,  19,   5 },
      {  -2,   6,  19,  34, -15,  -6,  -2,   0,   7,  26,   7,  -2,  51,  10, -14,  -3, -18,   0, -33, -22,  -6,  13,  -5,   0,  24,  45, -11,  21,   7,  13, -11,  -8, -27, -18, -21,  11,  -7,  17,  15,  -4,  -1,  10, -14,   5,   8, -12,  -5, -12,  -6,   0,   7,  -8,  17,  21,  23, -15,   1,  -4,   1,  -3,  -8,  -6,  17,  12 },
      {   4,  -4, -16, -22, -22,   5,  -3,   3,   4, -15, -14, -12, -36,  -6,  11,  -1,   7,  10,  22,  -2,  22,  -9,  -6,   2,  36,  34, -25,   0,   6,  31,   5,  -2,  41, -27, -21,   0, -10,   1,  22,   6,   5,  27, -42, -15,  -3, -10,  -5,  10, -11,  19,   2, -14, -12,  -2,  -2,  -8,   5,  -8,   6,   6,   1, -13,   8,  -3 },
      {  -3, -11, -22, -35,  37, -21,   7,   3,   2, -14,   0,  11,  15,   6,   8,  -7,  -6, -11,  13,  -9,  10, -24,   7,   4,  15,  10,   9,   8,  25,   4,  12,  -8,   0, -17,  -4, -18,   5, -23,  18,   9,  19,  -1,  22,  16,  23,  -7,  -4,  -2,   3, -11,   4, -17,  33,   3,  31,  27,  -2,  -3,  -2,  -9, -14,   8, -30,  42 },
      {   5,   1,   4,  19,  42,  -3,   3,  -3,  -1, -11, -13,  17,   7,  31,  -7,  -4,   9,  -5,  20,  35, -20,  22,   0,  -9,  25,  -1,  21,   3, -11,  22,   2,  -6,  -4, -40,  39,   0,   4,   9,  35, -12,   8, -33, -12,  -8,   6,   2,  13,  21,  -4,   7, -23,   5, -19,  11,  -6,   2,  -1,   5,  -2,  -4,   4, -10,  17,  -4 },
      {  -1,   7,   2,   1,  28, -19,   2,   3, -13, -14, -31, -10, -14,  10,   1,   3, -11,  -9,   6,   5,   7, -16,   3,   2, -19,   4, -18, -11,  10, -14,  -5,   8, -65, -23,  -9,  29,  10, -10, -21,   6,  14,   3, -47,  21,  -9,   8, -11,  -7,  -9,  29,  -4,   0,  15, -20,  -9,  -9,   1,  -1,   7,  -4,  16,   6,  -8,  -8 },
      {   0,   1, -14, -10,  57,  -8,  -2,  -3,   0,   2,   6,  28,  27, -15, -10,  -2,   8,  24,  14,  -3,  29,  20, -20,  -1,  -6,   6, -27,  11,  -8,  -1, -16,   2,  33,  18,   1,   6,   7,  15, -15,  -9, -25,   7, -17,  14,  -5,  -5,   1,   1,  -1,  -4,  -9,  -7,  31,  11, -25, -12,   0,  -5,   2,  -8,  -4,  33,  18, -15 },
      {  -1,   5,   2,   2,  33, -15,   4,  -7,  -1,   3,   4,  27,  -6,   7,  -8,   0,  -4,   8,  37,   4, -31,  10,   2,  -5,  11,  31,   0, -29, -14, -23,  17,   1, -14,  -7, -19,  -2, -17,  -1, -13,   8, -20,  34,  16, -37, -15,  -3, -10,  -9,  14, -21,  45,  13, -17,  -8,   7, -14,  -4,   4,  -1,  18,   3,  -7,  -2,   5 },
      { -12, -10, -17, -51, -31,  27,  -6,  -1, -10, -11,  -9,  15,  25, -41,   7,   0,   0,  14,  14,  27,  13,  37, -14,   2,   1, -12,   0,   4,   1,  19,  18,  -7, -56,  -2,  15,   6,  -4,   3,   9,   7, -12,  -6,  -2,   7,  -3,  -2, -10,   1,   0, -11,  19,   1,  -1,   9,   6, -11,  -5,   1,  -2,  -1,  -6, -10,  11,   2 },
      {   3,   1,   3,   9,   6, -22,  11,  -8,  -2,  -6, -20, -16,  21, -49,  -4,   0,   0,  -4,  -8,  40,  18, -14,  -5,  -3,   3,  12,  12,  15,  -7,  -1,  -5,  -3,  27,   1, -10,  33,  -5,   5,   6, -10,  30, -16,  26,  -6, -17, -10,  -1,   1,   6,  12,  18,  42,  10, -23,   8, -13,   7,  -6,  -1,   2,  36,  -6,   4,  18 },
      {  -5,  -4, -14,  -4,   1, -22,  -4,   2,   1,  11,   4,  16,  10,  -3,  12,  -1,  -4,   5,  -8,  15,  -4,  -8, -37,   9,   8,  17,   6,  26,  -1,  16,  -2,   6,  -6, -17,  -9,   2, -37,  -8, -29,  12,  20,  22,  29,  17,  -2,  20,  -2,  -9, -17,  -6, -36,  10, -27,  -1, -10,  23, -10,   6,  -6,  -1,  -6, -10, -21, -51 },
      {   0,   4,   4,  14,   4, -26,  -4,   0,   2,   7,   0,   1,   0, -21,   4,  -1,  -7, -12,   2,  15, -19,  31, -13,   4,  -6,  11,  55, -22,   9,  26,  11,  -8,   0,  54, -14, -36,  17, -16,  13,   9,  12,  26, -27,   6,   2,   3, -10,   8, -21,  25,  -1,  -8,  12,  -1,  -5,   2,   2,  -8,   8,  -5,   8,   2,  -2,  -5 },
      {   2,  -3,  -6, -16, -19,   6,  11,   2,   9,   3,   7,  26,  23,  23,  -3,  -2,   9,   5,  20, -12,  -8,  40,  13,  -9,  -2,   6, -15, -33,  26, -26,  21,   2,  23,  14, -32,  22,  -2,  -4, -10,   5,  49, -35, -12,  15,   6,  -1,   1,  -8,  10,  10, -22,  19,  -3,   4,   4,   5,   1,  -7,  -2,   4,   5, -18,   2,  -6 },
      {   1,  -2,  -1,  -2,  -6,   5, -25,  -3,   2,  -3,   2,   5,  -8,   6, -44,   1,   8,   2,   4,  -1,   1,   9, -22,   6,  -4, -17,  -3,   3, -12,  -8,   4, -11,  -1, -12,  23,   0,  -1,  14, -11, -12,  63,  26,  19, -13,   5,   3,  21, -17,  -1,  44,  37, -27,  16,   9,  -1,   8,   4,  -1,   2,  15, -15,  16,   8,  -4 },
      {  -5,  -3,  -6, -11,  -6,  -6,  -8,  -1,  -1,   3,   8,  12,  37,  22, -31,   1,  -3,  -3,   5,  -3,  10, -27,   2,   3, -10,  -2,   2,  11,  19, -24,  26,  -5,   7,  22,  24,  26, -15, -13,  26,  10, -42,  12,   7, -15, -17,  -2,  -8,  17, -30,  49, -12, -10, -14, -16,  -1,   1,   0,  -5,   3,   6,   7, -20,  -6,   2 },
      {   0,   0,  -5,  -6,  -6, -57,  33, -13,   2,   3,   7,  28,   3, -43,  17,   5,   2,   7,  -4, -24, -34,  -8,  17,   3,  -7, -19, -27,   4,  -9,  -3, -25,   9,   1,   5,  30,  -1,  -8,  14,   3,  -7,   9,   0, -25,   6,  10,  -2,  10,   6,  -2,  15,  13, -12, -24,   7,   6,  11,  -6,   6,   5,   4,  -7, -28,   3,  11 },
      {   0,   1,   1, -10, -11,  10,   7,   4,   2,   5,  -1,   1,  28,  -4,  11, -15,   6,   9,   4,  27,  -1, -23,  -3,   8,   0,   8,  12,  -7, -40, -23, -13,  14,  -5,   4,  -6, -28, -33,  -2, -15,   2,  13, -20, -25, -46,  -2,   4,  -8, -10,  20,  12, -39, -39,  12,  -8,  12, -17,   2,  11,   9,  -7,  -7,  11,  -2,  16 },
      {   3,   4,   9,  20,  13,  -2,  42,  -4,  -3,  -3,   1, -17, -27,  -9,  28,  -1,   4,   3,  -1,  -7,  31,  37,  -2,  -1,  -3,  -4,   0,  11,   6, -37,  28,  12, -16,   0,   8,   6, -20,  10,   2,  12,  12,  26,  21, -20,  -8,  -9,   2,  13, -29,   1, -29,  -5,   6,  18,   6,   3,  10, -18, -12, -14, -16,  -3,  23,  19 },
      {   1,   0,   1,  -8,  19,  38,  -3,  -2,   5,   2,  13,   8,  20,  -8,  43, -18,   0,   0,   1, -18, -15, -23, -10,  -9,  -3, -11,  -5, -10,  16,  -4,  -3, -15, -11,  -4,   8,  -8,   9,  -5,   7, -31,  11,  36,   8, -13,  32, -10,  24,  -9,   0,   4, -12,  13,   0, -20,   4, -18,  18, -21,  26, -10,  34, -15,  19, -20 },
      {  -3,  -6,  -7, -12, -30, -53,   0, -17,  -3,  -6,  -7,  -9, -12,  -8, -34,   6,   1,  -5,   5,  -4,  -7,  14, -15, -14,   1,  -5,  -8, -28,  -2,  -1,   9,  -8,  -3, -26, -18,  -9,   9,   0,  -4, -24, -31, -14,  29,  -9,  22,   2,  19,  -8, -21, -11, -30, -18,  18, -18,   8, -16,  13,   1,   8, -23,  15,   6,  12,  -9 },
      {   0,   2,   4,   1,   7,  31,   7, -15,  -2,   2,   1,   1,  18, -13,   5,   7,   0,   7,  10,  18,  -3, -19,  58, -11,  -1,  15,   2, -15, -28,  11,  11,  22,   5, -11, -25,  -7,  12,  10, -21,   2,  -3,  -5,  21,  23,   6,   2,  12,   1, -60,  21,  18, -13,  -4,   8,   0,  12,   2,   3,   9, -17, -14,  -2,  10, -10 },
      {   4,   1,   0,  10,  35,   7, -27,  -8,   3,   0,   9,   3, -21, -29,   3,  13,   0,  -8, -14, -24, -11,  -2,   1,   1,   4,   0,   4,  20,  20,   1,  33, -12,   3,   7,  -4,   7, -29,  -6,   2,  10,  12, -52,   0,   0, -14,   7, -11,  -9,  -5, -14,  16, -44,  -8, -15,   6, -31,  13,  -2,  20,  -8,  -5, -10,  10, -23 },
      {   0,  -3,  -2,  -5, -15, -27,   4, -15,   4,  -1,   1,  -4,  18,  31, -13,   6,   4,   0,   2,  11,  37,  -4,  38, -29,   0,  -2,  11,  23, -12,  -4,  14,  12,  -2,  11,  20, -14,   4, -11,  -2,  -1,  31,  31, -16,  16,  -5,  -3,   3,   6,  14, -51,  11, -21, -10, -14,  -5, -20,   8,  -5,   8,  -5,  16, -11,  13, -21 },
      {  -2,  -4,  -3,  -5, -20, -16,  26,  10,   1,   1,   2,   8,   9,  41,  42, -13,   0,  -4,   1,   9,  -9,   1, -23,   9,  -3,  -7,   2,  -6,   3,  12, -20,   1,   7,   1,   7,  21,   7,   4,  -2,   0,   7, -21,  12,  -9, -13,   9, -11,   7, -46, -17,  31, -16,   5, -14,  -5, -22,  -3, -35,  25,  26,  -8,  33,  -1,   2 },
      {  -1,   0,   2,   0,  -2,  -7, -31,  18,  -1,   1,   0,  -4,   0,  -5,  14,   8,   0,   2,   2,  -3,  -4,  -3,  14,  -9,  -4,  -2,  -8,  -9,   4,  -6,   2, -11,   5,   3,   3,  13,  -2,   0,  16,  -8,  19,  11,   1, -18, -14,   9,   5,  -1, -22, -17,  -8,   3,  16,   5,  -4,  -6, -98,  26,  11, -27,   6,   2,  10,   6 },
      {   1,   2,   1,   0,   3,  16, -27,  11,  -1,   0,  -2,   2,   8, -20, -28,  14,   1,   6,   2,  -1, -29, -11, -10,   1,  -1,  -2, -14, -23,   0,  -5, -19,  23,  -6,   4,  -8,  20,   1,  -2,  24, -10,  28,  22,   3,   8, -19,  19,  -9,  45, -16, -41, -23, -11, -14,   0, -12,   2,  36, -10,  -1,  -9,  -9,   8,  -8,  32 },
      {  -1,  -2,  -4,  -4,   5,   5, -26,  21,  -1,  -3,   0,   4,  -2,  -3, -19,  25,   0,  -1,  -3,  -7,   4,  21,  16,  -7,  -2,  -2,   3,  25,   7,  -2, -13,   8,   1,  -2,  14, -24, -28,   6, -32,   5,   2, -10, -26, -27,  45, -10,  -4,  -1, -51, -14,   0,  33,   0, -15,   5,  -2,   3, -16,  -5,  21,  22,   3, -20,  21 },
      {   1,   2,   2,   1,  -4,  15,  23, -15,   2,   4,   4,   5,  -1,   3,   0,  25,   1,   4,   2,  -6,  -8,  37,  28,   0,   5,   8,  -1,  12,  33,  36, -18,  26,   1,  -6,  13,  16, -10, -11, -17,  -3,  -2,  16,  26, -16,   6,  17, -15,  20,  31,  27,  -2, -19,   5, -19,  -7, -10,  -5,  20,  23, -22,  22,  24, -21,   9 },
      {   0,  -1,   3,   1,   6,  -8, -25,  30,   1,   2,   3,  -4,  -3,  -2,  11,   1,   1,   0,   3,  10,  25,  30,  -5,   6,   4,   6,  15,   4,   8, -29, -35,  20,   0, -11, -15,  -8,   6, -33,  -5, -24, -12,  -3,  16,  28,  -5, -23,  17,  -1,   2,  12,   2, -18, -28, -10,  -4,   0,  -2,  21,  42,  28, -13, -26,  18,  22 },
      {   2,   2,   5,   3,   4,  30,  42, -16,  -3,  -5,  -1,  -3,  -7,  -3, -44,   5,   2,   0,   8,   2,   9,   6, -21,  15,   0,   0,   5,  -3,  -2, -19, -43,  -3,   3,  -3,   0, -11,   0, -12,  14,   3,   1,  -6,  -7,  -1,  -6, -10,  -9,  -9, -11, -15,  15,  -4,   2,   4,  12,  10, -27, -33,  29, -35,   3, -34, -29, -24 },
      {   0,  -1,   2,  -1,  -3, -27, -55,  19,   1,  -2,  -1,  -4,  -1,  11,  27, -23,   2,   4,   6,   7,   6,  22,  30,  25,  -2,  -2,  -5,  -4,  -9, -12, -24,   5,   3,  -2,   4,   7,  -5,  26,  10,  10,  -8,  -1,  15, -12,   2,   7, -22,   5,  11,   8,  16,  -6,  18,  10,   6,  19,  26, -26,  -3, -42,  13, -24, -13, -24 },
      {   2,   3,   4,   3,   3,  -1,  10,  34,  -2,  -3,  -2,  -2,   2, -20, -12, -24,   3,   5,   5,   3,  -6,  10,  35, -22,  -3,  -5,  -2,  -6,  17,  19, -29, -25,   3,  -1,  17,  18,  13, -36,  -5,  -3,   5,   9,   2, -23, -26, -26,  -1, -27, -13,  -4, -16,  -4,  -1, -10,  25,  -9,  18,   9, -30,   5, -36,   7,  -9, -29 },
      {  -2,  -2,  -1,  -3, -11,  -5, -13, -37,  -1,  -1,  -2,  -1,   2,   0,   8, -12,   1,   0,   0,  -1,   2,   4,   7, -20,   3,   3,   4,   9,  16, -17, -12,   2,  -2,  -2,  -4,  -5, -37, -46,  16, -21,  -8,  -8, -24, -10, -18,  36,  29,  -9,  -6,   4,  23,  31,  -5,  23,   0,  28,  17,  -5,  12, -19,  -9,  42,  10,  -1 },
      {   0,   1,   1,  -2,   0,  10,  12, -10,   1,   2,   2,   3,   5,  16,  13,  10,   2,   4,   5,  -4,   6,   3, -18,  36,   0,   0,   2,  -3,  11, -11, -20,   2,   6,   3,  -2,  -7,  -2, -16,   8, -13,   7,  -5,  -5,  -1,   1,   7,   0,  -3, -31, -10,  27, -16,  10, -13,  -5,  -1,  17,  63, -65, -14,  31, -18,  15,  -3 },
      {  -1,  -3,  -2,  -1,  -1,  -6, -29, -36,   2,   2,   3,   2,   5,  12,   7,  18,   1,  -1,   0,   0,  -9,  -9, -20,  19,   1,  -4,   0,  -6,   7,  24,   0,  37,   4,   1,  17,  11,  27, -20, -28,  39,   4,  -3,  -4, -20, -21, -51,  18, -27,  -1,  -5,  -8,   8, -11,   9,   5,   3,   3, -12,   1, -31,  -1,  -3,  19,  21 },
      {   0,   2,   2,   1,  -4,  -6,   2,   5,   1,   2,   1,  -1,   1,  14,   9,  16,   2,   5,   6,   3,  10,  -1, -16,  -6,  -3,   0,  -4,  -1,  -9, -11,   6, -43,   3,  10,   6,   4,   4,  24,  -2,   9,   9,   3,  -8,   3,  11, -21, -12,   9, -10,   0,   9,  27, -25,  -5,  12,  -8,  30,  56,  43, -45, -27,  18, -24,   4 },
      {  -1,   0,   0,   2,   3,   6, -26, -77,  -2,  -2,  -3,  -3,  -9,  -4,  -2, -66,  -3,   0,  -2,  -3,   5,  14,   5,  -2,  -2,   0,  -2,   1,   4,  -9, -18,   4,   1,   3,  -1,   3,   5,  12,   6,  14,   7,   3,   4,   7,  17,  -9, -11,  18, -13,  -3, -13,   1,  -5, -17,   8, -22, -16,  19,  -3,  26, -13,   0, -22,   3 },
      {  -1,  -1,  -1,   0,   1,   4,  -4, -22,   0,   1,   2,   1,  -2,   4,  18,  20,   0,   0,  -1,   2,   2,   4,  -4, -48,  -2,  -2,  -2,  -2,  -7,   8, -11, -18,   1,   0,   5,   8,   6,   3,   2,   6,  -3,   1,   5, -14, -23,  12,  -8, -33, -12,  -6, -13, -10,  38,  34, -33,  11,  24,  19,  21,  30,  15, -44, -30,  17 },
      {   0,   0,   0,  -1,  -1,   4,   7,  31,  -1,  -1,   1,   1,  -4,  -2, -11, -61,  -2,  -2,  -3,  -9,  -5,   2, -23, -54,   0,   3,   0,   6,   3,   3,   6,  15,   1,   1,   1,  -3,  -2,   6, -14,   7,  -1,  -8,   1,  10,  -3, -10,  -5, -11,  -7,   6,  19, -22, -34,  12, -22,   2,  -4,   3, -13, -38,  39,   7, -11,  27 },
      {   0,   0,   1,  -1,   0,   1,  13,  28,  -1,  -1,  -1,  -2,  -2,  -5, -19, -27,  -3,  -2,  -3,  -5,  -7,  -6,  -9,  -8,  -1,  -2,  -5,   0,   5,   6,  -4,  36,   3,   1,   6,   3,  -1,  -3,  11,  63,   0,  -1,   1, -12,  14,  22,  21,  15,   0, -11,   9,  19,  39, -11,   7,  11,  12,  36,  35,   9,  -4, -14,  35, -19 },
      {   0,   1,   0,   0,   1,  -5,   8,  16,  -1,  -1,  -2,  -2,  -1,  -6,  -8, -17,   1,   2,  -1,   0,   1,   3,  12,  45,   2,  -2,   1,   2,   3,   1,  33,  48,   4,   1,   8,   1,   6,   7,  30, -24,   1,   6,  -3,   1,  -1,   5,  18, -45, -11,  -8,  -5,  14,  -4,  31,  -8, -40,   9,  16,  17,   5,   5,   8, -46,  -6 },
      {   0,   0,  -1,  -1,   1,   2,   3,   1,   0,   0,  -2,   2,   1,  -9,  -3,   5,   0,   0,   2,   1,   2,   5,  19,  25,  -3,  -1,  -5,  -4,  -6, -11, -16, -40,  -3,  -5,  -7,   2,  -1,  -1,   9,  48,   0,   5,   7,   8,  -9,  -8,  24,   1,  -4,  -8, -14, -23, -26,  35,  21,   7,   3,   5,  12,  26,  64,  40,   2,  -6 },
      {   0,   0,  -1,  -3,   2,  -1,  -9,   4,   2,   2,   1,   1,   3,   0,  19,  42,   0,   0,   0,   3,   2,   3, -15, -43,  -1,   0,   0,  -4,  -7, -19, -25,  21,   1,  -2,  -5,  -8,   2,  -1,  39,  41,   0,   0,  10,  13,  17,   8,  29,  -6,   8,  24,   8,  -1,  -7,  -3,   8, -49,  -6, -13, -34,  -7, -14,   9, -25, -12 },
      {   0,   0,   0,   1,  -2,  -4,   7,   5,   0,   0,  -1,   2,   2,   4,  -9, -18,   1,   1,   2,   2,  -1,  -2,  17,  32,   0,   2,   1,   3,   2,  10,  -8, -47,  -1,  -2,  -5,  -3, -11, -18, -25,  25,   0,   2,   5,   4,   5,  22,  45,  10,   4,  -1,   2,   8,   3,   3, -53, -43,   3, -15,  -6, -22, -20, -28,  -4,  32 },
      {   0,   0,   0,   1,   0,   1,   5,  -6,  -1,  -1,   0,   0,   2,  -4,  -6,   3,   1,   2,   3,   2,   1,   6,  15,  10,   2,   2,   0,   0,  -1,  -4, -12,  -7,   1,  -1,   3,   1,   0,  13,  38,  22,   3,   7,   7,   6,  20,  15, -33, -71,  -1,  -5, -11,   0, -23, -44, -40,  22,   2,  -6,   3,  -4, -10,  18,  32,  20 },
    },
    {
      { -99,  -8,  28,   5,   4,   4,   1,   1,  50, -25, -30,   8,   0,  -1,  -1,   0,  11,  25,  -7, -18,   0,   3,  -2,   0,   3,   3,  14,   0, -11,  -1,   1,  -1,   1,   1,   0,   8,   5,  -2,  -2,  -1,  -1,   0,   0,  -1,   2,   4,   1,   0,   1,  -1,  -1,   0,   0,  -2,   1,   0,  -1,   0,   0,   0,   0,   0,   0,   0 },
      { -49,  61, -29,  -5,   6,  -6,   3,   0,   3, -11,  53, -19, -11,  11,   2,   3,   4, -27, -20,  40,  -3,  -7,   7,   2, -10,   2, -22, -12,  16,   6,  -1,   1,  -1,  -3,  -9, -19, -22,  -2,   9,   2,   4,   1,   3,  -4,  -2, -23, -12,   0,   3,   1,   3,   4,   5,   6,  -8,  -6,   3,   1,   0,   0,   2,   0,   4,   1 },
      { -16, -24,  18,  18, -10,  -1,   4,   2,  47,  55,   5,  -8,  13,  -8,  -9,  -1, -37, -54,  32,  17,   1,  10,   5,   1,   5,  -9, -29,   9,  19,  -1,  -5,  -1, -17,  -9,  20, -13,   2,  17,   0,  -2,  -2, -10,  -2,  11, -11,   5,  13,   2,  -4,   1,   0,  -8,  -3,  -5,   0,   3,   0,  -1,   3,   0,   0,   2,  -1,   0 },
      {  -2, -15, -38,  23,  15,  -2,  -2,  -1,  20,  53, -24, -52, -15,  10,   3,  -1,  19,  48,  23,   9, -13, -23,  -5,  -1, -20,  -4,  15,   7,  22,   7,   5,   0,  13,   1, -22, -16,  -3,  -2,   6,   4,  11,   9,  -1,  -3,   1, -14, -16,  -2,   3,   0,   2,   9,   9,   6,  -1,   0,   2,  -2,  -3,   0,   0,  -2,  -1,   1 },
      {  -6,  25,   4,  -6, -15,  -2,   5,  -3, -18, -32, -10, -27,  22,  11,   7,   6,  -4,  -8,  57, -16, -51,  -1,   8,   3,   3, -28,  -7,  54,  -2, -16, -10,   0, -15,   0, -28,  13,   7,   0,   5,   0, -13,  -9,   3, -12,  -4, -12,   4,   3,   4,   0,   6,  -6,  -6,  -5, -12,  -1,   0,   3,   4,   2,   1,   2,   1,  -6 },
      {  35,  50,  -7,  -8,  -2,  -1,   1,   1,  65, -14,  -4,   5, -17,  -3,  -5,  -1, -69,  32,   0,  11,  12,   2,   0,  -2,  -5, -20,  15,   1, -10,  -3,   4,   0,   0,   0, -11,   5,  11,  -9,  -6,  -2,   0,   0,   8, -13,   8,  10,   3,  -2,   0,   0,   2,   1,  -5,   0,   5,  -1,  -2,  -2,  -2,   1,   0,  -1,  -1,   5 },
      {  14,  42,  48,  53,   5, -11,   0,  -2,  -1,  13,   4, -33,  17,   4,  -2,   3,  14,  -9, -37, -11, -16,  -2,   0,   1, -20,  10,  44,  -3,  12,  -9,  -7,  -3,  -7, -39,  14,  14,  11,  16,  -4,   2,  -6,  -1,  -7,  -2,   1,   9,   7,  -1,   1,  -9,  -2,  -4,  -9,  -3,  -1,  -1,  -4,   3,   3,  -1,   3,   3,   0,  -3 },
      { -26,  35, -43,   2,   2,  15,  -2,  -2, -20,  34,  19,   2, -10, -25,  -9,  -3,  18,  -4,  25,  -4,  21,  19,  -8,  -3,   6,   2,  15,   8, -15, -14,  11,   2,  14,  11,   3,  24,  47,   8, -18,  -7,  -4,   2, -11,  -9,   0,  38,  28,  -1,   0, -11,   0,  -1,  -7, -20,   9,  13,  -1,   2,  -6,  -2,  -2,   2,  -9,   0 },
      {   2,  -1, -13, -11,  -1,  -2,   1,   1,  20,  41,   5,  -6,  -2,   2,   4,   1,   8, -38, -16, -33, -10,   8,   8,   7,  -5,   2,  11,   3, -52,   3,  -5,  -3, -17,   6,  -7,  41, -11, -49,   3,   0,   2,  -6,  11,  -2,  39, -15, -24,  -5,   4,  -3,   5,   2,   7,  22,  -5,  -7,   1,   1,   2,  -2,  -4,   0,   8,  -2 },
      {   3, -64,  -2,   3,  -4,  -2,   2,  -4,  18, -24,  56,  -8, -14,   2,  -3,   6,   6,  10, -13,  46, -13,   5,   2,  -2, -21,   6,   5,  16,   5,  -8,   2,  -2,  -9, -17, -21,  36,  19,  -6,  -1,  -1, -15,  -3,  -8, -25,  16,  12,   9,   3,   5,  -6,   4,  -7, -16,  -3,   1,   2,  -2,   6,   2,  -5,   3,  -3,  -1,  -1 },
      {  -7,   8,  29,  22,   1,  -4,  -4,  -3, -21, -14, -12, -26, -13,  -3,   1,   1,  -4,  -5,  28,  32,  -2, -19, -12,  -4,   7,  -8, -27, -14,  -1,   6,   0,  -2,  23,   9,  18,   6,   1, -39, -18,  -1,  -7,   6,   8,  19,  39,  44, -25, -13,  -1,  -1,  -3,  -3, -17,  25,  35,  -4,  -3,  -1,  -2,   0,   1, -12,   2,   5 },
      {   2, -13, -54,  34,   4,  11,  12,   1,   3, -32, -15, -36,  21,  -1,  -9,  -1, -27,   3,  -3,  -6, -21,  10,  10,   1,  28,  22,  -7, -44,  -9, -15,  -5,   3,   0,   8,  34,  24, -19,  11,  -4,  -6,  -8,   3, -13,   5,  11, -12,  18,   5,   2,  -2,  14, -11,  -4,  -6, -10,   2,  -2,   2,  -4,   7,   0,   1,   1,  -2 },
      {  12,  17, -25, -40,  -7,  -8,  -3,  -5,  31,  -2,  -9,  17,   0,  15,   0,  -2,  42,   0,  14,   5, -11, -26,   7,   4,  18,  17,  16,   1,  29,  12, -23,  -2, -30, -15,  22,  22, -18,  19,   9,   0, -17, -17, -10,  19,   8,  17,   4,  -5,  -9,  -2,  -9, -13, -14,   9,  33,   7,   0,  -6,   6,  -1,  -2, -21,  -6,  15 },
      {   2,   7,  25,   3,   8,   3,  -6,  -8,  -2,   9,  14,   4,   2,  -6,  -6,   0,  18,  20,  28,   3,  20,  19,  -4,  -2, -19, -58,  -4, -33,  -5,  28,   5,   9,   0,   4,  -5,  33, -53,   7,   6,  -6, -13,   5, -13,   4,  14, -21,  28,   5,  12,  -8,  -4, -14,  -2, -13,  -7,  13,  -1,  10,  -3,   4,  -5,  -5, -13,  -8 },
      { -30,  -6,  -9,  10,  -4,  -5,   4,   6, -46,  24,  13,  31,   5, -11, -16,  -3, -42,  29,   7,  -5,   9, -17,  15,  -7,  21, -18,  26,  -1,  20,   0, -16,   0, -28, -29,   8,   9,  -7,   5,  14,   1,   9,  -6,  -1, -24,  21,  -2, -21,   8,   7,  11,  11,  -5, -16,  24,   2, -18,   3,   1,  -2,   5,   6,   0,  15,   7 },
      { -10, -21,  12, -37,  25,  -9,  -5,  -1,  -7,  -7,  26, -41, -21,  19,  13,   1, -32, -26,  11, -19,  12, -28,  -9,   7,  26, -13,  50,  -8,  -2,   9, -11,  -5,  29,   6,   5,  18, -10,  13,  -1,   0,   0,   4,  -2,   7, -25,   5,  -3,  -4,  -8,   0,  -9,  20,  -1, -12,   2,   1,   3,  -2,  -6,  -8,  14,   6,  -6,  -4 },
      { -16, -10, -25,  -6, -18,  -5,   7,  -7, -17,   6, -35,  12,  21,  24,  13,  12, -25, -27, -28,  -5,   3, -11,   8,   5, -30, -20,   5, -22,  23, -10,   1,   3,  10, -16, -41,   0,  -7,  -8, -18,   3, -16,  14,  20,  -9,   1,  12,  15, -24,   4, -16,   3,  -8, -10,  -7,  22,  27,  -2,  -1,   7,  -9, -11, -12, -24,   1 },
      {   0,  12,  31, -20,   3,   3,  -1, -17, -10,  21, -28,  10, -26,  -8,   0,   9,  -8,  -2,   0,  37, -12, -21,   5,  -2,  28,  31,   3, -19,  29, -23,   0,   4,  -1,  23, -19,  35,  17,   0,  -6,  -7,  -5,  -6,   3,   0,  13, -29,  16,   0,  21, -14,   2,  -3,  27,   6, -30,  14,  -7,   1,  -3,  -6,  -4,  15,   5, -11 },
      { -10,   6,  20, -34,   3, -12,  -4,   5, -18,  27,  -3, -16,  -6,  31,  16,  -6, -23,  24, -22,   1, -25,   8,  -4,   2, -14,  18, -28,  12, -15,  -4,   9,  -3, -10,  15,  21,   6,  -6,   5,   6,   6,  38,   5, -40, -17,  10,   5,  16,  -9,   2,  20,  13, -24, -12, -23,  20,  13,  12,  -5,  -1,  10,   6, -14, -14,   3 },
      {   1,  -2,   0,  30,  39,  -5,  -7, -10,   2,  -2,  39,  13,   2,  13,  13,   0,  -6,  20,  -2, -25, -20,  -1,   0,   3,  42,  -8, -16,  -1,   7,   1,  13,  -4, -43,  -2, -16,  -4,  11,  -8, -13,   5,  17,   9,  18,  28,   1,   3,   3, -19,  11,  -4, -17,   5,  27,  -5,  21,  34,  -3,  -3,  -4, -11, -15,   0, -18,  -1 },
      {  -5,  -4,  -7, -26,  18,  -2,  -2,   0,  -3,  16,   7,  -5,   0,  -7,  27,  11,   2,  28, -18,  -1, -16,  -6,   9,  -2, -13, -30, -10,  -5,  -3, -38,  -2,   6,  -1,  -3,  44, -21,  -5, -13, -27,  -2, -27, -49,  22,   7,  -4,  13,   3,   7,   4, -11, -23,  -2, -11,  -4, -24,  -4, -10,   1,  19,  -3,   1,  12,   3, -24 },
      {  -8,   1,  13, -20, -38,  -5,  11,   4,  -9,  17,  25,  -9,  29,  -6,  -8,  -6, -19,  30,  13,   9, -14,  40,  -6,  -4,   8,  39,  21,   4, -11,  -7,  19,  -2,  26, -14,  -6,  -9, -26,  -2,  -5,  10, -25,  -4,   9,  24,  -5, -15,  -2, -15,  -1, -25, -22,   3,  19,  13,  17,  -2,  -7,  -7,  -6,  -5,  -7, -16,   1,  16 },
      {   0,  -3,  -4, -14, -17, -10,   5, -16,   8, -11,  25,  -3,  11, -28, -14,   3,  -7,   4,  -4, -41,  -7,  -8,   1,   3, -47,  -5,  -9,   4,  38,  -9, -17,   3,  22,  27,  12,  15,  13,  22, -16, -15,  26,   2,  -5,  30,  15,  -1,  -1, -13,  30,  15,   4,   5,  20,  27,   9,   2,   9,   9,  -3,  -1,  -5,   0,  15,  -2 },
      {  -1,  -1,  -1, -32,   0,  12,   2,  -5,  -2, -19,  -8, -20, -10, -27,   3,  10,   0,  -5,  32,  -4,  19,   7,  -8,  -2, -18,  23,  11, -12,   5,  -9,  16,   5, -34, -64,   5, -13,  -8, -25, -20,   5,  32,  19,  -9,  13,   5,   5,  -4,  -7,   0,   7,  11, -10,   5,   0, -22,  16,   7,   8,   1,   2,  -4,  12, -10, -24 },
      {   1,   0,   2, -16,  74,  33,  -8,  -9,  -4,  -1, -14,   5,  22, -16,  12,  -3, -11,  -9,   8,   9, -10,  13,   4,  11, -25,   8,  -6,  -7,  -6,   7,   0,  -5,   2, -19,  -8,   3,   6,  10,   5,   7, -25, -14, -17,  -9, -11,   3, -12, -11,  34,   4,  13,   2,  21,   1,  26, -24,  -5,   3,  -4,  -1,  -3,   1,  17,  31 },
      {  -8,  -3, -11,  29,  -1,  -5,  12,   3, -20, -16, -10,  15, -33,   5,  -7,   3, -17,  -9,  18,  15,   4, -12,   0,  -2, -51,  19,  27,  23, -13,  11,   1,  -8, -23,  17,  32,   5, -13, -11,  15,   0,   3, -26,  28,  19,  -6, -10,  15,   0,  14,   5, -15,  -4,  26, -24,   8,   7,  -5, -10,  -2, -10,  11,   3, -18,  17 },
      {  -3,  -3,  -9,  -6,  10,  19, -23,   3,  -8,  -9, -14, -31,   2, -30,   4,   0, -15,  -9, -35,  -4,  15,  12, -37,  -5,  19,   1,   8,  43,  22,  39,   6, -10,  -7,  13,  -3, -15,  -6,   4,  19,  -3,  -3, -23,   7,  -1,  29,  -2,  19,   5,  10, -17, -13, -25, -10,  18,  -3,   6,   0,  19,   3,  -5, -19, -13, -10, -13 },
      {  -1,   6, -11,  19,  10,  -9, -23,   1,   9,  -1,  12,  42,  18,  15,  28,  -9, -17,  -1,  24, -19,  -3, -22, -20,   8, -15,  34, -16,   9,  14,  16,   7,  -7,  43, -19,  11,  16,  -1, -13,  12,   7,  -8,  -1, -28, -11,  14,  -7, -13,  11,  -2, -19, -20,  -1,  -5,  -1, -23,  12, -10,  -4,   2,   4,   6,   4, -12, -33 },
      {   3,  -3,   2, -19,  29, -14,  22,  17,  10,  -3,   4,   3,  51,   3, -14, -30,  16, -10,   9,  23,  -1, -18, -19,  -6,  -2,  -8,  19,   1,  -8, -18,   9,   2,   8,  12,  21,  -7,   5,   7,   3,  10,  26,  32,  41, -26,  17,  -2,   8,   7,  16,   4, -13, -24,  -2,  16, -11,  12, -16, -11, -15,   6,  -2,   6, -13,   7 },
      {   2,   1,   0,   1, -11, -45,   7, -18,   2,  -7,  -4, -21,  32,  19,  10,  14,   5,   2,  17,  -2,  42,  17,   2,  -1,   3,  16,   1, -28,   5,  13,  20,   1, -25,  24,  -7,   2,  20,  -6,   2,   7,   3, -43, -10, -29, -13,   6, -20,   4,  36,  13, -20,   8,  -9,   7,  -1,   2,   5,  11,   7, -17,   4,   0,   0,   1 },
      {   4,  -1, -10,   4,  -6, -21,  -5,   5,   8,  -7, -17,   8, -22,  -5,  -1,   1,  13, -16, -13,   3,  -2,  27,   8, -12,  26, -32,   5,  32,  28, -22,  34,   2,  23, -16,  38,  12, -16, -14,  -6,  25,  -5,  13, -19, -20,  -3, -16, -15, -11,  30,  12,   8,  15,  21,  -3,  15,   4,  -6,  12,  -7, -17,  11,  -1, -12,   6 },
      {   4,   1,  11,  -5,  44, -13,  34,   8,  -2,  -3,   0,  -3,  -4,  -1, -30,  -2,  -2,   1,   6, -23,  18,  15,  12,  -3, -14,  21, -10,  16,  36, -16,   2,  14, -18,  29,  -2,  -2, -20, -10, -13,  -5, -13,   2,  -3, -17,  14,  -2, -11,  -4, -41, -47,  13,  19, -15,  -7,   1,  -4,   5,   3,  -9,  -2,  16,  -8, -13,   2 },
      {   3,   0,  14,  -7, -14,  11, -10,  -9,   5,   0,  13, -12,  19, -28,  21,  19,   6,  11,   5,  -6,   6, -19,  33,   7,   5,   4,  -8,  -9,   0,  -2, -11,  -9,  19,  -8,  19, -10,  10, -18,  41,  -6,   9, -13,  21, -22,  14,  -8,   8,  22, -12, -14,  33,  24,  10, -16,  15,  28,   8,  17, -14, -11,  -9,   3, -31,  33 },
      {  -2,  -4,  -7,   4,  15, -15, -22,  -3,  -4,  -9,  -7,  14,  21,  15,   7,   8,  -4,  -9,   3,  34,  -2,  24,  26,   3,  -7, -23,  42,   5,  -7, -16,  -1,   3,   6,  21,  -4, -17,   5,   3,   5,   7,  26, -18, -44,  37,  20,   8, -13,  10, -36, -14,  14,  -7,  14,  17,  -9,  19,   1,   5,  -2,  -7, -20,   8,   5,  -1 },
      {   2,  -1,   6,  -4,  22,   0,  -8, -17,  -1,   0,   7,   1, -34,  40, -36,   4,   1,  -6,  15, -27,  -7,  28,  36,  -2,   0,  13,   8, -10,  14,   2,  15,   1,  33, -11,   3, -22,  16, -10,  19,  -1, -16,  -6,  26,  -1,   9,   1,  17,  17,   9,  26,   3, -36, -15,  15,  -6,   5,  15,  -3,   1,   9, -22,  -9,   3,  -2 },
      {   6,   0,  11,  14, -10,  50,  24,   2,  12,   2,  21,  13,  -3, -17,  10,   1,   5, -13,   8, -11, -12, -26,  -2,  -9,   7,   8,  31, -11,   4, -29,  12,  11,   2,  34, -20, -34, -23,  -8,   1,  20,  -6, -15, -30,  -8,  14,   4, -13, -10,  17,  27,  -5, -17, -16, -14,   5,   1,   7,  11,  10,   2,   6, -11, -21,  -2 },
      {  -1,   0,  -5,  -4,   3,  -1, -32, -25,   3,   4,  -5,  22,   6, -21,  -8,  16,  -1,  -7,   4,  12, -58,  28, -17,  -3,  -6,   3,  21, -23,  13,  18,  17,  -7,  -6,  17,   5,   2,  -7,  -2,   5,   3,  26,   2,  33,  -4, -14,   9, -10,  -4,  -2,  -3,   7,  30, -40, -11,  -2,  -6,  18,   4, -13,   2,  25, -11,   0, -17 },
      {  -2,   2,  -3, -17,   4, -12,  30,   5,  -5,   5,  -5,  -5,  19, -17, -51,  -8,  -6,   6, -23,   4, -15, -24,   3,   6,   3, -13, -10,  -1,  -1,  28,  30,   3,   7, -10, -11,  15,  24, -14,  19,  22, -14, -15, -13,  46,  -4,   9, -22,  19,   1,  11,  10,   1,  -2, -29, -15,   6,  -5,   4,  11,   4,  10,   5, -26,  -4 },
      {  -1,  -3,   4,  -7, -13,  22,  12,  24,  -6, -10,  -6,  -7, -32,  -2,  -4, -30,  -2, -11,   4, -10, -21,   6,  -7,   6, -18, -23,  -7, -37,   1,  -6,  11, -15, -15,  -8,  10,   0,  37,  15,  15,  13,   7, -18, -19, -15,  -2, -23, -14,   4, -11, -24, -33,  21,  -5,  28,   4,  11,  -8, -15, -23,  -4, -22, -25, -16,   1 },
      {   0,  -2,   3,  -2, -21,  20,  -1,  10,   2,  -2,  15,   0,  10,  24,   5, -10,   6,   4,   3,  -7,   0,  -4, -21,  12,   0, -12,  14,  -9,  27,  30,  24, -18, -11,  17,  21,  -4,  13, -23, -20,  12, -18, -19,   2, -11,  -6,   0,   6, -30, -10, -11,  57, -16,  18,   1,  -9,   1,   1, -26, -14,  23,  -2,  29,  19,   6 },
      {  -1,  -5,  -1,  -1,   3, -61, -14,  22,   0,  -1,   2,   6, -22, -19,  -8,  -8,   2,  -7,   8,  -8, -16,  -4, -14,   4,   2,   8,   7, -12, -23,   3, -22,  -4,  -6,  -2, -13, -45,  -1,  10,  -6,  -8, -31,  -4, -18,  -2,  20,   3,  18, -12,  33,  -2,  33,  23,  -6,   4,  -9,  17, -22,   3,  -6,  23,   1,  -2, -12,   5 },
      {  -4,  -1,   0, -10, -14,  -3, -11,  -2,  -2,   3,   5,  -3,  -7,  -2,  19,   9,  -9,   3,   2,  -5,  -9,   8,  11,  12, -12,  -1,  12, -10,  10,  12,  -2,   2, -30,  29,  24, -12,  22,  -4,  12,   9, -37,  72,  -6,   3,  -3,   6, -26,  22,   7, -19, -15, -23,   6, -12,  -5, -14, -14,  31,  15,   4,  -9,   7,  -4,  10 },
      {   0,  -2,   2,  -9, -12,  39,  -2,  -7,   1,   0,   0,   0,   8,  49,  -9,  -4,  -1,  -4,   0,  -1,   3,  -8,  14, -19,   1,  -1,   3,  11,  -4,  15,  17,  29,  -6,  -9,   9, -17,   6,   2,  -8, -23,  19,   3,  -9,  18,  27,   2,  14,  12,  41, -31,   0,  43, -17,  -8,  -4,  -6, -33,  15,  -8,   1,  18,   0,  16,   9 },
      {  -2,  -4,  -2,   1,   3, -22, -22,  -8,  -1,  -5,   2,  -7,  -1, -34,  31, -12,  -1, -10,  -1, -12,  -3, -10,  23,  -9,  -2,  -1,   4,  -5,  13,   9,  37,  29,  -1,  -1, -14,   2,  12, -15, -16,   1,   8,  -3,  -6,   7,   7, -26,  30,  10, -29,  14, -21, -15, -13,   3,   5, -16, -27, -25,   3,  23,  30,  -5,   9,  42 },
      {   0,   1,   3,  -3,  -6,   8, -34,  47,   0,  -2,   1, -14,  -3,  -4,   0, -24,  -1,  -1,   0,  -9,   4,  18,  14, -15,  -8,  13,  -5,  -7,  27, -17,  -4,   3,  -4,  11, -21,  22, -28,  -1,  24,  15,  11,  -7,  17,   2,  -4,  38, -16,  14,   9,  12,   0,  13,  -1, -23,   4,  14, -41, -18,  10,  11, -24,  25,  17,  -4 },
      {  -4,   3,   4,   2,   2,  15, -31, -23,  -3,   4,   5,  -6,  12,   4,  -2,   5,  -6,   6,  -4,   3,  25, -15,   0,  -2, -13,   9,  -3,  16, -13, -19,  19,   5, -21,  13,  10,  10,   0,  15,  -3,   6, -36,  14,   7,  13, -15, -16,  -9,   8, -11,  37,  34,  18, -16,  29,  -3,  15, -32, -19, -21, -13,  -5, -35, -18, -25 },
      {   0,  -1,  -1,  -3,  -7, -12, -47,  10,   2,  -3,   3,  -2,  -2,   4, -32, -13,   0,  -3,   8, -12,  -7, -33,  -7, -17,  -7,   2,  -9, -16, -19, -14,  31,   8,  -5,  -3,  -6,  -7,  -1,  23,   0,   3,   1, -12,   8, -12,   6,  -6,  -1, -11, -18, -20,  11, -31,  18,  -9,  23, -25,  -1,  37,   3, -50,   8,  15,  10, -10 },
      {  -1,  -1,   1,  -4,  -9,  14, -39,  -8,  -2,  -2,  -4,  -9,  17,  -5, -43, -20,  -1,   2,  -8,   8,  -5,  -5,  26,  40,   1,   1,   4,   5,   5,  13, -25,   5,  -8,   5,   1, -17,  -8, -25, -29,  -5, -12,   5, -20, -27,   9,   7,   8,   5,   0,  18, -41,  12,  22, -10,   1,   4,  16, -23, -26, -11,  10,  14, -11, -12 },
      {   0,  -1,  -1,  -6,   0,   0,  17, -51,  -1,  -1,   3,  -3,  -1,  -6, -21,   8,   0,  -3,   0,  -7, -14,   6, -14, -22,   2,  -8,   3,  -8,   1,   7,  -5, -22,   2,   1,   6, -15,  -6,  -9,  20, -25,   8,   5, -17, -30,   5,  13,   0,  -6, -19,   3,   1,   0,  28,  -5,   2,   4, -58, -33,  46, -19,   0,  -7,   8,  -8 },
      {   1,  -2,   2,   6,  -4,  -3,  10, -23,   1,  -2,   1,  -1, -14, -19,  -2,   1,   0,   0,  -1,   4,  -7, -10,  38,   9,   4,   8,  -1,  14, -21,  38, -11,  47,  12,   3,  10,   3,  -6,  20, -21,  33,  21, -10,   7, -22, -23,   1, -21, -27,   7, -26,   7, -20, -13,   0,   0,   8, -28,  -1,  -8,   8, -36,   1,  -4, -15 },
      {  -2,  -1,  -1,   5,  -2,   7, -20, -13,  -1,   2,   2,   2,  -4,  14, -29,  19,  -4,  -2,  -3,  -4,  -8, -12, -31,  15,  -2,   1,  -6,   3,  -4, -16,  13,  13,  -3,   0,   0,  12, -22,  -2, -13,   3,   9, -11,  -4, -18, -29,  22,   2,  22, -10,  -5,  -8,  10,  21,  37, -12,  10,  -4,  36,  21,  51,  -8,   2, -22,  43 },
      {  -1,  -1,  -5,  -2,  -5,  -6,   4, -32,   2,   1,  -1,  -1,  -9,  -2,  -3,  30,  -2,  -3,  -9, -13,  -5,  -5, -32, -28,  -5,  -3, -13,   3,  -8, -11,   0,  20,  -6,  -4,  -3,  -6, -13,  10,  19,  22, -16,   7, -13,   4,   9,  14,  -7,  19,   5, -11, -12,  -7, -14,   0,  -4,  13,  18, -41, -51, -17,  -8,  48,  14,  29 },
      {   2,  -3,  -1,   3,  -4,   4, -26, -20,   2,  -3,   3,  -5,  -3,   1, -14, -18,   5,   4,  -4,  -1,  20,  12, -14,  19,   7,  -8,  -2,  -1,   2, -35, -31,   3,   2,  -1,  -6,  -3,  10, -24,  21, -13,  17,  -7,  -3,  21, -28, -23, -22,  -6,  29, -38,  12, -39, -20,  -3,  15,  -4, -11, -22,   2,  15,  35,   2, -23,   2 },
      {   0,   4,  -1,  -1,  -1,  12,  14,  16,   0,   6,   0,   4,  -8,  13,   9,  13,   2,   7,   1,  -6, -11,  31, -17,  23,   2,   3,  -6,  -9,  11,  16, -32,  17,   9,  -4,  -9,   2,   6,   3,  -8,   8,   6, -10,   8,   4,   6,   7,   0,  -2, -18,  31,   5, -17,   0,  18, -16,  -3, -50,  27, -26, -44,  38,  17, -28,  16 },
      {  -2,  -2,  -2,  -1,  -8,   1,  -3, -44,  -3,  -3,  -3,  -6,  -8,   9,   9, -24,   1,   0,   1,   1,  18,  14,  -6,  32,   1,  -4,  -5,   0,  -2, -17,   3,   3,  -3,  -6,  -1,  -2,   5,  40,  19,  44,  -6,  -6,  25,   3,  55,   5,  -5, -24, -14,   6,  -6,  20,  14, -18,  -7, -14,   4,   3,  15,  31,  19,  10,  -6,  -4 },
      {   0,   2,   0,   4,  -3,   1,  17,  -9,   0,   1,  -2,  -2,  -9, -11,  18, -39,  -1,   3,  -4,  -2,  -6,  -4,  31,  13,  -2,   1,  -4,   3,  -9,   4,  22,  -1,  -6,   6,  -6,  -2, -18,   0,  34,   5,  -5,  11,  -5,   0, -23,  44,  29,   0,  12, -19,   9,   1,   6,  56,  -8,  14,  16,  -1,   8, -20,  42,  15,   4, -15 },
      {   2,   0,   2,  -3,   4,  15, -17,  45,   2,  -3,   3,  -5,   6,   0, -13,  67,   2,   1,   4,  -1,   3,   3,  14, -14,   0,  -2,  -4,  -5,  -9,  12,  -3,  12,   2,   3,   3,   0,  15,  22,   4,  27,   1,   8,   2,  -6,  13,  -5,   2, -11,   9, -14,  -4,   0,  16,  21,   5,   7,  12, -42,  37,   8,  39,  -1, -17, -11 },
      {  -2,   0,   1,  -1,  -1,   1,  -2,  -3,  -1,   0,   0,   0,   3,   7,  17, -41,  -5,  -3,  -2,   6, -12,  17, -21, -44,  -2,   1,   8,  -6,   4,  13, -20,  71,  -2,  -4,   6,   3,  15,  -4,   1, -23, -16,  -6,   7,  -6,  -1,  -3, -14,  13,  -7,  12,  16,   9,  26,   5,  11,   2,  33,  -8,  22,   0,  -4,   5, -31, -10 },
      {   0,  -1,   1,   3,  -5, -11,   4,   2,   2,  -1,   3,   2,  -4,   1,  12,   8,   2,  -3,   3,  -7,  -2,  -3, -12,  30,   6,  -3,   2,  -6,   8,  -5,   5,  14,  -1,  -2,  -8,  -1,  -9,   0, -10,  16,  18,  -3,   3,   1,   7,  29,  13,  49,  18, -18,  25,  -4,  30, -11, -12, -55,   4, -35, -10, -20, -16, -50, -12,  -5 },
      {  -1,   0,  -3,  -2,   3,  -2,  -5, -10,   1,   1,   2,  -2,  11,   6,  -7,  -2,  -1,   2,  -1,  -2,   5,  -6,   2, -55,  -4,  -2,  -4,  -6,   4,  -3, -42,  -5,  -3,   4,   0,   0,  13, -38,  16,  60,   2,   1,  -8,   5, -16,   6,  40, -16,  -6,  -3,  -7,   7,  17,   2,  -6, -22,  -7,  26, -25,  17,   4, -24,   6, -15 },
      {  -1,   0,   1,   4,   3,   2, -19,   9,  -2,  -2,   2,   0,   4,  -1, -10,  -8,  -2,   4,   2,  -1,   4,  -5, -11,  16,   0,  -2,  -3,   1,   3,  -7,  -5,  30,   1,   0,  12,   4,   5, -14,  21,   4,  -1,   8,  -3,   3,  -4,   5,   4, -50,   2,   1,  -3,   8, -16, -10, -58,  14,  17,  -8,  28, -40, -14, -34,  22,  46 },
      {   0,   2,   1,   0,   2,   5,  -5, -12,  -1,   1,   1,   2,  -3,   2,  -5, -24,  -1,   3,   6,   0,   0,   4,  -1,  -9,  -3,   7,  -2,   2,  -1,   7, -18, -32,   2,   6,   0,   2,  -3,  -2, -40,  44,   9,   7,  -3,   8,   1, -24,  18,  27,   7,  -4,  15,  19, -35,  14,   5, -11,   8,  -9,  51, -33, -19,  33, -22,  22 },
      {   0,   0,   1,   0,  -3,   7,  -7,  -9,   0,   0,  -1,  -4,  -6,  -1,   4, -21,   0,  -3,  -3,  -5,   1,   3,   7, -28,  -1,  -1,  -4,   1,  -6,  10, -15, -26,   3,  -3,  -6,  -2,  -4,  19, -39,  15,  -4,  -6,  11,  -6,   5,   6, -34,  32,   4,  -9,   7, -23,  18, -15, -22,  52,  11,   5,  -4,  -1,  41, -41,  27,  21 },
      {   0,   1,   0,   0,   3,  -1,  15,   5,  -1,   0,   0,   1,  -2,  -3,  -6,   6,  -2,   0,   3,   1,   0,   0, -31,  42,  -1,   0,   6,  -3,   1,  -5,  -8,  36,  -1,   0,   5,  -8,   7, -19,   7,  30,  -2,   0,   1,  -4,  -3, -23,  20,  33,  -5,   2,   6,   9,  -4,  -4,  45,  39,  -5,  11,  18,  -6,   8,  -7,  60, -22 },
    },
  },
};
#else
const int8_t g_lfnst8x8[ 4 ][ 2 ][ 16 ][ 48 ] = {
 {  //0
  {
  { -117,   28,   18,    2,    4,    1,    2,    1,   32,  -18,   -2,    0,   -1,    0,    0,    0,   14,   -1,   -3,    0,   -1,    0,    0,    0,    2,    0,    0,    0,    0,    0,    0,    0,    3,    0,   -1,    0,    1,    0,    0,    0,    1,    0,    0,    0,    1,    0,    0,    0  },
  {  -29,  -91,   47,    1,    9,    0,    3,    0,  -54,   26,   -8,    3,    0,    1,    0,    0,   33,    5,   -9,   -1,   -2,    0,   -1,    0,   -3,    3,    0,    0,    0,    0,    0,    0,    7,    2,   -2,    0,   -1,    1,    0,    0,    2,    1,   -1,    0,    0,    0,    0,    0  },
  {  -10,   62,  -11,   -8,   -2,   -2,   -1,   -1,  -95,    3,   32,    0,    4,    0,    2,    0,   32,  -30,   -4,    4,   -1,    1,    0,    0,    6,    2,   -5,    0,    0,    0,    0,    0,    6,   -3,    0,    0,    2,    0,   -1,    0,    2,   -1,    0,    0,    1,    0,    0,    0  },
  {  -15,   15,  -10,   -2,    1,    0,    1,    0,   10,  112,  -20,  -17,   -4,   -4,   -1,   -2,  -20,  -26,   31,    1,    0,    0,    0,    0,    2,  -16,   -1,    6,    0,    1,    0,    0,    1,   -4,    0,    0,    0,   -3,    0,    1,    0,   -1,    0,    0,    0,   -2,    0,    0  },
  {   32,   39,   92,  -44,    4,  -10,    1,   -4,   26,   12,  -15,   13,   -5,    2,   -2,    0,   29,  -16,  -22,    8,    0,    1,    0,    1,  -20,    6,    4,   -3,    1,    0,    0,    0,    1,   -4,   -3,    2,   -4,    1,    0,    0,    1,   -1,   -2,    1,   -2,    0,    0,    0  },
  {  -10,    1,   50,  -15,    2,   -3,    1,   -1,  -28,  -15,   14,    6,    1,    1,    1,    0,  -99,   -4,    9,    5,    5,    2,    2,    1,   44,  -10,  -11,    1,   -2,    0,   -1,    0,   -5,    4,   -3,    0,    8,   -1,   -2,    0,   -2,    1,   -1,    0,    4,    0,   -1,    0  },
  {    1,  -33,  -11,  -14,    7,   -2,    2,    0,   29,  -12,   37,   -7,   -4,    0,   -1,    0,    6,  -99,    3,   26,   -1,    5,    0,    2,   14,   30,  -27,   -2,    1,   -1,    0,   -1,   -6,    6,    6,   -3,    1,    3,   -3,    0,   -1,    1,    1,    0,    0,    1,   -1,    0  },
  {    0,    6,   -6,   21,   -4,    2,    0,    0,  -20,  -24, -104,   30,    5,    5,    1,    2,   -7,  -46,   10,  -14,    7,    0,    1,    0,    9,   21,    7,   -6,   -2,   -1,    0,   -1,    2,    2,    5,   -2,    0,    3,    4,   -1,    0,    0,    1,    0,    0,    1,    2,   -1  },
  {  -13,  -13,  -37, -101,   29,  -11,    8,   -3,  -12,  -15,  -20,    2,  -11,    5,   -2,    1,  -12,   10,   26,   12,   -6,    0,   -1,    0,  -32,   -2,   11,    3,    3,   -1,    1,    0,   11,   -5,   -1,    6,   -4,    2,    1,    0,    3,   -1,    1,    2,   -1,    0,    0,    0  },
  {    6,    1,  -14,  -36,    9,   -3,    2,    0,   10,    9,  -18,   -1,   -3,    1,    0,    0,   38,   26,  -13,   -1,   -5,   -1,   -1,    0,  102,    3,  -14,   -1,   -5,   -1,   -2,    0,  -29,   10,   10,    0,   10,   -4,   -1,    1,   -7,    1,    2,    1,    2,   -1,    0,    0  },
  {  -12,   -2,  -26,  -12,   -9,    2,   -1,    1,   -3,   30,    4,   34,   -4,    0,   -1,    0,  -30,    3,  -92,   14,   19,    0,    3,    0,  -11,   34,   21,  -33,    1,   -2,    0,   -1,   -9,   -4,   18,    3,    2,    0,    0,   -2,   -1,   -1,    3,    0,    0,    0,    0,   -1  },
  {    0,   -3,    0,   -4,  -15,    6,   -3,    1,   -7,  -15,  -28,  -86,   19,   -5,    4,   -1,   -5,  -17,  -41,   42,   -6,    2,   -1,    1,   -1,  -40,   37,   13,   -4,    2,   -1,    1,  -10,   13,   -1,   -4,    4,   -4,    3,    4,   -2,    2,   -1,   -1,    1,   -1,    1,    2  },
  {   -1,    9,   13,    5,   14,   -2,    2,   -1,   -8,    3,   -4,  -62,    4,    1,    1,    0,  -12,   23,   16,  -11,  -17,    0,   -1,    0,  -11,   97,   -3,   -3,    0,   -6,    0,   -2,  -21,   -5,   23,    0,    2,   -2,   -1,    6,   -3,   -3,    1,    0,    0,    0,    0,    2  },
  {    6,    2,   -3,    2,   10,   -1,    2,    0,    8,    3,   -1,  -20,    0,    1,    0,    0,   -4,    4,  -16,    0,   -2,    0,    1,    0,   34,   23,    6,   -7,   -4,   -2,   -1,    0,  108,   -5,  -30,    6,  -27,   10,    7,   -2,   11,   -3,   -1,    1,   -4,    1,    0,    1  },
  {    6,    9,   -2,   35,  110,  -22,   11,   -4,   -2,    0,   -3,    1,  -18,   12,   -3,    2,   -5,   -4,  -22,    8,  -25,    3,    0,    0,   -3,  -21,    2,   -3,    9,   -2,    1,    0,   -7,    1,    3,   -5,    3,    0,   -1,    0,    0,    1,    0,   -1,    1,    0,    0,    0  },
  {   -1,    7,   -2,    9,  -11,    5,   -1,    1,   -7,    2,  -22,    4,  -13,    0,   -1,    0,    0,   28,    0,   76,    4,   -6,    0,   -2,  -13,    5,  -76,   -4,   33,   -1,    3,    0,    9,   18,   -3,  -35,   -4,   -1,    6,    1,    1,    2,    0,   -3,   -1,    0,    2,    0  },
  },
  {
  { -108,   48,    9,    1,    1,    1,    0,    0,   44,   -6,   -9,   -1,   -1,    0,   -1,    0,    9,   -9,   -1,    1,    0,    0,    0,    0,    3,   -1,    1,    0,    0,    0,    0,    0,    1,   -1,    0,    0,    1,    0,    0,    0,    0,   -1,    0,    0,    0,    0,    0,    0  },
  {   55,   66,  -37,   -5,   -6,   -1,   -2,    0,   67,  -30,  -20,    4,   -2,    0,   -1,    0,  -31,  -19,   14,    4,    1,    1,    1,    0,   -6,    3,    5,   -2,    0,    0,    0,    0,   -7,   -1,    1,    0,   -1,    1,    1,    0,   -2,   -1,    1,    0,    0,    0,    0,    0  },
  {    2,   86,  -21,  -13,   -4,   -2,   -1,   -1,  -88,    5,    6,    4,    5,    1,    1,    0,   14,   -5,    0,    3,    0,    0,    0,    0,   10,   -5,   -2,    0,   -1,    0,    0,    0,    6,   -5,    0,    1,    2,   -1,    0,    0,    1,   -1,    0,    0,    1,    0,    0,    0  },
  {  -24,  -21,  -38,   19,    0,    4,   -1,    2,  -23,  -89,   31,   20,    2,    3,    1,    1,  -30,   26,   36,   -8,   -2,   -2,    0,   -1,   14,   18,   -7,   -9,   -1,   -1,    0,    0,    1,    3,   -2,   -1,    3,    2,   -2,   -1,    0,    1,    0,    0,    1,    1,   -1,    0  },
  {    9,   20,   98,  -26,   -3,   -5,    0,   -2,   -9,  -26,   15,  -16,    2,    0,    1,    0,  -61,   -3,   -2,    3,    7,    1,    1,    0,   12,   16,   -6,   -1,    0,   -1,    0,    0,    2,    0,   -8,    1,    3,    1,   -1,    1,    0,   -1,   -2,    0,    1,    0,   -1,    0  },
  {  -21,   -7,  -37,   10,    2,    2,   -1,    1,  -10,   69,   -5,   -7,   -2,   -2,    0,   -1,  -93,    2,   19,    0,    3,    0,    2,    0,   17,    4,    0,    0,   -1,    0,    0,    0,    5,   -4,   -2,    0,    4,   -2,    0,    1,    0,    0,    0,    0,    2,   -1,    0,    0  },
  {  -10,  -25,    4,  -17,    8,   -2,    2,   -1,  -27,  -17,  -71,   25,    8,    2,    1,    1,   -4,  -66,   28,   36,   -5,    3,    0,    1,  -10,   20,   33,  -13,   -8,    0,    0,   -1,    3,    6,   -3,   -7,   -1,    3,    3,   -1,    1,    0,   -1,    0,    0,    1,    1,   -1  },
  {    2,    5,   10,   64,   -9,    4,   -3,    1,   -4,    8,   62,    3,  -17,    1,   -2,    0,   -3,  -75,    5,  -14,    1,    4,    0,    1,  -36,    3,   18,   -4,    4,    0,    1,    0,    1,   14,   -2,   -8,   -2,    1,   -3,    0,    2,    2,   -1,   -2,    0,    1,   -1,    0  },
  {  -11,  -15,  -28,  -97,    6,   -1,    4,   -1,    7,    3,   57,  -15,   10,   -2,    0,   -1,   -1,  -27,   13,    6,    1,   -1,    0,    0,  -34,   -6,    0,    3,    4,    1,    2,    0,   -2,    8,    1,    5,   -2,    0,   -3,    1,    1,    1,    0,    2,   -1,    0,   -1,    0  },
  {    9,   13,   24,   -6,    7,   -2,    1,   -1,   16,   39,   20,   47,   -2,   -2,   -2,    0,   28,   23,   76,   -5,  -25,   -3,   -3,   -1,    6,   36,   -7,  -39,   -4,   -1,    0,   -1,    2,   -4,  -18,   -3,   -1,   -1,   -2,   -2,    1,   -2,   -2,    0,    0,    0,   -1,   -1  },
  {   -7,   11,   12,    7,    2,   -1,    0,   -1,  -14,   -1,  -24,   11,    2,    0,    0,    0,  -20,   48,   11,  -13,   -5,   -2,    0,   -1, -105,  -19,   17,    0,    6,    2,    3,    0,  -14,    8,    8,    2,    1,    2,   -1,   -2,    3,    0,   -1,    0,    0,    0,    0,    0  },
  {    0,    0,    7,   -6,   23,   -3,    3,   -1,    5,    1,   18,   96,   13,   -9,   -1,   -1,  -21,   -7,  -42,   14,  -24,   -3,    0,    0,   11,  -47,   -7,    3,   -5,    9,    1,    2,    0,   -1,   19,   -1,    1,    0,   -1,   -6,   -1,    1,    2,    0,    1,    0,    0,   -2  },
  {   -2,   -6,   -1,  -10,    0,    1,    1,    0,   -7,   -2,  -28,   20,  -15,    4,   -3,    1,   -2,  -32,   -2,  -66,    3,    7,    1,    2,  -11,   13,  -70,    5,   43,   -2,    3,    0,    8,  -14,   -3,   43,   -1,    2,    7,   -1,    1,   -2,    1,    3,   -1,    1,    1,    0  },
  {   -1,    6,  -16,    0,   24,   -3,    1,   -1,    2,    6,    6,   16,   18,   -7,    1,   -1,   -3,   11,  -63,    9,    4,   -5,    2,   -1,  -22,   94,   -4,   -6,   -4,   -4,    1,   -2,   10,   23,  -19,   -5,    0,   -6,   -4,    6,    3,   -2,    1,    1,    0,   -1,    0,    0  },
  {   -5,   -6,   -3,  -19, -104,   18,   -4,    3,    0,    6,    0,   35,  -41,   20,   -2,    2,   -2,   10,  -18,   16,   21,    3,   -2,    0,   -2,   11,    6,  -10,    6,   -3,   -1,    0,   -1,    5,   -1,   -6,   -1,   -1,   -1,   -1,   -1,    0,    0,    0,    0,    0,    0,   -1  },
  {   -1,   -2,    0,   23,   -9,    0,   -2,    0,    1,    1,    8,   -1,   29,    1,    1,    0,    3,   -6,   13,   76,   30,  -11,   -1,   -2,  -26,   -8,  -69,    7,   -9,   -7,    3,   -1,  -10,  -34,  -25,   13,   -1,    0,   11,    5,    1,   -1,    1,   -2,    0,    0,    2,    0  },
  }
 },
 {  //1
  {
  {  110,  -49,   -3,   -4,   -1,   -1,    0,   -1,  -38,   -1,   10,    0,    2,    0,    1,    0,   -9,   13,    1,   -2,    0,    0,    0,    0,   -4,    2,   -3,    0,    0,    0,    0,    0,   -2,    2,    0,    1,   -1,    1,    0,    0,   -1,    1,    0,    0,   -1,    0,    0,    0  },
  {  -43,  -19,   17,   -1,    3,    0,    1,    0,  -98,   46,   14,   -1,    2,    0,    1,    0,   26,   26,  -15,   -3,   -2,   -1,   -1,    0,   11,   -7,   -9,    2,    0,    0,    0,    0,    9,   -3,   -1,    2,    3,   -3,    0,    0,    4,   -1,    0,    0,    2,   -1,    0,    0  },
  {  -19,   17,   -7,    3,   -2,    1,   -1,    0,  -32,  -59,   29,    3,    4,    0,    2,    0,  -72,   43,   34,   -9,    3,   -2,    1,   -1,   13,   36,  -18,  -10,    0,   -2,    0,   -1,    3,    0,  -12,    3,    6,    1,   -3,    2,    1,   -1,   -2,    0,    3,    1,   -1,    1  },
  {  -35, -103,   39,    1,    7,    0,    2,    0,   38,  -13,   25,   -6,    1,   -1,    0,    0,   -1,    7,    6,   -7,    1,   -1,    0,    0,  -13,   14,    2,   -4,    2,   -1,    0,    0,   -2,   11,   -6,   -2,   -2,    4,   -3,    0,    0,    3,   -2,    0,   -1,    1,   -1,    0  },
  {    9,    5,   -6,   -1,   -1,    0,   -1,    0,   42,    4,   21,  -11,    1,   -3,    1,   -1,   21,   70,  -32,  -21,    0,   -4,   -1,   -1,   34,  -26,  -57,   11,    4,    2,    0,    1,   -4,  -32,    5,   24,    1,   -6,   12,    4,   -3,   -2,    4,   -2,    0,   -1,    0,    0  },
  {   -5,   -5,  -28,    9,   -3,    2,   -1,    1,  -20,  -78,   22,   16,    1,    3,    0,    1,   80,   -6,   25,   -5,   -4,   -1,   -1,    0,    6,  -24,    7,   -9,    0,    0,    0,    0,   -7,    3,   13,   -4,   -3,    5,    1,   -5,   -2,    3,    1,   -2,   -1,    2,   -1,   -2  },
  {   14,   17,   27,  -12,    1,   -3,    1,   -1,    8,   19,  -13,    4,   -2,    1,   -1,    0,   48,   -1,   48,  -15,   -4,   -2,   -1,   -1,    1,   60,  -28,  -42,    5,   -6,    1,   -2,   11,  -11,  -51,   11,   -2,  -10,   -2,   13,    2,   -6,   -4,    4,   -2,   -3,    2,    2  },
  {    7,   35,   17,   -4,   -1,    0,    0,    0,    3,    8,   54,  -17,    1,   -2,    1,   -1,   10,   14,  -11,  -34,    4,   -4,    1,   -1,  -80,   -7,   -6,    2,   15,    0,    3,    0,  -16,   46,    1,    3,    2,    7,  -24,    0,    2,   -2,   -5,    8,    1,   -1,   -2,    2  },
  {  -13,  -27, -101,   24,   -8,    6,   -3,    2,   11,   43,    6,   28,   -6,    3,   -1,    1,   -3,   14,   21,  -12,   -7,   -2,   -1,   -1,  -23,   10,   -4,  -12,    3,    0,    1,    0,    2,    9,  -10,    0,    1,   -5,   -4,    4,    2,   -2,    2,    2,    0,   -2,    1,    0  },
  {  -11,  -13,   -3,  -10,    3,   -1,    1,    0,  -19,  -19,  -37,    8,    4,    2,    0,    1,  -12,  -30,    3,   -9,    5,    0,    1,    0,  -56,   -9,  -47,    8,   21,    1,    4,    1,  -11,  -30,   10,   59,   -2,    8,   41,    8,    2,    5,    6,   -7,   -1,    3,    5,   -2  },
  {   -4,  -10,  -24,  -11,    3,   -2,    0,   -1,   -6,  -37,  -45,  -17,    8,   -2,    2,   -1,   17,   14,  -58,   14,   15,    0,    2,    0,  -10,   34,   -7,   28,    4,   -1,    1,    0,   23,   34,  -31,    4,   10,  -22,  -30,   22,    4,  -15,    9,   20,    2,   -5,    9,    4  },
  {   -2,    1,   13,  -17,    3,   -5,    1,   -2,    3,    0,  -55,   22,    6,    1,    1,    0,    8,   74,   21,   40,  -14,    0,   -2,    0,  -36,   -8,   11,  -13,  -23,    1,   -3,    0,  -36,    6,   16,  -14,    2,   19,   -4,  -12,   -1,    0,   -7,   -3,    0,    2,   -2,   -1  },
  {    3,    1,    5,  -15,    1,   -2,    1,   -1,    7,    4,   -7,   29,   -1,    2,   -1,    1,    8,    3,   12,  -14,   -9,   -1,   -1,    0,    4,   29,  -15,   31,   10,    4,    1,    1,   61,   22,   55,   14,   13,    3,   -9,  -65,    1,  -11,  -21,   -7,    0,    0,   -1,    3  },
  {   -4,   -8,   -1,  -50,    6,   -4,    2,   -2,   -1,    5,  -22,   20,    6,    1,    0,    0,  -16,  -15,   18,  -29,  -11,    2,   -2,    1,   40,  -45,  -19,  -22,   31,    2,    4,    1,  -25,   41,    0,   12,    9,    7,  -42,   12,   -3,  -14,    2,   28,    5,    1,    6,    2  },
  {    5,   -1,   26,  102,  -13,   12,   -4,    4,   -4,   -2,  -40,   -7,  -23,    3,   -5,    1,   -1,    5,    8,  -23,    7,    2,    1,    1,   10,  -11,  -13,   -3,   12,   -3,    2,    0,   -9,   23,    4,    9,   14,    9,  -14,   -4,    0,  -12,   -7,    6,    3,    0,    6,    3  },
  {   -5,   -6,  -27,  -22,  -12,    0,   -3,    0,   -5,    8,  -20,  -83,    0,    0,    0,    0,    9,    7,   24,  -20,   41,    3,    6,    1,   15,   20,   12,   11,   17,   -9,    1,   -2,  -26,   -1,   18,   -1,  -12,   32,    3,  -18,   -5,   10,  -25,   -5,   -2,    1,   -8,   10  },
  },
  {
  {   80,  -49,    6,   -4,    1,   -1,    1,   -1,  -72,   36,    4,    0,    1,    0,    0,    0,   26,    0,  -12,    2,   -2,    1,   -1,    0,   -7,   -9,    6,    1,    0,    0,    0,    0,    3,    5,   -1,   -2,   -2,   -2,   -1,    1,    1,    1,    0,    0,   -1,   -1,    0,    0  },
  {  -72,   -6,   17,    0,    3,    0,    1,    0,  -23,   58,  -21,    2,   -3,    1,   -1,    0,   55,  -46,   -1,    6,   -2,    1,   -1,    0,  -22,    7,   17,   -7,    2,   -1,    1,    0,    9,    5,  -12,    1,   -3,   -4,    4,    2,    4,    1,   -2,   -1,   -1,   -1,    1,    0  },
  {  -50,   19,  -15,    4,   -1,    1,   -1,    1,  -58,   -2,   30,   -3,    4,   -1,    2,    0,    6,   57,  -34,    0,   -2,    0,   -1,    0,   34,  -48,   -2,   14,   -4,    3,   -1,    1,  -10,    7,   21,  -10,    6,    1,  -11,    0,   -1,   -1,    4,    2,    3,    0,   -2,   -1  },
  {  -33,  -43,   28,   -7,    4,   -2,    2,   -1,  -38,   11,   -8,    4,    1,    1,    0,    0,  -55,   24,   26,   -5,    2,   -1,    1,    0,   15,   46,  -40,   -1,   -1,    0,   -1,    0,   17,  -38,    1,   17,   -3,   11,   15,  -11,    3,   -1,  -10,    1,    0,    1,    3,    2  },
  {   10,   66,  -21,   -3,   -3,    0,   -1,    0,  -53,  -41,   -2,   16,   -1,    4,   -1,    1,   36,   -5,   41,  -20,    3,   -3,    1,   -1,  -30,   26,  -32,   -3,    7,   -2,    2,   -1,   15,   -8,    1,   17,   -1,   -2,    4,   -8,    2,    0,   -1,    3,    0,    0,    0,   -1  },
  {   18,   14,   13,   -9,    2,   -2,    1,   -1,   34,   32,  -31,   12,   -5,    2,   -2,    1,   40,    4,   -4,   -9,   -3,   -2,   -1,   -1,   27,  -31,  -43,   19,   -2,    3,   -1,    1,    7,  -49,   52,   10,  -11,   22,    7,  -26,   -1,   -6,   -9,    6,   -2,    2,    4,   -2  },
  {   21,   66,   -1,    9,   -4,    2,   -1,    1,  -21,   41,  -30,  -10,    0,   -2,    0,   -1,  -35,  -17,   -3,   26,   -6,    5,   -2,    2,   56,    3,   18,  -25,   -1,   -2,   -1,   -1,  -15,  -13,  -27,    9,    9,   -6,   20,    5,   -3,    2,   -6,   -9,    3,   -3,    1,    5  },
  {    1,   -6,  -24,   17,   -5,    3,   -2,    1,   24,   10,   39,  -21,    5,   -4,    2,   -1,   33,   32,  -30,    4,   -3,   -1,   -1,    0,   -4,   13,  -16,  -10,    0,   -1,    0,    0,   24,  -26,  -37,   33,    5,  -32,   55,   -5,   -7,   22,  -14,  -22,    1,   -9,   -3,   13  },
  {    9,   33,  -24,    1,    4,    0,    1,    0,    6,   50,   26,    1,  -10,    0,   -2,    0,  -27,    1,  -28,  -21,   16,   -5,    3,   -2,  -23,   36,   -2,   40,  -17,    4,   -3,    1,   43,  -13,    4,  -41,  -19,   -2,  -24,   17,   11,   -4,    8,    4,   -3,   -3,   -3,   -3  },
  {   -7,   -9,  -32,   14,   -3,    3,   -1,    1,  -23,  -28,    0,   -5,   -1,    0,    0,    0,  -36,  -59,  -24,   14,    4,    2,    1,    1,  -23,  -26,   23,   26,   -3,    5,    0,    2,   10,  -26,   38,    7,  -12,   11,   42,  -22,   -5,   20,  -14,  -15,   -1,   -2,    1,    6  },
  {    6,   30,   69,  -18,    5,   -4,    3,   -1,   -3,  -11,  -34,  -16,    9,   -4,    2,   -1,  -16,   35,  -35,   30,   -9,    3,   -2,    1,  -57,  -13,    6,    4,   -5,    5,   -1,    1,   28,   10,    4,    7,    0,  -15,    7,  -10,   -1,    7,   -2,    2,    1,   -3,    0,    0  },
  {    1,   -8,   24,   -3,    7,   -2,    2,   -1,   -6,  -51,   -6,   -4,   -5,    0,   -1,    0,   38,   -1,    0,   25,    6,    2,    1,    1,   47,   20,   35,    1,  -27,    1,   -5,    0,   37,  -37,   -9,  -47,  -28,    5,    0,   18,    8,    6,    0,   -8,   -4,   -3,   -3,    1  },
  {    4,   10,    4,   17,   -9,    4,   -2,    1,    5,   14,   32,  -15,    9,   -3,    2,   -1,    7,   13,   19,   15,   -8,    1,   -1,    0,    3,   25,   30,  -18,    1,   -2,    0,   -1,   11,   24,   22,  -11,   -3,   37,  -13,  -58,   -5,   12,  -63,   26,    9,  -15,   11,    8  },
  {   -3,   -9,  -23,   10,  -10,    3,   -3,    1,   -5,  -14,  -16,  -27,   13,   -5,    2,   -1,   -1,  -13,  -30,   11,   -5,    2,   -1,    0,   -5,   -8,  -22,  -16,   10,    0,    1,    0,    0,  -29,  -27,    6,  -27,  -10,  -30,    9,   -3,  -10,   -7,   77,    9,  -13,   45,   -8  },
  {    2,   11,   22,    2,    9,   -2,    2,    0,   -6,   -7,   20,  -32,   -3,   -4,    0,   -1,   13,   -5,  -28,    6,   18,   -4,    3,   -1,  -26,   27,  -14,    6,  -20,    0,   -2,    0,  -76,  -26,   -4,   -7,   12,   51,    5,   24,    7,  -17,  -16,  -12,   -5,    4,    2,   13  },
  {    2,   -3,    8,   14,   -5,    3,   -1,    1,   -2,  -11,    5,  -18,    8,   -3,    2,   -1,   12,  -23,  -19,   22,    2,    0,    1,    0,   23,   41,   -7,   35,  -10,    4,   -1,    1,    5,    7,   23,    5,   69,  -38,   -8,  -32,  -15,  -31,   24,   11,    2,   18,   11,  -15  },
  }
 },
 {  //2
  {
  { -121,   33,    4,    4,    1,    2,    0,    1,   -1,   -1,    1,    0,    0,    0,    0,    0,   24,   -5,   -1,   -1,    0,    0,    0,    0,    5,   -1,    0,    0,    0,    0,    0,    0,    3,   -1,    0,    0,    2,   -1,    0,    0,    2,   -1,    0,    0,    1,    0,    0,    0  },
  {    0,   -2,    0,    0,    0,    0,    0,    0,  121,  -23,   -7,   -3,   -2,   -1,   -1,    0,   17,    1,   -2,    0,    0,    0,    0,    0,  -27,    4,    2,    0,    0,    0,    0,    0,  -12,    2,    1,    0,   -5,    1,    0,    0,   -1,    0,    0,    0,   -2,    0,    0,    0  },
  {  -20,   19,   -5,    2,   -1,    1,    0,    0,   16,    3,   -2,    0,    0,    0,    0,    0, -120,   14,    8,    1,    3,    1,    1,    0,  -18,   -2,    3,    0,    1,    0,    0,    0,   17,   -3,   -1,    0,    6,   -1,   -1,    0,    2,    0,    0,    0,    2,    0,    0,    0  },
  {   32,  108,  -43,   10,   -9,    3,   -3,    1,    4,   19,   -7,    1,   -1,    0,    0,    0,   11,  -30,    9,   -2,    1,   -1,    0,    0,    0,   -8,    2,    0,    0,    0,    0,    0,   -7,   -1,    2,    0,   -3,   -1,    1,    0,   -2,   -2,    1,    0,    0,    0,    0,    0  },
  {   -3,    0,   -1,    0,    0,    0,    0,    0,  -29,   11,   -2,    1,    0,    0,    0,    0,   12,    7,   -1,    0,    0,    0,    0,    0, -117,   12,    9,    1,    3,    0,    1,    0,  -32,   -3,    3,    0,   12,   -2,   -1,    0,    7,    0,    0,    0,    1,    0,    0,    0  },
  {   -4,  -12,   -3,    1,   -1,    0,    0,    0,   19,  105,  -31,    7,   -6,    1,   -2,    0,    9,   46,   -6,    0,    0,    0,    0,    0,    8,  -29,    9,   -3,    1,    0,    0,    0,   -3,  -19,    3,    0,   -4,   -6,    1,    0,    0,    0,    0,    0,    0,   -1,    0,    0  },
  {    7,    1,    2,    0,    0,    0,    0,    0,    4,    3,   -2,    0,    0,    0,    0,    0,   22,   -8,    1,   -1,    0,    0,    0,    0,  -28,   -9,    4,    0,    1,    0,    0,    0,  117,  -10,   -8,    0,   32,    1,   -4,    0,    3,    1,   -1,    0,   -3,    1,    0,    0  },
  {   -8,  -31,   14,   -4,    3,   -1,    1,    0,    9,   43,    0,    1,   -1,    0,    0,    0,  -13, -105,   17,   -2,    2,    0,    0,    0,   -8,  -25,   -3,    0,    0,    0,    0,    0,   -7,   32,   -5,    1,   -1,    4,    0,    0,    2,   -1,    0,    0,    1,    0,   -1,    0  },
  {  -15,  -43, -100,   23,  -12,    6,   -4,    2,   -6,  -17,  -48,   10,   -5,    2,   -1,    1,    1,   -5,   19,   -6,    3,   -1,    1,    0,    2,    7,   15,   -3,    1,   -1,    0,    0,    4,   10,    5,   -1,    0,    3,    1,    0,   -2,    1,    2,    0,   -1,    1,    1,    0  },
  {   -3,    1,    2,    0,    0,    0,    0,    0,   -6,    3,    1,    0,    0,    0,    0,    0,    0,    3,   -2,    0,    0,    0,    0,    0,  -20,    8,   -2,    0,    0,    0,    0,    0,   30,   13,   -3,    0, -116,    6,   10,    0,  -35,   -5,    4,    0,   -3,   -1,    0,    0  },
  {   -1,   -6,   -3,    2,   -1,    0,    0,    0,   -6,  -35,    9,    0,    2,    0,    0,    0,    1,   -6,   11,   -2,    2,    0,    1,    0,   -9, -100,   17,   -1,    1,    0,    0,    0,  -10,  -63,    1,    2,  -17,    3,   -4,    0,   -1,    9,   -1,    0,    3,    4,   -1,    0  },
  {   -5,  -14,  -48,    2,   -5,    1,   -2,    0,   10,   24,   99,  -17,   10,   -4,    3,   -1,    4,   14,   32,    0,    2,    0,    1,    0,   -4,    0,  -39,    6,   -4,    1,   -1,    0,    2,   -3,   -4,    0,    2,   -2,   -2,    0,    0,    0,   -1,    0,    0,   -1,   -1,    0  },
  {   -2,    0,    2,    0,    0,    0,    0,    0,   -2,    0,    1,    0,    0,    0,    0,    0,   -1,   -1,    1,   -1,    0,    0,    0,    0,   -1,   -4,    2,    0,    0,    0,    0,    0,   -8,   -2,   -1,    1,   30,    4,   -4,    1, -102,    4,    8,   -1,  -69,   -2,    6,   -1  },
  {   -2,  -10,   -4,    0,    0,    0,    0,    0,    3,   11,   -1,   -1,    0,    0,    0,    0,   -6,  -40,  -15,    6,   -2,    1,    0,    0,    5,   57,   -6,    2,    0,    0,    0,    0,    1,  -95,   18,   -6,  -10,  -34,   -2,    0,   -4,   17,   -2,    0,    0,    2,    1,    0  },
  {   -2,   -3,  -25,   -2,   -3,    0,   -1,    0,   -1,   -3,   -1,    4,   -2,    2,    0,    1,   -7,   -8,  -97,   17,   -9,    3,   -3,    1,   -8,  -26,  -61,   -1,   -3,   -1,   -1,   -1,    2,   10,   24,   -7,    5,    9,   19,   -1,    0,    1,    4,    0,   -2,    0,    1,    0  },
  {    4,   -4,   28,  103,  -42,   24,   -9,    7,    1,    2,    4,    0,    3,   -1,    0,    0,   -1,    0,   -9,  -42,   17,   -9,    3,   -2,   -1,    1,  -14,    6,   -4,    2,   -1,    0,   -1,   -2,   -4,    4,    0,    3,    1,   -1,    0,    2,    0,   -2,    2,    0,    0,    0  },
  },
  {
  {   87,  -41,    3,   -4,    1,   -1,    0,   -1,  -73,   28,    2,    1,    1,    1,    0,    0,   30,   -5,   -6,    1,   -1,    0,    0,    0,   -8,   -3,    3,    0,    0,    0,    0,    0,    3,    2,   -1,    0,   -2,   -1,    0,    0,    1,    1,    0,    0,   -1,    0,    0,    0  },
  {  -75,    4,    7,    0,    2,    0,    1,    0,  -41,   36,   -7,    3,   -1,    1,    0,    0,   72,  -29,   -2,    0,   -1,    0,   -1,    0,  -37,    6,    7,   -2,    1,    0,    0,    0,   12,    3,   -4,    0,   -3,   -2,    1,    0,    4,    0,    0,    0,   -1,    0,    0,    0  },
  {   26,  -44,   22,   -6,    4,   -2,    1,   -1,   77,   24,  -22,    2,   -4,    0,   -1,    0,    7,  -38,   10,    0,    1,    0,    0,    0,  -51,   27,    4,   -3,    2,   -1,    1,    0,   31,   -5,   -8,    3,  -14,    0,    5,   -1,    6,    1,   -3,    0,   -4,   -1,    1,    0  },
  {  -39,  -68,   37,   -7,    6,   -2,    2,    0,   -9,   56,  -21,    1,   -2,    0,   -1,    0,  -45,    4,   -3,    6,   -1,    2,    0,    1,   49,  -13,    3,   -3,   -1,    0,    0,    0,  -19,    2,    0,    0,    5,    1,    1,    0,   -2,    0,   -1,    0,    1,    0,    0,    0  },
  {   10,  -20,    2,    0,    1,    0,    0,    0,   50,   -1,    8,   -5,    1,   -1,    0,    0,   66,   17,  -24,    4,   -3,    1,   -1,    0,   13,  -49,   15,    1,    0,    0,    0,    0,  -53,   34,    6,   -5,   30,   -7,  -11,    3,  -11,   -2,    5,    1,    4,    2,   -1,   -1  },
  {  -21,  -45,    8,   -2,    3,   -1,    1,    0,   -7,  -30,   26,   -8,    3,   -1,    1,   -1,   -9,   69,  -33,    5,   -2,    0,   -1,    0,  -44,  -31,   10,    7,   -2,    2,    0,    1,   49,    7,    2,   -6,  -23,   -3,   -2,    2,    9,    4,    0,    0,   -2,   -1,   -1,    0  },
  {   -4,   -2,  -55,   28,   -8,    5,   -3,    2,   -2,   37,   43,  -19,    1,   -2,    1,   -1,  -47,  -34,  -27,    5,    4,   -1,    1,    0,  -39,   -2,   27,    4,   -2,    1,    0,    0,  -11,   32,   -8,   -7,   27,  -12,   -6,    6,  -13,    0,    4,   -3,    3,   -1,   -2,    1  },
  {    2,   19,   47,  -23,    6,   -4,    2,   -1,  -23,  -22,  -44,   17,   -2,    2,   -1,    0,  -33,    3,   22,   -2,   -4,    1,   -1,    0,  -58,  -17,    6,   -6,    7,   -1,    1,    0,  -23,   40,   -2,    5,   43,  -11,   -8,   -1,  -18,   -4,    5,    2,    4,    3,    0,   -1  },
  {  -19,  -62,   -9,    3,    0,    0,    0,    0,  -12,  -56,   27,   -7,    3,   -1,    1,    0,    7,   -8,   16,   -6,    4,   -2,    1,   -1,  -15,   54,  -23,    2,   -1,    0,    0,    0,  -42,  -25,    4,    6,   34,    8,    2,   -2,  -15,   -1,    0,   -1,    3,    2,    0,    1  },
  {    1,    9,   -5,    0,   -1,    0,    0,    0,    0,   22,   -1,    2,    0,    1,    0,    0,  -13,   17,    0,   -2,    0,   -1,    0,    0,  -46,  -10,  -10,    4,   -1,    1,    0,    0,  -80,  -27,   20,   -4,  -66,   23,   -2,   -2,   20,   -3,   -2,    3,  -14,    2,    3,   -1  },
  {    5,   17,   -9,    0,   -2,    1,    0,    0,   13,   54,   -2,    7,   -1,    1,    0,    0,    4,   51,   -3,   -6,   -1,   -1,    0,    0,  -20,    6,  -34,    9,   -2,    2,   -1,    0,   16,  -52,   28,    1,   59,   15,   -8,   -5,  -28,   -7,    2,    2,   10,    3,    0,   -1  },
  {    7,   27,   56,   -2,   10,   -3,    3,   -1,   -2,   -6,    8,  -28,    3,   -4,    1,   -1,   -1,   -4,  -68,   35,   -5,    5,   -2,    1,    0,   35,   43,   -4,   -6,    1,   -1,    0,  -14,  -38,  -12,  -10,    9,    5,    7,    6,   -9,    7,   -4,   -3,    4,   -4,    0,    3  },
  {    0,    0,   19,   -4,    3,   -2,    2,   -1,   -3,  -13,   10,   -4,    1,    0,    0,    0,   -6,  -37,  -18,   -5,    2,   -2,    1,   -1,    6,   -6,   -7,   25,   -6,    4,   -1,    1,   16,   10,   55,  -24,   15,   46,  -52,    1,   35,  -43,   10,   12,  -23,   13,    5,   -8  },
  {   -3,    0,  -27,  -80,   40,  -16,    6,   -4,    4,    3,   31,   61,  -22,    7,   -1,    1,   -4,   -7,  -26,   -6,  -10,    6,   -4,    1,    3,    8,   14,  -18,   15,   -5,    2,   -1,   -2,   -4,   -1,   13,    0,    2,   -4,   -3,    3,   -1,    2,    1,   -2,    0,   -2,   -1  },
  {    1,    2,   -8,    6,   -1,    1,    0,    0,    2,    8,   -5,   -1,    0,    0,    0,    0,    1,   24,    3,    5,   -1,    1,    0,    0,   -3,   12,    6,  -10,    1,   -1,    0,    0,   -9,   -1,  -25,   10,   45,  -11,   18,    2,   86,    1,  -13,   -4,  -65,   -6,    7,    2  },
  {   -4,  -18,  -57,    8,   -8,    1,   -3,    0,   -5,  -20,  -69,    7,   -6,    2,   -2,    1,    1,    4,    0,   33,   -7,    5,   -2,    1,    0,   -9,   53,  -22,    3,   -1,    0,    0,    4,  -27,   -2,   -9,    5,   36,  -13,    5,   -7,  -17,    1,    2,    4,    6,    4,   -1  },
  }
 },
 {  //3
  {
  { -115,   37,    9,    2,    2,    1,    1,    0,   10,  -29,    8,    0,    1,    0,    1,    0,   23,   -8,   -8,    1,   -1,    0,    0,    0,    3,    3,   -2,   -1,    0,    0,    0,    0,    4,    0,    0,   -1,    1,    1,    0,    0,    2,    0,    0,    0,    0,    0,    0,    0  },
  {   15,   51,  -18,    0,   -3,    0,   -1,    0,  -95,    7,   34,   -3,    5,   -1,    2,    0,   23,  -47,    1,    6,    0,    1,    0,    1,    8,    5,  -12,    0,   -1,    0,    0,    0,    3,   -3,    1,   -1,    2,    1,   -2,    0,    1,   -1,    0,    0,    1,    1,   -1,    0  },
  {   29,  -22,   16,   -6,    3,   -2,    1,   -1,   -4,  -80,   12,   15,    0,    3,    0,    1,   45,    7,  -59,    7,   -2,    1,   -1,    0,  -15,   41,   -3,  -16,    2,   -3,    0,   -1,    1,    0,    7,   -2,   -3,    6,    1,   -2,    0,    0,    1,    0,   -1,    2,    0,   -1  },
  {  -36,  -98,   25,    5,    4,    1,    2,    1,  -59,   11,  -17,    1,    1,    1,    0,    0,    6,  -13,    7,   -3,    0,    0,    0,    0,   14,   -4,  -14,    3,   -1,    0,    0,    0,    2,    8,   -3,   -5,    2,    0,    0,    0,    0,    3,    0,   -1,    1,    0,    0,    0  },
  {   -6,   18,    3,   -3,   -1,    0,    0,    0,  -50,   -5,  -38,   12,    0,    2,    0,    1,    3,   67,   -7,  -40,    3,   -6,    1,   -3,  -12,  -13,   65,   -3,  -10,    0,   -1,    0,    9,  -20,   -5,   22,   -2,    0,    0,   -1,    2,   -3,   -2,    3,   -1,    0,    1,    0  },
  {    4,   15,   52,  -13,    5,   -3,    2,   -1,  -17,  -45,   16,   24,   -2,    4,   -1,    2,  -87,   -8,  -14,    7,    8,    1,    2,    0,   23,  -35,   -6,   -3,    1,    1,    0,    0,    2,    5,  -17,    0,    3,   -1,   -1,   -5,    0,    1,   -4,    0,    1,    0,    0,   -2  },
  {  -20,   -7,  -43,    4,    0,    1,   -1,    1,   -7,   35,    0,   12,   -4,    1,   -1,    0,  -51,   -2,  -57,    5,   15,    0,    4,    0,    7,   39,    5,  -55,    1,   -7,    1,   -3,    1,  -10,   41,    2,    4,   -3,   -2,    3,   -1,   -2,    7,    1,    1,   -1,   -1,    0  },
  {    4,   29,    1,   26,   -5,    4,   -2,    1,  -17,   -7,  -73,    6,    6,    2,    1,    1,   -5,   21,   -3,    5,   -1,   -3,    0,   -1,  -11,    2,  -52,   -3,   27,   -2,    5,    0,    0,   27,    8,  -58,    2,   -5,   25,    3,    0,    3,    0,   -5,    0,   -2,    7,    0  },
  {   12,   13,   10,    2,   -1,    3,   -1,    1,   17,   -2,  -46,   12,    7,    0,    2,    0,   16,  -45,   -9,  -53,    6,    1,    1,    0,   70,   16,    8,   -4,  -37,    1,   -7,    0,  -12,   29,    3,   21,    4,    0,    5,   -1,   -3,    4,    1,    4,    2,    0,    1,    0  },
  {    5,   20,   90,  -17,    4,   -3,    2,   -1,    6,   66,    8,   28,   -7,    3,   -1,    1,   29,    5,  -19,   12,    9,   -1,    1,    0,  -10,   14,   -1,  -13,    7,    0,    1,    0,    0,   -6,   13,   -4,    0,   -4,    1,    5,    0,   -1,   -1,    1,    0,   -1,    0,    0  },
  {   -3,   -4,  -34,  -12,    2,   -1,   -1,    0,    5,   25,   11,   43,  -10,    4,   -2,    1,   23,   20,  -40,   12,   21,   -3,    4,   -1,   25,  -28,  -10,    5,    8,    6,    0,    2,   -4,   21,  -64,   -8,   -5,   19,   10,  -48,    3,   -1,   10,   -3,    0,    4,    3,   -6  },
  {   -1,   -3,    2,   19,   -2,    4,   -1,    2,    9,    3,  -35,   22,   11,    1,    2,    0,   -7,  -65,  -19,  -22,   11,    4,    2,    1,  -75,  -18,    3,   -1,  -10,    2,    0,    1,    2,  -35,  -27,    4,    1,    8,  -17,  -19,    3,    0,    3,   -6,    0,    2,   -1,   -2  },
  {   10,   -4,   -6,   12,    5,    1,    1,    0,   11,   -9,  -12,   -2,   -7,    0,   -1,    0,   33,  -10,   -4,   18,   18,   -4,    4,   -1,   28,  -72,    1,  -49,   15,    2,    2,    1,   56,  -23,   22,   -1,    4,   -1,  -15,   26,    6,    4,  -10,    0,    0,    2,   -3,    2  },
  {    4,    6,   14,   53,   -4,    4,    0,    2,    0,   -1,  -20,  -13,    3,    2,   -1,    1,   -3,    1,   -5,   35,  -16,   -6,   -1,   -2,   46,   29,   13,   21,   37,   -5,    4,   -1,  -10,  -53,  -18,    8,    9,   12,  -41,  -25,   -2,    2,   13,  -16,    4,    1,   -5,    1  },
  {    2,    9,   13,   37,   19,    6,    2,    2,   -9,   -3,   -9,  -28,  -20,   -4,   -3,   -1,    1,   18,    9,   28,   24,    6,    2,    2,  -20,   -5,  -25,  -33,  -36,    9,   -2,    2,  -13,   42,    1,   57,  -22,   -2,  -25,  -28,    5,    6,   19,  -12,   -5,   -3,   -2,    4  },
  {    3,   -3,   12,   84,  -12,    8,   -2,    3,    6,   13,   50,   -1,   45,    1,    7,    0,   -2,   18,  -22,  -37,  -13,   14,    0,    3,    1,  -12,   -3,    2,  -15,   -8,    1,   -1,   19,   14,   -4,  -12,   -4,    5,   17,    8,    2,   -4,   -4,    4,   -2,    2,    1,    0  },
  },
  {
  {  109,  -26,   -8,   -3,   -2,   -1,   -1,    0,  -50,   28,    2,    1,    0,    0,    0,    0,  -18,   -8,    6,    0,    1,    0,    1,    0,    6,   -2,   -3,    0,    0,    0,    0,    0,   -3,    2,    1,   -1,    0,    0,    0,    0,   -2,    0,    0,    0,    0,    0,    0,    0  },
  {  -39,   31,   -5,    2,   -1,    1,    0,    0,  -95,    6,   18,    0,    4,    0,    1,    0,   32,  -49,    5,    1,    1,    0,    0,    0,   27,   -1,  -14,    2,   -2,    1,   -1,    0,    3,    5,   -3,   -2,    4,    1,   -1,   -1,    2,    0,    0,    0,    2,    0,    0,    0  },
  {   29,   -3,   -2,   -2,    0,    0,    0,    0,    0,  -41,    9,    0,    2,    0,    1,    0,   86,    4,  -33,    2,   -6,    1,   -2,    0,  -32,   58,    1,   -7,    0,   -2,    0,   -1,  -14,   -8,   20,    0,   -2,   -3,    0,    4,   -1,   -1,    0,    0,   -1,    1,    0,    0  },
  {   18,   96,  -23,    2,   -5,    1,   -2,    0,  -10,    6,   10,   -2,    1,   -1,    1,    0,  -14,   26,    2,   -4,    1,   -1,    0,    0,  -43,   -9,   35,   -2,    4,   -1,    1,    0,   14,  -40,    1,   10,    2,    1,  -10,    1,    2,   -4,   -1,   -1,    0,    0,   -1,    0  },
  {  -29,  -60,   16,   -2,    3,   -1,    1,    0,  -52,    9,  -17,    5,   -2,    1,   -1,    1,   13,   56,   -2,   -9,    0,   -2,    0,   -1,  -34,  -18,   41,    0,    3,    0,    1,    0,   19,  -36,  -10,   13,    3,    6,  -14,   -1,    3,    1,   -1,   -3,    1,    1,   -1,   -1  },
  {  -23,   -5,  -15,    5,   -2,    1,   -1,    1,    2,   79,  -13,   -4,   -2,   -1,   -1,    0,   -9,    1,    5,   -1,    1,    0,    0,    0,   -4,   49,    2,  -14,    1,   -3,    0,   -1,  -31,  -14,   56,   -1,   13,  -37,   -4,   20,   -2,    2,  -10,    0,    2,   -4,    0,   -1  },
  {   -7,   -3,   12,   -3,    3,   -1,    1,    0,  -31,  -62,    8,    7,    0,    2,    0,    1,  -75,    9,  -45,    5,   -1,    1,   -1,    0,   14,   35,    0,  -23,    2,   -5,    1,   -2,    1,   -8,   32,   -1,    7,  -12,   -4,   10,    0,    2,   -6,   -1,    2,    0,    0,   -2  },
  {    1,  -26,    5,    0,    1,    0,    1,    0,   24,   -3,   43,   -6,    4,   -2,    1,   -1,   -7,  -64,    9,   14,    0,    3,    0,    1,  -12,   -4,    5,    3,   -1,    1,    0,    0,    8,  -59,   -3,   26,   14,    6,  -58,    6,   -5,   17,   -7,  -18,    3,    3,   -1,   -5  },
  {   11,   14,    6,   -3,    1,   -1,    1,    0,   10,   -7,   -9,    3,   -2,    1,   -1,    0,   22,   21,    1,  -21,    2,   -4,    1,   -2,   92,    1,   53,    0,   -9,    1,   -2,    0,  -21,  -11,    1,   40,   -5,   -4,  -24,    5,   -4,    5,   -6,   -5,    0,    0,    0,   -3  },
  {  -10,  -11,  -47,    3,   -4,    1,   -1,    0,    5,   28,   11,   -2,   -1,    0,    0,    0,  -12,   -2,  -38,    2,    0,    1,    0,    0,   16,   38,   11,  -16,   -1,   -3,    0,   -2,   12,   -9,  -22,    7,   -8,   60,    4,  -36,   -6,  -15,   54,    7,    3,   -7,   -8,   14  },
  {   -8,  -24,  -99,   11,  -10,    3,   -4,    1,   -5,  -36,   19,  -26,    4,   -5,    1,   -2,    0,   25,   41,    5,   -3,    1,    0,    0,   10,   -5,   -7,   12,    2,    1,    0,    0,   -1,    1,    9,   -3,   -3,  -14,   -3,   12,    2,    4,  -13,   -2,   -1,    3,    2,   -4  },
  {   -5,    1,   -1,    0,    1,    0,    0,    0,  -10,  -14,   -6,    8,    0,    1,    0,    0,  -17,   -2,    7,   -5,    3,   -1,    0,    0,  -16,   13,    3,   31,   -1,    6,    0,    2,  -93,  -15,  -46,   -3,   23,  -19,    0,  -47,    8,    4,    8,    3,    2,    3,    0,    0  },
  {    1,   12,  -20,   21,   -4,    5,   -2,    2,   -5,   -2,  -75,    9,   -1,    2,   -1,    1,   -1,   -2,  -16,   -4,    0,   -1,    0,    0,   -7,    7,  -31,    0,    3,    0,    0,    0,    4,   11,  -12,    4,  -12,   14,  -50,   -1,   -8,   32,   -4,  -54,    2,    0,   30,  -15  },
  {    2,   -9,  -18,    8,   -3,    3,   -1,    1,    3,  -25,  -62,   -6,    0,   -2,    0,   -1,   -6,  -61,   14,  -51,    2,   -6,    0,   -2,  -19,    0,   40,   -7,  -17,    0,   -3,    0,   13,   -4,   11,    9,   17,    0,   24,    5,    1,  -12,    4,   28,    0,    0,  -15,    8  },
  {    4,    9,   39,   18,    0,    2,    0,    1,   -6,  -16,  -22,  -37,    5,   -5,    1,   -2,   -5,   15,   63,    9,  -16,    0,   -3,    0,   18,   42,  -18,   27,   15,    1,    3,    1,   12,  -34,    9,  -24,    4,   28,   -2,    4,  -11,   -4,   30,    2,    5,  -13,   -4,   18  },
  {   -7,   -2,   15,   -6,    1,   -1,    1,   -1,  -11,   -3,   22,  -14,    0,   -2,    1,   -1,  -18,   -7,   30,   -9,   -4,    0,   -1,    0,  -35,   23,   23,   10,  -17,    1,   -3,    0,  -19,   53,    6,   48,  -65,   12,  -12,   11,   -8,  -16,   10,  -21,   -2,  -12,    6,    2  },
  }
 }
};

const int8_t g_lfnst4x4[ 4 ][ 2 ][ 16 ][ 16 ] = {
 {  //0
  {
  {  108,  -44,  -15,    1,  -44,   19,    7,   -1,  -11,    6,    2,   -1,    0,   -1,   -1,    0  },
  {  -40,  -97,   56,   12,  -11,   29,  -12,   -3,   18,   18,  -15,   -3,   -1,   -3,    2,    1  },
  {   25,  -31,   -1,    7,  100,  -16,  -29,    1,  -54,   21,   14,   -4,   -7,    2,    4,    0  },
  {  -32,  -39,  -92,   51,   -6,  -16,   36,   -8,    3,   22,   18,  -15,    4,    1,   -5,    2  },
  {    8,   -9,   33,   -8,  -16, -102,   36,   23,   -4,   38,  -27,   -5,    5,   16,   -8,   -6  },
  {  -25,    5,   16,   -3,  -38,   14,   11,   -3,  -97,    7,   26,    1,   55,  -10,  -19,    3  },
  {    8,    9,   16,    1,   37,   36,   94,  -38,   -7,    3,  -47,   11,   -6,  -13,  -17,   10  },
  {    2,   34,   -5,    1,   -7,   24,  -25,   -3,    8,   99,  -28,  -29,    6,  -43,   21,   11  },
  {  -16,  -27,  -39, -109,    6,   10,   16,   24,    3,   19,   10,   24,   -4,   -7,   -2,   -3  },
  {   -9,  -10,  -34,    4,   -9,   -5,  -29,    5,  -33,  -26,  -96,   33,   14,    4,   39,  -14  },
  {  -13,    1,    4,   -9,  -30,  -17,   -3,  -64,  -35,   11,   17,   19,  -86,    6,   36,   14  },
  {    8,   -7,   -5,  -15,    7,  -30,  -28,  -87,   31,    4,    4,   33,   61,   -5,  -17,   22  },
  {   -2,   13,   -6,   -4,   -2,   28,  -13,  -14,   -3,   37,  -15,   -3,   -2,  107,  -36,  -24  },
  {    4,    9,   11,   31,    4,    9,   16,   19,   12,   33,   32,   94,   12,    0,   34,  -45  },
  {    2,   -2,    8,  -16,    8,    5,   28,  -17,    6,   -7,   18,  -45,   40,   36,   97,   -8  },
  {    0,   -2,    0,  -10,   -1,   -7,   -3,  -35,   -1,   -7,   -2,  -32,   -6,  -33,  -16, -112  },
  },
  {
  {  119,  -30,  -22,   -3,  -23,   -2,    3,    2,  -16,    3,    6,    0,   -3,    2,    1,    0  },
  {  -27, -101,   31,   17,  -47,    2,   22,    3,   19,   30,   -7,   -9,    5,    3,   -5,   -1  },
  {    0,   58,   22,  -15, -102,    2,   38,    2,   10,  -13,   -5,    4,   14,   -1,   -9,    0  },
  {   23,    4,   66,  -11,   22,   89,   -2,  -26,   13,   -8,  -38,   -1,   -9,  -20,   -2,    8  },
  {  -19,   -5,  -89,    2,  -26,   76,  -11,  -17,   20,   13,   18,   -4,    1,  -15,    3,    5  },
  {  -10,   -1,   -1,    6,   23,   25,   87,   -7,  -74,    4,   39,   -5,    0,   -1,  -20,   -1  },
  {  -17,  -28,   12,   -8,  -32,   14,  -53,   -6,  -68,  -67,   17,   29,    2,    6,   25,    4  },
  {    1,  -24,  -23,    1,   17,   -7,   52,    9,   50,  -92,  -15,   27,  -15,  -10,   -6,    3  },
  {   -6,  -17,   -2, -111,    7,  -17,    8,  -42,    9,   18,   16,   25,   -4,    2,   -1,   11  },
  {    9,    5,   35,    0,    6,   21,   -9,   34,   44,   -3,  102,   11,   -7,   13,   11,  -20  },
  {    4,   -5,   -5,  -10,   15,   19,   -2,    6,    6,  -12,  -13,    6,   95,   69,  -29,  -24  },
  {   -6,   -4,   -9,  -39,    1,   22,    0,  102,  -19,   19,  -32,   30,  -16,  -14,   -8,  -23  },
  {    4,   -4,    7,    8,    4,  -13,  -18,    5,    0,    0,   21,   22,   58,  -88,  -54,   28  },
  {   -4,   -7,    0,  -24,   -7,    0,  -25,    3,   -3,  -30,    8,  -76,  -34,    4,  -80,  -26  },
  {    0,    6,    0,   30,   -6,    1,  -13,  -23,    1,   20,   -2,   80,  -44,   37,  -68,    1  },
  {    0,    0,   -1,    5,   -1,   -7,    1,  -34,   -2,    3,   -6,   19,    5,  -38,   11, -115  },
  }
 },
 {  //1
  {
  { -111,   39,    4,    3,   44,   11,  -12,   -1,    7,  -16,   -5,    2,    3,   -1,    4,    2  },
  {  -47,  -27,   15,   -1,  -92,   43,   20,   -2,   20,   39,  -16,   -5,   10,   -5,  -13,    2  },
  {  -35,  -23,    4,    4,  -17,  -72,   32,    6,  -59,   18,   50,   -6,    0,   40,    0,  -13  },
  {   13,   93,  -27,   -4,  -48,   13,  -34,    4,  -52,   11,    1,   10,    3,   16,   -3,    1  },
  {  -11,  -27,    1,    2,  -47,   -4,  -36,   10,   -2,  -85,   14,   29,  -20,   -2,   57,    4  },
  {    0,  -35,   32,   -2,   26,   60,   -3,  -17,  -82,    1,  -30,    0,  -37,   21,    3,   12  },
  {  -17,  -46,  -92,   14,    7,  -10,  -39,   29,  -17,   27,  -28,   17,    1,  -15,  -13,   17  },
  {    4,  -10,  -23,    4,   16,   58,  -17,   26,   30,   21,   67,    2,  -13,   59,   13,  -40  },
  {    5,  -20,   32,   -5,    8,   -3,  -46,   -7,   -4,    2,  -15,   24,  100,   44,    0,    5  },
  {   -4,   -1,   38,  -18,   -7,  -42,  -63,   -6,   33,   34,  -23,   15,  -65,   33,  -20,    2  },
  {   -2,  -10,   35,  -19,    5,    8,  -44,   14,  -25,   25,   58,   17,    7,  -84,  -16,  -18  },
  {    5,   13,   18,   34,   11,   -4,   18,   18,    5,   58,   -3,   42,   -2,  -10,   85,   38  },
  {   -5,   -7,  -34,  -83,    2,   -1,   -4,  -73,    4,   20,   15,  -12,    4,   -3,   44,   12  },
  {    0,    4,   -2,  -60,    5,    9,   42,   34,    5,  -14,    9,   80,   -5,   13,  -38,   37  },
  {   -1,    2,    7,  -57,    3,   -7,    9,   68,   -9,    6,  -49,  -20,    6,   -4,   36,  -64  },
  {   -1,    0,  -12,   23,    1,   -4,   17,  -53,   -3,    4,  -21,   72,   -4,   -8,   -3,  -83  },
  },
  {
  {   88,  -55,    6,   -3,  -66,   27,    9,   -2,   11,   11,  -13,    1,   -2,   -7,    1,    2  },
  {  -58,  -20,   27,   -2,  -27,   75,  -29,    0,   47,  -42,  -11,   11,   -9,   -3,   19,   -4  },
  {  -51,   23,  -22,    5,  -63,    3,   37,   -5,    1,   64,  -35,   -4,   29,  -31,  -11,   13  },
  {  -27,  -76,   49,   -2,   40,   14,    9,  -17,  -56,   36,  -25,    6,   14,    3,   -6,    8  },
  {   19,   -4,  -36,   22,   52,    7,   36,  -23,   28,  -17,  -64,   15,   -5,  -44,   48,    9  },
  {   29,   50,   13,  -10,    1,   34,  -59,    1,  -51,    4,  -16,   30,   52,  -33,   24,   -5  },
  {  -12,  -21,  -74,   43,  -13,   39,   18,   -5,  -58,  -35,   27,   -5,   19,   26,    6,   -5  },
  {   19,   38,  -10,   -5,   28,   66,    0,   -5,   -4,   19,  -30,  -26,  -40,   28,  -60,   37  },
  {   -6,   27,   18,   -5,  -37,  -18,   12,  -25,  -44,  -10,  -38,   37,  -66,   45,   40,   -7  },
  {  -13,  -28,  -45,  -39,    0,   -5,  -39,   69,  -23,   16,  -12,  -18,  -50,  -31,   24,   13  },
  {   -1,    8,   24,  -51,  -15,   -9,   44,   10,  -28,  -70,  -12,  -39,   24,  -18,   -4,   51  },
  {   -8,  -22,  -17,   33,  -18,  -45,  -57,  -27,    0,  -31,  -30,   29,   -2,  -13,  -53,   49  },
  {    1,   12,   32,   51,   -8,    8,   -2,  -31,  -22,    4,   46,  -39,  -49,  -67,   14,   17  },
  {    4,    5,   24,   60,   -5,  -14,  -23,   38,    9,    8,  -34,  -59,   24,   47,   42,   28  },
  {   -1,   -5,  -20,  -34,    4,    4,  -15,  -46,   18,   31,   42,   10,   10,   27,   49,   78  },
  {   -3,   -7,  -22,  -34,   -5,  -11,  -36,  -69,   -1,   -3,  -25,  -73,    5,    4,    4,  -49  },
  }
 },
 {  //2
  {
  { -112,   47,   -2,    2,  -34,   13,    2,    0,   15,   -7,    1,    0,    8,   -3,   -1,    0  },
  {   29,   -7,    1,   -1, -108,   40,    2,    0,  -45,   13,    4,   -1,    8,   -5,    1,    0  },
  {  -36,  -87,   69,  -10,  -17,  -33,   26,   -2,    7,   14,  -11,    2,    6,    8,   -7,    0  },
  {   28,   -5,    2,   -2,  -29,   13,   -2,    0,  103,  -36,   -4,    1,   48,  -16,   -4,    1  },
  {  -12,  -24,   15,   -3,   26,   80,  -61,    9,   15,   54,  -36,    2,    0,   -4,    6,   -2  },
  {   18,   53,   69,  -74,   14,   24,   28,  -30,   -6,   -7,  -11,   12,   -5,   -7,   -6,    8  },
  {    5,   -1,    2,    0,  -26,    6,    0,    1,   45,   -9,   -1,    0, -113,   28,    8,   -1  },
  {  -13,  -32,   18,   -2,   15,   34,  -27,    7,  -25,  -80,   47,   -1,  -16,  -50,   28,    2  },
  {   -4,  -13,  -10,   19,   18,   46,   60,  -48,   16,   33,   60,  -48,    1,    0,    5,   -2  },
  {   15,   33,   63,   89,    8,   15,   25,   40,   -4,   -8,  -15,   -8,   -2,   -6,   -9,   -7  },
  {   -8,  -24,  -27,   15,   12,   41,   26,  -29,  -17,  -50,  -39,   27,    0,   35,  -67,   26  },
  {   -2,   -6,  -24,   13,   -1,   -8,   37,  -22,    3,   18,  -51,   22,  -23,  -95,   17,   17  },
  {   -3,   -7,  -16,  -21,   10,   24,   46,   75,    8,   20,   38,   72,    1,    2,    1,    7  },
  {    2,    6,   10,   -3,   -5,  -16,  -31,   12,    7,   24,   41,  -16,  -16,  -41,  -89,   49  },
  {    4,    8,   21,   40,   -4,  -11,  -28,  -57,    5,   14,   31,   70,    7,   18,   32,   52  },
  {    0,    1,    4,   11,   -2,   -4,  -13,  -34,    3,    7,   20,   47,   -6,  -19,  -42, -101  },
  },
  {
  {  -99,   39,   -1,    2,   65,  -20,   -5,    0,  -15,   -2,    5,   -1,    0,    3,   -1,    0  },
  {   58,   42,  -33,    3,   33,  -63,   23,   -1,  -55,   32,    3,   -5,   21,   -2,   -8,    3  },
  {  -15,   71,  -44,    5,  -58,  -29,   25,    3,   62,   -7,   -4,   -4,  -19,    4,    0,    1  },
  {   46,    5,    4,   -6,   71,  -12,  -15,    5,   52,  -38,   13,   -2,  -63,   23,    3,   -3  },
  {  -14,  -54,  -29,   29,   25,   -9,   61,  -29,   27,   44,  -48,    5,  -27,  -21,   12,    7  },
  {   -3,    3,   69,  -42,  -11,  -50,  -26,   26,   24,   63,  -19,   -5,  -18,  -22,   12,    0  },
  {   17,   16,   -2,    1,   38,   18,  -12,    0,   62,    1,  -14,    5,   89,  -42,    8,   -2  },
  {   15,   54,   -8,    6,    6,   60,  -26,   -8,  -30,   17,  -38,   22,  -43,  -45,   42,   -7  },
  {   -6,  -17,  -55,  -28,    9,   30,   -8,   58,    4,   34,   41,  -52,  -16,  -36,  -20,   16  },
  {   -2,   -1,   -9,  -79,    7,   11,   48,   44,  -13,  -34,  -55,    6,   12,   23,   20,  -11  },
  {    7,   29,   14,   -6,   12,   53,   10,  -11,   14,   59,  -15,   -3,    5,   71,  -54,   13  },
  {   -5,  -24,  -53,   15,   -3,  -15,  -61,   26,    6,   30,  -16,   23,   13,   56,   44,  -35  },
  {    4,    8,   21,   52,   -1,   -1,   -5,   29,   -7,  -17,  -44,  -84,    8,   20,   31,   39  },
  {   -2,  -11,  -25,   -4,   -4,  -21,  -53,    2,   -5,  -26,  -64,   19,   -8,  -19,  -73,   39  },
  {   -3,   -5,  -23,  -57,   -2,   -4,  -24,  -75,    1,    3,    9,  -25,    6,   15,   41,   61  },
  {    1,    1,    7,   18,    1,    2,   16,   47,    2,    5,   24,   67,    3,    9,   25,   88  },
  }
 },
 {  //3
  {
  { -114,   37,    3,    2,  -22,  -23,   14,    0,   21,  -17,   -5,    2,    5,    2,   -4,   -1  },
  {  -19,  -41,   19,   -2,   85,  -60,  -11,    7,   17,   31,  -34,    2,  -11,   19,    2,   -8  },
  {   36,  -25,   18,   -2,  -42,  -53,   35,    5,   46,  -60,  -25,   19,    8,   21,  -33,   -1  },
  {  -27,  -80,   44,   -3,  -58,    1,  -29,   19,  -41,   18,  -12,   -7,   12,  -17,    7,   -6  },
  {  -11,  -21,   37,  -10,   44,   -4,   47,  -12,  -37,  -41,   58,   18,   10,  -46,  -16,   31  },
  {   15,   47,   10,   -6,  -16,  -44,   42,   10,  -80,   25,  -40,   21,  -23,   -2,    3,  -14  },
  {   13,   25,   79,  -39,  -13,   10,   31,   -4,   49,   45,   12,   -8,    3,   -1,   43,    7  },
  {   16,   11,  -26,   13,  -13,  -74,  -20,   -1,    5,   -6,   29,  -47,   26,  -49,   54,    2  },
  {   -8,  -34,  -26,    7,  -26,  -19,   29,  -37,    1,   22,   46,   -9,  -81,   37,   14,   20  },
  {   -6,  -30,  -42,  -12,   -3,    5,   57,  -52,   -2,   37,  -12,    6,   74,   10,    6,  -15  },
  {    5,    9,   -6,   42,  -15,  -18,   -9,   26,   15,   58,   14,   43,   23,  -10,  -37,   75  },
  {   -5,  -23,  -23,   36,    3,   22,   36,   40,   27,   -4,  -16,   56,  -25,  -46,   56,  -24  },
  {    1,    3,   23,   73,    8,    5,   34,   46,  -12,    2,   35,  -38,   26,   52,    2,  -31  },
  {   -3,   -2,  -21,  -52,    1,  -10,  -17,   44,  -19,  -20,   30,   45,   27,   61,   49,   21  },
  {   -2,   -7,  -33,  -56,   -4,   -6,   21,   63,   15,   31,   32,  -22,  -10,  -26,  -52,  -38  },
  {   -5,  -12,  -18,  -12,    8,   22,   38,   36,   -5,  -15,  -51,  -63,   -5,    0,   15,   73  },
  },
  {
  { -102,   22,    7,    2,   66,  -25,   -6,   -1,  -15,   14,    1,   -1,    2,   -2,    1,    0  },
  {   12,   93,  -27,   -6,  -27,  -64,   36,    6,   13,    5,  -23,    0,   -2,    6,    5,   -3  },
  {  -59,  -24,   17,    1,  -62,   -2,   -3,    2,   83,  -12,  -17,   -2,  -24,   14,    7,   -2  },
  {  -33,   23,  -36,   11,  -21,   50,   35,  -16,  -23,  -78,   16,   19,   22,   15,  -30,   -5  },
  {    0,  -38,  -81,   30,   27,    5,   51,  -32,   24,   36,  -16,   12,  -24,   -8,    9,    1  },
  {   28,   38,    8,   -9,   62,   32,  -13,    2,   51,  -32,   15,    5,  -66,   28,    0,   -1  },
  {   11,  -35,   21,  -17,   30,  -18,   31,   18,  -11,  -36,  -80,   12,   16,   49,   13,  -32  },
  {  -13,   23,   22,  -36,  -12,   64,   39,   25,  -19,   23,  -36,    9,  -30,  -58,   33,   -7  },
  {   -9,  -20,  -55,  -83,    3,   -2,    1,   62,    8,    2,   27,  -28,    7,   15,  -11,    5  },
  {   -6,   24,  -38,   23,   -8,   40,  -49,    0,   -7,    9,  -25,  -44,   23,   39,   70,   -3  },
  {   12,   17,   17,    0,   32,   27,   21,    2,   67,   11,   -6,  -10,   89,  -22,  -12,   16  },
  {    2,   -9,    8,   45,    7,   -8,   27,   35,   -9,  -31,  -17,  -87,  -23,  -22,  -19,   44  },
  {   -1,   -9,   28,  -24,   -1,  -10,   49,  -30,   -8,   -7,   40,    1,    4,   33,   65,   67  },
  {    5,  -12,  -24,  -17,   13,  -34,  -32,  -16,   14,  -67,   -7,    9,    7,  -74,   49,    1  },
  {    2,   -6,   11,   45,    3,  -10,   33,   55,    8,   -5,   59,    4,    7,   -4,   44,  -66  },
  {   -1,    1,  -14,   36,   -1,    2,  -20,   69,    0,    0,  -15,   72,    3,    4,    5,   65  },
  }
 }
};
#endif
#endif

//--------------------------------------------------------------------------------------------------
