/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2022, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     RomNSST.cpp
    \brief    Transformations
*/

#include "Rom.h"


// DCT-2
#define DEFINE_DCT2_P2_MATRIX(a) \
{ \
  {a,  a}, \
  {a, -a}  \
}

#define DEFINE_DCT2_P4_MATRIX(a,b,c) \
{ \
  { a,  a,  a,  a}, \
  { b,  c, -c, -b}, \
  { a, -a, -a,  a}, \
  { c, -b,  b, -c}  \
}

#define DEFINE_DCT2_P8_MATRIX(a,b,c,d,e,f,g) \
{ \
  { a,  a,  a,  a,  a,  a,  a,  a}, \
  { d,  e,  f,  g, -g, -f, -e, -d}, \
  { b,  c, -c, -b, -b, -c,  c,  b}, \
  { e, -g, -d, -f,  f,  d,  g, -e}, \
  { a, -a, -a,  a,  a, -a, -a,  a}, \
  { f, -d,  g,  e, -e, -g,  d, -f}, \
  { c, -b,  b, -c, -c,  b, -b,  c}, \
  { g, -f,  e, -d,  d, -e,  f, -g}  \
}

#define DEFINE_DCT2_P16_MATRIX(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o) \
{ \
  { a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a}, \
  { h,  i,  j,  k,  l,  m,  n,  o, -o, -n, -m, -l, -k, -j, -i, -h}, \
  { d,  e,  f,  g, -g, -f, -e, -d, -d, -e, -f, -g,  g,  f,  e,  d}, \
  { i,  l,  o, -m, -j, -h, -k, -n,  n,  k,  h,  j,  m, -o, -l, -i}, \
  { b,  c, -c, -b, -b, -c,  c,  b,  b,  c, -c, -b, -b, -c,  c,  b}, \
  { j,  o, -k, -i, -n,  l,  h,  m, -m, -h, -l,  n,  i,  k, -o, -j}, \
  { e, -g, -d, -f,  f,  d,  g, -e, -e,  g,  d,  f, -f, -d, -g,  e}, \
  { k, -m, -i,  o,  h,  n, -j, -l,  l,  j, -n, -h, -o,  i,  m, -k}, \
  { a, -a, -a,  a,  a, -a, -a,  a,  a, -a, -a,  a,  a, -a, -a,  a}, \
  { l, -j, -n,  h, -o, -i,  m,  k, -k, -m,  i,  o, -h,  n,  j, -l}, \
  { f, -d,  g,  e, -e, -g,  d, -f, -f,  d, -g, -e,  e,  g, -d,  f}, \
  { m, -h,  l,  n, -i,  k,  o, -j,  j, -o, -k,  i, -n, -l,  h, -m}, \
  { c, -b,  b, -c, -c,  b, -b,  c,  c, -b,  b, -c, -c,  b, -b,  c}, \
  { n, -k,  h, -j,  m,  o, -l,  i, -i,  l, -o, -m,  j, -h,  k, -n}, \
  { g, -f,  e, -d,  d, -e,  f, -g, -g,  f, -e,  d, -d,  e, -f,  g}, \
  { o, -n,  m, -l,  k, -j,  i, -h,  h, -i,  j, -k,  l, -m,  n, -o}  \
}

#define DEFINE_DCT2_P32_MATRIX(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z,A,B,C,D,E) \
{ \
  { a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a,  a}, \
  { p,  q,  r,  s,  t,  u,  v,  w,  x,  y,  z,  A,  B,  C,  D,  E, -E, -D, -C, -B, -A, -z, -y, -x, -w, -v, -u, -t, -s, -r, -q, -p}, \
  { h,  i,  j,  k,  l,  m,  n,  o, -o, -n, -m, -l, -k, -j, -i, -h, -h, -i, -j, -k, -l, -m, -n, -o,  o,  n,  m,  l,  k,  j,  i,  h}, \
  { q,  t,  w,  z,  C, -E, -B, -y, -v, -s, -p, -r, -u, -x, -A, -D,  D,  A,  x,  u,  r,  p,  s,  v,  y,  B,  E, -C, -z, -w, -t, -q}, \
  { d,  e,  f,  g, -g, -f, -e, -d, -d, -e, -f, -g,  g,  f,  e,  d,  d,  e,  f,  g, -g, -f, -e, -d, -d, -e, -f, -g,  g,  f,  e,  d}, \
  { r,  w,  B, -D, -y, -t, -p, -u, -z, -E,  A,  v,  q,  s,  x,  C, -C, -x, -s, -q, -v, -A,  E,  z,  u,  p,  t,  y,  D, -B, -w, -r}, \
  { i,  l,  o, -m, -j, -h, -k, -n,  n,  k,  h,  j,  m, -o, -l, -i, -i, -l, -o,  m,  j,  h,  k,  n, -n, -k, -h, -j, -m,  o,  l,  i}, \
  { s,  z, -D, -w, -p, -v, -C,  A,  t,  r,  y, -E, -x, -q, -u, -B,  B,  u,  q,  x,  E, -y, -r, -t, -A,  C,  v,  p,  w,  D, -z, -s}, \
  { b,  c, -c, -b, -b, -c,  c,  b,  b,  c, -c, -b, -b, -c,  c,  b,  b,  c, -c, -b, -b, -c,  c,  b,  b,  c, -c, -b, -b, -c,  c,  b}, \
  { t,  C, -y, -p, -x,  D,  u,  s,  B, -z, -q, -w,  E,  v,  r,  A, -A, -r, -v, -E,  w,  q,  z, -B, -s, -u, -D,  x,  p,  y, -C, -t}, \
  { j,  o, -k, -i, -n,  l,  h,  m, -m, -h, -l,  n,  i,  k, -o, -j, -j, -o,  k,  i,  n, -l, -h, -m,  m,  h,  l, -n, -i, -k,  o,  j}, \
  { u, -E, -t, -v,  D,  s,  w, -C, -r, -x,  B,  q,  y, -A, -p, -z,  z,  p,  A, -y, -q, -B,  x,  r,  C, -w, -s, -D,  v,  t,  E, -u}, \
  { e, -g, -d, -f,  f,  d,  g, -e, -e,  g,  d,  f, -f, -d, -g,  e,  e, -g, -d, -f,  f,  d,  g, -e, -e,  g,  d,  f, -f, -d, -g,  e}, \
  { v, -B, -p, -C,  u,  w, -A, -q, -D,  t,  x, -z, -r, -E,  s,  y, -y, -s,  E,  r,  z, -x, -t,  D,  q,  A, -w, -u,  C,  p,  B, -v}, \
  { k, -m, -i,  o,  h,  n, -j, -l,  l,  j, -n, -h, -o,  i,  m, -k, -k,  m,  i, -o, -h, -n,  j,  l, -l, -j,  n,  h,  o, -i, -m,  k}, \
  { w, -y, -u,  A,  s, -C, -q,  E,  p,  D, -r, -B,  t,  z, -v, -x,  x,  v, -z, -t,  B,  r, -D, -p, -E,  q,  C, -s, -A,  u,  y, -w}, \
  { a, -a, -a,  a,  a, -a, -a,  a,  a, -a, -a,  a,  a, -a, -a,  a,  a, -a, -a,  a,  a, -a, -a,  a,  a, -a, -a,  a,  a, -a, -a,  a}, \
  { x, -v, -z,  t,  B, -r, -D,  p, -E, -q,  C,  s, -A, -u,  y,  w, -w, -y,  u,  A, -s, -C,  q,  E, -p,  D,  r, -B, -t,  z,  v, -x}, \
  { l, -j, -n,  h, -o, -i,  m,  k, -k, -m,  i,  o, -h,  n,  j, -l, -l,  j,  n, -h,  o,  i, -m, -k,  k,  m, -i, -o,  h, -n, -j,  l}, \
  { y, -s, -E,  r, -z, -x,  t,  D, -q,  A,  w, -u, -C,  p, -B, -v,  v,  B, -p,  C,  u, -w, -A,  q, -D, -t,  x,  z, -r,  E,  s, -y}, \
  { f, -d,  g,  e, -e, -g,  d, -f, -f,  d, -g, -e,  e,  g, -d,  f,  f, -d,  g,  e, -e, -g,  d, -f, -f,  d, -g, -e,  e,  g, -d,  f}, \
  { z, -p,  A,  y, -q,  B,  x, -r,  C,  w, -s,  D,  v, -t,  E,  u, -u, -E,  t, -v, -D,  s, -w, -C,  r, -x, -B,  q, -y, -A,  p, -z}, \
  { m, -h,  l,  n, -i,  k,  o, -j,  j, -o, -k,  i, -n, -l,  h, -m, -m,  h, -l, -n,  i, -k, -o,  j, -j,  o,  k, -i,  n,  l, -h,  m}, \
  { A, -r,  v, -E, -w,  q, -z, -B,  s, -u,  D,  x, -p,  y,  C, -t,  t, -C, -y,  p, -x, -D,  u, -s,  B,  z, -q,  w,  E, -v,  r, -A}, \
  { c, -b,  b, -c, -c,  b, -b,  c,  c, -b,  b, -c, -c,  b, -b,  c,  c, -b,  b, -c, -c,  b, -b,  c,  c, -b,  b, -c, -c,  b, -b,  c}, \
  { B, -u,  q, -x,  E,  y, -r,  t, -A, -C,  v, -p,  w, -D, -z,  s, -s,  z,  D, -w,  p, -v,  C,  A, -t,  r, -y, -E,  x, -q,  u, -B}, \
  { n, -k,  h, -j,  m,  o, -l,  i, -i,  l, -o, -m,  j, -h,  k, -n, -n,  k, -h,  j, -m, -o,  l, -i,  i, -l,  o,  m, -j,  h, -k,  n}, \
  { C, -x,  s, -q,  v, -A, -E,  z, -u,  p, -t,  y, -D, -B,  w, -r,  r, -w,  B,  D, -y,  t, -p,  u, -z,  E,  A, -v,  q, -s,  x, -C}, \
  { g, -f,  e, -d,  d, -e,  f, -g, -g,  f, -e,  d, -d,  e, -f,  g,  g, -f,  e, -d,  d, -e,  f, -g, -g,  f, -e,  d, -d,  e, -f,  g}, \
  { D, -A,  x, -u,  r, -p,  s, -v,  y, -B,  E,  C, -z,  w, -t,  q, -q,  t, -w,  z, -C, -E,  B, -y,  v, -s,  p, -r,  u, -x,  A, -D}, \
  { o, -n,  m, -l,  k, -j,  i, -h,  h, -i,  j, -k,  l, -m,  n, -o, -o,  n, -m,  l, -k,  j, -i,  h, -h,  i, -j,  k, -l,  m, -n,  o}, \
  { E, -D,  C, -B,  A, -z,  y, -x,  w, -v,  u, -t,  s, -r,  q, -p,  p, -q,  r, -s,  t, -u,  v, -w,  x, -y,  z, -A,  B, -C,  D, -E}  \
}


#define DEFINE_DCT2_P64_MATRIX(aa, ab, ac, ad, ae, af, ag, ah, ai, aj, ak, al, am, an, ao, ap, aq, ar, as, at, au, av, aw, ax, ay, az, ba, bb, bc, bd, be, bf, bg, bh, bi, bj, bk, bl, bm, bn, bo, bp, bq, br, bs, bt, bu, bv, bw, bx, by, bz, ca, cb, cc, cd, ce, cf, cg, ch, ci, cj, ck) \
{ \
  { aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa,  aa }, \
  { bf,  bg,  bh,  bi,  bj,  bk,  bl,  bm,  bn,  bo,  bp,  bq,  br,  bs,  bt,  bu,  bv,  bw,  bx,  by,  bz,  ca,  cb,  cc,  cd,  ce,  cf,  cg,  ch,  ci,  cj,  ck, -ck, -cj, -ci, -ch, -cg, -cf, -ce, -cd, -cc, -cb, -ca, -bz, -by, -bx, -bw, -bv, -bu, -bt, -bs, -br, -bq, -bp, -bo, -bn, -bm, -bl, -bk, -bj, -bi, -bh, -bg, -bf }, \
  { ap,  aq,  ar,  as,  at,  au,  av,  aw,  ax,  ay,  az,  ba,  bb,  bc,  bd,  be, -be, -bd, -bc, -bb, -ba, -az, -ay, -ax, -aw, -av, -au, -at, -as, -ar, -aq, -ap, -ap, -aq, -ar, -as, -at, -au, -av, -aw, -ax, -ay, -az, -ba, -bb, -bc, -bd, -be,  be,  bd,  bc,  bb,  ba,  az,  ay,  ax,  aw,  av,  au,  at,  as,  ar,  aq,  ap }, \
  { bg,  bj,  bm,  bp,  bs,  bv,  by,  cb,  ce,  ch,  ck, -ci, -cf, -cc, -bz, -bw, -bt, -bq, -bn, -bk, -bh, -bf, -bi, -bl, -bo, -br, -bu, -bx, -ca, -cd, -cg, -cj,  cj,  cg,  cd,  ca,  bx,  bu,  br,  bo,  bl,  bi,  bf,  bh,  bk,  bn,  bq,  bt,  bw,  bz,  cc,  cf,  ci, -ck, -ch, -ce, -cb, -by, -bv, -bs, -bp, -bm, -bj, -bg }, \
  { ah,  ai,  aj,  ak,  al,  am,  an,  ao, -ao, -an, -am, -al, -ak, -aj, -ai, -ah, -ah, -ai, -aj, -ak, -al, -am, -an, -ao,  ao,  an,  am,  al,  ak,  aj,  ai,  ah,  ah,  ai,  aj,  ak,  al,  am,  an,  ao, -ao, -an, -am, -al, -ak, -aj, -ai, -ah, -ah, -ai, -aj, -ak, -al, -am, -an, -ao,  ao,  an,  am,  al,  ak,  aj,  ai,  ah }, \
  { bh,  bm,  br,  bw,  cb,  cg, -ck, -cf, -ca, -bv, -bq, -bl, -bg, -bi, -bn, -bs, -bx, -cc, -ch,  cj,  ce,  bz,  bu,  bp,  bk,  bf,  bj,  bo,  bt,  by,  cd,  ci, -ci, -cd, -by, -bt, -bo, -bj, -bf, -bk, -bp, -bu, -bz, -ce, -cj,  ch,  cc,  bx,  bs,  bn,  bi,  bg,  bl,  bq,  bv,  ca,  cf,  ck, -cg, -cb, -bw, -br, -bm, -bh }, \
  { aq,  at,  aw,  az,  bc, -be, -bb, -ay, -av, -as, -ap, -ar, -au, -ax, -ba, -bd,  bd,  ba,  ax,  au,  ar,  ap,  as,  av,  ay,  bb,  be, -bc, -az, -aw, -at, -aq, -aq, -at, -aw, -az, -bc,  be,  bb,  ay,  av,  as,  ap,  ar,  au,  ax,  ba,  bd, -bd, -ba, -ax, -au, -ar, -ap, -as, -av, -ay, -bb, -be,  bc,  az,  aw,  at,  aq }, \
  { bi,  bp,  bw,  cd,  ck, -ce, -bx, -bq, -bj, -bh, -bo, -bv, -cc, -cj,  cf,  by,  br,  bk,  bg,  bn,  bu,  cb,  ci, -cg, -bz, -bs, -bl, -bf, -bm, -bt, -ca, -ch,  ch,  ca,  bt,  bm,  bf,  bl,  bs,  bz,  cg, -ci, -cb, -bu, -bn, -bg, -bk, -br, -by, -cf,  cj,  cc,  bv,  bo,  bh,  bj,  bq,  bx,  ce, -ck, -cd, -bw, -bp, -bi }, \
  { ad,  ae,  af,  ag, -ag, -af, -ae, -ad, -ad, -ae, -af, -ag,  ag,  af,  ae,  ad,  ad,  ae,  af,  ag, -ag, -af, -ae, -ad, -ad, -ae, -af, -ag,  ag,  af,  ae,  ad,  ad,  ae,  af,  ag, -ag, -af, -ae, -ad, -ad, -ae, -af, -ag,  ag,  af,  ae,  ad,  ad,  ae,  af,  ag, -ag, -af, -ae, -ad, -ad, -ae, -af, -ag,  ag,  af,  ae,  ad }, \
  { bj,  bs,  cb,  ck, -cc, -bt, -bk, -bi, -br, -ca, -cj,  cd,  bu,  bl,  bh,  bq,  bz,  ci, -ce, -bv, -bm, -bg, -bp, -by, -ch,  cf,  bw,  bn,  bf,  bo,  bx,  cg, -cg, -bx, -bo, -bf, -bn, -bw, -cf,  ch,  by,  bp,  bg,  bm,  bv,  ce, -ci, -bz, -bq, -bh, -bl, -bu, -cd,  cj,  ca,  br,  bi,  bk,  bt,  cc, -ck, -cb, -bs, -bj }, \
  { ar,  aw,  bb, -bd, -ay, -at, -ap, -au, -az, -be,  ba,  av,  aq,  as,  ax,  bc, -bc, -ax, -as, -aq, -av, -ba,  be,  az,  au,  ap,  at,  ay,  bd, -bb, -aw, -ar, -ar, -aw, -bb,  bd,  ay,  at,  ap,  au,  az,  be, -ba, -av, -aq, -as, -ax, -bc,  bc,  ax,  as,  aq,  av,  ba, -be, -az, -au, -ap, -at, -ay, -bd,  bb,  aw,  ar }, \
  { bk,  bv,  cg, -ce, -bt, -bi, -bm, -bx, -ci,  cc,  br,  bg,  bo,  bz,  ck, -ca, -bp, -bf, -bq, -cb,  cj,  by,  bn,  bh,  bs,  cd, -ch, -bw, -bl, -bj, -bu, -cf,  cf,  bu,  bj,  bl,  bw,  ch, -cd, -bs, -bh, -bn, -by, -cj,  cb,  bq,  bf,  bp,  ca, -ck, -bz, -bo, -bg, -br, -cc,  ci,  bx,  bm,  bi,  bt,  ce, -cg, -bv, -bk }, \
  { ai,  al,  ao, -am, -aj, -ah, -ak, -an,  an,  ak,  ah,  aj,  am, -ao, -al, -ai, -ai, -al, -ao,  am,  aj,  ah,  ak,  an, -an, -ak, -ah, -aj, -am,  ao,  al,  ai,  ai,  al,  ao, -am, -aj, -ah, -ak, -an,  an,  ak,  ah,  aj,  am, -ao, -al, -ai, -ai, -al, -ao,  am,  aj,  ah,  ak,  an, -an, -ak, -ah, -aj, -am,  ao,  al,  ai }, \
  { bl,  by, -ck, -bx, -bk, -bm, -bz,  cj,  bw,  bj,  bn,  ca, -ci, -bv, -bi, -bo, -cb,  ch,  bu,  bh,  bp,  cc, -cg, -bt, -bg, -bq, -cd,  cf,  bs,  bf,  br,  ce, -ce, -br, -bf, -bs, -cf,  cd,  bq,  bg,  bt,  cg, -cc, -bp, -bh, -bu, -ch,  cb,  bo,  bi,  bv,  ci, -ca, -bn, -bj, -bw, -cj,  bz,  bm,  bk,  bx,  ck, -by, -bl }, \
  { as,  az, -bd, -aw, -ap, -av, -bc,  ba,  at,  ar,  ay, -be, -ax, -aq, -au, -bb,  bb,  au,  aq,  ax,  be, -ay, -ar, -at, -ba,  bc,  av,  ap,  aw,  bd, -az, -as, -as, -az,  bd,  aw,  ap,  av,  bc, -ba, -at, -ar, -ay,  be,  ax,  aq,  au,  bb, -bb, -au, -aq, -ax, -be,  ay,  ar,  at,  ba, -bc, -av, -ap, -aw, -bd,  az,  as }, \
  { bm,  cb, -cf, -bq, -bi, -bx,  cj,  bu,  bf,  bt,  ci, -by, -bj, -bp, -ce,  cc,  bn,  bl,  ca, -cg, -br, -bh, -bw,  ck,  bv,  bg,  bs,  ch, -bz, -bk, -bo, -cd,  cd,  bo,  bk,  bz, -ch, -bs, -bg, -bv, -ck,  bw,  bh,  br,  cg, -ca, -bl, -bn, -cc,  ce,  bp,  bj,  by, -ci, -bt, -bf, -bu, -cj,  bx,  bi,  bq,  cf, -cb, -bm }, \
  { ab,  ac, -ac, -ab, -ab, -ac,  ac,  ab,  ab,  ac, -ac, -ab, -ab, -ac,  ac,  ab,  ab,  ac, -ac, -ab, -ab, -ac,  ac,  ab,  ab,  ac, -ac, -ab, -ab, -ac,  ac,  ab,  ab,  ac, -ac, -ab, -ab, -ac,  ac,  ab,  ab,  ac, -ac, -ab, -ab, -ac,  ac,  ab,  ab,  ac, -ac, -ab, -ab, -ac,  ac,  ab,  ab,  ac, -ac, -ab, -ab, -ac,  ac,  ab }, \
  { bn,  ce, -ca, -bj, -br, -ci,  bw,  bf,  bv, -cj, -bs, -bi, -bz,  cf,  bo,  bm,  cd, -cb, -bk, -bq, -ch,  bx,  bg,  bu, -ck, -bt, -bh, -by,  cg,  bp,  bl,  cc, -cc, -bl, -bp, -cg,  by,  bh,  bt,  ck, -bu, -bg, -bx,  ch,  bq,  bk,  cb, -cd, -bm, -bo, -cf,  bz,  bi,  bs,  cj, -bv, -bf, -bw,  ci,  br,  bj,  ca, -ce, -bn }, \
  { at,  bc, -ay, -ap, -ax,  bd,  au,  as,  bb, -az, -aq, -aw,  be,  av,  ar,  ba, -ba, -ar, -av, -be,  aw,  aq,  az, -bb, -as, -au, -bd,  ax,  ap,  ay, -bc, -at, -at, -bc,  ay,  ap,  ax, -bd, -au, -as, -bb,  az,  aq,  aw, -be, -av, -ar, -ba,  ba,  ar,  av,  be, -aw, -aq, -az,  bb,  as,  au,  bd, -ax, -ap, -ay,  bc,  at }, \
  { bo,  ch, -bv, -bh, -ca,  cc,  bj,  bt, -cj, -bq, -bm, -cf,  bx,  bf,  by, -ce, -bl, -br, -ck,  bs,  bk,  cd, -bz, -bg, -bw,  cg,  bn,  bp,  ci, -bu, -bi, -cb,  cb,  bi,  bu, -ci, -bp, -bn, -cg,  bw,  bg,  bz, -cd, -bk, -bs,  ck,  br,  bl,  ce, -by, -bf, -bx,  cf,  bm,  bq,  cj, -bt, -bj, -cc,  ca,  bh,  bv, -ch, -bo }, \
  { aj,  ao, -ak, -ai, -an,  al,  ah,  am, -am, -ah, -al,  an,  ai,  ak, -ao, -aj, -aj, -ao,  ak,  ai,  an, -al, -ah, -am,  am,  ah,  al, -an, -ai, -ak,  ao,  aj,  aj,  ao, -ak, -ai, -an,  al,  ah,  am, -am, -ah, -al,  an,  ai,  ak, -ao, -aj, -aj, -ao,  ak,  ai,  an, -al, -ah, -am,  am,  ah,  al, -an, -ai, -ak,  ao,  aj }, \
  { bp,  ck, -bq, -bo, -cj,  br,  bn,  ci, -bs, -bm, -ch,  bt,  bl,  cg, -bu, -bk, -cf,  bv,  bj,  ce, -bw, -bi, -cd,  bx,  bh,  cc, -by, -bg, -cb,  bz,  bf,  ca, -ca, -bf, -bz,  cb,  bg,  by, -cc, -bh, -bx,  cd,  bi,  bw, -ce, -bj, -bv,  cf,  bk,  bu, -cg, -bl, -bt,  ch,  bm,  bs, -ci, -bn, -br,  cj,  bo,  bq, -ck, -bp }, \
  { au, -be, -at, -av,  bd,  as,  aw, -bc, -ar, -ax,  bb,  aq,  ay, -ba, -ap, -az,  az,  ap,  ba, -ay, -aq, -bb,  ax,  ar,  bc, -aw, -as, -bd,  av,  at,  be, -au, -au,  be,  at,  av, -bd, -as, -aw,  bc,  ar,  ax, -bb, -aq, -ay,  ba,  ap,  az, -az, -ap, -ba,  ay,  aq,  bb, -ax, -ar, -bc,  aw,  as,  bd, -av, -at, -be,  au }, \
  { bq, -ci, -bl, -bv,  cd,  bg,  ca, -by, -bi, -cf,  bt,  bn,  ck, -bo, -bs,  cg,  bj,  bx, -cb, -bf, -cc,  bw,  bk,  ch, -br, -bp,  cj,  bm,  bu, -ce, -bh, -bz,  bz,  bh,  ce, -bu, -bm, -cj,  bp,  br, -ch, -bk, -bw,  cc,  bf,  cb, -bx, -bj, -cg,  bs,  bo, -ck, -bn, -bt,  cf,  bi,  by, -ca, -bg, -cd,  bv,  bl,  ci, -bq }, \
  { ae, -ag, -ad, -af,  af,  ad,  ag, -ae, -ae,  ag,  ad,  af, -af, -ad, -ag,  ae,  ae, -ag, -ad, -af,  af,  ad,  ag, -ae, -ae,  ag,  ad,  af, -af, -ad, -ag,  ae,  ae, -ag, -ad, -af,  af,  ad,  ag, -ae, -ae,  ag,  ad,  af, -af, -ad, -ag,  ae,  ae, -ag, -ad, -af,  af,  ad,  ag, -ae, -ae,  ag,  ad,  af, -af, -ad, -ag,  ae }, \
  { br, -cf, -bg, -cc,  bu,  bo, -ci, -bj, -bz,  bx,  bl,  ck, -bm, -bw,  ca,  bi,  ch, -bp, -bt,  cd,  bf,  ce, -bs, -bq,  cg,  bh,  cb, -bv, -bn,  cj,  bk,  by, -by, -bk, -cj,  bn,  bv, -cb, -bh, -cg,  bq,  bs, -ce, -bf, -cd,  bt,  bp, -ch, -bi, -ca,  bw,  bm, -ck, -bl, -bx,  bz,  bj,  ci, -bo, -bu,  cc,  bg,  cf, -br }, \
  { av, -bb, -ap, -bc,  au,  aw, -ba, -aq, -bd,  at,  ax, -az, -ar, -be,  as,  ay, -ay, -as,  be,  ar,  az, -ax, -at,  bd,  aq,  ba, -aw, -au,  bc,  ap,  bb, -av, -av,  bb,  ap,  bc, -au, -aw,  ba,  aq,  bd, -at, -ax,  az,  ar,  be, -as, -ay,  ay,  as, -be, -ar, -az,  ax,  at, -bd, -aq, -ba,  aw,  au, -bc, -ap, -bb,  av }, \
  { bs, -cc, -bi, -cj,  bl,  bz, -bv, -bp,  cf,  bf,  cg, -bo, -bw,  by,  bm, -ci, -bh, -cd,  br,  bt, -cb, -bj, -ck,  bk,  ca, -bu, -bq,  ce,  bg,  ch, -bn, -bx,  bx,  bn, -ch, -bg, -ce,  bq,  bu, -ca, -bk,  ck,  bj,  cb, -bt, -br,  cd,  bh,  ci, -bm, -by,  bw,  bo, -cg, -bf, -cf,  bp,  bv, -bz, -bl,  cj,  bi,  cc, -bs }, \
  { ak, -am, -ai,  ao,  ah,  an, -aj, -al,  al,  aj, -an, -ah, -ao,  ai,  am, -ak, -ak,  am,  ai, -ao, -ah, -an,  aj,  al, -al, -aj,  an,  ah,  ao, -ai, -am,  ak,  ak, -am, -ai,  ao,  ah,  an, -aj, -al,  al,  aj, -an, -ah, -ao,  ai,  am, -ak, -ak,  am,  ai, -ao, -ah, -an,  aj,  al, -al, -aj,  an,  ah,  ao, -ai, -am,  ak }, \
  { bt, -bz, -bn,  cf,  bh,  ck, -bi, -ce,  bo,  by, -bu, -bs,  ca,  bm, -cg, -bg, -cj,  bj,  cd, -bp, -bx,  bv,  br, -cb, -bl,  ch,  bf,  ci, -bk, -cc,  bq,  bw, -bw, -bq,  cc,  bk, -ci, -bf, -ch,  bl,  cb, -br, -bv,  bx,  bp, -cd, -bj,  cj,  bg,  cg, -bm, -ca,  bs,  bu, -by, -bo,  ce,  bi, -ck, -bh, -cf,  bn,  bz, -bt }, \
  { aw, -ay, -au,  ba,  as, -bc, -aq,  be,  ap,  bd, -ar, -bb,  at,  az, -av, -ax,  ax,  av, -az, -at,  bb,  ar, -bd, -ap, -be,  aq,  bc, -as, -ba,  au,  ay, -aw, -aw,  ay,  au, -ba, -as,  bc,  aq, -be, -ap, -bd,  ar,  bb, -at, -az,  av,  ax, -ax, -av,  az,  at, -bb, -ar,  bd,  ap,  be, -aq, -bc,  as,  ba, -au, -ay,  aw }, \
  { bu, -bw, -bs,  by,  bq, -ca, -bo,  cc,  bm, -ce, -bk,  cg,  bi, -ci, -bg,  ck,  bf,  cj, -bh, -ch,  bj,  cf, -bl, -cd,  bn,  cb, -bp, -bz,  br,  bx, -bt, -bv,  bv,  bt, -bx, -br,  bz,  bp, -cb, -bn,  cd,  bl, -cf, -bj,  ch,  bh, -cj, -bf, -ck,  bg,  ci, -bi, -cg,  bk,  ce, -bm, -cc,  bo,  ca, -bq, -by,  bs,  bw, -bu }, \
  { aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa,  aa, -aa, -aa,  aa }, \
  { bv, -bt, -bx,  br,  bz, -bp, -cb,  bn,  cd, -bl, -cf,  bj,  ch, -bh, -cj,  bf, -ck, -bg,  ci,  bi, -cg, -bk,  ce,  bm, -cc, -bo,  ca,  bq, -by, -bs,  bw,  bu, -bu, -bw,  bs,  by, -bq, -ca,  bo,  cc, -bm, -ce,  bk,  cg, -bi, -ci,  bg,  ck, -bf,  cj,  bh, -ch, -bj,  cf,  bl, -cd, -bn,  cb,  bp, -bz, -br,  bx,  bt, -bv }, \
  { ax, -av, -az,  at,  bb, -ar, -bd,  ap, -be, -aq,  bc,  as, -ba, -au,  ay,  aw, -aw, -ay,  au,  ba, -as, -bc,  aq,  be, -ap,  bd,  ar, -bb, -at,  az,  av, -ax, -ax,  av,  az, -at, -bb,  ar,  bd, -ap,  be,  aq, -bc, -as,  ba,  au, -ay, -aw,  aw,  ay, -au, -ba,  as,  bc, -aq, -be,  ap, -bd, -ar,  bb,  at, -az, -av,  ax }, \
  { bw, -bq, -cc,  bk,  ci, -bf,  ch,  bl, -cb, -br,  bv,  bx, -bp, -cd,  bj,  cj, -bg,  cg,  bm, -ca, -bs,  bu,  by, -bo, -ce,  bi,  ck, -bh,  cf,  bn, -bz, -bt,  bt,  bz, -bn, -cf,  bh, -ck, -bi,  ce,  bo, -by, -bu,  bs,  ca, -bm, -cg,  bg, -cj, -bj,  cd,  bp, -bx, -bv,  br,  cb, -bl, -ch,  bf, -ci, -bk,  cc,  bq, -bw }, \
  { al, -aj, -an,  ah, -ao, -ai,  am,  ak, -ak, -am,  ai,  ao, -ah,  an,  aj, -al, -al,  aj,  an, -ah,  ao,  ai, -am, -ak,  ak,  am, -ai, -ao,  ah, -an, -aj,  al,  al, -aj, -an,  ah, -ao, -ai,  am,  ak, -ak, -am,  ai,  ao, -ah,  an,  aj, -al, -al,  aj,  an, -ah,  ao,  ai, -am, -ak,  ak,  am, -ai, -ao,  ah, -an, -aj,  al }, \
  { bx, -bn, -ch,  bg, -ce, -bq,  bu,  ca, -bk, -ck,  bj, -cb, -bt,  br,  cd, -bh,  ci,  bm, -by, -bw,  bo,  cg, -bf,  cf,  bp, -bv, -bz,  bl,  cj, -bi,  cc,  bs, -bs, -cc,  bi, -cj, -bl,  bz,  bv, -bp, -cf,  bf, -cg, -bo,  bw,  by, -bm, -ci,  bh, -cd, -br,  bt,  cb, -bj,  ck,  bk, -ca, -bu,  bq,  ce, -bg,  ch,  bn, -bx }, \
  { ay, -as, -be,  ar, -az, -ax,  at,  bd, -aq,  ba,  aw, -au, -bc,  ap, -bb, -av,  av,  bb, -ap,  bc,  au, -aw, -ba,  aq, -bd, -at,  ax,  az, -ar,  be,  as, -ay, -ay,  as,  be, -ar,  az,  ax, -at, -bd,  aq, -ba, -aw,  au,  bc, -ap,  bb,  av, -av, -bb,  ap, -bc, -au,  aw,  ba, -aq,  bd,  at, -ax, -az,  ar, -be, -as,  ay }, \
  { by, -bk,  cj,  bn, -bv, -cb,  bh, -cg, -bq,  bs,  ce, -bf,  cd,  bt, -bp, -ch,  bi, -ca, -bw,  bm,  ck, -bl,  bx,  bz, -bj,  ci,  bo, -bu, -cc,  bg, -cf, -br,  br,  cf, -bg,  cc,  bu, -bo, -ci,  bj, -bz, -bx,  bl, -ck, -bm,  bw,  ca, -bi,  ch,  bp, -bt, -cd,  bf, -ce, -bs,  bq,  cg, -bh,  cb,  bv, -bn, -cj,  bk, -by }, \
  { af, -ad,  ag,  ae, -ae, -ag,  ad, -af, -af,  ad, -ag, -ae,  ae,  ag, -ad,  af,  af, -ad,  ag,  ae, -ae, -ag,  ad, -af, -af,  ad, -ag, -ae,  ae,  ag, -ad,  af,  af, -ad,  ag,  ae, -ae, -ag,  ad, -af, -af,  ad, -ag, -ae,  ae,  ag, -ad,  af,  af, -ad,  ag,  ae, -ae, -ag,  ad, -af, -af,  ad, -ag, -ae,  ae,  ag, -ad,  af }, \
  { bz, -bh,  ce,  bu, -bm,  cj,  bp, -br, -ch,  bk, -bw, -cc,  bf, -cb, -bx,  bj, -cg, -bs,  bo,  ck, -bn,  bt,  cf, -bi,  by,  ca, -bg,  cd,  bv, -bl,  ci,  bq, -bq, -ci,  bl, -bv, -cd,  bg, -ca, -by,  bi, -cf, -bt,  bn, -ck, -bo,  bs,  cg, -bj,  bx,  cb, -bf,  cc,  bw, -bk,  ch,  br, -bp, -cj,  bm, -bu, -ce,  bh, -bz }, \
  { az, -ap,  ba,  ay, -aq,  bb,  ax, -ar,  bc,  aw, -as,  bd,  av, -at,  be,  au, -au, -be,  at, -av, -bd,  as, -aw, -bc,  ar, -ax, -bb,  aq, -ay, -ba,  ap, -az, -az,  ap, -ba, -ay,  aq, -bb, -ax,  ar, -bc, -aw,  as, -bd, -av,  at, -be, -au,  au,  be, -at,  av,  bd, -as,  aw,  bc, -ar,  ax,  bb, -aq,  ay,  ba, -ap,  az }, \
  { ca, -bf,  bz,  cb, -bg,  by,  cc, -bh,  bx,  cd, -bi,  bw,  ce, -bj,  bv,  cf, -bk,  bu,  cg, -bl,  bt,  ch, -bm,  bs,  ci, -bn,  br,  cj, -bo,  bq,  ck, -bp,  bp, -ck, -bq,  bo, -cj, -br,  bn, -ci, -bs,  bm, -ch, -bt,  bl, -cg, -bu,  bk, -cf, -bv,  bj, -ce, -bw,  bi, -cd, -bx,  bh, -cc, -by,  bg, -cb, -bz,  bf, -ca }, \
  { am, -ah,  al,  an, -ai,  ak,  ao, -aj,  aj, -ao, -ak,  ai, -an, -al,  ah, -am, -am,  ah, -al, -an,  ai, -ak, -ao,  aj, -aj,  ao,  ak, -ai,  an,  al, -ah,  am,  am, -ah,  al,  an, -ai,  ak,  ao, -aj,  aj, -ao, -ak,  ai, -an, -al,  ah, -am, -am,  ah, -al, -an,  ai, -ak, -ao,  aj, -aj,  ao,  ak, -ai,  an,  al, -ah,  am }, \
  { cb, -bi,  bu,  ci, -bp,  bn, -cg, -bw,  bg, -bz, -cd,  bk, -bs, -ck,  br, -bl,  ce,  by, -bf,  bx,  cf, -bm,  bq, -cj, -bt,  bj, -cc, -ca,  bh, -bv, -ch,  bo, -bo,  ch,  bv, -bh,  ca,  cc, -bj,  bt,  cj, -bq,  bm, -cf, -bx,  bf, -by, -ce,  bl, -br,  ck,  bs, -bk,  cd,  bz, -bg,  bw,  cg, -bn,  bp, -ci, -bu,  bi, -cb }, \
  { ba, -ar,  av, -be, -aw,  aq, -az, -bb,  as, -au,  bd,  ax, -ap,  ay,  bc, -at,  at, -bc, -ay,  ap, -ax, -bd,  au, -as,  bb,  az, -aq,  aw,  be, -av,  ar, -ba, -ba,  ar, -av,  be,  aw, -aq,  az,  bb, -as,  au, -bd, -ax,  ap, -ay, -bc,  at, -at,  bc,  ay, -ap,  ax,  bd, -au,  as, -bb, -az,  aq, -aw, -be,  av, -ar,  ba }, \
  { cc, -bl,  bp, -cg, -by,  bh, -bt,  ck,  bu, -bg,  bx,  ch, -bq,  bk, -cb, -cd,  bm, -bo,  cf,  bz, -bi,  bs, -cj, -bv,  bf, -bw, -ci,  br, -bj,  ca,  ce, -bn,  bn, -ce, -ca,  bj, -br,  ci,  bw, -bf,  bv,  cj, -bs,  bi, -bz, -cf,  bo, -bm,  cd,  cb, -bk,  bq, -ch, -bx,  bg, -bu, -ck,  bt, -bh,  by,  cg, -bp,  bl, -cc }, \
  { ac, -ab,  ab, -ac, -ac,  ab, -ab,  ac,  ac, -ab,  ab, -ac, -ac,  ab, -ab,  ac,  ac, -ab,  ab, -ac, -ac,  ab, -ab,  ac,  ac, -ab,  ab, -ac, -ac,  ab, -ab,  ac,  ac, -ab,  ab, -ac, -ac,  ab, -ab,  ac,  ac, -ab,  ab, -ac, -ac,  ab, -ab,  ac,  ac, -ab,  ab, -ac, -ac,  ab, -ab,  ac,  ac, -ab,  ab, -ac, -ac,  ab, -ab,  ac }, \
  { cd, -bo,  bk, -bz, -ch,  bs, -bg,  bv, -ck, -bw,  bh, -br,  cg,  ca, -bl,  bn, -cc, -ce,  bp, -bj,  by,  ci, -bt,  bf, -bu,  cj,  bx, -bi,  bq, -cf, -cb,  bm, -bm,  cb,  cf, -bq,  bi, -bx, -cj,  bu, -bf,  bt, -ci, -by,  bj, -bp,  ce,  cc, -bn,  bl, -ca, -cg,  br, -bh,  bw,  ck, -bv,  bg, -bs,  ch,  bz, -bk,  bo, -cd }, \
  { bb, -au,  aq, -ax,  be,  ay, -ar,  at, -ba, -bc,  av, -ap,  aw, -bd, -az,  as, -as,  az,  bd, -aw,  ap, -av,  bc,  ba, -at,  ar, -ay, -be,  ax, -aq,  au, -bb, -bb,  au, -aq,  ax, -be, -ay,  ar, -at,  ba,  bc, -av,  ap, -aw,  bd,  az, -as,  as, -az, -bd,  aw, -ap,  av, -bc, -ba,  at, -ar,  ay,  be, -ax,  aq, -au,  bb }, \
  { ce, -br,  bf, -bs,  cf,  cd, -bq,  bg, -bt,  cg,  cc, -bp,  bh, -bu,  ch,  cb, -bo,  bi, -bv,  ci,  ca, -bn,  bj, -bw,  cj,  bz, -bm,  bk, -bx,  ck,  by, -bl,  bl, -by, -ck,  bx, -bk,  bm, -bz, -cj,  bw, -bj,  bn, -ca, -ci,  bv, -bi,  bo, -cb, -ch,  bu, -bh,  bp, -cc, -cg,  bt, -bg,  bq, -cd, -cf,  bs, -bf,  br, -ce }, \
  { an, -ak,  ah, -aj,  am,  ao, -al,  ai, -ai,  al, -ao, -am,  aj, -ah,  ak, -an, -an,  ak, -ah,  aj, -am, -ao,  al, -ai,  ai, -al,  ao,  am, -aj,  ah, -ak,  an,  an, -ak,  ah, -aj,  am,  ao, -al,  ai, -ai,  al, -ao, -am,  aj, -ah,  ak, -an, -an,  ak, -ah,  aj, -am, -ao,  al, -ai,  ai, -al,  ao,  am, -aj,  ah, -ak,  an }, \
  { cf, -bu,  bj, -bl,  bw, -ch, -cd,  bs, -bh,  bn, -by,  cj,  cb, -bq,  bf, -bp,  ca,  ck, -bz,  bo, -bg,  br, -cc, -ci,  bx, -bm,  bi, -bt,  ce,  cg, -bv,  bk, -bk,  bv, -cg, -ce,  bt, -bi,  bm, -bx,  ci,  cc, -br,  bg, -bo,  bz, -ck, -ca,  bp, -bf,  bq, -cb, -cj,  by, -bn,  bh, -bs,  cd,  ch, -bw,  bl, -bj,  bu, -cf }, \
  { bc, -ax,  as, -aq,  av, -ba, -be,  az, -au,  ap, -at,  ay, -bd, -bb,  aw, -ar,  ar, -aw,  bb,  bd, -ay,  at, -ap,  au, -az,  be,  ba, -av,  aq, -as,  ax, -bc, -bc,  ax, -as,  aq, -av,  ba,  be, -az,  au, -ap,  at, -ay,  bd,  bb, -aw,  ar, -ar,  aw, -bb, -bd,  ay, -at,  ap, -au,  az, -be, -ba,  av, -aq,  as, -ax,  bc }, \
  { cg, -bx,  bo, -bf,  bn, -bw,  cf,  ch, -by,  bp, -bg,  bm, -bv,  ce,  ci, -bz,  bq, -bh,  bl, -bu,  cd,  cj, -ca,  br, -bi,  bk, -bt,  cc,  ck, -cb,  bs, -bj,  bj, -bs,  cb, -ck, -cc,  bt, -bk,  bi, -br,  ca, -cj, -cd,  bu, -bl,  bh, -bq,  bz, -ci, -ce,  bv, -bm,  bg, -bp,  by, -ch, -cf,  bw, -bn,  bf, -bo,  bx, -cg }, \
  { ag, -af,  ae, -ad,  ad, -ae,  af, -ag, -ag,  af, -ae,  ad, -ad,  ae, -af,  ag,  ag, -af,  ae, -ad,  ad, -ae,  af, -ag, -ag,  af, -ae,  ad, -ad,  ae, -af,  ag,  ag, -af,  ae, -ad,  ad, -ae,  af, -ag, -ag,  af, -ae,  ad, -ad,  ae, -af,  ag,  ag, -af,  ae, -ad,  ad, -ae,  af, -ag, -ag,  af, -ae,  ad, -ad,  ae, -af,  ag }, \
  { ch, -ca,  bt, -bm,  bf, -bl,  bs, -bz,  cg,  ci, -cb,  bu, -bn,  bg, -bk,  br, -by,  cf,  cj, -cc,  bv, -bo,  bh, -bj,  bq, -bx,  ce,  ck, -cd,  bw, -bp,  bi, -bi,  bp, -bw,  cd, -ck, -ce,  bx, -bq,  bj, -bh,  bo, -bv,  cc, -cj, -cf,  by, -br,  bk, -bg,  bn, -bu,  cb, -ci, -cg,  bz, -bs,  bl, -bf,  bm, -bt,  ca, -ch }, \
  { bd, -ba,  ax, -au,  ar, -ap,  as, -av,  ay, -bb,  be,  bc, -az,  aw, -at,  aq, -aq,  at, -aw,  az, -bc, -be,  bb, -ay,  av, -as,  ap, -ar,  au, -ax,  ba, -bd, -bd,  ba, -ax,  au, -ar,  ap, -as,  av, -ay,  bb, -be, -bc,  az, -aw,  at, -aq,  aq, -at,  aw, -az,  bc,  be, -bb,  ay, -av,  as, -ap,  ar, -au,  ax, -ba,  bd }, \
  { ci, -cd,  by, -bt,  bo, -bj,  bf, -bk,  bp, -bu,  bz, -ce,  cj,  ch, -cc,  bx, -bs,  bn, -bi,  bg, -bl,  bq, -bv,  ca, -cf,  ck,  cg, -cb,  bw, -br,  bm, -bh,  bh, -bm,  br, -bw,  cb, -cg, -ck,  cf, -ca,  bv, -bq,  bl, -bg,  bi, -bn,  bs, -bx,  cc, -ch, -cj,  ce, -bz,  bu, -bp,  bk, -bf,  bj, -bo,  bt, -by,  cd, -ci }, \
  { ao, -an,  am, -al,  ak, -aj,  ai, -ah,  ah, -ai,  aj, -ak,  al, -am,  an, -ao, -ao,  an, -am,  al, -ak,  aj, -ai,  ah, -ah,  ai, -aj,  ak, -al,  am, -an,  ao,  ao, -an,  am, -al,  ak, -aj,  ai, -ah,  ah, -ai,  aj, -ak,  al, -am,  an, -ao, -ao,  an, -am,  al, -ak,  aj, -ai,  ah, -ah,  ai, -aj,  ak, -al,  am, -an,  ao }, \
  { cj, -cg,  cd, -ca,  bx, -bu,  br, -bo,  bl, -bi,  bf, -bh,  bk, -bn,  bq, -bt,  bw, -bz,  cc, -cf,  ci,  ck, -ch,  ce, -cb,  by, -bv,  bs, -bp,  bm, -bj,  bg, -bg,  bj, -bm,  bp, -bs,  bv, -by,  cb, -ce,  ch, -ck, -ci,  cf, -cc,  bz, -bw,  bt, -bq,  bn, -bk,  bh, -bf,  bi, -bl,  bo, -br,  bu, -bx,  ca, -cd,  cg, -cj }, \
  { be, -bd,  bc, -bb,  ba, -az,  ay, -ax,  aw, -av,  au, -at,  as, -ar,  aq, -ap,  ap, -aq,  ar, -as,  at, -au,  av, -aw,  ax, -ay,  az, -ba,  bb, -bc,  bd, -be, -be,  bd, -bc,  bb, -ba,  az, -ay,  ax, -aw,  av, -au,  at, -as,  ar, -aq,  ap, -ap,  aq, -ar,  as, -at,  au, -av,  aw, -ax,  ay, -az,  ba, -bb,  bc, -bd,  be }, \
  { ck, -cj,  ci, -ch,  cg, -cf,  ce, -cd,  cc, -cb,  ca, -bz,  by, -bx,  bw, -bv,  bu, -bt,  bs, -br,  bq, -bp,  bo, -bn,  bm, -bl,  bk, -bj,  bi, -bh,  bg, -bf,  bf, -bg,  bh, -bi,  bj, -bk,  bl, -bm,  bn, -bo,  bp, -bq,  br, -bs,  bt, -bu,  bv, -bw,  bx, -by,  bz, -ca,  cb, -cc,  cd, -ce,  cf, -cg,  ch, -ci,  cj, -ck }, \
 }

// DCT-8
#define DEFINE_DCT8_P4_MATRIX(a,b,c,d) \
{ \
  {  a,  b,  c,  d,}, \
  {  b,  0, -b, -b,}, \
  {  c, -b, -d,  a,}, \
  {  d, -b,  a, -c,}, \
}

#define DEFINE_DCT8_P8_MATRIX(a,b,c,d,e,f,g,h) \
{ \
  {  a,  b,  c,  d,  e,  f,  g,  h,}, \
  {  b,  e,  h, -g, -d, -a, -c, -f,}, \
  {  c,  h, -e, -a, -f,  g,  b,  d,}, \
  {  d, -g, -a, -h,  c,  e, -f, -b,}, \
  {  e, -d, -f,  c,  g, -b, -h,  a,}, \
  {  f, -a,  g,  e, -b,  h,  d, -c,}, \
  {  g, -c,  b, -f, -h,  d, -a,  e,}, \
  {  h, -f,  d, -b,  a, -c,  e, -g,}, \
}

#define DEFINE_DCT8_P16_MATRIX(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p) \
{ \
  {  a,  b,  c,  d,  e,  f,  g,  h,  i,  j,  k,  l,  m,  n,  o,  p,}, \
  {  b,  e,  h,  k,  n,  0, -n, -k, -h, -e, -b, -b, -e, -h, -k, -n,}, \
  {  c,  h,  m, -p, -k, -f, -a, -e, -j, -o,  n,  i,  d,  b,  g,  l,}, \
  {  d,  k, -p, -i, -b, -f, -m,  n,  g,  a,  h,  o, -l, -e, -c, -j,}, \
  {  e,  n, -k, -b, -h,  0,  h,  b,  k, -n, -e, -e, -n,  k,  b,  h,}, \
  {  f,  0, -f, -f,  0,  f,  f,  0, -f, -f,  0,  f,  f,  0, -f, -f,}, \
  {  g, -n, -a, -m,  h,  f, -o, -b, -l,  i,  e, -p, -c, -k,  j,  d,}, \
  {  h, -k, -e,  n,  b,  0, -b, -n,  e,  k, -h, -h,  k,  e, -n, -b,}, \
  {  i, -h, -j,  g,  k, -f, -l,  e,  m, -d, -n,  c,  o, -b, -p,  a,}, \
  {  j, -e, -o,  a, -n, -f,  i,  k, -d, -p,  b, -m, -g,  h,  l, -c,}, \
  {  k, -b,  n,  h, -e,  0,  e, -h, -n,  b, -k, -k,  b, -n, -h,  e,}, \
  {  l, -b,  i,  o, -e,  f, -p, -h,  c, -m, -k,  a, -j, -n,  d, -g,}, \
  {  m, -e,  d, -l, -n,  f, -c,  k,  o, -g,  b, -j, -p,  h, -a,  i,}, \
  {  n, -h,  b, -e,  k,  0, -k,  e, -b,  h, -n, -n,  h, -b,  e, -k,}, \
  {  o, -k,  g, -c,  b, -f,  j, -n, -p,  l, -h,  d, -a,  e, -i,  m,}, \
  {  p, -n,  l, -j,  h, -f,  d, -b,  a, -c,  e, -g,  i, -k,  m, -o,}, \
}

#define DEFINE_DCT8_P32_MATRIX(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z,A,B,C,D,E,F) \
{ \
  {  a,  b,  c,  d,  e,  f,  g,  h,  i,  j,  k,  l,  m,  n,  o,  p,  q,  r,  s,  t,  u,  v,  w,  x,  y,  z,  A,  B,  C,  D,  E,  F,}, \
  {  b,  e,  h,  k,  n,  q,  t,  w,  z,  C,  F, -E, -B, -y, -v, -s, -p, -m, -j, -g, -d, -a, -c, -f, -i, -l, -o, -r, -u, -x, -A, -D,}, \
  {  c,  h,  m,  r,  w,  B,  0, -B, -w, -r, -m, -h, -c, -c, -h, -m, -r, -w, -B,  0,  B,  w,  r,  m,  h,  c,  c,  h,  m,  r,  w,  B,}, \
  {  d,  k,  r,  y,  F, -A, -t, -m, -f, -b, -i, -p, -w, -D,  C,  v,  o,  h,  a,  g,  n,  u,  B, -E, -x, -q, -j, -c, -e, -l, -s, -z,}, \
  {  e,  n,  w,  F, -y, -p, -g, -c, -l, -u, -D,  A,  r,  i,  a,  j,  s,  B, -C, -t, -k, -b, -h, -q, -z,  E,  v,  m,  d,  f,  o,  x,}, \
  {  f,  q,  B, -A, -p, -e, -g, -r, -C,  z,  o,  d,  h,  s,  D, -y, -n, -c, -i, -t, -E,  x,  m,  b,  j,  u,  F, -w, -l, -a, -k, -v,}, \
  {  g,  t,  0, -t, -g, -g, -t,  0,  t,  g,  g,  t,  0, -t, -g, -g, -t,  0,  t,  g,  g,  t,  0, -t, -g, -g, -t,  0,  t,  g,  g,  t,}, \
  {  h,  w, -B, -m, -c, -r,  0,  r,  c,  m,  B, -w, -h, -h, -w,  B,  m,  c,  r,  0, -r, -c, -m, -B,  w,  h,  h,  w, -B, -m, -c, -r,}, \
  {  i,  z, -w, -f, -l, -C,  t,  c,  o,  F, -q, -a, -r,  E,  n,  d,  u, -B, -k, -g, -x,  y,  h,  j,  A, -v, -e, -m, -D,  s,  b,  p,}, \
  {  j,  C, -r, -b, -u,  z,  g,  m,  F, -o, -e, -x,  w,  d,  p, -E, -l, -h, -A,  t,  a,  s, -B, -i, -k, -D,  q,  c,  v, -y, -f, -n,}, \
  {  k,  F, -m, -i, -D,  o,  g,  B, -q, -e, -z,  s,  c,  x, -u, -a, -v,  w,  b,  t, -y, -d, -r,  A,  f,  p, -C, -h, -n,  E,  j,  l,}, \
  {  l, -E, -h, -p,  A,  d,  t, -w, -a, -x,  s,  e,  B, -o, -i, -F,  k,  m, -D, -g, -q,  z,  c,  u, -v, -b, -y,  r,  f,  C, -n, -j,}, \
  {  m, -B, -c, -w,  r,  h,  0, -h, -r,  w,  c,  B, -m, -m,  B,  c,  w, -r, -h,  0,  h,  r, -w, -c, -B,  m,  m, -B, -c, -w,  r,  h,}, \
  {  n, -y, -c, -D,  i,  s, -t, -h,  E,  d,  x, -o, -m,  z,  b,  C, -j, -r,  u,  g, -F, -e, -w,  p,  l, -A, -a, -B,  k,  q, -v, -f,}, \
  {  o, -v, -h,  C,  a,  D, -g, -w,  n,  p, -u, -i,  B,  b,  E, -f, -x,  m,  q, -t, -j,  A,  c,  F, -e, -y,  l,  r, -s, -k,  z,  d,}, \
  {  p, -s, -m,  v,  j, -y, -g,  B,  d, -E, -a, -F,  c,  C, -f, -z,  i,  w, -l, -t,  o,  q, -r, -n,  u,  k, -x, -h,  A,  e, -D, -b,}, \
  {  q, -p, -r,  o,  s, -n, -t,  m,  u, -l, -v,  k,  w, -j, -x,  i,  y, -h, -z,  g,  A, -f, -B,  e,  C, -d, -D,  c,  E, -b, -F,  a,}, \
  {  r, -m, -w,  h,  B, -c,  0,  c, -B, -h,  w,  m, -r, -r,  m,  w, -h, -B,  c,  0, -c,  B,  h, -w, -m,  r,  r, -m, -w,  h,  B, -c,}, \
  {  s, -j, -B,  a, -C, -i,  t,  r, -k, -A,  b, -D, -h,  u,  q, -l, -z,  c, -E, -g,  v,  p, -m, -y,  d, -F, -f,  w,  o, -n, -x,  e,}, \
  {  t, -g,  0,  g, -t, -t,  g,  0, -g,  t,  t, -g,  0,  g, -t, -t,  g,  0, -g,  t,  t, -g,  0,  g, -t, -t,  g,  0, -g,  t,  t, -g,}, \
  {  u, -d,  B,  n, -k, -E,  g, -r, -x,  a, -y, -q,  h, -F, -j,  o,  A, -c,  v,  t, -e,  C,  m, -l, -D,  f, -s, -w,  b, -z, -p,  i,}, \
  {  v, -a,  w,  u, -b,  x,  t, -c,  y,  s, -d,  z,  r, -e,  A,  q, -f,  B,  p, -g,  C,  o, -h,  D,  n, -i,  E,  m, -j,  F,  l, -k,}, \
  {  w, -c,  r,  B, -h,  m,  0, -m,  h, -B, -r,  c, -w, -w,  c, -r, -B,  h, -m,  0,  m, -h,  B,  r, -c,  w,  w, -c,  r,  B, -h,  m,}, \
  {  x, -f,  m, -E, -q,  b, -t, -B,  j, -i,  A,  u, -c,  p,  F, -n,  e, -w, -y,  g, -l,  D,  r, -a,  s,  C, -k,  h, -z, -v,  d, -o,}, \
  {  y, -i,  h, -x, -z,  j, -g,  w,  A, -k,  f, -v, -B,  l, -e,  u,  C, -m,  d, -t, -D,  n, -c,  s,  E, -o,  b, -r, -F,  p, -a,  q,}, \
  {  z, -l,  c, -q,  E,  u, -g,  h, -v, -D,  p, -b,  m, -A, -y,  k, -d,  r, -F, -t,  f, -i,  w,  C, -o,  a, -n,  B,  x, -j,  e, -s,}, \
  {  A, -o,  c, -j,  v,  F, -t,  h, -e,  q, -C, -y,  m, -a,  l, -x, -D,  r, -f,  g, -s,  E,  w, -k,  b, -n,  z,  B, -p,  d, -i,  u,}, \
  {  B, -r,  h, -c,  m, -w,  0,  w, -m,  c, -h,  r, -B, -B,  r, -h,  c, -m,  w,  0, -w,  m, -c,  h, -r,  B,  B, -r,  h, -c,  m, -w,}, \
  {  C, -u,  m, -e,  d, -l,  t, -B, -D,  v, -n,  f, -c,  k, -s,  A,  E, -w,  o, -g,  b, -j,  r, -z, -F,  x, -p,  h, -a,  i, -q,  y,}, \
  {  D, -x,  r, -l,  f, -a,  g, -m,  s, -y,  E,  C, -w,  q, -k,  e, -b,  h, -n,  t, -z,  F,  B, -v,  p, -j,  d, -c,  i, -o,  u, -A,}, \
  {  E, -A,  w, -s,  o, -k,  g, -c,  b, -f,  j, -n,  r, -v,  z, -D, -F,  B, -x,  t, -p,  l, -h,  d, -a,  e, -i,  m, -q,  u, -y,  C,}, \
  {  F, -D,  B, -z,  x, -v,  t, -r,  p, -n,  l, -j,  h, -f,  d, -b,  a, -c,  e, -g,  i, -k,  m, -o,  q, -s,  u, -w,  y, -A,  C, -E,}, \
}


// DST-7
#define DEFINE_DST7_P4_MATRIX(a,b,c,d) \
{ \
  {  a,  b,  c,  d }, \
  {  c,  c,  0, -c }, \
  {  d, -a, -c,  b }, \
  {  b, -d,  c, -a }, \
}

#define DEFINE_DST7_P8_MATRIX(a,b,c,d,e,f,g,h) \
{ \
  {  a,  b,  c,  d,  e,  f,  g,  h,}, \
  {  c,  f,  h,  e,  b, -a, -d, -g,}, \
  {  e,  g,  b, -c, -h, -d,  a,  f,}, \
  {  g,  c, -d, -f,  a,  h,  b, -e,}, \
  {  h, -a, -g,  b,  f, -c, -e,  d,}, \
  {  f, -e, -a,  g, -d, -b,  h, -c,}, \
  {  d, -h,  e, -a, -c,  g, -f,  b,}, \
  {  b, -d,  f, -h,  g, -e,  c, -a,}, \
}

#define DEFINE_DST7_P16_MATRIX(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p) \
{ \
  {  a,  b,  c,  d,  e,  f,  g,  h,  i,  j,  k,  l,  m,  n,  o,  p,}, \
  {  c,  f,  i,  l,  o,  o,  l,  i,  f,  c,  0, -c, -f, -i, -l, -o,}, \
  {  e,  j,  o,  m,  h,  c, -b, -g, -l, -p, -k, -f, -a,  d,  i,  n,}, \
  {  g,  n,  l,  e, -b, -i, -p, -j, -c,  d,  k,  o,  h,  a, -f, -m,}, \
  {  i,  o,  f, -c, -l, -l, -c,  f,  o,  i,  0, -i, -o, -f,  c,  l,}, \
  {  k,  k,  0, -k, -k,  0,  k,  k,  0, -k, -k,  0,  k,  k,  0, -k,}, \
  {  m,  g, -f, -n, -a,  l,  h, -e, -o, -b,  k,  i, -d, -p, -c,  j,}, \
  {  o,  c, -l, -f,  i,  i, -f, -l,  c,  o,  0, -o, -c,  l,  f, -i,}, \
  {  p, -a, -o,  b,  n, -c, -m,  d,  l, -e, -k,  f,  j, -g, -i,  h,}, \
  {  n, -e, -i,  j,  d, -o,  a,  m, -f, -h,  k,  c, -p,  b,  l, -g,}, \
  {  l, -i, -c,  o, -f, -f,  o, -c, -i,  l,  0, -l,  i,  c, -o,  f,}, \
  {  j, -m,  c,  g, -p,  f,  d, -n,  i,  a, -k,  l, -b, -h,  o, -e,}, \
  {  h, -p,  i, -a, -g,  o, -j,  b,  f, -n,  k, -c, -e,  m, -l,  d,}, \
  {  f, -l,  o, -i,  c,  c, -i,  o, -l,  f,  0, -f,  l, -o,  i, -c,}, \
  {  d, -h,  l, -p,  m, -i,  e, -a, -c,  g, -k,  o, -n,  j, -f,  b,}, \
  {  b, -d,  f, -h,  j, -l,  n, -p,  o, -m,  k, -i,  g, -e,  c, -a,}, \
}

#define DEFINE_DST7_P32_MATRIX(a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z,A,B,C,D,E,F) \
{ \
  {  a,  b,  c,  d,  e,  f,  g,  h,  i,  j,  k,  l,  m,  n,  o,  p,  q,  r,  s,  t,  u,  v,  w,  x,  y,  z,  A,  B,  C,  D,  E,  F,}, \
  {  c,  f,  i,  l,  o,  r,  u,  x,  A,  D,  F,  C,  z,  w,  t,  q,  n,  k,  h,  e,  b, -a, -d, -g, -j, -m, -p, -s, -v, -y, -B, -E,}, \
  {  e,  j,  o,  t,  y,  D,  D,  y,  t,  o,  j,  e,  0, -e, -j, -o, -t, -y, -D, -D, -y, -t, -o, -j, -e,  0,  e,  j,  o,  t,  y,  D,}, \
  {  g,  n,  u,  B,  D,  w,  p,  i,  b, -e, -l, -s, -z, -F, -y, -r, -k, -d,  c,  j,  q,  x,  E,  A,  t,  m,  f, -a, -h, -o, -v, -C,}, \
  {  i,  r,  A,  C,  t,  k,  b, -g, -p, -y, -E, -v, -m, -d,  e,  n,  w,  F,  x,  o,  f, -c, -l, -u, -D, -z, -q, -h,  a,  j,  s,  B,}, \
  {  k,  v,  F,  u,  j, -a, -l, -w, -E, -t, -i,  b,  m,  x,  D,  s,  h, -c, -n, -y, -C, -r, -g,  d,  o,  z,  B,  q,  f, -e, -p, -A,}, \
  {  m,  z,  z,  m,  0, -m, -z, -z, -m,  0,  m,  z,  z,  m,  0, -m, -z, -z, -m,  0,  m,  z,  z,  m,  0, -m, -z, -z, -m,  0,  m,  z,}, \
  {  o,  D,  t,  e, -j, -y, -y, -j,  e,  t,  D,  o,  0, -o, -D, -t, -e,  j,  y,  y,  j, -e, -t, -D, -o,  0,  o,  D,  t,  e, -j, -y,}, \
  {  q,  E,  n, -c, -t, -B, -k,  f,  w,  y,  h, -i, -z, -v, -e,  l,  C,  s,  b, -o, -F, -p,  a,  r,  D,  m, -d, -u, -A, -j,  g,  x,}, \
  {  s,  A,  h, -k, -D, -p,  c,  v,  x,  e, -n, -F, -m,  f,  y,  u,  b, -q, -C, -j,  i,  B,  r, -a, -t, -z, -g,  l,  E,  o, -d, -w,}, \
  {  u,  w,  b, -s, -y, -d,  q,  A,  f, -o, -C, -h,  m,  E,  j, -k, -F, -l,  i,  D,  n, -g, -B, -p,  e,  z,  r, -c, -x, -t,  a,  v,}, \
  {  w,  s, -d, -A, -o,  h,  E,  k, -l, -D, -g,  p,  z,  c, -t, -v,  a,  x,  r, -e, -B, -n,  i,  F,  j, -m, -C, -f,  q,  y,  b, -u,}, \
  {  y,  o, -j, -D, -e,  t,  t, -e, -D, -j,  o,  y,  0, -y, -o,  j,  D,  e, -t, -t,  e,  D,  j, -o, -y,  0,  y,  o, -j, -D, -e,  t,}, \
  {  A,  k, -p, -v,  e,  F,  f, -u, -q,  j,  B,  a, -z, -l,  o,  w, -d, -E, -g,  t,  r, -i, -C, -b,  y,  m, -n, -x,  c,  D,  h, -s,}, \
  {  C,  g, -v, -n,  o,  u, -h, -B,  a,  D,  f, -w, -m,  p,  t, -i, -A,  b,  E,  e, -x, -l,  q,  s, -j, -z,  c,  F,  d, -y, -k,  r,}, \
  {  E,  c, -B, -f,  y,  i, -v, -l,  s,  o, -p, -r,  m,  u, -j, -x,  g,  A, -d, -D,  a,  F,  b, -C, -e,  z,  h, -w, -k,  t,  n, -q,}, \
  {  F, -a, -E,  b,  D, -c, -C,  d,  B, -e, -A,  f,  z, -g, -y,  h,  x, -i, -w,  j,  v, -k, -u,  l,  t, -m, -s,  n,  r, -o, -q,  p,}, \
  {  D, -e, -y,  j,  t, -o, -o,  t,  j, -y, -e,  D,  0, -D,  e,  y, -j, -t,  o,  o, -t, -j,  y,  e, -D,  0,  D, -e, -y,  j,  t, -o,}, \
  {  B, -i, -s,  r,  j, -A, -a,  C, -h, -t,  q,  k, -z, -b,  D, -g, -u,  p,  l, -y, -c,  E, -f, -v,  o,  m, -x, -d,  F, -e, -w,  n,}, \
  {  z, -m, -m,  z,  0, -z,  m,  m, -z,  0,  z, -m, -m,  z,  0, -z,  m,  m, -z,  0,  z, -m, -m,  z,  0, -z,  m,  m, -z,  0,  z, -m,}, \
  {  x, -q, -g,  E, -j, -n,  A, -c, -u,  t,  d, -B,  m,  k, -D,  f,  r, -w, -a,  y, -p, -h,  F, -i, -o,  z, -b, -v,  s,  e, -C,  l,}, \
  {  v, -u, -a,  w, -t, -b,  x, -s, -c,  y, -r, -d,  z, -q, -e,  A, -p, -f,  B, -o, -g,  C, -n, -h,  D, -m, -i,  E, -l, -j,  F, -k,}, \
  {  t, -y,  e,  o, -D,  j,  j, -D,  o,  e, -y,  t,  0, -t,  y, -e, -o,  D, -j, -j,  D, -o, -e,  y, -t,  0,  t, -y,  e,  o, -D,  j,}, \
  {  r, -C,  k,  g, -y,  v, -d, -n,  F, -o, -c,  u, -z,  h,  j, -B,  s, -a, -q,  D, -l, -f,  x, -w,  e,  m, -E,  p,  b, -t,  A, -i,}, \
  {  p, -F,  q, -a, -o,  E, -r,  b,  n, -D,  s, -c, -m,  C, -t,  d,  l, -B,  u, -e, -k,  A, -v,  f,  j, -z,  w, -g, -i,  y, -x,  h,}, \
  {  n, -B,  w, -i, -e,  s, -F,  r, -d, -j,  x, -A,  m,  a, -o,  C, -v,  h,  f, -t,  E, -q,  c,  k, -y,  z, -l, -b,  p, -D,  u, -g,}, \
  {  l, -x,  C, -q,  e,  g, -s,  E, -v,  j,  b, -n,  z, -A,  o, -c, -i,  u, -F,  t, -h, -d,  p, -B,  y, -m,  a,  k, -w,  D, -r,  f,}, \
  {  j, -t,  D, -y,  o, -e, -e,  o, -y,  D, -t,  j,  0, -j,  t, -D,  y, -o,  e,  e, -o,  y, -D,  t, -j,  0,  j, -t,  D, -y,  o, -e,}, \
  {  h, -p,  x, -F,  y, -q,  i, -a, -g,  o, -w,  E, -z,  r, -j,  b,  f, -n,  v, -D,  A, -s,  k, -c, -e,  m, -u,  C, -B,  t, -l,  d,}, \
  {  f, -l,  r, -x,  D, -C,  w, -q,  k, -e, -a,  g, -m,  s, -y,  E, -B,  v, -p,  j, -d, -b,  h, -n,  t, -z,  F, -A,  u, -o,  i, -c,}, \
  {  d, -h,  l, -p,  t, -x,  B, -F,  C, -y,  u, -q,  m, -i,  e, -a, -c,  g, -k,  o, -s,  w, -A,  E, -D,  z, -v,  r, -n,  j, -f,  b,}, \
  {  b, -d,  f, -h,  j, -l,  n, -p,  r, -t,  v, -x,  z, -B,  D, -F,  E, -C,  A, -y,  w, -u,  s, -q,  o, -m,  k, -i,  g, -e,  c, -a,}, \
}

#if JVET_R0351_HIGH_BIT_DEPTH_SUPPORT && RExt__HIGH_PRECISION_FORWARD_TRANSFORM
//--------------------------------------------------------------------------------------------------
// DCT-2
const TMatrixCoeff g_trCoreDCT2P2[TRANSFORM_NUMBER_OF_DIRECTIONS][2][2] =
{
  DEFINE_DCT2_P2_MATRIX(16384),
  DEFINE_DCT2_P2_MATRIX(64)
};

const TMatrixCoeff g_trCoreDCT2P4 [TRANSFORM_NUMBER_OF_DIRECTIONS][4][4]   =
{
  DEFINE_DCT2_P4_MATRIX  (16384, 21266,  9224),
  DEFINE_DCT2_P4_MATRIX  (   64,    83,    36)
};

const TMatrixCoeff g_trCoreDCT2P8[TRANSFORM_NUMBER_OF_DIRECTIONS][8][8] =
{
  DEFINE_DCT2_P8_MATRIX(16384, 21266,  9224, 22813, 19244, 12769,  4563),
  DEFINE_DCT2_P8_MATRIX(64,    83,    36,    89,    75,    50,    18)
};

const TMatrixCoeff g_trCoreDCT2P16[TRANSFORM_NUMBER_OF_DIRECTIONS][16][16] =
{
  DEFINE_DCT2_P16_MATRIX(16384, 21266,  9224, 22813, 19244, 12769,  4563, 23120, 22063, 20450, 17972, 14642, 11109,  6446,  2316),
  DEFINE_DCT2_P16_MATRIX(   64,    83,    36,    89,    75,    50,    18,    90,    87,    80,    70,    57,    43,    25,     9)
};

const TMatrixCoeff g_trCoreDCT2P32[TRANSFORM_NUMBER_OF_DIRECTIONS][32][32] =
{
  DEFINE_DCT2_P32_MATRIX(16384, 21266,  9224, 22813, 19244, 12769,  4563, 23120, 22063, 20450, 17972, 14642, 11109,  6446 , 2316, 23106, 22852, 22445, 21848, 20995, 19810, 18601, 17143, 15718, 13853, 11749,  9846,  7908,  5573,  3281,   946),
  DEFINE_DCT2_P32_MATRIX(   64,    83,    36,    89,    75,    50,    18,    90,    87,    80,    70,    57,    43,    25,     9,    90,    90,    88,    85,    82,    78,    73,    67,    61,    54,    46,    38,    31,    22,    13,     4)
};

const TMatrixCoeff g_trCoreDCT2P64[TRANSFORM_NUMBER_OF_DIRECTIONS][64][64] =
{
  DEFINE_DCT2_P64_MATRIX(16384, 21266,  9224, 22813, 19244, 12769,  4563, 23129, 22063, 20450, 17972, 14642, 11109,  6446,  2316, 23106, 22852, 22445, 21848, 20995, 19810, 18601, 17143, 15718, 13853, 11749,  9846,  7908,  5573,  3281,   946, 23360, 23053, 23048, 23023, 22610, 22339, 21936, 21502, 21266, 20730, 20251, 19726, 18731, 18201, 17638, 16604, 15881, 15084, 14322, 13340, 12238, 11330, 10493,  9428,  8426,  7100,  6151,  5101,  3848,  2734,  1754,   574),
  DEFINE_DCT2_P64_MATRIX(   64,    83,    36,    89,    75,    50,    18,    90,    87,    80,    70,    57,    43,    25,     9,    90,    90,    88,    85,    82,    78,    73,    67,    61,    54,    46,    38,    31,    22,    13,     4,    91,    90,    90,    90,    88,    87,    86,    84,    83,    81,    79,    77,    73,    71,    69,    65,    62,    59,    56,    52,    48,    44,    41,    37,    33,    28,    24,    20,    15,    11,     7,     2)
};

// DCT-8
const TMatrixCoeff g_trCoreDCT8P4[TRANSFORM_NUMBER_OF_DIRECTIONS][4][4] =
{
  DEFINE_DCT8_P4_MATRIX(21505, 18893, 14081,  7425),
  DEFINE_DCT8_P4_MATRIX(84,     74,     55,     29)
};
const TMatrixCoeff g_trCoreDCT8P8[TRANSFORM_NUMBER_OF_DIRECTIONS][8][8] =
{
  DEFINE_DCT8_P8_MATRIX(22018, 21790, 19958, 18154, 15363, 11754,  8148,  4350),
  DEFINE_DCT8_P8_MATRIX(   86,    85,    78,    71,    60,    46,    32,    17)
};
const TMatrixCoeff g_trCoreDCT8P16[TRANSFORM_NUMBER_OF_DIRECTIONS][16][16] =
{
  DEFINE_DCT8_P16_MATRIX(22569, 22542, 22202, 21664, 20754, 19738, 18787, 17369, 15781, 14044, 12209, 10360,  8498,  6421,  4295,  1967),
  DEFINE_DCT8_P16_MATRIX(   88,    88,    87,    85,    81,    77,    73,    68,    62,    55,    48,    40,    33,    25,    17,     8)
};
const TMatrixCoeff g_trCoreDCT8P32[TRANSFORM_NUMBER_OF_DIRECTIONS][32][32] =
{
  DEFINE_DCT8_P32_MATRIX(23065, 23136, 22715, 22533, 22544, 22053, 21901, 21463, 21131, 20385, 20019, 19708, 19007, 18415, 17448, 16894, 16143, 15230, 14312, 13616, 12679, 11526, 10770,  9720,  8606,  7734,  6623,  5414,  4478,  3225,  2291,  1043),
  DEFINE_DCT8_P32_MATRIX(   90,    90,    89,    88,    87,    86,    85,    84,    82,    80,    78,    77,    74,    72,    68,    66,    63,    60,    56,    53,    50,    46,    42,    38,    34,    30,    26,    21,    17,    13,     9,     4)
};

// DST-7
const TMatrixCoeff g_trCoreDST7P4[TRANSFORM_NUMBER_OF_DIRECTIONS][4][4] =
{
  DEFINE_DST7_P4_MATRIX( 7425, 14081, 18893, 21505),
  DEFINE_DST7_P4_MATRIX(   29,    55,    74,    84)
};
const TMatrixCoeff g_trCoreDST7P8[TRANSFORM_NUMBER_OF_DIRECTIONS][8][8] =
{
  DEFINE_DST7_P8_MATRIX( 4350,  8148, 11754, 15363, 18154, 19958, 21790, 22018),
  DEFINE_DST7_P8_MATRIX(   17,    32,    46,    60,    71,    78,    85,    86)
};
const TMatrixCoeff g_trCoreDST7P16[TRANSFORM_NUMBER_OF_DIRECTIONS][16][16] =
{
  DEFINE_DST7_P16_MATRIX(1967,  4295,  6421,  8498, 10360, 12209, 14044, 15781, 17369, 18787, 19738, 20754, 21664, 22202, 22542, 22569),
  DEFINE_DST7_P16_MATRIX(   8,    17,    25,    33,    40,    48,    55,    62,    68,    73,    77,    81,    85,    87,    88,    88)
};
const TMatrixCoeff g_trCoreDST7P32[TRANSFORM_NUMBER_OF_DIRECTIONS][32][32] =
{
  DEFINE_DST7_P32_MATRIX( 1043,  2291,  3225,  4478,  5414,  6623,  7734,  8606,  9720, 10770, 11526, 12679, 13616, 14312, 15230, 16143, 16894, 17448, 18415, 19007, 19708, 20019, 20385, 21131, 21463, 21901, 22053, 22544, 22533, 22715, 23136, 23065),
  DEFINE_DST7_P32_MATRIX(    4,     9,    13,    17,    21,    26,    30,    34,    38,    42,    46,    50,    53,    56,    60,    63,    66,    68,    72,    74,    77,    78,    80,    82,    84,    85,    86,    87,    88,    89,    90,    90)
};

#else
//--------------------------------------------------------------------------------------------------
// DCT-2
const TMatrixCoeff g_trCoreDCT2P2[TRANSFORM_NUMBER_OF_DIRECTIONS][2][2] =
{
  DEFINE_DCT2_P2_MATRIX(64),
  DEFINE_DCT2_P2_MATRIX(64)
};

const TMatrixCoeff g_trCoreDCT2P4[TRANSFORM_NUMBER_OF_DIRECTIONS][4][4] =
{
  DEFINE_DCT2_P4_MATRIX(64,    83,    36),
  DEFINE_DCT2_P4_MATRIX(64,    83,    36)
};

const TMatrixCoeff g_trCoreDCT2P8[TRANSFORM_NUMBER_OF_DIRECTIONS][8][8] =
{
  DEFINE_DCT2_P8_MATRIX(64,    83,    36,    89,    75,    50,    18),
  DEFINE_DCT2_P8_MATRIX(64,    83,    36,    89,    75,    50,    18)
};

const TMatrixCoeff g_trCoreDCT2P16[TRANSFORM_NUMBER_OF_DIRECTIONS][16][16] =
{
  DEFINE_DCT2_P16_MATRIX(64,    83,    36,    89,    75,    50,    18,    90,    87,    80,    70,    57,    43,    25,     9),
  DEFINE_DCT2_P16_MATRIX(64,    83,    36,    89,    75,    50,    18,    90,    87,    80,    70,    57,    43,    25,     9)
};

const TMatrixCoeff g_trCoreDCT2P32[TRANSFORM_NUMBER_OF_DIRECTIONS][32][32] =
{
  DEFINE_DCT2_P32_MATRIX(64,    83,    36,    89,    75,    50,    18,    90,    87,    80,    70,    57,    43,    25,     9,    90,    90,    88,    85,    82,    78,    73,    67,    61,    54,    46,    38,    31,    22,    13,     4),
  DEFINE_DCT2_P32_MATRIX(64,    83,    36,    89,    75,    50,    18,    90,    87,    80,    70,    57,    43,    25,     9,    90,    90,    88,    85,    82,    78,    73,    67,    61,    54,    46,    38,    31,    22,    13,     4)
};

const TMatrixCoeff g_trCoreDCT2P64[TRANSFORM_NUMBER_OF_DIRECTIONS][64][64] =
{
  DEFINE_DCT2_P64_MATRIX(64,    83,    36,    89,    75,    50,    18,    90,    87,    80,    70,    57,    43,    25,     9,    90,    90,    88,    85,    82,    78,    73,    67,    61,    54,    46,    38,    31,    22,    13,     4,    91,    90,    90,    90,    88,    87,    86,    84,    83,    81,    79,    77,    73,    71,    69,    65,    62,    59,    56,    52,    48,    44,    41,    37,    33,    28,    24,    20,    15,    11,     7,     2),
  DEFINE_DCT2_P64_MATRIX(64,    83,    36,    89,    75,    50,    18,    90,    87,    80,    70,    57,    43,    25,     9,    90,    90,    88,    85,    82,    78,    73,    67,    61,    54,    46,    38,    31,    22,    13,     4,    91,    90,    90,    90,    88,    87,    86,    84,    83,    81,    79,    77,    73,    71,    69,    65,    62,    59,    56,    52,    48,    44,    41,    37,    33,    28,    24,    20,    15,    11,     7,     2)
};
#if TU_256
const TMatrixCoeff g_trCoreDCT2P128[TRANSFORM_NUMBER_OF_DIRECTIONS][128][128] =
{
    {
      {     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64 },
      {     91,     90,     90,     90,     90,     90,     89,     89,     89,     88,     88,     87,     86,     86,     85,     84,     83,     82,     81,     80,     79,     78,     77,     76,     75,     73,     72,     71,     69,     68,     66,     65,     63,     62,     60,     58,     57,     55,     53,     51,     49,     47,     46,     44,     42,     40,     38,     36,     34,     32,     29,     27,     25,     23,     21,     19,     17,     14,     12,     10,      8,      6,      3,      1,     -1,     -3,     -6,     -8,    -10,    -12,    -14,    -17,    -19,    -21,    -23,    -25,    -27,    -29,    -32,    -34,    -36,    -38,    -40,    -42,    -44,    -46,    -47,    -49,    -51,    -53,    -55,    -57,    -58,    -60,    -62,    -63,    -65,    -66,    -68,    -69,    -71,    -72,    -73,    -75,    -76,    -77,    -78,    -79,    -80,    -81,    -82,    -83,    -84,    -85,    -86,    -86,    -87,    -88,    -88,    -89,    -89,    -89,    -90,    -90,    -90,    -90,    -90,    -91 },
      {     90,     90,     90,     89,     88,     87,     86,     84,     83,     81,     79,     76,     74,     71,     69,     66,     62,     59,     56,     52,     48,     45,     41,     37,     33,     28,     24,     20,     15,     11,      7,      2,     -2,     -7,    -11,    -15,    -20,    -24,    -28,    -33,    -37,    -41,    -45,    -48,    -52,    -56,    -59,    -62,    -66,    -69,    -71,    -74,    -76,    -79,    -81,    -83,    -84,    -86,    -87,    -88,    -89,    -90,    -90,    -90,    -90,    -90,    -90,    -89,    -88,    -87,    -86,    -84,    -83,    -81,    -79,    -76,    -74,    -71,    -69,    -66,    -62,    -59,    -56,    -52,    -48,    -45,    -41,    -37,    -33,    -28,    -24,    -20,    -15,    -11,     -7,     -2,      2,      7,     11,     15,     20,     24,     28,     33,     37,     41,     45,     48,     52,     56,     59,     62,     66,     69,     71,     74,     76,     79,     81,     83,     84,     86,     87,     88,     89,     90,     90,     90 },
      {     90,     90,     89,     88,     86,     83,     80,     77,     73,     69,     65,     60,     55,     49,     44,     38,     32,     25,     19,     12,      6,     -1,     -8,    -14,    -21,    -27,    -34,    -40,    -46,    -51,    -57,    -62,    -66,    -71,    -75,    -78,    -81,    -84,    -86,    -88,    -89,    -90,    -91,    -90,    -90,    -89,    -87,    -85,    -82,    -79,    -76,    -72,    -68,    -63,    -58,    -53,    -47,    -42,    -36,    -29,    -23,    -17,    -10,     -3,      3,     10,     17,     23,     29,     36,     42,     47,     53,     58,     63,     68,     72,     76,     79,     82,     85,     87,     89,     90,     90,     91,     90,     89,     88,     86,     84,     81,     78,     75,     71,     66,     62,     57,     51,     46,     40,     34,     27,     21,     14,      8,      1,     -6,    -12,    -19,    -25,    -32,    -38,    -44,    -49,    -55,    -60,    -65,    -69,    -73,    -77,    -80,    -83,    -86,    -88,    -89,    -90,    -90 },
      {     90,     90,     88,     85,     82,     78,     73,     67,     61,     54,     47,     39,     30,     22,     13,      4,     -4,    -13,    -22,    -30,    -39,    -47,    -54,    -61,    -67,    -73,    -78,    -82,    -85,    -88,    -90,    -90,    -90,    -90,    -88,    -85,    -82,    -78,    -73,    -67,    -61,    -54,    -47,    -39,    -30,    -22,    -13,     -4,      4,     13,     22,     30,     39,     47,     54,     61,     67,     73,     78,     82,     85,     88,     90,     90,     90,     90,     88,     85,     82,     78,     73,     67,     61,     54,     47,     39,     30,     22,     13,      4,     -4,    -13,    -22,    -30,    -39,    -47,    -54,    -61,    -67,    -73,    -78,    -82,    -85,    -88,    -90,    -90,    -90,    -90,    -88,    -85,    -82,    -78,    -73,    -67,    -61,    -54,    -47,    -39,    -30,    -22,    -13,     -4,      4,     13,     22,     30,     39,     47,     54,     61,     67,     73,     78,     82,     85,     88,     90,     90 },
      {     90,     89,     86,     82,     77,     71,     63,     55,     46,     36,     25,     14,      3,     -8,    -19,    -29,    -40,    -49,    -58,    -66,    -73,    -79,    -84,    -88,    -90,    -91,    -90,    -88,    -85,    -80,    -75,    -68,    -60,    -51,    -42,    -32,    -21,    -10,      1,     12,     23,     34,     44,     53,     62,     69,     76,     81,     86,     89,     90,     90,     89,     87,     83,     78,     72,     65,     57,     47,     38,     27,     17,      6,     -6,    -17,    -27,    -38,    -47,    -57,    -65,    -72,    -78,    -83,    -87,    -89,    -90,    -90,    -89,    -86,    -81,    -76,    -69,    -62,    -53,    -44,    -34,    -23,    -12,     -1,     10,     21,     32,     42,     51,     60,     68,     75,     80,     85,     88,     90,     91,     90,     88,     84,     79,     73,     66,     58,     49,     40,     29,     19,      8,     -3,    -14,    -25,    -36,    -46,    -55,    -63,    -71,    -77,    -82,    -86,    -89,    -90 },
      {     90,     88,     84,     79,     71,     62,     52,     41,     28,     15,      2,    -11,    -24,    -37,    -48,    -59,    -69,    -76,    -83,    -87,    -90,    -90,    -89,    -86,    -81,    -74,    -66,    -56,    -45,    -33,    -20,     -7,      7,     20,     33,     45,     56,     66,     74,     81,     86,     89,     90,     90,     87,     83,     76,     69,     59,     48,     37,     24,     11,     -2,    -15,    -28,    -41,    -52,    -62,    -71,    -79,    -84,    -88,    -90,    -90,    -88,    -84,    -79,    -71,    -62,    -52,    -41,    -28,    -15,     -2,     11,     24,     37,     48,     59,     69,     76,     83,     87,     90,     90,     89,     86,     81,     74,     66,     56,     45,     33,     20,      7,     -7,    -20,    -33,    -45,    -56,    -66,    -74,    -81,    -86,    -89,    -90,    -90,    -87,    -83,    -76,    -69,    -59,    -48,    -37,    -24,    -11,      2,     15,     28,     41,     52,     62,     71,     79,     84,     88,     90 },
      {     90,     88,     82,     75,     65,     53,     40,     25,     10,     -6,    -21,    -36,    -49,    -62,    -72,    -80,    -86,    -90,    -90,    -89,    -84,    -77,    -68,    -57,    -44,    -29,    -14,      1,     17,     32,     46,     58,     69,     78,     85,     89,     91,     89,     86,     79,     71,     60,     47,     34,     19,      3,    -12,    -27,    -42,    -55,    -66,    -76,    -83,    -88,    -90,    -90,    -87,    -81,    -73,    -63,    -51,    -38,    -23,     -8,      8,     23,     38,     51,     63,     73,     81,     87,     90,     90,     88,     83,     76,     66,     55,     42,     27,     12,     -3,    -19,    -34,    -47,    -60,    -71,    -79,    -86,    -89,    -91,    -89,    -85,    -78,    -69,    -58,    -46,    -32,    -17,     -1,     14,     29,     44,     57,     68,     77,     84,     89,     90,     90,     86,     80,     72,     62,     49,     36,     21,      6,    -10,    -25,    -40,    -53,    -65,    -75,    -82,    -88,    -90 },
      {     90,     87,     80,     70,     57,     43,     26,      9,     -9,    -26,    -43,    -57,    -70,    -80,    -87,    -90,    -90,    -87,    -80,    -70,    -57,    -43,    -26,     -9,      9,     26,     43,     57,     70,     80,     87,     90,     90,     87,     80,     70,     57,     43,     26,      9,     -9,    -26,    -43,    -57,    -70,    -80,    -87,    -90,    -90,    -87,    -80,    -70,    -57,    -43,    -26,     -9,      9,     26,     43,     57,     70,     80,     87,     90,     90,     87,     80,     70,     57,     43,     26,      9,     -9,    -26,    -43,    -57,    -70,    -80,    -87,    -90,    -90,    -87,    -80,    -70,    -57,    -43,    -26,     -9,      9,     26,     43,     57,     70,     80,     87,     90,     90,     87,     80,     70,     57,     43,     26,      9,     -9,    -26,    -43,    -57,    -70,    -80,    -87,    -90,    -90,    -87,    -80,    -70,    -57,    -43,    -26,     -9,      9,     26,     43,     57,     70,     80,     87,     90 },
      {     90,     86,     77,     65,     49,     32,     12,     -8,    -27,    -46,    -62,    -75,    -84,    -89,    -90,    -87,    -79,    -68,    -53,    -36,    -17,      3,     23,     42,     58,     72,     82,     89,     91,     88,     81,     71,     57,     40,     21,      1,    -19,    -38,    -55,    -69,    -80,    -88,    -90,    -89,    -83,    -73,    -60,    -44,    -25,     -6,     14,     34,     51,     66,     78,     86,     90,     90,     85,     76,     63,     47,     29,     10,    -10,    -29,    -47,    -63,    -76,    -85,    -90,    -90,    -86,    -78,    -66,    -51,    -34,    -14,      6,     25,     44,     60,     73,     83,     89,     90,     88,     80,     69,     55,     38,     19,     -1,    -21,    -40,    -57,    -71,    -81,    -88,    -91,    -89,    -82,    -72,    -58,    -42,    -23,     -3,     17,     36,     53,     68,     79,     87,     90,     89,     84,     75,     62,     46,     27,      8,    -12,    -32,    -49,    -65,    -77,    -86,    -90 },
      {     90,     84,     74,     59,     41,     20,     -2,    -24,    -45,    -62,    -76,    -86,    -90,    -89,    -83,    -71,    -56,    -37,    -15,      7,     28,     48,     66,     79,     87,     90,     88,     81,     69,     52,     33,     11,    -11,    -33,    -52,    -69,    -81,    -88,    -90,    -87,    -79,    -66,    -48,    -28,     -7,     15,     37,     56,     71,     83,     89,     90,     86,     76,     62,     45,     24,      2,    -20,    -41,    -59,    -74,    -84,    -90,    -90,    -84,    -74,    -59,    -41,    -20,      2,     24,     45,     62,     76,     86,     90,     89,     83,     71,     56,     37,     15,     -7,    -28,    -48,    -66,    -79,    -87,    -90,    -88,    -81,    -69,    -52,    -33,    -11,     11,     33,     52,     69,     81,     88,     90,     87,     79,     66,     48,     28,      7,    -15,    -37,    -56,    -71,    -83,    -89,    -90,    -86,    -76,    -62,    -45,    -24,     -2,     20,     41,     59,     74,     84,     90 },
      {     90,     83,     71,     53,     32,      8,    -17,    -40,    -60,    -76,    -86,    -90,    -88,    -79,    -65,    -46,    -23,      1,     25,     47,     66,     80,     89,     90,     86,     75,     58,     38,     14,    -10,    -34,    -55,    -72,    -84,    -90,    -89,    -82,    -69,    -51,    -29,     -6,     19,     42,     62,     77,     87,     91,     88,     78,     63,     44,     21,     -3,    -27,    -49,    -68,    -81,    -89,    -90,    -85,    -73,    -57,    -36,    -12,     12,     36,     57,     73,     85,     90,     89,     81,     68,     49,     27,      3,    -21,    -44,    -63,    -78,    -88,    -91,    -87,    -77,    -62,    -42,    -19,      6,     29,     51,     69,     82,     89,     90,     84,     72,     55,     34,     10,    -14,    -38,    -58,    -75,    -86,    -90,    -89,    -80,    -66,    -47,    -25,     -1,     23,     46,     65,     79,     88,     90,     86,     76,     60,     40,     17,     -8,    -32,    -53,    -71,    -83,    -90 },
      {     90,     82,     67,     47,     22,     -4,    -30,    -54,    -73,    -85,    -90,    -88,    -78,    -61,    -39,    -13,     13,     39,     61,     78,     88,     90,     85,     73,     54,     30,      4,    -22,    -47,    -67,    -82,    -90,    -90,    -82,    -67,    -47,    -22,      4,     30,     54,     73,     85,     90,     88,     78,     61,     39,     13,    -13,    -39,    -61,    -78,    -88,    -90,    -85,    -73,    -54,    -30,     -4,     22,     47,     67,     82,     90,     90,     82,     67,     47,     22,     -4,    -30,    -54,    -73,    -85,    -90,    -88,    -78,    -61,    -39,    -13,     13,     39,     61,     78,     88,     90,     85,     73,     54,     30,      4,    -22,    -47,    -67,    -82,    -90,    -90,    -82,    -67,    -47,    -22,      4,     30,     54,     73,     85,     90,     88,     78,     61,     39,     13,    -13,    -39,    -61,    -78,    -88,    -90,    -85,    -73,    -54,    -30,     -4,     22,     47,     67,     82,     90 },
      {     89,     80,     63,     40,     12,    -17,    -44,    -66,    -82,    -90,    -89,    -78,    -60,    -36,     -8,     21,     47,     69,     84,     90,     88,     76,     57,     32,      3,    -25,    -51,    -72,    -86,    -91,    -86,    -73,    -53,    -27,      1,     29,     55,     75,     87,     90,     85,     71,     49,     23,     -6,    -34,    -58,    -77,    -88,    -90,    -83,    -68,    -46,    -19,     10,     38,     62,     79,     89,     90,     81,     65,     42,     14,    -14,    -42,    -65,    -81,    -90,    -89,    -79,    -62,    -38,    -10,     19,     46,     68,     83,     90,     88,     77,     58,     34,      6,    -23,    -49,    -71,    -85,    -90,    -87,    -75,    -55,    -29,     -1,     27,     53,     73,     86,     91,     86,     72,     51,     25,     -3,    -32,    -57,    -76,    -88,    -90,    -84,    -69,    -47,    -21,      8,     36,     60,     78,     89,     90,     82,     66,     44,     17,    -12,    -40,    -63,    -80,    -89 },
      {     89,     79,     59,     33,      2,    -28,    -56,    -76,    -88,    -90,    -81,    -62,    -37,     -7,     24,     52,     74,     87,     90,     83,     66,     41,     11,    -20,    -48,    -71,    -86,    -90,    -84,    -69,    -45,    -15,     15,     45,     69,     84,     90,     86,     71,     48,     20,    -11,    -41,    -66,    -83,    -90,    -87,    -74,    -52,    -24,      7,     37,     62,     81,     90,     88,     76,     56,     28,     -2,    -33,    -59,    -79,    -89,    -89,    -79,    -59,    -33,     -2,     28,     56,     76,     88,     90,     81,     62,     37,      7,    -24,    -52,    -74,    -87,    -90,    -83,    -66,    -41,    -11,     20,     48,     71,     86,     90,     84,     69,     45,     15,    -15,    -45,    -69,    -84,    -90,    -86,    -71,    -48,    -20,     11,     41,     66,     83,     90,     87,     74,     52,     24,     -7,    -37,    -62,    -81,    -90,    -88,    -76,    -56,    -28,      2,     33,     59,     79,     89 },
      {     89,     77,     55,     25,     -8,    -40,    -66,    -84,    -91,    -85,    -68,    -42,    -10,     23,     53,     76,     89,     89,     78,     57,     27,     -6,    -38,    -65,    -83,    -90,    -86,    -69,    -44,    -12,     21,     51,     75,     88,     90,     79,     58,     29,     -3,    -36,    -63,    -82,    -90,    -86,    -71,    -46,    -14,     19,     49,     73,     88,     90,     80,     60,     32,     -1,    -34,    -62,    -81,    -90,    -87,    -72,    -47,    -17,     17,     47,     72,     87,     90,     81,     62,     34,      1,    -32,    -60,    -80,    -90,    -88,    -73,    -49,    -19,     14,     46,     71,     86,     90,     82,     63,     36,      3,    -29,    -58,    -79,    -90,    -88,    -75,    -51,    -21,     12,     44,     69,     86,     90,     83,     65,     38,      6,    -27,    -57,    -78,    -89,    -89,    -76,    -53,    -23,     10,     42,     68,     85,     91,     84,     66,     40,      8,    -25,    -55,    -77,    -89 },
      {     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89 },
      {     89,     73,     46,     10,    -27,    -60,    -82,    -91,    -83,    -62,    -29,      8,     44,     72,     88,     89,     75,     47,     12,    -25,    -58,    -81,    -90,    -84,    -63,    -32,      6,     42,     71,     88,     89,     76,     49,     14,    -23,    -57,    -80,    -90,    -85,    -65,    -34,      3,     40,     69,     87,     90,     77,     51,     17,    -21,    -55,    -79,    -90,    -86,    -66,    -36,      1,     38,     68,     86,     90,     78,     53,     19,    -19,    -53,    -78,    -90,    -86,    -68,    -38,     -1,     36,     66,     86,     90,     79,     55,     21,    -17,    -51,    -77,    -90,    -87,    -69,    -40,     -3,     34,     65,     85,     90,     80,     57,     23,    -14,    -49,    -76,    -89,    -88,    -71,    -42,     -6,     32,     63,     84,     90,     81,     58,     25,    -12,    -47,    -75,    -89,    -88,    -72,    -44,     -8,     29,     62,     83,     91,     82,     60,     27,    -10,    -46,    -73,    -89 },
      {     88,     71,     41,      2,    -37,    -69,    -87,    -89,    -74,    -45,     -7,     33,     66,     86,     90,     76,     48,     11,    -28,    -62,    -84,    -90,    -79,    -52,    -15,     24,     59,     83,     90,     81,     56,     20,    -20,    -56,    -81,    -90,    -83,    -59,    -24,     15,     52,     79,     90,     84,     62,     28,    -11,    -48,    -76,    -90,    -86,    -66,    -33,      7,     45,     74,     89,     87,     69,     37,     -2,    -41,    -71,    -88,    -88,    -71,    -41,     -2,     37,     69,     87,     89,     74,     45,      7,    -33,    -66,    -86,    -90,    -76,    -48,    -11,     28,     62,     84,     90,     79,     52,     15,    -24,    -59,    -83,    -90,    -81,    -56,    -20,     20,     56,     81,     90,     83,     59,     24,    -15,    -52,    -79,    -90,    -84,    -62,    -28,     11,     48,     76,     90,     86,     66,     33,     -7,    -45,    -74,    -89,    -87,    -69,    -37,      2,     41,     71,     88 },
      {     88,     69,     36,     -6,    -46,    -76,    -90,    -85,    -62,    -25,     17,     55,     81,     91,     80,     53,     14,    -27,    -63,    -86,    -90,    -75,    -44,     -3,     38,     71,     89,     88,     68,     34,     -8,    -47,    -77,    -90,    -84,    -60,    -23,     19,     57,     82,     90,     79,     51,     12,    -29,    -65,    -86,    -89,    -73,    -42,     -1,     40,     72,     89,     87,     66,     32,    -10,    -49,    -78,    -90,    -83,    -58,    -21,     21,     58,     83,     90,     78,     49,     10,    -32,    -66,    -87,    -89,    -72,    -40,      1,     42,     73,     89,     86,     65,     29,    -12,    -51,    -79,    -90,    -82,    -57,    -19,     23,     60,     84,     90,     77,     47,      8,    -34,    -68,    -88,    -89,    -71,    -38,      3,     44,     75,     90,     86,     63,     27,    -14,    -53,    -80,    -91,    -81,    -55,    -17,     25,     62,     85,     90,     76,     46,      6,    -36,    -69,    -88 },
      {     88,     67,     30,    -13,    -54,    -82,    -90,    -78,    -47,     -4,     39,     73,     90,     85,     61,     22,    -22,    -61,    -85,    -90,    -73,    -39,      4,     47,     78,     90,     82,     54,     13,    -30,    -67,    -88,    -88,    -67,    -30,     13,     54,     82,     90,     78,     47,      4,    -39,    -73,    -90,    -85,    -61,    -22,     22,     61,     85,     90,     73,     39,     -4,    -47,    -78,    -90,    -82,    -54,    -13,     30,     67,     88,     88,     67,     30,    -13,    -54,    -82,    -90,    -78,    -47,     -4,     39,     73,     90,     85,     61,     22,    -22,    -61,    -85,    -90,    -73,    -39,      4,     47,     78,     90,     82,     54,     13,    -30,    -67,    -88,    -88,    -67,    -30,     13,     54,     82,     90,     78,     47,      4,    -39,    -73,    -90,    -85,    -61,    -22,     22,     61,     85,     90,     73,     39,     -4,    -47,    -78,    -90,    -82,    -54,    -13,     30,     67,     88 },
      {     88,     65,     25,    -21,    -62,    -86,    -89,    -68,    -29,     17,     58,     85,     89,     71,     34,    -12,    -55,    -83,    -90,    -73,    -38,      8,     51,     81,     90,     76,     42,     -3,    -47,    -79,    -91,    -78,    -46,     -1,     44,     77,     90,     80,     49,      6,    -40,    -75,    -90,    -82,    -53,    -10,     36,     72,     90,     84,     57,     14,    -32,    -69,    -89,    -86,    -60,    -19,     27,     66,     88,     87,     63,     23,    -23,    -63,    -87,    -88,    -66,    -27,     19,     60,     86,     89,     69,     32,    -14,    -57,    -84,    -90,    -72,    -36,     10,     53,     82,     90,     75,     40,     -6,    -49,    -80,    -90,    -77,    -44,      1,     46,     78,     91,     79,     47,      3,    -42,    -76,    -90,    -81,    -51,     -8,     38,     73,     90,     83,     55,     12,    -34,    -71,    -89,    -85,    -58,    -17,     29,     68,     89,     86,     62,     21,    -25,    -65,    -88 },
      {     87,     62,     20,    -28,    -69,    -89,    -84,    -56,    -11,     37,     74,     90,     81,     48,      2,    -45,    -79,    -90,    -76,    -41,      7,     52,     83,     90,     71,     33,    -15,    -59,    -86,    -88,    -66,    -24,     24,     66,     88,     86,     59,     15,    -33,    -71,    -90,    -83,    -52,     -7,     41,     76,     90,     79,     45,     -2,    -48,    -81,    -90,    -74,    -37,     11,     56,     84,     89,     69,     28,    -20,    -62,    -87,    -87,    -62,    -20,     28,     69,     89,     84,     56,     11,    -37,    -74,    -90,    -81,    -48,     -2,     45,     79,     90,     76,     41,     -7,    -52,    -83,    -90,    -71,    -33,     15,     59,     86,     88,     66,     24,    -24,    -66,    -88,    -86,    -59,    -15,     33,     71,     90,     83,     52,      7,    -41,    -76,    -90,    -79,    -45,      2,     48,     81,     90,     74,     37,    -11,    -56,    -84,    -89,    -69,    -28,     20,     62,     87 },
      {     87,     60,     14,    -36,    -75,    -90,    -78,    -42,      8,     55,     85,     89,     65,     21,    -29,    -71,    -90,    -81,    -47,      1,     49,     82,     90,     69,     27,    -23,    -66,    -89,    -84,    -53,     -6,     44,     79,     90,     73,     34,    -17,    -62,    -88,    -86,    -58,    -12,     38,     76,     91,     77,     40,    -10,    -57,    -86,    -88,    -63,    -19,     32,     72,     90,     80,     46,     -3,    -51,    -83,    -89,    -68,    -25,     25,     68,     89,     83,     51,      3,    -46,    -80,    -90,    -72,    -32,     19,     63,     88,     86,     57,     10,    -40,    -77,    -91,    -76,    -38,     12,     58,     86,     88,     62,     17,    -34,    -73,    -90,    -79,    -44,      6,     53,     84,     89,     66,     23,    -27,    -69,    -90,    -82,    -49,     -1,     47,     81,     90,     71,     29,    -21,    -65,    -89,    -85,    -55,     -8,     42,     78,     90,     75,     36,    -14,    -60,    -87 },
      {     87,     57,      9,    -43,    -80,    -90,    -70,    -26,     26,     70,     90,     80,     43,     -9,    -57,    -87,    -87,    -57,     -9,     43,     80,     90,     70,     26,    -26,    -70,    -90,    -80,    -43,      9,     57,     87,     87,     57,      9,    -43,    -80,    -90,    -70,    -26,     26,     70,     90,     80,     43,     -9,    -57,    -87,    -87,    -57,     -9,     43,     80,     90,     70,     26,    -26,    -70,    -90,    -80,    -43,      9,     57,     87,     87,     57,      9,    -43,    -80,    -90,    -70,    -26,     26,     70,     90,     80,     43,     -9,    -57,    -87,    -87,    -57,     -9,     43,     80,     90,     70,     26,    -26,    -70,    -90,    -80,    -43,      9,     57,     87,     87,     57,      9,    -43,    -80,    -90,    -70,    -26,     26,     70,     90,     80,     43,     -9,    -57,    -87,    -87,    -57,     -9,     43,     80,     90,     70,     26,    -26,    -70,    -90,    -80,    -43,      9,     57,     87 },
      {     86,     55,      3,    -49,    -84,    -88,    -60,    -10,     44,     81,     89,     65,     17,    -38,    -78,    -90,    -69,    -23,     32,     75,     91,     73,     29,    -25,    -71,    -90,    -77,    -36,     19,     66,     90,     80,     42,    -12,    -62,    -89,    -83,    -47,      6,     57,     87,     86,     53,      1,    -51,    -85,    -88,    -58,     -8,     46,     82,     89,     63,     14,    -40,    -79,    -90,    -68,    -21,     34,     76,     90,     72,     27,    -27,    -72,    -90,    -76,    -34,     21,     68,     90,     79,     40,    -14,    -63,    -89,    -82,    -46,      8,     58,     88,     85,     51,     -1,    -53,    -86,    -87,    -57,     -6,     47,     83,     89,     62,     12,    -42,    -80,    -90,    -66,    -19,     36,     77,     90,     71,     25,    -29,    -73,    -91,    -75,    -32,     23,     69,     90,     78,     38,    -17,    -65,    -89,    -81,    -44,     10,     60,     88,     84,     49,     -3,    -55,    -86 },
      {     86,     52,     -2,    -56,    -87,    -84,    -48,      7,     59,     88,     83,     45,    -11,    -62,    -89,    -81,    -41,     15,     66,     90,     79,     37,    -20,    -69,    -90,    -76,    -33,     24,     71,     90,     74,     28,    -28,    -74,    -90,    -71,    -24,     33,     76,     90,     69,     20,    -37,    -79,    -90,    -66,    -15,     41,     81,     89,     62,     11,    -45,    -83,    -88,    -59,     -7,     48,     84,     87,     56,      2,    -52,    -86,    -86,    -52,      2,     56,     87,     84,     48,     -7,    -59,    -88,    -83,    -45,     11,     62,     89,     81,     41,    -15,    -66,    -90,    -79,    -37,     20,     69,     90,     76,     33,    -24,    -71,    -90,    -74,    -28,     28,     74,     90,     71,     24,    -33,    -76,    -90,    -69,    -20,     37,     79,     90,     66,     15,    -41,    -81,    -89,    -62,    -11,     45,     83,     88,     59,      7,    -48,    -84,    -87,    -56,     -2,     52,     86 },
      {     86,     49,     -8,    -62,    -89,    -79,    -36,     23,     72,     91,     71,     21,    -38,    -80,    -89,    -60,     -6,     51,     86,     85,     47,    -10,    -63,    -90,    -78,    -34,     25,     73,     90,     69,     19,    -40,    -81,    -89,    -58,     -3,     53,     87,     84,     46,    -12,    -65,    -90,    -77,    -32,     27,     75,     90,     68,     17,    -42,    -82,    -88,    -57,     -1,     55,     88,     83,     44,    -14,    -66,    -90,    -76,    -29,     29,     76,     90,     66,     14,    -44,    -83,    -88,    -55,      1,     57,     88,     82,     42,    -17,    -68,    -90,    -75,    -27,     32,     77,     90,     65,     12,    -46,    -84,    -87,    -53,      3,     58,     89,     81,     40,    -19,    -69,    -90,    -73,    -25,     34,     78,     90,     63,     10,    -47,    -85,    -86,    -51,      6,     60,     89,     80,     38,    -21,    -71,    -91,    -72,    -23,     36,     79,     89,     62,      8,    -49,    -86 },
      {     85,     47,    -13,    -67,    -90,    -73,    -22,     39,     82,     88,     54,     -4,    -61,    -90,    -78,    -30,     30,     78,     90,     61,      4,    -54,    -88,    -82,    -39,     22,     73,     90,     67,     13,    -47,    -85,    -85,    -47,     13,     67,     90,     73,     22,    -39,    -82,    -88,    -54,      4,     61,     90,     78,     30,    -30,    -78,    -90,    -61,     -4,     54,     88,     82,     39,    -22,    -73,    -90,    -67,    -13,     47,     85,     85,     47,    -13,    -67,    -90,    -73,    -22,     39,     82,     88,     54,     -4,    -61,    -90,    -78,    -30,     30,     78,     90,     61,      4,    -54,    -88,    -82,    -39,     22,     73,     90,     67,     13,    -47,    -85,    -85,    -47,     13,     67,     90,     73,     22,    -39,    -82,    -88,    -54,      4,     61,     90,     78,     30,    -30,    -78,    -90,    -61,     -4,     54,     88,     82,     39,    -22,    -73,    -90,    -67,    -13,     47,     85 },
      {     85,     44,    -19,    -72,    -90,    -65,     -8,     53,     88,     80,     34,    -29,    -78,    -89,    -57,      3,     62,     90,     75,     23,    -40,    -83,    -86,    -47,     14,     69,     91,     68,     12,    -49,    -87,    -82,    -38,     25,     76,     90,     60,      1,    -58,    -89,    -77,    -27,     36,     81,     88,     51,    -10,    -66,    -90,    -71,    -17,     46,     86,     84,     42,    -21,    -73,    -90,    -63,     -6,     55,     89,     79,     32,    -32,    -79,    -89,    -55,      6,     63,     90,     73,     21,    -42,    -84,    -86,    -46,     17,     71,     90,     66,     10,    -51,    -88,    -81,    -36,     27,     77,     89,     58,     -1,    -60,    -90,    -76,    -25,     38,     82,     87,     49,    -12,    -68,    -91,    -69,    -14,     47,     86,     83,     40,    -23,    -75,    -90,    -62,     -3,     57,     89,     78,     29,    -34,    -80,    -88,    -53,      8,     65,     90,     72,     19,    -44,    -85 },
      {     84,     41,    -24,    -76,    -89,    -56,      7,     66,     90,     69,     11,    -52,    -88,    -79,    -28,     37,     83,     86,     45,    -20,    -74,    -90,    -59,      2,     62,     90,     71,     15,    -48,    -87,    -81,    -33,     33,     81,     87,     48,    -15,    -71,    -90,    -62,     -2,     59,     90,     74,     20,    -45,    -86,    -83,    -37,     28,     79,     88,     52,    -11,    -69,    -90,    -66,     -7,     56,     89,     76,     24,    -41,    -84,    -84,    -41,     24,     76,     89,     56,     -7,    -66,    -90,    -69,    -11,     52,     88,     79,     28,    -37,    -83,    -86,    -45,     20,     74,     90,     59,     -2,    -62,    -90,    -71,    -15,     48,     87,     81,     33,    -33,    -81,    -87,    -48,     15,     71,     90,     62,      2,    -59,    -90,    -74,    -20,     45,     86,     83,     37,    -28,    -79,    -88,    -52,     11,     69,     90,     66,      7,    -56,    -89,    -76,    -24,     41,     84 },
      {     84,     38,    -29,    -80,    -87,    -46,     21,     76,     89,     53,    -12,    -71,    -90,    -60,      3,     65,     91,     66,      6,    -58,    -90,    -72,    -14,     51,     89,     77,     23,    -44,    -86,    -81,    -32,     36,     83,     85,     40,    -27,    -79,    -88,    -47,     19,     75,     89,     55,    -10,    -69,    -90,    -62,      1,     63,     90,     68,      8,    -57,    -90,    -73,    -17,     49,     88,     78,     25,    -42,    -86,    -82,    -34,     34,     82,     86,     42,    -25,    -78,    -88,    -49,     17,     73,     90,     57,     -8,    -68,    -90,    -63,     -1,     62,     90,     69,     10,    -55,    -89,    -75,    -19,     47,     88,     79,     27,    -40,    -85,    -83,    -36,     32,     81,     86,     44,    -23,    -77,    -89,    -51,     14,     72,     90,     58,     -6,    -66,    -91,    -65,     -3,     60,     90,     71,     12,    -53,    -89,    -76,    -21,     46,     87,     80,     29,    -38,    -84 },
      {     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84 },
      {     83,     32,    -40,    -86,    -79,    -23,     47,     89,     75,     14,    -55,    -90,    -69,     -6,     62,     91,     63,     -3,    -68,    -90,    -57,     12,     73,     89,     49,    -21,    -78,    -87,    -42,     29,     82,     84,     34,    -38,    -86,    -80,    -25,     46,     88,     76,     17,    -53,    -90,    -71,     -8,     60,     90,     65,     -1,    -66,    -90,    -58,     10,     72,     89,     51,    -19,    -77,    -88,    -44,     27,     81,     85,     36,    -36,    -85,    -81,    -27,     44,     88,     77,     19,    -51,    -89,    -72,    -10,     58,     90,     66,      1,    -65,    -90,    -60,      8,     71,     90,     53,    -17,    -76,    -88,    -46,     25,     80,     86,     38,    -34,    -84,    -82,    -29,     42,     87,     78,     21,    -49,    -89,    -73,    -12,     57,     90,     68,      3,    -63,    -91,    -62,      6,     69,     90,     55,    -14,    -75,    -89,    -47,     23,     79,     86,     40,    -32,    -83 },
      {     83,     28,    -45,    -88,    -74,    -11,     59,     90,     62,     -7,    -71,    -89,    -48,     24,     81,     84,     33,    -41,    -87,    -76,    -15,     56,     90,     66,     -2,    -69,    -90,    -52,     20,     79,     86,     37,    -37,    -86,    -79,    -20,     52,     90,     69,      2,    -66,    -90,    -56,     15,     76,     87,     41,    -33,    -84,    -81,    -24,     48,     89,     71,      7,    -62,    -90,    -59,     11,     74,     88,     45,    -28,    -83,    -83,    -28,     45,     88,     74,     11,    -59,    -90,    -62,      7,     71,     89,     48,    -24,    -81,    -84,    -33,     41,     87,     76,     15,    -56,    -90,    -66,      2,     69,     90,     52,    -20,    -79,    -86,    -37,     37,     86,     79,     20,    -52,    -90,    -69,     -2,     66,     90,     56,    -15,    -76,    -87,    -41,     33,     84,     81,     24,    -48,    -89,    -71,     -7,     62,     90,     59,    -11,    -74,    -88,    -45,     28,     83 },
      {     82,     25,    -49,    -90,    -68,      1,     69,     89,     47,    -27,    -83,    -81,    -23,     51,     90,     66,     -3,    -71,    -89,    -46,     29,     84,     80,     21,    -53,    -90,    -65,      6,     72,     89,     44,    -32,    -85,    -79,    -19,     55,     90,     63,     -8,    -73,    -88,    -42,     34,     86,     78,     17,    -57,    -90,    -62,     10,     75,     88,     40,    -36,    -86,    -77,    -14,     58,     91,     60,    -12,    -76,    -87,    -38,     38,     87,     76,     12,    -60,    -91,    -58,     14,     77,     86,     36,    -40,    -88,    -75,    -10,     62,     90,     57,    -17,    -78,    -86,    -34,     42,     88,     73,      8,    -63,    -90,    -55,     19,     79,     85,     32,    -44,    -89,    -72,     -6,     65,     90,     53,    -21,    -80,    -84,    -29,     46,     89,     71,      3,    -66,    -90,    -51,     23,     81,     83,     27,    -47,    -89,    -69,     -1,     68,     90,     49,    -25,    -82 },
      {     82,     22,    -54,    -90,    -61,     13,     78,     85,     30,    -47,    -90,    -67,      4,     73,     88,     39,    -39,    -88,    -73,     -4,     67,     90,     47,    -30,    -85,    -78,    -13,     61,     90,     54,    -22,    -82,    -82,    -22,     54,     90,     61,    -13,    -78,    -85,    -30,     47,     90,     67,     -4,    -73,    -88,    -39,     39,     88,     73,      4,    -67,    -90,    -47,     30,     85,     78,     13,    -61,    -90,    -54,     22,     82,     82,     22,    -54,    -90,    -61,     13,     78,     85,     30,    -47,    -90,    -67,      4,     73,     88,     39,    -39,    -88,    -73,     -4,     67,     90,     47,    -30,    -85,    -78,    -13,     61,     90,     54,    -22,    -82,    -82,    -22,     54,     90,     61,    -13,    -78,    -85,    -30,     47,     90,     67,     -4,    -73,    -88,    -39,     39,     88,     73,      4,    -67,    -90,    -47,     30,     85,     78,     13,    -61,    -90,    -54,     22,     82 },
      {     81,     19,    -58,    -90,    -53,     25,     84,     78,     12,    -63,    -90,    -47,     32,     86,     75,      6,    -68,    -89,    -42,     38,     88,     71,     -1,    -72,    -88,    -36,     44,     89,     66,     -8,    -76,    -86,    -29,     49,     90,     62,    -14,    -79,    -83,    -23,     55,     91,     57,    -21,    -82,    -80,    -17,     60,     90,     51,    -27,    -85,    -77,    -10,     65,     90,     46,    -34,    -87,    -73,     -3,     69,     89,     40,    -40,    -89,    -69,      3,     73,     87,     34,    -46,    -90,    -65,     10,     77,     85,     27,    -51,    -90,    -60,     17,     80,     82,     21,    -57,    -91,    -55,     23,     83,     79,     14,    -62,    -90,    -49,     29,     86,     76,      8,    -66,    -89,    -44,     36,     88,     72,      1,    -71,    -88,    -38,     42,     89,     68,     -6,    -75,    -86,    -32,     47,     90,     63,    -12,    -78,    -84,    -25,     53,     90,     58,    -19,    -81 },
      {     81,     15,    -62,    -90,    -45,     37,     88,     69,     -7,    -76,    -84,    -24,     56,     90,     52,    -28,    -86,    -74,     -2,     71,     87,     33,    -48,    -90,    -59,     20,     83,     79,     11,    -66,    -89,    -41,     41,     89,     66,    -11,    -79,    -83,    -20,     59,     90,     48,    -33,    -87,    -71,      2,     74,     86,     28,    -52,    -90,    -56,     24,     84,     76,      7,    -69,    -88,    -37,     45,     90,     62,    -15,    -81,    -81,    -15,     62,     90,     45,    -37,    -88,    -69,      7,     76,     84,     24,    -56,    -90,    -52,     28,     86,     74,      2,    -71,    -87,    -33,     48,     90,     59,    -20,    -83,    -79,    -11,     66,     89,     41,    -41,    -89,    -66,     11,     79,     83,     20,    -59,    -90,    -48,     33,     87,     71,     -2,    -74,    -86,    -28,     52,     90,     56,    -24,    -84,    -76,     -7,     69,     88,     37,    -45,    -90,    -62,     15,     81 },
      {     80,     12,    -66,    -89,    -36,     47,     90,     57,    -25,    -86,    -73,      1,     75,     85,     23,    -58,    -90,    -46,     38,     89,     65,    -14,    -81,    -79,    -10,     68,     88,     34,    -49,    -90,    -55,     27,     86,     72,     -3,    -76,    -84,    -21,     60,     90,     44,    -40,    -89,    -63,     17,     82,     78,      8,    -69,    -88,    -32,     51,     91,     53,    -29,    -87,    -71,      6,     77,     83,     19,    -62,    -90,    -42,     42,     90,     62,    -19,    -83,    -77,     -6,     71,     87,     29,    -53,    -91,    -51,     32,     88,     69,     -8,    -78,    -82,    -17,     63,     89,     40,    -44,    -90,    -60,     21,     84,     76,      3,    -72,    -86,    -27,     55,     90,     49,    -34,    -88,    -68,     10,     79,     81,     14,    -65,    -89,    -38,     46,     90,     58,    -23,    -85,    -75,     -1,     73,     86,     25,    -57,    -90,    -47,     36,     89,     66,    -12,    -80 },
      {     80,      9,    -70,    -87,    -26,     57,     90,     43,    -43,    -90,    -57,     26,     87,     70,     -9,    -80,    -80,     -9,     70,     87,     26,    -57,    -90,    -43,     43,     90,     57,    -26,    -87,    -70,      9,     80,     80,      9,    -70,    -87,    -26,     57,     90,     43,    -43,    -90,    -57,     26,     87,     70,     -9,    -80,    -80,     -9,     70,     87,     26,    -57,    -90,    -43,     43,     90,     57,    -26,    -87,    -70,      9,     80,     80,      9,    -70,    -87,    -26,     57,     90,     43,    -43,    -90,    -57,     26,     87,     70,     -9,    -80,    -80,     -9,     70,     87,     26,    -57,    -90,    -43,     43,     90,     57,    -26,    -87,    -70,      9,     80,     80,      9,    -70,    -87,    -26,     57,     90,     43,    -43,    -90,    -57,     26,     87,     70,     -9,    -80,    -80,     -9,     70,     87,     26,    -57,    -90,    -43,     43,     90,     57,    -26,    -87,    -70,      9,     80 },
      {     79,      6,    -73,    -84,    -17,     66,     88,     27,    -58,    -90,    -38,     49,     91,     47,    -40,    -90,    -57,     29,     88,     65,    -19,    -85,    -72,      8,     80,     78,      3,    -75,    -83,    -14,     68,     87,     25,    -60,    -89,    -36,     51,     90,     46,    -42,    -90,    -55,     32,     89,     63,    -21,    -86,    -71,     10,     81,     77,      1,    -76,    -82,    -12,     69,     86,     23,    -62,    -89,    -34,     53,     90,     44,    -44,    -90,    -53,     34,     89,     62,    -23,    -86,    -69,     12,     82,     76,     -1,    -77,    -81,    -10,     71,     86,     21,    -63,    -89,    -32,     55,     90,     42,    -46,    -90,    -51,     36,     89,     60,    -25,    -87,    -68,     14,     83,     75,     -3,    -78,    -80,     -8,     72,     85,     19,    -65,    -88,    -29,     57,     90,     40,    -47,    -91,    -49,     38,     90,     58,    -27,    -88,    -66,     17,     84,     73,     -6,    -79 },
      {     79,      2,    -76,    -81,     -7,     74,     83,     11,    -71,    -84,    -15,     69,     86,     20,    -66,    -87,    -24,     62,     88,     28,    -59,    -89,    -33,     56,     90,     37,    -52,    -90,    -41,     48,     90,     45,    -45,    -90,    -48,     41,     90,     52,    -37,    -90,    -56,     33,     89,     59,    -28,    -88,    -62,     24,     87,     66,    -20,    -86,    -69,     15,     84,     71,    -11,    -83,    -74,      7,     81,     76,     -2,    -79,    -79,     -2,     76,     81,      7,    -74,    -83,    -11,     71,     84,     15,    -69,    -86,    -20,     66,     87,     24,    -62,    -88,    -28,     59,     89,     33,    -56,    -90,    -37,     52,     90,     41,    -48,    -90,    -45,     45,     90,     48,    -41,    -90,    -52,     37,     90,     56,    -33,    -89,    -59,     28,     88,     62,    -24,    -87,    -66,     20,     86,     69,    -15,    -84,    -71,     11,     83,     74,     -7,    -81,    -76,      2,     79 },
      {     78,     -1,    -79,    -77,      3,     80,     76,     -6,    -81,    -75,      8,     82,     73,    -10,    -83,    -72,     12,     84,     71,    -14,    -85,    -69,     17,     86,     68,    -19,    -86,    -66,     21,     87,     65,    -23,    -88,    -63,     25,     88,     62,    -27,    -89,    -60,     29,     89,     58,    -32,    -89,    -57,     34,     90,     55,    -36,    -90,    -53,     38,     90,     51,    -40,    -90,    -49,     42,     90,     47,    -44,    -91,    -46,     46,     91,     44,    -47,    -90,    -42,     49,     90,     40,    -51,    -90,    -38,     53,     90,     36,    -55,    -90,    -34,     57,     89,     32,    -58,    -89,    -29,     60,     89,     27,    -62,    -88,    -25,     63,     88,     23,    -65,    -87,    -21,     66,     86,     19,    -68,    -86,    -17,     69,     85,     14,    -71,    -84,    -12,     72,     83,     10,    -73,    -82,     -8,     75,     81,      6,    -76,    -80,     -3,     77,     79,      1,    -78 },
      {     78,     -4,    -82,    -73,     13,     85,     67,    -22,    -88,    -61,     30,     90,     54,    -39,    -90,    -47,     47,     90,     39,    -54,    -90,    -30,     61,     88,     22,    -67,    -85,    -13,     73,     82,      4,    -78,    -78,      4,     82,     73,    -13,    -85,    -67,     22,     88,     61,    -30,    -90,    -54,     39,     90,     47,    -47,    -90,    -39,     54,     90,     30,    -61,    -88,    -22,     67,     85,     13,    -73,    -82,     -4,     78,     78,     -4,    -82,    -73,     13,     85,     67,    -22,    -88,    -61,     30,     90,     54,    -39,    -90,    -47,     47,     90,     39,    -54,    -90,    -30,     61,     88,     22,    -67,    -85,    -13,     73,     82,      4,    -78,    -78,      4,     82,     73,    -13,    -85,    -67,     22,     88,     61,    -30,    -90,    -54,     39,     90,     47,    -47,    -90,    -39,     54,     90,     30,    -61,    -88,    -22,     67,     85,     13,    -73,    -82,     -4,     78 },
      {     77,     -8,    -84,    -68,     23,     89,     57,    -38,    -90,    -44,     51,     90,     29,    -63,    -86,    -14,     73,     80,     -1,    -81,    -72,     17,     87,     62,    -32,    -90,    -49,     46,     90,     36,    -58,    -88,    -21,     69,     83,      6,    -78,    -76,     10,     85,     66,    -25,    -89,    -55,     40,     91,     42,    -53,    -89,    -27,     65,     86,     12,    -75,    -79,      3,     82,     71,    -19,    -88,    -60,     34,     90,     47,    -47,    -90,    -34,     60,     88,     19,    -71,    -82,     -3,     79,     75,    -12,    -86,    -65,     27,     89,     53,    -42,    -91,    -40,     55,     89,     25,    -66,    -85,    -10,     76,     78,     -6,    -83,    -69,     21,     88,     58,    -36,    -90,    -46,     49,     90,     32,    -62,    -87,    -17,     72,     81,      1,    -80,    -73,     14,     86,     63,    -29,    -90,    -51,     44,     90,     38,    -57,    -89,    -23,     68,     84,      8,    -77 },
      {     76,    -11,    -86,    -62,     33,     90,     45,    -52,    -89,    -24,     69,     83,      2,    -81,    -71,     20,     88,     56,    -41,    -90,    -37,     59,     87,     15,    -74,    -79,      7,     84,     66,    -28,    -90,    -48,     48,     90,     28,    -66,    -84,     -7,     79,     74,    -15,    -87,    -59,     37,     90,     41,    -56,    -88,    -20,     71,     81,     -2,    -83,    -69,     24,     89,     52,    -45,    -90,    -33,     62,     86,     11,    -76,    -76,     11,     86,     62,    -33,    -90,    -45,     52,     89,     24,    -69,    -83,     -2,     81,     71,    -20,    -88,    -56,     41,     90,     37,    -59,    -87,    -15,     74,     79,     -7,    -84,    -66,     28,     90,     48,    -48,    -90,    -28,     66,     84,      7,    -79,    -74,     15,     87,     59,    -37,    -90,    -41,     56,     88,     20,    -71,    -81,      2,     83,     69,    -24,    -89,    -52,     45,     90,     33,    -62,    -86,    -11,     76 },
      {     76,    -14,    -88,    -57,     42,     90,     32,    -65,    -84,     -3,     81,     69,    -25,    -90,    -47,     51,     89,     21,    -72,    -79,      8,     86,     62,    -36,    -91,    -38,     60,     86,     10,    -78,    -73,     19,     89,     53,    -46,    -90,    -27,     68,     82,     -1,    -83,    -66,     29,     90,     44,    -55,    -88,    -17,     75,     77,    -12,    -87,    -58,     40,     90,     34,    -63,    -85,     -6,     80,     71,    -23,    -89,    -49,     49,     89,     23,    -71,    -80,      6,     85,     63,    -34,    -90,    -40,     58,     87,     12,    -77,    -75,     17,     88,     55,    -44,    -90,    -29,     66,     83,      1,    -82,    -68,     27,     90,     46,    -53,    -89,    -19,     73,     78,    -10,    -86,    -60,     38,     91,     36,    -62,    -86,     -8,     79,     72,    -21,    -89,    -51,     47,     90,     25,    -69,    -81,      3,     84,     65,    -32,    -90,    -42,     57,     88,     14,    -76 },
      {     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75 },
      {     75,    -21,    -90,    -44,     58,     86,      3,    -83,    -63,     38,     90,     27,    -71,    -78,     14,     89,     49,    -53,    -88,    -10,     80,     68,    -32,    -91,    -34,     66,     81,     -8,    -87,    -55,     47,     89,     17,    -77,    -72,     25,     90,     40,    -62,    -84,      1,     85,     60,    -42,    -90,    -23,     73,     76,    -19,    -89,    -46,     57,     86,      6,    -82,    -65,     36,     90,     29,    -69,    -79,     12,     88,     51,    -51,    -88,    -12,     79,     69,    -29,    -90,    -36,     65,     82,     -6,    -86,    -57,     46,     89,     19,    -76,    -73,     23,     90,     42,    -60,    -85,     -1,     84,     62,    -40,    -90,    -25,     72,     77,    -17,    -89,    -47,     55,     87,      8,    -81,    -66,     34,     91,     32,    -68,    -80,     10,     88,     53,    -49,    -89,    -14,     78,     71,    -27,    -90,    -38,     63,     83,     -3,    -86,    -58,     44,     90,     21,    -75 },
      {     74,    -24,    -90,    -37,     66,     81,    -11,    -88,    -48,     56,     86,      2,    -84,    -59,     45,     89,     15,    -79,    -69,     33,     90,     28,    -71,    -76,     20,     90,     41,    -62,    -83,      7,     87,     52,    -52,    -87,     -7,     83,     62,    -41,    -90,    -20,     76,     71,    -28,    -90,    -33,     69,     79,    -15,    -89,    -45,     59,     84,     -2,    -86,    -56,     48,     88,     11,    -81,    -66,     37,     90,     24,    -74,    -74,     24,     90,     37,    -66,    -81,     11,     88,     48,    -56,    -86,     -2,     84,     59,    -45,    -89,    -15,     79,     69,    -33,    -90,    -28,     71,     76,    -20,    -90,    -41,     62,     83,     -7,    -87,    -52,     52,     87,      7,    -83,    -62,     41,     90,     20,    -76,    -71,     28,     90,     33,    -69,    -79,     15,     89,     45,    -59,    -84,      2,     86,     56,    -48,    -88,    -11,     81,     66,    -37,    -90,    -24,     74 },
      {     73,    -27,    -91,    -29,     72,     75,    -25,    -90,    -32,     71,     76,    -23,    -90,    -34,     69,     77,    -21,    -90,    -36,     68,     78,    -19,    -90,    -38,     66,     79,    -17,    -90,    -40,     65,     80,    -14,    -89,    -42,     63,     81,    -12,    -89,    -44,     62,     82,    -10,    -89,    -46,     60,     83,     -8,    -88,    -47,     58,     84,     -6,    -88,    -49,     57,     85,     -3,    -87,    -51,     55,     86,     -1,    -86,    -53,     53,     86,      1,    -86,    -55,     51,     87,      3,    -85,    -57,     49,     88,      6,    -84,    -58,     47,     88,      8,    -83,    -60,     46,     89,     10,    -82,    -62,     44,     89,     12,    -81,    -63,     42,     89,     14,    -80,    -65,     40,     90,     17,    -79,    -66,     38,     90,     19,    -78,    -68,     36,     90,     21,    -77,    -69,     34,     90,     23,    -76,    -71,     32,     90,     25,    -75,    -72,     29,     91,     27,    -73 },
      {     73,    -30,    -90,    -22,     78,     67,    -39,    -90,    -13,     82,     61,    -47,    -88,     -4,     85,     54,    -54,    -85,      4,     88,     47,    -61,    -82,     13,     90,     39,    -67,    -78,     22,     90,     30,    -73,    -73,     30,     90,     22,    -78,    -67,     39,     90,     13,    -82,    -61,     47,     88,      4,    -85,    -54,     54,     85,     -4,    -88,    -47,     61,     82,    -13,    -90,    -39,     67,     78,    -22,    -90,    -30,     73,     73,    -30,    -90,    -22,     78,     67,    -39,    -90,    -13,     82,     61,    -47,    -88,     -4,     85,     54,    -54,    -85,      4,     88,     47,    -61,    -82,     13,     90,     39,    -67,    -78,     22,     90,     30,    -73,    -73,     30,     90,     22,    -78,    -67,     39,     90,     13,    -82,    -61,     47,     88,      4,    -85,    -54,     54,     85,     -4,    -88,    -47,     61,     82,    -13,    -90,    -39,     67,     78,    -22,    -90,    -30,     73 },
      {     72,    -34,    -90,    -14,     82,     58,    -51,    -86,      6,     89,     42,    -66,    -77,     25,     91,     23,    -78,    -65,     44,     88,      3,    -86,    -49,     60,     81,    -17,    -90,    -32,     73,     71,    -36,    -90,    -12,     83,     57,    -53,    -85,      8,     89,     40,    -68,    -76,     27,     90,     21,    -79,    -63,     46,     88,      1,    -87,    -47,     62,     80,    -19,    -90,    -29,     75,     69,    -38,    -89,    -10,     84,     55,    -55,    -84,     10,     89,     38,    -69,    -75,     29,     90,     19,    -80,    -62,     47,     87,     -1,    -88,    -46,     63,     79,    -21,    -90,    -27,     76,     68,    -40,    -89,     -8,     85,     53,    -57,    -83,     12,     90,     36,    -71,    -73,     32,     90,     17,    -81,    -60,     49,     86,     -3,    -88,    -44,     65,     78,    -23,    -91,    -25,     77,     66,    -42,    -89,     -6,     86,     51,    -58,    -82,     14,     90,     34,    -72 },
      {     71,    -37,    -89,     -7,     86,     48,    -62,    -79,     24,     90,     20,    -81,    -59,     52,     84,    -11,    -90,    -33,     74,     69,    -41,    -88,     -2,     87,     45,    -66,    -76,     28,     90,     15,    -83,    -56,     56,     83,    -15,    -90,    -28,     76,     66,    -45,    -87,      2,     88,     41,    -69,    -74,     33,     90,     11,    -84,    -52,     59,     81,    -20,    -90,    -24,     79,     62,    -48,    -86,      7,     89,     37,    -71,    -71,     37,     89,      7,    -86,    -48,     62,     79,    -24,    -90,    -20,     81,     59,    -52,    -84,     11,     90,     33,    -74,    -69,     41,     88,      2,    -87,    -45,     66,     76,    -28,    -90,    -15,     83,     56,    -56,    -83,     15,     90,     28,    -76,    -66,     45,     87,     -2,    -88,    -41,     69,     74,    -33,    -90,    -11,     84,     52,    -59,    -81,     20,     90,     24,    -79,    -62,     48,     86,     -7,    -89,    -37,     71 },
      {     71,    -40,    -88,      1,     89,     38,    -72,    -69,     42,     88,     -3,    -89,    -36,     73,     68,    -44,    -87,      6,     89,     34,    -75,    -66,     46,     86,     -8,    -90,    -32,     76,     65,    -47,    -86,     10,     90,     29,    -77,    -63,     49,     85,    -12,    -90,    -27,     78,     62,    -51,    -84,     14,     90,     25,    -79,    -60,     53,     83,    -17,    -90,    -23,     80,     58,    -55,    -82,     19,     91,     21,    -81,    -57,     57,     81,    -21,    -91,    -19,     82,     55,    -58,    -80,     23,     90,     17,    -83,    -53,     60,     79,    -25,    -90,    -14,     84,     51,    -62,    -78,     27,     90,     12,    -85,    -49,     63,     77,    -29,    -90,    -10,     86,     47,    -65,    -76,     32,     90,      8,    -86,    -46,     66,     75,    -34,    -89,     -6,     87,     44,    -68,    -73,     36,     89,      3,    -88,    -42,     69,     72,    -38,    -89,     -1,     88,     40,    -71 },
      {     70,    -43,    -87,      9,     90,     26,    -80,    -57,     57,     80,    -26,    -90,     -9,     87,     43,    -70,    -70,     43,     87,     -9,    -90,    -26,     80,     57,    -57,    -80,     26,     90,      9,    -87,    -43,     70,     70,    -43,    -87,      9,     90,     26,    -80,    -57,     57,     80,    -26,    -90,     -9,     87,     43,    -70,    -70,     43,     87,     -9,    -90,    -26,     80,     57,    -57,    -80,     26,     90,      9,    -87,    -43,     70,     70,    -43,    -87,      9,     90,     26,    -80,    -57,     57,     80,    -26,    -90,     -9,     87,     43,    -70,    -70,     43,     87,     -9,    -90,    -26,     80,     57,    -57,    -80,     26,     90,      9,    -87,    -43,     70,     70,    -43,    -87,      9,     90,     26,    -80,    -57,     57,     80,    -26,    -90,     -9,     87,     43,    -70,    -70,     43,     87,     -9,    -90,    -26,     80,     57,    -57,    -80,     26,     90,      9,    -87,    -43,     70 },
      {     69,    -46,    -85,     17,     91,     14,    -86,    -44,     71,     68,    -47,    -84,     19,     90,     12,    -86,    -42,     72,     66,    -49,    -83,     21,     90,     10,    -87,    -40,     73,     65,    -51,    -82,     23,     90,      8,    -88,    -38,     75,     63,    -53,    -81,     25,     90,      6,    -88,    -36,     76,     62,    -55,    -80,     27,     90,      3,    -89,    -34,     77,     60,    -57,    -79,     29,     89,      1,    -89,    -32,     78,     58,    -58,    -78,     32,     89,     -1,    -89,    -29,     79,     57,    -60,    -77,     34,     89,     -3,    -90,    -27,     80,     55,    -62,    -76,     36,     88,     -6,    -90,    -25,     81,     53,    -63,    -75,     38,     88,     -8,    -90,    -23,     82,     51,    -65,    -73,     40,     87,    -10,    -90,    -21,     83,     49,    -66,    -72,     42,     86,    -12,    -90,    -19,     84,     47,    -68,    -71,     44,     86,    -14,    -91,    -17,     85,     46,    -69 },
      {     69,    -48,    -83,     24,     90,      2,    -89,    -28,     81,     52,    -66,    -71,     45,     84,    -20,    -90,     -7,     88,     33,    -79,    -56,     62,     74,    -41,    -86,     15,     90,     11,    -87,    -37,     76,     59,    -59,    -76,     37,     87,    -11,    -90,    -15,     86,     41,    -74,    -62,     56,     79,    -33,    -88,      7,     90,     20,    -84,    -45,     71,     66,    -52,    -81,     28,     89,     -2,    -90,    -24,     83,     48,    -69,    -69,     48,     83,    -24,    -90,     -2,     89,     28,    -81,    -52,     66,     71,    -45,    -84,     20,     90,      7,    -88,    -33,     79,     56,    -62,    -74,     41,     86,    -15,    -90,    -11,     87,     37,    -76,    -59,     59,     76,    -37,    -87,     11,     90,     15,    -86,    -41,     74,     62,    -56,    -79,     33,     88,     -7,    -90,    -20,     84,     45,    -71,    -66,     52,     81,    -28,    -89,      2,     90,     24,    -83,    -48,     69 },
      {     68,    -51,    -80,     32,     88,    -10,    -91,    -12,     88,     34,    -79,    -53,     66,     69,    -49,    -81,     29,     89,     -8,    -90,    -14,     87,     36,    -78,    -55,     65,     71,    -47,    -82,     27,     89,     -6,    -90,    -17,     86,     38,    -77,    -57,     63,     72,    -46,    -83,     25,     89,     -3,    -90,    -19,     86,     40,    -76,    -58,     62,     73,    -44,    -84,     23,     90,     -1,    -90,    -21,     85,     42,    -75,    -60,     60,     75,    -42,    -85,     21,     90,      1,    -90,    -23,     84,     44,    -73,    -62,     58,     76,    -40,    -86,     19,     90,      3,    -89,    -25,     83,     46,    -72,    -63,     57,     77,    -38,    -86,     17,     90,      6,    -89,    -27,     82,     47,    -71,    -65,     55,     78,    -36,    -87,     14,     90,      8,    -89,    -29,     81,     49,    -69,    -66,     53,     79,    -34,    -88,     12,     91,     10,    -88,    -32,     80,     51,    -68 },
      {     67,    -54,    -78,     39,     85,    -22,    -90,      4,     90,     13,    -88,    -30,     82,     47,    -73,    -61,     61,     73,    -47,    -82,     30,     88,    -13,    -90,     -4,     90,     22,    -85,    -39,     78,     54,    -67,    -67,     54,     78,    -39,    -85,     22,     90,     -4,    -90,    -13,     88,     30,    -82,    -47,     73,     61,    -61,    -73,     47,     82,    -30,    -88,     13,     90,      4,    -90,    -22,     85,     39,    -78,    -54,     67,     67,    -54,    -78,     39,     85,    -22,    -90,      4,     90,     13,    -88,    -30,     82,     47,    -73,    -61,     61,     73,    -47,    -82,     30,     88,    -13,    -90,     -4,     90,     22,    -85,    -39,     78,     54,    -67,    -67,     54,     78,    -39,    -85,     22,     90,     -4,    -90,    -13,     88,     30,    -82,    -47,     73,     61,    -61,    -73,     47,     82,    -30,    -88,     13,     90,      4,    -90,    -22,     85,     39,    -78,    -54,     67 },
      {     66,    -57,    -75,     46,     81,    -34,    -86,     21,     89,     -8,    -91,     -6,     90,     19,    -87,    -32,     82,     44,    -76,    -55,     68,     65,    -58,    -73,     47,     80,    -36,    -86,     23,     89,    -10,    -90,     -3,     90,     17,    -88,    -29,     83,     42,    -77,    -53,     69,     63,    -60,    -72,     49,     79,    -38,    -85,     25,     89,    -12,    -90,     -1,     90,     14,    -88,    -27,     84,     40,    -78,    -51,     71,     62,    -62,    -71,     51,     78,    -40,    -84,     27,     88,    -14,    -90,      1,     90,     12,    -89,    -25,     85,     38,    -79,    -49,     72,     60,    -63,    -69,     53,     77,    -42,    -83,     29,     88,    -17,    -90,      3,     90,     10,    -89,    -23,     86,     36,    -80,    -47,     73,     58,    -65,    -68,     55,     76,    -44,    -82,     32,     87,    -19,    -90,      6,     91,      8,    -89,    -21,     86,     34,    -81,    -46,     75,     57,    -66 },
      {     66,    -59,    -71,     52,     76,    -45,    -81,     37,     84,    -28,    -87,     20,     89,    -11,    -90,      2,     90,      7,    -90,    -15,     88,     24,    -86,    -33,     83,     41,    -79,    -48,     74,     56,    -69,    -62,     62,     69,    -56,    -74,     48,     79,    -41,    -83,     33,     86,    -24,    -88,     15,     90,     -7,    -90,     -2,     90,     11,    -89,    -20,     87,     28,    -84,    -37,     81,     45,    -76,    -52,     71,     59,    -66,    -66,     59,     71,    -52,    -76,     45,     81,    -37,    -84,     28,     87,    -20,    -89,     11,     90,     -2,    -90,     -7,     90,     15,    -88,    -24,     86,     33,    -83,    -41,     79,     48,    -74,    -56,     69,     62,    -62,    -69,     56,     74,    -48,    -79,     41,     83,    -33,    -86,     24,     88,    -15,    -90,      7,     90,      2,    -90,    -11,     89,     20,    -87,    -28,     84,     37,    -81,    -45,     76,     52,    -71,    -59,     66 },
      {     65,    -62,    -68,     58,     71,    -55,    -73,     51,     76,    -47,    -78,     44,     80,    -40,    -82,     36,     84,    -32,    -86,     27,     87,    -23,    -88,     19,     89,    -14,    -90,     10,     90,     -6,    -90,      1,     91,      3,    -90,     -8,     90,     12,    -89,    -17,     89,     21,    -88,    -25,     86,     29,    -85,    -34,     83,     38,    -81,    -42,     79,     46,    -77,    -49,     75,     53,    -72,    -57,     69,     60,    -66,    -63,     63,     66,    -60,    -69,     57,     72,    -53,    -75,     49,     77,    -46,    -79,     42,     81,    -38,    -83,     34,     85,    -29,    -86,     25,     88,    -21,    -89,     17,     89,    -12,    -90,      8,     90,     -3,    -91,     -1,     90,      6,    -90,    -10,     90,     14,    -89,    -19,     88,     23,    -87,    -27,     86,     32,    -84,    -36,     82,     40,    -80,    -44,     78,     47,    -76,    -51,     73,     55,    -71,    -58,     68,     62,    -65 },
      {     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64 },
      {     63,    -66,    -60,     69,     57,    -72,    -53,     75,     49,    -77,    -46,     79,     42,    -81,    -38,     83,     34,    -85,    -29,     86,     25,    -88,    -21,     89,     17,    -89,    -12,     90,      8,    -90,     -3,     91,     -1,    -90,      6,     90,    -10,    -90,     14,     89,    -19,    -88,     23,     87,    -27,    -86,     32,     84,    -36,    -82,     40,     80,    -44,    -78,     47,     76,    -51,    -73,     55,     71,    -58,    -68,     62,     65,    -65,    -62,     68,     58,    -71,    -55,     73,     51,    -76,    -47,     78,     44,    -80,    -40,     82,     36,    -84,    -32,     86,     27,    -87,    -23,     88,     19,    -89,    -14,     90,     10,    -90,     -6,     90,      1,    -91,      3,     90,     -8,    -90,     12,     89,    -17,    -89,     21,     88,    -25,    -86,     29,     85,    -34,    -83,     38,     81,    -42,    -79,     46,     77,    -49,    -75,     53,     72,    -57,    -69,     60,     66,    -63 },
      {     62,    -69,    -56,     74,     48,    -79,    -41,     83,     33,    -86,    -24,     88,     15,    -90,     -7,     90,     -2,    -90,     11,     89,    -20,    -87,     28,     84,    -37,    -81,     45,     76,    -52,    -71,     59,     66,    -66,    -59,     71,     52,    -76,    -45,     81,     37,    -84,    -28,     87,     20,    -89,    -11,     90,      2,    -90,      7,     90,    -15,    -88,     24,     86,    -33,    -83,     41,     79,    -48,    -74,     56,     69,    -62,    -62,     69,     56,    -74,    -48,     79,     41,    -83,    -33,     86,     24,    -88,    -15,     90,      7,    -90,      2,     90,    -11,    -89,     20,     87,    -28,    -84,     37,     81,    -45,    -76,     52,     71,    -59,    -66,     66,     59,    -71,    -52,     76,     45,    -81,    -37,     84,     28,    -87,    -20,     89,     11,    -90,     -2,     90,     -7,    -90,     15,     88,    -24,    -86,     33,     83,    -41,    -79,     48,     74,    -56,    -69,     62 },
      {     62,    -71,    -51,     78,     40,    -84,    -27,     88,     14,    -90,     -1,     90,    -12,    -89,     25,     85,    -38,    -79,     49,     72,    -60,    -63,     69,     53,    -77,    -42,     83,     29,    -88,    -17,     90,      3,    -90,     10,     89,    -23,    -86,     36,     80,    -47,    -73,     58,     65,    -68,    -55,     76,     44,    -82,    -32,     87,     19,    -90,     -6,     91,     -8,    -89,     21,     86,    -34,    -81,     46,     75,    -57,    -66,     66,     57,    -75,    -46,     81,     34,    -86,    -21,     89,      8,    -91,      6,     90,    -19,    -87,     32,     82,    -44,    -76,     55,     68,    -65,    -58,     73,     47,    -80,    -36,     86,     23,    -89,    -10,     90,     -3,    -90,     17,     88,    -29,    -83,     42,     77,    -53,    -69,     63,     60,    -72,    -49,     79,     38,    -85,    -25,     89,     12,    -90,      1,     90,    -14,    -88,     27,     84,    -40,    -78,     51,     71,    -62 },
      {     61,    -73,    -47,     82,     30,    -88,    -13,     90,     -4,    -90,     22,     85,    -39,    -78,     54,     67,    -67,    -54,     78,     39,    -85,    -22,     90,      4,    -90,     13,     88,    -30,    -82,     47,     73,    -61,    -61,     73,     47,    -82,    -30,     88,     13,    -90,      4,     90,    -22,    -85,     39,     78,    -54,    -67,     67,     54,    -78,    -39,     85,     22,    -90,     -4,     90,    -13,    -88,     30,     82,    -47,    -73,     61,     61,    -73,    -47,     82,     30,    -88,    -13,     90,     -4,    -90,     22,     85,    -39,    -78,     54,     67,    -67,    -54,     78,     39,    -85,    -22,     90,      4,    -90,     13,     88,    -30,    -82,     47,     73,    -61,    -61,     73,     47,    -82,    -30,     88,     13,    -90,      4,     90,    -22,    -85,     39,     78,    -54,    -67,     67,     54,    -78,    -39,     85,     22,    -90,     -4,     90,    -13,    -88,     30,     82,    -47,    -73,     61 },
      {     60,    -75,    -42,     85,     21,    -90,      1,     90,    -23,    -84,     44,     73,    -62,    -58,     76,     40,    -86,    -19,     90,     -3,    -89,     25,     83,    -46,    -72,     63,     57,    -77,    -38,     86,     17,    -90,      6,     89,    -27,    -82,     47,     71,    -65,    -55,     78,     36,    -87,    -14,     90,     -8,    -89,     29,     81,    -49,    -69,     66,     53,    -79,    -34,     88,     12,    -91,     10,     88,    -32,    -80,     51,     68,    -68,    -51,     80,     32,    -88,    -10,     91,    -12,    -88,     34,     79,    -53,    -66,     69,     49,    -81,    -29,     89,      8,    -90,     14,     87,    -36,    -78,     55,     65,    -71,    -47,     82,     27,    -89,     -6,     90,    -17,    -86,     38,     77,    -57,    -63,     72,     46,    -83,    -25,     89,      3,    -90,     19,     86,    -40,    -76,     58,     62,    -73,    -44,     84,     23,    -90,     -1,     90,    -21,    -85,     42,     75,    -60 },
      {     59,    -76,    -37,     87,     11,    -90,     15,     86,    -41,    -74,     62,     56,    -79,    -33,     88,      7,    -90,     20,     84,    -45,    -71,     66,     52,    -81,    -28,     89,      2,    -90,     24,     83,    -48,    -69,     69,     48,    -83,    -24,     90,     -2,    -89,     28,     81,    -52,    -66,     71,     45,    -84,    -20,     90,     -7,    -88,     33,     79,    -56,    -62,     74,     41,    -86,    -15,     90,    -11,    -87,     37,     76,    -59,    -59,     76,     37,    -87,    -11,     90,    -15,    -86,     41,     74,    -62,    -56,     79,     33,    -88,     -7,     90,    -20,    -84,     45,     71,    -66,    -52,     81,     28,    -89,     -2,     90,    -24,    -83,     48,     69,    -69,    -48,     83,     24,    -90,      2,     89,    -28,    -81,     52,     66,    -71,    -45,     84,     20,    -90,      7,     88,    -33,    -79,     56,     62,    -74,    -41,     86,     15,    -90,     11,     87,    -37,    -76,     59 },
      {     58,    -78,    -32,     89,      1,    -89,     29,     79,    -57,    -60,     77,     34,    -89,     -3,     90,    -27,    -80,     55,     62,    -76,    -36,     88,      6,    -90,     25,     81,    -53,    -63,     75,     38,    -88,     -8,     90,    -23,    -82,     51,     65,    -73,    -40,     87,     10,    -90,     21,     83,    -49,    -66,     72,     42,    -86,    -12,     90,    -19,    -84,     47,     68,    -71,    -44,     86,     14,    -91,     17,     85,    -46,    -69,     69,     46,    -85,    -17,     91,    -14,    -86,     44,     71,    -68,    -47,     84,     19,    -90,     12,     86,    -42,    -72,     66,     49,    -83,    -21,     90,    -10,    -87,     40,     73,    -65,    -51,     82,     23,    -90,      8,     88,    -38,    -75,     63,     53,    -81,    -25,     90,     -6,    -88,     36,     76,    -62,    -55,     80,     27,    -90,      3,     89,    -34,    -77,     60,     57,    -79,    -29,     89,     -1,    -89,     32,     78,    -58 },
      {     57,    -80,    -26,     90,     -9,    -87,     43,     70,    -70,    -43,     87,      9,    -90,     26,     80,    -57,    -57,     80,     26,    -90,      9,     87,    -43,    -70,     70,     43,    -87,     -9,     90,    -26,    -80,     57,     57,    -80,    -26,     90,     -9,    -87,     43,     70,    -70,    -43,     87,      9,    -90,     26,     80,    -57,    -57,     80,     26,    -90,      9,     87,    -43,    -70,     70,     43,    -87,     -9,     90,    -26,    -80,     57,     57,    -80,    -26,     90,     -9,    -87,     43,     70,    -70,    -43,     87,      9,    -90,     26,     80,    -57,    -57,     80,     26,    -90,      9,     87,    -43,    -70,     70,     43,    -87,     -9,     90,    -26,    -80,     57,     57,    -80,    -26,     90,     -9,    -87,     43,     70,    -70,    -43,     87,      9,    -90,     26,     80,    -57,    -57,     80,     26,    -90,      9,     87,    -43,    -70,     70,     43,    -87,     -9,     90,    -26,    -80,     57 },
      {     57,    -81,    -21,     91,    -19,    -82,     55,     58,    -80,    -23,     90,    -17,    -83,     53,     60,    -79,    -25,     90,    -14,    -84,     51,     62,    -78,    -27,     90,    -12,    -85,     49,     63,    -77,    -29,     90,    -10,    -86,     47,     65,    -76,    -32,     90,     -8,    -86,     46,     66,    -75,    -34,     89,     -6,    -87,     44,     68,    -73,    -36,     89,     -3,    -88,     42,     69,    -72,    -38,     89,     -1,    -88,     40,     71,    -71,    -40,     88,      1,    -89,     38,     72,    -69,    -42,     88,      3,    -89,     36,     73,    -68,    -44,     87,      6,    -89,     34,     75,    -66,    -46,     86,      8,    -90,     32,     76,    -65,    -47,     86,     10,    -90,     29,     77,    -63,    -49,     85,     12,    -90,     27,     78,    -62,    -51,     84,     14,    -90,     25,     79,    -60,    -53,     83,     17,    -90,     23,     80,    -58,    -55,     82,     19,    -91,     21,     81,    -57 },
      {     56,    -83,    -15,     90,    -28,    -76,     66,     45,    -87,     -2,     88,    -41,    -69,     74,     33,    -90,     11,     84,    -52,    -59,     81,     20,    -90,     24,     79,    -62,    -48,     86,      7,    -89,     37,     71,    -71,    -37,     89,     -7,    -86,     48,     62,    -79,    -24,     90,    -20,    -81,     59,     52,    -84,    -11,     90,    -33,    -74,     69,     41,    -88,      2,     87,    -45,    -66,     76,     28,    -90,     15,     83,    -56,    -56,     83,     15,    -90,     28,     76,    -66,    -45,     87,      2,    -88,     41,     69,    -74,    -33,     90,    -11,    -84,     52,     59,    -81,    -20,     90,    -24,    -79,     62,     48,    -86,     -7,     89,    -37,    -71,     71,     37,    -89,      7,     86,    -48,    -62,     79,     24,    -90,     20,     81,    -59,    -52,     84,     11,    -90,     33,     74,    -69,    -41,     88,     -2,    -87,     45,     66,    -76,    -28,     90,    -15,    -83,     56 },
      {     55,    -84,    -10,     89,    -38,    -69,     75,     29,    -90,     19,     80,    -62,    -47,     87,      1,    -88,     46,     63,    -79,    -21,     90,    -27,    -76,     68,     40,    -89,      8,     85,    -53,    -57,     83,     12,    -90,     36,     71,    -73,    -32,     90,    -17,    -81,     60,     49,    -86,     -3,     88,    -44,    -65,     78,     23,    -91,     25,     77,    -66,    -42,     89,     -6,    -86,     51,     58,    -82,    -14,     90,    -34,    -72,     72,     34,    -90,     14,     82,    -58,    -51,     86,      6,    -89,     42,     66,    -77,    -25,     91,    -23,    -78,     65,     44,    -88,      3,     86,    -49,    -60,     81,     17,    -90,     32,     73,    -71,    -36,     90,    -12,    -83,     57,     53,    -85,     -8,     89,    -40,    -68,     76,     27,    -90,     21,     79,    -63,    -46,     88,     -1,    -87,     47,     62,    -80,    -19,     90,    -29,    -75,     69,     38,    -89,     10,     84,    -55 },
      {     54,    -85,     -4,     88,    -47,    -61,     82,     13,    -90,     39,     67,    -78,    -22,     90,    -30,    -73,     73,     30,    -90,     22,     78,    -67,    -39,     90,    -13,    -82,     61,     47,    -88,      4,     85,    -54,    -54,     85,      4,    -88,     47,     61,    -82,    -13,     90,    -39,    -67,     78,     22,    -90,     30,     73,    -73,    -30,     90,    -22,    -78,     67,     39,    -90,     13,     82,    -61,    -47,     88,     -4,    -85,     54,     54,    -85,     -4,     88,    -47,    -61,     82,     13,    -90,     39,     67,    -78,    -22,     90,    -30,    -73,     73,     30,    -90,     22,     78,    -67,    -39,     90,    -13,    -82,     61,     47,    -88,      4,     85,    -54,    -54,     85,      4,    -88,     47,     61,    -82,    -13,     90,    -39,    -67,     78,     22,    -90,     30,     73,    -73,    -30,     90,    -22,    -78,     67,     39,    -90,     13,     82,    -61,    -47,     88,     -4,    -85,     54 },
      {     53,    -86,      1,     86,    -55,    -51,     87,     -3,    -85,     57,     49,    -88,      6,     84,    -58,    -47,     88,     -8,    -83,     60,     46,    -89,     10,     82,    -62,    -44,     89,    -12,    -81,     63,     42,    -89,     14,     80,    -65,    -40,     90,    -17,    -79,     66,     38,    -90,     19,     78,    -68,    -36,     90,    -21,    -77,     69,     34,    -90,     23,     76,    -71,    -32,     90,    -25,    -75,     72,     29,    -91,     27,     73,    -73,    -27,     91,    -29,    -72,     75,     25,    -90,     32,     71,    -76,    -23,     90,    -34,    -69,     77,     21,    -90,     36,     68,    -78,    -19,     90,    -38,    -66,     79,     17,    -90,     40,     65,    -80,    -14,     89,    -42,    -63,     81,     12,    -89,     44,     62,    -82,    -10,     89,    -46,    -60,     83,      8,    -88,     47,     58,    -84,     -6,     88,    -49,    -57,     85,      3,    -87,     51,     55,    -86,     -1,     86,    -53 },
      {     52,    -87,      7,     83,    -62,    -41,     90,    -20,    -76,     71,     28,    -90,     33,     69,    -79,    -15,     89,    -45,    -59,     84,      2,    -86,     56,     48,    -88,     11,     81,    -66,    -37,     90,    -24,    -74,     74,     24,    -90,     37,     66,    -81,    -11,     88,    -48,    -56,     86,     -2,    -84,     59,     45,    -89,     15,     79,    -69,    -33,     90,    -28,    -71,     76,     20,    -90,     41,     62,    -83,     -7,     87,    -52,    -52,     87,     -7,    -83,     62,     41,    -90,     20,     76,    -71,    -28,     90,    -33,    -69,     79,     15,    -89,     45,     59,    -84,     -2,     86,    -56,    -48,     88,    -11,    -81,     66,     37,    -90,     24,     74,    -74,    -24,     90,    -37,    -66,     81,     11,    -88,     48,     56,    -86,      2,     84,    -59,    -45,     89,    -15,    -79,     69,     33,    -90,     28,     71,    -76,    -20,     90,    -41,    -62,     83,      7,    -87,     52 },
      {     51,    -88,     12,     79,    -69,    -29,     90,    -36,    -65,     82,      6,    -86,     57,     46,    -89,     19,     76,    -73,    -23,     90,    -42,    -60,     85,     -1,    -84,     62,     40,    -90,     25,     72,    -77,    -17,     89,    -47,    -55,     87,     -8,    -81,     66,     34,    -91,     32,     68,    -80,    -10,     88,    -53,    -49,     89,    -14,    -78,     71,     27,    -90,     38,     63,    -83,     -3,     86,    -58,    -44,     90,    -21,    -75,     75,     21,    -90,     44,     58,    -86,      3,     83,    -63,    -38,     90,    -27,    -71,     78,     14,    -89,     49,     53,    -88,     10,     80,    -68,    -32,     91,    -34,    -66,     81,      8,    -87,     55,     47,    -89,     17,     77,    -72,    -25,     90,    -40,    -62,     84,      1,    -85,     60,     42,    -90,     23,     73,    -76,    -19,     89,    -46,    -57,     86,     -6,    -82,     65,     36,    -90,     29,     69,    -79,    -12,     88,    -51 },
      {     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50 },
      {     49,    -89,     23,     71,    -80,     -6,     85,    -63,    -34,     90,    -40,    -58,     87,    -12,    -77,     75,     17,    -88,     55,     44,    -90,     29,     66,    -83,      1,     82,    -68,    -27,     90,    -46,    -53,     89,    -19,    -73,     78,     10,    -86,     60,     38,    -91,     36,     62,    -86,      8,     79,    -72,    -21,     89,    -51,    -47,     90,    -25,    -69,     81,      3,    -84,     65,     32,    -90,     42,     57,    -88,     14,     76,    -76,    -14,     88,    -57,    -42,     90,    -32,    -65,     84,     -3,    -81,     69,     25,    -90,     47,     51,    -89,     21,     72,    -79,     -8,     86,    -62,    -36,     91,    -38,    -60,     86,    -10,    -78,     73,     19,    -89,     53,     46,    -90,     27,     68,    -82,     -1,     83,    -66,    -29,     90,    -44,    -55,     88,    -17,    -75,     77,     12,    -87,     58,     40,    -90,     34,     63,    -85,      6,     80,    -71,    -23,     89,    -49 },
      {     48,    -90,     28,     66,    -84,      7,     79,    -74,    -15,     87,    -59,    -37,     90,    -41,    -56,     88,    -20,    -71,     81,      2,    -83,     69,     24,    -89,     52,     45,    -90,     33,     62,    -86,     11,     76,    -76,    -11,     86,    -62,    -33,     90,    -45,    -52,     89,    -24,    -69,     83,     -2,    -81,     71,     20,    -88,     56,     41,    -90,     37,     59,    -87,     15,     74,    -79,     -7,     84,    -66,    -28,     90,    -48,    -48,     90,    -28,    -66,     84,     -7,    -79,     74,     15,    -87,     59,     37,    -90,     41,     56,    -88,     20,     71,    -81,     -2,     83,    -69,    -24,     89,    -52,    -45,     90,    -33,    -62,     86,    -11,    -76,     76,     11,    -86,     62,     33,    -90,     45,     52,    -89,     24,     69,    -83,      2,     81,    -71,    -20,     88,    -56,    -41,     90,    -37,    -59,     87,    -15,    -74,     79,      7,    -84,     66,     28,    -90,     48 },
      {     47,    -90,     34,     60,    -88,     19,     71,    -82,      3,     79,    -75,    -12,     86,    -65,    -27,     89,    -53,    -42,     91,    -40,    -55,     89,    -25,    -66,     85,    -10,    -76,     78,      6,    -83,     69,     21,    -88,     58,     36,    -90,     46,     49,    -90,     32,     62,    -87,     17,     72,    -81,      1,     80,    -73,    -14,     86,    -63,    -29,     90,    -51,    -44,     90,    -38,    -57,     89,    -23,    -68,     84,     -8,    -77,     77,      8,    -84,     68,     23,    -89,     57,     38,    -90,     44,     51,    -90,     29,     63,    -86,     14,     73,    -80,     -1,     81,    -72,    -17,     87,    -62,    -32,     90,    -49,    -46,     90,    -36,    -58,     88,    -21,    -69,     83,     -6,    -78,     76,     10,    -85,     66,     25,    -89,     55,     40,    -91,     42,     53,    -89,     27,     65,    -86,     12,     75,    -79,     -3,     82,    -71,    -19,     88,    -60,    -34,     90,    -47 },
      {     47,    -90,     39,     54,    -90,     30,     61,    -88,     22,     67,    -85,     13,     73,    -82,      4,     78,    -78,     -4,     82,    -73,    -13,     85,    -67,    -22,     88,    -61,    -30,     90,    -54,    -39,     90,    -47,    -47,     90,    -39,    -54,     90,    -30,    -61,     88,    -22,    -67,     85,    -13,    -73,     82,     -4,    -78,     78,      4,    -82,     73,     13,    -85,     67,     22,    -88,     61,     30,    -90,     54,     39,    -90,     47,     47,    -90,     39,     54,    -90,     30,     61,    -88,     22,     67,    -85,     13,     73,    -82,      4,     78,    -78,     -4,     82,    -73,    -13,     85,    -67,    -22,     88,    -61,    -30,     90,    -54,    -39,     90,    -47,    -47,     90,    -39,    -54,     90,    -30,    -61,     88,    -22,    -67,     85,    -13,    -73,     82,     -4,    -78,     78,      4,    -82,     73,     13,    -85,     67,     22,    -88,     61,     30,    -90,     54,     39,    -90,     47 },
      {     46,    -91,     44,     47,    -90,     42,     49,    -90,     40,     51,    -90,     38,     53,    -90,     36,     55,    -90,     34,     57,    -89,     32,     58,    -89,     29,     60,    -89,     27,     62,    -88,     25,     63,    -88,     23,     65,    -87,     21,     66,    -86,     19,     68,    -86,     17,     69,    -85,     14,     71,    -84,     12,     72,    -83,     10,     73,    -82,      8,     75,    -81,      6,     76,    -80,      3,     77,    -79,      1,     78,    -78,     -1,     79,    -77,     -3,     80,    -76,     -6,     81,    -75,     -8,     82,    -73,    -10,     83,    -72,    -12,     84,    -71,    -14,     85,    -69,    -17,     86,    -68,    -19,     86,    -66,    -21,     87,    -65,    -23,     88,    -63,    -25,     88,    -62,    -27,     89,    -60,    -29,     89,    -58,    -32,     89,    -57,    -34,     90,    -55,    -36,     90,    -53,    -38,     90,    -51,    -40,     90,    -49,    -42,     90,    -47,    -44,     91,    -46 },
      {     45,    -90,     48,     41,    -90,     52,     37,    -90,     56,     33,    -89,     59,     28,    -88,     62,     24,    -87,     66,     20,    -86,     69,     15,    -84,     71,     11,    -83,     74,      7,    -81,     76,      2,    -79,     79,     -2,    -76,     81,     -7,    -74,     83,    -11,    -71,     84,    -15,    -69,     86,    -20,    -66,     87,    -24,    -62,     88,    -28,    -59,     89,    -33,    -56,     90,    -37,    -52,     90,    -41,    -48,     90,    -45,    -45,     90,    -48,    -41,     90,    -52,    -37,     90,    -56,    -33,     89,    -59,    -28,     88,    -62,    -24,     87,    -66,    -20,     86,    -69,    -15,     84,    -71,    -11,     83,    -74,     -7,     81,    -76,     -2,     79,    -79,      2,     76,    -81,      7,     74,    -83,     11,     71,    -84,     15,     69,    -86,     20,     66,    -87,     24,     62,    -88,     28,     59,    -89,     33,     56,    -90,     37,     52,    -90,     41,     48,    -90,     45 },
      {     44,    -90,     53,     34,    -89,     62,     23,    -86,     69,     12,    -82,     76,      1,    -77,     81,    -10,    -71,     86,    -21,    -63,     89,    -32,    -55,     90,    -42,    -46,     90,    -51,    -36,     89,    -60,    -25,     87,    -68,    -14,     83,    -75,     -3,     78,    -80,      8,     72,    -85,     19,     65,    -88,     29,     57,    -90,     40,     47,    -91,     49,     38,    -90,     58,     27,    -88,     66,     17,    -84,     73,      6,    -79,     79,     -6,    -73,     84,    -17,    -66,     88,    -27,    -58,     90,    -38,    -49,     91,    -47,    -40,     90,    -57,    -29,     88,    -65,    -19,     85,    -72,     -8,     80,    -78,      3,     75,    -83,     14,     68,    -87,     25,     60,    -89,     36,     51,    -90,     46,     42,    -90,     55,     32,    -89,     63,     21,    -86,     71,     10,    -81,     77,     -1,    -76,     82,    -12,    -69,     86,    -23,    -62,     89,    -34,    -53,     90,    -44 },
      {     43,    -90,     57,     26,    -87,     70,      9,    -80,     80,     -9,    -70,     87,    -26,    -57,     90,    -43,    -43,     90,    -57,    -26,     87,    -70,     -9,     80,    -80,      9,     70,    -87,     26,     57,    -90,     43,     43,    -90,     57,     26,    -87,     70,      9,    -80,     80,     -9,    -70,     87,    -26,    -57,     90,    -43,    -43,     90,    -57,    -26,     87,    -70,     -9,     80,    -80,      9,     70,    -87,     26,     57,    -90,     43,     43,    -90,     57,     26,    -87,     70,      9,    -80,     80,     -9,    -70,     87,    -26,    -57,     90,    -43,    -43,     90,    -57,    -26,     87,    -70,     -9,     80,    -80,      9,     70,    -87,     26,     57,    -90,     43,     43,    -90,     57,     26,    -87,     70,      9,    -80,     80,     -9,    -70,     87,    -26,    -57,     90,    -43,    -43,     90,    -57,    -26,     87,    -70,     -9,     80,    -80,      9,     70,    -87,     26,     57,    -90,     43 },
      {     42,    -90,     62,     19,    -83,     77,     -6,    -71,     87,    -29,    -53,     91,    -51,    -32,     88,    -69,     -8,     78,    -82,     17,     63,    -89,     40,     44,    -90,     60,     21,    -84,     76,     -3,    -72,     86,    -27,    -55,     90,    -49,    -34,     88,    -68,    -10,     79,    -81,     14,     65,    -89,     38,     46,    -90,     58,     23,    -85,     75,     -1,    -73,     86,    -25,    -57,     90,    -47,    -36,     89,    -66,    -12,     80,    -80,     12,     66,    -89,     36,     47,    -90,     57,     25,    -86,     73,      1,    -75,     85,    -23,    -58,     90,    -46,    -38,     89,    -65,    -14,     81,    -79,     10,     68,    -88,     34,     49,    -90,     55,     27,    -86,     72,      3,    -76,     84,    -21,    -60,     90,    -44,    -40,     89,    -63,    -17,     82,    -78,      8,     69,    -88,     32,     51,    -91,     53,     29,    -87,     71,      6,    -77,     83,    -19,    -62,     90,    -42 },
      {     41,    -89,     66,     11,    -79,     83,    -20,    -59,     90,    -48,    -33,     87,    -71,     -2,     74,    -86,     28,     52,    -90,     56,     24,    -84,     76,     -7,    -69,     88,    -37,    -45,     90,    -62,    -15,     81,    -81,     15,     62,    -90,     45,     37,    -88,     69,      7,    -76,     84,    -24,    -56,     90,    -52,    -28,     86,    -74,      2,     71,    -87,     33,     48,    -90,     59,     20,    -83,     79,    -11,    -66,     89,    -41,    -41,     89,    -66,    -11,     79,    -83,     20,     59,    -90,     48,     33,    -87,     71,      2,    -74,     86,    -28,    -52,     90,    -56,    -24,     84,    -76,      7,     69,    -88,     37,     45,    -90,     62,     15,    -81,     81,    -15,    -62,     90,    -45,    -37,     88,    -69,     -7,     76,    -84,     24,     56,    -90,     52,     28,    -86,     74,     -2,    -71,     87,    -33,    -48,     90,    -59,    -20,     83,    -79,     11,     66,    -89,     41 },
      {     40,    -89,     69,      3,    -73,     87,    -34,    -46,     90,    -65,    -10,     77,    -85,     27,     51,    -90,     60,     17,    -80,     82,    -21,    -57,     91,    -55,    -23,     83,    -79,     14,     62,    -90,     49,     29,    -86,     76,     -8,    -66,     89,    -44,    -36,     88,    -72,      1,     71,    -88,     38,     42,    -89,     68,      6,    -75,     86,    -32,    -47,     90,    -63,    -12,     78,    -84,     25,     53,    -90,     58,     19,    -81,     81,    -19,    -58,     90,    -53,    -25,     84,    -78,     12,     63,    -90,     47,     32,    -86,     75,     -6,    -68,     89,    -42,    -38,     88,    -71,     -1,     72,    -88,     36,     44,    -89,     66,      8,    -76,     86,    -29,    -49,     90,    -62,    -14,     79,    -83,     23,     55,    -91,     57,     21,    -82,     80,    -17,    -60,     90,    -51,    -27,     85,    -77,     10,     65,    -90,     46,     34,    -87,     73,     -3,    -69,     89,    -40 },
      {     39,    -88,     73,     -4,    -67,     90,    -47,    -30,     85,    -78,     13,     61,    -90,     54,     22,    -82,     82,    -22,    -54,     90,    -61,    -13,     78,    -85,     30,     47,    -90,     67,      4,    -73,     88,    -39,    -39,     88,    -73,      4,     67,    -90,     47,     30,    -85,     78,    -13,    -61,     90,    -54,    -22,     82,    -82,     22,     54,    -90,     61,     13,    -78,     85,    -30,    -47,     90,    -67,     -4,     73,    -88,     39,     39,    -88,     73,     -4,    -67,     90,    -47,    -30,     85,    -78,     13,     61,    -90,     54,     22,    -82,     82,    -22,    -54,     90,    -61,    -13,     78,    -85,     30,     47,    -90,     67,      4,    -73,     88,    -39,    -39,     88,    -73,      4,     67,    -90,     47,     30,    -85,     78,    -13,    -61,     90,    -54,    -22,     82,    -82,     22,     54,    -90,     61,     13,    -78,     85,    -30,    -47,     90,    -67,     -4,     73,    -88,     39 },
      {     38,    -87,     76,    -12,    -60,     91,    -58,    -14,     77,    -86,     36,     40,    -88,     75,    -10,    -62,     90,    -57,    -17,     78,    -86,     34,     42,    -88,     73,     -8,    -63,     90,    -55,    -19,     79,    -85,     32,     44,    -89,     72,     -6,    -65,     90,    -53,    -21,     80,    -84,     29,     46,    -89,     71,     -3,    -66,     90,    -51,    -23,     81,    -83,     27,     47,    -89,     69,     -1,    -68,     90,    -49,    -25,     82,    -82,     25,     49,    -90,     68,      1,    -69,     89,    -47,    -27,     83,    -81,     23,     51,    -90,     66,      3,    -71,     89,    -46,    -29,     84,    -80,     21,     53,    -90,     65,      6,    -72,     89,    -44,    -32,     85,    -79,     19,     55,    -90,     63,      8,    -73,     88,    -42,    -34,     86,    -78,     17,     57,    -90,     62,     10,    -75,     88,    -40,    -36,     86,    -77,     14,     58,    -91,     60,     12,    -76,     87,    -38 },
      {     37,    -86,     79,    -20,    -52,     90,    -69,      2,     66,    -90,     56,     15,    -76,     87,    -41,    -33,     84,    -81,     24,     48,    -89,     71,     -7,    -62,     90,    -59,    -11,     74,    -88,     45,     28,    -83,     83,    -28,    -45,     88,    -74,     11,     59,    -90,     62,      7,    -71,     89,    -48,    -24,     81,    -84,     33,     41,    -87,     76,    -15,    -56,     90,    -66,     -2,     69,    -90,     52,     20,    -79,     86,    -37,    -37,     86,    -79,     20,     52,    -90,     69,     -2,    -66,     90,    -56,    -15,     76,    -87,     41,     33,    -84,     81,    -24,    -48,     89,    -71,      7,     62,    -90,     59,     11,    -74,     88,    -45,    -28,     83,    -83,     28,     45,    -88,     74,    -11,    -59,     90,    -62,     -7,     71,    -89,     48,     24,    -81,     84,    -33,    -41,     87,    -76,     15,     56,    -90,     66,      2,    -69,     90,    -52,    -20,     79,    -86,     37 },
      {     36,    -85,     81,    -27,    -44,     88,    -77,     19,     51,    -89,     72,    -10,    -58,     90,    -66,      1,     65,    -90,     60,      8,    -71,     90,    -53,    -17,     76,    -88,     46,     25,    -80,     86,    -38,    -34,     84,    -82,     29,     42,    -87,     78,    -21,    -49,     89,    -73,     12,     57,    -90,     68,     -3,    -63,     91,    -62,     -6,     69,    -90,     55,     14,    -75,     89,    -47,    -23,     79,    -86,     40,     32,    -83,     83,    -32,    -40,     86,    -79,     23,     47,    -89,     75,    -14,    -55,     90,    -69,      6,     62,    -91,     63,      3,    -68,     90,    -57,    -12,     73,    -89,     49,     21,    -78,     87,    -42,    -29,     82,    -84,     34,     38,    -86,     80,    -25,    -46,     88,    -76,     17,     53,    -90,     71,     -8,    -60,     90,    -65,     -1,     66,    -90,     58,     10,    -72,     89,    -51,    -19,     77,    -88,     44,     27,    -81,     85,    -36 },
      {     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35 },
      {     34,    -82,     86,    -42,    -25,     78,    -88,     49,     17,    -73,     90,    -57,     -8,     68,    -90,     63,     -1,    -62,     90,    -69,     10,     55,    -89,     75,    -19,    -47,     88,    -79,     27,     40,    -85,     83,    -36,    -32,     81,    -86,     44,     23,    -77,     89,    -51,    -14,     72,    -90,     58,      6,    -66,     91,    -65,      3,     60,    -90,     71,    -12,    -53,     89,    -76,     21,     46,    -87,     80,    -29,    -38,     84,    -84,     38,     29,    -80,     87,    -46,    -21,     76,    -89,     53,     12,    -71,     90,    -60,     -3,     65,    -91,     66,     -6,    -58,     90,    -72,     14,     51,    -89,     77,    -23,    -44,     86,    -81,     32,     36,    -83,     85,    -40,    -27,     79,    -88,     47,     19,    -75,     89,    -55,    -10,     69,    -90,     62,      1,    -63,     90,    -68,      8,     57,    -90,     73,    -17,    -49,     88,    -78,     25,     42,    -86,     82,    -34 },
      {     33,    -81,     87,    -48,    -15,     71,    -90,     62,     -2,    -59,     90,    -74,     20,     45,    -86,     83,    -37,    -28,     79,    -88,     52,     11,    -69,     90,    -66,      7,     56,    -89,     76,    -24,    -41,     84,    -84,     41,     24,    -76,     89,    -56,     -7,     66,    -90,     69,    -11,    -52,     88,    -79,     28,     37,    -83,     86,    -45,    -20,     74,    -90,     59,      2,    -62,     90,    -71,     15,     48,    -87,     81,    -33,    -33,     81,    -87,     48,     15,    -71,     90,    -62,      2,     59,    -90,     74,    -20,    -45,     86,    -83,     37,     28,    -79,     88,    -52,    -11,     69,    -90,     66,     -7,    -56,     89,    -76,     24,     41,    -84,     84,    -41,    -24,     76,    -89,     56,      7,    -66,     90,    -69,     11,     52,    -88,     79,    -28,    -37,     83,    -86,     45,     20,    -74,     90,    -59,     -2,     62,    -90,     71,    -15,    -48,     87,    -81,     33 },
      {     32,    -79,     89,    -55,     -6,     63,    -90,     73,    -21,    -42,     84,    -86,     46,     17,    -71,     90,    -66,     10,     51,    -88,     81,    -36,    -27,     77,    -89,     58,      1,    -60,     90,    -76,     25,     38,    -82,     87,    -49,    -12,     68,    -91,     69,    -14,    -47,     86,    -83,     40,     23,    -75,     90,    -62,      3,     57,    -89,     78,    -29,    -34,     80,    -88,     53,      8,    -65,     90,    -72,     19,     44,    -85,     85,    -44,    -19,     72,    -90,     65,     -8,    -53,     88,    -80,     34,     29,    -78,     89,    -57,     -3,     62,    -90,     75,    -23,    -40,     83,    -86,     47,     14,    -69,     91,    -68,     12,     49,    -87,     82,    -38,    -25,     76,    -90,     60,     -1,    -58,     89,    -77,     27,     36,    -81,     88,    -51,    -10,     66,    -90,     71,    -17,    -46,     86,    -84,     42,     21,    -73,     90,    -63,      6,     55,    -89,     79,    -32 },
      {     30,    -78,     90,    -61,      4,     54,    -88,     82,    -39,    -22,     73,    -90,     67,    -13,    -47,     85,    -85,     47,     13,    -67,     90,    -73,     22,     39,    -82,     88,    -54,     -4,     61,    -90,     78,    -30,    -30,     78,    -90,     61,     -4,    -54,     88,    -82,     39,     22,    -73,     90,    -67,     13,     47,    -85,     85,    -47,    -13,     67,    -90,     73,    -22,    -39,     82,    -88,     54,      4,    -61,     90,    -78,     30,     30,    -78,     90,    -61,      4,     54,    -88,     82,    -39,    -22,     73,    -90,     67,    -13,    -47,     85,    -85,     47,     13,    -67,     90,    -73,     22,     39,    -82,     88,    -54,     -4,     61,    -90,     78,    -30,    -30,     78,    -90,     61,     -4,    -54,     88,    -82,     39,     22,    -73,     90,    -67,     13,     47,    -85,     85,    -47,    -13,     67,    -90,     73,    -22,    -39,     82,    -88,     54,      4,    -61,     90,    -78,     30 },
      {     29,    -76,     90,    -66,     14,     44,    -83,     88,    -55,     -1,     57,    -88,     82,    -42,    -17,     68,    -90,     75,    -27,    -32,     77,    -90,     65,    -12,    -46,     84,    -87,     53,      3,    -58,     89,    -81,     40,     19,    -69,     90,    -73,     25,     34,    -78,     90,    -63,     10,     47,    -85,     86,    -51,     -6,     60,    -89,     80,    -38,    -21,     71,    -91,     72,    -23,    -36,     79,    -89,     62,     -8,    -49,     86,    -86,     49,      8,    -62,     89,    -79,     36,     23,    -72,     91,    -71,     21,     38,    -80,     89,    -60,      6,     51,    -86,     85,    -47,    -10,     63,    -90,     78,    -34,    -25,     73,    -90,     69,    -19,    -40,     81,    -89,     58,     -3,    -53,     87,    -84,     46,     12,    -65,     90,    -77,     32,     27,    -75,     90,    -68,     17,     42,    -82,     88,    -57,      1,     55,    -88,     83,    -44,    -14,     66,    -90,     76,    -29 },
      {     28,    -74,     90,    -71,     24,     33,    -76,     90,    -69,     20,     37,    -79,     90,    -66,     15,     41,    -81,     89,    -62,     11,     45,    -83,     88,    -59,      7,     48,    -84,     87,    -56,      2,     52,    -86,     86,    -52,     -2,     56,    -87,     84,    -48,     -7,     59,    -88,     83,    -45,    -11,     62,    -89,     81,    -41,    -15,     66,    -90,     79,    -37,    -20,     69,    -90,     76,    -33,    -24,     71,    -90,     74,    -28,    -28,     74,    -90,     71,    -24,    -33,     76,    -90,     69,    -20,    -37,     79,    -90,     66,    -15,    -41,     81,    -89,     62,    -11,    -45,     83,    -88,     59,     -7,    -48,     84,    -87,     56,     -2,    -52,     86,    -86,     52,      2,    -56,     87,    -84,     48,      7,    -59,     88,    -83,     45,     11,    -62,     89,    -81,     41,     15,    -66,     90,    -79,     37,     20,    -69,     90,    -76,     33,     24,    -71,     90,    -74,     28 },
      {     27,    -72,     90,    -76,     34,     21,    -68,     90,    -79,     40,     14,    -63,     89,    -82,     46,      8,    -58,     88,    -85,     51,      1,    -53,     86,    -87,     57,     -6,    -47,     83,    -89,     62,    -12,    -42,     80,    -90,     66,    -19,    -36,     77,    -90,     71,    -25,    -29,     73,    -91,     75,    -32,    -23,     69,    -90,     78,    -38,    -17,     65,    -89,     81,    -44,    -10,     60,    -88,     84,    -49,     -3,     55,    -86,     86,    -55,      3,     49,    -84,     88,    -60,     10,     44,    -81,     89,    -65,     17,     38,    -78,     90,    -69,     23,     32,    -75,     91,    -73,     29,     25,    -71,     90,    -77,     36,     19,    -66,     90,    -80,     42,     12,    -62,     89,    -83,     47,      6,    -57,     87,    -86,     53,     -1,    -51,     85,    -88,     58,     -8,    -46,     82,    -89,     63,    -14,    -40,     79,    -90,     68,    -21,    -34,     76,    -90,     72,    -27 },
      {     26,    -70,     90,    -80,     43,      9,    -57,     87,    -87,     57,     -9,    -43,     80,    -90,     70,    -26,    -26,     70,    -90,     80,    -43,     -9,     57,    -87,     87,    -57,      9,     43,    -80,     90,    -70,     26,     26,    -70,     90,    -80,     43,      9,    -57,     87,    -87,     57,     -9,    -43,     80,    -90,     70,    -26,    -26,     70,    -90,     80,    -43,     -9,     57,    -87,     87,    -57,      9,     43,    -80,     90,    -70,     26,     26,    -70,     90,    -80,     43,      9,    -57,     87,    -87,     57,     -9,    -43,     80,    -90,     70,    -26,    -26,     70,    -90,     80,    -43,     -9,     57,    -87,     87,    -57,      9,     43,    -80,     90,    -70,     26,     26,    -70,     90,    -80,     43,      9,    -57,     87,    -87,     57,     -9,    -43,     80,    -90,     70,    -26,    -26,     70,    -90,     80,    -43,     -9,     57,    -87,     87,    -57,      9,     43,    -80,     90,    -70,     26 },
      {     25,    -68,     89,    -83,     51,     -3,    -46,     80,    -90,     72,    -32,    -19,     63,    -88,     86,    -57,     10,     40,    -77,     91,    -76,     38,     12,    -58,     86,    -88,     62,    -17,    -34,     73,    -90,     79,    -44,     -6,     53,    -84,     89,    -66,     23,     27,    -69,     90,    -82,     49,     -1,    -47,     81,    -90,     71,    -29,    -21,     65,    -89,     85,    -55,      8,     42,    -78,     90,    -75,     36,     14,    -60,     87,    -87,     60,    -14,    -36,     75,    -90,     78,    -42,     -8,     55,    -85,     89,    -65,     21,     29,    -71,     90,    -81,     47,      1,    -49,     82,    -90,     69,    -27,    -23,     66,    -89,     84,    -53,      6,     44,    -79,     90,    -73,     34,     17,    -62,     88,    -86,     58,    -12,    -38,     76,    -91,     77,    -40,    -10,     57,    -86,     88,    -63,     19,     32,    -72,     90,    -80,     46,      3,    -51,     83,    -89,     68,    -25 },
      {     24,    -66,     88,    -86,     59,    -15,    -33,     71,    -90,     83,    -52,      7,     41,    -76,     90,    -79,     45,      2,    -48,     81,    -90,     74,    -37,    -11,     56,    -84,     89,    -69,     28,     20,    -62,     87,    -87,     62,    -20,    -28,     69,    -89,     84,    -56,     11,     37,    -74,     90,    -81,     48,     -2,    -45,     79,    -90,     76,    -41,     -7,     52,    -83,     90,    -71,     33,     15,    -59,     86,    -88,     66,    -24,    -24,     66,    -88,     86,    -59,     15,     33,    -71,     90,    -83,     52,     -7,    -41,     76,    -90,     79,    -45,     -2,     48,    -81,     90,    -74,     37,     11,    -56,     84,    -89,     69,    -28,    -20,     62,    -87,     87,    -62,     20,     28,    -69,     89,    -84,     56,    -11,    -37,     74,    -90,     81,    -48,      2,     45,    -79,     90,    -76,     41,      7,    -52,     83,    -90,     71,    -33,    -15,     59,    -86,     88,    -66,     24 },
      {     23,    -63,     87,    -88,     66,    -27,    -19,     60,    -86,     89,    -69,     32,     14,    -57,     84,    -90,     72,    -36,    -10,     53,    -82,     90,    -75,     40,      6,    -49,     80,    -90,     77,    -44,     -1,     46,    -78,     91,    -79,     47,     -3,    -42,     76,    -90,     81,    -51,      8,     38,    -73,     90,    -83,     55,    -12,    -34,     71,    -89,     85,    -58,     17,     29,    -68,     89,    -86,     62,    -21,    -25,     65,    -88,     88,    -65,     25,     21,    -62,     86,    -89,     68,    -29,    -17,     58,    -85,     89,    -71,     34,     12,    -55,     83,    -90,     73,    -38,     -8,     51,    -81,     90,    -76,     42,      3,    -47,     79,    -91,     78,    -46,      1,     44,    -77,     90,    -80,     49,     -6,    -40,     75,    -90,     82,    -53,     10,     36,    -72,     90,    -84,     57,    -14,    -32,     69,    -89,     86,    -60,     19,     27,    -66,     88,    -87,     63,    -23 },
      {     22,    -61,     85,    -90,     73,    -39,     -4,     47,    -78,     90,    -82,     54,    -13,    -30,     67,    -88,     88,    -67,     30,     13,    -54,     82,    -90,     78,    -47,      4,     39,    -73,     90,    -85,     61,    -22,    -22,     61,    -85,     90,    -73,     39,      4,    -47,     78,    -90,     82,    -54,     13,     30,    -67,     88,    -88,     67,    -30,    -13,     54,    -82,     90,    -78,     47,     -4,    -39,     73,    -90,     85,    -61,     22,     22,    -61,     85,    -90,     73,    -39,     -4,     47,    -78,     90,    -82,     54,    -13,    -30,     67,    -88,     88,    -67,     30,     13,    -54,     82,    -90,     78,    -47,      4,     39,    -73,     90,    -85,     61,    -22,    -22,     61,    -85,     90,    -73,     39,      4,    -47,     78,    -90,     82,    -54,     13,     30,    -67,     88,    -88,     67,    -30,    -13,     54,    -82,     90,    -78,     47,     -4,    -39,     73,    -90,     85,    -61,     22 },
      {     21,    -58,     83,    -90,     78,    -49,     10,     32,    -66,     87,    -89,     72,    -40,     -1,     42,    -73,     89,    -86,     65,    -29,    -12,     51,    -79,     90,    -82,     57,    -19,    -23,     60,    -84,     90,    -77,     47,     -8,    -34,     68,    -88,     89,    -71,     38,      3,    -44,     75,    -90,     86,    -63,     27,     14,    -53,     80,    -91,     81,    -55,     17,     25,    -62,     85,    -90,     76,    -46,      6,     36,    -69,     88,    -88,     69,    -36,     -6,     46,    -76,     90,    -85,     62,    -25,    -17,     55,    -81,     91,    -80,     53,    -14,    -27,     63,    -86,     90,    -75,     44,     -3,    -38,     71,    -89,     88,    -68,     34,      8,    -47,     77,    -90,     84,    -60,     23,     19,    -57,     82,    -90,     79,    -51,     12,     29,    -65,     86,    -89,     73,    -42,      1,     40,    -72,     89,    -87,     66,    -32,    -10,     49,    -78,     90,    -83,     58,    -21 },
      {     20,    -56,     81,    -90,     83,    -59,     24,     15,    -52,     79,    -90,     84,    -62,     28,     11,    -48,     76,    -90,     86,    -66,     33,      7,    -45,     74,    -89,     87,    -69,     37,      2,    -41,     71,    -88,     88,    -71,     41,     -2,    -37,     69,    -87,     89,    -74,     45,     -7,    -33,     66,    -86,     90,    -76,     48,    -11,    -28,     62,    -84,     90,    -79,     52,    -15,    -24,     59,    -83,     90,    -81,     56,    -20,    -20,     56,    -81,     90,    -83,     59,    -24,    -15,     52,    -79,     90,    -84,     62,    -28,    -11,     48,    -76,     90,    -86,     66,    -33,     -7,     45,    -74,     89,    -87,     69,    -37,     -2,     41,    -71,     88,    -88,     71,    -41,      2,     37,    -69,     87,    -89,     74,    -45,      7,     33,    -66,     86,    -90,     76,    -48,     11,     28,    -62,     84,    -90,     79,    -52,     15,     24,    -59,     83,    -90,     81,    -56,     20 },
      {     19,    -53,     78,    -90,     86,    -68,     38,     -1,    -36,     66,    -86,     90,    -79,     55,    -21,    -17,     51,    -77,     90,    -87,     69,    -40,      3,     34,    -65,     85,    -90,     80,    -57,     23,     14,    -49,     76,    -89,     88,    -71,     42,     -6,    -32,     63,    -84,     90,    -81,     58,    -25,    -12,     47,    -75,     89,    -88,     72,    -44,      8,     29,    -62,     83,    -91,     82,    -60,     27,     10,    -46,     73,    -89,     89,    -73,     46,    -10,    -27,     60,    -82,     91,    -83,     62,    -29,     -8,     44,    -72,     88,    -89,     75,    -47,     12,     25,    -58,     81,    -90,     84,    -63,     32,      6,    -42,     71,    -88,     89,    -76,     49,    -14,    -23,     57,    -80,     90,    -85,     65,    -34,     -3,     40,    -69,     87,    -90,     77,    -51,     17,     21,    -55,     79,    -90,     86,    -66,     36,      1,    -38,     68,    -86,     90,    -78,     53,    -19 },
      {     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18 },
      {     17,    -47,     72,    -87,     90,    -81,     62,    -34,      1,     32,    -60,     80,    -90,     88,    -73,     49,    -19,    -14,     46,    -71,     86,    -90,     82,    -63,     36,     -3,    -29,     58,    -79,     90,    -88,     75,    -51,     21,     12,    -44,     69,    -86,     90,    -83,     65,    -38,      6,     27,    -57,     78,    -89,     89,    -76,     53,    -23,    -10,     42,    -68,     85,    -91,     84,    -66,     40,     -8,    -25,     55,    -77,     89,    -89,     77,    -55,     25,      8,    -40,     66,    -84,     91,    -85,     68,    -42,     10,     23,    -53,     76,    -89,     89,    -78,     57,    -27,     -6,     38,    -65,     83,    -90,     86,    -69,     44,    -12,    -21,     51,    -75,     88,    -90,     79,    -58,     29,      3,    -36,     63,    -82,     90,    -86,     71,    -46,     14,     19,    -49,     73,    -88,     90,    -80,     60,    -32,     -1,     34,    -62,     81,    -90,     87,    -72,     47,    -17 },
      {     15,    -45,     69,    -84,     90,    -86,     71,    -48,     20,     11,    -41,     66,    -83,     90,    -87,     74,    -52,     24,      7,    -37,     62,    -81,     90,    -88,     76,    -56,     28,      2,    -33,     59,    -79,     89,    -89,     79,    -59,     33,     -2,    -28,     56,    -76,     88,    -90,     81,    -62,     37,     -7,    -24,     52,    -74,     87,    -90,     83,    -66,     41,    -11,    -20,     48,    -71,     86,    -90,     84,    -69,     45,    -15,    -15,     45,    -69,     84,    -90,     86,    -71,     48,    -20,    -11,     41,    -66,     83,    -90,     87,    -74,     52,    -24,     -7,     37,    -62,     81,    -90,     88,    -76,     56,    -28,     -2,     33,    -59,     79,    -89,     89,    -79,     59,    -33,      2,     28,    -56,     76,    -88,     90,    -81,     62,    -37,      7,     24,    -52,     74,    -87,     90,    -83,     66,    -41,     11,     20,    -48,     71,    -86,     90,    -84,     69,    -45,     15 },
      {     14,    -42,     65,    -81,     90,    -89,     79,    -62,     38,    -10,    -19,     46,    -68,     83,    -90,     88,    -77,     58,    -34,      6,     23,    -49,     71,    -85,     90,    -87,     75,    -55,     29,     -1,    -27,     53,    -73,     86,    -91,     86,    -72,     51,    -25,     -3,     32,    -57,     76,    -88,     90,    -84,     69,    -47,     21,      8,    -36,     60,    -78,     89,    -90,     82,    -66,     44,    -17,    -12,     40,    -63,     80,    -89,     89,    -80,     63,    -40,     12,     17,    -44,     66,    -82,     90,    -89,     78,    -60,     36,     -8,    -21,     47,    -69,     84,    -90,     88,    -76,     57,    -32,      3,     25,    -51,     72,    -86,     91,    -86,     73,    -53,     27,      1,    -29,     55,    -75,     87,    -90,     85,    -71,     49,    -23,     -6,     34,    -58,     77,    -88,     90,    -83,     68,    -46,     19,     10,    -38,     62,    -79,     89,    -90,     81,    -65,     42,    -14 },
      {     13,    -39,     61,    -78,     88,    -90,     85,    -73,     54,    -30,      4,     22,    -47,     67,    -82,     90,    -90,     82,    -67,     47,    -22,     -4,     30,    -54,     73,    -85,     90,    -88,     78,    -61,     39,    -13,    -13,     39,    -61,     78,    -88,     90,    -85,     73,    -54,     30,     -4,    -22,     47,    -67,     82,    -90,     90,    -82,     67,    -47,     22,      4,    -30,     54,    -73,     85,    -90,     88,    -78,     61,    -39,     13,     13,    -39,     61,    -78,     88,    -90,     85,    -73,     54,    -30,      4,     22,    -47,     67,    -82,     90,    -90,     82,    -67,     47,    -22,     -4,     30,    -54,     73,    -85,     90,    -88,     78,    -61,     39,    -13,    -13,     39,    -61,     78,    -88,     90,    -85,     73,    -54,     30,     -4,    -22,     47,    -67,     82,    -90,     90,    -82,     67,    -47,     22,      4,    -30,     54,    -73,     85,    -90,     88,    -78,     61,    -39,     13 },
      {     12,    -36,     57,    -73,     85,    -90,     89,    -81,     68,    -49,     27,     -3,    -21,     44,    -63,     78,    -88,     91,    -87,     77,    -62,     42,    -19,     -6,     29,    -51,     69,    -82,     89,    -90,     84,    -72,     55,    -34,     10,     14,    -38,     58,    -75,     86,    -90,     89,    -80,     66,    -47,     25,     -1,    -23,     46,    -65,     79,    -88,     90,    -86,     76,    -60,     40,    -17,     -8,     32,    -53,     71,    -83,     90,    -90,     83,    -71,     53,    -32,      8,     17,    -40,     60,    -76,     86,    -90,     88,    -79,     65,    -46,     23,      1,    -25,     47,    -66,     80,    -89,     90,    -86,     75,    -58,     38,    -14,    -10,     34,    -55,     72,    -84,     90,    -89,     82,    -69,     51,    -29,      6,     19,    -42,     62,    -77,     87,    -91,     88,    -78,     63,    -44,     21,      3,    -27,     49,    -68,     81,    -89,     90,    -85,     73,    -57,     36,    -12 },
      {     11,    -33,     52,    -69,     81,    -88,     90,    -87,     79,    -66,     48,    -28,      7,     15,    -37,     56,    -71,     83,    -89,     90,    -86,     76,    -62,     45,    -24,      2,     20,    -41,     59,    -74,     84,    -90,     90,    -84,     74,    -59,     41,    -20,     -2,     24,    -45,     62,    -76,     86,    -90,     89,    -83,     71,    -56,     37,    -15,     -7,     28,    -48,     66,    -79,     87,    -90,     88,    -81,     69,    -52,     33,    -11,    -11,     33,    -52,     69,    -81,     88,    -90,     87,    -79,     66,    -48,     28,     -7,    -15,     37,    -56,     71,    -83,     89,    -90,     86,    -76,     62,    -45,     24,     -2,    -20,     41,    -59,     74,    -84,     90,    -90,     84,    -74,     59,    -41,     20,      2,    -24,     45,    -62,     76,    -86,     90,    -89,     83,    -71,     56,    -37,     15,      7,    -28,     48,    -66,     79,    -87,     90,    -88,     81,    -69,     52,    -33,     11 },
      {     10,    -29,     47,    -63,     76,    -85,     90,    -90,     86,    -78,     66,    -51,     34,    -14,     -6,     25,    -44,     60,    -73,     83,    -89,     90,    -88,     80,    -69,     55,    -38,     19,      1,    -21,     40,    -57,     71,    -81,     88,    -91,     89,    -82,     72,    -58,     42,    -23,      3,     17,    -36,     53,    -68,     79,    -87,     90,    -89,     84,    -75,     62,    -46,     27,     -8,    -12,     32,    -49,     65,    -77,     86,    -90,     90,    -86,     77,    -65,     49,    -32,     12,      8,    -27,     46,    -62,     75,    -84,     89,    -90,     87,    -79,     68,    -53,     36,    -17,     -3,     23,    -42,     58,    -72,     82,    -89,     91,    -88,     81,    -71,     57,    -40,     21,     -1,    -19,     38,    -55,     69,    -80,     88,    -90,     89,    -83,     73,    -60,     44,    -25,      6,     14,    -34,     51,    -66,     78,    -86,     90,    -90,     85,    -76,     63,    -47,     29,    -10 },
      {      9,    -26,     43,    -57,     70,    -80,     87,    -90,     90,    -87,     80,    -70,     57,    -43,     26,     -9,     -9,     26,    -43,     57,    -70,     80,    -87,     90,    -90,     87,    -80,     70,    -57,     43,    -26,      9,      9,    -26,     43,    -57,     70,    -80,     87,    -90,     90,    -87,     80,    -70,     57,    -43,     26,     -9,     -9,     26,    -43,     57,    -70,     80,    -87,     90,    -90,     87,    -80,     70,    -57,     43,    -26,      9,      9,    -26,     43,    -57,     70,    -80,     87,    -90,     90,    -87,     80,    -70,     57,    -43,     26,     -9,     -9,     26,    -43,     57,    -70,     80,    -87,     90,    -90,     87,    -80,     70,    -57,     43,    -26,      9,      9,    -26,     43,    -57,     70,    -80,     87,    -90,     90,    -87,     80,    -70,     57,    -43,     26,     -9,     -9,     26,    -43,     57,    -70,     80,    -87,     90,    -90,     87,    -80,     70,    -57,     43,    -26,      9 },
      {      8,    -23,     38,    -51,     63,    -73,     81,    -87,     90,    -90,     88,    -83,     76,    -66,     55,    -42,     27,    -12,     -3,     19,    -34,     47,    -60,     71,    -79,     86,    -89,     91,    -89,     85,    -78,     69,    -58,     46,    -32,     17,     -1,    -14,     29,    -44,     57,    -68,     77,    -84,     89,    -90,     90,    -86,     80,    -72,     62,    -49,     36,    -21,      6,     10,    -25,     40,    -53,     65,    -75,     82,    -88,     90,    -90,     88,    -82,     75,    -65,     53,    -40,     25,    -10,     -6,     21,    -36,     49,    -62,     72,    -80,     86,    -90,     90,    -89,     84,    -77,     68,    -57,     44,    -29,     14,      1,    -17,     32,    -46,     58,    -69,     78,    -85,     89,    -91,     89,    -86,     79,    -71,     60,    -47,     34,    -19,      3,     12,    -27,     42,    -55,     66,    -76,     83,    -88,     90,    -90,     87,    -81,     73,    -63,     51,    -38,     23,     -8 },
      {      7,    -20,     33,    -45,     56,    -66,     74,    -81,     86,    -89,     90,    -90,     87,    -83,     76,    -69,     59,    -48,     37,    -24,     11,      2,    -15,     28,    -41,     52,    -62,     71,    -79,     84,    -88,     90,    -90,     88,    -84,     79,    -71,     62,    -52,     41,    -28,     15,     -2,    -11,     24,    -37,     48,    -59,     69,    -76,     83,    -87,     90,    -90,     89,    -86,     81,    -74,     66,    -56,     45,    -33,     20,     -7,     -7,     20,    -33,     45,    -56,     66,    -74,     81,    -86,     89,    -90,     90,    -87,     83,    -76,     69,    -59,     48,    -37,     24,    -11,     -2,     15,    -28,     41,    -52,     62,    -71,     79,    -84,     88,    -90,     90,    -88,     84,    -79,     71,    -62,     52,    -41,     28,    -15,      2,     11,    -24,     37,    -48,     59,    -69,     76,    -83,     87,    -90,     90,    -89,     86,    -81,     74,    -66,     56,    -45,     33,    -20,      7 },
      {      6,    -17,     27,    -38,     47,    -57,     65,    -72,     78,    -83,     87,    -89,     90,    -90,     89,    -86,     81,    -76,     69,    -62,     53,    -44,     34,    -23,     12,     -1,    -10,     21,    -32,     42,    -51,     60,    -68,     75,    -80,     85,    -88,     90,    -91,     90,    -88,     84,    -79,     73,    -66,     58,    -49,     40,    -29,     19,     -8,     -3,     14,    -25,     36,    -46,     55,    -63,     71,    -77,     82,    -86,     89,    -90,     90,    -89,     86,    -82,     77,    -71,     63,    -55,     46,    -36,     25,    -14,      3,      8,    -19,     29,    -40,     49,    -58,     66,    -73,     79,    -84,     88,    -90,     91,    -90,     88,    -85,     80,    -75,     68,    -60,     51,    -42,     32,    -21,     10,      1,    -12,     23,    -34,     44,    -53,     62,    -69,     76,    -81,     86,    -89,     90,    -90,     89,    -87,     83,    -78,     72,    -65,     57,    -47,     38,    -27,     17,     -6 },
      {      4,    -13,     22,    -30,     39,    -47,     54,    -61,     67,    -73,     78,    -82,     85,    -88,     90,    -90,     90,    -90,     88,    -85,     82,    -78,     73,    -67,     61,    -54,     47,    -39,     30,    -22,     13,     -4,     -4,     13,    -22,     30,    -39,     47,    -54,     61,    -67,     73,    -78,     82,    -85,     88,    -90,     90,    -90,     90,    -88,     85,    -82,     78,    -73,     67,    -61,     54,    -47,     39,    -30,     22,    -13,      4,      4,    -13,     22,    -30,     39,    -47,     54,    -61,     67,    -73,     78,    -82,     85,    -88,     90,    -90,     90,    -90,     88,    -85,     82,    -78,     73,    -67,     61,    -54,     47,    -39,     30,    -22,     13,     -4,     -4,     13,    -22,     30,    -39,     47,    -54,     61,    -67,     73,    -78,     82,    -85,     88,    -90,     90,    -90,     90,    -88,     85,    -82,     78,    -73,     67,    -61,     54,    -47,     39,    -30,     22,    -13,      4 },
      {      3,    -10,     17,    -23,     29,    -36,     42,    -47,     53,    -58,     63,    -68,     72,    -76,     79,    -82,     85,    -87,     89,    -90,     90,    -91,     90,    -89,     88,    -86,     84,    -81,     78,    -75,     71,    -66,     62,    -57,     51,    -46,     40,    -34,     27,    -21,     14,     -8,      1,      6,    -12,     19,    -25,     32,    -38,     44,    -49,     55,    -60,     65,    -69,     73,    -77,     80,    -83,     86,    -88,     89,    -90,     90,    -90,     90,    -89,     88,    -86,     83,    -80,     77,    -73,     69,    -65,     60,    -55,     49,    -44,     38,    -32,     25,    -19,     12,     -6,     -1,      8,    -14,     21,    -27,     34,    -40,     46,    -51,     57,    -62,     66,    -71,     75,    -78,     81,    -84,     86,    -88,     89,    -90,     91,    -90,     90,    -89,     87,    -85,     82,    -79,     76,    -72,     68,    -63,     58,    -53,     47,    -42,     36,    -29,     23,    -17,     10,     -3 },
      {      2,     -7,     11,    -15,     20,    -24,     28,    -33,     37,    -41,     45,    -48,     52,    -56,     59,    -62,     66,    -69,     71,    -74,     76,    -79,     81,    -83,     84,    -86,     87,    -88,     89,    -90,     90,    -90,     90,    -90,     90,    -89,     88,    -87,     86,    -84,     83,    -81,     79,    -76,     74,    -71,     69,    -66,     62,    -59,     56,    -52,     48,    -45,     41,    -37,     33,    -28,     24,    -20,     15,    -11,      7,     -2,     -2,      7,    -11,     15,    -20,     24,    -28,     33,    -37,     41,    -45,     48,    -52,     56,    -59,     62,    -66,     69,    -71,     74,    -76,     79,    -81,     83,    -84,     86,    -87,     88,    -89,     90,    -90,     90,    -90,     90,    -90,     89,    -88,     87,    -86,     84,    -83,     81,    -79,     76,    -74,     71,    -69,     66,    -62,     59,    -56,     52,    -48,     45,    -41,     37,    -33,     28,    -24,     20,    -15,     11,     -7,      2 },
      {      1,     -3,      6,     -8,     10,    -12,     14,    -17,     19,    -21,     23,    -25,     27,    -29,     32,    -34,     36,    -38,     40,    -42,     44,    -46,     47,    -49,     51,    -53,     55,    -57,     58,    -60,     62,    -63,     65,    -66,     68,    -69,     71,    -72,     73,    -75,     76,    -77,     78,    -79,     80,    -81,     82,    -83,     84,    -85,     86,    -86,     87,    -88,     88,    -89,     89,    -89,     90,    -90,     90,    -90,     90,    -91,     91,    -90,     90,    -90,     90,    -90,     89,    -89,     89,    -88,     88,    -87,     86,    -86,     85,    -84,     83,    -82,     81,    -80,     79,    -78,     77,    -76,     75,    -73,     72,    -71,     69,    -68,     66,    -65,     63,    -62,     60,    -58,     57,    -55,     53,    -51,     49,    -47,     46,    -44,     42,    -40,     38,    -36,     34,    -32,     29,    -27,     25,    -23,     21,    -19,     17,    -14,     12,    -10,      8,     -6,      3,     -1 },
    },
    {
      {     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     64 },
      {     91,     90,     90,     90,     90,     90,     89,     89,     89,     88,     88,     87,     86,     86,     85,     84,     83,     82,     81,     80,     79,     78,     77,     76,     75,     73,     72,     71,     69,     68,     66,     65,     63,     62,     60,     58,     57,     55,     53,     51,     49,     47,     46,     44,     42,     40,     38,     36,     34,     32,     29,     27,     25,     23,     21,     19,     17,     14,     12,     10,      8,      6,      3,      1,     -1,     -3,     -6,     -8,    -10,    -12,    -14,    -17,    -19,    -21,    -23,    -25,    -27,    -29,    -32,    -34,    -36,    -38,    -40,    -42,    -44,    -46,    -47,    -49,    -51,    -53,    -55,    -57,    -58,    -60,    -62,    -63,    -65,    -66,    -68,    -69,    -71,    -72,    -73,    -75,    -76,    -77,    -78,    -79,    -80,    -81,    -82,    -83,    -84,    -85,    -86,    -86,    -87,    -88,    -88,    -89,    -89,    -89,    -90,    -90,    -90,    -90,    -90,    -91 },
      {     90,     90,     90,     89,     88,     87,     86,     84,     83,     81,     79,     76,     74,     71,     69,     66,     62,     59,     56,     52,     48,     45,     41,     37,     33,     28,     24,     20,     15,     11,      7,      2,     -2,     -7,    -11,    -15,    -20,    -24,    -28,    -33,    -37,    -41,    -45,    -48,    -52,    -56,    -59,    -62,    -66,    -69,    -71,    -74,    -76,    -79,    -81,    -83,    -84,    -86,    -87,    -88,    -89,    -90,    -90,    -90,    -90,    -90,    -90,    -89,    -88,    -87,    -86,    -84,    -83,    -81,    -79,    -76,    -74,    -71,    -69,    -66,    -62,    -59,    -56,    -52,    -48,    -45,    -41,    -37,    -33,    -28,    -24,    -20,    -15,    -11,     -7,     -2,      2,      7,     11,     15,     20,     24,     28,     33,     37,     41,     45,     48,     52,     56,     59,     62,     66,     69,     71,     74,     76,     79,     81,     83,     84,     86,     87,     88,     89,     90,     90,     90 },
      {     90,     90,     89,     88,     86,     83,     80,     77,     73,     69,     65,     60,     55,     49,     44,     38,     32,     25,     19,     12,      6,     -1,     -8,    -14,    -21,    -27,    -34,    -40,    -46,    -51,    -57,    -62,    -66,    -71,    -75,    -78,    -81,    -84,    -86,    -88,    -89,    -90,    -91,    -90,    -90,    -89,    -87,    -85,    -82,    -79,    -76,    -72,    -68,    -63,    -58,    -53,    -47,    -42,    -36,    -29,    -23,    -17,    -10,     -3,      3,     10,     17,     23,     29,     36,     42,     47,     53,     58,     63,     68,     72,     76,     79,     82,     85,     87,     89,     90,     90,     91,     90,     89,     88,     86,     84,     81,     78,     75,     71,     66,     62,     57,     51,     46,     40,     34,     27,     21,     14,      8,      1,     -6,    -12,    -19,    -25,    -32,    -38,    -44,    -49,    -55,    -60,    -65,    -69,    -73,    -77,    -80,    -83,    -86,    -88,    -89,    -90,    -90 },
      {     90,     90,     88,     85,     82,     78,     73,     67,     61,     54,     47,     39,     30,     22,     13,      4,     -4,    -13,    -22,    -30,    -39,    -47,    -54,    -61,    -67,    -73,    -78,    -82,    -85,    -88,    -90,    -90,    -90,    -90,    -88,    -85,    -82,    -78,    -73,    -67,    -61,    -54,    -47,    -39,    -30,    -22,    -13,     -4,      4,     13,     22,     30,     39,     47,     54,     61,     67,     73,     78,     82,     85,     88,     90,     90,     90,     90,     88,     85,     82,     78,     73,     67,     61,     54,     47,     39,     30,     22,     13,      4,     -4,    -13,    -22,    -30,    -39,    -47,    -54,    -61,    -67,    -73,    -78,    -82,    -85,    -88,    -90,    -90,    -90,    -90,    -88,    -85,    -82,    -78,    -73,    -67,    -61,    -54,    -47,    -39,    -30,    -22,    -13,     -4,      4,     13,     22,     30,     39,     47,     54,     61,     67,     73,     78,     82,     85,     88,     90,     90 },
      {     90,     89,     86,     82,     77,     71,     63,     55,     46,     36,     25,     14,      3,     -8,    -19,    -29,    -40,    -49,    -58,    -66,    -73,    -79,    -84,    -88,    -90,    -91,    -90,    -88,    -85,    -80,    -75,    -68,    -60,    -51,    -42,    -32,    -21,    -10,      1,     12,     23,     34,     44,     53,     62,     69,     76,     81,     86,     89,     90,     90,     89,     87,     83,     78,     72,     65,     57,     47,     38,     27,     17,      6,     -6,    -17,    -27,    -38,    -47,    -57,    -65,    -72,    -78,    -83,    -87,    -89,    -90,    -90,    -89,    -86,    -81,    -76,    -69,    -62,    -53,    -44,    -34,    -23,    -12,     -1,     10,     21,     32,     42,     51,     60,     68,     75,     80,     85,     88,     90,     91,     90,     88,     84,     79,     73,     66,     58,     49,     40,     29,     19,      8,     -3,    -14,    -25,    -36,    -46,    -55,    -63,    -71,    -77,    -82,    -86,    -89,    -90 },
      {     90,     88,     84,     79,     71,     62,     52,     41,     28,     15,      2,    -11,    -24,    -37,    -48,    -59,    -69,    -76,    -83,    -87,    -90,    -90,    -89,    -86,    -81,    -74,    -66,    -56,    -45,    -33,    -20,     -7,      7,     20,     33,     45,     56,     66,     74,     81,     86,     89,     90,     90,     87,     83,     76,     69,     59,     48,     37,     24,     11,     -2,    -15,    -28,    -41,    -52,    -62,    -71,    -79,    -84,    -88,    -90,    -90,    -88,    -84,    -79,    -71,    -62,    -52,    -41,    -28,    -15,     -2,     11,     24,     37,     48,     59,     69,     76,     83,     87,     90,     90,     89,     86,     81,     74,     66,     56,     45,     33,     20,      7,     -7,    -20,    -33,    -45,    -56,    -66,    -74,    -81,    -86,    -89,    -90,    -90,    -87,    -83,    -76,    -69,    -59,    -48,    -37,    -24,    -11,      2,     15,     28,     41,     52,     62,     71,     79,     84,     88,     90 },
      {     90,     88,     82,     75,     65,     53,     40,     25,     10,     -6,    -21,    -36,    -49,    -62,    -72,    -80,    -86,    -90,    -90,    -89,    -84,    -77,    -68,    -57,    -44,    -29,    -14,      1,     17,     32,     46,     58,     69,     78,     85,     89,     91,     89,     86,     79,     71,     60,     47,     34,     19,      3,    -12,    -27,    -42,    -55,    -66,    -76,    -83,    -88,    -90,    -90,    -87,    -81,    -73,    -63,    -51,    -38,    -23,     -8,      8,     23,     38,     51,     63,     73,     81,     87,     90,     90,     88,     83,     76,     66,     55,     42,     27,     12,     -3,    -19,    -34,    -47,    -60,    -71,    -79,    -86,    -89,    -91,    -89,    -85,    -78,    -69,    -58,    -46,    -32,    -17,     -1,     14,     29,     44,     57,     68,     77,     84,     89,     90,     90,     86,     80,     72,     62,     49,     36,     21,      6,    -10,    -25,    -40,    -53,    -65,    -75,    -82,    -88,    -90 },
      {     90,     87,     80,     70,     57,     43,     26,      9,     -9,    -26,    -43,    -57,    -70,    -80,    -87,    -90,    -90,    -87,    -80,    -70,    -57,    -43,    -26,     -9,      9,     26,     43,     57,     70,     80,     87,     90,     90,     87,     80,     70,     57,     43,     26,      9,     -9,    -26,    -43,    -57,    -70,    -80,    -87,    -90,    -90,    -87,    -80,    -70,    -57,    -43,    -26,     -9,      9,     26,     43,     57,     70,     80,     87,     90,     90,     87,     80,     70,     57,     43,     26,      9,     -9,    -26,    -43,    -57,    -70,    -80,    -87,    -90,    -90,    -87,    -80,    -70,    -57,    -43,    -26,     -9,      9,     26,     43,     57,     70,     80,     87,     90,     90,     87,     80,     70,     57,     43,     26,      9,     -9,    -26,    -43,    -57,    -70,    -80,    -87,    -90,    -90,    -87,    -80,    -70,    -57,    -43,    -26,     -9,      9,     26,     43,     57,     70,     80,     87,     90 },
      {     90,     86,     77,     65,     49,     32,     12,     -8,    -27,    -46,    -62,    -75,    -84,    -89,    -90,    -87,    -79,    -68,    -53,    -36,    -17,      3,     23,     42,     58,     72,     82,     89,     91,     88,     81,     71,     57,     40,     21,      1,    -19,    -38,    -55,    -69,    -80,    -88,    -90,    -89,    -83,    -73,    -60,    -44,    -25,     -6,     14,     34,     51,     66,     78,     86,     90,     90,     85,     76,     63,     47,     29,     10,    -10,    -29,    -47,    -63,    -76,    -85,    -90,    -90,    -86,    -78,    -66,    -51,    -34,    -14,      6,     25,     44,     60,     73,     83,     89,     90,     88,     80,     69,     55,     38,     19,     -1,    -21,    -40,    -57,    -71,    -81,    -88,    -91,    -89,    -82,    -72,    -58,    -42,    -23,     -3,     17,     36,     53,     68,     79,     87,     90,     89,     84,     75,     62,     46,     27,      8,    -12,    -32,    -49,    -65,    -77,    -86,    -90 },
      {     90,     84,     74,     59,     41,     20,     -2,    -24,    -45,    -62,    -76,    -86,    -90,    -89,    -83,    -71,    -56,    -37,    -15,      7,     28,     48,     66,     79,     87,     90,     88,     81,     69,     52,     33,     11,    -11,    -33,    -52,    -69,    -81,    -88,    -90,    -87,    -79,    -66,    -48,    -28,     -7,     15,     37,     56,     71,     83,     89,     90,     86,     76,     62,     45,     24,      2,    -20,    -41,    -59,    -74,    -84,    -90,    -90,    -84,    -74,    -59,    -41,    -20,      2,     24,     45,     62,     76,     86,     90,     89,     83,     71,     56,     37,     15,     -7,    -28,    -48,    -66,    -79,    -87,    -90,    -88,    -81,    -69,    -52,    -33,    -11,     11,     33,     52,     69,     81,     88,     90,     87,     79,     66,     48,     28,      7,    -15,    -37,    -56,    -71,    -83,    -89,    -90,    -86,    -76,    -62,    -45,    -24,     -2,     20,     41,     59,     74,     84,     90 },
      {     90,     83,     71,     53,     32,      8,    -17,    -40,    -60,    -76,    -86,    -90,    -88,    -79,    -65,    -46,    -23,      1,     25,     47,     66,     80,     89,     90,     86,     75,     58,     38,     14,    -10,    -34,    -55,    -72,    -84,    -90,    -89,    -82,    -69,    -51,    -29,     -6,     19,     42,     62,     77,     87,     91,     88,     78,     63,     44,     21,     -3,    -27,    -49,    -68,    -81,    -89,    -90,    -85,    -73,    -57,    -36,    -12,     12,     36,     57,     73,     85,     90,     89,     81,     68,     49,     27,      3,    -21,    -44,    -63,    -78,    -88,    -91,    -87,    -77,    -62,    -42,    -19,      6,     29,     51,     69,     82,     89,     90,     84,     72,     55,     34,     10,    -14,    -38,    -58,    -75,    -86,    -90,    -89,    -80,    -66,    -47,    -25,     -1,     23,     46,     65,     79,     88,     90,     86,     76,     60,     40,     17,     -8,    -32,    -53,    -71,    -83,    -90 },
      {     90,     82,     67,     47,     22,     -4,    -30,    -54,    -73,    -85,    -90,    -88,    -78,    -61,    -39,    -13,     13,     39,     61,     78,     88,     90,     85,     73,     54,     30,      4,    -22,    -47,    -67,    -82,    -90,    -90,    -82,    -67,    -47,    -22,      4,     30,     54,     73,     85,     90,     88,     78,     61,     39,     13,    -13,    -39,    -61,    -78,    -88,    -90,    -85,    -73,    -54,    -30,     -4,     22,     47,     67,     82,     90,     90,     82,     67,     47,     22,     -4,    -30,    -54,    -73,    -85,    -90,    -88,    -78,    -61,    -39,    -13,     13,     39,     61,     78,     88,     90,     85,     73,     54,     30,      4,    -22,    -47,    -67,    -82,    -90,    -90,    -82,    -67,    -47,    -22,      4,     30,     54,     73,     85,     90,     88,     78,     61,     39,     13,    -13,    -39,    -61,    -78,    -88,    -90,    -85,    -73,    -54,    -30,     -4,     22,     47,     67,     82,     90 },
      {     89,     80,     63,     40,     12,    -17,    -44,    -66,    -82,    -90,    -89,    -78,    -60,    -36,     -8,     21,     47,     69,     84,     90,     88,     76,     57,     32,      3,    -25,    -51,    -72,    -86,    -91,    -86,    -73,    -53,    -27,      1,     29,     55,     75,     87,     90,     85,     71,     49,     23,     -6,    -34,    -58,    -77,    -88,    -90,    -83,    -68,    -46,    -19,     10,     38,     62,     79,     89,     90,     81,     65,     42,     14,    -14,    -42,    -65,    -81,    -90,    -89,    -79,    -62,    -38,    -10,     19,     46,     68,     83,     90,     88,     77,     58,     34,      6,    -23,    -49,    -71,    -85,    -90,    -87,    -75,    -55,    -29,     -1,     27,     53,     73,     86,     91,     86,     72,     51,     25,     -3,    -32,    -57,    -76,    -88,    -90,    -84,    -69,    -47,    -21,      8,     36,     60,     78,     89,     90,     82,     66,     44,     17,    -12,    -40,    -63,    -80,    -89 },
      {     89,     79,     59,     33,      2,    -28,    -56,    -76,    -88,    -90,    -81,    -62,    -37,     -7,     24,     52,     74,     87,     90,     83,     66,     41,     11,    -20,    -48,    -71,    -86,    -90,    -84,    -69,    -45,    -15,     15,     45,     69,     84,     90,     86,     71,     48,     20,    -11,    -41,    -66,    -83,    -90,    -87,    -74,    -52,    -24,      7,     37,     62,     81,     90,     88,     76,     56,     28,     -2,    -33,    -59,    -79,    -89,    -89,    -79,    -59,    -33,     -2,     28,     56,     76,     88,     90,     81,     62,     37,      7,    -24,    -52,    -74,    -87,    -90,    -83,    -66,    -41,    -11,     20,     48,     71,     86,     90,     84,     69,     45,     15,    -15,    -45,    -69,    -84,    -90,    -86,    -71,    -48,    -20,     11,     41,     66,     83,     90,     87,     74,     52,     24,     -7,    -37,    -62,    -81,    -90,    -88,    -76,    -56,    -28,      2,     33,     59,     79,     89 },
      {     89,     77,     55,     25,     -8,    -40,    -66,    -84,    -91,    -85,    -68,    -42,    -10,     23,     53,     76,     89,     89,     78,     57,     27,     -6,    -38,    -65,    -83,    -90,    -86,    -69,    -44,    -12,     21,     51,     75,     88,     90,     79,     58,     29,     -3,    -36,    -63,    -82,    -90,    -86,    -71,    -46,    -14,     19,     49,     73,     88,     90,     80,     60,     32,     -1,    -34,    -62,    -81,    -90,    -87,    -72,    -47,    -17,     17,     47,     72,     87,     90,     81,     62,     34,      1,    -32,    -60,    -80,    -90,    -88,    -73,    -49,    -19,     14,     46,     71,     86,     90,     82,     63,     36,      3,    -29,    -58,    -79,    -90,    -88,    -75,    -51,    -21,     12,     44,     69,     86,     90,     83,     65,     38,      6,    -27,    -57,    -78,    -89,    -89,    -76,    -53,    -23,     10,     42,     68,     85,     91,     84,     66,     40,      8,    -25,    -55,    -77,    -89 },
      {     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89,     89,     75,     50,     18,    -18,    -50,    -75,    -89,    -89,    -75,    -50,    -18,     18,     50,     75,     89 },
      {     89,     73,     46,     10,    -27,    -60,    -82,    -91,    -83,    -62,    -29,      8,     44,     72,     88,     89,     75,     47,     12,    -25,    -58,    -81,    -90,    -84,    -63,    -32,      6,     42,     71,     88,     89,     76,     49,     14,    -23,    -57,    -80,    -90,    -85,    -65,    -34,      3,     40,     69,     87,     90,     77,     51,     17,    -21,    -55,    -79,    -90,    -86,    -66,    -36,      1,     38,     68,     86,     90,     78,     53,     19,    -19,    -53,    -78,    -90,    -86,    -68,    -38,     -1,     36,     66,     86,     90,     79,     55,     21,    -17,    -51,    -77,    -90,    -87,    -69,    -40,     -3,     34,     65,     85,     90,     80,     57,     23,    -14,    -49,    -76,    -89,    -88,    -71,    -42,     -6,     32,     63,     84,     90,     81,     58,     25,    -12,    -47,    -75,    -89,    -88,    -72,    -44,     -8,     29,     62,     83,     91,     82,     60,     27,    -10,    -46,    -73,    -89 },
      {     88,     71,     41,      2,    -37,    -69,    -87,    -89,    -74,    -45,     -7,     33,     66,     86,     90,     76,     48,     11,    -28,    -62,    -84,    -90,    -79,    -52,    -15,     24,     59,     83,     90,     81,     56,     20,    -20,    -56,    -81,    -90,    -83,    -59,    -24,     15,     52,     79,     90,     84,     62,     28,    -11,    -48,    -76,    -90,    -86,    -66,    -33,      7,     45,     74,     89,     87,     69,     37,     -2,    -41,    -71,    -88,    -88,    -71,    -41,     -2,     37,     69,     87,     89,     74,     45,      7,    -33,    -66,    -86,    -90,    -76,    -48,    -11,     28,     62,     84,     90,     79,     52,     15,    -24,    -59,    -83,    -90,    -81,    -56,    -20,     20,     56,     81,     90,     83,     59,     24,    -15,    -52,    -79,    -90,    -84,    -62,    -28,     11,     48,     76,     90,     86,     66,     33,     -7,    -45,    -74,    -89,    -87,    -69,    -37,      2,     41,     71,     88 },
      {     88,     69,     36,     -6,    -46,    -76,    -90,    -85,    -62,    -25,     17,     55,     81,     91,     80,     53,     14,    -27,    -63,    -86,    -90,    -75,    -44,     -3,     38,     71,     89,     88,     68,     34,     -8,    -47,    -77,    -90,    -84,    -60,    -23,     19,     57,     82,     90,     79,     51,     12,    -29,    -65,    -86,    -89,    -73,    -42,     -1,     40,     72,     89,     87,     66,     32,    -10,    -49,    -78,    -90,    -83,    -58,    -21,     21,     58,     83,     90,     78,     49,     10,    -32,    -66,    -87,    -89,    -72,    -40,      1,     42,     73,     89,     86,     65,     29,    -12,    -51,    -79,    -90,    -82,    -57,    -19,     23,     60,     84,     90,     77,     47,      8,    -34,    -68,    -88,    -89,    -71,    -38,      3,     44,     75,     90,     86,     63,     27,    -14,    -53,    -80,    -91,    -81,    -55,    -17,     25,     62,     85,     90,     76,     46,      6,    -36,    -69,    -88 },
      {     88,     67,     30,    -13,    -54,    -82,    -90,    -78,    -47,     -4,     39,     73,     90,     85,     61,     22,    -22,    -61,    -85,    -90,    -73,    -39,      4,     47,     78,     90,     82,     54,     13,    -30,    -67,    -88,    -88,    -67,    -30,     13,     54,     82,     90,     78,     47,      4,    -39,    -73,    -90,    -85,    -61,    -22,     22,     61,     85,     90,     73,     39,     -4,    -47,    -78,    -90,    -82,    -54,    -13,     30,     67,     88,     88,     67,     30,    -13,    -54,    -82,    -90,    -78,    -47,     -4,     39,     73,     90,     85,     61,     22,    -22,    -61,    -85,    -90,    -73,    -39,      4,     47,     78,     90,     82,     54,     13,    -30,    -67,    -88,    -88,    -67,    -30,     13,     54,     82,     90,     78,     47,      4,    -39,    -73,    -90,    -85,    -61,    -22,     22,     61,     85,     90,     73,     39,     -4,    -47,    -78,    -90,    -82,    -54,    -13,     30,     67,     88 },
      {     88,     65,     25,    -21,    -62,    -86,    -89,    -68,    -29,     17,     58,     85,     89,     71,     34,    -12,    -55,    -83,    -90,    -73,    -38,      8,     51,     81,     90,     76,     42,     -3,    -47,    -79,    -91,    -78,    -46,     -1,     44,     77,     90,     80,     49,      6,    -40,    -75,    -90,    -82,    -53,    -10,     36,     72,     90,     84,     57,     14,    -32,    -69,    -89,    -86,    -60,    -19,     27,     66,     88,     87,     63,     23,    -23,    -63,    -87,    -88,    -66,    -27,     19,     60,     86,     89,     69,     32,    -14,    -57,    -84,    -90,    -72,    -36,     10,     53,     82,     90,     75,     40,     -6,    -49,    -80,    -90,    -77,    -44,      1,     46,     78,     91,     79,     47,      3,    -42,    -76,    -90,    -81,    -51,     -8,     38,     73,     90,     83,     55,     12,    -34,    -71,    -89,    -85,    -58,    -17,     29,     68,     89,     86,     62,     21,    -25,    -65,    -88 },
      {     87,     62,     20,    -28,    -69,    -89,    -84,    -56,    -11,     37,     74,     90,     81,     48,      2,    -45,    -79,    -90,    -76,    -41,      7,     52,     83,     90,     71,     33,    -15,    -59,    -86,    -88,    -66,    -24,     24,     66,     88,     86,     59,     15,    -33,    -71,    -90,    -83,    -52,     -7,     41,     76,     90,     79,     45,     -2,    -48,    -81,    -90,    -74,    -37,     11,     56,     84,     89,     69,     28,    -20,    -62,    -87,    -87,    -62,    -20,     28,     69,     89,     84,     56,     11,    -37,    -74,    -90,    -81,    -48,     -2,     45,     79,     90,     76,     41,     -7,    -52,    -83,    -90,    -71,    -33,     15,     59,     86,     88,     66,     24,    -24,    -66,    -88,    -86,    -59,    -15,     33,     71,     90,     83,     52,      7,    -41,    -76,    -90,    -79,    -45,      2,     48,     81,     90,     74,     37,    -11,    -56,    -84,    -89,    -69,    -28,     20,     62,     87 },
      {     87,     60,     14,    -36,    -75,    -90,    -78,    -42,      8,     55,     85,     89,     65,     21,    -29,    -71,    -90,    -81,    -47,      1,     49,     82,     90,     69,     27,    -23,    -66,    -89,    -84,    -53,     -6,     44,     79,     90,     73,     34,    -17,    -62,    -88,    -86,    -58,    -12,     38,     76,     91,     77,     40,    -10,    -57,    -86,    -88,    -63,    -19,     32,     72,     90,     80,     46,     -3,    -51,    -83,    -89,    -68,    -25,     25,     68,     89,     83,     51,      3,    -46,    -80,    -90,    -72,    -32,     19,     63,     88,     86,     57,     10,    -40,    -77,    -91,    -76,    -38,     12,     58,     86,     88,     62,     17,    -34,    -73,    -90,    -79,    -44,      6,     53,     84,     89,     66,     23,    -27,    -69,    -90,    -82,    -49,     -1,     47,     81,     90,     71,     29,    -21,    -65,    -89,    -85,    -55,     -8,     42,     78,     90,     75,     36,    -14,    -60,    -87 },
      {     87,     57,      9,    -43,    -80,    -90,    -70,    -26,     26,     70,     90,     80,     43,     -9,    -57,    -87,    -87,    -57,     -9,     43,     80,     90,     70,     26,    -26,    -70,    -90,    -80,    -43,      9,     57,     87,     87,     57,      9,    -43,    -80,    -90,    -70,    -26,     26,     70,     90,     80,     43,     -9,    -57,    -87,    -87,    -57,     -9,     43,     80,     90,     70,     26,    -26,    -70,    -90,    -80,    -43,      9,     57,     87,     87,     57,      9,    -43,    -80,    -90,    -70,    -26,     26,     70,     90,     80,     43,     -9,    -57,    -87,    -87,    -57,     -9,     43,     80,     90,     70,     26,    -26,    -70,    -90,    -80,    -43,      9,     57,     87,     87,     57,      9,    -43,    -80,    -90,    -70,    -26,     26,     70,     90,     80,     43,     -9,    -57,    -87,    -87,    -57,     -9,     43,     80,     90,     70,     26,    -26,    -70,    -90,    -80,    -43,      9,     57,     87 },
      {     86,     55,      3,    -49,    -84,    -88,    -60,    -10,     44,     81,     89,     65,     17,    -38,    -78,    -90,    -69,    -23,     32,     75,     91,     73,     29,    -25,    -71,    -90,    -77,    -36,     19,     66,     90,     80,     42,    -12,    -62,    -89,    -83,    -47,      6,     57,     87,     86,     53,      1,    -51,    -85,    -88,    -58,     -8,     46,     82,     89,     63,     14,    -40,    -79,    -90,    -68,    -21,     34,     76,     90,     72,     27,    -27,    -72,    -90,    -76,    -34,     21,     68,     90,     79,     40,    -14,    -63,    -89,    -82,    -46,      8,     58,     88,     85,     51,     -1,    -53,    -86,    -87,    -57,     -6,     47,     83,     89,     62,     12,    -42,    -80,    -90,    -66,    -19,     36,     77,     90,     71,     25,    -29,    -73,    -91,    -75,    -32,     23,     69,     90,     78,     38,    -17,    -65,    -89,    -81,    -44,     10,     60,     88,     84,     49,     -3,    -55,    -86 },
      {     86,     52,     -2,    -56,    -87,    -84,    -48,      7,     59,     88,     83,     45,    -11,    -62,    -89,    -81,    -41,     15,     66,     90,     79,     37,    -20,    -69,    -90,    -76,    -33,     24,     71,     90,     74,     28,    -28,    -74,    -90,    -71,    -24,     33,     76,     90,     69,     20,    -37,    -79,    -90,    -66,    -15,     41,     81,     89,     62,     11,    -45,    -83,    -88,    -59,     -7,     48,     84,     87,     56,      2,    -52,    -86,    -86,    -52,      2,     56,     87,     84,     48,     -7,    -59,    -88,    -83,    -45,     11,     62,     89,     81,     41,    -15,    -66,    -90,    -79,    -37,     20,     69,     90,     76,     33,    -24,    -71,    -90,    -74,    -28,     28,     74,     90,     71,     24,    -33,    -76,    -90,    -69,    -20,     37,     79,     90,     66,     15,    -41,    -81,    -89,    -62,    -11,     45,     83,     88,     59,      7,    -48,    -84,    -87,    -56,     -2,     52,     86 },
      {     86,     49,     -8,    -62,    -89,    -79,    -36,     23,     72,     91,     71,     21,    -38,    -80,    -89,    -60,     -6,     51,     86,     85,     47,    -10,    -63,    -90,    -78,    -34,     25,     73,     90,     69,     19,    -40,    -81,    -89,    -58,     -3,     53,     87,     84,     46,    -12,    -65,    -90,    -77,    -32,     27,     75,     90,     68,     17,    -42,    -82,    -88,    -57,     -1,     55,     88,     83,     44,    -14,    -66,    -90,    -76,    -29,     29,     76,     90,     66,     14,    -44,    -83,    -88,    -55,      1,     57,     88,     82,     42,    -17,    -68,    -90,    -75,    -27,     32,     77,     90,     65,     12,    -46,    -84,    -87,    -53,      3,     58,     89,     81,     40,    -19,    -69,    -90,    -73,    -25,     34,     78,     90,     63,     10,    -47,    -85,    -86,    -51,      6,     60,     89,     80,     38,    -21,    -71,    -91,    -72,    -23,     36,     79,     89,     62,      8,    -49,    -86 },
      {     85,     47,    -13,    -67,    -90,    -73,    -22,     39,     82,     88,     54,     -4,    -61,    -90,    -78,    -30,     30,     78,     90,     61,      4,    -54,    -88,    -82,    -39,     22,     73,     90,     67,     13,    -47,    -85,    -85,    -47,     13,     67,     90,     73,     22,    -39,    -82,    -88,    -54,      4,     61,     90,     78,     30,    -30,    -78,    -90,    -61,     -4,     54,     88,     82,     39,    -22,    -73,    -90,    -67,    -13,     47,     85,     85,     47,    -13,    -67,    -90,    -73,    -22,     39,     82,     88,     54,     -4,    -61,    -90,    -78,    -30,     30,     78,     90,     61,      4,    -54,    -88,    -82,    -39,     22,     73,     90,     67,     13,    -47,    -85,    -85,    -47,     13,     67,     90,     73,     22,    -39,    -82,    -88,    -54,      4,     61,     90,     78,     30,    -30,    -78,    -90,    -61,     -4,     54,     88,     82,     39,    -22,    -73,    -90,    -67,    -13,     47,     85 },
      {     85,     44,    -19,    -72,    -90,    -65,     -8,     53,     88,     80,     34,    -29,    -78,    -89,    -57,      3,     62,     90,     75,     23,    -40,    -83,    -86,    -47,     14,     69,     91,     68,     12,    -49,    -87,    -82,    -38,     25,     76,     90,     60,      1,    -58,    -89,    -77,    -27,     36,     81,     88,     51,    -10,    -66,    -90,    -71,    -17,     46,     86,     84,     42,    -21,    -73,    -90,    -63,     -6,     55,     89,     79,     32,    -32,    -79,    -89,    -55,      6,     63,     90,     73,     21,    -42,    -84,    -86,    -46,     17,     71,     90,     66,     10,    -51,    -88,    -81,    -36,     27,     77,     89,     58,     -1,    -60,    -90,    -76,    -25,     38,     82,     87,     49,    -12,    -68,    -91,    -69,    -14,     47,     86,     83,     40,    -23,    -75,    -90,    -62,     -3,     57,     89,     78,     29,    -34,    -80,    -88,    -53,      8,     65,     90,     72,     19,    -44,    -85 },
      {     84,     41,    -24,    -76,    -89,    -56,      7,     66,     90,     69,     11,    -52,    -88,    -79,    -28,     37,     83,     86,     45,    -20,    -74,    -90,    -59,      2,     62,     90,     71,     15,    -48,    -87,    -81,    -33,     33,     81,     87,     48,    -15,    -71,    -90,    -62,     -2,     59,     90,     74,     20,    -45,    -86,    -83,    -37,     28,     79,     88,     52,    -11,    -69,    -90,    -66,     -7,     56,     89,     76,     24,    -41,    -84,    -84,    -41,     24,     76,     89,     56,     -7,    -66,    -90,    -69,    -11,     52,     88,     79,     28,    -37,    -83,    -86,    -45,     20,     74,     90,     59,     -2,    -62,    -90,    -71,    -15,     48,     87,     81,     33,    -33,    -81,    -87,    -48,     15,     71,     90,     62,      2,    -59,    -90,    -74,    -20,     45,     86,     83,     37,    -28,    -79,    -88,    -52,     11,     69,     90,     66,      7,    -56,    -89,    -76,    -24,     41,     84 },
      {     84,     38,    -29,    -80,    -87,    -46,     21,     76,     89,     53,    -12,    -71,    -90,    -60,      3,     65,     91,     66,      6,    -58,    -90,    -72,    -14,     51,     89,     77,     23,    -44,    -86,    -81,    -32,     36,     83,     85,     40,    -27,    -79,    -88,    -47,     19,     75,     89,     55,    -10,    -69,    -90,    -62,      1,     63,     90,     68,      8,    -57,    -90,    -73,    -17,     49,     88,     78,     25,    -42,    -86,    -82,    -34,     34,     82,     86,     42,    -25,    -78,    -88,    -49,     17,     73,     90,     57,     -8,    -68,    -90,    -63,     -1,     62,     90,     69,     10,    -55,    -89,    -75,    -19,     47,     88,     79,     27,    -40,    -85,    -83,    -36,     32,     81,     86,     44,    -23,    -77,    -89,    -51,     14,     72,     90,     58,     -6,    -66,    -91,    -65,     -3,     60,     90,     71,     12,    -53,    -89,    -76,    -21,     46,     87,     80,     29,    -38,    -84 },
      {     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84,     84,     35,    -35,    -84,    -84,    -35,     35,     84 },
      {     83,     32,    -40,    -86,    -79,    -23,     47,     89,     75,     14,    -55,    -90,    -69,     -6,     62,     91,     63,     -3,    -68,    -90,    -57,     12,     73,     89,     49,    -21,    -78,    -87,    -42,     29,     82,     84,     34,    -38,    -86,    -80,    -25,     46,     88,     76,     17,    -53,    -90,    -71,     -8,     60,     90,     65,     -1,    -66,    -90,    -58,     10,     72,     89,     51,    -19,    -77,    -88,    -44,     27,     81,     85,     36,    -36,    -85,    -81,    -27,     44,     88,     77,     19,    -51,    -89,    -72,    -10,     58,     90,     66,      1,    -65,    -90,    -60,      8,     71,     90,     53,    -17,    -76,    -88,    -46,     25,     80,     86,     38,    -34,    -84,    -82,    -29,     42,     87,     78,     21,    -49,    -89,    -73,    -12,     57,     90,     68,      3,    -63,    -91,    -62,      6,     69,     90,     55,    -14,    -75,    -89,    -47,     23,     79,     86,     40,    -32,    -83 },
      {     83,     28,    -45,    -88,    -74,    -11,     59,     90,     62,     -7,    -71,    -89,    -48,     24,     81,     84,     33,    -41,    -87,    -76,    -15,     56,     90,     66,     -2,    -69,    -90,    -52,     20,     79,     86,     37,    -37,    -86,    -79,    -20,     52,     90,     69,      2,    -66,    -90,    -56,     15,     76,     87,     41,    -33,    -84,    -81,    -24,     48,     89,     71,      7,    -62,    -90,    -59,     11,     74,     88,     45,    -28,    -83,    -83,    -28,     45,     88,     74,     11,    -59,    -90,    -62,      7,     71,     89,     48,    -24,    -81,    -84,    -33,     41,     87,     76,     15,    -56,    -90,    -66,      2,     69,     90,     52,    -20,    -79,    -86,    -37,     37,     86,     79,     20,    -52,    -90,    -69,     -2,     66,     90,     56,    -15,    -76,    -87,    -41,     33,     84,     81,     24,    -48,    -89,    -71,     -7,     62,     90,     59,    -11,    -74,    -88,    -45,     28,     83 },
      {     82,     25,    -49,    -90,    -68,      1,     69,     89,     47,    -27,    -83,    -81,    -23,     51,     90,     66,     -3,    -71,    -89,    -46,     29,     84,     80,     21,    -53,    -90,    -65,      6,     72,     89,     44,    -32,    -85,    -79,    -19,     55,     90,     63,     -8,    -73,    -88,    -42,     34,     86,     78,     17,    -57,    -90,    -62,     10,     75,     88,     40,    -36,    -86,    -77,    -14,     58,     91,     60,    -12,    -76,    -87,    -38,     38,     87,     76,     12,    -60,    -91,    -58,     14,     77,     86,     36,    -40,    -88,    -75,    -10,     62,     90,     57,    -17,    -78,    -86,    -34,     42,     88,     73,      8,    -63,    -90,    -55,     19,     79,     85,     32,    -44,    -89,    -72,     -6,     65,     90,     53,    -21,    -80,    -84,    -29,     46,     89,     71,      3,    -66,    -90,    -51,     23,     81,     83,     27,    -47,    -89,    -69,     -1,     68,     90,     49,    -25,    -82 },
      {     82,     22,    -54,    -90,    -61,     13,     78,     85,     30,    -47,    -90,    -67,      4,     73,     88,     39,    -39,    -88,    -73,     -4,     67,     90,     47,    -30,    -85,    -78,    -13,     61,     90,     54,    -22,    -82,    -82,    -22,     54,     90,     61,    -13,    -78,    -85,    -30,     47,     90,     67,     -4,    -73,    -88,    -39,     39,     88,     73,      4,    -67,    -90,    -47,     30,     85,     78,     13,    -61,    -90,    -54,     22,     82,     82,     22,    -54,    -90,    -61,     13,     78,     85,     30,    -47,    -90,    -67,      4,     73,     88,     39,    -39,    -88,    -73,     -4,     67,     90,     47,    -30,    -85,    -78,    -13,     61,     90,     54,    -22,    -82,    -82,    -22,     54,     90,     61,    -13,    -78,    -85,    -30,     47,     90,     67,     -4,    -73,    -88,    -39,     39,     88,     73,      4,    -67,    -90,    -47,     30,     85,     78,     13,    -61,    -90,    -54,     22,     82 },
      {     81,     19,    -58,    -90,    -53,     25,     84,     78,     12,    -63,    -90,    -47,     32,     86,     75,      6,    -68,    -89,    -42,     38,     88,     71,     -1,    -72,    -88,    -36,     44,     89,     66,     -8,    -76,    -86,    -29,     49,     90,     62,    -14,    -79,    -83,    -23,     55,     91,     57,    -21,    -82,    -80,    -17,     60,     90,     51,    -27,    -85,    -77,    -10,     65,     90,     46,    -34,    -87,    -73,     -3,     69,     89,     40,    -40,    -89,    -69,      3,     73,     87,     34,    -46,    -90,    -65,     10,     77,     85,     27,    -51,    -90,    -60,     17,     80,     82,     21,    -57,    -91,    -55,     23,     83,     79,     14,    -62,    -90,    -49,     29,     86,     76,      8,    -66,    -89,    -44,     36,     88,     72,      1,    -71,    -88,    -38,     42,     89,     68,     -6,    -75,    -86,    -32,     47,     90,     63,    -12,    -78,    -84,    -25,     53,     90,     58,    -19,    -81 },
      {     81,     15,    -62,    -90,    -45,     37,     88,     69,     -7,    -76,    -84,    -24,     56,     90,     52,    -28,    -86,    -74,     -2,     71,     87,     33,    -48,    -90,    -59,     20,     83,     79,     11,    -66,    -89,    -41,     41,     89,     66,    -11,    -79,    -83,    -20,     59,     90,     48,    -33,    -87,    -71,      2,     74,     86,     28,    -52,    -90,    -56,     24,     84,     76,      7,    -69,    -88,    -37,     45,     90,     62,    -15,    -81,    -81,    -15,     62,     90,     45,    -37,    -88,    -69,      7,     76,     84,     24,    -56,    -90,    -52,     28,     86,     74,      2,    -71,    -87,    -33,     48,     90,     59,    -20,    -83,    -79,    -11,     66,     89,     41,    -41,    -89,    -66,     11,     79,     83,     20,    -59,    -90,    -48,     33,     87,     71,     -2,    -74,    -86,    -28,     52,     90,     56,    -24,    -84,    -76,     -7,     69,     88,     37,    -45,    -90,    -62,     15,     81 },
      {     80,     12,    -66,    -89,    -36,     47,     90,     57,    -25,    -86,    -73,      1,     75,     85,     23,    -58,    -90,    -46,     38,     89,     65,    -14,    -81,    -79,    -10,     68,     88,     34,    -49,    -90,    -55,     27,     86,     72,     -3,    -76,    -84,    -21,     60,     90,     44,    -40,    -89,    -63,     17,     82,     78,      8,    -69,    -88,    -32,     51,     91,     53,    -29,    -87,    -71,      6,     77,     83,     19,    -62,    -90,    -42,     42,     90,     62,    -19,    -83,    -77,     -6,     71,     87,     29,    -53,    -91,    -51,     32,     88,     69,     -8,    -78,    -82,    -17,     63,     89,     40,    -44,    -90,    -60,     21,     84,     76,      3,    -72,    -86,    -27,     55,     90,     49,    -34,    -88,    -68,     10,     79,     81,     14,    -65,    -89,    -38,     46,     90,     58,    -23,    -85,    -75,     -1,     73,     86,     25,    -57,    -90,    -47,     36,     89,     66,    -12,    -80 },
      {     80,      9,    -70,    -87,    -26,     57,     90,     43,    -43,    -90,    -57,     26,     87,     70,     -9,    -80,    -80,     -9,     70,     87,     26,    -57,    -90,    -43,     43,     90,     57,    -26,    -87,    -70,      9,     80,     80,      9,    -70,    -87,    -26,     57,     90,     43,    -43,    -90,    -57,     26,     87,     70,     -9,    -80,    -80,     -9,     70,     87,     26,    -57,    -90,    -43,     43,     90,     57,    -26,    -87,    -70,      9,     80,     80,      9,    -70,    -87,    -26,     57,     90,     43,    -43,    -90,    -57,     26,     87,     70,     -9,    -80,    -80,     -9,     70,     87,     26,    -57,    -90,    -43,     43,     90,     57,    -26,    -87,    -70,      9,     80,     80,      9,    -70,    -87,    -26,     57,     90,     43,    -43,    -90,    -57,     26,     87,     70,     -9,    -80,    -80,     -9,     70,     87,     26,    -57,    -90,    -43,     43,     90,     57,    -26,    -87,    -70,      9,     80 },
      {     79,      6,    -73,    -84,    -17,     66,     88,     27,    -58,    -90,    -38,     49,     91,     47,    -40,    -90,    -57,     29,     88,     65,    -19,    -85,    -72,      8,     80,     78,      3,    -75,    -83,    -14,     68,     87,     25,    -60,    -89,    -36,     51,     90,     46,    -42,    -90,    -55,     32,     89,     63,    -21,    -86,    -71,     10,     81,     77,      1,    -76,    -82,    -12,     69,     86,     23,    -62,    -89,    -34,     53,     90,     44,    -44,    -90,    -53,     34,     89,     62,    -23,    -86,    -69,     12,     82,     76,     -1,    -77,    -81,    -10,     71,     86,     21,    -63,    -89,    -32,     55,     90,     42,    -46,    -90,    -51,     36,     89,     60,    -25,    -87,    -68,     14,     83,     75,     -3,    -78,    -80,     -8,     72,     85,     19,    -65,    -88,    -29,     57,     90,     40,    -47,    -91,    -49,     38,     90,     58,    -27,    -88,    -66,     17,     84,     73,     -6,    -79 },
      {     79,      2,    -76,    -81,     -7,     74,     83,     11,    -71,    -84,    -15,     69,     86,     20,    -66,    -87,    -24,     62,     88,     28,    -59,    -89,    -33,     56,     90,     37,    -52,    -90,    -41,     48,     90,     45,    -45,    -90,    -48,     41,     90,     52,    -37,    -90,    -56,     33,     89,     59,    -28,    -88,    -62,     24,     87,     66,    -20,    -86,    -69,     15,     84,     71,    -11,    -83,    -74,      7,     81,     76,     -2,    -79,    -79,     -2,     76,     81,      7,    -74,    -83,    -11,     71,     84,     15,    -69,    -86,    -20,     66,     87,     24,    -62,    -88,    -28,     59,     89,     33,    -56,    -90,    -37,     52,     90,     41,    -48,    -90,    -45,     45,     90,     48,    -41,    -90,    -52,     37,     90,     56,    -33,    -89,    -59,     28,     88,     62,    -24,    -87,    -66,     20,     86,     69,    -15,    -84,    -71,     11,     83,     74,     -7,    -81,    -76,      2,     79 },
      {     78,     -1,    -79,    -77,      3,     80,     76,     -6,    -81,    -75,      8,     82,     73,    -10,    -83,    -72,     12,     84,     71,    -14,    -85,    -69,     17,     86,     68,    -19,    -86,    -66,     21,     87,     65,    -23,    -88,    -63,     25,     88,     62,    -27,    -89,    -60,     29,     89,     58,    -32,    -89,    -57,     34,     90,     55,    -36,    -90,    -53,     38,     90,     51,    -40,    -90,    -49,     42,     90,     47,    -44,    -91,    -46,     46,     91,     44,    -47,    -90,    -42,     49,     90,     40,    -51,    -90,    -38,     53,     90,     36,    -55,    -90,    -34,     57,     89,     32,    -58,    -89,    -29,     60,     89,     27,    -62,    -88,    -25,     63,     88,     23,    -65,    -87,    -21,     66,     86,     19,    -68,    -86,    -17,     69,     85,     14,    -71,    -84,    -12,     72,     83,     10,    -73,    -82,     -8,     75,     81,      6,    -76,    -80,     -3,     77,     79,      1,    -78 },
      {     78,     -4,    -82,    -73,     13,     85,     67,    -22,    -88,    -61,     30,     90,     54,    -39,    -90,    -47,     47,     90,     39,    -54,    -90,    -30,     61,     88,     22,    -67,    -85,    -13,     73,     82,      4,    -78,    -78,      4,     82,     73,    -13,    -85,    -67,     22,     88,     61,    -30,    -90,    -54,     39,     90,     47,    -47,    -90,    -39,     54,     90,     30,    -61,    -88,    -22,     67,     85,     13,    -73,    -82,     -4,     78,     78,     -4,    -82,    -73,     13,     85,     67,    -22,    -88,    -61,     30,     90,     54,    -39,    -90,    -47,     47,     90,     39,    -54,    -90,    -30,     61,     88,     22,    -67,    -85,    -13,     73,     82,      4,    -78,    -78,      4,     82,     73,    -13,    -85,    -67,     22,     88,     61,    -30,    -90,    -54,     39,     90,     47,    -47,    -90,    -39,     54,     90,     30,    -61,    -88,    -22,     67,     85,     13,    -73,    -82,     -4,     78 },
      {     77,     -8,    -84,    -68,     23,     89,     57,    -38,    -90,    -44,     51,     90,     29,    -63,    -86,    -14,     73,     80,     -1,    -81,    -72,     17,     87,     62,    -32,    -90,    -49,     46,     90,     36,    -58,    -88,    -21,     69,     83,      6,    -78,    -76,     10,     85,     66,    -25,    -89,    -55,     40,     91,     42,    -53,    -89,    -27,     65,     86,     12,    -75,    -79,      3,     82,     71,    -19,    -88,    -60,     34,     90,     47,    -47,    -90,    -34,     60,     88,     19,    -71,    -82,     -3,     79,     75,    -12,    -86,    -65,     27,     89,     53,    -42,    -91,    -40,     55,     89,     25,    -66,    -85,    -10,     76,     78,     -6,    -83,    -69,     21,     88,     58,    -36,    -90,    -46,     49,     90,     32,    -62,    -87,    -17,     72,     81,      1,    -80,    -73,     14,     86,     63,    -29,    -90,    -51,     44,     90,     38,    -57,    -89,    -23,     68,     84,      8,    -77 },
      {     76,    -11,    -86,    -62,     33,     90,     45,    -52,    -89,    -24,     69,     83,      2,    -81,    -71,     20,     88,     56,    -41,    -90,    -37,     59,     87,     15,    -74,    -79,      7,     84,     66,    -28,    -90,    -48,     48,     90,     28,    -66,    -84,     -7,     79,     74,    -15,    -87,    -59,     37,     90,     41,    -56,    -88,    -20,     71,     81,     -2,    -83,    -69,     24,     89,     52,    -45,    -90,    -33,     62,     86,     11,    -76,    -76,     11,     86,     62,    -33,    -90,    -45,     52,     89,     24,    -69,    -83,     -2,     81,     71,    -20,    -88,    -56,     41,     90,     37,    -59,    -87,    -15,     74,     79,     -7,    -84,    -66,     28,     90,     48,    -48,    -90,    -28,     66,     84,      7,    -79,    -74,     15,     87,     59,    -37,    -90,    -41,     56,     88,     20,    -71,    -81,      2,     83,     69,    -24,    -89,    -52,     45,     90,     33,    -62,    -86,    -11,     76 },
      {     76,    -14,    -88,    -57,     42,     90,     32,    -65,    -84,     -3,     81,     69,    -25,    -90,    -47,     51,     89,     21,    -72,    -79,      8,     86,     62,    -36,    -91,    -38,     60,     86,     10,    -78,    -73,     19,     89,     53,    -46,    -90,    -27,     68,     82,     -1,    -83,    -66,     29,     90,     44,    -55,    -88,    -17,     75,     77,    -12,    -87,    -58,     40,     90,     34,    -63,    -85,     -6,     80,     71,    -23,    -89,    -49,     49,     89,     23,    -71,    -80,      6,     85,     63,    -34,    -90,    -40,     58,     87,     12,    -77,    -75,     17,     88,     55,    -44,    -90,    -29,     66,     83,      1,    -82,    -68,     27,     90,     46,    -53,    -89,    -19,     73,     78,    -10,    -86,    -60,     38,     91,     36,    -62,    -86,     -8,     79,     72,    -21,    -89,    -51,     47,     90,     25,    -69,    -81,      3,     84,     65,    -32,    -90,    -42,     57,     88,     14,    -76 },
      {     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75,     75,    -18,    -89,    -50,     50,     89,     18,    -75,    -75,     18,     89,     50,    -50,    -89,    -18,     75 },
      {     75,    -21,    -90,    -44,     58,     86,      3,    -83,    -63,     38,     90,     27,    -71,    -78,     14,     89,     49,    -53,    -88,    -10,     80,     68,    -32,    -91,    -34,     66,     81,     -8,    -87,    -55,     47,     89,     17,    -77,    -72,     25,     90,     40,    -62,    -84,      1,     85,     60,    -42,    -90,    -23,     73,     76,    -19,    -89,    -46,     57,     86,      6,    -82,    -65,     36,     90,     29,    -69,    -79,     12,     88,     51,    -51,    -88,    -12,     79,     69,    -29,    -90,    -36,     65,     82,     -6,    -86,    -57,     46,     89,     19,    -76,    -73,     23,     90,     42,    -60,    -85,     -1,     84,     62,    -40,    -90,    -25,     72,     77,    -17,    -89,    -47,     55,     87,      8,    -81,    -66,     34,     91,     32,    -68,    -80,     10,     88,     53,    -49,    -89,    -14,     78,     71,    -27,    -90,    -38,     63,     83,     -3,    -86,    -58,     44,     90,     21,    -75 },
      {     74,    -24,    -90,    -37,     66,     81,    -11,    -88,    -48,     56,     86,      2,    -84,    -59,     45,     89,     15,    -79,    -69,     33,     90,     28,    -71,    -76,     20,     90,     41,    -62,    -83,      7,     87,     52,    -52,    -87,     -7,     83,     62,    -41,    -90,    -20,     76,     71,    -28,    -90,    -33,     69,     79,    -15,    -89,    -45,     59,     84,     -2,    -86,    -56,     48,     88,     11,    -81,    -66,     37,     90,     24,    -74,    -74,     24,     90,     37,    -66,    -81,     11,     88,     48,    -56,    -86,     -2,     84,     59,    -45,    -89,    -15,     79,     69,    -33,    -90,    -28,     71,     76,    -20,    -90,    -41,     62,     83,     -7,    -87,    -52,     52,     87,      7,    -83,    -62,     41,     90,     20,    -76,    -71,     28,     90,     33,    -69,    -79,     15,     89,     45,    -59,    -84,      2,     86,     56,    -48,    -88,    -11,     81,     66,    -37,    -90,    -24,     74 },
      {     73,    -27,    -91,    -29,     72,     75,    -25,    -90,    -32,     71,     76,    -23,    -90,    -34,     69,     77,    -21,    -90,    -36,     68,     78,    -19,    -90,    -38,     66,     79,    -17,    -90,    -40,     65,     80,    -14,    -89,    -42,     63,     81,    -12,    -89,    -44,     62,     82,    -10,    -89,    -46,     60,     83,     -8,    -88,    -47,     58,     84,     -6,    -88,    -49,     57,     85,     -3,    -87,    -51,     55,     86,     -1,    -86,    -53,     53,     86,      1,    -86,    -55,     51,     87,      3,    -85,    -57,     49,     88,      6,    -84,    -58,     47,     88,      8,    -83,    -60,     46,     89,     10,    -82,    -62,     44,     89,     12,    -81,    -63,     42,     89,     14,    -80,    -65,     40,     90,     17,    -79,    -66,     38,     90,     19,    -78,    -68,     36,     90,     21,    -77,    -69,     34,     90,     23,    -76,    -71,     32,     90,     25,    -75,    -72,     29,     91,     27,    -73 },
      {     73,    -30,    -90,    -22,     78,     67,    -39,    -90,    -13,     82,     61,    -47,    -88,     -4,     85,     54,    -54,    -85,      4,     88,     47,    -61,    -82,     13,     90,     39,    -67,    -78,     22,     90,     30,    -73,    -73,     30,     90,     22,    -78,    -67,     39,     90,     13,    -82,    -61,     47,     88,      4,    -85,    -54,     54,     85,     -4,    -88,    -47,     61,     82,    -13,    -90,    -39,     67,     78,    -22,    -90,    -30,     73,     73,    -30,    -90,    -22,     78,     67,    -39,    -90,    -13,     82,     61,    -47,    -88,     -4,     85,     54,    -54,    -85,      4,     88,     47,    -61,    -82,     13,     90,     39,    -67,    -78,     22,     90,     30,    -73,    -73,     30,     90,     22,    -78,    -67,     39,     90,     13,    -82,    -61,     47,     88,      4,    -85,    -54,     54,     85,     -4,    -88,    -47,     61,     82,    -13,    -90,    -39,     67,     78,    -22,    -90,    -30,     73 },
      {     72,    -34,    -90,    -14,     82,     58,    -51,    -86,      6,     89,     42,    -66,    -77,     25,     91,     23,    -78,    -65,     44,     88,      3,    -86,    -49,     60,     81,    -17,    -90,    -32,     73,     71,    -36,    -90,    -12,     83,     57,    -53,    -85,      8,     89,     40,    -68,    -76,     27,     90,     21,    -79,    -63,     46,     88,      1,    -87,    -47,     62,     80,    -19,    -90,    -29,     75,     69,    -38,    -89,    -10,     84,     55,    -55,    -84,     10,     89,     38,    -69,    -75,     29,     90,     19,    -80,    -62,     47,     87,     -1,    -88,    -46,     63,     79,    -21,    -90,    -27,     76,     68,    -40,    -89,     -8,     85,     53,    -57,    -83,     12,     90,     36,    -71,    -73,     32,     90,     17,    -81,    -60,     49,     86,     -3,    -88,    -44,     65,     78,    -23,    -91,    -25,     77,     66,    -42,    -89,     -6,     86,     51,    -58,    -82,     14,     90,     34,    -72 },
      {     71,    -37,    -89,     -7,     86,     48,    -62,    -79,     24,     90,     20,    -81,    -59,     52,     84,    -11,    -90,    -33,     74,     69,    -41,    -88,     -2,     87,     45,    -66,    -76,     28,     90,     15,    -83,    -56,     56,     83,    -15,    -90,    -28,     76,     66,    -45,    -87,      2,     88,     41,    -69,    -74,     33,     90,     11,    -84,    -52,     59,     81,    -20,    -90,    -24,     79,     62,    -48,    -86,      7,     89,     37,    -71,    -71,     37,     89,      7,    -86,    -48,     62,     79,    -24,    -90,    -20,     81,     59,    -52,    -84,     11,     90,     33,    -74,    -69,     41,     88,      2,    -87,    -45,     66,     76,    -28,    -90,    -15,     83,     56,    -56,    -83,     15,     90,     28,    -76,    -66,     45,     87,     -2,    -88,    -41,     69,     74,    -33,    -90,    -11,     84,     52,    -59,    -81,     20,     90,     24,    -79,    -62,     48,     86,     -7,    -89,    -37,     71 },
      {     71,    -40,    -88,      1,     89,     38,    -72,    -69,     42,     88,     -3,    -89,    -36,     73,     68,    -44,    -87,      6,     89,     34,    -75,    -66,     46,     86,     -8,    -90,    -32,     76,     65,    -47,    -86,     10,     90,     29,    -77,    -63,     49,     85,    -12,    -90,    -27,     78,     62,    -51,    -84,     14,     90,     25,    -79,    -60,     53,     83,    -17,    -90,    -23,     80,     58,    -55,    -82,     19,     91,     21,    -81,    -57,     57,     81,    -21,    -91,    -19,     82,     55,    -58,    -80,     23,     90,     17,    -83,    -53,     60,     79,    -25,    -90,    -14,     84,     51,    -62,    -78,     27,     90,     12,    -85,    -49,     63,     77,    -29,    -90,    -10,     86,     47,    -65,    -76,     32,     90,      8,    -86,    -46,     66,     75,    -34,    -89,     -6,     87,     44,    -68,    -73,     36,     89,      3,    -88,    -42,     69,     72,    -38,    -89,     -1,     88,     40,    -71 },
      {     70,    -43,    -87,      9,     90,     26,    -80,    -57,     57,     80,    -26,    -90,     -9,     87,     43,    -70,    -70,     43,     87,     -9,    -90,    -26,     80,     57,    -57,    -80,     26,     90,      9,    -87,    -43,     70,     70,    -43,    -87,      9,     90,     26,    -80,    -57,     57,     80,    -26,    -90,     -9,     87,     43,    -70,    -70,     43,     87,     -9,    -90,    -26,     80,     57,    -57,    -80,     26,     90,      9,    -87,    -43,     70,     70,    -43,    -87,      9,     90,     26,    -80,    -57,     57,     80,    -26,    -90,     -9,     87,     43,    -70,    -70,     43,     87,     -9,    -90,    -26,     80,     57,    -57,    -80,     26,     90,      9,    -87,    -43,     70,     70,    -43,    -87,      9,     90,     26,    -80,    -57,     57,     80,    -26,    -90,     -9,     87,     43,    -70,    -70,     43,     87,     -9,    -90,    -26,     80,     57,    -57,    -80,     26,     90,      9,    -87,    -43,     70 },
      {     69,    -46,    -85,     17,     91,     14,    -86,    -44,     71,     68,    -47,    -84,     19,     90,     12,    -86,    -42,     72,     66,    -49,    -83,     21,     90,     10,    -87,    -40,     73,     65,    -51,    -82,     23,     90,      8,    -88,    -38,     75,     63,    -53,    -81,     25,     90,      6,    -88,    -36,     76,     62,    -55,    -80,     27,     90,      3,    -89,    -34,     77,     60,    -57,    -79,     29,     89,      1,    -89,    -32,     78,     58,    -58,    -78,     32,     89,     -1,    -89,    -29,     79,     57,    -60,    -77,     34,     89,     -3,    -90,    -27,     80,     55,    -62,    -76,     36,     88,     -6,    -90,    -25,     81,     53,    -63,    -75,     38,     88,     -8,    -90,    -23,     82,     51,    -65,    -73,     40,     87,    -10,    -90,    -21,     83,     49,    -66,    -72,     42,     86,    -12,    -90,    -19,     84,     47,    -68,    -71,     44,     86,    -14,    -91,    -17,     85,     46,    -69 },
      {     69,    -48,    -83,     24,     90,      2,    -89,    -28,     81,     52,    -66,    -71,     45,     84,    -20,    -90,     -7,     88,     33,    -79,    -56,     62,     74,    -41,    -86,     15,     90,     11,    -87,    -37,     76,     59,    -59,    -76,     37,     87,    -11,    -90,    -15,     86,     41,    -74,    -62,     56,     79,    -33,    -88,      7,     90,     20,    -84,    -45,     71,     66,    -52,    -81,     28,     89,     -2,    -90,    -24,     83,     48,    -69,    -69,     48,     83,    -24,    -90,     -2,     89,     28,    -81,    -52,     66,     71,    -45,    -84,     20,     90,      7,    -88,    -33,     79,     56,    -62,    -74,     41,     86,    -15,    -90,    -11,     87,     37,    -76,    -59,     59,     76,    -37,    -87,     11,     90,     15,    -86,    -41,     74,     62,    -56,    -79,     33,     88,     -7,    -90,    -20,     84,     45,    -71,    -66,     52,     81,    -28,    -89,      2,     90,     24,    -83,    -48,     69 },
      {     68,    -51,    -80,     32,     88,    -10,    -91,    -12,     88,     34,    -79,    -53,     66,     69,    -49,    -81,     29,     89,     -8,    -90,    -14,     87,     36,    -78,    -55,     65,     71,    -47,    -82,     27,     89,     -6,    -90,    -17,     86,     38,    -77,    -57,     63,     72,    -46,    -83,     25,     89,     -3,    -90,    -19,     86,     40,    -76,    -58,     62,     73,    -44,    -84,     23,     90,     -1,    -90,    -21,     85,     42,    -75,    -60,     60,     75,    -42,    -85,     21,     90,      1,    -90,    -23,     84,     44,    -73,    -62,     58,     76,    -40,    -86,     19,     90,      3,    -89,    -25,     83,     46,    -72,    -63,     57,     77,    -38,    -86,     17,     90,      6,    -89,    -27,     82,     47,    -71,    -65,     55,     78,    -36,    -87,     14,     90,      8,    -89,    -29,     81,     49,    -69,    -66,     53,     79,    -34,    -88,     12,     91,     10,    -88,    -32,     80,     51,    -68 },
      {     67,    -54,    -78,     39,     85,    -22,    -90,      4,     90,     13,    -88,    -30,     82,     47,    -73,    -61,     61,     73,    -47,    -82,     30,     88,    -13,    -90,     -4,     90,     22,    -85,    -39,     78,     54,    -67,    -67,     54,     78,    -39,    -85,     22,     90,     -4,    -90,    -13,     88,     30,    -82,    -47,     73,     61,    -61,    -73,     47,     82,    -30,    -88,     13,     90,      4,    -90,    -22,     85,     39,    -78,    -54,     67,     67,    -54,    -78,     39,     85,    -22,    -90,      4,     90,     13,    -88,    -30,     82,     47,    -73,    -61,     61,     73,    -47,    -82,     30,     88,    -13,    -90,     -4,     90,     22,    -85,    -39,     78,     54,    -67,    -67,     54,     78,    -39,    -85,     22,     90,     -4,    -90,    -13,     88,     30,    -82,    -47,     73,     61,    -61,    -73,     47,     82,    -30,    -88,     13,     90,      4,    -90,    -22,     85,     39,    -78,    -54,     67 },
      {     66,    -57,    -75,     46,     81,    -34,    -86,     21,     89,     -8,    -91,     -6,     90,     19,    -87,    -32,     82,     44,    -76,    -55,     68,     65,    -58,    -73,     47,     80,    -36,    -86,     23,     89,    -10,    -90,     -3,     90,     17,    -88,    -29,     83,     42,    -77,    -53,     69,     63,    -60,    -72,     49,     79,    -38,    -85,     25,     89,    -12,    -90,     -1,     90,     14,    -88,    -27,     84,     40,    -78,    -51,     71,     62,    -62,    -71,     51,     78,    -40,    -84,     27,     88,    -14,    -90,      1,     90,     12,    -89,    -25,     85,     38,    -79,    -49,     72,     60,    -63,    -69,     53,     77,    -42,    -83,     29,     88,    -17,    -90,      3,     90,     10,    -89,    -23,     86,     36,    -80,    -47,     73,     58,    -65,    -68,     55,     76,    -44,    -82,     32,     87,    -19,    -90,      6,     91,      8,    -89,    -21,     86,     34,    -81,    -46,     75,     57,    -66 },
      {     66,    -59,    -71,     52,     76,    -45,    -81,     37,     84,    -28,    -87,     20,     89,    -11,    -90,      2,     90,      7,    -90,    -15,     88,     24,    -86,    -33,     83,     41,    -79,    -48,     74,     56,    -69,    -62,     62,     69,    -56,    -74,     48,     79,    -41,    -83,     33,     86,    -24,    -88,     15,     90,     -7,    -90,     -2,     90,     11,    -89,    -20,     87,     28,    -84,    -37,     81,     45,    -76,    -52,     71,     59,    -66,    -66,     59,     71,    -52,    -76,     45,     81,    -37,    -84,     28,     87,    -20,    -89,     11,     90,     -2,    -90,     -7,     90,     15,    -88,    -24,     86,     33,    -83,    -41,     79,     48,    -74,    -56,     69,     62,    -62,    -69,     56,     74,    -48,    -79,     41,     83,    -33,    -86,     24,     88,    -15,    -90,      7,     90,      2,    -90,    -11,     89,     20,    -87,    -28,     84,     37,    -81,    -45,     76,     52,    -71,    -59,     66 },
      {     65,    -62,    -68,     58,     71,    -55,    -73,     51,     76,    -47,    -78,     44,     80,    -40,    -82,     36,     84,    -32,    -86,     27,     87,    -23,    -88,     19,     89,    -14,    -90,     10,     90,     -6,    -90,      1,     91,      3,    -90,     -8,     90,     12,    -89,    -17,     89,     21,    -88,    -25,     86,     29,    -85,    -34,     83,     38,    -81,    -42,     79,     46,    -77,    -49,     75,     53,    -72,    -57,     69,     60,    -66,    -63,     63,     66,    -60,    -69,     57,     72,    -53,    -75,     49,     77,    -46,    -79,     42,     81,    -38,    -83,     34,     85,    -29,    -86,     25,     88,    -21,    -89,     17,     89,    -12,    -90,      8,     90,     -3,    -91,     -1,     90,      6,    -90,    -10,     90,     14,    -89,    -19,     88,     23,    -87,    -27,     86,     32,    -84,    -36,     82,     40,    -80,    -44,     78,     47,    -76,    -51,     73,     55,    -71,    -58,     68,     62,    -65 },
      {     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64,     64,    -64,    -64,     64 },
      {     63,    -66,    -60,     69,     57,    -72,    -53,     75,     49,    -77,    -46,     79,     42,    -81,    -38,     83,     34,    -85,    -29,     86,     25,    -88,    -21,     89,     17,    -89,    -12,     90,      8,    -90,     -3,     91,     -1,    -90,      6,     90,    -10,    -90,     14,     89,    -19,    -88,     23,     87,    -27,    -86,     32,     84,    -36,    -82,     40,     80,    -44,    -78,     47,     76,    -51,    -73,     55,     71,    -58,    -68,     62,     65,    -65,    -62,     68,     58,    -71,    -55,     73,     51,    -76,    -47,     78,     44,    -80,    -40,     82,     36,    -84,    -32,     86,     27,    -87,    -23,     88,     19,    -89,    -14,     90,     10,    -90,     -6,     90,      1,    -91,      3,     90,     -8,    -90,     12,     89,    -17,    -89,     21,     88,    -25,    -86,     29,     85,    -34,    -83,     38,     81,    -42,    -79,     46,     77,    -49,    -75,     53,     72,    -57,    -69,     60,     66,    -63 },
      {     62,    -69,    -56,     74,     48,    -79,    -41,     83,     33,    -86,    -24,     88,     15,    -90,     -7,     90,     -2,    -90,     11,     89,    -20,    -87,     28,     84,    -37,    -81,     45,     76,    -52,    -71,     59,     66,    -66,    -59,     71,     52,    -76,    -45,     81,     37,    -84,    -28,     87,     20,    -89,    -11,     90,      2,    -90,      7,     90,    -15,    -88,     24,     86,    -33,    -83,     41,     79,    -48,    -74,     56,     69,    -62,    -62,     69,     56,    -74,    -48,     79,     41,    -83,    -33,     86,     24,    -88,    -15,     90,      7,    -90,      2,     90,    -11,    -89,     20,     87,    -28,    -84,     37,     81,    -45,    -76,     52,     71,    -59,    -66,     66,     59,    -71,    -52,     76,     45,    -81,    -37,     84,     28,    -87,    -20,     89,     11,    -90,     -2,     90,     -7,    -90,     15,     88,    -24,    -86,     33,     83,    -41,    -79,     48,     74,    -56,    -69,     62 },
      {     62,    -71,    -51,     78,     40,    -84,    -27,     88,     14,    -90,     -1,     90,    -12,    -89,     25,     85,    -38,    -79,     49,     72,    -60,    -63,     69,     53,    -77,    -42,     83,     29,    -88,    -17,     90,      3,    -90,     10,     89,    -23,    -86,     36,     80,    -47,    -73,     58,     65,    -68,    -55,     76,     44,    -82,    -32,     87,     19,    -90,     -6,     91,     -8,    -89,     21,     86,    -34,    -81,     46,     75,    -57,    -66,     66,     57,    -75,    -46,     81,     34,    -86,    -21,     89,      8,    -91,      6,     90,    -19,    -87,     32,     82,    -44,    -76,     55,     68,    -65,    -58,     73,     47,    -80,    -36,     86,     23,    -89,    -10,     90,     -3,    -90,     17,     88,    -29,    -83,     42,     77,    -53,    -69,     63,     60,    -72,    -49,     79,     38,    -85,    -25,     89,     12,    -90,      1,     90,    -14,    -88,     27,     84,    -40,    -78,     51,     71,    -62 },
      {     61,    -73,    -47,     82,     30,    -88,    -13,     90,     -4,    -90,     22,     85,    -39,    -78,     54,     67,    -67,    -54,     78,     39,    -85,    -22,     90,      4,    -90,     13,     88,    -30,    -82,     47,     73,    -61,    -61,     73,     47,    -82,    -30,     88,     13,    -90,      4,     90,    -22,    -85,     39,     78,    -54,    -67,     67,     54,    -78,    -39,     85,     22,    -90,     -4,     90,    -13,    -88,     30,     82,    -47,    -73,     61,     61,    -73,    -47,     82,     30,    -88,    -13,     90,     -4,    -90,     22,     85,    -39,    -78,     54,     67,    -67,    -54,     78,     39,    -85,    -22,     90,      4,    -90,     13,     88,    -30,    -82,     47,     73,    -61,    -61,     73,     47,    -82,    -30,     88,     13,    -90,      4,     90,    -22,    -85,     39,     78,    -54,    -67,     67,     54,    -78,    -39,     85,     22,    -90,     -4,     90,    -13,    -88,     30,     82,    -47,    -73,     61 },
      {     60,    -75,    -42,     85,     21,    -90,      1,     90,    -23,    -84,     44,     73,    -62,    -58,     76,     40,    -86,    -19,     90,     -3,    -89,     25,     83,    -46,    -72,     63,     57,    -77,    -38,     86,     17,    -90,      6,     89,    -27,    -82,     47,     71,    -65,    -55,     78,     36,    -87,    -14,     90,     -8,    -89,     29,     81,    -49,    -69,     66,     53,    -79,    -34,     88,     12,    -91,     10,     88,    -32,    -80,     51,     68,    -68,    -51,     80,     32,    -88,    -10,     91,    -12,    -88,     34,     79,    -53,    -66,     69,     49,    -81,    -29,     89,      8,    -90,     14,     87,    -36,    -78,     55,     65,    -71,    -47,     82,     27,    -89,     -6,     90,    -17,    -86,     38,     77,    -57,    -63,     72,     46,    -83,    -25,     89,      3,    -90,     19,     86,    -40,    -76,     58,     62,    -73,    -44,     84,     23,    -90,     -1,     90,    -21,    -85,     42,     75,    -60 },
      {     59,    -76,    -37,     87,     11,    -90,     15,     86,    -41,    -74,     62,     56,    -79,    -33,     88,      7,    -90,     20,     84,    -45,    -71,     66,     52,    -81,    -28,     89,      2,    -90,     24,     83,    -48,    -69,     69,     48,    -83,    -24,     90,     -2,    -89,     28,     81,    -52,    -66,     71,     45,    -84,    -20,     90,     -7,    -88,     33,     79,    -56,    -62,     74,     41,    -86,    -15,     90,    -11,    -87,     37,     76,    -59,    -59,     76,     37,    -87,    -11,     90,    -15,    -86,     41,     74,    -62,    -56,     79,     33,    -88,     -7,     90,    -20,    -84,     45,     71,    -66,    -52,     81,     28,    -89,     -2,     90,    -24,    -83,     48,     69,    -69,    -48,     83,     24,    -90,      2,     89,    -28,    -81,     52,     66,    -71,    -45,     84,     20,    -90,      7,     88,    -33,    -79,     56,     62,    -74,    -41,     86,     15,    -90,     11,     87,    -37,    -76,     59 },
      {     58,    -78,    -32,     89,      1,    -89,     29,     79,    -57,    -60,     77,     34,    -89,     -3,     90,    -27,    -80,     55,     62,    -76,    -36,     88,      6,    -90,     25,     81,    -53,    -63,     75,     38,    -88,     -8,     90,    -23,    -82,     51,     65,    -73,    -40,     87,     10,    -90,     21,     83,    -49,    -66,     72,     42,    -86,    -12,     90,    -19,    -84,     47,     68,    -71,    -44,     86,     14,    -91,     17,     85,    -46,    -69,     69,     46,    -85,    -17,     91,    -14,    -86,     44,     71,    -68,    -47,     84,     19,    -90,     12,     86,    -42,    -72,     66,     49,    -83,    -21,     90,    -10,    -87,     40,     73,    -65,    -51,     82,     23,    -90,      8,     88,    -38,    -75,     63,     53,    -81,    -25,     90,     -6,    -88,     36,     76,    -62,    -55,     80,     27,    -90,      3,     89,    -34,    -77,     60,     57,    -79,    -29,     89,     -1,    -89,     32,     78,    -58 },
      {     57,    -80,    -26,     90,     -9,    -87,     43,     70,    -70,    -43,     87,      9,    -90,     26,     80,    -57,    -57,     80,     26,    -90,      9,     87,    -43,    -70,     70,     43,    -87,     -9,     90,    -26,    -80,     57,     57,    -80,    -26,     90,     -9,    -87,     43,     70,    -70,    -43,     87,      9,    -90,     26,     80,    -57,    -57,     80,     26,    -90,      9,     87,    -43,    -70,     70,     43,    -87,     -9,     90,    -26,    -80,     57,     57,    -80,    -26,     90,     -9,    -87,     43,     70,    -70,    -43,     87,      9,    -90,     26,     80,    -57,    -57,     80,     26,    -90,      9,     87,    -43,    -70,     70,     43,    -87,     -9,     90,    -26,    -80,     57,     57,    -80,    -26,     90,     -9,    -87,     43,     70,    -70,    -43,     87,      9,    -90,     26,     80,    -57,    -57,     80,     26,    -90,      9,     87,    -43,    -70,     70,     43,    -87,     -9,     90,    -26,    -80,     57 },
      {     57,    -81,    -21,     91,    -19,    -82,     55,     58,    -80,    -23,     90,    -17,    -83,     53,     60,    -79,    -25,     90,    -14,    -84,     51,     62,    -78,    -27,     90,    -12,    -85,     49,     63,    -77,    -29,     90,    -10,    -86,     47,     65,    -76,    -32,     90,     -8,    -86,     46,     66,    -75,    -34,     89,     -6,    -87,     44,     68,    -73,    -36,     89,     -3,    -88,     42,     69,    -72,    -38,     89,     -1,    -88,     40,     71,    -71,    -40,     88,      1,    -89,     38,     72,    -69,    -42,     88,      3,    -89,     36,     73,    -68,    -44,     87,      6,    -89,     34,     75,    -66,    -46,     86,      8,    -90,     32,     76,    -65,    -47,     86,     10,    -90,     29,     77,    -63,    -49,     85,     12,    -90,     27,     78,    -62,    -51,     84,     14,    -90,     25,     79,    -60,    -53,     83,     17,    -90,     23,     80,    -58,    -55,     82,     19,    -91,     21,     81,    -57 },
      {     56,    -83,    -15,     90,    -28,    -76,     66,     45,    -87,     -2,     88,    -41,    -69,     74,     33,    -90,     11,     84,    -52,    -59,     81,     20,    -90,     24,     79,    -62,    -48,     86,      7,    -89,     37,     71,    -71,    -37,     89,     -7,    -86,     48,     62,    -79,    -24,     90,    -20,    -81,     59,     52,    -84,    -11,     90,    -33,    -74,     69,     41,    -88,      2,     87,    -45,    -66,     76,     28,    -90,     15,     83,    -56,    -56,     83,     15,    -90,     28,     76,    -66,    -45,     87,      2,    -88,     41,     69,    -74,    -33,     90,    -11,    -84,     52,     59,    -81,    -20,     90,    -24,    -79,     62,     48,    -86,     -7,     89,    -37,    -71,     71,     37,    -89,      7,     86,    -48,    -62,     79,     24,    -90,     20,     81,    -59,    -52,     84,     11,    -90,     33,     74,    -69,    -41,     88,     -2,    -87,     45,     66,    -76,    -28,     90,    -15,    -83,     56 },
      {     55,    -84,    -10,     89,    -38,    -69,     75,     29,    -90,     19,     80,    -62,    -47,     87,      1,    -88,     46,     63,    -79,    -21,     90,    -27,    -76,     68,     40,    -89,      8,     85,    -53,    -57,     83,     12,    -90,     36,     71,    -73,    -32,     90,    -17,    -81,     60,     49,    -86,     -3,     88,    -44,    -65,     78,     23,    -91,     25,     77,    -66,    -42,     89,     -6,    -86,     51,     58,    -82,    -14,     90,    -34,    -72,     72,     34,    -90,     14,     82,    -58,    -51,     86,      6,    -89,     42,     66,    -77,    -25,     91,    -23,    -78,     65,     44,    -88,      3,     86,    -49,    -60,     81,     17,    -90,     32,     73,    -71,    -36,     90,    -12,    -83,     57,     53,    -85,     -8,     89,    -40,    -68,     76,     27,    -90,     21,     79,    -63,    -46,     88,     -1,    -87,     47,     62,    -80,    -19,     90,    -29,    -75,     69,     38,    -89,     10,     84,    -55 },
      {     54,    -85,     -4,     88,    -47,    -61,     82,     13,    -90,     39,     67,    -78,    -22,     90,    -30,    -73,     73,     30,    -90,     22,     78,    -67,    -39,     90,    -13,    -82,     61,     47,    -88,      4,     85,    -54,    -54,     85,      4,    -88,     47,     61,    -82,    -13,     90,    -39,    -67,     78,     22,    -90,     30,     73,    -73,    -30,     90,    -22,    -78,     67,     39,    -90,     13,     82,    -61,    -47,     88,     -4,    -85,     54,     54,    -85,     -4,     88,    -47,    -61,     82,     13,    -90,     39,     67,    -78,    -22,     90,    -30,    -73,     73,     30,    -90,     22,     78,    -67,    -39,     90,    -13,    -82,     61,     47,    -88,      4,     85,    -54,    -54,     85,      4,    -88,     47,     61,    -82,    -13,     90,    -39,    -67,     78,     22,    -90,     30,     73,    -73,    -30,     90,    -22,    -78,     67,     39,    -90,     13,     82,    -61,    -47,     88,     -4,    -85,     54 },
      {     53,    -86,      1,     86,    -55,    -51,     87,     -3,    -85,     57,     49,    -88,      6,     84,    -58,    -47,     88,     -8,    -83,     60,     46,    -89,     10,     82,    -62,    -44,     89,    -12,    -81,     63,     42,    -89,     14,     80,    -65,    -40,     90,    -17,    -79,     66,     38,    -90,     19,     78,    -68,    -36,     90,    -21,    -77,     69,     34,    -90,     23,     76,    -71,    -32,     90,    -25,    -75,     72,     29,    -91,     27,     73,    -73,    -27,     91,    -29,    -72,     75,     25,    -90,     32,     71,    -76,    -23,     90,    -34,    -69,     77,     21,    -90,     36,     68,    -78,    -19,     90,    -38,    -66,     79,     17,    -90,     40,     65,    -80,    -14,     89,    -42,    -63,     81,     12,    -89,     44,     62,    -82,    -10,     89,    -46,    -60,     83,      8,    -88,     47,     58,    -84,     -6,     88,    -49,    -57,     85,      3,    -87,     51,     55,    -86,     -1,     86,    -53 },
      {     52,    -87,      7,     83,    -62,    -41,     90,    -20,    -76,     71,     28,    -90,     33,     69,    -79,    -15,     89,    -45,    -59,     84,      2,    -86,     56,     48,    -88,     11,     81,    -66,    -37,     90,    -24,    -74,     74,     24,    -90,     37,     66,    -81,    -11,     88,    -48,    -56,     86,     -2,    -84,     59,     45,    -89,     15,     79,    -69,    -33,     90,    -28,    -71,     76,     20,    -90,     41,     62,    -83,     -7,     87,    -52,    -52,     87,     -7,    -83,     62,     41,    -90,     20,     76,    -71,    -28,     90,    -33,    -69,     79,     15,    -89,     45,     59,    -84,     -2,     86,    -56,    -48,     88,    -11,    -81,     66,     37,    -90,     24,     74,    -74,    -24,     90,    -37,    -66,     81,     11,    -88,     48,     56,    -86,      2,     84,    -59,    -45,     89,    -15,    -79,     69,     33,    -90,     28,     71,    -76,    -20,     90,    -41,    -62,     83,      7,    -87,     52 },
      {     51,    -88,     12,     79,    -69,    -29,     90,    -36,    -65,     82,      6,    -86,     57,     46,    -89,     19,     76,    -73,    -23,     90,    -42,    -60,     85,     -1,    -84,     62,     40,    -90,     25,     72,    -77,    -17,     89,    -47,    -55,     87,     -8,    -81,     66,     34,    -91,     32,     68,    -80,    -10,     88,    -53,    -49,     89,    -14,    -78,     71,     27,    -90,     38,     63,    -83,     -3,     86,    -58,    -44,     90,    -21,    -75,     75,     21,    -90,     44,     58,    -86,      3,     83,    -63,    -38,     90,    -27,    -71,     78,     14,    -89,     49,     53,    -88,     10,     80,    -68,    -32,     91,    -34,    -66,     81,      8,    -87,     55,     47,    -89,     17,     77,    -72,    -25,     90,    -40,    -62,     84,      1,    -85,     60,     42,    -90,     23,     73,    -76,    -19,     89,    -46,    -57,     86,     -6,    -82,     65,     36,    -90,     29,     69,    -79,    -12,     88,    -51 },
      {     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50,     50,    -89,     18,     75,    -75,    -18,     89,    -50,    -50,     89,    -18,    -75,     75,     18,    -89,     50 },
      {     49,    -89,     23,     71,    -80,     -6,     85,    -63,    -34,     90,    -40,    -58,     87,    -12,    -77,     75,     17,    -88,     55,     44,    -90,     29,     66,    -83,      1,     82,    -68,    -27,     90,    -46,    -53,     89,    -19,    -73,     78,     10,    -86,     60,     38,    -91,     36,     62,    -86,      8,     79,    -72,    -21,     89,    -51,    -47,     90,    -25,    -69,     81,      3,    -84,     65,     32,    -90,     42,     57,    -88,     14,     76,    -76,    -14,     88,    -57,    -42,     90,    -32,    -65,     84,     -3,    -81,     69,     25,    -90,     47,     51,    -89,     21,     72,    -79,     -8,     86,    -62,    -36,     91,    -38,    -60,     86,    -10,    -78,     73,     19,    -89,     53,     46,    -90,     27,     68,    -82,     -1,     83,    -66,    -29,     90,    -44,    -55,     88,    -17,    -75,     77,     12,    -87,     58,     40,    -90,     34,     63,    -85,      6,     80,    -71,    -23,     89,    -49 },
      {     48,    -90,     28,     66,    -84,      7,     79,    -74,    -15,     87,    -59,    -37,     90,    -41,    -56,     88,    -20,    -71,     81,      2,    -83,     69,     24,    -89,     52,     45,    -90,     33,     62,    -86,     11,     76,    -76,    -11,     86,    -62,    -33,     90,    -45,    -52,     89,    -24,    -69,     83,     -2,    -81,     71,     20,    -88,     56,     41,    -90,     37,     59,    -87,     15,     74,    -79,     -7,     84,    -66,    -28,     90,    -48,    -48,     90,    -28,    -66,     84,     -7,    -79,     74,     15,    -87,     59,     37,    -90,     41,     56,    -88,     20,     71,    -81,     -2,     83,    -69,    -24,     89,    -52,    -45,     90,    -33,    -62,     86,    -11,    -76,     76,     11,    -86,     62,     33,    -90,     45,     52,    -89,     24,     69,    -83,      2,     81,    -71,    -20,     88,    -56,    -41,     90,    -37,    -59,     87,    -15,    -74,     79,      7,    -84,     66,     28,    -90,     48 },
      {     47,    -90,     34,     60,    -88,     19,     71,    -82,      3,     79,    -75,    -12,     86,    -65,    -27,     89,    -53,    -42,     91,    -40,    -55,     89,    -25,    -66,     85,    -10,    -76,     78,      6,    -83,     69,     21,    -88,     58,     36,    -90,     46,     49,    -90,     32,     62,    -87,     17,     72,    -81,      1,     80,    -73,    -14,     86,    -63,    -29,     90,    -51,    -44,     90,    -38,    -57,     89,    -23,    -68,     84,     -8,    -77,     77,      8,    -84,     68,     23,    -89,     57,     38,    -90,     44,     51,    -90,     29,     63,    -86,     14,     73,    -80,     -1,     81,    -72,    -17,     87,    -62,    -32,     90,    -49,    -46,     90,    -36,    -58,     88,    -21,    -69,     83,     -6,    -78,     76,     10,    -85,     66,     25,    -89,     55,     40,    -91,     42,     53,    -89,     27,     65,    -86,     12,     75,    -79,     -3,     82,    -71,    -19,     88,    -60,    -34,     90,    -47 },
      {     47,    -90,     39,     54,    -90,     30,     61,    -88,     22,     67,    -85,     13,     73,    -82,      4,     78,    -78,     -4,     82,    -73,    -13,     85,    -67,    -22,     88,    -61,    -30,     90,    -54,    -39,     90,    -47,    -47,     90,    -39,    -54,     90,    -30,    -61,     88,    -22,    -67,     85,    -13,    -73,     82,     -4,    -78,     78,      4,    -82,     73,     13,    -85,     67,     22,    -88,     61,     30,    -90,     54,     39,    -90,     47,     47,    -90,     39,     54,    -90,     30,     61,    -88,     22,     67,    -85,     13,     73,    -82,      4,     78,    -78,     -4,     82,    -73,    -13,     85,    -67,    -22,     88,    -61,    -30,     90,    -54,    -39,     90,    -47,    -47,     90,    -39,    -54,     90,    -30,    -61,     88,    -22,    -67,     85,    -13,    -73,     82,     -4,    -78,     78,      4,    -82,     73,     13,    -85,     67,     22,    -88,     61,     30,    -90,     54,     39,    -90,     47 },
      {     46,    -91,     44,     47,    -90,     42,     49,    -90,     40,     51,    -90,     38,     53,    -90,     36,     55,    -90,     34,     57,    -89,     32,     58,    -89,     29,     60,    -89,     27,     62,    -88,     25,     63,    -88,     23,     65,    -87,     21,     66,    -86,     19,     68,    -86,     17,     69,    -85,     14,     71,    -84,     12,     72,    -83,     10,     73,    -82,      8,     75,    -81,      6,     76,    -80,      3,     77,    -79,      1,     78,    -78,     -1,     79,    -77,     -3,     80,    -76,     -6,     81,    -75,     -8,     82,    -73,    -10,     83,    -72,    -12,     84,    -71,    -14,     85,    -69,    -17,     86,    -68,    -19,     86,    -66,    -21,     87,    -65,    -23,     88,    -63,    -25,     88,    -62,    -27,     89,    -60,    -29,     89,    -58,    -32,     89,    -57,    -34,     90,    -55,    -36,     90,    -53,    -38,     90,    -51,    -40,     90,    -49,    -42,     90,    -47,    -44,     91,    -46 },
      {     45,    -90,     48,     41,    -90,     52,     37,    -90,     56,     33,    -89,     59,     28,    -88,     62,     24,    -87,     66,     20,    -86,     69,     15,    -84,     71,     11,    -83,     74,      7,    -81,     76,      2,    -79,     79,     -2,    -76,     81,     -7,    -74,     83,    -11,    -71,     84,    -15,    -69,     86,    -20,    -66,     87,    -24,    -62,     88,    -28,    -59,     89,    -33,    -56,     90,    -37,    -52,     90,    -41,    -48,     90,    -45,    -45,     90,    -48,    -41,     90,    -52,    -37,     90,    -56,    -33,     89,    -59,    -28,     88,    -62,    -24,     87,    -66,    -20,     86,    -69,    -15,     84,    -71,    -11,     83,    -74,     -7,     81,    -76,     -2,     79,    -79,      2,     76,    -81,      7,     74,    -83,     11,     71,    -84,     15,     69,    -86,     20,     66,    -87,     24,     62,    -88,     28,     59,    -89,     33,     56,    -90,     37,     52,    -90,     41,     48,    -90,     45 },
      {     44,    -90,     53,     34,    -89,     62,     23,    -86,     69,     12,    -82,     76,      1,    -77,     81,    -10,    -71,     86,    -21,    -63,     89,    -32,    -55,     90,    -42,    -46,     90,    -51,    -36,     89,    -60,    -25,     87,    -68,    -14,     83,    -75,     -3,     78,    -80,      8,     72,    -85,     19,     65,    -88,     29,     57,    -90,     40,     47,    -91,     49,     38,    -90,     58,     27,    -88,     66,     17,    -84,     73,      6,    -79,     79,     -6,    -73,     84,    -17,    -66,     88,    -27,    -58,     90,    -38,    -49,     91,    -47,    -40,     90,    -57,    -29,     88,    -65,    -19,     85,    -72,     -8,     80,    -78,      3,     75,    -83,     14,     68,    -87,     25,     60,    -89,     36,     51,    -90,     46,     42,    -90,     55,     32,    -89,     63,     21,    -86,     71,     10,    -81,     77,     -1,    -76,     82,    -12,    -69,     86,    -23,    -62,     89,    -34,    -53,     90,    -44 },
      {     43,    -90,     57,     26,    -87,     70,      9,    -80,     80,     -9,    -70,     87,    -26,    -57,     90,    -43,    -43,     90,    -57,    -26,     87,    -70,     -9,     80,    -80,      9,     70,    -87,     26,     57,    -90,     43,     43,    -90,     57,     26,    -87,     70,      9,    -80,     80,     -9,    -70,     87,    -26,    -57,     90,    -43,    -43,     90,    -57,    -26,     87,    -70,     -9,     80,    -80,      9,     70,    -87,     26,     57,    -90,     43,     43,    -90,     57,     26,    -87,     70,      9,    -80,     80,     -9,    -70,     87,    -26,    -57,     90,    -43,    -43,     90,    -57,    -26,     87,    -70,     -9,     80,    -80,      9,     70,    -87,     26,     57,    -90,     43,     43,    -90,     57,     26,    -87,     70,      9,    -80,     80,     -9,    -70,     87,    -26,    -57,     90,    -43,    -43,     90,    -57,    -26,     87,    -70,     -9,     80,    -80,      9,     70,    -87,     26,     57,    -90,     43 },
      {     42,    -90,     62,     19,    -83,     77,     -6,    -71,     87,    -29,    -53,     91,    -51,    -32,     88,    -69,     -8,     78,    -82,     17,     63,    -89,     40,     44,    -90,     60,     21,    -84,     76,     -3,    -72,     86,    -27,    -55,     90,    -49,    -34,     88,    -68,    -10,     79,    -81,     14,     65,    -89,     38,     46,    -90,     58,     23,    -85,     75,     -1,    -73,     86,    -25,    -57,     90,    -47,    -36,     89,    -66,    -12,     80,    -80,     12,     66,    -89,     36,     47,    -90,     57,     25,    -86,     73,      1,    -75,     85,    -23,    -58,     90,    -46,    -38,     89,    -65,    -14,     81,    -79,     10,     68,    -88,     34,     49,    -90,     55,     27,    -86,     72,      3,    -76,     84,    -21,    -60,     90,    -44,    -40,     89,    -63,    -17,     82,    -78,      8,     69,    -88,     32,     51,    -91,     53,     29,    -87,     71,      6,    -77,     83,    -19,    -62,     90,    -42 },
      {     41,    -89,     66,     11,    -79,     83,    -20,    -59,     90,    -48,    -33,     87,    -71,     -2,     74,    -86,     28,     52,    -90,     56,     24,    -84,     76,     -7,    -69,     88,    -37,    -45,     90,    -62,    -15,     81,    -81,     15,     62,    -90,     45,     37,    -88,     69,      7,    -76,     84,    -24,    -56,     90,    -52,    -28,     86,    -74,      2,     71,    -87,     33,     48,    -90,     59,     20,    -83,     79,    -11,    -66,     89,    -41,    -41,     89,    -66,    -11,     79,    -83,     20,     59,    -90,     48,     33,    -87,     71,      2,    -74,     86,    -28,    -52,     90,    -56,    -24,     84,    -76,      7,     69,    -88,     37,     45,    -90,     62,     15,    -81,     81,    -15,    -62,     90,    -45,    -37,     88,    -69,     -7,     76,    -84,     24,     56,    -90,     52,     28,    -86,     74,     -2,    -71,     87,    -33,    -48,     90,    -59,    -20,     83,    -79,     11,     66,    -89,     41 },
      {     40,    -89,     69,      3,    -73,     87,    -34,    -46,     90,    -65,    -10,     77,    -85,     27,     51,    -90,     60,     17,    -80,     82,    -21,    -57,     91,    -55,    -23,     83,    -79,     14,     62,    -90,     49,     29,    -86,     76,     -8,    -66,     89,    -44,    -36,     88,    -72,      1,     71,    -88,     38,     42,    -89,     68,      6,    -75,     86,    -32,    -47,     90,    -63,    -12,     78,    -84,     25,     53,    -90,     58,     19,    -81,     81,    -19,    -58,     90,    -53,    -25,     84,    -78,     12,     63,    -90,     47,     32,    -86,     75,     -6,    -68,     89,    -42,    -38,     88,    -71,     -1,     72,    -88,     36,     44,    -89,     66,      8,    -76,     86,    -29,    -49,     90,    -62,    -14,     79,    -83,     23,     55,    -91,     57,     21,    -82,     80,    -17,    -60,     90,    -51,    -27,     85,    -77,     10,     65,    -90,     46,     34,    -87,     73,     -3,    -69,     89,    -40 },
      {     39,    -88,     73,     -4,    -67,     90,    -47,    -30,     85,    -78,     13,     61,    -90,     54,     22,    -82,     82,    -22,    -54,     90,    -61,    -13,     78,    -85,     30,     47,    -90,     67,      4,    -73,     88,    -39,    -39,     88,    -73,      4,     67,    -90,     47,     30,    -85,     78,    -13,    -61,     90,    -54,    -22,     82,    -82,     22,     54,    -90,     61,     13,    -78,     85,    -30,    -47,     90,    -67,     -4,     73,    -88,     39,     39,    -88,     73,     -4,    -67,     90,    -47,    -30,     85,    -78,     13,     61,    -90,     54,     22,    -82,     82,    -22,    -54,     90,    -61,    -13,     78,    -85,     30,     47,    -90,     67,      4,    -73,     88,    -39,    -39,     88,    -73,      4,     67,    -90,     47,     30,    -85,     78,    -13,    -61,     90,    -54,    -22,     82,    -82,     22,     54,    -90,     61,     13,    -78,     85,    -30,    -47,     90,    -67,     -4,     73,    -88,     39 },
      {     38,    -87,     76,    -12,    -60,     91,    -58,    -14,     77,    -86,     36,     40,    -88,     75,    -10,    -62,     90,    -57,    -17,     78,    -86,     34,     42,    -88,     73,     -8,    -63,     90,    -55,    -19,     79,    -85,     32,     44,    -89,     72,     -6,    -65,     90,    -53,    -21,     80,    -84,     29,     46,    -89,     71,     -3,    -66,     90,    -51,    -23,     81,    -83,     27,     47,    -89,     69,     -1,    -68,     90,    -49,    -25,     82,    -82,     25,     49,    -90,     68,      1,    -69,     89,    -47,    -27,     83,    -81,     23,     51,    -90,     66,      3,    -71,     89,    -46,    -29,     84,    -80,     21,     53,    -90,     65,      6,    -72,     89,    -44,    -32,     85,    -79,     19,     55,    -90,     63,      8,    -73,     88,    -42,    -34,     86,    -78,     17,     57,    -90,     62,     10,    -75,     88,    -40,    -36,     86,    -77,     14,     58,    -91,     60,     12,    -76,     87,    -38 },
      {     37,    -86,     79,    -20,    -52,     90,    -69,      2,     66,    -90,     56,     15,    -76,     87,    -41,    -33,     84,    -81,     24,     48,    -89,     71,     -7,    -62,     90,    -59,    -11,     74,    -88,     45,     28,    -83,     83,    -28,    -45,     88,    -74,     11,     59,    -90,     62,      7,    -71,     89,    -48,    -24,     81,    -84,     33,     41,    -87,     76,    -15,    -56,     90,    -66,     -2,     69,    -90,     52,     20,    -79,     86,    -37,    -37,     86,    -79,     20,     52,    -90,     69,     -2,    -66,     90,    -56,    -15,     76,    -87,     41,     33,    -84,     81,    -24,    -48,     89,    -71,      7,     62,    -90,     59,     11,    -74,     88,    -45,    -28,     83,    -83,     28,     45,    -88,     74,    -11,    -59,     90,    -62,     -7,     71,    -89,     48,     24,    -81,     84,    -33,    -41,     87,    -76,     15,     56,    -90,     66,      2,    -69,     90,    -52,    -20,     79,    -86,     37 },
      {     36,    -85,     81,    -27,    -44,     88,    -77,     19,     51,    -89,     72,    -10,    -58,     90,    -66,      1,     65,    -90,     60,      8,    -71,     90,    -53,    -17,     76,    -88,     46,     25,    -80,     86,    -38,    -34,     84,    -82,     29,     42,    -87,     78,    -21,    -49,     89,    -73,     12,     57,    -90,     68,     -3,    -63,     91,    -62,     -6,     69,    -90,     55,     14,    -75,     89,    -47,    -23,     79,    -86,     40,     32,    -83,     83,    -32,    -40,     86,    -79,     23,     47,    -89,     75,    -14,    -55,     90,    -69,      6,     62,    -91,     63,      3,    -68,     90,    -57,    -12,     73,    -89,     49,     21,    -78,     87,    -42,    -29,     82,    -84,     34,     38,    -86,     80,    -25,    -46,     88,    -76,     17,     53,    -90,     71,     -8,    -60,     90,    -65,     -1,     66,    -90,     58,     10,    -72,     89,    -51,    -19,     77,    -88,     44,     27,    -81,     85,    -36 },
      {     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35,     35,    -84,     84,    -35,    -35,     84,    -84,     35 },
      {     34,    -82,     86,    -42,    -25,     78,    -88,     49,     17,    -73,     90,    -57,     -8,     68,    -90,     63,     -1,    -62,     90,    -69,     10,     55,    -89,     75,    -19,    -47,     88,    -79,     27,     40,    -85,     83,    -36,    -32,     81,    -86,     44,     23,    -77,     89,    -51,    -14,     72,    -90,     58,      6,    -66,     91,    -65,      3,     60,    -90,     71,    -12,    -53,     89,    -76,     21,     46,    -87,     80,    -29,    -38,     84,    -84,     38,     29,    -80,     87,    -46,    -21,     76,    -89,     53,     12,    -71,     90,    -60,     -3,     65,    -91,     66,     -6,    -58,     90,    -72,     14,     51,    -89,     77,    -23,    -44,     86,    -81,     32,     36,    -83,     85,    -40,    -27,     79,    -88,     47,     19,    -75,     89,    -55,    -10,     69,    -90,     62,      1,    -63,     90,    -68,      8,     57,    -90,     73,    -17,    -49,     88,    -78,     25,     42,    -86,     82,    -34 },
      {     33,    -81,     87,    -48,    -15,     71,    -90,     62,     -2,    -59,     90,    -74,     20,     45,    -86,     83,    -37,    -28,     79,    -88,     52,     11,    -69,     90,    -66,      7,     56,    -89,     76,    -24,    -41,     84,    -84,     41,     24,    -76,     89,    -56,     -7,     66,    -90,     69,    -11,    -52,     88,    -79,     28,     37,    -83,     86,    -45,    -20,     74,    -90,     59,      2,    -62,     90,    -71,     15,     48,    -87,     81,    -33,    -33,     81,    -87,     48,     15,    -71,     90,    -62,      2,     59,    -90,     74,    -20,    -45,     86,    -83,     37,     28,    -79,     88,    -52,    -11,     69,    -90,     66,     -7,    -56,     89,    -76,     24,     41,    -84,     84,    -41,    -24,     76,    -89,     56,      7,    -66,     90,    -69,     11,     52,    -88,     79,    -28,    -37,     83,    -86,     45,     20,    -74,     90,    -59,     -2,     62,    -90,     71,    -15,    -48,     87,    -81,     33 },
      {     32,    -79,     89,    -55,     -6,     63,    -90,     73,    -21,    -42,     84,    -86,     46,     17,    -71,     90,    -66,     10,     51,    -88,     81,    -36,    -27,     77,    -89,     58,      1,    -60,     90,    -76,     25,     38,    -82,     87,    -49,    -12,     68,    -91,     69,    -14,    -47,     86,    -83,     40,     23,    -75,     90,    -62,      3,     57,    -89,     78,    -29,    -34,     80,    -88,     53,      8,    -65,     90,    -72,     19,     44,    -85,     85,    -44,    -19,     72,    -90,     65,     -8,    -53,     88,    -80,     34,     29,    -78,     89,    -57,     -3,     62,    -90,     75,    -23,    -40,     83,    -86,     47,     14,    -69,     91,    -68,     12,     49,    -87,     82,    -38,    -25,     76,    -90,     60,     -1,    -58,     89,    -77,     27,     36,    -81,     88,    -51,    -10,     66,    -90,     71,    -17,    -46,     86,    -84,     42,     21,    -73,     90,    -63,      6,     55,    -89,     79,    -32 },
      {     30,    -78,     90,    -61,      4,     54,    -88,     82,    -39,    -22,     73,    -90,     67,    -13,    -47,     85,    -85,     47,     13,    -67,     90,    -73,     22,     39,    -82,     88,    -54,     -4,     61,    -90,     78,    -30,    -30,     78,    -90,     61,     -4,    -54,     88,    -82,     39,     22,    -73,     90,    -67,     13,     47,    -85,     85,    -47,    -13,     67,    -90,     73,    -22,    -39,     82,    -88,     54,      4,    -61,     90,    -78,     30,     30,    -78,     90,    -61,      4,     54,    -88,     82,    -39,    -22,     73,    -90,     67,    -13,    -47,     85,    -85,     47,     13,    -67,     90,    -73,     22,     39,    -82,     88,    -54,     -4,     61,    -90,     78,    -30,    -30,     78,    -90,     61,     -4,    -54,     88,    -82,     39,     22,    -73,     90,    -67,     13,     47,    -85,     85,    -47,    -13,     67,    -90,     73,    -22,    -39,     82,    -88,     54,      4,    -61,     90,    -78,     30 },
      {     29,    -76,     90,    -66,     14,     44,    -83,     88,    -55,     -1,     57,    -88,     82,    -42,    -17,     68,    -90,     75,    -27,    -32,     77,    -90,     65,    -12,    -46,     84,    -87,     53,      3,    -58,     89,    -81,     40,     19,    -69,     90,    -73,     25,     34,    -78,     90,    -63,     10,     47,    -85,     86,    -51,     -6,     60,    -89,     80,    -38,    -21,     71,    -91,     72,    -23,    -36,     79,    -89,     62,     -8,    -49,     86,    -86,     49,      8,    -62,     89,    -79,     36,     23,    -72,     91,    -71,     21,     38,    -80,     89,    -60,      6,     51,    -86,     85,    -47,    -10,     63,    -90,     78,    -34,    -25,     73,    -90,     69,    -19,    -40,     81,    -89,     58,     -3,    -53,     87,    -84,     46,     12,    -65,     90,    -77,     32,     27,    -75,     90,    -68,     17,     42,    -82,     88,    -57,      1,     55,    -88,     83,    -44,    -14,     66,    -90,     76,    -29 },
      {     28,    -74,     90,    -71,     24,     33,    -76,     90,    -69,     20,     37,    -79,     90,    -66,     15,     41,    -81,     89,    -62,     11,     45,    -83,     88,    -59,      7,     48,    -84,     87,    -56,      2,     52,    -86,     86,    -52,     -2,     56,    -87,     84,    -48,     -7,     59,    -88,     83,    -45,    -11,     62,    -89,     81,    -41,    -15,     66,    -90,     79,    -37,    -20,     69,    -90,     76,    -33,    -24,     71,    -90,     74,    -28,    -28,     74,    -90,     71,    -24,    -33,     76,    -90,     69,    -20,    -37,     79,    -90,     66,    -15,    -41,     81,    -89,     62,    -11,    -45,     83,    -88,     59,     -7,    -48,     84,    -87,     56,     -2,    -52,     86,    -86,     52,      2,    -56,     87,    -84,     48,      7,    -59,     88,    -83,     45,     11,    -62,     89,    -81,     41,     15,    -66,     90,    -79,     37,     20,    -69,     90,    -76,     33,     24,    -71,     90,    -74,     28 },
      {     27,    -72,     90,    -76,     34,     21,    -68,     90,    -79,     40,     14,    -63,     89,    -82,     46,      8,    -58,     88,    -85,     51,      1,    -53,     86,    -87,     57,     -6,    -47,     83,    -89,     62,    -12,    -42,     80,    -90,     66,    -19,    -36,     77,    -90,     71,    -25,    -29,     73,    -91,     75,    -32,    -23,     69,    -90,     78,    -38,    -17,     65,    -89,     81,    -44,    -10,     60,    -88,     84,    -49,     -3,     55,    -86,     86,    -55,      3,     49,    -84,     88,    -60,     10,     44,    -81,     89,    -65,     17,     38,    -78,     90,    -69,     23,     32,    -75,     91,    -73,     29,     25,    -71,     90,    -77,     36,     19,    -66,     90,    -80,     42,     12,    -62,     89,    -83,     47,      6,    -57,     87,    -86,     53,     -1,    -51,     85,    -88,     58,     -8,    -46,     82,    -89,     63,    -14,    -40,     79,    -90,     68,    -21,    -34,     76,    -90,     72,    -27 },
      {     26,    -70,     90,    -80,     43,      9,    -57,     87,    -87,     57,     -9,    -43,     80,    -90,     70,    -26,    -26,     70,    -90,     80,    -43,     -9,     57,    -87,     87,    -57,      9,     43,    -80,     90,    -70,     26,     26,    -70,     90,    -80,     43,      9,    -57,     87,    -87,     57,     -9,    -43,     80,    -90,     70,    -26,    -26,     70,    -90,     80,    -43,     -9,     57,    -87,     87,    -57,      9,     43,    -80,     90,    -70,     26,     26,    -70,     90,    -80,     43,      9,    -57,     87,    -87,     57,     -9,    -43,     80,    -90,     70,    -26,    -26,     70,    -90,     80,    -43,     -9,     57,    -87,     87,    -57,      9,     43,    -80,     90,    -70,     26,     26,    -70,     90,    -80,     43,      9,    -57,     87,    -87,     57,     -9,    -43,     80,    -90,     70,    -26,    -26,     70,    -90,     80,    -43,     -9,     57,    -87,     87,    -57,      9,     43,    -80,     90,    -70,     26 },
      {     25,    -68,     89,    -83,     51,     -3,    -46,     80,    -90,     72,    -32,    -19,     63,    -88,     86,    -57,     10,     40,    -77,     91,    -76,     38,     12,    -58,     86,    -88,     62,    -17,    -34,     73,    -90,     79,    -44,     -6,     53,    -84,     89,    -66,     23,     27,    -69,     90,    -82,     49,     -1,    -47,     81,    -90,     71,    -29,    -21,     65,    -89,     85,    -55,      8,     42,    -78,     90,    -75,     36,     14,    -60,     87,    -87,     60,    -14,    -36,     75,    -90,     78,    -42,     -8,     55,    -85,     89,    -65,     21,     29,    -71,     90,    -81,     47,      1,    -49,     82,    -90,     69,    -27,    -23,     66,    -89,     84,    -53,      6,     44,    -79,     90,    -73,     34,     17,    -62,     88,    -86,     58,    -12,    -38,     76,    -91,     77,    -40,    -10,     57,    -86,     88,    -63,     19,     32,    -72,     90,    -80,     46,      3,    -51,     83,    -89,     68,    -25 },
      {     24,    -66,     88,    -86,     59,    -15,    -33,     71,    -90,     83,    -52,      7,     41,    -76,     90,    -79,     45,      2,    -48,     81,    -90,     74,    -37,    -11,     56,    -84,     89,    -69,     28,     20,    -62,     87,    -87,     62,    -20,    -28,     69,    -89,     84,    -56,     11,     37,    -74,     90,    -81,     48,     -2,    -45,     79,    -90,     76,    -41,     -7,     52,    -83,     90,    -71,     33,     15,    -59,     86,    -88,     66,    -24,    -24,     66,    -88,     86,    -59,     15,     33,    -71,     90,    -83,     52,     -7,    -41,     76,    -90,     79,    -45,     -2,     48,    -81,     90,    -74,     37,     11,    -56,     84,    -89,     69,    -28,    -20,     62,    -87,     87,    -62,     20,     28,    -69,     89,    -84,     56,    -11,    -37,     74,    -90,     81,    -48,      2,     45,    -79,     90,    -76,     41,      7,    -52,     83,    -90,     71,    -33,    -15,     59,    -86,     88,    -66,     24 },
      {     23,    -63,     87,    -88,     66,    -27,    -19,     60,    -86,     89,    -69,     32,     14,    -57,     84,    -90,     72,    -36,    -10,     53,    -82,     90,    -75,     40,      6,    -49,     80,    -90,     77,    -44,     -1,     46,    -78,     91,    -79,     47,     -3,    -42,     76,    -90,     81,    -51,      8,     38,    -73,     90,    -83,     55,    -12,    -34,     71,    -89,     85,    -58,     17,     29,    -68,     89,    -86,     62,    -21,    -25,     65,    -88,     88,    -65,     25,     21,    -62,     86,    -89,     68,    -29,    -17,     58,    -85,     89,    -71,     34,     12,    -55,     83,    -90,     73,    -38,     -8,     51,    -81,     90,    -76,     42,      3,    -47,     79,    -91,     78,    -46,      1,     44,    -77,     90,    -80,     49,     -6,    -40,     75,    -90,     82,    -53,     10,     36,    -72,     90,    -84,     57,    -14,    -32,     69,    -89,     86,    -60,     19,     27,    -66,     88,    -87,     63,    -23 },
      {     22,    -61,     85,    -90,     73,    -39,     -4,     47,    -78,     90,    -82,     54,    -13,    -30,     67,    -88,     88,    -67,     30,     13,    -54,     82,    -90,     78,    -47,      4,     39,    -73,     90,    -85,     61,    -22,    -22,     61,    -85,     90,    -73,     39,      4,    -47,     78,    -90,     82,    -54,     13,     30,    -67,     88,    -88,     67,    -30,    -13,     54,    -82,     90,    -78,     47,     -4,    -39,     73,    -90,     85,    -61,     22,     22,    -61,     85,    -90,     73,    -39,     -4,     47,    -78,     90,    -82,     54,    -13,    -30,     67,    -88,     88,    -67,     30,     13,    -54,     82,    -90,     78,    -47,      4,     39,    -73,     90,    -85,     61,    -22,    -22,     61,    -85,     90,    -73,     39,      4,    -47,     78,    -90,     82,    -54,     13,     30,    -67,     88,    -88,     67,    -30,    -13,     54,    -82,     90,    -78,     47,     -4,    -39,     73,    -90,     85,    -61,     22 },
      {     21,    -58,     83,    -90,     78,    -49,     10,     32,    -66,     87,    -89,     72,    -40,     -1,     42,    -73,     89,    -86,     65,    -29,    -12,     51,    -79,     90,    -82,     57,    -19,    -23,     60,    -84,     90,    -77,     47,     -8,    -34,     68,    -88,     89,    -71,     38,      3,    -44,     75,    -90,     86,    -63,     27,     14,    -53,     80,    -91,     81,    -55,     17,     25,    -62,     85,    -90,     76,    -46,      6,     36,    -69,     88,    -88,     69,    -36,     -6,     46,    -76,     90,    -85,     62,    -25,    -17,     55,    -81,     91,    -80,     53,    -14,    -27,     63,    -86,     90,    -75,     44,     -3,    -38,     71,    -89,     88,    -68,     34,      8,    -47,     77,    -90,     84,    -60,     23,     19,    -57,     82,    -90,     79,    -51,     12,     29,    -65,     86,    -89,     73,    -42,      1,     40,    -72,     89,    -87,     66,    -32,    -10,     49,    -78,     90,    -83,     58,    -21 },
      {     20,    -56,     81,    -90,     83,    -59,     24,     15,    -52,     79,    -90,     84,    -62,     28,     11,    -48,     76,    -90,     86,    -66,     33,      7,    -45,     74,    -89,     87,    -69,     37,      2,    -41,     71,    -88,     88,    -71,     41,     -2,    -37,     69,    -87,     89,    -74,     45,     -7,    -33,     66,    -86,     90,    -76,     48,    -11,    -28,     62,    -84,     90,    -79,     52,    -15,    -24,     59,    -83,     90,    -81,     56,    -20,    -20,     56,    -81,     90,    -83,     59,    -24,    -15,     52,    -79,     90,    -84,     62,    -28,    -11,     48,    -76,     90,    -86,     66,    -33,     -7,     45,    -74,     89,    -87,     69,    -37,     -2,     41,    -71,     88,    -88,     71,    -41,      2,     37,    -69,     87,    -89,     74,    -45,      7,     33,    -66,     86,    -90,     76,    -48,     11,     28,    -62,     84,    -90,     79,    -52,     15,     24,    -59,     83,    -90,     81,    -56,     20 },
      {     19,    -53,     78,    -90,     86,    -68,     38,     -1,    -36,     66,    -86,     90,    -79,     55,    -21,    -17,     51,    -77,     90,    -87,     69,    -40,      3,     34,    -65,     85,    -90,     80,    -57,     23,     14,    -49,     76,    -89,     88,    -71,     42,     -6,    -32,     63,    -84,     90,    -81,     58,    -25,    -12,     47,    -75,     89,    -88,     72,    -44,      8,     29,    -62,     83,    -91,     82,    -60,     27,     10,    -46,     73,    -89,     89,    -73,     46,    -10,    -27,     60,    -82,     91,    -83,     62,    -29,     -8,     44,    -72,     88,    -89,     75,    -47,     12,     25,    -58,     81,    -90,     84,    -63,     32,      6,    -42,     71,    -88,     89,    -76,     49,    -14,    -23,     57,    -80,     90,    -85,     65,    -34,     -3,     40,    -69,     87,    -90,     77,    -51,     17,     21,    -55,     79,    -90,     86,    -66,     36,      1,    -38,     68,    -86,     90,    -78,     53,    -19 },
      {     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18,     18,    -50,     75,    -89,     89,    -75,     50,    -18,    -18,     50,    -75,     89,    -89,     75,    -50,     18 },
      {     17,    -47,     72,    -87,     90,    -81,     62,    -34,      1,     32,    -60,     80,    -90,     88,    -73,     49,    -19,    -14,     46,    -71,     86,    -90,     82,    -63,     36,     -3,    -29,     58,    -79,     90,    -88,     75,    -51,     21,     12,    -44,     69,    -86,     90,    -83,     65,    -38,      6,     27,    -57,     78,    -89,     89,    -76,     53,    -23,    -10,     42,    -68,     85,    -91,     84,    -66,     40,     -8,    -25,     55,    -77,     89,    -89,     77,    -55,     25,      8,    -40,     66,    -84,     91,    -85,     68,    -42,     10,     23,    -53,     76,    -89,     89,    -78,     57,    -27,     -6,     38,    -65,     83,    -90,     86,    -69,     44,    -12,    -21,     51,    -75,     88,    -90,     79,    -58,     29,      3,    -36,     63,    -82,     90,    -86,     71,    -46,     14,     19,    -49,     73,    -88,     90,    -80,     60,    -32,     -1,     34,    -62,     81,    -90,     87,    -72,     47,    -17 },
      {     15,    -45,     69,    -84,     90,    -86,     71,    -48,     20,     11,    -41,     66,    -83,     90,    -87,     74,    -52,     24,      7,    -37,     62,    -81,     90,    -88,     76,    -56,     28,      2,    -33,     59,    -79,     89,    -89,     79,    -59,     33,     -2,    -28,     56,    -76,     88,    -90,     81,    -62,     37,     -7,    -24,     52,    -74,     87,    -90,     83,    -66,     41,    -11,    -20,     48,    -71,     86,    -90,     84,    -69,     45,    -15,    -15,     45,    -69,     84,    -90,     86,    -71,     48,    -20,    -11,     41,    -66,     83,    -90,     87,    -74,     52,    -24,     -7,     37,    -62,     81,    -90,     88,    -76,     56,    -28,     -2,     33,    -59,     79,    -89,     89,    -79,     59,    -33,      2,     28,    -56,     76,    -88,     90,    -81,     62,    -37,      7,     24,    -52,     74,    -87,     90,    -83,     66,    -41,     11,     20,    -48,     71,    -86,     90,    -84,     69,    -45,     15 },
      {     14,    -42,     65,    -81,     90,    -89,     79,    -62,     38,    -10,    -19,     46,    -68,     83,    -90,     88,    -77,     58,    -34,      6,     23,    -49,     71,    -85,     90,    -87,     75,    -55,     29,     -1,    -27,     53,    -73,     86,    -91,     86,    -72,     51,    -25,     -3,     32,    -57,     76,    -88,     90,    -84,     69,    -47,     21,      8,    -36,     60,    -78,     89,    -90,     82,    -66,     44,    -17,    -12,     40,    -63,     80,    -89,     89,    -80,     63,    -40,     12,     17,    -44,     66,    -82,     90,    -89,     78,    -60,     36,     -8,    -21,     47,    -69,     84,    -90,     88,    -76,     57,    -32,      3,     25,    -51,     72,    -86,     91,    -86,     73,    -53,     27,      1,    -29,     55,    -75,     87,    -90,     85,    -71,     49,    -23,     -6,     34,    -58,     77,    -88,     90,    -83,     68,    -46,     19,     10,    -38,     62,    -79,     89,    -90,     81,    -65,     42,    -14 },
      {     13,    -39,     61,    -78,     88,    -90,     85,    -73,     54,    -30,      4,     22,    -47,     67,    -82,     90,    -90,     82,    -67,     47,    -22,     -4,     30,    -54,     73,    -85,     90,    -88,     78,    -61,     39,    -13,    -13,     39,    -61,     78,    -88,     90,    -85,     73,    -54,     30,     -4,    -22,     47,    -67,     82,    -90,     90,    -82,     67,    -47,     22,      4,    -30,     54,    -73,     85,    -90,     88,    -78,     61,    -39,     13,     13,    -39,     61,    -78,     88,    -90,     85,    -73,     54,    -30,      4,     22,    -47,     67,    -82,     90,    -90,     82,    -67,     47,    -22,     -4,     30,    -54,     73,    -85,     90,    -88,     78,    -61,     39,    -13,    -13,     39,    -61,     78,    -88,     90,    -85,     73,    -54,     30,     -4,    -22,     47,    -67,     82,    -90,     90,    -82,     67,    -47,     22,      4,    -30,     54,    -73,     85,    -90,     88,    -78,     61,    -39,     13 },
      {     12,    -36,     57,    -73,     85,    -90,     89,    -81,     68,    -49,     27,     -3,    -21,     44,    -63,     78,    -88,     91,    -87,     77,    -62,     42,    -19,     -6,     29,    -51,     69,    -82,     89,    -90,     84,    -72,     55,    -34,     10,     14,    -38,     58,    -75,     86,    -90,     89,    -80,     66,    -47,     25,     -1,    -23,     46,    -65,     79,    -88,     90,    -86,     76,    -60,     40,    -17,     -8,     32,    -53,     71,    -83,     90,    -90,     83,    -71,     53,    -32,      8,     17,    -40,     60,    -76,     86,    -90,     88,    -79,     65,    -46,     23,      1,    -25,     47,    -66,     80,    -89,     90,    -86,     75,    -58,     38,    -14,    -10,     34,    -55,     72,    -84,     90,    -89,     82,    -69,     51,    -29,      6,     19,    -42,     62,    -77,     87,    -91,     88,    -78,     63,    -44,     21,      3,    -27,     49,    -68,     81,    -89,     90,    -85,     73,    -57,     36,    -12 },
      {     11,    -33,     52,    -69,     81,    -88,     90,    -87,     79,    -66,     48,    -28,      7,     15,    -37,     56,    -71,     83,    -89,     90,    -86,     76,    -62,     45,    -24,      2,     20,    -41,     59,    -74,     84,    -90,     90,    -84,     74,    -59,     41,    -20,     -2,     24,    -45,     62,    -76,     86,    -90,     89,    -83,     71,    -56,     37,    -15,     -7,     28,    -48,     66,    -79,     87,    -90,     88,    -81,     69,    -52,     33,    -11,    -11,     33,    -52,     69,    -81,     88,    -90,     87,    -79,     66,    -48,     28,     -7,    -15,     37,    -56,     71,    -83,     89,    -90,     86,    -76,     62,    -45,     24,     -2,    -20,     41,    -59,     74,    -84,     90,    -90,     84,    -74,     59,    -41,     20,      2,    -24,     45,    -62,     76,    -86,     90,    -89,     83,    -71,     56,    -37,     15,      7,    -28,     48,    -66,     79,    -87,     90,    -88,     81,    -69,     52,    -33,     11 },
      {     10,    -29,     47,    -63,     76,    -85,     90,    -90,     86,    -78,     66,    -51,     34,    -14,     -6,     25,    -44,     60,    -73,     83,    -89,     90,    -88,     80,    -69,     55,    -38,     19,      1,    -21,     40,    -57,     71,    -81,     88,    -91,     89,    -82,     72,    -58,     42,    -23,      3,     17,    -36,     53,    -68,     79,    -87,     90,    -89,     84,    -75,     62,    -46,     27,     -8,    -12,     32,    -49,     65,    -77,     86,    -90,     90,    -86,     77,    -65,     49,    -32,     12,      8,    -27,     46,    -62,     75,    -84,     89,    -90,     87,    -79,     68,    -53,     36,    -17,     -3,     23,    -42,     58,    -72,     82,    -89,     91,    -88,     81,    -71,     57,    -40,     21,     -1,    -19,     38,    -55,     69,    -80,     88,    -90,     89,    -83,     73,    -60,     44,    -25,      6,     14,    -34,     51,    -66,     78,    -86,     90,    -90,     85,    -76,     63,    -47,     29,    -10 },
      {      9,    -26,     43,    -57,     70,    -80,     87,    -90,     90,    -87,     80,    -70,     57,    -43,     26,     -9,     -9,     26,    -43,     57,    -70,     80,    -87,     90,    -90,     87,    -80,     70,    -57,     43,    -26,      9,      9,    -26,     43,    -57,     70,    -80,     87,    -90,     90,    -87,     80,    -70,     57,    -43,     26,     -9,     -9,     26,    -43,     57,    -70,     80,    -87,     90,    -90,     87,    -80,     70,    -57,     43,    -26,      9,      9,    -26,     43,    -57,     70,    -80,     87,    -90,     90,    -87,     80,    -70,     57,    -43,     26,     -9,     -9,     26,    -43,     57,    -70,     80,    -87,     90,    -90,     87,    -80,     70,    -57,     43,    -26,      9,      9,    -26,     43,    -57,     70,    -80,     87,    -90,     90,    -87,     80,    -70,     57,    -43,     26,     -9,     -9,     26,    -43,     57,    -70,     80,    -87,     90,    -90,     87,    -80,     70,    -57,     43,    -26,      9 },
      {      8,    -23,     38,    -51,     63,    -73,     81,    -87,     90,    -90,     88,    -83,     76,    -66,     55,    -42,     27,    -12,     -3,     19,    -34,     47,    -60,     71,    -79,     86,    -89,     91,    -89,     85,    -78,     69,    -58,     46,    -32,     17,     -1,    -14,     29,    -44,     57,    -68,     77,    -84,     89,    -90,     90,    -86,     80,    -72,     62,    -49,     36,    -21,      6,     10,    -25,     40,    -53,     65,    -75,     82,    -88,     90,    -90,     88,    -82,     75,    -65,     53,    -40,     25,    -10,     -6,     21,    -36,     49,    -62,     72,    -80,     86,    -90,     90,    -89,     84,    -77,     68,    -57,     44,    -29,     14,      1,    -17,     32,    -46,     58,    -69,     78,    -85,     89,    -91,     89,    -86,     79,    -71,     60,    -47,     34,    -19,      3,     12,    -27,     42,    -55,     66,    -76,     83,    -88,     90,    -90,     87,    -81,     73,    -63,     51,    -38,     23,     -8 },
      {      7,    -20,     33,    -45,     56,    -66,     74,    -81,     86,    -89,     90,    -90,     87,    -83,     76,    -69,     59,    -48,     37,    -24,     11,      2,    -15,     28,    -41,     52,    -62,     71,    -79,     84,    -88,     90,    -90,     88,    -84,     79,    -71,     62,    -52,     41,    -28,     15,     -2,    -11,     24,    -37,     48,    -59,     69,    -76,     83,    -87,     90,    -90,     89,    -86,     81,    -74,     66,    -56,     45,    -33,     20,     -7,     -7,     20,    -33,     45,    -56,     66,    -74,     81,    -86,     89,    -90,     90,    -87,     83,    -76,     69,    -59,     48,    -37,     24,    -11,     -2,     15,    -28,     41,    -52,     62,    -71,     79,    -84,     88,    -90,     90,    -88,     84,    -79,     71,    -62,     52,    -41,     28,    -15,      2,     11,    -24,     37,    -48,     59,    -69,     76,    -83,     87,    -90,     90,    -89,     86,    -81,     74,    -66,     56,    -45,     33,    -20,      7 },
      {      6,    -17,     27,    -38,     47,    -57,     65,    -72,     78,    -83,     87,    -89,     90,    -90,     89,    -86,     81,    -76,     69,    -62,     53,    -44,     34,    -23,     12,     -1,    -10,     21,    -32,     42,    -51,     60,    -68,     75,    -80,     85,    -88,     90,    -91,     90,    -88,     84,    -79,     73,    -66,     58,    -49,     40,    -29,     19,     -8,     -3,     14,    -25,     36,    -46,     55,    -63,     71,    -77,     82,    -86,     89,    -90,     90,    -89,     86,    -82,     77,    -71,     63,    -55,     46,    -36,     25,    -14,      3,      8,    -19,     29,    -40,     49,    -58,     66,    -73,     79,    -84,     88,    -90,     91,    -90,     88,    -85,     80,    -75,     68,    -60,     51,    -42,     32,    -21,     10,      1,    -12,     23,    -34,     44,    -53,     62,    -69,     76,    -81,     86,    -89,     90,    -90,     89,    -87,     83,    -78,     72,    -65,     57,    -47,     38,    -27,     17,     -6 },
      {      4,    -13,     22,    -30,     39,    -47,     54,    -61,     67,    -73,     78,    -82,     85,    -88,     90,    -90,     90,    -90,     88,    -85,     82,    -78,     73,    -67,     61,    -54,     47,    -39,     30,    -22,     13,     -4,     -4,     13,    -22,     30,    -39,     47,    -54,     61,    -67,     73,    -78,     82,    -85,     88,    -90,     90,    -90,     90,    -88,     85,    -82,     78,    -73,     67,    -61,     54,    -47,     39,    -30,     22,    -13,      4,      4,    -13,     22,    -30,     39,    -47,     54,    -61,     67,    -73,     78,    -82,     85,    -88,     90,    -90,     90,    -90,     88,    -85,     82,    -78,     73,    -67,     61,    -54,     47,    -39,     30,    -22,     13,     -4,     -4,     13,    -22,     30,    -39,     47,    -54,     61,    -67,     73,    -78,     82,    -85,     88,    -90,     90,    -90,     90,    -88,     85,    -82,     78,    -73,     67,    -61,     54,    -47,     39,    -30,     22,    -13,      4 },
      {      3,    -10,     17,    -23,     29,    -36,     42,    -47,     53,    -58,     63,    -68,     72,    -76,     79,    -82,     85,    -87,     89,    -90,     90,    -91,     90,    -89,     88,    -86,     84,    -81,     78,    -75,     71,    -66,     62,    -57,     51,    -46,     40,    -34,     27,    -21,     14,     -8,      1,      6,    -12,     19,    -25,     32,    -38,     44,    -49,     55,    -60,     65,    -69,     73,    -77,     80,    -83,     86,    -88,     89,    -90,     90,    -90,     90,    -89,     88,    -86,     83,    -80,     77,    -73,     69,    -65,     60,    -55,     49,    -44,     38,    -32,     25,    -19,     12,     -6,     -1,      8,    -14,     21,    -27,     34,    -40,     46,    -51,     57,    -62,     66,    -71,     75,    -78,     81,    -84,     86,    -88,     89,    -90,     91,    -90,     90,    -89,     87,    -85,     82,    -79,     76,    -72,     68,    -63,     58,    -53,     47,    -42,     36,    -29,     23,    -17,     10,     -3 },
      {      2,     -7,     11,    -15,     20,    -24,     28,    -33,     37,    -41,     45,    -48,     52,    -56,     59,    -62,     66,    -69,     71,    -74,     76,    -79,     81,    -83,     84,    -86,     87,    -88,     89,    -90,     90,    -90,     90,    -90,     90,    -89,     88,    -87,     86,    -84,     83,    -81,     79,    -76,     74,    -71,     69,    -66,     62,    -59,     56,    -52,     48,    -45,     41,    -37,     33,    -28,     24,    -20,     15,    -11,      7,     -2,     -2,      7,    -11,     15,    -20,     24,    -28,     33,    -37,     41,    -45,     48,    -52,     56,    -59,     62,    -66,     69,    -71,     74,    -76,     79,    -81,     83,    -84,     86,    -87,     88,    -89,     90,    -90,     90,    -90,     90,    -90,     89,    -88,     87,    -86,     84,    -83,     81,    -79,     76,    -74,     71,    -69,     66,    -62,     59,    -56,     52,    -48,     45,    -41,     37,    -33,     28,    -24,     20,    -15,     11,     -7,      2 },
      {      1,     -3,      6,     -8,     10,    -12,     14,    -17,     19,    -21,     23,    -25,     27,    -29,     32,    -34,     36,    -38,     40,    -42,     44,    -46,     47,    -49,     51,    -53,     55,    -57,     58,    -60,     62,    -63,     65,    -66,     68,    -69,     71,    -72,     73,    -75,     76,    -77,     78,    -79,     80,    -81,     82,    -83,     84,    -85,     86,    -86,     87,    -88,     88,    -89,     89,    -89,     90,    -90,     90,    -90,     90,    -91,     91,    -90,     90,    -90,     90,    -90,     89,    -89,     89,    -88,     88,    -87,     86,    -86,     85,    -84,     83,    -82,     81,    -80,     79,    -78,     77,    -76,     75,    -73,     72,    -71,     69,    -68,     66,    -65,     63,    -62,     60,    -58,     57,    -55,     53,    -51,     49,    -47,     46,    -44,     42,    -40,     38,    -36,     34,    -32,     29,    -27,     25,    -23,     21,    -19,     17,    -14,     12,    -10,      8,     -6,      3,     -1 },
    },
};
#endif

// DCT-8
const TMatrixCoeff g_trCoreDCT8P4[TRANSFORM_NUMBER_OF_DIRECTIONS][4][4] =
{
  DEFINE_DCT8_P4_MATRIX(84,     74,     55,     29),
  DEFINE_DCT8_P4_MATRIX(84,     74,     55,     29)
};
const TMatrixCoeff g_trCoreDCT8P8[TRANSFORM_NUMBER_OF_DIRECTIONS][8][8] =
{
  DEFINE_DCT8_P8_MATRIX(86,     85,     78,     71,     60,     46,     32,     17),
  DEFINE_DCT8_P8_MATRIX(86,     85,     78,     71,     60,     46,     32,     17)
};
const TMatrixCoeff g_trCoreDCT8P16[TRANSFORM_NUMBER_OF_DIRECTIONS][16][16] =
{
  DEFINE_DCT8_P16_MATRIX(88,     88,     87,     85,     81,     77,     73,     68,     62,     55,     48,     40,     33,     25,     17,      8),
  DEFINE_DCT8_P16_MATRIX(88,     88,     87,     85,     81,     77,     73,     68,     62,     55,     48,     40,     33,     25,     17,      8)
};
const TMatrixCoeff g_trCoreDCT8P32[TRANSFORM_NUMBER_OF_DIRECTIONS][32][32] =
{
  DEFINE_DCT8_P32_MATRIX(90,     90,     89,     88,     87,     86,     85,     84,     82,     80,     78,     77,     74,     72,     68,     66,     63,     60,     56,     53,     50,     46,     42,     38,     34,     30,     26,     21,     17,     13,      9,      4),
  DEFINE_DCT8_P32_MATRIX(90,     90,     89,     88,     87,     86,     85,     84,     82,     80,     78,     77,     74,     72,     68,     66,     63,     60,     56,     53,     50,     46,     42,     38,     34,     30,     26,     21,     17,     13,      9,      4)
};
#if TU_256
const TMatrixCoeff g_trCoreDCT8P64[TRANSFORM_NUMBER_OF_DIRECTIONS][64][64] =
{
    {
      {     90,     90,     90,     90,     90,     89,     89,     89,     88,     88,     87,     87,     86,     85,     85,     84,     83,     82,     81,     80,     79,     78,     77,     76,     75,     73,     72,     71,     69,     68,     66,     65,     63,     62,     60,     59,     57,     55,     53,     52,     50,     48,     46,     44,     42,     40,     38,     36,     34,     32,     30,     28,     26,     24,     22,     20,     17,     15,     13,     11,      9,      7,      4,      2 },
      {     90,     90,     89,     87,     85,     83,     80,     77,     73,     69,     65,     60,     55,     50,     44,     38,     32,     26,     20,     13,      7,      0,     -7,    -13,    -20,    -26,    -32,    -38,    -44,    -50,    -55,    -60,    -65,    -69,    -73,    -77,    -80,    -83,    -85,    -87,    -89,    -90,    -90,    -90,    -90,    -89,    -87,    -85,    -83,    -80,    -77,    -73,    -69,    -65,    -60,    -55,    -50,    -44,    -38,    -32,    -26,    -20,    -13,     -7 },
      {     90,     89,     86,     82,     77,     71,     63,     55,     46,     36,     26,     15,      4,     -7,    -17,    -28,    -38,    -48,    -57,    -65,    -72,    -78,    -83,    -87,    -89,    -90,    -90,    -88,    -85,    -81,    -76,    -69,    -62,    -53,    -44,    -34,    -24,    -13,     -2,      9,     20,     30,     40,     50,     59,     66,     73,     79,     84,     87,     89,     90,     90,     88,     85,     80,     75,     68,     60,     52,     42,     32,     22,     11 },
      {     90,     87,     82,     75,     65,     53,     40,     26,     11,     -4,    -20,    -34,    -48,    -60,    -71,    -79,    -85,    -89,    -90,    -89,    -85,    -78,    -69,    -59,    -46,    -32,    -17,     -2,     13,     28,     42,     55,     66,     76,     83,     88,     90,     90,     87,     81,     73,     63,     52,     38,     24,      9,     -7,    -22,    -36,    -50,    -62,    -72,    -80,    -86,    -89,    -90,    -88,    -84,    -77,    -68,    -57,    -44,    -30,    -15 },
      {     90,     85,     77,     65,     50,     32,     13,     -7,    -26,    -44,    -60,    -73,    -83,    -89,    -90,    -87,    -80,    -69,    -55,    -38,    -20,      0,     20,     38,     55,     69,     80,     87,     90,     89,     83,     73,     60,     44,     26,      7,    -13,    -32,    -50,    -65,    -77,    -85,    -90,    -90,    -85,    -77,    -65,    -50,    -32,    -13,      7,     26,     44,     60,     73,     83,     89,     90,     87,     80,     69,     55,     38,     20 },
      {     89,     83,     71,     53,     32,      9,    -15,    -38,    -59,    -75,    -85,    -90,    -88,    -80,    -66,    -48,    -26,     -2,     22,     44,     63,     78,     87,     90,     87,     77,     62,     42,     20,     -4,    -28,    -50,    -68,    -81,    -89,    -90,    -85,    -73,    -57,    -36,    -13,     11,     34,     55,     72,     84,     90,     89,     82,     69,     52,     30,      7,    -17,    -40,    -60,    -76,    -86,    -90,    -88,    -79,    -65,    -46,    -24 },
      {     89,     80,     63,     40,     13,    -15,    -42,    -65,    -81,    -89,    -89,    -79,    -62,    -38,    -11,     17,     44,     66,     82,     90,     88,     78,     60,     36,      9,    -20,    -46,    -68,    -83,    -90,    -88,    -77,    -59,    -34,     -7,     22,     48,     69,     84,     90,     87,     76,     57,     32,      4,    -24,    -50,    -71,    -85,    -90,    -87,    -75,    -55,    -30,     -2,     26,     52,     72,     85,     90,     86,     73,     53,     28 },
      {     89,     77,     55,     26,     -7,    -38,    -65,    -83,    -90,    -85,    -69,    -44,    -13,     20,     50,     73,     87,     90,     80,     60,     32,      0,    -32,    -60,    -80,    -90,    -87,    -73,    -50,    -20,     13,     44,     69,     85,     90,     83,     65,     38,      7,    -26,    -55,    -77,    -89,    -89,    -77,    -55,    -26,      7,     38,     65,     83,     90,     85,     69,     44,     13,    -20,    -50,    -73,    -87,    -90,    -80,    -60,    -32 },
      {     88,     73,     46,     11,    -26,    -59,    -81,    -90,    -84,    -63,    -32,      4,     40,     69,     87,     89,     77,     52,     17,    -20,    -53,    -78,    -90,    -86,    -68,    -38,     -2,     34,     65,     85,     90,     80,     57,     24,    -13,    -48,    -75,    -89,    -88,    -72,    -44,     -9,     28,     60,     82,     90,     83,     62,     30,     -7,    -42,    -71,    -87,    -89,    -76,    -50,    -15,     22,     55,     79,     90,     85,     66,     36 },
      {     88,     69,     36,     -4,    -44,    -75,    -89,    -85,    -63,    -28,     13,     52,     79,     90,     82,     57,     20,    -22,    -59,    -83,    -90,    -78,    -50,    -11,     30,     65,     86,     89,     73,     42,      2,    -38,    -71,    -88,    -87,    -68,    -34,      7,     46,     76,     90,     85,     62,     26,    -15,    -53,    -80,    -90,    -81,    -55,    -17,     24,     60,     84,     90,     77,     48,      9,    -32,    -66,    -87,    -89,    -72,    -40 },
      {     87,     65,     26,    -20,    -60,    -85,    -89,    -69,    -32,     13,     55,     83,     90,     73,     38,     -7,    -50,    -80,    -90,    -77,    -44,      0,     44,     77,     90,     80,     50,      7,    -38,    -73,    -90,    -83,    -55,    -13,     32,     69,     89,     85,     60,     20,    -26,    -65,    -87,    -87,    -65,    -26,     20,     60,     85,     89,     69,     32,    -13,    -55,    -83,    -90,    -73,    -38,      7,     50,     80,     90,     77,     44 },
      {     87,     60,     15,    -34,    -73,    -90,    -79,    -44,      4,     52,     83,     89,     68,     26,    -24,    -66,    -89,    -84,    -53,     -7,     42,     78,     90,     75,     36,    -13,    -59,    -86,    -87,    -62,    -17,     32,     72,     90,     80,     46,     -2,    -50,    -82,    -89,    -69,    -28,     22,     65,     88,     85,     55,      9,    -40,    -77,    -90,    -76,    -38,     11,     57,     85,     88,     63,     20,    -30,    -71,    -90,    -81,    -48 },
      {     86,     55,      4,    -48,    -83,    -88,    -62,    -13,     40,     79,     90,     68,     22,    -32,    -75,    -90,    -73,    -30,     24,     69,     90,     78,     38,    -15,    -63,    -89,    -82,    -46,      7,     57,     87,     85,     53,      2,    -50,    -84,    -88,    -60,    -11,     42,     80,     89,     66,     20,    -34,    -76,    -90,    -72,    -28,     26,     71,     90,     77,     36,    -17,    -65,    -89,    -81,    -44,      9,     59,     87,     85,     52 },
      {     85,     50,     -7,    -60,    -89,    -80,    -38,     20,     69,     90,     73,     26,    -32,    -77,    -90,    -65,    -13,     44,     83,     87,     55,      0,    -55,    -87,    -83,    -44,     13,     65,     90,     77,     32,    -26,    -73,    -90,    -69,    -20,     38,     80,     89,     60,      7,    -50,    -85,    -85,    -50,      7,     60,     89,     80,     38,    -20,    -69,    -90,    -73,    -26,     32,     77,     90,     65,     13,    -44,    -83,    -87,    -55 },
      {     85,     44,    -17,    -71,    -90,    -66,    -11,     50,     87,     82,     38,    -24,    -75,    -90,    -62,     -4,     55,     88,     79,     32,    -30,    -78,    -89,    -57,      2,     60,     89,     76,     26,    -36,    -81,    -87,    -52,      9,     65,     90,     72,     20,    -42,    -84,    -85,    -46,     15,     69,     90,     68,     13,    -48,    -86,    -83,    -40,     22,     73,     90,     63,      7,    -53,    -88,    -80,    -34,     28,     77,     89,     59 },
      {     84,     38,    -28,    -79,    -87,    -48,     17,     73,     89,     57,     -7,    -66,    -90,    -65,     -4,     59,     90,     72,     15,    -50,    -88,    -78,    -26,     40,     85,     83,     36,    -30,    -80,    -87,    -46,     20,     75,     89,     55,     -9,    -68,    -90,    -63,     -2,     60,     90,     71,     13,    -52,    -88,    -77,    -24,     42,     85,     82,     34,    -32,    -81,    -86,    -44,     22,     76,     89,     53,    -11,    -69,    -90,    -62 },
      {     83,     32,    -38,    -85,    -80,    -26,     44,     87,     77,     20,    -50,    -89,    -73,    -13,     55,     90,     69,      7,    -60,    -90,    -65,      0,     65,     90,     60,     -7,    -69,    -90,    -55,     13,     73,     89,     50,    -20,    -77,    -87,    -44,     26,     80,     85,     38,    -32,    -83,    -83,    -32,     38,     85,     80,     26,    -44,    -87,    -77,    -20,     50,     89,     73,     13,    -55,    -90,    -69,     -7,     60,     90,     65 },
      {     82,     26,    -48,    -89,    -69,     -2,     66,     90,     52,    -22,    -80,    -84,    -30,     44,     88,     72,      7,    -63,    -90,    -55,     17,     78,     85,     34,    -40,    -87,    -75,    -11,     60,     90,     59,    -13,    -76,    -87,    -38,     36,     86,     77,     15,    -57,    -90,    -62,      9,     73,     88,     42,    -32,    -85,    -79,    -20,     53,     90,     65,     -4,    -71,    -89,    -46,     28,     83,     81,     24,    -50,    -89,    -68 },
      {     81,     20,    -57,    -90,    -55,     22,     82,     80,     17,    -59,    -90,    -53,     24,     83,     79,     15,    -60,    -90,    -52,     26,     84,     78,     13,    -62,    -90,    -50,     28,     85,     77,     11,    -63,    -90,    -48,     30,     85,     76,      9,    -65,    -89,    -46,     32,     86,     75,      7,    -66,    -89,    -44,     34,     87,     73,      4,    -68,    -89,    -42,     36,     87,     72,      2,    -69,    -88,    -40,     38,     88,     71 },
      {     80,     13,    -65,    -89,    -38,     44,     90,     60,    -20,    -83,    -77,     -7,     69,     87,     32,    -50,    -90,    -55,     26,     85,     73,      0,    -73,    -85,    -26,     55,     90,     50,    -32,    -87,    -69,      7,     77,     83,     20,    -60,    -90,    -44,     38,     89,     65,    -13,    -80,    -80,    -13,     65,     89,     38,    -44,    -90,    -60,     20,     83,     77,      7,    -69,    -87,    -32,     50,     90,     55,    -26,    -85,    -73 },
      {     79,      7,    -72,    -85,    -20,     63,     88,     32,    -53,    -90,    -44,     42,     90,     55,    -30,    -88,    -65,     17,     84,     73,     -4,    -78,    -80,     -9,     71,     85,     22,    -62,    -89,    -34,     52,     90,     46,    -40,    -90,    -57,     28,     87,     66,    -15,    -83,    -75,      2,     77,     81,     11,    -69,    -86,    -24,     60,     89,     36,    -50,    -90,    -48,     38,     89,     59,    -26,    -87,    -68,     13,     82,     76 },
      {     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78 },
      {     77,     -7,    -83,    -69,     20,     87,     60,    -32,    -90,    -50,     44,     90,     38,    -55,    -89,    -26,     65,     85,     13,    -73,    -80,      0,     80,     73,    -13,    -85,    -65,     26,     89,     55,    -38,    -90,    -44,     50,     90,     32,    -60,    -87,    -20,     69,     83,      7,    -77,    -77,      7,     83,     69,    -20,    -87,    -60,     32,     90,     50,    -44,    -90,    -38,     55,     89,     26,    -65,    -85,    -13,     73,     80 },
      {     76,    -13,    -87,    -59,     38,     90,     36,    -60,    -86,    -11,     77,     75,    -15,    -87,    -57,     40,     90,     34,    -62,    -85,     -9,     78,     73,    -17,    -88,    -55,     42,     90,     32,    -63,    -85,     -7,     79,     72,    -20,    -88,    -53,     44,     90,     30,    -65,    -84,     -4,     80,     71,    -22,    -89,    -52,     46,     90,     28,    -66,    -83,     -2,     81,     69,    -24,    -89,    -50,     48,     89,     26,    -68,    -82 },
      {     75,    -20,    -89,    -46,     55,     87,      9,    -80,    -68,     30,     90,     36,    -63,    -83,      2,     85,     60,    -40,    -90,    -26,     71,     78,    -13,    -88,    -52,     50,     88,     15,    -77,    -72,     24,     90,     42,    -59,    -85,     -4,     82,     65,    -34,    -90,    -32,     66,     81,     -7,    -86,    -57,     44,     89,     22,    -73,    -76,     17,     89,     48,    -53,    -87,    -11,     79,     69,    -28,    -90,    -38,     62,     84 },
      {     73,    -26,    -90,    -32,     69,     77,    -20,    -90,    -38,     65,     80,    -13,    -89,    -44,     60,     83,     -7,    -87,    -50,     55,     85,      0,    -85,    -55,     50,     87,      7,    -83,    -60,     44,     89,     13,    -80,    -65,     38,     90,     20,    -77,    -69,     32,     90,     26,    -73,    -73,     26,     90,     32,    -69,    -77,     20,     90,     38,    -65,    -80,     13,     89,     44,    -60,    -83,      7,     87,     50,    -55,    -85 },
      {     72,    -32,    -90,    -17,     80,     62,    -46,    -87,     -2,     86,     50,    -59,    -82,     13,     89,     36,    -69,    -75,     28,     90,     22,    -78,    -65,     42,     88,      7,    -85,    -53,     55,     84,     -9,    -89,    -40,     66,     77,    -24,    -90,    -26,     76,     68,    -38,    -89,    -11,     83,     57,    -52,    -85,      4,     88,     44,    -63,    -79,     20,     90,     30,    -73,    -71,     34,     90,     15,    -81,    -60,     48,     87 },
      {     71,    -38,    -88,     -2,     87,     42,    -68,    -73,     34,     89,      7,    -86,    -46,     65,     76,    -30,    -90,    -11,     85,     50,    -62,    -78,     26,     90,     15,    -83,    -53,     59,     80,    -22,    -90,    -20,     81,     57,    -55,    -82,     17,     90,     24,    -79,    -60,     52,     84,    -13,    -90,    -28,     77,     63,    -48,    -85,      9,     89,     32,    -75,    -66,     44,     87,     -4,    -89,    -36,     72,     69,    -40,    -88 },
      {     69,    -44,    -85,     13,     90,     20,    -83,    -50,     65,     73,    -38,    -87,      7,     90,     26,    -80,    -55,     60,     77,    -32,    -89,      0,     89,     32,    -77,    -60,     55,     80,    -26,    -90,     -7,     87,     38,    -73,    -65,     50,     83,    -20,    -90,    -13,     85,     44,    -69,    -69,     44,     85,    -13,    -90,    -20,     83,     50,    -65,    -73,     38,     87,     -7,    -90,    -26,     80,     55,    -60,    -77,     32,     89 },
      {     68,    -50,    -81,     28,     89,     -4,    -90,    -20,     85,     42,    -73,    -62,     57,     77,    -36,    -87,     13,     90,     11,    -87,    -34,     78,     55,    -63,    -72,     44,     84,    -22,    -90,     -2,     89,     26,    -82,    -48,     69,     66,    -52,    -80,     30,     88,     -7,    -90,    -17,     85,     40,    -75,    -60,     59,     76,    -38,    -86,     15,     90,      9,    -88,    -32,     79,     53,    -65,    -71,     46,     83,    -24,    -89 },
      {     66,    -55,    -76,     42,     83,    -28,    -88,     13,     90,      2,    -90,    -17,     87,     32,    -81,    -46,     73,     59,    -63,    -69,     52,     78,    -38,    -85,     24,     89,     -9,    -90,     -7,     89,     22,    -85,    -36,     79,     50,    -71,    -62,     60,     72,    -48,    -80,     34,     86,    -20,    -89,      4,     90,     11,    -88,    -26,     84,     40,    -77,    -53,     68,     65,    -57,    -75,     44,     82,    -30,    -87,     15,     90 },
      {     65,    -60,    -69,     55,     73,    -50,    -77,     44,     80,    -38,    -83,     32,     85,    -26,    -87,     20,     89,    -13,    -90,      7,     90,      0,    -90,     -7,     90,     13,    -89,    -20,     87,     26,    -85,    -32,     83,     38,    -80,    -44,     77,     50,    -73,    -55,     69,     60,    -65,    -65,     60,     69,    -55,    -73,     50,     77,    -44,    -80,     38,     83,    -32,    -85,     26,     87,    -20,    -89,     13,     90,     -7,    -90 },
      {     63,    -65,    -62,     66,     60,    -68,    -59,     69,     57,    -71,    -55,     72,     53,    -73,    -52,     75,     50,    -76,    -48,     77,     46,    -78,    -44,     79,     42,    -80,    -40,     81,     38,    -82,    -36,     83,     34,    -84,    -32,     85,     30,    -85,    -28,     86,     26,    -87,    -24,     87,     22,    -88,    -20,     88,     17,    -89,    -15,     89,     13,    -89,    -11,     90,      9,    -90,     -7,     90,      4,    -90,     -2,     90 },
      {     62,    -69,    -53,     76,     44,    -81,    -34,     85,     24,    -88,    -13,     90,      2,    -90,      9,     89,    -20,    -87,     30,     83,    -40,    -78,     50,     72,    -59,    -65,     66,     57,    -73,    -48,     79,     38,    -84,    -28,     87,     17,    -89,     -7,     90,     -4,    -90,     15,     88,    -26,    -85,     36,     80,    -46,    -75,     55,     68,    -63,    -60,     71,     52,    -77,    -42,     82,     32,    -86,    -22,     89,     11,    -90 },
      {     60,    -73,    -44,     83,     26,    -89,     -7,     90,    -13,    -87,     32,     80,    -50,    -69,     65,     55,    -77,    -38,     85,     20,    -90,      0,     90,    -20,    -85,     38,     77,    -55,    -65,     69,     50,    -80,    -32,     87,     13,    -90,      7,     89,    -26,    -83,     44,     73,    -60,    -60,     73,     44,    -83,    -26,     89,      7,    -90,     13,     87,    -32,    -80,     50,     69,    -65,    -55,     77,     38,    -85,    -20,     90 },
      {     59,    -77,    -34,     88,      7,    -90,     22,     83,    -48,    -68,     69,     46,    -84,    -20,     90,     -9,    -87,     36,     76,    -60,    -57,     78,     32,    -88,     -4,     90,    -24,    -82,     50,     66,    -71,    -44,     85,     17,    -90,     11,     87,    -38,    -75,     62,     55,    -79,    -30,     89,      2,    -89,     26,     81,    -52,    -65,     72,     42,    -85,    -15,     90,    -13,    -86,     40,     73,    -63,    -53,     80,     28,    -89 },
      {     57,    -80,    -24,     90,    -13,    -85,     48,     65,    -75,    -34,     89,     -2,    -88,     38,     72,    -68,    -44,     86,      9,    -90,     28,     78,    -60,    -53,     82,     20,    -90,     17,     83,    -52,    -62,     77,     30,    -89,      7,     87,    -42,    -69,     71,     40,    -87,     -4,     89,    -32,    -76,     63,     50,    -84,    -15,     90,    -22,    -81,     55,     59,    -79,    -26,     90,    -11,    -85,     46,     66,    -73,    -36,     88 },
      {     55,    -83,    -13,     90,    -32,    -73,     69,     38,    -89,      7,     85,    -50,    -60,     80,     20,    -90,     26,     77,    -65,    -44,     87,      0,    -87,     44,     65,    -77,    -26,     90,    -20,    -80,     60,     50,    -85,     -7,     89,    -38,    -69,     73,     32,    -90,     13,     83,    -55,    -55,     83,     13,    -90,     32,     73,    -69,    -38,     89,     -7,    -85,     50,     60,    -80,    -20,     90,    -26,    -77,     65,     44,    -87 },
      {     53,    -85,     -2,     87,    -50,    -57,     84,      7,    -88,     46,     60,    -82,    -11,     89,    -42,    -63,     80,     15,    -89,     38,     66,    -78,    -20,     90,    -34,    -69,     76,     24,    -90,     30,     72,    -73,    -28,     90,    -26,    -75,     71,     32,    -90,     22,     77,    -68,    -36,     90,    -17,    -79,     65,     40,    -89,     13,     81,    -62,    -44,     88,     -9,    -83,     59,     48,    -87,      4,     85,    -55,    -52,     86 },
      {     52,    -87,      9,     81,    -65,    -36,     90,    -26,    -72,     76,     20,    -89,     42,     60,    -84,     -2,     85,    -57,    -46,     89,    -15,    -78,     69,     30,    -90,     32,     68,    -79,    -13,     88,    -48,    -55,     86,     -4,    -83,     62,     40,    -90,     22,     75,    -73,    -24,     90,    -38,    -63,     82,      7,    -87,     53,     50,    -88,     11,     80,    -66,    -34,     90,    -28,    -71,     77,     17,    -89,     44,     59,    -85 },
      {     50,    -89,     20,     73,    -77,    -13,     87,    -55,    -44,     90,    -26,    -69,     80,      7,    -85,     60,     38,    -90,     32,     65,    -83,      0,     83,    -65,    -32,     90,    -38,    -60,     85,     -7,    -80,     69,     26,    -90,     44,     55,    -87,     13,     77,    -73,    -20,     89,    -50,    -50,     89,    -20,    -73,     77,     13,    -87,     55,     44,    -90,     26,     69,    -80,     -7,     85,    -60,    -38,     90,    -32,    -65,     83 },
      {     48,    -90,     30,     63,    -85,     11,     76,    -77,     -9,     85,    -65,    -28,     89,    -50,    -46,     90,    -32,    -62,     86,    -13,    -75,     78,      7,    -84,     66,     26,    -89,     52,     44,    -90,     34,     60,    -87,     15,     73,    -79,     -4,     83,    -68,    -24,     89,    -53,    -42,     90,    -36,    -59,     87,    -17,    -72,     80,      2,    -82,     69,     22,    -88,     55,     40,    -90,     38,     57,    -88,     20,     71,    -81 },
      {     46,    -90,     40,     52,    -90,     34,     57,    -89,     28,     62,    -87,     22,     66,    -85,     15,     71,    -83,      9,     75,    -80,      2,     78,    -77,     -4,     81,    -73,    -11,     84,    -69,    -17,     86,    -65,    -24,     88,    -60,    -30,     89,    -55,    -36,     90,    -50,    -42,     90,    -44,    -48,     90,    -38,    -53,     89,    -32,    -59,     88,    -26,    -63,     87,    -20,    -68,     85,    -13,    -72,     82,     -7,    -76,     79 },
      {     44,    -90,     50,     38,    -90,     55,     32,    -89,     60,     26,    -87,     65,     20,    -85,     69,     13,    -83,     73,      7,    -80,     77,      0,    -77,     80,     -7,    -73,     83,    -13,    -69,     85,    -20,    -65,     87,    -26,    -60,     89,    -32,    -55,     90,    -38,    -50,     90,    -44,    -44,     90,    -50,    -38,     90,    -55,    -32,     89,    -60,    -26,     87,    -65,    -20,     85,    -69,    -13,     83,    -73,     -7,     80,    -77 },
      {     42,    -90,     59,     24,    -85,     72,      4,    -77,     82,    -15,    -65,     88,    -34,    -50,     90,    -52,    -32,     88,    -66,    -13,     81,    -78,      7,     71,    -86,     26,     57,    -90,     44,     40,    -89,     60,     22,    -85,     73,      2,    -76,     83,    -17,    -63,     89,    -36,    -48,     90,    -53,    -30,     87,    -68,    -11,     80,    -79,      9,     69,    -87,     28,     55,    -90,     46,     38,    -89,     62,     20,    -84,     75 },
      {     40,    -89,     66,      9,    -77,     84,    -24,    -55,     90,    -53,    -26,     85,    -76,      7,     68,    -88,     38,     42,    -89,     65,     11,    -78,     83,    -22,    -57,     90,    -52,    -28,     85,    -75,      4,     69,    -88,     36,     44,    -89,     63,     13,    -79,     82,    -20,    -59,     90,    -50,    -30,     86,    -73,      2,     71,    -87,     34,     46,    -90,     62,     15,    -80,     81,    -17,    -60,     90,    -48,    -32,     87,    -72 },
      {     38,    -87,     73,     -7,    -65,     90,    -50,    -26,     83,    -80,     20,     55,    -90,     60,     13,    -77,     85,    -32,    -44,     89,    -69,      0,     69,    -89,     44,     32,    -85,     77,    -13,    -60,     90,    -55,    -20,     80,    -83,     26,     50,    -90,     65,      7,    -73,     87,    -38,    -38,     87,    -73,      7,     65,    -90,     50,     26,    -83,     80,    -20,    -55,     90,    -60,    -13,     77,    -85,     32,     44,    -89,     69 },
      {     36,    -85,     79,    -22,    -50,     89,    -71,      7,     62,    -90,     60,      9,    -72,     89,    -48,    -24,     80,    -85,     34,     38,    -86,     78,    -20,    -52,     89,    -69,      4,     63,    -90,     59,     11,    -73,     88,    -46,    -26,     81,    -84,     32,     40,    -87,     77,    -17,    -53,     90,    -68,      2,     65,    -90,     57,     13,    -75,     88,    -44,    -28,     82,    -83,     30,     42,    -87,     76,    -15,    -55,     90,    -66 },
      {     34,    -83,     84,    -36,    -32,     82,    -85,     38,     30,    -81,     85,    -40,    -28,     80,    -86,     42,     26,    -79,     87,    -44,    -24,     78,    -87,     46,     22,    -77,     88,    -48,    -20,     76,    -88,     50,     17,    -75,     89,    -52,    -15,     73,    -89,     53,     13,    -72,     89,    -55,    -11,     71,    -90,     57,      9,    -69,     90,    -59,     -7,     68,    -90,     60,      4,    -66,     90,    -62,     -2,     65,    -90,     63 },
      {     32,    -80,     87,    -50,    -13,     69,    -90,     65,     -7,    -55,     89,    -77,     26,     38,    -83,     85,    -44,    -20,     73,    -90,     60,      0,    -60,     90,    -73,     20,     44,    -85,     83,    -38,    -26,     77,    -89,     55,      7,    -65,     90,    -69,     13,     50,    -87,     80,    -32,    -32,     80,    -87,     50,     13,    -69,     90,    -65,      7,     55,    -89,     77,    -26,    -38,     83,    -85,     44,     20,    -73,     90,    -60 },
      {     30,    -77,     89,    -62,      7,     52,    -87,     83,    -42,    -17,     69,    -90,     71,    -20,    -40,     82,    -87,     53,      4,    -60,     89,    -78,     32,     28,    -76,     90,    -63,      9,     50,    -86,     84,    -44,    -15,     68,    -90,     72,    -22,    -38,     81,    -88,     55,      2,    -59,     89,    -79,     34,     26,    -75,     90,    -65,     11,     48,    -85,     85,    -46,    -13,     66,    -90,     73,    -24,    -36,     80,    -88,     57 },
      {     28,    -73,     90,    -72,     26,     30,    -75,     90,    -71,     24,     32,    -76,     90,    -69,     22,     34,    -77,     90,    -68,     20,     36,    -78,     90,    -66,     17,     38,    -79,     89,    -65,     15,     40,    -80,     89,    -63,     13,     42,    -81,     89,    -62,     11,     44,    -82,     88,    -60,      9,     46,    -83,     88,    -59,      7,     48,    -84,     87,    -57,      4,     50,    -85,     87,    -55,      2,     52,    -85,     86,    -53 },
      {     26,    -69,     90,    -80,     44,      7,    -55,     85,    -87,     60,    -13,    -38,     77,    -90,     73,    -32,    -20,     65,    -89,     83,    -50,      0,     50,    -83,     89,    -65,     20,     32,    -73,     90,    -77,     38,     13,    -60,     87,    -85,     55,     -7,    -44,     80,    -90,     69,    -26,    -26,     69,    -90,     80,    -44,     -7,     55,    -85,     87,    -60,     13,     38,    -77,     90,    -73,     32,     20,    -65,     89,    -83,     50 },
      {     24,    -65,     88,    -86,     60,    -17,    -30,     69,    -89,     84,    -55,     11,     36,    -73,     90,    -81,     50,     -4,    -42,     77,    -90,     78,    -44,     -2,     48,    -80,     90,    -75,     38,      9,    -53,     83,    -89,     71,    -32,    -15,     59,    -85,     88,    -66,     26,     22,    -63,     87,    -87,     62,    -20,    -28,     68,    -89,     85,    -57,     13,     34,    -72,     90,    -82,     52,     -7,    -40,     76,    -90,     79,    -46 },
      {     22,    -60,     85,    -89,     73,    -40,     -2,     44,    -76,     90,    -83,     57,    -17,    -26,     63,    -86,     89,    -71,     36,      7,    -48,     78,    -90,     81,    -53,     13,     30,    -66,     87,    -88,     68,    -32,    -11,     52,    -80,     90,    -79,     50,     -9,    -34,     69,    -88,     87,    -65,     28,     15,    -55,     82,    -90,     77,    -46,      4,     38,    -72,     89,    -85,     62,    -24,    -20,     59,    -84,     90,    -75,     42 },
      {     20,    -55,     80,    -90,     83,    -60,     26,     13,    -50,     77,    -90,     85,    -65,     32,      7,    -44,     73,    -89,     87,    -69,     38,      0,    -38,     69,    -87,     89,    -73,     44,     -7,    -32,     65,    -85,     90,    -77,     50,    -13,    -26,     60,    -83,     90,    -80,     55,    -20,    -20,     55,    -80,     90,    -83,     60,    -26,    -13,     50,    -77,     90,    -85,     65,    -32,     -7,     44,    -73,     89,    -87,     69,    -38 },
      {     17,    -50,     75,    -88,     89,    -76,     52,    -20,    -15,     48,    -73,     88,    -89,     77,    -53,     22,     13,    -46,     72,    -87,     89,    -78,     55,    -24,    -11,     44,    -71,     87,    -90,     79,    -57,     26,      9,    -42,     69,    -86,     90,    -80,     59,    -28,     -7,     40,    -68,     85,    -90,     81,    -60,     30,      4,    -38,     66,    -85,     90,    -82,     62,    -32,     -2,     36,    -65,     84,    -90,     83,    -63,     34 },
      {     15,    -44,     68,    -84,     90,    -86,     72,    -50,     22,      9,    -38,     63,    -81,     90,    -88,     76,    -55,     28,      2,    -32,     59,    -78,     89,    -89,     79,    -60,     34,     -4,    -26,     53,    -75,     87,    -90,     82,    -65,     40,    -11,    -20,     48,    -71,     85,    -90,     85,    -69,     46,    -17,    -13,     42,    -66,     83,    -90,     87,    -73,     52,    -24,     -7,     36,    -62,     80,    -89,     88,    -77,     57,    -30 },
      {     13,    -38,     60,    -77,     87,    -90,     85,    -73,     55,    -32,      7,     20,    -44,     65,    -80,     89,    -90,     83,    -69,     50,    -26,      0,     26,    -50,     69,    -83,     90,    -89,     80,    -65,     44,    -20,     -7,     32,    -55,     73,    -85,     90,    -87,     77,    -60,     38,    -13,    -13,     38,    -60,     77,    -87,     90,    -85,     73,    -55,     32,     -7,    -20,     44,    -65,     80,    -89,     90,    -83,     69,    -50,     26 },
      {     11,    -32,     52,    -68,     80,    -88,     90,    -87,     79,    -66,     50,    -30,      9,     13,    -34,     53,    -69,     81,    -88,     90,    -87,     78,    -65,     48,    -28,      7,     15,    -36,     55,    -71,     82,    -89,     90,    -86,     77,    -63,     46,    -26,      4,     17,    -38,     57,    -72,     83,    -89,     90,    -85,     76,    -62,     44,    -24,      2,     20,    -40,     59,    -73,     84,    -89,     90,    -85,     75,    -60,     42,    -22 },
      {      9,    -26,     42,    -57,     69,    -79,     86,    -90,     90,    -87,     80,    -71,     59,    -44,     28,    -11,     -7,     24,    -40,     55,    -68,     78,    -85,     89,    -90,     87,    -81,     72,    -60,     46,    -30,     13,      4,    -22,     38,    -53,     66,    -77,     85,    -89,     90,    -88,     82,    -73,     62,    -48,     32,    -15,     -2,     20,    -36,     52,    -65,     76,    -84,     89,    -90,     88,    -83,     75,    -63,     50,    -34,     17 },
      {      7,    -20,     32,    -44,     55,    -65,     73,    -80,     85,    -89,     90,    -90,     87,    -83,     77,    -69,     60,    -50,     38,    -26,     13,      0,    -13,     26,    -38,     50,    -60,     69,    -77,     83,    -87,     90,    -90,     89,    -85,     80,    -73,     65,    -55,     44,    -32,     20,     -7,     -7,     20,    -32,     44,    -55,     65,    -73,     80,    -85,     89,    -90,     90,    -87,     83,    -77,     69,    -60,     50,    -38,     26,    -13 },
      {      4,    -13,     22,    -30,     38,    -46,     53,    -60,     66,    -72,     77,    -81,     85,    -87,     89,    -90,     90,    -89,     88,    -85,     82,    -78,     73,    -68,     62,    -55,     48,    -40,     32,    -24,     15,     -7,     -2,     11,    -20,     28,    -36,     44,    -52,     59,    -65,     71,    -76,     80,    -84,     87,    -89,     90,    -90,     90,    -88,     86,    -83,     79,    -75,     69,    -63,     57,    -50,     42,    -34,     26,    -17,      9 },
      {      2,     -7,     11,    -15,     20,    -24,     28,    -32,     36,    -40,     44,    -48,     52,    -55,     59,    -62,     65,    -68,     71,    -73,     76,    -78,     80,    -82,     84,    -85,     87,    -88,     89,    -89,     90,    -90,     90,    -90,     90,    -89,     88,    -87,     86,    -85,     83,    -81,     79,    -77,     75,    -72,     69,    -66,     63,    -60,     57,    -53,     50,    -46,     42,    -38,     34,    -30,     26,    -22,     17,    -13,      9,     -4 },
    },
    {
      {     90,     90,     90,     90,     90,     89,     89,     89,     88,     88,     87,     87,     86,     85,     85,     84,     83,     82,     81,     80,     79,     78,     77,     76,     75,     73,     72,     71,     69,     68,     66,     65,     63,     62,     60,     59,     57,     55,     53,     52,     50,     48,     46,     44,     42,     40,     38,     36,     34,     32,     30,     28,     26,     24,     22,     20,     17,     15,     13,     11,      9,      7,      4,      2 },
      {     90,     90,     89,     87,     85,     83,     80,     77,     73,     69,     65,     60,     55,     50,     44,     38,     32,     26,     20,     13,      7,      0,     -7,    -13,    -20,    -26,    -32,    -38,    -44,    -50,    -55,    -60,    -65,    -69,    -73,    -77,    -80,    -83,    -85,    -87,    -89,    -90,    -90,    -90,    -90,    -89,    -87,    -85,    -83,    -80,    -77,    -73,    -69,    -65,    -60,    -55,    -50,    -44,    -38,    -32,    -26,    -20,    -13,     -7 },
      {     90,     89,     86,     82,     77,     71,     63,     55,     46,     36,     26,     15,      4,     -7,    -17,    -28,    -38,    -48,    -57,    -65,    -72,    -78,    -83,    -87,    -89,    -90,    -90,    -88,    -85,    -81,    -76,    -69,    -62,    -53,    -44,    -34,    -24,    -13,     -2,      9,     20,     30,     40,     50,     59,     66,     73,     79,     84,     87,     89,     90,     90,     88,     85,     80,     75,     68,     60,     52,     42,     32,     22,     11 },
      {     90,     87,     82,     75,     65,     53,     40,     26,     11,     -4,    -20,    -34,    -48,    -60,    -71,    -79,    -85,    -89,    -90,    -89,    -85,    -78,    -69,    -59,    -46,    -32,    -17,     -2,     13,     28,     42,     55,     66,     76,     83,     88,     90,     90,     87,     81,     73,     63,     52,     38,     24,      9,     -7,    -22,    -36,    -50,    -62,    -72,    -80,    -86,    -89,    -90,    -88,    -84,    -77,    -68,    -57,    -44,    -30,    -15 },
      {     90,     85,     77,     65,     50,     32,     13,     -7,    -26,    -44,    -60,    -73,    -83,    -89,    -90,    -87,    -80,    -69,    -55,    -38,    -20,      0,     20,     38,     55,     69,     80,     87,     90,     89,     83,     73,     60,     44,     26,      7,    -13,    -32,    -50,    -65,    -77,    -85,    -90,    -90,    -85,    -77,    -65,    -50,    -32,    -13,      7,     26,     44,     60,     73,     83,     89,     90,     87,     80,     69,     55,     38,     20 },
      {     89,     83,     71,     53,     32,      9,    -15,    -38,    -59,    -75,    -85,    -90,    -88,    -80,    -66,    -48,    -26,     -2,     22,     44,     63,     78,     87,     90,     87,     77,     62,     42,     20,     -4,    -28,    -50,    -68,    -81,    -89,    -90,    -85,    -73,    -57,    -36,    -13,     11,     34,     55,     72,     84,     90,     89,     82,     69,     52,     30,      7,    -17,    -40,    -60,    -76,    -86,    -90,    -88,    -79,    -65,    -46,    -24 },
      {     89,     80,     63,     40,     13,    -15,    -42,    -65,    -81,    -89,    -89,    -79,    -62,    -38,    -11,     17,     44,     66,     82,     90,     88,     78,     60,     36,      9,    -20,    -46,    -68,    -83,    -90,    -88,    -77,    -59,    -34,     -7,     22,     48,     69,     84,     90,     87,     76,     57,     32,      4,    -24,    -50,    -71,    -85,    -90,    -87,    -75,    -55,    -30,     -2,     26,     52,     72,     85,     90,     86,     73,     53,     28 },
      {     89,     77,     55,     26,     -7,    -38,    -65,    -83,    -90,    -85,    -69,    -44,    -13,     20,     50,     73,     87,     90,     80,     60,     32,      0,    -32,    -60,    -80,    -90,    -87,    -73,    -50,    -20,     13,     44,     69,     85,     90,     83,     65,     38,      7,    -26,    -55,    -77,    -89,    -89,    -77,    -55,    -26,      7,     38,     65,     83,     90,     85,     69,     44,     13,    -20,    -50,    -73,    -87,    -90,    -80,    -60,    -32 },
      {     88,     73,     46,     11,    -26,    -59,    -81,    -90,    -84,    -63,    -32,      4,     40,     69,     87,     89,     77,     52,     17,    -20,    -53,    -78,    -90,    -86,    -68,    -38,     -2,     34,     65,     85,     90,     80,     57,     24,    -13,    -48,    -75,    -89,    -88,    -72,    -44,     -9,     28,     60,     82,     90,     83,     62,     30,     -7,    -42,    -71,    -87,    -89,    -76,    -50,    -15,     22,     55,     79,     90,     85,     66,     36 },
      {     88,     69,     36,     -4,    -44,    -75,    -89,    -85,    -63,    -28,     13,     52,     79,     90,     82,     57,     20,    -22,    -59,    -83,    -90,    -78,    -50,    -11,     30,     65,     86,     89,     73,     42,      2,    -38,    -71,    -88,    -87,    -68,    -34,      7,     46,     76,     90,     85,     62,     26,    -15,    -53,    -80,    -90,    -81,    -55,    -17,     24,     60,     84,     90,     77,     48,      9,    -32,    -66,    -87,    -89,    -72,    -40 },
      {     87,     65,     26,    -20,    -60,    -85,    -89,    -69,    -32,     13,     55,     83,     90,     73,     38,     -7,    -50,    -80,    -90,    -77,    -44,      0,     44,     77,     90,     80,     50,      7,    -38,    -73,    -90,    -83,    -55,    -13,     32,     69,     89,     85,     60,     20,    -26,    -65,    -87,    -87,    -65,    -26,     20,     60,     85,     89,     69,     32,    -13,    -55,    -83,    -90,    -73,    -38,      7,     50,     80,     90,     77,     44 },
      {     87,     60,     15,    -34,    -73,    -90,    -79,    -44,      4,     52,     83,     89,     68,     26,    -24,    -66,    -89,    -84,    -53,     -7,     42,     78,     90,     75,     36,    -13,    -59,    -86,    -87,    -62,    -17,     32,     72,     90,     80,     46,     -2,    -50,    -82,    -89,    -69,    -28,     22,     65,     88,     85,     55,      9,    -40,    -77,    -90,    -76,    -38,     11,     57,     85,     88,     63,     20,    -30,    -71,    -90,    -81,    -48 },
      {     86,     55,      4,    -48,    -83,    -88,    -62,    -13,     40,     79,     90,     68,     22,    -32,    -75,    -90,    -73,    -30,     24,     69,     90,     78,     38,    -15,    -63,    -89,    -82,    -46,      7,     57,     87,     85,     53,      2,    -50,    -84,    -88,    -60,    -11,     42,     80,     89,     66,     20,    -34,    -76,    -90,    -72,    -28,     26,     71,     90,     77,     36,    -17,    -65,    -89,    -81,    -44,      9,     59,     87,     85,     52 },
      {     85,     50,     -7,    -60,    -89,    -80,    -38,     20,     69,     90,     73,     26,    -32,    -77,    -90,    -65,    -13,     44,     83,     87,     55,      0,    -55,    -87,    -83,    -44,     13,     65,     90,     77,     32,    -26,    -73,    -90,    -69,    -20,     38,     80,     89,     60,      7,    -50,    -85,    -85,    -50,      7,     60,     89,     80,     38,    -20,    -69,    -90,    -73,    -26,     32,     77,     90,     65,     13,    -44,    -83,    -87,    -55 },
      {     85,     44,    -17,    -71,    -90,    -66,    -11,     50,     87,     82,     38,    -24,    -75,    -90,    -62,     -4,     55,     88,     79,     32,    -30,    -78,    -89,    -57,      2,     60,     89,     76,     26,    -36,    -81,    -87,    -52,      9,     65,     90,     72,     20,    -42,    -84,    -85,    -46,     15,     69,     90,     68,     13,    -48,    -86,    -83,    -40,     22,     73,     90,     63,      7,    -53,    -88,    -80,    -34,     28,     77,     89,     59 },
      {     84,     38,    -28,    -79,    -87,    -48,     17,     73,     89,     57,     -7,    -66,    -90,    -65,     -4,     59,     90,     72,     15,    -50,    -88,    -78,    -26,     40,     85,     83,     36,    -30,    -80,    -87,    -46,     20,     75,     89,     55,     -9,    -68,    -90,    -63,     -2,     60,     90,     71,     13,    -52,    -88,    -77,    -24,     42,     85,     82,     34,    -32,    -81,    -86,    -44,     22,     76,     89,     53,    -11,    -69,    -90,    -62 },
      {     83,     32,    -38,    -85,    -80,    -26,     44,     87,     77,     20,    -50,    -89,    -73,    -13,     55,     90,     69,      7,    -60,    -90,    -65,      0,     65,     90,     60,     -7,    -69,    -90,    -55,     13,     73,     89,     50,    -20,    -77,    -87,    -44,     26,     80,     85,     38,    -32,    -83,    -83,    -32,     38,     85,     80,     26,    -44,    -87,    -77,    -20,     50,     89,     73,     13,    -55,    -90,    -69,     -7,     60,     90,     65 },
      {     82,     26,    -48,    -89,    -69,     -2,     66,     90,     52,    -22,    -80,    -84,    -30,     44,     88,     72,      7,    -63,    -90,    -55,     17,     78,     85,     34,    -40,    -87,    -75,    -11,     60,     90,     59,    -13,    -76,    -87,    -38,     36,     86,     77,     15,    -57,    -90,    -62,      9,     73,     88,     42,    -32,    -85,    -79,    -20,     53,     90,     65,     -4,    -71,    -89,    -46,     28,     83,     81,     24,    -50,    -89,    -68 },
      {     81,     20,    -57,    -90,    -55,     22,     82,     80,     17,    -59,    -90,    -53,     24,     83,     79,     15,    -60,    -90,    -52,     26,     84,     78,     13,    -62,    -90,    -50,     28,     85,     77,     11,    -63,    -90,    -48,     30,     85,     76,      9,    -65,    -89,    -46,     32,     86,     75,      7,    -66,    -89,    -44,     34,     87,     73,      4,    -68,    -89,    -42,     36,     87,     72,      2,    -69,    -88,    -40,     38,     88,     71 },
      {     80,     13,    -65,    -89,    -38,     44,     90,     60,    -20,    -83,    -77,     -7,     69,     87,     32,    -50,    -90,    -55,     26,     85,     73,      0,    -73,    -85,    -26,     55,     90,     50,    -32,    -87,    -69,      7,     77,     83,     20,    -60,    -90,    -44,     38,     89,     65,    -13,    -80,    -80,    -13,     65,     89,     38,    -44,    -90,    -60,     20,     83,     77,      7,    -69,    -87,    -32,     50,     90,     55,    -26,    -85,    -73 },
      {     79,      7,    -72,    -85,    -20,     63,     88,     32,    -53,    -90,    -44,     42,     90,     55,    -30,    -88,    -65,     17,     84,     73,     -4,    -78,    -80,     -9,     71,     85,     22,    -62,    -89,    -34,     52,     90,     46,    -40,    -90,    -57,     28,     87,     66,    -15,    -83,    -75,      2,     77,     81,     11,    -69,    -86,    -24,     60,     89,     36,    -50,    -90,    -48,     38,     89,     59,    -26,    -87,    -68,     13,     82,     76 },
      {     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78 },
      {     77,     -7,    -83,    -69,     20,     87,     60,    -32,    -90,    -50,     44,     90,     38,    -55,    -89,    -26,     65,     85,     13,    -73,    -80,      0,     80,     73,    -13,    -85,    -65,     26,     89,     55,    -38,    -90,    -44,     50,     90,     32,    -60,    -87,    -20,     69,     83,      7,    -77,    -77,      7,     83,     69,    -20,    -87,    -60,     32,     90,     50,    -44,    -90,    -38,     55,     89,     26,    -65,    -85,    -13,     73,     80 },
      {     76,    -13,    -87,    -59,     38,     90,     36,    -60,    -86,    -11,     77,     75,    -15,    -87,    -57,     40,     90,     34,    -62,    -85,     -9,     78,     73,    -17,    -88,    -55,     42,     90,     32,    -63,    -85,     -7,     79,     72,    -20,    -88,    -53,     44,     90,     30,    -65,    -84,     -4,     80,     71,    -22,    -89,    -52,     46,     90,     28,    -66,    -83,     -2,     81,     69,    -24,    -89,    -50,     48,     89,     26,    -68,    -82 },
      {     75,    -20,    -89,    -46,     55,     87,      9,    -80,    -68,     30,     90,     36,    -63,    -83,      2,     85,     60,    -40,    -90,    -26,     71,     78,    -13,    -88,    -52,     50,     88,     15,    -77,    -72,     24,     90,     42,    -59,    -85,     -4,     82,     65,    -34,    -90,    -32,     66,     81,     -7,    -86,    -57,     44,     89,     22,    -73,    -76,     17,     89,     48,    -53,    -87,    -11,     79,     69,    -28,    -90,    -38,     62,     84 },
      {     73,    -26,    -90,    -32,     69,     77,    -20,    -90,    -38,     65,     80,    -13,    -89,    -44,     60,     83,     -7,    -87,    -50,     55,     85,      0,    -85,    -55,     50,     87,      7,    -83,    -60,     44,     89,     13,    -80,    -65,     38,     90,     20,    -77,    -69,     32,     90,     26,    -73,    -73,     26,     90,     32,    -69,    -77,     20,     90,     38,    -65,    -80,     13,     89,     44,    -60,    -83,      7,     87,     50,    -55,    -85 },
      {     72,    -32,    -90,    -17,     80,     62,    -46,    -87,     -2,     86,     50,    -59,    -82,     13,     89,     36,    -69,    -75,     28,     90,     22,    -78,    -65,     42,     88,      7,    -85,    -53,     55,     84,     -9,    -89,    -40,     66,     77,    -24,    -90,    -26,     76,     68,    -38,    -89,    -11,     83,     57,    -52,    -85,      4,     88,     44,    -63,    -79,     20,     90,     30,    -73,    -71,     34,     90,     15,    -81,    -60,     48,     87 },
      {     71,    -38,    -88,     -2,     87,     42,    -68,    -73,     34,     89,      7,    -86,    -46,     65,     76,    -30,    -90,    -11,     85,     50,    -62,    -78,     26,     90,     15,    -83,    -53,     59,     80,    -22,    -90,    -20,     81,     57,    -55,    -82,     17,     90,     24,    -79,    -60,     52,     84,    -13,    -90,    -28,     77,     63,    -48,    -85,      9,     89,     32,    -75,    -66,     44,     87,     -4,    -89,    -36,     72,     69,    -40,    -88 },
      {     69,    -44,    -85,     13,     90,     20,    -83,    -50,     65,     73,    -38,    -87,      7,     90,     26,    -80,    -55,     60,     77,    -32,    -89,      0,     89,     32,    -77,    -60,     55,     80,    -26,    -90,     -7,     87,     38,    -73,    -65,     50,     83,    -20,    -90,    -13,     85,     44,    -69,    -69,     44,     85,    -13,    -90,    -20,     83,     50,    -65,    -73,     38,     87,     -7,    -90,    -26,     80,     55,    -60,    -77,     32,     89 },
      {     68,    -50,    -81,     28,     89,     -4,    -90,    -20,     85,     42,    -73,    -62,     57,     77,    -36,    -87,     13,     90,     11,    -87,    -34,     78,     55,    -63,    -72,     44,     84,    -22,    -90,     -2,     89,     26,    -82,    -48,     69,     66,    -52,    -80,     30,     88,     -7,    -90,    -17,     85,     40,    -75,    -60,     59,     76,    -38,    -86,     15,     90,      9,    -88,    -32,     79,     53,    -65,    -71,     46,     83,    -24,    -89 },
      {     66,    -55,    -76,     42,     83,    -28,    -88,     13,     90,      2,    -90,    -17,     87,     32,    -81,    -46,     73,     59,    -63,    -69,     52,     78,    -38,    -85,     24,     89,     -9,    -90,     -7,     89,     22,    -85,    -36,     79,     50,    -71,    -62,     60,     72,    -48,    -80,     34,     86,    -20,    -89,      4,     90,     11,    -88,    -26,     84,     40,    -77,    -53,     68,     65,    -57,    -75,     44,     82,    -30,    -87,     15,     90 },
      {     65,    -60,    -69,     55,     73,    -50,    -77,     44,     80,    -38,    -83,     32,     85,    -26,    -87,     20,     89,    -13,    -90,      7,     90,      0,    -90,     -7,     90,     13,    -89,    -20,     87,     26,    -85,    -32,     83,     38,    -80,    -44,     77,     50,    -73,    -55,     69,     60,    -65,    -65,     60,     69,    -55,    -73,     50,     77,    -44,    -80,     38,     83,    -32,    -85,     26,     87,    -20,    -89,     13,     90,     -7,    -90 },
      {     63,    -65,    -62,     66,     60,    -68,    -59,     69,     57,    -71,    -55,     72,     53,    -73,    -52,     75,     50,    -76,    -48,     77,     46,    -78,    -44,     79,     42,    -80,    -40,     81,     38,    -82,    -36,     83,     34,    -84,    -32,     85,     30,    -85,    -28,     86,     26,    -87,    -24,     87,     22,    -88,    -20,     88,     17,    -89,    -15,     89,     13,    -89,    -11,     90,      9,    -90,     -7,     90,      4,    -90,     -2,     90 },
      {     62,    -69,    -53,     76,     44,    -81,    -34,     85,     24,    -88,    -13,     90,      2,    -90,      9,     89,    -20,    -87,     30,     83,    -40,    -78,     50,     72,    -59,    -65,     66,     57,    -73,    -48,     79,     38,    -84,    -28,     87,     17,    -89,     -7,     90,     -4,    -90,     15,     88,    -26,    -85,     36,     80,    -46,    -75,     55,     68,    -63,    -60,     71,     52,    -77,    -42,     82,     32,    -86,    -22,     89,     11,    -90 },
      {     60,    -73,    -44,     83,     26,    -89,     -7,     90,    -13,    -87,     32,     80,    -50,    -69,     65,     55,    -77,    -38,     85,     20,    -90,      0,     90,    -20,    -85,     38,     77,    -55,    -65,     69,     50,    -80,    -32,     87,     13,    -90,      7,     89,    -26,    -83,     44,     73,    -60,    -60,     73,     44,    -83,    -26,     89,      7,    -90,     13,     87,    -32,    -80,     50,     69,    -65,    -55,     77,     38,    -85,    -20,     90 },
      {     59,    -77,    -34,     88,      7,    -90,     22,     83,    -48,    -68,     69,     46,    -84,    -20,     90,     -9,    -87,     36,     76,    -60,    -57,     78,     32,    -88,     -4,     90,    -24,    -82,     50,     66,    -71,    -44,     85,     17,    -90,     11,     87,    -38,    -75,     62,     55,    -79,    -30,     89,      2,    -89,     26,     81,    -52,    -65,     72,     42,    -85,    -15,     90,    -13,    -86,     40,     73,    -63,    -53,     80,     28,    -89 },
      {     57,    -80,    -24,     90,    -13,    -85,     48,     65,    -75,    -34,     89,     -2,    -88,     38,     72,    -68,    -44,     86,      9,    -90,     28,     78,    -60,    -53,     82,     20,    -90,     17,     83,    -52,    -62,     77,     30,    -89,      7,     87,    -42,    -69,     71,     40,    -87,     -4,     89,    -32,    -76,     63,     50,    -84,    -15,     90,    -22,    -81,     55,     59,    -79,    -26,     90,    -11,    -85,     46,     66,    -73,    -36,     88 },
      {     55,    -83,    -13,     90,    -32,    -73,     69,     38,    -89,      7,     85,    -50,    -60,     80,     20,    -90,     26,     77,    -65,    -44,     87,      0,    -87,     44,     65,    -77,    -26,     90,    -20,    -80,     60,     50,    -85,     -7,     89,    -38,    -69,     73,     32,    -90,     13,     83,    -55,    -55,     83,     13,    -90,     32,     73,    -69,    -38,     89,     -7,    -85,     50,     60,    -80,    -20,     90,    -26,    -77,     65,     44,    -87 },
      {     53,    -85,     -2,     87,    -50,    -57,     84,      7,    -88,     46,     60,    -82,    -11,     89,    -42,    -63,     80,     15,    -89,     38,     66,    -78,    -20,     90,    -34,    -69,     76,     24,    -90,     30,     72,    -73,    -28,     90,    -26,    -75,     71,     32,    -90,     22,     77,    -68,    -36,     90,    -17,    -79,     65,     40,    -89,     13,     81,    -62,    -44,     88,     -9,    -83,     59,     48,    -87,      4,     85,    -55,    -52,     86 },
      {     52,    -87,      9,     81,    -65,    -36,     90,    -26,    -72,     76,     20,    -89,     42,     60,    -84,     -2,     85,    -57,    -46,     89,    -15,    -78,     69,     30,    -90,     32,     68,    -79,    -13,     88,    -48,    -55,     86,     -4,    -83,     62,     40,    -90,     22,     75,    -73,    -24,     90,    -38,    -63,     82,      7,    -87,     53,     50,    -88,     11,     80,    -66,    -34,     90,    -28,    -71,     77,     17,    -89,     44,     59,    -85 },
      {     50,    -89,     20,     73,    -77,    -13,     87,    -55,    -44,     90,    -26,    -69,     80,      7,    -85,     60,     38,    -90,     32,     65,    -83,      0,     83,    -65,    -32,     90,    -38,    -60,     85,     -7,    -80,     69,     26,    -90,     44,     55,    -87,     13,     77,    -73,    -20,     89,    -50,    -50,     89,    -20,    -73,     77,     13,    -87,     55,     44,    -90,     26,     69,    -80,     -7,     85,    -60,    -38,     90,    -32,    -65,     83 },
      {     48,    -90,     30,     63,    -85,     11,     76,    -77,     -9,     85,    -65,    -28,     89,    -50,    -46,     90,    -32,    -62,     86,    -13,    -75,     78,      7,    -84,     66,     26,    -89,     52,     44,    -90,     34,     60,    -87,     15,     73,    -79,     -4,     83,    -68,    -24,     89,    -53,    -42,     90,    -36,    -59,     87,    -17,    -72,     80,      2,    -82,     69,     22,    -88,     55,     40,    -90,     38,     57,    -88,     20,     71,    -81 },
      {     46,    -90,     40,     52,    -90,     34,     57,    -89,     28,     62,    -87,     22,     66,    -85,     15,     71,    -83,      9,     75,    -80,      2,     78,    -77,     -4,     81,    -73,    -11,     84,    -69,    -17,     86,    -65,    -24,     88,    -60,    -30,     89,    -55,    -36,     90,    -50,    -42,     90,    -44,    -48,     90,    -38,    -53,     89,    -32,    -59,     88,    -26,    -63,     87,    -20,    -68,     85,    -13,    -72,     82,     -7,    -76,     79 },
      {     44,    -90,     50,     38,    -90,     55,     32,    -89,     60,     26,    -87,     65,     20,    -85,     69,     13,    -83,     73,      7,    -80,     77,      0,    -77,     80,     -7,    -73,     83,    -13,    -69,     85,    -20,    -65,     87,    -26,    -60,     89,    -32,    -55,     90,    -38,    -50,     90,    -44,    -44,     90,    -50,    -38,     90,    -55,    -32,     89,    -60,    -26,     87,    -65,    -20,     85,    -69,    -13,     83,    -73,     -7,     80,    -77 },
      {     42,    -90,     59,     24,    -85,     72,      4,    -77,     82,    -15,    -65,     88,    -34,    -50,     90,    -52,    -32,     88,    -66,    -13,     81,    -78,      7,     71,    -86,     26,     57,    -90,     44,     40,    -89,     60,     22,    -85,     73,      2,    -76,     83,    -17,    -63,     89,    -36,    -48,     90,    -53,    -30,     87,    -68,    -11,     80,    -79,      9,     69,    -87,     28,     55,    -90,     46,     38,    -89,     62,     20,    -84,     75 },
      {     40,    -89,     66,      9,    -77,     84,    -24,    -55,     90,    -53,    -26,     85,    -76,      7,     68,    -88,     38,     42,    -89,     65,     11,    -78,     83,    -22,    -57,     90,    -52,    -28,     85,    -75,      4,     69,    -88,     36,     44,    -89,     63,     13,    -79,     82,    -20,    -59,     90,    -50,    -30,     86,    -73,      2,     71,    -87,     34,     46,    -90,     62,     15,    -80,     81,    -17,    -60,     90,    -48,    -32,     87,    -72 },
      {     38,    -87,     73,     -7,    -65,     90,    -50,    -26,     83,    -80,     20,     55,    -90,     60,     13,    -77,     85,    -32,    -44,     89,    -69,      0,     69,    -89,     44,     32,    -85,     77,    -13,    -60,     90,    -55,    -20,     80,    -83,     26,     50,    -90,     65,      7,    -73,     87,    -38,    -38,     87,    -73,      7,     65,    -90,     50,     26,    -83,     80,    -20,    -55,     90,    -60,    -13,     77,    -85,     32,     44,    -89,     69 },
      {     36,    -85,     79,    -22,    -50,     89,    -71,      7,     62,    -90,     60,      9,    -72,     89,    -48,    -24,     80,    -85,     34,     38,    -86,     78,    -20,    -52,     89,    -69,      4,     63,    -90,     59,     11,    -73,     88,    -46,    -26,     81,    -84,     32,     40,    -87,     77,    -17,    -53,     90,    -68,      2,     65,    -90,     57,     13,    -75,     88,    -44,    -28,     82,    -83,     30,     42,    -87,     76,    -15,    -55,     90,    -66 },
      {     34,    -83,     84,    -36,    -32,     82,    -85,     38,     30,    -81,     85,    -40,    -28,     80,    -86,     42,     26,    -79,     87,    -44,    -24,     78,    -87,     46,     22,    -77,     88,    -48,    -20,     76,    -88,     50,     17,    -75,     89,    -52,    -15,     73,    -89,     53,     13,    -72,     89,    -55,    -11,     71,    -90,     57,      9,    -69,     90,    -59,     -7,     68,    -90,     60,      4,    -66,     90,    -62,     -2,     65,    -90,     63 },
      {     32,    -80,     87,    -50,    -13,     69,    -90,     65,     -7,    -55,     89,    -77,     26,     38,    -83,     85,    -44,    -20,     73,    -90,     60,      0,    -60,     90,    -73,     20,     44,    -85,     83,    -38,    -26,     77,    -89,     55,      7,    -65,     90,    -69,     13,     50,    -87,     80,    -32,    -32,     80,    -87,     50,     13,    -69,     90,    -65,      7,     55,    -89,     77,    -26,    -38,     83,    -85,     44,     20,    -73,     90,    -60 },
      {     30,    -77,     89,    -62,      7,     52,    -87,     83,    -42,    -17,     69,    -90,     71,    -20,    -40,     82,    -87,     53,      4,    -60,     89,    -78,     32,     28,    -76,     90,    -63,      9,     50,    -86,     84,    -44,    -15,     68,    -90,     72,    -22,    -38,     81,    -88,     55,      2,    -59,     89,    -79,     34,     26,    -75,     90,    -65,     11,     48,    -85,     85,    -46,    -13,     66,    -90,     73,    -24,    -36,     80,    -88,     57 },
      {     28,    -73,     90,    -72,     26,     30,    -75,     90,    -71,     24,     32,    -76,     90,    -69,     22,     34,    -77,     90,    -68,     20,     36,    -78,     90,    -66,     17,     38,    -79,     89,    -65,     15,     40,    -80,     89,    -63,     13,     42,    -81,     89,    -62,     11,     44,    -82,     88,    -60,      9,     46,    -83,     88,    -59,      7,     48,    -84,     87,    -57,      4,     50,    -85,     87,    -55,      2,     52,    -85,     86,    -53 },
      {     26,    -69,     90,    -80,     44,      7,    -55,     85,    -87,     60,    -13,    -38,     77,    -90,     73,    -32,    -20,     65,    -89,     83,    -50,      0,     50,    -83,     89,    -65,     20,     32,    -73,     90,    -77,     38,     13,    -60,     87,    -85,     55,     -7,    -44,     80,    -90,     69,    -26,    -26,     69,    -90,     80,    -44,     -7,     55,    -85,     87,    -60,     13,     38,    -77,     90,    -73,     32,     20,    -65,     89,    -83,     50 },
      {     24,    -65,     88,    -86,     60,    -17,    -30,     69,    -89,     84,    -55,     11,     36,    -73,     90,    -81,     50,     -4,    -42,     77,    -90,     78,    -44,     -2,     48,    -80,     90,    -75,     38,      9,    -53,     83,    -89,     71,    -32,    -15,     59,    -85,     88,    -66,     26,     22,    -63,     87,    -87,     62,    -20,    -28,     68,    -89,     85,    -57,     13,     34,    -72,     90,    -82,     52,     -7,    -40,     76,    -90,     79,    -46 },
      {     22,    -60,     85,    -89,     73,    -40,     -2,     44,    -76,     90,    -83,     57,    -17,    -26,     63,    -86,     89,    -71,     36,      7,    -48,     78,    -90,     81,    -53,     13,     30,    -66,     87,    -88,     68,    -32,    -11,     52,    -80,     90,    -79,     50,     -9,    -34,     69,    -88,     87,    -65,     28,     15,    -55,     82,    -90,     77,    -46,      4,     38,    -72,     89,    -85,     62,    -24,    -20,     59,    -84,     90,    -75,     42 },
      {     20,    -55,     80,    -90,     83,    -60,     26,     13,    -50,     77,    -90,     85,    -65,     32,      7,    -44,     73,    -89,     87,    -69,     38,      0,    -38,     69,    -87,     89,    -73,     44,     -7,    -32,     65,    -85,     90,    -77,     50,    -13,    -26,     60,    -83,     90,    -80,     55,    -20,    -20,     55,    -80,     90,    -83,     60,    -26,    -13,     50,    -77,     90,    -85,     65,    -32,     -7,     44,    -73,     89,    -87,     69,    -38 },
      {     17,    -50,     75,    -88,     89,    -76,     52,    -20,    -15,     48,    -73,     88,    -89,     77,    -53,     22,     13,    -46,     72,    -87,     89,    -78,     55,    -24,    -11,     44,    -71,     87,    -90,     79,    -57,     26,      9,    -42,     69,    -86,     90,    -80,     59,    -28,     -7,     40,    -68,     85,    -90,     81,    -60,     30,      4,    -38,     66,    -85,     90,    -82,     62,    -32,     -2,     36,    -65,     84,    -90,     83,    -63,     34 },
      {     15,    -44,     68,    -84,     90,    -86,     72,    -50,     22,      9,    -38,     63,    -81,     90,    -88,     76,    -55,     28,      2,    -32,     59,    -78,     89,    -89,     79,    -60,     34,     -4,    -26,     53,    -75,     87,    -90,     82,    -65,     40,    -11,    -20,     48,    -71,     85,    -90,     85,    -69,     46,    -17,    -13,     42,    -66,     83,    -90,     87,    -73,     52,    -24,     -7,     36,    -62,     80,    -89,     88,    -77,     57,    -30 },
      {     13,    -38,     60,    -77,     87,    -90,     85,    -73,     55,    -32,      7,     20,    -44,     65,    -80,     89,    -90,     83,    -69,     50,    -26,      0,     26,    -50,     69,    -83,     90,    -89,     80,    -65,     44,    -20,     -7,     32,    -55,     73,    -85,     90,    -87,     77,    -60,     38,    -13,    -13,     38,    -60,     77,    -87,     90,    -85,     73,    -55,     32,     -7,    -20,     44,    -65,     80,    -89,     90,    -83,     69,    -50,     26 },
      {     11,    -32,     52,    -68,     80,    -88,     90,    -87,     79,    -66,     50,    -30,      9,     13,    -34,     53,    -69,     81,    -88,     90,    -87,     78,    -65,     48,    -28,      7,     15,    -36,     55,    -71,     82,    -89,     90,    -86,     77,    -63,     46,    -26,      4,     17,    -38,     57,    -72,     83,    -89,     90,    -85,     76,    -62,     44,    -24,      2,     20,    -40,     59,    -73,     84,    -89,     90,    -85,     75,    -60,     42,    -22 },
      {      9,    -26,     42,    -57,     69,    -79,     86,    -90,     90,    -87,     80,    -71,     59,    -44,     28,    -11,     -7,     24,    -40,     55,    -68,     78,    -85,     89,    -90,     87,    -81,     72,    -60,     46,    -30,     13,      4,    -22,     38,    -53,     66,    -77,     85,    -89,     90,    -88,     82,    -73,     62,    -48,     32,    -15,     -2,     20,    -36,     52,    -65,     76,    -84,     89,    -90,     88,    -83,     75,    -63,     50,    -34,     17 },
      {      7,    -20,     32,    -44,     55,    -65,     73,    -80,     85,    -89,     90,    -90,     87,    -83,     77,    -69,     60,    -50,     38,    -26,     13,      0,    -13,     26,    -38,     50,    -60,     69,    -77,     83,    -87,     90,    -90,     89,    -85,     80,    -73,     65,    -55,     44,    -32,     20,     -7,     -7,     20,    -32,     44,    -55,     65,    -73,     80,    -85,     89,    -90,     90,    -87,     83,    -77,     69,    -60,     50,    -38,     26,    -13 },
      {      4,    -13,     22,    -30,     38,    -46,     53,    -60,     66,    -72,     77,    -81,     85,    -87,     89,    -90,     90,    -89,     88,    -85,     82,    -78,     73,    -68,     62,    -55,     48,    -40,     32,    -24,     15,     -7,     -2,     11,    -20,     28,    -36,     44,    -52,     59,    -65,     71,    -76,     80,    -84,     87,    -89,     90,    -90,     90,    -88,     86,    -83,     79,    -75,     69,    -63,     57,    -50,     42,    -34,     26,    -17,      9 },
      {      2,     -7,     11,    -15,     20,    -24,     28,    -32,     36,    -40,     44,    -48,     52,    -55,     59,    -62,     65,    -68,     71,    -73,     76,    -78,     80,    -82,     84,    -85,     87,    -88,     89,    -89,     90,    -90,     90,    -90,     90,    -89,     88,    -87,     86,    -85,     83,    -81,     79,    -77,     75,    -72,     69,    -66,     63,    -60,     57,    -53,     50,    -46,     42,    -38,     34,    -30,     26,    -22,     17,    -13,      9,     -4 },
    },
};
const TMatrixCoeff g_trCoreDCT8P128[TRANSFORM_NUMBER_OF_DIRECTIONS][128][128] =
{
    {
      {     90,     90,     90,     90,     90,     90,     90,     90,     90,     90,     90,     89,     89,     89,     89,     89,     89,     88,     88,     88,     88,     87,     87,     87,     86,     86,     86,     85,     85,     85,     84,     84,     83,     83,     82,     82,     81,     81,     81,     80,     79,     79,     78,     78,     77,     77,     76,     76,     75,     74,     74,     73,     72,     72,     71,     70,     70,     69,     68,     67,     67,     66,     65,     64,     64,     63,     62,     61,     60,     60,     59,     58,     57,     56,     55,     55,     54,     53,     52,     51,     50,     49,     48,     47,     46,     45,     44,     43,     42,     41,     40,     39,     38,     37,     36,     35,     34,     33,     32,     31,     30,     29,     28,     27,     26,     25,     24,     23,     22,     21,     20,     19,     18,     16,     15,     14,     13,     12,     11,     10,      9,      8,      7,      6,      4,      3,      2,      1 },
      {     90,     90,     90,     90,     89,     89,     88,     87,     86,     85,     84,     82,     81,     79,     78,     76,     74,     72,     70,     68,     66,     64,     61,     59,     56,     54,     51,     48,     45,     42,     39,     36,     33,     30,     27,     24,     21,     18,     14,     11,      8,      4,      1,     -2,     -6,     -9,    -12,    -15,    -19,    -22,    -25,    -28,    -31,    -34,    -37,    -40,    -43,    -46,    -49,    -52,    -55,    -57,    -60,    -62,    -64,    -67,    -69,    -71,    -73,    -75,    -77,    -78,    -80,    -81,    -83,    -84,    -85,    -86,    -87,    -88,    -89,    -89,    -90,    -90,    -90,    -90,    -90,    -90,    -90,    -89,    -89,    -88,    -88,    -87,    -86,    -85,    -83,    -82,    -81,    -79,    -77,    -76,    -74,    -72,    -70,    -67,    -65,    -63,    -60,    -58,    -55,    -53,    -50,    -47,    -44,    -41,    -38,    -35,    -32,    -29,    -26,    -23,    -20,    -16,    -13,    -10,     -7,     -3 },
      {     90,     90,     89,     88,     87,     85,     83,     81,     78,     76,     72,     69,     65,     61,     57,     53,     48,     43,     38,     33,     28,     23,     18,     12,      7,      1,     -4,    -10,    -15,    -21,    -26,    -31,    -36,    -41,    -46,    -51,    -55,    -60,    -64,    -67,    -71,    -74,    -77,    -80,    -82,    -85,    -86,    -88,    -89,    -90,    -90,    -90,    -90,    -90,    -89,    -88,    -86,    -84,    -82,    -79,    -77,    -74,    -70,    -67,    -63,    -59,    -55,    -50,    -45,    -40,    -35,    -30,    -25,    -20,    -14,     -9,     -3,      2,      8,     13,     19,     24,     29,     34,     39,     44,     49,     54,     58,     62,     66,     70,     73,     76,     79,     81,     84,     86,     87,     89,     89,     90,     90,     90,     90,     89,     88,     87,     85,     83,     81,     78,     75,     72,     68,     64,     60,     56,     52,     47,     42,     37,     32,     27,     22,     16,     11,      6 },
      {     90,     90,     88,     86,     84,     81,     77,     72,     67,     62,     56,     50,     43,     36,     29,     22,     14,      7,     -1,     -9,    -16,    -24,    -31,    -38,    -45,    -52,    -58,    -64,    -69,    -74,    -78,    -81,    -85,    -87,    -89,    -90,    -90,    -90,    -89,    -88,    -86,    -83,    -79,    -76,    -71,    -66,    -60,    -55,    -48,    -41,    -34,    -27,    -20,    -12,     -4,      3,     11,     19,     26,     33,     40,     47,     54,     60,     65,     70,     75,     79,     82,     85,     88,     89,     90,     90,     90,     89,     87,     85,     82,     78,     74,     70,     64,     59,     53,     46,     39,     32,     25,     18,     10,      2,     -6,    -13,    -21,    -28,    -35,    -42,    -49,    -55,    -61,    -67,    -72,    -76,    -80,    -83,    -86,    -88,    -89,    -90,    -90,    -90,    -89,    -87,    -84,    -81,    -77,    -73,    -68,    -63,    -57,    -51,    -44,    -37,    -30,    -23,    -15,     -8 },
      {     90,     89,     87,     84,     79,     74,     68,     61,     54,     45,     36,     27,     18,      8,     -2,    -12,    -22,    -31,    -40,    -49,    -57,    -64,    -71,    -77,    -81,    -85,    -88,    -90,    -90,    -90,    -88,    -86,    -82,    -77,    -72,    -65,    -58,    -50,    -41,    -32,    -23,    -13,     -3,      7,     16,     26,     35,     44,     53,     60,     67,     74,     79,     83,     87,     89,     90,     90,     89,     87,     84,     80,     75,     69,     62,     55,     46,     37,     28,     19,      9,     -1,    -11,    -21,    -30,    -39,    -48,    -56,    -64,    -70,    -76,    -81,    -85,    -88,    -90,    -90,    -90,    -89,    -86,    -82,    -78,    -72,    -66,    -59,    -51,    -42,    -33,    -24,    -14,     -4,      6,     15,     25,     34,     43,     52,     60,     67,     73,     78,     83,     86,     89,     90,     90,     89,     88,     85,     81,     76,     70,     63,     55,     47,     38,     29,     20,     10 },
      {     90,     89,     85,     81,     74,     67,     58,     48,     37,     26,     14,      2,    -10,    -22,    -33,    -44,    -55,    -64,    -72,    -78,    -84,    -88,    -90,    -90,    -89,    -87,    -82,    -77,    -70,    -61,    -52,    -41,    -30,    -19,     -7,      6,     18,     29,     40,     51,     60,     69,     76,     82,     86,     89,     90,     90,     88,     84,     79,     72,     64,     55,     45,     34,     23,     11,     -1,    -13,    -25,    -36,    -47,    -57,    -66,    -74,    -80,    -85,    -88,    -90,    -90,    -89,    -86,    -81,    -75,    -67,    -59,    -49,    -38,    -27,    -15,     -3,      9,     21,     32,     43,     54,     63,     71,     78,     83,     87,     90,     90,     89,     87,     83,     77,     70,     62,     53,     42,     31,     20,      8,     -4,    -16,    -28,    -39,    -50,    -60,    -68,    -76,    -81,    -86,    -89,    -90,    -90,    -88,    -85,    -79,    -73,    -65,    -56,    -46,    -35,    -24,    -12 },
      {     90,     88,     83,     77,     68,     58,     46,     33,     20,      6,     -9,    -23,    -36,    -49,    -60,    -70,    -78,    -85,    -89,    -90,    -90,    -87,    -82,    -75,    -66,    -55,    -43,    -30,    -16,     -2,     12,     26,     39,     52,     63,     72,     80,     86,     89,     90,     89,     86,     81,     73,     64,     53,     40,     27,     13,     -1,    -15,    -29,    -42,    -55,    -65,    -74,    -81,    -87,    -90,    -90,    -89,    -85,    -79,    -71,    -61,    -50,    -37,    -24,    -10,      4,     19,     32,     45,     57,     67,     76,     83,     88,     90,     90,     88,     84,     77,     69,     59,     47,     34,     21,      7,     -8,    -22,    -35,    -48,    -60,    -70,    -78,    -84,    -88,    -90,    -90,    -87,    -82,    -76,    -67,    -56,    -44,    -31,    -18,     -3,     11,     25,     38,     51,     62,     72,     79,     85,     89,     90,     89,     86,     81,     74,     64,     54,     41,     28,     14 },
      {     90,     87,     81,     72,     61,     48,     33,     18,      1,    -15,    -31,    -46,    -60,    -71,    -80,    -86,    -90,    -90,    -88,    -82,    -74,    -63,    -50,    -35,    -20,     -3,     13,     29,     44,     58,     70,     79,     86,     89,     90,     88,     83,     75,     64,     52,     37,     22,      6,    -11,    -27,    -42,    -56,    -68,    -78,    -85,    -89,    -90,    -89,    -84,    -76,    -66,    -54,    -39,    -24,     -8,      9,     25,     40,     55,     67,     77,     84,     89,     90,     89,     85,     77,     67,     55,     41,     26,     10,     -7,    -23,    -38,    -53,    -65,    -76,    -83,    -88,    -90,    -89,    -85,    -78,    -69,    -57,    -43,    -28,    -12,      4,     21,     36,     51,     64,     74,     82,     88,     90,     90,     86,     79,     70,     59,     45,     30,     14,     -2,    -19,    -34,    -49,    -62,    -73,    -81,    -87,    -90,    -90,    -87,    -81,    -72,    -60,    -47,    -32,    -16 },
      {     90,     86,     78,     67,     54,     37,     20,      1,    -18,    -35,    -52,    -66,    -77,    -85,    -90,    -90,    -87,    -79,    -69,    -55,    -39,    -22,     -3,     15,     33,     50,     64,     76,     85,     89,     90,     87,     81,     70,     57,     41,     24,      6,    -13,    -31,    -48,    -63,    -75,    -84,    -89,    -90,    -88,    -81,    -72,    -59,    -43,    -26,     -8,     11,     29,     46,     61,     74,     83,     89,     90,     88,     82,     73,     60,     45,     28,     10,     -9,    -27,    -44,    -60,    -72,    -82,    -88,    -90,    -89,    -83,    -74,    -62,    -47,    -30,    -12,      7,     25,     42,     58,     71,     81,     88,     90,     89,     84,     76,     64,     49,     32,     14,     -4,    -23,    -40,    -56,    -70,    -80,    -87,    -90,    -89,    -85,    -77,    -65,    -51,    -34,    -16,      2,     21,     38,     55,     68,     79,     86,     90,     90,     86,     78,     67,     53,     36,     19 },
      {     90,     85,     76,     62,     45,     26,      6,    -15,    -35,    -54,    -69,    -81,    -88,    -90,    -88,    -81,    -70,    -55,    -36,    -16,      4,     25,     44,     61,     75,     85,     90,     90,     85,     76,     63,     46,     27,      7,    -14,    -34,    -53,    -68,    -80,    -88,    -90,    -88,    -81,    -70,    -55,    -37,    -18,      3,     24,     43,     60,     74,     84,     89,     90,     86,     77,     64,     47,     28,      8,    -13,    -33,    -52,    -67,    -79,    -87,    -90,    -89,    -82,    -71,    -56,    -38,    -19,      2,     23,     42,     60,     74,     84,     89,     90,     86,     77,     64,     48,     29,      9,    -12,    -32,    -51,    -67,    -79,    -87,    -90,    -89,    -82,    -72,    -57,    -39,    -20,      1,     22,     41,     59,     73,     83,     89,     90,     86,     78,     65,     49,     30,     10,    -11,    -31,    -50,    -66,    -78,    -87,    -90,    -89,    -83,    -72,    -58,    -40,    -21 },
      {     90,     84,     72,     56,     36,     14,     -9,    -31,    -52,    -69,    -81,    -89,    -90,    -86,    -76,    -60,    -41,    -20,      3,     26,     47,     65,     79,     88,     90,     87,     78,     64,     46,     25,      2,    -21,    -42,    -61,    -76,    -86,    -90,    -89,    -81,    -68,    -51,    -30,     -8,     15,     37,     57,     73,     84,     90,     89,     83,     72,     55,     35,     13,    -10,    -32,    -53,    -70,    -82,    -89,    -90,    -85,    -75,    -60,    -40,    -19,      4,     27,     48,     66,     79,     88,     90,     87,     78,     64,     45,     24,      1,    -22,    -43,    -62,    -77,    -86,    -90,    -88,    -81,    -67,    -50,    -29,     -7,     16,     38,     58,     74,     85,     90,     89,     83,     71,     55,     34,     12,    -11,    -33,    -54,    -70,    -82,    -89,    -90,    -85,    -74,    -59,    -39,    -18,      6,     28,     49,     67,     80,     88,     90,     87,     77,     63,     44,     23 },
      {     89,     82,     69,     50,     27,      2,    -23,    -46,    -66,    -81,    -89,    -90,    -84,    -72,    -54,    -31,     -7,     19,     42,     63,     78,     88,     90,     86,     74,     57,     35,     11,    -14,    -38,    -60,    -76,    -87,    -90,    -87,    -77,    -60,    -39,    -15,     10,     34,     56,     74,     85,     90,     88,     79,     64,     43,     20,     -6,    -30,    -53,    -71,    -84,    -90,    -89,    -81,    -67,    -47,    -24,      1,     26,     49,     68,     82,     89,     90,     83,     70,     51,     28,      3,    -22,    -45,    -65,    -80,    -89,    -90,    -85,    -72,    -55,    -32,     -8,     18,     41,     62,     78,     88,     90,     86,     75,     58,     36,     12,    -13,    -37,    -59,    -76,    -86,    -90,    -87,    -77,    -61,    -40,    -16,      9,     33,     55,     73,     85,     90,     88,     79,     64,     44,     21,     -4,    -29,    -52,    -70,    -83,    -90,    -89,    -81,    -67,    -48,    -25 },
      {     89,     81,     65,     43,     18,    -10,    -36,    -60,    -77,    -88,    -90,    -84,    -70,    -50,    -25,      2,     29,     54,     73,     86,     90,     87,     75,     56,     32,      6,    -22,    -47,    -68,    -83,    -90,    -89,    -79,    -62,    -39,    -13,     14,     40,     63,     79,     89,     90,     82,     67,     46,     21,     -7,    -33,    -57,    -76,    -87,    -90,    -85,    -72,    -53,    -28,     -1,     26,     51,     71,     85,     90,     88,     77,     59,     35,      9,    -19,    -44,    -66,    -81,    -89,    -89,    -81,    -64,    -42,    -16,     11,     37,     60,     78,     88,     90,     84,     70,     49,     24,     -3,    -30,    -55,    -74,    -86,    -90,    -86,    -74,    -55,    -31,     -4,     23,     48,     69,     83,     90,     88,     78,     61,     38,     12,    -15,    -41,    -64,    -80,    -89,    -90,    -82,    -67,    -45,    -20,      8,     34,     58,     76,     87,     90,     85,     72,     52,     27 },
      {     89,     79,     61,     36,      8,    -22,    -49,    -71,    -85,    -90,    -86,    -72,    -50,    -23,      7,     35,     60,     79,     89,     89,     80,     62,     37,      9,    -21,    -48,    -70,    -85,    -90,    -86,    -72,    -51,    -24,      6,     34,     60,     78,     89,     89,     81,     63,     38,     10,    -20,    -47,    -70,    -85,    -90,    -86,    -73,    -52,    -25,      4,     33,     59,     78,     89,     90,     81,     64,     39,     11,    -19,    -46,    -69,    -84,    -90,    -87,    -74,    -53,    -26,      3,     32,     58,     77,     88,     90,     81,     64,     40,     12,    -18,    -45,    -68,    -84,    -90,    -87,    -74,    -54,    -27,      2,     31,     57,     77,     88,     90,     82,     65,     41,     13,    -16,    -44,    -67,    -83,    -90,    -87,    -75,    -55,    -28,      1,     30,     56,     76,     88,     90,     82,     66,     42,     14,    -15,    -43,    -67,    -83,    -90,    -88,    -76,    -55,    -29 },
      {     89,     78,     57,     29,     -2,    -33,    -60,    -80,    -90,    -88,    -76,    -54,    -25,      7,     37,     64,     82,     90,     87,     73,     50,     21,    -11,    -41,    -67,    -84,    -90,    -86,    -70,    -46,    -16,     15,     45,     70,     85,     90,     84,     67,     42,     12,    -20,    -49,    -72,    -87,    -90,    -82,    -64,    -38,     -8,     24,     53,     75,     88,     90,     81,     61,     34,      3,    -28,    -56,    -77,    -89,    -89,    -78,    -58,    -30,      1,     32,     60,     79,     89,     88,     76,     55,     26,     -6,    -36,    -63,    -81,    -90,    -87,    -74,    -51,    -22,     10,     40,     66,     83,     90,     86,     71,     47,     18,    -14,    -44,    -69,    -85,    -90,    -85,    -68,    -43,    -13,     19,     48,     72,     86,     90,     83,     65,     39,      9,    -23,    -52,    -74,    -88,    -90,    -81,    -62,    -35,     -4,     27,     55,     77,     89,     89,     79,     59,     31 },
      {     89,     76,     53,     22,    -12,    -44,    -70,    -86,    -90,    -81,    -60,    -31,      2,     35,     64,     83,     90,     85,     67,     40,      8,    -26,    -56,    -78,    -89,    -88,    -74,    -49,    -18,     16,     48,     73,     88,     90,     79,     57,     27,     -7,    -39,    -67,    -85,    -90,    -83,    -64,    -36,     -3,     30,     60,     81,     90,     87,     71,     45,     13,    -21,    -52,    -76,    -89,    -89,    -77,    -54,    -23,     11,     43,     70,     86,     90,     81,     61,     32,     -1,    -34,    -63,    -82,    -90,    -85,    -68,    -41,     -9,     25,     55,     78,     89,     88,     74,     50,     19,    -15,    -47,    -72,    -87,    -90,    -79,    -58,    -28,      6,     38,     66,     84,     90,     84,     65,     37,      4,    -29,    -59,    -80,    -90,    -87,    -72,    -46,    -14,     20,     51,     75,     88,     89,     77,     55,     24,    -10,    -42,    -69,    -86,    -90,    -82,    -62,    -33 },
      {     89,     74,     48,     14,    -22,    -55,    -78,    -90,    -87,    -70,    -41,     -7,     29,     60,     82,     90,     84,     64,     34,     -1,    -36,    -66,    -85,    -90,    -81,    -59,    -27,      9,     43,     71,     87,     89,     77,     53,     20,    -16,    -50,    -76,    -89,    -88,    -73,    -46,    -12,     24,     56,     79,     90,     86,     68,     39,      4,    -31,    -62,    -83,    -90,    -83,    -63,    -32,      3,     38,     67,     86,     90,     80,     57,     25,    -11,    -45,    -72,    -88,    -89,    -76,    -51,    -18,     19,     52,     77,     89,     88,     72,     44,     10,    -26,    -58,    -81,    -90,    -85,    -67,    -37,     -2,     33,     64,     84,     90,     82,     61,     30,     -6,    -40,    -69,    -86,    -90,    -79,    -55,    -23,     13,     47,     74,     88,     89,     75,     49,     15,    -21,    -54,    -78,    -90,    -87,    -70,    -42,     -8,     28,     60,     81,     90,     85,     65,     35 },
      {     88,     72,     43,      7,    -31,    -64,    -85,    -90,    -79,    -55,    -20,     19,     54,     79,     90,     85,     64,     32,     -6,    -42,    -72,    -88,    -89,    -73,    -44,     -8,     30,     63,     84,     90,     80,     55,     21,    -18,    -53,    -78,    -90,    -85,    -65,    -33,      4,     41,     71,     88,     89,     74,     45,      9,    -29,    -62,    -84,    -90,    -81,    -56,    -22,     16,     52,     78,     90,     86,     66,     34,     -3,    -40,    -70,    -88,    -89,    -74,    -46,    -10,     28,     61,     83,     90,     81,     57,     23,    -15,    -51,    -77,    -90,    -86,    -67,    -35,      2,     39,     70,     87,     89,     75,     47,     11,    -27,    -60,    -83,    -90,    -81,    -58,    -24,     14,     50,     77,     90,     86,     67,     36,     -1,    -38,    -69,    -87,    -89,    -76,    -48,    -12,     26,     60,     82,     90,     82,     59,     25,    -13,    -49,    -76,    -89,    -87,    -68,    -37 },
      {     88,     70,     38,     -1,    -40,    -72,    -89,    -88,    -69,    -36,      3,     42,     73,     89,     87,     67,     34,     -6,    -44,    -74,    -89,    -86,    -66,    -32,      8,     46,     76,     90,     86,     64,     30,    -10,    -48,    -77,    -90,    -85,    -63,    -28,     12,     50,     78,     90,     84,     61,     26,    -14,    -52,    -79,    -90,    -83,    -60,    -24,     16,     54,     80,     90,     82,     58,     22,    -19,    -55,    -81,    -90,    -81,    -56,    -20,     21,     57,     82,     90,     81,     55,     18,    -23,    -59,    -83,    -90,    -79,    -53,    -15,     25,     60,     84,     90,     78,     51,     13,    -27,    -62,    -85,    -90,    -77,    -49,    -11,     29,     64,     85,     90,     76,     47,      9,    -31,    -65,    -86,    -89,    -75,    -45,     -7,     33,     67,     87,     89,     74,     43,      4,    -35,    -68,    -87,    -89,    -72,    -41,     -2,     37,     70,     88,     88,     71,     39 },
      {     88,     68,     33,     -9,    -49,    -78,    -90,    -82,    -55,    -16,     26,     63,     86,     89,     73,     40,     -1,    -42,    -74,    -90,    -85,    -61,    -24,     19,     57,     83,     90,     77,     47,      7,    -35,    -70,    -88,    -87,    -67,    -31,     11,     51,     79,     90,     81,     54,     14,    -28,    -64,    -86,    -89,    -72,    -38,      3,     44,     76,     90,     84,     60,     22,    -21,    -59,    -84,    -90,    -76,    -45,     -4,     37,     71,     89,     87,     65,     29,    -13,    -53,    -81,    -90,    -80,    -52,    -12,     30,     66,     87,     89,     70,     36,     -6,    -46,    -77,    -90,    -83,    -58,    -20,     23,     60,     85,     90,     75,     43,      2,    -39,    -72,    -89,    -86,    -64,    -27,     15,     55,     81,     90,     79,     50,     10,    -32,    -67,    -88,    -88,    -69,    -34,      8,     48,     78,     90,     82,     56,     18,    -25,    -62,    -85,    -89,    -74,    -41 },
      {     88,     66,     28,    -16,    -57,    -84,    -90,    -74,    -39,      4,     47,     78,     90,     80,     50,      8,    -36,    -72,    -89,    -85,    -60,    -20,     25,     64,     87,     88,     68,     31,    -13,    -55,    -82,    -90,    -76,    -42,      1,     44,     77,     90,     81,     53,     11,    -33,    -70,    -89,    -86,    -62,    -23,     22,     61,     86,     89,     70,     34,    -10,    -52,    -81,    -90,    -77,    -45,     -2,     41,     75,     90,     83,     55,     14,    -30,    -67,    -88,    -87,    -64,    -26,     19,     59,     85,     89,     72,     37,     -7,    -49,    -79,    -90,    -79,    -48,     -6,     38,     73,     90,     84,     58,     18,    -27,    -65,    -87,    -88,    -67,    -29,     15,     56,     83,     90,     74,     40,     -3,    -46,    -78,    -90,    -81,    -51,     -9,     35,     71,     89,     85,     60,     21,    -24,    -63,    -86,    -89,    -69,    -32,     12,     54,     82,     90,     76,     43 },
      {     87,     64,     23,    -24,    -64,    -88,    -87,    -63,    -22,     25,     65,     88,     87,     62,     21,    -26,    -66,    -88,    -86,    -61,    -20,     27,     67,     88,     86,     60,     19,    -28,    -67,    -89,    -86,    -60,    -18,     29,     68,     89,     85,     59,     16,    -30,    -69,    -89,    -85,    -58,    -15,     31,     70,     89,     85,     57,     14,    -32,    -70,    -89,    -84,    -56,    -13,     33,     71,     89,     84,     55,     12,    -34,    -72,    -90,    -83,    -55,    -11,     35,     72,     90,     83,     54,     10,    -36,    -73,    -90,    -82,    -53,     -9,     37,     74,     90,     82,     52,      8,    -38,    -74,    -90,    -81,    -51,     -7,     39,     75,     90,     81,     50,      6,    -40,    -76,    -90,    -81,    -49,     -4,     41,     76,     90,     80,     48,      3,    -42,    -77,    -90,    -79,    -47,     -2,     43,     77,     90,     79,     46,      1,    -44,    -78,    -90,    -78,    -45 },
      {     87,     61,     18,    -31,    -71,    -90,    -82,    -50,     -3,     44,     79,     90,     75,     37,    -11,    -56,    -85,    -89,    -66,    -24,     25,     67,     89,     85,     55,     10,    -38,    -76,    -90,    -78,    -43,      4,     51,     82,     90,     70,     30,    -19,    -62,    -87,    -87,    -60,    -16,     32,     72,     90,     81,     49,      2,    -45,    -79,    -90,    -74,    -36,     12,     57,     85,     88,     65,     23,    -26,    -67,    -89,    -84,    -55,     -9,     39,     76,     90,     78,     42,     -6,    -52,    -83,    -89,    -70,    -29,     20,     63,     88,     86,     60,     15,    -33,    -72,    -90,    -81,    -48,     -1,     46,     80,     90,     74,     35,    -13,    -58,    -86,    -88,    -64,    -22,     27,     68,     89,     84,     54,      8,    -40,    -77,    -90,    -77,    -41,      7,     53,     83,     89,     69,     28,    -21,    -64,    -88,    -86,    -59,    -14,     34,     73,     90,     81,     47 },
      {     87,     59,     12,    -38,    -77,    -90,    -75,    -35,     15,     61,     88,     86,     56,      9,    -41,    -78,    -90,    -73,    -32,     19,     64,     88,     85,     54,      6,    -44,    -80,    -90,    -71,    -29,     22,     66,     89,     83,     51,      2,    -47,    -81,    -90,    -69,    -26,     25,     68,     89,     82,     48,     -1,    -50,    -83,    -89,    -67,    -23,     28,     70,     90,     81,     45,     -4,    -53,    -84,    -89,    -64,    -20,     31,     72,     90,     79,     42,     -8,    -55,    -85,    -88,    -62,    -16,     34,     74,     90,     77,     39,    -11,    -58,    -86,    -87,    -60,    -13,     37,     76,     90,     76,     36,    -14,    -60,    -87,    -86,    -57,    -10,     40,     78,     90,     74,     33,    -18,    -63,    -88,    -85,    -55,     -7,     43,     79,     90,     72,     30,    -21,    -65,    -89,    -84,    -52,     -3,     46,     81,     90,     70,     27,    -24,    -67,    -89,    -82,    -49 },
      {     86,     56,      7,    -45,    -81,    -89,    -66,    -20,     33,     75,     90,     74,     32,    -21,    -67,    -89,    -81,    -44,      8,     57,     87,     86,     55,      6,    -46,    -82,    -89,    -65,    -19,     34,     76,     90,     74,     31,    -22,    -67,    -90,    -81,    -43,      9,     58,     87,     86,     55,      4,    -47,    -82,    -89,    -64,    -18,     35,     76,     90,     73,     30,    -23,    -68,    -90,    -80,    -42,     10,     59,     87,     85,     54,      3,    -48,    -83,    -89,    -64,    -16,     36,     77,     90,     72,     29,    -24,    -69,    -90,    -79,    -41,     11,     60,     88,     85,     53,      2,    -49,    -83,    -89,    -63,    -15,     37,     77,     90,     72,     28,    -25,    -70,    -90,    -79,    -40,     12,     60,     88,     85,     52,      1,    -50,    -84,    -88,    -62,    -14,     38,     78,     90,     71,     27,    -26,    -70,    -90,    -78,    -39,     13,     61,     88,     84,     51 },
      {     86,     54,      1,    -52,    -85,    -87,    -55,     -3,     50,     85,     87,     57,      6,    -48,    -84,    -88,    -59,     -8,     46,     83,     88,     60,     10,    -44,    -82,    -89,    -62,    -12,     42,     81,     89,     64,     14,    -40,    -80,    -89,    -65,    -16,     38,     79,     90,     67,     19,    -36,    -78,    -90,    -68,    -21,     34,     77,     90,     70,     23,    -32,    -76,    -90,    -71,    -25,     30,     74,     90,     72,     27,    -28,    -73,    -90,    -74,    -29,     26,     72,     90,     75,     31,    -24,    -70,    -90,    -76,    -33,     22,     69,     90,     77,     35,    -20,    -67,    -90,    -78,    -37,     18,     66,     90,     79,     39,    -15,    -64,    -89,    -81,    -41,     13,     63,     89,     81,     43,    -11,    -61,    -89,    -82,    -45,      9,     60,     88,     83,     47,     -7,    -58,    -88,    -84,    -49,      4,     56,     87,     85,     51,     -2,    -55,    -86,    -86,    -53 },
      {     86,     51,     -4,    -58,    -88,    -82,    -43,     13,     64,     90,     78,     35,    -22,    -70,    -90,    -74,    -27,     30,     76,     90,     68,     19,    -38,    -80,    -89,    -62,    -10,     46,     84,     87,     55,      1,    -54,    -87,    -85,    -48,      8,     60,     89,     81,     40,    -16,    -67,    -90,    -77,    -32,     25,     72,     90,     72,     24,    -33,    -77,    -90,    -66,    -15,     41,     81,     89,     60,      7,    -49,    -85,    -86,    -53,      2,     56,     88,     83,     45,    -11,    -63,    -89,    -79,    -37,     20,     69,     90,     75,     29,    -28,    -74,    -90,    -70,    -21,     36,     79,     89,     64,     12,    -44,    -83,    -88,    -57,     -3,     52,     86,     85,     50,     -6,    -59,    -88,    -82,    -42,     14,     65,     90,     78,     34,    -23,    -71,    -90,    -73,    -26,     31,     76,     90,     67,     18,    -39,    -81,    -89,    -61,     -9,     47,     84,     87,     55 },
      {     85,     48,    -10,    -64,    -90,    -77,    -30,     29,     76,     90,     64,     11,    -47,    -85,    -86,    -49,      9,     63,     90,     77,     31,    -28,    -76,    -90,    -65,    -12,     46,     85,     86,     50,     -8,    -62,    -89,    -78,    -32,     27,     75,     90,     66,     13,    -45,    -84,    -86,    -51,      7,     61,     89,     78,     33,    -26,    -74,    -90,    -67,    -14,     44,     84,     87,     52,     -6,    -60,    -89,    -79,    -34,     25,     74,     90,     67,     15,    -43,    -83,    -87,    -53,      4,     60,     89,     79,     35,    -24,    -73,    -90,    -68,    -16,     42,     83,     87,     54,     -3,    -59,    -89,    -80,    -36,     23,     72,     90,     69,     18,    -41,    -82,    -88,    -55,      2,     58,     89,     81,     37,    -22,    -72,    -90,    -70,    -19,     40,     82,     88,     55,     -1,    -57,    -88,    -81,    -38,     21,     71,     90,     70,     20,    -39,    -81,    -88,    -56 },
      {     85,     45,    -15,    -69,    -90,    -70,    -16,     44,     85,     85,     46,    -14,    -68,    -90,    -70,    -18,     43,     84,     86,     47,    -13,    -67,    -90,    -71,    -19,     42,     84,     86,     48,    -12,    -67,    -90,    -72,    -20,     41,     83,     86,     49,    -11,    -66,    -90,    -72,    -21,     40,     83,     87,     50,    -10,    -65,    -90,    -73,    -22,     39,     82,     87,     51,     -9,    -64,    -90,    -74,    -23,     38,     82,     87,     52,     -8,    -64,    -90,    -74,    -24,     37,     81,     88,     53,     -7,    -63,    -90,    -75,    -25,     36,     81,     88,     54,     -6,    -62,    -90,    -76,    -26,     35,     81,     88,     55,     -4,    -61,    -90,    -76,    -27,     34,     80,     88,     55,     -3,    -60,    -89,    -77,    -28,     33,     79,     89,     56,     -2,    -60,    -89,    -77,    -29,     32,     79,     89,     57,     -1,    -59,    -89,    -78,    -30,     31,     78,     89,     58 },
      {     85,     42,    -21,    -74,    -90,    -61,     -2,     58,     89,     76,     25,    -38,    -83,    -86,    -46,     16,     71,     90,     64,      7,    -55,    -89,    -78,    -29,     34,     81,     87,     50,    -12,    -68,    -90,    -67,    -11,     51,     88,     81,     33,    -30,    -79,    -88,    -54,      8,     65,     90,     70,     15,    -47,    -86,    -82,    -37,     26,     77,     89,     57,     -3,    -62,    -90,    -73,    -20,     43,     85,     84,     41,    -22,    -74,    -90,    -60,     -1,     59,     89,     76,     24,    -39,    -83,    -86,    -45,     18,     72,     90,     64,      6,    -55,    -89,    -78,    -28,     35,     81,     87,     49,    -13,    -69,    -90,    -67,    -10,     52,     88,     80,     32,    -31,    -79,    -88,    -53,      9,     66,     90,     70,     14,    -48,    -87,    -82,    -36,     27,     77,     89,     56,     -4,    -63,    -90,    -72,    -19,     44,     85,     84,     40,    -23,    -75,    -90,    -60 },
      {     84,     39,    -26,    -78,    -88,    -52,     12,     70,     90,     63,      2,    -60,    -90,    -72,    -16,     48,     87,     80,     30,    -35,    -82,    -86,    -43,     22,     76,     89,     55,     -8,    -67,    -90,    -66,     -7,     56,     89,     75,     21,    -44,    -86,    -82,    -34,     31,     81,     87,     47,    -18,    -73,    -90,    -59,      3,     64,     90,     69,     11,    -53,    -89,    -77,    -25,     40,     85,     84,     38,    -27,    -78,    -88,    -51,     13,     70,     90,     62,      1,    -60,    -90,    -72,    -15,     49,     88,     79,     29,    -36,    -83,    -85,    -42,     23,     76,     89,     55,     -9,    -67,    -90,    -65,     -6,     57,     89,     74,     20,    -45,    -86,    -81,    -33,     32,     81,     87,     46,    -19,    -74,    -90,    -58,      4,     64,     90,     68,     10,    -54,    -89,    -77,    -24,     41,     85,     83,     37,    -28,    -79,    -88,    -50,     14,     71,     90,     61 },
      {     84,     36,    -31,    -81,    -86,    -41,     26,     79,     87,     46,    -21,    -76,    -89,    -51,     15,     73,     89,     55,    -10,    -70,    -90,    -60,      4,     66,     90,     64,      1,    -62,    -90,    -67,     -7,     58,     90,     71,     12,    -54,    -89,    -74,    -18,     49,     88,     77,     23,    -44,    -87,    -80,    -28,     39,     85,     82,     33,    -34,    -83,    -85,    -38,     29,     81,     86,     43,    -24,    -78,    -88,    -48,     19,     75,     89,     53,    -13,    -72,    -90,    -57,      8,     68,     90,     61,     -2,    -64,    -90,    -65,     -3,     60,     90,     69,      9,    -56,    -90,    -72,    -14,     52,     89,     76,     20,    -47,    -88,    -78,    -25,     42,     86,     81,     30,    -37,    -84,    -83,    -35,     32,     82,     85,     40,    -27,    -79,    -87,    -45,     22,     77,     88,     50,    -16,    -74,    -89,    -55,     11,     70,     90,     59,     -6,    -67,    -90,    -63 },
      {     83,     33,    -36,    -85,    -82,    -30,     39,     86,     81,     27,    -42,    -87,    -79,    -24,     45,     88,     77,     21,    -48,    -88,    -76,    -18,     51,     89,     74,     14,    -54,    -89,    -72,    -11,     56,     90,     70,      8,    -59,    -90,    -67,     -4,     61,     90,     65,      1,    -64,    -90,    -63,      2,     66,     90,     60,     -6,    -68,    -90,    -58,      9,     70,     90,     55,    -12,    -72,    -89,    -53,     15,     74,     89,     50,    -19,    -76,    -88,    -47,     22,     78,     87,     44,    -25,    -79,    -86,    -41,     28,     81,     85,     38,    -31,    -82,    -84,    -35,     34,     84,     83,     32,    -37,    -85,    -81,    -29,     40,     86,     80,     26,    -43,    -87,    -78,    -23,     46,     88,     77,     20,    -49,    -89,    -75,    -16,     52,     89,     73,     13,    -55,    -90,    -71,    -10,     57,     90,     69,      7,    -60,    -90,    -67,     -3,     62,     90,     64 },
      {     83,     30,    -41,    -87,    -77,    -19,     52,     89,     70,      7,    -61,    -90,    -62,      6,     70,     90,     53,    -18,    -77,    -87,    -42,     29,     82,     83,     31,    -40,    -87,    -78,    -20,     51,     89,     71,      8,    -60,    -90,    -63,      4,     69,     90,     54,    -16,    -76,    -88,    -43,     28,     82,     84,     32,    -39,    -86,    -78,    -21,     50,     89,     72,      9,    -60,    -90,    -64,      3,     68,     90,     55,    -15,    -76,    -88,    -44,     27,     81,     84,     33,    -38,    -86,    -79,    -22,     49,     89,     72,     10,    -59,    -90,    -64,      2,     67,     90,     55,    -14,    -75,    -88,    -45,     26,     81,     85,     34,    -37,    -86,    -79,    -23,     48,     89,     73,     11,    -58,    -90,    -65,      1,     67,     90,     56,    -13,    -74,    -88,    -46,     25,     81,     85,     35,    -36,    -85,    -80,    -24,     47,     89,     74,     12,    -57,    -90,    -66 },
      {     82,     27,    -46,    -89,    -72,     -7,     63,     90,     57,    -14,    -76,    -87,    -39,     34,     85,     79,     20,    -53,    -90,    -67,      1,     68,     90,     51,    -22,    -80,    -85,    -32,     41,     88,     75,     12,    -59,    -90,    -61,      9,     73,     88,     44,    -29,    -83,    -81,    -25,     48,     89,     70,      4,    -64,    -90,    -55,     16,     77,     86,     37,    -36,    -86,    -78,    -18,     55,     90,     65,     -3,    -70,    -89,    -49,     24,     81,     84,     30,    -43,    -88,    -74,    -10,     60,     90,     60,    -11,    -74,    -88,    -42,     31,     84,     81,     23,    -50,    -89,    -69,     -2,     66,     90,     54,    -19,    -78,    -86,    -35,     38,     87,     77,     15,    -56,    -90,    -64,      6,     71,     89,     47,    -26,    -82,    -83,    -28,     45,     89,     72,      8,    -62,    -90,    -58,     13,     76,     87,     40,    -33,    -85,    -79,    -21,     52,     90,     67 },
      {     82,     24,    -51,    -90,    -65,      6,     72,     88,     41,    -34,    -86,    -77,    -13,     60,     90,     57,    -16,    -78,    -85,    -31,     44,     89,     70,      2,    -67,    -89,    -48,     27,     83,     81,     21,    -54,    -90,    -63,      9,     74,     87,     38,    -37,    -87,    -75,    -10,     62,     90,     55,    -20,    -80,    -84,    -28,     47,     89,     68,     -1,    -70,    -89,    -45,     30,     85,     79,     18,    -56,    -90,    -60,     12,     76,     86,     35,    -40,    -88,    -73,     -7,     64,     90,     52,    -23,    -81,    -82,    -25,     50,     90,     66,     -4,    -72,    -88,    -42,     33,     86,     77,     14,    -59,    -90,    -58,     15,     78,     85,     32,    -43,    -89,    -71,     -3,     67,     90,     49,    -26,    -83,    -81,    -22,     53,     90,     64,     -8,    -74,    -88,    -39,     36,     87,     76,     11,    -61,    -90,    -55,     19,     79,     84,     29,    -46,    -89,    -69 },
      {     81,     21,    -55,    -90,    -58,     18,     80,     83,     24,    -53,    -90,    -60,     14,     78,     84,     27,    -50,    -90,    -63,     11,     77,     85,     30,    -47,    -90,    -65,      8,     75,     86,     33,    -44,    -89,    -67,      4,     73,     87,     36,    -41,    -89,    -70,      1,     71,     88,     39,    -38,    -88,    -72,     -2,     69,     89,     42,    -35,    -87,    -74,     -6,     67,     89,     45,    -32,    -86,    -76,     -9,     64,     90,     48,    -29,    -85,    -77,    -12,     62,     90,     51,    -26,    -84,    -79,    -15,     60,     90,     54,    -23,    -82,    -81,    -19,     57,     90,     56,    -20,    -81,    -82,    -22,     55,     90,     59,    -16,    -79,    -83,    -25,     52,     90,     61,    -13,    -78,    -85,    -28,     49,     90,     64,    -10,    -76,    -86,    -31,     46,     89,     66,     -7,    -74,    -87,    -34,     43,     89,     68,     -3,    -72,    -88,    -37,     40,     88,     70 },
      {     81,     18,    -60,    -90,    -50,     29,     86,     75,      6,    -68,    -89,    -39,     40,     89,     67,     -7,    -76,    -85,    -28,     51,     90,     59,    -19,    -81,    -81,    -16,     60,     90,     49,    -30,    -86,    -74,     -4,     69,     88,     38,    -41,    -89,    -67,      8,     76,     85,     27,    -52,    -90,    -58,     20,     82,     80,     15,    -61,    -90,    -48,     31,     86,     74,      3,    -70,    -88,    -37,     42,     89,     66,     -9,    -77,    -85,    -26,     53,     90,     57,    -21,    -82,    -79,    -14,     62,     90,     47,    -32,    -87,    -73,     -2,     70,     88,     36,    -43,    -89,    -65,     10,     77,     84,     25,    -54,    -90,    -56,     22,     83,     79,     13,    -63,    -90,    -46,     33,     87,     72,      1,    -71,    -88,    -35,     44,     89,     64,    -11,    -78,    -84,    -24,     55,     90,     55,    -23,    -83,    -78,    -12,     64,     90,     45,    -34,    -87,    -72 },
      {     81,     14,    -64,    -89,    -41,     40,     89,     64,    -13,    -80,    -81,    -15,     63,     89,     42,    -39,    -89,    -65,     12,     79,     81,     16,    -62,    -90,    -43,     38,     89,     66,    -11,    -79,    -82,    -18,     61,     90,     44,    -37,    -89,    -67,     10,     78,     82,     19,    -60,    -90,    -45,     36,     88,     67,     -9,    -78,    -83,    -20,     60,     90,     46,    -35,    -88,    -68,      8,     77,     83,     21,    -59,    -90,    -47,     34,     88,     69,     -7,    -77,    -84,    -22,     58,     90,     48,    -33,    -88,    -70,      6,     76,     84,     23,    -57,    -90,    -49,     32,     87,     70,     -4,    -76,    -85,    -24,     56,     90,     50,    -31,    -87,    -71,      3,     75,     85,     25,    -55,    -90,    -51,     30,     87,     72,     -2,    -74,    -85,    -26,     55,     90,     52,    -29,    -86,    -72,      1,     74,     86,     27,    -54,    -90,    -53,     28,     86,     73 },
      {     80,     11,    -67,    -88,    -32,     51,     90,     52,    -31,    -88,    -68,     10,     79,     81,     12,    -67,    -88,    -33,     50,     90,     53,    -30,    -87,    -69,      9,     79,     81,     13,    -66,    -88,    -34,     49,     90,     54,    -29,    -87,    -70,      8,     78,     81,     14,    -65,    -89,    -35,     48,     90,     55,    -28,    -87,    -70,      7,     78,     82,     15,    -64,    -89,    -36,     47,     90,     55,    -27,    -86,    -71,      6,     77,     82,     16,    -64,    -89,    -37,     46,     90,     56,    -26,    -86,    -72,      4,     77,     83,     18,    -63,    -89,    -38,     45,     90,     57,    -25,    -86,    -72,      3,     76,     83,     19,    -62,    -89,    -39,     44,     90,     58,    -24,    -85,    -73,      2,     76,     84,     20,    -61,    -89,    -40,     43,     90,     59,    -23,    -85,    -74,      1,     75,     84,     21,    -60,    -90,    -41,     42,     90,     60,    -22,    -85,    -74 },
      {     79,      8,    -71,    -86,    -23,     60,     89,     37,    -48,    -90,    -51,     34,     89,     63,    -20,    -85,    -73,      4,     78,     81,     11,    -69,    -87,    -26,     58,     90,     40,    -45,    -90,    -54,     31,     88,     65,    -16,    -83,    -75,      1,     76,     82,     14,    -67,    -88,    -29,     55,     90,     43,    -42,    -90,    -56,     28,     87,     67,    -13,    -82,    -77,     -2,     74,     84,     18,    -64,    -88,    -32,     53,     90,     46,    -39,    -90,    -59,     25,     86,     70,    -10,    -81,    -78,     -6,     72,     85,     21,    -62,    -89,    -35,     50,     90,     49,    -36,    -89,    -61,     22,     85,     72,     -7,    -79,    -80,     -9,     70,     86,     24,    -60,    -89,    -38,     47,     90,     52,    -33,    -89,    -64,     19,     84,     74,     -3,    -77,    -81,    -12,     68,     87,     27,    -57,    -90,    -41,     44,     90,     55,    -30,    -88,    -66,     15,     83,     76 },
      {     79,      4,    -74,    -83,    -13,     69,     86,     22,    -63,    -88,    -30,     56,     90,     38,    -49,    -90,    -46,     41,     90,     54,    -33,    -89,    -60,     25,     87,     67,    -16,    -84,    -72,      8,     81,     77,      1,    -76,    -81,    -10,     71,     85,     19,    -65,    -88,    -27,     59,     89,     35,    -52,    -90,    -43,     44,     90,     51,    -36,    -89,    -58,     28,     88,     64,    -20,    -85,    -70,     11,     82,     76,     -2,    -78,    -80,     -7,     73,     84,     15,    -67,    -87,    -24,     61,     89,     32,    -55,    -90,    -40,     47,     90,     48,    -39,    -90,    -55,     31,     89,     62,    -23,    -86,    -68,     14,     83,     74,     -6,    -79,    -78,     -3,     75,     82,     12,    -70,    -86,    -21,     64,     88,     29,    -57,    -90,    -37,     50,     90,     45,    -42,    -90,    -53,     34,     89,     60,    -26,    -87,    -66,     18,     85,     72,     -9,    -81,    -77 },
      {     78,      1,    -77,    -79,     -3,     76,     81,      6,    -75,    -81,     -8,     74,     82,     10,    -72,    -83,    -12,     71,     84,     14,    -70,    -85,    -16,     68,     86,     19,    -67,    -86,    -21,     65,     87,     23,    -64,    -88,    -25,     62,     88,     27,    -60,    -89,    -29,     59,     89,     31,    -57,    -89,    -33,     55,     90,     35,    -54,    -90,    -37,     52,     90,     39,    -50,    -90,    -41,     48,     90,     43,    -46,    -90,    -45,     44,     90,     47,    -42,    -90,    -49,     40,     90,     51,    -38,    -90,    -53,     36,     90,     55,    -34,    -89,    -56,     32,     89,     58,    -30,    -89,    -60,     28,     88,     61,    -26,    -88,    -63,     24,     87,     64,    -22,    -87,    -66,     20,     86,     67,    -18,    -85,    -69,     15,     85,     70,    -13,    -84,    -72,     11,     83,     73,     -9,    -82,    -74,      7,     81,     76,     -4,    -80,    -77,      2,     79,     78 },
      {     78,     -2,    -80,    -76,      7,     82,     73,    -11,    -84,    -70,     15,     85,     67,    -20,    -87,    -64,     24,     88,     61,    -28,    -89,    -58,     32,     89,     55,    -36,    -90,    -51,     40,     90,     47,    -44,    -90,    -43,     48,     90,     39,    -52,    -90,    -35,     55,     89,     31,    -59,    -89,    -27,     62,     88,     23,    -65,    -86,    -19,     68,     85,     14,    -71,    -83,    -10,     74,     81,      6,    -76,    -79,     -1,     78,     77,     -3,    -81,    -75,      8,     82,     72,    -12,    -84,    -70,     16,     86,     67,    -21,    -87,    -64,     25,     88,     60,    -29,    -89,    -57,     33,     90,     54,    -37,    -90,    -50,     41,     90,     46,    -45,    -90,    -42,     49,     90,     38,    -53,    -90,    -34,     56,     89,     30,    -60,    -88,    -26,     63,     87,     22,    -66,    -86,    -18,     69,     85,     13,    -72,    -83,     -9,     74,     81,      4,    -77,    -79 },
      {     77,     -6,    -82,    -71,     16,     86,     64,    -27,    -89,    -55,     37,     90,     46,    -47,    -90,    -36,     56,     89,     26,    -64,    -86,    -15,     72,     82,      4,    -78,    -77,      7,     83,     70,    -18,    -87,    -63,     28,     89,     55,    -38,    -90,    -45,     48,     90,     35,    -57,    -89,    -25,     65,     86,     14,    -72,    -81,     -3,     78,     76,     -8,    -83,    -70,     19,     87,     62,    -29,    -89,    -54,     39,     90,     44,    -49,    -90,    -34,     58,     88,     24,    -66,    -85,    -13,     73,     81,      2,    -79,    -76,      9,     84,     69,    -20,    -87,    -61,     30,     89,     53,    -40,    -90,    -43,     50,     90,     33,    -59,    -88,    -23,     67,     85,     12,    -74,    -81,     -1,     79,     75,    -10,    -84,    -68,     21,     88,     60,    -31,    -90,    -52,     41,     90,     42,    -51,    -90,    -32,     60,     88,     22,    -67,    -85,    -11,     74,     80 },
      {     77,     -9,    -85,    -66,     26,     89,     53,    -42,    -90,    -37,     57,     88,     21,    -70,    -82,     -3,     79,     74,    -14,    -86,    -62,     31,     90,     48,    -47,    -90,    -32,     61,     87,     15,    -73,    -80,      2,     82,     70,    -20,    -88,    -58,     36,     90,     43,    -52,    -89,    -27,     65,     85,     10,    -76,    -77,      8,     84,     67,    -25,    -89,    -54,     41,     90,     38,    -56,    -88,    -22,     69,     83,      4,    -79,    -74,     13,     86,     63,    -30,    -90,    -49,     46,     90,     33,    -60,    -87,    -16,     72,     81,     -1,    -81,    -71,     19,     88,     59,    -35,    -90,    -44,     51,     89,     28,    -64,    -85,    -11,     76,     78,     -7,    -84,    -67,     24,     89,     55,    -40,    -90,    -39,     55,     89,     23,    -68,    -83,     -6,     78,     75,    -12,    -86,    -64,     29,     90,     50,    -45,    -90,    -34,     60,     87,     18,    -72,    -81 },
      {     76,    -12,    -86,    -60,     35,     90,     40,    -56,    -88,    -18,     73,     79,     -7,    -85,    -64,     30,     90,     45,    -52,    -89,    -23,     70,     81,     -1,    -82,    -68,     25,     89,     50,    -47,    -90,    -28,     66,     84,      4,    -80,    -72,     20,     88,     55,    -42,    -90,    -33,     62,     86,     10,    -77,    -75,     14,     87,     59,    -37,    -90,    -38,     58,     87,     15,    -74,    -78,      9,     85,     63,    -32,    -90,    -43,     54,     89,     21,    -71,    -81,      3,     83,     67,    -27,    -90,    -48,     49,     89,     26,    -67,    -83,     -2,     81,     70,    -22,    -89,    -53,     44,     90,     31,    -64,    -85,     -8,     78,     74,    -16,    -88,    -57,     39,     90,     36,    -60,    -87,    -13,     76,     77,    -11,    -86,    -61,     34,     90,     41,    -55,    -88,    -19,     72,     79,     -6,    -84,    -65,     29,     90,     46,    -51,    -89,    -24,     69,     82 },
      {     76,    -15,    -88,    -55,     44,     90,     27,    -68,    -81,      3,     84,     64,    -33,    -90,    -38,     60,     86,      9,    -79,    -72,     22,     89,     49,    -50,    -89,    -21,     72,     78,    -10,    -86,    -59,     39,     90,     32,    -64,    -84,     -2,     82,     67,    -28,    -90,    -43,     55,     88,     14,    -76,    -75,     16,     88,     54,    -45,    -90,    -26,     69,     81,     -4,    -85,    -63,     34,     90,     37,    -60,    -86,     -8,     79,     71,    -23,    -89,    -48,     51,     89,     20,    -73,    -78,     11,     87,     58,    -40,    -90,    -31,     65,     83,      1,    -82,    -67,     29,     90,     42,    -56,    -87,    -13,     77,     74,    -18,    -88,    -53,     46,     90,     25,    -70,    -81,      6,     85,     62,    -35,    -90,    -36,     61,     85,      7,    -80,    -70,     24,     89,     47,    -52,    -89,    -19,     74,     77,    -12,    -87,    -57,     41,     90,     30,    -66,    -83 },
      {     75,    -19,    -89,    -48,     53,     88,     13,    -78,    -72,     24,     90,     43,    -57,    -86,     -8,     81,     68,    -29,    -90,    -38,     61,     85,      2,    -83,    -64,     34,     90,     33,    -65,    -82,      3,     85,     60,    -39,    -90,    -28,     69,     80,     -9,    -87,    -56,     44,     90,     23,    -72,    -77,     14,     88,     52,    -49,    -89,    -18,     76,     74,    -20,    -89,    -47,     54,     88,     12,    -78,    -71,     25,     90,     42,    -58,    -86,     -7,     81,     67,    -30,    -90,    -37,     62,     84,      1,    -83,    -64,     35,     90,     32,    -66,    -82,      4,     85,     60,    -40,    -90,    -27,     70,     79,    -10,    -87,    -55,     45,     89,     22,    -73,    -77,     15,     88,     51,    -50,    -89,    -16,     76,     74,    -21,    -89,    -46,     55,     87,     11,    -79,    -70,     26,     90,     41,    -59,    -86,     -6,     81,     67,    -31,    -90,    -36,     63,     84 },
      {     74,    -22,    -90,    -41,     60,     84,     -1,    -85,    -59,     43,     89,     20,    -76,    -73,     24,     90,     39,    -62,    -83,      3,     86,     57,    -45,    -89,    -18,     77,     72,    -26,    -90,    -37,     64,     82,     -6,    -86,    -55,     47,     89,     15,    -78,    -70,     28,     90,     35,    -65,    -81,      8,     87,     54,    -49,    -88,    -13,     79,     69,    -30,    -90,    -33,     67,     81,    -10,    -88,    -52,     51,     88,     11,    -80,    -67,     32,     90,     31,    -68,    -79,     12,     88,     50,    -53,    -87,     -9,     81,     66,    -34,    -90,    -29,     70,     78,    -14,    -89,    -48,     55,     87,      7,    -82,    -64,     36,     90,     27,    -71,    -77,     16,     89,     46,    -56,    -86,     -4,     83,     63,    -38,    -90,    -25,     72,     76,    -19,    -89,    -44,     58,     85,      2,    -84,    -61,     40,     90,     23,    -74,    -75,     21,     90,     42,    -60,    -85 },
      {     74,    -25,    -90,    -34,     67,     79,    -15,    -89,    -43,     60,     83,     -6,    -87,    -52,     53,     87,      4,    -84,    -60,     44,     89,     14,    -79,    -67,     35,     90,     24,    -74,    -73,     26,     90,     33,    -68,    -78,     16,     89,     42,    -61,    -83,      7,     87,     51,    -54,    -86,     -3,     84,     59,    -45,    -89,    -13,     80,     66,    -36,    -90,    -23,     75,     72,    -27,    -90,    -32,     69,     78,    -18,    -89,    -41,     62,     82,     -8,    -88,    -50,     55,     86,      2,    -85,    -58,     46,     89,     12,    -81,    -65,     37,     90,     22,    -76,    -72,     28,     90,     31,    -70,    -77,     19,     90,     40,    -63,    -82,      9,     88,     49,    -55,    -86,     -1,     85,     57,    -47,    -88,    -11,     81,     64,    -38,    -90,    -21,     76,     71,    -29,    -90,    -30,     70,     77,    -20,    -90,    -39,     64,     81,    -10,    -88,    -48,     56,     85 },
      {     73,    -28,    -90,    -27,     74,     72,    -29,    -90,    -26,     74,     72,    -30,    -90,    -25,     75,     71,    -31,    -90,    -24,     76,     70,    -32,    -90,    -23,     76,     70,    -33,    -90,    -22,     77,     69,    -34,    -90,    -21,     77,     68,    -35,    -90,    -20,     78,     67,    -36,    -90,    -19,     78,     67,    -37,    -90,    -18,     79,     66,    -38,    -90,    -16,     79,     65,    -39,    -89,    -15,     80,     64,    -40,    -89,    -14,     81,     64,    -41,    -89,    -13,     81,     63,    -42,    -89,    -12,     81,     62,    -43,    -89,    -11,     82,     61,    -44,    -89,    -10,     82,     60,    -45,    -88,     -9,     83,     60,    -46,    -88,     -8,     83,     59,    -47,    -88,     -7,     84,     58,    -48,    -88,     -6,     84,     57,    -49,    -87,     -4,     85,     56,    -50,    -87,     -3,     85,     55,    -51,    -87,     -2,     85,     55,    -52,    -86,     -1,     86,     54,    -53,    -86 },
      {     72,    -31,    -90,    -20,     79,     64,    -42,    -89,     -8,     84,     55,    -53,    -85,      4,     88,     45,    -62,    -81,     16,     90,     34,    -70,    -74,     28,     90,     23,    -77,    -67,     39,     89,     11,    -83,    -58,     50,     86,     -1,    -87,    -48,     60,     82,    -13,    -89,    -37,     68,     76,    -25,    -90,    -26,     76,     69,    -36,    -90,    -14,     81,     60,    -47,    -87,     -2,     86,     51,    -57,    -83,     10,     89,     40,    -66,    -78,     22,     90,     29,    -74,    -71,     33,     90,     18,    -80,    -63,     44,     88,      6,    -85,    -54,     55,     85,     -7,    -88,    -43,     64,     79,    -19,    -90,    -32,     72,     73,    -30,    -90,    -21,     78,     65,    -41,    -89,     -9,     84,     56,    -52,    -86,      3,     88,     46,    -61,    -81,     15,     90,     35,    -70,    -75,     27,     90,     24,    -77,    -67,     38,     89,     12,    -82,    -59,     49,     87 },
      {     72,    -34,    -90,    -12,     83,     55,    -55,    -84,     11,     89,     35,    -71,    -72,     33,     90,     13,    -83,    -56,     54,     84,    -10,    -89,    -36,     70,     73,    -32,    -90,    -14,     82,     57,    -53,    -85,      9,     89,     37,    -70,    -74,     31,     90,     15,    -82,    -58,     52,     85,     -8,    -89,    -38,     69,     74,    -30,    -90,    -16,     81,     59,    -51,    -85,      7,     89,     39,    -68,    -75,     29,     90,     18,    -81,    -60,     50,     86,     -6,    -89,    -40,     67,     76,    -28,    -90,    -19,     81,     60,    -49,    -86,      4,     88,     41,    -67,    -76,     27,     90,     20,    -80,    -61,     48,     86,     -3,    -88,    -42,     66,     77,    -26,    -90,    -21,     79,     62,    -47,    -87,      2,     88,     43,    -65,    -77,     25,     90,     22,    -79,    -63,     46,     87,     -1,    -88,    -44,     64,     78,    -24,    -90,    -23,     78,     64,    -45,    -87 },
      {     71,    -37,    -89,     -4,     87,     45,    -65,    -76,     29,     90,     13,    -84,    -53,     59,     81,    -21,    -90,    -22,     80,     60,    -52,    -84,     12,     90,     30,    -76,    -66,     44,     87,     -3,    -89,    -38,     70,     72,    -36,    -89,     -6,     86,     46,    -64,    -77,     28,     90,     14,    -83,    -54,     58,     81,    -20,    -90,    -23,     79,     60,    -51,    -85,     11,     90,     31,    -75,    -67,     43,     87,     -2,    -88,    -39,     70,     72,    -35,    -89,     -7,     86,     47,    -64,    -77,     27,     90,     15,    -83,    -55,     57,     81,    -19,    -90,    -24,     79,     61,    -50,    -85,     10,     90,     32,    -74,    -67,     42,     88,     -1,    -88,    -40,     69,     73,    -34,    -89,     -8,     86,     48,    -63,    -78,     26,     90,     16,    -82,    -55,     56,     82,    -18,    -90,    -25,     78,     62,    -49,    -85,      9,     89,     33,    -74,    -68,     41,     88 },
      {     70,    -40,    -88,      3,     89,     34,    -74,    -66,     46,     86,    -10,    -90,    -28,     78,     61,    -52,    -83,     16,     90,     22,    -81,    -56,     57,     81,    -23,    -90,    -15,     84,     51,    -62,    -77,     29,     90,      9,    -86,    -45,     67,     74,    -35,    -89,     -2,     88,     39,    -71,    -70,     41,     87,     -4,    -89,    -33,     75,     65,    -47,    -85,     11,     90,     27,    -78,    -60,     53,     83,    -18,    -90,    -21,     81,     55,    -58,    -80,     24,     90,     14,    -84,    -50,     63,     77,    -30,    -90,     -8,     86,     44,    -67,    -73,     36,     89,      1,    -88,    -38,     72,     69,    -42,    -87,      6,     89,     32,    -76,    -64,     48,     85,    -12,    -90,    -26,     79,     60,    -54,    -82,     19,     90,     20,    -82,    -55,     59,     79,    -25,    -90,    -13,     85,     49,    -64,    -76,     31,     89,      7,    -87,    -43,     68,     72,    -37,    -88 },
      {     70,    -43,    -86,     11,     90,     23,    -81,    -54,     61,     77,    -32,    -89,     -1,     89,     34,    -76,    -63,     52,     82,    -21,    -90,    -13,     85,     45,    -68,    -71,     41,     87,     -9,    -90,    -25,     81,     55,    -60,    -78,     30,     89,      3,    -88,    -36,     74,     64,    -50,    -83,     19,     90,     15,    -85,    -47,     67,     72,    -39,    -87,      7,     90,     27,    -79,    -57,     58,     79,    -28,    -90,     -6,     88,     38,    -73,    -66,     48,     84,    -16,    -90,    -18,     84,     49,    -65,    -74,     37,     88,     -4,    -89,    -29,     78,     59,    -56,    -80,     26,     90,      8,    -87,    -40,     72,     67,    -46,    -85,     14,     90,     20,    -83,    -51,     64,     75,    -35,    -88,      2,     89,     31,    -77,    -60,     55,     81,    -24,    -90,    -10,     86,     42,    -70,    -69,     44,     86,    -12,    -90,    -22,     82,     53,    -62,    -76,     33,     89 },
      {     69,    -46,    -84,     19,     90,     11,    -87,    -39,     74,     64,    -53,    -81,     26,     90,      3,    -89,    -32,     78,     58,    -59,    -77,     33,     88,     -4,    -90,    -25,     81,     52,    -64,    -73,     40,     86,    -12,    -90,    -18,     85,     45,    -70,    -68,     47,     84,    -20,    -90,    -10,     87,     38,    -74,    -63,     54,     81,    -27,    -89,     -2,     89,     31,    -78,    -57,     60,     77,    -34,    -88,      6,     90,     24,    -82,    -51,     65,     72,    -41,    -86,     13,     90,     16,    -85,    -44,     70,     67,    -48,    -83,     21,     90,      9,    -87,    -37,     75,     62,    -55,    -80,     28,     89,      1,    -89,    -30,     79,     56,    -60,    -76,     35,     88,     -7,    -90,    -23,     82,     50,    -66,    -72,     42,     86,    -14,    -90,    -15,     85,     43,    -71,    -67,     49,     83,    -22,    -90,     -8,     88,     36,    -76,    -61,     55,     79,    -29,    -89 },
      {     68,    -49,    -82,     26,     89,     -1,    -90,    -24,     83,     47,    -70,    -67,     51,     81,    -28,    -89,      3,     90,     22,    -84,    -45,     71,     65,    -53,    -80,     30,     89,     -6,    -90,    -20,     85,     43,    -72,    -64,     55,     79,    -32,    -88,      8,     90,     18,    -85,    -41,     74,     62,    -56,    -78,     34,     88,    -10,    -90,    -15,     86,     39,    -75,    -60,     58,     77,    -36,    -87,     12,     90,     13,    -87,    -37,     76,     59,    -60,    -76,     38,     86,    -14,    -90,    -11,     87,     35,    -77,    -57,     61,     74,    -40,    -86,     16,     90,      9,    -88,    -33,     78,     55,    -63,    -73,     42,     85,    -19,    -90,     -7,     88,     31,    -79,    -54,     64,     72,    -44,    -84,     21,     90,      4,    -89,    -29,     81,     52,    -66,    -70,     46,     83,    -23,    -90,     -2,     89,     27,    -81,    -50,     67,     69,    -48,    -82,     25,     89 },
      {     67,    -52,    -79,     33,     87,    -13,    -90,     -8,     89,     28,    -82,    -47,     71,     64,    -56,    -77,     38,     86,    -19,    -90,     -2,     89,     23,    -84,    -42,     74,     60,    -60,    -74,     43,     84,    -24,    -89,      3,     90,     18,    -86,    -37,     77,     55,    -64,    -70,     48,     81,    -29,    -88,      9,     90,     12,    -88,    -32,     80,     51,    -68,    -67,     53,     79,    -34,    -87,     14,     90,      7,    -89,    -27,     82,     46,    -72,    -63,     57,     76,    -39,    -85,     20,     90,      1,    -90,    -22,     85,     41,    -75,    -59,     61,     73,    -44,    -83,     25,     89,     -4,    -90,    -16,     86,     36,    -78,    -55,     65,     70,    -49,    -81,     30,     88,    -10,    -90,    -11,     88,     31,    -81,    -50,     69,     66,    -54,    -78,     35,     87,    -15,    -90,     -6,     89,     26,    -83,    -45,     72,     62,    -58,    -76,     40,     85,    -21,    -90 },
      {     67,    -55,    -77,     40,     84,    -25,    -89,      9,     90,      8,    -89,    -24,     85,     39,    -77,    -54,     67,     66,    -55,    -76,     41,     84,    -26,    -89,     10,     90,      7,    -89,    -23,     85,     38,    -78,    -53,     68,     65,    -56,    -76,     42,     83,    -27,    -88,     11,     90,      6,    -89,    -22,     85,     37,    -78,    -52,     69,     64,    -57,    -75,     43,     83,    -28,    -88,     12,     90,      4,    -89,    -21,     86,     36,    -79,    -51,     70,     64,    -58,    -74,     44,     82,    -29,    -88,     13,     90,      3,    -90,    -20,     86,     35,    -79,    -50,     70,     63,    -59,    -74,     45,     82,    -30,    -88,     14,     90,      2,    -90,    -19,     86,     34,    -80,    -49,     71,     62,    -60,    -73,     46,     81,    -31,    -87,     15,     90,      1,    -90,    -18,     87,     33,    -81,    -48,     72,     61,    -60,    -72,     47,     81,    -32,    -87,     16,     90 },
      {     66,    -57,    -74,     47,     80,    -36,    -85,     25,     88,    -13,    -90,      1,     90,     11,    -89,    -23,     86,     34,    -81,    -45,     75,     55,    -67,    -64,     59,     72,    -49,    -79,     38,     84,    -27,    -88,     15,     90,     -3,    -90,     -9,     89,     21,    -86,    -32,     82,     43,    -76,    -54,     69,     63,    -60,    -71,     51,     78,    -40,    -83,     29,     87,    -18,    -90,      6,     90,      7,    -89,    -19,     87,     30,    -83,    -41,     77,     52,    -70,    -61,     62,     70,    -53,    -77,     42,     82,    -31,    -87,     20,     89,     -8,    -90,     -4,     90,     16,    -88,    -28,     84,     39,    -78,    -50,     72,     60,    -64,    -68,     55,     76,    -44,    -81,     33,     86,    -22,    -89,     10,     90,      2,    -90,    -14,     88,     26,    -85,    -37,     79,     48,    -73,    -58,     65,     67,    -56,    -74,     46,     81,    -35,    -85,     24,     89,    -12,    -90 },
      {     65,    -60,    -70,     54,     75,    -47,    -79,     40,     82,    -33,    -85,     26,     88,    -19,    -89,     11,     90,     -3,    -90,     -4,     90,     12,    -89,    -20,     87,     27,    -85,    -34,     82,     41,    -78,    -48,     74,     55,    -70,    -60,     64,     66,    -59,    -71,     53,     76,    -46,    -79,     39,     83,    -32,    -86,     25,     88,    -18,    -89,     10,     90,     -2,    -90,     -6,     90,     13,    -89,    -21,     87,     28,    -85,    -35,     81,     42,    -78,    -49,     74,     55,    -69,    -61,     64,     67,    -58,    -72,     52,     76,    -45,    -80,     38,     83,    -31,    -86,     24,     88,    -16,    -89,      9,     90,     -1,    -90,     -7,     90,     14,    -89,    -22,     87,     29,    -84,    -36,     81,     43,    -77,    -50,     73,     56,    -68,    -62,     63,     67,    -57,    -72,     51,     77,    -44,    -81,     37,     84,    -30,    -86,     23,     88,    -15,    -90,      8,     90 },
      {     64,    -62,    -67,     60,     69,    -57,    -71,     55,     73,    -52,    -75,     49,     77,    -46,    -78,     43,     80,    -40,    -81,     37,     83,    -34,    -84,     31,     85,    -28,    -86,     25,     87,    -22,    -88,     19,     89,    -15,    -89,     12,     90,     -9,    -90,      6,     90,     -2,    -90,     -1,     90,      4,    -90,     -8,     90,     11,    -89,    -14,     89,     18,    -88,    -21,     88,     24,    -87,    -27,     86,     30,    -85,    -33,     83,     36,    -82,    -39,     81,     42,    -79,    -45,     77,     48,    -76,    -51,     74,     54,    -72,    -56,     70,     59,    -67,    -61,     65,     64,    -63,    -66,     60,     68,    -58,    -70,     55,     72,    -53,    -74,     50,     76,    -47,    -78,     44,     79,    -41,    -81,     38,     82,    -35,    -84,     32,     85,    -29,    -86,     26,     87,    -23,    -88,     20,     89,    -16,    -89,     13,     90,    -10,    -90,      7,     90,     -3,    -90 },
      {     64,    -64,    -63,     65,     62,    -66,    -61,     67,     60,    -67,    -60,     68,     59,    -69,    -58,     70,     57,    -70,    -56,     71,     55,    -72,    -55,     72,     54,    -73,    -53,     74,     52,    -74,    -51,     75,     50,    -76,    -49,     76,     48,    -77,    -47,     77,     46,    -78,    -45,     78,     44,    -79,    -43,     79,     42,    -80,    -41,     81,     40,    -81,    -39,     81,     38,    -82,    -37,     82,     36,    -83,    -35,     83,     34,    -84,    -33,     84,     32,    -85,    -31,     85,     30,    -85,    -29,     86,     28,    -86,    -27,     86,     26,    -87,    -25,     87,     24,    -87,    -23,     88,     22,    -88,    -21,     88,     20,    -88,    -19,     89,     18,    -89,    -16,     89,     15,    -89,    -14,     89,     13,    -89,    -12,     90,     11,    -90,    -10,     90,      9,    -90,     -8,     90,      7,    -90,     -6,     90,      4,    -90,     -3,     90,      2,    -90,     -1,     90 },
      {     63,    -67,    -59,     70,     55,    -74,    -50,     77,     45,    -79,    -40,     82,     35,    -84,    -30,     86,     25,    -88,    -20,     89,     14,    -90,     -9,     90,      3,    -90,      2,     90,     -8,    -90,     13,     89,    -19,    -88,     24,     86,    -29,    -85,     34,     82,    -39,    -80,     44,     77,    -49,    -74,     54,     71,    -58,    -67,     62,     64,    -66,    -60,     70,     55,    -73,    -51,     76,     46,    -79,    -41,     81,     36,    -84,    -31,     86,     26,    -87,    -21,     89,     15,    -89,    -10,     90,      4,    -90,      1,     90,     -7,    -90,     12,     89,    -18,    -88,     23,     87,    -28,    -85,     33,     83,    -38,    -81,     43,     78,    -48,    -75,     53,     72,    -57,    -68,     61,     64,    -65,    -60,     69,     56,    -72,    -52,     76,     47,    -78,    -42,     81,     37,    -83,    -32,     85,     27,    -87,    -22,     88,     16,    -89,    -11,     90,      6,    -90 },
      {     62,    -69,    -55,     75,     46,    -80,    -37,     84,     28,    -87,    -19,     89,      9,    -90,      1,     90,    -11,    -89,     21,     87,    -30,    -83,     39,     79,    -48,    -74,     56,     67,    -64,    -60,     70,     53,    -76,    -44,     81,     35,    -85,    -26,     88,     16,    -90,     -7,     90,     -3,    -90,     13,     89,    -23,    -86,     32,     82,    -41,    -78,     50,     72,    -58,    -66,     65,     59,    -72,    -51,     77,     42,    -82,    -33,     86,     24,    -88,    -14,     90,      4,    -90,      6,     90,    -15,    -88,     25,     85,    -34,    -81,     43,     77,    -52,    -71,     60,     64,    -67,    -57,     73,     49,    -78,    -40,     83,     31,    -86,    -22,     89,     12,    -90,     -2,     90,     -8,    -89,     18,     88,    -27,    -85,     36,     81,    -45,    -76,     54,     70,    -61,    -63,     68,     55,    -74,    -47,     79,     38,    -84,    -29,     87,     20,    -89,    -10,     90 },
      {     61,    -71,    -50,     79,     37,    -85,    -24,     89,     10,    -90,      4,     90,    -19,    -87,     32,     81,    -45,    -74,     57,     65,    -67,    -55,     76,     42,    -83,    -29,     88,     15,    -90,     -1,     90,    -13,    -88,     27,     84,    -40,    -77,     53,     69,    -64,    -59,     73,     47,    -81,    -34,     86,     21,    -89,     -7,     90,     -8,    -89,     22,     86,    -35,    -80,     48,     72,    -60,    -63,     70,     52,    -78,    -39,     84,     26,    -88,    -12,     90,     -2,    -90,     16,     87,    -30,    -82,     43,     76,    -55,    -67,     66,     56,    -75,    -44,     82,     31,    -87,    -18,     90,      3,    -90,     11,     89,    -25,    -85,     38,     78,    -51,    -70,     62,     60,    -72,    -49,     79,     36,    -85,    -23,     89,      9,    -90,      6,     89,    -20,    -86,     33,     81,    -46,    -74,     58,     64,    -68,    -54,     77,     41,    -83,    -28,     88,     14,    -90 },
      {     60,    -73,    -45,     82,     28,    -88,    -10,     90,     -9,    -89,     27,     83,    -44,    -74,     60,     61,    -72,    -46,     82,     29,    -88,    -11,     90,     -8,    -89,     26,     83,    -43,    -74,     59,     62,    -72,    -47,     81,     30,    -88,    -12,     90,     -7,    -89,     25,     84,    -42,    -75,     58,     63,    -71,    -48,     81,     31,    -88,    -13,     90,     -6,    -89,     24,     84,    -41,    -76,     57,     64,    -70,    -49,     81,     32,    -87,    -14,     90,     -4,    -89,     23,     85,    -40,    -76,     56,     64,    -70,    -50,     80,     33,    -87,    -15,     90,     -3,    -89,     22,     85,    -39,    -77,     55,     65,    -69,    -51,     79,     34,    -87,    -16,     90,     -2,    -90,     21,     85,    -38,    -77,     55,     66,    -68,    -52,     79,     35,    -86,    -18,     90,     -1,    -90,     20,     86,    -37,    -78,     54,     67,    -67,    -53,     78,     36,    -86,    -19,     90 },
      {     60,    -75,    -40,     85,     19,    -90,      4,     89,    -27,    -82,     48,     70,    -66,    -53,     79,     32,    -88,    -10,     90,    -13,    -87,     35,     78,    -55,    -64,     72,     45,    -83,    -24,     89,      1,    -90,     22,     84,    -43,    -73,     62,     57,    -77,    -37,     86,     15,    -90,      8,     88,    -30,    -81,     51,     67,    -68,    -50,     81,     29,    -89,     -7,     90,    -16,    -86,     38,     76,    -58,    -61,     74,     42,    -85,    -21,     90,     -2,    -89,     25,     83,    -46,    -71,     64,     55,    -78,    -34,     87,     12,    -90,     11,     88,    -33,    -79,     54,     65,    -70,    -47,     82,     26,    -89,     -3,     90,    -20,    -85,     41,     74,    -60,    -59,     76,     39,    -86,    -18,     90,     -6,    -89,     28,     81,    -49,    -69,     67,     52,    -80,    -31,     88,      9,    -90,     14,     87,    -36,    -77,     56,     63,    -72,    -44,     84,     23,    -90 },
      {     59,    -77,    -35,     88,      9,    -90,     19,     85,    -44,    -71,     66,     51,    -81,    -26,     89,     -1,    -89,     28,     81,    -53,    -64,     72,     42,    -85,    -16,     90,    -11,    -87,     37,     76,    -60,    -57,     78,     33,    -88,     -7,     90,    -21,    -84,     46,     70,    -67,    -49,     82,     24,    -90,      3,     89,    -30,    -79,     55,     63,    -74,    -40,     86,     14,    -90,     13,     86,    -39,    -74,     62,     55,    -79,    -31,     89,      4,    -90,     23,     83,    -48,    -68,     69,     47,    -83,    -22,     90,     -6,    -88,     32,     78,    -56,    -61,     75,     38,    -87,    -12,     90,    -15,    -86,     41,     73,    -64,    -54,     80,     29,    -89,     -2,     90,    -25,    -82,     50,     67,    -70,    -45,     84,     20,    -90,      8,     88,    -34,    -77,     58,     60,    -76,    -36,     87,     10,    -90,     18,     85,    -43,    -72,     65,     52,    -81,    -27,     89 },
      {     58,    -78,    -30,     89,     -1,    -89,     32,     77,    -60,    -56,     79,     28,    -89,      3,     88,    -34,    -76,     61,     55,    -81,    -26,     90,     -6,    -88,     36,     75,    -63,    -53,     81,     24,    -90,      8,     87,    -38,    -74,     64,     51,    -82,    -22,     90,    -10,    -87,     40,     72,    -66,    -49,     83,     20,    -90,     12,     86,    -42,    -71,     67,     47,    -84,    -18,     90,    -14,    -85,     44,     70,    -69,    -45,     85,     15,    -90,     16,     85,    -46,    -68,     70,     43,    -86,    -13,     90,    -19,    -84,     48,     67,    -72,    -41,     86,     11,    -90,     21,     83,    -50,    -65,     73,     39,    -87,     -9,     90,    -23,    -82,     52,     64,    -74,    -37,     88,      7,    -90,     25,     81,    -54,    -62,     76,     35,    -88,     -4,     90,    -27,    -80,     55,     60,    -77,    -33,     89,      2,    -89,     29,     79,    -57,    -59,     78,     31,    -89 },
      {     57,    -80,    -25,     90,    -11,    -86,     45,     67,    -72,    -38,     88,      3,    -89,     32,     76,    -63,    -51,     83,     18,    -90,     19,     83,    -52,    -62,     77,     31,    -89,      4,     88,    -39,    -72,     68,     44,    -86,    -10,     90,    -26,    -79,     58,     56,    -81,    -24,     90,    -12,    -85,     46,     67,    -73,    -37,     88,      2,    -89,     33,     76,    -64,    -50,     84,     16,    -90,     20,     82,    -53,    -61,     77,     30,    -89,      6,     87,    -40,    -71,     69,     43,    -86,     -9,     90,    -27,    -79,     59,     55,    -81,    -23,     90,    -13,    -85,     47,     66,    -74,    -36,     88,      1,    -89,     34,     75,    -64,    -49,     84,     15,    -90,     21,     82,    -54,    -60,     78,     29,    -90,      7,     87,    -41,    -70,     70,     42,    -87,     -8,     90,    -28,    -78,     60,     55,    -81,    -22,     90,    -14,    -85,     48,     65,    -74,    -35,     89 },
      {     56,    -81,    -20,     90,    -21,    -81,     57,     55,    -82,    -19,     90,    -22,    -81,     58,     55,    -82,    -18,     90,    -23,    -80,     59,     54,    -83,    -16,     90,    -24,    -79,     60,     53,    -83,    -15,     90,    -25,    -79,     60,     52,    -84,    -14,     90,    -26,    -78,     61,     51,    -84,    -13,     90,    -27,    -78,     62,     50,    -85,    -12,     90,    -28,    -77,     63,     49,    -85,    -11,     90,    -29,    -77,     64,     48,    -85,    -10,     90,    -30,    -76,     64,     47,    -86,     -9,     90,    -31,    -76,     65,     46,    -86,     -8,     89,    -32,    -75,     66,     45,    -86,     -7,     89,    -33,    -74,     67,     44,    -87,     -6,     89,    -34,    -74,     67,     43,    -87,     -4,     89,    -35,    -73,     68,     42,    -87,     -3,     89,    -36,    -72,     69,     41,    -88,     -2,     89,    -37,    -72,     70,     40,    -88,     -1,     88,    -38,    -71,     70,     39,    -88 },
      {     55,    -83,    -14,     90,    -30,    -75,     67,     41,    -88,      2,     87,    -45,    -64,     77,     26,    -90,     19,     81,    -59,    -52,     85,     10,    -89,     34,     72,    -70,    -37,     89,     -7,    -86,     49,     61,    -79,    -22,     90,    -23,    -79,     62,     48,    -86,     -6,     89,    -38,    -70,     73,     33,    -90,     11,     84,    -53,    -58,     81,     18,    -90,     27,     77,    -65,    -44,     87,      1,    -88,     42,     67,    -76,    -29,     90,    -15,    -82,     56,     55,    -83,    -13,     90,    -31,    -74,     68,     40,    -88,      3,     87,    -46,    -64,     78,     25,    -90,     20,     81,    -60,    -51,     85,      9,    -89,     35,     72,    -71,    -36,     89,     -8,    -85,     50,     60,    -80,    -21,     90,    -24,    -78,     63,     47,    -86,     -4,     89,    -39,    -69,     74,     32,    -90,     12,     84,    -54,    -57,     82,     16,    -90,     28,     76,    -66,    -43,     88 },
      {     55,    -84,     -9,     89,    -39,    -67,     76,     26,    -90,     23,     78,    -65,    -42,     88,     -6,    -85,     52,     57,    -83,    -12,     89,    -36,    -70,     74,     29,    -90,     20,     79,    -63,    -45,     88,     -2,    -86,     49,     60,    -81,    -15,     90,    -33,    -72,     72,     32,    -90,     16,     81,    -60,    -48,     87,      1,    -87,     46,     62,    -80,    -19,     90,    -30,    -74,     70,     35,    -90,     13,     82,    -58,    -51,     86,      4,    -88,     43,     64,    -78,    -22,     90,    -27,    -76,     68,     38,    -89,     10,     84,    -55,    -54,     85,      8,    -89,     40,     67,    -77,    -25,     90,    -24,    -77,     66,     41,    -89,      7,     85,    -53,    -56,     83,     11,    -89,     37,     69,    -75,    -28,     90,    -21,    -79,     64,     44,    -88,      3,     86,    -50,    -59,     82,     14,    -90,     34,     71,    -73,    -31,     90,    -18,    -81,     61,     47,    -87 },
      {     54,    -85,     -3,     87,    -48,    -59,     83,     10,    -89,     42,     64,    -80,    -16,     90,    -36,    -68,     77,     23,    -90,     30,     72,    -73,    -29,     90,    -24,    -76,     69,     35,    -90,     18,     79,    -64,    -41,     89,    -11,    -82,     60,     47,    -88,      4,     85,    -55,    -53,     86,      2,    -87,     49,     58,    -83,     -9,     89,    -43,    -63,     81,     15,    -90,     37,     67,    -77,    -22,     90,    -31,    -72,     74,     28,    -90,     25,     76,    -70,    -34,     90,    -19,    -79,     65,     40,    -89,     12,     82,    -60,    -46,     88,     -6,    -85,     55,     52,    -86,     -1,     87,    -50,    -57,     84,      8,    -88,     44,     62,    -81,    -14,     89,    -38,    -67,     78,     21,    -90,     32,     71,    -74,    -27,     90,    -26,    -75,     70,     33,    -90,     20,     78,    -66,    -39,     89,    -13,    -81,     61,     45,    -88,      7,     84,    -56,    -51,     86 },
      {     53,    -86,      2,     85,    -56,    -49,     88,     -7,    -83,     60,     45,    -89,     11,     81,    -63,    -41,     89,    -15,    -79,     66,     37,    -90,     20,     77,    -69,    -33,     90,    -24,    -75,     72,     29,    -90,     28,     72,    -74,    -25,     90,    -32,    -70,     77,     21,    -90,     36,     67,    -79,    -16,     89,    -40,    -64,     81,     12,    -89,     44,     60,    -83,     -8,     88,    -48,    -57,     85,      3,    -87,     52,     54,    -86,      1,     85,    -55,    -50,     87,     -6,    -84,     59,     46,    -88,     10,     82,    -62,    -42,     89,    -14,    -80,     65,     38,    -90,     19,     78,    -68,    -34,     90,    -23,    -76,     71,     30,    -90,     27,     73,    -74,    -26,     90,    -31,    -70,     76,     22,    -90,     35,     67,    -78,    -18,     90,    -39,    -64,     81,     13,    -89,     43,     61,    -82,     -9,     88,    -47,    -58,     84,      4,    -87,     51,     55,    -86 },
      {     52,    -87,      8,     82,    -64,    -38,     90,    -23,    -74,     74,     24,    -90,     37,     64,    -81,     -9,     88,    -51,    -53,     87,     -7,    -82,     63,     39,    -90,     22,     75,    -73,    -25,     90,    -36,    -65,     81,     10,    -88,     50,     54,    -87,      6,     83,    -62,    -40,     90,    -21,    -76,     72,     26,    -90,     35,     66,    -81,    -11,     88,    -49,    -55,     86,     -4,    -83,     61,     41,    -90,     20,     76,    -72,    -27,     90,    -34,    -67,     80,     12,    -88,     48,     55,    -86,      3,     84,    -60,    -42,     89,    -19,    -77,     71,     28,    -90,     33,     67,    -79,    -13,     89,    -47,    -56,     86,     -2,    -84,     60,     43,    -89,     18,     77,    -70,    -29,     90,    -32,    -68,     79,     14,    -89,     46,     57,    -85,      1,     85,    -59,    -44,     89,    -16,    -78,     70,     30,    -90,     31,     69,    -78,    -15,     89,    -45,    -58,     85 },
      {     51,    -88,     13,     78,    -70,    -27,     90,    -38,    -62,     84,      1,    -85,     60,     40,    -90,     25,     72,    -77,    -15,     89,    -49,    -53,     88,    -11,    -79,     69,     29,    -90,     36,     64,    -83,     -3,     85,    -59,    -42,     90,    -23,    -73,     76,     18,    -89,     47,     55,    -87,      9,     81,    -67,    -31,     90,    -34,    -65,     82,      6,    -86,     57,     44,    -89,     21,     74,    -75,    -20,     89,    -45,    -56,     86,     -7,    -81,     66,     33,    -90,     32,     67,    -81,     -8,     87,    -55,    -46,     89,    -19,    -76,     74,     22,    -90,     43,     58,    -86,      4,     82,    -64,    -35,     90,    -30,    -68,     80,     10,    -87,     54,     48,    -89,     16,     77,    -72,    -24,     90,    -41,    -60,     85,     -2,    -83,     63,     37,    -90,     28,     70,    -79,    -12,     88,    -52,    -50,     88,    -14,    -78,     71,     26,    -90,     39,     61,    -84 },
      {     50,    -89,     19,     74,    -76,    -15,     88,    -53,    -47,     89,    -22,    -72,     78,     12,    -87,     55,     44,    -90,     25,     70,    -79,     -9,     86,    -58,    -41,     90,    -28,    -68,     81,      6,    -85,     60,     38,    -90,     31,     66,    -82,     -2,     84,    -63,    -35,     90,    -34,    -64,     84,     -1,    -83,     65,     32,    -90,     37,     61,    -85,      4,     81,    -67,    -29,     90,    -40,    -59,     86,     -8,    -80,     70,     26,    -90,     43,     56,    -87,     11,     78,    -72,    -23,     89,    -46,    -54,     88,    -14,    -77,     74,     20,    -89,     49,     51,    -89,     18,     75,    -76,    -16,     88,    -52,    -48,     89,    -21,    -73,     77,     13,    -88,     55,     45,    -90,     24,     71,    -79,    -10,     87,    -57,    -42,     90,    -27,    -69,     81,      7,    -86,     60,     39,    -90,     30,     67,    -82,     -3,     85,    -62,    -36,     90,    -33,    -64,     83 },
      {     49,    -89,     24,     70,    -81,     -3,     84,    -65,    -30,     90,    -43,    -55,     88,    -18,    -74,     78,     10,    -86,     60,     36,    -90,     37,     60,    -86,     11,     77,    -74,    -16,     88,    -55,    -42,     90,    -31,    -64,     84,     -4,    -81,     70,     23,    -89,     50,     48,    -89,     25,     69,    -81,     -2,     83,    -66,    -29,     90,    -44,    -54,     88,    -19,    -73,     78,      9,    -86,     61,     35,    -90,     38,     59,    -87,     12,     77,    -75,    -15,     88,    -56,    -41,     90,    -32,    -64,     85,     -6,    -80,     71,     22,    -89,     51,     47,    -90,     26,     68,    -82,     -1,     83,    -67,    -28,     90,    -45,    -53,     89,    -20,    -72,     79,      8,    -85,     62,     34,    -90,     39,     58,    -87,     13,     76,    -76,    -14,     87,    -57,    -40,     90,    -33,    -63,     85,     -7,    -79,     72,     21,    -89,     52,     46,    -90,     27,     67,    -82 },
      {     48,    -90,     29,     64,    -85,      9,     77,    -76,    -12,     86,    -62,    -32,     90,    -45,    -51,     89,    -26,    -67,     84,     -6,    -79,     74,     15,    -87,     60,     35,    -90,     42,     54,    -89,     23,     69,    -82,      2,     81,    -72,    -19,     88,    -57,    -38,     90,    -39,    -56,     88,    -20,    -71,     81,      1,    -82,     70,     22,    -89,     55,     41,    -90,     36,     59,    -87,     16,     73,    -79,     -4,     83,    -67,    -25,     89,    -52,    -44,     90,    -33,    -61,     86,    -13,    -75,     78,      8,    -85,     65,     28,    -90,     49,     47,    -90,     30,     64,    -85,     10,     77,    -76,    -11,     86,    -63,    -31,     90,    -46,    -50,     89,    -27,    -66,     84,     -7,    -78,     74,     14,    -87,     60,     34,    -90,     43,     53,    -89,     24,     68,    -83,      3,     80,    -72,    -18,     88,    -58,    -37,     90,    -40,    -55,     88,    -21,    -70,     81 },
      {     47,    -90,     34,     59,    -88,     21,     69,    -83,      7,     77,    -77,     -8,     84,    -68,    -22,     88,    -58,    -35,     90,    -46,    -48,     90,    -33,    -60,     88,    -20,    -70,     83,     -6,    -78,     76,      9,    -84,     67,     23,    -88,     57,     36,    -90,     45,     49,    -90,     32,     60,    -87,     19,     70,    -82,      4,     78,    -76,    -10,     85,    -67,    -24,     89,    -56,    -37,     90,    -44,    -50,     90,    -31,    -61,     87,    -18,    -71,     82,     -3,    -79,     75,     11,    -85,     66,     25,    -89,     55,     38,    -90,     43,     51,    -90,     30,     62,    -87,     16,     72,    -81,      2,     79,    -74,    -12,     85,    -65,    -26,     89,    -55,    -39,     90,    -42,    -52,     89,    -29,    -63,     86,    -15,    -72,     81,     -1,    -80,     74,     13,    -86,     64,     27,    -89,     54,     40,    -90,     41,     53,    -89,     28,     64,    -86,     14,     73,    -81 },
      {     46,    -90,     39,     53,    -90,     32,     59,    -88,     25,     64,    -86,     18,     70,    -84,     10,     74,    -81,      2,     78,    -77,     -6,     82,    -72,    -13,     85,    -67,    -21,     87,    -62,    -28,     89,    -56,    -35,     90,    -50,    -42,     90,    -43,    -49,     90,    -36,    -55,     89,    -29,    -61,     88,    -22,    -67,     85,    -14,    -72,     82,     -7,    -76,     79,      1,    -80,     75,      9,    -83,     70,     16,    -86,     65,     24,    -88,     60,     31,    -89,     54,     38,    -90,     47,     45,    -90,     40,     52,    -90,     33,     58,    -89,     26,     64,    -87,     19,     69,    -84,     11,     74,    -81,      3,     78,    -77,     -4,     81,    -73,    -12,     85,    -68,    -20,     87,    -63,    -27,     89,    -57,    -34,     90,    -51,    -41,     90,    -44,    -48,     90,    -37,    -55,     89,    -30,    -60,     88,    -23,    -66,     86,    -15,    -71,     83,     -8,    -76,     79 },
      {     45,    -90,     44,     46,    -90,     43,     47,    -90,     42,     48,    -90,     41,     49,    -90,     40,     50,    -90,     39,     51,    -90,     38,     52,    -90,     37,     53,    -90,     36,     54,    -90,     35,     55,    -90,     34,     55,    -89,     33,     56,    -89,     32,     57,    -89,     31,     58,    -89,     30,     59,    -89,     29,     60,    -89,     28,     60,    -88,     27,     61,    -88,     26,     62,    -88,     25,     63,    -88,     24,     64,    -87,     23,     64,    -87,     22,     65,    -87,     21,     66,    -86,     20,     67,    -86,     19,     67,    -86,     18,     68,    -85,     16,     69,    -85,     15,     70,    -85,     14,     70,    -84,     13,     71,    -84,     12,     72,    -83,     11,     72,    -83,     10,     73,    -82,      9,     74,    -82,      8,     74,    -81,      7,     75,    -81,      6,     76,    -81,      4,     76,    -80,      3,     77,    -79,      2,     77,    -79,      1,     78,    -78 },
      {     44,    -90,     49,     39,    -90,     54,     34,    -89,     58,     29,    -88,     62,     24,    -87,     66,     19,    -85,     70,     13,    -83,     73,      8,    -81,     76,      2,    -78,     79,     -3,    -76,     81,     -9,    -72,     84,    -14,    -69,     86,    -20,    -65,     87,    -25,    -61,     89,    -30,    -57,     89,    -35,    -53,     90,    -40,    -48,     90,    -45,    -43,     90,    -50,    -38,     90,    -55,    -33,     89,    -59,    -28,     88,    -63,    -23,     87,    -67,    -18,     85,    -70,    -12,     83,    -74,     -7,     81,    -77,     -1,     78,    -79,      4,     75,    -82,     10,     72,    -84,     15,     68,    -86,     21,     64,    -88,     26,     60,    -89,     31,     56,    -90,     36,     52,    -90,     41,     47,    -90,     46,     42,    -90,     51,     37,    -90,     55,     32,    -89,     60,     27,    -88,     64,     22,    -86,     67,     16,    -85,     71,     11,    -82,     74,      6,    -80,     77 },
      {     43,    -90,     54,     32,    -89,     63,     21,    -85,     71,      9,    -81,     78,     -3,    -74,     83,    -15,    -67,     87,    -27,    -58,     90,    -38,    -48,     90,    -49,    -37,     89,    -59,    -26,     87,    -67,    -14,     83,    -75,     -2,     77,    -81,     10,     70,    -86,     22,     62,    -89,     33,     53,    -90,     44,     42,    -90,     55,     31,    -88,     64,     20,    -85,     72,      8,    -80,     78,     -4,    -74,     84,    -16,    -66,     88,    -28,    -57,     90,    -39,    -47,     90,    -50,    -36,     89,    -60,    -25,     87,    -68,    -13,     82,    -76,     -1,     77,    -81,     11,     70,    -86,     23,     61,    -89,     34,     52,    -90,     45,     41,    -90,     55,     30,    -88,     64,     19,    -85,     72,      7,    -79,     79,     -6,    -73,     84,    -18,    -65,     88,    -29,    -56,     90,    -40,    -46,     90,    -51,    -35,     89,    -60,    -24,     86,    -69,    -12,     82,    -76 },
      {     42,    -90,     58,     25,    -86,     71,      7,    -78,     81,    -12,    -67,     88,    -30,    -54,     90,    -47,    -37,     89,    -62,    -20,     84,    -74,     -1,     76,    -83,     18,     64,    -89,     35,     49,    -90,     52,     32,    -88,     66,     14,    -82,     77,     -4,    -72,     85,    -23,    -60,     90,    -40,    -44,     90,    -56,    -27,     87,    -70,     -9,     79,    -80,     10,     69,    -87,     28,     55,    -90,     45,     39,    -89,     60,     22,    -85,     73,      3,    -77,     82,    -15,    -65,     88,    -33,    -51,     90,    -50,    -34,     89,    -64,    -16,     83,    -76,      2,     74,    -85,     21,     61,    -89,     38,     46,    -90,     55,     29,    -87,     68,     11,    -81,     79,     -8,    -70,     86,    -26,    -57,     90,    -43,    -41,     90,    -59,    -24,     86,    -72,     -6,     78,    -81,     13,     67,    -88,     31,     53,    -90,     48,     36,    -89,     63,     19,    -84,     75 },
      {     41,    -89,     62,     18,    -82,     78,     -8,    -69,     88,    -32,    -50,     90,    -55,    -27,     86,    -72,     -2,     75,    -85,     23,     58,    -90,     46,     36,    -89,     66,     12,    -80,     81,    -13,    -65,     89,    -37,    -45,     90,    -59,    -22,     84,    -76,      3,     72,    -86,     28,     54,    -90,     51,     31,    -87,     70,      7,    -77,     83,    -19,    -61,     90,    -42,    -40,     89,    -63,    -16,     82,    -78,      9,     68,    -88,     33,     49,    -90,     55,     26,    -86,     73,      1,    -74,     85,    -24,    -57,     90,    -47,    -35,     88,    -67,    -11,     79,    -81,     14,     64,    -89,     38,     44,    -90,     60,     21,    -84,     76,     -4,    -71,     87,    -29,    -53,     90,    -52,    -30,     87,    -70,     -6,     77,    -83,     20,     60,    -90,     43,     39,    -89,     64,     15,    -81,     79,    -10,    -67,     88,    -34,    -48,     90,    -56,    -25,     85,    -74 },
      {     40,    -89,     66,     10,    -78,     83,    -22,    -57,     90,    -51,    -29,     86,    -74,      2,     71,    -87,     33,     47,    -90,     60,     18,    -81,     80,    -14,    -63,     90,    -44,    -36,     88,    -69,     -6,     76,    -85,     26,     54,    -90,     55,     25,    -85,     76,     -7,    -68,     88,    -37,    -43,     89,    -64,    -13,     79,    -82,     19,     60,    -90,     48,     32,    -87,     72,      1,    -73,     86,    -30,    -50,     90,    -58,    -21,     83,    -78,     11,     65,    -89,     41,     39,    -89,     67,      9,    -77,     84,    -23,    -56,     90,    -52,    -28,     86,    -74,      3,     70,    -88,     34,     46,    -90,     61,     16,    -81,     81,    -15,    -62,     90,    -45,    -35,     88,    -70,     -4,     75,    -85,     27,     53,    -90,     55,     24,    -84,     77,     -8,    -67,     89,    -38,    -42,     89,    -64,    -12,     79,    -82,     20,     59,    -90,     49,     31,    -87,     72 },
      {     39,    -88,     70,      2,    -72,     87,    -35,    -43,     89,    -67,     -7,     75,    -86,     31,     47,    -90,     64,     11,    -77,     85,    -27,    -51,     90,    -60,    -15,     79,    -83,     23,     55,    -90,     57,     20,    -81,     81,    -19,    -58,     90,    -54,    -24,     83,    -79,     14,     61,    -90,     50,     28,    -85,     77,    -10,    -64,     90,    -46,    -32,     86,    -74,      6,     67,    -89,     42,     36,    -88,     72,     -1,    -70,     88,    -38,    -40,     89,    -69,     -3,     73,    -87,     34,     44,    -89,     66,      8,    -76,     86,    -30,    -48,     90,    -63,    -12,     78,    -84,     26,     52,    -90,     60,     16,    -80,     82,    -22,    -55,     90,    -56,    -21,     82,    -81,     18,     59,    -90,     53,     25,    -84,     78,    -13,    -62,     90,    -49,    -29,     85,    -76,      9,     65,    -89,     45,     33,    -87,     74,     -4,    -68,     89,    -41,    -37,     88,    -71 },
      {     38,    -88,     73,     -6,    -66,     90,    -48,    -28,     84,    -79,     16,     58,    -90,     57,     18,    -79,     84,    -27,    -49,     90,    -65,     -7,     74,    -87,     37,     39,    -88,     72,     -4,    -67,     89,    -47,    -29,     85,    -78,     15,     59,    -90,     56,     19,    -80,     83,    -26,    -50,     90,    -64,     -8,     74,    -87,     36,     40,    -88,     72,     -3,    -67,     89,    -46,    -30,     85,    -78,     14,     60,    -90,     55,     20,    -81,     83,    -25,    -51,     90,    -64,     -9,     75,    -87,     35,     41,    -88,     71,     -2,    -68,     89,    -45,    -31,     85,    -77,     13,     60,    -90,     55,     21,    -81,     82,    -24,    -52,     90,    -63,    -10,     76,    -86,     34,     42,    -89,     70,     -1,    -69,     89,    -44,    -32,     86,    -77,     12,     61,    -90,     54,     22,    -81,     82,    -23,    -53,     90,    -62,    -11,     76,    -86,     33,     43,    -89,     70 },
      {     37,    -87,     76,    -13,    -59,     90,    -60,    -12,     76,    -87,     38,     36,    -86,     77,    -14,    -58,     90,    -60,    -11,     75,    -87,     39,     35,    -86,     77,    -15,    -57,     90,    -61,    -10,     74,    -88,     40,     34,    -86,     78,    -16,    -56,     90,    -62,     -9,     74,    -88,     41,     33,    -85,     78,    -18,    -55,     90,    -63,     -8,     73,    -88,     42,     32,    -85,     79,    -19,    -55,     90,    -64,     -7,     72,    -88,     43,     31,    -85,     79,    -20,    -54,     90,    -64,     -6,     72,    -89,     44,     30,    -84,     80,    -21,    -53,     90,    -65,     -4,     71,    -89,     45,     29,    -84,     81,    -22,    -52,     90,    -66,     -3,     70,    -89,     46,     28,    -83,     81,    -23,    -51,     90,    -67,     -2,     70,    -89,     47,     27,    -83,     81,    -24,    -50,     90,    -67,     -1,     69,    -89,     48,     26,    -82,     82,    -25,    -49,     89,    -68 },
      {     36,    -86,     79,    -21,    -51,     89,    -70,      4,     64,    -90,     58,     12,    -74,     88,    -44,    -28,     82,    -83,     29,     43,    -88,     75,    -13,    -57,     90,    -64,     -3,     69,    -90,     52,     20,    -78,     86,    -37,    -35,     85,    -79,     22,     50,    -89,     70,     -6,    -63,     90,    -59,    -11,     74,    -88,     45,     27,    -82,     83,    -30,    -42,     88,    -76,     14,     56,    -90,     65,      2,    -68,     90,    -53,    -19,     78,    -86,     38,     34,    -85,     80,    -23,    -49,     89,    -71,      7,     62,    -90,     60,     10,    -73,     89,    -46,    -26,     81,    -84,     31,     41,    -87,     76,    -15,    -55,     90,    -66,     -1,     67,    -90,     54,     18,    -77,     87,    -39,    -33,     85,    -81,     24,     48,    -89,     72,     -8,    -61,     90,    -60,     -9,     72,    -89,     47,     25,    -81,     84,    -32,    -40,     87,    -77,     16,     55,    -90,     67 },
      {     35,    -85,     81,    -28,    -42,     87,    -78,     21,     49,    -89,     74,    -13,    -55,     90,    -69,      6,     61,    -90,     64,      2,    -67,     90,    -58,    -10,     72,    -89,     52,     18,    -76,     88,    -45,    -25,     80,    -86,     38,     32,    -83,     83,    -31,    -39,     86,    -79,     24,     46,    -88,     76,    -16,    -53,     89,    -71,      9,     59,    -90,     66,     -1,    -64,     90,    -60,     -7,     70,    -90,     55,     14,    -74,     89,    -48,    -22,     78,    -87,     41,     29,    -82,     84,    -34,    -36,     85,    -81,     27,     43,    -87,     77,    -20,    -50,     89,    -73,     12,     56,    -90,     68,     -4,    -62,     90,    -63,     -3,     67,    -90,     57,     11,    -72,     89,    -51,    -19,     77,    -88,     44,     26,    -81,     85,    -37,    -33,     84,    -82,     30,     40,    -86,     79,    -23,    -47,     88,    -75,     15,     54,    -90,     70,     -8,    -60,     90,    -65 },
      {     34,    -83,     84,    -35,    -33,     83,    -84,     36,     32,    -82,     85,    -37,    -31,     82,    -85,     38,     30,    -81,     85,    -39,    -29,     81,    -86,     40,     28,    -81,     86,    -41,    -27,     80,    -86,     42,     26,    -79,     87,    -43,    -25,     79,    -87,     44,     24,    -78,     87,    -45,    -23,     78,    -88,     46,     22,    -77,     88,    -47,    -21,     77,    -88,     48,     20,    -76,     88,    -49,    -19,     76,    -89,     50,     18,    -75,     89,    -51,    -16,     74,    -89,     52,     15,    -74,     89,    -53,    -14,     73,    -89,     54,     13,    -72,     89,    -55,    -12,     72,    -90,     55,     11,    -71,     90,    -56,    -10,     70,    -90,     57,      9,    -70,     90,    -58,     -8,     69,    -90,     59,      7,    -68,     90,    -60,     -6,     67,    -90,     60,      4,    -67,     90,    -61,     -3,     66,    -90,     62,      2,    -65,     90,    -63,     -1,     64,    -90,     64 },
      {     33,    -82,     86,    -42,    -24,     77,    -88,     51,     14,    -72,     90,    -59,     -4,     65,    -90,     66,     -6,    -58,     90,    -72,     15,     50,    -88,     78,    -25,    -41,     85,    -82,     34,     32,    -81,     86,    -43,    -23,     77,    -89,     52,     13,    -71,     90,    -60,     -3,     64,    -90,     67,     -7,    -57,     90,    -73,     16,     49,    -88,     78,    -26,    -40,     85,    -83,     35,     31,    -81,     86,    -44,    -22,     76,    -89,     53,     12,    -70,     90,    -60,     -2,     64,    -90,     67,     -8,    -56,     89,    -74,     18,     48,    -88,     79,    -27,    -39,     85,    -83,     36,     30,    -81,     87,    -45,    -21,     76,    -89,     54,     11,    -70,     90,    -61,     -1,     63,    -90,     68,     -9,    -55,     89,    -74,     19,     47,    -87,     79,    -28,    -38,     84,    -84,     37,     29,    -80,     87,    -46,    -20,     75,    -89,     55,     10,    -69,     90,    -62 },
      {     32,    -81,     87,    -49,    -14,     70,    -90,     64,     -4,    -57,     89,    -76,     23,     41,    -85,     84,    -40,    -24,     76,    -89,     56,      6,    -64,     90,    -70,     13,     50,    -88,     80,    -31,    -33,     81,    -87,     48,     15,    -71,     90,    -63,      3,     58,    -89,     75,    -22,    -42,     85,    -84,     39,     25,    -77,     89,    -55,     -7,     65,    -90,     69,    -12,    -51,     88,    -79,     30,     34,    -81,     87,    -47,    -16,     72,    -90,     62,     -2,    -59,     90,    -74,     21,     43,    -85,     83,    -38,    -26,     77,    -89,     55,      8,    -66,     90,    -68,     11,     52,    -88,     79,    -29,    -35,     82,    -86,     46,     18,    -72,     90,    -61,      1,     60,    -90,     74,    -20,    -44,     86,    -83,     37,     27,    -78,     89,    -54,     -9,     67,    -90,     67,    -10,    -53,     88,    -78,     28,     36,    -82,     86,    -45,    -19,     73,    -90,     60 },
      {     31,    -79,     89,    -55,     -4,     62,    -90,     74,    -23,    -39,     83,    -86,     48,     13,    -68,     90,    -69,     14,     47,    -86,     83,    -40,    -22,     74,    -90,     63,     -6,    -55,     88,    -79,     32,     30,    -78,     89,    -56,     -3,     61,    -90,     75,    -24,    -38,     82,    -87,     49,     12,    -67,     90,    -70,     15,     46,    -86,     84,    -41,    -21,     73,    -90,     64,     -7,    -54,     88,    -80,     33,     29,    -78,     89,    -57,     -2,     60,    -90,     76,    -25,    -37,     82,    -87,     50,     11,    -67,     90,    -70,     16,     45,    -85,     84,    -42,    -20,     72,    -90,     64,     -8,    -53,     88,    -81,     34,     28,    -77,     89,    -58,     -1,     60,    -89,     76,    -26,    -36,     81,    -87,     51,     10,    -66,     90,    -71,     18,     44,    -85,     85,    -43,    -19,     72,    -90,     65,     -9,    -52,     88,    -81,     35,     27,    -77,     89,    -59 },
      {     30,    -77,     89,    -61,      6,     53,    -87,     82,    -40,    -20,     71,    -90,     69,    -16,    -43,     84,    -86,     50,      9,    -64,     90,    -76,     27,     33,    -79,     89,    -59,      2,     55,    -88,     81,    -37,    -23,     73,    -90,     67,    -13,    -46,     85,    -85,     47,     12,    -66,     90,    -74,     24,     36,    -81,     88,    -56,     -1,     58,    -89,     79,    -34,    -26,     75,    -90,     64,    -10,    -49,     86,    -84,     44,     15,    -68,     90,    -72,     21,     39,    -82,     88,    -54,     -4,     60,    -89,     78,    -31,    -29,     77,    -90,     62,     -7,    -52,     87,    -83,     41,     19,    -70,     90,    -70,     18,     42,    -83,     87,    -51,     -8,     63,    -90,     76,    -28,    -32,     78,    -89,     60,     -3,    -55,     88,    -81,     38,     22,    -72,     90,    -67,     14,     45,    -85,     86,    -48,    -11,     65,    -90,     74,    -25,    -35,     80,    -89,     57 },
      {     29,    -76,     90,    -67,     15,     42,    -82,     88,    -56,      1,     55,    -87,     83,    -44,    -13,     65,    -90,     77,    -31,    -27,     74,    -90,     68,    -18,    -40,     81,    -88,     58,     -3,    -53,     87,    -84,     46,     11,    -64,     90,    -78,     33,     25,    -73,     90,    -70,     20,     38,    -81,     89,    -60,      6,     51,    -86,     85,    -48,     -9,     62,    -89,     79,    -35,    -23,     72,    -90,     71,    -22,    -36,     79,    -89,     61,     -8,    -49,     85,    -86,     50,      7,    -60,     89,    -80,     37,     21,    -70,     90,    -72,     24,     34,    -78,     89,    -63,     10,     47,    -85,     86,    -52,     -4,     59,    -89,     81,    -39,    -19,     69,    -90,     74,    -26,    -32,     77,    -90,     64,    -12,    -45,     84,    -87,     54,      2,    -57,     88,    -82,     41,     16,    -67,     90,    -75,     28,     30,    -76,     90,    -66,     14,     43,    -83,     88,    -55 },
      {     28,    -74,     90,    -72,     25,     31,    -76,     90,    -70,     22,     34,    -77,     90,    -67,     19,     37,    -79,     90,    -65,     15,     40,    -81,     89,    -63,     12,     43,    -82,     89,    -60,      9,     46,    -83,     88,    -58,      6,     49,    -85,     87,    -55,      2,     52,    -86,     86,    -53,     -1,     55,    -87,     85,    -50,     -4,     57,    -88,     84,    -47,     -8,     60,    -88,     82,    -44,    -11,     62,    -89,     81,    -41,    -14,     64,    -89,     79,    -38,    -18,     67,    -90,     78,    -35,    -21,     69,    -90,     76,    -32,    -24,     71,    -90,     74,    -29,    -27,     73,    -90,     72,    -26,    -30,     75,    -90,     70,    -23,    -33,     77,    -90,     68,    -20,    -36,     78,    -90,     66,    -16,    -39,     80,    -89,     64,    -13,    -42,     81,    -89,     61,    -10,    -45,     83,    -88,     59,     -7,    -48,     84,    -87,     56,     -3,    -51,     85,    -86,     54 },
      {     27,    -72,     90,    -76,     34,     20,    -67,     90,    -80,     41,     12,    -61,     88,    -83,     48,      4,    -55,     86,    -86,     55,     -3,    -49,     84,    -88,     60,    -11,    -42,     81,    -89,     66,    -19,    -35,     77,    -90,     71,    -26,    -28,     72,    -90,     76,    -33,    -21,     67,    -90,     79,    -40,    -13,     62,    -89,     83,    -47,     -6,     56,    -87,     86,    -54,      2,     50,    -84,     88,    -60,     10,     43,    -81,     89,    -65,     18,     36,    -77,     90,    -70,     25,     29,    -73,     90,    -75,     32,     22,    -68,     90,    -79,     39,     14,    -63,     89,    -82,     46,      7,    -57,     87,    -85,     53,     -1,    -51,     85,    -88,     59,     -9,    -44,     81,    -89,     64,    -16,    -37,     78,    -90,     70,    -24,    -30,     74,    -90,     74,    -31,    -23,     69,    -90,     78,    -38,    -15,     64,    -89,     82,    -45,     -8,     58,    -87,     85,    -52 },
      {     26,    -70,     90,    -80,     43,      8,    -56,     86,    -87,     59,    -11,    -40,     78,    -90,     72,    -29,    -23,     67,    -89,     81,    -46,     -4,     54,    -85,     88,    -61,     14,     37,    -77,     90,    -74,     32,     20,    -65,     89,    -83,     49,      1,    -51,     84,    -89,     64,    -18,    -34,     75,    -90,     76,    -35,    -16,     63,    -88,     84,    -52,      2,     48,    -82,     89,    -66,     21,     31,    -73,     90,    -77,     38,     13,    -60,     88,    -85,     55,     -6,    -45,     81,    -90,     68,    -24,    -28,     71,    -90,     79,    -41,    -10,     58,    -87,     86,    -57,      9,     42,    -79,     90,    -70,     27,     25,    -69,     90,    -81,     44,      7,    -55,     86,    -87,     60,    -12,    -39,     78,    -90,     72,    -30,    -22,     67,    -89,     82,    -47,     -3,     53,    -85,     88,    -62,     15,     36,    -76,     90,    -74,     33,     19,    -64,     89,    -83,     50 },
      {     25,    -67,     89,    -83,     52,     -4,    -44,     79,    -90,     73,    -33,    -16,     61,    -87,     86,    -59,     13,     36,    -75,     90,    -78,     41,      8,    -55,     85,    -89,     65,    -22,    -28,     70,    -90,     82,    -49,      1,     47,    -81,     90,    -71,     30,     20,    -64,     88,    -85,     56,    -10,    -39,     77,    -90,     76,    -38,    -11,     57,    -86,     88,    -63,     19,     31,    -72,     90,    -81,     46,      2,    -50,     82,    -89,     69,    -27,    -23,     66,    -89,     84,    -54,      7,     42,    -78,     90,    -74,     35,     14,    -60,     87,    -87,     60,    -15,    -34,     74,    -90,     79,    -43,     -6,     53,    -84,     89,    -67,     24,     26,    -68,     89,    -83,     51,     -3,    -45,     80,    -90,     72,    -32,    -18,     62,    -88,     86,    -58,     12,     37,    -76,     90,    -77,     40,      9,    -55,     85,    -88,     64,    -21,    -29,     70,    -90,     81,    -48 },
      {     24,    -65,     88,    -86,     60,    -16,    -31,     70,    -89,     83,    -54,      9,     38,    -75,     90,    -80,     47,     -1,    -45,     79,    -90,     76,    -40,     -7,     52,    -82,     90,    -72,     33,     14,    -58,     85,    -89,     67,    -26,    -22,     64,    -88,     87,    -61,     19,     29,    -69,     89,    -84,     55,    -11,    -36,     74,    -90,     81,    -49,      3,     43,    -78,     90,    -77,     42,      4,    -50,     81,    -90,     73,    -35,    -12,     56,    -85,     89,    -68,     28,     20,    -62,     87,    -87,     63,    -21,    -27,     67,    -89,     85,    -57,     13,     34,    -72,     90,    -82,     51,     -6,    -41,     77,    -90,     78,    -44,     -2,     48,    -81,     90,    -74,     37,     10,    -55,     84,    -89,     70,    -30,    -18,     60,    -86,     88,    -64,     23,     25,    -66,     88,    -86,     59,    -15,    -32,     71,    -90,     83,    -53,      8,     39,    -76,     90,    -79,     46 },
      {     23,    -63,     87,    -88,     67,    -28,    -18,     59,    -85,     89,    -70,     33,     12,    -55,     83,    -90,     74,    -38,     -7,     50,    -81,     90,    -77,     43,      1,    -45,     78,    -90,     79,    -48,      4,     40,    -75,     90,    -82,     53,    -10,    -35,     72,    -89,     84,    -57,     15,     30,    -68,     89,    -86,     61,    -21,    -25,     64,    -87,     88,    -65,     26,     20,    -60,     86,    -89,     69,    -31,    -14,     56,    -84,     90,    -72,     36,      9,    -52,     81,    -90,     76,    -41,     -3,     47,    -79,     90,    -78,     46,     -2,    -42,     76,    -90,     81,    -51,      8,     37,    -73,     90,    -83,     55,    -13,    -32,     70,    -89,     85,    -60,     19,     27,    -66,     88,    -87,     64,    -24,    -22,     62,    -86,     88,    -67,     29,     16,    -58,     85,    -89,     71,    -34,    -11,     54,    -82,     90,    -74,     39,      6,    -49,     80,    -90,     77,    -44 },
      {     22,    -60,     85,    -89,     73,    -39,     -3,     45,    -77,     90,    -82,     55,    -15,    -28,     65,    -87,     88,    -69,     33,     10,    -51,     80,    -90,     79,    -50,      9,     34,    -70,     89,    -87,     64,    -27,    -16,     56,    -83,     90,    -76,     44,     -2,    -40,     74,    -90,     85,    -60,     21,     23,    -61,     85,    -89,     72,    -38,     -4,     46,    -77,     90,    -82,     55,    -14,    -29,     66,    -87,     88,    -68,     32,     11,    -52,     81,    -90,     79,    -49,      8,     35,    -70,     89,    -86,     64,    -26,    -18,     57,    -83,     90,    -76,     43,     -1,    -41,     74,    -90,     84,    -59,     20,     24,    -62,     86,    -89,     72,    -37,     -6,     47,    -78,     90,    -81,     54,    -13,    -30,     67,    -88,     88,    -67,     31,     12,    -53,     81,    -90,     78,    -48,      7,     36,    -71,     89,    -86,     63,    -25,    -19,     58,    -84,     90,    -75,     42 },
      {     21,    -58,     83,    -90,     78,    -50,     11,     30,    -65,     86,    -89,     73,    -41,      1,     39,    -72,     89,    -87,     67,    -32,     -9,     48,    -77,     90,    -84,     60,    -23,    -19,     56,    -82,     90,    -79,     52,    -13,    -28,     64,    -86,     89,    -74,     43,     -3,    -37,     70,    -88,     88,    -68,     34,      7,    -46,     76,    -90,     85,    -61,     25,     16,    -55,     81,    -90,     81,    -54,     15,     26,    -62,     85,    -90,     76,    -45,      6,     35,    -69,     88,    -88,     70,    -36,     -4,     44,    -75,     90,    -85,     63,    -27,    -14,     53,    -80,     90,    -81,     55,    -18,    -24,     60,    -84,     90,    -77,     47,     -8,    -33,     67,    -87,     89,    -71,     38,      2,    -42,     74,    -89,     86,    -64,     29,     12,    -51,     79,    -90,     82,    -57,     20,     22,    -59,     83,    -90,     78,    -49,     10,     31,    -66,     87,    -89,     72,    -40 },
      {     20,    -55,     81,    -90,     83,    -60,     25,     14,    -51,     78,    -90,     85,    -64,     30,      9,    -46,     75,    -89,     87,    -67,     35,      3,    -41,     72,    -88,     88,    -71,     40,     -2,    -36,     68,    -87,     89,    -74,     45,     -8,    -31,     64,    -85,     90,    -77,     50,    -13,    -26,     60,    -83,     90,    -80,     55,    -19,    -21,     56,    -81,     90,    -82,     59,    -24,    -15,     52,    -78,     90,    -85,     63,    -29,    -10,     47,    -76,     89,    -86,     67,    -34,     -4,     42,    -72,     89,    -88,     70,    -39,      1,     37,    -69,     87,    -89,     74,    -44,      7,     32,    -65,     86,    -90,     77,    -49,     12,     27,    -61,     84,    -90,     79,    -54,     18,     22,    -57,     81,    -90,     82,    -58,     23,     16,    -53,     79,    -90,     84,    -62,     28,     11,    -48,     76,    -90,     86,    -66,     33,      6,    -43,     73,    -89,     88,    -70,     38 },
      {     19,    -53,     78,    -90,     86,    -68,     38,     -2,    -34,     65,    -85,     90,    -80,     56,    -23,    -14,     49,    -76,     89,    -88,     71,    -42,      7,     30,    -62,     83,    -90,     82,    -60,     27,     10,    -45,     73,    -88,     89,    -74,     46,    -11,    -26,     59,    -81,     90,    -84,     63,    -31,     -6,     41,    -70,     87,    -89,     76,    -50,     15,     22,    -55,     79,    -90,     85,    -66,     35,      1,    -37,     67,    -86,     90,    -78,     54,    -20,    -18,     52,    -77,     90,    -87,     69,    -39,      3,     33,    -64,     85,    -90,     81,    -57,     24,     13,    -48,     75,    -89,     88,    -72,     43,     -8,    -29,     61,    -83,     90,    -82,     60,    -28,     -9,     44,    -72,     88,    -89,     74,    -47,     12,     25,    -58,     81,    -90,     84,    -64,     32,      4,    -40,     70,    -87,     89,    -77,     51,    -16,    -21,     55,    -79,     90,    -86,     67,    -36 },
      {     18,    -50,     75,    -89,     89,    -76,     51,    -19,    -16,     49,    -74,     88,    -89,     76,    -52,     20,     15,    -48,     74,    -88,     89,    -77,     53,    -21,    -14,     47,    -73,     88,    -89,     77,    -54,     22,     13,    -46,     72,    -88,     89,    -78,     55,    -23,    -12,     45,    -72,     87,    -90,     78,    -55,     24,     11,    -44,     71,    -87,     90,    -79,     56,    -25,    -10,     43,    -70,     87,    -90,     79,    -57,     26,      9,    -42,     70,    -86,     90,    -80,     58,    -27,     -8,     41,    -69,     86,    -90,     81,    -59,     28,      7,    -40,     68,    -86,     90,    -81,     60,    -29,     -6,     39,    -67,     85,    -90,     81,    -60,     30,      4,    -38,     67,    -85,     90,    -82,     61,    -31,     -3,     37,    -66,     85,    -90,     82,    -62,     32,      2,    -36,     65,    -84,     90,    -83,     63,    -33,     -1,     35,    -64,     84,    -90,     83,    -64,     34 },
      {     16,    -47,     72,    -87,     90,    -81,     62,    -34,      2,     30,    -59,     79,    -90,     88,    -74,     51,    -21,    -12,     43,    -69,     85,    -90,     83,    -65,     38,     -7,    -26,     55,    -77,     89,    -89,     77,    -55,     25,      8,    -39,     66,    -84,     90,    -85,     68,    -42,     11,     22,    -52,     75,    -88,     89,    -79,     58,    -29,     -3,     35,    -63,     82,    -90,     86,    -71,     46,    -15,    -18,     48,    -72,     87,    -90,     81,    -61,     33,     -1,    -31,     60,    -80,     90,    -88,     74,    -50,     20,     13,    -44,     70,    -86,     90,    -83,     64,    -37,      6,     27,    -56,     78,    -89,     89,    -76,     54,    -24,     -9,     40,    -67,     84,    -90,     85,    -67,     41,    -10,    -23,     53,    -76,     88,    -89,     78,    -57,     28,      4,    -36,     64,    -82,     90,    -86,     70,    -45,     14,     19,    -49,     73,    -87,     90,    -81,     60,    -32 },
      {     15,    -44,     68,    -84,     90,    -86,     72,    -49,     21,     10,    -39,     64,    -82,     90,    -88,     75,    -54,     26,      4,    -34,     60,    -79,     89,    -89,     78,    -58,     31,     -1,    -29,     56,    -77,     88,    -90,     81,    -62,     36,     -7,    -24,     52,    -74,     87,    -90,     83,    -66,     41,    -12,    -19,     47,    -70,     85,    -90,     85,    -70,     46,    -18,    -13,     42,    -67,     83,    -90,     87,    -73,     51,    -23,     -8,     37,    -63,     81,    -90,     88,    -76,     55,    -28,     -2,     32,    -59,     78,    -89,     89,    -79,     60,    -33,      3,     27,    -55,     76,    -88,     90,    -81,     64,    -38,      9,     22,    -50,     72,    -86,     90,    -84,     67,    -43,     14,     16,    -45,     69,    -85,     90,    -86,     71,    -48,     20,     11,    -40,     65,    -82,     90,    -87,     74,    -53,     25,      6,    -35,     61,    -80,     89,    -89,     77,    -57,     30 },
      {     14,    -41,     64,    -81,     89,    -89,     79,    -62,     38,    -11,    -18,     44,    -67,     82,    -90,     88,    -78,     60,    -35,      8,     21,    -47,     69,    -84,     90,    -88,     76,    -57,     32,     -4,    -24,     50,    -71,     85,    -90,     87,    -74,     55,    -29,      1,     27,    -53,     73,    -86,     90,    -86,     72,    -52,     26,      2,    -30,     55,    -75,     87,    -90,     85,    -70,     49,    -23,     -6,     33,    -58,     77,    -88,     90,    -83,     68,    -46,     20,      9,    -36,     60,    -78,     89,    -90,     82,    -66,     43,    -16,    -12,     39,    -63,     80,    -89,     89,    -81,     64,    -40,     13,     15,    -42,     65,    -81,     90,    -89,     79,    -61,     37,    -10,    -19,     45,    -67,     83,    -90,     88,    -77,     59,    -34,      7,     22,    -48,     70,    -84,     90,    -87,     76,    -56,     31,     -3,    -25,     51,    -72,     85,    -90,     86,    -74,     54,    -28 },
      {     13,    -38,     60,    -77,     88,    -90,     85,    -73,     55,    -31,      6,     21,    -45,     66,    -81,     89,    -90,     82,    -68,     48,    -24,     -2,     28,    -52,     71,    -84,     90,    -88,     79,    -63,     41,    -16,    -10,     35,    -58,     76,    -87,     90,    -86,     75,    -57,     34,     -9,    -18,     42,    -64,     79,    -89,     90,    -84,     70,    -51,     27,     -1,    -25,     49,    -69,     83,    -90,     89,    -81,     65,    -44,     20,      7,    -32,     55,    -74,     86,    -90,     87,    -77,     60,    -37,     12,     14,    -39,     61,    -78,     88,    -90,     85,    -72,     54,    -30,      4,     22,    -46,     67,    -81,     89,    -89,     82,    -67,     47,    -23,     -3,     29,    -53,     72,    -85,     90,    -88,     78,    -62,     40,    -15,    -11,     36,    -59,     76,    -87,     90,    -86,     74,    -56,     33,     -8,    -19,     43,    -64,     80,    -89,     90,    -83,     70,    -50,     26 },
      {     12,    -35,     56,    -73,     85,    -90,     89,    -81,     68,    -50,     28,     -4,    -20,     42,    -62,     77,    -87,     90,    -87,     78,    -63,     43,    -21,     -3,     27,    -49,     67,    -81,     89,    -90,     85,    -74,     57,    -36,     13,     11,    -34,     55,    -72,     84,    -90,     89,    -82,     69,    -51,     29,     -6,    -19,     41,    -61,     77,    -87,     90,    -88,     78,    -64,     44,    -22,     -2,     26,    -48,     67,    -81,     89,    -90,     85,    -74,     58,    -37,     14,     10,    -33,     55,    -72,     84,    -90,     89,    -82,     70,    -52,     30,     -7,    -18,     40,    -60,     76,    -86,     90,    -88,     79,    -64,     45,    -23,     -1,     25,    -47,     66,    -80,     88,    -90,     86,    -75,     59,    -38,     15,      9,    -32,     54,    -71,     83,    -90,     89,    -83,     70,    -53,     31,     -8,    -16,     39,    -60,     76,    -86,     90,    -88,     79,    -65,     46,    -24 },
      {     11,    -32,     52,    -68,     81,    -88,     90,    -87,     79,    -66,     49,    -29,      8,     14,    -35,     55,    -70,     82,    -89,     90,    -86,     77,    -64,     46,    -26,      4,     18,    -38,     57,    -72,     83,    -89,     90,    -85,     76,    -61,     43,    -23,      1,     21,    -41,     60,    -74,     85,    -90,     90,    -84,     74,    -59,     40,    -20,     -2,     24,    -44,     62,    -76,     86,    -90,     89,    -83,     72,    -56,     37,    -16,     -6,     27,    -47,     64,    -78,     87,    -90,     89,    -81,     70,    -54,     34,    -13,     -9,     30,    -50,     67,    -79,     88,    -90,     88,    -80,     67,    -51,     31,    -10,    -12,     33,    -53,     69,    -81,     88,    -90,     87,    -78,     65,    -48,     28,     -7,    -15,     36,    -55,     71,    -82,     89,    -90,     86,    -77,     63,    -45,     25,     -3,    -19,     39,    -58,     73,    -84,     89,    -90,     85,    -75,     60,    -42,     22 },
      {     10,    -29,     47,    -63,     76,    -85,     89,    -90,     86,    -78,     67,    -52,     34,    -15,     -4,     24,    -42,     59,    -72,     82,    -89,     90,    -88,     81,    -70,     56,    -39,     21,     -1,    -19,     37,    -55,     69,    -80,     87,    -90,     89,    -83,     74,    -60,     44,    -26,      7,     13,    -32,     50,    -65,     77,    -86,     90,    -90,     85,    -77,     64,    -49,     31,    -12,     -8,     27,    -45,     61,    -74,     84,    -89,     90,    -87,     79,    -68,     54,    -36,     18,      2,    -22,     40,    -57,     71,    -81,     88,    -90,     88,    -82,     72,    -58,     41,    -23,      3,     16,    -35,     53,    -67,     79,    -87,     90,    -89,     84,    -75,     62,    -46,     28,     -9,    -11,     30,    -48,     64,    -76,     85,    -90,     90,    -86,     78,    -66,     51,    -33,     14,      6,    -25,     43,    -60,     73,    -83,     89,    -90,     88,    -81,     70,    -55,     38,    -20 },
      {      9,    -26,     42,    -57,     70,    -79,     86,    -90,     90,    -87,     80,    -70,     58,    -43,     27,    -10,     -8,     25,    -41,     56,    -69,     79,    -86,     90,    -90,     87,    -81,     71,    -59,     44,    -28,     11,      7,    -24,     40,    -55,     68,    -78,     86,    -90,     90,    -87,     81,    -72,     60,    -45,     29,    -12,     -6,     23,    -39,     55,    -67,     78,    -85,     89,    -90,     88,    -81,     72,    -60,     46,    -30,     13,      4,    -22,     38,    -54,     67,    -77,     85,    -89,     90,    -88,     82,    -73,     61,    -47,     31,    -14,     -3,     21,    -37,     53,    -66,     77,    -85,     89,    -90,     88,    -82,     74,    -62,     48,    -32,     15,      2,    -20,     36,    -52,     65,    -76,     84,    -89,     90,    -88,     83,    -74,     63,    -49,     33,    -16,     -1,     19,    -35,     51,    -64,     76,    -84,     89,    -90,     89,    -83,     75,    -64,     50,    -34,     18 },
      {      8,    -23,     37,    -51,     63,    -73,     81,    -87,     90,    -90,     88,    -83,     76,    -67,     55,    -42,     28,    -13,     -2,     18,    -32,     46,    -59,     70,    -78,     85,    -89,     90,    -89,     85,    -79,     70,    -60,     47,    -33,     19,     -3,    -12,     27,    -41,     55,    -66,     76,    -83,     88,    -90,     90,    -87,     81,    -74,     64,    -52,     38,    -24,      9,      7,    -22,     36,    -50,     62,    -72,     81,    -86,     90,    -90,     88,    -84,     77,    -67,     56,    -43,     29,    -14,     -1,     16,    -31,     45,    -58,     69,    -78,     85,    -89,     90,    -89,     86,    -79,     71,    -60,     48,    -34,     20,     -4,    -11,     26,    -40,     54,    -65,     75,    -82,     88,    -90,     90,    -87,     82,    -74,     64,    -53,     39,    -25,     10,      6,    -21,     35,    -49,     61,    -72,     80,    -86,     89,    -90,     89,    -84,     77,    -68,     57,    -44,     30,    -15 },
      {      7,    -20,     32,    -44,     55,    -65,     74,    -81,     86,    -89,     90,    -90,     87,    -83,     77,    -69,     60,    -49,     37,    -25,     12,      1,    -14,     27,    -39,     51,    -61,     70,    -78,     84,    -88,     90,    -90,     89,    -85,     79,    -72,     64,    -54,     42,    -30,     18,     -4,     -9,     22,    -34,     46,    -57,     67,    -75,     81,    -86,     89,    -90,     89,    -87,     82,    -76,     67,    -58,     47,    -35,     23,    -10,     -3,     16,    -29,     41,    -53,     63,    -72,     79,    -85,     88,    -90,     90,    -88,     84,    -78,     71,    -62,     52,    -40,     28,    -15,      2,     11,    -24,     36,    -48,     59,    -68,     76,    -82,     87,    -90,     90,    -89,     86,    -81,     74,    -66,     56,    -45,     33,    -21,      8,      6,    -19,     31,    -43,     55,    -64,     73,    -80,     85,    -89,     90,    -90,     88,    -83,     77,    -70,     60,    -50,     38,    -26,     13 },
      {      6,    -16,     27,    -37,     47,    -56,     64,    -72,     78,    -83,     87,    -89,     90,    -90,     89,    -86,     81,    -76,     70,    -62,     54,    -44,     34,    -24,     13,     -2,     -9,     20,    -30,     40,    -50,     59,    -67,     74,    -79,     84,    -88,     90,    -90,     90,    -88,     85,    -80,     74,    -67,     60,    -51,     41,    -31,     21,    -10,     -1,     12,    -23,     33,    -43,     53,    -61,     69,    -76,     81,    -85,     88,    -90,     90,    -89,     87,    -83,     78,    -72,     65,    -57,     48,    -38,     28,    -18,      7,      4,    -15,     26,    -36,     46,    -55,     64,    -71,     77,    -82,     86,    -89,     90,    -90,     89,    -86,     82,    -77,     70,    -63,     55,    -45,     35,    -25,     14,     -3,     -8,     19,    -29,     39,    -49,     58,    -66,     73,    -79,     84,    -87,     89,    -90,     90,    -88,     85,    -81,     75,    -68,     60,    -52,     42,    -32,     22,    -11 },
      {      4,    -13,     22,    -30,     38,    -46,     54,    -60,     67,    -72,     77,    -81,     85,    -88,     89,    -90,     90,    -89,     88,    -85,     82,    -78,     73,    -67,     61,    -55,     47,    -39,     31,    -23,     14,     -6,     -3,     12,    -21,     29,    -37,     45,    -53,     60,    -66,     72,    -77,     81,    -85,     87,    -89,     90,    -90,     90,    -88,     86,    -82,     78,    -74,     68,    -62,     55,    -48,     40,    -32,     24,    -15,      7,      2,    -11,     20,    -28,     36,    -44,     52,    -59,     65,    -71,     76,    -81,     84,    -87,     89,    -90,     90,    -90,     88,    -86,     83,    -79,     74,    -69,     63,    -56,     49,    -41,     33,    -25,     16,     -8,     -1,     10,    -19,     27,    -35,     43,    -51,     58,    -64,     70,    -76,     80,    -84,     87,    -89,     90,    -90,     90,    -89,     86,    -83,     79,    -75,     70,    -64,     57,    -50,     42,    -34,     26,    -18,      9 },
      {      3,    -10,     16,    -23,     29,    -35,     41,    -47,     53,    -58,     63,    -67,     72,    -76,     79,    -82,     85,    -87,     88,    -89,     90,    -90,     90,    -89,     88,    -86,     84,    -81,     78,    -75,     71,    -67,     62,    -57,     52,    -46,     40,    -34,     28,    -22,     15,     -9,      2,      4,    -11,     18,    -24,     30,    -36,     42,    -48,     54,    -59,     64,    -68,     72,    -76,     79,    -82,     85,    -87,     89,    -90,     90,    -90,     90,    -89,     88,    -86,     84,    -81,     78,    -74,     70,    -66,     61,    -56,     51,    -45,     39,    -33,     27,    -21,     14,     -8,      1,      6,    -12,     19,    -25,     31,    -37,     43,    -49,     55,    -60,     64,    -69,     73,    -77,     80,    -83,     85,    -87,     89,    -90,     90,    -90,     90,    -89,     88,    -86,     83,    -81,     77,    -74,     70,    -65,     60,    -55,     50,    -44,     38,    -32,     26,    -20,     13,     -7 },
      {      2,     -7,     11,    -15,     20,    -24,     28,    -32,     36,    -40,     44,    -48,     52,    -55,     59,    -62,     65,    -68,     71,    -74,     76,    -78,     81,    -82,     84,    -86,     87,    -88,     89,    -90,     90,    -90,     90,    -90,     90,    -89,     88,    -87,     86,    -85,     83,    -81,     79,    -77,     74,    -72,     69,    -66,     63,    -60,     56,    -53,     49,    -45,     41,    -37,     33,    -29,     25,    -21,     16,    -12,      8,     -3,     -1,      6,    -10,     14,    -19,     23,    -27,     31,    -35,     39,    -43,     47,    -51,     55,    -58,     61,    -64,     67,    -70,     73,    -76,     78,    -80,     82,    -84,     85,    -87,     88,    -89,     89,    -90,     90,    -90,     90,    -90,     89,    -89,     88,    -86,     85,    -83,     81,    -79,     77,    -75,     72,    -70,     67,    -64,     60,    -57,     54,    -50,     46,    -42,     38,    -34,     30,    -26,     22,    -18,     13,     -9,      4 },
      {      1,     -3,      6,     -8,     10,    -12,     14,    -16,     19,    -21,     23,    -25,     27,    -29,     31,    -33,     35,    -37,     39,    -41,     43,    -45,     47,    -49,     51,    -53,     55,    -56,     58,    -60,     61,    -63,     64,    -66,     67,    -69,     70,    -72,     73,    -74,     76,    -77,     78,    -79,     80,    -81,     82,    -83,     84,    -85,     85,    -86,     87,    -87,     88,    -88,     89,    -89,     89,    -90,     90,    -90,     90,    -90,     90,    -90,     90,    -90,     90,    -90,     89,    -89,     89,    -88,     88,    -87,     86,    -86,     85,    -84,     83,    -82,     81,    -81,     79,    -78,     77,    -76,     75,    -74,     72,    -71,     70,    -68,     67,    -65,     64,    -62,     60,    -59,     57,    -55,     54,    -52,     50,    -48,     46,    -44,     42,    -40,     38,    -36,     34,    -32,     30,    -28,     26,    -24,     22,    -20,     18,    -15,     13,    -11,      9,     -7,      4,     -2 },
    },
    {
      {     90,     90,     90,     90,     90,     90,     90,     90,     90,     90,     90,     89,     89,     89,     89,     89,     89,     88,     88,     88,     88,     87,     87,     87,     86,     86,     86,     85,     85,     85,     84,     84,     83,     83,     82,     82,     81,     81,     81,     80,     79,     79,     78,     78,     77,     77,     76,     76,     75,     74,     74,     73,     72,     72,     71,     70,     70,     69,     68,     67,     67,     66,     65,     64,     64,     63,     62,     61,     60,     60,     59,     58,     57,     56,     55,     55,     54,     53,     52,     51,     50,     49,     48,     47,     46,     45,     44,     43,     42,     41,     40,     39,     38,     37,     36,     35,     34,     33,     32,     31,     30,     29,     28,     27,     26,     25,     24,     23,     22,     21,     20,     19,     18,     16,     15,     14,     13,     12,     11,     10,      9,      8,      7,      6,      4,      3,      2,      1 },
      {     90,     90,     90,     90,     89,     89,     88,     87,     86,     85,     84,     82,     81,     79,     78,     76,     74,     72,     70,     68,     66,     64,     61,     59,     56,     54,     51,     48,     45,     42,     39,     36,     33,     30,     27,     24,     21,     18,     14,     11,      8,      4,      1,     -2,     -6,     -9,    -12,    -15,    -19,    -22,    -25,    -28,    -31,    -34,    -37,    -40,    -43,    -46,    -49,    -52,    -55,    -57,    -60,    -62,    -64,    -67,    -69,    -71,    -73,    -75,    -77,    -78,    -80,    -81,    -83,    -84,    -85,    -86,    -87,    -88,    -89,    -89,    -90,    -90,    -90,    -90,    -90,    -90,    -90,    -89,    -89,    -88,    -88,    -87,    -86,    -85,    -83,    -82,    -81,    -79,    -77,    -76,    -74,    -72,    -70,    -67,    -65,    -63,    -60,    -58,    -55,    -53,    -50,    -47,    -44,    -41,    -38,    -35,    -32,    -29,    -26,    -23,    -20,    -16,    -13,    -10,     -7,     -3 },
      {     90,     90,     89,     88,     87,     85,     83,     81,     78,     76,     72,     69,     65,     61,     57,     53,     48,     43,     38,     33,     28,     23,     18,     12,      7,      1,     -4,    -10,    -15,    -21,    -26,    -31,    -36,    -41,    -46,    -51,    -55,    -60,    -64,    -67,    -71,    -74,    -77,    -80,    -82,    -85,    -86,    -88,    -89,    -90,    -90,    -90,    -90,    -90,    -89,    -88,    -86,    -84,    -82,    -79,    -77,    -74,    -70,    -67,    -63,    -59,    -55,    -50,    -45,    -40,    -35,    -30,    -25,    -20,    -14,     -9,     -3,      2,      8,     13,     19,     24,     29,     34,     39,     44,     49,     54,     58,     62,     66,     70,     73,     76,     79,     81,     84,     86,     87,     89,     89,     90,     90,     90,     90,     89,     88,     87,     85,     83,     81,     78,     75,     72,     68,     64,     60,     56,     52,     47,     42,     37,     32,     27,     22,     16,     11,      6 },
      {     90,     90,     88,     86,     84,     81,     77,     72,     67,     62,     56,     50,     43,     36,     29,     22,     14,      7,     -1,     -9,    -16,    -24,    -31,    -38,    -45,    -52,    -58,    -64,    -69,    -74,    -78,    -81,    -85,    -87,    -89,    -90,    -90,    -90,    -89,    -88,    -86,    -83,    -79,    -76,    -71,    -66,    -60,    -55,    -48,    -41,    -34,    -27,    -20,    -12,     -4,      3,     11,     19,     26,     33,     40,     47,     54,     60,     65,     70,     75,     79,     82,     85,     88,     89,     90,     90,     90,     89,     87,     85,     82,     78,     74,     70,     64,     59,     53,     46,     39,     32,     25,     18,     10,      2,     -6,    -13,    -21,    -28,    -35,    -42,    -49,    -55,    -61,    -67,    -72,    -76,    -80,    -83,    -86,    -88,    -89,    -90,    -90,    -90,    -89,    -87,    -84,    -81,    -77,    -73,    -68,    -63,    -57,    -51,    -44,    -37,    -30,    -23,    -15,     -8 },
      {     90,     89,     87,     84,     79,     74,     68,     61,     54,     45,     36,     27,     18,      8,     -2,    -12,    -22,    -31,    -40,    -49,    -57,    -64,    -71,    -77,    -81,    -85,    -88,    -90,    -90,    -90,    -88,    -86,    -82,    -77,    -72,    -65,    -58,    -50,    -41,    -32,    -23,    -13,     -3,      7,     16,     26,     35,     44,     53,     60,     67,     74,     79,     83,     87,     89,     90,     90,     89,     87,     84,     80,     75,     69,     62,     55,     46,     37,     28,     19,      9,     -1,    -11,    -21,    -30,    -39,    -48,    -56,    -64,    -70,    -76,    -81,    -85,    -88,    -90,    -90,    -90,    -89,    -86,    -82,    -78,    -72,    -66,    -59,    -51,    -42,    -33,    -24,    -14,     -4,      6,     15,     25,     34,     43,     52,     60,     67,     73,     78,     83,     86,     89,     90,     90,     89,     88,     85,     81,     76,     70,     63,     55,     47,     38,     29,     20,     10 },
      {     90,     89,     85,     81,     74,     67,     58,     48,     37,     26,     14,      2,    -10,    -22,    -33,    -44,    -55,    -64,    -72,    -78,    -84,    -88,    -90,    -90,    -89,    -87,    -82,    -77,    -70,    -61,    -52,    -41,    -30,    -19,     -7,      6,     18,     29,     40,     51,     60,     69,     76,     82,     86,     89,     90,     90,     88,     84,     79,     72,     64,     55,     45,     34,     23,     11,     -1,    -13,    -25,    -36,    -47,    -57,    -66,    -74,    -80,    -85,    -88,    -90,    -90,    -89,    -86,    -81,    -75,    -67,    -59,    -49,    -38,    -27,    -15,     -3,      9,     21,     32,     43,     54,     63,     71,     78,     83,     87,     90,     90,     89,     87,     83,     77,     70,     62,     53,     42,     31,     20,      8,     -4,    -16,    -28,    -39,    -50,    -60,    -68,    -76,    -81,    -86,    -89,    -90,    -90,    -88,    -85,    -79,    -73,    -65,    -56,    -46,    -35,    -24,    -12 },
      {     90,     88,     83,     77,     68,     58,     46,     33,     20,      6,     -9,    -23,    -36,    -49,    -60,    -70,    -78,    -85,    -89,    -90,    -90,    -87,    -82,    -75,    -66,    -55,    -43,    -30,    -16,     -2,     12,     26,     39,     52,     63,     72,     80,     86,     89,     90,     89,     86,     81,     73,     64,     53,     40,     27,     13,     -1,    -15,    -29,    -42,    -55,    -65,    -74,    -81,    -87,    -90,    -90,    -89,    -85,    -79,    -71,    -61,    -50,    -37,    -24,    -10,      4,     19,     32,     45,     57,     67,     76,     83,     88,     90,     90,     88,     84,     77,     69,     59,     47,     34,     21,      7,     -8,    -22,    -35,    -48,    -60,    -70,    -78,    -84,    -88,    -90,    -90,    -87,    -82,    -76,    -67,    -56,    -44,    -31,    -18,     -3,     11,     25,     38,     51,     62,     72,     79,     85,     89,     90,     89,     86,     81,     74,     64,     54,     41,     28,     14 },
      {     90,     87,     81,     72,     61,     48,     33,     18,      1,    -15,    -31,    -46,    -60,    -71,    -80,    -86,    -90,    -90,    -88,    -82,    -74,    -63,    -50,    -35,    -20,     -3,     13,     29,     44,     58,     70,     79,     86,     89,     90,     88,     83,     75,     64,     52,     37,     22,      6,    -11,    -27,    -42,    -56,    -68,    -78,    -85,    -89,    -90,    -89,    -84,    -76,    -66,    -54,    -39,    -24,     -8,      9,     25,     40,     55,     67,     77,     84,     89,     90,     89,     85,     77,     67,     55,     41,     26,     10,     -7,    -23,    -38,    -53,    -65,    -76,    -83,    -88,    -90,    -89,    -85,    -78,    -69,    -57,    -43,    -28,    -12,      4,     21,     36,     51,     64,     74,     82,     88,     90,     90,     86,     79,     70,     59,     45,     30,     14,     -2,    -19,    -34,    -49,    -62,    -73,    -81,    -87,    -90,    -90,    -87,    -81,    -72,    -60,    -47,    -32,    -16 },
      {     90,     86,     78,     67,     54,     37,     20,      1,    -18,    -35,    -52,    -66,    -77,    -85,    -90,    -90,    -87,    -79,    -69,    -55,    -39,    -22,     -3,     15,     33,     50,     64,     76,     85,     89,     90,     87,     81,     70,     57,     41,     24,      6,    -13,    -31,    -48,    -63,    -75,    -84,    -89,    -90,    -88,    -81,    -72,    -59,    -43,    -26,     -8,     11,     29,     46,     61,     74,     83,     89,     90,     88,     82,     73,     60,     45,     28,     10,     -9,    -27,    -44,    -60,    -72,    -82,    -88,    -90,    -89,    -83,    -74,    -62,    -47,    -30,    -12,      7,     25,     42,     58,     71,     81,     88,     90,     89,     84,     76,     64,     49,     32,     14,     -4,    -23,    -40,    -56,    -70,    -80,    -87,    -90,    -89,    -85,    -77,    -65,    -51,    -34,    -16,      2,     21,     38,     55,     68,     79,     86,     90,     90,     86,     78,     67,     53,     36,     19 },
      {     90,     85,     76,     62,     45,     26,      6,    -15,    -35,    -54,    -69,    -81,    -88,    -90,    -88,    -81,    -70,    -55,    -36,    -16,      4,     25,     44,     61,     75,     85,     90,     90,     85,     76,     63,     46,     27,      7,    -14,    -34,    -53,    -68,    -80,    -88,    -90,    -88,    -81,    -70,    -55,    -37,    -18,      3,     24,     43,     60,     74,     84,     89,     90,     86,     77,     64,     47,     28,      8,    -13,    -33,    -52,    -67,    -79,    -87,    -90,    -89,    -82,    -71,    -56,    -38,    -19,      2,     23,     42,     60,     74,     84,     89,     90,     86,     77,     64,     48,     29,      9,    -12,    -32,    -51,    -67,    -79,    -87,    -90,    -89,    -82,    -72,    -57,    -39,    -20,      1,     22,     41,     59,     73,     83,     89,     90,     86,     78,     65,     49,     30,     10,    -11,    -31,    -50,    -66,    -78,    -87,    -90,    -89,    -83,    -72,    -58,    -40,    -21 },
      {     90,     84,     72,     56,     36,     14,     -9,    -31,    -52,    -69,    -81,    -89,    -90,    -86,    -76,    -60,    -41,    -20,      3,     26,     47,     65,     79,     88,     90,     87,     78,     64,     46,     25,      2,    -21,    -42,    -61,    -76,    -86,    -90,    -89,    -81,    -68,    -51,    -30,     -8,     15,     37,     57,     73,     84,     90,     89,     83,     72,     55,     35,     13,    -10,    -32,    -53,    -70,    -82,    -89,    -90,    -85,    -75,    -60,    -40,    -19,      4,     27,     48,     66,     79,     88,     90,     87,     78,     64,     45,     24,      1,    -22,    -43,    -62,    -77,    -86,    -90,    -88,    -81,    -67,    -50,    -29,     -7,     16,     38,     58,     74,     85,     90,     89,     83,     71,     55,     34,     12,    -11,    -33,    -54,    -70,    -82,    -89,    -90,    -85,    -74,    -59,    -39,    -18,      6,     28,     49,     67,     80,     88,     90,     87,     77,     63,     44,     23 },
      {     89,     82,     69,     50,     27,      2,    -23,    -46,    -66,    -81,    -89,    -90,    -84,    -72,    -54,    -31,     -7,     19,     42,     63,     78,     88,     90,     86,     74,     57,     35,     11,    -14,    -38,    -60,    -76,    -87,    -90,    -87,    -77,    -60,    -39,    -15,     10,     34,     56,     74,     85,     90,     88,     79,     64,     43,     20,     -6,    -30,    -53,    -71,    -84,    -90,    -89,    -81,    -67,    -47,    -24,      1,     26,     49,     68,     82,     89,     90,     83,     70,     51,     28,      3,    -22,    -45,    -65,    -80,    -89,    -90,    -85,    -72,    -55,    -32,     -8,     18,     41,     62,     78,     88,     90,     86,     75,     58,     36,     12,    -13,    -37,    -59,    -76,    -86,    -90,    -87,    -77,    -61,    -40,    -16,      9,     33,     55,     73,     85,     90,     88,     79,     64,     44,     21,     -4,    -29,    -52,    -70,    -83,    -90,    -89,    -81,    -67,    -48,    -25 },
      {     89,     81,     65,     43,     18,    -10,    -36,    -60,    -77,    -88,    -90,    -84,    -70,    -50,    -25,      2,     29,     54,     73,     86,     90,     87,     75,     56,     32,      6,    -22,    -47,    -68,    -83,    -90,    -89,    -79,    -62,    -39,    -13,     14,     40,     63,     79,     89,     90,     82,     67,     46,     21,     -7,    -33,    -57,    -76,    -87,    -90,    -85,    -72,    -53,    -28,     -1,     26,     51,     71,     85,     90,     88,     77,     59,     35,      9,    -19,    -44,    -66,    -81,    -89,    -89,    -81,    -64,    -42,    -16,     11,     37,     60,     78,     88,     90,     84,     70,     49,     24,     -3,    -30,    -55,    -74,    -86,    -90,    -86,    -74,    -55,    -31,     -4,     23,     48,     69,     83,     90,     88,     78,     61,     38,     12,    -15,    -41,    -64,    -80,    -89,    -90,    -82,    -67,    -45,    -20,      8,     34,     58,     76,     87,     90,     85,     72,     52,     27 },
      {     89,     79,     61,     36,      8,    -22,    -49,    -71,    -85,    -90,    -86,    -72,    -50,    -23,      7,     35,     60,     79,     89,     89,     80,     62,     37,      9,    -21,    -48,    -70,    -85,    -90,    -86,    -72,    -51,    -24,      6,     34,     60,     78,     89,     89,     81,     63,     38,     10,    -20,    -47,    -70,    -85,    -90,    -86,    -73,    -52,    -25,      4,     33,     59,     78,     89,     90,     81,     64,     39,     11,    -19,    -46,    -69,    -84,    -90,    -87,    -74,    -53,    -26,      3,     32,     58,     77,     88,     90,     81,     64,     40,     12,    -18,    -45,    -68,    -84,    -90,    -87,    -74,    -54,    -27,      2,     31,     57,     77,     88,     90,     82,     65,     41,     13,    -16,    -44,    -67,    -83,    -90,    -87,    -75,    -55,    -28,      1,     30,     56,     76,     88,     90,     82,     66,     42,     14,    -15,    -43,    -67,    -83,    -90,    -88,    -76,    -55,    -29 },
      {     89,     78,     57,     29,     -2,    -33,    -60,    -80,    -90,    -88,    -76,    -54,    -25,      7,     37,     64,     82,     90,     87,     73,     50,     21,    -11,    -41,    -67,    -84,    -90,    -86,    -70,    -46,    -16,     15,     45,     70,     85,     90,     84,     67,     42,     12,    -20,    -49,    -72,    -87,    -90,    -82,    -64,    -38,     -8,     24,     53,     75,     88,     90,     81,     61,     34,      3,    -28,    -56,    -77,    -89,    -89,    -78,    -58,    -30,      1,     32,     60,     79,     89,     88,     76,     55,     26,     -6,    -36,    -63,    -81,    -90,    -87,    -74,    -51,    -22,     10,     40,     66,     83,     90,     86,     71,     47,     18,    -14,    -44,    -69,    -85,    -90,    -85,    -68,    -43,    -13,     19,     48,     72,     86,     90,     83,     65,     39,      9,    -23,    -52,    -74,    -88,    -90,    -81,    -62,    -35,     -4,     27,     55,     77,     89,     89,     79,     59,     31 },
      {     89,     76,     53,     22,    -12,    -44,    -70,    -86,    -90,    -81,    -60,    -31,      2,     35,     64,     83,     90,     85,     67,     40,      8,    -26,    -56,    -78,    -89,    -88,    -74,    -49,    -18,     16,     48,     73,     88,     90,     79,     57,     27,     -7,    -39,    -67,    -85,    -90,    -83,    -64,    -36,     -3,     30,     60,     81,     90,     87,     71,     45,     13,    -21,    -52,    -76,    -89,    -89,    -77,    -54,    -23,     11,     43,     70,     86,     90,     81,     61,     32,     -1,    -34,    -63,    -82,    -90,    -85,    -68,    -41,     -9,     25,     55,     78,     89,     88,     74,     50,     19,    -15,    -47,    -72,    -87,    -90,    -79,    -58,    -28,      6,     38,     66,     84,     90,     84,     65,     37,      4,    -29,    -59,    -80,    -90,    -87,    -72,    -46,    -14,     20,     51,     75,     88,     89,     77,     55,     24,    -10,    -42,    -69,    -86,    -90,    -82,    -62,    -33 },
      {     89,     74,     48,     14,    -22,    -55,    -78,    -90,    -87,    -70,    -41,     -7,     29,     60,     82,     90,     84,     64,     34,     -1,    -36,    -66,    -85,    -90,    -81,    -59,    -27,      9,     43,     71,     87,     89,     77,     53,     20,    -16,    -50,    -76,    -89,    -88,    -73,    -46,    -12,     24,     56,     79,     90,     86,     68,     39,      4,    -31,    -62,    -83,    -90,    -83,    -63,    -32,      3,     38,     67,     86,     90,     80,     57,     25,    -11,    -45,    -72,    -88,    -89,    -76,    -51,    -18,     19,     52,     77,     89,     88,     72,     44,     10,    -26,    -58,    -81,    -90,    -85,    -67,    -37,     -2,     33,     64,     84,     90,     82,     61,     30,     -6,    -40,    -69,    -86,    -90,    -79,    -55,    -23,     13,     47,     74,     88,     89,     75,     49,     15,    -21,    -54,    -78,    -90,    -87,    -70,    -42,     -8,     28,     60,     81,     90,     85,     65,     35 },
      {     88,     72,     43,      7,    -31,    -64,    -85,    -90,    -79,    -55,    -20,     19,     54,     79,     90,     85,     64,     32,     -6,    -42,    -72,    -88,    -89,    -73,    -44,     -8,     30,     63,     84,     90,     80,     55,     21,    -18,    -53,    -78,    -90,    -85,    -65,    -33,      4,     41,     71,     88,     89,     74,     45,      9,    -29,    -62,    -84,    -90,    -81,    -56,    -22,     16,     52,     78,     90,     86,     66,     34,     -3,    -40,    -70,    -88,    -89,    -74,    -46,    -10,     28,     61,     83,     90,     81,     57,     23,    -15,    -51,    -77,    -90,    -86,    -67,    -35,      2,     39,     70,     87,     89,     75,     47,     11,    -27,    -60,    -83,    -90,    -81,    -58,    -24,     14,     50,     77,     90,     86,     67,     36,     -1,    -38,    -69,    -87,    -89,    -76,    -48,    -12,     26,     60,     82,     90,     82,     59,     25,    -13,    -49,    -76,    -89,    -87,    -68,    -37 },
      {     88,     70,     38,     -1,    -40,    -72,    -89,    -88,    -69,    -36,      3,     42,     73,     89,     87,     67,     34,     -6,    -44,    -74,    -89,    -86,    -66,    -32,      8,     46,     76,     90,     86,     64,     30,    -10,    -48,    -77,    -90,    -85,    -63,    -28,     12,     50,     78,     90,     84,     61,     26,    -14,    -52,    -79,    -90,    -83,    -60,    -24,     16,     54,     80,     90,     82,     58,     22,    -19,    -55,    -81,    -90,    -81,    -56,    -20,     21,     57,     82,     90,     81,     55,     18,    -23,    -59,    -83,    -90,    -79,    -53,    -15,     25,     60,     84,     90,     78,     51,     13,    -27,    -62,    -85,    -90,    -77,    -49,    -11,     29,     64,     85,     90,     76,     47,      9,    -31,    -65,    -86,    -89,    -75,    -45,     -7,     33,     67,     87,     89,     74,     43,      4,    -35,    -68,    -87,    -89,    -72,    -41,     -2,     37,     70,     88,     88,     71,     39 },
      {     88,     68,     33,     -9,    -49,    -78,    -90,    -82,    -55,    -16,     26,     63,     86,     89,     73,     40,     -1,    -42,    -74,    -90,    -85,    -61,    -24,     19,     57,     83,     90,     77,     47,      7,    -35,    -70,    -88,    -87,    -67,    -31,     11,     51,     79,     90,     81,     54,     14,    -28,    -64,    -86,    -89,    -72,    -38,      3,     44,     76,     90,     84,     60,     22,    -21,    -59,    -84,    -90,    -76,    -45,     -4,     37,     71,     89,     87,     65,     29,    -13,    -53,    -81,    -90,    -80,    -52,    -12,     30,     66,     87,     89,     70,     36,     -6,    -46,    -77,    -90,    -83,    -58,    -20,     23,     60,     85,     90,     75,     43,      2,    -39,    -72,    -89,    -86,    -64,    -27,     15,     55,     81,     90,     79,     50,     10,    -32,    -67,    -88,    -88,    -69,    -34,      8,     48,     78,     90,     82,     56,     18,    -25,    -62,    -85,    -89,    -74,    -41 },
      {     88,     66,     28,    -16,    -57,    -84,    -90,    -74,    -39,      4,     47,     78,     90,     80,     50,      8,    -36,    -72,    -89,    -85,    -60,    -20,     25,     64,     87,     88,     68,     31,    -13,    -55,    -82,    -90,    -76,    -42,      1,     44,     77,     90,     81,     53,     11,    -33,    -70,    -89,    -86,    -62,    -23,     22,     61,     86,     89,     70,     34,    -10,    -52,    -81,    -90,    -77,    -45,     -2,     41,     75,     90,     83,     55,     14,    -30,    -67,    -88,    -87,    -64,    -26,     19,     59,     85,     89,     72,     37,     -7,    -49,    -79,    -90,    -79,    -48,     -6,     38,     73,     90,     84,     58,     18,    -27,    -65,    -87,    -88,    -67,    -29,     15,     56,     83,     90,     74,     40,     -3,    -46,    -78,    -90,    -81,    -51,     -9,     35,     71,     89,     85,     60,     21,    -24,    -63,    -86,    -89,    -69,    -32,     12,     54,     82,     90,     76,     43 },
      {     87,     64,     23,    -24,    -64,    -88,    -87,    -63,    -22,     25,     65,     88,     87,     62,     21,    -26,    -66,    -88,    -86,    -61,    -20,     27,     67,     88,     86,     60,     19,    -28,    -67,    -89,    -86,    -60,    -18,     29,     68,     89,     85,     59,     16,    -30,    -69,    -89,    -85,    -58,    -15,     31,     70,     89,     85,     57,     14,    -32,    -70,    -89,    -84,    -56,    -13,     33,     71,     89,     84,     55,     12,    -34,    -72,    -90,    -83,    -55,    -11,     35,     72,     90,     83,     54,     10,    -36,    -73,    -90,    -82,    -53,     -9,     37,     74,     90,     82,     52,      8,    -38,    -74,    -90,    -81,    -51,     -7,     39,     75,     90,     81,     50,      6,    -40,    -76,    -90,    -81,    -49,     -4,     41,     76,     90,     80,     48,      3,    -42,    -77,    -90,    -79,    -47,     -2,     43,     77,     90,     79,     46,      1,    -44,    -78,    -90,    -78,    -45 },
      {     87,     61,     18,    -31,    -71,    -90,    -82,    -50,     -3,     44,     79,     90,     75,     37,    -11,    -56,    -85,    -89,    -66,    -24,     25,     67,     89,     85,     55,     10,    -38,    -76,    -90,    -78,    -43,      4,     51,     82,     90,     70,     30,    -19,    -62,    -87,    -87,    -60,    -16,     32,     72,     90,     81,     49,      2,    -45,    -79,    -90,    -74,    -36,     12,     57,     85,     88,     65,     23,    -26,    -67,    -89,    -84,    -55,     -9,     39,     76,     90,     78,     42,     -6,    -52,    -83,    -89,    -70,    -29,     20,     63,     88,     86,     60,     15,    -33,    -72,    -90,    -81,    -48,     -1,     46,     80,     90,     74,     35,    -13,    -58,    -86,    -88,    -64,    -22,     27,     68,     89,     84,     54,      8,    -40,    -77,    -90,    -77,    -41,      7,     53,     83,     89,     69,     28,    -21,    -64,    -88,    -86,    -59,    -14,     34,     73,     90,     81,     47 },
      {     87,     59,     12,    -38,    -77,    -90,    -75,    -35,     15,     61,     88,     86,     56,      9,    -41,    -78,    -90,    -73,    -32,     19,     64,     88,     85,     54,      6,    -44,    -80,    -90,    -71,    -29,     22,     66,     89,     83,     51,      2,    -47,    -81,    -90,    -69,    -26,     25,     68,     89,     82,     48,     -1,    -50,    -83,    -89,    -67,    -23,     28,     70,     90,     81,     45,     -4,    -53,    -84,    -89,    -64,    -20,     31,     72,     90,     79,     42,     -8,    -55,    -85,    -88,    -62,    -16,     34,     74,     90,     77,     39,    -11,    -58,    -86,    -87,    -60,    -13,     37,     76,     90,     76,     36,    -14,    -60,    -87,    -86,    -57,    -10,     40,     78,     90,     74,     33,    -18,    -63,    -88,    -85,    -55,     -7,     43,     79,     90,     72,     30,    -21,    -65,    -89,    -84,    -52,     -3,     46,     81,     90,     70,     27,    -24,    -67,    -89,    -82,    -49 },
      {     86,     56,      7,    -45,    -81,    -89,    -66,    -20,     33,     75,     90,     74,     32,    -21,    -67,    -89,    -81,    -44,      8,     57,     87,     86,     55,      6,    -46,    -82,    -89,    -65,    -19,     34,     76,     90,     74,     31,    -22,    -67,    -90,    -81,    -43,      9,     58,     87,     86,     55,      4,    -47,    -82,    -89,    -64,    -18,     35,     76,     90,     73,     30,    -23,    -68,    -90,    -80,    -42,     10,     59,     87,     85,     54,      3,    -48,    -83,    -89,    -64,    -16,     36,     77,     90,     72,     29,    -24,    -69,    -90,    -79,    -41,     11,     60,     88,     85,     53,      2,    -49,    -83,    -89,    -63,    -15,     37,     77,     90,     72,     28,    -25,    -70,    -90,    -79,    -40,     12,     60,     88,     85,     52,      1,    -50,    -84,    -88,    -62,    -14,     38,     78,     90,     71,     27,    -26,    -70,    -90,    -78,    -39,     13,     61,     88,     84,     51 },
      {     86,     54,      1,    -52,    -85,    -87,    -55,     -3,     50,     85,     87,     57,      6,    -48,    -84,    -88,    -59,     -8,     46,     83,     88,     60,     10,    -44,    -82,    -89,    -62,    -12,     42,     81,     89,     64,     14,    -40,    -80,    -89,    -65,    -16,     38,     79,     90,     67,     19,    -36,    -78,    -90,    -68,    -21,     34,     77,     90,     70,     23,    -32,    -76,    -90,    -71,    -25,     30,     74,     90,     72,     27,    -28,    -73,    -90,    -74,    -29,     26,     72,     90,     75,     31,    -24,    -70,    -90,    -76,    -33,     22,     69,     90,     77,     35,    -20,    -67,    -90,    -78,    -37,     18,     66,     90,     79,     39,    -15,    -64,    -89,    -81,    -41,     13,     63,     89,     81,     43,    -11,    -61,    -89,    -82,    -45,      9,     60,     88,     83,     47,     -7,    -58,    -88,    -84,    -49,      4,     56,     87,     85,     51,     -2,    -55,    -86,    -86,    -53 },
      {     86,     51,     -4,    -58,    -88,    -82,    -43,     13,     64,     90,     78,     35,    -22,    -70,    -90,    -74,    -27,     30,     76,     90,     68,     19,    -38,    -80,    -89,    -62,    -10,     46,     84,     87,     55,      1,    -54,    -87,    -85,    -48,      8,     60,     89,     81,     40,    -16,    -67,    -90,    -77,    -32,     25,     72,     90,     72,     24,    -33,    -77,    -90,    -66,    -15,     41,     81,     89,     60,      7,    -49,    -85,    -86,    -53,      2,     56,     88,     83,     45,    -11,    -63,    -89,    -79,    -37,     20,     69,     90,     75,     29,    -28,    -74,    -90,    -70,    -21,     36,     79,     89,     64,     12,    -44,    -83,    -88,    -57,     -3,     52,     86,     85,     50,     -6,    -59,    -88,    -82,    -42,     14,     65,     90,     78,     34,    -23,    -71,    -90,    -73,    -26,     31,     76,     90,     67,     18,    -39,    -81,    -89,    -61,     -9,     47,     84,     87,     55 },
      {     85,     48,    -10,    -64,    -90,    -77,    -30,     29,     76,     90,     64,     11,    -47,    -85,    -86,    -49,      9,     63,     90,     77,     31,    -28,    -76,    -90,    -65,    -12,     46,     85,     86,     50,     -8,    -62,    -89,    -78,    -32,     27,     75,     90,     66,     13,    -45,    -84,    -86,    -51,      7,     61,     89,     78,     33,    -26,    -74,    -90,    -67,    -14,     44,     84,     87,     52,     -6,    -60,    -89,    -79,    -34,     25,     74,     90,     67,     15,    -43,    -83,    -87,    -53,      4,     60,     89,     79,     35,    -24,    -73,    -90,    -68,    -16,     42,     83,     87,     54,     -3,    -59,    -89,    -80,    -36,     23,     72,     90,     69,     18,    -41,    -82,    -88,    -55,      2,     58,     89,     81,     37,    -22,    -72,    -90,    -70,    -19,     40,     82,     88,     55,     -1,    -57,    -88,    -81,    -38,     21,     71,     90,     70,     20,    -39,    -81,    -88,    -56 },
      {     85,     45,    -15,    -69,    -90,    -70,    -16,     44,     85,     85,     46,    -14,    -68,    -90,    -70,    -18,     43,     84,     86,     47,    -13,    -67,    -90,    -71,    -19,     42,     84,     86,     48,    -12,    -67,    -90,    -72,    -20,     41,     83,     86,     49,    -11,    -66,    -90,    -72,    -21,     40,     83,     87,     50,    -10,    -65,    -90,    -73,    -22,     39,     82,     87,     51,     -9,    -64,    -90,    -74,    -23,     38,     82,     87,     52,     -8,    -64,    -90,    -74,    -24,     37,     81,     88,     53,     -7,    -63,    -90,    -75,    -25,     36,     81,     88,     54,     -6,    -62,    -90,    -76,    -26,     35,     81,     88,     55,     -4,    -61,    -90,    -76,    -27,     34,     80,     88,     55,     -3,    -60,    -89,    -77,    -28,     33,     79,     89,     56,     -2,    -60,    -89,    -77,    -29,     32,     79,     89,     57,     -1,    -59,    -89,    -78,    -30,     31,     78,     89,     58 },
      {     85,     42,    -21,    -74,    -90,    -61,     -2,     58,     89,     76,     25,    -38,    -83,    -86,    -46,     16,     71,     90,     64,      7,    -55,    -89,    -78,    -29,     34,     81,     87,     50,    -12,    -68,    -90,    -67,    -11,     51,     88,     81,     33,    -30,    -79,    -88,    -54,      8,     65,     90,     70,     15,    -47,    -86,    -82,    -37,     26,     77,     89,     57,     -3,    -62,    -90,    -73,    -20,     43,     85,     84,     41,    -22,    -74,    -90,    -60,     -1,     59,     89,     76,     24,    -39,    -83,    -86,    -45,     18,     72,     90,     64,      6,    -55,    -89,    -78,    -28,     35,     81,     87,     49,    -13,    -69,    -90,    -67,    -10,     52,     88,     80,     32,    -31,    -79,    -88,    -53,      9,     66,     90,     70,     14,    -48,    -87,    -82,    -36,     27,     77,     89,     56,     -4,    -63,    -90,    -72,    -19,     44,     85,     84,     40,    -23,    -75,    -90,    -60 },
      {     84,     39,    -26,    -78,    -88,    -52,     12,     70,     90,     63,      2,    -60,    -90,    -72,    -16,     48,     87,     80,     30,    -35,    -82,    -86,    -43,     22,     76,     89,     55,     -8,    -67,    -90,    -66,     -7,     56,     89,     75,     21,    -44,    -86,    -82,    -34,     31,     81,     87,     47,    -18,    -73,    -90,    -59,      3,     64,     90,     69,     11,    -53,    -89,    -77,    -25,     40,     85,     84,     38,    -27,    -78,    -88,    -51,     13,     70,     90,     62,      1,    -60,    -90,    -72,    -15,     49,     88,     79,     29,    -36,    -83,    -85,    -42,     23,     76,     89,     55,     -9,    -67,    -90,    -65,     -6,     57,     89,     74,     20,    -45,    -86,    -81,    -33,     32,     81,     87,     46,    -19,    -74,    -90,    -58,      4,     64,     90,     68,     10,    -54,    -89,    -77,    -24,     41,     85,     83,     37,    -28,    -79,    -88,    -50,     14,     71,     90,     61 },
      {     84,     36,    -31,    -81,    -86,    -41,     26,     79,     87,     46,    -21,    -76,    -89,    -51,     15,     73,     89,     55,    -10,    -70,    -90,    -60,      4,     66,     90,     64,      1,    -62,    -90,    -67,     -7,     58,     90,     71,     12,    -54,    -89,    -74,    -18,     49,     88,     77,     23,    -44,    -87,    -80,    -28,     39,     85,     82,     33,    -34,    -83,    -85,    -38,     29,     81,     86,     43,    -24,    -78,    -88,    -48,     19,     75,     89,     53,    -13,    -72,    -90,    -57,      8,     68,     90,     61,     -2,    -64,    -90,    -65,     -3,     60,     90,     69,      9,    -56,    -90,    -72,    -14,     52,     89,     76,     20,    -47,    -88,    -78,    -25,     42,     86,     81,     30,    -37,    -84,    -83,    -35,     32,     82,     85,     40,    -27,    -79,    -87,    -45,     22,     77,     88,     50,    -16,    -74,    -89,    -55,     11,     70,     90,     59,     -6,    -67,    -90,    -63 },
      {     83,     33,    -36,    -85,    -82,    -30,     39,     86,     81,     27,    -42,    -87,    -79,    -24,     45,     88,     77,     21,    -48,    -88,    -76,    -18,     51,     89,     74,     14,    -54,    -89,    -72,    -11,     56,     90,     70,      8,    -59,    -90,    -67,     -4,     61,     90,     65,      1,    -64,    -90,    -63,      2,     66,     90,     60,     -6,    -68,    -90,    -58,      9,     70,     90,     55,    -12,    -72,    -89,    -53,     15,     74,     89,     50,    -19,    -76,    -88,    -47,     22,     78,     87,     44,    -25,    -79,    -86,    -41,     28,     81,     85,     38,    -31,    -82,    -84,    -35,     34,     84,     83,     32,    -37,    -85,    -81,    -29,     40,     86,     80,     26,    -43,    -87,    -78,    -23,     46,     88,     77,     20,    -49,    -89,    -75,    -16,     52,     89,     73,     13,    -55,    -90,    -71,    -10,     57,     90,     69,      7,    -60,    -90,    -67,     -3,     62,     90,     64 },
      {     83,     30,    -41,    -87,    -77,    -19,     52,     89,     70,      7,    -61,    -90,    -62,      6,     70,     90,     53,    -18,    -77,    -87,    -42,     29,     82,     83,     31,    -40,    -87,    -78,    -20,     51,     89,     71,      8,    -60,    -90,    -63,      4,     69,     90,     54,    -16,    -76,    -88,    -43,     28,     82,     84,     32,    -39,    -86,    -78,    -21,     50,     89,     72,      9,    -60,    -90,    -64,      3,     68,     90,     55,    -15,    -76,    -88,    -44,     27,     81,     84,     33,    -38,    -86,    -79,    -22,     49,     89,     72,     10,    -59,    -90,    -64,      2,     67,     90,     55,    -14,    -75,    -88,    -45,     26,     81,     85,     34,    -37,    -86,    -79,    -23,     48,     89,     73,     11,    -58,    -90,    -65,      1,     67,     90,     56,    -13,    -74,    -88,    -46,     25,     81,     85,     35,    -36,    -85,    -80,    -24,     47,     89,     74,     12,    -57,    -90,    -66 },
      {     82,     27,    -46,    -89,    -72,     -7,     63,     90,     57,    -14,    -76,    -87,    -39,     34,     85,     79,     20,    -53,    -90,    -67,      1,     68,     90,     51,    -22,    -80,    -85,    -32,     41,     88,     75,     12,    -59,    -90,    -61,      9,     73,     88,     44,    -29,    -83,    -81,    -25,     48,     89,     70,      4,    -64,    -90,    -55,     16,     77,     86,     37,    -36,    -86,    -78,    -18,     55,     90,     65,     -3,    -70,    -89,    -49,     24,     81,     84,     30,    -43,    -88,    -74,    -10,     60,     90,     60,    -11,    -74,    -88,    -42,     31,     84,     81,     23,    -50,    -89,    -69,     -2,     66,     90,     54,    -19,    -78,    -86,    -35,     38,     87,     77,     15,    -56,    -90,    -64,      6,     71,     89,     47,    -26,    -82,    -83,    -28,     45,     89,     72,      8,    -62,    -90,    -58,     13,     76,     87,     40,    -33,    -85,    -79,    -21,     52,     90,     67 },
      {     82,     24,    -51,    -90,    -65,      6,     72,     88,     41,    -34,    -86,    -77,    -13,     60,     90,     57,    -16,    -78,    -85,    -31,     44,     89,     70,      2,    -67,    -89,    -48,     27,     83,     81,     21,    -54,    -90,    -63,      9,     74,     87,     38,    -37,    -87,    -75,    -10,     62,     90,     55,    -20,    -80,    -84,    -28,     47,     89,     68,     -1,    -70,    -89,    -45,     30,     85,     79,     18,    -56,    -90,    -60,     12,     76,     86,     35,    -40,    -88,    -73,     -7,     64,     90,     52,    -23,    -81,    -82,    -25,     50,     90,     66,     -4,    -72,    -88,    -42,     33,     86,     77,     14,    -59,    -90,    -58,     15,     78,     85,     32,    -43,    -89,    -71,     -3,     67,     90,     49,    -26,    -83,    -81,    -22,     53,     90,     64,     -8,    -74,    -88,    -39,     36,     87,     76,     11,    -61,    -90,    -55,     19,     79,     84,     29,    -46,    -89,    -69 },
      {     81,     21,    -55,    -90,    -58,     18,     80,     83,     24,    -53,    -90,    -60,     14,     78,     84,     27,    -50,    -90,    -63,     11,     77,     85,     30,    -47,    -90,    -65,      8,     75,     86,     33,    -44,    -89,    -67,      4,     73,     87,     36,    -41,    -89,    -70,      1,     71,     88,     39,    -38,    -88,    -72,     -2,     69,     89,     42,    -35,    -87,    -74,     -6,     67,     89,     45,    -32,    -86,    -76,     -9,     64,     90,     48,    -29,    -85,    -77,    -12,     62,     90,     51,    -26,    -84,    -79,    -15,     60,     90,     54,    -23,    -82,    -81,    -19,     57,     90,     56,    -20,    -81,    -82,    -22,     55,     90,     59,    -16,    -79,    -83,    -25,     52,     90,     61,    -13,    -78,    -85,    -28,     49,     90,     64,    -10,    -76,    -86,    -31,     46,     89,     66,     -7,    -74,    -87,    -34,     43,     89,     68,     -3,    -72,    -88,    -37,     40,     88,     70 },
      {     81,     18,    -60,    -90,    -50,     29,     86,     75,      6,    -68,    -89,    -39,     40,     89,     67,     -7,    -76,    -85,    -28,     51,     90,     59,    -19,    -81,    -81,    -16,     60,     90,     49,    -30,    -86,    -74,     -4,     69,     88,     38,    -41,    -89,    -67,      8,     76,     85,     27,    -52,    -90,    -58,     20,     82,     80,     15,    -61,    -90,    -48,     31,     86,     74,      3,    -70,    -88,    -37,     42,     89,     66,     -9,    -77,    -85,    -26,     53,     90,     57,    -21,    -82,    -79,    -14,     62,     90,     47,    -32,    -87,    -73,     -2,     70,     88,     36,    -43,    -89,    -65,     10,     77,     84,     25,    -54,    -90,    -56,     22,     83,     79,     13,    -63,    -90,    -46,     33,     87,     72,      1,    -71,    -88,    -35,     44,     89,     64,    -11,    -78,    -84,    -24,     55,     90,     55,    -23,    -83,    -78,    -12,     64,     90,     45,    -34,    -87,    -72 },
      {     81,     14,    -64,    -89,    -41,     40,     89,     64,    -13,    -80,    -81,    -15,     63,     89,     42,    -39,    -89,    -65,     12,     79,     81,     16,    -62,    -90,    -43,     38,     89,     66,    -11,    -79,    -82,    -18,     61,     90,     44,    -37,    -89,    -67,     10,     78,     82,     19,    -60,    -90,    -45,     36,     88,     67,     -9,    -78,    -83,    -20,     60,     90,     46,    -35,    -88,    -68,      8,     77,     83,     21,    -59,    -90,    -47,     34,     88,     69,     -7,    -77,    -84,    -22,     58,     90,     48,    -33,    -88,    -70,      6,     76,     84,     23,    -57,    -90,    -49,     32,     87,     70,     -4,    -76,    -85,    -24,     56,     90,     50,    -31,    -87,    -71,      3,     75,     85,     25,    -55,    -90,    -51,     30,     87,     72,     -2,    -74,    -85,    -26,     55,     90,     52,    -29,    -86,    -72,      1,     74,     86,     27,    -54,    -90,    -53,     28,     86,     73 },
      {     80,     11,    -67,    -88,    -32,     51,     90,     52,    -31,    -88,    -68,     10,     79,     81,     12,    -67,    -88,    -33,     50,     90,     53,    -30,    -87,    -69,      9,     79,     81,     13,    -66,    -88,    -34,     49,     90,     54,    -29,    -87,    -70,      8,     78,     81,     14,    -65,    -89,    -35,     48,     90,     55,    -28,    -87,    -70,      7,     78,     82,     15,    -64,    -89,    -36,     47,     90,     55,    -27,    -86,    -71,      6,     77,     82,     16,    -64,    -89,    -37,     46,     90,     56,    -26,    -86,    -72,      4,     77,     83,     18,    -63,    -89,    -38,     45,     90,     57,    -25,    -86,    -72,      3,     76,     83,     19,    -62,    -89,    -39,     44,     90,     58,    -24,    -85,    -73,      2,     76,     84,     20,    -61,    -89,    -40,     43,     90,     59,    -23,    -85,    -74,      1,     75,     84,     21,    -60,    -90,    -41,     42,     90,     60,    -22,    -85,    -74 },
      {     79,      8,    -71,    -86,    -23,     60,     89,     37,    -48,    -90,    -51,     34,     89,     63,    -20,    -85,    -73,      4,     78,     81,     11,    -69,    -87,    -26,     58,     90,     40,    -45,    -90,    -54,     31,     88,     65,    -16,    -83,    -75,      1,     76,     82,     14,    -67,    -88,    -29,     55,     90,     43,    -42,    -90,    -56,     28,     87,     67,    -13,    -82,    -77,     -2,     74,     84,     18,    -64,    -88,    -32,     53,     90,     46,    -39,    -90,    -59,     25,     86,     70,    -10,    -81,    -78,     -6,     72,     85,     21,    -62,    -89,    -35,     50,     90,     49,    -36,    -89,    -61,     22,     85,     72,     -7,    -79,    -80,     -9,     70,     86,     24,    -60,    -89,    -38,     47,     90,     52,    -33,    -89,    -64,     19,     84,     74,     -3,    -77,    -81,    -12,     68,     87,     27,    -57,    -90,    -41,     44,     90,     55,    -30,    -88,    -66,     15,     83,     76 },
      {     79,      4,    -74,    -83,    -13,     69,     86,     22,    -63,    -88,    -30,     56,     90,     38,    -49,    -90,    -46,     41,     90,     54,    -33,    -89,    -60,     25,     87,     67,    -16,    -84,    -72,      8,     81,     77,      1,    -76,    -81,    -10,     71,     85,     19,    -65,    -88,    -27,     59,     89,     35,    -52,    -90,    -43,     44,     90,     51,    -36,    -89,    -58,     28,     88,     64,    -20,    -85,    -70,     11,     82,     76,     -2,    -78,    -80,     -7,     73,     84,     15,    -67,    -87,    -24,     61,     89,     32,    -55,    -90,    -40,     47,     90,     48,    -39,    -90,    -55,     31,     89,     62,    -23,    -86,    -68,     14,     83,     74,     -6,    -79,    -78,     -3,     75,     82,     12,    -70,    -86,    -21,     64,     88,     29,    -57,    -90,    -37,     50,     90,     45,    -42,    -90,    -53,     34,     89,     60,    -26,    -87,    -66,     18,     85,     72,     -9,    -81,    -77 },
      {     78,      1,    -77,    -79,     -3,     76,     81,      6,    -75,    -81,     -8,     74,     82,     10,    -72,    -83,    -12,     71,     84,     14,    -70,    -85,    -16,     68,     86,     19,    -67,    -86,    -21,     65,     87,     23,    -64,    -88,    -25,     62,     88,     27,    -60,    -89,    -29,     59,     89,     31,    -57,    -89,    -33,     55,     90,     35,    -54,    -90,    -37,     52,     90,     39,    -50,    -90,    -41,     48,     90,     43,    -46,    -90,    -45,     44,     90,     47,    -42,    -90,    -49,     40,     90,     51,    -38,    -90,    -53,     36,     90,     55,    -34,    -89,    -56,     32,     89,     58,    -30,    -89,    -60,     28,     88,     61,    -26,    -88,    -63,     24,     87,     64,    -22,    -87,    -66,     20,     86,     67,    -18,    -85,    -69,     15,     85,     70,    -13,    -84,    -72,     11,     83,     73,     -9,    -82,    -74,      7,     81,     76,     -4,    -80,    -77,      2,     79,     78 },
      {     78,     -2,    -80,    -76,      7,     82,     73,    -11,    -84,    -70,     15,     85,     67,    -20,    -87,    -64,     24,     88,     61,    -28,    -89,    -58,     32,     89,     55,    -36,    -90,    -51,     40,     90,     47,    -44,    -90,    -43,     48,     90,     39,    -52,    -90,    -35,     55,     89,     31,    -59,    -89,    -27,     62,     88,     23,    -65,    -86,    -19,     68,     85,     14,    -71,    -83,    -10,     74,     81,      6,    -76,    -79,     -1,     78,     77,     -3,    -81,    -75,      8,     82,     72,    -12,    -84,    -70,     16,     86,     67,    -21,    -87,    -64,     25,     88,     60,    -29,    -89,    -57,     33,     90,     54,    -37,    -90,    -50,     41,     90,     46,    -45,    -90,    -42,     49,     90,     38,    -53,    -90,    -34,     56,     89,     30,    -60,    -88,    -26,     63,     87,     22,    -66,    -86,    -18,     69,     85,     13,    -72,    -83,     -9,     74,     81,      4,    -77,    -79 },
      {     77,     -6,    -82,    -71,     16,     86,     64,    -27,    -89,    -55,     37,     90,     46,    -47,    -90,    -36,     56,     89,     26,    -64,    -86,    -15,     72,     82,      4,    -78,    -77,      7,     83,     70,    -18,    -87,    -63,     28,     89,     55,    -38,    -90,    -45,     48,     90,     35,    -57,    -89,    -25,     65,     86,     14,    -72,    -81,     -3,     78,     76,     -8,    -83,    -70,     19,     87,     62,    -29,    -89,    -54,     39,     90,     44,    -49,    -90,    -34,     58,     88,     24,    -66,    -85,    -13,     73,     81,      2,    -79,    -76,      9,     84,     69,    -20,    -87,    -61,     30,     89,     53,    -40,    -90,    -43,     50,     90,     33,    -59,    -88,    -23,     67,     85,     12,    -74,    -81,     -1,     79,     75,    -10,    -84,    -68,     21,     88,     60,    -31,    -90,    -52,     41,     90,     42,    -51,    -90,    -32,     60,     88,     22,    -67,    -85,    -11,     74,     80 },
      {     77,     -9,    -85,    -66,     26,     89,     53,    -42,    -90,    -37,     57,     88,     21,    -70,    -82,     -3,     79,     74,    -14,    -86,    -62,     31,     90,     48,    -47,    -90,    -32,     61,     87,     15,    -73,    -80,      2,     82,     70,    -20,    -88,    -58,     36,     90,     43,    -52,    -89,    -27,     65,     85,     10,    -76,    -77,      8,     84,     67,    -25,    -89,    -54,     41,     90,     38,    -56,    -88,    -22,     69,     83,      4,    -79,    -74,     13,     86,     63,    -30,    -90,    -49,     46,     90,     33,    -60,    -87,    -16,     72,     81,     -1,    -81,    -71,     19,     88,     59,    -35,    -90,    -44,     51,     89,     28,    -64,    -85,    -11,     76,     78,     -7,    -84,    -67,     24,     89,     55,    -40,    -90,    -39,     55,     89,     23,    -68,    -83,     -6,     78,     75,    -12,    -86,    -64,     29,     90,     50,    -45,    -90,    -34,     60,     87,     18,    -72,    -81 },
      {     76,    -12,    -86,    -60,     35,     90,     40,    -56,    -88,    -18,     73,     79,     -7,    -85,    -64,     30,     90,     45,    -52,    -89,    -23,     70,     81,     -1,    -82,    -68,     25,     89,     50,    -47,    -90,    -28,     66,     84,      4,    -80,    -72,     20,     88,     55,    -42,    -90,    -33,     62,     86,     10,    -77,    -75,     14,     87,     59,    -37,    -90,    -38,     58,     87,     15,    -74,    -78,      9,     85,     63,    -32,    -90,    -43,     54,     89,     21,    -71,    -81,      3,     83,     67,    -27,    -90,    -48,     49,     89,     26,    -67,    -83,     -2,     81,     70,    -22,    -89,    -53,     44,     90,     31,    -64,    -85,     -8,     78,     74,    -16,    -88,    -57,     39,     90,     36,    -60,    -87,    -13,     76,     77,    -11,    -86,    -61,     34,     90,     41,    -55,    -88,    -19,     72,     79,     -6,    -84,    -65,     29,     90,     46,    -51,    -89,    -24,     69,     82 },
      {     76,    -15,    -88,    -55,     44,     90,     27,    -68,    -81,      3,     84,     64,    -33,    -90,    -38,     60,     86,      9,    -79,    -72,     22,     89,     49,    -50,    -89,    -21,     72,     78,    -10,    -86,    -59,     39,     90,     32,    -64,    -84,     -2,     82,     67,    -28,    -90,    -43,     55,     88,     14,    -76,    -75,     16,     88,     54,    -45,    -90,    -26,     69,     81,     -4,    -85,    -63,     34,     90,     37,    -60,    -86,     -8,     79,     71,    -23,    -89,    -48,     51,     89,     20,    -73,    -78,     11,     87,     58,    -40,    -90,    -31,     65,     83,      1,    -82,    -67,     29,     90,     42,    -56,    -87,    -13,     77,     74,    -18,    -88,    -53,     46,     90,     25,    -70,    -81,      6,     85,     62,    -35,    -90,    -36,     61,     85,      7,    -80,    -70,     24,     89,     47,    -52,    -89,    -19,     74,     77,    -12,    -87,    -57,     41,     90,     30,    -66,    -83 },
      {     75,    -19,    -89,    -48,     53,     88,     13,    -78,    -72,     24,     90,     43,    -57,    -86,     -8,     81,     68,    -29,    -90,    -38,     61,     85,      2,    -83,    -64,     34,     90,     33,    -65,    -82,      3,     85,     60,    -39,    -90,    -28,     69,     80,     -9,    -87,    -56,     44,     90,     23,    -72,    -77,     14,     88,     52,    -49,    -89,    -18,     76,     74,    -20,    -89,    -47,     54,     88,     12,    -78,    -71,     25,     90,     42,    -58,    -86,     -7,     81,     67,    -30,    -90,    -37,     62,     84,      1,    -83,    -64,     35,     90,     32,    -66,    -82,      4,     85,     60,    -40,    -90,    -27,     70,     79,    -10,    -87,    -55,     45,     89,     22,    -73,    -77,     15,     88,     51,    -50,    -89,    -16,     76,     74,    -21,    -89,    -46,     55,     87,     11,    -79,    -70,     26,     90,     41,    -59,    -86,     -6,     81,     67,    -31,    -90,    -36,     63,     84 },
      {     74,    -22,    -90,    -41,     60,     84,     -1,    -85,    -59,     43,     89,     20,    -76,    -73,     24,     90,     39,    -62,    -83,      3,     86,     57,    -45,    -89,    -18,     77,     72,    -26,    -90,    -37,     64,     82,     -6,    -86,    -55,     47,     89,     15,    -78,    -70,     28,     90,     35,    -65,    -81,      8,     87,     54,    -49,    -88,    -13,     79,     69,    -30,    -90,    -33,     67,     81,    -10,    -88,    -52,     51,     88,     11,    -80,    -67,     32,     90,     31,    -68,    -79,     12,     88,     50,    -53,    -87,     -9,     81,     66,    -34,    -90,    -29,     70,     78,    -14,    -89,    -48,     55,     87,      7,    -82,    -64,     36,     90,     27,    -71,    -77,     16,     89,     46,    -56,    -86,     -4,     83,     63,    -38,    -90,    -25,     72,     76,    -19,    -89,    -44,     58,     85,      2,    -84,    -61,     40,     90,     23,    -74,    -75,     21,     90,     42,    -60,    -85 },
      {     74,    -25,    -90,    -34,     67,     79,    -15,    -89,    -43,     60,     83,     -6,    -87,    -52,     53,     87,      4,    -84,    -60,     44,     89,     14,    -79,    -67,     35,     90,     24,    -74,    -73,     26,     90,     33,    -68,    -78,     16,     89,     42,    -61,    -83,      7,     87,     51,    -54,    -86,     -3,     84,     59,    -45,    -89,    -13,     80,     66,    -36,    -90,    -23,     75,     72,    -27,    -90,    -32,     69,     78,    -18,    -89,    -41,     62,     82,     -8,    -88,    -50,     55,     86,      2,    -85,    -58,     46,     89,     12,    -81,    -65,     37,     90,     22,    -76,    -72,     28,     90,     31,    -70,    -77,     19,     90,     40,    -63,    -82,      9,     88,     49,    -55,    -86,     -1,     85,     57,    -47,    -88,    -11,     81,     64,    -38,    -90,    -21,     76,     71,    -29,    -90,    -30,     70,     77,    -20,    -90,    -39,     64,     81,    -10,    -88,    -48,     56,     85 },
      {     73,    -28,    -90,    -27,     74,     72,    -29,    -90,    -26,     74,     72,    -30,    -90,    -25,     75,     71,    -31,    -90,    -24,     76,     70,    -32,    -90,    -23,     76,     70,    -33,    -90,    -22,     77,     69,    -34,    -90,    -21,     77,     68,    -35,    -90,    -20,     78,     67,    -36,    -90,    -19,     78,     67,    -37,    -90,    -18,     79,     66,    -38,    -90,    -16,     79,     65,    -39,    -89,    -15,     80,     64,    -40,    -89,    -14,     81,     64,    -41,    -89,    -13,     81,     63,    -42,    -89,    -12,     81,     62,    -43,    -89,    -11,     82,     61,    -44,    -89,    -10,     82,     60,    -45,    -88,     -9,     83,     60,    -46,    -88,     -8,     83,     59,    -47,    -88,     -7,     84,     58,    -48,    -88,     -6,     84,     57,    -49,    -87,     -4,     85,     56,    -50,    -87,     -3,     85,     55,    -51,    -87,     -2,     85,     55,    -52,    -86,     -1,     86,     54,    -53,    -86 },
      {     72,    -31,    -90,    -20,     79,     64,    -42,    -89,     -8,     84,     55,    -53,    -85,      4,     88,     45,    -62,    -81,     16,     90,     34,    -70,    -74,     28,     90,     23,    -77,    -67,     39,     89,     11,    -83,    -58,     50,     86,     -1,    -87,    -48,     60,     82,    -13,    -89,    -37,     68,     76,    -25,    -90,    -26,     76,     69,    -36,    -90,    -14,     81,     60,    -47,    -87,     -2,     86,     51,    -57,    -83,     10,     89,     40,    -66,    -78,     22,     90,     29,    -74,    -71,     33,     90,     18,    -80,    -63,     44,     88,      6,    -85,    -54,     55,     85,     -7,    -88,    -43,     64,     79,    -19,    -90,    -32,     72,     73,    -30,    -90,    -21,     78,     65,    -41,    -89,     -9,     84,     56,    -52,    -86,      3,     88,     46,    -61,    -81,     15,     90,     35,    -70,    -75,     27,     90,     24,    -77,    -67,     38,     89,     12,    -82,    -59,     49,     87 },
      {     72,    -34,    -90,    -12,     83,     55,    -55,    -84,     11,     89,     35,    -71,    -72,     33,     90,     13,    -83,    -56,     54,     84,    -10,    -89,    -36,     70,     73,    -32,    -90,    -14,     82,     57,    -53,    -85,      9,     89,     37,    -70,    -74,     31,     90,     15,    -82,    -58,     52,     85,     -8,    -89,    -38,     69,     74,    -30,    -90,    -16,     81,     59,    -51,    -85,      7,     89,     39,    -68,    -75,     29,     90,     18,    -81,    -60,     50,     86,     -6,    -89,    -40,     67,     76,    -28,    -90,    -19,     81,     60,    -49,    -86,      4,     88,     41,    -67,    -76,     27,     90,     20,    -80,    -61,     48,     86,     -3,    -88,    -42,     66,     77,    -26,    -90,    -21,     79,     62,    -47,    -87,      2,     88,     43,    -65,    -77,     25,     90,     22,    -79,    -63,     46,     87,     -1,    -88,    -44,     64,     78,    -24,    -90,    -23,     78,     64,    -45,    -87 },
      {     71,    -37,    -89,     -4,     87,     45,    -65,    -76,     29,     90,     13,    -84,    -53,     59,     81,    -21,    -90,    -22,     80,     60,    -52,    -84,     12,     90,     30,    -76,    -66,     44,     87,     -3,    -89,    -38,     70,     72,    -36,    -89,     -6,     86,     46,    -64,    -77,     28,     90,     14,    -83,    -54,     58,     81,    -20,    -90,    -23,     79,     60,    -51,    -85,     11,     90,     31,    -75,    -67,     43,     87,     -2,    -88,    -39,     70,     72,    -35,    -89,     -7,     86,     47,    -64,    -77,     27,     90,     15,    -83,    -55,     57,     81,    -19,    -90,    -24,     79,     61,    -50,    -85,     10,     90,     32,    -74,    -67,     42,     88,     -1,    -88,    -40,     69,     73,    -34,    -89,     -8,     86,     48,    -63,    -78,     26,     90,     16,    -82,    -55,     56,     82,    -18,    -90,    -25,     78,     62,    -49,    -85,      9,     89,     33,    -74,    -68,     41,     88 },
      {     70,    -40,    -88,      3,     89,     34,    -74,    -66,     46,     86,    -10,    -90,    -28,     78,     61,    -52,    -83,     16,     90,     22,    -81,    -56,     57,     81,    -23,    -90,    -15,     84,     51,    -62,    -77,     29,     90,      9,    -86,    -45,     67,     74,    -35,    -89,     -2,     88,     39,    -71,    -70,     41,     87,     -4,    -89,    -33,     75,     65,    -47,    -85,     11,     90,     27,    -78,    -60,     53,     83,    -18,    -90,    -21,     81,     55,    -58,    -80,     24,     90,     14,    -84,    -50,     63,     77,    -30,    -90,     -8,     86,     44,    -67,    -73,     36,     89,      1,    -88,    -38,     72,     69,    -42,    -87,      6,     89,     32,    -76,    -64,     48,     85,    -12,    -90,    -26,     79,     60,    -54,    -82,     19,     90,     20,    -82,    -55,     59,     79,    -25,    -90,    -13,     85,     49,    -64,    -76,     31,     89,      7,    -87,    -43,     68,     72,    -37,    -88 },
      {     70,    -43,    -86,     11,     90,     23,    -81,    -54,     61,     77,    -32,    -89,     -1,     89,     34,    -76,    -63,     52,     82,    -21,    -90,    -13,     85,     45,    -68,    -71,     41,     87,     -9,    -90,    -25,     81,     55,    -60,    -78,     30,     89,      3,    -88,    -36,     74,     64,    -50,    -83,     19,     90,     15,    -85,    -47,     67,     72,    -39,    -87,      7,     90,     27,    -79,    -57,     58,     79,    -28,    -90,     -6,     88,     38,    -73,    -66,     48,     84,    -16,    -90,    -18,     84,     49,    -65,    -74,     37,     88,     -4,    -89,    -29,     78,     59,    -56,    -80,     26,     90,      8,    -87,    -40,     72,     67,    -46,    -85,     14,     90,     20,    -83,    -51,     64,     75,    -35,    -88,      2,     89,     31,    -77,    -60,     55,     81,    -24,    -90,    -10,     86,     42,    -70,    -69,     44,     86,    -12,    -90,    -22,     82,     53,    -62,    -76,     33,     89 },
      {     69,    -46,    -84,     19,     90,     11,    -87,    -39,     74,     64,    -53,    -81,     26,     90,      3,    -89,    -32,     78,     58,    -59,    -77,     33,     88,     -4,    -90,    -25,     81,     52,    -64,    -73,     40,     86,    -12,    -90,    -18,     85,     45,    -70,    -68,     47,     84,    -20,    -90,    -10,     87,     38,    -74,    -63,     54,     81,    -27,    -89,     -2,     89,     31,    -78,    -57,     60,     77,    -34,    -88,      6,     90,     24,    -82,    -51,     65,     72,    -41,    -86,     13,     90,     16,    -85,    -44,     70,     67,    -48,    -83,     21,     90,      9,    -87,    -37,     75,     62,    -55,    -80,     28,     89,      1,    -89,    -30,     79,     56,    -60,    -76,     35,     88,     -7,    -90,    -23,     82,     50,    -66,    -72,     42,     86,    -14,    -90,    -15,     85,     43,    -71,    -67,     49,     83,    -22,    -90,     -8,     88,     36,    -76,    -61,     55,     79,    -29,    -89 },
      {     68,    -49,    -82,     26,     89,     -1,    -90,    -24,     83,     47,    -70,    -67,     51,     81,    -28,    -89,      3,     90,     22,    -84,    -45,     71,     65,    -53,    -80,     30,     89,     -6,    -90,    -20,     85,     43,    -72,    -64,     55,     79,    -32,    -88,      8,     90,     18,    -85,    -41,     74,     62,    -56,    -78,     34,     88,    -10,    -90,    -15,     86,     39,    -75,    -60,     58,     77,    -36,    -87,     12,     90,     13,    -87,    -37,     76,     59,    -60,    -76,     38,     86,    -14,    -90,    -11,     87,     35,    -77,    -57,     61,     74,    -40,    -86,     16,     90,      9,    -88,    -33,     78,     55,    -63,    -73,     42,     85,    -19,    -90,     -7,     88,     31,    -79,    -54,     64,     72,    -44,    -84,     21,     90,      4,    -89,    -29,     81,     52,    -66,    -70,     46,     83,    -23,    -90,     -2,     89,     27,    -81,    -50,     67,     69,    -48,    -82,     25,     89 },
      {     67,    -52,    -79,     33,     87,    -13,    -90,     -8,     89,     28,    -82,    -47,     71,     64,    -56,    -77,     38,     86,    -19,    -90,     -2,     89,     23,    -84,    -42,     74,     60,    -60,    -74,     43,     84,    -24,    -89,      3,     90,     18,    -86,    -37,     77,     55,    -64,    -70,     48,     81,    -29,    -88,      9,     90,     12,    -88,    -32,     80,     51,    -68,    -67,     53,     79,    -34,    -87,     14,     90,      7,    -89,    -27,     82,     46,    -72,    -63,     57,     76,    -39,    -85,     20,     90,      1,    -90,    -22,     85,     41,    -75,    -59,     61,     73,    -44,    -83,     25,     89,     -4,    -90,    -16,     86,     36,    -78,    -55,     65,     70,    -49,    -81,     30,     88,    -10,    -90,    -11,     88,     31,    -81,    -50,     69,     66,    -54,    -78,     35,     87,    -15,    -90,     -6,     89,     26,    -83,    -45,     72,     62,    -58,    -76,     40,     85,    -21,    -90 },
      {     67,    -55,    -77,     40,     84,    -25,    -89,      9,     90,      8,    -89,    -24,     85,     39,    -77,    -54,     67,     66,    -55,    -76,     41,     84,    -26,    -89,     10,     90,      7,    -89,    -23,     85,     38,    -78,    -53,     68,     65,    -56,    -76,     42,     83,    -27,    -88,     11,     90,      6,    -89,    -22,     85,     37,    -78,    -52,     69,     64,    -57,    -75,     43,     83,    -28,    -88,     12,     90,      4,    -89,    -21,     86,     36,    -79,    -51,     70,     64,    -58,    -74,     44,     82,    -29,    -88,     13,     90,      3,    -90,    -20,     86,     35,    -79,    -50,     70,     63,    -59,    -74,     45,     82,    -30,    -88,     14,     90,      2,    -90,    -19,     86,     34,    -80,    -49,     71,     62,    -60,    -73,     46,     81,    -31,    -87,     15,     90,      1,    -90,    -18,     87,     33,    -81,    -48,     72,     61,    -60,    -72,     47,     81,    -32,    -87,     16,     90 },
      {     66,    -57,    -74,     47,     80,    -36,    -85,     25,     88,    -13,    -90,      1,     90,     11,    -89,    -23,     86,     34,    -81,    -45,     75,     55,    -67,    -64,     59,     72,    -49,    -79,     38,     84,    -27,    -88,     15,     90,     -3,    -90,     -9,     89,     21,    -86,    -32,     82,     43,    -76,    -54,     69,     63,    -60,    -71,     51,     78,    -40,    -83,     29,     87,    -18,    -90,      6,     90,      7,    -89,    -19,     87,     30,    -83,    -41,     77,     52,    -70,    -61,     62,     70,    -53,    -77,     42,     82,    -31,    -87,     20,     89,     -8,    -90,     -4,     90,     16,    -88,    -28,     84,     39,    -78,    -50,     72,     60,    -64,    -68,     55,     76,    -44,    -81,     33,     86,    -22,    -89,     10,     90,      2,    -90,    -14,     88,     26,    -85,    -37,     79,     48,    -73,    -58,     65,     67,    -56,    -74,     46,     81,    -35,    -85,     24,     89,    -12,    -90 },
      {     65,    -60,    -70,     54,     75,    -47,    -79,     40,     82,    -33,    -85,     26,     88,    -19,    -89,     11,     90,     -3,    -90,     -4,     90,     12,    -89,    -20,     87,     27,    -85,    -34,     82,     41,    -78,    -48,     74,     55,    -70,    -60,     64,     66,    -59,    -71,     53,     76,    -46,    -79,     39,     83,    -32,    -86,     25,     88,    -18,    -89,     10,     90,     -2,    -90,     -6,     90,     13,    -89,    -21,     87,     28,    -85,    -35,     81,     42,    -78,    -49,     74,     55,    -69,    -61,     64,     67,    -58,    -72,     52,     76,    -45,    -80,     38,     83,    -31,    -86,     24,     88,    -16,    -89,      9,     90,     -1,    -90,     -7,     90,     14,    -89,    -22,     87,     29,    -84,    -36,     81,     43,    -77,    -50,     73,     56,    -68,    -62,     63,     67,    -57,    -72,     51,     77,    -44,    -81,     37,     84,    -30,    -86,     23,     88,    -15,    -90,      8,     90 },
      {     64,    -62,    -67,     60,     69,    -57,    -71,     55,     73,    -52,    -75,     49,     77,    -46,    -78,     43,     80,    -40,    -81,     37,     83,    -34,    -84,     31,     85,    -28,    -86,     25,     87,    -22,    -88,     19,     89,    -15,    -89,     12,     90,     -9,    -90,      6,     90,     -2,    -90,     -1,     90,      4,    -90,     -8,     90,     11,    -89,    -14,     89,     18,    -88,    -21,     88,     24,    -87,    -27,     86,     30,    -85,    -33,     83,     36,    -82,    -39,     81,     42,    -79,    -45,     77,     48,    -76,    -51,     74,     54,    -72,    -56,     70,     59,    -67,    -61,     65,     64,    -63,    -66,     60,     68,    -58,    -70,     55,     72,    -53,    -74,     50,     76,    -47,    -78,     44,     79,    -41,    -81,     38,     82,    -35,    -84,     32,     85,    -29,    -86,     26,     87,    -23,    -88,     20,     89,    -16,    -89,     13,     90,    -10,    -90,      7,     90,     -3,    -90 },
      {     64,    -64,    -63,     65,     62,    -66,    -61,     67,     60,    -67,    -60,     68,     59,    -69,    -58,     70,     57,    -70,    -56,     71,     55,    -72,    -55,     72,     54,    -73,    -53,     74,     52,    -74,    -51,     75,     50,    -76,    -49,     76,     48,    -77,    -47,     77,     46,    -78,    -45,     78,     44,    -79,    -43,     79,     42,    -80,    -41,     81,     40,    -81,    -39,     81,     38,    -82,    -37,     82,     36,    -83,    -35,     83,     34,    -84,    -33,     84,     32,    -85,    -31,     85,     30,    -85,    -29,     86,     28,    -86,    -27,     86,     26,    -87,    -25,     87,     24,    -87,    -23,     88,     22,    -88,    -21,     88,     20,    -88,    -19,     89,     18,    -89,    -16,     89,     15,    -89,    -14,     89,     13,    -89,    -12,     90,     11,    -90,    -10,     90,      9,    -90,     -8,     90,      7,    -90,     -6,     90,      4,    -90,     -3,     90,      2,    -90,     -1,     90 },
      {     63,    -67,    -59,     70,     55,    -74,    -50,     77,     45,    -79,    -40,     82,     35,    -84,    -30,     86,     25,    -88,    -20,     89,     14,    -90,     -9,     90,      3,    -90,      2,     90,     -8,    -90,     13,     89,    -19,    -88,     24,     86,    -29,    -85,     34,     82,    -39,    -80,     44,     77,    -49,    -74,     54,     71,    -58,    -67,     62,     64,    -66,    -60,     70,     55,    -73,    -51,     76,     46,    -79,    -41,     81,     36,    -84,    -31,     86,     26,    -87,    -21,     89,     15,    -89,    -10,     90,      4,    -90,      1,     90,     -7,    -90,     12,     89,    -18,    -88,     23,     87,    -28,    -85,     33,     83,    -38,    -81,     43,     78,    -48,    -75,     53,     72,    -57,    -68,     61,     64,    -65,    -60,     69,     56,    -72,    -52,     76,     47,    -78,    -42,     81,     37,    -83,    -32,     85,     27,    -87,    -22,     88,     16,    -89,    -11,     90,      6,    -90 },
      {     62,    -69,    -55,     75,     46,    -80,    -37,     84,     28,    -87,    -19,     89,      9,    -90,      1,     90,    -11,    -89,     21,     87,    -30,    -83,     39,     79,    -48,    -74,     56,     67,    -64,    -60,     70,     53,    -76,    -44,     81,     35,    -85,    -26,     88,     16,    -90,     -7,     90,     -3,    -90,     13,     89,    -23,    -86,     32,     82,    -41,    -78,     50,     72,    -58,    -66,     65,     59,    -72,    -51,     77,     42,    -82,    -33,     86,     24,    -88,    -14,     90,      4,    -90,      6,     90,    -15,    -88,     25,     85,    -34,    -81,     43,     77,    -52,    -71,     60,     64,    -67,    -57,     73,     49,    -78,    -40,     83,     31,    -86,    -22,     89,     12,    -90,     -2,     90,     -8,    -89,     18,     88,    -27,    -85,     36,     81,    -45,    -76,     54,     70,    -61,    -63,     68,     55,    -74,    -47,     79,     38,    -84,    -29,     87,     20,    -89,    -10,     90 },
      {     61,    -71,    -50,     79,     37,    -85,    -24,     89,     10,    -90,      4,     90,    -19,    -87,     32,     81,    -45,    -74,     57,     65,    -67,    -55,     76,     42,    -83,    -29,     88,     15,    -90,     -1,     90,    -13,    -88,     27,     84,    -40,    -77,     53,     69,    -64,    -59,     73,     47,    -81,    -34,     86,     21,    -89,     -7,     90,     -8,    -89,     22,     86,    -35,    -80,     48,     72,    -60,    -63,     70,     52,    -78,    -39,     84,     26,    -88,    -12,     90,     -2,    -90,     16,     87,    -30,    -82,     43,     76,    -55,    -67,     66,     56,    -75,    -44,     82,     31,    -87,    -18,     90,      3,    -90,     11,     89,    -25,    -85,     38,     78,    -51,    -70,     62,     60,    -72,    -49,     79,     36,    -85,    -23,     89,      9,    -90,      6,     89,    -20,    -86,     33,     81,    -46,    -74,     58,     64,    -68,    -54,     77,     41,    -83,    -28,     88,     14,    -90 },
      {     60,    -73,    -45,     82,     28,    -88,    -10,     90,     -9,    -89,     27,     83,    -44,    -74,     60,     61,    -72,    -46,     82,     29,    -88,    -11,     90,     -8,    -89,     26,     83,    -43,    -74,     59,     62,    -72,    -47,     81,     30,    -88,    -12,     90,     -7,    -89,     25,     84,    -42,    -75,     58,     63,    -71,    -48,     81,     31,    -88,    -13,     90,     -6,    -89,     24,     84,    -41,    -76,     57,     64,    -70,    -49,     81,     32,    -87,    -14,     90,     -4,    -89,     23,     85,    -40,    -76,     56,     64,    -70,    -50,     80,     33,    -87,    -15,     90,     -3,    -89,     22,     85,    -39,    -77,     55,     65,    -69,    -51,     79,     34,    -87,    -16,     90,     -2,    -90,     21,     85,    -38,    -77,     55,     66,    -68,    -52,     79,     35,    -86,    -18,     90,     -1,    -90,     20,     86,    -37,    -78,     54,     67,    -67,    -53,     78,     36,    -86,    -19,     90 },
      {     60,    -75,    -40,     85,     19,    -90,      4,     89,    -27,    -82,     48,     70,    -66,    -53,     79,     32,    -88,    -10,     90,    -13,    -87,     35,     78,    -55,    -64,     72,     45,    -83,    -24,     89,      1,    -90,     22,     84,    -43,    -73,     62,     57,    -77,    -37,     86,     15,    -90,      8,     88,    -30,    -81,     51,     67,    -68,    -50,     81,     29,    -89,     -7,     90,    -16,    -86,     38,     76,    -58,    -61,     74,     42,    -85,    -21,     90,     -2,    -89,     25,     83,    -46,    -71,     64,     55,    -78,    -34,     87,     12,    -90,     11,     88,    -33,    -79,     54,     65,    -70,    -47,     82,     26,    -89,     -3,     90,    -20,    -85,     41,     74,    -60,    -59,     76,     39,    -86,    -18,     90,     -6,    -89,     28,     81,    -49,    -69,     67,     52,    -80,    -31,     88,      9,    -90,     14,     87,    -36,    -77,     56,     63,    -72,    -44,     84,     23,    -90 },
      {     59,    -77,    -35,     88,      9,    -90,     19,     85,    -44,    -71,     66,     51,    -81,    -26,     89,     -1,    -89,     28,     81,    -53,    -64,     72,     42,    -85,    -16,     90,    -11,    -87,     37,     76,    -60,    -57,     78,     33,    -88,     -7,     90,    -21,    -84,     46,     70,    -67,    -49,     82,     24,    -90,      3,     89,    -30,    -79,     55,     63,    -74,    -40,     86,     14,    -90,     13,     86,    -39,    -74,     62,     55,    -79,    -31,     89,      4,    -90,     23,     83,    -48,    -68,     69,     47,    -83,    -22,     90,     -6,    -88,     32,     78,    -56,    -61,     75,     38,    -87,    -12,     90,    -15,    -86,     41,     73,    -64,    -54,     80,     29,    -89,     -2,     90,    -25,    -82,     50,     67,    -70,    -45,     84,     20,    -90,      8,     88,    -34,    -77,     58,     60,    -76,    -36,     87,     10,    -90,     18,     85,    -43,    -72,     65,     52,    -81,    -27,     89 },
      {     58,    -78,    -30,     89,     -1,    -89,     32,     77,    -60,    -56,     79,     28,    -89,      3,     88,    -34,    -76,     61,     55,    -81,    -26,     90,     -6,    -88,     36,     75,    -63,    -53,     81,     24,    -90,      8,     87,    -38,    -74,     64,     51,    -82,    -22,     90,    -10,    -87,     40,     72,    -66,    -49,     83,     20,    -90,     12,     86,    -42,    -71,     67,     47,    -84,    -18,     90,    -14,    -85,     44,     70,    -69,    -45,     85,     15,    -90,     16,     85,    -46,    -68,     70,     43,    -86,    -13,     90,    -19,    -84,     48,     67,    -72,    -41,     86,     11,    -90,     21,     83,    -50,    -65,     73,     39,    -87,     -9,     90,    -23,    -82,     52,     64,    -74,    -37,     88,      7,    -90,     25,     81,    -54,    -62,     76,     35,    -88,     -4,     90,    -27,    -80,     55,     60,    -77,    -33,     89,      2,    -89,     29,     79,    -57,    -59,     78,     31,    -89 },
      {     57,    -80,    -25,     90,    -11,    -86,     45,     67,    -72,    -38,     88,      3,    -89,     32,     76,    -63,    -51,     83,     18,    -90,     19,     83,    -52,    -62,     77,     31,    -89,      4,     88,    -39,    -72,     68,     44,    -86,    -10,     90,    -26,    -79,     58,     56,    -81,    -24,     90,    -12,    -85,     46,     67,    -73,    -37,     88,      2,    -89,     33,     76,    -64,    -50,     84,     16,    -90,     20,     82,    -53,    -61,     77,     30,    -89,      6,     87,    -40,    -71,     69,     43,    -86,     -9,     90,    -27,    -79,     59,     55,    -81,    -23,     90,    -13,    -85,     47,     66,    -74,    -36,     88,      1,    -89,     34,     75,    -64,    -49,     84,     15,    -90,     21,     82,    -54,    -60,     78,     29,    -90,      7,     87,    -41,    -70,     70,     42,    -87,     -8,     90,    -28,    -78,     60,     55,    -81,    -22,     90,    -14,    -85,     48,     65,    -74,    -35,     89 },
      {     56,    -81,    -20,     90,    -21,    -81,     57,     55,    -82,    -19,     90,    -22,    -81,     58,     55,    -82,    -18,     90,    -23,    -80,     59,     54,    -83,    -16,     90,    -24,    -79,     60,     53,    -83,    -15,     90,    -25,    -79,     60,     52,    -84,    -14,     90,    -26,    -78,     61,     51,    -84,    -13,     90,    -27,    -78,     62,     50,    -85,    -12,     90,    -28,    -77,     63,     49,    -85,    -11,     90,    -29,    -77,     64,     48,    -85,    -10,     90,    -30,    -76,     64,     47,    -86,     -9,     90,    -31,    -76,     65,     46,    -86,     -8,     89,    -32,    -75,     66,     45,    -86,     -7,     89,    -33,    -74,     67,     44,    -87,     -6,     89,    -34,    -74,     67,     43,    -87,     -4,     89,    -35,    -73,     68,     42,    -87,     -3,     89,    -36,    -72,     69,     41,    -88,     -2,     89,    -37,    -72,     70,     40,    -88,     -1,     88,    -38,    -71,     70,     39,    -88 },
      {     55,    -83,    -14,     90,    -30,    -75,     67,     41,    -88,      2,     87,    -45,    -64,     77,     26,    -90,     19,     81,    -59,    -52,     85,     10,    -89,     34,     72,    -70,    -37,     89,     -7,    -86,     49,     61,    -79,    -22,     90,    -23,    -79,     62,     48,    -86,     -6,     89,    -38,    -70,     73,     33,    -90,     11,     84,    -53,    -58,     81,     18,    -90,     27,     77,    -65,    -44,     87,      1,    -88,     42,     67,    -76,    -29,     90,    -15,    -82,     56,     55,    -83,    -13,     90,    -31,    -74,     68,     40,    -88,      3,     87,    -46,    -64,     78,     25,    -90,     20,     81,    -60,    -51,     85,      9,    -89,     35,     72,    -71,    -36,     89,     -8,    -85,     50,     60,    -80,    -21,     90,    -24,    -78,     63,     47,    -86,     -4,     89,    -39,    -69,     74,     32,    -90,     12,     84,    -54,    -57,     82,     16,    -90,     28,     76,    -66,    -43,     88 },
      {     55,    -84,     -9,     89,    -39,    -67,     76,     26,    -90,     23,     78,    -65,    -42,     88,     -6,    -85,     52,     57,    -83,    -12,     89,    -36,    -70,     74,     29,    -90,     20,     79,    -63,    -45,     88,     -2,    -86,     49,     60,    -81,    -15,     90,    -33,    -72,     72,     32,    -90,     16,     81,    -60,    -48,     87,      1,    -87,     46,     62,    -80,    -19,     90,    -30,    -74,     70,     35,    -90,     13,     82,    -58,    -51,     86,      4,    -88,     43,     64,    -78,    -22,     90,    -27,    -76,     68,     38,    -89,     10,     84,    -55,    -54,     85,      8,    -89,     40,     67,    -77,    -25,     90,    -24,    -77,     66,     41,    -89,      7,     85,    -53,    -56,     83,     11,    -89,     37,     69,    -75,    -28,     90,    -21,    -79,     64,     44,    -88,      3,     86,    -50,    -59,     82,     14,    -90,     34,     71,    -73,    -31,     90,    -18,    -81,     61,     47,    -87 },
      {     54,    -85,     -3,     87,    -48,    -59,     83,     10,    -89,     42,     64,    -80,    -16,     90,    -36,    -68,     77,     23,    -90,     30,     72,    -73,    -29,     90,    -24,    -76,     69,     35,    -90,     18,     79,    -64,    -41,     89,    -11,    -82,     60,     47,    -88,      4,     85,    -55,    -53,     86,      2,    -87,     49,     58,    -83,     -9,     89,    -43,    -63,     81,     15,    -90,     37,     67,    -77,    -22,     90,    -31,    -72,     74,     28,    -90,     25,     76,    -70,    -34,     90,    -19,    -79,     65,     40,    -89,     12,     82,    -60,    -46,     88,     -6,    -85,     55,     52,    -86,     -1,     87,    -50,    -57,     84,      8,    -88,     44,     62,    -81,    -14,     89,    -38,    -67,     78,     21,    -90,     32,     71,    -74,    -27,     90,    -26,    -75,     70,     33,    -90,     20,     78,    -66,    -39,     89,    -13,    -81,     61,     45,    -88,      7,     84,    -56,    -51,     86 },
      {     53,    -86,      2,     85,    -56,    -49,     88,     -7,    -83,     60,     45,    -89,     11,     81,    -63,    -41,     89,    -15,    -79,     66,     37,    -90,     20,     77,    -69,    -33,     90,    -24,    -75,     72,     29,    -90,     28,     72,    -74,    -25,     90,    -32,    -70,     77,     21,    -90,     36,     67,    -79,    -16,     89,    -40,    -64,     81,     12,    -89,     44,     60,    -83,     -8,     88,    -48,    -57,     85,      3,    -87,     52,     54,    -86,      1,     85,    -55,    -50,     87,     -6,    -84,     59,     46,    -88,     10,     82,    -62,    -42,     89,    -14,    -80,     65,     38,    -90,     19,     78,    -68,    -34,     90,    -23,    -76,     71,     30,    -90,     27,     73,    -74,    -26,     90,    -31,    -70,     76,     22,    -90,     35,     67,    -78,    -18,     90,    -39,    -64,     81,     13,    -89,     43,     61,    -82,     -9,     88,    -47,    -58,     84,      4,    -87,     51,     55,    -86 },
      {     52,    -87,      8,     82,    -64,    -38,     90,    -23,    -74,     74,     24,    -90,     37,     64,    -81,     -9,     88,    -51,    -53,     87,     -7,    -82,     63,     39,    -90,     22,     75,    -73,    -25,     90,    -36,    -65,     81,     10,    -88,     50,     54,    -87,      6,     83,    -62,    -40,     90,    -21,    -76,     72,     26,    -90,     35,     66,    -81,    -11,     88,    -49,    -55,     86,     -4,    -83,     61,     41,    -90,     20,     76,    -72,    -27,     90,    -34,    -67,     80,     12,    -88,     48,     55,    -86,      3,     84,    -60,    -42,     89,    -19,    -77,     71,     28,    -90,     33,     67,    -79,    -13,     89,    -47,    -56,     86,     -2,    -84,     60,     43,    -89,     18,     77,    -70,    -29,     90,    -32,    -68,     79,     14,    -89,     46,     57,    -85,      1,     85,    -59,    -44,     89,    -16,    -78,     70,     30,    -90,     31,     69,    -78,    -15,     89,    -45,    -58,     85 },
      {     51,    -88,     13,     78,    -70,    -27,     90,    -38,    -62,     84,      1,    -85,     60,     40,    -90,     25,     72,    -77,    -15,     89,    -49,    -53,     88,    -11,    -79,     69,     29,    -90,     36,     64,    -83,     -3,     85,    -59,    -42,     90,    -23,    -73,     76,     18,    -89,     47,     55,    -87,      9,     81,    -67,    -31,     90,    -34,    -65,     82,      6,    -86,     57,     44,    -89,     21,     74,    -75,    -20,     89,    -45,    -56,     86,     -7,    -81,     66,     33,    -90,     32,     67,    -81,     -8,     87,    -55,    -46,     89,    -19,    -76,     74,     22,    -90,     43,     58,    -86,      4,     82,    -64,    -35,     90,    -30,    -68,     80,     10,    -87,     54,     48,    -89,     16,     77,    -72,    -24,     90,    -41,    -60,     85,     -2,    -83,     63,     37,    -90,     28,     70,    -79,    -12,     88,    -52,    -50,     88,    -14,    -78,     71,     26,    -90,     39,     61,    -84 },
      {     50,    -89,     19,     74,    -76,    -15,     88,    -53,    -47,     89,    -22,    -72,     78,     12,    -87,     55,     44,    -90,     25,     70,    -79,     -9,     86,    -58,    -41,     90,    -28,    -68,     81,      6,    -85,     60,     38,    -90,     31,     66,    -82,     -2,     84,    -63,    -35,     90,    -34,    -64,     84,     -1,    -83,     65,     32,    -90,     37,     61,    -85,      4,     81,    -67,    -29,     90,    -40,    -59,     86,     -8,    -80,     70,     26,    -90,     43,     56,    -87,     11,     78,    -72,    -23,     89,    -46,    -54,     88,    -14,    -77,     74,     20,    -89,     49,     51,    -89,     18,     75,    -76,    -16,     88,    -52,    -48,     89,    -21,    -73,     77,     13,    -88,     55,     45,    -90,     24,     71,    -79,    -10,     87,    -57,    -42,     90,    -27,    -69,     81,      7,    -86,     60,     39,    -90,     30,     67,    -82,     -3,     85,    -62,    -36,     90,    -33,    -64,     83 },
      {     49,    -89,     24,     70,    -81,     -3,     84,    -65,    -30,     90,    -43,    -55,     88,    -18,    -74,     78,     10,    -86,     60,     36,    -90,     37,     60,    -86,     11,     77,    -74,    -16,     88,    -55,    -42,     90,    -31,    -64,     84,     -4,    -81,     70,     23,    -89,     50,     48,    -89,     25,     69,    -81,     -2,     83,    -66,    -29,     90,    -44,    -54,     88,    -19,    -73,     78,      9,    -86,     61,     35,    -90,     38,     59,    -87,     12,     77,    -75,    -15,     88,    -56,    -41,     90,    -32,    -64,     85,     -6,    -80,     71,     22,    -89,     51,     47,    -90,     26,     68,    -82,     -1,     83,    -67,    -28,     90,    -45,    -53,     89,    -20,    -72,     79,      8,    -85,     62,     34,    -90,     39,     58,    -87,     13,     76,    -76,    -14,     87,    -57,    -40,     90,    -33,    -63,     85,     -7,    -79,     72,     21,    -89,     52,     46,    -90,     27,     67,    -82 },
      {     48,    -90,     29,     64,    -85,      9,     77,    -76,    -12,     86,    -62,    -32,     90,    -45,    -51,     89,    -26,    -67,     84,     -6,    -79,     74,     15,    -87,     60,     35,    -90,     42,     54,    -89,     23,     69,    -82,      2,     81,    -72,    -19,     88,    -57,    -38,     90,    -39,    -56,     88,    -20,    -71,     81,      1,    -82,     70,     22,    -89,     55,     41,    -90,     36,     59,    -87,     16,     73,    -79,     -4,     83,    -67,    -25,     89,    -52,    -44,     90,    -33,    -61,     86,    -13,    -75,     78,      8,    -85,     65,     28,    -90,     49,     47,    -90,     30,     64,    -85,     10,     77,    -76,    -11,     86,    -63,    -31,     90,    -46,    -50,     89,    -27,    -66,     84,     -7,    -78,     74,     14,    -87,     60,     34,    -90,     43,     53,    -89,     24,     68,    -83,      3,     80,    -72,    -18,     88,    -58,    -37,     90,    -40,    -55,     88,    -21,    -70,     81 },
      {     47,    -90,     34,     59,    -88,     21,     69,    -83,      7,     77,    -77,     -8,     84,    -68,    -22,     88,    -58,    -35,     90,    -46,    -48,     90,    -33,    -60,     88,    -20,    -70,     83,     -6,    -78,     76,      9,    -84,     67,     23,    -88,     57,     36,    -90,     45,     49,    -90,     32,     60,    -87,     19,     70,    -82,      4,     78,    -76,    -10,     85,    -67,    -24,     89,    -56,    -37,     90,    -44,    -50,     90,    -31,    -61,     87,    -18,    -71,     82,     -3,    -79,     75,     11,    -85,     66,     25,    -89,     55,     38,    -90,     43,     51,    -90,     30,     62,    -87,     16,     72,    -81,      2,     79,    -74,    -12,     85,    -65,    -26,     89,    -55,    -39,     90,    -42,    -52,     89,    -29,    -63,     86,    -15,    -72,     81,     -1,    -80,     74,     13,    -86,     64,     27,    -89,     54,     40,    -90,     41,     53,    -89,     28,     64,    -86,     14,     73,    -81 },
      {     46,    -90,     39,     53,    -90,     32,     59,    -88,     25,     64,    -86,     18,     70,    -84,     10,     74,    -81,      2,     78,    -77,     -6,     82,    -72,    -13,     85,    -67,    -21,     87,    -62,    -28,     89,    -56,    -35,     90,    -50,    -42,     90,    -43,    -49,     90,    -36,    -55,     89,    -29,    -61,     88,    -22,    -67,     85,    -14,    -72,     82,     -7,    -76,     79,      1,    -80,     75,      9,    -83,     70,     16,    -86,     65,     24,    -88,     60,     31,    -89,     54,     38,    -90,     47,     45,    -90,     40,     52,    -90,     33,     58,    -89,     26,     64,    -87,     19,     69,    -84,     11,     74,    -81,      3,     78,    -77,     -4,     81,    -73,    -12,     85,    -68,    -20,     87,    -63,    -27,     89,    -57,    -34,     90,    -51,    -41,     90,    -44,    -48,     90,    -37,    -55,     89,    -30,    -60,     88,    -23,    -66,     86,    -15,    -71,     83,     -8,    -76,     79 },
      {     45,    -90,     44,     46,    -90,     43,     47,    -90,     42,     48,    -90,     41,     49,    -90,     40,     50,    -90,     39,     51,    -90,     38,     52,    -90,     37,     53,    -90,     36,     54,    -90,     35,     55,    -90,     34,     55,    -89,     33,     56,    -89,     32,     57,    -89,     31,     58,    -89,     30,     59,    -89,     29,     60,    -89,     28,     60,    -88,     27,     61,    -88,     26,     62,    -88,     25,     63,    -88,     24,     64,    -87,     23,     64,    -87,     22,     65,    -87,     21,     66,    -86,     20,     67,    -86,     19,     67,    -86,     18,     68,    -85,     16,     69,    -85,     15,     70,    -85,     14,     70,    -84,     13,     71,    -84,     12,     72,    -83,     11,     72,    -83,     10,     73,    -82,      9,     74,    -82,      8,     74,    -81,      7,     75,    -81,      6,     76,    -81,      4,     76,    -80,      3,     77,    -79,      2,     77,    -79,      1,     78,    -78 },
      {     44,    -90,     49,     39,    -90,     54,     34,    -89,     58,     29,    -88,     62,     24,    -87,     66,     19,    -85,     70,     13,    -83,     73,      8,    -81,     76,      2,    -78,     79,     -3,    -76,     81,     -9,    -72,     84,    -14,    -69,     86,    -20,    -65,     87,    -25,    -61,     89,    -30,    -57,     89,    -35,    -53,     90,    -40,    -48,     90,    -45,    -43,     90,    -50,    -38,     90,    -55,    -33,     89,    -59,    -28,     88,    -63,    -23,     87,    -67,    -18,     85,    -70,    -12,     83,    -74,     -7,     81,    -77,     -1,     78,    -79,      4,     75,    -82,     10,     72,    -84,     15,     68,    -86,     21,     64,    -88,     26,     60,    -89,     31,     56,    -90,     36,     52,    -90,     41,     47,    -90,     46,     42,    -90,     51,     37,    -90,     55,     32,    -89,     60,     27,    -88,     64,     22,    -86,     67,     16,    -85,     71,     11,    -82,     74,      6,    -80,     77 },
      {     43,    -90,     54,     32,    -89,     63,     21,    -85,     71,      9,    -81,     78,     -3,    -74,     83,    -15,    -67,     87,    -27,    -58,     90,    -38,    -48,     90,    -49,    -37,     89,    -59,    -26,     87,    -67,    -14,     83,    -75,     -2,     77,    -81,     10,     70,    -86,     22,     62,    -89,     33,     53,    -90,     44,     42,    -90,     55,     31,    -88,     64,     20,    -85,     72,      8,    -80,     78,     -4,    -74,     84,    -16,    -66,     88,    -28,    -57,     90,    -39,    -47,     90,    -50,    -36,     89,    -60,    -25,     87,    -68,    -13,     82,    -76,     -1,     77,    -81,     11,     70,    -86,     23,     61,    -89,     34,     52,    -90,     45,     41,    -90,     55,     30,    -88,     64,     19,    -85,     72,      7,    -79,     79,     -6,    -73,     84,    -18,    -65,     88,    -29,    -56,     90,    -40,    -46,     90,    -51,    -35,     89,    -60,    -24,     86,    -69,    -12,     82,    -76 },
      {     42,    -90,     58,     25,    -86,     71,      7,    -78,     81,    -12,    -67,     88,    -30,    -54,     90,    -47,    -37,     89,    -62,    -20,     84,    -74,     -1,     76,    -83,     18,     64,    -89,     35,     49,    -90,     52,     32,    -88,     66,     14,    -82,     77,     -4,    -72,     85,    -23,    -60,     90,    -40,    -44,     90,    -56,    -27,     87,    -70,     -9,     79,    -80,     10,     69,    -87,     28,     55,    -90,     45,     39,    -89,     60,     22,    -85,     73,      3,    -77,     82,    -15,    -65,     88,    -33,    -51,     90,    -50,    -34,     89,    -64,    -16,     83,    -76,      2,     74,    -85,     21,     61,    -89,     38,     46,    -90,     55,     29,    -87,     68,     11,    -81,     79,     -8,    -70,     86,    -26,    -57,     90,    -43,    -41,     90,    -59,    -24,     86,    -72,     -6,     78,    -81,     13,     67,    -88,     31,     53,    -90,     48,     36,    -89,     63,     19,    -84,     75 },
      {     41,    -89,     62,     18,    -82,     78,     -8,    -69,     88,    -32,    -50,     90,    -55,    -27,     86,    -72,     -2,     75,    -85,     23,     58,    -90,     46,     36,    -89,     66,     12,    -80,     81,    -13,    -65,     89,    -37,    -45,     90,    -59,    -22,     84,    -76,      3,     72,    -86,     28,     54,    -90,     51,     31,    -87,     70,      7,    -77,     83,    -19,    -61,     90,    -42,    -40,     89,    -63,    -16,     82,    -78,      9,     68,    -88,     33,     49,    -90,     55,     26,    -86,     73,      1,    -74,     85,    -24,    -57,     90,    -47,    -35,     88,    -67,    -11,     79,    -81,     14,     64,    -89,     38,     44,    -90,     60,     21,    -84,     76,     -4,    -71,     87,    -29,    -53,     90,    -52,    -30,     87,    -70,     -6,     77,    -83,     20,     60,    -90,     43,     39,    -89,     64,     15,    -81,     79,    -10,    -67,     88,    -34,    -48,     90,    -56,    -25,     85,    -74 },
      {     40,    -89,     66,     10,    -78,     83,    -22,    -57,     90,    -51,    -29,     86,    -74,      2,     71,    -87,     33,     47,    -90,     60,     18,    -81,     80,    -14,    -63,     90,    -44,    -36,     88,    -69,     -6,     76,    -85,     26,     54,    -90,     55,     25,    -85,     76,     -7,    -68,     88,    -37,    -43,     89,    -64,    -13,     79,    -82,     19,     60,    -90,     48,     32,    -87,     72,      1,    -73,     86,    -30,    -50,     90,    -58,    -21,     83,    -78,     11,     65,    -89,     41,     39,    -89,     67,      9,    -77,     84,    -23,    -56,     90,    -52,    -28,     86,    -74,      3,     70,    -88,     34,     46,    -90,     61,     16,    -81,     81,    -15,    -62,     90,    -45,    -35,     88,    -70,     -4,     75,    -85,     27,     53,    -90,     55,     24,    -84,     77,     -8,    -67,     89,    -38,    -42,     89,    -64,    -12,     79,    -82,     20,     59,    -90,     49,     31,    -87,     72 },
      {     39,    -88,     70,      2,    -72,     87,    -35,    -43,     89,    -67,     -7,     75,    -86,     31,     47,    -90,     64,     11,    -77,     85,    -27,    -51,     90,    -60,    -15,     79,    -83,     23,     55,    -90,     57,     20,    -81,     81,    -19,    -58,     90,    -54,    -24,     83,    -79,     14,     61,    -90,     50,     28,    -85,     77,    -10,    -64,     90,    -46,    -32,     86,    -74,      6,     67,    -89,     42,     36,    -88,     72,     -1,    -70,     88,    -38,    -40,     89,    -69,     -3,     73,    -87,     34,     44,    -89,     66,      8,    -76,     86,    -30,    -48,     90,    -63,    -12,     78,    -84,     26,     52,    -90,     60,     16,    -80,     82,    -22,    -55,     90,    -56,    -21,     82,    -81,     18,     59,    -90,     53,     25,    -84,     78,    -13,    -62,     90,    -49,    -29,     85,    -76,      9,     65,    -89,     45,     33,    -87,     74,     -4,    -68,     89,    -41,    -37,     88,    -71 },
      {     38,    -88,     73,     -6,    -66,     90,    -48,    -28,     84,    -79,     16,     58,    -90,     57,     18,    -79,     84,    -27,    -49,     90,    -65,     -7,     74,    -87,     37,     39,    -88,     72,     -4,    -67,     89,    -47,    -29,     85,    -78,     15,     59,    -90,     56,     19,    -80,     83,    -26,    -50,     90,    -64,     -8,     74,    -87,     36,     40,    -88,     72,     -3,    -67,     89,    -46,    -30,     85,    -78,     14,     60,    -90,     55,     20,    -81,     83,    -25,    -51,     90,    -64,     -9,     75,    -87,     35,     41,    -88,     71,     -2,    -68,     89,    -45,    -31,     85,    -77,     13,     60,    -90,     55,     21,    -81,     82,    -24,    -52,     90,    -63,    -10,     76,    -86,     34,     42,    -89,     70,     -1,    -69,     89,    -44,    -32,     86,    -77,     12,     61,    -90,     54,     22,    -81,     82,    -23,    -53,     90,    -62,    -11,     76,    -86,     33,     43,    -89,     70 },
      {     37,    -87,     76,    -13,    -59,     90,    -60,    -12,     76,    -87,     38,     36,    -86,     77,    -14,    -58,     90,    -60,    -11,     75,    -87,     39,     35,    -86,     77,    -15,    -57,     90,    -61,    -10,     74,    -88,     40,     34,    -86,     78,    -16,    -56,     90,    -62,     -9,     74,    -88,     41,     33,    -85,     78,    -18,    -55,     90,    -63,     -8,     73,    -88,     42,     32,    -85,     79,    -19,    -55,     90,    -64,     -7,     72,    -88,     43,     31,    -85,     79,    -20,    -54,     90,    -64,     -6,     72,    -89,     44,     30,    -84,     80,    -21,    -53,     90,    -65,     -4,     71,    -89,     45,     29,    -84,     81,    -22,    -52,     90,    -66,     -3,     70,    -89,     46,     28,    -83,     81,    -23,    -51,     90,    -67,     -2,     70,    -89,     47,     27,    -83,     81,    -24,    -50,     90,    -67,     -1,     69,    -89,     48,     26,    -82,     82,    -25,    -49,     89,    -68 },
      {     36,    -86,     79,    -21,    -51,     89,    -70,      4,     64,    -90,     58,     12,    -74,     88,    -44,    -28,     82,    -83,     29,     43,    -88,     75,    -13,    -57,     90,    -64,     -3,     69,    -90,     52,     20,    -78,     86,    -37,    -35,     85,    -79,     22,     50,    -89,     70,     -6,    -63,     90,    -59,    -11,     74,    -88,     45,     27,    -82,     83,    -30,    -42,     88,    -76,     14,     56,    -90,     65,      2,    -68,     90,    -53,    -19,     78,    -86,     38,     34,    -85,     80,    -23,    -49,     89,    -71,      7,     62,    -90,     60,     10,    -73,     89,    -46,    -26,     81,    -84,     31,     41,    -87,     76,    -15,    -55,     90,    -66,     -1,     67,    -90,     54,     18,    -77,     87,    -39,    -33,     85,    -81,     24,     48,    -89,     72,     -8,    -61,     90,    -60,     -9,     72,    -89,     47,     25,    -81,     84,    -32,    -40,     87,    -77,     16,     55,    -90,     67 },
      {     35,    -85,     81,    -28,    -42,     87,    -78,     21,     49,    -89,     74,    -13,    -55,     90,    -69,      6,     61,    -90,     64,      2,    -67,     90,    -58,    -10,     72,    -89,     52,     18,    -76,     88,    -45,    -25,     80,    -86,     38,     32,    -83,     83,    -31,    -39,     86,    -79,     24,     46,    -88,     76,    -16,    -53,     89,    -71,      9,     59,    -90,     66,     -1,    -64,     90,    -60,     -7,     70,    -90,     55,     14,    -74,     89,    -48,    -22,     78,    -87,     41,     29,    -82,     84,    -34,    -36,     85,    -81,     27,     43,    -87,     77,    -20,    -50,     89,    -73,     12,     56,    -90,     68,     -4,    -62,     90,    -63,     -3,     67,    -90,     57,     11,    -72,     89,    -51,    -19,     77,    -88,     44,     26,    -81,     85,    -37,    -33,     84,    -82,     30,     40,    -86,     79,    -23,    -47,     88,    -75,     15,     54,    -90,     70,     -8,    -60,     90,    -65 },
      {     34,    -83,     84,    -35,    -33,     83,    -84,     36,     32,    -82,     85,    -37,    -31,     82,    -85,     38,     30,    -81,     85,    -39,    -29,     81,    -86,     40,     28,    -81,     86,    -41,    -27,     80,    -86,     42,     26,    -79,     87,    -43,    -25,     79,    -87,     44,     24,    -78,     87,    -45,    -23,     78,    -88,     46,     22,    -77,     88,    -47,    -21,     77,    -88,     48,     20,    -76,     88,    -49,    -19,     76,    -89,     50,     18,    -75,     89,    -51,    -16,     74,    -89,     52,     15,    -74,     89,    -53,    -14,     73,    -89,     54,     13,    -72,     89,    -55,    -12,     72,    -90,     55,     11,    -71,     90,    -56,    -10,     70,    -90,     57,      9,    -70,     90,    -58,     -8,     69,    -90,     59,      7,    -68,     90,    -60,     -6,     67,    -90,     60,      4,    -67,     90,    -61,     -3,     66,    -90,     62,      2,    -65,     90,    -63,     -1,     64,    -90,     64 },
      {     33,    -82,     86,    -42,    -24,     77,    -88,     51,     14,    -72,     90,    -59,     -4,     65,    -90,     66,     -6,    -58,     90,    -72,     15,     50,    -88,     78,    -25,    -41,     85,    -82,     34,     32,    -81,     86,    -43,    -23,     77,    -89,     52,     13,    -71,     90,    -60,     -3,     64,    -90,     67,     -7,    -57,     90,    -73,     16,     49,    -88,     78,    -26,    -40,     85,    -83,     35,     31,    -81,     86,    -44,    -22,     76,    -89,     53,     12,    -70,     90,    -60,     -2,     64,    -90,     67,     -8,    -56,     89,    -74,     18,     48,    -88,     79,    -27,    -39,     85,    -83,     36,     30,    -81,     87,    -45,    -21,     76,    -89,     54,     11,    -70,     90,    -61,     -1,     63,    -90,     68,     -9,    -55,     89,    -74,     19,     47,    -87,     79,    -28,    -38,     84,    -84,     37,     29,    -80,     87,    -46,    -20,     75,    -89,     55,     10,    -69,     90,    -62 },
      {     32,    -81,     87,    -49,    -14,     70,    -90,     64,     -4,    -57,     89,    -76,     23,     41,    -85,     84,    -40,    -24,     76,    -89,     56,      6,    -64,     90,    -70,     13,     50,    -88,     80,    -31,    -33,     81,    -87,     48,     15,    -71,     90,    -63,      3,     58,    -89,     75,    -22,    -42,     85,    -84,     39,     25,    -77,     89,    -55,     -7,     65,    -90,     69,    -12,    -51,     88,    -79,     30,     34,    -81,     87,    -47,    -16,     72,    -90,     62,     -2,    -59,     90,    -74,     21,     43,    -85,     83,    -38,    -26,     77,    -89,     55,      8,    -66,     90,    -68,     11,     52,    -88,     79,    -29,    -35,     82,    -86,     46,     18,    -72,     90,    -61,      1,     60,    -90,     74,    -20,    -44,     86,    -83,     37,     27,    -78,     89,    -54,     -9,     67,    -90,     67,    -10,    -53,     88,    -78,     28,     36,    -82,     86,    -45,    -19,     73,    -90,     60 },
      {     31,    -79,     89,    -55,     -4,     62,    -90,     74,    -23,    -39,     83,    -86,     48,     13,    -68,     90,    -69,     14,     47,    -86,     83,    -40,    -22,     74,    -90,     63,     -6,    -55,     88,    -79,     32,     30,    -78,     89,    -56,     -3,     61,    -90,     75,    -24,    -38,     82,    -87,     49,     12,    -67,     90,    -70,     15,     46,    -86,     84,    -41,    -21,     73,    -90,     64,     -7,    -54,     88,    -80,     33,     29,    -78,     89,    -57,     -2,     60,    -90,     76,    -25,    -37,     82,    -87,     50,     11,    -67,     90,    -70,     16,     45,    -85,     84,    -42,    -20,     72,    -90,     64,     -8,    -53,     88,    -81,     34,     28,    -77,     89,    -58,     -1,     60,    -89,     76,    -26,    -36,     81,    -87,     51,     10,    -66,     90,    -71,     18,     44,    -85,     85,    -43,    -19,     72,    -90,     65,     -9,    -52,     88,    -81,     35,     27,    -77,     89,    -59 },
      {     30,    -77,     89,    -61,      6,     53,    -87,     82,    -40,    -20,     71,    -90,     69,    -16,    -43,     84,    -86,     50,      9,    -64,     90,    -76,     27,     33,    -79,     89,    -59,      2,     55,    -88,     81,    -37,    -23,     73,    -90,     67,    -13,    -46,     85,    -85,     47,     12,    -66,     90,    -74,     24,     36,    -81,     88,    -56,     -1,     58,    -89,     79,    -34,    -26,     75,    -90,     64,    -10,    -49,     86,    -84,     44,     15,    -68,     90,    -72,     21,     39,    -82,     88,    -54,     -4,     60,    -89,     78,    -31,    -29,     77,    -90,     62,     -7,    -52,     87,    -83,     41,     19,    -70,     90,    -70,     18,     42,    -83,     87,    -51,     -8,     63,    -90,     76,    -28,    -32,     78,    -89,     60,     -3,    -55,     88,    -81,     38,     22,    -72,     90,    -67,     14,     45,    -85,     86,    -48,    -11,     65,    -90,     74,    -25,    -35,     80,    -89,     57 },
      {     29,    -76,     90,    -67,     15,     42,    -82,     88,    -56,      1,     55,    -87,     83,    -44,    -13,     65,    -90,     77,    -31,    -27,     74,    -90,     68,    -18,    -40,     81,    -88,     58,     -3,    -53,     87,    -84,     46,     11,    -64,     90,    -78,     33,     25,    -73,     90,    -70,     20,     38,    -81,     89,    -60,      6,     51,    -86,     85,    -48,     -9,     62,    -89,     79,    -35,    -23,     72,    -90,     71,    -22,    -36,     79,    -89,     61,     -8,    -49,     85,    -86,     50,      7,    -60,     89,    -80,     37,     21,    -70,     90,    -72,     24,     34,    -78,     89,    -63,     10,     47,    -85,     86,    -52,     -4,     59,    -89,     81,    -39,    -19,     69,    -90,     74,    -26,    -32,     77,    -90,     64,    -12,    -45,     84,    -87,     54,      2,    -57,     88,    -82,     41,     16,    -67,     90,    -75,     28,     30,    -76,     90,    -66,     14,     43,    -83,     88,    -55 },
      {     28,    -74,     90,    -72,     25,     31,    -76,     90,    -70,     22,     34,    -77,     90,    -67,     19,     37,    -79,     90,    -65,     15,     40,    -81,     89,    -63,     12,     43,    -82,     89,    -60,      9,     46,    -83,     88,    -58,      6,     49,    -85,     87,    -55,      2,     52,    -86,     86,    -53,     -1,     55,    -87,     85,    -50,     -4,     57,    -88,     84,    -47,     -8,     60,    -88,     82,    -44,    -11,     62,    -89,     81,    -41,    -14,     64,    -89,     79,    -38,    -18,     67,    -90,     78,    -35,    -21,     69,    -90,     76,    -32,    -24,     71,    -90,     74,    -29,    -27,     73,    -90,     72,    -26,    -30,     75,    -90,     70,    -23,    -33,     77,    -90,     68,    -20,    -36,     78,    -90,     66,    -16,    -39,     80,    -89,     64,    -13,    -42,     81,    -89,     61,    -10,    -45,     83,    -88,     59,     -7,    -48,     84,    -87,     56,     -3,    -51,     85,    -86,     54 },
      {     27,    -72,     90,    -76,     34,     20,    -67,     90,    -80,     41,     12,    -61,     88,    -83,     48,      4,    -55,     86,    -86,     55,     -3,    -49,     84,    -88,     60,    -11,    -42,     81,    -89,     66,    -19,    -35,     77,    -90,     71,    -26,    -28,     72,    -90,     76,    -33,    -21,     67,    -90,     79,    -40,    -13,     62,    -89,     83,    -47,     -6,     56,    -87,     86,    -54,      2,     50,    -84,     88,    -60,     10,     43,    -81,     89,    -65,     18,     36,    -77,     90,    -70,     25,     29,    -73,     90,    -75,     32,     22,    -68,     90,    -79,     39,     14,    -63,     89,    -82,     46,      7,    -57,     87,    -85,     53,     -1,    -51,     85,    -88,     59,     -9,    -44,     81,    -89,     64,    -16,    -37,     78,    -90,     70,    -24,    -30,     74,    -90,     74,    -31,    -23,     69,    -90,     78,    -38,    -15,     64,    -89,     82,    -45,     -8,     58,    -87,     85,    -52 },
      {     26,    -70,     90,    -80,     43,      8,    -56,     86,    -87,     59,    -11,    -40,     78,    -90,     72,    -29,    -23,     67,    -89,     81,    -46,     -4,     54,    -85,     88,    -61,     14,     37,    -77,     90,    -74,     32,     20,    -65,     89,    -83,     49,      1,    -51,     84,    -89,     64,    -18,    -34,     75,    -90,     76,    -35,    -16,     63,    -88,     84,    -52,      2,     48,    -82,     89,    -66,     21,     31,    -73,     90,    -77,     38,     13,    -60,     88,    -85,     55,     -6,    -45,     81,    -90,     68,    -24,    -28,     71,    -90,     79,    -41,    -10,     58,    -87,     86,    -57,      9,     42,    -79,     90,    -70,     27,     25,    -69,     90,    -81,     44,      7,    -55,     86,    -87,     60,    -12,    -39,     78,    -90,     72,    -30,    -22,     67,    -89,     82,    -47,     -3,     53,    -85,     88,    -62,     15,     36,    -76,     90,    -74,     33,     19,    -64,     89,    -83,     50 },
      {     25,    -67,     89,    -83,     52,     -4,    -44,     79,    -90,     73,    -33,    -16,     61,    -87,     86,    -59,     13,     36,    -75,     90,    -78,     41,      8,    -55,     85,    -89,     65,    -22,    -28,     70,    -90,     82,    -49,      1,     47,    -81,     90,    -71,     30,     20,    -64,     88,    -85,     56,    -10,    -39,     77,    -90,     76,    -38,    -11,     57,    -86,     88,    -63,     19,     31,    -72,     90,    -81,     46,      2,    -50,     82,    -89,     69,    -27,    -23,     66,    -89,     84,    -54,      7,     42,    -78,     90,    -74,     35,     14,    -60,     87,    -87,     60,    -15,    -34,     74,    -90,     79,    -43,     -6,     53,    -84,     89,    -67,     24,     26,    -68,     89,    -83,     51,     -3,    -45,     80,    -90,     72,    -32,    -18,     62,    -88,     86,    -58,     12,     37,    -76,     90,    -77,     40,      9,    -55,     85,    -88,     64,    -21,    -29,     70,    -90,     81,    -48 },
      {     24,    -65,     88,    -86,     60,    -16,    -31,     70,    -89,     83,    -54,      9,     38,    -75,     90,    -80,     47,     -1,    -45,     79,    -90,     76,    -40,     -7,     52,    -82,     90,    -72,     33,     14,    -58,     85,    -89,     67,    -26,    -22,     64,    -88,     87,    -61,     19,     29,    -69,     89,    -84,     55,    -11,    -36,     74,    -90,     81,    -49,      3,     43,    -78,     90,    -77,     42,      4,    -50,     81,    -90,     73,    -35,    -12,     56,    -85,     89,    -68,     28,     20,    -62,     87,    -87,     63,    -21,    -27,     67,    -89,     85,    -57,     13,     34,    -72,     90,    -82,     51,     -6,    -41,     77,    -90,     78,    -44,     -2,     48,    -81,     90,    -74,     37,     10,    -55,     84,    -89,     70,    -30,    -18,     60,    -86,     88,    -64,     23,     25,    -66,     88,    -86,     59,    -15,    -32,     71,    -90,     83,    -53,      8,     39,    -76,     90,    -79,     46 },
      {     23,    -63,     87,    -88,     67,    -28,    -18,     59,    -85,     89,    -70,     33,     12,    -55,     83,    -90,     74,    -38,     -7,     50,    -81,     90,    -77,     43,      1,    -45,     78,    -90,     79,    -48,      4,     40,    -75,     90,    -82,     53,    -10,    -35,     72,    -89,     84,    -57,     15,     30,    -68,     89,    -86,     61,    -21,    -25,     64,    -87,     88,    -65,     26,     20,    -60,     86,    -89,     69,    -31,    -14,     56,    -84,     90,    -72,     36,      9,    -52,     81,    -90,     76,    -41,     -3,     47,    -79,     90,    -78,     46,     -2,    -42,     76,    -90,     81,    -51,      8,     37,    -73,     90,    -83,     55,    -13,    -32,     70,    -89,     85,    -60,     19,     27,    -66,     88,    -87,     64,    -24,    -22,     62,    -86,     88,    -67,     29,     16,    -58,     85,    -89,     71,    -34,    -11,     54,    -82,     90,    -74,     39,      6,    -49,     80,    -90,     77,    -44 },
      {     22,    -60,     85,    -89,     73,    -39,     -3,     45,    -77,     90,    -82,     55,    -15,    -28,     65,    -87,     88,    -69,     33,     10,    -51,     80,    -90,     79,    -50,      9,     34,    -70,     89,    -87,     64,    -27,    -16,     56,    -83,     90,    -76,     44,     -2,    -40,     74,    -90,     85,    -60,     21,     23,    -61,     85,    -89,     72,    -38,     -4,     46,    -77,     90,    -82,     55,    -14,    -29,     66,    -87,     88,    -68,     32,     11,    -52,     81,    -90,     79,    -49,      8,     35,    -70,     89,    -86,     64,    -26,    -18,     57,    -83,     90,    -76,     43,     -1,    -41,     74,    -90,     84,    -59,     20,     24,    -62,     86,    -89,     72,    -37,     -6,     47,    -78,     90,    -81,     54,    -13,    -30,     67,    -88,     88,    -67,     31,     12,    -53,     81,    -90,     78,    -48,      7,     36,    -71,     89,    -86,     63,    -25,    -19,     58,    -84,     90,    -75,     42 },
      {     21,    -58,     83,    -90,     78,    -50,     11,     30,    -65,     86,    -89,     73,    -41,      1,     39,    -72,     89,    -87,     67,    -32,     -9,     48,    -77,     90,    -84,     60,    -23,    -19,     56,    -82,     90,    -79,     52,    -13,    -28,     64,    -86,     89,    -74,     43,     -3,    -37,     70,    -88,     88,    -68,     34,      7,    -46,     76,    -90,     85,    -61,     25,     16,    -55,     81,    -90,     81,    -54,     15,     26,    -62,     85,    -90,     76,    -45,      6,     35,    -69,     88,    -88,     70,    -36,     -4,     44,    -75,     90,    -85,     63,    -27,    -14,     53,    -80,     90,    -81,     55,    -18,    -24,     60,    -84,     90,    -77,     47,     -8,    -33,     67,    -87,     89,    -71,     38,      2,    -42,     74,    -89,     86,    -64,     29,     12,    -51,     79,    -90,     82,    -57,     20,     22,    -59,     83,    -90,     78,    -49,     10,     31,    -66,     87,    -89,     72,    -40 },
      {     20,    -55,     81,    -90,     83,    -60,     25,     14,    -51,     78,    -90,     85,    -64,     30,      9,    -46,     75,    -89,     87,    -67,     35,      3,    -41,     72,    -88,     88,    -71,     40,     -2,    -36,     68,    -87,     89,    -74,     45,     -8,    -31,     64,    -85,     90,    -77,     50,    -13,    -26,     60,    -83,     90,    -80,     55,    -19,    -21,     56,    -81,     90,    -82,     59,    -24,    -15,     52,    -78,     90,    -85,     63,    -29,    -10,     47,    -76,     89,    -86,     67,    -34,     -4,     42,    -72,     89,    -88,     70,    -39,      1,     37,    -69,     87,    -89,     74,    -44,      7,     32,    -65,     86,    -90,     77,    -49,     12,     27,    -61,     84,    -90,     79,    -54,     18,     22,    -57,     81,    -90,     82,    -58,     23,     16,    -53,     79,    -90,     84,    -62,     28,     11,    -48,     76,    -90,     86,    -66,     33,      6,    -43,     73,    -89,     88,    -70,     38 },
      {     19,    -53,     78,    -90,     86,    -68,     38,     -2,    -34,     65,    -85,     90,    -80,     56,    -23,    -14,     49,    -76,     89,    -88,     71,    -42,      7,     30,    -62,     83,    -90,     82,    -60,     27,     10,    -45,     73,    -88,     89,    -74,     46,    -11,    -26,     59,    -81,     90,    -84,     63,    -31,     -6,     41,    -70,     87,    -89,     76,    -50,     15,     22,    -55,     79,    -90,     85,    -66,     35,      1,    -37,     67,    -86,     90,    -78,     54,    -20,    -18,     52,    -77,     90,    -87,     69,    -39,      3,     33,    -64,     85,    -90,     81,    -57,     24,     13,    -48,     75,    -89,     88,    -72,     43,     -8,    -29,     61,    -83,     90,    -82,     60,    -28,     -9,     44,    -72,     88,    -89,     74,    -47,     12,     25,    -58,     81,    -90,     84,    -64,     32,      4,    -40,     70,    -87,     89,    -77,     51,    -16,    -21,     55,    -79,     90,    -86,     67,    -36 },
      {     18,    -50,     75,    -89,     89,    -76,     51,    -19,    -16,     49,    -74,     88,    -89,     76,    -52,     20,     15,    -48,     74,    -88,     89,    -77,     53,    -21,    -14,     47,    -73,     88,    -89,     77,    -54,     22,     13,    -46,     72,    -88,     89,    -78,     55,    -23,    -12,     45,    -72,     87,    -90,     78,    -55,     24,     11,    -44,     71,    -87,     90,    -79,     56,    -25,    -10,     43,    -70,     87,    -90,     79,    -57,     26,      9,    -42,     70,    -86,     90,    -80,     58,    -27,     -8,     41,    -69,     86,    -90,     81,    -59,     28,      7,    -40,     68,    -86,     90,    -81,     60,    -29,     -6,     39,    -67,     85,    -90,     81,    -60,     30,      4,    -38,     67,    -85,     90,    -82,     61,    -31,     -3,     37,    -66,     85,    -90,     82,    -62,     32,      2,    -36,     65,    -84,     90,    -83,     63,    -33,     -1,     35,    -64,     84,    -90,     83,    -64,     34 },
      {     16,    -47,     72,    -87,     90,    -81,     62,    -34,      2,     30,    -59,     79,    -90,     88,    -74,     51,    -21,    -12,     43,    -69,     85,    -90,     83,    -65,     38,     -7,    -26,     55,    -77,     89,    -89,     77,    -55,     25,      8,    -39,     66,    -84,     90,    -85,     68,    -42,     11,     22,    -52,     75,    -88,     89,    -79,     58,    -29,     -3,     35,    -63,     82,    -90,     86,    -71,     46,    -15,    -18,     48,    -72,     87,    -90,     81,    -61,     33,     -1,    -31,     60,    -80,     90,    -88,     74,    -50,     20,     13,    -44,     70,    -86,     90,    -83,     64,    -37,      6,     27,    -56,     78,    -89,     89,    -76,     54,    -24,     -9,     40,    -67,     84,    -90,     85,    -67,     41,    -10,    -23,     53,    -76,     88,    -89,     78,    -57,     28,      4,    -36,     64,    -82,     90,    -86,     70,    -45,     14,     19,    -49,     73,    -87,     90,    -81,     60,    -32 },
      {     15,    -44,     68,    -84,     90,    -86,     72,    -49,     21,     10,    -39,     64,    -82,     90,    -88,     75,    -54,     26,      4,    -34,     60,    -79,     89,    -89,     78,    -58,     31,     -1,    -29,     56,    -77,     88,    -90,     81,    -62,     36,     -7,    -24,     52,    -74,     87,    -90,     83,    -66,     41,    -12,    -19,     47,    -70,     85,    -90,     85,    -70,     46,    -18,    -13,     42,    -67,     83,    -90,     87,    -73,     51,    -23,     -8,     37,    -63,     81,    -90,     88,    -76,     55,    -28,     -2,     32,    -59,     78,    -89,     89,    -79,     60,    -33,      3,     27,    -55,     76,    -88,     90,    -81,     64,    -38,      9,     22,    -50,     72,    -86,     90,    -84,     67,    -43,     14,     16,    -45,     69,    -85,     90,    -86,     71,    -48,     20,     11,    -40,     65,    -82,     90,    -87,     74,    -53,     25,      6,    -35,     61,    -80,     89,    -89,     77,    -57,     30 },
      {     14,    -41,     64,    -81,     89,    -89,     79,    -62,     38,    -11,    -18,     44,    -67,     82,    -90,     88,    -78,     60,    -35,      8,     21,    -47,     69,    -84,     90,    -88,     76,    -57,     32,     -4,    -24,     50,    -71,     85,    -90,     87,    -74,     55,    -29,      1,     27,    -53,     73,    -86,     90,    -86,     72,    -52,     26,      2,    -30,     55,    -75,     87,    -90,     85,    -70,     49,    -23,     -6,     33,    -58,     77,    -88,     90,    -83,     68,    -46,     20,      9,    -36,     60,    -78,     89,    -90,     82,    -66,     43,    -16,    -12,     39,    -63,     80,    -89,     89,    -81,     64,    -40,     13,     15,    -42,     65,    -81,     90,    -89,     79,    -61,     37,    -10,    -19,     45,    -67,     83,    -90,     88,    -77,     59,    -34,      7,     22,    -48,     70,    -84,     90,    -87,     76,    -56,     31,     -3,    -25,     51,    -72,     85,    -90,     86,    -74,     54,    -28 },
      {     13,    -38,     60,    -77,     88,    -90,     85,    -73,     55,    -31,      6,     21,    -45,     66,    -81,     89,    -90,     82,    -68,     48,    -24,     -2,     28,    -52,     71,    -84,     90,    -88,     79,    -63,     41,    -16,    -10,     35,    -58,     76,    -87,     90,    -86,     75,    -57,     34,     -9,    -18,     42,    -64,     79,    -89,     90,    -84,     70,    -51,     27,     -1,    -25,     49,    -69,     83,    -90,     89,    -81,     65,    -44,     20,      7,    -32,     55,    -74,     86,    -90,     87,    -77,     60,    -37,     12,     14,    -39,     61,    -78,     88,    -90,     85,    -72,     54,    -30,      4,     22,    -46,     67,    -81,     89,    -89,     82,    -67,     47,    -23,     -3,     29,    -53,     72,    -85,     90,    -88,     78,    -62,     40,    -15,    -11,     36,    -59,     76,    -87,     90,    -86,     74,    -56,     33,     -8,    -19,     43,    -64,     80,    -89,     90,    -83,     70,    -50,     26 },
      {     12,    -35,     56,    -73,     85,    -90,     89,    -81,     68,    -50,     28,     -4,    -20,     42,    -62,     77,    -87,     90,    -87,     78,    -63,     43,    -21,     -3,     27,    -49,     67,    -81,     89,    -90,     85,    -74,     57,    -36,     13,     11,    -34,     55,    -72,     84,    -90,     89,    -82,     69,    -51,     29,     -6,    -19,     41,    -61,     77,    -87,     90,    -88,     78,    -64,     44,    -22,     -2,     26,    -48,     67,    -81,     89,    -90,     85,    -74,     58,    -37,     14,     10,    -33,     55,    -72,     84,    -90,     89,    -82,     70,    -52,     30,     -7,    -18,     40,    -60,     76,    -86,     90,    -88,     79,    -64,     45,    -23,     -1,     25,    -47,     66,    -80,     88,    -90,     86,    -75,     59,    -38,     15,      9,    -32,     54,    -71,     83,    -90,     89,    -83,     70,    -53,     31,     -8,    -16,     39,    -60,     76,    -86,     90,    -88,     79,    -65,     46,    -24 },
      {     11,    -32,     52,    -68,     81,    -88,     90,    -87,     79,    -66,     49,    -29,      8,     14,    -35,     55,    -70,     82,    -89,     90,    -86,     77,    -64,     46,    -26,      4,     18,    -38,     57,    -72,     83,    -89,     90,    -85,     76,    -61,     43,    -23,      1,     21,    -41,     60,    -74,     85,    -90,     90,    -84,     74,    -59,     40,    -20,     -2,     24,    -44,     62,    -76,     86,    -90,     89,    -83,     72,    -56,     37,    -16,     -6,     27,    -47,     64,    -78,     87,    -90,     89,    -81,     70,    -54,     34,    -13,     -9,     30,    -50,     67,    -79,     88,    -90,     88,    -80,     67,    -51,     31,    -10,    -12,     33,    -53,     69,    -81,     88,    -90,     87,    -78,     65,    -48,     28,     -7,    -15,     36,    -55,     71,    -82,     89,    -90,     86,    -77,     63,    -45,     25,     -3,    -19,     39,    -58,     73,    -84,     89,    -90,     85,    -75,     60,    -42,     22 },
      {     10,    -29,     47,    -63,     76,    -85,     89,    -90,     86,    -78,     67,    -52,     34,    -15,     -4,     24,    -42,     59,    -72,     82,    -89,     90,    -88,     81,    -70,     56,    -39,     21,     -1,    -19,     37,    -55,     69,    -80,     87,    -90,     89,    -83,     74,    -60,     44,    -26,      7,     13,    -32,     50,    -65,     77,    -86,     90,    -90,     85,    -77,     64,    -49,     31,    -12,     -8,     27,    -45,     61,    -74,     84,    -89,     90,    -87,     79,    -68,     54,    -36,     18,      2,    -22,     40,    -57,     71,    -81,     88,    -90,     88,    -82,     72,    -58,     41,    -23,      3,     16,    -35,     53,    -67,     79,    -87,     90,    -89,     84,    -75,     62,    -46,     28,     -9,    -11,     30,    -48,     64,    -76,     85,    -90,     90,    -86,     78,    -66,     51,    -33,     14,      6,    -25,     43,    -60,     73,    -83,     89,    -90,     88,    -81,     70,    -55,     38,    -20 },
      {      9,    -26,     42,    -57,     70,    -79,     86,    -90,     90,    -87,     80,    -70,     58,    -43,     27,    -10,     -8,     25,    -41,     56,    -69,     79,    -86,     90,    -90,     87,    -81,     71,    -59,     44,    -28,     11,      7,    -24,     40,    -55,     68,    -78,     86,    -90,     90,    -87,     81,    -72,     60,    -45,     29,    -12,     -6,     23,    -39,     55,    -67,     78,    -85,     89,    -90,     88,    -81,     72,    -60,     46,    -30,     13,      4,    -22,     38,    -54,     67,    -77,     85,    -89,     90,    -88,     82,    -73,     61,    -47,     31,    -14,     -3,     21,    -37,     53,    -66,     77,    -85,     89,    -90,     88,    -82,     74,    -62,     48,    -32,     15,      2,    -20,     36,    -52,     65,    -76,     84,    -89,     90,    -88,     83,    -74,     63,    -49,     33,    -16,     -1,     19,    -35,     51,    -64,     76,    -84,     89,    -90,     89,    -83,     75,    -64,     50,    -34,     18 },
      {      8,    -23,     37,    -51,     63,    -73,     81,    -87,     90,    -90,     88,    -83,     76,    -67,     55,    -42,     28,    -13,     -2,     18,    -32,     46,    -59,     70,    -78,     85,    -89,     90,    -89,     85,    -79,     70,    -60,     47,    -33,     19,     -3,    -12,     27,    -41,     55,    -66,     76,    -83,     88,    -90,     90,    -87,     81,    -74,     64,    -52,     38,    -24,      9,      7,    -22,     36,    -50,     62,    -72,     81,    -86,     90,    -90,     88,    -84,     77,    -67,     56,    -43,     29,    -14,     -1,     16,    -31,     45,    -58,     69,    -78,     85,    -89,     90,    -89,     86,    -79,     71,    -60,     48,    -34,     20,     -4,    -11,     26,    -40,     54,    -65,     75,    -82,     88,    -90,     90,    -87,     82,    -74,     64,    -53,     39,    -25,     10,      6,    -21,     35,    -49,     61,    -72,     80,    -86,     89,    -90,     89,    -84,     77,    -68,     57,    -44,     30,    -15 },
      {      7,    -20,     32,    -44,     55,    -65,     74,    -81,     86,    -89,     90,    -90,     87,    -83,     77,    -69,     60,    -49,     37,    -25,     12,      1,    -14,     27,    -39,     51,    -61,     70,    -78,     84,    -88,     90,    -90,     89,    -85,     79,    -72,     64,    -54,     42,    -30,     18,     -4,     -9,     22,    -34,     46,    -57,     67,    -75,     81,    -86,     89,    -90,     89,    -87,     82,    -76,     67,    -58,     47,    -35,     23,    -10,     -3,     16,    -29,     41,    -53,     63,    -72,     79,    -85,     88,    -90,     90,    -88,     84,    -78,     71,    -62,     52,    -40,     28,    -15,      2,     11,    -24,     36,    -48,     59,    -68,     76,    -82,     87,    -90,     90,    -89,     86,    -81,     74,    -66,     56,    -45,     33,    -21,      8,      6,    -19,     31,    -43,     55,    -64,     73,    -80,     85,    -89,     90,    -90,     88,    -83,     77,    -70,     60,    -50,     38,    -26,     13 },
      {      6,    -16,     27,    -37,     47,    -56,     64,    -72,     78,    -83,     87,    -89,     90,    -90,     89,    -86,     81,    -76,     70,    -62,     54,    -44,     34,    -24,     13,     -2,     -9,     20,    -30,     40,    -50,     59,    -67,     74,    -79,     84,    -88,     90,    -90,     90,    -88,     85,    -80,     74,    -67,     60,    -51,     41,    -31,     21,    -10,     -1,     12,    -23,     33,    -43,     53,    -61,     69,    -76,     81,    -85,     88,    -90,     90,    -89,     87,    -83,     78,    -72,     65,    -57,     48,    -38,     28,    -18,      7,      4,    -15,     26,    -36,     46,    -55,     64,    -71,     77,    -82,     86,    -89,     90,    -90,     89,    -86,     82,    -77,     70,    -63,     55,    -45,     35,    -25,     14,     -3,     -8,     19,    -29,     39,    -49,     58,    -66,     73,    -79,     84,    -87,     89,    -90,     90,    -88,     85,    -81,     75,    -68,     60,    -52,     42,    -32,     22,    -11 },
      {      4,    -13,     22,    -30,     38,    -46,     54,    -60,     67,    -72,     77,    -81,     85,    -88,     89,    -90,     90,    -89,     88,    -85,     82,    -78,     73,    -67,     61,    -55,     47,    -39,     31,    -23,     14,     -6,     -3,     12,    -21,     29,    -37,     45,    -53,     60,    -66,     72,    -77,     81,    -85,     87,    -89,     90,    -90,     90,    -88,     86,    -82,     78,    -74,     68,    -62,     55,    -48,     40,    -32,     24,    -15,      7,      2,    -11,     20,    -28,     36,    -44,     52,    -59,     65,    -71,     76,    -81,     84,    -87,     89,    -90,     90,    -90,     88,    -86,     83,    -79,     74,    -69,     63,    -56,     49,    -41,     33,    -25,     16,     -8,     -1,     10,    -19,     27,    -35,     43,    -51,     58,    -64,     70,    -76,     80,    -84,     87,    -89,     90,    -90,     90,    -89,     86,    -83,     79,    -75,     70,    -64,     57,    -50,     42,    -34,     26,    -18,      9 },
      {      3,    -10,     16,    -23,     29,    -35,     41,    -47,     53,    -58,     63,    -67,     72,    -76,     79,    -82,     85,    -87,     88,    -89,     90,    -90,     90,    -89,     88,    -86,     84,    -81,     78,    -75,     71,    -67,     62,    -57,     52,    -46,     40,    -34,     28,    -22,     15,     -9,      2,      4,    -11,     18,    -24,     30,    -36,     42,    -48,     54,    -59,     64,    -68,     72,    -76,     79,    -82,     85,    -87,     89,    -90,     90,    -90,     90,    -89,     88,    -86,     84,    -81,     78,    -74,     70,    -66,     61,    -56,     51,    -45,     39,    -33,     27,    -21,     14,     -8,      1,      6,    -12,     19,    -25,     31,    -37,     43,    -49,     55,    -60,     64,    -69,     73,    -77,     80,    -83,     85,    -87,     89,    -90,     90,    -90,     90,    -89,     88,    -86,     83,    -81,     77,    -74,     70,    -65,     60,    -55,     50,    -44,     38,    -32,     26,    -20,     13,     -7 },
      {      2,     -7,     11,    -15,     20,    -24,     28,    -32,     36,    -40,     44,    -48,     52,    -55,     59,    -62,     65,    -68,     71,    -74,     76,    -78,     81,    -82,     84,    -86,     87,    -88,     89,    -90,     90,    -90,     90,    -90,     90,    -89,     88,    -87,     86,    -85,     83,    -81,     79,    -77,     74,    -72,     69,    -66,     63,    -60,     56,    -53,     49,    -45,     41,    -37,     33,    -29,     25,    -21,     16,    -12,      8,     -3,     -1,      6,    -10,     14,    -19,     23,    -27,     31,    -35,     39,    -43,     47,    -51,     55,    -58,     61,    -64,     67,    -70,     73,    -76,     78,    -80,     82,    -84,     85,    -87,     88,    -89,     89,    -90,     90,    -90,     90,    -90,     89,    -89,     88,    -86,     85,    -83,     81,    -79,     77,    -75,     72,    -70,     67,    -64,     60,    -57,     54,    -50,     46,    -42,     38,    -34,     30,    -26,     22,    -18,     13,     -9,      4 },
      {      1,     -3,      6,     -8,     10,    -12,     14,    -16,     19,    -21,     23,    -25,     27,    -29,     31,    -33,     35,    -37,     39,    -41,     43,    -45,     47,    -49,     51,    -53,     55,    -56,     58,    -60,     61,    -63,     64,    -66,     67,    -69,     70,    -72,     73,    -74,     76,    -77,     78,    -79,     80,    -81,     82,    -83,     84,    -85,     85,    -86,     87,    -87,     88,    -88,     89,    -89,     89,    -90,     90,    -90,     90,    -90,     90,    -90,     90,    -90,     90,    -90,     89,    -89,     89,    -88,     88,    -87,     86,    -86,     85,    -84,     83,    -82,     81,    -81,     79,    -78,     77,    -76,     75,    -74,     72,    -71,     70,    -68,     67,    -65,     64,    -62,     60,    -59,     57,    -55,     54,    -52,     50,    -48,     46,    -44,     42,    -40,     38,    -36,     34,    -32,     30,    -28,     26,    -24,     22,    -20,     18,    -15,     13,    -11,      9,     -7,      4,     -2 },
    },
};
#endif

// DST-7
const TMatrixCoeff g_trCoreDST7P4[TRANSFORM_NUMBER_OF_DIRECTIONS][4][4] =
{
  DEFINE_DST7_P4_MATRIX(29,    55,    74,    84),
  DEFINE_DST7_P4_MATRIX(29,    55,    74,    84)
};
const TMatrixCoeff g_trCoreDST7P8[TRANSFORM_NUMBER_OF_DIRECTIONS][8][8] =
{
  DEFINE_DST7_P8_MATRIX(17,    32,    46,    60,    71,    78,    85,    86),
  DEFINE_DST7_P8_MATRIX(17,    32,    46,    60,    71,    78,    85,    86)
};
const TMatrixCoeff g_trCoreDST7P16[TRANSFORM_NUMBER_OF_DIRECTIONS][16][16] =
{
  DEFINE_DST7_P16_MATRIX(8,    17,    25,    33,    40,    48,    55,    62,    68,    73,    77,    81,    85,    87,    88,    88),
  DEFINE_DST7_P16_MATRIX(8,    17,    25,    33,    40,    48,    55,    62,    68,    73,    77,    81,    85,    87,    88,    88)
};
const TMatrixCoeff g_trCoreDST7P32[TRANSFORM_NUMBER_OF_DIRECTIONS][32][32] =
{
  DEFINE_DST7_P32_MATRIX(4,     9,    13,    17,    21,    26,    30,    34,    38,    42,    46,    50,    53,    56,    60,    63,    66,    68,    72,    74,    77,    78,    80,    82,    84,    85,    86,    87,    88,    89,    90,    90),
  DEFINE_DST7_P32_MATRIX(4,     9,    13,    17,    21,    26,    30,    34,    38,    42,    46,    50,    53,    56,    60,    63,    66,    68,    72,    74,    77,    78,    80,    82,    84,    85,    86,    87,    88,    89,    90,    90)
};
#if TU_256
const TMatrixCoeff g_trCoreDST7P64[TRANSFORM_NUMBER_OF_DIRECTIONS][64][64] =
{
    {
      {      2,      4,      7,      9,     11,     13,     15,     17,     20,     22,     24,     26,     28,     30,     32,     34,     36,     38,     40,     42,     44,     46,     48,     50,     52,     53,     55,     57,     59,     60,     62,     63,     65,     66,     68,     69,     71,     72,     73,     75,     76,     77,     78,     79,     80,     81,     82,     83,     84,     85,     85,     86,     87,     87,     88,     88,     89,     89,     89,     90,     90,     90,     90,     90 },
      {      7,     13,     20,     26,     32,     38,     44,     50,     55,     60,     65,     69,     73,     77,     80,     83,     85,     87,     89,     90,     90,     90,     90,     89,     87,     85,     83,     80,     77,     73,     69,     65,     60,     55,     50,     44,     38,     32,     26,     20,     13,      7,      0,     -7,    -13,    -20,    -26,    -32,    -38,    -44,    -50,    -55,    -60,    -65,    -69,    -73,    -77,    -80,    -83,    -85,    -87,    -89,    -90,    -90 },
      {     11,     22,     32,     42,     52,     60,     68,     75,     80,     85,     88,     90,     90,     89,     87,     84,     79,     73,     66,     59,     50,     40,     30,     20,      9,     -2,    -13,    -24,    -34,    -44,    -53,    -62,    -69,    -76,    -81,    -85,    -88,    -90,    -90,    -89,    -87,    -83,    -78,    -72,    -65,    -57,    -48,    -38,    -28,    -17,     -7,      4,     15,     26,     36,     46,     55,     63,     71,     77,     82,     86,     89,     90 },
      {     15,     30,     44,     57,     68,     77,     84,     88,     90,     89,     86,     80,     72,     62,     50,     36,     22,      7,     -9,    -24,    -38,    -52,    -63,    -73,    -81,    -87,    -90,    -90,    -88,    -83,    -76,    -66,    -55,    -42,    -28,    -13,      2,     17,     32,     46,     59,     69,     78,     85,     89,     90,     89,     85,     79,     71,     60,     48,     34,     20,      4,    -11,    -26,    -40,    -53,    -65,    -75,    -82,    -87,    -90 },
      {     20,     38,     55,     69,     80,     87,     90,     89,     83,     73,     60,     44,     26,      7,    -13,    -32,    -50,    -65,    -77,    -85,    -90,    -90,    -85,    -77,    -65,    -50,    -32,    -13,      7,     26,     44,     60,     73,     83,     89,     90,     87,     80,     69,     55,     38,     20,      0,    -20,    -38,    -55,    -69,    -80,    -87,    -90,    -89,    -83,    -73,    -60,    -44,    -26,     -7,     13,     32,     50,     65,     77,     85,     90 },
      {     24,     46,     65,     79,     88,     90,     86,     76,     60,     40,     17,     -7,    -30,    -52,    -69,    -82,    -89,    -90,    -84,    -72,    -55,    -34,    -11,     13,     36,     57,     73,     85,     90,     89,     81,     68,     50,     28,      4,    -20,    -42,    -62,    -77,    -87,    -90,    -87,    -78,    -63,    -44,    -22,      2,     26,     48,     66,     80,     88,     90,     85,     75,     59,     38,     15,     -9,    -32,    -53,    -71,    -83,    -89 },
      {     28,     53,     73,     86,     90,     85,     72,     52,     26,     -2,    -30,    -55,    -75,    -87,    -90,    -85,    -71,    -50,    -24,      4,     32,     57,     76,     87,     90,     84,     69,     48,     22,     -7,    -34,    -59,    -77,    -88,    -90,    -83,    -68,    -46,    -20,      9,     36,     60,     78,     88,     90,     82,     66,     44,     17,    -11,    -38,    -62,    -79,    -89,    -89,    -81,    -65,    -42,    -15,     13,     40,     63,     80,     89 },
      {     32,     60,     80,     90,     87,     73,     50,     20,    -13,    -44,    -69,    -85,    -90,    -83,    -65,    -38,     -7,     26,     55,     77,     89,     89,     77,     55,     26,     -7,    -38,    -65,    -83,    -90,    -85,    -69,    -44,    -13,     20,     50,     73,     87,     90,     80,     60,     32,      0,    -32,    -60,    -80,    -90,    -87,    -73,    -50,    -20,     13,     44,     69,     85,     90,     83,     65,     38,      7,    -26,    -55,    -77,    -89 },
      {     36,     66,     85,     90,     79,     55,     22,    -15,    -50,    -76,    -89,    -87,    -71,    -42,     -7,     30,     62,     83,     90,     82,     60,     28,     -9,    -44,    -72,    -88,    -89,    -75,    -48,    -13,     24,     57,     80,     90,     85,     65,     34,     -2,    -38,    -68,    -86,    -90,    -78,    -53,    -20,     17,     52,     77,     89,     87,     69,     40,      4,    -32,    -63,    -84,    -90,    -81,    -59,    -26,     11,     46,     73,     88 },
      {     40,     72,     89,     87,     66,     32,     -9,    -48,    -77,    -90,    -84,    -60,    -24,     17,     55,     81,     90,     80,     53,     15,    -26,    -62,    -85,    -90,    -76,    -46,     -7,     34,     68,     87,     88,     71,     38,     -2,    -42,    -73,    -89,    -86,    -65,    -30,     11,     50,     78,     90,     83,     59,     22,    -20,    -57,    -82,    -90,    -79,    -52,    -13,     28,     63,     85,     89,     75,     44,      4,    -36,    -69,    -88 },
      {     44,     77,     90,     80,     50,      7,    -38,    -73,    -90,    -83,    -55,    -13,     32,     69,     89,     85,     60,     20,    -26,    -65,    -87,    -87,    -65,    -26,     20,     60,     85,     89,     69,     32,    -13,    -55,    -83,    -90,    -73,    -38,      7,     50,     80,     90,     77,     44,      0,    -44,    -77,    -90,    -80,    -50,     -7,     38,     73,     90,     83,     55,     13,    -32,    -69,    -89,    -85,    -60,    -20,     26,     65,     87 },
      {     48,     81,     90,     71,     30,    -20,    -63,    -88,    -85,    -57,    -11,     38,     76,     90,     77,     40,     -9,    -55,    -85,    -88,    -65,    -22,     28,     69,     89,     82,     50,      2,    -46,    -80,    -90,    -72,    -32,     17,     62,     87,     86,     59,     13,    -36,    -75,    -90,    -78,    -42,      7,     53,     84,     89,     66,     24,    -26,    -68,    -89,    -83,    -52,     -4,     44,     79,     90,     73,     34,    -15,    -60,    -87 },
      {     52,     85,     87,     59,      9,    -44,    -81,    -89,    -65,    -17,     36,     77,     90,     71,     26,    -28,    -72,    -90,    -76,    -34,     20,     66,     89,     80,     42,    -11,    -60,    -88,    -84,    -50,      2,     53,     85,     87,     57,      7,    -46,    -82,    -89,    -63,    -15,     38,     78,     90,     69,     24,    -30,    -73,    -90,    -75,    -32,     22,     68,     90,     79,     40,    -13,    -62,    -88,    -83,    -48,      4,     55,     86 },
      {     55,     87,     83,     44,    -13,    -65,    -90,    -77,    -32,     26,     73,     90,     69,     20,    -38,    -80,    -89,    -60,     -7,     50,     85,     85,     50,     -7,    -60,    -89,    -80,    -38,     20,     69,     90,     73,     26,    -32,    -77,    -90,    -65,    -13,     44,     83,     87,     55,      0,    -55,    -87,    -83,    -44,     13,     65,     90,     77,     32,    -26,    -73,    -90,    -69,    -20,     38,     80,     89,     60,      7,    -50,    -85 },
      {     59,     89,     77,     28,    -34,    -80,    -88,    -53,      7,     63,     90,     73,     22,    -40,    -83,    -86,    -48,     13,     68,     90,     69,     15,    -46,    -85,    -84,    -42,     20,     72,     90,     65,      9,    -52,    -87,    -81,    -36,     26,     76,     89,     60,      2,    -57,    -89,    -78,    -30,     32,     79,     88,     55,     -4,    -62,    -90,    -75,    -24,     38,     82,     87,     50,    -11,    -66,    -90,    -71,    -17,     44,     85 },
      {     62,     90,     69,     11,    -53,    -89,    -76,    -22,     44,     86,     81,     32,    -34,    -82,    -85,    -42,     24,     77,     88,     52,    -13,    -71,    -90,    -60,      2,     63,     90,     68,      9,    -55,    -89,    -75,    -20,     46,     87,     80,     30,    -36,    -83,    -85,    -40,     26,     78,     88,     50,    -15,    -72,    -90,    -59,      4,     65,     90,     66,      7,    -57,    -89,    -73,    -17,     48,     87,     79,     28,    -38,    -84 },
      {     65,     90,     60,     -7,    -69,    -90,    -55,     13,     73,     89,     50,    -20,    -77,    -87,    -44,     26,     80,     85,     38,    -32,    -83,    -83,    -32,     38,     85,     80,     26,    -44,    -87,    -77,    -20,     50,     89,     73,     13,    -55,    -90,    -69,     -7,     60,     90,     65,      0,    -65,    -90,    -60,      7,     69,     90,     55,    -13,    -73,    -89,    -50,     20,     77,     87,     44,    -26,    -80,    -85,    -38,     32,     83 },
      {     68,     89,     50,    -24,    -81,    -83,    -28,     46,     89,     71,      4,    -65,    -90,    -53,     20,     79,     85,     32,    -42,    -88,    -73,     -9,     62,     90,     57,    -15,    -77,    -86,    -36,     38,     87,     76,     13,    -59,    -90,    -60,     11,     75,     87,     40,    -34,    -85,    -78,    -17,     55,     90,     63,     -7,    -72,    -88,    -44,     30,     84,     80,     22,    -52,    -90,    -66,      2,     69,     89,     48,    -26,    -82 },
      {     71,     88,     38,    -40,    -88,    -69,      2,     72,     87,     36,    -42,    -89,    -68,      4,     73,     87,     34,    -44,    -89,    -66,      7,     75,     86,     32,    -46,    -89,    -65,      9,     76,     85,     30,    -48,    -90,    -63,     11,     77,     85,     28,    -50,    -90,    -62,     13,     78,     84,     26,    -52,    -90,    -60,     15,     79,     83,     24,    -53,    -90,    -59,     17,     80,     82,     22,    -55,    -90,    -57,     20,     81 },
      {     73,     85,     26,    -55,    -90,    -50,     32,     87,     69,     -7,    -77,    -83,    -20,     60,     90,     44,    -38,    -89,    -65,     13,     80,     80,     13,    -65,    -89,    -38,     44,     90,     60,    -20,    -83,    -77,     -7,     69,     87,     32,    -50,    -90,    -55,     26,     85,     73,      0,    -73,    -85,    -26,     55,     90,     50,    -32,    -87,    -69,      7,     77,     83,     20,    -60,    -90,    -44,     38,     89,     65,    -13,    -80 },
      {     76,     82,     13,    -68,    -87,    -26,     59,     89,     38,    -48,    -90,    -50,     36,     89,     60,    -24,    -86,    -69,     11,     81,     77,      2,    -75,    -83,    -15,     66,     87,     28,    -57,    -90,    -40,     46,     90,     52,    -34,    -89,    -62,     22,     85,     71,     -9,    -80,    -78,     -4,     73,     84,     17,    -65,    -88,    -30,     55,     90,     42,    -44,    -90,    -53,     32,     88,     63,    -20,    -85,    -72,      7,     79 },
      {     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78 },
      {     80,     73,    -13,    -85,    -65,     26,     89,     55,    -38,    -90,    -44,     50,     90,     32,    -60,    -87,    -20,     69,     83,      7,    -77,    -77,      7,     83,     69,    -20,    -87,    -60,     32,     90,     50,    -44,    -90,    -38,     55,     89,     26,    -65,    -85,    -13,     73,     80,      0,    -80,    -73,     13,     85,     65,    -26,    -89,    -55,     38,     90,     44,    -50,    -90,    -32,     60,     87,     20,    -69,    -83,     -7,     77 },
      {     82,     68,    -26,    -89,    -48,     50,     89,     24,    -69,    -81,      2,     83,     66,    -28,    -90,    -46,     52,     89,     22,    -71,    -80,      4,     84,     65,    -30,    -90,    -44,     53,     88,     20,    -72,    -79,      7,     85,     63,    -32,    -90,    -42,     55,     88,     17,    -73,    -78,      9,     85,     62,    -34,    -90,    -40,     57,     87,     15,    -75,    -77,     11,     86,     60,    -36,    -90,    -38,     59,     87,     13,    -76 },
      {     84,     62,    -38,    -90,    -28,     69,     79,    -11,    -87,    -53,     48,     89,     17,    -76,    -73,     22,     89,     44,    -57,    -86,     -7,     81,     66,    -32,    -90,    -34,     65,     82,     -4,    -85,    -59,     42,     90,     24,    -72,    -77,     15,     88,     50,    -52,    -88,    -13,     78,     71,    -26,    -90,    -40,     60,     85,      2,    -83,    -63,     36,     90,     30,    -68,    -80,      9,     87,     55,    -46,    -89,    -20,     75 },
      {     85,     55,    -50,    -87,     -7,     83,     60,    -44,    -89,    -13,     80,     65,    -38,    -90,    -20,     77,     69,    -32,    -90,    -26,     73,     73,    -26,    -90,    -32,     69,     77,    -20,    -90,    -38,     65,     80,    -13,    -89,    -44,     60,     83,     -7,    -87,    -50,     55,     85,      0,    -85,    -55,     50,     87,      7,    -83,    -60,     44,     89,     13,    -80,    -65,     38,     90,     20,    -77,    -69,     32,     90,     26,    -73 },
      {     87,     48,    -60,    -81,     15,     90,     34,    -71,    -73,     30,     90,     20,    -79,    -63,     44,     88,      4,    -85,    -52,     57,     83,    -11,    -89,    -38,     68,     76,    -26,    -90,    -24,     77,     66,    -40,    -89,     -9,     84,     55,    -53,    -85,      7,     88,     42,    -65,    -78,     22,     90,     28,    -75,    -69,     36,     89,     13,    -82,    -59,     50,     86,     -2,    -87,    -46,     62,     80,    -17,    -90,    -32,     72 },
      {     88,     40,    -69,    -72,     36,     89,      4,    -87,    -44,     66,     75,    -32,    -89,     -9,     85,     48,    -63,    -77,     28,     90,     13,    -84,    -52,     60,     79,    -24,    -90,    -17,     82,     55,    -57,    -81,     20,     90,     22,    -80,    -59,     53,     83,    -15,    -90,    -26,     78,     62,    -50,    -85,     11,     90,     30,    -76,    -65,     46,     86,     -7,    -89,    -34,     73,     68,    -42,    -87,      2,     88,     38,    -71 },
      {     89,     32,    -77,    -60,     55,     80,    -26,    -90,     -7,     87,     38,    -73,    -65,     50,     83,    -20,    -90,    -13,     85,     44,    -69,    -69,     44,     85,    -13,    -90,    -20,     83,     50,    -65,    -73,     38,     87,     -7,    -90,    -26,     80,     55,    -60,    -77,     32,     89,      0,    -89,    -32,     77,     60,    -55,    -80,     26,     90,      7,    -87,    -38,     73,     65,    -50,    -83,     20,     90,     13,    -85,    -44,     69 },
      {     89,     24,    -83,    -46,     71,     65,    -53,    -79,     32,     88,     -9,    -90,    -15,     86,     38,    -76,    -59,     60,     75,    -40,    -85,     17,     90,      7,    -88,    -30,     80,     52,    -66,    -69,     48,     82,    -26,    -89,      2,     90,     22,    -84,    -44,     72,     63,    -55,    -78,     34,     87,    -11,    -90,    -13,     87,     36,    -77,    -57,     62,     73,    -42,    -85,     20,     90,      4,    -89,    -28,     81,     50,    -68 },
      {     90,     15,    -87,    -30,     82,     44,    -75,    -57,     65,     68,    -53,    -77,     40,     84,    -26,    -88,     11,     90,      4,    -89,    -20,     86,     34,    -80,    -48,     72,     60,    -62,    -71,     50,     79,    -36,    -85,     22,     89,     -7,    -90,     -9,     89,     24,    -85,    -38,     78,     52,    -69,    -63,     59,     73,    -46,    -81,     32,     87,    -17,    -90,      2,     90,     13,    -88,    -28,     83,     42,    -76,    -55,     66 },
      {     90,      7,    -90,    -13,     89,     20,    -87,    -26,     85,     32,    -83,    -38,     80,     44,    -77,    -50,     73,     55,    -69,    -60,     65,     65,    -60,    -69,     55,     73,    -50,    -77,     44,     80,    -38,    -83,     32,     85,    -26,    -87,     20,     89,    -13,    -90,      7,     90,      0,    -90,     -7,     90,     13,    -89,    -20,     87,     26,    -85,    -32,     83,     38,    -80,    -44,     77,     50,    -73,    -55,     69,     60,    -65 },
      {     90,     -2,    -90,      4,     90,     -7,    -90,      9,     90,    -11,    -89,     13,     89,    -15,    -89,     17,     88,    -20,    -88,     22,     87,    -24,    -87,     26,     86,    -28,    -85,     30,     85,    -32,    -84,     34,     83,    -36,    -82,     38,     81,    -40,    -80,     42,     79,    -44,    -78,     46,     77,    -48,    -76,     50,     75,    -52,    -73,     53,     72,    -55,    -71,     57,     69,    -59,    -68,     60,     66,    -62,    -65,     63 },
      {     90,    -11,    -89,     22,     86,    -32,    -82,     42,     77,    -52,    -71,     60,     63,    -68,    -55,     75,     46,    -80,    -36,     85,     26,    -88,    -15,     90,      4,    -90,      7,     89,    -17,    -87,     28,     84,    -38,    -79,     48,     73,    -57,    -66,     65,     59,    -72,    -50,     78,     40,    -83,    -30,     87,     20,    -89,     -9,     90,     -2,    -90,     13,     88,    -24,    -85,     34,     81,    -44,    -76,     53,     69,    -62 },
      {     90,    -20,    -85,     38,     77,    -55,    -65,     69,     50,    -80,    -32,     87,     13,    -90,      7,     89,    -26,    -83,     44,     73,    -60,    -60,     73,     44,    -83,    -26,     89,      7,    -90,     13,     87,    -32,    -80,     50,     69,    -65,    -55,     77,     38,    -85,    -20,     90,      0,    -90,     20,     85,    -38,    -77,     55,     65,    -69,    -50,     80,     32,    -87,    -13,     90,     -7,    -89,     26,     83,    -44,    -73,     60 },
      {     89,    -28,    -80,     53,     63,    -73,    -40,     86,     13,    -90,     15,     85,    -42,    -72,     65,     52,    -81,    -26,     89,     -2,    -89,     30,     79,    -55,    -62,     75,     38,    -87,    -11,     90,    -17,    -85,     44,     71,    -66,    -50,     82,     24,    -90,      4,     88,    -32,    -78,     57,     60,    -76,    -36,     87,      9,    -90,     20,     84,    -46,    -69,     68,     48,    -83,    -22,     90,     -7,    -88,     34,     77,    -59 },
      {     88,    -36,    -73,     66,     46,    -85,    -11,     90,    -26,    -79,     59,     55,    -81,    -22,     90,    -15,    -84,     50,     63,    -76,    -32,     89,     -4,    -87,     40,     71,    -69,    -42,     87,      7,    -89,     30,     77,    -62,    -52,     83,     17,    -90,     20,     82,    -53,    -60,     78,     28,    -90,      9,     86,    -44,    -68,     72,     38,    -88,     -2,     89,    -34,    -75,     65,     48,    -85,    -13,     90,    -24,    -80,     57 },
      {     87,    -44,    -65,     77,     26,    -90,     20,     80,    -60,    -50,     85,      7,    -89,     38,     69,    -73,    -32,     90,    -13,    -83,     55,     55,    -83,    -13,     90,    -32,    -73,     69,     38,    -89,      7,     85,    -50,    -60,     80,     20,    -90,     26,     77,    -65,    -44,     87,      0,    -87,     44,     65,    -77,    -26,     90,    -20,    -80,     60,     50,    -85,     -7,     89,    -38,    -69,     73,     32,    -90,     13,     83,    -55 },
      {     86,    -52,    -55,     85,      4,    -87,     48,     59,    -83,     -9,     88,    -44,    -62,     81,     13,    -89,     40,     65,    -79,    -17,     90,    -36,    -68,     77,     22,    -90,     32,     71,    -75,    -26,     90,    -28,    -73,     72,     30,    -90,     24,     76,    -69,    -34,     90,    -20,    -78,     66,     38,    -89,     15,     80,    -63,    -42,     89,    -11,    -82,     60,     46,    -88,      7,     84,    -57,    -50,     87,     -2,    -85,     53 },
      {     85,    -59,    -44,     89,    -17,    -77,     71,     28,    -90,     34,     66,    -80,    -11,     88,    -50,    -53,     87,     -7,    -82,     63,     38,    -90,     24,     73,    -75,    -22,     90,    -40,    -62,     83,      4,    -86,     55,     48,    -88,     13,     79,    -68,    -32,     90,    -30,    -69,     78,     15,    -89,     46,     57,    -85,      2,     84,    -60,    -42,     89,    -20,    -76,     72,     26,    -90,     36,     65,    -81,     -9,     87,    -52 },
      {     83,    -65,    -32,     90,    -38,    -60,     85,     -7,    -80,     69,     26,    -90,     44,     55,    -87,     13,     77,    -73,    -20,     89,    -50,    -50,     89,    -20,    -73,     77,     13,    -87,     55,     44,    -90,     26,     69,    -80,     -7,     85,    -60,    -38,     90,    -32,    -65,     83,      0,    -83,     65,     32,    -90,     38,     60,    -85,      7,     80,    -69,    -26,     90,    -44,    -55,     87,    -13,    -77,     73,     20,    -89,     50 },
      {     81,    -71,    -20,     88,    -57,    -38,     90,    -40,    -55,     88,    -22,    -69,     82,     -2,    -80,     72,     17,    -87,     59,     36,    -90,     42,     53,    -89,     24,     68,    -83,      4,     79,    -73,    -15,     87,    -60,    -34,     90,    -44,    -52,     89,    -26,    -66,     84,     -7,    -78,     75,     13,    -86,     62,     32,    -90,     46,     50,    -89,     28,     65,    -85,      9,     77,    -76,    -11,     85,    -63,    -30,     90,    -48 },
      {     79,    -76,     -7,     82,    -72,    -13,     85,    -68,    -20,     87,    -63,    -26,     88,    -59,    -32,     89,    -53,    -38,     90,    -48,    -44,     90,    -42,    -50,     90,    -36,    -55,     89,    -30,    -60,     88,    -24,    -65,     86,    -17,    -69,     84,    -11,    -73,     81,     -4,    -77,     78,      2,    -80,     75,      9,    -83,     71,     15,    -85,     66,     22,    -87,     62,     28,    -89,     57,     34,    -90,     52,     40,    -90,     46 },
      {     77,    -80,      7,     73,    -83,     13,     69,    -85,     20,     65,    -87,     26,     60,    -89,     32,     55,    -90,     38,     50,    -90,     44,     44,    -90,     50,     38,    -90,     55,     32,    -89,     60,     26,    -87,     65,     20,    -85,     69,     13,    -83,     73,      7,    -80,     77,      0,    -77,     80,     -7,    -73,     83,    -13,    -69,     85,    -20,    -65,     87,    -26,    -60,     89,    -32,    -55,     90,    -38,    -50,     90,    -44 },
      {     75,    -84,     20,     62,    -89,     38,     46,    -90,     55,     28,    -87,     69,      9,    -79,     80,    -11,    -68,     87,    -30,    -53,     90,    -48,    -36,     89,    -63,    -17,     83,    -76,      2,     73,    -85,     22,     60,    -89,     40,     44,    -90,     57,     26,    -86,     71,      7,    -78,     81,    -13,    -66,     88,    -32,    -52,     90,    -50,    -34,     88,    -65,    -15,     82,    -77,      4,     72,    -85,     24,     59,    -90,     42 },
      {     72,    -87,     32,     48,    -90,     60,     17,    -81,     80,    -15,    -62,     90,    -46,    -34,     87,    -71,     -2,     73,    -86,     30,     50,    -90,     59,     20,    -82,     79,    -13,    -63,     89,    -44,    -36,     88,    -69,     -4,     75,    -85,     28,     52,    -90,     57,     22,    -83,     78,    -11,    -65,     89,    -42,    -38,     88,    -68,     -7,     76,    -85,     26,     53,    -90,     55,     24,    -84,     77,     -9,    -66,     89,    -40 },
      {     69,    -89,     44,     32,    -85,     77,    -13,    -60,     90,    -55,    -20,     80,    -83,     26,     50,    -90,     65,      7,    -73,     87,    -38,    -38,     87,    -73,      7,     65,    -90,     50,     26,    -83,     80,    -20,    -55,     90,    -60,    -13,     77,    -85,     32,     44,    -89,     69,      0,    -69,     89,    -44,    -32,     85,    -77,     13,     60,    -90,     55,     20,    -80,     83,    -26,    -50,     90,    -65,     -7,     73,    -87,     38 },
      {     66,    -90,     55,     15,    -76,     87,    -42,    -30,     83,    -82,     28,     44,    -88,     75,    -13,    -57,     90,    -65,     -2,     68,    -90,     53,     17,    -77,     87,    -40,    -32,     84,    -81,     26,     46,    -88,     73,    -11,    -59,     90,    -63,     -4,     69,    -89,     52,     20,    -78,     86,    -38,    -34,     85,    -80,     24,     48,    -89,     72,     -9,    -60,     90,    -62,     -7,     71,    -89,     50,     22,    -79,     85,    -36 },
      {     63,    -90,     65,     -2,    -62,     90,    -66,      4,     60,    -90,     68,     -7,    -59,     90,    -69,      9,     57,    -90,     71,    -11,    -55,     89,    -72,     13,     53,    -89,     73,    -15,    -52,     89,    -75,     17,     50,    -88,     76,    -20,    -48,     88,    -77,     22,     46,    -87,     78,    -24,    -44,     87,    -79,     26,     42,    -86,     80,    -28,    -40,     85,    -81,     30,     38,    -85,     82,    -32,    -36,     84,    -83,     34 },
      {     60,    -90,     73,    -20,    -44,     85,    -83,     38,     26,    -77,     89,    -55,     -7,     65,    -90,     69,    -13,    -50,     87,    -80,     32,     32,    -80,     87,    -50,    -13,     69,    -90,     65,     -7,    -55,     89,    -77,     26,     38,    -83,     85,    -44,    -20,     73,    -90,     60,      0,    -60,     90,    -73,     20,     44,    -85,     83,    -38,    -26,     77,    -89,     55,      7,    -65,     90,    -69,     13,     50,    -87,     80,    -32 },
      {     57,    -88,     80,    -36,    -24,     73,    -90,     66,    -13,    -46,     85,    -85,     48,     11,    -65,     90,    -75,     26,     34,    -79,     89,    -59,      2,     55,    -88,     81,    -38,    -22,     72,    -90,     68,    -15,    -44,     84,    -86,     50,      9,    -63,     90,    -76,     28,     32,    -78,     89,    -60,      4,     53,    -87,     82,    -40,    -20,     71,    -90,     69,    -17,    -42,     83,    -87,     52,      7,    -62,     89,    -77,     30 },
      {     53,    -86,     85,    -52,     -2,     55,    -87,     85,    -50,     -4,     57,    -87,     84,    -48,     -7,     59,    -88,     83,    -46,     -9,     60,    -88,     82,    -44,    -11,     62,    -89,     81,    -42,    -13,     63,    -89,     80,    -40,    -15,     65,    -89,     79,    -38,    -17,     66,    -90,     78,    -36,    -20,     68,    -90,     77,    -34,    -22,     69,    -90,     76,    -32,    -24,     71,    -90,     75,    -30,    -26,     72,    -90,     73,    -28 },
      {     50,    -83,     89,    -65,     20,     32,    -73,     90,    -77,     38,     13,    -60,     87,    -85,     55,     -7,    -44,     80,    -90,     69,    -26,    -26,     69,    -90,     80,    -44,     -7,     55,    -85,     87,    -60,     13,     38,    -77,     90,    -73,     32,     20,    -65,     89,    -83,     50,      0,    -50,     83,    -89,     65,    -20,    -32,     73,    -90,     77,    -38,    -13,     60,    -87,     85,    -55,      7,     44,    -80,     90,    -69,     26 },
      {     46,    -79,     90,    -76,     40,      7,    -52,     82,    -90,     72,    -34,    -13,     57,    -85,     89,    -68,     28,     20,    -62,     87,    -87,     63,    -22,    -26,     66,    -88,     85,    -59,     15,     32,    -71,     89,    -83,     53,     -9,    -38,     75,    -90,     80,    -48,      2,     44,    -78,     90,    -77,     42,      4,    -50,     81,    -90,     73,    -36,    -11,     55,    -84,     89,    -69,     30,     17,    -60,     86,    -88,     65,    -24 },
      {     42,    -75,     90,    -84,     59,    -20,    -24,     62,    -85,     89,    -72,     38,      4,    -46,     77,    -90,     82,    -55,     15,     28,    -65,     87,    -88,     69,    -34,     -9,     50,    -79,     90,    -80,     52,    -11,    -32,     68,    -88,     87,    -66,     30,     13,    -53,     81,    -90,     78,    -48,      7,     36,    -71,     89,    -86,     63,    -26,    -17,     57,    -83,     90,    -76,     44,     -2,    -40,     73,    -89,     85,    -60,     22 },
      {     38,    -69,     87,    -89,     73,    -44,      7,     32,    -65,     85,    -90,     77,    -50,     13,     26,    -60,     83,    -90,     80,    -55,     20,     20,    -55,     80,    -90,     83,    -60,     26,     13,    -50,     77,    -90,     85,    -65,     32,      7,    -44,     73,    -89,     87,    -69,     38,      0,    -38,     69,    -87,     89,    -73,     44,     -7,    -32,     65,    -85,     90,    -77,     50,    -13,    -26,     60,    -83,     90,    -80,     55,    -20 },
      {     34,    -63,     83,    -90,     84,    -65,     36,     -2,    -32,     62,    -82,     90,    -85,     66,    -38,      4,     30,    -60,     81,    -90,     85,    -68,     40,     -7,    -28,     59,    -80,     90,    -86,     69,    -42,      9,     26,    -57,     79,    -90,     87,    -71,     44,    -11,    -24,     55,    -78,     89,    -87,     72,    -46,     13,     22,    -53,     77,    -89,     88,    -73,     48,    -15,    -20,     52,    -76,     89,    -88,     75,    -50,     17 },
      {     30,    -57,     77,    -88,     89,    -80,     62,    -36,      7,     24,    -52,     73,    -87,     90,    -83,     66,    -42,     13,     17,    -46,     69,    -85,     90,    -85,     71,    -48,     20,     11,    -40,     65,    -82,     90,    -87,     75,    -53,     26,      4,    -34,     60,    -79,     89,    -89,     78,    -59,     32,     -2,    -28,     55,    -76,     88,    -90,     81,    -63,     38,     -9,    -22,     50,    -72,     86,    -90,     84,    -68,     44,    -15 },
      {     26,    -50,     69,    -83,     90,    -89,     80,    -65,     44,    -20,     -7,     32,    -55,     73,    -85,     90,    -87,     77,    -60,     38,    -13,    -13,     38,    -60,     77,    -87,     90,    -85,     73,    -55,     32,     -7,    -20,     44,    -65,     80,    -89,     90,    -83,     69,    -50,     26,      0,    -26,     50,    -69,     83,    -90,     89,    -80,     65,    -44,     20,      7,    -32,     55,    -73,     85,    -90,     87,    -77,     60,    -38,     13 },
      {     22,    -42,     60,    -75,     85,    -90,     89,    -84,     73,    -59,     40,    -20,     -2,     24,    -44,     62,    -76,     85,    -90,     89,    -83,     72,    -57,     38,    -17,     -4,     26,    -46,     63,    -77,     86,    -90,     89,    -82,     71,    -55,     36,    -15,     -7,     28,    -48,     65,    -78,     87,    -90,     88,    -81,     69,    -53,     34,    -13,     -9,     30,    -50,     66,    -79,     87,    -90,     88,    -80,     68,    -52,     32,    -11 },
      {     17,    -34,     50,    -63,     75,    -83,     88,    -90,     89,    -84,     76,    -65,     52,    -36,     20,     -2,    -15,     32,    -48,     62,    -73,     82,    -88,     90,    -89,     85,    -77,     66,    -53,     38,    -22,      4,     13,    -30,     46,    -60,     72,    -81,     87,    -90,     89,    -85,     78,    -68,     55,    -40,     24,     -7,    -11,     28,    -44,     59,    -71,     80,    -87,     90,    -90,     86,    -79,     69,    -57,     42,    -26,      9 },
      {     13,    -26,     38,    -50,     60,    -69,     77,    -83,     87,    -90,     90,    -89,     85,    -80,     73,    -65,     55,    -44,     32,    -20,      7,      7,    -20,     32,    -44,     55,    -65,     73,    -80,     85,    -89,     90,    -90,     87,    -83,     77,    -69,     60,    -50,     38,    -26,     13,      0,    -13,     26,    -38,     50,    -60,     69,    -77,     83,    -87,     90,    -90,     89,    -85,     80,    -73,     65,    -55,     44,    -32,     20,     -7 },
      {      9,    -17,     26,    -34,     42,    -50,     57,    -63,     69,    -75,     79,    -83,     86,    -88,     90,    -90,     90,    -89,     87,    -84,     80,    -76,     71,    -65,     59,    -52,     44,    -36,     28,    -20,     11,     -2,     -7,     15,    -24,     32,    -40,     48,    -55,     62,    -68,     73,    -78,     82,    -85,     88,    -89,     90,    -90,     89,    -87,     85,    -81,     77,    -72,     66,    -60,     53,    -46,     38,    -30,     22,    -13,      4 },
      {      4,     -9,     13,    -17,     22,    -26,     30,    -34,     38,    -42,     46,    -50,     53,    -57,     60,    -63,     66,    -69,     72,    -75,     77,    -79,     81,    -83,     85,    -86,     87,    -88,     89,    -90,     90,    -90,     90,    -90,     89,    -89,     88,    -87,     85,    -84,     82,    -80,     78,    -76,     73,    -71,     68,    -65,     62,    -59,     55,    -52,     48,    -44,     40,    -36,     32,    -28,     24,    -20,     15,    -11,      7,     -2 },
    },
    {
      {      2,      4,      7,      9,     11,     13,     15,     17,     20,     22,     24,     26,     28,     30,     32,     34,     36,     38,     40,     42,     44,     46,     48,     50,     52,     53,     55,     57,     59,     60,     62,     63,     65,     66,     68,     69,     71,     72,     73,     75,     76,     77,     78,     79,     80,     81,     82,     83,     84,     85,     85,     86,     87,     87,     88,     88,     89,     89,     89,     90,     90,     90,     90,     90 },
      {      7,     13,     20,     26,     32,     38,     44,     50,     55,     60,     65,     69,     73,     77,     80,     83,     85,     87,     89,     90,     90,     90,     90,     89,     87,     85,     83,     80,     77,     73,     69,     65,     60,     55,     50,     44,     38,     32,     26,     20,     13,      7,      0,     -7,    -13,    -20,    -26,    -32,    -38,    -44,    -50,    -55,    -60,    -65,    -69,    -73,    -77,    -80,    -83,    -85,    -87,    -89,    -90,    -90 },
      {     11,     22,     32,     42,     52,     60,     68,     75,     80,     85,     88,     90,     90,     89,     87,     84,     79,     73,     66,     59,     50,     40,     30,     20,      9,     -2,    -13,    -24,    -34,    -44,    -53,    -62,    -69,    -76,    -81,    -85,    -88,    -90,    -90,    -89,    -87,    -83,    -78,    -72,    -65,    -57,    -48,    -38,    -28,    -17,     -7,      4,     15,     26,     36,     46,     55,     63,     71,     77,     82,     86,     89,     90 },
      {     15,     30,     44,     57,     68,     77,     84,     88,     90,     89,     86,     80,     72,     62,     50,     36,     22,      7,     -9,    -24,    -38,    -52,    -63,    -73,    -81,    -87,    -90,    -90,    -88,    -83,    -76,    -66,    -55,    -42,    -28,    -13,      2,     17,     32,     46,     59,     69,     78,     85,     89,     90,     89,     85,     79,     71,     60,     48,     34,     20,      4,    -11,    -26,    -40,    -53,    -65,    -75,    -82,    -87,    -90 },
      {     20,     38,     55,     69,     80,     87,     90,     89,     83,     73,     60,     44,     26,      7,    -13,    -32,    -50,    -65,    -77,    -85,    -90,    -90,    -85,    -77,    -65,    -50,    -32,    -13,      7,     26,     44,     60,     73,     83,     89,     90,     87,     80,     69,     55,     38,     20,      0,    -20,    -38,    -55,    -69,    -80,    -87,    -90,    -89,    -83,    -73,    -60,    -44,    -26,     -7,     13,     32,     50,     65,     77,     85,     90 },
      {     24,     46,     65,     79,     88,     90,     86,     76,     60,     40,     17,     -7,    -30,    -52,    -69,    -82,    -89,    -90,    -84,    -72,    -55,    -34,    -11,     13,     36,     57,     73,     85,     90,     89,     81,     68,     50,     28,      4,    -20,    -42,    -62,    -77,    -87,    -90,    -87,    -78,    -63,    -44,    -22,      2,     26,     48,     66,     80,     88,     90,     85,     75,     59,     38,     15,     -9,    -32,    -53,    -71,    -83,    -89 },
      {     28,     53,     73,     86,     90,     85,     72,     52,     26,     -2,    -30,    -55,    -75,    -87,    -90,    -85,    -71,    -50,    -24,      4,     32,     57,     76,     87,     90,     84,     69,     48,     22,     -7,    -34,    -59,    -77,    -88,    -90,    -83,    -68,    -46,    -20,      9,     36,     60,     78,     88,     90,     82,     66,     44,     17,    -11,    -38,    -62,    -79,    -89,    -89,    -81,    -65,    -42,    -15,     13,     40,     63,     80,     89 },
      {     32,     60,     80,     90,     87,     73,     50,     20,    -13,    -44,    -69,    -85,    -90,    -83,    -65,    -38,     -7,     26,     55,     77,     89,     89,     77,     55,     26,     -7,    -38,    -65,    -83,    -90,    -85,    -69,    -44,    -13,     20,     50,     73,     87,     90,     80,     60,     32,      0,    -32,    -60,    -80,    -90,    -87,    -73,    -50,    -20,     13,     44,     69,     85,     90,     83,     65,     38,      7,    -26,    -55,    -77,    -89 },
      {     36,     66,     85,     90,     79,     55,     22,    -15,    -50,    -76,    -89,    -87,    -71,    -42,     -7,     30,     62,     83,     90,     82,     60,     28,     -9,    -44,    -72,    -88,    -89,    -75,    -48,    -13,     24,     57,     80,     90,     85,     65,     34,     -2,    -38,    -68,    -86,    -90,    -78,    -53,    -20,     17,     52,     77,     89,     87,     69,     40,      4,    -32,    -63,    -84,    -90,    -81,    -59,    -26,     11,     46,     73,     88 },
      {     40,     72,     89,     87,     66,     32,     -9,    -48,    -77,    -90,    -84,    -60,    -24,     17,     55,     81,     90,     80,     53,     15,    -26,    -62,    -85,    -90,    -76,    -46,     -7,     34,     68,     87,     88,     71,     38,     -2,    -42,    -73,    -89,    -86,    -65,    -30,     11,     50,     78,     90,     83,     59,     22,    -20,    -57,    -82,    -90,    -79,    -52,    -13,     28,     63,     85,     89,     75,     44,      4,    -36,    -69,    -88 },
      {     44,     77,     90,     80,     50,      7,    -38,    -73,    -90,    -83,    -55,    -13,     32,     69,     89,     85,     60,     20,    -26,    -65,    -87,    -87,    -65,    -26,     20,     60,     85,     89,     69,     32,    -13,    -55,    -83,    -90,    -73,    -38,      7,     50,     80,     90,     77,     44,      0,    -44,    -77,    -90,    -80,    -50,     -7,     38,     73,     90,     83,     55,     13,    -32,    -69,    -89,    -85,    -60,    -20,     26,     65,     87 },
      {     48,     81,     90,     71,     30,    -20,    -63,    -88,    -85,    -57,    -11,     38,     76,     90,     77,     40,     -9,    -55,    -85,    -88,    -65,    -22,     28,     69,     89,     82,     50,      2,    -46,    -80,    -90,    -72,    -32,     17,     62,     87,     86,     59,     13,    -36,    -75,    -90,    -78,    -42,      7,     53,     84,     89,     66,     24,    -26,    -68,    -89,    -83,    -52,     -4,     44,     79,     90,     73,     34,    -15,    -60,    -87 },
      {     52,     85,     87,     59,      9,    -44,    -81,    -89,    -65,    -17,     36,     77,     90,     71,     26,    -28,    -72,    -90,    -76,    -34,     20,     66,     89,     80,     42,    -11,    -60,    -88,    -84,    -50,      2,     53,     85,     87,     57,      7,    -46,    -82,    -89,    -63,    -15,     38,     78,     90,     69,     24,    -30,    -73,    -90,    -75,    -32,     22,     68,     90,     79,     40,    -13,    -62,    -88,    -83,    -48,      4,     55,     86 },
      {     55,     87,     83,     44,    -13,    -65,    -90,    -77,    -32,     26,     73,     90,     69,     20,    -38,    -80,    -89,    -60,     -7,     50,     85,     85,     50,     -7,    -60,    -89,    -80,    -38,     20,     69,     90,     73,     26,    -32,    -77,    -90,    -65,    -13,     44,     83,     87,     55,      0,    -55,    -87,    -83,    -44,     13,     65,     90,     77,     32,    -26,    -73,    -90,    -69,    -20,     38,     80,     89,     60,      7,    -50,    -85 },
      {     59,     89,     77,     28,    -34,    -80,    -88,    -53,      7,     63,     90,     73,     22,    -40,    -83,    -86,    -48,     13,     68,     90,     69,     15,    -46,    -85,    -84,    -42,     20,     72,     90,     65,      9,    -52,    -87,    -81,    -36,     26,     76,     89,     60,      2,    -57,    -89,    -78,    -30,     32,     79,     88,     55,     -4,    -62,    -90,    -75,    -24,     38,     82,     87,     50,    -11,    -66,    -90,    -71,    -17,     44,     85 },
      {     62,     90,     69,     11,    -53,    -89,    -76,    -22,     44,     86,     81,     32,    -34,    -82,    -85,    -42,     24,     77,     88,     52,    -13,    -71,    -90,    -60,      2,     63,     90,     68,      9,    -55,    -89,    -75,    -20,     46,     87,     80,     30,    -36,    -83,    -85,    -40,     26,     78,     88,     50,    -15,    -72,    -90,    -59,      4,     65,     90,     66,      7,    -57,    -89,    -73,    -17,     48,     87,     79,     28,    -38,    -84 },
      {     65,     90,     60,     -7,    -69,    -90,    -55,     13,     73,     89,     50,    -20,    -77,    -87,    -44,     26,     80,     85,     38,    -32,    -83,    -83,    -32,     38,     85,     80,     26,    -44,    -87,    -77,    -20,     50,     89,     73,     13,    -55,    -90,    -69,     -7,     60,     90,     65,      0,    -65,    -90,    -60,      7,     69,     90,     55,    -13,    -73,    -89,    -50,     20,     77,     87,     44,    -26,    -80,    -85,    -38,     32,     83 },
      {     68,     89,     50,    -24,    -81,    -83,    -28,     46,     89,     71,      4,    -65,    -90,    -53,     20,     79,     85,     32,    -42,    -88,    -73,     -9,     62,     90,     57,    -15,    -77,    -86,    -36,     38,     87,     76,     13,    -59,    -90,    -60,     11,     75,     87,     40,    -34,    -85,    -78,    -17,     55,     90,     63,     -7,    -72,    -88,    -44,     30,     84,     80,     22,    -52,    -90,    -66,      2,     69,     89,     48,    -26,    -82 },
      {     71,     88,     38,    -40,    -88,    -69,      2,     72,     87,     36,    -42,    -89,    -68,      4,     73,     87,     34,    -44,    -89,    -66,      7,     75,     86,     32,    -46,    -89,    -65,      9,     76,     85,     30,    -48,    -90,    -63,     11,     77,     85,     28,    -50,    -90,    -62,     13,     78,     84,     26,    -52,    -90,    -60,     15,     79,     83,     24,    -53,    -90,    -59,     17,     80,     82,     22,    -55,    -90,    -57,     20,     81 },
      {     73,     85,     26,    -55,    -90,    -50,     32,     87,     69,     -7,    -77,    -83,    -20,     60,     90,     44,    -38,    -89,    -65,     13,     80,     80,     13,    -65,    -89,    -38,     44,     90,     60,    -20,    -83,    -77,     -7,     69,     87,     32,    -50,    -90,    -55,     26,     85,     73,      0,    -73,    -85,    -26,     55,     90,     50,    -32,    -87,    -69,      7,     77,     83,     20,    -60,    -90,    -44,     38,     89,     65,    -13,    -80 },
      {     76,     82,     13,    -68,    -87,    -26,     59,     89,     38,    -48,    -90,    -50,     36,     89,     60,    -24,    -86,    -69,     11,     81,     77,      2,    -75,    -83,    -15,     66,     87,     28,    -57,    -90,    -40,     46,     90,     52,    -34,    -89,    -62,     22,     85,     71,     -9,    -80,    -78,     -4,     73,     84,     17,    -65,    -88,    -30,     55,     90,     42,    -44,    -90,    -53,     32,     88,     63,    -20,    -85,    -72,      7,     79 },
      {     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78,    -78,      0,     78,     78,      0,    -78 },
      {     80,     73,    -13,    -85,    -65,     26,     89,     55,    -38,    -90,    -44,     50,     90,     32,    -60,    -87,    -20,     69,     83,      7,    -77,    -77,      7,     83,     69,    -20,    -87,    -60,     32,     90,     50,    -44,    -90,    -38,     55,     89,     26,    -65,    -85,    -13,     73,     80,      0,    -80,    -73,     13,     85,     65,    -26,    -89,    -55,     38,     90,     44,    -50,    -90,    -32,     60,     87,     20,    -69,    -83,     -7,     77 },
      {     82,     68,    -26,    -89,    -48,     50,     89,     24,    -69,    -81,      2,     83,     66,    -28,    -90,    -46,     52,     89,     22,    -71,    -80,      4,     84,     65,    -30,    -90,    -44,     53,     88,     20,    -72,    -79,      7,     85,     63,    -32,    -90,    -42,     55,     88,     17,    -73,    -78,      9,     85,     62,    -34,    -90,    -40,     57,     87,     15,    -75,    -77,     11,     86,     60,    -36,    -90,    -38,     59,     87,     13,    -76 },
      {     84,     62,    -38,    -90,    -28,     69,     79,    -11,    -87,    -53,     48,     89,     17,    -76,    -73,     22,     89,     44,    -57,    -86,     -7,     81,     66,    -32,    -90,    -34,     65,     82,     -4,    -85,    -59,     42,     90,     24,    -72,    -77,     15,     88,     50,    -52,    -88,    -13,     78,     71,    -26,    -90,    -40,     60,     85,      2,    -83,    -63,     36,     90,     30,    -68,    -80,      9,     87,     55,    -46,    -89,    -20,     75 },
      {     85,     55,    -50,    -87,     -7,     83,     60,    -44,    -89,    -13,     80,     65,    -38,    -90,    -20,     77,     69,    -32,    -90,    -26,     73,     73,    -26,    -90,    -32,     69,     77,    -20,    -90,    -38,     65,     80,    -13,    -89,    -44,     60,     83,     -7,    -87,    -50,     55,     85,      0,    -85,    -55,     50,     87,      7,    -83,    -60,     44,     89,     13,    -80,    -65,     38,     90,     20,    -77,    -69,     32,     90,     26,    -73 },
      {     87,     48,    -60,    -81,     15,     90,     34,    -71,    -73,     30,     90,     20,    -79,    -63,     44,     88,      4,    -85,    -52,     57,     83,    -11,    -89,    -38,     68,     76,    -26,    -90,    -24,     77,     66,    -40,    -89,     -9,     84,     55,    -53,    -85,      7,     88,     42,    -65,    -78,     22,     90,     28,    -75,    -69,     36,     89,     13,    -82,    -59,     50,     86,     -2,    -87,    -46,     62,     80,    -17,    -90,    -32,     72 },
      {     88,     40,    -69,    -72,     36,     89,      4,    -87,    -44,     66,     75,    -32,    -89,     -9,     85,     48,    -63,    -77,     28,     90,     13,    -84,    -52,     60,     79,    -24,    -90,    -17,     82,     55,    -57,    -81,     20,     90,     22,    -80,    -59,     53,     83,    -15,    -90,    -26,     78,     62,    -50,    -85,     11,     90,     30,    -76,    -65,     46,     86,     -7,    -89,    -34,     73,     68,    -42,    -87,      2,     88,     38,    -71 },
      {     89,     32,    -77,    -60,     55,     80,    -26,    -90,     -7,     87,     38,    -73,    -65,     50,     83,    -20,    -90,    -13,     85,     44,    -69,    -69,     44,     85,    -13,    -90,    -20,     83,     50,    -65,    -73,     38,     87,     -7,    -90,    -26,     80,     55,    -60,    -77,     32,     89,      0,    -89,    -32,     77,     60,    -55,    -80,     26,     90,      7,    -87,    -38,     73,     65,    -50,    -83,     20,     90,     13,    -85,    -44,     69 },
      {     89,     24,    -83,    -46,     71,     65,    -53,    -79,     32,     88,     -9,    -90,    -15,     86,     38,    -76,    -59,     60,     75,    -40,    -85,     17,     90,      7,    -88,    -30,     80,     52,    -66,    -69,     48,     82,    -26,    -89,      2,     90,     22,    -84,    -44,     72,     63,    -55,    -78,     34,     87,    -11,    -90,    -13,     87,     36,    -77,    -57,     62,     73,    -42,    -85,     20,     90,      4,    -89,    -28,     81,     50,    -68 },
      {     90,     15,    -87,    -30,     82,     44,    -75,    -57,     65,     68,    -53,    -77,     40,     84,    -26,    -88,     11,     90,      4,    -89,    -20,     86,     34,    -80,    -48,     72,     60,    -62,    -71,     50,     79,    -36,    -85,     22,     89,     -7,    -90,     -9,     89,     24,    -85,    -38,     78,     52,    -69,    -63,     59,     73,    -46,    -81,     32,     87,    -17,    -90,      2,     90,     13,    -88,    -28,     83,     42,    -76,    -55,     66 },
      {     90,      7,    -90,    -13,     89,     20,    -87,    -26,     85,     32,    -83,    -38,     80,     44,    -77,    -50,     73,     55,    -69,    -60,     65,     65,    -60,    -69,     55,     73,    -50,    -77,     44,     80,    -38,    -83,     32,     85,    -26,    -87,     20,     89,    -13,    -90,      7,     90,      0,    -90,     -7,     90,     13,    -89,    -20,     87,     26,    -85,    -32,     83,     38,    -80,    -44,     77,     50,    -73,    -55,     69,     60,    -65 },
      {     90,     -2,    -90,      4,     90,     -7,    -90,      9,     90,    -11,    -89,     13,     89,    -15,    -89,     17,     88,    -20,    -88,     22,     87,    -24,    -87,     26,     86,    -28,    -85,     30,     85,    -32,    -84,     34,     83,    -36,    -82,     38,     81,    -40,    -80,     42,     79,    -44,    -78,     46,     77,    -48,    -76,     50,     75,    -52,    -73,     53,     72,    -55,    -71,     57,     69,    -59,    -68,     60,     66,    -62,    -65,     63 },
      {     90,    -11,    -89,     22,     86,    -32,    -82,     42,     77,    -52,    -71,     60,     63,    -68,    -55,     75,     46,    -80,    -36,     85,     26,    -88,    -15,     90,      4,    -90,      7,     89,    -17,    -87,     28,     84,    -38,    -79,     48,     73,    -57,    -66,     65,     59,    -72,    -50,     78,     40,    -83,    -30,     87,     20,    -89,     -9,     90,     -2,    -90,     13,     88,    -24,    -85,     34,     81,    -44,    -76,     53,     69,    -62 },
      {     90,    -20,    -85,     38,     77,    -55,    -65,     69,     50,    -80,    -32,     87,     13,    -90,      7,     89,    -26,    -83,     44,     73,    -60,    -60,     73,     44,    -83,    -26,     89,      7,    -90,     13,     87,    -32,    -80,     50,     69,    -65,    -55,     77,     38,    -85,    -20,     90,      0,    -90,     20,     85,    -38,    -77,     55,     65,    -69,    -50,     80,     32,    -87,    -13,     90,     -7,    -89,     26,     83,    -44,    -73,     60 },
      {     89,    -28,    -80,     53,     63,    -73,    -40,     86,     13,    -90,     15,     85,    -42,    -72,     65,     52,    -81,    -26,     89,     -2,    -89,     30,     79,    -55,    -62,     75,     38,    -87,    -11,     90,    -17,    -85,     44,     71,    -66,    -50,     82,     24,    -90,      4,     88,    -32,    -78,     57,     60,    -76,    -36,     87,      9,    -90,     20,     84,    -46,    -69,     68,     48,    -83,    -22,     90,     -7,    -88,     34,     77,    -59 },
      {     88,    -36,    -73,     66,     46,    -85,    -11,     90,    -26,    -79,     59,     55,    -81,    -22,     90,    -15,    -84,     50,     63,    -76,    -32,     89,     -4,    -87,     40,     71,    -69,    -42,     87,      7,    -89,     30,     77,    -62,    -52,     83,     17,    -90,     20,     82,    -53,    -60,     78,     28,    -90,      9,     86,    -44,    -68,     72,     38,    -88,     -2,     89,    -34,    -75,     65,     48,    -85,    -13,     90,    -24,    -80,     57 },
      {     87,    -44,    -65,     77,     26,    -90,     20,     80,    -60,    -50,     85,      7,    -89,     38,     69,    -73,    -32,     90,    -13,    -83,     55,     55,    -83,    -13,     90,    -32,    -73,     69,     38,    -89,      7,     85,    -50,    -60,     80,     20,    -90,     26,     77,    -65,    -44,     87,      0,    -87,     44,     65,    -77,    -26,     90,    -20,    -80,     60,     50,    -85,     -7,     89,    -38,    -69,     73,     32,    -90,     13,     83,    -55 },
      {     86,    -52,    -55,     85,      4,    -87,     48,     59,    -83,     -9,     88,    -44,    -62,     81,     13,    -89,     40,     65,    -79,    -17,     90,    -36,    -68,     77,     22,    -90,     32,     71,    -75,    -26,     90,    -28,    -73,     72,     30,    -90,     24,     76,    -69,    -34,     90,    -20,    -78,     66,     38,    -89,     15,     80,    -63,    -42,     89,    -11,    -82,     60,     46,    -88,      7,     84,    -57,    -50,     87,     -2,    -85,     53 },
      {     85,    -59,    -44,     89,    -17,    -77,     71,     28,    -90,     34,     66,    -80,    -11,     88,    -50,    -53,     87,     -7,    -82,     63,     38,    -90,     24,     73,    -75,    -22,     90,    -40,    -62,     83,      4,    -86,     55,     48,    -88,     13,     79,    -68,    -32,     90,    -30,    -69,     78,     15,    -89,     46,     57,    -85,      2,     84,    -60,    -42,     89,    -20,    -76,     72,     26,    -90,     36,     65,    -81,     -9,     87,    -52 },
      {     83,    -65,    -32,     90,    -38,    -60,     85,     -7,    -80,     69,     26,    -90,     44,     55,    -87,     13,     77,    -73,    -20,     89,    -50,    -50,     89,    -20,    -73,     77,     13,    -87,     55,     44,    -90,     26,     69,    -80,     -7,     85,    -60,    -38,     90,    -32,    -65,     83,      0,    -83,     65,     32,    -90,     38,     60,    -85,      7,     80,    -69,    -26,     90,    -44,    -55,     87,    -13,    -77,     73,     20,    -89,     50 },
      {     81,    -71,    -20,     88,    -57,    -38,     90,    -40,    -55,     88,    -22,    -69,     82,     -2,    -80,     72,     17,    -87,     59,     36,    -90,     42,     53,    -89,     24,     68,    -83,      4,     79,    -73,    -15,     87,    -60,    -34,     90,    -44,    -52,     89,    -26,    -66,     84,     -7,    -78,     75,     13,    -86,     62,     32,    -90,     46,     50,    -89,     28,     65,    -85,      9,     77,    -76,    -11,     85,    -63,    -30,     90,    -48 },
      {     79,    -76,     -7,     82,    -72,    -13,     85,    -68,    -20,     87,    -63,    -26,     88,    -59,    -32,     89,    -53,    -38,     90,    -48,    -44,     90,    -42,    -50,     90,    -36,    -55,     89,    -30,    -60,     88,    -24,    -65,     86,    -17,    -69,     84,    -11,    -73,     81,     -4,    -77,     78,      2,    -80,     75,      9,    -83,     71,     15,    -85,     66,     22,    -87,     62,     28,    -89,     57,     34,    -90,     52,     40,    -90,     46 },
      {     77,    -80,      7,     73,    -83,     13,     69,    -85,     20,     65,    -87,     26,     60,    -89,     32,     55,    -90,     38,     50,    -90,     44,     44,    -90,     50,     38,    -90,     55,     32,    -89,     60,     26,    -87,     65,     20,    -85,     69,     13,    -83,     73,      7,    -80,     77,      0,    -77,     80,     -7,    -73,     83,    -13,    -69,     85,    -20,    -65,     87,    -26,    -60,     89,    -32,    -55,     90,    -38,    -50,     90,    -44 },
      {     75,    -84,     20,     62,    -89,     38,     46,    -90,     55,     28,    -87,     69,      9,    -79,     80,    -11,    -68,     87,    -30,    -53,     90,    -48,    -36,     89,    -63,    -17,     83,    -76,      2,     73,    -85,     22,     60,    -89,     40,     44,    -90,     57,     26,    -86,     71,      7,    -78,     81,    -13,    -66,     88,    -32,    -52,     90,    -50,    -34,     88,    -65,    -15,     82,    -77,      4,     72,    -85,     24,     59,    -90,     42 },
      {     72,    -87,     32,     48,    -90,     60,     17,    -81,     80,    -15,    -62,     90,    -46,    -34,     87,    -71,     -2,     73,    -86,     30,     50,    -90,     59,     20,    -82,     79,    -13,    -63,     89,    -44,    -36,     88,    -69,     -4,     75,    -85,     28,     52,    -90,     57,     22,    -83,     78,    -11,    -65,     89,    -42,    -38,     88,    -68,     -7,     76,    -85,     26,     53,    -90,     55,     24,    -84,     77,     -9,    -66,     89,    -40 },
      {     69,    -89,     44,     32,    -85,     77,    -13,    -60,     90,    -55,    -20,     80,    -83,     26,     50,    -90,     65,      7,    -73,     87,    -38,    -38,     87,    -73,      7,     65,    -90,     50,     26,    -83,     80,    -20,    -55,     90,    -60,    -13,     77,    -85,     32,     44,    -89,     69,      0,    -69,     89,    -44,    -32,     85,    -77,     13,     60,    -90,     55,     20,    -80,     83,    -26,    -50,     90,    -65,     -7,     73,    -87,     38 },
      {     66,    -90,     55,     15,    -76,     87,    -42,    -30,     83,    -82,     28,     44,    -88,     75,    -13,    -57,     90,    -65,     -2,     68,    -90,     53,     17,    -77,     87,    -40,    -32,     84,    -81,     26,     46,    -88,     73,    -11,    -59,     90,    -63,     -4,     69,    -89,     52,     20,    -78,     86,    -38,    -34,     85,    -80,     24,     48,    -89,     72,     -9,    -60,     90,    -62,     -7,     71,    -89,     50,     22,    -79,     85,    -36 },
      {     63,    -90,     65,     -2,    -62,     90,    -66,      4,     60,    -90,     68,     -7,    -59,     90,    -69,      9,     57,    -90,     71,    -11,    -55,     89,    -72,     13,     53,    -89,     73,    -15,    -52,     89,    -75,     17,     50,    -88,     76,    -20,    -48,     88,    -77,     22,     46,    -87,     78,    -24,    -44,     87,    -79,     26,     42,    -86,     80,    -28,    -40,     85,    -81,     30,     38,    -85,     82,    -32,    -36,     84,    -83,     34 },
      {     60,    -90,     73,    -20,    -44,     85,    -83,     38,     26,    -77,     89,    -55,     -7,     65,    -90,     69,    -13,    -50,     87,    -80,     32,     32,    -80,     87,    -50,    -13,     69,    -90,     65,     -7,    -55,     89,    -77,     26,     38,    -83,     85,    -44,    -20,     73,    -90,     60,      0,    -60,     90,    -73,     20,     44,    -85,     83,    -38,    -26,     77,    -89,     55,      7,    -65,     90,    -69,     13,     50,    -87,     80,    -32 },
      {     57,    -88,     80,    -36,    -24,     73,    -90,     66,    -13,    -46,     85,    -85,     48,     11,    -65,     90,    -75,     26,     34,    -79,     89,    -59,      2,     55,    -88,     81,    -38,    -22,     72,    -90,     68,    -15,    -44,     84,    -86,     50,      9,    -63,     90,    -76,     28,     32,    -78,     89,    -60,      4,     53,    -87,     82,    -40,    -20,     71,    -90,     69,    -17,    -42,     83,    -87,     52,      7,    -62,     89,    -77,     30 },
      {     53,    -86,     85,    -52,     -2,     55,    -87,     85,    -50,     -4,     57,    -87,     84,    -48,     -7,     59,    -88,     83,    -46,     -9,     60,    -88,     82,    -44,    -11,     62,    -89,     81,    -42,    -13,     63,    -89,     80,    -40,    -15,     65,    -89,     79,    -38,    -17,     66,    -90,     78,    -36,    -20,     68,    -90,     77,    -34,    -22,     69,    -90,     76,    -32,    -24,     71,    -90,     75,    -30,    -26,     72,    -90,     73,    -28 },
      {     50,    -83,     89,    -65,     20,     32,    -73,     90,    -77,     38,     13,    -60,     87,    -85,     55,     -7,    -44,     80,    -90,     69,    -26,    -26,     69,    -90,     80,    -44,     -7,     55,    -85,     87,    -60,     13,     38,    -77,     90,    -73,     32,     20,    -65,     89,    -83,     50,      0,    -50,     83,    -89,     65,    -20,    -32,     73,    -90,     77,    -38,    -13,     60,    -87,     85,    -55,      7,     44,    -80,     90,    -69,     26 },
      {     46,    -79,     90,    -76,     40,      7,    -52,     82,    -90,     72,    -34,    -13,     57,    -85,     89,    -68,     28,     20,    -62,     87,    -87,     63,    -22,    -26,     66,    -88,     85,    -59,     15,     32,    -71,     89,    -83,     53,     -9,    -38,     75,    -90,     80,    -48,      2,     44,    -78,     90,    -77,     42,      4,    -50,     81,    -90,     73,    -36,    -11,     55,    -84,     89,    -69,     30,     17,    -60,     86,    -88,     65,    -24 },
      {     42,    -75,     90,    -84,     59,    -20,    -24,     62,    -85,     89,    -72,     38,      4,    -46,     77,    -90,     82,    -55,     15,     28,    -65,     87,    -88,     69,    -34,     -9,     50,    -79,     90,    -80,     52,    -11,    -32,     68,    -88,     87,    -66,     30,     13,    -53,     81,    -90,     78,    -48,      7,     36,    -71,     89,    -86,     63,    -26,    -17,     57,    -83,     90,    -76,     44,     -2,    -40,     73,    -89,     85,    -60,     22 },
      {     38,    -69,     87,    -89,     73,    -44,      7,     32,    -65,     85,    -90,     77,    -50,     13,     26,    -60,     83,    -90,     80,    -55,     20,     20,    -55,     80,    -90,     83,    -60,     26,     13,    -50,     77,    -90,     85,    -65,     32,      7,    -44,     73,    -89,     87,    -69,     38,      0,    -38,     69,    -87,     89,    -73,     44,     -7,    -32,     65,    -85,     90,    -77,     50,    -13,    -26,     60,    -83,     90,    -80,     55,    -20 },
      {     34,    -63,     83,    -90,     84,    -65,     36,     -2,    -32,     62,    -82,     90,    -85,     66,    -38,      4,     30,    -60,     81,    -90,     85,    -68,     40,     -7,    -28,     59,    -80,     90,    -86,     69,    -42,      9,     26,    -57,     79,    -90,     87,    -71,     44,    -11,    -24,     55,    -78,     89,    -87,     72,    -46,     13,     22,    -53,     77,    -89,     88,    -73,     48,    -15,    -20,     52,    -76,     89,    -88,     75,    -50,     17 },
      {     30,    -57,     77,    -88,     89,    -80,     62,    -36,      7,     24,    -52,     73,    -87,     90,    -83,     66,    -42,     13,     17,    -46,     69,    -85,     90,    -85,     71,    -48,     20,     11,    -40,     65,    -82,     90,    -87,     75,    -53,     26,      4,    -34,     60,    -79,     89,    -89,     78,    -59,     32,     -2,    -28,     55,    -76,     88,    -90,     81,    -63,     38,     -9,    -22,     50,    -72,     86,    -90,     84,    -68,     44,    -15 },
      {     26,    -50,     69,    -83,     90,    -89,     80,    -65,     44,    -20,     -7,     32,    -55,     73,    -85,     90,    -87,     77,    -60,     38,    -13,    -13,     38,    -60,     77,    -87,     90,    -85,     73,    -55,     32,     -7,    -20,     44,    -65,     80,    -89,     90,    -83,     69,    -50,     26,      0,    -26,     50,    -69,     83,    -90,     89,    -80,     65,    -44,     20,      7,    -32,     55,    -73,     85,    -90,     87,    -77,     60,    -38,     13 },
      {     22,    -42,     60,    -75,     85,    -90,     89,    -84,     73,    -59,     40,    -20,     -2,     24,    -44,     62,    -76,     85,    -90,     89,    -83,     72,    -57,     38,    -17,     -4,     26,    -46,     63,    -77,     86,    -90,     89,    -82,     71,    -55,     36,    -15,     -7,     28,    -48,     65,    -78,     87,    -90,     88,    -81,     69,    -53,     34,    -13,     -9,     30,    -50,     66,    -79,     87,    -90,     88,    -80,     68,    -52,     32,    -11 },
      {     17,    -34,     50,    -63,     75,    -83,     88,    -90,     89,    -84,     76,    -65,     52,    -36,     20,     -2,    -15,     32,    -48,     62,    -73,     82,    -88,     90,    -89,     85,    -77,     66,    -53,     38,    -22,      4,     13,    -30,     46,    -60,     72,    -81,     87,    -90,     89,    -85,     78,    -68,     55,    -40,     24,     -7,    -11,     28,    -44,     59,    -71,     80,    -87,     90,    -90,     86,    -79,     69,    -57,     42,    -26,      9 },
      {     13,    -26,     38,    -50,     60,    -69,     77,    -83,     87,    -90,     90,    -89,     85,    -80,     73,    -65,     55,    -44,     32,    -20,      7,      7,    -20,     32,    -44,     55,    -65,     73,    -80,     85,    -89,     90,    -90,     87,    -83,     77,    -69,     60,    -50,     38,    -26,     13,      0,    -13,     26,    -38,     50,    -60,     69,    -77,     83,    -87,     90,    -90,     89,    -85,     80,    -73,     65,    -55,     44,    -32,     20,     -7 },
      {      9,    -17,     26,    -34,     42,    -50,     57,    -63,     69,    -75,     79,    -83,     86,    -88,     90,    -90,     90,    -89,     87,    -84,     80,    -76,     71,    -65,     59,    -52,     44,    -36,     28,    -20,     11,     -2,     -7,     15,    -24,     32,    -40,     48,    -55,     62,    -68,     73,    -78,     82,    -85,     88,    -89,     90,    -90,     89,    -87,     85,    -81,     77,    -72,     66,    -60,     53,    -46,     38,    -30,     22,    -13,      4 },
      {      4,     -9,     13,    -17,     22,    -26,     30,    -34,     38,    -42,     46,    -50,     53,    -57,     60,    -63,     66,    -69,     72,    -75,     77,    -79,     81,    -83,     85,    -86,     87,    -88,     89,    -90,     90,    -90,     90,    -90,     89,    -89,     88,    -87,     85,    -84,     82,    -80,     78,    -76,     73,    -71,     68,    -65,     62,    -59,     55,    -52,     48,    -44,     40,    -36,     32,    -28,     24,    -20,     15,    -11,      7,     -2 },
    },
};
const TMatrixCoeff g_trCoreDST7P128[TRANSFORM_NUMBER_OF_DIRECTIONS][128][128] =
{
    {
      {      1,      2,      3,      4,      6,      7,      8,      9,     10,     11,     12,     13,     14,     15,     16,     18,     19,     20,     21,     22,     23,     24,     25,     26,     27,     28,     29,     30,     31,     32,     33,     34,     35,     36,     37,     38,     39,     40,     41,     42,     43,     44,     45,     46,     47,     48,     49,     50,     51,     52,     53,     54,     55,     55,     56,     57,     58,     59,     60,     60,     61,     62,     63,     64,     64,     65,     66,     67,     67,     68,     69,     70,     70,     71,     72,     72,     73,     74,     74,     75,     76,     76,     77,     77,     78,     78,     79,     79,     80,     81,     81,     81,     82,     82,     83,     83,     84,     84,     85,     85,     85,     86,     86,     86,     87,     87,     87,     88,     88,     88,     88,     89,     89,     89,     89,     89,     89,     90,     90,     90,     90,     90,     90,     90,     90,     90,     90,     90 },
      {      3,      7,     10,     13,     16,     20,     23,     26,     29,     32,     35,     38,     41,     44,     47,     50,     53,     55,     58,     60,     63,     65,     67,     70,     72,     74,     76,     77,     79,     81,     82,     83,     85,     86,     87,     88,     88,     89,     89,     90,     90,     90,     90,     90,     90,     90,     89,     89,     88,     87,     86,     85,     84,     83,     81,     80,     78,     77,     75,     73,     71,     69,     67,     64,     62,     60,     57,     55,     52,     49,     46,     43,     40,     37,     34,     31,     28,     25,     22,     19,     15,     12,      9,      6,      2,     -1,     -4,     -8,    -11,    -14,    -18,    -21,    -24,    -27,    -30,    -33,    -36,    -39,    -42,    -45,    -48,    -51,    -54,    -56,    -59,    -61,    -64,    -66,    -68,    -70,    -72,    -74,    -76,    -78,    -79,    -81,    -82,    -84,    -85,    -86,    -87,    -88,    -89,    -89,    -90,    -90,    -90,    -90 },
      {      6,     11,     16,     22,     27,     32,     37,     42,     47,     52,     56,     60,     64,     68,     72,     75,     78,     81,     83,     85,     87,     88,     89,     90,     90,     90,     90,     89,     89,     87,     86,     84,     81,     79,     76,     73,     70,     66,     62,     58,     54,     49,     44,     39,     34,     29,     24,     19,     13,      8,      2,     -3,     -9,    -14,    -20,    -25,    -30,    -35,    -40,    -45,    -50,    -55,    -59,    -63,    -67,    -70,    -74,    -77,    -79,    -82,    -84,    -86,    -88,    -89,    -90,    -90,    -90,    -90,    -90,    -89,    -88,    -86,    -85,    -82,    -80,    -77,    -74,    -71,    -67,    -64,    -60,    -55,    -51,    -46,    -41,    -36,    -31,    -26,    -21,    -15,    -10,     -4,      1,      7,     12,     18,     23,     28,     33,     38,     43,     48,     53,     57,     61,     65,     69,     72,     76,     78,     81,     83,     85,     87,     88,     89,     90,     90 },
      {      8,     15,     23,     30,     37,     44,     51,     57,     63,     68,     73,     77,     81,     84,     87,     89,     90,     90,     90,     89,     88,     86,     83,     80,     76,     72,     67,     61,     55,     49,     42,     35,     28,     21,     13,      6,     -2,    -10,    -18,    -25,    -32,    -39,    -46,    -53,    -59,    -64,    -70,    -74,    -78,    -82,    -85,    -87,    -89,    -90,    -90,    -90,    -89,    -88,    -85,    -82,    -79,    -75,    -70,    -65,    -60,    -54,    -47,    -40,    -33,    -26,    -19,    -11,     -3,      4,     12,     20,     27,     34,     41,     48,     55,     60,     66,     71,     76,     79,     83,     86,     88,     89,     90,     90,     90,     89,     87,     85,     81,     78,     74,     69,     64,     58,     52,     45,     38,     31,     24,     16,      9,      1,     -7,    -14,    -22,    -29,    -36,    -43,    -50,    -56,    -62,    -67,    -72,    -77,    -81,    -84,    -86,    -88,    -90,    -90 },
      {     10,     20,     29,     38,     47,     55,     63,     70,     76,     81,     85,     88,     89,     90,     90,     89,     86,     83,     78,     73,     67,     60,     52,     43,     34,     25,     15,      6,     -4,    -14,    -24,    -33,    -42,    -51,    -59,    -66,    -72,    -78,    -82,    -86,    -89,    -90,    -90,    -90,    -88,    -85,    -81,    -76,    -70,    -64,    -56,    -48,    -39,    -30,    -21,    -11,     -1,      9,     19,     28,     37,     46,     55,     62,     69,     75,     80,     84,     87,     89,     90,     90,     89,     87,     83,     79,     74,     67,     60,     53,     44,     35,     26,     16,      7,     -3,    -13,    -23,    -32,    -41,    -50,    -58,    -65,    -72,    -77,    -82,    -86,    -88,    -90,    -90,    -90,    -88,    -85,    -81,    -77,    -71,    -64,    -57,    -49,    -40,    -31,    -22,    -12,     -2,      8,     18,     27,     36,     45,     54,     61,     68,     74,     79,     84,     87,     89,     90 },
      {     12,     24,     35,     46,     56,     65,     73,     79,     85,     88,     90,     90,     89,     86,     81,     76,     68,     60,     50,     39,     28,     16,      4,     -8,    -20,    -31,    -42,    -53,    -62,    -70,    -77,    -83,    -87,    -89,    -90,    -90,    -87,    -83,    -78,    -71,    -63,    -54,    -43,    -32,    -21,     -9,      3,     15,     27,     38,     49,     59,     67,     75,     81,     86,     89,     90,     90,     88,     85,     80,     74,     66,     57,     47,     36,     25,     13,      1,    -11,    -23,    -34,    -45,    -55,    -64,    -72,    -79,    -84,    -88,    -90,    -90,    -89,    -86,    -82,    -76,    -69,    -60,    -51,    -40,    -29,    -18,     -6,      7,     19,     30,     41,     52,     61,     70,     77,     82,     87,     89,     90,     90,     88,     84,     78,     72,     64,     55,     44,     33,     22,     10,     -2,    -14,    -26,    -37,    -48,    -58,    -67,    -74,    -81,    -85,    -89,    -90 },
      {     14,     28,     41,     54,     64,     74,     81,     86,     89,     90,     89,     85,     79,     72,     62,     51,     38,     25,     11,     -3,    -18,    -31,    -44,    -56,    -67,    -76,    -82,    -87,    -90,    -90,    -88,    -84,    -78,    -70,    -60,    -48,    -35,    -22,     -8,      7,     21,     34,     47,     59,     69,     77,     84,     88,     90,     90,     88,     83,     76,     67,     57,     45,     32,     19,      4,    -10,    -24,    -37,    -50,    -61,    -71,    -79,    -85,    -89,    -90,    -90,    -87,    -81,    -74,    -65,    -55,    -42,    -29,    -15,     -1,     13,     27,     40,     53,     64,     73,     81,     86,     89,     90,     89,     86,     80,     72,     63,     52,     39,     26,     12,     -2,    -16,    -30,    -43,    -55,    -66,    -75,    -82,    -87,    -90,    -90,    -89,    -85,    -78,    -70,    -60,    -49,    -36,    -23,     -9,      6,     20,     33,     46,     58,     68,     77,     83,     88,     90 },
      {     16,     32,     47,     60,     72,     81,     87,     90,     90,     87,     81,     73,     62,     49,     34,     19,      2,    -14,    -30,    -45,    -59,    -70,    -79,    -86,    -90,    -90,    -88,    -82,    -74,    -64,    -51,    -36,    -21,     -4,     12,     28,     43,     57,     69,     78,     85,     89,     90,     88,     83,     76,     65,     53,     38,     23,      7,    -10,    -26,    -41,    -55,    -67,    -77,    -85,    -89,    -90,    -89,    -84,    -77,    -67,    -55,    -40,    -25,     -9,      8,     24,     39,     54,     66,     76,     84,     89,     90,     89,     85,     78,     68,     56,     42,     27,     11,     -6,    -22,    -37,    -52,    -64,    -75,    -83,    -88,    -90,    -89,    -86,    -79,    -70,    -58,    -44,    -29,    -13,      3,     20,     35,     50,     63,     74,     82,     88,     90,     90,     86,     80,     71,     60,     46,     31,     15,     -1,    -18,    -33,    -48,    -61,    -72,    -81,    -87,    -90 },
      {     19,     36,     53,     67,     78,     86,     90,     90,     86,     79,     68,     55,     38,     21,      2,    -16,    -34,    -51,    -65,    -77,    -85,    -89,    -90,    -87,    -80,    -70,    -56,    -40,    -23,     -4,     14,     32,     49,     64,     76,     84,     89,     90,     88,     81,     71,     58,     42,     25,      7,    -12,    -30,    -47,    -62,    -74,    -83,    -89,    -90,    -88,    -82,    -72,    -60,    -44,    -27,     -9,     10,     28,     45,     60,     73,     82,     88,     90,     89,     83,     74,     61,     46,     29,     11,     -8,    -26,    -43,    -59,    -72,    -81,    -88,    -90,    -89,    -84,    -75,    -63,    -48,    -31,    -13,      6,     24,     41,     57,     70,     81,     87,     90,     89,     85,     76,     64,     50,     33,     15,     -3,    -22,    -39,    -55,    -69,    -79,    -87,    -90,    -90,    -85,    -77,    -66,    -52,    -35,    -18,      1,     20,     37,     54,     67,     78,     86,     90 },
      {     21,     40,     58,     72,     83,     89,     90,     87,     78,     66,     50,     31,     11,    -10,    -30,    -49,    -65,    -78,    -86,    -90,    -89,    -83,    -73,    -59,    -41,    -22,     -1,     20,     39,     57,     72,     82,     89,     90,     87,     79,     67,     51,     32,     12,     -9,    -29,    -48,    -64,    -77,    -86,    -90,    -89,    -84,    -74,    -60,    -42,    -23,     -2,     19,     38,     56,     71,     82,     89,     90,     87,     79,     67,     52,     33,     13,     -8,    -28,    -47,    -64,    -77,    -86,    -90,    -89,    -84,    -74,    -60,    -43,    -24,     -3,     18,     37,     55,     70,     81,     88,     90,     88,     80,     68,     53,     34,     14,     -7,    -27,    -46,    -63,    -76,    -85,    -90,    -90,    -85,    -75,    -61,    -44,    -25,     -4,     16,     36,     55,     70,     81,     88,     90,     88,     81,     69,     54,     35,     15,     -6,    -26,    -45,    -62,    -76,    -85,    -90 },
      {     23,     44,     63,     77,     87,     90,     88,     80,     67,     49,     28,      6,    -18,    -39,    -59,    -74,    -85,    -90,    -89,    -82,    -70,    -54,    -33,    -11,     12,     34,     55,     71,     83,     89,     90,     85,     74,     58,     38,     16,     -7,    -29,    -50,    -67,    -81,    -88,    -90,    -86,    -77,    -62,    -43,    -22,      1,     24,     45,     64,     78,     87,     90,     88,     79,     66,     48,     27,      4,    -19,    -40,    -60,    -75,    -85,    -90,    -89,    -82,    -70,    -53,    -32,    -10,     13,     35,     55,     72,     83,     89,     90,     84,     73,     57,     37,     15,     -8,    -30,    -51,    -68,    -81,    -89,    -90,    -86,    -76,    -61,    -42,    -21,      2,     25,     46,     64,     78,     87,     90,     88,     79,     65,     47,     26,      3,    -20,    -41,    -60,    -76,    -86,    -90,    -89,    -81,    -69,    -52,    -31,     -9,     14,     36,     56,     72,     84,     90 },
      {     25,     48,     67,     81,     89,     90,     83,     70,     52,     29,      4,    -21,    -44,    -64,    -79,    -88,    -90,    -85,    -73,    -55,    -33,     -9,     16,     40,     61,     77,     87,     90,     86,     76,     59,     37,     13,    -12,    -36,    -58,    -75,    -86,    -90,    -88,    -78,    -62,    -41,    -18,      8,     32,     55,     72,     85,     90,     89,     80,     65,     45,     22,     -3,    -28,    -51,    -70,    -83,    -90,    -89,    -82,    -68,    -49,    -26,     -1,     24,     47,     67,     81,     89,     90,     84,     71,     53,     30,      6,    -20,    -43,    -64,    -79,    -88,    -90,    -85,    -74,    -56,    -34,    -10,     15,     39,     60,     77,     87,     90,     87,     76,     60,     38,     14,    -11,    -35,    -57,    -74,    -86,    -90,    -88,    -78,    -63,    -42,    -19,      7,     31,     54,     72,     84,     90,     89,     81,     66,     46,     23,     -2,    -27,    -50,    -69,    -82,    -89 },
      {     27,     52,     72,     85,     90,     87,     76,     58,     34,      8,    -20,    -45,    -67,    -82,    -90,    -89,    -80,    -64,    -41,    -15,     12,     38,     61,     78,     88,     90,     83,     69,     48,     23,     -4,    -31,    -55,    -74,    -86,    -90,    -86,    -74,    -55,    -30,     -3,     24,     49,     70,     84,     90,     88,     78,     60,     37,     11,    -16,    -42,    -64,    -81,    -89,    -89,    -81,    -66,    -44,    -19,      9,     35,     59,     77,     88,     90,     85,     71,     51,     26,     -1,    -28,    -53,    -72,    -85,    -90,    -87,    -76,    -57,    -33,     -7,     21,     46,     67,     82,     90,     89,     79,     63,     40,     14,    -13,    -39,    -62,    -79,    -89,    -90,    -83,    -68,    -47,    -22,      6,     32,     56,     75,     87,     90,     86,     73,     54,     29,      2,    -25,    -50,    -70,    -84,    -90,    -88,    -77,    -60,    -36,    -10,     18,     43,     65,     81,     89 },
      {     29,     55,     76,     88,     90,     83,     67,     43,     15,    -14,    -42,    -66,    -82,    -90,    -88,    -76,    -56,    -30,     -1,     28,     55,     75,     87,     90,     83,     67,     44,     16,    -13,    -41,    -65,    -82,    -90,    -88,    -77,    -57,    -31,     -2,     27,     54,     74,     87,     90,     84,     68,     45,     18,    -12,    -40,    -64,    -81,    -90,    -88,    -77,    -58,    -32,     -3,     26,     53,     74,     87,     90,     84,     69,     46,     19,    -11,    -39,    -64,    -81,    -90,    -89,    -78,    -59,    -33,     -4,     25,     52,     73,     86,     90,     85,     70,     47,     20,    -10,    -38,    -63,    -81,    -89,    -89,    -78,    -60,    -34,     -6,     24,     51,     72,     86,     90,     85,     70,     48,     21,     -9,    -37,    -62,    -80,    -89,    -89,    -79,    -60,    -35,     -7,     23,     50,     72,     86,     90,     85,     71,     49,     22,     -8,    -36,    -61,    -79,    -89 },
      {     31,     59,     79,     89,     89,     77,     55,     27,     -4,    -35,    -62,    -81,    -90,    -88,    -74,    -52,    -23,      9,     39,     65,     83,     90,     86,     72,     48,     19,    -13,    -43,    -68,    -85,    -90,    -85,    -69,    -44,    -14,     18,     47,     71,     86,     90,     83,     66,     40,     10,    -22,    -51,    -74,    -87,    -90,    -81,    -63,    -36,     -6,     26,     55,     76,     88,     89,     79,     60,     32,      1,    -30,    -58,    -78,    -89,    -89,    -77,    -56,    -28,      3,     34,     61,     81,     90,     88,     75,     53,     24,     -8,    -38,    -64,    -82,    -90,    -87,    -72,    -49,    -20,     12,     42,     67,     84,     90,     85,     70,     45,     15,    -16,    -46,    -70,    -86,    -90,    -84,    -67,    -41,    -11,     21,     50,     73,     87,     90,     82,     64,     37,      7,    -25,    -54,    -76,    -88,    -90,    -80,    -60,    -33,     -2,     29,     57,     78,     89 },
      {     33,     62,     82,     90,     86,     69,     42,     10,    -24,    -55,    -77,    -89,    -88,    -75,    -51,    -20,     14,     46,     72,     87,     90,     80,     59,     29,     -4,    -37,    -65,    -84,    -90,    -84,    -66,    -38,     -6,     28,     58,     79,     90,     87,     72,     47,     15,    -19,    -50,    -74,    -88,    -89,    -78,    -55,    -25,      9,     41,     68,     85,     90,     82,     63,     34,      1,    -32,    -61,    -81,    -90,    -86,    -70,    -43,    -11,     23,     54,     77,     89,     89,     76,     52,     21,    -13,    -45,    -71,    -87,    -90,    -81,    -60,    -30,      3,     36,     64,     83,     90,     85,     67,     39,      7,    -27,    -57,    -79,    -90,    -88,    -73,    -48,    -16,     18,     49,     74,     88,     89,     78,     56,     26,     -8,    -40,    -67,    -85,    -90,    -83,    -64,    -35,     -2,     31,     60,     81,     90,     86,     70,     44,     12,    -22,    -53,    -76,    -89 },
      {     35,     65,     85,     90,     81,     60,     28,     -8,    -42,    -70,    -87,    -90,    -78,    -54,    -21,     15,     49,     75,     89,     88,     74,     47,     13,    -23,    -55,    -79,    -90,    -86,    -69,    -40,     -6,     30,     61,     82,     90,     84,     64,     33,     -2,    -37,    -67,    -85,    -90,    -81,    -58,    -26,     10,     44,     72,     88,     89,     77,     52,     19,    -18,    -51,    -76,    -89,    -88,    -72,    -45,    -11,     25,     57,     80,     90,     86,     67,     38,      3,    -32,    -63,    -83,    -90,    -83,    -62,    -31,      4,     39,     68,     86,     90,     79,     56,     24,    -12,    -46,    -73,    -88,    -89,    -76,    -50,    -16,     20,     53,     77,     89,     87,     71,     43,      9,    -27,    -59,    -81,    -90,    -85,    -66,    -36,     -1,     34,     64,     84,     90,     82,     60,     29,     -7,    -41,    -70,    -87,    -90,    -78,    -55,    -22,     14,     48,     74,     89 },
      {     37,     68,     87,     89,     76,     49,     13,    -25,    -59,    -82,    -90,    -82,    -60,    -26,     12,     48,     76,     89,     87,     69,     38,      1,    -36,    -67,    -86,    -90,    -77,    -50,    -14,     24,     58,     81,     90,     83,     60,     27,    -11,    -47,    -75,    -89,    -87,    -70,    -39,     -2,     35,     67,     86,     90,     77,     51,     15,    -23,    -57,    -81,    -90,    -83,    -61,    -28,     10,     46,     74,     89,     88,     70,     40,      3,    -34,    -66,    -86,    -90,    -78,    -52,    -16,     22,     56,     81,     90,     84,     62,     29,     -9,    -45,    -74,    -89,    -88,    -71,    -41,     -4,     33,     65,     85,     90,     78,     53,     18,    -21,    -55,    -80,    -90,    -84,    -63,    -30,      8,     44,     73,     89,     88,     72,     42,      6,    -32,    -64,    -85,    -90,    -79,    -54,    -19,     20,     55,     79,     90,     85,     64,     31,     -7,    -43,    -72,    -88 },
      {     39,     71,     88,     88,     70,     37,     -2,    -41,    -72,    -89,    -87,    -68,    -35,      4,     43,     74,     89,     87,     67,     33,     -7,    -45,    -75,    -89,    -86,    -65,    -31,      9,     47,     76,     90,     85,     64,     29,    -11,    -49,    -77,    -90,    -85,    -62,    -27,     13,     51,     78,     90,     84,     60,     25,    -15,    -53,    -79,    -90,    -83,    -59,    -23,     18,     55,     81,     90,     82,     57,     21,    -20,    -56,    -81,    -90,    -81,    -55,    -19,     22,     58,     82,     90,     80,     54,     16,    -24,    -60,    -83,    -90,    -79,    -52,    -14,     26,     61,     84,     90,     78,     50,     12,    -28,    -63,    -85,    -90,    -77,    -48,    -10,     30,     64,     86,     90,     76,     46,      8,    -32,    -66,    -86,    -89,    -74,    -44,     -6,     34,     67,     87,     89,     73,     42,      3,    -36,    -69,    -88,    -89,    -72,    -40,     -1,     38,     70,     88 },
      {     41,     74,     89,     85,     62,     25,    -18,    -56,    -82,    -90,    -78,    -48,     -8,     34,     69,     88,     88,     67,     32,    -10,    -50,    -79,    -90,    -81,    -55,    -15,     27,     64,     86,     89,     72,     39,     -2,    -43,    -75,    -90,    -85,    -60,    -23,     20,     58,     83,     90,     77,     46,      6,    -36,    -70,    -89,    -87,    -66,    -30,     12,     52,     80,     90,     81,     53,     13,    -29,    -65,    -87,    -89,    -71,    -37,      4,     45,     76,     90,     84,     59,     21,    -22,    -60,    -84,    -90,    -76,    -44,     -3,     38,     72,     89,     86,     64,     28,    -14,    -54,    -81,    -90,    -79,    -51,    -11,     31,     67,     87,     88,     70,     35,     -7,    -47,    -77,    -90,    -83,    -57,    -19,     24,     61,     85,     90,     74,     42,      1,    -40,    -73,    -89,    -86,    -63,    -26,     16,     55,     82,     90,     78,     49,      9,    -33,    -68,    -88 },
      {     43,     76,     90,     82,     54,     12,    -32,    -69,    -89,    -86,    -63,    -24,     21,     60,     85,     89,     71,     35,     -9,    -51,    -81,    -90,    -78,    -46,     -3,     40,     74,     90,     83,     56,     15,    -29,    -67,    -88,    -87,    -65,    -27,     18,     58,     84,     90,     73,     38,     -6,    -48,    -79,    -90,    -79,    -49,     -7,     37,     72,     89,     85,     59,     19,    -26,    -64,    -87,    -88,    -67,    -30,     14,     55,     83,     90,     75,     41,     -2,    -45,    -77,    -90,    -81,    -52,    -10,     34,     70,     89,     86,     61,     22,    -23,    -62,    -86,    -89,    -70,    -33,     11,     53,     81,     90,     77,     44,      1,    -42,    -76,    -90,    -82,    -55,    -13,     31,     68,     88,     87,     64,     25,    -20,    -60,    -85,    -89,    -72,    -36,      8,     50,     80,     90,     78,     47,      4,    -39,    -74,    -90,    -84,    -57,    -16,     28,     66,     88 },
      {     45,     78,     90,     78,     44,     -1,    -46,    -79,    -90,    -77,    -43,      2,     47,     79,     90,     77,     42,     -3,    -48,    -80,    -90,    -76,    -41,      4,     49,     81,     90,     76,     40,     -6,    -50,    -81,    -90,    -75,    -39,      7,     51,     81,     90,     74,     38,     -8,    -52,    -82,    -90,    -74,    -37,      9,     53,     82,     90,     73,     36,    -10,    -54,    -83,    -90,    -72,    -35,     11,     55,     83,     90,     72,     34,    -12,    -55,    -84,    -89,    -71,    -33,     13,     56,     84,     89,     70,     32,    -14,    -57,    -85,    -89,    -70,    -31,     15,     58,     85,     89,     69,     30,    -16,    -59,    -85,    -89,    -68,    -29,     18,     60,     86,     89,     67,     28,    -19,    -60,    -86,    -88,    -67,    -27,     20,     61,     86,     88,     66,     26,    -21,    -62,    -87,    -88,    -65,    -25,     22,     63,     87,     88,     64,     24,    -23,    -64,    -87 },
      {     47,     81,     90,     73,     34,    -14,    -59,    -86,    -88,    -64,    -21,     28,     69,     89,     83,     53,      7,    -41,    -77,    -90,    -77,    -40,      8,     54,     84,     89,     68,     27,    -22,    -64,    -88,    -86,    -58,    -13,     35,     74,     90,     80,     46,     -1,    -48,    -81,    -90,    -72,    -33,     15,     60,     86,     88,     63,     20,    -29,    -70,    -89,    -83,    -52,     -6,     42,     78,     90,     76,     39,     -9,    -55,    -84,    -89,    -67,    -26,     23,     65,     88,     85,     57,     12,    -36,    -74,    -90,    -79,    -45,      2,     49,     81,     90,     72,     32,    -16,    -60,    -87,    -87,    -62,    -19,     30,     70,     90,     82,     51,      4,    -43,    -78,    -90,    -76,    -38,     10,     55,     85,     89,     67,     25,    -24,    -66,    -89,    -85,    -56,    -11,     37,     75,     90,     79,     44,     -3,    -50,    -82,    -90,    -71,    -31,     18,     61,     87 },
      {     49,     82,     89,     67,     24,    -27,    -70,    -90,    -81,    -46,      3,     52,     84,     89,     65,     21,    -30,    -72,    -90,    -79,    -43,      7,     55,     85,     88,     63,     18,    -33,    -74,    -90,    -78,    -40,     10,     57,     86,     87,     60,     14,    -36,    -76,    -90,    -76,    -37,     13,     60,     87,     86,     58,     11,    -39,    -77,    -90,    -74,    -34,     16,     62,     88,     85,     55,      8,    -42,    -79,    -90,    -72,    -31,     20,     64,     89,     84,     53,      4,    -45,    -81,    -90,    -70,    -28,     23,     67,     89,     83,     50,      1,    -48,    -82,    -89,    -68,    -25,     26,     69,     90,     81,     47,     -2,    -51,    -83,    -89,    -66,    -22,     29,     71,     90,     80,     44,     -6,    -54,    -85,    -88,    -64,    -19,     32,     73,     90,     78,     41,     -9,    -56,    -86,    -88,    -61,    -15,     35,     75,     90,     77,     38,    -12,    -59,    -87 },
      {     51,     84,     88,     61,     13,    -39,    -78,    -90,    -70,    -26,     27,     71,     90,     78,     38,    -14,    -62,    -88,    -84,    -50,      1,     52,     85,     88,     60,     12,    -40,    -79,    -90,    -70,    -25,     28,     72,     90,     77,     37,    -15,    -63,    -89,    -83,    -49,      2,     53,     85,     88,     60,     11,    -41,    -79,    -90,    -69,    -24,     29,     72,     90,     77,     36,    -16,    -64,    -89,    -83,    -48,      3,     54,     85,     87,     59,     10,    -42,    -80,    -90,    -68,    -23,     30,     73,     90,     76,     35,    -18,    -64,    -89,    -82,    -47,      4,     55,     86,     87,     58,      9,    -43,    -81,    -90,    -67,    -22,     31,     74,     90,     76,     34,    -19,    -65,    -89,    -82,    -46,      6,     55,     86,     87,     57,      8,    -44,    -81,    -89,    -67,    -21,     32,     74,     90,     75,     33,    -20,    -66,    -89,    -81,    -45,      7,     56,     86 },
      {     53,     86,     86,     55,      2,    -51,    -85,    -87,    -56,     -4,     49,     84,     88,     58,      7,    -47,    -83,    -88,    -60,     -9,     45,     82,     89,     61,     11,    -43,    -81,    -89,    -63,    -13,     41,     81,     89,     64,     15,    -39,    -79,    -90,    -66,    -18,     37,     78,     90,     67,     20,    -35,    -77,    -90,    -69,    -22,     33,     76,     90,     70,     24,    -31,    -75,    -90,    -72,    -26,     29,     74,     90,     73,     28,    -27,    -72,    -90,    -74,    -30,     25,     71,     90,     76,     32,    -23,    -70,    -90,    -77,    -34,     21,     68,     90,     78,     36,    -19,    -67,    -90,    -79,    -38,     16,     65,     89,     80,     40,    -14,    -64,    -89,    -81,    -42,     12,     62,     89,     82,     44,    -10,    -60,    -88,    -83,    -46,      8,     59,     88,     84,     48,     -6,    -57,    -87,    -85,    -50,      3,     55,     87,     85,     52,     -1,    -54,    -86 },
      {     55,     87,     84,     47,     -9,    -61,    -89,    -81,    -39,     18,     67,     90,     76,     31,    -26,    -73,    -90,    -71,    -23,     34,     78,     90,     65,     14,    -42,    -82,    -88,    -59,     -6,     50,     85,     86,     52,     -3,    -57,    -88,    -83,    -44,     12,     64,     89,     79,     36,    -21,    -70,    -90,    -74,    -28,     29,     75,     90,     69,     20,    -37,    -79,    -89,    -63,    -11,     45,     83,     88,     56,      2,    -53,    -86,    -85,    -49,      7,     60,     89,     81,     41,    -15,    -66,    -90,    -77,    -33,     24,     72,     90,     72,     25,    -32,    -77,    -90,    -67,    -16,     40,     81,     89,     60,      8,    -48,    -85,    -87,    -54,      1,     55,     87,     84,     46,    -10,    -62,    -89,    -80,    -38,     19,     68,     90,     76,     30,    -27,    -74,    -90,    -70,    -22,     35,     78,     90,     64,     13,    -43,    -82,    -88,    -58,     -4,     51,     86 },
      {     56,     88,     81,     39,    -20,    -70,    -90,    -71,    -21,     38,     81,     88,     57,      1,    -55,    -88,    -82,    -40,     19,     70,     90,     72,     22,    -37,    -81,    -89,    -58,     -2,     55,     88,     82,     41,    -18,    -69,    -90,    -72,    -23,     36,     80,     89,     59,      3,    -54,    -87,    -83,    -42,     16,     68,     90,     73,     24,    -35,    -79,    -89,    -60,     -4,     53,     87,     83,     43,    -15,    -67,    -90,    -74,    -25,     34,     79,     89,     60,      6,    -52,    -87,    -84,    -44,     14,     67,     90,     74,     26,    -33,    -78,    -89,    -61,     -7,     51,     86,     84,     45,    -13,    -66,    -90,    -75,    -27,     32,     78,     89,     62,      8,    -50,    -86,    -85,    -46,     12,     65,     90,     76,     28,    -31,    -77,    -90,    -63,     -9,     49,     86,     85,     47,    -11,    -64,    -90,    -76,    -29,     30,     77,     90,     64,     10,    -48,    -85 },
      {     58,     89,     78,     31,    -30,    -78,    -89,    -59,     -1,     57,     89,     79,     32,    -29,    -77,    -89,    -60,     -2,     56,     89,     79,     33,    -28,    -77,    -89,    -60,     -3,     55,     88,     80,     34,    -27,    -76,    -90,    -61,     -4,     55,     88,     81,     35,    -26,    -76,    -90,    -62,     -6,     54,     88,     81,     36,    -25,    -75,    -90,    -63,     -7,     53,     88,     81,     37,    -24,    -74,    -90,    -64,     -8,     52,     87,     82,     38,    -23,    -74,    -90,    -64,     -9,     51,     87,     82,     39,    -22,    -73,    -90,    -65,    -10,     50,     87,     83,     40,    -21,    -72,    -90,    -66,    -11,     49,     86,     83,     41,    -20,    -72,    -90,    -67,    -12,     48,     86,     84,     42,    -19,    -71,    -90,    -67,    -13,     47,     86,     84,     43,    -18,    -70,    -90,    -68,    -14,     46,     85,     85,     44,    -16,    -70,    -90,    -69,    -15,     45,     85 },
      {     60,     90,     75,     23,    -40,    -84,    -85,    -44,     19,     72,     90,     63,      4,    -56,    -89,    -77,    -27,     36,     82,     87,     48,    -14,    -70,    -90,    -66,     -9,     53,     88,     79,     31,    -32,    -80,    -88,    -52,     10,     67,     90,     69,     13,    -49,    -87,    -81,    -35,     28,     78,     89,     55,     -6,    -64,    -90,    -72,    -18,     45,     86,     83,     39,    -24,    -76,    -89,    -59,      1,     60,     90,     74,     22,    -41,    -84,    -85,    -43,     20,     73,     90,     62,      3,    -57,    -89,    -77,    -26,     37,     82,     86,     47,    -15,    -70,    -90,    -65,     -8,     54,     88,     79,     30,    -33,    -81,    -88,    -51,     11,     67,     90,     68,     12,    -50,    -87,    -81,    -34,     29,     78,     89,     55,     -7,    -64,    -90,    -71,    -16,     46,     86,     83,     38,    -25,    -76,    -89,    -58,      2,     61,     90,     74,     21,    -42,    -85 },
      {     61,     90,     71,     14,    -50,    -88,    -79,    -28,     37,     83,     85,     41,    -24,    -77,    -89,    -54,     10,     68,     90,     64,      4,    -58,    -90,    -74,    -19,     46,     87,     81,     32,    -33,    -81,    -86,    -45,     20,     74,     89,     57,     -6,    -65,    -90,    -67,     -9,     55,     89,     76,     23,    -42,    -85,    -83,    -36,     29,     79,     88,     49,    -15,    -72,    -90,    -60,      1,     62,     90,     70,     13,    -51,    -88,    -78,    -27,     38,     84,     85,     40,    -25,    -77,    -89,    -53,     11,     69,     90,     64,      3,    -59,    -90,    -73,    -18,     47,     87,     81,     31,    -34,    -82,    -86,    -44,     21,     75,     89,     56,     -7,    -66,    -90,    -67,     -8,     55,     89,     76,     22,    -43,    -86,    -82,    -35,     30,     80,     87,     48,    -16,    -72,    -90,    -60,      2,     63,     90,     70,     12,    -52,    -88,    -78,    -26,     39,     84 },
      {     63,     90,     67,      6,    -59,    -90,    -70,    -11,     55,     89,     74,     16,    -50,    -88,    -77,    -22,     45,     87,     79,     27,    -40,    -85,    -82,    -32,     35,     83,     84,     37,    -30,    -81,    -86,    -42,     25,     78,     88,     47,    -20,    -76,    -89,    -52,     14,     72,     90,     56,     -9,    -69,    -90,    -60,      3,     65,     90,     64,      2,    -61,    -90,    -68,     -8,     57,     90,     72,     13,    -53,    -89,    -75,    -19,     48,     88,     78,     24,    -43,    -86,    -81,    -29,     38,     85,     83,     34,    -33,    -82,    -85,    -39,     28,     80,     87,     44,    -23,    -77,    -88,    -49,     18,     74,     89,     54,    -12,    -71,    -90,    -58,      7,     67,     90,     62,     -1,    -64,    -90,    -66,     -4,     60,     90,     70,     10,    -55,    -89,    -73,    -15,     51,     89,     76,     21,    -46,    -87,    -79,    -26,     41,     86,     81,     31,    -36,    -84 },
      {     64,     90,     62,     -3,    -67,    -90,    -60,      7,     69,     90,     57,    -10,    -71,    -90,    -55,     13,     73,     89,     52,    -16,    -75,    -89,    -49,     20,     77,     88,     46,    -23,    -78,    -87,    -43,     26,     80,     86,     40,    -29,    -81,    -85,    -37,     32,     83,     84,     34,    -35,    -84,    -82,    -31,     38,     85,     81,     28,    -41,    -86,    -79,    -25,     44,     87,     78,     22,    -47,    -88,    -76,    -19,     50,     89,     74,     15,    -53,    -89,    -72,    -12,     55,     90,     70,      9,    -58,    -90,    -68,     -6,     60,     90,     66,      2,    -63,    -90,    -64,      1,     65,     90,     61,     -4,    -67,    -90,    -59,      8,     70,     90,     56,    -11,    -72,    -89,    -54,     14,     74,     89,     51,    -18,    -76,    -88,    -48,     21,     77,     88,     45,    -24,    -79,    -87,    -42,     27,     81,     86,     39,    -30,    -82,    -85,    -36,     33,     83 },
      {     66,     90,     57,    -12,    -74,    -89,    -47,     24,     80,     85,     36,    -35,    -85,    -81,    -25,     46,     88,     74,     13,    -56,    -90,    -67,     -1,     65,     90,     58,    -11,    -73,    -89,    -48,     23,     79,     86,     37,    -34,    -85,    -81,    -26,     45,     88,     75,     14,    -55,    -90,    -67,     -2,     64,     90,     59,    -10,    -72,    -89,    -49,     22,     79,     86,     38,    -33,    -84,    -81,    -27,     44,     88,     76,     15,    -55,    -90,    -68,     -3,     64,     90,     60,     -9,    -72,    -89,    -50,     21,     78,     86,     39,    -32,    -84,    -82,    -28,     43,     88,     76,     16,    -54,    -90,    -69,     -4,     63,     90,     60,     -8,    -71,    -89,    -51,     20,     78,     87,     40,    -31,    -83,    -82,    -29,     42,     87,     77,     18,    -53,    -90,    -70,     -6,     62,     90,     61,     -7,    -70,    -89,    -52,     19,     77,     87,     41,    -30,    -83 },
      {     67,     90,     52,    -21,    -79,    -85,    -33,     40,     87,     76,     13,    -58,    -90,    -62,      8,     72,     89,     45,    -28,    -83,    -82,    -26,     47,     89,     71,      6,    -64,    -90,    -56,     15,     77,     87,     38,    -35,    -86,    -78,    -19,     54,     90,     66,     -2,    -69,    -89,    -50,     23,     81,     84,     31,    -42,    -88,    -74,    -11,     60,     90,     60,    -10,    -74,    -88,    -43,     30,     84,     81,     24,    -49,    -89,    -70,     -3,     65,     90,     55,    -18,    -78,    -86,    -36,     37,     86,     77,     16,    -55,    -90,    -64,      4,     70,     89,     48,    -25,    -81,    -83,    -29,     44,     88,     73,      9,    -61,    -90,    -59,     12,     75,     88,     41,    -32,    -85,    -80,    -22,     51,     90,     68,      1,    -67,    -90,    -53,     20,     79,     85,     34,    -39,    -87,    -76,    -14,     57,     90,     63,     -7,    -72,    -89,    -46,     27,     82 },
      {     69,     89,     46,    -29,    -84,    -79,    -19,     55,     90,     61,    -11,    -76,    -87,    -36,     39,     88,     74,      8,    -64,    -90,    -53,     22,     81,     83,     26,    -49,    -90,    -67,      3,     71,     89,     43,    -32,    -85,    -78,    -15,     58,     90,     59,    -14,    -77,    -86,    -33,     42,     88,     72,      4,    -66,    -90,    -50,     25,     82,     81,     23,    -52,    -90,    -64,      7,     73,     88,     40,    -35,    -86,    -76,    -12,     60,     90,     56,    -18,    -79,    -85,    -30,     45,     89,     70,      1,    -68,    -89,    -47,     28,     84,     80,     20,    -55,    -90,    -62,     10,     75,     87,     37,    -38,    -87,    -74,     -9,     63,     90,     54,    -21,    -81,    -83,    -27,     48,     89,     67,     -2,    -70,    -89,    -44,     31,     85,     78,     16,    -57,    -90,    -60,     13,     77,     86,     34,    -41,    -88,    -72,     -6,     65,     90,     51,    -24,    -82 },
      {     70,     88,     40,    -37,    -88,    -72,     -3,     68,     89,     43,    -34,    -87,    -74,     -7,     66,     89,     46,    -31,    -86,    -76,    -10,     64,     90,     49,    -28,    -85,    -78,    -13,     61,     90,     52,    -25,    -83,    -79,    -16,     59,     90,     55,    -22,    -82,    -81,    -20,     56,     90,     57,    -19,    -81,    -82,    -23,     54,     90,     60,    -15,    -79,    -84,    -26,     51,     90,     62,    -12,    -77,    -85,    -29,     48,     90,     64,     -9,    -76,    -86,    -32,     45,     89,     67,     -6,    -74,    -87,    -35,     42,     89,     69,     -2,    -72,    -88,    -38,     39,     88,     71,      1,    -70,    -89,    -41,     36,     87,     73,      4,    -67,    -89,    -44,     33,     86,     75,      8,    -65,    -90,    -47,     30,     85,     77,     11,    -63,    -90,    -50,     27,     84,     78,     14,    -60,    -90,    -53,     24,     83,     80,     18,    -58,    -90,    -55,     21,     81 },
      {     72,     87,     34,    -45,    -90,    -64,     12,     78,     83,     23,    -55,    -90,    -55,     24,     84,     78,     11,    -64,    -89,    -44,     35,     88,     71,     -1,    -72,    -87,    -33,     46,     90,     63,    -13,    -79,    -83,    -22,     56,     90,     54,    -25,    -84,    -77,    -10,     65,     89,     43,    -36,    -88,    -70,      2,     73,     87,     32,    -47,    -90,    -62,     14,     79,     82,     21,    -57,    -90,    -53,     26,     85,     77,      9,    -66,    -89,    -42,     37,     88,     70,     -3,    -74,    -86,    -31,     48,     90,     61,    -15,    -80,    -82,    -20,     58,     90,     52,    -27,    -85,    -76,     -8,     67,     89,     41,    -38,    -88,    -69,      4,     74,     86,     30,    -49,    -90,    -60,     16,     81,     81,     19,    -59,    -90,    -51,     28,     85,     76,      7,    -67,    -89,    -40,     39,     89,     68,     -6,    -75,    -86,    -29,     50,     90,     60,    -18,    -81 },
      {     73,     86,     28,    -53,    -90,    -54,     27,     86,     74,      1,    -72,    -86,    -29,     52,     90,     55,    -26,    -85,    -74,     -2,     72,     87,     30,    -51,    -90,    -55,     25,     85,     75,      3,    -71,    -87,    -31,     50,     90,     56,    -24,    -85,    -76,     -4,     70,     87,     32,    -49,    -90,    -57,     23,     84,     76,      6,    -70,    -88,    -33,     48,     90,     58,    -22,    -84,    -77,     -7,     69,     88,     34,    -47,    -90,    -59,     21,     83,     77,      8,    -68,    -88,    -35,     46,     90,     60,    -20,    -83,    -78,     -9,     67,     88,     36,    -45,    -90,    -60,     19,     82,     78,     10,    -67,    -89,    -37,     44,     90,     61,    -18,    -82,    -79,    -11,     66,     89,     38,    -43,    -90,    -62,     16,     81,     79,     12,    -65,    -89,    -39,     42,     89,     63,    -15,    -81,    -80,    -13,     64,     89,     40,    -41,    -89,    -64,     14,     81 },
      {     74,     85,     22,    -60,    -90,    -42,     41,     90,     60,    -21,    -84,    -75,     -1,     74,     85,     23,    -59,    -90,    -43,     40,     89,     61,    -20,    -84,    -76,     -2,     73,     85,     24,    -58,    -90,    -44,     39,     89,     62,    -19,    -83,    -76,     -3,     72,     86,     25,    -57,    -90,    -45,     38,     89,     63,    -18,    -83,    -77,     -4,     72,     86,     26,    -56,    -90,    -46,     37,     89,     64,    -16,    -82,    -77,     -6,     71,     86,     27,    -55,    -90,    -47,     36,     89,     64,    -15,    -82,    -78,     -7,     70,     87,     28,    -55,    -90,    -48,     35,     89,     65,    -14,    -81,    -78,     -8,     70,     87,     29,    -54,    -90,    -49,     34,     88,     66,    -13,    -81,    -79,     -9,     69,     87,     30,    -53,    -90,    -50,     33,     88,     67,    -12,    -81,    -79,    -10,     68,     88,     31,    -52,    -90,    -51,     32,     88,     67,    -11,    -80 },
      {     76,     83,     15,    -66,    -88,    -30,     55,     90,     44,    -41,    -90,    -57,     27,     87,     68,    -12,    -81,    -77,     -3,     74,     84,     19,    -64,    -89,    -33,     52,     90,     47,    -38,    -89,    -60,     24,     86,     70,     -9,    -80,    -79,     -7,     72,     85,     22,    -61,    -89,    -36,     49,     90,     50,    -35,    -89,    -62,     21,     85,     72,     -6,    -78,    -81,    -10,     70,     86,     25,    -59,    -90,    -39,     46,     90,     53,    -32,    -88,    -64,     18,     84,     74,     -2,    -77,    -82,    -13,     67,     87,     28,    -56,    -90,    -42,     43,     90,     55,    -29,    -88,    -67,     14,     82,     76,      1,    -75,    -83,    -16,     65,     88,     31,    -54,    -90,    -45,     40,     90,     58,    -26,    -87,    -69,     11,     81,     78,      4,    -73,    -85,    -20,     63,     89,     34,    -51,    -90,    -48,     37,     89,     60,    -23,    -86,    -71,      8,     79 },
      {     77,     81,      9,    -72,    -85,    -18,     66,     87,     26,    -60,    -89,    -34,     53,     90,     42,    -45,    -90,    -50,     37,     90,     57,    -29,    -88,    -64,     21,     86,     70,    -12,    -82,    -75,      3,     78,     79,      6,    -74,    -83,    -14,     68,     86,     23,    -62,    -89,    -31,     55,     90,     39,    -48,    -90,    -47,     40,     90,     55,    -32,    -89,    -61,     24,     87,     67,    -15,    -84,    -73,      7,     80,     78,      2,    -76,    -82,    -11,     70,     85,     20,    -64,    -88,    -28,     58,     89,     36,    -51,    -90,    -44,     43,     90,     52,    -35,    -89,    -59,     27,     88,     65,    -19,    -85,    -71,     10,     81,     76,     -1,    -77,    -81,     -8,     72,     84,     16,    -67,    -87,    -25,     60,     89,     33,    -54,    -90,    -41,     46,     90,     49,    -38,    -90,    -56,     30,     88,     63,    -22,    -86,    -69,     13,     83,     74,     -4,    -79 },
      {     78,     79,      2,    -77,    -80,     -4,     76,     81,      7,    -74,    -82,     -9,     73,     83,     11,    -72,    -84,    -13,     70,     85,     15,    -69,    -85,    -18,     67,     86,     20,    -66,    -87,    -22,     64,     87,     24,    -63,    -88,    -26,     61,     88,     28,    -60,    -89,    -30,     58,     89,     32,    -56,    -89,    -34,     55,     90,     36,    -53,    -90,    -38,     51,     90,     40,    -49,    -90,    -42,     47,     90,     44,    -45,    -90,    -46,     43,     90,     48,    -41,    -90,    -50,     39,     90,     52,    -37,    -90,    -54,     35,     90,     55,    -33,    -89,    -57,     31,     89,     59,    -29,    -89,    -60,     27,     88,     62,    -25,    -88,    -64,     23,     87,     65,    -21,    -86,    -67,     19,     86,     68,    -16,    -85,    -70,     14,     84,     71,    -12,    -83,    -72,     10,     82,     74,     -8,    -81,    -75,      6,     81,     76,     -3,    -79,    -77,      1,     78 },
      {     79,     77,     -4,    -81,    -74,      9,     83,     72,    -13,    -85,    -69,     18,     86,     66,    -22,    -87,    -63,     26,     88,     60,    -30,    -89,    -56,     34,     90,     53,    -38,    -90,    -49,     42,     90,     45,    -46,    -90,    -41,     50,     90,     37,    -54,    -90,    -33,     57,     89,     29,    -60,    -88,    -25,     64,     87,     21,    -67,    -86,    -16,     70,     84,     12,    -72,    -82,     -8,     75,     81,      3,    -77,    -78,      1,     79,     76,     -6,    -81,    -74,     10,     83,     71,    -14,    -85,    -68,     19,     86,     65,    -23,    -88,    -62,     27,     89,     59,    -31,    -89,    -55,     35,     90,     52,    -39,    -90,    -48,     43,     90,     44,    -47,    -90,    -40,     51,     90,     36,    -55,    -89,    -32,     58,     89,     28,    -61,    -88,    -24,     64,     87,     20,    -67,    -85,    -15,     70,     84,     11,    -73,    -82,     -7,     76,     80,      2,    -78 },
      {     80,     74,    -11,    -85,    -67,     22,     88,     60,    -32,    -90,    -51,     42,     90,     41,    -52,    -90,    -31,     60,     88,     21,    -68,    -84,    -10,     75,     79,     -1,    -81,    -74,     12,     85,     67,    -23,    -88,    -59,     33,     90,     50,    -43,    -90,    -40,     53,     89,     30,    -61,    -87,    -20,     69,     84,      9,    -76,    -79,      2,     81,     73,    -13,    -85,    -66,     24,     88,     58,    -34,    -90,    -49,     44,     90,     39,    -54,    -89,    -29,     62,     87,     19,    -70,    -83,     -8,     76,     78,     -3,    -81,    -72,     14,     86,     65,    -25,    -89,    -57,     35,     90,     48,    -45,    -90,    -38,     55,     89,     28,    -63,    -87,    -18,     70,     83,      7,    -77,    -78,      4,     82,     72,    -15,    -86,    -64,     26,     89,     56,    -36,    -90,    -47,     46,     90,     37,    -55,    -89,    -27,     64,     86,     16,    -71,    -82,     -6,     77 },
      {     81,     72,    -18,    -87,    -60,     34,     90,     45,    -50,    -90,    -29,     64,     86,     12,    -75,    -78,      6,     83,     68,    -23,    -89,    -55,     39,     90,     40,    -55,    -89,    -24,     67,     84,      7,    -78,    -76,     11,     85,     64,    -28,    -89,    -51,     44,     90,     35,    -59,    -88,    -19,     71,     81,      1,    -81,    -72,     16,     87,     60,    -33,    -90,    -46,     49,     90,     30,    -63,    -86,    -13,     74,     79,     -4,    -83,    -69,     22,     88,     56,    -38,    -90,    -41,     54,     89,     25,    -67,    -84,     -8,     77,     76,    -10,    -85,    -65,     27,     89,     52,    -43,    -90,    -36,     58,     88,     20,    -70,    -82,     -2,     80,     73,    -15,    -87,    -61,     32,     90,     47,    -48,    -90,    -31,     62,     86,     14,    -74,    -79,      3,     82,     70,    -21,    -88,    -57,     37,     90,     42,    -53,    -89,    -26,     66,     85,      9,    -77 },
      {     82,     69,    -24,    -89,    -51,     46,     90,     29,    -65,    -84,     -6,     79,     72,    -19,    -88,    -55,     41,     90,     34,    -61,    -86,    -11,     77,     76,    -13,    -87,    -60,     36,     90,     39,    -57,    -88,    -16,     74,     78,     -8,    -85,    -64,     31,     90,     44,    -53,    -89,    -22,     70,     81,     -2,    -83,    -67,     26,     89,     49,    -48,    -90,    -27,     67,     83,      3,    -81,    -71,     21,     89,     54,    -43,    -90,    -32,     63,     85,      9,    -78,    -74,     15,     87,     58,    -38,    -90,    -37,     59,     87,     14,    -75,    -77,     10,     86,     62,    -33,    -90,    -42,     55,     88,     20,    -72,    -80,      4,     84,     66,    -28,    -90,    -47,     50,     89,     25,    -68,    -82,     -1,     81,     70,    -23,    -89,    -52,     45,     90,     30,    -64,    -85,     -7,     79,     73,    -18,    -88,    -56,     40,     90,     35,    -60,    -86,    -12,     76 },
      {     83,     66,    -30,    -90,    -41,     57,     87,     12,    -77,    -74,     19,     89,     52,    -47,    -89,    -24,     70,     80,     -7,    -85,    -61,     36,     90,     35,    -62,    -85,     -6,     81,     70,    -25,    -90,    -46,     53,     88,     18,    -74,    -77,     13,     87,     56,    -42,    -90,    -29,     67,     82,     -1,    -83,    -65,     31,     90,     40,    -58,    -87,    -11,     78,     73,    -20,    -89,    -51,     48,     89,     23,    -71,    -79,      8,     86,     60,    -37,    -90,    -34,     63,     85,      4,    -81,    -69,     26,     90,     45,    -54,    -88,    -16,     75,     76,    -14,    -88,    -55,     43,     90,     28,    -67,    -82,      2,     84,     64,    -32,    -90,    -39,     59,     86,     10,    -78,    -72,     21,     89,     50,    -49,    -89,    -22,     72,     79,     -9,    -86,    -60,     38,     90,     33,    -64,    -84,     -3,     81,     68,    -27,    -90,    -44,     55,     88,     15,    -76 },
      {     84,     63,    -36,    -90,    -31,     67,     81,     -6,    -86,    -59,     41,     90,     26,    -70,    -79,     11,     87,     55,    -46,    -89,    -21,     74,     76,    -16,    -89,    -50,     51,     88,     15,    -77,    -73,     22,     89,     45,    -55,    -87,    -10,     79,     70,    -27,    -90,    -40,     60,     85,      4,    -82,    -66,     32,     90,     35,    -64,    -83,      1,     84,     62,    -37,    -90,    -30,     67,     81,     -7,    -86,    -58,     42,     90,     25,    -71,    -78,     12,     88,     54,    -47,    -89,    -20,     74,     76,    -18,    -89,    -49,     52,     88,     14,    -77,    -72,     23,     90,     44,    -56,    -87,     -9,     80,     69,    -28,    -90,    -39,     60,     85,      3,    -82,    -65,     33,     90,     34,    -64,    -83,      2,     85,     61,    -38,    -90,    -29,     68,     81,     -8,    -86,    -57,     43,     90,     24,    -72,    -78,     13,     88,     53,    -48,    -89,    -19,     75 },
      {     85,     60,    -42,    -90,    -21,     75,     74,    -23,    -90,    -40,     61,     84,     -2,    -85,    -58,     44,     89,     19,    -76,    -72,     25,     90,     38,    -63,    -83,      4,     86,     56,    -46,    -89,    -16,     77,     71,    -27,    -90,    -36,     64,     82,     -7,    -87,    -55,     48,     89,     14,    -78,    -70,     29,     90,     34,    -66,    -81,      9,     87,     53,    -50,    -88,    -12,     79,     68,    -31,    -90,    -32,     67,     80,    -11,    -88,    -51,     52,     88,     10,    -81,    -67,     33,     90,     30,    -69,    -79,     13,     88,     49,    -54,    -87,     -8,     81,     65,    -35,    -90,    -28,     70,     78,    -15,    -89,    -47,     55,     86,      6,    -82,    -64,     37,     90,     26,    -72,    -77,     18,     89,     45,    -57,    -86,     -3,     83,     62,    -39,    -90,    -24,     73,     76,    -20,    -89,    -43,     59,     85,      1,    -84,    -60,     41,     90,     22,    -74 },
      {     85,     56,    -48,    -88,    -10,     81,     64,    -39,    -90,    -20,     77,     70,    -30,    -90,    -29,     71,     76,    -21,    -90,    -38,     64,     81,    -11,    -88,    -47,     57,     85,     -1,    -86,    -55,     49,     88,      9,    -82,    -63,     40,     90,     19,    -77,    -70,     31,     90,     28,    -72,    -76,     22,     90,     37,    -65,    -81,     12,     89,     46,    -58,    -85,      2,     86,     55,    -50,    -88,     -8,     82,     62,    -41,    -89,    -18,     78,     69,    -32,    -90,    -27,     72,     75,    -23,    -90,    -36,     66,     80,    -13,    -89,    -45,     59,     84,     -3,    -86,    -54,     51,     87,      7,    -83,    -61,     42,     89,     16,    -78,    -68,     33,     90,     26,    -73,    -74,     24,     90,     35,    -67,    -79,     14,     89,     44,    -60,    -84,      4,     87,     53,    -52,    -87,     -6,     83,     60,    -43,    -89,    -15,     79,     67,    -34,    -90,    -25,     74 },
      {     86,     53,    -54,    -86,      1,     86,     52,    -55,    -85,      2,     87,     51,    -55,    -85,      3,     87,     50,    -56,    -85,      4,     87,     49,    -57,    -84,      6,     88,     48,    -58,    -84,      7,     88,     47,    -59,    -83,      8,     88,     46,    -60,    -83,      9,     88,     45,    -60,    -82,     10,     89,     44,    -61,    -82,     11,     89,     43,    -62,    -81,     12,     89,     42,    -63,    -81,     13,     89,     41,    -64,    -81,     14,     89,     40,    -64,    -80,     15,     89,     39,    -65,    -79,     16,     90,     38,    -66,    -79,     18,     90,     37,    -67,    -78,     19,     90,     36,    -67,    -78,     20,     90,     35,    -68,    -77,     21,     90,     34,    -69,    -77,     22,     90,     33,    -70,    -76,     23,     90,     32,    -70,    -76,     24,     90,     31,    -71,    -75,     25,     90,     30,    -72,    -74,     26,     90,     29,    -72,    -74,     27,     90,     28,    -73 },
      {     87,     49,    -59,    -82,     12,     89,     38,    -67,    -77,     24,     90,     27,    -75,    -70,     35,     90,     15,    -81,    -61,     46,     88,      3,    -86,    -52,     56,     84,     -9,    -89,    -41,     65,     78,    -21,    -90,    -30,     73,     72,    -32,    -90,    -19,     79,     64,    -43,    -88,     -7,     85,     55,    -54,    -85,      6,     88,     44,    -63,    -80,     18,     90,     33,    -71,    -74,     29,     90,     22,    -78,    -66,     40,     89,     10,    -83,    -57,     51,     86,     -2,    -87,    -47,     60,     81,    -14,    -90,    -36,     69,     76,    -26,    -90,    -25,     76,     68,    -37,    -89,    -13,     82,     60,    -48,    -87,     -1,     86,     50,    -58,    -83,     11,     89,     39,    -67,    -77,     23,     90,     28,    -74,    -70,     34,     90,     16,    -81,    -62,     45,     88,      4,    -85,    -53,     55,     84,     -8,    -89,    -42,     64,     79,    -20,    -90,    -31,     72 },
      {     87,     45,    -64,    -78,     23,     90,     24,    -78,    -64,     44,     88,      1,    -87,    -46,     63,     79,    -22,    -90,    -25,     77,     65,    -43,    -88,     -2,     87,     47,    -62,    -79,     21,     90,     26,    -77,    -66,     42,     88,      3,    -86,    -48,     61,     80,    -20,    -90,    -27,     76,     67,    -41,    -88,     -4,     86,     49,    -60,    -81,     19,     90,     28,    -76,    -67,     40,     89,      6,    -86,    -50,     60,     81,    -18,    -90,    -29,     75,     68,    -39,    -89,     -7,     85,     51,    -59,    -81,     16,     90,     30,    -74,    -69,     38,     89,      8,    -85,    -52,     58,     82,    -15,    -90,    -31,     74,     70,    -37,    -89,     -9,     85,     53,    -57,    -82,     14,     90,     32,    -73,    -70,     36,     89,     10,    -84,    -54,     56,     83,    -13,    -90,    -33,     72,     71,    -35,    -89,    -11,     84,     55,    -55,    -83,     12,     90,     34,    -72 },
      {     88,     41,    -68,    -74,     33,     89,      9,    -85,    -49,     62,     78,    -25,    -90,    -18,     82,     56,    -55,    -82,     16,     90,     26,    -78,    -63,     48,     86,     -8,    -89,    -34,     73,     69,    -40,    -88,     -1,     88,     42,    -67,    -74,     32,     90,     10,    -85,    -50,     61,     79,    -24,    -90,    -19,     81,     57,    -55,    -83,     15,     90,     27,    -77,    -64,     47,     86,     -7,    -89,    -35,     72,     70,    -39,    -88,     -2,     87,     43,    -67,    -75,     31,     90,     11,    -85,    -51,     60,     79,    -23,    -90,    -20,     81,     58,    -54,    -83,     14,     90,     28,    -77,    -64,     46,     86,     -6,    -89,    -36,     72,     70,    -38,    -89,     -3,     87,     44,    -66,    -76,     30,     90,     12,    -84,    -52,     60,     80,    -22,    -90,    -21,     81,     59,    -53,    -84,     13,     90,     29,    -76,    -65,     45,     87,     -4,    -89,    -37,     71 },
      {     88,     37,    -72,    -68,     43,     87,     -7,    -89,    -31,     76,     64,    -49,    -85,     13,     90,     25,    -79,    -59,     55,     82,    -20,    -90,    -19,     82,     54,    -60,    -79,     26,     90,     12,    -85,    -48,     64,     76,    -32,    -89,     -6,     87,     42,    -69,    -72,     38,     88,     -1,    -89,    -36,     73,     67,    -44,    -86,      8,     90,     30,    -77,    -63,     50,     84,    -14,    -90,    -24,     80,     58,    -55,    -81,     21,     90,     18,    -83,    -53,     60,     78,    -27,    -90,    -11,     85,     47,    -65,    -75,     33,     89,      4,    -87,    -41,     70,     71,    -39,    -88,      2,     89,     35,    -74,    -67,     45,     86,     -9,    -90,    -29,     77,     62,    -51,    -84,     15,     90,     23,    -81,    -57,     56,     81,    -22,    -90,    -16,     83,     52,    -61,    -78,     28,     90,     10,    -86,    -46,     66,     74,    -34,    -89,     -3,     88,     40,    -70 },
      {     89,     33,    -76,    -62,     53,     82,    -22,    -90,    -12,     86,     44,    -69,    -70,     42,     86,    -10,    -90,    -24,     81,     55,    -60,    -77,     31,     89,      2,    -88,    -35,     75,     64,    -51,    -83,     20,     90,     14,    -85,    -46,     67,     72,    -40,    -87,      8,     90,     26,    -80,    -56,     59,     78,    -29,    -89,     -4,     88,     37,    -74,    -65,     49,     84,    -18,    -90,    -16,     84,     48,    -66,    -73,     38,     88,     -6,    -90,    -28,     79,     58,    -57,    -79,     27,     90,      7,    -87,    -39,     72,     67,    -47,    -85,     15,     90,     19,    -83,    -50,     64,     74,    -36,    -88,      3,     89,     30,    -78,    -60,     55,     81,    -25,    -90,     -9,     87,     41,    -71,    -68,     45,     85,    -13,    -90,    -21,     82,     52,    -63,    -76,     34,     89,     -1,    -89,    -32,     77,     61,    -54,    -81,     23,     90,     11,    -86,    -43,     70 },
      {     89,     29,    -79,    -55,     61,     76,    -36,    -88,      8,     90,     22,    -83,    -49,     67,     71,    -43,    -85,     15,     90,     14,    -86,    -42,     72,     66,    -50,    -82,     23,     90,      7,    -88,    -35,     76,     60,    -56,    -79,     30,     89,     -1,    -89,    -28,     80,     55,    -62,    -75,     37,     87,     -9,    -90,    -21,     83,     48,    -67,    -70,     44,     85,    -16,    -90,    -13,     86,     41,    -72,    -65,     51,     82,    -24,    -90,     -6,     88,     34,    -77,    -60,     57,     78,    -31,    -89,      2,     89,     27,    -81,    -54,     63,     74,    -38,    -87,     10,     90,     20,    -84,    -47,     68,     70,    -45,    -85,     18,     90,     12,    -86,    -40,     73,     64,    -52,    -81,     25,     90,      4,    -88,    -33,     77,     59,    -58,    -78,     32,     89,     -3,    -90,    -26,     81,     53,    -64,    -74,     39,     87,    -11,    -90,    -19,     84,     46,    -69 },
      {     89,     25,    -82,    -48,     69,     67,    -50,    -81,     27,     89,     -2,    -90,    -23,     83,     46,    -70,    -66,     52,     81,    -29,    -89,      4,     90,     21,    -84,    -44,     72,     64,    -54,    -79,     31,     88,     -7,    -90,    -19,     85,     42,    -73,    -63,     55,     78,    -33,    -88,      9,     90,     16,    -86,    -40,     74,     61,    -57,    -77,     35,     87,    -11,    -90,    -14,     86,     38,    -76,    -60,     59,     76,    -37,    -87,     13,     90,     12,    -87,    -36,     77,     58,    -60,    -75,     39,     86,    -15,    -90,    -10,     88,     34,    -78,    -56,     62,     74,    -41,    -85,     18,     90,      8,    -88,    -32,     79,     55,    -64,    -72,     43,     85,    -20,    -90,     -6,     89,     30,    -80,    -53,     65,     71,    -45,    -84,     22,     90,      3,    -89,    -28,     81,     51,    -67,    -70,     47,     83,    -24,    -90,     -1,     89,     26,    -82,    -49,     68 },
      {     90,     21,    -85,    -40,     76,     58,    -62,    -72,     45,     83,    -26,    -89,      6,     90,     15,    -87,    -35,     78,     54,    -66,    -69,     50,     81,    -31,    -88,     11,     90,     10,    -88,    -30,     81,     49,    -70,    -65,     55,     78,    -36,    -86,     16,     90,      4,    -89,    -25,     83,     44,    -73,    -61,     59,     75,    -41,    -85,     22,     90,     -1,    -90,    -20,     85,     39,    -76,    -57,     63,     72,    -46,    -82,     27,     89,     -7,    -90,    -14,     87,     34,    -79,    -53,     67,     68,    -51,    -80,     32,     88,    -12,    -90,     -9,     88,     29,    -81,    -48,     70,     64,    -55,    -77,     37,     86,    -18,    -90,     -3,     89,     24,    -84,    -43,     74,     60,    -60,    -74,     42,     84,    -23,    -89,      2,     90,     19,    -86,    -38,     77,     56,    -64,    -71,     47,     82,    -28,    -89,      8,     90,     13,    -87,    -33,     79,     52,    -67 },
      {     90,     16,    -87,    -32,     81,     47,    -72,    -60,     61,     72,    -48,    -81,     33,     87,    -18,    -90,      1,     90,     15,    -87,    -31,     81,     46,    -73,    -60,     62,     71,    -49,    -80,     34,     86,    -19,    -90,      2,     90,     14,    -88,    -30,     82,     45,    -74,    -59,     63,     70,    -50,    -79,     35,     86,    -20,    -90,      3,     90,     13,    -88,    -29,     82,     44,    -74,    -58,     64,     70,    -51,    -79,     36,     86,    -21,    -89,      4,     90,     12,    -88,    -28,     83,     43,    -75,    -57,     64,     69,    -52,    -78,     37,     85,    -22,    -89,      6,     90,     11,    -88,    -27,     83,     42,    -76,    -56,     65,     68,    -53,    -78,     38,     85,    -23,    -89,      7,     90,     10,    -89,    -26,     84,     41,    -76,    -55,     66,     67,    -54,    -77,     39,     85,    -24,    -89,      8,     90,      9,    -89,    -25,     84,     40,    -77,    -55,     67 },
      {     90,     12,    -89,    -24,     85,     35,    -81,    -46,     74,     56,    -67,    -65,     58,     73,    -48,    -79,     37,     85,    -26,    -88,     14,     90,     -2,    -90,    -10,     89,     22,    -86,    -33,     81,     44,    -76,    -55,     68,     64,    -60,    -72,     50,     78,    -39,    -84,     28,     88,    -16,    -90,      4,     90,      8,    -89,    -20,     87,     31,    -82,    -42,     77,     53,    -70,    -62,     61,     70,    -52,    -77,     41,     83,    -30,    -87,     19,     89,     -7,    -90,     -6,     90,     18,    -87,    -29,     83,     40,    -78,    -51,     71,     60,    -63,    -69,     54,     76,    -43,    -82,     32,     86,    -21,    -89,      9,     90,      3,    -90,    -15,     88,     27,    -84,    -38,     79,     49,    -72,    -59,     64,     67,    -55,    -75,     45,     81,    -34,    -86,     23,     89,    -11,    -90,     -1,     90,     13,    -88,    -25,     85,     36,    -80,    -47,     74,     57,    -66 },
      {     90,      8,    -90,    -15,     88,     23,    -86,    -30,     84,     37,    -81,    -44,     77,     51,    -72,    -57,     67,     63,    -62,    -68,     56,     73,    -50,    -77,     43,     81,    -36,    -84,     29,     87,    -22,    -89,     14,     90,     -7,    -90,     -1,     90,      9,    -89,    -16,     88,     24,    -86,    -31,     83,     38,    -80,    -45,     76,     52,    -72,    -58,     67,     64,    -61,    -69,     55,     74,    -49,    -78,     42,     81,    -35,    -85,     28,     87,    -21,    -89,     13,     90,     -6,    -90,     -2,     90,     10,    -89,    -18,     88,     25,    -86,    -32,     83,     39,    -79,    -46,     76,     53,    -71,    -59,     66,     64,    -60,    -70,     55,     74,    -48,    -78,     41,     82,    -34,    -85,     27,     87,    -20,    -89,     12,     90,     -4,    -90,     -3,     90,     11,    -89,    -19,     88,     26,    -85,    -33,     82,     40,    -79,    -47,     75,     54,    -70,    -60,     65 },
      {     90,      3,    -90,     -7,     90,     10,    -90,    -13,     89,     16,    -89,    -20,     88,     23,    -87,    -26,     86,     29,    -85,    -32,     84,     35,    -82,    -38,     81,     41,    -79,    -44,     78,     47,    -76,    -50,     74,     53,    -72,    -55,     70,     58,    -68,    -60,     66,     63,    -64,    -65,     61,     67,    -59,    -70,     56,     72,    -54,    -74,     51,     76,    -48,    -77,     45,     79,    -42,    -81,     39,     82,    -36,    -83,     33,     85,    -30,    -86,     27,     87,    -24,    -88,     21,     88,    -18,    -89,     14,     89,    -11,    -90,      8,     90,     -4,    -90,      1,     90,      2,    -90,     -6,     90,      9,    -90,    -12,     89,     15,    -89,    -19,     88,     22,    -87,    -25,     86,     28,    -85,    -31,     84,     34,    -83,    -37,     81,     40,    -80,    -43,     78,     46,    -77,    -49,     75,     52,    -73,    -55,     71,     57,    -69,    -60,     67,     62,    -64 },
      {     90,     -1,    -90,      2,     90,     -3,    -90,      4,     90,     -6,    -90,      7,     90,     -8,    -90,      9,     90,    -10,    -90,     11,     90,    -12,    -89,     13,     89,    -14,    -89,     15,     89,    -16,    -89,     18,     89,    -19,    -88,     20,     88,    -21,    -88,     22,     88,    -23,    -87,     24,     87,    -25,    -87,     26,     86,    -27,    -86,     28,     86,    -29,    -85,     30,     85,    -31,    -85,     32,     84,    -33,    -84,     34,     83,    -35,    -83,     36,     82,    -37,    -82,     38,     81,    -39,    -81,     40,     81,    -41,    -80,     42,     79,    -43,    -79,     44,     78,    -45,    -78,     46,     77,    -47,    -77,     48,     76,    -49,    -76,     50,     75,    -51,    -74,     52,     74,    -53,    -73,     54,     72,    -55,    -72,     55,     71,    -56,    -70,     57,     70,    -58,    -69,     59,     68,    -60,    -67,     60,     67,    -61,    -66,     62,     65,    -63,    -64,     64 },
      {     90,     -6,    -90,     11,     89,    -16,    -88,     22,     87,    -27,    -85,     32,     83,    -37,    -81,     42,     78,    -47,    -76,     52,     72,    -56,    -69,     60,     65,    -64,    -61,     68,     57,    -72,    -53,     75,     48,    -78,    -43,     81,     38,    -83,    -33,     85,     28,    -87,    -23,     88,     18,    -89,    -12,     90,      7,    -90,     -1,     90,     -4,    -90,     10,     89,    -15,    -89,     21,     87,    -26,    -86,     31,     84,    -36,    -81,     41,     79,    -46,    -76,     51,     73,    -55,    -70,     60,     66,    -64,    -62,     67,     58,    -71,    -54,     74,     49,    -77,    -44,     80,     39,    -82,    -34,     85,     29,    -86,    -24,     88,     19,    -89,    -13,     90,      8,    -90,     -2,     90,     -3,    -90,      9,     90,    -14,    -89,     20,     88,    -25,    -86,     30,     84,    -35,    -82,     40,     79,    -45,    -77,     50,     74,    -55,    -70,     59,     67,    -63 },
      {     90,    -10,    -89,     20,     87,    -29,    -84,     38,     79,    -47,    -74,     55,     68,    -63,    -61,     70,     54,    -76,    -45,     81,     36,    -85,    -27,     88,     18,    -89,     -8,     90,     -2,    -90,     12,     89,    -22,    -86,     31,     83,    -40,    -78,     49,     73,    -57,    -67,     64,     60,    -71,    -52,     77,     43,    -81,    -34,     85,     25,    -88,    -15,     90,      6,    -90,      4,     90,    -14,    -88,     24,     86,    -33,    -82,     42,     77,    -51,    -72,     59,     65,    -66,    -58,     72,     50,    -78,    -41,     82,     32,    -86,    -23,     89,     13,    -90,     -3,     90,     -7,    -90,     16,     88,    -26,    -85,     35,     81,    -44,    -76,     53,     70,    -60,    -64,     67,     56,    -74,    -48,     79,     39,    -83,    -30,     87,     21,    -89,    -11,     90,      1,    -90,      9,     89,    -19,    -87,     28,     84,    -37,    -80,     46,     75,    -55,    -69,     62 },
      {     90,    -14,    -88,     28,     83,    -41,    -77,     54,     68,    -64,    -58,     74,     46,    -81,    -33,     86,     20,    -89,     -6,     90,     -9,    -89,     23,     85,    -36,    -79,     49,     72,    -60,    -62,     70,     51,    -78,    -38,     85,     25,    -89,    -11,     90,     -3,    -90,     18,     87,    -31,    -82,     44,     75,    -56,    -66,     67,     55,    -76,    -43,     82,     30,    -87,    -16,     90,      2,    -90,     12,     88,    -26,    -84,     39,     78,    -52,    -70,     63,     60,    -72,    -48,     80,     35,    -86,    -22,     89,      8,    -90,      7,     89,    -21,    -86,     34,     81,    -47,    -73,     59,     64,    -69,    -53,     77,     40,    -84,    -27,     88,     13,    -90,      1,     90,    -15,    -88,     29,     83,    -42,    -76,     55,     67,    -65,    -57,     74,     45,    -81,    -32,     87,     19,    -90,     -4,     90,    -10,    -89,     24,     85,    -37,    -79,     50,     71,    -61 },
      {     90,    -19,    -86,     36,     78,    -53,    -67,     67,     54,    -78,    -37,     86,     20,    -90,     -1,     90,    -18,    -86,     35,     79,    -52,    -68,     66,     55,    -77,    -38,     85,     21,    -90,     -2,     90,    -16,    -87,     34,     79,    -51,    -69,     65,     55,    -77,    -39,     85,     22,    -89,     -3,     90,    -15,    -87,     33,     80,    -50,    -70,     64,     56,    -76,    -40,     85,     23,    -89,     -4,     90,    -14,    -87,     32,     81,    -49,    -70,     64,     57,    -76,    -41,     84,     24,    -89,     -6,     90,    -13,    -88,     31,     81,    -48,    -71,     63,     58,    -75,    -42,     84,     25,    -89,     -7,     90,    -12,    -88,     30,     81,    -47,    -72,     62,     59,    -74,    -43,     83,     26,    -89,     -8,     90,    -11,    -88,     29,     82,    -46,    -72,     61,     60,    -74,    -44,     83,     27,    -89,     -9,     90,    -10,    -88,     28,     82,    -45,    -73,     60 },
      {     90,    -23,    -84,     44,     72,    -63,    -56,     77,     36,    -87,    -14,     90,     -9,    -88,     31,     80,    -52,    -67,     69,     49,    -81,    -28,     89,      6,    -90,     18,     86,    -39,    -76,     59,     60,    -74,    -41,     85,     20,    -90,      3,     89,    -26,    -82,     47,     70,    -65,    -54,     79,     33,    -88,    -11,     90,    -12,    -87,     34,     78,    -55,    -64,     71,     46,    -83,    -25,     89,      2,    -90,     21,     85,    -42,    -74,     61,     58,    -76,    -38,     86,     16,    -90,      7,     89,    -29,    -81,     50,     68,    -67,    -51,     81,     30,    -88,     -8,     90,    -15,    -86,     37,     77,    -57,    -62,     73,     43,    -84,    -22,     90,     -1,    -89,     24,     83,    -45,    -72,     64,     55,    -78,    -35,     87,     13,    -90,     10,     88,    -32,    -79,     53,     66,    -70,    -48,     82,     27,    -89,     -4,     90,    -19,    -85,     40,     75,    -60 },
      {     89,    -27,    -81,     52,     65,    -72,    -43,     85,     18,    -90,     10,     87,    -36,    -76,     60,     58,    -77,    -34,     88,      8,    -90,     20,     84,    -45,    -70,     67,     50,    -82,    -25,     90,     -2,    -89,     29,     80,    -54,    -64,     73,     41,    -86,    -15,     90,    -12,    -87,     38,     75,    -61,    -56,     78,     32,    -88,     -6,     90,    -22,    -83,     47,     69,    -68,    -48,     83,     23,    -90,      4,     89,    -31,    -79,     55,     62,    -74,    -39,     86,     13,    -90,     14,     86,    -40,    -74,     63,     55,    -79,    -30,     89,      3,    -90,     24,     82,    -49,    -67,     70,     46,    -84,    -21,     90,     -7,    -88,     33,     78,    -57,    -60,     76,     37,    -87,    -11,     90,    -16,    -85,     42,     72,    -64,    -53,     81,     28,    -89,     -1,     89,    -26,    -81,     51,     66,    -71,    -44,     85,     19,    -90,      9,     88,    -35,    -77,     59 },
      {     89,    -31,    -78,     59,     57,    -79,    -29,     89,     -2,    -89,     33,     77,    -60,    -55,     80,     27,    -90,      4,     88,    -35,    -76,     62,     54,    -81,    -25,     90,     -7,    -88,     37,     74,    -64,    -52,     82,     23,    -90,      9,     87,    -39,    -73,     65,     50,    -83,    -21,     90,    -11,    -86,     41,     72,    -67,    -48,     84,     19,    -90,     13,     86,    -43,    -70,     68,     46,    -85,    -16,     90,    -15,    -85,     45,     69,    -70,    -44,     85,     14,    -90,     18,     84,    -47,    -67,     71,     42,    -86,    -12,     90,    -20,    -83,     49,     66,    -72,    -40,     87,     10,    -90,     22,     82,    -51,    -64,     74,     38,    -87,     -8,     90,    -24,    -81,     53,     63,    -75,    -36,     88,      6,    -90,     26,     81,    -55,    -61,     76,     34,    -88,     -3,     89,    -28,    -79,     56,     60,    -77,    -32,     89,      1,    -89,     30,     78,    -58 },
      {     89,    -35,    -74,     65,     48,    -85,    -14,     90,    -22,    -81,     55,     60,    -78,    -28,     90,     -8,    -87,     42,     70,    -70,    -41,     87,      7,    -90,     29,     78,    -60,    -54,     82,     21,    -90,     15,     84,    -49,    -64,     75,     34,    -89,      1,     88,    -36,    -74,     66,     47,    -85,    -13,     90,    -23,    -81,     55,     59,    -79,    -27,     90,     -9,    -86,     43,     69,    -71,    -40,     87,      6,    -89,     30,     77,    -61,    -53,     82,     20,    -90,     16,     84,    -50,    -64,     76,     33,    -89,      2,     88,    -37,    -73,     67,     46,    -85,    -12,     90,    -24,    -81,     56,     58,    -79,    -26,     90,    -10,    -86,     44,     68,    -72,    -39,     88,      4,    -89,     31,     77,    -62,    -52,     83,     19,    -90,     18,     83,    -51,    -63,     76,     32,    -89,      3,     88,    -38,    -72,     67,     45,    -86,    -11,     90,    -25,    -80,     57 },
      {     88,    -39,    -70,     71,     38,    -88,      1,     88,    -40,    -70,     72,     37,    -89,      2,     88,    -41,    -69,     72,     36,    -89,      3,     87,    -42,    -68,     73,     35,    -89,      4,     87,    -43,    -67,     74,     34,    -89,      6,     87,    -44,    -67,     74,     33,    -89,      7,     86,    -45,    -66,     75,     32,    -89,      8,     86,    -46,    -65,     76,     31,    -90,      9,     86,    -47,    -64,     76,     30,    -90,     10,     85,    -48,    -64,     77,     29,    -90,     11,     85,    -49,    -63,     77,     28,    -90,     12,     85,    -50,    -62,     78,     27,    -90,     13,     84,    -51,    -61,     78,     26,    -90,     14,     84,    -52,    -60,     79,     25,    -90,     15,     83,    -53,    -60,     79,     24,    -90,     16,     83,    -54,    -59,     80,     23,    -90,     18,     82,    -55,    -58,     81,     22,    -90,     19,     82,    -55,    -57,     81,     21,    -90,     20,     81,    -56 },
      {     88,    -43,    -66,     76,     28,    -90,     16,     82,    -57,    -54,     84,     12,    -90,     32,     74,    -69,    -39,     89,     -4,    -86,     47,     63,    -78,    -24,     90,    -21,    -80,     60,     50,    -85,     -8,     89,    -36,    -71,     72,     35,    -89,      9,     85,    -51,    -60,     81,     20,    -90,     25,     78,    -64,    -46,     87,      3,    -88,     40,     68,    -74,    -31,     90,    -13,    -83,     55,     56,    -82,    -15,     90,    -29,    -76,     67,     42,    -88,      1,     87,    -44,    -65,     77,     27,    -90,     18,     81,    -58,    -53,     84,     11,    -90,     33,     73,    -70,    -38,     89,     -6,    -86,     48,     62,    -79,    -23,     90,    -22,    -79,     61,     49,    -86,     -7,     89,    -37,    -70,     72,     34,    -89,     10,     85,    -52,    -59,     81,     19,    -90,     26,     77,    -64,    -45,     87,      2,    -88,     41,     67,    -75,    -30,     90,    -14,    -83,     55 },
      {     87,    -47,    -61,     81,     18,    -90,     31,     73,    -71,    -34,     90,    -14,    -82,     59,     50,    -86,     -3,     88,    -44,    -64,     79,     21,    -90,     28,     75,    -69,    -37,     89,    -11,    -83,     56,     53,    -85,     -7,     89,    -41,    -66,     77,     24,    -90,     25,     77,    -67,    -40,     89,     -8,    -85,     54,     55,    -84,    -10,     89,    -38,    -68,     76,     27,    -90,     22,     78,    -64,    -43,     88,     -4,    -86,     51,     58,    -82,    -13,     90,    -35,    -70,     74,     30,    -90,     19,     80,    -62,    -46,     87,     -1,    -87,     48,     60,    -81,    -16,     90,    -32,    -72,     72,     33,    -90,     15,     81,    -60,    -49,     86,      2,    -88,     45,     63,    -79,    -20,     90,    -29,    -74,     70,     36,    -89,     12,     83,    -57,    -52,     85,      6,    -88,     42,     65,    -78,    -23,     90,    -26,    -76,     67,     39,    -89,      9,     84,    -55 },
      {     86,    -51,    -56,     84,      7,    -88,     45,     61,    -81,    -13,     89,    -39,    -66,     78,     20,    -90,     33,     70,    -75,    -26,     90,    -27,    -74,     71,     32,    -90,     21,     78,    -67,    -38,     89,    -14,    -81,     62,     44,    -88,      8,     84,    -57,    -50,     87,     -1,    -86,     52,     55,    -85,     -6,     88,    -46,    -60,     82,     12,    -89,     40,     65,    -79,    -19,     90,    -34,    -70,     76,     25,    -90,     28,     74,    -72,    -31,     90,    -22,    -77,     67,     37,    -90,     15,     81,    -63,    -43,     89,     -9,    -83,     58,     49,    -87,      2,     86,    -53,    -55,     85,      4,    -88,     47,     60,    -82,    -11,     89,    -41,    -64,     79,     18,    -90,     35,     69,    -76,    -24,     90,    -29,    -73,     72,     30,    -90,     23,     77,    -68,    -36,     90,    -16,    -80,     64,     42,    -89,     10,     83,    -59,    -48,     87,     -3,    -85,     54 },
      {     86,    -55,    -51,     87,     -4,    -84,     58,     47,    -88,      9,     82,    -61,    -43,     89,    -13,    -81,     64,     39,    -90,     18,     78,    -67,    -35,     90,    -22,    -76,     70,     31,    -90,     26,     74,    -73,    -27,     90,    -30,    -71,     76,     23,    -90,     34,     68,    -78,    -19,     90,    -38,    -65,     80,     14,    -89,     42,     62,    -82,    -10,     88,    -46,    -59,     84,      6,    -87,     50,     55,    -85,     -1,     86,    -54,    -52,     87,     -3,    -85,     57,     48,    -88,      8,     83,    -60,    -44,     89,    -12,    -81,     64,     40,    -89,     16,     79,    -67,    -36,     90,    -21,    -77,     70,     32,    -90,     25,     74,    -72,    -28,     90,    -29,    -72,     75,     24,    -90,     33,     69,    -77,    -20,     90,    -37,    -66,     79,     15,    -89,     41,     63,    -81,    -11,     89,    -45,    -60,     83,      7,    -88,     49,     56,    -85,     -2,     86,    -53 },
      {     85,    -58,    -45,     89,    -15,    -78,     69,     31,    -90,     30,     70,    -78,    -16,     89,    -44,    -59,     85,      1,    -85,     57,     46,    -89,     14,     79,    -68,    -32,     90,    -29,    -70,     77,     18,    -89,     43,     60,    -84,     -2,     86,    -56,    -47,     89,    -13,    -79,     67,     33,    -90,     28,     71,    -77,    -19,     89,    -42,    -60,     84,      3,    -86,     55,     48,    -88,     12,     80,    -67,    -34,     90,    -27,    -72,     76,     20,    -90,     41,     61,    -83,     -4,     86,    -55,    -49,     88,    -11,    -81,     66,     35,    -90,     26,     72,    -76,    -21,     90,    -40,    -62,     83,      6,    -87,     54,     50,    -88,     10,     81,    -65,    -36,     90,    -25,    -73,     75,     22,    -90,     39,     63,    -82,     -7,     87,    -53,    -51,     88,     -9,    -81,     64,     37,    -90,     24,     74,    -74,    -23,     90,    -38,    -64,     82,      8,    -87,     52 },
      {     84,    -61,    -39,     90,    -26,    -71,     78,     14,    -88,     50,     52,    -88,     12,     79,    -70,    -28,     90,    -37,    -63,     83,      2,    -85,     60,     41,    -90,     24,     72,    -77,    -16,     89,    -48,    -54,     87,    -10,    -80,     68,     30,    -90,     35,     64,    -82,     -4,     86,    -58,    -43,     90,    -22,    -74,     76,     19,    -89,     46,     55,    -87,      8,     81,    -67,    -32,     90,    -33,    -66,     81,      7,    -86,     56,     45,    -89,     20,     75,    -74,    -21,     89,    -44,    -57,     86,     -6,    -82,     65,     34,    -90,     31,     67,    -81,     -9,     87,    -55,    -47,     89,    -18,    -76,     73,     23,    -90,     42,     59,    -85,      3,     83,    -64,    -36,     90,    -29,    -69,     79,     11,    -88,     53,     49,    -89,     15,     77,    -72,    -25,     90,    -40,    -60,     85,     -1,    -84,     62,     38,    -90,     27,     70,    -78,    -13,     88,    -51 },
      {     83,    -64,    -33,     90,    -36,    -62,     85,     -3,    -82,     67,     30,    -90,     39,     60,    -86,      7,     81,    -69,    -27,     90,    -42,    -57,     87,    -10,    -79,     71,     24,    -90,     45,     55,    -88,     13,     77,    -73,    -21,     89,    -48,    -52,     88,    -16,    -76,     75,     18,    -89,     51,     49,    -89,     20,     74,    -77,    -14,     88,    -54,    -46,     89,    -23,    -72,     78,     11,    -87,     56,     43,    -90,     26,     70,    -80,     -8,     86,    -59,    -40,     90,    -29,    -67,     81,      4,    -85,     61,     37,    -90,     32,     65,    -83,     -1,     84,    -64,    -34,     90,    -35,    -63,     84,     -2,    -82,     66,     31,    -90,     38,     60,    -85,      6,     81,    -68,    -28,     90,    -41,    -58,     86,     -9,    -79,     70,     25,    -90,     44,     55,    -87,     12,     78,    -72,    -22,     89,    -47,    -53,     88,    -15,    -76,     74,     19,    -89,     50 },
      {     82,    -67,    -27,     90,    -46,    -52,     89,    -21,    -72,     79,      7,    -85,     63,     33,    -90,     40,     57,    -87,     14,     76,    -76,    -13,     87,    -58,    -39,     90,    -34,    -62,     85,     -8,    -79,     72,     20,    -89,     53,     45,    -90,     28,     67,    -83,      1,     82,    -68,    -26,     90,    -47,    -51,     89,    -22,    -71,     80,      6,    -85,     64,     32,    -90,     41,     56,    -88,     15,     75,    -77,    -12,     87,    -59,    -38,     90,    -35,    -61,     86,     -9,    -78,     73,     19,    -88,     54,     44,    -90,     29,     66,    -83,      2,     81,    -69,    -25,     89,    -48,    -50,     89,    -23,    -70,     81,      4,    -84,     64,     31,    -90,     42,     55,    -88,     16,     74,    -77,    -11,     86,    -60,    -37,     90,    -36,    -60,     86,    -10,    -78,     74,     18,    -88,     55,     43,    -90,     30,     65,    -84,      3,     81,    -70,    -24,     89,    -49 },
      {     81,    -70,    -21,     88,    -55,    -40,     90,    -37,    -58,     88,    -18,    -72,     80,      3,    -83,     68,     24,    -89,     53,     43,    -90,     34,     60,    -87,     14,     74,    -78,     -7,     84,    -66,    -27,     89,    -50,    -46,     90,    -31,    -63,     86,    -11,    -76,     77,     10,    -85,     64,     30,    -90,     47,     49,    -90,     28,     65,    -85,      8,     78,    -75,    -13,     86,    -61,    -33,     90,    -44,    -52,     89,    -25,    -67,     83,     -4,    -79,     73,     16,    -87,     59,     36,    -90,     41,     55,    -89,     22,     70,    -82,      1,     81,    -71,    -20,     88,    -56,    -39,     90,    -38,    -57,     88,    -19,    -72,     81,      2,    -82,     69,     23,    -89,     54,     42,    -90,     35,     60,    -87,     15,     74,    -79,     -6,     84,    -67,    -26,     89,    -51,    -45,     90,    -32,    -62,     86,    -12,    -76,     77,      9,    -85,     64,     29,    -90,     48 },
      {     81,    -73,    -14,     86,    -64,    -28,     89,    -53,    -41,     90,    -40,    -54,     89,    -27,    -64,     86,    -13,    -74,     80,      1,    -81,     72,     15,    -86,     63,     29,    -89,     52,     42,    -90,     39,     55,    -89,     26,     65,    -85,     12,     74,    -79,     -2,     81,    -72,    -16,     87,    -62,    -30,     90,    -51,    -43,     90,    -38,    -55,     89,    -25,    -66,     85,    -11,    -75,     79,      3,    -82,     71,     18,    -87,     61,     31,    -90,     50,     44,    -90,     37,     56,    -89,     24,     67,    -85,     10,     76,    -78,     -4,     82,    -70,    -19,     87,    -60,    -32,     90,    -49,    -45,     90,    -36,    -57,     88,    -23,    -67,     84,     -9,    -76,     78,      6,    -83,     70,     20,    -88,     60,     33,    -90,     48,     46,    -90,     35,     58,    -88,     22,     68,    -84,      8,     77,    -77,     -7,     83,    -69,    -21,     88,    -59,    -34,     90,    -47 },
      {     79,    -76,     -8,     83,    -71,    -15,     86,    -66,    -23,     88,    -60,    -30,     89,    -55,    -37,     90,    -48,    -44,     90,    -41,    -51,     90,    -34,    -57,     89,    -27,    -63,     87,    -20,    -68,     85,    -12,    -73,     81,     -4,    -77,     78,      3,    -81,     74,     11,    -84,     69,     19,    -87,     64,     26,    -89,     58,     33,    -90,     52,     40,    -90,     45,     47,    -90,     38,     54,    -89,     31,     60,    -88,     24,     65,    -86,     16,     70,    -83,      9,     75,    -80,      1,     79,    -76,     -7,     82,    -72,    -14,     85,    -67,    -22,     88,    -61,    -29,     89,    -55,    -36,     90,    -49,    -43,     90,    -42,    -50,     90,    -35,    -56,     89,    -28,    -62,     87,    -21,    -67,     85,    -13,    -72,     82,     -6,    -77,     78,      2,    -81,     74,     10,    -84,     70,     18,    -86,     64,     25,    -88,     59,     32,    -90,     53,     39,    -90,     46 },
      {     78,    -78,     -1,     79,    -77,     -2,     79,    -77,     -3,     80,    -76,     -4,     81,    -76,     -6,     81,    -75,     -7,     81,    -74,     -8,     82,    -74,     -9,     82,    -73,    -10,     83,    -72,    -11,     83,    -72,    -12,     84,    -71,    -13,     84,    -70,    -14,     85,    -70,    -15,     85,    -69,    -16,     85,    -68,    -18,     86,    -67,    -19,     86,    -67,    -20,     86,    -66,    -21,     87,    -65,    -22,     87,    -64,    -23,     87,    -64,    -24,     88,    -63,    -25,     88,    -62,    -26,     88,    -61,    -27,     88,    -60,    -28,     89,    -60,    -29,     89,    -59,    -30,     89,    -58,    -31,     89,    -57,    -32,     89,    -56,    -33,     89,    -55,    -34,     90,    -55,    -35,     90,    -54,    -36,     90,    -53,    -37,     90,    -52,    -38,     90,    -51,    -39,     90,    -50,    -40,     90,    -49,    -41,     90,    -48,    -42,     90,    -47,    -43,     90,    -46,    -44,     90,    -45 },
      {     77,    -80,      6,     74,    -82,     11,     71,    -85,     16,     67,    -86,     22,     64,    -88,     27,     60,    -89,     32,     55,    -90,     37,     51,    -90,     42,     46,    -90,     47,     41,    -90,     52,     36,    -90,     56,     31,    -89,     60,     26,    -88,     64,     21,    -86,     68,     15,    -84,     72,     10,    -82,     75,      4,    -79,     78,     -1,    -77,     81,     -7,    -74,     83,    -12,    -70,     85,    -18,    -67,     87,    -23,    -63,     88,    -28,    -59,     89,    -33,    -55,     90,    -38,    -50,     90,    -43,    -45,     90,    -48,    -40,     90,    -53,    -35,     89,    -57,    -30,     89,    -61,    -25,     87,    -65,    -20,     86,    -69,    -14,     84,    -72,     -9,     81,    -76,     -3,     79,    -78,      2,     76,    -81,      8,     73,    -83,     13,     70,    -85,     19,     66,    -87,     24,     62,    -88,     29,     58,    -89,     34,     54,    -90,     39,     49,    -90,     44 },
      {     76,    -82,     12,     69,    -86,     24,     60,    -89,     35,     51,    -90,     46,     40,    -90,     56,     29,    -88,     65,     18,    -84,     73,      6,    -79,     79,     -7,    -72,     85,    -19,    -64,     88,    -30,    -55,     90,    -41,    -45,     90,    -52,    -34,     89,    -61,    -23,     86,    -70,    -11,     81,    -77,      1,     76,    -82,     13,     68,    -87,     25,     60,    -89,     36,     50,    -90,     47,     39,    -90,     57,     28,    -88,     66,     16,    -84,     74,      4,    -78,     80,     -8,    -72,     85,    -20,    -64,     88,    -31,    -55,     90,    -42,    -44,     90,    -53,    -33,     89,    -62,    -22,     86,    -70,    -10,     81,    -77,      2,     75,    -83,     14,     67,    -87,     26,     59,    -89,     37,     49,    -90,     48,     38,    -90,     58,     27,    -87,     67,     15,    -83,     74,      3,    -78,     81,     -9,    -71,     85,    -21,    -63,     89,    -32,    -54,     90,    -43 },
      {     75,    -84,     19,     63,    -89,     36,     48,    -90,     53,     31,    -88,     67,     13,    -81,     78,     -6,    -72,     86,    -24,    -59,     90,    -41,    -43,     90,    -57,    -26,     86,    -70,     -8,     79,    -81,     11,     68,    -87,     29,     55,    -90,     46,     38,    -89,     61,     21,    -85,     74,      2,    -76,     83,    -16,    -64,     89,    -34,    -50,     90,    -51,    -33,     88,    -65,    -15,     82,    -77,      3,     73,    -85,     22,     60,    -89,     39,     45,    -90,     55,     28,    -87,     69,     10,    -80,     79,     -9,    -70,     87,    -27,    -56,     90,    -44,    -40,     90,    -60,    -23,     85,    -72,     -4,     77,    -82,     14,     66,    -88,     32,     52,    -90,     49,     35,    -89,     64,     18,    -83,     76,     -1,    -74,     84,    -20,    -62,     89,    -37,    -47,     90,    -54,    -30,     88,    -67,    -12,     81,    -78,      7,     71,    -86,     25,     58,    -90,     42 },
      {     74,    -85,     25,     56,    -90,     48,     34,    -88,     67,     10,    -79,     81,    -15,    -64,     89,    -39,    -43,     90,    -60,    -20,     83,    -77,      6,     70,    -87,     30,     52,    -90,     53,     29,    -87,     71,      4,    -76,     84,    -21,    -60,     90,    -44,    -38,     89,    -64,    -14,     81,    -79,     11,     67,    -88,     35,     47,    -90,     57,     24,    -85,     74,     -1,    -73,     86,    -26,    -55,     90,    -49,    -33,     88,    -68,     -9,     78,    -82,     16,     63,    -89,     40,     42,    -90,     61,     19,    -83,     77,     -7,    -70,     87,    -31,    -51,     90,    -54,    -28,     86,    -72,     -3,     76,    -84,     22,     59,    -90,     45,     37,    -89,     65,     13,    -81,     80,    -12,    -66,     89,    -36,    -46,     90,    -58,    -23,     85,    -75,      2,     72,    -86,     27,     55,    -90,     50,     32,    -88,     69,      8,    -78,     82,    -18,    -62,     89,    -41 },
      {     72,    -87,     31,     49,    -90,     59,     20,    -82,     79,    -12,    -64,     89,    -42,    -38,     89,    -67,     -8,     77,    -84,     24,     55,    -90,     53,     27,    -85,     75,     -4,    -70,     88,    -35,    -45,     90,    -62,    -15,     81,    -81,     16,     61,    -90,     46,     34,    -88,     70,      3,    -74,     86,    -28,    -52,     90,    -56,    -23,     84,    -77,      9,     67,    -89,     39,     41,    -89,     65,     11,    -78,     83,    -21,    -58,     90,    -50,    -30,     86,    -73,      1,     72,    -87,     32,     48,    -90,     60,     19,    -82,     79,    -13,    -64,     89,    -43,    -37,     88,    -68,     -7,     76,    -85,     25,     55,    -90,     54,     26,    -85,     76,     -6,    -69,     88,    -36,    -44,     90,    -63,    -14,     80,    -81,     18,     60,    -90,     47,     33,    -87,     71,      2,    -74,     86,    -29,    -51,     90,    -57,    -22,     83,    -78,     10,     66,    -89,     40 },
      {     71,    -88,     37,     41,    -89,     68,      4,    -74,     87,    -33,    -45,     89,    -65,     -9,     76,    -85,     29,     49,    -90,     62,     13,    -78,     84,    -25,    -53,     90,    -59,    -18,     81,    -82,     21,     56,    -90,     55,     22,    -82,     80,    -16,    -60,     90,    -52,    -26,     84,    -78,     12,     63,    -90,     48,     30,    -86,     76,     -8,    -66,     89,    -44,    -34,     87,    -73,      3,     69,    -89,     40,     38,    -88,     70,      1,    -72,     88,    -36,    -42,     89,    -67,     -6,     74,    -86,     32,     46,    -90,     64,     10,    -77,     85,    -28,    -50,     90,    -61,    -14,     79,    -83,     24,     54,    -90,     58,     19,    -81,     81,    -20,    -57,     90,    -55,    -23,     83,    -79,     15,     60,    -90,     51,     27,    -85,     77,    -11,    -64,     90,    -47,    -31,     86,    -75,      7,     67,    -89,     43,     35,    -87,     72,     -2,    -70,     88,    -39 },
      {     70,    -89,     43,     33,    -86,     76,    -11,    -62,     90,    -53,    -23,     82,    -81,     22,     54,    -90,     61,     12,    -77,     86,    -32,    -44,     89,    -69,     -1,     70,    -89,     42,     34,    -86,     76,    -10,    -63,     90,    -52,    -24,     82,    -81,     21,     55,    -90,     60,     13,    -77,     85,    -31,    -45,     89,    -68,     -2,     71,    -88,     41,     35,    -87,     75,     -9,    -64,     90,    -51,    -25,     83,    -81,     20,     55,    -90,     60,     14,    -78,     85,    -30,    -46,     89,    -67,     -3,     72,    -88,     40,     36,    -87,     74,     -8,    -64,     90,    -50,    -26,     83,    -80,     19,     56,    -90,     59,     15,    -78,     85,    -29,    -47,     89,    -67,     -4,     72,    -88,     39,     37,    -87,     74,     -7,    -65,     90,    -49,    -27,     84,    -79,     18,     57,    -90,     58,     16,    -79,     84,    -28,    -48,     90,    -66,     -6,     73,    -88,     38 },
      {     68,    -89,     49,     25,    -82,     82,    -26,    -48,     89,    -69,      1,     67,    -90,     50,     24,    -81,     83,    -27,    -47,     89,    -70,      2,     67,    -90,     51,     23,    -81,     83,    -28,    -46,     89,    -70,      3,     66,    -90,     52,     22,    -81,     84,    -29,    -45,     89,    -71,      4,     65,    -90,     53,     21,    -80,     84,    -30,    -44,     89,    -72,      6,     64,    -90,     54,     20,    -79,     85,    -31,    -43,     88,    -72,      7,     64,    -90,     55,     19,    -79,     85,    -32,    -42,     88,    -73,      8,     63,    -90,     55,     18,    -78,     85,    -33,    -41,     88,    -74,      9,     62,    -90,     56,     16,    -78,     86,    -34,    -40,     88,    -74,     10,     61,    -90,     57,     15,    -77,     86,    -35,    -39,     87,    -75,     11,     60,    -90,     58,     14,    -77,     86,    -36,    -38,     87,    -76,     12,     60,    -90,     59,     13,    -76,     87,    -37 },
      {     67,    -90,     55,     16,    -77,     87,    -40,    -32,     84,    -81,     25,     47,    -89,     72,     -9,    -60,     90,    -61,     -8,     72,    -89,     48,     24,    -81,     85,    -33,    -39,     87,    -77,     18,     54,    -90,     67,     -1,    -66,     90,    -55,    -15,     76,    -87,     41,     31,    -84,     81,    -26,    -46,     89,    -73,     10,     60,    -90,     62,      7,    -71,     89,    -49,    -23,     80,    -85,     34,     38,    -86,     78,    -19,    -53,     90,    -68,      2,     65,    -90,     56,     14,    -76,     88,    -42,    -30,     83,    -82,     27,     45,    -88,     74,    -11,    -59,     90,    -63,     -6,     70,    -89,     50,     22,    -79,     85,    -35,    -37,     86,    -78,     20,     52,    -90,     69,     -3,    -64,     90,    -57,    -13,     75,    -88,     43,     29,    -83,     82,    -28,    -44,     88,    -74,     12,     58,    -90,     64,      4,    -70,     89,    -51,    -21,     79,    -86,     36 },
      {     65,    -90,     60,      8,    -70,     90,    -54,    -15,     75,    -88,     47,     23,    -79,     86,    -40,    -30,     82,    -84,     33,     37,    -85,     81,    -26,    -44,     88,    -77,     19,     51,    -89,     72,    -11,    -57,     90,    -67,      3,     63,    -90,     62,      4,    -68,     90,    -56,    -12,     73,    -89,     50,     20,    -77,     87,    -43,    -27,     81,    -85,     36,     34,    -84,     82,    -29,    -41,     87,    -78,     22,     48,    -89,     74,    -14,    -55,     90,    -70,      7,     60,    -90,     64,      1,    -66,     90,    -59,     -9,     71,    -89,     53,     16,    -76,     88,    -46,    -24,     79,    -86,     39,     31,    -83,     83,    -32,    -38,     86,    -80,     25,     45,    -88,     76,    -18,    -52,     89,    -72,     10,     58,    -90,     67,     -2,    -64,     90,    -61,     -6,     69,    -90,     55,     13,    -74,     89,    -49,    -21,     78,    -87,     42,     28,    -81,     85,    -35 },
      {     64,    -90,     64,     -1,    -63,     90,    -65,      2,     62,    -90,     66,     -3,    -61,     90,    -67,      4,     60,    -90,     67,     -6,    -60,     90,    -68,      7,     59,    -90,     69,     -8,    -58,     90,    -70,      9,     57,    -90,     70,    -10,    -56,     90,    -71,     11,     55,    -90,     72,    -12,    -55,     89,    -72,     13,     54,    -89,     73,    -14,    -53,     89,    -74,     15,     52,    -89,     74,    -16,    -51,     89,    -75,     18,     50,    -89,     76,    -19,    -49,     88,    -76,     20,     48,    -88,     77,    -21,    -47,     88,    -77,     22,     46,    -88,     78,    -23,    -45,     87,    -78,     24,     44,    -87,     79,    -25,    -43,     87,    -79,     26,     42,    -86,     80,    -27,    -41,     86,    -81,     28,     40,    -86,     81,    -29,    -39,     85,    -81,     30,     38,    -85,     82,    -31,    -37,     85,    -82,     32,     36,    -84,     83,    -33,    -35,     84,    -83,     34 },
      {     62,    -90,     69,    -10,    -55,     89,    -75,     20,     46,    -87,     80,    -29,    -37,     84,    -84,     38,     28,    -79,     87,    -47,    -19,     74,    -89,     55,      9,    -68,     90,    -63,      1,     61,    -90,     70,    -11,    -54,     89,    -76,     21,     45,    -87,     81,    -30,    -36,     83,    -85,     39,     27,    -79,     88,    -48,    -18,     74,    -89,     56,      8,    -67,     90,    -64,      2,     60,    -90,     70,    -12,    -53,     89,    -76,     22,     44,    -86,     81,    -31,    -35,     83,    -85,     40,     26,    -78,     88,    -49,    -16,     73,    -90,     57,      7,    -67,     90,    -64,      3,     60,    -90,     71,    -13,    -52,     89,    -77,     23,     43,    -86,     81,    -32,    -34,     82,    -85,     41,     25,    -78,     88,    -50,    -15,     72,    -90,     58,      6,    -66,     90,    -65,      4,     59,    -90,     72,    -14,    -51,     88,    -77,     24,     42,    -86,     82,    -33 },
      {     60,    -90,     73,    -19,    -45,     86,    -82,     36,     28,    -78,     88,    -53,    -10,     67,    -90,     67,     -9,    -54,     89,    -78,     27,     37,    -83,     86,    -44,    -20,     74,    -90,     60,      1,    -61,     90,    -72,     18,     46,    -86,     82,    -35,    -29,     79,    -88,     52,     11,    -68,     90,    -66,      8,     55,    -89,     77,    -26,    -38,     83,    -85,     43,     21,    -74,     90,    -59,     -2,     62,    -90,     72,    -16,    -47,     87,    -81,     34,     30,    -79,     88,    -51,    -12,     69,    -90,     65,     -7,    -55,     89,    -77,     25,     39,    -84,     85,    -42,    -22,     75,    -89,     58,      3,    -63,     90,    -71,     15,     48,    -87,     81,    -33,    -31,     80,    -88,     50,     13,    -70,     90,    -64,      6,     56,    -89,     76,    -24,    -40,     84,    -85,     41,     23,    -76,     89,    -57,     -4,     64,    -90,     70,    -14,    -49,     87,    -81,     32 },
      {     59,    -89,     77,    -27,    -35,     81,    -88,     52,      9,    -65,     90,    -72,     19,     43,    -85,     85,    -44,    -18,     71,    -90,     66,    -10,    -51,     87,    -81,     36,     26,    -76,     89,    -60,      1,     58,    -89,     77,    -28,    -34,     81,    -88,     53,      8,    -64,     90,    -72,     20,     42,    -84,     85,    -45,    -16,     70,    -90,     67,    -11,    -50,     87,    -82,     37,     25,    -76,     90,    -60,      2,     57,    -89,     78,    -29,    -33,     80,    -88,     54,      7,    -64,     90,    -73,     21,     41,    -84,     86,    -46,    -15,     70,    -90,     67,    -12,    -49,     87,    -82,     38,     24,    -75,     90,    -61,      3,     56,    -89,     78,    -30,    -32,     79,    -88,     55,      6,    -63,     90,    -74,     22,     40,    -83,     86,    -47,    -14,     69,    -90,     68,    -13,    -48,     86,    -83,     39,     23,    -74,     90,    -62,      4,     55,    -89,     79,    -31 },
      {     57,    -89,     80,    -35,    -25,     74,    -90,     65,    -11,    -48,     86,    -85,     45,     14,    -67,     90,    -72,     22,     38,    -81,     88,    -55,     -3,     60,    -89,     78,    -32,    -28,     76,    -90,     63,     -8,    -51,     87,    -83,     42,     18,    -70,     90,    -70,     19,     41,    -83,     87,    -52,     -7,     62,    -90,     77,    -29,    -31,     78,    -89,     60,     -4,    -54,     88,    -82,     39,     21,    -72,     90,    -68,     15,     44,    -84,     86,    -49,    -10,     64,    -90,     75,    -26,    -34,     79,    -89,     58,     -1,    -56,     88,    -81,     36,     24,    -74,     90,    -66,     12,     47,    -85,     85,    -46,    -13,     67,    -90,     73,    -23,    -37,     81,    -88,     55,      2,    -59,     89,    -79,     33,     27,    -76,     90,    -64,      9,     50,    -86,     84,    -43,    -16,     69,    -90,     71,    -20,    -40,     82,    -87,     53,      6,    -61,     89,    -77,     30 },
      {     55,    -88,     83,    -43,    -14,     66,    -90,     76,    -30,    -28,     75,    -90,     67,    -16,    -41,     82,    -88,     57,     -2,    -54,     87,    -84,     45,     12,    -64,     90,    -77,     32,     26,    -74,     90,    -69,     19,     39,    -81,     89,    -59,      4,     52,    -86,     85,    -47,    -10,     63,    -89,     78,    -34,    -24,     72,    -90,     70,    -21,    -37,     80,    -89,     60,     -7,    -50,     86,    -85,     49,      8,    -61,     89,    -79,     36,     22,    -71,     90,    -72,     23,     35,    -79,     89,    -62,      9,     48,    -85,     86,    -51,     -6,     60,    -89,     81,    -38,    -20,     70,    -90,     73,    -25,    -33,     78,    -90,     64,    -11,    -46,     84,    -87,     53,      3,    -58,     88,    -81,     40,     18,    -68,     90,    -74,     27,     31,    -77,     90,    -65,     13,     44,    -83,     87,    -55,     -1,     56,    -88,     82,    -42,    -15,     67,    -90,     76,    -29 },
      {     54,    -86,     85,    -51,     -3,     56,    -87,     84,    -48,     -7,     59,    -88,     83,    -45,    -10,     61,    -89,     81,    -42,    -13,     64,    -89,     80,    -39,    -16,     66,    -90,     78,    -36,    -20,     68,    -90,     77,    -33,    -23,     70,    -90,     75,    -30,    -26,     72,    -90,     73,    -27,    -29,     74,    -90,     71,    -24,    -32,     76,    -90,     69,    -21,    -35,     78,    -90,     67,    -18,    -38,     79,    -89,     64,    -14,    -41,     81,    -89,     62,    -11,    -44,     82,    -88,     60,     -8,    -47,     84,    -88,     57,     -4,    -50,     85,    -87,     55,     -1,    -53,     86,    -86,     52,      2,    -55,     87,    -85,     49,      6,    -58,     88,    -83,     46,      9,    -60,     89,    -82,     43,     12,    -63,     89,    -81,     40,     15,    -65,     90,    -79,     37,     19,    -67,     90,    -77,     34,     22,    -70,     90,    -76,     31,     25,    -72,     90,    -74,     28 },
      {     52,    -85,     87,    -58,      8,     45,    -82,     89,    -64,     15,     38,    -78,     90,    -69,     23,     31,    -74,     90,    -74,     30,     24,    -70,     90,    -78,     37,     16,    -64,     89,    -81,     44,      9,    -59,     88,    -85,     51,      1,    -53,     85,    -87,     57,     -7,    -46,     82,    -89,     63,    -14,    -39,     79,    -90,     68,    -22,    -32,     75,    -90,     73,    -29,    -25,     70,    -90,     77,    -36,    -18,     65,    -89,     81,    -43,    -10,     60,    -88,     84,    -50,     -2,     54,    -86,     87,    -56,      6,     47,    -83,     89,    -62,     13,     40,    -79,     90,    -67,     21,     33,    -76,     90,    -72,     28,     26,    -71,     90,    -77,     35,     19,    -66,     89,    -81,     42,     11,    -60,     88,    -84,     49,      3,    -55,     86,    -86,     55,     -4,    -48,     83,    -88,     61,    -12,    -41,     80,    -90,     67,    -20,    -34,     76,    -90,     72,    -27 },
      {     50,    -83,     89,    -64,     19,     33,    -74,     90,    -76,     36,     15,    -62,     88,    -85,     53,     -3,    -47,     82,    -89,     67,    -22,    -30,     72,    -90,     78,    -39,    -12,     60,    -87,     86,    -55,      7,     44,    -81,     90,    -69,     25,     27,    -70,     90,    -79,     42,      9,    -57,     86,    -87,     58,    -10,    -41,     79,    -90,     71,    -28,    -24,     68,    -90,     81,    -45,     -6,     55,    -85,     88,    -60,     13,     38,    -77,     90,    -73,     31,     21,    -66,     89,    -82,     48,      2,    -52,     84,    -88,     63,    -16,    -35,     76,    -90,     75,    -34,    -18,     64,    -89,     84,    -51,      1,     49,    -83,     89,    -65,     20,     32,    -74,     90,    -77,     37,     14,    -61,     88,    -85,     54,     -4,    -46,     81,    -89,     67,    -23,    -29,     72,    -90,     78,    -40,    -11,     59,    -87,     86,    -56,      8,     43,    -80,     90,    -70,     26 },
      {     48,    -81,     90,    -70,     29,     21,    -64,     88,    -85,     55,     -9,    -40,     77,    -90,     76,    -37,    -12,     58,    -86,     88,    -62,     18,     32,    -72,     90,    -80,     45,      3,    -51,     83,    -89,     68,    -26,    -24,     67,    -89,     84,    -53,      6,     43,    -79,     90,    -74,     34,     15,    -60,     87,    -87,     60,    -14,    -35,     74,    -90,     78,    -42,     -7,     54,    -84,     89,    -66,     23,     27,    -69,     89,    -82,     50,     -2,    -46,     81,    -90,     72,    -31,    -19,     63,    -88,     86,    -57,     11,     38,    -76,     90,    -77,     39,     10,    -56,     85,    -88,     64,    -20,    -30,     71,    -90,     81,    -47,     -1,     49,    -82,     90,    -70,     28,     22,    -65,     89,    -85,     55,     -8,    -41,     78,    -90,     75,    -36,    -13,     59,    -86,     87,    -61,     16,     33,    -73,     90,    -79,     44,      4,    -52,     83,    -89,     67,    -25 },
      {     46,    -79,     90,    -76,     39,      8,    -53,     83,    -90,     71,    -32,    -15,     59,    -86,     88,    -66,     25,     23,    -64,     88,    -86,     60,    -18,    -30,     70,    -89,     84,    -55,     10,     37,    -74,     90,    -81,     48,     -2,    -44,     78,    -90,     77,    -41,     -6,     51,    -82,     90,    -72,     34,     13,    -57,     85,    -89,     67,    -27,    -21,     63,    -87,     87,    -62,     20,     28,    -68,     89,    -85,     56,    -12,    -35,     73,    -90,     81,    -50,      4,     42,    -77,     90,    -78,     43,      3,    -49,     81,    -90,     74,    -36,    -11,     55,    -84,     89,    -69,     29,     19,    -61,     87,    -88,     64,    -22,    -26,     67,    -89,     85,    -58,     14,     33,    -72,     90,    -82,     52,     -7,    -40,     76,    -90,     79,    -45,     -1,     47,    -80,     90,    -75,     38,      9,    -54,     83,    -89,     70,    -31,    -16,     60,    -86,     88,    -65,     24 },
      {     44,    -77,     90,    -80,     49,     -6,    -39,     74,    -90,     82,    -54,     11,     34,    -71,     89,    -85,     58,    -16,    -29,     67,    -88,     86,    -62,     22,     24,    -64,     87,    -88,     66,    -27,    -19,     60,    -85,     89,    -70,     32,     13,    -55,     83,    -90,     73,    -37,     -8,     51,    -81,     90,    -76,     42,      2,    -46,     78,    -90,     79,    -47,      3,     41,    -76,     90,    -81,     52,     -9,    -36,     72,    -90,     84,    -56,     14,     31,    -69,     89,    -86,     60,    -20,    -26,     65,    -88,     87,    -64,     25,     21,    -61,     86,    -89,     68,    -30,    -15,     57,    -84,     89,    -72,     35,     10,    -53,     82,    -90,     75,    -40,     -4,     48,    -79,     90,    -78,     45,     -1,    -43,     77,    -90,     81,    -50,      7,     38,    -74,     90,    -83,     55,    -12,    -33,     70,    -89,     85,    -59,     18,     28,    -67,     88,    -87,     63,    -23 },
      {     42,    -75,     90,    -84,     58,    -19,    -25,     63,    -86,     89,    -71,     36,      7,    -48,     78,    -90,     81,    -53,     12,     31,    -67,     88,    -88,     67,    -30,    -13,     54,    -81,     90,    -78,     47,     -6,    -37,     72,    -89,     86,    -62,     24,     20,    -59,     84,    -90,     74,    -41,     -1,     43,    -76,     90,    -83,     57,    -18,    -26,     64,    -86,     89,    -70,     35,      8,    -49,     79,    -90,     81,    -52,     11,     32,    -68,     88,    -87,     66,    -29,    -14,     55,    -82,     90,    -77,     46,     -4,    -38,     72,    -89,     85,    -61,     23,     21,    -60,     85,    -90,     74,    -40,     -2,     44,    -76,     90,    -83,     56,    -16,    -27,     64,    -87,     89,    -70,     34,      9,    -50,     79,    -90,     80,    -51,     10,     33,    -69,     88,    -87,     65,    -28,    -15,     55,    -82,     90,    -77,     45,     -3,    -39,     73,    -89,     85,    -60,     22 },
      {     40,    -72,     89,    -87,     66,    -31,    -10,     49,    -78,     90,    -83,     59,    -22,    -20,     57,    -82,     90,    -79,     51,    -12,    -29,     64,    -86,     89,    -74,     42,     -2,    -38,     71,    -89,     87,    -67,     33,      8,    -47,     77,    -90,     84,    -60,     24,     18,    -55,     81,    -90,     80,    -53,     14,     27,    -63,     85,    -90,     75,    -44,      4,     36,    -70,     88,    -88,     69,    -35,     -6,     45,    -76,     90,    -85,     62,    -26,    -15,     54,    -81,     90,    -81,     55,    -16,    -25,     61,    -85,     90,    -76,     46,     -7,    -34,     68,    -88,     88,    -70,     37,      3,    -43,     74,    -89,     86,    -64,     28,     13,    -52,     79,    -90,     82,    -56,     19,     23,    -60,     84,    -90,     77,    -48,      9,     32,    -67,     87,    -89,     72,    -39,     -1,     41,    -73,     89,    -86,     65,    -30,    -11,     50,    -78,     90,    -83,     58,    -21 },
      {     38,    -70,     88,    -89,     73,    -43,      6,     33,    -66,     86,    -90,     76,    -48,     11,     28,    -62,     84,    -90,     79,    -53,     16,     23,    -58,     82,    -90,     81,    -57,     22,     18,    -54,     79,    -90,     84,    -61,     27,     12,    -49,     77,    -90,     86,    -65,     32,      7,    -44,     74,    -89,     87,    -69,     37,      1,    -39,     70,    -88,     89,    -72,     42,     -4,    -34,     67,    -86,     89,    -76,     47,    -10,    -29,     63,    -85,     90,    -78,     52,    -15,    -24,     59,    -82,     90,    -81,     56,    -21,    -19,     55,    -80,     90,    -83,     60,    -26,    -13,     50,    -77,     90,    -85,     64,    -31,     -8,     45,    -74,     89,    -87,     68,    -36,     -2,     40,    -71,     88,    -88,     72,    -41,      3,     35,    -67,     87,    -89,     75,    -46,      9,     30,    -64,     85,    -90,     78,    -51,     14,     25,    -60,     83,    -90,     81,    -55,     20 },
      {     36,    -67,     86,    -90,     79,    -55,     21,     16,    -51,     77,    -89,     87,    -70,     40,     -4,    -32,     64,    -84,     90,    -81,     58,    -25,    -12,     47,    -74,     89,    -88,     72,    -44,      9,     28,    -60,     82,    -90,     83,    -61,     29,      8,    -43,     72,    -88,     89,    -75,     48,    -13,    -24,     57,    -81,     90,    -85,     64,    -33,     -3,     39,    -69,     87,    -90,     77,    -52,     18,     20,    -54,     78,    -90,     86,    -67,     37,     -1,    -35,     66,    -85,     90,    -79,     55,    -22,    -15,     50,    -76,     89,    -87,     70,    -41,      6,     31,    -63,     84,    -90,     81,    -59,     26,     11,    -46,     74,    -89,     88,    -73,     45,    -10,    -27,     60,    -82,     90,    -83,     62,    -30,     -7,     42,    -71,     88,    -89,     76,    -49,     14,     23,    -56,     80,    -90,     85,    -65,     34,      2,    -38,     68,    -86,     90,    -78,     53,    -19 },
      {     34,    -64,     83,    -90,     84,    -64,     35,     -1,    -33,     63,    -83,     90,    -84,     65,    -36,      2,     32,    -62,     82,    -90,     85,    -66,     37,     -3,    -31,     61,    -82,     90,    -85,     67,    -38,      4,     30,    -60,     81,    -90,     85,    -67,     39,     -6,    -29,     60,    -81,     90,    -86,     68,    -40,      7,     28,    -59,     81,    -90,     86,    -69,     41,     -8,    -27,     58,    -80,     90,    -86,     70,    -42,      9,     26,    -57,     79,    -90,     87,    -70,     43,    -10,    -25,     56,    -79,     90,    -87,     71,    -44,     11,     24,    -55,     78,    -90,     87,    -72,     45,    -12,    -23,     55,    -78,     89,    -88,     72,    -46,     13,     22,    -54,     77,    -89,     88,    -73,     47,    -14,    -21,     53,    -77,     89,    -88,     74,    -48,     15,     20,    -52,     76,    -89,     88,    -74,     49,    -16,    -19,     51,    -76,     89,    -89,     75,    -50,     18 },
      {     32,    -60,     81,    -90,     87,    -73,     49,    -19,    -14,     45,    -70,     86,    -90,     82,    -64,     36,     -4,    -28,     57,    -78,     89,    -88,     76,    -53,     23,     10,    -41,     67,    -85,     90,    -84,     67,    -40,      9,     24,    -54,     76,    -89,     89,    -78,     56,    -27,     -6,     37,    -64,     83,    -90,     86,    -70,     44,    -13,    -20,     50,    -74,     88,    -90,     80,    -60,     31,      1,    -33,     61,    -81,     90,    -87,     72,    -48,     18,     15,    -46,     71,    -86,     90,    -82,     63,    -35,      3,     29,    -58,     79,    -89,     88,    -75,     52,    -22,    -11,     42,    -68,     85,    -90,     84,    -66,     39,     -8,    -25,     55,    -77,     89,    -89,     77,    -55,     26,      7,    -38,     65,    -83,     90,    -85,     69,    -43,     12,     21,    -51,     74,    -88,     90,    -79,     59,    -30,     -2,     34,    -62,     81,    -90,     87,    -72,     47,    -16 },
      {     30,    -57,     77,    -89,     89,    -80,     61,    -35,      6,     25,    -53,     74,    -87,     90,    -82,     65,    -40,     11,     20,    -48,     71,    -86,     90,    -85,     69,    -45,     16,     14,    -43,     67,    -84,     90,    -86,     72,    -50,     22,      9,    -38,     64,    -81,     90,    -88,     76,    -55,     27,      3,    -33,     60,    -79,     89,    -89,     78,    -59,     32,     -2,    -28,     55,    -76,     88,    -90,     81,    -63,     37,     -8,    -23,     51,    -73,     87,    -90,     83,    -67,     42,    -13,    -18,     46,    -70,     85,    -90,     85,    -70,     47,    -19,    -12,     41,    -66,     83,    -90,     87,    -74,     52,    -24,     -7,     36,    -62,     81,    -90,     88,    -77,     56,    -29,     -1,     31,    -58,     78,    -89,     89,    -79,     60,    -34,      4,     26,    -54,     75,    -88,     90,    -82,     64,    -39,     10,     21,    -49,     72,    -86,     90,    -84,     68,    -44,     15 },
      {     28,    -54,     74,    -86,     90,    -85,     72,    -51,     25,      3,    -31,     56,    -76,     87,    -90,     84,    -70,     48,    -22,     -7,     34,    -59,     77,    -88,     90,    -83,     67,    -45,     19,     10,    -37,     61,    -79,     89,    -90,     81,    -65,     42,    -15,    -13,     40,    -64,     81,    -89,     89,    -80,     63,    -39,     12,     16,    -43,     66,    -82,     90,    -89,     78,    -60,     36,     -9,    -20,     46,    -68,     83,    -90,     88,    -77,     58,    -33,      6,     23,    -49,     70,    -85,     90,    -87,     75,    -55,     30,     -2,    -26,     52,    -72,     86,    -90,     86,    -73,     53,    -27,     -1,     29,    -55,     74,    -87,     90,    -85,     71,    -50,     24,      4,    -32,     57,    -76,     88,    -90,     84,    -69,     47,    -21,     -8,     35,    -60,     78,    -88,     90,    -82,     67,    -44,     18,     11,    -38,     62,    -79,     89,    -89,     81,    -64,     41,    -14 },
      {     26,    -50,     70,    -83,     90,    -89,     80,    -64,     43,    -19,     -8,     33,    -56,     74,    -86,     90,    -87,     76,    -59,     36,    -11,    -15,     40,    -62,     78,    -88,     90,    -85,     72,    -53,     29,     -3,    -23,     47,    -67,     82,    -89,     89,    -81,     67,    -46,     22,      4,    -30,     54,    -72,     85,    -90,     88,    -78,     61,    -39,     14,     12,    -37,     60,    -77,     87,    -90,     86,    -74,     55,    -32,      7,     20,    -44,     65,    -81,     89,    -90,     83,    -69,     49,    -25,     -1,     27,    -51,     70,    -84,     90,    -89,     79,    -64,     42,    -18,     -9,     34,    -57,     75,    -86,     90,    -87,     76,    -58,     35,    -10,    -16,     41,    -63,     79,    -88,     90,    -84,     71,    -52,     28,     -2,    -24,     48,    -68,     82,    -90,     89,    -81,     66,    -45,     21,      6,    -31,     55,    -73,     85,    -90,     88,    -77,     60,    -38,     13 },
      {     24,    -46,     65,    -79,     88,    -90,     86,    -76,     60,    -39,     16,      8,    -31,     53,    -70,     83,    -89,     90,    -83,     71,    -54,     32,     -9,    -15,     38,    -59,     75,    -86,     90,    -88,     80,    -66,     47,    -25,      1,     23,    -45,     64,    -79,     88,    -90,     86,    -76,     60,    -40,     18,      7,    -30,     52,    -70,     82,    -89,     90,    -84,     72,    -55,     33,    -10,    -14,     37,    -58,     74,    -85,     90,    -89,     81,    -67,     48,    -26,      2,     22,    -44,     64,    -78,     88,    -90,     87,    -77,     61,    -41,     19,      6,    -29,     51,    -69,     82,    -89,     90,    -84,     72,    -55,     34,    -11,    -13,     36,    -57,     74,    -85,     90,    -89,     81,    -67,     49,    -27,      3,     21,    -43,     63,    -78,     87,    -90,     87,    -77,     62,    -42,     20,      4,    -28,     50,    -68,     81,    -89,     90,    -85,     73,    -56,     35,    -12 },
      {     22,    -42,     60,    -75,     85,    -90,     89,    -84,     73,    -58,     39,    -19,     -3,     25,    -45,     63,    -77,     86,    -90,     89,    -82,     71,    -55,     36,    -15,     -7,     28,    -48,     65,    -78,     87,    -90,     88,    -81,     69,    -53,     33,    -12,    -10,     31,    -51,     67,    -80,     88,    -90,     88,    -79,     67,    -50,     30,     -9,    -13,     34,    -54,     70,    -81,     89,    -90,     87,    -78,     64,    -47,     27,     -6,    -16,     37,    -56,     72,    -83,     89,    -90,     86,    -76,     62,    -44,     24,     -2,    -20,     40,    -59,     74,    -84,     90,    -90,     85,    -74,     60,    -41,     21,      1,    -23,     43,    -61,     76,    -85,     90,    -89,     83,    -72,     57,    -38,     18,      4,    -26,     46,    -64,     77,    -86,     90,    -89,     82,    -70,     55,    -35,     14,      8,    -29,     49,    -66,     79,    -87,     90,    -88,     81,    -68,     52,    -32,     11 },
      {     20,    -38,     55,    -70,     81,    -88,     90,    -89,     83,    -73,     60,    -43,     25,     -6,    -14,     33,    -51,     66,    -78,     86,    -90,     90,    -85,     76,    -64,     48,    -30,     11,      9,    -28,     46,    -62,     75,    -84,     89,    -90,     87,    -79,     67,    -53,     35,    -16,     -3,     23,    -41,     58,    -72,     82,    -88,     90,    -88,     81,    -71,     57,    -40,     22,     -2,    -18,     36,    -54,     68,    -79,     87,    -90,     89,    -84,     74,    -61,     45,    -27,      8,     12,    -31,     49,    -64,     77,    -85,     90,    -90,     86,    -77,     65,    -50,     32,    -13,     -7,     26,    -44,     60,    -74,     83,    -89,     90,    -87,     80,    -69,     55,    -37,     19,      1,    -21,     39,    -56,     70,    -81,     88,    -90,     89,    -82,     72,    -59,     42,    -24,      4,     15,    -34,     52,    -67,     78,    -86,     90,    -89,     85,    -76,     63,    -47,     29,    -10 },
      {     18,    -34,     50,    -64,     75,    -83,     89,    -90,     89,    -84,     76,    -64,     51,    -35,     19,     -1,    -16,     33,    -49,     63,    -74,     83,    -88,     90,    -89,     84,    -76,     65,    -52,     36,    -20,      2,     15,    -32,     48,    -62,     74,    -82,     88,    -90,     89,    -85,     77,    -66,     53,    -37,     21,     -3,    -14,     31,    -47,     61,    -73,     82,    -88,     90,    -89,     85,    -77,     67,    -54,     38,    -22,      4,     13,    -30,     46,    -60,     72,    -81,     88,    -90,     89,    -85,     78,    -67,     55,    -39,     23,     -6,    -12,     29,    -45,     60,    -72,     81,    -87,     90,    -90,     86,    -78,     68,    -55,     40,    -24,      7,     11,    -28,     44,    -59,     71,    -81,     87,    -90,     90,    -86,     79,    -69,     56,    -41,     25,     -8,    -10,     27,    -43,     58,    -70,     80,    -87,     90,    -90,     86,    -79,     70,    -57,     42,    -26,      9 },
      {     15,    -30,     44,    -57,     68,    -77,     84,    -89,     90,    -89,     86,    -80,     72,    -61,     49,    -35,     21,     -6,    -10,     25,    -39,     53,    -64,     74,    -82,     87,    -90,     90,    -88,     82,    -75,     65,    -54,     40,    -26,     11,      4,    -20,     34,    -48,     60,    -71,     79,    -86,     89,    -90,     89,    -85,     78,    -69,     58,    -45,     31,    -16,      1,     14,    -29,     43,    -56,     67,    -77,     84,    -88,     90,    -90,     86,    -81,     72,    -62,     50,    -36,     22,     -7,     -9,     24,    -38,     52,    -64,     74,    -81,     87,    -90,     90,    -88,     83,    -76,     66,    -55,     41,    -27,     12,      3,    -19,     33,    -47,     60,    -70,     79,    -85,     89,    -90,     89,    -85,     78,    -70,     59,    -46,     32,    -18,      2,     13,    -28,     42,    -55,     67,    -76,     83,    -88,     90,    -90,     87,    -81,     73,    -63,     51,    -37,     23,     -8 },
      {     13,    -26,     38,    -50,     60,    -70,     77,    -83,     88,    -90,     90,    -89,     85,    -80,     73,    -64,     55,    -43,     31,    -19,      6,      8,    -21,     33,    -45,     56,    -66,     74,    -81,     86,    -89,     90,    -90,     87,    -82,     76,    -68,     59,    -48,     36,    -24,     11,      2,    -15,     28,    -40,     52,    -62,     71,    -78,     84,    -88,     90,    -90,     88,    -85,     79,    -72,     63,    -53,     41,    -29,     16,     -3,    -10,     23,    -35,     47,    -58,     67,    -76,     82,    -87,     89,    -90,     89,    -86,     81,    -75,     67,    -57,     46,    -34,     22,     -9,     -4,     18,    -30,     42,    -54,     64,    -72,     79,    -85,     89,    -90,     90,    -88,     84,    -78,     70,    -61,     51,    -39,     27,    -14,      1,     12,    -25,     37,    -49,     60,    -69,     77,    -83,     87,    -90,     90,    -89,     86,    -81,     74,    -65,     55,    -44,     32,    -20,      7 },
      {     11,    -22,     32,    -42,     52,    -60,     68,    -75,     81,    -85,     88,    -90,     90,    -89,     87,    -84,     79,    -73,     66,    -58,     49,    -39,     29,    -19,      8,      3,    -14,     25,    -35,     45,    -55,     63,    -70,     77,    -82,     86,    -89,     90,    -90,     89,    -86,     82,    -77,     71,    -64,     55,    -46,     36,    -26,     15,     -4,     -7,     18,    -28,     38,    -48,     57,    -65,     72,    -78,     83,    -87,     89,    -90,     90,    -88,     85,    -81,     76,    -69,     61,    -53,     43,    -33,     23,    -12,      1,     10,    -21,     31,    -41,     51,    -60,     67,    -74,     80,    -85,     88,    -90,     90,    -90,     88,    -84,     79,    -74,     67,    -59,     50,    -40,     30,    -20,      9,      2,    -13,     24,    -34,     44,    -54,     62,    -70,     76,    -81,     86,    -89,     90,    -90,     89,    -87,     83,    -78,     72,    -64,     56,    -47,     37,    -27,     16,     -6 },
      {      9,    -18,     26,    -34,     42,    -50,     57,    -64,     70,    -75,     79,    -83,     86,    -89,     90,    -90,     90,    -89,     87,    -84,     80,    -76,     70,    -64,     58,    -51,     43,    -35,     27,    -19,     10,     -1,     -8,     16,    -25,     33,    -41,     49,    -56,     63,    -69,     74,    -79,     83,    -86,     88,    -90,     90,    -90,     89,    -87,     84,    -81,     76,    -71,     65,    -59,     52,    -44,     36,    -28,     20,    -11,      2,      7,    -15,     24,    -32,     40,    -48,     55,    -62,     68,    -74,     78,    -82,     86,    -88,     90,    -90,     90,    -89,     87,    -85,     81,    -77,     72,    -66,     60,    -53,     45,    -37,     29,    -21,     12,     -3,     -6,     14,    -23,     31,    -39,     47,    -55,     61,    -67,     73,    -78,     82,    -85,     88,    -89,     90,    -90,     89,    -88,     85,    -81,     77,    -72,     67,    -60,     54,    -46,     38,    -30,     22,    -13,      4 },
      {      7,    -13,     20,    -26,     32,    -38,     44,    -50,     55,    -60,     65,    -70,     74,    -77,     81,    -83,     86,    -88,     89,    -90,     90,    -90,     90,    -89,     87,    -85,     83,    -80,     77,    -73,     69,    -64,     60,    -55,     49,    -43,     37,    -31,     25,    -19,     12,     -6,     -1,      8,    -14,     21,    -27,     33,    -39,     45,    -51,     56,    -61,     66,    -70,     74,    -78,     81,    -84,     86,    -88,     89,    -90,     90,    -90,     90,    -89,     87,    -85,     82,    -79,     76,    -72,     68,    -64,     59,    -54,     48,    -42,     36,    -30,     24,    -18,     11,     -4,     -2,      9,    -15,     22,    -28,     34,    -40,     46,    -52,     57,    -62,     67,    -71,     75,    -78,     81,    -84,     86,    -88,     89,    -90,     90,    -90,     89,    -88,     87,    -85,     82,    -79,     76,    -72,     67,    -63,     58,    -53,     47,    -41,     35,    -29,     23,    -16,     10,     -3 },
      {      4,     -9,     13,    -18,     22,    -26,     30,    -34,     38,    -42,     46,    -50,     54,    -57,     60,    -64,     67,    -70,     72,    -75,     77,    -79,     81,    -83,     85,    -86,     88,    -89,     89,    -90,     90,    -90,     90,    -90,     89,    -89,     88,    -87,     85,    -84,     82,    -80,     78,    -76,     73,    -70,     67,    -64,     61,    -58,     55,    -51,     47,    -43,     39,    -35,     31,    -27,     23,    -19,     14,    -10,      6,     -1,     -3,      8,    -12,     16,    -21,     25,    -29,     33,    -37,     41,    -45,     49,    -53,     56,    -60,     63,    -66,     69,    -72,     74,    -77,     79,    -81,     83,    -85,     86,    -87,     88,    -89,     90,    -90,     90,    -90,     90,    -90,     89,    -88,     87,    -86,     84,    -82,     81,    -78,     76,    -74,     71,    -68,     65,    -62,     59,    -55,     52,    -48,     44,    -40,     36,    -32,     28,    -24,     20,    -15,     11,     -7,      2 },
      {      2,     -4,      7,     -9,     11,    -13,     15,    -18,     20,    -22,     24,    -26,     28,    -30,     32,    -34,     36,    -38,     40,    -42,     44,    -46,     48,    -50,     52,    -54,     55,    -57,     59,    -60,     62,    -64,     65,    -67,     68,    -70,     71,    -72,     74,    -75,     76,    -77,     78,    -79,     81,    -81,     82,    -83,     84,    -85,     86,    -86,     87,    -88,     88,    -89,     89,    -89,     90,    -90,     90,    -90,     90,    -90,     90,    -90,     90,    -90,     90,    -89,     89,    -89,     88,    -88,     87,    -87,     86,    -85,     85,    -84,     83,    -82,     81,    -80,     79,    -78,     77,    -76,     74,    -73,     72,    -70,     69,    -67,     66,    -64,     63,    -61,     60,    -58,     56,    -55,     53,    -51,     49,    -47,     45,    -43,     41,    -39,     37,    -35,     33,    -31,     29,    -27,     25,    -23,     21,    -19,     16,    -14,     12,    -10,      8,     -6,      3,     -1 },
    },
    {
      {      1,      2,      3,      4,      6,      7,      8,      9,     10,     11,     12,     13,     14,     15,     16,     18,     19,     20,     21,     22,     23,     24,     25,     26,     27,     28,     29,     30,     31,     32,     33,     34,     35,     36,     37,     38,     39,     40,     41,     42,     43,     44,     45,     46,     47,     48,     49,     50,     51,     52,     53,     54,     55,     55,     56,     57,     58,     59,     60,     60,     61,     62,     63,     64,     64,     65,     66,     67,     67,     68,     69,     70,     70,     71,     72,     72,     73,     74,     74,     75,     76,     76,     77,     77,     78,     78,     79,     79,     80,     81,     81,     81,     82,     82,     83,     83,     84,     84,     85,     85,     85,     86,     86,     86,     87,     87,     87,     88,     88,     88,     88,     89,     89,     89,     89,     89,     89,     90,     90,     90,     90,     90,     90,     90,     90,     90,     90,     90 },
      {      3,      7,     10,     13,     16,     20,     23,     26,     29,     32,     35,     38,     41,     44,     47,     50,     53,     55,     58,     60,     63,     65,     67,     70,     72,     74,     76,     77,     79,     81,     82,     83,     85,     86,     87,     88,     88,     89,     89,     90,     90,     90,     90,     90,     90,     90,     89,     89,     88,     87,     86,     85,     84,     83,     81,     80,     78,     77,     75,     73,     71,     69,     67,     64,     62,     60,     57,     55,     52,     49,     46,     43,     40,     37,     34,     31,     28,     25,     22,     19,     15,     12,      9,      6,      2,     -1,     -4,     -8,    -11,    -14,    -18,    -21,    -24,    -27,    -30,    -33,    -36,    -39,    -42,    -45,    -48,    -51,    -54,    -56,    -59,    -61,    -64,    -66,    -68,    -70,    -72,    -74,    -76,    -78,    -79,    -81,    -82,    -84,    -85,    -86,    -87,    -88,    -89,    -89,    -90,    -90,    -90,    -90 },
      {      6,     11,     16,     22,     27,     32,     37,     42,     47,     52,     56,     60,     64,     68,     72,     75,     78,     81,     83,     85,     87,     88,     89,     90,     90,     90,     90,     89,     89,     87,     86,     84,     81,     79,     76,     73,     70,     66,     62,     58,     54,     49,     44,     39,     34,     29,     24,     19,     13,      8,      2,     -3,     -9,    -14,    -20,    -25,    -30,    -35,    -40,    -45,    -50,    -55,    -59,    -63,    -67,    -70,    -74,    -77,    -79,    -82,    -84,    -86,    -88,    -89,    -90,    -90,    -90,    -90,    -90,    -89,    -88,    -86,    -85,    -82,    -80,    -77,    -74,    -71,    -67,    -64,    -60,    -55,    -51,    -46,    -41,    -36,    -31,    -26,    -21,    -15,    -10,     -4,      1,      7,     12,     18,     23,     28,     33,     38,     43,     48,     53,     57,     61,     65,     69,     72,     76,     78,     81,     83,     85,     87,     88,     89,     90,     90 },
      {      8,     15,     23,     30,     37,     44,     51,     57,     63,     68,     73,     77,     81,     84,     87,     89,     90,     90,     90,     89,     88,     86,     83,     80,     76,     72,     67,     61,     55,     49,     42,     35,     28,     21,     13,      6,     -2,    -10,    -18,    -25,    -32,    -39,    -46,    -53,    -59,    -64,    -70,    -74,    -78,    -82,    -85,    -87,    -89,    -90,    -90,    -90,    -89,    -88,    -85,    -82,    -79,    -75,    -70,    -65,    -60,    -54,    -47,    -40,    -33,    -26,    -19,    -11,     -3,      4,     12,     20,     27,     34,     41,     48,     55,     60,     66,     71,     76,     79,     83,     86,     88,     89,     90,     90,     90,     89,     87,     85,     81,     78,     74,     69,     64,     58,     52,     45,     38,     31,     24,     16,      9,      1,     -7,    -14,    -22,    -29,    -36,    -43,    -50,    -56,    -62,    -67,    -72,    -77,    -81,    -84,    -86,    -88,    -90,    -90 },
      {     10,     20,     29,     38,     47,     55,     63,     70,     76,     81,     85,     88,     89,     90,     90,     89,     86,     83,     78,     73,     67,     60,     52,     43,     34,     25,     15,      6,     -4,    -14,    -24,    -33,    -42,    -51,    -59,    -66,    -72,    -78,    -82,    -86,    -89,    -90,    -90,    -90,    -88,    -85,    -81,    -76,    -70,    -64,    -56,    -48,    -39,    -30,    -21,    -11,     -1,      9,     19,     28,     37,     46,     55,     62,     69,     75,     80,     84,     87,     89,     90,     90,     89,     87,     83,     79,     74,     67,     60,     53,     44,     35,     26,     16,      7,     -3,    -13,    -23,    -32,    -41,    -50,    -58,    -65,    -72,    -77,    -82,    -86,    -88,    -90,    -90,    -90,    -88,    -85,    -81,    -77,    -71,    -64,    -57,    -49,    -40,    -31,    -22,    -12,     -2,      8,     18,     27,     36,     45,     54,     61,     68,     74,     79,     84,     87,     89,     90 },
      {     12,     24,     35,     46,     56,     65,     73,     79,     85,     88,     90,     90,     89,     86,     81,     76,     68,     60,     50,     39,     28,     16,      4,     -8,    -20,    -31,    -42,    -53,    -62,    -70,    -77,    -83,    -87,    -89,    -90,    -90,    -87,    -83,    -78,    -71,    -63,    -54,    -43,    -32,    -21,     -9,      3,     15,     27,     38,     49,     59,     67,     75,     81,     86,     89,     90,     90,     88,     85,     80,     74,     66,     57,     47,     36,     25,     13,      1,    -11,    -23,    -34,    -45,    -55,    -64,    -72,    -79,    -84,    -88,    -90,    -90,    -89,    -86,    -82,    -76,    -69,    -60,    -51,    -40,    -29,    -18,     -6,      7,     19,     30,     41,     52,     61,     70,     77,     82,     87,     89,     90,     90,     88,     84,     78,     72,     64,     55,     44,     33,     22,     10,     -2,    -14,    -26,    -37,    -48,    -58,    -67,    -74,    -81,    -85,    -89,    -90 },
      {     14,     28,     41,     54,     64,     74,     81,     86,     89,     90,     89,     85,     79,     72,     62,     51,     38,     25,     11,     -3,    -18,    -31,    -44,    -56,    -67,    -76,    -82,    -87,    -90,    -90,    -88,    -84,    -78,    -70,    -60,    -48,    -35,    -22,     -8,      7,     21,     34,     47,     59,     69,     77,     84,     88,     90,     90,     88,     83,     76,     67,     57,     45,     32,     19,      4,    -10,    -24,    -37,    -50,    -61,    -71,    -79,    -85,    -89,    -90,    -90,    -87,    -81,    -74,    -65,    -55,    -42,    -29,    -15,     -1,     13,     27,     40,     53,     64,     73,     81,     86,     89,     90,     89,     86,     80,     72,     63,     52,     39,     26,     12,     -2,    -16,    -30,    -43,    -55,    -66,    -75,    -82,    -87,    -90,    -90,    -89,    -85,    -78,    -70,    -60,    -49,    -36,    -23,     -9,      6,     20,     33,     46,     58,     68,     77,     83,     88,     90 },
      {     16,     32,     47,     60,     72,     81,     87,     90,     90,     87,     81,     73,     62,     49,     34,     19,      2,    -14,    -30,    -45,    -59,    -70,    -79,    -86,    -90,    -90,    -88,    -82,    -74,    -64,    -51,    -36,    -21,     -4,     12,     28,     43,     57,     69,     78,     85,     89,     90,     88,     83,     76,     65,     53,     38,     23,      7,    -10,    -26,    -41,    -55,    -67,    -77,    -85,    -89,    -90,    -89,    -84,    -77,    -67,    -55,    -40,    -25,     -9,      8,     24,     39,     54,     66,     76,     84,     89,     90,     89,     85,     78,     68,     56,     42,     27,     11,     -6,    -22,    -37,    -52,    -64,    -75,    -83,    -88,    -90,    -89,    -86,    -79,    -70,    -58,    -44,    -29,    -13,      3,     20,     35,     50,     63,     74,     82,     88,     90,     90,     86,     80,     71,     60,     46,     31,     15,     -1,    -18,    -33,    -48,    -61,    -72,    -81,    -87,    -90 },
      {     19,     36,     53,     67,     78,     86,     90,     90,     86,     79,     68,     55,     38,     21,      2,    -16,    -34,    -51,    -65,    -77,    -85,    -89,    -90,    -87,    -80,    -70,    -56,    -40,    -23,     -4,     14,     32,     49,     64,     76,     84,     89,     90,     88,     81,     71,     58,     42,     25,      7,    -12,    -30,    -47,    -62,    -74,    -83,    -89,    -90,    -88,    -82,    -72,    -60,    -44,    -27,     -9,     10,     28,     45,     60,     73,     82,     88,     90,     89,     83,     74,     61,     46,     29,     11,     -8,    -26,    -43,    -59,    -72,    -81,    -88,    -90,    -89,    -84,    -75,    -63,    -48,    -31,    -13,      6,     24,     41,     57,     70,     81,     87,     90,     89,     85,     76,     64,     50,     33,     15,     -3,    -22,    -39,    -55,    -69,    -79,    -87,    -90,    -90,    -85,    -77,    -66,    -52,    -35,    -18,      1,     20,     37,     54,     67,     78,     86,     90 },
      {     21,     40,     58,     72,     83,     89,     90,     87,     78,     66,     50,     31,     11,    -10,    -30,    -49,    -65,    -78,    -86,    -90,    -89,    -83,    -73,    -59,    -41,    -22,     -1,     20,     39,     57,     72,     82,     89,     90,     87,     79,     67,     51,     32,     12,     -9,    -29,    -48,    -64,    -77,    -86,    -90,    -89,    -84,    -74,    -60,    -42,    -23,     -2,     19,     38,     56,     71,     82,     89,     90,     87,     79,     67,     52,     33,     13,     -8,    -28,    -47,    -64,    -77,    -86,    -90,    -89,    -84,    -74,    -60,    -43,    -24,     -3,     18,     37,     55,     70,     81,     88,     90,     88,     80,     68,     53,     34,     14,     -7,    -27,    -46,    -63,    -76,    -85,    -90,    -90,    -85,    -75,    -61,    -44,    -25,     -4,     16,     36,     55,     70,     81,     88,     90,     88,     81,     69,     54,     35,     15,     -6,    -26,    -45,    -62,    -76,    -85,    -90 },
      {     23,     44,     63,     77,     87,     90,     88,     80,     67,     49,     28,      6,    -18,    -39,    -59,    -74,    -85,    -90,    -89,    -82,    -70,    -54,    -33,    -11,     12,     34,     55,     71,     83,     89,     90,     85,     74,     58,     38,     16,     -7,    -29,    -50,    -67,    -81,    -88,    -90,    -86,    -77,    -62,    -43,    -22,      1,     24,     45,     64,     78,     87,     90,     88,     79,     66,     48,     27,      4,    -19,    -40,    -60,    -75,    -85,    -90,    -89,    -82,    -70,    -53,    -32,    -10,     13,     35,     55,     72,     83,     89,     90,     84,     73,     57,     37,     15,     -8,    -30,    -51,    -68,    -81,    -89,    -90,    -86,    -76,    -61,    -42,    -21,      2,     25,     46,     64,     78,     87,     90,     88,     79,     65,     47,     26,      3,    -20,    -41,    -60,    -76,    -86,    -90,    -89,    -81,    -69,    -52,    -31,     -9,     14,     36,     56,     72,     84,     90 },
      {     25,     48,     67,     81,     89,     90,     83,     70,     52,     29,      4,    -21,    -44,    -64,    -79,    -88,    -90,    -85,    -73,    -55,    -33,     -9,     16,     40,     61,     77,     87,     90,     86,     76,     59,     37,     13,    -12,    -36,    -58,    -75,    -86,    -90,    -88,    -78,    -62,    -41,    -18,      8,     32,     55,     72,     85,     90,     89,     80,     65,     45,     22,     -3,    -28,    -51,    -70,    -83,    -90,    -89,    -82,    -68,    -49,    -26,     -1,     24,     47,     67,     81,     89,     90,     84,     71,     53,     30,      6,    -20,    -43,    -64,    -79,    -88,    -90,    -85,    -74,    -56,    -34,    -10,     15,     39,     60,     77,     87,     90,     87,     76,     60,     38,     14,    -11,    -35,    -57,    -74,    -86,    -90,    -88,    -78,    -63,    -42,    -19,      7,     31,     54,     72,     84,     90,     89,     81,     66,     46,     23,     -2,    -27,    -50,    -69,    -82,    -89 },
      {     27,     52,     72,     85,     90,     87,     76,     58,     34,      8,    -20,    -45,    -67,    -82,    -90,    -89,    -80,    -64,    -41,    -15,     12,     38,     61,     78,     88,     90,     83,     69,     48,     23,     -4,    -31,    -55,    -74,    -86,    -90,    -86,    -74,    -55,    -30,     -3,     24,     49,     70,     84,     90,     88,     78,     60,     37,     11,    -16,    -42,    -64,    -81,    -89,    -89,    -81,    -66,    -44,    -19,      9,     35,     59,     77,     88,     90,     85,     71,     51,     26,     -1,    -28,    -53,    -72,    -85,    -90,    -87,    -76,    -57,    -33,     -7,     21,     46,     67,     82,     90,     89,     79,     63,     40,     14,    -13,    -39,    -62,    -79,    -89,    -90,    -83,    -68,    -47,    -22,      6,     32,     56,     75,     87,     90,     86,     73,     54,     29,      2,    -25,    -50,    -70,    -84,    -90,    -88,    -77,    -60,    -36,    -10,     18,     43,     65,     81,     89 },
      {     29,     55,     76,     88,     90,     83,     67,     43,     15,    -14,    -42,    -66,    -82,    -90,    -88,    -76,    -56,    -30,     -1,     28,     55,     75,     87,     90,     83,     67,     44,     16,    -13,    -41,    -65,    -82,    -90,    -88,    -77,    -57,    -31,     -2,     27,     54,     74,     87,     90,     84,     68,     45,     18,    -12,    -40,    -64,    -81,    -90,    -88,    -77,    -58,    -32,     -3,     26,     53,     74,     87,     90,     84,     69,     46,     19,    -11,    -39,    -64,    -81,    -90,    -89,    -78,    -59,    -33,     -4,     25,     52,     73,     86,     90,     85,     70,     47,     20,    -10,    -38,    -63,    -81,    -89,    -89,    -78,    -60,    -34,     -6,     24,     51,     72,     86,     90,     85,     70,     48,     21,     -9,    -37,    -62,    -80,    -89,    -89,    -79,    -60,    -35,     -7,     23,     50,     72,     86,     90,     85,     71,     49,     22,     -8,    -36,    -61,    -79,    -89 },
      {     31,     59,     79,     89,     89,     77,     55,     27,     -4,    -35,    -62,    -81,    -90,    -88,    -74,    -52,    -23,      9,     39,     65,     83,     90,     86,     72,     48,     19,    -13,    -43,    -68,    -85,    -90,    -85,    -69,    -44,    -14,     18,     47,     71,     86,     90,     83,     66,     40,     10,    -22,    -51,    -74,    -87,    -90,    -81,    -63,    -36,     -6,     26,     55,     76,     88,     89,     79,     60,     32,      1,    -30,    -58,    -78,    -89,    -89,    -77,    -56,    -28,      3,     34,     61,     81,     90,     88,     75,     53,     24,     -8,    -38,    -64,    -82,    -90,    -87,    -72,    -49,    -20,     12,     42,     67,     84,     90,     85,     70,     45,     15,    -16,    -46,    -70,    -86,    -90,    -84,    -67,    -41,    -11,     21,     50,     73,     87,     90,     82,     64,     37,      7,    -25,    -54,    -76,    -88,    -90,    -80,    -60,    -33,     -2,     29,     57,     78,     89 },
      {     33,     62,     82,     90,     86,     69,     42,     10,    -24,    -55,    -77,    -89,    -88,    -75,    -51,    -20,     14,     46,     72,     87,     90,     80,     59,     29,     -4,    -37,    -65,    -84,    -90,    -84,    -66,    -38,     -6,     28,     58,     79,     90,     87,     72,     47,     15,    -19,    -50,    -74,    -88,    -89,    -78,    -55,    -25,      9,     41,     68,     85,     90,     82,     63,     34,      1,    -32,    -61,    -81,    -90,    -86,    -70,    -43,    -11,     23,     54,     77,     89,     89,     76,     52,     21,    -13,    -45,    -71,    -87,    -90,    -81,    -60,    -30,      3,     36,     64,     83,     90,     85,     67,     39,      7,    -27,    -57,    -79,    -90,    -88,    -73,    -48,    -16,     18,     49,     74,     88,     89,     78,     56,     26,     -8,    -40,    -67,    -85,    -90,    -83,    -64,    -35,     -2,     31,     60,     81,     90,     86,     70,     44,     12,    -22,    -53,    -76,    -89 },
      {     35,     65,     85,     90,     81,     60,     28,     -8,    -42,    -70,    -87,    -90,    -78,    -54,    -21,     15,     49,     75,     89,     88,     74,     47,     13,    -23,    -55,    -79,    -90,    -86,    -69,    -40,     -6,     30,     61,     82,     90,     84,     64,     33,     -2,    -37,    -67,    -85,    -90,    -81,    -58,    -26,     10,     44,     72,     88,     89,     77,     52,     19,    -18,    -51,    -76,    -89,    -88,    -72,    -45,    -11,     25,     57,     80,     90,     86,     67,     38,      3,    -32,    -63,    -83,    -90,    -83,    -62,    -31,      4,     39,     68,     86,     90,     79,     56,     24,    -12,    -46,    -73,    -88,    -89,    -76,    -50,    -16,     20,     53,     77,     89,     87,     71,     43,      9,    -27,    -59,    -81,    -90,    -85,    -66,    -36,     -1,     34,     64,     84,     90,     82,     60,     29,     -7,    -41,    -70,    -87,    -90,    -78,    -55,    -22,     14,     48,     74,     89 },
      {     37,     68,     87,     89,     76,     49,     13,    -25,    -59,    -82,    -90,    -82,    -60,    -26,     12,     48,     76,     89,     87,     69,     38,      1,    -36,    -67,    -86,    -90,    -77,    -50,    -14,     24,     58,     81,     90,     83,     60,     27,    -11,    -47,    -75,    -89,    -87,    -70,    -39,     -2,     35,     67,     86,     90,     77,     51,     15,    -23,    -57,    -81,    -90,    -83,    -61,    -28,     10,     46,     74,     89,     88,     70,     40,      3,    -34,    -66,    -86,    -90,    -78,    -52,    -16,     22,     56,     81,     90,     84,     62,     29,     -9,    -45,    -74,    -89,    -88,    -71,    -41,     -4,     33,     65,     85,     90,     78,     53,     18,    -21,    -55,    -80,    -90,    -84,    -63,    -30,      8,     44,     73,     89,     88,     72,     42,      6,    -32,    -64,    -85,    -90,    -79,    -54,    -19,     20,     55,     79,     90,     85,     64,     31,     -7,    -43,    -72,    -88 },
      {     39,     71,     88,     88,     70,     37,     -2,    -41,    -72,    -89,    -87,    -68,    -35,      4,     43,     74,     89,     87,     67,     33,     -7,    -45,    -75,    -89,    -86,    -65,    -31,      9,     47,     76,     90,     85,     64,     29,    -11,    -49,    -77,    -90,    -85,    -62,    -27,     13,     51,     78,     90,     84,     60,     25,    -15,    -53,    -79,    -90,    -83,    -59,    -23,     18,     55,     81,     90,     82,     57,     21,    -20,    -56,    -81,    -90,    -81,    -55,    -19,     22,     58,     82,     90,     80,     54,     16,    -24,    -60,    -83,    -90,    -79,    -52,    -14,     26,     61,     84,     90,     78,     50,     12,    -28,    -63,    -85,    -90,    -77,    -48,    -10,     30,     64,     86,     90,     76,     46,      8,    -32,    -66,    -86,    -89,    -74,    -44,     -6,     34,     67,     87,     89,     73,     42,      3,    -36,    -69,    -88,    -89,    -72,    -40,     -1,     38,     70,     88 },
      {     41,     74,     89,     85,     62,     25,    -18,    -56,    -82,    -90,    -78,    -48,     -8,     34,     69,     88,     88,     67,     32,    -10,    -50,    -79,    -90,    -81,    -55,    -15,     27,     64,     86,     89,     72,     39,     -2,    -43,    -75,    -90,    -85,    -60,    -23,     20,     58,     83,     90,     77,     46,      6,    -36,    -70,    -89,    -87,    -66,    -30,     12,     52,     80,     90,     81,     53,     13,    -29,    -65,    -87,    -89,    -71,    -37,      4,     45,     76,     90,     84,     59,     21,    -22,    -60,    -84,    -90,    -76,    -44,     -3,     38,     72,     89,     86,     64,     28,    -14,    -54,    -81,    -90,    -79,    -51,    -11,     31,     67,     87,     88,     70,     35,     -7,    -47,    -77,    -90,    -83,    -57,    -19,     24,     61,     85,     90,     74,     42,      1,    -40,    -73,    -89,    -86,    -63,    -26,     16,     55,     82,     90,     78,     49,      9,    -33,    -68,    -88 },
      {     43,     76,     90,     82,     54,     12,    -32,    -69,    -89,    -86,    -63,    -24,     21,     60,     85,     89,     71,     35,     -9,    -51,    -81,    -90,    -78,    -46,     -3,     40,     74,     90,     83,     56,     15,    -29,    -67,    -88,    -87,    -65,    -27,     18,     58,     84,     90,     73,     38,     -6,    -48,    -79,    -90,    -79,    -49,     -7,     37,     72,     89,     85,     59,     19,    -26,    -64,    -87,    -88,    -67,    -30,     14,     55,     83,     90,     75,     41,     -2,    -45,    -77,    -90,    -81,    -52,    -10,     34,     70,     89,     86,     61,     22,    -23,    -62,    -86,    -89,    -70,    -33,     11,     53,     81,     90,     77,     44,      1,    -42,    -76,    -90,    -82,    -55,    -13,     31,     68,     88,     87,     64,     25,    -20,    -60,    -85,    -89,    -72,    -36,      8,     50,     80,     90,     78,     47,      4,    -39,    -74,    -90,    -84,    -57,    -16,     28,     66,     88 },
      {     45,     78,     90,     78,     44,     -1,    -46,    -79,    -90,    -77,    -43,      2,     47,     79,     90,     77,     42,     -3,    -48,    -80,    -90,    -76,    -41,      4,     49,     81,     90,     76,     40,     -6,    -50,    -81,    -90,    -75,    -39,      7,     51,     81,     90,     74,     38,     -8,    -52,    -82,    -90,    -74,    -37,      9,     53,     82,     90,     73,     36,    -10,    -54,    -83,    -90,    -72,    -35,     11,     55,     83,     90,     72,     34,    -12,    -55,    -84,    -89,    -71,    -33,     13,     56,     84,     89,     70,     32,    -14,    -57,    -85,    -89,    -70,    -31,     15,     58,     85,     89,     69,     30,    -16,    -59,    -85,    -89,    -68,    -29,     18,     60,     86,     89,     67,     28,    -19,    -60,    -86,    -88,    -67,    -27,     20,     61,     86,     88,     66,     26,    -21,    -62,    -87,    -88,    -65,    -25,     22,     63,     87,     88,     64,     24,    -23,    -64,    -87 },
      {     47,     81,     90,     73,     34,    -14,    -59,    -86,    -88,    -64,    -21,     28,     69,     89,     83,     53,      7,    -41,    -77,    -90,    -77,    -40,      8,     54,     84,     89,     68,     27,    -22,    -64,    -88,    -86,    -58,    -13,     35,     74,     90,     80,     46,     -1,    -48,    -81,    -90,    -72,    -33,     15,     60,     86,     88,     63,     20,    -29,    -70,    -89,    -83,    -52,     -6,     42,     78,     90,     76,     39,     -9,    -55,    -84,    -89,    -67,    -26,     23,     65,     88,     85,     57,     12,    -36,    -74,    -90,    -79,    -45,      2,     49,     81,     90,     72,     32,    -16,    -60,    -87,    -87,    -62,    -19,     30,     70,     90,     82,     51,      4,    -43,    -78,    -90,    -76,    -38,     10,     55,     85,     89,     67,     25,    -24,    -66,    -89,    -85,    -56,    -11,     37,     75,     90,     79,     44,     -3,    -50,    -82,    -90,    -71,    -31,     18,     61,     87 },
      {     49,     82,     89,     67,     24,    -27,    -70,    -90,    -81,    -46,      3,     52,     84,     89,     65,     21,    -30,    -72,    -90,    -79,    -43,      7,     55,     85,     88,     63,     18,    -33,    -74,    -90,    -78,    -40,     10,     57,     86,     87,     60,     14,    -36,    -76,    -90,    -76,    -37,     13,     60,     87,     86,     58,     11,    -39,    -77,    -90,    -74,    -34,     16,     62,     88,     85,     55,      8,    -42,    -79,    -90,    -72,    -31,     20,     64,     89,     84,     53,      4,    -45,    -81,    -90,    -70,    -28,     23,     67,     89,     83,     50,      1,    -48,    -82,    -89,    -68,    -25,     26,     69,     90,     81,     47,     -2,    -51,    -83,    -89,    -66,    -22,     29,     71,     90,     80,     44,     -6,    -54,    -85,    -88,    -64,    -19,     32,     73,     90,     78,     41,     -9,    -56,    -86,    -88,    -61,    -15,     35,     75,     90,     77,     38,    -12,    -59,    -87 },
      {     51,     84,     88,     61,     13,    -39,    -78,    -90,    -70,    -26,     27,     71,     90,     78,     38,    -14,    -62,    -88,    -84,    -50,      1,     52,     85,     88,     60,     12,    -40,    -79,    -90,    -70,    -25,     28,     72,     90,     77,     37,    -15,    -63,    -89,    -83,    -49,      2,     53,     85,     88,     60,     11,    -41,    -79,    -90,    -69,    -24,     29,     72,     90,     77,     36,    -16,    -64,    -89,    -83,    -48,      3,     54,     85,     87,     59,     10,    -42,    -80,    -90,    -68,    -23,     30,     73,     90,     76,     35,    -18,    -64,    -89,    -82,    -47,      4,     55,     86,     87,     58,      9,    -43,    -81,    -90,    -67,    -22,     31,     74,     90,     76,     34,    -19,    -65,    -89,    -82,    -46,      6,     55,     86,     87,     57,      8,    -44,    -81,    -89,    -67,    -21,     32,     74,     90,     75,     33,    -20,    -66,    -89,    -81,    -45,      7,     56,     86 },
      {     53,     86,     86,     55,      2,    -51,    -85,    -87,    -56,     -4,     49,     84,     88,     58,      7,    -47,    -83,    -88,    -60,     -9,     45,     82,     89,     61,     11,    -43,    -81,    -89,    -63,    -13,     41,     81,     89,     64,     15,    -39,    -79,    -90,    -66,    -18,     37,     78,     90,     67,     20,    -35,    -77,    -90,    -69,    -22,     33,     76,     90,     70,     24,    -31,    -75,    -90,    -72,    -26,     29,     74,     90,     73,     28,    -27,    -72,    -90,    -74,    -30,     25,     71,     90,     76,     32,    -23,    -70,    -90,    -77,    -34,     21,     68,     90,     78,     36,    -19,    -67,    -90,    -79,    -38,     16,     65,     89,     80,     40,    -14,    -64,    -89,    -81,    -42,     12,     62,     89,     82,     44,    -10,    -60,    -88,    -83,    -46,      8,     59,     88,     84,     48,     -6,    -57,    -87,    -85,    -50,      3,     55,     87,     85,     52,     -1,    -54,    -86 },
      {     55,     87,     84,     47,     -9,    -61,    -89,    -81,    -39,     18,     67,     90,     76,     31,    -26,    -73,    -90,    -71,    -23,     34,     78,     90,     65,     14,    -42,    -82,    -88,    -59,     -6,     50,     85,     86,     52,     -3,    -57,    -88,    -83,    -44,     12,     64,     89,     79,     36,    -21,    -70,    -90,    -74,    -28,     29,     75,     90,     69,     20,    -37,    -79,    -89,    -63,    -11,     45,     83,     88,     56,      2,    -53,    -86,    -85,    -49,      7,     60,     89,     81,     41,    -15,    -66,    -90,    -77,    -33,     24,     72,     90,     72,     25,    -32,    -77,    -90,    -67,    -16,     40,     81,     89,     60,      8,    -48,    -85,    -87,    -54,      1,     55,     87,     84,     46,    -10,    -62,    -89,    -80,    -38,     19,     68,     90,     76,     30,    -27,    -74,    -90,    -70,    -22,     35,     78,     90,     64,     13,    -43,    -82,    -88,    -58,     -4,     51,     86 },
      {     56,     88,     81,     39,    -20,    -70,    -90,    -71,    -21,     38,     81,     88,     57,      1,    -55,    -88,    -82,    -40,     19,     70,     90,     72,     22,    -37,    -81,    -89,    -58,     -2,     55,     88,     82,     41,    -18,    -69,    -90,    -72,    -23,     36,     80,     89,     59,      3,    -54,    -87,    -83,    -42,     16,     68,     90,     73,     24,    -35,    -79,    -89,    -60,     -4,     53,     87,     83,     43,    -15,    -67,    -90,    -74,    -25,     34,     79,     89,     60,      6,    -52,    -87,    -84,    -44,     14,     67,     90,     74,     26,    -33,    -78,    -89,    -61,     -7,     51,     86,     84,     45,    -13,    -66,    -90,    -75,    -27,     32,     78,     89,     62,      8,    -50,    -86,    -85,    -46,     12,     65,     90,     76,     28,    -31,    -77,    -90,    -63,     -9,     49,     86,     85,     47,    -11,    -64,    -90,    -76,    -29,     30,     77,     90,     64,     10,    -48,    -85 },
      {     58,     89,     78,     31,    -30,    -78,    -89,    -59,     -1,     57,     89,     79,     32,    -29,    -77,    -89,    -60,     -2,     56,     89,     79,     33,    -28,    -77,    -89,    -60,     -3,     55,     88,     80,     34,    -27,    -76,    -90,    -61,     -4,     55,     88,     81,     35,    -26,    -76,    -90,    -62,     -6,     54,     88,     81,     36,    -25,    -75,    -90,    -63,     -7,     53,     88,     81,     37,    -24,    -74,    -90,    -64,     -8,     52,     87,     82,     38,    -23,    -74,    -90,    -64,     -9,     51,     87,     82,     39,    -22,    -73,    -90,    -65,    -10,     50,     87,     83,     40,    -21,    -72,    -90,    -66,    -11,     49,     86,     83,     41,    -20,    -72,    -90,    -67,    -12,     48,     86,     84,     42,    -19,    -71,    -90,    -67,    -13,     47,     86,     84,     43,    -18,    -70,    -90,    -68,    -14,     46,     85,     85,     44,    -16,    -70,    -90,    -69,    -15,     45,     85 },
      {     60,     90,     75,     23,    -40,    -84,    -85,    -44,     19,     72,     90,     63,      4,    -56,    -89,    -77,    -27,     36,     82,     87,     48,    -14,    -70,    -90,    -66,     -9,     53,     88,     79,     31,    -32,    -80,    -88,    -52,     10,     67,     90,     69,     13,    -49,    -87,    -81,    -35,     28,     78,     89,     55,     -6,    -64,    -90,    -72,    -18,     45,     86,     83,     39,    -24,    -76,    -89,    -59,      1,     60,     90,     74,     22,    -41,    -84,    -85,    -43,     20,     73,     90,     62,      3,    -57,    -89,    -77,    -26,     37,     82,     86,     47,    -15,    -70,    -90,    -65,     -8,     54,     88,     79,     30,    -33,    -81,    -88,    -51,     11,     67,     90,     68,     12,    -50,    -87,    -81,    -34,     29,     78,     89,     55,     -7,    -64,    -90,    -71,    -16,     46,     86,     83,     38,    -25,    -76,    -89,    -58,      2,     61,     90,     74,     21,    -42,    -85 },
      {     61,     90,     71,     14,    -50,    -88,    -79,    -28,     37,     83,     85,     41,    -24,    -77,    -89,    -54,     10,     68,     90,     64,      4,    -58,    -90,    -74,    -19,     46,     87,     81,     32,    -33,    -81,    -86,    -45,     20,     74,     89,     57,     -6,    -65,    -90,    -67,     -9,     55,     89,     76,     23,    -42,    -85,    -83,    -36,     29,     79,     88,     49,    -15,    -72,    -90,    -60,      1,     62,     90,     70,     13,    -51,    -88,    -78,    -27,     38,     84,     85,     40,    -25,    -77,    -89,    -53,     11,     69,     90,     64,      3,    -59,    -90,    -73,    -18,     47,     87,     81,     31,    -34,    -82,    -86,    -44,     21,     75,     89,     56,     -7,    -66,    -90,    -67,     -8,     55,     89,     76,     22,    -43,    -86,    -82,    -35,     30,     80,     87,     48,    -16,    -72,    -90,    -60,      2,     63,     90,     70,     12,    -52,    -88,    -78,    -26,     39,     84 },
      {     63,     90,     67,      6,    -59,    -90,    -70,    -11,     55,     89,     74,     16,    -50,    -88,    -77,    -22,     45,     87,     79,     27,    -40,    -85,    -82,    -32,     35,     83,     84,     37,    -30,    -81,    -86,    -42,     25,     78,     88,     47,    -20,    -76,    -89,    -52,     14,     72,     90,     56,     -9,    -69,    -90,    -60,      3,     65,     90,     64,      2,    -61,    -90,    -68,     -8,     57,     90,     72,     13,    -53,    -89,    -75,    -19,     48,     88,     78,     24,    -43,    -86,    -81,    -29,     38,     85,     83,     34,    -33,    -82,    -85,    -39,     28,     80,     87,     44,    -23,    -77,    -88,    -49,     18,     74,     89,     54,    -12,    -71,    -90,    -58,      7,     67,     90,     62,     -1,    -64,    -90,    -66,     -4,     60,     90,     70,     10,    -55,    -89,    -73,    -15,     51,     89,     76,     21,    -46,    -87,    -79,    -26,     41,     86,     81,     31,    -36,    -84 },
      {     64,     90,     62,     -3,    -67,    -90,    -60,      7,     69,     90,     57,    -10,    -71,    -90,    -55,     13,     73,     89,     52,    -16,    -75,    -89,    -49,     20,     77,     88,     46,    -23,    -78,    -87,    -43,     26,     80,     86,     40,    -29,    -81,    -85,    -37,     32,     83,     84,     34,    -35,    -84,    -82,    -31,     38,     85,     81,     28,    -41,    -86,    -79,    -25,     44,     87,     78,     22,    -47,    -88,    -76,    -19,     50,     89,     74,     15,    -53,    -89,    -72,    -12,     55,     90,     70,      9,    -58,    -90,    -68,     -6,     60,     90,     66,      2,    -63,    -90,    -64,      1,     65,     90,     61,     -4,    -67,    -90,    -59,      8,     70,     90,     56,    -11,    -72,    -89,    -54,     14,     74,     89,     51,    -18,    -76,    -88,    -48,     21,     77,     88,     45,    -24,    -79,    -87,    -42,     27,     81,     86,     39,    -30,    -82,    -85,    -36,     33,     83 },
      {     66,     90,     57,    -12,    -74,    -89,    -47,     24,     80,     85,     36,    -35,    -85,    -81,    -25,     46,     88,     74,     13,    -56,    -90,    -67,     -1,     65,     90,     58,    -11,    -73,    -89,    -48,     23,     79,     86,     37,    -34,    -85,    -81,    -26,     45,     88,     75,     14,    -55,    -90,    -67,     -2,     64,     90,     59,    -10,    -72,    -89,    -49,     22,     79,     86,     38,    -33,    -84,    -81,    -27,     44,     88,     76,     15,    -55,    -90,    -68,     -3,     64,     90,     60,     -9,    -72,    -89,    -50,     21,     78,     86,     39,    -32,    -84,    -82,    -28,     43,     88,     76,     16,    -54,    -90,    -69,     -4,     63,     90,     60,     -8,    -71,    -89,    -51,     20,     78,     87,     40,    -31,    -83,    -82,    -29,     42,     87,     77,     18,    -53,    -90,    -70,     -6,     62,     90,     61,     -7,    -70,    -89,    -52,     19,     77,     87,     41,    -30,    -83 },
      {     67,     90,     52,    -21,    -79,    -85,    -33,     40,     87,     76,     13,    -58,    -90,    -62,      8,     72,     89,     45,    -28,    -83,    -82,    -26,     47,     89,     71,      6,    -64,    -90,    -56,     15,     77,     87,     38,    -35,    -86,    -78,    -19,     54,     90,     66,     -2,    -69,    -89,    -50,     23,     81,     84,     31,    -42,    -88,    -74,    -11,     60,     90,     60,    -10,    -74,    -88,    -43,     30,     84,     81,     24,    -49,    -89,    -70,     -3,     65,     90,     55,    -18,    -78,    -86,    -36,     37,     86,     77,     16,    -55,    -90,    -64,      4,     70,     89,     48,    -25,    -81,    -83,    -29,     44,     88,     73,      9,    -61,    -90,    -59,     12,     75,     88,     41,    -32,    -85,    -80,    -22,     51,     90,     68,      1,    -67,    -90,    -53,     20,     79,     85,     34,    -39,    -87,    -76,    -14,     57,     90,     63,     -7,    -72,    -89,    -46,     27,     82 },
      {     69,     89,     46,    -29,    -84,    -79,    -19,     55,     90,     61,    -11,    -76,    -87,    -36,     39,     88,     74,      8,    -64,    -90,    -53,     22,     81,     83,     26,    -49,    -90,    -67,      3,     71,     89,     43,    -32,    -85,    -78,    -15,     58,     90,     59,    -14,    -77,    -86,    -33,     42,     88,     72,      4,    -66,    -90,    -50,     25,     82,     81,     23,    -52,    -90,    -64,      7,     73,     88,     40,    -35,    -86,    -76,    -12,     60,     90,     56,    -18,    -79,    -85,    -30,     45,     89,     70,      1,    -68,    -89,    -47,     28,     84,     80,     20,    -55,    -90,    -62,     10,     75,     87,     37,    -38,    -87,    -74,     -9,     63,     90,     54,    -21,    -81,    -83,    -27,     48,     89,     67,     -2,    -70,    -89,    -44,     31,     85,     78,     16,    -57,    -90,    -60,     13,     77,     86,     34,    -41,    -88,    -72,     -6,     65,     90,     51,    -24,    -82 },
      {     70,     88,     40,    -37,    -88,    -72,     -3,     68,     89,     43,    -34,    -87,    -74,     -7,     66,     89,     46,    -31,    -86,    -76,    -10,     64,     90,     49,    -28,    -85,    -78,    -13,     61,     90,     52,    -25,    -83,    -79,    -16,     59,     90,     55,    -22,    -82,    -81,    -20,     56,     90,     57,    -19,    -81,    -82,    -23,     54,     90,     60,    -15,    -79,    -84,    -26,     51,     90,     62,    -12,    -77,    -85,    -29,     48,     90,     64,     -9,    -76,    -86,    -32,     45,     89,     67,     -6,    -74,    -87,    -35,     42,     89,     69,     -2,    -72,    -88,    -38,     39,     88,     71,      1,    -70,    -89,    -41,     36,     87,     73,      4,    -67,    -89,    -44,     33,     86,     75,      8,    -65,    -90,    -47,     30,     85,     77,     11,    -63,    -90,    -50,     27,     84,     78,     14,    -60,    -90,    -53,     24,     83,     80,     18,    -58,    -90,    -55,     21,     81 },
      {     72,     87,     34,    -45,    -90,    -64,     12,     78,     83,     23,    -55,    -90,    -55,     24,     84,     78,     11,    -64,    -89,    -44,     35,     88,     71,     -1,    -72,    -87,    -33,     46,     90,     63,    -13,    -79,    -83,    -22,     56,     90,     54,    -25,    -84,    -77,    -10,     65,     89,     43,    -36,    -88,    -70,      2,     73,     87,     32,    -47,    -90,    -62,     14,     79,     82,     21,    -57,    -90,    -53,     26,     85,     77,      9,    -66,    -89,    -42,     37,     88,     70,     -3,    -74,    -86,    -31,     48,     90,     61,    -15,    -80,    -82,    -20,     58,     90,     52,    -27,    -85,    -76,     -8,     67,     89,     41,    -38,    -88,    -69,      4,     74,     86,     30,    -49,    -90,    -60,     16,     81,     81,     19,    -59,    -90,    -51,     28,     85,     76,      7,    -67,    -89,    -40,     39,     89,     68,     -6,    -75,    -86,    -29,     50,     90,     60,    -18,    -81 },
      {     73,     86,     28,    -53,    -90,    -54,     27,     86,     74,      1,    -72,    -86,    -29,     52,     90,     55,    -26,    -85,    -74,     -2,     72,     87,     30,    -51,    -90,    -55,     25,     85,     75,      3,    -71,    -87,    -31,     50,     90,     56,    -24,    -85,    -76,     -4,     70,     87,     32,    -49,    -90,    -57,     23,     84,     76,      6,    -70,    -88,    -33,     48,     90,     58,    -22,    -84,    -77,     -7,     69,     88,     34,    -47,    -90,    -59,     21,     83,     77,      8,    -68,    -88,    -35,     46,     90,     60,    -20,    -83,    -78,     -9,     67,     88,     36,    -45,    -90,    -60,     19,     82,     78,     10,    -67,    -89,    -37,     44,     90,     61,    -18,    -82,    -79,    -11,     66,     89,     38,    -43,    -90,    -62,     16,     81,     79,     12,    -65,    -89,    -39,     42,     89,     63,    -15,    -81,    -80,    -13,     64,     89,     40,    -41,    -89,    -64,     14,     81 },
      {     74,     85,     22,    -60,    -90,    -42,     41,     90,     60,    -21,    -84,    -75,     -1,     74,     85,     23,    -59,    -90,    -43,     40,     89,     61,    -20,    -84,    -76,     -2,     73,     85,     24,    -58,    -90,    -44,     39,     89,     62,    -19,    -83,    -76,     -3,     72,     86,     25,    -57,    -90,    -45,     38,     89,     63,    -18,    -83,    -77,     -4,     72,     86,     26,    -56,    -90,    -46,     37,     89,     64,    -16,    -82,    -77,     -6,     71,     86,     27,    -55,    -90,    -47,     36,     89,     64,    -15,    -82,    -78,     -7,     70,     87,     28,    -55,    -90,    -48,     35,     89,     65,    -14,    -81,    -78,     -8,     70,     87,     29,    -54,    -90,    -49,     34,     88,     66,    -13,    -81,    -79,     -9,     69,     87,     30,    -53,    -90,    -50,     33,     88,     67,    -12,    -81,    -79,    -10,     68,     88,     31,    -52,    -90,    -51,     32,     88,     67,    -11,    -80 },
      {     76,     83,     15,    -66,    -88,    -30,     55,     90,     44,    -41,    -90,    -57,     27,     87,     68,    -12,    -81,    -77,     -3,     74,     84,     19,    -64,    -89,    -33,     52,     90,     47,    -38,    -89,    -60,     24,     86,     70,     -9,    -80,    -79,     -7,     72,     85,     22,    -61,    -89,    -36,     49,     90,     50,    -35,    -89,    -62,     21,     85,     72,     -6,    -78,    -81,    -10,     70,     86,     25,    -59,    -90,    -39,     46,     90,     53,    -32,    -88,    -64,     18,     84,     74,     -2,    -77,    -82,    -13,     67,     87,     28,    -56,    -90,    -42,     43,     90,     55,    -29,    -88,    -67,     14,     82,     76,      1,    -75,    -83,    -16,     65,     88,     31,    -54,    -90,    -45,     40,     90,     58,    -26,    -87,    -69,     11,     81,     78,      4,    -73,    -85,    -20,     63,     89,     34,    -51,    -90,    -48,     37,     89,     60,    -23,    -86,    -71,      8,     79 },
      {     77,     81,      9,    -72,    -85,    -18,     66,     87,     26,    -60,    -89,    -34,     53,     90,     42,    -45,    -90,    -50,     37,     90,     57,    -29,    -88,    -64,     21,     86,     70,    -12,    -82,    -75,      3,     78,     79,      6,    -74,    -83,    -14,     68,     86,     23,    -62,    -89,    -31,     55,     90,     39,    -48,    -90,    -47,     40,     90,     55,    -32,    -89,    -61,     24,     87,     67,    -15,    -84,    -73,      7,     80,     78,      2,    -76,    -82,    -11,     70,     85,     20,    -64,    -88,    -28,     58,     89,     36,    -51,    -90,    -44,     43,     90,     52,    -35,    -89,    -59,     27,     88,     65,    -19,    -85,    -71,     10,     81,     76,     -1,    -77,    -81,     -8,     72,     84,     16,    -67,    -87,    -25,     60,     89,     33,    -54,    -90,    -41,     46,     90,     49,    -38,    -90,    -56,     30,     88,     63,    -22,    -86,    -69,     13,     83,     74,     -4,    -79 },
      {     78,     79,      2,    -77,    -80,     -4,     76,     81,      7,    -74,    -82,     -9,     73,     83,     11,    -72,    -84,    -13,     70,     85,     15,    -69,    -85,    -18,     67,     86,     20,    -66,    -87,    -22,     64,     87,     24,    -63,    -88,    -26,     61,     88,     28,    -60,    -89,    -30,     58,     89,     32,    -56,    -89,    -34,     55,     90,     36,    -53,    -90,    -38,     51,     90,     40,    -49,    -90,    -42,     47,     90,     44,    -45,    -90,    -46,     43,     90,     48,    -41,    -90,    -50,     39,     90,     52,    -37,    -90,    -54,     35,     90,     55,    -33,    -89,    -57,     31,     89,     59,    -29,    -89,    -60,     27,     88,     62,    -25,    -88,    -64,     23,     87,     65,    -21,    -86,    -67,     19,     86,     68,    -16,    -85,    -70,     14,     84,     71,    -12,    -83,    -72,     10,     82,     74,     -8,    -81,    -75,      6,     81,     76,     -3,    -79,    -77,      1,     78 },
      {     79,     77,     -4,    -81,    -74,      9,     83,     72,    -13,    -85,    -69,     18,     86,     66,    -22,    -87,    -63,     26,     88,     60,    -30,    -89,    -56,     34,     90,     53,    -38,    -90,    -49,     42,     90,     45,    -46,    -90,    -41,     50,     90,     37,    -54,    -90,    -33,     57,     89,     29,    -60,    -88,    -25,     64,     87,     21,    -67,    -86,    -16,     70,     84,     12,    -72,    -82,     -8,     75,     81,      3,    -77,    -78,      1,     79,     76,     -6,    -81,    -74,     10,     83,     71,    -14,    -85,    -68,     19,     86,     65,    -23,    -88,    -62,     27,     89,     59,    -31,    -89,    -55,     35,     90,     52,    -39,    -90,    -48,     43,     90,     44,    -47,    -90,    -40,     51,     90,     36,    -55,    -89,    -32,     58,     89,     28,    -61,    -88,    -24,     64,     87,     20,    -67,    -85,    -15,     70,     84,     11,    -73,    -82,     -7,     76,     80,      2,    -78 },
      {     80,     74,    -11,    -85,    -67,     22,     88,     60,    -32,    -90,    -51,     42,     90,     41,    -52,    -90,    -31,     60,     88,     21,    -68,    -84,    -10,     75,     79,     -1,    -81,    -74,     12,     85,     67,    -23,    -88,    -59,     33,     90,     50,    -43,    -90,    -40,     53,     89,     30,    -61,    -87,    -20,     69,     84,      9,    -76,    -79,      2,     81,     73,    -13,    -85,    -66,     24,     88,     58,    -34,    -90,    -49,     44,     90,     39,    -54,    -89,    -29,     62,     87,     19,    -70,    -83,     -8,     76,     78,     -3,    -81,    -72,     14,     86,     65,    -25,    -89,    -57,     35,     90,     48,    -45,    -90,    -38,     55,     89,     28,    -63,    -87,    -18,     70,     83,      7,    -77,    -78,      4,     82,     72,    -15,    -86,    -64,     26,     89,     56,    -36,    -90,    -47,     46,     90,     37,    -55,    -89,    -27,     64,     86,     16,    -71,    -82,     -6,     77 },
      {     81,     72,    -18,    -87,    -60,     34,     90,     45,    -50,    -90,    -29,     64,     86,     12,    -75,    -78,      6,     83,     68,    -23,    -89,    -55,     39,     90,     40,    -55,    -89,    -24,     67,     84,      7,    -78,    -76,     11,     85,     64,    -28,    -89,    -51,     44,     90,     35,    -59,    -88,    -19,     71,     81,      1,    -81,    -72,     16,     87,     60,    -33,    -90,    -46,     49,     90,     30,    -63,    -86,    -13,     74,     79,     -4,    -83,    -69,     22,     88,     56,    -38,    -90,    -41,     54,     89,     25,    -67,    -84,     -8,     77,     76,    -10,    -85,    -65,     27,     89,     52,    -43,    -90,    -36,     58,     88,     20,    -70,    -82,     -2,     80,     73,    -15,    -87,    -61,     32,     90,     47,    -48,    -90,    -31,     62,     86,     14,    -74,    -79,      3,     82,     70,    -21,    -88,    -57,     37,     90,     42,    -53,    -89,    -26,     66,     85,      9,    -77 },
      {     82,     69,    -24,    -89,    -51,     46,     90,     29,    -65,    -84,     -6,     79,     72,    -19,    -88,    -55,     41,     90,     34,    -61,    -86,    -11,     77,     76,    -13,    -87,    -60,     36,     90,     39,    -57,    -88,    -16,     74,     78,     -8,    -85,    -64,     31,     90,     44,    -53,    -89,    -22,     70,     81,     -2,    -83,    -67,     26,     89,     49,    -48,    -90,    -27,     67,     83,      3,    -81,    -71,     21,     89,     54,    -43,    -90,    -32,     63,     85,      9,    -78,    -74,     15,     87,     58,    -38,    -90,    -37,     59,     87,     14,    -75,    -77,     10,     86,     62,    -33,    -90,    -42,     55,     88,     20,    -72,    -80,      4,     84,     66,    -28,    -90,    -47,     50,     89,     25,    -68,    -82,     -1,     81,     70,    -23,    -89,    -52,     45,     90,     30,    -64,    -85,     -7,     79,     73,    -18,    -88,    -56,     40,     90,     35,    -60,    -86,    -12,     76 },
      {     83,     66,    -30,    -90,    -41,     57,     87,     12,    -77,    -74,     19,     89,     52,    -47,    -89,    -24,     70,     80,     -7,    -85,    -61,     36,     90,     35,    -62,    -85,     -6,     81,     70,    -25,    -90,    -46,     53,     88,     18,    -74,    -77,     13,     87,     56,    -42,    -90,    -29,     67,     82,     -1,    -83,    -65,     31,     90,     40,    -58,    -87,    -11,     78,     73,    -20,    -89,    -51,     48,     89,     23,    -71,    -79,      8,     86,     60,    -37,    -90,    -34,     63,     85,      4,    -81,    -69,     26,     90,     45,    -54,    -88,    -16,     75,     76,    -14,    -88,    -55,     43,     90,     28,    -67,    -82,      2,     84,     64,    -32,    -90,    -39,     59,     86,     10,    -78,    -72,     21,     89,     50,    -49,    -89,    -22,     72,     79,     -9,    -86,    -60,     38,     90,     33,    -64,    -84,     -3,     81,     68,    -27,    -90,    -44,     55,     88,     15,    -76 },
      {     84,     63,    -36,    -90,    -31,     67,     81,     -6,    -86,    -59,     41,     90,     26,    -70,    -79,     11,     87,     55,    -46,    -89,    -21,     74,     76,    -16,    -89,    -50,     51,     88,     15,    -77,    -73,     22,     89,     45,    -55,    -87,    -10,     79,     70,    -27,    -90,    -40,     60,     85,      4,    -82,    -66,     32,     90,     35,    -64,    -83,      1,     84,     62,    -37,    -90,    -30,     67,     81,     -7,    -86,    -58,     42,     90,     25,    -71,    -78,     12,     88,     54,    -47,    -89,    -20,     74,     76,    -18,    -89,    -49,     52,     88,     14,    -77,    -72,     23,     90,     44,    -56,    -87,     -9,     80,     69,    -28,    -90,    -39,     60,     85,      3,    -82,    -65,     33,     90,     34,    -64,    -83,      2,     85,     61,    -38,    -90,    -29,     68,     81,     -8,    -86,    -57,     43,     90,     24,    -72,    -78,     13,     88,     53,    -48,    -89,    -19,     75 },
      {     85,     60,    -42,    -90,    -21,     75,     74,    -23,    -90,    -40,     61,     84,     -2,    -85,    -58,     44,     89,     19,    -76,    -72,     25,     90,     38,    -63,    -83,      4,     86,     56,    -46,    -89,    -16,     77,     71,    -27,    -90,    -36,     64,     82,     -7,    -87,    -55,     48,     89,     14,    -78,    -70,     29,     90,     34,    -66,    -81,      9,     87,     53,    -50,    -88,    -12,     79,     68,    -31,    -90,    -32,     67,     80,    -11,    -88,    -51,     52,     88,     10,    -81,    -67,     33,     90,     30,    -69,    -79,     13,     88,     49,    -54,    -87,     -8,     81,     65,    -35,    -90,    -28,     70,     78,    -15,    -89,    -47,     55,     86,      6,    -82,    -64,     37,     90,     26,    -72,    -77,     18,     89,     45,    -57,    -86,     -3,     83,     62,    -39,    -90,    -24,     73,     76,    -20,    -89,    -43,     59,     85,      1,    -84,    -60,     41,     90,     22,    -74 },
      {     85,     56,    -48,    -88,    -10,     81,     64,    -39,    -90,    -20,     77,     70,    -30,    -90,    -29,     71,     76,    -21,    -90,    -38,     64,     81,    -11,    -88,    -47,     57,     85,     -1,    -86,    -55,     49,     88,      9,    -82,    -63,     40,     90,     19,    -77,    -70,     31,     90,     28,    -72,    -76,     22,     90,     37,    -65,    -81,     12,     89,     46,    -58,    -85,      2,     86,     55,    -50,    -88,     -8,     82,     62,    -41,    -89,    -18,     78,     69,    -32,    -90,    -27,     72,     75,    -23,    -90,    -36,     66,     80,    -13,    -89,    -45,     59,     84,     -3,    -86,    -54,     51,     87,      7,    -83,    -61,     42,     89,     16,    -78,    -68,     33,     90,     26,    -73,    -74,     24,     90,     35,    -67,    -79,     14,     89,     44,    -60,    -84,      4,     87,     53,    -52,    -87,     -6,     83,     60,    -43,    -89,    -15,     79,     67,    -34,    -90,    -25,     74 },
      {     86,     53,    -54,    -86,      1,     86,     52,    -55,    -85,      2,     87,     51,    -55,    -85,      3,     87,     50,    -56,    -85,      4,     87,     49,    -57,    -84,      6,     88,     48,    -58,    -84,      7,     88,     47,    -59,    -83,      8,     88,     46,    -60,    -83,      9,     88,     45,    -60,    -82,     10,     89,     44,    -61,    -82,     11,     89,     43,    -62,    -81,     12,     89,     42,    -63,    -81,     13,     89,     41,    -64,    -81,     14,     89,     40,    -64,    -80,     15,     89,     39,    -65,    -79,     16,     90,     38,    -66,    -79,     18,     90,     37,    -67,    -78,     19,     90,     36,    -67,    -78,     20,     90,     35,    -68,    -77,     21,     90,     34,    -69,    -77,     22,     90,     33,    -70,    -76,     23,     90,     32,    -70,    -76,     24,     90,     31,    -71,    -75,     25,     90,     30,    -72,    -74,     26,     90,     29,    -72,    -74,     27,     90,     28,    -73 },
      {     87,     49,    -59,    -82,     12,     89,     38,    -67,    -77,     24,     90,     27,    -75,    -70,     35,     90,     15,    -81,    -61,     46,     88,      3,    -86,    -52,     56,     84,     -9,    -89,    -41,     65,     78,    -21,    -90,    -30,     73,     72,    -32,    -90,    -19,     79,     64,    -43,    -88,     -7,     85,     55,    -54,    -85,      6,     88,     44,    -63,    -80,     18,     90,     33,    -71,    -74,     29,     90,     22,    -78,    -66,     40,     89,     10,    -83,    -57,     51,     86,     -2,    -87,    -47,     60,     81,    -14,    -90,    -36,     69,     76,    -26,    -90,    -25,     76,     68,    -37,    -89,    -13,     82,     60,    -48,    -87,     -1,     86,     50,    -58,    -83,     11,     89,     39,    -67,    -77,     23,     90,     28,    -74,    -70,     34,     90,     16,    -81,    -62,     45,     88,      4,    -85,    -53,     55,     84,     -8,    -89,    -42,     64,     79,    -20,    -90,    -31,     72 },
      {     87,     45,    -64,    -78,     23,     90,     24,    -78,    -64,     44,     88,      1,    -87,    -46,     63,     79,    -22,    -90,    -25,     77,     65,    -43,    -88,     -2,     87,     47,    -62,    -79,     21,     90,     26,    -77,    -66,     42,     88,      3,    -86,    -48,     61,     80,    -20,    -90,    -27,     76,     67,    -41,    -88,     -4,     86,     49,    -60,    -81,     19,     90,     28,    -76,    -67,     40,     89,      6,    -86,    -50,     60,     81,    -18,    -90,    -29,     75,     68,    -39,    -89,     -7,     85,     51,    -59,    -81,     16,     90,     30,    -74,    -69,     38,     89,      8,    -85,    -52,     58,     82,    -15,    -90,    -31,     74,     70,    -37,    -89,     -9,     85,     53,    -57,    -82,     14,     90,     32,    -73,    -70,     36,     89,     10,    -84,    -54,     56,     83,    -13,    -90,    -33,     72,     71,    -35,    -89,    -11,     84,     55,    -55,    -83,     12,     90,     34,    -72 },
      {     88,     41,    -68,    -74,     33,     89,      9,    -85,    -49,     62,     78,    -25,    -90,    -18,     82,     56,    -55,    -82,     16,     90,     26,    -78,    -63,     48,     86,     -8,    -89,    -34,     73,     69,    -40,    -88,     -1,     88,     42,    -67,    -74,     32,     90,     10,    -85,    -50,     61,     79,    -24,    -90,    -19,     81,     57,    -55,    -83,     15,     90,     27,    -77,    -64,     47,     86,     -7,    -89,    -35,     72,     70,    -39,    -88,     -2,     87,     43,    -67,    -75,     31,     90,     11,    -85,    -51,     60,     79,    -23,    -90,    -20,     81,     58,    -54,    -83,     14,     90,     28,    -77,    -64,     46,     86,     -6,    -89,    -36,     72,     70,    -38,    -89,     -3,     87,     44,    -66,    -76,     30,     90,     12,    -84,    -52,     60,     80,    -22,    -90,    -21,     81,     59,    -53,    -84,     13,     90,     29,    -76,    -65,     45,     87,     -4,    -89,    -37,     71 },
      {     88,     37,    -72,    -68,     43,     87,     -7,    -89,    -31,     76,     64,    -49,    -85,     13,     90,     25,    -79,    -59,     55,     82,    -20,    -90,    -19,     82,     54,    -60,    -79,     26,     90,     12,    -85,    -48,     64,     76,    -32,    -89,     -6,     87,     42,    -69,    -72,     38,     88,     -1,    -89,    -36,     73,     67,    -44,    -86,      8,     90,     30,    -77,    -63,     50,     84,    -14,    -90,    -24,     80,     58,    -55,    -81,     21,     90,     18,    -83,    -53,     60,     78,    -27,    -90,    -11,     85,     47,    -65,    -75,     33,     89,      4,    -87,    -41,     70,     71,    -39,    -88,      2,     89,     35,    -74,    -67,     45,     86,     -9,    -90,    -29,     77,     62,    -51,    -84,     15,     90,     23,    -81,    -57,     56,     81,    -22,    -90,    -16,     83,     52,    -61,    -78,     28,     90,     10,    -86,    -46,     66,     74,    -34,    -89,     -3,     88,     40,    -70 },
      {     89,     33,    -76,    -62,     53,     82,    -22,    -90,    -12,     86,     44,    -69,    -70,     42,     86,    -10,    -90,    -24,     81,     55,    -60,    -77,     31,     89,      2,    -88,    -35,     75,     64,    -51,    -83,     20,     90,     14,    -85,    -46,     67,     72,    -40,    -87,      8,     90,     26,    -80,    -56,     59,     78,    -29,    -89,     -4,     88,     37,    -74,    -65,     49,     84,    -18,    -90,    -16,     84,     48,    -66,    -73,     38,     88,     -6,    -90,    -28,     79,     58,    -57,    -79,     27,     90,      7,    -87,    -39,     72,     67,    -47,    -85,     15,     90,     19,    -83,    -50,     64,     74,    -36,    -88,      3,     89,     30,    -78,    -60,     55,     81,    -25,    -90,     -9,     87,     41,    -71,    -68,     45,     85,    -13,    -90,    -21,     82,     52,    -63,    -76,     34,     89,     -1,    -89,    -32,     77,     61,    -54,    -81,     23,     90,     11,    -86,    -43,     70 },
      {     89,     29,    -79,    -55,     61,     76,    -36,    -88,      8,     90,     22,    -83,    -49,     67,     71,    -43,    -85,     15,     90,     14,    -86,    -42,     72,     66,    -50,    -82,     23,     90,      7,    -88,    -35,     76,     60,    -56,    -79,     30,     89,     -1,    -89,    -28,     80,     55,    -62,    -75,     37,     87,     -9,    -90,    -21,     83,     48,    -67,    -70,     44,     85,    -16,    -90,    -13,     86,     41,    -72,    -65,     51,     82,    -24,    -90,     -6,     88,     34,    -77,    -60,     57,     78,    -31,    -89,      2,     89,     27,    -81,    -54,     63,     74,    -38,    -87,     10,     90,     20,    -84,    -47,     68,     70,    -45,    -85,     18,     90,     12,    -86,    -40,     73,     64,    -52,    -81,     25,     90,      4,    -88,    -33,     77,     59,    -58,    -78,     32,     89,     -3,    -90,    -26,     81,     53,    -64,    -74,     39,     87,    -11,    -90,    -19,     84,     46,    -69 },
      {     89,     25,    -82,    -48,     69,     67,    -50,    -81,     27,     89,     -2,    -90,    -23,     83,     46,    -70,    -66,     52,     81,    -29,    -89,      4,     90,     21,    -84,    -44,     72,     64,    -54,    -79,     31,     88,     -7,    -90,    -19,     85,     42,    -73,    -63,     55,     78,    -33,    -88,      9,     90,     16,    -86,    -40,     74,     61,    -57,    -77,     35,     87,    -11,    -90,    -14,     86,     38,    -76,    -60,     59,     76,    -37,    -87,     13,     90,     12,    -87,    -36,     77,     58,    -60,    -75,     39,     86,    -15,    -90,    -10,     88,     34,    -78,    -56,     62,     74,    -41,    -85,     18,     90,      8,    -88,    -32,     79,     55,    -64,    -72,     43,     85,    -20,    -90,     -6,     89,     30,    -80,    -53,     65,     71,    -45,    -84,     22,     90,      3,    -89,    -28,     81,     51,    -67,    -70,     47,     83,    -24,    -90,     -1,     89,     26,    -82,    -49,     68 },
      {     90,     21,    -85,    -40,     76,     58,    -62,    -72,     45,     83,    -26,    -89,      6,     90,     15,    -87,    -35,     78,     54,    -66,    -69,     50,     81,    -31,    -88,     11,     90,     10,    -88,    -30,     81,     49,    -70,    -65,     55,     78,    -36,    -86,     16,     90,      4,    -89,    -25,     83,     44,    -73,    -61,     59,     75,    -41,    -85,     22,     90,     -1,    -90,    -20,     85,     39,    -76,    -57,     63,     72,    -46,    -82,     27,     89,     -7,    -90,    -14,     87,     34,    -79,    -53,     67,     68,    -51,    -80,     32,     88,    -12,    -90,     -9,     88,     29,    -81,    -48,     70,     64,    -55,    -77,     37,     86,    -18,    -90,     -3,     89,     24,    -84,    -43,     74,     60,    -60,    -74,     42,     84,    -23,    -89,      2,     90,     19,    -86,    -38,     77,     56,    -64,    -71,     47,     82,    -28,    -89,      8,     90,     13,    -87,    -33,     79,     52,    -67 },
      {     90,     16,    -87,    -32,     81,     47,    -72,    -60,     61,     72,    -48,    -81,     33,     87,    -18,    -90,      1,     90,     15,    -87,    -31,     81,     46,    -73,    -60,     62,     71,    -49,    -80,     34,     86,    -19,    -90,      2,     90,     14,    -88,    -30,     82,     45,    -74,    -59,     63,     70,    -50,    -79,     35,     86,    -20,    -90,      3,     90,     13,    -88,    -29,     82,     44,    -74,    -58,     64,     70,    -51,    -79,     36,     86,    -21,    -89,      4,     90,     12,    -88,    -28,     83,     43,    -75,    -57,     64,     69,    -52,    -78,     37,     85,    -22,    -89,      6,     90,     11,    -88,    -27,     83,     42,    -76,    -56,     65,     68,    -53,    -78,     38,     85,    -23,    -89,      7,     90,     10,    -89,    -26,     84,     41,    -76,    -55,     66,     67,    -54,    -77,     39,     85,    -24,    -89,      8,     90,      9,    -89,    -25,     84,     40,    -77,    -55,     67 },
      {     90,     12,    -89,    -24,     85,     35,    -81,    -46,     74,     56,    -67,    -65,     58,     73,    -48,    -79,     37,     85,    -26,    -88,     14,     90,     -2,    -90,    -10,     89,     22,    -86,    -33,     81,     44,    -76,    -55,     68,     64,    -60,    -72,     50,     78,    -39,    -84,     28,     88,    -16,    -90,      4,     90,      8,    -89,    -20,     87,     31,    -82,    -42,     77,     53,    -70,    -62,     61,     70,    -52,    -77,     41,     83,    -30,    -87,     19,     89,     -7,    -90,     -6,     90,     18,    -87,    -29,     83,     40,    -78,    -51,     71,     60,    -63,    -69,     54,     76,    -43,    -82,     32,     86,    -21,    -89,      9,     90,      3,    -90,    -15,     88,     27,    -84,    -38,     79,     49,    -72,    -59,     64,     67,    -55,    -75,     45,     81,    -34,    -86,     23,     89,    -11,    -90,     -1,     90,     13,    -88,    -25,     85,     36,    -80,    -47,     74,     57,    -66 },
      {     90,      8,    -90,    -15,     88,     23,    -86,    -30,     84,     37,    -81,    -44,     77,     51,    -72,    -57,     67,     63,    -62,    -68,     56,     73,    -50,    -77,     43,     81,    -36,    -84,     29,     87,    -22,    -89,     14,     90,     -7,    -90,     -1,     90,      9,    -89,    -16,     88,     24,    -86,    -31,     83,     38,    -80,    -45,     76,     52,    -72,    -58,     67,     64,    -61,    -69,     55,     74,    -49,    -78,     42,     81,    -35,    -85,     28,     87,    -21,    -89,     13,     90,     -6,    -90,     -2,     90,     10,    -89,    -18,     88,     25,    -86,    -32,     83,     39,    -79,    -46,     76,     53,    -71,    -59,     66,     64,    -60,    -70,     55,     74,    -48,    -78,     41,     82,    -34,    -85,     27,     87,    -20,    -89,     12,     90,     -4,    -90,     -3,     90,     11,    -89,    -19,     88,     26,    -85,    -33,     82,     40,    -79,    -47,     75,     54,    -70,    -60,     65 },
      {     90,      3,    -90,     -7,     90,     10,    -90,    -13,     89,     16,    -89,    -20,     88,     23,    -87,    -26,     86,     29,    -85,    -32,     84,     35,    -82,    -38,     81,     41,    -79,    -44,     78,     47,    -76,    -50,     74,     53,    -72,    -55,     70,     58,    -68,    -60,     66,     63,    -64,    -65,     61,     67,    -59,    -70,     56,     72,    -54,    -74,     51,     76,    -48,    -77,     45,     79,    -42,    -81,     39,     82,    -36,    -83,     33,     85,    -30,    -86,     27,     87,    -24,    -88,     21,     88,    -18,    -89,     14,     89,    -11,    -90,      8,     90,     -4,    -90,      1,     90,      2,    -90,     -6,     90,      9,    -90,    -12,     89,     15,    -89,    -19,     88,     22,    -87,    -25,     86,     28,    -85,    -31,     84,     34,    -83,    -37,     81,     40,    -80,    -43,     78,     46,    -77,    -49,     75,     52,    -73,    -55,     71,     57,    -69,    -60,     67,     62,    -64 },
      {     90,     -1,    -90,      2,     90,     -3,    -90,      4,     90,     -6,    -90,      7,     90,     -8,    -90,      9,     90,    -10,    -90,     11,     90,    -12,    -89,     13,     89,    -14,    -89,     15,     89,    -16,    -89,     18,     89,    -19,    -88,     20,     88,    -21,    -88,     22,     88,    -23,    -87,     24,     87,    -25,    -87,     26,     86,    -27,    -86,     28,     86,    -29,    -85,     30,     85,    -31,    -85,     32,     84,    -33,    -84,     34,     83,    -35,    -83,     36,     82,    -37,    -82,     38,     81,    -39,    -81,     40,     81,    -41,    -80,     42,     79,    -43,    -79,     44,     78,    -45,    -78,     46,     77,    -47,    -77,     48,     76,    -49,    -76,     50,     75,    -51,    -74,     52,     74,    -53,    -73,     54,     72,    -55,    -72,     55,     71,    -56,    -70,     57,     70,    -58,    -69,     59,     68,    -60,    -67,     60,     67,    -61,    -66,     62,     65,    -63,    -64,     64 },
      {     90,     -6,    -90,     11,     89,    -16,    -88,     22,     87,    -27,    -85,     32,     83,    -37,    -81,     42,     78,    -47,    -76,     52,     72,    -56,    -69,     60,     65,    -64,    -61,     68,     57,    -72,    -53,     75,     48,    -78,    -43,     81,     38,    -83,    -33,     85,     28,    -87,    -23,     88,     18,    -89,    -12,     90,      7,    -90,     -1,     90,     -4,    -90,     10,     89,    -15,    -89,     21,     87,    -26,    -86,     31,     84,    -36,    -81,     41,     79,    -46,    -76,     51,     73,    -55,    -70,     60,     66,    -64,    -62,     67,     58,    -71,    -54,     74,     49,    -77,    -44,     80,     39,    -82,    -34,     85,     29,    -86,    -24,     88,     19,    -89,    -13,     90,      8,    -90,     -2,     90,     -3,    -90,      9,     90,    -14,    -89,     20,     88,    -25,    -86,     30,     84,    -35,    -82,     40,     79,    -45,    -77,     50,     74,    -55,    -70,     59,     67,    -63 },
      {     90,    -10,    -89,     20,     87,    -29,    -84,     38,     79,    -47,    -74,     55,     68,    -63,    -61,     70,     54,    -76,    -45,     81,     36,    -85,    -27,     88,     18,    -89,     -8,     90,     -2,    -90,     12,     89,    -22,    -86,     31,     83,    -40,    -78,     49,     73,    -57,    -67,     64,     60,    -71,    -52,     77,     43,    -81,    -34,     85,     25,    -88,    -15,     90,      6,    -90,      4,     90,    -14,    -88,     24,     86,    -33,    -82,     42,     77,    -51,    -72,     59,     65,    -66,    -58,     72,     50,    -78,    -41,     82,     32,    -86,    -23,     89,     13,    -90,     -3,     90,     -7,    -90,     16,     88,    -26,    -85,     35,     81,    -44,    -76,     53,     70,    -60,    -64,     67,     56,    -74,    -48,     79,     39,    -83,    -30,     87,     21,    -89,    -11,     90,      1,    -90,      9,     89,    -19,    -87,     28,     84,    -37,    -80,     46,     75,    -55,    -69,     62 },
      {     90,    -14,    -88,     28,     83,    -41,    -77,     54,     68,    -64,    -58,     74,     46,    -81,    -33,     86,     20,    -89,     -6,     90,     -9,    -89,     23,     85,    -36,    -79,     49,     72,    -60,    -62,     70,     51,    -78,    -38,     85,     25,    -89,    -11,     90,     -3,    -90,     18,     87,    -31,    -82,     44,     75,    -56,    -66,     67,     55,    -76,    -43,     82,     30,    -87,    -16,     90,      2,    -90,     12,     88,    -26,    -84,     39,     78,    -52,    -70,     63,     60,    -72,    -48,     80,     35,    -86,    -22,     89,      8,    -90,      7,     89,    -21,    -86,     34,     81,    -47,    -73,     59,     64,    -69,    -53,     77,     40,    -84,    -27,     88,     13,    -90,      1,     90,    -15,    -88,     29,     83,    -42,    -76,     55,     67,    -65,    -57,     74,     45,    -81,    -32,     87,     19,    -90,     -4,     90,    -10,    -89,     24,     85,    -37,    -79,     50,     71,    -61 },
      {     90,    -19,    -86,     36,     78,    -53,    -67,     67,     54,    -78,    -37,     86,     20,    -90,     -1,     90,    -18,    -86,     35,     79,    -52,    -68,     66,     55,    -77,    -38,     85,     21,    -90,     -2,     90,    -16,    -87,     34,     79,    -51,    -69,     65,     55,    -77,    -39,     85,     22,    -89,     -3,     90,    -15,    -87,     33,     80,    -50,    -70,     64,     56,    -76,    -40,     85,     23,    -89,     -4,     90,    -14,    -87,     32,     81,    -49,    -70,     64,     57,    -76,    -41,     84,     24,    -89,     -6,     90,    -13,    -88,     31,     81,    -48,    -71,     63,     58,    -75,    -42,     84,     25,    -89,     -7,     90,    -12,    -88,     30,     81,    -47,    -72,     62,     59,    -74,    -43,     83,     26,    -89,     -8,     90,    -11,    -88,     29,     82,    -46,    -72,     61,     60,    -74,    -44,     83,     27,    -89,     -9,     90,    -10,    -88,     28,     82,    -45,    -73,     60 },
      {     90,    -23,    -84,     44,     72,    -63,    -56,     77,     36,    -87,    -14,     90,     -9,    -88,     31,     80,    -52,    -67,     69,     49,    -81,    -28,     89,      6,    -90,     18,     86,    -39,    -76,     59,     60,    -74,    -41,     85,     20,    -90,      3,     89,    -26,    -82,     47,     70,    -65,    -54,     79,     33,    -88,    -11,     90,    -12,    -87,     34,     78,    -55,    -64,     71,     46,    -83,    -25,     89,      2,    -90,     21,     85,    -42,    -74,     61,     58,    -76,    -38,     86,     16,    -90,      7,     89,    -29,    -81,     50,     68,    -67,    -51,     81,     30,    -88,     -8,     90,    -15,    -86,     37,     77,    -57,    -62,     73,     43,    -84,    -22,     90,     -1,    -89,     24,     83,    -45,    -72,     64,     55,    -78,    -35,     87,     13,    -90,     10,     88,    -32,    -79,     53,     66,    -70,    -48,     82,     27,    -89,     -4,     90,    -19,    -85,     40,     75,    -60 },
      {     89,    -27,    -81,     52,     65,    -72,    -43,     85,     18,    -90,     10,     87,    -36,    -76,     60,     58,    -77,    -34,     88,      8,    -90,     20,     84,    -45,    -70,     67,     50,    -82,    -25,     90,     -2,    -89,     29,     80,    -54,    -64,     73,     41,    -86,    -15,     90,    -12,    -87,     38,     75,    -61,    -56,     78,     32,    -88,     -6,     90,    -22,    -83,     47,     69,    -68,    -48,     83,     23,    -90,      4,     89,    -31,    -79,     55,     62,    -74,    -39,     86,     13,    -90,     14,     86,    -40,    -74,     63,     55,    -79,    -30,     89,      3,    -90,     24,     82,    -49,    -67,     70,     46,    -84,    -21,     90,     -7,    -88,     33,     78,    -57,    -60,     76,     37,    -87,    -11,     90,    -16,    -85,     42,     72,    -64,    -53,     81,     28,    -89,     -1,     89,    -26,    -81,     51,     66,    -71,    -44,     85,     19,    -90,      9,     88,    -35,    -77,     59 },
      {     89,    -31,    -78,     59,     57,    -79,    -29,     89,     -2,    -89,     33,     77,    -60,    -55,     80,     27,    -90,      4,     88,    -35,    -76,     62,     54,    -81,    -25,     90,     -7,    -88,     37,     74,    -64,    -52,     82,     23,    -90,      9,     87,    -39,    -73,     65,     50,    -83,    -21,     90,    -11,    -86,     41,     72,    -67,    -48,     84,     19,    -90,     13,     86,    -43,    -70,     68,     46,    -85,    -16,     90,    -15,    -85,     45,     69,    -70,    -44,     85,     14,    -90,     18,     84,    -47,    -67,     71,     42,    -86,    -12,     90,    -20,    -83,     49,     66,    -72,    -40,     87,     10,    -90,     22,     82,    -51,    -64,     74,     38,    -87,     -8,     90,    -24,    -81,     53,     63,    -75,    -36,     88,      6,    -90,     26,     81,    -55,    -61,     76,     34,    -88,     -3,     89,    -28,    -79,     56,     60,    -77,    -32,     89,      1,    -89,     30,     78,    -58 },
      {     89,    -35,    -74,     65,     48,    -85,    -14,     90,    -22,    -81,     55,     60,    -78,    -28,     90,     -8,    -87,     42,     70,    -70,    -41,     87,      7,    -90,     29,     78,    -60,    -54,     82,     21,    -90,     15,     84,    -49,    -64,     75,     34,    -89,      1,     88,    -36,    -74,     66,     47,    -85,    -13,     90,    -23,    -81,     55,     59,    -79,    -27,     90,     -9,    -86,     43,     69,    -71,    -40,     87,      6,    -89,     30,     77,    -61,    -53,     82,     20,    -90,     16,     84,    -50,    -64,     76,     33,    -89,      2,     88,    -37,    -73,     67,     46,    -85,    -12,     90,    -24,    -81,     56,     58,    -79,    -26,     90,    -10,    -86,     44,     68,    -72,    -39,     88,      4,    -89,     31,     77,    -62,    -52,     83,     19,    -90,     18,     83,    -51,    -63,     76,     32,    -89,      3,     88,    -38,    -72,     67,     45,    -86,    -11,     90,    -25,    -80,     57 },
      {     88,    -39,    -70,     71,     38,    -88,      1,     88,    -40,    -70,     72,     37,    -89,      2,     88,    -41,    -69,     72,     36,    -89,      3,     87,    -42,    -68,     73,     35,    -89,      4,     87,    -43,    -67,     74,     34,    -89,      6,     87,    -44,    -67,     74,     33,    -89,      7,     86,    -45,    -66,     75,     32,    -89,      8,     86,    -46,    -65,     76,     31,    -90,      9,     86,    -47,    -64,     76,     30,    -90,     10,     85,    -48,    -64,     77,     29,    -90,     11,     85,    -49,    -63,     77,     28,    -90,     12,     85,    -50,    -62,     78,     27,    -90,     13,     84,    -51,    -61,     78,     26,    -90,     14,     84,    -52,    -60,     79,     25,    -90,     15,     83,    -53,    -60,     79,     24,    -90,     16,     83,    -54,    -59,     80,     23,    -90,     18,     82,    -55,    -58,     81,     22,    -90,     19,     82,    -55,    -57,     81,     21,    -90,     20,     81,    -56 },
      {     88,    -43,    -66,     76,     28,    -90,     16,     82,    -57,    -54,     84,     12,    -90,     32,     74,    -69,    -39,     89,     -4,    -86,     47,     63,    -78,    -24,     90,    -21,    -80,     60,     50,    -85,     -8,     89,    -36,    -71,     72,     35,    -89,      9,     85,    -51,    -60,     81,     20,    -90,     25,     78,    -64,    -46,     87,      3,    -88,     40,     68,    -74,    -31,     90,    -13,    -83,     55,     56,    -82,    -15,     90,    -29,    -76,     67,     42,    -88,      1,     87,    -44,    -65,     77,     27,    -90,     18,     81,    -58,    -53,     84,     11,    -90,     33,     73,    -70,    -38,     89,     -6,    -86,     48,     62,    -79,    -23,     90,    -22,    -79,     61,     49,    -86,     -7,     89,    -37,    -70,     72,     34,    -89,     10,     85,    -52,    -59,     81,     19,    -90,     26,     77,    -64,    -45,     87,      2,    -88,     41,     67,    -75,    -30,     90,    -14,    -83,     55 },
      {     87,    -47,    -61,     81,     18,    -90,     31,     73,    -71,    -34,     90,    -14,    -82,     59,     50,    -86,     -3,     88,    -44,    -64,     79,     21,    -90,     28,     75,    -69,    -37,     89,    -11,    -83,     56,     53,    -85,     -7,     89,    -41,    -66,     77,     24,    -90,     25,     77,    -67,    -40,     89,     -8,    -85,     54,     55,    -84,    -10,     89,    -38,    -68,     76,     27,    -90,     22,     78,    -64,    -43,     88,     -4,    -86,     51,     58,    -82,    -13,     90,    -35,    -70,     74,     30,    -90,     19,     80,    -62,    -46,     87,     -1,    -87,     48,     60,    -81,    -16,     90,    -32,    -72,     72,     33,    -90,     15,     81,    -60,    -49,     86,      2,    -88,     45,     63,    -79,    -20,     90,    -29,    -74,     70,     36,    -89,     12,     83,    -57,    -52,     85,      6,    -88,     42,     65,    -78,    -23,     90,    -26,    -76,     67,     39,    -89,      9,     84,    -55 },
      {     86,    -51,    -56,     84,      7,    -88,     45,     61,    -81,    -13,     89,    -39,    -66,     78,     20,    -90,     33,     70,    -75,    -26,     90,    -27,    -74,     71,     32,    -90,     21,     78,    -67,    -38,     89,    -14,    -81,     62,     44,    -88,      8,     84,    -57,    -50,     87,     -1,    -86,     52,     55,    -85,     -6,     88,    -46,    -60,     82,     12,    -89,     40,     65,    -79,    -19,     90,    -34,    -70,     76,     25,    -90,     28,     74,    -72,    -31,     90,    -22,    -77,     67,     37,    -90,     15,     81,    -63,    -43,     89,     -9,    -83,     58,     49,    -87,      2,     86,    -53,    -55,     85,      4,    -88,     47,     60,    -82,    -11,     89,    -41,    -64,     79,     18,    -90,     35,     69,    -76,    -24,     90,    -29,    -73,     72,     30,    -90,     23,     77,    -68,    -36,     90,    -16,    -80,     64,     42,    -89,     10,     83,    -59,    -48,     87,     -3,    -85,     54 },
      {     86,    -55,    -51,     87,     -4,    -84,     58,     47,    -88,      9,     82,    -61,    -43,     89,    -13,    -81,     64,     39,    -90,     18,     78,    -67,    -35,     90,    -22,    -76,     70,     31,    -90,     26,     74,    -73,    -27,     90,    -30,    -71,     76,     23,    -90,     34,     68,    -78,    -19,     90,    -38,    -65,     80,     14,    -89,     42,     62,    -82,    -10,     88,    -46,    -59,     84,      6,    -87,     50,     55,    -85,     -1,     86,    -54,    -52,     87,     -3,    -85,     57,     48,    -88,      8,     83,    -60,    -44,     89,    -12,    -81,     64,     40,    -89,     16,     79,    -67,    -36,     90,    -21,    -77,     70,     32,    -90,     25,     74,    -72,    -28,     90,    -29,    -72,     75,     24,    -90,     33,     69,    -77,    -20,     90,    -37,    -66,     79,     15,    -89,     41,     63,    -81,    -11,     89,    -45,    -60,     83,      7,    -88,     49,     56,    -85,     -2,     86,    -53 },
      {     85,    -58,    -45,     89,    -15,    -78,     69,     31,    -90,     30,     70,    -78,    -16,     89,    -44,    -59,     85,      1,    -85,     57,     46,    -89,     14,     79,    -68,    -32,     90,    -29,    -70,     77,     18,    -89,     43,     60,    -84,     -2,     86,    -56,    -47,     89,    -13,    -79,     67,     33,    -90,     28,     71,    -77,    -19,     89,    -42,    -60,     84,      3,    -86,     55,     48,    -88,     12,     80,    -67,    -34,     90,    -27,    -72,     76,     20,    -90,     41,     61,    -83,     -4,     86,    -55,    -49,     88,    -11,    -81,     66,     35,    -90,     26,     72,    -76,    -21,     90,    -40,    -62,     83,      6,    -87,     54,     50,    -88,     10,     81,    -65,    -36,     90,    -25,    -73,     75,     22,    -90,     39,     63,    -82,     -7,     87,    -53,    -51,     88,     -9,    -81,     64,     37,    -90,     24,     74,    -74,    -23,     90,    -38,    -64,     82,      8,    -87,     52 },
      {     84,    -61,    -39,     90,    -26,    -71,     78,     14,    -88,     50,     52,    -88,     12,     79,    -70,    -28,     90,    -37,    -63,     83,      2,    -85,     60,     41,    -90,     24,     72,    -77,    -16,     89,    -48,    -54,     87,    -10,    -80,     68,     30,    -90,     35,     64,    -82,     -4,     86,    -58,    -43,     90,    -22,    -74,     76,     19,    -89,     46,     55,    -87,      8,     81,    -67,    -32,     90,    -33,    -66,     81,      7,    -86,     56,     45,    -89,     20,     75,    -74,    -21,     89,    -44,    -57,     86,     -6,    -82,     65,     34,    -90,     31,     67,    -81,     -9,     87,    -55,    -47,     89,    -18,    -76,     73,     23,    -90,     42,     59,    -85,      3,     83,    -64,    -36,     90,    -29,    -69,     79,     11,    -88,     53,     49,    -89,     15,     77,    -72,    -25,     90,    -40,    -60,     85,     -1,    -84,     62,     38,    -90,     27,     70,    -78,    -13,     88,    -51 },
      {     83,    -64,    -33,     90,    -36,    -62,     85,     -3,    -82,     67,     30,    -90,     39,     60,    -86,      7,     81,    -69,    -27,     90,    -42,    -57,     87,    -10,    -79,     71,     24,    -90,     45,     55,    -88,     13,     77,    -73,    -21,     89,    -48,    -52,     88,    -16,    -76,     75,     18,    -89,     51,     49,    -89,     20,     74,    -77,    -14,     88,    -54,    -46,     89,    -23,    -72,     78,     11,    -87,     56,     43,    -90,     26,     70,    -80,     -8,     86,    -59,    -40,     90,    -29,    -67,     81,      4,    -85,     61,     37,    -90,     32,     65,    -83,     -1,     84,    -64,    -34,     90,    -35,    -63,     84,     -2,    -82,     66,     31,    -90,     38,     60,    -85,      6,     81,    -68,    -28,     90,    -41,    -58,     86,     -9,    -79,     70,     25,    -90,     44,     55,    -87,     12,     78,    -72,    -22,     89,    -47,    -53,     88,    -15,    -76,     74,     19,    -89,     50 },
      {     82,    -67,    -27,     90,    -46,    -52,     89,    -21,    -72,     79,      7,    -85,     63,     33,    -90,     40,     57,    -87,     14,     76,    -76,    -13,     87,    -58,    -39,     90,    -34,    -62,     85,     -8,    -79,     72,     20,    -89,     53,     45,    -90,     28,     67,    -83,      1,     82,    -68,    -26,     90,    -47,    -51,     89,    -22,    -71,     80,      6,    -85,     64,     32,    -90,     41,     56,    -88,     15,     75,    -77,    -12,     87,    -59,    -38,     90,    -35,    -61,     86,     -9,    -78,     73,     19,    -88,     54,     44,    -90,     29,     66,    -83,      2,     81,    -69,    -25,     89,    -48,    -50,     89,    -23,    -70,     81,      4,    -84,     64,     31,    -90,     42,     55,    -88,     16,     74,    -77,    -11,     86,    -60,    -37,     90,    -36,    -60,     86,    -10,    -78,     74,     18,    -88,     55,     43,    -90,     30,     65,    -84,      3,     81,    -70,    -24,     89,    -49 },
      {     81,    -70,    -21,     88,    -55,    -40,     90,    -37,    -58,     88,    -18,    -72,     80,      3,    -83,     68,     24,    -89,     53,     43,    -90,     34,     60,    -87,     14,     74,    -78,     -7,     84,    -66,    -27,     89,    -50,    -46,     90,    -31,    -63,     86,    -11,    -76,     77,     10,    -85,     64,     30,    -90,     47,     49,    -90,     28,     65,    -85,      8,     78,    -75,    -13,     86,    -61,    -33,     90,    -44,    -52,     89,    -25,    -67,     83,     -4,    -79,     73,     16,    -87,     59,     36,    -90,     41,     55,    -89,     22,     70,    -82,      1,     81,    -71,    -20,     88,    -56,    -39,     90,    -38,    -57,     88,    -19,    -72,     81,      2,    -82,     69,     23,    -89,     54,     42,    -90,     35,     60,    -87,     15,     74,    -79,     -6,     84,    -67,    -26,     89,    -51,    -45,     90,    -32,    -62,     86,    -12,    -76,     77,      9,    -85,     64,     29,    -90,     48 },
      {     81,    -73,    -14,     86,    -64,    -28,     89,    -53,    -41,     90,    -40,    -54,     89,    -27,    -64,     86,    -13,    -74,     80,      1,    -81,     72,     15,    -86,     63,     29,    -89,     52,     42,    -90,     39,     55,    -89,     26,     65,    -85,     12,     74,    -79,     -2,     81,    -72,    -16,     87,    -62,    -30,     90,    -51,    -43,     90,    -38,    -55,     89,    -25,    -66,     85,    -11,    -75,     79,      3,    -82,     71,     18,    -87,     61,     31,    -90,     50,     44,    -90,     37,     56,    -89,     24,     67,    -85,     10,     76,    -78,     -4,     82,    -70,    -19,     87,    -60,    -32,     90,    -49,    -45,     90,    -36,    -57,     88,    -23,    -67,     84,     -9,    -76,     78,      6,    -83,     70,     20,    -88,     60,     33,    -90,     48,     46,    -90,     35,     58,    -88,     22,     68,    -84,      8,     77,    -77,     -7,     83,    -69,    -21,     88,    -59,    -34,     90,    -47 },
      {     79,    -76,     -8,     83,    -71,    -15,     86,    -66,    -23,     88,    -60,    -30,     89,    -55,    -37,     90,    -48,    -44,     90,    -41,    -51,     90,    -34,    -57,     89,    -27,    -63,     87,    -20,    -68,     85,    -12,    -73,     81,     -4,    -77,     78,      3,    -81,     74,     11,    -84,     69,     19,    -87,     64,     26,    -89,     58,     33,    -90,     52,     40,    -90,     45,     47,    -90,     38,     54,    -89,     31,     60,    -88,     24,     65,    -86,     16,     70,    -83,      9,     75,    -80,      1,     79,    -76,     -7,     82,    -72,    -14,     85,    -67,    -22,     88,    -61,    -29,     89,    -55,    -36,     90,    -49,    -43,     90,    -42,    -50,     90,    -35,    -56,     89,    -28,    -62,     87,    -21,    -67,     85,    -13,    -72,     82,     -6,    -77,     78,      2,    -81,     74,     10,    -84,     70,     18,    -86,     64,     25,    -88,     59,     32,    -90,     53,     39,    -90,     46 },
      {     78,    -78,     -1,     79,    -77,     -2,     79,    -77,     -3,     80,    -76,     -4,     81,    -76,     -6,     81,    -75,     -7,     81,    -74,     -8,     82,    -74,     -9,     82,    -73,    -10,     83,    -72,    -11,     83,    -72,    -12,     84,    -71,    -13,     84,    -70,    -14,     85,    -70,    -15,     85,    -69,    -16,     85,    -68,    -18,     86,    -67,    -19,     86,    -67,    -20,     86,    -66,    -21,     87,    -65,    -22,     87,    -64,    -23,     87,    -64,    -24,     88,    -63,    -25,     88,    -62,    -26,     88,    -61,    -27,     88,    -60,    -28,     89,    -60,    -29,     89,    -59,    -30,     89,    -58,    -31,     89,    -57,    -32,     89,    -56,    -33,     89,    -55,    -34,     90,    -55,    -35,     90,    -54,    -36,     90,    -53,    -37,     90,    -52,    -38,     90,    -51,    -39,     90,    -50,    -40,     90,    -49,    -41,     90,    -48,    -42,     90,    -47,    -43,     90,    -46,    -44,     90,    -45 },
      {     77,    -80,      6,     74,    -82,     11,     71,    -85,     16,     67,    -86,     22,     64,    -88,     27,     60,    -89,     32,     55,    -90,     37,     51,    -90,     42,     46,    -90,     47,     41,    -90,     52,     36,    -90,     56,     31,    -89,     60,     26,    -88,     64,     21,    -86,     68,     15,    -84,     72,     10,    -82,     75,      4,    -79,     78,     -1,    -77,     81,     -7,    -74,     83,    -12,    -70,     85,    -18,    -67,     87,    -23,    -63,     88,    -28,    -59,     89,    -33,    -55,     90,    -38,    -50,     90,    -43,    -45,     90,    -48,    -40,     90,    -53,    -35,     89,    -57,    -30,     89,    -61,    -25,     87,    -65,    -20,     86,    -69,    -14,     84,    -72,     -9,     81,    -76,     -3,     79,    -78,      2,     76,    -81,      8,     73,    -83,     13,     70,    -85,     19,     66,    -87,     24,     62,    -88,     29,     58,    -89,     34,     54,    -90,     39,     49,    -90,     44 },
      {     76,    -82,     12,     69,    -86,     24,     60,    -89,     35,     51,    -90,     46,     40,    -90,     56,     29,    -88,     65,     18,    -84,     73,      6,    -79,     79,     -7,    -72,     85,    -19,    -64,     88,    -30,    -55,     90,    -41,    -45,     90,    -52,    -34,     89,    -61,    -23,     86,    -70,    -11,     81,    -77,      1,     76,    -82,     13,     68,    -87,     25,     60,    -89,     36,     50,    -90,     47,     39,    -90,     57,     28,    -88,     66,     16,    -84,     74,      4,    -78,     80,     -8,    -72,     85,    -20,    -64,     88,    -31,    -55,     90,    -42,    -44,     90,    -53,    -33,     89,    -62,    -22,     86,    -70,    -10,     81,    -77,      2,     75,    -83,     14,     67,    -87,     26,     59,    -89,     37,     49,    -90,     48,     38,    -90,     58,     27,    -87,     67,     15,    -83,     74,      3,    -78,     81,     -9,    -71,     85,    -21,    -63,     89,    -32,    -54,     90,    -43 },
      {     75,    -84,     19,     63,    -89,     36,     48,    -90,     53,     31,    -88,     67,     13,    -81,     78,     -6,    -72,     86,    -24,    -59,     90,    -41,    -43,     90,    -57,    -26,     86,    -70,     -8,     79,    -81,     11,     68,    -87,     29,     55,    -90,     46,     38,    -89,     61,     21,    -85,     74,      2,    -76,     83,    -16,    -64,     89,    -34,    -50,     90,    -51,    -33,     88,    -65,    -15,     82,    -77,      3,     73,    -85,     22,     60,    -89,     39,     45,    -90,     55,     28,    -87,     69,     10,    -80,     79,     -9,    -70,     87,    -27,    -56,     90,    -44,    -40,     90,    -60,    -23,     85,    -72,     -4,     77,    -82,     14,     66,    -88,     32,     52,    -90,     49,     35,    -89,     64,     18,    -83,     76,     -1,    -74,     84,    -20,    -62,     89,    -37,    -47,     90,    -54,    -30,     88,    -67,    -12,     81,    -78,      7,     71,    -86,     25,     58,    -90,     42 },
      {     74,    -85,     25,     56,    -90,     48,     34,    -88,     67,     10,    -79,     81,    -15,    -64,     89,    -39,    -43,     90,    -60,    -20,     83,    -77,      6,     70,    -87,     30,     52,    -90,     53,     29,    -87,     71,      4,    -76,     84,    -21,    -60,     90,    -44,    -38,     89,    -64,    -14,     81,    -79,     11,     67,    -88,     35,     47,    -90,     57,     24,    -85,     74,     -1,    -73,     86,    -26,    -55,     90,    -49,    -33,     88,    -68,     -9,     78,    -82,     16,     63,    -89,     40,     42,    -90,     61,     19,    -83,     77,     -7,    -70,     87,    -31,    -51,     90,    -54,    -28,     86,    -72,     -3,     76,    -84,     22,     59,    -90,     45,     37,    -89,     65,     13,    -81,     80,    -12,    -66,     89,    -36,    -46,     90,    -58,    -23,     85,    -75,      2,     72,    -86,     27,     55,    -90,     50,     32,    -88,     69,      8,    -78,     82,    -18,    -62,     89,    -41 },
      {     72,    -87,     31,     49,    -90,     59,     20,    -82,     79,    -12,    -64,     89,    -42,    -38,     89,    -67,     -8,     77,    -84,     24,     55,    -90,     53,     27,    -85,     75,     -4,    -70,     88,    -35,    -45,     90,    -62,    -15,     81,    -81,     16,     61,    -90,     46,     34,    -88,     70,      3,    -74,     86,    -28,    -52,     90,    -56,    -23,     84,    -77,      9,     67,    -89,     39,     41,    -89,     65,     11,    -78,     83,    -21,    -58,     90,    -50,    -30,     86,    -73,      1,     72,    -87,     32,     48,    -90,     60,     19,    -82,     79,    -13,    -64,     89,    -43,    -37,     88,    -68,     -7,     76,    -85,     25,     55,    -90,     54,     26,    -85,     76,     -6,    -69,     88,    -36,    -44,     90,    -63,    -14,     80,    -81,     18,     60,    -90,     47,     33,    -87,     71,      2,    -74,     86,    -29,    -51,     90,    -57,    -22,     83,    -78,     10,     66,    -89,     40 },
      {     71,    -88,     37,     41,    -89,     68,      4,    -74,     87,    -33,    -45,     89,    -65,     -9,     76,    -85,     29,     49,    -90,     62,     13,    -78,     84,    -25,    -53,     90,    -59,    -18,     81,    -82,     21,     56,    -90,     55,     22,    -82,     80,    -16,    -60,     90,    -52,    -26,     84,    -78,     12,     63,    -90,     48,     30,    -86,     76,     -8,    -66,     89,    -44,    -34,     87,    -73,      3,     69,    -89,     40,     38,    -88,     70,      1,    -72,     88,    -36,    -42,     89,    -67,     -6,     74,    -86,     32,     46,    -90,     64,     10,    -77,     85,    -28,    -50,     90,    -61,    -14,     79,    -83,     24,     54,    -90,     58,     19,    -81,     81,    -20,    -57,     90,    -55,    -23,     83,    -79,     15,     60,    -90,     51,     27,    -85,     77,    -11,    -64,     90,    -47,    -31,     86,    -75,      7,     67,    -89,     43,     35,    -87,     72,     -2,    -70,     88,    -39 },
      {     70,    -89,     43,     33,    -86,     76,    -11,    -62,     90,    -53,    -23,     82,    -81,     22,     54,    -90,     61,     12,    -77,     86,    -32,    -44,     89,    -69,     -1,     70,    -89,     42,     34,    -86,     76,    -10,    -63,     90,    -52,    -24,     82,    -81,     21,     55,    -90,     60,     13,    -77,     85,    -31,    -45,     89,    -68,     -2,     71,    -88,     41,     35,    -87,     75,     -9,    -64,     90,    -51,    -25,     83,    -81,     20,     55,    -90,     60,     14,    -78,     85,    -30,    -46,     89,    -67,     -3,     72,    -88,     40,     36,    -87,     74,     -8,    -64,     90,    -50,    -26,     83,    -80,     19,     56,    -90,     59,     15,    -78,     85,    -29,    -47,     89,    -67,     -4,     72,    -88,     39,     37,    -87,     74,     -7,    -65,     90,    -49,    -27,     84,    -79,     18,     57,    -90,     58,     16,    -79,     84,    -28,    -48,     90,    -66,     -6,     73,    -88,     38 },
      {     68,    -89,     49,     25,    -82,     82,    -26,    -48,     89,    -69,      1,     67,    -90,     50,     24,    -81,     83,    -27,    -47,     89,    -70,      2,     67,    -90,     51,     23,    -81,     83,    -28,    -46,     89,    -70,      3,     66,    -90,     52,     22,    -81,     84,    -29,    -45,     89,    -71,      4,     65,    -90,     53,     21,    -80,     84,    -30,    -44,     89,    -72,      6,     64,    -90,     54,     20,    -79,     85,    -31,    -43,     88,    -72,      7,     64,    -90,     55,     19,    -79,     85,    -32,    -42,     88,    -73,      8,     63,    -90,     55,     18,    -78,     85,    -33,    -41,     88,    -74,      9,     62,    -90,     56,     16,    -78,     86,    -34,    -40,     88,    -74,     10,     61,    -90,     57,     15,    -77,     86,    -35,    -39,     87,    -75,     11,     60,    -90,     58,     14,    -77,     86,    -36,    -38,     87,    -76,     12,     60,    -90,     59,     13,    -76,     87,    -37 },
      {     67,    -90,     55,     16,    -77,     87,    -40,    -32,     84,    -81,     25,     47,    -89,     72,     -9,    -60,     90,    -61,     -8,     72,    -89,     48,     24,    -81,     85,    -33,    -39,     87,    -77,     18,     54,    -90,     67,     -1,    -66,     90,    -55,    -15,     76,    -87,     41,     31,    -84,     81,    -26,    -46,     89,    -73,     10,     60,    -90,     62,      7,    -71,     89,    -49,    -23,     80,    -85,     34,     38,    -86,     78,    -19,    -53,     90,    -68,      2,     65,    -90,     56,     14,    -76,     88,    -42,    -30,     83,    -82,     27,     45,    -88,     74,    -11,    -59,     90,    -63,     -6,     70,    -89,     50,     22,    -79,     85,    -35,    -37,     86,    -78,     20,     52,    -90,     69,     -3,    -64,     90,    -57,    -13,     75,    -88,     43,     29,    -83,     82,    -28,    -44,     88,    -74,     12,     58,    -90,     64,      4,    -70,     89,    -51,    -21,     79,    -86,     36 },
      {     65,    -90,     60,      8,    -70,     90,    -54,    -15,     75,    -88,     47,     23,    -79,     86,    -40,    -30,     82,    -84,     33,     37,    -85,     81,    -26,    -44,     88,    -77,     19,     51,    -89,     72,    -11,    -57,     90,    -67,      3,     63,    -90,     62,      4,    -68,     90,    -56,    -12,     73,    -89,     50,     20,    -77,     87,    -43,    -27,     81,    -85,     36,     34,    -84,     82,    -29,    -41,     87,    -78,     22,     48,    -89,     74,    -14,    -55,     90,    -70,      7,     60,    -90,     64,      1,    -66,     90,    -59,     -9,     71,    -89,     53,     16,    -76,     88,    -46,    -24,     79,    -86,     39,     31,    -83,     83,    -32,    -38,     86,    -80,     25,     45,    -88,     76,    -18,    -52,     89,    -72,     10,     58,    -90,     67,     -2,    -64,     90,    -61,     -6,     69,    -90,     55,     13,    -74,     89,    -49,    -21,     78,    -87,     42,     28,    -81,     85,    -35 },
      {     64,    -90,     64,     -1,    -63,     90,    -65,      2,     62,    -90,     66,     -3,    -61,     90,    -67,      4,     60,    -90,     67,     -6,    -60,     90,    -68,      7,     59,    -90,     69,     -8,    -58,     90,    -70,      9,     57,    -90,     70,    -10,    -56,     90,    -71,     11,     55,    -90,     72,    -12,    -55,     89,    -72,     13,     54,    -89,     73,    -14,    -53,     89,    -74,     15,     52,    -89,     74,    -16,    -51,     89,    -75,     18,     50,    -89,     76,    -19,    -49,     88,    -76,     20,     48,    -88,     77,    -21,    -47,     88,    -77,     22,     46,    -88,     78,    -23,    -45,     87,    -78,     24,     44,    -87,     79,    -25,    -43,     87,    -79,     26,     42,    -86,     80,    -27,    -41,     86,    -81,     28,     40,    -86,     81,    -29,    -39,     85,    -81,     30,     38,    -85,     82,    -31,    -37,     85,    -82,     32,     36,    -84,     83,    -33,    -35,     84,    -83,     34 },
      {     62,    -90,     69,    -10,    -55,     89,    -75,     20,     46,    -87,     80,    -29,    -37,     84,    -84,     38,     28,    -79,     87,    -47,    -19,     74,    -89,     55,      9,    -68,     90,    -63,      1,     61,    -90,     70,    -11,    -54,     89,    -76,     21,     45,    -87,     81,    -30,    -36,     83,    -85,     39,     27,    -79,     88,    -48,    -18,     74,    -89,     56,      8,    -67,     90,    -64,      2,     60,    -90,     70,    -12,    -53,     89,    -76,     22,     44,    -86,     81,    -31,    -35,     83,    -85,     40,     26,    -78,     88,    -49,    -16,     73,    -90,     57,      7,    -67,     90,    -64,      3,     60,    -90,     71,    -13,    -52,     89,    -77,     23,     43,    -86,     81,    -32,    -34,     82,    -85,     41,     25,    -78,     88,    -50,    -15,     72,    -90,     58,      6,    -66,     90,    -65,      4,     59,    -90,     72,    -14,    -51,     88,    -77,     24,     42,    -86,     82,    -33 },
      {     60,    -90,     73,    -19,    -45,     86,    -82,     36,     28,    -78,     88,    -53,    -10,     67,    -90,     67,     -9,    -54,     89,    -78,     27,     37,    -83,     86,    -44,    -20,     74,    -90,     60,      1,    -61,     90,    -72,     18,     46,    -86,     82,    -35,    -29,     79,    -88,     52,     11,    -68,     90,    -66,      8,     55,    -89,     77,    -26,    -38,     83,    -85,     43,     21,    -74,     90,    -59,     -2,     62,    -90,     72,    -16,    -47,     87,    -81,     34,     30,    -79,     88,    -51,    -12,     69,    -90,     65,     -7,    -55,     89,    -77,     25,     39,    -84,     85,    -42,    -22,     75,    -89,     58,      3,    -63,     90,    -71,     15,     48,    -87,     81,    -33,    -31,     80,    -88,     50,     13,    -70,     90,    -64,      6,     56,    -89,     76,    -24,    -40,     84,    -85,     41,     23,    -76,     89,    -57,     -4,     64,    -90,     70,    -14,    -49,     87,    -81,     32 },
      {     59,    -89,     77,    -27,    -35,     81,    -88,     52,      9,    -65,     90,    -72,     19,     43,    -85,     85,    -44,    -18,     71,    -90,     66,    -10,    -51,     87,    -81,     36,     26,    -76,     89,    -60,      1,     58,    -89,     77,    -28,    -34,     81,    -88,     53,      8,    -64,     90,    -72,     20,     42,    -84,     85,    -45,    -16,     70,    -90,     67,    -11,    -50,     87,    -82,     37,     25,    -76,     90,    -60,      2,     57,    -89,     78,    -29,    -33,     80,    -88,     54,      7,    -64,     90,    -73,     21,     41,    -84,     86,    -46,    -15,     70,    -90,     67,    -12,    -49,     87,    -82,     38,     24,    -75,     90,    -61,      3,     56,    -89,     78,    -30,    -32,     79,    -88,     55,      6,    -63,     90,    -74,     22,     40,    -83,     86,    -47,    -14,     69,    -90,     68,    -13,    -48,     86,    -83,     39,     23,    -74,     90,    -62,      4,     55,    -89,     79,    -31 },
      {     57,    -89,     80,    -35,    -25,     74,    -90,     65,    -11,    -48,     86,    -85,     45,     14,    -67,     90,    -72,     22,     38,    -81,     88,    -55,     -3,     60,    -89,     78,    -32,    -28,     76,    -90,     63,     -8,    -51,     87,    -83,     42,     18,    -70,     90,    -70,     19,     41,    -83,     87,    -52,     -7,     62,    -90,     77,    -29,    -31,     78,    -89,     60,     -4,    -54,     88,    -82,     39,     21,    -72,     90,    -68,     15,     44,    -84,     86,    -49,    -10,     64,    -90,     75,    -26,    -34,     79,    -89,     58,     -1,    -56,     88,    -81,     36,     24,    -74,     90,    -66,     12,     47,    -85,     85,    -46,    -13,     67,    -90,     73,    -23,    -37,     81,    -88,     55,      2,    -59,     89,    -79,     33,     27,    -76,     90,    -64,      9,     50,    -86,     84,    -43,    -16,     69,    -90,     71,    -20,    -40,     82,    -87,     53,      6,    -61,     89,    -77,     30 },
      {     55,    -88,     83,    -43,    -14,     66,    -90,     76,    -30,    -28,     75,    -90,     67,    -16,    -41,     82,    -88,     57,     -2,    -54,     87,    -84,     45,     12,    -64,     90,    -77,     32,     26,    -74,     90,    -69,     19,     39,    -81,     89,    -59,      4,     52,    -86,     85,    -47,    -10,     63,    -89,     78,    -34,    -24,     72,    -90,     70,    -21,    -37,     80,    -89,     60,     -7,    -50,     86,    -85,     49,      8,    -61,     89,    -79,     36,     22,    -71,     90,    -72,     23,     35,    -79,     89,    -62,      9,     48,    -85,     86,    -51,     -6,     60,    -89,     81,    -38,    -20,     70,    -90,     73,    -25,    -33,     78,    -90,     64,    -11,    -46,     84,    -87,     53,      3,    -58,     88,    -81,     40,     18,    -68,     90,    -74,     27,     31,    -77,     90,    -65,     13,     44,    -83,     87,    -55,     -1,     56,    -88,     82,    -42,    -15,     67,    -90,     76,    -29 },
      {     54,    -86,     85,    -51,     -3,     56,    -87,     84,    -48,     -7,     59,    -88,     83,    -45,    -10,     61,    -89,     81,    -42,    -13,     64,    -89,     80,    -39,    -16,     66,    -90,     78,    -36,    -20,     68,    -90,     77,    -33,    -23,     70,    -90,     75,    -30,    -26,     72,    -90,     73,    -27,    -29,     74,    -90,     71,    -24,    -32,     76,    -90,     69,    -21,    -35,     78,    -90,     67,    -18,    -38,     79,    -89,     64,    -14,    -41,     81,    -89,     62,    -11,    -44,     82,    -88,     60,     -8,    -47,     84,    -88,     57,     -4,    -50,     85,    -87,     55,     -1,    -53,     86,    -86,     52,      2,    -55,     87,    -85,     49,      6,    -58,     88,    -83,     46,      9,    -60,     89,    -82,     43,     12,    -63,     89,    -81,     40,     15,    -65,     90,    -79,     37,     19,    -67,     90,    -77,     34,     22,    -70,     90,    -76,     31,     25,    -72,     90,    -74,     28 },
      {     52,    -85,     87,    -58,      8,     45,    -82,     89,    -64,     15,     38,    -78,     90,    -69,     23,     31,    -74,     90,    -74,     30,     24,    -70,     90,    -78,     37,     16,    -64,     89,    -81,     44,      9,    -59,     88,    -85,     51,      1,    -53,     85,    -87,     57,     -7,    -46,     82,    -89,     63,    -14,    -39,     79,    -90,     68,    -22,    -32,     75,    -90,     73,    -29,    -25,     70,    -90,     77,    -36,    -18,     65,    -89,     81,    -43,    -10,     60,    -88,     84,    -50,     -2,     54,    -86,     87,    -56,      6,     47,    -83,     89,    -62,     13,     40,    -79,     90,    -67,     21,     33,    -76,     90,    -72,     28,     26,    -71,     90,    -77,     35,     19,    -66,     89,    -81,     42,     11,    -60,     88,    -84,     49,      3,    -55,     86,    -86,     55,     -4,    -48,     83,    -88,     61,    -12,    -41,     80,    -90,     67,    -20,    -34,     76,    -90,     72,    -27 },
      {     50,    -83,     89,    -64,     19,     33,    -74,     90,    -76,     36,     15,    -62,     88,    -85,     53,     -3,    -47,     82,    -89,     67,    -22,    -30,     72,    -90,     78,    -39,    -12,     60,    -87,     86,    -55,      7,     44,    -81,     90,    -69,     25,     27,    -70,     90,    -79,     42,      9,    -57,     86,    -87,     58,    -10,    -41,     79,    -90,     71,    -28,    -24,     68,    -90,     81,    -45,     -6,     55,    -85,     88,    -60,     13,     38,    -77,     90,    -73,     31,     21,    -66,     89,    -82,     48,      2,    -52,     84,    -88,     63,    -16,    -35,     76,    -90,     75,    -34,    -18,     64,    -89,     84,    -51,      1,     49,    -83,     89,    -65,     20,     32,    -74,     90,    -77,     37,     14,    -61,     88,    -85,     54,     -4,    -46,     81,    -89,     67,    -23,    -29,     72,    -90,     78,    -40,    -11,     59,    -87,     86,    -56,      8,     43,    -80,     90,    -70,     26 },
      {     48,    -81,     90,    -70,     29,     21,    -64,     88,    -85,     55,     -9,    -40,     77,    -90,     76,    -37,    -12,     58,    -86,     88,    -62,     18,     32,    -72,     90,    -80,     45,      3,    -51,     83,    -89,     68,    -26,    -24,     67,    -89,     84,    -53,      6,     43,    -79,     90,    -74,     34,     15,    -60,     87,    -87,     60,    -14,    -35,     74,    -90,     78,    -42,     -7,     54,    -84,     89,    -66,     23,     27,    -69,     89,    -82,     50,     -2,    -46,     81,    -90,     72,    -31,    -19,     63,    -88,     86,    -57,     11,     38,    -76,     90,    -77,     39,     10,    -56,     85,    -88,     64,    -20,    -30,     71,    -90,     81,    -47,     -1,     49,    -82,     90,    -70,     28,     22,    -65,     89,    -85,     55,     -8,    -41,     78,    -90,     75,    -36,    -13,     59,    -86,     87,    -61,     16,     33,    -73,     90,    -79,     44,      4,    -52,     83,    -89,     67,    -25 },
      {     46,    -79,     90,    -76,     39,      8,    -53,     83,    -90,     71,    -32,    -15,     59,    -86,     88,    -66,     25,     23,    -64,     88,    -86,     60,    -18,    -30,     70,    -89,     84,    -55,     10,     37,    -74,     90,    -81,     48,     -2,    -44,     78,    -90,     77,    -41,     -6,     51,    -82,     90,    -72,     34,     13,    -57,     85,    -89,     67,    -27,    -21,     63,    -87,     87,    -62,     20,     28,    -68,     89,    -85,     56,    -12,    -35,     73,    -90,     81,    -50,      4,     42,    -77,     90,    -78,     43,      3,    -49,     81,    -90,     74,    -36,    -11,     55,    -84,     89,    -69,     29,     19,    -61,     87,    -88,     64,    -22,    -26,     67,    -89,     85,    -58,     14,     33,    -72,     90,    -82,     52,     -7,    -40,     76,    -90,     79,    -45,     -1,     47,    -80,     90,    -75,     38,      9,    -54,     83,    -89,     70,    -31,    -16,     60,    -86,     88,    -65,     24 },
      {     44,    -77,     90,    -80,     49,     -6,    -39,     74,    -90,     82,    -54,     11,     34,    -71,     89,    -85,     58,    -16,    -29,     67,    -88,     86,    -62,     22,     24,    -64,     87,    -88,     66,    -27,    -19,     60,    -85,     89,    -70,     32,     13,    -55,     83,    -90,     73,    -37,     -8,     51,    -81,     90,    -76,     42,      2,    -46,     78,    -90,     79,    -47,      3,     41,    -76,     90,    -81,     52,     -9,    -36,     72,    -90,     84,    -56,     14,     31,    -69,     89,    -86,     60,    -20,    -26,     65,    -88,     87,    -64,     25,     21,    -61,     86,    -89,     68,    -30,    -15,     57,    -84,     89,    -72,     35,     10,    -53,     82,    -90,     75,    -40,     -4,     48,    -79,     90,    -78,     45,     -1,    -43,     77,    -90,     81,    -50,      7,     38,    -74,     90,    -83,     55,    -12,    -33,     70,    -89,     85,    -59,     18,     28,    -67,     88,    -87,     63,    -23 },
      {     42,    -75,     90,    -84,     58,    -19,    -25,     63,    -86,     89,    -71,     36,      7,    -48,     78,    -90,     81,    -53,     12,     31,    -67,     88,    -88,     67,    -30,    -13,     54,    -81,     90,    -78,     47,     -6,    -37,     72,    -89,     86,    -62,     24,     20,    -59,     84,    -90,     74,    -41,     -1,     43,    -76,     90,    -83,     57,    -18,    -26,     64,    -86,     89,    -70,     35,      8,    -49,     79,    -90,     81,    -52,     11,     32,    -68,     88,    -87,     66,    -29,    -14,     55,    -82,     90,    -77,     46,     -4,    -38,     72,    -89,     85,    -61,     23,     21,    -60,     85,    -90,     74,    -40,     -2,     44,    -76,     90,    -83,     56,    -16,    -27,     64,    -87,     89,    -70,     34,      9,    -50,     79,    -90,     80,    -51,     10,     33,    -69,     88,    -87,     65,    -28,    -15,     55,    -82,     90,    -77,     45,     -3,    -39,     73,    -89,     85,    -60,     22 },
      {     40,    -72,     89,    -87,     66,    -31,    -10,     49,    -78,     90,    -83,     59,    -22,    -20,     57,    -82,     90,    -79,     51,    -12,    -29,     64,    -86,     89,    -74,     42,     -2,    -38,     71,    -89,     87,    -67,     33,      8,    -47,     77,    -90,     84,    -60,     24,     18,    -55,     81,    -90,     80,    -53,     14,     27,    -63,     85,    -90,     75,    -44,      4,     36,    -70,     88,    -88,     69,    -35,     -6,     45,    -76,     90,    -85,     62,    -26,    -15,     54,    -81,     90,    -81,     55,    -16,    -25,     61,    -85,     90,    -76,     46,     -7,    -34,     68,    -88,     88,    -70,     37,      3,    -43,     74,    -89,     86,    -64,     28,     13,    -52,     79,    -90,     82,    -56,     19,     23,    -60,     84,    -90,     77,    -48,      9,     32,    -67,     87,    -89,     72,    -39,     -1,     41,    -73,     89,    -86,     65,    -30,    -11,     50,    -78,     90,    -83,     58,    -21 },
      {     38,    -70,     88,    -89,     73,    -43,      6,     33,    -66,     86,    -90,     76,    -48,     11,     28,    -62,     84,    -90,     79,    -53,     16,     23,    -58,     82,    -90,     81,    -57,     22,     18,    -54,     79,    -90,     84,    -61,     27,     12,    -49,     77,    -90,     86,    -65,     32,      7,    -44,     74,    -89,     87,    -69,     37,      1,    -39,     70,    -88,     89,    -72,     42,     -4,    -34,     67,    -86,     89,    -76,     47,    -10,    -29,     63,    -85,     90,    -78,     52,    -15,    -24,     59,    -82,     90,    -81,     56,    -21,    -19,     55,    -80,     90,    -83,     60,    -26,    -13,     50,    -77,     90,    -85,     64,    -31,     -8,     45,    -74,     89,    -87,     68,    -36,     -2,     40,    -71,     88,    -88,     72,    -41,      3,     35,    -67,     87,    -89,     75,    -46,      9,     30,    -64,     85,    -90,     78,    -51,     14,     25,    -60,     83,    -90,     81,    -55,     20 },
      {     36,    -67,     86,    -90,     79,    -55,     21,     16,    -51,     77,    -89,     87,    -70,     40,     -4,    -32,     64,    -84,     90,    -81,     58,    -25,    -12,     47,    -74,     89,    -88,     72,    -44,      9,     28,    -60,     82,    -90,     83,    -61,     29,      8,    -43,     72,    -88,     89,    -75,     48,    -13,    -24,     57,    -81,     90,    -85,     64,    -33,     -3,     39,    -69,     87,    -90,     77,    -52,     18,     20,    -54,     78,    -90,     86,    -67,     37,     -1,    -35,     66,    -85,     90,    -79,     55,    -22,    -15,     50,    -76,     89,    -87,     70,    -41,      6,     31,    -63,     84,    -90,     81,    -59,     26,     11,    -46,     74,    -89,     88,    -73,     45,    -10,    -27,     60,    -82,     90,    -83,     62,    -30,     -7,     42,    -71,     88,    -89,     76,    -49,     14,     23,    -56,     80,    -90,     85,    -65,     34,      2,    -38,     68,    -86,     90,    -78,     53,    -19 },
      {     34,    -64,     83,    -90,     84,    -64,     35,     -1,    -33,     63,    -83,     90,    -84,     65,    -36,      2,     32,    -62,     82,    -90,     85,    -66,     37,     -3,    -31,     61,    -82,     90,    -85,     67,    -38,      4,     30,    -60,     81,    -90,     85,    -67,     39,     -6,    -29,     60,    -81,     90,    -86,     68,    -40,      7,     28,    -59,     81,    -90,     86,    -69,     41,     -8,    -27,     58,    -80,     90,    -86,     70,    -42,      9,     26,    -57,     79,    -90,     87,    -70,     43,    -10,    -25,     56,    -79,     90,    -87,     71,    -44,     11,     24,    -55,     78,    -90,     87,    -72,     45,    -12,    -23,     55,    -78,     89,    -88,     72,    -46,     13,     22,    -54,     77,    -89,     88,    -73,     47,    -14,    -21,     53,    -77,     89,    -88,     74,    -48,     15,     20,    -52,     76,    -89,     88,    -74,     49,    -16,    -19,     51,    -76,     89,    -89,     75,    -50,     18 },
      {     32,    -60,     81,    -90,     87,    -73,     49,    -19,    -14,     45,    -70,     86,    -90,     82,    -64,     36,     -4,    -28,     57,    -78,     89,    -88,     76,    -53,     23,     10,    -41,     67,    -85,     90,    -84,     67,    -40,      9,     24,    -54,     76,    -89,     89,    -78,     56,    -27,     -6,     37,    -64,     83,    -90,     86,    -70,     44,    -13,    -20,     50,    -74,     88,    -90,     80,    -60,     31,      1,    -33,     61,    -81,     90,    -87,     72,    -48,     18,     15,    -46,     71,    -86,     90,    -82,     63,    -35,      3,     29,    -58,     79,    -89,     88,    -75,     52,    -22,    -11,     42,    -68,     85,    -90,     84,    -66,     39,     -8,    -25,     55,    -77,     89,    -89,     77,    -55,     26,      7,    -38,     65,    -83,     90,    -85,     69,    -43,     12,     21,    -51,     74,    -88,     90,    -79,     59,    -30,     -2,     34,    -62,     81,    -90,     87,    -72,     47,    -16 },
      {     30,    -57,     77,    -89,     89,    -80,     61,    -35,      6,     25,    -53,     74,    -87,     90,    -82,     65,    -40,     11,     20,    -48,     71,    -86,     90,    -85,     69,    -45,     16,     14,    -43,     67,    -84,     90,    -86,     72,    -50,     22,      9,    -38,     64,    -81,     90,    -88,     76,    -55,     27,      3,    -33,     60,    -79,     89,    -89,     78,    -59,     32,     -2,    -28,     55,    -76,     88,    -90,     81,    -63,     37,     -8,    -23,     51,    -73,     87,    -90,     83,    -67,     42,    -13,    -18,     46,    -70,     85,    -90,     85,    -70,     47,    -19,    -12,     41,    -66,     83,    -90,     87,    -74,     52,    -24,     -7,     36,    -62,     81,    -90,     88,    -77,     56,    -29,     -1,     31,    -58,     78,    -89,     89,    -79,     60,    -34,      4,     26,    -54,     75,    -88,     90,    -82,     64,    -39,     10,     21,    -49,     72,    -86,     90,    -84,     68,    -44,     15 },
      {     28,    -54,     74,    -86,     90,    -85,     72,    -51,     25,      3,    -31,     56,    -76,     87,    -90,     84,    -70,     48,    -22,     -7,     34,    -59,     77,    -88,     90,    -83,     67,    -45,     19,     10,    -37,     61,    -79,     89,    -90,     81,    -65,     42,    -15,    -13,     40,    -64,     81,    -89,     89,    -80,     63,    -39,     12,     16,    -43,     66,    -82,     90,    -89,     78,    -60,     36,     -9,    -20,     46,    -68,     83,    -90,     88,    -77,     58,    -33,      6,     23,    -49,     70,    -85,     90,    -87,     75,    -55,     30,     -2,    -26,     52,    -72,     86,    -90,     86,    -73,     53,    -27,     -1,     29,    -55,     74,    -87,     90,    -85,     71,    -50,     24,      4,    -32,     57,    -76,     88,    -90,     84,    -69,     47,    -21,     -8,     35,    -60,     78,    -88,     90,    -82,     67,    -44,     18,     11,    -38,     62,    -79,     89,    -89,     81,    -64,     41,    -14 },
      {     26,    -50,     70,    -83,     90,    -89,     80,    -64,     43,    -19,     -8,     33,    -56,     74,    -86,     90,    -87,     76,    -59,     36,    -11,    -15,     40,    -62,     78,    -88,     90,    -85,     72,    -53,     29,     -3,    -23,     47,    -67,     82,    -89,     89,    -81,     67,    -46,     22,      4,    -30,     54,    -72,     85,    -90,     88,    -78,     61,    -39,     14,     12,    -37,     60,    -77,     87,    -90,     86,    -74,     55,    -32,      7,     20,    -44,     65,    -81,     89,    -90,     83,    -69,     49,    -25,     -1,     27,    -51,     70,    -84,     90,    -89,     79,    -64,     42,    -18,     -9,     34,    -57,     75,    -86,     90,    -87,     76,    -58,     35,    -10,    -16,     41,    -63,     79,    -88,     90,    -84,     71,    -52,     28,     -2,    -24,     48,    -68,     82,    -90,     89,    -81,     66,    -45,     21,      6,    -31,     55,    -73,     85,    -90,     88,    -77,     60,    -38,     13 },
      {     24,    -46,     65,    -79,     88,    -90,     86,    -76,     60,    -39,     16,      8,    -31,     53,    -70,     83,    -89,     90,    -83,     71,    -54,     32,     -9,    -15,     38,    -59,     75,    -86,     90,    -88,     80,    -66,     47,    -25,      1,     23,    -45,     64,    -79,     88,    -90,     86,    -76,     60,    -40,     18,      7,    -30,     52,    -70,     82,    -89,     90,    -84,     72,    -55,     33,    -10,    -14,     37,    -58,     74,    -85,     90,    -89,     81,    -67,     48,    -26,      2,     22,    -44,     64,    -78,     88,    -90,     87,    -77,     61,    -41,     19,      6,    -29,     51,    -69,     82,    -89,     90,    -84,     72,    -55,     34,    -11,    -13,     36,    -57,     74,    -85,     90,    -89,     81,    -67,     49,    -27,      3,     21,    -43,     63,    -78,     87,    -90,     87,    -77,     62,    -42,     20,      4,    -28,     50,    -68,     81,    -89,     90,    -85,     73,    -56,     35,    -12 },
      {     22,    -42,     60,    -75,     85,    -90,     89,    -84,     73,    -58,     39,    -19,     -3,     25,    -45,     63,    -77,     86,    -90,     89,    -82,     71,    -55,     36,    -15,     -7,     28,    -48,     65,    -78,     87,    -90,     88,    -81,     69,    -53,     33,    -12,    -10,     31,    -51,     67,    -80,     88,    -90,     88,    -79,     67,    -50,     30,     -9,    -13,     34,    -54,     70,    -81,     89,    -90,     87,    -78,     64,    -47,     27,     -6,    -16,     37,    -56,     72,    -83,     89,    -90,     86,    -76,     62,    -44,     24,     -2,    -20,     40,    -59,     74,    -84,     90,    -90,     85,    -74,     60,    -41,     21,      1,    -23,     43,    -61,     76,    -85,     90,    -89,     83,    -72,     57,    -38,     18,      4,    -26,     46,    -64,     77,    -86,     90,    -89,     82,    -70,     55,    -35,     14,      8,    -29,     49,    -66,     79,    -87,     90,    -88,     81,    -68,     52,    -32,     11 },
      {     20,    -38,     55,    -70,     81,    -88,     90,    -89,     83,    -73,     60,    -43,     25,     -6,    -14,     33,    -51,     66,    -78,     86,    -90,     90,    -85,     76,    -64,     48,    -30,     11,      9,    -28,     46,    -62,     75,    -84,     89,    -90,     87,    -79,     67,    -53,     35,    -16,     -3,     23,    -41,     58,    -72,     82,    -88,     90,    -88,     81,    -71,     57,    -40,     22,     -2,    -18,     36,    -54,     68,    -79,     87,    -90,     89,    -84,     74,    -61,     45,    -27,      8,     12,    -31,     49,    -64,     77,    -85,     90,    -90,     86,    -77,     65,    -50,     32,    -13,     -7,     26,    -44,     60,    -74,     83,    -89,     90,    -87,     80,    -69,     55,    -37,     19,      1,    -21,     39,    -56,     70,    -81,     88,    -90,     89,    -82,     72,    -59,     42,    -24,      4,     15,    -34,     52,    -67,     78,    -86,     90,    -89,     85,    -76,     63,    -47,     29,    -10 },
      {     18,    -34,     50,    -64,     75,    -83,     89,    -90,     89,    -84,     76,    -64,     51,    -35,     19,     -1,    -16,     33,    -49,     63,    -74,     83,    -88,     90,    -89,     84,    -76,     65,    -52,     36,    -20,      2,     15,    -32,     48,    -62,     74,    -82,     88,    -90,     89,    -85,     77,    -66,     53,    -37,     21,     -3,    -14,     31,    -47,     61,    -73,     82,    -88,     90,    -89,     85,    -77,     67,    -54,     38,    -22,      4,     13,    -30,     46,    -60,     72,    -81,     88,    -90,     89,    -85,     78,    -67,     55,    -39,     23,     -6,    -12,     29,    -45,     60,    -72,     81,    -87,     90,    -90,     86,    -78,     68,    -55,     40,    -24,      7,     11,    -28,     44,    -59,     71,    -81,     87,    -90,     90,    -86,     79,    -69,     56,    -41,     25,     -8,    -10,     27,    -43,     58,    -70,     80,    -87,     90,    -90,     86,    -79,     70,    -57,     42,    -26,      9 },
      {     15,    -30,     44,    -57,     68,    -77,     84,    -89,     90,    -89,     86,    -80,     72,    -61,     49,    -35,     21,     -6,    -10,     25,    -39,     53,    -64,     74,    -82,     87,    -90,     90,    -88,     82,    -75,     65,    -54,     40,    -26,     11,      4,    -20,     34,    -48,     60,    -71,     79,    -86,     89,    -90,     89,    -85,     78,    -69,     58,    -45,     31,    -16,      1,     14,    -29,     43,    -56,     67,    -77,     84,    -88,     90,    -90,     86,    -81,     72,    -62,     50,    -36,     22,     -7,     -9,     24,    -38,     52,    -64,     74,    -81,     87,    -90,     90,    -88,     83,    -76,     66,    -55,     41,    -27,     12,      3,    -19,     33,    -47,     60,    -70,     79,    -85,     89,    -90,     89,    -85,     78,    -70,     59,    -46,     32,    -18,      2,     13,    -28,     42,    -55,     67,    -76,     83,    -88,     90,    -90,     87,    -81,     73,    -63,     51,    -37,     23,     -8 },
      {     13,    -26,     38,    -50,     60,    -70,     77,    -83,     88,    -90,     90,    -89,     85,    -80,     73,    -64,     55,    -43,     31,    -19,      6,      8,    -21,     33,    -45,     56,    -66,     74,    -81,     86,    -89,     90,    -90,     87,    -82,     76,    -68,     59,    -48,     36,    -24,     11,      2,    -15,     28,    -40,     52,    -62,     71,    -78,     84,    -88,     90,    -90,     88,    -85,     79,    -72,     63,    -53,     41,    -29,     16,     -3,    -10,     23,    -35,     47,    -58,     67,    -76,     82,    -87,     89,    -90,     89,    -86,     81,    -75,     67,    -57,     46,    -34,     22,     -9,     -4,     18,    -30,     42,    -54,     64,    -72,     79,    -85,     89,    -90,     90,    -88,     84,    -78,     70,    -61,     51,    -39,     27,    -14,      1,     12,    -25,     37,    -49,     60,    -69,     77,    -83,     87,    -90,     90,    -89,     86,    -81,     74,    -65,     55,    -44,     32,    -20,      7 },
      {     11,    -22,     32,    -42,     52,    -60,     68,    -75,     81,    -85,     88,    -90,     90,    -89,     87,    -84,     79,    -73,     66,    -58,     49,    -39,     29,    -19,      8,      3,    -14,     25,    -35,     45,    -55,     63,    -70,     77,    -82,     86,    -89,     90,    -90,     89,    -86,     82,    -77,     71,    -64,     55,    -46,     36,    -26,     15,     -4,     -7,     18,    -28,     38,    -48,     57,    -65,     72,    -78,     83,    -87,     89,    -90,     90,    -88,     85,    -81,     76,    -69,     61,    -53,     43,    -33,     23,    -12,      1,     10,    -21,     31,    -41,     51,    -60,     67,    -74,     80,    -85,     88,    -90,     90,    -90,     88,    -84,     79,    -74,     67,    -59,     50,    -40,     30,    -20,      9,      2,    -13,     24,    -34,     44,    -54,     62,    -70,     76,    -81,     86,    -89,     90,    -90,     89,    -87,     83,    -78,     72,    -64,     56,    -47,     37,    -27,     16,     -6 },
      {      9,    -18,     26,    -34,     42,    -50,     57,    -64,     70,    -75,     79,    -83,     86,    -89,     90,    -90,     90,    -89,     87,    -84,     80,    -76,     70,    -64,     58,    -51,     43,    -35,     27,    -19,     10,     -1,     -8,     16,    -25,     33,    -41,     49,    -56,     63,    -69,     74,    -79,     83,    -86,     88,    -90,     90,    -90,     89,    -87,     84,    -81,     76,    -71,     65,    -59,     52,    -44,     36,    -28,     20,    -11,      2,      7,    -15,     24,    -32,     40,    -48,     55,    -62,     68,    -74,     78,    -82,     86,    -88,     90,    -90,     90,    -89,     87,    -85,     81,    -77,     72,    -66,     60,    -53,     45,    -37,     29,    -21,     12,     -3,     -6,     14,    -23,     31,    -39,     47,    -55,     61,    -67,     73,    -78,     82,    -85,     88,    -89,     90,    -90,     89,    -88,     85,    -81,     77,    -72,     67,    -60,     54,    -46,     38,    -30,     22,    -13,      4 },
      {      7,    -13,     20,    -26,     32,    -38,     44,    -50,     55,    -60,     65,    -70,     74,    -77,     81,    -83,     86,    -88,     89,    -90,     90,    -90,     90,    -89,     87,    -85,     83,    -80,     77,    -73,     69,    -64,     60,    -55,     49,    -43,     37,    -31,     25,    -19,     12,     -6,     -1,      8,    -14,     21,    -27,     33,    -39,     45,    -51,     56,    -61,     66,    -70,     74,    -78,     81,    -84,     86,    -88,     89,    -90,     90,    -90,     90,    -89,     87,    -85,     82,    -79,     76,    -72,     68,    -64,     59,    -54,     48,    -42,     36,    -30,     24,    -18,     11,     -4,     -2,      9,    -15,     22,    -28,     34,    -40,     46,    -52,     57,    -62,     67,    -71,     75,    -78,     81,    -84,     86,    -88,     89,    -90,     90,    -90,     89,    -88,     87,    -85,     82,    -79,     76,    -72,     67,    -63,     58,    -53,     47,    -41,     35,    -29,     23,    -16,     10,     -3 },
      {      4,     -9,     13,    -18,     22,    -26,     30,    -34,     38,    -42,     46,    -50,     54,    -57,     60,    -64,     67,    -70,     72,    -75,     77,    -79,     81,    -83,     85,    -86,     88,    -89,     89,    -90,     90,    -90,     90,    -90,     89,    -89,     88,    -87,     85,    -84,     82,    -80,     78,    -76,     73,    -70,     67,    -64,     61,    -58,     55,    -51,     47,    -43,     39,    -35,     31,    -27,     23,    -19,     14,    -10,      6,     -1,     -3,      8,    -12,     16,    -21,     25,    -29,     33,    -37,     41,    -45,     49,    -53,     56,    -60,     63,    -66,     69,    -72,     74,    -77,     79,    -81,     83,    -85,     86,    -87,     88,    -89,     90,    -90,     90,    -90,     90,    -90,     89,    -88,     87,    -86,     84,    -82,     81,    -78,     76,    -74,     71,    -68,     65,    -62,     59,    -55,     52,    -48,     44,    -40,     36,    -32,     28,    -24,     20,    -15,     11,     -7,      2 },
      {      2,     -4,      7,     -9,     11,    -13,     15,    -18,     20,    -22,     24,    -26,     28,    -30,     32,    -34,     36,    -38,     40,    -42,     44,    -46,     48,    -50,     52,    -54,     55,    -57,     59,    -60,     62,    -64,     65,    -67,     68,    -70,     71,    -72,     74,    -75,     76,    -77,     78,    -79,     81,    -81,     82,    -83,     84,    -85,     86,    -86,     87,    -88,     88,    -89,     89,    -89,     90,    -90,     90,    -90,     90,    -90,     90,    -90,     90,    -90,     90,    -89,     89,    -89,     88,    -88,     87,    -87,     86,    -85,     85,    -84,     83,    -82,     81,    -80,     79,    -78,     77,    -76,     74,    -73,     72,    -70,     69,    -67,     66,    -64,     63,    -61,     60,    -58,     56,    -55,     53,    -51,     49,    -47,     45,    -43,     41,    -39,     37,    -35,     33,    -31,     29,    -27,     25,    -23,     21,    -19,     16,    -14,     12,    -10,      8,     -6,      3,     -1 },
    },
};
#endif
#endif
//--------------------------------------------------------------------------------------------------
